# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/Metrocluster.pm#30 $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking ProhibitFixedStringMatches)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)
##no critic (Modules::ProhibitMultiplePackages)
##no critic (RequireBracedFileHandleWithPrint)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MetroCluster Stub MTask Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Metrocluster;

use strict;
use warnings;
use Carp;

use List::Util qw(first) ;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Cluster ();
use WaflMCC::MCC::DR_Group ();
use WaflMCC::MCC::Utils ();

use NACL::C::SystemNode ();
use NACL::C::Cluster ();

use NACL::C::LoggerMgwdLog;

# the _primary_keys_validate_spec definition after this use statement is an ugly
# workaround for the MCC test lib use of NACL::C::Metrocluster (see
# BURT 589544).  Until NACL::C::Metrocluster is fixed, we want to nullify
# the effects of the current NACL::C::Metrocluster::_primary_keys_validate_spec
# method.
use NACL::C::Metrocluster ();
use NACL::CS::MetroclusterInterconnect ();
package NACL::C::Metrocluster;

#for now, redefine this sub if it's there, otherwise ignore
if (defined &_primary_keys_validate_spec) {
    sub _primary_keys_validate_spec {

        # we are redefining the NACL::C::Metrocluster::_primary_keys_validate_spec
        # method, which looks like this in NACL/C/Metrocluster.pm
        #
        # sub _primary_keys_validate_spec {
        #     return (
        #         'cluster' => { type => SCALAR },
        #         'node'    => { type => SCALAR },
        #     );
        # } ## end sub _primary_keys_validate_spec

        # here we just say there are no cluster or node primary keys to validate
        # against for our NACL::C::Metrocluster components.
        return ();
    }
}
#
# and back to our regularly scheduled package...
package WaflMCC::MCC::Metrocluster;

use NACL::APISet::Version;

use WaflMCC::MCC::Exceptions::InvalidMethodUsage ();

use WaflMCC::MCC::Log ();

use NACL::STask::Fiji ();  #for menu use only...

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);

use Tharn qw($Log subtest);
use Subtest;
use NATE::STAF::Peer qw($Parent);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use Readonly;

use Data::Dumper;
use List::Util ( );
use List::MoreUtils ( );


=head1 NAME

WaflMCC::MCC::MetroCluster

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.

    # Start test
    use WaflMCC::MCC::MetroCluster;
    ...

    my $mcc = WaflMCC::MCC::MetroCluster->new(
                               nodes => $arrayref_of_node_names_in_mcc_system);
    ...

    # Perform test here
    ...

    # End test

=cut

=head1 METHODS

=head2 new(...)

    my $mcc = WaflMCC::MCC::MetroCluster->new(
                               nodes => $arrayref_of_node_names_in_mcc_system);

Creates a new WaflMCC::MCC::Metrocluster object.  It is important to
understand that a Metrocluster component object executes commands on a
particular site's cluster only, so the library must understand which site
the user wishes to execute Metrocluster commands upon.

See also WaflMCC::MCC::DR_Coordinator for a more concrete example of this
dynamic.

=over

=item Options

=over

=item C<< nodes => $one_or_more_node_names_in_the_mcc_system >>

(Required.) An arrayref containing one or more nodes in an MCC system.

First, an initial A1 node will be chosen randomly from amongst the nodes listed
in the nodes parameter.  This will define both what is considered to be "site A"
as well as the A1 node for the first DR Group in a given test.

After automatic DR Group discovery, if the MCC system is found to contain
multiple DR Groups, then the A1 node for each of the remaining DR Groups will
be chosen randomly from the nodes in this DR Group at site A which are listed
in the nodes parameter.  If none of these nodes are listed, a random node in
this DR Group at site A will be chosen as the A1 node for this DR Group.

If nodes contains two or more nodes which are not in the same MCC system, an
exception will be thrown.

=back

=item Associated ENVIRONMENT Variables

=over

=item C<< DISABLE_SHARED_STORAGE_NETWORK_VALIDATION = 0|1 >>

(Optional, and not recommended except for specific exceptional cases.)

If this environment variable is set to 1 in a test environment, then the
automatic discovery and validation of the shared storage network will not
be executed.  This is not recommended unless the testbed has specific and
understood issues which need to be worked around in the shared storage
network.  The validation of the shared network before any test is an important
part of understanding and triaging various failures that can occur.

A WARNING will occur when this environment variable is used.

[NB: If the test you are running has any reason to call the method

    $dr_group->get_network_topology()

then DISABLE_SHARED_STORAGE_NETWORK_VALIDATION cannot be set.]

=item C<< DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION = 0|1 >>

(Optional, and not recommended except for specific exceptional cases.)

If this environment variable is set to 1 in a test environment, then the
automatic disk state validation that occurs during initialization and also often
occurs by default after DR operations will be disabled unless explicitly called
for in the test.

In other words, if the user sets DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION=1
in their test, but then explicitly calls

    $shared_storage->validate_disk_states()

then such validation will occur at that point, but will revert back to disabled
disk state validations afterwards.

Use of this environment variable is not recommended.  Changes in disk state
are a major part of MCC DR lifecycle operations and should always be validated
against unless specific and understood issues in the disk states need to be
worked around to allow other testing.

A WARNING will occur when this environment variable is used.

See also: WARN_ONLY_ON_INVALID_PLEX_ASSIGNMENT

=item C<< WARN_ONLY_ON_INVALID_PLEX_ASSIGNMENT = 0|1 >>

(Optional, and should only be used in a very specific case as detailed below.)

Since invalid plex assignment issues may be part of an MCC system
setup, this environment variable is used to make sure all shared storage disk
validations report on, but do not fail the test, when such issues are noted.

In other words, this is a workaround to a specific issue that should be
resolved at some point (by the RAID team?)

A WARNING will occur when this environment variable is used.

See also: DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION

=back

=back

=cut

# Automatically generate the constructor (includes a call to 'init').  This is
# a public interface.
#
# Also creates non-public members and non-public accesors for them.


use Class::MethodMaker [
    new    => [ '-init', 'new' ],
     scalar => [
        { -type => 'NACL::C::Metrocluster' },
        qw/ _site_A_metrocluster_component
            _site_B_metrocluster_component /,
              ],
    scalar => [
        { -type => 'WaflMCC::MCC::Cluster' },
        qw/ _cluster_A
            _cluster_B /,
              ],
    scalar => [
        qw/ _dr_groups _ontap_build_version /,
              ],
                       ];

# implicitly called by Class::MethodMaker's new (see -init)
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # log volumous constructor related info to MCC_Constructor.log
    my $constructor_log = Tharn::param('LOGDIR') . '/MCC_Constructor.log';
    $Log->comment(
        "Logging MCC::Metrocluster object construction and initial " .
        "validation to $constructor_log...");

    my %opts = validate_with(
        params => \@_,
        spec   => {
            nodes      => { type  => ARRAYREF },
            pick_nodes => { type  => SCALAR, default => 1 },
                  },
                            );

    my $node_names = delete($opts{'nodes'});
    my $pick_nodes = delete($opts{'pick_nodes'});

    unless (scalar(@{ $node_names })) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                "Invalid MCC::Metrocluster constructor call.  At least one\n" .
                "    node name must be passed in the 'nodes' parameter.");
    }

    # this is a private package which is defined at the end of this file
    $Log->comment('Acquiring bootstrap cluster and node command interfaces...');
    my $bootstrap_util;
    
    if($pick_nodes) {
        $Log->redirect(path => $constructor_log);
        $bootstrap_util = WaflMCC::MCC::Metrocluster::_Bootstrap_Util->new(
                                                    node_names => $node_names,
                                                    pick_nodes => 1);    
    } else {
        $Log->redirect(path => $constructor_log);
        $bootstrap_util = WaflMCC::MCC::Metrocluster::_Bootstrap_Util->new(
                                                    node_names => $node_names,
                                                    pick_nodes => 0);    
    }     




    # these are needed in various inits early on
    $self->_site_A_metrocluster_component(
                         $bootstrap_util->get_site_A_Metrocluster_component());
    $self->_site_B_metrocluster_component(
                         $bootstrap_util->get_site_B_Metrocluster_component());
    $Log->reset_path();

    $Log->comment('...done');

    # can now check if system is in a good MCC state
    $Log->comment('Metrocluster system validation, part 1...');

    $Log->redirect(path => $constructor_log);
    $self->_validate_initial_metrocluster_state__part_1();
    $Log->reset_path();

    $Log->comment('...ok');

    # create cluster objects
    $Log->comment('Creating cluster harness objects...');

    $Log->redirect(path => $constructor_log);
    $self->_cluster_A(WaflMCC::MCC::Cluster->new(metrocluster => $self,
                                                     site         => 'A'));
    $self->_cluster_B(WaflMCC::MCC::Cluster->new(metrocluster => $self,
                                                     site         => 'B'));

    $Log->reset_path();

    $Log->comment('...done');

    # create and init DR Groups' objects
    $Log->comment('Creating DR Group harness objects...');

    $Log->redirect(path => $constructor_log);
    $self->_dr_groups($self->_create_DR_groups(
              bootstrap_util                        => $bootstrap_util,
              constructor_log_path                  => $constructor_log));
    $Log->reset_path();

    $Log->comment('...done');

    # now validate passed parameters.  throw exception if user passed in a
    # strange/invalid set of values in nodes.  (e.g., if there
    # are nodes listed not in the same MCC system, etc.)  this validation
    # occurs now because we now have all the DR Group discovery of nodes
    # and their partners done at this point.
    $Log->comment('Validating all user passed nodes...');

    $Log->redirect(path => $constructor_log);
    $self->_validate_node_names_passed_by_user(node_names => $node_names);
    $Log->reset_path();

    $Log->comment('...ok');

    # all container objects in the system have now been constructed.  do
    # final initial validations... :)
    $Log->comment('Validating initial disk states...');

    $Log->redirect(path => $constructor_log);
    my $disaster_status =
                 $self->L_get_disaster_status(known_live_sites => [ 'A', 'B' ],
                                              include_HA       => 1);
    foreach my $dr_group (@{ $self->_dr_groups() }) {
        $dr_group->get_shared_storage()->L_validate_disk_states(
                                          disaster_status => $disaster_status);
    }
    $Log->reset_path();

    $Log->comment('...ok');

    # initialize mcc_version member
    $self->_ontap_build_version($self->_determine_ontap_version_info());

    $Log->comment('Metrocluster system validation, part 2...');

    $Log->redirect(path => $constructor_log);
    # can now continue check if system is in a good MCC state
    $self->_validate_initial_metrocluster_state__part_2();
    $Log->reset_path();

    $Log->comment('...ok');

    $Log->comment('Metrocluster system validation, part 3...');

    my $four_pack = 0;
    foreach my $dr_group (@{ $self->_dr_groups() }) {
        if ($dr_group->is_4_pack() and !$dr_group->L_is_vsim_group()){
            $four_pack = 1;
            last;
        }
    }

    if ($four_pack){
        $Log->redirect(path => $constructor_log);
        # can now continue check if MCC interconnect is in good state
        $self->_validate_initial_metrocluster_state__part_3();
        $Log->reset_path();
    } else {
        $Log->comment('Skipped interconnect validation since it is not real'.
                      'MCC 4-pack(s) test bed');
    }
    $Log->comment('...ok');

    $Log->redirect(path => $constructor_log);
    $self->_enable_debug_level_logs();
    $Log->reset_path();

    $Log->comment("\n\nHello from MetroCluster!\n\n");

    # show basic MCC system information
    $self->_log_some_mcc_system_info(disaster_status => $disaster_status,
                                     constructor_log => $constructor_log);

    return;
}


# this is a private method to check if mcc is enabled and in normal DR state
sub _validate_initial_metrocluster_state__part_1 {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $handy_ci = $self->_site_A_metrocluster_component()->command_interface();

    my @mc_states = NACL::CS::Metrocluster->fetch(
                            command_interface => $handy_ci,
                            requested_fields  => [ 'configuration-state' ,
                                                   'dr-operation-state' ] );

    # check if metrocluster is enabled and in normal state
    foreach my $mc_state (@mc_states) {
        my $config_state = $mc_state->{'configuration_state'};
        my $dr_op_state  = $mc_state->{'dr_operation_state'};
        if ((($config_state ne 'enabled') and
             ($config_state ne 'Enabled') and
             ($config_state ne '-')
         ) or
# temporary workaround due to ONTAP bug
#                                    ($dr_op_state  ne 'DRSOM_FSM_ST_NORMAL')) {
#TBD: reinstate proper check after ONTAP fix (see BURT 556067.)

#TBD: reinstate proper check after ONTAP fix for 'dr_operation_state' field (BURT 571681)
                    ($dr_op_state  !~ /^(?:DRSOM_FSM_ST_NORMAL|enabled|-|Normal|"Switchback completed")$/)) {
            my $error_msg =
                "ERROR: Metrocluster not enabled or not in normal state.\n" .
                "    STATE   : $config_state\n" .
                "    DR_STATE: $dr_op_state\n";
            WaflMCC::MCC::Exceptions::BaseException->throw($error_msg);
        }
    }

    return;
}

# this is a private method to check if mcc is enabled and in normal DR state
sub _validate_initial_metrocluster_state__part_2 {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate node_scope DR_SOM states
    foreach my $site (qw/ A B /) {
        my $cluster = $self->get_cluster(site => $site);
        my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

        foreach my $mcc_node (@{ $sites_nodes }) {
            # this will throw an exception if there are any inconsistencies.
            $mcc_node->L_validate_DRSOM_FSM_state(dr_state => 'normal');
        }
    }

    # do a workaround of volume DSID issue (BURT 526567)
    #unless (WaflMCC::MCC::Utils->workaround_expired(
    #expiration_date => '2012/10/31', burt => '526567')) {
    #$self->workaround__prevent_volume_DSID_collisions();
    #}

    return;
}

# this is a private method to check mcc interconnect status
sub _validate_initial_metrocluster_state__part_3 {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # do nothing if interconnect validation is disabled.
    if ($ENV{DISABLE_MCC_INTERCONNECT_STATE_VALIDATION}) {
        my $msg = "DISABLE_MCC_INTERCONNECT_STATE_VALIDATION set!\n" .
                  '    MCC interconnect NOT validated!';
        $Log->warn($msg);
        return;
    }

    my $handy_ci = $self->_site_A_metrocluster_component()->command_interface();

    # metrocluster interconnect show is good enough to check interconnect status
    my @ic_states = NACL::CS::MetroclusterInterconnect->fetch(
                            command_interface => $handy_ci,
                            requested_fields  => [ 'type',
						   'node',
						   'partner-type',
                                                   'mirror-status',
                                                   'adapter',
                                                   'status' ] );
    $Log->comment('   Interconnect States: ' .
                  WaflMCC::MCC::Utils->dump(@ic_states));

    my %ics_by_node = ();

    # Group the interconnect states by node so we can consider
    # iwarps "up" if one of the interfaces is up
    foreach my $ic_state (@ic_states) {
        # All interconnect cares about is about HA and DR.
        # Hence, ignoring "aux".
	( lc($ic_state->{'partner_type'}) eq 'aux') and next;
	push( @{ $ics_by_node{ $ic_state->{'node'} } }, $ic_state );
    }


    foreach my $ics_by_node ( values %ics_by_node ) {
	my ($mirror, $interface) = (0, 0);

	# Throw an exception if the mirror status is offline
	if ( my $interconn =
	     first { lc($_->{'mirror_status'}) ne 'online' }
	     @{ $ics_by_node } ) {
	    my $err = $self->_build_interconnect_error($interconn);
	    WaflMCC::MCC::Exceptions::BaseException->throw($err);
	}


	# If we are iwarp, continue if any adapters are up
	if ( lc( $ics_by_node->[0]->{'type'} ) eq 'iwarp' ) {
	    if ( first { lc($_->{'status'}) eq 'up' }
		 @{ $ics_by_node }
	       ) {
		next;
	    }
	    else {
		my $err = $self->_build_interconnect_error( $ics_by_node->[0] );
		WaflMCC::MCC::Exceptions::BaseException->throw($err);
	    }
	}

	# Otherwise, throw an exception for any offline adapter
	else {
	    if ( my $interconnect = first { lc($_->{'status'}) ne 'up' }
		 @{ $ics_by_node }
	       ) {
		my $err = $self->_build_interconnect_error( $ics_by_node->[0] );
		WaflMCC::MCC::Exceptions::BaseException->throw($err);
	    }
	    else {
		next;
	    }
	}

    }
    return;
}


# Private method to build the error message for interconnect failures
sub _build_interconnect_error {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my ($interconn) = @_;

    my @errs;

    if ( lc($interconn->{'status'}) ne 'up' ) {
	push(@errs,
	     'interconnect status for adapter '
	     . "$interconn->{'adapter'} is $interconn->{'status'}");
    }

    if ( lc($interconn-{'mirror_status'}) ne 'online') {
	push(@errs,
	     'mirror status for adapter '
	     . " $interconn->{'adapter'} is "
	     . " $interconn->{'mirror_status'}");
    }

    push(@errs, "partner type is $interconn->{'partner_type'}");

    return ucfirst( join(', ', @errs) );

}

#private method to enable debug level logging
sub _enable_debug_level_logs {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    unless (Tharn::param('DO_NOT_ENABLE_MCC_DEBUG_LOGS')) {

        $Log->comment('Enabling mcc debug logging ' .
                      '(disable with DO_NOT_ENABLE_MCC_DEBUG_LOGS=1) ...');

        # enable debug level logging for all nodes
        foreach my $site (qw/ A B /) {
            my $cluster = $self->get_cluster(site => $site);
            my $sites_nodes = $cluster->L_get_cluster_mcc_nodes();

            foreach my $mcc_node (@{ $sites_nodes }) {
                my $ci = $mcc_node->get_Node_component()->command_interface();
                foreach my $module qw (tables:metro* mcc* metro* tables:mcc*) {
                    try {
                        my $logger = NACL::C::LoggerMgwdLog->new(
                            command_interface => $ci,
                            module            => $module,
                        );
                        $logger->modify('level' => 'debug');
                    } catch NATE::BaseException with {
                            $Log->comment(
                                "Error setting debug level logging, ignoring:" .
                                shift->text()
                            );
                    }
                }
            }
        }
        $Log->comment('done...');
    } else {
        $Log->comment('NOT enabling mcc debug logging...');
    }

    return;
}

=head2 workaround__prevent_volume_DSID_collisions()

    # can be called on an instance . . .
    $mcc->workaround__prevent_volume_DSID_collisions();

    # or on the class by providing a node name from each cluster
    WaflMCC::MCC::Metrocluster->workaround__prevent_volume_DSID_collisions(
                   all_nodes => [ 'dpg-mcc-3270-01-a2', 'dpg-mcc-3270-01-b1',
                                  'dpg-mcc-3270-01-a1', 'dpg-mcc-3270-01-b2']);

Notes:

workaround of volume DSID issue (BURT 526567)
see test lib BURT 615343 for an explanation of the logic in this temp method

most of the complexity of this method is so we don't change anything that
doesn't need to be changed.


MSID notes:

now also adding a quickie fix for MSID ranges.  when DSID ranges are being set
as opposed to just being checked (because this method can be called at any
time), the MSID ranges will be set to the same ranges being used for the DSIDs
on that cluster with the exception that the highest bit for MSIDs will be 1 as
opposed to 0 for DSIDs (32 bit values.)  E.g., DSID = 0x00 010 00 00 would be
matched with MSID = 0x10 01 00 00.

new on 9/26/2012: must allow that volumes that are mirrored are allowed to
have the same DSIDs across sites.  We will take a shortcut here and say that
if the CRS_Vserver is found, then DSID ranges have already been applied, and
just to make sure, we will look at the vol DSIDs for the CRS_Vserver at each
site and compare them.  If they are in the different ranges, we will assume
all is well and taken care of in DSID range settings across the clusters.

=cut

use NACL::CS::Volume;    # remove when workaround removed.
sub workaround__prevent_volume_DSID_collisions {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $object_or_pkg = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            all_nodes => { type => ARRAYREF, optional => 1, },
                  },
                            );

    # DSID upper range value needs to be big, but high order bit must be 0
    # MSID will be DSID with high order bit = 1
    my $small_dsid = 0x00000400;  # 2^10  = 1024
    my $big_dsid   = 0x00100000;  # 2^20  = 1024 * 1024 =     1,048,576
    # small_msid would then be 0x10000400 = 2^31 +2^10  = 2,147,484,672
    # big_msid would then be   0x10100000 = 2^31 +2^20  = 2,148,532,224

    # this can be any 1 node on site A and 1 node on site B
    # population of this hash depends on how method was called.
    my %node_components = ( A => undef, B => undef );
    if (ref($object_or_pkg)) {
        # called on an instance.
        my $mcc = $object_or_pkg;

        my ($dr_group) = values(%{ $mcc->get_DR_groups() });   # arbitrary
        my $nodes_by_role = $dr_group->get_nodes_by_role();
        $node_components{A} = $nodes_by_role->{A1}->get_Node_component();
        $node_components{B} = $nodes_by_role->{B1}->get_Node_component();
    } elsif (exists($opts{all_nodes})) {
        # called on the class.
        my %node_component_vs_cluster_name = ();
        foreach my $node_name (@{ $opts{all_nodes} }) {
            my $host = Tharn::host($node_name);
            my $node_component = NACL::C::SystemNode->new(hostrec => $host);

            my $ClusterIdentity_state = NACL::CS::ClusterIdentity->fetch(
                                          command_interface => $node_component,
                                          requested_fields  => [ 'name' ] );
            my $cluster_name = $ClusterIdentity_state->name();

            $node_component_vs_cluster_name{$cluster_name} = $node_component;

            if (scalar(keys(%node_component_vs_cluster_name)) > 1) {
                # we've now gotten an arbitrary node in each cluster
                last;
            }
        }

        if (scalar(keys(%node_component_vs_cluster_name)) == 2) {
            # we don't really care which is at "A" vs "B" here, just that they
            # are one node at each site.
            ($node_components{A}, $node_components{B}) =
                                       values(%node_component_vs_cluster_name);
        }
    }

    unless (defined($node_components{A}) && defined($node_components{B})) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "Could not determine or create Node components!");
    }

    my @vol_states = NACL::CS::Volume->fetch(
                      command_interface => $node_components{A},
                      requested_fields  => [ 'dsid', 'aggregate', 'vserver', ],
                                            );

    my @A_dsids = ();

    # if vserver CRS_Vserver exists already, then we should have already been
    # through the DSID range assignment logic, and all we are going to do now
    # is look at the DSIDs associated with the CRS_Vserver vols as a double
    # check.  this avoids the issue of volumes that are mirrored across sites
    # having the same DSIDs (which they should).
    my $site_A_CRS_Vserver_found =
           List::MoreUtils::any { $_->vserver() eq 'CRS_Vserver' } @vol_states;
    if ($site_A_CRS_Vserver_found) {
        $Log->comment("vserver CRS_Vserver exists.  We assume then that \n" .
                      "DSID ranges have already been negotiated in prior\n" .
                      "MCC system build mechanisms, and will only do a\n" .
                      "cursory check of vol DSIDs associated with the\n" .
                      "CRS_Vserver at each site.");
        # filter on vols in CRS_Vserver
        my @CRS_Vserver_vol_states =
                           grep { $_->vserver() eq 'CRS_Vserver' } @vol_states;
        # get those vol's DSIDs
        @A_dsids = map { $_->dsid() } @CRS_Vserver_vol_states;
        $Log->comment('Existing site A CRS_Vserver volume DSIDs: ' .
                                                         join(q/,/, @A_dsids));
    } else {
        # here we assume we need to check all non-root DSIDs in the system
        my @root_aggr_states = NACL::CS::Aggregate->fetch(
                                 command_interface => $node_components{A},
                                 filter => { 'root' => 'true' },
                                 requested_fields  => [ 'aggregate', ],
                                                     );
        # grab each non-root numeric DSID
        foreach my $vol_state (@vol_states) {
            if (List::MoreUtils::any { $_->aggregate() eq
                                 $vol_state->aggregate() } @root_aggr_states) {
                # ignore this vol.  it is in a root aggr.
            } else {
                push(@A_dsids, $vol_state->dsid());
            }
        }
        $Log->comment('Existing site A (non-root-aggr) volume DSIDs: ' .
                                                         join(q/,/, @A_dsids));
    }

    # booleans
    my $site_A_has_low_dsids = scalar(@A_dsids) &&
                            List::MoreUtils::any { $_ <  $big_dsid } @A_dsids;
    my $site_A_has_high_dsids = scalar(@A_dsids) &&
                            List::MoreUtils::any { $_ >= $big_dsid } @A_dsids;

    # determine current setting for next DSID at site A
    my $result = $node_components{A}->apiset(interface => 'CLI')->
            execute_command(command =>
                           'set -confirmations off -rows 0 -privilege test; ' .
                           'debug smdb table nextIdTable show -fields dsid');
    my ($site_A_next_dsid) = ($result =~ /(\d+)/);
    $Log->comment("Current next volume DSID for site A: $site_A_next_dsid");
    my $site_A_next_dsid_is_low = ($site_A_next_dsid < $big_dsid);

    # and for site B...
    @vol_states = NACL::CS::Volume->fetch(
                                 command_interface => $node_components{B},
                                 requested_fields  => [ 'dsid', 'aggregate', ],
                                         );

    # (see site A process comments above)
    my @B_dsids = ();
    if ($site_A_CRS_Vserver_found) {
        # we better find a CRS_Vserver at site B as well...
        unless (List::MoreUtils::any { $_->vserver() eq 'CRS_Vserver' }
                                                                 @vol_states) {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                  "vserver 'CRS_Vserver' found at site A, but not at site B!");
        }

        # filter on vols in CRS_Vserver
        my @CRS_Vserver_vol_states =
                           grep { $_->vserver() eq 'CRS_Vserver' } @vol_states;
        # get those vol's DSIDs
        @B_dsids = map { $_->dsid() } @CRS_Vserver_vol_states;
        $Log->comment('Existing site B CRS_Vserver volume DSIDs: ' .
                                                         join(q/,/, @B_dsids));
    } else {
        my @root_aggr_states = NACL::CS::Aggregate->fetch(
                                 command_interface => $node_components{B},
                                 filter => { 'root' => 'true' },
                                 requested_fields  => [ 'aggregate', ],
                                                     );
        # grab each non-root numeric DSID
        foreach my $vol_state (@vol_states) {
            if (List::MoreUtils::any { $_->aggregate() eq
                                 $vol_state->aggregate() } @root_aggr_states) {
                # ignore this vol.  it is in a root aggr.
            } else {
                push(@B_dsids, $vol_state->dsid());
            }
        }
        $Log->comment('Existing site B (non-root-aggr) volume DSIDs: ' .
                                                         join(q/,/, @B_dsids));
    }

    # booleans
    my $site_B_has_low_dsids = scalar(@B_dsids) &&
                            List::MoreUtils::any { $_ <  $big_dsid } @B_dsids;
    my $site_B_has_high_dsids = scalar(@B_dsids) &&
                            List::MoreUtils::any { $_ >= $big_dsid } @B_dsids;

    # determine current setting for next DSID at site A
    $result = $node_components{B}->apiset(interface => 'CLI')->
            execute_command(command =>
                           'set -confirmations off -rows 0 -privilege test; ' .
                           'debug smdb table nextIdTable show -fields dsid');
    my ($site_B_next_dsid) = ($result =~ /(\d+)/);
    $Log->comment("Current next volume DSID for site B: $site_B_next_dsid");
    my $site_B_next_dsid_is_low = ($site_B_next_dsid < $big_dsid);


    # check for problem cases first
    if (            ($site_A_has_low_dsids  && $site_A_has_high_dsids) ||
                    ($site_A_has_low_dsids  && $site_B_has_low_dsids ) ||
                    ($site_A_has_high_dsids && $site_B_has_high_dsids) ||
                    ($site_B_has_low_dsids  && $site_B_has_high_dsids)) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                "WARNING: Clashing volume DSID ranges exist in this MCC\n" .
                "         system.  They must be cleaned up before continuing.");
    }

    # all the other possibilities are okay, but the next DSID at each site
    # may need to be tweaked...
    my $new_dsid = { 'A' => undef, 'B' => undef };

    # let's now make sure the next DSIDs at each site are in the same
    # ranges as any existing volume DSIDs at that site.
    # much of this logic is there so that we don't change anything unless
    # we need to.
    if ($site_A_has_low_dsids) {
        # next site A DSID better be low...
        if (! $site_A_next_dsid_is_low) {
            $new_dsid->{A} = List::Util::max(@A_dsids, $small_dsid) + 1;
        }
        # ...and next site B DSID better be high
        if ($site_B_next_dsid_is_low) {
            $new_dsid->{B} = List::Util::max(@B_dsids, $big_dsid) + 1;
        }
    } elsif ($site_A_has_high_dsids) {
        # next site A DSID better be high...
        if ($site_A_next_dsid_is_low) {
            $new_dsid->{A} = List::Util::max(@A_dsids, $big_dsid) + 1;
        }
        # ...and next site B DSID better be low
        if (! $site_B_next_dsid_is_low) {
            $new_dsid->{B} = List::Util::max(@B_dsids, $small_dsid) + 1;
        }
    # only get to here if site A had no allocated volume DSIDs...
    } elsif ($site_B_has_low_dsids) {
        # next site B DSID better be low...
        if (! $site_B_next_dsid_is_low) {
            $new_dsid->{B} = List::Util::max(@B_dsids, $small_dsid) + 1;
        }
        # ...and next site A DSID better be high
        if ($site_A_next_dsid_is_low) {
            $new_dsid->{A} = $big_dsid;
        }
    } elsif ($site_B_has_high_dsids) {
        # next site B DSID better be high...
        if ($site_B_next_dsid_is_low) {
            $new_dsid->{B} = List::Util::max(@B_dsids, $big_dsid) + 1;
        }
        # ...and next site A DSID better be low
        if (! $site_A_next_dsid_is_low) {
            $new_dsid->{A} = $small_dsid;
        }
    # only get here if neither site had allocated volume DSIDs
    } else {
        if ($site_A_next_dsid_is_low && $site_B_next_dsid_is_low) {
            # arbitrarily pick B site to have high volume DSIDs
            $new_dsid->{B} = $big_dsid;
        } elsif ((! $site_A_next_dsid_is_low) && (! $site_B_next_dsid_is_low)) {
            $new_dsid->{A} = $small_dsid;
        }
    }

    # only change next DSID (and MSID with it) setting if needed
    foreach my $site (qw/ A B /) {
        if (defined($new_dsid->{$site})) {
            my $dsid = $new_dsid->{$site};
            # highest order bit (32 bit) must be 1 for MSIDs.  will OR this in
            my $msid = $dsid | 0x80000000;
            $Log->warn("Setting next site $site DSID/MSID to be " .
                                  sprintf("0x%x / 0x%x", $dsid, $msid) . ' !');
            $node_components{$site}->apiset(interface => 'CLI')->
                execute_command(command =>
                           "set -confirmations off -rows 0 -privilege test; " .
                           "debug smdb table " .
                               "nextIdTable modify -dsid $dsid -msid $msid");
        }
    }

    return;
}

# this is a private method to determine which build is installed on mcc testbed
sub _determine_ontap_version_info{
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my $ci = $self->_site_A_metrocluster_component()->command_interface();
    my $version_info = {};
    $version_info->{branch} = $ci->get_version_manager()->
                                        get_version_attribute(
                                                    attribute => 'branch');
    $Log->comment('ONTAP BUILD INSTALLED: '.$version_info->{branch});

    return $version_info;
}

=head2 has_uichange

    my $has_change = $mcc->has_uichange(
                         uichange => "has_this_ui_change_string");


Returns a boolean indicating if the ontap build has the passed
UI change string

=cut

sub has_uichange {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            uichange => { type => SCALAR },
                  },
    );

    my $ci = $self->_site_A_metrocluster_component()->command_interface();
    my $has_change;
    try {
        $has_change =
            $ci->get_version_manager()->has_uichange(
                uichange => $opts{uichange}
            );
    } catch NATE::BaseException with { #TODO should we just let it throw this?
        $Log->comment( "problem with getting UI change strings ($opts{uichange})...\n");
    };

    return $has_change;
}

=head2 workaround_expired

    if (! $mcc->workaround_expired(
           uichange => "has_this_ui_change_string",
           burt => 123456,
       )) {
        ... do this workaround ...
    }


Returns a boolean indicating if the workaround has expired.  With the new
way workarounds are done, this means that a particular UI change string
exists in the ontap build.

If the ui change string exists in ontap, we are to no longer execute
the workaround.  Thus, this method returns true if the string is present
and false if the string is not present.

There may be a more complicated evaluation in the future.

=cut
sub workaround_expired {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    #TODO.  We MAY allow this to be a static method in the future (the older
    #workaround expired based on dates was static)
    #however, since we need ontap info it may not make sense (we'd
    #only need it if a script needed it outside of MCC usage really)

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            uichange => { type => SCALAR },
            burt     => { regex => qr/^\d\d\d\d\d\d$/ },
                  },
    );

    my $has_uichange = $self->has_uichange(
        uichange => $opts{uichange}
    );

    if (!$has_uichange) {
        $Log->comment("\n");
        $Log->comment("Workaround for BURT $opts{burt} being used " .
            "($opts{uichange} UI change string not present) ...");
        $Log->comment("\n");
        return 0;
    } else {
        $Log->comment("Workaround for BURT $opts{burt} has expired! " .
            "($opts{uichange} UI change string present) ...");
        return 1;
    }

}

=head2 get_ontap_version_info()

    my $build = $mcc->get_ontap_version_info();

    $build->{branch} has the branch information

This is a public method use to find which ONTAP build is installed over
hardware. Method returns the hashreference which contains build installed over
mcc hardware

=cut

sub get_ontap_version_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    return ( $self->_ontap_build_version);
}

=head2 get_Metrocluster_component()

    my $site_A_metrocluster_component =
                                 $mcc->get_Metrocluster_component(site => 'A');

This is a public utility accessor for both sites' Metrocluster components.
Returns a reference to the NACL::C::Metrocluster component object associated
with site A or site B.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Specifies which site's cluster is desired.

=back

=back

=cut

sub get_Metrocluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/ },
                  },
                            );

    my $metrocluster_component = ($opts{site} =~ /A/) ?
                       $self->_site_A_metrocluster_component() :
                       $self->_site_B_metrocluster_component();

    return $metrocluster_component;
}

=head2 get_cluster()

    my $cluster_A = $mcc->get_cluster(site => 'A');

Gets the WaflMCC::MCC::Cluster object associated with specified site.

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Required) Specifies which site's cluster is desired.

=back

=back

=cut

sub get_cluster {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            site => { regex => qr/^[AB]$/ },
                  },
                            );

    return ($opts{site} =~ /A/) ? $self->_cluster_A() : $self->_cluster_B();
}

=head2 get_disaster_status()

    my $disaster_status = $mcc->get_disaster_status();

    my $disaster_status = $mcc->get_disaster_status(include_HA => 1);

This method returns DR state information about the overall Metrocluster system.
Optionally it will also return HA state information for all available HA groups.

=over

=item Options

=over

=item C<< include_HA => 0|1  (default = 0) >>

(Optional) Includes HA states in all the HA Groups in the system if true.

=item C<< dr_group => $dr_group  >>

(Optional) Specify a particular DR Group status only if desired.  This is only
useful if include_HA => 1 is also passed, since all DR Groups have the same
DR status.  But this may be passed to filter the HA status checks to only a
particular DR Group.

=back

=back

The data structure returned has the following format:

    {
        DR_state      => $dr_state,
        live_sites    => \@queryable_sites,
        survivor_site => $survivor_site,
        disaster_site => $disaster_site,
        HA_states     => {
                $DR_group_id => {
## BEWARE that a site will not appear here if it is not in live_sites array ###
#                       A => {                                                #
#                               HA_state      => $ha_state,                   #
#                               survivor_node => $survivor_node,              #
#                               disaster_node => $disaster_node,              #
#                            },                                               #
#                       B => {                                                #
#                               HA_state      => $ha_state,                   #
#                               survivor_node => $survivor_node,              #
#                               disaster_node => $disaster_node,              #
#                            },                                               #
###############################################################################
                                },
                $DR_group_id => {
                        .
                        .
                        .
                                },
                         }
    }

where

 * DR_state is one of the values listed in
   WaflMCC::MCC::DR_Coordinator::get_DR_state()
 * live_sites is a list containing 'A' and/or 'B' (note that this list can
   contain both sites in both 'normal' and 'heal_roots_completed' DR states)
 * survivor_site is undef if DR_state is 'normal'
 * disaster_site is undef if DR_state is 'normal'
 * HA_state is one of the values listed in
   WaflMCC::MCC::HA_Group::get_HA_state()
 * survivor_node and disaster_node are WaflMCC::MCC::Node objects, or are
   undef if HA_state is 'normal'

=cut

sub get_disaster_status {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            include_HA => { regex    => qr/^[01]$/,
                            default  => 0, },
            dr_group   => { isa      => 'WaflMCC::MCC::DR_Group',
                            optional => 1, },
                  },
                            );

    return $self->L_get_disaster_status(%opts);
}

=head2 L_get_disaster_status()

This is the engine for the get_disaster_status method.  It is seperate from the
get_disaster_status() because a non-public option is available here.

=over

=item Options

See get_disaster_status() options, and add the following option(s) as well:

=over

=item C<< known_live_sites => \@live_sites >>

(Optional) @live_sites may include 'A' and/or 'B'.  This option should only ever
be used when the context of the call involves a known state of the queryability
of the two sites.

For example, at MCC object initialization, the system must be in a normal
state, which presupposes both sites A and B are live and queryable.  This option
should NEVER be used except when these sites' states are absolutely known.

This option allows for faster completion as it skips the subtests that "ping"
the two sites.

=back

=back

=cut

sub L_get_disaster_status {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            include_HA       => { regex    => qr/^[01]$/,
                                  default  => 0, },
            dr_group         => { isa      => 'WaflMCC::MCC::DR_Group',
                                  optional => 1, },
            known_live_sites => { type => ARRAYREF,
                                  optional => 1, },
                  },
                            );

    my $live_sites_ref = (defined($opts{known_live_sites})) ?
                     $opts{known_live_sites} : $self->L_determine_live_sites();
    my @live_sites = @{ $live_sites_ref };

    my %dr_states = ();
    foreach my $site (@live_sites) {
        my $cluster = $self->get_cluster(site => $site);
        $cluster->get_Metrocluster_component()->command_interface()->
                             refresh_command_interface(no_version_change => 1);
        $dr_states{$site} = $cluster->get_DR_Coordinator()->get_DR_state();
    }

    my $survivor_site = undef;
    my $disaster_site = undef;
    my $dr_state      = undef;

    # specific conditions for us to say we have a well defined dr_state
    if (scalar(@live_sites) == 2) {
        if (scalar(grep { $_ eq 'normal' } values(%dr_states)) == 2) {
            # 2 live sites claim "normal".  it is normal.
            $dr_state      = 'normal';
            $survivor_site = undef;
            $disaster_site = undef;
        } else {
            # the disaster site can be live @ heal_roots_completed
            foreach my $site (@live_sites) {
                if ($dr_states{$site} eq 'heal_roots_completed') {
                    $dr_state      = 'heal_roots_completed';
                    $survivor_site = $site;
                    $disaster_site = ($site eq 'A') ? 'B' : 'A';
                    last;
                }
            }
        }
    # beware that we expect that the "heal roots" operation is not complete
    # until the disaster site is live again.  heal_roots_completed should
    # only be detected in the case above where (scalar(@live_sites) == 2)
    } elsif (scalar(@live_sites) == 1) {
        if ($dr_states{$live_sites[0]} =~ m{^( switchover_completed |
                                               heal_aggrs_completed )$}x) {
            $dr_state      = $1;
            $survivor_site = $live_sites[0];
            $disaster_site = ($survivor_site eq 'A') ? 'B' : 'A';
        }
    }

    unless (defined($dr_state)) {
        # these dr states are not the ones we are looking for...
        WaflMCC::MCC::Exceptions::BaseException->throw(
                         "ERROR: Invalid DR states found:\n" .
                         WaflMCC::MCC::Utils->dump(%dr_states) . "\n\n");
    }

    my $disaster_status = {
            DR_state      => $dr_state,
            live_sites    => \@live_sites,
            survivor_site => $survivor_site,
            disaster_site => $disaster_site,
                          };

    # add HA disaster info if requested
    if ($opts{include_HA}) {
        $disaster_status->{HA_states} = { };
        my $specified_dr_group = delete($opts{dr_group});
        my $dr_groups = defined($specified_dr_group) ?
                     { $specified_dr_group->get_id() => $specified_dr_group } :
                     $self->get_DR_groups();
        foreach my $dr_group_id (keys(%{ $dr_groups })) {
            my $dr_group = $dr_groups->{$dr_group_id};
            $disaster_status->{HA_states}->{$dr_group_id} = { };
            foreach my $site (@live_sites) {
                my $ha_group = $dr_group->get_HA_groups()->{$site};
                $disaster_status->{HA_states}->{$dr_group_id}->{$site} =
                                                   $ha_group->L_get_HA_state();
            }
        }
    }

    return $disaster_status;
}

=head2 L_determine_live_sites()

Uses a subtest (which may time-out) for each site to find which sites are
currently in a state such that they may be queried for MCC information.

=cut

sub L_determine_live_sites {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @live_sites = ();
    # these subtests are quick queries to both sites one of them may fail
    my %subtest_handles = ();
    my $DR_STATE_QUERY_TIMEOUT = 120;
    foreach my $site (qw/ A B /) {
        my $cluster = $self->get_cluster(site => $site);
        my $dr_coordinator = $cluster->get_DR_Coordinator();

        # this should just try to do a simple query for the dr state.  all we
        # care about here is if the subtest succeeds or times out.  it is
        # essentially only a 'ping' to see if the Metrocluster CI is responding.
        $subtest_handles{$site} = subtest(
                sub {

                    # Leaving this in for documentation of an interesting use
                    # of subtest, where you can get some information out of one
                    # after it has completed (see follow up call commented out
                    # below).  turns out we don't end up needing this
                    # functionality here, but maybe sometime it will be useful
                    # to see how this worked when we were using it successfully.
                    #
                    # my $cluster = $self->get_cluster(site => $site);
                    # my $dr_coordinator = $cluster->get_DR_Coordinator();
                    # my $dr_state = $dr_coordinator->get_DR_state();
                    # $Parent->message_put(type    => 'DR_STATE_REPORT',
                    #                      message => $dr_state);

                    # replaced with:

                    try {
                        $cluster->get_Metrocluster_component()->
                             command_interface()->
                             refresh_command_interface(no_version_change => 1);
                        my $dr_state = $dr_coordinator->get_DR_state();
                        my $result = NATE::Result::Info->new("Site is up...");
                        $Tharn::Log->log($result);
                    } catch NATE::BaseException with {
                        print Dumper shift;
                        $Log->comment(
                              "Ignoring a NATE::BaseException most likely " .
                              "due to an ssh connection timeout.");
                        my $result = NATE::Result::Info->new("Site is down?...");
                        $Tharn::Log->log($result);
                        # force an eventual "natural" subtest timeout (i guess)
                        sleep($DR_STATE_QUERY_TIMEOUT * 2);
                    };

                },
                '-bg',
                '-timeout' => $DR_STATE_QUERY_TIMEOUT,
                '-iffatal' => '-ok',
                '-runid'   => 'DR_STATE_QUERY__site_' . $site,
                                         );
    }

    # we now get the results of each live site's dr state query
    my %dr_states = ();
    foreach my $site (keys(%subtest_handles)) {
        my $subtest = $subtest_handles{$site};
        $subtest->wait_finish();
        if ($subtest->exit_status() == 0) {
            # this commented out call is associated with the code commented out
            # above.  see notes therein.
            # $dr_states{$site}= $subtest->message_get(type=>'DR_STATE_REPORT');

            # the subtest returned successfully.  it is safe to query this site.
            push(@live_sites, $site);
        }
    }

    return \@live_sites;
}

=head2 _create_DR_groups()

This method creates the DR Group objects associated with the DR Groups that are
resident in this MCC system.  Returns an arrayref of WaflMCC::MCC::DR_Group
objects.

=cut

sub _create_DR_groups {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # validate opts
    my %opts = validate_with(
        params => \@_,
        spec   => {
            bootstrap_util                        => { isa   =>
                           'WaflMCC::MCC::Metrocluster::_Bootstrap_Util' },
            constructor_log_path                  => { type  => SCALAR },
                  },
                            );

    my $bootstrap_util       = delete($opts{bootstrap_util});
    my $constructor_log_path = delete($opts{constructor_log_path});

    my @dr_groups = ();
    foreach my $A1_node_name (@{ $bootstrap_util->get_A1_node_names() }) {
        push(@dr_groups, WaflMCC::MCC::DR_Group->new(
               metrocluster                  => $self,
               A1_node_name                  => $A1_node_name,
               constructor_log_path          => $constructor_log_path,
               site_A_Metrocluster_component =>
                        $bootstrap_util->get_site_A_Metrocluster_component()));
    }

    return \@dr_groups;
}


=head2 get_DR_groups()

Get the DR Groups present across clusters A and B.

    my %dr_groups = %{ $mcc->get_DR_groups() };

Returns a reference to a hash of the form:

    { $group_X_id => $dr_group_X,
      $group_Y_id => $dr_group_Y,
      ...                         }

where the keys are DR Group IDs (currently the name of the A1 node in a
particular DR Group), and the values are WaflMCC::MCC::DR_Group
objects.  These represent each of the DR Groups present in the Metrocluster
system across clusters A and B.

=cut

sub get_DR_groups {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %dr_groups = map { $_->get_id() => $_ } @{ $self->_dr_groups() };

    return \%dr_groups;
}

=head2 get_singleton_DR_group()

This is purely a utility method which makes it simpler to get the DR Group
when the Metrocluster contains only a single DR Group (see notes in
get_DR_groups() for details).  If more than 1 DR Group is found this method
will throw an exception.

When the multi-DR Group Metroclusters come into being, this method will likely
be deprecated from that ONTAP version onwards.

    my $singleton_dr_group = $mcc->get_singleton_DR_group();

Returns a WaflMCC::MCC::DR_Group object.

May throw a WaflMCC::MCC::Exceptions::InvalidMethodUsage exception.

=cut

sub get_singleton_DR_group {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # Get the DR Groups present across clusters A and B.  If there is only a
    # single DR Group found, return it.  If more than one DR Groups are found,
    # then the call to this method is no longer valid, and an exception is
    # thrown.

    # It is likely that this method should be deprecated when such exceptions
    # start being thrown.

    my @dr_groups = values(%{$self->get_DR_groups()});

    if (@dr_groups > 1) {
        WaflMCC::MCC::Exceptions::InvalidMethodUsage->throw(
             'get_singleton_DR_group() is no longer a valid method ' .
             'when Metrocluster elements may contain multiple ' .
             'DR Groups.  get_singleton_DR_group() likely needs to be ' .
             'deprecated from this ONTAP version onward.');
    }

    return $dr_groups[0];
}

=head2 get_all_mcc_nodes()

This method returns an arrayref of Metrocluster associated nodes in both
clusters.

    my @mcc_nodes = @{ $mcc->get_all_mcc_nodes() };

The objects in the array are WaflMCC::MCC::Node objects.

=cut

sub get_all_mcc_nodes {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @nodes = ();
    foreach my $site ('A', 'B') {
        my $cluster = $self->get_cluster(site => $site);
        push(@nodes, @{$cluster->L_get_cluster_mcc_nodes()} );
    }

    return \@nodes;
}

=head2 get_mcc_node()

This method finds a named node in the Metrocluster system if it exists.

    my $mcc_node = $mcc->get_mcc_node(
                       node_name => 'spectre-l11');

or

    my $mcc_node = $mcc->get_mcc_node(
                       node_uuid => 'd437c8cd-e931-11e0-aa13-c948a5272a17');

or

    my $mcc_node = $mcc->get_mcc_node(
                       node_nvramid => 'd437c8cd-e931-11e0-aa13-c948a5272a17');

One (and only one) of the 3 possible parameters must be passed.

=over

=item Options

=over

=item C< node_name >

Name of the node to find.  Be aware that node names CAN change, even in the
middle of a DR lifecycle.  In scripts, please assume that this value is
unreliable.  Node UUIds are reliable (as are NVRAM IDs generally.)

=item C< node_uuid >

Node UUID of the node to find.

=item C< node_nvramid >

Node NVRAM ID of the node to find.

=back

=back

Returns a WaflMCC::MCC::Node object.

=cut

sub get_mcc_node {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_name => { regex    => qr/^[\w\-]+$/, optional => 1 },
            node_uuid => { regex    => qr/^[\w\-]+$/, optional => 1 },
        },
                            );

    # one and only one of the parameters node_name and
    # node_uuid is required.  Check this requirement.
    {
        my $parameter_error_message =
            "WaflMCC::MCC::Node->metrocluster_enable() requires either\n" .
            "    the node_name parameter exclusive-or the\n" .
            "    node_uuid parameter to be passed.";

        if (defined($opts{node_name})) {
            if (defined($opts{node_uuid})) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                                                     $parameter_error_message);
            }
        } elsif (! defined($opts{node_uuid})) {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                                                     $parameter_error_message);
        }
    }

    my $found_node = undef;
    my @nodes = @{ $self->get_all_mcc_nodes() };
    foreach my $node (@nodes) {
        if (defined($opts{node_name})) {
            if ($node->get_name() eq $opts{node_name}) {
                $found_node = $node;
                last;
            }
        } elsif ($node->get_uuid() eq $opts{node_uuid}) {
            $found_node = $node;
            last;
        }
    }

    if (! defined($found_node)) {
        my $node_id = defined($opts{node_name}) ?
                                           $opts{node_name} : $opts{node_uuid};
        my $msg = "Node [$node_id] not found!";
        WaflMCC::MCC::Exceptions::BaseException->throw($msg);
    }

    return $found_node;
}

=head2 refresh_command_interfaces()

This method refreshes command interfaces for site A and/or B.

    $mcc->refresh_command_interfaces(site => 'A');

=over

=item Options

=over

=item C<< site => 'A'|'B' >>

(Optional) IF passed, this option flags which site's command interfaces should
be refreshed.  If not passed, both sites' command interfaces will be refreshed.

=back

=back

=cut

sub refresh_command_interfaces {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            site => { regex => qr/^[AB]$/, optional => 1, },
                  },
                            );

    foreach my $site ('A', 'B') {
        if ((! defined($opts{site})) || ($opts{site} eq $site)) {
            $self->get_Metrocluster_component(site => $site)->
                    command_interface()->refresh_command_interface(
                                                       no_version_change => 1);

            my $mcc_nodes =
                  $self->get_cluster(site => $site)->L_get_cluster_mcc_nodes();
            foreach my $mcc_node (@{ $mcc_nodes }) {
                $mcc_node->get_Node_component()->command_interface()->
                             refresh_command_interface(no_version_change => 1);
            }
        }
    }

    return;
}

=head2 check()

    my $result = $mcc->check(site => 'A');

Runs metrocluster check run (or show).  If there are any warnings for nodes or
aggreggrates then more detailed commands are run to gather what is failing.
Exceptions are then thrown.

=over

=item Options

=over

=item C<< site =E<gt> 'A'|'B' >>

Specifies which site's cluster the check command is initiated from.


=item C<< run =E<gt> '1'|'0' (default=1) >>

Specifies whether to do metrocluster check run or a metrocluster check show

=back

=back

=cut

sub check {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,@opts) = @_;

    #validate opts
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            site => { regex => qr/^[AB]$/ },
            run  => { regex => qr/^[01]$/, default => 1 },
                  },
               );
    my $run  = $opts{run};
    my $site = $opts{site};

    my $branch = $self->get_ontap_version_info->{branch};

    if ($branch ne 'mcc_config_repl') {
        $Log->comment("Skipping 'metrocluster check' test as the branch " .
                      "'$branch' doesn't support it.  If you think it does" .
                      "please email dl-mcc-dpg-qa so we can enable it.");
        return;
    }


    my $disaster_status = $self->get_disaster_status(include_HA => 1);

    my $check = $self->_metrocluster_check_show(
                       run  => $run,
                       site => $site,

                );

    $Log->comment(Dumper $check);

    my %fail;

    my %is_2_node;
    my $dr_groups = $self->get_DR_groups();
    foreach my $dr_group (values(%{ $dr_groups } )) {
        if (! $dr_group->is_4_pack()) {
            foreach my $node (@{ $dr_group->get_mcc_nodes() }) {
                $is_2_node{$node->get_name()} = $node->get_uuid();
            }

        }
    }
    #If NOT normal DR state then check all nodes,
    #otherwise ignore some errors: BURT 596281

    my %dr_disaster_node;
    if ($disaster_status->{DR_state} !~ /normal/) {
        $Log->comment(
            "metrocluster check will not function on all nodes " .
            " in this DR state");
        #We need a list of all the nodes that are down so that when
        #we examine the 'check' response, we can ignore errors related
        #to nodes being down for nodes that really ARE down
        #
        #TODO redo this with dr_group->get_nodes(site => ...) ?
        my $cluster = $self->get_cluster(site =>
            $disaster_status->{disaster_site});
        foreach my $node (@{$cluster->L_get_cluster_mcc_nodes()}) {
            $dr_disaster_node{$node->get_name()} = $disaster_status->{disaster_site};
        }
    }

    #don't ignore HA errors for actual HA disaster nodes (BURT 598768)
    my %ha_disaster_node;

    foreach my $ha_dr_group_id (values %{ $disaster_status->{HA_states} }) {
        foreach my $dr_site (values %{ $ha_dr_group_id }) {
            if ($dr_site->{disaster_node}) {
                $ha_disaster_node{$dr_site->{disaster_node}} =
                    $dr_site->{HA_state};
            }
        }
    }


    foreach my $item (@{$check}) {
        if (lc $item->{result} ne 'ok') {
            #$fail{$item->{component}} = $item; #move re: burt 593924

            if (lc $item->{component} eq 'nodes') {
                my $result =
                    $self->_metrocluster_check_node_show(site => $site,);
                foreach my $check (@{$result}) {
                    #most will be gone when BURTs are fixed...
                    ## no critic (ControlStructures::ProhibitCascadingIfElse)

                    if (lc $check->{result} ne 'ok') {
                        if (defined $dr_disaster_node{$check->{node}}) { #node down?
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "$check->{node} is unreachable (dr state " .
                                "$disaster_status->{DR_state})\n");
                        } elsif (defined $is_2_node{$check->{node}} &&
                            $check->{check} =~ /symmetric dr auxiliary/i) {
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "2 node systems have no HA partner"
                            );
                        } elsif (defined $is_2_node{$check->{node}} &&
                            $check->{check} =~ /symmetric ha relationship/i) {
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "2 node systems have no HA partner"
                            );
                        } elsif (!defined $ha_disaster_node{$check->{node}} &&
                            $check->{check} =~ /symmetric ha relationship/i) {
                            #these burt 598768 ignores will ignore more than
                            #necessary, remove when possible
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 598768"
                            );
                        } elsif (!defined $ha_disaster_node{$check->{node}} &&
                            $check->{check} =~ /ha partner/i) {
                            #these burt 598768 ignores will ignore more than
                            #necessary, remove when possible
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 598768"
                            );
                        } elsif (!defined $ha_disaster_node{$check->{node}} &&
                            $check->{check} =~ /dr partner in peer/i) {
                            #these burt 598768 ignores will ignore more than
                            #necessary, remove when possible
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 598768"
                            );
                        } elsif (!defined $ha_disaster_node{$check->{node}} &&
                            $check->{check} =~ /reachable/i) {
                            #these burt 598768 ignores will ignore more than
                            #necessary, remove when possible
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 598768"
                            );
                        } elsif (!defined $ha_disaster_node{$check->{node}} &&
                            $check->{check} =~ /metrocluster ready/i) {
                            #these burt 598768 ignores will ignore more than
                            #necessary, remove when possible
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 598768"
                            );
                        } elsif (($check->{check} =~ /symmetric dr/i) &&
                                 ($disaster_status->{DR_state} !~ /normal/)) {
                            #symmetric checks only work in normal DR state
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "$check->{node} is unreachable (dr state " .
                                "$disaster_status->{DR_state})\n");
                        } elsif ($check->{check} =~ /mirroring/i) {
                            $Log->comment(
                                "Ignoring $check->{check} error, " .
                                "see BURT 593924\n");
                        } else {
                            $fail{$item->{component}} = $item; # burt 593924
                            $Log->comment(
                                "$check->{node} : $check->{check} failed\n");
                            $fail{$check->{node}}->{$check->{check}} = $check;
                        }
                    }
                }


            } elsif (lc $item->{component} eq 'aggregates') {
                my $result =
                    $self->_metrocluster_check_aggregate_show(site => $site,);
                foreach my $check (@{$result}) {

                    if (lc $check->{result} ne 'ok') {
                        $Log->comment(
                            "$check->{node} : $check->{check} failed\n");
                        $fail{$check->{node}}->{$check->{check}} = $check;
                        $fail{$item->{component}} = $item; # burt 593924
                    }
                }
            }
        }
    }

    if (keys %fail) {
        $Log->comment(Dumper \%fail);
        WaflMCC::MCC::Exceptions::BaseException->throw(
                                     "ERROR: metrocluster check warnings:\n" .
                                     join q{,},keys %fail);


    }

    return 1; #OK if we didn't fail earlier
}

sub _metrocluster_check_show {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,@opts) = @_;

    #validate opts
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            site => { regex => qr/^[AB]$/ },
            run  => { regex => qr/^[01]$/, default => 1 },
                  },
               );
    my $run  = $opts{run};
    my $site = $opts{site};

    #TODO this needs update later when the API changes...(component vs apiset)

    my $component = ($site =~ /A/) ?
                     $self->_site_A_metrocluster_component() :
                     $self->_site_B_metrocluster_component();
    #for APISet
    my $hostobj = $component->{command_interface}->{version_manager}->{hostobj};
    #$name = $component->{command_interface}->
    #{version_manager}->{hostobj}->{hostname};


    my $apiset = NACL::APISet->new(
        hostobj   => $hostobj, #Tharn::host($cluster_name),
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode'
    );
    $apiset->set_timeout( 'connectrec-timeout' => 10*60 );


    if ($run) {
        my $return =  $apiset->metrocluster_check_run(
                          'privilege-level' => 'test',
                      );

        return $return->get_parsed_output();
    } else { #use show instead if not running
        my $show =  $apiset->metrocluster_check_show(
                          'privilege-level' => 'test',
                      );

        return $show->get_parsed_output();
    }
}

sub _metrocluster_check_aggregate_show {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,@opts) = @_;

    #validate opts
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            site => { regex => qr/^[AB]$/ },
                  },
               );
    my $site = $opts{site};

    #TODO this needs update later when the API changes...(component vs apiset)

    my $component = ($site =~ /A/) ?
                     $self->_site_A_metrocluster_component() :
                     $self->_site_B_metrocluster_component();
    #for APISet
    my $hostobj = $component->{command_interface}->{version_manager}->{hostobj};
    #$name = $component->{command_interface}->
    #{version_manager}->{hostobj}->{hostname};


    my $apiset = NACL::APISet->new(
        hostobj   => $hostobj, #Tharn::host($cluster_name),
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode'
    );

    $apiset->set_timeout( 'connectrec-timeout' => 10*60 );

    my $run =  $apiset->metrocluster_check_aggr_show(
                      'privilege-level' => 'test',
                  );
    #$Log->comment(Dumper $run->get_parsed_output());

    return $run->get_parsed_output();
}

sub _metrocluster_check_node_show {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,@opts) = @_;

    #validate opts
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            site => { regex => qr/^[AB]$/ },
                  },
               );

    my $site = $opts{site};

    #TODO this needs update later when the API changes...(component vs apiset)

    my $component = ($site =~ /A/) ?
                     $self->_site_A_metrocluster_component() :
                     $self->_site_B_metrocluster_component();
    #for APISet
    my $hostobj = $component->{command_interface}->{version_manager}->{hostobj};
    #$name = $component->{command_interface}->
    #{version_manager}->{hostobj}->{hostname};


    my $apiset = NACL::APISet->new(
        hostobj   => $hostobj, #Tharn::host($cluster_name),
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode'
    );

    $apiset->set_timeout( 'connectrec-timeout' => 10*60 );

    my $run =  $apiset->metrocluster_check_node_show(
                     'privilege-level' => 'test',
                  );
    #$Log->comment(Dumper $run->get_parsed_output());

    return $run->get_parsed_output();
}

## no critic (Subroutines::ProhibitUnusedPrivateSubroutines)
# this could be removed, just here for debugging for a while if someone
# needs it
sub _metrocluster_check_dump {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    #TODO this needs update later when the API changes...
    my $component = $self->_site_A_metrocluster_component();
    my $hostobj = $component->{command_interface}->{version_manager}->{hostobj};
    #$name = $component->{command_interface}->
    #{version_manager}->{hostobj}->{hostname};

    my $apiset = NACL::APISet->new(
        hostobj   => $hostobj, #Tharn::host($cluster_name),
        category  => 'Node',
        interface => 'CLI',
        set       => 'CMode'
    );
    $apiset->set_timeout( 'connectrec-timeout' => 10*60 );

    my $run =  $apiset->metrocluster_check_run(
                          'privilege-level' => 'test',
                      );
    $Log->comment(Dumper $run->get_parsed_output());
    my $show =  $apiset->metrocluster_check_show(
                          'privilege-level' => 'test',
                      );
    $Log->comment(Dumper $show->get_parsed_output());

    $Log->comment(Dumper $run->get_parsed_output());
    $show =  $apiset->metrocluster_check_show(
                          'privilege-level' => 'test',
                          'instance' => q{},
                      );
    $Log->comment(Dumper $show->get_parsed_output());

    $show =  $apiset->metrocluster_check_node_show(
                          'privilege-level' => 'test',
                      );
    $Log->comment(Dumper $show->get_parsed_output());

    $show =  $apiset->metrocluster_check_node_show(
                          'privilege-level' => 'test',
                          'instance' => q{},
                      );
    $Log->comment(Dumper $show->get_parsed_output());

    $show =  $apiset->metrocluster_check_aggr_show(
                          'privilege-level' => 'test',
                      );
    $Log->comment(Dumper $show->get_parsed_output());

    $show =  $apiset->metrocluster_check_aggr_show(
                          'privilege-level' => 'test',
                            'instance' => q{},
                        );
    $Log->comment(Dumper $show->get_parsed_output());

    return;
}

=head2 unconfigure()

This method runs the unconfigure command which destroys the metrocluster
configuration.  Without any arguments it destroys all metroclusters related
to this site.  A node argument will limit this to just the nodes associated
with the passed node.

    my $result = $mcc->unconfigure();
    my $result = $mcc->unconfigure(site => 'A');
    my $result = $mcc->unconfigure(node => 'node_to_remove');

=over

=item Options

=over

=item C<< wait_for_completion =E<gt> '0'|'1' (default=1) >>

Setting to 1 causes the method to wait until unconfigure has completed
before returning.  This will also activate additional internal checks
to verify that unconfigure is operating as expected (to some degree).

=item C<< site =E<gt> 'A'|'B' >>

Specifies which site's cluster the check command is initiated from.


=item C<< node =E<gt> E<lt>node nameE<gt> (optional) >>

If specified, limits unconfigure to just the DR partners related to passed node.

=item C<< metrocluster_node =E<gt> E<lt>node nameE<gt> (for static method) >>

required when calling unconfigure as a static call:

WaflMCC::MCC::Metrocluster-E<gt>unconfigure(
        site =E<gt> 'A',
        metrocluster_node =E<gt> $node-E<gt>name(),
        node =E<gt> $node-E<gt>name(),
        wait_for_completion =E<gt> 1,
);

=back

=back

=cut

sub unconfigure {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my ($self,@opts) = @_;

    my %opts;
    my ($component, $component_a, $component_b);
    my $hostobj;
    my $branch;
    my $site;

    if (ref $self) {
        #mcc object

        #validate opts
        %opts = validate_with(
            params => \@opts,
            spec   => {
                site => { regex => qr/^[AB]$/    },
                node => { type  => SCALAR | UNDEF, optional => 1 },
                wait_for_completion => { regex    => qr/^[01]$/, default => 1 },
                      },
                   );
        $site = $opts{site};
        $component_a = $self->_site_A_metrocluster_component();
        $component_b = $self->_site_B_metrocluster_component();

        $component = ($site =~ /A/) ?
                         $self->_site_A_metrocluster_component() :
                         $self->_site_B_metrocluster_component();

        $branch = $self->get_ontap_version_info->{branch};
    } else {
        #class method

        #validate opts
        %opts = validate_with(
            params => \@opts,
            spec   => {
                metrocluster_node   => { type  => SCALAR },
                node                => { type  => SCALAR | UNDEF,
                                         optional => 1 },
                site                => { regex => qr/^[AB]$/    },
                wait_for_completion => { regex => qr/^[01]$/, default => 1 },
                      },
                   );


        $site = $opts{site};

        my $bootstrap = WaflMCC::MCC::Metrocluster::_Bootstrap_Util->new(
                            node_names => [$opts{metrocluster_node}]);

        $component_a = $bootstrap->get_site_A_Metrocluster_component();
        $component_b = $bootstrap->get_site_B_Metrocluster_component();
        $component = ($site =~ /A/) ?
                         $bootstrap->get_site_A_Metrocluster_component() :
                         $bootstrap->get_site_B_Metrocluster_component() ;

        $hostobj = Tharn::host($opts{metrocluster_node});

        $branch = $component->command_interface()->get_version_manager()->
                                        get_version_attribute(
                                                    attribute => 'branch');

    }

    #for APISet
    $hostobj =
        $component->{command_interface}->{version_manager}->{hostobj};

    my $node = $opts{node};

        my $apiset = NACL::APISet->new(
             hostobj   => $hostobj, #Tharn::host($cluster_name),
             category  => 'Node',
             interface => 'CLI',
             set       => 'CMode'
         );


    my %node;
    if ($node) {
        $node{node} = $node;
    }


    my $run;

    try {
        $run =  $apiset->metrocluster_unconfigure(
                      'privilege-level' => 'test',
                      %node,


                  );
    } catch NATE::BaseException with {
        my $exception = shift;
        print Dumper $exception;
        my $text = $exception->text();
        print "caught wipeconfig/unconfigure:$text";
        $run =  $apiset->metrocluster_wipeconfig(
                      'privilege-level' => 'test',
                      %node,


                  );

    }
    $Log->comment(Dumper $run);
    my $job_id = $run->get_job_ids();
    print "job id: $job_id\n";


    my %job_key;
    #my %job_history_key; #TODO add history checks as backup?

    if ($job_id) {
      $job_key{id} = $job_id;
      #$job_history_key{id} = $job_id;
    } else {
        $job_key{name} = '"Metrocluster Unconfigure"';
        $job_key{'state'} = 'Success';

        #$job_history_key{name} = '"Metrocluster Unconfigure"';
        #$job_history_key{'event-type'} = 'Succeeded';
    }




    #
    #then optionally check for completion (check metrocluster show)
    #

    my $poll_time = 10;
    my $wait_time = 60*10;

    my $start_time = time;

    my $finished = 0;
    my $show_finished = 0;
    my $job_finished = 0;

    while ($opts{wait_for_completion} &&
           !$finished &&
           ((($start_time - time + $wait_time) > 0)) ) {
        my $loop_time = time;
        my @mcc_show;
        my $job_show;

        if (!$job_finished) {
            try {
                $job_show = NACL::CS::Job->fetch(
                    command_interface => $component->command_interface(),
                    filter => { %job_key },
                );
                $Log->comment(WaflMCC::MCC::Utils->dump(target=>$job_show));
                $job_finished = 1;
            } catch NACL::Exceptions::NoElementsFound with {

            };
        }

        if (!$show_finished) {
            try {
                @mcc_show = NACL::CS::Metrocluster->fetch(command_interface =>
                           $component->command_interface());
            } catch NACL::Exceptions::NoElementsFound with {
                $show_finished = 1;
            };
        }

        my $sleep = $poll_time - time + $loop_time;
        if (!$show_finished && !$job_finished && ($sleep > 0)) {
            foreach my $mcc_show (@mcc_show) {
                $Log->comment(WaflMCC::MCC::Utils->dump(target=>$mcc_show));
            }
            $Log->comment("sleeping $sleep till next poll...");
            sleep $sleep;
        }
        $finished = $show_finished && $job_finished;
    }


    #additional unconfigure checks to make sure unconfigure is operating
    #as expected
    #implicitly, this is only (potentially) run if we are waiting for completion

    if ($finished) {
        $Log->comment("unconfigure finished (job finished and metrocluster " .
            "show empty.)  Running additional internal checks...");

        my $extra_checks_fail = 0;


        my $cmd = 'set test;debug smdb table mccDRCOperation show';
        #should be: This table is currently empty.\r\n
        my $output =
            $component_a->command_interface()->apiset()->execute_command(
                command => $cmd );
        if ($output !~ /table is currently empty/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));
        $output = $component_b->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /table is currently empty/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));

        $cmd = 'set test;vserver peer show';
        #should be: There are no Vserver peering relationships
        $output = $component_a->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /There are no Vserver peer relationships/i) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));
        $output = $component_b->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /There are no Vserver peer relationships/i) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));

        $cmd = 'set test;debug crs stream show';
        #should be: This table is currently empty.\r\n
        $output = $component_a->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /table is currently empty/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));
        $output = $component_b->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /table is currently empty/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - table NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));

        #If you created servers, 'vserver show' should have
        #deleted sync-destination servers (not sync-source)
        #
        $cmd = 'set test;vserver show -subtype sync-destination';
        #should be: There are no entries matching your query
        $output = $component_a->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /are no entries matching your query/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - entries NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));
        $output = $component_b->command_interface()->apiset()->execute_command(
            command => $cmd );
        if ($output !~ /are no entries matching your query/) {
            $extra_checks_fail++;
            $Log->comment("$cmd failure - entries NOT empty as expected");
        }
        $Log->comment(WaflMCC::MCC::Utils->dump(target => $output));

        if ($extra_checks_fail) {

            WaflMCC::MCC::Exceptions::BaseException->throw(
                "unconfigure appeared to complete BUT failed $extra_checks_fail"
                . "  additional check(s)");
        }

    }

     if (!$finished && $opts{wait_for_completion}) {

         WaflMCC::MCC::Exceptions::BaseException->throw(
             "Did not finish unconfigure after $wait_time seconds");
     }

    return $run->get_parsed_output();
}

=head2 menu()

wipeconfig is deprecated, use unconfigure instead

=cut

sub wipeconfig {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;
    $Log->comment("wipeconfig is deprecated, please use unconfigure");
    $self->unconfigure(@_);

}
=head2 menu()

B<EXPERIMENTAL!>

This will be removed at some point to be replaced with a different structure
(put the same/similar functionality).

Displays an interactive cli utility menu to call common library functionality.


=cut

sub menu {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $exit = 0;

    my @macro;
    my $running_macro = 0;
    my $recording_macro = 0;
    my $macro_filename = '';
    my $macro_play_filename = '';
    my @macro_commands;
    my $wait_on_comment = 1;

    my $prompt      = '>';
    my $prompt_info = 'mcc';
    my $file        = 'mccmenu.macro';

#TBD: throughout, dr_state needs to be determined dynamically
    my $dr_state='normal'; #should query this I guess...

    my $setting;

    my $command = {

        'dr_1normal'     => { name => 'cycle to dr state: normal',
                              code => sub {
my $local_setting = shift;
$self->get_cluster(site=> $local_setting->{'survivor-site'})->cycle_to_DR_state(
                      state  => 'normal',
                      verify => { disks => $local_setting->{'verify-disks'} });
$dr_state = 'normal';
                                      },
                            },

        'dr_2switchover' => { name => 'cycle to dr state: switchover_completed',
                              code => sub {
my $local_setting = shift;
$self->get_cluster(site=> $local_setting->{'survivor-site'})->cycle_to_DR_state(
                      state  => 'switchover_completed',
                      verify => { disks => $local_setting->{'verify-disks'} });
$dr_state = 'switchover_completed';
                                          },
                            },

        'dr_3heal_aggrs' => { name => 'cycle to dr state: heal_aggrs_completed',
                              code => sub {
my $local_setting = shift;
$self->get_cluster(site=> $local_setting->{'survivor-site'})->cycle_to_DR_state(
                      state  => 'heal_aggrs_completed',
                      verify => { disks => $local_setting->{'verify-disks'} });
$dr_state = 'heal_aggrs_completed';
                                          },
                            },

        'dr_4heal_roots' => { name => 'cycle to dr state: heal_roots_completed',
                              code => sub {
my $local_setting = shift;
$self->get_cluster(site=> $local_setting->{'survivor-site'})->cycle_to_DR_state(
                      state  => 'heal_roots_completed',
                      verify => { disks => $local_setting->{'verify-disks'} });
$dr_state = 'heal_roots_completed';
                                          },
                            },

        'dump_mcc'       => { name => 'dump metrocluster object',
                              code => sub {
my $local_setting = shift;
$Log->comment(WaflMCC::MCC::Utils->dump(
                     target              => $self,
                     show_shared_storage => $local_setting->{'verify-disks'}));
                                          }
                            },

        'disk_verify'    => { name => 'do disk verification',
                              code => sub {
foreach my $dr (sort keys %{$self->get_DR_groups()}) {
    my $message = "validate shared storage in dr group $dr";
    $Log->comment($message);
    print ntest::NTEST_STDOUT "$message\n";

    $self->get_DR_groups()->{$dr}->get_shared_storage()->validate_disk_states();
}
                                          },
                            },

        'show_disasters' => { name => 'show DR and HA states',
                              code => sub {
try {
    my $disaster_status = $self->get_disaster_status(include_HA => 1);
    my $status_message = "overall disaster status:\n" .
                         WaflMCC::MCC::Utils->dump($disaster_status);
    $Log->comment($status_message);
}
catch WaflMCC::MCC::Exceptions::BaseException with {
    my $exception = shift;
    my $text = $exception->text();

    my $message = sprintf("%s", <<"END_PRINT");

Exception occurred during disaster status queries:

$text

This may be due to the DR or HA state currently transitioning, or it may be due
to an invalid DR or HA state.  You may safely repeat this query.  If this
query continues to fail in this manner for several minutes, it is likely that
the latter case is the reason for this failure, and you will need to triage
the MCC system manually.

END_PRINT

    $Log->comment($message);
};
                                          },
                            },
                  };

    my $setting_spec = {
                   'survivor-site' => { default => 'A', regex => qr/^[AB]$/},
                   'verify-disks'  => { default => 1, regex => qr/^[01]$/ },
                       };

    $setting = validate_with(
        params => $setting || {},
        spec   => $setting_spec,
    );

    my @history;

    my $fiji;
    my $fijifile = 'mcc.fiji';
    if (-e $fijifile) {

        if (open (FIJI,'<',$fijifile)) {
          while (my $line = <FIJI>) {
              my ($name,$filter) = split (/=/,$line,2);
              chomp $filter;
              $fiji->{$name}->{fiji} = $filter;
          }
          close FIJI;
          print ntest::NTEST_STDOUT "read fiji file:" . Dumper $fiji;
        }

    }
    while (!$exit) {

        print ntest::NTEST_STDOUT Dumper $setting;
        #do this better, this is a quick test here
        my @order;
        my %itemorder;
        my $count = 1;
        foreach my $item (sort keys %$command) {
            $order[$count] = $item;
            $itemorder{$item} = $count;
            print ntest::NTEST_STDOUT "$count: $command->{$item}->{name}\n";
            $count++;
        }

        print ntest::NTEST_STDOUT "f: fiji\n";
        print ntest::NTEST_STDOUT "l: get mcc filer logs\n";
        print ntest::NTEST_STDOUT
                                "m: " . ($recording_macro ? "stop" : "start") .
                                " recording macro" .
             ($recording_macro ? " [$macro_filename]\n" : "\n");
        print ntest::NTEST_STDOUT "p: play macro" .
                         ($running_macro ? " [$macro_play_filename]\n" : "\n");
        print ntest::NTEST_STDOUT "r: refresh command interfaces\n";
        print ntest::NTEST_STDOUT "q: quit\n";
        print ntest::NTEST_STDOUT "#: add comment to log\n";
        print ntest::NTEST_STDOUT "$prompt_info$prompt";

        my $in;
        if ($running_macro) {
            if (my $next = shift @macro) {
                chomp $next;
                if (defined $itemorder{$next}) {
                    $in = $itemorder{$next};
                } else {
                    if ($next =~ /^#/) {
                        print ntest::NTEST_STDOUT "$next\n";
                        if ($wait_on_comment) {
                          print ntest::NTEST_STDOUT "hit return to continue\n";
                          my $wait_in = <STDIN>;
                        }
                    }
                    elsif ($next eq 'm') {
                        #we aren't currently supporting recording macros during a macro play!
                        print ntest::NTEST_STDOUT "Did not execute record (m)acro command in macro\n";
                    }
                    elsif ($next eq 'q') {
                        $in = ''; #we don't want to quit do we?  or do we...
                        print ntest::NTEST_STDOUT "Did not execute (q)uit command in macro\n";

                    } else {
                        $in = $next;
                    }
                }

                print ntest::NTEST_STDOUT "macro command:$in\n";
            } else {
                print ntest::NTEST_STDOUT "macro completed\n";
                $running_macro = 0;
            }

        } else {
            $in = <STDIN>;
            chop $in;
            push @history,($in !~ /\D/ ? $order[$in] || $in : $in);
            push @macro_commands,$history[-1];
        }

        my @set = split /=/,$in,2; #if it's a set...

        if (lc $in eq 'q') {
            if ($dr_state eq 'normal') { #should query this...
                $exit = 1;
            } else {
                print ntest::NTEST_STDOUT "WARNING! Metrocluster NOT in normal state.  Switch to normal before quitting.\n";
            }
        } elsif ($in =~ /^#/) {
            $Log->comment("comment:$in");
        } elsif ($in eq 'c') {
            eval {$fiji->clear() if $fiji;};
        }
        elsif ($in eq 'f') {
            #fiji
            my $ci;
            if ($setting->{'survivor-site'} eq 'A') {
                $ci = $self->_site_A_metrocluster_component()->command_interface();
            } else {
                $ci = $self->_site_B_metrocluster_component()->command_interface();
            }

            my $fcount=1;
            my @forder;
            my %fitemorder;

            foreach my $item (sort keys %{$fiji}) {
                my $status = $fiji->{$item}->{status} || '-';
                print ntest::NTEST_STDOUT "$fcount: [$status] $item" .
                  ($item eq $fiji->{$item}->{fiji} ?
                                         "\n" : " [$fiji->{$item}->{fiji}]\n");
                  #($item eq $fiji->{$item}->{object}->{program} ?
                  #         "\n" : " [$fiji->{$item}->{object}->{program}]\n");
                $forder[$fcount] = $item;
                $fitemorder{$item} = $fcount;
                $fcount++;
            }



            print ntest::NTEST_STDOUT "c: clear fiji\n";
            #print ntest::NTEST_STDOUT "d: delete fiji\n"; #TODO
            print ntest::NTEST_STDOUT "n: new fiji\n";
            print ntest::NTEST_STDOUT "s: set fiji\n";
            print ntest::NTEST_STDOUT "<hit return to exit to main menu>\n";
            print ntest::NTEST_STDOUT "fiji$prompt";

            my $fin;

            if ($running_macro) {
                if (my $next = shift @macro) {
                    chomp $next;
                    if (defined $fitemorder{$next}) {
                        $fin = $fitemorder{$next};
                    } else {
                        if ($next =~ /^#/) {
                            print ntest::NTEST_STDOUT "$next\n";
                            if ($wait_on_comment) {
                              print ntest::NTEST_STDOUT "hit return to continue\n";
                              my $wait_in = <STDIN>;
                            }
                        } else {
                            $fin = $next;
                        }
                    }
                    print ntest::NTEST_STDOUT "macro command:$fin\n";
                }
            } else {
                $fin = <STDIN>;

                chop $fin;
                $fin = lc $fin;

                push @history,($fin !~ /\D/ ? $forder[$fin] || $fin : $fin);
                push @macro_commands,$history[-1];
            }

            if ($fin eq 'c') {
                print ntest::NTEST_STDOUT "fiji # to clear$prompt";
                my $number;
                if ($running_macro) {
                    if (my $next = shift @macro) {
                        chomp $next;
                        if (defined $fitemorder{$next}) {
                            $number = $fitemorder{$next};
                        } else {
                            if ($next =~ /^#/) {
                                print ntest::NTEST_STDOUT "$next\n";
                                if ($wait_on_comment) {
                                  print ntest::NTEST_STDOUT "hit return to continue\n";
                                  my $wait_in = <STDIN>;
                                }
                            } else {
                                $number = $next;
                            }
                        }
                        print ntest::NTEST_STDOUT "macro command:$number\n";
                    }
                } else {
                    $number = <STDIN>;

                    chop $number;
                    $number = lc $number;

                    push @history,($number !~ /\D/ ? $forder[$number] || $number : $number);
                    push @macro_commands,$history[-1];
                }
                #my $number = <STDIN>;
                #chop $number;
                my $item = $forder[$number] || next;
                eval {
                    print ntest::NTEST_STDOUT "fiji state?\n";
                    my $fs = $fiji->{$item}->{object}->state();
                    print ntest::NTEST_STDOUT "fiji state:" . Dumper $fs;
                };
                if ($fiji->{$item}->{status} eq 's') {
                    eval {
                        $fiji->{$item}->{object}->clear();
                        $fiji->{$item}->{status} = '';
                    };
                } else {
                    print ntest::NTEST_STDOUT "fiji $item not set, can't clear\n";
                }

            } elsif ($fin eq 'n') {
                print ntest::NTEST_STDOUT "fiji filter$prompt";
                my $fiji_program = <STDIN>;
                chop $fiji_program;
                print ntest::NTEST_STDOUT "fiji name/tag$prompt";
                my $fiji_tag = <STDIN>;
                chop $fiji_tag;
                $fiji_tag ||= $fiji_program;


                my $fijiobj = NACL::STask::Fiji->new(
                             command_interface => $ci,
                             tag               => $fiji_tag,
                             program           => $fiji_program,
                           );
                $fiji->{$fiji_tag}->{fiji} = $fiji_program;
                $fiji->{$fiji_tag}->{object} = $fijiobj;

                if (1) { #write fijis to file

                    if (open (FIJI,'>>',$fijifile)) {
                        my $filetag = $fiji_tag;
                        $filetag =~ s/=/_/g; #?? not sure if = is valid token
                        print FIJI "$filetag=$fiji_program\n";
                        close FIJI;
                    }
                }


                #print ntest::NTEST_STDOUT Dumper $fiji;
            } elsif ($fin eq 's') {
                print ntest::NTEST_STDOUT "fiji # to set$prompt";
                my $number;
                if ($running_macro) {
                    if (my $next = shift @macro) {
                        chomp $next;
                        if (defined $fitemorder{$next}) {
                            $number = $fitemorder{$next};
                        } else {
                            if ($next =~ /^#/) {
                                print ntest::NTEST_STDOUT "$next\n";
                                if ($wait_on_comment) {
                                  print ntest::NTEST_STDOUT "hit return to continue\n";
                                  my $wait_in = <STDIN>;
                                }
                            } else {
                                $number = $next;
                            }
                        }
                        print ntest::NTEST_STDOUT "macro command:$number\n";
                    }
                } else {
                    $number = <STDIN>;

                    chop $number;
                    $number = lc $number;

                    push @history,($number !~ /\D/ ? $forder[$number] || $number : $number);
                    push @macro_commands,$history[-1];
                }
                #my $number = <STDIN>;
                #chop $number;
                my $item = $forder[$number] || next;
                unless ($fiji->{$item}->{object}) {
                    $fiji->{$item}->{object} = NACL::STask::Fiji->new(
                                 command_interface => $ci,
                                 tag               => $item,
                                 program           => $fiji->{$item}->{fiji},
                               );

                }
                if ($fiji->{$item}->{status} ne 's') {
                    eval {
                        $fiji->{$item}->{object}->set();
                        $fiji->{$item}->{status} = 's';

                    };
                } else {
                    print ntest::NTEST_STDOUT "fiji $item not set, can't clear\n";
                }


            } else {
                #exit
            }


        } elsif ($in eq 'm') {
            if ($recording_macro) {

                print ntest::NTEST_STDOUT "\nAre you sure you want to stop recording macro $macro_filename? (y/[n]) ";

                my $verify = <STDIN>;
                if ($verify =~ /y/) {
                    print ntest::NTEST_STDOUT "macro filename (return to cancel)? ";

                    if (open (MACROFILE,'>>',$macro_filename.'.macro')) {
                        print MACROFILE join "\n",@macro_commands;
                        close MACROFILE;
                        $recording_macro = 0;
                        $macro_filename = '';
                    } else {
                        print ntest::NTEST_STDOUT "warning, could not open macro $file\n";
                    }
                }
            } else {
                print ntest::NTEST_STDOUT "macro filename (return to cancel)? ";
                $macro_filename = <STDIN>;
                chop $macro_filename;
                next if $macro_filename eq '';
                print ntest::NTEST_STDOUT "Recording macro $macro_filename\n";

                if (open (MACROFILE,'>',$macro_filename.'.macro')) {
                    print MACROFILE "version=1\n";

                    close MACROFILE;
                    $recording_macro = 1;
                    @macro_commands=(); #start over
                } else {
                    print ntest::NTEST_STDOUT "warning, could not open macro $macro_filename\n";
                }
            }

        } elsif ($in eq 'p') {
            print ntest::NTEST_STDOUT "macro filename (return to cancel)? ";
            my $macro_filename = <STDIN>;
            chop $macro_filename;
            next if $macro_filename eq '';
            $macro_filename .= '.macro' unless $macro_filename =~ /\.macro$/;
            if (open (FILE,'<',$macro_filename)) {
                my @macro = <FILE>;
                close FILE;
                print ntest::NTEST_STDOUT @macro;
            }
            print ntest::NTEST_STDOUT "Are you sure you want to play macro $macro_filename? (y/n) ";
            my $verify = <STDIN>;
            if ($verify =~ /y/) {
                $Log->comment("Replaying macro $file");
                print ntest::NTEST_STDOUT "Replaying macro $macro_filename\n";

                if (open (FILE,'<',$macro_filename)) {
                    @macro = <FILE>;
                    close FILE;
                    $running_macro = 1;
                    ($macro_play_filename = $macro_filename) =~ s/\.macro//;
                } else {
                    print ntest::NTEST_STDOUT "warning, could not open macro $macro_filename\n";
                }

            }
        } elsif ($in eq 'r') {
            $self->refresh_command_interfaces();
        } elsif ($in eq 'l') {
            ## no critic (InputOutput::ProhibitExplicitStdin)
            print ntest::NTEST_STDOUT
                "enter burt # <or leave blank to write logs to logdir>\n";
            my $submenu = 'log';
            print ntest::NTEST_STDOUT "$submenu$prompt";

            my ($burt) = <STDIN> =~ /(\d*)/;
            my $path = q{};

            if ($burt) {
                print ntest::NTEST_STDOUT "BURT: $burt\n";
                my $ci;
                if ($setting->{'survivor-site'} eq 'A') {
                    $ci = $self->_site_A_metrocluster_component()->
                                 command_interface();
                } else {
                    $ci = $self->_site_B_metrocluster_component()->
                                 command_interface();
                }

                #figure out the logs path
                my $show = NACL::CS::SystemNode->fetch(
                    command_interface => $ci,
                    requested_fields  => [ 'location' ] );
                my $location = $show->{location} || 'rtp';
                print ntest::NTEST_STDOUT "filer location:$location\n";

                $path = "/x/eng/$location-cores/burt.$burt";
            } else {
                $path = Tharn::param('LOGDIR') . '/logs';
            }
            print ntest::NTEST_STDOUT "log destination path:$path\n";

            print ntest::NTEST_STDOUT "p: paloma info\n";
            print ntest::NTEST_STDOUT
                's: controller scraper (adds san logs, core files, more -' .
                " can be SLOW -- 30 min)\n";
            print ntest::NTEST_STDOUT "<hit return to exit to main menu>\n";
            print ntest::NTEST_STDOUT "$submenu$prompt";
            my ($script) =  <STDIN> =~ /([p|s])/i;

            if ($script) {
                my $eval_return = eval {
                    $self->_get_logs( nodes     => $self->get_all_mcc_nodes(),
                                      log_path  => $path,
                                      use_tool  => lc $script  );
                    1;
                };
                unless ($eval_return) { print ntest::NTEST_STDOUT $@; };
            }
        } elsif ($in eq '') {
            #empty
        } elsif (defined $command->{$order[$in]}) {
            my $selected = $order[$in];
            if (defined $command->{$selected}) { # TODO REMOVE LINE
                my $commandref = $command->{$selected};
                print ntest::NTEST_STDOUT
                    "execute command: $commandref->{name} [$selected]\n";
                $Log->comment("execute command: $selected");
                if (defined $commandref->{code}) {

                    print ntest::NTEST_STDOUT "execute code: $selected\n";
                    eval {
                        #time just this menu call, whatever it is
                        my $trace_menu =
                            WaflMCC::MCC::Log::TRACELOG->
                            start_timer(
                                message => "menu command $selected"); #metrics
                        $commandref->{code}->($setting);
                    };
                    if ($@) { print ntest::NTEST_STDOUT $@; }
                }
            }

        } elsif (defined $setting->{$set[0]}) {
            print ntest::NTEST_STDOUT "set value: $in\n";
            if ($set[1] =~ /$setting_spec->{$set[0]}->{regex}/) {
                $setting->{$set[0]} = $set[1];
            } else {
                print ntest::NTEST_STDOUT "$set[1] not a valid value for $set[0]\n";
            }


        } else { #???
            print ntest::NTEST_STDOUT "unknown command: $in\n";
        }



    }

    #move this to a _method
    $Log->comment("command history: " . join ',',@history);
    if (open (FILE,'>',$file)) {
        print FILE "version=1\n";
        print FILE join "\n", @history;
        close FILE;
    } else {
        $Log->comment(
                "warning, could not open $file to save command history macro");
    }

    return;
}

sub _get_logs {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %raw_opts = @_;

    # validate opts
    my %opts = validate_with(
        params => \%raw_opts,
        spec   => {
            nodes            => { type => ARRAYREF },
            paloma_info_path => { type => SCALAR,
                                  default =>
                                      '/usr/software/test/bin/paloma_info.pl',
                                },
            use_tool         => { type => SCALAR, default => 'p' },
            scraper_path     => { type => SCALAR,
                                  default =>
                                      '/usr/software/test/agnostic/san/' .
                                      'fit/ControllerScraper.py',
                                },
            log_path         => { type => SCALAR,
                                  default => Tharn::param('LOGDIR') . '/logs'
                                },
            compress         => { regex => qr/^[01]$/, default => 0 },
            admin_password   => { type => SCALAR | UNDEF, optional => 1 },
            diag_password    => { type => SCALAR | UNDEF, optional => 1 },
                  },
        );

    my $nodes = delete $opts{nodes};

    # Start log pulling sessions in parallel
    my @subtest_handles = ();
    foreach my $node (@{ $nodes }) {
        my $node_name = $node->get_name();
        $Log->comment("Fetching logs from : $node_name");

        my $cmd;

        if ($opts{use_tool} eq 's') {
            my $toolname = 'Controller Scraper';
            print ntest::NTEST_STDOUT
                "Writing logs from $node_name to $opts{log_path} " .
                "with $toolname\n";
            $cmd = "$opts{scraper_path}";
            #$cmd .= ' -z' if $opts{compress}; #not an option
            $cmd .= ' -v';
            $cmd .= " -y $opts{admin_password}" if $opts{admin_password};
            $cmd .= " -z $opts{diag_password}" if $opts{diag_password};
            $cmd .= " -t $opts{log_path}" if $opts{log_path};
            $cmd .= " -n $node_name"; #or we can pass comma sep list??
        } else { #default to paloma
            my $toolname = 'paloma info';
            print ntest::NTEST_STDOUT
                "Writing logs from $node_name to $opts{log_path} " .
                "with $toolname\n";
            $cmd = "$opts{paloma_info_path}";
            $cmd .= ' -z' if $opts{compress};
            $cmd .= " -a $opts{admin_password}" if $opts{admin_password};
            $cmd .= " -d $opts{diag_password}" if $opts{diag_password};
            $cmd .= " -p $opts{log_path}" if $opts{log_path};
            $cmd .= " -f $node_name"; #must be LAST (well really just node)
        }

        print "$cmd\n";
        ## no critic (Modules::ProhibitConditionalUseStatements)
        push @subtest_handles, subtest(
            sub {
                use NACL::C::Client;
                use Net::Domain qw(hostname);
                # Execute the command on client running this script
                my $client_hostname = hostname()
                        or croak 'Unable to get client hostname';
                $Log->comment("Using client: $client_hostname");

                my $client = NACL::C::Client->new( name => $client_hostname );
                $client->apiset->set_timeout( 'connectrec-timeout' => 3600 );
                $client->execute( command => $cmd );
            },
            '-bg',
            '-runid', $node_name . '_log_extraction' );
    }
    Subtest::wait_finish( subtests => \@subtest_handles );

    return;


}

# just logs some information about the MCC system
sub _log_some_mcc_system_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status => { type => HASHREF },
            constructor_log => { type => SCALAR },
                  },
                            );

    my $disaster_status = delete($opts{disaster_status});
    my $constructor_log = delete($opts{constructor_log});

    my $dr_state_text = $disaster_status->{DR_state};
    ($disaster_status->{DR_state} ne 'normal') && ( $dr_state_text .=
                     ', survivor site: ' . $disaster_status->{survivor_site} );

    $Log->comment('Your Metrocluster system:');
    $Log->comment('DR State: ' . $dr_state_text);
    foreach my $dr_group (@{ $self->_dr_groups() }) {
        my $dr_group_id = $dr_group->get_id();

        $Log->comment('  DR Group ID: ' . $dr_group_id);

        $Log->comment(
              '    DR Group Node A1: ' . $dr_group->get_A1_node()->get_name());

        $Log->comment('    DR Partners: ( Site A  <===>  Site B )');
        my $nodes_by_role = $dr_group->get_nodes_by_role();
        $Log->comment(q/            / . $nodes_by_role->{A1}->get_name() .
                          ' <===> ' .  $nodes_by_role->{B1}->get_name());
        # skip for 2-node systems
        if (exists($nodes_by_role->{A2})) {
            $Log->comment(q/            / . $nodes_by_role->{A2}->get_name() .
                          ' <===> ' .  $nodes_by_role->{B2}->get_name());

            $Log->comment('    HA States: ' . WaflMCC::MCC::Utils->dump(
                               $disaster_status->{HA_states}->{$dr_group_id}));

            $Log->comment('    Shared Storage Intersite Fabrics: ');

            try {
                my $topology = $dr_group->get_network_topology();
                foreach my $fabric (@{ $topology->get_intersite_fabrics() }) {
                    $Log->redirect(path => $constructor_log);
                    my $site_A_switch_desc =
                           $fabric->get_switch(site => 'A')->get_description();
                    my $site_B_switch_desc =
                           $fabric->get_switch(site => 'B')->get_description();
                    $Log->reset_path();
                    $Log->comment(q/            / .
                        $site_A_switch_desc . ' <===> ' . $site_B_switch_desc);
                }
            } catch WaflMCC::MCC::Exceptions::BaseException with {
                my $exception = shift;

                $Log->comment(q{        } .
                     'Shared storage network topology discovery not enabled.');
            }
        }

        $Log->comment("\n");
    }

    return;
}

# validate passed parameters.  throw exception if user passed in a
# strange/invalid set of values in nodes.  (e.g., if there
# are nodes listed not in the same MCC system, etc.)
sub _validate_node_names_passed_by_user {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_names    => { type => ARRAYREF },
                  },
                            );

    my $node_names = delete($opts{node_names});

    my %check_list = map { $_ => 1 } @{ $node_names };
    # complain if any nodes in this list are not in one of the DR Groups
    foreach my $dr_group (values(%{ $self->get_DR_groups() })) {
        my $group_nodes = $dr_group->get_mcc_nodes();
        foreach my $mcc_node (@{ $group_nodes }) {
            if (exists($check_list{$mcc_node->get_name()})) {
                delete($check_list{$mcc_node->get_name()});
            }
        }
    }

    # anything left in list is a problem
    if (scalar(keys(%check_list))) {
        my $msg = "The following nodes passed in the nodes parameter do not\n" .
                  "    belong to any DR Group in this MCC system\n\n:" .
                  "        " . join("\n        ", keys(%check_list)) . "\n\n" .
                  "    Throwing an exception to prevent possible use of\n" .
                  "    incorrect system...";
        WaflMCC::MCC::Exceptions::BaseException->throw($msg);
    }

    return;
}

# END PACKAGE WaflMCC::MCC::Metrocluster

# private utility class for WaflMCC::MCC::Metrocluster
package WaflMCC::MCC::Metrocluster::_Bootstrap_Util;

use NATE::BaseException qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);

use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use Data::Dumper;

use Class::MethodMaker [
     new    => [ '-init', 'new' ],
     scalar => [
         { -type => 'NACL::C::SystemNode' },
         qw/ _temp_utility_site_A_node_component /,
               ],
     scalar => [
         qw/ _temp_cluster_A_name _original_node_names _A1_node_names /,
               ],
     scalar => [
         { -type => 'NACL::C::Metrocluster' },
         qw/ _site_A_metrocluster_component
             _site_B_metrocluster_component /,
               ],
                       ];

# implicitly called by Class::MethodMaker's new (see -init)
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      node_names => { type => ARRAYREF },
                      pick_nodes => { type => SCALAR , default => 0, optional => 1,},
                  },
                            );

    my $node_names = delete($opts{node_names});
    my $pick_nodes = delete($opts{pick_nodes});

    # hold onto these for a bit
    $self->_original_node_names($node_names);
    my $A1_node_name;

    if ($pick_nodes) {
        # randomly pick a node to define as one (or the) A1 node
        $A1_node_name = $node_names->[0];
    } else {
        # mica : statically pick nodes to define as one (or the) A1 node
        #$A1_node_name = $node_names->[0];
        $A1_node_name = $node_names->[int(rand(scalar(@{ $node_names })))];
    } 

    # this may contain more after more discovery is done...
    $self->_A1_node_names( [ $A1_node_name ] );

    # this is just to get a bootstrap CI
    my $host = Tharn::host($A1_node_name);

    $self->_temp_utility_site_A_node_component(
                                   NACL::C::SystemNode->new(hostrec => $host));

    unless ($self->_temp_utility_site_A_node_component_isset()) {
        my $msg = "Can't get a valid node component CI for discovery!";
        WaflMCC::MCC::Exceptions::BaseException->throw($msg);
    }
}

sub get_A1_node_names {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $Metrocluster_component_A = $self->get_site_A_Metrocluster_component();
    my @mcc_state_responses = NACL::CS::Metrocluster->fetch(
           command_interface => $Metrocluster_component_A->command_interface(),
           filter            => { cluster => $self->_temp_cluster_A_name() },
           requested_fields  => [ 'node', 'ha-partner' ]);

    my @discovered_A1s = @{ $self->_A1_node_names() };
    my @other_dr_group_node_names = ( );
    foreach my $state (@mcc_state_responses) {
        my $node_name  = $state->node();
        my $ha_partner = $state->ha_partner();
        if (List::MoreUtils::any { /^($node_name|$ha_partner)$/ }
                                                             @discovered_A1s) {
            # ignore these.  already have A1 for them.
        } else {
            # but these must be from other dr group(s).
            # check if these are in original node name list passed in
            my @choices_for_A1 = grep { /^($node_name|$ha_partner)$/ }
                                            @{ $self->_original_node_names() };
            # if no matches, put both in...
            unless (scalar(@choices_for_A1)) {
                @choices_for_A1 = ( $node_name, $ha_partner );
            }

            # select A1 randomly amongst choices (may be only 1)
            my $A1_node_name =
                           $choices_for_A1[int(rand(scalar(@choices_for_A1)))];

            # update discovered A1 node names
            push(@discovered_A1s, $A1_node_name);
        }
    }

    return \@discovered_A1s;
}

sub get_site_A_Metrocluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    unless ($self->_site_A_metrocluster_component_isset()) {

        my $node_component = $self->_temp_utility_site_A_node_component();
        my $ClusterIdentity_state = NACL::CS::ClusterIdentity->fetch(
                                          command_interface => $node_component,
                                          requested_fields  => [ 'name' ] );
        my $cluster_A_name = $ClusterIdentity_state->name();

        # cannot rely on a cluster nate_host file being available, so create
        # cluster mgmt hostrec from info in node nate_host file
        my $node_hostrec    = $node_component->{hostrec};
        my $cluster_hostrec = Hostrec->new(
            id               => $cluster_A_name,
            hosttype         => $node_hostrec->{'hosttype'},
            hostname         => $cluster_A_name,
            default_ip       => $node_hostrec->{'CLUSTER_MGMT_IP'},
            default_ip_gx    => $node_hostrec->{'CLUSTER_MGMT_IP'},
            default_username => $node_hostrec->{'default_username'} || q{admin},
            default_password => $node_hostrec->{'default_password'} ||
                                                                   q{netapp1!},
                                          );
        # create the cluster component using this hostrec object
        my $site_A_cluster_component =
                            NACL::C::Cluster->new(hostrec => $cluster_hostrec);

        $self->_site_A_metrocluster_component(NACL::C::Metrocluster->new(
                              command_interface => $site_A_cluster_component));

        # save for finding out about site B
        $self->_temp_cluster_A_name($cluster_A_name);
    }

    return $self->_site_A_metrocluster_component();
}

sub get_site_B_Metrocluster_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    unless ($self->_site_B_metrocluster_component_isset()) {
        # find an arbitrary node at site B
        my $site_A_mcc_comp = $self->get_site_A_Metrocluster_component();
        my $ci = $site_A_mcc_comp->command_interface();
        my @responses = NACL::CS::Metrocluster->fetch(
                                command_interface => $ci,
                                requested_fields  => [ 'cluster', 'node' ]);
        my $node_at_B_name = undef;
        my $cluster_B_name = undef;
        foreach my $response (@responses) {
            my $cluster_name = $response->{cluster};
            if ($cluster_name ne $self->_temp_cluster_A_name()) {
                $cluster_B_name = $cluster_name;
                $node_at_B_name = $response->{node};
                last;
            }
        }

        my $node_hostrec = Tharn::host($node_at_B_name);
        my $cluster_hostrec = Hostrec->new(
            id               => $cluster_B_name,
            hosttype         => $node_hostrec->{'hosttype'},
            hostname         => $cluster_B_name,
            default_ip       => $node_hostrec->{'CLUSTER_MGMT_IP'},
            default_ip_gx    => $node_hostrec->{'CLUSTER_MGMT_IP'},
            default_username => $node_hostrec->{'default_username'} || q{admin},
            default_password => $node_hostrec->{'default_password'} ||
                                                                   q{netapp1!},
                                          );
        # create the cluster component using this hostrec object
        my $site_B_cluster_component =
                            NACL::C::Cluster->new(hostrec => $cluster_hostrec);

        $self->_site_B_metrocluster_component(NACL::C::Metrocluster->new(
                              command_interface => $site_B_cluster_component));
    }

    return $self->_site_B_metrocluster_component();
}

# END PACKAGE WaflMCC::MCC::Metrocluster::_Bootstrap_Util

1;
