# $Id $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Network::Node Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Network::Node;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use NACL::APISet::Response;

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use Readonly;
use Data::Dumper;


Readonly my $TO_BE_DETERMINED => undef;


=head1 NAME

WaflMCC::MCC::Network::Node

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.

This class helps to define the network hardware mapping of an MCC DR Group.  An
MCC::Network::Node (not to be confused with the more general MCC::Node) is the
interface to get network specific information and objects associated with a
specific node in the DR Group, for example a node's set of FCVI ports used
in conjunction with the node's interconnect with its DR Partner.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();
    my $network_node = $topology->get_network_nodes()->{'B'}->{'1'};

    my @fcvi_ports = @{ $network_node->get_fcvi_ports() };

    ...
    # Perform test here
    ...

    # End test

Also please see the diagrams at

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for further clarification.

=cut

=head1 Public METHODS

There is no public constructor for this class.  A
WaflMCC::MCC::Network::Node object is obtained via calls such as

    my $network_node = $topology->get_network_nodes()->{'B'}->{'1'};

=cut

# automatically generate a (Non-public) constructor (includes a call to 'init')
# and (Non-public) member accessor methods
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'WaflMCC::MCC::DR_Group' },
                '_dr_group',
              ],
    scalar => [
                { -type => 'WaflMCC::MCC::Node' },
                '_mcc_node',
              ],
    scalar => [
                { -type => 'NACL::APISet::Response' },
                '_init_sysconfig_va_response',
              ],
    scalar => [
        qw/ _dr_group_topology_role /,
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group => { isa   => 'WaflMCC::MCC::DR_Group' },
            mcc_node => { isa   => 'WaflMCC::MCC::Node' },
        },
                            );

    my $dr_group = delete($opts{'dr_group'});
    my $mcc_node = delete($opts{'mcc_node'});

    $self->_dr_group($dr_group);
    $self->_mcc_node($mcc_node);

    # will be (A1|A2|B1|B2)
    $self->_dr_group_topology_role($self->_determine_network_topology_role());

    # the local nodescope sysconfig -va results at initialization
    #
    # this call is taking way too long in some cases.  complain about it if so.
    my $start_time = time();
    $self->_init_sysconfig_va_response(
                $mcc_node->get_nodescope_apiset()->sysconfig(
                                               'verbose'              => '1',
                                               'detailed-information' => '1'));
    my $delta_t = time() - $start_time;
    if ($delta_t > 55) {
        my $msg = "'sysconfig -va' on node " . $mcc_node->get_name() .
                  " took $delta_t seconds!  Report to BURT 581107 if still " .
                  "occurring after July (?)";
        $Log->warn($msg);
        print "WARNING: $msg\n";
    }

    return;
}

# determine the role (A1|A2|B1|B2) of this node in the dr group network topology
sub _determine_network_topology_role {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $role = undef;
    my $A1_node = $self->_dr_group()->get_A1_node();
    if ($self->_mcc_node() == $A1_node) {
        $role = 'A1';
    } else {
        my @dr_partnerships = @{ $self->_dr_group()->get_DR_partnerships() };
        foreach my $partnership (@dr_partnerships) {
            if ($partnership->get_partner_node(site => 'A') == $A1_node) {
                if ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'B')) {
                    $role = 'B1';
                }
            } else {
                if ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'A')) {
                    $role = 'A2';
                } elsif ($self->_mcc_node() ==
                                 $partnership->get_partner_node(site => 'B')) {
                    $role = 'B2';
                }
            }
        }
    }
    unless (defined($role)) {
        my $msg = q{Could not determine node [} .
                  $self->_mcc_node()->get_name() . "]'s DR Group role\n" .
                  q{    (e.g., 'A1', 'A2', 'B1', or 'B2').};
        WaflMCC::MCC::Exceptions::BaseException->throw($msg);
    }

    return $role;
}

=head2 L_is_vsim()

Returns true (1) if this node is a node in a vsim 4-pack DR Group.  False
otherwise.

This is here because the sysconfig info is here.  Maybe it should be straight
off in Node instead of Network::Node.

=cut

sub L_is_vsim {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    # seems only 1 entry, but for good style...
    my $entries = $self->_init_sysconfig_va_response()->get_parsed_output();
    foreach my $entry (@{ $entries }) {
        if (exists($entry->{slot_information})) {
            my @slot_entries = @{ $entry->{slot_information} };
            foreach my $slot_entry (@slot_entries) {
                if (exists($slot_entry->{model_name}) &&
                                     ($slot_entry->{model_name} eq 'SIMBOX')) {
                    return 1;
                }
            }
        }
    }

    return 0;
}



=head2 get_role()

    my $node_role = $mcc_network_node->get_role();

Returns the node's role in its DR group ('A1'|'A2'|'B1'|'B2').  Here A1 and
B1 are DR partners, and A2 and B2 are DR partners.

=cut

sub get_role {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_dr_group_topology_role();
}

=head2 get_shared_storage_connection_info()

    my $network_node = $mcc_node->L_get_network_node_functionality();
    my @fcvi_connection_info = @{ $network_node->
                    get_shared_storage_connection_info()->{fcvi_connections} };

Returns a hashref of data with the following format:

{
    fcvi_connections => 
        [
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabY',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:90',
            },
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabX',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:91',
            },
            ...
        ],
    storage_connections =>
        [
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabY',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:90',
            },
            {
                switch_name     => 'dpg-mcc-brcd6510-01-A-fabX',
                port_descriptor => '0',
                port_wwn        => '21:00:00:24:ff:32:06:91',
            },
            ...
        ],
    ha_connections =>
        [
            # empty/undefined for now
        ],
}

=cut

sub get_shared_storage_connection_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $info = { fcvi_connections    => [ ],
                 storage_connections => [ ],
                 ha_connections      => [ ], };

    # seems only 1 entry, but for good style...
    my $entries = $self->_init_sysconfig_va_response()->get_parsed_output();
    foreach my $entry (@{ $entries }) {
        if (exists($entry->{slot_information})) {
            my @slot_entries = @{ $entry->{slot_information} };
            foreach my $slot_entry (@slot_entries) {
                if (defined($slot_entry->{device})) {
                    if (   ($slot_entry->{device} =~ /FCVI Host Adapter/) &&
                           ($slot_entry->{physical_link} eq 'UP')) {
                        my ($switch_name, $port_descriptor) =
                              ($slot_entry->{switch_port} =~ /^([^:]+):(.+)$/);
                        my $port_wwn = $slot_entry->{fc_node_name};
                        push(@{ $info->{fcvi_connections} }, {
                                           switch_name     => $switch_name,
                                           port_descriptor => $port_descriptor,
                                           port_wwn        => $port_wwn,
                                                             } );
                    } elsif ($slot_entry->{device} =~ /FC Host Adapter.+<UP>/) {
                        my ($switch_name, $port_descriptor) =
                              ($slot_entry->{switch_port} =~ /^([^:]+):(.+)$/);
                        my $port_wwn = $slot_entry->{fc_port_name};
                        push(@{ $info->{storage_connections} }, {
                                           switch_name     => $switch_name,
                                           port_descriptor => $port_descriptor,
                                           port_wwn        => $port_wwn,
                                                             } );
                    }
                }
            }
        }
    }

    return $info;
}

=head2 get_intercluster_lifs()

Not yet Implemented.

    my @ic_lifs = @{ $network_node->get_intercluster_lifs() };

Returns an arrayref of WaflMCC::MCC::Network::LIF objects associated with
intercluster lifs on this node.

=cut

sub get_intercluster_lifs {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $TO_BE_DETERMINED;
}

1;
