##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

package WaflMCC::MCC::Network::Switch;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Log ();

use NACL::APISet;
use NACL::APISet::Switch;

use WaflMCC::MCC::Network::Port;
use WaflMCC::MCC::Utils;

use Tharn;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);

use Data::Dumper;


=head1 NAME

WaflMCC::MCC::Network::Switch

=head1 DESCRIPTION

This class is part of Metrocluster Network Topology.  It has a public
constructor so that it can be used in tests that may not require a Metrocluster
harness object (for DR Ops, etc), but access to the fabrics and switches in
the shared storage may still need to be monitored or manipulated.  This class
is expressly designed for MCC systems and should not be used for switch
control in other configurations.

=over

=item Examples

=back

    use WaflMCC::MCC::Network::Switch;

=cut

=head1 METHODS

=head2 new()

These constructors are public or MCC Test Lib only, as marked.  See acceptable
usage notes above in regards to using the public constructors outside the
construction of a WaflMCC::MCC::Metrocluster object.

=over

=item C<< Public Constructors >>

    # return access to a real or virtual switch of name switch_name
    my $switch = WaflMCC::MCC::Network::Switch->new(
                                   switch_name => $switch_name);

    # return access to a virtual switch operating on the real switch switch_name
    my $switch = WaflMCC::MCC::Network::Switch->new(
                                   switch_name => $switch_name,
                                   vswitch_id  => $vswitch_id);

=back

=over

=item C<<Library Access Only Constructors >>

    # return access to a virtual switch operating on the real switch switch_name
    my $switch = WaflMCC::MCC::Network::Switch->new(
                                   switch_name            => $switch_name,
                                   node_target_ports_spec => $port_spec);

    # return access to a virtual switch operating on the real switch switch_name
    my $switch = WaflMCC::MCC::Network::Switch->new(
                                   switch_name            => $switch_name,
                                   vswitch_isl_target_spec => $wwn_spec);

=back

Returns a WaflMCC::MCC::Network::Switch object.  This object may represent
either a real switch or a virtual switch.  The MCC system should not know or
care which.  But if it represents a real switch, the test user must be aware
that some actions they perform on this switch could affect other MCC systems
(or multiple DR Groups in their own MCC system) that may share the backend
fabrics' switches via virtual fabrics (or VSANs).  If this object represents
a virtual switch, such operations will access/affect only ports associated with
this particular virtual switch, allowing other virtual switches on this switch
to continue operating normally.

=over

=item Arguments

=over

=item C<< switch_name => $switch_name >>

(Required.) The text string name of the switch, which must have an associated
NATE_HOST file so that it can be found and connected to.

WARNING: If no vswitch associated parameters are passed then operations on this
switch can impact all of the virtual switches that may be running concurrently
on the physical switch.

=back

=over

=item C<< vswitch_id => $vswitch_id >>

(Optional, but cannot be passed with any other vswitch_* parameters.)  This ID
specifies the ID of the VFabric/VSAN of a "virtual switch" operating on the
switch $switch_name.  Passing this optional parameter implies that this
WaflMCC::MCC::Network::Switch object being constructed is purely a virtual
switch, and any operations subsequently executed on this virtual switch will
AFFECT ONLY THE PORTS ASSOCIATED WITH THIS VIRTUAL SWITCH.

=back

=over

=item C<< node_target_ports_spec => $port_spec >>

(Optional, but cannot be passed with any other vswitch_* parameters.)  This spec
is for any one of the FCVI ports in use by the virtual switch which this
WaflMCC::MCC::Network::Switch object will be configured to represent.
Passing this optional parameter implies that the object being constructed
represents a particular virtual switch running in the physical switch, and any
operations subsequently executed on this virtual switch will AFFECT ONLY THE
PORTS ASSOCIATED WITH THIS VIRTUAL SWITCH.

This option for building a Switch object representing a virtual switch is used
by the library after the library has determined a switch name and set of FCVI
(and other) ports a particular filer node is attached to via a call to the
node's response to 'sysconfig -va'.

See also the WaflMCC::MCC::Network::Fabric constructor for more details
regarding the usage and need for both WaflMCC::MCC::Network::Switch
constructors with the parameters node_target_ports_spec and
vswitch_isl_target_spec.

And the next step in the process of determining the virtual switch residing in
the same fabric across the ISL (at the other site) is to use this switch to find
the ISL target switch name and virtual switch WWN of the virtual switch on the
other side of the ISL.  Then this information is used to build another
WaflMCC::MCC::Network::Switch object using the Switch constructor which
takes the vswitch_isl_target_spec parameter.

$port_spec is a hashref structured as follows:
{
  'storage_connections' => [
                             {
                               'port_wwn' => '5:00a:098000:8a2c40',
                               'port_descriptor' => '1-1',
                               'switch_name' => 'dpg-mcc-cisco9148-03-A-fabX'
                             },
                             {
                               'port_wwn' => '5:00a:098100:8a2c40',
                               'port_descriptor' => '1-2',
                               'switch_name' => 'dpg-mcc-cisco9148-03-A-fabY'
                             },
                             {
                               'port_wwn' => '2:100:001b32:8e1e94',
                               'port_descriptor' => '1-3',
                               'switch_name' => 'dpg-mcc-cisco9148-03-A-fabX'
                             },
                             {
                               'port_wwn' => '2:101:001b32:ae1e94',
                               'port_descriptor' => '1-4',
                               'switch_name' => 'dpg-mcc-cisco9148-03-A-fabY'
                             }
                           ],
  'fcvi_connections' => [
                          {
                            'port_wwn' => '21:00:00:24:ff:31:ed:84',
                            'port_descriptor' => '1-13',
                            'switch_name' => 'dpg-mcc-cisco9148-03-A-fabX'
                          },
                          {
                            'port_wwn' => '21:00:00:24:ff:31:ed:85',
                            'port_descriptor' => '1-13',
                            'switch_name' => 'dpg-mcc-cisco9148-03-A-fabY'
                          }
                        ]
}


port_descriptor is a text string whose format depends on the switch make.  For
Brocade, it is a digit.  For Cisco, it is of the form: N/M where N and M are
digits.

port_wwn is a text string of the World Wide Name of the port.

=back

=over

=item C<< vswitch_isl_target_spec => $wwn_spec >>

(Optional, but cannot be passed with any other vswitch_* parameters.)  This ID
specifies the ID of the VFabric/VSAN of a "virtual switch" operating on the
switch $switch_name.  Passing this optional parameter implies that this
WaflMCC::MCC::Network::Switch object being constructed is purely a virtual
switch, and any operations subsequently executed on this virtual switch will
AFFECT ONLY THE PORTS ASSOCIATED WITH THIS VIRTUAL SWITCH.

=back

=back

=cut

use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                  { -type => 'WaflMCC::MCC::Network::Switch_Impl'},
                  '_switch_impl',
              ],
                       ];

sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            switch_name             => { regex => qr/^[\w\-]+$/ },
            vswitch_id              => { regex => qr/^[\w\-]+$/, optional => 1},
            node_target_ports_spec  => { type  => HASHREF,       optional => 1},
            vswitch_isl_target_spec => { type  => HASHREF,       optional => 1},
                  },
                            );
    # only a single vswitch_* parameter may be validly passed
    if (scalar(grep { $_ =~ /^(vswitch_|node_)/ } keys(%opts)) > 1) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "WaflMCC::MCC::Network::Switch paramters vswitch_id,\n" .
            "node_target_ports_spec, and vswitch_isl_target_spec are\n" .
            "mutually exclusive parameters.  More than one have been passed.");
    }

    my $host = Tharn::host($opts{switch_name});

    my $switch_implementation = undef;
    my $vendor = $host->{MANUFACTURER};
    if ($vendor =~ /brocade/i) {
        $switch_implementation =
            WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(%opts);
    } elsif ($vendor =~ /cisco/i) {
        $switch_implementation =
            WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl->new(%opts);
    }

    # sanity
    if (defined($switch_implementation)) {
        $self->_switch_impl($switch_implementation);
    } else {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                             WaflMCC::MCC::Utils->dump($self) .
                                       ":\n\tSwitch implemntation not found.");
    }

    return;
}

=head2 get_vendor()

    my $vendor_name = $switch->get_vendor();

Returns the name of the switch vendor, e.g., brocade, cisco, etc.

=cut

sub get_vendor {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl()->L_get_vendor();
}

=head2 enable()

    $switch->enable();

Enables the switch.

=cut

sub enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl()->L_enable();
}

=head2 disable()

    $switch->disable();

Disables the switch.

=cut

sub disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl()->L_disable();
}

=head2 update_switch_info()

    $switch->update_switch_info();

Updates the current status of the switch (ports up/down, etc)

=cut

sub update_switch_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl()->L_update_current_port_info();
}

=head2 is_enabled()

    if ($switch->is_enabled()) { ... }

Returns 0|1 signifying if the switch is enabled or not.

=cut

sub is_enabled {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl()->L_is_enabled();
}

=head2 get_switch_impl()

    my $switch_impl = $switch->get_switch_impl();

Returns the switch implementation object so that commands specific to this
type of switch can be accessed if needed.

=cut

sub get_switch_impl {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_impl();
}

=head2 get_switch_name()

    my $switch_name = $switch->get_switch_name();

Returns the name of the switch.

=cut

sub get_switch_name {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->get_switch_impl()->get_switch_name();
}

=head2 get_description()

    my $switch_description = $switch->get_description();

Returns the name and virtual switch id of the switch.

=cut

sub get_description {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->get_switch_impl()->get_description();
}

=head2 get_fcvi_ports()

    my $fcvi_ports = $switch->get_fcvi_ports();
    map { $_->disable() } @{ $fcvi_ports };

Returns an arrayref of WaflMCC::MCC::Netork::Port objects associated with
the fcvi ports found on the switch.

=cut

sub get_fcvi_ports {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
                     @{ $self->_switch_impl()->L_get_fcvi_port_descriptors() };

    my @ports = map { WaflMCC::MCC::Network::Port->new(
                            port_descriptor => $_, switch => $self) }
                                                             @port_descriptors;

    return \@ports;
}

=head2 get_storage_ports()

    my @storage_ports = $switch->get_storage_ports();
    map { $_->enable() } @{ $storage_ports };

Each filer at a site has connections to this switch which are used for data
I/O to and from the shared storage array.  This method returns an arrayref of
WaflMCC::MCC::Netork::Port objects associated with such ports.

Do not confuse these ports with ports out to the fibre-bridge array (see
get_fibreBridge_ports().

=cut

sub get_storage_ports {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
                  @{ $self->_switch_impl()->L_get_storage_port_descriptors() };

    my @ports = map { WaflMCC::MCC::Network::Port->new(
                            port_descriptor => $_, switch => $self) }
                                                             @port_descriptors;

    return \@ports;
}

=head2 get_fibreBridge_ports()

    my @fibreBridge_ports = $switch->get_fibreBridge_ports();

Returns an arrayref of WaflMCC::MCC::Netork::Port objects associated with
the ports connected out to the fibre-bridge (i.e., armadillos) or bridges.

=cut

sub get_fibreBridge_ports {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
              @{ $self->_switch_impl()->L_get_fibreBridge_port_descriptors() };

    my @ports = map { WaflMCC::MCC::Network::Port->new(
                            port_descriptor => $_, switch => $self) }
                                                             @port_descriptors;

    return \@ports;
}

=head2 get_isl_ports()

    my @isl_ports = $switch->get_isl_ports();

Returns an arrayref of WaflMCC::MCC::Netork::Port objects associated with
the ports connected to the inter-switch link(s) (ISLs).

=cut

sub get_isl_ports {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
                      @{ $self->_switch_impl()->L_get_isl_port_descriptors() };

    my @ports = map { WaflMCC::MCC::Network::Port->new(
                            port_descriptor => $_, switch => $self) }
                                                             @port_descriptors;

    return \@ports;
}

1;
