# -*-perl-*-

##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

package WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl;

use strict;
use warnings;
use Carp;
use base qw(WaflMCC::MCC::Network::Switch_Impl);
use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Switch;
use Tharn;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);

use List::MoreUtils qw{ any };


=head1 NAME

WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl

=head1 DESCRIPTION

This is a Brocade FC Switch Implementation Class. It interacts with Switch
APISet to run commands.

See new() for construction details.

=cut

=head2 new

This is the constructor of this class.  Only the
WaflMCC::MCC::Network::Switch class should ever access these constructors.

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(
                                        switch_name            => $switch_name);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(
                                        switch_name            => $switch_name,
                                        vswitch_id             => $vswitch_id);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(
                                        switch_name            => $switch_name,
                                        node_target_ports_spec => $spec);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(
                                        switch_name            => $switch_name,
                                        vswitch_isl_target_spec => $wwn_spec);

The vswitch associated parameters (vswittch_*) are all different ways in which
to either specify or to pass deterministic information regarding which virtual
fabric this object will be configured to represent.

Only one of the vswitch associated parameters (vswittch_*) may be validly
passed in a constructor call.  If more than one is passed, an exception will
be thrown.

=over

=item Arguments

All arguments in these constructors are identical to the arguments of the same
name in the WaflMCC::MCC::Network::Switch class.  Please see the
specification therein.

=back

=cut

use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'NACL::APISet::Switch' },
                '_switch_apiset',
              ],
    scalar => [
        qw/  _switch_name
             _default_vswitch_id
             _switch_timeout
             _port_info_update_time
             _port_info
             _port_info_cache /,
              ],
                        ];

sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            switch_name            => { regex => qr/^[\w\-]+$/ },
            vswitch_id             => { regex => qr/^\d+$/,     optional => 1 },
            node_target_ports_spec => { type  => HASHREF,       optional => 1 },
            vswitch_isl_target_spec => { type  => HASHREF,       optional => 1},
                  },
                            );

    my $switch_name = delete($opts{switch_name});

    # the rest of what might be in %opts is vswitch discovery related
    my %vswitch_opts = %opts;

    $self->_switch_name($switch_name);

    my $Host = host($switch_name);
    my $tries = 3;
    do {
        $self->_switch_apiset(NACL::APISet->new(
            hostobj   => $Host,
            category  => 'Switch',
            interface => 'CLI',
            set       => 'BrocadeFC',
            # guess i still don't understand why cdef_path is needed.  what is
            # wrong with the default path where all other cdefs are found?
            cdef_path => '/usr/software/test/agnostic/lib/' .
                         'NACL/APISet/Switch/CLI/BrocadeFC/default'));
    } while ($self->_disable_switch_timeout() && $tries--);

    # if virtual fabrics are enabled, find which VF we should be in and change
    # context for commands to run only in that VF
    if ($self->_get_vf_status() eq 'enabled') {
        $self->_default_vswitch_id($self->get_vswitch_id());

        my $vswitch_id = undef;
        if (exists($vswitch_opts{node_target_ports_spec}) ||
	                           exists($vswitch_opts{vswitch_isl_target_spec})) {
            $vswitch_id = $self->_determine_vswitch_id(%vswitch_opts);
        } elsif (exists($vswitch_opts{vswitch_id})) {
            $vswitch_id = $vswitch_opts{vswitch_id};
        }

        # now set switch context to the correct FID?
        if (defined($vswitch_id)) {
            $self->_set_vswitch_context(vswitch_id => $vswitch_id);
        }
    }

    # this will become stale if the switch context changes
    $self->_port_info_cache($self->_gather_port_info());
    $self->_port_info($self->_port_info_cache());

    return;
}

=head2 get_switch_APISet()

It gives you handle to the switch APISet object. With this handle
you can directly run commands to the switch

=over

=item Arguments

 None

=item Return Value

Handle to switch apiset

=back

    my $switch_apiset = $switch_impl->get_switch_APISet();

=cut

sub get_switch_APISet {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_apiset;
}


=head2 get_switch_name

It returns the name of the switch

=over

=item Arguments

 None

=item Return Value

name of the switch

=back

    my $switch_hostname = $switch_impl->get_switch_name();

=cut

sub get_switch_name {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_name;
}

=head2 _disable_switch_timeout

Disables switch timeout if there is one

=over

=item Arguments

 None

=item Return Value

If timeout was NOT disabled, return timeout value
If timeout already disabled, return 0

(This is important for brocade because the timeout modification doesn't
take effect until logging in again, so if a change was made, a new
switch object must be created.


=back

    my $current_switch_timeout = $switch_impl->_disable_switch_timeout();

=cut

sub _disable_switch_timeout {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my $response = $self->_switch_apiset()->timeout();
    #NACL switch parsing broken, use raw output for now
    #my ($timeout) = $response->get_parsed_output()->[0]->
    #                {current_idle_timeout}
    my ($timeout) = $response->get_raw_output()
                    =~ /(\d+)/;
    $self->_switch_timeout($timeout);

    if ($timeout) {
        $self->_switch_apiset()->timeout(timeout => 0);
        $response = $self->_switch_apiset()->timeout();
        #NACL switch parsing broken, use raw output for now
        #($timeout) = $response->get_parsed_output()->[0]->
        #                {current_idle_timeout}
        ($timeout) = $response->get_raw_output()
                   =~ /(\d+)/;
        if (!$timeout) {
            $Log->comment(
                $self->get_switch_name() . " switch timeout disabled");
        }
    }


    return $self->_switch_timeout;
}

=head2 get_description

    my $switch_description = $switch_impl->get_description();

This method returns the name and FID as a text string.

=over

=item Arguments

    None

=back

=cut

sub get_description {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $fid = ($self->get_vswitch_id() eq $self->get_default_vswitch_id()) ?
                                           'default' : $self->get_vswitch_id();
    return $self->get_switch_name() . " (FID: $fid)";
}

=head2 L_get_vendor()

    my $vendor_name = $switch_impl->L_get_vendor();

Returns the name of the switch vendor, e.g., brocade, cisco, etc.

=cut

sub L_get_vendor {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return 'brocade';
}

=head2 L_get_fcvi_port_descriptors()

    my $fcvi_port_descriptors = $switch_impl->L_get_fcvi_port_descriptors();
    foreach my $port_descriptor (@{ $fcvi_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the fcvi ports found on the switch.

Implementation detail: for Brocade switches, the port descriptor is the port
index, e.g., 1, 2, etc.

=cut

sub L_get_fcvi_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
           keys(%{ $self->_port_info_cache()->{nsshow_parsed}->{fcvi_ports} });

    return  \@port_descriptors;
}

=head2 L_get_storage_port_descriptors()

    my $storage_port_descriptors =
                                $switch_impl->L_get_storage_port_descriptors();
    foreach my $port_descriptor (@{ $storage_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Each filer at a site has connections to this switch which are used for data
I/O to and from the shared storage array.  This method returns an arrayref of
port descriptor strings associated with such ports.

Do not confuse these ports with ports out to the fibre-bridge array (see
get_fibreBridge_ports().

Implementation detail: for Brocade switches, the port descriptor is the port
index, e.g., 1, 2, etc.

=cut

sub L_get_storage_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
        keys(%{ $self->_port_info_cache()->{nsshow_parsed}->{storage_ports} });

    return  \@port_descriptors;
}

=head2 L_get_fibreBridge_port_descriptors()

    my $fibreBridge_port_descriptors =
                            $switch_impl->L_get_fibreBridge_port_descriptors();
    foreach my $port_descriptor (@{ $fibreBridge_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the fcvi ports found on the switch.

Implementation detail: for Brocade switches, the port descriptor is the port
index, e.g., 1, 2, etc.

=cut

sub L_get_fibreBridge_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors = keys(%{ $self->_port_info_cache()->
                                       {nsshow_parsed}->{fibreBridge_ports} });

    return  \@port_descriptors;
}

=head2 L_get_isl_port_descriptors()

    my $isl_port_descriptors = $switch_impl->L_get_isl_port_descriptors();
    foreach my $port_descriptor (@{ $isl_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the ISL ports found on the switch.

Implementation detail: for Brocade switches, the port descriptor is the port
index, e.g., 1, 2, etc.

=cut

sub L_get_isl_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors =
        keys(%{ $self->_port_info_cache()->{switchshow_parsed}->{isl_ports} });

    return  \@port_descriptors;
}

=head2 _gather_port_info()

this get some pretty raw data about ports and returns it as a hashref with
the following structure:

   {
       nsshow_raw    => 'a multiline string',
       nsshow_parsed => {
           fcvi_ports => {
               $port_descriptor => {
                   'share_area' => 'No',
                   'fabric_port_name' => '20:08:00:05:33:91:27:8e',
                   'nodename' => '20:00:00:24:ff:31:ec:ea',
                   'portsymb' => '[255] ".Netapp FCVI adapter  1a 0........."',
                   'portname' => '21:00:00:24:ff:31:ec:ea',
                   'partial' => 'No',
                   'device_shared_in_other_ad' => 'No',
                   'ttlsec' => 'na',
                   'port_index' => '8',
                   'cos' => '3',
                   'redirect' => 'No',
                   'pid' => '020800',
                   'permanent_port_name' => '21:00:00:24:ff:31:ec:ea',
                   'type' => 'N',
                   'meta__connection_target' => '21:00:00:24:ff:31:ec:ea'
                                   },
               ...
                         },
           fibreBridge_ports => {
               $port_descriptor => {
                   'share_area' => 'No',
                   'fabric_port_name' => '20:07:00:05:33:91:27:8e',
                   'nodesymb' => '[27] "ATTO FibreBridge 6500N A55R"',
                   'nodename' => '20:00:00:10:86:60:09:da',
                   'portname' => '21:00:00:10:86:60:09:da',
                   'partial' => 'No',
                   'device_shared_in_other_ad' => 'No',
                   'ttlsec' => 'na',
                   'port_index' => '7',
                   'cos' => '3',
                   'redirect' => 'No',
                   'pid' => '020700',
                   'fc4s' => 'FCP',
                   'permanent_port_name' => '21:00:00:10:86:60:09:da',
                   'type' => 'N',
                   'meta__connection_target' => '21:00:00:10:86:60:09:da'
                                   },
               ...
                                },
           storage_ports => {
               $port_descriptor => {
                   'share_area' => 'No',
                   'fabric_port_name' => '20:09:00:05:33:91:27:8e',
                   'nodename' => '20:00:00:1b:32:0c:ad:08',
                   'portsymb' =>
                      '[48] "NetApp FC Storage Adapter dpg-mcc-3270-01-a2:4a."',
                   'portname' => '21:00:00:1b:32:0c:ad:08',
                   'partial' => 'No',
                   'device_shared_in_other_ad' => 'No',
                   'ttlsec' => 'na',
                   'port_index' => '9',
                   'cos' => '3',
                   'redirect' => 'No',
                   'pid' => '020900',
                   'permanent_port_name' => '21:00:00:1b:32:0c:ad:08',
                   'type' => 'N',
                   'meta__connection_target' => 'dpg-mcc-3270-01-a2:4a'
                                   },
               ...
                            },
                        },
        switchshow_raw    => 'another multiline string',
        switchshow_parsed => {
            isl_ports => {
                $port_descriptor => {
                    'proto' => 'FC',
                    'wwpn' => '10:00:00:05:33:8a:21:65',
                    'port' => '12',
                    'state' => 'Online',
                    'index' => '12',
                    'media' => 'id',
                    'speed' => 'N16',
                    'porttype' => 'E-Port',
                    'address' => '020c00',
                    'meta__connection_target' => 'dpg-mcc-brcd6510-01-A-fabX'
                                    },
                ...
                         },
                             },
    }

=cut

sub _gather_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $port_info = { nsshow_raw        => undef,
                      nsshow_parsed     => { fcvi_ports        => { },
                                             storage_ports     => { },
                                             fibreBridge_ports => { },
                                             other_ports       => { },
                                           },
                      switchshow_raw    => undef,
                      switchshow_parsed => { isl_ports   => { },
                                             other_ports => { },
                                           },
                    };

    # nsshow returns information about the switch on per port basis. ###########
    my $response = $self->_switch_apiset()->nsshow();
    $port_info->{nsshow_raw} = $response->get_raw_output();

    # $parsed contains information about each port
    my $parsed_lines = $response->get_parsed_output();

    # for brocade, port_index is our "port descriptor"
    foreach my $parsed_line (@{ $parsed_lines }) {
        if (        defined($parsed_line->{portsymb}) ||
                    defined($parsed_line->{nodesymb})) {
            my $port_descriptor = $parsed_line->{port_index};

            # determine some metadata
            my $raw_symb_entry = defined($parsed_line->{portsymb}) ?
                           $parsed_line->{portsymb} : $parsed_line->{nodesymb};

            my ($type) = ($raw_symb_entry =~ /(Storage|FibreBridge|FCVI)/);
            my $connection_target = undef;
            if (defined($type)) {
                if ($type eq 'Storage') {
                    # gets target node_name:port
                    ($connection_target) =
                               ($raw_symb_entry =~ /Adapter\s+([\w\-]+\:\w+)/);
                    $type = 'storage_ports';
                } elsif ($type eq 'FibreBridge') {
                    # WWN only available.  should be port_name?  node_name?
                    ($connection_target) = $parsed_line->{permanent_port_name};
                    $type = 'fibreBridge_ports';
                } elsif ($type eq 'FCVI') {
                    # WWN only available.
                    ($connection_target) = $parsed_line->{permanent_port_name};
                    $type = 'fcvi_ports';
                }
            }
            # sanity check
            unless (defined($connection_target)) {
                my $msg = "connection_target undefined!\n" .
                          "    raw_symb_entry = $raw_symb_entry\n";
                WaflMCC::MCC::Exceptions::BaseException->throw($msg);
            }

            # add metadata to line hash
            $parsed_line->{meta__connection_target} = $connection_target;

            # save line hash in return structure
            $port_info->{nsshow_parsed}->{$type}->{$port_descriptor} =
                                                                  $parsed_line;
        }
    }

    # switchshow also returns some useful port information #####################
    $response = $self->_switch_apiset()->switchshow();
    my $switchshow_raw = $response->get_raw_output();
    $port_info->{switchshow_raw} = $switchshow_raw;

    $parsed_lines = $response->get_parsed_output();
    my $ports_info = undef;
    foreach my $entry (@{ $parsed_lines }) {
        if (exists($entry->{'ports-information'})) {
            $ports_info = $entry->{'ports-information'};
            last;
        }
    }

    foreach my $port_hash (@{ $ports_info }) {
        if (defined($port_hash->{porttype}) &&
                                          $port_hash->{state} ne 'No_Module') {
            my $index = $port_hash->{index};
            my $port_descriptor = $index;

            # determine some metadata
            my $connection_target = undef;
             my $type = undef;
            if ($port_hash->{porttype} eq 'F-Port') {
                # captures WWN only
                ($connection_target) =
                     ($switchshow_raw =~ /^\s*$index.*F-Port\s+([\w:]+)\s*$/m);
                $type = 'other_ports';
            } elsif ($port_hash->{porttype} eq 'E-Port') {
                # captures 'WWN "hostname" maybe_some_other_stuff_too'
                ($connection_target) = ($switchshow_raw =~
                           /^\s*$index.*E-Port\s+(.+)\s*$/m);
                $type = 'isl_ports';
            } else {
                $connection_target = q{};
            }
            # sanity check
            unless (defined($connection_target)) {
                my $msg = "connection_target undefined!\n" .
                          "    port $index in :\n" . $switchshow_raw;
                WaflMCC::MCC::Exceptions::BaseException->throw($msg);
            }

            # add metadata to line hash
            if ($connection_target) {
                $port_hash->{meta__connection_target} = $connection_target;
            }

            # save line hash in return structure
            $port_info->{switchshow_parsed}->{$type}->{$port_descriptor} =
                                                                    $port_hash;
        }
    }

    $self->_port_info_update_time(time); #update the last gathered time

    return $port_info;
}



=head2 L_update_current_port_info

updates current port info

=over

=item Arguments

=item Return Value

None

=back

    $switch_impl->L_update_current_port_info();

=cut

sub L_update_current_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    return $self->_port_info($self->_gather_port_info());

}

=head2 _get_current_port_info

Gets new or recent port info

=over

=item Arguments

  Required:
  Optional: <stale_time> int val

=item Return Value

None

=back

    $switch_impl->_get_current_port_info(stale_time => 2);

=cut

sub _get_current_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      stale_time   => { regex => qr/^\d+$/, default => 30 },
                  },
                            );

    my $last_time = $self->_port_info_update_time();

    if (time > $opts{stale_time} + $last_time) {
        #$self->_port_info($self->_gather_port_info());
        $self->L_update_current_port_info();
        #print "stale! [$opts{stale_time}]:" . Dumper $self->_port_info;
    }

    return $self->_port_info();

}

=head2 L_port_is_enabled

This returns whether port <port_descriptor> is enabled.

=over

=item Arguments

  Required: <port_descriptor> int val

=item Return Value

1|0

=back

    $switch_impl->L_port_enable(port_descriptor => '0');

=cut

sub L_port_is_enabled {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr/^\d+$/ },
                      stale_time      => { optional => 1 },
                  },
                            );
    my $enabled = 0;
    my $port_descriptor = delete $opts{port_descriptor};
    my $port_info =
        $self->_get_current_port_info( %opts );

    #TODO maybe add a single 'ports' hash to iterate over rather than searching

    foreach my $type
      (qw (storage_ports fibreBridge_ports fcvi_ports other_ports) ) {
        if (defined $port_info->{nsshow_parsed}->{$type}->{$port_descriptor}) {
            my $state =
                $port_info->{nsshow_parsed}->{$type}->{$port_descriptor}->
                {state};

            if ($state =~ /online/i) {
                $enabled = 1;
            }
        }
    }

    foreach my $type (qw (isl_ports other_ports) ) {
        if (defined
          $port_info->{switchshow_parsed}->{$type}->{$port_descriptor}) {
            my $state =
                $port_info->{switchshow_parsed}->{$type}->
                {$port_descriptor}->{state};

            if ($state =~ /online/i) {
                $enabled = 1;
            }
        }
    }

    return $enabled;
}

=head2 L_port_enable

This runs port enable <port_descriptor> and makes sure it is enabled.

=over

=item Arguments

  Required: <port_descriptor> int val
  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

    $switch_impl->L_port_enable(port_descriptor => '0');

=cut

sub L_port_enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr/^\d+$/ },
                      timeout         => { regex => qr/^\d+$/, default => 30 },
                      interval        => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    my $response =
             $self->_switch_apiset()->portenable(port=>$opts{port_descriptor});
    my $end_time = $opts{timeout} ?
                 Tharn::timeout2time($opts{timeout}) : Tharn::timeout2time(-1);

    my $time_remaining;
    # this block takes care of making sure its enabled else time out
    while (time() <= $end_time) {
        $response = $self->_switch_apiset()->switchshow();
        my $parsed = $response->get_parsed_output();
        my $state =
        $parsed->[0]->{'ports-information'}->[$opts{port_descriptor}]->{state};
        if ($state =~ /online/i) {
            $Log->comment("===============================================\n");
            $Log->comment("======== Port $opts{port_descriptor} is Enabled! " .
                          "=========\n");
            $Log->comment("===============================================\n");
            return;
        }

        $Log->comment(
                "Waiting for port: $opts{port_descriptor} to be enabled...\n");
        $Log->comment("Interval: $opts{interval}\n");
        Tharn::snooze($opts{interval});
        $time_remaining = $end_time - time();
        $Log->comment("Time out in : $time_remaining\n");
    }

# TBD: throw an exception
    confess ("Timed out waiting for port: $opts{port_descriptor} to be\n" .
             "Enabled! Please login to the switch for more Info.. \n");
}


=head2 L_port_disable

This runs port disable <port_descriptor> and makes sure it is disabled.

=over

=item Arguments

  Required: <port_descriptor> int val
  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

    $switch_impl->L_port_disable(port_descriptor => '0');

=cut

sub L_port_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr/^\d+$/ },
                      timeout         => { regex => qr/^\d+$/, default => 30 },
                      interval        => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    my $response =
            $self->_switch_apiset()->portdisable(port=>$opts{port_descriptor});
    my $end_time = $opts{timeout} ?
                 Tharn::timeout2time($opts{timeout}) : Tharn::timeout2time(-1);

    my $time_remaining;
    #this block takes care of making sure its enabled else time out
    while (time() <= $end_time) {
        $response = $self->_switch_apiset()->switchshow();
        my $parsed = $response->get_parsed_output();
        my $state = $parsed->[0]->{'ports-information'}->
                                          [$opts{port_descriptor}]->{porttype};
        if ($state =~ /disabled/i) {
            $Log->comment("===============================================\n");
            $Log->comment("======== Port $opts{port_descriptor} is Disabled! " .
                          "======= \n");
            $Log->comment("===============================================\n");
            return;
        }

        $Log->comment(
               "Waiting for port: $opts{port_descriptor} to be disabled...\n");
        $Log->comment("Interval: $opts{interval}\n");
        Tharn::snooze($opts{interval});
        $time_remaining = $end_time - time();
        $Log->comment("Time out in : $time_remaining\n");
    }

#TBD: throw exception
    confess ("Timed out waiting for port: $opts{port_descriptor} to be\n" .
             "Disabled! Please login to the switch for more Info.. \n");
}

=head2 L_enable

This runs switchenable and verifies it

=over

=item Arguments

  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

    my $sw = WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(switch_name =>$SWITCH);
    $sw->L_enable();

=cut

sub L_enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      timeout  => { regex => qr/^\d+$/, default => 30 },
                      interval => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    my $response = $self->_switch_apiset()->switchenable();
    my $end_time = $opts{timeout} ?
                 Tharn::timeout2time($opts{timeout}) : Tharn::timeout2time(-1);

    my $time_remaining;
    my $switch_name = $self->get_switch_name();
    # this block takes care of making sure its enabled else time out
    while (time() <= $end_time) {
        $response = $self->_switch_apiset()->switchshow();
        my $parsed = $response->get_parsed_output();
        my $state = $parsed->[0]->{switchstate};
        if ($state =~ /online/i) {
            $Log->comment("===============================================\n");
            $Log->comment("====== Switch: $switch_name is Enabled! =======\n");
            $Log->comment("===============================================\n");
            return;
        }

        $Log->comment("Waiting for switch: $switch_name() to be enabled...\n");
        $Log->comment("Interval: $opts{interval}\n");
        Tharn::snooze($opts{interval});
        $time_remaining = $end_time - time();
        $Log->comment("Time out in : $time_remaining\n");
    }

#TBD: throw exception
    confess ("Timed out waiting for switch: $switch_name() to be Enabled! Please login to the switch for more Info.. \n");
}

=head2 L_disable

This runs switchdisable and verifies it

=over

=item Arguments

  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

    my $sw = WaflMCC::MCC::Network::Switch_Impl::BrocadeFC_Impl->new(switch_name =>$SWITCH);
    $sw->L_disable();

=cut

sub L_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      timeout  => { regex => qr/^\d+$/, default => 30 },
                      interval => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    my $response = $self->_switch_apiset()->switchdisable();
    my $end_time = $opts{timeout} ?
                 Tharn::timeout2time($opts{timeout}) : Tharn::timeout2time(-1);
    my $time_remaining;
    my $switch_name = $self->get_switch_name();
    # this block takes care of making sure its enabled else time out
    while (time() <= $end_time) {
        $response = $self->_switch_apiset()->switchshow();
        my $parsed = $response->get_parsed_output();
        my $state = $parsed->[0]->{switchstate};
        if ($state =~ /offline/i) {
            $Log->comment("===============================================\n");
            $Log->comment("====== Switch: $switch_name is Disabled! ======\n");
            $Log->comment("===============================================\n");
            return;
        }
        $Log->comment("Waiting for switch: $switch_name() to be disabled...\n");
        $Log->comment("Interval: $opts{interval}\n");
        Tharn::snooze($opts{interval});
        $time_remaining = $end_time - time();
        $Log->comment("Time out in : $time_remaining\n");
    }

#TBD: throw exception
    confess ("Timed out waiting for switch: $switch_name() to be Disabled! Please login to the switch for more Info.. \n");
}

=head2 L_get_port_target_host_hints()

    my $port_connection_target_hints =
                 $switch_impl->L_get_port_target_host_hints(
                                          port_descriptor => $port_descriptor);

(Library method) Returns a hash of information about the host on the remote
side of the connection from the port described by port_descriptor.  This
information may not be completely reliable.  Some of it seems to be data
entered by hand by a network admin instead of being auto-generated by the
switch.  Thus the 'hints' in the method name.

This hash has the following form:

    {
        hostname     => 'dpg-mcc-brcd6510-01-A-fabY',
        host_port_id => '0d',
        host_type    => ('switch'|'node'|'fibreBridge'),
        WWN          => '21:00:00:24:ff:32:06:90',
    }

=over

=item Options

=over

=item C<< port_descriptor >>

(Required) The text identifier of a port on this switch.  In this case, it is
the port index.

=back

=back

=cut

sub L_get_port_target_host_hints {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            port_descriptor => { regex => qr/^\d+$/ },
                  },
                            );

    my $port_index = delete($opts{port_descriptor});

    my $port_info    = $self->_port_info_cache();
    my $fcvi_info    = $port_info->{nsshow_parsed}->    {fcvi_ports};
    my $storage_info = $port_info->{nsshow_parsed}->    {storage_ports};
    my $fb_info      = $port_info->{nsshow_parsed}->    {fibreBridge_ports};
    my $isl_info     = $port_info->{switchshow_parsed}->{isl_ports};

    my ($hostname, $host_port_id, $host_type, $WWN) =
                                                  (undef, undef, undef, undef);
    if (exists($fcvi_info->{$port_index})) {
        # FCVI port
        $host_type = 'node';
        $WWN       = $fcvi_info->{$port_index}->{meta__connection_target};
    } elsif (exists($storage_info->{$port_index})) {
        # storage port
        ($hostname,
         $host_port_id) =
            split(/:/, $storage_info->{$port_index}->{meta__connection_target});
        $host_type      = 'node';
        $WWN            = $storage_info->{$port_index}->{nodename};
    } elsif (exists($fb_info->{$port_index})) {
        # fibre-bridge port
        $host_type = 'fibreBridge';
        $WWN       = $fb_info->{$port_index}->{meta__connection_target};
    } elsif (exists($isl_info->{$port_index})) {
        # ISL port
        $host_type        = 'switch';
        ($WWN, $hostname) =
             split(/\s+/, $isl_info->{$port_index}->{meta__connection_target});
        $hostname =~ s/\"//g;
    } else {
        my $msg = "L_get_port_target_host_hints: No port information found \n" .
                  "    for port: $port_index\n";
        $Log->warn($msg);
    }

    my $target_info = { hostname     => $hostname,
                        host_port_id => $host_port_id,
                        host_type    => $host_type,
                        WWN          => $WWN, };

    return $target_info;
}

=head2 _determine_vswitch_id()

This is used to determine an FID associated with an FCVI port known to
connect to this virtual switch from a filer node.  This method requires that
one of these two parameters must be passed:
    node_target_ports_spec XOR vswitch_isl_target_spec.

    my $vswitch_id = $self->_determine_vswitch_id(
                                node_target_ports_spec => $vswitch_port_spec);

    where $vswitch_port_spec is a hashref with the following format:

        {
            port_descriptor => '0',
            port_wwn        => '21:00:00:24:ff:32:06:90',
        }



This is used to determine an FID associated with being the target across an ISL
from another switch in the same fabric.

    my $vswitch_id = $self->_determine_vswitch_id(
                       vswitch_isl_target_spec => $vswitch_target_wwn_spec);

    where $vswitch_target_wwn is the target WWN from a particular ISL.


(Private method) Returns the FID of the virtual fabric associated with either of
the conditions described in the two cases above.

=over

=item Options

=over

=item C<< node_target_ports_spec >>

(Required parameter as such: node_target_ports_spec XOR vswitch_isl_target_spec)
node_target_ports_spec is a  hashref specifying a port in the virtual fabric.
This hasref format is as described above.

=back

=over

=item C<< vswitch_isl_target_spec >>

(Required parameter as such: vswitch_isl_target_spec XOR node_target_ports_spec)
vswitch_isl_target_spec is a hashref containg a valid WWN.

=back

=back

=cut

sub _determine_vswitch_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_target_ports_spec  => { type => HASHREF,        optional => 1},
            vswitch_isl_target_spec => { type  => HASHREF,       optional => 1},
                  },
                            );

    my $vswitch_id = undef;
    if (scalar(keys(%opts)) == 1) {
        if (exists($opts{node_target_ports_spec})) {

            #now extract the port based on the switch name.  This is a
            #by product of the cisco switches requiring multiple vsans so
            #the call now passes in all ports, so we have to determine it now.

            my $node_port_spec = delete($opts{node_target_ports_spec});
            my $switch_name = $self->get_switch_name();
            my $port_spec = (grep { $_->{switch_name} eq $switch_name }
                @{ $node_port_spec->{fcvi_connections}} )[0] ;

            my $port_number = $port_spec->{port_descriptor};

            #print "port:$port_number\n";

            # now look for this port number in the output of 'lscfg --show'
            my $response = $self->_switch_apiset()->lscfg_show();
            my $parsed_lines = $response->get_parsed_output();
            # $parsed_lines should look like:
            # [
            #   {
            #     'created_switches' => [
            #                             '128(ds)',
            #                             '5',
            #                             '6'
            #                           ],
            #     'fabrics_info' => [
            #                         {
            #                           'fid' => 128,
            #                           'port' => [
            #                                       '24',
            #                                       ...,
            #                                       '47'
            #                                     ]
            #                         },
            #                         {
            #                           'fid' => 5,
            #                           'port' => [
            #                                       '0',
            #                                       ...,
            #                                       '11'
            #                                     ]
            #                         },
            #                         {
            #                           'fid' => 6,
            #                           'port' => [
            #                                       '12',
            #                                       ...,
            #                                       '23'
            #                                     ]
            #                         }
            #                       ]
            #   }
            # ]

            foreach my $element (@{ $parsed_lines }) {
                if (exists($element->{fabrics_info})) {
                    foreach my $fabrics_info_element (
                                               @{ $element->{fabrics_info} }) {
                        my @ports = @{ $fabrics_info_element->{port} };
                        # does port_number exist in this array?
                        if (any { $_ == $port_number } @ports) {
                            $vswitch_id = $fabrics_info_element->{fid};
                            last;
                        }
                    }
                    defined($vswitch_id) && last;
                }
            }

            unless (defined($vswitch_id)) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                        "Can't determine the virtual fabric ID for port\n\n" .
                        Dumper($port_spec) . "\n\n" .
                        "in switch " . $self->get_switch_name() . " !");
            }
        } elsif (exists($opts{vswitch_isl_target_spec})) {
            my $target_spec = delete($opts{vswitch_isl_target_spec});
            my $switch_name = $self->get_switch_name();
            my $target_wwn = (grep { $_->{switch_name} eq $switch_name }
                values %{ $target_spec } )[0]->{vswitch_isl_target_wwn} ;

            my $response = $self->_switch_apiset()->lscfg_show();
            my $parsed_lines = $response->get_parsed_output();
            # $parsed_lines should look like example above.

            foreach my $element (@{ $parsed_lines }) {
                if (exists($element->{'created_switches'})) {
                    # e.g., 'created_switches' => [ '128(ds)', '5', '6' ]
                    foreach my $fid_string (@{$element->{'created_switches'}}) {
                        my ($fid) = ($fid_string =~ /^\s*(\d+)/);

                        my $old_context =
                               $self->_set_vswitch_context(vswitch_id => $fid);

                        my $resp = $self->_switch_apiset()->switchshow();
                        my $parsed = $resp->get_parsed_output();
                        # $parsed should look like:
                        # [
                        #   {
                        #     'switchmode' => 'Native',
                        #     'switchwwn' => '10:00:00:05:33:88:9c:69',
                        #     'switchname' => 'dpg-mcc-brcd6510-05-B-fabX',
                        #     'zoning' => 'ON (mcc_vf1_x_cfg)',
                        #     'router' => 'OFF',
                        #     'switchrole' => 'Subordinate',
                        #     'switchbeacon' => 'OFF',
                        #     'ports-information' => [
                        #                              {
                        #                                'proto' => 'FC',
                        #                                'wwpn' => \
                        #                            '21:00:00:1b:32:81:af:b0',
                        #                                'port' => '0',
                        #                                'state' => 'Online',
                        #                                'index' => '0',
                        #                                'media' => 'id',
                        #                                'speed' => 'N4',
                        #                                'porttype' => 'F-Port',
                        #                                'address' => '020000'
                        #                              },
                        #                              ...,
                        #                              {
                        #                                'proto' => 'FC',
                        #                                'wwpn' => undef,
                        #                                'port' => '5',
                        #                                'state' => 'No_Light',
                        #                                'index' => '5',
                        #                                'media' => 'id',
                        #                                'speed' => 'N16',
                        #                                'porttype' => undef,
                        #                                'address' => '020500'
                        #                              {
                        #                                'proto' => 'FC',
                        #                                'wwpn' => \
                        #                            '10:00:00:05:33:8a:62:c5',
                        #                                'port' => '10',
                        #                                'state' => 'Online',
                        #                                'index' => '10',
                        #                                'media' => 'id',
                        #                                'speed' => 'N16',
                        #                                'porttype' => 'E-Port',
                        #                                'address' => '020a00'
                        #                              },
                        #                            ],
                        #     'use' => 'ON',
                        #     'switchtype' => '109.1',
                        #     'switchdomain' => '2',
                        #     'switchid' => 'fffc02',
                        #     'switchstate' => 'Online',
                        #     'attributes' => '[FID: 5, Base Switch: No, \
                        #                  Default Switch: No, Address Mode 0]'
                        #   }
                        # ];

                        foreach my $element_2 (@{ $parsed }) {
                            if (exists($element_2->{switchwwn})) {
                                if ($element_2->{switchwwn} eq $target_wwn) {
                                    # reset context for consistency
                                    $self->_set_vswitch_context(
                                                   vswitch_id => $old_context);
                                    $vswitch_id = $fid;
                                    last;
                                }
                            }
                        }
                        defined($vswitch_id) && last;
                    }
                }
                defined($vswitch_id) && last;
            }

            unless (defined($vswitch_id)) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                        "Can't determine the virtual fabric ID for ISL\n\n" .
                        "target WWN: $target_wwn\n\n" .
                        "in switch " . $self->get_switch_name() . " !");
            }
        }
    } else {
        # both parameters were passed.  not valid.
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "Both node_target_ports_spec and vswitch_isl_target_spec\n" .
            "parameters were passed.  Invalid call on this method.");
    }

    return $vswitch_id;
}

=head2 _set_vswitch_context()

    my $old_vswitch_id = $self->_set_vswitch_context(vswitch_id => $vswitch_id);

(Private method) Sets the context of this switch impl object to operate only
within the bounds of the virtual fabric defined by FID = $vswitch_id.

This method then confirms the context is set correctly or throws an exception.

Returns the value of the previous vswitch_id.

IMPORTANT NOTE for Lib developers!  When a change in FID context occurs as
usually happens when this method is called, the port data which is cached in
$self->_port_info_cache MUST be updated (and this is done implicitly in this
call.)

=over

=item Options

=over

=item C<< vswitch_id >>

(Required) The digit corresponding to a valid FID on this switch.

=back

=back

=cut

sub _set_vswitch_context {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            vswitch_id => { regex => qr/^\d+$/, },
                  },
                            );

    my $target_vswitch_id = delete($opts{vswitch_id});
    $Log->comment("Switch " . $self->get_switch_name() . " attempting\n" .
    "FID context switch to FID: $target_vswitch_id . . .\n");

    my $old_vswitch_id = $self->get_vswitch_id();
    if ($target_vswitch_id == $old_vswitch_id) {
        $Log->comment("  . . . context is already set to $target_vswitch_id.");
    } else {
        $self->_switch_apiset()->setcontext('fid-number' => $target_vswitch_id);

        my $vswitch_id = $self->get_vswitch_id();
        if ($vswitch_id == $target_vswitch_id) {
            $Log->comment("  . . . context successfully switched to " .
                          "FID: $target_vswitch_id.");

            # we now need to update data that rarely changes, but changes here.
            $self->_port_info_cache($self->_gather_port_info());
            $self->_port_info($self->_port_info_cache());
        } else {
            WaflMCC::MCC::Exceptions::BaseException->throw(
                "Switch " . $self->get_switch_name() . " failed to set " .
                "context to FID: $target_vswitch_id !!");
        }
    }

    return $old_vswitch_id;
}

=head2 get_vswitch_id()

    my $vswitch_id = $switch_impl->get_vswitch_id();

Returns the current FID context.

=over

=item Options

=over

=item None.

=back

=back

=cut

sub get_vswitch_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            # no valid parameters
                  },
                            );

    my $response = $self->_switch_apiset()->switchshow();
    my $parsed_lines = $response->get_parsed_output();
    my $current_vswitch_id = undef;
    foreach my $element (@{ $parsed_lines }) {
        if (exists($element->{attributes})) {
            ($current_vswitch_id) = ($element->{attributes} =~ /FID:\s*(\d+)/);
            last;
        }
        defined($current_vswitch_id) && last;
    }

    unless (defined($current_vswitch_id)) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "Can't determine the current fabric ID on switch" .
            $self->get_switch_name() . " !");
    }

    return $current_vswitch_id;
}

=head2 get_default_vswitch_id()

    my $default_vswitch_id = $switch_impl->get_default_vswitch_id();

Returns the switch's default FID context.

=over

=item Options

=over

=item None.

=back

=back

=cut

sub get_default_vswitch_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            # no valid parameters
                  },
                            );

    return $self->_default_vswitch_id();
}

=head2 _get_vf_status()

  This method run fosconfig --show to determine if vf is enabled
  on the switch. It returns 'enabled | disabled'

=over

=item Options

=over

=item None.

=back

=back

=cut

sub _get_vf_status {
  my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
  my $self = shift;
  my $response = $self->_switch_apiset()->fosconfig_show();
  my $parsed = $response->get_parsed_output();
  #its enabled or disabled!
  return $parsed->[0]->{virtual_fabric};
}

1;
