# -*-perl-*-
# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/Network/Switch_Impl/CiscoFC_Impl.pm#7 $

##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)
##no critic (RegularExpressions::ProhibitFixedStringMatches)

package WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl;

use strict;
use warnings;
use Carp;
use base qw(WaflMCC::MCC::Network::Switch_Impl);
use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Switch;
use Tharn;
use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);

use List::MoreUtils qw{ any };


=head1 NAME

WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl

=head1 DESCRIPTION

This is a Cisco FC Switch Implementation Class. It interacts with Switch
APISet to run commands.

See new() for construction details.

=cut

=head2 new

This is the constructor of this class.  Only the
WaflMCC::MCC::Network::Switch class should ever access these constructors.

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl->new(
                                        switch_name            => $switch_name);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl->new(
                                        switch_name            => $switch_name,
                                        vswitch_id             => $vswitch_id);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl->new(
                                        switch_name            => $switch_name,
                                        node_target_ports_spec => $spec);

     my $switch_impl =
             WaflMCC::MCC::Network::Switch_Impl::CiscoFC_Impl->new(
                                        switch_name            => $switch_name,
                                        vswitch_isl_target_spec => $wwn_spec);

The vswitch associated parameters (vswittch_*) are all different ways in which
to either specify or to pass deterministic information regarding which virtual
fabric this object will be configured to represent.

Only one of the vswitch associated parameters (vswittch_*) may be validly
passed in a constructor call.  If more than one is passed, an exception will
be thrown.

=over

=item Arguments

All arguments in these constructors are identical to the arguments of the same
name in the WaflMCC::MCC::Network::Switch class.  Please see the
specification therein.

=back

=cut

use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'NACL::APISet::Switch' },
                '_switch_apiset',
              ],
    scalar => [
        qw/  _switch_name
             _switch_timeout
             _vswitch_id
             _port_info_update_time
             _port_info
             _port_info_cache /,
              ],
                        ];

sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            switch_name             => { regex => qr/^[\w\-]+$/ },
            vswitch_id              => { regex => qr/^\d+$/,     optional => 1},
            node_target_ports_spec  => { type  => HASHREF,       optional => 1},
            vswitch_isl_target_spec => { type  => HASHREF,       optional => 1},
                  },
                            );

    my $switch_name = delete($opts{switch_name});

    # the rest of what might be in %opts is vswitch discovery related
    my %vswitch_opts = %opts;

    $self->_switch_name($switch_name);

    my $Host = host($switch_name);
    $self->_switch_apiset(NACL::APISet->new(
            hostobj   => $Host,
            category  => 'Switch',
            interface => 'CLI',
            set       => 'CiscoFC',
            # guess i still don't understand why cdef_path is needed.  what is
            # wrong with the default path where all other cdefs are found?
            cdef_path => '/usr/software/test/agnostic/lib/' .
                         'NACL/APISet/Switch/CLI/CiscoFC/default'
                                           ));

    $self->_disable_switch_timeout();

    my $vswitch_id = undef;
    if (exists($vswitch_opts{node_target_ports_spec}) ||
                               exists($vswitch_opts{vswitch_isl_target_spec})) {
        $vswitch_id = $self->_determine_vswitch_id(%vswitch_opts);
    } elsif (exists($vswitch_opts{vswitch_id})) {
        $vswitch_id = $vswitch_opts{vswitch_id};
    }

    # for cisco switches, we must pass the vsan id in queries from this point
    if (defined($vswitch_id)) {
        $self->_vswitch_id($vswitch_id);
    } else {
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "Could not determine the vsan ID for switch [$switch_name] !");
    }

    # this will become stale if the VSAN ID changes
    $self->_port_info_cache($self->_gather_port_info());
    $self->_port_info($self->_port_info_cache());

    return;
}

=head2 get_switch_APISet()

It gives you handle to the switch APISet object. With this handle
you can directly run commands to the switch

=over

=item Arguments

 None

=item Return Value

Handle to switch apiset

=back

    my $switch_apiset = $switch_impl->get_switch_APISet();

=cut

sub get_switch_APISet {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_apiset();
}


=head2 get_switch_name

It returns the name of the switch

=over

=item Arguments

 None

=item Return Value

name of the switch

=back

    my $switch_hostname = $switch_impl->get_switch_name();

=cut

sub get_switch_name {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_switch_name();
}

=head2 _disable_switch_timeout

Disables switch timeout if there is one

=over

=item Arguments

 None

=item Return Value

If timeout was NOT disabled, return timeout value
If timeout already disabled, return 0


=back

    my $current_switch_timeout = $switch_impl->_disable_switch_timeout();

=cut

sub _disable_switch_timeout {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my $cmd = 'show terminal';
    my $response = $self->_switch_apiset()->execute_raw_command(
                                          command => "$cmd");
    my ($timeout) = $response
                    =~ /Session Timeout:\s*(\d+) minutes/i;
    #NACL does not support this command yet...
    # my $response = $self->_switch_apiset()->show_terminal();
    #
    #NACL switch parsing broken, use raw output for now
    #my ($timeout) = $response->get_parsed_output()->[0]->
    #                {session_timeout}
    $self->_switch_timeout($timeout);

    if ($timeout) {
        #disable timeout by setting to 0
        $self->_switch_apiset()->terminal_session_timeout(timeout => 0);
        #NACL does not support this command yet...
        #my $response = $self->_switch_apiset()->show_terminal();
        $response = $self->_switch_apiset()->execute_raw_command(
                                          command => "$cmd");
        ($timeout) = $response
                    =~ /Session Timeout:\s*(\d+) minutes/i;
        if (!$timeout) {
            $Log->comment(
                $self->get_switch_name() . " switch timeout disabled");
        }
    }


    return $self->_switch_timeout;
}

=head2 get_description

    my $switch_description = $switch_impl->get_description();

This method returns the name and VSAN ID as a text string.

=over

=item Arguments

    None

=back

=cut

sub get_description {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->get_switch_name() . ' (VSAN: ' .
           ( join q{,}, keys %{$self->_vswitch_id()} ) . ')';
}

=head2 L_get_vendor()

    my $vendor_name = $switch_impl->L_get_vendor();

Returns the name of the switch vendor, e.g., brocade, cisco, etc.

=cut

sub L_get_vendor {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return 'cisco';
}

=head2 L_get_fcvi_port_descriptors()

    my $fcvi_port_descriptors = $switch_impl->L_get_fcvi_port_descriptors();
    foreach my $port_descriptor (@{ $fcvi_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the fcvi ports found on the switch.

Implementation detail: for Cisco switches, the port descriptor is the port's
port/slot combination, e.g., 1/1, 1/2, 1/4, etc.

=cut

sub L_get_fcvi_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my @port_descriptors = keys(%{ $self->_port_info_cache()->
                            {show_fcns_database_local_parsed}->{fcvi_ports} });

    return  \@port_descriptors;
}

=head2 L_get_storage_port_descriptors()

    my $storage_port_descriptors =
                                $switch_impl->L_get_storage_port_descriptors();
    foreach my $port_descriptor (@{ $storage_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Each filer at a site has connections to this switch which are used for data
I/O to and from the shared storage array.  This method returns an arrayref of
port descriptor strings associated with such ports.

Do not confuse these ports with ports out to the fibre-bridge array (see
get_fibreBridge_ports().

Implementation detail: for Cisco switches, the port descriptor is the port's
port/slot combination, e.g., 1/1, 1/2, 1/4, etc.

=cut

sub L_get_storage_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors = keys(%{ $self->_port_info_cache()->
                         {show_fcns_database_local_parsed}->{storage_ports} });

    return  \@port_descriptors;
}

=head2 L_get_fibreBridge_port_descriptors()

    my $fibreBridge_port_descriptors =
                            $switch_impl->L_get_fibreBridge_port_descriptors();
    foreach my $port_descriptor (@{ $fibreBridge_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the fcvi ports found on the switch.

Implementation detail: for Cisco switches, the port descriptor is the port's
port/slot combination, e.g., 1/1, 1/2, 1/4, etc.

=cut

sub L_get_fibreBridge_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors = keys(%{ $self->_port_info_cache()->
                     {show_fcns_database_local_parsed}->{fibreBridge_ports} });

    return  \@port_descriptors;
}

=head2 L_get_isl_port_descriptors()

    my $isl_port_descriptors = $switch_impl->L_get_isl_port_descriptors();
    foreach my $port_descriptor (@{ $isl_port_descriptors }) {
        my $port = WaflMCC::MCC::Netork::Port->new(
                        port_descriptor => $_, switch_impl => $switch_impl);
        ...
    }

Returns an arrayref of port descriptor strings associated with
the ISL ports found on the switch.

Implementation detail: for Cisco switches, the port descriptor is the port's
port/slot combination, e.g., 1/1, 1/2, 1/4, etc.

=cut

sub L_get_isl_port_descriptors {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my @port_descriptors = keys(%{
             $self->_port_info_cache()->{show_fcs_DATA_parsed}->{isl_ports} });

    return  \@port_descriptors;
}


=head2 _get_port_vsan()

This private method determines the vsan a port is in

=cut

sub _get_port_vsan {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr{^\d+/\d+$} },
                      port_info       => { type => HASHREF, optional => 1 },
                  },
                            );

    my $port_descriptor = delete($opts{port_descriptor});

    my $vsan;

    my $port_info = $opts{port_info} || $self->_port_info_cache();

    if (defined $port_info->{show_interface_brief_parsed}->
                {$port_descriptor}->{vsan}) {
        $vsan = $port_info->{show_interface_brief_parsed}->
                {$port_descriptor}->{vsan};
    }

    return $vsan;
}

=head2 _gather_port_info()

this get some pretty raw data about ports and returns it as a hashref with
the following structure:

    [NB: "meta" tags are not raw data from the switch queries, but are
         associated with data which has been processed in some manner
         before being explicitly placed in these data structures.]

    {
        show_fcns_database_local_raw    => 'a multiline string',
        show_fcns_database_local_parsed => {
            fcvi_ports => {
                $port_descriptor => {
                    'fabric_port_wwn' => '20:0e:54:7f:ee:05:08:40',
                    'fc4_types' => 'scsi-fcp:init fc-vi',
                    'connected_interface' => 'fc1/14',
                    'port_ip_addr' => '0.0.0.0',
                    'node_wwn' => '20:00:00:24:ff:31:ed:a2',
                    'switch_name_ip_address' =>
                                'dpg-mcc-cisco9148-03-A-fabX (10.228.160.217)',
                    'hard_addr' => '0x000000',
                    'symbolic_port_name' => 'Netapp FCVI adapter  1a 0',
                    'node_ip_addr' => '0.0.0.0',
                    'ipa' => 'ff ff ff ff ff ff ff ff',
                    'permanent_port_wwn_vendor' => '21:00:00:24:ff:31:ed:a2',
                    'class' => '3',
                    'port_type' => 'N',
                    'symbolic_node_name' => '',
                    'port_wwn_vendor' => '21:00:00:24:ff:31:ed:a2',
      (meta)        'meta__connection_target' => '21:00:00:24:ff:31:ed:a2'
                                    },
                $port_descriptor => {
                    . . .
                                    },
                . . .
                          },
            fibreBridge_ports => {
                $port_descriptor => {
                    'fabric_port_wwn' => '20:05:54:7f:ee:05:08:40',
                    'fc4_types' => 'scsi-fcp:target',
                    'connected_interface' => 'fc1/5',
                    'port_ip_addr' => '0.0.0.0',
                    'node_wwn' => '20:00:00:10:86:60:31:9e',
                    'switch_name_ip_address' =>
                                'dpg-mcc-cisco9148-03-A-fabX (10.228.160.217)',
                    'hard_addr' => '0x000000',
                    'symbolic_port_name' => '',
                    'node_ip_addr' => '0.0.0.0',
                    'ipa' => 'ff ff ff ff ff ff ff ff',
                    'permanent_port_wwn_vendor' => '21:00:00:10:86:60:31:9e',
                    'class' => '3',
                    'port_type' => 'N',
                    'symbolic_node_name' => 'ATTO FibreBridge 6500N A64P',
                    'port_wwn_vendor' => '21:00:00:10:86:60:31:9e'
      (meta)        'meta__connection_target' => '21:00:00:10:86:60:31:9e'
                                    },
                $port_descriptor => {
                    . . .
                                    },
                . . .
                                 },
            storage_ports => {
                $port_descriptor => {
                    'fabric_port_wwn' => '20:03:54:7f:ee:05:08:40',
                    'fc4_types' => 'scsi-fcp:init 248',
                    'connected_interface' => 'fc1/3',
                    'port_ip_addr' => '0.0.0.0',
                    'node_wwn' => '20:00:00:1b:32:8e:1e:94',
                    'switch_name_ip_address' =>
                                'dpg-mcc-cisco9148-03-A-fabX (10.228.160.217)',
                    'hard_addr' => '0x000000',
                    'symbolic_port_name' =>
                             'NetApp FC Storage Adapter dpg-mcc-3270-03-a1:4a',
                    'node_ip_addr' => '0.0.0.0',
                    'ipa' => 'ff ff ff ff ff ff ff ff',
                    'permanent_port_wwn_vendor' =>
                                            '21:00:00:1b:32:8e:1e:94 (Qlogic)',
                    'class' => '3',
                    'port_type' => 'N',
                    'symbolic_node_name' => '',
                    'port_wwn_vendor' => '21:00:00:1b:32:8e:1e:94 (Qlogic)'
      (meta)        'meta__connection_target' => 'dpg-mcc-3270-03-a1:4a'
                                    },
                $port_descriptor => {
                    . . .
                                    },
                . . .
                             },
                                           },
        show_fcs_database_raw    => 'a multiline string',
        show_fcs_ie_raw          => 'a multiline string',
        show_fcs_DATA_parsed => {
            isl_ports => {
                $port_descriptor => {
                    'interface' => 'fc1/15',
                    'type' => 'TE',
                    'attached_pwwn' => '20:0f:54:7f:ee:09:38:70',
                    'fwwn' => '20:0f:54:7f:ee:05:08:40'
      (meta)        'meta__target_switch_ip' => '10.228.160.219',
      (meta)        'meta__target_fabric_wwn' => '20:16:54:7f:ee:09:38:71',
      (meta)        'meta__target_switch_name' => 'dpg-mcc-cisco9148-03-B-fabX',
                                    },
                $port_descriptor => {
                    . . .
                                    },
                         },
                                    },
        show_interface_brief_parsed => {
                                          '1/11' => {
                                                  'port_channel' => '--',
                                                  'oper_speed_gbps' => '',
                                                  'status' => 'sfpAbsent',
                                                  'sfp' => '--',
                                                  'interface' => '1/11',
                                                  'admin_truck_mode' => 'on',
                                                  'vsan' => '22',
                                                  'admin_mode' => 'F',
                                                  'oper_mode' => '--'
                                                },
                                          '1/34' => {
                                                  'port_channel' => '--',
                                                  'oper_speed_gbps' => '',
                                                  'status' => 'sfpAbsent',
                                                  'sfp' => '--',
                                                  'interface' => '1/34',
                                                  'admin_truck_mode' => 'on',
                                                  'vsan' => '4094',
                                                  'admin_mode' => 'F',
                                                  'oper_mode' => '--'
                                                },
                                          . . .
        },
        show_interface_brief_raw    => 'a multiline string',

    }

=cut

sub _gather_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $port_info = { show_fcns_database_local_raw    => undef,
                      show_fcns_database_local_parsed => {
                                                     fcvi_ports        => { },
                                                     storage_ports     => { },
                                                     fibreBridge_ports => { },
                                                     other_ports       => { },
                                                         },
                      show_fcs_database_raw           => undef,
                      show_fcs_ie_raw                 => undef,
                      # show_fcs_DATA_parsed is a composite of fcs queries
                      show_fcs_DATA_parsed            => {
                                                     isl_ports         => { },
                                                         },
                      show_interface_brief_raw        => undef,
                      show_interface_brief_parsed     => undef,
                    };

    ##############################################
    # for port state
    ##############################################
    my $response = $self->_switch_apiset()->show_interface_brief();
    $port_info->{show_interface_brief_raw} = $response->get_raw_output();

    # %parsed contains information about each port
    my $parsed_lines = $response->get_parsed_output();
    my %parsed = map { $_->{interface} =~ s/^[^\d]+//; ($_->{interface}, $_) }
                     @{$parsed_lines}; # quick array -> hash
    $port_info->{show_interface_brief_parsed} = \%parsed;


    ##############################################
    # for FCVI, storage, and fibre bridge ports...
    ##############################################
    $response = $self->_switch_apiset()->show_fcns_database_local(
                                    detail => 1);
    $port_info->{show_fcns_database_local_raw} = $response->get_raw_output();

    # $parsed contains information about each port
    $parsed_lines = $response->get_parsed_output();

    # for cisco, port slot/index is our "port descriptor"
    foreach my $parsed_line (@{ $parsed_lines }) {
        if (defined($parsed_line->{fcid_info}->[0]->{connected_interface}) &&
                   (defined  $self->_vswitch_id()->{$parsed_line->{vsan}})) {
            my $fcid_info = $parsed_line->{fcid_info}->[0];

            my $port_descriptor = $fcid_info->{connected_interface};
            $port_descriptor =~ s/^fc//;    # removes 'fc' prefix

            # look at symbolic_port_name or symbolic_node_name?
            my $raw_symb_entry = (defined($fcid_info->{symbolic_port_name}) &&
                                  $fcid_info->{symbolic_port_name}) ?
                                             $fcid_info->{symbolic_port_name} :
                                             $fcid_info->{symbolic_node_name};
            my ($type) = ($raw_symb_entry =~ /(Storage|FibreBridge|FCVI)/);

            my $connection_target = undef;
            my $ports_type = undef;
            if (defined($type)) {
                if ($type eq 'Storage') {
                    # gets target node_name:port
                    ($connection_target) =
                               ($raw_symb_entry =~ /Adapter\s+([\w\-]+\:\w+)/);
                    $ports_type = 'storage_ports';
                } elsif ($type eq 'FibreBridge') {
                    # WWN only available.  is this the correct one?  here are
                    # the possibilities: fabric_port_wwn
                    #                    node_wwn
                    #                    permanent_port_wwn_vendor
                    #                    port_wwn_vendor
                    ($connection_target) =
                                     $fcid_info->{permanent_port_wwn_vendor};
                    $ports_type = 'fibreBridge_ports';
                } elsif ($type eq 'FCVI') {
                    # WWN only available.  same issue as above re which wwn?
                    ($connection_target) =
                                     $fcid_info->{permanent_port_wwn_vendor};
                    $ports_type = 'fcvi_ports';
                }
            }
            # sanity check
            unless (defined($connection_target)) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                          "connection_target undefined!\n" .
                          "    raw_symb_entry = $raw_symb_entry\n");
            }

            # add metadata to line hash
            $fcid_info->{meta__connection_target} = $connection_target;

            # save line hash in return structure
            $port_info->{show_fcns_database_local_parsed}->
                                {$ports_type}->{$port_descriptor} = $fcid_info;
        }
    }

    ##############################################
    # for isl ports...
    ##############################################
    $response = $self->_switch_apiset()->show_fcs_database();
    $port_info->{show_fcs_database_raw} = $response->get_raw_output();

    # $parsed contains information about each port
    $parsed_lines = $response->get_parsed_output();

    foreach my $parsed_line (@{ $parsed_lines }) {
        if (defined($parsed_line->{switch_ports}) &&
                   (defined  $self->_vswitch_id()->{$parsed_line->{vsan}})) {
            foreach my $interface_info (@{ $parsed_line->{switch_ports} }) {
                if ($interface_info->{type} =~ /^T?E$/) {
                    my $port_descriptor = $interface_info->{interface};
                    $port_descriptor =~ s/^fc//;    # removes 'fc' prefix

                    $port_info->{show_fcs_DATA_parsed}->
                             {isl_ports}->{$port_descriptor} = $interface_info;
                }
            }
        }
    }

    # supplement isl_port info with target switch related metadata

    # this is a workaround because there is no NACL call to execute
    # 'show fcs ie' yet (to get an already parsed data struct)
    my $fcs_ie_parsed = $self->_show_fcs_ie_query_workaround();

    # sanity check
    my @target_fields =
                   qw/ target_switch_name target_switch_ip target_fabric_wwn /;

    # see POD above for what it looks like after meta__XXX fields are inserted
    my $isl_port_info = $port_info->{show_fcs_DATA_parsed}->{isl_ports};
    foreach my $isl_port_desc (keys(%{ $isl_port_info })) {
        my $vsan = $self->_get_port_vsan(port_descriptor => $isl_port_desc,
                                         port_info       => $port_info,);
        foreach my $field (@target_fields) {
            unless (defined($fcs_ie_parsed->{$vsan}->{$field})) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                        "Could not determine ISL target switch:\n\n" .
                        $fcs_ie_parsed->{show_fcs_ie_raw});
            }
        }
        foreach my $field (@target_fields) {
            my $meta_key_string = "meta__$field";
            $isl_port_info->{$isl_port_desc}->{$meta_key_string} =
                                             $fcs_ie_parsed->{$vsan}->{$field};
            # e.g., $isl_port_info->{'1/15'}->{meta__target_switch_name} =
            #                        $fcs_ie_parsed->{22}->{target_switch_name}
        }
    }

    $self->_port_info_update_time(time); #update the last gathered time
    return $port_info;
}

=head2 _show_fcs_ie_query_workaround()

This is a workaround until NACL CiscoFC APISet supports the 'show fcs ie' query.

Returns a data struct of the following format:

    {
        show_fcs_ie_raw    => $unparsed raw string,
        $vsan_id           => {
            local_fabric_wwn   => '20:21:54:7f:ee:05:08:41',
            target_switch_name => 'dpg-mcc-cisco9148-03-B-fabX',
            target_switch_ip   => '10.228.160.219',
            target_fabric_wwn  => '20:16:54:7f:ee:09:38:71',
                              },
        $another_vsan_id   => {
            . . .
                              },
    }

If vsan option is passed, there will only be a key/value pair associated with
that particular vsan id in the returned data struct.

=cut

sub _show_fcs_ie_query_workaround {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      vsan => { regex => qr/^\d+$/, optional => 1 },
                  },
                            );

    my $raw_response = undef;
    unless (WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/03/12', burt => '626305')) {

        my $cmd = 'show fcs ie';
        if (defined($opts{vsan})) {
            $cmd .= ' vsan ' . $opts{vsan};
        }

        $raw_response = $self->_switch_apiset()->execute_raw_command(
                                          command => "$cmd");
    }

    my $fcs_ie_parsed = {
                            show_fcs_ie_raw => $raw_response,
                        };

    my @lines = split(/\n/, $raw_response);
    my $vsan = undef;
    while (defined(my $line = shift(@lines))) {

        # parsing something like this:
        #
        # . . .
        # IE List for VSAN: 22
        # ----------------------------------------------------------------------
        # IE-WWN                   IE     Mgmt-Id  Mgmt-Addr (Switch-name)
        # ----------------------------------------------------------------------
        # 20:16:54:7f:ee:05:08:41  S(Loc) 0xfffc57 10.228.160.217
        #                                        (dpg-mcc-cisco9148-03-A-fabX)
        # 20:16:54:7f:ee:09:38:71  S(Adj) 0xfffccf 10.228.160.219
        #                                        (dpg-mcc-cisco9148-03-B-fabX)
        # [Total 2 IEs in Fabric]
        #
        # IE List for VSAN: 33
        # ----------------------------------------------------------------------
        # IE-WWN                   IE     Mgmt-Id  Mgmt-Addr (Switch-name)
        # ----------------------------------------------------------------------
        # 20:21:54:7f:ee:05:08:41  S(Loc) 0xfffc00 10.228.160.217
        #                                        (dpg-mcc-cisco9148-03-A-fabX)
        # [Total 1 IEs in Fabric]
        #  . . .

        # determine the vsan id
        if ($line =~ m{^\s*IE List for VSAN:\s*(\d+)}) {
            $vsan = $1;
            $fcs_ie_parsed->{$vsan} = { };
        } elsif ($line =~ m{^\s* ([\w:]+) \s+ S[(]Loc[)]}x) {
            $fcs_ie_parsed->{$vsan}->{local_fabric_wwn} = $1;
        } elsif ($line =~
                  m{^\s* ([\w:]+) \s+ S[(]Adj[)] \s+ \S+ \s+ ([\d.]+) \s*$}x) {
            $fcs_ie_parsed->{$vsan}->{target_fabric_wwn} = $1;
            $fcs_ie_parsed->{$vsan}->{target_switch_ip}  = $2;

            # next line must be target switch name
            $line = shift(@lines);
            ($fcs_ie_parsed->{$vsan}->{target_switch_name}) =
                                      ($line =~ m{^\s* [(] ([^)]+) [)] \s*$}x);
        }
    }

    return $fcs_ie_parsed;
}

=head2 _port_enable_or_disable()

This private method is the common code for enabling or disabling a port.

=cut

sub _port_enable_or_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr{^\d+/\d+$} },
                      action          => { regex => qr{^(enable|disable)$} },
                      timeout         => { regex => qr{^\d+$} },
                      interval        => { regex => qr{^\d+$} },
                  },
                            );

    my $port_descriptor = delete($opts{port_descriptor});
    my $port_full_name  = 'fc' . $port_descriptor;
    my $action          = delete($opts{action});
    my $timeout         = delete($opts{timeout});
    my $interval        = delete($opts{interval});

    my $apiset = $self->_switch_apiset();

    # enable or disable the port.  via CLI, these commnads are:
    #   configure terminal
    #   interface fc <slot>/<port>
    #   no shutdown | shutdown    # enable vs disable
    #   end
    $apiset->configure(subsystem =>'terminal');
    $apiset->interface_fc(port => $port_descriptor);
    if ($action eq 'enable') {
        $apiset->shutdown(deconfigure => 1);
    } else {
        $apiset->shutdown();
    }
    $apiset->end();

    # this block takes care of making sure its enabled else time out
    my $timeout_end = time() + $timeout;
    my $success = 0;
    my $success_msg = "===================================================\n" .
                      "======= Port $port_descriptor is ${action}d! ======\n" .
                      "===================================================\n";
    while ((! $success) && (time() <= $timeout_end)) {
        my $response = $apiset->show_interface_fc(fc_portid=> $port_full_name);
        my $parsed = $response->get_parsed_output();

        foreach my $state_info (@{ $parsed }) {
            if (exists($state_info->{$port_full_name})) {
                my $state = $state_info->{$port_full_name};
                if (     (($action eq 'enable') && ($state =~ /up|trunking/)) ||
                         (($action eq 'disable') && ($state =~ /down/)) ) {
                    $Log->comment($success_msg);
                    $success = 1;
                } else {
                    $Log->comment(
                      "Waiting for port: $port_descriptor to be enabled...\n");
                    Tharn::snooze($interval);
                }
                last;
            }
        }

        # are we done?
        $success && last;
    }

    unless ($success) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
             "Timed out waiting for port: $port_descriptor to be\n" .
             "${action}d! Please login to the switch for more Info.. \n");
    }

    return;
}

=head2 L_update_current_port_info

updates current port info

=over

=item Arguments

=item Return Value

None

=back

    $switch_impl->L_update_current_port_info();

=cut

sub L_update_current_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    return $self->_port_info($self->_gather_port_info());

}

=head2 _get_current_port_info

Gets new or recent port info

=over

=item Arguments

  Required:
  Optional: <stale_time> int val

=item Return Value

None

=back

    $switch_impl->_get_current_port_info(stale_time => 2);

=cut

sub _get_current_port_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      stale_time   => { regex => qr/^\d+$/, default => 30 },
                  },
                            );

    my $last_time = $self->_port_info_update_time();

    if (time > $opts{stale_time} + $last_time) {
        $self->L_update_current_port_info();
    }

    return $self->_port_info();

}

=head2 L_port_is_enabled

This returns whether port <port_descriptor> is enabled.

=over

=item Arguments

  Required: <port_descriptor> int val
  Optional: <stale_time> int val

=item Return Value

1|0

=back

    $switch_impl->L_port_enable(port_descriptor => '0');
    $switch_impl->L_port_enable(port_descriptor => '0', stale_time=>2);

=cut

sub L_port_is_enabled {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr{^\d+/\d+$} },
                      stale_time      => { optional => 1 },
                  },
                            );
    my $enabled = 0;
    my $port_descriptor = delete $opts{port_descriptor};
    my $port_info =
        $self->_get_current_port_info( %opts );


    #print "PORT INFO:" . Dumper $port_info;

    if (defined $port_info->{show_interface_brief_parsed}->{$port_descriptor}) {
        my $state = $port_info->{show_interface_brief_parsed}->
            {$port_descriptor}->{status};

        if ($state =~ /up|trunking/) {
            $enabled = 1;
        }
    }

   return $enabled;


}

=head2 L_port_enable

This library-only method attempts to enable port_descriptor and checks if if
the port has been enabled successfully.

=over

=item Arguments

  Required: port_descriptor slot/port_id
  Optional: timeout int val
  Optional: interval int val

=item Return Value

None

=back

    $switch_impl->L_port_enable(port_descriptor => '1/3');

=cut

sub L_port_enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr{^\d+/\d+$} },
                      timeout         => { regex => qr{^\d+$}, default => 30 },
                      interval        => { regex => qr{^\d+$}, default =>  5 },
                  },
                            );

    return $self->_port_enable_or_disable(action => 'enable', %opts);
}

=head2 L_port_disable

This library-only method attempts to disable port_descriptor and checks if if
the port has been disabled successfully.

=over

=item Arguments

  Required: port_descriptor int val
  Optional: timeout int val
  Optional: interval int val

=item Return Value

None

=back

    $switch_impl->L_port_disable(port_descriptor => '0');

=cut

sub L_port_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      port_descriptor => { regex => qr{^\d+/\d+$} },
                      timeout         => { regex => qr{^\d+$}, default => 30 },
                      interval        => { regex => qr{^\d+$}, default =>  5 },
                  },
                            );

    return $self->_port_enable_or_disable(action => 'disable', %opts);
}

=head2 _enable_or_disable()

This private method is the common code for enabling or disabling this switch
vsan.

=cut

sub _enable_or_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      action          => { regex => qr{^(enable|disable)$} },
                      timeout         => { regex => qr{^\d+$} },
                      interval        => { regex => qr{^\d+$} },
                  },
                            );

    my $action   = delete($opts{action});
    my $timeout  = delete($opts{timeout});
    my $interval = delete($opts{interval});

    my $apiset       = $self->_switch_apiset();
    my $vsan_id      = $self->_vswitch_id();
    my $suspend_flag = ($action eq 'disable') ? 1 : 0;

    # enable or disable the port.  via CLI, these commnads are:
    #   configure terminal
    #   vsan database
    #   [no] vsan <vsan_id> suspend    # [enable] vs disable
    #   end
    $apiset->configure(subsystem =>'terminal');
    $apiset->vsan_database();
    #$apiset->vsan('vsan-id' => $vsan_id, suspend => $suspend_flag);

    # workaround for NACL APISet issue
    if (! $suspend_flag) {
        unless (WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2012/10/31', burt => '627837')) {
            $apiset->execute_raw_command(command => "no vsan $vsan_id suspend");
        }
    } else {
        # this is how the command should look no matter the value of
        # $suspend_flag
        $apiset->vsan('vsan-id' => $vsan_id, suspend => $suspend_flag);
    }
    $apiset->end();

    # this block takes care of making sure its enabled else time out
    my $timeout_end = time() + $timeout;
    my $success = 0;
    my $label =
        'Switch ' . $self->get_switch_name() . ' vsan ' . $self->_vswitch_id();
    my $success_msg = "===================================================\n" .
                      "======= $label is ${action}d! ======\n" .
                      "===================================================\n";
    while ((! $success) && (time() <= $timeout_end)) {
        my $response = $apiset->show_vsan();
        my $parsed = $response->get_parsed_output();

        foreach my $state_info (@{ $parsed }) {
            if (exists($state_info->{vsan_id}) &&
                                        ($state_info->{vsan_id} == $vsan_id)) {
                my $state = $state_info->{operational_state};
                if (     (($action eq 'enable')  && ($state =~ /up/)) ||
                         (($action eq 'disable') && ($state =~ /down/)) ) {
                    $Log->comment($success_msg);
                    $success = 1;
                } else {
                    $Log->comment("Waiting for $label to be ${action}d...\n");
                    Tharn::snooze($interval);
                }
                last;
            }
        }

        # are we done?
        $success && last;
    }

    unless ($success) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
             "Timed out waiting for $label to be\n" .
             "${action}d! Please login to the switch for more Info.. \n");
    }

    return;
}

=head2 L_enable

This library-only method enables the vsan associated with this switch object
and then verifies it has been enabled.

=over

=item Arguments

  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

=cut

sub L_enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      timeout  => { regex => qr/^\d+$/, default => 30 },
                      interval => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    return $self->_enable_or_disable(action => 'enable', %opts);
}

=head2 L_disable

This library-only method disables the vsan associated with this switch object
and then verifies it has been disabled.

=over

=item Arguments

  Optional: <timeout> int val
  Optional: <interval> int val

=item Return Value

None

=back

=cut

sub L_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      timeout  => { regex => qr/^\d+$/, default => 30 },
                      interval => { regex => qr/^\d+$/, default =>  5 },
                  },
                            );

    return $self->_enable_or_disable(action => 'disable', %opts);
}

=head2 L_get_port_target_host_hints()

    my $port_connection_target_hints =
                 $switch_impl->L_get_port_target_host_hints(
                                          port_descriptor => $port_descriptor);

(Library method) Returns a hash of information about the host on the remote
side of the connection from the port described by port_descriptor.  This
information may not be completely reliable.  Some of it seems to be data
entered by hand by a network admin instead of being auto-generated by the
switch.  Thus the 'hints' in the method name.

This hash has the following form:

    {
        hostname     => 'dpg-mcc-brcd6510-01-A-fabY',
        host_port_id => '0d',
        host_type    => ('switch'|'node'|'fibreBridge'),
        WWN          => '21:00:00:24:ff:32:06:90',
    }

=over

=item Options

=over

=item C<< port_descriptor >>

(Required) The text identifier of a port on this switch.  In this case, it is
the port slot/index.

=back

=back

=cut

sub L_get_port_target_host_hints {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            port_descriptor => { regex => qr{^\d+/\d+$} },
                  },
                            );

    my $port_desc = delete($opts{port_descriptor});

    my $fcns_info= $self->_port_info_cache()->{show_fcns_database_local_parsed};
    my $fcs_info = $self->_port_info_cache()->{show_fcs_DATA_parsed};

    my $fcvi_info    = $fcns_info->{fcvi_ports};
    my $storage_info = $fcns_info->{storage_ports};
    my $fb_info      = $fcns_info->{fibreBridge_ports};

    my $isl_info     = $fcs_info->{isl_ports};

    # some hints will remain undef depending on port type
    my ($hostname, $host_port_id, $host_type, $WWN) =
                                                  (undef, undef, undef, undef);
    if (exists($fcvi_info->{$port_desc})) {
        # FCVI port
        $host_type = 'node';
        $WWN       = $fcvi_info->{$port_desc}->{meta__connection_target};
    } elsif (exists($storage_info->{$port_desc})) {
        # storage port
        ($hostname, $host_port_id) =
            split(/:/, $storage_info->{$port_desc}->{meta__connection_target});
        $host_type = 'node';
        $WWN       = $storage_info->{$port_desc}->{permanent_port_wwn_vendor};
    } elsif (exists($fb_info->{$port_desc})) {
        # fibre-bridge port
        $host_type = 'fibreBridge';
        $WWN       = $fb_info->{$port_desc}->{meta__connection_target};
    } elsif (exists($isl_info->{$port_desc})) {
        # ISL port
        $host_type        = 'switch';
        $hostname = $isl_info->{$port_desc}->{meta__target_switch_name};
        $WWN = $isl_info->{$port_desc}->{meta__target_fabric_wwn};
    } else {
        my $msg = "L_get_port_target_host_hints: No port information found \n" .
                  "    for port: $port_desc\n";
        $Log->warn($msg);
    }

    my $target_info = { hostname     => $hostname,
                        host_port_id => $host_port_id,
                        host_type    => $host_type,
                        WWN          => $WWN, };

    return $target_info;
}

=head2 _determine_vswitch_id()

This is used to determine an VSAN ID associated with an FCVI port known to
connect to this virtual switch from a filer node.  This method requires that
one of these two parameters must be passed:
    node_target_ports_spec XOR vswitch_isl_target_spec.

    my $vswitch_id = $self->_determine_vswitch_id(
                                node_target_ports_spec => $vswitch_port_spec);

    where $vswitch_port_spec is a hashref with the following format:

        {
            port_descriptor => '1-3',
            port_wwn        => '21:00:00:24:ff:32:06:90',
        }

    (Beware that the '-' in '1-3' is how [slot]-[port_id] appear in node's
     sysconfig output, but the same port is referred to as '1/3' when looking
     at switch output.)


This is used to determine an VSAN ID associated with being the target across an
ISL from another switch in the same fabric.

    my $vswitch_id = $self->_determine_vswitch_id(
                       vswitch_isl_target_spec => $vswitch_target_wwn_spec);

    where $vswitch_target_wwn is the target fabric WWN of a particular ISL.


(Private method) Returns the VSAN ID of the virtual fabric associated with
either of the conditions described in the two cases above.

=over

=item Options

=over

=item C<< node_target_ports_spec >>

(Required parameter as such: node_target_ports_spec XOR vswitch_isl_target_spec)
node_target_ports_spec is a hashref specifying a port in the virtual fabric.
This hasref format is as described above.

=item C<< vswitch_isl_target_spec >>

(Required parameter as such: vswitch_isl_target_spec XOR node_target_ports_spec)
vswitch_isl_target_spec is a string containg a valid WWN.

=item Dev Notes

See _gather_port_info() for information about what data is
returned from the show_fcs_database() and show_fcs_ie() queries on the switch.

=back

=back

=cut

sub _determine_vswitch_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_target_ports_spec  => { type => HASHREF, optional => 1 },
            vswitch_isl_target_spec => { type => HASHREF, optional => 1 },
                  },
                            );
    my $vswitch_ids;
    if (scalar(grep {$_ =~ /^node|vswitch/} keys(%opts)) == 1) {
        if (exists($opts{node_target_ports_spec})) {
            my $potential_port_spec = $self->_filter_ports(%opts);

            foreach my $port_key (sort keys %{ $potential_port_spec }) {
                my $port_spec = $potential_port_spec->{$port_key};

                my $target_port_id = $port_spec->{port_descriptor};

                # translate port ID
                # from syscnfig rep (1-4) to switch rep (fc1/4)
                $target_port_id =~ s/-/\//;
                $target_port_id = 'fc' . $target_port_id;

                # now look for this port number in the output of
                # 'show fcs database'
                my $response = $self->_switch_apiset()->show_fcs_database();
                my $parsed_lines = $response->get_parsed_output();

                foreach my $element (@{ $parsed_lines }) {
                    if (exists($element->{switch_ports})) {
                        for my $port_element (@{ $element->{switch_ports} }) {
                            if ($port_element->{interface} eq $target_port_id) {
                                $vswitch_ids->{$element->{vsan}} = 1;
                                                 #->{$port_key} = $port_spec;
                            }
                        }
                    }
                }
            }
            unless (scalar keys %{$vswitch_ids}) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                    "Can't determine the virtual fabric ID(s) for ports\n\n" .
                    Dumper($potential_port_spec) . "\n\n" .
                    "in switch " . $self->get_switch_name() . " !");
            }
        } elsif (exists($opts{vswitch_isl_target_spec})) {
            my $target_spec = delete($opts{vswitch_isl_target_spec});

            # here we want to step through all vsans on this switch and find
            # the vsan with a local fabric wwn which matches the passed
            # vswitch_isl_target_spec.  this will mean we have found the
            # correct vsan on this switch which is associated with the
            # vsan in the already constructed switch in the intersite fabric.

            my $fcs_ie_parsed = $self->_show_fcs_ie_query_workaround();
            foreach my $key (keys(%{ $fcs_ie_parsed })) {
                next if $key eq 'show_fcs_ie_raw';
                if (defined($fcs_ie_parsed->{$key}->{local_fabric_wwn}) &&
                    defined $target_spec->{$fcs_ie_parsed->{$key}->
                            {local_fabric_wwn}} ) {
                    $vswitch_ids->{$key} = 1; # = $fcs_ie_parsed->{$key};
                }
            }

            unless (scalar keys %{$vswitch_ids}) {
                WaflMCC::MCC::Exceptions::BaseException->throw(
                    "Can't determine the virtual fabric ID(s) for ISL(s)\n\n" .
                    "target WWN(s)/switch:" . Dumper $target_spec
                    . " !");
            }
        }
    } else {
        # both parameters were passed.  not valid.
        WaflMCC::MCC::Exceptions::BaseException->throw(
            "Both node_target_ports_spec and vswitch_isl_target_spec\n" .
            "parameters were passed.  Invalid call on this method.");
    }

    return $vswitch_ids;
}

=head2 _filter_ports()


returns hashref with potential matching ports for switch name

=over

=item Options

=over

=item None.

=back

=back

=cut

sub _filter_ports {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_target_ports_spec => { type => HASHREF,        optional => 1 },
                  },
    );

    my $ports_spec = $opts{node_target_ports_spec};
    my $name       = $self->get_switch_name();

    my $ports;

    foreach my $type (keys %{$ports_spec} ) {
        foreach my $spec (@{$ports_spec->{$type}}) {
            if ($spec->{switch_name} eq $name) {
                $spec->{type} = $type;
                $spec->{port_descriptor} =~ /(\d+)-(\d+)/;
                my $descriptor = "$1/$2";
                $ports->{$descriptor} = $spec;
            }
        }
    }
    return $ports;
}


=head2 get_vswitch_id()

    my $vswitch_id = $switch_impl->get_vswitch_id();

Returns the current VSAN ID.

=over

=item Options

=over

=item None.

=back

=back

=cut

sub get_vswitch_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            # no valid parameters
                  },
                            );

    return $self->_vswitch_id();
}

1;
