# $Id $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Network::Topology Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Network::Topology;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Network::Fabric ();
use WaflMCC::MCC::Network::Node ();

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use Tharn qw($Log);

use Params::Validate
    qw(validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use Readonly;
use Data::Dumper;


Readonly my $TO_BE_DETERMINED => undef;


=head1 NAME

WaflMCC::MCC::Network::Topology

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.  This class defines
the network hardware mapping of an MCC DR Group.  It is used to gain access
to network entities such as fabrics, switches, ports, etc.


    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();
    my $fabric = $topology->get_intersite_fabrics()->[0];
    my $switch = $fabric->get_switch(site => 'A');

    ...
    # Perform test here
    ...

    # End test

=cut

=head1 Public METHODS

There is no public constructor for this class.  A
WaflMCC::MCC::Network::Topology object is obtained via calls such as

    my $topology = $mcc->get_DR_groups()->{$group_id}->get_network_topology();

=cut

# automatically generate a (Non-public) constructor (includes a call to 'init')
# and (Non-public) member accessor methods
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'WaflMCC::MCC::DR_Group' },
                '_dr_group',
              ],
    scalar => [
        qw/ _fabrics _fiber_bridges /,
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    $Log->enter();
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group     => { isa   => 'WaflMCC::MCC::DR_Group' },
        },
                            );

    $self->_dr_group($opts{'dr_group'});
    $self->_fabrics( $self->_discover_intersite_fabrics());

    $Log->exit();
    return;
}

=head2 _discover_intersite_fabrics()

This method performs a discovery operation to determine the intersite fabrics
associated with this DR Group network topology.  Please see

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for (hopefully) clarification on what such a fabric is.

Returns an arrayref containing WaflMCC::MCC::Network::Fabric objects.

Designer's Notes: This topology will consist of two redundant fabrics which
span between sites A and B.  Each of these fabrics will consist of a pair
of real or virtual switches, one at each site and connected to each other via
their ISLs.  Since these switches may be running any number of virtual
switches that may be used either in multiple DR Groups in the same MCC system,
or by multiple MCC systems, we need to discover which switches (real or
virtual) link to each other to form a particular fabric.  This method itself
makes use of the WaflMCC::MCC::Network::Switch constructor which specifies
the parameter 'vswitch_fcvi_port_spec' in order to determine a switch that is
being used by a particular filer node (A1 in the DR Group owning this topology.)

This ties one end of one fabric to this shared storage topology.  In order to
find the Switch on the other end (i.e., at site B) of this fabric, we will pass
the site A switch to the WaflMCC::MCC::Network::Fabric constructor.  The
WaflMCC::MCC::Network::Fabric constructor will make use of the
WaflMCC::MCC::Network::Switch constructor which specifies the parameter
'vswitch_isl_target_wwn' whose value can be found by querying the site A
switch.  See the WaflMCC::MCC::Network::Fabric constructor for more
details on this process.

In the end, each fabric will be composed of the pair of linked real or virtual
switches spanning sites A and B.

This shared storage network topology is designed so that ALL FABRIC AND SWITCH
OPERATIONS ON THESE PAIRS OF VIRTUAL SWITCHES WILL MONITOR OR AFFECT ONLY THE
PORTS ASSOCIATED WITH THESE VIRTUAL SWITCHES.  ALL OTHER VIRTUAL SWITCHES
RUNNING ON THE ASSOCIATED REAL SWITCHES WILL REMAIN UNAFFECTED AND UNSEEN
WITHIN THIS TOPOLOGY.

If this is found to not be the case at any point, then this class must be fixed
so that the above statement is true.

=cut

sub _discover_intersite_fabrics {
    $Log->enter();
    my $self = shift;

    my $A1_net_node = $self->_dr_group()->get_A1_node()->
                                            L_get_network_node_functionality();
    my $storage_net_info_as_seen_by_A1 =
                            $A1_net_node->get_shared_storage_connection_info();


    # A1 should see 2 switches at site A- one for each fabric
    my @fabrics = ();
    # track which switches have already been used to discover a fabric
    my %site_A_switches_done = ();
    foreach my $fcvi_connection_info
                   (@{ $storage_net_info_as_seen_by_A1->{fcvi_connections} }) {
        my $site_A_switch_name = $fcvi_connection_info->{switch_name};

        # don't create a fabric twice (in case there are multiple fcvi conns)
        if (exists($site_A_switches_done{$site_A_switch_name})) {
            next;
        } else {
            $site_A_switches_done{$site_A_switch_name} = 'done';
        }

        my $site_A_switch =
            WaflMCC::MCC::Network::Switch->new(
                switch_name            => $site_A_switch_name,
                node_target_ports_spec => $storage_net_info_as_seen_by_A1
			);


        # create the fabric object
        push(@fabrics, WaflMCC::MCC::Network::Fabric->new(
                                             site_A_switch => $site_A_switch));
    }

    # standard MCC shared storage network has 2 intersite fabrics
    unless (scalar(@fabrics) == 2) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                  "Strange number of fabrics discovered in shared storage:\n" .
                  Dumper(@fabrics));
    }

    $Log->exit();
    return \@fabrics;
}

=head2 get_network_nodes()

    my $network_nodes = $topology->get_network_nodes();

Returns a hashref with this format:

    {
        A1 => $A1_net_node,
        A2 => $A2_net_node,
        B1 => $B1_net_node,
        B2 => $B2_net_node,

    }

where all $XX_net_node are WaflMCC::MCC::Network::Node objects.

The nodes are defined as follows:

    A1 is the node that was passed as the A1 node for this DR Group.
    B1 is A1's DR Partner
    A2 is A1's HA Partner
    B2 is A2's DR Partner

=cut

sub get_network_nodes {
    $Log->enter();
    my $self = shift;

    # these are MCC::Nodes
    my $nodes_by_role = $self->_dr_group()->get_nodes_by_role();

    # want the MCC::Network::Nodes
    my %net_nodes_by_role =
        map { $_ => $nodes_by_role->{$_}->L_get_network_node_functionality() }
                                                     keys(%{ $nodes_by_role });

    $Log->exit();
    return \%net_nodes_by_role;
}

=head2 get_intersite_fabrics()

    my $fabrics = $topology->get_intersite_fabrics();

Returns a hashref with the following format:

    {
      'X' => $fabric_X,
      'Y' => $fabric_Y,
      . . .
    }

where the $fabric_C are WaflMCC::MCC::Network::Fabric objects.  Bear in
mind that the order of these fabrics is arbitrary and does not indicate which
nodes access which fabrics in any primary fashion.  The fabrics are completely
interchangable when working as intended.

=cut

sub get_intersite_fabrics {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $self->_fabrics();
}

=head2 get_fiber_bridge_arrays()

TBD: THIS METHOD NOT YET IMPLEMENTED.  BURT: 592069

    my $fiber_bridges = $topology->get_fiber_bridge_arrays();

Returns a hashref with this format:

    {
      'X' => {
                 'A' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ],
                 'B' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ]
             },
      'Y' => {
                 'A' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ],
                 'B' => [ $fiber_bridge_0, $fiber_bridge_1, . . . ]
             },
    }

where X/Y and A/B are the fabric and site specification needed to define
which array of fiber switches are targetted, and all $fiber_bridge_n are
WaflMCC::MCC::Network::Fiber_Bridge objects.

See the diagrams at

http://wikid.netapp.com/w/MCC_Automation/MCC_DPG_QA_Automation/Test_Library_Dev#MCC_Network_Topology_Discussion_and_Notes

for further clarification.

=cut

sub get_fiber_bridge_arrays {
    $Log->enter();
    my $self = shift;

    $Log->exit();
    return $TO_BE_DETERMINED;
}

1;
