# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/Node.pm#5 $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Node Stub MTask Module
## @author dl-mcc-dpg-qa@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Node;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Aggregate ();
use WaflMCC::MCC::Exceptions::BaseException;

use NACL::C::SystemNode ();

use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::BaseException qw(:try);

use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR);
use Readonly;
use Data::Dumper;
use List::MoreUtils ();


=head1 NAME

WaflMCC::MCC::Node

=head1 DESCRIPTION

This task is to be used for C-Mode MetroCluster testing.  The primary idea
behind the use of MCC::Nodes is that they guarantee the the command interface
used when making any call from this object is that of the associated filer
node.


    # Start test
    use WaflMCC::MCC::Node;
    ...
TBD
    ...
    # Perform test here
    ...

    # End test

=cut

=head1 Public METHODS

There is no public constructor for this class.  A
WaflMCC::MCC::Node object is obtained via calls such as

    my $node = TBD

=cut

# automatically generate a (Non-public) constructor (includes a call to 'init')
# and (Non-public) member accessor methods
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
                { -type => 'WaflMCC::MCC::DR_Group' },
                '_dr_group',
              ],
    scalar => [
                { -type => 'WaflMCC::MCC::HA_Group' },
                '_ha_group',
              ],
    scalar => [
                { -type => 'NACL::C::SystemNode' },
                '_system_node_component',
              ],
    scalar => [
                { -type => 'WaflMCC::MCC::Network::Node' },
                '_network_node_functionality',
              ],
    scalar => [
                { -type => 'WaflMCC::MCC::Aggregate' },
                '_root_aggregate',
              ],
    scalar => [
        qw/ _uuid _systemid _nvramid _name /,
              ],
                        ];

# validate the parameters passed in and what else?
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    # TBD?  We might want to allow construction via uuid or nvramid?
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_group  => { isa   => 'WaflMCC::MCC::DR_Group' },
            node_name => { regex => qr/^[\w\-]+$/ },
        },
                            );

    # do we need to do anything else in here?
    my $dr_group  = delete($opts{dr_group});
    my $node_name = delete($opts{node_name});

    $self->_dr_group($dr_group);
    # this is temporary.  we cannot count on unchanging node names!
    $self->_name($node_name);

    my $the_Node_component =
                  NACL::C::SystemNode->new(hostrec => Tharn::host($node_name));
    $self->_system_node_component($the_Node_component);

    # this is a temporary (?) fix to BURT 601682 where calls to sysconfig are
    # timing out in the APISet object
    {
        $self->get_nodescope_apiset()->set_timeout('connectrec-timeout' => 300);
    }

    my $node_state = $the_Node_component->state(
                       requested_fields => [ 'systemid', 'nvramid', 'uuid', ]);
    $self->_systemid($node_state->systemid());
    $self->_nvramid( $node_state->nvramid());
    $self->_uuid(    $node_state->uuid());

    $self->_root_aggregate($self->_determine_root_aggregate());

    # we cannot initialize $self->_network_node_functionality until the
    # DR Group is finished constructing all MCC::Node objects.
    # $self->_network_node_functionality will be init'd in MCC::Node objects
    # at the end of the DR_Group constructor.  a call therein to
    # each $mcc_node->L_init_network_node_handle() will be used for this.

    return;
}

# determine the HA group this node belongs in.  This value is determined here
# the first time this method is called because at the time of the Node
# constructor, the DR Group's HA Groups have not yet been created.
sub _get_ha_group {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    unless ($self->_ha_group_isset()) {
        my %ha_group_map = %{ $self->_dr_group()->get_HA_groups() };
        foreach my $site (keys(%ha_group_map)) {
            my $ha_group = $ha_group_map{$site};
            my $mcc_nodes = $ha_group->get_nodes();
            if (List::MoreUtils::any { $_->get_nvramid() eq
                                      $self->get_nvramid() } @{ $mcc_nodes }) {
                $self->_ha_group($ha_group);
            }
        }

        unless ($self->_ha_group_isset()) {
            WaflMCC::MCC::Exceptions::BaseException->throw("Cannot " .
                    "determine HA Group for node '" . $self->get_name() . "'");
        }
    }

    return $self->_ha_group();
}

# find the root aggregate for this node
sub _determine_root_aggregate {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my $node_name = $self->get_name();
    my $ci = $self->get_Node_component()->command_interface();
    my @root_aggrs = NACL::C::Aggregate->find(
                             command_interface => $ci,
                             filter            => { 'owner-name' => $node_name,
                                                    'root'       => 'true',
                                                  },
                                             );
    my @aggr_names =
                  map { $_->get_one_state_attribute('aggregate') } @root_aggrs;
    # sanity check
    unless (scalar(@aggr_names) == 1) {
        confess("Node $node_name has unusual number of root aggrs:\n\t" .
                    join("\n\t", @aggr_names));
    }

    my $root_aggr = WaflMCC::MCC::Aggregate->new(owner_mcc_node => $self,
                                             aggregate_name => $aggr_names[0]);

    return $root_aggr;
}

=head2 L_is_vsim()

Returns true (1) if this node is a vsim. False (0) otherwise.

=cut

sub L_is_vsim {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    return $self->_network_node_functionality()->L_is_vsim();
}

# this class also contains an object which is solely associated with this
# node's networking functionality in the storage backplane network
sub L_init_network_node_handle {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my $network_node = WaflMCC::MCC::Network::Node->new(
                            dr_group => $self->_dr_group(), mcc_node => $self);

    $self->_network_node_functionality($network_node);

    return $network_node;
}

=head2 L_get_root_aggregate()

Returns the NACL::MTask:MCC::Aggregate associated with the root aggregate of
this node.

=cut

sub L_get_root_aggregate {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_root_aggregate();
}

=head2 get_Node_component()

    my $node_comp = $mcc_node->get_Node_component();

Returns the component object (NACL::C::SystemNode) of the node.

=cut

sub get_Node_component {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_system_node_component();
}

=head2 get_nodescope_apiset()

    my $nodescope_apiset = $mcc_node->get_nodescope_apiset();

Returns the nodescope APISet object (NACL::APISet) associated with this node.

=cut

sub get_nodescope_apiset {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->get_Node_component()->get_7m_or_nodescope_apiset();
}

=head2 L_get_network_node_functionality()

    my $network_node = $mcc_node->L_get_network_node_functionality();

Returns the WaflMCC::MCC::Network::Node object associated with the
"network functionality" of this WaflMCC::MCC::Node.

=cut

sub L_get_network_node_functionality {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_network_node_functionality();
}

=head2 get_name()

    my $node_name = $mcc_node->get_name();

Returns the node's name (string).

=cut

sub get_name {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    # IMPORTANT TBD:  think about this.  problem is that the node cannot be
    # queried for the name if it is down due to some disaster cycle.  the query
    # will hang/timeout if this happens, as it often will in a DR lc.
    #
    # nodes can be renamed.  we cannot cache the name of a node, but need to
    # query for it every time we need it.
    #my $name = $self->get_Node_component()->get_one_state_attribute('node');
    my $name = $self->_name();

    return $name;
}

=head2 get_role()

    my $node_role = $mcc_node->get_role();

Returns the node's role in its DR group ('A1'|'A2'|'B1'|'B2').  Here A1 and
B1 are DR partners, and A2 and B2 are DR partners.

=cut

sub get_role {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_network_node_functionality()->get_role();
}

=head2 get_nvramid()

    my $node_nvramid = $mcc_node->get_nvramid();

Returns the node's nvramid (string).

=cut

sub get_nvramid {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_nvramid();
}

=head2 get_uuid()

    my $node_uuid = $mcc_node->get_uuid();

Returns the node's uuid (string).

=cut

sub get_uuid {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    return $self->_uuid();
}

=head2 metrocluster_enable()

This is a STATIC method because no MCC::Metrocluster object can be instantiated
before 'metrocluster enable' operations have been completed successfully.

    WaflMCC::MCC::Node->metrocluster_enable(
         command_interface_node_name => 'spectre-l11',
         node_name_to_enable         => 'spectre-l12',
         refresh                     => 0);

    WaflMCC::MCC::Node->metrocluster_enable(
         command_interface_node_name => 'spectre-l11',
         node_uuid_to_enable         => 'd437c8cd-e931-11e0-aa13-c948a5272a17');

When this method is called on a WaflMCC::MCC::Node, then
'metrocluster enable' is called directly via command_interface_node_name's node
command interface.  This method need only be called one one node in each of the
Metrocluster-to-be's DR partner nodes.

=over

=item Options

=over

=item C< command_interface_node_name >

(Required) Name of the node to use as the command interface for the call to
'metrocluster enable'.

=item C< node_name_to_enable >

(Either node_name_to_enable XOR node_uuid_to_enable are required) Name of the
node to be enabled.

=item C< node_uuid_to_enable >

(Either node_name_to_enable XOR node_uuid_to_enable are required) Node UUID of
the node to be enabled.

=item C< refresh => 0|1 (default = 1) >

(Optional) If true, refreshes the internal metrocluster state data.

=back

=back

=cut

sub metrocluster_enable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $pkg = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface_node_name => { regex    => qr/^[\w\-]+$/ },
            node_name_to_enable         => { regex    => qr/^[\w\-]+$/,
                                             optional => 1 },
            node_uuid_to_enable         => { regex    => qr/^[\w\-]+$/,
                                             optional => 1 },
            refresh                     => { default  => 1 },
        },
                            );

    # one and only one of the parameters node_name_to_enable and
    # node_uuid_to_enable is required.  Check this requirement.
    {
        my $parameter_error_message =
            "WaflMCC::MCC::Node->metrocluster_enable() requires either\n" .
            "    the node_name_to_enable parameter exclusive-or the\n" .
            "    node_uuid_to_enable parameter to be passed.";

        if (defined($opts{node_name_to_enable})) {
            if (defined($opts{node_uuid_to_enable})) {
                confess($parameter_error_message);
            }
        } elsif (! defined($opts{node_uuid_to_enable})) {
            confess($parameter_error_message);
        }
    }

    # get Metrocluster component based on command_interface_node_name
    my $node_ci = NACL::C::Node->new(hostrec =>
         Tharn::host($opts{command_interface_node_name}))->command_interface();
    my $metrocluster_component =
                     NACL::C::Metrocluster->new(command_interface => $node_ci);

    if (defined($opts{node_name_to_enable})) {
        $metrocluster_component->enable(node    => $opts{node_name_to_enable},
                                        refresh => $opts{refresh});
    } else {
        $metrocluster_component->enable(
                                    'node-uuid' => $opts{node_uuid_to_enable},
                                    refresh     => $opts{refresh});
    }

    return;
}

=head2 metrocluster_disable()

Unlike metrocluster_enable, metrocluster_disable is an instance method as it is
always called upon nodes in an active metrocluster system.

    # run 'metrocluster disable' on all site A nodes
    my $dr_groups = $self->get_DR_groups();
    foreach my $dr_group (values(%{ $dr_groups } )) {
        my $dr_partnerships = $dr_group->get_DR_partnerships();
        foreach my $partnership (@{$dr_partnerships}) {
            my $site_A_partner = $partnership->get_partner_node(site => 'A');
            $site_A_partner->metrocluster_disable();
        }
    }

When this method is called on a WaflMCC::MCC::Node, then
'metrocluster disable' is called directly via this node's command interface.

=cut

sub metrocluster_disable {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my $node_ci = $self->get_Node_component()->command_interface();
    my $metrocluster_component_using_this_nodes_ci =
                     NACL::C::Metrocluster->new(command_interface => $node_ci);

    $metrocluster_component_using_this_nodes_ci->disable(
        node => $self->get_name()
    );

    return;
}

=head2 metrocluster_show()

This method runs metrocluster show commmand - It prints the dump on stdin

=over

=item Arguments

 None

=item Return Value

Metrocluster show output

=item Example

=back

     my $mcc_node = WaflMCC::MCC::Node->new(node_name =>$NODEA);
     $mcc_show = $cluster_A->metrocluster_show();

=cut

sub metrocluster_show {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my ($self,%opts) = @_;
    my $node_comp = $self->get_Node_component();
    my $ci = $node_comp->command_interface();
    my $mcc_show = NACL::CS::Metrocluster->fetch(command_interface => $ci);
    return $mcc_show;
}

=head2 compare_metrocluster_show_fields()

This method compares the metrocluster show output attributes
with the attributes provided by the user in the form of hashref.
Mismatches upon finding are stored into a hashref. Mismatches
are returned to the user if throw_exception_on_mismatch is not set

=over

=item Arguments

  Required: fields_to_test => $fields  [HASHREF]
  Optional: throw_exception_on_mismatch [BOOLEAN], default: 1

=item Return Value

Mismatches are returned if throw_exception_on_mismatch is set to 0
Exception is thrown if exception flag is set

=item Example

=back

     my $fields = {
    'dr_partner' => $partner_name,
    'dr_cluster' => $site_name,
    'dr_operation_state' => 'switchover',
        'ha_partner' => 'fd8a2896-003d-4b07-ae82-ab0f7fc7a588',
         };
     my $mcc_node = WaflMCC::MCC::Node->new(node_name =>$NODEA);
     $mcc_node->compare_metrocluster_show_fields(fields_to_test=>$fields);

=cut

sub compare_metrocluster_show_fields {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my ($self,%opts) = @_;
    #validate opts
    #setting default to 1
    $opts{throw_exception_on_mismatch} = (exists $opts{throw_exception_on_mismatch} ?
                      $opts{throw_exception_on_mismatch} :
                      1);
    my %vopts = validate_with(
    params => \%opts,
    spec   => {
        fields_to_test => { type => HASHREF },
        throw_exception_on_mismatch => {regex => qr/^(0|1)$/ }
    }
                  );
    #getting the user provided fields out
    my $fields_to_test_hash_ref = $vopts{fields_to_test};
    #getting the node component
    my $mcc_show = $self->metrocluster_show();
    #removing uneccessary info from the hash
    delete $mcc_show->{'command_interface'};
    #this hash will have mismatches if found, empty initially
    my $mismatches = {};
    #iterate over the hash which is provided by the user
    #and compares those fields with metro show fields
    #we populate the mismatches into the returning hash - undef otherwise
    while (my ($key, $value) = each (%$fields_to_test_hash_ref)) {
    if(!exists ($mcc_show->{$key})){
        confess "Invalid metrocluster show attribute found: \'$key\' ".
        "Please pass in the correct attribute to compare. \n";
    }
    if($mcc_show->{$key} ne $fields_to_test_hash_ref->{$key}){
        $mismatches->{$key} = $fields_to_test_hash_ref->{$key};
    }
    }
    my $throw_exception_flag = $opts{throw_exception_on_mismatch};
    # if throw_exception_on_mismatch is set
    if($throw_exception_flag) {
    if(keys (%$mismatches)){
        #TBD confess to be replace by a proper exception
    confess 'mismatches found:'. Dumper($mismatches);
    }
    else {
        $Log->comment("No Mismatches found\n");
        return ;
    }
    } #throw_exception_on_mismatch block ends
    return $mismatches;
}

=head2 get_dr_fsm_state()

This method runs mcc status commmand over dblade to fetch FSM state

=over

=item Arguments

 None

=item Return Value

FSM state value obtained from mcc status command

=item Example

=back

     my $mcc_node = WaflMCC::MCC::Node->new(node_name =>$NODEA);
     $mcc_status = $mcc_node->get_dr_fsm_state();

=cut

sub get_dr_fsm_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my $node_component = $self->get_Node_component();
    my $cmd = '"priv set test;mcc status"';
    my $response = $node_component->run('command' => $cmd);

    my $drsom_state = undef;
    if ($response =~ /FSM state:([\s\w]+)/i) {
        $drsom_state = $1;
        $drsom_state =~ s/^\s+//;  $drsom_state =~ s/\s+$//;
    }

    unless (defined($drsom_state)) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                                           'FSM state not found:' . $response);
    }

    return $drsom_state;
}

=head2 L_validate_DRSOM_FSM_state

This method throws an exception if this node's DRSOM_FSM state is not consistent
with the mhost's DR state.

IMPORTANT: This method assumes that this node is in a cluster with a live
cluster management CI.  It uses this CI to determine if this node is queryable
as it may be down after an HA Takeover event.

=over

=item Arguments

=over

=item C< disaster_status >

(Required) DR state as defined by DR_Coordinator::get_DR_state().

=back

=back

=cut

#mica - need to handle here 
sub L_validate_DRSOM_FSM_state {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $self = shift;

    my %opts = validate_with(
        params => \@_,
        spec    => {
            dr_state => { type => SCALAR },
                   },
                            );

    my $dr_state = delete($opts{dr_state});

    my %DRSOM_STATE_EXPECTED = (
               switchover_completed => 'Switchover completed',
               heal_aggrs_completed => 'Aggr healing completed',
               heal_roots_completed => 'Root aggr switchback completed',
               normal               => 'Normal', );

    my %DRSOM_STATE_EXPECTED__deprecated = (
               switchover_completed => 'DRSOM_FSM_ST_SO_COMPLETE',
               heal_aggrs_completed => 'DRSOM_FSM_ST_SB_AGGR_HEALED',
               heal_roots_completed => 'DRSOM_FSM_ST_SB_ROOT_HEALED',
               normal               => 'DRSOM_FSM_ST_NORMAL', );

    # note that 2-node systems are always "HA_state == 'normal'"
    my $ha_group = $self->_get_ha_group();

    # mica - wafl changes 
    ################################################################################
    #my $ha_state = $self->_get_ha_group()->L_get_HA_state();
    $Log->comment("mica node.pm : wafl override...");
    my $ha_state = {     HA_state      => 'normal',
                         survivor_node => undef,
                         disaster_node => undef,
                    };
    return; 

    ################################################################################

    if (($ha_state->{HA_state} eq 'normal') ||
                  (($ha_state->{HA_state} eq 'takeover') &&
                   ($ha_state->{survivor_node}->get_nvramid() eq
                                                      $self->get_nvramid()))) {

        my $drsom_state = $self->get_dr_fsm_state();

        #unless (WaflMCC::MCC::Utils->workaround_expired(
        #expiration_date => '2012/11/30', burt => '000000')) {
        #if ($drsom_state =~ /^DRSOM_FSM_ST/) {
        #$Log->comment("Accepting old 'FSM state' string values...");
        #%DRSOM_STATE_EXPECTED = %DRSOM_STATE_EXPECTED__deprecated;
        #}
        #}

        unless ($drsom_state eq $DRSOM_STATE_EXPECTED{$dr_state}) {
            my $node_name = $self->get_name();
            WaflMCC::MCC::Exceptions::BaseException->throw(
                "Inconsistent DRSOM_FSM_state: Node [$node_name] has\n" .
                "state [$drsom_state] while DR State is reported to\n" .
                "be [$dr_state].");
        }
    }

    return;
}

1;
