# $Id: //depot/prod/test/main/lib/NACL/MTask/MCC/Shared_Storage.pm#9 $

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking ProhibitFixedStringMatches)
##no critic (ControlStructures::ProhibitUnlessBlocks)
##no critic (ValuesAndExpressions::ProhibitInterpolationOfLiterals)

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary MCC Shared_Storage Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Shared_Storage;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Exceptions::InconsistentDiskState (); #TODO - update??
use WaflMCC::MCC::Exceptions::DiskState ();
use WaflMCC::MCC::Exceptions::DiskState::FailedDisk ();
use WaflMCC::MCC::Exceptions::DiskState::NewDisk ();
use WaflMCC::MCC::Exceptions::DiskState::RaidState ();
use WaflMCC::MCC::Exceptions::DiskState::SanownState ();
use WaflMCC::MCC::Exceptions::DiskState::StorageLayerState ();
use WaflMCC::MCC::Log ();

use WaflMCC::MCC::Aggregate ();
use WaflMCC::MCC::Utils ();

use NACL::CS::StorageDisk ();

use NATE::BaseException qw(:try);
use Tharn qw($Log);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use Data::Dumper;
use Scalar::Util qw(weaken isweak);


=head1 NAME

WaflMCC::MCC::Shared_Storage

=head1 DESCRIPTION

This class is to be used for C-Mode Metrocluster testing.

    # Start test
    use WaflMCC::MCC::Metrocluster;
    ...

    my $dr_group = $mcc->get_DR_groups()->{$id};
    my $shared_storage = $dr_group->get_shared_storage();
    $shared_storage->validate_disk_states();
    ...
    # Perform test here
    ...
    # End test


This is a reasonably complex class for several reasons:

1)  There are many disk states to check, most of which have a different logic
    to follow for the current DR state.  Add HA states in the mix, and it gets
    another (binary) order of magnitude more complex.  I chose to break the
    checks for these disk states into many seperate methods, each of which
    only need to present the logic regarding a particular disk state(s):

        _validate_disk_states__owner_and_id()
        _validate_disk_states__home_and_id()
        _validate_disk_states__dr_home_and_id()
        _validate_disk_states__reserver_id()
        _validate_disk_states__pool()
        _validate_disk_states__plex()

    Each of these is passed the same set of disk data and return issues
    encontered in the same manner.  After all are called, problems are reported
    and exceptions are thrown if needed.

2)  This class encompasses all disks in the system instead of just 1, so the
    data structures involved have at least another layer of depth compared to
    treating individual disks as the base object.  This was a performance issue
    decision.  Querying ONTAP for each particular disk would have been
    prohibitive.  In retrospect though, now that I've seen it in action enough,
    I believe that it would be possible to still make the single query for all
    the disks, but then use that information to generate all the needed disk
    objects, and then validate individual disk objects, but that is pie in the
    sky for now...

    See pod on validate_disk_states() for the format of the primary disk data
    structure used.  In addition, data associated with each disk which requires
    an indirect methodology to determine is added into each disk's part of the
    data structure by the method _add_disk_meta_data().  Examples of such
    meta-data are

        meta__disk_locale
        meta__is_in_root_aggr
        meta__ha_takeover_node
        meta__ha_TO_node_name
        meta__dr_partner_if_orig_owner_at_disaster_site

and the infamous (see point 5 below)

        meta__TO_node_of_dr_partner_if_OOADS

3)  The fact that in many DR states, only one site can be queried re the
    storage, and a non-owning site does not (yet?) get all the information
    out of a disk query that an owner would see.

4)  When 2 sites can see all the storage, a "merge" of the disk data muust be
    made (while validating when different sites see different/incomplete disk
    state.)  Validating these 2 views is different than validating particular
    disk states in terms of what is expected at a particulat DR (+HA?) state.

5)  Since we have no Disk objects to hide the garbage in, we end up using
    egregious variables such as "$TO_node_of_dr_partner_if_OOADS", which
    means, _in the context of talking about a particular disk_, "takeover node
    of dr partner if original owner (of the disk) [is] at [the] disaster site".
    To wit, the original owner node of a disk may be at the disaster site, and
    would have a DR partner node which may have been in an HA takeover.  The
    node that did the takeover in this case is the
    $TO_node_of_dr_partner_if_OOADS.  It's almost TOO easy...  [sigh]

    Also, often times, the fact that one of these such variables is NOT
    UNdefined is used as a marker of a particular set of disks of interest.
    Be sure to understand the aspects of a disk in which one or the other of
    these variables is defined.  If such a value is not defined for a particular
    disk, then that disk is not in a particular category or environ in the
    shared storage.  For example, no disk which is originally owned by a
    survivor site node will be associated with a defined value of the
    meta__dr_partner_if_orig_owner_at_disaster_site hash key (nor the
    usual variable such a value is placed in: $dr_partner_if_OOADS.

=cut

=head1 Public Methods

There is no public constructor for this class.  WaflMCC::MCC::Shared_Storage
objects are obtained via calls such as

    my $dr_group = $mcc->get_DR_groups()->{$id};
    my $shared_storage = $dr_group->get_shared_storage();

=cut

# Automatically generate the (non-public) constructor (includes a call to
# 'init') and also creates non-public members and non-public accesors for
# them.
use Class::MethodMaker [
    new    => [ '-init', 'new' ],
    scalar => [
        { -type => 'WaflMCC::MCC::Metrocluster' },
        '_metrocluster',
              ],
    scalar => [
        { -type => 'WaflMCC::MCC::DR_Group' },
        '_dr_group',
              ],
    scalar => [
        qw/ _initial_disk_state_info
            _disk_serials_for_sampled_report
            _default_log_level /
              ],
                       ];

# validate the parameters passed in and what else?
sub init {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            metrocluster => { isa => 'WaflMCC::MCC::Metrocluster' },
            dr_group     => { isa => 'WaflMCC::MCC::DR_Group' },
        },
                            );

    # if this method is changed, beware of effects of setting for
    # $ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}
    if ($ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}) {
        $Log->warn(
            "User has set DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION\n" .
            "    Disk state validation has been disabled!\n" .
            "    No disk validation will occur in this test!\n" .
            "    This is not recommended unless necessary to\n" .
            "    proceed with tests which are otherwise impossible.\n");
    }

    # set members per the passed constructor arguments.
    $self->_metrocluster($opts{'metrocluster'});
    $self->_dr_group(    $opts{'dr_group'});
    weaken ($self->{_metrocluster});
    weaken ($self->{_dr_group});

    $self->_default_log_level('sample');

    # _initial_disk_state_info will be initialized in the first call to
    # L_validate_disk_states()

    return;
}

=head2 validate_disk_states()

    $shared_storage->validate_disk_states(
                # default values shown here
                failed_disks_ok => 0,
                new_disks_ok    => 0,
                                         );

This method validates that important storage layer, SANOWN layer, and RAID layer
values for all disks in the shared storage are valid with respect to the DR
state of the system.  If any disk state is found to be out of compliance with
the expected ownership state, then a InvalidDiskOwnershipState exception is
thrown.

This method also validates that the views of the shared storage from both sites
are consistent.

TBD: Throw InvalidDiskOwnershipState exception instead of exiting.

Disk metadata fields validated:

    Storage Layer:
        serial-number
        pool

    SANOWN Layer:
        owner
        owner-id
        home
        home-id
        dr-home
        dr-home-id
        reserver-id

    RAID Layer:
        plex

=over

=item Options

The following options have not been exercised much.  They may need to be
re-evaluated as to the usage scenarios and possibly removed or hardened
(bruce oct 2012)

=item C<< failed_disks_ok => 0|1 (default = 0) >>

(Optional) This option will write a WARNing report about disks it no longer sees
in the shared storage, but will not fail the test.

=item C<< new_disks_ok => 0|1 (default = 0) >>

(Optional) This option will write a WARNing report about new disks it sees in
the shared storage, but will not fail the test.

[NB: This method does nothing if
$ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}=1.]

=back

=over

=item Associated ENVIRONMENT Variables

=item C<< DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION = 0|1 >>

=item C<< WARN_ONLY_ON_INVALID_PLEX_ASSIGNMENT = 0|1 >>

[This option is likely obsolete.  It was introduced as a workaround when
 testbeds were being built with less control about which disk pools the plexes
 would draw from when being created. (bruce oct 2012)]

See Metrocluster constructor notes.

=back

Returns a hashref of information about all disks' normal and current states in
the format:

    {
        initial_states => {
                              JLVS2HAC => {
                                              'owner' => 'spectre-l12',
                                              'home-id' => '151708069',
                                              'dr-home' => '-',
                                              'serial' => 'JLVSY6VC',
                                              'plex' => '-',
                                              'pool' => 'Pool0',
                                              'owner-id' => '151708069',
                                              'dr-home-id' => '-',
                                              'home' => 'spectre-l12',
                                              'reserver-id' => '1574774938'
                                          },
                              . . .
                          },
        current_states => {
                              JLVS2HAC => {
                                              'owner' => 'spectre-l12',
                                              'home-id' => '151708069',
                                              'dr-home' => '-',
                                              'serial' => 'JLVSY6VC',
                                              'plex' => '-',
                                              'pool' => 'Pool0',
                                              'owner-id' => '151708069',
                                              'dr-home-id' => '-',
                                              'home' => 'spectre-l12',
                                              'reserver-id' => '1574774938'
                                          },
                              . . .
                          },
    }

=cut

sub validate_disk_states {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            failed_disks_ok     => { regex    => qr/^[01]$/, default  => 0, },
            new_disks_ok        => { regex    => qr/^[01]$/, default  => 0, },
            log_level           => { regex    => qr/^(errors|sample|full)$/,
                                     optional => 1, },
                   },
                            );

    # get the disaster status and add to arguments for L_validate_disk_states
    my $mcc      = $self->_metrocluster();
    my $dr_group = $self->_dr_group();
    my $disaster_status = $mcc->get_disaster_status(include_HA => 1,
                                                    dr_group   => $dr_group);
    $opts{disaster_status} = $disaster_status;

    return $self->L_validate_disk_states(\%opts);
}

=head2 set_default_log_level()

    my $old_log_level =
                   $shared_storage->set_default_log_level(log_level => 'full');

Logging of validation reports may be user set to 'full', 'sample', or 'errors'.

Returns the previous value of the log_level.

=over

=item Options

=over

=item C<< log_level => 'full'|'sample'|'errors' >>

(Required) Specifies the default disk validation log level a user wishes to
get for this shared storage array.

All log levels always report any and all disk inconsistencies and warnings.

'full' logs all validated fields of every disk in this shared storage array.

'sample' makes an attempt to log information about a smaller than full set
of disks that are samples of disks that should have different types of state
changes during a DR lifecycle.

=back

=back

=cut

sub set_default_log_level {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            log_level => { regex => qr/^(errors|sample|full)$/, },
                  },
                            );

    my $old_log_level = $self->_default_log_level();
    $self->_default_log_level(delete($opts{log_level}));

    return $old_log_level;
}

# L_validate_disk_states differs from validate_disk_states in that it requires
# the parameter disaster_status to be passed in, and a user should never be
# allowed to explicitly pass this in (I think.)
#
# disaster_status is expected to have the form as gotten from this call:
#
#       $disaster_status = $mcc->get_disaster_status(
#                  include_HA => 1,
#                  dr_group   => $the_dr_group_containing_this_shared_storage);
# or
#       $disaster_status = $mcc->get_disaster_status(include_HA => 1);
#
# Method does nothing if $ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}=1.
sub L_validate_disk_states {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            failed_disks_ok => { regex    => qr/^[01]$/, default  => 0, },
            new_disks_ok    => { regex    => qr/^[01]$/, default  => 0, },
            disaster_status => { type     => HASHREF },
            log_level       => { regex    => qr/^(errors|sample|full)$/,
                                 optional => 1, },
                   },
                            );

    # do nothing if disk validation is disabled.
    if ($ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}) {
        my $msg = "DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION set!\n" .
                  "    Disks NOT validated!";
        $Log->warn($msg);
        return $msg;
    }

    my $failed_disks_ok = delete($opts{failed_disks_ok});
    my $new_disks_ok    = delete($opts{new_disks_ok});

    # log_level has a user settable default.  get that here if log_level has
    # not been explicitly passed.
    my $log_level = defined($opts{log_level}) ? delete($opts{log_level}) :
                                                $self->_default_log_level();

    my $mcc      = $self->_metrocluster();
    my $dr_group = $self->_dr_group();

    my $disaster_status = defined($opts{disaster_status}) ?
                                  delete($opts{disaster_status}) :
                                  $mcc->get_disaster_status(
                                             include_HA => 1,
                                             dr_group   => $self->_dr_group());

    # this is a workaround for systems that have disks with incorrect plex
    # values.  eventually, these setting will invalidate the MCC setup, but for
    # now, they can still be used in mosts tests.
    my $warn_only_on_invalid_plex =
            defined($ENV{WARN_ONLY_ON_INVALID_PLEX_ASSIGNMENT}) ?
                                $ENV{WARN_ONLY_ON_INVALID_PLEX_ASSIGNMENT} : 0;

    # these two will be undef outside of DR
    my $survivor_site = $disaster_status->{survivor_site};
    my $disaster_site = $disaster_status->{disaster_site};

    my $dr_state   = $disaster_status->{DR_state};
    my $live_sites = $disaster_status->{live_sites};

    # execute reserver_id workaround (if needed)

    $mcc->get_cluster(site => @{$live_sites}[0])->
        L_workaround_force_reserver_id_update(
            known_live_sites => $live_sites);

    # first time through?
    my $is_initial_validation = 0;
    unless ($self->_initial_disk_state_info_isset()) {
        $is_initial_validation = 1;
        # throws an exception if DR/HA states are not normal
        my $initial_disk_state_info = $self->_get_initial_disk_state_info(
                                          disaster_status => $disaster_status);
        $self->_initial_disk_state_info($initial_disk_state_info);
    }

    # disks' states
    my $init_states = $self->_initial_disk_state_info();
    my $current_states = $is_initial_validation ? $init_states :
              $self->_get_disk_array_data(disaster_status => $disaster_status);

    # these same opts will be passed to each validation method
    my %common_opts = ( dr_state                  => $dr_state,
                        init_states               => $init_states,
                        current_states            => $current_states,
                      );

    # don't bother parsing this unless you have a masochistic streak.  it is
    # just something to print some condensed HA state info...
    {
        my %ha_states = %{$disaster_status->{HA_states}->{$dr_group->get_id()}};
        my @ha_state_items = map { $_ . ': ' . $ha_states{$_}->{HA_state} .
                     (($ha_states{$_}->{HA_state} ne 'normal') ? q{:} .
                     $ha_states{$_}->{survivor_node} : q{}) } keys(%ha_states);
        my $ha_state_text = join(q{, }, @ha_state_items);
        $Log->comment("\nValidating " . scalar(keys(%{ $current_states })) .
                      " disks' selected Storage/SANOWN/RAID layer states" .
                      " for DR = $dr_state / HA = $ha_state_text . . .");
    }

    # this report value will be returned by this method in case user wants to
    # parse it for some reason.
    my $report = { initial_states => $init_states,
                   current_states => $current_states };
    # log this report if log_level is 'full', or generate sampled report perhaps
    if ($log_level eq 'full') {
        $Log->comment("\n");
        $Log->comment($self->_dump_some_disk_data(disk_data => $report));
    } elsif ($log_level eq 'sample') {
        my $sampled_disk_report = $self->_generate_sampled_disk_report(
                                            current_states => $current_states);
        $Log->comment("\n");
        $Log->comment($sampled_disk_report);
    }

    # check that all disks are present that should be
    my $error_msg = q{};
    my %error;
    my $warning_msg = q{};
    my $disk_presence_check_messages = $self->_check_for_disk_presence(
                                            init_states    => $init_states,
                                            current_states => $current_states);
    if ($disk_presence_check_messages->{failed_disks}) {
        if ($failed_disks_ok) {
            $Log->comment($disk_presence_check_messages->{failed_disks});
        } else {
            $error{FailedDisk} = $disk_presence_check_messages->{failed_disks};
        }
    }
    if ($disk_presence_check_messages->{new_disks}) {
        if ($new_disks_ok) {
            $Log->comment($disk_presence_check_messages->{new_disks});
        } else {
            $error{NewDisk} = $disk_presence_check_messages->{new_disks};
        }
    }

    # here's the main part of the validation
    $error{StorageLayerState} .=
        $self->_validate_disk_states__pool(          %common_opts);

    $error{SanownState} .=
        $self->_validate_disk_states__owner_and_id(  %common_opts);
    $error{SanownState} .=
        $self->_validate_disk_states__home_and_id(   %common_opts);
    $error{SanownState} .=
        $self->_validate_disk_states__dr_home_and_id(%common_opts);
    $error{SanownState} .=
        $self->_validate_disk_states__reserver_id(   %common_opts);

    # plex validation logic needs to updated.  this workaround call is basically
    # just a placeholder to warn that work needs to be done to get this
    # validation back in place.
    $Log->comment("\n");
    unless (WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/03/12', burt => '648016')) {
        $Log->comment('Disk validation of plex value currently not being ' .
                      'validated!');
        #my $plex_msg .= $self->_validate_disk_states__plex(%common_opts);

        # likely the 'warn_only_on_invalid_plex' flag is obsolete now and should
        # be gotten rid of when this logic is re-addressed.

        #if ($warn_only_on_invalid_plex) {
        #    $warning_msg .= $plex_msg;
        #} else {
        #    $error{RaidState} .= $plex_msg;
        #}
    }

    my $exception_name;
    foreach my $key (sort keys %error) { #or do we care about order??
        if ($error{$key}) {
            $error_msg .= $error{$key};
            $exception_name = $exception_name || $key; #future:sort worst 1st?
        } else { # no actual error?
            delete $error{$key};
        }
    }

    # test fail if error message is non-empty
    if ($error_msg) {
        $Log->error($error_msg);
        if (scalar keys %error > 1) {
            #throw a general error if > 1 err type
            WaflMCC::MCC::Exceptions::DiskState->throw($error_msg);
        } else {
            my $exception =
                "WaflMCC::MCC::Exceptions::DiskState::$exception_name";
            ## no critic (TestingAndDebugging::ProhibitNoStrict)
            no strict "refs";
            $exception->throw($error{$exception_name});
        }

    } elsif ($warning_msg) {
        $Log->warn($warning_msg);
    } else {
        $Log->comment("\nDisk validation complete.");
    }

    return $report;
}

# gets initial disk state info if not in DR or HA event.
# otherwise throws exception.
sub _get_initial_disk_state_info {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status => { type => HASHREF, },
                   },
                            );

    my $disaster_status = delete($opts{disaster_status});

    my $initial_disk_state_info = undef;
    if ($disaster_status->{DR_state} eq 'normal') {
        my $all_HA_states_are_normal = 1;
        my %HA_states = %{ $disaster_status->{HA_states} };
      OUTER:
        foreach my $dr_group_name (keys(%HA_states)) {
            foreach my $site (qw/ A B /) {
                if ($HA_states{$dr_group_name}->{$site}->{HA_state}
                                                                 ne 'normal') {
                    $all_HA_states_are_normal = 0;
                    last OUTER;
                }
            }
        }

        # okay to init disk state info?
        if ($all_HA_states_are_normal) {
            $initial_disk_state_info = $self->_get_disk_array_data(
                                          disaster_status => $disaster_status);
        }
    }

    # check again now
    unless (defined($initial_disk_state_info)) {
        WaflMCC::MCC::Exceptions::BaseException->throw(
                "ERROR: Trying to initialize disk state information\n" .
                "       while in DR and/or HA disaster state:\n" .
                WaflMCC::MCC::Utils->dump($disaster_status) . "\n\n");
    }

    return $initial_disk_state_info;
}

=head2 _get_disk_array_data()

This method gets disk data for every disk in the shared storage of the dr_group,
as viewed by either the survivor site inside DR or by both sites when not in
DR.  In the case where both sites can view the storage, the data as seen by
both sites is validated against each other, and then it is consolidated and
returned as a single data structure.  The reason such consolidation is required
is because while storage and SANOWN layer data can be seen (and should be
identical) by each site, RAID and higher level data can only (currently) be
seen by the site which owns the disk.

TBD: Throw InvalidDiskOwnershipState exception instead of exiting.

=cut

sub _get_disk_array_data {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            disaster_status => { type => HASHREF },
                  },
                            );

    my $disaster_status = delete($opts{disaster_status});

    my @live_sites = @{ $disaster_status->{live_sites} };
    my %disk_data_by_site = ();
    foreach my $live_site (@live_sites) {
        my $site_ci = $self->_metrocluster()->get_Metrocluster_component(
                                      site => $live_site)->command_interface();

        my @disk_states = NACL::CS::StorageDisk->fetch(
                   command_interface => $site_ci,
                   requested_fields  => [ 'serial-number', 'uid',
                                          'disk',          'aggregate',
                                          'pool',          'plex',
                                          'orig-owner',    'reserver-id',
                                          'owner',         'owner-id',
                                          'home',          'home-id',
                                          'dr-home',       'dr-home-id',
                                        ],
                                               );


        $disk_data_by_site{$live_site} = { };
        for my $disk_state (@disk_states) {
            my $serial     = $disk_state->serial_number();

            my $home        = $disk_state->home();
            my $pool        = $disk_state->pool();
            my $orig_owner  = $disk_state->orig_owner();
            my $owner       = $disk_state->owner();
            my $owner_id    = $disk_state->owner_id();
            my $home_id     = $disk_state->home_id();
            my $reserver_id = $disk_state->reserver_id();
            my $plex        = $disk_state->plex();
            my $aggregate   = $disk_state->aggregate();

            ## disk and uid can be fetched if required in future
            ## my $disk = $disk_state->disk(); and
            ## my $uid  = $disk_state->uid();

            my $dr_home    =
                defined($disk_state->dr_home()) ? $disk_state->dr_home() : '-';

            my $dr_home_id =
                defined($disk_state->dr_home_id()) ?
                $disk_state->dr_home_id() : '-';

            # home and pool must have operational values, otherwise ignore
            if (      defined($home) && ($home ne '-') &&
                      defined($pool) && ($pool ne '-') && ($pool !~ /NONE/i)) {

                $disk_data_by_site{$live_site}->{$serial} = {
                                               'pool'        => $pool,
                                               'orig-owner'  => $orig_owner,
                                               'owner'       => $owner,
                                               'owner-id'    => $owner_id,
                                               'home'        => $home,
                                               'home-id'     => $home_id,
                                               'dr-home'     => $dr_home,
                                               'dr-home-id'  => $dr_home_id,
                                               'reserver-id' => $reserver_id,
                                               'plex'        => $plex,
                                               'aggregate'   => $aggregate,
                                                            };
            }
        }
    }

    # if views from both sites are available, validate them against each-other
    # and merge the valid information into 1 structure
    my $disk_array_data = (@live_sites == 1) ?
             $disk_data_by_site{$live_sites[0]} :
                 $self->_validate_and_consolidate_multi_site_disk_state_views(
                                     disk_data_by_site => \%disk_data_by_site);

    # add in metadata for disk states, like is_in_root_aggr, etc.
    $disk_array_data = $self->_add_disk_meta_data(
                                          disk_array_data => $disk_array_data,
                                          disaster_status => $disaster_status);

    return $disk_array_data;
}

sub _add_disk_meta_data {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            disk_array_data => { type => HASHREF },
            disaster_status => { type => HASHREF },
                  },
                            );

    my $disk_array_data = delete($opts{disk_array_data});
    my $disaster_status = delete($opts{disaster_status});

    # general data needed for every disks' metadata follows...

    # ...
    my $root_aggregates = $self->_determine_root_aggregates();

    # ...
    # each member of this array will be a hash with keys disaster_node and
    # survivor_node and MCC::Node values.
    my @HA_disaster_pairs = ();
    {
        my $ha_states =
                 $disaster_status->{HA_states}->{$self->_dr_group()->get_id()};
        foreach my $site (keys(%{ $ha_states })) {
            my $disaster_node = $ha_states->{$site}->{disaster_node};
            if (defined($disaster_node)) {
                my $survivor_node = $ha_states->{$site}->{survivor_node};
                push(@HA_disaster_pairs, { disaster_node => $disaster_node,
                                           survivor_node => $survivor_node });
            }
        }
    }

    # need to know which site each disk is at (and not just where its owner is)
    # (only placing this info in the init disk state struct)
    my %site_vs_node_names = ();
    if (! $self->_initial_disk_state_info_isset()) {
        my %nodes_by_role = %{ $self->_dr_group()->get_nodes_by_role() };
        foreach my $role (keys(%nodes_by_role)) {
            my $mcc_node = $nodes_by_role{$role};
            my ($site) = ( $role =~ /^([AB])/ );
            $site_vs_node_names{$mcc_node->get_name()} = $site;
        }
    }

    # ...
    # each member of this array will be a hash with keys disaster_node and
    # survivor_node and MCC::Node values.
    my $dr_state = $disaster_status->{DR_state};
    my @DR_disaster_pairs = ();
    if ($dr_state ne 'normal') {
        my @partnerships = @{ $self->_dr_group()->get_DR_partnerships() };
        foreach my $partnership (@partnerships) {
            my $disaster_node    = $partnership->get_partner_node(
                                    site => $disaster_status->{disaster_site});
            my $survivor_partner = $partnership->get_partner_node(
                                    site => $disaster_status->{survivor_site});
            push(@DR_disaster_pairs, { disaster_node    => $disaster_node,
                                       survivor_partner => $survivor_partner });
        }
    }

    # insert arcane metadata into each disk's hash where appropriate
    for my $serial (keys(%{ $disk_array_data })) {
        my $disk_state = $disk_array_data->{$serial};

        # meta__disk_locale (this is only placed in init disk data)
        if (! $self->_initial_disk_state_info_isset()) {
            my $owner_node_site = $site_vs_node_names{$disk_state->{owner}};
            my $pool            = $disk_state->{pool};
            if ($pool =~ /0$/) {
                # pool0 disks are at same site as orig owner
                $disk_state->{meta__disk_locale} = $owner_node_site;
            } else {
                # pool1 disks are housed at the dr partner site
                my $dr_partner_site = ($owner_node_site eq 'A') ? 'B' : 'A';
                $disk_state->{meta__disk_locale} = $dr_partner_site;
            }
        } else {
            $disk_state->{meta__disk_locale} =
                $self->_initial_disk_state_info()->{$serial}->
                    {meta__disk_locale};
        }

        # meta__disk_locale_is_up
        {
            $disk_state->{meta__disk_locale_is_up} =
                grep { $disk_state->{meta__disk_locale} eq $_ }
                     @{$disaster_status->{live_sites}};
        }

        # meta__disk_at_disaster_locale
        {
            my $disaster_site = $disaster_status->{disaster_site} || q{};
            $disk_state->{meta__disk_at_disaster_locale} =
                $disk_state->{meta__disk_locale} eq
                $disaster_site;
        }

        # meta__is_in_root_aggr
        {
            my $is_in_root_aggr =
                exists($root_aggregates->{$disk_state->{'aggregate'}}) ? 1 : 0;
            $disk_state->{meta__is_in_root_aggr} = $is_in_root_aggr;
        }

        # meta__ha_takeover_node
        {
            my $disk_in_root_aggr = $disk_state->{meta__is_in_root_aggr};
            my $HA_takeover_node = $self->_get_disk_HA_takeover_node(
                             disk_in_root_aggr => $disk_in_root_aggr,
                             disk_state        => $disk_state,
                             HA_disaster_pairs => \@HA_disaster_pairs, );
            if (defined($HA_takeover_node)) {
                $disk_state->{meta__ha_takeover_node} = $HA_takeover_node;
                # utility field when data struct dumped and filtered
                $disk_state->{meta__ha_TO_node_name} =
                                                 $HA_takeover_node->get_name();
            }
        }

        # meta__dr_partner_if_orig_owner_at_disaster_site
        {
            my $dr_partner_if_orig_owner_at_disaster_site = undef;
            if ($dr_state ne 'normal') {
                foreach my $disaster_pair (@DR_disaster_pairs) {
                    if ($disk_state->{'orig-owner'} eq
                                 $disaster_pair->{disaster_node}->get_name()) {
                        $dr_partner_if_orig_owner_at_disaster_site =
                                            $disaster_pair->{survivor_partner};
                        last;
                    }
                }
            }
            if (defined($dr_partner_if_orig_owner_at_disaster_site)) {
                $disk_state->{meta__dr_partner_if_orig_owner_at_disaster_site} =
                                    $dr_partner_if_orig_owner_at_disaster_site;
                # utility field when data struct dumped and filtered
                $disk_state->
                    {meta__dr_partner_name_if_orig_owner_at_disaster_site} =
                        $dr_partner_if_orig_owner_at_disaster_site->get_name();
            }
        }

        # meta__TO_node_of_dr_partner_if_OOADS (wanted to use:
        # meta__TO_node_of_dr_partner_if_orig_owner_at_disaster_site, but it was
        # too long, even for me. :)
        {
            my $dr_partner_if_OOADS =
                $disk_state->{meta__dr_partner_if_orig_owner_at_disaster_site};
            if (defined($dr_partner_if_OOADS)) {
                my $dr_group_id = $self->_dr_group()->get_id();
                my $ha_state = $disaster_status->{HA_states}->{$dr_group_id};
                foreach my $site (keys(%{ $ha_state })) {
                    if (    ($ha_state->{$site}->{HA_state} eq 'takeover') &&
                            ($ha_state->{$site}->{disaster_node} ==
                                                       $dr_partner_if_OOADS)) {
                        my $TO_node_of_dr_partner_if_OOADS =
                                           $ha_state->{$site}->{survivor_node};
                        if (defined($TO_node_of_dr_partner_if_OOADS)) {
                            $disk_state->{meta__TO_node_of_dr_partner_if_OOADS}=
                                               $TO_node_of_dr_partner_if_OOADS;
                            # utility field when data struct dumped and filtered
                            $disk_state->{meta__TO_node_name_of_DP_if_OOADS} =
                                   $TO_node_of_dr_partner_if_OOADS->get_name();
                        }
                        last;
                    }
                }
            }
        }
    }

    return $disk_array_data;
}

=head2 _validate_and_consolidate_multi_site_disk_state_views()

This method validates the consistency of the two clusters' views of the
disks' states when in normal MCC state.  (Not in DR, when only one cluster
is up/a valid viewer.)

Note that the validation that is done here is only a consistency check between
the views from the different sites.  The validity of this consolidated view
itself will be more fully checked later in the process.

It returns a consolidated data structure or throws an exception if the two
views are inconsistent with each other.

=cut

sub _validate_and_consolidate_multi_site_disk_state_views {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            disk_data_by_site => { type => HASHREF },
                  },
                            );

    my $disk_data_via_A = $opts{disk_data_by_site}->{'A'};
    my $disk_data_via_B = $opts{disk_data_by_site}->{'B'};

    my %golden_states = ();
    my %bad_disk_info = ();
    foreach my $serial (keys(%{ $disk_data_via_A })) {
        my $state_via_site_A = delete($disk_data_via_A->{$serial});
        my $state_via_site_B = delete($disk_data_via_B->{$serial});

        # storage layer (pool) seen by both sites
        my $pool = $state_via_site_A->{'pool'};
        unless ($pool eq $state_via_site_B->{'pool'}) {
            $bad_disk_info{'pool:'} = { $serial =>
                { 'A' => $state_via_site_A, 'B' => $state_via_site_B } };
        }

        # SANOWN layer field values seen by both sites
        my @sanown_fields = qw/ orig-owner reserver-id
                                home       home-id
                                owner      owner-id
                                dr-home    dr-home-id /;
        foreach my $sanown_field (@sanown_fields) {
            my $value_A = $state_via_site_A->{$sanown_field};
            unless ($value_A eq $state_via_site_B->{$sanown_field}) {
                $bad_disk_info{"$sanown_field:"} = { $serial =>
                    { 'A' => $state_via_site_A, 'B' => $state_via_site_B } };
            }
        }

        # RAID layer (plex, aggregate) seen ONLY by owner site!
        my %raid_fields = ( plex      => '-',
                            aggregate => '-', );
        foreach my $field (keys(%raid_fields)) {
            my $field_via_A = $state_via_site_A->{$field};
            my $field_via_B = $state_via_site_B->{$field};

            if (($field_via_A ne '-') && ($field_via_B ne '-')) {
                $bad_disk_info{"$field:"} = { $serial =>
                      { 'A' => $state_via_site_A, 'B' => $state_via_site_B } };
            } elsif ($field_via_A ne '-') {
                $raid_fields{$field} = $field_via_A;
            } elsif ($field_via_B ne '-') {
                $raid_fields{$field} = $field_via_B;
            } # else already taken care of with default '-' values above
        }

        # storage
        $golden_states{$serial} = { 'serial'     => $serial,
                                    'pool'       => $pool, };

        # SANOWN
        foreach my $sanown_field (@sanown_fields) {
            # for these, values seen by both sites assumed identical (or error
            # will be reported below.)
            $golden_states{$serial}->{$sanown_field} =
                                            $state_via_site_A->{$sanown_field};
        }

        # RAID
        foreach my $raid_field (keys(%raid_fields)) {
            $golden_states{$serial}->{$raid_field} = $raid_fields{$raid_field};
        }
    }

    # make sure both sites saw the same set of disks
    # (both of these hashes should be empty at this point)
    if (keys(%{ $disk_data_via_A }) || keys(%{ $disk_data_via_B })) {
        $bad_disk_info{'unmatched serials:'} = { 'A' => $disk_data_via_A };
        $bad_disk_info{'unmatched serials:'} = { 'B' => $disk_data_via_B };
    }

    # did we find any inconsistent disk info?
    if (keys(%bad_disk_info) > 0) {
          WaflMCC::MCC::Exceptions::InconsistentDiskState->throw(
               "ERROR: Inconsistent disk states as seen from sites A and B\n" .
               $self->_dump_some_disk_data(disk_data => \%bad_disk_info));
    }

    return \%golden_states;
}

=head2 _determine_root_aggregates()

Determines the root aggregates of each node in the dr group.  Returns a data
struct like this:

    {
        $root_aggr_1_name => {
                 node      => $mcc_node, (a WaflMCC::MCC::Node object)
                 aggregate => $mcc_aggr, (a WaflMCC::MCC::Aggregate object)
                           },
        $root_aggr_2_name => {
                 node      => $mcc_node, (a WaflMCC::MCC::Node object)
                 aggregate => $mcc_aggr, (a WaflMCC::MCC::Aggregate object)
                           },
        ...
    }

=cut

sub _determine_root_aggregates {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;

    my $root_aggrs_info = { };
    foreach my $site ('A', 'B') {
        my $mcc_nodes = $self->_dr_group()->get_mcc_nodes(site => $site);
        foreach my $mcc_node (@{ $mcc_nodes }) {
            my $root_aggr = $mcc_node->L_get_root_aggregate();
            $root_aggrs_info->{$root_aggr->get_name()} =
                                { node => $mcc_node, aggregate => $root_aggr };
        }
    }

    return $root_aggrs_info;
}

=head2 _check_for_disk_presence()

This method checks to make sure all disks that were in the shared storage at
start are still appearing.

=cut

sub _check_for_disk_presence {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            init_states    => { type  => HASHREF },
            current_states => { type  => HASHREF },
                   },
                            );

    my $init_states    = delete($opts{init_states});
    my $current_states = delete($opts{current_states});

    my %current_serials = map { $_ => 1 } keys(%{ $current_states });

    my @failed_disks = ();
    foreach my $serial (keys(%{ $init_states })) {
        if (exists($current_serials{$serial})) {
            delete($current_serials{$serial});
        } else {
            push(@failed_disks, $serial);
        }
    }

    my @new_disks = keys(%current_serials);

    my $return_msg = { failed_disks => q{}, new_disks => q{} };
    if (@failed_disks > 0) {
        my $msg = "Failed disks discovered:\n    (inital states):\n";
        foreach my $failed_serial (@failed_disks) {
            $msg .= $self->_dump_some_disk_data(
                           disk_data => $init_states->{$failed_serial}) . "\n";
        }
        $return_msg->{failed_disks} = $msg;
    }
    if (@new_disks > 0) {
        my $msg = "New disks discovered:\n    (current states):\n";
        foreach my $new_serial (@new_disks) {
            $msg .= $self->_dump_some_disk_data(
                              disk_data => $init_states->{$new_serial}) . "\n";
        }
        $return_msg->{new_disks} = $msg;
    }

    return $return_msg;
}

=head2 _validate_disk_states__owner_and_id()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__owner_and_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state                  => { type  => SCALAR },
            init_states               => { type  => HASHREF },
            current_states            => { type  => HASHREF },
                   },
                            );

    my $dr_state       = delete($opts{dr_state});
    my $init_states    = delete($opts{init_states});
    my $current_states = delete($opts{current_states});

    my @disk_state_errors = ();
    my @disk_state_warnings = ();
    # see _validate_disk_states__owner_and_id__footnote_1
    my $supposed_pre_DR_TO_node = undef;
    foreach my $serial (keys(%{ $current_states })) {
        my $owner    = $current_states->{$serial}->{'owner'};
        my $owner_id = $current_states->{$serial}->{'owner-id'};

        # both sites had this info at init
        my $orig_owner    = $init_states->{$serial}->{'owner'};
        my $orig_owner_id = $init_states->{$serial}->{'owner-id'};

        # the disk currently _claims_ its original owner is:
        unless ($current_states->{$serial}->{'orig-owner'} eq $orig_owner) {
            push(@disk_state_errors, $serial);
        }

        # if disk is in HA sfo, this is the takeover MCC::Node object
        my $ha_takeover_node =
                        $current_states->{$serial}->{'meta__ha_takeover_node'};

        # as far as ownership is concerned, if ha_takeover_node is defined, then
        # it has already been filtered to be part of an HA event on the survivor
        # site (or either live site if DR is normal), so owner will just be the
        # takeover node
        if (defined($ha_takeover_node)) {
            # part of an HA sfo.  owner should be home's HA partner
            unless (($owner    eq $ha_takeover_node->get_name()) &&
                    ($owner_id eq $ha_takeover_node->get_nvramid())) {
                push(@disk_state_errors, $serial);
            }
        } elsif ($dr_state eq 'normal') {
            # do the DR cases (HA is normal on survivor site and has no effect
            # here if it occurred previously on the disaster site)

            # there should be no change from initial values
            unless (($owner eq $orig_owner) && ($owner_id eq $orig_owner_id)) {
                push(@disk_state_errors, $serial);
            }
        } else {
            my $pool = $current_states->{$serial}->{'pool'};

            # here, $dr_partner_if_OOADS is defined only if $orig_owner is on
            # disaster side.  if defined, $dr_partner_if_OOADS is an MCC::Node
            # object.
            my $dr_partner_if_OOADS = $current_states->{$serial}->
                             {meta__dr_partner_if_orig_owner_at_disaster_site};

            # it is possible that the survivor site is in an HA takeover state.
            # this tells us if the dr partner of the original owner of this disk
            # has been taken over.  if so, the owner will be that dr partner's
            # takeover node.  if this is defined, it is an MCC::Node object.
            my $survivor_site_TO_node = $current_states->
                             {$serial}->{meta__TO_node_of_dr_partner_if_OOADS};

            if ($dr_state eq 'switchover_completed') {
                # if the disk is in pool1 of a disaster site node, the new
                # owner should be the dr-partner node of said node
                if (defined($dr_partner_if_OOADS) && ($pool ne 'Pool0')) {
                    if (defined($survivor_site_TO_node)) {
                        # ...or possibly just survivor node at survivor site
                        unless (($owner    eq
                                      $survivor_site_TO_node->get_name()) &&
                                ($owner_id eq
                                      $survivor_site_TO_node->get_nvramid())) {
                            push(@disk_state_errors, $serial);
                        }
                    } else {
                        unless (($owner    eq
                                        $dr_partner_if_OOADS->get_name()) &&
                                ($owner_id eq
                                        $dr_partner_if_OOADS->get_nvramid())) {
                            push(@disk_state_errors, $serial);
                        }
                    }
                } else {
                    # otherwise, there should be no change from initial vals
                    if (    ($owner    ne $orig_owner) ||
                            ($owner_id ne $orig_owner_id)) {
                        # See _validate_disk_states__owner_and_id__footnote_1
                        # below for description of this exceptional case.
                        if (! defined($supposed_pre_DR_TO_node)) {
                            if ($self->_are_an_HA_pair(
                                      node_names => [ $owner, $orig_owner ])) {
                                $supposed_pre_DR_TO_node = $owner;
                            }
                        }

                        # there can only be one supposed_pre_DR_TO_node
                        if (defined($supposed_pre_DR_TO_node) &&
                                        ($owner eq $supposed_pre_DR_TO_node)) {
                            # log a warning only
                            push(@disk_state_warnings, $serial);
                        } else {
                            # this is the usual case here, signifies an error
                            push(@disk_state_errors, $serial);
                        }
                    }
                }
            } elsif ($dr_state =~ /^heal_(aggrs|roots)_completed$/) {
                my $phase = $1;
                # if healing roots, and disk is a root aggr disk, owner should
                # be original owner
                my $disk_in_root_aggr =
                            $init_states->{$serial}->{'meta__is_in_root_aggr'};
                if (($phase eq 'roots') && $disk_in_root_aggr) {
                    unless (($owner    eq $orig_owner) &&
                            ($owner_id eq $orig_owner_id)) {
                        push(@disk_state_errors,
                            {   'serial' => $serial,
                                'violation' => [
                                    'root aggr owner/startup owner mismatch',
                                ],
                            }
                        );
                    }
                # if the disk is in either pool of a disaster site node, the
                # new owner should be the dr-partner node of said node...
                } elsif (defined($dr_partner_if_OOADS)) {
                    # ...or possibly just survivor node at survivor site
                    if (defined($survivor_site_TO_node)) {
                        unless (($owner    eq
                                      $survivor_site_TO_node->get_name()) &&
                                ($owner_id eq
                                      $survivor_site_TO_node->get_nvramid())) {
                            push(@disk_state_errors,
                                {   'serial' => $serial,
                                    'violation' => [
                                        'survivor site TO node mismatch',
                                    ],
                                }
                            );
                        }
                    } else {
                        unless (($owner    eq
                                        $dr_partner_if_OOADS->get_name()) &&
                                ($owner_id eq
                                        $dr_partner_if_OOADS->get_nvramid())) {
                            push(@disk_state_errors,
                                {   'serial' => $serial,
                                    'violation' => [
                                        'dr partner mismatch (no survivor TO)',
                                    ],
                                }
                            );
                        }
                    }
                } else {
                    # otherwise, there should be no change from initial vals
                    unless (($owner    eq $orig_owner) &&
                            ($owner_id eq $orig_owner_id)) {
                        push(@disk_state_errors,
                            {   'serial' => $serial,
                                'violation' => [
                                    'owner/startup owner mismatch',
                                ],
                            }
                        );
                    }
                }
            } else {
                confess("Unknown DR state: " . $dr_state . "\n");
            }
        }
    }

    if (scalar(@disk_state_warnings)) {
        my $message = <<END_PRINT;
The following disks have unusal (but possibly valid) 'owner' states.  You should
be sure they correspond to the states you expect for them in this particular
test:
END_PRINT

        $message .= "\t" . join("\n\t", @disk_state_warnings) . "\n";
        $Log->warn($message);
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'owner' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            my $ref;
            #thrown in for debugging, this should potentially be rearchitected.
            if (ref $serial eq 'HASH') {
                $ref = $serial;
                $serial = $ref->{'serial'};
            }
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "          pool: " .
                             $init_states->{$serial}->{'pool'} . "\n";
            $error_msg .= "            initial owner: " .
                             $init_states->{$serial}->{'owner'} . "\n";
            $error_msg .= "            currently _claimed_ original owner: " .
                             $current_states->{$serial}->{'orig-owner'} . "\n";
            $error_msg .= "            current owner: " .
                                $current_states->{$serial}->{'owner'} . "\n";
            $error_msg .= "            initial owner id: " .
                             $init_states->{$serial}->{'owner-id'} . "\n";
            $error_msg .= "            current owner id: " .
                                $current_states->{$serial}->{'owner-id'} . "\n";
            $error_msg .= "            current reserver id: " .
                           $current_states->{$serial}->{'reserver-id'} . "\n";
            if ($ref->{violation}) {
                #$Log->comment('current states: ' . WaflMCC::MCC::Utils->
                #    dump(target => $current_states->{$serial}));
                #$Log->comment("init states: " .
                #    Dumper $init_states->{$serial});
                $error_msg .= join(q{},
                    map { "            violation: $_\n" }
                    @{$ref->{violation}});
            }
        }
    } else {
        $Log->comment("\n SANOWN:   owner /   owner-id state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}
################################################################################
# _validate_disk_states__owner_and_id__footnote_1:
# there is a case here where an HA event may have occurred at the DR disaster
# site, but we cannot see this in the disaster status because site B is not
# queryable.  if an HA takeover had occurred at the DR disaster site, then the
# owner of non-root aggr pool 0 disks may be the HA partner of the original
# owner.  watch for this and print a warning to the logs so testers are aware of
# an "oddity" that MAY not be as it should.
################################################################################

# a quick check on whether 2 nodes are HA Partners or not
sub _are_an_HA_pair {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            node_names => { type => ARRAYREF },
                  },
                            );

    my $node_names = delete($opts{node_names});

    my @ha_groups = values(%{ $self->_dr_group()->get_HA_groups() });
    foreach my $ha_group (@ha_groups) {
        my $nodes = $ha_group->get_nodes();
        if (   (($nodes->[0]->get_name() eq $node_names->[0]) &&
                ($nodes->[1]->get_name() eq $node_names->[1])) ||
               (($nodes->[0]->get_name() eq $node_names->[1]) &&
                ($nodes->[1]->get_name() eq $node_names->[0]))  ) {
            return 1;
        }
    }

    return 0;
}


=head2 _validate_disk_states__dr_home_and_id()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__dr_home_and_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state                  => { type  => SCALAR },
            init_states               => { type  => HASHREF },
            current_states            => { type  => HASHREF },
                   },
                            );

    my $dr_state                  = delete($opts{dr_state});
    my $init_states               = delete($opts{init_states});
    my $current_states            = delete($opts{current_states});

    my @disk_state_errors = ();
    foreach my $serial (keys(%{ $current_states })) {
        my $dr_home    = $current_states->{$serial}->{'dr-home'};
        my $dr_home_id = $current_states->{$serial}->{'dr-home-id'};

        my $disk_in_root_aggr =
                            $init_states->{$serial}->{'meta__is_in_root_aggr'};

        # both sites had this info at init
        my $orig_owner    = $init_states->{$serial}->{'owner'};
        my $orig_owner_id = $init_states->{$serial}->{'owner-id'};

        # $dr_partner_if_OOADS is defined only if $orig_owner is on disaster
        # side.  if defined, $dr_partner_if_OOADS is an MCC::Node object.
        my $dr_partner_if_OOADS = $current_states->{$serial}->
                             {meta__dr_partner_if_orig_owner_at_disaster_site};

        # do the cases
        if ($dr_state eq 'normal') {
            # dr-home and id should be unset
            unless (($dr_home eq q{-}) && ($dr_home_id eq q{-})) {
                push(@disk_state_errors,
                    {   'serial' => $serial,
                        'violation' => [
                            'dr-home should be -: root disk',
                        ],
                    }
                );
            }
        } else {
            my $pool = $current_states->{$serial}->{'pool'};
            my $disk_locale_up =
                $current_states->{$serial}->{'meta__disk_locale_is_up'};
            my $disk_at_disaster_site =
                $current_states->{$serial}->{'meta__disk_at_disaster_locale'};

            if ($dr_state eq 'switchover_completed') {
                # if the disk is in pool1 of a disaster site node, the new
                # dr-home should be the original owner
                if (defined($dr_partner_if_OOADS) && ($pool ne 'Pool0')) {
                    unless (($dr_home    eq $orig_owner) &&
                            ($dr_home_id eq $orig_owner_id)) {
                        push(@disk_state_errors, $serial);
                    }
                } else {
                    # otherwise, they should not be set
                    unless (($dr_home eq q{-}) && ($dr_home_id eq q{-})) {
                        push(@disk_state_errors, $serial);
                    }
                }
            } elsif ($dr_state =~ /^heal_(aggrs|roots)_completed$/) {
                my $phase = $1;
                if (($phase eq 'roots') &&
                    $disk_in_root_aggr
                ) {
                    # if healing roots, and disk is a root disk on a disaster
                    # side node, AND disaster node is NOT booted,
                    # dr-home should be set to owner
                    #
                    #if (!$disaster_site_up) {
                    #for now we don't support the next state -
                    #heal roots + not booted ... only booted
                    if (0) {
                        unless (($dr_home    eq $orig_owner) &&
                                ($dr_home_id eq $orig_owner_id)) {
                            push(@disk_state_errors,
                                {   'serial' => $serial,
                                    'violation' => [
                                        "dr-home should be $orig_owner: " .
                                        'root disk owned by disaster site, ' .
                                        'disaster site NOT up',
                                    ],
                                }
                            );
                        }
                    } else {

                    # if healing roots, and disk is a root disk on a disaster
                    # side node, dr-home should be unset ('-')
                        unless (($dr_home    eq q/-/) &&
                                ($dr_home_id eq q/-/)) {
                            push(@disk_state_errors,
                                {   'serial' => $serial,
                                    'violation' => [
                                        'dr-home should be -: ' .
                                        'root disk owned by disaster site, ' .
                                        'disaster site up',
                                    ],
                                }
                            );
                        }
                    }
                # otherwise, if the disk is in either pool of a disaster
                # site node, dr-home should be set to the original owner
                } elsif (defined($dr_partner_if_OOADS)) {
                    unless (($dr_home    eq $orig_owner) &&
                            ($dr_home_id eq $orig_owner_id)) {
                        push(@disk_state_errors,
                            {   'serial' => $serial,
                                'violation' => [
                                    "dr-home should be $orig_owner" .
                                    ': disk originally owned by disaster site',
                                ],
                            }
                        );
                    }
                } else {
                    # otherwise, dr-home should should not be set
                    unless (($dr_home eq q{-}) && ($dr_home_id eq q{-})) {
                        push(@disk_state_errors,
                            {   'serial' => $serial,
                                'violation' => [
                                    'dr-home should be unset (-)',
                                ],
                            }
                        );
                    }
                }
            } else {
                confess("Unknown DR state: " . $dr_state . "\n");
            }
        }
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'dr-home' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            my $ref;
            #thrown in for debugging, this should potentially be rearchitected.
            if (ref $serial eq 'HASH') {
                $ref = $serial;
                $serial = $ref->{'serial'};
            }
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "          pool: " .
                              $init_states->{$serial}->{'pool'} . "\n";
            $error_msg .= "            initial owner: " .
                              $init_states->{$serial}->{'owner'} . "\n";
            $error_msg .= "            current owner: " .
                            $current_states->{$serial}->{'owner'} . "\n";
            $error_msg .= "            current dr-home: " .
                           $current_states->{$serial}->{'dr-home'} . "\n";
            $error_msg .= "            initial owner id: " .
                              $init_states->{$serial}->{'owner-id'} . "\n";
            $error_msg .= "            current dr-home id: " .
                           $current_states->{$serial}->{'dr-home-id'} . "\n";
            $error_msg .= "            current reserver id: " .
                           $current_states->{$serial}->{'reserver-id'} . "\n";
            if ($ref->{violation}) {
                #$Log->comment('current states: ' . WaflMCC::MCC::Utils->
                #    dump(target => $current_states->{$serial}));
                #$Log->comment('init states: ' .
                #    Dumper $init_states->{$serial});
                $error_msg .= join(q{},
                    map { "            violation: $_\n" }
                    @{$ref->{violation}});
            }
        }
    } else {
        $Log->comment("\n SANOWN: dr-home / dr-home-id state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}

=head2 _validate_disk_states__home_and_id()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__home_and_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state                  => { type  => SCALAR },
            init_states               => { type  => HASHREF },
            current_states            => { type  => HASHREF },
                   },
                            );

    my $dr_state                  = delete($opts{dr_state});
    my $init_states               = delete($opts{init_states});
    my $current_states            = delete($opts{current_states});

    my @disk_state_warnings = ();
    my @disk_state_errors = ();
    foreach my $serial (keys(%{ $current_states })) {
        my $home    = $current_states->{$serial}->{'home'};
        my $home_id = $current_states->{$serial}->{'home-id'};

        my $orig_home    = $init_states->{$serial}->{'home'};
        my $orig_home_id = $init_states->{$serial}->{'home-id'};

        # $dr_partner_if_OOADS is defined only if $orig_owner is on disaster
        # side.  if defined, $dr_partner_if_OOADS is an MCC::Node object.
        my $dr_partner_if_OOADS = $current_states->{$serial}->
                             {meta__dr_partner_if_orig_owner_at_disaster_site};

        # See _validate_disk_states__home_and_id__footnote_1.
        # this value is defined only if $dr_partner_if_OOADS above is defined,
        # and that node has been taken over in an HA event.  If defined, it is
        # an MCC::Node object.
        my $TO_node_of_dr_partner_if_OOADS =
            $current_states->{$serial}->{meta__TO_node_of_dr_partner_if_OOADS};

        # do the cases
        if ($dr_state eq 'normal') {
            # home should be the same as original home outside of DR
            unless (($home eq $orig_home) && ($home_id eq $orig_home_id)) {
                push(@disk_state_errors, $serial);
            }
        } elsif ($dr_state eq 'switchover_completed') {
            my $pool = $current_states->{$serial}->{'pool'};
            # if the disk is in pool1 of a disaster site node, the new
            # home should be the dr-partner node of said node
            if (defined($dr_partner_if_OOADS) && ($pool ne 'Pool0')) {
                if (    ($home    ne $dr_partner_if_OOADS->get_name()) ||
                        ($home_id ne $dr_partner_if_OOADS->get_nvramid())) {
                    # See _validate_disk_states__home_and_id__footnote_1
                    # below for description of this exceptional case.
                    if (defined($TO_node_of_dr_partner_if_OOADS) && ($home eq
                                $TO_node_of_dr_partner_if_OOADS->get_name())) {
                        # log a warning only
                        push(@disk_state_warnings, $serial);
                    } else {
                        # this is the usual case here, signifies an error
                        push(@disk_state_errors, $serial);
                    }
                }
            } else {
                # otherwise, there should be no change from initial vals
                unless (($home eq $orig_home) && ($home_id eq $orig_home_id)) {
                    push(@disk_state_errors, $serial);
                }
            }
        } elsif ($dr_state =~ /^heal_(aggrs|roots)_completed$/) {
            my $phase = $1;
            # if healing roots, and disk is a root aggr disk, home should
            # be original home
            my $disk_in_root_aggr =
                            $init_states->{$serial}->{'meta__is_in_root_aggr'};
            if (($phase eq 'roots') && $disk_in_root_aggr) {
                unless (($home    eq $orig_home) &&
                        ($home_id eq $orig_home_id)) {
                    push(@disk_state_errors, $serial);
                }
            # otherwise, if the disk is in either pool of a disaster site node,
            # the new home should be the dr-partner node of said node
            } elsif (defined($dr_partner_if_OOADS)) {
                if (    ($home    ne $dr_partner_if_OOADS->get_name()) ||
                        ($home_id ne $dr_partner_if_OOADS->get_nvramid())) {
                    # See _validate_disk_states__home_and_id__footnote_1
                    # below for description of this exceptional case.
                    if (defined($TO_node_of_dr_partner_if_OOADS) && ($home eq
                                $TO_node_of_dr_partner_if_OOADS->get_name())) {
                        # log a warning only
                        push(@disk_state_warnings, $serial);
                    } else {
                        # this is the usual case here, signifies an error
                        push(@disk_state_errors, $serial);
                    }
                }
            } else {
                # otherwise, there should be no change from initial vals
                unless (($home    eq $orig_home) &&
                        ($home_id eq $orig_home_id)) {
                    push(@disk_state_errors, $serial);
                }
            }
        } else {
            confess("Unknown DR state: " . $dr_state . "\n");
        }
    }

    if (scalar(@disk_state_warnings)) {
        my $message = <<END_PRINT;
The following disks have unusal (but possibly valid) 'home' states.  You should
be sure they correspond to the states you expect for them in this particular
test:
END_PRINT

        $message .= "\t" . join("\n\t", @disk_state_warnings) . "\n";
        $Log->warn($message);
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'home' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "          pool: " .
                           $init_states->{$serial}->{'pool'} . "\n";
            $error_msg .= "            current home: " .
                           $current_states->{$serial}->{'home'} . "\n";
            $error_msg .= "            current home id: " .
                           $current_states->{$serial}->{'home-id'} . "\n";
            $error_msg .= "            current dr-home: " .
                           $current_states->{$serial}->{'dr-home'} . "\n";
            $error_msg .= "            current dr-home id: " .
                           $current_states->{$serial}->{'dr-home-id'} . "\n";
            $error_msg .= "            current reserver id: " .
                           $current_states->{$serial}->{'reserver-id'} . "\n";
        }
    } else {
        $Log->comment("\n SANOWN:    home /    home-id state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}
################################################################################
# _validate_disk_states__home_and_id__footnote_1:
# there is a case here where an HA event may have occurred at the DR survivor
# site prior to the DR Switchover.  In this case the single survivor node on
# the DR survivor site will take over as the 'home' node for both disaster site
# nodes throughout the rest of the DR lifecycle.  watch for this and print a
# warning to the logs so testers are aware of an "oddity" that MAY not be as it
# should.
################################################################################

=head2 _validate_disk_states__reserver_id()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__reserver_id {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state                  => { type  => SCALAR },
            init_states               => { type  => HASHREF },
            current_states            => { type  => HASHREF },
                   },
                            );

    my $dr_state                  = delete($opts{dr_state});
    my $init_states               = delete($opts{init_states});
    my $current_states            = delete($opts{current_states});

    my @disk_state_warnings = ();
    my @disk_state_errors = ();
    foreach my $serial (keys(%{ $current_states })) {
        my $orig_owner_id = $init_states->{$serial}->{'owner-id'};

        my $pool        = $current_states->{$serial}->{'pool'};
        my $owner_id    = $current_states->{$serial}->{'owner-id'};
        my $reserver_id = $current_states->{$serial}->{'reserver-id'};

        # $dr_partner_if_OOADS is defined only if $orig_owner is on disaster
        # side.  if defined, $dr_partner_if_OOADS is an MCC::Node object.
        my $dr_partner_if_OOADS = $current_states->{$serial}->
                             {meta__dr_partner_if_orig_owner_at_disaster_site};
        # See _validate_disk_states__home_and_id__footnote_1.
        # this value is defined only if $dr_partner_if_OOADS above is defined,
        # and that node has been taken over in an HA event.  If defined, it is
        # an MCC::Node object.
        my $to_node_of_dr_partner_if_OOADS =
            $current_states->{$serial}->{meta__TO_node_of_dr_partner_if_OOADS};

        # do the cases
        if ($dr_state eq 'normal') {
            # reserver-id should be the same as owner id
            unless ($reserver_id eq $owner_id) {
                push(@disk_state_errors, $serial);
            }
        } elsif ($reserver_id == 0) {
            # reserver-id should be 0, but there are some times that it
            # could get into a weird state when reservations are released
            # ... however there is a burt where the owners are off, though
            # it may not be related at all to reserver_ids
            # also see line 1995 or so for another == 0 check
            push(@disk_state_warnings, $serial);
            $Log->comment("$serial reserver id is 0!");
            unless (WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2013/06/12', burt => '000000')) {
            }
        } elsif ($dr_state eq 'switchover_completed') {
            #IF surviving site is in HA takeover, the reserver_id should be
            #the surviving site node owner id,
            #else it is below (dr partner of disaster node)
            if (defined($to_node_of_dr_partner_if_OOADS)) {
                # this state: surviving site in HA takeover of dr partner of
                # original disk owner
                # remote disks originally owned by disaster site should now
                # have reserver-id set back to HA partner of dr partner of
                # original owner
                unless ($reserver_id eq
                        $to_node_of_dr_partner_if_OOADS->get_nvramid()) {
                    push(@disk_state_errors, $serial);
                }
            } elsif (defined($dr_partner_if_OOADS)) {
                # remote disks originally owned by disaster site should now
                # have reserver-id set back to original owner
                unless ($reserver_id eq $dr_partner_if_OOADS->get_nvramid()) {
                    push(@disk_state_errors, $serial);
                }
            } else {
                # Cdisks originally owned by survivor site- no change.
                unless ($reserver_id eq $owner_id) {
                    push(@disk_state_errors, $serial);
                }
            }
        } elsif ($dr_state =~ /^heal_(aggrs|roots)_completed$/) {
            # reserver-id should be same as owner-id
            unless ($reserver_id eq $owner_id) {
                push(@disk_state_errors, $serial);
            }
        } else {
            confess("Unknown DR state: " . $dr_state . "\n");
        }
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'reserver-id' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "          pool: " .
                           $init_states->{$serial}->{'pool'} . "\n";
            $error_msg .= "            current owner: " .
                           $current_states->{$serial}->{'owner'} . "\n";
            $error_msg .= "            current owner id: " .
                           $current_states->{$serial}->{'owner-id'} . "\n";
            $error_msg .= "            current home: " .
                           $current_states->{$serial}->{'home'} . "\n";
            $error_msg .= "            current home id: " .
                           $current_states->{$serial}->{'home-id'} . "\n";
            $error_msg .= "            current dr-home: " .
                           $current_states->{$serial}->{'dr-home'} . "\n";
            $error_msg .= "            current dr-home id: " .
                           $current_states->{$serial}->{'dr-home-id'} . "\n";
            $error_msg .= "            reserver id: " .
                           $current_states->{$serial}->{'reserver-id'} . "\n";
        }
    } else {
        $Log->comment("\n SANOWN:          reserver-id state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}

=head2 _validate_disk_states__pool()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__pool {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state       => { type  => SCALAR },
            init_states    => { type  => HASHREF },
            current_states => { type  => HASHREF },
                   },
                            );

    my $dr_state                  = delete($opts{dr_state});
    my $init_states               = delete($opts{init_states});
    my $current_states            = delete($opts{current_states});

    my @disk_state_errors = ();
    foreach my $serial (keys(%{ $current_states })) {
        my $pool      = $current_states->{$serial}->{'pool'};
        my $orig_pool = $init_states->{$serial}->{'pool'};

        # pool should never change
        unless ($pool eq $orig_pool) {
            push(@disk_state_errors, $serial);
        }
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'pool' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "            current pool:  " .
                           $current_states->{$serial}->{'pool'} . "\n";
            $error_msg .= "            original pool: " .
                           $init_states->{$serial}->{'pool'} . "\n";
        }
    } else {
        $Log->comment("\nStorage:                 pool state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}

=head2 _validate_disk_states__plex()

One of the disk state validation methods called by validate_disk_states().

=cut

sub _validate_disk_states__plex {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            dr_state                  => { type  => SCALAR },
            init_states               => { type  => HASHREF },
            current_states            => { type  => HASHREF },
                   },
                            );

    my $dr_state                  = delete($opts{dr_state});
    my $init_states               = delete($opts{init_states});
    my $current_states            = delete($opts{current_states});

    # for determining the cluster a node is in.
    my $dr_partnerships = $self->_dr_group()->get_DR_partnerships();

    my @disk_state_errors = ();
    # at each site, disks in pool1 may be in some plexN where N >= 1, but if
    # they are in this pool1 plexN, N must be the same for all disks at a given
    # site.  there may be a different plex number for pool1 disks at the other
    # site.
    my %the_non_zero_plex_IDs = ( A => undef, B => undef );
    foreach my $serial (keys(%{ $current_states })) {
        my $pool  = $current_states->{$serial}->{'pool'};
        my $owner = $current_states->{$serial}->{'owner'};

        # $dr_partner_if_OOADS is defined only if $orig_owner is on disaster
        # side.  if defined, $dr_partner_if_OOADS is an MCC::Node object.
        my $dr_partner_if_OOADS = $current_states->{$serial}->
                             {meta__dr_partner_if_orig_owner_at_disaster_site};

        # plex can only be reliably queried for by the owner of the disk.
        # this may not be possible if the current owner is a node on the
        # disaster side of a DR.
        #
        # after switchover, all disaster side pool0 disks are still owned
        # by disaster side nodes.  we will not check plex state on such
        # disks until heal_aggrs_complete, afterwhich CURRENT (not necessarily
        # initial) disk owners will always be up and in good state.
        if ( ! (($dr_state eq 'switchover_completed') &&
                       defined($dr_partner_if_OOADS) && ($pool eq 'Pool0')) ) {
            my $plex = $current_states->{$serial}->{'plex'};

            # plex should always be unset or be "plexN" where N is either
            #     a) N = 0 if in disk Pool0
            #     b) N > 0 if in disk Pool1, and additionally, all disks in
            #        a non-plex0 plexN should have the same value for N(site).
            #        (N can be different at each site.)
            if ($plex ne q{-}) {
                my ($pool_id) = ($pool =~ /^Pool(\d+)$/);  # note uppercase 'P'
                my ($plex_id) = ($plex =~ /^plex(\d+)$/);  # note lowercase 'p'

                if ($pool_id == 0) {
                    if ($plex_id != 0) {
                        push(@disk_state_errors, $serial);
                    }
                } else {
                    # $pool_id == 1
                    # $plex_id == N(site) > 0
                    my $disk_locale =
                                  $init_states->{$serial}->{meta__disk_locale};
                    if (defined($the_non_zero_plex_IDs{$disk_locale})) {
                        if ($plex_id != $the_non_zero_plex_IDs{$disk_locale}) {
                            push(@disk_state_errors, $serial);
                        }
                    } elsif ($plex_id > 0) {
                        # this is the first disk found from this plex
                        $the_non_zero_plex_IDs{$disk_locale} = $plex_id;
                    } else {
                        push(@disk_state_errors, $serial);
                    }
                }
            }
        }
    }

    my $error_msg = q{};
    if (scalar(@disk_state_errors)) {
        $error_msg .= "\n\nDisk 'plex' state errors:\n";
        foreach my $serial (@disk_state_errors) {
            $error_msg .= "        serial: $serial\n";
            $error_msg .= "            current plex:  " .
                           $current_states->{$serial}->{'plex'} . "\n";
            $error_msg .= "            current pool:  " .
                           $current_states->{$serial}->{'pool'} . "\n";
        }
    } else {
        $Log->comment("\n   RAID:                 plex state: All disks " .
                      "successfully validated.");
    }

    return $error_msg;
}

=head2 L_wait_for_slow_SANOWN_state_migration()

waits for dr_home states to all migrate to - before returning
(unless timeout is hit first)

This method is valid to be called only if in healed roots or normal DR state.

=cut

sub L_wait_for_slow_SANOWN_state_migration {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            timeout        => {type => SCALAR, default => 6000},
            poll_time      => {type => SCALAR, default => 30},
                   },
                            );

    # do nothing if disk validation is disabled.
    if ($ENV{DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION}) {
        $Log->warn("DISABLE_SHARED_STORAGE_DISK_STATE_VALIDATION set.\n" .
               "    Cannot check for slow dr-home migrartion after SB!");
        return q{};
    }

    my $timeout           = delete($opts{timeout});
    my $poll_time         = delete($opts{poll_time});

    my $end_time          = time + $timeout;
    my @disk_state_errors = ();

    my $error_msg = 'do at least one loop...';

    while ( (time < $end_time) && $error_msg ) {

        my $loop_start = time;
        my $current_states;
        try {
            # maybe we should do this ourselves as the exact errors may be
            # masked by this call?  [not sure what you mean.  bruce]

            # by definition both sites are live after heal roots or switchback
            my $disaster_status = $self->_metrocluster()->
                       L_get_disaster_status(known_live_sites => [ 'A', 'B' ]);
            $current_states = $self->_get_disk_array_data(
                                          disaster_status => $disaster_status);

            @disk_state_errors = ();
            $error_msg = q{};

            foreach my $serial (keys(%{ $current_states })) {
                my $dr_home = $current_states->{$serial}->{'dr-home'};
                # dr-home should be unset if in normal dr state
                if ($disaster_status->{DR_state} eq 'normal') {
                    unless ($dr_home eq '-') {
                        push(@disk_state_errors, $serial);
                    }
                }

                # reserver-id has the same/worse problem.  it should be set
                # to owner-id at heal roots and SB
                my $reserver_id = $current_states->{$serial}->{'reserver-id'};
                my $owner_id    = $current_states->{$serial}->{'owner-id'};
                if ($reserver_id == 0) {
                    $Log->comment("$serial reserver id is 0!");
                    #it really SHOULDN'T be 0...but it can be sometimes...
                    #this should probably be revisited... TODO
                    next;
                }

                unless ($reserver_id eq $owner_id) {
                    push(@disk_state_errors, $serial);
                }
            }

            if (scalar(@disk_state_errors)) {
                $error_msg = "Disk SANOWN state errors:\n";
                foreach my $serial (@disk_state_errors) {
                    $error_msg .= "        serial: $serial\n";
                    $error_msg .= "            current dr-home: " .
                            $current_states->{$serial}->{'dr-home'} . "\n";
                    $error_msg .= "            current owner-id: " .
                            $current_states->{$serial}->{'owner-id'} . "\n";
                    $error_msg .= "            current reserver-id: " .
                            $current_states->{$serial}->{'reserver-id'} . "\n";
                }
            }
        } catch WaflMCC::MCC::Exceptions::InconsistentDiskState with {
            my $exception = shift;
            $error_msg = "Invalid SANOWN (dr-home?) states detected...\n";
            $error_msg .= $exception->text();

            #for some reason this seemed like it didn't work all the time,
            #try repeat callings of this if it's not...
            $self->_metrocluster->get_cluster(site => 'A')->
                   L_workaround_force_reserver_id_update();
        };

        if ($error_msg) { #don't sleep if we are done...
            $Log->warn($error_msg . "\n    (report to BURT 541653?)");

            #this gives us a constant loop time (if tuned correctly)
            my $loop_wait_time = $loop_start + $poll_time - time;
            if ($loop_wait_time > 0) {
                $Log->comment("sleeping $loop_wait_time seconds " .
                              "($poll_time second loop time)");
                Tharn::sleep($loop_wait_time);
            } else {
                $Log->comment("not sleeping, ran over poll time by " .
                              (-1 * $loop_wait_time) .
                              " seconds ($poll_time second loop time)");
            }
        }
    }

    if (time > $end_time) {
        $Log->warn("Did not finish disk migrations before timing out!");
    }

    return $error_msg;
}


# checks if a disk is part of a non-root aggregate whose node has been taken
# over in an HA event.  returns the WaflMCC::MCC::Node object associated
# with the node that should have taken this disk over, or undef if not.
sub _get_disk_HA_takeover_node {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      disk_in_root_aggr => { regex => qr/^[01]$/ },
                      disk_state        => { type  => HASHREF },
                      HA_disaster_pairs => { type  => ARRAYREF },
                  },
                            );

    my $disk_in_root_aggr = delete($opts{disk_in_root_aggr});
    my $disk_state        = delete($opts{disk_state});
    my $HA_disaster_pairs = delete($opts{HA_disaster_pairs});

    my $HA_takeover_node = undef;
    # disks can go to sfo only if in a non-root aggregate.  spares NA.
    if ((! $disk_in_root_aggr) && ($disk_state->{'aggregate'} =~ /\w/)) {
        my $home = $disk_state->{'home'};

        foreach my $disaster_pair (@{ $HA_disaster_pairs }) {
            if ($disaster_pair->{disaster_node}->get_name() eq $home) {
                $HA_takeover_node = $disaster_pair->{survivor_node};
                last;
            }
        }
    }

    return $HA_takeover_node;
}

# utility for better dumps of disk data
sub _dump_some_disk_data {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      disk_data => { type  => HASHREF },
                  },
                            );

    # these are suppressed because they are objects which will dump large
    # quantities of data the user is likely not interested in sifting through
    my $fields_to_suppress = ['meta__ha_takeover_node',
                              'meta__dr_partner_if_orig_owner_at_disaster_site',
                              'meta__TO_node_of_dr_partner_if_OOADS' ];

    return WaflMCC::MCC::Utils->dump(
                          target                       => $opts{disk_data},
                          additional_suppress_patterns => $fields_to_suppress);
}

# this method tries to create a good sample of disks to report on in an
# abbreiviated form as opposed to the full report on all disks
sub _generate_sampled_disk_report {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $self = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
                      current_states => { type  => HASHREF },
                  },
                            );

#TBD

    return "Sampled Disk report not yet implemented.  Please use 'full'\n" .
           "log_level for more disk validation detail for now.";
}

# this code is not used currently.  but may be used in future, so keeping it in.
#
# private class used only by Shared_Storage class.
#package WaflMCC::MCC::_Shared_Storage_State_Snapshot;
#
#use NATE::BaseException qw(:try);
#use NACL::Exceptions::NoElementsFound qw(:try);
#
#use Tharn qw($Log);
#
#use Params::Validate
#    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
#
#use Class::MethodMaker [
#     new    => [ '-init', 'new' ],
#     scalar => [
#         qw/ _disk_state_info
#             _DR_state
#             _HA_state /,
#               ],
#                       ];
#
# implicitly called by Class::MethodMaker's new (see -init)
#sub init {
#    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

#    my $self = shift;
#    my %opts = validate_with(
#        params => \@_,
#        spec   => {
#                      A1 => { type => SCALAR },
#                  },
#                            );
#
#    # this is used only to get access to a site A NACL::C::Metrocluster object
#    $self->_a_site_A_node(NACL::C::SystemNode->new(
#                                           hostrec => Tharn::host($opts{A1})));
#
#    unless ($self->_a_site_A_node_isset()) {
#        confess("Can't get a site A node!");
#    }
#}

1;

