# $Id $

#
# Copyright (c) 2011-2021 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

# this is here due to our standard usage of @_ in Params::Validate methods.
##no critic (RequireArgUnpacking ProhibitFixedStringMatches)

## @summary MCC Utils Module
## @author dl-mcc-dpg@netapp.com
## @status shared
## @pod here

package WaflMCC::MCC::Utils;

use strict;
use warnings;
use Carp;

use base qw (NACL::MTask::MTask);

use WaflMCC::MCC::Log;

use Tharn qw($Log param);

use Params::Validate
    qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);

use List::MoreUtils ();
use Date::Calc ();

use Data::Dumper ();

=head1 NAME

WaflMCC::MCC::Utils

=head1 DESCRIPTION

This library-only class is to be used for C-Mode Metrocluster testing.

    use WaflMCC::MCC::Utils;
    ...
    print WaflMCC::MCC::Utils->dump(target => $my_object);

=cut

=head1 Public Methods

There is no constructor for this class.  It consists only of class methods,
always called as

    WaflMCC::MCC::Utils->utility_method(...);

=head2 dump()

    print WaflMCC::MCC::Utils->dump($my_object);

    print WaflMCC::MCC::Utils->dump(target => $my_object);

    print WaflMCC::MCC::Utils->dump(target              => $my_object,
                                        show_shared_storage => 1);

This is a replacement utility for Data::Dumper::Dumper() which is too verbose
when it comes to dumping objects which contain things like NACL command
interfaces and apisets, which are huge data structures very rarely of
interest to end users or library programmers.

This method uses Data::Dumper::Dumper() but filters out key/value pairs for all
the following matches as in key =~ /$value/ (which means the key '_apisets' will
be suppressed, for instance), no matter where they appear in the data structure
hierarchy:

    qw/ command_interface version_manager apiset hostrec /

All other data in the object or structure is still returned as usual.

=over

=item Options and/or Arguments

=over

=item C<< target => $data_structure >>

(Optional, kind of) target can be any sort of perl data.  If there is no target
specified, then the implied target is the one and only argument passed.

=item C<< show_shared_storage => (0|1)    (default = 0) >>

(Optional) Since dumping shared storage data can also be very lengthy, dump()
does not show shared storage data by default.  If you wish the dump to
show shared storage data, then you must use the longer form of the call to
dump, with both 'target' and 'show_shared_storage' explicitly specified.

=back

=back

=cut

# this is a global variable because the callback routine _S_dumper_filter
# needs to access it, but it needs to be able to be modified by the dump method.
my @key_suppress_list = qw/ command_interface
                             version_manager
                             apiset
                             hostrec
                             _sysconfig_va_response
                             nsshow_raw
                             switchshow_raw /;

sub dump {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $pkg = shift;

    # all this garbage is to allow a single non-keyed value to pass as "target"
    my @opts_v1 = @_;
    my @opts_v2 = undef;
    if (scalar(@opts_v1) == 1) {
        @opts_v2 = ('target', pop(@opts_v1));
    } elsif (List::MoreUtils::any { $_ eq 'target' } @opts_v1) {
        @opts_v2 = @opts_v1;
    } else {
        @opts_v2 = ('target', \@opts_v1);
    }

    my %opts = validate_with(
        params => \@opts_v2,
        spec   => {
             target              => 1,
             depth               => { regex => qr/^\d+$/,  optional => 1, },
             additional_suppress_patterns
                                 => { type  => ARRAYREF,   optional => 1, },
                  },
                            );

    my $data_object = delete($opts{target});

    my @original_list = @key_suppress_list;

    # more stuff to suppress?
    if (defined($opts{additional_suppress_patterns})) {
        push(@key_suppress_list, @{ $opts{additional_suppress_patterns} });
    }

    local $Data::Dumper::Sortkeys = \&_S_dumper_filter;
    local $Data::Dumper::Deepcopy = 1;
    local $Data::Dumper::Indent = 1;

    # set depth of dump?
    local $Data::Dumper::Maxdepth = defined($opts{depth}) ? $opts{depth} : 0;

    my $result = Data::Dumper::Dumper($data_object);

    @key_suppress_list = @original_list;

    return $result;
}

sub _S_dumper_filter {
    my ($hash) = @_;

    my @keys = keys(%{ $hash });
    my $keys_to_suppress_regex = join(q{|}, @key_suppress_list);

    my @keys_suppressed = grep { $_ =~ /(?:$keys_to_suppress_regex)/ } @keys;
    @keys_suppressed = map { "${_}__SUPPRESSED" } @keys_suppressed;

    my @keys_to_dump = grep { $_ !~ /(?:$keys_to_suppress_regex)/ } @keys;
    push(@keys_to_dump, @keys_suppressed);

    return \@keys_to_dump;
}

=head2 workaround_expired()

    if (! WaflMCC::MCC::Utils->workaround_expired(
                          expiration_date => '2012/04/28', burt => '123456')) {
        do_some_ugly_workaround_that_should_not_operate_in_perpetuity();
    }

This is a simple class method that just returns true if the expiration date is
earlier than the date that the test is being run.  In essence this is a
dead-man's switch to make workarounds stop working after a set date, the idea
being that the issue being worked around should have already been fixed by that
date.

When this method is called after the expiration date a warning is issued that
the workaround is no longer operating, and should either be given a new
expiration date or should be removed from the code.

=over

=item Options and/or Arguments

=over

=item C<< expiration_date => 'YYYY/MM/DD' >>

(Required) The expiration date after which workaround_expired() should return
true.

=item C<< burt => $burt_number >>

(Required) This is the BURT number of the issue that is being worked around.

=back

=back

=cut

sub workaround_expired {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics
    my $pkg_or_object = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      expiration_date => { regex => qr{^\d\d\d\d/\d+/\d+$} },
                      burt            => { regex => qr/^\d\d\d\d\d\d$/ },
                  },
                            );

    my @today = Date::Calc::Today([time()]);
    my @expiration_date = split(qr{/}, $opts{expiration_date});

    my $days_expired = Date::Calc::Delta_Days(@expiration_date, @today);

    if ($days_expired < 0) {
        $Log->comment("\n");
        $Log->comment('Workaround for BURT ' . $opts{burt} . ' being used...');
        $Log->comment("\n");
        return 0;
    } else {
        $Log->warn('Workaround for BURT ' . $opts{burt} . ' has expired!');
        return 1;
    }
}

=head2 S_workaround_unconfigure__delete_aggrs_vservers_on_all_nodes()

    WaflMCC::MCC::Utils->
            S_workaround_unconfigure__delete_aggrs_vservers_on_all_nodes(
                                         node_components => \@node_components);

This static method is a workaround to an issue related to metrocluster
unconfigure.  Currently, before 'metrocluster unconfigure' is called, we
must delete all cluster/data VServers and non-root aggregates on all nodes.

=over

=item Options and/or Arguments

=over

=item C<< node_components => $node_components >>

(Required) An arrayref of NACL::C::Node objects associated with each of the
nodes in the MCC system.

=back

=back

=cut

# placed here because these mods are specifc to this workaround
use NACL::C::StorageDisk ();
use NACL::CS::Vserver ();
use NACL::STask::Aggregate ();
use NACL::STask::Vserver ();

sub S_workaround_unconfigure__delete_aggrs_vservers_on_all_nodes {
    my $trace = WaflMCC::MCC::Log::TRACELOG->start_timer(); #metrics

    my $pkg = shift;

    my %opts = validate_with(
        params => \@_,
        spec   => {
                      node_components => { type => ARRAYREF },
                  },
                            );

    my $CIFS_ADMIN_USERNAME =
                  Tharn::param('CIFS_ADMIN_USERNAME', '-default', 'labclient');
    my $CIFS_ADMIN_PASSWORD =
                  Tharn::param('CIFS_ADMIN_PASSWORD', '-default', 'sundance');

    foreach my $node_component (@{ $opts{node_components} }){
        $Log->comment('Finding all non-root aggregates');
        ### Get all non-root aggrs
        my @aggrs = NACL::STask::Aggregate->find(
                                       command_interface => $node_component,
                                                  filter => { root => 'false' },
                                             allow_empty => 1 );

        $Log->comment('Find and delete all data/cluster vservers if ' .
                      'their root volumes are on non-root aggrs');

        ### Get all data/cluster Vservers
        my @vservers = NACL::CS::Vserver->fetch(
                                command_interface => $node_component,
                                           filter => { type => 'data|cluster' },
                                      allow_empty => 1 );

        if ( scalar(@vservers) && scalar(@aggrs) ){
            foreach my $vserver (@vservers){
                foreach my $aggr (@aggrs){
                    if( $vserver->aggregate eq $aggr->aggregate ){
                       NACL::STask::Vserver->purge(
                                    command_interface => $node_component,
                                              vserver => $vserver->vserver,
                               nacltask_cifs_username => $CIFS_ADMIN_USERNAME,
                               nacltask_cifs_password => $CIFS_ADMIN_PASSWORD );
                    }
                }
            }
        }

        ### Remove all non-root aggregates
        if ( scalar(@aggrs) ){
            $Log->comment('Removing all non-root aggregates');
            map { $_->purge() } @aggrs;

            # Start zeroing disks from deleted aggrs
            NACL::C::StorageDisk->zerospares(owner => $node_component->name,
                                 command_interface => $node_component);
        }
    }
}

1;
