# $Id$

# Library to support WAFL space test cases

#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @name WaflSpaceLib.pm
## @summary WaflSpaceLib having the subroutines used by Wafl Space Testcases
## Tests
## @status Development
## @author dl-rakesh-automation@netapp.com
## @pod here

=head1 NAME

WaflSpaceLib

=head1 DESCRIPTION

WaflSpaceLib -   Library that supports  different operations WAFL which are
helpful to automate WAFL test scenarios.

Uses NACL APISet/Components

=head1 METHODS

=cut

# Compiler directives.
use strict;
use warnings;

# =============================================================
# Module Import
# =============================================================

require Exporter;

package WaflSpaceLib;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF BOOLEAN UNDEF);
use vars qw (@ISA @EXPORT_OK);
use Data::Dumper;
use Tharn qw($Log);
use POSIX;

# =============================================================
# NACL Components Used
# =============================================================

use NACL::C::Volume;
use NACL::C::Node;
use NACL::C::Aggregate;
use NACL::C::Quota;
use NACL::C::Qtree;
use NACL::C::Statistics;
use NACL::CS::Statistics;
use NACL::C::VserverCifsShare;
use FS::EA::AggregateTask;
use FS::EA::VolumeTask;

#BEGIN {use vars qw{$Node}};
#$Node = NACL::C::Node->find();

# Method export

@ISA       = qw(Exporter);
@EXPORT_OK = qw(find_delete_all_volumes
    count_file_blocks
    convert_size
    find_delete_created_volumes
    enable_quota
    get_statistics
    create_share
    create_vol
    delete_cifs_share
    qtree_create
    set_security_style
    aggr_bit_type
    list_file_attributes
    get_counter
    set_file_attributes
    calculate_overwrite_reserve
    count_file_blocks_w_lvl
);

=head2 count_file_blocks()

count_file_blocks(fs_size => $file_size, aggr_type => $aggr_type);

calculates the number of blocks which can be used to compare OR and HR

=over

=item Arguments

=over

=item C<< fs_size=>$file_size >>

(Required) Size of the file in bytes

=item C<< aggr_type=>$aggr_type >>

(Required) Aggregate Block-Type 32/64

=item Return Value

returns the block count

=back

=back

=cut

sub count_file_blocks {

    my %opts      = @_;
    my $aggr_type = $opts{aggr_type};
    my $filesize  = $opts{fs_size};
    my ( $blocks, $count_blocks );

    ## for smaller files why we add 1 block
    $Log->comment("starting block count with $filesize bytes");
    my $l0_blocks = ( $filesize - ( $filesize % 4096 ) ) / 4096;

    if ( $filesize % 4096 > 0 ) {
        $l0_blocks = $l0_blocks + 1;
    }
    $blocks = $count_blocks = $l0_blocks;
    if ( $aggr_type == 32 ) {

        while ( $blocks > 8 ) {
            $count_blocks = $count_blocks + ceil( $blocks / 510 );
            $blocks       = ceil( $blocks / 510 );
        }
    }
    if ( $aggr_type == 64 ) {
        while ( $blocks > 5 ) {
            $count_blocks = $count_blocks + ceil( $blocks / 255 );
            $blocks       = ceil( $blocks / 255 );
        }
    }
    return $count_blocks;
}

=head2 count_file_blocks_w_lvl()

count_file_blocks_w_lvl(fs_size => $file_size, aggr_type => $aggr_type);

calculates the number of blocks which can be used to compare OR and HR

=over

=item Arguments

=over

=item C<< fs_size=>$file_size >>

(Required) Size of the file in bytes

=item C<< aggr_type=>$aggr_type >>

(Required) Aggregate Block-Type 32/64

=item Return Value

returns the block count and lvl value

=back

=back

=cut

sub count_file_blocks_w_lvl {
    my %opts      = @_;
    my $filesize  = $opts{fs_size};
    my $Aggr_type = $opts{aggr_type};
    my ( $blocks, $count_blocks );
    ## for smaller files why we add 1 block
    $Log->comment("starting block count with $filesize bytes");
    my $l0_blocks = ( $filesize - ( $filesize % 4096 ) ) / 4096;
    my $lvl = 0;

    if ( $filesize % 4096 > 0 ) {
        $l0_blocks = $l0_blocks + 1;
    }
    $blocks = $count_blocks = $l0_blocks;
    if ( $Aggr_type == 32 ) {
        while ( $blocks > 8 ) {
            $count_blocks = $count_blocks + ceil( $blocks / 510 );
            $blocks       = ceil( $blocks / 510 );
            $lvl          = $lvl + 1;
        }
        $Log->comment("block count is: $count_blocks blocks");
    }
    if ( $Aggr_type == 64 ) {
        while ( $blocks > 5 ) {
            $count_blocks = $count_blocks + ceil( $blocks / 255 );
            $blocks       = ceil( $blocks / 255 );
            $lvl          = $lvl + 1;
        }
        $Log->comment("block count is: $count_blocks blocks");
    }
    return ( $count_blocks, $lvl );
}

=head2 convert_size()

convert_size(1094k);

Converts the give size into bytes, Kb, Mb, Gb, Tb. Accepts argument in any
unit (Unit has to be append to number).

=over

=item Arguements

=over

=item C<< size >>

(Required) Size to be converted

=item C<< to_be_rounded >>

(Optional) <boolean> if true, returned value will be rounded else not.

=back

=back

=over

=item Return Value

Returns a reference to a hash which has Units as keys and converted values as
value of the element.

=back

=cut

sub convert_size {
    my ( $size, $to_be_runded ) = @_;
    my ( %size, $bytes );
    my %hash;
    $bytes = $size;

    if ( $size =~ /(\d+\.{0,1}\d{0,})m/i ) {
        $bytes = $1 * 1048576;
    }
    elsif ( $size =~ /(\d+\.{0,1}\d{0,})g/i ) {
        $bytes = $1 * 1073741824;
    }
    elsif ( $size =~ /(\d+\.{0,1}\d{0,})k/i ) {
        $bytes = $1 * 1024;
    }
    elsif ( $size =~ /(\d+\.{0,1}\d{0,})t/i ) {
        $bytes = $1 * 1099511627776;
    }

    #   map { $hash{$_} = ($bytes) and $bytes/=1024} qw(bytes Kb Mb Gb Tb);
    if ($to_be_runded) {
        map { $hash{$_} = _round($bytes) and $bytes /= 1024 }
            qw(bytes Kb Mb Gb Tb);
    }
    else {
        map { $hash{$_} = ($bytes) and $bytes /= 1024 } qw(bytes Kb Mb Gb Tb);
    }
    return \%hash;
}

# =============================================================
# This is a helper function to round off the value
# =============================================================

sub _round {
    my $number = shift;
    return int( $number + .5 * ( $number <=> 0 ) );
}

=head2 find_delete_created_volumes()

find_delete_created_volumes( $vol_hash, $Node);

Deletes all the volumes as per the vol_hash passed to the function

=over

=item Arguements

=over

=item C<< $vol_hash >>

(Required) Reference to a hash having volume parameters

=item C<< $Node >>

(Required) NACL Node object

=back

=item Return Value

Boolean

=back

=cut

sub find_delete_created_volumes {
    my ( $vol_hash, $node ) = @_;
    my $result = 1;
    my $Vol_obj = new VolumeTask( command_interface => $node );
    foreach my $index (@$vol_hash) {

        if ($Vol_obj->vol_exists(
                'volume'  => $index->{'volume'},
                'vserver' => $index->{'vserver'}
            )
            )
        {
            $Vol_obj->delete_flexvol($index);
        }
        else {
            $Log->comment("Test Fail: Volume is not get deleted");
            $result = 0;
        }
    }
    return $result;
}

=head2 delete_cifs_share()

delete_cifs_share( $share_params, $node_obj);

Deletes the CIFS Shares

=over

=item Arguements

=over

=item C<< $share_params >>

(Required) CIFS share params as hash reference

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

1 on Success

Exception if Failure

=back

=cut

sub delete_cifs_share {
    my ( $share_params, $node ) = @_;
    my @cifs_share = ();

    # Delete pre-existing cifs share
    @cifs_share
        = NACL::C::VserverCifsShare->find( command_interface => $node );
    foreach my $share (@cifs_share) {
        if ( $share->state()->share_name()
            =~ m/^$share_params->{'share-name'}$/i )
        {
            $share->delete();
        }    # if ($share->state()->share_name() .. ) ends here
    }    # foreach my $share (@cifs_share) ends here

    return 1;
}

=head2 create_vol()

create_vol( $vol_params, $node_obj);

Creates a volume based on the volume parameters passed in the $vol_params
hash reference. In addition to just creating a volume, it verifies if the given
aggregate exists or not, if doesn't exists, throws
NACL::APISet::Exceptions::ResponseException. Also checks if the given volume
already exists? if exists, it deletes and then creates the volume.

=over

=item Arguements

=over

=item C<< $vol_params >>

(Required) Volume params as hash reference

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

1 on Success

Exception if Failure

=back

=cut

sub create_vol {
    my ( $vol_params, $node ) = @_;
    my $vol_obj  = undef;
    my @vol_objs = ();

    my $mode = $node->mode();

    # Check for aggregate
    my $aggr_obj = new AggregateTask( command_interface => $node );
    $aggr_obj->set_aggr_obj( aggregate => $vol_params->{'aggregate'} );
    if (!$aggr_obj->aggr_exists( 'aggregate' => $vol_params->{'aggregate'} ) )
    {
        NACL::APISet::Exceptions::ResponseException->throw(
            text => "deux_init needs to be run to configure filer"
                . " prior to running script" );
    }
    $Log->comment( "Aggregate Exists : " . $vol_params->{'aggregate'} );

    # Find all the volumes
    @vol_objs = NACL::C::Volume->find( command_interface => $node );

    # If already present, delete it
    foreach my $vol (@vol_objs) {
        if ( $vol->volume() =~ m/^$vol_params->{'volume'}$/ ) {
            $vol->unmount() if ( $mode =~ /CMode/ );
            $vol->modify( state => 'offline' );
            $vol->delete();
        }
    }

    # Create Volume Task Object
    $vol_obj = new VolumeTask( command_interface => $node );

    # Check if the aggregate exists
    $vol_obj->create_wait(%$vol_params);

    return 1;
}

=head2 aggr_bit_type()

aggr_bit_type( $aggr_name, $node_obj);

Determines the aggrgate's block-type and returns it.

=over

=item Arguements

=over

=item C<< $aggr_name >>

(Required) Aggregate name

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

Aggregate's block-type

=back

=cut

sub aggr_bit_type {
    my ( $aggr_name, $node ) = @_;
    my $aggr_obj  = undef;
    my $aggr_type = undef;

    $aggr_obj = new AggregateTask( command_interface => $node );
    $aggr_obj->set_aggr_obj( aggregate => $aggr_name );
    my $aggr_state = $aggr_obj->get_aggr_obj()->state();

    # Calculate the aggregate type  (32 bit or 64 bit)
    $aggr_state->block_type() =~ /^(\d+)-\w+$/;
    $aggr_type = $1;
    $Log->comment("Aggregate type is: $aggr_type bit");
    return $aggr_type;
}

=head2 create_share()

create_share( $share_params, $node_obj);

Creates a CIFS share based on the share parameters passed in the $share_params
hash reference. In addition to just creating a CIFS share, it modifies the
access control as, user-or-group => everyone and permission => Full_Control.

=over

=item Arguements

=over

=item C<< $share_params >>

(Required) CIFS Share params as hash reference

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

1 on Success

Exception if Failure

=back

=cut

sub create_share {
    my ( $share_params, $node ) = @_;
    my $Apiset_obj = undef;

    NACL::C::VserverCifsShare->create(
        command_interface => $node,
        %$share_params
    );
    $Apiset_obj = NACL::APISet->new(
        name      => $node->name(),
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );
    $Apiset_obj->vserver_cifs_share_access_control_modify(
        share           => $share_params->{'share-name'},
        'user-or-group' => "everyone",
        permission      => "Full_Control",
        vserver         => $share_params->{'vserver'}
    );

    return 1;
}

=head2 get_statistics()

B<<Deprecated>>. Kindly use get_counter()

=cut

sub get_statistics {
    my $category  = shift;
    my $object    = shift;
    my $counter   = shift;
    my $node      = shift;
    my $stats_obj = undef;

    $stats_obj = NACL::C::Statistics->find(
        command_interface => $node,
        filter            => {
            node     => $node->name(),
            category => $category,
            object   => $object,
            counter  => $counter
        }
    );

    my $value = $stats_obj->state()->value();
    return $value;
}

=head2 enable_quota()

enable_quota( $vserver, $volume, $node_obj);

Turns on the quota on the given volume. If the quota is already set to ON, it
turns it OFF and then turns ON.

=over

=item Arguements

=over

=item C<< $vserver >>

(Required) Vserver on which the volume exists.

=item C<< $volume >>

(Required) Volume on which the quota to be turned ON

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

1 on Success

Exception if Failure

=back

=cut

sub enable_quota {
    my $vserver = shift;
    my $volume  = shift;
    my $node    = shift;

    my $quota_obj = NACL::C::Quota->find(
        command_interface => $node,
        filter            => {
            vserver => $vserver,
            volume  => $volume
        }
    );
    if ( $quota_obj->state()->state() eq "off" ) {
        $quota_obj->modify( state => "on" );
    }
    else {
        $quota_obj->modify( state => "off" );
        my $flag            = 0;
        my $polling_endtime = time() + 100;
        while ( time() < $polling_endtime ) {
            $quota_obj = NACL::C::Quota->find(
                command_interface => $node,
                filter            => {
                    vserver => $vserver,
                    volume  => $volume
                }
            );
            if ( $quota_obj->state()->state() == "off" ) {
                $flag = 1;
                last;
            }
            sleep 3;
        }

        if ($flag) {
            WaflSpaceException->throw(
                text => "Step Fail - Failed to enable quota" );
        }
        $quota_obj->modify( state => "on" );
    }

    return 1;
}

=head2 qtree_create()

qtree_create( $volume, $vserver, $node_obj);

Creates a Qtree based on the passed Qtree, volume and vserver parameters. If
the qtree exists, it is deleted and then created again.

=over

=item Arguements

=over

=item C<< $volume >>

(Required) Volume Name

=item C<< $vserver >>

(Required) Vserver Name

=item C<< $qtree >>

(Required) Qtree Name

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

1 on Success

Exception if Failure

=back

=cut

sub qtree_create {
    my $volume  = shift;
    my $vserver = shift;
    my $qtree   = shift;
    my $node    = shift;

    my @qtree_objs = NACL::C::Qtree->find( command_interface => $node );
    foreach (@qtree_objs) {
        my $state = $_->state();
        if (   ( $state->qtree() eq $qtree )
            && ( $state->volume()  eq $volume )
            && ( $state->vserver() eq $vserver ) )
        {
            $_->delete();
        }    # if ( ($_->state()->qtree() eq $qtree) && .. ) ends here
    }    # foreach (@qtree_objs) ends here

    # Qtree create
    NACL::C::Qtree->create(
        command_interface => $node,
        vserver           => $vserver,
        volume            => $volume,
        qtree             => $qtree
    );

    return 1;
}

=head2 set_security_style()

set_security_style( $sec_style, $node_obj);

Sets wafl default security style

=over

=item Arguements

=over

=item C<< $sec_style >>

(Required) Security Style

=item C<< $node >>

(Required) NACL Node object

=back

=item Return Value

Security Style before modifying to new Security Style

=back

=cut

sub set_security_style {
    my ( $sec_style, $node ) = @_;
    my $output
        = $node->run( command => "options wafl.default_security_style" );
    my ( $name, $old_sec_style ) = split( /\s+/, $output, 2 );
    if ( $old_sec_style !~ /$sec_style/ ) {
        chomp($old_sec_style);
        my $output = $node->run(
            command => "options wafl.default_security_style $sec_style" );
    }

    return $old_sec_style;
}

=head2 list_file_attributes()

list_file_attributes( node_obj => $Node, file => $file);

Lists the attributes of the given file using testsetattr.

=over

=item Arguements

=over

=item C<< node_obj => $Node >>

(Required) Node Component Object

=item C<< file => $file >>

(Required) File name of which the attributes need to be listed.

=back

=item Return Value

Hash of file attributes.

=back

=cut

sub list_file_attributes {

    my %args          = @_;
    my $node_comp_obj = $args{'node_obj'};
    my $mode          = $node_comp_obj->mode();
    my $hostrec       = $node_comp_obj->{'hostrec'};
    my $set           = $mode eq "CMode" ? "Nodescope" : "7Mode";
    my $Api_Set_Obj   = NACL::APISet->new(
        hostobj   => $hostrec,
        category  => "Node",
        interface => "CLI",
        set       => $set
    );
    my $file = $args{'file'};

    my $response_obj = $Api_Set_Obj->testsetattr(
        'get-attributes'     => 1,
        'path'               => $file,
        'privilege-level'    => 'test',
        'connectrec-timeout' => 600
    );

    # In CMode the allocation maps to allocation_size hence normalizing the
    # name of the attribute in the return list to allocation_size for both
    # CMode and 7Mode.

    my %file_attribs = %{ $response_obj->get_parsed_output->[0] };
    return %file_attribs;
}

=head2 set_file_attributes()

set_file_attributes( param1 => $value1, param2 => $value2, ...);

This is a function to set the file attributes using testsetattr.

=over 8

=item Synopsis

set_file_attributes {
        file              => <file with path>
        node_obj          => <Node Component Object>
        size              => <size>,
        alloc_size        => <alloc_size>,
        space_reserve     => [enable | disable] [fill | overwrite | cifs],
        cifs_sparse_flag  => 1,
        spinnp_open_req   => [trunc | create],
        spinnp_close_req  => [<fd>|true],
        get_attrib        => -l,
        restore           => -r,
        read_len          => <offset> <len>,
        dirty_len         => <offset> <len>,
        snapmirror_bypass => true
}

=back

=over

=item Arguements

=over

=item C<< file => $file >>

(Required) file with path

=item C<< node_obj => $Node >>

(Required) Node Component Object.

=item C<< size => $size >>

(Optional) Node Component Object.

=item C<< alloc_size => $alloc_size >>

(Optional) Allocation size.

=item C<< space_reserve => [enable | disable] [fill | overwrite | cifs] >>

(Optional) Space Reservations.

=item C<< cifs_sparse_flag => [1 | 0] >>

(Optional) Turn on CIFS SPARSE flag.

=item C<< spinnp_open_req => [trunc | create] >>

(Optional) Open a File Session with SPINNP_OPEN. Using -o by itself simply
opens the file. -o trunc will truncate the file to 0 size. -o create will
create the file if it does not exist

=item C<< spinnp_close_req => [<fd>|true] >>

(Optional) Close Sessions with SPINNP_CLOSE or SCAN_MATCHING_LOCKS.

=item C<< get_attrib => 1 >>

(Optional) List Attribute Values.

=item C<< restore => -r >>

(Optional) -r Option indicates for restore.

=item C<< read_len => <offset> <len> >>

(Optional) Reads 'len' bytes from the file at the specified 'offset' before
doing the setattr

=item C<< dirty_len => <offset> <len> >>

(Optional) Dirties 'len' bytes in the file at the specified 'offset' before
doing the setattr

=item C<< snapmirror_bypass => <true> >>

(Optional) Indicates snapmirror bypass redirector

=back

=item Return Value

N/A

=back

=cut

sub set_file_attributes {
    my %args            = @_;
    my %options_mapping = (
        'size'              => 'file-size',
        'alloc_size'        => 'file-allocation-size',
        'space_reserve'     => 'file-space-reservations',
        'cifs_sparse_flag'  => 'set-cifs-sparse-flag',
        'spinnp_open_req'   => 'open',
        'spinnp_close_req'  => 'close',
        'get_attrib'        => 'get-attributes',
        'restore'           => 'restore',
        'read_len'          => 'read-offset-and-length',
        'dirty_len'         => 'write-offset-and-length',
        'snapmirror_bypass' => 'bypass-snapmirror'
    );
    my $file          = delete $args{'file'};
    my $node_comp_obj = delete $args{'node_obj'};
    my $mode          = $node_comp_obj->mode();
    my $hostrec       = $node_comp_obj->{'hostrec'};
    my $set           = $mode eq "CMode" ? "Nodescope" : "7Mode";
    my $Api_Set_Obj   = NACL::APISet->new(
        hostobj   => $hostrec,
        category  => "Node",
        interface => "CLI",
        set       => $set
    );

    foreach my $key ( keys %args ) {
        if ( exists $args{$key} ) {
            my $value      = delete $args{$key};
            my $mapped_key = $options_mapping{$key};
            $args{$mapped_key} = $value;
        }
    }
    $args{'path'} = $file;
    $Api_Set_Obj->set_privilege( 'privilege-level' => "test" );
    $Api_Set_Obj->testsetattr(
        %args,
        'privilege-level'    => 'test',
        'connectrec-timeout' => 600
    );

    sleep 3;
    my $cmd = "sync";
    $Api_Set_Obj->sync();

}

=head2 get_counter()

get_counter( node_obj => $Node, object => $object, counter => $counter,
instance => $instance);

Returns the value of the given counter.

=over

=item Arguements

=over

=item C<< node_obj => $Node >>

(Required) Node Component Object

=item C<< object => $object >>

(Required) Name of the Object.

=item C<< counter => $counter >>

(Required) Name of the Counter.

=item C<< category_or_instance => $category_or_instance >>

(Required) Category or instance.

=back

=item Return Value

Value of the given counter

=back

=cut

sub get_counter {
    my %args = @_;

    # Determine the ONTAP Version
    my $node        = delete $args{'node_obj'};
    my $version_mgr = $node->get_version_manager();
    my $release
        = $version_mgr->get_version_attribute( attribute => "release" );
    my $category_or_instance = delete $args{'category_or_instance'}
        if defined( $args{'category_or_instance'} );
    my $object  = delete $args{'object'};
    my $counter = delete $args{'counter'};

    my $value;

    my $mode          = $node->mode();
    my $set           = $mode eq "CMode" ? "Nodescope" : "7Mode";
    my $Api_Set_Obj   = NACL::APISet->new(
        name   => $node->name(),
        category  => "Node",
        interface => "CLI",
        set       => $set
    );

    # Need to force a CP and wait before getting counter value.
    # VSIMs take a while to update.
    $Api_Set_Obj -> sync();
    sleep 10; 
    $value = NACL::CS::Statistics->fetch(
       command_interface => $node,
            filter            => {
                node     => $node->name(),
                instance => $category_or_instance,
                object   => $object,
                counter  => $counter,
                raw      => 'true'
            }
        )->value();
    
    return $value;
}

=head2 calculate_overwrite_reserve()

calculate_overwrite_reserve( <fractional reserve>, <rsrv_overwrite_counter>);

Returns the calculated value of overwrite reserve value.

=over

=item Arguements

=over

=item C<< $FR >>

(Required) Fractional Reserve

=item C<< $rsrv_overwrite_counter >>

(Required) Value of the wvblk_rsrv_overwrite_counter counter.

=back

=item Return Value

Calculated overwrite reserve value

=back

=cut

##==========================================================================
## Function:    calculate_overwrite_reserve
##
## Description: Helper function to calculate the OR value
##
## Synopsis:
##        calculate_overwrite_reserve(<fractional reserve>,
##        <wvblk_rsrv_overwrite_counter>);
##
## Returns:     Calculated overwrite reserve value in bytes
##
##==========================================================================

sub calculate_overwrite_reserve {
    my ( $FR, $wvblk_rsrv_overwrite_counter ) = @_;
    my $OR = ( int( $FR * $wvblk_rsrv_overwrite_counter ) * 4096 );

    return $OR;
}

# =============================================================
# Exception Package
# =============================================================

package WaflSpaceException;
use base qw(NACL::APISet::Exceptions::ResponseException);

1;
