#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/WaflUtils.pm#1 $
#!/usr/local/bin/perl
## @summary Common functions used by wafl test scripts  
##          
## @description
##   Common wafliron and wafl_check functions used among wafl test scripts 
##
## 1)   wafliron_run - run wafliron from the dbladecli on a specified aggregate and return once the wafliron is complete, or timeout  
##
## 2)   wafliron_start - start wafliron (dbladecli) on a specified aggregate  
##
## 3)   wafliron_stop - stop wafliron (dbladecli) on a specified aggregate  
##
## 4)   wafliron_status -check wafliron status (dbladecli) on a specified aggregate  
##
## 5)   wafliron_boot - start wafliron from the boot menu  
##
## 6)   wafl_check_run -start wafl_check from IBM menu on a specified aggregate  
##
## 7)   get_aggr_name - get the correct aggregate name for a coral aggregate (method 1)  
##
## 8)   get_aggr_plex - get the correct aggregate name for a coral aggregate (method 2)  
##
## @status public
## @author jackm@netapp.com

## History jackm	07/04/13	changed the prompt string in wafl_check_run
## 	   jackm	07/12/13	added new prompt string for wafliron starting
##	   jackm	08/05/07	switch dblade priv set to dblade_set_privilege 
##	   jackm	08/05/15	changes for updated version of transit_to

package WaflUtils;

use strict;
use vars qw(@ISA @EXPORT);
use Tharn;

use Server;
use SayWrap qw(prompt_loop);

## @notes
## All routines use the TCD.pm return values for CONFIG, PASS and FAIL. 
## This allows the scripts to be 
## compatible in terms of error handling even if TCD is not used to
## develop the scipt. CONFIG is only used to indicate problems setting up
## the functions, such as not being able to create a Server object.
##
## Also note that the Nate logging functions are used instead of the TCD
## logging functions to maintain compatability as well.

my $TCD_CONFIG   =  -7;
my $TCD_PASS     =  1;
my $TCD_FAIL     =  0;

require Exporter;

use Data::Dumper;

@ISA = qw(Exporter);
@EXPORT = qw(
             wafliron_run wafliron_start wafliron_stop wafliron_status
             wafliron_boot wafl_check_run get_aggr_name get_aggr_plex
             );

# ------------------ API for WAFL QA internal use --------------------------

## @usage ::wafliron_run::
## Runs wafliron on an aggregate and waits for it to complete.
## Function returns a $TCD::PASS when the iron is complete, and 
## $TCD::FAIL if an error occurs.
##
##   wafliron_run($server, $filerspec, $filer_aggregate, $timeout);
##   Timeout defaults to 300 secs.
##

sub wafliron_run {
    my ($Server, $Filer, $Aggr, $timeout) = @_;
    my ($fail_cnt,
        $cmd,
        $output,
        @response,
        $found,
        $data,
        @datarray,
        @aggrs);
 
    $fail_cnt = 0;

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($Aggr)) {
    	$Aggr = param ("AGGR");
    }
    if (!defined ($timeout)) {
        # (240 iterations * 15 second sleep = 3600 seconds which is 1 hour)
	$timeout = 240; 
    }

     logcomment ("Run wafliron on $Filer against aggregate $Aggr");

    eval {
	$Server->set_execution_server(-Server => $Filer); 
    };
    if ($@) {
	logcomment ("Error creating Server object: $@");
	return $TCD_CONFIG;
   }

    # Comment: Fail if aggregate does not exist

    eval {
	@aggrs = $Server->storage_aggregate_show(aggregate=>$Aggr);
    };
    if ($@) {
        eval {
            # If you are dealing with Coral you need to check members also
	    @aggrs = $Server->storage_aggregate_member_show(aggregate=>$Aggr);
        };
        if ($@) {
	    logcomment ("Can not find aggregate: $Aggr");
	    return $TCD_CONFIG;
        }
    }

    # Comment: ("Run wafliron on aggregate $Aggr");

    eval {
        $Server->dblade_set_privilege(level=> "advanced",quiet=> "on");
    };
    if ($@) {
	logcomment ("Could not change to privilege mode advanced: $@");
	return $TCD_CONFIG;
    } else { 

        eval {
            $Server->dblade(cmd => "aggr wafliron start $Aggr");
        };
        if ($@) {
	    logcomment ("Could not start wafliron on aggregate $Aggr: $@");
	    return $TCD_CONFIG;
        } else { 

	    while ($timeout > 0) {

        	@response = $Server->dblade(cmd => "aggr wafliron status $Aggr");

                # If the response from this is that the dbladecli is Already Connected,
                # IGNORE the error, wait 15 seconds and try again, as the problem will
                # eventually go away.
                $found = grep /Already Connected/, (@response);
     	        if ($found == 1) { 
	            logcomment ("dbladecli Failure on $timeout"); 
	        } else {

		    # If not, the response from this should be one of the following:
		    # aggr wafliron status: wafliron is in the process of starting
		    # wafliron is active on aggregate: nqa_f3050_17_aggr1 \n Scanning (X% done)
		    # aggr wafliron status nqa_f3050_17_aggr1: wafliron is not currently active
		    # aggr wafliron status: no waflirons in progress

     	            if ($found == 0) { 
                        $found = grep /[Ww]afliron is in the process of starting/, (@response);
                    }
     	            if ($found == 0) { 
                        $found = grep /[Ww]afliron is active on aggregate/, (@response); 
                    }
    	            if ($found == 0) { 
                        # wafliron does not appear to be running, but check for specific
                        # messages to make sure it is really not running.
		        $found = grep /[Ww]afliron is not currently active/, (@response);
     	                if ($found == 1) {$timeout = 0;}
		        $found = grep /no waflirons in progress/, (@response);
     	                if ($found == 1) {$timeout = 0;}
     	                # If you did not find a specific message then there is some type of problem
     	                # so break out of the loop and force an error return 
     	                if ($timeout != 0) {logcomment ("wafliron state unknown"); $timeout = 0; $fail_cnt++;}
       	            } 
       	        } 
                sleep(15);
                $timeout--;
	    }
        }
    }

    return ($fail_cnt == 0) ? $TCD_PASS : $TCD_FAIL;
} 

## @usage ::wafliron_start::
##        Starts wafliron on an aggregate but DOES NOT waits for it to complete
##              Function returns a $TCD::PASS when the iron start command completes, and 
##              $TCD::FAIL if an error occurs.
##
##   wafliron_start($server, $filerspec, $filer_aggregate);
##

sub wafliron_start {
    my ($Server, $Filer, $Aggr) = @_;
    my ($fail_cnt,
        $cmd,
        $output);
 
    $fail_cnt = 0;

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($Aggr)) {
    	$Aggr = param ("AGGR");
    }

    logcomment ("Start wafliron on $Filer against aggregate $Aggr");

    eval {
	$Server->set_execution_server(-Server => $Filer); 
    };
    if ($@) {
	logcomment ("Error creating Server object: $@");
	return $TCD_CONFIG;
    }

    eval {
        $Server->dblade_set_privilege(level=> "advanced",quiet=> "on");
    };
    if ($@) {
	logcomment ("Could not change to privilege mode advanced: $@");
	return $TCD_CONFIG;
    } else { 
        eval {
            $output = $Server->dblade(cmd => "aggr wafliron start $Aggr");
	};
        if ($@) {
	    logcomment ("Could not start wafliron on aggregate $Aggr: $@");
       	    $fail_cnt++;
        } 
    }

    return ($fail_cnt == 0) ? $TCD_PASS : $TCD_FAIL;
} 

## @usage ::wafliron_stop::
##        Stops wafliron on an aggregate 
##              Function returns a $TCD::PASS when the stop command completes, and 
##              $TCD::FAIL if an error occurs.
##
##   wafliron_stop($server, $filerspec, $filer_aggregate);
##

sub wafliron_stop {
    my ($Server, $Filer, $Aggr) = @_;
    my ($fail_cnt,
        $output);
 
     $fail_cnt = 0;

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($Aggr)) {
    	$Aggr = param ("AGGR");
    }

    logcomment ("Stop wafliron on $Filer against aggregate $Aggr");

    eval {
	$Server->set_execution_server(-Server => $Filer); 
    };
    if ($@) {
	logcomment ("Error creating Server object: $@");
	return $TCD_CONFIG;
    }
    eval {
        $Server->dblade_set_privilege(level=> "advanced",quiet=> "on");
    };
    if ($@) {
	logcomment ("Could not change to privilege mode advanced: $@");
	return $TCD_CONFIG;
    } else { 

        eval {
            $output = $Server->dblade(cmd => "aggr wafliron stop $Aggr");
        };
        if ($@) {
	    logcomment ("Could not stop wafliron on aggregate $Aggr: $@");
       	    $fail_cnt++;
        } else {
            sleep (10); # Give it some time to actually stop
        }
    }
    sleep (10); # Give it some time to actually
    return ($fail_cnt == 0) ? $TCD_PASS : $TCD_FAIL;
} 

## @usage ::wafliron_status::
##        Checks wafliron status on an aggregate 
##              Function returns a $TCD::PASS when the wafliron is ACTIVE,  
##              $TCD::FAIL if it is NOT ACTIVE, and dies if an error occurs.
##
##   wafliron_status($server, $filerspec, $filer_aggregate);
##

sub wafliron_status {
    my ($Server, $Filer, $Aggr) = @_;
    my ($cmd,
        $fail_cnt,
        $output,
        $data,
        @datarray,
        @aggrs,
        @active,
        @inactive,
        $timeout,
        $state,
        $done);
 
     $fail_cnt = 0;

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($Aggr)) {
    	$Aggr = param ("AGGR");
    }

    logcomment ("Check wafliron status on $Filer, for aggregate $Aggr if specified");

    eval {
	$Server->set_execution_server(-Server => $Filer); 
    };
    if ($@) {
	logcomment ("Error creating Server object: $@");
	return $TCD_CONFIG;
    }

    if (defined($Aggr)) {
    	eval {
		@aggrs = $Server->storage_aggregate_show(aggregate=>$Aggr);
    	};
    	if ($@) {
        eval {
            # If you are dealing with Coral you need to check members also
	    @aggrs = $Server->storage_aggregate_member_show(aggregate=>$Aggr);
        };
        if ($@) {
	    logcomment ("Can not find aggregate: $Aggr");
	    return $TCD_CONFIG;
        }
    	}
	logcomment ("Check wafliron status for $Aggr on $Filer");
    } else {
        $Aggr = "";
       	logcomment ("Check wafliron status for all aggregates on $Filer");
    }
    eval {
        $Server->dblade_set_privilege(level=> "advanced",quiet=> "on");
    };
    if ($@) {
	logcomment ("Could not change to privilege mode advanced: $@");
	return $TCD_CONFIG;
    } else { 
        eval {
            $output = $Server->dblade(cmd => "aggr wafliron status $Aggr");
        };
        if ($@) {
	    logcomment ("Could not check wafliron status: $@");
	    return $TCD_CONFIG;
        } else { 

	    # The response from this should be one of the following:
	    
	    # aggr wafliron status nqa_f3050_17_aggr1: wafliron is not currently active
	    # aggr wafliron status: no waflirons in progress
	    # wafliron is active on aggregate: fee_fi_fo_fum \n Scanning (X% done)
	    # ANYTHING else is assumed to be wafliron is active on aggregate
	    
	    @datarray = split("\n", $output);
	    @inactive = grep(/wafliron is not currently active/, @datarray);
	    @active = grep(/"wafliron is active"/, @datarray);
	    
	    # Comment: ("Output: $output :");
	    if (scalar(@active) > 0) { 
		$state = "wafliron is active: $timeout";
		return $TCD_FAIL;
	    } else {
     	        if (scalar(@inactive) > 0) { 
		    $state = "wafliron not active: $timeout";
		    return $TCD_PASS;
     	        } else {
		    $state = "wafliron state unknown: $timeout \n @datarray";
		    return $TCD_FAIL;
       	        }
	    }
        }
    }
    
    return ($fail_cnt == 0) ? $TCD_PASS : $TCD_FAIL;
} 

## @usage ::wafliron_boot::
##       Starts wafliron on all aggregates from the boot menu (IBM) 
##             Function ALWAYS returns a $TCD::PASS, unless an error occurs while
##             trying to reach the IBM state, trying to start wafliron or trying
##	       to get to the login prompt. The function leaves you at the LOGIN
##	       prompt, and does NOT wait for wafliron completion. You should use 
##	       wafliron_status with no aggregate name to do that. 
##
##   wafliron_boot($server, $filerspec, $timeout);
##

sub wafliron_boot {
    my ($Server, $Filer, $timeout) = @_;
    my ($Test,
        $fail_cnt,
        $delay,
        $state,
        $done,
        $output,
        $cmd,
        $rtn,
        @datarray,
        @active,
        @inactive,
        $console);

    $fail_cnt = 0;
    $delay = 10;

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($timeout)) {
    	$timeout = 300;
    }

    logcomment ("Start wafliron on $Filer against ALL aggregates");

    eval {
	$Server->set_execution_server(-Server => $Filer); 
    };
    if ($@) {
	logcomment("ERROR: Error creating Server object: $@");
	return $TCD_CONFIG;
   }

    # $Test->nlog("Transit To IBM");

    eval {
        $Server->transit_to(state => $Server->IBM());
    };
    if ($@) {
	logcomment("ERROR on transit_to IBM: $@");
       	$fail_cnt++;
    }
    sleep ($delay);
    
    eval {
    $console = $Server->get_connection(-Server => $Filer, type => 'console_connection');
    };
    if ($@) {
	logcomment("ERROR on get connection: $@");
	return $TCD_CONFIG;
    }

    # Starting wafliron from the boot menu causes the system to restart
    # with all aggregates being ironed in parallel. You can then log in and run dbladecli
    # aggr wafliron status to see if they are finished. You can also change the wafl scan
    # speed to help it along.

    $cmd = "wafliron";

    eval {
	say($console, $cmd, -blind);
    };
    sleep ($delay);

    # $Test->nlog("Transit To CLI via LOGIN");

    eval {
        $Server->transit_to(state => $Server->CLI(), wait_for => $Server->LOGIN());
    };
    if ($@) {
	logcomment("ERROR on transit_to CLI via LOGIN: $@");
       	$fail_cnt++;
    }
    sleep ($delay);

    return ($fail_cnt == 0) ? $TCD_PASS : $TCD_FAIL;
}

## @usage ::wafl_check_run::
##       Starts wafl_check from the boot menu (IBM) and waits for it to complete
##              Function returns a $TCD::PASS when the check is complete, and 
##              $TCD::FAIL if an error occurs.
##
##   wafl_check_run($server, $filerspec, $timeout);
##

sub wafl_check_run {
    my ($Server, $Filer, $Aggr, $timeout) = @_;
    my ($fail_cnt,
        $delay,
        $console,
        $cmd,
        $prompt,
        $rtn,
        @datarray,
        $match, $before, $after, $prompt);
 
    $fail_cnt = 0;
    $delay = 10;

    logcomment ("Start wafl_check on $Filer against specified aggregate");

    if (!defined($Filer)) {
    	$Filer = param ("FILER1");
    }
    if (!defined($Aggr)) {
    	logcomment ("NO aggregate defined for WAFL_check");
	return $TCD_CONFIG;
    }
    if (!defined($timeout)) {
    	$timeout = 300;
    }

    eval {
	$Server->set_execution_server(-Server => $Filer); 
	# I have no idea where I got this from as it is not a server method
	# $Server->node_reboot(node => $Filer,'inhibit-takeover' => 'true');
    };
    if ($@) {
	logcomment("ERROR on get connection: $@");
	return $TCD_CONFIG;
    }

    # Transit To IBM
	#logcomment("TRANSITING TO CFE");
    #eval {
    	#$Server->transit_to(state => $Server->CFE(),
	#		    wait_for => $Server->CFE());
    #};
	logcomment("TRANSITING TO IBM");
    eval {
    	$Server->transit_to(state => $Server->IBM());
    };

    $console = $Server->get_connection(-Server => $Filer, 
				type => 'console_connection');

    sleep ($delay); # On general principal, wait a couple of seconds  

    # Starting wafl_check from the boot menu causes the system to restart.
    # Once the system is up it will run wafl_check on the specified aggregate.
    # If no aggregate is defined the list of available aggregates is specified
    # and you can select one. Once you selct one with all aggregates being 
    # ironed in parallel. You can then log in and run dbladecli aggr wafliron 
    # status to see if they are finished. You can also change the wafl scan
    # speed to help it along.

     $cmd = "wafl_check $Aggr";
     
    eval {
	    
		say($console, "wafl_check $Aggr", '-blind');
		prompt_loop(connection => $console,
		    table => [
			      "Continue\\?\\?" => "yes",
			      ],
		    done => [
			     "Press any key to reboot",
			     "Press Enter to reboot system",
			     ],
		    max_tries => 5,
		    say_opts => ['-timeout' => 600]);
     };
     if ($@) {
 		logcomment("ERROR running wafl_check on $Aggr: $@");
	 }
     

    sleep ($delay);

    # Transit To CLI -  HOPEFULLY this will pick up that the system is rebooting and
    #                   wait until it is back up before it returns
    eval {
        $Server->transit_to(state => $Server->CLI());
    };
    if ($@) {
        logcomment("ERROR on transit_to CLI: $@");
        $fail_cnt++;
    }
    return ($fail_cnt == 0) ? $TCD::PASS : $TCD::FAIL;
} 

## @usage ::get_aggr_name::
##              When using an aggregate name in the dblade the Coral aggregates
##              need to have the correct suffix (usually _XXX) added to the name
##              the aggregate is created with. For example, nqa_f6070_5_aggr2 is
##		actually nqa_f6070_5_aggr2_000. 
##		The way to make sure you are using the correct name is to run
##		"storage aggregate member show ${aggregate}*" and see if it 
##		returns anything. If it does not then the aggregate is flex and
##		you can use the name as is. If it does then the list it returns
##		is the list of coral aggregates. If you create the aggregate with
##		only 1 node listed then you only get 1 aggregate_000, but if you
##		create it with more than 1 node (and you can use same node name 
##		more than once) you will get _000, _001, ...
##
##   $aggr_plex = get_aggr_name($server, $filer, $aggregate_root);
##   Returns an array of names if the aggregate exists (for flex it is 1 name long)
##   Returns null array if no values found for aggregate name or if an error occurs
##   

sub get_aggr_name {
    my ($Server, $Filer, $Aggr) = @_;
    my (@output,
        $data,
        @datarray);
 
    #logcomment ("Determine name value based on Aggregate Name");
    @datarray = ();

   
    if (!defined($Aggr)) {
	logcomment ("Aggregate Root Required");
	return @datarray;
    }

    eval {
	@output = $Server->storage_aggregate_member_show(aggregate=>"$Aggr"."*");
    };
    if ($@) {		# if there were no members
    	eval {
	    @output = $Server->storage_aggregate_show(aggregate=>"$Aggr");
    	};
    	if ($@) {
	    logcomment ("Error on storage_aggregate_member_show $Aggr: $@");
	    return @datarray;
    	}
	# else # in this cases means it is probably a non-Coral aggregate
    }

    my $size = @output;
    #logcomment ("$size aggregate references for $Aggr");
    for (my $i = 0; $i < scalar @output; $i++) {
       push(@datarray, $output[$i]{aggregate});
       #logcomment ("${i}: @datarray[$i]");
    }

    return @datarray;
} 

## @usage ::get_aggr_plex::
##              When using an aggregate name in the dblade the Coral aggregates
##              need to have the correct suffix (usually _000) added to the name
##              the aggregate is created with. For example, nqa_f6070_5_aggr2 is
##		actually nqa_f6070_5_aggr2_000. The way to get this is to do an
##		aggregate show -instance and parse the name in the Plexes: field.
##   
##   $aggr_plex = get_aggr_plex($server, $filer, $aggregate_root);
##   Returns null if no Plexex values found for Aggregate name or
##   if an error occurs
##


sub get_aggr_plex {
    my ($Server, $Filer, $Aggr) = @_;
    my (@output,
        $data,
        @datarray);
 
    #logcomment ("Determine Plex value based on  Aggregate Name");
    $data = "";
   
    if (!defined($Aggr)) {
	logcomment ("Aggregate Root Required");
	return $data;
    }

#    We will assume the correct Filer is already set
#    eval {
#	$Server->set_execution_server(-Server => $Filer); 
#    };

    eval {
	@output = $Server->storage_aggregate_show(aggregate=>$Aggr);
    };
    if ($@) {
	logcomment ("Can not find aggregate $Aggr: $@");
	return $data;
    }

    @datarray = split("/",$output[0]{plexes});
    $data = @datarray[1];
    logcomment ("Plex value for $Aggr: $data");

    return $data;
} 

