# Hey emacs, this is basically -*- perl -*-
#
# $Id$

#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Module that provides methods for wafliron test cases
## @author dl-rakesh-autmation@netapp.com
## @status shared

## TODO:
##
## [Burt 424262] Once the ApiSet of retrieving /mroot/etc/mlog/messages.log is
## ready, have to catch the message for wafliron from starting to ending.
##
## [Burt 430345] Once the Component is supported is added to the 'aggregate'
## component, need to add the function:  delete_striped_aggregate_member

=head1 NAME

WaflironLib

=head1 DESCRIPTION

This library provides functions to test various WAFL related test scenarios.

=head1 SYNOPSIS

use WaflironLib;

=head1 DEPENDENCY

This library uses the following libraries.

1. NACL::C::Aggregate;

2. NACL::C::Volume;

3. NACL::C::Wafliron;

4. NACL::C::WaflironMember;

5. NACL::C::VolumeMember;

6. NACL::C::Node;

7. NACL::C::Snapmirror;

8. NACL::APISet::Exceptions::CommandFailedException qw(:try);

9. NACL::APISet::Exceptions::ResponseException qw(:try);

10. FS::EA::EventlogTask;

11. FS::EA::VolumeTask;

12. FS::EA::SnapmirrorTask;

=head1 KNOWN ISSUES

=head1 BUGS

=head1 Methods

=cut

package WaflironLib;

use strict;
use warnings;

use FS::EA::EventlogTask;
use FS::EA::VolumeTask;
use FS::EA::SnapmirrorTask;
use WaflironLibrary::AccessSyslog;
use FS::EA::Exceptions::WaflironException;
use NACL::Transit;

require Exporter;

use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF BOOLEAN UNDEF OBJECT);
use vars qw (@ISA @EXPORT_OK);
use Tharn qw($Log sleep host subtest);
our $SUDO_PATH = 'sudo';

##############################################################################
##
##  Components used.
##
##############################################################################

use NACL::C::Aggregate;
use NACL::C::Volume;
use NACL::C::Wafliron;
use NACL::C::WaflironMember;
use NACL::CS::WaflironMember;
use NACL::CS::Volume;
use NACL::C::VolumeMember;
use NACL::C::Node;
use NACL::C::Snapmirror;
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::Mkfile;
use constant BLOCK_SIZE => 4096;

@ISA = qw(Exporter);

@EXPORT_OK = qw(initialize_volume
    log_step
    log_check_point
    wafliron_finalize
    unmount_hammer
    wafliron_on_mirror_finalize
    create_vol_and_populate_data
    populate_percent_data
    convert_size
    get_vol_flag
    set_vol_flag
    run_hammer
    purge_vol
    shrink_root_vol
    get_root_object
    check_aggregate_members_exists
    start_hammer
    stop_hammer
    is_licensed
    root_vol_recovery
    create_striped_aggregate_member
    create_LS_mirror
    create_DP_mirror
    get_vol_move_job_id
    get_event_logs
    access_wafl_scan_speed
    get_wafl_flags
    set_wafl_flags
    wafliron_cluster_commit_reject
    populate_volume_with_snapshots
    wafliron_node_start
    wafliron_cluster_start
    wafliron_node_start_boot_menu
);

##===========================================================================
## Function:    check_aggregate_members_exists
##
## Description: This is a function to get the count of the member aggregates
##              existing.
##
## Synopsys:    check_aggregate_members_exists(<command_interface>, <count>,
##              <aggregate>)
##
## Returns:     Boolean
##
##
##===========================================================================

=pod

=head2 check_aggregate_members_exists ()

This is a function to get the count of the existing member aggregates .

=over

=item Synopsis

check_aggregate_members_exists(<command_interface>, <count>, <aggregate>);

=item Arguments

=over

=item C<< command_interface=>$command_interface >>

=item C<< count=>$count >>

=item C<< aggregate=>$aggr >>

=back

=item Return Value

=over

Boolean

=back

=back

=cut

sub check_aggregate_members_exists {
    $Log->enter();
    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        count             => { type => SCALAR },
        aggregate         => { type => SCALAR },
        };

    my @member_aggregate;
    my @all_member_aggregate = NACL::C::AggregateMember->find(
        'command_interface' => $opts{command_interface}

            #filter              => { aggregate => $opts{aggregate} . "*" }
    );

    foreach (@all_member_aggregate) {
        my $pattern        = $opts{aggregate} . '_' . '*';
        my $curr_aggr_name = $_->state()->aggregate();

        if ( $curr_aggr_name =~ /$pattern/ ) {

            push( @member_aggregate, $_ );

        }

    } ## end foreach (@all_member_aggregate)

    if ( ( $#member_aggregate + 1 ) < $opts{count} ) {
        return 1;
    }
    return 0;
} ## end sub check_aggregate_members_exists

############################################################################
## Name: log_step()
##
## Description:
## Method to print the log messages in a formated output
##
## Synopsis:
##          log_step(<"log message">);
##
##
## Returns: N/A
##
#############################################################################

sub log_step {
    my $string = shift;
    my $border = <<END;
==============================================================================
END
    $string =~ s/ {2,}/ /g;
    $Log->comment($border);
    $Log->comment( "Step: " . $string );
    $Log->comment($border);
} ## end sub log_step

############################################################################
## Name: wafliron_finalize()
##
## Description:
## This method is the main method which will be exposed. This method is used
## to run the Wafliron Finalize testcase.
## This method, based on the type of the aggregate provided, calls one of the
## sub functions to execute the steps.
##
## Synopsis:
##    wafliron_finalize(
##        command_interface => $node,
##        aggregate         => 'stripedaggr_0',
##        vserver           => 'vs0',
##        volume            => 'striped_vol1',
##        idle_time         => 60,
##        files_to_validate => $files,
##        node              => 'chethanm-vsim1'
##    );
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##           <vserver>            : Vserver
##           <volume>             : Volume in which the test files exists.
##           <idle_time>          : Counter to wait for a wafliron state
##                                  Total Wait Cycle = <idle_time> * 5 Secs
##           <files_to_validate>  : Array reference containing the list of
##                                  files to be tested for existence
##           <node>               : Node Name
##           <client>             : Client Host name
##
## Returns: N/A
##
#############################################################################

=head2 wafliron_finalize ()

This is a function which covers the whole Wafliron test case WAFL022-029-007
10M_wafliron finalize.

I<WAFL022-029-007> is been written as a function, as this was used as setup
in many of the WAFL test cases. This test case uses many helper functions like,
_wafliron_finalize_striped_vol() and _wafliron_finalize_flex_vol() which are
not actually exposed by this library as they are not generic.

For details of the steps this script is executing, kindly refer QC for the
test case I<WAFL022-029-007>

=over

=item Synopsis

    wafliron_finalize(
        command_interface => $node,
        aggregate         => 'stripedaggr_0',
        vserver           => 'vs0',
        volume            => 'striped_vol1',
        idle_time         => 60,
        files_to_validate => $files,
        node              => 'chethanm-vsim1'
    );

=item Arguments

=over

=item C<< command_interface=>$command_interface >>

=item C<< aggregate=>$aggregate >>

=item C<< vserver=>$vserver >>

=item C<< volume=>$volume >>

=item C<< idle_time=>$idle_time >>

=item C<< files_to_validate=>\[file1 file2 ...] >>

=item C<< node=>$node >>

=item C<< client=>$client >>

=back

=item Return Value

=over

Throws an FS::EA::Exceptions::WaflironException if failed at any stage.

=back

=back

=cut

sub wafliron_finalize {

    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        aggregate         => { type => SCALAR },
        volume            => { type => SCALAR },
        vserver           => { type => SCALAR },
        idle_time         => { type => SCALAR, default => 60 },
        files_to_validate => { type => ARRAYREF },
        node              => { type => SCALAR },
        client            => { type => SCALAR },
        time_stamp        => { type => SCALAR, optional => 1 }
        };

    my @aggr = NACL::C::Aggregate->find(
        command_interface => $opts{'command_interface'},
        filter            => { 'aggregate' => $opts{'aggregate'} }
    );

    if ( $aggr[0]->state()->voltype() =~ /striped/ ) {
        _wafliron_finalize_striped_vol(%opts);
    } else {
        _wafliron_finalize_flex_vol(%opts);
    }
} ## end sub wafliron_finalize

############################################################################
## Name: _wafliron_finalize_striped_vol()
##
## Description:
## This method is a subfunction to the main method. This method is called for
## the striped aggregate
##
## Synopsis:
##    _wafliron_finalize_striped_vol(
##        command_interface => $node,
##        aggregate         => 'stripedaggr_0',
##        vserver           => 'vs0',
##        volume            => 'striped_vol1',
##        idle_time         => 60,
##        files_to_validate => $files,
##        node              => 'chethanm-vsim1'
##    );
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##           <vserver>            : Vserver
##           <volume>             : Volume in which the test files exists.
##           <idle_time>          : Counter to wait for a wafliron state
##                                  Total Wait Cycle = <idle_time> * 5 Secs
##           <files_to_validate>  : Array reference containing the list of
##                                  files to be tested for existence
##           <node>               : Node Name
##
## Returns: N/A
##
#############################################################################

sub _wafliron_finalize_striped_vol {
    my %args = @_;

    # Step 1
    #        Validates 'Current wafliron status is on "scanning" for
    #        striped/member aggregates

    $Log->comment("Validating current wafliron status");
    log_step( "Validates Current wafliron status is on \"scanning\" for "
            . "striped/member aggregates" );

    my $wafl_state = _get_wafliron_status( $args{'command_interface'},
        $args{'aggregate'} );

    if (( $wafl_state->iron_status() =~ /check_members|not_running/ )
        && ( $wafl_state->iron_last_start_error_info()
            =~ /At least one member aggregate has some wafliron information/ )
        )
    {
        $Log->comment(
            "Wafliron Status for $args{'aggregate'} is \"check_members or not
            running\""
        );
        $Log->comment(
            "Wafliron last start error info for $args{'aggregate'} is:
            \"At least one member aggregate has some wafliron information\""
        );
    } else {
        FS::EA::Exceptions::WaflironException->throw(
            text => "Wafliron status for  $args{'aggregate'} is not
                \"check_members\" or not_running"
        );
    } ## end else [ if ( ( $wafl_state->iron_status...

    # Step 2
    #       'Once the wafliron is done, the 'show" command returns the
    #       "not_running" on wafliron current status with 0 error.

    $Log->comment("Validating current wafliron status");

    log_step(
        "Once the wafliron is done, the \"show\" command returns the
              \"not_running\" on wafliron current status with 0 error."
    );

    #   Wait till the wafliron finishes or timeout happens
    $wafl_state = _get_wafliron_status( $args{'command_interface'},
        $args{'aggregate'} );

    my $endtime = time() + $args{'idle_time'};
    while (
        ( $wafl_state->iron_status() !~ /not_running/ )
        && ( $wafl_state->iron_last_start_errno()
            !~ /At least one member aggregate has some wafliron information/ )
        )
    {
        sleep 5;

        # Display wafliron member status in the logs
        $Log->comment(
            "Wafliron Member Status for the all the member aggregates..");
        NACL::CS::WaflironMember->fetch(
            'command_interface' => $args{'command_interface'} );

        if ( time() > $endtime ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Wafliron did not completed its scan  for "
                    . "$args{'aggregate'}" );
        }
        $wafl_state = _get_wafliron_status( $args{'command_interface'},
            $args{'aggregate'} );
    } ## end while ( ( $wafl_state->iron_status...

    $Log->comment(
        "Wafliron has completed the scan on " . $args{'aggregate'} );
    $Log->comment( "Checking the wafliron status on member aggregates of "
            . $args{'aggregate'} );
    my @mem_aggrs = _get_member_aggregates( $args{'command_interface'},
        $args{'aggregate'} );

    foreach my $aggr_mem (@mem_aggrs) {
        $Log->comment( "Checking the wafliron status on member aggregates "
                . $aggr_mem->aggregate() );
        my $aggr_mem_state
            = _get_wafliron_member_status( $args{'command_interface'},
            $aggr_mem->aggregate() );
        if (   ( $aggr_mem_state->iron_status() !~ /not_running/ )
            || ( $aggr_mem_state->iron_last_start_errno() != 0 ) )
        {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Wafliron did not completed its scan  for member "
                    . "aggregate $aggr_mem->aggregate()" );
        } ## end if ( ( $aggr_mem_state...
    } ## end foreach my $aggr_mem (@mem_aggrs)

    sleep 30;
    log_step( "Check event log and show the summary as list after finishing "
            . "wafliron" );
    my $wafl_start_time = $args{'time_stamp'}
        || _get_wafl_start_time( $args{'command_interface'} );
    my $wafliron_messages
        = _get_all_wafl_messages( $args{'command_interface'},
        $wafl_start_time );

    log_step(
        "Verifying the Event Log for wafliron messages on member Aggregates");

    @mem_aggrs = _get_member_aggregates( $args{'command_interface'},
        $args{'aggregate'} );

    foreach my $m_aggr (@mem_aggrs) {
        $Log->comment( "Verifying the wafl status for the member aggregate"
                . $m_aggr->aggregate() );
        _check_event_log(
            'name'     => $m_aggr->aggregate(),
            'type'     => 'aggregate',
            'messages' => $wafliron_messages
        );
    } ## end foreach my $m_aggr (@mem_aggrs)
    $Log->comment( "Verified the Event Log for wafliron messages on member"
            . " Aggregates " );
    $Log->comment(
        "Verifying the event log for wafliron messages on member volumes ");

    my @mem_vols = NACL::C::VolumeMember->find(
        command_interface => $args{'command_interface'},
        filter            => {
            'volume'  => $args{'volume'} . '*',
            'vserver' => $args{'vserver'}
        }
    );

    foreach my $m_vol (@mem_vols) {
        $Log->comment( "Verifying the wafl status for the member volume"
                . $m_vol->volume() );
        _check_event_log(
            'name'     => $m_vol->volume(),
            'type'     => 'volume',
            'messages' => $wafliron_messages
        );
    } ## end foreach my $m_vol (@mem_vols)

    $Log->comment(
        "Listing all the event log messages" . $wafliron_messages );
    $Log->comment("Wafliron finished on aggregate and member aggregates!");
    NACL::C::WaflironMember->find(
        command_interface => $args{'command_interface'} );

    # Step 3
    #           check member aggregate and volume state for striped/member
    #           aggregates
    #

    log_step(
        "check member aggregate and volume state for striped/member
            aggregates"
    );

    #   Check volume and aggregates are online after wafliron run

    $Log->comment(
        "Verifying if the volume: " . $args{'volume'} . " is online" );
    _verify_volume_online( $args{'command_interface'},
        $args{'volume'}, $args{'vserver'} );
    $Log->comment("Pass: Volume state verifyed");

    $Log->comment(
        "Verifying if the aggregate: " . $args{'aggregate'} . " is online" );
    _verify_aggregate_online( $args{'command_interface'},
        $args{'aggregate'} );
    $Log->comment("Pass: Aggregate state verifyed");

    #   Check all the members under the striped aggregates are online after
    #   wafliron run

    $Log->comment( "Verifying if all the member aggregates of "
            . $args{'aggregate'}
            . " are online" );
    _check_aggregate_members_online( $args{'command_interface'},
        $args{'aggregate'} );
    $Log->comment("Pass: Aggregate Member's state verifyed");

    #   Check all the volume members under the striped aggregates are online
    #   after wafliron run

    $Log->comment( "Verifying if all the member volumes of "
            . $args{'volume'}
            . " are online" );
    _check_volume_members_online( $args{'command_interface'},
        $args{'volume'}, $args{'vserver'} );
    $Log->comment("Pass: Aggregate Member's state verifyed");

    # Step 4
    #              Check if the files are available

    log_step("Check if the files are available");
    _validate_files(
        $args{'volume'},
        $args{'files_to_validate'},
        $args{'command_interface'},
        $args{client}, $args{vserver}
    );

} ## end sub _wafliron_finalize_striped_vol

############################################################################
## Name: _wafliron_finalize_flex_vol()
##
## Description:
## This method is a subfunction to the main method. This method is called for
## the normal aggregate
##
## Synopsis:
##    _wafliron_finalize_flex_vol(
##        command_interface => $node,
##        aggregate         => 'test_aggr1',
##        vserver           => 'vs0',
##        volume            => 'test_vol1',
##        idle_time         => 60,
##        files_to_validate => $files,
##        node              => 'chethanm-vsim1'
##    );
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##           <vserver>            : Vserver
##           <volume>             : Volume in which the test files exists.
##           <idle_time>          : Wait time in seconds
##           <files_to_validate>  : Array reference containing the list of
##                                  files to be tested for existence
##           <node>               : Node Name
##
## Returns: N/A
##
#############################################################################

sub _wafliron_finalize_flex_vol {
    my %args = @_;

    # Step 1
    #              Validates 'Current wafliron status is on "scanning" for
    #              regular aggregates.

    $Log->comment("Validating current wafliron status");
    log_step(
        "Validates Current wafliron status is on \"scanning\" for
               regular aggregates."
    );

    my $wafl_state = _get_wafliron_status( $args{'command_interface'},
        $args{'aggregate'} );

    if ( $wafl_state->iron_status() =~ /scanning/ ) {
        $Log->comment(
            "Wafliron Status for $args{'aggregate'} is \"Scanning\"");
    } else {
        FS::EA::Exceptions::WaflironException->throw( text =>
                "Wafliron status for  $args{'aggregate'} is not \"scanning\""
        );
    }

    # Step 2
    #              'Once the wafliron is done, the 'show" command returns the
    #              "not_running" on wafliron current status with 0 error.

    log_step(
        "Once the wafliron is done, the \"show\" command returns the \
            \"not_running\" on wafliron current status with 0 error."
    );

    my $endtime = time() + $args{'idle_time'};
    while (
        _get_wafliron_status(
            $args{'command_interface'}, $args{'aggregate'}
        )->iron_status() !~ /not_running/
        )
    {
        sleep 5;
        if ( time() > $endtime ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Wafliron Status for $args{'aggregate'} is still "
                    . _get_wafliron_status( $args{'command_interface'},
                    $args{'aggregate'} )->iron_status()
            );
        } ## end if ( time() > $endtime)
    } ## end while ( _get_wafliron_status...

    # Verify event log messages after the wafliron is finished

    log_step( "Check event log and show the summary as list after finishing "
            . "wafliron" );
    my $wafl_start_time = _get_wafl_start_time( $args{'command_interface'} );
    my $wafliron_messages
        = _get_all_wafl_messages( $args{'command_interface'},
        $wafl_start_time );

    log_step("Verifying the Event Log for wafliron messages on Aggregate");

    _check_event_log(
        'name'     => $args{'aggregate'},
        'type'     => 'aggregate',
        'messages' => $wafliron_messages
    );

    $Log->comment(
        "Verified the Event Log for wafliron messages on" . " Aggregate " );
    $Log->comment("Verifying the event log for wafliron messages on volume");

    _check_event_log(
        'name'     => $args{'volume'},
        'type'     => 'volume',
        'messages' => $wafliron_messages
    );
    $Log->comment("Listing all the event log messages");
    $Log->debug("$wafliron_messages");

    # Step 3
    #              Check Aggregate and Volume State

    log_step("Check Aggregate and Volume State");
    _verify_volume_online( $args{'command_interface'},
        $args{'volume'}, $args{'vserver'} );
    $Log->comment("Volume $args{'volume'} is online after wafliron run");

    _verify_aggregate_online( $args{'command_interface'},
        $args{'aggregate'}, $args{'vserver'} );
    $Log->comment(
        "Aggregate $args{'aggregate'} is online after wafliron run");

    # Step 4
    #              Check if the files are available

    log_step("Check if the files are available");
    _validate_files(
        $args{'volume'},
        $args{'files_to_validate'},
        $args{'command_interface'},
        $args{client}, $args{vserver}
    );
} ## end sub _wafliron_finalize_flex_vol

############################################################################
## Name: wafliron_on_mirror_finalize()
##
## Description:
## This method is the main method which will be exposed. This method is used
## to run the Wafliron on mirror finalize testcase.
## This method, based on the type of the aggregate provided, calls one of the
## sub functions to execute the steps.
##
## Synopsis:
##    wafliron_on_mirror_finalize(
##        command_interface => $node,
##        aggregate         => 'test_aggr1',
##        vserver           => 'vs0',
##        volume            => 'test_vol1',
##        idle_time         => 60,
##        data_file         => $files,
##        node              => 'chethanm-vsim1',
##        source_path       => 'Cluster://vs0/source_vol',
##        type              => 'DP',
##        data_file         => 'file1',
##        destination_volume        => 'dest_vol',
##        destination_cluster       => 'Cluster',
##        destination_vserver       => 'vs0',
##        wafliron_include_mirror   => 'true',
##        syslog_obj                => \$syslog_obj
##    );
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##           <vserver>            : Vserver
##           <volume>             : Volume on which wafliron runs
##           <idle_time>          : Counter to wait for a wafliron state
##                                  Total Wait Cycle = <idle_time> * 5 Secs
##           <source_path>        : Snapmorror Source Path
##           <type>               : Relationship Type
##           <node>               : Node Name
##           <messagename>        : Event log message names
##           <data_file>          : File on source volume for validating data
##           <destination_volume> : Destination Volume
##           <destination_cluster>: Destination Cluster
##           <destination_vserver>: Destination Vserver
##           <wafliron_include_mirror> : Force checking on mirrors
##                                       flag {true|false}
##           <syslog_obj>         : Object refernce
##
## Returns: N/A
##
#############################################################################

=head2 wafliron_on_mirror_finalize ()

This is a function which covers the whole Wafliron test case WAFL022-029-007
10M_wafliron on-mirror finalize.

I<WAFL022-029-007 10M_wafliron on mirror finalize> is been written as a
function, as this was used as setup in many of the WAFL test cases. This test
case uses many helper function _wafliron_on_mirror_finalize_flex_vol()
which are not actually exposed by this library as they are not generic.

For details of the steps this script is executing, kindly refer QC for the
test case I<WAFL022-029-007>

=over

=item Synopsis

	wafliron_on_mirror_finalize(
		command_interface       => $node,
		aggregate               => 'test_aggr1',
		vserver                 => vs0,
		volume                  => 'test_vol1',
		idle_time               => 60,
		data_file               => $files,
		node                    => wqa-gx71,
		source_path             => Cluster://vs0/source_vol,
		type                    => DP,
		data_file               => file1,
		destination_volume      => dest_vol,
		destination_cluster     => Cluster,
		destination_vserver     => vs0,
		wafliron_include_mirror => true,
        syslog_obj              => \$syslog_obj
	);

=item Arguments

=over

=item C<< command_interface=>$command_interface >>

=item C<< command_interface=>$command_interface >>

=item C<< aggregate=>$aggregate >>

=item C<< vserver=>$vserver >>

=item C<< volume=>$volume >>

=item C<< idle_time=>$idle_time >>

=item C<< data_file=>$data_file >>

=item C<< node=>$node >>

=item C<< source_path=>$source_path >>

=item C<< type=>$type >>

=item C<< data_file=>$data_file >>

=item C<< destination_volume=>$destination_volume >>

=item C<< destination_cluster=>$destination_cluster >>

=item C<< destination_vserver=>$destination_vserver >>

=item C<< wafliron_include_mirror=>$wafliron_include_mirror >>

=item C<< syslog_obj=>\$syslog_obj >>

=back

=item Return Value

=over

Throws an FS::EA::Exceptions::WaflironException if failed at any stage.

=back

=back

=cut

sub wafliron_on_mirror_finalize {

    my %opts = validate @_,
        {
        command_interface       => { isa  => 'NACL::C::CommandInterface' },
        aggregate               => { type => SCALAR },
        time_stamp              => { type => SCALAR, optional => 1 },
        vserver                 => { type => SCALAR },
        messagename             => { type => SCALAR, optional => 1 },
        idle_time               => { type => SCALAR, default => 60 },
        source_path             => { type => SCALAR },
        destination_volume      => { type => SCALAR },
        destination_vserver     => { type => SCALAR },
        destination_cluster     => { type => SCALAR },
        wafliron_include_mirror => { type => SCALAR },
        snapmirror_type         => { type => SCALAR },
        client                  => { type => SCALAR },
        data_file               => { type => SCALAR },
        syslog_obj              => { type => HASHREF, optional => 1 }
        };

    my @aggr = NACL::C::Aggregate->find(
        command_interface => $opts{'command_interface'},
        filter            => { 'aggregate' => $opts{'aggregate'} }
    );
    my $aggr_type
        = ( $aggr[0]->state()->voltype() =~ /striped/ )
        ? 'striped'
        : 'flex';
    if ( $aggr_type =~ /flex/ ) {
        _wafliron_on_mirror_finalize_flex_vol(%opts);
    } else {
        $Log->comment(
                  "Wafliron on mirror finalize is not supported for striped"
                . " volume" );
    }

} ## end sub wafliron_on_mirror_finalize

############################################################################
## Name: _wafliron_on_mirror_finalize_flex_vol()
##
## Description:
## This method is the sub function, which will be called by
## wafliron_on_mirror_finalize for flex volumes.
##
## Synopsis:
##    Finalize::wafliron_on_mirror_finalize(
##        command_interface => $node,
##        aggregate         => 'test_aggr1',
##        vserver           => 'vs0',
##        volume            => 'test_vol1',
##        idle_time         => 60,
##        data_file         => $files,
##        node              => 'chethanm-vsim1',
##        source_path       => 'Cluster://vs0/source_vol',
##        type              => 'DP',
##        data_file         => 'file1',
##        destination_volume        => 'dest_vol',
##        destination_cluster       => 'Cluster',
##        destination_vserver       => 'vs0',
##        wafliron_include_mirror   => 'true',
##        syslog_obj                => \$syslog_obj
##    );
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##           <vserver>            : Vserver
##           <volume>             : Volume on which wafliron runs
##           <idle_time>          : Counter to wait for a wafliron state
##                                  Total Wait Cycle = <idle_time> * 5 Secs
##           <source_path>        : Snapmorror Source Path
##           <type>               : Relationship Type
##           <node>               : Node Name
##           <data_file>          : File on source volume for validating data
##           <destination_volume> : Destination Volume
##           <destination_cluster>: Destination Cluster
##           <destination_vserver>: Destination Vserver
##           <wafliron_include_mirror> : Force checking on mirrors
##                                       flag {true|false}
##           <syslog_obj>         : Object refernce
##
## Returns: N/A
##
#############################################################################

sub _wafliron_on_mirror_finalize_flex_vol {

    my %args = @_;
    $Log->comment("Validating current wafliron status");

    # Step 1
    #              Keep checking the wafliron status till it finishes
    #

    log_step("Keep checking the wafliron status till it finishes");

    my $endtime = time() + $args{'idle_time'};
    while (
        _get_wafliron_status(
            $args{'command_interface'}, $args{'aggregate'}
        )->iron_status() !~ /not_running/
        )
    {
        sleep 5;
        if ( time() > $endtime ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Wafliron Status for $args{'aggregate'} is still "
                    . _get_wafliron_status( $args{'command_interface'},
                    $args{'aggregate'} )->iron_status()
            );
        } ## end if ( time() > $endtime)
    } ## end while ( _get_wafliron_status...
    log_check_point(
        "Check Passed: Wafliron completed on aggr $args{'aggregate'}");
    sleep 60;

    # Step 2
    #              Check event log wafliron message
    #

    my $match;
    my $log_msgs;
    if ( defined $args{messagename} ) {
        my $elog_obj = EventlogTask->new(
            command_interface => $args{'command_interface'} );

        $args{'messagename'} =~ s/\s+//g;
        my @messages = split( /,/, $args{'messagename'} );
        my $captured_log = undef;

        foreach (@messages) {
            $captured_log = undef;

            # Introducing sleep so that event messages gets generated
            # before capturing particular event.
            sleep 30;
            $captured_log = $elog_obj->capture_all_event_logs(
                messagename => $_,
                init_time   => $args{'time_stamp'}
            );
            if ( defined $captured_log ) {
                $match .= $captured_log;
            }
        } ## end foreach (@messages)
    } elsif ( defined $args{syslog_obj} ) {
        foreach ( keys %{ $args{syslog_obj} } ) {
            $match .= $args{syslog_obj}->{$_}->syslog_tail();
        }
    } else {
        log_check_point( "Warning : Either messagename or syslog_obj needs"
                . " to be defined" );
    }

    if ( $match !~ m[(Aggregate $args{'aggregate'}, wafliron completed)]i ) {
        FS::EA::Exceptions::WaflironException->throw(
            text => "Event log is not showing wafliron "
                . " complete msgs on aggregate" );
    }
    $log_msgs .= $1 . "\n";

    my @volumes = NACL::C::Volume->find(
        command_interface => $args{'command_interface'},
        filter            => {
            vserver   => $args{'vserver'},
            aggregate => $args{'aggregate'}
        }
    );

    log_check_point( "This is mroot message list:\n" . $match );
    foreach (@volumes) {
        my $vol_name = $_->volume();
        if (   $_->volume() eq $args{'destination_volume'}
            && $args{'wafliron_include_mirror'} eq 'false' )
        {
            if ( $match !~ m[(Read only.*?$vol_name.*?not checked)]i ) {
                FS::EA::Exceptions::WaflironException->throw(
                    text => "Mroot log is not showing vol skipped mesg" );
            }
            $log_msgs .= $1 . "\n";
        } elsif ( $_->volume() eq $args{'destination_volume'}
            && $args{'wafliron_include_mirror'} eq 'true'
            && $args{'snapmirror_type'} =~ m[ls]i )
        {
            if ( $match !~ m[(Read only.*?$vol_name.*?not checked)]i ) {
                FS::EA::Exceptions::WaflironException->throw(
                    text => "Mroot log is not showing vol skipped mesg" );
            }
            $log_msgs .= $1 . "\n";
        } else {
            if ( $match !~ m[($vol_name.*?wafliron\s+completed\.)]i ) {
                $Log->warn(
                    "Mroot log is not showing wafliron completed msg on $vol_name"
                );
            } else {
                $log_msgs .= $1 . "\n";
            }
        } ## end else [ if ( $_->volume() eq $args...
    } ## end foreach (@volumes)
    log_check_point( "Below mroot msgs are verified:\n" . $log_msgs );

    log_check_point("Check Passed: Mroot msg log check is successful!");

    # Step 3
    #              Check aggregates and volumes status
    #

    log_step("Check Aggregate and Volume State");

    foreach (@volumes) {
        if ( $_->state()->state() !~ m[online]i ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Volume state is not online ($_->volume())" );
        }
    } ## end foreach (@volumes)
    log_check_point("Check Passed: All the volumes are in online state");

    _verify_aggregate_online( $args{'command_interface'},
        $args{'aggregate'}, $args{'vserver'} );
    log_check_point( "Check Passed: Aggregate $args{'aggregate'}"
            . " is online after wafliron run" );

    # Step 4
    #           Check Snapmirror status, if Broken-off resync it
    #

    log_step("Check Snapmirror status");

    my $destination_path
        = $args{'destination_cluster'} . '://'
        . $args{'destination_vserver'} . '/'
        . $args{'destination_volume'};
    my $snapmirror_state = (
        NACL::C::Snapmirror->find(
            command_interface => $args{'command_interface'},
            filter            => { 'destination-path' => $destination_path }
        )
    )->state();

    if ( $snapmirror_state->state() !~ /Snapmirrored/ ) {
        NACL::C::Snapmirror->resync(
            'command_interface'   => $args{'command_interface'},
            'destination-cluster' => $args{'destination_cluster'},
            'destination-vserver' => $args{'destination_vserver'},
            'destination-volume'  => $args{'destination_volume'}
        );
    } ## end if ( $snapmirror_state...
    _wait_for_snapmirror_status( $args{'command_interface'},
        $destination_path, $args{'idle_time'} );
    log_check_point("Check Passed: Snapmirror state is Smapmirrored");

    # Step 5
    #           Verify data
    #

    log_step("Varifying Data.. ");

    my $source_vol = ( split( /\//, $args{'source_path'} ) )[-1];

    my $source_file_content = NACL::C::Node->run(
        command_interface => $args{'command_interface'},
        node              => $args{'command_interface'}->name(),
        command           => "rdfile /vol/$source_vol/$args{'data_file'}"
    );
    my $dest_file_content = NACL::C::Node->run(
        command_interface => $args{'command_interface'},
        node              => $args{'command_interface'}->name(),
        command =>
            "rdfile /vol/$args{'destination_volume'}/$args{'data_file'}"
    );

    $source_file_content =~ s/system node run .* rdfile \/vol.*$//m;
    $dest_file_content   =~ s/system node run .* rdfile \/vol.*$//m;
    $source_file_content =~ s/\n$//m;
    $source_file_content =~ s/^\n//m;
    $dest_file_content   =~ s/\n$//m;
    $dest_file_content   =~ s/^\n//m;

    $Log->comment( "Source content :" . $source_file_content );
    $Log->comment( "Dest content :" . $dest_file_content );

    if ( $source_file_content ne $dest_file_content ) {
        FS::EA::Exceptions::WaflironException->throw(
                  text => "Content of file on source volume : "
                . $source_vol
                . " is not matching with content of the file on destination"
                . " volume: "
                . $args{'destination_volume'} );
    } ## end if ( $source_file_content...
    log_check_point("Check Passed: Data varification is done succesfully!");
} ## end sub _wafliron_on_mirror_finalize_flex_vol

############################################################################
#
#                             HELPER FUNCTIONS
#
############################################################################

############################################################################
## Name: _wait_for_snapmirror_status()
##
## Description:
## Method to wait for the Snapmirror status - "Snapmirrored".
## Waits still the snapmirror to finish. If snapmirror doesn't finishes
## within the given wait_time, throws as exception.
##
## Synopsis:
##         _wait_for_snapmirror_status(<command_interface>,
##                                     <source_path>,
##                                     <wait_time>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <destination_path>   : Destination path as given for the snapmirror
##                                  initialize.
##           <wait_time>          : Wait counter for expected status
##                                  Wait cycle = <wait_time> * 5 Secs
##
## Note:     This function is deprecated in WaflironLib as it has been moved
##           to FS::EA::SnapmirrorTask as "wait_for_snapmirror_to_finish".
##
#############################################################################
sub _wait_for_snapmirror_status {
    my ( $command_interface, $destination_path, $wait_time ) = @_;

    my $snapmirror_task_obj
        = SnapmirrorTask->new( 'command_interface' => $command_interface );

    $snapmirror_task_obj->wait_for_snapmirror_to_finish(
        'destination-path' => $destination_path,
        'wait_time'        => $wait_time
    );
    return 1;
} ## end sub _wait_for_snapmirror_status

############################################################################
## Name: _get_wafliron_status()
##
## Description:
## Method to get the wafliron status for the given aggregate
##
## Synopsis:
## $wafliron_obj = _get_wafliron_status(<command_interface>,<aggregate_name>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate on which wafliron
##                                  is run.
##
## Returns: Wafliron Status Object
##
#############################################################################

sub _get_wafliron_status {
    my ( $command_interface, $aggregate ) = @_;
    my $wafliron_obj = NACL::C::Wafliron->find(
        command_interface => $command_interface,
        filter            => { 'aggregate' => $aggregate }
    );
    return $wafliron_obj->state();
} ## end sub _get_wafliron_status

############################################################################
## Name: _get_wafliron_member_status()
##
## Description:
## Method to get the wafliron status for the given member aggregate
##
## Synopsis:
## $wafliron_obj = _get_wafliron_member_status(<command_interface>,
##                 <aggregate_name>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Member Aggregate on which
##                                  wafliron is run.
##
## Returns: Wafliron Member Status Object
##
#############################################################################

sub _get_wafliron_member_status {
    my ( $command_interface, $aggregate ) = @_;
    my @wafliron_objs = NACL::C::WaflironMember->find(
        command_interface => $command_interface );
    foreach my $wafl_mem_obj (@wafliron_objs) {
        if ( $wafl_mem_obj->aggregate() =~ /$aggregate/ ) {
            return $wafl_mem_obj->state();
        }
    } ## end foreach my $wafl_mem_obj (@wafliron_objs)
    return 0;
} ## end sub _get_wafliron_member_status

############################################################################
## Name: _get_member_aggregates()
##
## Description:
## Method to get all the member aggregates under given aggregate
##
## Synopsis:
## $wafliron_obj = _get_member_aggregates(<command_interface>,
##                 <aggregate_name>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate of which member
##                                  aggregates need to be found.
##
## Returns:
##        1. Array of Member Aggregate Status Objects if multiple member
##           aggregates are found.
##        2. Returns scalar value if single member aggregate is found.
##
##
#############################################################################

sub _get_member_aggregates {
    my ( $command_interface, $aggregate ) = @_;
    my @aggr_members  = ();
    my @wafliron_objs = NACL::C::AggregateMember->find(
        command_interface => $command_interface );
    $aggregate .= "_";

    foreach (@wafliron_objs) {
        if ( $_->{'aggregate'} =~ m[$aggregate\d+] ) {
            push @aggr_members, $_;
        }
    }
    return wantarray ? @wafliron_objs : $wafliron_objs[0];
} ## end sub _get_member_aggregates

############################################################################
## Name: _verify_volume_online()
##
## Description:
## Method to check if the given volume is online else raises a Wafliron
## Exception
##
## Synopsis:
## $wafliron_obj = _verify_volume_online(<command_interface>,
##                 <volume>, <vserver>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <volume>             : Name of the volume of which member
##                                  volumes need to be checked.
##           <vserver>            : Vserver to which volume belongs to.
##
## Returns: Exception if volume is not online
##
#############################################################################

sub _verify_volume_online {
    my ( $command_interface, $volume, $vserver ) = @_;
    my $vol_obj = NACL::C::Volume->find(
        command_interface => $command_interface,
        filter            => { 'volume' => $volume, 'vserver' => $vserver }
    );

    if ( $vol_obj->state()->state() !~ /online/ ) {
        $Log->comment("Volume $volume is not online after wafliron run.");
        FS::EA::Exceptions::WaflironException->throw(
            text => "Volume status for  $volume is: "
                . $vol_obj->state()->state() );
    } ## end if ( $vol_obj->state()...
} ## end sub _verify_volume_online

############################################################################
## Name: _verify_aggregate_online ()
##
## Description:
## Method to check if the given aggregate is online else raises a Wafliron
## Exception
##
## Synopsis:
## $wafliron_obj = _verify_aggregate_online (<command_interface>,
##                 <aggregate>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate
##
## Returns: N/A
##
#############################################################################

sub _verify_aggregate_online {
    my ( $command_interface, $aggregate ) = @_;
    my $aggr_obj = NACL::C::Aggregate->find(
        command_interface => $command_interface,
        filter            => { 'aggregate' => $aggregate }
    );

    if ( $aggr_obj->state()->state() !~ /online/ ) {
        $Log->comment(
            "Aggregate $aggregate is not online after wafliron run.");
        FS::EA::Exceptions::WaflironException->throw(
            text => "Aggregate status for  $aggregate is: "
                . $aggr_obj->state()->state() );
    } ## end if ( $aggr_obj->state(...
} ## end sub _verify_aggregate_online

############################################################################
## Name: _check_aggregate_members_online()
##
## Description:
## Method to check if the member aggregates of given aggregate are online.
## If not online raises a Wafliron Exception.
##
## Synopsis:
## $wafliron_obj = _check_aggregate_members_online(<command_interface>,
##                 <aggregate>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <aggregate>          : Name of the Aggregate
##
## Returns: N/A
##
#############################################################################

sub _check_aggregate_members_online {
    my ( $command_interface, $aggregate ) = @_;

    my @mem_aggrs = _get_member_aggregates( $command_interface, $aggregate );

    foreach my $aggr_mem (@mem_aggrs) {

        if ( $aggr_mem->state()->state() =~ /online/ ) {
            $Log->comment(
                "Aggregate $aggregate is online after wafliron run.");
        } else {
            FS::EA::Exceptions::WaflironException->throw(
                text => "Aggregate status for  $aggregate is: "
                    . $aggr_mem->state()->state() );
        }
    } ## end foreach my $aggr_mem (@mem_aggrs)
} ## end sub _check_aggregate_members_online

############################################################################
## Name: _check_volume_members_online()
##
## Description:
## Method to check if the member volumes of given volume are online.
## If not online raises a Wafliron Exception.
##
## Synopsis:
## $wafliron_obj = _check_volume_members_online(<command_interface>,
##                 <volume>, <vserver>);
##
## Arguments:
##           <command_interface>  : Command Interface Object
##           <volume>             : Name of the Volume
##           <vserver>            : Vserver to which volume belongs to.
##
## Returns: N/A
##
#############################################################################

sub _check_volume_members_online {
    my ( $command_interface, $volume, $vserver ) = @_;
    my @vol_objs = NACL::C::VolumeMember->find(
        command_interface => $command_interface,
        filter            => { 'vserver' => $vserver }
    );

    foreach my $vol_obj (@vol_objs) {
        my $vol_state = $vol_obj->state();
        next if ( $vol_state->volume() !~ /$volume/ );
        $Log->comment( "Verifying if the member volume "
                . $vol_state->volume()
                . " is online" );

        if ( $vol_state->state() =~ /online/ ) {
            $Log->comment( "Pass: Member volume "
                    . $vol_state->volume()
                    . " is online" );
        } else {
            $Log->comment(
                "Member Volume: $vol_state->volume() is not online after "
                    . "wafliron run." );
            FS::EA::Exceptions::WaflironException->throw(
                text => "Volume status for  $vol_state->volume() is: "
                    . $vol_state->state()->state() );
        } ## end else [ if ( $vol_state->state...
    } ## end foreach my $vol_obj (@vol_objs)
} ## end sub _check_volume_members_online

############################################################################
## Name: _validate_files()
##
## Description:
## Method to check if the given files exist in the given volume
## If files are not found, raises a Wafliron Exception.
##
## Synopsis:
## $wafliron_obj = _validate_files(<volume>, <files_array_ref>, <node_obj>);
##
## Arguments:
##           <volume>             : Name of the Volume
##           <files_array_ref>    : Reference to an array containing list of
##                                  files
##           <node_obj>           : Node Object
##           <client>             : Client host name
##           <vserver>            : Vserver name
##
## Returns: N/A
##
#############################################################################

sub _validate_files {
    my ( $vol, $files_aref, $node_obj, $client, $vserver ) = @_;
    $Log->comment("Validating files...");

    my $Volume_task = VolumeTask->new( command_interface => $node_obj );
    my $files = $Volume_task->get_all_files(
        client  => $client,
        volume  => $vol,
        vserver => $vserver
    );

    foreach my $s_file ( @{$files_aref} ) {
        if ( !( grep /$s_file/, @{$files} ) ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => $s_file . " : Not found after wafliron run!" );
        }
    } ## end foreach my $s_file ( @{$files_aref...

    $Log->comment("All the files are validated and found that they exist!");

} ## end sub _validate_files

############################################################################
## Name: log_check_point()
##
## Description:
## Method to print the log in the format
##
## Synopsis:
##          log_check_point(<"log message">);
##
##
## Returns: N/A
##
#############################################################################
sub log_check_point {
    my $string = shift;
    my $border = <<END;
-----------------------------------------------------------------------------
END
    $string =~ s/ {2,}/ /g;
    $Log->comment($border);
    $Log->comment($string);
    $Log->comment($border);
} ## end sub log_check_point

############################################################
## Name: run_hammer()
##
## Description:
##		Method does
##		1) Create a new mount point (vol_name+pid).
##		2) Mount volume on client.
##		3) run hammer (runs in foreground).
##		4) Create a sample data file.
##		5) unmount
##		6) delete the mount point
##
## Synopsis:
##
##	run_hammer(
##		'command_interface' => $node,
##		'volume_params'     => $inputs->{'Volume'}->[0],
##		client              => $CLIENT,
##		'hammer_params'     => { 'nofilercore' => '', 'runtime' => '10' }
##	);
##
## Arguments:
##           <command_interface>  : MANDATORY - command interface to execute
##                                  command.
##           <volume_params>      : MANDATORY - parameters for volume.
##           <client>             : MANDATORY - Name of client.
##			 <hammer_params>	  : MANDATORY - parametrs for hammer.
##
## Returns: N/A
## Notes:  Hammer runs in foreground.
##
############################################################

sub run_hammer {
    $Log->enter();
    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        volume_params     => { type => HASHREF },
        client            => { type => SCALAR },
        hammer_params     => { type => HASHREF, optional => 1 },
        };
    my $vol_obj = NACL::C::Volume->find(
        'command_interface' => $opts{'command_interface'},
        filter              => {
            volume  => $opts{'volume_params'}->{'volume'},
            vserver => $opts{'volume_params'}->{'vserver'}
        }
    );
    my $volume_state = $vol_obj->state();

    if (   $volume_state->junction_active() ne 'true'
        || $volume_state->junction_path() !~ /\w+/ )
    {
        FS::EA::Exceptions::WaflironException->throw( text =>
                "Check 'junction-active and junction-path of the given volume"
        );
    } ## end if ( $volume_state->junction_active...
    my $Api_Set_Obj = NACL::APISet->new(
        name      => $opts{'client'},
        category  => "Host",
        interface => "CLI",
        connid    => 'telnet',
        set       => _get_client_conntype( $opts{'client'} ),
    );

    my $mount_point = "/t/$opts{'client'}/$opts{'volume_params'}->{'volume'}";
    _remove_stale_mount( $Api_Set_Obj, $mount_point );

    $opts{'hammer_params'}->{dir} = $mount_point;
    $Api_Set_Obj->mkdir( 'make-parents' => 1, 'paths' => $mount_point );
    my $response = $Api_Set_Obj->mount(
        'source-path' => $opts{'command_interface'}->{'hostrec'}->{'DATA_IP1'}
            . ":/$opts{'volume_params'}->{volume}",
        'mount-point' => $mount_point,
        "privilege-level" => "root"
    );
    $Api_Set_Obj->hammer( %{ $opts{'hammer_params'} } );

    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/echo Test data > $mount_point/test1" );
    $Api_Set_Obj->umount(
	'paths' => $mount_point,
	'force' => 'true',
	'lazy'  => 'true',
	'privilege-level' => 'root'
	);
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/rmdir $mount_point" );
    $Log->exit();
    return $mount_point;

} ## end sub run_hammer

##===========================================================================
## Function:    unmount_hammer
##
## Description: This is a function to unmount the volume and delete the
##              mount point/dir
##
## Synopsys:   unmount_hammer(<Client>, <mount dir>);
##
## Returns:    Exception on failure
##
##
##===========================================================================

sub unmount_hammer {
    $Log->enter();
    my %opts = validate @_,
        {
        'client'      => { type => SCALAR },
        'mount_point' => { type => SCALAR }
        };
    my $Api_Set_Obj = NACL::APISet->new(
        name      => $opts{'client'},
        category  => "Host",
        interface => "CLI",
        set       => "Linux",
        connid    => 'telnet',
        conntype  => ''
    );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/umount -lf $opts{'mount_point'}" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/rmdir $opts{'mount_point'}" );

} ## end sub unmount_hammer
############################################################
## Name: start_hammer()
##
## Description:
##              Method does
##              1) Create a new mount point (vol_name+pid).
##              2) Mount volume on client.
##              3) start hammer (runs in background).
##              4) Create sample data files.
##
## Synopsis:
##          start_hammer('command_interface' => $node, 'volume' => 'vol1',
##          'client_api_obj' => $Apiset_client );
##
## Arguments:
##           <command_interface>  : MANDATORY - Command interface
##           <volume>             : MANDATORY - Volume to run hammer.
##           <vserver>            : MANDATORY - Vserver of the volume
##           <client_api_obj>     : MANDATORY - Client apiset object.
##
## Returns: (mount_point, hammer_pid)
## Note   : Hammer runs in background.
##
############################################################

sub start_hammer {
    $Log->enter();
    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        volume            => { type => SCALAR },
        vserver           => { type => SCALAR },
        client_api_obj    => { isa  => 'NACL::APISet' }
        };

    my $Api_Set_Obj = $opts{'client_api_obj'};
    $Api_Set_Obj->set_privilege("privilege-level" => "root", 'answer-password' => 1);
    my $vol_obj     = NACL::C::Volume->find(
        'command_interface' => $opts{'command_interface'},
        filter              => {
            volume  => $opts{'volume'},
            vserver => $opts{'vserver'}
        }
    );
    my $volume_state = $vol_obj->state();
    my $mount_name   = "/tmp/" . $opts{'volume'} . $$;
    $Api_Set_Obj->mkdir( 'make-parents' => 1, 'paths' => $mount_name );
    my $response = $Api_Set_Obj->mount(
        'source-path' => $opts{'command_interface'}->{'hostrec'}->{'DATA_IP1'}
            . ":/$opts{'volume'}",
        'mount-point' => $mount_name
    );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/echo Test data1 > $mount_name/test1 &" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/echo Test data2 > $mount_name/test2 &" );
    my $pid = $Api_Set_Obj->execute_raw_command( 'command' =>
            "hammer $mount_name -filercore default_lib.dmp -bg -fill &" );
    $pid =~ /\[[0-9]+\] ([0-9]+)/;
    return ( $mount_name, $1 );
    $Log->exit();
}
############################################################
## Name: stop_hammer()
##
## Description:
##              Method does
##              1) Kills an hammer.
##              2) Unmounts the mount directory.
##              3) Deletes the mount directory.
##
## Synopsis:
##          stop_hammer( 'mount_dir' => '/ tmp / temp ', ' pid ' => 2332,
##          'client_api_obj' => $Apiset_client );
##
## Arguments:
##           <mount_dir>      : MANDATORY - Directory where the volume
##                                          is mounted
##           <client_api_obj> : MANDATORY - Client apiset object.
##           <pid>            : MANDATORY - Process id of the hammer
##
## Returns: N/A
##
############################################################

sub stop_hammer {
    $Log->enter();
    my %opts = validate @_,
        {
        mount_dir      => { type => SCALAR },
        pid            => { type => SCALAR },
        client_api_obj => { isa  => 'NACL::APISet' }
        };

    my $Api_Set_Obj = $opts{'client_api_obj'};
    $Api_Set_Obj->execute_raw_command(
        'command' => "sudo /bin/kill -9 $opts{'pid'}" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "sudo /bin/umount -lf $opts{'mount_dir'}" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "sudo /bin/rm -rf $opts{'mount_dir'}" );

    $Log->exit();
} ## end sub stop_hammer

############################################################
## Name: populate_percent_data()
##
## Description:
##		Method does
##		1) Create a new mount point (vol_name+pid).
##		2) Mount volume on client.
##		3) run dd comand.
##		4) Create a sample data file.
##		5) unmount
##		6) delete the mount point
##
## Synopsis:
##
##	populate_percent_data(
##		'command_interface' => $node,
##		'volume_params'     => $inputs->{'Volume'}->[0],
##		client              => $CLIENT,
##		'file_count'        => 5,
##		percentage          => "50%"
##	);
##
## Arguments:
##           <command_interface>  : MANDATORY - command interface to execute
##                                  command.
##           <volume_params>      : MANDATORY - parameters for volume.
##           <client>             : MANDATORY - Name of client.
##			 <file_count>		  : MANDATORY - No of files to be created.
##			 <percentage>		  : MANDATORY - Percentage of volume to be
##                                  filled.
##           <trace_log>          : OPTIONAL  - Avoid logging the command
##                                  output
##
## Returns: N/A
##
############################################################

sub populate_percent_data {

    $Log->enter();
    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        volume_params     => { type => HASHREF },
        connection_type   => { type => SCALAR, default => 'ssh-persist' },
        client            => { type => SCALAR },
        file_count        => { type => SCALAR, optional => 1 },
        percentage        => { type => SCALAR },
        trace_log         => { type => SCALAR, optional => 1 },
        write_command     => { type => SCALAR, optional => 1 },
        file_size         => { type => SCALAR, optional => 1 },
        };
    my ( $Api_Set_Obj, $percentage, $mount_name, $vol_obj, $vol_size,
        $file_size, $vol_percent_used, $percent_to_be_filled );

    $vol_obj = NACL::C::Volume->find(
        'command_interface' => $opts{'command_interface'},
        filter              => {
            volume  => $opts{'volume_params'}->{'volume'},
            vserver => $opts{'volume_params'}->{'vserver'}
        }
    );
    my $volume_state = $vol_obj->state();

    if (   $volume_state->junction_active() ne 'true'
        || $volume_state->junction_path() !~ /\w+/ )
    {
        FS::EA::Exceptions::WaflironException->throw( text =>
                "Check 'junction-active and junction-path of the given volume"
        );
    } ## end if ( $volume_state->junction_active...
    $vol_size         = $volume_state->size();
    $vol_percent_used = $volume_state->percent_used();
    $vol_percent_used =~ s/\%//;
    $percentage = $opts{'percentage'};
    $percentage =~ s/%//;
    $percent_to_be_filled = $percentage - $vol_percent_used;
    $percent_to_be_filled = 1 if (( $percent_to_be_filled < 0 )
                 && ( $vol_percent_used < 90 ));

    if ( $percent_to_be_filled < 0 ) {
        FS::EA::Exceptions::WaflironException->throw(
            text => "Volume is already $vol_percent_used % filled." );
    }
    $Api_Set_Obj = NACL::APISet->new(
        name      => $opts{'client'},
        category  => "Host",
        interface => "CLI",
        connid    => $opts{"connection_type"},
        set       => _get_client_conntype( $opts{'client'} ),
    );
	$Api_Set_Obj->set_privilege("privilege-level" => "root", 'answer-password' => 1);

    my $mount_point = "/t/$opts{'client'}/$opts{'volume_params'}->{'volume'}";
    _remove_stale_mount( $Api_Set_Obj, $mount_point );
    $Api_Set_Obj->mkdir( 'make-parents' => 1, 'paths' => $mount_point );
    $Api_Set_Obj->mount(
        'source-path' => $opts{'command_interface'}->{'hostrec'}->{'DATA_IP1'}
            . ":/$opts{'volume_params'}->{'junction-path'}",
        'mount-point' => $mount_point
    );

    $Log->comment("Vol size : $vol_size\n");
    $Api_Set_Obj->set_timeout( 'connectrec-timeout' => 600 );
    if ( defined $opts{'write_command'}
        && $opts{'write_command'} eq 'mkfile' )
    {
        my $size_to_fill
            = int( ( $volume_state->total() / 100 ) * $percent_to_be_filled );
        my $file_count = int(
            $size_to_fill / convert_size( $opts{'file_size'} )->{'bytes'} );
        $file_size = $opts{'file_size'};
        for ( my $i = 0; $i < $file_count; $i++ ) {
            $Api_Set_Obj->execute_raw_command( 'command' =>
                    "/usr/local/bin/mkfile $file_size $mount_point/$i" );
        }
    } else {
        $file_size = int( ( $vol_size / ( $opts{'file_count'} * BLOCK_SIZE ) )
            * ( $percent_to_be_filled / 100 ) );
        $Log->comment("File size : $file_size\n");

        for ( my $i = 0; $i < $opts{'file_count'}; $i++ ) {
            if ( defined $opts{'trace_log'} ) {
                $Api_Set_Obj->execute_raw_command(
                    'command' => "/bin/dd if=/dev/zero of=$mount_point/$i"
                        . " bs=4096 count=$file_size",
                    "output_log_level" => $opts{'trace_log'}
                );
            } else {
                $Api_Set_Obj->execute_raw_command(
                    'command' => "/bin/dd if=/dev/zero of=$mount_point/$i"
                        . " bs=4096 count=$file_size" );
            }
        } ## end for ( my $i = 0; $i < $opts...
    } ## end else [ if ( defined $opts{'write_command'...
    $Api_Set_Obj->clear_timeout();
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/echo Test data > $mount_point/test1" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/umount $mount_point" );
    $Api_Set_Obj->execute_raw_command(
        'command' => "/bin/rmdir $mount_point" );
    $Log->exit();
    return 1;
} ## end sub populate_percent_data

=over

=item populate_volume_with_snapshots()

This function fills given % full of data on a volume and creates given number
of snapshots while populating data. Volume which needs to be filled has to have
a junction-path and junction-active enabled as it needs to be mounted on the
client. Client and Filer host specs will be taken from the instance of this
class and hence need not be passed.

Following is the process flow this function works,

=over

=over

=item Remove the stale mounts if any and create the mount path

=item Mount the volume

=item Populate the volume to given % using unix dd command.

=item While the data is being populated, create several [as given] snapshots.

=item List the snapshots created

=item Validate the volume is filled to given %.

=back

=back

=over

=item Options

=over

=item C<< "api_set_obj_Client" => $client_api_set >>

(Required) The Clinet APISet Object

=item C<< node=>$command_interface >>

(Required) Node object or command_interface

=item C<< "volume" => $volume >>

(Required) The name of the volume to be populated

=item C<< "vserver" => $vserver >>

(Required) The name of vserver on which the volume resides

=item C<< "percent_to_fill" => $percent_to_fill >>

(Required) Percentage of the volume to be filled.

=item C<< "file_count" => $file_count >>

(Optional) (default => 50) Number of files to create while populating the data

=item C<< "no_of_snapshots" => $no_of_snapshots >>

(Optional) (default => 3) Number of snapshots to create while populating the data

=back

=item Return Value

Returns exception on any failure.

=item Dependency

This function uses the helper function _remove_stale_mount() to remove the
stale mount if any.

=item Author

Chethan Kumar M

=back

=back

=cut

sub populate_volume_with_snapshots {
    $Log->enter();
    $Log->comment("In WaflironLib:populate_volume_with_snapshots");

    # Declare the variables
    my ($vol_state,                    $size_to_fill,
        $blocks_to_fill,               $client,
        $block_count,                  $junction_path,
        $mount_ip,                     $mount_point,
        %opts,                         $percent,
        $percentage,                   $percent_to_fill,
        $percent_used,                 $size,
        $subtest_handle,               $write_files,
        $percent_used_after_data_fill, $resp_obj,
        $err_string,
    );

    # Validate the arguments
    # api_set_obj_Client is to maintain backward compatability
    my %Input = validate_with(
        params => \@_,
        spec   => {
            'api_set_obj_Client' => { isa => 'NACL::APISet', optional => 1 },
            'node'    => { isa  => 'NACL::C::CommandInterface::ONTAP' },
            'volume'  => { type => SCALAR },
            'vserver' => { type => SCALAR },
            'percent_to_fill' => { type => SCALAR },
            'file_count'      => { type => SCALAR, default => 50 },
            'write_command'   => { type => SCALAR, optional => 1 },
            'no_of_snapshots' => { type => SCALAR, default => 3 }

        },
        allow_extra => 1
    );
	$Input{'api_set_obj_Client'}->set_privilege(
                         'privilege-level' => "root",
                         'answer-password' => 1)
                 if defined $Input{'api_set_obj_Client'};

    # Instantiate a VolumeTask object
    $Input{'vol_task_obj'}
        = new VolumeTask( command_interface => $Input{'node'} );

    # Set volume object in volume task object
    $Input{'vol_task_obj'}->set_vol_obj(
        'volume'  => $Input{'volume'},
        'vserver' => $Input{'vserver'}
    );
    $vol_state = $Input{'vol_task_obj'}->{'vol_obj'}->state();

    # Validate the junction_active and junction_path of the volume
    if (   $vol_state->junction_active() ne 'true'
        || $vol_state->junction_path() !~ /\w+/ )
    {
        NATE::BaseException->throw(
            "Check 'junction-active and junction-path of the given volume");
    } ## end if ( $vol_state->junction_active...

    # Calculate the size to fill
    $size         = $vol_state->size();
    $percent_used = $vol_state->percent_used();
    $percent_used =~ s/\%//;
    $percent = $Input{'percent_to_fill'};
    $percent =~ s/%//;
    $percent_to_fill = ($percent - $percent_used);

    if ( $percent_to_fill < 0 ) {
        NATE::BaseException->throw(
            "Volume is already $percent_used" . "% filled." );
    }

    # Log the percentage filled before populating the data and creating the
    # snapshots.
    $Log->comment( "Percentage used before populating data and creating the "
            . "snapshots is: $percent_used"
            . '%' );

    my ($new_file_count, $new_file_size);
    if (defined $Input{'file_count'} && $Input{'file_count'}) {
        $percent_to_fill = $percent_to_fill + 4;
        $size_to_fill
             = int( ($vol_state->total() / 100) * $percent_to_fill );
        $new_file_count  = $Input{'file_count'};
        $new_file_size   = int( $size_to_fill / $new_file_count );
    } else {
        $new_file_size  = "100m";
        $size_to_fill
             = int( ($vol_state->total() / 100) * $percent_to_fill );
        $new_file_count = int( $size_to_fill
                    / convert_size($new_file_size)->{'bytes'} ) + 1;
    }

    # Loop count based on number of snapshots
    if (defined $Input{'no_of_snapshots'} && $Input{'no_of_snapshots'}) {
        my $file_count = int ($new_file_count / $Input{'no_of_snapshots'});
		
		# when there is a situation where $new_file_count is less than the 'no_of_snapshots"
		# specified, $file_count becomes 0 which causes the undesired results.
		# hence taking the $new_file_count into another variable "$no_of_files" so that we
		# endup creating $new_file_count number of files and create snapshots and once the 
		# required number of files are create(enough to fill the volume to specified %), we just
		# create snapshots (without creating the files).
		my $no_of_files;
		if($file_count == 0) {
			$no_of_files = $new_file_count;
		}
		
        for my $i (1..$Input{'no_of_snapshots'}) {
			if($no_of_files || $file_count){
				NACL::C::Mkfile->mkfile(
						command_interface => $Input{'node'},
						size              => $new_file_size,
						filename          => "/vol/".$Input{'volume'}."/f$i",
						filecount         => $file_count || 1,
						'method-timeout'  => 1800,
				);
				--$no_of_files;
			}
            # Create snapshot
            try {
                $Input{'vol_task_obj'}
                        ->create_snapshot( 'snapshot' => 'snap' . $i );
            }
            catch NATE::BaseException with {
                my $exception_object = shift;
                $Log->warn("Snapshot creation failed: "
                                    . $exception_object->text());
            };
		}
    } else {
        NACL::C::Mkfile->mkfile(
                    command_interface => $Input{'node'},
                    size              => $new_file_size,
                    filename          => "/vol/".$Input{'volume'}."/f",
                    filecount         => $new_file_count,
                    'method-timeout'  => 1800,
        );
    }

    # Validate the % of volume filled is correct?
    $Log->comment(
        "Validating the \% of volume used after populating the data");
    $vol_state = $Input{'vol_task_obj'}->{'vol_obj'}->state();
    $percent_used_after_data_fill = $vol_state->percent_used();
    $percent_used_after_data_fill =~ s/\%//;

    $Log->comment(
            "Volume is $percent_used_after_data_fill" . "% filled" );
    if ( $percent_used_after_data_fill < ( $percent - 2 )) {
        NATE::BaseException->throw(
            "Volume did not get filled to $percent" . '%' );
    }

    $Log->comment("Data write is been completed successfully");
    $Log->comment("WaflironLib:populate_volume_with_snapshots Completed");
    $Log->exit();
} ## end sub populate_volume_with_snapshots

=over

=item wafliron_node_start()

This is basically a function to start wafliron on nodescope with optional
commit true and prev cp option from node shell. Prior to start wafliron,
aggregate is brought offline. Takes care of different selection options
available. Eg: snapid, r,c

Validation for wafliron started is performed if not started, NATE::BaseException
 will be thrown.

Waits till the wafliron is completed and returns nothing.

On an any failure, NATE::BaseException will be thrown with appropriate message.

=over

=item Options

=over

=item C<< "volume" => $volume >>

(Optional) Volume Name

=item C<< "Aggregate" => $aggr >>

(Required) Aggregate Name

=item C<< node=>$command_interface >>

(Required) Node object or command_interface

=item C<< "Optional-commit" => $boolean >>

(Optional) (Default to 1)

=item C<< "force" => $boolean >>

(Optional) (Default to 1)

=item C<< "prev_cp" => $arr_ref >>

(Required) Reference to an array of volume with prev_cp options

=back

=item Return Value

Returns NATE::BaseException on failure

=item Author

Chethan Kumar M

=back

=back

=cut

sub wafliron_node_start {

    $Log->enter();

    # Declare the variables
    my ( $node, $command, $raw_output, $volumes, $prev_cp, $wafl_comp_obj );

    # Validate the arguments
    my %Input = validate_with(
        params => \@_,
        spec   => {
            'aggregate'       => { type => SCALAR },
            'optional-commit' => { type => BOOLEAN, default => 1 },
            'volume'          => { type => ARRAYREF, optional => 1 },
            'force'           => { type => BOOLEAN, default => 0 },
            'prev_cp'         => { type => ARRAYREF },
            'node' => { isa => 'NACL::C::CommandInterface::ONTAP' },
        }
    );
    $Log->comment("In WaflironLib:wafliron_node_start");

    # Get the node object from the instance
    $node = $Input{node};

    set_wafl_flags(
        'node'                              => $node->name(),
        'wafl_iron_optional_commit_enabled' => 1,
        'wafl_iron_panic_on_corruption'     => 0,
        'wafl_inconsistent_panic_testing'   => 1
    );

    # Bring the aggregate offline

    # Instantiate a AggregateTask object

    $Input{'aggr_task_obj'}
        = new AggregateTask( command_interface => $Input{'node'} );

    # Set the aggregate for the AggregateTask Object
    $Input{'aggr_task_obj'}
        ->set_aggr_obj( 'aggregate' => $Input{'aggregate'} );

    # Execute aggregate offline using component object stored in task object
    $Input{'aggr_task_obj'}->{aggr_obj}->modify( 'state' => 'offline' );

    # Wait till the aggregate state is offline
    $Input{'aggr_task_obj'}->check_state(
        'aggregate'      => $Input{'aggregate'},
        'expected_state' => 'offline'
    );

    # Start Wafliron
    $volumes = join( ',', @{ $Input{volume} } ) if defined($volumes);
    $prev_cp = join( ',', @{ $Input{'prev_cp'} } );

    $command = "\"priv set test; aggr wafliron start " . $Input{'aggregate'};
    $command .= ' -o'                   if $Input{'optional-commit'};
    $command .= ' -v ' . $volumes       if defined($volumes);
    $command .= ' -f'                   if $Input{'force'};
    $command .= ' -prev_cp ' . $prev_cp if $prev_cp;
    $command .= "\"";

    $raw_output = $node->run( 'command' => $command );
    $Log->comment("$raw_output");

    # Check if the wafliron is running
    $wafl_comp_obj = new NACL::C::Wafliron(
        'command_interface' => $node,
        'aggregate'         => $Input{'aggregate'}
    );

    if ( $wafl_comp_obj->state->iron_status() !~ /not_running/ ) {
        $Log->comment("Wafliron started!");
    } else {
        NATE::BaseException->throw(
            "Wafliron is not started! Wafliron Status is: "
                . $wafl_comp_obj->state->iron_status() );
    }

    # Wait for wafliron to finish
    $Input{'aggr_task_obj'}
        ->wait_wafliron_status( 'expected_status' => 'not_running' );

    $Log->comment("WaflironLib:wafliron_node_start Completed");
    $Log->exit();
} ## end sub wafliron_node_start

##############################################################################
#
# This method is deprecated. Use create_vol_and_populate_data instead.
#
##############################################################################

sub initialize_volume {
    $Log->warn(
              "The function 'initialize_volume' is deprecated. Replace with "
            . "'create_vol_and_populate_data'" );
    create_vol_and_populate_data(@_);
} ## end sub initialize_volume

############################################################
## Name: create_vol_and_populate_data()
##
## Description:
##		Method does
##		1) Create a volume (striped/flex).
##      2) runs hammer or dd based on inputs from user.
##
## Synopsis:
##
##	create_vol_and_populate_data (
##		'command_interface' => $node,
##		'volume_params'     => {
##			volume             => 'test1',
##			vserver            => 'vs0',
##			aggregate          => 'aggr1',
##			'unix-permissions' => "777",
##			'junction-path'    => "/test1",
##			'junction-active'  => "true"
##		},
##		client          => $CLIENT,
##		'hammer_params' => { 'nofilercore' => '', 'runtime' => '10' }
##	);
##	create_vol_and_populate_data (
##		'command_interface' => $node,
##		'volume_params'     => {
##			volume             => 'test1',
##			vserver            => 'vs0',
##			aggregate          => 'aggr1',
##			'unix-permissions' => "777",
##			'junction-path'    => "/test1",
##			'junction-active'  => "true"
##		},
##		client       => $CLIENT,
##		'file_count' => 5,
##		percentage   => "50%"
##	);
##
## Arguments:
##           <command_interface>  : MANDATORY - command interface to execute
##                                  command.
##           <volume_params>      : MANDATORY - parameters for volume.
##           <client>             : MANDATORY - Name of client.
##			 <hammer_params>	  : OPTIONAL -  parameters for hammer.
##
## Returns: N/A
##
############################################################

sub create_vol_and_populate_data {
    my %opts = validate_with(
        params => \@_,
        spec   => {
            command_interface => { isa  => 'NACL::C::CommandInterface' },
            volume_params     => { type => HASHREF },
            client            => { type => SCALAR },
            hammer_params     => { type => HASHREF, optional => 1 }
        },
        allow_extra => 1
    );
    my $vol = VolumeTask->new(
        'command_interface' => $opts{'command_interface'} );
    $vol->create( $opts{'volume_params'} );
    if ( exists( $opts{'hammer_params'} ) ) {
        run_hammer(%opts);
    } else {
        $Log->comment("Going to populate data in the volume");
        populate_percent_data(%opts);
    }
    return 1;
} ## end sub create_vol_and_populate_data

=over

=item wafliron_cluster_start()

Starts wafliron on cluster shell with optional commit true and prev cp options.
Prior to start wafliron, aggregate is brought offline. Takes care of different
 selection options available. Eg: snapid, r,c

Validation for wafliron started is performed if not started, NATE::BaseException
 will be thrown.

Waits till the wafliron is completed and returns nothing.

On an any failure, NATE::BaseException will be thrown with appropriate message.

=over

=item Options

=over

=item C<< "Aggregate" => $aggr >>

(Required) Aggregate Name


=item C<< node=>$command_interface >>

(Required) Node object or command_interface

=item C<< "Optional-commit" => $boolean >>

(Required) Optional Commit mode

=item C<< "include-mirrors" => $boolean >>

(Optional) Force checking on mirrors

=item C<< "prev_cp" => $arr_ref >>

(Required) Reference to an array of volume and prev_cp options

Ex: 'prev_cp' => ['vol1' => 'r', 'vol2' => 'c'....]

=back

=item Return Value

Returns NATE::BaseException on failure

=item Author

Chethan Kumar M

=back

=back

=cut

sub wafliron_cluster_start {

    my ( $prev_cp_val, $hash, %hash, %wafliron_options, $wafl_comp_obj );

    $Log->enter();
    $Log->comment("In WaflironLib:wafliron_cluster_start");
    my %Input = validate_with(
        params => \@_,
        spec   => {
            'node'      => { isa  => 'NACL::C::CommandInterface::ONTAP' },
            'aggregate' => { type => SCALAR },
            'include-mirrors' => { type => BOOLEAN, optional => 1 },
            'optional-commit' => { type => BOOLEAN, default  => "false" },
            'previous-cp'     => { type => BOOLEAN, default  => "false" },
            'aggr_resp'       => { type => HASHREF, optional => 1 },      #pod
            'vol_resp'        => { type => HASHREF, optional => 1 },      #pod
            'interactive'     => { type => BOOLEAN, default  => 0 },      #pod
            'cmd_timeout'     => { type => SCALAR,  default  => '90' }
        },
        allow_extra => 1
    );
    my $hostspec = $Input{'node'}->node();

    # Validate the parameters.
    # If interactive is "true" validat that prev_cp is set

    if ( $Input{'interactive'} && ( $Input{'previous-cp'} eq "false" ) ) {
        NATE::BaseException->throw(
            "WaflironLib::wafliron_cluster_start - Invalid option value for "
                . "'interactive' - $Input{'interactive'} when 'previous-cp' "
                . ": $Input{'previous-cp'}" );
    } ## end if ( $Input{'interactive'...

    # For prev_cp in interactive mode, both aggr_resp and vol_resp must be
    # provided
    if (   $Input{'previous-cp'} eq "true"
        && $Input{'interactive'}
        && (!(     defined( $Input{'aggr_resp'} )
                || defined( $Input{'vol_resp'} )
            )
        )
        )
    {
        NATE::BaseException->throw( "WaflironLib::wafliron_cluster_start - "
                . "For wafliron to start in previous-cp interactive mode, "
                . "parameters: aggr_resp and vol_resp must be provided" );
    } ## end if ( $Input{'previous-cp'...

    # For prev_cp in non-interactive mode, aggr_resp must be provided
    if (   $Input{'previous-cp'} eq "true"
        && $Input{'interactive'}
        && ( !defined $Input{'aggr_resp'} ) )
    {
        NATE::BaseException->throw( "WaflironLib::wafliron_cluster_start - "
                . "For wafliron to start in previous-cp non-interactive, "
                . "parameter aggr_resp must be provided" );
    } ## end if ( $Input{'previous-cp'...

    # Build the wafliron params
    %wafliron_options = {
        'aggregate'       => $Input{'aggregate'},
        'previous-cp'     => $Input{'previous-cp'},
        'optional-commit' => $Input{'optional-commit'}
    };
    $wafliron_options{'include-mirrors'} = $Input{'include-mirrors'}
        if defined( $Input{'include-mirrors'} );

    # Create and APISet object to run prev_cp in interactive mode.
    # Component does not support interactive mode [burt:]
    my $Host        = host($hostspec);
    my $Api_Set_Obj = NACL::APISet->new(
        hostobj   => $Host,
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );

    # Bring the aggregate offline
    $Log->comment( "WaflironLib::wafliron_cluster_start - "
            . "Bringing the aggregate - $Input{'aggregate'} offline" );

    # Instantiate a AggregateTask object
    $Input{'aggr_task_obj'}
        = new AggregateTask( command_interface => $Input{'node'} );

    # Set the aggregate for the AggregateTask Object
    $Input{'aggr_task_obj'}
        ->set_aggr_obj( 'aggregate' => $Input{'aggregate'} );

    # Execute aggregate offline using component object stored in task object
    $Input{'aggr_task_obj'}->{aggr_obj}->modify( 'state' => 'offline' );

    # Wait till the aggregate state is offline
    $Input{'aggr_task_obj'}->check_state(
        'aggregate'      => $Input{'aggregate'},
        'expected_state' => 'offline'
    );
    my $prev_volinfo_vbn_dep = 0;

    # various prompts for prev_cp
    my $inofile_prompt = qr/Which inofile should be used for (\w+)\?/i;
    my $vol_prompt
        = qr/Which flexvols in aggregate \w+ should perform prev_cp/i;
    my $vol_inofile_prompt1
        = qr/Which inofile should be used for (?:\w+):(\w+)\?/i;
    my $vol_inofile_prompt2 = qr/Selection?/i;
    my $volinfo_vbn_prmpt   = qr/prev_volinfo_vbn\s*:\s*(\d+)/i;
    my $invalid_sel_prompt  = qr/invalid inofile selection|bad volinfo_vbn/i;

    # list of vvols for interative mode
    my $list_of_vols = join( ',', keys %{ $Input{'vol_resp'} } );

    # list of vvol and inofile for interative mode
    my $vol_ni_resp_list
        = join( ',', map { $_ . ':' . $hash{$_}; } keys %hash );

    my $validate_and_answer = sub {

      # if the previous prompt was answered '#', this prompt is invalid. Hence
      # log an error and exit
        NATE::BaseException->throw( "WaflironLib::wafliron_cluster_start - "
                . "Invalid vvol matching done!" )
            if $prev_volinfo_vbn_dep;

        # if prmpted for aggregate inofile, answer the aggr inofile
        return $Input{'aggr_resp'}->{$1}
            if ( defined $Input{'aggr_resp'}->{$1} );

  # when #0 is the vvol option, it should be invalidated and better option
  # to use is 'r' instead
  # when '#' is the vvol option send '#' as answer and handle prev_volinfo_vbn
  # in the prev_volinfo_vbn prompt with #prev_volinfo_vbn

        if ( $Input{'vol_resp'}->{$1} eq "#0" ) {
            $Log->comment( "WaflironLib::wafliron_cluster_start - "
                    . "Invalid prev_vol_vbn no; using r instead" );
            return 'r';
        } elsif ( $Input{'vol_resp'}->{$1} eq "#" ) {
            $prev_volinfo_vbn_dep = 1;
            return '#';
        } else {
            return $Input{'vol_resp'}->{$1};
        }
    };

    my $answer_prev_volinfo_vbn = sub {
        if ($prev_volinfo_vbn_dep) {
            $prev_volinfo_vbn_dep = 0;
            return "#" . $1;
        } else {
            NATE::BaseException->throw(
                "WaflironLib::wafliron_cluster_start - Invalid prompt found");
        }
    };
    my $answer_invalid_sel_prompt = sub {
        return 'r';
    };

    if ( $Input{'interactive'} ) {
        $Log->comment( "WaflironLib::wafliron_cluster_start - Processing the "
                . "prev_cp interactive request" );

        my $match_table = [
            $inofile_prompt      => $validate_and_answer,
            $vol_prompt          => $list_of_vols,
            $vol_inofile_prompt1 => $validate_and_answer,
            $vol_inofile_prompt2 => $list_of_vols,
            $volinfo_vbn_prmpt   => $answer_prev_volinfo_vbn,
            $invalid_sel_prompt  => $answer_invalid_sel_prompt
        ];

        $Api_Set_Obj->storage_aggregate_wafliron_start(
            %wafliron_options,
            'connectrec-match_table' => $match_table,
            'connectrec-timeout'     => $Input{'cmd_timeout'}
        );
    } else {
        $Log->comment( "WaflironLib::wafliron_cluster_start - Processing the "
                . "prev_cp non-interactive request" );

        my %command_args = %wafliron_options;
        $command_args{'connectrec-match_table'}
            = [ $inofile_prompt => $Input{'aggr_resp'}->{$1} ]
            if defined( $Input{'aggr_resp'} );

        $Api_Set_Obj->storage_aggregate_wafliron_start( %command_args,
            'connectrec-timeout' => $Input{'cmd_timeout'} );
    } ## end else [ if ( $Input{'interactive'...

    $wafl_comp_obj = NACL::C::Wafliron->new(
        'command_interface' => $Input{'node'},
        'aggregate'         => $Input{'aggregate'}
    );

    # Check if the wafliron is running

    if ( $wafl_comp_obj->state->iron_status() !~ /not_running/ ) {
        $Log->comment(
            "WaflironLib::wafliron_cluster_start - Wafliron started!");
    } else {
        NATE::BaseException->throw( "WaflironLib::wafliron_cluster_start - "
                . "Wafliron is not started!" );
    }

    # Wait for wafliron to finish
    $Input{'aggr_task_obj'}
        ->wait_wafliron_status( 'expected_status' => 'not_running' );

    $Log->comment("WaflironLib:wafliron_cluster_start Completed");
    $Log->exit();
} ## end sub wafliron_cluster_start

##===========================================================================
## Function:    create_striped_aggregate_member
##
## Description: This is a function to create member aggregates.
##
## Synopsys:    create_striped_aggregate_member(<command_interface>,
##              <aggr_params => hashref of member aggr params>,
##
## Returns:     Boolean
##
##
##===========================================================================

=pod

=head2 create_striped_aggregate_member ()

This is a function to create member aggregates..

=over

=item Synopsis

 create_striped_aggregate_member(
     command_interface => $command_interface,
     aggr_params       => $href_aggr_params
 );

=item Arguments

=over

=item C<< command_interface=>$command_interface >>

=item C<< aggr_params=>$href_aggr_params >>

=back

=item Return Value

=over

True if passed.
Throws an exception if failed.

=back

=back

=cut

sub create_striped_aggregate_member {
    $Log->enter();
    my %opts = validate @_,
        {
        command_interface => { isa  => 'NACL::C::CommandInterface' },
        aggr_params       => { type => HASHREF },
        };
    my @member_aggregate = NACL::C::AggregateMember->find(
        command_interface => $opts{'command_interface'} );
    if ( $#member_aggregate > 1 ) {
        return 0;
    } else {
        NACL::C::AggregateMember->create(
            command_interface    => $opts{'command_interface =>'},
            'add-member-volumes' => "true",
            aggregate            => $opts{'aggr_params'}->{aggregate},
            nodes                => $opts{'aggr_params'}->{nodes},
            foreground           => "false"
        );
    } ## end else [ if ( $#member_aggregate...
    return 1;
} ## end sub create_striped_aggregate_member

=pod

=head2 get_vol_flag ()

This is a function which returns the value of the requested volume flag.

=over

=item Synopsis

get_vol_flag(<volume>, <vol_flag> <node_obj>);

=item Arguments

=over

=item C<< $volume >>

=item C<< $vol_flag >>

=item C<< $node_obj >>

=back

=item Return Value

=over

Returns value of the requested volume flag

Retrurns zero on failure

=back

=item Limitations

As of now this function covers only two volume flags (i.e. I<spcdata_last_fbn>
and I<spcdata_flags>). If needed users can request for enhancement of this
functionto cover more volume flags.

contact: dl-rakesh-automation@netapp.com

=back

=back

=cut

sub get_vol_flag {
    my ( $vol, $flag, $Node ) = @_;
    my $raw_output = $Node->run( 'command' => "vol vvlabel $vol" );
    if ( $flag eq 'spcdata_last_fbn' ) {
        $raw_output =~ /Upgrade scanner last fbn ([0-9]+)/m;
        return $1;
    } elsif ( $flag eq 'spcdata_flags' ) {
        $raw_output =~ /Upgrade scanner flags (0x[0-9a-f]+)/m;
        return $1;
    } else {
        return 0;
    }
} ## end sub get_vol_flag

=over

=item get_wafl_flags()

This function is used to get the wafliron flags.
This function takes multiple flags separated by spaces for the parameter 'flags'

=over

=item Options

=over

=item C<< "hostspec" => $hostspec >>

(Required) Hostspec of the filer

=item C<< "flags" => $flags >>

(Required) list of flags to get the value for delimited by space

=item C<< priv_level=>$priv_level >>

(Optional) privilege level to be used when executing printflag (Default: test)

=back

=item Return Value

If multiple flags are sent as an argument, returns reference to a hash
containing flag as key and falg value as 'value'.

If single flag is sent as an argument, returns only its value as return value.

=item Author

Chethan Kumar M

=back

=back

=cut

sub get_wafl_flags {
    $Log->enter();

    # Declare the variables
    my ($hostspec, $api_set_obj, $raw_output,
        $volumes,  $prev_cp,     $wafl_comp_obj
    );

    # Validate the arguments
    my %Input = validate_with(
        params => \@_,
        spec   => {
            'hostspec'   => { type => SCALAR },
            'flags'      => { type => SCALAR },
            'priv_level' => { type => SCALAR, default => 'test' },
        }
    );
    $Log->comment("In WaflironLib:get_wafl_flags");

    $hostspec = delete $Input{'hostspec'};
    my $priv_level  = delete $Input{'priv_level'};
    my $flags       = delete $Input{'flags'};
    my $no_of_flags = scalar split( ' ', $flags );

    $api_set_obj = NACL::APISet->new(
        name      => $hostspec,
        category  => "Node",
        interface => "CLI",
        set       => 'Nodescope',
    );
    $api_set_obj->set_privilege( "privilege-level" => $priv_level );
    my $parsed_output
        = $api_set_obj->printflag( 'flags' => $flags )->get_parsed_output();
    my %res_hash = %{ $parsed_output->[0] };
    $api_set_obj = undef;
    $Log->comment("WaflironLib:get_wafl_flags Completed");

    $Log->exit();

    return ( $no_of_flags > 1 ) ? \%res_hash : $res_hash{$flags};
} ## end sub get_wafl_flags

#===========================================================================
# Function:    set_vol_flag
#
# Description: This is a helper function which sets the value of the
#              given volume flag
#
# Parameters:
#              volume          : Volume name
#              flag            : Flag Name
#              value           : Value to be set
#              NodeObj         : Node Component Object [command_interface]
#
# Example: set_vol_flag('vol1', 'spcdata_flags', '255');
# Returns: Requested flag value
#          Raises an exception on failure
#
#===========================================================================

sub set_vol_flag {
    my ( $vol, $flag, $value, $Node ) = @_;
    my $raw_output
        = $Node->run( 'command' => "vol vvlabel $vol $flag $value" );

    # Validated if the value is set properly
    my $set_val = get_vol_flag( $vol, $flag, $Node );
    $set_val = oct($set_val) if $set_val =~ /^0/;
    $Log->log("The $flag value is set to : $set_val");

    if ( $set_val == $value ) {
        $Log->log("Volume flag $flag is set to $value succesfully");
        return 1;
    } else {
        NACL::APISet::Exceptions::ResponseException->throw(
            'text' => "Failed to set the volume flag: $flag is to $value" );
    }
} ## end sub set_vol_flag

#===========================================================================
# Function:    root_vol_recovery
#
# Description: This is a helper function which does root volume recovery for
#              a single node cluster.
#
# Parameters:
#              node                => Node Name
#              api_set_obj => APISet object with console connection
#              user        => User Name
#              password    => Password
#
# Returns:     N/A
# Synopsys:    root_vol_recovery ( 'node' => 'wqa-gx78',
#                                  'api_set_obj' => $Api_set_obj,
#                                  'user'        => 'admin',
#                                  'password'    => 'admin123'
#                                );
#
# Notes:
#   - This works only for a single node cluster.
#===========================================================================

sub root_vol_recovery {
    my %args = @_;
    my ( $raw_output, $transit_obj, $console_connection );

    # Construct an API Object for BSD
    my $bsd_api = NACL::APISet->new(
        name      => $args{'node'},
        category  => "Node",
        interface => "CLI",
        set       => 'Systemshell'
    );

    try {
        $raw_output = $bsd_api->kenv( pass_through => "bootarg.init.boot_recovery" )
                ->get_raw_output();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $raw_output = $exception_object->text;
    };
    if ( $raw_output !~ /kenv: unable to get bootarg.init.boot_recovery/m ) {
        $bsd_api->sysctl(
                pass_through      => "kern.bootargs=--bootarg.init.boot_recovery",
                'privilege-level' => 'root'
        );
    }

    try {
        $raw_output = $bsd_api->ls(
                        paths => "/mroot/etc/cluster_config/monitor_mroot.nvfail" )
                        ->get_raw_output();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $raw_output = $exception_object->text;
    };
    if ( $raw_output !~ /No such file or directory/m ) {
        $bsd_api->rm("/mroot/etc/cluster_config/monitor_mroot.nvfail");
    }

    $bsd_api = undef;

    ## Create a transit obj here .
    $console_connection = $args{'api_set_obj'}->get_connection();
    $transit_obj = NACL::Transit->new( name => $console_connection );
    $transit_obj->reboot();

    $args{'api_set_obj'}->reinit();
    $args{'api_set_obj'}->set_privilege( "privilege-level" => "test" );
} ## end sub root_vol_recovery

##===========================================================================
## Function:    get_root_object
##
## Description: This is a function to get the root vol/aggr
##
## Synopsys:   get_root_object(<APISet Obj>, <command_interface>, <vol/aggr>);
##
## Returns:    Returns the root volume/aggregate
##
##
##===========================================================================

sub get_root_object {
    my ( $api, $node, $object_type ) = @_;
    my ( $cmd_output, $response_object );
    try {
        $cmd_output = $api->system_node_run(
            node    => $node,
            command => "$object_type status"
        )->get_raw_output();
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Log->comment( $exception_object->text );
    };

    my @lines = split( '\n', $cmd_output );    # split into lines
    foreach my $line (@lines) {
        $line =~ s/^\s*//;
        $line =~ s/,\s*/,/g;
        my ( $obj, $state, $status, $opts ) = split( /\s+/, $line );

        if ( defined($opts) && ( $opts =~ /^root/ ) ) {
            return ($obj);
        }
    } ## end foreach my $line (@lines)
    NACL::APISet::Exceptions::ResponseException->throw(
        'text' => "No root object found!" );
} ## end sub get_root_object

#===========================================================================
## Function:    is_licensed
##
## Description: Check if the 7Mode license is installed
##
## Parameters:
##                   feature - Feature for which the license needs to be
##                             checked
##                   node    - Node Name
##
## Returns:     type with license code
##              undef if the license is not installed
##
## Notes:
##   - Global parameters referenced: $taskobj->{'SystemNode'}
##===========================================================================

sub is_licensed {
    my ( $feature, $node ) = @_;
    my $api_set_obj = NACL::APISet->new(
        name      => $node,
        category  => "Node",
        interface => "CLI",
        set       => 'Nodescope',
    );
    my $licenses = $api_set_obj->license->get_parsed_output();
    foreach my $license ( @{$licenses} ) {

        if (   ( defined $license->{$feature} )
            && ( $license->{$feature} ne 'not licensed' ) )
        {
            return $license->{$feature};
        }
    } ## end foreach my $license ( @{$licenses...
    return undef;
} ## end sub is_licensed

#===========================================================================
## Function:    purge_vol
##
## Description: Delete the volume using APISet passed.
##
## Parameters:
##                   api_set - APISet Object
##                   vol     - Volume to delete
##
## Returns:     1 on success 0 on failure
##
##
## Notes:
##   - This purge is usefull when cleanup is required after the reboot and
##     deleting the volume through APISet is inevitable.
##
##===========================================================================

sub purge_vol {
    my ( $apiset, $volume, $vserver ) = @_;
    try {
        $apiset->volume_unmount(
            'volume'  => $volume,
            'vserver' => $vserver
        );
        $apiset->volume_modify(
            'volume'  => $volume,
            'vserver' => $vserver,
            'state'   => 'offline'
        );
        $apiset->volume_delete(
            'volume'  => $volume,
            'vserver' => $vserver,
            'force'   => 'true'
        );
        return 1;
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Log->comment( $exception_object->text );
        return 0;
    };
} ## end sub purge_vol

##===========================================================================
## Function:    convert_size
##
## Description: This is a helper function to get the size in all the units.
##
## Synopsys:    convert_size(100mb);
##
## Returns:     Reference to an hash containing size in all units.
##
## Note: This function is deprecated in this lib. Its been moved to Task.
##
##===========================================================================

=pod

=head2 convert_size ()

This is a function to get the given size conveted into different units of size.
This function takes size value in any units and converts it into KB, MB, GB and
TB and returns the same.

=over

=item Synopsis

convert_size(100mb);

=item Arguments

=over

=item C<< $size >>

=back

=item Return Value

=over

Reference to a hash having size units as kyes and converted values as
corresponding values.

=back

=back

=cut

sub convert_size {
    my $size = shift;
    my ( %size, $bytes );
    my %hash;
    if ( $size =~ /(\d+\.{0,1}\d{0,})m/i ) {
        $bytes = $1 * 1048576;
    } elsif ( $size =~ /(\d+\.{0,1}\d{0,})g/i ) {
        $bytes = $1 * 1073741824;
    } elsif ( $size =~ /(\d+\.{0,1}\d{0,})k/i ) {
        $bytes = $1 * 1024;
    } elsif ( $size =~ /(\d+\.{0,1}\d{0,})t/i ) {
        $bytes = $1 * 1099511627776;
    } else {
        $bytes = $size;
    }

    map { $hash{$_} = _round($bytes) and $bytes /= 1024 }
        qw(bytes Kb Mb Gb Tb);
    return \%hash;
} ## end sub convert_size

##===========================================================================
## Function:    round
##
## Description: This is a helper function to get the given number round up to
##              one decimal digit.
##
## Synopsys:    round(100.432);
##
## Returns:     rounded number
##
##
##===========================================================================

sub _round {
    my $number = shift;
    return int( $number + .05 * ( $number <=> 0 ) );
}

##===========================================================================
## Function:    shrink_root_vol
##
## Description: This is a function to shrink the root volume to the given size
##
## Synopsys:    shrink_root_vol(<node>, <root volume>, <size>);
##
## Returns:     1 if succeeds, else returns error string.
##
##
##===========================================================================
sub shrink_root_vol {
    my ( $node, $vol, $size ) = @_;
    my $min_size;
    my $retval      = 1;
    my $api_set_obj = NACL::APISet->new(
        name      => $node,
        category  => "Node",
        interface => "CLI",
        set       => 'Nodescope',
    );
    my $raw_output;
    try {
        $api_set_obj->vol_size( 'volume' => $vol, 'size' => $size );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        $retval = $exception->text();
    };
    return $retval;
} ## end sub shrink_root_vol

##===========================================================================
## Function:    get_vol_move_job_id
##
## Description: This is a function to move a volume from one aggregate to
##              another aggregate.
##
## Synopsys:    get_vol_move_job_id(<node>, <volume>, <vserver>,
##              <target-aggregate>, <member>);
##
## Returns:     Job ID of the volume move job
##
##
##===========================================================================

sub get_vol_move_job_id {
    $Log->enter();
    my $vol_move_obj = undef;
    my %opts         = validate @_,
        {
        node               => { type => SCALAR },
        volume             => { type => SCALAR },
        vserver            => { type => SCALAR },
        'target-aggregate' => { type => SCALAR },
        'member'           => { type => SCALAR, optional => 1 }
        };
    my $Apiset = NACL::APISet->new(
        name      => $opts{'node'},
        category  => "Node",
        interface => "CLI",
        set       => "CMode"
    );

    if ( defined $opts{'member'} ) {
        $vol_move_obj = $Apiset->volume_member_move(
            volume                  => $opts{volume},
            'destination-aggregate' => $opts{'target-aggregate'},
            vserver                 => $opts{'vserver'}
        );
    } else {
        $vol_move_obj = $Apiset->volume_move(
            volume                  => $opts{volume},
            'destination-aggregate' => $opts{'target-aggregate'},
            vserver                 => $opts{'vserver'}
        );
    } ## end else [ if ( defined $opts{'member'...
    if ( $vol_move_obj->{'processed_output'}
        !~ /\[Job (\d+)\] Job is queued:/i )
    {
        FS::EA::Exceptions::WaflironException->throw(
            'text' => "Job Id is not found" );
    }
    return $1;
} ## end sub get_vol_move_job_id
##===========================================================================
## Function:    _get_client_conntype
##
## Description: This is a function to get the conntype for unix client based
##              on host file.
##
## Synopsys:    _get_client_conntype(<client>);
##
## Returns:     hosttype of the client
##
## Note   :   As of now the NACL APISet has only 'Linux' as conntype for all
##            unix types. When there is a change in APISet, this can be
##            updated.
##
##===========================================================================

sub _get_client_conntype {
    my $client_spec = shift;
    my $h           = host($client_spec);
    my $type        = $h->hosttype();

    my %hosttype_map = (
        'unix-solaris'      => 'Linux',
        'unix-linux'        => 'Linux',
        'unix-hpux'         => 'Linux',
        'unix-aix'          => 'Linux',
        'unix-freebsd'      => 'Freebsd',
        'unix-*'            => 'Linux',
        'unix-linux-redhat' => 'Linux'

    );

    return $hosttype_map{$type};

} ## end sub _get_client_conntype

=over

=item _remove_stale_mount()

This is the helper function to remove the stale mount based on given
 mount point.

=over

=item Options

=over

=item C<< $apiset_obj>>

Client APISet object

=item C<< $mount_point>>

Mount Point for the volume

=back

=item Author

Chethan Kumar M

=back

=back

=cut

sub _remove_stale_mount {
    my ( $api_set, $mount_point ) = @_;
    my $mounted;
	$api_set->set_privilege("privilege-level" => "root", 'answer-password' => 1);
    try {
        my $existing_mounts = $api_set->mount()->get_parsed_output();

        foreach my $mount ( @{$existing_mounts} ) {
            if ( $mount->{'mount_point'} eq $mount_point ) {
                $api_set->umount(
                    'paths' => $mount_point,
                    'force' => 'true',
                    'lazy'  => 'true'
                );
            } ## end if ( $mount->{'mount_point'...
        } ## end foreach my $mount ( @{$existing_mounts...
    } ## end try
    catch NATE::BaseException with {
        my $exception_object = shift;
        $Log->comment( $exception_object->text() );
    }
} ## end sub _remove_stale_mount

##===========================================================================
## Function:    _make_dir
##
## Description: This is a helper function to create a directory which is
##              similar to '-p' option in linux 'mkdir'
##              no error if existing, makes parent directories as needed.
##
## Synopsys:    _make_dir(<APISet_Obj>, <dir_path>);
##
## Returns:     NACL::APISet::Exceptions::CommandFailedException on failure
##
## Note: There is a burt (429245) for adding 'ls' and 'mkdir' command support
##       for APISet. Once support is there we can remove execute_raw_command
##===========================================================================

sub _make_dir {
    my ( $api_set, $dir_path ) = @_;
    my $directory;
    my ( undef, @dirs ) = split /\//, $dir_path;
    foreach my $dir (@dirs) {
        $directory .= $dir;
        eval {
            $api_set->execute_raw_command( 'command' => "ls $directory" );
        };
        if ( $@ !~ /No such file or directory/ ) {
            $api_set->execute_raw_command(
                'command' => "sudo mkdir $directory" );
        }
        $directory .= '/';
    } ## end foreach my $dir (@dirs)
    chop $directory;
    eval { $api_set->execute_raw_command( 'command' => "ls $directory" ); };
    if ($@) {
        NACL::APISet::Exceptions::CommandFailedException->throw(
            'text' => "Could not create a directory $directory" );
    }
} ## end sub _make_dir

#=============================================================================
# Sub-routine  : _get_event_logs
#
# Description  : Get All event log messages for the given message name
#
# Synopsys     : _get_event_logs(<time>, <message_name>, <command_interface>);
#
# Return Value : Message string containing all event logs
#=============================================================================

sub _get_event_logs {
    my ( $time, $msg, $ci ) = @_;
    my $eventlog_obj = new EventlogTask( 'command_interface' => $ci );
    my $logs = $eventlog_obj->capture_all_event_logs(
        'init_time' => $time,
        messagename => $msg
    );
    $eventlog_obj = undef;
    return $logs;
} ## end sub _get_event_logs

#=============================================================================
# Sub-routine  : _get_wafl_start_time
#
# Description  : Get the start time of the wafliron based on latest message in
#                the event log.
#
# Synopsys     : _get_wafl_start_time(<command_interface>);
#
# Return Value : Time of wafliron start
#=============================================================================

sub _get_wafl_start_time {
    my ($ci) = shift;
    my $el = NACL::C::Eventlog->find(
        command_interface => $ci,
        'filter'          => { messagename => 'wafl.iron.start' }
    );
    my $time = $el->state()->time();
    $time =~ s/"//g;
    return $time;
} ## end sub _get_wafl_start_time

#=============================================================================
# Sub-routine  : _get_all_wafl_messages
#
# Description  : Returns all wafliron scan related messages
#
# Synopsys     : _get_wafl_start_time(<command_interface>);
#
# Return Value : Time of wafliron start
#=============================================================================

sub _get_all_wafl_messages {
    my ( $ci, $time ) = @_;
    $Log->step("Check for event log message details");
    my @messages = ( 'wafl.scan.iron.done', 'wafl.iron.start' );
    my $elog_obj = new EventlogTask( 'command_interface' => $ci );

    my $captured_logs  = undef;
    my $event_log_msgs = undef;

    foreach my $msg (@messages) {
        $captured_logs = undef;
        $captured_logs = $elog_obj->capture_all_event_logs(
            'messagename' => $msg,
            'init_time'   => $time
        );
        $event_log_msgs .= $captured_logs if ( defined $captured_logs );
    } ## end foreach my $msg (@messages)
    log_step("These are the Event Logs captured in this run ");

    $Log->comment($event_log_msgs);
    return $event_log_msgs;

} ## end sub _get_all_wafl_messages

#=============================================================================
# Sub-routine  : _check_event_log
#
# Description  : Verify wafl scan messages in event log for given volume or
#                aggregate
#
# Synopsis     :
#
#                 _check_event_log(
#                     'name'     => 'aggr1',
#                     'type'     => 'aggregate',
#                     'messages' => $msg_log
#                 );
# Return value : Thorows FS::EA::Exceptions::WaflironException on failure.
#=============================================================================

sub _check_event_log {

    my (%opts) = @_;

    my $name = $opts{'name'};
    my $type = $opts{'type'};
    my $log  = $opts{'messages'};

    if ( $type =~ /aggregate/i ) {

        if ( $log !~ m[Aggregate $name, wafliron completed] ) {
            FS::EA::Exceptions::WaflironException->throw(
                text => "STEP FAIL - Event log is not showing wafliron "
                    . " complete msgs on aggregate $name" );
        } else {
            $Log->comment("WAFLIRON completed on Aggregate $name");
        }
    } ## end if ( $type =~ /aggregate/i)

    if ( $type =~ /volume/i ) {
        if ( $log !~ m{$name.*?wafliron\s+completed\.}i ) {
            FS::EA::Exceptions::WaflironException->throw(
                      text => "STEP FAIL - Event log is not showing wafliron "
                    . " complete msgs on vol"
                    . $name );
        } else {
            $Log->comment("WAFLIRON completed on Volume $name");

        }
    } ## end if ( $type =~ /volume/i)
} ## end sub _check_event_log

#===========================================================================
# Function:    access_wafl_scan_speed
#
# Description: This function is used to get/set the wafliron scan speed.
#
# Parameters:
#              action => 'get'/'set'
#              speed  => '10'
#
# Returns:     Current wafliron scan speed if <action> is get
#              Throws exception on invalid arguements.
#
# Note   :    As there is no support for wafliron commands in 7Mode APISet
#             CMode node command is used.
#             There is a burt raised for this support 431827. Once fixed,
#             this method can use 7Mode APISet.
#
#===========================================================================
sub access_wafl_scan_speed {

    my %args        = @_;
    my $Api_set_obj = NACL::APISet->new(
        name      => $args{'host'},
        category  => "Node",
        interface => "CLI",
        set       => 'CMode',
    );

    if ( $args{'action'} eq 'get' ) {
        my $response_output = $Api_set_obj->system_node_run(
            'node'    => $args{'host'},
            'command' => "wafl scan speed"
        )->get_raw_output();
        $response_output =~ /WAFL scan speed is ([0-9]+)/m;
        return $1;
    } elsif ( $args{'action'} eq 'set' ) {
        if ( defined $args{'speed'} ) {
            $Api_set_obj->system_node_run(
                'node'    => $args{'host'},
                'command' => "wafl scan speed $args{'speed'}"
            );
        } else {
            NACL::APISet::Exceptions::ResponseException->throw(
                'text' => "Arguement \'speed\' is not defined." );
        }
    } else {
        NACL::APISet::Exceptions::ResponseException->throw(
            'text' => "Unknown option for \'action\'" );
    }

} ## end sub access_wafl_scan_speed

#===========================================================================
# Function:    set_wafl_flags
#
# Description: This function is used to set the wafliron flags.
#              This function takes multiple 'flag' => 'value' pairs and sets
#
# Parameters:
#              'node'       => <Node Name>
#              'flag'       => <Value>,
#              'priv_level' => 'test'
#
# Returns:     N/A
#
# Note   :    As there is no support for setflag/printflag commands in 7Mode
#             APISet, CMode node command is used.
#             There is a burt raised for this support 411849. Once fixed,
#             this method can use 7Mode APISet.
#
#===========================================================================

sub set_wafl_flags {
    my %args       = @_;
    my $hostspec   = delete $args{'node'};
    my $priv_level = delete $args{'priv_level'} || 'test';

    my $api_set_obj = NACL::APISet->new(
        name      => $hostspec,
        category  => "Node",
        interface => "CLI",
        set       => 'CMode',
    );
    $api_set_obj->set_privilege( "privilege-level" => $priv_level );

    foreach my $flag ( keys %args ) {
        $api_set_obj->system_node_run(
            'node'    => $hostspec,
            'command' => "setflag $flag $args{$flag}",
        );
        my $output = $api_set_obj->system_node_run(
            'node'    => $hostspec,
            'command' => "printflag $flag"
        )->get_raw_output();
        $Log->comment($output);
    } ## end foreach my $flag ( keys %args)
    $api_set_obj = undef;
} ## end sub set_wafl_flags

#===========================================================================
# Function:    wafliron_cluster_commit_reject
#
# Description: This function is used to commit/reject the IOC changes.
#
# Required Parameters:
#              node => node object
#              aggr_name => name of aggregate
#              operation => commit/reject operation to perform
#
# Returns:     N/A.
#              Throws exception on invalid arguements.
#===========================================================================
sub wafliron_cluster_commit_reject {
    my %Input = validate @_,
        {
        node      => { type => OBJECT },
        aggr_name => { type => SCALAR },
        operation => { type => SCALAR }
        };
    my ( $node, $result );
    $node = $Input{node};
    $Log->comment("In WaflironLib:wafliron_cluster_commit");
    if ( $Input{operation} =~ /commit/ ) {
        NACL::C::Wafliron->commit(
            command_interface => $node,
            aggregate         => $Input{aggr_name}
        );
    } elsif ( $Input{operation} =~ /reject/ ) {
        NACL::C::Wafliron->reject(
            command_interface => $node,
            aggregate         => $Input{aggr_name}
        );
    } else {
        NACL::APISet::Exceptions::ResponseException->throw(
            'text' => "Unknown option for \'operation\'" );
    }
} ## end sub wafliron_cluster_commit_reject

=over

=item wafliron_node_start_boot_menu()

Run the wafliron with optional commit and prev_cp from boot menu on an aggregate

=over

=item Options

=over

=item C<< "apiset_obj_filer_console" => $apiset_obj_filer_console >>

(Required) An ApiSet Object handle to the filer where the connection type is

'console'

=item C<< "aggregate" => $aggregate >>

(Required) Name of the aggregate on which the wafliron will be run from

the boot menu

=item C<< "node_obj" => $node >>

(Required) Handle to the node object .

=item C<< "optional-commit" => 1 >>

(Optional)  If the wafliron should be run with optional commit. By default

optional commit is not set for wafliron from boot menu.

=item C<< "prev-cp" => 1 >>

(Optional)  If the wafliron should be run with prev_cp option. By default

wafliron is not run with prev_cp from the boot menu.

=back

=cut

sub wafliron_node_start_boot_menu {

    my %Input = validate @_,
        {
        'apiset_obj_filer_console' => { isa  => 'NACL::APISet' },
        'aggregate'                => { type => SCALAR, optional => 1 },
        'node_obj'        => { isa  => 'NACL::C::CommandInterface::ONTAP' },
        'optional-commit' => { type => SCALAR, optional => 1 },
        'prev-cp'         => { type => SCALAR, optional => 1 },
        'force'           => { type => BOOLEAN, optional => 1 }
        };

    $Log->comment("In WaflironLib:wafliron_node_start_boot_menu");

    my ( $apiset_obj_filer_console, $userid, $password, $aggregate, $node );
    my ( $node_name, $opt_commit, $prev_cp );
    my ( $flag, $command, $prompt1, $prompt2 );

    ## The %Input contains apiset obj with console type connection
    ## and the name of the aggregate for test
    $apiset_obj_filer_console = $Input{'apiset_obj_filer_console'};
    $aggregate                = $Input{'aggregate'};
    $node                     = $Input{'node_obj'};
    $opt_commit               = "";
    $prev_cp                  = "";

    if ( defined( $Input{'optional-commit'} ) ) {
        $opt_commit = ' -o';
    }

    if ( defined( $Input{'prev-cp'} ) ) {
        $prev_cp = ' -prev_cp';
    }

    ##Taking the Userid and password of the filer for use
    $userid = $apiset_obj_filer_console->{'session'}->{'connection'}
        ->{'username'};
    $password = $apiset_obj_filer_console->{'session'}->{'connection'}
        ->{'password'};
    $node_name = $node->{'hostrec'}->hostname();

    # Reboot the filer and run wafliron on root aggregate
    $Log->comment( "Reboot the filer and run wafliron on the "
            . " aggregate $aggregate " );

    $flag    = 1;
    $command = "wafliron" . $opt_commit . $prev_cp;
    $command .= " -f " if ( defined $Input{'force'} && $Input{'force'} );
    $command .= $Input{'aggregate'} if ( defined $Input{'aggregate'} );

    $apiset_obj_filer_console->system_node_reboot(
        'node'                   => $node_name,
        'connectrec-match_table' => [
            'Press Ctrl-C for Boot Menu' =>
                sub { return ( $flag == 1 ? "/\cC" : "" ); },
            'Please make a selection:' =>
                sub { sleep 10; $flag = 0; return $command; },
            'Please choose one of the following:' =>
                sub { sleep 20; $flag = 0; return $command; },
            'Do you still want to continue? {y|n}:' => "y",
            'Selection\?'                           => "r",
            "LOADER.*"                              => "boot_ontap",
            "login: \$"   => sub { sleep 10; return $userid if ( !$flag ); },
            "Password:\$" => $password,
        ],
        'connectrec-timeout' => '2500'
    );

    ##Reinitialize the console connection
    $apiset_obj_filer_console->reinit();

    $Log->comment(
        "The following command was run from the boot" . " menu : $command" );
    $Log->comment("The filer has successfully booted after the operation");

} ## end sub wafliron_node_start_boot_menu

=over

=item wafliron_node_status()

The method verifies the status of the wafliron from node scope for the

given aggregate and waits for the same to complete

=over

=item Options

=over

=item C<< "aggregate" => $aggregate >>

(Required) Name of the aggregate which has to be tested for wafliron status

=item C<< "node_obj" => $node >>

(Required) Handle to the node object

=item C<< "wait_time" => $wait_time >>

(Optional) Number of seconds to wait before the wafliron completes on the

aggregate

=back

=cut

sub wafliron_node_status {

    my %Input = validate @_,
        {
        'node_obj'  => { isa  => 'NACL::C::CommandInterface::ONTAP' },
        'aggregate' => { type => SCALAR },
        'wait_time' => { type => SCALAR, optional => 1 }
        };

    $Log->comment("WaflironLib : wafliron_node_status");

    my ( $node_obj, $aggregate, $wait_time );
    my ( $cmd, $output );

    $node_obj  = $Input{'node_obj'};
    $aggregate = $Input{'aggregate'};
    $wait_time = $Input{'wait_time'} || '600';

    my ($wi_msg_scan,       $wi_msg_active,
        $wi_msg_not_active, $wi_msg_opt_commit
    );

    $wi_msg_scan       = "Scanning \d+\% done";
    $wi_msg_active     = "Wafliron is active on aggregate $aggregate";
    $wi_msg_opt_commit = "wafliron in optional commit mode has "
        . "completed on aggregate \'$aggregate\'";
    $wi_msg_not_active = "aggr wafliron status $aggregate: wafliron "
        . "is not currently active";

    $cmd = "aggr wafliron status $aggregate";
    $output = $node_obj->run( command => qq("priv set -q test;$cmd") );

    ## If wafliron is not at active raise an exception
    if ( $output =~ /$wi_msg_not_active/si ) {
        NACL::APISet::Exceptions::ResponseException->throw(
            text => "Wafliron is not active on the aggregate $aggregate" );
    }

    ## If wafliron with optional commit is already complete,
    ## and it is pending review then log the same and return success

    if ( $output =~ /$wi_msg_opt_commit/si ) {
        $Log->comment( "wafliron in optional mode on aggregate "
                . " $aggregate is already complete" );
        $Log->comment("Wafliron commit or reject is pending");

        return 0;
    } ## end if ( $output =~ /$wi_msg_opt_commit/si)

    ## Wait for wafliron to complete from node scope here

    my $end_time = time() + $wait_time;

    $output = $node_obj->run( command => qq("priv set -q test; $cmd") );

    while ( $output =~ /$wi_msg_scan/si || $output =~ /$wi_msg_active/si ) {
        Tharn::sleep 10;
        $output = $node_obj->run( command => qq("priv set -q test; $cmd") );
        if ( time() > $end_time ) {
            NACL::APISet::Exceptions::ResponseException->throw(
                      text => "Failure:: Wafliron on aggregate $aggregate "
                    . "could not reach the status complete scanning"
                    . " before timeout " );
        } ## end if ( time() > $end_time)
    }    ##while loop ends here

    $output = $node_obj->run( command => qq("priv set -q test; $cmd") );

    ## Make sure that wafliron is complete with/without optional commit
    ## Throw an exception if not.

    if ( $output =~ /$wi_msg_opt_commit/si ) {
        $Log->comment( "wafliron in optional mode on aggregate "
                . " $aggregate is complete now" );
        $Log->comment("Wafliron commit or reject is pending");
        return 0;
    } elsif ( $output =~ /$wi_msg_not_active/si ) {
        $Log->comment(
            "wafliron on the aggregate " . " $aggregate is complete now" );
        return 0;
    } else {
        NACL::APISet::Exceptions::ResponseException->throw(
            text => "Expected  message is not displayed at the "
                . " end of the wafliron . Verify ." );
    }

} ## end sub wafliron_node_status

1;
