# $Id: //depot/prod/test/Rsierranevada/lib/WaflironLibrary.pm#1 $

#
# Copyright (c) 2010 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary Wafliron library
## @author chethanm@netapp.com
## @status public

package WaflironLibrary;
use Tharn qw(sleep $Log subtest host param_merge);
use warnings;
use strict;
use Data::Dumper;
use POSIX;
use Params::Validate
    qw(validate validate_with SCALAR ARRAYREF HASHREF BOOLEAN UNDEF OBJECT);
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::Wafliron;
use NACL::APISet;
use WaflironLibrary::Utils;
use NACL::APISet::Version;
use NACL::Transit;
use WaflironLibrary::AccessSyslog;

# Exceptions
use NACL::Transit::Exceptions::TransitException qw(:try);
use NACL::Transit::Exceptions::RetriesExceeded qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);
use NATE::Exceptions::Argument qw(:try);
use NACL::Exceptions::NoElementsFound qw(:try);
use WaflironLibrary::Exceptions::WaflironLibException qw(:try);

## @pod here

=head1 NAME

new()

=head2 DESCRIPTION

Creates an object of the WaflironLibrary Class. The new() method also creates
many Component,APISet and connection objects used by the library.

=head2 PARAMETERS

=over

=item C<< container >>

Type: Scalar    Optional:No

Name of a volume/aggregate


=item C<< hostname >>

Type:String    Optional:No

Hostspec of the filer

=item C<< cmd_timeout >>

Type:String    Optional:Yes    Default:90

Command timeout value for any command execution.

=back

=item C<< client >>

Type:String    Optional:Yes    Default:localhost

hostname of a client.

=back

=head2 SYNOPSIS

my $wafliron_lib_obj = new WaflironLibrary(
            'container'   => $FILER_AGGR,
			'hostname'    => $FILER,
			'cmd_timeout' => 600
		);

=cut

use Class::MethodMaker [
    new    => [ '-hash', '-init', 'new' ],
    scalar => 'container',
    scalar => 'hostname',
    scalar => 'cmd_timeout',
    scalar => 'probe_timeout',
    scalar => [
        { -type => 'NACL::C::CommandInterface::ONTAP' },
        'command_interface'
    ],
    scalar => [ { -type => 'Connectrec' },         'conn' ],
    scalar => [ { -type => 'Connectrec' },         'conn_console' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_api' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_bsd_ssh' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_api_console' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_api_maint' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_api_nodescope' ],
    scalar => [ { -type => 'NACL::APISet' },       'filer_api_firmware' ],
    scalar => [ { -type => 'NACL::C::Volume' },    'vol_comp' ],
    scalar => [ { -type => 'NACL::C::Aggregate' }, 'aggr_comp' ],
    scalar => [ { -type => 'NACL::C::Wafliron' },  'wafl_comp' ],
    scalar => [ { -type => 'NACL::Transit' },      'transit_obj' ],
    scalar => 'mode',
    scalar => 'container_type',
    scalar => 'ontap_release',
    scalar => 'ontap_release_codename',
    scalar => 'syslog_file',
    scalar => 'default_scan_speed',
    scalar => 'client',
];

sub init {
    my ( $self, %opts ) = @_;
    my %v_args = validate_with(
        params => \%opts,
        spec   => {
            'command_interface' =>
                { isa => 'NACL::C::CommandInterface::ONTAP' },
            'container'   => { type => SCALAR },
            'hostname'    => { type => SCALAR },
            'cmd_timeout' => { type => SCALAR, default => '90' },
            'probe_timeout' => {
                type     => ARRAYREF,
                optional => 1,
                default  => [ 1, 2, 4, 8, 16, 32, 64, 128 ]
            },
            'client' => { type => SCALAR, default => 'localhost' },
        }
    );
    my $host_obj = host( $v_args{'hostname'} );
    my $mode       = $self->command_interface()->mode();
    my $conn_type  = $mode eq "CMode" ? "ssh" : "rsh";
    my $connection = $host_obj->connect( 'template' => $conn_type );
    $self->client( $v_args{'client'} );

    $self->conn($connection);
    $self->_logmsg( "Connecting console", "comment" );
    $self->conn_console( Tharn::connect("#console", '-login') );

    if ( !defined $mode ) {
        NATE::Exceptions::Argument->throw(
            "Invalid filer host - " . $self->hostname() );
    }
    $self->probe_timeout( $v_args{'probe_timeout'} );
    $self->cmd_timeout( $v_args{'cmd_timeout'} );
    $self->_logmsg( "Creating an Transit Object", "comment" );
    $self->transit_obj( NACL::Transit->new( name => $self->hostname() ) );
    $self->mode($mode);

    $self->_get_ontap_version();
    $self->_create_apiset_objs();
    my $container_type = $self->_get_container_type($mode);
    $self->container_type($container_type);
    if ( $container_type eq "aggregate" ) {
        $self->aggr_comp(
            NACL::C::Aggregate->new(
                'command_interface' => $self->command_interface(),
                'aggregate'         => $self->container()
            )
        );
        $self->wafl_comp(
            NACL::C::Wafliron->new(
                'command_interface' => $self->command_interface(),
                'aggregate'         => $self->container()
            )
        );
    }

    # Get the default wafl scan speed of the release on current filer host.
    my $d_scan_speed = $self->_get_default_scan_speed();

    # For different version of ontap the ems and syslog files are different
    # Defining the ems and syslog files

    if ( $self->ontap_release() =~ /IronCity/i ) {
        $self->syslog_file("/etc/messages");
    }
    else {
        $self->syslog_file("/mroot/etc/mlog/messages.log")
            if ( $mode eq "CMode" );
        $self->syslog_file("/mroot/etc/messages") if ( $mode eq "7Mode" );
    }
}

=head1 NAME

wafliron_start()

=head2 DESCRIPTION

start wafliron on aggregate or traditional volume

=head2 PARAMETERS

=over

=item C<< hostname >>

Type:String    Optional:No

Host name of the filer


=item C<< command_interface >>

Type:String    Optional:No

NACL Command Interface Object [NACL::C::CommandInterface::ONTAP]


=item C<< node_name >>

Type:String    Optional:No

Node Name


=item C<< container_name >>

Type:String    Optional:No

Name of aggregate, traditional volume or flexible volume

=item C<< wait >>

Type:Boolean Default:1 Optional:Yes

if wait is set to 1, the wafliron_start() will wait for wafliron
complete; otherwise, return immediately after start.


=item C<< force >>

Type:Boolean    Optional:Yes

command line option -f


=item C<< triage >>

Type:Boolean    Optional:Yes

command line option -t


=item C<< compression >>

Type:Boolean    Optional:Yes

command line option -c


=item C<< volume_order >>

Type:String    Optional:Yes

volumes list (command line option -v, available from 7.3, e.g vol3,vol1,vol2).


=item C<< optional_commit >>

Type:Boolean    Default:0    Optional:Yes

if oc is 1, wafliron is run in Optional Commit mode if oc is prev_cp,
wafliron is run in Optional Commit prev_cp mode. if oc is 0, regular
wafliron is run.


=item C<< response_list >>

Type:String   Optional:Yes

a hash with key-value pair as vol/aggr name and reponse to be sent to the
vol/aggr during prev_cp


=item C<< inject_file >>

Type:String   Optional:Yes

command line option -i <inject-file>


=item C<< timeout >>

Type:Number   Optional:Yes

wafliron timeout value, $WAFLIRON_TIMEOUT from commonParam.pm is
recommended


=item C<< scan_speed >>

Type:Number   Default:10000    Optional:Yes

Speed of the scan, should be between 0 and 30000. Scan_speed=0 means to
suspend scanning


=item C<< iod >>

Type:Boolean    Optional:Yes   Default: Disable

Enable/Disable IOD flag


=item C<< setflags >>

Type:Hashref    Optional:Yes

A reference to a hash with a list of flags and values to be set before
running wafliron. The script has to explicitly restore the flags to their
original value. The key is the name of the flag and the value is the
value to be set.


=item C<< root_vol >>

Type:String    Optional:Yes

For IRON_ON_ROOT this parameter should contain the name of the volume on
the aggr that should become the root volume when ironing.
For the corruption suite, the object of the iron is an aggr, but for
IRON_ON_ROOT we need to setup the corrupted volume as the root.
This parameter contains the name of the corrupted volume.


=item C<< tmp_root >>

Type:String    Optional:Yes

For IRON_ON_ROOT this parameter indicates a temporary root volume has
already been established and therefore should not be done inside this
library. For the corruption suite, when the corruption is on the root
aggr, a temporary root volume is created by Corruption.pm, so there
is no need to create one here in this library


=item C<< DSC >>

Type:Boolean    Optional:Yes    Default: Enable

Enable/Disable DSC (Defer Summary Check) flag


=item C<< DAC >>

Type:Boolean    Optional:Yes    Default: Disable

Enable/Disable DAC (Defer Active Check) flag


=item C<< COD >>

Type:Boolean    Optional:Yes    Default: Enable

Enable/Disable COD (Container On Demand) flag


=item C<< LSF >>

Type:Boolean    Optional:Yes    Default: Disable

Enable/Disable LSF (Local Status Files) flag


=item C<< panic_on_corruption >>

Type:Boolean    Optional:Yes

Enable/Disable iron panic on corruption


=item C<< maxidle >>

Type:Number    Optional:Yes

After wafliron start, how long should we wait when the console output
pauses. Default to $START_MAXIDLE


=item C<< context >>

Type:String    Optional:No

This specifies if the commands should be run in 7Mode, CMode or
Nodescope context.


=item C<< prev_cp >>

Type:Hashref    Optional:Yes

This is the theory for prev_cp

"aggr_resp"

Type:Hashref    Optional:Yes

Response list for aggregate prompt

"vol_resp"

Type:Hashref    Optional:Yes

Response list for volume prompts

"vol_with_snap_id_list"

Type:Scalar    Optional:Yes

Response list for volume prompts where the response contains both volume and 
snap ids separated by ','.

"ic_vol_resp"

Type:Scalar     Optional:Yes

This referes to incomplete volume response list in partial interactive
pre_cp mode.

"interactive"

Type:Boolean    Optional:No

If this is set to 0, prev_cp will be executed in non-interactive mode.
If this is set to 1, prev_cp will be executed in interactive mode.
If this is set to 2, prev_cp will be executed in semi-interactive mode.


Example:
               $Wafliron_obj->wafliron_start(
                   'optional_commit' => 'prev_cp',
                   'context'         => "CMode",
                   'prev_cp'         => {
                       'interactive' => 2,
                       'aggr_resp'   => { 'testaggr' => '1' },
                       'vol_resp'    => { 'vol1' => 1, 'vol2' => 2 },
                       'ic_vol_resp' => "vol1,vol2,vol3:1,vol4:1"
                   },
                   'wait' => 0
               );


=back

=item <Return Values>

=over

if wait => 0, returns (<syslog_obj>, <cp_hash>)
if wait => 1, returns (<corruption_found>, <return_message>, <cp_hash>)

=back

=cut 

sub wafliron_start {
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'include_mirrors' => { type => BOOLEAN,         optional => 1 },
            'optional_commit' => { type => SCALAR,          default  => 0 },
            'prev_cp'         => { type => HASHREF | UNDEF, optional => 1 },
            'timeout'         => { type => SCALAR,          optional => 1 },
            'wait'            => { type => BOOLEAN,         default  => 1 },
            'force'           => { type => BOOLEAN,         default  => 0 },
            'triage'          => { type => BOOLEAN,         optional => 1 },
            'compression'     => { type => BOOLEAN,         optional => 1 },
            'volume_order'    => { type => ARRAYREF,        optional => 1 },
            'inject_file'     => { type => SCALAR,          optional => 1 },
            'scan_speed'      => { type => SCALAR | UNDEF,  optional => 1 },
            'iod'                   => { type => BOOLEAN, optional => 1 },
            'setflags'              => { type => HASHREF, optional => 1 },
            'root_vol'              => { type => SCALAR,  optional => 1 },
            'tmp_root'              => { type => SCALAR,  optional => 1 },
            'DSC'                   => { type => BOOLEAN, default  => 1 },
            'DAC'                   => { type => BOOLEAN, default  => 1 },
            'COD'                   => { type => BOOLEAN, default  => 1 },
            'LSF'                   => { type => BOOLEAN, default  => 0 },
            'panic_on_corruption'   => { type => BOOLEAN, optional => 1 },
            'maxidle'               => { type => SCALAR,  optional => 1 },
            'context'               => { type => SCALAR,  optional => 1 },
            'powercycle'            => { type => BOOLEAN, default  => 0 },
            'validate_logs'         => { type => BOOLEAN, default  => 1 },
            'vol_with_snap_id_list' => { type => SCALAR,  optional => 1 },
        },
    );
    my ( %flags_to_set, $logmsg, $cp_hash, $scan_speed );
    my $start_match_type = $pkg->container_type();
    my $container_name   = $pkg->container();
    my $context          = $opts{'context'} || $pkg->_get_default_context();
    my $vcontext_regex
        = ( $pkg->mode() eq "CMode" ) ? qr/Nodescope|CMode/ : "7Mode";
    my $default_context = $pkg->_get_default_context();
    my $mode        = $pkg->mode();
    my $api_set_obj = $mode eq "CMode"
                            ? $pkg->filer_api_nodescope()
                            : $pkg->filer_api();


    if ( $context !~ /$vcontext_regex/ ) {
        $pkg->_logmsg(
            "invalid value provided for context : $context on $pkg->mode() filer",
            "warn"
        );
        $pkg->_logmsg(
            "Hence falling back to $default_context as context for $pkg->mode() filer.",
            "warn"
        );
        $context = $default_context;
    }

    $opts{'context'} = $context;

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # Enable/Disable IOD option
    $flags_to_set{'wafl_iron_buf_on_demand'}
        = $opts{'IOD'} ? $opts{'IOD'} : "";

    # Enable/Disbale Defer Summary Check
    $flags_to_set{'wafl_iron_defer_summary_check'}
        = $opts{'DSC'} ? $opts{'DSC'} : 1;

    # Enable/Disable Defer Active Check
    $flags_to_set{'wafl_iron_defer_active_check'}
        = $opts{'DAC'} ? $opts{'DAC'} : 1;

    # Enable/Disbale Container On Demand
    $flags_to_set{'wafl_iron_vvol_buftree_on_demand'}
        = $opts{'COD'} ? $opts{'COD'} : 1;

    my $wait = defined( $opts{wait} ) ? $opts{wait} : 1;

    # get the flag which decides to validate logs or not.
    my $validate_logs = $opts{'validate_logs'};

    # Enable/Disable local status files
    $flags_to_set{'wafl_iron_status_local'} = $opts{'LSF'} ? $opts{'LSF'} : 0;

    # Enable/Disable panic_on_corruption
    $flags_to_set{'wafl_iron_panic_on_corruption'}
        = $opts{'panic_on_corruption'} ? $opts{'panic_on_corruption'} : 0;

    # tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    # When the function wafliron_pick_new_root is called and eventhough after
    # setting the flags If there is any panic when onlining
    # the aggregate when filer is up, then powercycle can be set so that the
    # aggregate will be onlined from maintainance mode
    my $powercycle = defined( $opts{powercycle} ) ? "$opts{powercycle}" : 0;

    # New parameters for IRON_ON_ROOT
    # root is the name of the volume that will become the root volume
    #  when IRON_ON_ROOT=core
    my $root
        = defined( $opts{root_vol} ) ? "$opts{root_vol}" : "$::FILER_VOL";

    # Get the timeout
    my $timeout
        = defined( $opts{timeout} ) ? $opts{timeout} : LOCAL_WAFLIRON_TIMEOUT;

    if ( defined $::IRON_ON_ROOT && $::IRON_ON_ROOT =~ /core/i ) {
        my $prev_cp_mode = $opts{'prev_cp'}->{'interactive'};
        if ( $prev_cp_mode == 0 || $prev_cp_mode == 2 ) {

# REF: http://wikid.netapp.com/w/WAFL/iron/Optional_Commit/IOC_on_C-mode_Volumes/Functional_Specification#Non-requirements

            WaflironLibrary::Exceptions::WaflironLibException->throw(
                "RA IOC prev_cp does not support query mode" );
        }
        if ( $start_match_type eq "volume" ) {
            if ( $pkg->_vol_type($container_name) eq "trad" ) {
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    "RA IOC prev_cp does not support trad vol" );
            }
        }
    }

    # Make sure passed in scan_speed is a valid value.
    if (defined $opts{'scan_speed'}) {
        $scan_speed = $opts{'scan_speed'};
        if ( $scan_speed != -1 && $scan_speed < 0 ) {
            $pkg->_logmsg(
                "wafl scan speed can't be less than minus one, hence setting to 1",
                    "debug" );
            $scan_speed = 1;
        }
        elsif ( $scan_speed > 10000 ) {
            my $msg = "As a standard max wafl scan speed is considered 10000,"
                . " hence setting the wafl scan speed to 10000";
            $pkg->_logmsg( $msg, "debug" );
            $scan_speed = 10000;
        }
    } else {
        # Get current scan speed
        my $parsed_output
            = $api_set_obj->wafl_scan_speed( 'privilege-level' => 'diag' )
                  ->get_parsed_output();
        $scan_speed = $parsed_output->[0]->{'wafl_scan_speed'};
    }

    # Set the wafl scan speed if needed
    if (defined $opts{'scan_speed'}) {
        $logmsg = "Setting the wafl scan speed to : " . $scan_speed;
        $pkg->_logmsg( $logmsg, 'comment' );
        $pkg->_set_scan_speed($scan_speed);
    }

    # Save the original root volume and aggregate
    my $orig_root_aggr   = $pkg->_get_root_obj('aggregate');
    my $orig_root_volume = $pkg->_get_root_obj('volume');
    my $msg;

    if ($::IRON_ON_ROOT) {

        if ( $::IRON_ON_ROOT =~ /core/ ) {
            $msg
                = "Calling wafliron_on_root on account of IRON_ON_ROOT=$::IRON_ON_ROOT";
            $pkg->_logmsg( $msg, 'comment' );
            if ( $root eq "" ) {
                $msg = "root_vol is required for IRON_ON_ROOT=core;";
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    $msg);
            }
            $pkg->_wafliron_pick_new_root(
                'aggr'          => $container_name,
                'vol'           => $root,
                'reboot'        => 0,
                'scan_speed'    => $scan_speed,
                'tmp_root_done' => $tmp_root_done,
                'powercycle'    => $powercycle,
            );
            $Log->comment(
                "control has been returned from _wafliron_pick_new_root",
                'comment' );
            my %wafl_on_root_args = %opts;
            $wafl_on_root_args{'root_vol'} = $root;
            $wafl_on_root_args{'oc'}       = $opts{'optional_commit'};
            $wafl_on_root_args{'force'}    = $opts{'force'};
            $wafl_on_root_args{'timeout'}  = $timeout;
            $wafl_on_root_args{'wait'}     = $wait;
            $wafl_on_root_args{'type'}     = $pkg->container_type();
            $pkg->_logmsg(
                "Arguments before calling are : "
                    . Dumper( \%wafl_on_root_args ),
                'comment'
            );

            if ( !$wait ) {
                $pkg->_wafliron_on_root(%wafl_on_root_args);
                return ( undef, undef, undef );
            }
            else {
                my ( $corruption_found, $return_message )
                    = $pkg->_wafliron_on_root(%wafl_on_root_args);
                return ( $corruption_found, $return_message );

            }
        }
        elsif ( $::IRON_ON_ROOT =~ /multi|spare/ ) {
            my $exception_msg
                = "IRON_ON_ROOT is currently not supported for multi/spare";
            WaflironLibrary::Exceptions::WaflironLibException->throw(
                $exception_msg);
        }

    }

    # Set the wafliron flags
    my %wafl_flags_to_set = %flags_to_set;
    %wafl_flags_to_set = %{ $opts{'setflags'} }
        if defined $opts{'setflags'};
    $logmsg = "Setting the wafl flags as below : \n"
        . Dumper( \%wafl_flags_to_set );
    $pkg->_logmsg( $logmsg, 'comment' );

    $pkg->_set_flags( \%wafl_flags_to_set );

    # fail if wafliron is already running
    if ( $pkg->_get_wafliron_status($context) ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "WaflironLibrary::wafliron_start - wafliron is running on $container_name"
        );
    }
    my $syslog_obj;
    if ($validate_logs) {

        # it's possible for a newly zeroed filer to not have any messages
        # file at all. create an empty messages file so that this script
        # will not fail because it can't find the file.

        my $messages_file = $pkg->syslog_file();

        # Create AccessSyslog object
        $syslog_obj = new WaflironLibrary::AccessSyslog(
            'conn_obj' => $pkg->conn(),
            'log_file' => $messages_file
        );

        if ( !$syslog_obj->syslog_exists() ) {
            $syslog_obj->syslog_create( 'log_file' => '/mroot/etc/mlog' );
        }

   # capture current position in the syslog (for wafl iron start notification)
        $pkg->_logmsg( "marking syslog", 'trace' );
        $syslog_obj->syslog_get_mark();
        $pkg->{'log_lines'} = undef;

    }

    # track wafliron state
    my $wafliron_started = 0;
    if ( !$opts{'optional_commit'} ) {
        $pkg->_start_wafliron(%opts);
    }
    else {

        $logmsg
            = "Changing the state of the aggregate "
            . $container_name
            . " to offline";
        $pkg->_logmsg( $logmsg, 'comment' );
        $pkg->_aggr_offline( $pkg->{'container'}, $context );

        if ( $opts{'optional_commit'} =~ /^1$/ ) {
            $logmsg
                = "Starting the wafliron on  "
                . $container_name
                . " in optional commit mode";
            $pkg->_logmsg( $logmsg, 'comment' );

            $pkg->_start_wafliron(%opts);
        }
        elsif ( $opts{'optional_commit'} eq "prev_cp" ) {

            # start wafliron with prev_cp in semi-interactive mode
            $cp_hash = $pkg->_do_prev_cp_semi_interactive(%opts)
                if ( $opts{'prev_cp'}->{'interactive'} == 2 );

            # start wafliron with prev_cp in interactive mode
            $cp_hash = $pkg->_do_prev_cp_interactive(%opts)
                if ( $opts{'prev_cp'}->{'interactive'} == 1 );

            # start wafliron with prev_cp in non-interactive mode
            $pkg->_do_prev_cp_non_interactive(%opts)
                if ( !$opts{'prev_cp'}->{'interactive'} );

            # throw exception if invalid value for 'interactive' parameter
            if ( $opts{'prev_cp'}->{'interactive'} !~ /[0-2]/ ) {
                NATE::Exceptions::Argument->throw(
                    "Invalid value -  $opts{'interactive'} for option 'interactive'"
                );
            }

            # Go handle the password prompt
            $pkg->_wi_login();
        }
        else {
            NATE::Exceptions::Argument->throw(
                "WaflironLibrary::wafliron_start - Invalid value passed for optional_commit!"
            );
        }
    }

    my $start_timeout = 600;
    my $endtime       = time() + $start_timeout;
    my @wafl_check_info_events;
    $pkg->_logmsg( "Waiting for wafliron to start", 'trace' );

outer:
    while ( time() < $endtime ) {
        $pkg->_logmsg( "...in start wait loop...", 'trace' );

        # if validate_logs is false, no need to validate the logs
        if ( !$validate_logs ) {
            my $wi_status = $pkg->_get_wafliron_status($context);
            if ( $wi_status == 1 || $wi_status == 2 ) {
                $wafliron_started = 1;
                last;
            }
            sleep 10;
            next;
        }
        my $log_lines = $syslog_obj->syslog_tail();
        $pkg->{'log_lines'} .= $log_lines;
        $pkg->_logmsg( "events are\n" . $log_lines . "\n", 'trace' );
        if ($syslog_obj->check_pattern(
                'text'    => $log_lines,
                'pattern' => 'wafl.iron.start'
            )
            || $syslog_obj->check_pattern(
                     'text'    => $log_lines,
                     'pattern' => 'wafl.iron.mounted.volume'
            )
            )
        {
            $wafliron_started = 1;
            last outer;
        }
        elsif (
            $syslog_obj->check_pattern(
                'text'    => $log_lines,
                'pattern' => 'wafl.check.info'
            )
            && $syslog_obj->check_pattern(
                'text'    => $log_lines,
                'pattern' => qr/\s+$start_match_type\s+$container_name/
            )
            )
        {
            my $pattern
                = "not enough free space|cannot start iron|no space left on device";
            if ($syslog_obj->check_pattern(
                    'text'    => $log_lines,
                    'pattern' => $pattern
                )
                )
            {
                my $error
                    = "failed with error (check logs for related wafl.check.info events)\n"
                    . Dumper($log_lines);
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    $error);
            }
            else {
                push( @wafl_check_info_events, $log_lines );
            }
        }
        sleep 10;
    }
    if (@wafl_check_info_events) {
        $msg
            = "matched wafl.check.info event(s) for $start_match_type"
            . "$container_name. these may (or may not) be errors. if wafliron fails"
            . " to start, these events may (or may not) contain the reason. if "
            . "wafliron starts successfully, then this warning can probably be "
            . "ignored. matched events:\n"
            . Dumper( \@wafl_check_info_events );
        $pkg->_logmsg( $msg, 'warn' );
    }

    # did it start, or did we time out?
    if ( !$wafliron_started ) {
        $msg = "Wafliron::wafliron_start - wafliron did not start within the "
            . "timeout period: $timeout";
        WaflironLibrary::Exceptions::WaflironLibException->throw($msg);
    }

    $msg = "Now, wafliron has started on '$container_name'.";
    $pkg->_logmsg( $msg, 'trace' );

    # Depending on the 'wait' => 0|1, either wait for wafliron to complete or
    # return
    if ( $opts{'wait'} ) {
        if ($validate_logs) {
            my ( $corruption_found, $return_message ) = $pkg->wafliron_wait(
                'syslog_obj' => $syslog_obj,
                'timeout'    => $timeout
            );
            return ( $corruption_found, $return_message, $cp_hash );
        }
        else {
            my $status = $pkg->_wafliron_wait_wo_log_check(
                'timeout' => $timeout,
                'context' => $context
            );
            if ( $status == 1 ) {
                return ( undef, undef, undef );
            }
        }
    }
    else {

        # wafliron_state() needs to return the errors found and console
        # messages that is still need to be implemented.
        if ($validate_logs) {
            return ( undef, $syslog_obj, $cp_hash );
        }
        else { return ( undef, undef, undef ); }
    }
}

=head1 NAME

wafliron_review()

=head2 DESCRIPTION

=over

Run the 'aggr wafliron review' command for the specified aggregate and return
the output.

=back

=head2 PARAMETERS

=over

 N/A
 container_name set while instantiating the WaflironLibrary Object is being
 used as the aggregate param for wafliron review.

=back

=head2 RETURNS

=over

Returns wafliron review messages in sclar context

=back

=cut

sub wafliron_review {
    my $pkg            = shift;
    my %opts           = @_;
    my $review_msg     = "";
    my $mode           = $pkg->mode();
    my $container_name = $pkg->container();
    my $context        = $opts{'context'} || $pkg->_get_default_context();
    my $priv_level     = 'test';
    my $logmsg;

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # Get the relevent apiset obj based on context
    my $api_set_obj = $pkg->_get_context_based_api($context);

    # phrase the command based on context
    my $cmd
        = $context =~ /7Mode|Nodescope/i
        ? "aggr_wafliron_review"
        : "storage_aggregate_wafliron_review";

    # Iron_on_root in-core does not produce a review file
    if ( defined $::IRON_ON_ROOT && $::IRON_ON_ROOT =~ /spare|multi/ ) {
        my $messages_file;

        # Define the oc log file based on ontap version
        if ( $pkg->ontap_release() eq "IronCity" ) {
            $messages_file = "/etc/log/wafliron/oc-$container_name-current";
        }
        else {
            $messages_file
                = "/mroot/etc/log/wafliron/oc-$container_name-current";
        }

        # Create AccessSyslog object
        my $syslog_obj = new WaflironLibrary::AccessSyslog(
            'conn_obj' => $pkg->conn(),
            'log_file' => $messages_file
        );

        # Check the log file if exists? and get the review messages
        if ( $syslog_obj->syslog_exists() ) {
            $review_msg = $syslog_obj->syslog_dump();
        }
        else {
            $logmsg = $messages_file . " does not exists!";
            $pkg->_logmsg( $logmsg, 'debug' );
        }

        $logmsg = "review message is : $review_msg";
        $pkg->_logmsg( $logmsg, 'comment' );
    }
    else {
        my $response_obj = $api_set_obj->$cmd(
            'aggregate'       => $container_name,
            'privilege-level' => $priv_level
        );
        $review_msg = $response_obj->get_raw_output();
    }
    return $review_msg;
}

=head1 NAME

wafliron_status()

=head2 DESCRIPTION

Is wafliron running on specified aggr or not?

If the wafliron is running on the aggregate, it returns 1.

If wafliron is idle returns 2.

If no wafliron in progress returns 0.

=head2 PARAMETERS

=over

N/A
container_name set while instantiating the WaflironLibrary Object is being
used as the aggregate param for wafliron status.

=back

=head2 RETURNS

=over

Returns status as below,

1 - Wafliron is running

2 - Wafliron is idle

0 - Wafliron is not running

=back

=cut

sub wafliron_status {
    my ( $pkg, %opts ) = @_;
    my $context = $opts{'context'} || $pkg->_get_default_context();

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # capture result in a var to force scalar context
    my $status = $pkg->_get_wafliron_status($context);

    return $status;
}

=head1 NAME

wafliron_reject()

=head2 DESCRIPTION

Reject the optional-commit status files for the aggregate/tradvol.
container_name set while instantiating the WaflironLibrary Object is being
used as the aggregate param for wafliron status.

=head2 PARAMETERS

=over

=item C<< context >>

Type:String    Optional:No
This specifies if the commands should be run in 7Mode, CMode or
Nodescope context.

=back

=head2 RETURNS

=over

 Returns optional commit messages

=back

=cut

sub wafliron_reject {
    my ( $pkg, %opts ) = @_;
    my ( $oc_message, $messages_file, $priv_level, $logmsg );
    my $context = $opts{'context'} || $pkg->_get_default_context();

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # Get the relevent apiset obj based on context
    my $api_set_obj = $pkg->_get_context_based_api($context);

    $priv_level = "test";

    #  'validate_logs' decides if the log messages need to be validated in the
    #  logs or not. It defaults to 1.

    my $validate_logs
        = defined( $opts{validate_logs} ) ? "$opts{validate_logs}" : 1;

    # phrase the command based on context
    my $cmd
        = $context =~ /7Mode|Nodescope/i
        ? "aggr_wafliron_reject"
        : "storage_aggregate_wafliron_reject";

    # tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    my $container_name = $pkg->container();
    my $end_time       = time() + LOCAL_WAFLIRON_TIMEOUT;

    ## Assume that wafliron scanners are done. It's the caller's
    ## responsibility to make sure of that.

    # See if we need to enter the password
    # Handle any password prompt
    $pkg->_wi_login();

 # If we are running IRON_ON_ROOT=core, then we are in the boot menu and need
 # to send the command to the console and capture all the output in oc_message
    if ( defined $::IRON_ON_ROOT && $::IRON_ON_ROOT =~ /core/ ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "WaflironLibrary::wafliron_start - IRON_ON_ROOT is currently not supported yet"
        );
    }
    else {

        my $syslog_obj;

        if ($validate_logs) {

            # Create AccessSyslog object
            $messages_file = $pkg->syslog_file();
            $syslog_obj    = new WaflironLibrary::AccessSyslog(
                'conn_obj' => $pkg->conn(),
                'log_file' => $messages_file
            );

            # Check the log file if exists? and get the review messages
            if ( !$syslog_obj->syslog_exists() ) {
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    $messages_file . " does not exists!" );
            }

            $syslog_obj->syslog_get_mark( 'log_file' => $messages_file );
        }

        my $response_obj = $api_set_obj->$cmd(
            'aggregate'          => $container_name,
            'privilege-level'    => $priv_level,
            'connectrec-timeout' => LOCAL_WAFLIRON_TIMEOUT
        );

        # wait until reject is done or timeout
        while (( time() < $end_time )
            && ( $pkg->_get_wafliron_status($context) != 0 ) )
        {
            sleep(15);
        }
        if ( !$validate_logs ) {
            if ( $pkg->_get_wafliron_status($context) == 0 ) {
                return $response_obj->get_raw_output();
            }
            else {
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    "Wafliron failed to reject the changes");
            }
        }
        $oc_message .= $syslog_obj->syslog_tail();

        # Deletion of the change log files can be confirmed through
        # $chglog_del_msg. So, we need to wait for $chglog_del_msg and
        # also it should get append with $oc_message
        my $chglog_del_msg
            = "Deleted change log files for aggregate $container_name";
        my $log_fetch_timer = time() + 30;

        # wait till you get the message or 30 seconds whichever is earlier
        until (    ( $oc_message =~ m/$chglog_del_msg/g )
                || ( time() > $log_fetch_timer ) )
        {

            # append the messages to $oc_message
            $oc_message .= $syslog_obj->syslog_tail();

            # sleep a bit to avoid frequent check
            sleep(1);
        }

        # $oc_message to be NATE logged, refer to enhancement burt 336996
        $logmsg = "wafliron reject messages : $oc_message";
        $pkg->_logmsg( $logmsg, 'comment' );

        # Verify if wafliron rejected the changes.
        if ($oc_message =~ /Rejecting\s+change\s+log\s+files\s+for\s+\w+\s+
                     	 $container_name.*|Rejected\schanges\sfor\saggregate/ixs
            )
        {
            $pkg->_logmsg( "Wafliron rejected changes", 'comment' );
        }
        else {
            WaflironLibrary::Exceptions::WaflironLibException->throw(
                "Wafliron failed to reject the changes");
        }

        # Verify the deletion of change log files.
        if ($oc_message
            =~ /Deleted\s+change\s+log\s+files\s+for\s+aggregate\s+
                          $container_name.*/ixs
            )
        {
            $pkg->_logmsg( "Change logs files have been deleted", 'comment' );
        }
        else {
            WaflironLibrary::Exceptions::WaflironLibException->throw(
                "Change logs files have not been deleted!");
        }
    }

    # If any kind of IRON_ON_ROOT we need to restore the original root volume
    # after rejecting
    if ( defined $::IRON_ON_ROOT && !$tmp_root_done ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "WaflironLibrary::wafliron_start - IRON_ON_ROOT is currently not supported"
        );
    }
    return $oc_message;
}

=head1 NAME

wafliron_reject()

=head2 DESCRIPTION

Checks to see if logs files exist for a volume specified, if volume is not
defined, current root volume will be used.

=head2 PARAMETERS

=over

=item C<< wafliron_type >>

Type:String    Optional:Yes    Default: "oc"
This specifies wafliron Type.

=back

=head2 RETURNS

=over

 Returns 1 if the log files are present for the given Volume else returns 0.

=back

=cut

sub wafliron_logfiles_present {
    my ( $pkg, %opts ) = @_;

    # Declare the variables
    my ($iron_log_dir);

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # Initial local variables
    my ($wafliron_type)
        = defined( $opts{wafliron_type} ) ? $opts{wafliron_type} : "oc";
    my ($volume)
        = defined( $opts{volume} )
        ? $opts{wafliron_type}
        : $pkg->container();
    my $apiset_obj
        = $pkg->mode() eq "CMode"
        ? $pkg->filer_api()
        : $pkg->filer_api_nodescope();

    # Enable metadata visible flag so we can check the wafliron metadir
    $pkg->_logmsg( "Enable wafl_metadata_visible", 'debug' );

    my $old_wafl_metadata_flag
        = $pkg->_misc_printflag( flag => 'wafl_metadata_visible' );

    my $ret_val = $pkg->_set_flags( \{ 'wafl_metadata_visible' => 1 } );

    # Check if the flag setting successful
    if ( !($ret_val) ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "Unable to set wafl_metadata_visible");
    }

    if ( defined($volume) ) {
        $iron_log_dir = $volume;
    }
    else {

        # For now, the log files reside in the root directory
        $pkg->_logmsg( "Get Root Directory", 'debug' );

        $iron_log_dir = $pkg->_get_root_obj("volume");
    }

    $iron_log_dir = "/vol/$iron_log_dir/metadir/irondir/"
        if ( $pkg->mode() eq "7Mode" );

    $pkg->_logmsg( "Check for log files", 'debug' );
    $ret_val = $apiset_obj->ls(
        'paths'           => $iron_log_dir,
        'privilege-level' => "test"
    )->get_raw_output();
    $pkg->_logmsg( "ls $iron_log_dir results\n" . $ret_val, 'debug' );

    # Restoring old value
    $pkg->_logmsg(
        "Restoring wafl_metadata_visible to $old_wafl_metadata_flag",
        'debug' );
    $pkg->_set_flags(
        \{ 'wafl_metadata_visible' => $old_wafl_metadata_flag } );

    my @files = split( /\n/, $ret_val );

    if ( $wafliron_type eq "nlif" ) {
        if ( grep ( /iron_container_bitmap/, @files ) ) {
            return 1;
        }
    }
    elsif ( $wafliron_type eq "oc" ) {
        if ( grep ( /OC-Change-Log/, @files ) ) {
            return 1;
        }
    }
    else {
        return 0;
    }
}

=head1 NAME

wafliron_wait()

=head2 DESCRIPTION

This function works with walfiron_start(). It waits for wafliron to complete
on the specified container.Returns (corrution_found, corruption_messages);

$corruption_found == 1 means "found corruption"

$corruption_messages:
The wafliron corruption messages can have many different patterns, this function
 will not try to match any specific pattern. Instead it will return any message
saved from wafliron Starting to wafliron completed on specified container.
The following is an example of the corruption_message:

Mar 29 12:22:12 GMT [wafl.iron.start:notice]: Starting wafliron on \
aggregate wafl_check_iron_aggr.
Tue Mar 29 12:22:12 GMT [wafl.check.info:error]: WAFLIRON, aggregate \
wafl_check_iron_aggr: Directory 103, block 0, unallocated entry 2 \
with nonzero fields. Clearing.
Tue Mar 29 12:22:12 GMT [wafl.vv.vol.config:notice]: Unable to configure \
volume in aggregate 'wafl_check_iron_aggr': missing container inode
Tue Mar 29 12:22:12 GMT [wafl.vv.vol.config:notice]: Unable to configure \
volume in aggregate 'wafl_check_iron_aggr': missing container inode
Tue Mar 29 12:22:13 GMT [wafl.delete.restart:info]: Debug-build-only \
message - restarting 0 zombies, aggr wafl_check_iron_aggr 0x95f33098
Tue Mar 29 12:22:53 GMT [wafl.iron.start:notice]: Starting wafliron on \
volume lost_volume1.
Tue Mar 29 12:22:54 GMT [wafl.delete.restart:info]: Debug-build-only \
message - restarting 0 zombies, vol lost_volume1 0x95f62098
Tue Mar 29 12:22:54 GMT [wafl.scan.start:info]: Starting container block \
reclamation on volume lost_volume1.
Tue Mar 29 12:29:04 GMT [wafl.scan.iron.done:info]: Volume lost_volume1, \
wafliron completed.
Tue Mar 29 12:33:48 GMT [wafl_filesystem_sca:info]: DBG: iron pvol \
wafl_check_iron_aggr - completing
Tue Mar 29 12:33:53 GMT [wafl.check.info:error]: WAFLIRON, aggregate \
wafl_check_iron_aggr: 1 lost files or directories relinked into lost+found
Tue Mar 29 12:33:53 GMT [wafl.scan.iron.done:info]: Aggregate \
wafl_check_iron_aggr, wafliron

=head2 PARAMETERS

=over

=item syslog_obj

WaflironLibrary::AccessSyslog Object created.

=item timeout

wafliron timeout value, $WAFLIRON_TIMEOUT from commonParam.pm is
recommended

=back

=head2 RETURNS

=over

Returns (corrution_found, corruption_messages);

=back

=cut

sub wafliron_wait {
    my ( $pkg, %opts ) = @_;
    my $container_name = $pkg->container();
    my $tail           = $pkg->{'log_lines'};
    my $syslog_obj     = $opts{'syslog_obj'};
    my $timeout        = $opts{'timeout'} || LOCAL_WAFLIRON_TIMEOUT;
    my $endtime        = time() + $timeout;
    my $context        = $opts{'context'} || $pkg->_get_default_context();
    my $mode           = $pkg->mode();

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    # if iron_on_root look for wafliron completed messages from console
    if ( defined $::IRON_ON_ROOT && $::IRON_ON_ROOT =~ /core/ ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "IRON_ON_ROOT=core is not supported in this library yet!");
    }
    else {
        my $lines              = $tail;
        my $wi_completed_msg   = qr/(Call home for WAFL.*?)\n/is;
        my $start_msg_varified = 0;
        my $end_msg_varified   = 0;

        # wait until wafliron complete
        while ( time() < $endtime ) {
            if ( $pkg->_get_wafliron_status($context) == 1 ) {
                sleep(45);
            }
            else {
                sleep(45);
                $lines .= $syslog_obj->syslog_tail();

             # Because of excessive syslog messages (ex. in case of summarymap
             # corruption) the following line matching could hang.
             # To avoid, we can split lines and search for start message.
                my @lines = split /\n/, $lines;
                my $length = scalar @lines;
                for ( my $i = 0; $i <= $length; $i += 50 ) {
                    my $lines_few;
                    if ( ($i + 50) > $length ) {
                        $lines_few = join( "\n", @lines[ $i .. $length - 1 ] );
                    } else {
                        $lines_few = join( "\n", @lines[ $i .. $i + 50 ] );
                    }

                    # matching the starting message for walfiron
                    if ( $lines_few
                        !~ /\[.*\]\s+Starting\s+wafliron\s+on\s+\w+\s+$container_name.*/ixs
                        )
                    {
                        $pkg->_logmsg(
                            "Did not find wafliron start message:\n$lines_few",
                            'warn'
                        ) if ( !$start_msg_varified );
                    }
                    else {
                        $lines_few =~ s/(.*\s*\w+\s+\w+\s+\d+\s+\d+:\d+:\d+)
                                  (\s+\w+\s+.*\[.*\]:\s+
                                  Starting\s+wafliron\s+on\s+\w+\s+
                                  $container_name.*)/$1$2/ixs;
                        $pkg->_logmsg(
                            "Found message - Starting wafliron on...",
                            'comment' );
                        $start_msg_varified = 1;
                        last if ( $mode ne "CMode" );
                    }
                    if (   ( $mode eq "CMode" )
                        && ( $lines_few =~ /$wi_completed_msg/ ) )
                    {
                        $end_msg_varified = 1;
                        last
                            if ( ($start_msg_varified)
                            && ( $mode eq "CMode" ) );
                    }
                }
                my $found = 0;
                if ( $lines !~ s/(.*?$wi_completed_msg).*/$1/s ) {
                    my $msg
                        = "Did not find wafliron completed message. This "
                        . "is expected in rare cases involving a lot of "
                        . "wafliron output, e.g. massive corruptions. For "
                        . "those cases, this condition can be ignored within"
                        . " the test script code. Otherwise, it should be "
                        . "considered a product bug";
                    $pkg->_logmsg( $msg, 'warn' );
                }
                if ($lines =~ /wafl.check.info:error|wafl.check.info:ALERT/
                    || (   $lines =~ /wafl.iron.*:CRITICAL/
                        && $lines
                        !~ /wafl.iron.oc.deletedChangeLog:CRITICAL|wafl.iron.oc.noCorruption:CRITICAL/
                    )
                    )
                {
                    $found = 1;
                }
                return ( $found, $lines );
            }
        }
        $opts{'syslog_obj'} = undef;
        $pkg->{'log_lines'} = undef;
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "wafliron timeout.");
    }
}

=head1 NAME

wafliron_commit()

=head2 DESCRIPTION

This function mimics the 'aggr wafliron commit'.
container_name set while instantiating the WaflironLibrary Object is being
used as the aggregate param for wafliron commit.

=head2 PARAMETERS

=over

=item C<< context >>

Type:String    Optional:No
This specifies if the commands should be run in 7Mode, CMode or
Nodescope context.

=item C<< force >>

Type:boolean    Optional:yes
This enables the commit with force.

=back

=item C<< timeout >>

Type:number    Optional:yes
Timeout value for wafliron_commit. Default is: 2400

=item C<< root_vol >>

Type:String    Optional:yes
Name of the volume that will become the root volume when IRON_ON_ROOT=core

=item C<< validate_logs >>

Type:number    Optional:yes  Default: 1
This flag value decides if the log messages need to be validated or not.

=item C<< delete_all_snaps >>

Type:Number   Optional:yes  Default: 0
If the delete_all_snaps option is set then we must get into boot menu and
delete all snaps before the system will come up.

=head2 RETURNS

=over

 Returns optional commit messages

=back

=cut

sub wafliron_commit {
    my ( $pkg, %opts ) = @_;

    my ( $oc_message, $force, $tail, $messages_file, $mount_root,
        $already_halted );

    # Verify the connections and reconnect if required
    $pkg->_check_connection_state();

    $force = defined( $opts{force} ) ? $opts{force} : 1;

    # Get the timeout
    my $timeout
        = defined( $opts{timeout} ) ? $opts{timeout} : LOCAL_WAFLIRON_TIMEOUT;

    # root is the name of the volume that will become the root volume
    # when IRON_ON_ROOT=core

    my $root
        = defined( $opts{root_vol} ) ? "$opts{root_vol}" : "$::FILER_VOL";

    #  tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    #  'validate_logs' decides if the log messages need to be validated in the
    #  logs or not. It defaults to 1.

    my $validate_logs
        = defined( $opts{validate_logs} ) ? "$opts{validate_logs}" : 1;

    # If the delete_all_snaps option is set then we must get into boot menu
    #  and delete all snaps before the system will come up
    my $delete_all_snaps
        = defined( $opts{delete_all_snaps} ) ? "$opts{delete_all_snaps}" : 0;

    my $container_name = $pkg->container();

# If IRON_ON_ROOT with delete_all_snaps then we need to halt the filer
# and delete the existing snaps before we can allow the filer to come all the way up
    $already_halted = 0;

    ## Assume that wafliron scanners are done. It's the caller's
    ## responsibility to make sure of that.

    # Get the relevent apiset obj based on context
    my $context = $opts{'context'} || $pkg->_get_default_context();
    my $api_set_obj = $pkg->_get_context_based_api($context);

    # phrase the command based on context
    my $cmd
        = $context =~ /7Mode|Nodescope/i
        ? "aggr_wafliron_commit"
        : "storage_aggregate_wafliron_commit";

    # make sure if the filer into the CLI prompt
    $pkg->_wi_login();

 # If we are running IRON_ON_ROOT=core, then we are in the boot menu and need
 # to send the command to the console and capture all the output in oc_message
    if ( defined $::IRON_ON_ROOT && $::IRON_ON_ROOT ) {

        # TODO: implement IRON_ON_ROOT
        my $error_msg
            = "IRON_ON_ROOT is not supported for wafliron commit yet!";
        WaflironLibrary::Exceptions::WaflironLibException->throw($error_msg);
    }
    else {
        my ( $messages_file, $syslog_obj );

        if ( $validate_logs == 1 ) {

            # get a handle on the /etc/messages file
            $messages_file = $pkg->syslog_file();

            # Create AccessSyslog object
            $syslog_obj = new WaflironLibrary::AccessSyslog(
                'conn_obj' => $pkg->conn(),
                'log_file' => $messages_file
            );

            # Check the log file if exists?
            if ( !$syslog_obj->syslog_exists() ) {
                WaflironLibrary::Exceptions::WaflironLibException->throw(
                    $messages_file . " does not exists!" );
            }

            $syslog_obj->syslog_get_mark( 'log_file' => $messages_file );
        }
        my $prompted_for_commit;
        $api_set_obj->$cmd(
            'aggregate'              => $container_name,
            'privilege-level'        => 'test',
            'connectrec-match_table' => [
                eval RE_COMMIT_CHANGES => sub {
                    $prompted_for_commit = 1;
                    return ( $force == 1 ) ? "Y" : "N";
                },
                eval RE_SNAPMIRRORED_VOLS => sub {
                    return ( $force == 1 ) ? "Y" : "N";
                }
            ],
            'connectrec-timeout' => $timeout,
        );

        # wait until commit is done or timeout
        my $end_time = time() + LOCAL_WAFLIRON_TIMEOUT;

        while (( time() < $end_time )
            && ( $pkg->_get_wafliron_status($context) != 0 ) )
        {
            sleep(15);
        }
        if ( $pkg->_get_wafliron_status($context) != 0 ) {
            WaflironLibrary::Exceptions::WaflironLibException->throw(
                "wafliron is not completed during "
                    . LOCAL_WAFLIRON_TIMEOUT );
        }

        if ( $validate_logs == 0 ) {

            my $container_state = $pkg->_get_aggr_status($context);
            if ( $container_state ne 'offline' ) {
                $pkg->_logmsg( "container status is: $container_state",
                    'warn' );
            }
            return 1;
        }
        $oc_message = $syslog_obj->syslog_tail();

        my $req_end_msg
            = "Deleted change log files for aggregate $container_name";
        my $log_fetch_timer = time() + 30;

        # wait till you get the message or 30 seconds whichever is earlier
        until (    ( $oc_message =~ m/$req_end_msg/g )
                || ( time() > $log_fetch_timer ) )
        {

            # append the messages to $oc_message
            $oc_message .= $syslog_obj->syslog_tail();

            # sleep a bit to avoid frequent check
            sleep(2);
        } ## end until ( ( $oc_message =~ ...

        # $oc_message to be NATE logged
        $pkg->_logmsg( "wafliron commit messages: \n" . $oc_message . "\n",
            'debug' );

        # Verify if wafliron committed the changes.

        if ($oc_message =~ /Committed\s+change\s+log\s+files\s+for\s+\w+\s+
                     	 .*$container_name.*/ixs
            )
        {
            $pkg->_logmsg( "Wafliron committed changes", "comment" );
        }
        elsif ( $prompted_for_commit && ( $force == 0 ) ) {
            $pkg->_logmsg(
                "Wafliron did not commit changes when prompted: force was $force",
                "comment"
            );
        }
        else {
            die("Wafliron failed to commit changes");
        }

    } ## end else [ if ( $::IRON_ON_ROOT =~...

    # If any kind of IRON_ON_ROOT we need to restore the original root volume
    # after committing
    if ( $::IRON_ON_ROOT && !$tmp_root_done ) {

        # TODO: to be implemented
    } ## end if ( $::IRON_ON_ROOT &&...

    return ($oc_message);
}

=head1 NAME

recover_root_vol()

=head2 DESCRIPTION

This function can be used to recover root volume of a single node cluster.
Once the root recovery flags are cleared, filer is rebooted and brought into
CLI state.

=head2 PARAMETERS

=over

=item C<< powercycle >>

Type:Number   Optional:yes  Default: 0
If set to 1 and if transition fails once, try to cycle power on the host and then try exactly once more.

=back

=cut

sub recover_root_vol {
    my ( $pkg, %opts ) = @_;

    my $mode = $pkg->mode();
    if ( $mode eq "7Mode" ) {
        WaflironLibrary::Exceptions::WaflironLibException->throw(
            "This function call is invalid on 7Mode host");
    }

    my $transit_obj = $pkg->transit_obj();
    $transit_obj->change_state( 'to' => 'CLI' );
    $pkg->_check_connection_state();

    $pkg->_logmsg( "Starting the root vol recovery..", 'comment' );

    my $apiset_obj     = $pkg->filer_bsd_ssh();
    my $transit_to_obj = $pkg->transit_obj();
    my $powercycle = defined( $opts{powercycle} ) ? "$opts{powercycle}" : 0;

    my $raw_output
        = $apiset_obj->kenv( pass_through => "bootarg.init.boot_recovery" )
        ->get_raw_output();
    if ( $raw_output !~ /kenv: unable to get bootarg.init.boot_recovery/ ) {

        # burt497419
        $apiset_obj->execute_raw_command( 'command' =>
                "sudo sysctl kern.bootargs=--bootarg.init.boot_recovery" );

        $raw_output
            = $apiset_obj->ls(
            'paths' => "/mroot/etc/cluster_config/monitor_mroot.nvfail" )
            ->get_raw_output();
        if ( $raw_output !~ /No such file or directory/m ) {
            $apiset_obj->rm("/mroot/etc/cluster_config/monitor_mroot.nvfail");
        }
    }
    $pkg->_logmsg( "Finished the root vol recovery..", 'comment' );
    $pkg->_logmsg( "Now rebooting...",                 'comment' );
    $transit_to_obj->reboot( 'powercycle' => $powercycle );
    $transit_to_obj->change_state( 'to' => 'CLI' );
    $pkg->_check_connection_state();
}

##
## @name ::cold_start_filer
## Cold start a filer, basically does a newfs on the filer
## args are the same as for the Filer/newfs.thpl script
##
## This is a generic method, hence it needs to be moved unders some proper
## NACL lib path in future.
##

sub cold_start_filer {
    my ( $pkg, %args ) = @_;

    my $newfs = defined( $args{newfs} ) ? $args{newfs} : 0;

    my $fh          = $pkg->conn()->hostp();
    my $apiset_obj  = $pkg->filer_api_console();
    my $transit_obj = $pkg->transit_obj();

    if ($newfs) {
        $pkg->_logmsg( "Cold starting the filer", 'debug' );
    }
    else {
        $pkg->_logmsg( "Warm starting the filer", 'debug' );
    }

    my ( $cmd, $vol_type, $filer_args );
    my @opts;
    my %fields = $fh->param_fields();

    push( @opts, "FILER_GATEWAY=$fields{GATEWAY}" );
    push( @opts, "FILER_DNS=Y" );
    push( @opts,
              "FILER_DNS_SERVER_LIST="
            . $fields{MGMT_DNS} . ","
            . $fields{MGMT_DNS2} );
    push( @opts, "FILER_NEWFS_COMMAND=$cmd" );
    defined($main::FILER_ROOT_VOL_TYPE)
        ? push( @opts, "FILER_ROOT_VOL_TYPE=$main::FILER_ROOT_VOL_TYPE" )
        : push( @opts, "FILER_ROOT_VOL_TYPE=f" );

    defined( $fields{ "IFC_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "IFC_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_IFC_" . $fields{DATA_PORT1} . "=" . $fields{DATA_PORT1} );
    defined( $fields{ "IP_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "IP_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_IP_" . $fields{DATA_PORT1} . "=" . $fields{DATA_IP1} );
    defined( $fields{ "MASK_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "MASK_" . $fields{DATA_PORT1} } )
        : push( @opts,
        "FILER_MASK_" . $fields{DATA_PORT1} . "=" . $fields{DATA_MASK1} );
    defined( $fields{ "PARTNER_" . $fields{DATA_PORT1} } )
        ? push( @opts,
              "FILER_PARTNER_"
            . $fields{DATA_PORT1} . "="
            . $fields{ "PARTNER_" . $fields{DATA_PORT1} } )
        : undef;

    defined( $fields{ "IFC_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "IFC_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_IFC_" . $fields{DATA_PORT2} . "=" . $fields{DATA_PORT2} );
    defined( $fields{ "IP_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "IP_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_IP_" . $fields{DATA_PORT2} . "=" . $fields{DATA_IP2} );
    defined( $fields{ "MASK_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "MASK_" . $fields{DATA_PORT2} } )
        : push( @opts,
        "FILER_MASK_" . $fields{DATA_PORT2} . "=" . $fields{DATA_MASK2} );
    defined( $fields{ "PARTNER_" . $fields{DATA_PORT2} } )
        ? push( @opts,
              "FILER_PARTNER_"
            . $fields{DATA_PORT2} . "="
            . $fields{ "PARTNER_" . $fields{DATA_PORT2} } )
        : undef;

    defined( $fields{ "IFC_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "IFC_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{CLUSTER_PORT1} );
    defined( $fields{ "IP_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "IP_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
        "FILER_IP_" . $fields{CLUSTER_PORT1} . "=" . $fields{CLUSTER_IP1} );
    defined( $fields{ "MASK_" . $fields{CLUSTER_PORT1} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{ "MASK_" . $fields{CLUSTER_PORT1} } )
        : push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT1} . "="
            . $fields{CLUSTER_MASK1} );

    defined( $fields{ "IFC_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "IFC_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
              "FILER_IFC_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{CLUSTER_PORT2} );
    defined( $fields{ "IP_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "IP_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
        "FILER_IP_" . $fields{CLUSTER_PORT2} . "=" . $fields{CLUSTER_IP2} );
    defined( $fields{ "MASK_" . $fields{CLUSTER_PORT2} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{ "MASK_" . $fields{CLUSTER_PORT2} } )
        : push( @opts,
              "FILER_MASK_"
            . $fields{CLUSTER_PORT2} . "="
            . $fields{CLUSTER_MASK2} );

    defined( $fields{ "IFC_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_IFC_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "IFC_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_IFC_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_PORT} );
    defined( $fields{ "IP_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_IP_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "IP_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_IP_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_IP} );
    defined( $fields{ "MASK_" . $fields{MGMT_PORT} } )
        ? push( @opts,
              "FILER_MASK_"
            . $fields{MGMT_PORT} . "="
            . $fields{ "MASK_" . $fields{MGMT_PORT} } )
        : push( @opts,
        "FILER_MASK_" . $fields{MGMT_PORT} . "=" . $fields{MGMT_MASK} );

    eval {
        $pkg->_logmsg( "Calling _basic_setup", "comment" );
        $pkg->_basic_setup(
            'prefix'          => "FILER",
            'issue_setup_cmd' => 1,
            'timeout'         => 560,
            'host_obj'        => $fh
        );
    };
    if ($@) {
        if ( $@ =~ /Err=TIMEOUT/ ) {
            $pkg->_logmsg(
                "Calling basic_setup again after poking the console",
                'debug' );
            $pkg->_basic_setup(
                'prefix'          => "FILER",
                'issue_setup_cmd' => 0,
                'timeout'         => 60,
                'host_obj'        => $fh
            ) or die("basic setup failed for a second time. $@");
        }
        else {
            die("basic setup failed for something other than a timeout. $@");
        }
    }
    $transit_obj->change_state( to => 'CLI' );
    $pkg->_set_security_defaults();

    my $filer_pw = $fh->default_password();

    if ( $filer_pw ne '' ) {
        my $cmd = "passwd";
        $apiset_obj->execute_raw_command(
            'command' => $cmd,
            'connectrec-match_table' =>
                [ 'New password:|Retype new password:' => $filer_pw ]
        );
    }

    $pkg->_license( [ 'nfs', 'cifs', 'http' ] );
}
1;
