# $Id: //depot/prod/test/Rsierranevada/lib/WaflironLibrary/AccessSyslog.pm#1 $

#
# Copyright (c) 2010 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary This module provides methods to process logs on the filer.
## @author chethanm@netapp.com
## @status Public

=head1 NAME

WaflironLibrary::AccessSyslog

=head1 DESCRIPTION

This library provides methods to access the syslog files on the filer using
BSD interface. Syslog files are accessed using the BSD APISet and mainly uses
sed and unix commands to process the logs.

=head2 EXCEPTIONS USED

=over

=item WaflironLibrary::Exceptions::AccessSyslogException

An exception for when any failure related to processing the syslog files.

=item See Also

Package inherits from the BaseException module in NATE.

=back

=cut

package WaflironLibrary::AccessSyslog;

use WaflironLibrary::Exceptions::AccessSyslogException;

use Params::Validate
    qw(validate_with SCALAR ARRAYREF HASHREF BOOLEAN UNDEF OBJECT);
use NACL::APISet::Version;
use Tharn qw(sleep $Log subtest host);
use Data::Dumper;

=head1 NAME

new()

=head2 DESCRIPTION

Creates an object of the AccessSyslog Class. Also creates a BSD APISet object.

=head2 PARAMETERS

=over

=item C<< conn_obj >>

Type: Connectrec Object    Optional:No

Connection object of the filer


=item C<< log_file >>

Type:String    Optional:No

Name of the syslog file

=back

=head2 SYNOPSIS

my $syslog_obj = new WaflironLibrary::AccessSyslog(
            'conn_obj' => $conn_obj,
			'log_file' => $messages_file
);

=back

=cut

use Class::MethodMaker [
    new => [ '-hash', '-init', 'new' ],
    scalar => [ { -type => 'NACL::APISet' }, 'api_bsd' ],
    scalar => [ { -type => 'Connectrec' },   'conn_obj' ],
    scalar => 'line_no',
    scalar => 'log_file',
    scalar => 'mode',
    scalar => 'ontap_release',
    scalar => 'mount_root',
    scalar => [ {-default => 7 }, 'wait_time']
];

sub init {
    my ( $self, %opts ) = @_;
    validate_with(
        params => \%opts,
        spec   => {
            'conn_obj' => { isa  => 'Connectrec' },
            'log_file' => { type => SCALAR, optional => 0 },
            'line_no'  => { type => SCALAR, optional => 1 },
        },
       allow_extra => 1
    );
    my $host_type = $self->conn_obj->hostp->hosttype();
    my $mode = ($host_type =~ /ng-vsim|ng-cfe/) ? "CMode" : "7Mode";
    $self->mode($mode);
    $self->_get_ontap_release();
    if ( $self->ontap_release() !~ /IronCity/ ) {
        $self->api_bsd(
            NACL::APISet->new(
                hostobj   => $self->conn_obj->hostp(),
                category  => "Node",
                interface => "CLI",
                set       => "Systemshell",
            )
        );
    }
    else {
        require "FS/Util/Setup.pm";
        my $mount_root = FS::Util::Setup::mount_root_vol(
            $self->conn_obj->hostp(), undef, undef,
            filer_conn => "#rsh",
            client     => "localhost",
            timeout    => 100,
            cmdtimeout => 120
        ) or croak($@);
        my $file_name = $mount_root . $self->log_file();
        $self->mount_root($mount_root);
        $self->log_file($file_name);
    }

}

=head1 NAME

syslog_get_mark()

=head2 DESCRIPTION

Updates the object with the line mark from where we need the logs lines from.
The same is used when we tail the log, it tails from that line till the end of
file.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=back

=item C<< timeout >>

Type:Number   Optional:Yes

Timeout used while getting the line count. Helps in the cases where the file size is large.

-back

=head2 SYNOPSIS

$acces_syslog_obj->syslog_get_mark();

=back

=cut

sub syslog_get_mark {
    my ( $pkg, %opts ) = @_;
    my $log_file = $opts{'log_file'} || $pkg->log_file();
   my $timeout = $opts{'timeout'} || 120;
    if ( !defined $log_file ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "Syslog file is mandatory!");
    }
    my $cmd = "wc -l $log_file";
    my $log_mark;
    my $raw_output;
    if ( $pkg->ontap_release() eq "IronCity" ) {
        {
            open( CMD_OUT, "$cmd 2>&1|" )
                or die "Can't open the file handle for the $cmd. $!";
            local $/ = undef;
            $raw_output = <CMD_OUT>;
            close CMD_OUT;
        }
    }
    else {
        my $apiset_obj = $pkg->api_bsd();
        $raw_output = $apiset_obj->execute_raw_command( 'command' => $cmd, 'connectrec-timeout' => $timeout );
    }
    if ( $raw_output =~ /\s*(\d+?)\s/ ) {
        $log_mark = $1;
    }
    else {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "$cmd Failed!");
    }
    $pkg->line_no($log_mark);
    $pkg->log_file($log_file);
}

=head1 NAME

syslog_tail()

=head2 DESCRIPTION

Tails the log file from the marked line till end of file.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=item C<< line_no >>

Type:String    Optional:Yes

Line number from where we need tail the log file.

=back

=head2 SYNOPSIS

my $log_lines = $acces_syslog_obj->syslog_tail();

=back

=head2 RETURNS

=over

Returns the lines from the log file, from the marked line number till end
of file.

=back

=cut

sub syslog_tail {
    my ( $pkg, %opts ) = @_;
    my $log_file  = $opts{'log_file'} || $pkg->log_file();
    my $line_mark = $opts{'line_no'}  || $pkg->line_no();
    my $sleep_time = $opts{'wait_time'} || $pkg->wait_time();
    if ( !defined $log_file ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "Syslog file is mandatory!");
    }
    if ( !$line_mark ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "line_no is a mandatory !!");
    }
    $pkg->syslog_get_mark();
    my $cmd = "sed -n '$line_mark,\$p' <$log_file";
    my $raw_output;
    sleep $sleep_time;
    if ( $pkg->ontap_release() eq "IronCity" ) {
        {
            open( CMD_OUT, "$cmd 2>&1|" )
                or die "Can't open the file handle for the $cmd. $!";
            local $/ = undef;
            $raw_output = <CMD_OUT>;
            close CMD_OUT;
        }
    }
    else {
        my $apiset_obj = $pkg->api_bsd();
        $raw_output = $apiset_obj->execute_raw_command( 'command' => $cmd );
    }

    return $raw_output;
}

=head1 NAME

syslog_dump()

=head2 DESCRIPTION

Returns the content of the whole file.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=back

=head2 SYNOPSIS

my $log_lines = $acces_syslog_obj->syslog_dump();

=back

=head2 RETURNS

=over

Returns the all the lines from the log file.

=back

=cut

sub syslog_dump {
    my ( $pkg, %opts ) = @_;
    my $log_file = $opts{'log_file'} || $pkg->log_file();
    if ( !defined $log_file ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "Syslog file is mandatory!");
    }
    my $cmd = "cat $log_file";
    my $raw_output;
    if ( $pkg->ontap_release() eq "IronCity" ) {
        {
            open( CMD_OUT, "$cmd 2>&1|" )
                or die "Can't open the file handle for the $cmd. $!";
            local $/ = undef;
            $raw_output = <CMD_OUT>;
            close CMD_OUT;
        }
    }
    else {
        my $apiset_obj = $pkg->api_bsd();
        $raw_output = $apiset_obj->execute_raw_command( 'command' => $cmd );
    }

    return $raw_output;
}

=head1 NAME

syslog_exists()

=head2 DESCRIPTION

Checks if the given file exists? If the file exists returns 1, else returns 0.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=back

=head2 SYNOPSIS

if($acces_syslog_obj->syslog_exists()) {
		# process the file
}

=back

=head2 RETURNS

=over

Returns 1 if the given file exists, else returns 0.

=back

=cut

sub syslog_exists {
    my ( $pkg, %opts ) = @_;
    my $log_file = $opts{'log_file'} || $pkg->log_file();
    if ( !defined $log_file ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "Syslog file is mandatory!");
    }
    my $cmd = "ls $log_file";
    my $raw_output;
    if ( $pkg->ontap_release() eq "IronCity" ) {
        {
            open( CMD_OUT, "$cmd 2>&1|" )
                or die "Can't open the file handle for the $cmd. $!";
            local $/ = undef;
            $raw_output = <CMD_OUT>;
            close CMD_OUT;
        }
    }
    else {
        my $apiset_obj = $pkg->api_bsd();
        $raw_output = $apiset_obj->execute_raw_command( 'command' => $cmd );
    }

    if ( $raw_output =~ /No such file or directory/i ) {
        return 0;
    }
    else {
        $pkg->log_file($log_file);
        return 1;
    }
}

=head1 NAME

syslog_create()

=head2 DESCRIPTION

Creates an empty syslog flie with the given name.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=back

=head2 SYNOPSIS

if($acces_syslog_obj->syslog_exists()) {
		$acces_syslog_obj->syslog_create('log_file' => '/mroot/etc/mlog');
}

=back

=head2 RETURNS

=cut

sub syslog_create {
    my ( $pkg, %opts ) = @_;
    my $log_file = $opts{'log_file'} || $pkg->log_file();
    if ( !defined $log_file ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "Syslog file is mandatory!");
    }
    my $apiset_obj = $pkg->api_bsd();
    my $cmd        = "sudo touch $log_file";
    my $raw_output;
    if ( $pkg->ontap_release() eq "IronCity" ) {
        {
            open( CMD_OUT, "$cmd 2>&1|" )
                or die "Can't open the file handle for the $cmd. $!";
            local $/ = undef;
            $raw_output = <CMD_OUT>;
            close CMD_OUT;
        }
    }
    else {
        my $apiset_obj = $pkg->api_bsd();
        $raw_output = $apiset_obj->execute_raw_command( 'command' => $cmd );
    }
    if ( $raw_output =~ /No such file or directory/ ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "$cmd : Failed!");
    }
    return 1;
}

=head1 NAME

check_pattern()

=head2 DESCRIPTION

Checks if the given pattern exists in the given multiline text string.

=head2 PARAMETERS

=over

=item C<< log_file >>

Type:String    Optional:Yes

Name of the syslog file

=back

=head2 SYNOPSIS

if($acces_syslog_obj->check_pattern('text' => $text, 'pattern' => $pattern)) {
		$Log->comment("Pattern found!")
}

=back

=head2 RETURNS

   Returns 1 if the pattern is found else returns 0

=cut

sub check_pattern {
    my ( $pkg, %opts ) = @_;
    my $text    = $opts{'text'};
    my $pattern = $opts{'pattern'};
    if ( !defined $text || !defined $pattern ) {
        WaflironLibrary::Exceptions::AccessSyslogException->throw(
            "text string and pattern are mandatory!");
    }
    if ( $text =~ /$pattern/mi ) {
        return 1;
    }
    return 0;
}

# Figure out the ONTAP version

sub _get_ontap_release {
    my ( $pkg, %opts ) = @_;
    my $category   = "Node";
    my $conn       = $pkg->conn_obj();
    my $shell_type = $pkg->mode();

    my $version_manager = NACL::APISet::Version->new(
        hostobj    => $conn->hostp(),
        category   => "Node",
        shell_type => $shell_type
    );
    my $release
        = $version_manager->get_version_attribute( attribute => "release" );
    my $mapped_release
        = $version_manager->map_release_to_codename( release => $release );
    if ( $mapped_release =~ /rollingrock/ ) {
        $pkg->ontap_release("RollingRock");
    }
    elsif ( $mapped_release =~ /boilermaker/ ) {
        $pkg->ontap_release("Boilermaker");
    }
    elsif ( $mapped_release =~ /sierranevada/ ) {
        $pkg->ontap_release("SierraNevada");
    }
    elsif ( $mapped_release =~ /fullsteam/ ) {
        $pkg->ontap_release("Fullsteam");
    }
    elsif ( $mapped_release =~ /longboard/ ) {
        $pkg->ontap_release("LongBoard");
    }
    else {
        $pkg->ontap_release("IronCity");
    }
}

#sub DESTROY {
#    my $self = shift;
#    $Log->comment("Destroying the AccessSyslog Object...");
#    if ( $self->ontap_release() eq "IronCity" ) {
#        my $mount_root = $self->mount_root();
#        $Log->comment( "Mount root is: " . $mount_root );
#        require "mount.pm";
#        mount::umount($mount_root);
#    }
#}

1;
