# $Id: //depot/prod/test/Rsierranevada/lib/WinClient/Disk.pm#1 $
#
# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.
#
# This library creates disk objects which are then used by IO tools.
# This lib also handles partitioning disks, converting them, formatting
# them, and mounting.
#
# TODO: Complete documentation pending.
#
## @summary
##  Wrapper for WinClient disk functionality.
##
## @author  kenny speer
## @status  private
##

use strict;

package WinClient::Disk;

# define the routines used in this package
sub new;                # creates new instance of object
sub convert;            # converts disk to/from mbr/gpt
sub partition;          # partitions disks
sub assign;             # assigns drive letter or mountpoint
sub remove;             # removes drive letter or mountpoint
sub format;             # formats disk
sub chkdsk;             # runs chkdsk on filesystem
sub getStats;           # gets performance stats for disks
sub startIO;            # starts IO on disk
sub empty;              # remove all files from a disk
sub destroy;            # destroys disk configuration information
sub getDiskInfo;        # populates disk object info
sub setLBPolicy;        # sets the load balancing policy

use Error;
use WinClient::Util;
use WinClient::DT;
use WinClient::SIO;
use File::Temp;
use File::Basename;

use vars qw( @ISA @EXPORT );
@ISA = qw( 
          Error 
          WinClient::Util 
          WinClient::DT 
          WinClient::SIO 
          );

#
#
#
sub new 
{
    my ($class, %args) = @_;
    my $this = {};

    my ($msg, $device, $host, $timeout);

    # bless this object
    bless $this, $class;

    # initially clear errors
    $this->_clear_err();

    foreach (keys %args)
    {
        if (/^-?device$/)
        {
            if (!length($device = $args{$_}))
            {
                $msg = "Disk::new Invalid value for -device: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
            }
            $device =~ s/\s//g;
            next;
        }
        if (/^-?host$/)
        {
            if (!length($host = $args{$_}))
            {
                $msg = "Disk::new Invalid value for -host: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
            }
            next;
        }
        if (/^-?timeout$/)
        {
            if (!length($timeout = $args{$_}))
            {
                $msg = "Disk::new Invalid value for -timeout: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
            }
            next;
        }
    }

    $timeout = 300 if (!defined($timeout));

    # set globals
    $this->{host} = $host;
    $this->{connection} = $this->{host}->{connection};

    # clear any host errors
    $this->_clear_err();

    if (!defined($device))
    {
        $msg = "Disk::new - device not defined for disk.";
        &main::logcomment($msg);
        $this->_set_err($msg);
    }

    # set disk properties
    $this->{rawName}  = "\\\\.\\PHYSICALDRIVE" . $device;
    $this->{device}   = $device;
    $this->{timeout}  = $timeout;
    $this->{protocol} = $this->{host}->{protocol};

    # properties only available via Drexel+ using deviceinfo
    $this->{policy}   = undef;
    $this->{numPaths} = undef;
    $this->{serial}   = undef;
    $this->{cfmode}   = undef;
    $this->{filer}    = undef;
    $this->{lunPath}  = undef;
    $this->{lunType}  = undef;

    # assume some disk properties until proven otherwise
    $this->{isClean}   = 0;
    $this->{isParted}  = 1; 

    # volume and mount info
    $this->{lun}       = undef;
    $this->{lunSize}   = undef;
    $this->{lunState}  = undef;
    $this->{boot}      = undef;
    $this->{vendor}    = undef;
    $this->{partStyle} = undef;
    $this->{diskType}  = undef;
    $this->{mount}     = undef;
    $this->{fsType}    = undef;
    $this->{fsLabel}   = undef;
    $this->{sig}       = undef;

    # undefined as of yet
    $this->{volume}  = undef;

new_EXIT:

    return $this;

} # END new

#
#
#
sub convert 
{
    my ($this, %args) = @_;

    # initially clear errors
    $this->_clear_err();

    my $timeout = $this->{timeout};
    my $device = $this->{device};
    
    my ($status, @output, $partStyle, $msg);

    # parse arguments, we only take one
    foreach (keys %args)
    {
        if (/-?partStyle/i)
        {
            if (!length($partStyle = $args{$_}))
            {
                $msg = "Disk::convert - Invalid value for -partStyle: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto convert_EXIT;
            }
        }
        last;
    }

    if ($partStyle !~ /gpt|mbr/i)
    {
        $msg = "Disk::convert - Invalid type for partStyle: $partStyle";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto convert_EXIT;
    }

    if ($this->{partStyle} =~ /$partStyle/i)
    {
        &main::logcomment("Disk::convert - Disk already type: $partStyle");
        goto convert_EXIT;
    }

    if (($this->{host}->{hostver} =~ /windows 2000/i) && ($partStyle =~ /gpt/i))
    {
        &main::logcomment("Windows 2000 does not support GPT style partitions.\n" .
            "Continueing with MBR style partitions.");
        goto convert_EXIT;
    }

    &main::logcomment("Converting disk $device to $partStyle");

    # we cannot have volumes or partitions, disk must be empty
    # don't clean since that's too drastic
    $this->destroy() if ($this->{isParted});

    ($status, @output) = $this->_diskpartCmd(
            timeout  => $timeout,
            commands => [("select disk $device",
                          "convert $partStyle",
                          "exit")],
            comment  => "Converting disk $device to $partStyle");
    
    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::convert - Diskpart cmd 'convert $partStyle' failed.");
        $this->_set_err(@output);
        goto convert_EXIT;
    }


convert_EXIT:

    return;
 
} # END convert

#
#
#
sub partition
{
    my ($this, $align) = @_;
    
    # initially clear errors
    $this->_clear_err();

    my $timeout = $this->{timeout};
    my $device = $this->{device};
    my $partition = $this->{isParted};
    my $volume = $this->{volume};

    my ($status, @output, $line, @line);
    
    # set the alignment value if it was passed in
    if (not defined $align) {
        $align = "";
    } else {
        $align = "align=$align";
    }
    
    if ($partition)
    {
        &main::logcomment("Disk::partition - disk ($device) contains partition ($partition).");

        if (defined($volume))
        {
            &main::logcomment("Disk::partition - disk ($device) contains volume ($volume).");
        }

        goto partition_EXIT;
    }
    else
    {
        # 2008 *sometimes* has the disks offline by default, so online them here, otherwise use old method
        # of configuring the disks.
        if ($this->{host}->{hostver} =~ /2008/ or
            $this->{host}->{hostver} =~ /windows server 7/ or
	    $this->{host}->{hostver} =~ /2012/)
        {
            ($status, @output) = $this->_diskpartCmd(
                    timeout  => $timeout,
                    commands => [("select disk $device",
                                  "attributes disk clear readonly",
                                  "online disk",
                                  "exit")],
                    comment  => "Disk::partition - Set device online $device");

            ($status, @output) = $this->_diskpartCmd(
                    timeout  => $timeout,
                    commands => [("select disk $device",
                                  "attributes disk clear readonly",
                                  "create partition primary $align",
                                  "list partition",
                                  "detail disk",
                                  "exit")],
                    comment  => "Disk::partition - Creating primary partition on disk $device");
        }
        else
        {
            ($status, @output) = $this->_diskpartCmd(
                    timeout  => $timeout,
                    commands => [("select disk $device",
                                  "create partition primary $align",
                                  "list partition",
                                  "detail disk",
                                  "exit")],
                    comment  => "Disk::partition - Creating primary partition on disk $device");
        }
     

        if ((grep(/DiskPart was unable to create the specified partition/, @output)) ||
            (grep(/DiskPart has encountered an error: The media is write protected/, @output)))
        {
            &main::logcomment("Disk::partition - Failed to create partition." .
                "\nCapture disk info via getDiskInfo.");
            $this->getDiskInfo();

            if (0 == $this->{isParted})
            {
                &main::logcomment("Disk::partition - Disk cannot be partitioned.");
                $this->_set_err(@output);
                goto partition_EXIT;
            }
        }
        elsif ($status != $Host::TRUE)
        {
            &main::logcomment("Disk::partition - Diskpart command failed.");
            $this->_set_err(@output);
            goto partition_EXIT;
        }

        foreach $line (@output)
        {

           # Get the Disk ID first, its on its own line
           if ($line =~ /Disk\sID:\s(\S+)/) {
                $this->{sig} = $1;
                next;
            }

            # split the line
            @line = split(/\s+/, $line);

            # check for partitions and verify whether gpt or mbr and
            # dynamic or basic disks
            if (($line[1] =~ /^Partition$/i) && ($line[2] =~ /^\d+$/))
            {
                $this->{isClean} = 0;

                if ($line[3] =~ /Dynamic/)
                {
                    $this->{diskType} = "dynamic";
                    &main::logcomment("Disk::partition - Cannot create primary partitions on dynamic disks");
                    $this->_set_err(@output);
                    goto partition_EXIT;
                }
                elsif ($line[3] =~ /^Reserved$/)
                {
                    $this->{partStyle} = "gpt";
                }
                elsif ($line[3] =~ /^Primary$/)
                {
                    # add the partition number of the primary partition
                    $this->{isParted} = $line[2];
                }
            }
            elsif (($line[1] =~ /^Volume$/i) && ($line[2] =~ /^\d+$/))
            {
                $this->{volume} = $line[2];
            }
        }
    }

partition_EXIT:

    return;
   
} # END partition

#
#
#
sub online
{
    my ($this, %args) = @_;

    my $timeout = $this->{timeout};
    my $device = $this->{device};

    # initially clear errors
    $this->_clear_err();

    &main::logcomment("Disk::online - onlining disk ($device).");

    my ($status, @output) = $this->_diskpartCmd(
                            timeout  => $timeout,
                            commands => [("select disk $device",
                                          "attributes disk clear readonly",
                                          "online disk",
                                          "exit")],
                            comment  => "Disk::online - Set device online $device");

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::online - Diskpart (online) command failed.");
        $this->_set_err(@output);
    }

    return;

}

#
#
#
sub assign
{
    my ($this, %args) = @_;

    # initially clear errors
    $this->_clear_err();

    my $timeout = $this->{timeout};
    my $device = $this->{device};
    my $partition = $this->{isParted};

    my ($mountpoint, $msg, $status, @output, $type);

    foreach (keys %args)
    {
        if (/^-?mountpoint$/)
        {
            if (!length($mountpoint = $args{$_}))
            {
                $msg = "Disk::assign - invalid value for -mountpoint: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto assign_EXIT;
            }
            next;
        }
        if (/^-?type$/)
        {
            if (!length($type = $args{$_}))
            {
                $msg = "Disk::assign - invalid value for -type: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto assign_EXIT;
            }
            next;
        }
    }

    # default to using drive letters if the user didn't specify
    $type = "letter" if (!defined($type));

    # don't do anything, drive already mounted
    if (!defined($mountpoint) && $this->{mount} =~ /^[A-Z]:?/i)
    {
        &main::logcomment("Disk::assign - device $device already assigned $this->{mount}");
        &main::logcomment("Disk::assign - Use remove() to first to change the mount type");
        goto assign_EXIT;
    }

    # user defined a mount and the device is already mounted, if they
    # don't match, then remove the current map and fall through to the
    # assignment for currently unmounted devices
    if (defined($mountpoint) && $this->{mount} != undef)
    {
        if ($this->{mount} != $mountpoint)
        {
            &main::logcomment("Disk::assign - User requested mount '$mountpoint' not the same as current mount '$this->{mount}'");
            $this->remove();
        }
        else
        {
            &main::logcomment("Disk::assign - Device already assigned requested mount point '$mountpoint'");
            goto assign_EXIT;
        }
    }

    # drive not mounted, and user has no preference, use next available
    # drive letter from array or create mount point
    if (!defined($mountpoint) && $this->{mount} == undef)
    {
        if ($type =~ /letter/)
        {
            # Windows 2000 won't allow A|B as they are reserved for
            # removable media (not sure of side effects of using)
            while (!defined($mountpoint) || $mountpoint =~ /^[A-C]$/)
            {
                # avoid a race condition where we've lost track of the
                # used and available drive letters
                if (!defined($this->{host}->{availLetters}) ||
                    !defined($this->{host}->{usedLetters}) ||
                    (@{$this->{host}->{availLetters}} == 0))
                {
                    $msg = "\nERROR\n" .
                        "No available or used drive letters defined.\n" .
                        "Please execute getDevices() or remove the mount points";
                    &main::logcomment($msg);
                    $this->_set_err($msg);
                    goto assign_EXIT;
                }

                $mountpoint = shift(@{$this->{host}->{availLetters}});
                push(@{$this->{host}->{usedLetters}}, $mountpoint);
            }
        }
        elsif ($type =~ /mount/)
        {
            # verify the mount dir exists
            $mountpoint = $Host::MOUNT . "\\Disk" . $device;
            $this->execute("mkdir $mountpoint", $this->{timeout});
        }
    }
    elsif (defined($mountpoint) && $this->{mount} == undef)
    {
        if ($type =~ /letter/)
        {
            # drive letter assignments don't have :
            $mountpoint =~ s/://;
            if (grep(/$mountpoint/i, @{$this->{host}->{usedLetters}}) > 0)
            {
                $msg = "Disk::assign - Requested drive letter '$mountpoint' already assigned to another device";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto assign_EXIT;
            }
        }
    }
    else
    {
        $msg = "Disk::assign - SEVERE ERROR: unsure how we got here, but this needs to be fixed, logic does not compute.";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto assign_EXIT;
    }

    # now assign the mountpoint to the device
    ($status, @output) = $this->_diskpartCmd(
            timeout  => $timeout,
            commands => [("select disk $device",
                          "select partition $partition",
                          "assign $type=$mountpoint",
                          "exit")],
            comment  => "Assigning $type=$mountpoint to disk $device.");
        
    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::assign - Diskpart cmd 'assign $type=$mountpoint' failed.");
        $this->_set_err(@output);
        goto assign_EXIT;
    }

    $this->{mount} = $mountpoint;

assign_EXIT:

    return;

} # END assign

#
#
#
sub remove
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my $timeout = $this->{timeout};
    my $device = $this->{device};
    my $partition = $this->{isParted};

    my ($status, @output, $msg);

    goto remove_EXIT if (!defined($this->{mount}));

    ($status, @output) = $this->_diskpartCmd(
            timeout  => $timeout,
            commands => [("select disk $device",
                          "select partition $partition",
                          "remove",
                          "exit")],
            comment  => "Removing mountpoint from disk $device.");
    
    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::remove - Diskpart cmd 'remove' failed.");
        $this->_set_err(@output);
        goto remove_EXIT;
    }

    $this->{mount} = undef;

remove_EXIT:

    return;
} # END remove

#
#
#
sub format
{
    my ($this, %args) = @_;

    # initially clear errors
    $this->_clear_err();

    my $device = $this->{device};
    my $protocol = $this->{protocol};
    my $mount = $this->{mount};

    # add the : if this is a drive letter
    $mount .= ":" if ($mount =~ /^[A-Z]$/);

    my ($status, @output, $msg, $cmd);
    my ($type, $blockSize, $fragmentSize, $size, $label);

    my $tmpname = File::Temp::tmpnam();
    $tmpname = File::Basename::basename($tmpname);

    my $remFile = "$Host::TEMP\\formatInput_${tmpname}.txt";
    # delete since previous run might contain incorrect data
    $this->deleteRemoteFile($remFile);

    foreach (keys %args)
    {
        if (/^-?type$/i)
        {
            $type = $args{$_};
            if (!length($type) || $type !~ /^(?:NTFS|FAT32|REFS)$/i)
            {
                $msg = "Disk::format - invalid value for -type: $args{$_}";      
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto format_EXIT;
            }
            next;
        }
        if (/^-?blockSize$/i)
        {
            if (!length($blockSize = $args{$_}))
            {
                $msg = "Disk::format - invalid value for -blockSize: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto format_EXIT;
            }
            next;
        }
        if (/^-?fragmentSize$/i)
        {
            if (!length($fragmentSize = $args{$_}))
            {
                $msg = "Disk::format - invalid value for -fragmentSize: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto format_EXIT;
            }
            next;
        }
        if (/^-?size$/i)
        {
            if (!length($size = $args{$_}))
            {
                $msg = "Disk::format - invalid value for -size: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto format_EXIT;
            }
            next;
        }
        if (/^-?label$/i)
        {
            if (!length($label = $args{$_}))
            {
                $msg = "Disk::format - invalid value for -label: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto format_EXIT;
            }
            next;
        }
    }

    if (!defined($label))
    { 
        $label = "NATE_F_" . $device . "_" . $mount if ($protocol =~ /fcp/i);
        $label = "NATE_I_" . $device . "_" . $mount if ($protocol =~ /iscsi/i);
        $label = "NATE_D_" . $device . "_" . $mount if ($protocol =~ /none/i);
    }

    # default to NTFS
    $type = "NTFS" if (!defined($type));

    # create command and include responses from file
    $cmd = "format $mount /FS:$type /Q ";
    $cmd .= " < $remFile";

    &main::logcomment("Disk::format - executing command: $cmd");

    # create remote file
    $this->createRemoteFile($remFile, "Y", $label);

    # execute format command
    ($status, @output) = $this->execute($cmd, 900);

    # first we try once more in case we're asked for the current label
    if (grep(/An incorrect volume label was entered/, @output))
    {
        $this->deleteRemoteFile($remFile);
        $this->createRemoteFile($remFile, $this->{fsLabel}, "Y", $label);

        # execute format command
        ($status, @output) = $this->execute($cmd, $this->{timeout});

        # we need to fail after the second attent at a drive letter
        if (grep(/incorrect volume label was entered/, @output) or
            $status != $Host::TRUE)
        {
            &main::logcomment("Disk::format - Command failed: $cmd");
            $this->_set_err(@output);
            goto format_EXIT;
        }
    }
    elsif ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::format - Command failed: $cmd");
        $this->_set_err(@output);
        goto format_EXIT;
    }

    # format was successful; set some properties
    $this->{fsType} = $type;
    $this->{fsLabel} = $label;

    # if this is iscsi, then bind the LUNs
    if ($protocol =~ /iscsi/i)
    {
        $cmd = "$this->{host}->{cmd_iscsicli} BindPersistentVolumes";
        ($status, @output) = $this->execute($cmd, $this->{timeout});

        if ($status != $Host::TRUE)
        {
            &main::logcomment("Disk::format - Command failed: $cmd");
            $this->_set_err(@output);
            goto format_EXIT;
        }
    }

format_EXIT:

    $this->deleteRemoteFile($remFile);

    return;
} # END format

#
#
#
sub chkdsk
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my $device = $this->{device};
    my $mount = $this->{mount};

    $mount .= ":" if ($mount =~ /^[A-Z]$/);

    my ($status, @output, $cmd);

    $cmd = "chkdsk $mount"; 

    &main::logcomment("Disk::chkdsk - Executing chkdsk on disk $device, $mount");
    
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::chkdsk - Command failed: $cmd");
        $this->_set_err(@output);
        goto chkdsk_EXIT;
    }

chkdsk_EXIT:

    return;

} # END chkdsk

#
#
#
sub getStats
{
} # END getStats

#
#
#
sub startIO
{
    my ($this, %args) = @_;

    # initially clear errors
    $this->_clear_err();

    my ($msg, $program, $iotype, $status, @output);
    my ($runLength, $fileSize, $blockSize);

    foreach (keys %args)
    {
        if (/^-?program$/i)
        {
            if (!length($program = $args{$_}) || $program !~ /^(?:dt|sio)$/i)
            {
                $msg = "Disk::startIO - invalid value for -program: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startIO_EXIT;
            }
            next;
        }
        if (/^-?iotype$/i)
        {
            if (!length($iotype = $args{$_}) || $iotype !~ /^(?:fs|raw)$/i)
            {
                $msg = "Disk::startIO - invalid value for -iotype: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startIO_EXIT;
            }
            next;
        }
    }

    # set the location properly if this is raw or mount
    if ($iotype =~ /^raw$/i)
    {
        $args{"-location"} = $this->{rawName};
    }
    else
    {
        if ($this->{mount} !~ /^[A-Z]:?/)
        {
            $msg = "Disk::startIO - invalid mount point: $this->{mount}";
            &main::logcomment($msg);
            $this->_set_err($msg);
            goto startIO_EXIT;
        }
        else
        {
            if ($this->{mount} =~ /^[A-Z]$/)
            {
                $args{"-location"} = $this->{mount} . ":" . "\\nate_test";
            }
            else
            {
                $args{"-location"} = $this->{mount} . "\\nate_test";
            }
        }
    }


    $this->_startDT(%args) if ($program =~ /^dt$/i);
    $this->_startSIO(%args) if ($program =~ /^sio$/i);

startIO_EXIT:
    
    return;

} # END startIO

#
#
#
sub empty
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my $target = $this->{mount} . ":\\*";

    my ($status, @output) = $this->deleteRemoteFile("/S $target");

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Disk::empty - Failed for Volume ($target):\n@output");
        $this->_set_err(@output);
        goto empty_EXIT;
    }

empty_EXIT:

    return;

} # END empty

#
#
#
sub destroy
{
    my ($this, %args) = @_;

    # initially clear errors
    $this->_clear_err();

    my $device = $this->{device};
    my $diskType = $this->{diskType};
    my $timeout = $this->{timeout};
    my $partition = $this->{isParted};

    my ($status, @output, $line, $msg);
    my (@newCmds, @failedVolumes, $failedVolume, @missingDisks, $missingDisk);

    if (!defined($device))
    {
        $msg = "Disk::destroy requires a device.";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto destroy_EXIT;
    }

    if ($this->{diskType} =~ /none/i)
    {
        &main::logcomment("Disk::destroy - Nothing to destroy on Disk $device.");
        goto destroy_EXIT;
    }

    # if clean is defined, this is a brutal cleaning of the disk
    ($status, @output) = $this->_diskpartCmd(
            timeout  => $timeout,
            commands => [("select disk $device",
                            "clean",
                            "exit")],
            comment  => "Cleaning all disk information from Disk $device"); 
    
    if ($status != $Host::TRUE)
    {
        &main::logcomment(
            "Disk::destroy - Diskpart command 'select disk $device' failed.");
        $this->_set_err(@output);
        goto destroy_EXIT;
    }

    # cleaning a dynamic disk can leave disks and volumes orphaned,
    # so clean them up too
    if ($diskType =~ /dynamic/i)
    {
        &main::logcomment(
            "Disk::destroy - This disk was dynamic.  Search for failed volumes.");

        ($status, @output) = $this->_diskpartCmd(
                timeout  => $timeout,
                commands => [("list volume", "exit")],
                comment  => "Searching for failed volumes ...");

        if ($status != $Host::TRUE)
        {
            &main::logcomment(
                "Disk::destroy - Diskpart command 'list volume' failed.");
            $this->_set_err(@output);
            goto destroy_EXIT;
        }

        foreach $line (@output)
        {
            if ($line =~ /^\S?\s+Volume\s(\d+)\s+.+Failed/)
            {
                push(@failedVolumes, $1);
            }
        }

        &main::logcomment(
            "Disk::destroy - Found failed volumes: @failedVolumes");

        # need to reverse sort the volumes, since Windows
        # dynamically renames them as you remove them
        @failedVolumes = reverse(@failedVolumes);
        
        # re-initialize the array of commands
        @newCmds = ();

        if (defined($failedVolumes[0]))
        {
            foreach $failedVolume (@failedVolumes)
            {
                push(@newCmds, "select volume $failedVolume", "delete volume");
            }
            push(@newCmds, "exit");

            ($status, @output) = $this->_diskpartCmd(
                    timeout  => $timeout,
                    commands => [(@newCmds)],
                    comment  => "Deleting failed volumes (@failedVolumes)");
            if ($status != $Host::TRUE)
            {
                &main::logcomment(
                    "Disk::destroy - Diskpart cmd 'delete volume' failed.");
                $this->_set_err(@output);
                goto destroy_EXIT;
            }
        }

        # search and destroy orphaned disks
        ($status, @output) = $this->_diskpartCmd(
                timeout  => $timeout,
                commands => [("list disk", "exit")],
                comment  => "Searching for orphaned disks ...");

        if ($status != $Host::TRUE)
        {
            &main::logcomment(
                "Disk::destroy - Diskpart command 'list disk' failed.");
            $this->_set_err(@output);
            goto destroy_EXIT;
        }

        foreach $line (@output)
        {
            if ($line =~ /^\S?\s+Disk\s([M]\d+)\s+Missing/)
            {
                push(@missingDisks, $1);
            }
        }

        &main::logcomment(
            "Disk::destroy - Found orphaned disks (@missingDisks).");

        # need to reverse sort this array since Windows dynamically
        # renames them as they are deleted
        @missingDisks = reverse(@missingDisks);

        # re-initialize the array of commands
        @newCmds = ();

        if (defined($missingDisks[0]))
        {
            foreach $missingDisk (@missingDisks)
            {
                push(@newCmds, "select disk $missingDisk", "delete disk");
            }
            push(@newCmds, "exit");

            ($status, @output) = $this->_diskpartCmd(
                    timeout  => $timeout,
                    commands => [(@newCmds)],
                    comment  => "Deleting orphaned disks (@missingDisks)");

            if ($status != $Host::TRUE)
            {
                &main::logcomment(
                    "Disk::destroy - Diskpart cmd 'delete disk' failed.");
                $this->_set_err(@output);
                goto destroy_EXIT;
            }
        }
    }

    # reset some properties, since this is no longer dynamic
    # clean indicates mbr, basic, gpt with no partitions
    $this->{isClean} = 1;
    $this->{diskType} = "basic";
    $this->{partStyle} = "mbr";
    $this->{isParted} = 0;
    $this->{mount} = undef;
    $this->{fsType} = undef;
    $this->{fsLabel} = undef;

destroy_EXIT:

    return;

} # END destroy 

#
# Offline this disk object with disk part
#
sub offline
{
    my ($this) = @_;

    $this->_clear_err();

    my $device = $this->{device};
    my $timeout = $this->{timeout};

    # offline this disk
    my ($status, @output) = $this->_diskpartCmd(
        timeout  => $timeout,
        commands => [("select disk $device",
                      "offline disk",
                      "exit")],
        comment  => "Disk::offline - Set device offline $device");
    if (grep(/The disk you specified is not valid\./i, @output) || $status != $Host::TRUE)
    {
        &main::logcomment("Disk::offline - Diskpart command failed.");
        $this->_set_err(@output);
        goto offline_EXIT;
    }

offline_EXIT:

    return;     
} # END offline

#
# DEPRECATED, replaced with mpc functionality
#
sub getDiskInfo
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my $device = $this->{device};
    my $timeout = $this->{timeout};

    my ($status, @output, $line, @line);

    #
    # get partition information and populate disk properties
    #

    ($status, @output) = $this->_diskpartCmd(
            timeout  => $timeout,
            commands => [("select disk $device",
                          "list partition",
                          "detail disk",
                          "exit")],
            comment  => "Listing partitions for disk $device");

    if (grep(/The disk you specified is not valid\./i, @output) || $status != $Host::TRUE)
    {
        &main::logcomment("Disk::getDiskInfo - Diskpart command failed.");
        $this->_set_err(@output);
        goto getDiskInfo_EXIT;
    }

    # since the above command worked, clear the disk properties so that
    # we can se them below; diskType and partStyle are set to defaults
    # and then overriden if needed
    $this->{isParted} = 0;
    $this->{diskType} = "basic";
    $this->{partStyle} = "mbr";
    $this->{fsType} = undef;
    $this->{fsLabel} = undef;
    $this->{mount} = undef;
       
    foreach $line (@output)
    {

        # split the line by spaces
        @line = split(/\s+/, $line);

        # Get the Disk ID first, its on its own line.
        # If this routine was called through ->format, we should
        # already have this information.
        if ((not length($this->{sig})) and ($line =~ /Disk\sID:\s(\S+)/)) {
            $this->{sig} = $1;
            next;
        }


        # check for partitions and verify whether gpt or mbr 
        # and dynamic or basic
        if (($line[1] =~ /^Partition$/i) && ($line[2] =~ /^\d+$/))
        {
            $this->{isClean} = 0;

            if ($line[3] =~ /Dynamic/)
            {
                $this->{diskType} = "dynamic";
            }
            elsif ($line[3] =~ /^Reserved$/)
            {
                $this->{partStyle} = "gpt";
            }
            elsif ($line[3] =~ /^Primary$/)
            {
                # add the partition number of the primary partition
                $this->{isParted} = $line[2];
            }
        }
        elsif (($line[1] =~ /^Volume$/i) && ($line[2] =~ /^\d+$/))
        {
            $this->{volume} = $line[2];

            # iterate through the array since it could have blanks
            for (my $i = 0; $i < @line; $i++)
            {
                $this->{mount} = $line[$i] if ($line[$i] =~ /^[A-Z]$/);
                $this->{fsType} = $line[$i] if ($line[$i] =~ /^(?:RAW|NTFS|FAT32)$/i);
                $this->{fsLabel} = $line[$i] if ($line[$i] =~ /^NATE_[A-Z]_\d+$/);
            }

        }
        elsif ($line[1] =~ /^([A-Z]:.*)\\?$/)
        {
            $this->{mount} = $1;
        }
    }

    &main::logcomment("Disk::getDiskInfo - Found disk: $this->{device} ;" .
        "type: $this->{diskType} ; style: $this->{partStyle} ; signature ; $this->{sig}");
    &main::logcomment("Disk::getDiskInfo - Found volume: $this->{volume} ;" .
        "fsType: $this->{fsType} ; fsLabel: $this->{fsLabel} ; mount: $this->{mount}");

getDiskInfo_EXIT:

    return; 

} # END getDiskInfo


#
#
#
sub setLBPolicy
{
    my ($this, %args) = @_;

    my ($status, @output, $cmd, $device_number, $device, $lbpolicy, $msg);

    # initially clear errors
    $this->_clear_err();

    # pull out the device number and append 'disk' to it
    $device_number = $this->{device};
    $device = "disk" . $device_number;

    # parse the args hash
    foreach (keys %args)
    {
        if (/^-?lbpolicy$/i)
        {
            if (!length($lbpolicy = $args{$_}) || $lbpolicy !~ /^[1234568]$/i)
            {
                $msg = "Disk::setLBPolicy - Invalid value for -lbpolicy: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto setLBPolicy_EXIT;
            }
            next;
        }
    }

    # build the mpc command to set the load balancing policy and execute
    $cmd = "\%SystemDrive\%\\TOOLS\\mpc.exe set lbpolicy $device $lbpolicy";
    
    &main::logcomment("Disk::setLBPolicy - Executing command: $cmd");
    
    ($status, @output) = $this->execute($cmd);
    if ($status != $Host::TRUE or grep(/Failed: set lbpolicy disk/i, @output)) 
    {
        $msg = "ERROR: Disk::setLBPolicy - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto setLBPolicy_EXIT;
    }
        
    &main::logcomment("Disk::setLBPolicy - Load Balancing Policy set to " . 
                      "'$lbpolicy' for '$device'");


setLBPolicy_EXIT:
    
    return;

} # END setLBPolicy


1;
