
# $Id: //depot/prod/test/Rsierranevada/lib/WinClient/Host.pm#6 $
#
# Copyright (c) 2009 Network Appliance, Inc.
# All rights reserved.
#
# This library is used to populate the host object created by
# WinClient.pm and determines features, such as Windows Version.
#
# TODO: Complete documentation pending.
#
## @summary
##  Creates Host object and imports associated libraries.  All architecture
##  specific properties are defined here and extrapolated elsewhere.
##
## @author  kenny speer
## @status  public
##

use strict;

package WinClient::Host;

# define the subroutines used in this package
sub new;                    # creates new host object
sub cleanup;                # cleanup all host objects
sub rescanPNP;              # rescans the entire PnP Bus
sub rescanDisks;            # rescan only disk subsystem
sub getDevices;             # updates host obj with hash of disks
sub configDevices;          # method to configure all the disks
sub updateLetters;          # updates array of avail/used letters
sub verifyDomain;           # Verify that $this->{domainName} is a domain and is accessible
sub reboot;                 # reboot the host
sub install;                # install something on the host
sub getiSCSINodename;       # retrieves a list of iSCSI Nodenames
sub getiSCSIInitiators;     # retrieves a list of iSCSI Initiators
sub getFcpInitiators;       # retrieves a list of FCP Initiators
sub runFcconfig;            # run fcconfig command to capture fcp adapter vendor and settings
sub startTracing;           # start driver tracing on the client
sub stopTracing;            # stop the driver trace session
sub collectData;            # capture debugging information to c:\temp
sub getSystemStats;         # collect system statistics (CPU&Mem Utilization)
sub hvcontrol;              # interface to the hvcontrol utility
sub nsanity;              # interface to the nsanity utility
sub _getHostInfo;           # populates host object with info
sub _hostSetup;             # setup the host with required files, etc

use Error;
use WinClient::Util;
use WinClient::Disk;
use WinClient::DsmTest;
use WinClient::SDCLI;
use WinClient::Install;
use WinClient::MPC;
use WinClient::ISCSI;
use WinClient::Wrshdnt;
use San::SanLog;
use San::SanClient::ClientUtil;
use Scalar::Util;
use vars qw( @ISA @EXPORT );
@ISA = qw(
          Error
          WinClient::Util
          WinClient::DsmTest
          WinClient::SDCLI
          WinClient::Install
          WinClient::MPC
          WinClient::ISCSI
          WinClient::Wrshdnt
          SanLog
          );

# set some global variables we may need later
$Host::SYSDRIVE     = "\%SystemDrive\%";
$Host::SYSROOT      = "\%SystemRoot\%";
$Host::TEMP         = "\%SystemDrive\%\\temp";
$Host::TOOLS        = "\%SystemDrive\%\\TOOLS";
$Host::MOUNT        = "\%SystemDrive\%\\mnt";
$Host::PATH_SEP     = "\\";
$Host::TRUE         = 0;
$Host::FALSE        = 1;
$Host::MAX_TIMEOUT  = 3600;
$Host::COMMON       = "\\SAN\\bin\\common";

# common commands
$Host::RM           = "del /F /Q";
$Host::COPY         = "copy /Y /V /Z";

#
#
#
sub new
{
    my ($class, %args) = @_;

    my ($this, $msg);

    if (defined $args{-self})
    {
        $this = $args{-self};
    }
    else
    {
        bless $this, $class;
        $msg = "WinClient::Host::new not called with WinClient object.";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto new_EXIT;
    }

    # need to bless this object
    bless $this, $class;

    # clear the errors first
    $this->_clear_err();

    # if an object template was specified, copy all the contents over and return
    # the new object
    if (defined $args{-obj_template}) {
        foreach (keys %{$args{-obj_template}}) {
            $this->{$_} = $args{-obj_template}->{$_};
        }
        # need to fix this stupid pointer
        Scalar::Util::weaken($this->{host} = $this);
        return $this;
    }

    # point back to ourselves for compat and calling host methods from
    # various objects
    Scalar::Util::weaken($this->{host} = $this);

    # set some host properties
    $this->{hostver}       = undef;     # host version
    $this->{build}         = undef;     # build number of the OS from ver
    $this->{full_build}    = undef;     # full build number of the OS from ver
    $this->{serverModel}   = undef;     # The model of the server, supplid by systeminfo
    $this->{arch}          = undef;     # host architecture
    $this->{domainName}    = undef;     # The domain ( or workgroup ) the computer belongs to
    $this->{usedLetters}   = undef;     # array of drive letters used
    $this->{availLetters}  = undef;     # array of available drive letters
    $this->{disks}         = undef;     # hash of disk objects
    $this->{nodes}         = undef;     # hash of cluster nodes
    $this->{fcpAdapters}   = undef;     # fcp adapters
    $this->{iscsiAdapters} = undef;     # iscsi adapters
    $this->{dsms}          = ( );       # array to hold the dsms available on the node

    # test specific tools and utils
    $this->{bin}                  = undef;   # architecture specific bin dir
    $this->{cmd_diskpart}         = undef;   # arch specific diskpart
    $this->{cmd_devcon}           = undef;   # arch specific devcon
    $this->{cmd_fcconfig}         = undef;   # fcconfig command
    $this->{cmd_iscsicli}         = undef;   # arch specific iscsicli
    $this->{cmd_dev_rescan}       = undef;   # uses devcon
    $this->{cmd_disk_rescan}      = undef;   # uses devcon
    $this->{cmd_sc}               = undef;   # service control utility
    $this->{cmd_sdcli}            = undef;   # snapdrive command line utility
    $this->{cmd_mpc}              = undef;   # multi-path control program
    $this->{cmd_dsmtest}          = undef;   # dsmtest utility
    $this->{cmd_dt}               = undef;   # data test io program
    $this->{cmd_sio}              = undef;   # sio io program
    $this->{cmd_spt}              = undef;   # spt program
    $this->{cmd_win_system_stats} = undef;   # System Usage collector
    $this->{cmd_hvcontrol}        = undef;   # Hyper-V command line utility
    $this->{cmd_nsanity}          = undef;   # nSANity command line utility
    $this->{cmd_custom_nsanity}   = undef;   # nSANity command line utility
    $this->{nsanity_commands_xml} = undef;   # nSANity command line utility xml commands file

    # set sd properties
    $this->{sdVersion} = undef;         # snapdrive version
    $this->{sdBuild}   = undef;         # snapdrive build

    # setup some host params based on arch and version
    $this->_getHostInfo();
    goto new_EXIT if ($this->error());

    # setup the host with specifc paths and info
    $this->_hostSetup();
    goto new_EXIT if ($this->error());

    # populate and enable some sd properties if we're using snapdrive
    $this->sdSysconfig() if ($this->{useSD});

    # get DSM/MPIO information - only call on server platforms were
    # a DSM is supported.
    if ($this->{hostver} =~ /2012|2008|2000|2003/i) {
        $this->_getDSMInfo();
        goto new_EXIT if ($this->error());
    }

    # dump out mpc version information for debug/audit purposes
    # purposely do not check status
    if ($this->{hostver} =~ /2012|2008|2000|2003/i) {
        $this->execute("$this->{cmd_mpc} show version", 120);
    }

new_EXIT:

    return $this;

} # END new

#
#
#
sub cleanup
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    foreach my $key (keys %{$this->{disks}})
    {
        $this->{disks}{$key}->destroy();
        if ($this->{disks}{$key}->error)
        {
            $this->_set_err(@{$this->{disks}{$key}->{errmsg}});
        }
    }

    if (!$this->error)
    {
        $this->execute("rmdir /Q /S $Host::MOUNT");
        $this->execute("rmdir /Q /S $Host::TEMP");
    }

} # END cleanup


#
#
#
sub rescanPNP
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my ($status, @output, $cmd);

    &main::logcomment("WinClient::Host::rescanPNP - Performing system device rescan.");

    ($status, @output) = $this->execute($this->{cmd_dev_rescan}, $this->{timeout});
    if ($status != $Host::TRUE)
    {
        &main::logcomment("$this->{cmd_dev_rescan} failed.");
        $this->_set_err(@output);
        goto rescanPNP_EXIT;
    }

    &main::logcomment("WinClient::Host::rescanPNP - System device rescan completed: ", @output);

rescanPNP_EXIT:

    return;

} # END rescanPNP

#
#
#
sub rescanDisks
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my ($status, @output);

    &main::logcomment("WinClient::Host::rescanDisks - Performing disk rescan.");

    # generate disk object to get access to diskpartCmd
    my $diskObj = WinClient::Disk->new(-host => $this);

    ($status, @output) = $diskObj->_diskpartCmd(
                    timeout  => 600,
                    commands => [("rescan",
                                  "exit")],
                    comment  => "Host::rescanDisks - rescan disks");

    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::rescanDisks - diskpart rescan failed.");
        $this->_set_err(@output);
        goto rescanDisks_EXIT;
    }

    &main::logcomment("WinClient::Host::rescanDisks - Disk rescan completed: ", @output);

rescanDisks_EXIT:

    return;

} # END rescanDisks

#
#
#
sub updateLetters
{
    my ($this) = @_;

    my ($line, $status, @output, $cmd, @usedLetters, @availLetters);

    $cmd = "$this->{cmd_mpc} show driveletters";
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::updateLetters - $cmd failed.");
        $this->_set_err(@output);
        goto updateLetters_EXIT;
    }

    foreach $line (@output)
    {
        if ($line =~ /^Used\s+Drive\s+Letters\s+\(\d+\):\s+(.*$)/i)
        {
            @usedLetters = split(/\s/, $1);
            $this->{usedLetters} = \@usedLetters;
        }
        elsif ($line =~ /^Available\s+Drive\s+Letters\s+\(\d+\):\s+(.*$)/i)
        {
            @availLetters = split(/\s/, $1);
            $this->{availLetters} = \@availLetters;
        }
    }

    &main::logcomment("WinClient::Host::updateLetters - Updated Drive Letters");

updateLetters_EXIT:

    return;

} # END updateLetters

#
#
#
sub getDevices
{
    my ($this, $numLuns, $skip_show_luns, $rescan_timeout) = @_;

    # initially clear errors
    $this->_clear_err();

    # update the drive letter arrays before beginning
    $this->updateLetters();

    my ($status, @output, $msg);
    my ($line, $diskObj, @disks, @usedLetters, @availLetters);
    my ($disk, $boot, $lun, $serial, $cfmode, $filer, $lunPath);
    my ($numDisks, $numDisksPrev, $retry, $cmd, $numLunsSet, $retries);

    if (defined($numLuns))
    {
        $numLunsSet = $Host::TRUE;
        $retries = 60;
    }
    else
    {
        $numLunsSet = $Host::FALSE;
        $numLuns = 0;
        $retries = 5;
    }

    # give a default to rescan_timeout if not passed
    if (not defined($rescan_timeout))
    {
        $rescan_timeout = 120;
    }

    # before configuring anything, make sure we have enough disks
    # retry for 30 minutes and stop, this is needed for lots of LUNs
    $numDisksPrev = 0;
    my $match_count = 0;
    for ($retry = 0; $retry < $retries; $retry++)
    {
        $numDisks = 0;

        # rescan the disks before each attempt of finding them
        $this->rescanDisks();

        # sleep rescan_timeout seconds to give PnP play event to complete
        # and and have all disks/LUNs be available
        sleep($rescan_timeout);

        $cmd = "$this->{cmd_mpc} show disks";
        ($status, @output) = $this->execute($cmd, 3900);

        if ($status != $Host::TRUE)
        {
            &main::logcomment("WinClient::Host::getDevices - $cmd failed.");
            $this->_set_err(@output);
            goto getDevices_EXIT;
        }

        foreach $line (@output)
        {
            $numDisks++ if ($line =~ /^Disk\d+\s+/);
        }

        last if ((0 == $numLuns) && ($Host::TRUE == $numLunsSet));
        last if (($numDisks >= $numLuns) && ($Host::TRUE == $numLunsSet));

        if ($Host::FALSE == $numLunsSet) {
            my $need_retry = 0;
            # check to see if MPC through an exception, if it did, retry again.
            # This can happen when all PnP events are not complete (after discovery,
            # reboot, or fault injection)
            if (grep(/An error occurred/i, @output) > 0) {
                &main::logcomment("WinClient::Host::getDevices - An exception occurred running mpc, " .
                                  "retrying rescan and show disks");
                $need_retry = 1;
            }

            # If the user did not pass in numLuns to be seen, call rescan and
            # mpc show disks until two repeating calls verify the same number
            # of disks. If it does not use the retry logic
            if ($numDisks == 0 or $numDisks > $numDisksPrev) {
                &main::logcomment("WinClient::Host::getDevices - Number of disks discovered ($numDisks) is greater " .
                                  "than previous number discovered ($numDisksPrev),\n" .
                                  "or number of discovered disks = 0\n" .
                                  "retrying rescan and show disks");
                $numDisksPrev = $numDisks;
                $need_retry = 1;
            }

            if ($need_retry == 0) {
                last;
            }
        } else {
            # the user specified a specific number of luns that they want to discover
            # then if the disks that we discovered did not change since the last attempt, then break out
            if ($numDisks == $numDisksPrev) {
                if ($match_count > 3) {
                    &main::logcomment("WinClient::Host::getDevices - The number of disks discovered ($numDisks) has not changed " .
                                          "since the last three discovery attempts, giving up.");
                    last;
                } else {
                    $match_count++;
                }
            } else {
                # reset the match count
                $match_count = 0;
            }
            $numDisksPrev = $numDisks;

        }

        &main::logcomment("WinClient::Host::getDevices - found ($numDisks) disks of ($numLuns) " .
                          "requested\n retry: $retry");
        sleep(10);
    }

    if (defined($numLuns) && ($numDisks < $numLuns))
    {
        $msg = "WinClient::Host::getDevices - all devices were not found" .
            "\nLUNs Found: $numDisks\nLUNs Required: $numLuns";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto getDevices_EXIT;
    }

    # iterate through the list of disks and create them if they
    # don't exist, else update them in the next loop
    foreach $line (@output)
    {
        # match the line which contains a disk in the OK state
        if ($line =~ /^Disk(\d+)\s+/i)
        {
            # save the disk number
            $disk = $1;

            # update the array of disks being used
            push(@disks, $disk);

            # create disk objects
            if (!defined($this->{disks}{$disk}))
            {
                &main::logcomment(
                    "WinClient::Host::getDevices - Creating disk: Disk $disk");

                $diskObj = WinClient::Disk->new(-host   => $this,
                                                -device => $disk);

                if ($diskObj->error())
                {
                    &main::logcomment(
                        "WinClient::Host::getDevices - Disk::new returned an error");
                    $this->_set_err(@{$diskObj->{errmsg}});
                    goto getDevices_EXIT;
                }

                $this->{disks}{$disk} = $diskObj;
            }
            else
            {
                &main::logcomment(
                    "WinClient::Host::getDevices - Device already exists: Disk $disk");
                &main::logcomment("WinClient::Host::getDevices - Updating disk $disk");
            }

            # match the line which contains a disk
            # output is \t delimited and of the form:
            #
            # Disk\tLUN\tSize\tState\tBoot\tVendor\tStyle\tType\tLetter\tFileSys\tLabel\tQuorum\tClusterName\tClusterNode
            # ----\t---\t----\t-----\t----\t------\t-----\t----\t------\t-------\t-----\t------\t-----------\t-----------
            # Disk1\t1\t5245 M\tOK\tFalse\tNETAPP\tMBR\tNone\tNone\tNone\tNone\tNone\tNone\tNone
            #
            # NOTE: we use the regex instead of split since the Label
            # and disk size both *could* contain a \s

            if ($line =~ /^Disk(\d+)\s+(\d+)\s+(\d+\s+\w+)\s+(\w+)\s+(True|False)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w*)\s+(\w+)\s+\[(.*)\]\s+(True|False)\s+(\S+)\s+(.*)/)
            {
                # save the disk number
                $disk = $1;

                # populate this disks properties based on disk info
                $this->{disks}{$disk}->{lun}          = $2;
                $this->{disks}{$disk}->{lunSize}      = $3;
                $this->{disks}{$disk}->{lunState}     = $4;
                $this->{disks}{$disk}->{boot}         = $5;
                $this->{disks}{$disk}->{vendor}       = $6;
                $this->{disks}{$disk}->{partStyle}    = $7;
                $this->{disks}{$disk}->{diskType}     = $8;
                $this->{disks}{$disk}->{mount}        = $9;
                $this->{disks}{$disk}->{fsType}       = $10;
                $this->{disks}{$disk}->{fsLabel}      = $11;
                $this->{disks}{$disk}->{isQuorum}     = $12;
                $this->{disks}{$disk}->{clusterName}  = $13;
                $this->{disks}{$disk}->{clusterNode}  = $14;

                # for good measure due to the ridiculous regex above,
                # logcomment on what we found
                &main::logcomment("
                    lun:          $this->{disks}{$disk}->{lun}
                    lunSize:      $this->{disks}{$disk}->{lunSize}
                    lunState:     $this->{disks}{$disk}->{lunState}
                    boot:         $this->{disks}{$disk}->{boot}
                    vendor:       $this->{disks}{$disk}->{vendor}
                    partStyle:    $this->{disks}{$disk}->{partStyle}
                    diskType:     $this->{disks}{$disk}->{diskType}
                    mount:        $this->{disks}{$disk}->{mount}
                    fsType:       $this->{disks}{$disk}->{fsType}
                    fsLabel:      $this->{disks}{$disk}->{fsLabel}
                    isQuorum:     $this->{disks}{$disk}->{isQuorum}
                    clusterName:  $this->{disks}{$disk}->{clusterName}
                    clusterNode:  $this->{disks}{$disk}->{clusterNode}
                    ");

                # fix the label and remove erroneous spaces
                $this->{disks}{$disk}->{fsLabel} =~ s/\s//g;

                # set the boot property based on boolean value
                if ($this->{disks}{$disk}->{boot} =~ /true/i)
                {
                    $this->{disks}{$disk}->{boot} = 1;
                }
                else
                {
                    $this->{disks}{$disk}->{boot} = 0;
                }

                if ($this->{disks}{$disk}->{mount} !~ /none/i)
                {
                    $this->{disks}{$disk}->{isParted} = 1;
                }
                else
                {
                    $this->{disks}{$disk}->{isParted} = 0;
                    $this->{disks}{$disk}->{mount}    = undef;
                    $this->{disks}{$disk}->{fsType}   = undef;
                    $this->{disks}{$disk}->{fsLabel}  = undef;
                }

                # reset isQuorum to boolen value
                if ($this->{disks}{$disk}->{isQuorum} =~ /True/i)
                {
                    $this->{disks}{$disk}->{isQuorum} = 1;
                }
                else
                {
                    $this->{disks}{$disk}->{isQuorum} = 0;
                }
            }
        }
    }

    if (defined($skip_show_luns) and $skip_show_luns == 1)
    {
        goto getDevices_skip_show_luns;
    }

    # Add to a retry loop in case exception is thrown
    # Have seen this in various configs, and retry usually helps
    my $lun_show_retries = 0;
    while ($lun_show_retries <= 5) {
        # execute mpc show luns for lun specific info, this sends an inquiry
        # to each LUN which could take a long time on a loaded filer with
        # many luns
        $cmd = "$this->{cmd_mpc} show luns";
        ($status, @output) = $this->execute($cmd, 3900);

        # determine if an exception occurred
        if ($lun_show_retries <= 5) {
            if (grep(/An error occurred|STACKTRACE/i, @output) > 0) {
                &main::logcomment("WinClient::Host::getDevices - Exception occurred, retrying mpc show luns ($lun_show_retries of 5)");
                $lun_show_retries++;
                sleep(10);
            } else {
                last;
            }
        }
    }

    if ($status != $Host::TRUE) {
        &main::logcomment("WinClient::Host::getDevices - $cmd failed.");
        $this->_set_err(@output);
        goto getDevices_EXIT;
    }

    # only return the disks requested to be used via userDevices
    foreach $line (@output)
    {
        # Hack for no vdiskgetattr
        my $no_vdisk_get_attr = 0;
        if ($line =~ /toaster\s+please\s+implement\s+scsit_conf_lu_path/i) {
            $line =~ s/toaster\s+please\s+implement\s+scsit_conf_lu_path/NOT_IMPLEMENTED/;
            $no_vdisk_get_attr = 1;
        }

        # match the line which contains a disk
        # Disk\tLUN\tDSM  \tPolicy\tPaths\tSerial\t\tCFMode\tFiler\t\tProto\tLunType\t\tPath
        # ----\t---\t-----\t------\t------\t-----\t\t------\t-----\t\t-----\t-------\t\t----
        # Disk1\t1\tontap\tFO\t2\tVqlpHJ9gZhAT\tStandBy\tfas960-fak01\t\tFCP\twindows\t\t/vol/vol1/nate_wbc2b2_lun1
        #
        # NOTE: we use the regex instead of split since the Label
        # and disk size both *could* contain a \s
        if ($line =~ /^Disk(\d+)\s+(\d+)\s+(\w+)\s+(\w+)\s+(\S+)\s+(\S+)\s+(\w+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/ or
            $line =~ /^Disk(\d+)\s+(\d+)\s+(\w+)\s+(\w+)\s+(\d+)\s+(\S+)\s+(\w+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(NOT_IMPLEMENTED)/) {
            # save the disk number
            $disk = $1;

            # populate this disks properties based on disk info
            # fields - Disk LUN DSM Policy Paths Serial CFMode Filer Path
            $this->{disks}{$disk}->{lun}      = $2;
            $this->{disks}{$disk}->{dsm}      = $3;
            $this->{disks}{$disk}->{policy}   = $4;
            $this->{disks}{$disk}->{numPaths} = $5;
            $this->{disks}{$disk}->{serial}   = $6;
            $this->{disks}{$disk}->{cfmode}   = $7;

            # hack for ONTAP w/o vdisk getattr()
            if ($no_vdisk_get_attr == 1) {
                $this->{disks}{$disk}->{filer}   = "no_vdisk_get_attr";
                $this->{disks}{$disk}->{lunPath} = "no_vdisk_get_attr";
                $this->{disks}{$disk}->{lunType} = "no_vdisk_get_attr";
                $this->{disks}{$disk}->{protocol} = "no_vdisk_get_attr";
            } else {
                $this->{disks}{$disk}->{filer}    = $8;
                $this->{disks}{$disk}->{lunPath}  = $11;
                $this->{disks}{$disk}->{lunType}  = $10;
                $this->{disks}{$disk}->{protocol} = $9;
            }

            # for good measure due to the ridiculous regex above,
            # logcomment on what we found
            &main::logcomment("
                disk:     Disk $disk
                lun:      $this->{disks}{$disk}->{lun}
                dsm:      $this->{disks}{$disk}->{dsm}
                policy:   $this->{disks}{$disk}->{policy}
                numPaths: $this->{disks}{$disk}->{numPaths}
                serial:   $this->{disks}{$disk}->{serial}
                cfmode:   $this->{disks}{$disk}->{cfmode}
                filer:    $this->{disks}{$disk}->{filer}
                lunPath:  $this->{disks}{$disk}->{lunPath}
                lunType:  $this->{disks}{$disk}->{lunType}
                protocol: $this->{disks}{$disk}->{protocol}
                ");

        }
    }

getDevices_skip_show_luns:

    # if this host is using SnapDrive then we can gather mount points
    # via sdcli
    if ($this->{useSD})
    {
        $cmd = "$this->{cmd_sdcli} disk list";
        ($status, @output) = $this->execute($cmd, $this->{timeout});

        if ($status != $Host::TRUE)
        {
            &main::logcomment("WinClient::Host::getDevices - $cmd failed.");
            $this->_set_err(@output);
            goto getDevices_EXIT;
        }

        my ($match, $diskID, $mount);
        foreach $line (@output)
        {
            if ($line =~ /\s+DiskID:\s+(\d+)$/i)
            {
                $diskID = $1;
                $match = 1;
            }
            elsif ($line =~ /\s+Mount\s+points:\s+(.+)$/i && $match)
            {
                $mount = $1;
                $mount =~ s/:\\// if ($mount =~/[A-Z]:\\$/);
                $this->{disks}{$diskID}->{isParted} = 1;
                $this->{disks}{$diskID}->{mount}    = $mount;
                $this->{disks}{$diskID}->{fsType}   = "ntfs";
                $match = 0;
            }
        }
    }

    # update and remove any disks no longer seen by the host
    foreach my $key (keys %{$this->{disks}})
    {
        delete($this->{disks}{$key}) if (not grep(/^$key$/, @disks));
    }

    # notify the user which disks which are going to use
    &main::logcomment("WinClient::Host::getDevices - Using these disks: @disks");

    # if there are no disks then we need to set a warning
    &main::logcomment("WARNING: No devices were found on this system.") if (!@disks);

getDevices_EXIT:

    return @disks;

} # END getDevices

#
#
#
sub configDevices
{
    my ($this, %args) = @_;

    my ($msg, $deviceList, $partStyle, $fsType, @lunIds);

    # iterate through the args
    foreach (keys %args)
    {
        if (/-?deviceList/i)
        {
            # $deviceList has an overload of "all" which covers all
            # discovered disks exept the boot disk
            if (!length($deviceList = $args{$_}))
            {
                $msg = "WinClient::Host::configDevices() - Invalid value for -deviceList: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto configDevices_EXIT;
            }
            next;
        }
        if (/-?partStyle/i)
        {
            if (!length($partStyle = $args{$_}))
            {
                $msg = "WinClient::Host::configDevices() - Invalid value for -partStyle: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto configDevices_EXIT;
            }
            next;
        }
        if (/-?fsType/i)
        {
            if (!length($fsType = $args{$_}))
            {
                $msg = "WinClient::Host::configDevices() - Invalid value for -fsType: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto configDevices_EXIT;
            }
            next;
        }
    }

    # parse the device list and create an array of lunIds
    my @tmpLuns = split(/:/, $deviceList);
    foreach my $tmpLun (@tmpLuns)
    {
        if ($tmpLun =~ /(\d+)-(\d+)/)
        {
            for ($1..$2)
            {
                push(@lunIds, $_);
            }
        }
        elsif ($tmpLun =~ /^\d+$/)
        {
            push(@lunIds, $tmpLun);
        }
    }

    # track whether we need to update the disk info
    my $updateDevices = 0;

    # rescan the bus and get disks via $hostObj->rescanDisks();
    $this->rescanDisks();

    # then scan and acquire the disks
    if ($deviceList !~ /all/i)
    {
        $this->getDevices(@lunIds);
    }
    else
    {
        $this->getDevices();
    }
    goto configDevices_EXIT if ($this->error());

    # iterate through the list of disks
    foreach my $key (sort keys %{$this->{disks}})
    {
        # don't touch the boot lun
        next if ($this->{disks}{$key}->{boot});

        # save time and only config the disk if it's defined in
        # CLIENT_IO_DEVS ($deviceList)
        if (grep(/^$this->{disks}{$key}->{lun}$/, @lunIds)|
            $deviceList =~ /all/i)
        {
            # make sure the disk is clean if we have to convert, or if
            # the fsType is raw
            if ($partStyle !~ /none/i | $fsType =~ /raw/i)
            {
                # if the partition style doesn't match the current
                # config OR if we need raw devices but we have a
                # partition or filesystem
                if ($this->{disks}{$key}->{partStyle} !~ /$partStyle/i |
                    ($this->{disks}{$key}->{mount} !~ /none/i &
                    defined($this->{disks}{$key}->{mount}) &
                    $fsType =~ /raw/i))
                {
                    # destroy the current disk configuration
                    $this->{disks}{$key}->destroy();
                    if ($this->{disks}{$key}->error())
                    {
                        $this->_set_err($this->{disks}{$key}->errmsg);
                        goto configDevices_EXIT;
                    }

                    # convert the disk to the proper partition style
                    $this->{disks}{$key}->convert(-partStyle => $partStyle);
                    if ($this->{disks}{$key}->error())
                    {
                        $this->_set_err($this->{disks}{$key}->errmsg);
                        goto configDevices_EXIT;
                    }

                    # we need to update the device properties before
                    # formatting the devices
                    $updateDevices = 1;
                }
            }
        }
    }

    # update the disks with new info if they changed
    if ($updateDevices)
    {
        $this->getDevices();
        goto configDisk_EXIT if ($this->error());
        $updateDevices = 0; # reset to false
    }

    # our disks should be ready for additional config, but only perform
    # it if it's required by fs type
    if ($fsType =~ /fs/)
    {
        # partition and format the luns
        foreach my $key (sort keys %{$this->{disks}})
        {
            # regardless, we always skip the boot lun
            next if ($this->{disks}{$key}->{boot});

            # create partition and mount point if the mount point does
            # not exist
            if ($this->{disks}{$key}->{mount} =~ /none/i |
                !defined($this->{disks}{$key}->{mount}))
            {
                $this->{disks}{$key}->partition();
                $this->{disks}{$key}->assign(-type => "letter");
                if ($this->{disks}{$key}->error())
                {
                    $this->_set_err($this->{disks}{$key}->errmsg);
                    goto configDevices_EXIT;
                }
            }

            # format the disk if it's not currently formated, else just
            # empty the disk
            if ($this->{disks}{$key}->{fsType} =~ /none|undef/i |
                !defined($this->{disks}{$key}->{fsType}))
            {
                $this->{disks}{$key}->format();
                if ($this->{disks}{$key}->error())
                {
                    $this->_set_err($this->{disks}{$key}->errmsg);
                    goto configDevices_EXIT;
                }
            }
            else
            {
                $this->{disks}{$key}->empty();
                if ($this->{disks}{$key}->error())
                {
                    $this->_set_err($this->{disks}{$key}->errmsg);
                    goto configDevices_EXIT;
                }
            }
        }
    }

configDevices_EXIT:

    return;

} # END configDevices
#
#
#
sub reboot
{
    my ($this) = @_;

    # initially clear errors
    $this->_clear_err();

    my $connection = $this->{connection};
    my ($status, $output, $msg);
    my ($cmd, $timeout, $endtime);

    # reboot the machine under test
    $cmd = "cmd /c " . $this->{cmd_reboot};
    &main::say($connection, $cmd, '-blind');

    # default to 15 minute timeout, since i/o could be in flight and IA64
    # is the slowest pig i've ever seen attempt to boot
    $timeout = 900;
    $endtime = time() + $timeout;

    # sleep for 120 seconds to give the host a chance to reboot
    # ia64 takes forever to shutdown, so we wait > 1 min
    sleep(120);

    # check the host to determine it's availability, the most reliable
    # method I've found is to use RPC and the 'net use' command
    $cmd = "cmd /c net use";

    while (time() < $endtime)
    {
        (undef, undef, $output) = &main::say($connection, $cmd,
            '-eval_error', '-timeout', 30);

        if ($output =~ /^New.*connections/)
        {
            $this->_isMinWin();
            goto reboot_EXIT;
        }
        sleep(30);
    }

    # if we hit this, then we didn't complete the reboot
    $msg = "WinClient::Host::reboot - FATAL: reboot did not complete";
    &main::logcomment($msg);
    $this->_set_err($msg);

reboot_EXIT:

    return;

} # END reboot

#
#
#
sub install
{
    my ($this, %args) = @_;

    my ($msg, $application);

    foreach (keys %args)
    {
        if (/^-application$/)
        {
            if (!length($application = $args{$_}))
            {
                $msg = "WinClient::Host::install - invalid value for -application: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto install_EXIT;
            }
            next;
        }
    }

    if (!defined($application) || ($application !~ /^(?:mpio)$/i))
    {
        $msg = "WinClient::Host::install - Required argument '-application' invalid";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto install_EXIT;
    }

    $this->_installMPIO(%args) if ($application =~ /mpio/);

install_EXIT:

    return;

} # END install

#
#
#
sub startTracing
{
    my ($this, %args) = @_;

    my ($msg, $cmd, $status, @output, $logSize);
    my ($buffer_size, $buffer_count_min, $buffer_count_max, $format, $clock);
    my ($driver, $level, $flags, $guid, $outFile);

    # parse the arguments
    foreach (keys %args)
    {
        if (/^-?driver/i)
        {
            if (!length($driver = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -driver: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?level/i)
        {
            if (!length($level = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -level: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?flags/i)
        {
            if (!length($flags = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -flags: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?guid/i)
        {
            if (!length($guid = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -guid: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?outFile/i)
        {
            if (!length($outFile = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -outFile: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?logSize/i)
        {
            if (!length($logSize = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -logSize: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?buffer_size/i)
        {
            if (!length($buffer_size = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -buffer_size: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?buffer_count_min/i)
        {
            if (!length($buffer_count_min = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -buffer_count_min: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?buffer_count_max/i)
        {
            if (!length($buffer_count_max = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -buffer_count_max: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?format/i)
        {
            if (!length($format = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -format: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
        elsif (/^-?clock/i)
        {
            if (!length($clock = $args{$_}))
            {
                $msg = "WinClient::Host::startTracing - invalid value for -clock: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto startTracing_EXIT;
            }
            next;
        }
    }

    # parse the required and default arguments
    if (!defined($driver) || !defined($level) || !defined($guid))
    {
        $msg = "WinClient::Host::startTracing - required parameters missing";
        &main::logcomment($msg);
        goto startTracing_EXIT;
    }

    # take into consideration OEM_REBRAND
    if (defined($this->{general}->{OEM_REBRAND}) and
        $this->{general}->{OEM_REBRAND} =~ /ibm/i) {
        $Host::DSMPATH      = "$Host::SYSDRIVE\\progra~1\\IBM\\mpio";
    } else {
        $Host::DSMPATH      = "$Host::SYSDRIVE\\progra~1\\netapp\\mpio";
    }

    $guid = "$Host::DSMPATH\\ontapdsm.ctl" if (!defined($guid) || $guid =~ /none/i);
    $outFile = "$Host::TEMP\\ontapdsm.etl" if (!defined($outFile) || $outFile =~ /none/i);

    for my $dsm ( @{$this->{dsms}} ) {
        if ($dsm->{dsmname} =~ /^Data ONTAP DSM/i) {

            # DSM == 4.0, use logman.exe
            if ($dsm->{dsmversion} =~ /^4\.0/) {
                $cmd = "$this->{cmd_logman} start ontapdsm ";
                $cmd .= "-p {$guid} $flags $level ";
                $cmd .= "-o $outFile ";
                $cmd .= "-ets -nb $buffer_count_min $buffer_count_max ";
                $cmd .= "-bs $buffer_size -max $logSize ";
                $cmd .= "-f $format -ct $clock -mode globalsequence";
            }
            else {
                # DSM <= 3.5, use tracelog.exe
                $cmd = "$this->{cmd_tracelog} -start ";
                $cmd .= "$driver -guid $guid ";
                $cmd .= "-f $outFile ";
                $cmd .= "-rt -level $level -flags $flags -cir 250 ";
            }
            last;
        }
    }

    # send the command and check the output
    ($status, @output) = $this->execute($cmd, 60);
    if ($status != $Host::TRUE)
    {
        $msg = "WinClient::Host::startTracing - failed to execute command: $cmd";
        &main::logcomment($msg);
        $this->_set_err(@output);
        goto startTracing_EXIT;
    }
    elsif (!grep(/completed successfully/i, @output))
    {
        $msg = "Failed to enable driver tracing for $driver";
        &main::logcomment($msg);
        $this->_set_err(@output);
        goto startTracing_EXIT;
    }
    else
    {
        $msg = "Successfully enabled driver tracing for $driver";
        &main::logcomment($msg);
    }

startTracing_EXIT:

    return;

} # END startTracing

sub stopTracing
{
    my ($this, %args) = @_;

    my ($msg, $cmd, $status, @output);
    my ($driver);

    # parse the arguments
    foreach (keys %args)
    {
        if (/^-?driver/i)
        {
            if (!length($driver = $args{$_}))
            {
                $msg = "WinClient::Host::stopTracing - invalid value for -driver: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                goto stopTracing_EXIT;
            }
            next;
        }
        else
        {
            $msg = "WinClient::Host::stopTracing - unknown parameter: $args{$_}";
            &main::logcomment($msg);
            next;
        }
    }

    # parse the required and default arguments
    if (!defined($driver))
    {
        $msg = "WinClient::Host::stopTracing - required parameters missing";
        &main::logcomment($msg);
        goto stopTracing_EXIT;
    }

    # build the command
    for my $dsm ( @{$this->{dsms}} ) {
        if ($dsm->{dsmname} =~ /^Data ONTAP DSM/i) {

            # DSM == 4.0, use logman.exe
            if ($dsm->{dsmversion} =~ /^4\.0/) {
                $cmd = "$this->{cmd_logman} stop ontapdsm -ets";
            }
            else {
                # DSM <= 3.5, use tracelog.exe
                $cmd = "$this->{cmd_tracelog} -stop $driver";
            }
            last;
        }
    }

    # send the command and check the output
    ($status, @output) = $this->execute($cmd, 600);
    if ($status != $Host::TRUE)
    {
        $msg = "WinClient::Host::stopTracing - failed to execute command: $cmd";
        &main::logcomment($msg);
        $this->_set_err(@output);
        goto stopTracing_EXIT;
    }
    elsif (!grep(/completed successfully/i, @output))
    {
        $msg = "Failed to terminate driver tracing for $driver";
        &main::logcomment($msg);
        $this->_set_err(@output);
        goto stopTracing_EXIT;
    }
    else
    {
        $msg = "Successfully terminated driver tracing for $driver";
        &main::logcomment($msg);
    }

stopTracing_EXIT:

    return;

} # END stopTracing

#
#
#
sub collectData
{
    my ($this) = @_;

    # define our locations and file names
    my $dataDir = $Host::TEMP . "\\DEBUG_" . `date +%y%m%d_%H%M%S`;
    my $mpc_general = $dataDir . "\\mpc_general.txt";
    my $mpc_mpio = $dataDir . "\\mpc_mpio.txt";
    my $mpc_ontapdsm = $dataDir . "\\mpc_ontapdsm.txt";
    my $mpc_fcp = $dataDir . "\\mpc_fcp.txt";

    # create the directory to store the output
    $this->execute("mkdir $dataDir", 60);

    # capture the event logs
    $this->execute("$this->{cmd_mpc} events capture", 600);
    $this->execute("move EventLog* $dataDir\\", 60);

    # capture mpc information
    $this->execute("$this->{cmd_mpc} show version >> $mpc_general", 600);
    $this->execute("$this->{cmd_mpc} show disks >> $mpc_general", 600);
    $this->execute("$this->{cmd_mpc} show luns >> $mpc_general", 600);
    $this->execute("$this->{cmd_mpc} mpio show >> $mpc_mpio", 600);
    $this->execute("$this->{cmd_mpc} mpio show summary >> $mpc_mpio", 600);
    $this->execute("$this->{cmd_mpc} mpio show disks >> $mpc_mpio", 600);
    $this->execute("$this->{cmd_mpc} mpio show lunpaths >> $mpc_mpio", 600);
    $this->execute("$this->{cmd_mpc} ontapdsm show >> $mpc_ontapdsm", 600);
    $this->execute("$this->{cmd_mpc} ontapdsm show lunpaths >> $mpc_ontapdsm", 600);
    $this->execute("$this->{cmd_mpc} ontapdsm show perfstats >> $mpc_ontapdsm", 600);
    $this->execute("$this->{cmd_mpc} fcp show adapters >> $mpc_fcp", 600);
    $this->execute("$this->{cmd_mpc} fcp show switches >> $mpc_fcp", 600);
    $this->execute("$this->{cmd_mpc} fcp show discoveredports >> $mpc_fcp", 600);

    # etl should be in c:\temp, but we'll move it to the datadir
    $this->execute("move $Host::TEMP\\*.etl $dataDir\\", 600);

    return;

} # END collectData

#
#
#
sub getiSCSINodename
{
    my ($this) = @_;

    my ($nodename, $status, @output, $cmd, $line, $msg);

    # initially clear errors
    $this->_clear_err();

    # get the iqn name
    $cmd = "$this->{cmd_mpc} iscsi show adapters";
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE)
    {
        $msg = "WinClient::Host::getiSCSINodename - $cmd failed " . "\n@output";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto getiSCSINodename_EXIT;
    }

    foreach $line (@output)
    {
        if ($line =~ /^iSCSINodeName: (.*)/)
        {
            $nodename = $1;
            last;
        }
    }

getiSCSINodename_EXIT:

    return $nodename;

} # END getiSCSINodename

#
#
#
sub getiSCSIInitiators
{
    my ($this) = @_;

    my (@initiators, @adapter, $status, $line, $cmd, @output, $msg);

    # initially clear errors
    $this->_clear_err();

    # get the iqn name
    $cmd = "$this->{cmd_mpc} iscsi show adapters";
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE)
    {
        $msg = "WinClient::Host::getiSCSINodename - $cmd failed " . "\n@output";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto getiSCSINodename_EXIT;
    }

    foreach $line (@output)
    {
        next if $line =~ /^iSCSINodeName.*/;   # skip iqn name
        next if $line =~ /^\s*$/;               # skip empty lines
        next if $line =~ /^AdapterName.*/;      # skip header line 1
        next if $line =~ /^-----------.*/;      # skip header line 2

        # format of output is \t separated:
        # AdapterName Index Tag Type Address
        @adapter = split (/\s+/, $line);
        push(@initiators, @adapter);
    }

getiSCSIInitiators_EXIT:

    # populate our iscsiAdapters property
    $this->{iscsiAdapters} = \@initiators if (@initiators);

    return @initiators;

} # END getiSCSIInitiators

#
#
#
sub getFcpInitiators
{
    my ($this) = @_;
    my ($status, @output, @output_detail, $msg);
    my ($cmd, $line, @adapter, @internalAdapters, @initiators);

    # initially clear errors
    $this->_clear_err();

    # get a list of fcp adapters
    $cmd = "$this->{cmd_mpc} fcp show adapters";
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE or grep(/Failed: fcp show adapters/i, @output))
    {
        $msg = "WinClient::Host::getFcpInitiators - $cmd failed  " . "\n@output";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto getFcpInitiators_EXIT;
    }

    foreach $line (@output)
    {
        next if $line =~ /^\s*$/;           # skip empty lines
        next if $line =~ /^AdapterName.*/;  # skip header line 1
        next if $line =~ /^-----------.*/;  # skip header line 2
        next if $line =~ /^Success: fcp/;  # skip success line

        # format of output is \t separated (but Solaris uses expandtab):
        # AdapterName State Type NodeName PortName
        @adapter = split(/\s+/, $line);
        push(@internalAdapters, @adapter);

        # get detailed information for an fcp adapter
        $cmd = "$this->{cmd_mpc} fcp show adapter detail $adapter[0]";
        ($status, @output_detail) = $this->execute($cmd, $this->{timeout});

        if ($status != $Host::TRUE or grep(/Failed: fcp show adapter detail/i, @output_detail))
        {
            $msg = "WinClient::Host::getFcpInitiators - $cmd failed  " .
                "\n@output_detail";
            &main::logcomment($msg);
            $this->_set_err($msg);
            goto getFcpInitiators_EXIT;
        }
        push (@initiators, @output_detail);
    }

getFcpInitiators_EXIT:

    # populate our fcpAdapters property using the generic fcp adapter
    # output, instead of the verbose detailed output
    $this->{fcpAdapters} = \@internalAdapters if (@internalAdapters);

    return @initiators;

} # END getFcpInitiators

#
#
#
sub runFcconfig
{
    my ($this) = @_;
    my ($status, @output, $msg);

    # initially clear errors
    $this->_clear_err();

    # run the fcconfig command
    my $cmd = "$this->{cmd_fcconfig} -g";
    ($status, @output) = $this->execute($cmd, $this->{timeout});

    if ($status != $Host::TRUE)
    {
        $msg = "WinClient::Host::runFcconfig: $cmd failed  " . "\n@output";
        &main::logcomment($msg);
        $this->_set_err($msg);
        return $status, @output;
    } else {
        return $status, @output;
    }

} # END runFcconfig

#
# Copy over the nsanity binaries from the X drive
#
sub load_nsanity {
    my ($this) = @_;

    # Copy over the nsanity Command line tool
    my ($status, @output) = $this->copyRemoteFile($this->{host_common_driveletter} . "$Host::COMMON\\Nsanity\\$this->{cmd_nsanity}",
                                                  "$Host::TOOLS\\$this->{cmd_nsanity}");
    if ($status != $Host::TRUE) {
        $this->_set_err("Failed to copy remote file $this->{cmd_nsanity}");
        return;
    }

    # copy the nsanity binary to the custom area
    ($status, @output) = $this->copyRemoteFile("$Host::TOOLS\\$this->{cmd_nsanity}",
                                               "$Host::TOOLS\\Nsanity_custom\\$this->{cmd_nsanity}");

    # update the nSANity commands
    $this->{cmd_custom_nsanity} = "$Host::TOOLS\\Nsanity_custom\\$this->{cmd_nsanity}";
    $this->{cmd_nsanity} = "$Host::TOOLS\\$this->{cmd_nsanity}";

    # copy the nsanity commands.xml file to the custom area
    ($status, @output) = $this->copyRemoteFile($this->{host_common_driveletter} . "$Host::COMMON\\Nsanity\\$this->{nsanity_commands_xml}",
                                               "$Host::TOOLS\\Nsanity_custom\\$this->{nsanity_commands_xml}");
    if ($status != $Host::TRUE) {
        $this->_set_err("Failed to copy remote file $this->{nsanity_commands_xml}");
        return;
    }

    # update the nsanity commands xml path
    $this->{nsanity_commands_xml} = "$Host::TOOLS\\Nsanity_custom\\$this->{nsanity_commands_xml}";

    return;
}

#
#
#
sub _getHostInfo
{
    my ($this) = @_;

    my ($os_type, $status, @output);

    # If this is in a pNATE environment set the drive letter to C:
    if ($this->{pnate_environment} == 1) {
        $this->{host_common_driveletter} = 'C:\\NETAPP';
    } else {
        $this->{host_common_driveletter} = 'X:';
    }

    my $hostName = $this->{hostname};

    # verify we have X mounted and accessible
    $this->_isMinWin();
    goto _getHostInfo_EXIT if ($this->error());

    # these tools need to be run locally, not from a share
    # full path to these is changed during _hostSetup()
    $this->{cmd_mpc}     = "mpc.exe";
    $this->{cmd_spt}     = "spt.exe";
    $this->{cmd_dt}      = $this->{host_common_driveletter} . "$Host::COMMON\\dt.exe";
    $this->{cmd_sio}     = $this->{host_common_driveletter} . "$Host::COMMON\\sio.exe";
    $this->{cmd_fcconfig}= "fcconfig.exe";
    $this->{cmd_hvcontrol} = "hvcontrol.exe";
    $this->{cmd_nsanity} = "Nsanity.exe";
    $this->{nsanity_commands_xml} = "commands.xml";

    # these are used for driver tracing
    $this->{cmd_tracefmt} = "tracing\\tracefmt.exe";
    $this->{cmd_tracelog} = "tracing\\tracelog.exe";
    $this->{cmd_logman}   = "logman.exe";

    $Host::DSMPATH      = "$Host::SYSDRIVE\\progra~1\\netapp\\mpio";

    # snapdrive info and tools
    $Host::SNAPPATH    = "$Host::SYSDRIVE\\progra~1\\netapp\\snapdr~1";
    $this->{cmd_sdcli} = "$Host::SNAPPATH\\sdcli.exe";

    # arch specific utilities and tools
    $this->{cmd_powercfg} = "powercfg.exe";
    $this->{cmd_devcon}   = "devcon.exe";
    $this->{cmd_iscsicli} = "$Host::SYSROOT\\system32\\iscsicli.exe";
    $this->{cmd_diskpart} = "$Host::SYSROOT\\system32\\diskpart.exe";
    $this->{cmd_sc}       = "$Host::SYSROOT\\system32\\sc.exe";

    # data collection tools
    $this->{cmd_win_system_stats} = "win_system_stats.exe";
    $this->{cmd_mpcdata} = "mpcData.bat";
    $this->{cmd_hba_info} = "hba_info.exe";

    # first run 'ver' and determine if this is windows 2000, then
    # run systeminfo if it's 2003 or 2008 or win7, and determine
    # i386, x64, ia64
    ($status, @output) = $this->execute("ver", $this->{timeout});
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_getHostInfo - $hostName failed to obtain version.");
        $this->_set_err(@output);
        goto _getHostInfo_EXIT;
    }
    # Adding the build string to the WinClient object
    # Expecting output such as this: Microsoft Windows XP [Version 5.1.2600]
    my $ver = (grep /Version/, @output)[0];
    $ver =~ m/Version (\d.+)\]/;
    $this->{build} = $1;

    # store the full build version
    if ($ver =~ /\[Version (\S+?)\]/) {
        $this->{full_build} = $1;
    } else {
        $this->{full_build} = $this->{build};
    }

    if (grep(/Windows 2000/, @output))
    {
        &main::logcomment("WinClient::Host::_getHostInfo - Detected Windows 2000 (x86) installation on: $hostName\n");

        # set the hostver and set the path to bin directory
        $this->{hostver}      = "windows 2000 x86";
        $this->{arch}         = "i386";
        $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\i386\\";
        $this->{cmd_diskpart} = "$this->{bin}\\diskpart-20.exe";
        $this->{cmd_sc}       = "$this->{bin}\\sc-20.exe";
    }
    else
    {
        # this must be either windows 2003, 2008 or XP, determine the version
        ($status, @output) = $this->execute("systeminfo", 600);
        if ($status != $Host::TRUE)
        {
            &main::logcomment("WinClient::Host::_getHostInfo - $hostName failed to obtain version.");
            $this->_set_err(@output);
            goto _getHostInfo_EXIT;
        }

        # Add the server model to the winclient object.
        # Expecting something like: System Model:              eserver xSeries 346 -[884011U]-
        my $serverModel = (grep /System\sModel/, @output)[0];
        $serverModel =~ m/System Model:\s+?(\w.+)/i;
        $this->{serverModel} = $1;

        # Add the server domain (or workgroup) to the winclient object.
        # Expecting something like: Domain:                    hq.netapp.com
        my $serverDomain = (grep /^Domain:/i, @output)[0];
        $serverDomain =~ m/Domain:\s+(\S+)/i;
        $this->{domainName} = $1;

        if (grep(/OS\sName:.+2008/, @output) or
            grep(/OS\sName:.+2012/, @output) or
            grep(/OS\sName:.+7\s+Enterprise/i, @output) or
            grep(/OS\sName:.+Server\s+7/i, @output))
        {

            if (grep(/OS\sName:\s+Windows\s+Server\s+7/i, @output)) {
                $os_type = "windows server 7";
            } elsif (grep(/OS\s+Name:\s+Microsoft\s+Windows\s+7\s+Enterprise/i, @output)) {
                $os_type = "windows 7";
            } elsif (grep(/OS\s+Name:\s+Microsoft\s+Windows\s+Server\s+2012\s+Release\s+Candidate\s+(Datacenter|Standard)/i, @output)) {
                $os_type = "windows 2012";
            } elsif (grep(/OS\s+Name:\s+Microsoft\s+Windows\s+Server\s+2012\s+(Datacenter|Standard)/i, @output)) {
                $os_type = "windows 2012";
            } elsif (grep(/OS\s+Name:\s+Microsoft\s+Windows\s+Server 2008 R2\s+(Enterprise|Standard)/i, @output)) {
                $os_type = "windows 2008 R2";
            } else {
                $os_type = "windows 2008";
            }

            # If OS type does not include a SP, look through output and see if an SP is installed and add it to OS type
            if ($os_type !~ /SP/i) {
                my @lines = grep(/OS\s+Version:.*Service\s+Pack\s+\d+\s+Build\s+/i, @output);
                foreach my $line (@lines) {
                    if ($line =~ /OS\s+Version:.*Service\s+Pack\s+(\d+)\s+Build\s+/i) {
                        $os_type .= " SP$1";
                        last;
                    }
                }
            }

            if (grep(/System\sType:.+x64/, @output))
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected $os_type (x64) installation on: $hostName");

                # set the hostver and set the path to bin directory
                $this->{hostver}      = "$os_type x64";
                $this->{arch}         = "AMD64";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\x64\\";
                $this->{cmd_sc}       = "$this->{bin}\\sc.exe";
            }
            elsif (grep(/System\sType:.+Itanium/, @output))
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected $os_type (ia64) installation on: $hostName");
                # set hostver and set the path to bin directory
                $this->{hostver}      = "$os_type ia64";
                $this->{arch}         = "IA64";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\ia64\\";
                $this->{cmd_sc}       = "$this->{bin}\\sc.exe";
            }
            else
            {
                &main::logcomment("WinClient::Host::_gethostInfo - Detected $os_type (i386) installation on: $hostName");
                $this->{hostver}      = "$os_type x86";
                $this->{arch}         = "i386";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\i386\\";
            }
        }
        elsif (grep(/OS\sName:.+Windows(\(R\)|)\s+XP/, @output))
        {
            $os_type = "windows xp";
            if (grep(/System\sType:.+x64/, @output))
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected $os_type (x64) installation on: $hostName");
                # set the hostver and set the path to bin directory
                $this->{hostver}      = "$os_type x64";
                $this->{arch}         = "AMD64";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\x64\\";
                $this->{cmd_diskpart} = "$this->{bin}\\diskpart.exe";
            }
            else
            {
                &main::logcomment("WinClient::Host::_gethostInfo - Detected $os_type (i386) installation on: $hostName");
                # set the hostver and set the path to bin directory
                $this->{hostver}      = "$os_type x86";
                $this->{arch}         = "i386";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\i386\\";
            }
        }
        else
        {
            # this must be windows 2003
            if (grep(/OS\sName:.+x64/, @output))
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected Windows 2003 (x64) installation on: $hostName");

                # set the hostver and set the path to bin directory
                $this->{hostver}      = "windows 2003 x64";
                $this->{arch}         = "AMD64";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\x64\\";
                $this->{cmd_sc}       = "$this->{bin}\\sc.exe";
            }
            elsif (grep(/OS\sName:.+Itanium/, @output))
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected Windows 2003 (ia64) installation on: $hostName");

                # set hostver and set the path to bin directory
                $this->{hostver}      = "windows 2003 ia64";
                $this->{arch}         = "IA64";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\ia64\\";
                $this->{cmd_diskpart} = "$this->{bin}\\diskpart.exe";
                $this->{cmd_sc}       = "$this->{bin}\\sc.exe";
            }
            else
            {
                &main::logcomment("WinClient::Host::_getHostInfo - Detected Windows 2003 (i386) installation on: $hostName");

                # check for the service pack
                $this->{hostver} = "windows 2003 x86";
                foreach my $line (@output) {
                    if ($line =~ /OS\s+Version\:.+?Service Pack (\d+)/i) {
                        $this->{hostver} = "windows 2003 SP" . $1 . " x86";
                    }
                }

                # set the path to bin directory
                $this->{arch}         = "i386";
                $this->{bin}          = $this->{host_common_driveletter} . "\\SAN\\bin\\i386\\";
            }
        }
    }

    # If this is in a pNATE environment set the drive letter to C:
    if ($this->{pnate_environment} == 1) {
        # for diskpart, if 2008 or 2003/xp x86, use installed version
        # with OS version.
        if ($this->{hostver} =~ /2008/i or
            $this->{hostver} =~ /2012/i or
            $this->{hostver} =~ /windows server 7/i or
            $this->{hostver} =~ /windows 2003/i or
            $this->{hostver} =~ /windows xp x86/i) {
        } else {
            if ($this->{arch} =~ /IA64/i) {
                $this->{cmd_diskpart} = "c:\\NETAPP\\\\SAN\\bin\\ia64\\diskpart.exe";
            } else {
                $this->{cmd_diskpart} = "c:\\NETAPP\\diskpart.exe";
            }
        }

        $this->{cmd_tracelog} = "c:\\NETAPP\\SAN\\bin\\common\\" . $this->{cmd_tracelog};
        $this->{cmd_devcon}   = "$this->{bin}\\devcon.exe";
        $this->{cmd_iscsicli} = "$Host::SYSROOT\\system32\\iscsicli.exe";
        $this->{cmd_mpc}      = "c:\\NETAPP\\SAN\\bin\\common\\mpc.exe";
        $this->{cmd_mpcdata} = "echo No MPCDATA with PNATE";
    }

_getHostInfo_EXIT:

    return;

} # END _getHostInfo

#
#
#
sub _getDSMInfo
{
    my ($this) = @_;

    # log DSM information
    &main::logcomment("WinClient::Host::_getDSMInfo - Getting MPIO/DSM information");

    # This command will appear in the logs and then add relevant data to the WinClient object
    my ($status, @dsm_output) = $this->execute("$this->{cmd_mpc} show dsms", 120);
    if ($status != 0){
       &main::logcomment("WinClient::_getDSMInfo - WARNING - An error occurred attempting to run the mpc command");
    }

    # If the mpc utility reports on error, it indicates that there are
    # no dsms installed. It does not indicate an execution problem.
    unless (grep (/ERROR/, @dsm_output)) {
        my @dsm_names = grep (/DsmName:/, @dsm_output);
        my @dsm_versions = grep (/DsmVersion:/, @dsm_output);
        # subract one from the number of array elements, so that this lines up
        # with the foreach loop over the array below which pushes on to the local array.
        my $dsm_count = (scalar(@dsm_names) - 1);
        my @dsm_info = ();

        # Here is a sample of the output of the mpc show dsms command
        # Registered DSMs
        # ---------------
        # Number of DSMs: 1
        #
        #  DsmName: Microsoft iSCSI DSM
        #  DSM Context: 18446738436233988960 (0xFFFFFADF6C458B60)
        #  DsmVersion: 5.2.3790.3640
        #
        # <Additional DSMs may appear>

        foreach my $index (0 .. $dsm_count) {
           # Substitute off the titles of the data supplied by mpc.exe
           # They will be titled on the resulting hash
           $dsm_names[$index] =~ s/\s+DsmName:\s+//;
           $dsm_versions[$index] =~ s/\s+DsmVersion:\s+//;
           # Create an anonymous hash to store the dsm name and version
           my %dsmhash = ('dsmname' => $dsm_names[$index],
                          'dsmversion' => $dsm_versions[$index]);
           # store the hash on the local array since we will overwrite the hash next time through
           push (@dsm_info, \%dsmhash)
        }

        # Store the populated array on the WinClient object.
        @{$this->{dsms}} = @dsm_info;
    }
}

#
#
#
sub _hostSetup
{
    my ($this) = @_;

    if ($this->{pnate_environment} == 1) {
        goto _hostSetup_EXIT;
    }

    my ($status, @output);

    # copy the required utilities to local disk
    # this ensures they are present for manual tests and debugging

    ($status, @output) = $this->copyRemoteFile("$this->{bin}\\$this->{cmd_devcon}",
                                               "$Host::TOOLS\\$this->{cmd_devcon}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_devcon} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_devcon} = "$Host::TOOLS\\$this->{cmd_devcon}";
    }

    my $mpc_source;
    my $win_system_stats_source;
    my $hvcontrol_source;
    if ($this->{full_build} =~ /^6\.[2-9]/) {
        # load the .NET4 builds
        if ($this->{full_build} =~ /^6\.[3-9]/) {
            # load the .NET4.5 mpc builds for win2k12r2 and later
            $mpc_source = $this->{host_common_driveletter} . "$Host::COMMON\\mpc_bin\\mpc_net45.exe";
        } else {
            # load the .NET4 mpc builds for win2k12
            $mpc_source = $this->{host_common_driveletter} . "$Host::COMMON\\mpc_bin\\mpc_net4.exe";
        }
        $win_system_stats_source = $this->{host_common_driveletter} . "$Host::COMMON\\win_system_stats_bin\\win_system_stats_net4.exe";
        $hvcontrol_source = $this->{host_common_driveletter} . "$Host::COMMON\\hvcontrol_bin\\hvcontrol_net4.exe";
    } else {
        # load the .NET3.5 builds for windows 2k8r2 and eariler
        $mpc_source = $this->{host_common_driveletter} . "$Host::COMMON\\mpc_bin\\$this->{cmd_mpc}";
        $win_system_stats_source = $this->{host_common_driveletter} . "$Host::COMMON\\win_system_stats_bin\\$this->{cmd_win_system_stats}";
        $hvcontrol_source = $this->{host_common_driveletter} . "$Host::COMMON\\hvcontrol_bin\\$this->{cmd_hvcontrol}";
    }

    ($status, @output) = $this->copyRemoteFile("$mpc_source",
                                               "$Host::TOOLS\\$this->{cmd_mpc}");

    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_mpc} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_mpc} = "$Host::TOOLS\\$this->{cmd_mpc}";
    }

    # Copy over the spt.exe binary
    ($status, @output) = $this->copyRemoteFile($this->{host_common_driveletter} . "$Host::COMMON\\$this->{cmd_spt}",
                                               "$Host::TOOLS\\$this->{cmd_spt}");

    if ($status != $Host::TRUE)
        {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_spt} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_spt} = "$Host::TOOLS\\$this->{cmd_spt}";
    }

    # Copy over the win system stats binary
    ($status, @output) = $this->copyRemoteFile($win_system_stats_source,
                                               "$Host::TOOLS\\$this->{cmd_win_system_stats}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_win_system_stats} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_win_system_stats} = "$Host::TOOLS\\$this->{cmd_win_system_stats}";
    }

    # Copy over the Hyper-V Command line utility
    ($status, @output) = $this->copyRemoteFile($hvcontrol_source,
                                               "$Host::TOOLS\\$this->{cmd_hvcontrol}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_hvcontrol} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_hvcontrol} = "$Host::TOOLS\\$this->{cmd_hvcontrol}";
    }

    # copy over tools required for driver tracing from Windows DDK
    #  this includes: tracefmt.exe tracelog.exe traceprt.dll

    # copy traceprt.dll

    ($status, @output) = $this->copyRemoteFile("$this->{bin}\\tracing\\traceprt.dll",
                                               "$Host::TOOLS\\traceprt.dll");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of traceprt.dll failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }

    # copy tracefmt.exe

    ($status, @output) = $this->copyRemoteFile("$this->{bin}\\$this->{cmd_tracefmt}",
                                               "$Host::TOOLS\\$this->{cmd_tracefmt}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_tracefmt} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_tracefmt} = "$Host::TOOLS\\$this->{cmd_tracefmt}";
    }

    # copy tracelog.exe

    ($status, @output) = $this->copyRemoteFile("$this->{bin}\\$this->{cmd_tracelog}",
                                               "$Host::TOOLS\\$this->{cmd_tracelog}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_tracelog} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_tracelog} = "$Host::TOOLS\\$this->{cmd_tracelog}";
    }
    # copy mpcData.bat

    ($status, @output) = $this->copyRemoteFile($this->{host_common_driveletter} . "$Host::COMMON\\$this->{cmd_mpcdata}",
                                               "$Host::TOOLS\\$this->{cmd_mpcdata}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_mpcdata} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_mpcdata} = "$Host::TOOLS\\$this->{cmd_mpcdata}";
    }
    # copy hba_info.exe

    ($status, @output) = $this->copyRemoteFile("$this->{bin}\\$this->{cmd_hba_info}",
                                               "$Host::TOOLS\\$this->{cmd_hba_info}");
    if ($status != $Host::TRUE)
    {
        &main::logcomment("WinClient::Host::_hostSetup - Copy of $this->{cmd_hba_info} failed: \n@output");
        $this->_set_err(@output);
        goto _hostSetup_EXIT;
    }
    else
    {
        $this->{cmd_hba_info} = "$Host::TOOLS\\$this->{cmd_hba_info}";
    }

    # make sure the sw iscsi service is enabled (only on win2k8+) and non hypervisor based virtual machines
    if ($this->{hostver} =~ /2008|2012|windows server 7|8/ and $this->{protocol} =~ /iscsi/i and not defined($this->{hypervisor_type})) {
        ($status, @output) = $this->execute("net start \"Microsoft iSCSI Initiator Service\"");
    }

    # For windows 7 systems, disable standby timer by executing powercfg -X -standby-timeout-ac 0
    if ($this->{hostver} =~ /Windows\s+7/i) {
        ($status, @output) = $this->execute("$this->{cmd_powercfg} -X -standby-timeout-ac 0", 60);
        if ($status != 0){
            &main::logcomment("WinClient::Host::_hostSetup - failed to disable standby timeout");
            $this->_set_err("WinClient::Host::_hostSetup - failed to disable standby timeout.");
            goto _hostSetup_EXIT;
        }
    }

_hostSetup_EXIT:

    # common alias'
    $this->{cmd_reboot}      = $this->{cmd_devcon} . " reboot";
    $this->{cmd_dev_rescan}  = $this->{cmd_devcon} . " rescan";
    $this->{cmd_disk_rescan} = $this->{cmd_devcon} . " rescan disks";

    return;

} # END _hostSetup

#
#
#
sub _isMinWin
{
    my ($this) = @_;

    my $hostName = $this->{hostname};

    # If this is a pNATE environment, do not check MinWin
    if ($this->{pnate_environment} == 1) {
        return;
    }
    my $persistent_map = $this->{connection}->{hostp}->{PC_PERSISTENT_MAP_XDRIVE};
    if (!&TestEnv::MinWin::minwin_compatible($hostName, $persistent_map))
    {
        &main::logcomment("WinClient::Host::_isMinWin - MinWin compatibility failed for $hostName, trying again.");
        sleep(30);
        if (!&TestEnv::MinWin::minwin_compatible($hostName))
        {
            &main::logcomment("WinClient::Host::_isMinWin - $hostName is NOT minwin compatible");
            $this->_set_err("WinClient::Host::_isMinWin, $hostName " .
                "is NOT minwin compatible.");
        }
    }

    return;

} # END _isMinwin

#
# Obtain system stats by running win system stats
#
sub getSystemStats
{
    my ($this, $interval) = @_;

    # Collect system stats over the specified interval period
    my ($status, @output) = $this->execute("$this->{cmd_win_system_stats} $interval", $interval + 100);
    if ($status != 0){
       &main::logcomment("WinClient::getSystemStats - failed collect system statisitcs.");
       $this->_set_err("WinClient::Host::getSystemStats, failed to collect system statistics.");
       return (wantarray() ? () : []);
    }

    return (wantarray() ? @output : \@output);
}

#
# Run the HVControl command line utility with the specified arguments
#
sub hvcontrol
{
    my ($this, $args, $timeout) = @_;

    $this->_clear_err();

    # default the timeout
    $timeout = $this->{timeout} if (not defined($timeout));

    # execute the hvcontrol utility with the passed in arguments
    my ($status, @output) = $this->execute("$this->{cmd_hvcontrol} $args", $timeout);
    if ($status != 0){
       &main::logcomment("WinClient::hvcontrol - failed to execute hvcontrol with args ($args).");
       $this->_set_err("Failed to execute hvcontrol with args ($args).");
       return ($status, @output);
    }

    return ($status, @output);
}

#
# Run the nsanity command line utility with the specified arguments
#
sub nsanity
{
    my ($this, $args, $use_internal_commands, $timeout) = @_;

    $this->_clear_err();

    # default the timeout
    $timeout = $this->{timeout} if (not defined($timeout));

    # default the internal commands flag to off
    if (not defined($use_internal_commands)) {
        $use_internal_commands = 0;
    }

    my $cmd;
    if ($use_internal_commands == 1) {
        # use the regular nsanity binary
        $cmd = $this->{cmd_nsanity};
    } else {
        # use the nsanity binary that has the custom commands.xml file
        $cmd = $this->{cmd_custom_nsanity};
    }

    # execute the nsanity utility with the passed in arguments
    my ($status, @output) = $this->execute("$cmd $args", $timeout);
    if ($status != 0) {
       $this->_set_err("Failed to execute nsanity with args ($args).");
       return ($status, @output);
    }

    return ($status, @output);
}

#
# Verify that $this->domainName is really a domain (not just a workgroup)
# and that the client is a member of said domain
sub verifyDomain
{
    my ($this) = @_;

    # Take the domainName and try to read the netlogon directory
    # Access Denied means you are not part of the domain
    # A network failure means the domain doesn't exist. Its probably a workgroup.
    my $cmd = "dir \\\\$this->{domainName}\\netlogon";
    my ($status, @output) = $this->execute($cmd);
    if ($status != 0) {
       $this->_set_err("WinClient::Host::verifyDomain - failed to execute command: $cmd.");
       return "";
    }

    if ((grep (/Configuration\s+information\s+could\s+not/i, @output) or (grep (/Logon\s+failure/i, @output)))) {
       $this->_set_err("WinClient::Host::verifyDomain - failed to confirm domain membership " .
                       "in $this->{domainName}.");
       return "";
    } elsif (grep (/not\s+found/i, @output)) {
       return "workgroup";
    }
    return "domain";
}

############################################################################
# Method:    stopService
# Objective: stop a Windows service
############################################################################
sub stopService{
    my ($this, %args) = @_;
    my ($service_name, $msg);
    # Process arguments.
    foreach (keys %args) {
        if (/^-service_name$/i) {
            if (!length($service_name = $args{$_})) {
                $msg = "Host::stopService - invalid value for -service_name: $args{$_}";
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    my $cmd = "sc stop " . $service_name;
    my ($status, @output) = $this->execute($cmd);
    if ($status != $Host::TRUE) {
         $msg = "Host::stopService Command failed: $cmd";
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }
    my $stop_count = 0;
    while($this->isServiceRunning(-service_name => $service_name)){
        &SanUtil::nate_sleep(10);
        if($stop_count >= 3){
            $msg = "Host::stopService - Command failed: stop service ".
                      "timed out";
            $this->_set_err($msg);
            return $SanUtil::FALSE;
        }
        $stop_count++;
    }
    return $SanUtil::TRUE;
} # END stopService

############################################################################
# Method:    startService
# Objective: start a Windows service
# Details:   See POD documentation at the end of this file.
############################################################################
sub startService{
    my ($this, %args) = @_;
    my ($service_name, $msg);

    # Process arguments.
    foreach (keys %args) {
        if (/^-service_name$/i) {
            if (!length($service_name = $args{$_})) {
                $msg = "Host::startService - invalid value for -service_name: $args{$_}";
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    my $cmd = "sc start " . $service_name;
    my ($status, @output) = $this->execute($cmd);
    if ($status != $Host::TRUE) {
        $msg = "Host::startService Command failed: $cmd";
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }
    my $running_count = 0;
    while(not $this->isServiceRunning(-service_name => $service_name)){
        &SanUtil::nate_sleep(10);
        if($running_count >= 3){
            $msg = "Host::startService - Command failed: start service ".
                      "timed out";
            $this->_set_err($msg);
            return $SanUtil::FALSE;
        }
        $running_count++;
    }
    return $SanUtil::TRUE;
} # END startService

############################################################################
# Method:    setServiceStartParam
# Objective: set the start option for a Windows service
# Details:   See POD documentation at the end of this file.
############################################################################
sub setServiceStartParam{
    my ($this, %args) = @_;
    my ($service_name, $start_param, $msg);

    # Process arguments.
    foreach (keys %args) {
        if (/^-service_name$/i) {
            if (!length($service_name = $args{$_})) {
                $msg = "Host::setServiceStartParam - invalid value for -service_name: $args{$_}";
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
        if (/^-start_param$/i) {
            if (!length($start_param = $args{$_})) {
                $msg = "Host::setServiceStartParam - invalid value for -start_param: $args{$_}";
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    my $cmd = "sc config " . $service_name . " start= " . $start_param;
    my ($status, @output) = $this->execute($cmd);
    if ($status != $Host::TRUE or !grep(/SUCCESS/i, @output)) {
        $msg = "Host::setServiceStartParam Command failed: $cmd";
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # END setServiceStartParam

############################################################################
# Method:    isServiceRunning
# Objective: check if a windows service is running or not
# Details:   See POD documentation at the end of this file.
############################################################################
sub isServiceRunning{
    my ($this, %args) = @_;
    my ($service_name, $msg);

    # Process arguments.
    foreach (keys %args) {
        if (/^-service_name$/i) {
            if (!length($service_name = $args{$_})) {
                $msg = "Host::isServiceRunning - invalid value for -service_name: $args{$_}";
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    my $cmd = "sc query " . $service_name;
    my ($status, @output) = $this->execute($cmd);

    if ($status ne $Host::TRUE) {
        $msg = "ISCSI::isServiceRunning Command failed: $cmd";
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }
    foreach my $line (@output) {
        if ($line =~ /STATE/i){
            if ($line =~ /RUNNING/){
                return $SanUtil::TRUE;
            }
        }
    }
    return $SanUtil::FALSE;
} # END isServiceRunning

############################################################################
# Method:    get_service_config
# Objective: get information about a Windows service configuration
# Details:   See POD documentation at the end of this file.
############################################################################
sub get_service_config{
    my ($this, %args) = @_;
    my $return_value = "";
    my ($service_name, $config_param, $msg);

    # Process arguments.
    foreach (keys %args) {
        if (/^-service_name$/i) {
            if (!length($service_name = $args{$_})) {
                $msg = "Host::get_service_config - invalid value for -service_name: $args{$_}";
                $this->_set_err($msg);
                return $return_value;
            }
            next;
        }
        if (/^-config_param$/i) {
            if (!length($config_param = $args{$_})) {
                $msg = "Host::get_service_config - invalid value for -config_param: $args{$_}";
                $this->_set_err($msg);
                return $return_value;
            }
            next;
        }
    }

    my $cmd = "sc qc " . $service_name;
    my ($status, @output) = $this->execute($cmd);
    if ($status != $Host::TRUE or !grep(/SUCCESS/i, @output)) {
        $msg = "Host::get_service_config Command failed: $cmd";
        $this->_set_err($msg);
        return $return_value;
    }
    foreach my $line (@output) {
        chomp($line);
        if ($line =~ /$config_param/i){
            $line =~ /.*:\s+\d\s+(\w+)/;
            $return_value = $1;
            last;
        }
    }
    return $return_value;
} # END get_service_config


1;
