# $Id: //depot/prod/test/main/lib/WinClient/ISCSI.pm $/
#
# Copyright (c) 2007-2009 Network Appliance, Inc.
# All rights reserved.
#
#
## @summary
## Library for implementing ISCSI CLI for Windows. This library is used
## to execute iscsi commands using mpc and parse results. Each of these
## subroutines can be called by a Winclient Host object.
##
## @author  Lopa Sarangi: sarangi@netapp.com
## @author Robert Hunter: hrivera@netapp.com
## @status  public
## @pod here
##

package WinClient::ISCSI;

#
# Compiler directives
#
use strict;
use lib();
use San::SanUtil;

########################################################################
#                            Public Methods
########################################################################

sub iscsiShowAdapters;
sub iscsiShow;
sub iscsiShowConnections;
sub iscsiAddConnection;
sub iscsiRemoveConnection;
sub iscsiAddTargetPortal;
sub iscsiRemoveTargetPortal;
sub iscsiLogin;
sub iscsiLogout;
sub iscsiSetMutualCHAPSecret;
sub iscsiAddiSNSServer;
sub iscsiRemoveiSNSServer;

########################################################################
# Method:    iscsiShowAdapters
# Objective: Display details of iscsi adapters on the Windows Host
# Details:   See POD documentation at the end of this file.
########################################################################

sub iscsiShowAdapters
{
    my ($this) = @_;
	
    $this->_clear_err();

    my ($status, @output, $cmd, $msg, $line);
	
    # verify if the protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiShowAdapters_EXIT;
    }

    # build the iscsicli command
    $cmd = "$this->{cmd_mpc} iscsi show adapters";

    # execute the command
    ($status, @output) = $this->execute($cmd);

    if ($status != $Host::TRUE || !grep(/Success: iscsi show adapters/i, @output))
    {
        $msg = "ERROR: ISCSI::iscsiShowAdapters - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiShowAdapters_EXIT;
    }


iscsiShowAdapters_EXIT:
	
    return($status, @output);

} # END iscsiShowAdapters

############################################################################
# Method:    iscsiShow
# Objective: Display the details of configured iscsi targetportals, targets
#            and active sessions on the Windows Host
# Details:   See POD documentation at the end of this file.
############################################################################

sub iscsiShow
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my($status, @output, $line, $cmd, $object, $msg);

    foreach (keys %args)
    {
        if(/^-object$/i)
        {
            if (!length($object = $args{$_}) || $object !~ /(?:targetportals|targets|sessions|isnsservers)/i)
            {
                $msg = "ISCSI::iscsiShow - invalid value for -object: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE; 
                goto iscsiShow_EXIT;
            }
            next;
        }
    }

    # make sure atleast one object is defined as it is a strict argument
    if (!defined($object) or $object=~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiShow - Strict parameter $object not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiShow_EXIT;
    }

    # verify if the protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiShow_EXIT;
    }

    if ($object =~ /isnsservers/i and ($this->{hostver} =~ /2008/ or 
                                       $this->{hostver} =~ /windows server 7/i)) {
        # manually use wmic to determine the isnsserver list for win2k8 machines only
        # known issue with "mpc iscsi show isnsservers" that does not display isnsservers properly in win2k8
        $cmd = "wmic /NAMESPACE:\\\\root\\wmi PATH MSiSCSIInitiator_iSNSServerClass GET iSNSServerAddress /VALUE";
        # execute the command
        ($status, @output) = $this->execute($cmd);
        if ($status != $Host::TRUE) {
            $msg = "ISCSI::iscsiShow - Command failed: $cmd";
            &main::logcomment($msg);
            $this->_set_err($msg);
            goto iscsiShow_EXIT;            
        }
    } else {
        # build the command in mpc
        $cmd = "$this->{cmd_mpc} iscsi show $object ";
        
        # execute the command
        ($status, @output) = $this->execute($cmd);
        if ($status != $Host::TRUE || !grep(/Success: iscsi show $object/i, @output)) {
            $msg = "ISCSI::iscsiShow - Command failed: $cmd";
            &main::logcomment($msg);
            $this->_set_err($msg);
            goto iscsiShow_EXIT;
        }
    }

iscsiShow_EXIT:

    return($status, @output);

} # END iscsiShow


################################################################################
# Method:    iscsiShowConnections
# Objective: Display the details of active iscsi connections on the Windows Host
# Details:   See POD documentation at the end of this file.
################################################################################

sub iscsiShowConnections
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $line, $cmd, $sessionid, $msg);

    foreach (keys %args)
    {
        if(/^-sessionid$/i)
        {
            if (!length($sessionid = $args{$_})) 
            {
                $msg = "ISCSI::iscsiShowConnections - invalid value for -sessionid: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiShowConnections_EXIT;
            }
            next;
        } 
    }

    # verify if the protocol used is iscsi  
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiShowConnections_EXIT;

    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi show connections $sessionid";

    # execute the command
    ($status, @output) = $this->execute($cmd);

    if ($status != $Host::TRUE || !grep(/Success: iscsi show connections/i, @output)) 
    {
        $msg = "ISCSI::iscsiShowConnections - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiShowConnections_EXIT;
    }


iscsiShowConnections_EXIT:

    return($status, @output);

} # END iscsiShowConnections

###########################################################################
# Method:    iscsiAddConnection
# Objective: Add a new connection between an iscsi target and an initiator
#            with the specified parameters.
# Details:   See POD documentation at the end of this file.
###########################################################################

sub iscsiAddConnection
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $line, $cmd, $sessionid, $targetindex, $initiatorindex, $msg);

    my ($mpio, $persistent, $headerdigest, $datadigest, $auth_type, $chap_username, $chap_password,
        $default_time2wait, $default_time2retain, $maximumconnections);

    $mpio = 0; 
    $persistent = 0;
    $headerdigest = 0;
    $datadigest = 0;
    $chap_username = "";
    $chap_password = "";
    $auth_type = "none";
    
    foreach (keys %args)
    {
        if(/^-sessionid$/i)
        {
            if (!length($sessionid = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -sessionid: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if(/^-targetindex$/i)
        {
            if (!length($targetindex = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -targetindex: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if(/^-initiatorindex$/i)
        {
            if (!length($initiatorindex = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -initiatorindex: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if(/^-mpio$/i)
        {
            if (!length($mpio = $args{$_}) ||  $mpio !~ /(?:1|0)/i)
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -mpio: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if (/^-persistent$/i)
        {
            if (!length($persistent = $args{$_}) ||  $persistent !~ /(?:1|0)/i)
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -persistent: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if (/^-headerdigest$/i)
        {
            if (!length($headerdigest = $args{$_}) ||  $headerdigest !~ /(?:1|0)/i)
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -headerdigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        }
        if (/^-datadigest$/i)
        {
            if (!length($datadigest = $args{$_}) || $datadigest !~ /(?:1|0)/i)
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -datadigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;
     	}
        if (/^-auth_type$/i)
        {
            if (!length($auth_type = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddConnection - invalid value for -auth_type: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddConnection_EXIT;
            }
            next;

        } 
        if (/^-default_time2retain$/i)
        {
            $default_time2retain = $args{$_};
            next;
        }
        if (/^-default_time2wait$/i)
        {
            $default_time2wait = $args{$_};
            next;
        }
        if (/^-maxconnections$/i)
        {
            $maximumconnections = $args{$_};
            next;
        }
        if (/^-chap_username$/i)
        {
            $chap_username = $args{$_};
            next;
        } 

        if (/^-chap_password$/i)
        {
            $chap_password = $args{$_};
            next;
        } 
    }

    # make sure sessionid is defined as it is a strict argument
    if (!defined($sessionid) or $sessionid =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiAddConnection - Strict argument $sessionid not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddConnection_EXIT;
    }

    # make sure targetindex is defined as it is a strict argument
    if (!defined($targetindex) or $targetindex =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiAddConnection - Strict argument $targetindex not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddConnection_EXIT;
    }

    # make sure initiatorindex is defined as it is a strict argument
    if (!defined($initiatorindex) or $initiatorindex =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiAddConnection - Strict argument $initiatorindex not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddConnection_EXIT;
    }

    #verify if protocol is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddConnection_EXIT;
    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi add connection $sessionid $targetindex $initiatorindex ";

    if($mpio)
    {
        $cmd = $cmd . "mpio ";
    }
    if($persistent)
    {
        $cmd = $cmd . "persistent ";
    }
    if($headerdigest)
    {
        $cmd = $cmd . "headerdigest ";
    }
    if($datadigest)
    {
        $cmd = $cmd . "datadigest ";
    }
    if ($auth_type =~ /^(uni|bi)directional$/i)
    {
        my $type = $1;

        $auth_type = 1 if ($type =~ /uni/i);
        $auth_type = 2 if ($type =~ /bi/i);
        
        $cmd = $cmd . "authtype $auth_type ";
        if ($chap_username ne "") {
            $cmd = $cmd . "chap_username $chap_username ";
        }
        if ($chap_password ne "") {
            $cmd = $cmd . "chap_password $chap_password ";
        }
    }
    if ($default_time2retain ne "") {
        $cmd = $cmd . "defaulttime2retain $default_time2retain ";
    }
    if ($default_time2wait ne "") {
        $cmd = $cmd . "defaulttime2wait $default_time2wait ";
    }
    if ($maximumconnections ne "") {
        $cmd = $cmd . "maximumconnections $maximumconnections ";
    }

    # execute the command
    ($status, @output) = $this->execute($cmd);

    if($status != $Host::TRUE || !grep(/Success: iscsi add connection/i, @output)) 
    {
        $msg = "ISCSI::iscsiAddConnection - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiAddConnection_EXIT;
    }

iscsiAddConnection_EXIT:

    return($status, @output);

} # END iscsiAddConnection

########################################################################
# Method:    iscsiRemoveConnection
# Objective: Remove an existing connection between an iscsi target and a
#            Windows iscsi host/initiator
# Details:   See POD documentation at the end of this file.
########################################################################

sub iscsiRemoveConnection
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $line, $cmd, $connectionid, $msg);
    my $retry_count = 1;

    foreach (keys %args)
    {
        if(/^-connectionid$/i)
        {
            if (!length($connectionid = $args{$_}))
            {
                $msg = "ISCSI::iscsiRemoveConnection - invalid value for -connectionid: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiRemoveConnection_EXIT;
            }
            next;
        } 
        elsif (/^-retry_count$/i) 
        {
            if (!length($retry_count = $args{$_}))
            {
                $msg = "ISCSI::iscsiRemoveConnection - invalid value for -retry_count: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiRemoveConnection_EXIT;
            }
            next;
        }
    }

    # make sure connectionid is defined as it is a strict argument
    if (!defined($connectionid) or $connectionid =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiRemoveConnection - Strict argument $connectionid not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveConnection_EXIT;
    }
	
    #verify if protocol is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveConnection_EXIT;
    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi remove connection $connectionid ";
    my $success = 0;
       
    # retry the command $retry_count times
    for (my $i = 1; $i <= $retry_count; $i++) {
        # execute the command 
        ($status, @output) = $this->execute($cmd);
        
        if($status != $Host::TRUE || !grep(/Success: iscsi remove connection/i, @output)) {
            # sleep and retry if not on the last attempt
            last if ($i == $retry_count);
            $msg = "ISCSI::iscsiRemoveConnection - Retrying Command ($cmd) (retry_count=$i/$retry_count)";
            &main::logcomment($msg);
            &SanUtil::nate_sleep(10);
            next;
        } 
        $success = 1;
        last;
    }

    if (not $success) {
        $msg = "ISCSI::iscsiRemoveConnection - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
    }

iscsiRemoveConnection_EXIT:

    return($status, @output);

} # END iscsiRemoveConnection


############################################################################
# Method:    iscsiAddTargetPortal
# Objective: Add a targetportal to the iscsi Windows host with the specified
#            parameters
# Details:   See POD documentation at the end of this file.
############################################################################

sub iscsiAddTargetPortal
{

    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $line, $cmd, $filer, $msg);

    my ($mpio, $headerdigest, $datadigest, $initiator_index, $chap_username, 
        $chap_password, $auth_type, $default_time2wait, $default_time2retain, $maximumconnections);

    $mpio = 0;
    $headerdigest = 0;
    $datadigest = 0;
    $initiator_index = "";
    $chap_username = "";
    $chap_password = "";
    $auth_type = "none";

    foreach (keys %args)
    {
        if(/^-filer$/i)
        {
            if (!length($filer = $args{$_}))
            { 
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -filer: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;
                	
        }
        if(/^-mpio$/i)
        {
            if (!length($mpio = $args{$_}) ||  $mpio !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -mpio: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;

        }
        if(/^-headerdigest$/i)
        {
            if (!length($headerdigest = $args{$_}) ||  $headerdigest !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -headerdigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;	

        }
        if(/^-datadigest$/i)
        {
            if (!length($datadigest = $args{$_}) ||  $datadigest !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -datadigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;

        } 
        if(/^-initiator_index$/i)
        {
            if (!length($initiator_index = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -initiator_index: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;

        } 

        if (/^-auth_type$/i)
        {
            if (!length($auth_type = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddTargetPortal - invalid value for -auth_type: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiAddTargetPortal_EXIT;
            }
            next;

        } 
        if (/^-default_time2retain$/i)
        {
            $default_time2retain = $args{$_};
            next;

        }
        if (/^-default_time2wait$/i)
        {
            $default_time2wait = $args{$_};
            next;
        }
        if (/^-maxconnections$/i)
        {
            $maximumconnections = $args{$_};
            next;
        }
        if (/^-chap_username$/i)
        {
            $chap_username = $args{$_};
            next;
        }
        if (/^-chap_password$/i)
        {
            $chap_password = $args{$_};
            next;
        }

    }

    # make sure filername/IP is defined as it is a strict argument
    if (!defined($filer) or $filer =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiAddTargetPortal - Strict argument $filer not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddTargetPortal_EXIT;
    }

    #verify if protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiAddTargetPortal_EXIT;
    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi add targetportal $filer ";

    if ($initiator_index ne "") 
    {
        $cmd = $cmd . "$initiator_index ";
    }
    if($mpio == 1)
    {
        $cmd = $cmd . "mpio ";
    }
    if($headerdigest == 1)
    {
        $cmd = $cmd . "headerdigest ";
    }
    if($datadigest == 1)
    {
        $cmd = $cmd . "datadigest ";
    }
    if ($auth_type =~ /^(uni|bi)directional$/i)
    {
        my $type = $1;

        $auth_type = 1 if ($type =~ /uni/i);
        $auth_type = 2 if ($type =~ /bi/i);
        
        $cmd = $cmd . "authtype $auth_type ";
        if ($chap_username ne "") {
            $cmd = $cmd . "chap_username $chap_username ";
        }
        if ($chap_password ne "") {
            $cmd = $cmd . "chap_password $chap_password ";
        }
    }
    if ($default_time2retain ne "") {
        $cmd = $cmd . "defaulttime2retain $default_time2retain ";
    }
    if ($default_time2wait ne "") {
        $cmd = $cmd . "defaulttime2wait $default_time2wait ";
    }
    if ($maximumconnections ne "") {
        $cmd = $cmd . "maximumconnections $maximumconnections ";
    }

    # execute the command 
    ($status, @output) = $this->execute($cmd);

    if ($status != $Host::TRUE || grep(/Failed: iscsi add targetportals/i, @output))
    {
        $msg = "ISCSI::iscsiAddTargetPortal - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiAddTargetPortal_EXIT;
    }
	

iscsiAddTargetPortal_EXIT:

    return($status, @output);

} # END AddTargetPortal

########################################################################
# Method:    iscsiRemoveTargetPortal
# Objective: Remove an existing targetportal from the Windows iscsi host
# Details:   See POD documentation at the end of this file.
########################################################################

sub iscsiRemoveTargetPortal
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my($status, @output, $line, $cmd, $filer, $msg);

    foreach (keys %args)
    {
        if(/^-filer$/i)
        { 
            if (!length($filer = $args{$_}))
            {
                $msg = "ISCSI::iscsiRemoveTargetPortal - invalid value for -filer: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiRemoveTargetPortal_EXIT;
            }
            next;
        }
    }

    # make sure filername/IP is defined as it is a strict argument
    if (!defined($filer) or $filer =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiRemoveTargetPortal - Strict argument $filer not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveTargetPortal_EXIT;
    }

    #verify if protocol is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveTargetPortal_EXIT;
    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi remove targetportal $filer ";

    # execute the command 
    ($status, @output) = $this->execute($cmd);

    if($status != $Host::TRUE || !grep(/Success: iscsi remove targetportal/i, @output)) 
    {
        $msg = "ISCSI::iscsiRemoveTargetPortal - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiRemoveTargetPortal_EXIT;
    }

iscsiRemoveTargetPortal_EXIT:

    return($status, @output);

} # END iscsiRemoveTargetPortal

#########################################################################
# Method:    iscsiLogin
# Objective: Login to an iscsi target from a Windows iscsi host/initiator
# Details:   See POD documentation at the end of this file.
#########################################################################

sub iscsiLogin
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $line, $cmd, $targetname, $targetindex, $initiatorindex, $msg);

    my ($mpio, $persistent, $headerdigest, $datadigest, $chap_username,
        $chap_password, $auth_type, $default_time2wait, $default_time2retain, $maximumconnections);

    $mpio = 0; 
    $persistent = 0;
    $headerdigest = 0;
    $datadigest = 0;
    $chap_username = "";
    $chap_password = "";
    $auth_type = "none";

    foreach (keys %args)
    {
        if(/^-targetname$/i)
        {
            if (!length($targetname = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -targetname: $args{$_}";
                &main::logcomment($msg);
                this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-targetindex$/i)
        {
            if (!length($targetindex = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -targetindex: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-initiatorindex$/i)
        {
            if (!length($initiatorindex = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -initiatorindex: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-mpio$/i)
        {
            if (!length($mpio = $args{$_}) ||  $mpio !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -mpio: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-persistent$/i)
        {  
            if (!length($persistent = $args{$_}) ||  $persistent !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -persistent: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-headerdigest$/i)
        {
            if (!length($headerdigest = $args{$_}) ||  $headerdigest !~ /(?:0|1)/i)
            { 
                $msg = "ISCSI::iscsiLogin - invalid value for -headerdigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        }
        if(/^-datadigest$/i)
        {
            if (!length($datadigest = $args{$_}) ||  $datadigest !~ /(?:0|1)/i)
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -datadigest: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;
        }
        if (/^-auth_type$/i)
        {
            if (!length($auth_type = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogin - invalid value for -auth_type: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogin_EXIT;
            }
            next;

        } 
        if (/^-default_time2retain$/i)
        {
            $default_time2retain = $args{$_};
            next;
        }
        if (/^-default_time2wait$/i)
        {
            $default_time2wait = $args{$_};
            next;
        }
        if (/^-maxconnections$/i)
        {
            $maximumconnections = $args{$_};
            next;
        }
        if (/^-chap_username$/i)
        {
            $chap_username = $args{$_};
            next;
        }
        if (/^-chap_password$/i)
        {
            $chap_password = $args{$_};
            next;
        }
    }    

    # make sure targetname is defined as it is a strict argument
    if (!defined($targetname) or $targetname =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiLogin - Strict argument $targetname not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogin_EXIT;
    }

    # make sure targetindex is defined as it is a strict argument
    if (!defined($targetindex) or $targetindex =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiLogin - Strict argument $targetindex not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogin_EXIT;
    }

    # make sure initiatorindex is defined as it is a strict argument
    if (!defined($initiatorindex) or $initiatorindex =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiLogin - Strict argument $initiatorindex not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogin_EXIT;
    }

    #verify if protocol is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    { 
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogin_EXIT;
    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi login $targetname $targetindex $initiatorindex ";

    if($mpio == 1)
    {
        $cmd = $cmd . "mpio ";
    }
    if($persistent == 1)
    {
        $cmd = $cmd . "persistent ";
    }
    if($headerdigest == 1)
    {
        $cmd = $cmd . "headerdigest ";
    }
    if($datadigest == 1)
    {
        $cmd = $cmd . "datadigest ";
    }
    if ($auth_type =~ /^(uni|bi)directional$/i)
    {
        my $type = $1;

        $auth_type = 1 if ($type =~ /uni/i);
        $auth_type = 2 if ($type =~ /bi/i);
        
        $cmd = $cmd . "authtype $auth_type ";
        if ($chap_username ne "") {
            $cmd = $cmd . "chap_username $chap_username ";
        }
        if ($chap_password ne "") {
            $cmd = $cmd . "chap_password $chap_password ";
        }
    }
    if ($default_time2retain ne "") {
        $cmd = $cmd . "defaulttime2retain $default_time2retain ";
    }
    if ($default_time2wait ne "") {
        $cmd = $cmd . "defaulttime2wait $default_time2wait ";
    }
    if ($maximumconnections ne "") {
        $cmd = $cmd . "maximumconnections $maximumconnections ";
    }

    # execute the command 
    ($status, @output) = $this->execute($cmd);

    if($status != $Host::TRUE || !grep(/Success: iscsi login target/i, @output)) 
    {
        $msg = "ISCSI::iscsiLogin - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiLogin_EXIT;
    }

iscsiLogin_EXIT:

    return($status, @output);

} # END iscsiLogin

##########################################################################
# Method:    iscsiLogout
# Objective: Logout of an active iscsi session between an iscsi target and
#            a Windows iscsi host/initiator.
# Details:   See POD documentation at the end of this file.
##########################################################################

sub iscsiLogout
{
    my($this, %args) = @_;

    $this->_clear_err();

    my($status, @output, $line,$cmd, $sessionid, $msg);
    my $retry_count = 5;

    foreach (keys %args)
    {
        if(/^-sessionid$/i)
        {
            if (!length($sessionid = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogout - invalid value for -sessionid: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogout_EXIT;
            }
            next;
        }
        elsif (/^-retry_count$/i) 
        {
            if (!length($retry_count = $args{$_}))
            {
                $msg = "ISCSI::iscsiLogout - invalid value for -retry_count: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiLogout_EXIT;
            }
            next;
        }
    }

    # make sure sessionid is defined as it is a strict argument
    if (!defined($sessionid) or $sessionid =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiLogout - Strict argument $sessionid not specified";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogout_EXIT;
    }

    # verify if protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiLogout_EXIT;

    }

    # build the command
    $cmd = "$this->{cmd_mpc} iscsi logout $sessionid ";
    my $success = 0;

    # retry the command $retry_count times
    for (my $i = 1; $i <= $retry_count; $i++) {
        # execute the command 
        ($status, @output) = $this->execute($cmd);
        
        if($status != $Host::TRUE || !grep(/Success: iscsi logout target/i, @output)) {
            # sleep and retry if not on the last attempt
            last if ($i == $retry_count);
            $msg = "ISCSI::iscsiLogout - Retrying Command ($cmd) (retry_count=$i/$retry_count)";
            &main::logcomment($msg);
            &SanUtil::nate_sleep(10);
            next;
        } 
        $success = 1;
        last;
    }

    if (not $success) {
        $msg = "ISCSI::iscsiLogout - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
    }

iscsiLogout_EXIT:

    return($status, @output);
} # END iscsiLogout

#########################################################################
# Method:    iscsiSetMutualCHAPSecret
# Objective: Set the mutual chap secret on the initiator
# Details:   See POD documentation at the end of this file.
#########################################################################

sub iscsiSetMutualCHAPSecret
{
    my ($this, %args) = @_;
    my ($msg, $status, @output);

    $this->_clear_err();

    my $secret = "";

    foreach (keys %args)
    {
        if(/^-secret$/i)
        {
            if (!length($secret = $args{$_}))
            {
                $msg = "ISCSI::iscsiSetMutualCHAPSecret - invalid value for -secret: $args{$_}";
                &main::logcomment($msg);
                this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiSetMutualCHAPSecret_EXIT;
            }
            next;

        }
    }    

    # make sure targetname is defined as it is a strict argument
    if (!defined($secret) or $secret =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiSetMutualCHAPSecret - Strict argument -secret not defined";
        &main::logcomment($msg);
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiSetMutualCHAPSecret_EXIT;
    }

    my $cmd = "$this->{cmd_mpc} iscsi add mutualchapsecret $secret";
    # execute the command 
    ($status, @output) = $this->execute($cmd);

    if($status != $Host::TRUE || !grep(/Success: iscsi add mutualchapsecret/i, @output)) 
    {
        $msg = "ISCSI::iscsiSetMutualCHAPSecret - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        goto iscsiSetMutualCHAPSecret_EXIT;
    }

iscsiSetMutalCHAPSecret_EXIT:

    return($status, @output);

} # END iscsiSetMutualCHAPSecret

############################################################################
# Method:    iscsiAddiSNSServer
# Objective: Add an iSNS Server to the host via mpc
# Details:   See POD documentation at the end of this file.
############################################################################
sub iscsiAddiSNSServer
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $cmd, $ip_address, $msg);

    # verify if protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        &main::logcomment($msg);
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }

    foreach (keys %args)
    {
        if(/^-ip_address$/i)
        {
            if (!length($ip_address = $args{$_}))
            {
                $msg = "ISCSI::iscsiAddiSNSServer - invalid value for -ip_address: $args{$_}";
                &main::logcomment($msg);
                $this->_set_err($msg);
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    # get the initial firewall start config so that it can be set back
    # to the original after the firewall setting is started.
    # start param can be <boot,system,auto,demand,disabled,delayed-auto>
    my $initial_start_config = $this->get_service_config(-service_name => "mpssvc",
                                                         -config_param => "start_type" );
    if ($initial_start_config eq ""){
        return $SanUtil::FALSE;
    }

    my $service_initial_state = 0;
    # query the firewall service to see if it is running
    if( $SanUtil::TRUE ne ($service_initial_state = $this->isServiceRunning(-service_name => "mpssvc"))){
        # if the firewall service isn't running, set start service to on
        # demand, and start the service
        if (not $this->setServiceStartParam(-service_name => "mpssvc", -start_param => "demand")){
            return $SanUtil::FALSE;
        }
        if (not $this->startService(-service_name => "mpssvc")){
            return $SanUtil::FALSE;
        }
    }

    if ($this->{hostver} =~ /2008 R2/ or $this->{hostver} =~ /windows 7/i) {
        #create firewall exempt command
        my $cmd = "$this->{cmd_iscsicli} firewallexemptisnsserver";

        # execute the command
        my ($status, @output) = $this->execute($cmd);
        if ($status != $Host::TRUE or !grep(/completed successfully/i, @output)) {
            my $msg = "ISCSI::firewallexemptisnsserver - Command failed: $cmd";
            $this->_set_err($msg);
            return $SanUtil::FALSE;
        }
    }
    # build the command
    $cmd = "$this->{cmd_mpc} iscsi add isnsserver $ip_address";

    # execute the command
    ($status, @output) = $this->execute($cmd);

    if ($status != $Host::TRUE or grep(/Failed: iscsi add isnsserver/i, @output))
    {
        $msg = "ISCSI::iscsiAddiSNSServer - Command failed: $cmd";
        &main::logcomment($msg);
        $this->_set_err($msg);
        return $SanUtil::FALSE;
    }

    # create a hash of legal service start command
    # status output to command inputs
    my $service_start_param_hash = {AUTO_START   => "auto",
                                    DEMAND_START => "demand",
                                    DISABLED     => "disabled"};

    # get the start command input based off the initial config
    my $init_start_cmd_param = $service_start_param_hash->{$initial_start_config};
    unless (($init_start_cmd_param eq "") or ($init_start_cmd_param eq undef)) {
        if (not $this->setServiceStartParam(-service_name => "mpssvc",
                                            -start_param  => $init_start_cmd_param)){
            return $SanUtil::FALSE;
        }
    }

    # if the firewall was stopped before running these functions,
    # stop it again at the end.
    if ($service_initial_state ne 1){
        if (not $this->stopService(-service_name => "mpssvc")){
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} # END iscsiAddiSNSServer

############################################################################
# Method:    iscsiRemoveiSNSServer
# Objective: Remove an iSNS Server from the host via mpc
# Details:   See POD documentation at the end of this file.
############################################################################
sub iscsiRemoveiSNSServer
{
    my ($this, %args) = @_;

    $this->_clear_err();

    my ($status, @output, $cmd, $ip_address, $msg);

    # verify if protocol used is iscsi
    if ($this->{protocol} !~ /iscsi/i)
    {
        $msg = "ERROR: Unsupported protocol: $this->{protocol}";
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveiSNSServer_EXIT;
    }

    foreach (keys %args)
    {
        if(/^-ip_address$/i)
        {
            if (!length($ip_address = $args{$_}))
            {
                $msg = "ISCSI::iscsiRemoveiSNSServer - invalid value for -ip_address: $args{$_}";
                $this->_set_err($msg);
                $status = $Host::FALSE;
                goto iscsiRemoveiSNSServer_EXIT;
            }
            next;
        }
    }

    # check for required parameters
    if (!defined($ip_address) or $ip_address =~ /^\s*$/)
    {
        $msg = "ISCSI::iscsiRemoveiSNSServer - Strict argument -ip_address not specified";
        $this->_set_err($msg);
        $status = $Host::FALSE;
        goto iscsiRemoveiSNSServer_EXIT;
    }

    # for all non 2k8 machines, use mpc to remove the isnsserver
    if ($this->{hostver} !~ /2008/ and $this->{hostver} !~ /windows server 7/i) {
        $cmd = "$this->{cmd_mpc} iscsi remove isnsserver $ip_address";
        
        # execute the command 
        ($status, @output) = $this->execute($cmd);

        if ($status != $Host::TRUE or !grep(/Instance deletion successful/i, @output)) {
            $msg = "ISCSI::iscsiRemoveiSNSServer - Command failed: $cmd";
            $this->_set_err($msg);
            goto iscsiRemoveiSNSServer_EXIT;
        }

    } else { 
        # build the command using wmic, there is a known issue with mpc iscsi remove isnsservers and win2k8 machines
        $cmd = "wmic /NAMESPACE:\\\\root\\wmi PATH MSiSCSIInitiator_iSNSServerClass WHERE iSNSServerAddress=\"$ip_address\" DELETE";
        
        # execute the command 
        ($status, @output) = $this->execute($cmd);
        
        if ($status != $Host::TRUE or !grep(/Instance deletion successful/i, @output)) {
            $msg = "ISCSI::iscsiRemoveiSNSServer - Command failed: $cmd";
            $this->_set_err($msg);
            goto iscsiRemoveiSNSServer_EXIT;
        }
    }
        
iscsiRemoveiSNSServer_EXIT:

    return($status, @output);

} # END iscsiRemoveiSNSServer

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

ISCSI.pm

=head1 Author

Lopamudra Sarangi

=head1 Description

ISCSI.pm is used for implementing ISCSI CLI for Windows. It has subroutines
to execute iscsi commands using mpc and parse results. Each of these
subroutines can be called by a WinClient Host object.

=head1 See Also

In order to create a Windows Host object the following modules might be used.

B<WinClient.pm>

B<WinClient::Host.pm>

=head1 Public Methods

=head2 B<iscsiShowAdapters()>

Used to display the details of iscsi adapters on the Windows host/initiator.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item Outputs

=item * -@output

Output on execution of "iscsi show adapters"

=item * -$status

Status of the Host after execution.

=item Examples

#
# Create a Windows host object<br>
#
use WinClient;

my $connection = &main::connect($CLIENT.$CLIENT_CONN);
my $hostObj = WinClient->new(-connection => $connection,
                             -protocol   => $PROTOCOL,
                             -multipath  => $MULTIPATH
                            );

#
# Displaying the list of ISCSI adapters
#
($status,@output) = $hostObj->iscsiShowAdapters();

=back

=head2 B<iscsiShow()>

Used to display the details of configured iscsi targetportals, targets
and active sessions on the Windows Host

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$object

Strict parameter: Can be either targets/targetportals/sessions

=item Outputs

=item * -@output

Output on execution of "iscsi show $object"

=item * -$status

Status after the command execution

=item Examples

#
# Displaying the list of configured ISCSI targetportals 
#
($status,@output) = $hostObj->iscsiShow(-object => "targetportals");

#
# Displaying all available targets
#
($status,@output) = $hostObj->iscsiShow(-object => "targets");

#
# Displaying all currently active sessions
#
($status,@output) = $hostObj->iscsiShow(-object => "sessions");

=back

=head2 B<iscsiShowConnections()>

Used to display the details of active iscsi connections on the Windows host.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$sessionid

Strict parameter: The session id for which the connection details have to be displayed

=item Outputs

=item * -@output

Output on execution of "iscsi show connections "

=item * -$status

Status after the command execution

=item Examples

#
# Displaying all connections for specified session id
#
($status,@output) = $hostObj->iscsiShowConnections(-sessionid => $sessions[0]);

=back

=head2 B<iscsiAddConnection()>

Used to add a new connection between an iscsi target and an Windows iscsi host/initiator
with the specified parameters.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$targetindex

Strict parameter: Index of the iscsi target

=item * -$sessionid

Strict parameter: Session ID to add a connection

=item * -$initiatorindex

Strict parameter: Index of the iscsi initiator

=item * -$mpio

Optional parameter: boolean, if specified mpio is enabled

=item * -$persistent

Optional parameter: boolean, if specified persistent is enabled

=item * -$headerdigest

Optional parameter: boolean, if specified headerdigest is enabled

=item * -$datadigest

Optional parameter: boolean, if specified datadigest is enabled

=item Outputs

=item * -@output

Output on execution of "iscsi add connection"

=item * -$status

Status after the command execution

=item Examples
 
#
# Adding a connection with mpio enabled
#
($status, @output) = $hostObj->iscsiAddConnection(-sessionid => $essionid,
                                                  -targetindex => $targetindex,
                                                  -initiatorindex => $initiatorindex,
                                                  -mpio => $Host::TRUE);

=back

=head2 B<iscsiRemoveConnection()>

Used to remove an existing connection between an iscsi target and a Windows iscsi host/initiator

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$connectionid

Strict parameter: The connection ID of the iscsi connection to be removed

=item Outputs

=item * -@output

Output on execution of "iscsi remove connection"

=item * -$status

Status after the command execution

=item Examples

#
# Removing a connection
#
($status, @output) = $hostObj->iscsiRemoveConnection(-connectionid => $connectionid);

=back

=head2 B<iscsiAddTargetPortal()>

Used to add a targetportal to the iscsi Windows host/initiator with the specified parameters

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$filer

Strict parameter: IP address of the filer to be added as targetportal

=item * -$mpio

Optional parameter: boolean, if specified mpio is enabled

=item * -$persistent

Optional parameter: boolean, if specified persistent is enabled

=item * -$headerdigest

Optional parameter: boolean, if specified headerdigest is enabled

=item * -$datadigest

Optional parameter: boolean, if specified datadigest is enabled

=item Outputs

=item * -@output

Output on execution of "iscsi add targetportal"

=item * -$status

Status after the command execution

=item Examples

#
# Adding a targetportal with mpio enabled
#
($status, @output) = $hostObj->iscsiAddTargetPortal(-filer => $filer,
                                                    -mpio => $Host::TRUE);

=back

=head2 B<iscsiRemoveTargetPortal()> 

Used to remove an existing targetportal from the Windows iscsi host/initiator.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$filer

Strict parameter: IP address of the filer to be removed from targetportals

=item Outputs

=item * -@output

Output on execution of "iscsi remove targetportal"

=item * -$status

Status after the command execution

=item Examples

#
# Removing a target portal
#
($status, @output) = $hostObj->iscsiRemoveTargetPortal(-filer => $filerip);

=back

=head2 B<iscsiLogin()>

Used to login to an iscsi target from a Windows iscsi host/initiator.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$targetname

Strict parameter: Name of the iscsi target to login to

=item * -$targetindex

Strict parameter: Index of the iscsi target to login to

=item * -$initiatorindex

Strict parameter: Index of the iscsi initiator

=item * -$mpio

Optional parameter: boolean, if specified mpio is enabled

=item * -$persistent

Optional parameter: boolean, if specified persistent is enabled

=item * -$headerdigest

Optional parameter: boolean, if specified headerdigest is enabled

=item * -$datadigest

Optional parameter: boolean, if specified datadigest is enabled

=item Outputs

=item * -@output

Output on execution of "iscsi login"

=item * -$status

Status after the command execution

=item Examples

#
# ISCSI Login with datadigest enabled
#
($status, @output) = $hostObj->iscsiLogin(-targetname => $targetname,
                                          -targetindex => $targetindex,
                                          -initiatorindex => initiatorindex
                                          -datadigest => $Host::TRUE);

=back

=head2 B<iscsiLogout()>

Used to logout of an active iscsi session between an iscsi target and a Windows iscsi host/initiator.

=over 4

=item Inputs

=item * -$this

Windows Host Object

=item * -$sessionid

Session ID of the session to logout from

=item Outputs

=item * -@output

Output on execution of "iscsi logout"

=item * -$status

Status after the command execution

=item Examples

#
# ISCSI Logout
#
($status, @output) = $hostObj->iscsiLogout(-sessionid => $sessions[0]);

=back

=cut

1;
