# $Id: //depot/prod/test/Rsierranevada/lib/WinClient/Util.pm#1 $
#
# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.
#
# This lib is a collection of routines used by the Windows Host
# libraries.  These are designed to be called from libraries and not
# from a thpl directly.
#
# TODO: Complete documentation pending.
#
## @summary
##  Wrapper for WinClient host functionality.
##
## @author  kenny speer
## @status  private
##

use strict;

package WinClient::Util;

# define the subroutines used in this package
sub execute;                # execute a command on the host
sub createRemoteFile;       # creates/writes a file on the host
sub readRemoteFile;         # reads a file on the host
sub deleteRemoteFile;       # removes a file on the host
sub copyRemoteFile;         # copy a file on the host;
sub readRegistryKey;        # read a registry key on the host
sub writeRegistryKey;       # write to a registry key on the host
sub _diskpartCmd;           # executes diskpart commands on the host

use lib();
use San::SanClient::ClientUtil;
use File::Temp;
use File::Basename;

# Define temp file post fix constant and value.
# This helps make the temp files used unique to this run
# and keeps multiple WinClient objects from stomping on
# each other when accessing temp files.
use constant TEMP_FILE_POSTFIX => time();

#
#
#
sub execute 
{
    my ($this, $cmd, $timeout, $allowTimeout) = @_;

    my $connection = $this->{connection};

    # default the timeout to the max timeout
    if (not defined($timeout) or $timeout =~ /^\s*$/) {
        $timeout = $Host::MAX_TIMEOUT;
    }

    # invoke the  execute function for windows located in ClientUtil.pm
    my ($status, @output) = &ClientUtil::_execute_windows($connection, 
                                                          $cmd, 
                                                          $timeout, 
                                                          $allowTimeout);

    return ($status, @output);
} # END execute

#
#
#
sub createRemoteFile 
{
    my ($this, $remFile, @filebuf) = @_;

    my ($status, @output, $line, $path);
    my $retry_count = 0;

    if (!defined($remFile))
    {
        &main::logcomment("Remote file name not specified.");
    }

createRemoteFile_START:

    # ensure the path exists
    $path = $remFile;
    $path =~ s/[^\\]*$//;
    $this->execute("mkdir $path", 60);

    # if user specified the contents of the file, then use it to create
    # the file, else create an empty file
    if (@filebuf)
    {
        foreach $line (@filebuf)
        {
            $line =~ s/\$/\\\$/g;
            $line =~ s/\"/\\\"/g;

            ($status, @output) = $this->execute("echo $line >> $remFile", 60);

            if ($status != $Host::TRUE)
            {
                &main::logcomment("Command failed: echo $line >> $remFile");
                # if we failed any of the echo commands, remove the file and attempt to create again
                $this->deleteRemoteFile($remFile);
                if ($retry_count > 4) {
                    # retry count was exhausted, set error and fail.
                    $this->_set_err("WinClient::Util::createRemoteFile() Error: failed to create " . 
                                    "remote file '$remFile'");
                    goto createRemoteFile_EXIT;
                } else {
                    # retry again
                    $retry_count++;
                    &main::logcomment("Will attempt to create the file from scratch. (retry count: $retry_count of 4)");
                    goto createRemoteFile_START;
                }

            }
        }
    }
    else
    {
        ($status, @output) = $this->execute("echo >> $remFile", 60);

        if ($status != $Host::TRUE)
        {
            &main::logcomment("Failed to create file: $remFile");
            goto createRemoteFile_EXIT;
        }
    }

createRemoteFile_EXIT:

    return ($status, @output);

} # END createRemoteFile 

#
#
#
sub readRemoteFile 
{
    my ($this, $remFile) = @_;

    my ($status, @output, $line);

    if (!defined($remFile))
    {
        &main::logcomment("Filename not specified.");
        $status = $Host::FALSE;
        push(@output, "Filename not specified.");
        goto readRemoteFile_EXIT;
    }

    ($status, @output) = $this->execute("dir $remFile", 120);

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Command failed: dir $remFile");
        goto readRemoteFile_EXIT;
    }

    ($status, @output) = $this->execute("type $remFile", 60);

readRemoteFile_EXIT:

    return ($status, @output);

} # END readRemoteFile

#
#
#
sub deleteRemoteFile 
{
    my ($this, $remFile) = @_;

    my ($status, @output);

    ($status, @output) = $this->execute("$Host::RM $remFile", 60);

    return ($status, @output);

} # END deleteRemoteFile

#
#
#
sub copyRemoteFile 
{
    my ($this, $srcFile, $dstFile) = @_;

    my ($status, @output, $path);

    if (!defined($srcFile))
    {
        &main::logcomment("Source file name not specified.");
    }

    # ensure the source file exists
    ($status, @output) = $this->execute("dir $srcFile", 120);

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Command failed: dir $srcFile");
        goto copyRemoteFile_EXIT;
    }

    # ensure the destination path exists
    $path = $dstFile;
    $path =~ s/[^\\]*$//;
    $this->execute("mkdir $path", 60);

    # copy the file
    ($status, @output) = $this->execute("$Host::COPY $srcFile $dstFile", 300);

    if ($status != $Host::TRUE)
    {
        &main::logcomment("Command failed: $Host::COPY $srcFile $dstFile");
        goto copyRemoteFile_EXIT;
    }

copyRemoteFile_EXIT:

    return ($status, @output);

} # END copyRemoteFile

################################################################################
# Subroutine Name: getRegistryPathEntries
# Objective:       Will get registry key's value/type of a instance of a path
#                  and populate the hash
# Details:         See POD documentation at the end of this file
################################################################################
sub getRegistryPathEntries {
    my ($this, %args) = @_;

    my ($path, $cmd, $status, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?path/) {
            if (not defined ($path = $args{$_})) {
                $this->_set_err("Invalid value for -path => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Check that required parameter '-devices' was provided
    if (not defined $path and $path ne "") {
        $this->_set_err("Missing required argument '-path'");
        return $ClientUtil::FALSE;
    }

    #------------------
    # Here is an output example of "reg query" cmd
    #------------------
    # C:\Documents and Settings\Administrator.SANINTEROP>reg query HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}

    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}
    #    Class    REG_SZ    Volume
    #    (Default)    REG_SZ    Storage volumes
    #    EnumPropPages32    REG_SZ    StorProp.Dll,VolumePropPageProvider
    #    SilentInstall    REG_SZ    1
    #    NoInstallClass    REG_SZ    1
    #    Icon    REG_SZ    -53
    #    Installer32    REG_SZ    syssetup.dll,VolumeClassInstaller
    #    UpperFilters    REG_MULTI_SZ    vxcache

    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0000
    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0001
    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0002
    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0003
    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0004
    #HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\{71A27CDD-812A-11D0-BEC7-08002BE2092F}\0005
    #------------------

    # Get the registry info for path $path
    $cmd = "reg query $path";
    ($status, @output) = $this->execute($cmd, 120);
    if ($status > 0 or (grep(/ERROR:/, @output))) {
        $this->_set_err("Failed to get resitry entry for path $path");
        return $ClientUtil::FALSE;
    }

    foreach my $dev_instance_path (@output) {
        # skip empty lines
        next if ($dev_instance_path =~ /^$/);
        # will populate hash for each instance
        my @path_arr = split (/\\/, $path);
        my @instance_arr = split (/\\/, $dev_instance_path);
        if ($instance_arr[0] eq $path_arr[0] and scalar(@instance_arr) > scalar(@path_arr)) {
            # will populate hash for each instance
            $this->getRegistryPathEntries(-path => $dev_instance_path);
            return $ClientUtil::FALSE if ($this->error);
        } elsif ($instance_arr[0] ne $path_arr[0]) {
            my @key_info_arr = split (/\s+/, $dev_instance_path);
            my $crossed = 0;
            my ($name, $type, $value);
            foreach my $line (@key_info_arr) {
                # Remove \n at the end if present
                chomp($line);
                if ($crossed == 1) {
                    $value .= $line . " ";
                } elsif ($line =~ /REG_/) {
                    $type = $line;
                    $crossed = 1;
                } else {
                    $name .= $line . " ";
                }
            }
            # Remove the last space
            chop($name);
            chop($value);
            # Populate the hash
            $this->{reg}->{$path}->{$name}->{type} = $type;
            $this->{reg}->{$path}->{$name}->{value} = $value;
        }
    }
}

################################################################################
# Subroutine Name: readRegistryKey
# Objective:       Will read the value of the registry key and return it
# Input:           Key Path, Key Name
# Details:         See POD documentation at the end of this file
################################################################################
sub readRegistryKey {
    my ($this, %args) = @_;

    my ($path, $name, $options);
    my ($cmd, $status, @output, $timeout);

    $timeout = 120;
    
    foreach (keys %args) {
        if (/-?path/) {
            if (not length($path = $args{$_})) {
                $this->_set_err("Invalid value for -path: $args{$_}");
                return "";
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $this->_set_err("Invalid value for -name: $args{$_}");
                return "";
            }
            next;
        }
    }

    # Check that all required parameter are provided
    if (not defined $name or $name eq "") {
        $this->_set_err("Missing required argument '-name'");
        return "";
    }

    if (not defined $path or $path eq "") {
        $this->_set_err("Missing required argument '-path'");
        return "";
    }

    $cmd = "reg query $path /v $name $options";
    ($status, @output) = $this->execute($cmd, $timeout);
    if ($status != $Host::TRUE or (grep(/^ERROR:/, @output))) {
        $this->_set_err("Failed to get value for registry key " .
                        "$path" . "\\" . "$name");
        return "";
    } else {
        my ($key_info) = (grep (/\s+$name\s+/, @output));
        my @key_info_arr = split (/\s+/, $key_info);
        my $crossed = 0;
        my ($reg_name, $reg_type, $reg_value);
        foreach my $line (@key_info_arr) {
            # Remove \n at the end if present
            chomp($line);
            if ($crossed == 1) {
                $reg_value .= $line . " ";
            } elsif ($line =~ /REG_/) {
                $reg_type = $line;
                $crossed = 1;
            } else {
                $reg_name .= $line . " ";
            }
        }
        # Remove the last space
        chop($reg_name);
        chop($reg_value);

        return $reg_value;
    }

} # END readRegistryKey

################################################################################
# Subroutine Name: writeRegistryKey
# Objective:       Will write/set the value of the registry key
# Input:           Key Path, Key Name, Key Value and Key Type
# Details:         See POD documentation at the end of this file
################################################################################
sub writeRegistryKey {
    my ($this, %args) = @_;

    my ($path, $name, $value, $type);
    my ($cmd, $status, @output, $options, $timeout);

    $timeout = 120;
    foreach (keys %args) {
        if (/-?path/) {
            if (not length($path = $args{$_})) {
                $this->_set_err("Invalid value for -path: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $this->_set_err("Invalid value for -options: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?value/) {
            if (not length($value = $args{$_})) {
                $this->_set_err("Invalid value for -value: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $this->_set_err("Invalid value for -type: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $this->_set_err("Invalid value for -name: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }
    
    # Check that all required parameter are provided
    if (not defined $value or $value eq "") {
        $this->_set_err("Missing required argument '-value'");
        return $ClientUtil::FALSE;
    }

    if (not defined $type or $type eq "") {
        $this->_set_err("Missing required argument '-type'");
        return $ClientUtil::FALSE;
    }

    if (not defined $name or $name eq "") {
        $this->_set_err("Missing required argument '-name'");
        return $ClientUtil::FALSE;
    }

    if (not defined $path or $path eq "") {
        $this->_set_err("Missing required argument '-path'");
        return $ClientUtil::FALSE;
    }

    $cmd = "reg add $path /v $name /t $type /d $value $options";
    ($status, @output) = $this->execute($cmd, $timeout);
    if ($status != $Host::TRUE or (grep(/^ERROR:/, @output))) {
        $this->_set_err("Failed to set value($value) for registry key " .
                        "$path" . "\\" . "$name");
        return $ClientUtil::FALSE; 
    } else {
        &main::logcomment("$cmd was successfull: ", @output); 
        return $ClientUtil::TRUE;
    }

} # END writeRegistryKey


#
#
#
sub _diskpartCmd
{
    my ($this, %args) = @_;

    my @cmds = @{$args{commands}};
    my $timeout = $args{timeout};
    my $lastDiskpartTime = time();

    my ($status, @output, $waittime, $cmd);

    # need a remote filename so use the first command
    my $remfile = "diskpart_input_";

    # edit the remfile for something we can use, remove spaces
    $remfile =~ s/\s+/_/g;
    $remfile = "$Host::TEMP\\$remfile";

    # Append the temp file postfix
    my $tmpname = File::Temp::tmpnam();
    $remfile .= File::Basename::basename($tmpname);

    # print comments if defined
    if (defined($args{comment}))
    {
        &main::logcomment($args{comment}, "$this->{host}->{cmd_diskpart} input: ", @cmds);
    }
    else
    {
        &main::logcomment("$this->{host}->{cmd_diskpart} input:", @cmds);
    }

    # create remote file for use with diskpart /s
    $this->deleteRemoteFile($remfile);
    $this->createRemoteFile($remfile, @cmds);

    # wait at least 15 seconds between diskpart commands for Windows
    $waittime = $lastDiskpartTime - time() + 15;
    sleep($waittime) if ($waittime > 0);

    # look at timeout
    if (! defined($timeout)) {
        $timeout  = $this->{timeout};
    }

    # set and execute the command
    $cmd = "$this->{host}->{cmd_diskpart} /s $remfile";
    ($status, @output) = $this->execute($cmd, $timeout);
    $lastDiskpartTime = time();

    if (grep(/specified partition is not valid/, @output)  or
        grep(/disk management services could not complete/, @output) or
        grep(/unable to create the specified partition/, @output) or
        grep(/Specified drive does not exist/, @output) or
        grep(/arguments you specified for this command are not valid/, @output) or
        $status != $Host::TRUE)
    {
        &main::logcomment("Disk::_diskpartCmd failed: ", @output);
        $status = $Host::FALSE;
        goto _diskpartCmd_EXIT;
    }

    &main::logcomment("$this->{host}->{cmd_diskpart} was successful: ", @output);

_diskpartCmd_EXIT:

    # remove the file we used
    $this->deleteRemoteFile($remfile);

    return ($status, @output);

} # END _diskpartCmd

1;
