##
## Copyright (c) 2008 NetApp, Inc.
## All rights reserved.
##
##  This library is used to control the user context of the rshd process.
##
##
## @summary
##  This library is used to control the user context of the Denicomp Systems 
##  rshd process. The processs can be run as the local machine administrator 
##  account or LocalSystem account. Enabling as the local system account also 
##  enables the interact with desktop feature.
##
## @Dependencies
## The administrator password is expected to be netapp
## This only works with the wrshdnt rsh daemon from Denicomp Systems, 
## which is available in the MinWin area of the x: drive. For example, 
## in RTP: \\rtpinf02\localtest\arch\win32\san\clientsetup\common\rshdnt_225.exe
## 
## @author  dascenj@netapp.com
## @status  public
##

package WinClient::Wrshdnt;

use strict;

use San::SanClient;
use San::SanClient::ClientUtil;
use Error;

###########################################################################
# Input: WinClient object / administrator or system 
# Output: returns TRUE or FALSE
# Tasks: * Assess the current context of the RSHD process
#        * Change the context to the requested context by restarting the service
############################################################################
sub wrshdnt_change_context {
    my ($this, $new_context) = @_;
    
    my @sc_commands;
    my $status;
    my @output;
    
    # initially clear errors
    $this->_clear_err();
 
    #Check for the proper args 
    if ($new_context !~ /system|administrator/i) {
        my $msg = "Error: Winclient::wrshdnt - The context \"$new_context\" " . 
                  "is invalid.You must supply administrator, or system";
        $this->_set_err($msg);
        return $ClientUtil::FALSE;
    }

    # Make sure that there is a need to do the work by checking the current and
    # requested context. 
    my $current_context = $this->wrshdnt_check_context;
    
    # Check the error on the object.
    if ( $this->error() ) {
        $this->sanlog(msg => "Error: WinClient::Wrshdnt - $this->{error}");
        return $ClientUtil::FALSE;
    }

    if ($current_context eq $new_context) {
        $this->sanlog(msg => "Requested $new_context  context. 
            The rsh daemon is already running as $current_context. There is nothing to do");
        return $ClientUtil::TRUE;
    }
    
    # Check to make sure that the value in $this->{domainName} really is a 
    # domain, # and not just a workgroup. If the machine is not a member of a
    # domain, the service will be configured to use the local administrator 
    # password. The script expects the local admin password to be netapp.
    # verifyDomain does not return a value. Checking the error on the object 
    # instead. If there is no error message, we succeeded. 
    my $groupingSolution = $this->verifyDomain();

    my $domainName;
    # Check the error on the object.
    if ( $this->error() ) {
        $this->sanlog(msg => "Warning: WinClient::Wrshdnt - $this->{error}");
        return $SanClient::FALSE;
    } 
    # Use the domain name if we were able to confirm that its really a domain.
    if ($groupingSolution =~ /domain/i) {  
        
        $domainName = $this->{domainName};
        # Have to manually grant the logon as service right to the administrator
        # account
        ($status, @output) = &ClientUtil::execute($this->{connection},
                'x:\bin\ffdata\Progra~1\Resour~1\ntrights.exe -u ' . $domainName . 
                '\administrator +r SeServiceLogonRight');
        if (($status != 0) or (! grep (/success/i, @output))) {
            my $msg = "Error: Winclient::wrshdnt - Could not grant the LogonAsService to " . 
                      $domainName . "administrator";
            $this->_set_err($msg);
            return $ClientUtil::FALSE;
        }
    } else {
        # Use the local administrator account
        $domainName = $this->{hostname};
    }   
    
    # Unmap the X drive so the mapping doesn't get "stuck" when the rsh 
    # service is restarted 
    ($status, @output) = &ClientUtil::execute($this->{connection}, "net use x: /delete /yes");
    if (($status != 0) or (! grep (/success/i, @output))) {
        my $msg = "Error: Winclient::wrshdnt - Failed to Unmap the x drive.";
        $this->_set_err($msg);
        return $ClientUtil::FALSE;
    }
    
    # The sc command controls services on 2k3 and 2k8. Here we are configuring
    # the "remote shell daemon" service to restart on failure after 30000
    # milliseconds. Enabling the restart on the service is harmless, and 
    # helpful...Lets just do it all the time.  
    push (@sc_commands, 'sc failure "remote shell daemon" reset= 10 actions= ' . 
                        'restart/30000/restart/30000/restart/30000');

    if ($new_context =~ /administrator/i) {
        # Changes the security context of the service to the 
        # domain based administrator account
        push (@sc_commands, 'sc config "remote shell daemon" obj= ' .
                            $domainName . '\administrator password= ' .
                            'netapp type= own');
    } elsif ($new_context =~ /system/i) {
        # Changes the security context of the service to the localsystem 
        # account and re-enables desktop interaction
        push (@sc_commands, 'sc config "remote shell daemon" type= own type= interact obj= LocalSystem');
    }
 
    # Execute the commands and check the status
    foreach my $command (@sc_commands) { 
        # There is no checking for the output of the command here. This is
        # omitted because after we kill the service, we expect it to restart.
        # After the restart, we check to see that the context is as we 
        # expected it to be. If its not, then the commands on the @sc_commmands 
        # array were not successful. 
        ($status, @output) = &ClientUtil::execute($this->{connection}, $command);
        if ($status != 0) {
             my $msg = "Error: Winclient::wrshdnt - failed to execute: \"$command\" The status is $status";
             $this->_set_err($msg);
             return $ClientUtil::FALSE;
        }
    }
    # All is well if we got here
    # kill the service exe so that the service will restart
    # Cannot use the sanclient kill function, the retries will get in the way.
    # For the same reasons as above, we are not checking errors here. 
    # This is because the next step is to verify that the rsh proc has 
    # changed context as we expect. We can use the result of that 
    # process to determine wether or not the process actually died properly. 
    # Furthermore, the proper return status could be in doubt, since we would 
    # actually be killing the process we hope will respond. 
    $status = $this->wrshdnt_kill();
  
    # Give rsh some time to restart
    sleep 15;

    # Verify the context was changed as expected. It is reasonable to infer that
    # if this check succeeds, the sc commands above worked properly. 
    my $changed_context = $this->wrshdnt_check_context();
    if ( $this->error() ) {
       $this->sanlog(msg => "Warning: WinClient::Wrshdnt - " . 
                                      $this->{error});
        return $ClientUtil::FALSE;
    }
    if ($changed_context != $new_context) {
        my $msg = "Error: Winclient::wrshdnt - Failed to modify RSH context";
        $this->_set_err($msg);
        return $ClientUtil::FALSE;
    }
    $this->sanlog(msg => "WinClient::Wrshdnt - The user context of process wrshdnt.exe was" . 
                                 " successfully changed from $current_context to $changed_context.");

    # We should have successfully restarted the service now. Verify. 
    # Minwin will remap the x: drive, thus verifing that the rshd is working.
    $this->_isMinWin;
    # Check the error on the object.
    if ( $this->error() ) {
        $this->sanlog(msg => "Warning: WinClient::Wrshdnt - $this->{error}");
        return $SanClient::FALSE;
    } 
    # If we got here, all is well. 
    $this->sanlog(msg => "WinClient::Wrshdnt - Completed Successfully");
    return $ClientUtil::TRUE;

} 

###########################################################################
# Name: wrshdnt_check_context
# Input: WinClient object
# Output: returns the current context of the RSHD process (system/administrator)
# Tasks: * Assess the current context of the RSHD process
############################################################################
sub wrshdnt_check_context {
    my ($this) = @_;

    my $whoami = "whoami";

    # Check the current context
    my ($status, @current_context) = &ClientUtil::execute($this->{connection}, $whoami);

    # Check the status of the executed command
    if ($status !=  0 ) {
        $this->sanlog(msg => "Error: Winclient::wrshdnt - There was an error running the rsh command");
        return "";
    }
    if (grep (/system/i, @current_context)) {
        return 'system';
    }elsif (grep (/administrator/i, @current_context)) {
        return 'administrator';
    }
    my $msg = "Error: Winclient::wrshdnt - The context of the service did not return as system or administrator";
    $this->_set_err($msg);

    return "";
}

###########################################################################
# Name: wrshdnt_kill
# Input: WinClient object
# Output: TRUE or FALSE
# Tasks: Terminate the wrshdnt.exe process without retrying. 
############################################################################
sub wrshdnt_kill {
   my ($this) = @_;

   # This is not using the ClientUtil::process_kill methods because the 
   # retries get in the way of the process restart. 
   # The windows process directives are meant to restart the wrshdnt.exe 
   # executable, the process will restart (as we expect) before the kill
   # method exits. This can cause the script to exhaust the retry count
   # and generate a failure.
   my $cmd = 'taskkill.exe /f /im wrshdnt.exe';
   my ($status, @output) = &ClientUtil::execute( $this->{connection}, $cmd);
   
   # We don't do any checking here in the output of the command, because
   # we will check later to see that the service has restarted by making
   # sure that the context has changed as we expect. This only happens if
   #  the service is sucessfully restarted. 
   if ($status !=  0 ) {
       $this->sanlog(msg => "Error Winclient::wrshdnt - There was a problem trying to run the taskkill command");
       return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

1;
