#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary    Parse workload definition files in XML 
##
## @author      Jd dascenj@netapp.com
##
## @pod here

=head1 NAME

system_test::workload::client::WorkloadParser - workload xml parser

=head1 SYNOPSIS

    This library is used by various thpls to get workload 
    details from xml defintion files. The workload to obtain 
    the details about is typically handed out by ConfigBroker

    See Topology.pm and ConfigBrokerTopo.pm

=head1 DESCRIPTION

Parse a workload definition file to return details on three 
types of interesting components. 

 Tool - Relates to the binary, or script that will be called to do 
        the required work.

 Workload - This is a logical grouping. Resources, as well as workitems are put together to form workloads. 
            A workload may contain one or more workitems, and resources. 

 Workitem - The most basic piece of work. This is the specific instationation of a given script or tool. 
            For example, there are many ways to call dt. One workitem may dictate that dt is called with 
            a documented blocksize, io pattern or other options.
 
More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

I am sure you can come up with something

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Probably. 

=head1 STATUS

  Awesome

=cut

package WorkloadParser;

BEGIN {
    use XML::Simple;
    use Data::Dumper;
    use Module::Common qw(validate_mixed);
    use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
    use Storable 'dclone';
    use Tharn qw($Log script_result fail_result);
}

=head1 METHODS

=head2 new 

    All other methods here are intended to be called as object methods

    The new() method should be used to instantiate the parser object.
    It expects a 'filename" argument that will tell the parser which 
    xml file to read in. 

     Return A workload parser object.

=head2 all_data

     Gets the whole of the xml file defined. It will exactly reflect the xml
     file that is parsed.  

      Return HASHREF

=head2 tools

   Get all the tools defined by the xml

=head3 Options 

    Takes only a 'details' boolean value. If boolean is false, 
    an array ref of only the names of the tools is returned. 

    When details => 1 is specified, all the known details are 
    also returned. 

    Return ARRAYREF

=head2 workloads

   Get all the workloads defined by the xml

=head3 Options 

    Takes only a 'details' boolean value. If boolean is false, 
    an array ref of only the names of the workloads is returned. 

    When details => 1 is specified, all the known details are 
    also returned. 

    Return ARRAYREF

=head2 workitems
 
   Get all the workitems defined by the xml

=head3 Options 

    Takes only a 'details' boolean value. If boolean is false, 
    an array ref of only the names of the workitems i returned. 

    When details => 1 is specified, all the known details are 
    also returned. 

    Return ARRAYREF

=head1 Workload Definition Example

   Example of the expected format of the xml from tool to workitem

<?xml version="1.0"?>
<tools>
  <cthon>
    <tool>cthon</tool>
    <defaults>
      <toolwillmount>1</toolwillmount>
      <executable>stress-scale/nfs/stress/cthon.thpl</executable>
      <NFS_VERSION>3</NFS_VERSION>
      <NFS_TRANSPORT>tcp</NFS_TRANSPORT>
    </defaults>
    <workitems>
      <cthon_nfs4_tcp>
        <NFS_VERSION>4</NFS_VERSION>
        <workloads>
          <nstp_wkld_nfs_cthon_1/>
        </workloads>
      </cthon_nfs4_tcp>
      <cthon_nfs3_tcp>
        <NFS_VERSION>3</NFS_VERSION>
        <workloads>
          <nstp_wkld_nfs_cthon_1/>
        </workloads>
      </cthon_nfs3_tcp>
    </workitems>
  </cthon>
</tools>

=cut

use strict;
use warnings;

$Data::Dumper::Sortkeys = 1;

sub new {
    my ($invocant, %opts) =
          validate_mixed(@_,
                  [{type => SCALAR},],
                  {filename  => {type => SCALAR,},},);

    my $class = ref($invocant) || $invocant;
    my $self = {};
    $Log->enter();
    bless($self, $class);

    my $filename = $opts{filename};
    my $wf = NATE::Inc::find_file($filename) ? NATE::Inc::find_file($filename) : $filename;
    if (not -e $wf) { 
        die fail_result ("Unable to find the specified workload definition file: $wf");
    }
    $self->{filename} = $filename;
    $self->{xml} = new XML::Simple;
    $self->{workloads} = $self->_parse(filename => $wf);

    $Log->exit();
    return $self;
}

sub all_data { 
    my ($self) = @_;
    return $self->{workloads};
}

sub _parse { 
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {filename    => {type => SCALAR,},},);

    my $filename = $opts{filename};
    die "could not find file $filename" if (not -e $filename);
    my $xml = $self->{xml};
    my $all = $xml->XMLin($filename);
    return $all;
}

sub tools { 
    $Log->enter();
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name => {type => SCALAR, default => 0, optional => 1},
                   details => {type => SCALAR, default => 0, optional => 1},},);

    my $all_data = dclone $self->all_data();

    $opts{details} = 1 if ($opts{name});
    
    my @tools; 
    if ($opts{details}) { 
        if ($opts{name}) { 
            push (@tools, $all_data->{$opts{name}});

        } else { 
            foreach my $key (keys %{$all_data}) { 
                next if (($opts{name}) and ($key !~ /^$opts{name}$/));

                $all_data->{$key}{name} = $key;   
                push (@tools, $all_data->{$key});
            }
        }
        $Log->debug("\ntools\n" . Dumper(\@tools));

    } else { 
        @tools = keys(%{$all_data}); 
        $Log->debug("\ntools\n" . Dumper(\@tools));
    }
   
    return \@tools;
}

sub workitems {
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name => {type => SCALAR, default => 0, optional => 1},
                   details => {type => SCALAR, default => 0, optional => 1},},);

    my @workitems;
    my $all_tools = $self->tools(details => 1);
    $opts{details} = 1 if ($opts{name});
   
    if ($opts{details}) { 
        foreach my $tool (@$all_tools) { 
            foreach my $wi (keys %{$tool->{workitems}}) { 
                next if (($opts{name}) and ($wi !~ /^$opts{name}$/));

                my $workitemhash = $tool->{workitems}{$wi};
                $workitemhash->{name} = $wi;
                $workitemhash->{tool} = $tool->{name};
                foreach my $tooldefkey (keys %{$tool->{defaults}}) { 
                    if (not defined($workitemhash->{$tooldefkey})) { 
                        $workitemhash->{$tooldefkey} = $tool->{defaults}{$tooldefkey};
                    }
                }

                push (@workitems, $tool->{workitems}{$wi}); 
             }
        }

    } else { 
        foreach my $tool (@$all_tools) { 
            foreach my $wikey (keys(%{$tool->{workitems}})) { 
                push (@workitems, $wikey);
            }
        }
    }

    $Log->debug("\nwork items: \n" .  Dumper(\@workitems));

    return \@workitems;
}

sub workloads {
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {name => {type => SCALAR, default => 0, optional => 1},
                   details => {type => SCALAR, default => 0, optional => 1},},);

    my @workloads;
    my $all_workitems = $self->workitems(details => 1);
    $opts{details} = 1 if ($opts{name});

    my $unique = {};
    foreach my $wi (@$all_workitems) { 
        foreach my $wl (keys(%{$wi->{workloads}})) { 
            # make sure we can identify the anonymous hash with a name
            # also supports options defined for the workload anywhere in teh 
            # workloads xml file. i.e. an option must only be defined once for 
            # any given workload
            $wi->{workloads}{$wl}{name} = $wl;
            if ($unique->{$wl}) { 
               foreach my $key (keys %{$wi->{workloads}{$wl}}) { 
                   $unique->{$wl}{$key} = $wi->{workloads}{$wl}{$key};
                }

            } else { 
               $unique->{$wl} = $wi->{workloads}->{$wl};
            }
        }
    }

    my @keys = keys(%{$unique});
    if ($opts{details}) { 
        foreach my $key (@keys) { 
           next if (($opts{name}) and ($key !~ /^$opts{name}$/));

           $unique->{$key}{workitems} = $self->_get_workitem_by_workload(workloadname => $key);
           push(@workloads, $unique->{$key});
        }

    } else { 
        @workloads = @keys; 
    }

    $Log->debug("\nworkloads: \n" .  Dumper(\@workloads));

    return \@workloads;
}

sub _get_workitem_by_workload {
    my ($self, %opts) =
          validate_mixed(@_,
                  [{type => OBJECT},],
                  {workloadname => {type => SCALAR, default => 0, optional => 1},},);

    my %workitems;
    my %workload_names;
    my $all_workitems = $self->workitems(details => 1);
    foreach my $wi (@{$all_workitems}) {
        my @wk_keys = keys (%{$wi->{workloads}});
        if (grep(/$opts{workloadname}/, @wk_keys)) { 
           $workitems{$wi->{name}} = 1;
        }
    }

    return [keys %workitems];

}
       
1;


