package Workspace;

=head1 package Workspace 

OO package to manage the Workspace queue data structure.  This data structure
was created to manage the queue of workspaces that the smoke system keeps around.
We currently keep workspaces for 2 reasons:

 1) a core file was collected during a bundle run, and the workspace is needed to quickly analyse that core.
 2) if the bundle run passed on all flavors, we keep the workspace for additional QA and smoke team testing.

Currently used by: tools/smoke/lastpass.thpl
Author: mgarrett


<after a few runs, paste the data structure here to make it easier to understand>

=cut

use strict;
use Data::Dumper;
use Storable;
use Net::SMTP;

=head1 User Methods

=over 8

=item new

Creates the object and loades the persistant data from $self->{filename}.
 ex. new(filename=> <filename>)

 p4dir      the p4 directory that contain a number of directories with names reflecting
            change numbers.  Here is what today's dir holds: 
                mglinux__prompt__>pwd
                /u/smoke/p4/ontapng/mainng

                mglinux__prompt__>ls
                854675@  854975@  mgok@

             This directory becomes the base directory for where wsget puts its files,
             and where the archive process tars its files up from.  (the 8 image.tgz's)
 
 bundledir   this variable is the root directory for all information related to this 
             branch.  See this tree dump of what goes in here:

                bundledir/
                |-- archive
                |   `-- 856006
                |       |-- final
                |       |   `-- bedrock
                |       |       `-- export
                |       |           |-- x86
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86.debug
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86.debug.sim
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86.sim
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86_64
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86_64.debug
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           |-- x86_64.debug.sim
                |       |           |   `-- tarball
                |       |           |       `-- image.tgz
                |       |           `-- x86_64.sim
                |       |               `-- tarball
                |       |                   `-- image.tgz
                |       `-- ontap
                |           `-- test
                |-- highbar_allpass -> /u/smoke,spin/p4/ontapng/mainng/856007
                |-- lastpass
                |   |-- buildinfo
                |   |   |-- 856006
                |   |   `-- 856007
                |   |-- lastpass.log
                |   |-- x86
                |   |-- x86.debug
                |   |-- x86.debug.sim
                |   |-- x86.sim
                |   |-- x86_64
                |   |-- x86_64.debug
                |   |-- x86_64.debug.sim
                |   `-- x86_64.sim
                `-- workspace.store

=cut


sub new {
	my $class = shift;
	my $self  = {
        p4dir=>undef, #see above
        bundledir=>undef, #see above
		data=>undef, #filled in by reading the Storable 'filename'
        testonly=>undef, #don't actually run the wsdel, just testing!
        p4view=>undef,
		@_,
	};
    my $newfile=0;

    $self->{filename}="$self->{bundledir}/workspace.store";
    $self->{link2allpass}="$self->{bundledir}/highbar_allpass";
	#Define, then read in the build information if the stored file already exists.
	if (-e $self->{filename} ) { $self->{data}=Storable::lock_retrieve $self->{filename}; }
	else {
        print "$self->{filename} wasn't found.  Starting a new one!\n";
		$self->{data}{max_fifo_size}=5;
        $newfile=1;
	}

	bless($self,$class);
    $self->_save2file() if $newfile;
	return $self;
}


=item copy_images(mode=>'full',src_buildroot=>'/u/ptcbuild/build.daemon/ontapng/mainng/ok',dst_archivedir=>'/u/smoke/bundle/ontapng/archive');

Make a tar copy of the important files in the src_buildroot.  Here is a sample of what we copy:

 if mode=>full we copy this:
 ==================================
 /u/mgarrett/archive/monkey
 |-- P4ENV
 |-- P4SYNC
 `-- final
     `-- bedrock
         `-- export
             |-- x86
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86.debug
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86.debug.sim
             |   |-- simiso
             |   |   `-- simsetup.iso
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86.sim
             |   |-- simiso
             |   |   `-- simsetup.iso
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86_64
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86_64.debug
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             |-- x86_64.debug.sim
             |   |-- simiso
             |   |   `-- simsetup.iso
             |   `-- tarball
             |       |-- image.md5
             |       `-- image.tgz
             `-- x86_64.sim
                 |-- simiso
                 |   `-- simsetup.iso
                 `-- tarball
                     |-- image.md5
                     `-- image.tgz
 
 
 if mode=>vsim_no_iso we copy this:
 ==================================
 /u/mgarrett/archive/next
 |-- P4ENV
 |-- P4SYNC
 `-- final
     `-- bedrock
        `-- export
            |-- x86
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86.debug
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86.debug.sim
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86.sim
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86_64
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86_64.debug
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            |-- x86_64.debug.sim
            |   `-- tarball
            |       |-- image.md5
            |       `-- image.tgz
            `-- x86_64.sim
                `-- tarball
                    |-- image.md5
                    `-- image.tgz
 

=cut 

sub copy_images {
	my $self = shift;
    my %parms = (
        mode => 'full', # could also be 'vsim_no_iso'
        @_
    );

    if(!exists($parms{src_buildroot}) or (!defined($parms{src_buildroot}))) { 
        print "Error: Workspace->copy_images() needs the param src_buildroot.  Doing nothing.\n";
    } elsif (!exists($parms{dst_archivedir}) or (!defined($parms{dst_archivedir}))) { 
        print "Error: Workspace->copy_images() needs the param dst_archivedir.  Doing nothing.\n";
    } else {
        #We have the 2 params we need.

        print "cd $parms{src_buildroot}\n";
        if (chdir $parms{src_buildroot}) {
            if (system("mkdir -p $parms{dst_archivedir}")==0) {
                print "Directory '$parms{dst_archivedir}' was successfully created.\n";

                my $cmd;
                if ($parms{mode} eq 'full') {
                    $cmd="tar cf - tools/build/.codeline final/bedrock/export/common/p4info/p4info final/bedrock/export/x86*/tarball/image.* final/bedrock/export/x86*/simiso/simsetup.iso final/bedrock/export/x86*/simiso/simsetup.iso P4* | tar xC $parms{dst_archivedir} -f -";
                } elsif ($parms{mode} eq 'vsim_no_iso') {
                    $cmd="tar cf - tools/build/.codeline final/bedrock/export/common/p4info/p4info final/bedrock/export/x86*/tarball/image.* P4* | tar xC $parms{dst_archivedir} -f -";
                } else {
                    print "Error: Workspace::copy_images() param 'mode' must be either 'full' or 'vsim_no_iso'\n";
                }

                print "$cmd\n";
                system $cmd;
            } else {
                print "Error: failed to create the directory '$parms{dst_archivedir}'.  This means I failed to archive!\n"; 
            }
        } else {
            print "Error: Failed to cd to '$parms{src_buildroot}'.\n";
        }
    }
}



=item update_allpass(changenum=>553344,force=>0)

When a build passes its highbar test on all flavors, this method does 5 things:
 1) coppies the old allpass workspace to an archive directory for future use (just the image.tgz and simiso files)
 2) removes the workspace from the old "passed_on_all_flavors" build (wsdel)
 3) Updates the entry in it's stored database $self->{data}{passed_on_all_flavors}.
 4) saves the changes to file.
 5) updates the symlink to point to the new changenum

If force=1, it will update the link even if it is the same build number.  Used in developement.

=cut

sub update_allpass {
	my $self = shift;
    my %parms = (
        changenum => undef,
        force => 0,
        @_
    );
    my @tolog;

	if (exists $self->{data}{passed_all_flavors}{changenum}) {
		print "Now that we have a new build that passed on all flavors, we can delete the old one. ($self->{data}{passed_all_flavors}{changenum})\n";
		if ($self->{data}{passed_all_flavors}{changenum} eq $parms{changenum}) {
            if ($parms{force}) {
                print "I see that $parms{changenum} was already saved as the passed_all_flavors.  Not deleteing the old one, but updating link since force was specified!\n";
            } else {
                print "I see that $parms{changenum} was already saved as the passed_all_flavors.  Not deleteing the old one, and not updating link!\n";
                push(@tolog,"I see that $parms{changenum} was already saved as the passed_all_flavors.  Not deleteing the old one, and not updating link!\n");
                return(@tolog);
            }
		} else {
            if ($self->is_saved($self->{data}{passed_all_flavors}{changenum})) {
                print "Wanted to delete the workspace, but it is listed in the 'saved' queue.\n";
            } elsif ($self->has_cores($self->{data}{passed_all_flavors}{changenum})) {
                print "Wanted to delete the workspace, but it is listed in the 'fifo' queue. (meaning at least one run had a core)\n";
            } else {
                push(@tolog,"Removing allpass build number $self->{data}{passed_all_flavors}{changenum} and replacing it with $parms{changenum}.\n");
                $self->wsdel(changenum=>$self->{data}{passed_all_flavors}{changenum});
            }
		}
	} 

	#update our database record for passed_all_flavors
	$self->{data}{passed_all_flavors}={
		date=>scalar(localtime),
		changenum=>$parms{changenum}
	};
	$self->_save2file;

    #archive the new allpass for safe keeping
    push(@tolog,"Archiving change number $parms{changenum} to $self->{bundledir}/archive/$self->{data}{passed_all_flavors}{changenum} for safe keeping.\n");
    
    print "\n\n\n\n";
    print "Job 3) - Archive and link this new all pass to highbar_allpass\n";
    print "#"x80 ."\n";
    print "src_buildroot=$self->{p4dir}/$parms{changenum}\n";
    print "dst_archivedir=$self->{bundledir}/archive/$self->{data}{passed_all_flavors}{changenum}\n";
    $self->copy_images(mode=>'full',src_buildroot=>"$self->{p4dir}/$parms{changenum}",dst_archivedir=>"$self->{bundledir}/archive/$self->{data}{passed_all_flavors}{changenum}");


    sub print_and_run {
        my $cmd=shift;
        print "cmd: $cmd\n";
        my @local_result=`$cmd`;
        print @local_result;
        return @local_result;
    }

    if (-l $self->{link2allpass}) {
        if (unlink $self->{link2allpass}) {
            print "Removed $self->{link2allpass} to make way for a new one. (" . scalar(localtime()) . ")\n";
        } else {
            #I probably should turn this into a return instead of a print, but I'm curious to see
            #if I will ever fail this remove, but succeed at the link that follows.
            push(@tolog,"Error: Couldn't remove $self->{link2allpass} to make way for a new one. ($!)\n");
        }
    }

	if (symlink "$self->{bundledir}/archive/$self->{data}{passed_all_flavors}{changenum}",$self->{link2allpass}) {
        print "symlink succesfully created from $self->{bundledir}/archive/$self->{data}{passed_all_flavors}{changenum} to $self->{link2allpass} (" . scalar(localtime()) . ")\n";
        print "\n\n\n\n";
        print "Job 4) - Update BammBamm with a new --test-high\n";
        print "#"x80 ."\n";

        my @high_result=print_and_run("/usr/software/rats/bin/bb_admin_state $self->{p4view} --test-high $parms{changenum} 2>&1");

        if (grep(/ERROR: 'test-high' must be <= 'test-low'/,@high_result)) {
            print "I see that bbadmin errored out saying that test_low needs to be at or above test_high.\nI'll set test_low too!\n";
            print_and_run("/usr/software/rats/bin/bb_admin_state $self->{p4view} --test-low $parms{changenum} 2>&1");
            print_and_run("/usr/software/rats/bin/bb_admin_state $self->{p4view} --test-high $parms{changenum} 2>&1");
        }

		push(@tolog,"link $self->{p4dir}/$parms{changenum} to $self->{link2allpass} is done.\n");
        return @tolog;
	} else {
		push(@tolog,"Error: link $self->{p4dir}/$parms{changenum} to $self->{link2allpass} failed! ($!)\n");
        return @tolog;
	}
}

=item show_allpass()

This function returns the change number of the last build that passed on all flavors.

=cut

sub show_allpass {
    my $self = shift;
    return $self->{data}{passed_all_flavors}{changenum};
}

=item wsdel(changenum=><changenum>)

Run a workspace delete on this change number.  Currently hardcoded to run as user smoke.

Looks something like this in real life:
 smkbc1b14{testbeds}: sudo su - smoke -c  "/usr/software/rats/bin/wsdel /u/smoke/p4/ontapng/mainng/795536"
 wsdel: Check p4 for opened files
 wsdel: Delete P4 Client: smoke:smoke_795536
 wsdel: Deleted client smoke:smoke_795536 succesfully
 wsdel: Delete Cloned Workspace...
 wsdel: Remove Symlink: /u/smoke/p4/ontapng/mainng/795536
 wsdel: completed in 39 seconds

=cut

sub wsdel {
	my $self = shift;

    my %parms = (
        changenum => undef,
        @_
    );

    if (!defined($parms{changenum})) {
        print "Error: sub routine Workspace->wsdel was called without defining 'changenum'.  Doing nothing!\n";
        return 0;
    }

    my $envUser;
    if (!exists($ENV{'USER'}) || !defined($ENV{'USER'}) || $ENV{'USER'} =~ /^\s*$/) {
	$envUser = `whoami`;
	chomp($envUser);
    } else {
	$envUser = $ENV{'USER'};
    }
    if ($envUser ne "smoke") {
        die("Error: wsdel must be run as the smoke user. Not as \"" . $ENV{USER} . "\" FYI: combined output of whoami and id is " . `/usr/bin/whoami` . " and " . `/usr/bin/id -a`);
        #Note:
        # Originally, I was running 'sudo' to the smoke user here.  But found that 
        # sudo isn't allowed to run in a cron.  
    }


	my $tempfile="$self->{bundledir}/.$parms{changenum}.deleting";

	if (-e "$self->{p4dir}/$parms{changenum}") {
		if (-e $tempfile) {
			print "Warning: I was asked to delete the workspace for build $parms{changenum}, but I see someone is already deleting it. (found file '$tempfile')\n";
		} else {
			system "touch $tempfile";
			my $cmd = "/usr/software/rats/bin/wsdel $self->{p4dir}/$parms{changenum} 2>&1"; 
			print "$cmd\n";
            if ($self->{testonly}) {
                print "Skipping this command since 'testonly' is turned on.\n";
            } else {
                my @output=`$cmd`;
                print @output;

                if ((not grep(/wsdel\: completed in/,@output)) or (grep(/Client deletion failed/,@output))) {
                    print "wsdel failed!  Sending a mail to MikeG so he will look at it...\n";
                    my $smtp = Net::SMTP->new("mail.mgmt.smoke");
                    $smtp->mail('smoke@netapp.com');
                    $smtp->to('mgarrett@netapp.com');
                    $smtp->data();
                    $smtp->datasend("To: mgarrett\@netapp.com\n");
                    $smtp->datasend("Subject: Lastpass: wsdel failed for build " . $parms{changenum} . "\n");
                    $smtp->datasend("\n");
                    $smtp->datasend("http://web.smoke.spin.eng.netapp.com/natejobs_gx.cgi?testbed=/u/smoke/presub&logdir=$ENV{JOBID}\n");
                    $smtp->datasend("\n");

                    $smtp->datasend(@output);
                    $smtp->datasend("hostname=" . `hostname`);
                    $smtp->dataend();
                    $smtp->quit;
                }

            }

			system "rm -f $tempfile";
		}
	}
    elsif (-l "$self->{p4dir}/$parms{changenum}") {
        print "\tWarning: there was a link, but it pointed to a dead directory.  Deleting link...\n";
        system "rm -f $self->{p4dir}/$parms{changenum}";
    } else {
		print "Warning: I was asked to delete this workspace '$self->{p4dir}/$parms{changenum}', but it doesn't exist.\n";
	}
}

=item delete_unmanaged_workspaces()

Sometimes bundles fail, and leave workspaces that aren't managed.  This cleans them up.

=cut

sub delete_unmanaged_workspaces {
	my $self = shift;

    #Take any params passed in, and treat them like build numbers that must be saved!
    my @keep_these=@_;
    my %keep_these;
    for my $one (@keep_these) {
        $keep_these{$one}=undef;
    }

	chdir "$self->{p4dir}";
	my @dirs=<*>;
	@dirs=grep /^\d+$/, @dirs; #if the whole dir is a number, I assume it is a change number, and thus a workspace

	my $counter=0;

    #Calculate 12 hours ago
    my $twelve_hours=12*60*60; 
    my $now=time();
    my $twelve_hours_ago=$now-$twelve_hours;

    foreach my $onedir (@dirs) {
        if (exists $keep_these{$onedir}) {
            print "Skipping this one ($onedir) since it was on your list to skip.\n";
        }
        elsif ($self->_unmanaged($onedir)) {
            my $dirdate=(stat($onedir))[10];
            if ($dirdate>$twelve_hours_ago) {
                print "keeping '$onedir' because it isn't 12 hours old yet\n";
                my @lpruns=`grep $onedir $self->{bundledir}/lastpass/lastpass.log`;
                for (@lpruns) { print "\t$_"; }
            } else {
                print "removing '$onedir' because it is 12 hours old and not in a mangaged list.\n";
                my @lpruns=`grep $onedir $self->{bundledir}/lastpass/lastpass.log`;
                for (@lpruns) { print "\t$_"; }

                $self->wsdel(changenum=>$onedir);
                $counter++;
            }
        } else {
            print "keeping ($onedir) since it is in my managed list.\n";
        }
    }
    print "I cleaned up $counter workspaces.\n";
}


sub delete_old_managed_workspaces {
	my $self = shift;

	chdir "$self->{p4dir}";
	my @dirs=<*>;
	@dirs=grep /^\d+$/, @dirs; #if the whole dir is a number, I assume it is a change number, and thus a workspace

	my $counter=0;

    #Calculate 30 days ago
    my $now=time();
    my $thirty_days=30*24*60*60;
    my $thirty_days_ago=$now-$thirty_days;

    foreach my $onedir (@dirs) {
        my $dirdate=(stat($onedir))[10];
        unless ($self->_unmanaged($onedir)) { 
            if ($dirdate>$thirty_days_ago) {
                print "keeping ($onedir) since it is in my managed list, and it is not yet 30 days old.\n";
            } else {
                print "deleting ($onedir) even though it is on the managed list, because it is 30 days old.\n";
                $counter++;

                #delete the workspace
                $self->wsdel(changenum=>$onedir);

                #remove it from our managed list
                if (exists($self->{data}{saved}{$onedir})) {
                    delete $self->{data}{saved}{$onedir};
                }

                if (exists($self->{data}{fifo}{$onedir})) {
                    delete $self->{data}{fifo}{$onedir};
                }
            }
        }
    }
    print "\nI cleaned up $counter workspaces.\n";

    #if we deleted any, save the file.
    $self->_save2file if $counter;

}



#convinience routine to look through the data structure and find if a change is managed, or in this case not.
sub _unmanaged {
	my $self = shift;
	my $changenum= shift;

	if ((exists $self->{data}{saved}{$changenum})
			or
		(exists $self->{data}{fifo}{$changenum})
			or
		($self->{data}{passed_all_flavors}{changenum}==$changenum)) {
		return 0;
	} else {
		return 1; #it is unmanaged
	}
}

=item dump() 

dump $self (mostly for debug)

=cut

sub dump {
	my $self = shift;
	print "Contents of object Workspace (note: only the {data} key is persistant):\n";
	print Dumper $self;
}

=item delete_all()

Used when a branch's life is ended, and there is no more need to keep any of the workspaces.

=cut

sub delete_all {
	my $self = shift;
    print "Deleting all items from the fifo queue...\n";
    for my $build ( keys %{ $self->{data}{fifo} }) {
        print "#"x10 . "deleting build=$build" . "#"x10 .  "\n";
        $self->wsdel(changenum=>$build);
    }

    print "Finally, removing the last allpass that was saved...\n";
    print "#"x10 . "build=$self->{data}{passed_all_flavors}{changenum}" . "#"x10 . "\n";
    $self->wsdel(changenum=>$self->{data}{passed_all_flavors}{changenum});
}

#Find the size of the fifo array
sub _fifo_size {
	my $self = shift;
	if (exists $self->{data}{index}) {
		return scalar(@{$self->{data}{index}});
	} else {
		return 0;
	}
}

#Find the size of the saved hash
sub _saved_size {
	my $self = shift;
	if (exists $self->{data}{fifo}) {
		return scalar( keys %{$self->{data}{saved}});
	} else {
		return 0;
	}
}

=item add2fifo(changenum=> <num>, runid=> <jobname>)

Adds a new changenum record to the fifo queue and pushes the oldest one out. (if there are slots available)
Then it saves the changes to file.
 ex. $obj->add2fifo(changenum=>'793113',runid=>'vsim-mainng64-converged-nondebug-smoke.1211320637');

=cut

sub add2fifo {
	my $self = shift;
	my %parms = (
		changenum=>undef,
		runid=>undef,
		@_
	);

	if (my $ref=$self->_is_duplicate($parms{changenum})) {
		print "This build was already saved because it had a core during a smoke run.  Duplicate!\n";
		push @{$ref->{runs_with_cores}}, $parms{runid}; #just add to runs_with_cores
		$self->_save2file;
	} else { #This changenum hasn't been added before
		if ($self->_saved_size >= $self->{data}{max_fifo_size}) { #if all of the slots are used up in 'saved'
			print "User saved entries have used all of our slots.  None left for the fifo queue.\n";
		} else {
			if (($self->_fifo_size + $self->_saved_size) >= $self->{data}{max_fifo_size}) { #the fifo queue is at it's max
				$self->_remove_first_in; #make some room
			} 

			#add the changenum's info to the fifo hash
			$self->{data}{fifo}{$parms{changenum}} = {
				date_saved => scalar(localtime),
				runs_with_cores => [$parms{runid}]
			};

			#add the changenum's info to the index array
			push @{ $self->{data}{index} }, $parms{changenum};

			$self->_save2file;
		}
	}
}

=item move2saved(changenum=> <changenum>, note=> <user specified notes field> )

Takes an entry from the $self->{data}{fifo} queue, and moves it to the $self->{data}{saved} hash
and saves the changes to file.
 ex. $obj->move2saved(changenum=>'796192', note=>"Don't delete until Semke says its ok.  (mgarrett)");

=cut

sub move2saved {
	my $self = shift;
	my %parms=(
		changenum=>undef,
		note=>undef,
		@_
	);

	my $ref=undef;
	if (exists $self->{data}{fifo}{$parms{changenum}}) {
		$ref=$self->{data}{fifo}{$parms{changenum}};
		$ref->{note}=$parms{note};
		$self->{data}{saved}{$parms{changenum}}=$ref;
		delete $self->{data}{fifo}{$parms{changenum}};
	}

	@{ $self->{data}{index} } = grep !/^$parms{changenum}$/,  @{ $self->{data}{index} };
	$self->_save2file;
}

=item unsave(<changenum>)

Removes a previously saved entry from the saved hash, and saves the changes to file.

=cut

sub unsave {
	my $self = shift;
	my $changenum=shift;
	if (exists $self->{data}{saved}{$changenum}) { #has it already been saved?
		delete $self->{data}{saved}{$changenum};
		$self->_save2file;
	}
}

=item is_saved(<changenum>)

Returns true if the changenum is listed in the saved hash.

=cut

sub is_saved {
	my $self = shift;
	my $changenum=shift;

	if (exists $self->{data}{saved}{$changenum} ) { #has it already been saved?
		return 1;
	} else {
		return 0;
	}
}

=item has_cores(<changenum>)

Returns true if the changenum is listed in the fifo queue.

=cut


sub has_cores {
	my $self = shift;
	my $changenum=shift;

	if (exists $self->{data}{fifo}{$changenum}) { #is it already in the fifo queue?
		return 1;
	}

	return 0;
}

#answers the question, is this changenum already in either the fifo queue, or the saved hash?
#ex. $self->_is_duplicate(<changenum>)
#
sub _is_duplicate {
	my $self = shift;
	my $changenum=shift;

	if (exists $self->{data}{saved}{$changenum}) { #has it already been saved?
		return $self->{data}{saved}{$changenum};
	}

	if (exists $self->{data}{fifo}{$changenum}) { #is it already in the fifo queue?
		return $self->{data}{fifo}{$changenum};
	}

	return 0;
}

#Writes the storable part of the object (ie $self->{data}) to a file.
sub _save2file {
	my $self = shift;
	my $retval=Storable::lock_nstore $self->{data}, $self->{filename};
}

#removes the first entry in the fifo queue to make room for a new one
sub _remove_first_in { 
	my $self = shift;
	my $changenum=shift @{ $self->{data}{index} };
	delete $self->{data}{fifo}{$changenum};

	print "To make more room in the FIFO queue, we need to delete the oldest change number in the queue ($changenum).\n";
    if ($self->{testonly}) {
        print "skip the wsdel since 'testonly' is set.\n";
    } else {
        $self->wsdel(changenum=>$changenum);
    }
}

1;
