# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Wreck.pm#1 $

## @summary Routines for using 'wreck' corruption command
##
## @description
##    The 'wreck' command corrupts filesystems for testing purposes.
##    This library helps you use it.
##
## @synopsis
##    use Wreck;
##    wreck($conn, $volname, type => $type, inject_script => $inject_script,
##          timeout => $timeout);
##
## @examples
##    use Wreck;
##    my $h = host($FILER);
##    my $c = connect("#console");
##    my $ret = wreck($c, 'aggr', 'myaggr', 'inode_type inum=64 type=1');
##    if (defined($ret)) {
##        tstlog(-fail, "Failed to corrupt myaggr rootdir: $ret");
##    }
##
## @usage
##
##    wreck($conn, $name, type => $type, inject_script => $inject_script,
##	    timeout => $timeout, pseudo_bad => 1);
##
##    $conn           - connection
##
##    $type           - 'aggr', 'flex', or 'trad'
##
##    $volname        - name of target volume/aggr
##
##    $inject_script  - text of inject script to apply
##
##    $timeout        - seconds for 'wreck' command (optional, default 600)
##
##    $pseudo_bad     - write modified block as pseudo_bad (optional, default 0)
##
##    $unverified     - write modified block as unverified (optional, default 0)
##
##    $iron    	  	  - 1 implies start wafliron after injecting error.
##    
##    Note: The pseudo_bad option should be used only with initialized parity disks
##
##    The inject script is written to /etc/inject-volname.
##    wreck [-a] $volname /etc/inject-volname is done.
##    Then the inject file is removed again.
##
##    If the wreck command says 'Finished' an undefined value is returned.
##    Otherwise, an error string is returned.
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status public

package Wreck;

use Carp;
use strict;

use Exporter qw(import);
BEGIN {
    our @EXPORT = ('wreck');
};

use Tharn;
use CLI::FileIO qw(wrfile);
use SayWrap;
use Utils qw(make_conn_and_complain);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR);
use Aggregate qw(aggr_status);
use Volume qw(vol_status);
use Wafliron qw(wafliron_start);

# ----------------------------------------------------------------------

sub wreck
{
	# Validate arguments
	my $conn = make_conn_and_complain(shift);
	my ($name, %opts) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		  type => { type => SCALAR, optional => 0,
			    regexp => qr/aggr|flex|trad/i  },
		  inject_script => { type=> SCALAR, optional => 0 },
		  timeout => { type => SCALAR, optional => 1,
			       default => 600 },
		  pseudo_bad => { type => SCALAR, optional => 1,
		               default => 0},
                  unverified => { type => SCALAR, optional => 1,
		               default => 0},
		  iron => { type => SCALAR, optional => 1,
						default => 0}
		}
	);

	return eval {
		my $file = "/etc/inject-$name";
		my $cmd;
		# If we fail to write the file, bail out
		wrfile($conn, $file, $opts{inject_script}) 
			or croak("Failed to write wreck injection script: $@");

		if ($opts{type} =~ /aggr/) {
			$cmd = "wreck -A $name $file";
		} elsif ($opts{type} =~ /flex|trad/i) {
			$cmd = "wreck $name $file";
		} else {
			# We shouldn't ever reach here. the
			# type is being validated at the top
			croak("Illegal wreck type: $opts{type}");
		}
                
                # checking wheter both unverified and pseudo_bad are 1
                # as per wreck syntax at a time only one of the switch 
                # (pseudo_bad|unverified) is possible
                if ($opts{pseudo_bad} =~ /1/ && $opts{unverified}=~ /1/)
                {
                    croak("Illegal use of switch pseudo_bad and unverified together");
                }
                
                
		if ($opts{pseudo_bad} =~ /1/ ||  $opts{inject_script} =~ /pseudo_bad/ )
                {
                    
                    # Checking if parity disk is initialized.
                    # pseudo_bad option should not be used with 
                    # uninitialized disks

                    my $status;
                    $Log->comment("Inside parity checking");
                    if($opts{type} =~ /flex/i) {
                        # Status for flevol
                        $status = vol_status($conn, $name)->status;
                    } else {
                        # Status for aggregate or tradvol
                        $status = aggr_status($conn, $name)->status;
                    }

                    if( $status =~ /parity uninit'd/i ) {
                        croak("pseudo_bad option should be exercised only "
                              . "on containers which have the parity disk "
                              . "initialized");
                    }
                    if ($opts{pseudo_bad} =~ /1/)
                    {
                        $cmd = $cmd . " -pseudo_bad";
                    }
                }
                
                if ($opts{unverified} =~ /1/ || $opts{inject_script} =~ /unverified/)
                {
                    # Checking if parity disk is initialized.
                    # pseudo_bad option should not be used with 
                    # uninitialized disks

                    my $status;
                    $Log->comment("Inside parity checking");
                    if($opts{type} =~ /flex/i) {
                        # Status for flevol
                        $status = vol_status($conn, $name)->status;
                    } else {
                        # Status for aggregate or tradvol
                        $status = aggr_status($conn, $name)->status;
                    }

                    if( $status =~ /parity uninit'd/i ) {
                        croak("pseudo_bad option should be exercised only "
                              . "on containers which have the parity disk "
                              . "initialized");
                    }
                    if ($opts{unverified} =~ /1/)
                    {
                        $cmd = $cmd . " -unverified";
                    }
                    
                }
		# '-q' option reduces redundant output from filer
		my $output = sayEOF($conn, "priv set -q test; $cmd",
				    TIMEOUT => $opts{timeout});
		sayEOF($conn, "priv set -q test; rm $file");

		if ($output !~ /Finished/s || $output !~/Wrote vbn .*/s ) {
			croak("Wreck failed to corrupt filesystem: $output");
		}

		# wrack users typically call iron to fix the corruption. therefore, 
		# disable the panic on corruption mode since that's what wrack users 
		# probably want
		sayEOF($conn, "priv set -q test; setflag wafl_iron_panic_on_corruption 0",
				    TIMEOUT => $opts{timeout});

		if ($opts{iron} =~ /1/) {

			my ($found, $error) = wafliron_start(
                                            $conn,
                                            $name,
                                            wait => 0,
                                            scan_speed => 10);
		}
		return 1;
	};
}


1;
