# $Id: //depot/prod/test/Rsierranevada/lib/Xen/Utils.pm#1 $
##
## @summary Functions for working with Xen simulators
## @description
## Functions that allow scripts to create, manipulate, boot, halt and destroy Xen simulator instances.
## of common resources.
## @usage
##  use Xen::Utils;
## @author ewing@netapp.com
## @status in-progress

package Xen::Utils;

use strict;
use Tharn;
use File::Copy;
use SayWrap qw(sayEOF);

our @ISA = qw(Exporter);
our @EXPORT = qw(
    &boot_xen_domain
    &configure_xen_domain
    &get_active_xen_domains
    &destroy_xen_domain
    &fix_xenargs_gdbserver
    &halt_xen_domain
    );

sub boot_xen_domain
{
    my %opt = (
        connection => undef(),
        xen_directory => undef(),
        xen_user => 'root',
        xen_pass => "\r",
        xen_memory_size => 850,
        xen_build_root => undef(),
        xen_boot_timeout => 300,
        @_
    );

    if (!defined($opt{xen_build_root})) {
        if (!defined($ENV{ROOT})) {
            die(config_result("configure_xen_domain: xen_build_root must be supplied explicitly or in your environment"));
        } else {
            logcomment("xen_build_root is not defined entering boot_xen_domain, defaulting to environment: $ENV{ROOT}");
            $opt{xen_build_root} = $ENV{ROOT};
        }
    }

    if (!defined($opt{connection}) || !defined($opt{xen_directory})) {
        die(fail_result("Connection and xen directory are required to boot a xen simulator."));
    }

    say($opt{connection}, "cd $opt{xen_directory} && ROOT=$opt{xen_build_root} ./xenargs -m $opt{xen_memory_size}", '-wait', "ogin:", '-noprompt', '-timeout', $opt{xen_boot_timeout});
    say($opt{connection}, $opt{xen_user}, '-wait', 'Password:', '-noprompt');
    say($opt{connection}, $opt{xen_pass}, '-wait', "$opt{connection}->{hostp}->{hostname}#", '-nocr', '-noprompt');
    logcomment("Xen domain running.");
}

sub fix_xenargs_gdbserver
{
    my %opt = (
        xen_directory => undef(),
        @_
    );

    my $nohup = `which nohup`;
    chomp($nohup);

    if (! -e $nohup) {
        die(fatal_result("Cannot find a nohup binary."));
    }

    logcomment("Found nohup: \"$nohup\"");

    open(XENARGS, "+<$opt{xen_directory}/xenargs") or die($!);
    open(OUT, ">$opt{xen_directory}/xenargs.new") or die($!);

    while (<XENARGS>) {
        if ($_ =~ m/(\s+)(\$ROOT\/tools\/xen\/bin\/xengdbserver)$/) {
            print OUT "$1 echo \"Running xengdbserver: $nohup $2\"\n";
            print OUT "$1$nohup $2\n";
        } else {
            print OUT $_;
        }
    }

    close(XENARGS);
    close(OUT);

    move("$opt{xen_directory}/xenargs", "$opt{xen_directory}/xenargs.org") or die($!);
    move("$opt{xen_directory}/xenargs.new", "$opt{xen_directory}/xenargs") or die ($!);

    return 1;
}

sub halt_xen_domain
{
    my %opt = (
        connection => undef(),
        @_,
    );

    say($opt{connection}, "halt", '-wait', 'Rebooting...', '-noprompt');
}

sub get_active_xen_domains
{
    my %opt = (
        connection => undef(),
        xm_binary => '/usr/bin/xm',
        @_
    );
    
    my %xen_domains;

    if (!defined($opt{connection})) {
        die(fail_result("Connection required to get active xen domains."));
    }

    my $out = sayEOF($opt{connection}, "$opt{xm_binary} list");
    
    my @lines = split(/\n/, $out);
    
    shift(@lines);

    foreach my $l (@lines) {

        if ($l =~ m/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)?/) {
            $xen_domains{$1}->{id} = $2;
            $xen_domains{$1}->{mem} = $3;
            $xen_domains{$1}->{cpu} = $4;
            $xen_domains{$1}->{state} = $5;
            $xen_domains{$1}->{time} = $6;
            $xen_domains{$1}->{console} = $7;
        } 
    }

    return (wantarray() ? keys(%xen_domains) : \%xen_domains);
}

sub destroy_xen_domain
{
    my %opt = (
        connection => undef(),
        xm_binary => '/usr/bin/xm',
        xen_host_name => undef(),
        @_
    );
    
    if (!defined($opt{connection}) || !defined($opt{xen_host_name})) {
        die(fail_result("Connection and host name required to destroy a xen domain."));
    }

    my $out = sayEOF($opt{connection}, "$opt{xm_binary} destroy $opt{xen_host_name}",  {timeout => 480});
}

sub configure_xen_domain 
{
    my %opt = (
        xen_directory => undef(),
        xen_build_root => undef(),
        xen_config_script => "tools/xen/bin/tricky_configure",
        xen_rootfs => undef(),
        no_disks => 4,
        disks_type => 9,
        host_name => undef(),
        host_mac => undef(),
        nblade_mac => undef(),
        dblade_mac => undef(),
        @_
    );

    logcomment("Starting configuration of Xen host $opt{host_name} in $opt{xen_directory}.");

    if (!defined($opt{xen_build_root})) {
        if (!defined($ENV{ROOT})) {
            die(config_result("configure_xen_domain: xen_build_root must be supplied explicitly or in your environment"));
        } else {
            logcomment("xen_build_root is not defined entering configure_xen_domain, defaulting to environment: $ENV{ROOT}");
            $opt{xen_build_root} = $ENV{ROOT};
        }
    }

    eval {
	my $config = $opt{xen_build_root}."/".$opt{xen_config_script};
	if (! -e $config) {
            die(fatal_result("$config does not exist"));
	}
	logcomment("$config file exists");
	
        require "$opt{xen_build_root}/$opt{xen_config_script}";

        my $ip = gethostbyname($opt{host_name}) or die(fatal_result("configure_xen_domain: Cannot lookup host $opt{host_name}."));

        our %tricky_config = (
            xen_host_name => {
                value => $opt{host_name},
            },
            xen_host_ip => {
                value => inet_ntoa($ip),
            },
            xen_no_disks => {
                value => $opt{no_disks},
            },
            xen_disks_type => {
                value => $opt{disks_type},
            },
            xen_rootfs=> {
                value => undef(),
            },
            xen_mac0_addr => {
                value => $opt{host_mac},
                interface => 'xbr-sim',
            },
            xen_mac1_addr => {
                value => $opt{nblade_mac},
                interface => 'xbr-data',
            },
            xen_mac2_addr => {
                value => $opt{dblade_mac},
                interface => 'xbr-clus',
            },
            xen_root => { 
                value => $opt{xen_directory}
            },
            verbosity => { 
                value => 2 
            },
            xen_build_root => { 
                value => $opt{xen_build_root}
            },
            xen_mgmt_interfaces => {
                value => 1
            },
            xen_data_interfaces => {
                value => 1
            },
            xen_cluster_interfaces => {
                value => 1
            }
        );

        xen_setup::uninstall_xen_directory(\%tricky_config, 0);
        xen_setup::create_xen_home(\%tricky_config);
    };

    if ($@) { die($@); }

    logcomment("Finished configuration of Xen host $opt{host_name}.");
}

1;

