# 
#  $Id: //depot/prod/test/Rsierranevada/lib/Zapi.pm#2 $
# 
#  Copyright (c) 2002 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary Support routines for Zapi tests
##
## @description
##  This file contains routines that
##  let you exercise the XML/HTTP API of the filer.
##
## @author tjt@netapp.com
## @burt 354475
## @change nsesha@netapp.com    Added new function parse_raw_xml, required for ZapiCli.pm
## @burt 385056
## @change astraub@netapp.com   Add option to use LWP::UserAgent for http connections
## @burt 399920
## @change isshwar@netapp.com   Added accessor methods to specify the server type which decides DTD
## @burt 423018
## @change nonga@netapp.com    Added APIs set_zapi_name, set_zapi_parm and 
##                             invoke_zapi to allow specification of complex 
##                             output
## @change davidlow@netapp.com Added handling of UTF8 charactors in ZAPI input.  This
##                              change is needed to execute this type of ZAPIs on a Linux
##                              client using HTTP or HTTPS.

package     Zapi;

use strict;
use misc;
use Socket;
use LWP::UserAgent;
use XML::Parser;
use NaElement;
use TestEnv::MinWin qw(:all);
use Encode;
use vars ('@ISA', '@EXPORT');
require     Exporter;
@ISA        = qw(Exporter);
@EXPORT     = qw(invoke);
@EXPORT     = qw(invoke_elem);
@EXPORT     = qw(set_zapi_name);
@EXPORT     = qw(set_zapi_param);
@EXPORT     = qw(invoke_zapi);
use log_zapi;

# Globals here
$::ZAPI_xmlns = "http://www.netapp.com/filer/admin";

# DTD file names for various server types
$::ZAPI_dtd = "file:/etc/netapp_filer.dtd";
$::ZAPI_nc_dtd = "file:/etc/netapp.dtd";
$::ZAPI_dfm_dtd = 'file:/etc/netapp_dfm.dtd';
$::ZAPI_agent_dtd = 'file:/etc/netapp_agent.dtd';

# url for various server types
$::ZAPI_url = "/servlets/netapp.servlets.admin.XMLrequest_filer";
$::ZAPI_nc_url = "/servlets/netapp.servlets.admin.XMLrequest";
$::ZAPI_dfm_url = "/apis/XMLrequest";
$::ZAPI_agent_url = "/apis/XMLrequest";

# default port no for various server types
$::ZAPI_port = 80;
$::ZAPI_nc_port = 80;
$::ZAPI_dfm_port = 8088;
$::ZAPI_agent_port = 4092;

$::ZAPI_snoop = 0;
$::ZAPI_noauth = 0;

$Zapi::OK  = 0;
$Zapi::FAIL = -1;
$Zapi::ANY = "__ANY__";

# If the user of a Zapi object doesn't override the routines
# used for doing logging and indicating failure, these two routines
# are used by default.

sub fail_default {
    my $msg = shift;
    print STDERR "Zapi FAIL: ".$msg."\n";
}

sub log_default {
    my $msg = shift;
    print STDERR "Zapi LOG: ".$msg."\n";
}

my $ctx = "";

sub new {
    my ($class)  = shift;
    my ($server) = shift;
    my ($user)   = shift;
    my ($password) = shift;
    my ($timeout) = shift;

    my ($fail) = \&fail_default;
    my ($log) = \&log_default;
    my ($reserved_port) = 0;
    my ($major_version) = 1;
    my ($minor_version) = 0;
    my ($vfiler_server) = "";
    my ($url) = $::ZAPI_url;
    my ($port) = $::ZAPI_port;
    my ($legacy_http_conn) = 1;
    
    
    # If the password is not provided, and the server name
    # is not an IP address, then we can look up the default password for it

    if ( !defined($password)
        && ! ( "$server" =~ /[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/ ) ) {

        my $h = ::host("$server");
        $password = $h->{default_password};
        $user = "root";
        # The code down below assumes that the $server value is
        # an IP address, so we overwrite the server name with the IP.
        $server = $h->{default_ip};
    }

    # If admin username is not provided, we assume we're using
    # hosts.equiv authentication.
    if ( !defined($user) ) {
        $user = "root";
        $password = "";
        $reserved_port = 1;
    }

    # If timeout is not provided, we default it to 30 seconds
    if ( !defined($timeout) ) {
        $timeout = 30;
    }

    my $self = {
        server => $server,
        user   => $user,
        password => $password,
        reserved_port => $reserved_port,
        fail => $fail,
        log => $log,
        major_version => $major_version,
        minor_version => $minor_version,
        vfiler_server => $vfiler_server,
        url => $url,
        port => $port,
        transport_type => "HTTP",
        legacy_http_conn => $legacy_http_conn,
        server_type => "filer",
        timeout => $timeout
    };
    
    # Although there's a method to set the transport type to CLI,
    # this is a little hack that makes testing easier, you 
    # can just set an environment variable to enable it.
    if ( defined($ENV{ZAPI_USES_CLI}) && $ENV{ZAPI_USES_CLI} eq "yes" ) {
        $self->{transport_type} = "CLI";
    }
    if ( defined($ENV{ZAPI_USES_HTTPS}) && $ENV{ZAPI_USES_HTTPS} eq "yes" ) {
        # require this here so we work for tests that don't need SSL, and clients
        # that don't have SSL.  will still fail for tests that need SSL and clients
        # that don't have it.
        require Net::SSLeay;
        $self->{transport_type} = "HTTPS";
        $self->{port} = 443;
          Net::SSLeay::load_error_strings();
          Net::SSLeay::SSLeay_add_ssl_algorithms();
        # Random seed. 
          Net::SSLeay::randomize("", time ^ $$); 
        $ctx = Net::SSLeay::CTX_new() or 
            die "Failed to create SSL_CTX $!";
          Net::SSLeay::ERR_clear_error();
          Net::SSLeay::CTX_set_options($ctx, &Net::SSLeay::OP_ALL)
          and Net::SSLeay::die_if_ssl_error("ssl ctx set options");
    }
    if ( defined($ENV{ZAPI_USES_RPC}) && $ENV{ZAPI_USES_RPC} eq "yes" ) {
        $self->{transport_type} = "RPC";
        if ( defined($ENV{ZAPI_USES_RPC_HOST}) ) {
            $self->{transport_rpc_host} = $ENV{ZAPI_USES_RPC_HOST};
        } else {
            $self->{transport_rpc_host} = "localhost";
        }
         }


    bless $self, $class;
    return $self;
}

sub set_version {
    my $self = shift;
    $self->{major_version} = shift;
    $self->{minor_version} = shift;
}

sub get_major_version {
    my $self = shift;
    return( $self->{major_version} );
}

sub get_minor_version {
    my $self = shift;
    return( $self->{minor_version} );
}

sub set_fail_routine {
    my $self = shift;
    $self->{fail} = shift;
}

sub set_log_routine {
    my $self = shift;
    $self->{log} = shift;
}

sub set_transport_type {
    my $self = shift;
    my $t = shift;
    if ( $t ne "HTTP" && $t ne "CLI" && $t ne "HTTPS") {
        $self->{log}->("Zapi set_transport_type "
            . "given invalid value, must be HTTP, CLI, or HTTPS.  "
            . "Defaulting to HTTP\n");
        $self->{transport_type} = "HTTP";
    } else {
        $self->{transport_type} = $t;
    }
}

sub get_transport_type {
    my $self = shift;
    return( $self->{transport_type} );
}

sub set_port( $ ) {
    my $self = shift;
    $self->{port} = shift;
}

sub get_port( ) {
    my $self = shift;
    return( $self->{port} );
}

sub set_dtd( $ ) {
    my $self = shift;
    $self->{dtd} = shift;
}

sub get_dtd( ) {
    my $self = shift;
    return( $self->{dtd} );
}

sub set_vfiler_server( $ ) {
    my $self = shift;
    $self->{vfiler_server} = shift;
    # Vfiler tunneling requires version 1.7 or above
    my $major_v;
    my $minor_v;
    if ($self->{vfiler_server}) {
        $major_v = $self->get_major_version();
        $minor_v = $self->get_minor_version();
        if ($major_v < 2) {
            if ($minor_v < 7) {
                $self->set_version(1,7);
        }
        }
    }

}

sub get_vfiler_server( ) {
    my $self = shift;
    return( $self->{vfiler_server} );
}

sub set_url( $ ) {
    my $self = shift;
    $self->{url} = shift;
}

sub get_url( ) {
    my $self = shift;
    return( $self->{url} );
}

sub server {
    my $self = shift;
    return ( $self->{server} );
}

sub user {
    my $self = shift;
    return ( $self->{user} );
}

sub password {
    my $self = shift;
    return ( $self->{password} );
}

sub reserved_port {
    my $self = shift;
    return ( $self->{reserved_port} );
}

sub set_reserved_port {
    my $self = shift;
    $self->{reserved_port} = 1;
}

sub set_snoop {
    my $self = shift;
    my $level = shift;

    my $curr_level = $::ZAPI_snoop;
    $::ZAPI_snoop = $level;
    return $curr_level;
}

sub set_snoop_default {
    # Check the value of THARN_VERBOSE
    # and if it's 2, sets the snoop level so that
    # ->invoke() methods will provide more information.
    my $self = shift;
    if ( $::THARN_VERBOSE >= 2 ) {
        $self->set_snoop(1);
    }
}

sub use_cli {
    my $self = shift;
    if ( $self->{transport_type} eq "CLI" ) {
        return 1;
    } else {
        return 0;
    }
}

sub use_https {
    my $self = shift;
    if ( $self->{transport_type} eq "HTTPS" ) {
        return 1;
    } else {
        return 0;
    }
}

sub use_rpc {
    my $self = shift;
    if ( $self->{transport_type} eq "RPC" ) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::set_zapi_param
## @synopsis Set the parameters of the ZAPI API 
## @description Save the reference to the parameters of the ZAPI
##      to be invoked
## @param $apiParam      Reference to the hash specifying the API
##         parameters       
##
## Expects to be used as a method.

sub set_zapi_param {
    my $self = shift;
    my $apiParam = shift;
    $self->{api_param} = $apiParam;

}

## @name ::set_zapi_name
## @synopsis Set the API name of the API to invoke
## @description Save the name of the API to invoke
##     
## @param $apiName      Name of the API to invoke
##
## Expects to be used as a method.

sub set_zapi_name {
    my $self = shift;
    my $apiName = shift;
    $self->{api_name} = $apiName;
}

## @name ::get_legacy_http_conn
## @synopsis    Returns the http connection method used by the lib
## @description If set to 0, the LWP::UserAgent lib will be used
##      If value is 1, the original legacy mode will be used (default).

sub get_legacy_http_conn {
    my $self = shift;
    return( $self->{legacy_http_conn} );
}

## @name ::set_legacy_http_conn
## @synopsis    Controls the http connection method used by the lib
## @description If set to 0, the LWP::UserAgent lib will be used
##      If value is 1, the original legacy mode will be used (default).
## @param $bool     Boolean value to set legacy mode.      

sub set_legacy_http_conn {
    my $self = shift;
    $self->{legacy_http_conn} = shift;
}

## @name ::set_server_type
## @synopsis    Determines the corresponding DTD file name and set it
## @description Based on the given server type, the external DTD declaration
##     to be wrapped in DOCTYPE definition of input xml will be determined.
##     Possible values for server type can be filer (default), dfm, agent
##     and netcache. This will also set the default port no and default url
##     corresponding to the server type. If users want to override these
##     values, then they can make use of set_port() and set_url() to set the
##     overrides after making a call to this method

sub set_server_type {
    my $self = shift;
    $self->{server_type} = shift;
    my $type = lc($self->{server_type});
    my ($dtd_file, $url, $port);
    if ($type =~ /netcache/i) {
        $dtd_file = $::ZAPI_nc_dtd;
        $url = $::ZAPI_nc_url;
        $port = $::ZAPI_nc_port;
    } elsif ($type =~ /dfm/i) {
        $dtd_file = $::ZAPI_dfm_dtd;
        $url = $::ZAPI_dfm_url;
        $port = $::ZAPI_dfm_port;
    } elsif ($type =~ /agent/i) {
        $dtd_file = $::ZAPI_agent_dtd;
        $url = $::ZAPI_agent_url;
        $port = $::ZAPI_agent_port;
    } else {
        # If the server type is "filer" or any other invalid value is passed,
        # then the default value will be taken for DTD.
        # This how ontapi also works
        $dtd_file = $::ZAPI_dtd;;
        $url = $::ZAPI_url;
        $port = $::ZAPI_port;
    }
    $self->set_dtd($dtd_file);
    $self->set_url($url);
    $self->set_port($port);
}

## @name ::get_server_type
## @synopsis    Returns the server type which decides the DTD for input xml
## @description Based on the given server type, the external DTD declaration
##     to be wrapped in DOCTYPE definition of input xml will be determined.
##     Possible values for server type can be filer (default), dfm, agent
##     and netcache.

sub get_server_type {
    my $self = shift;
    return( $self->{server_type} );
}

## @name ::fail_response
## @synopsis    Return an NaElement which is a <results> containing a failure.
## @description Return an NaElement which is a <results> containing a failure.
##      This is used when the transmission path fails, and we don't actually
##      get back any XML from the server.

sub fail_response {
    my $self    = shift;
    my $errno   = shift;
    my $reason  = shift;

    my $n = new NaElement("results");
    $n->attr_set("status","failed");
    $n->attr_set("reason","$reason");
    $n->attr_set("errno","$errno");
    return $n;
}

## @name ::invoke_http
## @synopsis    Submit an XML request using HTTP(S)
## @description Prepares an XML API request, submits it with
##      HTTP(S), and returns an NaElement object which 
##      contains all the results.
##      If set_reserved_port has been used, the source
##      port of the socket will be set to a low (1023 or below)
##      port number, so that /etc/hosts.equiv authentication can be used.
## @param $req      An NaElement object containing the request.
## @param $expected Optional, this value indicates the expected status.
##          If this value is "passed" or "failed", and the
##          expected status does not match it,
##          we generate a test failure.
##          If this value is "any", then we allow any
##          result, and don't complain about whatever it is.

sub invoke_http ( $$@ ) {
    my $self    = shift;
    my $req     = shift;
    my $expected    = shift;

    my $server  = $self->server;
    my $user    = $self->user;
    my $password    = $self->password;

    my ($sockaddr);
    my ($name, $aliases, $proto, $port, $type, $len, $thisaddr);
    my ($thisport, $thatport);
    my ($thataddr);

    my $using_ssl = $self->use_https(); 
    my $ssl;
    my $xml;
    my $response;
    if ( ! defined($expected) ) {
        $expected = "passed";
    }

    my $xmlrequest = $req->toEncodedString();
    
        # This is the filer url, in a form acceptable
    # to the method line of an HTTP transaction.

        my $url = $self->{url};
    $port = $self->{port};

    if ( $::ZAPI_snoop > 0 ) {
        $self->{log}->("Zapi invoke_http "
            . "server=$server using_ssl=$using_ssl\n"
            . "legacy_http_mode=$self->{legacy_http_conn}\n"
            . "user=$user password=$password port = $port\n"
            . "url=$url\n"
            . "xmlrequest=$xmlrequest\n");
    }
       
    #XXX This constant should really be set in .ntest
    if (!$self->{dtd}) {
        $self->{dtd} = $::ZAPI_dtd;
    }
    my $content;
    $content = "<?xml version='1.0' encoding='utf-8' ?>"
        ."<!DOCTYPE netapp SYSTEM '".$self->{dtd}."'>"
        ."<netapp version='"
        .$self->{major_version}.".".$self->{minor_version}
        ."' xmlns='$::ZAPI_xmlns";
    if ($self->{vfiler_server}) {
        $content .= "' vfiler='$self->{vfiler_server}";
    }
    $content .= "'>"
             .$xmlrequest
             ."</netapp>";

    unless ($self->{legacy_http_conn}) {
        if ( $self->reserved_port() ) {
           return $self->fail_response(13001,
               "in Zapi::invoke, unable to bind "
               ."to reserved port, not implemented ");
        }
        my $http_ua = new LWP::UserAgent;
        $http_ua->timeout($self->{timeout});
        my $proto = $self->use_https() ? "https" : "http";
        my $http_url = "${proto}://${server}:${port}${url}";
        my $http_request = HTTP::Request->new('POST', "$http_url");
        # Normally we provide authorization, but this is a hook that
        # a test can use to explicitly avoid sending authorization
        if ( $::ZAPI_noauth == 0 ) {
            $http_request->authorization_basic($user, $password);
        }
        if (Encode::is_utf8($content)) {
            my $octets = encode("utf8",$content);
            $content = $octets;
        }
        $http_request->content($content);
        $http_request->content_length(length($content));
        $http_request->content_type("text/xml");
        $response = $http_ua->request($http_request);
        
        $xml = $response->content();

    } else {
        #Use the old style connection
        #
        # Establish socket connection
        #
        $sockaddr = 'S n a4 x8';
        if ($using_ssl) {
            ($name,$aliases,$proto)=getprotobyname('ssl');
            $proto = 0;
        } else {
            ($name,$aliases,$proto)=getprotobyname('tcp');
        }
    
        ($name,$aliases,$type,$len,$thataddr)=gethostbyname($server);
    
        $thatport=pack($sockaddr, &AF_INET, $port, $thataddr);
        if ( !socket(S, &PF_INET,&SOCK_STREAM,$proto) ) {
            return $self->fail_response(13001,
                "in Zapi::invoke, cannot create socket");
        }
    
        #
        # If we are being asked to use a reserved port (used when we
        # are doing hosts.equiv authentication), then we search to
        # find an available port number below 1024.
        #
        if ( $self->reserved_port() ) {
            my $lowport;
            for ($lowport=1023; $lowport > 0; $lowport--) {
                $thisport=pack($sockaddr, &AF_INET,$lowport);
                if (bind(S,$thisport)) {
                    last;
                }
            }
            if ($lowport == 0) {
                return $self->fail_response(13001,
                    "in Zapi::invoke, unable to bind "
                    ."to reserved port, you must be "
                    ."executing as root");
            }
        } else {
            $thisport=pack($sockaddr, &AF_INET,0);
            if (!bind(S,$thisport)) {
                return $self->fail_response(13001,
                    "in Zapi::invoke, unable to bind to port");
            }
        }
    
        if ( ! connect(S,$thatport) ) {
            return $self->fail_response(13001,
                "in Zapi::invoke, cannot connect to socket");
        }
    
        S->autoflush(1);              # Turn on autoflushing
        
            # Create an HTTP request.
            my $request = HTTP::Request->new('POST',"$url");
    
        # Normally we provide authorization, but this is a hook that
        # a test can use to explicitly avoid sending authorization
        if ( $::ZAPI_noauth == 0 ) {
            $request->authorization_basic($user,$password);
        }
   
        if (Encode::is_utf8($content)) {
            $content = encode("utf8",$content);
        }
 
        $request->content($content);
        $request->content_length(length($content));
    
        my $methline = $request->method()." ".$request->uri()." HTTP/1.1\n";
        my $headers = $request->headers_as_string();
    
        if ($using_ssl) {
            $ssl = Net::SSLeay::new($ctx) or return $self->fail_response(13001,
                "in Zapi::invoke, failed to create SSL $!");
          Net::SSLeay::set_fd($ssl, fileno(S));   # Must use fileno
          Net::SSLeay::connect($ssl) or return $self->fail_response(13001,
                "in Zapi::invoke failed to connect SSL $!");
        }
    
    
        if ( $::ZAPI_snoop > 0 ) {
            $self->{log}->("Zapi Request:::\n");
            $self->{log}->("$methline");
            $self->{log}->("$headers");
            $self->{log}->("\n");
            $self->{log}->($request->content()."\n");
            if ($using_ssl) {
                $self->{log}->("Cipher '".Net::SSLeay::get_cipher($ssl) 
                       . "'\n");
            }
            $self->{log}->(":::\n");
        }
    
        if ($using_ssl) {
          Net::SSLeay::ssl_write_all($ssl, $methline);
          Net::SSLeay::ssl_write_all($ssl, $headers);
          Net::SSLeay::ssl_write_all($ssl, "\n");
          Net::SSLeay::ssl_write_all($ssl, $request->content());
            
        } else {
            print S $methline;
            print S $headers;
            print S "\n";
            print S $request->content();
        }
    
        $xml = "";
    
        # Inside this loop we will read the response line and all headers
        # found in the response.
    
        my $state = 0;  # 1 means we're in headers, 2 means we're in content
        my ($key, $val);
        my $line;
        while (1) {
                if ($using_ssl) {
                $line = Net::SSLeay::ssl_read_CRLF($ssl);
            } else {
                $line = <S>;
            }
                
            if ( !defined($line)  || $line eq "" ) {
                last;
            }
            if ( $::ZAPI_snoop > 0 ) {
                $self->{log}->("Response:::".$line."\n");
            }
            if ( $state == 0 ) {
                if ($line =~ s/^(HTTP\/\d+\.\d+)[ \t]+(\d+)[ \t]*([^\012]*)\012//) {
                    # HTTP/1.0 response or better
                    my($ver,$code,$msg) = ($1, $2, $3);
                    $msg =~ s/\015$//;
                    $response = HTTP::Response->new($code, $msg);
                    $response->protocol($ver);
                    $state = 1;
                    next;
                } else {
                    return $self->fail_response(13001,
                        "in Zapi::invoke, unable to parse "
                        ."status response line - $line");
                }
            } elsif ( $state == 1 ) {
                # ensure that we have read all headers.
                # The headers will be terminated by two blank lines
                if ( $line =~ /^\r*\n*$/ ) {
                    $state = 2;
                } else {
                    if ($line =~ /^([a-zA-Z0-9_\-.]+)\s*:\s*(.*)/) {
                        $response->push_header($key, $val) if $key;
                        ($key, $val) = ($1, $2);
                    } elsif ($line =~ /^\s+(.*)/ && $key) {
                        $val .= " $1";
                    } else {
                        $response->push_header(
                            "Client-Bad-Header-Line" => $line);
                    }
                }
            } elsif ( $state == 2 ) {
                $xml .= $line;
            } else {
                return $self->fail_response(13001,
                    "in Zapi::invoke, bad state value "
                    ."while parsing response - $state\n");
            }
        }
    
        if ($using_ssl) {
          Net::SSLeay::free ($ssl);               # Tear down connection
        }
    }
    if ( ! defined($response) ) {
        $self->{fail}->("No response received?");
        $xml = "<results reason=\"No response received\" "
                ."status=\"failed\" errno=\"13001\" />";
    } else {
        my $code = $response->code();
        if ( $code == 401 ) {
            return $self->fail_response(13002,"Authorization failed");
        }
        #Catch a general failure in the http response
        if (!$response->is_success) {
            return $self->fail_response(13001,
                "in Zapi::invoke, Http session failure:".$response->status_line);
        }
    }

    my $obj = $self->parse_xml($xml,$expected,$xmlrequest);
    # Log all the zapi details in NATE db
    log_zapi::write_in_db( input => $xmlrequest, output => $obj->sprintf());
    return $obj;
}

## @name ::invoke_cli
## @synopsis    Submit an XML request using the CLI
## @description Prepares an XML API request, submits it with rsh,
##      and returns an NaElement object which contains all the results.
##      
## @param $req      An NaElement object containing the request.
## @param $expected Optional, this value indicates the expected status.
##          If this value is "passed" or "failed", and the
##          expected status does not match it,
##          we generate a test failure.
##          If this value is "any", then we allow any
##          result, and don't complain about whatever it is.

sub invoke_cli ( $$@ ) {
    my $self    = shift;
    my $req     = shift;
    my $expected    = shift;

    my $server  = $self->server;
    my $user    = $self->user;
    my $password    = $self->password;

    if ( ! defined($expected) ) {
        $expected = "passed";
    }

    my $xmlrequest = $req->toEncodedString();

    if (!$self->{dtd}) {
        $self->{dtd} = $::ZAPI_dtd;
    }

    my $content;
    $content = "<?xml version='1.0' encoding='utf-8' ?>"
        ."<!DOCTYPE netapp SYSTEM '".$self->{dtd}."'>"
        ."<netapp version='"
        .$self->{major_version}.".".$self->{minor_version}
        ."' xmlns='$::ZAPI_xmlns";
    if ($self->{vfiler_server}) {
        $content .= "' vfiler='$self->{vfiler_server}";
    }
    $content .= "'>"
        .$xmlrequest
        ."</netapp>";

    my $f = "$server ; default_password=$password; hosttype=filer";
    my $xml = cli_advanced($f,"ontapi execute \"$content\"");
    
    my $obj = $self->parse_xml($xml,$expected,$xmlrequest);
	
    # Log all the zapi details in NATE db
    log_zapi::write_in_db( input => $xmlrequest, output => $obj->sprintf());

    return $obj;
}

## @name ::invoke_rpc
## @synopsis    Submit an XML request using the RPC on the Windows host
## @description Prepares an XML API request, submits it with RPC,
##      and returns an NaElement object which contains all the results.
##      
## @param $req      An NaElement object containing the request.
## @param $expected Optional, this value indicates the expected status.
##          If this value is "passed" or "failed", and the
##          expected status does not match it,
##          we generate a test failure.
##          If this value is "any", then we allow any
##          result, and don't complain about whatever it is.

sub invoke_rpc ( $$@ ) {
    my $self    = shift;
    my $req     = shift;
    my $expected    = shift;

    my $server  = $self->server;
    my $user    = $self->user;
    my $password    = $self->password;

    if ( ! defined($expected) ) {
        $expected = "passed";
    }

    my $xmlrequest = $req->toEncodedString();

    if (!$self->{dtd}) {
        $self->{dtd} = $::ZAPI_dtd;
    }

    my $content = $xmlrequest;

    my $rpc_spec = "$self->{transport_rpc_host};hosttype=windows";

    my $c = Tharn::connect("$rpc_spec#rsh");

    my $xml;

    # We cannot use runcmd.pl utility here as it interfers with
    # xml string specification. For this to work, target Windows
    # machine has to have enough privileges to establish RCP pipe
    # with the target filer.

    (undef, undef, $xml) = Tharn::say($c,
            "cmd.exe /c x:\\bin\\ontapi -ri $server $user $password \"$content\"");

    my $obj = $self->parse_xml($xml,$expected,$xmlrequest);
	     
    # Log all the zapi details in NATE db	 
    log_zapi::write_in_db( input => $xmlrequest, output => $obj->sprintf());
    return $obj;
}
#################################################################
## @name ::parse_raw_xml
## @synopsis    Take an XML response and parse it.
## @description Prepares an XML API response, parses it,
##              and returns an NaElement object 
## @param $xml          complete XML response
#################################################################
sub parse_raw_xml($$) {
 
    my $self        = shift;
    my $xml         = shift;

    $::ZAPI_stack = [];
 
    my $p = new XML::Parser(ErrorContext => 2);
    $p->setHandlers(
        Start => \&server_start_handler,
        Char => \&server_char_handler,
        End => \&server_end_handler
    );
    $p->parse($xml);

    if ( $#$::ZAPI_stack < 0 ) {
        return $self->fail_response(13001,
        "Zapi::parse_xml - no elements on stack");
    }
    my $r = pop(@$::ZAPI_stack);

    return $r;
}
#################################################################


## @name ::parse_xml
## @synopsis    Take an XML response and parse it.
## @description Prepares an XML API response, parses it,
##      and returns an NaElement object which 
##      contains all the results.
## @param $xml      complete XML response
## @param $expected The expectation to pass or fail.
##          Value is "passed", "failed", or "any".
## @param $xmlrequest   original XML request

sub parse_xml {

    my $self    = shift;
    my $xml     = shift;
    my $expected    = shift;
    my $xmlrequest  = shift;

    if ( $::ZAPI_snoop > 0 ) {
        $self->{log}->("Zapi parse_xml xml=".$xml."\n");
    }

    $::ZAPI_stack = [];

    my $p = new XML::Parser(ErrorContext => 2);
    $p->setHandlers(
        Start => \&server_start_handler,
        Char => \&server_char_handler,
        End => \&server_end_handler
        );
    $p->parse($xml);

    if ( $#$::ZAPI_stack < 0 ) {
        return $self->fail_response(13001,
            "Zapi::parse_xml - no elements on stack");
    }
    my $r = pop(@$::ZAPI_stack);

    if ( $::ZAPI_snoop > 0 ) {
        $self->{log}->("Zapi::invoke_reserved_port raw output="
                    .$r->toString()."\n");
    }

    if ( $self->use_rpc == 0 && $r->name() ne "netapp" ) {
        return $self->fail_response(13001,
            "Zapi::parse_xml - Expected <netapp> element, "
            ."but got ".$r->name());
    }

    # We return the first element inside the <netapp>, which
    # should be <results>.  This only works if the original request
    # has only a single request.
    my $c;
    if ( $self->use_rpc == 0) {
        $c = $r->child_get_nth(0);
    } else {
        $c = $r;
    }
    if ( $c->name() ne "results" ) {
        return $self->fail_response(13001,
            "Zapi::parse_xml - Unexpected element "
            ."in output - expected results, got ".$c->name());
    }

    if ( defined($expected) && $expected ne "any" ) {
        my $st = $c->attr_get("status");
        if ( $st ne $expected ) {
            my $reason;
            if ( $st eq "failed" ) {
                $reason = $c->results_reason();
            } else {
                $reason = "It worked";
            }
            return $self->fail_response(13001,"Zapi::parse_xml - "
                ."Unexpected status - expected $expected, "
                ."got $st.  Reason=[[$reason]]  Original "
                ."request was [[".$xmlrequest."]]");
        }
    }

    return $c;
}

sub server_start_handler ($$@) {
    my $xp = shift;
    my $el = shift;

    my $n = new NaElement("$el");
    push(@$::ZAPI_stack,$n);

    my $sz = $#$::ZAPI_stack;

    %::ZAPI_atts = ();
    while ( @_ ) {
        my $att = shift;
        my $val = shift;
        $::ZAPI_atts{$att} = $val;
        $n->attr_set($att,$val);
    }
}

sub server_char_handler {
    my $xp = shift;
    my $data = shift;

    my $i = $#$::ZAPI_stack;
    $::ZAPI_stack->[$i]->content_add($data);
}

sub server_end_handler {
    my $xp = shift;
    my $el = shift;

    # We leave the last element on the stack.
    if ( $#$::ZAPI_stack > 0 ) {
        my $sz = $#$::ZAPI_stack;

        # Pop the element and add it as a child
        # to its parent.
        my $n = pop(@$::ZAPI_stack);
        my $ns = $n->toString();
        my $i = $#$::ZAPI_stack;

        $::ZAPI_stack->[$i]->child_add($n);
    }
}

## @name ::invoke
## @synopsis    Submit an XML request using HTTP.
## @description Prepares an XML API request, takes the parameters, 
##      builds an element structure, submits it with
##      HTTP, and returns an NaElement object which 
##      contains all the results.
## @param $api      The API to be invoked.
## @param $parm0-n  The API parameters to be invoked.
## 
## Expects to be used as a method.

sub invoke {
    my $self = shift;
    my $api  = shift;

    my $num_parms = @_;
    my $i;
    my $key;
    my $value;

    $self->set_snoop_default();

    if ( ($num_parms & 1) != 0 ) {
        return $self->fail_response(13001,
            "in Zapi::invoke, invalid number of parameters");
    }

    my $xi = new NaElement($api);

    for ($i = 0; $i < $num_parms; $i += 2) {
        $key = shift;
        $value = shift;
        $xi->child_add(new NaElement($key, $value));
    }
    if ( $self->use_cli() ) {
        return $self->invoke_cli($xi);
    }
    if ( $self->use_rpc() ) {
        return $self->invoke_rpc($xi);
    } else {
        return $self->invoke_http($xi);
    }
}

## @name ::invoke_api
## @synopsis    Submit an XML request using HTTP.
## @description Prepares an XML API request, takes the parameters, builds 
##      an element structure, submits it with HTTP, and returns an
##      NaElement object which contains all the results.  The results
##      must be checked after the call.  This method does not check
##      the results.
## @param $api      The API to be invoked.
## @param $parm0-n  The API parameters to be invoked.
## 
## Expects to be used as a method.

sub invoke_api {
    my $self     = shift;
    my $api      = shift;

    my $num_parms = @_;
    my $i;
    my $key;
    my $value;

    $self->set_snoop_default();

    if ( ($num_parms & 1) != 0 ) {
        return $self->fail_response(13001,
            "in Zapi::invoke_api, invalid number of parameters");
    }

    my $xi = new NaElement($api);

    for ($i = 0; $i < $num_parms; $i += 2) {
        $key = shift;
        $value = shift;
        $xi->child_add(new NaElement($key, $value));
    }

    if ( $self->use_cli() ) {
        return $self->invoke_cli($xi, "any");
    }
    if ( $self->use_rpc() ) {
        return $self->invoke_rpc($xi, "any");
    } else {
        return $self->invoke_http($xi, "any");
    }
}

## @name ::construct_api
## @synopsis    Build the XML request for the API to invoke
## @description 
##

sub construct_api() {
    my $self = shift;

    my $ele      = new NaElement($self->{api_name});
    my $finalEle = $self->construct_element($ele, $self->{api_param});

    return $finalEle;
}

## @name ::construct_elements
## @synopsis    Recursively construct element from the hash
## @description Recursively construct XML elements based on the content of the hash
##
## Expects to be used as a method.

sub construct_element($$) {

    my $self  = shift;
    my $ele   = shift;
    my $param = shift;

    my %pHash = %$param;

    # Add key and value to element
    foreach my $key (keys %pHash) {
        my $value = $pHash{$key};

        # if there is another hash in the value variable
        # which means it has child
        # go further level to abstract the key and value
        if (ref($value) eq "HASH") {
            my $newEle = new NaElement($key);
            $ele->child_add($newEle);
            my $ele = $self->construct_element($newEle, $value);


        } elsif (ref($value) eq "ARRAY") {
            # If the reference of value is an array,
            # which means there are more than one children.
            # They might be a leaf or contain other hash.
            # Check the first child and see if it points to a hash.
            # If it is, means each item in the array
            # point to a hash, then further process is needed
            if (ref($value->[0]) eq "HASH") {

                # element 0 point to a hash
                # Then it is not leaf element
                # create a new element for each entry in the array
                foreach my $ref (@$value) {
                    my $newEle = new NaElement($key);
                    $ele->child_add($newEle);

                    # get the value of each item
                    $self->construct_element($newEle, $ref);
                }
            } else {

                # It is leaf level
                foreach my $val (@$value) {

                    # Leave, add  child element
                    $ele->child_add_string($key, $val);
                }
            }
        } else {

            # Add child element
            $ele->child_add_string($key, $value);
        }
    } ## end foreach my $key (keys %pHash)
    return $ele;
} ## end sub construct_element($$)


## @name ::check_result
## @synopsis    Check the response compared to expected results
## @description Check status, errno and or reason returned compared to 
##      expected results
##     
## @param xo: XML response
## @param exp_status: expected status. Zapi::ANY means value of status does not matter
## @param exp_errno: expected errno. Zapi::ANY means value of errno does not matter
## @param exp_reason: expecetd reason: Zapi::ANY means value of reason does not matter
##
## Expects to be used as a method.

sub check_result {
        my $self = shift;
        my $xo = shift;
        my $exp_status = shift;
        my $exp_errno = shift;
        my $exp_reason = shift;

        my $status =  $xo->attr_get("status");
        if (( $exp_status ne $Zapi::ANY ) && ( $status ne $exp_status)) {
		my $reason_msg=" ";
		if ($exp_status eq "passed") {
                   $reason_msg = "errno:" . $xo->attr_get("errno") 
                                . ", reason: " . $xo->attr_get("reason");
                } 
                return ($Zapi::FAIL, "Zapi returned status is: \'$status\',"
                        . " but  was expecting: \'$exp_status\'. $reason_msg\n");

        }

        if (defined($exp_reason)) {
                my $reason =  $xo->attr_get("reason");
                if (( $exp_reason ne $Zapi::ANY ) && ( $reason ne $exp_reason)) {
                        return ($Zapi::FAIL, 
				"Zapi returned reason is: \'$reason\',"
                        . " but  was expecting: \'$exp_reason\'");
                }
        }

        if (defined($exp_errno)) {
                my $errno =  $xo->attr_get("errno");
                if (( $exp_errno ne $Zapi::ANY ) && ( $errno ne $exp_errno)) {
                        return ($Zapi::FAIL, "Zapi returned errno is: \'$errno\',"
                        . " but  was expecting: \'$exp_errno\'");
                }
        }

        return ($Zapi::OK, "Result OK");
}

## @name ::invoke_zapi
## @synopsis    Submit an XML request .
## @description Prepares and send an XML request corresponding to an API
##      Before this method is invoked, the API and its parameter should be 
##      as follow:
##      1- call set_zapi_name method to set the name of the API
##      2- call set_zapi_param to specify a reference to the hash 
##         containing the API parameters
##      3- call invoke_zapi methid to build and send the XML request
##      4- Optionally call check_result method to check the result of the request
## 
## Expects to be used as a method.

sub invoke_zapi {
   my $self = shift;

   $self->set_snoop_default();

   my $xi = $self->construct_api();
    if ( $self->use_cli() ) {
        return $self->invoke_cli($xi, "any");
    }
    if ( $self->use_rpc() ) {
        return $self->invoke_rpc($xi, "any");
    } else {
        return $self->invoke_http($xi, "any");
    }
}



## @name ::invoke_elem
## @synopsis    Submit an XML request using HTTP.
## @description Prepares an XML API request, submits it with
##      HTTP, and returns an NaElement object which 
##      contains all the results.
## @param $xi       The elements to be invoked.
## @param $expected Optional argument to specify what result status
##          to expect.  Value is "passed", "failed", or "any".
##          If this argument is not provided, the default is "any",
##          meaning that the caller must check the result.
## 
## Expects to be used as a method.

sub invoke_elem (@) {
    my $self     = shift;
    my $xi       = shift;
    my $expected = shift;   # expected status: "passed", "failed", or "any"

    $self->set_snoop_default();

    if ( ! defined($expected) ) {
        $expected = "any";
    }
    if ( $self->use_cli() ) {
        return $self->invoke_cli($xi, $expected);
    }
    if ( $self->use_rpc() ) {
        return $self->invoke_rpc($xi, $expected);
    } else {
        return $self->invoke_http($xi, $expected);
    }
}
	  
1;
