#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/ZapiUtil.pm#1 $
#
#Copyright (c) 2004 Network Appliance, Inc.
#All rights reserved.
#
## @summary Library of functions for using the ONTAP Zephyr API (ZAPI)
## @description
## This library provides 3 types of functions for interfacing
## with NetApp filers or Netcaches via ZAPI
## 1) A basic function to support ZAPI requests.
## 2) A set of functions to support filer CLI commands over ZAPI
##  using any filer privilege.
## 3) A function, zapi_error_check_say, for invoking ZAPI requests
## and testing for their expected return code and output text.
## This function is the one recommended to verify
## expected error message text from a ZAPI executed command.
## This function logs each call as a PASS or FAIL test case completion.
## It also returns a 1 or 0 to indicate PASS or FAIL.
##
## All calls to these functions that do not specify valid parameters
## will result in a call to ExitScriptError that also terminates the test script.

## This package was based on the lib/misc.pm
## and lib/ZapiTest.pm modules with the following changes:
## 1) Name the package to make ZAPI support obvious
## 2) remove use of the deprecated tstlog function
## 3) Add log output of invoked ZAPI commands
## 4) Add log output of ZAPI command output
## 5) Count ZAPI requests to enable test metrics

## @usage
##   use ZapiUtil;
## @keywords ZAPI, Zephyr
## @dependencies Filer: 
## @author Charlie Hedstrom (hedstrom@netapp.com)
## @status public
## @change 2004-11-03 hedstrom: created
## @change 2005-04-07 korry: add zapi_parse_output routine
## @change 2005-04-15 korry: fix zapi_parse_output to return (none) string for empty value for its tag.
## @change 2005-05-03 korry: fix zapi_parse_output to return undef right away if first tag does not exist.
## @change 2006-01-27 korry: remove ^ in the match pattern for list-entries tag.

# When adding new functions, please:
# 1) Maintain the alphabetical order of the functions
# 2) Define new function names that start with zapi_ and
# 3) Add additional descriptive nouns to the function name to define the 
#  portion of ZAPI that the new function deals with and
# 4) End the function name with a verb. (Try to reuse the same verbs that appear on other functions)
#    The following verbs  are reserved for functions that can cope with unknown initial states:
#    initialize - completely sets up a service, returns 0 if the service is not available or cannot be started
#    make - create a new object if it does not exist, fail only if the object does not exist at the end
#    vaporize - destroys an object only reporting a failure if the object still exists at the end
# 6) Check input parameter values. Use ExitScriptError to report out of range values.
# 7) Use ExitError to report the inability to perform a request that would change the filer state.
#    or a serious problem that indicates an unexpected change in filer state
#    (For example, lost of the ability to communicate with the filer)
# 8) Use ExitSoftError to report other unexpected results.

use strict;

package ZapiUtil;

use Burt;
use NaElement;
use Tharn;
use Zapi;

use vars qw(
   @ISA @EXPORT
   $ONTAP_ZAPI_COUNT
   );

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw( 
   zapi_cli_advanced_say 
   zapi_cli_diag_say 
   zapi_cli_say 
   zapi_cli_test_say
   zapi_error_check
   zapi_say
   zapi_parse_output
   );

use vars ('@ISA', '@EXPORT');

sub new {
    my $class = shift;
    my $self = new Zapi(@_);
    $self->set_fail_routine(\&Burt::ExitSoftError);
    $self->set_log_routine(\&OntapZapi::nate_log);
    bless $self, $class;
}

sub nate_log {
        my $msg = shift;
    Tharn::tstlog('-commline',$msg);
}

$ONTAP_ZAPI_COUNT   = 0;

sub zapi_cli_priv_say( $$@ );
sub zapi_say ($$;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$);

# ----------------------------------------------------------------------
## @name ::zapi_cli_advanced_say
## @synopsis
## ($output, $value) = zapi_cli_advanced_say($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the advanced
##    priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no parameters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli_advanced($server, "license", "UYTCNCG");
##
# ----------------------------------------------------------------------
sub zapi_cli_advanced_say( $@ )
{
    my $server = shift;

    my ($output, $value) = zapi_cli_priv_say($server, "advanced", @_);
    return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::zapi_cli_diag_say
## @synopsis
## ($output, $value) = zapi_cli_diag_say($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the diag
##    priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no parameters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli_advanced($server, "license", "UYTCNCG");
##
# ----------------------------------------------------------------------
sub zapi_cli_diag_say( $@ )
{
    my $server = shift;

    my ($output, $value) = zapi_cli_priv_say($server, "diag", @_);
    return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::zapi_cli_test_say
## @synopsis
## ($output, $value) = zapi_cli_advanced_say($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the test
##    priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no parameters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli_advanced($server, "license", "UYTCNCG");
##
# ----------------------------------------------------------------------
sub zapi_cli_test_say( $@ )
{
    my $server = shift;

    my ($output, $value) = zapi_cli_priv_say($server, "test", @_);
    return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::zapi_cli_priv_say
## @synopsis
## ($output, $value) = zapi_cli_priv_say($server, $priv, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API.
##
## @params
##    $server:  Required: A zapi server object obtained via new OntapZapi( ).
##    $priv: Required: The privileged of the command.
##    cmd: Required: The command with no paramenters.
##    parm[1-n]: Optional: any command parameter.
##
# ----------------------------------------------------------------------
sub zapi_cli_priv_say( $$@ )
{
    my $server = shift;
    my $priv   = shift;
    my $num_parms = @_;

    my $xi;
    my $xo;
    my $args;
    my $i;
    my $arg;
    my $api = "system-cli";

    $xi = new NaElement($api);
    $xi->child_add(new NaElement("priv", $priv));
    $args = new NaElement("args");
    my $trace_message = "ZAPI: api=$api priv=$priv" ;

    # load up the command and the arguments

    for ($i = 0; $i < $num_parms; $i += 1 ) {
        $arg = shift;
        $args->child_add(new NaElement("arg",$arg));
        $trace_message .= " arg=$arg";
    }

    # add the command and arguments to the api
    $xi->child_add($args);

    # execute the command
    $ONTAP_ZAPI_COUNT++;  # Count the commands
    my $oe = $server->invoke_elem($xi, "any");

    if ( ! $oe->results_ok() ) {
        ExitSoftError("server->invoke failed.");
        return undef, undef;
    }

    my $output = $oe->child_get_string("cli-output");
    my $value = $oe->child_get_string("cli-result-value");
    logcomment("output=$output","result=$value");

    return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::zapi_cli_say
## @synopsis
## ($output, $value) = z_cli($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the admin priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no paramenters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli($server, "qtree", "status");
##
# ----------------------------------------------------------------------
sub zapi_cli_say( $@ )
{
    my $server = shift;

    my ($output, $value) = zapi_cli_priv_say($server, "admin", @_);
    return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::zapi_error_check_say
## @synopsis
## ($ok) = zapi_error_check_say($i,$o,$expected)
##
## @description
##  This function makes a ZAPI call and verifies
##  an expected error number
##  and/or message text is returned.
##  It returns 1 if the expected results were was found,
##  a 0 otherwise.
##
## @params
##    filer Required: filer to which to send the API.
##    expectedrc Required: although undef will check for non-zero
##    expectedmsg Required: undef will ignore returned text
##    api Required: api name
##    key1 Required: first (optional) parameter keyword
##    val1 Required: first (optional) parameter value
##
# ----------------------------------------------------------------------
sub zapi_error_check_say ($$$;$) {

    my $filer = shift;
    my $expectederrno = shift;
    my $expected = shift;
    my $api = shift;
    my @parameters = @_;

    my ($errno,$reason,$out) = zapi_say($filer,$api,@parameters);

    if (defined($expectederrno)) {
        if ($expectederrno eq '0') {
            # There were no expected error messages
            if ($errno ne '0') {
                ExitSoftError("ZAPI - "
                    ."Expected no errors, got return code $errno "
                    ."output text='$reason', "
                    ."got Output=$out"
                    ."from Input=$api ".join(' ',@parameters)
                    );
                return(0);
            } else {
                logresult('PASS',"Verified return code: $errno "
                          ."from Input=$api ".join(' ',@parameters)
                            );
                return(1);
            }
        } else {
            # There were expected error messages
            if ($errno ne '0') {
                # Fall through this logic, test expected text
            } else {
                ExitSoftError("ZAPI - "
                    ."Expected an error, got return code $errno "
                    ."output text='$reason', "
                    ."got Output=$out"
                    ."from Input=$api ".join(' ',@parameters)
                    );
                return(0);
            }
        }
    }

    # Any non-zero errno is OK
    if ($errno ne '0') {
        # Now make sure that expected text was delivered
        if (defined($expected)) {
            if ($reason =~ m/$expected/) {
                logresult('PASS',"Verified non-zero return code $errno and expected text "
                          ."from Input=$api ".join(' ',@parameters)
                            );
                return(1);
            } else {
                ExitSoftError("ZAPI - "
                    ."Expected output text='$expected', "
                    ."got Output=$out"
                    ."from Input=$api ".join(' ',@parameters)
                    );
                return(0);
            }
        } else {
            logresult('PASS',"Verified non-zero return code: $errno "
                      ."from Input=$api ".join(' ',@parameters)
                        );
            return(1);
        }
    } else {
        # Some error number was expected, got none
        ExitSoftError("ZAPI - "
            ."Unexpected result - expected non-zero error number, "
            ."got 0. "
            ."Input=$api ".join(' ',@parameters)
            ."Output=$out");
        return(0);
    }
}

# ----------------------------------------------------------------------
## @name ::zapi_say
## @synopsis
## ($err,$reason) = zapi($FILER,"sytem-get-version",...)
##
## @description
##  A simple routine to executes a zephyr-style API which has inputs only,
##  Returns:
##  - the full output 
##  - the errno,
##  - the reason string (Do not use if errno = 0)
##
## @params
##    filer Required: filer to which to send the API.
##    api Required: api name
##    key1 Required: first (optional) parameter keyword
##    val1 Required: first (optional) parameter value
##
# ----------------------------------------------------------------------
sub zapi_say ($$;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$) {

    my $filer = shift;
    my $api = shift;

    my $host = host($filer);

    my $ip = $host->{default_ip};
    my $password = $host->{default_password};

    my $server = new ZapiTest($ip, "root", $password);
    if ( ! defined $server ) {
        ExitError("Unable to connect via ZAPI to filer $filer at $ip");
        return undef;
    }
    my $e = new NaElement($api);
    my $trace_message = "ZAPI: api=$api";
    my ($key, $value);
    while ( 1 ) {
        $key = shift;
        $value = shift;
        if ( ! defined $key || ! defined $value ) {
            last;
        }
        $e->child_add(new NaElement($key,$value));
        $trace_message .= " $key=$value";
    }

    $ONTAP_ZAPI_COUNT++;  # Count the commands
    logcomment($trace_message);
    my $o = $server->invoke_elem($e,"any");
    my $errno = $o->results_errno();
    my $reason = $o->results_reason();
    my $out = $o->toString();
    logcomment("output=$out","result=$errno");

    return ($out,$errno,$reason);
}

#----------------------------------------------------------------------
## @name ::zapi_parse_output
## @synopsis
##    zapi_parse_output($zapi_output, tag1,...)
##
## @description
##  A routine to parse tag value(s) from a ZAPI output string
##  Returns:
##  - undef          - if tag1 is not found in string or alias name is unassigned
##  - $tagvalue      - if passing only tag1 or return $alias_name, if alias name is assigned 
##  - %hash          - if passing tag1 and tag2 - $hash{$key} = value
##  - %hash of hash  - if passing 3 or more tags - $hash{$key}{$tagN} = valueN
##
## @params
##    $zapi_output Required: returned output string from ZAPI call
##    $tag1 Required: tag1 to parse for its value (key)
##    $tag2 Optional: tag2 to parse for $hash{$key}{$tag2}=value1 or $hash{$key}=value1 if !defined(tag3)
##    $tag3 Optional: tag3 to parse for $hash{$key}{$tag3}=value2
# ----------------------------------------------------------------------
sub zapi_parse_output ($@) {
   my $line = shift;
   my @tag = @_;
   
   my %Hash;
   my ($tagkey, @tagvalue);

   # check begin of output string
   if($line =~ m/^<results status=.*?>/) {

      # $tag[0] does not exist, return undef
      if($line !~ m/<$tag[0]>/) {
         return undef;
      }

      # check for tpgroup-list-entry-info
      while ($line =~ m/<$tag[0]>/) {  # looking for tag[0]
         $line =~ s/<(.*?<$tag[0])>//;   # remove upto tag[0]
         if ($line =~ m/(.*?$tag[0]>)/) { # looking for end tag[0]
            $tagkey = $1;                 # get verything included end of tag[0]
            $tagkey =~ s/<(.*?$tag[0])>//;  # remove end of tag[0]
            if($tag[0] eq "is-alias-assigned") {
               if($tagkey eq "true") {
                  # get alias-name
                  if(defined($tag[1]) && $line =~ m/<$tag[1]>/) {
                     $line =~ s/(.*?$tag[1])>//;   # remove upto tag[1]
                     if ($line =~ m/(.*?$tag[1]>)/) { # looking for end of tag[1]
                        $tagvalue[0] = $1; # get verything included tag[1]
                        $tagvalue[0] =~ s/<(.*?$tag[1])>//; # remove end of tag[1]
                        return $tagvalue[0];
                     }
                  } else {
                     return undef;
                  }
               } else {
                  return undef;     # return undef for unassigned alias
               }
            } elsif (!defined($tag[1])) {
               return $tagkey;          # return $tagkey if !defined tag[1]
            }
            $line =~ s/(.*?$tag[1])>// if ($line =~ m/<$tag[1]>/);   # remove upto tag[1]
            if ($line =~ m/(.*?$tag[1]>)/) { # looking for end of tag[1]
               $tagvalue[0] = $1; # get verything included tag[1]
               $tagvalue[0] =~ s/<(.*?$tag[1])>//; # remove end of tag[1]
               $line =~ s/(.*?$tag[1])>//;  # remove upto end of tag[1]
               
               $tagvalue[0] = "enabled" if ($tagvalue[0] eq "true");
               $tagvalue[0] = "disabled" if ($tagvalue[0] eq "false");

               # if there are more tags, need hash of hash
               if(!defined($tag[2])) {
                  $Hash{$tagkey} = $tagvalue[0]; # hash
               } else {
                  $Hash{$tagkey}{$tag[1]} = $tagvalue[0]; # hash of hash
               }
               

            } else {
               ExitSoftError("Missing <$tag[1]> end tag from: $line\n");
            }
            #logcomment("tagkey: $tagkey, tagvalue[0]: $tagvalue[0]\n");

            # getting tag[2] and so forth
            for (my $cnt = 2; $cnt <= $#tag; $cnt++) {
               # getting tag3 if exists
               if(defined($tag[$cnt])) {
                  # if tag does not have a value , set to (none)
                  if ($line =~ m/<\w+-list-entries><\/\w+-list-entries>/ ) {
                     $Hash{$tagkey}{$tag[$cnt]} = '(none)';
                     next;
                  }
                  $line =~ s/(.*?$tag[$cnt])>// if ($line =~ m/<$tag[$cnt]>/);   # remove upto tag[cnt]
                  if ($line =~ m/(.*?$tag[$cnt]>)/) { # looking for end of tag[cnt]
                     $tagvalue[$cnt-1] = $1; # get verything included tag[cnt]
                     #print("tagvalue[$cnt-1]: $tagvalue[$cnt-1]\n");
                     $tagvalue[$cnt-1] =~ s/<(.*?$tag[$cnt])>//; # remove end of tag[cnt]
                     $line =~ s/(.*?$tag[$cnt])>//;  # remove upto end of tag[cnt]
                     $Hash{$tagkey}{$tag[$cnt]} = $tagvalue[$cnt-1];
                  } else {
                     ExitSoftError("Missing <$tag[$cnt]> end tag from: $line\n");
                  }
               }  
            }  
         } else {
            ExitSoftError("Missing <$tag[0]> end tag from: $line\n");
         }
      }

   } else {
      ExitSoftError("Unexpected result status from zapi:($line)");
   }

   return %Hash;
}

1;


#print "\n ".__FILE__." LOADED \n";

END {
    my $save_return_code = $?;
    #print("Starting END routine for ".__FILE__." with exit code $save_return_code\n");

    # This is the place to process statistics that were gathered
    # during the use of this module.
    if (defined($ONTAP_ZAPI_COUNT) and $ONTAP_ZAPI_COUNT ne '0') {
        # Only do this if there is something to report
        my @params;
        push(@params,"ONTAP_ZAPI_COUNT=$ONTAP_ZAPI_COUNT");
        my $result_obj = Rslt-> new(type=>'INFO',@params);
        $result_obj -> store();
    }

    # Restore the initial return code
    $? = $save_return_code;
}

1;
