# 
#  $Id: //depot/prod/test/Rsierranevada/lib/ZquotaEntry.pm#1 $
# 
#  Copyright (c) 2002 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary A zapi quota entry object
##
## @description
##	This file contains methods to manipulate a quota entry.  A
##	ZquotaEntry object is a ZquotaId object plus limits and represents
##	a quota entry in /etc/quotas.  To obtain the values from /etc/quotas,
##	the object must be created and the get_entry() method called.  The
##	set_entry() method is called to place the values in /etc/quotas.
##	A QuotaId object is needed by this object.
##
## @example
##	The example below creates a QuotaEntry object and gets the values
##	from /etc/quotas.
##
##	my $qid = new QuotaId($server, "user", "/home/tjt", $root_vol, "");
##	if ( ! defined($qid_tjt) ) {
##		tstlog( "-fail", "QuotaId object creation failed");
##		return;
##	}
##
##	my $qe = new QuotaEntry( $qid );
##	if ( ! defined($qe) ) {
##		tstlog( "-fail", "QuotaEntry object creation failed for " .
##			$qid->toString() );
##		return;
##	}
##	
##	my $result = $qe->get_entry();
##	if ( defined($result) ) {
##		tstlog( "-fail",
##			"quota-get-entry failed for " . $qe->toString() .
##			" with " . $result );
##	}
##
## @author rle@netapp.com

package     ZquotaEntry;

use	strict;
use	NaElement;
use	ZquotaId;
use	vars ('@ISA', '@EXPORT');
require     Exporter;
@ISA        = qw(Exporter);

## @name ZquotaEntry ::new
## @synopsis	creates a ZquotaEntry object.
## @description Creates a ZquotaEntry object from a QuotaId object, and
##		option key-value pair parameters..
## @param $qref the QuotaId object.
## @param @	optional key-value pairs, for example "disk-limit", "20000".
##		The default for limits is unlimited.

sub new {
	my $class  = shift;
	my $qref   = shift;   # QuotaId object
	my %parms  = @_;      # intial key-value pairs

	my $key;
	my $value;

	my $self = {
		"quota_ref"       => $qref,
		"disk-limit"      => "-",
		"file-limit"      => "-",
		"threshold"       => "-",
		"soft-disk-limit" => "-",
		"soft-file-limit" => "-",
		"user_mapping"    => 0
	};

	# overwrite the initial values with input parameters
	#
	while ( ( $key, $value) = each %parms ) {
		$self->{"$key"} = $parms{"$key"};
	}

	bless $self, $class;
	return $self;
}

## @name ZquotaEntry ::get_entry
## @synopsis	gets a quota entry from /etc/quotas
## @description Obtains a quota entry from /etc/quotas, and places the
##		information into the object's data.
## @param $qref    the QuotaId object.

sub get_entry {
        my $self = shift;
	my $qref = $self->{"quota_ref"};

	my ($results, $xo) = $qref->get_entry();
	if ( defined($results) ) {
		return $results;
	}
	
	$self->{"disk-limit"} = $xo->child_get_string("disk-limit");
	$self->{"file-limit"} = $xo->child_get_string("file-limit");
	$self->{"threshold"} = $xo->child_get_string("threshold");
	$self->{"soft-disk-limit"} = $xo->child_get_string("soft-disk-limit");
	$self->{"soft-file-limit"} = $xo->child_get_string("soft-file-limit");

	return undef;
}

## @name ZquotaEntry ::set_entry
## @synopsis	sets a quota entry to /etc/quotas
## @description Adds or modifies a quota entry in /etc/quotas from the
##		object's data.
## @param $qref    the QuotaId object.

sub set_entry {
        my $self = shift;
	my $qref = $self->{"quota_ref"};
	
	return $qref->set_entry("disk-limit", $self->{"disk-limit"},
				"file-limit", $self->{"file-limit"},
				"threshold", $self->{"threshold"},
				"soft-disk-limit", $self->{"soft-disk-limit"},
				"soft-file-limit", $self->{"soft-file-limit"});
}

## @name ZquotaEntry ::del_entry
## @synopsis	deletes a quota entry from /etc/quotas
## @description Removes a quota entry from /etc/quotas.
## @param $qref    the QuotaId object.

sub del_entry {
        my $self = shift;
	my $qref = $self->{quota_ref};

	return $qref->del_entry();
}

## @name ZquotaEntry ::add_entry
## @synopsis	adds a quota entry to /etc/quotas
## @description Adds a quota entry in /etc/quotas from the object's data.
## @param $qref    the QuotaId object.

sub add_entry {
        my $self = shift;
	my $qref = $self->{"quota_ref"};
	
	return $qref->add_entry("disk-limit", $self->{"disk-limit"},
				"file-limit", $self->{"file-limit"},
				"threshold", $self->{"threshold"},
				"soft-disk-limit", $self->{"soft-disk-limit"},
				"soft-file-limit", $self->{"soft-file-limit"});
}

## @name ZquotaEntry ::mod_entry
## @synopsis	modifyss a quota entry to /etc/quotas
## @description Modifies an existing quota entry in /etc/quotas from the
##		object's data.
## @param $qref    the QuotaId object.

sub mod_entry {
        my $self = shift;
	my $qref = $self->{"quota_ref"};
	
	return $qref->mod_entry("disk-limit", $self->{"disk-limit"},
				"file-limit", $self->{"file-limit"},
				"threshold", $self->{"threshold"},
				"soft-disk-limit", $self->{"soft-disk-limit"},
				"soft-file-limit", $self->{"soft-file-limit"});
}

## @name ZquotaEntry ::get_disk_limit
## @synopsis	gets the disk limit
## @description Returns the disk limit from the object's data.

sub get_disk_limit {
        my $self = shift;
        return ( $self->{"disk-limit"} );
}

## @name ZquotaEntry ::get_file_limit
## @synopsis	gets the file limit
## @description Returns the file limit from the object's data.

sub get_file_limit {
        my $self = shift;
        return ( $self->{"file-limit"} );
}

## @name ZquotaEntry ::get_threshold
## @synopsis	gets the threshold
## @description Returns the threshold from the object's data.

sub get_threshold {
        my $self = shift;
        return ( $self->{"threshold"} );
}

## @name ZquotaEntry ::get_soft_disk_limit
## @synopsis	gets the soft disk limit
## @description Returns the soft disk limit from the object's data.

sub get_soft_disk_limit {
        my $self = shift;
        return ( $self->{"soft-disk-limit"} );
}

## @name ZquotaEntry ::get_soft_file_limit
## @synopsis	gets the soft file limit
## @description Returns the soft file limit from the object's data.

sub get_soft_file_limit {
        my $self = shift;
        return ( $self->{"soft-file-limit"} );
}

## @name ZquotaEntry ::set_disk_limit
## @synopsis	sets the disk limit
## @description Sets the disk limit in the object's data.
## @param $_[0]  A disk limit string.

sub set_disk_limit {
        my $self = shift;
	$self->{"disk-limit"} = shift;
}

## @name ZquotaEntry ::set_file_limit
## @synopsis	sets the file limit
## @description Sets the file limit in the object's data.
## @param $_[0]  A file limit string.

sub set_file_limit {
        my $self = shift;
	$self->{"file-limit"} = shift;
}

## @name ZquotaEntry ::set_threshold
## @synopsis	sets the threshold limit
## @description Sets the threshold in the object's data.
## @param $_[0]  A threshold string.

sub set_threshold {
        my $self = shift;
	$self->{"threshold"} = shift;
}

## @name ZquotaEntry ::set_soft_disk_limit
## @synopsis	sets the soft disk limit
## @description Sets the soft disk limit in the object's data.
## @param $_[0]  A soft disk limit string.

sub set_soft_disk_limit {
        my $self = shift;
	$self->{"soft-disk-limit"} = shift;
}

## @name ZquotaEntry ::set_soft_file_limit
## @synopsis	sets the soft file limit
## @description Sets the soft file limit in the object's data.
## @param $_[0]  A soft file limit string.

sub set_soft_file_limit {
        my $self = shift;
	$self->{"soft-file-limit"} = shift;
}

## @name ZquotaEntry ::toString
## @synopsis	returns a string representation of the object
## @description returns QuotaId object string representation of the object.

sub toString {
        my $self = shift;
	my $qref = $self->{"quota_ref"};
	return $qref->toString();
}

1
