# 
#  $Id: //depot/prod/test/Rsierranevada/lib/ZquotaId.pm#1 $
# 
#  Copyright (c) 2002 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary A zapi quota id object
##
## @description
##	This file contains methods to manipulate a quota Id.  A
##	ZquotaId object is used to uniquely identify a quota.  To
##	uniquely identify the quota, the ZquotaId object contains
##	a Zapi object, the quota type, the quota target, the volume
##	and qtree the quota resides in.  The quota Id is used to
##	create QuotaEntry objects and can be returned from QuotaList
##	objects.
##
## @author rle@netapp.com

package     ZquotaId;

use	strict;
use	Zapi;
use	vars ('@ISA', '@EXPORT');
require     Exporter;
@ISA        = qw(Exporter);

## @name ZquotaId ::new
## @synopsis	creates a ZquotaId object
## @description Creates a ZquotaId object from a Zapi object, quota type,
##		quota target, volume name, and qtree name.
## @param $type    a given quota type string.
## @param $target  a given quota target string.
## @param $volume  a given volume string.
## @param $qtree   a give qtree string.

sub new {
	my $class  = shift;
	my $server = shift;   # Zapi object
	my $type   = shift;   # quota type string: user, group, tree
	my $target = shift;   # quota target string
	my $volume = shift;   # volume name string
	my $qtree   = shift;  # quota tree string

	my $self = {
		"server" => $server,
		"type"   => $type,
		"target" => $target,
		"volume" => $volume,
		"qtree"   => $qtree

	};

	bless $self, $class;
	return $self;
}

## @name ZquotaId ::toString
## @synopsis	returns a string representation of the object
## @description returns a string representation of the object.

sub toString {
        my $self = shift;
	
	return "$self->{type} $self->{target} $self->{volume}  $self->{qtree}";
}


## @name ZquotaId ::compare
## @synopsis	compares the object's to given string.
## @description compares the object's qtype, qtarget, volume, and qtree,
##		to given strings.  Returns 1 if equal, 0 if unequal.
## @param $qtype   a given quota type string.
## @param $qtarget a given quota target string.
## @param $volume  a given volume string.
## @param $qtree   a give qtree string.

sub compare {
        my $self = shift;
	my $qtype = shift;
	my $qtarget = shift;
	my $volume = shift;
	my $qtree = shift;
	
	return ( ($self->{type} eq $qtype) && ($self->{target} eq $qtarget) &&
	         ($self->{volume} eq $volume) && ($self->{qtree} eq $qtree) );
}

## @name ZquotaId ::get_quota_type
## @synopsis	returns the object's quota type
## @description returns the object's quota type string

sub get_quota_type {
        my $self = shift;
	return $self->{type};
}

## @name ZquotaId ::get_quota_target
## @synopsis	returns the object's quota type
## @description returns the object's quota type string

sub get_quota_target {
        my $self = shift;
	return $self->{target};
}

## @name ZquotaId ::get_volume
## @synopsis	returns the object's volume
## @description returns the object's volume string

sub get_volume {
        my $self = shift;
	return $self->{volume};
}

## @name ZquotaId ::get_tree
## @synopsis	returns the object's qtree
## @description returns the object's qtree string

sub get_tree {
        my $self = shift;
	return $self->{qtree};
}

## @name ZquotaId ::get_entry
## @synopsis	invokes a quota-get-entry
## @description A quota-get-entry is invoked with the object's Zapi server
##		object, quota type, quota target, volume, and qtree.  A
##		tuple of results and xml ouput is returned.  If results
##		are defined then an error occured.

sub get_entry {
        my $self = shift;

	my $server = $self->{server};
	my $xo = $server->invoke_api("quota-get-entry",
				     "quota-type",   $self->{type},
				     "quota-target", $self->{target},
				     "volume",       $self->{volume},
				     "qtree",        $self->{qtree} );
	if ( ! $xo->results_ok() ) {
		my $reason = $xo->results_reason();
		if ( ! defined($reason) ) {
			my $reason = "unknown invoke error for quota-get-entry";
		}
		return $reason, undef;
	}

	return undef, $xo;
}

## @name ZquotaId ::set_entry
## @synopsis	invokes a quota-set-entry
## @description A quota-set-entry is invoked with the object's Zapi server
##		object, quota type, quota target, volume, qtree, and
##		key-value pairs.  The results are returned. If results
##		are defined then an error occured.
## @param @_	optional key-value pairs, like "file-limit", "8000".

sub set_entry {
        my $self = shift;

	my $xo = $self->{server}->invoke_api("quota-set-entry",
			 		     "quota-type",   $self->{type},
			 		     "quota-target", $self->{target},
			 		     "volume",       $self->{volume},
			 		     "qtree",        $self->{qtree},
					     @_);
	if ( ! $xo->results_ok() ) {
		my $reason = $xo->results_reason();
		if ( ! defined($reason) ) {
			my $reason = "unknown invoke error for quota-set-entry";
		}
		return $reason;
	}

	return $xo->results_reason();
}

## @name ZquotaId ::del_entry
## @synopsis	invokes a quota-del-entry
## @description A quota-del-entry is invoked with the object's Zapi server
##		object, quota type, quota target, volume, and qtree.  The
##		results are returned. If results are defined then an error
##		occured.

sub del_entry {
        my $self = shift;

	my $xo = $self->{server}->invoke_api("quota-delete-entry",
					     "quota-type",   $self->{type},
					     "quota-target", $self->{target},
					     "volume",       $self->{volume},
					     "qtree",        $self->{qtree} );
	if ( ! $xo->results_ok() ) {
		my $reason = $xo->results_reason();
		if ( ! defined($reason) ) {
			my $reason = "unknown invoke error for quota-del-entry";
		}
		return $reason;
	}

	return $xo->results_reason();
}

## @name ZquotaId ::add_entry
## @synopsis	invokes a quota-add-entry
## @description A quota-add-entry is invoked with the object's Zapi server
##		object, quota type, quota target, volume, qtree, and
##		key-value pairs.  The results are returned. If results
##		are defined then an error occured.
## @param @_	optional key-value pairs, like "file-limit", "8000".

sub add_entry {
        my $self = shift;

	my $xo = $self->{server}->invoke_api("quota-add-entry",
			 		     "quota-type",   $self->{type},
			 		     "quota-target", $self->{target},
			 		     "volume",       $self->{volume},
			 		     "qtree",        $self->{qtree},
					     @_);
	if ( ! $xo->results_ok() ) {
		my $reason = $xo->results_reason();
		if ( ! defined($reason) ) {
			my $reason = "unknown invoke error for quota-add-entry";
		}
		return $reason;
	}

	return $xo->results_reason();
}

## @name ZquotaId ::mod_entry
## @synopsis	invokes a quota-modify-entry
## @description A quota-modify-entry is invoked with the object's Zapi server
##		object, quota type, quota target, volume, qtree, and
##		key-value pairs.  The results are returned. If results
##		are defined then an error occured.
## @param @_	optional key-value pairs, like "file-limit", "8000".

sub mod_entry {
        my $self = shift;

	my $xo = $self->{server}->invoke_api("quota-modify-entry",
			 		     "quota-type",   $self->{type},
			 		     "quota-target", $self->{target},
			 		     "volume",       $self->{volume},
			 		     "qtree",        $self->{qtree},
					     @_);
	if ( ! $xo->results_ok() ) {
		my $reason = $xo->results_reason();
		if ( ! defined($reason) ) {
			my $reason = "unknown invoke error for quota-modify-entry";
		}
		return $reason;
	}

	return $xo->results_reason();
}

1
