# 
#  $Id: //depot/prod/test/Rsierranevada/lib/ZquotaList.pm#1 $
# 
#  Copyright (c) 2002 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary A zapi quota list object
##
## @description
##	This file contains methods to manipulate a Zephyr API quota-list-entries.
##
## @author rle@netapp.com

package     ZquotaList;

use	strict;
use	NaElement;
use	ZquotaId;
use	Zapi;
use	vars ('@ISA', '@EXPORT');
require     Exporter;
@ISA        = qw(Exporter);

## @name ZquotaList ::new
## @synopsis	creates a ZquotaList object
## @description Creates a ZquotaList object from a Zapi object.
## @param	Zapi object
## @param	entry-ouput, optional.  The value can be "true" or "false, and
##		defaults to "false".  If the value is "true" then the whole
##		quota entry is included in quota-entry.

sub new {
	my $class       = shift;
	my $server      = shift;
	my $output_line = shift;

	my $xo;
	if ( defined($output_line) ) {
		if ( ($output_line ne "true") && ($output_line ne "false") ) {
			return undef;
		}
		$xo = $server->invoke("quota-list-entries",
				      "include-output-entry", "$output_line");
	}
	else {
		$xo = $server->invoke("quota-list-entries");
		$output_line = "false";
	}
	if ( ! $xo->results_ok() ) {
		# XXX - return error ???
		return undef;
	}

	my $results = $xo->results_reason();
	if ( defined($results) ) {
		# XXX - return error ???
		return undef;
	}
	
	my $qentries = $xo->child_get("quota-entries");

	my $self = {
		server        => $server,       # the Zapi object
		quota_entries => $qentries,     # the quota list XML
		output_line   => $output_line,  # true if ouput_line defined
		quota_x       => 0              # the current quota index
	};

	bless $self, $class;
	return $self;
}

## @name ZquotaList ::get_quota_entries
## @synopsis	returns the quotas info
## @description Returns the quotas info in a XML format which contains all
##		the quota entries.

sub get_quota_entries {
        my $self = shift;
        return ( $self->{quota_entries} );
}

## @name ZquotaList ::get_num_quotas
## @synopsis	returns the number of quotas in the list
## @description Returns the number of quota entries in the list.

sub get_num_quotas {
        my $self = shift;

	my $qentries = $self->{quota_entries};
	if ( ! defined($qentries) ) {
		print "*** get_num_quotas(): quota-entries undefined.\n";
		return -1;
	}

        return ( $qentries->child_max_number() );
}

## @name ZquotaList ::set_next_quota_index
## @synopsis	sets the next quota index
## @description Sets the next quota index to be used by get_next_quota().
## @param $i    the quota index.

sub set_next_quota_index {
        my $self = shift;
	my $i = shift;

	$self->{quota_x} = $i;
}

## @name ZquotaList ::get_next_quota
## @synopsis	gets the next quota
## @description Obtains the next quota as specified by quota_x and returns
##		a QuotaId object.

sub get_next_quota {
        my $self = shift;

	my $i = $self->{quota_x};
        if ( ($i >= $self->get_num_quotas() ) || ($i < 0) ) {
		print "*** get_next_quota(): Invalid index: $i.\n".
		return undef;
	}

	$self->{quota_x} += 1;
	my $qinfo = $self->{quota_entries}->child_get_nth($i);
	return ( new QuotaId( $self->{server},
                               $qinfo->child_get_string("quota-type"),
                               $qinfo->child_get_string("quota-target"),
                               $qinfo->child_get_string("volume"),
                               $qinfo->child_get_string("qtree") ) );
}

## @name ZquotaList ::get_next_quota_line
## @synopsis	gets the next quota line from an entry
## @description Returns the next quota line as specified by quota_x.

sub get_next_quota_line {
        my $self = shift;

	if ( $self->{output_line} ne "true" ) {
		print "*** quota line output not set ***\n";
		return undef;
	}
	my $i = $self->{quota_x};
        if ( ($i >= $self->get_num_quotas() ) || ($i < 0) ) {
		print "*** get_next_quota(): Invalid index: $i.\n".
		return undef;
	}

	$self->{quota_x} += 1;
	my $qinfo = $self->{quota_entries}->child_get_nth($i);
	return ( $qinfo->child_get_string("quota-line") );
}

## @name ZquotaList ::find_quota
## @synopsis	finds a quota in the list object
## @description Locates a quota in the list object and returns 1 for
##		found and 0 for not found.
## @param $qref  The QuotaId object to locate. 

sub find_quota {
	my $self = shift;
	my $qref = shift;

	my $qinfo;
	my $match = 0;
	my $i = 0;
	while ( $i < $self->get_num_quotas() ) {
		$qinfo = $self->{quota_entries}->child_get_nth($i);
		$match = $qref->compare(
                               $qinfo->child_get_string("quota-type"),
                               $qinfo->child_get_string("quota-target"),
                               $qinfo->child_get_string("volume"),
                               $qinfo->child_get_string("qtree") );
		if ( $match == 1 ) {
			last;
		}
		$i += 1;
	}

	return $match;
}

## @name ZquotaList ::find_error
## @synopsis	finds the first error in the quota list
## @description Attempts to find an error in the quota list.  If an error
##		is found, the error number, error reason and error detail
##		are returned, else undef.
## @param $qref  The QuotaId object to locate. 

sub find_error {
	my $self = shift;
	my $qref = shift;

	my $qinfo;
	my $qerror;
	my $i = 0;
	while ( $i < $self->get_num_quotas() ) {
		$qinfo = $self->{quota_entries}->child_get_nth($i);
		$qerror = $qinfo->child_get("quota-error");
		if ( defined($qerror) ) {
			return $qerror->child_get_string("errno"),
			       $qerror->child_get_string("reason"),
			       $qerror->child_get_string("detail");
		}
		$i += 1;
	}

	return undef, undef, undef;
}

## @name ZquotaList ::toString
## @synopsis	returns a string representation of the object
## @description returns a string representation of the object with each
##		quota entry separated by a \n.

sub toString {
	my $self = shift;

	my $qinfo;
	my $outstr = "";
	my $i = 0;
	while ( $i < $self->get_num_quotas() ) {
		$qinfo = $self->{quota_entries}->child_get_nth($i);
		$outstr .= "  " . $qinfo->child_get_string("quota-type") . " " .
                           $qinfo->child_get_string("quota-target") . " " .
                           $qinfo->child_get_string("volume") . " " .
                           $qinfo->child_get_string("qtree") . "\n";
		$i += 1;
	}

	return $outstr;
}

## @name ZquotaList ::toFile
## @synopsis	returns the quotas file as a collection of lines
## @description returns the quotas file as a collestion of lines from
##		the <line> tag separated by a \n.

sub toFile {
	my $self = shift;

	if ($self->{output_line} ne "true") {
		return " ";
	}

	my $qinfo;
	my $outstr = "";
	my $i = 0;
	while ( $i < $self->get_num_quotas() ) {
		$qinfo = $self->{quota_entries}->child_get_nth($i);
		$outstr .= "  " . $qinfo->child_get_string("line") . "\n";
		$i += 1;
	}

	return $outstr;
}

1
