
# 
#  $Id: //depot/prod/test/Rsierranevada/lib/ZquotaOp.pm#1 $
# 
#  Copyright (c) 2002 Network Appliance, Inc.
#  All rights reserved.
#

# -----------------------------------------------------------
##
## @summary A zapi quota operator object
##
## @description
##	This file contains methods to manipulate a quota-* Zephyr APIs
##	for the various operations like "status", "on", "off".
##
## @dependencies
##	Zapi.pm, NaElement.pm
##
## @author rle@netapp.com

package     ZquotaOp;

use	strict;
use	NaElement;
use	Zapi;
use	vars ('@ISA', '@EXPORT');
require     Exporter;
@ISA        = qw(Exporter);

## @name ZquotaOp ::new
## @synopsis	creates a ZquotaOp object
## @description Creates a ZquotaOp object from a Zapi object.
## @param	Zapi object
## @param	volume
## @param	num_checks, optional, defaults to 5 checks
## @param	sleep_time, optional, defaults to 1 second interval

sub new {
	my $class  = shift;
	my $server = shift;
	my $volume = shift;
	my $num_checks = shift;
	my $sleep_time = shift;

	my $self = {
		server     => $server,  # the Zapi object
		volume     => $volume,  # the volume
		num_checks => 120,      # number of times to check the status
		sleep_time => 2         # time in secs to sleep between checks
	};

	if ( defined($num_checks) ) {
		$self->{num_checks} = $num_checks;
	}
	if ( defined($sleep_time) ) {
		$self->{sleep_time} = $sleep_time;
	}

	bless $self, $class;
	return $self;
}

## @name ZquotaOp ::get_num_checks
## @synopsis	gets the number of times to check the status
## @description Returns the number of times that the status is checked
##		from the object's data.

sub get_num_checks {
        my $self = shift;
        return ( $self->{"num_checks"} );
}

## @name ZquotaOp ::set_num_checks
## @synopsis	sets the number of times to check the status
## @description Sets the number of times to check the status in the
##		object's data.
## @param $_[0]  The number of times to check the status.

sub set_num_checks {
        my $self = shift;
	$self->{"num_checks"} = shift;
}

## @name ZquotaOp ::get_sleep_time
## @synopsis	gets the sleep time
## @description Returns the number of seconds to sleep between checks.

sub get_sleep_time {
        my $self = shift;
        return ( $self->{"sleep_time"} );
}

## @name ZquotaOp ::set_sleep_time
## @synopsis	sets the sleep time
## @description Sets the number of seconds to sleep between checks.
## @param $_[0] The sleep time in seconds. 

sub set_sleep_time {
        my $self = shift;
	$self->{"sleep_time"} = shift;
}

## @name ZquotaOp ::get_op_status
## @synopsis	returns the quota op status
## @description Returns the quota op status.

sub get_op_status {
        my $self = shift;

	my $xo = $self->{server}->invoke( "quota-status",
					  "volume", $self->{volume} );
	if ( ! defined($xo) ) {
		return "failed: invoke for quota-status";
	}

	my $result = $xo->results_reason();
	if ( defined($result) ) {
		Tharn::tstlog( "-fail",
			"quota-status failed on volume $self->{volume}: " .
			"$result");
		return;
	}

	Tharn::tstlog( "-comment",
		"quota-status succeeded on volume $self->{volume}" );
	return $xo->child_get_string("status");
}

## @name ZquotaOp ::get_op_full_status
## @synopsis	returns the all of the quota op status
## @description Returns the results tag which includes the status
##		and substatus, and includes either reason or
##		percent-complete.

sub get_op_full_status {
        my $self = shift;

	my $xo = $self->{server}->invoke( "quota-status",
					  "volume", $self->{volume} );
	if ( ! defined($xo) ) {
		return "failed: invoke for quota-status";
	}

	my $result = $xo->results_reason();
	if ( defined($result) ) {
		Tharn::tstlog( "-fail",
			"quota-status failed on volume $self->{volume}: " .
			"$result");
		return;
	}

	Tharn::tstlog("-comment",
	       "quota-status succeeded on volume $self->{volume}");
	return $xo
}


## @name ZquotaOp ::do_quota_op
## @synopsis	do a quota op function
## @description Perform a quota op function like quota on, off, or resize.
##		The quota status must be checked to verify that function
##		succeeded.  Returns the reason if an error occurs, else
##		undef for success.
## @param $_[0]  the operation to preform

sub do_quota_op {
        my $self = shift;
	my $op   = shift;

	my $xo = $self->{server}->invoke( $op, "volume", $self->{volume} );
	if ( ! $xo->results_ok() ) {
		return "failed: invoke for $op";
	}

	my $result = $xo->results_reason();
	if ( defined($result) ) {
		Tharn::tstlog("-fail", "$op failed on vol $self->{volume}: $result");
		return;
	}

	return undef;
}

## @name ZquotaOp ::turn_quotas_on
## @synopsis	turns quotas on
## @description Sets quotas on and checks the status.  Since turning
##		quotas on is an asynchronous process, this subroutine
##		sets the quotas on and then checks the status.
## 		Returns the results tag which includes the status
##		and substatus, and includes either reason or
##		percent-complete.

sub turn_quotas_on {
        my $self = shift;

	my $result = $self->do_quota_op("quota-on");
	if ( defined($result) ) {
		return;  # failure already printed
	}

	my $qstatus = "";
	my $i = 0;
	while ( $i < $self->{num_checks} ) {
		$qstatus = $self->get_op_status();
		if ( $qstatus eq "on" ) {
			last;
		}
		sleep( $self->{sleep_time} );
		$i += 1;
	}

	if ( $i >= $self->{num_checks} ) {
		return $self->get_op_full_status();
	}

	return undef;
}

## @name ZquotaOp ::turn_quotas_off
## @synopsis	turns quotas off
## @description Sets quotas off and checks the status.  Since turning
##		quotas off is an asynchronous process, this subroutine
##		sets the quotas off and then checks the status.
## 		Returns the results tag which includes the status
##		and substatus, and includes either reason or
##		percent-complete.

sub turn_quotas_off {
        my $self = shift;

	my $result = $self->do_quota_op("quota-off");
	if ( defined($result) ) {
		return;  # failure already printed
	}

	my $qstatus = "";
	my $i = 0;
	while ( $i < $self->{num_checks} ) {
		$qstatus = $self->get_op_status();
		if ( $qstatus eq "off" ) {
			last;
		}
		sleep( $self->{sleep_time} );
		$i += 1;
	}

	if ( $i >= $self->{num_checks} ) {
		return $self->get_op_full_status();
	}

	return undef;
}

1
