#-*-perl-*-

## @summary Routines for issuing aggr commands to the Filer
##
## @description
##      The aggr.pm functions are wrappers around NATE's say() function that
##      send a 'aggr xxx' command to the filer and return 1 for success or
##      undef for failure.  In the case of failure, an error string
##      is also returned in $EVAL_ERROR ($@)
##
## @synopsis
##      use aggr;
##      $success = aggr_add( &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_copy ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_create( &nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = create_aggregate ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_destroy( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_join( &nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_mirror( &nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_offline( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_online( &nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_options ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_rename( &nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_restrict( $fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_scrub ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_split( &nbsp;&nbsp;&nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_status ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $success = aggr_verify ( &nbsp;$fullconnspec, $aggr_name [, <i>optional_params</i>] );
##      $opts = get_raid_create_opts ( &nbsp;$fullconnspec, [, <i>optional_params</i>] );
##
## @usage
##      <tt>$success = <i>aggrfunction</i>( $fullconnspec, $aggr_name [, options => ""] [, timeout => -1]);</tt>
##
##      where <i><tt>aggrfunction</tt></i></b> is one of the following:
##      <pre><ul>
##      aggr_add
##      aggr_copy
##      aggr_create
##      create_aggregate
##      aggr_destroy
##      aggr_mirror
##      aggr_offline
##      aggr_online
##      aggr_options
##      aggr_rename
##      aggr_restrict
##      aggr_scrub
##      aggr_split
##      aggr_status
##      aggr_verify
##      get_raid_create_opts
##      </ul></pre>
##
##      <tt>$fullconnspec</tt> identifies the host and/or connection to be
##      used to submit the appropriate 'aggr xxx' command.  It can be a host
##      object, connection object, hostspec, connspec or hostspec#connspec. 
##      If a connection is not specified, #rsh is used.
##
##      <tt>$aggr_name</tt> is the name of the aggregate on which <i>aggrfunction</i>
##      will be performed.
##
##      The <tt><i>optional_params</i></tt> argument is a string containing
##      any desired options for the particular aggr command.  It is null by
##      default.  If supplied, it must conform to that aggr command's syntax.
##
##      The <tt>timeout</tt> argument specifies the maximum amount of time,
##      in seconds, for the function to return.  If that maximum is
##      exceeded, a timeout error is returned.  The default timeout is -1
##      (infinite).
##
## @examples
##      To create a aggr with two disks:
##      <blockquote>
##      use aggr;
##      my $h = host($FILER);
##      aggr_create($h, "testaggr", options=>2, timeout=>300)
##      &nbsp;&nbsp;&nbsp; or logresult('SCRIPT', $@);
##      # aggr_create succeeded: continue with test
##      </blockquote>
## 
## @status deprecated, use Aggregate.pm instead
##
## @keywords nebuchadnezzar, vvol
##
## @author dl-johny@netapp.com (FSS WAFL Quality)

package aggr;

use strict;
use Module::Common;
our ($Validation, $VERSION);

BEGIN
{
    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @aggr_subs = qw(
            aggr_add 
            aggr_copy
            aggr_create
            create_aggregate
            aggr_destroy
            aggr_join 
            aggr_mirror 
            aggr_offline
            aggr_online 
            aggr_options
            aggr_rename
            aggr_restrict 
            aggr_scrub 
            aggr_split 
            aggr_status
            aggr_verify
            aggr_purge
            aggr_wait_all
            aggr_destroy_prefix
            create_prefix
            aggr_exists
            wait_for_aggr
            get_raid_create_opts
            get_vols_of_aggr
            aggr_compare
    );

    Module::Common::setup_export_tags(
        aggr => [ @aggr_subs ]
    );
}

use Tharn;

sub wait_for_aggr ( $$$ );
sub get_raid_create_opts ( $% );

use Params::Validate qw(validate_pos SCALAR);
use Utils qw(make_conn make_conn_and_complain);
use SayWrap qw(sayEOF);
use RaidInfo;
use vvol qw(vvol_purge vvol_offline vvol_container vvol_status vvol_restrict);
use fileSystem qw(treecompare_vol);

sub aggr_options ($$%);
sub get_vols_of_aggr($$);

# Tab width - 4
# no tabs - spaces only
# connection type - rsh only
# dont use logresult/tstlog ... to exit... use die with eval
# functions have to return 1 on success 
# add local subroutines at the end.




# ---------------------------------
# Exported sub-routines begin here
# ---------------------------------

## @name ::aggr_create()

## @summary create an aggregate

## @synopsis
##   aggr_create($spec, $aggr [,options=>] [,timeout=>])

## @notes To create aggreates of various types use create_aggregate. 
##        create_aggregate has an option to use an existing aggregate. 
##        It also to creates aggregates of diff checksum style, 
##        raid type and mirror.
##        Reusing exisiting aggregates if possible will save
##        time when disks are not zeroed.
##        When multiple tests are executed in serial 
##        they could use the same aggregate (if possible)
##        and save on creation time.

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to be sent down to aggr create
## @arg timeout Optional: time in sec to wait for the aggr creation 
##              to complete
##              Default: uses default of wait_for_aggr 1800 (sec)

## @see-also lib/aggr.pm::create_aggregate

## @status public

sub aggr_create ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {
    my $after = say_cmd($spec,'create',$aggr,$args{timeout},$args{options})
        or die($@);

    if ($after =~ /has\s+been\s+initiated/ || $after =~ /has\s+completed/) {
        return wait_for_aggr($spec,$aggr,$args{timeout});
    } else {
        die $after;
    }
    };
}


## @name ::create_aggregate()

## @description creates an aggregate.
##       Basically a wrapper around aggr_create.
##       If the volume / aggregate is already present and if purge_existing 
##       is '1' then the exisiting volume / aggregate will be purged and
##       a new one created. The volume / aggregate creation heavily relies
##       on the additional args disks, raid_mirror, cksum_style and raid_type.
##       So depending on the type of volume / aggregate asked for, the correct
##       spares are chosen and the volume / aggregate created.

## @synopsis create_aggregate($spec, $aggr, 
##                       [, purge_existing=>0]
##                       [, create_opts=>'']
##                       [, disks=>2]
##                       [, raid_mirror=>'normal']
##                       [, cksum_style=>'']
##                       [, raid_type=>'raid4']);

## @arg spec  Required: Hostspec of filer
## @arg aggr  Required: Name of aggregate to be created
## @arg purge_existing Optional: Flag to denote if existing vol / aggr
##            must be purged and a new one created in its place.
##            Default: 0
## @arg create_opts Optional: Options to create the aggregate.
##            If provided then the args disks, cksum_style,
##            raid_type, raid_mirror will be ignored.
##            These options will be directly passed to aggr::aggr_create
##            or vvol::vvol_create. No error checking will be done.
##            Default: ''
## @arg disks Optional: Number of disks to use for volume / aggregate creation
##            Default: 2
## @arg raid_mirror Optional: To create a normal or mirrored volume / aggregate.
##            Possible values: 'normal' / 'mirrored'
##            Default: 'normal'
## @arg cksum_style Optional: Style of checksum to use for volume / aggregate.
##            Possible values: 'zoned' / 'block' / ''
##            When '' then ontap will decide.
##            Default: ''
## @arg raid_type Optional: Type of raid group to use for volume / aggregate creation.
##            Possible values: 'raid4' / 'raid_dp'
##            Default: 'raid4'
## @arg timeout Optional: Time (in sec) to wait for aggregate to be created.
##            Default: 1000 (sec)

## @returns 1 if the aggregate was created by this routine
##          0 if aggregate was not created in this routine
##          will die with $@ if the routine could not complete for some reason

## @dependencies Need equal number of disks in each plex for mirrored volume / aggregate. 
##               We assume number of plexes is 2, so we need even number 
##               of disks to make a mirrored volume / aggregate
## @dependencies Need a min of 2 disks to make a raid4 volume / aggregate
## @dependencies Need a min of 3 disks to make a raid_dp volume / aggregate

## @notes does not take into consideration the size of the spares.
## @notes does not take into consideration the % of zeroing of the spares.
## @notes virtual volume is a wafl object and is not created by this routine

## @status testing

sub create_aggregate ( $$% ) {

    my ($spec, $aggr, %args) = @_;


    eval {
    my $fc = make_conn($spec);

    # set the defaults for the optional arguments
    # TBD possibly could add more error handling for the optional args
    my $purge_existing = defined($args{purge_existing}) ? $args{purge_existing} : 0;
    my $disks = defined($args{disks}) ? $args{disks} : 2;
    my $timeout = defined($args{timeout}) ? $args{timeout} : 1000;
    my $create_opts = defined($args{create_opts}) ? $args{create_opts} : '';
    my $cksum_style = defined($args{cksum_style}) ? $args{cksum_style} : '';
    my $raid_mirror = (defined($args{raid_mirror}) && ($args{raid_mirror} eq 'mirrored')) 
                        ?  $args{raid_mirror} : 'normal';
    my $raid_type = (defined($args{raid_type}) && ($args{raid_type} eq 'raid_dp')) 
                        ? $args{raid_type} : 'raid4';

    logcomment("Will create aggregate:$aggr disks:$disks cksum_style:$cksum_style". 
                " raid_mirror:$raid_mirror raid_type:$raid_type");

    # Check if volume / aggregate already exists
    my $exists = aggr_exists($fc, $aggr, timeout=>100);

    if ($exists) {
        if ($purge_existing) { 
            aggr_purge($fc, $aggr) or
                die("Purge of aggregate \'$aggr\' failed :$@\n");
            $exists=0;
        } else {
            # check if aggregate is online
            my %status = aggr_status($spec,$aggr,timeout=>$args{timeout})
                or die($@);
            die("Aggregate '$aggr' exists but is not online. Will not be able to use it.")
                if (!defined($status{State}) || $status{State} !~ /online/);

            logcomment("DBG: Will use existing aggregate \'$aggr\'");
        }
    }

    my $we_created = 0;
    if (!$exists) {
        if (!$create_opts) {
            # get raid create options
            $create_opts = get_raid_create_opts($spec, 
                            disks=>$args{disks},
                            cksum_style=>$args{cksum_style},
                            raid_type=>$args{raid_type},
                            raid_mirror=>$args{raid_mirror},
                            timeout=>$timeout);
            die($@) if ($@);
        }

        # Create an aggregate
        tstlog('-commline', "DBG: Aggregate create options are $create_opts");
        $we_created = aggr_create($fc, $aggr, timeout=>$timeout, options=>$create_opts) or
            die("Aggregate $aggr could not be created :$@");

    }
    return($we_created);

    };
}


## @name ::aggr_rename()

## @summary rename an aggregrate

## @synopsis
##   aggr_rename($spec, $aggr [,options=>] [,timeout=>] )

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: specify newname that the aggr is to be renamed to
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @example aggr_rename('tada-u22', 'aggr1', options=>'aggr2');

## @status public

sub aggr_rename( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {
    my $after = say_cmd($spec,'rename',$aggr,$args{timeout},$args{options})
        or die($@);

    my $expected = "'".$aggr."' renamed to '".$args{options}."'";
    if ($after =~ /$expected/) {
        # double check to see if new aggr is really present
        my %status = aggr_status($spec,$args{options},timeout=>$args{timeout})
            or die($@);
        # reached here so aggr was found
        return(1);
    } else {
        die("Expected $expected. Got $after.");
    }
    };
}



## @name ::aggr_split()

## @summary splits a mirrored aggregate

## @synopsis
##   aggr_split($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options that are sent to aggr split
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status in-progress

sub aggr_split ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {
    if ($aggr =~ /[^\/]/){
        #Assuming a plex is not given, but only the aggr name which is to be 
        #split, so get a random plex of the aggr to split
        $aggr = get_plex_of_aggr($spec,$aggr);
    }
    my $after = say_cmd($spec,'split',$aggr,$args{timeout},$args{options});

    if ($after =~ /new\s+aggregate.*\s+created/g) {
        return 1;
    } else {
        die $after;
    }
    };
}

## @name ::aggr_destroy()

## @summary destroys an aggregate.

## @notes Caller must include '-f' in 
## options in order to suppress the prompt
## "are you sure ?", which this routine does not handle

## @synopsis
##   aggr_destroy($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to aggr destroy command
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status public

sub aggr_destroy ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {
    my $after = say_cmd($spec,'destroy',$aggr,$args{timeout},$args{options});
    if ($after =~ /destroyed/){
        return 1;
    } else {
        die $after;
    }
    };
}

## @name ::aggr_online()

## @summary makes an aggregate online

## @synopsis
##   aggr_online($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of aggregrate/plex
##              Format : aggrname or /<aggrname>/<plexname>
## @arg options Optional: options to aggr online command
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @example aggr_online('tada-u22', 'aggr1');
## @example aggr_online('tada-u22', '/aggr1/plex1');

# usage is ... aggr online { aggrname | plexname } [ -f ]

# plex can be specified.... /<aggrname>/<plexname> or <aggrname>/<plexname>
# get_plex_info() expects the format /<aggrname>/<plexname>
# ontap does not care and accepts either
# but the messages from the command are of format <aggrname>/<plexname>
# so add hacks to make this work

## @status public

# TBD Add a flag to online all volumes in aggr if required by user


sub aggr_online ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {
    my $after = say_cmd($spec,'online',$aggr,$args{timeout},$args{options})
        or die($@);

    # remove beginning "/" if present in $aggr for matching
    # if "/" not present add it for get_plex_info()...
    my $tmp = ($aggr =~ /^\//) ? substr($aggr,1,(length($aggr)-1)) : $aggr;
    my $expected = "(Plex|Aggregate) \'$tmp\' is (now|already) online";

    if ($after =~ /$expected/) {
        my $type = $1;
        if ($type =~ /Plex/) {
            my $p = get_plex_info($spec, $tmp)
                or die("Could not get info for plex $tmp");

            return(($p->is_online) ? 1 : 0);
        } else {
            my %status = aggr_status($spec,$aggr,timeout=>$args{timeout})
                or die($@);
            if ($status{State} =~ /online/) {
                return(1);
            } else {
                die("State of $aggr is not online but ".$status{State});
            }
        }
    } else {
        die("Expected $expected. Got $after.");
    }
    };
}

## @name ::aggr_offline()

## @summary makes an aggregate offline
## First offlines all volumes in aggr
## then makes aggr offline

## @synopsis
##   aggr_offline($spec, $aggr [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status public

sub aggr_offline ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {

    # get all volumes on aggr
    # offline all of them then offline the aggr itself
    my (@vols) = get_vols_of_aggr($spec, $aggr);
    die if ($@);
    if (@vols) {
        foreach my $vol (@vols) {
            my %status = vvol_status($spec,$vol,timeout=>$args{timeout})
                or die($@);
            vvol_offline($spec, $vol)
                if (defined($status{State}) && $status{State} =~ /online/);
            die if ($@);
        }
    }

    # now offline the aggr itself
    my $after = say_cmd($spec,'offline',$aggr,$args{timeout})
        or die($@);

    # remove beginning "/" if present in $aggr for matching
    # if "/" not present add it for get_plex_info()...
    my $tmp = ($aggr =~ /^\//) ? substr($aggr,1,(length($aggr)-1)) : $aggr;
    my $expected = "(Plex|Aggregate) \'$tmp\' is (now|already) offline";

    if ($after =~ /$expected/) {
        my $type = $1;
        if ($type =~ /Plex/) {
            my $p = get_plex_info($spec, $aggr )
                or die("Could not get info for plex $aggr");

            return(($p->is_online) ? 0 : 1);
        } else {
            my %status = aggr_status($spec,$aggr,timeout=>$args{timeout})
                or die($@);
            if ($status{State} =~ /offline/) {
                return(1); 
            } else {
                die("State of $aggr is not offline but ".$status{State});
            }
        }
    } else {
        die("Expected $expected. Got $after.");
    }
    };
}

## @name ::aggr_restrict()

## @summary makes an aggregate restricted
## Will first restrict all volumes in aggregate,
## then restrict the aggregate itself.

## @synopsis
##   aggr_restrict($spec, $aggr [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status public

sub aggr_restrict ( $$% ) {

    my ($spec,$aggr,%args) = @_;

    return eval {

    # get all volumes on aggr
    # restrict all of them then restrict the aggr itself
    my (@vols) = get_vols_of_aggr($spec, $aggr);
    die if ($@);
    if (@vols) {
        foreach my $vol (@vols) {
            my %status = vvol_status($spec,$vol,timeout=>$args{timeout})
                or die($@);
            vvol_restrict($spec, $vol)
                if (defined($status{State}) && $status{State} =~ /online/);
            die if ($@);
        }
    }

    my $after = say_cmd($spec,'restrict',$aggr,$args{timeout});

    my $expected = "Aggregate $aggr is now restricted";
    if ($after =~ /$expected/) {
        my %status = aggr_status($spec,$aggr,timeout=>$args{timeout});
        if ($status{State} =~ /restricted/){
            return(1);
        } else {
            die("State of $aggr is not restricted but ".$status{State});
        }
    } else {
        die("Expected $expected. Got $after.");
    }
    };
}


## @name ::aggr_mirror()

## @summary mirrors an aggregate with disk list 
## specifed in options

## @synopsis
##   aggr_mirror($spec, $aggr [,options=>''] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to be sent down to aggr mirror
## @arg timeout Optional: time in sec to wait for aggr mirror to complete
##              Default: uses default of wait_for_aggr 1800 (sec)

## @status public

sub aggr_mirror ($$%) {
    my ($spec,$aggr,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'mirror',$aggr,$args{timeout},$args{options});
    if ($after =~ /has\s+completed/ || $after =~ /has\s+been\s+initiated/ || !length($after)) {
        return wait_for_aggr($spec,$aggr,$args{timeout});
    } elsif($args{options} =~ /-n/){
        return($after);
    }
    else{
        die $after;
    }
    };
}


## @name ::aggr_join()

## @summary Joins the plexes of a split mirrored aggr 

## @synopsis
##   aggr_join($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to be sent down to aggr mirror
## @arg timeout Optional: time in sec to wait for aggr mirror to complete
##              Default: uses default of wait_for_aggr 1800 (sec)

## @notes Specify the timeout argument when using this 
## routine, to break out of the while loop, if aggr join fails.

## @status in-progress

sub aggr_join ($$%) {
    my ($spec,$aggr,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'mirror',$aggr,$args{timeout},$args{options});
    if ($after =~ /Invalid\s+operation/ || $after =~ /join/) {
        logcomment ("Cannot join the plexes");
        return $after;
    } else {
        return wait_for_aggr($spec,$aggr,$args{timeout});
        }
    };
}

## @name ::aggr_scrub()

## @summary Start,suspend,resume or stop scrubbing, as specified in options.

## @synopsis
##   aggr_scrub($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to command aggr scrub
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status in-progress

sub aggr_scrub ($$%) {
    my ($spec,$aggr,%args) = @_;
    return eval {
    if ($args{options} eq 'start') {
    my $after = say_cmd($spec,'scrub',$aggr,$args{timeout},$args{options},1);
        if ($after =~ /aggr\s+scrub:\s+started\s+scrub\s+of\s+\/$aggr*/g) {
            return 1;
        } else {
            die  $after;
        }
    }elsif($args{options} eq 'stop') {
    my $after = say_cmd($spec,'scrub',$aggr,$args{timeout},$args{options},1);
        if ($after =~ /aggr\s+scrub:\s+stopped\s+scrub\s+of\s+\/$aggr*/g) {
            return 1;
        } else {
            die "$after";
        }
    }elsif($args{options} eq 'suspend') {
    my $after = say_cmd ($spec,'scrub',$aggr,$args{timeout},$args{options},1);
        if ($after =~ /aggr\s+scrub:\s+suspended\s+scrub\s+of\s+\/$aggr*/g) {
            return 1;
        }else {
            die $after;
        }
    }elsif($args{options} eq 'resume') {
    my $after = say_cmd($spec,'scrub',$aggr,$args{timeout},$args{options},1);
        if ($after =~ /aggr\s+scrub:\s+resumed\s+scrub\s+of\s+\/$aggr*/g) {
            return 1;
        } else {
            die "$after";
        }
    }
    };
}

## @name ::aggr_add()

## @summary adds disks specified in options to an aggregate

## @synopsis
##   aggr_add($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: options to be sent down to aggr add
## @arg timeout Optional: time in sec to wait for aggr add to complete
##              Default: uses default of wait_for_aggr 1800 (sec)

## @status in-progress

sub aggr_add ($$%) {
    my ($spec,$aggr,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'add',$aggr,$args{timeout},$args{options});
    if ($after =~ /has\s+been\s+initiated/ || $after =~ /has\s+completed/) {
        return wait_for_aggr($spec,$aggr,$args{timeout});
    } else {
        die $after;
    }
    };
}

## @name ::aggr_options()

## @summary returns the value of an option after 
## setting it to the specified value

## @synopsis
##   aggr_options($spec, $aggr, %args)

## @arg spec Required: hostspec of the filer
## @arg aggr Required: name of existing aggregrate
## @arg args Optional: additional arguments like
##          timeout=>
##          options=>
##
## @notes Syntax of this command is
##      aggr options aggrname [ optname optval ]
## Allowed optname and optvals are
##      dsize number
##      aggr_size_fixed on | off
##      nosnap on | off
##      snapmirrored off
##      ignore_inconsistent on | off
##      resyncsnaptime number
##
## @returns If successful returns @options

## @status untested

#aggr_size_fixed=off
#nosnap=off
#snapmirrored=off,
#resyncsnaptime=60

#raidtype=raid4
#raidsize=8
# root, diskroot, nosnap=off, raidtype=raid4, raidsize=8,
#snapmirrored=off, resyncsnaptime=60, aggr_size_fixed=off

sub aggr_options ( $$% ) {
    my ($spec,$aggr,%args) = @_;

    # this command has 2 behaviours
    # 1. Get all the options / values for the aggregate
    # 2. Set an options to value for the aggregate
    eval {
    if ($args{options}) {

        # reached here because caller wants to SET a particular option
        # dont know what extent of error handling is to be done here
        my $out = say_cmd($spec,'options',$aggr,$args{timeout},$args{options});

        die($out) if ($out !~ /is not a legal aggregate option/i);

        # TBD 
        # this part can become very very complex.
        # verifying if each set worked or not 
        # Also set section is not working too well, refer burt95052

        # verify if the set was successful
        my ($optname, $optval) = ($args{options} =~ /^\s*(\S*)\s*(\S*)/);
        $optval='on'  if ($optval =~ /on|1|true/i);
        $optval='off' if ($optval =~ /off|0|false/i);
        
        my (@options) = aggr_options($spec,$aggr,$args{timeout}) or
            die("Could not get aggr options for $aggr : $@");

        my $found=0;
        for my $one (@options) {
            for my $name (keys %$one) {
                print "\nNAME ", $name, "\nVALUE ",  $one->{$name};
                if ($name eq $optname) {
                    $found=1;
                    die("Tried to set option $optname to $optval. ".
                        "But its set to ". $one->{$name}) 
                        if ($one->{$name} ne $optval);
                    last;
                }
            }
        }
        die("Could not find option $optname for $aggr") if (!$found);

    } else {
        # Reached here because caller wants to get/check all aggr options
        my $after = say_cmd($spec,'options',$aggr,$args{timeout});
        my $expected = '((root|'.
                            'diskroot|'.
                            'nosnap=(on|off)|'.
                            'raid_type=\w+|'.
                            'raid_size=\d+|'.
                            'snapmirrored=(on|off)'.
                            'resyncsnaptime=\d+'.
                            'aggr_size_fixed=(on|off)'.
                        '),?\s?[\n|\r]?)+';

        # TBD
        # Now do we have to test if the options seen are correct or not
        # e.g root aggr / vol has extra options
        # offline vol has lesser options
        # there can be many many more instances

        if ($after !~ /$expected/) {
            die("Expected \'$expected\'. Got \'$after\'");
        } else {
            my @options;
            foreach my $line (split(',',$after)) {
                $line =~ /(\w+)=?(\w+)?/;
                my $one = {};
                $one->{$1} = $2;
                push(@options,$one);
            }
            return(@options);
        }
    }
    }; 
}


## @name ::aggr_status()

## @summary gets aggregate status

## @synopsis
##   aggr_status($spec, $aggr, %args)

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection
## @arg args    Optional: additional arguments like
##              options=> (only '-b' option is implemented for now)
##              timeout=>
##                       
## @example
##    ## %status contains hashkeys 'Aggr', 'State', and 'Status' only
##    my %status = aggr_status($filer, $aggr);
##    tstlog('-comment', "$status{Aggr}: $status{State}: $status{Status}");
##
##    ## %status contains hashkeys 'block_size', 'aggr_size', 'fs_size' only
##    my %status = aggr_status($filer, $aggr, options=>'-b');
##
## @returns %status (options will determine what hashkeys will be populated)
## @status deprecated, Use Aggregate::aggr_status instead

sub aggr_status ($$%) {
    my ($spec,$aggr,%args) = @_;

    logresult('warn', "aggr::aggr_status is deprecated. Migrate to Aggregate::aggr_status");

    return eval {
    my $after = say_cmd($spec,'status',$aggr,$args{timeout},$args{options});

    if ($after =~ /^\s*aggr\s+status:\s+No\s+aggregate\s+named\s+\'$aggr\'\s+exists/) {
        die($after);
    }

    # '-b' option to show the size of the aggregate in blocks
    if (defined($args{options}) &&
        ($args{options} =~ /\s*\-b\s*/)) {

        my %status_size;

        # filer> aggr status vol0 -b
        # Aggr  Block Size (bytes)  Aggr Size (blocks)  FS Size (blocks)
        # ----  ------------------  ------------------  ----------------
        # vol0  4096                256000              256000   

        foreach my $line (split(/\n/, $after)) {
	    if ($line =~ /^\s*$aggr\s+(\d+)\s+(\d+)\s+(\d+)/) {
                $status_size{block_size} = $1;
                $status_size{aggr_size}  = $2;
                $status_size{fs_size}    = $3;
                return %status_size;
            }
        }
        
        die("Unable to parse text for \'$args{options}\': $after");

    } elsif ($after !~ /aggr\s+status:/) {
        my %status;
        my ($line) = ($after =~ /($aggr\s+.*?.+\S)/);
        $line =~ /\s*($aggr)\s+(\w+)\s+((\w,?\s?)+)\s*/;
        ($status{Aggr},$status{State},$status{Status}) = ($1, $2, $3); 
        if ($status{Status} =~ /aggr\b/) {
            $status{Type} = 'aggr';
        } elsif ($status{Status} =~ /trad\b/) {
            $status{Type} = 'trad';
        } else {
            die "aggr::aggr_status(): Could not obtain type 'aggr' or 'trad'";
        }
        return(%status);
    } else {
        die($after);
    }
    };
}

## @name ::aggr_verify()

## @summary Start, suspend, resume, stop, or get aggr verification status, 
##          as specified in options.

## @synopsis
##   aggr_verify($spec, $aggr [,options=>] [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg options Optional: subcommand to aggr verify
##              Possible values: start, suspend, resume, stop, status
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status incomplete. see burt 95044. will only work if $aggr is empty.

sub aggr_verify ($$%) {
    my ($spec, $aggr, %args) = @_;

    return eval {
    my $option = $args{options};
    my $after = say_cmd($spec, "verify", $aggr, $args{timeout}, $args{options}, 1)
        or die($@);

    my $regexp;

    if ($option eq "start") {
        $regexp = 'aggr\s+verify\s+start\:\s+verify\s+started';
        if ($after =~ /$regexp/) {
            return 1;
        }

    } elsif ($option eq "suspend") {
        $regexp = 'aggr\s+verify\s+suspend\:\s+verify\s+suspended';
        if ($after =~ /$regexp/) {
            return 1;
        }
    
    } elsif ($option eq "resume") {
        $regexp  = 'aggr\s+verify\s+resume\:\s+verify\s+resumed';
        if ($after =~ /$regexp/) {
            return 1;
        }

    } elsif ($option eq "stop") {
        $regexp  = 'aggr\s+verify\s+stop\:\s+verify\s+stopped';
        if ($after =~ /$regexp/) {
            return 1;
        }
    
    } elsif ($option eq "status") {
        if ($aggr) {
            $regexp = 'aggr\s+verify\s+' . $aggr . '\s+is\s+[\d]+\%\s+complete';
        } else {
            $regexp = 'aggr\s+verify\s+[\w]+\s+is\s+[\d]+\%\s+complete';
        }

        my $err = 0;
        foreach my $line (split(/\n/, $after)) {
            if ($line !~ /$regexp/) {
                $err = 1;
                last;
            }
        }

        if (!$err) {
            return 1;
        }

    } else {
        die "Unknown option: $option";
    }

    die "Expected $regexp. Got $after.";

    }; # return eval { ...
}


## @name ::aggr_purge()

## @summary Purge aggregate and all its contents.
## Will not fail if $aggr does not exist.

## @synopsis
##   aggr_purge($spec, $aggr [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status public

# @notes TBD Have to add checking for cloned volumes
# since clone cannot be destroyed while its parent is offline
# and parent cannot be destroyed while clone is alive

sub aggr_purge ( $$% ) {
    my ($spec, $aggr, %args) = @_;

    return eval {
    my $ri = new RaidInfo($spec);
    my $aggregate = $ri->aggregate($aggr); 

    if ($aggregate) {
        # offline all volumes in aggr if any
        my  %volumes = $aggregate->all_volumes();
        if (%volumes) {
            foreach my $v (keys(%volumes)) {
		# final frontier requires flexible vols to be offlined *and*
		# destroyed prior to destroying an aggregate.
                vvol_purge($spec,$v) or die($@);
            }
        } 
        # offline aggregate
        aggr_offline($spec,$aggr,timeout=>$args{timeout})
            or die($@);
        # destroy aggregate
        aggr_destroy($spec,$aggr,timeout=>$args{timeout},options=>'-f')
            or die($@);
    } else {
        print("\nAggregate $aggr does not exist");
    }
    return(1);
    };
}


## @name ::wait_for_aggr()

## @synopsis wait_for_aggr($spec, $aggr, $timeout);

## @description loop, calling 'aggr status $aggr' until the state
##              and status match $state and $status.

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for aggr state and status match

## @status in-progress

sub wait_for_aggr ( $$$ ) {

    my ($spec, $aggr, $timeout) = @_;

    my $endtime = timeout2time( ($timeout) ? $timeout : 1800);
    my $sleeptime = 1;

    while (time() < $endtime) {
        my @complaint;
        my $conn=make_conn($spec);
        my (undef,undef,$after) = say ($conn,
                    "priv set -q diag; raid_config info liststatus",
                    '-timeout',$timeout, -swallow,
                    '-eval', sub { @complaint = @_;1; });
        if (@complaint) {
            die join('\n',@complaint);
        }

        my (@info) = split (/\r\n/, $after);
        my ($normal) = 1;

        foreach my $line (@info) {
            
        if ($line =~ /,name=$aggr,/) {
        if ( ($line =~ /,initializing=1,/) || ($line =~ /,invalid=1,/)
             || ($line =~ /,growing=1,/ )||($line =~ /,mdegraded=1,/)||($line =~/,resyncing=1,/) ) {
            $normal = 0;
            last;
        }
        
        last;
            }
        }

        if ($normal) {
            return 1;
        } else {
            sleep $sleeptime;
            $sleeptime = ($sleeptime > 60) ? 60 : $sleeptime * 2;
        }
    }
    die "Timed out waiting for aggr '$aggr' to become (online,normal)\n";
}


## @name ::aggr_wait_all()

## @summary wait for fsm_state of all aggregates to become NORMAL

## @synopsis
##   aggr_wait_all($spec, [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg timeout Optional: time in sec to wait for aggrs to become normal
##              Default: 1800

## @status public


sub aggr_wait_all ( $% ) {

    my ($spec,%args) = @_;
    my ($sleeptime) = 1;
    my ($done) = 0;

    my $endtime = timeout2time( ($args{timeout}) ? $args{timeout} : 1800);
    while ( !$done && (time() < $endtime) ) {

        my ($ri) = new RaidInfo($spec);
        my %aggrs =  $ri->all_aggregates();

        $done = 1;
        foreach my $name (keys %aggrs) {
            my $aggr = $ri->aggregate($name);

            if ($aggr->parameter("fsm_state") !~ /NORMAL/) {
                $sleeptime = ($sleeptime > 60) ? 60 : ($sleeptime * 2);
                sleep($sleeptime);
                $done = 0;

                last;
            }

        } # foreach-loop

    } # while-loop
}

## @name ::aggr_destroy_prefix()

## @summary destroys all aggregates which begin with the specified prefix

## @synopsis
##   aggr_destroy_prefix($spec, $aggr [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg prefix  Required: prefix of existing aggregrate names
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of aggr_wait_all - 1800 (sec)

## @status public


sub aggr_destroy_prefix ( $$% ) {

    my ($spec, $prefix, %args) = @_;

    aggr_wait_all($spec, %args);
    my ($ri) = new RaidInfo($spec);
    my %aggrs =  $ri->all_aggregates();
    my %vols =  $ri->all_volumes();
    foreach my $a (keys %aggrs) {
        my @comppaint;
        my $after;
        if ($a =~ /^$prefix/) {
            foreach my $v (keys %vols) {
                my $hostaggr = vvol_container($spec, $v, %args);
                if ($hostaggr eq $a) {
                    vvol_offline($spec, $v);
                }
            }
            aggr_offline($spec, $a, %args);
            aggr_destroy($spec, $a, options=>"-f", %args);
        }
    }
}

# @name ::create_prefix()

# @summary creates a unique prefix for aggregate names

sub create_prefix( $ ) {
    my ($type) = @_;

    return "vvolreg_${type}";
}

## @name ::aggr_exists()

## @summary return 1 if specified aggregate does not exists, 0 otherwise

## @synopsis
##   aggr_exists($spec, $aggr [,timeout=>])

## @arg spec    Required: hostspec of the filer
## @arg aggr    Required: name of existing aggregrate
## @arg timeout Optional: time in sec to wait for command to complete
##              Default: uses default value of the connection

## @status public

sub aggr_exists( $$% ) {
    my ($spec, $aggr, %args) = @_;
    my $conn = make_conn($spec);

    my $after = say_cmd($conn,'status',$aggr,$args{timeout});
    if ($after =~ /aggr status: No aggregate named '$aggr' exists/) {
        return 0;
    }
    return 1;
}

## @name ::get_raid_create_opts()

## @description This routine will query the filer and decide the
##           create options string that can be used to create
##           the desired raid object (paired volume / aggregate)

## @synopsis get_raid_create_opts($spec,
##                       [, disks=>2]
##                       [, raid_mirror=>'normal']
##                       [, cksum_style=>'']
##                       [, raid_type=>'raid4']);

## @arg spec  Required: Hostspec of filer
## @arg disks Optional: Number of disks to use for volume / aggregate creation
##            Default: 2
## @arg raid_mirror Optional: To create a normal or mirrored volume / aggregate.
##            Possible values: 'normal' / 'mirrored'
##            Default: 'normal'
## @arg cksum_style Optional: Style of checksum to use for volume / aggregate.
##            Possible values: 'zoned' / 'block' / ''
##            When '' then ontap will decide.
##            Default: ''
## @arg raid_type Optional: Type of raid group to use for volume / aggregate creation.
##            Possible values: 'raid4' / 'raid_dp'
##            Default: 'raid4'

## @returns a string of the create options that can be used to create the vol/aggr
##          will die with $@ if the routine could not complete for some reason

## @dependencies Need equal number of disks in each plex for mirrored volume / aggregate. 
##               We assume number of plexes is 2, so we need even number 
##               of disks to make a mirrored volume / aggregate
## @dependencies Need a min of 2 disks to make a raid4 volume / aggregate
## @dependencies Need a min of 3 disks to make a raid_dp volume / aggregate

## @notes does not take into consideration the size of the spares.
## @notes does not take into consideration the % of zeroing of the spares.
## @notes virtual volume is a wafl object and is not created by this routine

## @status testing

sub get_raid_create_opts ( $% ) {

    my ($spec, %args) = @_;

    eval {
    my $fc = make_conn($spec);

    # set the defaults for the optional arguments
    my $disks = defined($args{disks}) ? $args{disks} : 2;
    my $cksum_style = defined($args{cksum_style}) ? $args{cksum_style} : '';
    my $raid_mirror = (defined($args{raid_mirror}) && ($args{raid_mirror} eq 'mirrored')) 
                        ?  $args{raid_mirror} : 'normal';
    my $raid_type = (defined($args{raid_type}) && ($args{raid_type} eq 'raid_dp')) 
                        ? $args{raid_type} : 'raid4';

    tstlog('-commline', "DBG: Will generate string to create raid object ". 
                            "disks:$disks cksum_style:$cksum_style". 
                            " raid_mirror:$raid_mirror raid_type:$raid_type");

    #-------------------------
    # do inital error checking
    #-------------------------
    # calculate minumum disks that are required for the volume / aggregate
    my $num_plex = 2;  # for now ontap supports only 2 plexes
    my $min_disks = 2; # min number of disks to create a basic aggr is 2
    $min_disks += 1 if ($raid_type eq 'raid_dp');
    $min_disks *= $num_plex if ($raid_mirror eq 'mirrored');

    die("Need minimum of $min_disks disks to make this raid object\n")
        if ($disks < $min_disks);

    # exit if asked for mirrored volume / aggregate but # disks is odd
    die("Need even number of disks ". 
        "\(you asked for $disks disks\) to make a mirrored raid object\n")
        if (($raid_mirror eq 'mirrored') && ($disks % 2));

    my $ri = new RaidInfo($fc) or 
        die("Unable to get raid info\n");

    # check if number of spares are available
    my %spare_disks =  $ri->spare_disks();
    die("Not enough spare disks. Require $disks in spare pool.\n")
        if (scalar(keys(%spare_disks)) < $disks);

    # check if syncmirror is setup if aggr to be created is to be mirrored
    my (%spare_disks_pool0, %spare_disks_pool1);
    my $disks_pool;
    if ($raid_mirror eq 'mirrored') {
        # check if syncmirror is licensed
        my (undef, undef, $out) = say($fc, "license");
            die("Raid mirror (syncmirror_local) is not licensed.\n".
                "Pls license, reboot and re-run this test\n")
                if ($out =~ /syncmirror_local\s+not licensed/);

        # check if there are 2 pools on the filer
        %spare_disks_pool0 = $ri->spare_disks_pool0();
        %spare_disks_pool1 = $ri->spare_disks_pool1();
        die("Disks are not in 2 pools.\nUnable to created a mirrored raid object\n")
            unless (scalar(keys(%spare_disks_pool1)) > 0);

        # check if there are enough spares in both pools
        $disks_pool = $disks / 2;
        die("Not enough \(require $disks_pool\) disks in both spare pools.\n")
            if ((scalar(keys(%spare_disks_pool0)) < $disks_pool) ||
                (scalar(keys(%spare_disks_pool1)) < $disks_pool));
    }

    #-------------------------
    # start assembling the create options for the volume / aggregate
    #-------------------------
    my $create_opts = "";
    $create_opts .= "-m " if ($raid_mirror eq 'mirrored');
    $create_opts .= "-t $raid_type ";

    #-------------------------
    # branch based on the cksum_style
    #-------------------------
    if ($cksum_style eq '') {

        # let ontap take care and pick the disks
        $create_opts .= "$disks";

    } else {

        # Check if we have required checksum style disks; else exit
        # For 'zoned' all disks (512/520 bps) can be configured in zoned checkum style.
        # whereas 'block' checksum style requires disks of 520bps.

        # get a list of 520 and 512 bps disks
        my (@bps_512_spares, @bps_520_spares,
            @bps_512_spares_pool0, @bps_520_spares_pool0,
            @bps_512_spares_pool1, @bps_520_spares_pool1);

        @bps_512_spares = get_matching_sub_list ( 'bps', '512', %spare_disks );
        @bps_520_spares = get_matching_sub_list ( 'bps', '520', %spare_disks );
        tstlog('-commline',"DBG: 512bps spares are @bps_512_spares");
        tstlog('-commline',"DBG: 520bps spares are @bps_520_spares");

        if ($raid_mirror eq 'mirrored') {
            @bps_512_spares_pool0 = get_matching_sub_list ( 'bps', '512', %spare_disks_pool0 );
            @bps_512_spares_pool1 = get_matching_sub_list ( 'bps', '512', %spare_disks_pool1 );
            @bps_520_spares_pool0 = get_matching_sub_list ( 'bps', '520', %spare_disks_pool0 );
            @bps_520_spares_pool1 = get_matching_sub_list ( 'bps', '520', %spare_disks_pool1 );
            tstlog('-commline',"DBG: 512bps spares in pool0 are @bps_512_spares_pool0");
            tstlog('-commline',"DBG: 512bps spares in pool1 are @bps_512_spares_pool1");
            tstlog('-commline',"DBG: 520bps spares in pool0 are @bps_520_spares_pool0");
            tstlog('-commline',"DBG: 520bps spares in pool1 are @bps_520_spares_pool1");
        }

        if ($cksum_style eq 'block') {

            if (scalar(@bps_512_spares)) {
                # randomly pick required number of disks from the list 
                # here we take control and select the disks to be used
                if ($raid_mirror eq 'mirrored') {

                    # TBD we could use -f option to acheive this
                    die("Not enough \(require $disks_pool\) 520 bps in ".
                        "each spare pool to create $cksum_style raid object\n")
                        if ((scalar(@bps_520_spares_pool0) < $disks_pool) || 
                            (scalar(@bps_520_spares_pool1) < $disks_pool));

                    # could look for zeroed disks
                    my @chosen_disks_pool0 = get_rand_sublist($disks_pool, @bps_520_spares_pool0);
                    my @chosen_disks_pool1 = get_rand_sublist($disks_pool, @bps_520_spares_pool1);
                    $create_opts .= "-d @chosen_disks_pool0 -d @chosen_disks_pool1";

                } else {
                    die("Not enough \(require - $disks\) spares of 520 ". 
                        "bps to create $cksum_style raid object\n")
                        if (scalar(@bps_520_spares) < $disks);
                    my @chosen_disks = get_rand_sublist($disks, @bps_520_spares);
                    $create_opts .= "-d @chosen_disks";
                }
            } else {
                # there are no 512 bps spare disks on the filer so just go ahead and
                # create the vol / aggr with $disks and ontap will do the disk selection
                $create_opts .= "$disks";
            }

        } elsif ($cksum_style eq 'zoned') {

            if (scalar(@bps_520_spares)) {
                if ($raid_mirror eq 'mirrored') {
                    die("No spares \(require $disks_pool\) ". 
                        "of 512 bps in both spare pools to create ". 
                        "$cksum_style raid object\n")
                        unless ((@bps_512_spares_pool1) && (@bps_512_spares_pool0));

                    if (scalar(@bps_512_spares_pool0) < $disks_pool) {
                        # not enough disks in 512 pool0 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares pool0 list
                        if (scalar(@bps_520_spares_pool0) > (scalar(@bps_512_spares_pool0) - $disks_pool)) {
                            tstlog('-commline', "DBG: Did not get enough spares 512 bps in pool0. ". 
                                                "\nWill look in the 520 bps pool0 spare list");
                            my @additional_spares = get_rand_sublist(scalar(@bps_512_spares_pool0) - $disks_pool, 
                                                                    @bps_520_spares_pool0);
                            push(@bps_512_spares_pool0, @additional_spares);
                        } else {
                            die("Not enough \(require $disks_pool\) spares", 
                                        " of required bps in pool0 to". 
                                        " create $cksum_style raid object\n");
                        }
                    }
                    if (scalar(@bps_512_spares_pool1) < $disks_pool) {
                        # not enough disks in 512 pool0 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares pool1 list
                        if (scalar(@bps_520_spares_pool1) > 
                            (scalar(@bps_512_spares_pool1) - $disks_pool)) {
                            tstlog('-commline', "DBG: Did not get enough ". 
                                        "spares 512 bps in pool1. \nWill look ". 
                                        "in the 520 bps pool1 spare list");
                            my @additional_spares = get_rand_sublist(scalar(@bps_512_spares_pool1) 
                                                        - $disks_pool, @bps_520_spares_pool1);
                            push(@bps_512_spares_pool1, @additional_spares);
                        } else {
                            die("Not enough \(require $disks_pool\) spares", 
                                        " of required bps in pool1 to". 
                                        " create $cksum_style raid object\n");
                        }
                    }
                    my @chosen_disks_pool0 = get_rand_sublist($disks_pool, @bps_512_spares_pool0);
                    my @chosen_disks_pool1 = get_rand_sublist($disks_pool, @bps_512_spares_pool1);
                    $create_opts .= "-d @chosen_disks_pool0 -d @chosen_disks_pool1";

                } else {

                    die("No spares \(require $disks\) of 512 bps ". 
                        " to create $cksum_style raid object\n")
                        unless (@bps_512_spares);

                    if (scalar(@bps_512_spares) < $disks) {
                        # not enough disks in 512 to make the zoned checksum vol/aggr
                        # we can see if there are some in the bps_520_spares list
                        # because the 520 when combined with 512 for an vol/aggr
                        # behaves as a zoned checksum style
                        if (scalar(@bps_520_spares) > (scalar(@bps_512_spares) - $disks)) {
                            tstlog('-commline', "DBG: Did not get enough ". 
                                        "spares in 512 bps. ". 
                                        "Will look in the 520 bps list");
                            my @additional_spares = get_rand_sublist(
                                                        scalar(@bps_512_spares) - $disks, 
                                                        @bps_520_spares);
                            push(@bps_512_spares, @additional_spares);
                        } else {
                            die("Not enough \(require $disks\) spares ".
                                "of required bps to create ".
                                "$cksum_style raid object\n");
                        }
                    }
                    # randomly pick required number of disks from the list 
                    # here we take control and select the disks to be used
                    my @chosen_disks = get_rand_sublist($disks, @bps_512_spares);
                    $create_opts .= "-d @chosen_disks";
                }

            } else {
                # there are no 520 bps disks on the filer so just go ahead and
                # create the vol/aggr with $disks and ontap will do disk selection
                $create_opts .= "$disks";
            }
        }
    }

    return($create_opts);

    };
}


## @name ::aggr_compare

## @synopsis Compare two aggregates

## @arg $conn Required: Connection object to filer
## @arg $aggr1 Required: Name of first aggregate
## @arg $aggr2 Required: Name of second aggregate
## @arg $conn2 Optional: Connection object to second filer
##             Default: $conn

## @notes if aggregates are empty routine will pass
## @notes used to compare aggregates after Aggregate snapmirror
## @notes snapmirror must be quiesced and broken off before
##        calling this routine - else will fail
## @notes routine depends heavily on names of volumes
##        to decide which volumes to compare. So renaming them might
##        result in the routine failing

sub aggr_compare ( $$$% ) {

    my $conn = make_conn_and_complain(shift);
    my ($aggr1, $aggr2, %args) = validate_pos(
        @_,
        [ {type=>SCALAR}, {type=>SCALAR} ],
        {
            conn2 => { isa=>'Connectrec', optional=>1 }
        }
    );

    my $conn2 = ($args{conn2}) ? make_conn_and_complain($args{conn2}) : $conn;

	return eval {

        # get vols for both aggrs
        # check for existance of aggregate is done in get_vols_of_aggr
        my (@vols1) = get_vols_of_aggr($conn,$aggr1);
        die if ($@);
        my (@vols2) = get_vols_of_aggr($conn2,$aggr2);
        die if ($@);
        tstlog('-commline', "DBG: Vols of aggregate \'$aggr1\' are @vols1");
        tstlog('-commline', "DBG: Vols of aggregate \'$aggr2\' are @vols2");

        # die if the num of vols dont match
        die("Aggregates have different number of volumes.".
            "\n$aggr1 has ".scalar(@vols1)." - @vols1".
            "\n$aggr2 has ".scalar(@vols2)." - @vols2\n")
            if (scalar(@vols1) != scalar(@vols2));

        # return if the aggrs are empty
        if (!@vols1 && !@vols2) {
            print("No volumes to compare. Aggregates \'$aggr1\' and \'$aggr2\' are empty");
            return(1);
        }

        # pair up the volumes to compare them
        # treecompare each vol of first aggregate with corresponding vol on
        # second aggregate
        foreach my $vol (@vols1) {
            my @mirror_vol = grep(/$vol\(?\d*\)?/,@vols2);
            if (@mirror_vol) {
                print("\nCompare $aggr1:$vol with $aggr2:$mirror_vol[0]");
                my $filer = $conn->hostp->id();
                my $filer_dst = $conn2->hostp->id();
                tstlog('-commline', "DBG: FILER SRC is $filer, FILER DST is $filer_dst");
                my $response = treecompare_vol($filer, $vol,
                                            $mirror_vol[0],
                                            filer_dst=>$filer_dst,
                                            timeout=>600);
                if (!$response) {
                    die("Mismatch in contents of $aggr1:$vol and $aggr2:@mirror_vol :$@");
                }
            } else {
                die("Volume \'$vol\' on aggregate \'$aggr1\' does not exist in aggregate \'$aggr2\'");
            }
        }

        # aggregates match if it reached here
        return(1);
    };
    
}


## @name ::get_vols_of_aggr

## @description Will get all the volumes of the aggregate
## Volumes that are offline, restricted will also 
## be returned
## if aggr is not online (i.e offline / restricted)  
## ONTAP will not be able to list the volumes in the aggregate.

## @synopsis get_vols_of_aggr($spec, $aggr);

## @arg $spec Required: Connection obj to filer
## @arg $aggr Required: Aggregate whose volumes are required.

## @returns an array containing list of volumes


sub get_vols_of_aggr($$) {

    my $conn = make_conn_and_complain(shift);
    my ($aggr) = validate_pos( @_, 1 );

	return eval {

        # we could use aggr status $aggr to get the same info
        my $ri = new RaidInfo($conn);
        my $aggregate = $ri->aggregate($aggr); 
        die("Aggregate $aggr does not exist\n") if (!$aggregate);

        my %vols =  $aggregate->all_volumes();
        my @vols = keys(%vols);

        return(@vols);

	};
}


# -----------------------------
# Local sub-routines begin here
# -----------------------------

# @name ::get_plex_info()

# @summary gets info about a plex

# @synopsis
#   get_plex_info($spec, $plex)

# @arg spec Required: hostspec of the filer
# @arg plex Required: name of plex
#           format is "/<aggrname>/<plexname>"

 
sub get_plex_info ( $$ ) {

    my ($spec,$plex) = @_;
    return eval {
    my $tmp_plex = '/';
    if ($plex !~ /^\//) {
        substr($tmp_plex,1) = $plex;
    } else {
        $tmp_plex=$plex;
    }
    my $conn = make_conn($spec);
    my $ri = new RaidInfo($conn);
    my %plexs =  $ri->all_plexes();
    my $p = $ri->plex($tmp_plex);

    if ($p) {
        return($p);
    } else {
        die("$tmp_plex not found");
    }
    };
}

# @name ::get_plex_of_aggr()

# @summary gets a random plex of an aggregate

# @synopsis
#   get_plex_of_aggr($spec, $aggr)

# @arg spec Required: hostspec of the filer
# @arg $aggr Required: name of aggr for which you need the plex
# return value /aggr

sub get_plex_of_aggr( $$ ) {

    my ($spec,$aggr) = @_;
    my $conn = make_conn($spec);
    my $ri = new RaidInfo($conn);
    my %aggregates =  $ri->all_aggregates();
    my %plexes = $aggregates{$aggr}->all_plexes();
    my @plex_list = (keys %plexes);
    return($plex_list[0]);
}

# @name ::get_rand_sublist

# Randomly selects $num unique elements from @list
# TBD Improve to pick the ones that are already zeroed

sub get_rand_sublist ( $@ ) {

    my ($num, @list) = @_;

    eval {
    my @sublist;
    die ("Not enough elements in original list.\n")
        if (scalar(@list) < $num);

    return(@list) if (scalar(@list) == $num);

    while (($num) && (@list))  {
        my $index = int(rand(@list));
        my $element = $list[$index];
        tstlog('-commline', "DBG: Adding $element to sub list");
        push(@sublist, $element);
        $num--;
        splice(@list,$index,1);
    }
    return(@sublist);
    };
}

# @name ::get_matching_sub_list

sub get_matching_sub_list ( $$% ) {

    my ($item, $match, %hash) = @_;

    my @new_list;
    my @keys = keys(%hash);
    foreach my $index (@keys) {
        push(@new_list, $index) if ($hash{$index}{$item} eq $match);
    }
    return (@new_list);
}


# @name ::say_cmd()

# wrapper to Tharn::say() with more stuff

sub say_cmd ( $$$$;$$ ) {

    my ($spec,$subcmd,$aggr,$timeout,$options,$flag) = @_;

    $options = $options ? $options : "";
    $flag = $flag ? $flag : "";

    my $conn = make_conn($spec);
    $timeout = $timeout ? $timeout : $conn->{timeout};
    my $saytime = timeout2time($timeout);
    my $cmd;
    if ($flag) {
        $cmd = "aggr $subcmd $options $aggr";
    } else {
        $cmd = "aggr $subcmd $aggr $options";
    }
    
    for(my $t = 0; $t < 6; ++$t) {
	my @complaint;
	my (undef,undef,$after) = say ($conn,"$cmd",
                    '-timeout',$saytime,
                    '-eval', sub { @complaint = @_;1; });

	if ($after =~ /in use by other operations/) {
		sleep 10;
		next;
	}

	if (@complaint) {
        	die join('\n',@complaint);
	} else {
		return($after);
	}
    }
}

1;
