## @summary Routines to boot/halt/etc. filers/caches
##
## @description
##    This is a library to boot, halt, enter maintenance mode, etc.,
##    with various kinds of appliances (filers or caches) and
##    appliance simulators.
##
## @synopsis
##    use boot;
##
##    boot_state($a, ...options...);
##    transit_to($a, $to, ...options...);
##    boot($a, ...options...);
##    halt($a, ...options...);
##    reboot_to($a, $to, ...options...);
##    reboot($a, ...options...);
##    netboot_setup($a,$localip,$serverip,$filename,...options...);
##
## @examples
##
##   Boot an appliance up if it is not up already.  If this fails, it's a script error.
##
##   my $h = host($FILER);
##   boot($h) or tstlog('-script',$@);
##
##   Zero disks from ontap booted using devalias "myalias".  We do not
##   know what state it is already in, so cycle power if necessary:
##
##   my $h = host($FILER);
##   my ($res, $state) = reboot_to($h,ZERO_DISKS,bootimage=&gt;'myalias',powercycle=&gt;1);
##
## @usage
##
## <HR><center><b>Appliance Hosts and hosttypes and bootimages</b></center>
##
## Appliances come in approximately ten flavors, five for filers and
## five for netcaches.  When this library boots an appliance it uses
## the value of the field "hosttype" in that appliance's host record
## to figure out which flavor to expect, and the field "bootimage" to
## describe any parameters needed for booting.  A blank bootimage
## generally means to try to boot the default way, otherwise the value
## depends on the flavor of host.
##
## Here are the ten flavors:
##
## Filers or caches with Open FirmWare (models F2xx, F3xx, F5xx, F6xx,
## F7xx, F8xx, FAS960, and so on).  "hosttype" should be <b>filer</b> or
## <b>cache</b>, and "bootimage" is a string appropriate as an
## argument to OFW's "boot" command.
##
## Filer or cache simulators.  The hosttype should be
## <b>filer-simulator</b> or <b>cache-simulator</b> and "bootimage"
## must be the simulator executable itself and any arguments it should
## be passed.  Bootimage cannot be blank in this case.
##
## A filer or cache with a BIOS which uses GRUB for booting (F87,
## C1xxx, C2xxx).  The hosttype should be <b>filer-grub</b> or
## <b>cache-grub</b> and bootimage can be blank (meaning to try to let
## the host boot on its own), or a number (to always choose that menu
## item when presented with a grub boot menu), or a string of the form
## "(xxx)yyy" (which means to run GRUB commands "root (xxx)" then
## "kernel (xxx)yyy" then "boot"), or anything else (which means to
## interpret $bootimage as a newline-separated list of GRUB commands
## to run).
##
## A filer or cache with a BIOS which uses the the "netboot" floppy
## (F85 or C1100.  See //depot/tools/netboot/... and
## http://web.netapp.com/engineering/projects/chupa/project/calypso/netboot/).
## The netboot floppy comes in both interactive and non-interactive
## forms.  The hosttype should be <b>filer-netboot</b> or
## <b>cache-netboot</b>.  The bootimage only matters for the
## interactive version, where it's blank or "H" for a hard disk boot,
## and "N,<i>local_ip</i>,<i>server-ip</i>,<i>file</i>" for a network
## boot.  For the noninteractive version, you can set up BOOTP/DHCP to
## provide such information (and whatever bootimage is passed to these
## library routines will have no effect).
##
## A filer or cache with CFE firmware (FAS2xx, FAS30x0) or 
## Boot Loader firmware (Excelsior, so far).  The hosttype should
## be <b>filer-cfe</b> or <b>cache-cfe</b>.  The bootimage will
## be the sequence of commands to enter at the CFE prompt in order to
## boot, separated by semicolons.  If bootimage is left unspecified,
## the command used if a CFE prompt appears will be "boot" with no
## arguments.
##
## <HR><center><b>Constants for Boot States</b></center>
##
## boot exports the following constants to describe the various
## overall states that an appliance can seem to be in (at least all the
## ones that this library can deal with)
## 
## Some states are listed as <i>transient</i>.  These states are
## events which are important for routines monitoring booting to
## notice, but probably are not of interest to a caller who just wants
## to wait for the appliance to finish ending up in whatever state it will
## end up in.  (see function "boot_state", below)
##
## <b>UP</b>:
## (logged in on the console the console, i.e. at a normal, takeover,
## or partner mode prompt)
##
## <b>FIRMWARE</b>:
## (at the 'ok' prompt in the case of Open FirmWare, at the unix
## prompt in the case of a simulator, or at the fake 'ok' prompt
## that occurs after halting in the case of BIOS-based appliances)
##
## <b>MINI_MONITOR</b>:
## (this is an intermediate state for filers with Boot loader firmware.
## The mini-monitor is loaded when the system is halted after which
## the boot loader is loaded.)
##
## <b>HALTED</b>:
## (this is an intermediate state for clusterd ontap.  It prints "The
## operating system has halted.  Please press any key to reboot." when
## the system is halted (not when rebooted), after which the boot
## loader is loaded.)
##
## <b>BOOT_MENU</b>:
## (the "Selection (1-5)?" prompt, i.e. the "floppy boot menu" or the
## "special boot menu")
##
## <b>NORC</b>:
## (booted into "no rc mode", i.e. option "2" from the BOOT_MENU)
##
## <b>ZERO_DISKS</b>:
## (booted into zeroing disks mode, i.e. the state just after option
## "4" from the BOOT_MENU and answering yes to both of the "are you
## sure?" questions.  This state is transient.)
##
## <b>ZERO_DISKS_QUERY</b>:
## (asking one of the "are you sure?" questions before zeroing disks)
##
## <b>MAINT</b>:
## (maintainance mode, "5" from the BOOT_MENU)
##
## <b>MAINT_RELEASE_DISKS_QUERY</b>:
## (confirmation prompt from "storage release disks" in maintainance mode)
##
## <b>NONSTD_BOOT_QUERY</b>:
## (the "continue with boot?" question that one is asked before
## entering maintenance mode on a clustered filer)
##
## <b>TAKEN_OVER_WAITING_OFW</b>:
## The "Waiting for giveback" prompt, issued in open firmware when
## trying to boot from a disk when that disk has a reservation
## asserted by the partner.
##
## <b>TAKEN_OVER_WAITING_ONTAP</b>:
## The "Waiting for giveback" prompt, issued by ontap when ontap has
## already started but some disks have reservations asserted by the
## partner.
##
## <b>TAKEN_OVER_QUERY</b>:
## A "Do you wish to halt this node rather than wait [y/n]?" prompt,
## issued when this node was previously taken over and partner is
## detected as alive.
##
## <b>TAKEN_OVER_DEAD</b>:
## A "YOUR FILESYSTEMS MAY BE DESTROYED ... do you wish to continue
## [y/n]?" prompt which asks if you wish to take back the filesystem
## when the partner has taken over but seems to be dead.
##
## <b>TAKEN_OVER_TRANSIENT</b>:
## A "HALT: cluster partner has taken over" message which appears when
## the partner takes over unexpectedly while this node is already up.
##
## <b>FLOPPY_BOOT</b>:
## (the "floppy_boot?" prompt to which one answers yes, on simulators
## and maybe still on Open FirmWare appliances when env var floppy-boot?
## is set to true, to get the BOOT_MENU)
##
## <b>USERNAME</b>:
## (the login: prompt that an appliance may present when booting is
## complete but before the UP prompt may be reached)
##
## <b>PASSWORD</b>:
## (the Password: prompt that an appliance may present when booting is
## complete but before the UP prompt may be reached)
##
## <b>GRUB_MENU</b>:
## (the GRand Unified Bootloader main menu)
##
## <b>GRUB_CMD</b>:
## (the GRand Unified Bootloader "grub&gt;" prompt)
##
## <b>NETBOOT_METHOD</b>:
## (prompt from the interactive netboot floppy for whether to boot
## from [h]arddisk or [n]etwork)
##
## <b>NETBOOT_LOCAL_IP</b>
## (prompt from the interactive netboot floppy for local IP address)
##
## <b>NETBOOT_SERVER_IP</b>
## (prompt from the interactive netboot floppy for tftp server's IP address)
##
## <b>NETBOOT_FILE</b>
## (prompt from the interactive netboot floppy for what file to ask for via tftp)
##
## <b>NETBOOT_QUERY</b>
## (prompt from the interactive netboot floppy for whether the local
## IP, server IP, and file values entered were correct)
##
## <b>CTRL_C</b>:
## (the "Press CTRL-C for special boot menu" or equivalent message which
## appears as ontap is booted on anything other than a 7G simulator.
## This state is transient)
##
## <b>OFW_BANNER</b>:
## (the "Open Firmware by FirmWorks" message which appears as Open
## FirmWare appliances reset.  This state is transient)
##
## <b>CFE_BANNER</b>:
## (the "CFE version x.x.x based on Broadcom CFE" message which
## appears as CFE applicances reset.  This state is transient)
##
## <b>LOADER_BANNER</b>:
## (the "Boot Loader version x.x" message which appears as 
## boot loader applicances reset.  This state is transient)
##
## <b>LOADER_ABORT_AUTOBOOT</b>:
## (the "Starting AUTOBOOT press Ctrl-C to abort" message which
## appears as boot loader applicances prepare to boot.  This state is
## transient)
##
## <b>SIMLOADER_BANNER</b>:
## (the "Hit [Enter] to boot immediately, or any other key for command
## prompt." message which appears as a vsim simulator resets.  This
## state is transient)
##
## <b>NVRAM_FW_UPDATE</b>:
## (the "Firmware version 5 has been successfully downloaded on to the
## NVRAM5 card. The system will automatically reboot now" message
## which may appear during boot.  This state is transient.  It exists
## to help the boot library detect that a subsequent lapse into
## firmware is not an error)
##
## <b>DIAG_MAIN_PROMPT</b>
## main menu prompt presented right after booting diagnostic
## from this prompt, 'exit' is entered to get to FIRMWARE
## prompt when halting.
##
## <b>DIAG_SUB_PROMPT</b>
## diags sub menu prompt presented after each successive
## diags cmds are entered from DIAG_MAIN_PROMPT. from this
## prompt '99' is entered to successively return to
## previous DIAG_SUB_PROMPT, if any, until eventually
## DIAG_MAIN_PROMPT is reached from which halt can  get
## the system into FIRMWARE prompt.
##
## <b>RESTORE_SYSTEM_CONFIG_QUERY</b>
## A "Are you sure you want to continue?" question before replacing
## all boot-card-based configuration with the last disk backup.
##
## @author dl-central-libs
## @status public

# The overall design here is that boot_state() can interpret INPUT
# from an appliance to know what state it's in, and next_message() knows
# what to OUTPUT to go to the next state, and transit_to() uses these
# to forge a path to the desired state.  (and then there are a bunch
# of miscellaneous wrappers around transit_to).

package boot;

use strict;

our @BOOTSTATES;
our %States;

BEGIN {
    # For each state, the value is 1 if it's a state where ontap is
    # running and 0 if it is a state where ontap is not (pre-booted, or
    # after halted, etc.)
    %States = (
         'UP'                        => { is_ontap=>1 , is_transient=>0 },
         'FIRMWARE'                  => { is_ontap=>0 , is_transient=>0 },
         'BOOT_MENU'                 => { is_ontap=>1 , is_transient=>0 },
         'NORC'                      => { is_ontap=>1 , is_transient=>0 },
         'ZERO_DISKS'                => { is_ontap=>1 , is_transient=>1 },
         'ZERO_DISKS_QUERY'          => { is_ontap=>1 , is_transient=>0 },
         'MAINT'                     => { is_ontap=>1 , is_transient=>0 },
         'MAINT_RELEASE_DISKS_QUERY' => { is_ontap=>1 , is_transient=>0 },
         'NONSTD_BOOT_QUERY'         => { is_ontap=>1 , is_transient=>0 },
         'TAKEN_OVER_WAITING_OFW'    => { is_ontap=>0 , is_transient=>0 },
         'TAKEN_OVER_WAITING_ONTAP'  => { is_ontap=>1 , is_transient=>0 },
         'TAKEN_OVER_QUERY'          => { is_ontap=>1 , is_transient=>0 },
         'TAKEN_OVER_DEAD'           => { is_ontap=>1 , is_transient=>0 },
         'TAKEN_OVER_TRANSIENT'      => { is_ontap=>1 , is_transient=>1 },
         'FLOPPY_BOOT'               => { is_ontap=>1 , is_transient=>0 },
         'USERNAME'                  => { is_ontap=>1 , is_transient=>0 },
         'PASSWORD'                  => { is_ontap=>1 , is_transient=>0 },
         'GRUB_MENU'                 => { is_ontap=>0 , is_transient=>1 },
         'GRUB_CMD'                  => { is_ontap=>0 , is_transient=>0 },
         'NETBOOT_METHOD'            => { is_ontap=>0 , is_transient=>0 },
         'NETBOOT_LOCAL_IP'          => { is_ontap=>0 , is_transient=>0 },
         'NETBOOT_SERVER_IP'         => { is_ontap=>0 , is_transient=>0 },
         'NETBOOT_FILE'              => { is_ontap=>0 , is_transient=>0 },
         'NETBOOT_QUERY'             => { is_ontap=>0 , is_transient=>0 },
         'CTRL_C'                    => { is_ontap=>1 , is_transient=>1 },
         'OFW_BANNER'                => { is_ontap=>0 , is_transient=>1 },
         'CFE_BANNER'                => { is_ontap=>0 , is_transient=>1 },
         'LOADER_BANNER'             => { is_ontap=>0 , is_transient=>1 },
         'LOADER_ABORT_AUTOBOOT'     => { is_ontap=>0 , is_transient=>1 },
         'SIMLOADER_BANNER'          => { is_ontap=>0 , is_transient=>1 },
         'NVRAM_FW_UPDATE'           => { is_ontap=>1 , is_transient=>1 },
         'MINI_MONITOR'              => { is_ontap=>1 , is_transient=>0 },
         'HALTED'                    => { is_ontap=>1 , is_transient=>0 },
         'DIAG_MAIN_PROMPT'          => { is_ontap=>1 , is_transient=>0 },
         'DIAG_SUB_PROMPT'           => { is_ontap=>1 , is_transient=>0 },
         'RESTORE_SYSTEM_CONFIG_QUERY'  => { is_ontap=>1 , is_transient=>0 },
    );

    @BOOTSTATES = keys %States;
    # define a function for each state which just returns that state
    # as a string.  (perl is smart enough to optimize constant
    # functions into constant values)
    foreach my $state (@BOOTSTATES) {
        eval "sub $state { \"$state\" }";
    }

    use Exporter qw(import);
    our @EXPORT = qw(
        boot_state transit_to
        boot halt reboot_to reboot
        netboot_setup wait_for_rsh
    );
    push @EXPORT, @BOOTSTATES;
}

use Tharn;
use NATE::Util qw(derefarray);
use Tie::IxHash ();

# ----------------------------------------------------------------------

# convert $spec (which might be a hostspec or connspec) to a
# connection object (to console, if not specified)

sub make_conn ( $ ) {
    my ($spec) = @_;
    if (ref($spec) =~ /Connectrec/) {
        return $spec;
    } elsif (ref($spec) =~ /Hostrec/) {
        return connect($spec->{id} . "#console");
    } else {
        return connect($spec);
    }
}

# convert $spec (which might be a hostspec or connspec) to a
# host object

sub make_host ( $ ) {
    my ($spec) = @_;
    if (ref($spec) =~ /Connectrec/) {
        return $spec->{hostp};
    } else {
        return host($spec);
    }
}


# C1100/C1200/C2100 hack: for "netboot" and "grub" hosts, convert a
# regular expression that would match prompts sent to serial port by a
# sane BIOS to one which also works when such prompts are sent by a
# C1100/C1200/C2100.  (This used to only occur for "netboot" hosts,
# but it appears that more recent GRUB setups on our appliances make
# use of BIOS serial console support instead of GRUB's own)

# The BIOS sends vt100 escape sequences to update from the old screen
# contents to the new screen contents, putting in a cursor jump
# command whenever old and new screen match (it's always a jump to an
# absolute cursor address, and it occurs even if only one character
# matches.  sheesh).  In this routine, we assume that all newly
# printed lines replace blank lines (whenever not scrolling) or that
# adjacent lines will be identical only where they both have blanks
# (whenever scrolling), and so the vt100 escape sequences will always
# be inserted to jump over spaces and never over non-spaces, so we
# replace all spaces in the regular expression with expressions which
# can also match an absolute cursor position jump sequence (escape
# open-bracket digits semicolon digits "f" )

sub netboot_regex {
    my ($re) = @_;
    $re =~ s/ /(?: |\e\\[\\d+;\\d+f)/g;
    return $re;
}

# parse a bootimage (with the syntax used for filer-netboot
# or cache-netboot hosts) and return the requested field.
#
# The bootimage syntax is
#   blank or "H" for hard disk boots
#   "N,local-ip,server-ip,file" for network boots


sub netboot_bootimage {
    my ($bootimage,$field) = @_;
    my %hash;
    if (!$bootimage) {
        $hash{'method'}="H";
    } else {
        ($hash{'method'},$hash{'local_ip'},$hash{'server_ip'}, $hash{'file'})
            = split(/,/,$bootimage);
    }
    if (!defined($hash{$field})) {
        return eval { die "field \"$field\" could not be determined from bootimage \"$bootimage\"\n" };
    } else {
        return $hash{$field} . "\r";
    }
}


# return the regexen that boot_state should look for, mapped to what
# each regex means.

sub make_prompt_hash ( $$ ) {
    my ($host,$transients) = @_;
    my $hosttype = $host->{hosttype};
    my %hash;
    my $ixhash = tie( %hash,"Tie::IxHash" );

    # burt 557355: "Node/NG/Transit.pm and/or lib/boot.pm detect
    # "LOADER>" prompt as a normal rather than firmware prompt"
    # Use a Tie::IxHash and add the firmware prompts first, so that
    # they will be matched first.

    my @fwprompts;
    if ($hosttype =~ /vsim/) {
        @fwprompts = ('^SIMLOADER> ', '^VLOADER> ');
    } elsif ($hosttype =~ /vsa/) {
        @fwprompts = ('^VSALOADER> ', '^VLOADER> ');
    } elsif ($hosttype =~ /cfe|ng/) {
        @fwprompts = ('^CFE> ', '^LOADER(?:-\\w+)?\\*?> ');
        $hash{"^ok"} = MINI_MONITOR;
    } elsif ($hosttype =~ /simulator/) {
        @fwprompts = derefarray($host->prompt_fw());
    } else {
        @fwprompts = ('^ok ');
    }
    foreach my $re (@fwprompts) {
        $hash{$re} ||= FIRMWARE;
    }

    foreach my $re (derefarray($host->prompt_norm(),$host->prompt_take(),$host->prompt_virt())) {
        $hash{$re} ||= UP;
    }

    $hash{"Selection \\(1-\\d\\)\\?"} = BOOT_MENU;
    $hash{"Please make a selection:"} = BOOT_MENU;
    $hash{"^>"} = NORC;
    $hash{"Zero disks and install a new file system\\? "} = ZERO_DISKS_QUERY;
    $hash{"Install a new file system\\? "} = ZERO_DISKS_QUERY;
    $hash{"This will erase all the data on the disks, are you sure\\? "} = ZERO_DISKS_QUERY;
    $hash{"^\\*> "} = MAINT;
    $hash{"Continue with boot\\? "} = NONSTD_BOOT_QUERY;
    $hash{"(?s:Hit \\^C to abort.*Waiting for giveback\\.\\.\\.)"} = TAKEN_OVER_WAITING_OFW;
    $hash{"Waiting for giveback\\.\\.\\.\\(Press Ctrl-C to abort wait\\)"} = TAKEN_OVER_WAITING_ONTAP;
    $hash{"Waiting for giveback"} = TAKEN_OVER_WAITING_ONTAP;
    $hash{"Press Ctrl-C for Maintenance menu to release disks"} = TAKEN_OVER_WAITING_ONTAP;
    $hash{"Do you wish to halt this node rather than wait \\[y/n\\]\\? "} = TAKEN_OVER_QUERY;
    $hash{"(?s:YOUR FILESYSTEMS MAY BE DESTROYED.*Do you wish to continue \\[y/n\\]\\? )"} = TAKEN_OVER_DEAD;
    $hash{"(?s:This command will release all disk reservations.*Do you wish to continue\\? )"} = MAINT_RELEASE_DISKS_QUERY;
    $hash{"floppy boot\\? "} = FLOPPY_BOOT;
    $hash{"login: "} = USERNAME;
    $hash{"Password:"} = PASSWORD;
    $hash{"Password: "} = PASSWORD;
    $hash{"Enter Diag, Command or Option: "} = DIAG_MAIN_PROMPT;
    $hash{"Select test or feature by number "} = DIAG_SUB_PROMPT;
    $hash{"The operating system has halted\\.\\s*Please press any key to reboot\\."} = HALTED;
    $hash{"This will replace all flash-based configuration with the last backup to\r\ndisks\\. Are you sure you want to continue\\?"} = RESTORE_SYSTEM_CONFIG_QUERY;


    if ($hosttype =~ /grub/) {
        $hash{netboot_regex("grub> ")} = GRUB_CMD;
    } elsif ($hosttype =~ /netboot/) {
        $hash{netboot_regex("Choose Boot Method:")} = NETBOOT_METHOD;
        $hash{netboot_regex("Local IP:")} = NETBOOT_LOCAL_IP;
        $hash{netboot_regex("Server IP:")} = NETBOOT_SERVER_IP;
        $hash{netboot_regex("/tftpboot/<file>:")} = NETBOOT_FILE;
        # C1100 hack: this next regex cannot be longer, since the 'e'
        # in value (as well as maybe some other characters depending
        # on what user types on previous line) may match the previous
        # line ("/tftpboot/<file>:") and so the assumption in
        # netboot_regex about non-spaces wouldn't hold:
        $hash{netboot_regex("Are these valu")} = NETBOOT_QUERY;
    }

    if ($transients) {
        $hash{"Zeroing disks takes about \\d+ minutes"} = ZERO_DISKS;
        $hash{"Press CTRL-C for (?:special|floppy) boot menu"} = CTRL_C; # (7G)
        $hash{"(?s:Press Ctrl-C for Boot Menu.*".("\\*"x31).")"} = CTRL_C; # (converged ontap)
        $hash{"Firmware version \\d+ has been successfully downloaded on to the \\S+ card. The system will automatically reboot now."} = NVRAM_FW_UPDATE;
	$hash{"Updating nvram firmware, battery is (?:off|on). The system will automatically reboot when the update is complete."} = NVRAM_FW_UPDATE;
        if ($hosttype =~ /(vsim|vsa)/) {
            $hash{"Hit \\[Enter\\] to boot immediately, or any other key for command prompt."} = SIMLOADER_BANNER;
        } elsif ($hosttype =~ /^(filer|cache)$/) {
            $hash{"Open Firmware by FirmWorks"} = OFW_BANNER;
        } elsif ($hosttype =~ /grub/) {
            $hash{netboot_regex("or \\'c\\' for a command-line\\.")} = GRUB_MENU;
        } elsif ($hosttype =~ /cfe|ng/) {
            $hash{"CFE version \\d+\\.\\d+\\.\\d+ based on Broadcom CFE"} = CFE_BANNER;
            $hash{"Boot Loader version \\d+\\.\\w+"} = LOADER_BANNER;
            $hash{"Starting AUTOBOOT press Ctrl-C to abort"} = LOADER_ABORT_AUTOBOOT;
        }
        $hash{"HALT: cluster partner has taken over"} = TAKEN_OVER_TRANSIENT;
    }
    return \%hash;
}

# A helper routine to perform most of "boot_state", documented below.

sub boot_state_once {
    my ($spec, %opts) = @_;
    my $conn = make_conn($spec) or return undef;
    my $append_to = $opts{append_to};
    my @probe_timeout = defined($opts{probe_timeout}) ?	@{$opts{probe_timeout}} : (5, 120);
    
    return eval {
       my $prompthash = make_prompt_hash($conn->{hostp},$opts{transient});
       my @prompts = keys(%$prompthash);
       
       AGAIN: {
           my $regular_timeout = Tharn::timeout2time($opts{timeout} || 120);
           my $timeout;
           my $do_probe_timeout;
           if (@probe_timeout && (Tharn::timeout2time($probe_timeout[0]) < $regular_timeout)) {
               $timeout=$probe_timeout[0];
               $do_probe_timeout=1;
           } else {
               $timeout=$regular_timeout;
               $do_probe_timeout=0;
           }
           my $maxidle = $opts{maxidle} || 120;
           my $timed_out = 0;
           my $idled = 0;
           my $at_eof = 0;
           my $before = "";
           my $matching_regex;
           
 # burt 328123: When upgrading from ontap 7.x to 8.0, something
 # like this may be printed, where the ">" at start of line
 # would otherwise be confused with a NORC prompt:
 #    [...]
 #    CPU Type: Dual Core AMD Opteron(tm) Processor 885
 #    > setenv AUTOBOOT_FROM primary
 #    > setenv USE_SECONDARY false
 #    [...]
 # Thus, we ignore it with '-ignore'.  The ignore string is
 # anchored to 'CPU Type' rather than to 'setenv' because it
 # must ignore even when all we have read so far is the first
 # ">".
           
           my $match = $conn->expect(
                                     match => \@prompts,
                                     timeout => $timeout,
                                     max_idle => $maxidle,
                                     ignore => ["^CPU Type: .*(?:\n>.*)+"],
                                     on_timeout => sub { $timed_out=1; undef },
                                     on_idle => sub { $idled=1; undef },
                                     on_eof => sub { $at_eof=1; undef },
                                     matching_regex => \$matching_regex,
                                     before => \$before,
                                    );
           
           if ($timed_out && $do_probe_timeout) {
               $conn->send("\r");
               shift @probe_timeout if (@probe_timeout > 1);
               if ($opts{_probed}) {
                   ${$opts{_probed}}=1;
               }
               redo AGAIN;
           } elsif ($timed_out || $idled) {
			   die fatal_result(["Timed out waiting for recognizable output to identify current boot state",
								 "  Timeout: ".(($timed_out) ? "$timeout (timeout)": "$maxidle (idle)"),
								 "  Connection: " . $conn->{id},
								 "  Looking for one of: '" . join("' '",@prompts)."'",
								 ]);
           } elsif ($at_eof) {
               if ($append_to) {
                   $$append_to .= $before;
               }
               die config_result(["Unexpected EOF while reading console to identify current boot state",
                                  "  Connection: " . $conn->{id},
                                 ]);
           } elsif ($matching_regex) {
               if ($append_to) {
                   $$append_to .= $before . $match;
               }
               my $state = $prompthash->{$matching_regex};
               if (!$state) {
                   die "boot_state_once internal error: matched, but no regexp matches \"$matching_regex\"\n";
               }
			   return $state;
           } else {
               die script_result("boot_state_once internal error: timed_out=$timed_out idled=$idled at_eof=$at_eof matching_regex=$matching_regex match=$match");
           }
       }
   }
}


# boot_state_delay:

# As boot_state_once, but also deal with possible false matches, as
# follows.

# (this function is a helper for boot_state).

# If we think we're in one of the states listed in this table, we
# should listen a bit longer and see if the state changes.  If the
# state doesn't change within that time, though, we return the state
# we initially thought we saw.

# For NETBOOT_* and GRUB_*, it might be a false hit because some
# bioses (C1xxx, C2xxx, but not F8x) send serial output as diffs to
# the screen, which not only makes it *really* hard to read, but we
# end up seeing previous output over and over again (because scrolling
# is emulated quite inefficiently)

# For UP, a false prompt may appear as a side effect of how ttymux
# emits the last console line again whenever a console/telnet user
# logs in.

# For any non-transient state, it might be a false hit if we
# previously speculatively sent ctrl-C as input.

my %MIGHT_BE_FALSE = (
                      BOOT_MENU => 2,
                      UP => 5,
                      GRUB_CMD => 5,
                      NETBOOT_METHOD => 5,
                      NETBOOT_LOCAL_IP => 5,
                      NETBOOT_SERVER_IP => 5,
                      NETBOOT_FILE => 5,
                      NETBOOT_QUERY => 5,
                     );

sub boot_state_delay {
    my ($spec, %opts) = @_;
    my $state = boot_state_once($spec, %opts);
    while (1) {
        last if !$state;
        my $retry = $MIGHT_BE_FALSE{$state};
        if (!defined($retry)
            && !($States{$state}{is_transient})
            && $opts{_probed}
            && ${$opts{_probed}}) {

            # We have recently speculatively sent a carraige return to
            # try to scare up a prompt, but this may have caused more
            # prompts i.e. the prompt we're looking at now might not
            # be the latest.

            $retry=5;
        }
        last if !$retry;
        my $newstate = boot_state_once($spec,%opts,timeout=>$retry,
                                       probe_timeout=>[]);
        last if !$newstate;
        $state = $newstate;
    }
    if ($state) {
        $@ = undef;
    }
    if (!$States{$state}{is_transient} && $opts{_probed}) {
        ${$opts{_probed}}=0;
    }

    return $state;
}

## @name ::boot_state
## @summary determine current state from console output

## @description
## Listen to the appliance $a's console to figure what state it is in.
## (By default, and depending on the probe_timeout option,
## we might send carriage returns to encourage
## the console to divulge a prompt).
##
## On success, return the constant described above for whatever state
## $a is in.  On error, return undef, and return the text of the error
## in $@.

## @synopsis
##   boot_state($a, [transient=&gt;0], [timeout=&gt;120], [maxidle=&gt;120], [probe_timeout=&gt;[5,120]], [append_to=&gt;\$var])

## @arg $a Required: a host record or connection record or connspec.  If
## it's a host record, we will connect to that host's "#console"
## connection.

## @arg transient Optional: if 0, which is the default, then don't
## return any of the states listed as "transient"; instead, keep going
## and wait for a non-transient state.  If 1, then return such a state
## if seen.

## @arg timeout Optional: how long to wait for something recognizeable
## before giving up. Default is 120 seconds. 

## @arg maxidle Optional: how long to suffer total lack of *any*
## output by host before giving up.  Default is 120 seconds (and so
## only interesting if timeout, above, is increased beyond its default
## of 120)

## @arg probe_timeout Optional: an array reference.  If non-empty,
## then it lists times at which to send carriage returns in order to
## try to get a prompt.  The default is at 5 seconds and 120 seconds.
## If the list runs out, the last value in the list is used repeatedly.
## (The 5 second setting is intended to handle the case where
## boot_state is called on a console that was already sitting at a
## prompt before boot_state was called; leave it out when you know
## that the appliance has already been pushed into changing states and
## should be printing something recognizable shortly.  The 120 second
## setting is intended to recover, inefficiently and to the extent
## that one can, from a prompt never having been seen because the
## console output was lost or garbled.  Leave it out when the product
## does not have bugs that cause this, but unfortunately our product
## does: see burt 308673, burt 296403, etc.)

## @arg append_to Optional: if given, the value should be a reference
## to a scalar.  The output of the connection to $a will be appended
## to this scalar.

## @author lincoln@netapp.com

use constant BOOT_STATE_OPTS_REGEX => qr/^(?:timeout|maxidle|transient|append_to|probe_timeout|_probed)$/;

sub boot_state {
    my ($spec, @opts) = @_;
    my %opts = @opts;
    foreach my $key (keys %opts) {
        if ($key !~ BOOT_STATE_OPTS_REGEX) {
            return eval { die "boot_state: unknown option \"$key\"\n"; };
        }
    }
    return boot_state_delay($spec,@opts);
}

# return a list of strings to be sent next, while in state $from, to
# get one step closer to state $to.  Also return whether the next
# state following $from is allowed to have changed from being an ontap
# state to a non-ontap state or vice versa.

# next_message returns an empty list if it couldn't figure what to
# send; $@ is filled in to describe why.

# next_message may return more than one message, in the case where
# this call to next_message knows better than future calls to
# next_message what messages to send in the future.  (this is mainly
# for typing in multiple commands at the same prompt; future
# next_message calls wouldn't know how far along we are, since the
# prompt looks the same each time)

sub next_message ( $$$$$$$$$ ) {
    my ($host, $from, $to, $username, $password, $bootimage, $haltcmd, $iftakenover, $restore_system_config) = @_;
    my $hosttype = $host->{hosttype};
    if ($from eq $to) {
        return (0, "");
    } elsif ($from eq MINI_MONITOR) {
        return(1, "bye\r");
    } elsif ($from eq HALTED) {
        return(1, " ");
    } elsif ($from eq MAINT && $iftakenover eq "takeback") {
        # release disks before leaving maintenance mode
        return (0, "storage release disks\r", 0, "yes\r", 1, "$haltcmd\r");
    } elsif ($from =~ /^(UP|NORC|ZERO_DISKS|MAINT)$/) {
        # states one can't get out of except via "halt" or "reboot"
        return (1, "$haltcmd\r");
    } elsif ($from eq FIRMWARE) {
        if ($hosttype =~ /vsim/) {
            if ($bootimage) {
                return (map { (1, "$_\r") } (split(/;/,$bootimage)));
            } else {
                return (1, "boot\r");
            }
        } elsif ($hosttype =~ /cfe|ng|vsa/) {
            if ($bootimage) {
                return (map { (1, "$_\r") } (split(/;/,$bootimage)));
            } else {
                return (1, "boot_ontap\r");
            }
        } elsif ($hosttype =~ /(grub|netboot)/) {
            # The "ok" prompt is the 'mini-monitor' after halt, not
            # something that happens on powerup
            return (1, "bye\r");
        } elsif ($hosttype =~ /simulator/) {
            return (1, "cd " . $host->{simdir} . " && /usr/local/bin/sudo $bootimage\r");
        } else {
            return $bootimage ? (1, "boot $bootimage\r") : (1, "boot\r");
        }
    } elsif ($from eq BOOT_MENU) {
        if ($to =~ /^(UP|USERNAME|PASSWORD)$/) {
            return $restore_system_config ? (0, "6\r") : (0, "1\r");
        } elsif ($to eq NORC) {
            return (0, "2\r");
        } elsif ($to eq ZERO_DISKS || $to eq ZERO_DISKS_QUERY) {
            return (0, "4\r");
        } elsif ($to eq MAINT) {
            # the extra return is because the appliance seems to print one
            # prompt of ">" before printing "*>", so it looks like
            # it's reached NORC rather than MAINT.
            return (0, "5\r", 0, "\r");
        } else {
            # we need to boot again to go backwards.  Fastest/safest
            # way to halt from here is probably maintenance mode
            return (0, "5\r");
        }
    } elsif ($from eq RESTORE_SYSTEM_CONFIG_QUERY) {
        return ($to =~ /^(UP|USERNAME|PASSWORD)$/) ? (1, "yes\r") : (0, "no\r");
    } elsif ($from eq ZERO_DISKS_QUERY) {
        return ($to eq ZERO_DISKS) ? (0, "yes\r") : (1, "no\r");
    } elsif ($from eq NONSTD_BOOT_QUERY) {
        # the extra return is because the appliance seems to print one
        # prompt of ">" before printing "*>", so it looks like
        # it's reached NORC rather than MAINT.
        return (0, "yes\r", 0, "\r");
    } elsif ($from =~ /TAKEN_OVER/) {
        if ($iftakenover eq "error") {
            eval { die "partner has taken over\n"; };
            return ();
        } elsif ($iftakenover eq "wait") {
            # wait means "Assume the takeover will abate soon, because
            # some other test will make the partner give back".  Whatever
            # TAKEN_OVER_* prompt we're at will be cancelled by the act of
            # giveback, so we just have to wait for subsequent prompts.
            return (1, "");
        } elsif ($iftakenover eq "takeback") {
            # takeback means "Assume the partner is dead or having its
            # disks reinitialized so that it's up to us to clear
            # reservations and take back our disks".
            if ($from eq TAKEN_OVER_WAITING_OFW) {
                # Break out of "waiting for giveback", then release
                # disks at the OFW prompt.
                return (1, "\cC", 0, "release-disks\r");
            } elsif ($from eq TAKEN_OVER_WAITING_ONTAP) {
                # Break out and go to maintenance mode, so we can
                # release disks.  (See MAINT state for the
                # continuation).  ctrl-C for hardware, ctrl-X for
                # simulators.
                return (1, $host->{irptchar});
            } elsif ($from eq TAKEN_OVER_QUERY) {
                # Partner is still alive and thinks it has taken over.
                # Wait for it to give back or be halted, etc.
                return (1, "");
            } elsif ($from eq TAKEN_OVER_DEAD) {
                # Partner asks if we wish to take back, say yes.
                return (1, "y\r");
            } elsif ($from eq TAKEN_OVER_TRANSIENT) {
                # This host has halted because it was just taken over.
                # Assume we will get our chance to take back later.
                # Transient state, so no prompt to answer.
                return (1, "");
            } else {
                eval { die "lib/boot.pm: internal error: unhandled TAKEN_OVER state \"$from\"\n"; };
                return ();
            }
        } else {
            eval { die "unrecognized value for iftakenover option: \"$iftakenover\"\n"; };
            return ();
        }
    } elsif ($from eq FLOPPY_BOOT) {
        return (0, ($to =~ /^(UP|USERNAME|PASSWORD)$/) ? "no\r" : "yes\r");
    } elsif ($from eq USERNAME) {
        return (0, $username . "\r");
    } elsif ($from eq PASSWORD) {
        return (0, $password . "\r");
    } elsif ($from eq GRUB_MENU) {
        if ($to eq GRUB_CMD) {
            return (0, "c");
        } elsif (!$bootimage) {
            return (1, "\r");
        } elsif ($bootimage =~ /^\d+$/) {
            return (1, ("v" x $bootimage) . "\r");
        } else {
            # anything $bootimage more complex must be dealt with at
            # the grub command line
            return (0, "c");
        }
    } elsif ($from eq GRUB_CMD) {
        if ($to eq GRUB_MENU) {
            return (0, "\e");
        } elsif (!$bootimage or $bootimage =~ /^\d*$/) {
            # this bootimage is more appropriate for the menu
            return (0, "\e");
        } elsif ($bootimage =~ /^\((.+?)\)(.*)/) {
            return (0, "root ($1)\r", 0, "kernel ($1)$2\r", 1, "boot\r");
        } else {
            return (map { (1, "$_\r") } (split(/\n/,$bootimage)));
        }
    } elsif ($from eq NETBOOT_METHOD) {
        return (1, netboot_bootimage($bootimage,'method'));
    } elsif ($from eq NETBOOT_LOCAL_IP) {
        # C1100 hack: we shouldn't see this state just before a boot
        # (so the first return value should be 0), but because of how
        # scrolling works poorly on a C1100, we sometimes see this
        # state after we've answered NETBOOT_QUERY, so we permit this
        # to preceed an ontap state.
        return (1, netboot_bootimage($bootimage,'local_ip'));
    } elsif ($from eq NETBOOT_SERVER_IP) {
        # C1100 hack: as above.
        # Also, the machines print "Server IP:" when booting, so we
        # falsely detect a NETBOOT_SERVER_IP prompt just before ontap
        # boots, so this needs to be counted as allowing an change into ontap
        # afterward.
        return (1, netboot_bootimage($bootimage,'server_ip'));
    } elsif ($from eq NETBOOT_FILE) {
        # C1100 hack: as above
        return (1, netboot_bootimage($bootimage,'file'));
    } elsif ($from eq NETBOOT_QUERY) {
        return (1, "y\r");
    } elsif ($from eq CTRL_C) {
        # Gigan/Rodan hack: the filer is not quite ready for us to
        # send ctrl-C at the point where it says it can, so let's
        # sleep just a little bit here.
        sleep(1);
        return ($to =~ /^(?:UP|USERNAME|PASSWORD)$/) ? (0, "") : (0, $host->{irptchar});
    } elsif ($from =~ /_BANNER$|_ABORT_AUTOBOOT$/) {
        if ($bootimage || ($to eq FIRMWARE)) {
            # Break out of the boot sequence to attempt to reach a
            # firmware prompt, either because we want to boot in a
            # nonstandard away ($bootimage is nonempty) or FIRMWARE
            # was itself the desired destination.

            # For LOADER_*, we send nothing at LOADER_BANNER and wait
            # until LOADER_ABORT_AUTOBOOT to break out if it appears.
            # (If we send ctrl-c at LOADER_BANNER time when the
            # firmware was not planning to boot automatically, then
            # either the LOADER> prompt arrives so quickly or has a
            # good enough memory such that the ctrl-C will end up
            # producing a second LOADER> prompt.

            # For real CFE, we cannot do the same thing (even though
            # there is a CFE "press any key to abort autoboot" prompt,
            # it occurs too late to actually have ever been usable by
            # anyone) but then we don't have to (an extra space being
            # sent is no problem, unlike an extra ctrl-C).

            my $interrupt_character;
            if ($from eq OFW_BANNER) {
                $interrupt_character="\c?";
            } elsif ($from eq LOADER_BANNER) {
                $interrupt_character=""; # send nothing
            } elsif ($from eq LOADER_ABORT_AUTOBOOT) {
                $interrupt_character="\cC";
            } else {
                $interrupt_character=" ";
            }
            return (0, $interrupt_character);
        } else {
            # Send nothing (allow machine to autoboot to ontap if it wants to)
            return (1, "");
        }
    } elsif ($from eq MAINT_RELEASE_DISKS_QUERY) {
        # answer to "storage release disks" query.  This code should
        # almost never be reached, since we add a subsequent "yes"
        # response at the same time as we add "storage release
        # disks".
        return (0, "yes\r");
    } elsif ($from eq NVRAM_FW_UPDATE) {
        # The filer announced that it will now reboot automatically.
        # Don't send anything, but return that going to non-ontap
        # state at this point would be fine.
        return (1, "");
    } elsif ($from eq DIAG_MAIN_PROMPT) {
        return (0, "exit\r");
    } elsif ($from eq DIAG_SUB_PROMPT) {
        return (0, "99\r");
    }
    eval { die "lib/boot.pm: internal error, no code to handle transit from \"$from\" to \"$to\"\n"; };
    return ();
}
    

# send appropriate strings to go from state $from to state $to.
# return 1 or undef for whether the transit succeeded.  if wantarray,
# also return the state we ended up in (or undef if we don't know).

sub transit_to_nopowercycle ( $$$@ ) {
    my ($conn, $to, $paramsref, %opts) = @_;
    if (!$States{$to}) {
        return eval { die "unknown boot state \"$to\"\n"; };
    }

    my $host = $conn->{hostp};

    my $bootimage=make_bootimage($host->id,$opts{prefix},$paramsref);
    my $username=$paramsref->{USERNAME} || 'root';
    my $password=$paramsref->{PASSWORD};
    my $timeout=defined($opts{timeout}) ? $opts{timeout} : 600;
    $timeout=timeout2time($timeout);
    my $maxidle=defined($opts{maxidle}) ? $opts{maxidle} : 120;
    my $maxcycle=defined($opts{maxcycle}) ? $opts{maxcycle} : 20;
    my $haltcmd;
    if ($host->{hosttype} =~ /filer-ng/) {
    $haltcmd=defined($opts{haltcmd}) ? $opts{haltcmd} : "run local halt -f -t 0";
    }
    else {
    $haltcmd=defined($opts{haltcmd}) ? $opts{haltcmd} : "halt -f -t 0";
    }
    my $iftakenover=defined($opts{iftakenover}) ? $opts{iftakenover} : "error";
    my $restore_system_config=defined($opts{restore_system_config}) ? $opts{restore_system_config} : 0;
    my $_probed_scalar;
    my $_probed=\$_probed_scalar;

    my @bootstate_opts = (timeout => $timeout, maxidle => $maxidle,
                          transient => 1, append_to => $opts{append_to},
                          _probed => $_probed);
    my $from = $opts{from} || undef;

    my $res = eval {
        if (!$from) {
            $from = boot_state($conn, @bootstate_opts, probe_timeout=>[5, 120]) or die "$@";
        }
        $Log->trace("transit_to: want to reach '$to' state from '$from' state");

        my @msgs;
        for (my $cycle=0;$cycle<$maxcycle;$cycle++) {
            $Log->trace("transit_to: reached boot state \"$from\"");
            if ($from eq $to) {
                return 1;
            }
            if (!@msgs) {
                push @msgs, next_message($host,$from,$to,$username,$password,$bootimage,$haltcmd,$iftakenover,$restore_system_config);
            }
            if (!@msgs) {
                die "Cannot transit from $from to $to state: $@";
            }
            my $can_change_ontap = shift @msgs;
            my $msg = shift @msgs;


            # GRUB/Rolex hack: GRUB appears to not like taking
            # command input too quickly (I notice it getting only the
            # first 16 characters of a command, and it is probably not
            # a coincidence that serial port hardware decended from
            # the 16550 contains 16 character wide buffers.)

            if (($conn->{hostp}->{hosttype} =~ /grub/) && (length($msg)>16)) {
                # chop $msg into 16 character max chunks
                my (@chunks) = ($msg =~ /.{1,16}/sg);
                foreach my $m (@chunks) {
                    say($conn,$m,'-nocr','-blind','-ok');
                    sleep(2);
                }
            } else {
                # the non-hack case...
                say($conn,$msg,'-nocr','-blind','-ok');
            }
            my $previous = $from;
            $from = boot_state($conn, @bootstate_opts, probe_timeout=>[120]) or die $@;
        
            if (!$can_change_ontap) {
                if ($States{$previous}{is_ontap} && !$States{$from}{is_ontap}) {
                    die "Command \"$msg\" should not have gone from ontap state \"$previous\" to non-ontap state \"$from\"\n";
                }
                if (!$States{$previous}{is_ontap} && $States{$from}{is_ontap}) {
                    die "Command \"$msg\" should not have gone from non-ontap state \"$previous\" to ontap state \"$from\"\n";
                }
            }
        }
        die "cycled through too many boot states without reaching the right one ($maxcycle times)\n";
    };
    return wantarray ? ($res, $from) : $res;
}

## @name ::transit_to
## @summary enter commands to reach a specific boot state
##
## @description
## Repeatedly determine what state the appliance $a is in and then enter
## commands to try to reach state $to.
##
## Return a true value if we end up in the state the user asked for,
## otherwise return false and set $@ to contain a string describing
## the failure.  If transit_to is called in an array context, the
## second element returned will be the state we ended up in (or undef
## if that is not known).
##
## (An example of when we might end up in a different state than what
## was asked for is when the iftakenover option is set to "error" and
## we reach one of the TAKEN_OVER_* states).

## @synopsis
##   transit_to($a, $to, [option=&gt;value, [option=&gt;value [,...]]])

## @arg $a Required: a host record or connection record or connspec.
## If it's a host record, we will connect to that host's "#console"
## connection.

## @arg $to Required: what state the appliance should go to.  (see the
## constants described for lib/boot.pm in general)

## @arg prefix=&gt;... Optional: if provided, various other options to
## this function default to params that begin with this prefix.

## @arg timeout=&gt;... Optional: transit_to should finish with an
## error if it cannot reach its goal within this many seconds.
## Default is 600 seconds.  (this is a usual NATE timeout: number of
## seconds, or absolute time, with -1 meaning infinite).

## @arg maxidle=&gt;... Optional: transit_to should finish with an
## error if an appliance ever prints absolutely nothing for a span of
## this many seconds.  Default is 120 seconds.

## @arg maxcycle=&gt;... Optional: transit_to should finish with an
## error if it has gone through more than this many boot states
## without reaching the state it wants.  The default is 20.

## @arg powercycle=&gt;... Optional: if 1 and if transit_to fails
## once, try to cycle power on the host and then try exactly once
## more.  The default is the value of param $prefix_POWERCYCLE if it
## exists, or else the field POWERCYCLE in the host record for $a if
## it exists, otherwise 0.  <p><i>(Note about power cycling: the
## routine used to cycle power is the "powercycle" method on the host
## object for $a, which in turn is probably "power_cycle
## &gt;hostname&lt;" or whatever the "powercmd" field in that host
## object is set to.  For simulators, a powercycle means to interrupt
## the simulator process by typing ctrl-C at the simulator's
## console)</i>

## @arg netboot=&gt;... Optional: if 1, and if $a is not a simulator,
## then boot over the network (as specified by arguments
## netboot_local_ip, netboot_file, netboot_device, tftphost,
## tftphost_ip, and tftphost_tftpdir).  If 0, or when $a is a
## simulator, then the form of booting (netbooting or otherwise) is
## performed according to argument "bootimage".  This defaults to the
## param $prefix_NETBOOT if that exists, or else the field NETBOOT in
## the host record for $a if that exists, otherwise 0.

## @arg netboot_device=&gt;... Optional: what ethernet device to use
## for network booting, when netboot=&gt;1.  This defaults to the
## value of param $prefix_NETBOOT_DEVICE if it exists, or the field
## NETBOOT_DEVICE in the host record of $a if that exists, or else it
## defaults to "net" when using open firmware or "e0a" when using
## CFE.  (The value of this does not matter for GRUB or for the
## interactive netboot floppy, as one cannot select interactively
## which interface they will use).

## @arg netboot_local_ip=&gt;... Optional: what IP address $a will use
## as it netboots.  This defaults to the value of param
## $prefix_NETBOOT_LOCAL_IP if it exists, or to the field NETBOOT_DEVICE
## in the host record of $a if that exists, or to param $prefix_IP if
## that exists, or else to the "default_ip" field in the host record
## of $a.

## @arg netboot_netmask=&gt;... Optional: what netmask $a will use
## as it netboots.  This defaults to the value of param
## $prefix_NETBOOT_NETMASK if it exists, or to param $prefix_NETMASK
## if that exists

## @arg netboot_gateway=&gt;... Optional: what gateway $a will use
## as it netboots.  This defaults to the value of param
## $prefix_NETBOOT_GATEWAY if it exists, or to param $prefix_GATEWAY
## if that exists

## @arg netboot_file=&gt;... Optional: what kernel image to request
## when netbooting.  The default is the value of param
## $prefix_NETBOOT_FILE if that exists, or the field NETBOOT_FILE in
## the host record of $a if that exists, or else the default is
## "Boot_" followed by the hostid of $a.

## @arg tftphost=&gt;... Optional: the hostspec of the host that will
## serve tftp.  This is only used to provide better defaults for
## tftphost_ip and tftphost_tftpdir.  The default is the value of
## param $prefix_TFTPHOST if that exists, or the value of field
## "TFTPHOST" in the host record of $a if that exists.

## @arg tftphost_ip=&gt;... Optional: the IP address of the tftp
## server from which $a should netboot.  This defaults to the value of
## param $prefix_TFTPHOST_IP if that exists, or to the value of
## TFTPHOST_IP in the host record for $a if that exists, or to the
## "default_ip" field in the host record for the host given in the
## "tftphost" option if that exists.

## @arg tftphost_tftpdir=&gt;... Optional: the subdirectory from which
## the tftp server serves.  This is only needed when netbooting
## using GRUB, which insists on using absolute paths when requesting
## files.  This defaults to the value of param
## $prefix_TFTPHOST_TFTPDIR if that exists, or to the value of
## TFTPHOST_TFTPDIR in the host record for $a if that exists, or to
## the "TFTPDIR" field in the host record for the host given in the
## "tftphost" option if that exists, or else the default is
## "/tftpboot".

## @arg bootimage=&gt;... Optional: a description of what kernel to
## boot when 'netboot' is false.  What value is useful depends on the
## hosttype (see discussion of hosttypes in the documentation for this
## library as a whole).  This defaults to the param $prefix_BOOTIMAGE
## if that exists, else the field "bootimage" in the host record
## corresponding to $a.  If booting a simulator and if the default
## would otherwise be blank, a value will be created from
## "maytag_suffix", "ontapdir", "brand", and "simargs" options like so:
## "$ONTAPDIR/prod/build/$BRAND/maytag.$MAYTAG_SUFFIX $SIMARGS".

## @arg brand=&gt;... Optional: the brand of kernel to be booted
## (such as "netapp" or "ibm") .  Used to construct a bootimage for
## simulators if no non-empty value is given.  If the value is
## "netapp", then the bootimage will point to "prod/build",
## otherwise it will be under the given subdirectory of
## "prod/build".  Defaults to the value of param $prefix_BRAND if
## that exists, or to the value of BRAND in the host record of $a if
## that exists


## @arg maytag_suffix=&gt;... Optional: the build flags used to build
## the flavor of kernel to be booted.  Used to construct a bootimage
## for simulators if no non-empty value is given.  Defaults to the
## value of param $prefix_MAYTAG_SUFFIX if that exists, or to the
## value of MAYTAG_SUFFIX in the host record of $a if that exists.

## @arg ontapdir=&gt;... Optional: the top-level directory of the
## ontap source tree corresponding to the kernel to be booted.  Used
## to construct a bootimage for simulators if no non-empty value is
## given.  Defaults to the value of param $prefix_ONTAPDIR if that
## exists, or to the value of ONTAPDIR in the host record of $a if
## that exists.

## @arg simargs=&gt;... Optional: extra arguments to pass to the ontap
## simulator.  Used to construct a bootimage for simulators if no
## non-empty value is given, otherwise has no effect.  Defaults to the
## value of param $prefix_SIMARGS if that exists, or to the value of
## SIMARGS in the host record of $a if that exists.

## @arg username=&gt;... Optional: the user to log in as, if a
## "login:" prompt appears (USERNAME state).  The default is the value
## of param $prefix_USERNAME if that exists, or else the connection's
## "username" field, which tends to default to "root" if not
## overridden.

## @arg password=&gt;... Optional: the password for login, if a
## "password:" prompt appears (PASSWORD state).  The default is the
## value of param $prefix_PASSWORD if that exists, or else the
## connection's "password" field, which tends to default to the host's
## "default_password" field, which defaults to empty.

## @arg haltcmd=&gt;... Optional: the command used to leave UP (or
## similar) state.  The default is "halt -f -t 0".

## @arg from=&gt;... Optional: the current state of the appliance.
## The boot library tries to guess this state, and so this option is
## better left out.

## @arg iftakenover=&gt;... Optional: what to do if this node has been
## taken over.  If "error", then transit_to fails as soon as it
## reaches any of the TAKEN_OVER_* states.  If "wait", then transit_to
## waits whenever it sees TAKEN_OVER_* states (useful if the partner
## will do giveback soon).  If "takeback", then the behavior is as for
## "wait", but it also tries to release disk reservations and take
## back its filesystem whenever the opportunity arises (useful if the
## partner will halt or be reinitialized soon).  When waiting, it only
## waits as long as "maxidle" or "timeout" options allow.

## @arg append_to=&gt;... Optional: if given, the value should be a
## reference to a scalar.  The output of the connection to $a will be
## appended to this scalar.

## @arg restore_system_config=&gt;... Optional: if 1, then option 6 
## "Update flash from backup config." is selected at the BOOTMENU while
## transitioning to the UP state. The default is 0, and this selects
## option 1 at the BOOTMENU i.e "Normal Boot".

## @author lincoln@netapp.com

# as transit_to_nopowercycle, but if the first attempts to transit
# fail, and if the 'powercycle' option is set, cycle power and then
# try once more.

use constant TRANSIT_TO_OPTS_REGEX => qr/^(?:prefix|timeout|maxidle|maxcycle|powercycle|netboot|netboot_device|netboot_local_ip|netboot_netmask|netboot_gateway|netboot_file|tftphost|tftphost_ip|tftphost_tftpdir|bootimage|maytag_suffix|ontapdir|simargs|username|password|haltcmd|from|iftakenover|append_to|restore_system_config)$/;

sub transit_to ( $$@ ) {
    my ($spec, $to, @opts) = @_;
    my (%opts) = @opts;

    my $conn = make_conn($spec) or return undef;
    my $host = $conn->{hostp};

    my %connparams;
    foreach my $opt ("username" , "password") {
        if (exists $conn->{$opt}) {
            $connparams{$opt} = $conn->{$opt};
        }
    }

    my %params = param_merge(
                             hostrec => $host,
                             hash_uc => \%connparams,
                             param_prefix => $opts{prefix},
                             hash_uc => \%opts,
                            );

    foreach my $key (keys %opts) {
        if ($key !~ TRANSIT_TO_OPTS_REGEX) {
            return eval { die "transit_to: unknown option \"$key\"\n"; };
        }
    }

    if (!$params{POWERCYCLE}) {
        return transit_to_nopowercycle($conn,$to,\%params,@opts);
    } else {
        my ($res, $state) = transit_to_nopowercycle($conn,$to,\%params,@opts);
        if (!$res) {
            $host->powercycle();
            return transit_to_nopowercycle($conn,$to,\%params,@opts,from=>undef);
        } else {
            return wantarray ? ($res, $state) : $res;
        }
    }
}

## @name ::boot
## @summary transit to UP state (boot an appliance if necessary)
## @description
## Make sure that an appliance is UP (but don't bother halting first, or
## anything; if it's already all the way up, do nothing)
##
## @synopsis
## <tt>boot($a,...options...)</tt>
##
## @usage
## This is equivalent to <tt>transit_to($a,UP,...options...)</tt>
##
## @author lincoln@netapp.com

sub boot ( $@ ) {
    my ($spec, @opts) = @_;
    return transit_to($spec,UP,@opts);
}

## @name ::halt
## @summary transit to FIRMWARE state (halt an appliance if necessary)
##
## @description
## Try to make sure that an appliance has reached something like a firmware
## prompt (so a unix shell prompt for simulators, and the mini-monitor
## for BIOS-based appliances).  If it's not already UP or in FIRMWARE, it
## might have to boot some more before it can reach that state.
##
## @synopsis
## halt($a,...options...)
##
## @usage
## This is equivalent to <tt>transit_to($a,FIRMWARE,...options...);</tt>
##
## @author lincoln@netapp.com

sub halt ( $@ ) {
    my ($spec, @opts) = @_;
    return transit_to($spec,FIRMWARE,@opts);
}

## @name ::reboot_to
## @summary transit out of ontap, transit to given state
## @description
## As transit_to, but if the appliance seems to already be running OnTAP,
## make sure to halt it first.
##
## @synopsis
## reboot_to($a,$to,...options...)
##
## @usage
## This is equivalent to calling <tt>halt($a,...options...)</tt> only
## if the current state is not FIRMWARE or GRUB_something or
## NETBOOT_something, and then calling
## <tt>transit_to($a,$to,...options...)</tt>
##
## @author lincoln@netapp.com

sub reboot_to ( $$@ ) {
    my ($spec, $to, @opts) = @_;
    my $conn = make_conn($spec) or return undef;

    my %opts = @opts;

    my ($res,$state);

    my @bootstate_opts = map { $_ => $opts{$_} } ('timeout','maxidle','append_to');
    $state = boot_state($conn,@bootstate_opts,transient=>1);
    if ($state && $state !~ /^(FIRMWARE|GRUB_|OFW_|CFE_|LOADER_|SIMLOADER_|VLOADER_|NETBOOT_|MINI_MONITOR|HALTED)/) {
        # We're in a state that may be in ontap.  Jump to a state
        # outside of ontap, so that we can be sure we've left the
        # current ontap boot and will be entering a new one.

        ($res,$state) = transit_to($conn,FIRMWARE,@opts,from=>$state);

        if (!$res) {
            return wantarray ? ($res, $state) : $res;
        }

        # Since the above made progress, future failures will not be
        # of a kind that power cycling was meant to help.
        push @opts, (powercycle => 0);

    }
    return transit_to($conn,$to,@opts,from=>$state);
}

## @name ::reboot
## @summary transit out of ontap, transit to UP state (reboot an appliance)
## @description
## Make sure that an appliance is UP, but if it's running ontap already,
## make sure to halt it first.
##
## @synopsis
## reboot($a,...options...)
##
## @usage
## This is equivalent to calling <tt>halt($a,...options...)</tt> only
## if the current state is not FIRMWARE or GRUB_something or
## NETBOOT_something, and then calling
## <tt>boot($a,$to,...options...)</tt>
##
## @author lincoln@netapp.com

sub reboot ( $@ ) {
    my ($spec, @opts) = @_;
    return reboot_to($spec,UP,@opts);
}

# return a suitable bootimage value.  The value provide by netboot*
# options take precedence (except for simulators, where this is
# meaningless).  Otherwise, we use the value chosen by 'bootimage' as
# a param or hostrec.  If there is none, we try to guess at a good
# value for a simulator.  Otherwise, return the empty string (which
# should successfully boot the default way on any non-simulator).

sub make_bootimage {
    my ($hostid, $prefix, $params) = @_;
    my $hosttype = $params->{TYPE};
    if ($hosttype !~ /simulator/ && $params->{NETBOOT}) {
        return make_bootimage_netboot($hostid,$prefix,$params);
    } elsif (defined($params->{BOOTIMAGE}) && length($params->{BOOTIMAGE})) {
        return $params->{BOOTIMAGE};
    } elsif ($hosttype =~ /simulator/) {
        return make_bootimage_simulator($hostid,$prefix,$params);
    }
    return "";
}

# Return a bootimage built from NETBOOT_* and TFTPHOST_* params,
# suitable for this kind of host.

sub make_bootimage_netboot {
    my ($hostid, $prefix, $params) = @_;
    my $hosttype = $params->{TYPE};
    my %tftphost_params;

    # get param information from TFTPHOST, if we can
    if ($prefix && $params->{TFTPHOST}) {
        my $tftphost = host($params->{TFTPHOST},'-eval_error')
            or die $@;
        %tftphost_params = param_merge(
                                       hostrec=>$tftphost,
                                       param_prefix=>$prefix . "_TFTPHOST"
                                       );
    }

    my $localip = $params->{NETBOOT_LOCAL_IP} || $params->{IP} || die "netboot requested but IP address for filer not known (IP or default_ip for filer not set)\n";
    my $serverip = $params->{TFTPHOST_IP} || $tftphost_params{IP} || die "netboot requested but tftp server IP address not known (TFTPHOST_IP or default_ip for TFTPHOST not set)\n";
    my $netmask = $params->{NETBOOT_NETMASK} || $params->{NETMASK};
    my $gateway = $params->{NETBOOT_GATEWAY} || $params->{GATEWAY};
    my $filename = $params->{NETBOOT_FILE} || "Boot_" . $hostid;

    if ($hosttype =~ /grub/) {
        # Older NetApp-specific versions of grub have a "localhost"
        # command.  Newer versions have an "ifconfig" command
        # (inherited from upstream) and have gotten rid of
        # "localhost".  "tftpserver" exists in both but is deprecated
        # in newer GRUB in favor of "ifconfig".  The version of GRUB
        # in use may not correspond to the version of GRUB used by the
        # OS we wish to boot, so we need to support both.  The version
        # number of newer grub is almost unreadable on C1200 or C2100
        # because of how we've set the newer grub to use the BIOS's
        # console redirection.  We'll deal with this by sending both
        # "localhost"/"tftpserver" and "ifconfig" commands.

        # unfortunately GRUB requires a file argument to the
        # "kernel" command to be an absolute path (I guess
        # relative paths mean something else to GRUB), so we need
        # to pass in the directory from which the tftp server
        # serves.

        my $tftpdir = $params->{TFTPHOST_TFTPDIR} || $tftphost_params{TFTPDIR} || "/tftpboot";

        return
            "localhost $localip\n".
            "tftpserver $serverip\n".
            "ifconfig --address=$localip --server=$serverip".
               ($netmask ? " --mask=$netmask" : "").
               ($gateway ? " --gateway=$gateway" : "").
               "\n".
            "root (nd)\n".
            "kernel (nd)$tftpdir/$filename\n".
            "boot\n";
    } elsif ($hosttype =~ /netboot/) {
        return "N,$localip,$serverip,$filename";
    } elsif ($hosttype =~ /^(filer|cache)$/) {
        my $device = $params->{NETBOOT_DEVICE} || "net";
        return "$device:$serverip,$filename,$localip";
    } elsif ($hosttype =~ /cfe|ng/) {
        my $device = $params->{NETBOOT_DEVICE} || "e0a";
        my $ifconfig_cmd = "ifconfig $device -addr=$localip";
        $ifconfig_cmd .= " -mask=$netmask" if $netmask;
        $ifconfig_cmd .= " -gw=$gateway" if $gateway;
        # netboot is an alias for "setenv BOOTED_FROM netboot; boot -elfxx"
        return "$ifconfig_cmd;netboot tftp://$serverip/$filename";
    } else {
        die "I don't know how hosttype \"$hosttype\" does network booting\n";
    }
}

# Return a bootimage built from ONTAPDIR, MAYTAG_SUFFIX, and SIMARGS params

sub make_bootimage_simulator {
    my ($hostid, $prefix, $params) = @_;
    my $hosttype = $params->{TYPE};
    my $maytag_suffix = $params->{MAYTAG_SUFFIX};
    my $ontapdir = $params->{ONTAPDIR};
    my $brand = $params->{BRAND} || "netapp";
    my $build_dir = ($brand eq "netapp") ? "prod/build" : "prod/build/$brand";
    my $simargs = defined($params->{SIMARGS}) ? $params->{SIMARGS} :
        ($hosttype =~ /cache/) ? "-n" : "";
    if (!$ontapdir || !$maytag_suffix) {
        die "Do not know how to invoke simulator (need non-empty bootimage, or non-empty ontapdir and maytag_suffix)\n";
    }
    my $result = "$ontapdir/$build_dir/maytag.$maytag_suffix";

    if (! (-e $result)) {
        die "File does not exist: $result\n";
    }

    $result .= " $simargs" if length($simargs);
    
    return $result;
}

## @name ::netboot_setup
## @summary guess bootimage for netbooting, make tftp link
##
## @deprecated
## This function is deprecated.  Instead of using this function to
## produce a bootimage in a hosttype-independent way, use the
## 'netboot' option and other companion options when calling
## transit_to() or its many wrappers.  Instead of using this function
## to make symbolic links in /tftpboot, use the test
## Filer/netboot_setup.thpl.
##
## @description
## return a proper bootimage string to network boot $a.  Also, if
## "linkto" is given, create a symbolic link in /tftpboot.
##
## @synopsis 
##   netboot_setup($a,$localip,$serverip,$filename,
##   [linkto=&gt;""], [device=&gt;"net"])
##
## @usage
## <b>$a</b> - a host record or connection record or hostspec of
## appliance to boot.
##
## <b>$localip</b> - What IP address the appliance should use as its
## own when it downloads via TFTP.  (this value is encoded in the
## bootimage string returned)
##
## <b>$serverip</b> - What IP address the appliance should use for the
## TFTP server when it downloads via TFTP (this value is encoded in
## the bootimage string returned)
##
## <b>$filename</b> - What filename the appliance should request when
## it downloads via TFTP.  (this value is encoded in the bootimage
## string returned).
##
## <b>device</b> - When appliance downloads via TFTP, it should use
## this network device.  (this value is encoded in the bootimage
## string returned).  Currently only makes sense when using Open
## FirmWare.  The default is "net", a devalias which already exists on
## many models of appliance.
##
## <b>linkto</b> - If set, $filename should first be removed (if it
## exists) and recreated as a symbolic link to this value.
##
## @bugs
##   This routine assumes that the tftp server serves out of /tftpboot
##   and is the host that is running the test.
##
## @status deprecated
## @author lincoln@netapp.com

sub netboot_setup ( $$$$@ ) {
    my ($spec, $localip, $serverip, $filename, %opts) = @_;
    return eval {
        my $h = make_host($spec);
        my $hosttype = $h->{hosttype};

        my $linkto = defined($opts{linkto}) ? $opts{linkto} : "";
        my $device = defined($opts{device}) ? $opts{device} : "net";

        if ($linkto) {
            my $fullfilename = "/tftpboot/" . $filename;
            if (!unlink($fullfilename) && $! !~ /No such file or directory/i) {
                die "Cannot unlink $fullfilename: $!\n";
            }
            symlink($linkto,$fullfilename)
                or die "Could not create symlink $filename: $!\n";
        }
        if ($hosttype =~ /grub/) {
            # our local netapp version of grub has a "localhost" command;
            # the official GRUB has since added an "ifconfig" command, but
            # we haven't merged in newer grubs it seems

            # unfortunately GRUB requires a file argument to be an
            # absolute path (I guess relative paths mean something else to
            # GRUB), so we assume "/tftpboot/" onto the beginnings of the
            # $filename passed to "kernel".

            return 
                "localhost $localip\n".
                "tftpserver $serverip\n".
                "root (nd)\n".
                "kernel (nd)/tftpboot/$filename\n".
                "boot\n";
        } elsif ($hosttype =~ /netboot/) {
            return "N,$localip,$serverip,$filename";
        } elsif ($hosttype =~ /^(filer|cache)$/) {
            return "$device:$serverip,$filename,$localip";
        } elsif ($hosttype =~ /cfe|ng/) {
            return "ifconfig e0a -addr=$localip;netboot tftp://$serverip/$filename";
        } else {
            die "netboot_setup: I don't know how hosttype \"$hosttype\" does network booting\n";
        }
    };
}

## @name ::wait_for_rsh
## @summary probe for rsh service to become available

## @description

##   Probe using rsh commands until rsh service becomes available.
##   The command used will be "version".

##   (In recent ontap, rsh may not being listening until after the
##   normal filer prompt is printed by a completely booted and
##   logged-in filer)

## @synopsis
##   wait_for_rsh($aspec, [timeout=&gt;...], [delta=&gt;...])

## @arg $aspec Required: hostspec of host to probe

## @arg timeout=&gt;... Optional: after how many seconds to give up
## probing.  The default is 60 seconds.

## @arg delta=&gt;... Optional: how many seconds between probes.  The
## default is 5 seconds.

sub wait_for_rsh ( $@ ) {
    my ($aspec, %opts) = @_;

    return eval {
        my $ahost = host($aspec);
        my $aconn = connect("#rsh");

        my $timeout = defined($opts{timeout}) ? $opts{timeout} : 60;
        my $endtime = Tharn::timeout2time($timeout);
        my $delta = defined($opts{delta}) ? $opts{delta} : 5;

        while (time() < $endtime) {
            #Added support of IBM Release String 
            say($aconn,"version","NetApp Release|Data ONTAP Release",'-eval_error','-timeout', 10);
            return 1 unless $@;
            sleep($delta);
        }

        die "rsh did not succeed within $timeout seconds: $@\n";
    };
}

# Local Variables:
# indent-tabs-mode: nil
# End:
