# Hey emacs, this is basically -*- perl -*-
# $Id:$

## @summary Collection of routines used by CAT
## @keywords CAT, lib
## @author archana@netapp.com

package cat;

use strict;
use Module::Common;

BEGIN
{
    my @subs = qw(
        run_test
        is_cifs_share_present
        get_vfiler_details
        vfiler_destroy
        vfiler_create
        do_shell
        append_unique_line
        contains
        get_traffic
        get_fcp_adapter_info
        get_igroup_info
        is_vol_useful
        get_root_vol
        get_vol_status
    );

	Module::Common::setup_export_tags(
		cat  => [ @subs ]
	);
}

use Fcntl;

use Tharn;

## @name ::run_test
## @description : execute the list of tests passed as a param as subtests
## @usage run_test($bg,@test)
sub run_test($@)
{
    my ($bg, @test) = @_;

    my $testid;

    foreach my $cmd (@test) {
        my $teststring="";
        foreach my $c (@{ $cmd }) {
            if ($teststring eq "") {
                ($testid) = split(".thpl",join("_", split("/", $c)));
                $teststring = qq{"$c"};
            } else {
                $teststring .= qq{, "$c"};
            }
        }
        my $tm0 = time();
        # Start the test as a subtest
        $teststring = "subtest " . qq{$teststring} . qq{, '$bg', -runid, "$testid"} .
                ", -iffatal, -fail, " . q{"Fatal Failure"} . ", -ifscript, -fail, " . q{"Script Error"};
        tstlog ("-comment", "$teststring");
        eval $teststring;
        if ($bg =~ /nobg/) {
            my $tm1 = time() - $tm0;
            logcomment("Finished: $testid took ".$tm1." seconds."); 
        } else { 
            logcomment("Launched: $testid in background");
        } 
    }
}

## @name ::get_vol_status
## @description : get status of volume
## @usage get_vol_status($filer,$volname)
## returns (%status)
sub get_vol_status ($$)
{
    my ($filer,$vol) = @_;

    my %status;

    if ($vol eq '') {
       $vol = get_root_vol($filer);
    }
    my $fc = connect("$filer#rsh");
    my (undef, undef, $output) = say($fc, "vol status $vol");
    my ($vol_stat) = ($output =~ /($vol?.+\S+)/g);
    ($status{volume},$status{state},$status{status}) = split(/\s+/,$vol_stat);
    return %status;
}

## @name ::is_vol_useful
## @description checks volume specified to see if its present,
## online status and state normal/reconstruct
## @status Deprecated
sub is_vol_useful($$)
{
    my ($filer,$volume) = @_;

    logresult('CONFIG', "Volume: $volume does not exist on $filer")
        if (!is_vol_present($filer,$volume));

    my %status = get_vol_status($filer,$volume);
    logresult('CONFIG', "Volume: $volume on $filer is either not in normal state or not in online/reconstruct status")
        if ($status{state} !~ /online/);

        # removed this because with virtual volume code changes
        # the vol status is of no use
        #if (($status{state} !~ /online/) || ($status{status} !~ /(normal|reconstruct)/));
}


## @name ::is_vol_present
## @description : check presence or absence of volume on filer
## @usage is_vol_present($filer,$volname)
## returns 1 if present, 0 otherwise
## @status Deprecated
sub is_vol_present ($$)
{
    my ($filer,$volname) = @_;

        eval {

        die("Specify volume name") if ($volname eq '');
        my $fc = connect("$filer#rsh");
        my (undef, undef, $out) = say($fc, "vol status $volname");
        die($out) if ($out =~ /unrecognized command/);
        return(($out =~ /No volume named.*exists/) ? 0 : 1);
    } 
}

## @name ::get_root_vol
## @description : get name of root volume
## @usage get_root_vol($filer)
## returns $rootvol
## @status to be Deprecated
sub get_root_vol($)
{
    my ($filer) = @_;
    my $fc = connect("$filer#rsh");
    my (undef, undef, $output) = say($fc, "vol status");
    my @lines = split('\n',$output);		# split into lines
    foreach my $line (@lines) {
        my ($vol,$state,$status,$options) = split(" ",$line);
        return ($vol) if (defined($options) && ($options =~ /^root/));
    }
}

## @name ::is_cifs_share_present
## @description : check presence or absence of cifs share on filer
## @usage is_cifs_share_present($filer,$share)
## returns $rootvol
sub is_cifs_share_present($$)
{
    my ($filer,$share) = @_;

    my $fc = connect("$filer#rsh");
    my (undef, undef, $output) = say($fc, "cifs shares $share -a");
    return(($output =~ /There is no share named/) ? 0 : 1);
}

## @name ::get_vfiler_details
## @description : obtain vfiler params
# Parameters : 
#	filer
#	vfilername
# Returns:
#	Status
#	1 - if vfiler found
#	0 - if vfiler not found
# if found
#	IP Address
#	Vol name
#	Qtree name
#
# This routine is highly dependant on the UI - can be improved if
# there was a vfiler diag command providing all details
# Currenly the UI looks like the foll....here is a sample
#tada-u3> vfiler status -r vf1
#vf1                              running
#   ipspace: default-ipspace
#   IP address: 172.16.72.103 
#   Path: /vol/vol1/qtree1
#   UUID: 1bffb8ea-d0cb-11d5-a5ea-0002b32302c5
#
# This subroutine is tailored for additonal vfilers only - will not work for vfiler0
sub get_vfiler_details($$)
{
	my ($filer,$vfilerN) = @_;
    my $fc = connect("$filer#rsh");
	my (undef, undef, $output) = say($fc, "vfiler status $vfilerN");
	if ($output =~ /not found/) {
		return(0);
	} else {
		my (undef, undef, $output) = say($fc, "vfiler status -r $vfilerN");
		my @lines = split('\n',$output);		# split into lines

		# this script is HIGHLY dependant on the UI

		# parse the first line containing status of vfiler
		my (undef, $status) = split(" ",$lines[0]);
		# parse the third line for IP address
		my (undef, undef, $ip_addr) = split(" ",$lines[2]);
		# parse the fourth line for vol name and qtree name
		my (undef, undef, $vol, $qtree) = split("/",$lines[3]);

		return(1,$status,$ip_addr,$vol,$qtree);
	}
}

## @name ::vfiler_destroy
## @description : To destroy vfilers
# Parameters : 
#    For vfiler destroy:
#	               vfilername
#	               vfiler interface
#    For vfiler create:
#                      vfilername
#                      vfiler ipaddr
#                      vfiler  interface
#                      vfiler volume name
#                      vfiler qtree
# Returns:
#	Status
#	1 - if vfiler has been created/destroyed
#	0 - if vfiler creation/destruction has failed
#
# This routine is highly dependant on the UI - Infact for
# me it changed in a day
#
# The way to go abt it is:
#
# For destroying : 
#         -Bring the interface associated with it down
#         -Stop the vfiler
#         -destroy the vfiler 
#         -remove the qtree dir associated with it else there
#          are problems associated with creating new qtree dirs 
#         -unlicense the vfiler feature
#         -Pass the basic things as results back so that creation is possible
sub vfiler_destroy($$$)
{
    my $filer = host(shift);
    my $fc = connect("#rsh");
    my $vfname = shift;
    my $vfif = shift;
    
    tstlog("-comment", "$vfname");
    my @result = get_vfiler_details($filer, $vfname);
    
    if ($result[0] != 1) {
	return(0);
    }
    # The vfiler data has been obtained
    # Now let's destroy it 
    my ($vfstatus, $vf_ipaddr, $vf_vol, $vf_qtree)=($result[1], $result[2], $result[3], $result[4]);
    tstlog("-comment", "$vfstatus", "$vf_ipaddr", "$vf_vol", "$vf_qtree");
    if ($vfstatus =~ /run/) {
	my $output = say($fc, "priv set advanced;ifconfig $vfif down");
	(undef, undef, $output) = say($fc, "vfiler stop $vfname");
	if ($output !~ /stopped/) {
	    tstlog("-comment", "vfiler stop command failed\n $output");
	    return(0);
	}
    }
    my (undef, undef, $output) = say($fc, "vfiler destroy $vfname -f");
    if ($output !~ /moved/) {
	tstlog("-comment", "vfiler destroy command failed");
	return(0);
    }
    (undef, undef, $output) = say($fc, "priv set advanced; license vfiler=DISABLE");
    if ($output !~ /disabled/) {
	tstlog("-comment", "vfiler license disable failed");
	return(0);
    }
    
    tstlog("-comment","Vfiler $vfname successfully destroyed");
    return(1,$vf_ipaddr, $vf_vol, $vf_qtree);
}

## @name ::vfiler_create
# For creation :
#         -check for the vol status
#         -create the qtree
#         -Bring the interface down
#         -Create the vfiler
#         -Bring the interface up
sub vfiler_create($$$$$$)
{
    my $filer    = host(shift);
    my $fc       = connect("#rsh");
    my $vf_name  = shift;
    my $vf_ip    = shift;
    my $vf_if    = shift;
    my $vf_vol   = shift;
    my $vf_qtree = shift;
    
    my (undef, undef, $output) = say($fc, "priv set advanced;license vfiler=essrxff");
    if ($output !~ /enabled/)
    {
	tstlog("-comment", "License command failed");
	return(0);
    }
    (undef, undef, $output) = say($fc, "vol status $vf_vol");
    if ($output !~ /normal/i) {
	tstlog("-comment", "Did not find $vf_vol\n");
	return(0);
    }
    (undef, undef, $output) = say($fc, "qtree create /vol/$vf_vol/$vf_qtree");
    if (($output ne "") && ($output !~ /exists/)) { 
	tstlog("-comment","Qtree command failed\n");
	return(0);
    }
    say($fc, "priv set advanced; ifconfig $vf_if down");
    #  Hoping that the vf_if was brought down gracefully
    (undef, undef, $output) = say($fc, "vfiler create $vf_name -i $vf_ip /vol/$vf_vol/$vf_qtree");
    # to check if the vfiler was created we need to do some check
    (undef, undef, $output) = say($fc, "vfiler status $vf_name");
    if ($output =~ /not found/) {
	tstlog("-comment","Vfiler creation command has failed");
	return(0);
    }
    say($fc, "priv set advanced; ifconfig $vf_if up");
    #reached here then success
    tstlog("-comment","Vfiler $vf_name succesfuly created");
    
    return(1);
}


## @name ::do_shell
## @description :Run the shell command in a more sane way
sub do_shell ( $ )
{
    my ($cmd) = @_;
    logcomment("running command: $cmd");
    my $out = `$cmd 2>&1`;
    my @msg;
    if ($?) {
	push @msg, ("command exit: " . decode_result($?));
    }
    if (length($out)) {
	push @msg, ("command output:",$out);
    }
    if (@msg) {
	logcomment(@msg);
    }
    return(@msg);
}

## @name ::append_unique_line
## @usage append_unique_line($filename,$line)
## @description :append this line to this file, but only append if it's not already there.
## - sharad@netapp.com
sub append_unique_line($$)
{
    my ($filename, $line) = @_;

    my $fh = new FileHandle($filename,O_RDONLY) 
        or logresult('FATAL',"Could not open file ${filename} for read: $!");
    my @data = <$fh>;
    $fh->close() 
        or logresult('FATAL',"Error closing file ${filename}: $!");
    
    my $match = Tharn::string2regex($line);
    if (grep { /^$match$/ } @data) {
        # already there, do nothing
    } else {
        $fh = new FileHandle($filename,O_WRONLY|O_APPEND) 
            or logresult('FATAL',"Could not open file ${filename} for write: $!");
        syswrite($fh,$line."\n",length($line)+1,0) 
            or logresult('FATAL',"Error writing to file ${filename}: $!");
        $fh->close() 
            or logresult('FATAL',"Error closing file ${filename}: $!");
    }
}


## @name ::decode_result
sub decode_result ( $ )
{
    my ($result) = @_;
    my $status = $result>>8;
    my $signal = $result&127;
    my $core = $result&128;
    return ($signal ? "signal $signal" : "status $status").
	(($core) ? " (core dumped)" : "");
}

## @name ::contains
## usage: contains($item, @list)
## Returns whether the given items is in the given list
sub contains ($@) {
    my $item = shift @_;
    return(scalar(grep { $_ eq $item } @_));
}

## @name ::df_util
## @usage df_util($mntpoint)
## @description : gets the disk utilisation for a supplied mountpoint
##                on the localhost. This routine does a uname -a to get
##                OS type so that correct system call is made. Thus it can
##                give df stats only for fs mounted on the localhost. The 
##                function returns an array whose first element is the number 
##                of free blocks and the second element is number of free inodes.
##                Any user must take care of both to ascertain whether any further
##                file creation is possible or not.
## @status to be deprecated
sub df_util($) {
    my ($path) = @_;
    eval {
        my $ostype = `uname -a`;
        if ($ostype =~ /Sun/) {
            my @array1 = 1..15;
            my @array2 = "c"x16;
            my @array3 = 1..2;
            my @array4 = "c"x32;
            my @array5 = 1..16;
            my $buf = pack("L15A16L2A32L16", @array1, @array2, @array3, @array4, @array5); 
            # buf has been packed now let's syscall 
            syscall(218, $path, $buf);
            my @array = unpack("L15A16L2A32L16", $buf);
            return($array[7], $array[13]);
        } elsif ($ostype =~ /Linux/) {
            my @array1 = 1..16;
            my $buf = pack("L16", @array1);
            syscall(99, $path, $buf);
            my @array = unpack("L16", $buf);
            return($array[4], $array[6]);
        } else {
            die "do not know how to df for $ostype\n";
        }
    }
}

## @name ::get_traffic($filer,$protocol
## @usage get_traffic($filer,$protocol
##         [,count=>]
##         [,interval=>])
## @returns %ops
sub get_traffic($$%)
{
    my ($filer,$protocol,%args) = @_;
    eval {
    my $fc = connect("$filer#rsh");

    my $count = defined($args{count}) ? $args{count} : 10;
    my $interval = defined($args{interval}) ? $args{interval} : 1;

    my %column = ( # 0 indexed
         'nfs'   => 1,
         'cifs'  => 2,
         'http'  => 3,
         'dafs'  => 16,
         'fcp'   => 17,
         'iscsi' => 18,
    );

    my (undef, undef, $out) = say($fc,"sysstat -x -c $count -s $interval");
    my $skip = $column{"$protocol"};

    my %ops;
    (undef, $ops{min}) = ($out =~ /Min.*(\s+\S+){$skip}\s+(\S+)/gm);
    (undef, $ops{avg}) = ($out =~ /Avg.*(\s+\S+){$skip}\s+(\S+)/gm);
    (undef, $ops{max}) = ($out =~ /Max.*(\s+\S+){$skip}\s+(\S+)/gm);

    return(%ops);
    };
}

## @name ::get_fcp_adapter_info
## @usage get_fcp_adapter_info($filer,$adapter)
## @returns ($state,$standby,$nodename,$portname)
sub get_fcp_adapter_info($$)
{
    my ($filer,$adapter) = @_;

    eval {
    my $fc = connect("$filer#rsh");

    my (undef, undef, $out) = say($fc, "fcp show adapter $adapter");
    die("FCP not licensed on $filer") if ($out =~ /not licensed/);
    die($out) if ($out =~ /not found/i);

    my ($status, $standby, $nodename, $portname);
    foreach my $line (split(/\r\n/, $out)) {
        if ($line =~ /^Status:\s+(.*)/)  { $status=$1;}
        if ($line =~ /^Standby:\s+(.*)/) { $standby = $1;}
        if ($line =~ /^FC Nodename:\s+\S+\s+\((\w+)\)/) { $nodename = $1;}
        if ($line =~ /^FC Portname:\s+\S+\s+\((\w+)\)/) { $portname = $1;}
    }

    logcomment("FCP Adapter $adapter - STATUS:$status STANDBY:$standby".
                " NODENAME:$nodename PORTNAME:$portname");
    return($status,$standby,$nodename,$portname);

    }
}

# Get igroup information
sub get_igroup_info($$) 
{
    my ($filer,$igroup) = @_;

    eval {
    my @list;
    my $fc = connect("$filer#rsh");

    my (undef,undef,$out) = say($fc, "igroup show $igroup");
    die($out) if ($out =~ /not found|No such.*exists/i);

    while ($out =~ /(\w\w\:\w\w:\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)\s+\((.*)\)/gm) {
        my ($nodename, $status) = ($1, $2);
        if (defined($nodename) && defined($status)) {
            push(@list, ($nodename,$status));
        }
    }
    return(@list);
    };
}

1;
