# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/chain_db.pm#1 $

## @summary persistent separate-chain hash
## @author remi@netapp.com
## @status public
## @pod here

=head1 NAME

chain_db.pm - a disk-backed chain database

=head1 DESCRIPTION

This package provides a disk-backed hash, which allows you to store a
chain of values under each hash key (rather than just a single value).
Such a chain can contain any number of scalar values.  When retrieving
values, the caller can either get the entire chain of values for a
specific key, or else needs to provide a routine to identify which of
the values should be returned.

A chain database is disk-backed by using the C<DB_File> package.
Normally, it exists in a temporary file which is deleted when the
program exits.  Optionally, the backing file can be saved for later
use.  If C<DB_File> is not present, then the chain database will only
exist in-memory, and there is no way to save the data in a file.

For convenience, a chain database can store objects of one specific
type, instead of just scalar values, provided you give it a little
more information about how to serialize and rebuild the those objects.
This is determined when the chain database is constructed, and cannot
be changed later.  On constructed this way, the chain database will
assume that all values stored and retrieved are objects of the
specific type.

=head1 FUNCTIONS

=cut

package chain_db;
use warnings;
use strict;

use Carp;
use Fcntl qw(:DEFAULT); # O_*
use File::Temp qw(tempfile);
use DataManip::cmd_utils qw(serialize);

# Keys for internal hash.
use constant db => __PACKAGE__ . "::hashref";
use constant file => __PACKAGE__ . "::filename";
use constant obj => __PACKAGE__ . "::object_class";
use constant serializer => __PACKAGE__ . "::serializer";
use constant constructor => __PACKAGE__ . "::constructor";

my $DISK_BACKED;

BEGIN {
  eval {
    require DB_File;
    import DB_File qw(:DEFAULT);
  };
  if ($@) {
    if ($@ =~ /\ACan\'t locate DB_File.pm in \@INC/) {
      $DISK_BACKED = 0;
      warn "DB_File not found, chain_db will not be disk-backed.";
    } else {
      die $@;
    }
  } else {
    $DISK_BACKED = 1;
  }
}

=head2 new(%options)

The constructor for a chain database just takes a set of options.

Options:

=over 4

=item FILE

Path to a file to use or create for the chain database.  If this
option is provided, then the backing file will not be automatically
deleted.

=item OBJECT_CLASS

Package name for objects to be stored and retrieved from the chain
database, if the chain database is to use objects of some specific
type instead of scalar values.

=item OBJECT_PACKAGE

Alternate package name for a file in which OBJECT_CLASS is defined
(in case the package OBJECT_CLASS is in a file of a different name).

=item SERIALIZER

Name of the method in OBJECT_CLASS which should be called on objects
to obtain serialized versions of the objects when storing them.
(default=C<serialize>)

=item CONSTRUCTOR

Name of the method in OBJECT_CLASS which should be given serialized
versions of objects in order to rebuild the original objects when
retrieving them.  (default=C<new>)

=back

=cut

sub new {
  my $class = shift;
  my %db = ();
  my %opts = @_;

  my $self = {};
  $self->{&db} = \%db;
  $self->{&serializer} = (exists $opts{SERIALIZER} ? $opts{SERIALIZER} : 'serialize');
  $self->{&constructor} = (exists $opts{CONSTRUCTOR} ? $opts{CONSTRUCTOR} : 'new');
  $self->{&obj} = undef;

  if (exists $opts{OBJECT_CLASS}) {
    $self->{&obj} = $opts{OBJECT_CLASS};

    # Require the named package.
    my $package = $self->{&obj};
    if (exists $opts{OBJECT_PACKAGE}) {
      $package = $opts{OBJECT_PACKAGE};
    }
    eval "require $package;";
    croak $@ if ($@);

    # Do some primitive checking to make sure the named methods exist
    # (even if they're not functions).
    foreach (serializer, constructor) {
      my $sym = $self->{&obj} . "::" . $self->{$_};
      no strict 'refs';
      unless (defined(*$sym)) {
	croak "$sym is not defined\n";
      }
      use strict 'refs';
    }
  }

  if ($DISK_BACKED) {
    # Setup the disk-backed hash.
    if ($opts{FILE}) {
      $self->{&file} = $opts{FILE};
    } else {
      my ($fh, $file) = tempfile(UNLINK => 1);
      unless ($fh && $file) {
	croak "cannot open temporary database file: $!\n";
      }
      $self->{&file} = $file;
      close $fh or croak "$!\n";
    }
    tie(%db, 'DB_File', $self->{&file}, O_RDWR);
  } else {
    # Just use an in-core hash.
    $self->{&file} = undef;
  }

  return bless($self, $class);
}

# This is probably overkill, but it shouldn't hurt.
sub DESTROY {
  my $self = shift;

  if ($DISK_BACKED) {
    untie %{$self->{&db}};
    $self->{&db} = undef;
  }
}

=head2 backing_file

Retrieves the path to the backing file for the chain database,
or returns undef if the chain database is not disk-backed.

=cut

sub backing_file {
  my $self = shift;
  return $self->{&file};
}

=head2 each

Iterates through all the keys in the chain database.

=cut

sub each {
  my $self = shift;
  return scalar(each %{$self->{&db}});
}

=head2 keys

Returns a list of all the keys currently stored in the
chain database.

=cut

sub keys {
  my $self = shift;
  return keys %{$self->{&db}};
}

=head2 store($key, @values)

Stores one or more new values in the chain database, under the given
key.  Possibly serializes each new value first, if it is an
OBJECT_CLASS object.

=cut

sub store( $$$ ) {
  my $self = shift;
  my ($key, @new_values) = @_;
  unless (defined $key) {
    croak "no key argument provided\n";
  }

  # Get the current values stored under this key.
  my @values;
  if (exists $self->{&db}->{$key}) {
    @values = eval $self->{&db}->{$key};
    croak $@ if ($@);
  } else {
    @values = ();
  }

  # Add on the new values.
  if ($self->{&obj}) {
    no strict 'refs';
    push @values, map { &{$self->{&obj}.'::'.$self->{&serializer}}($_) } @new_values;
    use strict 'refs';
  } else {
    push @values, @new_values;
  }

  # Store the new list of values.
  $self->{&db}->{$key} = serialize(@values);
}

=head2 lookup($key, $which, @args)

=head2 remove($key, $which, @args)

Locates or removes values from the chain database.  The located
or removed values are returned.

When just the $key is specified, returns all the values stored under
that key.  Otherwise, $which is a routine which will be used to
determine which value to return.  The routine will be called on each
value, with @args being passed as additional arguments to the routine.
The first value for which the routine returns true is returned.

=cut

foreach my $function (qw(lookup remove)) {
  my $method = sub {
    my $self = shift;
    my ($key, $wanted_func, @args) = @_;
    unless (defined $key) {
      croak "no key argument provided\n";
    }

    # Get all the values for this key.
    unless (exists $self->{&db}->{$key}) {
      return (wantarray ? () : undef);
    }

    my @values = eval $self->{&db}->{$key};
    croak $@ if ($@);

    # Return all the values if we weren't given a way
    # to pick the right one.
    unless (defined $wanted_func) {
      if ($function eq 'remove') {
	delete $self->{&db}->{$key};
      }
      if ($self->{&obj}) {
	no strict 'refs';
	my @objs = map {
	  &{$self->{&obj}.'::'.$self->{&constructor}}($self->{&obj}, $_);
	} @values;
	use strict 'refs';
	return @objs;
      } else {
	return @values;
      }
    }

    # Run through all the values and call $wanted_func on each
    # until we find the right one to return.
    for (my $i = 0; $i < scalar(@values); $i++) {
      my $entry;
      if ($self->{&obj}) {
	no strict 'refs';
	$entry = &{$self->{&obj}.'::'.$self->{&constructor}}($self->{&obj}, $values[$i]);
	use strict 'refs';
      } else {
	$entry = $values[$i];
      }
      if ($wanted_func->($entry, @args)) {
	# Value at this offset matches.
	if ($function eq 'remove') {
	  splice(@values, $i, 1);
	  $self->{&db}->{$key} = serialize(@values);
	}
	return $entry;
      }
    }

    # None of the values matched.
    return (wantarray ? () : undef);
  };

  no strict 'refs';
  *$function = $method;
  use strict 'refs';
}


=head1 BUGS

Arguably, C<lookup> and C<remove> should behave more like C<grep>,
in that they should return all the values for which the function
returns true, instead of just the first.

=cut


1;
