# $Id$

## @summary Library to reserve clients for CIFS sanity tests.
## @author mueller@netapp.com

package cifs_client_reservation;
use strict;

BEGIN {
  use Tharn;
  use List::Util 'shuffle';
  use Module::Common;
  Module::Common::setup_export_tags(
    ops => [ qw(reserve_client free_client) ]
    );
}

my $basedir = '/x/eng/localtest/arch/win32/bin/cifs/sanity';

# There are errors that can kill the sanity test without giving it the
# ability to properly clean-up. While these are rare, we know that any
# clients that are still reserved after 45 minutes are no longer being
# used, so it's safe to clean them up.
sub clean_old_reservations
{
	opendir(DIR, "$basedir/in-use") or return;

	while (my $client = readdir(DIR)) {
		if ($client =~ /^(\.|\.\.)/) {
			next;
		}
		my $reservation_age = (time() - (stat("$basedir/in-use/$client"))[9]);
		if ($reservation_age > (60*45)) {
			logcomment("Expiring old reservation for $client ($reservation_age seconds old)");
			unlink("$basedir/in-use/$client") or logcomment("unlink $basedir/in-use/$client: $!");
		}
	}

	closedir(DIR);
}

sub reserve_client
{
	my $wait_seconds = 60;
	my $client;
	my @client_list;

	clean_old_reservations();

	opendir(DIR, "$basedir/clients") or return '';
	while (defined($client = readdir(DIR))) {
		push @client_list, $client;
	}
	closedir(DIR);

	@client_list = shuffle(@client_list);

	while ($wait_seconds != 0)
	{
		foreach $client (@client_list) {
			if (!(-e "$basedir/in-use/$client")) {
				touch("$basedir/in-use/$client");
				logcomment("Windows client $client reserved.");
				return $client;
			}
		}

		$wait_seconds--;
		sleep(1);
	}
}

sub free_client
{
	my ($client) = shift;
	$client =~ s/;.*//;
	if (!(-e "$basedir/in-use/$client")) {
		logcomment("Unable to locate reserved client $client in $basedir/in-use!");
	} else {
		unlink("$basedir/in-use/$client");
		logcomment("Windows client $client has been returned.");
	}
}

sub touch
{
	my $now = time;
	local (*TMP);
	foreach my $file (@_) {
		utime ($now, $now, $file)
		|| open (TMP, ">>$file")
		|| warn ("Couldn't touch file: $!\n");
	}
}

1;
