# $Id$

package cluster;

## @summary Library of routines to boot clusters (CFO/SFO pairs) and query their states
##
## @synopsis
##   use cluster;
##
##   boot_with_giveback($filerA, $filerB, ...options...)
##   get_cf_state($filer[, $rate]);
##   wait_state($filer, $want_state[, $timeout[, $rate]]);
##   cf_enable($filer);
##   cf_takeover($filer,$dasheff);
##   cf_giveback($filer,[boot=&gt;0],[partner=&gt;$filerB]);
##   cf_disable($filer);
##
## @example
##
##   # a simple repeated takeover and giveback script
##
##   use cluster;
##   param('FILERA','-mesg',"hostspec for one filer");
##   param('FILERB','-mesg',"hostspec for its partner");
##   param('TIMEOUT','-default',180);
##   param('SLEEP_BETWEEN','-default',60);
##
##   my $filerA = host($FILERA);
##   my $filerB = host($FILERB);
##   boot_with_giveback($filerA,$filerB) or tstlog('-fatal',$@);
##   cf_enable($filerA) or tstlog('-fatal',$@);
##   cf_enable($filerB) or tstlog('-fatal',$@);
##
##   while (1) {
##       cf_takeover($filerA) or tstlog('-fail',$@);
##       wait_state($filerA,"TAKEOVER",$TIMEOUT) or tstlog('-fail',$@);
##       wait_state($filerB,"DEAD",$TIMEOUT) or tstlog('-fail',$@);
##       sleep($SLEEP_BETWEEN);
##       cf_giveback($filerA,partner=&gt;$filerB) or tstlog('-fail',$@);
##       wait_state($filerA,"UP",$TIMEOUT) or tstlog('-fail',$@);
##       wait_state($filerB,"UP",$TIMEOUT) or tstlog('-fail',$@);
##       sleep($SLEEP_BETWEEN);
##   }
##
## @usage
##
## <HR><center><b>boot_with_giveback</b></center>
##
## <tt><b>boot_with_giveback($filerA, $filerB, ... boot options ...)</b></tt>
##
## Attempt to boot $filerA and $filerB to the normal filer prompt if
## they are not already.  If this is not possible due to one node
## having taken over the other, then ask it to give back, and attempt
## to boot again.
##
## This uses both console (for booting) and rsh (for determining
## current state).
##
## This returns 1 on success or undef on failure (with $@ set to the error)
##
## <b>$filerA,$filerB</b> - host records or connection records for two
## filers which are a CFO pair.
##
## <b>... boot options ...</b> - further options will be passed to
## boot() when booting.  See the boot library (lib/boot.pm) for more
## information.
##

## <HR><center><b>get_cf_state</b></center>
##
## <tt><b>get_cf_state($filer[, $rate])</b></tt>
##
## Use "cf status" and "cf monitor" to determine the state of $filer
## with respect to clustering.
##
## return "UP" if CFO is enabled on filer but it has not taken over or
## been taken over, "TAKEOVER" if a filer has taken over, "DEAD" if a
## filer has been taken over, "DISABLED" if CFO is disabled on this
## host, or "UNREACH" if we cannot successfully talk to $filer.
##
## If takeover or giveback are in progress, this command waits (tries
## again repeatedly).
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$rate</b> - how many seconds to wait in between each poll for the
## current state.  The default is 10 seconds.

## <HR><center><b>wait_state</b></center>
##
## <tt><b>wait_state($filer, $want_state[, $timeout[, $rate]])</b></tt>
##
## Wait for the current state of $filer (according to get_cf_state) to
## match $want_state.
##
## In a scalar context, return 1 if the wanted state was reached, or
## undef if not (in which case $@ will contain an appropriate error
## message that wait_state timed out waiting)
##
## In an array context, return the aforementioned as the first value,
## and the current state at the time wait_state finished as the second
## value.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$want_state</b> - one of UP, TAKEOVER, DEAD, DISABLED, or
## UNREACH, as for get_cf_state. The actual match is by regex, so you
## can specify more than one state (for example, "DISABLED|UP").
##
## <b>$timeout</b> - how many seconds (or an absolute time_t value) to
## wait for the wanted state to be reached before giving up with an
## error.  The default is -1 (infinite).
##
## <b>$rate</b> - how many seconds to wait in between each poll for the
## current state.  The default is 10 seconds.

## <HR><center><b>cf_enable</b></center>
##
## <tt><b>cf_enable($filer)</b></tt>
##
## Issue "cf enable" to $filer.
##
## Return 1 if the response suggests CFO is now enabled or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).

## <HR><center><b>cf_disable</b></center>
##
## <tt><b>cf_disable($filer)</b></tt>
##
## Issue "cf disable" to $filer.
##
## Return 1 if the response suggests CFO is now disabled or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).

## <HR><center><b>cf_takeover</b></center>
##
## <tt><b>cf_takeover($filer,[$cmd])</b></tt>
##
## Issue "cf takeover -f" (or other cf takeover command) to $filer
##
## Return 1 if the response suggests takeover has started, or undef
## otherwise, with an appropriate error in $@.
##
## <b>$filer</b> - host record or connection record for this filer (if
## a host record, then the connection #rsh will be used).
##
## <b>$cmd</b> - what command to use for takeover.  Defaults to "cf
## takeover -f", but could be changed to something else (e.g. boring,
## negotiated "cf takeover").

## <HR><center><b>cf_giveback</b></center>
##
## <tt><b>cf_giveback($filer, [boot=&gt;0], [partner=&gt;$filerB])</b></tt>
##

## Issue "cf giveback -f" to a filer, and possibly help its partner to
## reboot.

## <b>$filer</b> - host record or connection record of filer that will
## give back (if a host record, then the connection #rsh will be
## used).

## <b>boot</b> - whether to help $filer's partner to boot.  The
## default is 0 for non-simulators (which generally need no
## assistance) and 1 for simulators (which generally need assistance).

## <b>partner</b> - the argument or arguments to boot() to use when
## booting the partner (see boot() in lib/boot.pm).  Required if boot
## is 1, ignored if boot is 0.

## @bugs 
##   Both the boot library and this library use "UP" to mean slightly
##   different things ("UP" in a CFO and failover monitor sense is
##   that a node has not taken over or been taken over, but is enabled
##   to do so if need be.  "UP" for the boot library means at the
##   normal filer prompt)

## @see-also lib/boot.pm

## @author dl-central-libs
## @change amal@netapp.com added cmode support for cf_enabled 
package cluster;

use strict;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('boot_with_giveback','boot_with_giveback_one','get_cf_state','wait_state','wait_fw_state',
	   'cf_enable','cf_disable','cf_takeover','cf_giveback','ic_connected','disks_on_adapter','cf_enabled','cf_hwassist_precheck','cf_status_precheck','cf_options_precheck','cf_hwassist_stats_precheck','setup_partner' , 'set_ha_mode' );

use Tharn;
use boot;
use Utils qw(make_conn);
use SayWrap qw(sayEOF);
use CLI::FileIO qw(rdfile wrfile);
use Storage::Misc qw(filer_cmd);


# given a host or connection, return a connection (when just a host is
# given, favor whatever is given in the 'favor' option or #rsh if not
# specified).

sub make_connection ( $@ ) {
    my ($spec, %opts) = @_;	
    if (ref($spec) =~ /Connectrec/) {
	return $spec;
    } elsif (ref($spec) =~ /Hostrec/) {
	my $conn = $opts{'favor'} || "#rsh";
	return connect($spec->{id} . $conn);
    } else {
	my $host = host($spec,'-ok');
	if ($host) {
	    my $conn = $opts{'favor'} || "#rsh";
	    return connect($spec . $conn);
	} else {
	    return connect($spec);
	}
    }
}

# given a host or connection, return a host.

sub make_host ( $ ) {
    my ($spec) = @_;	
    if (ref($spec) =~ /Connectrec/) {
	return $spec->{hostp};
    } else {
	return host($spec);
    }
}

# Bring both filers up (if they're not already), and make sure any
# which are in takeover state have given back to each other.

sub boot_with_giveback ( $$@ ) {
    my ($filerA, $filerB, @options) = @_;
    my @filers = ($filerA, $filerB);

    return eval {
	my @hosts = map { make_host($_); } @filers;
	my @sobj;
	foreach my $h (@hosts) {
	    my $id = $h->{id};
	    push (@sobj,
		  subtest(sub { boot_with_giveback_one($h,\&boot,@options)
				    or tstlog('-fail',$@); }, 
			  '-runid', "boot-$id",
			  '-bg',
			  '-iffatal','-ok','-ifscript','-ok'
			 )
		 );
	}
        Subtest::waitFinish(subtests => [@sobj]);

	foreach my $test (@sobj) {
	    my $result = Rslt::worst($test->rslts());
	    next if !$result;
	    my $runid = $test->runid();
	    # (Want to error for CONFIG or worse definitely, and I
	    # think INSPECT and NA too, but not WARN)
	    if ($result->severity() > NATE::Result::Warn->severity()) {
		die "Boot failed in subtest \"$runid\": "
                    . join("\n",$result->type,$result->msgs) . "\n";
	    }
	}
	return 1;
    };
}

# Do the work of boot_with_giveback on just one filer (each boots up
# and waits for the other to give back, then if it boots up all the
# way it is told to give back.)

# (needed by Filer/boot.thpl, which is a test that performs what
# boot_with_giveback() does, but with a slightly different interface)

sub boot_with_giveback_one {
    my ($host, $boot_sub, @options) = @_;
    return eval {
	host($host);
	my $c = connect("#console");
	my $r = connect("#rsh");
	$boot_sub->($c,iftakenover=>'wait',@options) or die $@;
	my $try = 0;
	while (get_cf_state($r) eq 'TAKEOVER') {
	    say($r,"cf giveback -f",'-eval_error');
	    die $@ if $@;
	    if ($try > 10) {
		die("Tried giveback via $r->{id} too many times ($try)\n");
	    }
	}
	return 1;
    };
}

# State will either be "UP", "TAKEOVER", or "DEAD" (a subset of the
# strings that "cf monitor" will print) or "UNREACH" if we can't seem
# to talk to a filer to find out its state, or "DISABLED" if
# clustering is disabled.

# If a takeover is in progress or a giveback is in progress, this
# command will keep retrying.

# About what the states mean: We don't accept "DOWN" and "REBOOTING"
# as states from cf monitor, for some reason.  This makes sense to me
# for REBOOTING (we'll keep trying, and it sounds like a transitory
# state) but not DOWN, but perhaps 'cf monitor' can never print DOWN
# since maybe DOWN implies the filer is halted and not taken over and
# hence neither filer has that IP address to be able to reply to rsh
# of "cf monitor".  "DEAD" might be more appropriately termed "UNDEAD"
# or "TAKEN OVER": someone at this filer's IP address was alive enough
# to tell us that this filer is dead, so it must be that the partner
# has taken over.

sub get_cf_state ( $;$ ) {
    my ($filer,$rate) = @_;
    $rate = $rate || 10;
    my $conn = make_connection($filer);

    my $cur_state;

    while (1) {
	my $retstr;

	(undef, undef, $retstr) = 
	    say($conn, "cf status",'-eval_error');
	return 'UNREACH' if ($@);
        return 'NON_HAMODE' if ($retstr =~ /Node is in non_ha mode/ );
        return 'NO_LICENSE' if ($retstr =~ /Controller failover must be licensed/ );
	return 'DISABLED' if ($retstr =~ /Cluster disabled/);
	return 'DISABLED' if ($retstr =~ /Controller Failover disabled/);
	return 'DISABLED' if ($retstr =~ /takeover disabled/);
	return 'DISABLED' if ($retstr =~ /has disabled takeover by/);

	last if (length($retstr) &&
		 ($retstr !~ /attempting takeover/) &&
		 ($retstr !~ /giveback in progress/));

	sleep $rate;
    }

    my $out;
    do {
	(undef, undef, $out) = 
	    say($conn,"cf monitor",'-eval_error');
	if ($out =~ /(UP|TAKEOVER|DEAD)/) {
	    $cur_state = $1;
	}
    } while ($out =~ /Try again/ && !defined($cur_state));

    if (!defined($cur_state)) {
	return 'UNREACH';
    } else {
	return $cur_state;
    }
}

# wait for up to $timeout seconds ($timeout being undefined means
# infinite) for $filer to be in state $target_state.  Whether or not
# it does so, return the last state seen

sub wait_state ( $$;$$ ) {
    my ($filer, $want_state, $timeout, $rate) = @_;
    my $cur_state;
    my $res = eval {
	my $id = make_host($filer)->{id};
	my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
	my $rate = $rate || 10;
	while (time() <= $end_time) {
	    $cur_state = get_cf_state($filer,$rate);
            return 1 if $cur_state =~ m/$want_state/x;
	    sleep $rate;
	};
	die "Timed out waiting for state \"$want_state\", current state is \"$cur_state\"\n";
    };
    return wantarray ? ($res, $cur_state) : $res;
}

# wait for up to $timeout seconds ($timeout being undefined means
# infinite) for $filer to be in desired fw state.  Whether or not
# it does so, return the last state seen
# The last var $partner means tells the function whether to check its 
# own firmware state or it's partner's fw state. This is because 
# we cannot read the fw state of a node that is down by doing a cf 
# readfw on itself. So we have to go thru its partner.

sub wait_fw_state ( $$;$$$ ) {
    my ($filer, $want_state, $timeout, $rate, $partner) = @_;
    my $cur_state;
    my $res = eval {
	my $id = make_host($filer)->{id};
    my $conn = make_connection($filer);
	my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
	my $rate = $rate || 10;
	while (time() <= $end_time) {
	    (undef, undef, $cur_state) = say($conn,"priv set -q test; cf readfw","-eval_error");
        if ($partner) {
            ($cur_state) = ($cur_state =~ /PARTNER: state (SF[A-Z_]*), progress/);
        } else {
            ($cur_state) = ($cur_state =~ /LOCAL: state (SF[A-Z_]*), progress/);
        }
	    last if ($cur_state eq $want_state);
	    sleep $rate;
	};
	if ($cur_state ne $want_state) {
	    die "Timed out waiting for state \"$want_state\", current state is \"$cur_state\"\n";
	}
	return 1;
    };
    return wantarray ? ($res, $cur_state) : $res;
}

# check for filer sanity by taking an adapter number as input
# and checking if disks exist on that adapter or not
sub disks_on_adapter ( $$$ ) {
	my ($filer, $adapter, $timeout) = @_;
	#tstlog('-comment', "filer=$filer adapter=$adapter timeout=$timeout");
	my $conn = make_connection($filer);
	my $end_time = $timeout ? Tharn::timeout2time($timeout) : Tharn::timeout2time(-1);
	my $rate = 10;	
	my $count = 0;
	my $out;
	while (time() <= $end_time) {
		(undef, undef, $out) = say($conn,"priv set -q advanced;fcadmin device_map $adapter","-eval_error");
	        if ($out =~ /Shelf Unknown: [0-9 X]+/) {
                                tstlog('-comment',"disk adapter $adapter back (but SES not) in $count tries");
                                return 1;
                }
		if ($out =~ /Shelf \d+: [0-9 X]+/) {
			tstlog('-comment',"disk adapter $adapter back in $count tries");
			return 1;
		}
		sleep($rate);
		$count = $count + 1;
	}
	tstlog('-comment',"disk adapter $adapter not back in $count tries");
	#die "Timed out but did not see expected disks on adapter\n";
}

# check whether IC is connected or not. Params are 2 filers, primary timeout1. If that 
# timeout triggers, we reset the NIC and try again for another timeout2 period.
sub ic_connected($$$$) {
	my ($filer1, $filer2, $timeout1, $timeout2) = @_;
	my $conn1 = make_connection($filer1);
	my $conn2 = make_connection($filer2);
	my $end_time = $timeout1 ? Tharn::timeout2time($timeout1) : Tharn::timeout2time(-1);
	my $rate = 10;
	my $count = 0;
	my $out1;
	my $out2;
 	# skip all this if in takeover state. (this probably doesn't work if we are currently in the middle
	# of a takeover...)
	(undef, undef, $out1) = say($conn1,"cf status","-eval_error");
        if ($out1 =~ /taken over/) {
                  tstlog('-comment',"IC not back - in takeover state");
                  return 1;
	}
	while (time() <= $end_time) {
		(undef, undef, $out1) = say($conn1,"priv set -q advanced;ic stats error","-eval_error");
		(undef, undef, $out2) = say($conn2,"priv set -q advanced;ic stats error","-eval_error");
		# only one side has to say "CONNECTED" for things to be ok. (right?) Hell, no. BOTH
		if ($out1 =~ /RV connection state :\s+CONNECTED/) {
			 if ($out2 =~ /RV connection state :\s+CONNECTED/) {
               		        tstlog('-comment',"IC back in $count tries");
                        	return 1;
               		 }
		}
		sleep($rate);
		$count = $count + 1;
	}
	#things are hosed.
	tstlog('-comment',"IC not back in $count tries");
	say($conn1,"priv set -q test; ic dump hw_stats","-eval_error");
	say($conn2,"priv set -q test; ic dump hw_stats","-eval_error");
	say($conn1,"priv set -q test; ic reset nic","-eval_error");
	say($conn2,"priv set -q test; ic reset nic","-eval_error");
        
	#try again after a reset. Really, this shouldn't be here, because these routines
	# should just look at status, but we need to know is reset does any good....
	$end_time = $timeout2 ? Tharn::timeout2time($timeout2) : Tharn::timeout2time(-1);
	while (time() <= $end_time) {
                (undef, undef, $out1) = say($conn1,"priv set -q advanced;ic stats error","-eval_error");
                (undef, undef, $out2) = say($conn2,"priv set -q advanced;ic stats error","-eval_error");
                if ($out1 =~ /RV connection state :\s+CONNECTED/) {
                        if ($out2 =~ /RV connection state :\s+CONNECTED/) {
                                tstlog('-comment',"IC back after resets in $count tries");
                                return 1;
                	}
		}
                sleep($rate);
                $count = $count + 1;
        }

	tstlog('-comment',"IC still not back in $count tries");
	#die "Timed out while expecting to see IC CONNECTED\n";
}

#check whether clustering is enabled correctly after a NIC- or disk-destroying
# event. Enabled is good. Disabled is good if originally
# induced by cf disable; and form of disabling due to unsync log
# or MB disks or some such is bad. 
sub cf_enabled($$$$) {
        my ($filer1, $filer2, $timeout1, $timeout2) = @_;
        #my $conn1 = make_connection($filer1);
        my $host_rec1 = host($filer1);
        #my $conn2 = make_connection($filer2);
        my $host_rec2 = host($filer2);
        my $end_time = $timeout1 ? Tharn::timeout2time($timeout1) : Tharn::timeout2time(-1);
        my $rate = 10;
        my $count = 0;
        my $out;
        while (time() <= $end_time) {
                #(undef, undef, $out) = say($conn1,"cf status","-eval_error");
                $out = filer_cmd(host=>$host_rec1, cmd=>"cf status");
                # look for cf enabled  on filer1
                if ($out =~ /Cluster enabled|Controller Failover enabled/) {
                        tstlog('-comment',"clustering on A back in $count tries");
                        return 1;
                }
                # look for has taken over or has been taken over
                if ($out =~ /taken over/) {
                        tstlog('-comment',"clustering not back - in takeover state");
                        return 1;
                }
		if ($out =~ /Cluster disabled|Controller Failover disabled/) {
                        tstlog('-comment',"clustering not back on A - disabled");
                        return 1;
                }
		sleep($rate);
                $count = $count + 1;
        }
	
        # this should not be necessary, but just in case...
	$end_time = $timeout2 ? Tharn::timeout2time($timeout2) : Tharn::timeout2time(-1);
	while (time() <= $end_time) {
                #(undef, undef, $out) = say($conn2,"cf status","-eval_error");
                $out = filer_cmd(host=>$host_rec2, cmd=>"cf status");
                # look for cf enabled on filer2
                if ($out =~ /Cluster enabled|Controller Failover enabled/) {
                        tstlog('-comment',"clustering on B back in $count tries");
                        return 1;
                }
                if ($out =~ /Cluster disabled|Controller Failover disabled/) {
                        tstlog('-comment',"clustering on B disabled");
                        return 1;
                }
                sleep($rate);
                $count = $count + 1;
        }

	tstlog('-comment',"clustering still not back in $count tries");
	#die...
}

# (arbitrarily decide it takes 10 seconds max cf enable to fully take
# effect)
use constant CF_ENABLE_TIMELIMIT => 10;

sub cf_enable ( $ ) {
    my ($filer) = @_;
    my $conn = make_connection($filer);
    say($conn,"cf enable",'-eval_error');
    return ($@) ? undef : 1;
}

sub cf_disable ( $ ) {
    my ($filer) = @_;
    my $conn = make_connection($filer);
    say($conn,"cf disable",'-eval_error');
    return ($@) ? undef : 1;
}


sub cf_takeover ( $;$ ) {
    my ($filer, $cmd) = @_;
    if (!$cmd) {
        $cmd = "cf takeover -f";
    }
    my $conn = make_connection($filer);

    # If the cf command fails due to vfiler failing to run, retry the command
    # until sucessfull
    my $retry = 1;
    while ($retry) {
        my (undef, undef, $output) = say($conn, $cmd, '-eval_error');
        if ($output !~ /try again later/) { $retry = 0; }
        sleep 1;
    }
    return ($@) ? undef : 1;
} ## end sub cf_takeover ( $;$ )

# Issue a command (cf giveback -f) to give back $filer.  If boot=>0,
# that's it.  If boot=>1, then the 'partner' option is required and we
# attempt to help the partner to reboot after the giveback.
# Simulators are usually the only hosts which need such help, so this
# option defaults to 1 if $filer has hosttype "filer-simulator" and 0
# otherwise. If force=>1 (default), force the giveback (use -f flag).

sub cf_giveback ( $;% ) {
    my ($filer,%options) = @_;
    return eval {
	my $fhost = make_host($filer);
	my $fconn = make_connection($filer);
	my $do_boot = exists($options{boot}) ? $options{boot} :
	    ($fhost->hosttype =~ /^filer-simulator/);
        my $force = (!exists $options{force} || $options{force}) ? '-f' : '';

        # If the cf command fails due to vfiler failing to run, retry the 
	# command until sucessfull
        my $retry = 1;
        while ($retry) {
            my (undef, undef, $output) =
              say($fconn, "cf giveback $force", '-eval_error');
            if ($output !~ /try again later/) { $retry = 0; }
            sleep 1;
        }
	die $@ if $@;
	if ($do_boot) {
	    my $partner = $options{partner};
	    if (!$partner) {
		die "'partner' option required for giveback() when 'boot' option is true\n";
	    }
	    my (@partner) = (ref($partner) eq 'ARRAY') ? @$partner : ($partner);
	    &boot(@partner) or die $@;
	}
	return 1;
    };
}

sub cf_status_precheck ( $$;$ ) {
    my ($filerA,$filerB,$test_check) = @_;
    
    logcomment ("Pre-Check to check the Status of the filers") if ($test_check == 1);

    my $fhostA = make_host($filerA);
    my $fconnA = make_connection($filerA);
    my $fhostB = make_host($filerB);
    my $fconnB = make_connection($filerB);
    
    my $boot_stateA = boot_state($fhostA);
    my $boot_stateB = boot_state($fhostB);

    logcomment ("Boot state of $filerA now is $boot_stateA, and $filerB is $boot_stateB");

    if ($boot_stateA eq "FIRMWARE" && $boot_stateB eq "FIRMWARE")
    {
         logcomment ("Both nodes $filerA and $filerB are down, rebooting both nodes.");

         reboot($fhostA);
         wait_for_rsh($fhostA) or logresult('FAIL'."Rsh did not came up for fhostA");
         reboot($fhostB);
         wait_for_rsh($fhostB) or logresult('FAIL'."Rsh did not came up for fhostB");

         logcomment ("sleeping for 300 seconds");
         sleep(300);

         $boot_stateA = boot_state($fhostA);
         $boot_stateB = boot_state($fhostB);
    }
    elsif ($boot_stateA eq "FIRMWARE" && $boot_stateB eq "UP")
    {
         logcomment ("$filerA is down and $filerB is UP, rebooting node $filerA.");

         reboot($fhostA);
         wait_for_rsh($fhostA) or logresult('FAIL'."Rsh did not came up for fhostA");

         logcomment ("sleeping for 300 seconds");
         sleep(300);

         $boot_stateA = boot_state($fhostA);
         $boot_stateB = boot_state($fhostB);
    }
    elsif ($boot_stateA eq "UP" && $boot_stateB eq "FIRMWARE")
    {
         logcomment ("$filerB is down and $filerA is UP, rebooting node $filerB.");

         reboot($fhostB);
         wait_for_rsh($fhostB) or logresult('FAIL'."Rsh did not came up for fhostB");

         logcomment ("sleeping for 300 seconds");
         sleep(300);

         $boot_stateA = boot_state($fhostA);
         $boot_stateB = boot_state($fhostB);
    }

    my $state_of_nodeA = get_cf_state ($filerA);
    my $state_of_nodeB = get_cf_state ($filerB);

    logcomment ("state of $filerA now is $state_of_nodeA, and $filerB is $state_of_nodeB");
    logcomment ("Boot state of $filerA now is $boot_stateA, and $filerB is $boot_stateB");
    
    if ($boot_stateA eq "" || $boot_stateB eq "")
    {
	 logresult ('FATAL',"Either or both of the nodes $filerA and $filerB could not be brought up, please look into it.");
    }

    if ($boot_stateA eq "FIRMWARE" && $boot_stateB eq "FIRMWARE")
    {
         logresult ('FATAL',"Either or both of the nodes $filerA and $filerB could not be brought up, please look into it.");
    }
    elsif ($state_of_nodeA eq "UP" && $state_of_nodeB eq "UP")
    {
         logcomment ("Both filers are up");
    }
    elsif ($state_of_nodeA eq "DISABLED")
    {
         logcomment ("cf is disabled, enabling it");
         cf_enable($filerA);
    }
    elsif ($state_of_nodeA eq "TAKEOVER")
    {
         if ($boot_stateB eq "FIRMWARE")
         {
              logcomment ("Node $filerB is down, rebooting it");
              reboot($fhostB);
              logcomment ("sleeping for 300 seconds");
              sleep(300);
         }
         elsif ($boot_stateB eq "TAKEN_OVER_WAITING_ONTAP")
         {
              logcomment ("Node $filerB has been takenover, giving it back");
              cf_giveback($filerA);
              sleep 60;
         }
         else
         {
              logresult ('WARN',"Check the status of the filers.");
         }
    }
    elsif ($state_of_nodeB eq "TAKEOVER")
    {
         if ($boot_stateA eq "FIRMWARE")
         {
              logcomment ("Node $filerA is down, rebooting it");
              reboot($fhostA);
              logcomment ("sleeping for 300 seconds");
              sleep(300);
         }
         elsif ($boot_stateA eq "TAKEN_OVER_WAITING_ONTAP")
         {
              logcomment ("Node $filerA has been takenover, giving it back");
              cf_giveback($filerB);
              sleep 60;
         }
         else
         {
              logresult ('WARN',"Check the status of the filers.");
         }
    }
    else
    {
         logresult ('WARN',"Unknown state of nodes $filerA and $filerB.  Please check the filers.");
    }
}


sub cf_options_precheck ( $$$$$ ) {
    my ($filerA,$filerB,$port_number,$test_check,$ip_type) = @_;
    my $ip;
    (($ip_type ==1) ? ($ip = "ipv6") : ($ip = "ipv4"));
    logcomment("The ip type is $ip");


    logcomment ("Pre-Check to check the Options of the filers") if ($test_check == 1);

    my $fhostA = make_host($filerA);
    my $fconnA = make_connection($filerA);
    my $fhostB = make_host($filerB);
    my $fconnB = make_connection($filerB);

    my $part_default_ip = $fhostB->default_ip();
    my $part_default_ipv6 = $fhostB->{'default_ipv6'};
    my $id = $fhostA->id();
    #my $rsh = connect("$filerA#rsh");
    my (undef, undef, $out) = say($fconnA,"options cf.hw_assist",'-eval_error');
    my (undef, undef, $out_interval) = say($fconnA,"options cf.hw_assist.health_check_interval",'-eval_error');

    # Do the necessary checking
    my ($enable) = ($out =~ /enable\s+(\S+)/);
    my ($address) = ($out =~ /address\s+(\S+)/);
    my ($port) = ($out =~ /port\s+(\S+)/);
    my ($interval) = ($out_interval=~ /health_check_interval\s+(\S+)/);
    if($ip_type == 0) {
        logcomment("Continuing with ipv4...");

    if($enable ne 'on' || $address ne $part_default_ip || $port ne '4444')# || $interval ne '60')
    {
         logcomment("Setting hw_assist options to the following default values on node $id prior to the tests:
         hw_assist.enable on
         hw_assist.partner.address $part_default_ip
         hw_assist.partner.port $port_number 
         hw_assist.partner.health_check_interval 60");
        
         my (undef, undef, $option) = say($fconnA,"options cf.hw_assist.enable on",'-eval_error');
         my (undef, undef, $address) = say($fconnA,"options cf.hw_assist.partner.address $part_default_ip",'-eval_error');
         my (undef, undef, $port) = say($fconnA,"options cf.hw_assist.partner.port $port_number",'-eval_error');#4444");
         my (undef, undef, $interval) = say($fconnA,"options cf.hw_assist.health_check_interval 60",'-eval_error');
    }
    else
    {
         logcomment("hw_assist options were already set to their default values on node $id.");
    }

}

elsif($ip_type == 1) {

        logcomment(" ipv6 is ENABLED; Continuing with ipv6...");

    if($enable ne 'on' || $address ne $part_default_ipv6 || $port ne '4444')# || $interval ne '60')
    {
         logcomment("Setting hw_assist options to the following default values on node $id prior to the tests:
         hw_assist.enable on
         hw_assist.partner.address $part_default_ipv6
         hw_assist.partner.port $port_number
         hw_assist.partner.health_check_interval 60");

         my (undef, undef, $option) = say($fconnA,"options cf.hw_assist.enable on",'-eval_error');
         my (undef, undef, $address) = say($fconnA,"options cf.hw_assist.partner.address $part_default_ipv6",'-eval_error');
         my (undef, undef, $port) = say($fconnA,"options cf.hw_assist.partner.port $port_number",'-eval_error');#4444");
         my (undef, undef, $interval) = say($fconnA,"options cf.hw_assist.health_check_interval 60",'-eval_error');
    }
    else
    {
         logcomment("hw_assist options were already set to their default values on node $id.");

    }
  }
}


sub cf_hwassist_stats_precheck ( $$;$ ) {
    my ($filerA,$filerB,$test_check) = @_;

    logcomment ("Pre-Check to check the HWASSIST STATS of the filers") if($test_check == 1);

    my $fhostA = make_host($filerA);
    my $fconnA = make_connection($filerA);
    my $fhostB = make_host($filerB);
    my $fconnB = make_connection($filerB);

    my $part_default_ip = $fhostB->default_ip();
    my $id = $fhostA->id();
    #my $rsh = connect("$filerA#rsh");
    my (undef, undef, $out) = say($fconnA,"cf hw_assist stats",'-eval_error');

    my @lines = split(/\n/, $out);
    
    my $power_loss;
    my $L2_watchdog_reset;
    my $power_off_via_rlm;
    my $power_cycle_via_rlm;
    my $reset_via_rlm;
    my $post_error;
    my $abnormal_reboot;
    my $loss_of_heartbeat;
    my $periodic_message;
    my $test;

    foreach my $line1(@lines) 
    {
         $power_loss = $line1 if($line1 =~ /power_loss/);
         $L2_watchdog_reset = $line1 if($line1 =~ /L2_watchdog_reset/);
         $power_off_via_rlm = $line1 if($line1 =~ /power_off_via_rlm/);
         $power_cycle_via_rlm = $line1 if($line1 =~ /power_cycle_via_rlm/);
         $reset_via_rlm = $line1 if($line1 =~ /reset_via_rlm/);
         $post_error = $line1 if($line1 =~ /post_error/);
         $abnormal_reboot = $line1 if($line1 =~ /abnormal_reboot/);
         $loss_of_heartbeat = $line1 if($line1 =~ /loss_of_heartbeat/);
         $periodic_message = $line1 if($line1 =~ /periodic_message/);
         $test = $line1 if($line1 =~ /test/);
    }

    # Capturing the info from the "cf hw_assist stats" command.

    my ($power_loss_alert_type, $power_loss_alert_event, $power_loss_count,$power_loss_takeover, $power_loss_last_received) = ($power_loss =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($L2_watchdog_reset_alert_type, $L2_watchdog_reset_alert_event, $L2_watchdog_reset_count, $L2_watchdog_reset_takeover, $L2_watchdog_reset_last_received) = ($L2_watchdog_reset =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($power_off_via_rlm_alert_type, $power_off_via_rlm_alert_event, $power_off_via_rlm_count, $power_off_via_rlm_takeover, $power_off_via_rlm_last_received) = ($power_off_via_rlm =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($power_cycle_via_rlm_alert_type, $power_cycle_via_rlm_alert_event, $power_cycle_via_rlm_count, $power_cycle_via_rlm_takeover, $power_cycle_via_rlm_last_received) = ($power_cycle_via_rlm =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($reset_via_rlm_alert_type, $reset_via_rlm_alert_event, $reset_via_rlm_count, $reset_via_rlm_takeover, $reset_via_rlm_last_received) = ($reset_via_rlm =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($post_error_alert_type, $post_error_alert_event, $post_error_count,$post_error_takeover, $post_error_last_received) = ($post_error =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($abnormal_reboot_alert_type, $abnormal_reboot_alert_event, $abnormal_reboot_count, $abnormal_reboot_takeover, $abnormal_reboot_last_received) = ($abnormal_reboot =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($loss_of_heartbeat_alert_type, $loss_of_heartbeat_alert_event, $loss_of_heartbeat_count, $loss_of_heartbeat_takeover, $loss_of_heartbeat_last_received) = ($loss_of_heartbeat =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($periodic_message_alert_type, $periodic_message_alert_event, $periodic_message_count, $periodic_message_takeover, $periodic_message_last_received) = ($periodic_message =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);
    my ($test_alert_type, $test_alert_event, $test_count, $test_takeover, $test_last_received) = ($test =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)/);

    if($power_loss_count != 0 || $L2_watchdog_reset_count != 0 || $power_off_via_rlm_count != 0 || $power_cycle_via_rlm_count != 0 || $reset_via_rlm_count != 0 || $post_error_count != 0 || $abnormal_reboot_count != 0 || $loss_of_heartbeat_count != 0 || $periodic_message_count != 0 || $test_count != 0)
    {
         logcomment("Clearing the stats on node $id prior to the tests");
         my (undef, undef, $stats_clear) = say($fconnA,"cf hw_assist stats clear",'-eval_error');
         say($fconnA,"cf hw_assist stats",'-eval_error');
    }
    else
    {
         logcomment("hw_assist stats were already clear on node $id.");
    }
}



sub cf_hwassist_precheck ( $$$$ ) {
    my ($filerA,$filerB,$test_check,$ip_type) = @_;

    logcomment ("Getting in routine precheck to check the configuration of the filers before starting the tests:") if($test_check ==1);

    cf_status_precheck($filerA,$filerB,$test_check);
    cf_options_precheck($filerA,$filerB,4500,$test_check,$ip_type);
    cf_options_precheck($filerB,$filerA,4444,$test_check,$ip_type);
    cf_hwassist_stats_precheck($filerA,$filerB,$test_check);
    cf_hwassist_stats_precheck($filerB,$filerA,$test_check);
    
    logcomment ("FILERS are UP and READY to USE for testing") if($test_check ==1);
}


#
#Configures takeover IP with shared interface and enables negotiated failovers.
#In this case the primary IP of the Filer will have primary IP of the partner 
#as the failover IP
#
sub setup_partner ( $$ ) {
    my ($filer,$h) = @_;
    my %hash = %$h;
    my $partner = $hash{part_host};
    my $p = host($partner);
    my %part_hash = %$p;
    my $conn =  make_conn($filer);
    my $data;

    if( defined($hash{IFC}) ) { 
      $data = rdfile($conn,"/etc/rc");
      my @lines = split(/\n/,$data);
      foreach (@lines) {
         if ( m/^ifconfig $hash{IFC}/ && !m/partner/) {
            $_.=" partner $part_hash{default_ip} nfo";
         }
      }
      $data = join("\n",@lines);
      wrfile($conn, "/etc/rc", $data."\n");
      sayEOF($conn,"options cf.takeover.on_network_interface_failure on");
      sayEOF($conn, "source /etc/rc");
     } else {
       logresult ('WARN',"IFC not defined on the nate host file"); 
     }
}
#
# HA feature in SierraNevada will be available without the need 
# for a HA license key.
# CLI will be available to turn on/off HA.
# This is done to support both HA pairs and non-HA nodes.
# This function will set the filer to ha mode
## Note: Remove the below after SN MS5.
## Reboot of nodes after setting cf.mode to HA will be removed in SN MS.5 testing

sub set_ha_mode ($$) {
    my ($filerA, $filerB) = @_;
    my $connA = make_connection($filerA);
    my $connB = make_connection($filerB);
    my (undef, undef, $ha_mode_A) = say($connA,"options cf.mode ha", '-eval_error');
    my (undef, undef, $ha_mode_B) = say($connB,"options cf.mode ha", '-eval_error');
    if ( $ha_mode_A =~ /A reboot is required for node to be in ha mode/i || $ha_mode_B =~ /A reboot is required for node to be in ha mode/i ) {
        reboot($filerA);
        reboot($filerB);
    }
    ## Make sure that cf.mode is set to HA
    my ( undef , undef , $ha_modeA ) = say($connA,"options cf.mode" , '-eval_error');
    my ( undef , undef , $ha_modeB ) = say($connB,"options cf.mode" , '-eval_error');
    my @list1 = split(/ +/ ,$ha_modeA,2); chomp (@list1);
    my @list2 = split(/ +/ ,$ha_modeB,2); chomp (@list2);
    my $modeA = pop(@list1);
    my $modeB = pop(@list2);
    if ( $modeA =~ /^ha/  && $modeB =~ /^ha/) {
        tstlog('-comment',"Sucessfully set cf.mode to HA");
        return 1;
    }else {
        tstlog('-comment',"Failed to set cf.mode to HA");
        return 0;
    }

}

1;
