package compression;

## @summary Compression basic functionality
##
## @description
##    This module provides basic compression functionality.
##
## @author dl-compression-qa@netapp.com
##
## @note
## Any modification to this module requires approval from the maintainers (dl-compression-qa@netapp.com).


use strict;
use filer;
use Burt;
use Tharn qw(connect host say logcomment logresult param tstlog);
use Utils qw(make_conn make_host);
use Data::Dumper qw(Dumper);
use Admin::License;
use Dense::Status;
use Dense::Constant::Command;
use Dense::Constant::Message;
use Params::Validate qw(validate validate_pos OBJECT SCALAR HASHREF ARRAYREF UNDEF BOOLEAN);
BEGIN {
        Params::Validate::validation_options(
            on_fail => sub {my @msgs = @_; chomp(@msgs);
                Carp::confess(Tharn::script_result(@msgs))}
           );


        ###############################################################################
        require Exporter;
        our @ISA= qw(Exporter);
        our @EXPORT = qw (
                                license_compression
                                is_compression_licensed
                                enable_compression_option
                                disable_compression_option
                                is_compression_enabled
                                is_compression_supported
                                vol_decompress
                                get_savings
                                compression_status
                                store_vbns
                                store_counter_compression
                                store_counter_volume
                                compare_vbns
                                compare_counter
                                start_counter_manager
                                stop_counter_manager
                                store_stats
                                print_vbns
                                print_counter
                                do_sync
                                do_rsh
                           );
        
        ###############################################################################
}


use constant C_C_MIN_ONTAP_MAY_SUPPORT_COMPR => 8.0;

################
# TODO:
# these are to be deprecated too, along with the unsupported subroutines
# and with the applicable includes
#These need to be moved to constants.pm later
use constant CMD_STAT_COUNTER_COMPRESS => "priv set -q test; stats show compression ";
use constant CMD_STAT_COUNTER_VOLUME => "priv set -q test; stats show volume ";
use constant CMD_WAFL_SHOWVBNS => "priv set -q test; wafl showvbns ";
use constant CMD_START_COUNTER_MANAGER => "priv set -q test; stats start ";
use constant CMD_STOP_COUNTER_MANAGER => "priv set -q test; stats stop";

my $VERBOSE = 0;
my @cmd_list = ();

#my %counter_volume_before = ();
#my %counter_volume_after = ();
#my %counter_compression_before = ();
#my %counter_compression_after = ();
#my %vvbns_before = ();
#my %vvbns_after = ();

#do_sync($filer);
#
#store_counter_volume(\%counter_volume_before, $filer);
#store_counter_compression(\%counter_compression_before, $filer);
#
#start_counter_manager($filer, "compression");
#
#print `r rm file.txt`;
#my $buf = `r perl /u/sureshb/data_files/make_filesize.pl 32 $path/file.txt 0 0`;
#print $buf;
#
#do_sync($filer);
#
#my %counter_compression_after = stop_counter_manager($filer, "compression");
#
#store_counter_volume(\%counter_volume_after, $filer);
#
#print_counter(\%counter_volume_before, "Volume before");
#print_counter(\%counter_volume_after, "Volume after");
#compare_counter(\%counter_volume_before, \%counter_volume_after, "Volume counters");
#
#print_counter(\%counter_compression_before, "Compression before");
#print_counter(\%counter_compression_after, "Compression after");
#compare_counter(\%counter_compression_before,\%counter_compression_after, "Compression counters");
# End TODO (deprecate)
################



## @name ::license_compression
## @summary Installs the compression license.
##
## @description
## Installs the compression license.
##
## @arg filer Required: Filer name.
##
## @arg license_dependencies Optional: Boolean flag to be used
## if licensing dependecies is desired. Default: 0.
##
## @arg verbose Optional: Boolean flag to enable status comments.
## Default: 0.
##
## @examples
## license_compression(filer => 'my_filer', license_dependencies => 1);
##
## @author dl-compression-qa@netapp.com
##
sub license_compression ($@)
{
        my %args = validate(@_,
        {
                filer => { type => SCALAR, },
                license_dependencies => { type => BOOLEAN, default => 0 },
                verbose => { type => BOOLEAN, default => 0 },
        } );

        my $filer = $args{filer};
        my $license_dependencies = $args{license_dependencies};
        my $verbose = $args{verbose};
        
        my $conn = make_conn($filer, "#rsh");
        my $licenses_to_install = [];
        
        if (!is_compression_supported(filer => $filer, verbose => $verbose))
        {
                logresult('NA', "Compression is not supported on this ontap version");
        }
        
        if ($license_dependencies)
        {
                my $ontap = _get_ontap_version($conn);
                $licenses_to_install = _get_license_dependencies($ontap);
        }
        
        push(@$licenses_to_install, 'compression');
        
        foreach my $license (@$licenses_to_install)
        {
                logcomment("Installing the \"$license\" license") if ($verbose);
                if (!Admin::License::license($conn, $license))
                {
                        logresult('fatal', "Failed to license \"$license\":\n" . $@);
                }
        }
}


## @name ::is_compression_licensed
## @summary Checks if the compression license is installed.
##
## @description
## Checks if the compression license is installed.
##
## @arg filer Required: Filer name.
##
## @arg check_dependencies Optional: Boolean flag to be used
## if checking the licensing dependecies is desired. Default: 0.
##
## @arg verbose Optional: Boolean flag to enable status comments.
## Default: 0.
##
## @return Boolean value indicating if compression is licensed
## (may be 2, if site license is detected).
##
## @examples
## is_compression_licensed(filer => 'my_filer', check_dependencies => 1);
##
## @author dl-compression-qa@netapp.com
##
sub is_compression_licensed
{
        my %args = validate(@_,
        {
                filer => { type => SCALAR, },
                check_dependencies => { type => BOOLEAN, default => 0 },
                verbose => { type => BOOLEAN, default => 0 },
        } );

        my $filer = $args{filer};
        my $check_dependencies = $args{check_dependencies};
        my $verbose = $args{verbose};
        
        my $is_licensed = 1;
        my $conn = make_conn($filer, "#rsh");
        my $licenses_to_check = [];
        
        if ($check_dependencies)
        {
                my $ontap = _get_ontap_version($conn);
                $licenses_to_check = _get_license_dependencies($ontap);
        }
        
        push(@$licenses_to_check, 'compression');
        
        foreach my $license (@$licenses_to_check)
        {
                logcomment("Checking the \"$license\" license") if $verbose;
                $is_licensed = ($is_licensed) && (filer_service_is_licensed($conn, $license));
        }
        
        return $is_licensed;
}


## @name ::is_compression_supported
## @summary Checks if compression is supported.
##
## @description
## Checks if compression is supported.
##
## @arg filer Required: Filer name.
##
## @arg volname Optional: Volume name.
##
## @arg verbose Optional: Boolean flag to enable status comments.
## Default: 0.
##
## @return Boolean value indicating if compression is supported.
##
## @examples
## my $is_compr_vol = is_compression_supported(filer => 'my_filer', volname => 'my_vol');
##
## @author dl-compression-qa@netapp.com
##
sub is_compression_supported
{
	my %opts = validate (@_,
        {
                filer => { type=>SCALAR },
                volume => { type=>SCALAR, optional => 1 },
                verbose => { type=>SCALAR, default => 0 }, 
	});
        
	my $filer = $opts{filer};
	my $volume = $opts{volume};
        my $verbose = $opts{verbose};
        
	my $conn = make_conn($filer, "#rsh");
        my $ontap = _get_ontap_version($conn);
        my $cmd_type = 'is_supported';
        
        if ($ontap < C_C_MIN_ONTAP_MAY_SUPPORT_COMPR)
        {
                logcomment("Ontap version $ontap does not support compression") if ($verbose);
                return 0;
        }
        
        # For some ontap versions we may need some extra checks
        my $expected_is_supported_string = _get_is_supported_string(ontap_ver => $ontap);
        if ($expected_is_supported_string eq '')
        {
                # No extra checks needed
                logcomment("Compression is supported on ontap version $ontap (no extra check needed)") if ($verbose);
                return 1;
        }
        
        # If no volume is passed in, get any existent volume to check if the option is present
        # Warning: we should only do this as long as the "check action" does not change the state of the volume in any way
        if (!defined($volume))
        {
                logcomment("Auto-detecting an existent volume to perform the extra check for compression") if ($verbose);
                my @volume_list = filer_vols_get($conn);
                $volume = pop(@volume_list);
        }
        
        my $command = _get_command_string(ontap_ver => $ontap,
                                          cmd_type => $cmd_type,
                                          volume => $volume,);
        
        if ($command eq '$conn')
        {
                logresult('script', "Extra checks needed to check if compression is supported, but no command is defined for ontap version: $ontap");
        }
        
        my $output = filer_cli_say($conn, $command);
        
	if ($output =~ /$expected_is_supported_string/)
        {
                logcomment("Compression is supported on ontap version $ontap (extra check completed)") if ($verbose);
                return 1;
        }
        
        logcomment("Ontap version $ontap does not support compression (extra check failed)") if ($verbose);
        return 0;
}


## @name ::is_compression_enabled
## @summary Checks if compression is enabled on the volume.
##
## @description
## Checks if compression is enabled on the volume.
##
## @arg filer Required: Filer name.
##
## @arg volname Required: Volume name.
##
## @arg verbose Optional: Boolean flag to enable status comments.
## Default: 0.
##
## @return Boolean value indicating if compression is enabled.
##
## @examples
## my $is_compr_vol = is_compression_enabled(filer => 'my_filer', volname => 'my_vol');
##
## @author dl-compression-qa@netapp.com
##
sub is_compression_enabled
{
	my %opts = validate (@_,
        {
                filer => { type=>SCALAR },
                volname => { type=>SCALAR },
                verbose => { type=>SCALAR, default => 0 },
        });
        
	my $filer = $opts{filer};
	my $volume = $opts{volname};
        my $verbose = $opts{verbose};
        
	my $conn = make_conn($filer, "#rsh");
        my $ontap = _get_ontap_version($conn);
        my $cmd_type = 'is_enabled';
        
        logcomment("Checking if compression is enabled for volume $volume (filer $filer)") if ($verbose);

	my $command = _get_command_string(ontap_ver => $ontap,
                                          cmd_type => $cmd_type,
                                          volume => $volume,);
        
        my $output = filer_cli_say($conn, $command);
        
	return _parse_output_string(ontap_ver => $ontap,
                                    cmd_type => $cmd_type,
                                    output => $output);
}


## @name ::enable_compression_option
## @summary Enables compression for the volume.
##
## @description
## Enables compression for the volume.
##
## @arg filer Required: Filer name.
##
## @arg volname Required: Volume name.
##
## @examples
## enable_compression_option(filer => 'my_filer', volname => 'my_vol');
##
## @note
## Starting with RR, compression is no longer considered a volume option,
## but an addition to the sis functionality.
## This means that sis needs to be enabled on this volume (must be done before enabling compression).
## Even for these releases, this subroutine handles configuring compression for the volume (sis must be already on).
##
## @author dl-compression-qa@netapp.com
##
sub enable_compression_option
{
	my %opts = validate (@_,
        {
                filer => { type=>SCALAR },
                volname => { type=>SCALAR },
                verbose => { type=>SCALAR, default => 0 },
	});
        
	my $filer = $opts{filer};
	my $volume = $opts{volname};
        my $verbose = $opts{verbose};
        
	my $conn = make_conn($filer, "#rsh");
        my $ontap = _get_ontap_version($conn);
        my $cmd_type = 'enable';

	logcomment("Enabling compression on volume $volume (filer $filer)");

	if (!is_compression_supported(filer => $filer, volume => $volume, verbose => $verbose))
        {
                logresult('NA', "Compression is not supported on this ontap version: $ontap");
        }

	my $command = _get_command_string(ontap_ver => $ontap,
                                          cmd_type => $cmd_type,
                                          volume => $volume,);
        
        filer_cli_say($conn, $command);
        
        logcomment("Verifying the compression setting on volume $volume (filer $filer)") if ($verbose);
	my $current_setting = is_compression_enabled(filer => $filer, volname => $volume);
        if ($current_setting == 0)
        {
                logresult('fatal', "Failed to enable compression on volume $volume\n" .
                                   "Possible causes of failure:\n" .
                                   "- Sis is not on");
        }
        
        return 1;
}

## @name ::disable_compression_option
## @summary Disables compression for the volume.
##
## @description
## Disables compression for the volume.
##
## @arg filer Required: Filer name.
##
## @arg volname Required: Volume name.
##
## @examples
## disable_compression_option(filer => 'my_filer', volname => 'my_vol');
##
## @note
## Starting with RR, compression is no longer considered a volume option,
## but an addition to the sis functionality.
## This means that sis needs to be enabled on this volume.
## Even for these releases, this subroutine handles disabling compression for the volume.
##
## @author dl-compression-qa@netapp.com
##
sub disable_compression_option
{
	my %opts = validate (@_,
        {
                filer => { type=>SCALAR },
                volname => { type=>SCALAR },
                verbose => { type=>SCALAR, default => 0 },
        });
        
	my $filer = $opts{filer};
	my $volume = $opts{volname};
        my $verbose = $opts{verbose};
        
	my $conn = make_conn($filer, "#rsh");
        my $ontap = _get_ontap_version($conn);
        my $cmd_type = 'disable';
	
	logcomment("Disabling Compression on volume $volume (filer $filer)");

	if (!is_compression_supported(filer => $filer, volume => $volume, verbose => $verbose))
        {
                logresult('NA', "Compression is not supported on this ontap version: $ontap");
        }
        
        # Verify the setting
        my $command = _get_command_string(ontap_ver => $ontap,
                                          cmd_type => $cmd_type,
                                          volume => $volume,);
        
        filer_cli_say($conn, $command);
        
        logcomment("Verifying the compression setting on volume $volume (filer $filer)") if ($verbose);
        my $current_setting = is_compression_enabled(filer => $filer, volname => $volume);
        if ($current_setting == 1)
        {
                logresult('fatal', "Failed to disable compression on volume $volume");
        }
        
        return 1;
}


######################
# Internal subroutines
######################

sub _get_ontap_version
{
        my $conn = shift;
        
        my $host = make_host($conn);
        return Filer::Version::get_ontap_version($host);
}

sub _get_license_dependencies
{
        my $ontap = shift;
        
        my $licensing_info =
	{
		'default' => ['a_sis'],
	};
        
        my $release_specific_info = (exists($licensing_info->{$ontap})) ? $licensing_info->{$ontap} : $licensing_info->{default};
	
        return $release_specific_info;
}

sub _get_command_string
{
        my %args = validate(@_,
        {
                ontap_ver => { type => SCALAR, },
                cmd_type => { type => SCALAR, },
                volume => { type => SCALAR, },
        } );
        
        # Only support here actions for a specific volume
        my $ontap = $args{ontap_ver};
        my $cmd_type = $args{cmd_type};
        my $volume = $args{volume};
        
        # Any cmd_type entry in here MUST have the 'default' hash
        my $commands =
        {
                'enable' => {'default' => 'sis config -C true -I true /vol/<volume>',
                             '8.0'=>'vol options <volume> compression on'},
                # Note: the default command for disable should in fact be only 'sis config -C false -I false /vol/<volume>'
                # The SW does not allow this yet. To change accordingly when the functionality is there.
                'disable' => {'default' => 'priv set -q diag;sis config -C true -I false /vol/<volume>;sync;sync;sis config -C false -I false /vol/<volume>',
                              '8.0'=>'vol options <volume> compression off'},
                'is_enabled' => {'default' => 'sis status -l /vol/<volume>',
                                 '8.0' => 'vol options <volume>',},
                # For is_supported, an empty command string means that there is no extra check needed
                # If an extra check is needed, this needs to be a neutral action,
                # otherwise the logic in is_compression_supported needs to change for the case of
                # an undefined volume name
                'is_supported' => {'default' => '',
                                   '8.0' => 'vol options <volume>',},
        };
        
        if (!exists($commands->{$cmd_type}))
        {
                logresult('script', "Commands for compression action \"$cmd_type\" are not defined. Cannot continue");
        }
        
        my $cmd_entries = $commands->{$cmd_type};
        
        my $command = (defined($cmd_entries->{$ontap})) ? $cmd_entries->{$ontap} : $cmd_entries->{default};
        $command =~ s/<volume>/$volume/g;
        
        if (!defined($command))
        {
                logresult('script', "Command for compression action \"$cmd_type\" is not defined. Cannot continue");
        }
        
        return $command;
}

sub _parse_output_string
{
        my %args = validate(@_,
        {
                ontap_ver => { type => SCALAR, },
                cmd_type => { type => SCALAR, },
                output => { type => SCALAR, },
        } );
        
        # Only support here actions for a specific volume
        my $ontap = $args{ontap_ver};
        my $cmd_type = $args{cmd_type};
        my $output = $args{output};
        
        # Any cmd_type entry in here SHOULD have the 'default' hash
        my $known_outputs =
        {
                'is_enabled' => {'default' => {'No status entry found' => 0,
                                               'Compression:\s+Disabled\s*\nInline Compression:\s+Disabled' => 0,
                                               # This corresponds to bg-on/inline-off. Leave this as 'unexpected output'
                                               # for now, since we only enable both or none in here
                                               #'Compression:\s+Enabled\s*\nInline Compression:\s+Disabled' => 0,
                                               'Compression:\s+Enabled\s*\nInline Compression:\s+Enabled' => 1,},
                                 '8.0' => {'compression=off' => 0,
                                           'compression=on' => 1,},},
        };
        
        if (!exists($known_outputs->{$cmd_type}))
        {
                logresult('script', "Expected outputs for compression action \"$cmd_type\" are not defined. Cannot continue");
        }
        
        my $action_outputs = $known_outputs->{$cmd_type};
        
        my $this_release_outputs = (exists($action_outputs->{$ontap})) ? $action_outputs->{$ontap} : $action_outputs->{default};
                
        foreach my $possible_match (keys %$this_release_outputs)
        {
                if ($output =~ /$possible_match/i)
                {
                        return $this_release_outputs->{$possible_match};
                }
        }
        
        logresult('fatal', "Unexpected output for compression action \"$cmd_type\":\n$output\n");
}

sub _get_is_supported_string
{
        my %args = validate(@_,
        {
                ontap_ver => { type => SCALAR, },
        } );
        
        my $ontap = $args{ontap_ver};
        
        # For any release that has an expected "is supported" string,
        # there should also be the correspondent command in _get_command_string for 'is_supported'
        my $is_supported_strings =
        {
                '8.0' =>  'compression=',
        };
        
        my $this_release_outputs = (defined($is_supported_strings->{$ontap})) ? $is_supported_strings->{$ontap} : '';
        return $this_release_outputs;
}


sub testme
{
        package main;
        use vars qw($FILER $AGGREGATE);
        
        use strict;
        use warnings;
        use Data::Dumper;
        
        use Volume qw(vol_create_flex vol_purge);
        use compression qw(is_compression_licensed
                           is_compression_supported
                           is_compression_enabled
                           license_compression
                           enable_compression_option
                           disable_compression_option);
        use Tharn qw(connect host say logcomment logresult param tstlog);
        use Utils qw(make_conn);
        use Admin::License;
        use Dense qw(dense_toggle);
        
        param("FILER", "-mesg", "Filer name");
        param("AGGREGATE", "-mesg", "Aggregate name");
        
        my $fc = make_conn($FILER, "#rsh");
        
        my $aggr_name = $AGGREGATE;
        my $vol_name = 'testvol';
        my $result;
        
        
        logcomment("\n##########################\n" .
                   "# Creating the test volume\n" .
                   "##########################");
        vol_create_flex($fc, $vol_name, $aggr_name, "20m") or die  $@;
        
        
        logcomment("\n##################################################\n" .
                   "# Checking if compression is supported (no volume)\n" .
                   "##################################################");
        is_compression_supported(filer => $FILER,
                                 verbose => 1);
        
        
        logcomment("\n###############################################\n" .
                   "# Checking if compression is supported (volume)\n" .
                   "###############################################");
        is_compression_supported(filer => $FILER,
                                 volume => $vol_name,
                                 verbose => 1);
        
        
        logcomment("\n##################################\n" .
                   "# Removing the compression license\n" .
                   "##################################");
        if (!Admin::License::unlicense($fc, 'compression'))
        {
                logresult('config', "Failed to remove the compression license\n" .
                                    "Please make sure there are no compression-enabled volumes on this filer");
        }
        
        
        logcomment("\n#############################################################\n" .
                   "# Checking if compression is licensed (checking dependencies)\n" .
                   "#############################################################");
        $result = is_compression_licensed(filer => $FILER, check_dependencies => 1, verbose => 1);
        if ($result != 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is not licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned 0");
        
        
        logcomment("\n#####################################\n" .
                   "# Checking if compression is licensed\n" .
                   "#####################################");
        $result = is_compression_licensed(filer => $FILER, verbose => 1);
        if ($result != 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is not licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned 0");
        
        
        logcomment("\n###########################\n" .
                   "# Removing the asis license\n" .
                   "###########################");
        Admin::License::unlicense($fc, 'a_sis');
        
        
        logcomment("\n#####################################\n" .
                   "# Checking if compression is licensed\n" .
                   "#####################################");
        $result = is_compression_licensed(filer => $FILER, verbose => 1);
        if ($result != 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is not licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned 0");
        
        
        logcomment("\n#############################################################\n" .
                   "# Checking if compression is licensed (checking dependencies)\n" .
                   "#############################################################");
        $result = is_compression_licensed(filer => $FILER, check_dependencies => 1, verbose => 1);
        if ($result != 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is not licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned 0");
        
        
        logcomment("\n###########################################\n" .
                   "# Licensing compression (with dependencies)\n" .
                   "###########################################");
        license_compression(filer => $FILER,
                            license_dependencies => 1,
                            verbose => 1);
        
        
        logcomment("\n#############################################################\n" .
                   "# Checking if compression is licensed (checking dependencies)\n" .
                   "#############################################################");
        $result = is_compression_licensed(filer => $FILER, check_dependencies => 1, verbose => 1);
        if ($result == 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned true");
        
        
        logcomment("\n##############################################################\n" .
                   "# Checking if compression is licensed (not checking dependencies)\n" .
                   "#################################################################");
        $result = is_compression_licensed(filer => $FILER, verbose => 1);
        if ($result == 0)
        {
                logresult('fatal', "is_compression_licensed did not detect that compression is licensed");
        }
        logcomment("SUCCESS: is_compression_licensed returned true");
        
        
        logcomment("\n###############\n" .
                   "# Turning sis on\n" .
                   "################");
        dense_toggle(host => $FILER,
                     path_name => '/vol/' . $vol_name,
                     state => 'on');
        
        for (my $i = 1; $i <= 5; $i++)
        {
                logcomment("\n#########################\n" .
                           "# Enabling compression: $i\n" .
                           "##########################");
                enable_compression_option(filer => $FILER, volname => $vol_name);
                
                
                logcomment("\n##########################\n" .
                           "# Checking compression: $i\n" .
                           "##########################");
                $result = is_compression_enabled(filer => $FILER, volname => $vol_name);
                if ($result != 1)
                {
                        logresult('fatal', "is_compression_enabled did not detect that compression is enabled");
                }
                logcomment("SUCCESS: is_compression_enabled returned 1");
                
                
                logcomment("\n###########################\n" .
                           "# Disabling compression: $i\n" .
                           "###########################");
                disable_compression_option(filer => $FILER, volname => $vol_name);
                
                
                logcomment("\n##########################\n" .
                           "# Checking compression: $i\n" .
                           "##########################");
                $result = is_compression_enabled(filer => $FILER, volname => $vol_name);
                if ($result != 0)
                {
                        logresult('fatal', "is_compression_enabled did not detect that compression is disabled");
                }
                logcomment("SUCCESS: is_compression_enabled returned 0");
        }
        
        vol_purge($fc, $vol_name);
}



#########################################
# Below this line: will not be maintained
#########################################

# Not exported, leave it undocumented. Will be deprecated too
sub get_vol_option_compression
{
	my %opts = validate (@_, { filer => { type=>SCALAR },
				   volname => { type=>SCALAR } }
			     );
	my $filer = $opts{filer};
	my $volume = $opts{volname};
	my $conn = make_conn($filer, "#rsh");

	my @out = filer_vol_options_get($conn, $volume);

	my $value = "";	
	foreach my $val (@out)
	{
		$value = $val;
		if($val =~ /compression=/i)
		{
			last;
		}
	}
	return $value;
}

## @name ::vol_decompress
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# vol_decompress ( filer=>$filer, volname=>$volname )
sub vol_decompress ($@) 
{
        my %opts = validate (@_, { filer => { type=>SCALAR },
                                   volname => { type=>SCALAR },
				   wait => { type=>SCALAR, default=>1, regex=>qr/^(0|1)$/ } }
                             );
        my $filer = $opts{filer};
        my $volume = $opts{volname};
	my $wait = $opts{wait};
        my $conn = make_conn($filer, "#rsh");

	my $cmd = "vol decompress start" . " " . $volume;

        logcomment("Starting decompress on volume $volume on filer $filer");
	filer_cli_say($conn, $cmd);
	
	if($wait) {
		logcomment("Waiting for decompress to finish ...");
		my $decomp;
		do {
			logcomment("Decompression still in progress.");
			$decomp = compression_status( filer=>$filer, volname=>$volume );
		} while($decomp <= 0); 
	}
	else {
		logcomment("Not waiting for decompress to finish.");
	}
}

## @name ::compression_status
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# compression_status ( filer=>$filer, volname=>$volname )
sub compression_status ($@) 
{

	my %opts = validate (@_, { filer => { type=>SCALAR },
                                   volname => { type=>SCALAR } }
                             );

        my $filer = $opts{filer};
        my $volume = $opts{volname};
        my $conn = make_conn($filer, "#rsh");
        my $cmd = "vol decompress status";
        
        my ($return_text, $result_code) = filer_cli_say($conn, $cmd);
        
        # logcomment("return_text is $return_text");
        # logcomment("return code is $result_code");
        
        my (@process, %decomp_status);
            
        my $nodecomp = "No volume decompression in progress.";

        # Parse the return value and return approiately  
        if($result_code eq '255') {

        	# logcomment("Result code is 255");
                chomp($return_text);
        	if($return_text eq $nodecomp) { 
			logcomment("No decompression in progress");  
			return 1; 
		}
        	else {  
                	logcomment("$return_text is not equal to $nodecomp");
                 	@process = filer_cli_split($return_text); 
        	}

        	my $header = qr/^\s*Filesystem\s+Progress\b/;
         
            	if ($process[0] !~ /$header/){
                	die "Unrecognized format for $header: $process[0]\n";
            	}else{
                	shift @process;
            	}
        
        	foreach my $line (@process) {
                	#skip first empty line
	                # Parse them into hash 
        
       		         if ($line =~ /^\s*(\S+)\s+(.+)\s*$/){
                	    	my $key = $1;
                    		my $progress = $2;
        
                    		# get rid of empty space at the end, 
                    		$key =~ s/\s*$//;
        
                    		$decomp_status{$key}{Progress} = $progress;
        
                	}else{
                    		die "Unable to parse status:$line\n";
                	}
            	} # end foreach
        }
       	if ($decomp_status{$volume}{Progress} >= 0) {
               	logcomment("Decompress in progress for volume $volume");
               	return 0;
       	}
       	else { 
		logcomment("No decompress in progress for volume $volume"); 
		return 1; 
	}
}

#=head2 get_savings(host=>$filer, vol_name=>$vol_name)
#
#Get the ASIS and Compression savings of a volume(if provided) or all the volumes
#
#=over 4
#
#=item host
#
#Filer HostSpec (Required)
#
#=item vol_name
#
#Volume name (Optional)
#
#=item Returns:
#
#Success
#
#=over 4
#
#Returns 1 and hash containing ASIS and compression savings
#
#=back
#
#Failure
#
#=over 4
#
#Returns 0
#
#=back
#
#=back
#
#=cut

## @name ::get_savings
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
####################### df -S ####################

sub get_savings
{
        my (%opts) = validate(@_, { host => { type=>SCALAR|HASHREF },
                                   vol_name => { type=>SCALAR, optional=>1 }}
                             );
        
        
        my $cfiler;
        if (ref($opts{host})) {
           	$cfiler = $opts{host};
        } else {
           	my $filer =$opts{host};
           	$cfiler = connect("#rsh");
        }
        my $cmd="";
        if($opts{vol_name})
        {
        	$cmd = "df -S" . " $opts{vol_name}";
        }
        else
        {
        	$cmd = "df -S";
        }
        
        my ($df_result, $df_exit__code) = filer_cli_say($cfiler,$cmd);
        
        my %df_result=();
        $df_result{result}=0;
        
        # Return failure if df command failed.
        return (%df_result)  if ((split(/\n/,$df_result))[0] !~ /^Filesystem\s+used\s+compressed\s+a-sis\s+\%saved/);
        # Test Passed. Fill success bit (1) in $df_result{result} and populate %df_result
        $df_result{result}=1;
        
        # Returning the result in the hash which looks like this
        #       %df_result={    fs=>{  vol1=>{used=>value,compressed=>value,asis=>value,,%saved=>value},
        #                              vol2=>{used=>value,compressed=>value,asis=>value,,%saved=>value},
        #                           }
        #                       result=>0/1
        #                  }
        #
        
        foreach my $line (split(/\n/, $df_result))
        {
            if($line =~ /^\s*(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+\%)\s*/)
            {
                    $df_result{fs}->{$1}->{"used"}=$2;
                    $df_result{fs}->{$1}->{"compressed"}=$3;
                    $df_result{fs}->{$1}->{"asis"}=$4;
                    $df_result{fs}->{$1}->{"saved"}=$5;
            }
        }
        return %df_result;
}

## @name ::printfh
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# printfh (buf=>$buf)
sub printfh ($@)
{
	my %opts = validate (@_, { buf => { type=>SCALAR } } );
	my ($buf) = $opts{buf};

	#add to debug message later
	print $buf;
	#print fh $buf;
}

## @name ::do_rsh
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# do_rsh (filer=>$filer, cmd=>$cmd; options=>\@options)
sub do_rsh ($@)
{
	my %opts = validate (@_, { filer => { type=>SCALAR },
				   cmd => { type=>SCALAR },
				   options => { type=>ARRAYREF | UNDEF } }
			     );
	my ($filer , $cmd, $optionref, @options); 
	$filer = $opts{filer};
	$cmd = $opts{cmd};
	$optionref = $opts{options};
	@options = @$optionref;

	printfh("\nCMD = $cmd @options\n");
	push @cmd_list, "rsh $filer -l root '$cmd @options'";
	my $buf = `rsh $filer -l root "$cmd" @options`;
	printfh($buf) if($VERBOSE);
	return $buf;
}

## @name ::construct_hash_counter
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# construct_hash_counter (buffer=>$buf)
sub construct_hash_counter ($@)
{
	my %opts = validate (@_, { buffer => { type=>SCALAR } } );
	my ($buf) = $opts{buffer};

	my %val = ();

	print "\n==Parsing output==\n" if($VERBOSE);
	print "\n--Null buffer--\n" if($buf eq "" and $VERBOSE);
	foreach my $line (split /\n/, "$buf")
	{
		print "\nLine = $line\n" if($VERBOSE);
		my @vals = split /:/, $line;
		my $val = $vals[$#vals];
		my $key = $vals[$#vals-1];
		#my $prefix = "";
		#foreach my $i (0..($#vals-2))
		#{ 
		#	$prefix .= "{$vals[$i]}"; 
		#}
		$val{$key} = $val;
		print "{$key},$val, $key = $val{$key}" if($VERBOSE);
	}
	print "\n==Done==\n" if($VERBOSE);
	return %val;
}

## @name ::construct_hash_vbns
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# construct_hash_vbns (buffer => $buf)
sub construct_hash_vbns ($@)
{
	my %opts = validate (@_, { buffer => { type=>SCALAR } } );
        my ($buf) = $opts{buffer};

	my %val = ();
	my @bufs = split /\n/, $buf;
	foreach my $line (@bufs)
	{
		#print "\n$line" if($VERBOSE);
		#next if($line =~ /[4294967292|4294967294]/);
		if($line =~ /fbn (\d+) is vbn (\d+), vvbn (\d+)/)
		{
			$val{$1} = {};
			$val{$1}{vbn} = $2;
			$val{$1}{vvbn} = $3;
			print "\nStoring {$1}{vbn} = $val{$1}{vbn}, {$1}{vvbn} = $val{$1}{vvbn}\n" if($VERBOSE);
		} else {
			print "\nInvalid output format: Expected => fbn (d+) is vbn (d+), vvbn (d+), Obtained => $line\n";
		}
	}
	return %val;
}

## @name ::store_vbns
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# store_vbns (ref_ds=>$ref_ds, filer=>$filer, volume=>$volume, dir=>$dir, file=>$file)
sub store_vbns ($@)
{
	my %opts = validate (@_, { ref_ds => { type=>HASHREF },
				   filer => { type=>SCALAR },
				   volume => { type=>SCALAR },
				   dir => { type=>SCALAR | UNDEF },
				   file => { type=>SCALAR } }
			     );
				    
	my ($ref_ds, $filer, $volume, $dir, $file);

	$ref_ds = $opts{ref_ds};
	$filer = $opts{filer};
	$volume = $opts{volume};
	$dir = $opts{dir};
	$file = $opts{file};

	$dir = "." if(!defined $dir);
	my $SHOWVBNS_OPTION = "/vol/$volume/$dir/$file";

	%{$ref_ds} = 
	construct_hash_vbns (
		do_rsh ( $filer, (CMD_WAFL_SHOWVBNS . $SHOWVBNS_OPTION))
	);
}


## @name ::store_counter
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# store_counter (ref_ds=>$ref_ds, object=>$object, filer=>$filer, volume=>$volume, adnl_cmd=>$cmd)
sub store_counter ($@)
{
	my %opts = validate (@_, { ref_ds => { type=>HASHREF },
				   object => { type=>SCALAR },
                                   filer => { type=>SCALAR },
                                   volume => { type=>SCALAR },
				   adnl_cmd => { type=>SCALAR | UNDEF } }
                             );

	my ($ref_ds, $object, $filer, $volume, $additional_cmd);
	
	$ref_ds = $opts{ref_ds};
	$object = $opts{object};
        $filer = $opts{filer};
        $volume = $opts{volume};
	$additional_cmd = $opts{adnl_cmd};

	my $cmd = ($object eq "compression")?
			CMD_STAT_COUNTER_COMPRESS :
			($object eq "volume")?
				CMD_STAT_COUNTER_VOLUME : "";

	$cmd = $additional_cmd if($cmd eq "");

	%{$ref_ds} =
	construct_hash_counter (
			do_rsh ( $filer, $cmd)
	);
}

## @name ::store_counter_compression
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# store_counter_compression (ref_ds=>$ref_ds, filer=>$filer, volume=>$volume)
sub store_counter_compression ($@)
{
	 my %opts = validate (@_, { ref_ds => { type=>HASHREF },
                                   filer => { type=>SCALAR },
                                   volume => { type=>SCALAR } }
			      );

	my ($ref_ds, $filer, $volume);

	$ref_ds = $opts{ref_ds};
        $filer = $opts{filer};
        $volume = $opts{volume};

	store_counter($ref_ds, "compression", $filer, $volume);
}

## @name ::store_counter_volume
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# store_counter_volume (ref_ds=>$ref_ds, filer=>$filer, volume=>$volume)
sub store_counter_volume ($@)
{
	 my %opts = validate (@_, { ref_ds => { type=>HASHREF },
                                   filer => { type=>SCALAR },
                                   volume => { type=>SCALAR } }
                              );

        my ($ref_ds, $filer, $volume);

        $ref_ds = $opts{ref_ds};
        $filer = $opts{filer};
        $volume = $opts{volume};

	store_counter($ref_ds, "volume", $filer, $volume);
}

## @name ::compare_vbns
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# compare_vbns (hashref1=>$ref1, hashref2=>$ref2, desc=>$desc, state=>$state, invert=>$invert)
sub compare_vbns ($@)
{
	my %opts = validate (@_, { hashref1 => { type=>HASHREF },
				   hashref2 => { type=>HASHREF },
				   desc => { type=>SCALAR | UNDEF },
				   state => { type=>SCALAR | UNDEF },
				   invert => { type=>SCALAR | UNDEF } }
			    );

	my ($ref1, $ref2, $desc, $state, $inverted);

	$ref1 = $opts{hashref1};
	$ref2 = $opts{hashref2};
	$desc = $opts{desc};
	$state = $opts{state};
	$inverted = $opts{invert};

	$state = 1 if(!defined $state);

	my @keys1 = keys %{$ref1};
	my @keys2 = keys %{$ref2};

	if($#keys1 < $#keys2)
	{
		compare_vbns($ref2, $ref1, $desc, $state, 1);
	} else {
	printfh("\n=========Compare $desc ===========\n");
	printfh("\n====AFTER==================================BEFORE====\n") if($inverted);
	printfh("\n====BEFORE=================================AFTER====\n") if(!defined $inverted or !$inverted);

	my @arr = keys %{$ref1};
	my $cg = 1;
	my $cg_counter = 1;
	foreach my $val (0..$#arr)
	{
		if( ${$ref1}{$val}{vbn} == ${$ref2}{$val}{vbn} and
		    ${$ref1}{$val}{vvbn} == ${$ref2}{$val}{vvbn} and 
		    ${$ref1}{$val}{vbn} ne "" and ${$ref2}{$val}{vbn} ne "" and
		    ${$ref1}{$val}{vvbn} ne "" and ${$ref2}{$val}{vvbn} ne "" )
		{
			printfh("\nCG$cg - FBN $val: vbn[${$ref1}{$val}{vbn}] - vvbn[${$ref1}{$val}{vbn}] - SAME") if($VERBOSE or $state);
		}elsif(${$ref1}{$val}{vbn} eq "" and 
			${$ref1}{$val}{vbn} eq "" and $inverted) {
			printfh("\nCG$cg - FBN $val: vbn and vvbn DELETED");
		}elsif(${$ref2}{$val}{vbn} eq "" and 
			${$ref2}{$val}{vbn} eq "" and !$inverted) {
			printfh("\nCG$cg - FBN $val: vbn and vvbn TRUNCATED");
		}elsif(${$ref2}{$val}{vbn} eq "" and 
			${$ref2}{$val}{vbn} eq "" and $inverted) {
			printfh("\nCG$cg - FBN $val: vbn[${$ref1}{$val}{vbn}] and vvbn[${$ref1}{$val}{vvbn}] - APPEND"); # if($VERBOSE or $state);
		}else{
			printfh("\nCG$cg - FBN $val: vbn[${$ref1}{$val}{vbn}] - vvbn[${$ref1}{$val}{vbn}] - DIFER with vbn[${$ref2}{$val}{vbn}] - vvbn[${$ref2}{$val}{vvbn}]") if($VERBOSE or $state);
		}
		if(($cg_counter%8) == 0){
			$cg++;
			$cg_counter = 0;
			printfh("\n--------------------------");
		}
		$cg_counter++;
	}

	printfh("\n=========Compare $desc - Done===========\n");
	}
}

## @name ::compare_counter
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# compare_counter (hashref1=>$ref1, hashref2=>$ref2, desc=>$desc, state=>$state)
sub compare_counter ($@)
{
	 my %opts = validate (@_, { hashref1 => { type=>HASHREF },
                                   hashref2 => { type=>HASHREF },
                                   desc => { type=>SCALAR | UNDEF },
                                   state => { type=>SCALAR | UNDEF } }
			     );

	my ($ref1, $ref2, $desc, $state);

	$ref1 = $opts{hashref1};
        $ref2 = $opts{hashref2};
        $desc = $opts{desc};
        $state = $opts{state};

	$state = 1 if(!defined $state);

	my @differ = ();
	my @same = ();
	printfh("\n=========Compare $desc ===========\n");
	foreach my $val (keys %{$ref1})
	{
		if( ${$ref1}{$val} eq ${$ref2}{$val} or 
			${$ref1}{$val} == ${$ref2}{$val} )
		{
			printfh("\n$val - SAME") if($VERBOSE or !$state);
			push @same, $val;
		}else{
			printfh("\n$val before[${$ref1}{$val}] vs after[${$ref2}{$val}] - DIFFER, Change = ". (${$ref2}{$val} - ${$ref1}{$val})) if($VERBOSE or $state);
			push @differ, $val;
		}
	}

	printfh("\n=========Compare $desc - Done===========\n");

	if($state) { return @differ; } else { return @same; }
}

## @name ::print_vbns
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# print_vbns (ref_vbns=>$ref_vbns, desc=>$desc)
sub print_vbns ($@)
{
	my %opts = validate (@_, { ref_vbns => { type=>HASHREF },
				   desc => { type=>SCALAR | UNDEF } }
			    );

	my ($ref_vbns, $desc);
	
	$ref_vbns = $opts{ref_vbns};
	$desc = $opts{desc};

	my @arr = (keys %{$ref_vbns});
	my $count = $#arr;
	printfh("\n=========$desc===========\n");
	foreach my $val (0..$count)
	{
		printfh("Id=$val,vbn=${$ref_vbns}{$val}{vbn}, vvbn=${$ref_vbns}{$val}{vvbn}\n");
	}

	printfh("\n=========$desc - Done===========\n");
}

## @name ::print_counter
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# print_counter (ref_counter=>$ref_counter, desc=>$desc)
sub print_counter ($@)
{
	my %opts = validate (@_, { ref_counter => { type=>HASHREF },
				  desc => { type=>SCALAR | UNDEF } }
			    );

	my ($ref_counter, $desc);
	
	$ref_counter = $opts{ref_counter};
	$desc = $opts{desc};

	printfh("\n=========$desc===========\n");
	foreach my $val (keys %{$ref_counter})
	{
		printfh("\n$val => ${$ref_counter}{$val}");
	}

	printfh("\n=========$desc - Done===========\n");
}

## @name ::start_counter_manager
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# start_counter_manager (filer=>$filer, counterobj=>$counter_object)
sub start_counter_manager ($@)
{
	my %opts = validate (@_, { filer => { type=>SCALAR },
				   counterobj => { type=>SCALAR } }
			    );

	my ($filer, $counter_object);
	
	$filer = $opts{filer};
	$counter_object = $opts{counterobj};

	my $CMD_OPTION = "$counter_object";
	do_rsh($filer, CMD_START_COUNTER_MANAGER . $CMD_OPTION);
}

## @name ::stop_counter_manager
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# stop_counter_manager (filer=>$filer,counterobj=>$counter_object)
sub stop_counter_manager ($@)
{
	my %opts = validate (@_, { filer => { type=>SCALAR },
                                   counterobj => { type=>SCALAR | UNDEF } }
                            );

        my ($filer, $counter_object);

        $filer = $opts{filer};
        $counter_object = $opts{counterobj};

	my %scm = ();
	store_counter(\%scm, "", $filer, "", CMD_STOP_COUNTER_MANAGER);

	return %scm;
}

## @name ::do_sync
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# do_sync (filer=>$filer)
sub do_sync ($@)
{
	my %opts = validate (@_, { filer => { type=>SCALAR } } );

	my ($filer) = $opts{filer};

	do_rsh($filer, "priv set -q test; sync;sync");
}

## @name ::store_stats
##
## @author dl-compression-qa@netapp.com
## @status deprecated
##
# store_stats (filer=>$filer, vol=>$vol, ptr=>$ptr, options=>\@options)
sub store_stats ($@)
{
	my %args = validate (@_, { filer => { type=>SCALAR }, 
				   vol => { type=>SCALAR },
				   ptr => { type=>HASHREF },
				   options => { type=> HASHREF | UNDEF } }
			    );

	my ($filer, $vol, $ptr, $optsref, %opts);
	
	$filer = $args{filer};
	$vol = $args{vol};
	$ptr = $args{ptr};
	$optsref = $args{options};
	%opts = %$optsref;
	
	#my $opt = (!defined $opts{flags} or $opts{flags} eq "")? " -S " : $opts{flags};
	my $opt = " -S ";
	open(DF, "rsh $filer -l root: df $opt /vol/$vol |") || die "df";
	while (<DF>) {
		print;
#nss-u155> df -S
#Filesystem                used       compressed         a-sis     %saved
#/vol/vol0/            98473268                0             0         0%
		if(/(\w+)+(\s)+(\d+)(\s)+(\d+)(\s)+(\d+)/){
			(${$ptr}{vol}, ${$ptr}{used}, ${$ptr}{compressed}, ${$ptr}{saved}) = split;
			#(${$ptr}{vol}, ${$ptr}{used}, ${$ptr}{compressed}, ${$ptr}{saved}) = ($1, $2, $3, $5);
			#print "\n($1, $2, $3, $5)";
			last if ${$ptr}{vol} eq "/vol/$vol";
		}
	}
	close(DF);

	logcomment("Stats = ${$ptr}{vol}, ${$ptr}{used}, ${$ptr}{compressed}, ${$ptr}{saved}");
}





1;
