#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/config_base.pm#2 $
#
#Copyright (c) 2002 Network Appliance, Inc.
#All rights reserved.
#
## @summary Library of functions used by other configuration packages 

## @description
## This script defines common NATE parameters 
## that are independent of FILER or CLIENT parameters.
##
## It also defines values for test script return codes,
## a set of functions to exit from a test script,
## and a function to announce detected burts.
## The output from these functions is compatible with the nuffo.thpl
## program for collecting and storing test results as a NATE POSTTEST.
## @usage
##   use config_base ;
##
## @param ITERATIONS Optional: 
## Looping on these tests is stopped after this number of iterations. Default: 0.
## @param RUNLENGTH Optional: 
## Looping on these tests is stopped after this number of seconds. Default: 0.
##
## @author Charlie Hedstrom (hedstrom@netapp.com)
## @keywords burt return 
## @status public
## @change 2002-02-12 hedstrom: created
## @change 2002-02-15 hedstrom: added ONTAP command line return codes
## @change 2002-03-14 hedstrom: use strict
## @change 2002-03-29 hedstrom: function name conventions
## @change 2002-04-08 hedstrom: add use Tharn for CONFIGS params
## @change 2002-04-15 hedstrom: allow both / and _ config type separators
## Add additional support for burt processing
## @change 2002-04-18 hedstrom: add burt_needed support to ExitError and ExitSoftError functions
## @change 2002-05-21 hedstrom: add caller to ExitError and ExitSoftError
## @change 2002-07-21 hedstrom: report all callers until Tharn is mentioned in Exit functions, fix redundant CONFIG: text
## @change 2002-07-26 hedstrom: support burt number arguments on ExitError and ExitSoftError functions
## @change 2002-07-30 hedstrom: fix FILER_CLUSTER_SUFFIX use in test bed plans
## @change 2002-07-31 hedstrom: fix Perl warning in -burt handling
## @change 2002-08-05 hedstrom: fix bug in -burt arg handling
## @change 2002-10-07 hedstrom: avoid Perl error if TESTPLAN_CONFIGS is empty
## @change 2002-11-05 hedstrom: calculate combined test bed configuration types like nuffo.thpl
## @change 2002-11-21 hedstrom: issue Symptoms not detected message for every verification that passes until an error is seen
## Issue Symptoms detected message for every verification that fails.
## This allows nuffo.thpl to detect scripts that do not track burts
## @change 2003-02-17 hedstrom: add implied configuration types, and filer models
## @change 2003-03-27 hedstrom: fix parsing of TESTPLAN_CONFIGS with underscores
## @change 2003-06-13 hedstrom: add Jivaros configuration types
## @change 2003-06-16 hedstrom: add testbed_config_types function
## @change 2003-06-17 hedstrom: use ConfigTypes package
## @change 2003-07-23 hedstrom: fix $FILER_MODEL use in TESTPLAN_CONFIGS
## @change 2003-09-08 hedstrom: increase power on timeout to 4 minutes
## @change 2003-09-17 hedstrom: perform power operations in parallel on multiple hosts
## @change 2003-09-25 vadim: use Net::Ping instead of system cmd in host_is_alive()
## @change 2004-06-30 olinger: Adjust regexp on "not found" error from power_cycle.
## @change 2004-07-07 hedstrom: fix for new power_cycle output
## @change 2004-08-20 hedstrom: increase host_wait timeout to 180 seconds (burt139686)
## @change 2005-02-01 hedstrom: drop processing of network group's parameters
## @change 2006-03-01 korry: add support for rpower output.
## @change 2005-02-02 gajria : removed NIS and DNS params, a new Network/commonParams.pl will hold these variables

use strict;
use Net::Ping;

package main;
use vars qw(
   $ITERATIONS $RUNLENGTH
   $FILER_MODE $FILER_MODEL $FILER_CLUSTER_SUFFIX
   $TESTBED $TESTBED_CONFIGS $TESTPLAN_CONFIGS
   );


use Tharn;
# Parameters common to all NATE configurations
param ('RUNLENGTH',  '-default', 0 );
param ('ITERATIONS', '-default', 0 );
param ('FILER_MODE', '-default', 'mono' );
param ('FILER_MODEL', '-default', undef );
param ('FILER_CLUSTER_SUFFIX',  -default, "c");
if ($FILER_MODE ne 'cf') {
   $FILER_CLUSTER_SUFFIX = '';
}
my $filer_model = undef;
if (defined($FILER_MODEL)) {
    $filer_model = $FILER_MODEL.$main::FILER_CLUSTER_SUFFIX;
} else {
    # burt150903 requires that FILER_MODEL not even be initialized to undef
    param ('FILER_MODEL', '-unset');
}

param ('TESTBED',  '-default', '' );
param ('TESTBED_CONFIGS',  '-default', '' );
param ('TESTPLAN_CONFIGS', -default, '');
if ($TESTPLAN_CONFIGS ne '') {
   # Perform late variable substitution to get current values (not initial values)
   no strict 'refs';
   while ($TESTPLAN_CONFIGS =~ s/\$(\w+)/${$1}/g) {};	# Substitute parameters
   use strict 'refs';
   $TESTPLAN_CONFIGS =~ s#//#/#g;
   logcomment("TESTPLAN_CONFIGS=$TESTPLAN_CONFIGS","TESTBED_CONFIGS=$TESTBED_CONFIGS");
}

package config_base;

use vars qw(@ISA @EXPORT
   );

require Exporter;
@ISA = ('Exporter');

@EXPORT = qw(
          applies_to_testbed

          client_param_get

          host_is_alive host_power host_wait

          EXIT_PASS EXIT_NOT_APPLICABLE EXIT_ERROR EXIT_SCRIPT_ERROR EXIT_CONFIG_ERROR
          EXIT_CHECKME

          burt_begin burt_end
          burt_fixed burt_found burt_get burt_needed

          burt_read burt_state

          ExitNA ExitError ExitSoftError ExitScriptError ExitConfigError
          );

use English qw(-no_match_vars);
use Carp;
use Burt;
use Tharn;
use SharedTestBed::ConfigTypes;

sub host_power(@);

## @name ::applies_to_testbed
## @summary Skips the current test case if the test bed does not meet test plan needs
## @description
## If both the TESTPLAN_CONFIGS and TESTBED_CONFIGS parameters have non-empty values,
## its entries are compared against the TESTBED_CONFIGS values.
## If no entries match the test case is skipped (considered as Not Applicable)
sub applies_to_testbed  (;$@) {
   my $testbed_configs = shift;
   my $testplan_configs = shift;

   if (!defined($testbed_configs)) {
      $testbed_configs = $main::TESTBED_CONFIGS;
   }
   if (!defined($testplan_configs)) {
      $testplan_configs = $main::TESTPLAN_CONFIGS;
   }
   my $test_config = '';
   if ($testplan_configs ne '' && $testbed_configs ne '') {
      if ($testplan_configs =~ m/^ANY$/i) {
         # All tests are run once, anywhere
         $test_config = 'any';
      } else {
         # TESTPLAN_CONFIGS holds a list of configuration type combinations
         my @testplan_combo = split(",",$testplan_configs);

         # TESTBED_CONFIGS holds a list of configuration types
         param ('TESTBED_CONFIGS', -default, '');
         my @testbed_configs = split(",",$testbed_configs);

         # Add the filer model as a configuration types
         @testbed_configs = ( @testbed_configs, $filer_model);
         
         # Pick up configuration types implied from the test bed specification
         @testbed_configs = testbed_config_types(@testbed_configs);

         # Add the test bed name as another configuration type
         # (But do not allow a test bed name to imply other types)
         if ($main::TESTBED ne '') {
            @testbed_configs = ( @testbed_configs, $main::TESTBED);
         }

         #print("test bed configs=".join(' ',@testbed_configs)."\n");#DEBUG

         # Look for duplicate TESTPLAN_CONFIGS entries
         # Make sure all specified configuration types are valid
         # Determine that the current test bed has at least one of these combinations
         my %testplan_config_entry;
         foreach my $this_testplan_combo (@testplan_combo) {
             if (defined($testplan_config_entry{"$this_testplan_combo"})) {
                 if ($testplan_config_entry{"$this_testplan_combo"} == 1) {
                     logresult('WARN',"TESTPLAN_CONFIGS has more than one instance of: $this_testplan_combo");
                 } else {
                     $testplan_config_entry{"$this_testplan_combo"}++;
                 }
                 next;
             } else {
                 $testplan_config_entry{"$this_testplan_combo"} = 1;
             }
            # See if every configuration type in this combination is included
            # in the current test bed
            my @testplan_types = split("/|_",$this_testplan_combo);
            my $testplan_type;
            #print("test plan types=".join(' ',@testplan_types)."\n");#DEBUG
            my @invalid_types;
            foreach $testplan_type (@testplan_types) {
               if (!testbed_config_type_check($testplan_type)) {
                  push(@invalid_types,$testplan_type);
               }
            }
            if (@invalid_types) {
               ExitScriptError("An entry in TESTPLAN_CONFIGS has the following invalid configuration types: ".join(',',@invalid_types));
            } else {
               my $has_all = 1;
               foreach $testplan_type (@testplan_types) {
                  my $got_testplan_type = 0;
                  foreach my $this_config (@testbed_configs) {
                     if ($this_config eq $testplan_type) {
                        #print("Got match on '$this_config'\n");#DEBUG
                        $got_testplan_type = 1;
                        last;
                     }
                  }
                  if ($got_testplan_type == 0) {
                     #print("Found no test bed configuration match on '$testplan_type'\n");#DEBUG
                     $has_all = 0;
                     last;
                  }
               }
               if ($has_all) {
                  $test_config = $this_testplan_combo;
                  last;
               }
               if ($test_config ne '') {
                  last;
               }
            }
         }
         if ($test_config eq '') {
            ExitNA ("Test bed configuration does not include any entries from TESTPLAN_CONFIGS=".$testplan_configs);
         } else {
            #print("Test configuration is $test_config\n");#DEBUG
         }
      }
   }
   # Return the matching test plan configuration
   if (wantarray) {
      my @return_configs = split("[_/]",$test_config);
      return @return_configs;
   } else {
      return $test_config;
   }
}

## @name ::client_param_get
## @summary Returns the param name that defines the specified client id
## @description
## Returns the 'CLIENTn' param name that corresponds to the specified client name.
## Returns undef if no such name can be found.
## @arg $clientname Required: The client name
sub client_param_get  ($;$) {
   my($clientname) = shift;
   my($use_clientn) = shift; # if this is set, the CLIENTn entry is preferred over CLIENT

   if (!defined($use_clientn)) {
       $use_clientn = 0; # Return 'CLIENT' if it applies
   }
   if (!$use_clientn) {
       # see if basic CLIENT is the one named
       if ($main::CLIENT eq $clientname) {
          return 'CLIENT';
       }
   }

   # Look for all defined clients
   my $clientnum = 1;
   while (1) {

      no strict 'refs';
      param('CLIENT'.$clientnum,'-default',undef);
      if (!defined(${'main::CLIENT'.$clientnum}) or ${'main::CLIENT'.$clientnum} eq '') {
         $clientnum = 0;
         last;
      } elsif (${'main::CLIENT'.$clientnum} eq $clientname) {
         return 'CLIENT'.$clientnum;
      }
      use strict 'refs';

      $clientnum++;
   }
   # see if basic CLIENT is the one named
   if ($main::CLIENT eq $clientname) {
      return 'CLIENT';
   }
   # Either no clients are defined, or none match the name given
   return undef;
}

=head2 host_is_alive($;$)

    Determine if a host is alive (responds to ping).
    
    Parameters include:
        $address The host name or IP address or a hostspec
        $timeout How long to wait for reply from $address. Optional
     
=cut
sub host_is_alive($;$) {
    my ($address, $timeout) = @_;

    # deal with hostspecs.  It is unfortunate that this function is
    # ambiguous as to whether it takes a hostname or a hostspec, since
    # a hostspec which looks like it could be a hostname isn't
    # guaranteed to be usable as a hostname.

    if (UNIVERSAL::isa($address,"Hostrec") or $address =~ /[;{}]/) {
	my $host = Tharn::host($address);
	$address = $host->default_ip();
    }

    # A ping time-out parameter is required since some pings may not end
    $timeout = 15 unless defined($timeout);

    # Ping by attempting to connect to ssh (port 22), since GX hosts
    # filter out the default choice of echo (7), or at least they had
    # in the past.  Note that the host need not be set up to serve
    # ssh, since "connection refused" is a successful ping, and so we
    # need only choose a port which is not filtered/firewalled.  The
    # port is referred to by number, since getsrvbyname does not know
    # "ssh" on Windows.
    #
    # Use the external ping system call to perform ping to avoid some
    # ONTAP firewall issues.

    my $ping = Net::Ping->new('external');
    $ping->{port_num} = 22;
    my $is_alive = $ping->ping($address, $timeout);
    return $is_alive;
}

## @name ::host_power
## @summary Power a host (filer, client, switch) off, on or both
## @arg $action Optional: defaults to 'reboot' ('off' then 'on')
## @arg $hc Required: A list of host or host connection objects
sub host_power(@) {
   my $action = 'reboot'; # The default action
   my @hosts;   # The array of host connections

   foreach (@_) {
      if ($_ =~ m/^reboot$|^on$|^off$/i) {
         $action = $_ ;
      } else {
         push(@hosts,$_);
      }
   }

   my @hostids;  # List of all host ID or IPs to verify
   if (!defined($action)) {
      $action = 'reboot';
   }

   if ($action =~ m/^reboot$/i) {
      host_power('off',@hosts);
      host_power('on',@hosts);
   } else {
      # First perform the requested operation of all hosts
      # Then verify all hosts have completed
      foreach my $hc (@hosts) {
         my $power_cmd = '';
         my $hh = $hc->{'hostp'};
         if (!defined($hh)) {
            # Assume a host object was passed instead of a connection object
            $power_cmd = $hc->powercmd();
            $hh = $hc;
         } else {
            $power_cmd = $hh->powercmd();
         }
         if (!$power_cmd) {
            # Assume the 'power_cycle' command if none is defined
            $power_cmd = 'power_cycle';
         }

         my $hostid = $hh->id();
         my $dual_power_supply = 0;   # Assume only single power supply is allowed
         logcomment("Powering $action host $hostid");

         # Format the power command
         # Three known power commands are handled
         # RTP power_cycle
         # usage: /usr/rtp/scripts/power_cycle machine <action>
         #    /usr/rtp/scripts/power_cycle rpb[1|2|3] <port> <action>
         #    /usr/rtp/scripts/power_cycle rtp_rpb[1|2|3] <port> <action>
         #    where <action> = on | off | reboot (default is reboot)
         # RTP rpb
         # usage: /usr/rtp/scripts/rpb hostname
         # Sunnyvale power_cycle
         # usage: /usr/local/bin/power_cycle <filer_name> [action]
         #    /usr/local/bin/power_cycle <power_switch_name> <port> [action]
         #
         #    where [action] = on | off | reboot (default is reboot)

         # Unfortunately $powercmd may have full command syntax
         $power_cmd =~ s/^\s//g;                       # strip leading spaces
         $power_cmd =~ s/\s$//g;                       # strip trailing space
         $power_cmd =~ s/\s+on$|\s+off$|\s+reboot$//;  # strip action
         # Determine if machine or 'switch port' is specified
         if ($power_cmd =~ m/\s+\S+\s+/) {
            # Assume this is a power_switch port command
         } else {
            $power_cmd =~ s/\s+$hostid$//;                # strip hostid
            # Replace rpb with basic power_cycle from RTP
            $power_cmd =~ s#\S*rpb\s*#/usr/rtp/scripts/power_cycle# ;
            # Add hostid
            $power_cmd = "$power_cmd $hostid";
            # flag for dual power supply support
            $dual_power_supply = 1; # If request fails, try with '1' and '2' suffix
         }

         my $powerlog = `$power_cmd $action`;
         chomp($powerlog); # Remove trailing whitespace, newlines
         if ($powerlog =~ qr/$hostid not found/) {
            if ($dual_power_supply) {
               my $power_cmd1 = $power_cmd.'1';
               my $powerlog1 = `$power_cmd1 $action`;
               chomp($powerlog1); # Remove trailing whitespace, newlines
               if ($powerlog1 =~ qr/\S+ not found/) {
                  ExitConfigError("Host $hostid has not been defined for the powercmd: $power_cmd $action");
               } else {
                  # Looks like first power supply was defined, try second one
                  my $power_cmd2 = $power_cmd.'2';
                  my $powerlog2 = `$power_cmd2 $action`;
                  chomp($powerlog2); # Remove trailing whitespace, newlines
                  if ($powerlog2 =~ qr/\S+ not found/) {
                     logresult('WARN',"Second power supply has not been defined for the powercmd: $power_cmd2 $action");
                  } elsif ($powerlog2 =~ /This will take about 10 seconds to return/) {
                  } elsif ($powerlog2 =~ /Processing - please wait/) {
                  } elsif ($powerlog2 =~ /spawn telnet/) {
                  } elsif ($powerlog2 =~ m/^\n*\r*$/) {
                  } else  {
                      ExitError("Error during '$power_cmd $action': $powerlog2");
                  }
               }
            } else { # Dual power supplies is not an option
               ExitConfigError("Host $hostid has not been defined for the powercmd: $power_cmd $action");
            }
         } elsif ($powerlog =~ /This will take about 10 seconds to return/) {
         } elsif ($powerlog =~ /Processing - please wait/) {
         } elsif ($powerlog =~ /spawn telnet/) {
         } elsif ($powerlog =~ /found entry/) {
         } elsif ($powerlog =~ m/^\n*\r*$/) {
         } else  {
            ExitError("Error during powercmd $action: $powerlog");
         }
         push(@hostids,$hostid);  # Remember which hosts need to be verified
      }
      # OK, time to verify the operation worked on each host
      my $power_time = time();
      foreach my $hostid (@hostids) {
         if ($action =~ m/^on$/i) {
            while () {
               sleep(10);
               if (host_is_alive($hostid)) {
                  my $nowtime = time();
                  my ($sec,$min,$hour,$mday,$mon,$year) = (localtime($nowtime))[0..5];
                  $year = $year + 1900;
                  my $mm = $mon + 1;
                  # And format it for two digits
                  $mm = '0'.$mm;     
                  my $dd = '0'.$mday;
                  $dd =~ s/.*(\d\d)$/$1/;
                  my $hh = '0'.$hour;
                  $hh =~ s/.*(\d\d)$/$1/;
                  my $nn = '0'.$min;
                  $nn =~ s/.*(\d\d)$/$1/;
                  my $ss = '0'.$sec;
                  $ss =~ s/.*(\d\d)$/$1/;
                  $mm =~ s/.*(\d\d)$/$1/;
                  my $time = "$year$mm$dd $hh$nn$ss";
                  logcomment("Host $hostid was alive at $time");
                  last;
               }
               if (time() > $power_time+480) {
                  logresult('WARN',"Host $hostid could not be ping'ed 8 minutes after it was powered on");
                  last;
               }
            }
         } elsif ($action =~ m/^off$/i) {
sleep 5;
if(0)
{
            while () {
               sleep(5);
               if (!host_is_alive($hostid)) {
                  logcomment("Host $hostid has been powered off.");
                  last;
               }
               if (time() > $power_time+90) {
                  ExitConfigError("Host $hostid appears live after it was powered off. Please correct remote power management.");
               }
            }

}
         }
      }

      if ($action =~ m/^on$/i) {
         sleep(60);  # Wait a little more for rsh daemon to start
      }
   }

}

## @name ::host_wait
## @summary Wait for a host to be alive (responds to ping)
## @arg $address The host name or IP address or a hostspec
## @arg $timeout How long to wait for reply from $address. Optional.
sub host_wait($;$) {
   my ($address, $timeout) = @_;

   if (!defined($timeout)) {
      $timeout = 180;    # (burt139686)
   }
   my $alive = 0;
   my $endtime = time() + $timeout;

   # deal with hostspecs
   $address = (split(';', $address))[0] if ($address =~ ';');

   while (time() < $endtime) {
      if (host_is_alive($address)) {
         $alive = 1;
         last;
      } else {
         sleep(10);
      }
   }
   return $alive;
}

1;
