## @summary DataAnalyzer test library to handle SCA data analyzer functions.
##
## @description
##
## $Id: //depot/prod/test/Rsierranevada/lib/configadvisor/DataAnalyzer.pm#1 $<br>
## $DateTime: 2013/01/22 11:22:46 $<br>
## $Author: build $<br>
##
## With this module various task which can be performed
## with DataAnalyzer.
##
## @usage
##   use DataAnalyzer;
##
## @keywords Tools
## @dependencies none
##
## @author Rajiv Gupta (rajivg@netapp.com)
## @status public
## @change 2006-09-11 rajivg: created
##

#Package Name
#################
package DataAnalyzer;

#Package Requirements
######################
use strict;
use Tharn;
use Data::Dumper;    #use to print the hash usage print Dumper($var)

#Load San Packages
############################
use San::SanLog;
use Error;
use San::SanFiler;
use San::SanClient;
use San::SanSwitch;
use San::SanUtil;
use configadvisor::SCAUtil;
use configadvisor::DataCollector;

#Load Java Loader Module to load Java Objects
#############################################
use configadvisor::Java_Loader;

use vars qw (@ISA);

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

#######################################################
#####################Constructor#######################
#######################################################
## @name ::new
## @description Create a new DataCollector Object. It will initializes the data structure,
## get DataCollector specific information (populate global variables)
## @synopsis
## $obj = DataAnalyzer->new();
##
## @example
## $dc_obj = DataAnalyzer->new()
##
sub new () {
    my ($class, %args) = @_;

    # Declare and bless empty hash for DataCollector.
    my $self = {};
    bless($self, $class);
    $self->_clear_err();

    #############################################################
    ### Declare global variables Specific to Config Advisor######
    #############################################################

    my ($dc_obj,      $xml_parsed, $xml_unparsed, $clients_da, $client_vms, $filers_da,
        $switches_da, $daclient,   $protocol,     $pkgname,    $pkgdir,     $lab_info);

    my %san_client_args_hash = ();

    ###############################################################
    ### Get supplied arguments ###################################
    ###############################################################

    foreach (keys %args) {
        if (/-?datacollector$/i) {
            $dc_obj = $args{$_};
            if (ref($dc_obj) !~ /DataCollector/) {
                $self->_set_err(
                              'Error:DataAnalyzer.pm:new():Supplied DC object $dc_obj is not the Datacollector object');
                return $self;
            }
        } elsif (/-?xmlparsed/i) {
            if (not length($xml_parsed = $args{$_})) {
                $self->_set_err("Error:DataAnalyzer.pm:new():Value for \"xmlparsed\": is not supplied");
                return $self;
            }
        } elsif (/-?xmlunparsed/i) {
            if (not length($xml_unparsed = $args{$_})) {
                $self->_set_err("Error:DataAnalyzer.pm:new():Value for \"xmlunparsed\": is not supplied");
                return $self;
            }
        } elsif (/-?daclient/i) {
            if (not length($daclient = $args{$_})) {

                #                DA client not required at this time. I will be requred when GUI will be used.
                #                $self->_set_err("Error:DataAnalyzer.pm:new():Value for \"scaclient\": is not supplied");
                #                return $self;
            }
        } elsif (/-?clients/i) {
            $clients_da = $args{$_};
        } elsif (/-?client_vms/i) {
            $client_vms = $args{$_};
            if (length($client_vms)) {
                %san_client_args_hash = (-mp_type => $client_vms);
            }
        } elsif (/-?filers/i) {
            $filers_da = $args{$_};
        } elsif (/-?switches/i) {
            $switches_da = $args{$_};
        } elsif (/-?protocol/i) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error:DataAnalyzer.pm:new()::Protocol not specified");
                return $self;
            }
        } elsif (/-?pkgname/i) {
            if (not length($pkgname = $args{$_})) {

                #                $self->_set_err("Error:DataAnalyzer.pm:new():Data Analyzer package name is not provided.");
                #                return $self;
            }
        } elsif (/-?pkgdir/i) {
            if (not length($pkgdir = $args{$_})) {

                #                $self->_set_err("Error:DataAnalyzer.pm:new():Data Analyzr package directory is not provided.");
                #                return $self;
            }
        } elsif (/-?lab/i) {
            $lab_info = $args{$_};
        }
    }

    #Check if all mandatory parameters are provided.

    if (not length($xml_parsed) and not length($xml_unparsed)) {
        $self->_set_err("Error:DataAnalyzer.pm:new():Data Analzyer unparsed and parsed xml is not supplied.");
        return $self;
    }

    $self->{'XML_PARSED'}   = $xml_parsed;
    $self->{'XML_UNPARSED'} = $xml_unparsed;

    #########################################################
    ##
    ## Intentionally left blank for DA client implementation.
    ##
    ##########################################################

    my $san_obj;    # Empty hash reference for storing san objects.

    if (defined $dc_obj) {
        $self->{'CLIENTS_CNT'}    = $dc_obj->{'CLIENTS_CNT'};
        $self->{'CLIENTS_STR'}    = $dc_obj->{'CLIENTS_STR'};
        $self->{'SWITCHES_CNT'}   = $dc_obj->{'SWITCHES_CNT'};
        $self->{'SWITCHES_STR'}   = $dc_obj->{'SWITCHES_STR'};
        $self->{'FILERS_CNT'}     = $dc_obj->{'FILERS_CNT'};
        $self->{'FILERS_STR'}     = $dc_obj->{'FILERS_STR'};
        $self->{'PKGDIR'}         = $dc_obj->{'PKGDIR'};
        $self->{'PKGNAME'}        = $dc_obj->{'PKGNAME'};
        $self->{'LAB_INFO'}       = $dc_obj->{'LAB_INFO'};
        $self->{'SAN_COMPONENTS'} = $dc_obj->{'SAN_COMPONENTS'};
    } else {

        #Check atleast one component is present.
        if (    not length($clients_da)
            and not length($filers_da)
            and not length($switches_da)) {
            $self->_set_err(
                "Error:DataAnalyzer.pm:new():All clients, filer and switch strings are 0, atleast one component should be provided."
            );
            return $self;
        }

        # Create Clients, Filer and Switch array.
        my @dc_node_clients  = split(/:/, $clients_da);
        my @dc_node_filers   = split(/:/, $filers_da);
        my @dc_node_switches = split(/:/, $switches_da);

        my $numofclients  = scalar(@dc_node_clients);
        my $numofswitches = scalar(@dc_node_switches);
        my $numoffilers   = scalar(@dc_node_filers);

        $self->{'CLIENTS_CNT'}  = $numofclients;
        $self->{'CLIENTS_STR'}  = $clients_da;
        $self->{'SWITCHES_CNT'} = $numofswitches;
        $self->{'SWITCHES_STR'} = $switches_da;
        $self->{'FILERS_CNT'}   = $numoffilers;
        $self->{'FILERS_STR'}   = $filers_da;
        $self->{'PKGDIR'}       = $pkgdir;
        $self->{'PKGNAME'}      = $pkgname;
        $self->{'LAB_INFO'}     = $lab_info;

        ############################################################
        #### Populate DA hash with Clients #######################
        ############################################################

        foreach my $client (@dc_node_clients) {
            $self->sanlog(-msg => "Creating Client object for client '" . $client . "'");
            print ntest::NTEST_STDOUT "Creating client object : $client\n";
            %san_client_args_hash = (%san_client_args_hash,
                                     -host      => $client,
                                     -protocol  => $protocol,
                                     -conn_type => "THARNHOST",);
            my $client_obj = SanClient->new(%san_client_args_hash);
            if ($client_obj->error) {
                my $errmsg = join("\n", $client_obj->errmsg());
                $self->_set_err("Error:DataCollector.pm:new():", "Instantiating client object for $client.", "$errmsg");
                return $self;
            }
            push @{$san_obj->{clients}}, $client_obj;    #Populate San Client array.
        }

        ############################################################
        #### Populate DA hash with switches #######################
        ############################################################

        foreach my $switch (@dc_node_switches) {
            $self->sanlog(-msg => "Creating Switch object for switch '" . $switch . "'");
            print ntest::NTEST_STDOUT "Creating switch object : $switch\n";
            my $switch_obj = new SanSwitch(-host       => $switch,
                                           -persistent => 1,
                                           -conn_type  => "THARNHOST");
            if ($switch_obj->error) {
                my $errmsg = join("\n", $switch_obj->errmsg());
                $self->_set_err("Error:DataCollector.pm:new():", "Instantiating switch object for $switch.", "$errmsg");
                return $self;
            }
            $switch_obj->disconnect();
            push @{$san_obj->{switches}}, $switch_obj;
        }

        ############################################################
        #### Populate DA hash with Filers    #######################
        ############################################################

        foreach my $filer (@dc_node_filers) {
            $self->sanlog(-msg => "Creating filer object for filer '" . $filer . "'");
            print ntest::NTEST_STDOUT "Creating filer object : $filer\n";
            my $filer_obj = new SanFiler(-filer => $filer);
            if ($filer_obj->error) {
                my $errmsg = join("\n", $filer_obj->errmsg());
                $self->_set_err("Error:DataCollector.pm:new():", "Instantiating filer object for $filer.", "$errmsg");
                return $self;
            }
            push @{$san_obj->{filers}}, $filer_obj;
        }

        ############################################################
        #### Populate DA hash with SAN components  #################
        ############################################################

        $self->{'SAN_COMPONENTS'} = $san_obj;
    }

    return $self;
}

###############################################################################
###############################################################################
#  Required Java 1.5 to work. It won't work with java 1.6.
#
###############################################################################
###############################################################################

sub get_unparsed_command_value() {

    my ($self, $component, $cmd, $java_int) = @_;
    my ($out);

    $self->_clear_err();

    &main::tstlog(-comment => "Java Loaded Successfully");
    my @in_args = $self->{'XML_UNPARSED'};
    push @in_args, $component;
    push @in_args, $cmd;

    #$out = $java_int->hello();
    $out = $java_int->get_method("getCmdOutput",@in_args);
    
    if ($java_int->error) {
        my $errmsg = $java_int->errmsg();
        $self->_set_err("ERROR:DataAnalyzer:get_unparsed_command_value:$errmsg");
    }
    return $out;
}

# This function is implemented because of burt 322402 We cannot pass the object to subtest. Because of that We have
# to implement this functionality in the object itself. This is basically the verification functionality which shoud
# be the part of test script.
sub command_validate() {

    my ($self, $component, $component_type, $cmd, $java_int) = @_;

    $self->_clear_err();

    #Get the output of the command from the unparsed xml.
    my $cmd_out_xml = $self->get_unparsed_command_value($component, $cmd, $java_int);
    my @cmd_xml_array_print = split "\n", $cmd_out_xml;
	if(@cmd_xml_array_print > 50){
		$self->sanlog(-msg => "The Output xml of the command $cmd is :\n");
		for (my $i=0; $i<50; $i++){
			$self->sanlog(-msg => "$cmd_xml_array_print[$i]\n"); #The output is very long it failed to display on the log file.
		}
		$self->sanlog(-msg => ".... The Output xml of the command $cmd is more than 50 lines");
	}else{
		$self->sanlog(-msg => "The Output xml of the command $cmd is :\n $cmd_out_xml\n");
	}

    # $self->sanlog(-msg => "The Output xml of the command $cmd is : $cmd_out_xml\n"); #The output is very long it failed to display on the log file.
    if ($cmd_out_xml =~ /DAError:/ || $self->error) {
        my $errmsg = $cmd_out_xml;
        $errmsg = $errmsg . $self->errmsg() if($self->error);
        $self->_set_err("Error:DataCollector.pm:command_validate():", "XML File read Error", "$errmsg");
        return $self;
    }

    if ($component_type =~ /client/i) {
        my $cmd_out_actual = undef;    # to Store actual commond output
        my $client         = undef;
        foreach my $c (@{$self->{'SAN_COMPONENTS'}->{'clients'}}) {
            if (   ($c->get_connection()->{hostp}->{default_ip} =~ /$component/)
                or ($c->hostname() =~ /$component/)) {
                $client = $c;
                last;
            }
        }


        if (defined $client) {
            if ($client->hosttype() =~ /windows/) {
                $self->_set_err("Error:DataCollector.pm:command_validate():",
                                "Windows is not the supported host in this automation");
                return $self;

            } else {

                #check if Client object is Sanclient the to get array ref use array and get the array ref.
                if (ref($client) =~ /SanClient/){
                	my @arry_out = $client->execute($cmd);
                	($cmd_out_actual) = \@arry_out;
                }else{
                    ($cmd_out_actual) = $client->execute($cmd);	
                }
                

                if ($client->error) {

                    #check if cmd output is zero lenth, and xml out is also zero then no error.
                    if (not @$cmd_out_actual) {
                        if (not length $cmd_out_xml) {
                            return $self;
                        } else {
                            $self->_set_err(
                                "Error:DataCollector.pm:command_validate():",
                                "NATE Command output is error and blank that is not matched with cmd out in XML file.\n",
                                "$cmd_out_xml");
                            return $self;
                        }
                    } else {
                        my @cmd_xml_array = split "\n", $cmd_out_xml;
                        my $out = &SCAUtil::compare_arrays($cmd_out_actual, \@cmd_xml_array);
                        if ($out == 1) {
                            return $self;
                        } else {
                            $self->_set_err("Error:DataCollector.pm:command_validate():",
                                    "Command output is error that is not matched the command output in the XML file.\n",
                                    "$out");
                            return $self;
                        }
                    }
                }

                my @cmd_xml_array = split "\n", $cmd_out_xml;

                my $out = &SCAUtil::compare_arrays($cmd_out_actual, \@cmd_xml_array);

                if ($out == 1) {
                    return $self;
                } else {
                    $self->_set_err(
                        "Error:DataCollector.pm:command_validate():",
                        "Command output is not matched. match command output executed from the NATE log file and match with XML file.\n",
                        "$out");
                    return $self;
                }
            }
        } else {
            $self->_set_err("Error:DataCollector.pm:command_validate():",
                            "Component $component not found in DA object,",);
            return $self;
        }
    } elsif ($component_type =~ /switch/i) {
        my $switch;
        my @cmd_out_actual;    # to Store actual commond output
        foreach my $s (@{$self->{'SAN_COMPONENTS'}->{'switches'}}) {
            if (   ($s->get_connection()->{hostp}->{default_ip} =~ /$component/)
                or ($s->hostname() =~ /$component/)) {
                $switch = $s;
                last;
            }
        }

        if (defined $switch) {
            (@cmd_out_actual) = $switch->execute($cmd);

            ##Due to burt 325379 for brocade switch add the command with the switch output. When this burt will fix remove the below code.
            if ($switch->get_connection()->{hostp}->{hosttype} =~ /brocade/) {
                unshift @cmd_out_actual, $cmd;
            }
            ##Delete till here

            if ($switch->error) {

                #check if cmd output is zero lenth, and xml out is also zero then no error.
                if (not @cmd_out_actual) {
                    if (not length $cmd_out_xml) {
                        return $self;
                    } else {
                        $self->_set_err(
                               "Error:DataCollector.pm:command_validate():",
                               "NATE Command output is error and blank that is not matched with cmd out in XML file.\n",
                               "$cmd_out_xml");
                        return $self;
                    }
                } else {
                    my @cmd_xml_array = split "\n", $cmd_out_xml;
                    my $out = &SCAUtil::compare_arrays(\@cmd_out_actual, \@cmd_xml_array);
                    if ($out == 1) {
                        return $self;
                    } else {
                        $self->_set_err("Error:DataCollector.pm:command_validate():",
                                    "Command output is error that is not matched the command output in the XML file.\n",
                                    "$out");
                        return $self;
                    }
                }
            }

            my @cmd_xml_array = split "\n", $cmd_out_xml;

            my $out = &SCAUtil::compare_arrays(\@cmd_out_actual, \@cmd_xml_array);

            if ($out == 1) {
                return $self;
            } else {
                $self->_set_err("Error:DataCollector.pm:command_validate():",
                                "Command output is not matched.\n", "$out");
                return $self;
            }
        } else {
            $self->_set_err("Error:DataCollector.pm:command_validate():",
                            "Component $component not found in DA object,",);
            return $self;
        }
    } elsif ($component_type =~ /filer/i) {
        my $filer = undef;
        my $cmd_out_actual;    # to Store actual commond output
        foreach my $f (@{$self->{'SAN_COMPONENTS'}->{'filers'}}) {
            if (   ($f->connection()->{hostp}->{default_ip} =~ /$component/)
                or ($f->connection()->{hostp}->{hostname} =~ /$component/)) {
                $filer = $f;
                last;
            }
        }

        unless (defined $filer) {
            $self->_set_err("Error:DataCollector.pm:command_validate():",
                            "Component $component not found in DA object,",);
            return $self;
        }

        $self->sanlog(-msg => "Filer command executing $cmd\n");

        ## Set Transport type based on the command

        my ($errorno, $reason, $cmd_out) = &SCAUtil::filer_zapi_cmd($filer, $cmd);

        my @cmd_xml_array = split "\n", $cmd_out_xml;
        my @cmd_act_array = split "\n", $cmd_out;

        my $out = &SCAUtil::compare_arrays(\@cmd_act_array, \@cmd_xml_array);

        if ($out == 1) {
            return $self;
        } else {
            $self->_set_err("Error:DataCollector.pm:command_validate():", "Command output is not matched.\n", "$out");
            return $self;
        }
    } else {
        $self->_set_err("Error:DataCollector.pm:command_validate():", "Invalid component type $component_type");
        return $self;
    }

}
1;

###############################################################################
###################### POD Documents ##########################################
###############################################################################

=pod

=head1 Package Name

DataAnalyzer.pm

=head1 Author

Rajiv Gupta

=head1 Description

DataAnalyzer test library to handle San Config Advisor Data Analyzer functions and properties.

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

B<SanLog.pm>

B<Error.pm>

=head1 Attributes

=head2 B<CLIENTS_CNT>

Number = Number of SAN Clients.

=head2 B<CLIENTS_STR>

string = Clients String supplied as parameter. Supplied through 'CLIENTSS' from the testbed.

=head2 B<SWITCHES_CNT>

Number = Number of SAN switches.

=head2 B<SWITCHES_STR>

string = switch String supplied as parameter. Supplied through 'SWITCHES' from the testbed.

=head2 B<FILERS_CNT>

Number = Number of SAN filer.

=head2 B<FILERS_STR>

string = Filer String supplied as parameter. Supplied through 'FILERS' from the testbed. 

=head2 B<OUTDIR>

String = Output Directory, Mandatory parameter no default.

=head2 B<PKGDIR>

string = Directory where Data collector is installed. Default is c:\progam~1\NtapSCA. Please notice the '~1' in the default.
         If at any level the directory name is more than 8 chars then use '~1' notation.

=head2 B<PKGNAME>

string = Data collector package name. Default is "ntapsca.exe"

=head2 B<PASS>

string = Temporarily password for Data collector execution. Default is ""1q2w3e4r5t6y7u8i9o0p""

=head2 B<SCA_CLIENT>

string = Data Collector Windows client. It has to be windows. Mandatory parameters.

=head2 B<SAN_COMPONENTS>

hash = Hash for all san components, keys for this hash is clients, filers and switches and values are in arrays. The component values are
       in hash will be described in DCClient, DCFiler and DCSwitches.

=head2 B<DC_NODE_STRING>

string = Default node string for the components in testbed.

=head2 B<DC_OUT>

string = Carry the Datacollector execution output array reference.

=head2 B<CDF0>

string = Information of unparsed CDF file.

=head2 B<CDF1>

string = Information of parsed CDF file.

=head2 B<ERROR_FILE>

string = Information of Error Message file.


=head1 Public Methods

=head2 B<new(%args)>

Creates a new DataCollector object.  Uses the Testbed information and creates Datacollector Object. As per arguments create
output directory if required. Initialized all SAN objects.

=over 4

=item Inputs

=item * -createoutdir

Create output directory. Input value is yes|no|1|0. Default is yes|1.

=item * -clients

":" seperated List of clients on which data collector will run for example client1:client2:client3

=item * -filers 

 ":" seperated List of filer on which data collector will run in format filer1:filer2:filer3
 
 =item * -switches 

 ":" seperated List of switches on which data collector will run in format switch1:switch2:switch3
 
=item * -protocol

Protocol used FCP:ISCSI - Required argument

=item * -password

Temprarory passpharase. - Required argument

=item * -outdirlocal

Local output directory where data collector output is installed. - Required argument

=item * -outdirremote

Remote output directory where data collector output will be coped

=item * -pkgname

Data Collector package name - required argument

=item * -pkgdir

Data Collector package directory - required argument

=item * -lab

Lab information hash in format 

	my $LAB_INFO = {'LAB'           => $LAB,
                    'TESTBED'       => $TESTBED,
                    'LOGDIR'        => $LOGDIR,
                    'TESTBED_OWNER' => $TESTBED_OWNER,
                    'TESTBED_USER'  => $TESTBED_USER,
                    'USER'          => $ENV{'LOGNAME'},
                   }; 

=item Outputs

Returns a DataAnalyzer object with all SAN client and information required to execute data analyzer. 

=item Examples

 $Da = new DataAnalyzer(-datacollector => $dc_obj);

  if ($Da->error) {
     $Test->nlog($Client->errmsg());     
 }

=back

=head2 B<get_unparsed_command_value()>

Get the unparesed command value from the unparsed xml file.

=over 4

=item Inputs

=item *Component

=item *Command

=item *java loader object


=item Outputs

Returns String. Command output

=item Examples

  $dc = new DataCollector();
  
  #### Modify San objects like access methods etc.#####
  
  $nodestr = $dc->get_unparsed_command_vaue($component, $command, $java_ini);
       
=back

=head2 B<command_validate()>

Validate command by executing command to the component and get command output from unparsed xml file.

=over 4

=item Inputs

=item *Component

=item *Component_type

=item *Command

=item *java loader object


=item Outputs

Return 1 or 0;

=item Examples

  $dc = new DataCollector();
  
  #### Modify San objects like access methods etc.#####
  
  $nodestr = $dc->command_validate($component, $component_type $command, $java_ini);
       
=back

