## @summary DataCollector test library to handle San Config Advisor Data collector functions and properties.
##
## @description More details available in pod.
##
## $Id: //depot/prod/test/Rsierranevada/lib/configadvisor/DataCollector.pm#1 $<br>
## $DateTime: 2013/01/22 11:22:46 $<br>
## $Author: build $<br>
##
## These test functions use various tasks which can be performed
## with Datacollector.
##
## @usage
##   use DataColletor;
##
## @keywords Tools
## @dependencies San, SanLog, Error, DCClient, DCSwitch, DCFiler
##
## @author Rajiv Gupta (rajivg@netapp.com)
## @status public
##

#Package Name
##################
package DataCollector;

#Package Requirements
######################
use strict;
use Tharn;

# use Data::Dumper;  #use to print the hash usage print Dumper($var)

#Load San Packages
############################
use San::SanLog;
use Error;
use configadvisor::DCFiler;
use configadvisor::DCClient;
use configadvisor::DCSwitch;
use configadvisor::SCAUtil;
use Data::Dumper;

use San::SanClient::ClientUtil;

#Load Java Loader Module to load Java Objects
#############################################
#use configadvisor::lib::Java_Loader;

use vars qw (@ISA);
use XML::Simple;

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

my $Dc_scadc_process  = "sancfgdc.exe";
my $Dc_sancol_process = "sancol.exe";
my $Lock_file_name    = "\%TEMP\%\\dc_auto_lock";

#######################################################
#####################Constructor#######################
#######################################################
## @name ::new
## @description Create a new DataCollector Object. This will initializes the data structure for DataCollector object
## get DataCollector specific information (populate DC global variables). See more details in POD.
## @synopsis
## $obj = DataCollector->new(%args);
##
## @example
## $dc_obj = DataCollector->new(-scaclient    => $SCA_CLIENT,
##                              -clients      => $CLIENTS,
##                              -filers       => $FILERS,
##                              -switches     => $SWITCHES,
##                              -protocol     => $PROTOCOL,
##                              -password     => $PASS,
##                              -outdirlocal  => $OUTDIRLOCAL,
##                              -outdirremote => $OUTDIRREMOTE,
##                              -pkgname      => $PKGNAME,
##                              -pkgdir       => $PKGDIR,
##                              -createoutdir => $OUTDIR_CREATE,
##                              -lab          => $LAB_INFO
##                              -dctimeout    => $DC_TIMEOUT
##                              );
##
###########################################################
sub new () {
    my ($class, %args) = @_;

    # Declare and bless empty hash for DataCollector.
    my $self = {};
    bless($self, $class);
    $self->_clear_err();

    #############################################################
    ### Declare global variables Specific to Config Advisor######
    #############################################################

    my ($clients_dc,  $client_vms,   $filers_dc, $switches_dc, $scaclient, $protocol, $pass,
        $outdirlocal, $outdirremote, $dctimeout, $pkgname,     $pkgdir,    $lab_info, $force_execute);

    my %san_client_args_hash = ();

    my $copyoutdir    = 0;    #if Remote directory is given and different than local directory then set to 1
    my $outdir_create = 1;    #Create Output directory before starting execution.

    ###############################################################
    ### Get supplied arguments ###################################
    ###############################################################

    foreach (keys %args) {
        if (/-?createoutdir$/i) {
            if (not length($outdir_create = $args{$_})
                or $outdir_create !~ /no|yes|1|0/i) {
                $self->sanlog('-msg', 'Setting parameter to create the output directory');
                $outdir_create = 1;
            } else {
                if ($outdir_create =~ /yes|1/i) {
                    $outdir_create = 1;
                } else {
                    $outdir_create = 0;
                }
            }
        } elsif (/-?scaclient/i) {
            if (not length($scaclient = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Value for \"scaclient\": is not supplied");
                return $self;
            }
        } elsif (/-?clients/i) {
            if (length $args{$_}) {
                $clients_dc = $args{$_};
            }
        } elsif (/-?client_vms/i) {
            $client_vms = $args{$_};
            if (length($client_vms)) {
                %san_client_args_hash = (-mp_type => $client_vms);
            }
        } elsif (/-?filer/i) {
            if (length $args{$_}) {
                $filers_dc = $args{$_};
            }
        } elsif (/-?switches/i) {
            if (length $args{$_}) {
                $switches_dc = $args{$_};
            }
        } elsif (/-?timeout/i) {
            if (not length($dctimeout = $args{$_})
                or $dctimeout == 0) {
                $dctimeout = undef;
            } else {
                $dctimeout = $args{$_};
            }
        } elsif (/-?protocol/i) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Protocol not specified");
                return $self;
            }
        } elsif (/-?password/i) {
            if (not length($pass = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Password is not provided not specified");
                return $self;
            }
        } elsif (/-?outdirlocal/i) {
            if (not length($outdirlocal = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Output directory is not provided");
                return $self;
            }
        } elsif (/-?outdirremote/i) {
            $outdirremote = $args{$_};
        } elsif (/-?pkgname/i) {
            if (not length($pkgname = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Data Collector package name is not provided.");
                return $self;
            }
        } elsif (/-?pkgdir/i) {
            if (not length($pkgdir = $args{$_})) {
                $self->_set_err("Error:DataCollector.pm:new():Data Collector package directory is not provided.");
                return $self;
            }
        } elsif (/-?lab/i) {
            $lab_info = $args{$_};
        } elsif (/-?force_execute/i) {
            $force_execute = $args{$_};
        }
    }

##Check all mandatory parameters are supplied.

    if (not length($scaclient)) {
        $self->_set_err("Error:DataCollector.pm:new():Mandatory parameter \"scaclient\": is not supplied");
        return $self;
    }

    if (not length($pkgdir)) {
        $self->_set_err(
                  "Error:DataCollector.pm:new():Mandatory parameter Data Collector package directory is not provided.");
        return $self;
    }

    if (not length($pkgname)) {
        $self->_set_err("Error:DataCollector.pm:new():Data Collector package name is not provided.");
        return $self;
    }

    if (not length($protocol)) {
        $self->_set_err("Error:DataCollector.pm:new():Protocol not specified");
        return $self;
    }
    if (not length($pass)) {
        $self->_set_err("Error:DataCollector.pm:new():Password is not provided not specified");
        return $self;
    }

    $outdirremote = &SCAUtil::trim($outdirremote);
    $outdirlocal  = &SCAUtil::trim($outdirlocal);
    $copyoutdir   = 1

      # Copy directory if Remote directory is given and different than the local directory
      if ((length($outdirremote) != 0) and ($outdirlocal ne $outdirremote));

    my $san_obj;    # Empty hash reference for storing san objects.

    # Create Clients, Filer and Switch array.
    my @dc_node_clients  = split(/:/, $clients_dc);
    my @dc_node_filers   = split(/:/, $filers_dc);
    my @dc_node_switches = split(/:/, $switches_dc);

    my $numofclients  = scalar(@dc_node_clients);
    my $numofswitches = scalar(@dc_node_switches);
    my $numoffilers   = scalar(@dc_node_filers);

    $self->{'CLIENTS_CNT'}  = $numofclients;
    $self->{'CLIENTS_STR'}  = $clients_dc;
    $self->{'SWITCHES_CNT'} = $numofswitches;
    $self->{'SWITCHES_STR'} = $switches_dc;
    $self->{'FILERS_CNT'}   = $numoffilers;
    $self->{'FILERS_STR'}   = $filers_dc;
    $self->{'PKGDIR'}       = $pkgdir;
    $self->{'PKGNAME'}      = $pkgname;
    $self->{'PASS'}         = $pass;
    $self->{'LAB_INFO'}     = $lab_info;
    $self->{'DC_TIMEOUT'}   = $dctimeout;

    $force_execute = 0 if (not defined $force_execute or not length $force_execute);
    $self->{'FORCE_EXECUTE'} = $force_execute;

    #############################################################
    ###### Create Client object for SCA host    #################
    #############################################################

    my $sca_client = DCClient->new(-host => $scaclient,
                                   -DC   => 1);

    if ($sca_client->error()) {
        my $errmsg = join("\n", $sca_client->errmsg());
        $self->_set_err("Error:DataCollector.pm:new():", "Instantiating SCA client object for $scaclient.", "$errmsg");
        return $self;
    }

    $self->{'SCA_CLIENT'} = $sca_client;

    #############################################################
    ##### Test that SCA Host is supported host ##################
    #############################################################

    if ($sca_client->{'dc_client_type'} !~ /windows/i) {
        my $errmsg =
          "Not supported host type \"$sca_client->{'dc_client_type'}\" for configadvisor. The SCA_CLIENT client should be windows.";
        $self->_set_err("Error:DataCollector.pm:new():", "Instantiating client object for $scaclient", "$errmsg");
        return $self;
    }

    ##################################################################
    #### Test CFG DC installed in the specified dir              #####
    #### Test if output dir is existing if not then creates it   #####
    ##################################################################

    my $cfgdc_full_path = "\"" . $pkgdir . "\\" . $pkgname . "\"";

    if (!($sca_client->file_test($cfgdc_full_path))) {
        my $errmsg = "No Data collector \"$pkgname\" at location \"$pkgdir\"";
        $self->_set_err("Error:DataCollector.pm:new():", "Package not found $cfgdc_full_path", "$errmsg",);
        return $self;
    }

    ##################################################################
    #### Get Data Collector Version                              #####
    ####                                                         #####
    ##################################################################

    $self->{'VERSION'} = $self->get_dc_version();
    if ($self->error) {
        my $errmsg = join("\n", $self->errmsg());
        $self->_set_err("Error:DataCollector.pm:new():", "Fail to get data collector version", "$errmsg");
        return $self;
    }

    ############################################################
    #### Populate DC hash with Clients #######################
    ############################################################

    foreach my $client (@dc_node_clients) {
        $self->sanlog(-msg => "Creating Client object for client '" . $client . "'");
        print ntest::NTEST_STDOUT "Creating client object : $client\n";
        %san_client_args_hash = (%san_client_args_hash,
                                 -host      => $client,
                                 -protocol  => $protocol,
                                 -conn_type => "THARNHOST",);
        my $client_obj = DCClient->new(%san_client_args_hash);
        if ($client_obj->error) {
            my $errmsg = join("\n", $client_obj->errmsg());
            $self->_set_err("Error:DataCollector.pm:new():", "Instantiating client object for $client.", "$errmsg");
            return $self;
        }

        push @{$san_obj->{clients}}, $client_obj;    #Populate San Client array.
    }

    ############################################################
    #### Populate DC hash with switches #######################
    ############################################################

    foreach my $switch (@dc_node_switches) {
        $self->sanlog(-msg => "Creating Switch object for switch '" . $switch . "'");
        print ntest::NTEST_STDOUT "Creating switch object : $switch\n";
        my $switch_obj = new DCSwitch(-host      => $switch,
                                      -conn_type => "THARNHOST");
        if ($switch_obj->error) {
            my $errmsg = join("\n", $switch_obj->errmsg());
            $self->_set_err("Error:DataCollector.pm:new():", "Instantiating switch object for $switch.", "$errmsg");
            return $self;
        }

        push @{$san_obj->{switches}}, $switch_obj;
    }

    ############################################################
    #### Populate DC hash with Filers    #######################
    ############################################################

    foreach my $filer (@dc_node_filers) {
        $self->sanlog(-msg => "Creating filer object for filer '" . $filer . "'");
        print ntest::NTEST_STDOUT "Creating filer object : $filer\n";
        my $filer_obj = new DCFiler(-filer => $filer);
        if ($filer_obj->error) {
            my $errmsg = join("\n", $filer_obj->errmsg());
            $self->_set_err("Error:DataCollector.pm:new():", "Instantiating filer object for $filer.", "$errmsg");
            return $self;
        }

        push @{$san_obj->{filers}}, $filer_obj;
    }

    ############################################################
    ### If not defined DC timeout as parameter           #######
    ### Calculate DC Timeout based on Following parameter#######
    ### Each Host Execution time = 3 min #######################
    ### Each Switch Execution time = 5min ######################
    ### Each Filer Execution time = 2 min ######################
    ############################################################

    if (   $self->{'DC_TIMEOUT'} == 0
        or $self->{'DC_TIMEOUT'} == ""
        or not length($self->{'DC_TIMEOUT'})) {
        my $timeout = $self->{'SWITCHES_CNT'} * 600;    #5 mins per switch.
        $timeout = $timeout + ($self->{'CLIENTS_CNT'} * 600);    #5 mins per client
        $timeout = $timeout + ($self->{'FILERS_CNT'} * 600);     # 3 mins per filer
        $self->{'DC_TIMEOUT'} = $timeout;
    }

    ############################################################
    #### Populate DC hash with SAN components  #################
    ############################################################

    $self->{'SAN_COMPONENTS'} = $san_obj;

    ##############################################################
    ### SCA not supports mapped drive to collect data if SSH  ####
    ### is used as access method for any SAN Components.      ####
    ### Because of this reason create local directory at C:\  ####
    ### Copy data from local directory to OUTDIR variable and ####
    ### Delete local directory. After execution               ####
    ##############################################################

    my $timestamp = &SCAUtil::get_timestamp();

    if ($outdirlocal =~ /\\+$/) {
        $outdirlocal = $outdirlocal . $timestamp;
    } else {
        $outdirlocal = $outdirlocal . "\\" . $timestamp;
    }

    if ($outdirremote =~ /\\+$/) {
        $outdirremote = $outdirremote . $timestamp;
    } else {
        $outdirremote = $outdirremote . "\\" . $timestamp;
    }

    # Generate Destination directory name,

    my ($des_dir, $client_string, $filer_string, $switch_string);

    # create unique host name string. If there are multiple host present with same hosttype only one hosttype will be taken.
    foreach my $client (@{$self->{'SAN_COMPONENTS'}->{'clients'}}) {
        $client_string = $client_string . "_" . $client->hosttype() if ($client->hosttype() !~ m/$client_string/);
        $client_string =~ s/^_+//;
        $client_string =~ s/_+$//;
    }

    foreach my $filer (@{$self->{'SAN_COMPONENTS'}->{'filers'}}) {
        $filer_string = $filer_string . "_" . $filer->get_model() if ($filer->get_model() !~ m/$filer_string/);
        $filer_string =~ s/^_+//;
        $filer_string =~ s/_+$//;

    }

    foreach my $switch (@{$self->{'SAN_COMPONENTS'}->{'switches'}}) {
        $switch_string = $switch_string . "_" . $switch->hosttype() if ($switch->hosttype() !~ m/$switch_string/);
        $switch_string =~ s/^_+//;
        $switch_string =~ s/_+$//;
    }

    $des_dir = $client_string                  if (defined $client_string);
    $des_dir = $des_dir . "_" . $filer_string  if (defined $filer_string);
    $des_dir = $des_dir . "_" . $switch_string if (defined $switch_string);

    $outdirlocal = $outdirlocal . "\\" . $des_dir;

    $outdirremote = $outdirremote . "\\" . $des_dir if (length($outdirremote) != 0);

    $self->{'OUTDIRLOCAL'} = $outdirlocal;

    $self->{'OUTDIRREMOTE'} = $outdirremote;

    $self->{'COPYOUTDIR'} = $copyoutdir;

    #######################################################################
    #### Check for output directory if not exist then create it       #####
    #### Create directory if create dir parameter is set to "yes|1    #####
    #######################################################################

    if ($outdir_create) {
        $sca_client->create_directory($outdirlocal);
        if ($sca_client->error()) {
            my $errmsg = join("\n", $sca_client->errmsg());
            $self->_set_err("Error:DataCollector.pm:new():", "Cannot Create output cirector", "$errmsg",);
            return $self;
        }
    } else {
        my $dir_priv_level;
        if ($sca_client->{dc_client_type} =~ m/(windows)/i) {
            $dir_priv_level = $outdirlocal;
            $dir_priv_level =~ s/(.*)\\(.*)$/$1/;
            if ($sca_client->is_dir_writable($dir_priv_level)) {
                $self->sanlog(-msg => "Directory $dir_priv_level is writable");
            } else {
                $self->_set_err("Error:DataCollector.pm:new():Directory $dir_priv_level is not writable");
                return $self;
            }
        } else {
            $dir_priv_level = $outdirlocal;
            $dir_priv_level =~ s/(.*)\/(.*)$/$1/;
            if ($sca_client->is_dir_writable($dir_priv_level)) {
                $self->sanlog(-msg => "Directory $dir_priv_level is writable");
            } else {
                $self->_set_err("Error:DataCollector.pm:new():Directory $dir_priv_level is not writable");
                return $self;
            }
        }
    }

    #######################################################
    ###### CREATE and populate DC_NODE STRING #############
    #######################################################

    $self->{'DC_NODE_STRING'} = $self->create_node($san_obj);
    $self->{'DC_EXE_CMD'}     = $self->create_default_cmd();

    return $self;

}    ### END New

###########################################################################
# Subroutine to create default Datacollector command.
###########################################################################
## @name :: create_default_cmd
## @description :: Based on parameter create node for config advisor.
##                 Subroutine will create the default cli command.
##
## Returns: String
##
##
## @synopsis
##  DataCollector->create_default_cmd()
###########################################################################
sub create_default_cmd() {
    my ($self, %args) = @_;
    my $pkgdir = $self->{'PKGDIR'};
    $pkgdir =~ s/(program\sfiles)/progra~1/i;    #Temprarory solution for "program files"

    my $command =
        $pkgdir . "\\"
      . $self->{'PKGNAME'}
      . " -n   \""
      . $self->{'DC_NODE_STRING'}
      . "\"  -o "
      . $self->{'OUTDIRLOCAL'} . " -p "
      . $self->{'PASS'};

    return $command;
}    #end create_detault_cmd

##########################################################################
# Subroutine to create the node string
# ########################################################################
## @name :: create_node
## @description :: Based on parameter create node for config advisor. Node is
## "netapp-controller <IP/Host> username:pass connection method>"
##
## Returns: Node
##
##
## @synopsis
##  DataCollector->create_node()
##########################################################################
sub create_node () {
    my $self = shift;
    my (@node_str_arr, $str);

    # Create filer node
    foreach my $filer_local (@{$self->{'SAN_COMPONENTS'}->{filers}}) {
        $str =
            $filer_local->{dc_filer_type} . " "
          . $filer_local->{dc_filer_ip} . " "
          . $filer_local->{dc_username} . ":"
          . $filer_local->{dc_password} . "   "
          . $filer_local->{dc_node_access} . " ";
        push @node_str_arr, $str;
    }

    # Create Client node.
    foreach my $client_local (@{$self->{'SAN_COMPONENTS'}->{clients}}) {
        $str =
            $client_local->{dc_client_type} . " "
          . $client_local->{dc_client_ip} . " "
          . $client_local->{dc_username} . ":"
          . $client_local->{dc_password} . " "
          . $client_local->{dc_node_access} . " ";
        push @node_str_arr, $str;
    }

    # Create switch node
    foreach my $switch_local (@{$self->{'SAN_COMPONENTS'}->{switches}}) {
        $str =
            $switch_local->{dc_switch_type} . " "
          . $switch_local->{dc_switch_ip} . " "
          . $switch_local->{dc_username} . ":"
          . $switch_local->{dc_password} . " "
          . $switch_local->{dc_node_access} . " ";
        push @node_str_arr, $str;
    }

    my $node_string = join('; ', @node_str_arr);

    $node_string =~ s/^;//;
    return $node_string;
}    # end create_node

##########################################################################
# Subroutine to Execute Data collector through Command line
# ########################################################################
## @name :: execute_sca_cli
## @description :: Based on default command or the command supplied execute Data collector through CLI.
##
## Returns: 1|0
##
##
## @synopsis
##  DataCollector->execute_sca_cli(-command => $cmd)
##########################################################################
sub execute_sca_cli () {
    my ($self, %args) = @_;
    $self->_clear_err();
    my $sca_client = $self->{'SCA_CLIENT'};
    my $command;
    my $timeout = 1500;
    my $console_msg;
    my ($status, $result) = "";

    # Run the command if supplied in the argument
    foreach (keys %args) {
        if (/-?(command)$/i) {
            if (not length($command = $args{$_})) {
                $self->sanlog('-msg' => 'Command argument required command value running default command');
            }
        }
        if (/-?(timeout)$/i) {
            if (not length($timeout = $args{$_})) {
                $self->sanlog('-msg' => 'Timeout required timeout value running default timeout');
            }
        }
        if (/-?(console_msg)$/i) {
            if (not length($console_msg = $args{$_})) {
                $self->sanlog('-msg' => 'Console_msg required timeout value running without console msg');
            }
        }
    }

    if (not length($command)) {
        $self->sanlog('-msg' => 'No command argument passed, using default command');
        $command = $self->{'DC_EXE_CMD'};
    }

    if (ref($sca_client) =~ /DCClient/i) {

        # check if Data collector is running on the host.
        $self->is_dc_running();
        

        if ($self->error) {
            my $errmsg = $self->errmsg();
            print ntest::NTEST_STDOUT ("$errmsg\n");
            $self->sanlog('-msg' => '$errmsg\n');
            if ($self->{'FORCE_EXECUTE'}) {
                print ntest::NTEST_STDOUT ("Doing Force clean up, killing all DC processes on the host \n");
                $self->sanlog('-msg' => 'Doing Force clean up, killing all DC processes on the host DC\n');
                if (!$self->dc_cleanup()) {
                    my $errmsg = $self->errmsg();
                    $self->_set_err(
                               "Error:DataCollector.pm:execute_sca_cli():DataCollector is already running on the client"
                                 . "Cannot kill the processes\n $errmsg");
                    return $SCAUtil::FALSE;
                }
            } else {
                my $errmsg = $self->errmsg();
                $self->_set_err(
                    "Error:DataCollector.pm:execute_sca_cli():DataCollector is already running on the client\n$errmsg");
                return $SCAUtil::FALSE;
            }
        }

        # check if orphanage Sancol process running on the system or not 
        $self->remove_sancol_process();
        if ($self->error) {
        	return $SCAUtil::FALSE;  #return in case of error.
        }

        # Create lock file at windows in windows temp directory
        my $file_name = $Lock_file_name;
        my @file_buff;
        push @file_buff, &SCAUtil::get_timestamp();
        
        # Write information about user and lab 
        push @file_buff, "LAB:" . $self->{'LAB_INFO'}->{'LAB'};
        push @file_buff, "TESTBED:" . $self->{'LAB_INFO'}->{'TESTBED'};
        push @file_buff, "USER:" . $self->{'LAB_INFO'}->{'USER'};

        $sca_client->remove_remote_file($sca_client->{dc_conn}, $file_name);  #tried remove any previously existing file
        $sca_client->_clear_err();

        # Create remote file at Windows Tem location.
        if (!$sca_client->create_remote_file($sca_client->{dc_conn}, $file_name, @file_buff)) {
            # print error message if fail to create the lock file.
            print ntest::NTEST_STDOUT ("Error:DataCollector.pm:execute_sca_cli():Failed to create lock file \n");
            my $errmsg = $sca_client->errmsg();
            print ntest::NTEST_STDOUT ("Error:DataCollector.pm:execute_sca_cli():Error message $errmsg \n");
            $self->_set_err("Error:DataCollector.pm:execute_sca_cli():Failed to create the lock file \n" . $errmsg);
            return $SCAUtil::FALSE;
        }

        print ntest::NTEST_STDOUT ($console_msg);
        ($result) = $sca_client->execute($command, $timeout);                 #Execute data collector command.

		# remove the lock file after execution.
        $sca_client->remove_remote_file($sca_client->{dc_conn}, $file_name);
        if ($sca_client->error) {
            print ntest::NTEST_STDOUT ("Error:DataCollector.pm:execute_sca_cli():Fail to remove lock file \n");
            $self->sanlog('-msg' =>
                "Error:DataCollector.pm:execute_sca_cli():After executing data collector failed to remove the lock file $file_name\n"
                . "Please delete is manually");
        }
    } else {
        my $errmsg = "Unknown/Blank SCA client type \"$sca_client\" :" . ref($sca_client);
        $self->_set_err("Error:DataCollector.pm:execute_sca_cli():$errmsg");
        return $SCAUtil::FALSE;
    }

    # Update DC_OUT with execution results.

    $self->{'DC_OUT'}->{'VALUE'} = $result;

    if ($sca_client->error) {
        my $dc_value = join('\n', @{$self->{'DC_OUT'}->{'VALUE'}});
        my $errmsg = "Failed to execute the data collector\n
		              The Output of Data collector is\n"
          . $dc_value;
        $self->_set_err("Error:DataCollector.pm:execute_sca_cli():", $errmsg);

        # Update Node with information of XML file location and out of DC.
        return $SCAUtil::FALSE;
    }
    return $SCAUtil::TRUE;
}

sub remove_sancol_process() {
    my ($self)         = @_;
    my $sancol_process = $Dc_sancol_process;
    my $sca_client     = $self->{'SCA_CLIENT'};
    my $dc_client      = $sca_client->{'dc_client_name'};
    my $dc_conn        = $sca_client->{'connection'};

    my $process_instance_sancol = &ClientUtil::count_process_instances($dc_client, $sancol_process, 1500, $dc_conn);

    # if failed at routine level.
    if ($process_instance_sancol == -1) {
        my $errmsg = "Cannot run the process command to get the process information\n";
        $self->_set_err("Error:DataCollector.pm:remove_sancol_process():$errmsg");
    }

    # Sancol process is only running means orphange process which should be killed.
    if ($process_instance_sancol) {
        if (!$self->dc_cleanup()) {
            $self->_set_err(
                "Error:DataCollector.pm:sancol process found running on dc client $dc_client, which was not able to be killed.\n"
                  . "Please kill this process manually");
        }
    }
}

#################################################################################
# Method    : check_dc_running
# Objective : Check if DC is already running or not
#################################################################################
sub is_dc_running {
    my ($self) = @_;
    $self->_clear_err();
    my $sca_client = $self->{'SCA_CLIENT'};
    my $dc_client  = $sca_client->{'dc_client_name'};
    my $dc_conn    = $sca_client->{'connection'};

    my $scadc_process = $Dc_scadc_process;

    my $file_name = $Lock_file_name;

    # First check if sancfgdc.exe is running or not.
    my $process_instance = &ClientUtil::count_process_instances($dc_client, $scadc_process, 1500, $dc_conn);

    if ($process_instance == -1) {
        my $errmsg = "Cannot run the process command to get the process information\n";
        $self->_set_err("Error:DataCollector.pm:is_dc_running():$errmsg");
        return $SCAUtil::TRUE;
    }

    if ($process_instance) {
        $self->_set_err("DataCollector already running on the host \n");

        # Check if data collector is runnig from Automation.
        if ($sca_client->file_test($file_name)) {
            my $out       = $sca_client->read_remote_file($file_name);
            my $exe_time  = shift @$out;
            my $curr_time = &SCAUtil::get_timestamp();
            my $diff      = &SCAUtil::get_diff_time($exe_time, $curr_time);
            my $lab_info  = join("\n", @$out);
            my $errmsg =
                "Data collector lock file found. Seems Data collector is already running"
              . "from last $diff \n"
              . "Following is the user information: \n"
              . $lab_info
              . "\nTo run the data collector please manually kill the processes \"sandfgdc.exe\" & \"sancol.exe\""
              . "or use force option DC_FORCE=1 with command line parameter";

            $self->_set_err("Error:DataCollector.pm:is_dc_running():$errmsg");
        } else {
            my $errmsg = "Data collector lock file not found. Seems Data collector ran manually\n";
            $self->_set_err("Error:DataCollector.pm:is_dc_running():$errmsg");
        }
    }

}

#################################################################################
# Method    : cleanup the dc orphan processes
# Objective : Check if DC is already on the host and delete the processes.
#################################################################################
sub dc_cleanup {
    my ($self) = @_;
    $self->_clear_err();
    my $scadc_process  = $Dc_scadc_process;
    my $sancol_process = $Dc_sancol_process;
    my $file_name      = $Lock_file_name;

    my $sca_client = $self->{'SCA_CLIENT'};
    my $dc_client  = $sca_client->{'dc_client_name'};
    my $dc_conn    = $sca_client->{'connection'};

    my $kill_success = $SCAUtil::TRUE;

    if (&ClientUtil::count_process_instances($dc_client, $scadc_process, 1500, $dc_conn)) {
        if (!&ClientUtil::kill_process_instances($dc_client, $scadc_process)) {
            $kill_success = $SCAUtil::FALSE;
            $self->_set_err(
                      "Error:DataCollector.pm:dc_cleanup:Fail to kill the $scadc_process on the dc client $sca_client");
        } else {
            $kill_success = $SCAUtil::TRUE;
        }
    }

    # check if sancol process is running
    if (&ClientUtil::count_process_instances($dc_client, $sancol_process, 1500, $dc_conn)) {
        if (!&ClientUtil::kill_process_instances($dc_client, $sancol_process)) {
            $kill_success = $SCAUtil::FALSE;
            $self->_set_err(
                     "Error:DataCollector.pm:dc_cleanup:Fail to kill the $sancol_process on the dc client $sca_client");
        } else {
            $kill_success = $SCAUtil::TRUE;
        }
    }

    # check for lock file
    if ($sca_client->file_test($file_name)) {
        if (!$sca_client->remove_remote_file()) {
            my $errmsg = $sca_client->errmsg();
            $self->_set_err("Error:DataCollector.pm:dc_cleanup:Failed to remove the lock file $file_name. \n $errmsg");
        }
    }
    return $kill_success;
}

#################################################################################
# Method    : run_dc_cli_testbed_full
# Objective : Run DataCollector on the default testbed and generate summary.
################################################################################

sub run_dc_cli_testbed_full {
    my $self = shift;
    $self->_clear_err();
    my $command = $self->{'DC_EXE_CMD'};
    my $timeout;

    if (defined $self->{'DC_TIMEOUT'}) {
        $timeout = $self->{'DC_TIMEOUT'};
    } else {
        $timeout = 1500;    # High timeout to for DC to run fully.
    }

    my $timeout_min = $timeout / 60;

    $self->sanlog('-msg' => "Executing Data Collector, Please wait, this may take maximum $timeout_min minutes");

    my $console_msg = "Executing Data Collector, Please wait, this may take $timeout_min minutes";

    $self->execute_sca_cli(-command     => $command,
                           -timeout     => $timeout,
                           -console_msg => $console_msg);
    $self->_parse_dc_out();
    $self->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY_BRIEF'} = $self->_dc_generate_summary();
    my $success = $self->_is_dc_executed_successfully();

    # Copy summary to the data collection location.
    my $filename        = $self->{'OUTDIRLOCAL'} . "\\dc_auto_summary.txt";
    my $sca_client_conn = $self->{'SCA_CLIENT'}->{'dc_conn'};
    my $sca_client      = $self->{'SCA_CLIENT'};
    my @summ_array      = @{$self->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY_BRIEF'}};

    foreach my $s_line (@summ_array) {
        $s_line =~ s/\n//g;
        $s_line =~ s/\\/\\\\/g;
        $s_line =~ s/\'/\\\'/g;
        $sca_client->create_remote_file($sca_client_conn, $filename, $s_line);
    }

    if ($sca_client->error()) {
        my $errmsg = join("\n", $self->{'SCA_CLIENT'}->errmsg());
        $self->_set_err("Error:DataCollector.pm:execute_sca_cli():",
                        "Fail to write the summary file to the location $filename",
                        "$errmsg");
    }

    $self->{'DC_OUT'}->{'SUCCESS'} = $success;
    return $success;
}

#################################################################################
# Method    : _is_dc_executed_successfully
# Objective : Check for error and warnings.
################################################################################
sub _is_dc_executed_successfully() {

    my ($self) = @_;
    $self->_clear_err();
    my $success = 0;
    my @error_messages;

    # Check if Type 1 CDF is generated.

    if ($self->{'DC_OUT'}->{'OUT_XML_LOCAL'} ne "") {
        if ($self->{'SCA_CLIENT'}->file_test($self->{'DC_OUT'}->{'OUT_XML_LOCAL'})) {
            $success = 1;
            $self->sanlog('-msg' => "Output XML file generated by Data Collector is found");
        } else {
            $success = 0;
            my $xml_file = $self->{'DC_OUT'}->{'OUT_XML_LOCAL'};
            my $prv_error = join('\n', $self->errmsg);
            my $errmsg =
              ": Data Collector XML file :$xml_file not found\n" . "  File is not generated please check errors\n";
            push @error_messages, $errmsg;
            $success = 0;
        }
    } else {
        my $prv_error = join('\n', $self->errmsg);
        my $errmsg =
            ": Data Collector PARSED XML file not generated\n"
          . "  May be DC ran with error or killed by other user and file was not generated please check the errors\n"
          . "In the log file\n";

        push @error_messages, $errmsg;
        $success = 0;
    }

    # Check if Type 0 CDF is generated.

    if ($self->{'CDF0_LOCAL'} ne "") {
        if ($self->{'SCA_CLIENT'}->file_test($self->{'CDF0_LOCAL'})) {
            $success = 1;
            $self->sanlog('-msg' => "Unparsed XML file found");
        } else {
            $success = 0;
            my $xml_file  = $self->{'CDF0_LOCAL'};
            my $prv_error = join('\n', $self->errmsg);
            my $errmsg    = ": Data Collector Unparsed XML file :$xml_file not generated\n"
              . "May be DC ran with error or killed by other user and file is not generated please check the errors\n";
            push @error_messages, $errmsg;
            $success = 0;
        }
    } else {
        my $prv_error = join('\n', $self->errmsg);
        my $errmsg = ": Data Collector UNPARSED XML file in the output not generated\n"
          . "May be DC ran with error or killed by other user and file is not generated please check the errors\n";

        push @error_messages, $errmsg;
        $success = 0;
    }

    if (scalar(@{$self->{'DC_OUT'}->{'SUMMARY'}->{'ERROR'}}) != 0) {

        my $errors = join('\n', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'ERROR'}});

        my $prv_error = join('\n', $self->errmsg);
        my $errmsg =
            ": Data Collector executed with Error\n"
          . "Please check the error below\n"
          . "\nErrors are:\n"
          . $errors
          . "\nEnd of error message\n";

        push @error_messages, $errmsg;

        # Test if errormessages.txt file is generated.

        if ($self->{'SCA_CLIENT'}->file_test($self->{'DC_OUT'}->{'ERROR_FILE_LOCAL'})) {
            $self->sanlog('-msg' => "Error message file: \"$self->{DC_OUT}->{ERROR_FILE_LOCAL}\" created ");
        } else {
            my $error_file = $self->{'DC_OUT'}->{'ERROR_FILE_LOCAL'};
            my $prv_error = join('\n', $self->errmsg);
            my $errmsg =
              ": Data Collector Error file should be created. This is an error condition where error file $error_file was not found\n"
              . "File is not generated please check previous errors\n";
            push @error_messages, $errmsg;
            $success = 0;
        }
        $success = 0;
    }

    if (scalar(@{$self->{'DC_OUT'}->{'SUMMARY'}->{'OTHERERROR'}}) != 0) {

        my $errors = join('\n', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'OTHERERROR'}});
        my $errmsg =
            ": Data Collector executed with Error\n"
          . "Please check the log file and errors below\n"
          . "\nErrors are:\n"
          . $errors
          . "\nEnd of error message\n";
        push @error_messages, $errmsg;
        $success = 0;
    }

    if (length @{$self->{'DC_OUT'}->{'SUMMARY'}->{'WARNING'}} != 0) {
        my $warnings = join('\n', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'WARNING'}});
        $self->sanlog(-msg => "Data collector executed with warnings $warnings\n");
    }

    if (!$success) {
        $self->_set_err(@error_messages);
    }
    return $success;
}

#################################################################################
# Method    : _parse_dc_out()
# Objective : Parse the output from the DC execution and populate Data Collector.
################################################################################
sub _parse_dc_out() {

    my ($self, @dc_out) = @_;

    # Initialize variable for DC_OUT hash.

    my (@warnings, @errors, @other_errors, @info, @extra, @dc_summary, $xml_file, @dc_comps);

    if (!(@dc_out)) {
        @dc_out = @{$self->{'DC_OUT'}->{'VALUE'}};
    }

    foreach my $line (@dc_out) {
        if ($line =~ m/(.*):\s?(WARNING)\s:\s(.*)/) {
            push @warnings, $line;
        } elsif ($line =~ m/(.*):\s?(ERROR)\s:\s(.*)/) {
            push @errors, $line;
        } elsif ($line =~ m/(.*):\s?(INFO)\s:\s(.*)/) {
            push @info, $line;
        } elsif ($line =~ m/(.*):\s?(STATUS)\s:\s(.*)/) {

            # Do nothing.
        } elsif ($line =~ m/(.*):\s?#############################################/) {
            last;    # reached to summary end.
        } else {
            push @extra, $line;
        }

        if ($line =~ m/(.*):\s+(INFO)\s+:\s+(Successfully\s+generated\s+cdf\s+file:)\s+(.*)/) {
            $xml_file = $4;
        }

        if ($line =~ m/(.*):\s+(INFO)\s+:\s+(Starting\sdata\scollection\son)\s(.*)/) {
            push @dc_comps, $4;
        }
    }

    foreach my $line (@extra) {
        if ($line =~ /exception/i) {
            push @other_errors, $line;
        } elsif ($line =~ /Java/i) {
            push @other_errors, $line;
        }
    }

    my $end_counter   = scalar(@dc_out);
    my $start_counter = $end_counter - 7;

    for (my $i = $start_counter; $i <= $end_counter; $i++) {
        push @dc_summary, $dc_out[$i];
        push @dc_summary, "\n";
    }

    if ($self->error) {
        my @self_error = $self->errmsg();
        push @other_errors, @self_error;
    }

    my $xml_file_only = $xml_file;

    $xml_file_only =~ s/(.*)\\(.*)$/$2/;

    my $err_file_local = $self->{'OUTDIRLOCAL'} . "\\errormessages.txt";
    my $err_file       = $self->{'OUTDIRREMOTE'} . "\\errormessages.txt";

    my $remote_xml_file = $self->{'OUTDIRREMOTE'} . "\\" . $xml_file_only if (length $xml_file);

    my $type0_cdf_local = $xml_file;
    $type0_cdf_local =~ s/(.*)\.xml/$1\_unparsed\.xml/;

    my $type0_cdf = $remote_xml_file;
    $type0_cdf =~ s/(.*)\.xml/$1\_unparsed\.xml/;

    $self->{'DC_OUT'}->{'SUMMARY'}->{'WARNING'}    = \@warnings;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'ERROR'}      = \@errors;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'OTHERERROR'} = \@other_errors;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'INFO'}       = \@info;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'EXTRA'}      = \@extra;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'COMPONENTS'} = \@dc_comps;
    $self->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY'}    = \@dc_summary;
    $self->{'DC_OUT'}->{'OUT_XML_LOCAL'}           = $xml_file;
    $self->{'DC_OUT'}->{'OUT_XML'}                 = $remote_xml_file;
    $self->{'DC_OUT'}->{'ERROR_FILE'}              = $err_file;
    $self->{'DC_OUT'}->{'ERROR_FILE_LOCAL'}        = $err_file_local;
    $self->{'CDF1'}                                = $remote_xml_file;
    $self->{'CDF1_LOCAL'}                          = $xml_file;
    $self->{'CDF0'}                                = $type0_cdf;
    $self->{'CDF0_LOCAL'}                          = $type0_cdf_local;

    return $SCAUtil::TRUE;
}

#################################################################################
# Method    : _dc_generate_summary()
# Objective : Generate summary for dc execution.
# Summary will look like this:
#
# Test Case passed
#
#
# Data Collector Execution LAB Summary:
#
# LAB NAME         :btc-san-tools
# Test Bed Name    :lnx42_f218_f222_b202_q204
# Test Bed Log Dir :/x/eng/testbeds/btc-san-tools/lnx42_f218_f222_b202_q204/logs/20080909_171724
# Test Bed Owner   :rajivg
# Test Bed User    :rajivg
# Current User     :rajivg
#
#
# San Conifiguration Advisor Host Information:
#
# Sca Host Name      :w2k3s200-30
# Sca Host IP        :10.72.200.30
#
#
#
# Data Collector Execution Summary:
#
# Data collector output XML file locally is: E:\sca_logs\20080809_171811\unix-linux_FAS3020_fcp-switch-qlogic_switch-brocade\config_advisor_20080909_171359.xml
# Data collector output XML remote copy location: s:\automation_test\20080809_171811\unix-linux_FAS3020_fcp-switch-qlogic_switch-brocade\config_advisor_20080909_171359.xml
# Number of Warnings :1
# Number of Errors   :0
# Components for which Data Collected:
# netapp-controller ( 10.72.198.103 ) using ontapi\n netapp-controller ( 10.72.198.92 ) using ontapi\n linux ( 10.72.198.42 ) using telnet\n qlogic ( 10.72.198.31 ) using telnet\n brocade ( 10.72.198.22 ) using telnet
# Warnings:
# 2008099_171602 : WARNING : iscli not found in default location and in user's path on 10.72.198.42, iSCSI HBA data can not be collected
# Errors:
# No Errors
#
#
# SAN Host Details:
#
# San Host Name            :sles198-69
# San Host IP              :10.72.198.42
# San Host Protocol        :FCP
# San Host Type            :unix-linux
# San Host OS version      :
# San Host Sanlun Verion   :3.0
# San Host Sanlun Build    :566312
# San Host Multipath Info  :1
#
# SAN FILER Details:
#
# Name:    f3020-199-218
# Model:   FAS3020
# Version: 7.2.2
# Cfmode:  single_image
# Partner: f3020-198-92
#
# SAN Switch Details:
#
# San Switch Name            :SANbox_Qlogic_204
# San Switch IP              :10.72.198.31
# San Switch Type            :fcp-switch-qlogic
# San Switch FW version      :V6.8.0.3.0
# San Switch Model           :SANbox 5602 FC Switch
#
# San Switch Name            :brd4900-199-203
# San Switch IP              :10.72.198.22
# San Switch Type            :switch-brocade
# San Switch FW version      :v5.3.0d
# San Switch Model           :SilkWorm4900
################################################################################
################################################################################
sub _dc_generate_summary () {

    my $self           = shift;
    my $xml_file       = $self->{'DC_OUT'}->{'OUT_XML'};
    my $xml_file_local = $self->{'DC_OUT'}->{'OUT_XML_LOCAL'};

    my $errors_no  = scalar(@{$self->{'DC_OUT'}->{'SUMMARY'}->{'ERROR'}});
    my $errors     = join('\n', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'ERROR'}});
    my $warning_no = scalar(@{$self->{'DC_OUT'}->{'SUMMARY'}->{'WARNING'}});
    my $warning    = join('\n', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'WARNING'}});
    my $coms_num   = scalar(@{$self->{'DC_OUT'}->{'SUMMARY'}->{'COMPONENTS'}});
    my $coms       = join('\n ', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'COMPONENTS'}});
    my $summary    = join(' ', @{$self->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY'}});

    $warning = "No Warnings"   if ($warning_no == 0);
    $errors  = "No Errors"     if ($errors_no == 0);
    $coms    = "No Components" if ($coms_num == 0);

    my @lab_summary;
    if (defined($self->{'LAB_INFO'})) {
        my $lab_name = $self->{'LAB_INFO'}->{'LAB'};
        my $tb_name  = $self->{'LAB_INFO'}->{'TESTBED'};
        my $tb_owner = $self->{'LAB_INFO'}->{'TESTBED_OWNER'};
        my $tb_user  = $self->{'LAB_INFO'}->{'TESTBED_USER'};
        my $log_dir  = $self->{'LAB_INFO'}->{'LOGDIR'};
        my $cur_user = $self->{'LAB_INFO'}->{'USER'};

        @lab_summary = ("\n\n",
                        "\nData Collector Execution LAB Summary:  \n",
                        "\nLAB NAME         :" . $lab_name,
                        "\nTest Bed Name    :" . $tb_name,
                        "\nTest Bed Log Dir :" . $log_dir,
                        "\nTest Bed Owner   :" . $tb_owner,
                        "\nTest Bed User    :" . $tb_user,
                        "\nCurrent User     :" . $cur_user,
                        "\n\n");
    }

    my @summary_brief = ("@lab_summary",
                         "\nSAN Configuration Advisor Host Information:   \n",
                         "\nSCA Host Name      :" . $self->{'SCA_CLIENT'}->{dc_client_name},
                         "\nSCA Host IP        :" . $self->{'SCA_CLIENT'}->{dc_client_ip},
                         "\nDC version         :" . $self->{'VERSION'},
                         "\n\n",
                         "\n\nData Collector Execution Summary: \n",
                         "\nData collector output XML file locally is: $xml_file_local",
                         "\nData collector output XML remote copy location: $xml_file",
                         "\nSancfgdc Execution Summary:\n $summary",
                         "\nNumber of Warnings :$warning_no",
                         "\nNumber of Errors   :$errors_no",
                         "\nComponents for which Data Collected:\n",
                         $coms,
                         "\nWarnings:\n",
                         $warning,
                         "\nErrors:\n",
                         $errors,
                         "\n\n");

    if ($errors != 0) {
        push @summary_brief, "\n\nError log:" . $self->{'DC_OUT'}{'ERROR_FILE'};
    }

    push @summary_brief, "\nSAN Host Details:\n";

    # Generate Host summary
    foreach my $client (@{$self->{'SAN_COMPONENTS'}{'clients'}}) {
        my $toolkit_version = $client->get_hu_toolkit_version();
        my $sanlun_version  = $client->get_hu_sanlun_version();
        my $mp_type         = $client->get_mp_type();
        push @summary_brief, "\nSAN Host Name            :" . $client->hostname();
        push @summary_brief, "\nSAN Host IP              :" . $client->{dc_client_ip};
        push @summary_brief, "\nSAN Host Protocol        :" . $client->get_protocol();
        push @summary_brief, "\nSAN Host Type            :" . $client->hosttype();
        push @summary_brief, "\nSAN Host OS version      :" . $client->os_version();
        push @summary_brief, "\nSAN Host Sanlun Verion   :" . $sanlun_version;
        push @summary_brief, "\nSAN Host Toolkit version :" . $toolkit_version;
        push @summary_brief, "\nSAN Host Multipath Info  :" . $mp_type;
        push @summary_brief, "\n";
    }

    push @summary_brief, "\nSAN FILER Details:\n";

    # Generate Filers Summary
    my @filer_summary;
    foreach my $filer (@{$self->{'SAN_COMPONENTS'}{'filers'}}) {
        foreach my $line ($filer->info()) {
            push @summary_brief, "\n";
            push @summary_brief, $line;
        }
        push @summary_brief, "\n";
    }

    push @summary_brief, "\nSAN Switch Details:\n";

    # Generate Switches Summary
    foreach my $switch (@{$self->{'SAN_COMPONENTS'}{'switches'}}) {
        push @summary_brief, "\nSAN Switch Name            :" . $switch->hostname();
        push @summary_brief, "\nSAN Switch IP              :" . $switch->{dc_switch_ip};
        push @summary_brief, "\nSAN Switch Type            :" . $switch->hosttype();
        push @summary_brief, "\nSAN Switch FW version      :" . $switch->fw_version();
        push @summary_brief, "\nSAN Switch Model           :" . $switch->model();
        push @summary_brief, "\n";
    }

    return \@summary_brief;
}

sub get_dc_version {

    my $self = shift;
    return $self->{'VERSION'} if (defined $self->{'VERSION'});

    my $pkgdir = $self->{'PKGDIR'};
    $pkgdir =~ s/(program\sfiles)/progra~1/i;    #Temprarory solution for "program files"

    my $version;

    my $command = $pkgdir . "\\" . $self->{'PKGNAME'} . " -v  ";

    my $timeout = 60;
    $self->sanlog('-msg' => "Executing Data Collector, to get DC version information");

    if (
        $self->execute_sca_cli(-command => $command,
                               -timeout => $timeout,)
      ) {
        ## NetApp SAN Configuration Advisor - Data Collector v1.1.2
        $version = $self->{'DC_OUT'}->{'VALUE'}[0];
    } else {
        my $dc_value;
        if (defined $self->{'DC_OUT'}->{'VALUE'}) {
            $dc_value = join('\n', @{$self->{'DC_OUT'}->{'VALUE'}});
        }

        my $errmsg =
            "Failed to execute the data collector to get version Information."
          . "The Output of Data collector is\n"
          . $self->errmsg()
          . $dc_value;
        $self->_set_err("Error:DataCollector.pm:get_dc_version():", $errmsg);

        # Update Node with information of XML file location and out of DC.
        $version = 0.0;
    }
    return $version;
}

sub get_client_count() {
    my $self = shift;
    return $self->{'CLIENTS_CNT'};
}

sub get_switch_count() {
    my $self = shift;
    return $self->{'SWITCHES_CNT'};
}

sub get_filer_count() {
    my $self = shift;
    return $self->{'FILERS_CNT'};
}

sub get_out_dir() {
    my $self = shift;
    return $self->{'OUTDIRLOCAL'};
}

sub get_package_dir() {
    my $self = shift;
    return $self->{'PKGDIR'};
}

sub get_package_name() {
    my $self = shift;
    return $self->{'PKGNAME'};
}

sub set_type1_cdf() {
    my ($self, $type1CDF) = @_;
    $self->{'CDF1'} = $type1CDF;
}

sub set_type0_cdf() {
    my ($self, $type0CDF) = @_;
    $self->{'CDF0'} = $type0CDF;
}

sub get_type1_cdf() {
    my $self = @_;
    return $self->{'CDF1'};
}

sub get_type0_cdf() {
    my $self = @_;
    return $self->{'CDF0'};
}

sub get_san_components() {
    my ($self) = @_;
    return $self->{'SAN_COMPONENTS'};
}

sub get_dc_out() {
    my $class = shift;
    return $class->{'DC_OUT'};
}

sub set_default_cmd() {
    my ($self, $command) = @_;
    $self->{'DC_EXE_CMD'} = $command;
}

sub get_default_cmd() {
    my $self = shift;
    return $self->{'DC_EXE_CMD'};
}

sub set_DC_out() {
    my ($self, $dc_out) = @_;
    $self->{'DC_OUT'} = $dc_out->{'DC_OUT'};
}

###############################################################################
###################### POD Documents ##########################################
###############################################################################

=pod

=head1 Package Name

DataCollector.pm

=head1 Author

Rajiv Gupta

=head1 Description

DataCollector test library to handle San Config Advisor Data collector functions and properties.

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

B<SanLog.pm>

B<Error.pm>

=head1 Attributes

=head2 B<CLIENTS_CNT>

Number = Number of SAN Clients.

=head2 B<CLIENTS_STR>

string = Clients String supplied as parameter. Supplied through 'CLIENTSS' from the testbed.

=head2 B<SWITCHES_CNT>

Number = Number of SAN switches.

=head2 B<SWITCHES_STR>

string = switch String supplied as parameter. Supplied through 'SWITCHES' from the testbed.

=head2 B<FILERS_CNT>

Number = Number of SAN filer.

=head2 B<FILERS_STR>

string = Filer String supplied as parameter. Supplied through 'FILERS' from the testbed. 

=head2 B<OUTDIR>

String = Output Directory, Mandatory parameter no default.

=head2 B<PKGDIR>

string = Directory where Data collector is installed. Default is c:\progam~1\NtapSCA. Please notice the '~1' in the default.
         If at any level the directory name is more than 8 chars then use '~1' notation.

=head2 B<PKGNAME>

string = Data collector package name. Default is "ntapsca.exe"

=head2 B<PASS>

string = Temporarily password for Data collector execution. Default is ""1q2w3e4r5t6y7u8i9o0p""

=head2 B<SCA_CLIENT>

string = Data Collector Windows client. It has to be windows. Mandatory parameters.

=head2 B<SAN_COMPONENTS>

hash = Hash for all san components, keys for this hash is clients, filers and switches and values are in arrays. The component values are
       in hash will be described in DCClient, DCFiler and DCSwitches.

=head2 B<DC_NODE_STRING>

string = Default node string for the components in testbed.

=head2 B<DC_OUT>

string = Carry the Datacollector execution output array reference.

=head2 B<CDF0>

string = Information of unparsed CDF file.

=head2 B<CDF1>

string = Information of parsed CDF file.

=head2 B<ERROR_FILE>

string = Information of Error Message file.


=head1 Public Methods

=head2 B<new(%args)>

Creates a new DataCollector object.  Uses the Testbed information and creates Datacollector Object. As per arguments create
output directory if required. Initialized all SAN objects.

=over 4

=item Inputs

=item * -createoutdir

Create output directory. Input value is yes|no|1|0. Default is yes|1.

=item * -clients

":" seperated List of clients on which data collector will run for example client1:client2:client3

=item * -filers 

 ":" seperated List of filer on which data collector will run in format filer1:filer2:filer3
 
 =item * -switches 

 ":" seperated List of switches on which data collector will run in format switch1:switch2:switch3
 
=item * -protocol

Protocol used FCP:ISCSI - Required argument

=item * -password

Temprarory passpharase. - Required argument

=item * -outdirlocal

Local output directory where data collector output is installed. - Required argument

=item * -outdirremote

Remote output directory where data collector output will be coped

=item * -pkgname

Data Collector package name - required argument

=item * -pkgdir

Data Collector package directory - required argument

=item * -lab

Lab information hash in format 

	my $LAB_INFO = {'LAB'           => $LAB,
                    'TESTBED'       => $TESTBED,
                    'LOGDIR'        => $LOGDIR,
                    'TESTBED_OWNER' => $TESTBED_OWNER,
                    'TESTBED_USER'  => $TESTBED_USER,
                    'USER'          => $ENV{'LOGNAME'},
                   }; 

=item Outputs

Returns a DataCollector object with all SAN client and information required to execute data collector. 

=item Examples

 $Dc = new DataCollector(
        -createoutdir => 1);

  if ($Dc->error) {
     $Test->nlog($Client->errmsg());     
 }

=back

=head2 B<create_node()>

Creates node string for DataCollector based on components defined in SAN_COMPONENTS.

=over 4

=item Inputs

None

=item Outputs

Returns String. Node String.

=item Examples

  $dc = new DataCollector();
  
  #### Modify San objects like access methods etc.#####
  
  $nodestr = $dc->create_node();
       
=back

=head2 B<create_default_cmd(%args)>

Creates default CLI command for data collector.

=over 4

=item Inputs

=none

=item Outputs

Returns default cli command with option -n.

=item Examples

 # Create default command.

  $dc = new DataCollector();
  
  #### Modify San objects like access methods etc.#####
  
  $nodestr = $dc->create_default_cmd();
  
=back

=head2 B<execute_sca_cil($cmd)>

Executes Data collector through CLI

=over 4

=item Inputs

=item * $cmd

Data collector command to execute. If not supplied use default CLI command.

=item Outputs

Returns 0|1 based on execution. Execute Data Collector, analyze the output create summary and return pass or fail.   

=item Examples

  $dc = new DataCollector();
  
  #### Modify San objects like access methods etc.#####
  
 
 if($dc->execute_sca_cli()){
	&tstlog(-pass => "Test Case passed:\n $dc->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY_BRIEF'}\n");	
	}else{
	my $errmsg = join "\n", $dc->errmsg();
	&tstlog(-fail => "Test Case failed \n Error message is $errmsg \n $dc->{'DC_OUT'}->{'SUMMARY'}->{'SUMMARY_BRIEF'}");	
	}

=back

=head2 B<get_client_count()>

This method returns the number of SAN clients.

=over 4

=item Outputs

Returns a Clients count

=item Examples

 #Retrieve execution time
 $cnt = $dc->get_client_count();

=back

=head2 B<get_filer_count()>

This method returns the number of SAN filers.

=over 4

=item Outputs

Returns a filers count

=item Examples

 #Retrieve execution time
 $cnt = $dc->get_filer_count();

=back

=head2 B<get_switch_count()>

This method returns the number of SAN switch.

=over 4

=item Outputs

Returns a switch count

=item Examples

 #Retrieve execution time
 $cnt = $dc->get_switch_count();

=back

=head2 B<get_out_dir()>

This method returns the output directory for SCA.

=over 4

=item Outputs

Returns a output directory

=item Examples

 #Retrieve execution time
 $outdir = $dc->get_out_dir();

=back

=head2 B<get_package_dir()>

This method returns the package installation directory for SCA.

=over 4

=item Outputs

Returns a sca package directory

=item Examples

 #Retrieve execution time
 $outdir = $dc->get_package_dir();

=back

=head2 B<get_package_name()>

This method returns the package name for SCA.

=over 4

=item Outputs

Returns a sca package name

=item Examples

 #Retrieve execution time
 $name = $dc->get_package_name();

=back

=head2 B<get_type0_cdf()>

This method returns type0 cdf.

=over 4

=item Outputs

Returns type0 cdf

=item Examples

 #Retrieve execution time
 $type0 = $dc->get_type0_cdf();

=back

=head2 B<get_type1_cdf()>

This method returns type1 cdf.

=over 4

=item Outputs

Returns type1 cdf

=item Examples

 #Retrieve execution time
 $type1 = $dc->get_type1_cdf();

=back

=head2 B<get_san_component()>

hash = This method returns san components

=over 4

=item Outputs

Returns san component hash

=item Examples

 #Retrieve execution time
 $san_obj = $dc->get_type0_cdf();

=back

=head2 B<get_protocol()>

String: This method returns protocol

=over 4

=item Outputs

Returns protocol string (FCP|ISCSI)

=item Examples

 #Retrieve execution time
 $protocol = $dc->get_protocol();

=back

=head2 B<get_dc_out()>

hash: This method returns the output of data collector execution

=over 4

=item Outputs

Returns data collector out hash 

=item Examples

 #Retrieve execution time
 $output = $dc->get_dc_out();

=back

=head2 B<get_default_cmd()>

String: This method returns the default command to run the data collector

=over 4

=item Outputs

Returns data collector default command. 

=item Examples

 #Retrieve execution time
 $cmd = $dc->get_default_cmd();

=back

=head2 B<set_default_cdm()>

This method sets the default command for data collector.

=over 4

=item Input

String = Command

=item Outputs

none 

=item Examples

 
 $dc->set_default_cmd($cmd);

=back

=head2 B<set_DC_out()>

This method sets the data collector output. If you have output and would just like to process this DC_out

=over 4

=item Input

String = String buffer

=item Outputs

none 

=item Examples

 
 $dc->set_dc_out(\@output);

=back


=cut

1;

