#-*-perl-*-
#  $Id:$

## @summary Routines for issuing reallocate commands to the Filer
##
## @description
##      The defrag.pm functions are wrappers around NATE's say() function that
##      send a 'reallocate xxx' command to the filer and return 1 for success or
##      undef for failure.  In the case of failure, an error string
##      is also returned in $EVAL_ERROR ($@)
##
## @synopsis
##      use defrag;
##
## @usage
##
## @examples
##
## 
## @author rfair
##
## @keywords defrag

package defrag;

use strict;
use Module::Common qw();
use Volume;
use Burt;
use filer;
our ($Validation, $VERSION);

BEGIN
{
    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @defrag_subs = qw(
	defrag_file_make
	defrag_on
	defrag_off
	confirm_defrag_on
	confirm_defrag_off
	confirm_defrag_schedule
	defrag_start
	defrag_schedule
	defrag_stop
	defrag_quiesce
	defrag_restart
	get_defrag_scan_state
	confirm_defrag_scan_state
	defrag_wait_scan
	defrag_file_make_scramble
	defrag_file_destroy_scramble
	defrag_wait_scan_state
	defrag_start_bad
	defrag_stop_bad
	defrag_quiesce_bad
	defrag_restart_bad
	defrag_status_bad
	defrag_schedule_bad
	confirm_defrag_scan_interval
	confirm_defrag_state
    );

    Module::Common::setup_export_tags(
        defrag => [ @defrag_subs ]
   );
}

sub defrag_file_make($$$) 
{
	my $conn = shift;
	my $path = shift;
	my $size = shift;

	my $cmd = "mkfile $size $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd, 180);
	print ("DBG::: $text\n");
}

sub defrag_file_make_scramble($$$) 
{
	my $conn = shift;
	my $path = shift;
	my $size = shift;

	my $cmd = "lun create -s $size -t linux $path";

	# Create a big sparse file for 'randomize' to chomp on,
	# say as a shared lun
	#
	my ($text, $result) = filer_cli_diag_say($conn, $cmd, 180);
	print ("DBG::: $text\n");
	if ($result != 0) {
	     ExitError("Unexpected error from '$cmd': $text");
	}

	$cmd = "lun share $path all";
	($text, $result) = filer_cli_diag_say($conn, $cmd, 180);
	print ("DBG::: $text\n");
	if ($result != 0) {
	     ExitError("Unexpected error from '$cmd': $result, $text");
	}

	# Randomize the file, may take a bit
	#
	$cmd = "randomize  $path";
	($text, $result) = filer_cli_test_say($conn, $cmd, 180);
	print ("DBG::: $text\n");
	if ($result != 0) {
	     ExitError("Unexpected error from '$cmd': $result, $text");
	}
	$cmd = "randomize status";
	($text, $result) = filer_cli_test_say($conn, $cmd, 180);
	my $busy='false';
	do {
		$busy='false';
		my @info = filer_cli_split($text);
		foreach my $line (@info) {
		    if ($line =~ /Working.*/) {
			$busy='true';
			last;
		    } 
		}
		($text, $result) = filer_cli_test_say($conn, $cmd, 180);
		sleep(2);
	} while ($result == 0 and $busy eq 'true');
}
#
# This should be used to destroy a file created with make_scramble,
# since it may have interesting attributes (like be a lun)
#
sub defrag_file_destroy_scramble($$) 
{
	my $conn = shift;
	my $path = shift;
	my $cmd = "lun destroy $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd, 180);

	print ("DBG::: $text\n");
	if ($result != 0) {
	     ExitError("Unexpected error from '$cmd': $result, $text");
	}
}

sub defrag_on($)
{
	my $conn = shift;
	my $cmd = "reallocate on";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

}

sub defrag_off($)
{
	my $conn = shift;
	my $cmd = "reallocate off";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

}
sub confirm_defrag_on($)
{
	my $conn = shift;
	my $cmd = "reallocate status";
	my $found = "false";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	my @info = filer_cli_split($text);
	my $field = '';
	foreach my $line (@info) {
	  if ($line =~ /defrag:\s/) {
	     ExitSoftError("Unexpected error from '$cmd': $line");
	  } elsif ($line =~ /Reallocation scans are on/) {
	     $found="true";
	     last;
	  }
	}
	if ($found ne "true" ) {
		ExitError("reallocate is off when expected to be on.");
	}

}
sub confirm_defrag_off($)
{
	my $conn = shift;
	my $cmd = "reallocate status";
	my $found = "false";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	my @info = filer_cli_split($text);
	my $field = '';
	foreach my $line (@info) {
	  if ($line =~ /defrag:\s/) {
	     ExitSoftError("Unexpected error from '$cmd': $line");
	  } elsif ($line =~ /Reallocation scans are off/) {
	     $found="true";
	     last;
	  }
	}
	if ($found ne "true" ) {
		ExitError("reallocate is on when expected to be off.");
	}
}

sub confirm_defrag_schedule($$$)
{
	my $conn = shift;
	my $path = shift;
	my $expected = shift;
	my $cmd = "reallocate status $path";
	my $found = "false";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	my @info = filer_cli_split($text);
	my $field = '';
	foreach my $line (@info) {
	  if ($line =~ /defrag:\s/) {
	     ExitSoftError("Unexpected error from '$cmd': $line");
	  } elsif ($line =~ /\sSchedule:.*/) {
	     $field = $line;
	     $field =~ s/^\s+Schedule:\s+(.*)$/$1/;
	     if ( $field ne $expected ) {
		ExitError("reallocate schedule for $path was '$field' but expected '$expected'");
	     }
	     $found="true";
	     last;
	  }
	}
	if ($found ne "true" ) {
		ExitError("reallocate schedule for $path was not found");
	}
}

sub defrag_restart($$)
{
	my $conn = shift;
	my $path = shift;
	my $cmd = "reallocate restart $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	print ("DBG::: $text\n");
	if ($result != 0) {
		ExitError("reallocate restart failed: $text");
	}
	sleep(5);
}
sub defrag_quiesce($$)
{
	my $conn = shift;
	my $path = shift;
	my $cmd = "reallocate quiesce $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	print ("DBG::: $text\n");
	if ($result != 0) {
		ExitError("reallocate quiesce failed: $text");
	}
	sleep(5);
}

sub defrag_stop($$)
{
	my $conn = shift;
	my $path = shift;
	my $cmd = "reallocate stop $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	print ("DBG::: $text\n");
	if ($result != 0) {
		ExitError("reallocate stop failed: $text");
	}
	sleep(5);
}

sub defrag_schedule($$$)
{
	my $conn = shift;
	my $path = shift;
	my $options = shift;
	my $cmd = "reallocate schedule $options $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	print ("DBG::: $text\n");
	if ($result != 0) {
		ExitError("reallocate schedule failed: $text");
	}
}

sub defrag_start($$$)
{
	my $conn = shift;
	my $path = shift;
	my $options = shift;
	my $cmd = "reallocate start $options $path";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	print ("DBG::: $text\n");
	if ($result != 0) {
		ExitError("reallocate start failed: $text");
	}
}

sub get_defrag_scan_state($$) {
	my $conn = shift;
	my $path = shift;
	my $cmd = "reallocate status $path";
	my $field = "NotFound";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	my @info = filer_cli_split($text);
	foreach my $line (@info) {
	  if ($line =~ /defrag:\s/) {
	     ExitSoftError("Unexpected error from '$cmd': $line");
	  } elsif ($line =~ /^\s+State:.*/) {
	     $field = $line;
	     $field =~ s/^\s+State:\s+(\w+).*/$1/;
	     last;
	  } elsif ($line =~ /^\s+State:.*:/) {
	     $field = $line;
	     $field =~ s/^\s+State:\s+(\w+):.*/$1/;
	     last;
	  }
	}
	return $field;
}

sub get_defrag_scan_interval($$) {
	my $conn = shift;
	my $path = shift;
	my $cmd = "reallocate status $path";
	my $field = "NotFound";

	my ($text, $result) = filer_cli_diag_say($conn, $cmd);

	my @info = filer_cli_split($text);
	foreach my $line (@info) {
	  if ($line =~ /defrag:\s/) {
	     ExitSoftError("Unexpected error from '$cmd': $line");
	  } elsif ($line =~ /^\s+Interval:.*/) {
	     $field = $line;
	     $field =~ s/^\s+Interval:\s+(.*)/$1/;
	     last;
	  } 
	}
	return $field;
}

sub confirm_defrag_scan_interval($$$)
{
	my $conn = shift;
	my $path = shift;
	my $interval = shift;
	my $intervalIs; 
	$intervalIs=get_defrag_scan_interval($conn, $path);

	if ($intervalIs ne $interval) {
		ExitError("reallocate interval for '$path' is '$intervalIs', but wanted '$interval'");
		
	}
}
sub confirm_defrag_scan_state($$$) {
	my $conn = shift;
	my $path = shift;
	my $state = shift;
	my $stateIs = get_defrag_scan_state($conn, $path);

	if ($stateIs ne  $state) {
		ExitError("reallocate state for '$path' is '$stateIs', but wanted '$state'");
	}
}

sub defrag_wait_scan($$) {
	my $conn = shift;
	my $path = shift;
	my $count = 0;
	
	while (get_defrag_scan_state($conn, $path) eq "Idle") {
		$count++;
		if ($count > 600) {
			ExitError("defrag_wait_scan: didn't enter scan\n");
		}
		sleep 1;
	}
}

sub defrag_wait_scan_state($$$) {
	my $conn = shift;
	my $path = shift;
	my $state = shift;
	my $count = 0;

	while (get_defrag_scan_state($conn, $path) ne $state ) {
		$count++;
		if ($count > 600) {
			ExitError("defrag_wait_scan: didn't enter scan\n");
		}
		sleep 1;
	}
}

sub defrag_start_bad($$$$) {
	my $conn = shift;
	my $path = shift;
	my $options = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate start $options $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate start' but didn't see it.");
	}
}

sub defrag_stop_bad($$$) {
	my $conn = shift;
	my $path = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate stop $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate stop' but didn't see it.");
	}
}

sub defrag_quiesce_bad($$$) {
	my $conn = shift;
	my $path = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate quiesce $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate quiesce' but didn't see it.");
	}
}

sub defrag_restart_bad($$$) {
	my $conn = shift;
	my $path = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate restart $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate restart' but didn't see it.");
	}
}

sub defrag_status_bad($$$$) {
	my $conn = shift;
	my $path = shift;
	my $options = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate status $options $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate status' but didn't see it.");
	}
}

sub defrag_schedule_bad($$$$) {
	my $conn = shift;
	my $path = shift;
	my $options = shift;
	my $error = shift;
	my $found = 'false';

	my $cmd = "reallocate schedule $options $path";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	my @info = filer_cli_split($text);

        foreach my $line (@info) {
          if ($line =~ /.*$error.*/) {
		$found='true';
		last;
          }
        }

	if ($found eq 'false' ) {
		ExitError("Expected error '$error' in 'reallocate schedule' but didn't see it.");
	}
}


sub confirm_defrag_state($$)
{
	my $conn = shift;
	my $file = shift;
    my ($optimization, $hotspots);

	my $ret_value = 0;
	my $field = "NotFound";
	
	defrag_on($conn);
	
	my $cmd = "reallocate measure $file";
	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	sleep(2);
	
	my $cmd = "reallocate status $file";
 	my ($text, $result) = filer_cli_diag_say($conn, $cmd);
	
	my @info = filer_cli_split($text);
	foreach my $line (@info) {

		if ($line =~ /defrag:\s/) {
	     	ExitSoftError("Unexpected error from '$cmd': $line");
	     	
	  	} elsif ($line =~ /^\s+Optimization:.*/) {
	     	$field = $line;
            $field =~ /Optimization:(\s+)(\d+)(.*hot-spots:(\s+)(\d+))?/;
            $optimization = $2;
            $hotspots = $5;

	     	last;
	  	} 
	}

	defrag_stop($conn, $file);	
    if ($@) {    
		ExitError("confess $@");
	}
	
    return ($optimization, $hotspots);
}



1;
