#-*-perl-*-

## @summary Routines for disk
##
## @description
##      This package contains various helpful disk-related routines.
##
## @status in-progress
## @author dl-johny (FSS WAFL QA)
## @keywords nebuchadnezzar

package disk;

use strict;
use vars qw(@ISA @EXPORT);
use Tharn;
use RaidInfo;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
            verify_spare_disks
            raid_group_count
            spare_disks_info
            get_disks
            get_zeroed_spares
            get_nonzeroed_spares
            poll_for_zeroed_spares
);

## @name ::verify_spare_disks()

## @summary verifies if at least specified number of disks are available in
## spare pools
#
## @synopsis
##   verify_spare_disks($spec, [  pool{0|1}_spares => X 
##                             [, pool{0|1}_spares => Y]
##                             [, poolAny_spares => Y]
##                             [, total_spares => Z] ] ]
##
##   where, X, Y and Z are integral number of disks
##
## @example verify_spare_disks($fc, pool0_spares => 2, pool1_spares => 2,
##                               total_spares => 4)

sub verify_spare_disks ( $% ) {

    my ($spec,%args) = @_;

    my ($ri) = new RaidInfo($spec);
    my %disks =  $ri->all_disks();
    my ($total_spares, $pool0_spares, $pool1_spares) = (0, 0, 0);
    my $pass = 1;

    foreach my $name (keys %disks) {
        my $disk = $ri->disk($name);
        if ($disk->parameter("spare") eq "1") {
            ++$total_spares;
            if ($disk->parameter("tree") eq "POOL:Pool0") {
                ++$pool0_spares;
            } elsif ($disk->parameter("tree") eq "POOL:Pool1") {
                ++$pool1_spares;
            }
        }
    }

    if (defined($args{pool0_spares}) && ($args{pool0_spares} > $pool0_spares)) {
        tstlog(-config, "There should be at least $args{pool0_spares} spare "
                      . "disks in Pool0");
	$pass = 0;
    }

    if (defined($args{pool1_spares}) && ($args{pool1_spares} > $pool1_spares)) {
        tstlog(-config, "There should be at least $args{pool1_spares} spare "
                      . "disks in Pool1");
	$pass = 0;
    }

    if (defined($args{total_spares}) && ($args{total_spares} > $total_spares)) {
        tstlog(-config, "There should be at least $args{total_spares} spare "
                      . "disks");
	$pass = 0;
    }

    if (defined($args{poolAny_spares})
            && ($args{poolAny_spares} > $pool0_spares)
            && ($args{poolAny_spares} > $pool1_spares)) {
        tstlog(-config, "There should be at least $args{poolAny_spares} spare "
                      . "disks in any pool");
	$pass = 0;
    }

    return $pass;
}

sub raid_group_count($;$;)
{
    my ($spec, $aggrname) = @_;

    my $ri = new RaidInfo($spec);
    my $aggr = $ri->aggregate($aggrname);
    my %groups = $aggr->all_raid_groups();
    return scalar keys %groups;
}

sub spare_disks_info($;)
{
    my ($spec) = @_;

    my $ri = new RaidInfo($spec);
    my %disks = $ri->all_disks();
    my %result = ();

    foreach my $name (keys %disks) {
        my $d = $ri->disk($name);

        my $size = int (($d->parameter("raid_size")) * 4 / 1000);
        if (defined($result{$size})) {
	    $result{$size} .= "$name,";
        } else {
	    $result{$size} = "";
        }

        my $tree = $d->parameter("tree");
        if (defined($result{"$size:$tree"})) {
            $result{"$size:$tree"} .= "$name,";
        } else {
            $result{"$size:$tree"} = "";
        }
    }

    return %result;
}

sub get_disks($;$;*;)
{
	my ($spec, $num, $sizeref) = @_;

	my @disks = ();
	my %diskinfo = spare_disks_info($spec);
	my @disksizes = grep !/POOL:Pool/, (keys %diskinfo);

	@disks = ();
	loop: foreach my $size (@disksizes) {
		for (my $pool = 0; $pool < 2; ++$pool) {
			my $index = "$size:POOL:Pool$pool";
			if (defined($diskinfo{$index})) {
				@disks = split /,/, $diskinfo{$index};
				if (@disks >= $num) {
					${$sizeref} = $size;
					last loop;
				}
			}
		}
	}
	return @disks;
}

## @name ::get_zeroed_spares()
## @summary Find spare disks which have been zeroed already.
## @synopsis
##   use disk;
##   my @zeroed_disks = get_zeroed_spares($spec)
##
## @arg spec    
##   Required: hostspec of the filer
##
## @returns an array of disk names containing a list of spare disks that are 
##          zeroed already.
## @status public

sub get_zeroed_spares($) {
    my ($spec) = @_;

    my $ri = new RaidInfo($spec);
    my %disks = $ri->all_disks();
    my @disk_list = ();

    foreach my $name (keys %disks) {
        my $disk = $ri->disk($name);

        if ( ($disk->parameter("spare") eq "1") &&
             ($disk->parameter("fsm_state") eq "ZEROED") ) {
            push(@disk_list, $name);
        }
    }
    return @disk_list;
}

## @name ::get_nonzeroed_spares()
## @summary Find spare disks that are not zeroed yet
## @synopsis
##   use disk;
##   my @nonzeroed_disks = get_nonzeroed_spares($spec)
##
## @arg spec    
##   Required: hostspec of the filer
##
## @returns an array of disk names containing a list of spare disks that are 
##          *not* zeroed.
## @status public

sub get_nonzeroed_spares($) {
    my ($spec) = @_;

    my $ri = new RaidInfo($spec);
    my %disks = $ri->all_disks();
    my @disk_list = ();

    foreach my $name (keys %disks) {
        my $disk = $ri->disk($name);

        if ( ($disk->parameter("spare") eq "1") &&
             ($disk->parameter("fsm_state") eq "NOTZEROED") ) {
            push(@disk_list, $name);
        }
    }
    return @disk_list;
}

## @name ::poll_for_zeroed_spares
## @summary Poll raid info to ensure all spare disks are zeroed
## @description 
##   This function is used right after when you call 'disk zero spares'. It
##   will poll raid info to ensure that the required no. of spare disks are
##   zeroed before proceeding on into the testscript.
##
## @synopsis
##   use disk;
##
##   poll_for_zeroed_spares($spec, [spares=>-1], [timeout=>3600],
##                                 [hostref=>...]);
##
## @usage
##   use disk;
##
##   say($fc, "disk zero spares");
##
##   # ensure all spare disks are zeroed first before moving on
##   my $num_zeroed_spares = poll_for_zeroed_spares($fc);
##
##   # OR ensure at least 2 spare disks are zeroed before moving on
##   my $num_zeroed_spares = poll_for_zeroed_spares($fc, spares=>2, 
##                                                  timeout=>100,
##                                                  host=>$fh);
##
## @arg spec    
##   Required: hostspec of the filer
## @arg spares
##   Optional: The number of spares (to be zeroed) to poll for.
##             Default: -1 (all disks; a value of 0 also mean all disks)
## @arg timeout
##   Optional: The timeout value for this function to poll. 
##             Default: 3600 (1 hr)
## @arg hostrec
##   Optional: This value refers to the Hostrec object, *NOT* the hostname
##             of the filer. By passing the object, the function is 'smarter'
##             in determining the sleep value between each polling. Simulators
##             are polled much faster while hardware filers are polled
##             every 5 mins. 
##             Default: [no value] (poll as though hardware filers are used)
##
## @returns Returns the no. of spare disks that have been zeroed. Otherwise
##          returns -1 if it timed out.
## @status TESTME

sub poll_for_zeroed_spares($%) {
    my ($spec, %opts) = @_;

    my $spares_required = defined($opts{spares}) ? $opts{spares} : -1;
    my $timeout = defined($opts{timeout}) ? $opts{timeout} : 3600;
    my $hostrec = defined($opts{hostrec}) ? $opts{hostrec} : "filer";

    my $ri = new RaidInfo($spec);
    my %disks = $ri->all_disks();

    my $cur_time = time();
    my $fut_time = $cur_time + $timeout;

    # someone is trying to be funny here
    # if spares required is zero, then caller actually wants all of the spares
    # zeroed
    if ($spares_required == 0) {
        $spares_required = -1;
    }

    my $spares_zeroed = 0;
    my $spares_nonzeroed = 0;
    my $done = 0;

    while ($cur_time < $fut_time) {
        # refresh raid info
        $ri->refresh();

        foreach my $name (keys %disks) {
            my $disk = $ri->disk($name);

            # look only for spare disks
            if ($disk->parameter("spare") ne "1") {
                next;
            }

            if ($disk->parameter("fsm_state") eq "ZEROED") {
                ++$spares_zeroed;
                
                # check if the required no. of disks have been zeroed
                if ($spares_required > 0 && 
                    $spares_zeroed > $spares_required) {
                    $done = 1;
                    last;
                }
                        
            } elsif ($disk->parameter("fsm_state") eq "NOTZEROED") {
                ++$spares_nonzeroed;

                # we require all spare disks to be zeroed. if one of them is
                # not even zeroed, forget abt checking the rest. break from
                # this loop and poll again later
                if ($spares_required < 0) {
                    $done = 0;
                    last;
                }
            }
        }

        if ($spares_required < 0) {
            # ok, we want all spares to be zero, therefore there
            # should be no spares that are nonzeroed. si?
            if ($spares_nonzeroed == 0 && $spares_zeroed >= 0) {
                $done = 1;
            }
        }

        # all the disks have been zeroed
        # return the no. of spare disks that were zeroed
        if ($done) {
            return $spares_zeroed;

        } else {

            # not done yet. time to snooze...
            # determine if the test is running on a filer or simulator
            # we adjust the sleep time accordingly 
            if ($hostrec ne "filer" && 
                $hostrec->hosttype() =~ /simulator/) {
                sleep 5;
            } else {
                sleep 300;
            }

            $spares_zeroed = 0;
            $spares_nonzeroed = 0;
            $done = 0;
        }
    }

    # timed out. not done yet...
    return -1;
}

1;
