#-*-perl-*-

## @summary general purpose routines used for populating volume
##          comparing data sets
##
## @author neelima@netapp.com
##
## @status public
## @pod here

package fileSystem;
require Exporter;
use vars ('@ISA','@EXPORT');

BEGIN {
  @ISA = qw(Exporter);
  @EXPORT = qw(
	     populate_vol
	     populate_qtree
	     treecompare
	     treecompare_vol
	     modify_data
	     unpopulate_qtree
	     delete_nfs_data
	     delete_cifs_data
	     create_file
	     write_file
	     remote_cmd
	     populate_nfs_data 
	     populate_fcp_data
	     populate_cifs_data
	     createScratchDir
	     backupFiles
	     restoreFiles
	     restoreFilesPreserveBackups
	     create_win_data
	     convert_size_to_blocks
        );
} # end BEGIN

use Module::Common qw(validate_mixed);
use SPARSE::Util::Logger;
use Params::Validate
	qw(validate_pos validate SCALAR ARRAYREF HASHREF UNDEF BOOLEAN);

use Utils::Units qw(:convert);
use Volume qw(vol_exists);
use waflobjCreator qw(:windows
                create_directory
                check_directory
                create_fifo
                create_symlink
                create_hardlink
                create_dev
                check_file
                check_filesize
                rename_obj
                create_socket
                create_deep_directory
                create_wide_directory
);

use mount qw(mount umount);
use MapShare qw(map_cifs_share unmap_cifs_share);
use SnapCmds qw(snapshot_delete snapshot_create snapshot_exists);
use ExportFS qw(export_path share_delete);
use CIFS::Setup qw(is_cifs_setup);
use SAN qw(mkfs_mount_fcp create_lun);
use Admin::License qw(license);
use QtreeCmds qw(qtreePresent qtreeSecurity qtreeCreate);
use wafl_utils qw(df);
use Utils::Units qw(:convert);
use Utils qw( make_conn_and_complain make_host_and_complain make_host make_conn);
use Carp qw(croak);
use DataManip qw(delete_tree);
use Tharn;
use strict;


=head1 NAME

fileSystem - Populate/modify/treecompare different Data sets on a volume/qtree

=head1 DESCRIPTION

This package is mainly used to populate different kinds of data sets, modifying,
deleting and comparing them


=head1 SYNOPSIS
    
    use fileSystem qw(populate_vol populate_cifs_data populate_qtree treecompare treecompare_vol);
    my $h = host($FILER);
    my $fc = connect("#rsh");
    populate_vol($fc, vol1,
                    [,size => 30m] 
		    [,wclient => tada-c30]
    ) or logresult('fail', "$@");

    treecompare_vol($cfiler, 
		    "src", "dst",
                    [operation=>dr] ,
                    [skip=>["crtime"]])or logcomment($@);

    populate_qtree($cfiler,"src","qtree1",	
	 	    [wclient => 'tada-c30'],
		    [domain => 'ad'],
		    [win_user=>'administrator'],
		    [win_passwd=>'sundance']
     ) or logcomment($@);

=head1 ROUTINES

=cut


=head2 populate_vol

=over

=item Arguments

=over

=item filer

(Required) conn spec of the filer
else makes a conn spec and complains

=item vol

(Required): Volume which you want to  populate

=item size 

(Optional): Will populate approximately $size, Default populates 15m
can be specified in (k|m|g|t), if no units given assumed as bytes
Note: This routine cannot populate less than 15m,(minimum is 15m)
If you want to populate a volume exactly and don't care about 
different kind of files, you should use routines from DataManip
or waflobjCreator.

=item wclient

(Optional):Windows client through which volume or path is populated via cifs,
If not provided, volume will not be populated via cifs 
Note:Windows client should be minwin_compatible

=item uclient 

(Optional):Client through which volume is populated via nfs
Defaults to localhost.

=item sanclient

(Optional):Client through which fcp data is populated,
client and filer should be setup for fcp,
client should be running sanlun software. 

=item filer_ip

(Optional):Ip of the filer, Defaults to host record's
default ip
 
=item prefix

(Optional): If this is defined, all the elements in the volume
created will have this prefix.
Defaults to ''

=item nfs_path

(Optional): If this is given, All the nfs files will be created in this
path(this is a mounted path on the client)

=item cifs_path

(Optional): If this is given, All the cifs files, ie acls and streams
will be created in this path, This should be a path mapped on the windows 
client 

=item fcp_path

(Optional): If this is given, we expect that this is a lun, 
converted to a unix filesystem,

=item win_user

(Optional): username used for authentication to the filer via cifs

=item win_passwd

(Optional): password for win_user

=item workgroup 

(Optional): If using workgroup authentication

=item domain

(Optional): If using domain authentication

=back

=item Description

This routine populates a volume via different protocols,
'nfs' , 'cifs' , 'fcp'  based on the arguments passed to
it

=item Examples
    
    populates volume via 'nfs' and 'cifs'
    populate_vol($fc,"pop_vol",wclient => "tada-c30) or logresult('fail', "$@");

    populates volume via 'nfs','cifs','fcp'
    populate_vol("tada-u4","pop_vol",uclient=>"tada-c11"
             wclient=>"tada-c30",sanclient=>tada-c63)

    populates volume approximately of size 20m, by default populates 
    nfs, cifs and fcp files,by default it'll populate aprox 15m, you 
    cannot use this routine to populate volumes lesser than this.
    populate_vol("tada-u4","pop_vol",size=>"20m");

=back

=for html 
<hr> </hr>

=cut

sub populate_vol {

    my $cfiler = make_conn(shift);
    my $h = make_host($cfiler);  
    my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		     size => { type => SCALAR, default =>undef },
		     wclient => { type => SCALAR|HASHREF, default =>  '' },
		     uclient => { type => SCALAR|HASHREF, default => 'localhost' },
		     ostype => { type => SCALAR, default => 'solaris' }, 
		     filer_ip => { type => SCALAR, default => "$h->{default_ip}" },
		     sanclient => { type => SCALAR|HASHREF, default => '' },
		     prefix =>  { type => SCALAR, default => '' },
		     nfs_path => { type => SCALAR, default => undef },
		     cifs_path => { type =>SCALAR, default =>  undef },
		     fcp_path => { tpe => SCALAR, default => undef },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		     win_user => {type => SCALAR, default => undef},
		     win_passwd => {type => SCALAR, default => undef},
		     domain => {type=>SCALAR, optional=>1, default=>'',
			         depends=>['win_user','win_passwd'] },
		     workgroup => {type=>SCALAR, optional=>1, default =>''}
		}
	    );
    
    # Local variables used in the routine
    my $source_root;
    #Check if the volume exists, croak if it dosen't
    return eval{
	if(!vol_exists($cfiler,$vol)){
	    croak("$@");
	}
	#export the volume with the right permissions for the files to get created
	#By default, it exports with nosuid, we want it with anon=0
	export_path($cfiler, "/vol/$vol",  OPTIONS =>"anon=0") or croak("$@");
	#*******************************************************#
	#populate with nfs based files                          #
	#*******************************************************#
	populate_nfs_data($cfiler, $vol,
			  uclient => $args{uclient}, 
			  prefix => $args{prefix},
			  path => $args{nfs_path},
			  timeout => $args{timeout})
	    or Carp::croak("Unable to populate data via nfs: $@");                           
	
	#********************************************************#
	#Populate with  cifs data ,mainly   with                 #    
	#acls and streams                                        #
	#********************************************************#
	
	if(defined($args{wclient}) &&  $args{wclient} ne ''){
	    populate_cifs_data($cfiler,$vol,
			       wclient => $args{wclient},
			       workgroup => $args{workgroup},
			       domain => $args{domain},
			       win_user=>$args{win_user},
			       win_passwd=>$args{win_passwd}, 
			       prefix =>$args{prefix}, 
			       path=>$args{cifs_path},
			       timeout=>$args{timeout})
		or Carp::croak("Unable to populate cifs data : $@");
	}
	
	#********************************************#
	#Populate volume with FCP Data(SLEDGEHAMMER) #
	#********************************************#
	if(defined($args{sanclient}) &&  $args{sanclient} ne ''){
	    populate_fcp_data($cfiler,$vol,
			      sanclient => $args{sanclient}, 
			      ostype => $args{ostype},	
			      prefix => $args{prefix},
			      path => $args{fcp_path},
			      timeout => $args{timeout})
		or Carp::croak("Unable to populate fcp data $@");
	} else {
	    #create a  lun on the filer in a mixed style qtree
	    my $qtreename = $args{prefix} . "qtree_mixed";
	    my $qtreepath = "/vol/". $vol ."/" .$qtreename;
	    #create a qtree in the volume
	    if(!qtreePresent($cfiler,$qtreepath)){
		qtreeCreate($cfiler, $qtreepath) or croak("Unable to create the qtree: $@");
	    }
	    #set the security of the qtree to mixed style
	    qtreeSecurity($cfiler,$qtreepath,"mixed") 
		or croak ("Unable to set the security style of qtree $@");
	    my $lunpath="$qtreepath" . "/$args{prefix}" . "testlun";
	    my (undef, undef, $after) = say($cfiler, "lun create -s 4m  -t $args{ostype} $lunpath",-timeout, $args{timeout});
	    if ($after =~ /(no space)|(too small)/i){
		croak("Unable to create lun  $lunpath of size 4m : $after")
	    }
	    logcomment("Using exising lun $lunpath") if ($after =~ /LUN exists/i);
	}
	if (!defined($args{nfs_path})){
	    $source_root = mount($args{filer_ip},"/vol/$vol",client=>$args{uclient} ,timeout=>$args{timeout});
	    if (!defined($source_root)) {
		croak("Could not mount /vol/$vol: $@");
	    }
	    
	}else {
	    $source_root =$args{nfs_path};
	}
	#change the permission of source_root to allow write
	if(defined($args{size})){
	    # write remaining size of $size to the volume,
	    host($args{uclient});
	    my $client_conn = connect('#rsh');
	    exec_shell($client_conn,_which_sudo() . " chmod 0777 $source_root");
	    my %df = df($cfiler, $vol);
	    my $populated_size = "$df{volume}->{used}" . "k";
	    my $diff_size = convert_to_bytes($args{size})-convert_to_bytes($populated_size);
	    my $avail_size = "$df{volume}->{avail}" . "k";
	    $avail_size = convert_to_bytes($avail_size);
	    if($diff_size < $avail_size && $diff_size > 0 ){
		my $filename = $args{prefix} . "file" . time();
		$filename = $source_root . "/" . $filename;
		waflobjCreator::create_file($filename,$diff_size, 
					    client=>$args{uclient},
					    timeout=>$args{timeout})
		or croak("Unable to create a file $@") ;
	    }
	}
	my %df = df($cfiler, $vol);
	return(1);
    };
}


=head2 populate_cifs_data

=over

=item Arguments

=over

=item filer

(Required) conn spec of the filer
else makes a conn spec and complains

=item vol

(Required): Volume which you want to  populate

=item workgroup 

(Optional): If using workgroup to authenticate 

=item domain

(Optional): If using domain to authenticate

=item win_user

(Optional): username used for authentication to the filer via cifs

=item win_passwd

(Optional): password for win_user

=item prefix

(Optional): If this is defined, all the elements in the volume
created will have this prefix.
Defaults to ''

=item timeout

(Optional): Timeout of say 
Defaults to 60 sec

=item wclient

(Optional): client which populates cifs data
Defaults to wclient.

=back

=item Examples

    populate_cifs_data("tada-u4", "pop_vol", wclient=>"tada-c30")
    populate_cifs_data("tada-u4", "pop_vol", wclient=>"tada-c30" ,domain=>"domain1",
		       win_user=>"me", win_passwd=>"me_again")
    populate_cifs_data("tada-u4", "pop_vol", wclient=>"tada-c30" ,prefix=>"1st") 
                      or logresult('fail'," could not populate cifs data");

=item Dependencies 

windows client needs to be minwin_compatible.
Cifs needs to be setup

=back

=for html 
<hr> </hr>

=cut

sub populate_cifs_data {
    my $cfiler = make_conn_and_complain(shift);
    my $h = make_host($cfiler);  
    my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		     wclient => { type => SCALAR, default =>  'localhost' },
		     filer_ip => { type => SCALAR, default => "$h->{default_ip}" },
		     prefix =>  { type => SCALAR, default => '' },
		     path => { type =>SCALAR|UNDEF, default =>  undef },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		     win_user => {type => SCALAR|UNDEF, default => undef},
		     win_passwd => {type => SCALAR|UNDEF, default => undef},
		     domain => {type=>SCALAR, optional=>1, default=>'',
                        depends=>['win_user','win_passwd'] },
		     workgroup => {type=>SCALAR|UNDEF, optional=>1,default=>''}
		}
	    );
    #local variables
    my ($dir_name,$qtreename,$path,$share_drive);
    return eval {
	#If path is not defined, create a default path
	if (!defined($args{path})) {
	    #license cifs
	    license($cfiler,"cifs") or croak("Could not license cifs: $@"); 
	    if(!is_cifs_setup(host => $cfiler->hostp())){
		croak("cifs is not setup");
	    }
	    #create a qtree in the volume, ntfs style
	    $qtreename=$args{prefix} . "qtree_ntfs_" .  time();
	    my $qtreepath= "/vol/". $vol ."/" .$qtreename;
	    qtreeCreate($cfiler, $qtreepath) or 
		croak("Unable to create qtree '$qtreepath': $@\n");
	    qtreeSecurity($cfiler,$qtreepath,'ntfs') or 
		croak ("Unable to set security style of qtree '$qtreepath' to 'ntfs': $@\n");

	    #create a cifs share on the Filer and populate it with Acls and Streams
	    #creating a unique id since share name can be only 8chars,else they'll
	    my ($sec,$min,$hour) = gmtime(time());
	    my $unique_id = join('',$sec,$min,$hour);
	    my $share_name="sh" . $unique_id; 
	    create_cifsshare($cfiler,$share_name,"$qtreepath" ,
                timeout=>$args{timeout}) or croak("Unable to create a cifs share $@");
	    if ($args{domain}){
            $share_drive = map_cifs_share(server => $args{filer_ip}, client => $args{wclient}, share => $share_name, 
					      username => $args{win_user}, password => $args{win_passwd} , 
					      domain => $args{domain},timeout=>$args{timeout})or croak("$@");
	    }else{
            $share_drive = map_cifs_share(server => $args{filer_ip}, client => $args{wclient}, share => $share_name, 
					      username => $args{win_user}, password => $args{win_passwd},
					      workgroup => $args{workgroup} , timeout=>$args{timeout})or croak("$@");

	    }
	} else {
	    $qtreename = "cifsFiles".time();
	    $share_drive = $args{path} . "\\". $qtreename;
	}

	logcomment("Will populate '$share_drive' with CIFS data");
	create_win_data($share_drive, client=>$args{wclient},acl_user => $args{win_user}, 
                        no_of_files=>2 ,timeout=>$args{timeout}) 
		or croak("Unable to create streams or acls : $@\n");
	return $qtreename;
    };
}


=head2 populate_nfs_data

=over

=item Arguments

=over

=item filer

(Required) : filer.

=item vol

(Required) : volume to populate.

=item path

(Optional) : path to populate. This path must exist and has to be mounted.

=item prefix

(Optional) : prefix used in path creation.

=item filer_ip

(Optional) : IP address of filer to use while mounting.
Defaults to default_ip of filer.

=item uclient

(Optional) : unix client from where the data will be written via nfs.

=item timeout

(Optional) : timeout

=back

=item Examples

    populate_nfs_data("tada-u4","pop_vol",uclient=>"tada-c16")

    populate_nfs_data("tada-u4","pop_vol")

Parent directory is created with a prefix "1st"
    populate_nfs_data("tada-u4", "pop_vol",prefix=>"1st")


=item Dependencies 

If path is used, path should be  mounted path on a unix machine

=back

=for html 
<hr> </hr>

=cut

sub populate_nfs_data($$;%){

    my $cfiler = make_conn_and_complain(shift);
    my $h = make_host($cfiler);  
    my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		     uclient => { type => SCALAR, default =>  'localhost' },
		     filer_ip => { type => SCALAR, default => "$h->{default_ip}" },
		     prefix =>  { type => SCALAR, default => '' },
		     path => { type =>SCALAR|UNDEF, default =>  undef },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		}
	    );

    #local variables
    my ($qtreename, $path);
    # NFS mount volume,mount the created vol
    #export the volume with the right permissions for the files to get created
    #By default, it exports with nosuid, we want it with anon=0
    license($cfiler,"nfs") or croak("Could not license nfs: $@"); 
    export_path($cfiler, "/vol/$vol",  OPTIONS =>"anon=0") or croak("$@");

    return eval {
	
	#If path is not defined, create a default path
	if(!defined($args{path})){
	    #mount the volume
	    my $source_root = mount($args{filer_ip},"/vol/$vol",client=>$args{uclient} ,timeout=>$args{timeout});
	    if (!defined($source_root)) {
		croak("Could not mount /vol/$vol: $@");
	    }
	    #create a qtree in the volume, unix style
	    $qtreename = $args{prefix} ."qtree_unix_" . time();
	    my $qtreepath = "/vol/". $vol ."/" .$qtreename;
	    qtreeCreate($cfiler, $qtreepath) or 
		croak("Unable to create qtree '$qtreepath': $@");

	    #Set the root  Directory to a default
	    $path="$source_root/$qtreename/nfsFiles" . time();
	} else {
        # path specified must exist
	    croak("$args{path} does not exist")
            if (!check_directory($args{path},client=>$args{uclient},timeout=>$args{timeout}));
	    $qtreename="nfsFiles".time();
	    $path = $args{path} . "/". $qtreename;
	}

	logcomment("Will populate '$path' with NFS data");
	if(!check_directory($path ,timeout=>$args{timeout})){
	    create_directory($path,
				client=>$args{uclient},
				timeout=>$args{timeout}) or
		Carp::croak("Unable to create the directory $path : $@") ;
	}

	waflobjCreator::create_file("$path/File", "1k", 
				client=>$args{uclient},
				timeout=>$args{timeout})
		or croak("Unable to create a file $@") ;
	say($cfiler, "exportfs", -timeout, $args{timeout});
	create_dev("$path/block_dev",1,1,client=>$args{uclient},
				type=>'b',
				timeout=>$args{timeout})
		or croak("Unable to create a block device $@") ;
	create_dev("$path/char_dev",1,1,client=>$args{uclient},
				type=>'c',
				timeout=>$args{timeout})
		or croak("Unable to create the char device $@") ;
	create_symlink("$path/File","$path/symlinktext",
				client=>$args{uclient},
				timeout=>$args{timeout}) 
		or croak("Unable to create a symlink $@") ;
	create_hardlink("$path/File","$path/hardlinktext",
				client=>$args{uclient},
				timeout=>$args{timeout}) 
		or croak("Unable to create a symlink $@") ;
	create_fifo("$path/pipefile",client=>$args{uclient},
				timeout=>$args{timeout})
		or croak("Unable to create a FIFO $@") ;
	create_socket("$path/sockfile",
				client=>$args{uclient},
				timeout=>$args{timeout})
		or croak("$@");

	return $qtreename;
    };
}


=head2 populate_fcp_data

=over

=item Arguments

=over

=item filer

(Required) : filer.

=item vol

(Required) : volume to populate.

=item path

(Optional) : path to populate. This path must exist and has to be mounted.

=item prefix

(Optional) : prefix used in path creation, if path is not set.


=item filer_ip

(Optional) : IP address of filer to use while mounting.
Defaults to default_ip of filer.

=item sanclient

(Optional) : unix(solaris) client, where the lun is converted to 
a ufs filesystem, mount and populate with data.

=item timeout

(Optional) : timeout

=back

=item Examples

    the parent directory is created with a prefix.
    populate_fcp_data("tada-u4", "pop_vol",prefix=>"1st")


=item Dependencies 

works only on a solaris client.
Filer and sanclient  must be setup for FCP 


=back

=for html 
<hr> </hr>

=cut

sub populate_fcp_data($$;%){
    my $cfiler = make_conn_and_complain(shift);
    my $h =make_host($cfiler);
    my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		     sanclient => { type => SCALAR, default =>  'localhost' },
		     ostype => { type => SCALAR, default => 'solaris' }, 
		     filer_ip => { type => SCALAR, default => "$h->{default_ip}" },
		     prefix =>  { type => SCALAR, default => '' },
		     path => { type =>BOOLEAN, default=>undef },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		}
	    );


    #local arguments
    my ($path,$qtreename,$rawlun,$lunpath,$lun_mountpt);

    license($cfiler,"fcp") or croak("Could not license fcp: $@"); 

    #If path is not defined, creates a default path
    return eval {
	if(!defined($args{path})){
	    croak("$args{path} does not exist")
		if (!check_directory($args{path},client=>$args{sanclient},timeout=>$args{timeout}));
	    $qtreename = $args{prefix} ."qtree_mixed_" . time();
	    my $qtreepath = "/vol/". $vol ."/" .$qtreename;
	    #create a qtree in the volume
	    logcomment("Creating qtree '$qtreepath'");
	    qtreeCreate($cfiler, $qtreepath) or 
		croak("Unable to create qtree '$qtreepath': $@");

	    #set the security of the qtree to mixed style
	    qtreeSecurity($cfiler,$qtreepath,"mixed") or 
		croak ("Unable to set the security style of qtree '$qtreepath' : $@");

	    my $lunname= $args{prefix} ."testlun" . time();
	    $lunpath="$qtreepath". "/" . $lunname;
	    #Set the root  Directory to a default
	    logcomment("Creating lun '$lunpath'");
	    $rawlun=create_lun($cfiler,$args{sanclient},ostype=>$args{ostype},lun_path=>$lunpath,lun_size=>'4m')or croak("$@");
	    $lun_mountpt=mkfs_mount_fcp($cfiler,$args{sanclient},$rawlun,timeout=>$args{timeout})or croak("$@");
	    $path=$lun_mountpt;
	}else {
        # path specified must exist
	    croak("$args{path} does not exist")
            if (!check_directory($args{path},client=>$args{sanclient},timeout=>$args{timeout}));
	    $path = $args{path};
	}

	# chmod the mount point
	my $cc = make_conn($args{sanclient});
	exec_shell($cc, _which_sudo() . " chmod 777 $path"); # 

	#*******************************************************#
	#populate fcp path with nfs based files                 #
	#*******************************************************#
	$qtreename = populate_nfs_data($cfiler, $vol,
			  uclient => $args{sanclient}, 
			  path => $path,
			  timeout => $args{timeout})
	    or Carp::croak("Unable to populate lun via nfs: $@");                           
	
    return $qtreename;
    };
}


=head2 populate_qtree

=over

=item Arguments

=over

=item filer

(Required) : connspec of a filer

=item vol

(Required) : volume to populate.

=item qtree

(Optional) : this routine populates in the qtree.

=item prefix

(Optional) : If prefix is given all the files created are given this prefix.

=item filer_ip

(Optional) : IP address of filer to use while mounting.
Defaults to default_ip of filer.

=item sanclient

(Optional) : unix(solaris) client, where the lun is converted to 
a ufs filesystem,mounted  and gets populated via nfs.

=item wclient

(Optional):Windows client through which volume or path is populated via cifs,
If not provided, volume will not be populated via cifs 
Note:Windows client should be minwin_compatible

=item uclient 

(Optional):Client through which volume is populated via nfs
Defaults to localhost.

=item win_user

(Optional): username used for authentication to the filer via cifs

=item win_passwd

(Optional): password for win_user

=item workgroup 

(Optional): If using workgroup authentication

=item domain

(Optional): If using domain authentication

=item timeout

(Optional) : timeout

=back

=item Examples

    #the parent directory is created with a prefix.
    populate_qtree("tada-u4", "pop_vol","qtree1",prefix=>"1st")


=item Dependencies 

sanclient is tested only on a solaris box.
sanclient and filer must be setup for FCP.
if wclient is given, it should be minwin_compatible.


=back

=for html 
<hr> </hr>

=cut

sub populate_qtree {

    my $cfiler = make_conn(shift);
    my $h = make_host($cfiler);  
    my($vol,$qtree, %args) = validate_mixed(
		@_,
		[ { type => SCALAR },{ type => SCALAR } ],
		{
		     uclient => { type => SCALAR, default => 'localhost' },
		     wclient => { type => BOOLEAN, default =>  '' },
		     ostype => { type => SCALAR, default => 'solaris' },
		     sanclient => { type => BOOLEAN, default => '' },
		     filer_ip => { type => SCALAR, default => "$h->{default_ip}" },
		     prefix =>  { type => SCALAR, default => '' },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
                     nfs_path => { type => SCALAR, default => undef },
		     win_user => {type => BOOLEAN, default => undef},
		     win_passwd => {type => BOOLEAN, default => undef},
		     domain => {type=>BOOLEAN, optional=>1, default=>'',
                         depends=>['win_user','win_passwd'], },
		     workgroup => {type=>SCALAR, optional=>1,default => ''}
		}
	    );
    
    # Local variables used in the routine
    my ($source_root,$qtree_path, $nfs_path);
	my %return;

    #Check if the volume exists, croak if it dosen't
    $qtree_path = "/vol/$vol/$qtree";
    return eval{
	if(!vol_exists($cfiler,$vol)){
	    croak("$@");
	} elsif (!qtreePresent($cfiler,$qtree_path)){
	    logcomment("Qtree '$qtree_path' does not exist, creating it");
	    qtreeCreate($cfiler, $qtree_path) or croak("Unable to create the qtree: $@");
	}
	
	#change the security to mixed
	qtreeSecurity($cfiler, $qtree_path, "mixed") or croak($@);
	#export the volume with the right permissions for the files to get created
	#By default, it exports with nosuid, we want it with anon=0
	export_path($cfiler, "/vol/$vol",  OPTIONS =>"anon=0") or croak("$@");
	
	
	#*******************************************************#
	#populate with nfs based files                          #
	#*******************************************************#
	#mount the qtree

        if (!defined($args{nfs_path})){
	    $nfs_path = mount($args{filer_ip},"$qtree_path",
                        client=>$args{uclient},
                        timeout=>$args{timeout});
	    if (!defined($nfs_path)) {
	        croak("Could not mount $qtree_path: $@");
	    }
        } else {
            $nfs_path = "$args{nfs_path}/$qtree";
        }
	    
	$return{nfs} = populate_nfs_data($cfiler, $vol, path=>$nfs_path, 
			  uclient=>$args{uclient}, 
			  prefix =>$args{prefix},
			  timeout=>$args{timeout})
	    or croak("$@");
	logcomment("Populated data through NFS. Path is $return{nfs}");
	
	#********************************************************#
	#Populate volume with  NT data ,populate  with           #    
	#acls and streams                                        #
	#********************************************************#
	if($args{wclient} ne ''){
	    #creating a unique id since share name can be only 8chars,else they'll
	    my ($sec,$min,$hour) = gmtime(time());
	    my $unique_id = join('',$sec,$min,$hour);
	    my $share_name="sh" . $unique_id; 
	    create_cifsshare($cfiler,$share_name,$qtree_path,
                        timeout=>$args{timeout}) 
		or croak("Unable to create a cifs share $@");
	    my $share_drive = map_cifs_share(server=>$args{filer_ip},
					     client => $args{wclient},
					     share => $share_name,
					     username => $args{win_user},
					     password => $args{win_passwd},
					     workgroup => $args{workgroup} ,
					     domain => $args{domain},
					     timeout=>$args{timeout})
		or croak("$@");
        $return{share} = $share_name;
        $return{share_drive} = $share_drive;
        logcomment("Created cifs share $share_name");

	    logcomment("Attempting to populate through cifs");
	    $return{cifs} = populate_cifs_data($cfiler,$vol,wclient=>$args{wclient},
			       win_user=>$args{win_user},
			       win_passwd=>$args{win_passwd}, 
			       prefix =>$args{prefix}, 
			       path=>$share_drive ,
			       timeout=>$args{timeout})
		or croak("$@");
	    logcomment("Populated data through CIFS. Path is $return{cifs}");
	}
	
	#********************************************#
	#Populate volume with FCP Data(SLEDGEHAMMER) #
	#********************************************#
	#set the security of the qtree to mixed style
	my $lunname = $args{prefix} . 'testlun' . time();
	my $lunpath = $qtree_path . "/" . $lunname;
	my (undef, undef, $after) = say($cfiler, "lun create -s 4m  -t $args{ostype} $lunpath",-timeout, $args{timeout});
	if ($after =~ /(no space)|(too small)/i){
	    croak("Unable to create lun $lunpath of size 4m : $after")
	}
	$return{lun} = $lunname;
	logcomment("Created lun $lunname");
	
	if($args{sanclient} ne ''){
	    my $rawlun = create_lun($cfiler,$args{sanclient},ostype=>$args{ostype},lun_path=>$lunpath,lun_size=>'4m')
		or croak("$@");
	    my $lun_mountpt = mkfs_mount_fcp($cfiler,$args{sanclient},
					     $rawlun,timeout=>$args{timeout})
                or croak("$@\n");
	    
	    my $conn_san = make_conn($args{sanclient});
	    my (undef, undef, $out) = say($conn_san, _which_sudo() . " chmod 0777 $lun_mountpt");
	    croak("Could not chmod $lun_mountpt on $args{sanclient} : $out")
		if ($out);
	    
	    my $fcp_path = "$lun_mountpt/allFiles";
	    create_directory($fcp_path,
			     client=>$args{sanclient},
			     timeout=>$args{timeout}) or
				 croak("Unable to create the directory '$fcp_path' : $@\n");
	    
	    $return{fcp} = populate_fcp_data($cfiler,$vol,
					     sanclient=>$args{sanclient},
					     ostype=>$args{ostype}, 
					     prefix =>$args{prefix}, 
					     path=>$fcp_path,
					     timeout=>$args{timeout})
		or croak("$@\n");
	    logcomment("Populated data through FCP. Path is $return{fcp}");
	}
	return(%return);
    };
}


=head2 treecompare

=over

=item Arguments

=over

=item filer

(Required) : conn spec of the filer


=item filer_dst

(Optional) : conn spec of the destination filer, if it is other than 
filer, defaults to filer

=item src_snappath

(Required): treecompare compares snapshots,
src_snappath should be the path on the local filer(path on $filer),
the filer on which treecompare is run. 

=item dest_snappath

(Required): Path to the destination snapshot to be compared.

=item @skip

(Optional): Array of sentences that you want to skip in the 
treecompare output, defaults to undef....
Ex:treecompare($cfiler, $src, $dst, 
   skip=>["avoid space reservation", "avoid\\s+vvol\\s+stuff"], args=>'-i')

=item args 

(Optional): Flags of  treecompare

                          -A Don't compare ACLs.
                          -c Don't compare ctime.
                          -i Compare inode number, generation and treeid.
                          -V Don't compare vdisk metadata/timesa
                          -L Specify check level.
                          -k Specify throttle value.

=item operation 

(Optional):treecompare output could be valid based on the operation we ran 
For dumprestore, on dest side there is an extra file, so have to skip it 
By default operation is other(stands for other than dump/restore operations)
If you are comparing dump/restore src and dest, you need to use 
operation=>"dr"

=back

=item Examples                         

treecompare($filer,$src_snappath,$dest_snappath,timeout=>'600',filer_dst=>"tada-u23")
If  filer_dst is not specified ,filer_dst defaults to $filer.

=item Return

(1 if no mismatches were found, else croak's)

=back

=for html 
<hr> </hr>

=cut

sub treecompare {
    my $cfiler = make_conn_and_complain(shift);
    my $src_host = make_host($cfiler);  
    my($src_snappath,$dest_snappath, %args) = validate_mixed(
		@_,
		[ { type => SCALAR },  { type => SCALAR }],
		{
		     skip => { default => undef, type => ARRAYREF|UNDEF },
		     filer_dst => {type => SCALAR|HASHREF, default => $cfiler},
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		     operation => {type => SCALAR, default => "other"},
		     args => {type => SCALAR, default => ""},
		}
	    );

    my $src_hostname = $src_host->hostname;
    $src_hostname =~ s/\..*//;
    my $dest_host = make_host($args{filer_dst});
    my $dest_hostname = $dest_host->hostname;
    $dest_hostname =~ s/\..*//;
    #local variables
    my ($output,@falseops);
    return eval{
	# Keep just the hostname name if the full domain name was given
	# For dumprestore, on the destination restore symbol table is 
	# created, skip this from the output for dumprestore....
	if ($args{operation} eq 'dr'){
	    logcomment("Skipping comparison of ctimes '-c'");
	    $args{args} .= ' -c';
	}
	(undef,undef,$output) = say($cfiler,"priv set -q diag; treecompare start $args{args} ".
				    "$dest_hostname:$dest_snappath $src_hostname:$src_snappath",
				    '-timeout',$args{timeout});
	
	#Break into lines
	#Check if treecompare failed, if fail croak here
	if ($output =~ /treecompare\s+failed/){
	    croak("$output");
	}
	my @lines = split(/\n/ , $output);
	my @filter_lines;
	if ($args{skip}){
	    my $skip_array_ref = $args{skip};
	LINE: for my $line (@lines) {
		foreach my $skip_string (@$skip_array_ref){
		    if ($line =~ /$skip_string/){
			logcomment("SKIPPED: $line");
			next LINE;
		    }
		}
		push @filter_lines, $line;
	    }
	    @lines=@filter_lines;
	}
	for my $line (@lines) {
	    if ($line =~ /Found.*/) {
		if ($args{operation} eq 'dr') { 
		    if ($line !~ /Found\s+\w+\s+only\s+directory\s+entry\s+\.\/restore_symboltable/){
			push @falseops , $line;
		    } else {
			logcomment("skipped restore_symboltable mismatch");
		    }
		}else{
		    push @falseops , $line;
		}  
	    }elsif($line =~ /tree1_only\s+=\s+(\d+),\s+tree2_only\s+=\s+(\d+).*/){
		my $tree1 = $1;
		my $tree2 = $2;
		if ($args{operation} eq 'dr'){
			if(($tree1-$tree2==1) or ($tree2-$tree1 == 1)){
    			logcomment("skipped one difference for restore symbol table");
		    }else{
			if($tree1 == 0 && $tree2 == 0){
			    push @falseops , "Restore symbol table is not found\n";
			}else{
			    push @falseops , "There are missing dirs or files\n";
			}
		    }
		}elsif (($tree1 != 0 || $tree2 != 0) && $args{operation} ne 'dr'){
		    push @falseops , "There are missing dirs or files (See $line)\n";
		}
	    }
	}
	if ($#falseops >= 0) {
	    for my $falseop (0..$#falseops) {
		logcomment("$falseops[$falseop]");
	    }
	    croak("There are some mismatches, please check treecompare results @falseops\n");
	} else {
	    logcomment("No differences found between ".
		       "$dest_hostname:$dest_snappath  $src_hostname:$src_snappath ");
	    return(1);
	}
    };
}

=head2 treecompare_vol

=over

=item Description

sub routine to  treecompare at volume or qtree level, treecompare_vol 
takes source and dest volume (path in case of qtree) names and 
passes down to treecompare routine.


=item Arguments

=over

=item filer

(Required) : conn spec of the filer

=item filer_dst

(Optional) : conn spec of the destination filer, if it is other than 
filer, defaults to filer

=item src_path

(Required): src volume in case of volume or path in case of qtree compare

=item dst_path

(Required): dest volume in case of volume or path in case of qtree compare

=item @skip

(Optional): Array of sentences that you want to skip in the 
treecompare output, defaults to undef....
Ex:treecompare($cfiler, $src, $dst, 
skip=>["avoid space reservation", "avoid\\s+vvol\\s+stuff"], args=>'-i')

=item args 

(Optional): Flags of  treecompare

                          -A Don't compare ACLs.
                          -c Don't compare ctime.
                          -V Don't compare vdisk metadata/times
                          -i Compare inode number, generation and treeid.
                          -L Specify check level.
                          -k Specify throttle value.

=item operation 

(Optional):treecompare output could be valid based on the operation we ran 
For dumprestore, on dest side there is an extra file, so have to skip it 
By default operation is other(stands for other than dump/restore operations)
If you are comparing dump/restore src and dest, you need to use 
operation=>"dr"

=item type

(Optional) : type of data to compare. Possible values : 'qtree' / 'vol'. 
Default: 'vol'

=back

=item Examples                         

	host('tada-u22');
	my $conn_src = connect("#rsh");

	# compare volumes
	treecompare_vol($conn_src,'vol1','vol2',timeout=>600,filer_dst=>'tada-u23')

	# compare qtrees
	treecompare_vol($conn_src,'/vol/vol1/qtree1','/vol/vol2/qtree2',type=>'qtree',filer_dst=>'tada-u23')


=item Return

(1 if no mismatches were found, else croak's)

=back

=for html 
<hr> </hr>

=cut

sub treecompare_vol {

    my $cfiler_src = make_conn_and_complain(shift);
    my ($src_path, $dst_path, %args) = validate_mixed(
		@_,
		[ { type => SCALAR }, { type => SCALAR }
		], {
		     skip => { default => undef, type => ARRAYREF },
		     filer_dst => {type => SCALAR|HASHREF, default => $cfiler_src},
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		     operation => {type => SCALAR, default => 'other'},
		     args => {type => SCALAR, default => ''},
		     type => {type => SCALAR, default => 'vol', regex=>qr/^qtree|vol$/},
		}
	    );

    my $src_host = make_host($cfiler_src);  
    my $src_hostname = $src_host->hostname;
    my $dest_host = make_host($args{filer_dst});
    my $dest_hostname = $dest_host->hostname;
    my $cfiler_dst =  make_conn($args{filer_dst});

    my $src_snap='src_snap'.time();
    my $dst_snap='dst_snap'.time();

	my ($src_vol, $dst_vol);
	my ($src_qtree, $dst_qtree);
	if ($args{type} eq 'qtree') {
		($src_vol, $src_qtree) = ($src_path =~ /\/vol\/(\w+)\/?(\S*)/);
		($dst_vol, $dst_qtree) = ($dst_path =~ /\/vol\/(\w+)\/?(\S*)/);
		croak("Was unable to extract src and dst volume from path\n")
			if (!$src_vol || !$dst_vol);
		logcomment("SRC vol is '$src_vol' and DST vol is '$dst_vol'");
		croak("Was unable to extract src and dst qtree from path\n")
			if (!$src_qtree || !$dst_qtree);
		logcomment("SRC qtree is '$src_qtree' and DST qtree is '$dst_qtree'");
	} else {
		$src_vol = $src_path;
		$dst_vol = $dst_path;
		logcomment("SRC vol is '$src_vol' and DST vol is '$dst_vol'");
	}

    return eval {
	#create snapshots
	snapshot_create($cfiler_src,$src_vol,$src_snap) or 
        croak("Could not create a snapshot '$src_snap' of src volume $src_vol $@\n");
	snapshot_create($cfiler_dst,$dst_vol,$dst_snap) or 
        croak("Could not create a snapshot '$dst_snap' of dst volume $dst_vol $@\n");
	my $count=0;
	do{
	    $count++;
	    sleep(20);
	} while (!snapshot_exists($cfiler_src,$src_vol,$src_snap) && 
	         !snapshot_exists($cfiler_dst,$dst_vol,$dst_snap) && $count<5);

	my $src_snappath = '/vol/' . $src_vol . '/.snapshot/' . $src_snap;
	my $dst_snappath = '/vol/' . $dst_vol . '/.snapshot/' . $dst_snap;
	if ($args{type} eq 'qtree') {
		$src_snappath .= '/' . $src_qtree;
		$dst_snappath .= '/' . $dst_qtree;
	}
	# This takes snapshot paths as parameters
	treecompare($cfiler_src,$src_snappath, $dst_snappath, 
		    timeout=>$args{timeout}, 
		    skip=>$args{skip}, 
		    args=>$args{args}, 
		    filer_dst=>$cfiler_dst, 
		    operation=>$args{operation}) or croak("there are treecompare differences $@");
	# Cleanup 
	snapshot_delete($cfiler_src,$src_vol,$src_snap) or 
            croak("Could not delete source snapshot '$src_snap' : $@\n");
	snapshot_delete($cfiler_dst,$dst_vol,$dst_snap) or
            croak("Could not delete dest snapshot '$dst_snap' : $@\n");
	return 1;
    };
}

## @name ::shared_drive_syntax
##
## @example shared_drive_syntax('Z:')
##  Checks that shared volume begins with \\(but using  \\\\, since back slash is an escape 
##                                       char in perl) or a letter followed by a colon
##
## @arg $drive(Required): shared drive

sub shared_drive_syntax($) {
	my ($drive) = @_;
	my $first_two_letters = substr($drive, 0, 2);
	if ($first_two_letters eq '\\\\' | $first_two_letters eq '\\') {
		return 1;
	}
	if ($first_two_letters =~ /[a-zA-Z]\:/ ) {
		return 1;
	}
	
	tstlog(-warn => 
	       "drive $drive does not begin with \\\\ or <letter>:");
	return 0;
}

=head2 create_win_data

=over

=item Description

sub routine to create windows data ,
currently it  creates  streams and acls 

=item Arguments

=over

=item share_drive

(Required):  mapped path on windows/cifs client 

=item win_client

(Optional): windows client

=item no_of_files

(Optional) : creates $no_of_files no of acl and stream file dirs

=item populate_streams

(Optional) : creates or skips streams creation based on the flag
                        Defaults to 1 , to create streams 

=item populate_acls

(Optional) : creates or skips acl creation based on the flag
                        Defaults to 1 , to create files/dirs and sets acls 

=back 

=item Dependencies

root user needs to be there on the PC client
rshd needs to be installed on the client 

=item Examples

 create_win_data($share_drive,client => $winclient) 

 if you want to create specific number of files
 create_win_data($share_drive,client => $winclient,no_of_files=>3)

=item Returns 

(1 if everything passes  else croak's)

=back

=for html 
<hr> </hr>

=cut

sub create_win_data{

    my($share_drive, %args) = validate_mixed(
		@_,
		[ { type => SCALAR }],
		{
		     no_of_files => {type => SCALAR, optional => 1, default => 2},
		     timeout => {type => SCALAR, optional => 1, default => 60},   # timeout for say
                     client => {type => SCALAR, optional => 1, default => 60},
		     acl_user => {type => SCALAR, optional => 1, default => 'root'},
		     populate_streams => {type => SCALAR, optional =>1, default => '1'},
		     populate_acls    => {type => SCALAR, optional =>1, default => '1'}

		}
	    );

    my $i;

    #array of acls strings that'll be applied to set acls on  files and directories
    my @acls = ( "/e /d $args{acl_user}", "/e /g $args{acl_user}:R","/e /g $args{acl_user}:C","/e /g $args{acl_user}:F",
		 "/e /d $args{acl_user}", "/e /g $args{acl_user}:C","/e /g $args{acl_user}:W","/e /g $args{acl_user}:W"
	       );
    #--------------------------------------------------------#
    #create $no_of_files directories with streams
    #create another directory named filestreamdir
    #create $no_of_files files with streams in filestreamdir
    #--------------------------------------------------------#

    
    return eval{
        if($share_drive !~ /^\s*\\\\/ && !shared_drive_syntax($share_drive)) {
		logresult('warn',"incorrect drive syntax for drive $share_drive");
	}
        if($args{populate_streams}){
	    my $win_dir=$share_drive . "\\NtStreams";
	    create_dironWin($win_dir,$args{client} ,timeout=>$args{timeout}) or croak("could not create a directory from PC: $@");
	    #creating $no_of_files dirstreamfile
	    my ($curdir,$curfile,@streams);
	    for my $dirindex (0 .. $args{no_of_files}) {
		$curdir="dir_with_" . $dirindex . "streams";
		create_dironWin("$win_dir\\$curdir",$args{client} ,timeout=>$args{timeout}) or croak("could not create a directory from PC: $@");
		#flush the array
		$#streams=-1;
		for($i=0;$i <= $dirindex;$i++) {
		    push @streams,"dirstream".$i;
		    
		}
		create_ntstream($win_dir,$curdir,\@streams,$args{client} ,timeout=>$args{timeout}) or croak("Unable to create NT streams $@");
	    }
	    #--------------------------------------------------------#
	    #creating a directory with $no_of_files files with streams
	    #--------------------------------------------------------#
	    
	    create_dironWin("$win_dir\\filestreamdir",$args{client} ,timeout=>$args{timeout}) or croak("could not create a directory from PC: $@");
	    for my $fileindex (0 .. $args{no_of_files}) {
		$curfile="file_with_" . $fileindex . "streams";
		create_fileonWin("$win_dir\\filestreamdir\\$curfile",$args{client} ,timeout=>$args{timeout})  or croak("could not create a file from PC: $@");
		#flush the array
		$#streams=-1;
		for ($i=0;$i <= $fileindex;$i++) {
		    $streams[$i]="filestream".$i;
		}
		create_ntstream("$win_dir\\filestreamdir",$curfile,\@streams,$args{client} ,timeout=>$args{timeout}) or croak("Unable to create NT streams $@");;
	    }
	}
	if($args{populate_acls}){
    
	    #--------------------------------------------------------#
	    #create $no_of_files directories
	    #set acls on them
	    #--------------------------------------------------------#
	    my $win_dir=$share_drive . "\\NtAcls";
	    create_dironWin($win_dir,$args{client} ,timeout=>$args{timeout}) or croak("$@");
	    for my $dirindex (0 .. $args{no_of_files}) {
		my $curdir="dir_with_" . $dirindex . "acls";
		create_dironWin("$win_dir\\$curdir",$args{client} ,timeout=>$args{timeout}) or croak("$@");
		my $aclchangestr= $acls[rand @acls];
		create_acls("$win_dir\\$curdir",$aclchangestr,$args{client} ,timeout=>$args{timeout}) or croak("$@");
	    }
	    
	    #--------------------------------------------------------#
	    #create another directory fileaclsdir
	    #create $no_of_files files with different acls
	    #--------------------------------------------------------#
	    create_dironWin("$win_dir\\fileacldir",$args{client} ,timeout=>$args{timeout}) or croak("$@");
	    for my $fileindex (0 .. $args{no_of_files}) {
		my $curfile="file_with_" . $fileindex . "acls";
		create_fileonWin("$win_dir\\fileacldir\\$curfile",$args{client} ,timeout=>$args{timeout}) or croak("$@");
		my $aclchangestr= $acls[rand @acls];
		create_acls("$win_dir\\fileacldir\\$curfile",$aclchangestr,$args{client} ,timeout=>$args{timeout}) or croak("$@");
	    }
	}
	return(1);
    };
}

=head2 modify_data

=over

=item Description
 
Creates new data in the existing volume.

=item Examples 

    modify_data($filer,$vol,client=>$client,prefix=>"clone1")

    this function creates qtree with prefix starting with clone1
    and all the parent directories(they become unique in the path) 
    will get prefixed with clone1		   

    @returns (1 or croaks if it cannot modify the vol)

=item Arguments

=over

=item filer

(Required) : connspec of a filer

=item vol

(Required) : volume to populate.

=item prefix

(Optional) : If prefix is given all the files created are given this prefix.

=item filer_ip

(Optional) : IP address of filer to use while mounting.
Defaults to default_ip of filer.

=item sanclient

(Optional) : unix(solaris) client, where the lun is converted to 
a ufs filesystem, mounted  and populate the lun via nfs.

=item path

(Optional): If this is given, All the nfs files will be created in this
path (this is a mounted path on the client)

=item timeout

(Optional) : timeout

=back

=item status

deprecated, use populate_qtree/populate_nfs_data/populate_cifs_data/populate_fcp_data/populate_vol
with different prefix.

=back

=for html 
<hr> </hr>

=cut

sub modify_data {
    my $cfiler = make_conn(shift);
    my $h = make_host($cfiler);  
    my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
		     client => { type => SCALAR, default => 'localhost' },
		     ostype => { type => SCALAR, default => 'solaris' },	
		     filer_ip => { type => SCALAR, default => $h->{default_ip} },
		     prefix =>  { type => SCALAR, default => '' },
		     path => { type => SCALAR, default => undef },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		}
	    );

    #Create a qtree and create a random file in it 
    #write some random data into the clone

    #local variables
    my ($source_root);
    return eval{
	#If path is given, just use the path...
	if(!defined($args{path})){
	    #create a qtree in the volume, unix style
	    my $qtreename=$args{prefix} . "qtree_unix";
	    my $qtreepath = "/vol/". $vol ."/" .$qtreename;
	    if(qtreePresent($cfiler, $qtreepath)){
		$qtreename .= time();
		$qtreepath = "/vol/". $vol ."/" .$qtreename;
	    }
	    # qtree will be created in populate_qtree itself
	    populate_qtree($cfiler,$vol,$qtreename,ostype=>$args{ostype},filer_ip=>$args{filer_ip},
			   prefix=>$args{prefix},timeout=>$args{timeout}) 
		or croak("Modify Failed: $@");
	}else {
	    #we expect a mounted path , and populate path via nfs
	    populate_nfs_data($cfiler,$vol,prefix => $args{prefix},path => $args{path},
			 timeout => $args{timeout},filer_ip => $args{filer_ip},uclient => $args{client}) 
		or croak("Modify Failed: $@");
	    
	}	
    };
}


=head2 delete_nfs_data

=over 

=item Description

Deletes an existing tree in a volume via a nfs client and
tree contains only data populated via nfs (i.e. no data
populated via cifs or lun)


=item Arguments

=over

=item filer 

(Required): connspec of a filer

=item treepath

(Required): path of the tree which is to be deleted.
The tree can also be a qtree.

=item uclient 

(Optional): name of unix client. Default: localhost

=item filer_ip 

(Optional): IP address to filer. Default: default_ip

=item timeout

(Optional): timeout to say(). Default: 60

=back

=item Returns

returns (1 or croaks if it cannot delete the tree)
If tree did not exist will return 1

=back

=for html 
<hr> </hr>

=cut
sub delete_nfs_data ( $$% ) {

    my $filer = make_conn(shift);
    my $h = make_host($filer);  

    my ($treepath, %args) = validate_mixed(
		@_,
		[ { type => SCALAR }, ],
		{
		     uclient => { type => SCALAR, default => 'localhost' },
		     filer_ip => { type => SCALAR, default => $h->{default_ip} },
		     timeout => {type => SCALAR, default => 60},   # timeout for say
		}
	    );

    return eval{

        my ($vol, $treename) = ($treepath =~ qr/^\/vol\/(\w+)\/(.*)$/);
	
	# mount the volume via NFS
	my $mountpoint = mount($args{filer_ip},"/vol/$vol",
                        client=>$args{uclient},
                        timeout=>$args{timeout});
	croak("Could not mount volume '$vol' :$@") if (!defined($mountpoint));
	
	# remove the tree
	my $path = $mountpoint.'/'.$treename;
	logcomment("DBG: Will remove nfs tree '$path'");
        my $uh = host($args{uclient});
        delete_tree($uh, $path, CONTENTS_ONLY=>0, VERBOSE=>5);
        my $err = $@;
        # Ignore unmount error
        umount($mountpoint, force=>1, client=>$args{uclient}, 
               timeout=>$args{timeout});
	croak("Could not delete nfs tree '$path'. :$err") if ($err);
	return(1);
    };
}

=head2 delete_cifs_data

=over 

=item Description

Deletes an existing tree in a volume via a windows (cifs) client.

=item Arguments

=over

=item filer 

(Required): connspec of a filer

=item treepath

(Required): path of the tree which is to be deleted.
The tree can also be a qtree.

=item share 

(Required): name of share.

=item wclient 

(Optional): name of windows client. Default: localhost

=item filer_ip 

(Optional): IP address to filer. Default: default_ip

=item timeout

(Optional): timeout to say(). Default: 60

=item win_user

(Optional): windows user.

=item win_passwd

(Optional): windows password.

=item domain

(Optional): name of domain to use.

=item workgroup

(Optional): name of workgroup to use.

=back

=item Returns

returns (1 or croaks if it cannot delete the tree)
If tree did not exist will return 1

=back

=for html 
<hr> </hr>

=cut
sub delete_cifs_data ( $$% ) {

    my $conn = make_conn(shift);
    my $h = make_host($conn);  

    my ($treepath, %args) = validate_mixed(
		@_,
		[ { type=>SCALAR } ],
		{
		     wclient => { type=>SCALAR, default=>'localhost' },
		     filer_ip => { type=>SCALAR, default=>$h->{default_ip} },
		     timeout => {type=>SCALAR, default=>60},   # timeout for say
		     win_user => {type=>SCALAR|UNDEF, default=>undef},
		     win_passwd => {type=>SCALAR|UNDEF, default=>undef},
		     domain => {type=>SCALAR, optional=>1, default=>'',
                        depends=>['win_user','win_passwd'] },
		     workgroup => {type=>SCALAR|UNDEF, optional=>1, default=>''}
		}
	    );

    return eval{

    	my ($vol, $basetree, $treename) = ($treepath =~ qr/^\/vol\/(\w+)\/(\w+)\/(.*)$/);

    	# create a cifs share on the Filer
    	# creating a unique id since share name can be only 8chars,else they'll
        my ($sec,$min,$hour) = gmtime(time());
        my $unique_id = join('',$sec,$min,$hour);
        my $share = 'sh' . $unique_id; 
        logcomment("Creating temp share '$share' of '/vol/$vol/$basetree'");
        create_cifsshare($conn, $share , "/vol/$vol/$basetree", timeout=>200)
		or croak("Unable to create a cifs share '$share' : $@");
	
        # map the CIFS share
        logcomment("Mapping temp share '$share' on win client $args{wclient}");
        my $share_drive = map_cifs_share(server=>$args{filer_ip},
                     client=>$args{wclient},
                     share=>$share,
                     username=>$args{win_user},
                     password=>$args{win_passwd},
                     workgroup=>$args{workgroup},
                     domain=>$args{domain},
                     timeout=>200,
                     drive=>"*"
                    )
                or croak("$@\n");
        logcomment("Share drive is $share_drive");

	# remove the tree
	my $path = $share_drive . '\\' . $treename;
	logcomment("DBG: Will remove cifs tree '$path'");
        my $wh = host($args{wclient});
        delete_tree($wh, $path, CONTENTS_ONLY=>0, VERBOSE=>5);
        my $err = $@;
        # Ignore unmapping error
        unmap_cifs_share(drive => $share_drive, client=>$args{wclient});
	croak("Could not delete cifs tree '$path'. :$err") if ($err);

        logcomment("Will remove temp cifs share '$share'");
        share_delete($conn, $share)
            or croak("$@\n");
        logcomment("Deleted temp share '$share'");

	return(1);
    };
}


=head2 unpopulate_qtree

=over 

=item Description

Opposite of populate_qtree routine. Will remove the data in the qtree via
nfs, cifs, delete the cifs share and also remove the lun created.
It will purge the qtree also by default. This can be changed by setting purge_tree=>0

=item Arguments

=over

=item filer 

(Required): connspec of a filer

=item vol

(Required): volume on which this qtree lives.

=item qtreename

(Required): name of qtree

=item items

(Required): hash of items to be deleted. Basically the return value of populate_qtree.

=item uclient 

(Optional): name of unix client. Default: localhost.

=item sanclient 

(Optional): name of SAN (fcp) client.

=item wclient 

(Optional): name of windows client. Default: localhost

=item purge_tree

(Optional): flag denoting if the tree itself is to be removed or not. Default : 1

=item filer_ip 

(Optional): IP address to filer. Default: default_ip

=item timeout

(Optional): timeout to say(). Default: 60

=item win_user

(Optional): windows user.

=item win_passwd

(Optional): windows password.

=item domain

(Optional): name of domain to use.

=item workgroup

(Optional): name of workgroup to use.

=back

=item Returns

returns (1 or croaks if it cannot delete the tree)
If tree did not exist will return 1

=items Example

    my %items = populate_qtree($fc, $FILER_VOL, $qtree, 
                            uclient=>$CLIENT,
                            timeout=>500,
                            wclient=>$WCLIENT,
                            win_user=>$CIFS_USER,
                            win_passwd=>$CIFS_PASSWORD,
                            sanclient=>$SANCLIENT,
                            );
    if ($@) {
        logresult('fail',"populate volume '$FILER_VOL' :$@");
    } else {
        logresult('pass',"populate volume '$FILER_VOL'");
    }

    unpopulate_qtree($fc, $FILER_VOL, $qtree, items=>\%items, 
                                    uclient=>$CLIENT, 
                                    wclient=>$WCLIENT,
                                    win_user=>$CIFS_USER,
                                    win_passwd=>$CIFS_PASSWORD,
                                    sanclient=>$SANCLIENT,
                                );
    if ($@) {
        logresult('fail',"unpopulate volume '$FILER_VOL' :$@");
    } else {
        logresult('pass',"unpopulated volume '$FILER_VOL'");
    }


=back

=for html 
<hr> </hr>

=cut
sub unpopulate_qtree ( $$$% ) {

    my $conn = make_conn(shift);
    my $h = make_host($conn);  

    my ($vol, $qtreename, %args) = validate_mixed(
		@_,
		[ { type => SCALAR }, 
		  { type => SCALAR } ],
		{
		     items => { type => HASHREF, optional=>0 },
		     uclient => { type => SCALAR, optional=>1, default => 'localhost' },
		     sanclient => { type => SCALAR, optional=>1 },
		     wclient => { type => SCALAR, optional=>1 },
		     win_user => {type => SCALAR|UNDEF, default => undef},
		     win_passwd => {type => SCALAR|UNDEF, default => undef},
		     domain => { type=>SCALAR, optional=>1, default=>'',
                        depends=>['win_user','win_passwd'] },
		     workgroup => { type=>SCALAR|UNDEF, optional=>1, default=>''},
		     purge_tree => { type => SCALAR, optional=>1, default=>1, regex=>qr/^0|1$/ },
		     filer_ip => { type => SCALAR, optional=>1, default => $h->{default_ip} },
		     timeout => { type => SCALAR, optional=>1, default => 60 },   # timeout for say
	    }
	    );

    return eval {

	# set defaults for the options args
	my $qtreepath = "/vol/" . $vol . "/" . $qtreename;
	if (!qtreePresent($conn, $qtreepath )) {
	    # return without failure if qtree did not exist in the first place
	    logcomment("DBG: Qtree '$qtreename' does not exist");
	    return(1);
	}

    # The items hash will can have 5 keys namely : nfs, cifs, share, lun, fcp
    # designed to take items hash that is the return value from populate_qtree
	
    # use delete_tree to remove the nfs data
    if ($args{items}{nfs}) {
        logcomment("Will remove nfs directory '$qtreepath/$args{items}{nfs}'");
        delete_nfs_data($conn, "$qtreepath/$args{items}{nfs}");
        if ($@) {
            croak("Could not remove nfs directory '$args{items}{nfs}' from  qtree $qtreepath : $@\n");
        } else {
            logcomment("Removed nfs directory '$args{items}{nfs}' on path '$qtreepath'");
        }
    }

    # use delete_tree to remove the cifs data
    if ($args{items}{cifs}) {
        logcomment("Will remove cifs directory '$qtreepath/$args{items}{cifs}'");
        die("wclient is not specified. So cannot destroy cifs dir '$qtreepath/$args{items}{cifs}'")
            if (!$args{wclient});
        delete_cifs_data($conn, "$qtreepath/$args{items}{cifs}", 
                                #$args{items}{share},
                                wclient=>$args{wclient},
                                win_user=>$args{win_user},
                                win_passwd=>$args{win_passwd},
                                domain=>$args{domain},
                                workgroup=>$args{workgroup},
                                );
        if ($@) {
            croak("Could not remove cifs directory '$args{items}{cifs}' from  qtree $qtreepath : $@\n");
        } else {
            logcomment("Removed cifs directory '$args{items}{cifs}' on path '$qtreepath'");
        }
    }

    if ($args{items}{share}) {
        logcomment("Will remove cifs share '$args{items}{share}'");
        share_delete($conn, $args{items}{share})
            or croak("$@\n");
        logcomment("Deleted share '$args{items}{share}'");
    }

    # force destroy the lun 
    if ($args{items}{lun}) {
        logcomment("Will remove lun '$qtreepath/$args{items}{lun}'");
        my (undef, undef, $out) = say($conn, "lun destroy -f $qtreepath/$args{items}{lun}");
        if ($out) {
            croak("Unable to remove lun '$qtreepath/$args{items}{lun}' : $out\n") 
        } else {
            logcomment("Destroyed lun '$qtreepath/$args{items}{lun}'");
        }
    }

    if ($args{purge_tree}) {
        # remove the qtree now that it is empty
        my (undef, undef, $out) = say($conn, "priv set -q test; qtree delete $qtreepath");
        croak("Was unable to remove qtree '$qtreepath' : $out") if ($out);
    }

    return(1);

    };
}



## @name ::create_file()
##
## @description Creates a file of specified size with optional hole of
##        specified size on a specified client host.
##
## @example create_file($filepath, $sizestr, %options)
##
## @returns 1 on success, 0 on failure.
## @author Arvind

sub create_file($$%) {
    my ($filepath, $sizestr, %options) = @_;

    my ($client) = $options{client} ?  $options{client} : 'localhost';
    my ($fillbyte) = $options{fillbyte} ?  $options{fillbyte} : '0';
    my ($holeposstr) = $options{holepos} ?  $options{holepos} : 0;
    my ($holesizestr) = $options{holesize} ?  $options{holesize} : 0;
    my ($uid) = $options{uid} ?  $options{uid} : -1;
    my ($gid) = $options{gid} ?  $options{gid} : -1;

    my $t = subtest('{
	my ($filepath, $filesize, $holepos, $holesize, $fillbyte,
		$uid, $gid) = @_;
	my ($count);

	# change euid to the requested uid, gid before file system operations
	if ($uid >= 0 || $gid >= 0) {
		$> = 0;
		$) = $gid if ($gid >= 0);
		$> = $uid if ($uid >= 0);
	}

        open(FILE, ">$filepath")
	    or return addresult(type=>"PASS",
			"opening file $filepath failed: $!");

        my ($fillstr) = $fillbyte x 1024;
        my ($numblocks) = int($holepos / 1024);
        my ($fragment) = $holepos % 1024;

        for ($count = 0; $count < $numblocks; ++$count) {
            print FILE $fillstr
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
        if ($fragment) {
            print FILE $fillbyte x $fragment
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
	truncate(FILE, $holepos + $holesize);
        seek(FILE, $holepos + $holesize, 0)
	    or return addresult(type=>"PASS", "seek to offset "
                                . ($holepos + $holesize) . " failed: $!");
        $numblocks = int(($filesize - $holepos - $holesize) / 1024);
        $fragment = ($filesize - $holepos - $holesize) % 1024;

        for ($count = 0; $count < $numblocks; ++$count) {
            print FILE $fillstr
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
        if ($fragment) {
            print FILE $fillbyte x $fragment
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
        close(FILE)
            or return addresult(type=>"PASS",
			"closing file $filepath failed: $!");

        return addresult(type=>"PASS", "X-PASS");
      }',
      '-runid', "create_file",
      '-client', $client,
      '-ifscript', '-ok',
      '-iffatal', '-ok',
      '--',
      $filepath,
      convert_to_bytes($sizestr),
      convert_to_bytes($holeposstr),
      convert_to_bytes($holesizestr),
      $fillbyte,
      $uid,
      $gid
    );

    my @results = $t->own_results();
    my @msgs = $results[0]->msgs();
    if ($msgs[0] =~ /X-PASS/) {
        return 1;
    } else {
	logStep("create_file failed: $msgs[0]");
        return 0;
    }
}

## @name ::write_file()
##
## @description Creates a file of specified size with optional hole of
##        specified size on a specified client host.
##
## @example write_file($filepath, $sizestr, %options)
##
## @returns 1 on success, 0 on failure.
## @author Arvind

sub write_file($$$%) {
    my ($filepath, $offsetstr, $sizestr, %options) = @_;

    my ($client) = $options{client} ?  $options{client} : 'localhost';
    my ($fillbyte) = $options{fillbyte} ?  $options{fillbyte} : '0';
    my ($uid) = $options{uid} ?  $options{uid} : -1;
    my ($gid) = $options{gid} ?  $options{gid} : -1;

    my $t = subtest('{
	my ($filepath, $offset, $size, $fillbyte, $uid, $gid) = @_;
	my ($count);

	# change euid to the requested uid, gid before file system operations
	if ($uid >= 0 || $gid >= 0) {
		$> = 0;
		$) = $gid if ($gid >= 0);
		$> = $uid if ($uid >= 0);
	}

        open(FILE, "+<$filepath")
	    or return addresult(type=>"PASS",
			"opening file $filepath failed: $!");

        my ($fillstr) = $fillbyte x 1024;

        seek(FILE, $offset, 0)
	    or return addresult(type=>"PASS",
			"seek to offset $offset failed: $!");

        my ($numblocks) = int($size / 1024);
        my ($fragment) = $size % 1024;

        for ($count = 0; $count < $numblocks; ++$count) {
            print FILE $fillstr
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
        if ($fragment) {
            print FILE $fillbyte x $fragment
	        or return addresult(type=>"PASS",
                    "writing to file $filepath failed: $!");
        }
        close(FILE)
            or return addresult(type=>"PASS",
			"closing file $filepath failed: $!");

        return addresult(type=>"PASS", "X-PASS");
      }',
      '-runid', "write_file",
      '-client', $client,
      '-ifscript', '-ok',
      '-iffatal', '-ok',
      '--',
      $filepath,
      convert_to_bytes($offsetstr),
      convert_to_bytes($sizestr),
      $fillbyte,
      $uid,
      $gid
    );

    my @results = $t->own_results();
    my @msgs = $results[0]->msgs();
    if ($msgs[0] =~ /X-PASS/) {
        return 1;
    } else {
	logStep("write_file failed: $msgs[0]");
        return 0;
    }
}

# @name ::remote_cmd()
# @description executes a remote command on a client machine
# @example remote_cmd($command, $client);
## @author Arvind

sub remote_cmd($$) {
    my ($command, $client) = @_;
    my $t = subtest('{
        my ($command) = @_;
        my ($ret) = system("$command");
        return addresult(type=>"PASS", "X-REMOTE-RESULT=$ret");
        }',
        '-runid', "remote_cmd",
        '-client', $client,
        '-ifscript', '-ok',
        '-iffatal', '-ok',
        '--',
        $command
    );

    my @results = $t->own_results();
    my @msgs = $results[0]->msgs();
    if ($msgs[0] =~ /X-REMOTE-RESULT=(\d+)/) {
	return ($1 == 0) ? 1 : 0;
    } else {
        return 0;
    }
}

## @name ::createScratchSir()

## @description
## Creates a unique scratch directory under specified path. This function will
## The created directory, by default, will be with 0777 permission mode.
##
## @returns
## Returns the name of the created scratch directory
## @author Arvind

sub createScratchDir($$;%)
{
    my ($spec, $mntpath, %opt) = @_;
    my ($client) = (defined($ENV{CLIENT}) && $ENV{CLIENT}) ? $ENV{CLIENT} :
						"localhost";

    # 'client' option passed in hash overrides anything
    $client = $opt{client} if (defined($opt{client}));

#    my $t = subtest('{
#
#        my ($mntpath) = @_;

    my $sudo_cmd=_which_sudo();
	my (@dirlist) = `$sudo_cmd /bin/ls $mntpath`;
	croak("could not list dir $mntpath on $client") if ($? != 0);

	# searching a unique directory name to be created under $mntpath
	my ($scratch);
	my ($found) = 0;

	for(my $i = 0; !$found; ++$i) {
	    $found = 1;
	    $scratch = "scratch.$i";
	    foreach my $name (@dirlist) {
		if ($scratch eq $name) {
		    $found = 0;
		    last;
		}
	    }
	}

	croak("could not create $mntpath/$scratch on $client")
	    if system("$sudo_cmd /bin/mkdir $mntpath/$scratch");

       croak("could not chmod $mntpath/$scratch on $client")
	    if system("$sudo_cmd /bin/chmod a+rwx $mntpath/$scratch");

	return "$mntpath/$scratch";

#        return addresult(type=>"PASS", "X-RESULT=$mntpath/$scratch");
#
#      }',
#      '-runid', "createScratchDir",
#      '-client', $client,
#      '-ifscript', '-ok',
#      '-iffatal', '-ok',
#      '--',
#      $mntpath
#    );
#
#    my @results = $t->own_results();
#    my @msgs = $results[0]->msgs();
#    if ($msgs[0] =~ /X-RESULT=(\w+)/) {
#	return $1;
#    } else {
#        return undef;
#    }
}

## @author Arvind
sub backupFiles($@) {
    my ($rootmntdir, @files) = @_;

    return eval {
    	logcomment("backupFiles -- Files to backup:\n@files");
	if (!-d $rootmntdir) {
	    croak("backupFiles: rootmntdir \"$rootmntdir\" not a directory");
	}
	use File::Copy;
	my ($saved_euid) = $>; $> = 0;
        if ($> == 0) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
	        if (! -f $orig_file) {
		    logresult('INFO', "File $orig_file does not exist");
		    system("touch $orig_file") == 0
		        or croak("could not touch (create) $orig_file");
                }
	        copy($orig_file, $backup_file)
		    or croak("could not backup $orig_file");
            }
	    $> = $saved_euid;
	} elsif (-x _which_sudo() ) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
		if (! -e $orig_file) {
		    logresult('INFO', "File $orig_file does not exist");
	            (system(_which_sudo() . " touch $orig_file") == 0)
		        or croak("could not touch (create) $orig_file");
                }
                (system(_which_sudo() . " cp $orig_file $backup_file") == 0)
		    or croak("could not backup $orig_file");
            }
	} else {
            croak("could not backup files @files");
	}
	return 1;
    };
}


## @author Arvind
sub restoreFiles($@) {
    my ($rootmntdir, @files) = @_;

    return eval {
    	logcomment("restoreFiles -- Files to backup:\n@files");
	if (!-d $rootmntdir) {
	    croak("restoreFiles: rootmntdir \"$rootmntdir\" not a directory");
	}
	my ($saved_euid) = $>; $> = 0;
        if ($> == 0) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
	        if (-f "$backup_file") {
		    rename($backup_file, $orig_file)
		        or croak("could not restore $orig_file");
                }
	    }
	    $> = $saved_euid;
	} elsif (-x _which_sudo()) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
                (system(_which_sudo() . " mv $backup_file $orig_file") == 0)
		        or croak("could not restore $orig_file");
            }
	} else {
            croak("could not restore files @files");
	}
	return 1;
    }
}


#############################################################################
#
#      'restoreFilesPreserveBackups' is similar to 'restoreFiles',
#      but 'restoreFiles' is intended to be called once, during 'cleanup,
#      and it deletes the 'backup' temp. files. 'restoreFiles' is not
#      idempotent -- it can be called only once.
#
#     'restoreFilesPreserveBackups' is intended to be called any number
#     of times during a test. It 'preserves' the backup files, so
#     that subsequent calls to either 'restoreFilesPreserveBackups' or
#     'restoreFiles' can be made.
#
#    After the test is finished, a call to 'restoreFiles'  should be
#    made during 'cleanup'. 'restoreFilesPreserveBackups' should not
#    be used for 'cleanup'
#
#    The implementation of 'restoreFilesPreserverBackups' is based
#    on 'backupFiles'.
#
#############################################################################

sub restoreFilesPreserveBackups($@) {
    my ($rootmntdir, @files) = @_;

    return eval {
    	logcomment("restoreFilesPreserveBackups -- Files to restore:\n@files");
	if (!-d $rootmntdir) {
	    croak("restoreFilesPreserveBackups: rootmntdir \"$rootmntdir\" not a directory");
	}
	use File::Copy;
	my ($saved_euid) = $>; $> = 0;
        if ($> == 0) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
	        if (! -e $backup_file) {
		    croak("restorePreserveBackup  Line: ".__LINE__."  could not restore $orig_file");
		}
	        copy($backup_file, $orig_file)
		    or croak("restorePreserveBackup  Line: ".__LINE__."  could not restore $orig_file");
            }
	    $> = $saved_euid;
	} elsif (-x _which_sudo() ) {
	    foreach my $name (@files) {
	        my $orig_file = "$rootmntdir$name";
		my $backup_file = "$rootmntdir$name" . ".backup";
                (system(_which_sudo() . " cp $backup_file $orig_file") == 0)
		    or croak("restorePreserveBackup  Line: ".__LINE__."  could not restore $orig_file");
            }
	} else {
            croak("restorePreserveBackup  Line: ".__LINE__."  could not restore files @files");
	}
	return 1;
    };
}


# say() a shell command to a unix host and die if it returns non-zero
# exit status.

sub exec_shell {
    my ($conn,$sendstr,@rest) = @_;
    say($conn,"$sendstr && echo _''OK_","_OK_",@rest,'-eval_error');
    if ($@) {
	croak "Command \"$sendstr\" failed: $@";
    }
}

# convert file size to number of blocks needed by the file - including
# all indirect blocks and data blocks.

use constant TRAD_INODE_INDIRECT_COUNT => 16;
use constant TRAD_INDIRECT_COUNT => 1024;
use constant FLEX_INODE_INDIRECT_COUNT => 8;
use constant FLEX_INDIRECT_COUNT_64 => 255;
use constant FLEX_INDIRECT_COUNT => 510;
use constant INODE_DATA_SIZE => 64;
use constant BLOCK_SIZE => 4096;

sub convert_size_to_blocks ($;$@) {
	my ($size, $aggrstatus, %args) = validate_mixed(
                @_,
                [ { type => SCALAR },  { type => SCALAR, optional => 1}],
		{
			vol_type => {type => SCALAR, optional => 1,
				default => "flex"}
		}
	    );

        my $size_bytes = convert_to_bytes($size);
        my $flex_indirect_count = FLEX_INDIRECT_COUNT;

        return 0 if ($size_bytes <= INODE_DATA_SIZE);
        
        if ($aggrstatus =~m /64-bit/){
        $flex_indirect_count = FLEX_INDIRECT_COUNT_64;
        }

	my $inode_indirect_count = ($args{vol_type} eq "flex") ?
		FLEX_INODE_INDIRECT_COUNT : TRAD_INODE_INDIRECT_COUNT;
	my $indirect_count = ($args{vol_type} eq "flex") ?
		$flex_indirect_count : TRAD_INDIRECT_COUNT;
 
	my $toplevel = int(($size_bytes + BLOCK_SIZE - 1) / BLOCK_SIZE);
	my $nblocks = $toplevel;

	while ($toplevel > $inode_indirect_count) { # Need another level
		$toplevel = int(($toplevel + $indirect_count - 1)
					/ $indirect_count);
		$nblocks += $toplevel;
	}

	return $nblocks;
}

sub _which_sudo {
    my $sudo_cmd = param('SUDO_PATH', '-mesg','(NATE is expected to have set this)');
    return $sudo_cmd;
}

1;
