# -*-perl-*x
#
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for parsing `file check` output
##
## @description
##
##     file_check - Calls file check on a particular file.
##     file_check_read_output - Parses the output of filecheck
##
## @usage
##     use filecheck;
##
## @author Karthik Ravichandra (karth@netapp.com) 
##
## @status private

package filecheck;

use Carp qw(croak);
use Tharn qw($Log);
use misc qw(cli);
use Params::Validate qw (validate validate_pos SCALAR OBJECT);
use Module::Common qw(validate_mixed);
use Utils qw(make_conn_and_complain);
use SayWrap;
use Filer::Version qw(:releases get_wafl_version);
use FS::Util::Utilities qw(version_check_ic2 version_check_ic3);
use SnapCmds;

use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(file_check file_check_read_output);
 
## @name ::file_check($PATH)
##
## @summary
#i     Calls file check on a particular file
## @description
##     
##
## @arg $fc Required: filer connection object.
## @arg $args Required: arguments to `file check` command.
##
## @return value
##      Returns errcode = 0 - Success
##                      < 0 - Failure
##              result = On Success, path of file, where output of file check is stored.
##                     = On Failure, error message.
##   Usage: 
##               (errcode, $path) = file_check( conn => $fc, args => "/vol/vol1/testfile","test");                      
sub file_check (%)
{
       my %opts=validate(
	@_,
	{
		conn => {type => OBJECT },
		args => {type => SCALAR },
                mode => {type => SCALAR, optional =>1},
	}
   
    );
    
    my $fc = $opts{conn};
    my $filecheck_str = $opts{args};

    #Arguments passed will have a single space in between.
    $filecheck_str =~ s/\s+/ /g;

    #Splitting the filecheck_str to get an array of arguments
    my @arr = split(' ',$filecheck_str);
    my $mode = $opts{mode};

    if (!defined($opts{mode})){
       $mode = "diag";
    }
 
    my ($result, $inum, $vol, $snap_id, $snap, $Snap, $path, $vol_snap, $private, $msg,
        $errcode, $i, $neg_flag, %parser_fc_args, $invalid_snap_flag, $path_of_file);
    
    #
    #Initializing Values
    #
    $msg = "Output will be saved to file ";
    $errcode = -3;
    $i = 0;
    $invalid_snap_flag = 0;
    $path = ""; 
    $vol = "";
    $vol_snap = "";
    $inum = "";
    $snap = "";
    $snap_id = "";
    $Snap = "";
    $private = "";
  
    #Parsing the file check arguments provided by the user. 
    
    for($i = 0;$i < @arr;$i++) {

        #For negative senario.(Avoid overwriting in to the hash data structure created)

        #E.g. if the user is passing -s snap1 -s snap2 which should not be overwritten.

        if (!defined($parser_fc_args{$arr[$i]})) {  
           #Keys to match in the array inorder to assign the corresponding value.           
 
	   if ($arr[$i] =~ m!(/)|-[v|s|i]|(-private)!) { 
              if ($1) { 
                 #Argument starting with a / is nothing but path of a file provided for file check.
                 #Assigning the path provided,to the hash by setting the key as -path. 

                 $parser_fc_args{-path} = $arr[$i]; 
              }
              elsif($2) {  
                 #If -private is provided.hash will have "-private" as a key and value.

                 $parser_fc_args{$arr[$i]} = "-private";
              }
              else {
                 #If the arguments are having -v or -s or -i then the values which
                 #are next to this arguments in the array will have the corresponding
                 #values(E.g. -v <volname>) so,assign that to a hash  as a key ,value pair.
 
                 #Since -v and -vol are treated same in file check, If the argument 
                 #contains -vol, then it is handled by assigning the value next to -vol in 
                 #the array as a value for the key -v in the hash.

		 if ($arr[$i] =~ m!-vol!) {
                    $parser_fc_args{-v} = $arr[$i+1];
                    $i++;
		 } 
                 else {
                    $parser_fc_args{$arr[$i]}= $arr[$i+1]; 
	            $i++;                                 
                 }
              }                                
           }
            else {
              #If the array is having any other keys other than -v etc.,set the flag.  

              $neg_flag =1;
           }
        }
        else {
           #If 2 keys are same set the negative flag.

           $neg_flag =1;
        }
    }
    
    #If the negative flag($neg_flag) is set to one means,the arguments passed by the user 
    #is having some wrong data.So pass it as it is to file check and return the appropriate
    #errorcode to the caller.
 
    if (!defined($neg_flag)) {
    
       #Producing the strings inorder to pass to a file check command.

       if (defined($parser_fc_args{-S})) {
          #Snapshot name. The pattern will be -S <snapshot_name>

          $Snap = "-S $parser_fc_args{-S}";
       }

       if (defined($parser_fc_args{-i})) {
          #Inode number of the file.The pattern will be -i <inode number>
          $inum = "-i $parser_fc_args{-i}";
       }

       if (defined($parser_fc_args{-v})) {
          #Volume name.The pattern will be -v <volname>
          $vol = "-v $parser_fc_args{-v}";
       }

       if (defined($parser_fc_args{-path})) {
          #Path of the file.The pattern will be /vol/<vol>/<file>

          $path = "$parser_fc_args{-path}";
       }

       if (defined($parser_fc_args{-private})) {
          #"-private" option.The pattern is -private.

          $private = "$parser_fc_args{-private}";
       }
       
       if (defined($parser_fc_args{-s})) {
          #Snapshot id of the snapshot.The pattern will be -s <snapid>

          if ($parser_fc_args{-s} =~ m/^[0-9]+$/) {
             $snap = "-s $parser_fc_args{-s}";
          }
          else {
             #Incase user provided valid snapname convert it in to snapid.
             #For invalid snapnames "-s <invalidsnapname>" just pass it as it is.

             #Getting the volume name to check provided snapshot exists or not.

             if (defined($vol)) {
                $vol_snap = $parser_fc_args{-v};
             }
             if (defined($path) && $parser_fc_args{-path} =~ m!/vol/(.*?)/(.*)!) {
                $vol_snap = $1;
             }


             #Checking whether the snapshot provided exists or not.
             if (snapshot_exists($fc, $vol_snap, $parser_fc_args{-s}) == 1) {
                #If the snapshot exists and path of the file is not provided/defined, 
                #find the snapid.Incase of IC.2 filecheck syntax if -s <snapname> <path>
                #will through an error.Inorder to solve 'generate snapshot path of the file'
                #and pass to file check as an input.
  
                if ($parser_fc_args{-s} =~ m/\w+/ && !defined($parser_fc_args{-path})) {
                    @snap_status = snapshot_status($fc,$vol_snap);
                    foreach (@snap_status) {
                       if ($$_{name} eq $parser_fc_args{-s}) {
                          $snap_id = $$_{snapid};
                          $snap = "-s $snap_id";
                       }
                    }
                }

                else {
                   if ((defined($parser_fc_args{-path})) && $parser_fc_args{-s} =~ m/\w+/) {
                      #If the path provided is a snapshot file path, then dont pass the snapshot 
                      #name to file check to avoid error in IC.2.

                    if ($path =~ m!/vol/(.*?)/(.*)!) {
		       $vol_snap = $1;
                       $path_of_file = $2;
		    }
                    if ($path_of_file =~ m/.snapshot/) {
                       $snap="";
                    }
                    else {
                       $path = "/vol/" . $vol_snap . "/.snapshot/" . $parser_fc_args{-s} . "/" . $path_of_file;
                       $snap = ""; 
                    }
                   }
                }
             }
             else {
                #Snapshot name provided is not exists on the filer.So pass it as it is, and file check 
                #will through the error.

                $invalid_snap_flag = 1;
                $snap = "-s $parser_fc_args{-s}";
             }
          }
       }
       #Checking the version and generating the input string /argument sting to the file check command.

       my $fh = $fc->{'hostp'};
       my $version= get_wafl_version($fh);
       my $ver = version_check_ic2(conn => $fc);

       if (($version >= WV_BOILERMAKER_MIN && $version <= WV_BOILERMAKER_MAX) ||
          ($version >= WV_ROLLINGROCK_MIN && $version <= WV_ROLLINGROCK_MAX)) { 
          #The version is BR.x or RR

          if(version_check_ic3(conn => $fc)) {
             #The version is BR.1.
             $filecheck_str="$path $vol $inum $snap $Snap $private";
          }
          elsif ($Snap eq "") {
             #The version is not BR.1 , hence It shouldnt have -S in file check command.
             $filecheck_str = "$path $vol $inum $snap $private";
          }
          else {
             #Croak if the user passes -S for file check. 
             croak("BR/RR is not having -S option in file check");
          }
       } 
       elsif (($version >= WV_IRONCITY_MIN) && ($version <= WV_IRONCITY_MAX)) {
          if ($ver == 1) {
             #The version is IC.2
             if ($invalid_snap_flag == 1) {
                $snap = "";
                $Snap = "-S $parser_fc_args{-s}";
             }
             $filecheck_str="$path $vol $inum $snap $Snap";
          }
          else {
             #The version is IC.1
             if ($Snap eq "") {
                $filecheck_str = "$snap $path $vol $inum";
             }
             else {
                croak("IC.1 is not having -S option in file check");
             }
          }
       }
      
    }#end of  if(!defined($neg_flag)) {
  
    my $output = sayEOF($fc,"priv set -q $mode; file check ". $filecheck_str);
    my @lines = split /\n/, $output;

    if($output eq "") {
        $errcode = -3;
        $result = "Bad output string: ".$output;
        return ($errcode,$result);
    }

    foreach (@lines) {

       if($_ =~ m/^(File|Snapshot|Volume|Snapid)\s+(\S+)\s+doesn't\s+exist/i){
          $errcode = -1;
          $result = $_;
          last;
       }

       if($_ =~ /Usage:/) {
          $errcode = -2;
          $result = $_;
          last; 
       }

       if($_ =~ /$msg(\'.*\')/){
           $errcode = 0;
           $result = $1;
           last;
       }
    }

    return ($errcode,$result);
}


## @name ::file_check_read_ouput($fc, $PATH)
##
## @summary
#i     Parses output of `file check`
## @description
##     Returns the parsed output of file check log.

## ##
## @arg $fc Required: filer connection object.
## @arg $path Required: path of the log, returned when file_check() is called
##
## @return value
##      Returns  $count = Number of corrupted blocks
##               $total = Number of blocks scanned
##               @vbnlist = Reference to Array of vbns corrupted.
##               @snapshot_list = Reference to Array of snapshots that are possibly corrupt.
##    Usage:
##               my ($count, $total,@vvbnlist, @snapshot_list) = file_check_read_output(conn => $fc, path => $file_check_path);
sub file_check_read_output (%)
{
    my %opts = validate(
	@_,
	{
		conn => {type => OBJECT },
		path => {type => SCALAR },
	}
    );

    my $fc = $opts{conn};
	my $path = $opts{path};

    my $lost_blocks_msg = "Lost blocks: ";
    my $snapshot_msg = "The following snapshots contain any of the listed blocks:";
    my $snapshots_corrupt;

    my $count=0;
    my $total =0;
    my @vvbn_list;
    my @snapshot_list;
    my ($level, $errno, @fbns);
    $level = 0;
    $errno = 0;

    my $output = sayEOF($fc, "rdfile ".$path);
    my @lines = split /\n/, $output;

    foreach (@lines) {
      # Store the corrupt blocks in a list.
      if($_ =~ /$lost_blocks_msg(.*)/) {
        my @list = split / /, $1;

        foreach (@list) {
          push @vvbn_list, $_;
        }
      } elsif($_ =~ m/Lost\sVVBN:\s(\d+),/) {
        # New format in IC.3
        # Lost VVBN: 4150, Disk: 0a.18, Volume: testvol_flex, Inode: 100

        push @vvbn_list, $1;
      }	 

      if($_ =~ m/.*Level:\s+(\d+),\s+Fbn:\s+(\d+),(.*)Error:\s+(\d+)/i) {
        # New format in BR.1 and RR.0 when we enabled compression on the volume.
        # Lost VVBN: 1941867, Disk: 0a.20, Volume: vol1, Inode: 96, Level: 0, Fbn: 0, Error: 117 <==BR.1
        # Lost VVBN: 1941867, Disk: 0a.20, Volume: vol1, Inode: 96, Level: 0, Fbn: 0, References: 0, Error: 117 <==RR.0
        # Level and errno are same for all the block in CG since level and errno are
        # assigned to scalar variable.
        $level = $1;
        push @fbns, $2;
        $errno = $4;
      }

      # Store the number of lost blocks
      if($_ =~ m/Total\s(\d+)\slost\sblock/) {
        $count = $1;
      }

      # Determine if there are any corrupt blocks in the snapshots.
      if($_ =~ /$snapshot_msg/) {
        $snapshots_corrupt = 1;
      }

      # If there are snapshots, then add them to a list.
      if(($_ =~ /\d+:\d+:\d+/) && ($snapshots_corrupt == 1)) {
        my @line = split / /, $_;
        push @snapshot_list, @line[0];
      }

      # Store the total number of blocks scanned
      if($_ =~ m/Total\s(\d+)\sblock\(s\)\sscanned/) {
        $total = $1;
      }

    }
    return ($count, $total, \@vvbn_list, \@snapshot_list, $level, $errno, \@fbns);
}

1;
