## this is basically -*- perl -*-
## vi: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
##------------------------------------------------------------------
## @summary    FIT client check wrapper
## Script that is a FIT wrapper for client_check to run in Push Button
## type framework. This script would check if a required number
## of clients pass client_check based on the test that is running.
## In cases where there are 35 clients defined and the test requires
## only 30 clients and if two of those clients fail client check the
## remaining test can still run. This script should run in place of
## client_check to validate the clients and reset the NATE CLIENT param
## with the ones that pass client check
##
## This lib was originally a thpl. It was ported into a lib, so that return 
## values could be passed directly to FCJ, instead of having to mess with 
## adjusting environment variables. 
##
## MIN_WIN_COUNT and MIN_UNIX_COUNT are Mutually exclusive with MAX_FAIL_ALLOW
## percent. If MAX_FAIL_ALLOW is passed it takes priority over the WIN/UNIX.
##
## @author      santhos3@netapp.com
##
###############################################################################

package fit_client_validate;

use strict;
use warnings;
use Options;
use POSIX qw(ceil floor);
use Tharn qw($Log subtest host);
use NATE::Result;
use Data::Dumper;
use NATE::ParamSet;
use Params::Validate qw(validate SCALAR BOOLEAN);
use NACL::C::Client;

# We need this, because we are still grabbing "CLIENT" just in case 
# Params::Common stuck stuff in there
my $Options = new Options();
# Define parameters
$Options->parameters(definitions =>
    [
    {
        name            => "PROC_LIMIT",
        description     => "Set the maximum number of sub-processes to be launched simultaneously.  Defaults to 50",
        spec            => "s",
        default         => "50",
    },
]);
$Options->process();
my $OptsHash = $Options->get_hash();

sub _arrange_client_list { 
    $Log->enter();
    my %dupcheck;
    my @clients;

    # NACL returns lots of client data in a hash. We only want the 
    # list of client names.
    foreach my $clienthash (NACL::C::Client->find(allow_empty => 1)) { 
        push(@clients, $clienthash->{client});
        $dupcheck{$clienthash->{client}} = 1;
    } 
    $Log->debug("These clients were obtained from NACL::C::Client->find()" . Dumper(\%dupcheck));

    # Get rid of the dups that might come in from sourcing both possibilities
    $Log->debug("These clients were obtained from CLIENTFILE:");
    foreach my $client (@{$OptsHash->{CLIENT}}) { 
        next if ($dupcheck{$client});


        $dupcheck{$client} = 1;
        $Log->debug($client);
        push (@clients, $client);
    }

    $Log->exit();
    return (\@clients);
}

    
sub client_validate { 
   my (%opts) =
     validate(@_,
         {max_fail_allow => {type => SCALAR, default => undef},
          skip_client_validate => {type => BOOLEAN, default => 0},
          max_proc_limit => {type => SCALAR, default => $OptsHash->{PROC_LIMIT}},
          min_win_count => {type => SCALAR, default => 0},
          min_unix_count => {type => SCALAR, default => 0},},);

    my $clients = &_arrange_client_list();
    my %PassClients;
    my %FailClients;
    my @WinClients;
    my @UnixClients;
    my $MinNfs;
    my $MinCifs;
    my $FailPercent;
    my $PassPercent;
    my $Threshold;

    # Params::Common will put everything in CLIENT and set CLIENTS to undef
    my $paramset = NATE::ParamSet->new(global => 1);

    # bail out now if we aren't doing any checking
    if($opts{'skip_client_validate'}) { 
        $Log->comment("Skipping client validation due to skip_client_validate");
        return $clients;
    }


    #################### Subtest client_check ################################
    
    # Split clients into an array of arrays so we don't launch more than max_proc_limit sub-processes at the same time
    my $splitsize = ceil(scalar(@{$clients}) / $opts{'max_proc_limit'});
    my $splitClients = &_evenlySplitArray($splitsize, $clients);
    
    my $count = 1;
    foreach my $segment (@{$splitClients}) {
        $Log->comment("Checking the health of clients ".$count."-".($count + scalar(@{$segment}) - 1)." of ".scalar(@{$clients}));
        
        my @Subtests;
        foreach my $client (@{$segment}) {
            my $subtest;
            eval {
        	$subtest = subtest(
        			   "tools/client_check.thpl", 
        			   "CLIENTFILE="."",
        			   "CLIENT=".$client,
        			   "client_check_CLIENT=".$client,
        			   "CLEANUP_MOUNTS=1",
        			   "REBOOT=reactively",
        			   "REBOOT_METHOD=rpower",
        			   -runid => 'client_validate/' . $client,
        			   -onexit => sub {},
        			   '-bg',
        			   );
            };
            if ($@) {
                $Log->error("Encountered unexpected error while trying to subtest ".
                  "client_check.thpl");
                $Log->comment("Error:" . $@);
                die();
            }
    
            push(@Subtests,$subtest);
        }
        
        Subtest::wait_finish(Subtests => \@Subtests);
        
        for (my $ii=0; $ii<scalar(@Subtests); $ii++) {
            my $idx = $ii + ($count - 1);
            my $st = $Subtests[$ii];
            my @result = $st->results();
       
            my $worst = NATE::Result::worst(@result);
            my $type = $worst->type();
    
            if (not $type =~ /^PASS|WARN$/) {
                $FailClients{$$clients[$idx]} = $worst->{'messages'}[0];
            } else {
                $PassClients{$$clients[$idx]} = $worst->{'messages'}[0];
            }
        }
        
        $count += scalar(@{$segment});
    }
    $Log->debug("pass results #: ".Dumper(\%PassClients));
    $Log->debug("fail results #: ".Dumper(\%FailClients));
    
    my @newclients = keys (%PassClients);
    my @failclients = keys (%FailClients);
    
    #If there were no clients passed,ie if number of clients that passed and failed client check is 0,we warn and continue
    if ((scalar(@newclients) == 0) && (scalar(@failclients) == 0)) {
       $Log->warn("No clients were passed to fit client validate. Skipping fit client validate");
    }
    
   #If clients were specified but they all failed we should die. 
   elsif ((scalar(@newclients) == 0) && (scalar(@failclients) > 0)) {
        $Log->error("ALL Clients failed client check");
        die();
    }
    $Log->comment('CLIENTS TOTAL#: '.scalar(@$clients));
    $Log->comment('CLIENTS PASSED# : '.scalar(@newclients));

    ####################### Validate Count ###################################
    if ($opts{'max_fail_allow'}) {
        ### Compute percent ####
        my $fail = scalar(keys %FailClients);
        my $pass = scalar(keys %PassClients);
        my $total = scalar(@$clients);
    
        $Log->comment("Fail : ".$fail);
        $Log->comment("Pass : ".$pass);
        $Log->comment("Total: ".$total);
    
        $FailPercent = ceil(($fail/$total) * 100);
        $PassPercent = int(($pass/$total) * 100);
        $Threshold = int($opts{'max_fail_allow'});
    
        if ($FailPercent > $Threshold) {
    	$Log->error("Maximum allowed client failure rate is: ".
    		    $Threshold."% and the failure is: ".$FailPercent."%");
    	die();
        }
        $Log->comment("Pass Percent: ".$PassPercent);
    } else {
        $MinCifs = $opts{'min_win_count'};
        $MinNfs = $opts{'min_unix_count'};
    
        foreach my $client (@newclients) {
    	my $h = host($client);
    	my $hosttype = $h->hosttype();
    	if ($hosttype =~ /unix/i) {
    	    push(@UnixClients,$client);
    	} elsif ($hosttype =~ /windows/i) {
    	    push(@WinClients,$client);
    	}
        }
    
        if (scalar(@UnixClients) < $MinNfs) {
    	$Log->error("The number of nfs clients required to run ".
    		    "the test is: ".$MinNfs.
    		    "Clients provided: ".scalar(@UnixClients));
    	die();
        }
        $Log->comment("NFS clients passed# : ".scalar(@UnixClients));
    
        if (scalar(@WinClients) < $MinCifs) {
    	$Log->error("The number of cifs clients required to run ".
    		    "the test is: ".$MinCifs. 
    		    "Clients provided: ".scalar(@WinClients));
    	die();
        }
        $Log->comment("CIFS clients passed# : ".scalar(@WinClients));
    }
    
    #################### RESET ENV USING PARAMSET ############################

    my $clients_str = join(',',@newclients);
    my $client_str = $newclients[0];
    $Log->comment("client str:".$client_str);
    $Log->comment("clients str:".$clients_str);
    
    return \@newclients;
        
}

# Takes a number(n) + arrayref and splits the array into n arrays
# Returns a ref to array of arrays
sub _evenlySplitArray {
    my ($n, $array) = @_;

    my @parts;
    my $i = 0;
    foreach my $elem (@$array) {
        push @{ $parts[$i++ % $n] }, $elem;
    };
    
    return \@parts;
}

1;
            
