#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Install module for gx_boot

package gx_install;

use strict;
use vars ('@ISA','@EXPORT');

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
            install
            );

# Make sure we have NATE >= 3.0
use NATE 3.000;

use Server;
use Options;
use Data::Dumper;
use Hostrec;
use boot;
use TestUtils qw(get_site_info);
use gx_utils;
use NATE::BaseException qw(:try);
use Tharn qw(param);
use TracePkg;


=head1 NAME

gx_install

=head1 DESCRIPTION

Provides functionality to install an image and optionally reboot a VSIM
or filer.

=head1 ATTRIBUTES

=head2 none

=cut

=head1 METHODS

=head2 install

    gx_install::install(
        @_,
        myParams     => $Params,
        bootMenuType => \%BootMenuType,
        Server       => $Server,
        prompt_fw    => $prompt_fw,
        Trace_level  => $Trace_level,
        Result       => $Result,
    );

  (Class method)
  Installs an image on a filer or VSIM based on either a parameter,
  or from the site_info.dat file.

=over

=item Options

=over

=item C<< myParams => $Params >>

 (Required) 
 Parameters related to hostrec and promoted install paths.

=item C<< Options => $Options >>

 (Required) 
 Contains parameters, compatibility definitions, test cases, etc.

=item C<< bootMenuType => \%BootMenuType >>

 (Required) 
 Specifies the type of boot menu.

=item C<< Server => $Server >>

 (Required) 
 The Server object which will be used to transit to, as well as
 interact with, via its console.

=item C<< prompt_fw => $prompt_fw >>

 (Required) 
 Prompt forwarding string(s).  

=item C<< Trace_level => $Trace_level >>

 (Required) 
 Set trace level to fatal, warn, etc. 

=item C<< Result => $Result >>

 (Required) 
 Set results as failures or warnings. 

=back

=back

=cut

sub install {
    #Create Trace Object#
    my $Trace = new TracePkg();

    $Trace->entry();
    my %params = (
		  node => undef,
		  fake => undef,
		  reboot_after => undef,
		  promoted => 0,
		  force => 0,
		  @_,
		  );

    ##Site Specific Data Structure##
    my %Siteinfo = get_site_info();

    #Valid 32bit HW
    my @Supported32 = qw(FAS3050 FAS2040 N5500);

    #Valid 64bit HW
    my @Supported64 = qw(SA300 SA310 FAS3040 FAS3070 FAS3140 FAS3170 FAS6030 FAS6040 FAS6070 FAS6080 FAS6210 FAS6240 FAS6280 FAS3260 FAS3240 FAS3270 FAS3210 FAS3160 FAS2220 N5300 N5600 N6040 N6060 N6070 N6210 N6240 N6260 N7550T N7600 N7700 N7750T N7800 N7900 N7950T FAS2240-2 FAS2240-4 N3220 N3240 N3150);

    my $Params = $params{myParams};
    
    my $local_trace_level = $params{Trace_level};

    my $image = thirty_two_sixty_four(node => $params{node},
                                      Options => $params{Options});
    $Trace->comment(msg => "IMAGE to be installed:\n'$image'");

    #Allow for http offset with http root is not "/";
    if ($Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->{install_proto} eq "http") {
        my $http_root_dir = $Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->{http_root_dir};
        if ($http_root_dir ne ""){
            if ($image =~ m/$http_root_dir/){
                $image = $';
            }
        }
    }

    if ($params{promoted}) { #$image is used unless this is a promoted_install
        #if 32bit hw then load 32bit code
        if (grep(/$Params->{HOSTRECS}{$params{node}}{MODEL}/, @Supported32)) {

	    if ($Params->{PROMOTED_INSTALL_32PATH}) {
		$image = "$Params->{PROMOTED_INSTALL_32PATH}\/"."image.tgz";
                if ($image !~ /debug/i) {
                    $Trace->warn(msg => "PROMOTED_INSTALL_32PATH does not ".
                        " have debug in the path, please make sure it is a ".
                        " debug build ");
                }
	    } else {
                if (!$Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->
                        {promoted_install_32path}) {
                    $Trace->fatal(msg => "No promoted_install_32path defined ".
                        "for the site. Unable to provide default. ");

                }

		$image = "$Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->{promoted_install_32path}\/".
		"image.tgz";
	    }
        #else, must be 64bit hw
        } else {
	    if ($Params->{PROMOTED_INSTALL_64PATH}) {
        	$image = "$Params->{PROMOTED_INSTALL_64PATH}\/"."image.tgz";
                if ($image !~ /debug/i) {
                    $Trace->warn(msg => "PROMOTED_INSTALL_32PATH does not ".
                        " have debug in the path, please make sure it is a ".
                        " debug build ");
                }
	    } else {
	        if (!$Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->
                        {promoted_install_64path}) {
                    $Trace->trace(msg => "Siteinfo Data Structure:\n".
                          Dumper($Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}));
                    $Trace->fatal(msg => "No promoted_install_64path defined ".
                        "for the site. Unable to provide default. ");

                }
		$image = "$Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}->{promoted_install_64path}\/".
		"image.tgz";
	    }
        }

    }
    my $install_url;
    if ($image =~ /^(ftp|http):\/\//) {
        $install_url = $image;
    } else {
        $install_url = "$Siteinfo{($Params->{HOSTRECS}{$params{node}}{SITE})}->{install_proto}:\/\/".
                       "$Siteinfo{($Params->{HOSTRECS}{$params{node}}{SITE})}->{install_user}:".
                       "$Siteinfo{($Params->{HOSTRECS}{$params{node}}{SITE})}->{install_pass}\@".
                       "$Siteinfo{($Params->{HOSTRECS}{$params{node}}{SITE})}->{install_server}\/".
                       "$image";  #build up the install URL from %Siteinfo and IMAGE
    }
    my $reboot_after;             #must set $params{reboot_after} to Y or N thus need a local var

    $Trace->debug(msg => "Params Data Structure:\n".
                          Dumper($Params->{HOSTRECS}{$params{node}}));
    $Trace->trace(msg => "Siteinfo Data Structure:\n".
                          Dumper($Siteinfo{$Params->{HOSTRECS}{$params{node}}{SITE}}));
    $Trace->trace(msg => "install() params:\n" . Dumper(%params));

    _transit_to(state => $params{Server}->IBM, #make sure that we are in the boot menu
                no_connections    => 1,
                state_cr_interval => 60,
                state_newlines    => 5,
                boot_ontap => 0,
                node => $params{node},
                clustered => $params{clustered},
		bootMenuType => $params{bootMenuType},
               );

    if ($params{reboot_after}) {  #transfer 0/1 to y/n
    	$reboot_after = "y"
    } else {
    	$reboot_after = "n"
    }

    # The pre-requisite for calling this routine is that we are already at the
    # boot menu
    my $numeric_boot   = is_numeric_boot_menu( node => $params{node},
                                               bootMenuType => $params{bootMenuType},
                                             );
    my $converged_boot = is_converged_boot_menu( node => $params{node},
                                                 bootMenuType => $params{bootMenuType},
					       );

    my $cmd = $numeric_boot ? "7" : "install";
    if ($params{fake}) {
        if ( $converged_boot || !$numeric_boot ) {
            $cmd = "fake_install";
        }
    }

    # The check here is a little different than other places where we check the converged
    # boot menu flag.  Most other places we need to switch from "normal" or "maint" to
    # option numbers (like 1 or 5), but for the converged boot menu, we have fake_install
    # but 7 mode non-converged does not.

    if ($params{clustered} || $numeric_boot) {  #10mode  or converged boot menu install
        #non-converged 10mode bootmenu:setup and init or fake_init must be run prior to running this command!
        #converged bootmenu:install (opt 7) does not require setup and init to be run first.
        # BURT 471705: GXSMOKE: gxboot not handling bad ftp download resulting in smokes failure
        # Checking package contents... failed.\r\n
        # Added "Checking package contents... failed" to install_prompts with action as
        # FATAL_PROMPT
        my %install_prompts =
           ("Do you wish to continue"         => "y",
            "Do you want to continue"         => "y",
            "want to continue"                => "y",
            "Do you want to restore the backup configuration" => "n",
            "use for the download"            => "$Params->{HOSTRECS}{$params{node}}{MGMT_PORT}",
            "the network port you want to use for the download"
                                              => "$Params->{HOSTRECS}{$params{node}}{MGMT_PORT}",
            "Enter IP address:"               => "$Params->{HOSTRECS}{$params{node}}{MGMT_IP}",
            "Enter the IP address"            => "$Params->{HOSTRECS}{$params{node}}{MGMT_IP}",
            "Enter netmask"                   => "$Params->{HOSTRECS}{$params{node}}{MGMT_MASK}",
            "Enter the netmask"               => "$Params->{HOSTRECS}{$params{node}}{MGMT_MASK}",
            "IP address of default gateway"   => "$Params->{HOSTRECS}{$params{node}}{MGMT_GATEWAY}",
            "Reboot now"                      => "y",
            "What is the URL for the package" => $install_url,
             "Enter URL to install image"     => "$install_url", #old prompt
             "URL for the package"            => "$install_url",
            "to continue booting"             => 'FINAL_PROMPT',
            "either reboot or retry install"  => 'FINAL_PROMPT',
            "Do you want to set the newly"    => "y",
            "The node must be rebooted"       => 'FINAL_PROMPT',
             "reboot now"                     => 'FINAL_PROMPT',
            "Installation completed"          => 'FINAL_PROMPT', #for Tricky
             "Software install SUCCESSFUL"    => "FINAL_PROMPT",
            "$params{prompt_fw}"                      => 'boot_ontap',
            "Starting AUTOBOOT press"         => "\cc",
            "any other key for command prompt" => "\cc",
            "Invalid package URL"             => 'FATAL_PROMPT',
            "Checking package contents... failed" => 'FATAL_PROMPT',
            "Install Failed"                  => 'FATAL_PROMPT',
            "Couldn't resolve host"           => 'FATAL_PROMPT',
            "Unrecognized archive format"     => 'FATAL_PROMPT',
            "access denied"                   => 'FATAL_PROMPT',
            "Failed to open file"             => 'FATAL_PROMPT',
            "out of swap space"               => 'FATAL_PROMPT',
            "Selection not recognized"        => 'FATAL_PROMPT',
            "You must use 'setup' to"         => 'FATAL_PROMPT',
             "Failed to retrieve install image" => "FATAL_PROMPT",
             "etry "                           => "y",
             'etry\? '                          => "y",
             "if any"                         => "\n",
             "subsequent reboots"             => "y",

           );

        #do fake (one CF location), real (both CF locations), or force install
        if ($params{force} && !$numeric_boot) {
            my $current_attempt = 1;
            my $max_attempts = 3;
            while ($current_attempt <= $max_attempts) {
                eval {
                    $params{Server}->console(cmd=>"force_install",
                                     additional_prompts_timeout => "1980",
                                     noprompt => 1,
                                     additional_prompts => {%install_prompts},);
                };
                if ($@) {
                    $Trace->warn(msg => "BootMenu:force_install is not available in the build you are running\n".
                                        "A Non-forced install will be tried before giving up."
                                 );
                } else {
                    $params{force} = 2; #the force_install worked so set a flag
                    last;
                }
            }
        }
        if ($params{force} != 2) {
            my $current_attempt = 1;
            my $max_attempts = 3;
            while ($current_attempt <= $max_attempts) {
                eval {
                    $params{Server}->console(cmd=>$cmd,
                                     additional_prompts_timeout => "3960",
                                     noprompt => 1,
                                     additional_prompts => {%install_prompts},);
                };
                if ($@) {
                    if ($@ =~ /failed because: Install Failed/) {
                        $Trace->warn(msg => "Attempt $current_attempt failed to install the image.");
                        ++$current_attempt;
                        if ($current_attempt > $max_attempts) {
                        	my $msg = "We tried $max_attempts time to install the image.  Giving up!";
                            $Trace->$local_trace_level(msg=>$msg, result=>$params{Result});
                            job_retry(
                                Params            => $Params,
                                max_fail_body_add => $msg
                            );
                        } else {
                            $Trace->comment(msg => "Let's try this again!");
                            next;
                        }
                    } else {
                        my $msg = "Install *FAILED*\nHere is more info on why:\n$@";
                        $Trace->$local_trace_level(msg=>$msg, result=>$params{Result});
                        job_retry(
                            Params            => $Params,
                            max_fail_body_add => $msg
                        );
                    }
                } else {
                    last;
                }
            }
        }
        $params{Server}->console(cmd=>"$reboot_after",
                         additional_prompts_timeout => "300",
                         noprompt => 1,
                         additional_prompts => {
                                 "Do you want to reboot now"  => "$reboot_after",
                                 "reboot now"                 => "$reboot_after",
                                 "Starting AUTOBOOT press"    => "\cc",
                                 "any other key for command prompt" => "\cc",
                                 "Software install SUCCESSFUL"=> "FINAL_PROMPT",
                                 "Please choose one of the "  => "FINAL_PROMPT",
                                 "reboot or retry install"    => "FINAL_PROMPT",
                                                   $params{prompt_fw} => "FINAL_PROMPT",
                                               },
                       );
    } else { #7mode install, there is no fake_install so if task is fake_install or install do the same thing
		my %install_prompts =
			("want to continue"	=> "y",
			 "use for the download" => "$Params->{HOSTRECS}{$params{node}}{MGMT_PORT}",
			 "Enter IP address:" => "$Params->{HOSTRECS}{$params{node}}{MGMT_IP}",
			 "Enter the IP address" => "$Params->{HOSTRECS}{$params{node}}{MGMT_IP}",
			 "Enter netmask" => "$Params->{HOSTRECS}{$params{node}}{MGMT_MASK}",
			 "Enter the netmask" => "$Params->{HOSTRECS}{$params{node}}{MGMT_MASK}",
			 "IP address of default gateway" => "$Params->{HOSTRECS}{$params{node}}{MGMT_GATEWAY}",
			 "Enter URL to install image" => "$install_url", #old prompt
			 "URL for the package" => "$install_url",
			 "if any" => "\n",
                         "Do you want to restore the backup configuration" => "n",
			 "Software install SUCCESSFUL" => "FINAL_PROMPT",
			 "Failed to retrieve install image" => "FATAL_PROMPT",
			 "subsequent reboots" => "y",
			 "Reboot now" => "y",
			 "Retry" => "y",
			 "etry" => "y",
			 "Retr" => "y",
			 "$params{prompt_fw}" => "boot_ontap",
			 "reboot now" => "FINAL_PROMPT",
			);
        if ($params{force}) {
            my $attempt = 1;
            my $max_attempts = 3;
            while ($attempt <= $max_attempts) {
                eval {
                    $params{Server}->console(cmd=>"force_install",
                                     additional_prompts_timeout => "900",
                                     noprompt => 1,
                                     additional_prompts => {%install_prompts},);
                };
                if ($@) {
                    if ($@ =~ /failed because: Failed to retrieve install image/) {
                        $Trace->warn(msg => "BootMenu:force_install failed on attempt '$attempt' out of '$max_attempts'!");
                        ++$attempt;
                    } else {
                        $Trace->warn(msg => "BootMenu:force_install is not available in the build you are running\n".
                                            "A Non-forced install will be tried before giving up."
                                     );
                        last;
                    }
                } else {
                    $params{force} = 2; #the force_install worked so set a flag
                    last;
                }
            } #end while
            if ($attempt > $max_attempts) {
            	my $msg = "BootMenu:force_install failed after multiple attempts!:\n'$@'";
                $Trace->$local_trace_level(msg=>$msg, result=>$params{Result});
                job_retry(
                    Params            => $Params,
                    max_fail_body_add => $msg
                );
            }
        }
        if ($params{force} != 2) {
            my $attempt = 1;
            my $max_attempts = 3;
            while ($attempt <= $max_attempts) {
                eval {
                    $params{Server}->console(cmd=>$cmd,
                                     additional_prompts_timeout => "960",
                                     noprompt => 1,
                                     additional_prompts => {%install_prompts},);
                };

                if ($@) {
                    $Trace->warn(msg => "BootMenu:install failed on attempt '$attempt' out of '$max_attempts'!");
                    ++$attempt;
                } else {
                    last;
                }
            } #end while
            if ($attempt > $max_attempts) {
                my $msg = "BootMenu:install failed after multiple attempts!:\n'$@'";
                $Trace->$local_trace_level(msg=>$msg, result=>$params{Result});
                job_retry(
                    Params            => $Params,
                    max_fail_body_add => $msg
                );
            }
        }
        $params{Server}->console(cmd=>"$reboot_after",
                         additional_prompts_timeout => "300",
                         noprompt => 1,
                         additional_prompts => {
                                 "Do you want to reboot now"  => "$reboot_after",
                                 "Starting AUTOBOOT press"    => "\cc",
                                 "any other key for command prompt" => "\cc",
                                 "Software install SUCCESSFUL"=> "FINAL_PROMPT",
                                 "Please choose one of the "  => "FINAL_PROMPT",
                                                   $params{prompt_fw} => "FINAL_PROMPT",
                                               },
                       );
    }
    if ($reboot_after eq "y") {
        _transit_to(state  => $params{Server}->CFE,
                    no_connections    => 1,
                    state_try_passive => 1,
                    node => $params{node},
                    clustered => $params{clustered},
		    bootMenuType => $params{bootMenuType},
                   );
    } elsif ($reboot_after eq "n") {
        _transit_to(state  => $params{Server}->IBM,
                    no_connections    => 1,
                    state_try_passive => 1,
                    node => $params{node},
                    clustered => $params{clustered},
		    bootMenuType => $params{bootMenuType},
                   );
    }
    $Trace->exit();
}

1;
