# -*- perl -*-
# $Id: //depot/prod/test/pvt_sn_warm/lib/gx_utils.pm#2 $
#
## @summary gx_utils library functions to be used by gx_boot and deux_init
##
## @description
##   This library contains routines shared by gx_boot.thpl and deux_init.thpl
##   The routines in this library replicate some of the functionality that
##   would be available to someone using Server via a cli connection.
##
##   The gx_boot.thpl script only has a console session and deux_init.thpl only
##   has a cli session, so this library uses $Server->say_wrapper() to invoke all
##   cli commands after determining which type of session we have.
##   Also, a reference to the Server object must be passed in rather than creating 
##   a local object based on the Server singleton because gx_boot.thpl calls Server
##   with no_cli and deux_init does not, so we would not know how to call the
##   constructor to satisfy both usages.
##
## @usage
##   use gx_utils;
##   use gx_utils qw(lifinfo);
##
## @keywords  gx_utils test library
##
## @dependencies
## use config_base
## use Tharn
## 
## @author dl-cluster-qa@netapp.com
## @status testme
## @change 2010-06-01 radish: created
## 


package gx_utils;

use strict;
use vars ('@ISA','@EXPORT');

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
            lifinfo
            cluster_mgmt_lifinfo
            resolve_port_config
            ports_by_role
            change_port_config
            has_change
            check_initialization_permission
            );

use Options;
use Params::Common;
use Server;
use Socket;
use Tharn;
use NATE::Log::Call qw(call_trace);
use Data::Dumper;
use gx_licenses;
use NACL::C::Node;
use NACL::CS::NetworkIpspace;

our $Log;

my %uichanges;

my $options = new Options();

$options->parameters( definitions =>
    [
    # need a way to prevent gx_boot from happening on nodes someone doesn't want it to.  This is
    # setup to allow for future expansion but should never allow an option to lets someone use less
    # than the "simple" mechanism.
    {
        name            => 'INITIALIZE_PERMISSION_TYPE',
        description     => 'How should check_initialization_permission verify you are permitted to init nodes. ' .
            'Defaults to "simple" nate_hosts verification and assumes if you passed in ' .
            'the nodes on the command line, and no one has blocked the nodes in their nate_hosts ' .
            'files you have permission to execute the code you are using against them.',
        default         => 'simple',
        valid_values    => [ qw( simple ) ],
    },
    # this parameter is here just to suck it in to the Options object and is only used during
    # permisson checks
    {
        name        => 'TESTBED',
        description => 'Name of the TESTBED we may be operating within.',
        spec        => 's',
    },
    ]);

$options->process(hostrec=>1);

sub portsort {
    $a =~ /(\d+)$/;
    my $an = $1;
    $b =~ /(\d+)$/;
    my $bn = $1;

    $an <=> $bn;
}

## @name lifinfo
## @description Returns array of hashs containing vif information of a 
##              particular interface type for a single node.  Looks through
##              the HOSTREC hash for applicable configuration information.
##
## @synopsis 
## lifinfo( server => $node, interface_type => "CLUSTER" );
##
## @arg server         Required: The node to act upon
## @arg interface_type Required: The interface type to retrieve
##                               one of CLUSTER|MGMT|DATA|CLUSTER_MGMT|INTERCLUSTER
## @arg ports          Optional: A list of ports if <interface_type>_PORTx
##                               are not defined
## @arg required_keys  Optional: A list of keys which, if missing will cause an error
##                                one of <port|address|netmask|gateway|mtu>
##                                Note: routine will attempt to default ports if required but
##                                      not found in the nate hosts file by looking at the ports
##                                      of that interface type currently defined on the node.
## @arg default_mtu    Optional: Default mtu to value given if not specified in the nate hosts file
##                               The default is 1500.
##
## Note: 
##     These are the nate hosts variables that this routine looks for:
##
##     <interface_type>_PORTx
##     <interface_type>_IPx
##     <interface_type>_MASKx
##     <interface_type>_GATEWAYx
##     <interface_type>_MTUx
##     DATA_PROTOCOLx
##
##     For interface_types CLUSTER, INTERCLUSTER and DATA, where x = 1,2,3,4,etc...
##
##     <interface_type>_PORT
##     <interface_type>_IP
##     <interface_type>_MASK
##     <interface_type>_GATEWAY
##     <interface_type>_MTU
##
##     For interface_types MGMT and CLUSTER_MGMT
##
##     Alternatively, a more compact form was developed a few years ago,
##     but it is not widely used:
##
##     <interface_type>_CONFIGx="port=e0a address=172.31.13.18 netmask=255.255.192.0 gateway=172.31.0.1 mtu=1500"
##
##  NOTE: to avoid a nate hosts variable that looks like DATA_DATA_PROTOCOL1 we shortcut the
##        variable down to just DATA_PROTOCOLx to control the "data-protocol" of a DATA LIF.  Internally
##        we pass around 'data-protocol' to match the structure the product has.  It is only valid
##        to set 'data-protocol' on a DATA LIF, so other types of LIFS do not support that parameter.
##
##  You can limit which variables we look at in the nate hosts file by passing in
##  a list of required_keys.  If a required key is not present, the routine will error out.
##  If one of the required keys is port, this routine will attempt to default the ports
##  to ports of that role that are currently defined on the node.
##
sub lifinfo {
    # This sub makes an array of hashes containing vif information for a particular interface type for a single node.
    my %params = (
        server => undef,
        interface_type => undef,
        ports => undef,
        required_keys => undef,
        default_mtu => "1500",
        source_of_port => undef,
        with_v6 => 0,
        @_
    );

    if ( !($params{server} && $params{interface_type}) ) {
        $Log->comment( "lifinfo args = ".Dumper(\%params) );
        $Log->error( die script_result("Must supply server & interface_type"));
    }

    my $source_of_port;
    my $Server = $params{server};
    my $node = $Server->get_execution_server();
    my $option_ref = $options->get_hash();

    if ( !defined $option_ref->{HOSTRECS}{$node} ) {
        $option_ref->{HOSTRECS}{$node} = host($node);
    }

    # make sure that any hostrec overrides in NODES have been applied
    _apply_hostrec_overrides( $option_ref );

    my @retval;
    my @lif_hash;

    my %nettype_converter = %Hostrec::Appliance::Nettype_Converter;

    # Check if we must handle v6 records too
    if ( $params{with_v6} ) {
        # Push inet6 records into my @retval
        @lif_hash = $option_ref->{HOSTRECS}{$node}->net_filter( 
            include => { 
                family => 'inet6',
                nettype => $nettype_converter{$params{interface_type}}, 
            } 
        );

        my $lif_len = scalar( @lif_hash );
        for( my $i = 0 ; $i < $lif_len; $i++ ) {
            $lif_hash[$i]->{name} = _set_if_name( $node,$params{interface_type},($i+1) );

            push( @retval, {
                    'gateway' => $lif_hash[$i]->{gateway},
                    'name' => $lif_hash[$i]->{name},
                    'netmask' => $lif_hash[$i]->{netmask},
                    'port' => $lif_hash[$i]->{interface},
                    'role' => uc ($lif_hash[$i]->{nettype}),
                    'source_of_port' => 'config',
                    'address' => $lif_hash[$i]->{address},
                    'mtu' => $lif_hash[$i]->{mtu},
                    'family' => $lif_hash[$i]->{family},
                }
            );
        }
    }

    $Log->debug("params = ".Dumper(\%params) );
    my @ports;
    my @required_keys = qw(name port address netmask gateway mtu);
    if ( $params{required_keys} ) {
        @required_keys = @{$params{required_keys}};
    }

    my @keys;
    # Retrieve all host configuration related to lifs
    # NOTE: MGMT_IP or CLUSTER_MGMT_IP don't have a numeric suffix
    @keys = grep(/^$params{interface_type}_(IP|PORT|MTU|MASK|GATEWAY|CONFIG)\d*$/, 
        keys %{$option_ref->{HOSTRECS}{$node}});

    $Log->debug("Port role $params{interface_type} keys = ".Dumper(\@keys) );

    # If caller didn't specify port defaults or ports =\(), see if ports of this role
    # already exist
    if ( !defined $params{ports} || !@{$params{ports}} ) {

        # normalize the interface_type so it will match one
        # of the hash keys
        my $it = lc $params{interface_type};
        if ( $it =~ /^mgmt$|^node-mgmt$/i ) {
            $it = "mgmt";
        }

        my %ports_by_role = ports_by_role( server => $Server );
        @ports = ($ports_by_role{$it}) ? @{$ports_by_role{$it}} : ();

        # If no ports of this role exist, default them to something reasonable
        if ( !@ports ) {
            # setup defaults appropriate for a given role
            if ( $it =~ /^cluster$/ ) {
                my $model = $option_ref->{HOSTRECS}{$node}->{MODEL};

                if ( $model =~ /8040|8060|2868-CN1|2868-CN2|8080|2868-EN4/ ) {

                    # Bimota and Augusta support a default of 4 ports
                    @ports = qw( e0a e0b e0c e0d );
                } else {
                    @ports = qw( e0a e0b );
                }
            } elsif ( $it =~ /^mgmt$/ ) {
                @ports = qw( e4a e4b );
            } else {
                @ports = qw( e0c e0d );
            }
            $source_of_port = "default";
        }
    } else {
        @ports = @{$params{ports}};
    }

    my %encountered;
 
    # @keys should now have everything that we need to operate over, but do them in order
    foreach my $key (sort portsort @keys){
        $key =~ /(\d+)$/;

        # Note: For MGMT_ or CLUSTER_MGMT, instance will be undef...
        my $instance = $1;

        $Log->debug("Port Role |$params{interface_type}| key = |$key|, instance = |$instance|");

        if ($key =~ /IP|PORT|MTU|MASK|GATEWAY|CONFIG|PROTOCOL/){
            # Have we already processed this instance?
            if ( $encountered{$instance} ) {
                next;
            } else {
                $Log->debug("Will process Port Role $params{interface_type} key = $key, instance = $instance");
                $encountered{$instance} = 1;
            }
        }

        my %data = (
            mtu => $params{default_mtu},
        );

        $data{name} = _set_if_name( $node,$params{interface_type},$instance );

        if ($key =~ /IP|PORT|MTU|MASK|GATEWAY|PROTOCOL/){
            # Older format where everything is in individual keys
            $data{port}    = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_PORT$instance"};
            $data{address} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_IP$instance"};
            $data{netmask} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MASK$instance"};
            $data{gateway} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_GATEWAY$instance"};

            if (    exists($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU$instance"})
                && defined($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU$instance"})){
                $data{mtu} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU$instance"};
            }

            # for DATA lifs we could also need to know the data-protocol(s) the LIF will support
            if ( $params{interface_type} =~ /data/i ) {
                if ( exists($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_PROTOCOL$instance"}) && 
                    defined($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_PROTOCOL$instance"}) ) {
                    $data{data_protocol} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_PROTOCOL$instance"};
                    # Make sure what we have here makes sense.  IE NAS and SAN can't live together, and
                    # none can't live with anything else.
                    my $protocol_ok = 1;
                    my %protocols = (
                        # Name         # type
                        "nfs"       => "nas",
                        "cifs"      => "nas",
                        "fcache"    => "nas",
                        "none"      => "none",
                        "iscsi"     => "san",
                        "fcp"       => "san",
                    );
                    my $lif_family;
                    foreach my $protocol ( split /,/, $data{data_protocol} ) {
                        if ( grep(/$protocol$/, keys(%protocols)) == 0 ) { # make sure the protocol is valid
                            $Log->error( die  config_result(
                                    "Invalid data_protocol '$protocol' for a LIF in the hostrec of node '$node' " .
                                    "for '$params{interface_type}_PROTOCOL$instance' Please validate that the " .
                                    "entries in the hostrec for this node are correct.  Valid values are '" .
                                    join(",", keys %protocols) . "' and SAN (iscsi,fcp) and NAS " .
                                    "(nfs,cifs,fcache) protocols cannot be mixed on the same LIF."
                                ) );
                        } else {
                            if ( $lif_family ) {
                                # make sure this protocol matches the previous family
                                if ( $lif_family ne $protocols{$protocol} ) { 
                                    $Log->error( die config_result(
                                            "Encountered a LIF that is trying to include incompatible " .
                                            "data_protocol entries in the hostrec for node '$node' " .
                                            "'$params{interface_type}_PROTOCOL$instance' set to " .
                                            "'$data{data_protocol}' which includes an illegal mix of " .
                                            "NAS (nfs,cifs,fcache), SAN (fcp, iscsi) or No (none) protocols."  
                                        ) );
                                }
                            } else {
                                # set the family of this lif to the first type we see
                                $lif_family = $protocols{$protocol};
                            }
                        }
                    }
                } # end date role lif with PROTOCOL set
            } # end data role lif

            # If MGMT or CLUSTER_MGMT, should be no numeric suffix
            if ( $params{interface_type} =~ /mgmt/i ) {
                if ( !$data{port} ) {
                    $data{port}    = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_PORT"};
                }
                if ( !$data{address} ) {
                    $data{address} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_IP"};
                }
                if ( !$data{netmask} ) {
                    $data{netmask} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MASK"};
                }
                if ( !$data{gateway} ) {
                    $data{gateway} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_GATEWAY"};
                }

                if (    exists($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU"})
                    && defined($option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU"})){
                    $data{mtu} = $option_ref->{HOSTRECS}{$node}->{$params{interface_type}."_MTU"};
                }
            }
        } elsif ($key =~ /CONFIG/){
            # Newer format where the config is more like what is used in CFE
            foreach my $pair (split(/\s+/,$option_ref->{HOSTRECS}{$node}->{$key})){
                my ($k, $v) = split(/=/, $pair);
                $data{$k} = $v;
            }
        } # No else since we matched the keys before

        # This is only here to show how we got the port
        # i.e. did we get it from the nate host variables (config),
        #      we got it from the filer (node), we defaulted it to
        #      some "well-known" list of ports (default), or
        #      we don't know where this came from (unknown).
        if ( $data{port} ) {
            $data{source_of_port} = "config";
        } else {
            if ( $params{source_of_port} ) {
                $data{source_of_port} = $params{source_of_port};
            } else {
                $data{source_of_port} = "unknown";
            }
        }

        $data{role} = $params{interface_type};

        $Log->trace("Data gleaned from nate host, $node and defaults is " . Dumper(\%data));
        my @missing;
        my $num_ports = scalar(@ports);
        foreach my $required_key (@required_keys){
            if ( !exists($data{$required_key}) ||
                !defined($data{$required_key}) ||
                 !length($data{$required_key})){
                if ($required_key =~ /port/) {
                    my $port_var_name;
                    my $config_var_name;
                    if ( $params{interface_type} =~ /mgmt/i ) {
                        $port_var_name   = (uc $params{interface_type})."_PORT";
                        $config_var_name = (uc $params{interface_type})."_CONFIG";
                    } else {
                        $port_var_name   = (uc $params{interface_type})."_PORT".$instance;
                        $config_var_name = (uc $params{interface_type})."_CONFIG".$instance;
                    }

                    if ( $instance <= $num_ports ) {
                        # For DATA_, CLUSTER_, INTERCLUSTER_
                        # instance is 1-relative, array index is 0-relative,
                        # subtracting 1 before the modulo will ensure that the lowest
                        # instance matches with the first port in the list
                        if ( $instance >= 1 ) {
                            $data{$required_key} = $ports[($instance -1) % @ports];
                        } else {
                            # For MGMT and CLUSTER_MGMT, $instance will be 0 (undef)
                            $data{$required_key} = $ports[$instance % @ports];
                        }
                        # if source of port is not default or config, then we must have discovered
                        # it from the node
                        if ( !$data{source_of_port} ) {
                            $data{source_of_port} = "node";
                        }

                        # Thought about making this a warn, but MANY nate_hosts files define DATA_PORTx but no CLUSTER_PORTx
                        $Log->comment("  \nNo $port_var_name (or \'port\' in $config_var_name) defined in the nate host variables.  ".
                                      "Defaulting $port_var_name to $data{$required_key}.  ".
                                      "If this is not what you want, please define $port_var_name in the $node nate host variables.");
                    } else {
                        $Log->comment("  \nDebugging info for the following error.\n".
                                          "  Data from $node nate host file, node and defaults is " . Dumper(\%data));
                        $Log->error( die config_result(
                                        "  \nNo $port_var_name (or \'port\' in $config_var_name) defined in the nate host variables.  ".
                                        "Cannot default $port_var_name because only (@ports) $params{interface_type} ports were ".
                                        "found on $node or defined in the $node nate host variables"));
                    }
                } else {
                    my $it              = uc $params{interface_type};
                    my $missing_varname = uc $required_key;

                    if ( $required_key =~ /address/i ) {
                        $missing_varname = $it."_IP";
                    } elsif ( $required_key =~ /netmask/i ) {
                        $missing_varname = $it."_MASK";
                    } else {
                        $missing_varname = $it."_".$missing_varname;
                    }

                    if ( $instance ) {
                        $missing_varname = $missing_varname.$instance;
                    }
                    push @missing, $missing_varname;
                }
            }
        }
        if (@missing){
            $Log->error( die config_result( "  \nWhile processing nate_host definitions for $node, could not find required variables => " . join(',', @missing)."\n".Dumper(\%params)));
        }

        push (@retval, \%data);
    }

    $Log->debug("lifinfo = ".Dumper(\@retval) );
    return @retval;
}

## @name cluster_mgmt_lifinfo
## @description Returns hash containing lif information for the
##              cluster management lif.  Looks through the HOSTREC 
##              hash for applicable configuration information or defaults
##              based off of the MGMT_IP.
##
## @synopsis 
## ::cluster_mgmt_lifinfo(first_node => $node );
##
## @arg first_node     Required: The node on which you will create the cluster
##
##
sub cluster_mgmt_lifinfo {
    # This sub makes an array of hashes containing vif information for a particular interface type for a single node.
    my %params = (
        first_node => undef,
        @_
    );

    if ( !$params{first_node} ) {
        $Log->error( die script_result("Must supply first_node"));
    }

    my $first_node = $params{first_node};
    my %retval;

    my $option_ref = $options->get_hash();
    if ( !defined $option_ref->{HOSTRECS}{$first_node} ) {
        $option_ref->{HOSTRECS}{$first_node} = host($first_node);
    }

    # make sure that any hostrec overrides in NODES have been applied
    _apply_hostrec_overrides( $option_ref );

    my $test = $option_ref->{HOSTRECS}{$first_node}->{MGMT_IP};
    my @octets = split (/\./,$test);

    # Pick the up the lif info from the nate hosts file if it exists, otherwise build info from
    # 172.16 and the last 2 octets of the MGMT_IP
    $retval{cm_port}    = $option_ref->{HOSTRECS}{$first_node}->{CLUSTER_MGMT_PORT}    || $option_ref->{HOSTRECS}{$first_node}->{DATA_PORT1};
    $retval{cm_ip}      = $option_ref->{HOSTRECS}{$first_node}->{CLUSTER_MGMT_IP}      || "172.16.$octets[2].$octets[3]";
    $retval{cm_mask}    = $option_ref->{HOSTRECS}{$first_node}->{CLUSTER_MGMT_MASK}    || "255.255.255.0";
    $retval{cm_gateway} = $option_ref->{HOSTRECS}{$first_node}->{CLUSTER_MGMT_GATEWAY} || "172.16.$octets[2].$octets[3]";

    # BURT 433004
    # Check if MGMT_IP and Mask results in a valid n/w address by ANDing
    # the ip address with the mask
    my $raw_masked = inet_aton($retval{cm_ip}) & inet_aton($retval{cm_mask});
    my $masked = inet_ntoa($raw_masked);

    # Check for ADDRESS == (ADDRESS & MASK)
    if($retval{cm_ip} eq $masked)
    {
        $Log->error( die config_result("Cluster Mgmt IP (".$retval{cm_ip}.
        ") ANDed with Cluster Mgmt Mask (".$retval{cm_mask}.") for node $first_node result in network address $masked. Cannot specify n/w address"));
    }

    if ( !$retval{cm_port} ) {            
        $Log->warn("Node '$first_node' does not have a CLUSTER_MGMT_PORT or DATA_PORT1 defined.  Falling back to 'e0c' as the CLUSTER_MGMT_PORT");
        $retval{cm_port} = "e0c";
    }

    return %retval;
}

## @name ::resolve_port_config
## @description Changes node port config based on definitions in your
##              nate hosts file.  Currently modifies port role and/or MTU.
##
## @synopsis 
## gx_utils::resolve_port_config( server => $server );
##
## @arg server   Required: The server object to use
## @arg role     Optional: If 1, change port role based on nate hosts config.
##                         (see description for strict below)
## @arg mtu      Optional: If 1, change port mtu based on nate hosts config.
## @arg strict   Optional: If 1, try to set Node-Mgmt and Intercluster
##                         port roles as node-mgmt/intercluster IF those
##                         ports are not also specified as some other
##                         port type.  In other words, if you have
##
##                             INTERCLUSTER_PORT1 = e0d
##                             DATA_PORT2 = e0d
##
##                         then e0d will be typed as "data".  If only
##                         INTERCLUSTER_PORT1 referenced e0d and strict=1,
##                         AND the filer supports intercluster roles, then
##                         e0d will be configured as an intercluster role.
##                         If strict is 0, then all ports other than 
##                         cluster ports will have their roles set to "data".
##
##  Note: Cluster-Mgmt lifs can be created on node-mgmt or data ports
##        Node-Mgmt    lifs can be created on node-mgmt or data ports
##        Cluster      lifs can only be created on cluster ports
##        Data         lifs can only be created on data ports
##        Intercluster lifs can be created on intercluster or data ports
##
##        If node-mgmt, cluster-mgmt or intercluster config is specified
##        in the nate hosts file, and the port that is specified is not
##        used for anything else, and the "strict" option was specified,
##        then we should change the port role to that specific port role.
##        Otherwise, those port roles will be set to data.
##
sub resolve_port_config {

    my %params = (
        server => undef,
        strict => undef,
        role   => undef,
        mtu    => undef,
        @_
    );

    if ( !($params{server} && ($params{role} || $params{mtu}) ) ) {
        $Log->comment( "resolve_port_config args = ".Dumper(\%params) );
        $Log->error( die script_result("Must supply server and role|mtu"));
    }

    my $Server = $params{server};
    my $node   = $Server->get_execution_server();

    # Get existing configured port roles
    my %ports_by_role = ports_by_role( server => $Server );

    my $option_ref = $options->get_hash();
    if ( !defined $option_ref->{HOSTRECS}{$node} ) {
        $option_ref->{HOSTRECS}{$node} = host($node);
    }

    # make sure that any hostrec overrides in NODES have been applied
    _apply_hostrec_overrides( $option_ref );

    $Log->debug("ports_by_role = ".Dumper(\%ports_by_role) );

    # Use these as defaults in case ports are not defined in nate hosts
    my @cluster_exist      = ($ports_by_role{cluster})        ? @{$ports_by_role{cluster}}      : ();
    my @mgmt_exist         = ($ports_by_role{mgmt})           ? @{$ports_by_role{mgmt}}         : ();
    my @data_exist         = ($ports_by_role{data})           ? @{$ports_by_role{data}}         : ();
    my @ic_exist           = ($ports_by_role{intercluster})   ? @{$ports_by_role{intercluster}} : ();

    # build array of CLUSTER, DATA, CLUSTER_MGMT and MGMT, ports as they are defined in the nate hosts file.
    # The ports we are passing in are only default values
    my ( @cluster_config, @data_config, @cluster_mgmt_config, @mgmt_config, @ic_config ); 

    # Ports passed in to lifinfo are only used if no ports of that type are specified in the nate hosts file
    # We're passing default_mtu as "unspecified" so we can tell if the MTU was set in the nate hosts
    # file, rather than using lifinfo's default of 1500
    my @required_keys = qw(name port mtu);
    @cluster_config      = gx_utils::lifinfo( server => $Server, interface_type => "CLUSTER",      
                                              required_keys => \@required_keys, default_mtu => "unspecified",
                                              source_of_port => "node",
                                              ports => \@cluster_exist );

    # slice cluster ports out of array of hashes
    my @cluster_config_ports      = map $_->{port}, @cluster_config;

    # remove the cluster_config ports from other "exist" lists.  These other lists are
    # primarily used as defaults in case a given port role type (MGMT_PORT for example) 
    # is not specified in the nate_hosts file.
    # For example, if e0a,e0b,e0c,e0d were all originally roled as data on the node, but 
    # we then specified e0a,e0b as our cluster ports, we would not also want to pass in e0a,e0b 
    # as default data ports.

    foreach my $cluster_config_port ( @cluster_config_ports ) {
        @data_exist          = grep(!/^$cluster_config_port$/, @data_exist );
        @mgmt_exist          = grep(!/^$cluster_config_port$/, @mgmt_exist );
        @ic_exist            = grep(!/^$cluster_config_port$/, @ic_exist );
        @cluster_mgmt_config = grep(!/^$cluster_config_port$/, @cluster_mgmt_config );
    }

    # Every lif type other than cluster lifs can live on a data port
    # so if they conflict, no big deal (i.e. if we specify MGMT_PORT=e0c and DATA_PORT1=e0c)

    @mgmt_config         = gx_utils::lifinfo( server => $Server, interface_type => "MGMT",
                                              required_keys => \@required_keys, default_mtu => "unspecified",
                                              source_of_port => "node",
                                              ports => \@mgmt_exist );

    @ic_config           = gx_utils::lifinfo( server => $Server, interface_type => "INTERCLUSTER",
                                              required_keys => \@required_keys, default_mtu => "unspecified",
                                              source_of_port => "node",
                                              ports => \@ic_exist );

    @cluster_mgmt_config = gx_utils::lifinfo( server => $Server, interface_type => "CLUSTER_MGMT",
                                              required_keys => \@required_keys, default_mtu => "unspecified",
                                              source_of_port => "node",
                                              ports => \@mgmt_exist );

    # Ports that are specified for each type (slice ports out of array of hashes)
    my @mgmt_config_ports         = map $_->{port}, @mgmt_config;
    my @ic_config_ports           = map $_->{port}, @ic_config;
    my @cluster_mgmt_config_ports = map $_->{port}, @cluster_mgmt_config;

    # Order @default_data_ports based on whether or not it would need to be shared with 
    # other role types.  If possible, we'd rather NOT have a port conflict in case
    # we want to enforce strict port roles, so if we have a large number of ports that
    # are currently 'data', move the ports which we are going to configure as mgmt, intercluster
    # and cluster mgmt ports at the end of the list.  

    # For example, say we have e0c,e1a,e1b,e1c,e1d currently defined as data ports, but we
    # want e0c as our node-mgmt port, and e1a,e1b as our intercluster ports, we reorder the
    # @default_data_ports list so it looks like e1c,e1d,e1a,e1b,e0c so that we pick
    # ports that aren't being used any other way first.

    my @default_data_ports;
    my @conflicting_ports;
    foreach my $port ( @data_exist ) {
        if ( grep(/^$port$/, @cluster_mgmt_config_ports, @ic_config_ports, @mgmt_config_ports ) ) {
            push( @conflicting_ports, $port );
        } else {
            push( @default_data_ports, $port );
        }
    }    

    # If a port which is currently defined as DATA, is also specified as one of these other port types
    # add it to the end of the default data port list.
    @default_data_ports = ( @default_data_ports, @conflicting_ports );

    @data_config         = gx_utils::lifinfo( server => $Server, interface_type => "DATA",
                                              required_keys => \@required_keys, default_mtu => "unspecified",
                                              source_of_port => "node",
                                              ports => \@default_data_ports );

    my @data_config_ports         = map $_->{port}, @data_config;

    $Log->debug("cluster_exist "     .Dumper(\@cluster_exist)."\n".
                "mgmt_exist "        .Dumper(\@mgmt_exist)."\n".
                "data_exist "        .Dumper(\@data_exist)."\n".
                "ic_exist "          .Dumper(\@ic_exist)."\n");

    $Log->debug("cluster_config "     .Dumper(\@cluster_config)."\n".
                "cluster_mgmt_config ".Dumper(\@cluster_mgmt_config)."\n".
                "mgmt_config "        .Dumper(\@mgmt_config)."\n".
                "data_config "        .Dumper(\@data_config)."\n".
                "ic_config "          .Dumper(\@ic_config)."\n");

    $Log->debug("cluster_config_ports "     .Dumper(\@cluster_config_ports)."\n".
                "cluster_mgmt_config_ports ".Dumper(\@cluster_mgmt_config_ports)."\n".
                "mgmt_config_ports "        .Dumper(\@mgmt_config_ports)."\n".
                "data_config_ports "        .Dumper(\@data_config_ports)."\n".
                "ic_config_ports "          .Dumper(\@ic_config_ports)."\n");

    my %change_role;
    my %change_mtu;

    if ( $params{role} ) {
        # Check to see if we have the "mgmt-role-split" UI change
        my $have_split_mgmt_change = gx_utils::has_change( server => $Server, change => "mgmt-role-split");
        my $conn_type = _determine_server_connection_type( server => $Server );
    
        # question mark immediately shows the possible roles, then
        # "prompt::> network port show -role" will actually be returned
        my $port_roles = $Server->say_wrapper( nocr => 1, type => $conn_type, cmd => "network port show -role ?" );

        # Example  
        #  prompt::*> network port show -role ?
        #  cluster
        #  data
        #  node-mgmt
        #  intercluster
        #  prompt::*> network port show -role

        eval {
            # This will fail, but I don't care
            # If I didn't enter return, "network port show -role" would actually
            # be left over on the console as residue.

            $Server->say_wrapper( nocr => 1, type => $conn_type, cmd => "\r" );
        };

        my @roles_supported = split( /\s+/, $port_roles );

        my $has_node_mgmt    = grep(/^node-mgmt$|^mgmt$/, @roles_supported );
        my $has_intercluster = grep(/^intercluster$/,     @roles_supported );

        $Log->debug(" has_node_mgmt = $has_node_mgmt, has_intercluster = $has_intercluster");

        # Process cluster ports

        # If we have ports currently defined as cluster ports that do not
        # correspond to CLUSTER_PORTx entries in the nate hosts file, we
        # need to change their role to be data so they do not interfere with
        # the cluster ports defined in the nate hosts file.
        foreach my $config (@cluster_config_ports) {
            # If this port already is a cluster port, no need to change
            if ( grep(/^$config$/, @cluster_exist ) ) {
                # remove current port from cluster_exist
                @cluster_exist = grep(!/^$config$/, @cluster_exist );
            } else {
                $change_role{$config} = "cluster";
                $Log->debug("change_role of cluster port $config = $change_role{$config}");
            }

            # If any other port type is defined on the same port as a cluster port, this is an error
            my $c_conflict = grep(/^$config$/, @cluster_mgmt_config_ports );
            my $m_conflict = grep(/^$config$/, @mgmt_config_ports );
            my $d_conflict = grep(/^$config$/, @data_config_ports );
            my $i_conflict = grep(/^$config$/, @ic_config_ports );

            if ( $c_conflict || $m_conflict || $d_conflict || $i_conflict ) {

                my $conflict_type="";

                $conflict_type = ($c_conflict) ? "CLUSTER_MGMT_PORT,".$conflict_type  : $conflict_type;
                $conflict_type = ($m_conflict) ? "MGMT_PORT,".$conflict_type          : $conflict_type;
                $conflict_type = ($d_conflict) ? "DATA_PORTx,".$conflict_type         : $conflict_type;
                $conflict_type = ($i_conflict) ? "INTERCLUSTER_PORTx,".$conflict_type : $conflict_type;

                $Log->error( "CLUSTER_PORTx=$config conflicts with following port types ($conflict_type)");

                $Log->comment("cluster_exist "     .Dumper(\@cluster_exist)."\n".
                              "mgmt_exist "        .Dumper(\@mgmt_exist)."\n".
                              "data_exist "        .Dumper(\@data_exist)."\n".
                              "ic_exist "          .Dumper(\@ic_exist)."\n");
            
                $Log->comment("cluster_config "     .Dumper(\@cluster_config)."\n".
                              "cluster_mgmt_config ".Dumper(\@cluster_mgmt_config)."\n".
                              "mgmt_config "        .Dumper(\@mgmt_config)."\n".
                              "data_config "        .Dumper(\@data_config)."\n".
                              "ic_config "          .Dumper(\@ic_config)."\n");
            
                $Log->error( die config_result( 
                                 "Resolve conflict between CLUSTER_PORTx=$config and the following types ".
                                 "($conflict_type) ".
                                 "by explicitly defining the appropriate variables in the $node nate host file."));
            }
        }

        # if there are any cluster ports left over in @cluster_exist, 
        # change their role to be data.
        foreach my $exist (@cluster_exist) {
            $change_role{$exist} = "data";
            $Log->debug("change_role of existing cluster port $exist = $change_role{$exist}");
        }
        if ( grep(/^e0M$/, @data_config_ports, @ic_config_ports) ) {
                $Log->warn( "An attempt was made to change e0M to $change_role{e0M} "
                        . "Since this is not supported, e0M's port role will continue to be node-mgmt"
                );
        }
        # Process cluster management ports
        # Currently I believe that cluster management lifs MUST be created
        # on data ports, maybe later on allowed on node-mgmt?
        foreach my $config (@cluster_mgmt_config_ports) {
            $change_role{$config} = "data";
            $Log->debug("change_role of cluster-mgmt port $config = $change_role{$config}");
        }

        my $strict_mgmt_role = $params{strict};

        # MGMT_PORT_ROLE is a legacy option from deux_init.  If "strict" 
        # argument passed in is false, but MGMT_PORT_ROLE is true, then 
        # we will treat MGMT_PORT_ROLE as a strict=true specifically for
        # node mgmt ports.  Same rule apply, though, if this same port is
        # used in another way (say as a data port), we will not change
        # the role.
        # Allow MGMT_PORT_ROLE to be defined globally, or per node
        if ( !$strict_mgmt_role && 
             ($option_ref->{MGMT_PORT_ROLE} || 
              $option_ref->{HOSTRECS}{$node}->{MGMT_PORT_ROLE}) ) {
            $strict_mgmt_role = 1;
        }

        # Process node management ports
        # This looks a little tricky, but we want to handle the case
        # where we only change the role if we must (i.e. if the current
        # role is unacceptable)
        foreach my $config (@mgmt_config_ports) {

            my $port_role = undef;
            # if strict mgmt role was set or the port is an e0M port
            # set the port role to node-mgmt or mgmt
            if ( $strict_mgmt_role && $has_node_mgmt || ($config =~ /^e0M$/)) {
                if ( $have_split_mgmt_change ) {
                    $port_role = "node-mgmt";
                } else {
                    $port_role = "mgmt";
                }
            } else {
                # If we have specified this port any other way, change it to data
                if ( grep(/^$config$/, @cluster_mgmt_config_ports, @data_config_ports, @ic_config_ports ) ) {
                $port_role = "data";
                }
            }

            my $is_currently_data = grep(/^$config$/, @data_exist );
            my $is_currently_mgmt = grep(/^$config$/, @mgmt_exist );

            my $is_either = ($is_currently_data || $is_currently_mgmt);

            # If we didn't force a port role, but current role isn't node-mgmt or
            # data, force to "data".
            if ( !$port_role && !$is_either ) {
                $port_role = "data";
            }

            # Only change if not already that type of port
            if ( (($port_role =~ /mgmt/) && !$is_currently_mgmt) ||
                 (($port_role =~ /data/) && !$is_currently_data) ) {
                $change_role{$config} = $port_role;
                $Log->debug("change_role of mgmt port $config = $change_role{$config}");
            } else {
                # burt709637: in this case we don't want to change this mgmt port's role because
                # it is currently the role we intended it to be, remove it from the set of ports
                # that will get their roles changed just in case it was previously
                # added to the hash
                delete($change_role{$config});
            }
        }

        # Process intercluster ports
        foreach my $config (@ic_config_ports) {
            my $port_role = "data";

            # If we have specified this port any other way, leave it data
            if ( !grep(/^$config$/, @mgmt_config_ports, @data_config_ports, @cluster_mgmt_config_ports ) ) {
                if ( $params{strict}  && $has_intercluster ) {
                    $port_role = "intercluster";
                }
            }
    
            $change_role{$config} = $port_role;
            $Log->debug("change_role of intercluster port $config = $change_role{$config}");
        }

        # Process data ports
        # data ports can host everything except cluser lifs
        foreach my $config (@data_config_ports) {
            $change_role{$config} = "data";
            $Log->debug("change_role data port $config = $change_role{$config}");
        }
    }

    my $mtu_conflict = 0;
    if ( !$params{mtu} ) {
        undef %change_mtu;
    } else {

        # Warn if cluster port MTUs aren't all the same
        my $cluster_mtu;
        my $cluster_mtu_conflict;
        foreach my $config (@cluster_config) {
            if ( $config->{mtu} && ($config->{mtu} !~ /unspecified/) ) {
                if ( "$cluster_mtu" ne "$config->{mtu}" ) {
                    if ( $cluster_mtu ) {
                        $cluster_mtu_conflict = 1;
                        $Log->warn("Conflicting MTU definitions for CLUSTER ports. Port $config->{port} ".
                                   "specifies $config->{mtu}, but a previous CLUSTER port specified $cluster_mtu.\n".
                                   "Setting all CLUSTER port MTUs to $cluster_mtu.  If this is not what you want\n".
                                   "explicitly define all CLUSTER_MTUx in your $node nate host file." );
                    } else {
                        $cluster_mtu = $config->{mtu};
                    }
                }
            }
        }

        my $mtu;
        if ( $cluster_mtu_conflict ) {
            $mtu = $cluster_mtu;
        }
        foreach my $config (@cluster_config) {

            # Only set the mtu if it was specified in a nate hosts file
            if ( ($config->{mtu} && ($config->{mtu} !~ /unspecified/)) || $cluster_mtu_conflict ) {
                if ( !$cluster_mtu_conflict ) {
                    $mtu = $config->{mtu};
                }
                if ( defined $change_mtu{$config->{port}} ) {
                    if ( "$change_mtu{$config->{port}}" ne "$mtu" ) {
                        $Log->warn("Conflicting MTU definitions for PORT $config->{port}, role $config->{role} ".
                                   "specifies $mtu, but was previously $change_mtu{$config->{port}}.\n".
                                   "Possibly port $config->{port} was specified as multiple different roles in $node nate host configuration?" );
                        $mtu_conflict = 1;
                    }
                }
                $change_mtu{$config->{port}} = $mtu;
            }
        }

        foreach my $config (@data_config, @cluster_mgmt_config, @mgmt_config, @ic_config) {

            # Only set the mtu if it was specified in a nate hosts file
            if ( $config->{mtu} && ($config->{mtu} !~ /unspecified/) ) {
                if ( defined $change_mtu{$config->{port}} ) {
                    if ( "$change_mtu{$config->{port}}" ne "$config->{mtu}" ) {
                        $Log->warn("Conflicting MTU definitions for $config->{role} PORT $config->{port}.  ".
                                   "Specified as $mtu, but was previously $change_mtu{$config->{port}}.\n".
                                   "Possibly port $config->{port} was specified as multiple different roles in $node nate host configuration?" );
                        $mtu_conflict = 1;
                    }
                }
                $change_mtu{$config->{port}} = $config->{mtu};
            }
        }

        if ( $mtu_conflict ) {
            $Log->comment( "MTU conflict discovered:");
            $Log->comment( "CLUSTER CONFIG = ".Dumper(\@cluster_config) );
            $Log->comment( "DATA CONFIG = ".Dumper(\@data_config) );
            $Log->comment( "CLUSTER MGMT CONFIG = ".Dumper(\@cluster_mgmt_config) );
            $Log->comment( "MGMT CONFIG = ".Dumper(\@mgmt_config) );
            $Log->comment( "INTERCLUSTER CONFIG = ".Dumper(\@ic_config) );
        }
    }

    # %change_role & %change_mtu may contain different ports, so run through both hashes
    # In other words, we only set change_role if the role needs to change, but we set
    # change_mtu always (if user wants to change the mtu)
    if ( %change_role ) {
        foreach my $port (keys %change_role) {
            if ( $change_mtu{$port} ) {
                change_port_config( server => $Server, port => $port, role => $change_role{$port}, mtu => $change_mtu{$port} );
                $change_mtu{$port} = undef;
            } else {
                change_port_config( server => $Server, port => $port, role => $change_role{$port} );
            }
        }
    }

    if ( %change_mtu ) {
        foreach my $port (keys %change_mtu) {
            if ( $change_mtu{$port} ) {
                change_port_config( server => $Server, port => $port, mtu => $change_mtu{$port} );
            }
        }
    }
}

## @name _determine_server_connection
## @description Return connection type of Server object.
##
## @synopsis 
## gx_utils::_determine_server_connection( server => $server );
##
##
sub _determine_server_connection_type {

    my %params = (
        server => undef,
        @_
    );

    if ( !$params{server} ) {
        $Log->error( die script_result( "Must supply server") );
    }

    my $Server = $params{server};

    my $connection_type = 'cli_connection';
    if ( $Server->{no_cli} ) {
        $connection_type = 'console_connection';
    }

    $Log->debug("_determine_server_connection_type server connection type = $connection_type" );

    return $connection_type;
}

## @name ::ports_by_role
## @description Returns hash of ports keyed by port role.
##
## @synopsis 
## ports_by_role( server => $Server );
##
## @arg server         Required: The node to act upon
##
##
sub ports_by_role {
    # This sub returns a hash of ports keyed by role
    my %params = (
        server => undef,
        @_
    );

    if ( !$params{server} ) {
        $Log->error( die script_result( "Must supply server") );
    }

    my $Server = $params{server};
    my $node = $Server->get_execution_server();
    my @ports;
    my @all_roles = qw( mgmt data cluster cluster-mgmt intercluster );
    my @cluster_ports;
    my @default_ports;
    my %by_role;

    my $conn_type = _determine_server_connection_type( server => $Server );
    my $ipspace_change = gx_utils::has_change( server => $Server, change => "ipspaces-ms4" );

    if(!$ipspace_change) {
        my $output = $Server->say_wrapper( type => $conn_type, cmd => "network port show -node $node -fields port,role" );
        my @output = split( /\n/, $output );
        my $fields = $output[0];
        my @fields = split( /\s+/, $fields );

        # remove last line which says how many rows are displayed
        pop( @output );

        $Log->debug("output = ".Dumper(\@output) );

        my @existing_ports;
        my $index = 0;
        foreach my $line (@output) {
            if ( $index > 1 ) {
                my @data = split( /\s+/, $line );
                my %data;
                my $ind = 0;
                foreach my $data (@data) {
                    $data{$fields[$ind]} = $data;
                    $ind++;
                }
                push( @existing_ports, { %data } );
            }
            $index++;
        }

        foreach my $port (@existing_ports) {
            my $role = $port->{role};

            if ( $role =~ /^node-mgmt$|^mgmt$/i ) {
                $role = "mgmt";
            } 

            if ( exists $by_role{$role} ) {
                if ( !grep( /^$port->{port}$/, @{$by_role{$role}} ) ) {
                    push( @{$by_role{$role}}, $port->{port} );
                }
            } else {
                $by_role{$role} = [ $port->{port} ];
            } 
        }
    } else {
        my %ports_by_ipspace = gx_utils::ports_by_ipspace( server => $Server);

        # Build the Role=>@(Ports) hash
        foreach my $role (@all_roles) {
            if( $role =~ /^cluster$/i ) {
                $by_role{$role} = [ @{$ports_by_ipspace{Cluster}} ];
            } else {
                $by_role{$role} = [ @{$ports_by_ipspace{Default}} ];
            }
        }
    }

    return %by_role;
}

## @name ports_by_ipspace
## @description Returns a hash keyed by ipspace names and contains array 
## of ports that belong to those ipspace names
##
## @synopsis
## ports_by_ipspace( server => $Server );
## 
## arg server          Required: the node to act upon
sub ports_by_ipspace {
    my %params = (
        @_
    );

    my @default_ipspaces = qw( Default Cluster );
    my %ports_by_ipspace;
    my $this_node        = NACL::C::Node->new(node => $params{server}->get_execution_server());


    my @ipspace_objs = NACL::CS::NetworkIpspace->fetch(
        command_interface => $this_node,
        filter            => { ipspace => 'Default|Cluster' },
        requested_fields  => ['ports'],
        allow_empty       => 1,
        apiset_must       => { connid => 'console' }
    );

    if (scalar(@ipspace_objs)) {
      foreach my $ipspace (@ipspace_objs) {
          my $type = ucfirst( $ipspace->ipspace() );
          my @ports = $ipspace->ports();
          my @ports_without_nodename = map { $_ =~ s/^.+://g ; $_ ;} @ports;
          push @{ $ports_by_ipspace{$type} }, @ports_without_nodename;
      }
    }

    # Make sure we have at least the default ipspaces
    foreach my $default_name (@default_ipspaces) {
        if ( !exists $ports_by_ipspace{$default_name} ) {
            die config_result("gx_utils: No entry was found for ipspace $default_name");
        }
    }

    return %ports_by_ipspace;
}

## @name change_port_config
## @description Returns hash of ports keyed by port role.
##
## @synopsis 
## change_port_config( server => $Server, port => $port, role => $role );
##
## @arg server         Required: The node to act upon
## @arg port           Required: The port to act upon
## @arg role           Optional: The role to be changed
## @arg mtu            Optional: The mtu to be changed
##
##
sub change_port_config {
    my %params = (
        server => undef,
        port => undef,
        role => undef,
        mtu => undef,
        @_
    );

    my $Server = $params{server};
    my $node = $Server->get_execution_server();
    my $conn_type = _determine_server_connection_type( server => $Server );

    my $ipspace_change = gx_utils::has_change( server => $Server, change => "ipspaces-ms4" );

    my $opts = "";

    if ( !($params{server} && $params{port} && ($params{role} || $params{mtu})) ) {
        $Log->comment( "change_port_config args = ".Dumper(\%params) );
        $Log->error( die script_result( "Must supply server, port and one of role|mtu"));
    }

    my @port_modify_options = qw ( role mtu );
    if ( $ipspace_change ) {
        @port_modify_options = qw( mtu );
    }


    foreach my $opt (@port_modify_options) {
        if ( $params{$opt} ) {
            $opts = $opts." -$opt $params{$opt}"; 
        }
    }

    if ( $opts ) {
        $Server->say_wrapper( type => $conn_type, cmd => "network port modify -node $node -port $params{port} ".$opts );
    }

    if( $ipspace_change && $params{role} ) {
        my %ports_by_ipspace = gx_utils::ports_by_ipspace( server => $Server);

        # Set the ports to their new ipspaecs
        if( $params{role} =~ /^cluster$/i ) {
            if( ! grep ($_ eq $params{port}, @{$ports_by_ipspace{Cluster}}) ) { # Change to cluster
                $Server->say_wrapper( type => $conn_type, cmd => "network ipspace assign-port -ipspace Cluster -node $node -port $params{port}" );
            }
        } else {
            if( ! grep ($_ eq $params{port}, @{$ports_by_ipspace{Default}}) ) { # Change to default
                $Server->say_wrapper( type => $conn_type, cmd => "network ipspace assign-port -ipspace Default -node $node -port $params{port}" );
            }
        }
    }
}    

## @name has_change
## @description Searches the uichanges list for a given change name
##
##  NOTE: I realize that Server.pm has this functionality, but you need to 
##        have a CLI connection, which we will not have if we are called
##        from gx_boot.thpl.
##
## @synopsis 
## change_port_config( server => $Server, port => $port, role => $role );
##
## @arg server         Required: The node to act upon
## @arg change         Required: The change to search for
##
##
sub has_change {
    my %params = (
        server => undef,
        change => undef,
        @_
    );

    if ( !($params{server} && $params{change}) ) {
        $Log->error( die script_result( "Must supply server and change") );
    }

    my $has_change = 0;
    my $Server = $params{server};
    my $node = $Server->get_execution_server();

    if ( !$uichanges{$node} ) {

        my $conn_type = _determine_server_connection_type( server => $Server );
        my $output;
        $Server->say_wrapper( type => $conn_type, cmd=>"set -confirmations off -rows 0 -privilege diag" );

        # Now get the known UI changes
        eval { $Server->say_wrapper( type => $conn_type, cmd=>"system uichanges -node $node", output=>\$output ); };
    
        if ($@ || $output =~ /ERROR/) {
            eval {
                $Server->say_wrapper( type => $conn_type, cmd=>"node uichanges -node $node", output=>\$output );
            };
        }

        if ($@ || $output =~ /ERROR: invalid argument \"-node\"/ ) {
            eval {
                $Server->say_wrapper( type => $conn_type, cmd=>"filer uichanges -filer $node", output=>\$output );
            };
        }

        # Remove extraneous characters
        $output =~ s/:/,/g;
        $output =~ s/ //g;
        $output =~ s/\r//g;
        $output =~ s/\n//g;

        my @changes = split(/,/, $output );
        # get rid of first 3 items, they are:
        # Node 
        # nodename
        # UIChanges 
        shift( @changes );
        shift( @changes );
        shift( @changes );

        $Log->debug("uichanges = |@changes|");
        @{$uichanges{$node}} = @changes;
    } 

    return grep( /^$params{change}$/, @{$uichanges{$node}} );

}    

## @name trim
## @description Removes whitespace from start and end of string
##
## @synopsis 
## $string = trim($string);
##
##
sub trim($) {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

## @name _set_if_name
## @description sets the interface name depending on the type of interface
##
## @synopsis
## my $name = _set_if_name( $node,$params{interface_type},$instance );
##
##
sub _set_if_name {
    my ($node,$interface,$instance) = @_;

    my $name;

    if ( $interface =~ /^cluster$/i ) {
        $name = "clus$instance";
    } elsif ( $interface =~ /^cluster_mgmt$/i ) {
        $name = "cluster_mgmt";
    } elsif ( $interface =~ /^mgmt$|^node-mgmt$/i ) {
        $name = "mgmt1";
    } else {
        $name = $node . "_" . (lc $interface) . "$instance";
    }

    return $name;
}


## @name check_initialization_permission
## @description Validate that the caller has permissions to execute initialization
##              procedures on the nodes we are acting upon.
##
## @synopsis
##    my $permitted = check_initialization_permission();
##
## This routine will inspect each Hostrec that Server knows about and check to see if
## the current user or testbed has permission to execute whatever is running against
## this node.  This routine is called by both gx_boot and deux_init and is primarily
## used to prevent accidental wipe and initialization of a node or cluster.  The routine
## will die with a CONFIG error if it fails to permit the caller to initialize the node.
##
## The routine inspects the Hostrec for the following params:
##
## INIT_ALLOW_NONE='This node should never be initialized!  Owner: dl-nst-lab@netapp.com'
##
## This parameter will cause no one to be allowed to initialize this node.  The string which
## the parameter is set to will be returned to the person and logged as a message from the
## owner of the node to the person or process trying to execute initialization.  It is
## highly encouraged that you include a dl or other email address to direct the user to contact
## you in the event they believe this block is incorrect.  If this is set to anything at all,
## even an empty string, no one will be allowed to initialize the nodes.  We do not inspect the
## allow list at all if this param is set.
##
## INIT_ALLOW_LIST='ctladmin,rtpeng_csiqa_schirato_1'
##
## This parameter provides a list of users or testbeds which are permitted to execute
## initialization of the nodes.  The list should be a , or : seperated list of usernames
## and testbed names or a combination of both of those.  If INIT_ALLOW_NONE is set that trumps
## this parameter and no one will be allowed to initialize the node.
##
## In the event that neither of these are set we assume the old default behavior that says "if
## someone gave me nodes to init - they knew what they were doing and I'll init them for them."
##
sub check_initialization_permission {
    my $call = call_trace(
        enter_message   => "Checking initialization permissions.",
        exit_message    => "Permission check complete.",
        show_eval_error => 0,
    );

    my %params = (
        server  => undef,
        @_,
    );

    my $node_names = delete $params{node_names};

    # make sure we have either a Server object or node names
    if ( !$params{server}
         && ( !defined $node_names
         or ref $node_names ne 'ARRAY'
         or !@$node_names )
       ) {
        die script_result("Must supply either server object or node names as array ref");
    }

    my $perm_type = $options->get(option => 'INITIALIZE_PERMISSION_TYPE');
    my @nodes = $params{server} ? $params{server}->servers() : @$node_names;
    my $called_by = $Tharn::Scriptname;

    # Handle simple permissions where we look at two 
    if ( $perm_type eq 'simple' ) {
        my $hostrecs    = $options->get(option => 'HOSTRECS');
        my $block_execution = 0;
        foreach my $node ( @nodes ) {
            # See if there is a generic block set in the nate_hosts file for this node that blocks
            # everyone from executing gx_boot on this node 
            if ( $hostrecs->{$node}{INIT_ALLOW_NONE} ) {
                $Log->warn("Node '$node' has had all execution of $called_by blocked via " .
                    "the INIT_ALLOW_NONE mechanism.\nPlease read the following notice from the owner " .
                    "of '$node':\n'$hostrecs->{$node}{INIT_ALLOW_NONE}'");
                $block_execution++;
                next; # skip to the next node - this one has been blocked
            }
            # See if the current user or testbed matches something in the allow list.  
            # This allows someone to limit the usage to a smaller subset of users or a testbed.
            if ( defined $hostrecs->{$node}{INIT_ALLOW_LIST} ) {
                my $allowed_list = $hostrecs->{$node}{INIT_ALLOW_LIST};
                $Log->trace("Found INIT_ALLOW_LIST for node '$node' set to '$allowed_list'");
                # We explicitly set these strings to something that will have a very low chance of matching
                # anything in the string that could be in the nate_hosts file.  If we can't determine the
                # testbed or username we really don't want to accidentally match what is in the nate_hosts
                # file.  But if they one of them does match we are going to let whatever is calling this code
                # execute.  So we don't die if we aren't in a testbed or if we can't for some reason
                # lookup the username of the person running the test calling this code in case the
                # other one does get set, and it matches what is in the nate_hosts file.
                my $testbed     = $options->get(option => 'TESTBED') || 'NOT-SET-ASSUMING-NOT-IN-A-TESTBED';
                my $username    = scalar(getpwuid($<)) || 'FAILED-TO-LOOKUP-CURRENT-USER-NAME';
                my $permitted   = 0;
                # Split the entry in the hostrec on , and : and look for a specific match for the
                # current user or the testbed
                foreach my $allowed ( split(/[,:]/, $allowed_list) ) {
                    $Log->debug("Inspecting allowed entry '$allowed' for permission to initialize.");
                    if ( $allowed =~ /^($testbed|$username)$/ ) {
                        $Log->comment("Node '$node' has an INIT_ALLOW_LIST set to '$allowed' " .
                            "which '$1' matched. Allowing $called_by to proceed based on this match.");
                        $permitted = 1;
                        last; # don't bother checking any other items from the variable
                    }
                }
                # If we didn't explicitly find a match and set $permitted to true - then deny permissions
                if ( !$permitted ) {
                    $block_execution++;
                    $Log->warn("Node '$node' has an INIT_ALLOW_LIST set to '$allowed_list' " .
                        "which does not match the current user '$username' or the current " .
                        "testbed '$testbed'. $called_by will not run against this node because  " .
                        "you do not have permission to initialize based on this parameter.  Please " .
                        "contact the owner of '$node' to determine how you should proceed."
                    );
                }
            }
        }
        # Now that we've gone through all the nodes, if we had _any_ that tripped the block - we break
        # out here
        if ( $block_execution ) {
            die config_result("One or more nodes have been blocked from " .
                "execution of $called_by via the INIT_ALLOW_* procedure.  Please check the logs to " .
                "determine who owns this node and why it was blocked.  Contact the owner before " .
                "trying to run $called_by again on these nodes.");
        }
    }
}

## @name _parse_hosts
## @description Parse a hostspec into a hash keyed by hostname
##
## @synopsis
##    my %hash = _parse_hosts();
##
sub _parse_hosts {
    my ($hostspecs) = @_;
    my @hostspec_pairs;
    while (1) {
        last if ($hostspecs =~ /\G$/cg); # exit loop on empty string
            my $oldpos = pos($hostspecs) || 0;
            my $hostid = Hostrec::parse_host_spec
                (
                 {},                       # partial_hostrec
                 \$hostspecs,              # contentref
                 0,                        # allow_connspec
                 0,                        # comment_allowed
                 undef,                    # hostid
                 "host spec '$hostspecs'", # source
                 1,                        # allow_many_hosts
                );
            my $newpos = pos($hostspecs);
            my $hostspec = substr($hostspecs,$oldpos,$newpos-$oldpos);
            $hostspec =~ s/[:,]$//g;
            push @hostspec_pairs, ($hostid => $hostspec);
    }
    return @hostspec_pairs;
}

## @name _apply_hostrec_overrides
## @description Make sure that all hostrec overrides provided by the NODES
##              parameter are applied to the Hostrec object for the nodes
##
## @synopsis
##    _apply_hostrec_overrides( $option_href );
##
sub _apply_hostrec_overrides {
    my ($option_ref)   = @_;
    my $whoami         = (caller(0))[3];

    foreach my $nodespec_string ( @{$option_ref->{NODES}} ) {
        my %nodespecs = _parse_hosts( $nodespec_string );
        foreach my $node (keys %nodespecs) {
            my $override_string = $nodespecs{$node};
            $override_string =~ s/{//g;
            $override_string =~ s/}//g;
            my @overrides = split(';', $nodespecs{$node});
            if ( scalar(@overrides) > 1 ) {
                shift @overrides;
                foreach my $override (@overrides) {
                    my @param = split('=', $override );
                    if ( scalar(@overrides) > 1 ) {
                        if ( !exists $option_ref->{HOSTRECS}{$node}->{$param[0]} ) {
                            $Log->comment("$whoami: $node ".$param[0]." = ".$param[1]." was missing, adding it");
                            $option_ref->{HOSTRECS}{$node}->{$param[0]} = $param[1];
                        }
                    }
                }
            }
        }
    }
}

1;
