#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/ha_blocks_filer.pm#1 $
#$DateTime: 2013/01/22 11:22:46 $ $Author: build $
#
#Copyright (c) 2002 Network Appliance, Inc.
#All rights reserved.
#
## @summary Filer error insertion library functions for Sledgehammer.
##
## @description
##   These test functions use various ONTAP Testpoints and features to insert errors in SCSI adapters, targets, disks, and luns during NATE testing
##
## @usage
##   use ha_blocks_filer;
##
## @keywords SAN FCP iSCSI Blocks
## @dependencies
##   blocks_misc package;
##
## @author pressley@netapp.com (Blocks QA)
## @status released
## @change 2002-02-27 johnm: created
## @change 2002-03-08 johnm: completed
## @change 2002-03-13 johnm: Changed following routines:<br>
##                                    ha_blocks_filer_scsi_cmd_delete<br>
##                                    ha_blocks_filer_scsi_cmd_timeout<br>
##                                    ha_blocks_filer_scsi_rsp_timeout<br>
##                                    ha_blocks_filer_coredump_delay<br><br>
##                           Added new routines and global variables:<br>
##                                 ha_blocks_filer_scsi_rsp_data<br>
##	                           ha_blocks_filer_scsi_clear<br>
##	                           $SCSI_STATUS_RESERVATION_CONFLCT<br>
##	                           $SCSI_STATUS_HARDWARE_ERROR<br>
##	                           $SCSI_STATUS_WRITE_ERROR_HARD<br>
##	                           $SCSI_STATUS_WRITE_ERROR_SOFT<br>
##	                           $SCSI_STATUS_READ_ERROR_HARD<br>
##	                           $SCSI_STATUS_READ_ERROR_SOFT<br>
## @change 2002-03-14 johnm: misc corrections to the documentation
## @change 2002-03-14 johnm: fixed $SCSI_STATUS_READ_ERROR_SOFT definition
## @change 2002-08-02 johnm: added new subroutines:
##	     ha_filer_cmd
##	     ha_filer_igroup
##	     ha_filer_volume
##	     ha_filer_lun
##	     ha_filer_create_qtree
##	     ha_filer_parse_lun_list
## @change 2005-01-24 pressley: Changed default timeout in ha_filer_cmd().
## @change 2005-06-01 archana: Updated author tag
##
use strict;

package ha_blocks_filer;

use vars ('@ISA','@EXPORT');

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
	     ha_blocks_filer_cp_panic
	     ha_blocks_filer_coredump_delay
	     ha_blocks_filer_giveback_delay
	     ha_blocks_filer_panic
	     ha_blocks_filer_cp
	     ha_blocks_filer_disk_scrub
	     ha_blocks_filer_vol_disk_get
         ha_blocks_filer_disk_pull
         ha_blocks_filer_disk_push
	     ha_blocks_filer_adapter_config
	     ha_blocks_filer_adapter_get
	     ha_blocks_filer_scsi_cmd_delete
	     ha_blocks_filer_scsi_cmd_timeout
	     ha_blocks_filer_scsi_rsp_timeout
	     ha_blocks_filer_scsi_rsp_data
	     ha_blocks_filer_scsi_clear
	     ha_blocks_filer_nvram_fail
	     ha_blocks_filer_scsitarget_fail
	     $SCSI_STATUS_RESERVATION_CONFLCT
	     $SCSI_STATUS_HARDWARE_ERROR
	     $SCSI_STATUS_WRITE_ERROR_HARD
	     $SCSI_STATUS_WRITE_ERROR_SOFT
	     $SCSI_STATUS_READ_ERROR_HARD
	     $SCSI_STATUS_READ_ERROR_SOFT
	     ha_filer_cmd
	     ha_filer_igroup
	     ha_filer_volume
	     ha_filer_lun
	     ha_filer_create_qtree
	     ha_filer_parse_lun_list
	     );

use blocks_misc;
use config_base;
use Tharn;

use blocks_filer qw(filer_lun_create filer_lun_destroy filer_lun_map filer_lun_offline);
use blocks_filer;

## @name ::ha_blocks_filer_cp_panic
## @description Set/clear a panic_before_CP Testpoint
## @synopsis 
## ha_blocks_filer_cp_panic {$conn, $action}
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "on" of "off"
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_cp_panic($fc, "on")) { exit 1 } # Error
## ha_blocks_filer_cp_panic($fc, "off");
##
sub ha_blocks_filer_cp_panic ($$) {
   my($conn, $action) = @_;

   if (!defined($conn) || !defined($action)) { return 0; }

   if ($action eq "on") {
       say($conn, "priv set test;tp panic wafl panic_before_CP 1;tp list",
	   "panic_before_CP",
	   '-ignore',"tp panic wafl panic_before_CP 1",
	   '-fatal', "Couldn't set panic_before_CP 1");
   } elsif ($action eq "off") {
       say($conn, "priv set test;tp disable wafl panic_before_CP");
   } else {
       print "ha_blocks_filer_cp_panic: invalid request $action\n";
       return 0;
   }
   return 1;
}

## @name ::ha_blocks_filer_coredump_delay
## @description Program a coredump delay Testpoint.
## The following (testpoint_names) can be set:
##
## (invalidate_core_header) -   
##        make the filer fail to invalide the core headers;        
##        invoke with panic or savecore -k.<br>
## (missing_header) - 
##         make the filer think there are no core headers present;
##        invoke panic or savecore, savecore -s.<br>
## (find_core) -               
##        make the filer claim there is a core on disk to recover;
##        invoke panic or savecore, savecore -s.<br>
## (find_core_again) -         
##        make the filer claim we've tried to save this core > 1 times; 
##        invoke panic or savecore, savecore -s.<br>
## (find_no_core) -     
##        make the filer claim there is no core;
##        invoke panic or savecore, savecore -s.<br>
## (assimilation) -    
##        simulates trying to dump core before raid lables have been read;
##        invoke with panic.<br>
## (no_disks) -                 
##        simulates a filer having no disks to dump to; 
##        invoke with panic.<br>
## (double_panic) -     
##        causes a real panic inside of dumpcore; 
##        invoke with panic.<br>
## (compress) -                 
##        causes a compression failure;
##        invoke a panic w/o enough disks for compression or savecore.<br>
## (compress_write) -   
##        causes a write to fail ;
##        invoke a panic w/o enough disks for compression or savecore.<br>
## (write_headers) -    
##        causes core header write to fail 
##        invoke panic<br>
## (spray_write) -             
##        causes a sprayed write to fail 
##        invoke a panic w/ a sprayed core<br>
## (no_disk_space) - 
##        simulate an out of space problem 
##        invoke a panic<br>
## (dump_write_buf) -
##        simple TESTPT to allow delay insertion into every core write<br>
## (start) -
##        simple TESTPT to allow delay insertion at start of dumpcore_internal<br>
## (dump_core_segment) -
##        simple TESTPT to allow delay insertion for an entire core segment<br>
##
## More at BURT 65805
## @synopsis 
## ha_blocks_filer_coredump_delay {$conn, $action, $testpoint, $count, $seconds}
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "on" or "off"
## @arg $testpoint Required: "" (null) for default or "(testpoint_name)"
## @arg $count Optional: number of times to hit
## @arg $seconds Optional: number of seconds to delay 
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_coredump_delay($fc, "on", "start", 1, 60)) { exit 1 }
## ha_blocks_filer_panic($fc);
##
## ha_blocks_filer_coredump_delay($fc, "off", "start");
##

my $_DUMPCORE_TPS="invalidate_core_header|missing_header|find_core|find_core_again|find_no_core|assimilation|no_disks|double_panic|compress|compress_write|write_headers|spray_write|no_disk_space|dump_write_buf|start|dump_core_segment";

sub ha_blocks_filer_coredump_delay ($$$;$;$;) {
   my($conn, $action, $testpoint, $count, $seconds) = @_;
   my $name="ha_blocks_filer_coredump_delay";

   if (!defined($conn) || !defined($action)) { return 0; }
   if (!($action = check_param($name, $action, "on|off"))) { return 0; }
   if (!defined($testpoint)) { $testpoint = "dump_write_buf"; }
   if (!($testpoint = check_param($name, $testpoint, $_DUMPCORE_TPS))) {
       return 0;
   }

   if ($action eq "on") {
       if (!($count = check_param($name, $count, "\\d+"))) { return 0; }
       if (!($seconds = check_param($name, $seconds, "\\d+"))) { return 0; }

#      Testpoint tp spin dumpcore <ID> <# times to hit> <seconds>
       say($conn, "priv set test;tp spin dumpcore $testpoint $count $seconds;tp list",
	   "$testpoint",
	   '-ignore',"tp spin dumpcore $testpoint $count $seconds",
	   '-fatal', "Couldn't set dumpcore $testpoint $count $seconds");
   } else {
       say($conn, "priv set test;tp disable dumpcore $testpoint");
   }

   return 1;
}

## @name ::ha_blocks_filer_giveback_delay
## @description Program a giveback delay, in seconds; on or off.
## @synopsis 
## ha_blocks_filer_giveback_delay {$conn, $action, $seconds}
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "on" or "off"
## @arg $seconds Optional: number of seconds to delay before giveback is complete
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_giveback_delay($fc, "on", 60)) { exit 1 } # Error
## say($fc, "cf giveback");
## ha_blocks_filer_giveback_delay($fc, "off", undef);
##
sub ha_blocks_filer_giveback_delay ($$;$) {
   my($conn, $action, $seconds) = @_;

   if (!defined($conn) || !defined($action)) { return 0; }

   if (!($action =~ /on|off/)) {
       print "ha_blocks_filer_coredump_delay: invalid request: \"$action\"\n";
       return 0;
   }

   if ($action eq "on") {
       if (!defined($seconds)) { 
	   print "ha_blocks_filer_giveback_delay: \$seconds is undefined!\n";
	   return 0;	   
       } else {
	   $_=$seconds;
	   if (!(/\d+/)) {
	       print "ha_blocks_filer_giveback_delay: \"$seconds\" is invalid!\n";
	       return 0;	   
	   }
	   $seconds =~ /\d+/;
       }

       say($conn, "priv set test;tp error fm giveback_reservation_delay 1 $seconds;tp list",
	   "giveback_reservation_delay",
	   '-ignore',"tp error fm giveback_reservation_delay 1 $seconds",
	   '-fatal', "Couldn't set fm giveback_reservation_delay $seconds");
   } else {
       say($conn, "priv set test;tp disable fm giveback_reservation_delay");
   }

   return 1;
}

## @name ::ha_blocks_filer_panic
## @description Panic the filer
## @synopsis 
## ha_blocks_filer_panic {$conn}
## @arg $conn Required: The filer connection to send this operation over
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_panic($fc)) { exit 1 } # Error filer didn't panic
##
sub ha_blocks_filer_panic ($) {
   my($conn) = @_;
   if (!defined($conn)) { return 0; }
   say($conn, "priv set diag;panic", -blind);
   return 1;
}

## @name ::ha_blocks_filer_cp
## @description Cause a wafl Check Point (CP) on the filer
## @synopsis 
## ha_blocks_filer_cp {$conn}
## @arg $conn Required: The filer connection to send this operation over
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_cp($fc)) { exit 1 } # Error filer didn't CP
##
sub ha_blocks_filer_cp ($) {
   my($conn) = @_;
   if (!defined($conn)) { return 0; }
   say($conn, "priv set diag;sync");
   return 1;
}

## @name ::ha_blocks_filer_disk_scrub
## @description Cause disk scrub
## @synopsis 
## ha_blocks_filer_disk_scrub {$conn, $action}
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "stop" or "start" 
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_disk_scrub($fc, "start")) { exit 1 } # Disk scrub didn't start
## sleep(30);
## ha_blocks_filer_disk_scrub($fc, "stop");
sub ha_blocks_filer_disk_scrub ($$) {
   my($conn, $action) = @_;
   if (!defined($conn) || !defined($action)) { return 0; }
   if (!($action eq "start") && !($action eq "stop")) {
       print "ha_blocks_filer_disk_scrub: $action is not a valid parameter!\n";
       return 0;
   }
   say($conn, "priv set diag; disk scrub $action");
   return 1;
}

###############################################################################

sub get_fcp_config_list ($) {
   my($conn) = @_;
   my($match,$before,$after) = say($conn, "fcp config;echo DoNE",
				   "DoNE",
				   '-ignore',"echo DoNE",
				   '-fatal', "Couldn't get fcp config output!");   
   return ($before);
}

sub get_sysconfig_r_list ($) {
   my($conn) = @_;
   my($match,$before,$after) = say($conn, "sysconfig -r;echo DoNE",
				   "DoNE",
				   '-ignore',"echo DoNE",
				   '-fatal', "Couldn't get sysconfig -r");   
   return ($before);
}

## @name ::ha_blocks_filer_vol_disk_get
## @description Returns the name of the first data disk for the specified volume, else undef. 
## @synopsis 
## ha_blocks_filer_vol_disk_get {$conn, $vol}
## @arg $conn Required: The filer connection to send this operation over
## @arg $vol Required: The volume name to search for (e.g. /vol/vol1)
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## my ($disk);
## if (!defined $disk = ha_blocks_filer_vol_disk_get($fc, "vol1")) { 
##      print "No disk found for vol1\n";
## } 
sub ha_blocks_filer_vol_disk_get ($$) {
   my($conn, $vol) = @_;

   if (!defined($conn) || !defined($vol)) { return undef; }

   my($list) = get_sysconfig_r_list($conn);   
   my(@info) = split(/\n/, $list);
   my($disk, $line, $index, $i);

   for ($index=0; $index<@info; $index++) {
       $line = $info[$index];
       if ($line =~ /^Volume\s+$vol/) {
	   print "$line\n";
	   for ($i=$index; $i<@info; $i++) {
	       $line = $info[$i];
	       if ($line =~ /^\s+data\s+(\w+\.\w+)/) {
		   $disk = $1;
		   print "Found disk $disk\n";
		   return $disk;
	       }
	   }
       }
   }
   
   printf "Volume $vol not found!\n";
   return undef;
}

## @name ::ha_blocks_filer_disk_pull
## @description Fail the specified disk with the `disk fail' CLI. If no disk is passed, the first data disk from the firt raid set on the filer is used.  Returns the disk (adapter.disk) number of the failed disk. 
## @synopsis 
## ha_blocks_filer_disk_pull {$conn, $disk}
## @arg $conn Required: The filer connection to send this operation over
## @arg $disk Optional: The disk to fail
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## my ($disk);
## if (!defined $disk = ha_blocks_filer_disk_pull($fc, 8a.25)) { 
##      print "disk 8a.25 not found\n";
## } 
##
## if (defined $disk = ha_blocks_filer_disk_pull($fc)) {
##      print "found or failed disk $disk\n";
## }
##
sub ha_blocks_filer_disk_pull ($$) {
   my($conn, $disk) = @_;

   if (!defined($conn)) { return 0; }

   my($list) = get_sysconfig_r_list($conn);   
   my(@info) = split(/\n/, $list);
   my ($line);

   if (!defined($disk)) {
       foreach $line (@info) {
	   if ($line =~ /^failed\s+(\w+\.\w+)/) {
	       print "Found disk $1 in the failed state, returning $1\n";
	       $disk = $1;
	       return $disk;
	   }
       }

       foreach $line (@info) {
	   if ($line =~ /^\s+data\s+(\w+\.\w+)/) {
	       $disk = $1;
	       last;
	   }
       }
   } else {
       my ($found);
       foreach $line (@info) {
	   if ($line =~ /^\s+data\s+$disk/) {
	       $found = 1;
	       last;
	   }
       }
       if (!defined($found)) {
	   return undef;
       }
   }

   print "Failing disk $disk\n";

   say($conn, "priv set diag;disk fail $disk",
       -noprompt, '-wait',"Really fail disk",
       '-fatal', "Couldn't fail disk $disk");

   say($conn, "y", -noprompt, '-wait', "raid.disk.unload.done",
       '-fatal', "disk fail $disk never completed unload");

   return $disk;

}

## @name ::ha_blocks_filer_disk_push
## @description Un-fail the specified disk with the `disk unfail' CLI. If no disk is passed, the first failed disk from the firt raid set on the filer is used.  Returns the disk (adapter.disk) number of the unfailed disk. 
## @synopsis 
## ha_blocks_filer_disk_push {$conn, $disk}
## @arg $conn Required: The filer connection to send this operation over
## @arg $disk Optional: The disk to unfail
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## my ($disk);
## if (!defined $disk = ha_blocks_filer_disk_push($fc, 8a.25)) { 
##      print "disk 8a.25 not found\n";
## } 
##
## if (defined $disk = ha_blocks_filer_disk_push($fc)) {
##      print "unfailed disk $disk\n";
## }
##
sub ha_blocks_filer_disk_push ($$) {
   my($conn, $disk) = @_;

   if (!defined($conn)) { return 0; }

   my($list) = get_sysconfig_r_list($conn);   
   my(@info) = split(/\n/, $list);
   my ($line);

   if (!defined($disk)) {
       foreach my $line (@info) {
	   if ($line =~ /^failed\s+(\w+\.\w+)/) {
	       print "Found disk $1 in the failed state\n";
	       $disk = $1;
	       last;
	   }
       }
   } else {
       my ($found);
       foreach my $line (@info) {
	   if ($line =~ /^failed\s+$disk/) {
	       $found = 1;
	       last;
	   }
       }
       if (!defined($found)) {
	   return 0;
       }
   }

   print "Unfailing disk $disk\n";

   say($conn, "priv set diag;disk unfail $disk");

   return 1;

}

sub build_fiji_cmd ($$$$$$) {
   my($cmdstr, $lun, $count, $cmd, $parse, $code) = @_;
   my ($fiji) = "\"";

   if (!defined($cmdstr)) { return undef; }

   if (defined($cmd)) {
       $_=$cmd;
       if (/$parse/) {
	   $cmd =~ /$parse/;
	   $cmd =~ s/read/(scsi_command SCSI_READ)/;
	   $cmd =~ s/read10/(scsi_command SCSI_READ_X)/;
	   $cmd =~ s/write/(scsi_command SCSI_WRITE)/;
	   $cmd =~ s/write10/(scsi_command SCSI_WRITE_X)/;
	   if ($cmd eq "all") {undef $cmd};
       } elsif (/0x\w+/) {
	   $cmd =~ /0x\w+/;
       } else {
	   print "build_fiji_cmd: invalid scsi cmd: \"$_\"\n";
	   return undef;
       }
   }

   if (defined($lun)) {
       if ($lun =~ /all|\d+/) {
#	   if ($lun eq "all") {$lun=16384;}
	   if ($lun eq "all") {undef $lun;}
       } else {
	   print "build_fiji_cmd: invalid lun: \"$lun\"\n";
	   return undef;
       }
   }

   if (defined($count)) { 
       if (!($count =~ /\d+/)) {
	   print "build_fiji_cmd: invalid count: \"$count\"\n";
	   return undef;
       }
   }

   if (defined($count)) { $fiji = $fiji."repeat ".$count." "; }

   $fiji = $fiji."(".$cmdstr." ";

   if (defined($lun) || defined($cmd)) {
       $fiji = $fiji."(&& ";
   } else { 
       $fiji = $fiji."(true) ";
   }

   if (defined($lun)) { $fiji = $fiji."(== \@lun ".$lun.")";}
  
   if (defined($cmd)) { $fiji = $fiji."(== \@cmd ".$cmd.")";}

   if (defined($lun) || defined($cmd)) { $fiji = $fiji.") ";}

   if (!defined($code)) {
       $fiji = $fiji."(return 1))\"";
   } else {
       $fiji = $fiji."(return $code))\"";
   }

   print "\$fiji = $fiji\n";

   return $fiji;

}

## @name ::ha_blocks_filer_scsi_clear
## @description Clears FIJI Testpoint.
## These FIJI filters can only be set one at a time.  Running
## one of these filters will clear any other out.  However
## when you are done testing you should clear the filter
## to avoid affecting another test that does not expect a filter
##
## @synopsis 
## ha_blocks_filer_scsi_clear {$conn}
## @arg $conn Required: The filer connection to send this operation over
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## ha_blocks_filer_scsi_rsp_data($fc, 1, 60, "write", $SCSI_STATUS_RESERVATION_CONFLICT);
## ha_blocks_filer_scsi_clear($fc);
##
sub ha_blocks_filer_scsi_clear ($) {
   my($conn) = @_;
   my ($name) = "ha_blocks_filer_scsi_clear";

   if (!defined($conn)) { return 0; }

   say($conn, "fiji clear",
       "filter cleared",
       '-ignore',"fiji clear",
       '-fatal', "Couldn't clear fiji filters");
   
   return 1;
}

## @name ::ha_blocks_filer_scsi_rsp_data
## @description Returns the specified SCSI status, sense key, additional
## sense code and additional sense code qualifier for the given command.
## SCSI status and sense data is represented by a 32 bit word passed via
## the $scsi_status parameter.  These 32 bits are encoded as follows:
##
## o Byte 1, low-order 7 bits: SCSI status<br>
## o Byte 2: SKEY<br>
## o Byte 3: ASC<br>
## o Byte 4: ASCQ
##
## For example, to encode: Status=0x02, SKEY=0x11, ASC=0x22, ASCQ=0x33, 
## you would pass "0x33221102"
##
## For convenience, this perl module exports the fillowing global variables:
##
## $SCSI_STATUS_RESERVATION_CONFLCT<br>
## $SCSI_STATUS_HARDWARE_ERROR<br>
## $SCSI_STATUS_WRITE_ERROR_HARD<br>
## $SCSI_STATUS_WRITE_ERROR_SOFT<br>
## $SCSI_STATUS_READ_ERROR_HARD<br>
## $SCSI_STATUS_READ_ERROR_SOFT<br>
##
## These definitions can be used in the $scsi_status field.
##
## @synopsis 
## ha_blocks_filer_scsi_rsp_data {$conn, $lun, $count, $scsi_cmd, $scsi_status}
## @arg $conn Required: The filer connection to send this operation over
## @arg $lun Required: The filer lun (number) to use, or "all", of `undef'
## @arg $count Required: Then number of commands to delete, or `undef'
## @arg $scsi_cmd Required: The type of command: "read", "write", "all", the hexidecimal scsi opcode, or `undef'
## @arg $scsi_status Required: A predefined `SCSI_STATUS' scalar, or the hexadecimal encoded SCSI Status and Sense data.
##
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## ha_blocks_filer_scsi_rsp_data($fc, 1, 60, "write", $SCSI_STATUS_WRITE_ERR);
##
sub ha_blocks_filer_scsi_rsp_data ($$$$$) {
   my($conn, $lun, $count, $cmd, $code) = @_;
   my ($fiji_cmd) = "scsitarget.cmd.rsp.code";
   my ($name) = "ha_blocks_filer_scsi_rsp_data";
   my ($fiji);

   if (!defined($conn) || !defined($code)) { return 0; }

   if (!defined($fiji = build_fiji_cmd($fiji_cmd,
				       $lun, $count, $cmd,
				       "read|write|all", $code))) {
       print "$name: ($fiji_cmd) failed!\n";
       return 0;
   }
   my $mfiji = quotemeta($fiji);

   say($conn, "fiji clear;fiji filter $fiji;fiji dumpfilter",
       "$fiji_cmd",
       '-ignore',"$mfiji",
       '-fatal', "Couldn't set fiji filter: $fiji");

   return 1;
}

## @name ::ha_blocks_filer_scsi_cmd_delete
## @description Cause command timeouts on a filer lun - one or more SCSI CDBs are received, but removed from Lun queue or dropped
## @synopsis 
## ha_blocks_filer_scsi_cmd_delete {$conn, $lun, $count, $scsi_cmd}
## @arg $conn Required: The filer connection to send this operation over
## @arg $lun Required: The filer lun (number) to use, or "all", of `undef'
## @arg $count Required: Then number of commands to delete, or `undef'
## @arg $scsi_cmd Required: The type of command: "read", "write", "all", the hexidecimal scsi opcode, or `undef'
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_scsi_cmd_delete($fc, 1, 60, "all")) { exit 1 } # Error
##
## if (!ha_blocks_filer_scsi_cmd_delete($fc, 0, 1, 0xa2)) { exit 1 }
##
## if (!ha_blocks_filer_scsi_cmd_delete($fc, undef, undef, undef)) { exit 1 }
##
sub ha_blocks_filer_scsi_cmd_delete ($$$$) {
   my($conn, $lun, $count, $cmd) = @_;
   my ($fiji_cmd) = "scsitarget.cmd.delete";
   my ($name) = "ha_blocks_filer_scsi_cmd_delete";
   my ($fiji);

   if (!defined($conn)) { return 0; }

   if (!defined($fiji = build_fiji_cmd($fiji_cmd,
				       $lun, $count, $cmd,
				       "read|write|all", undef))) {
       print "$name: ($fiji_cmd) failed!\n";
       return 0;
   }
   my $mfiji = quotemeta($fiji);

   say($conn, "fiji clear;fiji filter $fiji;fiji dumpfilter",
       "$fiji_cmd",
       '-ignore',"$mfiji",
       '-fatal', "Couldn't set fiji filter: $fiji");

   return 1;
}

## @name ::ha_blocks_filer_scsi_cmd_timeout
## @description Cause command timeouts on a filer lun - one or more SCSI CDBs are received, but the target never goes to data or status phase..
## @synopsis 
## ha_blocks_filer_scsi_cmd_timeout {$conn, $lun, $count, $scsi_cmd}
## @arg $conn Required: The filer connection to send this operation over
## @arg $lun Required: The filer lun (number) to use, or "all", of `undef'
## @arg $count Required: Then number of commands to delete, or `undef'
## @arg $scsi_cmd Required: The type of command: "read", "write", "all", the hexidecimal scsi opcode, or `undef'
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_scsi_cmd_timeout($fc, 1, 60, "all")) { exit 1 }
##
## if (!ha_blocks_filer_scsi_cmd_timeout($fc, 0, 1, 0xa2)) { exit 1 }
##
## if (!ha_blocks_filer_scsi_cmd_timeout($fc, undef, undef, undef)) { exit 1 }
##
sub ha_blocks_filer_scsi_cmd_timeout ($$$$) {
   my($conn, $lun, $count, $cmd) = @_;
   my ($fiji_cmd) = "scsitarget.cmd.timeout";
   my ($name) = "ha_blocks_filer_scsi_cmd_timeout";
   my ($fiji);

   if (!defined($conn)) { return 0; }

   if (!defined($fiji = build_fiji_cmd($fiji_cmd,
				       $lun, $count, $cmd,
				       "read|write|all", undef))) {
       print "$name: ($fiji_cmd) failed!\n";
       return 0;
   }

   my $mfiji = quotemeta($fiji);

   say($conn, "fiji clear;fiji filter $fiji;fiji dumpfilter",
       "$fiji_cmd",
       '-ignore',"$mfiji",
       '-fatal', "Couldn't set fiji filter: $fiji");

   return 1;
}

## @name ::ha_blocks_filer_scsi_rsp_timeout
## @description Cause command timeouts on a filer lun - Receives one or more SCSI CDBs and complete data phases but never send SCSI status 
## @synopsis 
## ha_blocks_filer_scsi_rsp_timeout {$conn, $lun, $count, $scsi_cmd}
## @arg $conn Required: The filer connection to send this operation over
## @arg $lun Required: The filer lun (number) to use, or "all", of `undef'
## @arg $count Required: Then number of commands to delete, or `undef'
## @arg $scsi_cmd Required: The type of command: "read", "write", "all", the hexadecimal scsi command opcode, or `undef'
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_scsi_rsp_timeout($fc, 1, 60, "all")) { exit 1 }
## ha_blocks_filer_scsi_rsp_timeout($fc, 1, 60, undef);
##
## if (!ha_blocks_filer_scsi_rsp_timeout($fc, 0, 1, 0xa2)) { exit 1 }
##
## if (!ha_blocks_filer_scsi_rsp_timeout($fc, undef, undef, undef)) { exit 1 }
##
sub ha_blocks_filer_scsi_rsp_timeout ($$$$) {
   my($conn, $lun, $count, $cmd) = @_;
   my ($fiji_cmd) = "scsitarget.cmd.rsp.timeout";
   my ($name) = "ha_blocks_filer_scsi_rsp_timeout";
   my ($fiji);

   if (!defined($conn)) { return 0; }

   if (!defined($fiji = build_fiji_cmd($fiji_cmd,
				       $lun, $count, $cmd,
				       "read|write|all", undef))) {
       print "$name: ($fiji_cmd) failed!\n";
       return 0;
   }

   my $mfiji = quotemeta($fiji);

   say($conn, "fiji clear;fiji filter $fiji;fiji dumpfilter",
       "$fiji_cmd",
       '-ignore',"$mfiji",
       '-fatal', "Couldn't set fiji filter: $fiji");

   return 1;
}

## @name ::ha_blocks_filer_adapter_config
## @description Control HBA enable/disable/reset
## @synopsis 
## ha_blocks_filer_adapter_config {$conn, $adapter, $action}
## @arg $conn Required: The filer connection to send this operation over
## @arg $adapter Required: The filer adapter to reset/enable/disable
## @arg $action Required: "up" or "down" or "reset"
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_adapter_config($fc, "5a", "down")) { exit 1 } # Error
## if (!ha_blocks_filer_adapter_config($fc, "8b", "reset")) { exit 1 } # Error
 
sub ha_blocks_filer_adapter_config ($$$) {
    my($conn, $adapter, $action) = @_;
    my($name) = "ha_block_filer_adapter_config";
    
    if (!defined($conn) || !defined($adapter) || !defined($action)) {
	return 0; 
    }

    if (!($_ = check_param($name, $action, "up|down|reset"))) {
	return 0;
    }

    if (/down/) {
        say($conn, "fcp config $adapter down");
    } elsif (/up/) {
        say($conn, "fcp config $adapter up");
    } else {
        say($conn, "fcp config $adapter down");
        say($conn, "fcp config $adapter up");
    }

    return 1;
}

sub ha_blocks_filer_adapter_get ($$) {
    my($conn, $type) = @_;
    my($name) = "ha_block_filer_adapter_get";
    
    if (!defined($conn) || !defined($type)) { return undef; }
    
    if (!($type = check_param($name, $type, "fcp"))) { return undef; }
    
    my($list) = get_fcp_config_list($conn);   
    my(@info) = split(/[\w+\s+]partner\s+adapter\s+\w+\n/, $list);
    my($adapter, $line, $index, $i);
    
    foreach $line (@info) {
	print "\$line = $line\n";
    }
    
    return undef;
}

## @name ::ha_blocks_filer_nvram_fail
## @description Cause nvram to fail and have to recompute parity on reboot
## @synopsis 
## ha_blocks_filer_nvram_fail {$conn, $fail_type}
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "jettison" or "disable" or "trash"
## @example
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_nvram_fail($fc, "jettison")) { exit 1 } # Error
##
##*** WARNING *** This command will corrupt NVRAM and force a reboot.
##                Do you still want to execute this command? (yes/no)
##
##nvram_munge: Forcing a reboot.
##Expect dirty shutdown without NVRAM protection message 
##and parity recompute upon reboot! 


sub ha_blocks_filer_nvram_fail ($$) {
   my($conn, $fail_type) = @_;

   if (!defined($conn) || !defined($fail_type)) { return 0; }

   say($conn, "priv set test;platform_error nvram $fail_type", -blind);

   say($conn, "yes", -blind);
   say($conn, "", -swallow);

   return 1;
}

## @name ::ha_blocks_filer_scsitarget_fail
## @description Cause scsi target to fail with resource errors
## @synopsis 
## ha_blocks_filer_scsitarget_fail {$conn, $fail_type, $num_failures} 
## @arg $conn Required: The filer connection to send this operation over
## @arg $action Required: "fail_alloc_cmdblk" or "fail_alloc_async_msg" 
##                        "fail_wafl_steal"
## @arg $num_failures Optional: Number of these failure to do in a row
##                              Default = 1
## @example to fail the next 10 scsi command block allocations
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## if (!ha_blocks_filer_scsitarget_fail($fc, "fail_alloc_cmdblk", 10)) { exit 1 } # Error
##
sub ha_blocks_filer_scsitarget_fail ($$$) {
   my($conn, $fail_type, $num_failures) = @_;

   if (!defined($conn) || !defined($fail_type)) { return 0; }

   if (!defined($num_failures)) { $num_failures = 1; }

   say($conn, "priv set test;tp error scsitarget $fail_type $num_failures 1");

   return 1;
}

## @name :: ha_filer_cmd
## @description Execute command on the filer and return output.
## 
## @synopsis 
## ha_filer_cmd {$conn, $cmd, $timeout} 
##
## @arg $conn Required: The filer connection to send this operation over
## @arg $cmd Required: File CLI command to send
## @arg $timeout Optional: Timeout in seconds (defaults to 125) 
##
## @example 
## my $fa = host($FILER);
## my $fc = connect ($FILER_CONN);
## my $output = ha_filer_cmd($fc, "lun show", 125);
##
sub ha_filer_cmd($$;$) {
    my ($conn, $cmd, $timeout) = @_;
    my $name="ha_filer_cmd";
    my $output = undef;
    my $metacmd = quotemeta($cmd);

    tstlog(-fatal => "$name: hosttype not filer!")
	unless ($conn->{hostp}->{hosttype} =~ /filer/);

    $timeout = 125 unless($timeout);

    if ($timeout =~ /blind/) {
	say($conn, "$cmd", '-blind');
    } else {
	(undef, undef, $output) =
	    say($conn, "$cmd", '-timeout', $timeout, 
		'-ignore', "$metacmd\r\n",
		'-fatal', "$name: \"$cmd\" failed on $conn->{hostp}->{id}");

#	print "$name: \$output = $output\n";

    }

    return ($output);
}

## @name :: ha_filer_igroup
## @description Execute command on the filer and return output.
## 
## 
## @synopsis 
## ha_filer_igroup {$action, $conn, $igroup, $param, $wwpns} 
##
## @arg $action Required: "create" | "destroy" | "replace"
## @arg $conn Required: The filer connection to send this operation over
## @arg $igroup Required: The igroup name
## @arg $param Optional: igroup command aguments (required w/create|replace)
## @arg $wwpns Optional: pointer to an array of WWPNs (required w/create|replace)
##
## @example 
## my $IGROUP_TYPE="-f -t solaris";
## my @wwpns = ("10000000c9291e76", "10000000c92db0fd");
## my $fc = connect ("lettuce"."#telnet");
## ha_filer_igroup("create", $fc, "dumbo", $IGROUP_TYPE, \@wwpns);
##
##  Creates FCP igroup named `dumbo' of type solaris with the wwpns:
##  "10000000c9291e76" and "10000000c92db0fd"
##
sub ha_filer_igroup($$$;$$) {
    my ($action, $fc, $igroup, $param, $wwpns) = @_;
    my $name="ha_filer_igroup";
    my ($cmd, $wwpn, $output, @wwpns);

    ExitScriptError( "$name: Invalid parameter passed ($action)")
	if (!($action = check_param($name, $action, "create|destroy|replace")));

    if ($action =~ /create|replace/) {
	tstlog(-fatal => "$name: missing paramter (\$param)")
	    unless (defined($param) && defined($wwpns));
	@wwpns = @$wwpns;
    }

  again:

    $output = ha_filer_cmd($fc, "igroup show $igroup");
    if ($output =~ /No such initiator group exists/) {	

	if ($action =~ /destroy/) {
	    tstlog(-comment => "$name: $igroup doesn't exist");
	    return;
	}

	$cmd = "igroup create ".$param." ".$igroup;
	foreach $wwpn (@wwpns) {
	    $wwpn = wwn_set($wwpn);
	    $cmd .= " ".$wwpn;
	}

	tstlog(-fatal => "$name: \"$cmd\" failed\n \$output = $output\n")
	    if ($output = ha_filer_cmd($fc, "$cmd"));
	
	tstlog(-comment => "$name: $igroup created");

    } else {

	if ($action =~ /destroy|replace/) {
	    $cmd = "igroup destroy ".$igroup;

	    tstlog(-fatal => "$name: \"$cmd\" failed\n \$output = $output\n")
		if ($output = ha_filer_cmd($fc, "$cmd"));

	    tstlog(-comment => "$name: $igroup destroyed");

	    return if ($action =~ /destroy/);
	    goto again;
	}

	# asked to create but one already exists, verify wwpns match.
	foreach $wwpn (@wwpns) {
	    $wwpn = wwn_set($wwpn);
	    tstlog(-fatal =>
		   "$name: no $wwpn in igroup $igroup\n \$output = $output\n")
		unless (string_search($output, "find", $wwpn));
	}

	tstlog(-comment => "$name: $igroup exists");

    }

}

## @name :: ha_filer_volume
## @description Creates and destroys volumes on filer.  Exits on error.  
## Returns nothing (undef) on success.
## 
## @synopsis 
## ha_filer_volume {$action, $conn, $volname, $option, $opt_val, $timeout} 
##
## @arg $action Required: "create" | "destroy" | "replace" | "set"
## @arg $conn Required: The filer connection to send this operation over
## @arg $volname Required: The volume name (e.g. vol1)
## @arg $option Optional: The volume option to set (required w/set)
## @arg $opt_val Optional: The volume option value to set (required w/set)
## @arg $timeout Optional: Timeout in seconds (defaults to 6000) 
##
## @example 
## my $fa = host($CLIENT);
## my $fc = connect ($CLIENT_CONN);
## ha_filer_volume("replace", $fc, "vol1");
## ha_filer_volume("set", $fc, "vol1", "space_reservations", "on");
##
sub ha_filer_volume($$$;$;$;$) {
    my ($action, $fc, $volname, $option, $opt_state, $timeout) = @_;
    my $name="ha_filer_volume";
    my ($line, $output);
    my ($end, $cmd);

    my $valid_options ="nosnap|nosnapdir|minra|no_atime_update|raidsize|nvfail|snapmirrored|resyncsnaptime|create_ucode|convert_ucode|maxdirsize|fs_size_fixed|space_reservations";

    my $valid_actions = "verify|create|set|destroy|replace";

    $option=0 unless($option);
    $opt_state="on" unless($opt_state);
    $timeout = 6000 unless ($timeout);

    ExitScriptError( "$name: Invalid parameter passed ($action)")
	if (!($action = check_param($name, $action, "$valid_actions")));

    if ($option) {
	tstlog(-fatal => "$name: Invalid paramter passed ($option)")
	    if (!($option = check_param($name, $option, "$valid_options")));

    }

  again:

    if ($line = string_search(ha_filer_cmd($fc, "vol status"), "find", $volname)) {

	my ($vol, $state, $status, $opts) = 
	    $line =~ /^\s+(\w+)\s+(\w+)\s+(\w+)/;

#	print "\$vol = $vol, \$state = $state, \$status = $status\n";

	if ($action =~ /replace|destroy/) {

	    if ($status =~ /initializing|growing/) {
		if ($action =~ /replace/) {
		    $action = "create";
		    goto again;
		}
	    }

	    tstlog(-fatal => "$name: $volname status is $status!")
		unless ($status =~ /normal/);

	    $end = quotemeta("Volume \'".$volname."\' is now offline");
	    $cmd = "vol offline ".$volname;

	    say($fc, "$cmd", '-wait', "$end", '-timeout', 10,
		'-fatal', "$name: \"$cmd\" failed!")
		unless ($state =~ /offline/);

	    $end = quotemeta("Volume \'".$volname."\' destroyed");
	    $cmd = "vol destroy ".$volname." -f";

	    say($fc, "$cmd", '-wait', "$end", '-timeout', 10,
		'-fatal', "$name: \"$cmd\" failed!");

	    tstlog(-comment => "$name: $volname destroyed");
	    
	    return if ($action =~ /destroy/);
	    goto again;

	} else {

	    unless ($state =~ /online/) {

		$end = quotemeta("Volume \'".$volname."\' is now online");
		$cmd = "vol online ".$volname." -f";

		say($fc, "$cmd", '-wait', "$end", '-timeout', 10,
		    '-fatal', "$name: \"$cmd\" failed!");

		tstlog(-comment => "$name: set $volname online");

	    }

	    if ($option) {
		
		$cmd = "vol options ".$volname." ".$option." ".$opt_state;
		$output = ha_filer_cmd($fc, "$cmd");
		tstlog(-fatal => "$name: \"$cmd\" failed\n \$output = $output\n")
		    if ($output && $output =~ /vol options:/);

		tstlog(-comment => "$name: set $volname $option $opt_state");

	    }

	    tstlog(-comment => "$name: $volname $action $option");
	    return 1;
	}

    } else {

	if ($action =~ /destroy|set|verify/) {
	    tstlog(-comment => "$name: $volname doesn't exit");
	    return 0;
	}

	my $count;

	$count=string_search(ha_filer_cmd($fc, "sysconfig -r"), "count", "spare");
    
	tstlog(-fatal => "$name: only $count spare disks")
	    unless ($count >= 3);

	$count--;

	# Create the Volume and wait for it to grow to normal status.

	$line = string_search(ha_filer_cmd($fc, "vol status"), "find", $volname);
	$cmd = "vol create /vol/".$volname." -f $count";

	say($fc, "$cmd", '-timeout', 10, '-fatal', "$name: \"$cmd\" failed!")
	    unless ($line);

	my $tot_runlength = time() + $timeout;
	while (time() <= $tot_runlength) {
	    $line = string_search(ha_filer_cmd($fc, "vol status"),
				 "find", $volname);

	    tstlog(-fatal => "$name: $volname not created!") unless ($line);

	    my (undef, undef, $status) = $line =~ /^\s+(\w+)\s+(\w+)\s+(\w+)/;

	    last if ($status =~ /normal/);

	    sleep 30;
	}

	goto again;

    }
}

## @name :: ha_filer_lun
## @description Creates and destroys luns on filer.  Exits on error.  
## Returns nothing (undef).
##
## Lun information is passed via an array of scalar strings (the
## $lun_list).  Each scaler string contains a comma separated list with
## three values: <lun-id>, <size> and <type>. These values are used
## together with the $lun_prefix and $igroup information to build a
## lun-path using the convention:
##
##  <lun-path> = {$lun_prefix}{$igroup}_lun<lun-id>
##
## This information is passed directly to the Filer lun create CLI:
##
##  lun create -s <size> -t <type> <lun-path>
##
## The <lun_path> and $igroup information are then used to map the lun
## to the initiator-group with the lun map CLI:
##
##  lun map <lun-path> $igroup <lun-id>
##
## @synopsis 
## ha_filer_lun {$action, $conn, $igroup, $lun_prefix, $lun_list} 
##
## @arg $action Required: "create" | "destroy" | "replace"
## @arg $conn Required: The filer connection to send this operation over
## @arg $lun_prefix Required: The vlun directory path (e.g. /vol/vol1/q0)
## @arg $lun_list Required: Address of lun_list array
##
## @example 
## my $fc = connect ("lettuce"."#telnet");
## my $LUN_LIST="0,2g,solaris 1,1g,default 2,30g,solaris 10,250m,windows";
## my @lun_list = split(/ /, $LUN_LIST);
## ha_filer_lun("replace", $fc, "bonito", "/vol/vol1/", \@lun_list);
##
## This will create 4 luns on filer lettuce, mapped to igroup bonito:
##
##  /vol/vol1/bonito_lun0      (map=0, size=2g, type=solaris)
##  /vol/vol1/bonito_lun1      (map=1, size=1g, type=default)
##  /vol/vol1/bonito_lun2      (map=2, size=30g, type=solaris)
##  /vol/vol1/bonito_lun10     (map=10, size=250m, type=windows)
##
sub ha_filer_lun($$$$;$) {
    my ($action, $fc, $igroup, $lun_prefix, $lun_list) = @_;
    my $name="ha_filer_lun";
    my ($cmd, $output, $size, $line, $end, @lun_list);
    my (undef, undef, $volname, $qtree) = split(/\//, $lun_prefix);

    my $valid_actions = "create|destroy|replace";

    ExitScriptError( "$name: Invalid parameter passed ($action)")
	if (!($action = check_param($name, $action, "$valid_actions")));

    ExitScriptError( "$name: lun prefix \"$lun_prefix\" contains \"$igroup\"")
	if ($lun_prefix =~ /$igroup/);

    @lun_list = split(/:/, $lun_list) if (defined($lun_list));

  again:

    if ($line = string_search(ha_filer_cmd($fc, "lun show all"),"find", $igroup)) {
	
	if ($action =~ /destroy|replace/) {

	    while ($line = string_search(ha_filer_cmd($fc, "lun show all"),
					"find", $igroup)) {
		
    		#/vol/vol1/q0/lun0   250m (262144000)     (r/w, online, mapped)
    
    		my ($lun_prefix, $size, $bytes, $state) =
    		    $line =~ /^\s+(\S+)\s+(\S+)\s+\((\d+)\)\s+\S+\s+(\S+)[,\)].*/;
    
    		if ($state =~ /online/) {
                &filer_lun_offline($fc,$lun_prefix);
    
    		    logcomment("$name: $lun_prefix offlined");
    		}
    
    		&filer_lun_destroy($fc,$lun_prefix);
    		logcomment("$name: $lun_prefix destroyed");
	    }
	    
	    goto again;
	}

    } else {

	return if ($action =~ /destroy/);

        # determine the os type
        my $type = filer_igroup_os_get($fc, $igroup);
        $type =~ s/unix-//;

	
	# create new luns

	foreach my $param (@lun_list) {
	    my ($n, $size) = split(/,/,$param);
	    my $lun = $lun_prefix.$igroup."_lun".$n;

	    tstlog(-comment => "create lun:size:type - $lun: $size : $type");
	    filer_lun_create($fc, $lun, $size, $type);	
	    tstlog(-comment => "$name: created $lun");

            filer_lun_map($fc, $lun, $igroup, $n);
		
	    tstlog(-comment => "$name: mapped $lun to $igroup");
	}

    }
}

## @name :: ha_filer_create_qtree
## @description Create qtree on provided volume.  Exits on error.  
## Returns nothing (undef) on success.
##
## @synopsis 
## ha_filer_create_qtree {$conn, $volname, $qtree} 
##
## @arg $conn Required: The filer connection to send this operation over
## @arg $volname Required: The volume name (e.g. vol1)
## @arg $qtree Required: The qtree name
##
## @example 
## my $fc = connect ("lettuce"."#telnet");
## ha_filer_create_qtree($fc, "vol1", "q1");
##
## Creates qtree q0 on /vol/vol1
##
sub ha_filer_create_qtree($$$) {
    my ($fc, $volname, $qtree) = @_;
    my $name="ha_filer_create_qtree";
    my ($output, $line);
    my $qtree_name = "/vol/".$volname."/".$qtree;
    
  again:

    if ($line = string_search(ha_filer_cmd($fc, "qtree"), "find", $qtree)) {

	my ($vol, $qname) = $line =~ /^(\w+)\s+(\w+)/;
#	print "\$vol = $vol, \$qname = $qname\n";

	tstlog(-fatal => "$name: $qtree not on $volname")
	    unless ($vol eq $volname);

	tstlog(-comment => "$name: $qtree_name exits");

    } else {

	my $cmd = "qtree create $qtree_name";
	tstlog(-fatal => "$name: \"cmd\" failed\n \$output = $output")
	    if ($output = ha_filer_cmd($fc, "$cmd"));

	tstlog(-comment => "$name: $qtree_name created");

    }
}

## @name :: ha_filer_parse_lun_list
## @description Convert LUN_LIST stanza string into lun_list array.
## Exit on error.  Returns @lun_list and @lun arrays on success.
##
## Lun information is passed via a scalar string.  Each scaler string
## contains a space delimited group of comma separated values: 
## 
##  <lun-id>, <size>
##
## more at example.
##
## @synopsis 
## ha_filer_parse_lun_list {$action, $LUN_LIST} 
##
## @arg $action Required: "luns" | "list"
## @arg $lun_list Required: lun_list scalar variable
##
## @example 
## my $LUN_LIST="$LUN_LIST="0,2g:1-10,100m:12,250m:15,30g";
## my @luns = ha_filer_parse_lun_list("luns", $LUN_LIST);
##
sub ha_filer_parse_lun_list($$) {
    my ($action, $lun_list) = @_;
    my $name="ha_filer_parse_lun_list";
    my (@luns, @list);

    ExitScriptError( "$name: Invalid parameter passed ($action)")
	if (!($action = check_param($name, $action, "luns|list")));

    my @lun_list = split(/:/, $lun_list);

    print "$name:\n";
    foreach my $param (@lun_list) {
         my ($id, $size) = split(/,/,$param);
         if ($id =~ /\-/) {
             my ($begin, $end) = split(/\-/,$id);
             for (my $i=$begin; $i<=$end; $i++) {
                 push (@list, join(',', $i, $size));
                 push (@luns, $i);
                 print "lun-id = $i, size = $size\n";
             }
         } else {
             push (@list, join(',', $id,$size));
             push (@luns, $id);
             print "lun-id = $id, size = $size\n";
         }

    }

    return (@list) if ($action =~ /list/);
    return (@luns);  
}

no strict "vars";

$SCSI_STATUS_RESERVATION_CONFLCT= "0x00000018";
$SCSI_STATUS_HARDWARE_ERROR=      "0x01150402";
$SCSI_STATUS_WRITE_ERROR_HARD=    "0x020c0302";
$SCSI_STATUS_WRITE_ERROR_SOFT=    "0x010c0102";
$SCSI_STATUS_READ_ERROR_HARD=     "0x03110302";
$SCSI_STATUS_READ_ERROR_SOFT=     "0x02180102";

1;




