# Let emacs know that we are editing -*-perl-*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/i2p.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library interface for the inode to path (i2p) feature
##
## @description
##     The i2p library consists of the following functions:
##
##     i2p_is_supported - is i2p supported for the specified filer?
##
##     i2p_fiji_is_supported - is i2p fiji point supported for the specified filer?
##
##     i2p_is_enabled - is i2p enabled for the specified volume?
##
##     i2p_enable - enable i2p for the specified volume
##
##     i2p_disable - disable i2p for the specified volume
##
##     i2p_scan_start - start the i2p scanner for the specified volume
##
##     i2p_scan_status - query i2p scanner status for the specified volume
##
##     i2p_scan_stop - stop the i2p scanner for the specified volume
##
##     i2p_scan_wait - wait for the i2p scanner to complete for the specified
##                     volume
##
##     i2p_scan_speed - set/get the i2p scanner speed
##
##     i2p_inode_to_path - query parent/path info for the specified inode
##
##     i2p_path_to_inode - query inode info for the specified path
##
##     i2p_t_option_exists - check if -t option (burt203123) exists for inodepath.
##
## @usage
##     use i2p;
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
##
## @status public


# XXX NOTE: This library is still a work in progress.  To do:
#
# Add perldoc support
#
# Should probably add parameter validation to the functions.  The
# necessary modules to import are included but commented out below.


package i2p;

use strict;
use Carp qw(croak);
use Exporter;
#use Module::Common qw(validate_mixed);
#use Params::Validate qw(validate_pos BOOLEAN SCALAR);
use SPARSE::Util::Logger qw(logDbg);
use Volume qw(vol_exists vol_options);
use SayWrap qw(sayEOF);
use Filer::Version qw(get_wafl_version
                      WV_TSINGTAO_MIN WV_GORDONBIERSCH_MIN UNSUPPORTED_FEATURE_EXCEPTION);

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(&i2p_is_supported
		    &i2p_fiji_is_supported
		    &i2p_is_enabled
		    &i2p_enable
		    &i2p_disable
		    &i2p_scan_start
		    &i2p_scan_status
		    &i2p_scan_stop
		    &i2p_scan_wait
		    &i2p_scan_speed
		    &i2p_inode_to_path
		    &i2p_path_to_inode
            &i2p_t_option_exists
                 );

## @name ::i2p_is_supported($fh)
##
## @summary
##     Is i2p supported?
##
## @description
##     If i2p is supported in the filer version, returns 1; otherwise
##     raises an exception.
##
## @arg $fh Required: filer host object.

sub i2p_is_supported ($)
{
    my($fh) = @_;

    my $version = get_wafl_version($fh);
    if ($version < WV_TSINGTAO_MIN) {
        croak(UNSUPPORTED_FEATURE_EXCEPTION .
              "i2p is supported for Tsing Tao on.");
    }
    return 1;
}

## @name ::i2p_fiji_is_supported($fh)
##
## @summary
##     Is i2p fiji point supported?
##
## @description
##     If i2p fiji point is supported in the filer version, returns 1; otherwise
##     raises an exception.
##
## @arg $fh Required: filer host object.
sub i2p_fiji_is_supported ($)
{
    my($fh) = @_;

    my $version = get_wafl_version($fh);
    # i2p fiji point is supported from tt.1 on
    if ($version < (WV_TSINGTAO_MIN+1)) {
        croak(UNSUPPORTED_FEATURE_EXCEPTION .
              "i2p fiji point is supported for TT.1 and GB on.");
    }
    return 1;
}

## @name ::i2p_is_enabled($fc, $vol)
##
## @summary
##     Is i2p enabled?
##
## @description
##     If i2p is enabled for the specified volume, returns 1; otherwise
##     returns 0.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.

sub i2p_is_enabled ($$)
{
    my ($fc, $vol) = @_;

    my $opts = vol_options($fc, $vol) or croak $@;
    return($opts->{no_i2p} eq 'off' ? 1 : 0);
}

## @name ::i2p_enable($fc, $vol)
##
## @summary
##     Enable i2p.
##
## @description
##     Enable i2p for the specified volume.  Returns 1 for success,
##     0 for fail.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.

sub i2p_enable ($$)
{
    my ($fc, $vol) = @_;

    vol_options($fc, $vol, no_i2p => 'off') or croak $@;

    return(i2p_is_enabled($fc, $vol));
}

## @name ::i2p_disable($fc, $vol)
##
## @summary
##     Disable i2p.
##
## @description
##     Disable i2p for the specified volume.  Returns 1 for success,
##     0 for fail.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.

sub i2p_disable ($$)
{
    my ($fc, $vol) = @_;

    vol_options($fc, $vol, no_i2p => 'on') or croak $@;

    return(!i2p_is_enabled($fc, $vol));
}

## @name ::i2p_scan_start($fc, $vol [,%opts])
##
## @summary
##     Start the i2p scanner.
##
## @description
##     Starts the i2p scanner for the specified volume.  Returns 1 for
##     success; croaks otherwise.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg wait Optional: 0 or 1.  Defaults to 0.
##                     Wait for the scanner to complete?
##
## @arg timeout Optional: if wait is set, how long to wait.  Defaults to 600.

sub i2p_scan_start ($$;%)
{
    my ($fc, $vol, %opts) = @_;

    my $cmd = "wafl scan i2p $vol";
    my $text = sayEOF($fc, "priv set -q advanced; $cmd; priv set");
    $text eq "" or croak "$cmd failed: $text";

    sleep(30);

    if (defined($opts{wait}) && $opts{wait} == 1) {
        my $timeout = $opts{timeout} || 1200;
        return(i2p_scan_wait($fc, $vol, timeout => 1200));
    }

    return 1;
}

## @name ::i2p_scan_status($fc, $vol)
##
## @summary
##     Query i2p scanner status for the specified volume.
##
## @description
##     Queries the i2p scanner status for the specified volume and returns a
##     hash reference to the information.  The hash keys are id, scanned_inodes,
##     and total_inodes which correspond to the data reported by the filer's
##     wafl scan status command.  Sample filer output is:
##
##     Volume v:
##      Scan id                   Type of scan     progress
##           13 inode to parent initialization     22099 of 2175327
##
##     If the i2p scanner is not currently running for the specified volume,
##     the function returns 0.  For any unexpected behavior, the function
##     croaks.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.

# XXX Workaround alert:
#
# There appears to be a timing problem with the wafl scan command.  If no
# scans are running, it normally reports that.  If one or more scans are
# running, it normally reports status data.  However, on occasion, it seems
# to get in an in-between state such that it thinks a scanner is running but
# doesn't have data to report.  In those cases, it reports a header with no
# data.  The following wrapper is a temporary workaround.  If the problem is
# fixed, the wrapper should be removed and _i2p_scan_status() should be
# renamed to i2p_scan_status().

sub i2p_scan_status($$)
{
    my ($fc, $vol) = @_;

    my $tries = 0;
    while ($tries <= 5) {
        my $rc;
        eval { $rc = _i2p_scan_status($fc, $vol); };
        if ($@) {
            if ($@ =~ /Unexpected output/) {
                ++$tries;
                sleep(5);
                next;
            }
            else {
                croak $@;
            }
        }
        else {
            return($rc);
        }
    }
}

sub _i2p_scan_status ($$)
{
    my ($fc, $vol) = @_;

    my $idle_re = qr(No\s+WAFL\s+scans\s+in\s+progress);
    my $vol_re = qr(Volume\s$vol:);
    my $hdr_re = qr(Scan\s+id\s+Type\s+of\s+scan\s+progress);
    my $entry_re =
        qr((\d+)\s+inode\s+to\s+parent\s+initialization\s+(\d+)\s+of\s+(\d+));

    my $br_entry_re =
        qr((\d+)\s+inode\s+to\s+parent\s+initialization\s+public\s+inode\s+(\d+)\s+of\s+(\d+));
        
    # There are two expected cases:
    #
    # 1. No scanners in progress: return 0
    # 2. Scanner status starting with the Volume line: parse output and
    #    return accordingly
    #
    # Any other type of output is unexpected and will result in the function
    # croaking.

    my $cmd = "wafl scan status $vol";
    my $text = sayEOF($fc, "priv set -q advanced; $cmd; priv set");

    croak "No output returned for $cmd" if (!defined($text) || $text eq "");

    # No scans in progress?
    return 0 if ($text =~ /$idle_re/);

    my @lines = split('\n', $text);

    # Some sanity checks.  There should be at least 3 lines with the first
    # being the volume name and the second being the header.
    croak "Unexpected output for $cmd: $text"
        unless (@lines >= 3 && $lines[0] =~ $vol_re && $lines[1] =~ $hdr_re);

    shift; shift; # list just contains status entries now

    foreach my $entry (@lines) {
        if ($entry =~ /$entry_re/) {
            return { id => $1, scanned_inodes => $2, total_inodes => $3 };
        }
        if ($entry =~ /$br_entry_re/) {
            return { id => $1, scanned_inodes => $2, total_inodes => $3 };
        }
        # Otherwise, it must be a non-i2p scanner entry
    }

    # If we get here, it means there are scanners running for the volume,
    # but none of them are i2p.  Therefore, we just return 0.

    return 0;
}

## @name ::i2p_scan_stop($fc, $vol)
##
## @summary
##     Stop the i2p scanner.
##
## @description
##     Stops the i2p scanner for the specified volume.  Returns 1 for
##     success (or if there's no scanner running on the volume); croaks
##     otherwise.
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.

sub i2p_scan_stop ($$)
{
    my ($fc, $vol) = @_;

    my $status;

    # The abort subcommand requires a scan id
    eval { $status = i2p_scan_status($fc, $vol); };
    croak $@ if ($@);

    return 1 if ($status == 0); # scanner isn't running

    my $cmd = "wafl scan abort $status->{id}";
    my $text = sayEOF($fc, "priv set -q advanced; $cmd; priv set");

    $text eq "" or croak "$cmd failed: $text";

    return 1;
}

## @name ::i2p_scan_wait($fc, $vol [,%opts])
##
## @summary
##     Wait for the i2p scanner to complete.
##
## @description
##     Waits for the i2p scanner to complete for the specified volume.
##     Returns 1 for success; croaks on timeout. 
##
## @arg $fc Required: filer connection object.
##
## @arg $vol Required: volume name.
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg timeout Optional: how long to wait.  Defaults to 600.

sub i2p_scan_wait ($$;%)
{
    my ($fc, $vol, %opts) = @_;

    my $timeout = $opts{timeout} || 1200;
    my $etime = time() + $timeout;

    while (i2p_scan_status($fc, $vol)) {
        if (time() > $etime) {
            croak "Timed out waiting for i2p scanner to complete";
        }
        sleep(30);
    }

    return 1;
}

## @name ::i2p_scan_speed($fc [, %opts])
##
## @summary
##     Set/get the i2p scanner speed
##
## @description
##     If speed is not specified, return the current wafl scanner speed
##     setting.  If speed is specified, set the scanner speed to that value.
##     In the latter case, the function will return the old value as a
##     convenience for possible later restoral.  For anything unexpected,
##     the function croaks.
##
##     NOTES:
##     1.  Wafl speed settings are global in nature; therefore, modifying
##         the speed affects all system scanners.
##
##     2.  The contents of this function should probably be relocated to
##         a more generic area as there's currently nothing i2p-specific
##         within it.  However, this function should probably still live on
##         as a wrapper in case the wafl scan speed ever becomes settable
##         on a per-scanner basis.
##
## @arg $fc Required: filer connection object.
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg speed Optional: scanner speed.  Ranges from 1 to 100000.  Defaults
##                      to 2000.  The value 'default' can also be specified
##                      in which case the speed will be set to the system
##                      default.

sub i2p_scan_speed ($;%)
{
    my ($fc, %opts) = @_;

    # Perhaps this can be queried somehow rather than hardcoding?
    use constant DEFAULT_SPEED => 2000;

    my $speed_re = qr(WAFL\s+scan\s+speed\s+is\s+(\d+));
    my $cmd = "priv set -q advanced; wafl scan speed";

    # Get the current setting
    my $output = sayEOF($fc, $cmd);
    my ($curr_speed) = $output =~ /$speed_re/;

    defined($curr_speed) or croak "$cmd failed: $output";

    return ($curr_speed) if (!defined($opts{speed}));

    # Change the setting
    $opts{speed} = DEFAULT_SPEED if ($opts{speed} eq 'default');
    $cmd .= " $opts{speed}";
    $output = sayEOF($fc, $cmd);
    croak "$cmd failed: $output" if (sayEOF($fc, $cmd) ne "");
   
    return $curr_speed;
}

## @name ::i2p_inode_to_path($fc, $vol, $inum [, %options])
##
## @summary
##     Query path info.
##
## @description
##     For the specified volume and inode number, return all requested
##     i2p path info.  Required parameters are positional; optional
##     parameters are named.
##
##     Path info is returned as a list reference.  Each item in the list
##     is a hash reference corresponding to parent/path info.  The primary
##     path is always the first list element; the remaining elements are
##     secondary paths.
##
##     Example of how to access path info for volume v, inode 100:
##
##     $info = i2p_inode_to_path($fc, 'v', 100)
##     foreach my $p (@$info) {
##         $name = $p->{name};
##         $path = $p->{path};
##         $par_inode = $p->{par_inode};
##         $par_cookie = $p->{par_cookie};
##     }
##
##     To report primary only:
##
##     $info = i2p_inode_to_path($fc, 'v', 100, secondary => 0) 
##
##     To retrieve all inode info from a snapshot:
##
##     $info = i2p_inode_to_path($fc, 'v', 100, snap => 'hourly.0') 
##
##
## @arg fc Required: filer connection object.
##
## @arg vol Required: volume name.
##
## @arg inum Required: inode number.
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg snap Optional: snapshot name.
##                     Retrieve inode info from specified snapshot
##
## @arg secondary Optional: 0 or 1.  Defaults to 1.
##                                   Report secondary paths?
##
## @arg nametype Optional: nfs or cifs.
##     nfs,  is the default behavior,
##     cifs, returns a Unicode name, if available.
##
##     This argument is added to support the new 'inodepath CLI'
##     option, [ -t { NFS | CIFS } ], that is added (currently) 
##     to IC.3, as part of the fix to burt203123.
##
## @arg fsid  Optional: vol fsid .
##                     Use volume fsid instead of volume name.
##
## XXX Note: Ideally, it should have been either -v option or
##           -f option, but for backward compatibility, we
##           maintain the existing calling sequence, i.e. have
##           volume name as a required parameter, even if it
##           is required that inodepath be called with -f option.
##           We overwrite the default $vol_arg (-v $vol)  with
##           "-f $opts{fsid}".
##           e.g
##           $info = i2p_inode_to_path($fc, 'v', 100, 
##                                     fsid => '0x158a274')
##
## @arg stty_type Optional: nfs, sgml or utf8
##           'stty encoding' in which the output 
##           (filename / pathname) from the inodepath
##           command is desired.
##           By Default, inodepath command is called
##           without setting the 'stty encoding'.
##          

sub i2p_inode_to_path ($$$;%)
{
    my ($fc, $vol, $inum, %opts) = @_;

    # spell it out...easier to read/modify...
    my $cmd = 'inodepath';
    my $flags =
        (defined($opts{secondary}) && !$opts{secondary}) ? '-in' : '-ain';
    my $vol_arg = defined($opts{fsid}) ? "-f $opts{fsid}" : "-v $vol";
    my $snap_arg = defined($opts{snap}) ? "-s $opts{snap}" : "";
    my $snapid_arg = $opts{snapid} ? "-d $opts{snapid}" : "";
    my $nametype_arg = $opts{nametype} ? "-t $opts{nametype}" : "" ;
    my $stty_type_cmd = $opts{'stty_type'} 
                            ? "stty encoding $opts{'stty_type'};"
                            : "";

    my $cmd_str = "priv set -q advanced; $stty_type_cmd $cmd $flags $vol_arg "
                . "$snap_arg $snapid_arg $nametype_arg $inum";

    my $text = sayEOF($fc, $cmd_str, timeout=>360);

    # Errors always start with the cmd name followed by a colon
    if ($text =~ /^$cmd:/) {
        croak $text;
    }

    return(_parse_path_info(\$text, $flags));
}

## @name ::_parse_path_info($text_ref)
##
## @summary
##     Parse inodepath output (internal function)
##
## @description
##     Take inodepath output text as input, extract inode parent and
##     pathname info, build the data structure, and return a reference
##     to it.
##
## @arg text_ref Required: text reference

sub _parse_path_info ($$)
{
    my($text_ref, $_flags) = @_;

    my $path_cnt_re = qr(Inode\s+\d+.*?has\s+(\d+)\s+name);
    my $name_re = qr(name\s+=\s+(.*)\n);
    my $path_re = qr(pathname\s+=\s+(.*)\n);
    my $par_inode_re = qr(parent\s+inode\s+=\s+(\d+));
    my $par_cookie_re = qr(parent\s+dir\s+cookie\s+=\s+(\d+));

    # The [    n] field serves as a good delimiter as it leaves the header
    # in paths[0] and each path entry in paths[i].
    my $delimiter = qr(\[\s+\d+\]);
    my @path_entries = split(/$delimiter/, $$text_ref);

    # Extract path count
    my ($path_cnt) = ($path_entries[0] =~ /$path_cnt_re/);
    defined($path_cnt) or croak "Error parsing path count: $path_entries[0]";

    logDbg("path entries parsed: $#path_entries; path count reported: $path_cnt");

    # Vaporize header
    shift @path_entries;

    # Sanity check: reported path count should equal number of path entries,
    # if all the paths (-a) are requested.
    if ($_flags eq "-ain") {
        if ($path_cnt != @path_entries) {
            croak "Path count does not match number of paths: $$text_ref";
        }
    }

    my @paths;

    # Parse paths and build data structure
    foreach my $path (@path_entries) {
        my ($name, $pathname, $par_inode, $par_cookie) =
            ($path =~ /$name_re.*?
                       $path_re.*?
                       $par_inode_re.*?
                       $par_cookie_re/sx);

        defined($name) &&
        defined($pathname) &&
        defined($par_inode) &&
        defined($par_cookie) or croak "Error parsing paths: $path";

       push(@paths,
            { name       => $name,
              path       => $pathname,
              par_inode  => $par_inode,
              par_cookie => $par_cookie });
    }
    
    return(\@paths);
}

## @name ::i2p_path_to_inode($fc, $path)
##
## @summary
##     Query inode number
##
## @description
##     For the specified path, return the corresponding inode number.
##
##     This is just a filer side convenience function.  It may be
##     preferrable to retrieve inode numbers from the client side via
##     stat().
##
##     Examples:
##     $my inum = i2p_path_to_inode($fc, '/vol/v/file')
##     $my inum = i2p_path_to_inode($fc, '/vol/v/dir/subdir/file')
##
## @arg fc Required: filer connection object.
##
## @arg path Required: file path (including the volume).

sub i2p_path_to_inode ($$)
{
    my ($fc, $path) = @_;

    my $inode_re = qr(fileid=0x0*(\w+)); # hex() doesn't like leading 0x

    my $cmd_str = "priv set -q advanced; showfh $path";
    my $output = sayEOF($fc, $cmd_str);
    
    my ($inum) = ($output =~ /$inode_re/);
    defined($inum) or croak $output;

    return(hex($inum)) # yes, hex() actually returns base 10
}

## @name ::i2p_t_option_exists($fc)
##
## @summary
##     Check if -t option exists for inodepath command.
##
## @description
##     This function checks if -t option exists for 
##     CLI command 'inodepath'; this option is added
##     to 'inodepath' as part of fix to burt203123.
##
##     Returns 'true', if -t option exists, false 
##     otherwise.
##
## @arg fc Required: filer connection object.
##

sub i2p_t_option_exists($) 
{
    my ($fc) = @_ ;
    my $cmd = "priv set -q advanced; inodepath" ;

    ## If -t option is present for inodepath CLI command, it
    ## would be displayed by inodepath usage.
    ## match for "[ -t { NFS | CIFS } ]", with leading space.
    my $t_option_re = qr(\s+\[\s+-t\s\{\s+NFS\s+\|\s+CIFS\s+}\s+\]) ;
    my $output = sayEOF($fc, $cmd) ;

    # return 'true' on match.
    return ($output =~ m/$t_option_re/) ;
}

1;
