# $Id: //depot/prod/test/Rsierranevada/lib/install.pm#2 $

## @summary Routines to (re)create an ontap installation
##
## @description
##    This library contains individual routines to create simulator
##    files, install a new filesystem, interact with the initial setup
##    command, set licenses, and install/download another version of
##    ontap.
##
## @synopsis
##    use install; 
##
## @usage
##    Each routine provided only performs a small portion of the steps
##    in any likely scenario, in order to provide new tests a fine
##    grain of control.  For example, because download_files doesn't
##    perform the "reboot" that would usually follow when installing
##    ontap, it could be used to perform the ontap installation that
##    would precede using "revert_to".
##
##    Tests that use these routines in their most typical ways can be
##    found in the "Filer" directory.
##
## @keywords 
##    install, filer install, setup, zero, simulator
##
## @author dl-central-libs
## @change 2003-05-16 hedstrom: cope with current filer version not supporting software command
## @change 2005-01-10 hedstrom: use timeout argument in copy_kernel
## @change 2005-01-19 hedstrom: support Deux, use Burt.pm to track known burt
## @change 2005-06-23 hedstrom: support Excelsior with amd64 architecture and 2 maytag suffix

package install;

use vars ('@ISA', '@EXPORT');

BEGIN {
    require Exporter;
    @ISA = qw(Exporter);
    @EXPORT = qw(basic_setup download_files license unlicense disk_assign newfs make_sim_files check_version guess_buildtype);
};

use Tharn;
use strict;
use mount ('mount');
use boot;
use Burt;
use ModelInfo;
use Volume qw(vol_root);
use SayWrap qw(prompt_loop);

sub make_host ( $ );
sub make_conn ( $$ );

# (for use by license, below)

my %Filer_licenses = (
					  cifs => "GTLYRVH",
					  cluster => "KDQLGBN",
					  dafs => "OBOLDQC",
					  fcp => "ICRNCZG",
					  http => "AODUWOJ",
					  iscsi => "IKVAREM",
					  nfs => "UYTCNCG",

					  # Next 3 are same actual license
					  cluster_remote => "GTFJHSI",
					  syncmirror_remote => "GTFJHSI",
					  remotesyncmirror => "GTFJHSI",

					  snapmanager => "GFMYRVH",
					  snapmanagerexchange => "GFMYRVH",
					  snapmirror => "UUNLDQC",
					  snapmirror_sync => "IEOAOTB",
					  snaprestore => "EOYPBIL",
					  snapvault_client => "KXQPNOJ",
					  snapvault_server => "YSPCZIE",

					  # Next two are same license
					  syncmirror_local => "UELJHSI",
					  syncmirror => "UELJHSI",

					  vfiler => "SPQLGBN",
					  vld => "QYOAREM",
					  volcopy => "AAWPYWA",
	                                  flex_clone => "QNFYRVH"
					 );

# (for use by license, and a little bit by basic_setup, below)

my %Cache_licenses = (
					  netcache => "GXVPYWA",
					  nntp => "WJSRXFF",
					  mms => "EKMAREM",
					  real => "YMWRXFF",
					  quicktime => "MOWPYWA",
					  icap => "GFCJHSI",
					  grm => "WRQLGBN",
					  ncagent => "GJOLDQC",
					  https => "IQWRXFF",
					  pnfs => "SZNNCZG",
					 );

#-------------------------------------------------------------------------

## @name ::basic_setup()
## @summary talk to the filer "setup" command
## @description 
##
## This function interacts with an appliance (filer or cache) "setup"
## command, to perform basic, initial setup.  It can be called to
## answer prompts when the appliance enters setup atomatically on boot
## (such as after zeroing disks), or it can interact with the "setup"
## command on an already-set-up filer.
##
## @synopsis
##   basic_setup($aspec, [issue_setup_cmd=>0], [setup=>{}], [prefix=>...]);
##
## @arg aspec Required: A hostspec or connection record of the
## appliance to be set up.  If a host spec is given, the "#telnet"
## connection will be used.

## @arg issue_setup_cmd Optional: If set to 1, issue the "setup"
## command.  The default is 0, which is to expect that setup will be
## entered automatically (such as by an appliance which just finished
## zeroing disks).

## @arg setup Optional: A hash, whose fields will be used as input to
## setup.  The default is an empty hash.  See usage.

## @arg prefix Optional: If specified, parameters and environment
## variables whose names begin with this prefix will also be used as
## input to setup.  See usage.

## @arg brand Optional: the brand of maytag image to use.  Possible
## values are "netapp" or "ibm".  Defaults to "$prefix_BRAND" if that
## exists or to "BRAND" in the host record of $aspec if that exists,
## or else a value will be guessed from the model option.


## @usage
## The values used in responding to setup come from a number of
## different sources.
##
## (1) Fields in the "setup" argument to this function.
##
## (2) If the "prefix" argument is given, parameters and environment
## variables that begin with that prefix followed by an underscore
## (_).  So, for example, if prefix=&gt;FILER, then the field HOSTNAME
## will take its value from the parameter or environment variable
## "FILER_HOSTNAME".
##
## (3) Fields in the host record for this appliance (as provided by
## the "aspec" argument).
##
## (4) For values that may not appear in a host record, a constant
## default value for that parameter, which is usually a blank ("") or
## no ("n").
##
## When a field appears in multiple sources, the value used will be
## from the source that's listed earlier in the above list.
##
## The following fields are defined:
##
## HOSTNAME : The hostname of the filer.  (defaults to the host record's
## hostname or HOSTNAME fields).
##
## IFC_xxx : The name of a network interface, where "xxx" can be
## anything.  One of these would exist for each interface to be
## handled by setup, and other fields that end in the same "xxx" will
## provide further information about that interface.
##
## IP_xxx : The IP address of the interface given in IFC_xxx.
## (default is "")
##
## MASK_xxx : The netmask of the interface given in IFC_xxx.
## (default is "255.255.255.0").
##
## NETMASK_xxx : A deprecated synonym for MASK_xxx
##
## PARTNER_xxx : For a CFO pair of filers, the name of an interface or
## IP address on the partner which IFC_xxx will substitute for on
## takeover.  (default is "")
##
## MEDIATYPE_xxx : The mediatype of the interface given in IFC_xxx.
## (default is "")
##
## FLOWCONTROL_xxx : The flow control value for the interface given in
## IFC_xxx.  (default is "")
##
## JUMBOFRAMES_xxx : Whether jumbo frames should be supported by the
## interface given in IFC_xxx.  (default is "")
##
## GATEWAY : The IP address of the default gateway (default is "")
##
## LOCATION : The physical location of this appliance.  (default is
## "")
##
## DNS : Whether the appliance should use DNS (default is "y" if
## DNS_DOMAIN and DNS_SERVER_LIST have been provided and "n" if
## they have not been.  Note that netcache appliances do not ask--they
## assume yes.)
##
## DNS_DOMAIN : What domain name to append to hostnames for DNS
## lookups.  (default is "", which setup will probably not allow if
## DNS is "y" or this is a netcache)
##
## DNS_DOMAINNAME : A deprecated synonym for DNS_DOMAIN
##
## DNS_SERVER_LIST : Comma separated list of IP addresses of DNS
## servers.  (default is "", which setup will probably not allow if
## DNS is "y" or this is a netcache)
##
## NIS : Whether the appliance should use NIS (default is "y" if
## NIS_DOMAIN and NIS_SERVER_LIST have been provided and "n" if
## they have not been).
##
## NIS_DOMAIN : What domain to use for NIS queries.  (default is
## "", which setup may not allow if NIS is "y")
##
## NIS_DOMAINNAME : A deprecated synonym for NIS_DOMAIN
##
## NIS_SERVER_LIST : Comma separated list of IP addresses of NIS
## servers.  (default is "", which setup may not allow if NIS is "y")
##
## TIMEZONE : What time zone the filer is in (default is "")
##
## ROOT_PASS : The root user's password.  If this is not passed in
## the default_password_7g of the host is used.
##
## PRIMARY_CONTACT_NAME : Required : Primary contact person for this filer.
## (default is "")
## This is required by IBM branded machines.
##
## PRIMARY_CONTACT_PHONE : Required : Phone number of primary contact 
## person for this filer. (default is "")
## This is required by IBM branded machines.
##
## PRIMARY_CONTACT_ALT_PHONE : Required : Alternate phone number of 
## primary contact person for this filer. (default is "")
## This is required by IBM branded machines.
## 
## PRIMARY_CONTACT_EMAIL : Required : Email address or IBM WebID of 
## primary contact person for this filer. (default is "")
## This is required by IBM branded machines.
## 
## SECONDARY_CONTACT_NAME : Required : Secondary contact person for this filer.
## (default is "")
## This is required by IBM branded machines.
##
## SECONDARY_CONTACT_PHONE : Required : Phone number of secondary 
## contact person for this filer. (default is "")
## This is required by IBM branded machines.
##
## SECONDARY_CONTACT_ALT_PHONE : Required : Alternate phone number of 
## secondary contact person for this filer. (default is "")
## This is required by IBM branded machines.
## 
## SECONDARY_CONTACT_EMAIL : Required : Email address or IBM WebID of 
## secondary contact person for this filer. (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_NAME : Required : Name of the business. (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_ADDR_STREET : Required : Business address. (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_ADDR_CITY : Required : City where business resides.
## (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_ADDR_STATE : State where business resides. (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_ADDR_COUNTRY : Required : 2-character country code where
## business resides.  (default is "")
## This is required by IBM branded machines.
##
## BUSINESS_ADDR_POSTALCODE : Postal code where business resides.
## (default is "")
## This is required by IBM branded machines.
##
## @bugs
##
## This routine deals with a number of different setup options, but
## there are some filer features it cannot handle except to disable
## them, so that if you want them you'll have to (re)enable them
## after setup is complete.  They are:
##
## - Virtual network interfaces (VIFs)
##
## - An admin host
##
## - Autosupport (the "autosupport.enable" version.  the
##   "autosupport.support.enable" version remains untouched)
##
## - IPv6
##
## @dependencies
## In case setup is to be done on a simulator the simulator should be 
## up and running and the test should not be run from the simhost of the 
## simulator being set up.



sub basic_setup($@) {
    my ($aspec, %opt) = @_;
    my $setup = $opt{setup} || {}; 
    my $ahost = make_host($aspec);
    my $aconn = make_conn($aspec,"#telnet");
	my $timeout = $opt{timeout} || 60;

	# Pull the parameters from various sources into a single hash:

    # BURT 449454:
    # Changes needed to enable ACP. We add two entries to the defaults
    # ACP=>"n" and ACP_PORT=>""
    # We also need to add the following query responses to @prompts_setup
    #       i.    Please enter the domain for Network Interface => ""
    #       ii.   Please enter the netmask for Network Interface =>"" 
    # We are fine with the default values for the above queries.

    my %defaults = (
					GATEWAY=>"",
					LOCATION=>"",
					DNS=>"", DNS_SERVER_LIST=>"",
					DNS_DOMAIN=>"", DNS_DOMAINNAME=>"", 
					NIS=>"", NIS_SERVER_LIST=>"",
					NIS_DOMAIN=>"", NIS_DOMAINNAME=>"", 
					AUTOSUPPORT=>"n",
					TIMEZONE=>"",
					ADMIN_HOST=>"",
					PRIMARY_CONTACT_NAME=>"",
					PRIMARY_CONTACT_PHONE=>"",
					PRIMARY_CONTACT_ALT_PHONE=>"",
					PRIMARY_CONTACT_EMAIL=>"",
					SECONDARY_CONTACT_NAME=>"",
					SECONDARY_CONTACT_PHONE=>"",
					SECONDARY_CONTACT_ALT_PHONE=>"",
					SECONDARY_CONTACT_EMAIL=>"",
					BUSINESS_NAME=>"",
					BUSINESS_ADDR_STREET=>"",
					BUSINESS_ADDR_CITY=>"",
					BUSINESS_ADDR_STATE=>"",
					BUSINESS_ADDR_COUNTRY=>"",
					BUSINESS_ADDR_POSTALCODE=>"",
                    ACP =>"n",
                    ACP_PORT=>"",
				   );

	my %params = param_merge(
							 hash => \%defaults,
							 hostrec => $ahost,
							 param_prefix => $opt{prefix},
							 hash => $setup,
							);

	# Shuffle deprecated parameters into non-deprecated ones
	foreach my $key (keys %params) {
		my ($suffix) = ($key =~ /^NETMASK(.*)/);
		next if !defined($suffix);
		$params{"MASK$suffix"} ||= $params{"NETMASK$suffix"};
	}
	$params{DNS_DOMAIN} ||= $params{DNS_DOMAINNAME};
	$params{NIS_DOMAIN} ||= $params{NIS_DOMAINNAME};

    # Construct an array of DNS servers from DNS_SERVER_LIST

	my @dns_server_list = split(/,/, $params{DNS_SERVER_LIST});
	my $dns = $params{DNS} || (($params{DNS_DOMAIN} && @dns_server_list) ? "y" : "n");
    my $root_pass = $ahost->default_password();
    if (defined($params{ROOT_PASS})) {
        $root_pass = $params{ROOT_PASS};
    }

    # Construct comma separated list of NIS servers in NIS_SERVER_LIST
    # (remove extra spaces, as the filer does not allow them)
	my $nis_server_list = $params{NIS_SERVER_LIST};
	$nis_server_list =~ s/\s+//g;
	my $nis = $params{NIS} || (($params{NIS_DOMAIN} && $nis_server_list) ? "y" : "n");

    if ($opt{issue_setup_cmd}) {
        say ($aconn, "setup", '-blind'); 
    }

    # Let's define a pattern to match the interface types that setup
    # might query us for.
	my $failure;
    my @prompts_setup = 
		("Are you sure you want to continue"         => "y",
         "Do you want to configure interface groups" => "n",
		 "new hostname"                              => $params{HOSTNAME},
		 "(e|E)nable IPv6\\?"                        => "n",
		 "virtual network interfaces"                => "n",
		 "IP address for (?:FC Network Interface|VI Network Interface|Network Interface|SPANS interface) (\\w+)\\W"
                                                     => sub { get_ifc_info(\%params,$1,"IP") || ""; },
		 "netmask for (VI )?Network Interface (\\w+)\\W" => sub { get_ifc_info(\%params,$2,"MASK") || "255.255.255.0"; },
         "domain for (VI )?Network Interface (\\w+)\\W" => sub { get_ifc_info(\%params,$2,"DOMAIN") || ""; },
		 "Should interface (\\w+) take over "        => sub { get_ifc_info(\%params,$1,"PARTNER") ? "y" : "n"; },
		 "Should adapter (\\w+) take over ELANs"     => sub { get_ifc_info(\%params,$1,"PARTNER") ? "y" : "n"; },
		 "name to be taken over by (\\w+)\\W"           => sub { get_ifc_info(\%params,$1,"PARTNER"); },
		 "Please enter the ATM adapter to take over" => sub { get_ifc_info(\%params,$1,"PARTNER"); },
		 "media type for (\\w+)\\W"                     => sub { get_ifc_info(\%params,$1,"MEDIATYPE") || ""; },
		 "flow control for (\\w+)\\W"                   => sub { get_ifc_info(\%params,$1,"FLOWCONTROL") || ""; },
		 "want (\\w+) to support jumbo frames\\?"    => sub { get_ifc_info(\%params,$1,"JUMBOFRAMES") || ""; },
		 "enter name of ELAN to add to ATM adapter"  => "",
		 "setup through the web interface"           => "n",
		 "default gateway"                           => $params{GATEWAY},
		 "IP address of the administration host"     => $params{ADMIN_HOST},
		 "filer located"                             => $params{LOCATION},
		 "enter timezone"                            => $params{TIMEZONE},
		 "root directory for HTTP files"             => "",
		 "DNS resolver"                              => $dns,
		 "DNS domain name"                           => $params{DNS_DOMAIN},
		 "IP address for [a-z]+ nameserver"          => sub { shift(@dns_server_list) || ""; },
		 "another nameserver"                        => sub { @dns_server_list ? "y" : "n"; },
		 "NIS client"                                => $nis,
		 "NIS domain name"                           => $params{NIS_DOMAIN},
		 "NIS servers"                               => $nis_server_list,
		 "run autosupport"                           => "n",
         "(N|n)ew password"                          => $root_pass, #this covers both 'New password' and 'Retype new password'
         "Password:"                                 => $root_pass, #if running via console you get prompted to login with new pass
		 "What language .* for multi-protocol files" => "",
		 "Number of disks to add to root volume"     => "0", # (pre-redhook only)
		 "Would you like to configure the RMC LAN interface" => "n",
		 "Would you like to configure the RLM LAN interface" => "n",
		 "Would you like to configure the SP LAN interface" => "n",
		 "Would you like to configure the BMC"       => "n",
		 "Do you want to configure the Shelf Alternate Control Path Management interface for SAS shelves" => $params{ACP},
         "Enter the network interface you want to use for the Alternate Control Path Management" => $params{ACP_PORT},
         #hitting ctrl-c here because basic setup should not configure CIFS
         #there are other libraries that exists to run the CIFS setp wizard
         #this prompt does not always appear, except in special cases
         "Ctrl-C to exit without committing changes" => "\cc",
		 # more prompts to add later
		 #"Would you like to enable DHCP on the RMC LAN interface [y]:"
		 #"Please enter media type for the RMC {100tx-fd, tp-fd, 100tx, tp, auto} [auto]:"
		 "Errors during setup, no changes made"      => sub {
			 # don't die yet; there's a filer prompt that our caller
			 # still has to consume.
			 $failure="setup aborted with \"Errors during setup, no changes made\"";
			 return undef;
		 },
		 #console collisions make looking for "Press retrun key to continue" 
		 #impossible thus looking for smaller subset of the text
		 "key to continue"          => "",
		 # (After change 200344.  It is letting us choose between
		 # "netcache" and "pnfs"; we always choose netcache for now)
		 "Please enter your license: "               => $Cache_licenses{netcache},
		 # (Before change 200344)
		 "enter your NetCache license"               => $Cache_licenses{netcache},
		 # (end change 200344)
		 "enter your NNTP license"                   => "",
		 # (After change 284296)
		 "WM \\(.*\\) protocol license"              => "",
		 # (Before change 284296)
		 "MMS \\(.*\\) protocol license"             => "",
		 # (end change 284296)
		 "MMS Pro license"                           => "",
		 "RTSP \\(\\S+\\) license"                   => "",
		 "ICAP license"                              => "",
		 "GRM Server license"                        => "",
		 "GRM Agent license"                         => "",
		 "HTTPS proxy license"                       => "",
		 "Telnet proxy license"                      => "",
		 "PNFS license"                              => "",
		 "SOCKS license"                             => "",
		 "HTTP Compression license"                  => "",
		 # prompts required for IBM branded machines
		 "Name of primary contact"                   => $params{PRIMARY_CONTACT_NAME},
		 "Phone number of primary contact"           => $params{PRIMARY_CONTACT_PHONE},
		 "Alternate phone number of primary contact" => $params{PRIMARY_CONTACT_ALT_PHONE},
		 "Primary Contact e-mail address or IBM WebID" => $params{PRIMARY_CONTACT_EMAIL},
		 "Name of secondary contact"                 => $params{SECONDARY_CONTACT_NAME},
		 "Phone number of secondary contact"         => $params{SECONDARY_CONTACT_PHONE},
		 "Alternate phone number of secondary contact" => $params{SECONDARY_CONTACT_ALT_PHONE},
		 "Secondary Contact e-mail address or IBM WebID" => $params{SECONDARY_CONTACT_EMAIL},
		 "Business name"                             => $params{BUSINESS_NAME},
		 "Business address"                          => $params{BUSINESS_ADDR_STREET},
		 "City where business resides"               => $params{BUSINESS_ADDR_CITY},
		 "State where business resides"              => $params{BUSINESS_ADDR_STATE}, # optional
		 "2-character country code"                  => $params{BUSINESS_ADDR_COUNTRY},
		 "Postal code where business resides"        => $params{BUSINESS_ADDR_POSTALCODE}, # optional
		);


	return eval {
		prompt_loop(connection => $aconn,
					table => \@prompts_setup,
					done => [$aconn->prompts()],
					max_tries => 100,
					say_opts => ['-timeout',$timeout]);
		if ($failure) {
			die "$failure\n";
		}
		return 1;
    };
}

# Look up key $field for interface $ifname in a hash containing
# IFC_xxx, IP_xxx, etc., or return the empty string if it doesn't
# exist.

# (that is, we know $hash->{IFC_xxx} will be $ifname for some value of
# xxx.  Find which xxx that is, and return $hash->{${field}_xxx}.  As
# a special case, xxx and the _ before it might be the empty string,
# i.e. we return $hash->{$field} if $hash->{IFC} is $ifname.  Yes,
# this is all sort-of crazy, but it's what Filer/setup.thpl made
# the convention)

sub get_ifc_info ( $$$ ) {
	my ($hash, $ifname, $field) = @_;
	my $result=undef;
	# (don't exit this loop early, since the 'each' won't start at the
	# beginning next time unless we get to the end this time...)
	while (my ($key, $value) = each %$hash) {
		if ($key =~ /^IFC(_.*)?$/ && $value eq $ifname) {
			# (if _.* is empty, $key matches, but $1 is undef, apparently)
			my $suffix = defined($1) ? $1 : "";
			$result=$hash->{$field . $suffix};
		}
	}
	return $result;
}

## @name ::license()
## @summary issue "license" command for the given licenses
## @deprecated
##   This function has not been maintained (it does not support
##   nearstore licenses, for example) and does not check whether the
##   license to be installed was successfully installed.  You should
##   instead use license function in the package Admin::License (see
##   Admin/License.pm)
##
## @description
##   Give a filer or cache a site license for particular services.
##
## @synopsis
##    license($aspec
##            [,prefix=&gt;...]
##            [,licenses=&gt;...]
##           )
##
## @arg aspec Required: A hostspec or connection record of the
## appliance to be licensed.  If a host spec is given, the "#rsh"
## connection will be used.
##
## @arg licenses Optional: A list of names of services to assign to
## this filer or cache.  Can either be a reference to an array of
## those services or a string containing a comma-separated list.  The
## default is "$prefix_LICENSES" if the "prefix" option is given and
## that parameter exists, or else it is "nfs,netcache,nntp,mms,real"
## for caches and "nfs,cifs,http" for filers.
##
## @arg prefix Optional: Provides another default for "licenses", as
## described above.
##
## @example
##     # License nfs, cifs, http, and clustering
##     license($h,licenses=>[qw(nfs cifs http clustering)]) or logresult('fail',$@);
##
##     # Same thing
##     license($h,licenses=>"nfs,cifs,http,clustering");
##
##     # Same thing, with possible override by user/parent
##     param('FILER','-mesg',"hostspec of filer");
##     param('FILER_LICENSES','-default',"nfs,cifs,http,clustering");
##     my $h = host($FILER);
##     license($h,prefix=>"FILER");

sub license {
	my ($aspec, %opt) = @_;

	my $ahost = make_host($aspec);
	my $aconn = make_conn($aspec,"#rsh");
	my %defaults = (
					LICENSES=>[
							  ($ahost->hosttype() =~ /cache/) ? 
							  qw(nfs netcache nntp mms real) :
							  qw(nfs cifs http),
							 ],
				   );
	my %params = param_merge(
							 hash => \%defaults,
							 hostrec => $ahost,
							 param_prefix => $opt{prefix},
							);
	my $licenses = $opt{licenses} || $params{LICENSES};
	if (!ref($licenses)) {
		$licenses=[split(/\s*,\s*/,$licenses)];
	}
	return eval {
		foreach my $service (@$licenses) {
			my $filer_code = $Filer_licenses{$service};
			my $cache_code = $Cache_licenses{$service};
			my ($code,$special);
			if ($ahost->hosttype() =~ /cache/) {
				if ($filer_code) {
					$special=1;
					$code=$filer_code;
				} else {
					$code=$cache_code;
				}
			} else {
				$code=$filer_code;
			}
			if (!$code) {
				die "Unknown service \"$service\"\n"; 
			}

			my $operation = $opt{delete} ? "delete" : "add";
			$code = $service if $opt{delete};

			my $send = $special ? 
				"priv set -q advanced; license -Special $operation $code" :
				"priv set -q advanced; license $operation $code";
			say($aconn, $send, '-eval_error');
			die $@ if $@;
		}
		return 1;
	};
}

## @name ::unlicense()
## @summary issue "license delete" command for the given service
##
## @deprecated
##   You should instead use the unlicense function in the package
##   Admin::License (see Admin/License.pm)
## @description
##   Delete a filer or cache site license for particular services.
##
## @synopsis
##    unlicense($aspec
##            [,prefix=&gt;...]
##            [,licenses=&gt;...]
##           )
##
## @arg aspec Required: A hostspec or connection record of the
## appliance to be unlicensed.  If a host spec is given, the "#rsh"
## connection will be used.
##
## @arg licenses Optional: A list of names of services to assign to
## this filer or cache.  Can either be a reference to an array of
## those services or a string containing a comma-separated list.  The
## default is "$prefix_LICENSES" if the "prefix" option is given and
## that parameter exists, or else it is "nfs,netcache,nntp,mms,real"
## for caches and "nfs,cifs,http" for filers.
##
## @arg prefix Optional: Provides another default for "licenses", as
## described above.
##
## @example
##     # UnLicense nfs, cifs, http, and clustering
##     unlicense($h,licenses=>[qw(nfs cifs http clustering)]) or logresult('fail',$@);
##
##     # Same thing
##     unlicense($h,licenses=>"nfs,cifs,http,clustering");
##
##     # Same thing, with possible override by user/parent
##     param('FILER','-mesg',"hostspec of filer");
##     param('FILER_LICENSES','-default',"nfs,cifs,http,clustering");
##     my $h = host($FILER);
##     unlicense($h,prefix=>"FILER");

sub unlicense {
	my ($aspec, %opt) = @_;

	# call license() with "delete" key in the option hash argument

	$opt{delete} = 1;
	return license($aspec, %opt);
}

# Mount the root of the given appliance and return the mountpoint
# mounted (or undef on error with $@ set to the error).  Also, for
# caches, make sure nfs is running and the root volume has been
# exported.

sub do_mount ( $$$$ ) {
	my ($aipaddr, $path, $chost, $nfsvers) = @_;
	my @mountargs = ($aipaddr,$path,client=>$chost);
	if (defined($nfsvers)) {
		push @mountargs, (vers=>$nfsvers);
	}
	return &mount(@mountargs);
}

sub netcache_enable_nfs {
	my ($aconn) = @_;
	say ($aconn, "priv set -q advanced; license -Special NFS=GDPPBIL; nfs on",'-eval_error');
	die $@ if $@;
	say ($aconn, "priv set -q advanced; exportfs -i -o anon=0 /",'-eval_error');
	die $@ if $@;
}

# invoke the netcache "install" command on the filer and die if it
# fails.

sub do_install_cmd {
	my ($atelnet, $cmd, $timeout) = @_;
	say($atelnet, $cmd, '-blind');
	prompt_loop(connection => $atelnet,
				table => [

				 # (this "Do you want to continue" question is asked on C1200, C2100)
				  "Do you want to continue\\? \\(y/n\\) \\[.?\\]: " => "y",
				  "download failed" => sub { die "download failed (printed \"download failed\")\n"; },
				  "download: Failed" => sub { die "download failed (printed \"download: Failed\")\n"; },
				],
				done => ["You must now reboot the appliance"],
				max_tries => 5,
				say_opts => ['-timeout',$timeout]);
}

## @name ::download_files()
## @summary load kernel and system files and invoke 'download'

## @description
##   (1) Copy system files (using NFS, CIFS, or HTTP), then possibly
##   (2) copy a kernel using NFS (if the kernel requested is different
##   from what the system files would provide) and then (3) possibly
##   download if that didn't happen in step (1).
##
##   For netcaches, installs use the "install" command (and for NFS a
##   hidden license for NFS).  The CIFS installs use Visual Test
##   automation of setup.exe (see dependencies).  The HTTP install for
##   filers uses the "software" command.

## @synopsis
##   download_files($aspec, $cspec, [...options...])

## @arg aspec Required: hostspec of the filer on which install is to
## be done.

## @arg cspec Required: hostspec of the client which will copy ontap
## to the filer.  The copy will be done using CIFS if the hosttype is
## "windows" and using NFS, otherwise.  (Or, if ontap_url or
## ontap_base_url are provided, the filer will copy using HTTP and
## the value of this argument will be ignored).

## @arg prefix=&gt;... Optional: if given, defaults for some of the
## following arguments related to the filer or cache will be taken
## from parameters with the given prefix (if prefix is "FILER", for
## example, then the default for ontapdir will be taken from
## FILER_ONTAPDIR)

## @arg ontapdir=&gt;... Optional: The top-level ontap directory under
## which to find system files and kernel images.  Defaults to
## "$prefix_ONTAPDIR" if that exists, or ONTAPDIR in the host record
## of $aspec if that exists, otherwise is required if using NFS or
## CIFS in step (1) or when performing step (2) or when installing for
## simulators (to get the boot time filesystem image).

## @arg ontap_url Optional: The URL to the .zip or .exe file
## containing system files to use in an HTTP install (using the
## "install" command on caches or "software" command on filers).
## Defaults to "$prefix_ONTAP_URL" if that exists, or else to
## ONTAP_URL in the host record of $aspec if that exists, or else if
## option ontap_base_url has a value it defaults to the system files
## image implied by "ontap_base_url", "ontap_releases_dir", and "arch"
## options, or else there is no default, and install will use NFS or
## CIFS instead of HTTP.

## @arg ontap_base_url Optional: The URL to the top directory of an
## ontap tree.  Used as part of the default to ontap_url, so as to
## specify a version of ontap for HTTP install while not also
## specifying the specific architecture (that goes in "arch") or
## flavor of build (that goes in "ontap_releases_dir").  Defaults to
## "$prefix_ONTAP_BASE_URL" if that exists.

## @arg maytag_suffix Optional: The suffix of the kernel that is
## desired to be installed.  This value is required for simulators (to
## properly create the "bfs" link).  For non-simulators, if a value is
## available and the kernel with this suffix is not what the system
## files would contain, step (2) is performed.  Otherwise step (2) is
## skipped.  Defaults to param "$prefix_MAYTAG_SUFFIX", or else has no
## value.

## @arg arch Optional: The architecture of system files to install for
## this host: "x86", "alpha", "mips", or "sim".  Defaults to
## "$prefix_ARCH" if that exists or to "ARCH" in the host record of
## $aspec if that exists, or else a value will be guessed from the
## maytag_suffix option, or else a value will be guessed from the
## model option.  Because "sim" system files are rare and simulators
## do not actually care which architecture of system files are used,
## the guesses for simulators will be "x86" instead of "sim".

## @arg ontap_releases_dir Optional: The name of the directory to use
## under $ontapdir instead of "releases".  Possible values are
## "releases", "releases_debug", or "releases_pbo".  Defaults to
## "$prefix_ONTAP_RELEASES_DIR" if that exists or "ONTAP_RELEASES_DIR"
## in the host record of $aspec, or to "releases_debug" or
## "releases_pbo" if "maytag_suffix" has a value and contains "d" or
## "O" respectively, or else "releases".

## @arg download Optional: If set to 0, copy the system files but do
## not perform a subsequent "download" command if possible (a download
## may still be performed if installing for netcache, as the netcache
## "install" command does this inescapably).  The default is 1, which
## means to perform a subsequent download command.

## @arg install_system_files Optional: If set to 1 (the default), copy
## the system files (tar_floppy_$arch, zip_image_netapp.zip, or
## setup.exe).  If set to 0, only copy maytag_kern.xx and
## na_bootcard.xx files.

## @arg nfsvers Optional: The NFS version to use when installing using
## NFS.  Defaults to undef (let mount program choose version).

## @arg cifs_user Optional: For CIFS installs, the user to use when
## mapping the filer's "C$" share.  Defaults to "$prefix_CIFS_USER" or
## "CIFS_USER" in the host record of $aspec.

## @arg cifs_password Optional: For CIFS installs, the password to use
## when mapping the filer's "C$" share.  Defaults to
## "$prefix_CIFS_PASSWORD" or "CIFS_PASSWORD" in the host record of
## $aspec.

## @arg timeout Optional: Fail with an error if file copy and download
## together take longer than this many seconds.  Default is 1800 (30
## minutes).

## @arg brand Optional: the brand of maytag image to use.  Possible
## values are "netapp" or "ibm".  Defaults to "$prefix_BRAND" if that
## exists or to "BRAND" in the host record of $aspec if that exists,
## or else a value will be guessed from the model option.


## @dependencies
##   For filers, assumes that the root volume is exported and can be
##   mounted.
##

## @dependencies
##   For CIFS, $cspec requires the Visual Test runtime to be installed
##   (available in cifs/setup/runtime.exe).  $cspec must also be
##   running an rsh server which supports rcp and interactive use,
##   such as the NATE rshd service.  The rsh server must run as a user
##   who can access network shares and who has an associated display
##   (which rules out rsh being run as a service).  The drive letter
##   "n:" on $cspec will be mapped to the filer's root (which among
##   other things means two filers cannot be installed to in parallel
##   by one $cspec).  All of the other Windows on $cspec's display
##   (especially any console windows) should be closed or minimized or
##   the visual test script may be confused (causing it to stall and
##   this routine to time out).

## @bugs
##   It is not possible with the netcache "install" command to unpack
##   system files and download them as seprate steps, so in order to
##   install a kernel which has no corresponding system files (e.g. a
##   "3dA" kernel using "3d" system files) this function ends up
##   causing download twice, and the extra download (because the
##   "download" command is not normally available) is done using "priv
##   -set test", which means it will not work for promoted builds and
##   is not guaranteed to continue to work.  The normal way to get
##   nonstandard, internal builds to be downloaded is presumably to
##   modify the .zip file before installing it, but this libarary does
##   not take that approach.

## @see-also cifs/setup/start_setup.pc6
## @see-also cifs/setup/run_setup.pc6
## @see-also cifs/setup/runtime.exe

sub download_files ( $$@ ) {
    my ($aspec, $cspec, @opt) = @_;
	my %opt=@opt;

	return eval {
		my $ahost = make_host($aspec);
		my $arsh = make_conn($ahost,"#rsh");

		my $paramsref = dl_getargs($ahost,%opt);

		my $ontapdir = $paramsref->{ONTAPDIR};
		my $hosttype = $ahost->hosttype();
		my $maytag_suffix = $paramsref->{MAYTAG_SUFFIX};
		my $simdir = $paramsref->{SIMDIR};
		my $releases_dir = $paramsref->{ONTAP_RELEASES_DIR};
		my $brand = $paramsref->{BRAND};
                my $build_dir = ($brand eq "netapp") ? "prod/build" : "prod/build/$brand" ;

		my $endtime = Tharn::timeout2time(exists $opt{timeout} ? $opt{timeout} : 1800);
		my $download = (exists $opt{download} ? $opt{download} : 1);
		my $downloaded_already=0;
		my $install_system_files = (exists $opt{install_system_files} ? $opt{install_system_files} : 1);

		# First, copy and unpack software (and, if we can't avoid it,
		# download as well)

		if ($install_system_files) {
			my $res;
			($res, $downloaded_already) = copy_system_files($aspec,$cspec,@opt,params=>$paramsref,timeout=>$endtime);
			die $@ if $@;
		}

		if ($hosttype=~ /simulator/) {
			# Use #rlogin to the simulator, which talks to the simhost.
			# (other machines, such as simhost is the only host guaranteed to be able to see
			# files in $simdir).

			if (!defined($maytag_suffix)) {
				die "MAYTAG_SUFFIX must be provided for simulators (need to know which boot file system to link to)\n";
			}
			if (!defined($ontapdir)) {
				die "ONTAPDIR must be provided for simulators (need to know which boot file system to link to)\n";
			}
			my $sconn = make_conn($ahost,"#rlogin");
			my $to_remove = $simdir . "/bfs.$maytag_suffix";
			my $to_link = $ontapdir . "/prod/build/bfs.$maytag_suffix";
			say_confirm($sconn,
						"/usr/local/bin/sudo rm -f $to_remove");
			say_confirm($sconn,
						"/usr/local/bin/sudo ln -s $to_link $to_remove");
			$download = 0;

		} elsif (defined($maytag_suffix) &&
				 !$install_system_files ||
				 !same_maytag_as_release($maytag_suffix,$releases_dir,$brand)) {

			# We do this only if $install_system_files is false (so we
			# haven't yet installed new maytag_kern or na_bootcard
			# files) or if the ones we already installed were wrong
			# (because the $maytag_suffix requested doesn't match the
			# files files from the $releases_dir that we just used).

			copy_kernel($aspec,$cspec,@opt,params=>$paramsref,timeout=>$endtime);
			die $@ if $@;
			# Whatever downloads we might have been forced to do when
			# copying system files are invalidated again...
			$downloaded_already=0;
		}

		if ($download && !$downloaded_already) {
			do_download($aspec,$cspec,@opt,params=>$paramsref);
			die $@ if $@;
		}
		return 1;
	};
}

# This does step #1 of download_files(), where it installs the
# complete set of system files, but does not fix up the maytag_kern or
# do the download.  It has a reasonable interface for being exported
# on its own, though this hasn't been documented or tested.

sub copy_system_files {
	my ($aspec, $cspec, %opt) = @_;
	return eval {
		my $downloaded_already=0;
		my $ahost = make_host($aspec);
		my $arsh = make_conn($ahost,"#rsh");
		my $atelnet = make_conn($ahost,"#telnet");
		my $chost = make_host($cspec);
		my $paramsref = dl_getargs($ahost,%opt);
		my $ontap_url = $paramsref->{ONTAP_URL};
		my $ontap_base_url = $paramsref->{ONTAP_BASE_URL};
		my $nfsvers=$opt{nfsvers};
		my $endtime = Tharn::timeout2time(exists $opt{timeout} ? $opt{timeout} : 1800);

		# Map IP_7G or IP_GX to IP to fix burt 293543
		if (defined($paramsref->{IP_7G})) {
			$paramsref->{IP} = $paramsref->{IP_7G}; 
		} else {
			$paramsref->{IP} = $paramsref->{IP_GX}; 
		}

		if ($ahost->hosttype() =~ /cache/) {
			if ($ontap_url || $ontap_base_url) {
				netcache_install_HTTP($atelnet,$paramsref,$endtime);
			} else {
				netcache_install_NFS($ahost,$atelnet,$cspec,$paramsref,$endtime,$nfsvers);
			}
			$downloaded_already=1;
		} else {
         my $copy_is_complete = 0;
			if ($ontap_url || $ontap_base_url) {
            # Attempt to use software command, if not there, fall back to CIFS or NFS after returning here
				$copy_is_complete = copy_system_files_HTTP($arsh,$paramsref,$endtime);
			} 
         if ($copy_is_complete == 0) {
            if ($chost->hosttype() =~ /windows/) {
   				copy_system_files_CIFS($ahost,$cspec,$paramsref,$endtime);
   			} else {
   				copy_system_files_NFS($ahost,$cspec,$paramsref,$endtime,$nfsvers);
            }
			}
		}
                disconnect($atelnet);
		return wantarray ? (1, $downloaded_already) : 1;
	};
}

# (helper for copy_system_files, which does netcache and HTTP in particular)

sub netcache_install_HTTP ( $$$ ) {
	my ($atelnet, $paramsref, $endtime) = @_;
	my $rel_arch = rel_arch($paramsref->{ARCH});
	my $filename = "zip_image_$rel_arch.zip";
	my $url = dl_url($filename,$paramsref);
	do_install_cmd($atelnet,"install -dc $url",$endtime);
}

# (helper for copy_system_files, which does netcache and NFS in particular)

sub netcache_install_NFS ( $$$$$$ ) {
	my ($ahost,$atelnet,$cspec,$paramsref,$endtime,$nfsvers) = @_;
	my $chost = make_host($cspec);
	my $cconn = make_conn($cspec,"#rlogin");
	netcache_enable_nfs($atelnet);
	my $mountpoint = do_mount($paramsref->{IP},"/",$chost,$nfsvers)
		or die $@;
	my $rel_arch=rel_arch($paramsref->{ARCH});
	my $filename="zip_image_$rel_arch.zip";
	my $to_unzip = dl_filename($filename,$paramsref);
	say_confirm($cconn,
				"cd $mountpoint && r mkdir -p software && r cp $to_unzip software/$filename",
				'-timeout',$endtime);
	do_install_cmd($atelnet,
				   "install -c $filename",
				   $endtime);
}

# (helper for copy_system_files, which does filers and HTTP, or
# conceivably other URLs, in particular)

sub copy_system_files_HTTP ( $$$ ) {
	my ($arsh, $paramsref, $endtime) = @_;
	my $filename = "setup.exe";
	my $url = dl_url($filename,$paramsref);
   my $unsupported_command = 'software not found';
	my ($match,$before,$after) = say($arsh,"software install $url -f",
		'-wait', [
				  # Boilermaker or later
				  "Installation Complete\\.",
				  # Gordonbiersch.1 or later
				  "software: installation of .* completed",
				  # Scrimshaw, Anchorsteam, Gordonbiersch.0
				  "software: installation completed",
				  # FullSail
				  "software install: installation completed",
				  # pre-FullSail - this copy operation cannot be performed
				  $unsupported_command,
 				 ],
		'-timeout',$endtime,
		'-eval_error');
	die $@ if $@;
   if ($match =~ m/$unsupported_command/) {
      return 0;  # Ask for this to be retried with CIFS or NFS
   } else {
      return 1;
   }
}

# (helper for copy_system_files, which does filers and CIFS in
# particular)

sub copy_system_files_CIFS ( $$$$ ) {
	my ($ahost, $cspec, $paramsref, $endtime) = @_;
	my $chost = make_host($cspec);
	my $cconn = make_conn($cspec,"#rsh");
	my $cifs_user=$paramsref->{CIFS_USER};
	my $cifs_password=$paramsref->{CIFS_PASSWORD};
	my $ipaddr=$paramsref->{IP};

	# Connect a drive letter to filer's root.  Must use
	# drive n:, as this is hard-coded in Visual Test scripts.

	say($cconn,"net use n: /y /delete",'-eval_error');
	die $@ if $@;
	my $useroption = ($cifs_user) ? " /user:$cifs_user" : "";
	my $passwordoption = defined($cifs_password) ? " $cifs_password" : "";
	say($cconn,
		"net use n: \\\\$ipaddr\\c\$$useroption$passwordoption",
		'-wait',"command completed successfully",
		'-eval_error');
	die $@ if $@;

	# Copy setup.exe and the Visual Test scripts to run it
	rcp($chost,dl_filename("setup.exe",$paramsref),
		"n:\\setup.exe",$endtime)
		or die $@;
	foreach my $vtscript ("run_setup.pc6","start_setup.pc6") {
		my $from = findtest("cifs/setup/$vtscript")
			or die $@;
		rcp($chost,$from,"n:\\$vtscript",$endtime)
			or die $@;
	}

	# Delete logfile created by previous run of Visual Test script.
	say($cconn,"cmd /c del n:\\auto_setup.log",'-eval_error');
	die $@ if $@;
	say($cconn,"cmd /c type n:\\auto_setup.log",
		"The system cannot find the file specified",
		'-eval_error');
	die $@ if $@;

	# Invoke "run_setup.pc6" Visual Test script (which in
	# turn invokes start_setup.pc6?)

	say($cconn,"cmd /c n:\\run_setup.pc6",
		'-timeout', $endtime,'-eval_error');
	die $@ if $@;

	# Determine success from Visual Test script log
	say($cconn,"cmd /c type n:\\auto_setup.log",
		'-wait',"\[Suite Result\]       PASS",
		'-eval_error');
	die $@ if $@;
}


# (helper for copy_system_files, which does filers and NFS in
# particular)

sub copy_system_files_NFS ( $$$$$ ) {
	my ($ahost, $cspec, $paramsref, $endtime, $nfsvers) = @_;
	my $chost = make_host($cspec);
	my $cconn = make_conn($cspec,"#rlogin");

	# get root_vol
	my $arsh = make_conn($ahost,"#rsh");
	my $root = vol_root($arsh) or die $@;

	my $mountpoint = do_mount($paramsref->{IP},"/vol/$root",$chost,$nfsvers)
		or die $@;
	my $rel_arch=rel_arch($paramsref->{ARCH});
	my $to_untar = dl_filename("tar_floppy_$rel_arch",$paramsref);
	my $sendstr = "cd $mountpoint && /usr/local/bin/sudo tar xvf $to_untar && echo _''OK_" ;
        my (undef, $before) = say($cconn, $sendstr, "_OK_", '-timeout',
                                $endtime,'-eval_error');
	print "undef and before is $before \n";
        die "Command \"$sendstr\" failed: $@" if $@;
        my ($install_script) = ($before =~ /(install_\w+)/);
        if (!$install_script) {
                die("Could not find \"install_...\" script in \"tar\" command   output. Output was: $before\n");
        }

	# Remove the tar_image that might already exist in the
	# filer root, or else install_netapp might prompt us
	# about whether to remote it.
	say($cconn, "cd $mountpoint && /usr/local/bin/sudo rm -f tar_image",
		'-eval_error');
	die $@ if $@;
	say_confirm($cconn, "cd $mountpoint && /usr/local/bin/sudo ./$install_script -k .",
                    '-timeout', $endtime);
}

# Copy appropriate maytag_kern file, as well as na_bootcard

# This does step #2 of download_files(), but it does it always
# (download_files is responsible for calling this only when needed).

# This function has a reasonable interface for being exported
# on its own, though this hasn't been documented or tested.

sub copy_kernel {
	my ($aspec,$cspec,%opt) = @_;

	return eval {
		my $ahost = make_host($aspec);
		my $arsh = make_conn($ahost,"#rsh");
		my $paramsref = dl_getargs($ahost,%opt);
		my $endtime = Tharn::timeout2time(exists $opt{timeout} ? $opt{timeout} : 1800);

		my $maytag_suffix = $paramsref->{MAYTAG_SUFFIX};
		my $ontapdir = $paramsref->{ONTAPDIR};
		my @Dir = split(/\//, $ontapdir); #split on the slash 
		my $release = pop(@Dir); #being the last element
		my $brand = $paramsref->{BRAND};
                my $build_dir = ($brand eq "netapp") ?  "prod/build" : "prod/build/$brand" ;

		if (!$maytag_suffix) {
			die "param MAYTAG_SUFFIX required (needed to find maytag_kern file to copy)\n";
		}
		if (!$ontapdir) {
			die "param ONTAPDIR required (needed to determine directory from which to copy maytag_kern.$maytag_suffix)\n";
		}

		my $chost = make_host($cspec);

		if ($chost->hosttype() =~ /windows/) {
			# generalizing the below to use rcp is horrid, and we
			# wouldn't expect it to be used (CIFS installs are
			# mainly for testing setup.exe, so one presumably
			# always wants exactly the kernels it provides)
			die("Installing individual maytag files using Windows not supported".
				"(use UNIX, or have maytag_suffix match what setup.exe gives you).");
		}

		my $cconn = make_conn($cspec,"#rlogin");

		my $path;
		if ($ahost->hosttype() =~ /cache/) {
			netcache_enable_nfs($arsh);
			$path = "/";
		}
		if ($ahost->hosttype() =~ /filer/) {
			# get root volume
			my $root = vol_root($arsh) or die $@;
			$path = "/vol/$root";
		}
		my $mountpoint = do_mount($paramsref->{IP},$path,$chost,$opt{nfsvers})
			or die $@;

		my $link_arch = link_arch($paramsref->{ARCH});

		my @copyandlink = ("netapp-$link_arch" => 
						   "maytag_kern.$maytag_suffix");

		# (It's hard to know when the boot card is really needed or was
		# able to be built and when it's not, so let's just copy it if we
		# can find it and not copy it if we can't)
		if (-e "$ontapdir/$build_dir/na_bootcard-$link_arch.$maytag_suffix") {
			push @copyandlink, ("na_bootcard-$link_arch" => 
								"na_bootcard-$link_arch.$maytag_suffix");
		}
		#
		# Copy the mapfile for this build when installing a new build
		if (-e "$ontapdir/$build_dir/mapfile.$maytag_suffix") {
			push @copyandlink,("mapfile_$release.$maytag_suffix" =>
								"mapfile.$maytag_suffix");
		}

		while (@copyandlink) {
			my $link = shift @copyandlink;
			my $regfile = shift @copyandlink;
			say($cconn,
				"cd $mountpoint/etc/boot/ && \\\r".
				"/usr/local/bin/sudo rm -f $link && \\\r".
				"/usr/local/bin/sudo cp $ontapdir/$build_dir/$regfile $regfile && \\\r" .
				"/usr/local/bin/sudo ln -s $regfile $link && \\\r".
				"echo _''OK_",
				'-wait', "_OK_",'-eval_error',
                '-timeout',$endtime);
			die "failed to copy and link $regfile: $@" if $@;
		}
	};
}

# Invoke the download command and confirm success.

# This does step #3 of download_files(), and does it always
# (download_files is responsible for calling this only when needed)

# This function has a reasonable interface for being exported
# on its own, though this hasn't been documented or tested.

sub do_download {
 	my ($aspec, $cspec, %opt) = @_;
	return eval {
		my $ahost = make_host($aspec);
		my $arsh = make_conn($ahost,"#rsh");
		my $atelnet = make_conn($ahost,"#telnet");		
		my $paramsref = dl_getargs($ahost,%opt);
		my $endtime = Tharn::timeout2time(exists $opt{timeout} ? $opt{timeout} : 1800);

		my $cmd = ($ahost->hosttype() =~ /cache/) ?
			"priv set -q test ; download" : "download";
        my $copied_to_disk=0;
		my $download_failure;

        say($atelnet, $cmd, '-blind');

		PROMPT_LOOP: {
            my $table = [
                         "Operator requested download completed" =>
                         sub {
                             # Does "completed" imply "successfully"?
                             # We have to assume so unless some other
                             # recognizable failures messages are
                             # seen.
                             last PROMPT_LOOP;
                         },
                         "Operator requested download failed" =>
                         sub {
                             my $reason = $&;
                             $download_failure=fatal_result("download syslog message printed '$reason'");
                             last PROMPT_LOOP;
                         },
                         "download: Failed[^\\n]*\\n" =>
                         sub {
                             my $reason = $&;
                             if ($reason =~ quotemeta("download: Failed to open source file /etc/boot/x86_elf/diag/diag.krn: No such file or directory") ||
                                 $reason =~ quotemeta("download: Failed to complete download instruction at line number 3 of download")) {
                                 # Report this as a failure and
                                 # continue (it was a known bug, but
                                 # not a bug that prevents
                                 # downloading)
                                 ExitSoftError('-burt',147687,"Missing /etc/boot/x86_elf/diag/diag.krn");
                             
                             } elsif ($reason =~ quotemeta("download: Failed to backup primary kernel image 1")) {
                                 # Ignore this particular error, as it
                                 # legitimately occurs for 7G ontap
                                 # (IC) downloading a clustered ontap
                                 # (BR) image for upgrade.  See burt
                                 # 329951
                                 
                             } else {
                                 # download failed, but we still need
                                 # to keep listening for output until
                                 # the download command finishes and
                                 # the "completed" or "failed" syslog
                                 # messages appear
                                 $download_failure=fatal_result("download printed '$reason'");
                             }
                             return undef;
                         },
                         "Most installations will\\s*complete in 20-60 minutes" => 
                         sub {
                             # This build will take longer than normal
                             # to download (upgrade to clustered
                             # ontap), and so the default timeout
                             # needs to be increased above normal.
                             # Let's choose 90 minutes (5400 seconds).
                             $endtime = Tharn::timeout2time(exists $opt{timeout} ? $opt{timeout} : 5400);
                             redo PROMPT_LOOP;
                         },
                        ];
            prompt_loop(connection=>$atelnet,
                        table => $table,
                        done=>[],
                        max_tries=>10,
                        say_opts => ['-timeout' => $endtime]);
        };

        if ($download_failure) {
            die $download_failure;
        }
        disconnect($atelnet);
	};
 }

# Common function for getting arguments or params used by #
# download_files() (or copy_system_files, copy_kernel, or do_download
# if those are exported).  Also happens to work for check_version, so
# we're using it there as well.

sub dl_getargs ( $% ) {
	my ($ahost, %opt) = @_;
	if ($opt{params}) {
		return $opt{params};
	} else {
		my %params = param_merge(
								 hostrec => $ahost,
								 param_prefix => $opt{prefix},
								 hash_uc => \%opt,
								 );
		($params{MAYTAG_SUFFIX},$params{ARCH},$params{ONTAP_RELEASES_DIR},$params{BRAND}) =
			guess_buildtype(
							$params{MAYTAG_SUFFIX},
							$params{ARCH},
							$params{ONTAP_RELEASES_DIR},
							$ahost->hosttype(),
							$params{MODEL},
							$params{BRAND}
						   );
		return \%params;
	}
}

# Unlike alpha, mips, or x86-64, x86 images have two strings: "pc" or
# "pc_elf" are used to refer to the architecture in filenames in the
# "releases/" directory (what this library calls REL_ARCH), and "x86"
# is used in the kernel filename which the 'download' command expects
# to find (what this library calls LINK_ARCH).

# Before "pc_elf" existed, there was an unambiguous conversion between
# the two, and we preferred to default to ARCH=x86 over ARCH=pc
# (rel_arch and link_arch would be happy with either one, and "x86"
# is/was a more sensible name).

# However, now "pc_elf" also exists, and it also has a LINK_ARCH of
# x86 (even though the -3/pc or -8/pc_elf executables that this
# filename links to cannot be substituted for one another.  grumble).
# For backwards compatibility, ARCH=x86 or ARCH=pc are guessed to mean
# "pc", and if people want REL_ARCH to be "pc_elf", they have to pass
# ARCH=pc_elf.

# REL_ARCH is ARCH, but "pc" instead of "x86".

sub rel_arch {
	my ($arch) = @_;
	if (!$arch) {
		die "Unknown architecture (ARCH not provided and could not be guessed from MAYTAG_SUFFIX or MODEL)\n";
	}
	return ($arch eq "x86") ? "pc" : $arch;
}

# LINK_ARCH is ARCH, but "x86" instead of "pc" or "pc_elf"

sub link_arch {
	my ($arch) = @_;
	if (!$arch) {
		die "Unknown architecture (ARCH not provided and could not be guessed from MAYTAG_SUFFIX or MODEL)\n";
	}
	return ($arch =~ /^(pc|pc_elf)$/) ? "x86" : $arch;
}
	

# Return the full path to a file in the appropriate "dist" directory
# (appropriateness determined by arguments to download_files() or one
# if its minions: copy_system_files, copy_kernel, or do_download)

sub dl_filename {
	my ($filename,$paramsref) = @_;
	if (!$paramsref->{ONTAPDIR}) {
		die "param ONTAPDIR required (needed to find \"$filename\" in order to install ontap)\n";
	}
	return join("/",
				$paramsref->{ONTAPDIR},
				$paramsref->{ONTAP_RELEASES_DIR},
				rel_arch($paramsref->{ARCH}),
				"dist",
				$filename);
}

# Return the full URL to a file in the appropriate "dist" directory
# (appropriateness determined by arguments to download_files() or one
# of its minions: copy_system_files, copy_kernel, or do_download).
# Yes, if ONTAP_URL is defined it returns the complete URL, ignoring
# whatever $filename is.

sub dl_url {
	my ($filename,$paramsref) = @_;
	my $url=$paramsref->{ONTAP_URL};
	if (!$url) {
		$url=join("/",
				  $paramsref->{ONTAP_BASE_URL},
				  $paramsref->{ONTAP_RELEASES_DIR},
				  rel_arch($paramsref->{ARCH}),
				  "dist",
				  $filename);
	}
	return $url;
}

## @name ::guess_buildtype
## @summary fill in values for whichever of MAYTAG_SUFFIX, ARCH, 
## ONTAP_RELEASES_DIR and BUILD have no value
##
## @synopsis ($maytag_suffix, $arch, $ontap_releases_dir,$brand) = guess_buildtype($maytag_suffix,$arch,$ontap_releases_dir,$hosttype,$model,$brand)
##
## @arg $maytag_suffix optional: 
## @arg $arch optional: 
## @arg $brand optional:
## @arg $ontap_releases_dir optional: 
## @arg $hosttype optional: required:
## @arg $model optional: 
##
## @usage
##   $arch is guessed to be "x86" for simulators instead of "sim",
##   since no nightly build builds "sim" architecture files, and which
##   architecture of these is actually is used doesn't too much
##   matter.
##
##   $maytag_suffix cannot be guessed for simulators
##
## @author lincoln

sub guess_buildtype {
	my ($maytag_suffix, $arch, $ontap_releases_dir, $hosttype, $model,$brand) = @_;

	if (!$arch && $maytag_suffix) {
		$arch="pc";
		$arch="alpha" if ($maytag_suffix =~ /1/);
		$arch="x86-64" if ($maytag_suffix =~ /2/);
		$arch="mips" if ($maytag_suffix =~ /6/);
		$arch="pc_elf" if ($maytag_suffix =~ /8/);
	}
	if (!$arch && $model) {
		my $mi = new ModelInfo($model);
		$arch=$mi->{arch};
	}
	if (!$arch && ($hosttype =~ /simulator/)) {
		$arch="pc";
	}
	 if (!$brand && $model) {
                my $mi = new ModelInfo($model);
                $brand = $mi->{brand};
        }

        if (!$brand) {
                $brand = "netapp";
        }

	if (!$ontap_releases_dir) {
		$ontap_releases_dir="releases";
		if ($maytag_suffix) {
			$ontap_releases_dir="releases_debug" if ($maytag_suffix =~ /d/);
			$ontap_releases_dir="releases_pbo" if ($maytag_suffix =~ /O/);
		}
	}
	if (!$maytag_suffix && $arch && $ontap_releases_dir) {
		$maytag_suffix="1" if ($arch eq "alpha");
		$maytag_suffix="2" if ($arch eq "x86-64");
		$maytag_suffix="2" if ($arch eq "amd64");
		$maytag_suffix="3" if ($arch eq "pc");
		$maytag_suffix="3" if ($arch eq "x86");
		$maytag_suffix="6" if ($arch eq "mips");
		$maytag_suffix="8" if ($arch eq "pc_elf");
		$maytag_suffix.="d" if ($ontap_releases_dir =~ /debug/);
		$maytag_suffix.="O" if ($ontap_releases_dir =~ /pbo/);
	}
	return ($maytag_suffix, $arch, $ontap_releases_dir,$brand);
}

# return true if the maytag with $maytag_suffix as its compile flags
# and $brand as its brand would be one of the main ontap builds
# installed by the system files archives in $releases_dir.  The focus
# is less on architecture (1 vs. 3. vs. 6) and more on extra flags
# (nothing vs. 'd' vs. 'O' vs. something else)

sub same_maytag_as_release {
        my ($maytag_suffix,$releases_dir,$brand) = @_;
        foreach my $tuple (
                                ["releases",qr/^\d$/,"netapp"],
                                ["releases_debug",qr/^\dd$/,"netapp"],
                                ["releases_pbo",qr/^\dO$/,"netapp"],
                                ["releases_ibm",qr/^\d$/,"ibm"],
                        ) {
                my ($t_dir, $t_maytag_suffix, $t_brand) = @$tuple;
                return 1 if ($releases_dir eq $t_dir &&
                        $maytag_suffix =~ $t_maytag_suffix &&
                        $brand eq $t_brand);
        }
        return 0;
}

# (Search @INC for a test named "$filename".  Return the full path to
# that test on success, or undef on failure with an error message in
# $@)

sub findtest {
	my ($filename) = @_;
 	return eval {
		foreach my $dir (@INC) {
			my $fullpath = "$dir/$filename";
			return $fullpath if (-e $fullpath);
		}
		die "Could not find \"$filename\" in \@INC (" . join(" ",@INC) . ")\n";
	};
}

# Copy $fromfile on local host to $tofile on $chost, by invoking rcp.

# We use say() instead of system() so that we can detect timeouts.
# Copying 57MB setup.exe to a simulator from a moderately fast Win2k
# client takes about 120 seconds, so the timeout is set to 300
# seconds.  In most situations it should take far less time.

sub rcp {
	my ($chost,$fromfile,$tofile,$timeout) = @_;
	my $h = host("localhost");
	my $c = connect("#rsh");
	my $dest = $chost->default_ip();
	return say($c,"rcp $fromfile $dest:$tofile && echo _OK_",
			   '-timeout',$timeout,
			   '-wait',"_OK_",'-eval_error');
}

## @name ::disk_assign()
## @summary assign disks in maintenance mode
## @description
##   assume we are already in maintenance mode, run disk assign
##
## @synopsis 
##   disk_assign($aspec, [command=&gt;], [...boot options...]);
##
## @arg aspec Required: hostspec of the filer to zero
##

## @arg * Optional: all other options are passed to transit_to() in
## lib/boot.pm to control how the filer reboots to the boot menu (what
## kernel to boot, whether to power cycle on failure, etc.).  See
## lib/boot.pm::transit_to for more information.

## @see-also lib/boot.pm::transit_to

sub disk_assign {
    my ($aspec, %opt) = @_;
    my $output;
    my $P_MAINT = "\\*\>";
	return eval {
		my $ahost = make_host($aspec);
		my $aconn = make_conn($ahost,"#console");

        (undef, undef, $output) = say($aconn, "disk show", '-prompt', $P_MAINT);
        my $system_id = 0;
        if ( $output =~ /Local System ID: ([+-]?\d+)/ ) {
           $system_id = $1;
        }

        (undef, undef, $output) = say($aconn, "disk assign all -s $system_id", '-prompt', $P_MAINT) if ( $system_id ne 0 );
}
}

## @name ::newfs()
## @summary Install new filesystem (4 or 24/7)
## @description
##   (Re)boot to the boot menu and install a new filesystem.  Waits
##   for disk zeroing to finish, but returns just before the setup
##   prompts.
##
## @synopsis 
##   newfs($aspec, [command=&gt;], [...boot options...]);
##
## @arg aspec Required: hostspec of the filer to zero
##

## @arg root_vol_type Optional: What to answer if asked to choose
## between traditional volumes ("t") or flexible volumes ("f").  The
## default is "t" for traditional.

## @arg command Optional: what command to use at the boot menu to
## initialize a new filesystem (such as "4", "4a", "24/6", "24/7", or
## "24/8").  The default is "4" if root_vol_type is "t" or "4a" if
## root_vol_type is "f".

## @arg * Optional: all other options are passed to transit_to() in
## lib/boot.pm to control how the filer reboots to the boot menu (what
## kernel to boot, whether to power cycle on failure, etc.).  See
## lib/boot.pm::transit_to for more information.

## @see-also lib/boot.pm::transit_to

sub newfs {
    my ($aspec, %opt) = @_;
	return eval {
		my $ahost = make_host($aspec);
		my $aconn = make_conn($ahost,"#console");

		my $root_vol_type = $opt{root_vol_type} || 't';
		delete $opt{root_vol_type};
		my $default_command = ($root_vol_type eq "t") ? "4" : "4a";
		my $command = $opt{command} || $default_command;
		delete $opt{command};
		# Reach the boot menu
		transit_to($aconn,BOOT_MENU,iftakenover=>'takeback',%opt);

		# If transit_to fails, maybe filer is in initial
		# setup.  Go through setup to attempt to escape.
		
		if ($@) {
			basic_setup($aconn,%opt); 
			transit_to($aconn,BOOT_MENU,iftakenover=>'takeback',%opt) or die($@); 
		}		    

		# Ask to install a new filesystem and confirm
		say($aconn,$command,'-blind','-eval_error');
		my @prompts_confirm_zero =
			(
			 "Install new file system without zeroing disks\\?" => sub { "yes" },
			 "Install a new file system\\?" => sub {"yes" },
			 "Zero disks(, reset config)? and install a new file system\\?" => sub { "yes" },
			 "REALLY install file system without zeroing disks\\?" => sub { "yes" },
			 "This will erase all the data on the disks, are you sure\\?" => sub {"yes" },
			 "Please select one \\('t' or 'f'\\): " => $root_vol_type,
			 );

		my $ret_prompt =
			prompt_loop(connection => $aconn,
						table => \@prompts_confirm_zero,
						done => ["Zeroing disks takes about \\d+ minutes",
								 "Zeroing disks\\.",
								 "There are \\d+ disks in the system"],
						max_tries => 5,
                        say_opts => ['-timeout',Tharn::timeout2time($opt{timeout} || 120)],
					   );
		my ($filer_est_time) = 
			($ret_prompt =~ /Zeroing disks takes about (\d+) minutes/);

		my @prompts_fs_layout;
		my $zero_timeout; 

		if ($command eq "24/7") {
			# Now, if 24/7, it will ask us questions about number of
			# spares, volumes, raid groups.  For compatibility between
			# (4) and (24/7), and to be conservative, always choose
			# the maximum number of spares.

			# Note that it may not ask us for number of spares (seen
			# on a C1100 with 1 disk.  I do not know if this also
			# happens on filers and/or with exactly 2 disks)

			# Note also that it may, if there are broken disks, lie to
			# us the first time when it says "You can have up to N
			# spares", so we compute the real, (presumably lower)
			# number of disks from the error message given in that
			# case and let ourselves go through the same prompts
			# again.

			# Note that "sub { $spares; }" is used instead of just
			# "$spares", so that the value of $spares used will be its
			# value when that prompt is hit, not just when
			# @prompts_fs_layout is first created.

			# Order matters in the following (the first three messages
			# may come in almost on top of each other, so we list the
			# first ones that will come first so that they'll all be
			# seen)

			# $spares is initialized with an extremely large number.
			# The specific value should not matter, since a prompt to
			# replace it with a valid value should occur before a
			# prompt which requires its value.  (This used to be
			# Infinity, but I cannot find an obvious constant for that
			# which is portable between v5.6 and v5.8.  1.2e+999 is
			# large enough to be a non-obvious constant for Infinity,
			# and is in any case larger than any actual number of
			# spares could be)

			my $spares=1.2e+999;
			@prompts_fs_layout =
				(
				 "\\d+ spares, \\d+ vols, \\d+ plexes, \\d+ groups = \\d+; avail = (\\d+)"
				     => sub { $spares = ($1 - 2); return undef; },
				 "You can have up to (\\d+) spares" => sub { $spares = $1 if ($1<$spares); return undef; },
				 "How many disks should be designated as spares \\[\\d+\\]: " => sub { $spares; },
				 "How many volumes \\[\\d+\\]: " => "",
				 "Number of RAID groups per volume \\[\\d+\\]: " => "",
				);

			# Depite 24/7 being supposedly nearly instant, we choose a
			# 600 second timeout.  This is to wait for the filer or
			# simulator to potentially attempt to configure using DHCP
			# (occurs after the above loop of prompts) which takes
			# several minutes before giving up if DHCP not available.
			# Also, fake zeroing disks may take longer than a minute
			# for simulators.
			$zero_timeout=600;
		} else {
			@prompts_fs_layout =
				(
				 "Do you want to use RAID DP\\?" => "",
				 "How many RAID groups do you want to create" => "",
				 "RAID group size" => "",
				 "Do you want to initialize NetCache with these parameters\\?" => "",
				 );

			# calculate a timeout based on the filer's estimate to
			# zero the disks.  Allow half again as much time as the
			# filer estimates, plus 1/2 hour (the extreme case being a
			# simulator which might take that long even though it
			# estimates 0 minutes)
			$zero_timeout=(($filer_est_time*60) * 1.5) + 1800; 
		}

		logcomment("Waiting up to $zero_timeout seconds for zeroing to finish");

		# Keep answering prompts until we see "NetApp Release",
		# which is printed as part of sysconfig just before setup.

		# The ignore of "auto_mode_setup" is to ignore a
		# particular dbg_printf which has a tendancy to break up
		# "NetApp Release".  See bug 156312.
		prompt_loop(connection => $aconn,
					table => \@prompts_fs_layout,
					done => ["NetApp Release",
							 "Data ONTAP Release",
							],
					max_tries => 15,
					say_opts => ['-timeout',$zero_timeout,
								 '-ignore',["auto_mode_setup:  \\.\\.\\.\\r*\\n"],
								],
				   );

        return 1;
      };
}


## @name ::make_sim_files
## @summary Populate a simulator directory with disks and configuration

## @description
##   This routine creates the various files that a simulator needs:
##   ",disks" and ",tapes" directory, disk files in ",disks",
##   ",mac-address", ",serialno", and for clusters the
##   ",local-interconnect" file and ",partner" symlink to the
##   partner's simulator directory.
##
##   See
##      <A HREF="http://web.netapp.com/engineering/hhg/Software-4.0/Procedures-3.0/Simulator.html">http://web.netapp.com/engineering/hhg/Software-4.0/Procedures-3.0/Simulator.html</A>
##   for more infromation about the files simulators need.

## @synopsis
##   make_sim_files($aspec
##                  [,prefix=&gt;...]
##                  [,simdir=&gt;...]
##                  [,system_id=&gt;...]
##                  [,mac_address=&gt;...]
##                  [,makedisks=&gt;...]
##                  [,partner=&gt;...]
##                  [,partner_prefix=&gt;...]
##                  [,partner_simdir=&gt;...]
##                  [,partner_system_id=&gt;...]
##                 );

## @arg $aspec Required: hostspec of the simulator to create

## @arg prefix Optional: if given, various other options to this
## function will take their defaults from NATE parameters beginning
## with this prefix (e.g. if prefix=&gt;"FILER" is given, simdir will
## default to $FILER_SIMDIR).

## @arg simdir Required: directory to populate (directory in
## which the simulator will run).  A value is required, but will
## default to "$prefix_SIMDIR" if prefix option is given and that
## value is present, or "simdir" in the simulator's host record if
## that value is present.

## @arg system_id Optional: System ID for this simulator (can be nearly
## any natural number, but should differ between this host and its
## partner).  A value is required when setting up CFO, but will
## default to "$prefix_SYSTEM_ID" if prefix option is given and that
## value is present, or "system_id" in the simulator's host record if
## that value is present.

## @arg mac_address Required: mac address for this simulator.  A
## value is required, but will default is "$prefix_MAC_ADDRESS" if prefix
## option is given and that value is present, or "mac_address" in the
## simulator's host record if that value is present.

## @arg makedisks Optional: command-line to use to create
## ",disks" files, usually "makedisks.main" followed by appropriate
## arguments.  Defaults to "$prefix_MAKEDISKS" if prefix option is
## given and that value is present, or "makedisks" in the simulator's
## host record, which NATE provides a default for (currently
## "makedisks.main").

## @arg partner Optional: for CFO simulators, hostspec of
## $aspec's partner.  By default, simulator is assumed to not be CFO,
## and the following CFO options are ignored.

## @arg partner_prefix Optional: As for "prefix" option, but
## parameters with the given prefix provides defaults for options
## related to the partner of this simulator rather than this
## simulator.

## @arg partner_simdir Required: directory to populate
## (directory in which the simulator will run).  A value is required
## when setting up for CFO (i.e. when "partner" option is given), but
## will default to "${partner_prefix}_SIMDIR" if prefix option is
## given and that value is present, or "simdir" in the partner's host
## record if that value is present.

## @arg partner_system_id Required: System ID for the partner (can be
## nearly any natural number, but should differ from that of the
## simulator being set up).  A value is required when setting up CFO,
## but defaults to "${partner_prefix}_SYSTEM_ID" if partner_prefix
## option is given and that value is present, or "system_id" in the
## partner's host record if that value is present.

##
## @arg timeout (Optional): Defaults to the connection timeout 
##

## @usage
##   The return value will be 1 on success, or undef on failure with
##   $@ set to an error message describing the failure.

## @example
##    # Make one simulator with many disks, some 512bps and some 520bps
##
##    make_sim_files($h,prefix=&gt;"FILER",makedisks=&gt;"makedisks.main -t 8 -n 28 && makedisks.main -t 21 -n 28") or logresult('fatal',$@);

## @example
##    # Make a CFO pair
##
##    param('FILERA','-mesg',"hostspec of Filer A");
##    param('FILERB','-mesg',"hostspec of Filer B");
##    make_sim_files($FILERA,partner=&gt;$FILERB,prefix=&gt;"FILERA",partner_prefix=&gt;"FILERB") or logresult('fatal',$@);
##    make_sim_files($FILERB,partner=&gt;$FILERA,prefix=&gt;"FILERB",partner_prefix=&gt;"FILERA") or logresult('fatal',$@);

## @author lincoln@netapp.com

sub make_sim_files {
	my ($aspec,%opt) = @_;
	return eval {
		my $ahost = make_host($aspec);
		my $aconn = make_conn($ahost,"#rlogin");
		my %aparams = param_merge(
								  hostrec=>$ahost,
								  param_prefix=>$opt{prefix}
								 );
		my $phost = make_host($opt{partner}) if ($opt{partner});
		my $asimdir = ($opt{simdir} || 
					   $aparams{SIMDIR} ||
					   die "Value must be provided for local simulator directory (simdir)\n");
		my $asystem_id = ($opt{system_id} ||
						 $aparams{SYSTEM_ID});
		my $mac_address = ($opt{mac_address} ||
					   $aparams{MAC_ADDRESS} ||
					   die "Value must be provided for MAC Address (mac_address)\n");
		my $makedisks = ($opt{makedisks} ||
						 $aparams{MAKEDISKS} ||
						 die "Value must be provided for makedisks\n");
		my ($psimdir,$psystem_id);
		my $timeout = defined($opt{timeout}) ? $opt{timeout} : $aconn->{timeout};
		my $endtime = Tharn::timeout2time($timeout);
		if ($phost) {
			if (!defined($asystem_id)) {
				die "Value must be provided for local System ID (system_id) when creating a CFO simulator";
			}
			my %pparams = param_merge(
									  hostrec=>$phost,
									  param_prefix=>$opt{partner_prefix}
									 );
			$psimdir = ($opt{partner_simdir} ||
						$pparams{SIMDIR} ||
						die "Value must be provided for partner's simulator directory (simdir)\n");
			$psystem_id = ($opt{partner_system_id} ||
						  $pparams{SYSTEM_ID} ||
						  die "Value must be provided for partner's System ID (system_id)\n");
		}
		# Change directory to the sim directory
		say_confirm($aconn,"cd $asimdir",'-timeout',$endtime);
                # Capture contents of the sim directory
                my @sim_conf = <$asimdir/*>;
                # Clean up sim directory; leave cores and mac-address.
                foreach(@sim_conf){
                        if($_ !~ m/core|mac-address/){
                                say_confirm($aconn,"rm -rf $_ || true",'-timeout',$endtime);
                        }
                }
                # Restore minimum sim directory configuration
                say_confirm($aconn,"mkdir ,disks",'-timeout',$endtime);
                say_confirm($aconn,"mkdir ,tapes",'-timeout',$endtime);
                say_confirm($aconn,"echo '$mac_address' > ,mac-address",'-timeout',$endtime);
                # Run makedisks
		say_confirm($aconn,$makedisks,'-timeout',$endtime);
		# (yes, the file ",serialno" is used for what is now known as the "System ID")
		if ($asystem_id) {
			say_confirm($aconn,"echo '$asystem_id' > ,serialno",'-timeout',$endtime);
		}
		if ($phost) {
			say_confirm($aconn,"echo '$psystem_id' >> ,serialno",'-timeout',$endtime);
			say_confirm($aconn,"dd if=/dev/zero of=,local-interconnect bs=1k count=1k",'-timeout',$endtime);
			say_confirm($aconn,"ln -s '$psimdir' ,partner",'-timeout',$endtime);
		}
		return 1;
	};
}

# say() a shell command to a unix host and die if it returns non-zero
# exit status.

sub say_confirm {
	my ($conn,$sendstr,@rest) = @_;
	say($conn,"$sendstr && echo _''OK_","_OK_",@rest,'-eval_error');
	die "Command \"$sendstr\" failed: $@" if $@;
}

## @name ::check_version
## @summary confirm that ontap version installed is as expected

## @description
##   This checks that the output of the 'version' command on the filer
##   matches the version given (or a version taken from the
##   appropriate vers-*.c file in the ontap build directory)

## @synopsis
##    check_version ($aspec, ...options...);

## @arg aspec Required: A hostspec or connection record of the
## appliance to be licensed.  If a host spec is given, the "#rsh"
## connection will be used.

## @arg prefix=&gt;... Optional: if given, various other options to this
## function will take their defaults from NATE parameters beginning
## with this prefix (e.g. if prefix=&gt;"FILER" is given,
## ontap_version will default to $FILER_ONTAP_VERSION).

## @arg ontap_version=&gt;... Optional: The expected ontap version,
## including product name and possibly including the date.  The filer
## is considered to have the correct version installed if this string
## exactly matches the beginning of one of the lines of "version"
## command output.  Useful values might be
##   <pre>NetApp Release 1.0:</pre>
## to ignore the exact date or something like
##   <pre>NetApp Release /x/eng/build/daemon/main/240651: Thu Jan  9 19:08:51 PST 2003</pre>.
## to include it.  If no value is provided, the default is the value
## assigned to the variable "shortVersion" in the corresopnding vers-XX.c
## file.

## @arg maytag_suffix=&gt;... Optional: build flags of the expected
## ontap version.  This is required only if ontap_version is not
## provided; it is used to determine which vers-XX.c file to use.  The
## default is the value of param $prefix_ONTAP_VERSION if that exists,
## or the field "ONTAP_VERSION" in the host record of $aspec if that
## exists, or else if not a simulator a value may be guessed from
## "ARCH", "ONTAP_RELEASES_DIR", and/or "MODEL" options (lower case),
## parameters (upper case with $prefix prepended), or host record
## fields (upper case).

## @arg ontapdir=&gt;... Optional: top-level ontap directory of the
## expected ontap build.  This is required only if ontap_version is
## not provided; it is used to determine which vers-XX.c file to use.
## The default is the value of param $prefix_ONTAPDIR if that exists,
## or the field "ONTAPDIR" in the host record of $aspec if that exists
## (for consistency, not recommended), or the value of param $ONTAPDIR
## if that exists.

## @author lincoln@netapp.com

sub check_version ( $@ ) {
    my ($aspec, %opt) = @_;
	return eval {
		my $ahost = make_host($aspec);
		my $aconn = make_conn($ahost,"#rsh");

		my $paramsref = dl_getargs($ahost,%opt);

		my $version = $paramsref->{ONTAP_VERSION};
		my $brand = $paramsref->{BRAND};
                my $build_dir = ($brand eq "netapp") ? "prod/build" : "prod/build/$brand" ;
		if (!$version) {
			foreach my $p (qw(ONTAPDIR MAYTAG_SUFFIX)) {
				if (!$paramsref->{$p}) {
					die "cannot verify desired ontap version: no value provided for $p";
				}
			}
			my $ontapdir=$paramsref->{ONTAPDIR};
			my $maytag_suffix=$paramsref->{MAYTAG_SUFFIX};
            # <8> builds use the <3> version file.  See burt 150413
            $maytag_suffix =~ s/8/3/g;
			$version=guess_version("$ontapdir/$build_dir/vers-$maytag_suffix.c") or die $@;
		}
		my (undef, undef, $output) = say($aconn,"version",'-eval_error');
		die $@ if $@;
		if ($output !~ ("(?m:^".string2regex($version).")")) {
			die "Desired ontap version (\"$version\") did not match version command output (\"$output\")\n";
		}
		return 1;
	};
}

# helper for check_version.  Reads the given prod/build/vers-*.c file
# to extract the full ontap version string.

sub guess_version ( $ ) {
    my ($filename) = @_;
	return eval {
		my $in = new IO::File($filename,"r");
		if (!$in) {
			die "could not open file \"$filename\" to determine ontap version string: $!";
		}
		my $version;
		while (<$in>) {

			# The first string definition in the vers-*.c file which
			# matches what the full version string, like what the
			# version command will print (but with the date part
			# dropped, because that doesn't seem to always match
			# between the build and the last vers-*.c built)

			# This works for or needs to work for:

			# Anchorsteam or earlier:
			# char version[VERS_VERSION_MAX_LEN] __attribute__((section (".ontap_version"))) ="NetApp Release /x/eng/build4/daemon/anchorsteam/405348: Fri Mar 11 17:23:17 PST 2005";
			# GB or later:
			# struct version_s ... {
			#       "3d",
			#    	"NetApp Release /x/eng/build5/daemon/main/405337: Fri Mar 11 16:29:40 PST 2005",
			
			if($_ =~ /\"(NetApp Release.*?):.*\"/){
				$version=$1;
		       	#Including support for IBM filers .
			#TBD: Enhance to check it with filer model number
			} elsif ($_ =~ /\"(Data ONTAP Release.*?):.*\"/){
				$version=$1;
			}
		}
		if (!$version) {
			die "could not find version in file \"$filename\"";
		}
		return $version;
	};
}

sub make_conn ( $$ ) {
    my ($spec,$default_conn) = @_;
	if (ref($spec) =~ /Connectrec/) {
	    return $spec;
	} elsif (ref($spec) =~ /Hostrec/) {
		return connect($spec->id() . $default_conn);
	} else {
        my $host = host($spec);
        return connect($host->id() . $default_conn);
    }
}

sub make_host ( $ ) {
    my $spec = shift;
    if (ref($spec) =~ /Connectrec/) {
        return $spec->hostp();
    } elsif (ref($spec) =~ /Hostrec/) {
        return $spec;
    } else {
        return host($spec);
    }
}

# Local Variables:
# tab-width: 4
# End:
