
# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/iterator.pm#1 $

## @summary Iterator that basic tests can use for running and reporting simple tests. 
##
## @description  Iterator accepts a CLIENT  and array of tests as an arguments.
##               Test array specifies a series of subtests to run. Each subtest contains
##               a basic command to execute, what result to expect and how to 
##               report success/failure. Subtests are executed serialy on CLIENT and
##               the results are than reported accordingly.
##
## @keywords iterator, NATE, tests, client, remote
##
## @status public
##
## @author vadim@netapp.com
package iterator;

use strict;
use Tharn;

use vars qw(@ISA @EXPORT);
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( iterator);
## @name  ::iterator
## @summary this is an attempt to promote a common
## test data structure, something that will be sufficient
## for most simple NATE tests. This iterator 
## accepts a CLIENT string and an array of tests.
## Then the tests are seraily executed on client(s).
## @usage  &nbsp; iterator($client_string,@subtests);
## where @subtests is array representing tests: 
##   '@subtests' = (                       
##
## 	     {
## 	      name=>'does client have perl in path?',
##
## 	      cmd=>"print \'AAAA\'", #what to do
##
## 	      cmd_type=>"perl",      #how to do it
##
## 	      error=>"could not print AAA", #error msg
## 	      
##            #are we expecting to fail?
##
##            fail=>0, 
## 	      
##            #when to report in respect to test behaivor.
##           'all' | 'expected' | 'unexpected' | 'none'
##            
##            report=>'all', 
##     
##            #how to report when reporting
##            #what arg to pass to tslog if 
##            #subtest behaves as expected
##            '-pass', '-fail', '-comment', '-fatal'
##            
##            expect_action=>'-pass'
##            },
## 	     
##           {
## 	      name=>'lets try ls on /usr',
## 	     
##            cmd=>"ls /usr", #what to do
##
## 	      cmd_type=>"sh",      #how to do it
##
## 	      error=>"could not ls /usr", #error msg
##
## 	      fail=>1, #are we expecting to fail?
##
##            report=>'unexpected',
##            
##            #useless in this case since only unexpected action
##            #will get reported
##
##            expect_action=>'-comment',
##            
##            unexpect_action=>'-fatal',
## 	      }
## 	    );
##
## and $client = "orbit;hosttype=unix:gaff;hosttype=unix-linux";
## 
## @returns 1 if no failures occured, 0 otherwise
## 
## @status public 

sub iterator {
    my ($clients, @subtests) = @_;
    my $final_retval = 1;
    my @clients = split(':', $clients);
    
    foreach my $client (@clients) {
	foreach my $subtest (@subtests) {
	    my $name     = $subtest->{name};
	    my $cmd      = $subtest->{cmd};
	    my $error    = $subtest->{error};
	    my $fail     = $subtest->{fail};
	    my $cmd_type = $subtest->{cmd_type};

	    my $expect_action = defined($subtest->{expect_action}) ?
		$subtest->{expect_action} : '-pass';

	    my $unexpect_action = defined($subtest->{unexpect_action}) ?
		$subtest->{unexpect_action} : '-fail';

	    my $report = defined($subtest->{report}) ? 
		$subtest->{report} : 'all';

	    die "One of the parameteres was not provided"
		if (!(defined($name) && defined($cmd) && defined($error) 
		      && defined($fail) && defined($cmd_type)));

	    my $retval;
	    if ($cmd_type eq  'perl') {
		$retval = subtest("perl", -client, $client, -search, 
				  [], "--", "-e", "$cmd");
	    }elsif ($cmd_type eq 'sh') {
		$retval = subtest("sh", -client, $client, -search, 
				  [], "--", "-c", "$cmd");
	    } #XXX..... more stuff will go here later
	    else {
		tstlog(-fatal, "iterator.pm:unknown command type $cmd_type\n");
	    }
	    iterator_report($fail, $retval->exit_status, 
			    $name, $client, 
			    $report, $expect_action, 
			    $unexpect_action, $error);
	}
    }
}


#--------------------------------------------------
##@name    ::iterator_report
##@summary This is an internal helper function that
##          is only used by iterator for examining
##          results and for logging.
##
##@synopsis Examines results of subtest, decides if 
##          logging needs to be done, and if it does, 
##          generates appropriate tstlog message.
##          Returns 1 if subtest acted as expected,
##          0 otherwise.
##@usage   iterator_report($expected_failure, $exit_status, 
##	                   $name, $client, $report,$expect_action, 
##                         $unexpect_action, $error);
##
##@returns 1 if subtest acted as expected,
##          0 otherwise.
##
##@status private
#----------------------------------------------------	
sub iterator_report {
    my ($expected_failure, $exit_status, 
	$name, $client, $report, 
	$expect_action, $unexpect_action, $error) = @_;
    
    my $tstlog_arg;
    my $expected;
    my $retval = 0;
    
    if ($expected_failure) {
	if ($exit_status) {
	    $retval =1;
	    #expected failure and got it
	    #if we are reporting expected, report it
	    if ($report eq 'all' || $report eq 'expected') {
		#expected behaivor is success
		tstlog("$expect_action", "$name succeeded on $client");
	    }
	} else {
	    #did not get an error when expected one
	    if ($report eq 'all' || $report eq 'unexpected') {
		tstlog("$unexpect_action", "$name $error on client $client");
	    }
	}
    } else {
	#expected success
	if ($exit_status) {
	    #failed while expecting success
	    if ($report eq 'all' || $report eq 'unexpected') {
		tstlog("$unexpect_action", "$name $error (status:$exit_status) on client $client");
	    }
	} else {
	    $retval = 1;
	    #succeeded while expecting success
	    if ($report eq 'all' || $report eq 'expected') {
		tstlog("$expect_action", "$name succeeded on $client");
	    }
	}	
    }
    return $retval;
}

1;
