# $Id: //depot/prod/test/main/lib/log_zapi.pm#1 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#

## @summary This library is used to log the zapi details: 
##          zapi-name, input and output into NateDB.
## @author swati.bajaj@netapp.com, dl-nacl-dev
## @status shared
## @pod here

package log_zapi;
use NaElement;
use NATE::BaseException qw(:try);
use Tharn qw($Log);
use Params::Validate qw(:all);
use Data::Dumper;
&Tharn::param("LOG_ZAPI",  "-default", 1);


###############################################################################
# Method :    write_in_db
# Objective:  Log the ZAPI info in NATE DB
# Details:    Whenever ZAPIs are invoked by scripts/libs, they will get logged
#             into NATE DB and this will happen transparently. QC would read
#             the information from NATE DB and will store these for future
#             references and for metrics purposes.
###############################################################################

sub write_in_db {

    $Log->enter();
    my %args = @_;
    my $input = $args{input};
    my $output = $args{output};
    if ($Tharn::Params{"LOG_ZAPI"}->{value}) {
        my $is_nate = _is_nate_running();
        if ( $is_nate ) {
            # Get the ZAPI name from the input XML by reading the root node
            my $zapi;
            my $xml = $input;
            ($zapi) = ($xml =~ /<(.+?)>/);
            if (defined $zapi) {
                $zapi =~ s:\s*/\s*::;
            } else {
                # If we couldn't obtain the ZAPI name, then we cannot log any entry
                # into the DB as 'name' is a mandatory param for register_zapi.
                # So warn users and exit the method.
                $Log->warn("Couldn't obtain the API name from the input XML to log ".
                    " in NATE DB. The input XML is\n". $xml
                );
                $Log->exit();
                return;
            }
            unless (Tharn->can('register_zapi')) {
                $Log->debug("An older version of NATE prior to 4.2.54 is used");
                $Log->exit();
                return;
            }
            
            try {
                # Though register_zapi asynchronously logs the info in DB, it's
                # still safe to have try-catch block to handle any of the errors
                # thrown from the processing code present before invoking the
                # async call.
                Tharn::register_zapi(
                    name => $zapi,
                    input => $xml,
                    other => $output
                );
            } catch NATE::BaseException with {
                my $ex = shift;
                $Log->warn("Failed to log zapi input and output ".
                    "in NATE DB. The reason is\n". $ex->text);
            };
        }
    
    }
    
    $Log->exit();
                      
}   

# Helper routine to check whether the scripts are executed using NATE
# For PNATE customers we will not enable the logging
sub _is_nate_running {

    my $is_nate_running = 0;
    my $is_from_ntest = 0;
    no strict 'subs';
    # Check if running using ntest
    eval {
        require NATE;
    };
    if (!$@) {
    
        my $nate_ver = &NATE::MAJOR_VERSION.'.'. &NATE::MINOR_VERSION;
        if($nate_ver > 4.3) # 4.4.0+
        {
            require NATE::Database::Test;
            $is_from_ntest = int(defined(NATE::Database::Test::testdb_global));
        }
        else
        {
            require NATE::TestDB;
            $is_from_ntest = int(defined(NATE::TestDB::testdb_global));
        }
    
        # check if NATE or PNATE
        if ($is_from_ntest) {
            $is_nate_running = 1 unless (_is_pnate());
        }
    } 
    return $is_nate_running;
}   

# Helper routine to check if running PNATE
sub _is_pnate {

    require NATE::ParamSet;
    my $var = NATE::ParamSet->new(global=>1);
    my $is_pnate = int($var->get('PNATE') || $ENV{PNATE} || 0);
    return $is_pnate;
}
    
	
#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

log_zapi.pm

=head1 Author

dl-nacl-dev

=head1 Description

This library is used to log the zapi details: zapi-name, input and output 
into NateDB.

=head1 B<Public Methods>

=head2 B<write_in_db()>

Logs the input xml and the output xml in NateDB. If the command line param
LOG_ZAPI is set to 0, then the zapi data will not be logged. By default 
LOG_ZAPI is set to 1.

    write_in_db( input => $xmlrequest, output => $obj->sprintf());

=over

=item Inputs

=over

=item input

The input xml request that has to be logged.

=back

=over

=item output

The output xml, or any other message that needs to be logged
corresponding to the input xml request.

=back

=back

=cut
	
1;  
