# ----------------------------------------------------------------------
# $Id: //depot/prod/test/Rsierranevada/lib/misc.pm#1 $
# ----------------------------------------------------------------------

## @summary Miscellaneous library -- a diverse collection of subroutines
##
## @description
##   This package contains usable miscellaneous routines that will
##   organically grow into a diverse and multi-function library.
##
##   In other words, please add your subroutines here.
##
## @author Rick Ehrhart
## @author Tim Thompson
##
## @status public

package misc;

use Tharn;
use ZapiTest;
use Socket;
use NaElement;
use strict;
use utf8;
use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
		get_ips_and_hosts
		get_net_interfaces
		cli
		cli_admin
		cli_advanced
		cli_partner
		zapi
		zapi_check_expected
		cli_localhost
		subtest_continuous
		z_cli
		z_cli_advanced
		z_verify_child_contains
		z_verify_child_contains_typecheck
		z_verify_child_contains_values
		z_registry_get
		z_registry_delete
		z_registry_set
		);

# ----------------------------------------------------------------------
## @name ::get_ips_and_hosts
## @synopsis
## ($ips, $hosts) = get_ips_and_hosts()
##
## @description
##    Obtains all the IP addresses and host names of the current host.
##    Returns a comma-separated string of IP addressess and a comma-
##    separated string of host names.
##
## @params
##    none
# ----------------------------------------------------------------------

sub get_ips_and_hosts( ) {
    my $ip = "";
    my $ips = "";
    my $packed_ip = 0;
    my $name = "";
    my $names = "";
    my $sep = "";

    my $ifc = ifconfigpath();
    my $ifconfig_output = qx( $ifc -a );

    my $i = 0;
    foreach my $line (split /\n/, $ifconfig_output) {
        #
        # matches either the Linux or Sun outputs

        if ($line =~ m/inet (addr:)?(\d+.\d+.\d+.\d+)/) {
            # print "$i: $line\n";

            $ip = $2;
            # print "$i: ip: $ip\n" ;

            # if the IP address is local host (127), ignore it.

            if ( "$ip" !~ m/^127\./ ) {
                $packed_ip = inet_aton( $ip );
                $name = gethostbyaddr( $packed_ip, AF_INET );
		if ( $name ) {
			$ips = $ips .  $sep . $ip;
			$names = $names . $sep . $name;
			$sep = ",";
		} else {
			# Unable to gethostbyaddr for $packed_ip
		}
            }

            $i += 1;
        }
    }
    return $ips, $names;
    
}


# ----------------------------------------------------------------------
## @name ::get_net_interfaces
## @synopsis
## $ifs = get_net_interfaces($ifconfig_output)
##
## @description
##    Obtains all the network interface names from a given ifconfig output.
##    Returns a comman separated string of network interface names.
##
## @params
##    ifconfig_output Required: The output from an ifconfig command.
##
## @example
##    $ifconfig_output = qx( ifconfig -a );
##    my $ifs = get_net_interface($ifconfig_output);
# ----------------------------------------------------------------------

sub get_net_interfaces( $ ) {
    my $if = "";
    my $ifs = "";

    my $ifconfig_output = $_[0];

    foreach my $line (split /\n/, $ifconfig_output) {
        if ( $line =~ m/^(\w+:) flags/ ) {
            $if = $1;
            #print "if: $if\n";
            if ( $if !~ m/lo/ ) {
                chop( $if );   # chop off the colon
                $ifs = $ifs . $if . ",";
            }
        }
    }
    
    chop( $ifs );  # chop off the comma
    return( $ifs );
    }

# ----------------------------------------------------------------------
## @name ::cli
## @synopsis
## $output = cli($FILER,"cmd");
##
## @description
##    Executes a CLI command and returns the output.
# ----------------------------------------------------------------------

$::rsh_count = 0;

sub cli ($$;$$) {

	my ($filer,$call,$tmout,$password) = @_;

	if (!defined($tmout)) {
		$tmout = 300;
	}
	my $host;
	my $conn;

	$host = host($filer);
	if (!defined($password)) {
		$password = $host->{default_password};
	}

	# If we try to use different passwords in different calls to cli(),
	# it won't work unless we use a unique #rsh value on each.
	# So, we just artificially construct one.
	$::rsh_count++;
	$conn = connect("$filer#rsh".$::rsh_count,
	    , "conntype=rsh-persay"
	    , "username=root"
	    , "password=$password"
	    , "connport="
	    , "connhost=$host->{default_ip}"
	    , "-fatal", "misc::cli(): unable to connect to filer $filer");

	my ($mid,$after);
        (undef,$mid,$after) = say ($conn,"$call",'-timeout',$tmout);
	return $after;
}

sub make_connection($) {
	my ($spec) = @_;	
    return eval {
	if (ref($spec) =~ /Connectrec/) {
		return $spec;
	}elsif(ref($spec) =~ /Hostrec/) {
		return Tharn::connect($spec->{id} . "#rsh");
	}else{
		return Tharn::connect($spec);
	}
    };
}


# ----------------------------------------------------------------------
## @name ::cli_partner
## @synopsis
## $output = cli_partner($FILER,"cmd")
##
## @description
##    Executes a CLI command on a partner (during takeover) and returns
##    the output.
# ----------------------------------------------------------------------
sub cli_partner {
	my ($filer, $cmd) = @_;
	return(cli($filer,"partner $cmd"));
}

# ----------------------------------------------------------------------
## @name ::cli_advanced
## @synopsis
## $output = cli_advanced($FILER,"cmd")
##
## @description
##    Executes a "priv advanced" CLI command and returns the output.
# ----------------------------------------------------------------------
sub cli_advanced {
	my ($filer, $cmd) = @_;
	return(cli($filer,"priv -level secret; $cmd"));
}

# ----------------------------------------------------------------------
## @name ::cli_admin
## @synopsis
## $output = cli_admin($FILER,"cmd")
##
## @description
##    Executes a "priv admin" CLI command and returns the output.
##    It *guarantees* that it's executed in 'admin' priv level, even if
##    the console is currently in the advanced level.
# ----------------------------------------------------------------------
sub cli_admin {
	my ($filer, $cmd) = @_;
	return(cli($filer,"priv set admin; $cmd"));
}

# ----------------------------------------------------------------------
## @name ::cli_advanced_partner
## @synopsis
## $output = cli_advanced_partner($FILER,"cmd")
##
## @description
##    Executes an advanced CLI command on a partner (during takeover)
##    and returns the output.
# ----------------------------------------------------------------------
sub cli_advanced_partner {
	my ($filer, $cmd) = @_;

	# XXX - I can't figure out how to do a "priv -level secret" AND
	# a subsequent command on the partner.  Doing the obvious
	# 	partner priv -level secret; partner $cmd
	# doesn't seem to work - the code in the partner command
	# pushes and pops the privilege level each time it's executed.
	#
	# The method used here is a kludge - it assumes that the
	# partner's console starts out in "normal" command mode,
	# and will not work properly when multiple invocations of
	# this are are done simultaneously.

	return(cli($filer,"partner priv -level secret; partner $cmd; partner priv -level secret"));
}

# ----------------------------------------------------------------------
## @name ::check_java_enabled
## @synopsis
## if ( check_java_enabled($FILER) ) {
##       tstlog ("-fail", "Java is not enabled!  This test needs Java.");
##       return 1;
## }
## @description
##    Returns non-zero if java is not enabled on $FILER.
# ----------------------------------------------------------------------
sub check_java_enabled {
	my ($filer) = @_;
	my $o = cli($filer,"java");
	if ( "$o" =~ /Usage:/ ) {
		return 0;
	} else {
		# It will say "Java not enabled"
		return 1;
	}
}

# ----------------------------------------------------------------------
## @name ::check_dns_enabled
## @synopsis
## if ( check_dns_enabled($FILER) ) {
##       tstlog ("-fail", "DNS is not enabled!  This test needs DNS.");
##       return 1;
## }
## @description
##    Returns non-zero if DNS is not enabled on $FILER.
# ----------------------------------------------------------------------
sub check_dns_enabled {
	my ($filer) = @_;
	my $o = cli($filer,"dns info");
	if ( "$o" =~ /enabled/ ) {
		return 0;
	} else {
		# It will say "DNS is disabled"
		return 1;
	}
}

# The following two routines should share the same code.
sub nslookuppath {
	my $f = "/usr/sbin/nslookup";
	if ( -f $f ) {
		return $f;
	}
	$f = "/usr/bin/nslookup";
	if ( -f $f ) {
		return $f;
	}
	$f = "/sbin/nslookup";
	if ( -f $f ) {
		return $f;
	}
	$f = "/bin/nslookup";
	if ( -f $f ) {
		return $f;
	}
	tstlog ("-fail", "nslookup not found!  This test needs nslookup.");
	return $f;
}

# This routine should share code with the routine above.
sub ifconfigpath {
	my $f = "/usr/sbin/ifconfig";
	if ( -f $f ) {
		return $f;
	}
	$f = "/usr/bin/ifconfig";
	if ( -f $f ) {
		return $f;
	}
	$f = "/sbin/ifconfig";
	if ( -f $f ) {
		return $f;
	}
	$f = "/bin/ifconfig";
	if ( -f $f ) {
		return $f;
	}
	tstlog ("-fail", "ifconfig not found!  This test needs ifconfig.");
	return $f;
}

# ----------------------------------------------------------------------
# returns the IP address of a host name
# ----------------------------------------------------------------------

sub nslookup( $ ) {
    my $ns = nslookuppath();
    my $nslookup_output = qx($ns $_[0]);

    my $addr_output;
    foreach my $line (split /\n/, $nslookup_output) {
      if ($line =~ /Address:/) {
         $addr_output = $line;
      }
    }

    my($addr_label, $addr) = split( /:/, $addr_output );

    chomp($addr);
    $addr =~ s/\s+//;
    return $addr;
}

# ----------------------------------------------------------------------
## @name ::zapi
## @synopsis
## ($err,$reason) = zapi($FILER,"sytem-get-version",...)
##
## @description
## 	A simple routine to executes a zephyr-style API which has inputs only,
##	Returns two values - the errno, and the reason string.
##	If the errno is 0, the reason string should not be used.
##
## @params
##    filer Required: filer to which to send the API.
##    api Required: api name
##    key1 Required: first (optional) parameter keyword
##    val1 Required: first (optional) parameter value
##
# ----------------------------------------------------------------------

sub zapi ($$;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$) {

	my $filer = shift;
	my $api = shift;

	my $host = host($filer);

	my $ip = $host->{default_ip};
	my $password = $host->{default_password};
	my $comment;


	my $server = new ZapiTest($ip, "root", $password);
	if ( ! defined $server ) {
		tstlog("-fatal","misc::zapi(): unable to connect to filer $filer");
		return undef;
	}

	$comment = "Zapi: host :" . $host->hostname() . " ZAPI: $api";
	
	my $e = new NaElement($api);
	my ($key, $value);
	while ( 1 ) {
		$key = shift;
		$value = shift;
		if ( ! defined $key || ! defined $value ) {
			last;
		}
		$comment = $comment . " $key $value";
		$e->child_add(new NaElement($key,$value));
	}

	logcomment($comment);

	my $o = $server->invoke_elem($e,"any");
	my $errno = $o->results_errno();
	my $reason = $o->results_reason();
	my $out = $o->toString();
	return ($errno,$reason,$out);
}

# ----------------------------------------------------------------------
## @name ::zapi_check_expected
## @synopsis
## ($ok) = zapi_check_expected($i,$o,$expected)
##
## @description
## 	A simple routine to check the results of an api call for
##	an expected result - failed or passed.
##	The return value is 0 if the expected result was NOT found.
##
## @params
##    i Required: input NaElement of API call
##    o Required: output NaElement of API call
##    expected Required: either "failed" or "passed"
##
# ----------------------------------------------------------------------

sub zapi_check_expected ($$$;$) {

	my $i = shift;
	my $o = shift;
	my $expected = shift;
	my $expectederrno = shift;

	my $st = $o->attr_get("status");
	my $errno = $o->results_errno();
	my $reason;
	if ( $st eq "failed" ) {
		$reason = $o->results_reason();
	} else {
		$reason = "It worked";
	}
	if ( $st ne $expected ) {
		tstlog("-fail","ZAPI - "
			."Unexpected result - expected $expected, "
			."got $st.  Reason=[[$reason]]  "
			."Input=[[".$i->toString."]] "
			."Output=[[".$o->toString."]]");
		return(0);
	} elsif ( $expected eq "failed" && defined($expectederrno) && $expectederrno != $errno ) {
		tstlog("-fail","ZAPI - "
			."Unexpected result - expected errno=$expectederrno, "
			."got $errno.  Reason=[[$reason]]  "
			."Input=[[".$i->toString."]] "
			."Output=[[".$o->toString."]]");
		return(0);
	} else {
		return(1);
	}
}

# ----------------------------------------------------------------------
## @name ::cli_localhost
## @synopsis
## $output = cli_localhost("cmd");
##
## @description
##    Executes a CLI command on the local host and returns the output.
##
## @params
##    cmd Required: The shell command-line to execute.
##
## @example
##    $out = cli_localhost("ls -l $mountpoint");
# ----------------------------------------------------------------------

sub cli_localhost ($) {

	my ($cmd) = @_;

	if ( !open(CMD,"$cmd |") ) {
		tstlog("-fail", "Unable to open pipe to cmd=$cmd");
		return "";
	}
	my $out = "";
	while ( <CMD> ) {
		$out .= $_;
	}
	close(CMD);
	return $out;
}

# ----------------------------------------------------------------------
## @name ::z_do_cli
## @synopsis
## ($output, $value) = z_do_cli($server, $priv, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    $priv: Required: The privileged of the command.
##    cmd: Required: The command with no paramenters.
##    parm[1-n]: Optional: any command parameter.
##
# ----------------------------------------------------------------------

sub z_do_cli( $$@ )
{
	my $xi;
	my $xo;
	my $args;
	my $i;
	my $arg;

	my $server = shift;
	my $priv   = shift;
	my $num_parms = @_;

	$xi = new NaElement("system-cli");
	$xi->child_add(new NaElement("priv", $priv));
	$args = new NaElement("args");

	# load up the command and the arguments

	for ($i = 0; $i < $num_parms; $i += 1 ) {
		$arg = shift;
		$args->child_add(new NaElement("arg",$arg));
	}

	# add the command and arguments to the api

	$xi->child_add($args);

	# execute the command

	my $oe = $server->invoke_elem($xi, "any");

	if ( ! $oe->results_ok() ) {
		tstlog("-fail", "server->invoke failed.");
		return undef, undef;
	}

	my $output = $oe->child_get_string("cli-output");
	my $value = $oe->child_get_string("cli-result-value");

	return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::z_cli
## @synopsis
## ($output, $value) = z_cli($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the admin priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no paramenters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli($server, "qtree", "status");
##
# ----------------------------------------------------------------------
sub z_cli( $@ )
{
	my $output;
	my $value;

	my $server = shift;

	($output, $value) = z_do_cli($server, "admin", @_);
	return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::z_cli_advanced
## @synopsis
## ($output, $value) = z_cli($server, "cmd", "parm1", ...)
##
## @description
##    Executes a CLI command via system-cli Zephyr API at the advance
##    priv level.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    cmd: Required: The command with no paramenters.
##    parm[1-n]: Optional: any command parameter.
##
## @example
##    ($output, $value) = z_cli_advanced($server, "license", "UYTCNCG");
##
# ----------------------------------------------------------------------

sub z_cli_advanced( $@ )
{
	my $output;
	my $value;

	my $server = shift;

	($output, $value) = z_do_cli($server, "advanced", @_);
	return $output, $value;
}

# ----------------------------------------------------------------------
## @name ::z_registry_get
## @synopsis
## $value = z_registry_get($server, $key)
##
## @description
##    Get a registry value.  If the value doesn't exist, undef is returned.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    key: Required: The key name.
##
## @example
##    $value = z_registry_get($server, "state.api.validate.enable");
##
# ----------------------------------------------------------------------

sub z_registry_get( $@ )
{
	my $output;
	my $value;

	my $server = shift;
	my $key = shift;

	($output, $value) = z_cli_advanced($server, "registry", "get", "$key");
	if ( $output =~ "does not exist" ) {
		return undef;
	}
	$output =~ s/.*=//;
	$output =~ s/[\r\n]*$//;
	return $output;
}

# ----------------------------------------------------------------------
## @name ::z_registry_delete
## @synopsis
## $value = z_registry_delete($server, $key)
##
## @description
##    Delete a registry value.  If an error occurs, a string describing
##    the error is returned, otherwise "" is returned.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    key: Required: The key name.
##
## @example
##    z_registry_delete($server, "state.api.validate.enable");
##
# ----------------------------------------------------------------------

sub z_registry_delete( $@ )
{
	my $server = shift;
	my $key = shift;
	my $output;
	my $value;

	($output, $value) = z_cli_advanced($server, "registry", "delete", "$key");
	if ( $output =~ "error occurred" ) {
		return $output;
	} else {
		return "";
	}
	return;
}

# ----------------------------------------------------------------------
## @name ::z_registry_set
## @synopsis
## $value = z_registry_set($server, $key, $val)
##
## @description
##    Set a registry value.  If an error occurs, a string describing
##    the error is returned, otherwise "" is returned.
##
## @params
##    $server:  Required: A zapi server object obtained via new Zapi( ).
##    key: Required: The key name.
##    val: Required: The value.
##
## @example
##    z_registry_set($server, "state.api.validate.enable", "true");
##
# ----------------------------------------------------------------------

sub z_registry_set( $@ )
{
	my $server = shift;
	my $key = shift;
	my $val = shift;
	my $output;
	my $value;

	($output, $value) = z_cli_advanced($server, "registry", "set", "$key", "$val");
	if ( $output =~ "error occurred" ) {
		return $output;
	} else {
		return "";
	}
}

# ----------------------------------------------------------------------
## @name ::subtest_continuous
## @synopsis
## 	subtest_continuous("admin/zapi_basic");
##
## @description
## 	Run a subtest continously.
##
## @params
##    cmd Required: The name of the test to run.
##
## @example
## 	subtest_continuous("admin/zapi_basic");
# ----------------------------------------------------------------------

sub subtest_continuous ($) {
	my ($testname) = shift;
	my ($runtimes) = shift;
	my ($runlength) = shift;

	if ( ! $runtimes ) {
		$runtimes = 9999999;
	}
	if ( ! $runlength ) {
		$runlength = 0;
	}
	my $loopcount = 0;
	my $begtime = time;
	while ( ($runtimes != 0 && $loopcount < $runtimes)
		|| ($runlength != 0 && ((time-$begtime) < $runlength)) ) {

		$loopcount++;

		tstlog -comment, "subtest_continous, running $testname, loopcount=$loopcount";

		# Run the test

		# Replace / with _ in the test name, to synthesize a runid id.
		my $id = $testname;
		$id =~ s/\//_/g;

		subtest $testname.".thpl", -runid, "$id", -ifscript, -ok, -iffatal, -ok;
	}
}


# ----------------------------------------------------------------------
## @name ::z_verify_elem_exists
## @synopsis
## 	z_verify_elem_exists($elem, $child, $sub_child, ...);
##
## @description
## 	verifies whether or not the sub (child) elements exists, under
##      the given element.  In the example, it'll return 1 if "name" 
##      is a child of "disk-info", who is a child of "disks", who is
##      a child of $elem, otherwise it'll return 0.
##
## @params
##    cmd Required: The top element $elem, and a abitrary list of sub
##                  elements.
##
## @example
## 	z_verify_elem_exists($elem, "disks", "disk-info", "name");
# ----------------------------------------------------------------------

sub z_verify_elem_exists {
    my ($elem) = @_;
    my $size = @_;
    my @api_args = @_[1..($size-1)];
    $size = @api_args;

    my $elems_all = $elem->name();
    my $tmp;
    my $prev = $elem;
    for (my $i = 0; $i < $size; $i++) {
	$elems_all .= "->" . $api_args[$i];
	$tmp = $prev->child_get($api_args[$i]);
	if (!defined($tmp)) {
	    tstlog("-fail", "verification failed, ".$api_args[$i]." does NOT exist in $elems_all.");
	    return 0;
	}
	$prev = $tmp;
    }
    
    tstlog("-comment", "verification success : $elems_all exists.");
    return 1;
}



# ----------------------------------------------------------------------
## @name ::z_verify_elem_not_exists
## @synopsis
## 	z_verify_elem_not_exists($elem, $child, $sub_child, ...);
##
## @description
## 	verifies whether or not the sub (child) elements exists, under
##      the given element.  In the example, it'll return 0 if "name" 
##      is a child of "disk-info", who is a child of "disks", who is
##      a child of $elem, otherwise it'll return 1.
##
## @params
##    cmd Required: The top element $elem, and a abitrary list of sub
##                  elements.
##
## @example
## 	z_verify_elem_not_exists($elem, "disks", "disk-info", "name");
# ----------------------------------------------------------------------
sub z_verify_elem_not_exists {
    my ($elem) = @_;
    my $size = @_;
    my @api_args = @_[1..($size-1)];
    $size = @api_args;

    my $elems_all = $elem->name();
    my $tmp;
    my $prev = $elem;
    for (my $i = 0; $i < $size; $i++) {
	$elems_all .= "->" . $api_args[$i];
	$tmp = $prev->child_get($api_args[$i]);
	if (!defined($tmp)) {
	   tstlog("-comment", "verification success : ".$api_args[$i]." does not exist in $elems_all as expected.");
	   return 1;
	}
	$prev = $tmp;
    }
    
    my $name = $elem->name();
    tstlog("-fail", "verification failed, element $elems_all exists while not expected.");
    return 0;
}


# ----------------------------------------------------------------------
## @name ::z_verify_elem
## @synopsis
## 	z_verify_elem($elem, $api, $content);
##
## @description
## 	verify the content of the given element for expected content.
##      In the case above, it'll verify that the content of 
##      element $elem is $content, and log the corresponding msg.
##
## @params
##    cmd Required: an element $elem, its expected content $content,
##                  and the API $api used for logging.
##
## @example
## 	z_verify_elem($elem, "fcp-adapter-list-info", "online");
# ----------------------------------------------------------------------
sub z_verify_elem {
    my ($elem, $api, $content) = @_;

    if (!defined($elem)) {
	tstlog("-fail", "verification failed : $api, expected element ($elem) does not exist.");
	return 0;
    } elsif ($elem->content() ne $content) {
	my $badcontent = $elem->content();
	tstlog("-fail", "verification failed : $api, expected $content, got $badcontent");
	return 0;
    }

    tstlog("-comment", "verification successful : $api verified");
    return 1;
}


# ----------------------------------------------------------------------
## @name ::z_verify_elem_multi
## @synopsis
## 	z_verify_elem_multi($elem, $api, $expected_1, $expected_2, ...);
##
## @description
## 	verify the content of the given element for expected contents.
##      In the example, it'll verify that the content of 
##      element $elem is either "online" or "running", and log 
##      the corresponding msg.
##
## @params
##    cmd Required: an element $elem, the API $api used for logging.
##                  and list of possible expected contents.
##
## @example
## 	z_verify_elem_multi($elem, $api, "online", "running");
# ----------------------------------------------------------------------
sub z_verify_elem_multi {
    my ($elem, $api) = @_;

    my $size = @_;
    my @contents = @_[2..($size-1)];
    $size = @contents;

    if (!defined($elem)) {
	tstlog("-fail", "verification failed : $api, expected element ($elem) does not exist");
	return 0;
    } else {
       my $found = 0;
       for (my $i = 0; $i < $size; $i = $i+1) {
	   if ($elem->content() eq $contents[$i]) {
	      $found = 1;
	      last;
	   }
       }
       if ($found == 1) {
	      tstlog("-comment", "verification success : $api verified");
	      return 1;
       } else {
	      my $badcontent = $elem->content();
	      tstlog("-fail", "verification failed : $api, expected @contents, got $badcontent");
	      return 0;
       }
    }
}



# ----------------------------------------------------------------------
## @name ::do1api
## @synopsis
## 	z_do1api($server, $api, $elem, $expected);
##
## @description
##      invokes an api on host $server, with $elem and log the corresponding
##      msg.  $api is used for logging and $expected is either "any", "passed",
##      or "failed".
##
## @params
##      cmd Required: the server $server to invoke from, the name of the API $api
##      to invoke (used only for logging), the xml element (request) to send to 
##      server, and an optional $expected result, if not specified, "passed"
##      is the default.
##
## @example
## 	z_do1api($server, $api, $elem, "passed");
# ----------------------------------------------------------------------
sub z_do1api {
	my ($server,$api,$e,$expected) = @_;

	if (!defined($server)) {
		tstlog("-fail", "can not invoke anyting");
	}
	if (!defined($expected) ) {
		$expected = "passed";
	}

	my $a = $api;
	my $o = $server->invoke_elem($e, "any");
	my $r = $o->results_reason();

	if ( $expected eq "passed" && ! $r ) {
		tstlog("-comment", "$a passed as expected");
	} elsif ( $expected eq "failed" && $r ) {
		tstlog("-comment", "$a failed as expected, reason=$r");
	} elsif ( $expected eq "passed" && $r ) {
		tstlog("-fail", "$a did not pass as expected, it failed with reason=$r");
	} elsif ( $expected eq "failed" && ! $r ) {
		tstlog("-fail", "$a did not fail as expected, it passed?");
	} elsif ( $expected eq "any" ) {
		tstlog("-comment", "$a completed");
	} else {
		tstlog("-fail", "Unexpected internal error in test: expected=$expected r=$r");
	}
	return $o;
}



# ----------------------------------------------------------------------
## @name ::z_get_elem
## @synopsis
## 	z_get_elem($elem, $child, $sub_child, ...);
##
## @description
## 	get the sub (child) element from the given element.  In 
##      the example, it'll return element "name" under
##      "disk-info", who is a child of "disks", who is
##      a child of $elem.  If that element does not exist
##      it'll return undef.
##
## @params
##    cmd Required: The top element $elem, and a abitrary list of sub
##                  elements.
##
## @example
## 	z_get_elem($elem, "disks", "disk-info", "name");
# ----------------------------------------------------------------------

sub z_get_elem {
    my ($elem) = @_;
    my $size = @_;
    my @api_args = @_[1..($size-1)];
    $size = @api_args;

    my $tmp;
    my $prev = $elem;
    for (my $i = 0; $i < $size; $i++) {
	if (!defined($prev)) {
	   return undef;
	}
	$tmp = $prev->child_get($api_args[$i]);
	$prev = $tmp;
    }
    return $prev;
}


# ----------------------------------------------------------------------
## @name ::z_get_elem_content
## @synopsis
## 	z_get_elem_content($elem, $child, $sub_child, ...);
##
## @description
## 	get the content of sub (child) element from the given
##      element.  In the example, it'll return the content of 
##      element "name" under "disk-info", who is a child of "disks",
##      who is a child of $elem.  If that element does not exist
##      it'll return undef.
##
## @params
##    cmd Required: The top element $elem, and a abitrary list of sub
##                  elements.
##
## @example
## 	z_get_elem($elem, "disks", "disk-info", "name");
# ----------------------------------------------------------------------
sub z_get_elem_content {
    my $elem_tmp = z_get_elem(@_);
    if (defined($elem_tmp)) {
       return $elem_tmp->content();
    }
    return undef;
}

# ----------------------------------------------------------------------
## @name ::z_verify_child_contains
## @synopsis
## 	z_verify_child_contains($elem, $childname1, $childname2, ...);
##
## @description
## 	Verify that an element contains children with the
##	specified names.
##
## @params
##    elem Required: The top element $elem, and a abitrary list of sub
##                  elements.
##
## @example
## 	z_verify_child_contains($elem, "output1", "output2");
# ----------------------------------------------------------------------
sub z_verify_child_contains {
	my $o = shift;
	my $element = shift;
	my $name;
	while ( defined ($name=shift) ) {
		if ( ! defined($o->child_get_string($name)) ) {
			tstlog("-fail", "$element did not contain"
				." $name value");
		} 
	}
}

# ----------------------------------------------------------------------
## @name ::z_verify_child_contains_typecheck
## @synopsis
## 	z_verify_child_contains_typecheck($elem, $child1, $type1, ...);
##
## @description
## 	Verify that an element contains child values appropriate to
##	the types of those values.  Currently, the only type supported
##	is "boolean".
##
## @params
##    elem Required: The top element $elem, and a abitrary list of sub
##                  elements and types.
##
## @example
## 	z_verify_child_contains_typecheck($elem, "output1", "boolean");
# ----------------------------------------------------------------------
sub z_verify_child_contains_typecheck {
	my $o = shift;
	my $element = shift;
	my $name;
	my $type;
	while ( defined ($name=shift) && defined ($type=shift) ) {
		if ( ! defined($o->child_get_string($name)) ) {
			tstlog("-fail", "$element did not contain"
				." $name value");
			next;
		}
		my $v = $o->child_get_string($name);
		my $f = 0;
		if ( $type eq "boolean" ) {
			if ( $v ne "true" && $v ne "false" ) {
				$f = 1;
			}
		} else {
			tstlog("-fail", "Unknown type given to verify_child_contains_typecheck");
			next;
		}
		if ( $f != 0 ) {
			tstlog("-fail", "The value ($v) of the $name element "
				."in $element is not a proper $type value.");
		}
	}
}

# ----------------------------------------------------------------------
## @name ::z_verify_child_contains_values
## @synopsis
## 	z_verify_child_contains_values($elem, $childname, $value1, $value2, ...);
##
## @description
## 	Verify that a particular child element has a value that matches
##	one of the possible values for that type.
##
## @params
##    elem Required: The top element $elem, child name $childname
##		and a abitrary list of values.
##
## @example
## 	z_verify_child_contains_values($elem, "output1", "running", "idle", "off");
# ----------------------------------------------------------------------
sub z_verify_child_contains_values {
	my $o = shift;
	my $element = shift;
	my $name = shift;
	if ( ! defined($o->child_get_string($name)) ) {
		tstlog("-fail", "$element did not contain"
			." $name value");
		next;
	}
	my $val = $o->child_get_string($name);
	my $v;
	my $found = 0;
	while ( ($v=shift) ) {
		if ( "$v" eq "$val" ) {
			$found = 1;
		}
	}
	if ( ! $found ) {
		tstlog("-fail", "The value ($val) of the $name element "
			."did not have one of the expected values.");
	}
}

# ----------------------------------------------------------------------
## @name ::nice_utf8_string
## @synopsis
## 	nice_utf8_string($string)
##
## @description
##    Converts an UTF8 string into hex format, so you can print it out.
##
## @params
##    String you want to view
##
## @example
## 	nice_utf8_string("some weird utf8 string);
# ----------------------------------------------------------------------
sub nice_utf8_string {
    join("",
         map { $_ > 255 ?                  # if wide character...
               sprintf("\\x{%04X}", $_) :  # \x{...}
               chr($_) =~ /[[:cntrl:]]/ ?  # else if control character ...
               sprintf("\\x%02X", $_) :    # \x..
               quotemeta(chr($_))          # else quoted or as themselves
         } unpack("U*", $_[0]));           # unpack Unicode characters
   }


# ----------------------------------------------------------------------
## @name ::random_utf8_string
## @synopsis
## 	random_utf8_string($length);
##
## @description
##    Creates a string filled with random utf8 characters from 0xC280 to 0xD280
##    of the given length.
##
## @params
##    Length of the string you want to create.
##
## @example
## 	random_utf8_string(100);
# ----------------------------------------------------------------------
sub random_utf8_string {
    my $length = shift;
    my $i = 0;
    my $ch;
    my $ret = "";
    while ($i < $length) {
#	$ch = rand(0xDFBF-0xC280)+0xC280; # this is the bounds of normal UTF8
	$ch = rand(0x1000)+0xC280;
	if ($i + 1 >= $length) {
	    $ret = "U".$ret;
	    return $ret;
	}
	$i ++;
	if ($ch > 0xFF) {
	    $i++;
	}
	$ret = $ret.chr($ch)
    }
    return $ret;
}

1;  # return a true value
