#  $Id: //depot/prod/test/Rsierranevada/lib/mount.pm#1 $

## @summary Invoke mount(8) and umount(8) idempotently
## @author lincoln@netapp.com
## @pod here

=head1 NAME

mount - invoke "mount" and "umount" idempotently

=head1 SYNOPSIS

    use mount qw(mount umount mounts);

    my $mountpoint = mount($server, $path, %options);
    my $mount_list = mounts($client);
    umount($mountpoint, %options);

=head1 DESCRIPTION

A library of functions to create, list, and destroy mountpoints.

This library is intended to supersede F<NFS/Mount.thpl>,
F<NFS/Mount.pm>, and F<lib/mount_old.pm>.  (F<NFS/Mount.pm> and
F<mount_old.pm> both suffer from hanging the test whenever NFS
accesses to the mountpoint would hang.  However, this library makes
use of NATE connections to perform its work and avoid that effect,
something which may not work for tests which run as root, for
example).

This library is perhaps superseded by F<lib/GenericMount.pm>

=head1 FUNCTIONS

=cut

package mount;

use strict;
use Params::Validate qw(validate validate_pos SCALAR);
use Utils qw(make_host make_conn);
use Tharn qw(say logcomment logresult connect param);
use SayWrap qw(sayEOF);
use base qw(Exporter);
use Net::IP qw(ip_get_version);

my $addr_type=param('ADDR_TYPE',"-default","inet4");


our @EXPORT = qw(mount umount);
our @EXPORT_OK = qw(mounts);

# Prototypes.

sub mount ( $$% );
sub _get_dev ( $$$ );
sub _try_mount ( $$$$$% );
sub _mountpoint ( $$% );
sub _mount_command ( $$$$% );
sub _real_nfs_mount_opts ( $$$$$ );

sub umount ( $% );
sub _umount_command ( $$% );


# As quote_command(), but do the right thing for rsh instead of the
# right thing for rlogin (don't break lines with \r expecting to be
# giving input to a tty; instead expect that line length is
# unlimited).

sub my_quote_one ( $ ) {
    my ($arg) = @_;
    if ($arg =~ /^[-A-Za-z0-9_,.\/]+$/) {
	return $arg;
    } else {
	$arg =~ s{\'}{\'\\\'\'};
	return "'$arg'";
    }
}

sub my_quote_command ( @ ) {
    return join(" ",map { my_quote_one($_) } @_);
}

# As Utils::make_conn, except when the host seems to be localhost use
# a more direct style of connection, and be sure to die on connection
# failure instead of return-undef-and-set-$@.

sub my_make_conn {
    my ($ch, $client) = @_;
    my $cc;
    if ($ch->id eq "localhost") {
	# Make a connection called "local" which thinks it will be
	# an rsh connection, but pass commands to a local shell
	# instead
	$cc = connect("#local","conntype=rsh-persay");
	$cc->conncmd("/bin/sh","-c");
    } else {
	$cc = make_conn($client) or die $@;
    }
    return $cc;
}

# "say, with kill on timeout"

# Do the usual say on the connection, but send an interrupt if the
# error is a timeout.  Also, the complainspec will be '-eval_error',
# and the value returned will be $after (third return value of say).

# This is to try to make sure that a command which failed due to NFS
# hangs is more likely to be dead rather than hanging around.  (Just
# closing the rsh connection doesn't kill anything, and the kill tends
# to need to be SIGINT rather than any other signal, because Linux is
# very picky about what can interrupt something blocked on an 'intr'
# mount)

sub say_kot {
    my ($cc, @sayopts) = @_;
    my (undef, undef, $out) = say($cc, @sayopts, '-eval_error');
    if ($@ && $@ =~ /TIMEOUT/) {
	my $save = $@;
	$cc->send_intr();
	$@=$save;
    }
    return $out;
}

=head2 mount

  my $mountpoint = mount($server, $path, %options);

C<mount> attempts to cause a mount point to be established regardless
of previous state.  In particular, it will create the mount point
directory if it does not already exist, will invoke C<mount> binary
only if the mountpoint does not already have a mount, and will
C<umount> and try C<mount> again if the first attempt to mount fails.
It will retry only one time unless C<retries> is set.

C<mount> returns the mountpoint on success.  On error, it returns
C<undef> with the error in C<$@>.

It is intended that several independent tests may all share the same
mountpoints.  To accomplish this, C<mount> chooses a mountpoint based
on C<$server>, C<$path>, C<proto> and C<vers> values, and callers
should generally allow C<mount> to choose the mountpoint value.  Also,
callers who wish to permit other tests to run simultaneously should
generally ignore failures from C<umount> (because the mount may still
be in use by another test on the same machine) or not call C<umount>
at all (because another test may need the mountpoint but not be using
it at this exact moment).

In order to handle all possible wacky errors from the operating system
mount command but still succeed if the output contains mere warnings,
C<mount> ignores the output of the mount command.  Instead, it
succeeds iff the chosen mountpoint has become mounted.  The presence of a
mounted filesystem is detected by examining the C<st_dev> fields in
C<stat> of the mountpoint and its parent.  So far I have found no
common shell script tools which can do this and only this, so C<mount>
uses a bit of perl code by invoking C<perl -e>.

If an error of any kind occurs in the first C<mount> attempt (usually
that accessing the mountpoint returns a "stale file handle" error, or
the given timeout is exceeded) then C<mount> automatically calls
C<umount> (with the same timeout and with the C<force> and C<lazy>
options) and tries the same mount again.  Any errors with the
C<umount> are ignored.  Any errors with the second C<mount> are not
retried and are returned as an error.

=over

=item Arguments:

Mount has two types of operation (C<nfs> or C<device>) which affect
the meanings of its two positional arguments.  The default type is
C<nfs>, but one can choose type C<device> instead by passing the option
C<< type=>device >>.

=over

=item $server

For type C<nfs>, the DNS name or IP address of a file server to mount.
For type C<device>, a more informal name of the server providing the
device (used in choosing the name of the mountpoint, logging, etc.)

=item $path

For type C<nfs>, the export point on that file server to mount.  For
type C<device>, the name of a block device (local to client) to mount.

=back

=item Options

=over

=item type

Either C<nfs> or C<device>.  Default is C<nfs>.

=item client

Hostspec or Connectrec of a client that should perform the mount.  A
host object (value returned by C<host>) is preferred, in which case
the connection C<#rsh> to that host will be used.  The default is
"localhost", which uses rsh looped back to perform the mount locally.

If a Connectrec is given it must be for a C<per-say> type of
connection i.e. one which effectively connects/disconnects for every
command.  (Thus, for example, the Connectrec returned by
connect("#rlogin") or connect("#telnet") could not be used).

=item proto

For NFS mounts, what RPC transport protocol to use, such as "udp" or
"tcp".  Default is undef (which implies protocol left unspecified to
mount(8)).

=item vers

For NFS mounts, what version of NFS to use, such as "2" or "3" or
"4").  Default is undef (which implies version left unspecified to
mount(8))

=item mountopts

What mount options to use (what to pass to the C<-o> option of the
C<mount> executable).  For type C<device>, this value will be passed
to C<mount> unmodified, and the default is empty.  For type C<nfs>,
the default is 'hard,intr,noquota' and before the value is passed to
the C<mount> executable the C<proto> and C<vers> options given above
will be added and the options possibly translated.

The way the C<nfs>-mode option translation works is that options
should be given in Solaris style, and if the host performing the mount
is Linux or FreeBSD, known options will be translated into that
operating system style.  For Linux, options are left alone except
"noquota" is dropped if present and "vers=4" is dropped and expressed
by changing C<fstype> instead.  For FreeBSD/NetBSD/OpenBSD, the
options "hard", "soft", "intr", "bg", "proto", and "vers" are
translated into their equivalents.  No translation is yet performed
for HP/UX or AIX.

=item fstype

The file system type to pass to the C<mount> executable (with "-t" or
"-F" option, depending on operating system).  The default is not to
pass any filesystem name, as the OS choice of defaults is generally
correct (except that on Linux when C<type> is C<nfs> and C<vers> is
C<4> then the fstype defaults to "nfs4", as that is what current
NFSv4-capable C<mount> executables on Linux require).

=item mountpoint

What directory to mount this filesystem onto (the second non-option
argument to the C<mount> executable).  The default is
"/t/server/extra" where "server" is the given C<$server> value, and
"extra" is a string generated from the C<$path> and the values given
for C<proto> and C<vers> options.

As described above, one should generally let C<mount> choose the value
of this option or else do at least as good a job of avoiding ever
having separate tests choose identical mountpoint paths but
non-identical mount options.

=item timeout

How many seconds to wait for a mount or umount attempt to finish.  The
default is 60 seconds.  (C<mount> may take a total time which is
longer than this if it has to retry, since the timeout is for each
mount and umount)

=item chmod

If given, what mode (permissions) to set on the mountpoint after
mounting.  This can be any value accepted by the C<chmod(1)> Unix
command.

=item chown

If '1', specifies that the mountpoint should be 'chowned' to the
current euid. The default is '0'.

=item retries

If given, the number of retries to attempt before failing on the 
mount.  Defaults to 1.

=back

=back

=cut

sub mount ( $$% ) {
    my ($server, $path, @opts) =
	validate_pos(@_,
		     { type => SCALAR },
		     { type => SCALAR }, 
		     (0) x (@_ - 2) );
    my %opts = validate (@opts, {
	type => { regex => qr/^(nfs|device)$/, default => "nfs" },
	proto => 0,
	vers => 0,
	mountopts => 0,
	mountpoint => 0,
	client => { default => "localhost" },
	fstype => 0,
	timeout => { regex => qr/^-?\d+$/, default => 60 },
	chmod => 0,
        chown => {optional => 1, default => 0},
	retries => { default => 1 },
    });

    return eval {
	my $ch = make_host($opts{client});
	my $cc = my_make_conn($ch, $opts{client});

	my $mountpoint = _mountpoint($server, $path, %opts);

	# Attempt to mount up to $opt{retries} times, with umount in
	# between.  Ignore errors until the very last mount attempt.

	while (1) {
	    eval {
		_try_mount($ch,$cc,$server,$path,$mountpoint,%opts);
	    };
	    if ($@) {
		die $@ if ($opts{retries} <= 0);
		logcomment("Retrying Mount... ($opts{retries} attempt[s] remaining)");
		umount($mountpoint,client=>$ch,force=>1,lazy=>1,timeout=>$opts{timeout});
		$opts{retries}--;
		sleep 1;
	    } else {
		return $mountpoint;
	    }
	}
    };
}


# Return the 'dev' field of the st_dev value in stat($path)

# Communications problems are handled with 'die', but errors which are
# probably stat() returning an error are returned as the second return
# value, to make it easier to sometimes ignore the latter kind of
# error but never ignore the former.

sub _get_dev ( $$$ ) {
    my ($cc, $path, $timeout) = @_;
    my $ch = $cc->hostp();
    my $quoted_path = my_quote_command($path);
    my $cmd = q{perl -MEnglish -e '($dev)=stat($ARGV[0]) or die $OS_ERROR; print qq(-$dev-)'};
    if ($ch->hosttype() !~ /^unix-(solaris|linux)/) {
	# This is not necessary on Solaris or most imaginable Linux
	# distributions, since they come with perl in /usr/bin.
	$cmd = q{env 'PATH=/usr/local/bin:$PATH' } . $cmd;
    }
    my $out = say_kot($cc,"$cmd $quoted_path",
		      '-timeout',$timeout);
    die $@ if $@;
    my ($st_dev) = ($out =~ /^-(-?\d+)-/m);
    if (!defined($st_dev)) {
	chomp $out;
	return (undef, "remote stat of \"$path\" on \"" . $ch->id() . "\" failed: $out\n");
    }
    return ($st_dev);
}

# The meat of the mount() call.  This function is split from mount()
# because it may be called a second time (after trying to umount) if
# the first call has an exception.

sub _try_mount ( $$$$$% ) {
    my ($ch, $cc, $server, $path, $mountpoint, %opts) = @_;

    my $sudo_path = param('SUDO_PATH',
			  '-mesg','(NATE is expected to have set this)');
    # Must finish by this absolute time
    my $timeout = $opts{timeout} || 60;
    my $endtime = Tharn::timeout2time($opts{timeout});
	
    my $abovepoint = $mountpoint;
    $abovepoint =~ s@/[^/]+/*$@@;

    my ($above_dev, $mount_dev, $out);

    ($above_dev) = _get_dev($cc,$abovepoint,$endtime);
    if (defined($above_dev)) {
	($mount_dev) = _get_dev($cc,$mountpoint,$endtime);
    }

    if (!defined($mount_dev)) {
	# stat of mount directory failed.  Let us try to create it (or
	# fail trying).

	my $mkdir_command = my_quote_command($sudo_path,"mkdir","-p",$mountpoint);

	my $out = say_kot($cc,
			  "$mkdir_command  && echo '_mkdir_ok_'",
			  '-timeout',$endtime);
	if ($@) {
	    die "mount: error while creating directory \"$mountpoint\": $@";
	}
	if ($out !~ "_mkdir_ok_") {
	    die "mount: failed to create directory \"$mountpoint\".  ".
		"mkdir printed: $out\n";
	}
    }

    unless (defined($mount_dev) && ($above_dev != $mount_dev)) {
	# unless the mountpoint had already existed (before the
	# possible mkdir attempt, above) and had had a mount, we need
	# to mount and check that the mountpoint now has a mount.

	my $mount_command = _mount_command($ch->hosttype(), $mountpoint, $server, $path, %opts)
	    . " && echo _-mount_ok-_";

	my $out = say_kot($cc,$mount_command,
			  '-timeout',$endtime);
	if ($@) {
	    die "mount: error invoking mount: $@";
	}
	my $err;

	($above_dev,$err) = _get_dev($cc,$abovepoint,$endtime);
	die $err if $err;
	($mount_dev,$err) = _get_dev($cc,$mountpoint,$endtime);
	die $err if $err;

	if ($above_dev == $mount_dev) {
	    die "mount: nothing is mounted at \"$mountpoint\".  mount must have failed.\ncommand: $mount_command\noutput: $out\n";
	}
    }

    # (at this point there is a mounted mountpoint.
    # chown and chmod mountpoint if it was requested)

    if ($opts{chown}) {
	# 'chown' mountpoint to our euid, to avoid 'permission denied' problems later.
	my $chown_command = my_quote_command($sudo_path,"chown", "$>", $mountpoint);

	my $chown_out = say_kot($cc,
			  "$chown_command  && echo '_chown_ok_'",
			  '-timeout',$endtime);

	if ($@ || ($chown_out !~ "_chown_ok_")) {
		logresult('WARN', "Could not 'chown' mountpoint \"$mountpoint\": $@");
	}
    }

    if ($opts{chmod}) {
	my $chmod_command =
	    my_quote_command($sudo_path,
			  "chmod",$opts{chmod},$mountpoint) .
	        " && echo _-chmod_ok-_ ";
	my $out = say_kot($cc,$chmod_command,
			  '-timeout',$endtime);
	if ($@) {
	    die "mount: error invoking chmod: $@";
	}
	if ($out !~ /_-chmod_ok-_/) {
	    die "mount: error setting mode of mountpoint.\ncommand:$chmod_command\noutput:$out\n";
	}
    }
}

# Compute the mountpoint given various mount arguments

sub _mountpoint ( $$% ) {
    my ($server, $path, %opts) = @_;
    if ($opts{mountpoint}) {
	return $opts{mountpoint};
    }
    my $extra = $path;
    $extra =~ s@^/@@;
    if (!length($extra)) {
	$extra = "root";
    }
    if ($opts{proto}) {
	$extra .= "-$opts{proto}";
    }
    if ($opts{vers}) {
	$extra .= "-$opts{vers}";
    }
    $extra =~ tr@ /@_@;
    return "/t/$server/$extra";
}


# Return the 'mount' command-line given various mount() function
# arguments

sub _mount_command ( $$$$% ) {
    my ($hosttype, $mountpoint, $server, $path, %opts) = @_;

    my $sudo_path = param('SUDO_PATH',
			  '-mesg','(NATE is expected to have set this)');
    my $type=$opts{type};
    my $mountopts = $opts{mountopts};
    my $fstype = $opts{fstype};

    if ($type eq "nfs") {
        ($fstype, $mountopts) = _real_nfs_mount_opts($fstype, $mountopts, $hosttype, $opts{proto}, $opts{vers});
    }

    my ($fsopt, $mount) = _system_mount_command($hosttype);

    my @cmd = ($sudo_path,$mount);
    if ($fstype) {
        push @cmd, ($fsopt,$fstype);
    }
    if ($mountopts) {
        push @cmd, ('-o',$mountopts);
    }

    if ($server =~ /^\[(\S+)$\]/ && ip_get_version($1) == 6) {
        $server = $server;
    } else {
        my $inet = ip_get_version($server);
        
        if (! defined($inet)) {
            my $host = Tharn::host($server);
            $server = $host->net_best_address(include=>{family=>$addr_type});
            if (ip_get_version($server) == 6) {
                $server = qq{\[$server\]};
            }
        } elsif ($inet == 6) {
            $server = qq{\[$server\]};
        }
    }

    push @cmd, ($type eq "nfs") ? "$server\:$path" : $path;
    push @cmd, $mountpoint;

    return my_quote_command(@cmd);
}

=head2 mounts

  my $mount_list = mounts($client);

C<mounts> returns a hash reference listing the filesystems currently
mounted on $client.  The keys of the hash are host:volume mount specs,
and each value is a hash containing the following key/value pairs:

    src_host	the source hostname
    src_path	path of the source volume
    mountpoint	path to where the filesystem is mounted
    fstype	the type of filesystem (nfs, ufs, etc)
    mount_date	when the mount was created
    fsopts	a list of the options associated with the mount
    fsopts_str	the original string from which the fsopts list is
		derived
    client_hosttype	the client hosttype (useful for parsing
		    the fsopts_str)

C<mounts> is not exported by default. It returns a hash reference,
which might be empty if there are no mounts matching the given criteria.
The function dies if it does not know how to parse the mount command
output for the current hosttype.

=over

=item Arguments

=over

=item $client

The DNS name or IP address of the client system whose mounts should be
enumerated.

=back

=back

=cut

sub mounts($) {
    my ($client_hostspec) = @_;

    my $ch = make_host($client_hostspec);
    my $cc = my_make_conn($ch,$client_hostspec);

    my $hosttype = $ch->hosttype();

    my ($unused, $mount, $enum_opt) = _system_mount_command($hosttype);

    my $cmd = "$mount $enum_opt";

    my $text = sayEOF($cc, $cmd);

    return _parse_mount_listing($hosttype, $text);
}

# Return the system-specific 'mount' command and options

sub _system_mount_command($) {
    my ($hosttype) = @_;

    my $mount;
    my $fsopt;
    my $enum_opt = '-v';

    if ($hosttype =~ /^unix-(aix|hpux|solaris)/) {
	$mount = "/usr/sbin/mount";
	$fsopt = "-F";
	$enum_opt = '' if ($hosttype =~ /^unix-aix/);
    } elsif ($hosttype =~ /^unix-linux/) {
	$mount = "/bin/mount";
	$fsopt = "-t";
    } elsif ($hosttype =~ /^unix-(net|free|open)bsd/) {
	$mount = "/sbin/mount";
	$fsopt = "-t";
    } else {
	$fsopt = "-t";
	$mount = "mount";
    }
    return($fsopt, $mount, $enum_opt);
}
    
sub _parse_mount_listing {
    my ($hosttype, $mount_output) = @_;

    my %mount_list;

    my @lines = split("\n", $mount_output);

    if ($hosttype =~ /^unix-aix/) {
	# discard the first two (header) lines
	shift @lines;
	shift @lines;
    }

    foreach my $line (@lines) {
	my @fields = split(/\s+/, $line, 6);

	my ($mount, $mountpoint, $fstype, $fsopts, $src_host, $src_path, $mount_date);
	if ($hosttype =~ /^unix-(hpux|solaris|linux)/) {
	    ($mount, $mountpoint, $fstype, $fsopts) = @fields[0,2,4,5];
	} elsif ($hosttype =~ /^unix-aix/) {
	    if (scalar(@fields) == 7) {
		unshift @fields, 'localhost';
	    }
	    my ($month, $date, $time);
	    ($src_host, $src_path, $mountpoint, $fstype, $month, $date, $time, $fsopts) = @fields;
	    $mount = "${src_host}:${src_path}";
	    $mount_date = join(' ', $month, $date, $time);
	} elsif ($hosttype =~ /^unix-(net|free|open)bsd/) {
	    ($mount, $mountpoint, $fsopts) = @fields[0,2,3];
	} else {
	    die "don't know how to parse mount listing for hosttype '$hosttype'";
	}

	$mount = "localhost:$mount" if ($mount !~ /:/);

	if ($hosttype =~ /^unix-(hpux|solaris|linux|(net|free|open)bsd)/) {
	    ($src_host, $src_path) = split(':', $mount);
	    if (!$src_path) {
		$src_path = $src_host;
		$src_host = 'localhost';
	    }
	}

	my @fsopts;
	# split the fs options string
	if ($hosttype =~ /^unix-solaris/) {
	    ($fsopts, $mount_date) = (split(/\s+/, $fsopts, 3))[0,2];
	    @fsopts = split('/', $fsopts);
	} elsif (
	    ($hosttype =~ /^unix-(hpux|linux|aix|(net|free|open)bsd)/)
	) {
	    # remove any enclosing parens
            my @chars = split('', $fsopts);
            if (($chars[0] eq '(') and ($chars[$#chars] eq ')')) {
                @chars = @chars[1..scalar($#chars-1)];
            }
            $fsopts = join('', @chars);

            @fsopts = split(',', $fsopts);
	} else {
	    # nothing to do, leave $fsopts as is
	}

	$mount_list{$mount} = {
	    src_host	=> $src_host,
	    src_path	=> $src_path,
	    mountpoint	=> $mountpoint,
	    fstype	=> $fstype,
	    mount_date	=> $mount_date || 'UNKNOWN',
	    fsopts	=> \@fsopts,
	    # include these so callers can always parse fsopts themselves
	    client_hosttype => $hosttype,
	    fsopts_str	=> $fsopts,
	};
    }

    return \%mount_list;
};

# Given Solaris-style mount options for NFS, return what a $hosttype
# host should be using as mount options and filsystem type instead.

sub _real_nfs_mount_opts ( $$$$$ ) {
    my ($fstype, $mountopts, $hosttype, $proto, $vers) = @_;

    if (!$mountopts) {
	$mountopts = "hard,intr,noquota";
    }
    if ($proto) {
	$mountopts .= ",proto=$proto";
    }
    if ($vers) {
	$mountopts .= ",vers=$vers";
    }

    if ($hosttype =~ /^unix-(aix|hpux|solaris)/) {
	# (Only a few changes needed, since we mostly define the mount
    # options so that there would be no fixup needed for Solaris, 
    # and from the changes others have made to NFS/Mount.thpl, 
    # apparently AIX and HP/UX don't need any fixup either?)
    my @opts;
    for (split(/,/,$mountopts)) {
        if (/^vers=(\d+)/) {
            # No distinction among the various possible minor
            # versions ("4", "4.0", "4.1", etc. all become just
            # "4").
            push @opts, "vers=$1";
        } else {
            # assume any other options are equivalent
            push @opts, $_;
        }
    }
    $mountopts = join(',',@opts);
    } elsif ($hosttype =~ /^unix-linux/) {
	my @opts;
	for (split(/,/,$mountopts)) {
	    if (/^noquota$/) {
		# (no corresponding option)
		} elsif (/^vers=4/) {
        # Instead of a "vers=4" mount option, Linux
        # distributions have a separate mount program
        # (separate filesystem type) for NFSv4.

		    $fstype="nfs4";

        # Translate any specification of minor version number
        # (4.0, 4.1, 4.2, etc.) into a "minorversion" option,
        # which is apparently new with NFS v4.1 support
	        if (/^vers=4\.(\d+)/) {
	            push @opts, "minorversion=$1";
	        }
		} else {
		# assume any other options are equivalent
		push @opts, $_;
	    }
	}
	$mountopts = join(',',@opts);
    } elsif ($hosttype =~ /^unix-(net|free|open)bsd/) {
	my @opts;
	for (split(/,/,$mountopts)) {
	    if (/^noquota$/) {
		# (no corresponding option)
	    } elsif (/^hard$/) {
		# (no corresponding option; hard is assumed)
	    } elsif (/^intr$/) {
		push @opts, "-i";
	    } elsif (/^bg$/) {
		push @opts, "-b";
	    } elsif (/^soft$/) {
		push @opts, "-s";
	    } elsif (/^proto=udp/i) {
		push @opts, 'mntudp';
	    } elsif (/^proto=tcp/i) {
		push @opts, 'tcp';
	    } elsif (/^vers=(\w+)/) {
		push @opts, "nfsv$1";
	    } else {
		# (otherwise, assume it's the same option on FreeBSD as Solaris)
		push @opts, $_;
	    }
	}
	$mountopts = join(',',@opts);
    }
    return ($fstype, $mountopts);
}

=head2 umount

  umount($mountpoint, %options);

C<umount> invokes the operating system C<umount> command and
interprets its results.

It returns a true value on success.  On error it returns C<undef> with
the error message in C<$@>.

=over

=item Arguments

=over

=item $mountpoint

The path to unmount

=back

=item Options

=over

=item client

Hostspec or Connectrec of a client that should perform the umount (see
the corresponding entry for C<mount>).

=item force

If true, and if the system is known to have such a flag (Solaris,
Linux), pass the '-f' flag to umount to ask for umount to be forced,
which should cause any existing users of the mountpoint to get I/O
errors.  This is more effective on Solaris (the mountpoint is
unmounted even if busy; open files and cwd on the old mount get errors
from now on) than on Linux (pending I/O at the time of the umount is
cancelled with errors, with the hope that this will cause processes
keeping the mount busy to let go)

=item lazy

If true, and if the system is known to have such a flag (Linux), pass
the '-l' flag to umount to ask for umount to be lazy, which should
make a busy mountpoint become private to those processes still using
it, effectively unmounted from the point of view of most processes

=item timeout

How many seconds to wait for a mount or umount attempt to finish.  The
default is 60 seconds.

=back

=back

=cut

sub umount ( $% ) {
    my ($mountpoint, @opts) =
	validate_pos(@_,
		     { type => SCALAR }, 
		     (0) x (@_ - 1) );
    my %opts = validate (@opts, {
	client => { default => "localhost" },
	force => { default => 0 },
	lazy => { default => 0 },
	timeout => { default => 60 },
    });

    return eval {
	my $ch = make_host($opts{client});
	my $cc = my_make_conn($ch, $opts{client});

	AGAIN: {
	    my $umount_command = _umount_command($ch->hosttype(),$mountpoint,%opts);
	    my $out = say_kot($cc,"$umount_command && echo _-umount_ok-_",
			      '-timeout',$opts{timeout});
	    die $@ if $@;
	    if ($out =~ /illegal option -- f/) {
		# We guessed incorrectly that Solaris was capable of
		# taking a force flag: older versions cannot.
		$opts{force}=0;
		goto AGAIN;
	    }
	    if ($out =~ /illegal option -- l/) {
		# We guessed incorrectly that Linux was capable of
		# taking an umount flag.
		$opts{lazy}=0;
		goto AGAIN;
	    }
	    if ($out !~ /_-umount_ok-_/) {
		$out =~ s/^\s+//;
		$out =~ s/\s+$//;
		die "umount of \"$mountpoint\" failed.\ncommand: $umount_command\noutput: $out\n";
	    }
	}
	return 1;
    };
}

# Return the 'umount' command line to use given various umount()
# function arguments

sub _umount_command ( $$% ) {
    my ($hosttype, $mountpoint, %opts) = @_;
    my @cmd;
    my ($can_force, $can_lazy);
    my $sudo_path = param('SUDO_PATH',
			  '-mesg','(NATE is expected to have set this)');
    my $umount;
    if ($hosttype =~ /^unix-solaris/) {
	$umount = "/usr/sbin/umount";
	$can_force=1;
    } elsif ($hosttype =~ /^unix-(aix|hpux)/) {
	$umount = "/usr/sbin/umount";
    } elsif ($hosttype =~ /^unix-linux/) {
	$umount = "/bin/umount";
	$can_force=1;
	$can_lazy=1;
    } elsif ($hosttype =~ /^unix-(net|free|open)bsd/) {
	$umount = "/sbin/umount";
    } else {
	$umount = "umount";
    }
    @cmd = ($sudo_path,$umount);
    if ($can_force && $opts{force}) {
	push @cmd, ("-f");
    }
    if ($can_lazy && $opts{lazy}) {
	push @cmd, ("-l");
    }
    push @cmd, $mountpoint;
    return my_quote_command(@cmd);
}

1;

=head1 EXAMPLE

    use mount;

    param('FILER','-mesg',"hostspec for a filer");
    my $h = host($FILER);
    param('FILER_IP','-default',$h->default_ip());
    param('CLIENT','-default',"");
    param('DEVICE','-default',"/dev/dsk/c3t3d14s6");

    my $mountpoint = mount($FILER_IP,"/",client=>$CLIENT)
        or die $@;

    my $mountpoint = mount($h->id(),$DEVICE,type=>"device",client=>$CLIENT)
        or die $@;

    my $mounts = mount::mounts($CLIENT);
    foreach my $m (keys %{$mounts}) {
	print "$m is mounted on " . $m->{mountpoint} . "\n";
    }

    umount($mountpoint)
        or logresult('WARN', $@);

=cut
