# this is basically -*- perl -*-
# vim: set ft=perl tabstop=4 shiftwidth=4 expandtab shiftround: 
#
# $Id:$
#
# Copyright (c) 2007 Network Appliance Inc.
# All rights reserved.
#

=head1 NAME

nblade::CIFS::Auth - Routines for testing CIFS Authentication

=head1 SYNOPSIS

   use nblade::CIFS::Auth;

=head1 DESCRIPTION

This package provides routines for testing authentication using various CIFS methods

=head1 METHODS



=cut

package nblade::CIFS::Auth;

use strict;
use warnings;
use Data::Dumper;
use File::Slurp;
use TracePkg;
use Options;
use Params::Common;
use Params::CIFS;
use Params::Validate qw(:all);
use MapShare qw(:all);
use Server; # do not create a global $Server object though - this library is used by threaded tests which will
            # make hundreds of SSH connections back to the Server

my $Checked_Kerberos_Config = 0;
my $Cleaned_Win32_Mappings = 0;

my $Trace = TracePkg->new();
my $Options = Options->new();

$Options->process();
my $Params=$Options->get_hash();


#
# Handle using smbclient to do authentication requests
#
sub Validate_Auth_SMBClient {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            CIFS_Server => { optional => 0, type => SCALAR, default => $Params->{CIFS_SERVER} },
            CIFS_Share  => { optional => 0, type => SCALAR, default => $Params->{CIFS_SHARE} },
            CIFS_Domain => { optional => 0, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            CIFS_User   => { optional => 0, type => SCALAR, default => $Params->{CIFS_USER} },
            CIFS_Password   => { optional => 0, type => SCALAR, default => $Params->{CIFS_PASSWORD} },
            Data_Interfaces => { optional => 0, type => ARRAYREF, default => $Params->{DATA_INTERFACES} },
            SMB_Client      => { optional => 1, type => SCALAR, default => $Params->{SMBCLIENT} },
            Kerberos_Realm  => { optional => 1, type => SCALAR, default => $Params->{KERBEROS_REALM}},
            use_kerberos    => { optional => 1, type => BOOLEAN, default => 0, depends => ["Kerberos_Realm"] },
            get_kerberos_creds => { optional => 1, type => BOOLEAN, default => 1 },
            destroy_kerberos_creds => { optional => 1, type => BOOLEAN, default => 1 },
            enable_signing  => { optional => 1, type => BOOLEAN, default => 0 },
            cmd_to_run  => { optional => 1, type => SCALAR, default => '' },
        },
    );
    $Trace->debug(msg => "Input input_vars = " . Data::Dumper::Dumper(\%input_vars));

    # if we are supposed to use kerberos, then get our credentials
    if ( $input_vars{use_kerberos} == 1 ) {
        if ($input_vars{get_kerberos_creds} == 1) {
            eval {
                nblade::CIFS::Auth::Get_Kerberos_Tickets(
                    Kerberos_Realm  => $input_vars{Kerberos_Realm},
                    CIFS_User       => $input_vars{CIFS_User},
                    CIFS_Password   => $input_vars{CIFS_Password},
                );
            };
            if ( $@ ) {
                $Trace->fatal(msg =>
                    "Failed to get kerberos credentials for user '$input_vars{CIFS_User}\@$input_vars{Kerberos_Realm}' '$@'");
            }
        } else {
            my $output = `$Params->{KLIST}`;
            $Trace->debug(msg => "Assuming we have kerberos credentials already '$output'");
        }
        # log in to the server
        my $cmd = "$input_vars{SMB_Client} -I $input_vars{Data_Interfaces}[0] -W $input_vars{CIFS_Domain} -U $input_vars{CIFS_User} -k -d 4 -c \"$input_vars{cmd_to_run}\" //$input_vars{CIFS_Server}/$input_vars{CIFS_Share} $input_vars{CIFS_Password} 2>&1";
        $Trace->trace(msg => "Logging into '//$input_vars{CIFS_Server}/$input_vars{CIFS_Share}' as '$input_vars{CIFS_User}\@$input_vars{CIFS_Domain}' with command '$cmd'");
        my $output = `$cmd`;
        if ( $? > 0 ) {
            $Trace->fatal(msg => "Kerberos Authentication failed '$output' '$?'");
        } else {
            $Trace->trace(msg => "Kereros Authentication succeeded");
            $Trace->debug(msg => $output);
        }

        # nuke our kerberos credentials
        if ( $input_vars{destroy_kerberos_creds} == 1 ) { 
            eval {
                nblade::CIFS::Auth::Destroy_Kerberos_Tickets();
            };
            if ( $@ ) {
                $Trace->fatal(msg =>
                    "Unable to destroy kerberos credentials for user '$input_vars{CIFS_User}\@$input_vars{Kerberos_Realm}' '$@'");
            }
        } else {
            $Trace->debug(msg => "Leaving kerberos credentitals alone");
        }
    } else {
        # log in to the server
        my $cmd = "$input_vars{SMB_Client} -I $input_vars{Data_Interfaces}[0] -W $input_vars{CIFS_Domain} -U $input_vars{CIFS_User} -d 4 -c \"$input_vars{cmd_to_run}\" //$input_vars{CIFS_Server}/$input_vars{CIFS_Share} $input_vars{CIFS_Password} 2>&1";
        $Trace->trace(msg => "Logging into '//$input_vars{CIFS_Server}/$input_vars{CIFS_Share}' as '$input_vars{CIFS_User}\@$input_vars{CIFS_Domain}' with command '$cmd'");
        my $output = `$cmd`;
        if ( $? > 0 ) {
            $Trace->fatal(msg => "NTLM Authentication failed '$output' '$?'");
        } else {
            $Trace->trace(msg => "NTLM AUthentication succeeded");
            $Trace->debug(msg => $output);
        }
    }
    $Trace->exit();
} # end Validate_Auth_SMBClient

#
# Handle getting kerberos credentials for a user in a given domain
#
sub Get_Kerberos_Tickets {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            Kerberos_Realm  => { optional => 0, type => SCALAR },
            CIFS_User       => { optional => 0, type => SCALAR },
            CIFS_Password   => { optional => 0, type => SCALAR },
        },
    );
    $Trace->debug(msg => "Input input_vars = " . Data::Dumper::Dumper(\%input_vars));
    # if we are trying to log in with kerberos, check the realm 
    if ( $Checked_Kerberos_Config == 0 ) {
        Check_Kerberos_Config( Kerberos_Realm => $input_vars{Kerberos_Realm});
    }
    my $output = `$Params->{KLIST}`;
    $Trace->debug(msg => "KLIST: '$output'");

    # get kerberos credentials
    my $cmd = "$Params->{KINIT} $input_vars{CIFS_User}\@$input_vars{Kerberos_Realm}";
    $Trace->debug(msg => "KINIT: '$cmd'");
    open(KINIT, "|$cmd");
    # we might have some goofy characters we need to escape to pass in for the password
    print KINIT $input_vars{CIFS_Password};
    close(KINIT);
    $output = `$Params->{KLIST}`;
    if ( $output !~ /$input_vars{Kerberos_Realm}/ ) {
        $Trace->fatal(msg => "Unable to get kerberos tickets with '$cmd' using password '$input_vars{CIFS_Password}' '$output'");
    } else {
        $Trace->debug(msg => "KLIST: '$output'");
    }

    $Trace->exit();
} # end Get_Kerberos_Tickets

#
# Handle destroying kerberos credentials for a user
#
sub Destroy_Kerberos_Tickets {
    $Trace->entry();
    my $output = `$Params->{KDESTROY}`;
    if ( $? > 0 ) {
        $Trace->fatal(msg => "Failure during kerberos credentials destroy '$Params->{KDESTROY}' '$?' '$output'");
    } else {
        $Trace->trace(msg => "Removed kerberos credentials successfully.");
    }
    $Trace->exit();
} # end Destroy_Kerberos_Tickets

#
# Check out the local clients configurationg to see if kerberos should work
#
sub Check_Kerberos_Config {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,                                                                          
        {                                                                                                                     
            Kerberos_Realm  => { optional => 0, type => SCALAR, default => $Params->{KERBEROS_REALM} },
        },
    );
    $Trace->debug(msg => "input_vars = " . Data::Dumper::Dumper(\%input_vars));
    if ( $Checked_Kerberos_Config == 1 ) {
        $Trace->debug(msg => "Already checked the kerberos configuration once");
    } else {
        # we only really want to do this once, so set a global to keep us from checking too many times
        $Checked_Kerberos_Config = 1;
        my $krb5_conf;
        eval {
            $krb5_conf = File::Slurp::read_file("/etc/krb5.conf");
        };
        if ( $@ ) {
            $Trace->warn(msg => "Unable to read kerberos conf '/etc/krb5_conf' '$^E'");
        } elsif ( ($krb5_conf !~ /$input_vars{Kerberos_Realm}/) && ($krb5_conf !~ /dns_.*= true/) ) {
            $Trace->warn(msg => "This client has no knowledge of kerberos realm '$input_vars{Kerberos_Realm}' and doesn't fall back to DNS to look it up.  Most likely the Kerberos authentication is going to fail.  Consider enabling dns lookup or adding the realm to /etc/krb5.conf");                                                     
        } else {
            $Trace->trace(msg => "Client will most likely be able to log into Kerberos Realm");
        }
    }
    $Trace->exit();
}

# use protoconfig to test out authentication by_passing the NBlade and directly validating authenticate
sub Validate_Auth_Protoconfig {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            CIFS_User       => { optional => 0, type => SCALAR, default => $Params->{CIFS_USER} },
            CIFS_Password   => { optional => 0, type => SCALAR, default => $Params->{CIFS_PASSWORD} },
            Vserver_ID      => { optional => 1, type => SCALAR },
            CIFS_Server     => { optional => 1, type => SCALAR, default => $Params->{CIFS_SERVER} },
            use_kerberos    => { optional => 1, type => BOOLEAN, default => 0 },
        },
    );
    $Trace->debug(msg => "input_vars = " . Dumper(\%input_vars));
    # we need to have a Vserver_ID or CIFS_Server in order to run
    if ( !defined $input_vars{Vserver_ID} && $input_vars{CIFS_Server} eq "" ) {
        $Trace->fatal(msg => "A 'CIFS_Server' or 'Vserver_ID' is required to call this routine.", 
            result => "SCRIPT");
    } elsif ( !defined $input_vars{Vserver_ID} ) { # go figure out the Vserver_ID for this CIFS_Server
        eval {
            $input_vars{Vserver_ID} = nblade::CIFS::Auth::Get_Vserver_ID(CIFS_Server => $input_vars{CIFS_Server});
        };
        if ( $@ ) {
            $Trace->fatal(msg => "Unable to determine the Vitrual Server ID for CIFS Server '$input_vars{CIFS_Server}' '$@'");
        }
    }
    # now attempt to authenticate the user with protoconfig
    my $cmd = "/usr/sbin/protoconfig -Auth";
    if ( $input_vars{use_kerberos} == 1 ) {
        $Trace->warn(msg => "Kerberos Authentication with protoconfig does not work at this time and will most likely fail.");
        $cmd .= " -kerberos ";
    } else {
        $cmd .= " -ntlm ";
    }
    $cmd .= " -vserverid=$input_vars{Vserver_ID} -username=$input_vars{CIFS_User} -password=$input_vars{CIFS_Password}";
    my $output = "";
    eval {
        my $server = Server->new();
        $server->shell(cmd => $cmd, output => \$output);
    };
    if ( $@ ) {
        $Trace->fatal(msg => "Failed to authenticate user '$input_vars{CIFS_User}' with password '$input_vars{CIFS_Password}' '$output' '$@'");
    } else {
        # we need to parse the output to know if the user actually got logged in.  protoconfig only sets error
        # if the command line is wrong or something like that
        my $cifs_user = quotemeta($input_vars{CIFS_User}); # it is possible a user will have "funky" characters in their name
        $Trace->debug(msg => "protoconfig output: '$output'");
        if ( $output =~ /the authentication request failed with DC error (.+)/i ) {
            $Trace->fatal(msg => "Unable to log in '$input_vars{CIFS_User}' '$1' '$output'");
        } elsif( $output =~ /No description exists for this error/i ) {
            $Trace->fatal(msg => "Unable to log in '$input_vars{CIFS_User}' '$output'");
        } elsif ( ($output =~ /User Name:\s+$cifs_user/i) &&  
            ( $output =~ /Session Key: \s+(\S.)+/i ) && ( $output =~ /Primary Group Sid:/i ) &&
            ( $output =~ /Domain Sid\/Rids/i ) ) {
            $Trace->trace(msg => "User '$input_vars{CIFS_User}' successfully authenticated");
        } else {
            $Trace->comment(msg => "Unsure if user '$input_vars{CIFS_User}' was authenticated or not '$output'", result => "INSPECT");
        }
    }

    $Trace->exit();
} 

# given a CIFS Server, determine the Virtual Server ID that the nblade uses for this server
sub Get_Vserver_ID {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            CIFS_Server => { optional => 0, type => SCALAR, default => $Params->{CIFS_SERVER} },
        },
    );
    $Trace->debug(msg => "input_vars = " . Dumper(\%input_vars));
    
    my $output;
    my $vserver_id = 0;
    my $server = Server->new();
    $server->shell(cmd => "protoconfig -Config -Server -l", output => \$output);

    foreach my $line ( split(/\n/, $output) ) {
        $Trace->debug(msg => "line = '$line'");
        if ( $line =~ /Virtual Server ID:\s+(\d+)\s$/ ) {
            $Trace->debug(msg => "Found Virtual Server ID: '$1'");
            $vserver_id = $1;
        } elsif ( $line =~ /Name:\s+(.+)\s$/ ) {
            # if we match the name of the vserver_id kick out now
            $Trace->debug(msg => "Found CIFS Server Name: '$1'");
            if ( $1 =~ /$input_vars{CIFS_Server}/i ) {
                last;
            } else {
                $vserver_id = 0;
            }
        }
    }
    if ( $vserver_id == 0 ) {
        $Trace->fatal(msg => "Unable to map the CIFS Server '$input_vars{CIFS_Server}' to a VServer ID on the stream from '$output'", result => "CONFIG");
    } else {
        return $vserver_id;
    }
    $Trace->exit();
}

# validate a windows client can authenticate to the server - simply uses MapShare to map a drive with the
# approprtiate username/password and unmap the share rather than reimplementing things.
sub Validate_Auth_Win32 {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            Client  => { optional => 0, type => SCALAR, default => $Params->{CLIENT}[0] },
            CIFS_Server => { optional => 0, type => SCALAR, default => $Params->{CIFS_SERVER} },
            CIFS_Share  => { optional => 0, type => SCALAR, default => $Params->{CIFS_SHARE} },
            CIFS_User   => { optional => 0, type => SCALAR, default => $Params->{CIFS_USER} },
            CIFS_Password   => { optional => 0, type => SCALAR, default => $Params->{CIFS_PASSWORD} },
            CIFS_Domain     => { optional => 1, type => SCALAR, default => $Params->{CIFS_DOMAIN} },
            CIFS_Drive      => { optional => 1, type => SCALAR, default => $Params->{CIFS_DRIVE} },
            timeout     => {optional => 1, type => SCALAR, default => 100},
        },
    );
    $Trace->debug(msg => "input_vars = " . Data::Dumper::Dumper(\%input_vars));

    # make sure we don't have any other mappings for this server or else we might get an error
    if ( $Cleaned_Win32_Mappings == 0 ) {
        eval {
            Clean_Win32_Mappings(CIFS_Server => $input_vars{CIFS_Server}, Client => $input_vars{Client}, force => 1);
        };
        if ( @$ ) {
            $Trace->fatal(msg => "Tried to clean any mappings to server '$input_vars{CIFS_Server} on the client and encountered an error. '$@'");
        }
    }
    # try to map the drive
    my $drive;
    eval {
        $drive = MapShare::map_cifs_share(
            server  => $input_vars{CIFS_Server},
            share   => $input_vars{CIFS_Share},
            client  => $input_vars{Client},
            domain  => $input_vars{CIFS_Domain},
            username    => $input_vars{CIFS_User},
            password    => $input_vars{CIFS_Password},
            drive   => $input_vars{CIFS_Drive},
            timeout => $input_vars{timeout},
        );
    };
    if ( $@ || !defined $drive ) {
        $Trace->fatal(msg => "Unable to map share for server '$input_vars{CIFS_Server}' as user
            '$input_vars{CIFS_Domain}\\$input_vars{CIFS_User}' '$@'");
    } else {
        $Trace->trace(msg => "Successfully mapped a share for user '$input_vars{CIFS_Domain}\\$input_vars{CIFS_User}' on server '$input_vars{CIFS_Server}' to drive '$drive'");
    }

    # remove the mapping
    eval {
        MapShare::unmap_cifs_share(drive => $drive, client => $input_vars{Client} );
    };
    if ( $@ ) {
        $Trace->fatal(msg => "Unable to unmap drive '$drive' from client '$input_vars{Client} '$@'");
        $Cleaned_Win32_Mappings = 0;
    } else {
        $Trace->debug(msg => "Successfully unmapped drive '$drive' from client '$input_vars{Client}");
        $Cleaned_Win32_Mappings = 1;
    }

    $Trace->exit();
}

# given a server, delete any mappings to that server
sub Clean_Win32_Mappings {
    $Trace->entry();
    my %input_vars = Params::Validate::validate( @_,
        {
            Client  => { optional => 0, type => SCALAR, default => $Params->{CLIENT} },
            CIFS_Server => { optional => 0, type => SCALAR, default => $Params->{CIFS_SERVER} },
            timeout     => { optional => 1, type => SCALAR, default => 100 },
			force       => { optional => 1, type => SCALAR, default => 0 }
        },
    );
    $Trace->debug(msg => "input_vars = " . Dumper(\%input_vars));
    eval {
        MapShare::unmap_cifs_shares_for_server(
            server  => $input_vars{CIFS_Server}, 
            client  => $input_vars{Client},
            timeout => $input_vars{timeout},
            force   => $input_vars{force},
        );
    };
    if ( $@ ) {
        $Trace->fatal(msg => "Unable to clear all shares for server '$input_vars{CIFS_Server}' from client '$input_vars{Client}' '$@'");
    } else {
        $Trace->trace(msg => "Successfully unmapped all drives mapped to '$input_vars{CIFS_Server}'");
    }

    $Trace->exit();
}

