# There are certain things that everyone of my ActivePerl scripts does.
# I do not assume that we have the SpinTestTool loaded at all.
# vim: set tabstop=4 shiftwidth=4 expandtab shiftround: 

package CIFS::Standard;

use strict;
use warnings;
use Win32;
use File::Spec;
use FindBin;

# Netapp Specific libraries that we pick up from NATELIB
use RandomData;
use TracePkg;

my $Trace = TracePkg->new(separateNewline => 1);

local %CIFS::Standard::ClientInfo;
local %CIFS::Standard::ErrorInfo;

# the buffering in the windows command window sucks, this makes it more like
# a line mode unix buffered style that is much more desirable
# TODO 03/21/07:DPS: I'm not sure this is needed with NATE....
select STDERR ; $|=1;
select STDOUT ; $|=1;


# GatherClientStats()
# This routine grabs all the information that we could need about a client
# and stores it away in a hash that we can use to report back to the user at
# a later time.
sub GatherClientStats {
    $Trace->entry();
    $CIFS::Standard::ClientInfo{OS} = [Win32::GetOSVersion()];
    $CIFS::Standard::ClientInfo{OSName} = Win32::GetOSName;
    $CIFS::Standard::ClientInfo{ActivePerl} = "$] " . Win32::BuildNumber();
    $CIFS::Standard::ClientInfo{CPU} = "$ENV{NUMBER_OF_PROCESSORS} @ $ENV{PROCESSOR_IDENTIFIER}";
    $CIFS::Standard::ClientInfo{UserName} = Win32::LoginName();
    $CIFS::Standard::ClientInfo{MachineName} = Win32::NodeName();

    # these are some things we can only fill when on NT/2000/XP
    if ( Win32::IsWinNT() ) {
        $CIFS::Standard::ClientInfo{MachineDomain}=Win32::DomainName();
    }

    # get some data out of the registry
    use Win32::TieRegistry( Delimiter => "/");
    my $setup = new Win32::TieRegistry "LMachine/SYSTEM/Setup", { Delimiter=>"/", Access=>Win32::TieRegistry::KEY_READ() };
    if ( defined $setup ) {
        $CIFS::Standard::ClientInfo{ImageInfo} = $setup->GetValue("OemDuplicatorString") || "None";
        $CIFS::Standard::ClientInfo{ClonedDate} = $setup->GetValue("CloneTag") || "Not Cloned";
    } else {
        $Trace->warn(msg =>"Unable to create Registry Tie to gather Clone values. '$^E'");
        $CIFS::Standard::ClientInfo{ImageInfo} = "UNKNOWN";
        $CIFS::Standard::ClientInfo{ClonedDate} = "UNKNOWN";
    }
    # we want to explicitly delete anything that we had for the registry so that we aren't taking up that memory
    # and that no one hurts themselves.
    undef($setup);
    no Win32::TieRegistry;

    $Trace->exit();
    return 1;
}

# PrintClientStats
# This routine prints out the client statistics gathered during testing
sub PrintClientStats {
    $Trace->entry();

    $Trace->comment(msg => "Machine Name: $CIFS::Standard::ClientInfo{MachineName}");
    $Trace->comment(msg => "Machine Domain: $CIFS::Standard::ClientInfo{MachineDomain}") if $CIFS::Standard::ClientInfo{MachineDomain};
    $Trace->comment(msg => "Windows Version: $CIFS::Standard::ClientInfo{OSName} $CIFS::Standard::ClientInfo{OS}[1]\.$CIFS::Standard::ClientInfo{OS}[2]\.$CIFS::Standard::ClientInfo{OS}[3] $CIFS::Standard::ClientInfo{OS}[0]");
    $Trace->comment(msg => "CPU: $CIFS::Standard::ClientInfo{CPU}");
    $Trace->comment(msg => "ActivePerl build: $CIFS::Standard::ClientInfo{ActivePerl}");
    $Trace->comment(msg => "Running as user: $CIFS::Standard::ClientInfo{UserName}");
    $Trace->comment(msg => "Clone Info: $CIFS::Standard::ClientInfo{ImageInfo}");
    $Trace->comment(msg => "Image Date: $CIFS::Standard::ClientInfo{ClonedDate}");

    $Trace->exit();
    return 1;
}


# Given a path name, return the maximum available path available for file and directory name creation
# Assume the path given is in a correct format already (IE back slashed correctly, or uses forward slashes, and every
# character is a valid path character)
# returns -1 if the path is too long to add to
sub MaxAvailablePath {
    $Trace->entry();
    my $path = shift;
    my $max_path = 248;  # this is close enough, should be queryiable from Windows though

    my $path_len = length $path;

    if ( $path_len >= $max_path ) {
        $Trace->trace(msg => "path_len = '$path_len' and max_path = '$max_path' so max_available_path is 0");
        $max_path = 0;
    }

    # we need to be able to add at least 2 characters to the path unless the path that was passed in ended in a
    # path delimiting character (like \\ or /)
    if ( $path =~ /[\\|\/]$/ ) {
        if ( ($path_len > ($max_path - 1)) || ($path_len > ($max_path - 2)) ) {
            $Trace->trace(msg => "path_len = '$path_len' and doesn't have enough room to add a delimiter but seems to need one.");
            $max_path = -1;
        } else {
            $max_path = $max_path - $path_len;
        }
    } else {
        $max_path = $max_path - $path_len;
    }
    $Trace->exit(msg => "max_path => $max_path");
    return ($max_path);
}

# Given a filename and an existance check flag, determine if the filename is a valid filename, and possibly
# if it exists already or not.  Works for both directories and files Return pass or fail based on the results 
sub checkFileName {
    $Trace->entry();
    my $file = shift;
    my $can_exist = shift;
    my $valid = 1;

    my ( $volume, $dirname, $filename ) = File::Spec->splitpath($file);
    $Trace->trace(msg => "Checking file '$filename' in directory '$dirname'");
    # first verify the filename is valid
    if ( ($filename =~ /^\\.$/) || ($filename =~/^\.\.$/) ) {
        $Trace->trace(msg => "'.' and '..' are not valid filenames.");
        $valid = 0;
    } elsif ( ($filename =~ /^\s+/) || ($filename =~ /\s+$/) ) {
        $Trace->trace(msg => "Files/Dirs cannot begin or end in whitespace.");
        $valid = 0;
    } elsif ( ($filename =~ /^nul$/i) || ($filename =~ /^prn$/i) || ($filename =~ /^con$/i) || ($filename =~ /^aux$/i) ) {
        $Trace->trace(msg => "Files/Dirs cannot be called 'NUL', 'PRN', 'CON', or 'AUX'.");
        $valid = 0;
    }

    # now check if the file/dir already exists if asked to do so and its a valid name
    # (-e should catch both files and dirs, but I've seen it sneak by sometimes)
    if ( !$can_exist && $valid ) {
        if ( -e $file ) {
            $Trace->trace(msg => "File/Dir '$file' already exists.");
            $valid = 0;
            # we know it exists, let see if its a file or a dir
            if ( -d $file ) {
                $Trace->trace(msg => "Dir '$file' already exists.");
            } elsif ( -f $file ) {
                $Trace->trace(msg => "File '$file' already exists.");
            }
        }
    }

    # return what we think is valid or not
    $Trace->exit(msg => "File '$file' is " . ($valid ? "" : "not " ) . "valid");
    return $valid;
}

# createFileName
# Given a leading directory and some options will return a filename that could be used as a directory or file name
sub createFileName {
    $Trace->entry();
    my %params = (
        parent	    => undef,
        maxsize	    => undef,
        cifsutf8    => 1,
        random	    => 1,
        checkexist  => undef,
        verbose	    => undef,
        debug	    => undef,
        @_,
    );
    my $file = "";

    if ( !defined $params{maxsize} ) { # we need to set the max size if it wasn't passed to us
        $params{maxsize} = MaxAvailablePath($params{parent});
    }

    my $rd = new randomString(size => $params{maxsize}, cifsutf8 => $params{cifsutf8}, randomsize => $params{random});
    my $fileOK = 0;
    until ( $fileOK ) {
        $file = $rd->String;
        if ( $params{checkexist} ) {
            if ( checkFileName("$params{parent}/$file", 0) ) {
                $fileOK = 1;
            }
        } else {
            if ( checkFileName("$params{parent}/$file", 1) ) {
                $fileOK = 1;
            }
        }
    }
    $Trace->exit(msg => "file => '$file'");
    return $file;
}

# createRandomFile
# given a parent directory, create a random file,  in the directory and exit.  if given a file size, fill file to that size, otherwise 
# the file will be left empty.  returns the name of the file created
sub createRandomFile {
    $Trace->entry();
    my %params = (
        parent	=> undef,
        name	=> undef,
        size	=> undef,
        blocksize => 65536,
        cifsutf8 => undef,
        checkexist => 1,
        verbose	=> undef,
        debug	=> undef,
        sparse	=> undef,
        @_);

    if ( !defined $params{parent} ) {
        $Trace->fatal(msg => "Parent directory required for this call.");
    }

    # generate a random name
    if ( !defined $params{name} ) {
        $params{name} = CIFS::Standard::createFileName(parent => $params{parent}, cifsutf8 => $params{cifsutf8}, random => 1, checkexist => $params{checkexist});
    }

    # now create the file
    my $file = "$params{parent}/$params{name}";
    # if the file already exists, and the caller didn't care return
    if ( !$params{checkexist} && -e $file) {
        $Trace->trace(msg => " file '$file' exists and that is ok.");
    } elsif ( -e $file ) {
        $Trace->trace(msg => "  file '$file' already exists and can not.");
        $params{name} = undef; # we are going to be returning this to the caller
    } else {
        my $FH = Win32API::File::createFile($file, "w"); # open for write, share none
        unless ( $FH ) {
            $Trace->comment(msg => "Unable to open file '$file' PathSize = " . length $params{name} . " '$^E'");
            $params{name} = undef; # we are going to be returning this to the caller
        } else {
            if ( defined $params{size} ) {
                if ( defined $params{sparse} ) {
                    my $high = $params{size} / 4096;
                    my $low = ((($params{size} % 4096)*1048576)-1) || 4096*1048576;
                    unless ( defined Win32API::File::SetFilePointer($FH, $low, $high, Win32API::File::FILE_BEGIN()) ) {
                        $Trace->trace(msg => "Could not seek to: '$high' '$low' '$^E'");
                         $params{name} = undef; # we are going to be returning this to the caller
                    } else {
                        unless ( Win32API::File::WriteFile($FH, "a", 0, [], []) ) {
                            $Trace->trace(msg => " could not write at position '$high' '$low' in file '$file' '$^E'");
                            $params{name} = undef; # we are going to be returning this to the caller
                        } 
                    }
                } else {
                    my $rd = new randomString(size => $params{blocksize});
                    my $text = $rd->String;
                    for ( my $x=0; $x<int($params{size}/$params{blocksize}) ; $x++) {
                        if ( !Win32API::File::WriteFile( $FH, $text, 0, [], [] ) ) {
                            $Trace->trace(msg => "Failed to write. '$^E'");
                        }
                    }
                    $text = $rd->String(size => ($params{size} % $params{blocksize}));
                    if ( !Win32API::File::WriteFile( $FH, $text, 0, [], [] ) ) {
                        $Trace->trace(msg => "Failed to write. '$^E'");
                    }
                }
                $Trace->trace(msg => "File '$file' created with size '" . (stat($file))[7] . "' bytes.");
            }
            if ( Win32API::File::CloseHandle($FH) ) {
                unless ( -f $file ) {
                    $Trace->comment(msg => "Unable to lookup file '$file' after creation. '$^E'");
                    $params{name} = undef; # we are going to be returning this to the caller
                }
            } else {
                $Trace->comment(msg => "Unable to close file '$file' after creation. '$^E'");
                $FH = undef; # try really hard to dump this fileHandle
            }
        }
    }
    if ( defined $params{name} ) {
        $Trace->exit(msg => "file => '$params{name}'");
    } else {
        $Trace->exit(msg => "Returning 'undef'");
    }
    return $params{name};
}

# createRandomDir
# given a parent directory, create a random directory in the directory and exit.  
sub createRandomDir {
    $Trace->entry();
    my %params = (
        parent	=> undef,
        name	=> undef,
        cifsutf8 => undef,
        checkexist => 1,
        verbose => 0,
        debug	=> 0,
        @_);

    if ( !defined $params{parent} ) {
        $Trace->fatal(msg => "Parent directory required for this call.");
    }

    # createFileName doesn't realize that dirs need 1 more character that a file (IE the API won't let me create a directory that I can't put files in)
    my $maxsize = MaxAvailablePath($params{parent}) - 1;

    if ( $maxsize == 0 ) {
        $Trace->comment(msg => "Cannot create a directory at this depth.");
        $params{name} = undef;
    } elsif ( !defined $params{name} ) {
        $params{name} = CIFS::Standard::createFileName(parent => $params{parent}, cifsutf8 => $params{cifsutf8}, random => 1, checkexist => $params{checkexist}, maxsize => $maxsize);
    }

    if ( defined $params{name} ) {
        # now create the dir
        my $dir = "$params{parent}/$params{name}";

        # if the dir already exists, and the caller didn't care return
        if ( !$params{checkexist} && -d $dir) {
            $Trace->trace(msg => "Directory '$dir' exists and that is ok.");
        } elsif ( !mkdir $dir ) {
            $Trace->comment(msg => "Unable to create directory '$dir' '$^E'");
            $params{name} = undef;
        } else {
            unless ( -d $dir ) {
                $Trace->comment(msg => "Unable to lookup dir '$dir' after creation. '$^E'");
                $params{name} = undef;
            }
        }
    }
    if ( defined $params{name} ) {
        $Trace->exit(msg => "FileName = '$params{name}'");
    } else {
        $Trace->exit(msg => "Returning 'undef'");
    }
    return $params{name};
}


# delete an entire subdirectory tree removing everything in its path
# currently we just call the rmtree command to do this.  eventually we will allow you to pass in what style to delete
# where we can chose multiple different styles of delete code.
sub deltree {
    $Trace->entry();
    my $path = shift;
    my $style = shift || 0;
    my $rtn = -1;

    unless ( !-d $path ) { 
        $Trace->trace(msg => "Deleting directory tree at '$path'");
        if ( ($style == 1) || ($^V gt 5.8.1) ) { # we can't use the version of File::Path that comes with 5.8.2 and up
            $Trace->trace(msg => "Using private deltree for perl '$]' rather than File::Path::rmtree");
            if ( !-d $path ) {
                unless ( unlink $path ) {
                    $Trace->comment(msg => "Unable to remove '$path' '$^E'");
                } else {
                    $rtn = 1;
                }
            } else {
                opendir(DIR, $path);
                foreach my $file ( readdir(DIR) ) {
                    next if ( $file =~ /^\.\.?$/ );
                    if ( !-d "$path/$file" ) {
                        unless ( unlink "$path/$file" ) {
                            $Trace->comment(msg => "Unable to remove '$path/$file' '$^E'");
                        } else {
                            $rtn = 1;
                        }
                    } else {
                        if ( CIFS::Standard::deltree("$path/$file", 1) < 0 ) {
                            $Trace->comment(msg => "Failed to remove a file '$path/$file' so I cannot continue.");
                        } else {
                            $rtn = 1;
                        }
                    }
                }
                if ( rmdir "$path" ) {
                    $rtn = 1;
                } else {
                    $Trace->comment(msg => "Delete of $path root failed. Reason:$!");
                }
            }
        } else {
            $Trace->comment(msg => "Using File::Path::rmtree for perl '$]'");
            $path =~ s/\//\\/g; # we need \ not / as the dir seperator to call to rmtree because it seems to be able to deal with that better
            if (File::Path::rmtree($path, 0, 0) > 0) {
                $rtn = 1;
            } else {
                $Trace->comment(msg => "Failed to remove '$path'.");
            }
        }
    } else {
        $Trace->trace(msg => "There doesn't seem to be a directory tree at '$path'");
        $rtn = -1; # there was nothing to delete	
    }
    $Trace->exit(msg => "Returning '$rtn'");
    return $rtn;
}

# There is a tool dir for windows that we may want to use outside of NATE (IE without subtest).
# Walk down through @INC and see if we can't find this tool dir, and optionally a specific tool
sub find_windows_tools {
    $Trace->entry();
    my $tool = shift || "tools/windows";
    my $tool_dir = undef;
    foreach my $dir ( @INC ) {
        if ( -e "$dir/$tool" ) {
            $tool_dir = "$dir/$tool";
        }
    }
    if ( defined $tool_dir ) {
        $Trace->exit(msg => "Returning '$tool_dir'");
    } else {
        $Trace->exit(msg => "Returning 'undef'");
    }
    return $tool_dir;
}

1;
