# $Id: //depot/prod/test/Rsierranevada/lib/nblade/ExportVerifier.pm#1 $

## @summary Routines used to verify client access to exports
## @author snider@netapp.com
## @status Shared
## @pod here

package nblade::ExportVerifier;

use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR UNDEF);
use NFSClient::Linux;
use strict;
use English;
use TracePkg;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
		 setup
		 verify
		 cleanup
                 );
our @EXPORT_OK = qw(
		    setup
		    verify
		    cleanup
                    );

my $Trace = new TracePkg;
my $data =  "Hello, World!";
my $myfile = "myfile";
my %nfs_versions = (
		    2 => {
			client => new NFSClient::Linux(verbosity=>2, name => "root", version=>2),
			mount_fh => undef,
			file_handle => undef,
		    },
		    3 => {
			client => new NFSClient::Linux(verbosity=>2, name => "root", version=>3),
			mount_fh => undef,
			file_handle => undef,
		    },
		    );
my $ok_to_exist = 1;

sub setup {
    $Trace->entry();
    my %params = validate_mixed(
				@_,
				[],
				{
				    export=> { type => SCALAR },
				}
				);
    my $setup_write_status = "";
    my $setup_mount_status = "";
    my $file_handle ;
    my $mount_fh = "";
    
    foreach my $version (keys %nfs_versions) {
	my $client = $nfs_versions{$version}->{client};
	my $uid = $client->{nfs}->{uid};
	$client->{nfs}->{uid} = 0;
	$mount_fh = $client->mount(export=> $params{export});
	if (! $mount_fh) {
	    $Trace->fatal(msg=>"mount failed when it should have succeeded");
	}
	eval{
	    my %stat = $client->stat(handle=> $mount_fh);
	    if ($stat{mode} != 0777) {
		$client->chmod(handle=> $mount_fh, file_handle=> $mount_fh, mode=> "0777");
	    }
	    $file_handle = $client->create(ok_to_exist=> $ok_to_exist, handle=> $mount_fh, file=> $myfile);
	    $client->chmod(handle=> $mount_fh, file_handle=> $file_handle, mode=> "0777");
	    $client->write(handle=> $mount_fh, data=> $data, file_handle=> $file_handle);
	};
	if ($@) {
	    $Trace->fatal(msg=>"write failed when it should have succeeded: $@");
	}
	$nfs_versions{$version}->{mount_fh} = $mount_fh;
	$nfs_versions{$version}->{file_handle} = $file_handle;
	$client->{nfs}->{uid} = $uid;
    }
    $Trace->exit();
}

sub verify {
    $Trace->entry();
    my @status;
    my $code;
    my %params = validate_mixed(
				@_,
				[],
				{
				 export=> { type => SCALAR },
				 read => { type => SCALAR | UNDEF, default => 0, optional => 1 },
				 write => { type => SCALAR | UNDEF, default => 0, optional => 1 },
				 mount => { type => SCALAR | UNDEF, default => 0, optional => 1 },
				 client_ip => { type => SCALAR | UNDEF, optional => 1 },
				 nfsv2 => { type => SCALAR, optional => 1 },
				 nfsv3 => { type => SCALAR, optional => 1 },
				 auth => { type => SCALAR | UNDEF, default => "sys", optional => 1 },
				 user => { type => SCALAR | UNDEF, default => 0, optional => 1 },
				 auth_user => { type => SCALAR | UNDEF, default => 0, optional => 1 },
				}
				);

    if (!$params{nfsv2} && !$params{nfsv3}) {
	$Trace->fatal(msg=>"Either 'nfsv2' or 'nfsv3' must be set");
    }

    my $allowed_attempts = 5;
    my $attempts;
    my $version;

    if ($params{nfsv2}) {
	$version = 2;
    }
    if ($params{nfsv3}) {
	$version = 3;
    }
    my $client = $nfs_versions{$version}->{client};
    my $mount_fh = $nfs_versions{$version}->{mount_fh};
    my $file_handle = $nfs_versions{$version}->{file_handle};

    $client->set_uid(uid=>0);

    eval {
	my $mount_handle = $client->mount(export=> $params{export}, src_ip => $params{client_ip});
    };
    $code = $client->nfs_code();
    $attempts=$client->attempts();
    if ( defined($code) && $params{mount} != $code ) {
	push(@status, "mount failed because the server returned status code $code rather than $params{mount}");
    }
    if ( $attempts  > $allowed_attempts) {
	push(@status, "mount took $attempts attempts to finally respond");
    }

    $client->set_uid(uid=>$params{auth_user});

    eval{
	my $myfile = "write_file";
        $client->write(handle=> $mount_fh, data=> $data, file=>$myfile, src_ip => $params{client_ip});
	$client->remove(handle=>$mount_fh, file=> $myfile);
    };
    $code = $client->nfs_code();
    $attempts=$client->attempts();
    my %attrs = $client->get_attributes();
    if ( defined($code) && $params{write} != $code ) {
	push(@status, "write failed because the server returned status code $code rather than $params{write}");
    }
    if ( $attempts  > $allowed_attempts) {
	push(@status, "write took $attempts attempts to finally respond");
    }
    if ( $params{user} != 0 && $params{user} != $attrs{uid}) {
	push(@status, "the created file was owned by '$attrs{uid}' rather than '$params{user}'");
    }

    eval{
        $client->read(handle=> $mount_fh, file_handle=>$file_handle, src_ip => $params{client_ip});
    };
    $code = $client->nfs_code();
    $attempts=$client->attempts();
    if ( defined($code) && $params{read} != $code ) {
	push(@status, "read failed because the server returned status code $code rather than $params{read}");
    }
    if ( $attempts  > $allowed_attempts) {
	push(@status, "read took $attempts attempts to finally respond");
    }

    if (scalar @status) {
	$Trace->fatal(msg=>join("\n", @status));
    }
    $Trace->exit();
}

sub cleanup {
    $Trace->entry();
    my %params = validate_mixed(
				@_,
				[],
				{
				    export=> { type => SCALAR },
				}
				);


    my $client = $nfs_versions{3}->{client};

    $client->{uid} = 0;
    my $mount_fh = $nfs_versions{3}->{mount_fh};
    $client->remove(handle=>$mount_fh, file=> $myfile);
    foreach my $version (keys %nfs_versions) {
	my $client = $nfs_versions{$version}->{client};
	$client->{uid} = 0;
	$client->umount(export=> $params{export});
    }
    $Trace->exit();
}

1;
