# $Id: //depot/prod/test/Rsierranevada/lib/nblade/FileHandle.pm#1 $

## @summary Base class for an NFS filehandle object
## @author snider@netapp.com
## @status Shared
## @pod here

=head1 NAME

nblade::FileHandle - Base class for an NFS filehandle object

=head1 SYNOPSIS

    use nblade::FileHandle;
    my $fho = new nblade::FileHandle(filehandle=>$handle);
    $fho->print();
    my $msid = $fho->local_msid();

=head1 DESCRIPTION

C<nblade::FileHandle> implements an object which is used to manipulate
an NFS filehandle.

=back

=head1 METHODS

=cut

package nblade::FileHandle;
use strict;
use Data::Dumper;
use TracePkg;

use overload q("") => \&_print;

my $Trace = new TracePkg;

my $bitmask = {
               NULL => 0x0,        #SPINNP_FH_FLAG_NULL
               STREAMDIR => 0x1,   #SPINNP_FH_FLAG_STREAMDIR
               SNAPDIR => 0x2,     #SPINNP_FH_FLAG_SNAPDIR
               SNAPDIR_ENT => 0x4, #SPINNP_FH_FLAG_SNAPDIR_ENT
               EMPTY => 0x8,       #SPINNP_FH_FLAG_EMPTY
               STRIPED => 0x10,    #SPINNP_FH_FLAG_STRIPED
               ONTAPC => 0x20,     #SPINNP_FH_FLAG_ONTAPC
              };


my @Definition = (
                  {
                   name => "utility", length => 1, 
                   bitmask => {
                               tree_w => 0x80,          # /.admin path
                               broken_junction => 0x40, # VLDB error
                              },
                   mask => {
                            version => 0x3f,         # the version of the filehandle
                           },
                  },
                  {
                   # how many volumes have been traversed by the client
                   name => "vol_cnt", length => 1,
                  },
                  {
                   # the Coral epoch
                   name => "epoch", length => 2,
                  },
                  {
                   parent => "spinfh_w",
                   children => [
                                {
                                 name => "msid", length => 4,
                                },
                                {
                                 name => "cluster_id", length => 2,
                                },
                                {
                                 name => "reserved", length => 1,
                                },
                                {
                                 name => "flags", length => 1,
                                 bitmask => $bitmask,
                                },
                                {
                                 name => "fileid", length => 4,
                                },
                                {
                                 name => "uniq", length => 4,
                                },
                               ]
                  },
                  {
                   parent => "spinfh_mntpnt",
                   children => [
                                {
                                 name => "msid", length => 4,
                                },
                                {
                                 name => "cluster_id", length => 2,
                                },
                                {
                                 name => "reserved", length => 1,
                                },
                                {
                                 name => "flags", length => 1,
                                 bitmask => $bitmask,
                                },
                                {
                                 name => "fileid", length => 4,
                                },
                                {
                                 name => "uniq", length => 4,
                                },
                               ]
                  },
                  {
                   # export point
                   name => "fileid_exppt", length => 4,
                  },
                  {
                   # export point
                   name => "uniq_exppt", length => 4,
                  },
                 );
=head2 new

   my $fho = new nblade::FileHandle(filehandle=>$handle);

B<new> will create and return a filehandle object. If the C<filehandle> is
specified, then it will go ahead and parse the handle and store it internally.

The following named arguments are accepted:

=over

=item filehandle

(Optional) : the filehandle string. The string must be a 44-character 
representation of the hexadecimal filehandle.

=back

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
                filehandle => '',
                @_,
                };
    bless($self, $class);
    
    $Trace->entry(msg=>"filehandle=$self->{filehandle}");

    if ($self->{filehandle}) {
        $self->filehandle(filehandle=>$self->{filehandle});
    }

    $Trace->exit(msg=>"return $self");
    return $self;
} # end new


sub filehandle {
    my $self = shift;
    my %params = @_;
    my $object_name;
    $Trace->entry(msg=>"filehandle=$params{filehandle}");

    if ($params{object_name}) {
	$object_name = $params{object_name};
    } else {
	$object_name = 'FileHandle';  # this is the default
    }

    # clear FileHandle hash if we are building a new one
    if ($object_name eq 'FileHandle') {
	$Trace->debug(msg=>"DEBUG: CLEARING FileHandle!!!");
	$self->{FileHandle} = undef;
    }
    my $fh = $params{filehandle};

    if (length($fh) != 88) {
        $Trace->fatal(msg=>"The filehandle '$fh' is not a valid NFSv3 filehandle");
    }

    my $offset = 0;
    $self->{max_name_length} = 0;
    foreach my $field (@Definition) {
        if ($field->{parent}) {
            foreach my $child (@{$field->{children}}) {
		$self->_process_filehandle($child, \$fh, \$offset, $object_name, $field->{parent});
            }
        } else {
	    $self->_process_filehandle($field, \$fh, \$offset, $object_name);
        }
    }
    $Trace->debug(msg=>"@@@@@@@@@@@@@@@@@@@");
    $Trace->debug(msg=>"\$self->{$object_name}: ".Dumper(\$self->{$object_name}) );
    $Trace->debug(msg=>"@@@@@@@@@@@@@@@@@@@");
    $Trace->exit();
} # end filehandle

sub compare {
    my $self = shift;
    my %params = @_;
    my @object_names;
    my $fh;
    my @differ;
    my @maydiffer = split(",", $params{maydiffer});
    my @should_match;
    $Trace->entry();
    $Trace->debug(msg=>"DEBUG: filehandle=$params{filehandle}");
    $Trace->debug(msg=>"DEBUG: filehandle=$params{filehandle2compare}");

    # load the FileHandle2compare object and the FileHandle if it isn't already
    if ($params{filehandle2compare}) {
	$self->{FileHandle2compare} = undef;
	if ($self->{FileHandle} == undef) {
	    #TODO: need better checking here. This is not catching a "no FileHandle" state
	    $Trace->fatal(msg=>"Filehandle object must be pre-loaded or provided before using the compare method!");
	}
	push (@object_names,'FileHandle2compare');
    } 
    if ($params{filehandle}) {
	$self->{FileHandle} = undef;
	push (@object_names,'FileHandle');
    }

    # process filehandle and/or filehandle2compare
    foreach my $object_name (@object_names) {
	# set $fh to the appropriate filehandle
	if ($object_name eq 'FileHandle2compare') {
	    $fh = $params{filehandle2compare};
	} else {
	    $fh = $params{filehandle};
	}
	#$Trace->debug(msg=>"DEBUG: Prepairing to process $object_name; FH:\n$fh"); 	
	# call filehandle sub with the proper fh and object name
	$self->filehandle(filehandle=>$fh, object_name=>$object_name);
    }

    $Trace->debug(msg=>"***************");
    $Trace->debug(msg=>"\$self->{FileHandle}: ".Dumper(\$self->{FileHandle}) );
    $Trace->debug(msg=>"***************");
    $Trace->debug(msg=>"---------------");
    $Trace->debug(msg=>"\$self->{FileHandle2compare}: ".Dumper(\$self->{FileHandle2compare}) );
    $Trace->debug(msg=>"---------------");
    
    ### do the compare
    
    # compare each field of the objects and find any which differ
    foreach my $field (@Definition) {
	if ($field->{parent}) {
	    my $parent = $field->{parent};
	    #$Trace->debug(msg=>"DEBUG: \$parent is: $parent");
	    foreach my $child (@{$field->{children}}) {
		#$Trace->debug(msg=>"DEBUG: Checking ".$child->{name}."...");
		if ($self->{FileHandle}->{$parent}{$child->{name}} ne 
		    $self->{FileHandle2compare}->{$parent}{$child->{name}}) {
		    #$Trace->debug(msg=>"DEBUG: !!! ".$self->{FileHandle}->{$parent}{$child->{name}}." does not match ".$self->{FileHandle2compare}->{$parent}{$child->{name}}." !!!");
		    push (@differ, $parent."_".$child->{name});
		}
	
	    }
	} else {
	    #$Trace->debug(msg=>"DEBUG:Checking FIELD: ".$field->{name}."...");
	    my $parent = $field->{parent};
	    if ($self->{FileHandle}->{$parent}{$field->{name}} ne 
		$self->{FileHandle2compare}->{$parent}{$field->{name}}) {
		#$Trace->debug(msg=>"DEBUG: !!! ".$self->{FileHandle}->{$parent}{$field->{name}}." does not match ".$self->{FileHandle2compare}->{$parent}{$field->{name}}." !!!");
		push (@differ, $field->{name});
	    }

	}
    }

    # return the field name of and differing fields unless they are passed
    # in as values to maydiffer
    foreach my $diffield (@differ) {
	if (!grep($diffield =~ /^$_/, @maydiffer)) {
	    push (@should_match, $diffield);
	} else { 
	    $Trace->comment(msg=>"Check of $diffield has been overridden.");
	}

    }

    # clean up; remove the filehandle2compare hash from the object
    $self->{FileHandle2compare} = undef;
    
    # return an array ref of the different fields
    return \@should_match;

    $Trace->exit();
} # end compare

=head2 print

   $fho->print();

B<print> will print out all of the fields of the parsed filehandle.

=cut

sub print {
    my $self = shift;
    print $self;
}

sub _process_filehandle {
    my ($self, $field, $fh, $offset, $object_name, $parent) = @_;
    #$Trace->debug(msg=>"DEBUG: _process input:\n\$field: $field\n\$fh: $$fh\n\$offset: $$offset\n\$object_name: $object_name\n$parent: $parent\n");
    my $value = substr($$fh, $$offset, $field->{length}*2);
    #$Trace->debug(msg=>"DEBUG: The value of ".$field->{name}." is: $value");
    my $lvalue = "";
    while (length $value) {
        $lvalue .= substr($value, -2, 2, '');
    }
    #$Trace->debug(msg=>"DEBUG: The lvalue of ".$field->{name}." is: 0x$lvalue");
    $$offset += $field->{length}*2;
    my %mask = %{$field->{bitmask}} if ($field->{bitmask});
    if ($field->{mask}) {
        while (my ($key, $value) = each %{$field->{mask}}) {
            $mask{$key} = $value;
	    #$Trace->debug(msg=>"DEBUG: The mask of ".$field->{name}."'s $key is: $value");
        }
    }
    if (keys %mask) {
        while (my ($subfield, $mask) = each %mask) {
            $self->{max_name_length} = length($subfield) if (length($subfield) > $self->{max_name_length});
	    #$self->{FileHandle}{$parent}{$subfield} = hex($lvalue) & $mask;
            $self->{$object_name}{$parent}{$subfield} = hex($lvalue) & $mask;
	    #$Trace->debug(msg=>"DEBUG: The value of ".$field->{name}."'s $subfield is: ".$self->{$object_name}{$parent}{$subfield});

        }
    } else {
        $self->{max_name_length} = length($field->{name}) if (length($field->{name}) > $self->{max_name_length});
    }
    $self->{$object_name}->{$parent}{$field->{name}} = $lvalue;
    #$Trace->debug(msg=>"The value of ".$field->{name}." is: 0x".$self->{$object_name}->{$parent}{$field->{name}});

} # end _process_filehandle

sub _print {
    my $self = shift;

    my $result;
    foreach my $field (@Definition) {
        if ($field->{parent}) {
            $result .= sprintf("%$self->{max_name_length}s:\n", $field->{parent});
            foreach my $child (@{$field->{children}}) {
                $result .= $self->_print_field($child, $field->{parent});
            }
        } else {
            $result .= $self->_print_field($field);
        }
    }
    return $result;
}

sub _print_field {
    my ($self, $field, $parent) = @_;
    my $header = ($parent) ? "\t" : "";
    my $result = sprintf("$header%$self->{max_name_length}s %+8s %12d", 
                         $field->{name}, $self->{FileHandle}->{$parent}{$field->{name}}, 
                         hex($self->{FileHandle}->{$parent}{$field->{name}}));
    if ($field->{bitmask}) {
        $result .= " : ";
        foreach my $subfield (keys %{$field->{bitmask}}) {
            if ($self->{FileHandle}->{$parent}{$subfield}) {
                $result .= "$subfield ";
            }
        }
    } 
    if ($field->{mask}) {
        $result .= "\n";
        foreach my $subfield (keys %{$field->{mask}}) {
            $result .= sprintf("$header\t%$self->{max_name_length}s  %-12d %8s\n", 
                               $subfield, hex($self->{FileHandle}->{$parent}{$subfield}), 
                               $self->{FileHandle}->{$parent}{$subfield});
        }
    }
    return $result."\n";
}

sub AUTOLOAD {
    my $method = our $AUTOLOAD;
    $method =~ s/.*:://;

    my $function;
#    if (exists($self->{FileHandle}->{$method})) {
#        $function = sub { return $self->{FileHandle}->{$method} };
#    } else {
        print "Method '$method' is not defined\n";
        $function = sub { $Trace->fatal(result=>"SCRIPT",
                                        msg=>"method '$method' not defined\n") };
#    }
    goto &$function;
}

sub DESTROY {}

=head1 PARAMS 

=head1 SEE ALSO

=head1 BUGS

=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/nblade/FileHandle.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
