## @summary     NATE test that verifies the NLMv3 Lock procedure
##
## @description Make Lock calls with different values
##
## @param       nbladeIP Required. The IP address of the server.
##
## @synopsis    ntest nbladeIP=<IP> Lock.thpl
##
## @example     ntest nbladeIP=10.98.200.200 nblade/nlm3/Lock.thpl
##
## @status      public
##
## @attribute   listtestcases=yes
##
## @notes       This test will execute a list of testcases
##
## @author      snider@netapp.com

package nblade::nlm::Lock;

use strict 'vars';
use nblade::nfsErrors;
use RandomData;
use TracePkg;
use Exporter;

use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);
@EXPORT = qw(NlmLockTests testAndPrintStatus);

my ($nfs, $mountFH);
my $nlm;

my ($filehandle, $dirhandle);
my $testcase;
my $Trace = new TracePkg();

sub testAndPrintStatus {
    my ($cookie, $svid, $options) = @_;
    $nlm->Test("Quiet$testcase", "Test", "--filehandle $filehandle --exclusive 1 $options", 1);
    if ($nlm->{returnStatus}) {
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid $options", 0);
	$nlm->printResult(1, !$nlm->{error}, $testcase, 1);
    } else {
	$nlm->printResult(1, 0, $testcase, 1);
    }
}

sub blockingTest {
    my $owner = shift;
    my $blockingArgs = shift;

    my $owned_by;

    # acquire, client 1
    $nlm->Test("Quiet$testcase", 'Lock',   "--filehandle $filehandle --locksvid 1 --exclusive 1", 0);
    # request and block, client 2
    $nlm->Test("Quiet$testcase", 'Lock',   "--filehandle $filehandle --locksvid 2 --exclusive 1 --block 1 --lockowner $owner $blockingArgs", 3);
    # release, client 1
    $nlm->Test("Quiet$testcase", 'Unlock', "--filehandle $filehandle --locksvid 1", 0);
    # verify client 1 does not hold it, and that client 2 does
    $nlm->Test("Quiet$testcase", 'Test',   "--filehandle $filehandle --locksvid 1 --exclusive 1", 1);
    ($owned_by) = ($nlm->{outputFromCmd} =~ m/Owner\s*=\s*(.*?)$/sm);

    if ($owned_by eq $owner) {
	# success
	$nlm->printResult(1, !$nlm->{error}, $testcase, 1);
    } else {
	$nlm->printResult(1, 0, $testcase, 1);
    }

    # release, client 2
    $nlm->Test("Quiet$testcase", 'Unlock', "--filehandle $filehandle --locksvid 2 --lockowner $owner $blockingArgs", 0);

}

sub NlmLockTests {
    $nfs	= shift;
    $mountFH	= shift;
    $nlm	= shift;

    my ($data, $string);
    my $dir = "lock$$.".int(rand(0xff));

    my ($maxOffset, $longOffsetError, $wrappedOffsetError, $maxQuiet);
    my $skip = 0;
    my $platform = $nlm->{Platform};
    my $is_7mode = 0;

    $Trace->comment(msg => "Testing against $platform platform");
    if ($platform =~ /ontap/i) {
	$nlm->ConnectToServer();
        my $release = $nlm->{serverobject}->get_release(format=>"cooked");
        $Trace->comment(msg => "Testing against a $release build.");
	if ($nlm->{'version'} >= 4 && $release eq 'Tricky.0') {
	    $skip = 1;
	}
	if ($release =~ /^7\./) {
	    $is_7mode = 1;
	} elsif ($release =~ /^8\./) {
	    $is_7mode = $nfs->{is_7mode};
	}
    }

    if ($is_7mode) {
	$Trace->comment(msg => "Using 7-mode return values");
    } else {
	$Trace->comment(msg => "Using 10-mode return values");
    }

    if ($nlm->{'version'} < 4) {
	$maxOffset = '0xffffffff';
	$longOffsetError = 0;
	$wrappedOffsetError = 0;
    } else {
	$maxOffset = '0xffffffffffffffff';
	# 7-mode does not return an error when the lock offset is equal to -1
	# 7-mode/GX return different error codes if the sum of the offset and length are too big
	if ($is_7mode) {
	    $longOffsetError = 0;
	    $wrappedOffsetError = 8; # NLM_BIG
	} else {
	    $longOffsetError = 9;
	    $wrappedOffsetError = 9;
	}
    }

    if (($dirhandle = $nfs->Test("ALWAYS", "Mkdir", "--mkdirFH $mountFH --dirname $dir --filemode 0040777", 0)) eq "") {
	$dirhandle = $nfs->Test("ALWAYS", "Lookup", "--filename $dir --dirhandle $mountFH", 0);
    }
    if (($filehandle = $nfs->Test("ALWAYS", "Create", "--createFH $dirhandle --createfilename file --filemode 0100666 0", 0)) eq "") {
	$filehandle = $nfs->Test("ALWAYS", "Lookup", "--filename file --dirhandle $dirhandle", 0);
    }
    if ($filehandle eq "" || $dirhandle eq "") {
	die "Could not setup test\n";
    }

    my $size = ($nfs->{fileLength}) ? $nfs->{fileLength} : int(rand(1024))+1;
    $nfs->VPrint(2, "filesize(size) set to $size\n");

# Test 0: OK
    my $cookie = int(rand(0xffffffff)) + 1;
    my $svid = int(rand(0xffffffff)) + 1;
    if ($nlm->checkForTest(0)) {
	$testcase = "0a";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random'");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 0);
	testAndPrintStatus($cookie, $svid);

	$testcase = "0b";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie \"\" --locksvid $svid", 0);
	testAndPrintStatus("\"\"", $svid);

	$testcase = "0c";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random hex string'");
	$data = randomString->new(size=>$size*2, hex => 1);
	$string = $data->String();
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0);
	testAndPrintStatus($string, $svid);

	$testcase = "0d";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='1K all 0s'");
	$string = '0'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0);
	testAndPrintStatus($string, $svid);

	$testcase = "0e";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='1K all 1s'");
	$string = 'f'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0);
	testAndPrintStatus($cookie, $svid);

	$testcase = "0f";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' exclusive=0xffffffff block=0xffffffff reclaim=0xffffffff");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --exclusive 0xffffffff --block 0xffffffff --reclaim 0xffffffff", 0);
	testAndPrintStatus($cookie, $svid);

	$testcase = "0g";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' state=0");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --state 0", 0);
	testAndPrintStatus($cookie, $svid);

	$testcase = "0h";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' state=0xffffffff");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --state 0xffffffff", 0);
	testAndPrintStatus($cookie, $svid);

	$testcase = "0i";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' client=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --clientname \"\"", 0);
	testAndPrintStatus($cookie, $svid, "--clientname \"\"");

	$testcase = "0j";
	$data = randomString->new(size=>$size);
	$string = $data->String();
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' client='random'");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --clientname $string", 0);
	testAndPrintStatus($cookie, $svid, "--clientname $string");

	$testcase = "0k";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' owner=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner \"\"", 0);
	testAndPrintStatus($cookie, $svid, "--lockowner \"\"");

	$testcase = "0l";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' owner='random'");
	$data = randomString->new(size=>$size*2, hex => 1);
	$string = $data->String();
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0);
	testAndPrintStatus($cookie, $svid, "--lockowner $string");

	$testcase = "0m";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' owner='1K all 0s'");
	$string = '0'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0);
	testAndPrintStatus($cookie, $svid, "--lockowner $string");

	$testcase = "0n";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' owner='1K all 1s'");
	$string = 'f'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0);
	testAndPrintStatus($cookie, $svid, "--lockowner $string");

	unless ($skip) {
	    $testcase = "0o";
	    if (!$longOffsetError) {
		$maxQuiet = 'Quiet';
	    } else {
		$maxQuiet = '';
	    }
	    $nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' offset=$maxOffset");
	    $nlm->Test("$maxQuiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockoffset $maxOffset", $longOffsetError);
	    if ($maxQuiet) {
		testAndPrintStatus($cookie, $svid, "--lockoffset $maxOffset");
	    }
	}

	$testcase = "0p";
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' length=$maxOffset");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --locklength $maxOffset", 0);
	testAndPrintStatus($cookie, $svid, "--locklength $maxOffset");

	unless ($skip) {
	    $testcase = "0q";
	    if (!$wrappedOffsetError) {
		$maxQuiet = 'Quiet';
	    } else {
		$maxQuiet = '';
	    }
	    $nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' offset=$maxOffset length=$maxOffset");
	    $nlm->Test("$maxQuiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockoffset $maxOffset --locklength $maxOffset", $wrappedOffsetError);
	    if ($maxQuiet) {
		testAndPrintStatus($cookie, $svid, "--lockoffset $maxOffset --locklength $maxOffset");
	    }
	}

	$testcase = "0r";
	$string = 'f'x1024;
	$nlm->AddTestcase($testcase, "lock fh locksvid='random' cookie='random' client='512-character name'");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --clientname $string", 0);
	testAndPrintStatus($cookie, $svid, "--clientname $string");
    }

    if ($nlm->checkForTest(7)) {
	# There are behavioral differences between 7-mode and GX:
	# When given a random FH of length < sizeof(nfs_File_handle) (32)
	#   GX returns an RPC error (code=7)
	#   7-mode returns an NLM error (code=1)
	# When given a random FH of length >= sizeof(nfs_File_handle) (32) and <= MAX_NETOBJ_SZ (1024)
	#   GX returns an RPC error (code=7)
	#   7-mode accepts the request
	# When given a random FH of length > MAX_NETOBJ_SZ (1024)
	#   GX returns an RPC error (code=7)
	#   7-mode returns an RPC error (code=4)
	#
	# For test cases that might pass (currently 7d/7e), we need to supply the svid so
	# the owner matches in the unlock request.
	my @short_garbage_code = (7, "Garbage Arguments");
	my @medium_garbage_code = (7, "Garbage Arguments");
	my @long_garbage_code = (7, "Garbage Arguments");
	if ($is_7mode) {
	    @short_garbage_code = (1);
	    @medium_garbage_code = (0);
	    @long_garbage_code = (4, "Garbage Arguments");
	}

	$testcase = "7a";
	$nlm->AddTestcase($testcase, "lock fh=''");
	$nlm->Test($testcase, "Lock", "--filehandle \"\"", @short_garbage_code);

	$testcase = "7b";
	$nlm->AddTestcase($testcase, "lock fh='1 byte string'");
	$data = randomString->new(size=>2, hex => 1); # 1 byte
	$string = $data->String();
	$nlm->Test($testcase, "Lock", "--filehandle $string", @short_garbage_code);

	$testcase = "7c";
	$nlm->AddTestcase($testcase, "lock fh='12 byte string'");
	$data = randomString->new(size=>24, hex => 1); # 12 byte
	$string = $data->String();
	$nlm->Test($testcase, "Lock", "--filehandle $string", @short_garbage_code);

	$testcase = "7d";
	$nlm->AddTestcase($testcase, "lock fh='32 byte string'");
	$data = randomString->new(size=>64, hex => 1); # 32 byte
	$string = $data->String();
	$nlm->Test($testcase, "Lock", "--filehandle $string --locksvid $svid", @medium_garbage_code);
	if ($is_7mode) {
	    # since this passes in 7-mode, we need to send unlock
	    $nlm->Test("Quiet$testcase", "Unlock", "--filehandle $string --locksvid $svid", 0);
	}

	$testcase = "7e";
	$nlm->AddTestcase($testcase, "lock fh='128 byte string'");
	$data = randomString->new(size=>256, hex => 1); # 128 byte
	$string = $data->String();
	$nlm->Test($testcase, "Lock", "--filehandle $string --locksvid $svid", @medium_garbage_code);
	if ($is_7mode) {
	    # since this passes in 7-mode, we need to send unlock
	    $nlm->Test("Quiet$testcase", "Unlock", "--filehandle $string --locksvid $svid", 0);
	}

	$testcase = "7f";
	$nlm->AddTestcase($testcase, "lock fh='12 bytes all 1'");
	$string = 'f'x24;
	$nlm->Test($testcase, "Lock", "--filehandle $string", @short_garbage_code);

	$testcase = "7mode";
	$nlm->AddTestcase($testcase, "lock fh='12 bytes all 0'");
	$string = '0'x24;
	$nlm->Test($testcase, "Lock", "--filehandle $string", @short_garbage_code);

	$testcase = "7h";
	$nlm->AddTestcase($testcase, "lock fh='1025 byte hex string'");
	$data = randomString->new(size=>2050, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Lock", "--filehandle $string", @long_garbage_code);

	$testcase = "7i";
	$nlm->AddTestcase($testcase, "lock fh name='512 bytes' namelength=5");
	$string = 'f'x1024;
	$nlm->Test($testcase, "Lock", "--filehandle $filehandle --clientname $string --namelength 5", @long_garbage_code);

	$testcase = "7j";
	$nlm->AddTestcase($testcase, "lock fh name='512 bytes' namelength=0xffffffff");
	$string = 'f'x1024;
	$nlm->Test($testcase, "Lock", "--filehandle $filehandle --clientname $string --namelength 0xffffffff", @long_garbage_code);
    }

    # These tests are heavily dependent on client behavior
    # Many of our clients will reject the NLM_GRANTED sent by the filer
    # On 7-mode, this causes the lock to be released and the subsequent TEST is granted
    # This results in a test failure even though the filer behavior is correct
    # Disable these tests for 7-mode so QA can use the script to easily verify other bug fixes
    if ($nlm->checkForTest(8) && !$is_7mode) {
	$testcase = '8a';
	$string = 'a'x10;
	$nlm->AddTestcase($testcase, 'lock block=1');
	blockingTest($string, "");
	
	$testcase = '8b';
	$string = 'a'x1024;
	$nlm->AddTestcase($testcase, 'lock block=1 max clientname, lockowner, cookie, ');
	blockingTest("$string$string", "--clientname $string --cookie $string$string");
    }

    $testcase = "9";
    $nlm->AddTestcase($testcase, "lock fh [file was just removed]");
    $nfs->Test("ALWAYS", "Remove", "--removeFH $dirhandle --removefilename file", 0);
    $nlm->Test($testcase, "Lock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 7, 0);
    $nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 0);

    $nfs->Test("ALWAYS", "Rmdir", "--rmdirFH $mountFH --rmdirfilename $dir", 0);

    return $nlm;
}

# return value, so perl doesn't have a cow
1;
