## @summary     NATE test that verifies the NLMv3 Test procedure
##
## @description Make Test calls with different values
##
## @param       nbladeIP Required. The IP address of the server.
##
## @synopsis    ntest nbladeIP=<IP> Test.thpl
##
## @example     ntest nbladeIP=10.98.200.200 nblade/nlm3/Test.thpl
##
## @status      public
##
## @attribute   listtestcases=yes
##
## @notes       This test will execute a list of testcases
##
## @author      snider@netapp.com

package nblade::nlm::Test;

use strict 'vars';
use nblade::nfsErrors;
use RandomData;
use TracePkg;
use Exporter;

use vars qw(@ISA @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT_OK = qw(NlmTestTests);

my ($nfs, $mountFH);
my $nlm;

my ($filehandle, $dirhandle);
my $testcase;
my $Trace = new TracePkg();

sub NlmTestTests {
    $nfs	= shift;
    $mountFH	= shift;
    $nlm	= shift;

    my ($data, $string);
    my $dir = "test$$.".int(rand(0xff));

    my ($maxOffset, $maxOffsetError);

    if ($nlm->{'version'} < 4) {
	$maxOffset = '0xffffffff';
	$maxOffsetError = 0;
    } else {
	$maxOffset = '0xffffffffffffffff';
	$maxOffsetError = 9;
    }

    my $skip = 0;
    if ($nlm->{'version'} >= 4 && $nlm->{Platform} =~ /ontap/i) {
	$nlm->ConnectToServer();
        my $release = $nlm->{serverobject}->get_release(format=>"cooked");
        $Trace->comment(msg => "Testing against a $release build.");
	if ($release eq 'Tricky.0') {
	    $skip = 1;
	}
    }

    if (($dirhandle = $nfs->Test("ALWAYS", "Mkdir", "--mkdirFH $mountFH --dirname $dir --filemode 0040777", 0)) eq "") {
	$dirhandle = $nfs->Test("ALWAYS", "Lookup", "--filename $dir --dirhandle $mountFH", 0);
    }
    if (($filehandle = $nfs->Test("ALWAYS", "Create", "--createFH $dirhandle --createfilename file --filemode 0100666 0", 0)) eq "") {
	$filehandle = $nfs->Test("ALWAYS", "Lookup", "--filename file --dirhandle $dirhandle", 0);
    }
    if ($filehandle eq "" || $dirhandle eq "") {
	die "Could not setup test\n";
    }

# Test 0: OK
    if ($nlm->checkForTest(0)) {
	$testcase = "0a";
	$nlm->AddTestcase($testcase, "test fh");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle", 0);

	$testcase = "0b";
	$nlm->AddTestcase($testcase, "test fh cookie=''");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie \"\"", 0);

	$testcase = "0c";
	$nlm->AddTestcase($testcase, "test fh cookie='random byte'");
	$data = randomString->new(size=>2, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string", 0);

	$testcase = "0d";
	$nlm->AddTestcase($testcase, "test fh cookie='1K bytes");
	$data = randomString->new(size=>2048, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string", 0);

	$testcase = "0e";
	$nlm->AddTestcase($testcase, "test fh cookie='random'");
	$data = randomString->new(size=>int(rand(2048))+1, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string", 0);

	$testcase = "0f";
	$nlm->AddTestcase($testcase, "test fh cookie='1K bytes of all 0'");
	$string = '0'x2048;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string", 0);

	$testcase = "0g";
	$nlm->AddTestcase($testcase, "test fh cookie='1K bytes of all 1");
	$string = 'f'x2048;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string", 0);

	$testcase = "0h";
	$nlm->AddTestcase($testcase, "test fh exclusive=0");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --exclusive 0", 0);

	$testcase = "0i";
	$nlm->AddTestcase($testcase, "test fh exclusive=0xffffffff");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --exclusive 0xffffffff", 0);

	$testcase = "0j";
	$nlm->AddTestcase($testcase, "test fh client='1K if all 0'");
	$string = '0'x1024;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname $string", 0);

	$testcase = "0k";
	$nlm->AddTestcase($testcase, "test fh client='1 byte'");
	$data = randomString->new(size=>1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname $string", 0);

	$testcase = "0l";
	$nlm->AddTestcase($testcase, "test fh client='1K bytes'");
	$data = randomString->new(size=>1024);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname $string", 0);

	$testcase = "0m";
	$nlm->AddTestcase($testcase, "test fh client=''");
	$data = randomString->new(size=>int(rand(1024))+1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname \"\"", 0);

	$testcase = "0n";
	$nlm->AddTestcase($testcase, "test fh owner=''");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner \"\"", 0);

	$testcase = "0o";
	$nlm->AddTestcase($testcase, "test fh owner='1 byte'");
	$data = randomString->new(size=>2, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner $string", 0);

	unless ($skip) {
	    $testcase = "0p";
	    $nlm->AddTestcase($testcase, "test fh owner='1K bytes'");
	    $data = randomString->new(size=>2048, hex => 1);
	    $string = $data->String();
	    $nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner $string", 0);
	}

	$testcase = "0q";
	$nlm->AddTestcase($testcase, "test fh owner='random'");
	$data = randomString->new(size=>int(rand(2048))+1, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner $string", 0);

	$testcase = "0r";
	$nlm->AddTestcase($testcase, "test fh owner='1K of all 0s'");
	$string = '0'x2048;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner $string", 0);

	$testcase = "0s";
	$nlm->AddTestcase($testcase, "test fh '1K of all 1s'");
	$string = 'f'x2048;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --lockowner $string", 0);

	$testcase = "0t";
	$nlm->AddTestcase($testcase, "test fh svid='random'");
	$string = int(rand(0xffffffff)+1);
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --locksvid $string", 0);

	unless ($skip) {
	    $testcase = "0u";
	    $nlm->AddTestcase($testcase, "test fh offset=$maxOffset");
	    $nlm->Test($testcase, "Test", "--filehandle $filehandle --lockoffset $maxOffset", $maxOffsetError);
	}

	$testcase = "0v";
	$nlm->AddTestcase($testcase, "test fh length=$maxOffset");
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --locklength $maxOffset", 0);

	unless ($skip) {
	    $testcase = "0w";
	    $nlm->AddTestcase($testcase, "test fh offset=$maxOffset length=$maxOffset");
	    $nlm->Test($testcase, "Test", "--filehandle $filehandle --lockoffset $maxOffset --locklength $maxOffset", $maxOffsetError);
	}

	$testcase = "0x";
	$nlm->AddTestcase($testcase, "test fh cookie='1K bytes' name='1K bytes' owner='1K bytes'");
	$data = randomString->new(size=>1024, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --cookie $string --clientname $string --lockowner $string", 0);
    }

    if ($nlm->checkForTest(7)) {
	$testcase = "7a";
	$nlm->AddTestcase($testcase, "test fh=''");
	$nlm->Test($testcase, "Test", "--filehandle \"\"", 7, "Garbage Arguments");

	$testcase = "7b";
	$nlm->AddTestcase($testcase, "test fh='1 byte string'");
	$data = randomString->new(size=>2, hex => 1); # 1 byte
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7c";
	$nlm->AddTestcase($testcase, "test fh='12 byte string'");
	$data = randomString->new(size=>24, hex => 1); # 12 byte
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7d";
	$nlm->AddTestcase($testcase, "test fh='32 byte string'");
	$data = randomString->new(size=>64, hex => 1); # 32 byte
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $string", 7);

	$testcase = "7e";
	$nlm->AddTestcase($testcase, "test fh='128 byte string'");
	$data = randomString->new(size=>256, hex => 1); # 128 byte
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7f";
	$nlm->AddTestcase($testcase, "test fh='12 bytes of all 1'");
	$string = 'f'x24;
	$nlm->Test($testcase, "Test", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7g";
	$nlm->AddTestcase($testcase, "test fh='12 bytes of all 0'");
	$string = '0'x24;
	$nlm->Test($testcase, "Test", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7h";
	$nlm->AddTestcase($testcase, "test fh='1250 byte string'");
	$data = randomString->new(size=>2050, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Test", "--filehandle $string", "Garbage Arguments");

	$testcase = "7i";
	$nlm->AddTestcase($testcase, "test fh name='512 bytes' namelength=5");
	$string = 'f'x1024;
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname $string --namelength 5", "Garbage Arguments");
    }

    if ($nlm->checkForTest(8)) {
	$testcase = '8b';
	$string = 'a'x1024;
	$nlm->AddTestcase($testcase, 'lock block=1 max clientname, lockowner, cookie, ');
	$nlm->Test($testcase, "Test", "--filehandle $filehandle --clientname $string --cookie $string$string --lockowner $string$string", 0);

    }

    $nfs->Test("ALWAYS", "Remove", "--removeFH $dirhandle --removefilename file", 0);
    $testcase = "7j";
    $nlm->AddTestcase($testcase, "test fh [file was just removed]");
    $nlm->Test($testcase, "Test", "--filehandle $filehandle", 7, 0);

    $nfs->Test("ALWAYS", "Rmdir", "--rmdirFH $mountFH --rmdirfilename $dir", 0);

    return $nlm;
}

# return value, so perl doesn't have a cow
1;
