## @summary     NATE test that verifies the NLMv3 Unlock procedure
##
## @description Make Unlock calls with different values
##
## @param       nbladeIP Required. The IP address of the server.
##
## @synopsis    ntest nbladeIP=<IP> Unlock.thpl
##
## @example     ntest nbladeIP=10.98.200.200 nblade/nlm3/Unlock.thpl
##
## @status      public
##
## @attribute   listtestcases=yes
##
## @notes       This test will execute a list of testcases
##
## @author      snider@netapp.com

package nblade::nlm::Unlock;

use strict 'vars';
use nblade::nfsErrors;
use RandomData;
use TracePkg;
use Exporter;

use vars qw(@ISA @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT_OK = qw(NlmUnlockTests);

my ($nfs, $mountFH);
my $nlm;
my $testcase;
my $Trace = new TracePkg();

my ($filehandle, $dirhandle);
my $dir = "unlock$$.".int(rand(0xff));

sub testAndPrintStatus {
    my ($cookie, $svid) = @_;
    $nlm->Test("Quiet$testcase", "Test", "--filehandle $filehandle --exclusive 1 ", 0);
    if ($nlm->{returnStatus}) {
	$nlm->printResult(1, 0, $testcase, 1);
    } else {
	$nlm->printResult(1, 1, $testcase, 1);
    }
}    

sub NlmUnlockTests {
    $nfs	= shift;
    $mountFH	= shift;
    $nlm	= shift;

    my ($data, $string);
    my $dir = "lock$$.".int(rand(0xff));

    my ($maxOffset, $maxOffsetError, $maxQuiet);

    if ($nlm->{'version'} < 4) {
	$maxOffset = '0xffffffff';
	$maxOffsetError = 0;
	$maxQuiet = 'Quiet';
    } else {
	$maxOffset = '0xffffffffffffffff';
	$maxOffsetError = 9;
	$maxQuiet = '';
    }

    my $skip = 0;
    if ($nlm->{'version'} >= 4 && $nlm->{Platform} =~ /ontap/i) {
	$nlm->ConnectToServer();
        my $release = $nlm->{serverobject}->get_release(format=>"cooked");
        $Trace->comment(msg => "Testing against a $release build.");
	if ($release eq 'Tricky.0') {
	    $skip = 1;
	}
    }

    if (($dirhandle = $nfs->Test("ALWAYS", "Mkdir", "--mkdirFH $mountFH --dirname $dir --filemode 0040777", 0)) eq "") {
	$dirhandle = $nfs->Test("ALWAYS", "Lookup", "--filename $dir --dirhandle $mountFH", 0);
    }
    if (($filehandle = $nfs->Test("ALWAYS", "Create", "--createFH $dirhandle --createfilename file --filemode 0100666 0", 0)) eq "") {
	$filehandle = $nfs->Test("ALWAYS", "Lookup", "--filename file --dirhandle $dirhandle", 0);
    }
    if ($filehandle eq "" || $dirhandle eq "") {
	die "Could not setup test\n";
    }

# Test 0: OK
    my $cookie = int(rand(0xffffffff)) + 1;
    my $svid = int(rand(0xffffffff)) + 1;
    if ($nlm->checkForTest(0)) {
	$testcase = "0a";
	$nlm->AddTestcase($testcase, "unlock fh cookie='random' svid='random'");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 0) 
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0b";
	$nlm->AddTestcase($testcase, "unlock fh cookie=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie \"\" --locksvid $svid", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie \"\" --locksvid $svid", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus("", $svid);

	$testcase = "0c";
	$nlm->AddTestcase($testcase, "unlock fh cookie='1K string'");
	$data = randomString->new(size=>int(rand(2048))+1, hex => 1);
	$string = $data->String();
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $string --locksvid $svid", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($string, $svid);

	$testcase = "0d";
	$nlm->AddTestcase($testcase, "unlock fh cookie='1K of all 0s'");
	$string = '0'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $string --locksvid $svid", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($string, $svid);

	$testcase = "0e";
	$nlm->AddTestcase($testcase, "unlock fh cookie='1K of all 1s'");
	$string = 'f'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $string --locksvid $svid", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $string --locksvid $svid", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($string, $svid);

	$testcase = "0f";
	$nlm->AddTestcase($testcase, "unlock fh client=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --clientname \"\"", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --clientname \"\"", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0g";
	$nlm->AddTestcase($testcase, "unlock fh client='1K string'");
	$data = randomString->new(size=>int(rand(1024))+1);
	$string = $data->String();
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --clientname $string", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --clientname $string", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0h";
	$nlm->AddTestcase($testcase, "unlock fh owner=''");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockowner \"\"", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner \"\"", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0i";
	$nlm->AddTestcase($testcase, "unlock fh owner='1K string'");
	$data = randomString->new(size=>int(rand(2048))+1, hex => 1);
	$string = $data->String();
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockowner $string", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0j";
	$nlm->AddTestcase($testcase, "unlock fh owner='1K all 0s'");
	$string = '0'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockowner $string", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0k";
	$nlm->AddTestcase($testcase, "unlock fh owner='1K all 1s'");
	$string = 'f'x2048;
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockowner $string", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid --lockowner $string", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid);

	$testcase = "0l";
	$nlm->AddTestcase($testcase, "unlock fh svid='random'");
	$string = int(rand(0xffffffff)+1);
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $string", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $string", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $string);

	unless ($skip) {
	    $testcase = "0m";
	    $nlm->AddTestcase($testcase, "unlock fh [offset=$maxOffset]");
	    $nlm->Test("$maxQuiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockoffset $maxOffset", $maxOffsetError);
	    if ($maxQuiet) {
		$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 0);
		testAndPrintStatus($cookie, $svid, "--lockoffset $maxOffset");
	    }
	}

	$testcase = "0n";
	$nlm->AddTestcase($testcase, "unlock fh [length=$maxOffset]");
	$nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --locklength $maxOffset", 0);
	$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid  --locklength $maxOffset", 0)
	unless ($nlm->{returnStatus});
	testAndPrintStatus($cookie, $svid, "--locklength $maxOffset");

	unless ($skip) {
	    $testcase = "0o";
	    $nlm->AddTestcase($testcase, "unlock fh [offset=$maxOffset length=$maxOffset]");
	    $nlm->Test("$maxQuiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid --lockoffset $maxOffset --locklength $maxOffset", $maxOffsetError);
	    if ($maxQuiet) {
		$nlm->Test("Quiet$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 0);
		testAndPrintStatus($cookie, $svid, "--lockoffset $maxOffset --locklength $maxOffset");
	    }
	}
    }

    if ($nlm->checkForTest(7)) {
	$testcase = "7a";
	$nlm->AddTestcase($testcase, "unlock fh=''");
	$nlm->Test($testcase, "Unlock", "--filehandle \"\"", 7, "Garbage Arguments");

	$testcase = "7b";
	$nlm->AddTestcase($testcase, "unlock fh='1 byte string'");
	$data = randomString->new(size=>2, hex => 1); # 1 byte
	$string = $data->String();
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7c";
	$nlm->AddTestcase($testcase, "unlock fh='12 byte string'");
	$data = randomString->new(size=>24, hex => 1); # 12 byte
	$string = $data->String();
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7d";
	$nlm->AddTestcase($testcase, "unlock fh='32 byte string'");
	$data = randomString->new(size=>64, hex => 1); # 32 byte
	$string = $data->String();
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7);

	$testcase = "7e";
	$nlm->AddTestcase($testcase, "unlock fh='128 byte string'");
	$data = randomString->new(size=>256, hex => 1); # 128 byte
	$string = $data->String();
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7f";
	$nlm->AddTestcase($testcase, "unlock fh='12 bytes all 1s'");
	$string = 'f'x24;
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7g";
	$nlm->AddTestcase($testcase, "unlock fh='12 bytes all 0s'");
	$string = '0'x24;
	$nlm->Test($testcase, "Unlock", "--filehandle $string", 7, "Garbage Arguments");

	$testcase = "7h";
	$nlm->AddTestcase($testcase, "unlock fh='1250 bytes'");
	$data = randomString->new(size=>2050, hex => 1);
	$string = $data->String();
	$nlm->Test($testcase, "Unlock", "--filehandle $string", "Garbage Arguments");

	$testcase = "7i";
	$nlm->AddTestcase($testcase, "unlock fh client='512 bytes' namelength=5");
	$string = 'f'x1024;
	$nlm->Test($testcase, "Unlock", "--filehandle $filehandle --clientname $string --namelength 5", "Garbage Arguments");
    }

    $testcase = "7j";
    $nlm->AddTestcase($testcase, "unlock fh [file was just removed]");
    $nlm->Test("Quiet$testcase", "Lock", "--filehandle $filehandle  --exclusive 1 --cookie $cookie --locksvid $svid", 0);
    $nfs->Test("ALWAYS", "Remove", "--removeFH $dirhandle --removefilename file", 0);
    $nlm->Test("$testcase", "Unlock", "--filehandle $filehandle --cookie $cookie --locksvid $svid", 7, 0);

    $nfs->Test("ALWAYS", "Rmdir", "--rmdirFH $mountFH --rmdirfilename $dir", 0);

    return $nlm;
}

# return value, so perl doesn't have a cow
1;
