# File Name     :       TestTemplateGen.pm
#
# Summary       :       Contains methods to create a thpl file template 
#		
#
# Creation Date :       17th August 2010
#
# Author        :       tanzim , dl-sm-nacl-dev@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#       Summary: Enhancements to template
#		 Template provided by Jagadeesh
#       Date   : 1st September 2010
#       Author : tanzim
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#

#
# Package Declaration
#

package CodeModules::TestTemplateGen;

#
# Compiler Directives
#

use strict;
use warnings;

#
# Perl Modules
#

use Data::Dumper;
use Cwd;
use IO::File;

#
# User Specific Modules
#


# Class Objects


# Constructor Definition

sub new {
    my $class = shift;
    my $self  = {};
    bless $self, $class;
    return $self;
}


################################################################################
# Subroutine  : create_test_file                                               #
#                                                                              #
# Description : This method creates a  test thpl file                          #
#                                                                              #
# Returns     : Success / Failure					       #
#		File Handler                                                   #
#                                                                              #
################################################################################

sub create_test_file {

    my $self      = shift;
    my $command   = shift;
    my $code_path = shift;
    my $interface = shift;	

    my $success = 1;
    my $fh_ut;
    $command =~ s/\s+/_/g;
    #my  $file_name = $command . "_cli.thpl";
    my $file_name =  $command."_".lc $interface.".thpl";	

#    print "\n===> Generating  Test THPL Template: $file_name at location : $code_path\n"; 

# The error handling is done in this way so that the tool does not exit here as there may be other commands remaining
# Moreover the setup uninitialization also has to be invoked
    if ( -e "$code_path/$file_name" ) {

        print
            "\n\t WARNING : An old version of  test file \"$file_name\" exists at \"$code_path\"\n\tDeleting the file ...\n";
        unless ( unlink("$code_path/$file_name") ) {
            warn
                "\n\tERROR : Unable to delete old file \"$file_name\" at \"$code_path\" : $!\n";
            $success = 0;
        }
    }

    if ($success) {
        $fh_ut = IO::File->new("+>> $code_path/$file_name");
        unless ( defined $fh_ut ) {
            warn
                "\n\tERROR : Unable to create file $file_name at $code_path : $!\n";
            $success = 0;
        }
    }

    #Better way to return - ?
    return ( $success, $fh_ut, $file_name );
}

################################################################################
# Subroutine  : close_test_file                                                #
#                                                                              #
# Description : This method closes the file handle to a  test thpl file        #
#                                                                              #
# Returns     : Nothing                                                        #
#                                                                              #
################################################################################

sub close_test_file {

    my $self = shift;
    my $fh   = shift;

    undef $fh;

}

################################################################################
# Subroutine  : gen_test_template                                              #
#                                                                              #
# Description : This method generates the  test template for DFM               #
#               The template resides in a thpl file consisting                 #
#                       - Tharndoc                                             #
#                       - main()                                               #
#                       - init()                                               #
#                       - setup()                                              #
#                       - cleanup()                                            #
#                       - uninit                                               #
#                                                                              #
# Returns     : Generates the test code template                               #
#                                                                              #
################################################################################

sub gen_test_template {

    my ( $self, %args ) = @_;
 
    ( $args{"interface"} =~ /zapi/i ) ? ($self->_write_tharndoc_zapi(\%args)) : ($self->_write_tharndoc(\%args));

    $self->_write_module_import(\%args);	
	
    $self->_write_placeholder(\%args); 

    $self->_write_main(\%args);
	
    $self->_write_init(\%args);	

    $self->_write_setup(\%args); 	

    $self->_write_cleanup(\%args); 	

    $self->_write_uninit(\%args);	

}

################################################################################
# Subroutine  : gen_test_case_template                                         #  
#                                                                              #
# Description : This method generates the  test case template for DFM          #
#               unit test cases                                                #
#                                                                              #
# Returns     : Generates the test code template                               #
#                                                                              #
################################################################################

sub gen_test_case_template{

	my ( $self, %args ) = @_;

	#print "$args{testcase_name} : $args{testcase_description} : $args{expected_output} : $args{testcase_type} : $args{file_handle} ";
	#print Dumper ($args{parameter_set});

        if( $args{'interface'} =~ /zapi/i ) {

           ($args{'testcase_type'} =~ /positive/) ? $self->_gen_positive_tc_template_zapi(\%args) : $self->_gen_negative_tc_template_zapi(\%args) ;
        } else {
	
    	    ($args{'testcase_type'} =~ /positive/) ? $self->_gen_positive_tc_template(\%args) : $self->_gen_negative_tc_template(\%args);	
	}

}


################################################################################
# Subroutine  : _write_tharndoc                                                #
#                                                                              #
# Description : This method writes tharndoc code                               #
#                                                                              #
################################################################################

sub _write_tharndoc {

    my ( $self, $args ) = @_;
    my $fh = $args->{"file_handle"};	
    my $command = $1 if ($args->{"file_name"} =~ /(.*)_$args->{"interface"}\.thpl/i);	
    $command =~ s/_/ /;	
    my ($day, $month, $year) = (localtime)[3,4,5];	
    my $creation_time = $day."-".($month+1)."-".($year+1900);	

    my $user_message = ">>>>>>USER NOTE>>>> Placeholder for test case id and name";

print $fh <<"FOO" or warn "\n\tWARNING : $!\n";
# \$Id:  \$
#
# Copyright (c) 2010 NetApp
# All rights reserved.
#
##
## \@summary Unit test script for '$command'
##
## \@description
##  Basic unit testing for '$command'
##
## \@description Objective:
##   Test the '$command'
##
## \@synopsis
##   ntest $args->{"file_name"}
#
#  Test Control Parameters
#
## \@param TESTCASES
##   Optional: List of test cases to execute.  This can be "all", a
##   comma separated list of test cases, or the index/id of test cases.
##   The user can include switches and multipliers as well.  See the
##   documentation for the TCD lib for a full description of this
##   parameter.  By default all test cases will be executed one time.
##
## \@param LISTTESTCASES
##   Optional: If this is set to yes, the script will not execute the
##   test cases, but instead will print a list of the test cases
##   and their descriptions that correspond to the other input
##   parameters.  Default is 0.
##
## \@param NUFFO_MULTITEST_LOGS
##   Optional: This parameter is used by nuffo.thpl to log
##   execution resuts in NUFFO, set this to "yes".
#
#
#  Test Case Specific Parameters
#
## \@param DFM
##   Required: Hostspec of the DFM Server
##
## \@param FILER
##   Optional: Hostspec of the FILER
##
## \@param CDEF_PATH
##   Optional: Custom CDEF Location
##
## \@example
##   Run the $args->{"file_name"}
##
##   # run all the test cases
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"}
##
##   # list of test cases to execute - used by nip
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         LISTTESTCASES=1
##
##   # run all the test cases use -q "quiet" so you don't get STDOUT output
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES="-q,all"
##
##   # run specific test cases
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"}
##         TESTCASES=testcase_1,testcase_2
##
##   # run test cases that begin with "testcase" using the "%" wild card
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=testcase%
##
##   # execute test case testcase_1 50 times
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=testcase_1*50
##
##   # execute specific tests and skipover setup and cleanup
##   # (-nsc=no setup or cleanup and overrides testcase specific params)
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=-nsc,testcase_1
##
##   This test includes multiple test cases whose execution can be
##   controlled by the user.  Unless otherwise specified, all tests
##   are executed by default. Test control is accomplished by using
##   the Test Case Driver (TCD.pm).  The following test cases
##   (or purposes) can be executed by supplying the TESTCASES parameters
##   with either the test name or id value(s) on the command
##   line.  Please see the TCD.pm POD documentation for other command
##   line options.
##
## \@duration # %% {Time taken to execute all the testcases} % #
##
## \@keywords '$command', $args->{"interface"}
##
## \@see-also # %% see also list % #
##
## \@burt # %% BURTID:"Description of the burt" % #
##
## \@status review
##   
## \@attributes listtestcases=yes
##
## \@author dl-sm-nacl-dev\@netapp.com
## \@change $creation_time  dl-sm-nacl-dev:   Template created


FOO
}

################################################################################
# Subroutine  : _write_tharndoc_zapi                                                #
#                                                                              #
# Description : This method writes tharndoc code                               #
#                                                                              #
################################################################################

sub _write_tharndoc_zapi {

    my ( $self, $args ) = @_;
    my $fh = $args->{"file_handle"};
    my $command = $1 if ($args->{"file_name"} =~ /(.*)_$args->{"interface"}\.thpl/i);
    my $code_type = $args->{"code_type"};  

    $command =~ s/_/-/g;
    my ($day, $month, $year) = (localtime)[3,4,5];
    my $creation_time = $day."-".($month+1)."-".($year+1900);

    my @zapi_list = @{$args->{"zapi_list"}};
    my %zapi_error_desc = %{$args->{"zapi_error_description"}};

    # Make a string of the zapi list
    my $zapi_list_string = "";
    foreach my $zapi ( @zapi_list ) {
       $zapi_list_string .= "\n##  ". $zapi;
    }

    my $zapi_error_string = "";
    foreach my $err ( keys( %zapi_error_desc ) ) {  

        $zapi_error_string .= "\n##  ". $err . " : ". $zapi_error_desc{$err};
    }

    # Variable to hold the name of the parameter string
    # It is set to DFM by default. If $code_type is "filer", then appropriate changes are made later. 
    my $filer_param = "DFM";  
 
print $fh <<"FOO" or warn "\n\tWARNING : $!\n";
# \$Id:  \$
#
# Copyright (c) 2011 NetApp
# All rights reserved.
#
##
## \@summary ZAPI Unit test script for '$command'
##
## \@description
##  Basic unit testing for '$command'
##
##  This Script has been generated for version: $args->{"version"}  
## 
##  # The ZAPIs getting tested in this script are:  
##$zapi_list_string  
##
##  # Description of error numbers thrown by the ZAPIs tested in the script:
##$zapi_error_string
## 
## \@description Objective:
##   Test the '$command'
##
## \@synopsis
##   ntest $args->{"file_name"}
#
#  Test Control Parameters
#
## \@param TESTCASES
##   Optional: List of test cases to execute.  This can be "all", a
##   comma separated list of test cases, or the index/id of test cases.
##   The user can include switches and multipliers as well.  See the
##   documentation for the TCD lib for a full description of this
##   parameter.  By default all test cases will be executed one time.
##
## \@param LISTTESTCASES
##   Optional: If this is set to yes, the script will not execute the
##   test cases, but instead will print a list of the test cases
##   and their descriptions that correspond to the other input
##   parameters.  Default is 0.
##
## \@param NUFFO_MULTITEST_LOGS
##   Optional: This parameter is used by nuffo.thpl to log
##   execution resuts in NUFFO, set this to "yes".
#
#
#  Test Case Specific Parameters
#
#
FOO

if ($code_type eq "dfm") {

print $fh <<"FOO" or warn "\n\tWARNING : $!\n";
## \@param DFM
##   Required: Hostspec of the DFM Server
##
## \@param FILER
##   Optional: Hostspec of the FILER
##
## \@param CDEF_PATH
##   Optional: Custom CDEF Location
##
FOO

} elsif ($code_type eq "filer") {

    my $family = $args->{"family"} ? $args->{"family"} : "7Mode" ;

    # Setting the Filer parameter depending on the family name.

    if ( $family =~ /7Mode|ontapD/ ) {
        $filer_param = "FILER";
    } elsif ( $family eq "CMode") {
        $filer_param = "FILER";
    } elsif ( $family eq "Vfiler") {
        $filer_param = "VFILER";
    } elsif ( $family eq "Vserver") {
        $filer_param = "VSERVER";
    }
print $fh <<"FOO" or warn "\n\tWARNING : $!\n";
## \@param $filer_param
##   Required: Hostspec of $filer_param.
##
FOO
}

print $fh <<"FOO" or warn "\n\tWARNING : $!\n";
## \@example
##   Run the $args->{"file_name"}
##
##   # run all the test cases
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"}
##
##   # list of test cases to execute - used by nip
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         LISTTESTCASES=1
##
##   # run all the test cases use -q "quiet" so you don't get STDOUT output
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES="-q,all"
##
##   # run specific test cases
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"}
##         TESTCASES=testcase_1,testcase_2
##
##   # run test cases that begin with "testcase" using the "%" wild card
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=testcase%
##
##   # execute test case testcase_1 50 times
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=testcase_1*50
##
##   # execute specific tests and skipover setup and cleanup
##   # (-nsc=no setup or cleanup and overrides testcase specific params)
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \
##         TESTCASES=-nsc,testcase_1
##
##   # execute with tharnhost files specified
##   ntest NUFFO_MULTITEST_LOGS=yes -noconsole $args->{"file_name"} \  
##         $filer_param=<tharnhost_file>
##
##   This test includes multiple test cases whose execution can be
##   controlled by the user.  Unless otherwise specified, all tests
##   are executed by default. Test control is accomplished by using
##   the Test Case Driver (TCD.pm).  The following test cases
##   (or purposes) can be executed by supplying the TESTCASES parameters
##   with either the test name or id value(s) on the command
##   line.  Please see the TCD.pm POD documentation for other command
##   line options.
##
## \@duration # %% {Time taken to execute all the testcases} % #
##
## \@keywords '$command', $args->{"interface"}
##
## \@see-also # %% see also list % #
##
## \@burt # %% BURTID:"Description of the burt" % #
##
## \@status review
##
## \@attributes listtestcases=yes
##
## \@author dl-ncage-dev\@netapp.com
## \@change $creation_time  dl-ncage-dev:   Template created


FOO
}


################################################################################
# Subroutine  : _write_module_import                                           #
#                                                                              #
# Description : This method writes module import code                          #
#                                                                              #
################################################################################

sub _write_module_import {

	my ( $self, $args ) = @_;
	my $fh = $args->{"file_handle"};
	my $interface = $args->{"interface"};

print $fh <<'MOD' or warn "\n\tWARNING : $!\n";

#
# Compiler directives.
#
use strict;
use warnings;

#
# Module import.
#
use TCD;
use NACL::APISet;
MOD
if( $interface =~ /CLI/i ) {

   print $fh <<'MOD' or warn "\n\tWARNING : $!\n";
use Data::Dumper;
MOD
}
print $fh <<'MOD' or warn "\n\tWARNING : $!\n";
use DfmUtils;
use XML::Simple;
use ZapiCli;

#
# Import exceptions
#

use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::CDEFNotFoundException qw(:try);
use NACL::APISet::Exceptions::MissingRequiredParamException qw(:try);
use NACL::APISet::Exceptions::MethodNotFoundException qw(:try);
use NACL::APISet::Exceptions::UnsupportedOptionException qw(:try);

MOD
}

################################################################################
# Subroutine  : _write_placeholder                                             #
#                                                                              #
# Description : This method writes placeholder code for global vars and tc list#
#                                                                              #
################################################################################

sub _write_placeholder {

    my ( $self, $args ) = @_;
    my $fh = $args->{"file_handle"};
    my $code_type = $args->{"code_type"}; 

    my $user_message = ">>>>>>USER NOTE>>>> Placeholder for test case id and name";

    my $filer_param = "DFM"; 

    if ($args->{"interface"} =~ /zapi/i && $code_type eq "filer") {

        my $family = $args->{"family"} ? $args->{"family"} : "7Mode" ;

        # Setting the Filer parameter depending on the family name.

        if ( $family =~ /7Mode|ontapD/ ) {
            $filer_param = "FILER";
        } elsif ( $family eq "CMode") {
            $filer_param = "FILER";
        } elsif ( $family eq "Vfiler") {
            $filer_param = "VFILER";
        } elsif ( $family =~ /Vserver/i ) {
            $filer_param = "VSERVER";
        }
    }  

print $fh <<"PH" or warn "\n\tWARNING : $!\n";

#
# Global Variables.
#
use vars qw(

    \@TC_LIST
    \$Test
    \$Api_Set_Obj_DFM
    \$Api_Set_Obj_FILER	
    \$Cdef_Path
    \$$filer_param
);

#
# Test cases to execute
#
my \@TC_LIST = (
		_NG_TESTCASE_INFO
		);

PH
}

################################################################################
# Subroutine  : _write_main                                                    #
#                                                                              #
# Description : This method writes main function code                          #
#                                                                              #
################################################################################

sub _write_main {

	  my ( $self, $args ) = @_;
         my $fh = $args->{"file_handle"};

print $fh <<'MAIN' or warn "\n\tWARNING : $!\n";

#
# Run program
#
&main();
 
########## MAIN    ###################################################
sub main {
 
    #
    # Standard break point for debugging - don't remove
    #
    $DB::single = 2;
    $Test = new TCD( -testcases => \@TC_LIST );
 
    if ( $Test->error() ) {
        $Test->log( $Test->errmsg );
        return $TCD::FAIL;
    }
    $Test->run_test();
 
    if ( $Test->error() ) {
        $Test->log( $Test->errmsg );
        return $TCD::FAIL;
    }
 
    exit(0);
}


MAIN
}

################################################################################
# Subroutine  : _write_init                                                    #
#                                                                              #
# Description : This method writes init function code                          #
#                                                                              #
################################################################################

sub _write_init {

	 my ( $self, $args ) = @_;
         my $fh = $args->{"file_handle"};

	# Take the code type
         my $code_type = $args->{"code_type"}? $args->{"code_type"} : "dfm";

	if($code_type eq "dfm") {  

print $fh <<'INIT' or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : init()
# Input        : None
# Return value : $TCD::PASS by default
# Summary      : init() is automatically called before executing tests to get
#                test parameters and to create any needed resources
###############################################################################
sub init() {
    $Test->description( "Retrieve variables from the test harness and "
            . "instantiate objects needed for test execution" );
    $DFM    = $Test->get_param("DFM");
    #$FILER  = $Test->get_param("FILER");
    $Cdef_Path = $Test->get_param("CDEF_PATH");
 
    # %%         User Specific init code can go here!               % #
 
    return $TCD::PASS;
}

INIT
    } elsif ($code_type eq "filer") {

        my $family = $args->{"family"} ? $args->{"family"} : "7Mode" ;

        # Setting the Filer parameter depending on the family name.

        my $filer_param;
        if ( $family =~ /7Mode|ontapD/ ) {
            $filer_param = "FILER";
        } elsif ( $family eq "CMode") {
            $filer_param = "FILER";
        } elsif ( $family eq "Vfiler") {
            $filer_param = "VFILER";
        } elsif ( $family =~ /Vserver/i ) {
            $filer_param = "VSERVER";
        }

print $fh <<"INIT" or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : init()
# Input        : None
# Return value : \$TCD::PASS by default
# Summary      : init() is automatically called before executing tests to get
#                test parameters and to create any needed resources
###############################################################################
sub init() {
    \$Test->description( "Retrieve variables from the test harness and "
            . "instantiate objects needed for test execution" );

    \$$filer_param  = \$Test->get_param("$filer_param");

    # %%         User Specific init code can go here!               % #

    return \$TCD::PASS;
}

INIT
    }

}

################################################################################
# Subroutine  : _write_setup                                                   #
#                                                                              #
# Description : This method writes setup function code                         #
#                                                                              #
################################################################################

sub _write_setup {

	my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};
	my $interface = uc($args->{"interface"});

        # Take the code type
        my $code_type = $args->{"code_type"}? $args->{"code_type"} : "dfm";

	my $user_message = ">>>>>>USER NOTE>>>> Placeholder for setup related code";	

    if($code_type eq "dfm") {

print $fh <<'SETUP' . <<"SETUP_INTERP" or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : setup()
# Inpu t        : None
# Return value : $TCD::PASS
# Summary      : setup() is automatically called before executing tests
###############################################################################
sub setup() {
 
    $Test->description(
        "Creating APISet Object required to connect Filer and DFM Server." );
    my $result = $TCD::PASS;
    my $Host   = host($DFM);	
    my $connection_object  = connect ("#ssh");	
    $Test->nlog("Establishing an APISet connection to DFM.");
SETUP
   \$Api_Set_Obj_DFM = NACL::APISet->new(
	connection => \$connection_object,
        category  => "Host",
        interface => "$interface",
        set       => "DFM",
        cdef_path => \$Cdef_Path,
    );   
  

    #my \$Filer = host(\$FILER);
    #\$Test->nlog("Establishing an APISet connection to Storage System.");
    #\$Api_Set_Obj_FILER = NACL::APISet->new(
    #    hostobj   => \$Filer,
    #    category  => "Node",
    #    interface => "$interface",
    #    set       => "CMode",
    #);
 
    #unless (\$Api_Set_Obj_FILER || \$Api_Set_Obj_DFM ) {
     unless ( \$Api_Set_Obj_DFM ) {	
        \$result = \$TCD::FAIL;    
    }
SETUP_INTERP
if( $interface =~ /CLI/i ) {

    print $fh <<'SETUP' or warn "\n\tWARNING : $!\n";
 
    # DfmUtils::is_dfm_alive( api_set_obj_dfm => $Api_Set_Obj_DFM ) ? undef : return \$TCD::CONFIG : "" ;
SETUP
}
print $fh <<"SETUP_INTERP" or warn "\n\tWARNING : $!\n";

    # %%         User Specific setup code can go here!              % #
 
    return \$result;

}
	
SETUP_INTERP
} elsif ($code_type eq "filer") {

    my $family = $args->{"family"} ? $args->{"family"} : "7Mode" ;

        # Setting the Filer parameter depending on the family name.

        my $filer_param;
        if ( $family =~ /7Mode|ontapD/ ) {
            $filer_param = "FILER";
        } elsif ( $family eq "CMode") {
            $filer_param = "FILER";
        } elsif ( $family eq "Vfiler") {
            $filer_param = "VFILER";
        } elsif ( $family =~ /Vserver/i ) {
            $filer_param = "VSERVER";
        }

print $fh <<"SETUP" or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : setup()
# Input        : None
# Return value : \$TCD::PASS
# Summary      : setup() is automatically called before executing tests
###############################################################################
sub setup() {

    \$Test->description(
        "Creating APISet Object required to connect to Filer" );

    \$Test->nlog("Establishing an APISet connection to Storage System.");

    my \$Filer = host(\$$filer_param);
SETUP

if( $family =~ /Vserver/ ) {

print $fh <<"VSERVER" or warn "\n\tWARNING : $!\n";
   my \$Vserver = "vserver_name";
VSERVER

} # end of if family is Vserver

print $fh <<"SETUP" or warn "\n\tWARNING : $!\n";

    \$Api_Set_Obj_FILER = NACL::APISet->new(
        hostobj   => \$Filer,
        category  => "Node",
        interface => "$interface",
        set       => "$family",
SETUP

if( $family =~ /Vserver/ ) {

print $fh <<"VSERVER" or warn "\n\tWARNING : $!\n";
        context_definition => {name => \$Vserver,},
VSERVER

} # end of if family is Vserver

print $fh <<"SETUP" or warn "\n\tWARNING : $!\n";
    );

    unless ( \$Api_Set_Obj_FILER) {
        return \$TCD::CONFIG;
    }

    # %%         User Specific setup code can go here!              % #

    return \$TCD::PASS;

}

SETUP
}
}


################################################################################
# Subroutine  : _write_cleanup                                                 #
#                                                                              #
# Description : This method writes cleanup function code                       #
#                                                                              #
################################################################################

sub _write_cleanup {

	my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};

	# Take the code type
        my $code_type = $args->{"code_type"}? $args->{"code_type"} : "dfm";

	my $user_message = ">>>>>>USER NOTE>>>> Placeholder for cleanup related code";

	if($code_type eq "dfm") {

print $fh <<'CLEANUP' or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : cleanup
# Input        : None
# Return value : $TCD::PASS
# Summary      : cleanup() is automatically called after executing tests
###############################################################################
sub cleanup() {
    $Test->description("Cleanup");
    my $result     = $TCD::PASS;
 
    # %%         User Specific clean up code can go here!           % #
 
    #$Api_Set_Obj_FILER->cleanup_APISet();
    $Api_Set_Obj_DFM->cleanup_APISet();
    return $result;
}

CLEANUP
    } elsif ($code_type eq "filer") {

print $fh <<'CLEANUP' or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : cleanup
# Input        : None
# Return value : $TCD::PASS
# Summary      : cleanup() is automatically called after executing tests
###############################################################################
sub cleanup() {
    $Test->description("Cleanup");

    # %%         User Specific clean up code can go here!           % #

    $Api_Set_Obj_FILER->cleanup_APISet();

    return $TCD::PASS;
}

CLEANUP
    }
}

################################################################################
# Subroutine  : _write_uninit                                                  #
#                                                                              #
# Description : This method writes uninit function code                        #
#                                                                              #
################################################################################


sub _write_uninit {

	my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};

print $fh <<'UNINIT' or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : uninit()
# Input        : None
# Return value : $TCD::PASS
# Summary      : uninit() is automatically called after executing tests
###############################################################################
sub uninit() {
    $Test->description("uninit");
 
    # %%         User Specific unint code can go here!              % #
 
    return $TCD::PASS;
}

UNINIT
}

################################################################################
# Subroutine  : _write_get_failure_type                                        #
#                                                                              #
# Description : This method writes the _get_failure_type function code         #
#                                                                              #
################################################################################

sub write_get_failure_type {

        my ( $self, %args ) = @_;
        my $fh = $args{"file_handle"};

print $fh <<'FAILURE_TYPE' or warn "\n\tWARNING : $!\n";


###############################################################################
# Function     : _get_failure_type
# Input        : The Subtest result Object.
# Return value : Returns the Result object type in TCD style.
# Summary      : The function will get the type of the result in a test case.
###############################################################################
 
sub _get_failure_type {
 
    my ($rslt, $result, @results);
 
    $rslt    = shift;
    @results = $rslt->rslts();
 
    foreach $result (@results) {
 
        if    ($result->type() =~ /PASS/i)    { return $TCD::PASS; }
        elsif ($result->type() =~ /PARAM/i)   { return $TCD::PASS; }
        elsif ($result->type() =~ /INFO/i)    { return $TCD::PASS; }
        elsif ($result->type() =~ /WARN/i)    { return $TCD::FAIL; }
        elsif ($result->type() =~ /NA/i)      { return $TCD::NA; }
        elsif ($result->type() =~ /INSPECT/i) { return $TCD::FAIL; }
        elsif ($result->type() =~ /FAIL/i)    { return $TCD::FAIL; }
        elsif ($result->type() =~ /FATAL/i)   { return $TCD::FAIL; }
        elsif ($result->type() =~ /CONFIG/i)  { return $TCD::CONFIG; }
        elsif ($result->type() =~ /SCRIPT/i)  { return $TCD::SCRIPT; }
        else                                  { return $TCD::FAIL; }
    }
}

FAILURE_TYPE

}

## New for ZAPIS ##

################################################################################
# Subroutine  : _gen_positive_tc_template_zapi                                 #
#                                                                              #
# Description : This method writes positive test case template code            #
#                                                                              #
# Inputs      : 'code_type' - Valid values: 'dfm' or 'ontap'                   #
#               'file_handle' - File handle of the file to which content is    #
#			   	being appended 				       #
#		'zapi_stack' - Data structure containing info of zapi calls    #
#	        'index'  - Variable telling the depth of zapi calls            #
#               'testcase_name' - String containing the test case name         #
#		'testcase_description' - String containing tc description      #
#		'testcase_type' - String describing the type of test case      #
#		'expected_output' - String containing the regular expression   #
#				    that validates the output of the string    #
#		'setup_hash' - Reference to a hash containing calls to the     #
#				setup and cleanup 			       #
#               'iter_zapi' - Flag value telling if the zapi is an iter or not #  
#									       #	
################################################################################


sub _gen_positive_tc_template_zapi {

        my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};

        my @variable_list;
        my @user_message = ("#>>>>>>USER NOTE>>>> Test case Setup/Cleanup to be handled","#>>>>>>USER NOTE>>>> Expected Output to be validated or added depending on the test case requirements");

	# Take the code type
	my $code_type = $args->{"code_type"};
	
	# Suffix for the APISet object name  
	my $suffix; 
	if( $code_type =~ /dfm/i ) {
		$suffix = "DFM";
	
	} else {  
		$suffix = "FILER";  

	}   
        # Take the zapis to be executed
        my @zapi_string = ( );
        my @array = @{$args->{"zapi_stack"}};
        my $index = $args->{"index"};

        my %setup_hash =  %{$args->{"setup_hash"}};
      
	# String for variable initialization      
	my $init_string = "";

	for ( my $i = $index -1; $i > 0 ; $i-- ) {
	
	   $array[$i] =~ s/(Val|Bool)://g;
	   $init_string .= "\t\t" . $array[$i] . "\n";

	} 

	if( $init_string ne "" ) {  

		$init_string = "\n\t\t#####  Variable Initialisation  #####\n". $init_string;  
	} 
  
	# Variable for api call
	$array[0] =~ s/(Val|Bool)://g;
	my $api_call = $array[0];

	# API Call for setup
	my $setup_code = "";
	if( defined $setup_hash{"setup"}) {   

	    if( $args->{"iter_zapi"} =~ /1/ ) {

                   $setup_code .= "\n\t\t###  Setup Code  ####\n". 
				  "\t\tmy \$setup_resp = \$Api_Set_Obj_$suffix->" . $setup_hash{"setup"} .
				  "\n\t\tmy \@parsed_output = \@{\$setup_resp->get_parsed_output()}; ".
                                  "\n\t\tmy \$tag0 = \$parsed_output[0]->{\'tag\'};  ";
            } else {  
 
	           $setup_code =  "\n\t\t###  Setup Code  ####\n".   
			   "\t\t## Please uncomment and use ##\n" .  
		 	   "\t\t#my \$setup_resp = \$Api_Set_Obj_$suffix->" . $setup_hash{"setup"};

	    }  
            $setup_code .= "\n\n\t\t#########################\n";     
	}

	#API Call for clean up
	my $cleanup_code = "";
	if( defined $setup_hash{"cleanup"}) {    

	    $cleanup_code = "\t\t\t# Code for cleanup goes here\n".
			    "\t\t\ttry {\n\n".   
			    "\t\t\t\t# Cleanup APISet call\n\n".   
                            "\t\t\t\t# Please initialize the variables and uncomment the cleanup code.\n".
			    "\t\t\t\t#\$Api_Set_Obj_$suffix->" . $setup_hash{"cleanup"} . "\n\n".      
			    "\t\t\t} catch  NACL::APISet::Exceptions::ResponseException with {\n".   
			    "\t\t\t\tmy \$exception_object = shift;\n".   
			    "\t\t\t\t\$response = \$exception_object->get_response_object();\n".   
			    "\t\t\t\t\$Log->log(\"Cleanup step Failed.\");\n".   
			    "\t\t\t\t\$Log->log(\"Exception details: \".Dumper(\$exception_object));\n".   
			    "\t\t\t}\n";
	}  

	# Code for DATA VALIDATION #
	my $data_validation = "";

	if( defined $setup_hash{"data_validate"}{"check"} ) {  

	    $data_validation = "\n\t\t\t    ### Data Validation done here ###\n". 
			       "\t\t\t    my \$dv_resp = \$Api_Set_Obj_$suffix->". $setup_hash{"data_validate"}{"start_call"} . "\n\n".
			       "\t\t\t    my \@parsed_output = \@{\$dv_resp->get_parsed_output()}; ". "\n".
				"\t\t\t    my \$tag0 = \$parsed_output[0]->{\'tag\'}; "."\n".
				"\t\t\t    my \$recs = \$parsed_output[0]->{\'records\'}; "."\n\n". 
				"\t\t\t    my \$count = 10;  " . "\n" . 
				"\t\t\t    my \$pass_flag = 0;  " . "\n" . 
				"\t\t\t    my \$expected_data = \"". $setup_hash{"data_validate"}{"check"} . "\";". "\n" .  
				"\t\t\t    for( my \$loop_rec = \$recs; \$loop_rec > 0 ; \$loop_rec-=\$count ) {  " . "\n\n" .  
			  	"\t\t\t\t    \$dv_resp = \$Api_Set_Obj_$suffix->" . $setup_hash{"data_validate"}{"next_call"} . "( \"tag\" => \$tag0, \"maximum\" => \$count ); ". "\n\n" .    
				"\t\t\t\t    my \$dv_raw_output = \$dv_resp->get_raw_output(); ". "\n".
			        "\t\t\t\t    if( \$dv_raw_output =~ /\$expected_data/s ) { ". "\n". 
				"\t\t\t\t         \$pass_flag = 1;  " . "\n" .  
				"\t\t\t\t         last;  " . "\n\n" . 
				"\t\t\t\t    }    ". "\n".  
				"\t\t\t    }   ". "\n\n" .   
				"\t\t\t    if( \$pass_flag == 1 ) {  " . "\n".  
				"\t\t\t         \$Log->log(pass_result(\$runid));  " . "\n" . 
			        "\t\t\t         \$Log->log(\"Data Validation PASSED!\"); ". "\n" . 
				"\t\t\t    } else {  ". "\n" .  
				"\t\t\t           \$Log->log(\"Data Validation FAILED!\"); ". "\n" .
				"\t\t\t           \$Log->log(\"Expected pattern :   \"\.\$expected_data);  " . "\n" .  
				"\t\t\t           \$Log->log(fail_result(\$runid)); " . "\n" .
				"\t\t\t    }  " . "\n\n" .
				"\t\t\t    \$Api_Set_Obj_$suffix->". $setup_hash{"data_validate"}{"end_call"} . "( \"tag\" => \$tag0 ); " . "\n" ;	

	}
 
	# Take the expected output
	my $expected_output = $args->{"expected_output"};
 
print $fh <<"POSITIVE_INTERP"  or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : $args->{testcase_name}
# Return value : \$TCD::PASS or \$TCD::FAIL
# Summary      : $args->{testcase_description}
# Test Type    : $args->{testcase_type}
###############################################################################

sub $args->{testcase_name} {
    \$Test->description();
    my \$$args->{testcase_name} = << 'ET' ;
    {
        my \$runid = \$Test->get_testcase();
        my \$expected_output ;
        try {
$init_string $setup_code  
	      ##### Expected output #####
              \$expected_output = \"$expected_output\";

	      ### API CALL  ###
              my \$response = \$Api_Set_Obj_$suffix->$api_call
	      my \$execution_time = \$response->get_execution_time();
              \$Log->log("Time taken by the ZAPI::"
                       . "$args->{testcase_name}"
                       . "::\$execution_time");

              my \$raw_output = \$response->get_raw_output();

              \$Log->log("Expected Output Pattern : ".Dumper(\$expected_output));

	      my \$result = &comparatorXML( "console"    => \$raw_output,
					    "expected"   => \$expected_output,  
					    "attributes" => "true",  
					    "text"       => "true",  
					    "regex"      => "true" );  
 
	      if (\$result->[0]){  
                     \$Log->log(pass_result(\$runid));
$data_validation
$cleanup_code
              } else {       
                     \$Log->log("Actual output doesn't match with the expected output");
                     \$Log->log(fail_result(\$runid));
              }
		   
         }
         catch NACL::APISet::Exceptions::ResponseException with {
              my \$exception_object = shift;
              my \$response = \$exception_object->get_response_object();
              my \$raw_output = \$response->get_raw_output();
              \$Log->log("Actual Output : ".Dumper(\$raw_output));
              \$Log->log("Expected Output Pattern : ".Dumper(\$expected_output));
              \$Log->log("Exception details: ".Dumper(\$exception_object));
              \$Log->log(fail_result(\$runid));
         };
     }
ET
     my \$result = subtest(\$$args->{testcase_name},
               -runid, \$Test->get_testcase(), -onexit, sub { }, "--");
     return \&_get_failure_type(\$result);
}

POSITIVE_INTERP

}


################################################################################
# Subroutine  : _gen_negative_tc_template_zapi                                 #
#                                                                              #
# Description : This method writes negative test case template code            #
#                                                                              #
# Inputs      : 'code_type' - Valid values: 'dfm' or 'ontap'                   #
#               'file_handle' - File handle of the file to which content is    #
#                               being appended                                 #
#               'zapi_stack' - Data structure containing info of zapi calls    #
#               'index'  - Variable telling the depth of zapi calls            #
#               'testcase_name' - String containing the test case name         #
#               'testcase_description' - String containing tc description      #
#               'testcase_type' - String describing the type of test case      #
#               'expected_output' - String containing the regular expression   #
#                                   that validates the output of the string    #
#               'setup_hash' - Reference to a hash containing calls to the     #
#                               setup and cleanup                              #
#                                                                              #
################################################################################


sub _gen_negative_tc_template_zapi {

        my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};

        # Take the code type
        my $code_type = $args->{"code_type"};

        # Suffix for the APISet object name
        my $suffix;
        if( $code_type =~ /dfm/i ) {
                $suffix = "DFM";

        } else {
                $suffix = "FILER";

        }

        # Take the zapis to be executed
        my @zapi_string = ( );
        my @array = @{$args->{"zapi_stack"}};
        my $index = $args->{"index"};

        my %setup_hash =  %{$args->{"setup_hash"}};

        # String for variable initialization
        my $init_string = "";

        for( my $i = $index-1; $i > 0; $i-- ) {

           $array[$i] =~ s/(Val|Bool)://g;
           $init_string .= "\t\t" . $array[$i] . "\n";

        }

        if( $init_string ne "" ) {

                $init_string = "\n\t\t#####  Variable Initialisation  #####\n". $init_string;
        }

        # Variable for api call
	$array[0] =~ s/(Val|Bool)://g; 
        my $api_call = $array[0];

        # API Call for setup
        my $setup_code = "";
        if( defined $setup_hash{"setup"}) {

            $setup_code =  "\n\t\t#########  Setup Code  ########\n".
			   "\t\t## Please uncomment and use  ##\n" . 
                           "\t\t#\$Api_Set_Obj_$suffix->" . $setup_hash{"setup"}.
                           "\n\n\t\t#################################\n";
	}

        #API Call for clean up
        my $cleanup_code = "";
        if( defined $setup_hash{"cleanup"}) {

            $cleanup_code = "\t\t\t#### Code for cleanup goes here ####\n".
                            "\t\t\ttry {\n\n".
                            "\t\t\t\t# Cleanup APISet call\n\n".
			    "\t\t\t\t# Please initialize the variables and uncomment the cleanup code.".
                            "\t\t\t\t#\$Api_Set_Obj_$suffix->" . $setup_hash{"cleanup"} . "\n\n".
                            "\t\t\t} catch  NACL::APISet::Exceptions::ResponseException with {\n".
                            "\t\t\t\tmy \$exception_object = shift;\n".
                            "\t\t\t\t\$response = \$exception_object->get_response_object();\n".
                            "\t\t\t\t\$Log->log(\"Cleanup step Failed.\");\n".
                            "\t\t\t\t\$Log->log(\"Exception details: \",Dumper(\$exception_object));\n".
                            "\t\t\t}\n";
        }

        # Take the expected output
	my %error_output = %{$args->{"expected_output"}};

print $fh <<"NEGATIVE_INTERP"  or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : $args->{testcase_name}
# Return value : \$TCD::PASS or \$TCD::FAIL
# Summary      : $args->{testcase_description}
# Test Type    : $args->{testcase_type}
###############################################################################

sub $args->{testcase_name} {
    \$Test->description();
    my \$$args->{testcase_name} = << 'ET' ;
    {
        my \$runid = \$Test->get_testcase();
        my \$response ;
        try {
$init_string  $setup_code  
	      #### API CALL ####
              \$response = \$Api_Set_Obj_$suffix->$api_call

              my \$raw_output = \$response->get_raw_output();

              \$Log->log("Actual Output : ".Dumper(\$raw_output));
              \$Log->log('ERROR : The exit status of the command is 0.');
              \$Log->log(fail_result(\$runid));

         }
        catch NACL::APISet::Exceptions::ResponseException with {
              my \$exception_object = shift;
              \$response = \$exception_object->get_response_object();
              my \$error_number = \$response->get_error_number();
              my \$error_reason = \$response->get_error_reason();
	      my \$raw_output = \$response->get_raw_output();
 
	      \$Log->log("Actual Output : ".Dumper(\$raw_output));
	      \$Log->log('Expected error reason: $error_output{error_reason}');
	      \$Log->log('Expected error number: $error_output{error_number}');

              if (\$error_number =~ /$error_output{'error_number'}/si && \$error_reason =~ /$error_output{'error_reason'}/ ){
                     \$Log->log(pass_result(\$runid));

$cleanup_code
                 } else {
                     \$Log->log(fail_result(\$runid));
                 }
        };
      
    }
ET
    my \$result = subtest(\$$args->{testcase_name},
               -runid, \$Test->get_testcase(), -onexit, sub { }, "--");
    return \&_get_failure_type(\$result);
}

NEGATIVE_INTERP


}


################################################################################
# Subroutine  : _gen_positive_tc_template                                      #
#                                                                              #
# Description : This method writes positive test case template code            #
#                                                                              #
################################################################################


sub _gen_positive_tc_template{

	my ( $self, $args ) = @_;
        my $fh = $args->{"file_handle"};
	#print "\n$args->{command}\n";
	
	#print Dumper ($args->{parameter_set});getc();

	my @variable_list;
        my @user_message = ("#>>>>>>USER NOTE>>>> Test case Setup/Cleanup to be handled","#>>>>>>USER NOTE>>>> Expected Output to be validated or added depending on the test case requirements");
	my $param_string = "";
	my $init_string = "";

	while(my ($param_key, $param_value) = each(%{$args->{parameter_set}})) {
		$param_value =~ s/-/_/g;
		$param_string.= " \"$param_key\" => $param_value,"; 
		push (@variable_list,$param_value) if ($param_value =~ /^\$.*/i);
	}	
	$param_string =~ s/,$//;
	#print $param_string;getc();

	foreach my $variable(@variable_list){
        	$init_string.= " my $variable = \"\"\;\n\t      ";
	}
	#print $init_string;getc();

	#print "\nTC NAME : $args->{testcase_name} \n TC DES : $args->{testcase_description} \n EXP_OP :  $args->{expected_output} \n TC TYPE :  $args->{testcase_type} \n FH :$args->{file_handle} \n";
        #print "PARAM SET : ". Dumper ($args->{parameter_set}). "\n";getc();

	my @test_description = split ("\n",$args->{testcase_description});

print $fh <<"POSITIVE_INTERP"  or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : $args->{testcase_name}
# Return value : \$TCD::PASS or \$TCD::FAIL
# Summary      : $args->{testcase_description}
# Test Type    : $args->{testcase_type}
###############################################################################

sub $args->{testcase_name} {
    \$Test->description();
    my \$$args->{testcase_name} = << 'ET' ;
    {
        my \$runid = \$Test->get_testcase();
        my \$response ;
        try {    
	      # %%         $args->{config_status}               % #

	      # %%         Variable Initialisation              % #	
	      $init_string	
              \$response = \$Api_Set_Obj_DFM->$args->{command}($param_string);
              my \$raw_output = \$response->get_raw_output();
	      
	      \$Log->log("Actual Output : ".Dumper(\$raw_output));	
	      \$Log->log('Expected Output Pattern : $args->{expected_output}');

              if ( \$raw_output =~ /$args->{expected_output}/si){
                     \$Log->log(pass_result(\$runid));   
              } else {
                     \$Log->log(fail_result(\$runid));
              }   
         }
         catch NACL::APISet::Exceptions::ResponseException with {
              my \$exception_object = shift;
	      \$response = \$exception_object->get_response_object();
              my \$raw_output = \$response->get_raw_output();
              \$Log->log("Actual Output : ".Dumper(\$raw_output));
              \$Log->log('Expected Output Pattern : $args->{expected_output}');
              \$Log->log("Exception details: ",Dumper(\$exception_object));
              \$Log->log(fail_result(\$runid));
         };
     }
ET
     my \$result = subtest(\$$args->{testcase_name},
               -runid, \$Test->get_testcase(), -onexit, sub { }, "--");
     return \&_get_failure_type(\$result);
}

POSITIVE_INTERP
}


################################################################################
# Subroutine  : gen_negative_test_api_template                                 #
#                                                                              #
# Description : This method writes negative test case template code            #
#               (For commands execute via apiset method calls)                 #
################################################################################

sub gen_negative_test_api_template{ 

	my ( $self, %args ) = @_;
        my $fh = $args{"file_handle"};
	my @variable_list;
	my @user_message = ("#>>>>>>USER NOTE>>>> Test case Setup/Cleanup to be handled","#>>>>>>USER NOTE>>>> Expected Output to be validated or added depending on the test case requirements");
        my $param_string = "";
	my $init_string = "";
	
	while(my ($param_key, $param_value) = each(%{$args{parameter_set}})) {
		$param_value =~ s/-/_/g;
               $param_string.= " \"$param_key\" => $param_value,";
		push (@variable_list,$param_value) if ($param_value =~ /^\$.*/i);
        }
        $param_string =~ s/,$//;
	
	#print "\nVariable List : @variable_list\n";getc();
	unless ($args{testcase_type} eq 'negative_required_parameter_object_exists'){
	
		foreach my $string_variable(@variable_list){
			$string_variable =~ s/\$//;
			$string_variable =~ s/\-|\s+/_/g;
                        $init_string.= " my \$$string_variable = \"test_$string_variable\"\;\n\t      ";
		}

	}else{
		#foreach my $empty_variable(@{$args{"variable_list"}}){
		foreach my $empty_variable(@variable_list){
                	$init_string.= " my $empty_variable = \"\"\;\n\t      ";
        	}
	}
        #print $init_string;getc();

	my @test_description = split ("\n",$args{testcase_description});


	print $fh <<"NEGATIVE_INTERP"  or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : $args{testcase_name}
# Return value : \$TCD::PASS or \$TCD::FAIL
# Summary      : $args{testcase_description}
# Test Type    : $args{testcase_type}
###############################################################################

sub $args{testcase_name} {
    \$Test->description();	
    my \$$args{testcase_name} = << 'ET' ;
    {
        my \$runid = \$Test->get_testcase();
        my \$response ;       
        try { 
              # %%         $args{config_status}               % #

              # %%         Variable Initialisation              % #
              $init_string 
              \$response = \$Api_Set_Obj_DFM->$args{command}($param_string) ;
	      my \$raw_output = \$response->get_raw_output();
   
              \$Log->log("Actual Output : ".Dumper(\$raw_output));
              \$Log->log('Expected Output Pattern : $args{expected_output}');	
	      \$Log->log('ERROR : The exit status of the command is 0.');	
              \$Log->log(fail_result(\$runid));
 
         }
        catch NACL::APISet::Exceptions::ResponseException with {
              my \$exception_object = shift;
              \$response = \$exception_object->get_response_object();
              my \$raw_output = \$response->get_raw_output();
              \$Log->log("Actual Output : ".Dumper(\$raw_output));	
              \$Log->log('Expected Output Pattern : $args{expected_output}');
  
 
              if (\$raw_output =~ /$args{expected_output}/si ){
                     \$Log->log(pass_result(\$runid));   
                 } else {
                     \$Log->log(fail_result(\$runid));
                 }                   
              };
    }
ET
    my \$result = subtest(\$$args{testcase_name},
               -runid, \$Test->get_testcase(), -onexit, sub { }, "--");
    return \&_get_failure_type(\$result);
}

NEGATIVE_INTERP

}

################################################################################
# Subroutine  : gen_negative_test_raw_template                                 #
#                                                                              #
# Description : This method writes negative test case template code            #
#               (For commands execute via execute_raw_command  method )        #
################################################################################

sub gen_negative_test_raw_template{

	my ( $self, %args ) = @_;
        my $fh = $args{"file_handle"};
	
	my @variable_list;
	my @user_message = ("#>>>>>>USER NOTE>>>> Test case Setup/Cleanup to be handled","#>>>>>>USER NOTE>>>> Expected Output to be validated or added depending on the test case requirements");
        my $param_string = "";
	my $init_string = "";

        while(my ($param_key, $param_value) = each(%{$args{parameter_set}})) {
                $param_string.= " $param_key => $param_value,";
		push (@variable_list,$param_value) if ($param_value =~ /^\$.*/i);
        }
        $param_string =~ s/,$//;

	foreach my $string_variable(@variable_list){
                        $string_variable =~ s/\$//;
                        $string_variable =~ s/\-|\s+/_/g;
                        $init_string.= " my \$$string_variable = \"test_$string_variable\"\;\n\t";
        }

        my @test_description = split ("\n",$args{testcase_description});

	#print "\n Test Case Name : $args{testcase_name} / Raw Command : $args{raw_command} / Test Description : $args{testcase_description} / Test Type : $args{testcase_type}\n";getc();

	print $fh <<"NEGATIVE_RAW_INTERP"  or warn "\n\tWARNING : $!\n";

###############################################################################
# Function     : $args{testcase_name}
# Return value : \$TCD::PASS or \$TCD::FAIL
# Summary      : $args{testcase_description}
# Test Type    : $args{testcase_type}
###############################################################################

sub $args{testcase_name} {
    \$Test->description();
    my \$$args{testcase_name} = << 'ET' ;
    {
	# %%         $args{config_status}               % #

	# %%         Variable Initialisation              % #
        $init_string
        my \$runid = \$Test->get_testcase();
        my \$raw_output = \$Api_Set_Obj_DFM->execute_raw_command(
                               command => "$args{raw_command}"
                            );
        \$Log->log("Actual Output : ".Dumper(\$raw_output));
        \$Log->log('Expected Output Pattern : $args{expected_output}');
 
        if (\$raw_output =~ /$args{expected_output}/si){
                \$Log->log(pass_result(\$runid));  
        } else {
                \$Log->log(fail_result(\$runid));
        }                        
   }
ET
   my \$result = subtest(\$$args{testcase_name},
               -runid, \$Test->get_testcase(), -onexit, sub { }, "--");
    return \&_get_failure_type(\$result);
}

NEGATIVE_RAW_INTERP

}

1;

