# File Name     :       Setup_Cleanup.pm
#
# Summary       :       The package parses zapidoc files and returns setup,cleanup,type for the api's .
#
# Creation Date :       24th November 2010
#
# Author        :       srinivas.prabhu@netapp.com, dl-sm-nacl-dev@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#

#
# Package Declaration
#

package CodeModules::ZAPI::SetupCleanup;

#
# Compiler Directives
#

use strict;
use warnings;

#
# Perl Modules
#

use Data::Dumper;
use XML::Parser;
use XML::Simple;

#
# User Specific Modules
#
#use ParseEngines::CDEFParseEngine;
use ParseEngines::CommonParseEngine;

#
# Class Objectives
#
my $zapi_parser = ParseEngines::CommonParseEngine->new;

# Constructor Definition

sub new {
    my $class = shift;
    my $self  = {
        'filenames' => undef,
        'logger'    => undef,
    };
    bless $self, $class;
    return $self;
} ## end sub new

#############################################################################################
# $cleanup_mapping_hash and $setup_mapping_hash contains the list of the keyword and their
# eqvivalent mapping to their opposities
#############################################################################################
my $cleanup_mapping_hash = {
    'create' => [ "destroy", "delete" ],
    'add'    => [ "remove",  "delete" ],
    'commit' => "rollback",
    'start'   => [ "stop",    "cancel", "abort" ],
    'enable'  => "disable",
    'up'      => "down",
    'online'  => "offline",
    'swap'    => "unswap",
    'suspend' => "resume",
    'bind'    => "unbind",
    'map'     => "unmap",
    'fail'    => "unfail",
    'inherit' => "disinherit",
    'modify'  => [ "destroy", "delete", "remove" ]
};

my $setup_mapping_hash = {
    'destroy'    => "create",
    'remove'     => "add",
    'rollback'   => "commit",
    'disable'    => "enable",
    'down'       => "up",
    'offline'    => "online",
    'unswap'     => "swap",
    'resume'     => "suspend",
    'unbind'     => "bind",
    'unmap'      => "map",
    'unfail'     => "fail",
    'remove'     => "add",
    'modify'     => [ "create", "add" ],
    'delete'     => [ "create", "add" ],
    'cancel'     => "start",
    'stop'       => "start",
    'abort'      => "start",
    'disinherit' => "inherit"
};

##########################################################################################
# Subroutine : get_setup_cleanup
# Description : Will find the 'type' of zapi and return the appropriate setup/cleanup api call
# Return     : Nested Hash
##########################################################################################
sub get_setup_cleanup {
    my ( $self,%input_args) = @_;
    my ( $zapi_type, $zapi_setup, $zapi_cleanup );

  #  _validate_input(%input_args);

    my $category_name = $input_args{'category_name'};
    my $parse_hash = $input_args{'zapi_data_ref'};

    # Get the list of zapi belonging to the category #
    my @cmd_list = $self->_parse_zapidata( $category_name,$parse_hash );

    # Declaring the Output Hash #
    my %api;
    my $array_ref = \@cmd_list;

    foreach my $val (@cmd_list) {
        if ( $val =~ /(-list|-info|-iter|-stats|-status)/ ) {
            $zapi_type    = "State";
            $zapi_setup   = undef;
            $zapi_cleanup = undef;
        } elsif ( $val
            =~ /(-create|-destroy|-add|-delete|-commit|-rollback|-remove|-modify|-start|-end|-stop|-down|-up|-enable|-disable|-online|-offline|-swap|-unswap|-resume|-suspend|-bind|-unbind|-map|-unmap|-fail|-unfail|-restart|-reboot|-cancel|-inherit|-disinherit)/
            )
        {
            $zapi_type = "Non-Passive";

            # Get the cleanup requirement of the zapi #
            $zapi_cleanup
                = $self->_decision_maker( $val, $array_ref, "cleanup" );

            # Get the setup requirement of the zapi #
            $zapi_setup = $self->_decision_maker( $val, $array_ref, "setup" );
        } elsif ( $val =~ /(-get|-set)/ ) {
            $zapi_type    = "Passive";
            $zapi_setup   = undef;
            $zapi_cleanup = undef;
        } else {
            $zapi_type    = "Unknown";
            $zapi_setup   = undef;
            $zapi_cleanup = undef;
        }

        $api{ ${val} }{'type'}    = $zapi_type;
        $api{ ${val} }{'setup'}   = $zapi_setup;
        $api{ ${val} }{'cleanup'} = $zapi_cleanup;
    } ## end foreach my $val (@cmd_list)

    return ( \%api );

}    #

################################################################################################
#
#
#################################################################################################
sub _decision_maker {

    my ( $self, $val, $ref_array, $type ) = @_;
    my ( $zapi_name, $key, $temp, $temp1 );
    my @cmd_array = @$ref_array;

   # Declaring a temporary hash which will hold the appropriate mapping hash #
    my %temp_hash;

    # Load the mapping hash into temporary hash based on input type #
    if    ( $type =~ /setup/ )   { %temp_hash = %$setup_mapping_hash; }
    elsif ( $type =~ /cleanup/ ) { %temp_hash = %$cleanup_mapping_hash; }

    foreach my $match ( keys %temp_hash ) {
        if ( $val =~ /-$match/ ) {
            $key = $match;
            if ( $val =~ /((.*?)-$key(.*?)$)/ ) {
                $temp  = $2;
                $temp1 = $3;
            }
            foreach my $arr (@cmd_array) {
                if ( $temp_hash{$key} =~ /ARRAY\(0x/ ) {
                    my $arr_ref = $temp_hash{$key};
                    my @tmp_arr = @$arr_ref;
                    foreach my $option (@tmp_arr) {
                        if ( $arr =~ /$temp-$option$temp1$/ ) {
                            $zapi_name = $arr;
                            last;
                        }
                    } ## end foreach my $option (@tmp_arr)
                } else {
                    if ( $arr =~ /$temp-$temp_hash{$key}$temp1$/ ) {
                        $zapi_name = $arr;
                        last;
                    }
                } ## end else [ if ( $temp_hash{$key} ...
            } ## end foreach my $arr (@cmd_array)
        } ## end if ( $val =~ /-$match/)
    } ## end foreach my $match ( keys %temp_hash)
    return $zapi_name;

}    #

###################################################################################################
#
#
#####################################################################################################

sub _parse_zapidata {

    my ( $self, $category_name, $hash_ref ) = @_;

    my @api_list;
    my %hash = %$hash_ref;
    delete $hash{'zapi_typedefs'};
    my $zapi_ref = $hash{'zapi_category'};
    my %temp     = %$zapi_ref;
   
   my $t = $temp{$category_name};
   my @temp_arr = @$t ;
   return @temp_arr;
} ## end sub _parse_zapidoc

1;
