# File Name     :       UnitTestZAPIGen.pm
#
# Summary       :       Contains methods to generate DFM ZAPI unit tests using NACL APISet method
#
# Creation Date :        2010
#
# Author        :       Deepthi.Shetty@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#       Summary: Main engine to generate test cases  
#       Date   : 
#       Author : Deepthi.Shetty@netapp.com   
#
# Copyright (c) 2011 NetApp, Inc.
# All rights reserved.
#

#
# Package Declaration
#

package CodeModules::ZAPI::UnitTestZAPIGen;

#
# Compiler Directives
#

use strict;
use warnings;

#
# Class Inheritance
#

use base qw(CodeModules::TestTemplateGen);

#
# Perl Modules
#

use Data::Dumper;
use Tie::IxHash;
use Tie::File;
use IO::File;

#
# User Specific Modules
#
use ParseEngines::CommonParseEngine;
use CodeModules::ZAPI::SetupCleanup;
use CodeModules::ZAPI::DataValidation;  

#
# Class Objectives
#

my $parse_obj =  ParseEngines::CommonParseEngine->new();

# Constructor Definition

sub new {
    my $class = shift;
    my $self  = {};
    bless $self, $class;
    return $self;
}

# Global variables
my %tc_list = ( );

################################################################################
# Subroutine  : gen_unit_tests                                                 #
#                                                                              #
# Description : This method generates DFM ZAPI unit tests using NACL APISet    #
#		        methods							                               #
#                                                                              #
# Returns     : Nothing                                                        #
#               Generates the unit test files                                  #
################################################################################


sub gen_unit_tests {

    my $self = shift;
    my $conf_hash = shift;

     # Variable initialization	
     my $code_type = $conf_hash->{'code_type'};	
     my $code_path = $conf_hash->{'code_path'}; 	
     my $file_path = $conf_hash->{'code'}->{$code_type}->{'file_path'};
     my $commands  = $conf_hash->{'code'}->{$code_type}->{'command'};
     my $interface = $conf_hash->{'code'}->{$code_type}->{'interface'};
     my $version   = $conf_hash->{'code'}->{$code_type}->{'version'}; 
     my $generate_test_type = $conf_hash->{'code'}->{$code_type}->{'test_type'};   

     # Take type and family from the user config
     my $type = $conf_hash->{'code'}->{$code_type}->{'type'};
     $type = lc($type); 

     tie my %test_case_num, "Tie::IxHash"; 	
     tie %tc_list, "Tie::IxHash";
     
     # Hash to hold the summary of the files. For every file the total number of positive and negatice test cases generated
     tie my %summary_hash, "Tie::IxHash"; 

     # Invoke the parser and collect all the information in the variable $zapi_parsed_hash
     # This will have four types of mapping:
     # 1. Categories and all the zapis under it     2. Error strings with error numbers
     # 3. Every zapi and all details of it          4. Every zapi typedef and it's details
 
     my $zapi_parsed_hash = $parse_obj->parse_file("file_path" => $file_path, "CommandName" => $commands , "interface" => $interface, "type" => $type);

    # Get the category and the commands for which test cases have to be generated. This information is sent as a hash.
    my $zapi_list= $self->_get_command_list( $commands, $zapi_parsed_hash, $type);	

    unless ( keys(%{$zapi_list}) ) {

        print "\nERROR: There are no valid ZAPIs specified to generate .thpl files";
        print "\n\nExiting..\n";

        exit(1);

    }
  
   ##### IMPORTANT !!!! ###### 
   # Change "ontap" to "filer"
   $type = "filer" if( $type =~ /ontap/ );
 
   ### Hold the total count for testcases ###  
   my ($total_pos, $total_neg, $total_tot);

    ################################### 
    ##   When the type is DFM        ##  
    ###################################  
    if( $type =~ /dfm/i ) {

	# Create .thpl files for every key in $zapi_list
	foreach my $zapi ( keys( %{$zapi_list} ) ) {  

	    ### Empty the tc list  ###
	    %tc_list = (); 

	    ### Hold count of testcases ###
	    my ($pos, $neg, $tot);  

	    ### Get the hash with error numbers and error descriptions ###  
	    my $error_desc = $self->_get_error_list( "zapi_list" => $zapi_list->{$zapi}, "parsed_structure" => $zapi_parsed_hash );  

	    # Form the file name 
	    my $thpl = $type . "_" . $zapi ;  
	    # Remove hyphens and convert to underscores  
	    $thpl =~ s/\-/_/;  

	    # Remove more than one occurence of 'dfm_' 
	    if( $thpl =~ /(dfm_){2,}/ ) {   
		$thpl =~ s/dfm_//;  
	    }

	    my ($success,$ut_fh,$file_name) = $self->create_test_file( $thpl, $code_path, $interface);  

	    if ($success){  

		$self->gen_test_template("file_handle" => $ut_fh ,"interface" => $interface, "zapi_list" => $zapi_list->{$zapi}, "zapi_error_description"=> $error_desc, "file_name" => $file_name, "version" => $version, "code_type" => $type ); 

		# Generate test cases 
	   	($pos, $neg) = $self->_gen_test_cases_zapi( "file_handle" => $ut_fh, "file_name" => $file_name, "code_type" => $type, "zapi_list" => $zapi_list->{$zapi}, "parsed_structure" => $zapi_parsed_hash, "code_path" => $code_path, "test_type" => $generate_test_type ); 

		$tot = $pos + $neg; 
		$total_pos += $pos; 
		$total_neg += $neg; 
		$total_tot += $tot;  

		print "----------------------------------------------------------------------------------------------------" . "\nGenerated $thpl.thpl with:\n";
		print "\nPositive  :  $pos"; 
		print "\nNegative  :  $neg"; 
		print "\n----------------";  
		print "\nTotal     :  $tot\n\n";  

	        # Update the summary hash 
		$summary_hash{"$thpl.thpl"}{"pos"} = $pos;
		$summary_hash{"$thpl.thpl"}{"neg"} = $neg;  
		$summary_hash{"$thpl.thpl"}{"tot"} = $tot; 

	    } # end of if success

	} # end of foreach $zapi  

     } else {
     ###################################
     ##   When the type is ONTAP      ##
     ###################################

       # Create an array of file handlers
       my @ontap_fh = ();
       
       foreach my $family ( keys( %{$zapi_list} ) ) {

	    # Create .thpl files for every key in $zapi_list for that family  
	    foreach my $zapi ( keys( %{$zapi_list->{$family}} ) ) {   

		### Empty the tc list  ###
                %tc_list = ();  

		### Hold count of testcases ### 
		my ($pos, $neg, $tot);  

		### Get the hash with error numbers and error descriptions ###  
		my $error_desc = $self->_get_error_list( "zapi_list" => $zapi_list->{$family}->{$zapi}, "parsed_structure" => $zapi_parsed_hash );  

		# Form the file name 
		my $thpl = $family . "_" . $zapi ; 
		# Remove hyphens and convert to underscores  
		$thpl =~ s/\-/_/; 

		my ($success,$ut_fh,$file_name) = $self->create_test_file( $thpl, $code_path, $interface);  

		if ($success){  

		    $self->gen_test_template("file_handle" => $ut_fh ,"interface" => $interface, "zapi_list" => $zapi_list->{$family}->{$zapi}, "zapi_error_description"=> $error_desc, "file_name" => $file_name, "version" => $version, "code_type" => $type, "family" => $family );

		    # Generate test cases 
		    ($pos, $neg) = $self->_gen_test_cases_zapi( "file_handle" => $ut_fh, "file_name" => $file_name, "code_type" => $type, "zapi_list" => $zapi_list->{$family}->{$zapi}, "parsed_structure" => $zapi_parsed_hash, "code_path" => $code_path, "test_type" => $generate_test_type );

		    $tot = $pos + $neg;
		    $total_pos += $pos; 
		    $total_neg += $neg;  
		    $total_tot += $tot;  

		    print "----------------------------------------------------------------------------------------------------" . "\nGenerated $thpl.thpl with:\n";   
		    print "\nPositive  :  $pos"; 
		    print "\nNegative  :  $neg";  
		    print "\n----------------";  
		    print "\nTotal     :  $tot\n\n";  

		    # Update the summary hash
                    $summary_hash{"$thpl.thpl"}{"pos"} = $pos;
                    $summary_hash{"$thpl.thpl"}{"neg"} = $neg;
		    $summary_hash{"$thpl.thpl"}{"tot"} = $tot;  

		} # end of if success  

	    } # end of foreach $zapi 

	} # end of foreach $family 

    } # end of else   

        #####################################################################
        # Generate the zapi xml files          #
        #####################################################################

        $self->_generate_zapi_cdefs ( "code_path" => $code_path, "file_path" => $file_path );


	print "\n****************************************************************";  
	print "\n*\t\t\t\t\t\t\t\t*";
	print "\n*\t\t\tS U M M A R Y\t\t\t\t*";  
	print "\n*\t\t\t\t\t\t\t\t*"; 
	print "\n************************************************************************\n\n"; 

	printf '%36s    %s   %s   %s', " Thpl file     ", "Positive", "Negative", "Total     ";   

	print "\n************************************************************************\n"; 

	my $file_count = 0; 
	foreach my $file ( keys( %summary_hash ) ) {  

	    $file_count++;
	    printf '%3d . %-30s    %6d    %6d    %6d ', $file_count, $file, $summary_hash{$file}{pos}, $summary_hash{$file}{neg}, $summary_hash{$file}{tot};  
	    printf "\n"; 

	}   
        print "\n\n***********************************************************************\n";  	
        #print "\n  TOTAL\t\t$total_pos\t$total_neg\t$total_tot \n";
        printf '%10s %26s    %6d    %6d    %6d ', "    ", "TOTAL", $total_pos, $total_neg, $total_tot; 
        print "\n***********************************************************************";

        print "\n\nGenerated test scripts can be found at PATH: \"$code_path\".\n " ;  

	
}

################################################################################
# Subroutine  : _gen_test_cases_zapi                                           #
#                                                                              #
# Description : This method generates test cases                               #
#                                                                              #
# Inputs      : File handle, File name, Zapi list, Parsed structure            #
#                                                                              #
# Returns     :                                                                #
#                                					       #
#                                                                              #
################################################################################


sub _gen_test_cases_zapi {

        my ( $self, %args ) = @_;

        my $ut_fh = $args{"file_handle"};
        my $file_name = $args{"file_name"};
	my $code_path = $args{"code_path"};
        my @zapis_list = @{ $args{"zapi_list"}};
        my $parsed_structure = $args{"parsed_structure"};
	my $type = $args{"code_type"};
        my $generate_test_type = $args{"test_type"};  

        my $zapi_command = $zapis_list[0];
       
	my ($positive_count, $negative_count) = (0,0); 
        ########################################
        # Generate positive test cases        ##
        ########################################

        $positive_count = $self->_gen_positive_zapi( "file_handle" => $ut_fh, "file_name" => $file_name, "code_type" => $type, "zapi_list" => \@zapis_list, "parsed_structure" => $parsed_structure ) if ( $generate_test_type =~ /positive|all/ ) ;

	########################################
        # Generate negative test cases        ##
        ########################################

	$negative_count = $self->_gen_negative_zapi( "file_handle" => $ut_fh, "file_name" => $file_name, "code_type" => $type, "zapi_list" => \@zapis_list, "parsed_structure" => $parsed_structure ) if ( $generate_test_type =~ /negative|all/ ) ;

	#####################################################################
        # Insert test case list and write failure type in the thpl file    ##
	#####################################################################

	my $new_ut_fh = $self->_insert_testcase_list( "file_handle" => $ut_fh, "file_name" => $file_name, "code_path" => $code_path, "tc_list" => \%tc_list );  
	$self->write_get_failure_type("file_handle" => $new_ut_fh);
        close($new_ut_fh);

	# Return the testcase count
	return ($positive_count, $negative_count);

}


#####################################
# Global Array and other variables  #
#####################################
my @array;
my $global_index = 0;

my %outer_count = ();
my %inner_count = ();

################################################################################
# Subroutine  : _gen_positive_zapi                                             #
#                                                                              #
# Description : This method generates positive test cases                      #
#                                                                              #
# Inputs      : File handle, File name, Zapi list, Parsed structure            #
#                                                                              #
# Returns     : The number of positive test cases generated                    #
#                                                                              #
#                                                                              #
################################################################################

sub _gen_positive_zapi {

        my ( $self, %args ) = @_;

        my $ut_fh = $args{"file_handle"};
        my $file_name = $args{"file_name"};
        my @zapis_list = @{ $args{"zapi_list"}};
        my $parsed_structure = $args{"parsed_structure"};
	my $code_type = $args{"code_type"};

        # Take the zapi strcuture
        my %zapi_main = %{$parsed_structure->{"zapis"}};
        my %zapi_typedefs = %{$parsed_structure->{"zapi_typedefs"}};

	# Variable to keep the count of the number of test cases
	my $test_case_count = 0;

	# Variable to store the test case description
	my $test_case_desc;

        # Variable to store the expected output for all commands
	my $expected_output;
		
	# Variable to hold the test case name
	my $test_case_name;
	
	# Hash reference to hold setup and/or cleanup
	my $setup_hash;

	# Variab;e to hold test case type
	my $test_type;
	my $value_type;

	# Flags for integer range and boolean range
	my ($integer_range_flag, $boolean_flag);
	
	 foreach my $zapi_command ( @zapis_list ) {

		    ##### If it is not an iter zapi ######
		    if( $zapi_command !~ /\-iter\-(start|next|end)/ ) {
 
			#################################################
			# 1. Only Required parameters at all levels        #
			#################################################

			$test_type = "positive_only_required";
			$value_type = "LOW_RANGE";

			# Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

			%outer_count = ();
			%inner_count = ();

			compute_stack( "parsed_structure" => $parsed_structure,
					"zapi_name" => $zapi_command,
					"code_type" => $code_type,
               				"test_type" => $test_type,
					"depth" => "0",
					"value_type" => $value_type,
				);
			_rearrange_stack (); 

			# Get the expected output 
			$expected_output = $self->get_expected_output_tags( "zapi_command" => $zapi_command,
									    "parsed_structure" => $parsed_structure,
									    "code_type" => $code_type,
									   );
	 
			# Make the call for finding setup and or cleanup
                        $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
      								"category"      => $zapi_main{$zapi_command}{'APIcategory'}, 
								"zapi_data_ref" => $parsed_structure,    
                                                               );

			$test_case_desc = "Test ZAPI with only required parameters";

                        # Check if integer or boolean values were included
                        my $stack_as_string = Dumper( \@array );

                        $integer_range_flag = ( $stack_as_string =~ /Val:/) ? 1 : 0;
                        $boolean_flag = ( $stack_as_string =~ /Bool:/ ) ? 1 : 0;

		        if( $integer_range_flag ) {
				$test_type .= "_order1";
				$test_case_desc .= ", with lower value range for integers";
			}

			$test_case_name = $zapi_command . "_" . $test_type;
			$test_case_name =~ s/\-/_/g;

			# Store in hash %tc_list
			$tc_list{$test_case_name} = $test_case_desc;	

			# Increase the count
			$test_case_count++;    
		
			# Function call to write contents into the file 
			 $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0 ); 


			if( $integer_range_flag ) {

			    ####################################################################################
			    # 1.a. With Only required parameters ( with other range value of integer/boolean)  #
			    ####################################################################################

		   	    $test_type = "positive_only_required";	
			    $value_type = "HIGH_RANGE";

                            # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
			    _rearrange_stack (); 

			    # Expected output and Setup/Cleanup hash remains the same

                            $test_case_name = $zapi_command . "_" . $test_type . "_order2";
                            $test_case_name =~ s/\-/_/g;

			    $test_case_desc = "Test ZAPI with only it's required parameters and give integers it's higher end of range value and boolean as false.";

			    # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

			    # Increase the count
                            $test_case_count++;     	
			    
			    # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0  );

			    ####################################################################################	
			    # 1.b. With Only required parameters (No range values explicitly given)            #
			    ####################################################################################

			    $test_type = "positive_only_required";
			    $value_type = "NO_VALUE";

			    # Empty the hash and reset global_top
			    @array = ( );
			    $global_index = 0;

			    %outer_count = ();
			    %inner_count = ();

			    compute_stack( "parsed_structure" => $parsed_structure,
						"zapi_name" => $zapi_command,
						"code_type" => $code_type,
						"test_type" => $test_type,
						"depth" => "0",
						"value_type" => $value_type,	
				);
			    _rearrange_stack (); 

			    # Expected output and Setup/Cleanup hash remains the same 
		
			    $test_case_name = $zapi_command . "_" . $test_type; 
			    $test_case_name =~ s/\-/_/g;

			    $test_case_desc = "Execute the ZAPI with only it's required parameters, with user given values for integers and boolean";

			    # Store in hash %tc_list
			    $tc_list{$test_case_name} = $test_case_desc;

			    # Increase the count
			    $test_case_count++;

			    # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0 );
			    
			}
			
			#################################################
			# 2. Required plus optional options             #
			#################################################

			$test_type = "positive_with_optional";
			$value_type = "LOW_RANGE";

			######### Check if there Are optional options to create the testcase  ########    
			if ( @{$zapi_main{$zapi_command}{'optional_inputs'}} ) {
			
                            # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
			    _rearrange_stack (); 

			    # Make the call for finding setp and/or cleanup
			    $setup_hash = $self->_get_setup_cleanup( 
								"zapi"          => $zapi_command,
								"test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
								"zapi_data_ref" => $parsed_structure, 
								);

                            $test_case_desc = "Test ZAPI with both required and optional parameters";
			    
                            # Check if integer or boolean values were included
                            $stack_as_string = Dumper( \@array );

                            $integer_range_flag = ( $stack_as_string =~ /Val:/) ? 1 : 0;
                            $boolean_flag = ( $stack_as_string =~ /Bool:/ ) ? 1 : 0;

                            if( $integer_range_flag ) {
                                $test_type .= "_order1";
                                $test_case_desc .= ", with lower value range for integers";
                            }

			    $test_case_name = $zapi_command."_" . $test_type;  
			    $test_case_name =~ s/\-/_/g;

			    # Increase the count
                            $test_case_count++;    

			    # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

			    # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0 );


                            if( $integer_range_flag ) {

                                ############################################################################################
                                # 2.a. With required and optional parameters ( with other range value of integer/boolean)  #
                                ############################################################################################

				$test_type = "positive_with_optional"; 
			        $value_type = "HIGH_RANGE";

                        	# Empty the hash and reset global_top
                        	@array = ( );
                        	$global_index = 0;

                        	%outer_count = ();
                        	%inner_count = ();

                        	compute_stack( "parsed_structure" => $parsed_structure,
                                        	"zapi_name" => $zapi_command,
                                        	"code_type" => $code_type,
                                        	"test_type" => $test_type,
                                        	"depth" => "0",
                                        	"value_type" => $value_type,
                                );
				_rearrange_stack ();

				# Expected output and Setup/Cleanup hash remains the same
				
				$test_case_name = $zapi_command. "_" . $test_type. "_order2";
                            	$test_case_name =~ s/\-/_/g;

                                # Increase the count
                                $test_case_count++;   

			        $test_case_desc = "Test ZAPI with required and optional parameters, with higher end value of range given to integers and boolean values are false";

				# Store in hash %tc_list
                                $tc_list{$test_case_name} = $test_case_desc;

                                # Function call to write contents into the file
				$self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0 );

				############################################################################################
                                # 2.b. With required and optional parameters (No range values explicitly given)            #
                                ############################################################################################

                                $test_type = "positive_with_optional";
                                $value_type = "NO_VALUE";

                                # Empty the hash and reset global_top
                                @array = ( );
                                $global_index = 0;

                                %outer_count = ();
                                %inner_count = ();

                                compute_stack( "parsed_structure" => $parsed_structure,
                                                "zapi_name" => $zapi_command,
                                                "code_type" => $code_type,
                                                "test_type" => $test_type,
                                                "depth" => "0",
                                                "value_type" => $value_type,
                                );
				_rearrange_stack (); 

                                # Expected output and Setup/Cleanup hash remains the same

                                $test_case_name = $zapi_command. "_" . $test_type;
                                $test_case_name =~ s/\-/_/g;

                                # Increase the count
                                $test_case_count++;

                                $test_case_desc = "Test ZAPI with required and optional parameters";

                                # Store in hash %tc_list
                                $tc_list{$test_case_name} = $test_case_desc;

                                # Function call to write contents into the file
                                $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 0 );


			    } # end of range check 
 
			} # end of check for if optional params exist     
			
		    } else {

			# Skip if the command is an 'iter-start' or 'iter-end'. These will be covered in 'iter-next' anyway.
			if( $zapi_command =~ /start/ or $zapi_command =~ /end/ ) {
				next;
			}

			#################################################
			#    3. Iter ZAPIs                              #
			#################################################

			$test_type = "positive_only_required";
			$value_type = "LOW_RANGE";

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
			_rearrange_stack (); 

			# Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
			for( my $i = 0; $i < $global_index; $i++ ) {   
				
				if( $array[$i] =~ /my \$tag0/ ) {
				    $array[$i] = " ";
				} elsif( $array[$i] =~ /my \$maximum0/ ) {
				    $array[$i] = "my \$maximum0 = \"\"; ";
				}
			}

                        $test_case_desc = "Test iter-next zapi, with iter-start as setup and iter-end as cleanup";

			# Variable that holds zapi name and modified for test case name etc
			my $temp; 

		        #### CHECK TYPE ####
 
    				$zapi_command =~ /(.*)\-iter\-.*/;
			        $temp = $1;
	
				# Form names for start, next and end iter ZAPIs
				my $zapi_start = $temp . "-iter-start";
				my $zapi_next = $temp . "-iter-next";
				my $zapi_end = $temp . "-iter-end";

				# Make the call for finding setp and/or cleanup
				$setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_start,
                                                                "test_type"     => "iter",
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                                );


                        # Get the expected output
                        $expected_output = $self->get_expected_output_tags( "zapi_command" => $zapi_command,
                                                                            "parsed_structure" => $parsed_structure,
                                                                            "code_type" => $code_type,
                                                                           );


         		# Check if integer or boolean values were included
                        my $stack_as_string = Dumper( \@array );

                        $integer_range_flag = ( $stack_as_string =~ /Val:/) ? 1 : 0;
                        $boolean_flag = ( $stack_as_string =~ /Bool:/ ) ? 1 : 0;

                        if( $integer_range_flag ) {
                            $test_type .= "_order1";
                            $test_case_desc .= ", with lower value range for integers";
                        }

			$test_case_name = $temp . "_iter" . "_". $test_type;
			$test_case_name =~ s/\-/_/g; 

			# Increase the count
                        $test_case_count++; 	

			# Store in hash %tc_list
                        $tc_list{$test_case_name} = $test_case_desc;

			# Function call to write contents into the file
			$self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );

                        if( $integer_range_flag  ) {


			    ###########################################################################
                            #    3a. Iter ZAPIs - with higher range for integer and false for boolean #
                            ###########################################################################

			    $test_type = "positive_only_required"; 
			    $value_type = "HIGH_RANGE";

                            # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
			    _rearrange_stack (); 

                            # Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
                            for( my $i = 0; $i < $global_index; $i++ ) {

                                    if( $array[$i] =~ /my \$tag0/ ) {
                                        $array[$i] = " ";
                                    } elsif( $array[$i] =~ /my \$maximum0/ ) {
                                        $array[$i] = "my \$maximum0 = \"\"; ";
                                    }
                            }

			    # Test case name
			    $test_case_name = $temp . "_iter" . "_". $test_type . "_order2";
			    $test_case_name =~ s/\-/_/g;

			    # Increase the count
                            $test_case_count++;

			    $test_case_desc = "Test iter-next ZAPIs, with iter-start as setup and iter-end as cleanup, with higher end range value for integers and false for booleans";

			    # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

                            # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );


			    ###########################################################################
                            #    3b. Iter ZAPIs (No explicit values given)                            #
                            ###########################################################################

                            $test_type = "positive_only_required";
                            $value_type = "NO_VALUE";

                            # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );

                            # Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
                            for( my $i = 0; $i < $global_index; $i++ ) {

                                    if( $array[$i] =~ /my \$tag0/ ) {
                                        $array[$i] = " ";
                                    } elsif( $array[$i] =~ /my \$maximum0/ ) {
                                        $array[$i] = "my \$maximum0 = \"\"; ";
                                    }
                            }

                            # Test case name
                            $test_case_name = $temp . "_iter" . "_". $test_type;
                            $test_case_name =~ s/\-/_/g;

                            # Increase the count
                            $test_case_count++;

                            $test_case_desc = "Test iter-next ZAPIs, with iter-start as setup and iter-end as cleanup";

                            # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

                            # Function call to write contents into the file
                            $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );


			} # end of if $integer_range_flag

		        #################################################
                        #    3c. Iter ZAPIs - with optional params      #
                        #################################################

                        ######### Check if there Are optional options to create the testcase  ########
                        if ( @{$zapi_main{$zapi_command}{'optional_inputs'}} ) {

			    $test_type = "positive_with_optional";
                            $value_type = "LOW_RANGE";
	
			    # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );

			    # Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
                            for( my $i = 0; $i < $global_index; $i++ ) {

                                    if( $array[$i] =~ /my \$tag0/ ) {
                                        $array[$i] = " ";
                                    } elsif( $array[$i] =~ /my \$maximum0/ ) {
                                        $array[$i] = "my \$maximum0 = \"\"; ";
                                    }
                            }

			    $test_case_desc = "Execute iter-next ZAPIs, with iter-start as setup and iter-end as cleanup, with optional parameters";

                            # Check if integer or boolean values were included
                            my $stack_as_string = Dumper( \@array );

                            $integer_range_flag = ( $stack_as_string =~ /Val:/) ? 1 : 0;
                            $boolean_flag = ( $stack_as_string =~ /Bool:/ ) ? 1 : 0;

                            if( $integer_range_flag ) {
                                $test_type .= "_order1";
                                $test_case_desc .= ", with lower value range for integers";
                            }

                            $test_case_name = $temp . "-iter" . "_". $test_type;
                            $test_case_name =~ s/\-/_/g;

                            # Increase the count
                            $test_case_count++;

                            # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

                            # Function call to write contents into the file
                            $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );


			    #######################################################################################
                            #    3d. Iter ZAPIs w/ optional - with higher range for integer and false for boolean #
                            #######################################################################################

			    if( $integer_range_flag  ) {
			
			   	$test_type = "positive_with_optional"; 
        		    	$value_type = "HIGH_RANGE";
				
                                # Empty the hash and reset global_top
                        	@array = ( );
                                $global_index = 0;

                                %outer_count = ();
                                %inner_count = ();

                                compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
		
			        # Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
                                for( my $i = 0; $i < $global_index; $i++ ) {

                                    if( $array[$i] =~ /my \$tag0/ ) {
                                        $array[$i] = " ";
                                    } elsif( $array[$i] =~ /my \$maximum0/ ) {
                                        $array[$i] = "my \$maximum0 = \"\"; ";
                                    }
                                }
	
                                $test_case_desc = "Test iter-next ZAPIs w/ optional params, with iter-start as setup and iter-end as cleanup, with higher end range value for integers and false for booleans";

			        $test_case_name = $temp . "-iter" . "_". $test_type . "_order2";
			        $test_case_name =~ s/\-/_/g; 

                                # Increase the count
                                $test_case_count++;

                                # Store in hash %tc_list
                                $tc_list{$test_case_name} = $test_case_desc;

                                # Function call to write contents into the file
                                $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );


				#######################################################################################
                                #    3e. Iter ZAPIs w/ optional (No explicit values given)                            #
                                #######################################################################################

                                $value_type = "NO_VALUE";

                                # Empty the hash and reset global_top
                                @array = ( );
                                $global_index = 0;

                                %outer_count = ();
                                %inner_count = ();

                                compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );

                                # Some modification to the array to remove $tag declaration amd make $maximum initialsed to null
                                for( my $i = 0; $i < $global_index; $i++ ) {

                                    if( $array[$i] =~ /my \$tag0/ ) {
                                        $array[$i] = " ";
                                    } elsif( $array[$i] =~ /my \$maximum0/ ) {
                                        $array[$i] = "my \$maximum0 = \"\"; ";
                                    }
                                }

                                $test_case_desc = "Test iter-next ZAPIs w/ optional params, with iter-start as setup and iter-end as cleanup, with user given values for inputs";

                                $test_case_name = $temp . "-iter" . "_". $test_type;
                                $test_case_name =~ s/\-/_/g;

                                # Increase the count
                                $test_case_count++;

                                # Store in hash %tc_list
                                $tc_list{$test_case_name} = $test_case_desc;

                                # Function call to write contents into the file
                                $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "positive", "expected_output" => $expected_output, "setup_hash" => $setup_hash, "iter_zapi" => 1 );

			    } # end of if integer
    
			} # end of check for if optional present
			
		    } # end of if-iter   
		}

	# Return the test case count
        return $test_case_count;	
}

################################################################################
# Subroutine  : _gen_negative_zapi                                             #
#                                                                              #
# Description : This method generates negative test cases                      #
#                                                                              #
# Inputs      : File handle, File name, Zapi list, Parsed structure            #
#                                                                              #
# Returns     : The number of negative test cases generated                    #
#                                                                              #
#                                                                              #
################################################################################

sub _gen_negative_zapi {

        my ( $self, %args ) = @_;

        my $ut_fh = $args{"file_handle"};
        my $file_name = $args{"file_name"};
        my @zapis_list = @{ $args{"zapi_list"}};
        my $parsed_structure = $args{"parsed_structure"};
        my $code_type = $args{"code_type"};

        # Take the zapi strcuture
        my %zapi_main = %{$parsed_structure->{"zapis"}};
        my %zapi_typedefs = %{$parsed_structure->{"zapi_typedefs"}};

	# Variable to hold the 'type'
        my $type;

	# Variable to hold the testcase's description
	my $test_case_desc;

        # Variable to keep the count of the number of test cases
        my $test_case_count = 0;

        # Variable to store the expected output for all commands
        my %expected_output;

        # Variable to hold the test case name
        my $test_case_name;

        # Hash reference to hold setup and/or cleanup
        my $setup_hash;
	
	# Variable to hold test type & value type
	my ( $test_type, $value_type); 

        # Flags for integer range and boolean range
        my ($integer_range_flag, $boolean_flag);

        foreach my $zapi_command ( @zapis_list ) {

                ##### If it is not an iter zapi ######
                if( $zapi_command !~ /\-iter/ ) {

	            #### Check if there are required parameters and optional parameters ####
		    if( @{$zapi_main{$zapi_command}{'required_inputs'}} && @{$zapi_main{$zapi_command}{'optional_inputs'}} ) {
	
                        #################################################
                        # 1. Missing required parameters at all levels        #
                        #################################################

			$test_type = "negative_missing_required";
			$value_type = "LOW_RANGE";

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );
			_rearrange_stack (); 


			# Check if integer or boolean values were included
                        my $stack_as_string = Dumper( \@array );

                        $integer_range_flag = ( $stack_as_string =~ /Val:/) ? 1 : 0;
                        $boolean_flag = ( $stack_as_string =~ /Boolean:/ ) ? 1 : 0;

                        %expected_output = $self->get_expected_error_output( "zapi_command" => $zapi_command,
                                                                             "parsed_structure" => $parsed_structure,
									     "test_type" => $test_type,
                                                                            );

                        # Make the call for finding setp and/or cleanup
                        $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                               );

			$test_case_desc = "Test ZAPI without giving it's required parameters"; 

			if ( $integer_range_flag ) {

				$test_type .= "_order1";
				$test_case_desc .= ", with lower value range for integers"; 
			}

			# Name of the test case
                        $test_case_name = $zapi_command . "_" . $test_type;  
			$test_case_name =~ s/\-/_/g;

			# Store in hash %tc_list
                        $tc_list{$test_case_name} = $test_case_desc;

                        # Increase the count
                        $test_case_count++;

                        # Function call to write contents into the file
                         $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

		        #################################################################################
                        # 1a. Missing required parameters at all levels - with the other level of range #
                        #################################################################################

			if( $integer_range_flag or $boolean_flag ) {
			
			    $test_type = "negative_missing_required";  
			    $value_type = "HIGH_RANGE";

                            # Empty the hash and reset global_top
                            @array = ( );
                            $global_index = 0;

                            %outer_count = ();
                            %inner_count = ();

                            compute_stack( "parsed_structure" => $parsed_structure,
                                           "zapi_name"        => $zapi_command,
                                           "code_type"        => $code_type,
                                           "test_type"        => $test_type,
                                           "depth"            => "0",
                                           "value_type"       => $value_type,
                                );
			    _rearrange_stack ();

			    # Expected error output and setup/cleanup hash will remain the same 

			    # Name of the test case
                            $test_case_name = $zapi_command . "_" . $test_type . "_order2";

                            $test_case_name =~ s/\-/_/g;

                            # Increase the count
                            $test_case_count++;

			    $test_case_desc = "Execute the ZAPI w/o required parameters, with higher end range value for integers and false for booleans";

			    # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

                            # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

			} # end of if integer or boolean value present
	
		      } # end of if check for required and optional params 

		      ###########################################################################
		      #  2. Value already exists                                       ##########
		      ###########################################################################
		      if( $zapi_command =~ /create|add|start|enable|online/ ) {

			$test_type = "negative_already_exists"; 
			$value_type = "LOW_RANGE"; 

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name"       => $zapi_command,
                                        "code_type"       => $code_type,
                                        "test_type"       => $test_type,
                                        "depth"           => "0",
                                        "value_type"      => $value_type,
                                );

			%expected_output = $self->get_expected_error_output( "zapi_command"     => $zapi_command,
                                                                             "parsed_structure" => $parsed_structure,
                                                                             "test_type"        => $test_type,
                                                                            );

                        # Make the call for finding setp and/or cleanup
                        $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                               );

                        # Name of the test case
                        $test_case_name = $zapi_command . "_" . $test_type;  

                        $test_case_name =~ s/\-/_/g;

			$test_case_desc = "Negative case where the object already exists";

			# Store in hash %tc_list
                        $tc_list{$test_case_name} = $test_case_desc;

                        # Increase the count
                        $test_case_count++;

                        # Function call to write contents into the file
			$self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

		      } # end of if create|add|start|enable|online

		      ###########################################################################
                      #  3. Value does not exist                                       ##########
                      ###########################################################################

                      elsif ( $zapi_command =~ /destroy|delete|remove|stop|disable|offline/i ) {  

			$test_type = "negative_does_not_exist";

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name"       => $zapi_command,
                                        "code_type"       => $code_type,
                                        "test_type"       => $test_type,
                                        "depth"           => "0",
                                        "value_type"      => $value_type,
                                );
			_rearrange_stack (); 

                        %expected_output = $self->get_expected_error_output( "zapi_command"     => $zapi_command,
                                                                             "parsed_structure" => $parsed_structure,
                                                                             "test_type"        => $test_type,
                                                                            );

                        # Make the call for finding setp and/or cleanup
                        $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                               );

                        # Name of the test case
                        $test_case_name = $zapi_command . "_". $test_type;   

                        $test_case_name =~ s/\-/_/g;

                        # Increase the count
			$test_case_count++;

			$test_case_desc = "Negative case where the object does not exist";

			# Store in hash %tc_list
                        $tc_list{$test_case_name} = $test_case_desc;

                        # Function call to write contents into the file
			$self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

		      } # end of if destroy|delete|remove|stop|disable|offline      

		    } # end of iter check

		      #################################################################
		      ##  4. Integer - out of range                               #####
		      #################################################################
	
			$test_type = "negative_out_of_range";
			$value_type = "OUT_OF_RANGE_LOW";

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name"       => $zapi_command,
                                        "code_type"       => $code_type,
                                        "test_type"       => $test_type,
                                        "depth"           => "0",
                                        "value_type"      => $value_type,
                                );
			_rearrange_stack ();  

			# Check if integer or boolean values were included
                        my $stack_as_string = Dumper( \@array );

                        $integer_range_flag = ( $stack_as_string =~ /Val:/i) ? 1 : 0;

			# If integer value exists
			if( $integer_range_flag ) {

                            %expected_output = $self->get_expected_error_output( "zapi_command"     => $zapi_command,
                                                                                 "parsed_structure" => $parsed_structure,
                                                                                 "test_type"        => $test_type,
                                                                                );

                            # Make the call for finding setp and/or cleanup
                            $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                               );

                            # Name of the test case
                            $test_case_name = $zapi_command . "_" . $test_type. "_order1";   

                            $test_case_name =~ s/\-/_/g;

                            # Increase the count
                            $test_case_count++;

                            $test_case_desc = "Negative case where integers are given value below their lower end range";

			    # Store in hash %tc_list
                            $tc_list{$test_case_name} = $test_case_desc;

                            # Function call to write contents into the file
			    $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );
	
			   ########## Other invalid range  ###################

				$value_type = "OUT_OF_RANGE_HIGH";

                                # Empty the hash and reset global_top
                        	@array = ( );
                        	$global_index = 0;

                        	%outer_count = ();
                        	%inner_count = ();

                        	compute_stack( "parsed_structure" => $parsed_structure,
                                               "zapi_name"        => $zapi_command,
                                               "code_type"        => $code_type,
                                               "test_type"        => $test_type,
                                               "depth"            => "0",
                                               "value_type"       => $value_type,
                                );
				_rearrange_stack ();  

				# Setup & clean up and expected output remain the same

				# Name of the test case
				$test_case_name = $zapi_command . "_" . $test_type. "_order2";

                         	$test_case_name =~ s/\-/_/g;

                          	# Increase the count
                          	$test_case_count++;

                          	$test_case_desc = "Negative case where integers are given value above their higher end range";

				# Store in hash %tc_list
                        	$tc_list{$test_case_name} = $test_case_desc;

                            	# Function call to write contents into the file
			        $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

             	          #################################################################
                          ##  5. Invalid value - string for integer                   #####
                          #################################################################

			$test_type = "negative_invalid_integer"; 
			$value_type = "INTEGER_INVALID"; 

                        # Empty the hash and reset global_top
                        @array = ( );
                        $global_index = 0;

                        %outer_count = ();
                        %inner_count = ();

                        compute_stack( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $zapi_command,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => "0",
                                        "value_type" => $value_type,
                                );


                        # Setup & clean up and expected output remain the same

                        # Name of the test case
                        $test_case_name = $zapi_command . "_" . $test_type . "_value";   

                         $test_case_name =~ s/\-/_/g;

                          # Increase the count
                          $test_case_count++;

                          $test_case_desc = "Negative case with invalid (string) values given to integers";

			  # Store in hash %tc_list
                          $tc_list{$test_case_name} = $test_case_desc;

			  # Function call to write contents into the file
			  $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );

			} # end of if integer check
		
	        ###############################################################
		#### 6. Invalid Value - No value is given                ######
		###############################################################

		if( @{$zapi_main{$zapi_command}{'required_inputs'}} ) {   

			$test_type = "negative_no_value"; 
	
			@array = ();
			$global_index = 0;
			
			$array[0] = $zapi_command . "( ";
			$array[0] =~ s/-/_/g; 
	
			my @inputs = @{$zapi_main{$zapi_command}{'required_inputs'}}; 

			foreach my $input ( @inputs ) {

			    my $type = $zapi_main{$zapi_command}{$input}{'type'};

			    $input =~ s/IN//;
			    $array[0] .= "\"". $input. "\" => ";
			
			    if( $type =~ /\[/ ) {
				$array[0] .= " \[ \"\", \], ";
			    } elsif ( $type !~ /(string|boolean|integer)/i ) {  
				$array[0] .= "\{ \}, ";
			    } else  {  
				$array[0] .= " \"\", ";
			    } 

			}	
			$array[0] .= " ); ";

                        # Get the expected output 
			%expected_output = $self->get_expected_error_output( "zapi_command"     => $zapi_command,
                                                                             "parsed_structure" => $parsed_structure,
                                                                             "test_type"        => $test_type,
                                                                            );

                            # Make the call for finding setp and/or cleanup
                            $setup_hash = $self->_get_setup_cleanup(
                                                                "zapi"          => $zapi_command,
                                                                "test_type"     => $test_type,
                                                                "category"      => $zapi_main{$zapi_command}{'APIcategory'},
                                                                "zapi_data_ref" => $parsed_structure,
                                                               );

                        # Name of the test case
                        $test_case_name = $zapi_command . "_". $test_type; 

                         $test_case_name =~ s/\-/_/g;

                          # Increase the count
                          $test_case_count++;

                          $test_case_desc = "Negative case where no value is given to parameters";

			  # Store in hash %tc_list
                          $tc_list{$test_case_name} = $test_case_desc;

                          # Function call to write contents into the file
		   	  $self->gen_test_case_template( 'interface' => "zapi", 'code_type' => $code_type, 'file_handle' => $ut_fh, 'zapi_stack' => \@array, 'index' => $global_index, 'testcase_name' => $test_case_name, 'testcase_description' => $test_case_desc, 'testcase_type' => "negative", "expected_output" => \%expected_output, "setup_hash" => $setup_hash );
 

		} # end of if required params exist
	
		############################################################
		####   7. No license to run the ZAPI            ############
		############################################################	

		if( $zapi_main{$zapi_command}{'APIlicense'} =~ /unnecessary/ ) {

		     my @all_inputs = @{$zapi_main{$zapi_command}{'required_inputs'}};
		     push( @all_inputs, @{$zapi_main{$zapi_command}{'optional_inputs'}} ); 

		}
	} # end of for loop for each zapi command  

        # Return the test case count
        return $test_case_count;
}



####################################################################################################
#
# Compute stack recursion 
#
# Inputs: 'zapi_name' - Name of the zapi/typedef  
#         'parsed_structure' - Reference to the parsed hash of zapi files   
#         'test_type' - String containing the type of test case   
#         'value_flag' - Takes value "0" or "1" or "-1". 
#                         LOW_RANGE 0 : For Integer: Lower range, for Boolean: 0 (false) is taken
#                         HIGH_RANGE 1 : For Integer: higher range, for Boolean: 1 (true) is taken
#                         OUT_OF_RANGE_LOW 2 : Out of range values for integers - below lower end 
#                         OUT_OF_RANGE_HIGH 3 : Out of range values for integers - above higher end
#                         INTEGER_INVALID 4 : Invalid values (string val) for integers
#                         NO_VALUE 5 : No value is hardcoded.
#         'options' - Takes values "only required", "only optional" and "with optional". Decides what options must be taken as input
#         'code_type' - Value of the code type. Possible values are "dfm" or "ontap"
# 
# Notes: Whenever the parameter is an integer and contains the substring '-id', no value from range is added
# 
####################################################################################################

	     
sub compute_stack {

    my ( %args ) = @_;

    my $command = $args{"zapi_name"};
    my $parsed_structure = $args{"parsed_structure"};
    my $code_type = $args{"code_type"};
    my $test_type = $args{"test_type"};
    my $depth = $args{"depth"};
    my $value_type = $args{"value_type"};

    my %zapi_structure;

    # Some corrections might need to be made
    $command =~ s/\.//;

    # Deciding to use zapi or typedef
    if( $depth == 0 ) {
        %zapi_structure = %{$parsed_structure->{'zapis'}};

    } else {
        %zapi_structure = %{$parsed_structure->{'zapi_typedefs'}};
    }

    my @all_inputs = ();

    if( $depth == 0 ) {

        if( $test_type =~ /positive_only_required|negative_already_exists|negative_does_not_exist|negative_out_of_range/ ) {

                @all_inputs = @{$zapi_structure{$command}{'required_inputs'}};

        } elsif( $test_type =~ /positive_with_optional/ ) {

                @all_inputs = @{$zapi_structure{$command}{'required_inputs'}};
                push( @all_inputs, @{$zapi_structure{$command}{'optional_inputs'}} );

        }

    } elsif( $depth < 3 ) {

        $command = $code_type . ":" .$command;
        if( $test_type =~ /positive_only_required/ ) {
 
                @all_inputs = @{$zapi_structure{$command}{'required_options'}};

        } elsif( $test_type =~ /positive_with_optional/ ) {

                @all_inputs = @{$zapi_structure{$command}{'required_options'}};
                push( @all_inputs, @{$zapi_structure{$command}{'optional_options'}} );

        }

    } else {
        $command = $code_type . ":" .$command;
        @all_inputs = @{$zapi_structure{$command}{'required_options'}};
    }

    # Important!!
    my $level = $global_index++;

    # The main APISet call
    if( $level == 0 ) {

        $array[$level] = $command . " ( ";

	$array[$level] =~ s/-/_/g;

        foreach my $input ( @all_inputs ) {

           my $input_type = $zapi_structure{$command}{$input}{'type'};

	   my $range = ( defined $zapi_structure{$command}{$input}{'range'}) ? ($zapi_structure{$command}{$input}{'range'}) : ( "" );

           $input =~ s/IN//;
           $array[$level] .= "\"". $input . "\" => ";

           my $input_k = $input;
           $input_k =~ s/-/_/g;

           # Check array or not
           if( $input_type =~ /\[\s*\]/ ) {

                if( $input_type =~ /(integer|string)/i ) {

                    $input_type = $1;

                    ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                    $array[$level] .= "\\\@". $input_k . $inner_count{$input_k} . ", ";

                    ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);
                    ( exists $inner_count{$input_type} ) ? ( $inner_count{$input_type}++ ) : ($inner_count{$input_type} = 0 );
                    ( exists $outer_count{$input_type} ) ? ( $outer_count{$input_type}++ ) : ($outer_count{$input_type} = 0 );

                    $array[$global_index++] = "my \@". $input_k . $outer_count{$input_k} . " = ( \$". $input_type . $inner_count{$input_type} . ", );";
                    $array[$global_index++] = "my \$". $input_type . $outer_count{$input_type} . " = \"". $input_type . $outer_count{$input_type} . "\";  ";

                } else {

                    $input_type =~ s/\[\s*\]//;
                    $input_type =~ s/\s*,\s*optional//;
		    $input_type =~ s/\s*,\s*validate//;

                    my $input_type_k = $input_type;
                    $input_type_k =~ s/-/_/g;

                    ( exists $inner_count{$input_type_k} ) ? ( $inner_count{$input_type_k}++ ) : ( $inner_count{$input_type_k} = 0 );
                    $array[$level] .= "\\\@". $input_type_k . $inner_count{$input_type_k} . ", ";

                    ( exists $outer_count{$input_type_k} ) ? ( $outer_count{$input_type_k}++ ) : ($outer_count{$input_type_k} = 0 );

                    # Start declaration here
                    my $temp_index = $global_index;
                    $array[$temp_index] = "my \@". $input_type_k . $outer_count{$input_type_k} . " = ( { ";

                        compute_stack ( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $input_type,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => $depth + 1,
                                        "value_type" => $value_type,
                              );

                    $array[$temp_index] .= " }, );   ";
                }

           } else {
           # Not array

                if( $input_type =~ /integer/i ) {

                    if( $input !~ /id/ && $range =~ /\.\./ && $value_type !~ /NO_VALUE/ ) {

                        my $temp = $range;
                        $temp =~ /(.*)\.\.(.*)/;

                        my $low = $1;
                        my $high = $2;

                        # Add comment about the Range value
                        $array[$global_index++] = "# \"$input\" accepts range: $low to $high ";  

                        if( $value_type =~ /LOW_RANGE/ ) {
                                $temp = $low;
                        } elsif( $value_type =~ /HIGH_RANGE/ ) {
                                $temp = $high;
                        }elsif( $value_type =~ /OUT_OF_RANGE_LOW/ ) {
                                $temp = $low-1;
                        } elsif( $value_type =~ /OUT_OF_RANGE_HIGH/ ) {
                                $temp = $high+1;
                        } elsif( $value_type =~ /INTEGER_INVALID/ ) {
                                $temp = "abcd";
                        }

                        $array[$level] .= "\"Val:". $temp . "\", ";

                    } else {

                        ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                        $array[$level] .= "\$". $input_k . $inner_count{$input_k} . ", ";

                        ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);
                        $array[$global_index++] = "my \$". $input_k . $outer_count{$input_k} . " = \"\";  ";

                    }

                } elsif ( $input_type =~ /string/i ) {

                        ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                        $array[$level] .= "\$". $input_k . $inner_count{$input_k} . ", ";

                        ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);
                        $array[$global_index++] = "my \$". $input_k . $outer_count{$input_k} . " = \"". $input_k . $outer_count{$input_k} . "\";  ";

                } elsif ( $input_type =~ /boolean/i ) {

                        my $value ;
                        ( $value_type =~ /HIGH_RANGE/  ) ? ( $value = "1" ) : ( $value = "0" );


                        $array[$level] .= "\"Bool:". $value . "\", ";

                } else {

                        ## TYPEDEFS ##
                        $input_type =~ s/\s*,\s*optional//;
			$input_type =~ s/\s*,\s*validate//;

                        my $input_type_k = $input_type;
                        $input_type_k =~ s/-/_/g;

                        ( exists $inner_count{$input_type_k} ) ? ( $inner_count{$input_type_k}++ ) : ( $inner_count{$input_type_k} = 0 );
                        $array[$level] .= "\\\%" . $input_type_k . $inner_count{$input_type_k} . ", ";

                        ( exists $outer_count{$input_type_k} ) ? ( $outer_count{$input_type_k}++ ) : ($outer_count{$input_type_k} = 0);

                        # Start declaration here
                        my $temp_index = $global_index;
                        $array[$temp_index] = "my \%". $input_type_k . $outer_count{$input_type_k} . " = ( ";

                        compute_stack ( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $input_type,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => $depth + 1,
                                        "value_type" => $value_type,
                              );

                        $array[$temp_index] .= ");   ";
                }

           } # end of array-or-not check

        } # end of foreach $input

    } # end of if $level
    else {

        #$array[$level] .= " { ";

        foreach my $input ( @all_inputs ) {

           my $input_type = $zapi_structure{$command}{$input}{'type'};

	   my $range = $zapi_structure{$command}{$input}{'range'};

           $input =~ s/ELEMENT//;
           $array[$level] .= "\"". $input . "\" => ";

	   my $input_k = $input;
           $input_k =~ s/-/_/g;

           # Check array or not
           if( $input_type =~ /\[\s*\]/ ) {

                if( $input_type =~ /(integer|string)/i ) {

                    $input_type = $1;

                    ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                    $array[$level] .= "\\\@". $input_k . $inner_count{$input_k} . ", ";

                    ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);
                    ( exists $inner_count{$input_type} ) ? ( $inner_count{$input_type}++ ) : ($inner_count{$input_type} = 0 );
                    ( exists $outer_count{$input_type} ) ? ( $outer_count{$input_type}++ ) : ($outer_count{$input_type} = 0 );

                    $array[$global_index++] = "my \@". $input_k . $outer_count{$input_k} . " = ( \$". $input_type . $inner_count{$input_type} . ", );";
                    $array[$global_index++] = "my \$". $input_type . $outer_count{$input_type} . " = \"". $input_type . $outer_count{$input_type} . "\";  ";

                } else {

                    $input_type =~ s/\[\s*\]//;
                    $input_type =~ s/\s*,\s*optional//;
		    $input_type =~ s/\s*,\s*validate//;

                    my $input_type_k = $input_type;
                    $input_type_k =~ s/-/_/g;

                    ( exists $inner_count{$input_type_k} ) ? ( $inner_count{$input_type_k}++ ) : ( $inner_count{$input_type_k} = 0 );
                    $array[$level] .= "\\\@". $input_type_k . $inner_count{$input_type_k} . ", ";

                    ( exists $outer_count{$input_type_k} ) ? ( $outer_count{$input_type_k}++ ) : ($outer_count{$input_type_k} = 0 );

                    # Start declaration here
                    my $temp_index = $global_index;
                    $array[$temp_index] = "my \@". $input_type_k . $outer_count{$input_type_k} . " = ( {";

                        compute_stack ( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $input_type,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => $depth + 1,
                                        "value_type" => $value_type,
                              );

                    $array[$temp_index] .= " }, );   ";
                }


           } else {
           # Not array

                if( $input_type =~ /integer/i ) {

                    if( $input !~ /id/ && $range =~ /\.\./ && $value_type !~ /NO_VALUE/ ) {

                        my $temp = $range;
                        $temp =~ /(.*)\.\.(.*)/;

                        my $low = $1;
                        my $high = $2;

                        # Add comment about the Range value
                        $array[$global_index++] = "# \"$input\" accepts Range: $low to $high ";

                        if( $value_type =~ /LOW_RANGE/ ) {
                                $temp = $low;
                        } elsif( $value_type =~ /HIGH_RANGE/ ) {
                                $temp = $high;
                        }elsif( $value_type =~ /OUT_OF_RANGE_LOW/ ) {
                                $temp = --$low;
                        } elsif( $value_type =~ /OUT_OF_RANGE_HIGH/ ) {
                                $temp = ++$high;
                        } elsif( $value_type =~ /INTEGER_INVALID/ ) {
                                $temp = "abcd";
                        }

                        $array[$level] .= "\"Val:". $temp . "\", ";

                    } else {

                        ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                        $array[$level] .= "\$". $input_k . $inner_count{$input_k} . ", ";

                        ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);
                        $array[$global_index++] = "my \$". $input_k . $outer_count{$input_k} . " = \"\";  ";

                    }

                } elsif ( $input_type =~ /string/i ) {

                        ( exists $inner_count{$input_k} ) ? ( $inner_count{$input_k}++ ) : ( $inner_count{$input_k} = 0 );
                        $array[$level] .= "\$". $input_k . $inner_count{$input_k} . ", ";

                        ( exists $outer_count{$input_k} ) ? ( $outer_count{$input_k}++ ) : ($outer_count{$input_k} = 0);

                        $array[$global_index++] = "my \$". $input_k . $outer_count{$input_k} . " = \"". $input_k . $outer_count{$input_k} . "\";  ";

                } elsif ( $input_type =~ /boolean/i ) {

                        my $value ;
                        ( $value_type =~ /HIGH_RANGE/ ) ? ( $value = "1" ) : ( $value = "0" );

                        $array[$level] .= "\"Bool:". $value . "\", ";

                } else {

                        ## TYPEDEFS ##
                        $input_type =~ s/\s*,\s*optional//;
			$input_type =~ s/\s*,\s*validate//;

                        my $input_type_k = $input_type;
                        $input_type_k =~ s/-/_/g;

                        ( exists $inner_count{$input_type_k} ) ? ( $inner_count{$input_type_k}++ ) : ( $inner_count{$input_type_k} = 0 );
                        $array[$level] .= "\\\%" . $input_type_k . $inner_count{$input_type_k} . ", ";

                        ( exists $outer_count{$input_type_k} ) ? ( $outer_count{$input_type_k}++ ) : ($outer_count{$input_type_k} = 0);

                        # Start declaration here
                        my $temp_index = $global_index;
                        $array[$temp_index] = "my \%". $input_type_k . $outer_count{$input_type_k}.  " = ( ";

                        compute_stack ( "parsed_structure" => $parsed_structure,
                                        "zapi_name" => $input_type,
                                        "code_type" => $code_type,
                                        "test_type" => $test_type,
                                        "depth" => $depth + 1,
                                        "value_type" => $value_type,
                              );

                        $array[$temp_index] .= ");   ";
                }

           } # end of array-or-not check

        } # end of foreach $input

        #$array[$level] .= " } ";

    } # end of level check

    if ( $level == 0) {
        $array[0] .= " );  ";

    }

}

################################################################################
# Subroutine  : _rearrange_stack                                               #
#                                                                              #
# Description : Rearranges values to put comments in the beginning and         #
#		declaration slater     					       #
#                                                                              #
# Inputs      : None                                                           #
#                                                                              #
# Returns     : None                                                           #
#                                                                              #
################################################################################

sub _rearrange_stack {

	my ( $i, $j, $k);
 	my @arr_comments = ();
	my @arr_declarations = ();

	for( $i = 0, $j=0, $k=0; $i < $global_index; $i++ ) { 

	    if( $array[$i] =~ /#.*/ ) { 

		$arr_comments[$j++] = $array[$i];

	    } else {  
	
		$arr_declarations[$k++] =$array[$i];

	    }
	} 

	@array = ();  
	
	@array = @arr_declarations; 
	push( @array, @arr_comments ); 
	
}   

################################################################################
# Subroutine  : _get_setup_cleanup                                             #
#                                                                              #
# Description : This method computes the setup and/or cleanup required         # 
#                                                                              #
# Inputs      : Takes a hash with the following keys:                          #
#               'zapi'  -  Contains the zapi which is being unit tested        #
#               'test_type' - Contains a string which describes the type of    #
#                             test case being generated. Valid values are:     #
#                             1. positive_only_required                        #
#                             2. positive_with_optional                        #
#                             3. negative_missing_required                     #
#                             4. negative_already_exists                       #
#                             5. negative_does_not_exist                       #
#                             6. negative_out_of_range                         #
#                             7. negative_no_value                             #  
#			      8. negative_no_license                           #  
#			      9. negative_wrong_family                         #
#			     10. iter                                          #  
#               'category' - The category of the zapi                          #
#               'zapi_data_ref' - Reference to the DS of CPE                   #
#                                                                              #
# Returns     : A hash reference with the following keys:                      #
#               'setup' - Contains a string which is the APISet call for setup #
#               'cleanup' - Contains a string which is the APISet call for     #  
#                           cleanup                                            #
#                                                                              #
################################################################################
 
sub _get_setup_cleanup {

	my ( $self, %args ) = @_;

	my $zapi = $args{"zapi"};  
	my $test_type = $args{"test_type"};  
	my $zapi_category = $args{"category"};
	my $parsed_structure = $args{"zapi_data_ref"};

	# Create an object of setup/cleanup class
	my $parser = CodeModules::ZAPI::SetupCleanup->new;

	# Hash to collect data 
	my %data_hash = %{$parser->get_setup_cleanup(   "category_name" => $zapi_category,
							"zapi_data_ref" => $parsed_structure,
						    )}; 

	# Hash to return
	my %return_hash = ( );

	# Strings that hold the API call for setup and cleanup and DATA VALIDATION part  
	my ($setup_call, $cleanup_call, $dv); 

	# String to hold the now zapi name. i.e. setup/cleanup API
	my $now_zapi;
 
	if( $test_type =~ /(positive_only_required)|(positive_with_optional)/ ) {

		# If the zapi -s non-passive, only then call setup cleanup  
		if( $data_hash{$zapi}{'type'} =~ /Non\-Passive/ ) {

                    if( defined $data_hash{$zapi}{'setup'} ) {   
			
			$now_zapi = $data_hash{$zapi}{'setup'};
		
			my @req_inputs = @{$parsed_structure->{'zapis'}{$now_zapi}{'required_inputs'}};

			$setup_call = $now_zapi . "( ";
			$setup_call =~ s/\-/_/g;

			foreach my $input ( @req_inputs ) {

			    if( $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){ 
			        my $input_k = $input; 
				$input_k =~ s/\-/_/g;	
				if( $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'} =~ /\[\s*\]/ ) {  
					$setup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
				} else {

					$setup_call .= "\"$input\" => \$" . $input_k  . "0". ", "; 
				} 
			    
			    } else {
				
				my $input_k = $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'};
				$input_k =~ s/\-/_/g;
				if( $input_k =~ /\[\s*\]/ ) {  

					$setup_call .= "\"$input\" => \\\@" . $input_k . "0". ", ";
				} else {
					$setup_call .= "\"$input\" => \\\%" . $input_k . "0". ", "; 
				}
			    } 
			}
			$setup_call .= " );";
			$setup_call =~ s/IN//g;   
			$setup_call =~ s/\[//; $setup_call =~ s/\]//;
	
		    }
		    if( defined $data_hash{$zapi}{'cleanup'} ) {  

			$now_zapi = $data_hash{$zapi}{'cleanup'};

			my @req_inputs = @{$parsed_structure->{'zapis'}{$now_zapi}{'required_inputs'}};

			$cleanup_call = $now_zapi . "( ";
			$cleanup_call =~ s/\-/_/g;  		

			foreach my $input ( @req_inputs ) {   

			    if( $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){  
				my $input_k = $input;
                                $input_k =~ s/\-/_/g;
                                if( $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                        $cleanup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                                } else {

                                        $cleanup_call .= "\"$input\" => \$" . $input_k  . "0". ", ";
                                }

			    } else {
				
                                my $input_k = $parsed_structure->{'zapis'}{$now_zapi}{$input}{'type'};  
				$input_k =~ s/\-/_/g;   
                                if( $input_k =~ /\[\s*\]/ ) {

                                        $cleanup_call .= "\"$input\" => \\\\@" . $input_k . "0". ", ";
                                } else {
                                        $cleanup_call .= "\"$input\" => \\\\%" . $input_k . "0". ", ";
                                }
			    }
			}
			$cleanup_call .= " );";  
			$cleanup_call =~ s/IN//g;	
			$cleanup_call =~ s/\[//; $cleanup_call =~ s/\]//;  
		    } 	

		    # Get data validation part  
		    $dv = $self->_get_data_validation(
                                                                "zapi_command" => $zapi,
                                                                "parsed_structure" => $parsed_structure,
                                                            );

		} # end of if non-Passive   

	} elsif ( $test_type =~ /(negative_missing_required)|(negative_out_of_range)|(negative_no_value)/ ) {

	    # No setup/clean up required - as the test case is bound to fail anyway

	} elsif ( $test_type =~ /negative_already_exists/ ) {

	    $setup_call = "";
            $cleanup_call = "";

	    my @req_inputs = @{$parsed_structure->{'zapis'}{$zapi}{'required_inputs'}};

	    # Same command for setup
	    $setup_call = $zapi . "( ";

	    $setup_call =~ s/\-/_/g;

            foreach my $input ( @req_inputs ) {

                if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){
   	              my $input_k = $input;
                      $input_k =~ s/\-/_/g;
                      if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $setup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                      } else {

                                $setup_call .= "\"$input\" => \$" . $input_k  . "0". ", ";
                      }

                } else {

                      my $input_k = $parsed_structure->{'zapis'}{$zapi}{$input}{'type'};
                      $input_k =~ s/\-/_/g;
                      if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $setup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                      } else {

                                $setup_call .= "\"$input\" => \\\%" . $input_k  . "0". ", ";
                      }
                }
            }
            $setup_call .= " );";
            $setup_call =~ s/IN//g;
	    $setup_call =~ s/\[//; $setup_call =~ s/\]//;  

	    if( defined $data_hash{$zapi}{'cleanup'} ) {

                $cleanup_call = $data_hash{$zapi}{'cleanup'}. "( ";
                $cleanup_call =~ s/\-/_/g;

                foreach my $input ( @req_inputs ) {

                    if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){
                        my $input_k = $input;
                        $input_k =~ s/\-/_/g;
                        if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $cleanup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                        } else {

                                $cleanup_call .= "\"$input\" => \$" . $input_k  . "0". ", ";
                        }

                    } else {

                        my $input_k = $parsed_structure->{'zapis'}{$zapi}{$input}{'type'};
                        $input_k =~ s/\-/_/g;
                        if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $cleanup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                        } else {

                                $cleanup_call .= "\"$input\" => \\\%" . $input_k  . "0". ", ";
                        }

                    }
                }
                $cleanup_call .= " );";
                $cleanup_call =~ s/IN//g;
		$cleanup_call =~ s/\[//; $cleanup_call =~ s/\]//;  
            }


	} elsif ( $test_type =~ /negative_does_not_exist/ ) {

		# None
	
	} elsif ( $test_type =~ /negative_no_license/ ) {  

	       #TBD

	} elsif ( $test_type =~ /negative_wrong_family/ ) {

		# TBD  

	} elsif ( $test_type =~ /iter/ ) {  

	    $setup_call = "";
            $cleanup_call = "";

            my @req_inputs = @{$parsed_structure->{'zapis'}{$zapi}{'required_inputs'}};

            # Same command for setup
            $setup_call = $zapi . "( ";

            $setup_call =~ s/\-/_/g;

            foreach my $input ( @req_inputs ) {

                if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){
                      my $input_k = $input;
                      $input_k =~ s/\-/_/g;
                      if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $setup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                      } else {

                                $setup_call .= "\"$input\" => \$" . $input_k  . "0". ", ";
                      }

                } else {

                      my $input_k = $parsed_structure->{'zapis'}{$zapi}{$input}{'type'};
                      $input_k =~ s/\-/_/g;
                      if( $parsed_structure->{'zapis'}{$zapi}{$input}{'type'} =~ /\[\s*\]/ ) {
                                $setup_call .= "\"$input\" => \\\@" . $input_k  . "0". ", ";
                      } else {

                                $setup_call .= "\"$input\" => \\\%" . $input_k  . "0". ", ";
                      }
                }
            }
            $setup_call .= " );";
            $setup_call =~ s/IN//g;
            $setup_call =~ s/\[//; $setup_call =~ s/\]//;

	    # For cleanup
	    $zapi =~ s/(.*)\-start/$1/;
	    $zapi .= "-end";

	    $cleanup_call = $zapi . "( );  "; 
	        
	}

	# Return  
	$return_hash{'setup'} = $setup_call;    
        $return_hash{'cleanup'} = $cleanup_call;  	
	$return_hash{'data_validate'} = $dv;  
	return \%return_hash;  

}

################################################################################
# Subroutine  : get_expected_output_tags                                       #
#                                                                              #
# Description : This method computes the expected output tags for a zapi       #
#                                                                              #
# Inputs      : Zapi command, Parsed structure                                 #
#                                                                              #
# Returns     : A string which contains the regular expression format for an   #
#               XML structure. The XML structure in question is the output     #
#               of the Zapi.                                                   #  
#                                                                              #
################################################################################

my $arrayflag = 0;

sub get_expected_output_tags {

        my ( $self, %args ) = @_;

        my $zapiname =  $args{"zapi_command"};
        my $hash_ref = $args{"parsed_structure"};
        my $zapitype = $args{"code_type"};

        my $outputxml;
        my $finalhashref;
        my %zapihash;

        my %globalzapihash    = %{ $hash_ref->{"zapis"} };
        my %globaltypedefhash = %{ $hash_ref->{"zapi_typedefs"} };

        if ( exists( $globalzapihash{$zapiname} ) ) {

                %zapihash = %{ $globalzapihash{$zapiname} };
        } else {

                print "please specify a proper zapi name\n";
                return;
        }

        my @requiredoutputarray = @{ ( $zapihash{"required_outputs"} ) };

        if (@requiredoutputarray) {

                foreach (@requiredoutputarray) {

                    my $convertedvar = $_;
                    if ( $convertedvar =~ /OUT/ ) {
                        $convertedvar =~ s/OUT//g;
                     }
                    if ( $convertedvar =~ /ELEMENT/ ) {
                        $convertedvar =~ s/ELEMENT//g;
                    }
                    if ( exists( $zapihash{$_} ) ) {

                        my %hash = %{ $zapihash{$_} };
                        if ( $hash{"type"} =~ /integer|string|boolean/ ) {
			    if($& =~ /integer/) {
				$finalhashref->{$convertedvar} = $hash{"type"}.$hash{"range"};  
			    } else {  
                                $finalhashref->{$convertedvar} = $hash{"type"};
			    }  

                        } else {

                                $finalhashref->{$convertedvar} = parsesubzapi( $hash_ref,$hash{"type"}, $zapitype);
                        }
                    }
                }
        } else {

                $outputxml = "<results status=\\\"passed\\\"></results>";
                return $outputxml;
        }

        my %finalhash = %$finalhashref;

        while ( my ( $key, $value ) = each(%finalhash) ) {
                next if (!defined $finalhash{$key} );
            $outputxml = processxml( $key, $value,$outputxml  );
            $key       = processkey($key,);
            $outputxml = $outputxml . "<\/$key>";


        }

        #$outputxml =~ s/\{([a-zA-Z0-9\-\\\+\.\*<>\/\[\]\|]*)\}/$1/mg;  
        #$outputxml =~ s/\{([a-zA-Z0-9\-\\\+\.\*<>\/\[\]\|]*)\}/$1/mg;  

        $outputxml = "<results status=\\\"passed\\\">".$outputxml."</results>";

        return $outputxml;

}

################################################################################
# Subroutine  : parsesubzapi                                                   #
#                                                                              #
# Description : Parses through the zapidoc hash                                #
#                                                                              #
#                                                                              #
# Returns :   Hash containing required output of the given ZAPI                #
#                                                                              #
################################################################################

sub parsesubzapi {

    my ($self,%args) = @_;
    my $hashref = shift;

    my $uniquetype = shift;
    my $zapitype   = shift;

    my $subhashref;
    my $count;
    my %globalzapihash    = %{ $hashref->{"zapis"} };
    my %globaltypedefhash = %{ $hashref->{"zapi_typedefs"} };
    my $arrayinfo         = $uniquetype;
    if ( $uniquetype =~ /\[\]/ ) {

        $uniquetype =~ s/[\[\]]//g;
    }

    if ( $zapitype eq "dfm" ) {

        $uniquetype = "dfm:" . $uniquetype;
    } else {
        $uniquetype = "filer:" . $uniquetype;
    }

    if ( exists( $globaltypedefhash{$uniquetype} ) ) {

        my %typehash        = %{ $globaltypedefhash{$uniquetype} };
        my @reqdoptionarray = @{ $typehash{"required_options"} };

        if (@reqdoptionarray) {
            foreach (@reqdoptionarray) {  #foreach element of the reqd options
                my $convertedvar = $_;
                if ( $convertedvar =~ /OUT/ ) {
                    $convertedvar =~ s/OUT//g;
                }
                if ( $convertedvar =~ /ELEMENT/ ) {
                    $convertedvar =~ s/ELEMENT//g;
                }
                my %subtypehash = %{ $typehash{$_} };
                if ( $subtypehash{"type"} =~ /integer|string|boolean/ ) {
                    if ( $subtypehash{"type"} =~ /integer/ ) {
                        $subhashref->{$arrayinfo}->{$convertedvar}
                            = $subtypehash{"type"} . $subtypehash{"range"};
                    } else {
                        $subhashref->{$arrayinfo}->{$convertedvar}
                            = $subtypehash{"type"};
                    }
                } else {
                    $count++;
                    if ( $count <= 4 ) {
                        $subhashref->{$arrayinfo}->{$convertedvar}
                            = parsesubzapi( $hashref,$subtypehash{"type"}, $zapitype );

                    } else {
                        return $subhashref;
                    }
                }
            }

        } else {
            $subhashref->{$arrayinfo} = undef;

        }
    }
    return $subhashref;
}


################################################################################
# Subroutine  : processxml                                                         #
#                                                                              #
# Description : Parses through the data structure generated by parsesubzapi    #
#               function  and generates xml                                    #
#                                                                              #
#                                                                              #
# Returns :   Scalar variable containing the regular expression for expected   #
#             xml output                                                       #
#                                                                              #
################################################################################

sub processxml {

    my $count;
    my $key   = shift;
    my $value = shift;
    my $outputxml = shift;

    $key = processkey($key);

    $outputxml = $outputxml . "<$key>";

    if ( $value =~ /string|boolean|integer/ ) {

        if ( $value =~ /integer(.*)/ ) {
            my $range = $1;
            if ( $range =~ /2\^/ ) {
                $outputxml = $outputxml . "\\\\d\+";

	    } elsif ( $range =~ /(\d+)..(\d+)/ ) { 

		my $min = $1;  
		my $max = $2;  
	 	my $var = "(\\\\d\+)(??{\\\$1&gt;=$min &amp;&amp;\\\$1&lt;=$max?&apos;&apos;:&apos;(?!)&apos;})\\\$";  

		$outputxml = $outputxml .$var;  

            } else {
                $outputxml = $outputxml . "\\\\d\+";
            }
        } elsif ( $value =~ /string/ ) {
            $outputxml = $outputxml . "\.\*";
        } else {
            $outputxml = $outputxml . "\(true|false|0|1\)";
        }

    }

    else {

        $count++;

        if ( $count <= 5 ) {

            my %subxml = %$value;    #subhash

            while ( my ( $key, $value ) = each(%subxml) ) {
                        next if (!defined $subxml{$key});
                if ( $key =~ /\[\]/ ) {
                    #$outputxml = $outputxml . "{";
                    $outputxml = processxml( $key, $value ,$outputxml);
                    $arrayflag = 1;
                } else {
                    $outputxml = processxml( $key, $value,$outputxml );
                }

                $key = processkey($key);

                $outputxml = $outputxml . "<\/$key>";
                if ( $arrayflag == 1 ) {
                    #$outputxml = $outputxml . "}";
                    $arrayflag = 0;
                }
            }


        }
    }

    return $outputxml;
}

################################################################################
# Subroutine  : processkey                                                         #
#                                                                              #
# Description : Removes the tags OUT and ELEMENT from the option names         #
#                                                                              #
#                                                                              #
# Returns :   Processed option names                                           #
#                                                                              #
################################################################################

sub processkey {
    my $key = shift;

    if ( $key =~ /OUT/ ) {
        $key =~ s/OUT//g;

    } elsif ( $key =~ /\[\]/ ) {
        $key =~ s/\[\]//g;

        #$arrayflag  = 1;
    } elsif ( $key =~ /ELEMENT/ ) {
        $key =~ s/ELEMENT//g;

    }
    return $key;
}

################################################################################
# Subroutine  : get_expected_error_output                                      #
#                                                                              #
# Description : This method computes the expected error ouput for a zapi       #
#                                                                              #
# Inputs      : Takes a hash with the following keys:                          #
#               'zapi_command' - Contains the zapi which is being unit tested  #
#               'test_type' - Contains a string which describes the type of    #
#                             test case being generated. Valid values are:     #
#                             1. negative_missing_required                     #
#                             2. negative_already_exists                       #
#                             3. negative_does_not_exist                       #
#                             4. negative_out_of_range                         #
#                             5. negative_no_value                             #
#                             6. negative_no_license                           #
#                             7. negative_wrong_family                         #
#               'parsed_structure' - Reference to the DS of CPE                #
#                                                                              #
# Returns    :  A hash with the following keys:                                #
#               'error_number' - Contains regex witht the possible error nos   #
#               'error_message' - Contains regex for the possible error reason #
#                                                                              #
################################################################################ 

sub get_expected_error_output {

        my ( $self, %args ) = @_;

        my $zapi_command =  $args{"zapi_command"};
        my $parsed_structure = $args{"parsed_structure"};
	my $test_type = $args{"test_type"};
	
        # String which will contain the expected output in XML format
	my %error_output = ( );

        # Other variables used
	my ($error_message, $error_number); 

	if ( $test_type =~ /negative_missing_required/ ) {

	   # Missing required inputs
	   $error_message = "Missing (|required\\s*)input(|\(s\))"; 
	   $error_number = "13006";

        } elsif ( $test_type =~ /(negative_already_exists)|(negative_does_not_exist)|(negative_out_of_range)|(negative_no_value)/ ) {

           $error_message = ".*";  
	   $error_number = "";

	   my @arr = @{$self->_get_errors_codes_of_zapi( "zapi_command" => $zapi_command, "parsed_structure" => $parsed_structure) };

	   # Make an OR list of these errors
	   foreach my $err ( @arr ) {

		next unless $err; 
		$error_number .= $err . "|";
	   }
	   $error_number =~ s/\|$//;


        } elsif ( $test_type =~ /negative_no_license/ ) {


        } elsif ( $test_type =~ /negative_wrong_family/ ) {



        }

	$error_output{'error_reason'} = $error_message;
	$error_output{'error_number'} = $error_number;

	# Return the error message
	return %error_output;
}

################################################################################
# Subroutine  : _get_data_validation                                           #
#                                                                              #
# Description : This method gives all possible errors thrown by a zapi         #
#                                                                              #
# Inputs      : Takes a hash with the following keys:                          #
#               'zapi_command' - Contains the zapi which is being unit tested  #
#               'parsed_structure' - Reference to the DS of CPE                #
#                                                                              #
# Returns    :  A reference to an array containing the error codes             #
#                                                                              #
################################################################################

sub _get_data_validation {

        my ( $self, %args ) = @_;

        my $zapi_command =  $args{"zapi_command"};
        my $parsed_structure = $args{"parsed_structure"};

	# String that holds the APISet call to ZAPI list
	my ($dv_call1, $dv_call2, $dv_call3, $dv_check);

	# Hash for data validation
	my %data_validate = ();


        # Create an object of DataValidation class
        my $parser = CodeModules::ZAPI::DataValidation->new;   

	$dv_call2 = $parser->data_validate(
			 		  "category" => $parsed_structure->{'zapis'}{$zapi_command}{'APIcategory'},
					  "zapi"     => $zapi_command,
					  "hashref"  => $parsed_structure,
					); 
	if ( defined $dv_call2 ) {  # $dv_call2 ne "" ) {

		my $start_zapi = $dv_call2;
        	$start_zapi =~ s/next/start/;

        	$dv_call3 = $dv_call2;
        	$dv_call3 =~ s/next/end/;
        	$dv_call3 =~ s/-/_/g;

        	my @req_inputs = @{$parsed_structure->{'zapis'}{$start_zapi}{'required_inputs'}};

	        # Call 1 to start iter zapi
       	 	$dv_call1 = $start_zapi . "( ";
        	$dv_call1 =~ s/\-/_/g;

        foreach my $input ( @req_inputs ) {

            if( $parsed_structure->{'zapis'}{$start_zapi}{$input}{'type'} =~ /(integer|string|boolean)/i ){

                my $input_k = $input;
                $input_k =~ s/\-/_/g;

                if( $parsed_structure->{'zapis'}{$start_zapi}{$input}{'type'} =~ /\[\s*\]/ ) {

                    $dv_call1 .= "\"$input\" => \\\@" . $input_k  . "0". ", ";

                } else {

                    $dv_call1 .= "\"$input\" => \$" . $input_k  . "0". ", ";

                }

            } else {

                my $input_k = $parsed_structure->{'zapis'}{$start_zapi}{$input}{'type'};
                $input_k =~ s/\-/_/g;

                if( $input_k =~ /\[\s*\]/ ) {

                    $dv_call1 .= "\"$input\" => \\\@" . $input_k . "0". ", ";

                } else {

                    $dv_call1 .= "\"$input\" => \\\%" . $input_k . "0". ", ";

                }
            }

        }
        $dv_call1 .= " );";
        $dv_call1 =~ s/IN//g;
        $dv_call1 =~ s/\[//;
        $dv_call1 =~ s/\]//;


		# Call 2 for next iter zapi
        $dv_call2 =~ s/-/_/g;

        @req_inputs = @{$parsed_structure->{'zapis'}{$zapi_command}{'required_inputs'}};

        # Regex for checking if variable value is present or not
        $dv_check = "";

        ## Strings to prepend n attach for regex ##
        my ( $pre, $post);

        if( $zapi_command =~ /destroy|delete|remove|disinherit/ ) {

            $pre = "?=^.*^((?!(>\$";
            $post = "0<)).)*\$)";

        } else {

            $pre = "(?=^.*>\$";
            $post = "0<)";

        }
        foreach my $input ( @req_inputs ) {

            if( $parsed_structure->{'zapis'}{$zapi_command}{$input}{'type'} =~ /(integer|string|boolean)/i ){

                my $input_k = $input;
                $input_k =~ s/\-/_/g;
                $input_k =~ s/IN//;

                # Consider only if NOT ARRAY
                if( $parsed_structure->{'zapis'}{$zapi_command}{$input}{'type'} !~ /\[\s*\]/ ) {

                    $dv_check .= $pre . $input_k . $post;

                }

            } else {

              # TBD

            }

        }

	} # end of if check null    

	# Construct the hash to return 
	$data_validate{'start_call'} = $dv_call1;
        $data_validate{'next_call'} = $dv_call2;  
	$data_validate{'end_call'} = $dv_call3;  
	$data_validate{'check'} = $dv_check;

		
	# Return the hash  
	return \%data_validate;  

}

################################################################################
# Subroutine  : _get_errors_codes_of_zapi                                      #
#                                                                              #
# Description : This method gives all possible errors thrown by a zapi         #
#                                                                              #
# Inputs      : Takes a hash with the following keys:                          #
#               'zapi_command' - Contains the zapi which is being unit tested  #
#               'parsed_structure' - Reference to the DS of CPE                #
#                                                                              #
# Returns    :  A reference to an array containing the error codes             #
#                                                                              #
################################################################################

sub _get_errors_codes_of_zapi {

        my ( $self, %args ) = @_;

        my $zapi_command =  $args{"zapi_command"};
        my $parsed_structure = $args{"parsed_structure"};

        # Array output 
        my @errors = ( );
	  
	my $str = $parsed_structure->{'zapis'}{$zapi_command}{'APIerrnos'}; 

	if( $str =~ /,/ ) {  
		@errors = split( ",", $str);
	} else {
        	$errors[0] = $str;
	}

	my @err_numbers = ();

	foreach my $err ( @errors ) {  

		next unless $err;  
	
		push( @err_numbers, $parsed_structure->{'zapi_errors'}{$err}{'number'} );

	}

	return \@err_numbers;

}

################################################################################
# Subroutine  : _get_command_list                                              #
#                                                                              #
# Description : This method generates a user specified command list            #
#                                                                              #
#                                                                              #
# Returns     :  A Hash with keys having the names of the category             #
#                and values being an array containing the ZAPIs                #
#                                                                              #
################################################################################

sub _get_command_list{
      
        my $self = shift;
        my $commands = shift;
        my $zapi_parsed_hash = shift;
	my $code_type = shift; 
 
	# New zapi list that is to be formed
        my %zapi_list;

	#################################
	##   If code_type is DFM       ##
	################################# 
	if( $code_type =~ /dfm/i ) {

	    if( $commands =~ /\*/ ) {  

	   	return $zapi_parsed_hash->{'zapi_category'};
	    }

            # Populate user specified config list
            my @command_list = split (/,/,$commands);

            foreach my $cmd ( @command_list ) {

                if( exists $zapi_parsed_hash->{'zapi_category'}{$cmd} ) {

                    my @temp_array = @{ $zapi_parsed_hash->{'zapi_category'}{$cmd} };
                    # Add into the new hash
                    $zapi_list{$cmd} = \@temp_array;

            	} elsif( exists $zapi_parsed_hash->{'zapis'}{$cmd} ) {

                    # Get it's category
                    my $category = $zapi_parsed_hash->{'zapis'}{$cmd}{'APIcategory'};
                    my @temp_array = ( );

                    if( exists $zapi_list{$category} ) {
      
                    	@temp_array = @{ $zapi_list{$category} };
                    } 
                    push( @temp_array, $cmd );
                    $zapi_list{$category} = \@temp_array;

                }

            }   

	} else {  
	#################################
        ##   If code_type is Ontap     ## 
	#################################  

	    if( $commands =~ /\*/ ) {  

		foreach my $category ( keys(%{$zapi_parsed_hash->{'zapi_category'}}) ) {   

 		    foreach my $zapi ( @{$zapi_parsed_hash->{'zapi_category'}{$category}} ) {

			if( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /ontap_cluster/ ) {

			    push( @{$zapi_list{'CMode'}{$category}}, $zapi); 
	
			} 
			if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /(ontap_c,|ontap_c$)/ ) {  
	
			    push( @{$zapi_list{'7Mode'}{$category}}, $zapi); 

			} 
			if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /vserver/ ) {

                            push( @{$zapi_list{'Vserver'}{$category}}, $zapi);   

			}
   			if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /ontap_d/ ) {

			    push( @{$zapi_list{'ontapD'}{$category}}, $zapi);  

			}  

 
		    } # end of foreach $zapi  

		}  # end of foreach $category  

	    } # end of if command  

	    else {  

	        # Populate user specified config list
                my @command_list = split (/,/,$commands);

                foreach my $cmd ( @command_list ) {

		    if( exists $zapi_parsed_hash->{'zapi_category'}{$cmd} ) {  

			my @all_zapis = @{$zapi_parsed_hash->{'zapi_category'}{$cmd}}; 

			foreach my $zapi ( @all_zapis ) {  

			    if( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /ontap_cluster/ ) {

				push( @{$zapi_list{'CMode'}{$cmd}}, $zapi);
			    }
			    if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /(ontap_c,|ontap_c$)/ ) { 
		
				push( @{$zapi_list{'7Mode'}{$cmd}}, $zapi); 
			    }
			    if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /vserver/ ) {  
				
				push( @{$zapi_list{'Vserver'}{$cmd}}, $zapi);  
			    } 
			    if ( $zapi_parsed_hash->{'zapis'}{$zapi}{'APIfamily'} =~ /ontap_d/ ) {  

				push( @{$zapi_list{'ontapD'}{$cmd}}, $zapi);  
			    }   
			} # end of foreach 

		    } elsif( exists $zapi_parsed_hash->{'zapis'}{$cmd} ) {  

			my $category = $zapi_parsed_hash->{'zapis'}{$cmd}{'APIcategory'};   

			if( $zapi_parsed_hash->{'zapis'}{$cmd}{'APIfamily'} =~ /ontap_cluster/ ) {  
			    
			    push( @{$zapi_list{'CMode'}{$category}}, $cmd);
			}
			if( $zapi_parsed_hash->{'zapis'}{$cmd}{'APIfamily'} =~ /(ontap_c,|ontap_c$)/ ) {  

			    push( @{$zapi_list{'7Mode'}{$category}}, $cmd); 
		 	}
			if( $zapi_parsed_hash->{'zapis'}{$cmd}{'APIfamily'} =~ /vserver/ ) { 

			    push( @{$zapi_list{'Vserver'}{$category}}, $cmd);  
			}
			if( $zapi_parsed_hash->{'zapis'}{$cmd}{'APIfamily'} =~ /ontap_d/ ) { 

			    push( @{$zapi_list{'ontapD'}{$category}}, $cmd);  
			}  

		    }   

	        } # end of foreach $cmd 


	    } # end if else when ! * 
	} # end of else   

        # Return the new list
        return \%zapi_list;

}

################################################################################
# Subroutine  : _get_error_list                                                #
#                                                                              #
# Description : This method gives all possible errors thrown by a zapi         #
#                                                                              #
# Inputs      : Takes a hash with the following keys:                          #
#               'zapi_list' - Contains list of zapis which are being tested    #
#               'parsed_structure' - Reference to the DS of CPE                #
#                                                                              #
# Returns    :  A reference to a hash:                                         #
#               key - Number representing an error code                        #
#               value - Description of that error code                         #
#                                                                              #
################################################################################

sub _get_error_list {

        my ( $self, %args ) = @_;

        my @zapi_list =  @{$args{"zapi_list"}};
        my $parsed_structure = $args{"parsed_structure"};

        # Array output
        my @errors = ();

        my $str = "";

	foreach my $zapi_command (@zapi_list) {

	    $str .= $parsed_structure->{'zapis'}{$zapi_command}{'APIerrnos'} . ",";

	}

        if( $str =~ /,/ ) {
                @errors = split( ",", $str);
        } else {
                $errors[0] = $str;
        }

        # Hash output
        my %error_hash = ( );

        # Copy error hash
        my %parsed_error_hash = %{$parsed_structure->{'zapi_errors'}};


	foreach my $err ( @errors ) {  
	
	    next unless $err;  

	    if ( exists $parsed_error_hash{$err} ) {

		if( exists $parsed_error_hash{$err}{'number'} ) { 

	            $error_hash{$err} = $parsed_error_hash{$err}{'description'}; 
	        } else {
		    #print "\nNot found. |$err|". Dumper( $parsed_error_hash{$err} ); 
	        }

	    } else {  

		print "\nWarning : Error number $err not found in the error definition files.\n";
	    }   
	}

	# Return the hash reference
	return \%error_hash;    
}

################################################################################
# Subroutine  : _insert_testcase_list                                          #
#                                                                              #
# Description : This method generates a user specified command list            #
#                                                                              #
#                                                                              #
# Returns     :  A Hash with keys having the names of the category             #
#                and values being an array containing the ZAPIs                #
#                                                                              #
################################################################################

sub _insert_testcase_list {

	my ( $self, %args ) = @_;
        my $ut_fh = $args{"file_handle"};
	my $tc_list = $args{"tc_list"};
	my $code_path = $args{"code_path"};
	my $file_name = $args{"file_name"};
      
        my $new_ut_fh;
        # Insert the test case list into thpl file

        # Close and reopen in read mode
        close( $ut_fh );

        open( FH, "< $code_path/$file_name");
        my @file_contents = <FH>;
         close(FH);

        open($new_ut_fh, "> $code_path/$file_name");

        my $len1= @file_contents - 1;
        my $len2 = scalar(keys %$tc_list);

        foreach( my $i =0; $i< @file_contents; $i++) {

                if( $file_contents[$i] =~ /_NG_TESTCASE_INFO/) {

                        for( my $j= $len1 ; $j >= $i ; $j-- ){

                                $file_contents[$j+$len2-1] = $file_contents[$j];
                        }
                        $len1 += $len2;
                        while(my ($name,$description) = each(%$tc_list)){

                                $description =~ s/#.*//;
                                $description =~ s/"//g;
                                chomp($description);
                                $tc_list->{$name} = $description;
                                $file_contents[$i] = "$name =>\"$description \", ";
                                $i++;
                        }
                }
        }
        print $new_ut_fh @file_contents;
        return $new_ut_fh;

}

################################################################################
# Subroutine  : _generate_zapi_cdefs                                           #
#                                                                              #
# Description : This method generates the CDEFs for the all the ZAPIs present  #
#               in the zapidoc.all file, specified by the user                 #
#                                                                              #
#                                                                              #
# Returns     :  A Hash with keys having the names of the category             #
#                and values being an array containing the ZAPIs                #
#                                                                              #
################################################################################

sub _generate_zapi_cdefs {

    my ( $self, %args ) = @_;

    my $code_path = $args{"code_path"};
    my $file_path = $args{"file_path"};

    my $stdout;

    # Check whether it's needed to append slash at the end of file_path string
    if ( $file_path !~ /\/$/ ) {
        $file_path .= "/";
    }

    # Execute the zapitestxml to generate XMLs
    $stdout = `perl /u/deepthis/NATE/p4/nacldev/lib/NACL/APISet/Tools/zapitestxml -b /x/eng/rlse/DOT/mainN -d ${code_path}/xml -Z -a all ${file_path}zapidoc.all`;

    # Check the if the creation of xmls was successful or not
    if( !$stdout || $stdout =~ /ERROR/ ) {  

       print "\nCould not generate the CDEFs for the ZAPIs in zapidoc.all. " .
             "\nReason : \"$stdout\""; 
    } else {

       $stdout = `pwd`; 
       chomp($stdout);
       $code_path =  $stdout . $code_path;  
       $code_path =~ s/\.\//\//;

       print "\nSucessfully created the CDEFs for the ZAPIs in the zapidoc.all." .
            "\nThe CDEFs can be found in the location: ${code_path}/xml\n";
    }

}


1;
