# File Name     :       ConfigParse.pm
#
# Summary       :       Contains method to parse different tool input configuration files
#
# Creation Date :       4th August 2010
#
# Author        :       Jagadeesh , dl-sm-nacl-dev@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#       Summary:
#       Date  :
#       Author:
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#

#
# Package Declaration
#
package CommonLibs::ConfigParse;

#
# Compiler Directives
#
use strict;
use warnings;

#
# Perl Modules
#
use Data::Dumper;
use XML::Parser;
use XML::Simple;

#####################################################################################
# $MASTER_CONFIG hash contains the required & optional params list along with default
# values of all optional parameters. 
#####################################################################################

our $MASTER_CONFIG  = {
           'required_params' => {
				'cdef'       => ["fileName_cmd","interface"],
		                'functest'    => "",
				'task'        => "",
				'libcheck'    => "lib_name",
                        	'component'   => "dfm_entity",
                        	'unittest'    => ["interface","command","file_path","type", "version" ],
                     	     },
	   'optional_params' => {
				'common'   => {
						'log_level' => 'INFO',
					   },
                     		'cdef'     => {
                              		    'dfm_ip'       => 'localhost',
                               		    'dfm_login'    => 'root',
                                            'zapidoc_path' => 'TBD',
                                            'dfm_password' => 'autosmqa',
                                          },
                               'functest'  => {},
                               'libcheck'  => {},
                    	       'component' => {},
                   	       'task'      => {},
                               'unittest'  => {
                                             'family' => 'vserver',
					     'test_type' => 'all',  
                                         }
          },
};


#
# Constructor Definition
#
sub new {
    my $class = shift;
    my $self  = {};
    bless $self, $class;
    return $self;
}


################################################################################
# Subroutine  : parse_config                                                   #
#                                                                              #
# Description : This method parses the input xml config file                   #
#                                                                              #
# Returns     : Nested hash                                                    #
#                                                                              #
################################################################################
sub parse_config {

	my ( $self, $config_hash, %cli_options ) = @_;
        my $code_type = $cli_options{'code_type'};
 
	if(_is_valid_xml ()){
     	       $config_hash = _parse_xml();
	}else {
        	die "\n\tERROR: Unable to open input config file.\n";
        }
	
	# Hash where the values need to be overrided  

	$config_hash->{'code'}->{$code_type}->{'file_path'} = $cli_options{'file_path'} if defined $cli_options{'file_path'}; 

	$config_hash->{'code'}->{$code_type}->{'interface'} = $cli_options{'interface'} if defined $cli_options{'interface'};  

	$config_hash->{'code'}->{$code_type}->{'type'} = $cli_options{'type'} if defined $cli_options{'type'};  

	$config_hash->{'code'}->{$code_type}->{'family'} = $cli_options{'family'} if defined $cli_options{'family'};  

	$config_hash->{'code'}->{$code_type}->{'command'} = $cli_options{'command'} if defined $cli_options{'command'};  

	$config_hash->{'code'}->{$code_type}->{'version'} = $cli_options{'version'} if defined $cli_options{'version'};  

        $config_hash->{'code'}->{$code_type}->{'test_type'} = $cli_options{'test_type'} if defined $cli_options{'test_type'};     

	#
	# Keep only the required code type.
	#
	_remove_unwanted_blocks($config_hash, $code_type);
	
	#
	# Verify the required and set the default value to optional params
	#
	$config_hash = _verify_and_set_params ($config_hash, $code_type);
	
	#print "\nDEBUG: Config hash :- ".Dumper ($config_hash);
        return $config_hash;	
}

##################################################################################
# Subroutine  : _is_valid_xml						         #
#                                                                                #
# Description : This method checks the existance and correctness of the xml file.#
#                                                                                #
# Returns     : Returns undef in case of failure, otherwise returns one.         #
#                                                                                #
##################################################################################
sub _is_valid_xml () {
        my $xmlfile = "./Config/UserConfig.xml";
	unless (-e $xmlfile) {
		return undef;
	}
     
        # initialize parser object and parse the string
	my $parser = XML::Parser->new( ErrorContext => 2 );
	eval { 
		$parser->parsefile( $xmlfile ); 
	};
 
	if( $@ ) {
    		$@ =~ s/at \/.*?$//s;              
    		print STDERR "\nERROR in '$xmlfile':\n$@\n";
		return undef;
	} 
        return 1 ;
}

################################################################################
# Subroutine  : parse_xml                                                      #
#                                                                              #
# Description : This method parses the input xml config file                   #
#                                                                              #
# Returns     : Nested hash                                                    #
#                                                                              #
################################################################################
sub _parse_xml() {
	my $xml = new XML::Simple;
	my $data = $xml->XMLin("./Config/UserConfig.xml");
	return ($data);
}

################################################################################
# Subroutine  : _remove_unwanted_blocks 				       #
#                                                                              #
# Description : This method unwanted blocks in the config file.                #
#                                                                              #
# Returns     : none	 						       #
#                                                                              #
################################################################################
sub _remove_unwanted_blocks () {
	my $config_hash = shift;
	my $code_type = shift;
	
	#
	# Verify the code_type tag is present in $config_hash
	#		
	unless (defined $config_hash->{'code'}->{$code_type} ) {
		die  "\n\tERROR: Tag $code_type is not defined in UserConfig.xml\n";	
	}
	
	#
	# Remove unwanted keys in $config_hash
	#
	my @keys = keys %{$MASTER_CONFIG->{'required_params'}}; 
	foreach my $key (@keys) {
		unless ( $key =~ /$code_type/ ){
			delete $config_hash->{'code'}->{$key};
		}else {
		#	print "DEBUG: Keeping only $key in the config hash\n";
		}
		
	}
	
}

################################################################################
# Subroutine  : _verify_and_set_params					       #
#                                                                              #
# Description : This method check the required parameters and set the default  #
#               value for undefined optional paramater values.                 #
# Returns     : config hash                                                    #
#                                                                              #
################################################################################
sub _verify_and_set_params() {
	my $config_hash = shift;
        my $code_type  = shift;
        my @required_params = ();
	
	#
	# Verfiy all the required params are specified by user for the selected code_type
	# die otherwise.
	#

        #@required_params = split (' ', $MASTER_CONFIG->{'required_params'}->{$code_type});
	@required_params = @{$MASTER_CONFIG->{'required_params'}->{$code_type}};
        foreach my $param (@required_params) {
             my $val = $config_hash->{'code'}->{$code_type}->{$param};
             if ( $val =~ /HASH\(0x/i) {
	          die "\n\tERROR : Requried Param $param is missing in $code_type block.\n";
	     };
        }
	
	#
	# Check ZAPIdoc location in case of ZAPI interface
	#
        if ( 	$code_type =~ /cdef/ && 
		$config_hash->{'code'}->{'cdef'}->{'interface'} =~ /ZAPI/i && 
		$config_hash->{'code'}->{'cdef'}->{'zapidoc_path'} =~ /HASH\(0x/i 
	   ) {
			die "\n\tERROR: Param zapidoc_path is required for ZAPI interface.\n";
	}	
        
        #
	# Set the default value for each optional value, if user did not supply.
	#
        while ( my ($optional_param, $default_value ) = 
		each (%{$MASTER_CONFIG->{'optional_params'}->{$code_type}})
		) {
                  
		  if ($config_hash->{'code'}->{$code_type}->{$optional_param} =~ /HASH\(0x/i) {
			$config_hash->{'code'}->{$code_type}->{$optional_param} = $default_value ;   
			
			#
			# replace _IF_ in cdef location with proper interface details
			#
			if ($optional_param =~ /cdef_path/){
				$config_hash->{'code'}->{$code_type}->{$optional_param} =~ 
					s/_IF_/$config_hash->{'code'}->{$code_type}->{'interface'}/;	
			}
		  }
        }	
	
	#
	# Set Default value for Log Level.
	#
	if ( $config_hash->{'common'}->{'log_level'} =~  /HASH\(0x/i) {
		$config_hash->{'common'}->{'log_level'} = 
			$MASTER_CONFIG->{'optional_params'}->{'common'}->{'log_level'};
	}
	
	#
	# Update $config_hash with $code_type
	#
	$config_hash->{'code_type'} = $code_type ;
	
        return $config_hash ;
}

1;


