# File Name     :       SetupInit.pm
#
# Summary       :       Contains methods related to environment setup , cleanup or validation
#
# Creation Date :       5th August 2010
#
# Author        :       tanzim,dl-sm-nacl-dev@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#       Summary:
#       Date  :
#       Author:
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#

#
# Package Declaration
#

package CommonLibs::SetupInit;

#
# Compiler Directives
#

use strict;
use warnings;

#
# Perl Modules
#

use Data::Dumper;
use Net::Ping;
use Expect;
use File::Path qw(make_path);


#
# User Specific Modules
#


# Constructor Definition

sub new {
    my $class = shift;
    my $self  = {};
    bless $self, $class;
    return $self;
}


################################################################################
# Subroutine  : setup_init                                                     #
#                                                                              #
# Description : This method verifies environment setup configurations required #
#		by the tool.						       #	
#				    					       #
#                                                                              #
# Returns     : Displays error and exits tool - if verification fails          #
#               Returns configuration hash with                                #     
#		- ssh session id - if host type is remote		       #
#		- host type						       #	
#		- directory path for generated code   			       #		 
#                                                                              #
################################################################################


sub setup_init {
    my $self = shift;
    my $config_hash = shift;

    my ($host_type,$ssh_session_id);
    my $code_type = $config_hash->{'code_type'};		
    my $interface = $config_hash->{'code'}->{$code_type}->{'interface'};  

    my ($sec,$min,$hour,$day,$mon,$year,undef,undef,undef) = localtime(time);
    $mon = $mon+1;
    $mon = "0".$mon if($mon =~ /^(\d)$/);
    $day = "0".$day if($day =~ /^(\d)$/);
    $year += 1900;

    my $timestamp = $year.$mon.$day."_".$hour.$min.$sec;
    my $user = getlogin();

    #my $code_path = "/u/$user/NCAGE/$interface/".ucfirst($code_type."_". $timestamp);
    my $code_path = "./Code/" . ucfirst($code_type."_". $timestamp);  

        #print "\nSetup Initialization\n";

    # Decide if initialization is required depending on the code type
        # Check if the code_type key is correct	

     	# Will Support in Later Releases
	if(0){
	 unless ($config_hash->{'code_type'} =~ /task|functest|libcheck/){

		$host_type = ($config_hash->{'code'}->{$code_type}->{'dfm_ip'} eq "localhost") ?  "local" : "remote";
		$config_hash->{'host_type'} = $host_type;

		if ($host_type eq "local"){

			$self->_local_init($config_hash);
			print "\nLocal Setup pre-configuration successful\n";
			#return $config_hash;

		}
		else{

			$config_hash = $self->_remote_init($config_hash);
			# shifted to _remote_init method
			#$config_hash->{'ssh_session_id'} = $ssh_session_id;
			print "\nRemote Setup pre-configuration successful\n";
			#return $config_hash;
		}
		
	}
	}
	 $self->_create_file_path($code_path);
	 $config_hash->{'code_path'} = $code_path;
	 return $config_hash;

}

################################################################################
# Subroutine  : setup_uninit                                                   #
#                                                                              #
# Description : This method cleans up environment setup configurations  done   #
#               by the tool.                                                   #
#                                                                              #
#                                                                              #
# Returns     : Nothing                                                        #
#                                                                              #
################################################################################


sub setup_uninit {
    my $self = shift;
    my $config_hash = shift;

        #print "\nSetup UnInitialization\n";

	# Will Support in Later Releases
	if(0){
        if ($config_hash->{'host_type'} eq "remote"){

		$config_hash->{'ssh_session_id'}->soft_close();
	}
	}
	
	# Create a symbolic link
	# Linux specific
	
	# Need to add a check if a symlink by the name Latest exists
	# system ("unlink Latest");
	system ("ln -s $config_hash->{'code_path'} Latest ");
	# print "\nGenerated Code for $config_hash->{'code_type'} can be found at the directory : \"Latest\"\n"; 

 return;
}


################################################################################
# Subroutine  : _local_init                                                    #
#                                                                              #
# Description : This method verifies setup initialization on a local dfm host  #
#                                                                              #
# Returns     : config hash                                                    #
#                                                                              #
################################################################################


sub _local_init{
    my $self = shift;
    my $config_hash = shift;
        print "\nLocal Initialization\n";
     # Check DFM Status 
	
    $self-> _check_dfm_service($config_hash);

	return;

}


################################################################################
# Subroutine  : _remote_init                                                   #
#                                                                              #
# Description : This method verifies setup initialization on a remote dfm host #
#                                                                              #
# Returns     : config hash                                                    #
#                                                                              #
################################################################################


sub _remote_init{
    my $self = shift;
    my $config_hash = shift;
    my $code_type = $config_hash->{'code_type'}; 	
    my $ssh_session_id;
        print "\nRemote Initialization\n";

	$self->_check_ping($config_hash->{'code'}->{$code_type}->{'dfm_ip'});
	$ssh_session_id = $self->_ssh_target($config_hash->{'code'}->{$code_type}->{'dfm_ip'},$config_hash->{'code'}->{$code_type}->{'dfm_login'},$config_hash->{'code'}->{$code_type}->{'dfm_password'});	
        
	$config_hash->{'ssh_session_id'} = $ssh_session_id;      

	$self-> _check_dfm_service($config_hash);

	return $config_hash;
}






################################################################################
# Subroutine  : _check_dfm_service                                             #
#                                                                              #
# Description : This method verifies whether all service on DFM are started    #
#                                                                              #
# Returns     : Nothing , Errors out of the tool if DFM services are not up    #
#                                                                              #
################################################################################


sub _check_dfm_service{
    my $self = shift;
    my $config_hash = shift;
    my $command = "dfm service list";
    my $dfm_out;
        print "\nChecking DFM status\n";

    my $host_type = $config_hash->{'host_type'};

   if ($host_type eq "local"){

		$dfm_out = `$command`;
    }else{

	     $dfm_out= 	$self->_exec_remote_cmd($command , $config_hash);
   }	

	if ($dfm_out =~ /.*not\s+started.*|Error.*/i){

            die "\nERROR : All DFM services not started \n \t or \n $dfm_out\n";
        }
	
   return;	  	

}



################################################################################
# Subroutine  : _check_ping                                                    #
#                                                                              #
# Description : This method verifies whether a DFM host is up                  #
#                                                                              #
# Returns     : Returns nothing			                                       #
#               Displays error message and exits the tool if the host is downw #
#                                                                              #
################################################################################


sub _check_ping{
    my $self = shift;
    my $dfm_ip = shift;
        print "\nChecking DFM host connectivity status\n";

	my $ping_server;
    my $FLAG_PING = 0;
    my $num_tries = 0;
    my $MAX_TRIES = 3;

    $ping_server = Net::Ping->new("icmp");

    while ( ( !($FLAG_PING) ) and ( $num_tries <= $MAX_TRIES ) ) {
        if ( $ping_server->ping( $dfm_ip, 5 ) ) {
            $FLAG_PING = 1;
            print "\nPing to Server Successful\n";
        }
        else {
            print "\nAttempting PING $num_tries time\n";
            $num_tries = $num_tries + 1;
        }
    }
    die  "\n\nERROR :  Unable to ping $dfm_ip\n" if ( !$FLAG_PING );
    
    return;	
}


################################################################################
# Subroutine  : _ssh_target                                                    #
#                                                                              #
# Description : This module establishes an ssh session with the node           #
#											    						       #	 
# Returns : Spawn id                                                           #
#           Displays error and exits the tool                                  #
#                                                                              #
################################################################################

sub _ssh_target {

    my $self       = shift;
    my $connect_to = shift;
    my $user       = shift;
    my $password   = shift;
    my $timeout    = 5;

    my $command   = "ssh $user\@$connect_to";
    my $err_num   = '0';
    my $pass_flag = '0';
    my $spawn_ok  = '0';


    my $spawnID = Expect->spawn($command);    
    $spawnID->expect(
        $timeout,
        [
            qr'.continue connecting.',
            sub {
                $spawn_ok = 1;
                my $fh = shift;
                $fh->send("yes\n");
                exp_continue;
              }
        ],
        [
            'password:',
            sub {
                my $fh = shift;
                print $fh "$password\n";
                $pass_flag = 1;
                exp_continue;
              }
        ],

        [
            'Permission denied, please try again.',
            sub {
                $err_num = '-5';
                return $err_num;
              }
        ],
 [
            eof => sub {
                if ($spawn_ok) {
                    $err_num = '-2';
                    print "\nEOF : ERR NUM = -2\n";
                    return $err_num;
                }
                else {
                    $err_num = '-3';
                    print "\nEOF : ERR NUM = -3\n";
                    return $err_num;
}
              }
        ],

        [
            timeout => sub {
                if ( $pass_flag == 1 ) {
                    $pass_flag = 0;
                    $err_num   = 0;

                    #print "\nTIMEOUT : ERR NUM = 0\n";
                    return ($err_num);
                }

                $err_num = '-4';
                print "\nTIMEOUT : ERR NUM = -4\n";
                return $err_num;
              }
        ],
        '-re',
        qr'./# $',
    );
    sleep(2);
    if ( $err_num == '0' ) {
        return $spawnID;
    }
    else {
        die "\n ERROR : SSH Open Session to $connect_to failed... returning Err $err_num\n";
        #return ($err_num);
    }

}



################################################################################
# Subroutine  : _exec_remote_cmd                                               #
#                                                                              #
# Description : This method executes a command on a remote host                #
#                                                                              #
# Returns     : Command Output                                                 #
#                                                                              #
################################################################################


sub _exec_remote_cmd {

        my $self = shift;
        my $cmd  = shift;
        my $confref = shift;
        my $info;
        my $flag = "0";

        #$confref->{'logger'}->info("**************Executing command : $cmd**************");

        $confref->{'ssh_session_id'}->clear_accum();
        $confref->{'ssh_session_id'}->send("$cmd\n");
	    $confref->{'ssh_session_id'}->expect( "5",-re, qr'./*> $' );
	# Not handling interactive commands for now
        #  $confref->{'ssh_session_id'}->expect(undef,
        #                                [
        #                              '{y|n}:',
        #                                sub {
        #                                        my $fh = shift;
        #                                        print $fh "n\n";
        #                                        $pass_flag = 1;
        #                                        exp_continue;
        #                                }
        #                                ]
        #                                ,'-re', qr'./*> $',
        #                                );
        #
        $info = $confref->{'ssh_session_id'}->before();

	return $info;
}


sub _create_file_path {
	
	my $self = shift;
        my $code_path = shift;  
            make_path($code_path, {error => \my $err});

	if (@$err) {
   	   for my $diag (@$err) {
        	  my ($file, $message) = %$diag;
          	if ($file eq '') {
              		print "ERROR : $message\n";
          	}
          	else {
              		print "ERROR : problem unlinking $file: $message\n";
          	}
      	  }
  	}
  	else {
 	     	print "Directory $code_path successfully created\n";
  	}

}


1;
