# File Name     :       CommonParseEngine.pm
#
# Summary       :       The package parses different types of files ( CDEFs for CLI, zapidoc.all and zapi error files for ZAPI) into a well-defined data structure .
#
# Creation Date :       4th August 2010
#
# Author        :       deepthis@netapp.com, dl-sm-nacl-dev@netapp.com
#
# Version History
# ---------------
#
#  Version 1
#       Summary : Created the basic parser for cli
#       Date    : Aug 10, 2010
#       Author  : deepthis
#
# Copyright (c) 2010 Network Appliance, Inc.
# All rights reserved.
#

#
# Package Declaration
#

package ParseEngines::CommonParseEngine;

#
# Compiler Directives
#

use strict;
use warnings;

#
# Perl Modules
#

use Data::Dumper;
use XML::Parser;
use XML::Simple;

#
# User Specific Modules
#
BEGIN {
    push @INC,
        "/x/eng/rlse/DOT/mainN/final/bedrock/export/common/zapiprod/lib";
}

require "element_instance.pm";
require "element_typedef.pm";
require "zapi.pm";
require "zapiattr.pm";
require "zerrno.pm";
require "zcategory.pm";

# Constructor Definition

sub new {
    my $class = shift;
    my $self  = {
        'filenames' => undef,
        'logger'    => undef,
        'file_path' => undef,
        'type'      => undef,
    };
    bless $self, $class;
    return $self;
}

################################################################################
# Subroutine  : parse_file                                                     #
#                                                                              #
# Description : This method parses CLI CDEF files into a data structure        #
#                                                                              #
# Input       : Takes the interface, file_path and CommandName(optional) which  #
#               is passed in the form of a hash.                               #
#                                                                              #
# Returns     : Nested hash                                                    #
#                                                                              #
################################################################################

sub parse_file {

    my ( $self, %args ) = @_;

    # Initialize the logger variable
    #$Log = $args{'logger'};

    if ( $self->_validate_inputs(%args) ) {

        die "\nExiting as the inputs aren't valid\n";
    }

    # The reference to the Final data structure to be returned
    my $final_hash_ref;

    # Check the interface and call the appropriate parsing sub routine
    if ( $args{"interface"} =~ /CLI/i ) {

        $final_hash_ref = $self->_parse_cli_cdef();

    } elsif ( $args{"interface"} =~ /ZAPI/i ) {

        $final_hash_ref = $self->_parse_zapidoc();

    }

    # Return the hash reference
    return $final_hash_ref;
}

################################################################################
# Subroutine  : _validate_inputs                                               #
#                                                                              #
# Description : Validates the inputs given to the tool                         #
#                                                                              #
# Inputs      : Host(optional),Set(optional),Interface(required),              #
#               file_path, CommandName                                          #
#                                                                              #
# Outputs     : A defined/undef value indication true or false                 #
#                                                                              #
################################################################################

sub _validate_inputs {

    #Get the argumemts
    my ( $self, %args ) = @_;

    # Flag for error
    my $error_flag = 0;

    if ( defined $args{"file_path"} && !( -d $args{"file_path"} ) ) {
        $error_flag++;

        #$Log->error("The file_path mentioned does not exist!");
	print "\n\tERROR : The file_path mentioned ($args{\"file_path\"}) does not exist!\n"; 
	return $error_flag; 

    }

    # Assign the path to 'file_path'
    $self->{'file_path'} = $args{"file_path"};

    # Check whether it's needed to append slash at the end of file_path string
    if ( $self->{'file_path'} !~ /\/$/ ) {
        $self->{'file_path'} .= "/";
    }

    # Remove leading and trailing white spaces
    $self->{'file_path'} =~ s/^\s+//;
    $self->{'file_path'} =~ s/\s+$//;

    # Assign type to 'type'
    $self->{'type'} = $args{"type"};

    my @filenames = ();

    # Check for the existence of CLI Cdefs
    if ( $args{"interface"} =~ /CLI/i ) {

        if ( defined $args{"CommandName"} ) {

            # Split of comma, if multiple commands are specified
            my @filenames_temp = split( ",", $args{"CommandName"} );

            foreach my $cdef (@filenames_temp) {
                my @all_files = ();

                # Remove leading and trailing white spaces
                $cdef =~ s/^\s+//;
                $cdef =~ s/\s+$//;

                if ( $args{"interface"} =~ /CLI/i ) {

                    $cdef =~ s/\s+/_/g;
                }

                $cdef = $args{"file_path"} . $cdef;

                # Command is Linux-specific
                @all_files = `ls $cdef*`;
                if ( @all_files == 0 ) {
                    $error_flag++;

                    #$Log->error("There are no cdefs starting with $cdef");
                    last;
                }
                push( @filenames, @all_files );
            }

        } else {
            my $cdef = $args{"file_path"};

            # All the files in the folder are considered
            @filenames = `ls $cdef`;

            if ( $args{"interface"} =~ /CLI/i ) {
                foreach (@filenames) { $_ = $args{"file_path"} . $_; }
            }
        }
        foreach (@filenames) { chomp($_); }

        # Assign filesnames to 'filenames'
        $self->{'filenames'} = \@filenames;
    }

    # Check for the presence of zapi files
    if ( $args{"interface"} =~ /ZAPI/i ) {

        if ( defined $args{"CommandName"} ) {

            # Check if there's any regular expression present. If present,
            # throw an error message  
            if( $args{"CommandName"} =~ /\*|\+/ && $args{"CommandName"} !~ /^\*$/ ) {

                print "\n\tERROR : The command list ($args{'CommandName'}) contains special characters." .
                    "\n\tPlease mention only the ZAPI name or the ZAPI category name. Or mention \"*\" to include ALL the ZAPIs.\n";
                $error_flag++;  
	    }

            # Split of comma, if multiple commands are specified
            my @commands = split( ",", $args{"CommandName"} );
            foreach my $cmd (@commands) {

                # Remove leading and trailing white spaces
                $cmd =~ s/^\s+//;
                $cmd =~ s/\s+$//;
            }

            # Assign commands to 'filenames'
            $self->{'filenames'} = \@commands;
        }

        if (!(     -e ( $self->{'file_path'} . "zapidoc.all" )
                && -e ( $self->{'file_path'} . "zapierrno.h" )
            )
            )
        {

	    print "\n\tERROR : The zapidoc.all and/or zapierrno.h is missing from the file_path ($self->{'file_path'}) location.\n"; 
            $error_flag++;

        } elsif ( ( $args{"type"} =~ /DFM/i )
            and !( -e $self->{'file_path'} . "zapidoc.errdefs" ) )
        {
	    print "\n\tERROR : The zapidoc.errdefs is missing from the file_path ($self->{'file_path'}) location.\n";  
            $error_flag++;

        }
    }

# Check whether $error_flag is still 0 (which means no error) or >0 (which means error)
# Return undef for FAIL and defined for PASS
#return $error_flag ? defined : undef;
    return $error_flag ;
}

################################################################################
# Subroutine  : _parse_cli_cdef                                                #
#                                                                              #
# Description : This method parses CLI CDEF files into a data structure        #
#                                                                              #
# Input       : None.                                                          #
#                                                                              #
# Returns     : Nested hash                                                    #
#                                                                              #
################################################################################

sub _parse_cli_cdef {

    my $self = shift;

    #$Log->debug("\nParsing CLI CDEF");

    # The final data structure to be returned
    my %final_nested_hash;

    my @filenames = @{ $self->{'filenames'} };

    foreach my $file (@filenames) {

        require $file;

        our ( %CommandDefinitions, @Boolean_Options );

        # Copy the hash into a local variable
        my %single_command_hash = %CommandDefinitions;

        # For the primary keys and options ref
        my ( @prim_keys, $options_hash_ref );

        # Get the parameter type - hash reference
        my $param_type_hash_ref = $self->_get_cli_parameter_type();

        foreach my $key ( keys(%single_command_hash) ) {

            #Delete the command key
            delete $single_command_hash{$key}->{'command'};

            # Delete the keys key, after copying contents to @prim_keys array
            @prim_keys = keys( %{ $single_command_hash{$key}->{'keys'} } );
            delete $single_command_hash{$key}->{'keys'};

            # Delete the options key after getting the options hash ref
            $options_hash_ref = $single_command_hash{$key}->{'options'};
            delete $single_command_hash{$key}->{'options'};

            my @bool_options = ();

            # Populate the boolean_options key
            foreach
                my $opt ( @{ $single_command_hash{$key}->{'option_order'} } )
            {

                if ( grep { $_ eq $opt } @Boolean_Options ) {
                    push( @bool_options, $opt );

                }

            }

            # Add the boolean_options key
            $single_command_hash{$key}->{'boolean_options'} = \@bool_options;
        }

        # Add primary keys
        $single_command_hash{'Primary Keys'} = \@prim_keys;

        # Add Param Type
        $single_command_hash{'Param Type'} = $param_type_hash_ref;

        # Add counter_commands key
        $single_command_hash{'counter_commands'}
            = $self->_get_counter_commands_hash(
            "hash_ref" => \%single_command_hash );

        # Add the options map
        $single_command_hash{'options_map'} = $options_hash_ref;

        # Get the key name from the cdef name
        $file =~ s/(.*)(df.*)(\.cdefs)/$2/;
        $file =~ s/_/ /g;

# Assign the nested hash structure as value to the hash with key as the command
        $final_nested_hash{"$file"} = \%single_command_hash;
    }

#$Log->debug("The final data structure: ".Dumper(\%final_nested_hash));getc();

    return \%final_nested_hash;
}

################################################################################
# Subroutine  : _parse_zapidoc                                                 #
#                                                                              #
# Description : This method parses zapidoc.all file into a data structure      #
#                                                                              #
# Input       : Path of the zapidoc.all file                                   #
#                                                                              #
# Returns     : Nested hash                                                    #
#                                                                              #
################################################################################

sub _parse_zapidoc {
    my $self = shift;

    #$Log->debug("\nParsing zapidoc.all file");

    my @filenames = @{ $self->{'filenames'} };

    # Zapidoc file location
    my $zapi_file_path = $self->{'file_path'} . "zapidoc.all";

    # The final data structure to be returned
    my %final_nested_hash;

    # Parse type
    my $parse_type = $self->{'type'};

    # Parsing
    zapi::scanforstuff("$zapi_file_path");

    # Store the typedefs in the data structure
    $final_nested_hash{'zapi_typedefs'} = \%::Ztypes;

    #########################################
    ##       TYPEDEFS PART                 ##
    #########################################
    # Make the list of required and optional Options and also fill in the Ranges for the zapi TYPEDEFS
    foreach my $zapi_typedef ( keys( %{ $final_nested_hash{'zapi_typedefs'} } ) )    
    {
	# Changing from netcache: to filer: 
	if ( $zapi_typedef =~ /netcache:/ ) {
	
	    my $temp_zapi_typedef = $zapi_typedef; 
	    $temp_zapi_typedef =~ s/netcache:/filer:/;

	    my $temp_value = $final_nested_hash{'zapi_typedefs'}{$zapi_typedef};  
	    delete $final_nested_hash{'zapi_typedefs'}{$zapi_typedef};  
	    $final_nested_hash{'zapi_typedefs'}{$temp_zapi_typedef} = $temp_value;

	    # Change the value of the key variable 
	    $zapi_typedef = $temp_zapi_typedef;  
	}
		
        my @required_options = ();
        my @optional_options = ();

        # Delete 'api'
        delete $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'api'};

        foreach my $option ( keys( %{ $final_nested_hash{'zapi_typedefs'}{$zapi_typedef} } ) )   
        {

            # If it's an "element"
            if ( $option =~ /^ELEMENT/ ) {

                # Check if it is required/optional
                if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} =~ /optional/ )    
                {
                    push( @optional_options, $option );
                } else {
                    push( @required_options, $option );
                }

                # Check if integer
                if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} =~ /integer/ )   
                {

                    # Check if Range is already specified in 'range' field
                    if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} =~ /\.\./ )   
                    {

                        my ( $high, $low );
			my $minus = 0; 
                        if ($final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} =~ /\[\s*([+\-]*.+)\s*\.\.\s*(.*)\s*\]/ )      
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high); 

                            # Store it in the data structure
                            $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} = $low . ".." . $high; 

                        } elsif ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'}=~ /([+\-]*\d+)\s*to\s*([+\-]*\d+)/ )    
                        {

                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high); 

                            # Store it in the data structure
                            $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} = $low . ".." . $high; 

                        }
                    }

                    # Check if description has range
                    elsif ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'desc'} =~ /Range/i )   
                    {

                        my ( $low, $high );
			my $minus = 0;  
                        if ($final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'desc'} =~ /\[\s*([+\-]*.+)\s*\.\.\s*(.*)\s*\]/ )   
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);     

			    # Store it in the data structure 
			    $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} = $low . ".." . $high;  

                        } elsif ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'desc'} =~ /([+\-]*\d+)\s*to\s*([+\-]*\d+)/ )  
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);

                            # Store it in the data structure
                            $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} = $low . ".." . $high; 

                        }
                    }
                }

                # Check if string
                if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} =~ /string/ )   
                {
                    # Check if description has enumerations
                    my $desc_string = $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'desc'}; 

                    if ($desc_string =~ /Possible\s*values|Allowed\s*values|Legitimate\s*values/i ) {
      
                        my @enum_array = ();

                        # look for all strings inside ""
                        @enum_array = $desc_string =~ /\"(\w*)\"/g;

                        # If above array is empty look for all strings inside ''                
                        @enum_array = $desc_string =~ /\'(\w*)\'/g unless (@enum_array);

                        #removing duplicate entries
                        my %seen;
                        my @uniq_enums = grep {! $seen{$_}++} @enum_array;
                        @enum_array = @uniq_enums;

                        # Store it in the data structure
                        
                        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'enum'} = \@enum_array;   

                    }

                }

            }

        }

        # Check for range in the main 'desc'
        if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'desc'}=~ /Range/i )   
        {
            my ( $low, $high );
	    my $minus = 0;  
            if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'desc'} =~ /\[\s*([+\-]*.+)\s*\.\.\s*(.*)\s*\]/ )  
            {
                $low  = $1;
                $high = $2;
		($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high); 

		# Store it in the data structure 
		$final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'range'} = $low . ".." . $high; 

            } elsif ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'desc'} =~ /([+\-]*\d+)\s*to\s*([+\-]*\d+)/ )     
            {
                $low  = $1;
                $high = $2;
		($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high); 

                # Store it in the data structure
                $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'range'} = $low . ".." . $high; 
            }

        }

        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'required_options'} = \@required_options;   
        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{'optional_options'} = \@optional_options;   
    }

    # To put the range & type - for indirect reference

    foreach  my $zapi_typedef ( keys( %{ $final_nested_hash{'zapi_typedefs'} } ) )  
    {

        foreach my $option ( keys( %{ $final_nested_hash{'zapi_typedefs'}{$zapi_typedef} } ) )      
        {

            # If it's an "element"
            if ( $option =~ /^ELEMENT/ ) {

                # If a reference to another thing, which is an integer/string
                if ( $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} !~ /(integer|string|boolean)/i )    
                {

                    my $is_array    = 0;
                    my $is_optional = 0;
                    my $type = $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'};    
                    $is_array    = 1 if ( $type =~ /\[/ );
                    $is_optional = 1 if ( $type =~ /, optional/ );

                    $type =~ s/\[//;
                    $type =~ s/\]//;
                    $type =~ s/, optional//;
                    my $tempp
                        = ( ( $self->{'type'} eq "dfm" )
                        ? ( "dfm:" . $type )
                        : ( "filer:" . $type ) );

                    #my $tempp = "dfm:".$type;
                    if (defined $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'} )    
                    {

                        # Update type
                        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'};   
                        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} .= "[]"  if ($is_array);     
                        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'type'} .= ", optional" if ($is_optional);      

                        # Update range
                        $final_nested_hash{'zapi_typedefs'}{$zapi_typedef}{$option}{'range'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'range'};      

                    }

                }
            }
        }
    }

    #########################################
    ##       MAIN ZAPIS PART               ##
    #########################################
    my %zapi_hash = %::Zapis;

    # Simplify the data structure
    foreach my $zapi_command ( keys(%zapi_hash) ) {

        my $flag = 1;

        # Put the key for category in 'APIcategory'
        my $tempp = $zapi_hash{$zapi_command}{'APIcategory'}{'name'};
        $zapi_hash{$zapi_command}{'APIcategory'} = $tempp;

        # Delete these keys as they are not needed
        delete $zapi_hash{$zapi_command}{'APIreadonly'};
        delete $zapi_hash{$zapi_command}{'APIcfunc'};
        delete $zapi_hash{$zapi_command}{'APIontapng'};
        delete $zapi_hash{$zapi_command}{'APIstreaming'};
        delete $zapi_hash{$zapi_command}{'APIjava'};
        delete $zapi_hash{$zapi_command}{'TESTTBD'};

    }

    # Make list category wise i.e key is category, value is a reference to an array of ZAPIs that belong to that category
    my %zapi_category;
    foreach my $zapi_command ( keys(%zapi_hash) ) {

        # Take the category of each zapi
        my $category = $zapi_hash{$zapi_command}{'APIcategory'};

        # Put it in a hash with category as key and command being appended to value
        $zapi_category{$category} .= $zapi_command . ",";

    }

    # Convert the comma seperated string into an array
    # and make the category key in the hash have reference to the array as value
    foreach my $category ( keys(%zapi_category) ) {

        my $list = $zapi_category{$category};
        my @array = split( ",", $list );

        $zapi_category{$category} = \@array;

    }

    # Assign it to final nested hash
    $final_nested_hash{'zapi_category'} = \%zapi_category;

    # Make the list of required and optional input keys and also fill in the Ranges for the zapis
    foreach my $zapi_command ( keys(%zapi_hash) ) {

        # A flag for 'Range specified for a required input'
        # Set the flag to '0' (false) initially
        $zapi_hash{$zapi_command}{'integer_range'} = 0;

        my @required_inputs  = ();
        my @optional_inputs  = ();
        my @required_outputs = ();
        my @optional_outputs = ();

        foreach my $option ( keys( %{ $zapi_hash{$zapi_command} } ) ) {

            #########################
            #	Option is an INPUT  #
            #########################
            if ( $option =~ /^IN/ ) {

                # To make arrays for required and optional Inputs
                if ($zapi_hash{$zapi_command}{$option}{'type'} =~ /optional/ )
                {
                    push( @optional_inputs, $option );
                } else {
                    push( @required_inputs, $option );
                }

                # Check if integer
                if ( $zapi_hash{$zapi_command}{$option}{'type'} =~ /integer/ )
                {

                    # Check if description has range
                    if ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /Range/i )  
                    {

                        # Put the value in the 'range' field of the input
                        my ( $low, $high );
			my $minus = 0;  
                        if ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /\[\s*(\d+)\s*\.\.\s*(.*)\s*\]/ )    
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);  

                            # Store it in the data structure
                            $zapi_hash{$zapi_command}{$option}{'range'} = $low . ".." . $high;   

                        } elsif ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /([+\-]*\d+)\s*to\s*([+\-]*\d+)/ )   
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);  

                            # Store it in the data structure
                            $zapi_hash{$zapi_command}{$option}{'range'} = $low . ".." . $high;   

                        }

                        # Set the flag field saying range is specified for a required integer input
   			# Find out if this option belongs to 'required_inputs'
                        if ( $zapi_hash{$zapi_command}{$option}{'type'} !~ /optional/ )  
                        {
                            $zapi_hash{$zapi_command}{'integer_range'} = 1;
                        }
                    } else {

                        # Put the 'range' field as undef
                        $zapi_hash{$zapi_command}{$option}{'range'} = undef;
                    }
                }

                # To put the range & type - for indirect reference
                if ( $zapi_hash{$zapi_command}{$option}{'type'} !~ /(integer|string|boolean)/i )  
                {

                    my $is_array    = 0;
                    my $is_optional = 0;
                    my $type = $zapi_hash{$zapi_command}{$option}{'type'};
                    $is_array    = 1 if ( $type =~ /\[/ );
                    $is_optional = 1 if ( $type =~ /, optional/ );

                    $type =~ s/\[//;
                    $type =~ s/\]//;
                    $type =~ s/, optional//;
                    my $tempp
                        = ( ( $self->{'type'} eq "dfm" )
                        ? ( "dfm:" . $type )
                        : ( "filer:" . $type ) );

                    #my $tempp = "dfm:".$type;
                    if (defined $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'} )     
                    {

                        # Update type
                        $zapi_hash{$zapi_command}{$option}{'type'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'};    
                        $zapi_hash{$zapi_command}{$option}{'type'} .= "[]" if ($is_array); 
                        $zapi_hash{$zapi_command}{$option}{'type'} .= ", optional"  if ($is_optional);     

                        # Update range
                        $zapi_hash{$zapi_command}{$option}{'range'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'range'};      
                    }

                }

            } elsif ( $option =~ /^OUT/ ) {
                #########################
                #	Option is an OUTPUT  #
                #########################

                # To make arrays for required and optional Inputs
                if ($zapi_hash{$zapi_command}{$option}{'type'} =~ /optional/ )
                {
                    push( @optional_outputs, $option );
                } else {
                    push( @required_outputs, $option );
                }

                # Check if integer
                if ( $zapi_hash{$zapi_command}{$option}{'type'} =~ /integer/ )
                {

                    # Check if description has range
                    if ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /Range/i )   
                    {

                        my ( $low, $high );
			my $minus = 0; 
                        if ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /\[\s*(\d+)\s*\.\.\s*(.*)\s*\]/ )   
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);  
=cut 
			    $low =~ s/ //g;
                            $high =~ s/ //g;

                            $minus = 1  if( $low =~ /^\-/ );
                            # If $low is 2 to the power something..
                            if( $low =~ /2\^(\d+)\-(\d)/ ) {
                                $low = (2 ** $1 ) - $2;
                            }
                            if( $minus ) {
                                $low = "-" . $low;
                                $minus = 0;
                            }

                            $minus = 1  if( $high =~ /^\-/ );
                            # If $high is 2 to the power something..
                            if( $high =~ /2\^(\d+)\-(\d)/ ) {
                                $high = (2 ** $1 ) - $2;
                            }
                            if( $minus ) {
                                $high = "-" . $high;   
                            }  
=cut 
                            # Store it in the data structure
                            $zapi_hash{$zapi_command}{$option}{'range'} = $low . ".." . $high;     

                        } elsif ( $zapi_hash{$zapi_command}{$option}{'desc'} =~ /([+\-]*\d+)\s*to\s*([+\-]*\d+)/ )   
                        {
                            $low  = $1;
                            $high = $2;
			    ($low, $high) = $self->_compute_high_low( "low" => $low, "high" => $high);  
=cut 
			    $low =~ s/ //g;
                            $high =~ s/ //g;

                            $minus = 1  if( $low =~ /^\-/ );
                            # If $low is 2 to the power something..
                            if( $low =~ /2\^(\d+)\-(\d)/ ) {
                                $low = (2 ** $1 ) - $2;
                            }
                            if( $minus ) {
                                $low = "-" . $low;
                                $minus = 0;
                            }

                            $minus = 1  if( $high =~ /^\-/ );
                            # If $high is 2 to the power something..
                            if( $high =~ /2\^(\d+)\-(\d)/ ) {
                                $high = (2 ** $1 ) - $2;
                            }
                            if( $minus ) {
                                $high = "-" . $high;
                            }
=cut 

                            # Store it in the data structure
                            $zapi_hash{$zapi_command}{$option}{'range'} = $low . ".." . $high;   
                        }

                    }
                }

                # To put the range & type - for indirect reference
                if ( $zapi_hash{$zapi_command}{$option}{'type'} !~ /(integer|string|boolean)/i )    
                {

                    my $is_array    = 0;
                    my $is_optional = 0;
                    my $type = $zapi_hash{$zapi_command}{$option}{'type'};
                    $is_array    = 1 if ( $type =~ /\[/ );
                    $is_optional = 1 if ( $type =~ /, optional/ );

                    $type =~ s/\[//;
                    $type =~ s/\]//;
                    $type =~ s/, optional//;
                    my $tempp
                        = ( ( $self->{'type'} eq "dfm" )
                        ? ( "dfm:" . $type )
                        : ( "filer:" . $type ) );
                    ####	my $tempp = "dfm:".$type;
                    if (defined $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'} )  
                    {

                        # Update type
                        $zapi_hash{$zapi_command}{$option}{'type'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'type'};     
                        $zapi_hash{$zapi_command}{$option}{'type'} .= "[]" if ($is_array);    
                        $zapi_hash{$zapi_command}{$option}{'type'} .= ", optional" if ($is_optional);       

                        # Update range
                        $zapi_hash{$zapi_command}{$option}{'range'} = $final_nested_hash{'zapi_typedefs'}{$tempp}{'range'};     
                    }

                }
            }

        }
        $zapi_hash{$zapi_command}{'required_inputs'}  = \@required_inputs;
        $zapi_hash{$zapi_command}{'optional_inputs'}  = \@optional_inputs;
        $zapi_hash{$zapi_command}{'required_outputs'} = \@required_outputs;
        $zapi_hash{$zapi_command}{'optional_outputs'} = \@optional_outputs;
    }

    # Store the information in the final data structure
    $final_nested_hash{'zapis'} = \%zapi_hash;

    #########################################
    ##       ZAPI ERROR STRING PART        ##
    #########################################

    # Get the error data structure
    my %error_hash = $self->_parse_zapi_error_files();
    $final_nested_hash{'zapi_errors'} = \%error_hash;

    #$Log->debug("The final data structure: ".Dumper(\%final_nested_hash));
    return \%final_nested_hash;
}

################################################################################
# Subroutine  : _compute_high_low                                              #
#                                                                              #
# Description : Computes values of $high and $low and returns                  #
#                                                                              #
################################################################################ 

sub _compute_high_low {

    # Get the arguments 
    my ( $self, %args ) = @_;  

    my $low = $args{'low'};
    my $high = $args{'high'};
    my $minus = 0; 

    $low =~ s/ //g;
    $high =~ s/ //g;

    $minus = 1  if( $low =~ /^\-/ );
    # If $low is 2 to the power something..
    if( $low =~ /2\^(\d+)\-(\d)/ ) {
        $low = (2 ** $1 ) - $2;
    }
    if( $minus ) {
          $low = "-" . $low;
          $minus = 0;
    }

    $minus = 1  if( $high =~ /^\-/ );
    # If $high is 2 to the power something..
    if( $high =~ /\(*2\^(\d+)\)*\-(\d)/ ) {
              $high = (2 ** $1 ) - $2;
    }
     if( $minus ) {
           $high = "-" . $high;
      }
    return ($low, $high);
}

################################################################################
# Subroutine  : _parse_zapi_error_files                                         #
#                                                                              #
# Description : This method parses the error files zapidoc.errdefs and         #
#               zapierrno.h and prepares a data structure                      #
#                                                                              #
# Input       : Location of the error files                                    #
#                                                                              #
# Returns     : Reference to a hash containing error definitions as keys       #
#               which are mapped to the corresponding error numbers & desc     #
#                                                                              #
# Note        : Both the zapidoc.errdefs and zapierrno.h have to be in the     #
#               same folder                                                    #
################################################################################

sub _parse_zapi_error_files {

    my $self = shift;

    #$Log->debug("\nParsing zapidoc.errdefs and zapierrno.h files");

    # Zapi error files' location
    my $ontap_errorfile = $self->{'file_path'} . "zapierrno.h";

    # The final error hash
    my %errhash;

    # Count variable
    my $count;

    # Processing the Ontap error file zapierrno.h
    my $ontaperrors = `cat $ontap_errorfile`;
    my @ontaperrors = split( '#define', $ontaperrors );

    $count = 0;

    foreach (@ontaperrors) {

        if ( $_ =~ /(E[a-zA-Z_0-9]*)(\s+)([\d]+)(.*)(\/\*)(.*)(\*\/)/s ) {
            my $errorname = $1;
            $errhash{$errorname}{number} = $3;
            my $desc = $6;
            $desc =~ s/\n+//g;
            $desc =~ s/\s+/ /g;
            $errhash{$errorname}{description} = $desc;
        }

    }

    # Processing the DFM error file zapidoc.errdefs
    if ( $self->{'type'} =~ /dfm/i ) {

        my $dfm_errorfile = $self->{'file_path'} . "zapidoc.errdefs";

        my $dfmerrors        = `cat $dfm_errorfile`;
        my @errordefinitions = split( '\/\*', $dfmerrors );
        my $dupcount         = 0;

        foreach (@errordefinitions) {
            my $errorname;
            if ( $_ =~ /\@errdef(.*)\@desc(.*)/s ) {
                my $errornameandnumber = $1;
                my $desc               = $2;
                if ( $desc =~ /[\*\/]+/ ) {
                    $desc =~ s/[\*\/]+/ /g;
                }

                $desc =~ s/\n+//g;
                $desc =~ s/\s+/ /g;

                if ( $errornameandnumber =~ /(E[A-Za-z_0-9]*)(\s+)([\d]+)/ ) {

                    #check if the key already exists
                    if ( exists( $errhash{$1} ) ) {
                        my $oldid = $errhash{$1}{number};

                        #if exists , check if the id is same
                        if ( $oldid == $3 ) {
                            $errhash{$1}{number} = $3;

                            $errhash{$1}{description} = $desc;
                            $errorname = $1;
                        } else {
                            print
                                "The error id of $1 is not matching in both error definition files\n";
                            $errhash{$1}{number} = $3;

                            $errhash{$1}{description} = $desc;
                            $errorname = $1;

                        }

                    } else {
                        $errhash{$1}{number} = $3;

                        $errhash{$1}{description} = $desc;
                        $errorname = $1;
                    }

                }
            }

        }
    }

    # Return the data structure
    return %errhash;
}

################################################################################
# Subroutine  : _get_counter_commands_hash                                     #
#                                                                              #
# Description : This method parses command hash and forms a hash with counter  #
#               commands                                                       #
#                                                                              #
# Input       : The single command hash                                        #
#                                                                              #
# Returns     : Reference to a hash containing counter commands                #
#                                                                              #
# Note        : An assumption is made that the main command contains only      #
#               add/delete etc commands. If >1 present, they would be a        #
#               seperate component and hence in a different hash.              #
################################################################################

sub _get_counter_commands_hash {

    #Get the argumemts
    my ( $self, %args ) = @_;

    # Get the hash to work on
    my %command_hash = %{ $args{"hash_ref"} };

    # The counter hash to be constructed and returned
    my %counter_hash = ();

    # A temporary variable to hold values during assignment
    my $counter_temp_var;

    # Check if create command is present
    if ( my @temp1 = grep {/create/} keys(%command_hash) ) {

        # Check if add command is also present
        if ( my @temp2 = grep {/add/} keys(%command_hash) ) {

            my @temp3 = grep {/destroy/} keys(%command_hash);
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp3[0];
            $counter_hash{ $temp3[0] }       = $counter_temp_var
                if ( ( scalar @temp3 ) > 0 );

            @temp3 = grep {/delete/} keys(%command_hash);
            $counter_temp_var = $temp2[0];
            $counter_hash{$counter_temp_var} = $temp3[0];
            $counter_hash{ $temp3[0] } = $counter_temp_var;

        } else {    # Check if only create is present
            if ( @temp2 = grep {/destroy/} keys(%command_hash) ) {
                $counter_temp_var                = $temp1[0];
                $counter_hash{$counter_temp_var} = $temp2[0];
                $counter_hash{ $temp2[0] }       = $counter_temp_var;

            } elsif (
                @temp2 = grep {
                    /delete/
                } keys(%command_hash)
                )
            {
                $counter_temp_var                = $temp1[0];
                $counter_hash{$counter_temp_var} = $temp2[0];
                $counter_hash{ $temp2[0] }       = $counter_temp_var;
            }
        }
    } elsif (
        @temp1 = grep {
            /add/
        } keys(%command_hash)
        )
    {

        # If only add is present
        if ( my @temp2 = grep {/destroy/} keys(%command_hash) ) {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;

        } elsif (
            @temp2 = grep {
                /delete/
            } keys(%command_hash)
            )
        {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;
        }
    }

    # If start command is present
    if ( my @temp1 = grep {/start/} keys(%command_hash) ) {

        if ( my @temp2 = grep {/stop/} keys(%command_hash) ) {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;
        }
    }

    # If a disable command is present
    if ( my @temp1 = grep {/disable/} keys(%command_hash) ) {

        if ( my @temp2 = grep {/enable/} keys(%command_hash) ) {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;

        } elsif (
            @temp2 = grep {
                /install/
            } keys(%command_hash)
            )
        {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;
        }
    }

    # If a inherit command is present
    if ( my @temp1 = grep {/inherit/} keys(%command_hash) ) {

        if ( my @temp2 = grep {/disinherit/} keys(%command_hash) ) {
            $counter_temp_var                = $temp1[0];
            $counter_hash{$counter_temp_var} = $temp2[0];
            $counter_hash{ $temp2[0] }       = $counter_temp_var;
        }
    }

    # Return the hash reference
    return \%counter_hash;

}

################################################################################
# Name    : _get_cli_parameter_type()                                          #
#                                                                              #
# Purpose : To find out the data type of CLI options                           #
#                                                                              #
# Inputs  : TBD                                                                #
#                                                                              #
# Outputs : TBD                                                                #
################################################################################

sub _get_cli_parameter_type() {

    return undef;
}

1;

