# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/Acl.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
package nfs::ACL::Acl;

## @summary Library for the NFS ACL tests
##
## @description
##   Common routines for NFS ACL testing.
##
## @usage
##   use nfs::ACL::Acl;
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;
use English;
use Carp qw(croak);
use Cwd;
use Fcntl qw(:mode);
use File::stat;
use Tharn;
use FS::Util::Setup qw(create_vol common_cleanup);
use TestEnv::MinWin qw(minwin_compatible);
use MapShare qw(map_cifs_share unmap_cifs_share);
use Network::Utils qw(run_getXXbyYY);
use GenericMount qw(mount umount);
use ExportFS qw(share_add share_delete);
use Users qw(addCIFSUser removeCIFSUser);
use SPARSE::Util::Logger;
use filer qw(filer_options_get filer_options_set filer_qtree_security_set
             filer_cifs_lookup);
use Params::Validate qw(validate ARRAYREF CODEREF HASHREF OBJECT SCALAR UNDEF);
use Admin::License qw(license is_licensed unlicense);
use SnapMirror qw(snapmirror_enable snapmirror_disable);
use nfs::GenericExport qw(export_path unexport_path);
use nfs::Export::NFSexports qw(add_result);
use nfs::ACL::AclNFSV4 qw(:ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);
use CLI::FileIO qw(rdfile wrfile);
use FileManip qw(read_file write_file);
use config_base qw(host_is_alive);
use SnapCmds qw(snapshot_create snapshot_exists snapshot_delete
                snaprestore_volume);
use boot qw(transit_to);
use Storable qw(dclone);
use nfs::Utils qw(setup_cluster);
use cluster qw(cf_disable);
use CIFS::Setup qw(cifs_setup);

#-------------------------------------------------------
# Global variables
#-------------------------------------------------------
use vars qw(
  $VERSION
  @ISA
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  run_remote
  run_local
  acl_validate_client_os
  acl_access_check
  acl_get_v4_id_domain
  acl_enable_v4_acl
  acl_disable_v4_acl
  flip_mode_bits
  set_user_group
  get_file_owner_group
  get_file_owner
  validate_file_owner
  get_ace_array
  get_users
  wait_for_file
  acl_clear_inheritance
);

#-------------------------------------------------------
# Package global variables
#-------------------------------------------------------
my $VERSION = '0.01';
my $create_vol_called = 0;
my $qtree_security_completed = 0;
my $nfs_export_completed = 0;
my $nfs_mount_completed = 0;
my $main_mount_completed = 0;
my $main_mountpoint;
my $nfsv4_id_domain_client;
my $cifs_share_completed = 0;
my $cifs_map_completed = 0;
my $uid_gid_changed = 0;
my $snapmirror_enabled = 0;
my $snapmirror_licensed = 0;
my $sudo = '/usr/software/bin/sudo';

my $clean_snapshot = '__clean__';
my $filer_vol_path;
my @results;
my @filer_settings;
my $usermap = '/etc/usermap.cfg';
my $filer_usermap_text = 'NOTSET';

# Enable/disable V4 delegations options according to
# NFS_V4_DELEGATIONS parameter, defaults to 'on'
my $NFS_V4_DELEGATIONS =
  ($main::NFS_V4_DELEGATIONS) ? $main::NFS_V4_DELEGATIONS : 'on';

# The options to set on the filer. The key is the option name
# and the value is the value to set the option to. Options not
# set are deleted by the _acl_set_env() so that _acl_clean_env()
# only resets the options that were changed.
my %filer_options = (
    'nfs.v4.enable'           => 'on',
    'nfs.v4.id.domain'        => 'on',
    'nfs.v4.acl.enable'       => 'on',
    'nfs.v4.read_delegation'  => $NFS_V4_DELEGATIONS,
    'nfs.v4.write_delegation' => $NFS_V4_DELEGATIONS,
);

# Test User Accounts
# Note that group name and group id should all be the same
my $test_group_name_unix = 'engr';
my $test_gid = 30;
my @users = (
    {
        user_name  => 'testu1',
        uid        => 3196,
        group_name => $test_group_name_unix,
        gid        => $test_gid,
        cifspass   => 'testp1',
        cifsgroup  => 'Users',
    },
    {
        user_name  => 'testu2',
        uid        => 3197,
        group_name => $test_group_name_unix,
        gid        => $test_gid,
        cifspass   => 'testp2',
        cifsgroup  => 'Users',
    },
    {
        user_name  => 'testu3',
        uid        => 3198,
        group_name => $test_group_name_unix,
        gid        => $test_gid,
        cifspass   => 'testp3',
        cifsgroup  => 'Users',
    },
    {
        user_name  => 'testu4',
        uid        => 3199,
        group_name => $test_group_name_unix,
        gid        => $test_gid,
        cifspass   => 'testp4',
        cifsgroup  => 'Users',
    },
    {
        user_name  => 'testu5',
        uid        => 3200,
        group_name => $test_group_name_unix,
        gid        => $test_gid,
        cifspass   => 'testp5',
        cifsgroup  => 'Users',
    }
);

# Array to track the CIFS users so they can be cleaned up
my @cifs_users_created;

# Constant hashes that should not be modified by the user
# Do not change the ACE hashes unless you are sure that the tests that use
# them will still work.

# ACEs used to initialize the standard ACL for a file
my @file_aces = (
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_READ_DATA | ACE4_WRITE_DATA | ACE4_APPEND_DATA |
                 ACE4_READ_NAMED_ATTRS | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_READ_ATTRIBUTES | ACE4_WRITE_ATTRIBUTES |
                 ACE4_DELETE | ACE4_READ_ACL | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_SYNCHRONIZE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_READ_DATA | ACE4_WRITE_DATA | ACE4_APPEND_DATA |
                 ACE4_WRITE_NAMED_ATTRS | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_DELETE | ACE4_READ_ACL |
                 ACE4_WRITE_ACL,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_WRITE_NAMED_ATTRS | ACE4_EXECUTE |
                 ACE4_WRITE_ATTRIBUTES | ACE4_WRITE_OWNER | ACE4_SYNCHRONIZE,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_DELETE | ACE4_READ_ACL |
                 ACE4_WRITE_ACL,
        who   => 'EVERYONE@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_WRITE_DATA | ACE4_APPEND_DATA | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_WRITE_ATTRIBUTES | ACE4_WRITE_OWNER |
                 ACE4_SYNCHRONIZE,
        who   => 'EVERYONE@'
    }
);

# File inheritable ACEs
my @file_inherit_aces = (
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_FILE_INHERIT_ACE,
        mask  => ACE4_READ_DATA | ACE4_WRITE_DATA | ACE4_APPEND_DATA |
                 ACE4_READ_NAMED_ATTRS | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_READ_ATTRIBUTES | ACE4_WRITE_ATTRIBUTES |
                 ACE4_DELETE | ACE4_READ_ACL | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_FILE_INHERIT_ACE,
        mask  => ACE4_SYNCHRONIZE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_FILE_INHERIT_ACE,
        mask  => ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS | ACE4_READ_ATTRIBUTES |
                 ACE4_READ_ACL | ACE4_WRITE_ACL,
        who => 'EVERYONE@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_FILE_INHERIT_ACE,
        mask  => ACE4_WRITE_DATA | ACE4_APPEND_DATA | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_WRITE_ATTRIBUTES | ACE4_DELETE |
                 ACE4_WRITE_OWNER | ACE4_SYNCHRONIZE,
        who => 'EVERYONE@'
    }
);

# ACEs used to initialize the standard ACL for a dir
my @dir_aces = (
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_LIST_DIRECTORY | ACE4_ADD_FILE | ACE4_ADD_SUBDIRECTORY |
                 ACE4_READ_NAMED_ATTRS | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_DELETE_CHILD | ACE4_READ_ATTRIBUTES |
                 ACE4_WRITE_ATTRIBUTES | ACE4_DELETE | ACE4_READ_ACL |
                 ACE4_WRITE_ACL | ACE4_WRITE_OWNER,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_SYNCHRONIZE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_LIST_DIRECTORY | ACE4_ADD_FILE | ACE4_ADD_SUBDIRECTORY |
                 ACE4_READ_NAMED_ATTRS | ACE4_READ_ATTRIBUTES | ACE4_READ_ACL,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_WRITE_NAMED_ATTRS | ACE4_EXECUTE | ACE4_DELETE_CHILD |
                 ACE4_WRITE_ATTRIBUTES | ACE4_DELETE | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER | ACE4_SYNCHRONIZE,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_LIST_DIRECTORY | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_READ_ACL,
        who   => 'EVERYONE@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => 0,
        mask  => ACE4_ADD_FILE | ACE4_ADD_SUBDIRECTORY |
                 ACE4_WRITE_NAMED_ATTRS | ACE4_EXECUTE | ACE4_DELETE_CHILD |
                 ACE4_WRITE_ATTRIBUTES | ACE4_DELETE | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER | ACE4_SYNCHRONIZE,
        who   => 'EVERYONE@'
    }
);

# Directory inheritable ACEs
my @dir_inherit_aces = (
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_DIRECTORY_INHERIT_ACE,
        mask  => ACE4_LIST_DIRECTORY | ACE4_ADD_FILE | ACE4_ADD_SUBDIRECTORY |
                 ACE4_READ_NAMED_ATTRS | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_EXECUTE | ACE4_DELETE_CHILD | ACE4_READ_ATTRIBUTES |
                 ACE4_WRITE_ATTRIBUTES | ACE4_DELETE | ACE4_READ_ACL |
                 ACE4_WRITE_ACL | ACE4_WRITE_OWNER,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_DIRECTORY_INHERIT_ACE,
        mask  => ACE4_SYNCHRONIZE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_DIRECTORY_INHERIT_ACE,
        mask  => ACE4_LIST_DIRECTORY | ACE4_ADD_FILE | ACE4_ADD_SUBDIRECTORY |
                 ACE4_READ_NAMED_ATTRS | ACE4_READ_ATTRIBUTES |
                 ACE4_READ_ACL | ACE4_WRITE_ACL,
        who => 'EVERYONE@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flags => ACE4_INHERIT_ONLY_ACE | ACE4_DIRECTORY_INHERIT_ACE,
        mask  => ACE4_WRITE_NAMED_ATTRS | ACE4_EXECUTE | ACE4_DELETE_CHILD |
                 ACE4_WRITE_ATTRIBUTES | ACE4_DELETE | ACE4_WRITE_OWNER |
                 ACE4_SYNCHRONIZE,
        who => 'EVERYONE@'
    }
);

## @name ::run_remote
## @description
##  Use subtest to run the specified test on the specified clients.
##  This subroutine will do the setup and cleanup for the test.
##
## @arg %opts	Required: The hash of options that include:
##
##  client_list		Required: List of clients to run test on.
##
##  id_domain		Required: NFS V4 id domain to use.
##
##  max_run_time	Optional: Maximum number of seconds to wait until
##			a test finishes. Default is 1800 seconds.
##
##  nis_domain		Required: NIS domain for both filer and client.
##
##  nis_domain_server	Required: NIS domain server for filer. Only used
##			if the filer NIS domain is changed.
##
##  num_users		Optional: Number of test user accounts on the filer
##			machine required for the test. Defaults to 0.
##
##  qtree_security	Optional: Security setting for a qtree. Defaults
##			to unix.
##
##  snapmirror		Optional: Values are enable or disable. Enable/disable
##			snapmirror feature. Default is to not enable it.
##
##  test_msg		Required: Test specific result message to log.
##
##  test_script		Required: Relative path of test thpl script
##			to run on all the clients.
##
##  username		Optional: Username to run tests as on all the clients.
##
##  unix_only		Optional: If set to yes, indicates that all clients
##			must be UNIX machines. Default is no.
##
##  secondary_filers    Optional: Array reference to hostspecs of filers
##                      other than $main::FILER that need to have the
##                      environment set. This only includes setting
##                      nfs.v4 options, acl options, and CIFS related
##                      setup. Volume creation/export/sharing is only
##                      done on $main::FILER
##
## @return
##  nothing

sub run_remote( % ) {
    my %opts = validate(
        @_,
        {
            client_list       => { type => SCALAR },
            id_domain         => { type => SCALAR },
            max_run_time      => { type => SCALAR, default => 1800 },
            nis_domain        => { type => SCALAR },
            nis_domain_server => { type => SCALAR },
            num_users         => { type => SCALAR, default => 0 },
            qtree_security    => { type => SCALAR, default => 'unix' },
            snapmirror        => { type => SCALAR, default => 'disable',
                                   regex => qr/^(enable|disable)$/ },
            test_msg          => { type => SCALAR },
            test_script       => { type => SCALAR },
            username          => { type => SCALAR, optional => 1 },
            unix_only         => { type => SCALAR, default => 'no' },
            secondary_filers  => { type => ARRAYREF, optional => 1 },
        }
    );
    my (@clients, $clnt, $err, $fc, $h, $i, $msg, @null_arg, $ping,
        @res, $rslt_obj, $snap, $sobj, @subtests, $uref, @user_arg, $worst);

    logStep("Setup Test Environment");

    $filer_vol_path = "/vol/$main::FILER_VOL";

    eval {
        $fc = _acl_set_env(id_domain         => $opts{id_domain},
                           nis_domain        => $opts{nis_domain},
                           nis_domain_server => $opts{nis_domain_server},
                           num_users         => $opts{num_users},
                           qtree_security    => $opts{qtree_security},
                           secondary_filers  => $opts{secondary_filers});

        if ($opts{snapmirror} eq 'enable') {
            if ( ! is_licensed($fc, 'snapmirror') ) {
                ## @step License snapmirror
                logStep('License snapmirror');
                license($fc, 'snapmirror') or croak $@;
                $snapmirror_licensed = 1;
            }

            ## @step Enable snapmirror
            logStep('Enable snapmirror');
            my $tmp = filer_options_get($fc, 'snapmirror.enable');
            if ($tmp eq 'off') {
                snapmirror_enable(host => $fc->hostp());
                $snapmirror_enabled = 1;
            }
            $snap = 1;
        } else {
            $snap = 0;
        }
    };
    if ($@) {
        $err = $@;
        _acl_clean_env();
        logresult('CONFIG', $err);
        return undef;
    }

    if (defined($opts{username})) {
        @user_arg = ( '-username', $opts{username} );
    }

    @clients = split(' ', $opts{client_list});

    # If the client_list was empty then set it to localhost
    if ($#clients == -1) {
        @clients = ('localhost');
    }

    # Make sure any CIFS clients on the list are minwin compatabile
    $msg = '';
    for $clnt (@clients) {
        $h = host($clnt);
        $ping = undef;
        eval {
            $ping = host_is_alive($clnt, 5);
        };
        if (!$ping || $@) {
            $msg .= 'The host ' . $h->hostname() .
                    " is not responding to a ping: $@\n";
        }
        if ($h->hosttype() !~ /unix/i && $opts{unix_only} eq 'yes') {
            $msg .= 'The host ' . $h->hostname() .  " is not running UNIX\n";
        }
        if ($h->hosttype() =~ /windows/i) {
            if (!minwin_compatible($clnt)) {
                $msg .= 'The host ' . $h->hostname() .
                        " is not MinWin compatible\n";
            }
        }
    }
    if ($msg ne '') {
        logresult('CONFIG', $msg);
        return undef;
    }

    $i = 1;
    for $clnt (@clients) {
        $h = host($clnt);

        # If the hosttype is windows then do not set the -username
        # parameter because NATE trys to use runas and it isn't
        # successful causing the test to fail.
        if ($h->hosttype() =~ /windows/i) {
            $uref = \@null_arg;
        } else {
            $uref = \@user_arg;
        }
        eval {
            push @subtests, subtest($opts{test_script},
                                    -bg,
                                    -client, $clnt,
                                    @$uref,
                                    '-onexit', \&_exit_handler,
                                    "CLIENT_LIST=localhost",
                                    "CLIENT_NUM=$i",
                                    "SNAPMIRROR=$snap",
                                    "NFS_V4_ID_DOMAIN=$opts{id_domain}",
                                    "NIS_DOMAIN=$opts{nis_domain}");
        };
        # Log any errors from the subtest call
        if ($@) {
            push @results, add_result($@, undef);
        }
        $i++;
    }

    # Eval is needed here because if one of the subtest croaks
    # the wait_finish will croak and we want to call our cleanup.
    # Do not care about error returned from wait_finish as the
    # subtest error will have been logged.
    logStep("Wait for the tests to finish");
    foreach $sobj (@subtests) {
        eval {
            $sobj->wait_finish(timeout => $opts{max_run_time},
                            iftimeout => ['-fatal',
                            "Test timeout after $opts{max_run_time} seconds"]);
        };
        if ($@) {
            if (!ref($@)) {
                # Results object
                push @results, $@;
            } elsif ($@) {
                # Not a results object
                push @results, fatal_result($@);
            }
        }

        # If for some reason the subtest is still running then kill it.
        # Try 5 times before giving up.
        $i = 5;
        while ($sobj->is_alive() && $i-- > 0) {
            logStep('Forced kill of subtest after waiting for completion');
            $sobj->kill();
            # Yield the processor, let subtest die
            sleep(2);
        }

        # Save any results returned by the subtest
        @res = $sobj->own_results();
        if ($#res != -1) {
            push @results, @res;
        }
    }

    _acl_clean_env();

    # Log all low severity results,
    # except for pass results
    foreach my $rslt (@results) {
        if ( $rslt->severity() < 2 and $rslt->type() ne 'PASS' ) {
            logresult($rslt->type(), $rslt->messages());
        }
    }

    # Get worst result with FAIL results taking precedence over FATAL
    # and a result with a Burt taking precedence over everything
    $worst = pop @results;
    if (defined $worst && join(" ", $worst->messages()) !~ /Burt/i) {
        foreach my $rslt (@results) {
            if (join(" ", $rslt->messages()) =~ /Burt/i) {
                $worst = $rslt;
                last;
            }
            if (($rslt->severity() > $worst->severity()) ||
                ($worst->type() =~ /FATAL/i && $rslt->type() =~ /FAIL/i)) {
                $worst = $rslt;
            }
        }
    }

    if (!defined($worst)) {
        # This happens when the subtest never executes the test
        # due to bad parameters, IPC issues, ...
        logresult('SCRIPT', 'No results were logged');
    } elsif ( $worst->severity() > 2 ) {
        # Log worst high severity result
        logresult($worst->type(), $worst->messages());
    } else {
        logresult('PASS', $opts{test_msg});
    }

    return;
}

## @name ::run_local
## @description
##  Run the the specified test subroutine locally.
##  This subroutine will do the setup and cleanup for the test.
##  The test subroutine will be called with the mountpoint
##  passed in as the only parameter.
##
## @arg %opts	Required: The hash of options that include:
##
##  cleanup		Optional: Reference to a function to run during
##			the cleanup phase. This routine is only called
##			if test_script was exectued.
##
##  mount_suffix	Optional: Path to append to the default mountpoint.
##
##  mount_sharing	Optional: Indicates the type of mount sharing to use.
##
##  num_users		Optional: Number of test user accounts on the local
##			machine required for the test. Defaults to 0.
##
##  test_msg		Required: Test specific result message to log.
##
##  test_script		Required: Reference of the test subroutine to invoke.
##
## @return
##  nothing

sub run_local( % ) {
    my %opts = validate(
        @_,
        {
            cleanup       => { type => CODEREF, optional => 1 },
            mount_suffix  => { type => SCALAR, optional => 1 },
            mount_sharing => { type => SCALAR, default => 'exclusive',
                               regex => qr/^(shared|exclusive)$/ },
            num_users     => { type => SCALAR, default => 0 },
            test_msg      => { type => SCALAR },
            test_script   => { type => CODEREF },
        }
    );
    my ($clean_env, $err, $fc, $h, $i, $mountpoint, $rslt_obj, $saved_uid,
        $saved_gid, $status, $user_entry);

    logStep("Setup Local Test Environment");

    $filer_vol_path = "/vol/$main::FILER_VOL";

    # The ACL tests may change the uid/gid. Save them here and restore
    # them after the test run.
    $saved_uid = $>;
    ($saved_gid) = split(" ", $), 2);

    eval {
        acl_validate_client_os();

        if ($opts{num_users} > 0) {
            logStep("Validate the test user accounts on client");
            $status = _validate_users(num_users => $opts{num_users});
            if ($status != 0) {
                croak "test user account validation failed";
            }
        }

        logStep("Connect to the filer");
        $h = host($main::FILER);
        $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
        $fc->timeout($main::CMDTIMEOUT);

        # If it is a windows client then do CIFS related testing.
        # Otherwise assume that it is a NFS client
        $h = host('localhost');
        if ($h->hosttype() !~ /windows/i) {
            $clean_env = \&_acl_clean_nfs_env;
            my %set_opts;
            $set_opts{mount_suffix} = $opts{mount_suffix} if $opts{mount_suffix};
            $set_opts{mount_sharing} = $opts{mount_sharing};
            $mountpoint = _acl_set_nfs_env(fc => $fc, %set_opts);
        } else {
            $clean_env = \&_acl_clean_cifs_env;
            $mountpoint = _acl_set_cifs_env(fc => $fc);
        }
    };
    if ($@) {
        $err = $@;
        if (defined($clean_env)) {
            &{$clean_env}(mountpoint => $mountpoint);
        }
        logresult('CONFIG', $err);
        return;
    }

    eval {
        &{$opts{test_script}}($filer_vol_path, $mountpoint);
    };

    $rslt_obj = $@;

    eval {
        # If the uid changed, then restore it before doing the cleanup
        # Ignore any failures and still try to do the cleanup.
        if ($uid_gid_changed != 0) {
            logStep("Reset the username");
            $h = host('localhost');
            if ($h->hosttype() !~ /windows/i) {
                $user_entry = {
                    user_name => '',
                    uid => $saved_uid,
                    group_name => '',
                    gid => $saved_gid
                };
            } else {
                $user_entry = {
                    user_name => $main::CIFS_USER,
                    uid => 0,
                    group_name => 'Administrators',
                    gid => 0,
                    cifspass => $main::CIFS_PASSWD
                };
            }
            set_user_group(mountpoint => $mountpoint,
                           user_entry => $user_entry);
        }
    };
    logWarn("run_local: $@") if $@;

    if (defined($clean_env)) {
        &{$clean_env}(cleanup => $opts{cleanup},
                      mountpoint => $mountpoint);
    }

    if ( ref($rslt_obj) ) {
        # Results object
        logresult($rslt_obj);
    } elsif ($rslt_obj) {
        # Not a results object
        logresult('FATAL', $rslt_obj);
    } else {
        logresult('PASS', $opts{test_msg});
    }
    return;
}

# @name ::_exit_handler
# @description
#  Exit handler for subtest: it appends all results to a global
#  list of results to be processed after all subtests are done.
#  INTERNAL SUBROUTINE
#
# @arg
#
#  sobj			Required: Subtest object reference.
#
# @return
#  nothing

sub _exit_handler($;) {
    my $sobj = shift;

    # Append results from this subtest to list
    push @results, $sobj->results();
    return;
}

# @name ::_acl_set_env
# @description
#  Set up environment for NFS ACL tests. That includes creating
#  the volume on the filer, a snapshot used for cleanup, exporting the
#  volume, and sharing it.
#  Internal function
#
# @arg %opts	Required: The hash of options that include:
#
#  id_domain		Required: NFS V4 ID domain to use.
#
#  nis_domain		Required: NIS domain for both filer and client.
#
#  nis_domain_server	Required: NIS domain server for filer. Only used
#			if the filer NIS domain is changed.
#
#  num_users		Optional: Number of test user accounts on the filer
#			machine required for the test. Defaults to 0.
#
#  qtree_security	Required: Security setting for a qtree. Defaults
#			to unix.
#
#  secondary_filers    Optional: Array reference to hostspecs of filers
#                      other than $main::FILER that need to have the
#                      environment set. This only includes setting
#                      nfs.v4 options, acl options, and CIFS related
#                      setup. Volume creation/export/sharing is only
#                      done on $main::FILER
#
# @return
#  Connection to the filer on error croaks
#
# @notes
#  After successful completion the volume will be exported from
#  the filer. This routine should only be called once per test
#  not for each subtest invocation of a test.

sub _acl_set_env( % ) {
    my %opts = validate(
        @_,
        {
            id_domain         => { type => SCALAR },
            nis_domain        => { type => SCALAR },
            nis_domain_server => { type => SCALAR },
            num_users         => { type => SCALAR, default => 0 },
            qtree_security    => { type => SCALAR, default => 'unix' },
            secondary_filers  => { type => ARRAYREF|UNDEF, optional => 1 },
        }
    );
    my ($cifs_on, $conn, $fc, $filer, @filers, $host, $hostname, $i,
        $msg, $new_text, $out, $option, $tmp, $volume);

    logStep("BEGIN: _acl_set_env: setup the environment");

    logStep("Validate the client is a unix client");
    # We require the initial test invocation to be from a unix host
    # to ensure setup is completed correctly. We need a unix host
    # to mount and set permissions on the volume.
    if (host('localhost')->hosttype() !~ /unix/i) {
        croak "Test must be run from a unix host, even if the test " .
          "client(s) are CIFS only. This is to ensure the environment " .
            "is set correctly";
    }

    logStep("Create the volume");
    $create_vol_called = 1;
    ($fc, $volume) = create_vol() or croak $@;

    # create a snapshot of FILER_VOL that we can revert to during cleanup
    if (snapshot_exists($fc, $volume, $clean_snapshot)) {
        snapshot_delete($fc, $volume, $clean_snapshot);
        croak($@) if $@;
    }
    snapshot_create($fc, $volume, $clean_snapshot);
    croak($@) if $@;

    # Set the qtree security mode on the filer volume
    if (filer_qtree_security_set($fc, $filer_vol_path, $opts{qtree_security})
        == 0) {
        croak("QTREE set failed on $filer_vol_path for $opts{qtree_security}");
    }
    $qtree_security_completed = 1;

    @filers = ($main::FILER);
    if (defined $opts{secondary_filers}) {
        push @filers, @{$opts{secondary_filers}};
    }

    # Copy the NFS V4 ID and NIS domain into the filer options that
    # need to be set
    $filer_options{'nfs.v4.id.domain'} = $opts{id_domain};
    $filer_options{'nis.domainname'} = $opts{nis_domain};

    foreach $filer (@filers) {
        $host = host($filer);
        $conn = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
        $conn->timeout($main::CMDTIMEOUT);
        $hostname = $conn->hostp()->hostname();

        my %filer_info = %filer_options; # Copy the hash

        # For NIS see if the current domain is different, if so then
        # add the NIS servers to the hash so they get set.
        $tmp = filer_options_get($conn, 'nis.domainname');
        if ($tmp ne $opts{nis_domain}) {
            $filer_info{'nis.servers'} = $opts{nis_domain_server};
        }

        logStep("Change options on the filer as needed");
        foreach $option (keys %filer_info) {
            # Save current value of option
            $tmp = filer_options_get($conn, $option);

            if ($filer_info{$option} ne $tmp) {
                # Set option to value specified
                filer_options_set($conn, $option, $filer_info{$option});
                # Save the original option for resetting
                $filer_info{$option} = $tmp;
            } else {
                # Option was not set, remove it so it does not get cleaned up
                delete $filer_info{$option};
            }
        }

        # All NIS related options need to be set first, in loop above,
        # before enabling NIS if needed.
        logStep("Enable NIS on filer if needed");
        $tmp = filer_options_get($conn, 'nis.enable');
        if ($tmp eq 'off') {
            filer_options_set($conn, 'nis.enable', 'on');
        }

        $filer_info{filer} = $filer;
        $filer_info{fname} = $hostname;
        $filer_info{filer_usermap_text} = 'NOTSET';

        push @filer_settings, \%filer_info;
    }

    # File export/sharing based on qtree type
    #          export    share
    #  unix     Yes       No
    #  ntfs     No        Yes
    #  mixed    Yes       Yes
    if ($opts{qtree_security} eq 'mixed' || $opts{qtree_security} eq 'unix') {
        logStep("Export the volume from the filer");
        export_path($fc, $filer_vol_path, OPTIONS=>"rw,anon=0");
        croak $@ if $@;
        $nfs_export_completed = 1;
    }
    if ($opts{qtree_security} eq 'mixed' || $opts{qtree_security} eq 'ntfs') {
        # setup cifs on all filers
        foreach $filer (@filer_settings) {
            $main::CIFS_DOMAIN =~ s/\s+//g;
            if ($main::CIFS_DOMAIN eq '') {
                # Using WORKGROUP
                # Create a mapping from CIFS_USER to root on the filer
                $new_text = "\n$main::CIFS_USER == root\n";

                logStep("Setup $usermap file on $filer->{fname}");
                $filer_usermap_text = rdfile($conn, $usermap);
                $filer->{filer_usermap_text} = $filer_usermap_text;
                wrfile($conn, $usermap, "$filer_usermap_text$new_text")
                  or croak("can't write $usermap: $@");

            }
            logStep('Setup CIFS on the filer');
            _acl_cifs_setup(filer => $filer->{filer},
                            mode  => 'multi');
            # Set a flag so we know to check user accounts on the filer
            # on the CIFS side.
            $cifs_on = 1;
        }

        # Create the share only on the main test filer
        logStep('Create the share on the filer');
        share_add($fc, $main::CIFS_SHARE, $filer_vol_path,
                  caching => 'no', verify => 1, purge_existing => 1, umask => 0)
            or croak("failed to share $main::CIFS_SHARE");
        $cifs_share_completed = 1;
    }

    # Validate the current user and the required test users on the filer
    $msg = '';
    logStep('Verify filer knows current user from NFS side');
    $out = run_getXXbyYY($fc, 'getpwbyuid_r', $>);
    if ( $out !~ /pw_name = / ) {
        $msg .= "current user not resolvable filer NFS side uid: $>\n";
    }

    if ($opts{num_users} > 0) {
        logStep("Validate the test user accounts on filer");
        $i = @users;
        if ($opts{num_users} > $i) {
            croak "Too many users requests: req: $opts{num_users} avail: $i";
        }

        $filer = $filer_settings[0];
        for ($i = 0; $i < $opts{num_users}; $i++) {
            $out = run_getXXbyYY($fc, 'getpwbyname_r', $users[$i]->{user_name});
            if ( $out !~ /pw_name = / ) {
                $msg .= "account $users[$i]->{user_name} not resolvable " .
                        "filer NFS side\n";
            }
            if ($cifs_on) {
                # Check if the cifs user can be looked up. If not and
                # using CIFS_DOMAIN then generate an error. If not and
                # using WORKGROUP then just create the CIFS user on the filer.
                $out = filer_cifs_lookup($fc, $users[$i]->{user_name});
                if ($out =~ /lookup failed/i) {
                    if ($main::CIFS_DOMAIN ne '') {
                        $msg .= "test user not resolvable filer CIFS side " .
                                "name: $users[$i]->{user_name}\n";
                    } else {
                        # Test account passwords may not meet filer
                        # minimum password length requirements. Just set
                        # the minimum length to 3.
				        if (!$filer->{'security.passwd.rules.minimum'}) {
                            $filer->{'security.passwd.rules.minimum'} =
        	                filer_options_get($fc, 'security.passwd.rules.minimum'); 
						    $filer->{'security.passwd.rules.maximum'} =
							filer_options_get($fc, 'security.passwd.rules.maximum');
							filer_options_set($fc, 'security.passwd.rules.maximum', '255');
							filer_options_set($fc, 'security.passwd.rules.minimum', '3'); 
					    } 
                        addCIFSUser($fc, $users[$i]->{user_name},
                                    $users[$i]->{cifspass},
                                    $users[$i]->{cifsgroup});
                        if ($@) {
                            $msg .= "$@\n";
                        } else {
                            push @cifs_users_created, $users[$i]->{user_name};
                        }
                    }
                }
            }
        }
    }
    croak $msg if $msg;

    # Open the permissions on the volume so that test users
    # have access to it. Set the mode to 777 here, rather than
    # when each client mounts the volume. This avoids multiple clients
    # all doing the same thing, and facilitates granting general
    # permissions for cifs clients. Since this mount is performed on
    # whatever client the test was started on then use NFS V3 for the
    # mount so that clients that do not have a good NFS V4 client can
    # still be used to initiate ACL tests (i.e. where ntest is run).

    logStep("Mount the volume and set the permissions to 0777");
    $main_mountpoint = mount($fc->connhost(), "/vol/$volume",
                             chmod => '0777', mntopts => "vers=3,actimeo=0");
    croak $@ if $@;
    if (!defined($main_mountpoint)) {
        croak "_acl_set_env: Mount failed\n";
    }
    $main_mount_completed = 1;

    # Unmount the volume
    umount($main_mountpoint);
    # If the unmount succeeds, we don't need to clean it up later
    # If it fails, we'll just try to clean up at the end of the test
    $main_mount_completed = 0 if !$@;

    logStep("END: _acl_set_env:");

    return ($fc);
}

# @name ::_acl_set_nfs_env
# @description
#  Set up environment for the current NFS client (localhost).
#
# @arg %opts	Required: The hash of options that include:
#
#  fc			Connection to filer
#
#  mount_sharing	Optional: Indicates the type of mount sharing to use.
#			The default is exclusive.
#
#  mount_suffix	Optional: Path to append to the default mountpoint.
#
# @return
#  The full path of the mountpoint.
#
# @notes
#  After successful completion the volume will be mounted from this client.

sub _acl_set_nfs_env( % ) {
    my %opts = validate(
        @_,
        {
            fc            => { type => OBJECT },
            mount_sharing => { type => SCALAR,
                               regex => qr/^(shared|exclusive)$/ },
            mount_suffix  => { type => SCALAR, optional => 1 },
        }
    );
    my (@mountarg, $mountpoint, $out, $ret_mountpoint);

    logStep("BEGIN: _acl_set_nfs_env:");

    # Make sure the client is in the same NIS domain as what
    # the test asked for
    logStep("Verify clients NIS domain");
    $out = qx(domainname);
    chomp $out;
    if ($out ne $main::NIS_DOMAIN) {
        croak "_acl_set_nfs_env: client NIS domain: $out NIS_DOMAIN: " .
              "$main::NIS_DOMAIN";
    }

    # Set the clients NFS V4 ID domain if it is not set to what
    # the test asked for
    logStep("Set clients NFS V4 ID domain if needed");
    $nfsv4_id_domain_client =
        _acl_set_client_nfs_id_domain($main::NFS_V4_ID_DOMAIN);

    if ($opts{mount_sharing} eq 'exclusive') {
        $mountpoint = "/t/$main::FILER_VOL/ACL";
        if (defined($opts{mount_suffix})) {
            $mountpoint .= "/$opts{mount_suffix}";
        }
        @mountarg = ( 'mountpoint', "$mountpoint" );
    }

    logStep("Have localhost mount exported volume");

    # All clients will be using the same exported volume for testing
    # so pass in the mountpoint to mount which also gives exclusive
    # access on a per client basis.
    $ret_mountpoint = mount($opts{fc}->connhost(), $filer_vol_path,
                            @mountarg, mntopts => "vers=4,actimeo=0",
                            sharing => $opts{mount_sharing}); croak $@ if $@;
    if (!defined($ret_mountpoint)) {
        croak "_acl_set_nfs_env: Mount failed\n";
    }
    $nfs_mount_completed = 1;

    logStep("END: _acl_set_nfs_env:");
    return $ret_mountpoint;
}

# @name ::_acl_set_cifs_env
# @description
#  Set up environment for current CIFS client (localhost).
#
# @arg %opts	Required: The hash of options that include:
#
#  fc			Connection to filer
#
# @return
#  The full path of the mountpoint.
#
# @notes
#  After successful completion the volume will be mapped from this client.

sub _acl_set_cifs_env( % ) {
    my %opts = validate(
        @_,
        {
            fc           => { type => OBJECT },
        }
    );
    my ($cifs_drive);

    logStep("BEGIN: _acl_set_cifs_env:");

    my $host = $opts{fc}->connhost();
    $cifs_drive = map_cifs_share(
                server   => $host,
                share    => $main::CIFS_SHARE,
                client   => 'localhost',
                drive    => $main::CIFS_DRIVE,
                username => $main::CIFS_USER,
                password => $main::CIFS_PASSWD
                )
            or croak($@);
    logDbg("mapped drive $cifs_drive");
    if (!$cifs_drive) {
        croak("_acl_set_cifs_env: could not map the drive");
    }

    $cifs_drive .= ":" unless $cifs_drive =~ /:$/;
    logDbg("Share $main::CIFS_SHARE of $host mounted as $cifs_drive");

    $cifs_map_completed = 1;

    logStep("END: _acl_set_cifs_env:");
    return $cifs_drive;
}

# @name ::_acl_clean_env
# @description
#  Global cleanup after a test run if necessary. Use this cleanup
#  routine if you use _acl_set_env.
#
# @arg
#  None
#
# @return
#  Nothing, on error logs a warning.

sub _acl_clean_env() {
    my ($conn, $filer, $fname, $fc, $fh, $i, $msg, $option, $status);

    logStep("CLEANUP == $main::CLEANUP");

    eval { transit_to(host($main::FILER), 'UP', timeout => 1200); };
    logWarn("_acl_clean_env: $@") if $@;

    if ($main::FILER_MODE eq 'cf') {
        logStep('Try to leave cluster in a usable state');
        eval { setup_cluster(); };
        logWarn("_acl_clean_env: $@") if $@;
        eval { cf_disable($main::FILERA); };
        logWarn("_acl_clean_env: $@") if $@;
    }

    eval {
        logStep("Reconnect to the filer");
        eval {
            $fh = host($main::FILER);
            $fc = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
            $fc->timeout($main::CMDTIMEOUT);
        };
        if ($@) {
            undef $fc;
            logWarn("_acl_clean_env: $@");
        }

        if ( $main::CLEANUP ne 'none' ) {
            if ($main_mount_completed && defined $main_mountpoint) {
                logStep("Unmount $main_mountpoint");
                umount($main_mountpoint);
                logWarn("_acl_clean_env: umount $main_mountpoint failed: $@")
                  if $@;
            }

            if ($fc && $nfs_export_completed != 0) {
                logStep("Unexport $filer_vol_path from the filer");
                eval {
                    unexport_path ($fc, $filer_vol_path);
                };
                logWarn("_acl_clean_env: $@") if $@;
            }

            if ($fc && $cifs_share_completed != 0) {
                logStep("unsharing the CIFS volume");
                eval {
                    share_delete($fc, $main::CIFS_SHARE);
                };
                logWarn("_acl_clean_env: $@") if $@;
            }

            # Restore the clean snapshot if it exists.
            if (snapshot_exists($fc, $main::FILER_VOL, $clean_snapshot)) {
                my $lic = 0;
                logStep("Restore volume to $clean_snapshot");
                if (!is_licensed($fc, 'snaprestore')) {
                    license($fc, 'snaprestore') or logWarn ($@);
                    $lic = 1;
                }
                snaprestore_volume($fc, $main::FILER_VOL, $clean_snapshot);
                logWarn($@) if $@;
                if ($lic) {
                    unlicense($fc, 'snaprestore') or logWarn ($@);
                }
            }

            if ( $fc && $#cifs_users_created != -1 ) {
                logStep("Remove CIFS users added to the filer");
                eval {
                    foreach my $user (@cifs_users_created) {
                        # Remove CIFS user from the filer
                        removeCIFSUser($fc, $user);
                    }
                };
                logWarn("acl_clean_env: $@") if $@;
            }

            if ($snapmirror_enabled == 1) {
                ## @step Disable snapmirror
                logStep('Disable snapmirror');
                eval { snapmirror_disable(host => $fc->hostp()) };
                logWarn("_acl_clean_env: $@") if $@;
            }

            if ( $snapmirror_licensed ) {
                ## @step Unlicense snapmirror
                logStep('Unlicense snapmirror');
                unlicense($fc, 'snapmirror') or logWarn("_acl_clean_env: $@");
            }

            # For each filer restore the usermap.cfg file if needed and
            # restore whatever options were changed.
            foreach $filer (@filer_settings) {
                $fname = $filer->{fname};
                logStep("Reconnect to the filer: $fname");
                host($filer->{filer});
                $conn = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
                $conn->timeout($main::CMDTIMEOUT);

                if ($conn && $filer->{filer_usermap_text} ne 'NOTSET') {
                    logStep("_acl_clean_env: Restore $usermap file on filer");
                    wrfile($conn, $usermap, $filer->{filer_usermap_text})
                      or logWarn("_acl_clean_env: filer $fname: " .
                                 "can't write $usermap: $@");
                }

                # Delete the keys we do not need
                delete $filer->{filer};
                delete $filer->{fname};
                delete $filer->{filer_usermap_text};
                foreach $option ( keys %{$filer} ) {
                    eval { filer_options_set($conn, $option,
                                             $filer->{$option}) };
                    logWarn("_acl_clean_env: $@") if $@;
                }
            }
        }

        # Call common cleanup if create_vol was called
        if ($fc && $create_vol_called != 0) {
            # Set the qtree security mode on the filer volume back to unix
            if ($qtree_security_completed != 0) {
                logStep("Set qtree security mode volume back to unix");
                eval {
                    filer_qtree_security_set($fc, $filer_vol_path,
                                             'unix');
                };
                logWarn("_acl_clean_env: $@") if $@;
            }

            logStep("Run common cleanup");
            common_cleanup($fc);
            logWarn("_acl_clean_env: $@") if $@;
        }
    };
    logWarn("_acl_clean_env: $@") if $@;
    logStep("END: _acl_clean_env:");
    return;
}

# @name ::_acl_clean_nfs_env
# @description
#  Local clean up environment for the current NFS client (localhost).
#  Use this cleanup routine if you use _acl_set_nfs_env.
#
# @arg %opts	Required: The hash of options that include:
#
#  cleanup		Optional: Reference to a function to run during
#			the cleanup phase. This routine is only called
#			if the mount was completed.
#
#  mountpoint		Optional: Mountpoint to mount on.
#
# @return
#  Nothing, on error logs a warning.

sub _acl_clean_nfs_env( % ) {
    my %opts = validate(
        @_,
        {
            cleanup    => { type => CODEREF | UNDEF, optional => 1 },
            mountpoint => { type => SCALAR | UNDEF, optional => 1 },
        }
    );
    logStep("BEGIN: _acl_clean_nfs_env:");

    eval {
        if ($main::CLEANUP ne 'none') {
            if ($nfs_mount_completed != 0) {
                if ( $opts{cleanup} ) {
                    # Run specific cleanup for test
                    # Eval the specific cleanup function just in
                    # case the function does not return eval, so
                    # the error can be properly log as a warning.
                    eval {
                        &{$opts{cleanup}}();
                        # Just in case the function does a return eval
                        croak $@ if $@;
                    };
                    logWarn("_acl_clean_nfs_env: $@") if $@;
                }
            }
            if ($nfs_mount_completed != 0 && $opts{mountpoint}) {
                logStep("unmounting the NFS mountpoint");
                umount($opts{mountpoint});
                logWarn("_acl_clean_nfs_env: $@") if $@;
            }
            if (defined($nfsv4_id_domain_client)) {
                eval { _acl_set_client_nfs_id_domain($nfsv4_id_domain_client) };
                logWarn("_acl_clean_nfs_env: $@") if $@;
            }
        }
    };
    logWarn("_acl_clean_nfs_env: $@") if $@;
    logStep("END: _acl_clean_nfs_env:");
    return;
}

# @name ::_acl_clean_cifs_env
# @description
#  Local clean up environment for the current NFS client (localhost).
#  Use this cleanup routine if you use _acl_set_cifs_env.
#
# @arg %opts	Required: The hash of options that include:
#
#  cleanup		Optional: Reference to a function to run during
#			the cleanup phase. This routine is only called
#			if the mount was completed.
#
#  mountpoint		Optional: Mountpoint to mount on.
#
# @return
#  Nothing, on error logs a warning.

sub _acl_clean_cifs_env( % ) {
    my %opts = validate(
        @_,
        {
            cleanup    => { type => CODEREF | UNDEF, optional => 1 },
            mountpoint => { type => SCALAR | UNDEF, optional => 1 },
        }
    );
    logStep("BEGIN: _acl_clean_cifs_env:");

    eval {
        if ($main::CLEANUP ne 'none') {
            if ($cifs_map_completed != 0) {
                if ( $opts{cleanup} ) {
                    # Run specific cleanup for test
                    # Eval the specific cleanup function just in
                    # case the function does not return eval, so
                    # the error can be properly log as a warning.
                    eval {
                        &{$opts{cleanup}}();
                        # Just in case the function does a return eval
                        croak $@ if $@;
                    };
                    logWarn("_acl_clean_cifs_env: $@") if $@;
                }
            }

            if ($cifs_map_completed != 0 && $opts{mountpoint}) {
                logStep("unmapping the CIFS mountpoint");
                eval {
                    unmap_cifs_share(mappath => $opts{mountpoint});
                };
                logWarn("_acl_clean_cifs_env: $@") if $@;
            }
        }
    };
    logWarn("_acl_clean_cifs_env: $@") if $@;
    logStep("END: _acl_clean_cifs_env:");
    return;
}

## @name ::acl_validate_client_os
## @description
##  Makes sure that the client operating system that we are running
##  on is at the right level to run the ACL tests. Croaks if the client
##  fails validation
##
## @return
##  Nothing

sub acl_validate_client_os {
    my ($h, $hostname, $i, @osrel, $rel, $release, $subrel, $type, $version);

    # Get the hosttype
    $h = host('localhost');
    $type = $h->hosttype();

    (undef, $hostname, $release, $version, undef) = POSIX::uname();
    if (!defined($release)) {
        croak "acl_validate_client_os: Could not get OS uname";
    }
    logStep("Validate client operating system correct for ACL testing:\n" .
            "\ttype: $type\n\thost: $hostname\n\trelease: $release\n" .
            "\tversion: $version");

    if ($type =~ /unix-solaris/i) {
        # Solaris 5.10 or later

        ($rel, $subrel) = split(/\./, $release);
        if ($rel < 5 || ($rel == 5 && $subrel < 10)) {
            croak "acl_validate_client_os: Solaris $release not supported " .
                  "by NFS V4 ACL tests. Must use 5.10 or later\n";
        }
    } elsif($type =~ /unix-linux/i) {
        # Linux releases post 2.6.18 (RHEL5) support NFSv4 Acls.

        my @rel_sup = (2, 6, 18);
        # Note that a release does not have to have 3 numbers, we will check
        # up to first 3 numbers. It does have to have at least 2 numbers.
        @osrel = $release =~ /^(\d+)\.(\d+)(?:\.(\d+))?.*/;
        if (@osrel == 0) {
            croak "acl_validate_client_os: Could not handle Linux OS Release format";
        }
        for ($i = 0; $i < @rel_sup; $i++) {
            if ($osrel[$i] < $rel_sup[$i]) {
                croak "acl_validate_client_os: Only Linux releases post " .
                      "2.6.18 are supported by NFSv4 ACL tests. Prior " .
                      "releases such as $release are not!!!\n";
            }
            next if ($osrel[$i] == $rel_sup[$i]);
            last if ($osrel[$i] > $rel_sup[$i]);
        }
    } elsif ($type =~ /unix-aix/i) {
        # AIX 5.3 or later

        if (!defined($version)) {
            croak "acl_validate_client_os: Could not get OS version on AIX";
        }
        if ($version < 5 || ($version ==5 && $release < 3)) {
            croak "acl_validate_client_os: AIX $version.$release not " .
                "supported by NFS V4 ACL tests. Must use 5.3 or later\n";
        }
    } elsif ($type =~ /windows/i) {
        # Any release of windows will work.
        ;
    } else {
        croak "acl_validate_client_os: $type not supported by " .
              "NFS V4 ACL tests\n";
    }
    return;
}

# @name ::_set_v4_id_domain_solaris
# @description
#  Set the NFS V4 ID domain on Solaris. If the domain is already set to the
#  value passed in then it just returns. Croak if it fails to set it.
#  INTERNAL SUBROUTINE
#
# @return
#  previous domain or undef if not changed

sub _set_v4_id_domain_solaris($) {
    my ($id_domain) = @_;
    my ($old);
    my $msg = '';

    # Solaris 5.10 or later
    my $cmd = "$sudo cat /var/run/nfs4_domain";
    my $out = qx($cmd);
    # Remove trailing newline
    chomp $out;
    if ($out ne $id_domain) {
        # Save the value to be returned
        $old = $out;

        my $file = '/etc/default/nfs';
        my $key = 'NFSMAPID_DOMAIN';

        # Read the /etc/default/nfs file, modify the NFSMAPID_DOMAIN line
        # and rewrite the file
        my $contents;
        eval { $contents = read_file(file => $file, user => 'root') };
        croak "_set_v4_id_domain_solaris: $@" if $@;
        my $contents_orig = $contents;
        $contents =~ s/$key=$old/$key=$id_domain/;

        eval { write_file(file     => $file,
                          contents => $contents,
                          user     => 'root') };
        croak "_set_v4_id_domain_solaris: $@" if $@;

        # Restart the /network/nfs/mapid service
        # We don't try to catch errors here. We'll just check whether
        # the domain was set or not.
        qx($sudo /usr/sbin/svcadm restart svc:/network/nfs/mapid);

        # Verify the domain was set
        my $start_time = time;
        while (time - $start_time < 30) {
            $out = qx($cmd);
            chomp $out;
            return ($old) if $out eq $id_domain;
            sleep 5;
        }
        # The domain was not set correctly, write the original
        # contents back.
        eval { write_file(file     => $file,
                          contents => $contents,
                          user     => 'root') };
        # Writing the original contents back failed.
        # Indicate this in the message.
        $msg = " Contents of $file for $key may be corrupt." if $@;

        # The domain was not set correctly, croak with an error.
        croak "_set_v4_id_domain_solaris: Could not set " .
          "NFS V4 ID domain to $id_domain. $msg";
    }
    return ($old);
}

# @name ::_set_v4_id_domain_aix
# @description
#  Set the NFS V4 ID domain on AIX. If the domain is already set to the
#  value passed in then it just returns. Croak if it fails to set it.
#  INTERNAL SUBROUTINE
#
# @return
#  previous domain or undef if not changed
#  Nothing

sub _set_v4_id_domain_aix($) {
    my ($id_domain) = @_;
    my ($old);

    # AIX 5.3 or later
    my $cmd = "$sudo /usr/sbin/chnfsdom";
    my $out = qx($cmd);
    $out =~ /local domain: (\S+)\s*$/;
    if ($1 ne $id_domain) {
        # Save the value to be restored
        $old = $1;
        qx($cmd $id_domain);

        # Verify the domain was set
        $out = qx($cmd);
        $out =~ /local domain: (\S+)\s*$/;
        if ($1 ne $id_domain) {
            croak "_set_v4_id_domain_aix: Could not set " .
                  "NFS V4 ID domain to $id_domain";
        }
    }
    return ($old);
}

# @name ::_set_v4_id_domain_linux
# @description
#  Check the NFS V4 ID domain on Linux. If the domain is not set to the
#  value passed in, then croak.
#  INTERNAL SUBROUTINE
#
# @return
#  previous domain or undef if not changed
#  Nothing

sub _set_v4_id_domain_linux($) {
    my ($id_domain) = @_;
    my ($old);
    my $contents;
    my $file = '/etc/idmapd.conf';

    # Linux 2.6.18 or later

    # Get the contents of /etc/idmapd.conf and determine
    # the current id domain.
    eval { $contents = read_file(file => $file, user => 'root') };
    croak "_set_v4_id_domain_linux: $@" if $@;
    $contents =~ /Domain = (\S+)/;
    if ($1 ne $id_domain) {
        # We cannot reliably set the domain without rebooting the client.
        croak "_set_v4_id_domain_linux: Linux client NFS V4 ID domain is " .
          "$1, needs to be $id_domain. There is no way to reliably change " .
            "the ID domain without a client reboot.";
    }
    return ($old);
}

# @name ::_acl_set_client_nfs_id_domain
# @description
#  Makes sure that the client operating system has the correct NFS
#  V4 ID Domain setting. Croak if the client doesn't. This routine
#  doesn't try to reset the ID domain as that usually requires a reboot
#  which the NATE test libraries do not provide thorough methods for
#  client booting (like they do for filer booting).
#  INTERNAL SUBROUTINE
#
# @return
#  previous domain or undef if not changed

sub _acl_set_client_nfs_id_domain($;) {
    my ($id_domain) = @_;
    my ($h, $old, $type);

    croak "_acl_set_client_nfs_id_domain: NFS V4 ID domain not passed in " .
          "correctly" if (!$id_domain);

    # Get the hosttype
    $h = host('localhost');
    $type = $h->hosttype();

    if ($type =~ /unix-solaris/i) {
        $old = _set_v4_id_domain_solaris($id_domain);
    } elsif($type =~ /unix-linux/i) {
        $old = _set_v4_id_domain_linux($id_domain);
    } elsif ($type =~ /unix-aix/i) {
        $old = _set_v4_id_domain_aix($id_domain);
    } elsif ($type =~ /windows/i) {
        # Never have to set it on windows
        ;
    } else {
        croak "_acl_set_client_nfs_id_domain: $type not supported by " .
              "NFS V4 ID domain client check\n";
    }
    return ($old);
}

## @name ::acl_access_check
## @description
##  Check that the current user/group can access the given file or
##  directory for the specified access type.
##
## @arg %opts	Required: The hash of options that include:
##
##  mask		Required: NFS V4 ACE access type
##
##  filename		Required: Name of file to check access on.
##
## @return
##  0 if access granted, 1 if access denied

sub acl_access_check( % ) {
    my %opts = validate(
        @_,
        {
            mask     => { type => SCALAR },
            filename => { type => SCALAR },
        }
    );
    my (@cont, $cwd, $op, $subdir_dir, $subdir_file);

    eval {
        logDbg("acl_access_check: uid: $> gid: $) file: $opts{filename}");
        if (-f $opts{filename}) {
            if ($opts{mask} == ACE4_READ_DATA) {
                $op = open(IN, '<', $opts{filename}) or croak "$OS_ERROR";
                @cont = <IN>;
                close(IN) or croak "$OS_ERROR";
            } elsif ($opts{mask} == ACE4_WRITE_DATA) {
                $op = open(OUT, '>', $opts{filename}) or croak "$OS_ERROR";
                print OUT "echo acl_access_check: ACE4_WRITE_DATA\n";
                close(OUT) or croak "$OS_ERROR";
            } elsif ($opts{mask} == ACE4_APPEND_DATA) {
                $op = open(OUT, '>>', $opts{filename}) or croak "$OS_ERROR";
                print OUT "echo acl_access_check: ACE4_APPEND_DATA\n";
                close(OUT) or croak "$OS_ERROR";
            } elsif ($opts{mask} == ACE4_EXECUTE) {
                $op = `$opts{filename} && echo _-exec_ok-_`;
                croak "exec failed" if ($op !~ /_-exec_ok-_/);
            } else {
                croak "acl_access_check: Invalid mask value: $opts{mask}";
            }
        } elsif (-d $opts{filename}) {
            if ($opts{mask} == ACE4_LIST_DIRECTORY) {
                opendir(DIR, $opts{filename}) or croak "$OS_ERROR";
                @cont = readdir DIR;
                closedir DIR;
            } elsif ($opts{mask} == ACE4_ADD_FILE) {
                $subdir_file = "$opts{filename}/acl_access_check.$$";
                open TST, "+> $subdir_file" or croak "$OS_ERROR";
                close TST;
                unlink $subdir_file;
            } elsif ($opts{mask} == ACE4_ADD_SUBDIRECTORY) {
                $subdir_dir = "$opts{filename}/acl_access_check.$$";
                mkdir $subdir_dir or croak "$OS_ERROR";
                rmdir $subdir_dir;
            } elsif ($opts{mask} == ACE4_EXECUTE) {
                $cwd = getcwd();
                chdir $opts{filename} or croak "$OS_ERROR";
                chdir $cwd;
            } else {
                croak "acl_access_check: Invalid mask value: $opts{mask}";
            }
        } else {
            croak "acl_access_check: couldn't determine file type";
        }
    };

    if ($@) {
        logDbg("acl_access_check: Error: $@");
        return (1);
    } else {
        return (0);
    }
}

## @name ::acl_get_v4_id_domain
## @description
##  Get the NFS V4 ID domain from the filer.
##
## @arg %opts	Required: The hash of options that include:
##
##  fc			Required: Connection to filer
##
## @return
##  V4 ID domain string

sub acl_get_v4_id_domain( % ) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    return (filer_options_get($opts{fc}, 'nfs.v4.id.domain'));
}

## @name ::acl_enable_v4_acl
## @description
##  Enable NFS V4 ACLs on the filer.
##
## @arg %opts	Required: The hash of options that include:
##
##  fc			Required: Connection to filer
##
## @return
##  nothing

sub acl_enable_v4_acl( % ) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    filer_options_set($opts{fc}, 'nfs.v4.acl.enable', 'on');
    return undef;
}

## @name ::acl_disable_v4_acl
## @description
##  Disable NFS V4 ACLs on the filer.
##
## @arg %opts	Required: The hash of options that include:
##
##  fc			Required: Connection to filer
##
## @return
##  nothing

sub acl_disable_v4_acl( % ) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    filer_options_set($opts{fc}, 'nfs.v4.acl.enable', 'off');
    return;
}

## @name ::flip_mode_bits
## @description
##  Flip the mode bits for the specified mode and type. If the current
##  bits for the type are 7 then they will be set to 0.
##
## @arg %opts	Required: The hash of options that include:
##
##  mode	Required: The current mode of the file. It can be an
##		octal number or string, e.g., 0755, '755'
##
##  type	Required: The permission bits to flip. It is either
##		owner, group, or everyone.
##
## @return
##  mode

sub flip_mode_bits( % ) {
    my %opts = validate(
        @_,
        {
            mode => { type => SCALAR },
            type => { type => SCALAR, regex => qr/^(owner|group|everyone)$/ },
        }
    );
    my ($perms, $shift);
    my $mode = $opts{mode};

    # Convert mode into an octal number
    if ((~$mode & $mode) ne '0') {
        # Mode was passed as an octal string, convert it to a number
        $mode = oct($mode);
    }

    # Figure out which bits to flip based on the type specified
    if ($opts{type} eq 'owner') {
        $shift = 6;
    } elsif ($opts{type} eq 'group') {
        $shift = 3;
    } else {
        $shift = 0;
    }

    $perms = ($mode >> $shift) & 07;
    $perms = (~$perms & 07) << $shift;

    # Clear out the old permissions and or in the new
    $mode &= ~(07 << $shift);
    $mode |= $perms;

    return($mode);
}

# @name ::_set_user_group_unix
# @description
#  Set the user and group id. The underlying assumption is that the
#  test was run as root to allow this to happen. If the current uid
#  is not root then first change to root before changing to the new
#  uid. This call will clear out all the groups in the group list and
#  just set it to the passed in group.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: The hash of options that include:
#
#  user_entry	Required: A user entry hash containing user and group info.
#
# @return
#  nothing

sub _set_user_group_unix( % ) {
    my %opts = @_;

    # Check to make sure we are currently root. If not switch to root,
    # if that fails then croak.
    if ($> != 0) {
        $> = 0;
        $) = "0 0";
        croak "_set_user_group_unix: couldn't change to root" if ($> != 0);
    }

    # Switch the gid first while still root then switch the uid
    $) = "$opts{user_entry}->{gid} $opts{user_entry}->{gid}";
    $> = $opts{user_entry}->{uid};

    # Verify that the changes were made
    if ($> != $opts{user_entry}->{uid}) {
        croak "_set_user_group_unix: uid: $> exp uid: $opts{user_entry}->{uid}"
    }
    if ("$)" ne "$opts{user_entry}->{gid} $opts{user_entry}->{gid}") {
        croak "_set_user_group_unix: gid: $) exp gid: $opts{user_entry}->{gid}"
    }
    return;
}

# @name ::_set_user_group_windows
# @description
#  Set the user name. This is done by unsharing the current drive
#  and resharing it as the new user. The users password should be
#  $CIFS_PASSWD.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: The hash of options that include:
#
#  mountpoint	Required: The mountpoint where files are going to be
#			accessed by the caller. This is needed when
#			mapping shares on Windows.
#
#  user_entry	Required: A user entry hash containing user and group info.
#
# @return
#  nothing

sub _set_user_group_windows( % ) {
    my %opts = @_;
    my ($cifs_drive, $host, $out, @tmp, $u, @user);

    # Get the server name from the current mapping
    $out = `net use $opts{mountpoint}`;
    ($host) = $out =~ /Remote name\s+\\\\(\S+)\\/;

    # unmap the drive
    unmap_cifs_share(mappath => $opts{mountpoint});

    # Reset global cleanup variable in case it dies during mapping
    # so that cleanup will not try to unmap the drive.
    $cifs_map_completed = 0;

    # If the CIFS_DOMAIN is set and the user is not CIFS_USER then prepend
    # the domain to the user name
    if (($main::CIFS_DOMAIN ne '') &&
        ($main::CIFS_USER ne $opts{user_entry}->{user_name})) {

        $user[0] = "$main::CIFS_DOMAIN\\$opts{user_entry}->{user_name}";
        # If the domain consists of subdomains then just prepend
        # the first subdomain name
        @tmp = split('\.', $main::CIFS_DOMAIN);
        if (@tmp > 1) {
            $user[1] = "$tmp[0]\\$opts{user_entry}->{user_name}";
        }
    } else {
        $user[0] = "$opts{user_entry}->{user_name}";
    }

    # If more than one user was setup to try out then if the mapping
    # fails, try out the next user.
    foreach $u (@user) {
        # Now map it back at the same drive letter
        $cifs_drive = map_cifs_share(
                    server   => $host,
                    share    => $main::CIFS_SHARE,
                    client   => 'localhost',
                    drive    => $opts{mountpoint},
                    username => $u,
                    password => $opts{user_entry}->{cifspass}
                    );
        last if (!$@);
    }
    croak($@) if ($@);

    logDbg("_set_user_group_windows: mapped drive $cifs_drive");
    if (!$cifs_drive) {
        croak("_set_user_group_windows: could not map the drive");
    }
    $cifs_map_completed = 1;

    $cifs_drive .= ":" unless $cifs_drive =~ /:$/;
    return;
}

## @name ::set_user_group
## @description
##  Set the effective user and group. Determines if the localhost is Windows
##  or UNIX and calls the appropriate routine.
##
## @arg %opts	Required: The hash of options that include:
##
##  mountpoint	Required: The mountpoint where files are going to be
##			accessed by the caller. This is needed when
##			mapping shares on Windows.
##
##  user_entry	Required: A user entry hash containing user and group info.
##
##  user_reset	Optional: If set to 1 then reset the saved username
##			before run_local() calls _acl_clean_env(). Default
##			is 1.
##
## @return
##  nothing

sub set_user_group( % ) {
    my %opts = validate(
        @_,
        {
            mountpoint => { type => SCALAR },
            user_entry => { type => HASHREF },
            user_reset => { type => SCALAR, default => 1 },
        }
    );

    my $h = host('localhost');
    if ($h->hosttype() =~ /windows/i) {
        _set_user_group_windows(%opts);
    } else {
        _set_user_group_unix(%opts);
    }
    if ($opts{user_reset} == 1) {
        # Set a flag to let the cleanup know that the user was changed
        $uid_gid_changed = 1;
    }
    return;
}

# @name ::_get_file_owner_group_unix
# @description
#  Return the Unix file owner and group for the specified file.
#  INTERNAL SUBROUTINE
#
# @arg %opts		Required: The hash of options that include:
#
#  filename		Required: File name to get owner and group for.
#
# @return
#  ($owner, $group) on success, croak on failure to get info

sub _get_file_owner_group_unix( % ) {
    my %opts = @_;
    my ($group, $owner, $sb);

    $sb = stat($opts{filename}) or
        croak "_get_file_owner_group_unix: stat failed: $opts{filename}: $!";
    ($owner) = getpwuid($sb->uid);
    ($group) = getgrgid($sb->gid);
    if (!defined($owner) || !defined($group)) {
        croak "_get_file_owner_group_unix: undef owner: $owner group: $group";
    }

    return ($owner, $group);
}

# @name ::_get_file_owner_group_windows
# @description
#  Return the Windows file owner and group for the specified file.
#  Uses setACL.exe to get the file owner and group.
#  INTERNAL SUBROUTINE
#
# @arg %opts		Required: The hash of options that include:
#
#  filename		Required: File name to get owner and group for.
#
# @return
#  ($owner, $group) on success, croak on failure to get info

sub _get_file_owner_group_windows( % ) {
    my %opts = @_;

    my (@fields, $filename, $group, @lines, $out, $owner);

    # Change pathname to windows format
    $filename = $opts{filename};
    $filename =~ s/\//\\/g;

    # Use setACL to get the file owner and group
    $out = `x:\\bin\\setacl.exe -on $filename -ot file -actn list -lst \"f:csv;w:o,g\"`;
    if ($out !~ /finished successfully/i) {
        croak "get_file_owner_windows: could not get owner\n$out";
    }

    # There are multiple output lines but we only care about the first
    @lines = split(/^/, $out);

    # Remove the trailing newline on the first line
    chomp($lines[0]);

    # First parse for different fields items:
    #   "objectName",objectType(1 for file),"ACEs";Owner:SD;Group:SD
    @fields = $lines[0] =~ /"(.+)",(\d),"(.+)"/;

    # Seperate owner and group from the ACEs
    @lines = split(';', $fields[2]);

    # When getting the owner this is the local machines view of who
    # owns the file. If the SID is not found in active directory, or
    # the local machine is not running active directory then it gets
    # mapped to the Builtin\Administrators which is a group. Since
    # a way is not known to get the credentials used to login when
    # mapping the share we don't know how to reliably determine if
    # an SID should map to OWNER@ or GROUP@.

    # Get the owner
    ($owner) = $lines[1] =~ /Owner:(\S+)/;

    # Get the group
    ($group) = $lines[2] =~ /Group:(\S+)/;

    if (!defined($owner) || !defined($group)) {
        croak "_get_file_owner_group_windows: undef owner: $owner group: $group";
    }
    return ($owner, $group);
}

## @name ::get_file_owner_group
## @description
##  Return the file owner and group for the specified file.
##
## @arg %opts		Required: The hash of options that include:
##
##  filename		Required: File name to get owner and group for.
##
## @return
##  ($owner, $group) on success, croak on failure to get info

sub get_file_owner_group( % ) {
    my %opts = validate(
        @_,
        {
            filename => { type => SCALAR },
        }
    );

    my $h = host('localhost');
    if ($h->hosttype() =~ /windows/i) {
        return (_get_file_owner_group_windows(%opts));
    } else {
        return (_get_file_owner_group_unix(%opts));
    }
}

# @name ::_validate_file_owner_unix
# @description
#  Validate that the file is owned by the user and group name
#  passed in.
#  INTERNAL SUBROUTINE
#
# @arg %opts		Required: The hash of options that include:
#
#  filename		Required: File name to validate owner of.
#
#  user_entry		Required: A user entry hash with user and group info.
#
#  validate_group	Optional: Define to validate group also.
#
# @return
#  0 if the validation passed 1 otherwise

sub _validate_file_owner_unix( % ) {
    my %opts = @_;
    my ($sb);

    $sb = stat($opts{filename}) or
        croak "validate_file_owner: stat failed: $opts{filename}: $!";
    if ($sb->uid != $opts{user_entry}->{uid} ||
        (defined($opts{validate_group})
         && $sb->gid != $opts{user_entry}->{gid})) {
        logDbg("_validate_file_owner_unix: file: $opts{filename} exp uid: " .
               "$opts{user_entry}->{uid} uid: " . $sb->uid . " exp gid: " .
               "$opts{user_entry}->{gid} gid: " . $sb->gid);
        return (1);
    } else {
        return (0);
    }
}

# @name ::_validate_file_owner_windows
# @description
#  Validate that the file is owned by the user name passed in.
#  INTERNAL SUBROUTINE
#
# @arg %opts		Required: The hash of options that include:
#
#  filename		Required: File name to validate owner of.
#
#  user_entry		Required: A user entry hash with user and group info.
#
#  validate_group	Optional: Define to validate group also.
#
# @return
#  0 if the validation passed 1 otherwise

sub _validate_file_owner_windows( % ) {
    my %opts = @_;
    my (@fields, $filename, $group, @lines, $out, $owner);

    ($owner, $group) = get_file_owner_windows(filename => $opts{filename});

    if ($opts{user_entry}->{user_name} ne $owner ||
        (defined($opts{validate_group}) &&
         $opts{user_entry}->{group_name} ne $group)) {
        logDbg("_validate_file_owner_windows: file: $opts{filename} exp user:" .
               "$opts{user_entry}->{user_name} user: $owner exp group: " .
               "$opts{user_entry}->{group_name} group: $group");
        return (1);
    } else {
        return (0);
    }
}

## @name ::validate_file_owner
## @description
##  Validate that the file is owned by the user and group name
##  passed in.
##
## @arg %opts		Required: The hash of options that include:
##
##  filename		Required: File name to validate owner of.
##
##  not_windows		Optional: Define to indicate no check should be
##			done on windows just return 0.
##
##  user_entry		Required: A user entry hash with user and group info.
##
##  validate_group	Optional: Define to validate group also.
##
## @return
##  0 if the validation passed 1 otherwise

sub validate_file_owner( % ) {
    my %opts = validate(
        @_,
        {
            filename       => { type => SCALAR },
            not_windows    => { type => SCALAR, optional => 1 },
            user_entry     => { type => HASHREF },
            validate_group => { type => SCALAR, optional => 1 },
        }
    );
    my ($status);

    my $h = host('localhost');
    if ($h->hosttype() =~ /windows/i) {
        if ($opts{not_windows}) {
            $status = 0;
        } else {
            $status = _validate_file_owner_windows(%opts);
        }
        return ($status);
    } else {
        return (_validate_file_owner_unix(%opts));
    }
}

# @name ::_validate_users
# @description
#  Validate that the number of test user accounts exist on the
#  local machine.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: The hash of options that include:
#
#  num_users	Required: Number of test user accounts needed.
#
# @return
#  0 if the validation passed 1 otherwise

sub _validate_users( % ) {
    my %opts = validate(
        @_,
        {
            num_users => { type => SCALAR },
        }
    );
    my ($gid, $i, $uid);

    $i = @users;
    if ($opts{num_users} > $i) {
        croak "_validate_users: Too many users requests: req: " .
              "$opts{num_users} avail: $i";
    }

    my $h = host('localhost');
    if ($h->hosttype() =~ /windows/i) {
        # For windows there is nothing to do since the share mapping is
        # verified through the filer and _acl_set_env() already verified
        # accounts on the filer.
        ;
    } else {
        for ($i = 0; $i < $opts{num_users}; $i++) {
            (undef, undef, $uid, $gid) = getpwnam($users[$i]->{user_name});
            if (!defined($uid)) {
                logDbg("_validate_users_unix: user $users[$i]->{user_name} " .
                       "must be defined");
                return (1);
            }
            $users[$i]->{uid} = $uid;
            $users[$i]->{gid} = $gid;
        }
    }
    return (0);
}

## @name ::get_users
## @description
##  Return the user array of hashs.
##
## @arg
##  none
##
## @return
##  The user array of hashs

sub get_users {
    return (@users);
}

## @name ::get_ace_array
## @description
##  Return an array of ACEs used
##
## @arg %opts	Required: The hash of options that include:
##
##  type	Required: The type of ACEs to return. The valid types are:
##			file_aces - standard file init ACEs
##			dir_aces - standard directory init ACEs
##			dir_inherit_aces - Directory Inherit ACEs
##			file_inherit_aces - Directory File Inherit ACEs
##
## @return
##  The array of ACEs

sub get_ace_array {
    my %opts = validate(
        @_,
        {
            type => { type => SCALAR,
                     regex =>
                  qr/^(file_aces|dir_aces|dir_inherit_aces|file_inherit_aces)$/
                    },
        }
    );
    my $tmp;

    # Since the ace arrays contain hash references, all copies of the
    # same array point to the same hash refs. Changing a hash ref in
    # one array will change all copies of that array. To avoid this,
    # we deep clone the array first and return the clone.
    if ($opts{type} eq 'file_aces') {
        $tmp = dclone(\@file_aces);
        return (@$tmp);
    } elsif ($opts{type} eq 'dir_aces') {
        $tmp = dclone(\@dir_aces);
        return (@$tmp);
    } elsif ($opts{type} eq 'dir_inherit_aces') {
        $tmp = dclone(\@dir_inherit_aces);
        return (@$tmp);
    } else {
        $tmp = dclone(\@file_inherit_aces);
        return (@$tmp);
    }
}

## @name ::wait_for_file
## @description
##  Wait for the specified operation to occur on a file. This subroutine
##  will poll to see if the operation occurred and if not then sleep for
##  a second. It will keep checking until the operation occurs or
##  the timeout value has been exceeded. The assumption here is that
##  another thread or process on this host or another host will make
##  the operation happen on the file.
##
## @arg %opts	Required: The hash of options that include:
##
##  filename	Required: The path of the filename to wait for. Can be
##		path of any UNIX filetype.
##
##  op		Required: The file operation to wait for. Valid values
##		are 'create' and 'remove'. If set to create then it waits
##		for the file to appear. If set to remove then it waits
##		for the file to be removed.
##
##  timeout	Optional: The number of seconds to wait for the file
##		operation to take place. The default value is 120 secs.
##
## @return
##  0 if the file operation occurred or 1 if it timedout

sub wait_for_file( % ) {
    my %opts = validate(
        @_,
        {
            filename => { type => SCALAR },
            op       => { type => SCALAR, regex => qr/^(create|remove)$/ },
            timeout  => { type => SCALAR, default => 120 },
        }
    );
    my $time_left = $opts{timeout};

    while ($time_left > 0) {
        if (-e $opts{filename}) {
            if ($opts{op} eq 'create') {
                return (0);
            }
        } else {
            if ($opts{op} eq 'remove') {
                return (0);
            }
        }
        $time_left--;
        sleep 1;
    }

    return(1);
}

## @name ::acl_clear_inheritance
## @description
##  Removes all file and directory inheritance flags from the ACL
##  of an object. Retrieves the current ACL from the object, then
##  scans the ACEs on the ACL, removing all ACE4_FILE_INHERIT_ACE
##  and ACE4_DIRECTORY_INHERIT_ACE flags. The ACE itself is not
##  removed or cleared because the ACE may still apply to the object
##  itself.
##
## @arg %opts	Required: The hash of options that include:
##
##  fc			Required: Connection to filer
##
##  filename		Required: Name of current file or dir to get the
##			ACL for, including local mountpoint.
##
##  filer_path		Required: Full pathname of volume on filer.
##
##  filer_vol		Required: Volume name on filer.
##
##  mountpoint		Required: Mountpoint where the file is located.
##
## @return
##  nothing

sub acl_clear_inheritance( % ) {
    my %opts = validate(
        @_,
        {
            fc         => { type => OBJECT },
            filename   => { type => SCALAR },
            filer_path => { type => SCALAR },
            filer_vol  => { type => SCALAR },
            mountpoint => { type => SCALAR },
        }
    );

    # Get the ACL from the object
    my $acl = nfs::GenericAcl::get_acl(fc => $opts{fc},
                                       filename => $opts{filename},
                                       filer_path => $opts{filer_path},
                                       filer_vol => $opts{filer_vol},
                                       mountpoint => $opts{mountpoint});

    # Remove ACE4_FILE_INHERIT_ACE and ACE4_DIRECTORY_INHERIT_ACE
    # from all flags
    foreach (@{$acl->{ace}}) {
        if ($_->{flags} & ACE4_FILE_INHERIT_ACE) {
            $_->{flags} ^= ACE4_FILE_INHERIT_ACE;
        }
        if ($_->{flags} & ACE4_DIRECTORY_INHERIT_ACE) {
            $_->{flags} ^= ACE4_DIRECTORY_INHERIT_ACE;
        }
    }

    # Set the modified ACL on the object
    my $status = nfs::GenericAcl::set_acl(acl => $acl);
    if ($status ne 'NFS4_OK') {
        nfs::GenericAcl::print_acl(acl => $acl);
        croak("acl_clear_inheritance for $opts{filename} failed: $status");
    }

    return;
}

## @name ::_acl_cifs_setup
## @description
## Wrapper function that calls cifs_setup to setup CIFS on the
## filer with the specified mode. This wrapper is used to isolate the details
## of setting up in a domain environment or on WORKGROUP.
##
## Will setup cifs in a domain environment or under WORKGROUP. If domain
## is an empty string will setup under WORKGROUP otherwise will attempt to
## setup under a domain using passed in or default values. Croaks if their
## is an error in call to cifs_setup.
##
## @arg %opts	Required: The hash of options that include:
##
##  filer		Required: Hostspec of filer to connect to.
##
##  mode		Optional: Protocol mode to setup cifs. Either 'Multi'
##			protocol or 'Ntfs' only. Defaults to 'Multi' protocol.
##
##  ad_setup		Optional: If defined and set to 'yes' will add filer
##			to an active directory domain.
##
##  domain		Optional: Domain name to add filer to, used in
##			cifs_setup. Allows a domain name with just spaces for
##			testing purposes. Defaults to $main::CIFS_DOMAIN.
##
##  domain_user		Optional: Domain user that administers filer in
##			domain,used in cifs_setup.
##			Defaults to $main::CIFS_DOMAIN_ADMIN.
##
##  domain_pass		Optional: Domain user password,used in cifs_setup.
##			Defaults to $main::CIFS_DOMAIN_ADMIN_PASSWD.
##
## @return
##  nothing

sub _acl_cifs_setup(%) {
    my %opts = validate(
        @_,
        {
            filer       => { type => SCALAR },
            mode        => { type => SCALAR, default => 'multi',
                             regex => qr/^(ntfs|multi)$/ },
            domain      => { type => SCALAR,
                             default => $main::CIFS_DOMAIN },
            domain_user => { type => SCALAR,
                             default => $main::CIFS_DOMAIN_ADMIN },
            domain_pass => { type => SCALAR,
                             default => $main::CIFS_DOMAIN_ADMIN_PASSWD },
        }
    );

    my (%args, $fc, $fh, $err_msg, @proto);

    # Determine whether to use the domain or the workgroup
    $opts{domain} =~ s/\s+//g;
    if ($opts{domain} ne '') {
        # Special hook to allow passing in an empty domain name
        if ($opts{domain} eq 'XX_empty_XX') {
            $opts{domain} = $opts{dns_name} = '    ';
        }

        %args = (domain      => $opts{domain},
                 domain_user => $opts{domain_user},
                 domain_pass => $opts{domain_pass},
                 'auth-type' => 'ad',);
    } else {
        %args = (workgroup   => 'WORKGROUP',
                 'auth-type' => 'workgroup',);
    }

    # If the protocol requested is ntfs then cifs_setup has to be run
    # twice, the first time to multi then the second time to ntfs. This
    # gurantees that the filer is in NTFS-only mode.
    if ($opts{mode} eq 'ntfs') {
        @proto = ( 'multiprotocol', 'ntfs' );
    } elsif ($opts{mode} eq 'multi') {
        @proto = ( 'multiprotocol' );
    } else {
        @proto = ( $opts{mode} );
    }

    $fh = host($opts{filer});
    foreach my $p (@proto) {
        cifs_setup(cifs_pass        => $main::CIFS_PASSWD,
                   host             => $fh,
                   'server-name'    => $fh->hostname,
                   passwd_group     => 'yes',
                   'security-style' => $p,
                   %args);
        if ($@) {
            # save the cifs_setup error
            $err_msg = $@;

            # cifs_setup should cleanup after itself but it can leave
            # the command 'cifs setup' running and prompting for input.
            # When 'cifs setup' gets converted to ZAPI this problem should
            # go away. For now grab the telnet connection to send a ^C to
            # the filer.
            eval {
                host($opts{filer});
                $fc = connect("#telnet", -timeout, $main::TIMEOUT);
                $fc->timeout($main::CMDTIMEOUT);
                $fc->send_intr();
                $fc->disconnect();
            };
            croak("_acl_cifs_setup: cifs setup failed $err_msg");
        }
    }

    return;
}
1;
