# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/AclBase.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::ACL::AclBase;

## @summary
##  Base class to the OS dependent ACL classes
##
## @description
##  Provides a default behavior for the subroutines that operate
##  on ACLs. These subroutines should only be called directly or
##  indirectly from the nfs::GenericAcl.pm library. The base class uses
##  the synthetic NFS V4 client for all operations. This should be
##  the parent class for any OS-dependent class.
##
##  The principle routines for Acl operations are defined in the
##  nfs::ACL::V4tClient and nfs::ACL::PerlSyntheticClient libraries.
##  The base class for AclBase is defined by one of these two
##  libraries, which is determined by the value of the global
##  environment variable SYNTH_CLIENT_TYPE.
##  See documentation in nfs::ACL::AclClass for details.
##
##  Windows clients cannot use either nfs::ACL::V4tClient or
##  nfs::ACL::PerlSyntheticClient. nfs::ACL::Windows will always
##  be used for Windows clients, regardless of SYNTH_CLIENT_TYPE.
##
## @usage
##  use nfs::ACL::AclBase;
##
## @status	review
##
## @author      dl-nfsqa-only (NFS QA Sunnyvale)
## @notes       Created by EP Network Storage Performance Lab, Inc.

use strict;

use nfs::ACL::V4tClient;
use nfs::ACL::PerlSyntheticClient;

use vars qw( @ISA );

if ($ENV{SYNTH_CLIENT_TYPE} =~ /v4t/i) {
    # Use v4t client programs
    @ISA = qw(nfs::ACL::V4tClient);
} else {
    # Use perl synthetic client libraries (nfs::NfsClient)
    @ISA = qw(nfs::ACL::PerlSyntheticClient);
}

## @name ::new
##
## @description
##  ACL base class constructor which initializes default behavior.
##  Class behavior is controlled by the blessed hash of the constructor.
##  V4t information is set on the ACL base object, even if
##  v4t is not being used, to keep the class definition consistent
##  regardless of client type (v4t or perl synthetic).
##
## @synopsis
##  $acl_obj = nfs::ACL::AclBase::->new( %override_defaults );
##
## @arg
##  options		Optional: Any named parameter found in a ACL
##			can be specified here as the default value.
##
##  get_acl_cmd		Optional: Full path of get ACL command.
##
##  set_acl_cmd		Optional: Full path of set ACL command.
##
##  sudo		Optional: Full path of sudo command.
##
##  v4t_path		Optional: Path to the v4t commands
##
##  v4t_prefix		Optional: Prefix to place before the v4t
##			command. Currently used to set LDR_CNTRL
##			environment variable on AIX.
##
## @return
##  ACL base object.
##
## @example
##  $obj = nfs::ACL::AclBase::->new( set_acl_cmd => '/bin/setfacl' );

sub new( $% ) {
    my $class = shift;

    bless {
        # Defaults
        get_acl_cmd => '/usr/bin/getfacl',
        set_acl_cmd => '/usr/bin/setfacl',
        sudo        => '/usr/software/bin/sudo',
        v4t_path    => '/usr/local/test/bin',
        v4t_prefix  => '',

        # Override defaults
        @_
    }, $class;
}

1;
