# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/AclClass.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
package nfs::ACL::AclClass;

## @summary
##   Provide an OS independent NFS V4 Access Control List (ACL) class.
##
## @description
##   Provides an interface for NFS V4 ACL access. The class
##   provides a generic interface for accessing the ACL and hides
##   the details from the test programs.
##
##   User can specify SYNTH_CLIENT_TYPE=(perl_only|perl|v4t_only|v4t).
##   'perl' and 'perl_only' will use the perl synthetic client library
##   nfs::NfsClient. 'v4t' and 'v4t_only' will use the v4t_* client
##   programs. 'perl_only' or 'v4t_only' will strictly use the synthetic
##   client for all operations. Otherwise, system specific functionality
##   will be used when available and synthetic client functionality
##   will be used elsewhere. Not specifying SYNTH_CLIENT_TYPE is
##   equivalent to 'perl'. Windows clients are not affected by the
##   SYNTH_CLIENT_TYPE setting, and will always use Windows native
##   routines.
##
## @usage
##   use nfs::ACL::AclClass;
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;

use Tharn;
use Carp qw(croak);
use nfs::ACL::Aix;
use nfs::ACL::Hpux;
use nfs::ACL::Linux;
use nfs::ACL::Solaris;
use nfs::ACL::Windows;
use vars qw( @ISA );

if (defined $ENV{SYNTH_CLIENT_TYPE} &&
    $ENV{SYNTH_CLIENT_TYPE} !~ /^(perl|perl_only|v4t|v4t_only)$/i) {
    croak "'$ENV{SYNTH_CLIENT_TYPE}' is an invalid value for SYNTH_CLIENT_TYPE";
}

## @name ::new
## @description
##  Creates a new ACL Object. This routine will create a new
##  object by calling the platform specific 'new' constructor
##  (in the platform specific ACL package (i.e. Solaris.pm)). This way
##  the user uses a single API to manipulate ACLs and the code underneath
##  creates the appropriate object based on the platform type.
##
## @synopsis
## 	$obj = nfs::ACL::AclClass->new(%opts);
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
##  opts	Required: An options hash this is passed through to
##		the class constructor. The caller should have
##		already validated the options.
##
## @return
##  Platform specific ACL object

sub new ( $% ) {
    my $class = shift;
    my ($h, $type);

    $h = host('localhost');
    $type = $h->hosttype();

    if ($type =~ /windows/i) {
        # Windows will only use nfs::ACL::Windows regardless
        # of the SYNTH_CLIENT_TYPE setting
        @ISA = qw(nfs::ACL::Windows);
    } elsif ($ENV{SYNTH_CLIENT_TYPE} =~ /^(perl_only|v4t_only)$/i) {
        @ISA = qw(nfs::ACL::AclBase);
    } elsif ($type =~ /unix-solaris/i) {
        @ISA = qw(nfs::ACL::Solaris);
    } elsif ($type =~ /unix-hpux/i) {
        @ISA = qw(nfs::ACL::Hpux);
    } elsif ($type =~ /unix-aix/i) {
        @ISA = qw(nfs::ACL::Aix);
    } elsif ($type =~ /unix-linux/i) {
        @ISA = qw(nfs::ACL::Linux);
    } else {
        # If we don't directly support the hosttype then just use the
        # synthetic client.
        @ISA = qw(nfs::ACL::AclBase);
    }
    # Call the constructor for the OS dependent class
    return ( SUPER::new $class @_ );
}

1;
