# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/Aix.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::ACL::Aix;

## @summary  
##   Class to handle OS dependent portion of ACLs on AIX.
##
## @description  
##   This is a sub-class of the AclClass package which provides
##   support for AIX specific portion of ACLs.
##
## @usage
##   use nfs::ACL::Aix
##
## @status	review
##
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
##
## @notes
##    Written for AIX 5.3

use strict;
use Fcntl qw(:mode);
use File::stat;
use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::ACL::AclNFSV4 qw(:ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);
use nfs::ACL::AclBase;
use vars qw( @ISA );

#---------------------------------------------------------------------
# Inherit functionality from AclBase class.
# Override the get_acl and set_acl methods.
#---------------------------------------------------------------------
our @ISA = qw(nfs::ACL::AclBase);

## @name ::new
## @description
##  AIX specific class constructor. It overrides default
##  behavior of the base exports class.
##
## @return
##  AIX specific export object.

sub new ( $ ) {
    my $class = shift;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class (
        get_acl_cmd    => '/usr/bin/aclget',
        set_acl_cmd    => '/usr/bin/aclput',
        v4t_prefix     => 'LDR_CNTRL=\'MAXDATA=0x80000000\'',

        # Override defaults
        @_
    );
}

## @name ::get_acl
## @description
##  Uses the aclget command for getting an ACL. Gets the ACL output
##  from the specified file and then parses it into a generic ACL.
##  If use_synthetic is set in the ACL then call the SUPER class.
##  Updates fields in the ACL object.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##  
## @return
##  Nothing

sub get_acl ( $ ) {
    my ($this) = @_;
    my ($cmd, $out, $sb);

    if ($this->{use_synthetic} == 1) {
        return SUPER::get_acl $this;
    }

    $cmd = "$this->{get_acl_cmd} $this->{filename} && echo _-aclget_ok-_";
    logDbg("Aix::get_acl: $cmd");

    $out = `$cmd`;
    if ($out !~ /_-aclget_ok-_/) {
        croak "get_acl: AIX could not get attributes\n$out";
    }

    aix_to_acl($this, $out);

    # If this file is a directory then set the isdir flag
    if (-d $this->{filename}) {
        $this->{isdir} = 1;
    }
    return;
}

## @name ::set_acl
## @description
##  Set the ACL associated with a given file using the aclput command.
##  If use_synthetic is set in the ACL then call the SUPER class.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return	
##  NFS4_OK on success or NFS V4 error string

sub set_acl ( $ ) {
    my ($this) = @_;
    my ($aclfile, $cmd, $err, $out);

    if ($this->{use_synthetic} == 1) {
        return SUPER::set_acl $this;
    }

    # Convert from ACL class format to AIX format
    $aclfile = "/tmp/set_acl.$$";
    acl_to_aix($this, $aclfile);

    # Log the contents of the file used to set the ACL
    my $t = `cat $aclfile`;
    logDbg("Aix::set_acl: file: $aclfile\n$t\n");

    # Set the ACL
    $cmd = "$this->{set_acl_cmd} -i $aclfile $this->{filename}";
    logDbg("Aix::set_acl: $cmd");

    $out = `$cmd`;
    $err = $?;

    unlink $aclfile;

    if ($err == 0) {
        return ('NFS4_OK');
    } else {
        logDbg("Aix::set_acl: $out");
        return ('NFS4ERR_XXX');
    }
}

# name ::aix_to_acl
# description
#  Creates an ACL object using the output from the aclget command.
#  Could have gone table driven to keep from having to specify a
#  token twice (i.e. like Allow) here and in acl_to_aix but this
#  approach is more obvious. Any changes here most likely need to
#  also be made in acl_to_aix.
#
# arg
#  this	Required: The name of the class. Automatically passed
#		in by PERL. Updates fields in the acl object
#  
#  out		Required: Output from aclget command.
#
# return
#  Nothing

sub aix_to_acl( $$ ) {
    my ($this, $out) = @_;
    my ($flags, $flags_str, $id_type, $id_name, $id_who, $line, @lines,
        $mask, $mask_str, @tmp, $type, $type_str);
    my $ace_cnt = 0;

    $this->{ace_cnt} = 0;
    # Parse the output from the aclget command and put into the acl
    @lines = split(/^/, $out);
    foreach $line (@lines) {
        chomp($line);
        $line =~ s/\r$//;
        # Skip lines that start with *
        next if ($line =~ /^\*/);
        
        # Parse the line
        ($id_type, $id_name, $id_who, $type_str, $mask_str, $flags_str) =
            $line =~ /^\s*(u|g|s):\s*(\S*)\s*\(\s*(\S+)\s*\):\s+(\S)\s+(\S+)\s*(\S*)/;

        # If there is no ID type then skip the line
        next if (!defined($id_type));

        # Set the type from the type_str
        if ($type_str eq 'a') {
            $type = ACE4_ACCESS_ALLOWED_ACE_TYPE;
        } elsif ($type_str eq 'd') {
            $type = ACE4_ACCESS_DENIED_ACE_TYPE;
        } elsif ($type_str eq 'u') {
            $type = ACE4_SYSTEM_AUDIT_ACE_TYPE;
        } elsif ($type_str eq 'l') {
            $type = ACE4_SYSTEM_ALARM_ACE_TYPE;
        } else {
            croak "aix_to_acl: Invalid ACE Type: $type_str";
        }
        $this->{ace}[$ace_cnt]{type} = $type;

        # Set the flags from the flags_str
        $flags = 0;
        if ($flags_str =~ /fi/) {
            $flags |= ACE4_FILE_INHERIT_ACE;
        }
        if ($flags_str =~ /di/) {
            $flags |= ACE4_DIRECTORY_INHERIT_ACE;
        }
        if ($flags_str =~ /ni/) {
            $flags |= ACE4_NO_PROPAGATE_INHERIT_ACE;
        }
        if ($flags_str =~ /oi/) {
            $flags |= ACE4_INHERIT_ONLY_ACE;
        }
        if ($flags_str =~ /sf/) {
            $flags |= ACE4_SUCCESSFUL_ACCESS_ACE_FLAG;
        }
        if ($flags_str =~ /ff/) {
            $flags |= ACE4_FAILED_ACCESS_ACE_FLAG;
        }
        if ($id_type eq 'g') {
            $flags |= ACE4_IDENTIFIER_GROUP;
        }
        # XXX The AIX aclget command doesn't seem to be setting the
        # 'g' flag to indicate a group. Set the flag if who is GROUP@
        if ($id_who eq 'GROUP@') {
            $flags |= ACE4_IDENTIFIER_GROUP;
        }
        $this->{ace}[$ace_cnt]{flags} = $flags;

        # Set the mask from the mask_str
        $mask = 0;
        if ($mask_str =~ /r/) {
            # Also takes care of ACE4_LIST_DIRECTORY
            $mask |= ACE4_READ_DATA;
        }
        if ($mask_str =~ /w/) {
            # Also takes care of ACE4_ADD_FILE
            $mask |= ACE4_WRITE_DATA;
        }
        if ($mask_str =~ /p/) {
            # Also takes care of ACE4_ADD_SUBDIRECTORY
            $mask |= ACE4_APPEND_DATA;
        }
        if ($mask_str =~ /R/) {
            $mask |= ACE4_READ_NAMED_ATTRS;
        }
        if ($mask_str =~ /W/) {
            $mask |= ACE4_WRITE_NAMED_ATTRS;
        }
        if ($mask_str =~ /x/) {
            $mask |= ACE4_EXECUTE;
        }
        if ($mask_str =~ /D/) {
            $mask |= ACE4_DELETE_CHILD;
        }
        if ($mask_str =~ /a/) {
            $mask |= ACE4_READ_ATTRIBUTES;
        }
        if ($mask_str =~ /A/) {
            $mask |= ACE4_WRITE_ATTRIBUTES;
        }
        if ($mask_str =~ /d/) {
            $mask |= ACE4_DELETE;
        }
        if ($mask_str =~ /c/) {
            $mask |= ACE4_READ_ACL;
        }
        if ($mask_str =~ /C/) {
            $mask |= ACE4_WRITE_ACL;
        }
        if ($mask_str =~ /o/) {
            $mask |= ACE4_WRITE_OWNER;
        }
        if ($mask_str =~ /s/) {
            $mask |= ACE4_SYNCHRONIZE;
        }
        $this->{ace}[$ace_cnt]{mask} = $mask;

        # Ignore id_name and just id_who
        $this->{ace}[$ace_cnt]{who} = $id_who;
        $ace_cnt++;
    }
    $this->{ace_cnt} = $ace_cnt;
    return;
}

# name ::acl_to_aix
# description
#  Converts an ACL object into output that can be used by aclput.
#  Could have gone table driven to keep from having to specify a
#  token twice (i.e. like Allow) here and in aix_to_acl but this
#  approach is more obvious. Any changes here most likely need to
#  also be made in aix_to_acl.
#
# arg
#  this	Required: The name of the class. Automatically passed
#		in by PERL.
#  
#  aclfile	Required: Name of file to write ACL to.
#
# return
#  Nothing

sub acl_to_aix( $$ ) {
    my ($this, $aclfile) = @_;
    my ($flags, $i, $line, @lines, $mask, $outstr, $sb, $str, @tmp, $type,
        $group, $owner, $who);

    # Open up the file that holds the ACL
    open ACL, "> $aclfile" or croak "acl_to_aix: Cannot create $aclfile";

    # Stat the file so we can get owner and group
    $sb = stat($this->{filename}) or
        croak "acl_to_aix: stat failed: $this->{filename}: $!";

    # Create the header for the file
    print ACL "*\n* ACL_type   NFS4\n*\n*\n";
    ($owner) = getpwuid($sb->uid);
    ($group) = getgrgid($sb->gid);
    print ACL "* Owner: $owner\n";
    print ACL "* Group: $group\n*\n";

    # Step through each ACE and print it to the file
    for ($i = 0; $i < $this->{ace_cnt}; $i++) {
        $type = $this->{ace}[$i]{type};
        $flags = $this->{ace}[$i]{flags};
        $mask = $this->{ace}[$i]{mask};
        $who = $this->{ace}[$i]{who};

        # If the mask is 0 then skip this ACE
        if ($mask == 0) {
            next;
        }

        # Print out the type of identity 
        if ($who eq 'OWNER@' || $who eq 'GROUP@' || $who eq 'EVERYONE@') {
            $outstr = 's';
        } elsif ($flags & ACE4_IDENTIFIER_GROUP) {
            $outstr = 'g';
        } else {
            $outstr = 'u';
        }
        $outstr .= ":($who): ";

        # Process ACE Type
        if ($type == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
            $outstr .= 'a ';
        } elsif ($type == ACE4_ACCESS_DENIED_ACE_TYPE) {
            $outstr .= 'd ';
        } elsif ($type == ACE4_SYSTEM_AUDIT_ACE_TYPE) {
            $outstr .= 'u ';
        } elsif ($type == ACE4_SYSTEM_ALARM_ACE_TYPE) {
            $outstr .= 'l ';
        } else {
            close ACL;
            unlink $aclfile;
            croak "acl_to_aix: Bad ACL Type: $type";
        }
        
        # Process ACE Mask
        $str = '';
        if ($mask & ACE4_READ_DATA) {
            # Also takes care of ACE4_LIST_DIRECTORY
            $str .= 'r';
        }
        if ($mask & ACE4_WRITE_DATA) {
            # Also takes care of ACE4_ADD_FILE
            $str .= 'w';
        }
        if ($mask & ACE4_APPEND_DATA) {
            # Also takes care of ACE4_ADD_SUBDIRECTORY
            $str .= 'p';
        }
        if ($mask & ACE4_READ_NAMED_ATTRS) {
            $str .= 'R';
        }
        if ($mask & ACE4_WRITE_NAMED_ATTRS) {
            $str .= 'W';
        }
        if ($mask & ACE4_EXECUTE) {
            $str .= 'x';
        }
        if ($mask & ACE4_DELETE_CHILD) {
            $str .= 'D';
        }
        if ($mask & ACE4_READ_ATTRIBUTES) {
            $str .= 'a';
        }
        if ($mask & ACE4_WRITE_ATTRIBUTES) {
            $str .= 'A';
        }
        if ($mask & ACE4_DELETE) {
            $str .= 'd';
        }
        if ($mask & ACE4_READ_ACL) {
            $str .= 'c';
        }
        if ($mask & ACE4_WRITE_ACL) {
            $str .= 'C';
        }
        if ($mask & ACE4_WRITE_OWNER) {
            $str .= 'o';
        }
        if ($mask & ACE4_SYNCHRONIZE) {
            $str .= 's';
        }
        if ($str eq '') {
            $str = '-';
        }
        $outstr .= "$str ";

        # Process ACE Flags
        if ($flags & ACE4_FILE_INHERIT_ACE) {
            $outstr .= 'fi';
        }
        if ($flags & ACE4_DIRECTORY_INHERIT_ACE) {
            $outstr .= 'di';
        }
        if ($flags & ACE4_NO_PROPAGATE_INHERIT_ACE) {
            $outstr .= 'ni ';
        }
        if ($flags & ACE4_INHERIT_ONLY_ACE) {
            $outstr .= 'oi';
        }
        if ($flags & ACE4_SUCCESSFUL_ACCESS_ACE_FLAG) {
            $outstr .= 'sf';
        }
        if ($flags & ACE4_FAILED_ACCESS_ACE_FLAG) {
            $outstr .= 'ff';
        }

        # Print out the string
        print ACL "$outstr\n";
    }
    close ACL;
    return;
}

1;
