# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/PerlSyntheticClient.pm#2 $
#
# Copyright (c) 2007 Network Appliance, Inc.
# All rights reserved.

package nfs::ACL::PerlSyntheticClient;

## @summary
##  Perl synthetic client routines to be used as a base class to
##  the OS dependent ACL classes
##
## @description
##  Provides a default behavior for the subroutines that operate
##  on ACLs. These subroutines should only be called directly or
##  indirectly from the nfs::GenericAcl.pm library. This base class
##  uses the perl synthetic NFS client (nfs::NfsClient.pm) using v4
##  protocols for all operations.
##
## @usage
##  use nfs::ACL::PerlSyntheticClient;
##
## @status	review
##
## @author      dl-nfsqa-only (NFS QA Sunnyvale)
## @notes       Created by EP Network Storage Performance Lab, Inc.

use strict;
use File::Basename;
use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::NfsClient;
use nfs::NfsDef qw(:NFSCONSTANTS bitmap4_convert);
use File::Basename;

my $ng_check = 0;
my $h1 = host($main::FILER);
if ($h1->hosttype() =~ /filer-ng/) {
    $ng_check = 1;
    require nfs_setup_gx;
}



#-------------------------------------------------------
# Local subroutine prototypes
#-------------------------------------------------------
sub set_nfs_obj( $$ );
sub synth_to_acl( $ );
sub acl_to_synth( $ );

# We may get an NFS4ERR_DELAY from the filer and need to retry the
# command. The number of retries and the time to sleep between
# retires is specified here.
my $num_retries = 3;
my $retry_sleep_sec = 5;



# If an NfsClient object does not exist for the current ACL object,
# create one. If the filehandle for the desired filepath has been
# looked up, set it as the current filehandle, otherwise look up
# the filehandle and store it. Returns the NfsClient object.
sub set_nfs_obj( $$ ) {
    my ($this, $filepath) = @_;
    my ($ipaddr, $nfs_clnt);

    if (defined $this->{nfs_obj}) {
        # If an nfs_obj already exists on this acl, make sure
        # the authentication is set to the current user
        $this->{nfs_obj}->rpc->auth_sys_set;
    } else {

        if ($ng_check == 0) {
           $ipaddr = $this->{fc}->connhost();
        }

        if ( $main::ADDR_TYPE eq 'inet6' ) {
            my $h = host($main::FILER);
            $ipaddr = $h->net_best_address(sort_preference=>{family=>'inet6'});
            logcomment("IPv6 address used for synthetic client: $ipaddr");
        }

        # This has been modified to work for cmode
        if ($ng_check == 1) {
            #Extracting data IP
            $ipaddr = nfs_setup_gx::get_data_ip_cmode($main::FILER);
        }

        # Create the NfsClient object
        my $connected = 0;
        my $timer = time;
        while (! $connected) {

            eval {
                $this->{nfs_obj} = nfs::NfsClient->new(hostname => $ipaddr,
                                                       version  => 4,
                                                       minor_version => $main::NFS4_MINOR_VERSION,
                                                       proto    => 'tcp',
                                                       timeout  => $main::CMDTIMEOUT,
                                                       retry    => 5,
                                                       reconnect => 1,
                                                       qsize    => 1);
            }; 

            if ($@) { # Returned an error, process it
                if ($this->{retry_conn} &&
                    $@ =~ /Can't bind to socket|Connection reset/i &&
                    (time - $timer) < 300) {
                    # If {retry_conn} is set and the error recieved is
                    # "Can't bind to socket..." and our elapsed time is
                    # less than 5 minutes, sleep for a bit and try again
                    sleep 15;
                    $connected = 0;
                } else {
                    # Either {retry_conn} is not set, the error is not
                    # a socket binding error, or too much time has passed
                    croak "PerlSyntheticClient::set_nfs_obj: could not create " .
                      "synthetic client object: $@" if $@;
                }
            } else {
                # No error, we have an NfsClient. Stop trying.
                $connected = 1;
            }
        }
    }

    if (defined $this->{"fh:$filepath"}) {
        $this->{nfs_obj}{curr_fh} = $this->{"fh:$filepath"};
    } else {
        if ($main::NFS4_MINOR_VERSION != 0) {
            #Increase the maxrequestsize for max aces to work over v4.1
            eval { $this->{nfs_obj}->mount(path => $filepath,maxrequestsize => 71680) };
        } else {
            eval { $this->{nfs_obj}->mount(path => $filepath) };
        }
        croak "PerlSyntheticClient::set_nfs_obj: lookup of $filepath " .
          "failed: $@" if $@;
        $this->{"fh:$filepath"} = $this->{nfs_obj}{curr_fh};
    }

    return $this->{nfs_obj};
}

## @name ::create_with_acl
## @description
##  Create a file or directory with a the specified ACL using the
##  perl synthetic client library.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Updates fields in the acl object
##
## @return
##  nothing

sub create_with_acl ( $% ) {
    my ($this, %opts) = @_;
    my ($attrs, $dir, $filename, $nfs_clnt, $out);

    # Convert from ACL class format to Synthetic Client format
    acl_to_synth($this);
    $attrs->{FATTR4_ACL()} = $this->{ace};

    # Log the contents of the ACL
    logDbg("PerlSyntheticClient::create_with_acl: ACL:");

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ($this->{filename} =~ /\Q$this->{mountpoint}\E\/(\S+)/);

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ($dir eq '/.') {
        $dir = '';
    }
    $filename = basename($filename);

    if (defined($opts{mode})) {
        if ( (~$opts{mode} & $opts{mode}) ne '0' ) {
            # Mode was passed as an octal string,
            # convert it to an octal number.
            $opts{mode} = oct($opts{mode});
        }
        my $mode = sprintf "0%lo", $opts{mode};

        $attrs->{FATTR4_MODE()} = $mode;
    }

    # Lookup the filehandle for the directory where we will create
    $nfs_clnt = set_nfs_obj($this, "$this->{filer_path}$dir");

    if ($opts{filetype} eq 'file') {
        eval {
            $nfs_clnt->open(name    => $filename,
                            cr_mode => UNCHECKED,
                            access  => OPEN4_SHARE_ACCESS_WRITE,
                            attr    => $attrs,
                            retry   => 12);
        };
        croak "create_with_acl: synthetic client open failed: $@" if $@;
    } else {
        eval {
            $nfs_clnt->create(name => $filename,
                              type => NFDIR,
                              attr => $attrs);
        };
        croak "create_with_acl: synthetic client create failed: $@" if $@;
    }

    if (logCond("dbg")) {
        $out = `ls -ld $this->{filename}`;
        logDbg("$out");
    }

    # Convert from Synthetic Client format back to ACL class format
    synth_to_acl($this);

    return;
}

## @name ::get_acl
## @description
##  Uses the perl synthetic client library for getting an ACL. Gets
##  the ACL output from the specified file and then parses it into
##  a generic ACL.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Updates fields in the acl object
##
## @return
##  Nothing

sub get_acl ( $ ) {
    my ($this) = @_;
    my ($dir, $filename, $i, $ipaddr, $nfs_clnt, $out);

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.              
    #
    ($filename) = ($this->{filename} =~ /\Q$this->{mountpoint}\E\/(\S+)/);

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ($dir eq '/.') {
        $dir = '';
    }
    $filename = basename($filename);

    # Lookup the file
    $nfs_clnt = set_nfs_obj($this, "$this->{filer_path}$dir/$filename");

    # Retry the command if the filer is returning NFS4ERR_DELAY
    $i = $num_retries;
    do {
        if ($i != $num_retries) {
            logDbg('PerlSyntheticClient::get_acl: ' .
                   'retrying command due to NFS4ERR_DELAY');
            sleep ($retry_sleep_sec);
        }
        eval { $out = $nfs_clnt->getattr(attr => [ FATTR4_ACL() ]) };
    } while ($i-- > 0 && $@ =~ /NFS4ERR_DELAY/);
    croak "get_acl: synthetic client could not get attributes: $@" if $@;

    # Put the ACL on the object
    $this->{ace} = $out->{fattr}{FATTR4_ACL()};
    synth_to_acl($this);

    return;
}

## @name ::get_acl_array
## @description
##  Uses the perl synthetic client library for getting ACLs for all files
##  in a directory. Gets the ACL output from the specified directory and
##  then parses it putting the output from each file as a generic ACL in
##  an array of ACLs.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Used as a basis for constructing the command.
##
## @return
##  An array of ACLs

sub get_acl_array ( $ ) {
    my ($this) = @_;
    my ($acl, @acls, @attrs, $cookie, $cookieverf, $count, $dir, $eof,
        $filename, $nfs_clnt, $out, %readdir_opts, $rd);

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ($this->{filename} =~ /\Q$this->{mountpoint}\E\/(\S+)/);

    # Prepend "/" to the filename and check for current directory "/."
    $dir = "/" . $filename;
    if ($dir eq '/.') {
        $dir = '';
    }

    # Build a hash of additional options
    $readdir_opts{maxcount} = (defined($this->{total_max})) ?
      $this->{total_max} : 8192;
    $readdir_opts{dircount} = $readdir_opts{maxcount};
    $this->{total_max} = undef;

    # Call the synthetic client for a directory listing

    # Lookup the directory
    $nfs_clnt = set_nfs_obj($this, "$this->{filer_path}$dir");

    $count = 0;
    $cookie = 0;
    $cookieverf = '';
    $eof = 0;

    push @attrs, FATTR4_ACL();

    # Add an additional attribute to the readdir request, if specified
    if (defined $this->{add_attribute}) {
        # Get the supported attributes from the filer
        eval { $out = $nfs_clnt->getattr(attr => [ FATTR4_SUPPORTED_ATTRS() ]) };
        croak "get_acl_array: Could not get list of supported attributes " .
          "for including (random) attribute: $@" if $@;

        # Convert the attributes to an array
        my @supp_attr =
          @{bitmap4_convert(attrmask =>
                            $out->{fattr}{FATTR4_SUPPORTED_ATTRS()})};

        if ($this->{add_attribute} >= 0) {
            # The user has specified a specific attribute, verify that
            # it is supported and add it to @attrs if it is
            my $attr_ok = 0;
            foreach (@supp_attr) {
                if ($_ == $this->{add_attribute}) {
                    $attr_ok = 1;
                    last;
                }
            }
            if ($attr_ok) {
                push @attrs, $this->{add_attribute};
            } else {
                logWarn("Additional attribute <$this->{add_attribute}> is not " .
                        "supported. Not adding attribute to readdir request");
            }
        } else {
            # The user specified a negative value, select a random
            # attribute (other than FATTR4_ACL) from the supported attributes
            # and add it to @attrs
            srand(time);
            my $random = FATTR4_ACL();
            while ($random == FATTR4_ACL()) {
                # Keep selecting a random attr until we do not match FATTR4_ACL
                my $indx = int(rand(scalar @supp_attr));
                $random = $supp_attr[$indx];
            }
            logcomment "PerlSyntheticClient::get_acl_array calling readdir " .
              "for ACLs with additional attribute request <$random>. See " .
                "test documentation for information on passing this value " .
                  "into the test.";
            push @attrs, $random;
        }
    }

    # Make repeated readdir calls until 'eof' received, and record
    # the files/ACLs from the responses
    while (!$eof) {
        eval {
            $out = $nfs_clnt->readdir(cookie => $cookie,
                                      cookieverf => $cookieverf,
                                      attr => \@attrs,
                                      %readdir_opts);
        };
        croak "get_acl_array: readdir request <" . $count+1 . "> failed: $@" if $@;

        # Put the READDIR response length in the readdir_len array if passed in.
        if (defined($this->{readdir_len})) {
            # Find readdir result length
            $rd = (grep { $_->{op} == OP_READDIR } @{$out->{result_sizes}})[0];
            push @{$this->{readdir_len}}, $rd->{size};
        }

        foreach my $file (@{$out->{list}}) {
            $acl = nfs::ACL::AclClass->new(filename =>
                                           "$this->{filename}/$file->{name}");
            $acl->{ace} = $file->{fattr}{attr_vals}{FATTR4_ACL()};
            synth_to_acl($acl);
            $acls[$count] = $acl;
            $count++;
            $cookie = $file->{cookie};
        }

        $eof = $out->{eof};
        $cookieverf = $out->{cookieverf};
    }

    return (@acls);
}

# Convert ace list to compatible keys for GenericAcl library.
# GenericAcl library uses 'flags' instead of 'flag'
sub synth_to_acl( $ ) {
    my ($this) = shift;
    my $ace_cnt = 0;

    foreach (@{$this->{ace}}) {
        $ace_cnt++;
        $_->{flags} = $_->{flag};
        delete $_->{flag};
    }
    $this->{ace_cnt} = $ace_cnt;

    return;
}

## @name ::get_acl_support_attr
## @description
##  Uses the perl synthetic client library for getting the ACL
##  support attribute.
##
## @synopsis
##  $status = $acl->get_acl_support_attr();
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  The ACL support attribute or 0 if not returned by server

sub get_acl_support_attr ( $ ) {
    my ($this) = @_;
    my ($dir, $filename, $nfs_clnt, $out);

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ($this->{filename} =~ /\Q$this->{mountpoint}\E\/(\S+)/);

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ($dir eq '/.') {
        $dir = '';
    }
    $filename = basename($filename);

    # Lookup the file
    $nfs_clnt = set_nfs_obj($this, "$this->{filer_path}$dir/$filename");

    # Retrieve the ACLSUPPORT attribute
    eval {
        eval { $out = $nfs_clnt->getattr(attr => [ FATTR4_ACLSUPPORT() ]) };
    };
    croak "get_acl_support_attr: " .
      "synthetic client could not get attributes: $@" if $@;

    # Convert the returned decimal value to hex to match the
    # original return format from using v4t_
    return sprintf("%x", $out->{fattr}{FATTR4_ACLSUPPORT()});
}

## @name ::set_acl
## @description
##  Set the ACL associated with a given file using the perl synthetic
##  client library to write the ACL to the server. If the ace_cnt of
##  the passed in ACL is 0 (ace is an empty array) then the ACL is cleared out.
##
## @synopsis
##  $status = $acl->set_acl();
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  NFS4_OK on success or NFS V4 error string

sub set_acl ( $ ) {
    my ($this) = @_;
    my ($attrs, $dir, $err, $filename, $i, $nfs_clnt);

    # Convert from ACL class format to Synthetic Client format
    acl_to_synth($this);
    $attrs->{FATTR4_ACL()} = $this->{ace};

    # Log the contents of the file used to set the ACL
    logDbg("PerlSyntheticClient::set_acl: ACL:");

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to the
    # filer volume.
    ($filename) = ($this->{filename} =~ /\Q$this->{mountpoint}\E\/(\S+)/);

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ($dir eq '/.') {
        $dir = '';
    }
    $filename = basename($filename);

    # Lookup the file
    $nfs_clnt = set_nfs_obj($this, "$this->{filer_path}$dir/$filename");

    # Retry the command if the filer is returning NFS4ERR_DELAY
    $i = $num_retries;
    do {
        if ($i != $num_retries) {
            logDbg('PerlSyntheticClient::set_acl: ' .
                   'retrying command due to NFS4ERR_DELAY');
            sleep ($retry_sleep_sec);
        }

        eval {
            $nfs_clnt->setattr(attr => $attrs);
        };
    } while ($i-- > 0 && $@ =~ /NFS4ERR_DELAY/);

    # Convert from Synthetic Client format back to ACL class format
    synth_to_acl($this);

    # Return the NFS4 status received from the call. If any error is received,
    # it should be from the SETATTR, but it could come from a failed
    # PUTFH. This lets the user know where the problem is.
    if (!$@) {
        return 'NFS4_OK';
    } else {
        logDbg("PerlSyntheticClient::set_acl failed: $@");
        # If the error string doesn't contain an NFS4ERR then croak
        die $@ if ($@ !~ /NFS4ERR_/);
        ($err) = ($@ =~ /NFS4ERR_(\S+)/);
        $err =~ s/://; # Strip trailing :
        return 'NFS4ERR_' . $err;
    }
}

# Convert ace list to compatible keys for perl synthetic client.
# Synthetic client uses 'flag' instead of 'flags'
sub acl_to_synth( $ ) {
    my ($this) = shift;

    foreach (@{$this->{ace}}) {
        $_->{flag} = $_->{flags};
        delete $_->{flags};
    }

    return;
}

1;
