# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/V4tClient.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::ACL::V4tClient;

## @summary
##  V4t synthetic client routines to be used as a base class to
##  the OS dependent ACL classes
##
## @description
##  Provides a default behavior for the subroutines that operate
##  on ACLs. These subroutines should only be called directly or
##  indirectly from the nfs::GenericAcl.pm library. This base class uses
##  the synthetic NFS V4 client for all operations.
##
## @usage
##  use nfs::ACL::V4tClient;
##
## @status	review
##
## @author      dl-nfsqa-only (NFS QA Sunnyvale)
## @notes       Created by EP Network Storage Performance Lab, Inc.

use strict;
use File::Basename;
use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::ACL::AclNFSV4 qw(:ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);
use Data::Dumper;

#-------------------------------------------------------
# Global variables
#-------------------------------------------------------
use vars qw(
    @ISA
    @EXPORT
    @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
    create_with_acl
    get_acl
    get_acl_array
    set_acl
    get_acl_support_attr
    v4t_to_acl
    acl_to_v4t
);

#-------------------------------------------------------
# Local subroutine prototypes
#-------------------------------------------------------
sub v4t_to_acl( $$ );
sub acl_to_v4t( $$ );

# We may get an NFS4ERR_DELAY from the filer and need to retry the
# command. The number of retries and the time to sleep between
# retires is specified here.
my $num_retries     = 3;
my $retry_sleep_sec = 5;

## @name ::create_with_acl
## @description
##  Create a file or directory with a the specified ACL using the
##  v4t_crf synthetic client program.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Updates fields in the acl object
##
## @return
##  nothing
##
## @notes
##  Due to burt 166898 it does not look like I can specify an ACL when
##  creating a file so there is a workaround in place to have it
##  create the file and then do a setacl afterwards.

sub create_with_acl ( $% ) {
    my ( $this, %opts ) = @_;
    my ( $aclfile, $cmd, $dir, $err, $filename, $ipaddr, $out );
    my $extraflags = '';

    # Convert from ACL class format to v4t format
    $aclfile = "/tmp/create_with_acl.$$";
    acl_to_v4t( $this, $aclfile );

    # Log the contents of the file used to set the ACL
    my $t = `cat $aclfile`;
    logDbg("V4tClient::create_with_acl: aclfile: $aclfile\n$t\n");

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ( $this->{filename} =~ m{$this->{mountpoint}/(\S+)} );

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ( $dir eq '/.' ) {
        $dir = '';
    }
    $filename = basename($filename);

    $ipaddr = $this->{fc}->connhost();

    if ( defined( $opts{mode} ) ) {
        if ( ( ~$opts{mode} & $opts{mode} ) ne '0' ) {

            # Mode was passed as an octal string,
            # convert it to an octal number.
            $opts{mode} = oct( $opts{mode} );
        } ## end if ( ( ~$opts{mode} & ...
        my $mode = sprintf "0%lo", $opts{mode};
        $extraflags .= "-b $mode ";
    } ## end if ( defined( $opts{mode...
    if ( $opts{filetype} eq 'dir' ) {
        $extraflags .= '-r ';
    }

    if ( $opts{filetype} eq 'file' ) {

        # XXX - Hack to work around burt 166898
        unlink $aclfile;
        $cmd = "$this->{v4t_prefix} $this->{v4t_path}/v4t_crf -C "
            . "$this->{filer_path}$dir -I $ipaddr -F $filename $extraflags";
        logDbg("V4tClient::create_with_acl: $cmd");
        $out = `$cmd`;
        if ( $out =~ /close\(0\) -- went OK/i ) {
            $err = $this->set_acl();

            # try it again under certain error conditions
            if ( $err eq 'NFS4ERR_NOENT' ) {
                $err = $this->set_acl();
            }
            if ( $err ne 'NFS4_OK' ) {
                croak "create_with_acl: v4t_setacl failed: $err";
            }
        } else {
            croak "create_with_acl: v4t_crf failed\n$out";
        }
    } else {
        $cmd
            = "$this->{v4t_prefix} $this->{v4t_path}/v4t_crf -C "
            . "$this->{filer_path}$dir -I $ipaddr -F $filename "
            . "-a $aclfile $extraflags";
        logDbg("V4tClient::create_with_acl: $cmd");
        $out = `$cmd`;
        unlink $aclfile;
        if ( $out !~ /create-dir\S*?\(0\) -- went OK/i ) {
            croak "create_with_acl: v4t_crf failed\n$out";
        }
    } ## end else [ if ( $opts{filetype} eq...

    if ( logCond("dbg") ) {
        $out = `ls -ld $this->{filename}`;
        logDbg("$out");
    }
    return;
} ## end sub create_with_acl ( $% )

## @name ::get_acl
## @description
##  Uses the v4t_ls command for getting an ACL. Gets the ACL output
##  from the specified file and then parse it into a generic ACL.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Updates fields in the acl object
##
## @return
##  Nothing

sub get_acl ( $ ) {
    my ($this) = @_;
    my ( $cmd, $dir, $filename, $i, $ipaddr, $out );

    # print "Dumper Values : " . Dumper(\$this);
    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ( $this->{filename} =~ m{$this->{mountpoint}/(\S+)} );

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.

    $dir = "/" . dirname($filename);
    if ( $dir eq '/.' ) {
        $dir = '';
    }
    $filename = basename($filename);
    $Log->comment("\n The Filename that we are passing : $filename\n");
    $ipaddr = $this->{data_lif};
    $Log->comment("\n The Data_Lif that we are using : $ipaddr\n");
    $cmd
        = "$this->{v4t_prefix} $this->{v4t_path}/v4t_ls -a -C $this->{filer_path}$dir -I $ipaddr -F $filename";
    $Log->comment("\n\n\Entire get acl command : $cmd\n\n");
    logDbg("V4tClient::get_acl: $cmd");

    # Retry the command if the filer is returning NFS4ERR_DELAY
    $i = $num_retries;
    do {
        if ( $i != $num_retries ) {
            logDbg(
                'V4tClient::get_acl: retrying command due to NFS4ERR_DELAY');
            sleep($retry_sleep_sec);
        }
        $out = `$cmd`;
        logInfo("$cmd: $out");
    } while ( $i-- > 0 && $out =~ /\s18270000\s/ );# output will not capture the string "NFS4ERR_DELAY" but when we check the error 
	                                               # for NFS4ERR_DELAY the output is caputred the unique number "18270000" which will represent NFS4ERR_DELAY
    if ( $out !~ /Getting attributes -- went OK/i ) {
        croak "get_acl: v4t_ls could not get attributes\n$out";
    }

    v4t_to_acl( $this, $out );
    return $out;
} ## end sub get_acl ( $ )

## @name ::get_acl_array
## @description
##  Uses the v4t_ls command for getting ACLs for all files in a directory.
##  Gets the ACL output from the specified directory and then parses it
##  putting the output from each file as a generic ACL in an array of ACLs.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Used as a basis for constructing the command.
##
## @return
##  An array of ACLs

sub get_acl_array ( $ ) {
    my ($this) = @_;
    my ($acl,   @acls,   $cmd, $count,  $dir,     $filename,
        @files, $ipaddr, $out, $lenout, $options, $stderr
    );

    # Generate filename to hold stderr output from synthetic client
    $stderr = "/tmp/get_acl_array.$$";

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ( $this->{filename} =~ m{$this->{mountpoint}/(\S+)} );

    # Prepend "/" to the filename and check for current directory "/."
    $dir = "/" . $filename;
    if ( $dir eq '/.' ) {
        $dir = '';
    }

    # Build a string of additional options
    $options .= " -t $this->{total_max}" if defined( $this->{total_max} );
    $this->{total_max} = undef;

    # Call the synthetic client for a directory listing
    $ipaddr = $this->{fc}->connhost();
    $cmd
        = "$this->{v4t_prefix} $this->{v4t_path}/v4t_ls -a -D $this->{filer_path}$dir -I $ipaddr $options 2> $stderr";
    logDbg("V4tClient::get_acl_array: $cmd");

    $out = `$cmd`;
    if ( $out =~ /Step/i ) {
        unlink $stderr;
        croak
            "get_acl_array: v4t_ls could not get attributes for directory listing\n$out";
    }

    if ($out) {

        # Process the READDIR response length from stderr and save in
        # the readdir_len array if passed in.
        if ( defined( $this->{readdir_len} ) ) {
            $lenout = `cat $stderr`;
            ( @{ $this->{readdir_len} } )
                = $lenout =~ /READDIR response length: (\d+)/gi;
        }
        unlink $stderr;

        # Parse output for each file listing and store the acl
        @files = split( /\d+: /, $out );
        $count = 0;
        foreach my $file (@files) {
            if ($file) {
                if ( $file !~ /^(\S+) \[/ ) {
                    croak "get_acl_array: error parsing output for filename\n"
                        . "substring:---\n$file\n\nfrom output:---\n$out";
                }
                $acl = nfs::ACL::AclClass->new(
                    filename => "$this->{filename}/$1" );
                v4t_to_acl( $acl, $file );
                $acls[$count] = $acl;
                $count++;
            } ## end if ($file)
        } ## end foreach my $file (@files)
    } else {
        unlink $stderr;
    }

    return (@acls);
} ## end sub get_acl_array ( $ )

# name ::v4t_to_acl
# description
#  Creates an ACL object using the output from the v4t_ls command.
#  Could have gone table driven to keep from having to specify a
#  token twice (i.e. like Allow) here and in acl_to_v4t but this
#  approach is more obvious. Any changes here most likely need to
#  also be made in acl_to_v4t.
#  INTERNAL SUBROUTINE
#
# arg
#  this	Required: The name of the class. Automatically passed
#		in by PERL. Updates fields in the acl object
#
#  out		Required: Output from v4t_ls command.
#
# return
#  Nothing

sub v4t_to_acl( $$ ) {
    my ( $this, $out ) = @_;
    my ( $line, @lines, @tmp );
    my $ace_cnt = 0;

    $this->{ace_cnt} = 0;

    # Parse the output from the v4t_ls command and put into the acl
    @lines = split( /\r?\n/, $out );
    foreach $line (@lines) {
        if ( $line =~ /^\s+Acl:\s+(\d+)\s+entries/ ) {
            $this->{ace_cnt} = $1;
            next;
        }
        if ( $line =~ /^\s+Ace$/ ) {
            $ace_cnt++;
            next;
        }
        if ( $line =~ /^\s+Type:\s+(\S+)/ ) {
            my $type = $1;
            if ( $type =~ /allow/i ) {
                $type = ACE4_ACCESS_ALLOWED_ACE_TYPE;
            } elsif ( $type =~ /deny/i ) {
                $type = ACE4_ACCESS_DENIED_ACE_TYPE;
            } elsif ( $type =~ /audit/i ) {
                $type = ACE4_SYSTEM_AUDIT_ACE_TYPE;
            } elsif ( $type =~ /alarm/i ) {
                $type = ACE4_SYSTEM_ALARM_ACE_TYPE;
            } else {
                croak "v4t_to_acl: No valid ACE Type";
            }
            $this->{ace}[ $ace_cnt - 1 ]{type} = $type;
            next;
        } ## end if ( $line =~ /^\s+Type:\s+(\S+)/)
        if ( $line =~ s/^\s+Flags:\s+// ) {
            my $flags = 0;
            @tmp = split( /\s+/, $line );
            foreach (@tmp) {
            SWITCH: {
                    /FI/i && do {
                        $flags |= ACE4_FILE_INHERIT_ACE;
                        last SWITCH;
                    };
                    /DI/i && do {
                        $flags |= ACE4_DIRECTORY_INHERIT_ACE;
                        last SWITCH;
                    };
                    /NPI/i && do {
                        $flags |= ACE4_NO_PROPAGATE_INHERIT_ACE;
                        last SWITCH;
                    };
                    /IONLY/i && do {
                        $flags |= ACE4_INHERIT_ONLY_ACE;
                        last SWITCH;
                    };
                    /S/i && do {
                        $flags |= ACE4_SUCCESSFUL_ACCESS_ACE_FLAG;
                        last SWITCH;
                    };
                    /F/i && do {
                        $flags |= ACE4_FAILED_ACCESS_ACE_FLAG;
                        last SWITCH;
                    };
                    /GR/i && do {
                        $flags |= ACE4_IDENTIFIER_GROUP;
                        last SWITCH;
                    };
                } ## end SWITCH:
            } ## end foreach (@tmp)
            $this->{ace}[ $ace_cnt - 1 ]{flags} = $flags;
            next;
        } ## end if ( $line =~ s/^\s+Flags:\s+//)
        if ( $line =~ s/^\s+Mask:\s+// ) {
            my $mask = 0;
            @tmp = split( /\s+/, $line );
            foreach (@tmp) {
            SWITCH: {
                    /Rdata/i && do {
                        $mask |= ACE4_READ_DATA;
                        last SWITCH;
                    };
                    /Ldir/i && do {
                        $mask |= ACE4_LIST_DIRECTORY;
                        $this->{isdir} = 1;
                        last SWITCH;
                    };
                    /Wdata/i && do {
                        $mask |= ACE4_WRITE_DATA;
                        last SWITCH;
                    };
                    /Afile/i && do {
                        $mask |= ACE4_ADD_FILE;
                        $this->{isdir} = 1;
                        last SWITCH;
                    };
                    /Adata/i && do {
                        $mask |= ACE4_APPEND_DATA;
                        last SWITCH;
                    };
                    /Asubd/i && do {
                        $mask |= ACE4_ADD_SUBDIRECTORY;
                        $this->{isdir} = 1;
                        last SWITCH;
                    };
                    /Rna/i && do {
                        $mask |= ACE4_READ_NAMED_ATTRS;
                        last SWITCH;
                    };
                    /Wna/i && do {
                        $mask |= ACE4_WRITE_NAMED_ATTRS;
                        last SWITCH;
                    };
                    /Ex/i && do {
                        $mask |= ACE4_EXECUTE;
                        last SWITCH;
                    };
                    /Dch/i && do {
                        $mask |= ACE4_DELETE_CHILD;
                        last SWITCH;
                    };
                    /Rattr/i && do {
                        $mask |= ACE4_READ_ATTRIBUTES;
                        last SWITCH;
                    };
                    /Wattr/i && do {
                        $mask |= ACE4_WRITE_ATTRIBUTES;
                        last SWITCH;
                    };
                    /Del/i && do {
                        $mask |= ACE4_DELETE;
                        last SWITCH;
                    };
                    /Racl/i && do {
                        $mask |= ACE4_READ_ACL;
                        last SWITCH;
                    };
                    /Wacl/i && do {
                        $mask |= ACE4_WRITE_ACL;
                        last SWITCH;
                    };
                    /Wown/i && do {
                        $mask |= ACE4_WRITE_OWNER;
                        last SWITCH;
                    };
                    /Sync/i && do {
                        $mask |= ACE4_SYNCHRONIZE;
                        last SWITCH;
                    };
                } ## end SWITCH:
            } ## end foreach (@tmp)
            $this->{ace}[ $ace_cnt - 1 ]{mask} = $mask;
            next;
        } ## end if ( $line =~ s/^\s+Mask:\s+//)
        if ( $line =~ /^\s+User:\s+(\S+)/ ) {
            $this->{ace}[ $ace_cnt - 1 ]{who} = $1;
        }
        if ( $line =~ /^\s+Group:\s+(\S+)/ ) {
            $this->{ace}[ $ace_cnt - 1 ]{who} = $1;
        }
    } ## end foreach $line (@lines)
    if ( $this->{ace_cnt} != $ace_cnt ) {
        croak "v4t_to_acl: Number ACEs specified: $this->{ace_cnt} != to "
            . "number passed in: $ace_cnt";
    }
    return;
} ## end sub v4t_to_acl( $$ )

## @name ::get_acl_support_attr
## @description
##  Uses the v4t_ls command for getting the ACL support attribute.
##  This is always run as root.
##
## @synopsis
##  $status = $acl->get_acl_support_attr();
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  The ACL support attribute or 0 if not returned by server

sub get_acl_support_attr ( $ ) {
    my ($this) = @_;
    my ( $dir, $filename, $ipaddr, $out );

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to
    # the filer volume.
    ($filename) = ( $this->{filename} =~ m{$this->{mountpoint}/(\S+)} );

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ( $dir eq '/.' ) {
        $dir = '';
    }
    $filename = basename($filename);

    $ipaddr = $this->{fc}->connhost();
    $out
        = `$this->{v4t_prefix} $this->{v4t_path}/v4t_ls -e -C $this->{filer_path}$dir -I $ipaddr -F $filename`;
    if ( $out !~ /Getting attributes -- went OK/i ) {
        croak "get_acl_support_attr: v4t_ls could not get attributes\n$out";
    }

    if ( $out =~ /\s+AclSupport:\s+0x(\d+)/ ) {
        return ($1);
    } else {
        return (0);
    }
} ## end sub get_acl_support_attr ( $ )

## @name ::set_acl
## @description
##  Set the ACL associated with a given file. This is done by writing
##  the ACL into a temporary file and using v4t_setacl to write the ACL
##  to the server. If the ace_cnt of the passed in ACL is 0 then the
##  ACL is cleared out.
##
## @synopsis
##  $status = $acl->set_acl();
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  NFS4_OK on success or NFS V4 error string

sub set_acl ( $ ) {
    my ($this) = @_;

    # print "\$this =" . Dumper($this);
    my ( $aclfile, $cmd, $dir, $err, $filename, $i, $ipaddr, $out );

    # Convert from ACL class format to v4t format
    # $aclfile = "/tmp/set_acl.$$";
    $aclfile = $this->{aclfile};
    $Log->comment("\nThe acl file that we are passing to set acl :$aclfile\n");

    # Log the contents of the file used to set the ACL
    my $t = `cat $aclfile`;
	logDbg("V4tClient::set_acl: aclfile: $aclfile\n$t\n");

    # The filename will contain the full pathname relative to the mount.
    # Remove the mountpoint from the path to get the pathname relative to the
    # filer volume.
    ($filename) = ( $this->{filename} =~ m{$this->{mountpoint}/(\S+)} );

    # The filename could still have a directory component. Strip it off
    # to add to the filer directory and then get just the filename.
    $dir = "/" . dirname($filename);
    if ( $dir eq '/.' ) {
        $dir = '';
    }
    $filename = basename($filename);
    $ipaddr = $this->{data_lif};
    $cmd
        = "$this->{v4t_prefix} $this->{v4t_path}/v4t_setacl -C $this->{filer_path}$dir -I $ipaddr -F $filename -f $aclfile -c 1";
    $Log->comment("\n\n\Entire set acl command : $cmd\n\n");
    logDbg("V4tClient::set_acl: $cmd");

    # Retry the command if the filer is returning NFS4ERR_DELAY
    $i = $num_retries;
    do {
        if ( $i != $num_retries ) {
            logDbg(
                'V4tClient::set_acl: retrying command due to NFS4ERR_DELAY');
            sleep($retry_sleep_sec);
        }
        $out = `$cmd`;
        logInfo("$cmd: $out");
    } while ( $i-- > 0 && $out =~ /\sNFS4ERR_DELAY\s/ );
    unlink $aclfile;

    # The SETATTR could have failed due to a bad lookup or for other reasons.
    # If the command failed then see if there is a setattr error. If not,
    # grab the first NFS4ERR and return it. Worst case just return NFS4ERR_
    if ( $out =~ /set attr -- went OK/i ) {
        $err = 'NFS4_OK';
    } else {
        logDbg("V4tClient::set_acl failed: $out");
        ($err) = ( $out =~ /\sSETATTR\s\S+\sreturned\s(\S+)\s/m );
        if ( !defined($err) ) {
            ($err) = ( $out =~ /\sNFS4ERR_(\S+)\s/m );
            $err = 'NFS4ERR_' . $err;
        }
    } ## end else [ if ( $out =~ /set attr -- went OK/i)
    return ($err);
    return ($out);
} ## end sub set_acl ( $ )

# name ::acl_to_v4t
# description
#  Converts an ACL object into output that can be used by v4t_setacl.
#  Could have gone table driven to keep from having to specify a
#  token twice (i.e. like Allow) here and in v4t_to_acl but this
#  approach is more obvious. Any changes here most likely need to
#  also be made in v4t_to_acl.
#  INTERNAL SUBROUTINE
#
# arg
#  this	Required: The name of the class. Automatically passed
#		in by PERL.
#
#  aclfile	Required: Name of file to write ACL to.
#
# return
#  Nothing

sub acl_to_v4t( $$ ) {
    my ( $this, $aclfile ) = @_;
    print "acl to v4t \$this " . Dumper($this);

    # print "\$aclfile  = $aclfile" . Dumper($aclfile);
    my ( $flags, $i, $line, @lines, $mask, @tmp, $type );

    # Open up the file that holds the ACL
    open ACL, "> $aclfile"
        or croak "V4tClient::acl_to_v4t: Cannot create $aclfile: $!";

    # Step through each ACE and print it to the file
    for ( $i = 0; $i < $this->{ace_cnt}; $i++ ) {
        print ACL "Ace\nType: ";

        # Process ACE Type
        $type = $this->{ace}[$i]{type};
        if ( $type == ACE4_ACCESS_ALLOWED_ACE_TYPE ) {
            print ACL "Allow\n";
        } elsif ( $type == ACE4_ACCESS_DENIED_ACE_TYPE ) {
            print ACL "Deny\n";
        } elsif ( $type == ACE4_SYSTEM_AUDIT_ACE_TYPE ) {
            print ACL "Audit\n";
        } elsif ( $type == ACE4_SYSTEM_ALARM_ACE_TYPE ) {
            print ACL "Alarm\n";
        } else {
            close ACL;
            unlink $aclfile;
            croak "acl_to_v4t: Bad ACL Type: $type";
        }

        # Process ACE Flags
        print ACL "Flags: ";
        $flags = $this->{ace}[$i]{flags};
        if ( $flags & ACE4_FILE_INHERIT_ACE ) {
            print ACL "FI ";
        }
        if ( $flags & ACE4_DIRECTORY_INHERIT_ACE ) {
            print ACL "DI ";
        }
        if ( $flags & ACE4_NO_PROPAGATE_INHERIT_ACE ) {
            print ACL "NPI ";
        }
        if ( $flags & ACE4_INHERIT_ONLY_ACE ) {
            print ACL "IONLY ";
        }
        if ( $flags & ACE4_SUCCESSFUL_ACCESS_ACE_FLAG ) {
            print ACL "S ";
        }
        if ( $flags & ACE4_FAILED_ACCESS_ACE_FLAG ) {
            print ACL "F ";
        }
        if ( $flags & ACE4_IDENTIFIER_GROUP ) {
            print ACL "GR ";
        }
        if ( $flags == 0 ) {
            print ACL "None ";
        }

        # Process ACE Mask
        print ACL "\nMask: ";
        $mask = $this->{ace}[$i]{mask};
        if ( $mask & ACE4_READ_DATA ) {
            if ( $this->{isdir} ) {
                print ACL "Ldir ";
            } else {
                print ACL "Rdata ";
            }
        } ## end if ( $mask & ACE4_READ_DATA)
        if ( $mask & ACE4_WRITE_DATA ) {
            if ( $this->{isdir} ) {
                print ACL "Afile ";
            } else {
                print ACL "Wdata ";
            }
        } ## end if ( $mask & ACE4_WRITE_DATA)
        if ( $mask & ACE4_APPEND_DATA ) {
            if ( $this->{isdir} ) {
                print ACL "Asubd ";
            } else {
                print ACL "Adata ";
            }
        } ## end if ( $mask & ACE4_APPEND_DATA)
        if ( $mask & ACE4_READ_NAMED_ATTRS ) {
            print ACL "Rna ";
        }
        if ( $mask & ACE4_WRITE_NAMED_ATTRS ) {
            print ACL "Wna ";
        }
        if ( $mask & ACE4_EXECUTE ) {
            print ACL "Ex ";
        }
        if ( $mask & ACE4_DELETE_CHILD ) {
            print ACL "Dch ";
        }
        if ( $mask & ACE4_READ_ATTRIBUTES ) {
            print ACL "Rattr ";
        }
        if ( $mask & ACE4_WRITE_ATTRIBUTES ) {
            print ACL "Wattr ";
        }
        if ( $mask & ACE4_DELETE ) {
            print ACL "Del ";
        }
        if ( $mask & ACE4_READ_ACL ) {
            print ACL "Racl ";
        }
        if ( $mask & ACE4_WRITE_ACL ) {
            print ACL "Wacl ";
        }
        if ( $mask & ACE4_WRITE_OWNER ) {
            print ACL "Wown ";
        }
        if ( $mask & ACE4_SYNCHRONIZE ) {
            print ACL "Sync ";
        }
        if ( $mask == 0 ) {
            print ACL "None ";
        }

        # Print out who field
        if ( $flags & ACE4_IDENTIFIER_GROUP ) {
            print ACL "\nGroup: ";
        } else {
            print ACL "\nUser: ";
        }
        print ACL "$this->{ace}[$i]{who}\n";
    } ## end for ( $i = 0; $i < $this...
    close ACL;
    return;
} ## end sub acl_to_v4t( $$ )

1;
