# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/ACL/Windows.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
package nfs::ACL::Windows;

## @summary
##   Class to handle OS dependent portion of ACLs on Windows.
##
## @description
##   This is a sub-class of the AclClass package which provides
##   support for Windows specific portion of ACLs. This module
##   needs to handle converting ACLs from NT to NFS V4 format.
##
##   Here is the table used to convert an NT ACL to a NFS V4 ACL:
##
##   NT ACL                             NFS V4 ACL
##   ================================================================
##   ***********************    Types    ****************************
##   ACCESS_ALLOWED_ACE_TYPE 0x00               ACE4_ACCESS_ALLOWED_ACE_TYPE 0x00
##   ACCESS_DENIED_ACE_TYPE 0x01                ACE4_ACCESS_DENIED_ACE_TYPE 0x01
##   SYSTEM_AUDIT_ACE_TYPE 0x02                 ACE4_SYSTEM_AUDIT_ACE_TYPE 0x02
##   SYSTEM_ALARM_ACE_TYPE 0x03                 ACE4_SYSTEM_ALARM_ACE_TYPE 0x03
##   ACCESS_ALLOWED_COMPOUND_ACE_TYPE 0x04	ACE4_ACCESS_ALLOWED_ACE_TYPE 0x00
##   ACCESS_ALLOWED_OBJECT_ACE_TYPE 0x05	ACE4_ACCESS_ALLOWED_ACE_TYPE 0x00
##   ACCESS_DENIED_OBJECT_ACE_TYPE 0x06         ACE4_ACCESS_DENIED_ACE_TYPE 0x01
##   SYSTEM_AUDIT_OBJECT_ACE_TYPE 0x07          ACE4_SYSTEM_AUDIT_ACE_TYPE 0x02
##   SYSTEM_ALARM_OBJECT_ACE_TYPE 0x08          ACE4_SYSTEM_ALARM_ACE_TYPE 0x03
##
##   ***********************    Flags    ****************************
##   OBJECT_INHERIT_ACE	0x01		ACE4_FILE_INHERIT_ACE 0x01
##   CONTAINER_INHERIT_ACE 0x02		ACE4_DIRECTORY_INHERIT_ACE 0x02
##   NO_PROPAGATE_INHERIT_ACE 0x04	ACE4_NO_PROPAGATE_INHERIT_ACE 0x04
##   INHERIT_ONLY_ACE 0x08		ACE4_INHERIT_ONLY_ACE 0x08
##   INHERITED_ACE 0x10                 --
##   VALID_INHERIT_FLAGS 0x1F           --
##   SUCCESSFUL_ACCESS_ACE_FLAG 0x40	ACE4_SUCCESSFUL_ACCESS_ACE_FLAG 0x10
##   FAILED_ACCESS_ACE_FLAG 0x80	ACE4_FAILED_ACCESS_ACE_FLAG 0x20
##   --					ACE4_IDENTIFIER_GROUP 0x40
##
##   ***********************    Masks    ****************************
##   FILE_READ_DATA 0x01		ACE4_READ_DATA 0x01
##   FILE_LIST_DIRECTORY 0x01		ACE4_LIST_DIRECTORY 0x01
##   FILE_WRITE_DATA 0x02		ACE4_WRITE_DATA 0x02
##   FILE_ADD_FILE 0x02			ACE4_ADD_FILE 0x02
##   FILE_APPEND_DATA 0x04		ACE4_APPEND_DATA 0x04
##   FILE_ADD_SUBDIRECTORY 0x04		ACE4_ADD_SUBDIRECTORY 0x04
##   FILE_READ_EA 0x08			ACE4_READ_NAMED_ATTRS 0x08
##   FILE_WRITE_EA 0x10			ACE4_WRITE_NAMED_ATTRS 0x10
##   FILE_EXECUTE 0x20			ACE4_EXECUTE 0x20
##   FILE_DELETE_CHILD 0x40		ACE4_DELETE_CHILD 0x40
##   FILE_READ_ATTRIBUTES 0x80		ACE4_READ_ATTRIBUTES 0x80
##   FILE_WRITE_ATTRIBUTES 0x100	ACE4_WRITE_ATTRIBUTES 0x100
##   DELETE 0x10000			ACE4_DELETE 0x10000
##   READ_CONTROL 0x20000		ACE4_READ_ACL 0x20000
##   STANDARD_RIGHTS_READ 0x20000	ACE4_READ_ACL 0x20000
##   STANDARD_RIGHTS_WRITE 0x20000	ACE4_READ_ACL 0x20000
##   STANDARD_RIGHTS_EXECUTE 0x20000	ACE4_READ_ACL 0x20000
##   WRITE_DAC 0x40000			ACE4_WRITE_ACL 0x40000
##   WRITE_OWNER 0x80000		ACE4_WRITE_OWNER 0x80000
##   SYNCHRONIZE 0x100000               ACE4_SYNCHRONIZE 0x100000
##
##      These are mapped access rights. The value is converted by
##      windows to mean the same thing as a group of individual
##      rights, but the value of the map is not the 'sum' of the
##      values of the individual rights. These will need to be
##      converted during execution.
##   GENERIC_READ 0x80000000
##   GENERIC_READ = STANDARD_RIGHTS_READ | FILE_READ_DATA |
##                  FILE_READ_ATTRIBUTES | FILE_READ_EA | SYNCHRONIZE
##
##   GENERIC_WRITE 0x40000000
##   GENERIC_WRITE = STANDARD_RIGHTS_WRITE | FILE_WRITE_DATA |
##                   FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA |
##                   FILE_APPEND_DATA | SYNCHRONIZE
##
##   GENERIC_EXECUTE 0x20000000
##   GENERIC_EXECUTE = STANDARD_RIGHTS_EXECUTE | FILE_READ_ATTRIBUTES |
##                     FILE_EXECUTE | SYNCHRONIZE
##
##   GENERIC_ALL 0x10000000
##   GENERIC_ALL = STANDARD_RIGHTS_ALL | GENERIC_READ | GENERIC_WRITE |
##                 GENERIC_EXECUTE | FILE_DELETE_CHILD
##
##      These are straight mappings, i.e. the components ORed together
##      generate the value of the map, so we don't need to convert
##      them during the execution.
##   STANDARD_RIGHTS_ALL 0x1F0000
##   STANDARD_RIGHTS_ALL = DELETE | READ_CONTROL | WRITE_DAC |
##                         WRITE_OWNER | SYNCHRONIZE
##
##   STANDARD_RIGHTS_REQUIRED 0xF0000
##   STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER
##
## @usage
##   use nfs::ACL::Windows
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @notes
##  Windows XP supports 1820 ACEs.
##
##  This implementation is by no means complete. There are a lot of
##  nuances to how Windows handles NT ACLs and trying to map them to
##  and from NFS V4 ACLs is a learning experience. Let us know about
##  cases that this library needs to handle.

use strict;
use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::ACL::AclNFSV4 qw(:ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);
use nfs::ACL::AclBase;
use vars qw( @ISA );

# Define the Win32 ACE constants
# Types
use constant ACCESS_ALLOWED_ACE_TYPE          => 0x00;
use constant ACCESS_DENIED_ACE_TYPE           => 0x01;
use constant SYSTEM_AUDIT_ACE_TYPE            => 0x02;
use constant SYSTEM_ALARM_ACE_TYPE            => 0x03;
use constant ACCESS_ALLOWED_COMPOUND_ACE_TYPE => 0x04;
use constant ACCESS_ALLOWED_OBJECT_ACE_TYPE   => 0x05;
use constant ACCESS_DENIED_OBJECT_ACE_TYPE    => 0x06;
use constant SYSTEM_AUDIT_OBJECT_ACE_TYPE     => 0x07;
use constant SYSTEM_ALARM_OBJECT_ACE_TYPE     => 0x08;

# Flags
use constant OBJECT_INHERIT_ACE         => 0x01;
use constant CONTAINER_INHERIT_ACE      => 0x02;
use constant NO_PROPAGATE_INHERIT_ACE   => 0x04;
use constant INHERIT_ONLY_ACE           => 0x08;
use constant INHERITED_ACE              => 0x10;
use constant VALID_INHERIT_FLAGS        => 0x1F;
use constant SUCCESSFUL_ACCESS_ACE_FLAG => 0x40;
use constant FAILED_ACCESS_ACE_FLAG     => 0x80;

# ACE mask
use constant FILE_READ_DATA           => 0x01;
use constant FILE_LIST_DIRECTORY      => 0x01;
use constant FILE_WRITE_DATA          => 0x02;
use constant FILE_ADD_FILE            => 0x02;
use constant FILE_APPEND_DATA         => 0x04;
use constant FILE_ADD_SUBDIRECTORY    => 0x04;
use constant FILE_READ_EA             => 0x08;
use constant FILE_WRITE_EA            => 0x10;
use constant FILE_EXECUTE             => 0x20;
use constant FILE_DELETE_CHILD        => 0x40;
use constant FILE_READ_ATTRIBUTES     => 0x80;
use constant FILE_WRITE_ATTRIBUTES    => 0x100;
use constant DELETE                   => 0x10000;
use constant READ_CONTROL             => 0x20000;
use constant STANDARD_RIGHTS_READ     => 0x20000;
use constant STANDARD_RIGHTS_WRITE    => 0x20000;
use constant STANDARD_RIGHTS_EXECUTE  => 0x20000;
use constant WRITE_DAC                => 0x40000;
use constant WRITE_OWNER              => 0x80000;
use constant SYNCHRONIZE              => 0x100000;
use constant GENERIC_ALL              => 0x10000000;
use constant GENERIC_EXECUTE          => 0x20000000;
use constant GENERIC_WRITE            => 0x40000000;
use constant GENERIC_READ             => 0x80000000;
use constant STANDARD_RIGHTS_ALL      => 0x1F0000;
use constant STANDARD_RIGHTS_REQUIRED => 0xF0000;

# (UN)PROTECTED_DACL_SECURITY_INFORMATION - used for clearing aces
use constant UNPROTECTED_DACL_SECURITY_INFORMATION => 0x20000000;
use constant PROTECTED_DACL_SECURITY_INFORMATION   => 0x80000000;

#---------------------------------------------------------------------
# Inherit functionality from AclBase class.
# Override the get_acl and set_acl methods.
#---------------------------------------------------------------------
our @ISA = qw(nfs::ACL::AclBase);

## @name ::new
## @description
##  Windows specific class constructor. It overrides default
##  behavior of the base exports class.
##
## @return
##  Windows specific export object.

sub new ( $ ) {
    my $class = shift;

    # Include the Win32::Security modules
    require Win32::Security::NamedObject;
    require Win32::Security::ACL;
    require Win32::Security::Raw;
    require Win32::Security::SID;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class (
        get_acl_cmd    => 'Win32::Security::NamedObject::dacl',
        set_acl_cmd    => 'Win32::Security::NamedObject::dacl',

        # Override defaults
        @_
    );
}

## @name ::create_with_acl
## @description
##  Create a file or directory with a the specified ACL. Since Windows
##  does not have synthetic client, just create the file or directory
##  and set the ACL on it.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL. Updates fields in the acl object
##
##  opts	Required: The options from the create_with_acl()
##		call in GenericAcl.pm.
##
## @return
##  nothing

sub create_with_acl ( $% ) {
    my ($this, %opts) = @_;
    my ($filename, $object, $status);

    # Create the file or directory
    if ($opts{filetype} eq 'dir') {
        mkdir("$this->{filename}") or
            croak "create_with_acl: mkdir failed: $this->{filename}: $!";
        $this->{isdir} = 1;
    } else {
        open TST, "+> $this->{filename}" or
            croak "create_with_acl: create failed: $this->{filename}: $!";
        close TST;
    }

    if (defined($opts{mode})) {
        logWarn("create_with_acl: Windows does not support mode option");
    }

    # Since the ACL could not be created at the same time the file
    # was and since set_acl needs to know the windows OWNER and GROUP
    # SIDs, then get this information.

    # Change pathname to windows format
    $filename = $this->{filename};
    $filename =~ s/\//\\/g;

    # Create a Win32::Security object for the filename, which will
    # extract the owner and group sids
    $object = make_security_object($this, 'SE_FILE_OBJECT', $filename);

    # Set the ACL on the file or directory
    $status = $this->set_acl();
    if ($status ne 'NFS4_OK') {
        croak "create_with_acl: set_acl failed: $status";
    }
    return;
}

## @name ::get_acl
## @description
##  Uses the Win32::Security::NamedObject functionality to retrieve
##  the ACL. Calls ntacl_to_acl to get and store the ACL in generic
##  format. Since Windows does not have a synthetic client ignore the
##  use_synthetic flag and proceed with the operation.
##  Updates fields in the ACL object.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  Nothing

sub get_acl ( $ ) {
    my ($this) = @_;
    my ($filename, $object);

    # If this file is a directory then set the isdir flag
    if (-d $this->{filename}) {
        $this->{isdir} = 1;
    }

    # Change pathname to windows format
    $filename = $this->{filename};
    $filename =~ s/\//\\/g;

    $object = make_security_object($this, 'SE_FILE_OBJECT', $filename);

    ntacl_to_acl($this, $object, 0);

    return;
}

## @name ::set_acl
## @description
##  Set the ACL associated with a given file using the
##  Win32::Security::NamedObject functionality.
##  Since Windows does not have a synthetic client ignore the
##  use_synthetic flag and proceed with the operation.
##
## @arg
##  this	Required: The name of the class. Automatically passed
##		in by PERL.
##
## @return
##  NFS4_OK on success or NFS4ERR_??? on error since it is not really
##  talking to an NFS V4 server.

sub set_acl ( $ ) {
    my ($this) = @_;
    my ($filename, $object);
    my (@aces, $acl_obj);

    # Change pathname to windows format
    $filename = $this->{filename};
    $filename =~ s/\//\\/g;

    # Clear all non-inherited ACEs from the ACL
    # We cannot ignore any errors here, if the clearing of the
    # ACLs fails, then setting the ACL afterwards may produce
    # unexpected results.
    eval {
        $object = make_security_object($this, 'SE_FILE_OBJECT', $filename);

        # If we simply whack the ACL from the object, then there is a
        # good chance we won't have permission to re-write the ACL
        # with something else, so we create an ACL that only contains
        # Rattr, Racl, Wacl for EVERYONE@.
        my $set_acl = nfs::ACL::AclClass->new();
        my $acl_mask = ACE4_READ_ATTRIBUTES|ACE4_READ_ACL|ACE4_WRITE_ACL;
        nfs::GenericAcl::set_acl_ace(acl => $set_acl,
                    type => ACE4_ACCESS_ALLOWED_ACE_TYPE, consistent => 'no',
                    mask => $acl_mask, mask_how => 'SET', who => 'EVERYONE@');
        @aces = acl_to_ntacl($set_acl);

        $acl_obj = Win32::Security::ACL->new($object->{objectType}, @aces);
        $object->dacl($acl_obj, UNPROTECTED_DACL_SECURITY_INFORMATION);
    };

    if ($@) {
        logDbg("Windows::set_acl: Could not clear all non-inherited ACEs: $@");
        return("NFS4ERR_???: $@");
    }

    # Get the ACEs to be set from the class object
    @aces = acl_to_ntacl($this);

    # Set the ACL
    eval {
        # Create a null ACL object
        $acl_obj = Win32::Security::ACL->new($object->{objectType}, '');
        # Add the ACEs (this reorders them according to Win32 requirements)
        $acl_obj->addAces(@aces);
        # Set the ACL on the security object (the file or directory)
        $object->dacl($acl_obj, UNPROTECTED_DACL_SECURITY_INFORMATION);
    };

    if (! $@) {
        return ('NFS4_OK');
    } else {
        logDbg("Windows::set_acl: $@");
        return ("NFS4ERR_???: $@");
    }

}

# name ::ntacl_to_acl
# description
#  Using a Win32::Security::NamedObject object, retrieves
#  the ACL information from the object and stores it in the class
#  object in generic format.
#
# arg
#  this         Required: The name of the class. Automatically passed
#		in by PERL. Updates fields in the acl object
#
#  object	Required: A Win32::Security::NamedObject.
#
#  owner_only   Required: If set, only retrieves the owner/owner_sid
#               for the object.
#
# return
#  Nothing

sub ntacl_to_acl( $$$ ) {
    my ($this, $object, $owner_only) = @_;
    my ($acl_obj, @aces, $ace, $ace_flags, $ace_mask, $ace_type, $flags,
        $mask, $type, $who, $who_sid);
    my $ace_cnt = 0;

    # Save the binary representation of the SID for use in print_acl
    $this->{owner_sid_binary} = $object->ownerSid();

    $this->{owner_sid} = $object->ownerTrustee();
    if ($this->{owner_sid} eq 'S-1-5-32-544') {
        $this->{owner} = 'BUILTIN\Administrators';
    } elsif ($this->{owner_sid} eq 'S-1-1-0') {
        $this->{owner} = 'EVERYONE'
    } else {
        $this->{owner} = $this->{owner_sid};
    }

    if ($owner_only) {
        return;
    }

    $acl_obj = $object->dacl();
    @aces = $acl_obj->aces();

    foreach $ace (@aces) {
        $flags = 0x00;

        # The trustee on the ace is the 'who'
        # Win32::Security uses sid as the trustee
        $who = $ace->trustee();

        # Use the Win32::Security::SID library to determine the
        # SID for the who. This may be reassigned in the following
        # code, but do it for robustness here. If the information
        # does not map in the if/elsif section below, at least the who
        # and who_sid will be what Windows sees.
        $who_sid = Win32::Security::SID::ConvertSidToStringSid(
                        Win32::Security::SID::ConvertNameToSid($who));

        # Get the sid for owner and group
        my $sid_owner = $this->{owner_sid};
        my $sid_group = $this->{group_sid};

        # Map to the well known NFS V4 who names
        if ($who eq $sid_owner) {
            $who = 'OWNER@';
        } elsif ($who eq $sid_group) {
            $who = 'GROUP@';
            $flags = ACE4_IDENTIFIER_GROUP;
        } elsif ($who =~ /^everyone$/i || $who =~ /^S-1-1-0$/) {
            $who = 'EVERYONE@';
            $who_sid = 'S-1-1-0';
        } elsif ($this->{owner} =~ /^BUILTIN\\Administrators$/i &&
            $who_sid =~ /^S-1-5-32-544$/) {
            # Sometimes windows will list the file owner or group as
            # BUILTIN\Administrators which is a group. This can happen because
            # the sesssion is logged in as an Administrator which is also a
            # member of the group. The code below will map these cases
            # to OWNER@ or GROUP@.
            $this->{owner} = '\administrator';
            $who = 'OWNER@';
            $this->{owner_sid} = $who_sid;
        } elsif (!$this->{owner} && $who !~ /^BUILTIN\\/i) {
            # If the owner is not set and this ACE SID is not a builtin
            # SID then set the owner to this.
            # This technically should never happen.
            $this->{owner} = $who;
            $who = 'OWNER@';
            $this->{owner_sid} = $who_sid;
        } elsif (!$this->{group} && $who_sid =~ /^S-1-5-32-544$/i) {
            # If the group is not set and this ACE SID is the standard
            # builtin group then set the group to this.
            $this->{group} = $who;
            $who = 'GROUP@';
            $flags = ACE4_IDENTIFIER_GROUP;
            $this->{group_sid} = $who_sid;
        }
        $this->{ace}[$ace_cnt]{who} = $who;
        $this->{ace}[$ace_cnt]{who_sid} = $who_sid;

        # Set the NFS V4 ACL mask
        $mask = 0x00;
        $ace_mask = $ace->rawAccessMask();

        # Convert mapped mask values to individual components
        # Add the component values and remove the mapped mask value
        # Start with GENERIC_ALL, which includes the others, so the
        # mappings will propagate correctly.
        if ($ace_mask & GENERIC_ALL) {
            $ace_mask |= STANDARD_RIGHTS_ALL | GENERIC_READ | GENERIC_WRITE |
                         GENERIC_EXECUTE | FILE_DELETE_CHILD;
            $ace_mask ^= GENERIC_ALL;
        }
        if ($ace_mask & GENERIC_EXECUTE) {
            $ace_mask |= STANDARD_RIGHTS_EXECUTE | FILE_READ_ATTRIBUTES |
                         FILE_EXECUTE | SYNCHRONIZE;
            $ace_mask ^= GENERIC_EXECUTE;
        }
        if ($ace_mask & GENERIC_WRITE) {
            $ace_mask |= STANDARD_RIGHTS_WRITE | FILE_WRITE_DATA |
                         FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA |
                         FILE_APPEND_DATA | SYNCHRONIZE;
            $ace_mask ^= GENERIC_WRITE;
        }
        if ($ace_mask & GENERIC_READ) {
            $ace_mask |= STANDARD_RIGHTS_READ | FILE_READ_DATA |
                         FILE_READ_ATTRIBUTES | FILE_READ_EA | SYNCHRONIZE;
            $ace_mask ^= GENERIC_READ;
        }

        # Convert remaining NT ACL bits to NFS V4 ACL values
        if ($ace_mask & FILE_READ_DATA) {
            # Also maps to
            #   FILE_LIST_DIRECTORY -> ACE4_LIST_DIRECTORY
            $mask |= ACE4_READ_DATA;
            $ace_mask ^= FILE_READ_DATA;
        }
        if ($ace_mask & FILE_WRITE_DATA) {
            # Also maps to
            #   FILE_ADD_FILE -> ACE4_ADD_FILE
            $mask |= ACE4_WRITE_DATA;
            $ace_mask ^= FILE_WRITE_DATA;
        }
        if ($ace_mask & FILE_APPEND_DATA) {
            # Also maps to
            #   FILE_ADD_SUBDIRECTORY -> ACE4_ADD_SUBDIRECTORY
            $mask |= ACE4_APPEND_DATA;
            $ace_mask ^= FILE_APPEND_DATA;
        }
        if ($ace_mask & FILE_READ_EA) {
            $mask |= ACE4_READ_NAMED_ATTRS;
            $ace_mask ^= FILE_READ_EA;
        }
        if ($ace_mask & FILE_WRITE_EA) {
            $mask |= ACE4_WRITE_NAMED_ATTRS;
            $ace_mask ^= FILE_WRITE_EA;
        }
        if ($ace_mask & FILE_EXECUTE) {
            $mask |= ACE4_EXECUTE;
            $ace_mask ^= FILE_EXECUTE;
        }
        if ($ace_mask & FILE_DELETE_CHILD) {
            $mask |= ACE4_DELETE_CHILD;
            $ace_mask ^= FILE_DELETE_CHILD;
        }
        if ($ace_mask & FILE_READ_ATTRIBUTES) {
            $mask |= ACE4_READ_ATTRIBUTES;
            $ace_mask ^= FILE_READ_ATTRIBUTES;
        }
        if ($ace_mask & FILE_WRITE_ATTRIBUTES) {
            $mask |= ACE4_WRITE_ATTRIBUTES;
            $ace_mask ^= FILE_WRITE_ATTRIBUTES;
        }
        if ($ace_mask & DELETE) {
            $mask |= ACE4_DELETE;
            $ace_mask ^= DELETE;
        }
        if ($ace_mask & READ_CONTROL) {
            # Also maps to
            #   STANDARD_RIGHTS_READ -> ACE4_READ_ACL
            #   STANDARD_RIGHTS_WRITE -> ACE4_READ_ACL
            #   STANDARD_RIGHTS_EXECUTE -> ACE4_READ_ACL
            $mask |= ACE4_READ_ACL;
            $ace_mask ^= READ_CONTROL;
        }
        if ($ace_mask & WRITE_DAC) {
            $mask |= ACE4_WRITE_ACL;
            $ace_mask ^= WRITE_DAC;
        }
        if ($ace_mask & WRITE_OWNER) {
            $mask |= ACE4_WRITE_OWNER;
            $ace_mask ^= WRITE_OWNER;
        }
        if ($ace_mask & SYNCHRONIZE) {
            $mask |= ACE4_SYNCHRONIZE;
            $ace_mask ^= SYNCHRONIZE;
        }
        $this->{ace}[$ace_cnt]{mask} = $mask;

        # Make sure all mask bits were processed
        if ($ace_mask) {
            croak "ntacl_to_acl: Extra mask bits found: $ace_mask";
        }

        # Set the NFS V4 ACL flags
        # Ignore INHERITED_ACE and VALID_INHERIT_FLAGS flags
        # (no NFS V4 ACL correspondence)
        $ace_flags = $ace->rawAceFlags();
        if ($ace_flags & OBJECT_INHERIT_ACE) {
            $flags |= ACE4_FILE_INHERIT_ACE;
        }
        if ($ace_flags & CONTAINER_INHERIT_ACE) {
            $flags |= ACE4_DIRECTORY_INHERIT_ACE;
        }
        if ($ace_flags & NO_PROPAGATE_INHERIT_ACE) {
            $flags |= ACE4_NO_PROPAGATE_INHERIT_ACE;
        }
        if ($ace_flags & INHERIT_ONLY_ACE) {
            $flags |= ACE4_INHERIT_ONLY_ACE;
        }
        if ($ace_flags & SUCCESSFUL_ACCESS_ACE_FLAG) {
            $flags |= ACE4_SUCCESSFUL_ACCESS_ACE_FLAG;
        }
        if ($ace_flags & FAILED_ACCESS_ACE_FLAG) {
            $flags |= ACE4_FAILED_ACCESS_ACE_FLAG;
        }
        $this->{ace}[$ace_cnt]{flags} = $flags;

        # Set the NFS V4 ACL type
        $ace_type = $ace->rawAceType();
        if ($ace_type == ACCESS_ALLOWED_ACE_TYPE ||
            $ace_type == ACCESS_ALLOWED_COMPOUND_ACE_TYPE ||
            $ace_type == ACCESS_ALLOWED_OBJECT_ACE_TYPE) {
            $type = ACE4_ACCESS_ALLOWED_ACE_TYPE;
        } elsif ($ace_type == ACCESS_DENIED_ACE_TYPE ||
                 $ace_type == ACCESS_DENIED_OBJECT_ACE_TYPE) {
            $type = ACE4_ACCESS_DENIED_ACE_TYPE;
        } elsif ($ace_type == SYSTEM_AUDIT_ACE_TYPE ||
                 $ace_type == SYSTEM_AUDIT_OBJECT_ACE_TYPE) {
            $type = ACE4_SYSTEM_AUDIT_ACE_TYPE;
        } elsif ($ace_type == SYSTEM_ALARM_ACE_TYPE ||
                 $ace_type == SYSTEM_ALARM_OBJECT_ACE_TYPE) {
            $type = ACE4_SYSTEM_ALARM_ACE_TYPE;
        } else {
            croak "ntacl_to_acl: Invalid ACE Type: $ace_type";
        }
        $this->{ace}[$ace_cnt]{type} = $type;

        $ace_cnt++;
    }

    $this->{ace_cnt} = $ace_cnt;
    return;
}

# name ::acl_to_ntacl
# description
#  Converts an ACL object into an array of ACEs that can be used
#  by a Win32::Security::NamedObject object to set its ACL.
#
# arg
#  this         Required: The name of the class. Automatically passed
#		in by PERL.
#
# return
#  Array of aces in proper format for use with
#  Win32::Security::NamedObject->dacl().

sub acl_to_ntacl( $ ) {
    my ($this) = @_;
    my ($ace, @aces, $flags, $i, $mask, $sid, $this_flags, $this_mask,
        $this_type, $type, $who, $who_sid);

    # Step through each ACE and print it to the file
    for ($i = 0; $i < $this->{ace_cnt}; $i++) {
        $this_type = $this->{ace}[$i]{type};
        $this_flags = $this->{ace}[$i]{flags};
        $this_mask = $this->{ace}[$i]{mask};
        $who = $this->{ace}[$i]{who};
        $who_sid = $this->{ace}[$i]{who_sid};

        # If the mask is 0 or just set to synchronize then skip this ACE
        # Otherwise SetACL or windows sets default values on it and it
        # collides with other ACEs.
        if ($this_mask == 0 || $this_mask == ACE4_SYNCHRONIZE) {
            next;
        }

        # Set the NT ACL flags
        $flags = 0x00;
        if ($this_flags & ACE4_FILE_INHERIT_ACE) {
            $flags |= OBJECT_INHERIT_ACE;
        }
        if ($this_flags & ACE4_DIRECTORY_INHERIT_ACE) {
            $flags |= CONTAINER_INHERIT_ACE;
        }
        if ($this_flags & ACE4_NO_PROPAGATE_INHERIT_ACE) {
            $flags |= NO_PROPAGATE_INHERIT_ACE;
        }
        if ($this_flags & ACE4_INHERIT_ONLY_ACE) {
            $flags |= INHERIT_ONLY_ACE;
        }
        if ($this_flags & ACE4_SUCCESSFUL_ACCESS_ACE_FLAG) {
            $flags |= SUCCESSFUL_ACCESS_ACE_FLAG;
        }
        if ($this_flags & ACE4_FAILED_ACCESS_ACE_FLAG) {
            $flags |= FAILED_ACCESS_ACE_FLAG;
        }

        # Set the NT ACL type
        # NFS V4 ACL types 0x00 .. 0x03 map directly to
        # NT ACL types 0x00 .. 0x03.
        $type = $this_type;

        # Set the NT ACL mask
        $mask = 0x00;
        if ($this_mask & ACE4_READ_DATA) {
            # Also maps to
            #   ACE4_LIST_DIRECTORY -> FILE_LIST_DIRECTORY
            $mask |= FILE_READ_DATA;
            $this_mask ^= ACE4_READ_DATA;
        }
        if ($this_mask & ACE4_WRITE_DATA) {
            # Also maps to
            #   ACE4_ADD_FILE -> FILE_ADD_FILE
            $mask |= FILE_WRITE_DATA;
            $this_mask ^= ACE4_WRITE_DATA;
        }
        if ($this_mask & ACE4_APPEND_DATA) {
            # Also maps to
            #   ACE4_ADD_SUBDIRECTORY -> FILE_ADD_SUBDIRECTORY
            $mask |= FILE_APPEND_DATA;
            $this_mask ^= ACE4_APPEND_DATA;
        }
        if ($this_mask & ACE4_READ_NAMED_ATTRS) {
            $mask |= FILE_READ_EA;
            $this_mask ^= ACE4_READ_NAMED_ATTRS;
        }
        if ($this_mask & ACE4_WRITE_NAMED_ATTRS) {
            $mask |= FILE_WRITE_EA;
            $this_mask ^= ACE4_WRITE_NAMED_ATTRS;
        }
        if ($this_mask & ACE4_EXECUTE) {
            $mask |= FILE_EXECUTE;
            $this_mask ^= ACE4_EXECUTE;
        }
        if ($this_mask & ACE4_DELETE_CHILD) {
            $mask |= FILE_DELETE_CHILD;
            $this_mask ^= ACE4_DELETE_CHILD;
        }
        if ($this_mask & ACE4_READ_ATTRIBUTES) {
            $mask |= FILE_READ_ATTRIBUTES;
            $this_mask ^= ACE4_READ_ATTRIBUTES;
        }
        if ($this_mask & ACE4_WRITE_ATTRIBUTES) {
            $mask |= FILE_WRITE_ATTRIBUTES;
            $this_mask ^= ACE4_WRITE_ATTRIBUTES;
        }
        if ($this_mask & ACE4_DELETE) {
            $mask |= DELETE;
            $this_mask ^= ACE4_DELETE;
        }
        if ($this_mask & ACE4_READ_ACL) {
            # Also maps to
            #   ACE4_READ_ACL -> STANDARD_RIGHTS_READ
            #   ACE4_READ_ACL -> STANDARD_RIGHTS_WRITE
            #   ACE4_READ_ACL -> STANDARD_RIGHTS_EXECUTE
            $mask |= READ_CONTROL;
            $this_mask ^= ACE4_READ_ACL;
        }
        if ($this_mask & ACE4_WRITE_ACL) {
            $mask |= WRITE_DAC;
            $this_mask ^= ACE4_WRITE_ACL;
        }
        if ($this_mask & ACE4_WRITE_OWNER) {
            $mask |= WRITE_OWNER;
            $this_mask ^= ACE4_WRITE_OWNER;
        }
        if ($this_mask & ACE4_SYNCHRONIZE) {
            $mask |= SYNCHRONIZE;
            $this_mask ^= ACE4_SYNCHRONIZE;
        }

        # Make sure all mask bits were processed
        # We probably don't need to do this, because all ACE4_* types
        # are known. The initial reason for reporting extra bits was
        # to ensure we had processed all valid bits. However, extra bits
        # may be left due to a REPLACE call on setting an ACE, so
        # we leave this out for now.
        #if ($this_mask) {
        #    croak "acl_to_ntacl: Extra mask bits found: $this_mask";
        #}

        if (!defined $who_sid) {
            if ($who eq 'OWNER@') {
                $who_sid = $this->{owner_sid};
            } elsif ($who eq 'GROUP@') {
                $who_sid = $this->{group_sid};
            } elsif ($who eq 'EVERYONE@') {
                $who_sid = 'S-1-1-0';
            } elsif ($who =~ /^BUILTIN\\Administrators$/i) {
                $who_sid = 'S-1-5-32-544';
            } elsif ($who =~ /WIN_MAP_TO_CURRENT/) {
                # Map the marker above to the current user
                my $tmp = `x:\\bin\\whoami /user /sid`;
                ($who, $who_sid) = $tmp =~ /"(\S+)"\s+(\S+)/;
                $this->{ace}[$i]{who} = $who;
                $this->{ace}[$i]{who_sid} = $who_sid;
                $who = $who_sid;
            } else {
                # If the who is not one of the well known ids, and
                # is not WIN_MAP_TO_CURRENT, try to figure out
                # the sid for that user.
                $who_sid = Win32::Security::SID::ConvertSidToStringSid(
                               Win32::Security::SID::ConvertNameToSid($who));
                croak "acl_to_ntacl: Could not determine SID for $who"
                  if (!defined $who_sid);
            }
        }

        $ace = Win32::Security::ACE->new('SE_FILE_OBJECT',
                                         $type, $flags, $mask, $who_sid);

        push @aces, $ace;
    }

    return @aces;
}

# name ::make_security_object
# description
#  Initializes a Win32::Security::NamedObject object of the given
#  type and filename. Stores the owner and group SIDs for the object
#  in the class object.
#
# arg
#  this         Required: The name of the class. Automatically passed
#		in by PERL.
#
# return
#  Reference to a Win32::Security::NamedObject object.

sub make_security_object( $$$ ) {
    my ($this, $type, $filename) = @_;
    my ($group, $object, @security_info, $sid, $sid_length, $trustee);

    # Create the NamedObject
    $object = Win32::Security::NamedObject->new($type, $filename);

    # If owner_sid or group_sid are not set, find the needed values
    # In Windows, 'sid' represents a binary form of the security identifier
    # and 'trustee' represents the string form (e.g. S-1-1-0), which we
    # refer to as the 'sid' in the ACL libraries.
    if (!defined $this->{owner_sid} || !defined $this->{group_sid}) {
        @security_info =
          Win32::Security::Raw::GetNamedSecurityInfo(
                  $object->{objectName},
                  $object->{objectType},
                  'OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION');

        if (!defined $this->{owner_sid}) {
            $this->{owner_sid} = $object->ownerTrustee();
            if ($this->{owner_sid} =~ /S-/) {
                my $validate =
                  Win32::Security::SID::ConvertStringSidToSid(
                                                    $this->{owner_sid});
                if (!defined $validate) {
                    croak "Burt 217893: Unable to parse Trustee/SID: " .
                      "$this->{owner_sid}";
                }
            }
        }

        if (!defined $this->{group_sid}) {
            $trustee = undef;
            $group = $security_info[1];
            $sid_length = Win32::Security::Raw::GetLengthSid($group);
            $sid = Win32::Security::Raw::CopyMemory_Read($group, $sid_length);
            $trustee = Win32::Security::SID::ConvertSidToName($sid);
            $this->{group_sid} = $trustee;
        }
    }

    return $object;
}

1;
