# -*- perl -*-

# ---------------------------------------------------
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
# Any other use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
# ---------------------------------------------------
#{[(
package nfs::Auth_Utils;
##
## @summary Authentication utilities which contains sub routines for
##          modifying UID, GID entries in AD and NIS.
##
## @usage use nfs::Auth_Utils;
##
## @status new
##
## @see-also lib/TCD.pm {name of related test case}
##
## @author Sanjay Shanthakumar (sanjay09@netapp.com)
##
## @pod here

#-------------------------------------------------------------------------------
# Compiler directives.
#-------------------------------------------------------------------------------
use strict;
use base qw(Exporter);
use warnings;

#-------------------------------------------------------------------------------
# Modules Import
#-------------------------------------------------------------------------------

use Tharn;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF OBJECT);
use Net::LDAP;
use Carp qw(croak);
use Data::Dumper;
use nfs::Kerberos;
use NACL::C::Vserver;
use NACL::CS::Vserver;
use NACL::C::DiagSecdAuthentication;
use NACL::C::Exceptions::Vserver::DoesNotExist qw{:try};
use NACL::C::Exceptions::VserverServicesDns::AlreadyExists qw{:try};
use nfs::CommonUtilities qw{client_execute $Node_cmd $Client_cmd check_client_and_sudo return_client };

BEGIN {
    use vars qw(
        $YP_CONF_PATH
        $SECOND_YP_PATH
        $YPBIND_PATH
        $USER_PASSWD
        $LOGIN_SHELL
        $UNIX_HOME_DIR
        $GROUP_ID
        $SUDO_PATH
        $UID_LOWER_LIMIT
        $UID_UPPER_LIMIT
    );

    param('GROUP_ID',       '-mesg',    'Group id to which the user has to be added');
    param('YP_CONF_PATH',   '-default', '/etc/yp.conf');
    param('SECOND_YP_PATH', '-default', '/var/yp/binding');
    param('UNIX_HOME_DIR',  '-default', '/home');
    param('LOGIN_SHELL',    '-default', '/bin/sh');
    param('YPBIND_PATH',    '-default', '/etc/init.d/ypbind');
    param('USER_PASSWD',    '-default', 'netapp1!');
    param('SUDO_PATH',      '-default', '/usr/software/bin/sudo');
    param('UID_LOWER_LIMIT','-default', '10000' );
    param('UID_UPPER_LIMIT','-default', '1000000' );
}

our @EXPORT = qw(
    Aux_gid_LDAP_uid_modify
    Aux_gid_LDAP_group_create
    establish_ldap
    Aux_gid_filer_setup
    Aux_gid_client_setup
    user_login_and_verify_groups
    clear_client_config
);

our @EXPORT_OK = @EXPORT;
our $previous_yp_domain;
our $previous_ns_switch;
our $script_user;
our $script_user_added;

=head1 METHODS

=head2 Aux_gid_LDAP_uid_modify

This function checks if the user exists in the AD, if not creates the user
or else it modifies the user account and other attributes like first name
user principle and so on.

Aux_gid_LDAP_uid_modify( kdc_AD => \%<KDC-Details> , user => $<Username> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=item user

(Required) : Name of the user who has to be added into AD (if not present)

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub Aux_gid_LDAP_uid_modify( % ) {
    my %opts = validate(
        @_,
        {   kdc_AD => {type => HASHREF},
            user   => {type => SCALAR},
        }
    );
    $Log->enter();

    # Establish the NACL::C::Client object to the KDC server
    my $KDC_Connection_Obj = establish_kdc_connection($opts{kdc_AD}->{kdc_server});

    # Making sure the server handle retrieved is a windows client.
    if (!$KDC_Connection_Obj->is_windows()) {
        croak("KDC_Server_Setup : Incorrect or invalid client is detected for the KDC server:\n$opts{ kdc }->{ kdc_server }");
    }

    my $DC = kerberos_create_dc(kdc_realm => $opts{kdc_AD}->{kdc_realm});
    my $i;
    # Obtain the name of the user who is running the script
    $script_user = `echo \$USER`;
    chomp( $script_user );

    # First add the user who is running the script and then add the actual user
    for( $i = 0 ; $i < 2 ; $i++ ){
        # Add client name in AD computers account & Create Service Principal and verify that is registered
        my $base_dn        = "$opts{ kdc_AD }->{ kdc_ou },$DC";
        my $user           = ( $i == 0 ) ? $script_user : $opts{user};
        my $user_principal = "cn=$user,$base_dn";
        my $AD_user_info   = $KDC_Connection_Obj->execute(command => "dsget user \"$user_principal\"");

        #verify if the dsget user failed, then add the user
        if ($AD_user_info =~ /dsget succeeded/) {
            if( $i == 0 ){
                $script_user_added = 0;
            }
            # Do nothing. Its successful
        } elsif ($AD_user_info =~ /dsget failed:Directory object not found/) {
            my $user_create_output = $KDC_Connection_Obj->execute(command => "dsadd user \"$user_principal\"");

            #verify  if the dsget user failed then abort the script exection
            $AD_user_info = $KDC_Connection_Obj->execute(command => "dsget user \"$user_principal\"");
            if ($AD_user_info !~ /dsget succeeded/) {
                croak("AD user creation failed : $user_create_output");
            }
            if( $i == 0 ){
                $script_user_added = 1;
            }
        } else {
            croak("dsget is failing with a different error: $AD_user_info");
        }

        # modifying the user attributes in order to enable the user account,
        # add first name, user principal name and to change the options to
        # passwd never expires
        my $command = "dsmod user \"$user_principal\" -disabled no ";
        $command .= " -pwdneverexpires yes -display $opts{user} ";
        $command .= " -fn $opts{user} -upn $opts{user}\@$opts{kdc_AD}->{kdc_domain} ";
        $command .= " -pwd $::USER_PASSWD ";
        my $modify_user_output = $KDC_Connection_Obj->execute(command => $command);
        if ($modify_user_output !~ /dsmod succeeded/) {
            croak('Unable to modify the user attributes in the AD');
        }

        my $ldap = establish_ldap(kdc_AD => $opts{kdc_AD});

        #search if the kdc_admin_user is present in LDAP, if so
        #obtain its group ID
        $Log->comment("Checking if ".$opts{kdc_AD}->{kdc_user}." is present in the AD");
        my $groupID;
        my $result = $ldap->search(
            base   => $base_dn,
            scope  => 'sub',
            filter => "(& (ObjectClass=user) (cn=$opts{kdc_AD}->{kdc_user}) )",
            attrs  => ['cn', 'primaryGroupID', 'uid'],
        );

        $Log->comment('Obtaining the primaryGroupID of user: ' . $opts{kdc_AD}->{kdc_user});
        if ($result->entries) {
            foreach my $entry ($result->entries) {
                $groupID = ${$entry->get('primaryGroupID')}[0];
            }
        } else {
            croak("The requested user : $opts{kdc_AD}->{kdc_user} was not found\n");
        }

        #search for the user that was added before and modify its GID
        # Check if the id <user> output contains the UID of the user
        my $user_dn = "cn=$user," . $base_dn;
        my $domain  = _get_domain_name($opts{kdc_AD}->{kdc_domain});

        # Generate a UID number > UID_LOWER_LIMIT
        my $uid;
        if( $i == 0 ){
            $uid = _return_script_user_uid( $user );
        } else {
            $uid = _generate_uid( $ldap , $base_dn );
        }

        # We are modifying the groupid, uidNumber, the default shell
        # homedirectory and other unix related attributes without which
        # the user on a unix system wont be authenticated in the AD.

        my $result = $ldap->modify(
            $user_dn,
            replace => {
                'msSFU30NisDomain'  => $domain,
                'gidNumber'         => $::GROUP_ID,
                'unixUserPassword'  => $::USER_PASSWD,
                'uid'               => $user,
                'msSFU30Name'       => $user,
                'unixHomeDirectory' => $::UNIX_HOME_DIR . '/' . $user,
                'loginShell'        => $::LOGIN_SHELL,
                'uidNumber'         => $uid,
            }
        );
        if ($result->error ne 'Success') {    #error occured
            croak("Error while updating user values: " . $result->error);
        }
        $ldap->unbind;
    }

    $Log->exit();
    return;
} ## end sub Aux_gid_LDAP_uid_modify( % )

# This subroutine generates UID for the user
sub _generate_uid($$){
    my $ldap = shift;
    my $base_dn = shift; 

    while(1){
        my $uid = int (rand( $::UID_UPPER_LIMIT ) + $::UID_LOWER_LIMIT);

        my $result = $ldap->search(
            base => $base_dn ,
            scope => 'sub',
            filter => "(& (ObjectClass=user) (uidNumber=$uid) )",
            attrs => [ 'cn' , 'primaryGroupID' , 'uid', 'msSFU30Name' , 'uidNumber' , 'unixUserPassword'],
        );

        if( !$result->entries ){
            $Log->comment("No user exists with uid: $uid. Hence this uid can be assigned to the user");
            return $uid;
        }
    }
}

#Obtain the uid of the user running the script
sub _return_script_user_uid($){
    my $user = shift;
   
    my $id_message = `id $user`;
    my $user_id;

    # Check if the id <user> output contains the UID of the user
    $Log->comment('Obtaining UID of user: '.$user);
    if ($id_message =~ /uid=(\d+)\($user\)/) {
        $user_id = $1; 
    } else { # 2 issues. 1. unable to run the command or 2. invalid username provided
        croak(
            "The client could not even run the id command for user $user"
        );
    }   
 
    return $user_id;
}

=head1 METHODS

=head2 establish_ldap

This function establishes the ldap connection with the AD. It uses
net::LDAP library to create a new object and bind to the AD.

establish_ldap( kdc_AD => \%<KDC-Details> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=back

=item Return

LDAP object is returned on success or croaks on failure

=back

=cut

# Establish an LDAP connection using Net::LDAP library
sub establish_ldap(%) {
    my %opts = validate(@_, {kdc_AD => HASHREF,});

    $Log->enter();

    # Add client name in AD computers account & Create Service Principal and verify that is registered
    my $DC = kerberos_create_dc(kdc_realm => $opts{kdc_AD}->{kdc_realm});
    my $base_dn = "$opts{ kdc_AD }->{ kdc_ou },$DC";

    #make an LDAP connection to the windows AD
    $Log->comment('Creating new connection to AD');
    my $ldap = Net::LDAP->new($opts{kdc_AD}->{kdc_fqn}, port => $::LDAP_PORT)
        or croak("Unable to connect to the LDAP server:$@");
    my $dn = 'cn=' . $opts{kdc_AD}->{kdc_user} . ',' . $base_dn;
    $Log->comment('Bind the connection with AD');
    my $message = $ldap->bind(dn => $dn, password => $opts{kdc_AD}->{kdc_passwd});
    if ($message->code) {    # if bind fails then fail the script
        croak("Failure during LDAP bind: " . $message->error);
    }

    $Log->exit();
    return $ldap;
} ## end sub establish_ldap(%)

=head1 METHODS

=head2 Aux_gid_LDAP_group_create

This function checks if the required number of groups with the given name 
exists in the AD, if not creates the groups or else it adds the user account
into each of those groups.

Aux_gid_LDAP_group_create( kdc_AD => \%<KDC-Details> , user => $<Username> ,
                           group_name => $<group-prefix-name> , number_of_groups => $<num-of-groups> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=item user

(Required) : Name of the user who has is present in the AD (if not present)

=item group_name

(Required) : Prefix of the group name 

=item number_of_groups

(Required) : Total number of groups that has to be verified in the AD.

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub Aux_gid_LDAP_group_create( % ) {
    my %opts = validate(
        @_,
        {   kdc_AD           => {type => HASHREF},
            user             => {type => SCALAR},
            group_name       => {type => SCALAR},
            number_of_groups => {type => SCALAR},
        }
    );
    $Log->enter();

    # Establish the NACL::C::Client object to the KDC server
    my $KDC_Connection_Obj = establish_kdc_connection($opts{kdc_AD}->{kdc_server});

    # Making sure the server handle retrieved is a windows client.
    if (!$KDC_Connection_Obj->is_windows()) {
        croak("KDC_Server_Setup : Incorrect or invalid client is detected for the KDC server: " . $opts{kdc}->{kdc_server});
    }

    my $DC = kerberos_create_dc(kdc_realm => $opts{kdc_AD}->{kdc_realm});
    my $ldap = establish_ldap(kdc_AD => $opts{kdc_AD});
    my $base_dn = "$opts{ kdc_AD }->{ kdc_ou },$DC";

    $Log->comment('Check if the user: ' . $opts{user} . ' is present ');
    $Log->comment('before adding the user into the groups');

    my $result = $ldap->search(
        base   => $base_dn,
        scope  => 'sub',
        filter => "(& (ObjectClass=user) (cn=$opts{user}) )",
        attrs  => ['cn', 'uid'],
    );

    if (!$result->entries) {
        croak("$opts{user} user does not exist " . $result->error());
    }

    # Check if the groupname exists, if not create the requested number of groups.
    my $i = 0;
    for ($i = 1; $i <= $opts{number_of_groups}; $i++) {
        my $base_dn         = "$opts{ kdc_AD }->{ kdc_ou },$DC";
        my $group_name      = $opts{group_name} . $i;
        my $group_principal = "cn=$group_name,$base_dn";
        $Log->comment('Check if group: ' . $group_name . ' is present in AD');
        my $AD_group_info = $KDC_Connection_Obj->execute(command => "dsget group \"$group_principal\"");

        #verify if the dsget group succeeded or failed, if it fails, then create the group
        if ($AD_group_info =~ /dsget failed:Directory object not found/) {
            $Log->comment('Group not found.So, create ' . $group_name);
            my $group_create_output = $KDC_Connection_Obj->execute(command => "dsadd group \"$group_principal\"");

            #verify  if the dsget group failed then abort the script exection
            my $AD_group_create = $KDC_Connection_Obj->execute(command => "dsget group \"$group_principal\"");

            if ($AD_group_create =~ /dsget succeeded/) {

                # do nothing
            } else {
                croak("AD group creation failed : $group_create_output");
            }
        } elsif ($AD_group_info =~ /dsget succeeded/) {

            # Do nothing
        } else {
            croak("dsget failed with another reason: $AD_group_info");
        }

        # search for the group that was added and then add the user
        $Log->comment('Search for group: ' . $group_name . ' and add the user: ' . $opts{user});
        my $result = $ldap->search(
            base   => $base_dn,
            scope  => 'sub',
            filter => "(| (objectclass => group ) (cn=$group_name))",
            attrs  => ['cn', 'msSFU30NisDomain'],
        );

        # If the group was added successfully then the search will be a success
        if ($result->entries) {
            foreach my $entry ($result->entries) {
                $entry->add('memberUid' => [$opts{user}]);
                $entry->update($ldap);
                $Log->comment("${$entry->get('memberUid')}[0] was successfully added into $group_name\n");
            }
        } else {    #indicate that the search failed
            croak("The requested group : $group_name was not found even after adding the group through dsadd");
        }
    }    #end of for

    $ldap->unbind;
    $Log->exit();

    return;
} ## end sub Aux_gid_LDAP_group_create( % )

=head1 METHODS

=head2 Aux_gid_filer_setup

This subroutine performs dns setup, ns-switch modifications and secd
authentication translate in order to verify if the user has already 
been added to AD.

Aux_gid_filer_setup( kdc_AD => \%<KDC-Details> , vserver => $<Vserver-name> , filer_name => $<Filer-name> ,
                     user => $<User-name> , ns_switch => $<Ns-switch> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=item user

(Required) : Name of the user who has is present in the AD (if not present)

=item vserver

(Required) : Name of the vserver present on the filer.

=item filer_name

(Required) : Name of the filer

=item ns_switch

(Required) : Ns_switch that has to be configured on the vserver. Either nis or ldap.

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub Aux_gid_filer_setup(%) {
    my %opts = validate(
        @_,
        {   vserver    => {type => SCALAR},
            filer_name => {type => SCALAR},
            user       => {type => SCALAR},
            kdc_AD     => {type => HASHREF},
            ns_switch  => {
                type     => SCALAR,
                optional => 1,
                default  => 'ldap',
            },
        }
    );

    # for Auth_gid testing ns_switch should not be file
    if ($opts{ns_switch} =~ /file/) {
        croak('ldap or nis is expected and not file for ns_switch');
    }

    # configure the dns on the filer/vsim to use the kdc domain
    $Log->comment("Create the dns for the vserver");
    try {
        NACL::C::VserverServicesDns->create(
            command_interface => $Node_cmd,
            vserver           => "$opts{vserver}",
            domains           => [$opts{kdc_AD}->{kdc_domain}],
            'name-servers'    => [$opts{kdc_AD}->{kdc_ip}],
            state             => 'enabled',
            timeout           => "2",
            attempts          => "1"
        );
    }
    catch NACL::C::Exceptions::VserverServicesDns::AlreadyExists with {
        $Log->comment("AlreadyExists so Modify");
        NACL::C::VserverServicesDns->modify(
            command_interface => $Node_cmd,
            vserver           => "$opts{vserver}",
            domains           => [$opts{kdc_AD}->{kdc_domain}],
            'name-servers'    => [$opts{kdc_AD}->{kdc_ip}],
            state             => 'enabled',
            timeout           => "2",
            attempts          => "1"
        );
    };

    # Modify the ns-switch on the vserver to use nis/ldap
    my $vserver = NACL::CS::Vserver->fetch(
        command_interface => $Node_cmd,
        filter            => {vserver => $opts{vserver}},
        requested_fields  => ['ns-switch'],
    );
    $previous_ns_switch = $vserver->ns_switch();
    my @ns_switch = split(',', $opts{ns_switch});
    _modify_ns_switch(ns_switch => \@ns_switch, vserver => $opts{vserver});

    if ($opts{ns_switch} =~ /ldap/) {
        KRB_Filer_LDAP_create(
            filer_name    => $opts{filer_name},
            filer_vserver => $opts{vserver},
            kdc           => $opts{kdc_AD},
        );
    }
    
    # authenticate the user using secd authenticate
    NACL::C::DiagSecdAuthentication->translate(
        command_interface => $Node_cmd,
        vserver           => $opts{vserver},
        node              => $Node_cmd->node(),
        'unix-user-name'  => $opts{user},
    );

    $Log->comment("User: $opts{user} was successfully added to Nis/Ldap");

    return;
} ## end sub Aux_gid_filer_setup(%)

=head1 METHODS

=head2 Aux_gid_client_setup

This sub routine performs client setup by performing yp.conf entries, change the domainname to that of
AD/NIS and then perform a ypbind restart.

Aux_gid_client_setup( kdc_AD => \%<KDC-Details> , user => $<Username> , client_object => $<NACL-Object> ,
                      group_name => $<group-prefix-name> , number_of_groups => $<num-of-groups> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=item user

(Required) : Name of the user who has is present in the AD (if not present)

=item group_name

(Required) : Prefix of the group name 

=item number_of_groups

(Required) : Total number of groups that has to be verified in the AD.

=item client_object

(Required) : NACL::C::Client object needs to be passed into the sub routine.

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub Aux_gid_client_setup(%) {
    my %opts = validate(
        @_,
        {   client_object    => {type => OBJECT},
            kdc_AD           => {type => HASHREF},
            user             => {type => SCALAR},
            group_name       => {type => SCALAR},
            number_of_groups => {type => SCALAR},
        }
    );

    $Log->enter();

    # Validate the client obhject and enter the sudo mode
    # Add these functions into nfs::CommonUtilities.pm
    check_client_and_sudo($opts{client_object});

    if ($opts{client_object}->type() =~ /unix-linux/) {

        #linux client
        _setup_linux_client_yp_conf(
            client_object    => $opts{client_object},
            kdc_AD           => $opts{kdc_AD},
            user             => $opts{user},
            group_name       => $opts{group_name},
            number_of_groups => $opts{number_of_groups},
        );
    } elsif ($opts{client_object}->type() =~ /unix-solaris/) {

        #solaris client
        croak('Currently no support for Solaris clients');
    } else {
        croak("The object's hosttype passed into the setup_client_yp_conf subroutine is not supported: " . $opts{client_object}->type());
    }
    return_client();
    $Log->exit();

    return;
} ## end sub Aux_gid_client_setup(%)

#setup the yp config in a Linux/Solaris client
sub _setup_linux_client_yp_conf(%) {
    my %opts = validate(
        @_,
        {   client_object    => {type => OBJECT},
            kdc_AD           => {type => HASHREF},
            user             => {type => SCALAR},
            group_name       => {type => SCALAR},
            number_of_groups => {type => SCALAR},
        }
    );

    #Creating an alias
    my $client  = $opts{client_object};
    my $primary = 1;

    # check if the config files are in the standard primary location
    # of /etc/yp.conf
    my $yp_conf_output = client_execute("ls -l $::YP_CONF_PATH");

    if (   $yp_conf_output =~ /No such file or directory/
        && $yp_conf_output !~ /$::YP_CONF_PATH/) {

        # this is an unlikely scenario for linux client but still this
        # check is necessary
        $yp_conf_output = client_execute("ls -l $::SECOND_YP_PATH");
        if ($yp_conf_output =~ /No such file or directory/) {

            # fail if the yp file is not in any of the locations
            $Log->comment("Standard locations: $::YP_CONF_PATH and $::SECOND_YP_PATH");
            croak("Unable to locate yp config files at the standard locations");
        }
        $primary = 0;
    }

    if ($primary) {    # yp.conf is present in /etc/yp.conf
        _configure_primary_yp_location(
            client_object    => $opts{client_object},
            kdc_AD           => $opts{kdc_AD},
            user             => $opts{user},
            group_name       => $opts{group_name},
            number_of_groups => $opts{number_of_groups},
        );
    } else {           # unlikely secondary location
        croak("Currently library has no support for yp conf location in any other directory");
    }

    return;
} ## end sub _setup_linux_client_yp_conf(%)

#Configure the /etc/yp.conf file and restart the ypbind.
sub _configure_primary_yp_location(%) {
    my %opts = validate(
        @_,
        {   client_object    => {type => OBJECT},
            kdc_AD           => {type => HASHREF},
            user             => {type => SCALAR},
            group_name       => {type => SCALAR},
            number_of_groups => {type => SCALAR},
        }
    );

    # This temporrary location will contain the new yp.conf file
    my $temp_yp_location = $::YP_CONF_PATH . '_';

    # obtain the domain name of the AD/NIS
    my $domain = _get_domain_name($opts{kdc_AD}->{kdc_domain});

    client_execute(qq{\\grep -v '$opts{kdc_AD}->{kdc_server}\\|^\$' $::YP_CONF_PATH > $temp_yp_location });

    $previous_yp_domain = client_execute(qq{\\domainname});

    $Log->comment('Add domain entry into yp.conf file');

    # create the yp.conf entry and add it into yp.conf
    my $yp_domain = 'domain ' . $domain . ' server ' . $opts{kdc_AD}->{kdc_server} . '.' . $opts{kdc_AD}->{kdc_domain};
    client_execute("\\mv $temp_yp_location $::YP_CONF_PATH");

    # add the entry into the yp.conf
    client_execute("\\echo $yp_domain >> $::YP_CONF_PATH");
    etc_hosts_modify(kdc_AD => $opts{kdc_AD},);

    # change the yp domain on the client to the AD domain
    client_execute(qq{\\domainname $domain});

    # stop the ypbind service and start it again
    # if ypbind is not present in the location specified
    # then fail the script
    $Log->comment('Check the location of ypbind executable');
    my $ypbind_location = client_execute("\\ls -l $::YPBIND_PATH");
    if (   $ypbind_location =~ /No such file or directory/
        || $ypbind_location !~ /$::YPBIND_PATH/) {
        $Log->comment("ypbind was not located in $::YPBIND_PATH");
        croak('Please pass a proper ypbind path to the script');
    }

    _restart_and_verify_ypbind($opts{client_object});

    return;
} ## end sub _configure_primary_yp_location(%)

# Obtain the domainname from the FQN
sub _get_domain_name($) {
    my $domain_name = shift;
    if ($domain_name eq '') {
        croak('No domain name was passed into the module _get_domain_name');
    }

    my @temp_domain = split('\.', $domain_name);
    return $temp_domain[0];
} ## end sub _get_domain_name($)

=head1 METHODS

=head2 clear_client_config

This subroutine changes the domainname back to the previous domainname entry
ane then restarts the ypbind domain name

clear_client_config( client_object => $<NACL-Object> , vserver => $<Vserver-name> , kdc_AD => \%<KDC-Details> );

=over

=item Arguments

=over

=item kdc_AD

(Required) : Reference to the KDC hash which contains various information as mentioned in the design document
             http://wikid.netapp.com/w/NOS_Automation/NFS/Kerberos_Library_Design_Document#AD_Setup

=item client_object

(Required) : NACL::C::Client object needs to be passed into the sub routine.

=item vserver

(Required) : Name of the vserver present on the filer.

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub clear_client_config(%) {
    my %opts = validate(
        @_,
        {   client_object => {type => OBJECT},
            vserver       => {type => SCALAR},
            kdc_AD        => {type => HASHREF}, 
        }
    );

    # restore the domain name back to its previous value
    check_client_and_sudo($opts{client_object});
    
    # Delete the script user from AD
    if( $script_user ne '' ){
        _delete_script_user( $opts{kdc_AD} );
    }

    if ($previous_ns_switch ne '') {
        _modify_ns_switch(
            ns_switch => $previous_ns_switch,
            vserver   => $opts{vserver}
        );
    }

    if ($previous_yp_domain ne '') {
        chomp($previous_yp_domain);
        $Client_cmd->execute(command => "domainname $previous_yp_domain");

        # restart ypbind
        _restart_and_verify_ypbind($opts{client_object});
    }
    return_client();

    return;
} ## end sub clear_client_config(%)

# Change the ns_switch on the filer to either NIS or LDAP
# or revert it back to your previous setting.

sub _modify_ns_switch($) {
    my %opts = validate(
        @_,
        {   ns_switch => {type => ARRAYREF},
            vserver   => {type => SCALAR},
        }
    );

    NACL::C::Vserver->modify(
        command_interface => $Node_cmd,
        vserver           => $opts{vserver},
        'ns-switch'       => $opts{ns_switch},
    );

    return;
} ## end sub _modify_ns_switch($)

sub _delete_script_user($){
    my $kdc = shift;
    
    # Establish the NACL::C::Client object to the KDC server
    my $KDC_Connection_Obj = establish_kdc_connection($kdc->{kdc_server});

    # Making sure the server handle retrieved is a windows client.
    if (!$KDC_Connection_Obj->is_windows()) {
        croak("KDC_Server_Setup : Incorrect or invalid client is detected for the KDC server: ".$kdc->{ kdc_server });
    }

    my $DC = kerberos_create_dc(kdc_realm => $kdc->{kdc_realm});
    my $base_dn = "$kdc->{ kdc_ou },$DC";
    my $user_dn = "cn=$script_user,$base_dn";
    my $AD_user_info   = $KDC_Connection_Obj->execute(command => "dsget user \"$user_dn\"");

    #verify if the dsget user succeeded, if so delete the user
    if ($AD_user_info =~ /dsget succeeded/) {
        $KDC_Connection_Obj->execute(command => "dsrm -noprompt \"$user_dn\"");
        $AD_user_info   = $KDC_Connection_Obj->execute(command => "dsget user \"$user_dn\"");

        if( $AD_user_info =~ /dsget failed:Directory object not found/){
            $Log->comment("Script user $script_user was successfully removed from the AD");
        } else {
            $Log->warn("Unable to delete Script user $script_user from the AD: $AD_user_info");
        }
    }

    return;
}

=head1 METHODS

=head2 user_login_and_verify_groups

This subroutine verifies if the user is part of the requested number of groups.

user_login_and_verify_groups( user => $<Username> , client_object => $<NACL-Object> ,
                      group_name => $<group-prefix-name> , number_of_groups => $<num-of-groups> );

=over

=item Arguments

=over

=item user

(Required) : Name of the user who has is present in the AD (if not present)

=item group_name

(Required) : Prefix of the group name 

=item number_of_groups

(Required) : Total number of groups that has to be verified in the AD.

=item client_object

(Required) : NACL::C::Client object needs to be passed into the sub routine.

=back

=item Return

Nothing is returned on success or croaks on failure

=back

=cut

sub user_login_and_verify_groups(%) {
    my %opts = validate(
        @_,
        {   user             => {type => SCALAR},
            group_name       => {type => SCALAR},
            number_of_groups => {type => SCALAR},
            client_object    => {type => OBJECT},
        }
    );

    #verify if the user was added successfully into the groups
    check_client_and_sudo($opts{client_object});
    my $check_groups = client_execute("\\id $opts{user}");
    my $i            = 1;
    for (; $i <= $opts{number_of_groups}; $i++) {
        my $group = $opts{group_name} . $i;
        if ($check_groups !~ qr/$group/) {
            croak("Unable to add user: $opts{user} into the group: $group");
        }
    }

    $Log->comment("User: $opts{user} has been successfully added into $opts{number_of_groups}");
    return_client();

    return;
} ## end sub user_login_and_verify_groups(%)

# Stop and start the ypdomain. Verify the ypbind status
# if stop or start is successful.

sub _restart_and_verify_ypbind($) {
    my $client = shift;

    # stop and start the ypbind service
    $Log->comment('Stop the ypbind daemon process');
    my $ypbind_stop   = $client->execute(command => "$::YPBIND_PATH stop");
    my $ypbind_status = $client->execute(command => "$::YPBIND_PATH status");
    if ($ypbind_status !~ qr/ypbind is stopped/) {
        croak("Unable to stop the ypbind process: $ypbind_status");
    }

    $Log->comment('Start the ypbind process');
    $client->execute(command => "$::YPBIND_PATH start");
    $ypbind_status = $client->execute(command => "$::YPBIND_PATH status");
    if ($ypbind_status !~ qr/ypbind .+ is running/) {
        croak("Unable to stop the ypbind process: $ypbind_status");
    }

    return;
} ## end sub _restart_and_verify_ypbind($)

#modify the /etc/hosts to add the AD entry
sub etc_hosts_modify(%) {
    my %opts = validate(@_, {kdc_AD => {type => HASHREF},});

    my $filer_fqn_short_name = get_fqn('junk', $opts{kdc_AD}->{kdc_ip}, 0);
    my $client_etc_hosts = client_execute('cat /etc/hosts');

    # Add the kdc details into /etc/hosts
    if ($client_etc_hosts =~ /$opts{kdc_AD}->{kdc_ip}/) {
        eval {
            client_execute(qq{\\grep -v '$opts{kdc_AD}->{kdc_ip}\\|^\$' /etc/hosts > /etc/hosts_});
            client_execute(qq{\\mv /etc/hosts_ /etc/hosts});
        };
        if ($@) {
            $Log->warn(qq{Ignoring a possible error from grep, if there is no matching line present in the /etc/hosts});
        }
    }
    my $command = "echo '$opts{kdc_AD}->{kdc_ip} $opts{kdc_AD}->{kdc_fqn} $opts{kdc_AD}->{kdc_server}' >> /etc/hosts";
    client_execute($command);
    client_execute('cat /etc/hosts');

    return;
} ## end sub etc_hosts_modify(%)

#}])
1;    # Do not remove this line
