# $Id: //depot/prod/DOT/dev/test/lib/nfs/CIT/Export.pm#1 $

#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Export Module
## @author dl-svl-nosauto@netapp.com
## @status shared
## @pod here

=head1 NAME

nfs::CIT::Export

=head1 DESCRIPTION

This module contains common functions related to Export.

=cut

package nfs::CIT::Export;

use nfs::CommonParams;
use NACL::APISet;
use NACL::STask::VserverServicesUnixUser;
use NACL::C::ExportPolicyRule;
use NATE::BaseException;
use nfs_setup_gx;
use Tharn qw($Log);
use Options;
use Exporter 'import';

use strict;
use warnings;

our @EXPORT_OK = qw(get_vserver_id
                    flush_access_caches
                    set_export_policy
                    open_up_export_policy);


##############################################################################
# Name          : get_vserver_id
# Description   : Find the vserver id for the vserver from the smdb table
# Input         : $vserver_obj - vserver object whose ID is to be found
# Returns       : Integer: vserver ID
#                 0 - Error (Not found)
##############################################################################
sub get_vserver_id {
    my ($vserver_obj) = @_;
    my $node_obj = $vserver_obj->command_interface();

    $Log->comment('Getting the vserver id for '.$vserver_obj->vserver());

    my $vserver_id = 0;
    my $response
        = $node_obj->apiset->debug_smdb_table_show(table => 'vserver_by_name');
    my $parsed_output = $response->get_parsed_output();

    foreach my $row (@{$parsed_output}) {
        my $vserver_row = $row->{'vserver'};
        if ($vserver_row eq $vserver_obj->vserver()) {
            $vserver_id = $row->{'id-unsigned'};
            last;
        }
    }

    $Log->comment('vserver_id for '.$vserver_obj->vserver()." is $vserver_id");

    return $vserver_id;
}


##############################################################################
# Name          : flush_access_caches
# Description   : Flush all of the relevant access caches
# Input         : $vserver_obj - Vserver object that we are running on
#                 $vserver_id  - Vserver ID that we are running on
#                 $rule_index  - What rule inside the policy?
# Returns       : 1 on success
#                 NACL exception on failure
##############################################################################
sub flush_access_caches {
    my ($vserver_obj, $vserver_id, $rule_index) = @_;
    my $node_obj = $vserver_obj->command_interface();

    $Log->comment('Flushing caches for '.$vserver_obj->vserver().
                  " - id $vserver_id");

    #Prep for clearing the mhost caches:
    my $user_obj = NACL::STask::VserverServicesUnixUser->find_or_create(
                                        command_interface => $node_obj,
                                        user              => 'hastings',
                                        vserver     => $vserver_obj->vserver(),
                                        create_specific_options => {
                                            'primary-gid' => 1066,
                                            id            => 1066,
                                            'full-name'   => '"Mr. Battle of"',
                                        });

    #Now clear the Nblade caches:
    my $system_apiset_obj = NACL::APISet->new(hostobj   => $node_obj->hostrec(),
                                              category  => 'Node',
                                              interface => 'CLI',
                                              set       => 'Systemshell');
    $system_apiset_obj->set_timeout('timeout' => 300);

    my $cmd = "nitrocfg 0 invalidateAccessCache $vserver_id $rule_index";
    my $response = $system_apiset_obj->execute_raw_command(command => $cmd);
    $Log->comment("nitrocfg response is $response");

    #Now really clear the mhost caches:
    $user_obj->delete();


    return 1;
}


##############################################################################
# Name          : set_export_policy
# Description   : Deletes policy with the parameters provided, if exists
#                 Creates policy with the parameters provided
# Input         : Hash with following keys:
#                 command_interface - What node are we running on?
#                 policyname - What policy are we using?
#                 clientmatch - What is the clientmatch?
#                 ro_rule - What is the ROrule?
#                 rw_rule - What is the RWrule?
#                 protocol - What is the access protocol?
#                 vserver - What vserver are we running on?
#                 anon - Who is the anon user?
#                 superuser - Who is the super user?
#                 ruleindex - What rule inside the policy?
#                 $allow_suid - Do we allow setuid? (true/false)
#                 $allow_dev - Do we allow devices to be created? (true/false)
# Returns       : True on success
#                 NACL Exception on failure
##############################################################################
sub set_export_policy {
    my (%params) = @_;

    my $policy_rule_obj = NACL::C::ExportPolicyRule->find(
                               command_interface => $params{command_interface},
                               filter => {
                                          policyname => $params{policyname},
                                          vserver    => $params{vserver},
                                          ruleindex  => $params{ruleindex},
                                         },
                               allow_empty => 1,
                          );
    if ($policy_rule_obj) {
        $policy_rule_obj->delete();
    }

    if ($params{anon} eq q{-}) {
        $params{anon} = $::EXPORT_ANON;
    }

    NACL::C::ExportPolicyRule->create(%params);

    return 1;
}


##############################################################################
# Name          : open_up_export_policy
# Description   : Opens up the export policy with following
#                 clientmatch - '0.0.0.0/0'
#                 ro_rule - 'any'
#                 rw_rule - 'any'
#                 protocol - 'any'
#                 anon - 0
#                 superuser - 'any'
#                 $allow_suid - true
#                 $allow_dev - true
# Input         : Hash with following keys:
#                 command_interface - What node are we running on?
#                 policyname - What policy are we using?
#                 clientmatch - What is the clientmatch?
#                 vserver     - What is the vserver?
# Returns       : True on success
#                 NACL Exception on failure
##############################################################################
sub open_up_export_policy {
    my (%params) = @_;

    my $client_match =
      nfs_setup_gx::correct_address_family() eq 'ipv4' ? '0.0.0.0/0' : '::0/0';

    set_export_policy(clientmatch     => $client_match,
                      rorule          => ['any'],
                      rwrule          => ['any'],
                      protocol        => ['any'],
                      anon            => 0,
                      superuser       => ['any'],
                      'allow-suid'      => 'true',
                      'allow-dev'       => 'true',
                      %params);

    return 1;
}


1;

