package nfs::CIT::params;

## @summary
##  Parameters used in cit-nfs tests.
##
##  WARNING:
##  Adding new parameters is fine as long as no other parameters are used the same way.
##  Avoid changing any existing parameters since it may change behaviours of other scripts.

use Params::Common;
use Options;

my $Options = new Options;
$Options->parameters(
    definitions => [
        {   name => "RESETCLUSTER",
            description =>
                "Reset the cluster, this will delete all volumes and vservers",
            spec    => "boolean",
            default => 0,
        },
        {   name         => "NFS_VSERVERS",
            description  => "List of vservers to act on",
            spec         => "s@",
            default      => ["vs1", "vs2"],
            valid_values => ["vs1", "vs2"],
        },
        {   name          => "AGGR_CREATE_WAIT",
            description   => "Total wait time in seconds of create_aggr test",
            spec          => "s",
            default       => 7200,
            greater_equal => 0,
        },
        {   name => "AGGR_CHECK_INTERVAL",
            description =>
                "Interval in seconds between each Aggregate state check",
            spec          => "s",
            default       => 30,
            greater_equal => 5,
        },
        {   name          => "TIMEOUT",
            description   => "Set the command timeout value",
            spec          => "s",
            default       => 600,
            greater_equal => 0,
        },
        {   name        => "VSERVER",
            description => "Vserver name",
            spec        => "s",
            default     => "vs0",
        },
        {   name        => "SECURITY_STYLE",
            description => "Security style for Vserver and volume",
            spec        => "s",
            default     => "mixed",
        },
        {   name        => "DATA_LIF_PROTOCOL",
            description => "data-protocol for the data lif",
            spec        => "s@",
            default     => ['nfs', 'cifs', 'fcache'],
        },
        {   name         => "NFS_PROTOCOL",
            description  => "List of NFS protocols to enable",
            spec         => "s@",
            default      => ["v3", "v4.0", "v4.1", "v4.1-pnfs"],
            valid_values => ["v3", "v4.0", "v4.1", "v4.1-pnfs"],
        },
        {   name        => "FLEXVOL_VSERVERS",
            description => "List of flexvol vservers",
            spec        => "s@",
            default     => [],
        },
        {   name        => "FLEXCACHE_VSERVERS",
            description => "List of flexcache vservers",
            spec        => "s@",
            default     => [],
        },
        {   name         => "V4_ACL",
            description  => "NFSv4.0 ACL Support",
            spec         => "s",
            default      => "enabled",
            valid_values => ["enabled", "disabled"],
        },
        {   name         => "V4_READ_DELEGATION",
            description  => "NFSv4.0 Read Delegation Support",
            spec         => "s",
            default      => "disabled",
            valid_values => ["enabled", "disabled"],
        },
        {   name         => "V4_WRITE_DELEGATION",
            description  => "NFSv4.0 Write Delegation Support",
            spec         => "s",
            default      => "disabled",
            valid_values => ["enabled", "disabled"],
        },
        {   name        => "V4_ID_DOMAIN",
            description => "NFSv4 ID Mapping Domain",
            spec        => "s",
            default     => "netapp.com",
        },
        {   name         => "V4_1_ACL",
            description  => "NFSv4.1 ACL Support",
            spec         => "s",
            default      => "enabled",
            valid_values => ["enabled", "disabled"],
        },
        {   name        => "FEATURES",
            description => "List of Features to add to the cluster",
            spec        => "s@",
            default     => ["NFS", "CIFS", "Snapmirror_DP", "SnapRestore"],
        },
        {   name        => "UNIX_PERMISSIONS",
            description => "unix premissions for vserver and infinvol",
            spec        => "s",
            default     => "---rwxrwxrwx",
        },
        {   name        => "LOCAL_UNIX_USERS",
            description => "Local unix user",
            spec        => "s",
            default =>
                'root:0:0:"Local Root",nobody:65534:65534:nobody,smoke:18811:30:smoke',
        },
        {   name        => "LOCAL_UNIX_GROUPS",
            description => "Local unix group",
            spec        => "s",
            default     => 'root:0,nobody:65534,engr:30',
        },
        {   name => "DNS_SERVER_IP",
            description =>
                "DNS Server IP, this will override the one provide by the site file",
            spec => "s@",
        },
        {   name => "DNS_DOMAIN",
            description =>
                "DNS Server DOMAIN, this will override the one provide by the site file",
            spec => "s@",
        },
        {   name          => "AGGR_DISK_COUNT",
            description   => "Number of disk per aggregate",
            spec          => "s",
            default       => 10,
            greater_equal => 8,
        },
        {   name        => "AGGR_PREFIX",
            description => "prefix for aggregate name",
            spec        => "s",
            default     => "aggr",
        },
        {   name         => "AGGR_RAID_TYPE",
            description  => "Aggregate raid type, can be raid4 or raid_dp",
            spec         => "s",
            default      => "raid4",
            valid_values => ["raid_dp", "raid4"],
        },
        {   name        => "REPLACE_NA_WITH_PASS",
            description => "Force NA test to be PASS test",
            spec        => "boolean",
            default     => 0,
        },
        {   name        => "DEBUG_LOGGING",
            description => "Enable changes to the logging level",
            spec        => "boolean",
            default     => 1,
        },
        {   name         => "LOGGING_LEVEL",
            description  => "The loging level for mgwd log and job log",
            spec         => "s",
            default      => "debug",
            valid_values => [
                "emerg",   "alert",  "crit", "err",
                "warning", "notice", "info", "debug"
            ],
        },
        {   name => 'NIS_DOMAIN',
            description =>
                'NIS Domain name, if not provide, will try to get it from site info files',
            spec    => 's',
            default => 'netapp.com',
        },
        {   name => 'NIS_SERVERS',
            description =>
                'NIS Domain Servers IP, if not provide, will try to get it from site info files',
            spec    => 's@',
            default => ['10.60.132.113'],
        },
        {   name        => "NTP_SERVER_IP",
            description => "NTP Server IP address",
            spec        => "s@",
            default     => [qw(10.98.203.74 10.98.18.43 10.98.13.43)],
        },
        {   name => "AD_SERVER_USER",
            description =>
                "AD server admin name, needed for remove cifs vserver",
            spec    => "s",
            default => "adadmin",
        },
        {   name => "AD_SERVER_PASSWORD",
            description =>
                "AD server admin password, needed for remove cifs vserver",
            spec    => "s",
            default => "adadmin",
        },
        {   name        => "CIFS_DOMAIN",
            description => "Domain name for CIFS server",
            spec        => "s",
            default     => "CNC",
        },
        {   name        => "CIFS_DOMAIN_LONG",
            description => "Domain name for CIFS server",
            spec        => "s",
            default     => "CNC.SPIN.ENG.NETAPP.COM",
        },
        {   name => "AD_SERVER_IP",
            description =>
                "List of IP addresss of the Active Directory servers",
            spec    => "s@",
            default => ["10.98.203.74", "10.98.202.208"],
        },
        {   name        => "CIFS_SERVER",
            description => "CIFS server name",
            spec        => "s",
        },
        {   name        => "CIFS_SHARE",
            description => "CIFS SHARE name",
            spec        => "s",
            default     => "cifs_share",
        },
        {   name        => "CIFS_SUPERUSERS",
            description => "CIFS Super User",
            spec        => "s@",
            default     => [],
        },
        {   name        => "CREATE_CIFS",
            description => "Create CIFS",
            spec        => "boolean",
            default     => 0,
        },
        {   name        => "DATA_PATH",
            description => "Path to be used for tests for mount operation",
            spec        => "s",
            default     => "/data",
        },
        {   name        => "DATA_POLICY_NAME",
            description => "Export policy name for data mounts",
            spec        => "s",
            default     => "data",
        },
        {   name        => "DATA_RULE_INDEX",
            description => "Export policy rule index for data mounts",
            spec        => "s",
            default     => "10",
        },
    ]
);
$Options->process();

1;
