# -*- perl -*-
#
# Copyright (c) 2013 Network Appliance, Inc.
# All Rights reserved
#
## @summary
##  Script will setup reference NFS configuration for cit-nfs
##
## @description
##  This script will setup 2 vservers for NFS testing for cit-nfs.
##  The result should match that as shown in
##   http://wikid.netapp.com/w/NFS/cit-nfs
##  This may require undoing the vs0 created by deux_init.
##
## @synopsis
##  ntest
##      NODES=thomas-vsim1,thomas-vsim2
##      RESETCLUSTER=1
##      FLEXVOL_VSERVERS=vs_vol,vs_cache
##      nfs/CIT/setup.thpl
##
## @param
##  NODES
##  Required: Your Filer name (get from tharn_hosts)
##
##  RESETCLUSTER
##  Optional: Delete all Vservers, Volumes, and data interfaces
##            in the cluster.
##  Default:  0
##
## @dependencies
##  tharn host must contain DATA_IP, DATA_MASK, DATA_PORT, DATA_GATEWAY
##
## @type
##  test (nate)
##
## @author
##  thomas@netapp.com
##

#---------------------------------------------------------------------
# Initialize and validate parameters
#---------------------------------------------------------------------

package nfs::CIT::setup;

package main;
use nfs::CIT::params;
use strict;
use warnings;
no warnings 'redefine';

# Module import.
use TCD;
use Params::Validate qw(validate_with UNDEF HASHREF SCALAR HASHREF);

use NACL::STask::SystemLicense;
use NACL::C::Node;
use NATE::BaseException qw(:try);
use NATE::Error;

use NACL::C::Node;
use NACL::C::SystemNode;
use NACL::Exceptions::NoElementsFound qw{:try};

use NACL::C::Aggregate;
use NACL::C::NetworkInterface;
use NACL::C::Vserver;
use NACL::C::Volume;
use NACL::C::VolumeFlexcache;
use NACL::C::VolumeFlexcacheCachePolicy;
use NACL::C::VserverNfs;
use NACL::C::ExportPolicyRule;
use NACL::C::ExportPolicy;

use NACL::C::VserverServicesUnixUser;
use NACL::C::VserverServicesUnixGroup;
use NACL::C::VserverServicesNisDomain;
use NACL::C::VserverNameMapping;
use NACL::C::Interface;
use NACL::C::Route;
use NACL::C::VserverServicesDns;
use NACL::C::Disk;
use NACL::C::JobLog;
use NACL::C::Job;
use NACL::C::SecurityLogin;
use NACL::APISet::Exceptions::TimeoutException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::Exceptions::HostUnreachable;
use NACL::Exceptions::NoElementsFound;
use NACL::C::Network;
use NACL::C::SystemServicesNtpServer;
use NACL::STask::Volume;
use Digest::SHA1 qw(sha1_hex);

use TestUtils;

# Global Variables.

my $aggr_base   = "aggr_";
my $volume_base = "data_";

use vars
    qw($Options $Node $Apiset %SiteData $Site $NA $Vsadmim_name @Vserver_Name_List);

#################
# Name: UseCommonSetup
# Description: Prep for all functions
# Input:
# Output:
#################
sub UseCommonSetup {
    $Options = new Options;

    $Options->parameters(
        definitions => [
            {   name        => 'NFS_VOL_SIZE',
                description => 'Volume size',
                default     => '1G',
            },
        ]
    );

    $Options->process();
    $Node   = NACL::C::Node->find();
    $Apiset = $Node->apiset(
        category  => "Node",
        interface => "CLI",
        set       => "CMode",
    );

    $NA
        = $Options->get(option => 'REPLACE_NA_WITH_PASS')
        ? $TCD::PASS
        : $TCD::NA;
    %SiteData = get_site_info();

# Find out the site, default it to svl if it's not define and can't be set from tharnhost
    if ($Options->defined(option => 'SITE')) {
        $Site = $Options->get(option => 'SITE');
    } else {
        $Site = $Node->hostrec()->SITE() ? $Node->hostrec()->SITE() : 'svl';
    }
} ## end sub UseCommonSetup

#################
# Name: configure_everything
# Description: Configure the two vservers for cit-nfs
# Input:
# Output: Pass or failure indicator
#################
sub configure_everything {
    $Test->description("Configure the two vservers for cit-nfs");

    my @cmdinfs = NACL::C::Node->find();
    if ($#cmdinfs != 1) {
        $Test->nlog('There need to be two nodes!');
        return $TCD::FAIL;
    }

    my $node1 = $cmdinfs[0];
    my $node2 = $cmdinfs[1];

    my $nodename1 = $node1->{node};

    my $nodename2 = $node2->{node};

    my @flexVol_vservers   = @{$Options->get(option => "FLEXVOL_VSERVERS")};
    my @flexCache_vservers = @{$Options->get(option => "FLEXCACHE_VSERVERS")};

    my $i = 0;
    $Test->nlog("flexVol vservers");
    foreach my $key (@flexVol_vservers) {
        $Test->nlog("\t" . $i++ . " " . $key);
    }

    $i = 0;
    $Test->nlog("flexCache vservers");
    foreach my $key (@flexCache_vservers) {
        $Test->nlog("\t" . $i++ . " " . $key);
    }

    my $num_vservers = @flexVol_vservers + @flexCache_vservers;
    if ($num_vservers == 0) {
        $Test->nlog('No vservers defined!');
        return $TCD::FAIL;
    } elsif ($num_vservers > 2) {
        $Test->nlog('Cannot define more than two vservers!');
        return $TCD::FAIL;
    } elsif ($num_vservers == 1) {
        $Test->nlog(
            'One vserver is not enough, need to define two vservers!');
        return $TCD::FAIL;
    }

    my ($vserver1, $vserver2) = (@flexVol_vservers, @flexCache_vservers);
    my $vserver1_has_cache;
    my $vserver2_has_cache;
    if (defined $vserver1) {
        $vserver1_has_cache = grep { $_ eq $vserver1 } @flexCache_vservers;
        my $flexType = $vserver1_has_cache ? "FlexCache" : "FlexVol";
        $Test->nlog("$vserver1 is a $flexType");
    } else {
        $Test->nlog('Failed to define first vserver!');
        return $TCD::FAIL;
    }

    if (defined $vserver2) {
        $vserver2_has_cache = grep { $_ eq $vserver2 } @flexCache_vservers;
        my $flexType = $vserver2_has_cache ? "FlexCache" : "FlexVol";
        $Test->nlog("$vserver2 is a $flexType");
    } else {
        $Test->nlog('Failed to define second vserver!');
        return $TCD::FAIL;
    }

    $Test->step("Setting up: $vserver1 and $vserver2");

    configure_aggregate($node1, $nodename1);
    configure_aggregate($node2, $nodename2);

    # Create vserver with root on the other node
    configure_vserver($node1, $vserver1, $nodename1, $vserver1_has_cache);
    configure_vserver($node2, $vserver2, $nodename2, $vserver2_has_cache);

    configure_data_interface($node1, $vserver1, 1, "_d1_v1");
    configure_data_interface($node1, $vserver2, 2, "_d2_v2");
    configure_data_interface($node2, $vserver1, 2, "_d2_v1");
    configure_data_interface($node2, $vserver2, 1, "_d1_v2");

    # Create data_vs2 on vsim1 and aggr_1
    configure_data_volume($node1, $nodename1, $vserver1, $vserver1_has_cache);

    # Create data_vs1 on vsim2 and aggr_2
    configure_data_volume($node2, $nodename2, $vserver2, $vserver2_has_cache);

    # Now configure NFS on the vservers
    configure_nfs($node1, $vserver1);
    configure_nfs($node2, $vserver2);

    # Configure the exports
    configure_export_rule($node1, $vserver1);
    configure_export_rule($node2, $vserver2);

    # Configure the NIS
    configure_nis($node1, $vserver1);
    configure_nis($node2, $vserver2);

    # Configure the DNS
    configure_dns($node1, $vserver1);
    configure_dns($node2, $vserver2);

    add_ntp_servers($node1);
    add_ntp_servers($node2);

    return $TCD::PASS;
} ## end sub configure_everything

#################
# Name: configure_vserver
# Description: Configure a vserver
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
#   nodename - The name of the node
#   has_cache - Uses FlexCache?
# Output: Pass or failure indicator
#################
sub configure_vserver {
    $Test->description("Configuring a vserver");

    my ($node, $vserver, $nodename, $has_cache)
        = (shift, shift, shift, shift);

    $Test->nlog("Configuring vserver $vserver on node $nodename");

    my @ns_vals = ("file", "nis");
    my @nm_vals = ("file");

    my $vs = NACL::C::Vserver->create(
        command_interface => $node,
        "vserver"         => $vserver,
        "rootvolume"      => "root_" . $vserver,
        "aggregate"       => $aggr_base . _normalize($nodename),
        "rootvolume-security-style" =>
            $Options->get(option => 'SECURITY_STYLE'),
        "ns-switch" => \@ns_vals,
        "nm-switch" => \@nm_vals,
        'rootvolume-security-style' =>
            $Options->get(option => 'SECURITY_STYLE'),
        'method-timeout' => $Options->get(option => 'TIMEOUT'),
    );

    $Test->step("Modifying vserver root volume permissions");
    NACL::STask::Volume->modify(
        command_interface  => $node,
        'unix-permissions' => $Options->get(option => 'UNIX_PERMISSIONS'),
        'volume'           => "root_" . $vserver,
        'vserver'          => $vserver,
        "size"             => "100M",
    );

    my $reason;
    if (!$node->is_feature_supported(feature => "flexcache", reason_ref => \$reason)) {
        $has_cache = 0;
        $Test->log(qq{This node does not support flexcache due to: $reason});
    }

    if ($has_cache) {
        $Test->nlog("Setting cache policy to prefer local cache");

        NACL::C::VolumeFlexcacheCachePolicy->modify(
            command_interface    => $node,
            'vserver'            => $vserver,
            'policy'             => "default",
            'prefer-local-cache' => "true",
        );

        $Test->nlog("Creating caches for vserver root");

        NACL::C::VolumeFlexcache->create(
            command_interface => $node,
            'vserver'         => $vserver,
            'origin-volume'   => "root_" . $vserver,
        );
    }

    return $TCD::PASS;
} ## end sub configure_vserver

#################
# Name:
# Description: Create a data volume on the vserver for the vsim
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
#   nodename - The name of the node
#   has_cache - Uses FlexCache?
# Output: Pass or failure indicator
#################
sub configure_data_volume {
    $Test->description("Creating a FlexVol for data");

    my ($node, $nodename, $vserver, $has_cache)
        = (shift, shift, shift, shift);
    my $volume = $volume_base . $vserver;

    $Test->nlog(
        "Creating volume $volume on aggr $aggr_base$nodename for vserver $vserver"
    );

    my $vol = NACL::C::Volume->create(
        command_interface => $node,
        aggregate         => $aggr_base . _normalize($nodename),
        volume            => $volume,
        vserver           => $vserver,
        "size"            => $Options->get(option => 'NFS_VOL_SIZE')
    );

    # This is the workaround for 751024.
    my $retries = 5;
    while ($retries) {
        try {
            $vol->mount("junction-path"   => $Options->get(option => 'DATA_PATH'));
            $retries = 0;
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            my $excp = shift;
            if ($retries) {
                $retries--;
                sleep 5;
            } else {
                $excp->throw();
            }
        }
    }

    my $reason;
    if (!$node->is_feature_supported(feature => "flexcache", reason_ref => \$reason)) {
        $has_cache = 0;
        $Test->log(qq{This node does not support flexcache due to: $reason});
    }

    if ($has_cache) {
        $Test->nlog("Creating caches for volume $volume");

        NACL::C::VolumeFlexcache->create(
            command_interface => $node,
            'vserver'         => $vserver,
            'origin-volume'   => $volume
        );
    }

    return $TCD::PASS;
} ## end sub configure_data_volume

#################
# Name: _normalize
# Description: Make sure the string is normalized
#   I.e., a goood node name.
# Input: String to normalize
# Output: Normalized version of string
#################
sub _normalize {
    my $string = shift();

    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    $string =~ s/-/_/g;

    return $string;
} ## end sub _normalize

#################
# Name: configure_export_rule
# Description:
#   Configure the default export policy for a vserver.
#   Configure the "data" export policy for a vserver.
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
# Output: Pass or failure indicator
#################
sub configure_export_rule {
    $Test->description("Create default export rules");

    my ($node, $vserver) = (shift, shift);

    NACL::C::ExportPolicyRule->create(
        command_interface => $node,
        'policyname'      => 'default',
        'clientmatch'     => '0.0.0.0/0',
        'rorule'          => ['any'],
        'rwrule'          => ['any'],
        'protocol'        => ['any'],
        'vserver'         => $vserver,
        'anon'            => '0',
        'superuser'       => ['any'],
        'ruleindex'       => '1',
        'allow-suid'      => 'true',
        'allow-dev'       => 'true',
    );

    NACL::C::ExportPolicy->create(
        command_interface => $node,
        'policyname'      => $Options->get(option => 'DATA_POLICY_NAME'),
        'vserver'         => $vserver,
    );

    NACL::C::ExportPolicyRule->create(
        command_interface => $node,
        'policyname'      => $Options->get(option => 'DATA_POLICY_NAME'),
        'clientmatch'     => '0.0.0.0/0',
        'rorule'          => ['any'],
        'rwrule'          => ['any'],
        'protocol'        => ['any'],
        'vserver'         => $vserver,
        'anon'            => '0',
        'superuser'       => ['any'],
        'ruleindex'       => $Options->get(option => 'DATA_RULE_INDEX'),
        'allow-suid'      => 'true',
        'allow-dev'       => 'true',
    );

    return $TCD::PASS;
} ## end sub configure_export_rule

#################
# Name: configure_nfs
# Description: Create NFS Services
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
# Output: Pass or failure indicator
#################
sub configure_nfs {
    $Test->description("Create NFS Services");

    my ($node, $vserver) = (shift, shift);

    my $default_protocols = {
        'v3'        => 'disabled',
        'v4.0'      => 'disabled',
        'v4.1'      => 'disabled',
        'v4.1-pnfs' => 'disabled',
    };

    my $default_options = ();

    my @required_protocols = @{$Options->get(option => "NFS_PROTOCOL")};

    $default_options->{'v4.0-acl'} = $Options->get(option => "V4_ACL")
        if ($Options->get(option => "V4_ACL"));
    $default_options->{'v4.0-read-delegation'}
        = $Options->get(option => "V4_READ_DELEGATION")
        if ($Options->get(option => "V4_READ_DELEGATION"));
    $default_options->{'v4.0-write-delegation'}
        = $Options->get(option => "V4_WRITE_DELEGATION")
        if ($Options->get(option => "V4_WRITE_DELEGATION"));
    $default_options->{'v4-id-domain'}
        = $Options->get(option => "V4_ID_DOMAIN")
        if ($Options->get(option => "V4_ID_DOMAIN"));
    $default_options->{'v4.1-acl'} = $Options->get(option => "V4_1_ACL")
        if ($Options->get(option => "V4_1_ACL"));

    foreach my $protocol (@required_protocols) {
        $default_protocols->{$protocol} = 'enabled';
    }

    my @nfs = NACL::C::VserverNfs->find(
        command_interface => $node,
        filter            => {'vserver' => $vserver},
        allow_empty       => 1,
    );

    if (scalar(@nfs) > 0) {
        $Test->nlog("Deleting old NFS settings");
        NACL::C::VserverNfs->delete(
            command_interface => $node,
            'vserver'         => $vserver,
        );
    }

    NACL::C::VserverNfs->create(
        command_interface => $node,
        'access'          => 'true',
        'vserver'         => $vserver,
        %$default_protocols,
        %$default_options,
    );

    _vserver_nfs_modify(
        vserver          => $vserver,
        'mount-rootonly' => 'disabled',
        'nfs-rootonly'   => 'disabled',
    );

    return $TCD::PASS;
} ## end sub configure_nfs

#################
# Name: configure_aggregate
# Description: Configure an aggreagate for the node
# Input:
#   node - What node am I working on
#   nodename - The name of the node
# Output: Pass or failure indicator
#################
sub configure_aggregate {
    my ($node, $nodename) = (shift, shift);

    my $status = $NA;
    my $cnt    = 0;

    my $aggr_disk     = $Options->get(option => "AGGR_DISK_COUNT");
    my $timeout_timer = $Options->get(option => 'AGGR_CREATE_WAIT');
    my $sleep_time    = $Options->get(option => 'AGGR_CHECK_INTERVAL');

    my $normalized_nodename = _normalize($nodename);

    my $aggr = $aggr_base . $normalized_nodename;

    $Test->description("Creating an aggregate for the node $node");

    # Create aggregate here
    my $tmp = NACL::C::Aggregate->find(
        command_interface => $node,
        filter            => {aggregate => $aggr},
        allow_empty       => 1,
    );

    my $create_aggr = 1;
    if ($tmp && $tmp->get_one_state_attribute('nodes')->[0] =~ /$nodename/) {
        $Test->step("Aggregate $aggr already exists.");
    } else {
        if ($tmp) {
            if ($tmp->get_one_state_attribute('volcount') ne '0') {
                $Test->nlog(
                    "Can't delete aggregate $aggr because there are volumes in it"
                );
                $create_aggr = 0;

            } else {
                $Test->step(
                    "Found aggregate $aggr but it's not on node $nodename, deleting it"
                );
                $tmp->delete(
                    "method-timeout" => $Options->get(option => "TIMEOUT"));
            }
        }

        if ($create_aggr) {
            $Test->step("Creating aggregate " . $aggr . " on node " . $nodename);
            NACL::C::Aggregate->create(
                command_interface => $Node,
                aggregate         => $aggr,
                diskcount         => $aggr_disk,
                "method-timeout"  => $Options->get(option => "TIMEOUT"),
                "nodes"           => [$nodename],
                "raidtype"        => $Options->get(option => "AGGR_RAID_TYPE"),
            );
        }

        $status = $TCD::PASS;
    }

    # Wait for aggregate to go online
    $Test->step("Waiting for " . $aggr . " to go online");
    $tmp = NACL::C::Aggregate->find(
        command_interface => $node,
        filter            => {aggregate => $aggr},
    );

    while ($tmp->get_one_state_attribute('state') ne 'online'
        && $timeout_timer > 0) {
        if ($tmp->get_one_state_attribute('state') eq '-') {
            $Test->nlog('cannot get the state of the aggr, node may be down');
            return $TCD::FATAL;
        }
        sleep $sleep_time;
        $timeout_timer -= $sleep_time;
    }

    if ($tmp->get_one_state_attribute('state') ne 'online') {
        $Test->step(
            'Aggregate create time out, may be change AGGR_CREATE_WAIT and try again'
        );
        NACL::APISet::Exceptions::TimeoutException->throw(
            "Aggregate create timeout");
    }

    return $status;
} ## end sub configure_aggregate

#################
# Name: add_ntp_servers
# Description: Adding NTP Servers
# Input:
#   node - What node am I working on
# Output: Pass or failure indicator
#################
sub add_ntp_servers {
    $Test->description('Adding NTP Servers');

    my ($node) = shift();

    # Site config doesn't have ntp servers
    foreach my $s (@{$Options->get(option => 'NTP_SERVER_IP')}) {
        my $ntp_server = NACL::C::SystemServicesNtpServer->find(
            command_interface => $node,
            filter            => {
                node   => $node,
                server => $s,
            },
            allow_empty => 1
        );

        if (!$ntp_server) {
            NACL::C::SystemServicesNtpServer->create(
                command_interface => $Node,
                node              => $node,
                server            => $s,
            );
        }
    }

    return $TCD::PASS;
} ## end sub add_ntp_servers

#################
# Name: configure_nis
# Description: Adding NIS Domain to vserver
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
# Output: Pass or failure indicator
#################
sub configure_nis {
    $Test->description("Adding NIS Domain to vserver");

    my ($node, $vserver) = (shift, shift);

    my $nis_domain;
    my @nis_servers;

    if ($Options->defined('option' => 'NIS_DOMAIN')) {
        $nis_domain = $Options->get('option' => 'NIS_DOMAIN');
    } else {
        $Test->step("Getting NIS Domain from Site Data file");
        $nis_domain = $SiteData{$Site}->{nis_domain};
    }

    if ($Options->defined('option' => 'NIS_SERVERS')) {
        @nis_servers = @{$Options->get('option' => 'NIS_SERVERS')};
    } else {
        $Test->step("Getting NIS Servers IP from Site Data file");
        @nis_servers = split(/\s+/, $SiteData{$Site}->{nis_servers});
    }

    my @localusers
        = split(/,/, $Options->get('option' => 'LOCAL_UNIX_USERS'));

    foreach my $userdata (@localusers) {
        my ($name, $uid, $gid, $fullname) = split(/:/, $userdata);
        $Test->step(
            "Add user: $name, UID: $uid, GID $gid, Full name: $fullname if user does not exist"
        );
        NACL::C::VserverServicesUnixUser->find_or_create(
            command_interface       => $node,
            'user'                  => $name,
            'vserver'               => $vserver,
            create_specific_options => {
                'primary-gid' => $gid,
                'id'          => $uid,
                'full-name'   => $fullname,
            }
        );
    }

    my @localgroups
        = split(/,/, $Options->get('option' => 'LOCAL_UNIX_GROUPS'));
    foreach my $groupdata (@localgroups) {
        my ($name, $id) = split(/:/, $groupdata);
        $Test->step("Create group: $name, ID: $id if group does not exist");
        NACL::C::VserverServicesUnixGroup->find_or_create(
            command_interface       => $node,
            'name'                  => $name,
            'vserver'               => $vserver,
            create_specific_options => {'id' => $id,}
        );
    }

    $Test->step("Join NIS Domain: $nis_domain");
    NACL::C::VserverServicesNisDomain->create(
        command_interface => $node,
        'domain'          => $nis_domain,
        'servers'         => \@nis_servers,
        'vserver'         => $vserver,
        'active'          => "true",
    );

    return $TCD::PASS;
} ## end sub configure_nis

#################
# Name: configure_dns
# Description: Adding $Site DNS to vserver
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
# Output: Pass or failure indicator
#################
sub configure_dns {
    $Test->description("Adding $Site DNS to vserver");

    my ($node, $vserver) = (shift, shift);

    my @dns_servers;
    my @dns_domains;
    if (   $Options->defined('option' => 'DNS_SERVER_IP')
        && $Options->defined('option' => 'DNS_DOMAIN')) {
        $Test->nlog(
            'Overriding DNS Servers IP and DNS Domain Name provided by site data file'
        );
        @dns_servers = @{$Options->get('option' => 'DNS_SERVER_IP')};
        @dns_domains = @{$Options->get('option' => 'DNS_DOMAIN')};
    } else {
        @dns_servers = $SiteData{$Site}->{mgmt_dns1};
        if ($SiteData{$Site}->{mgmt_dns2}) {
            push(@dns_servers, $SiteData{$Site}->{mgmt_dns2});
        }
        @dns_domains = split(/\s+/, $SiteData{$Site}->{dns_domains});
    }

    NACL::C::VserverServicesDns->create(
        command_interface => $node,
        'domains'         => \@dns_domains,
        'name-servers'    => \@dns_servers,
        'vserver'         => $vserver,
    );

    return $TCD::PASS;
} ## end sub configure_dns

#################
# Name: _get_bytes_from_string
# Description: Translate from KB, MB, GB, TB, or PB into bytes
# Input: A string with a size
# Output: An expanded version
#################
sub _get_bytes_from_string {
    my $string_size = lc(shift);
    my $size;

    # Translate from KB, MB, GB, TB, or PB into bytes
    if ($string_size =~ /k/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * 1024;
    } elsif ($string_size =~ /m/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * (1024**2);
    } elsif ($string_size =~ /g/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * (1024**3);
    } elsif ($string_size =~ /t/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * (1024**4);
    } elsif ($string_size =~ /p/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size * (1024**5);
    } elsif ($string_size =~ /b/) {
        $string_size =~ s/^\+([0-9]+).*/$1/;
        $size = $string_size;
    } else {

        # This is an error case.  We could not parse
        # the string as either KB, MB, GB, TB, or PB.
        #
        # Disregard the string by setting $size to 0.
        $Test->nlog("Could not parse CVOL_SIZE option: " . $string_size);
        $size = 0;
    }

    return $size;
} ## end sub _get_bytes_from_string()

#################
# Name: configure_data_interface
# Description:
#   Create a data interface based on the
#   information provided by the tharn host,
# Input:
#   node - What node am I working on
#   vserver - The vserver to configure
#   ip_num - Address to use
#   lif_ext - Name of lif
# Output: Pass or failure indicator
#################
sub configure_data_interface {
    my ($node, $vserver, $ip_num, $lif_ext) = (shift, shift, shift, shift);

    my $lif;

    $Test->description(
        "Create a data lif for $vserver on DATA_IP" . $ip_num . " $lif_ext");

    if ($ip_num > 2) {
        $Test->nlog("Whoops, we are limited to 2 data lifs");
        return $TCD::FAIL;
    }

    my $host   = $node->hostrec();
    my %params = $host->param_fields();
    $Test->step("Create data lif on "
            . $host->id()
            . " for vserver $vserver and DATA_IP"
            . $ip_num . " as "
            . $params{"DATA_IP" . $ip_num});

    $Test->nlog("DATA_IP1 = "
            . $params{'DATA_IP1'}
            . " and DATA_IP2 = "
            . $params{'DATA_IP2'});

    $lif = NACL::C::Interface->find(
        command_interface => $node,
        filter            => {'address' => $params{"DATA_IP" . $ip_num},},
        allow_empty       => 1,
    );

    if ($lif) {
        my $role = $lif->get_one_state_attribute('role');
        if (   $role eq 'cluster'
            || $role eq 'node-mgmt'
            || $role eq 'cluster_mgmt') {
            $Test->nlog(
                "Could not assign a data lif, it was already being used for $role"
            );
            return $TCD::PASS;
        } elsif ($role eq 'data') {
            $Test->nlog(
                "We have a data lif already in use when we were not expecting one."
            );
            return $TCD::FAIL;
        } else {
            $Test->nlog(
                "Could not assign a data lif, it was already being used for $role"
            );
            return $TCD::FAIL;
        }
    }

    $lif = NACL::C::Interface->create(
        command_interface => $node,
        'vserver'         => $vserver,
        'lif'             => $vserver . $lif_ext,
        'role'            => 'data',
        'data-protocol'   => \@{$Options->get(option => 'DATA_LIF_PROTOCOL')},
        'home-node'       => $host->id(),
        'home-port'       => $params{"DATA_PORT" . $ip_num},
        'address'         => $params{"DATA_IP" . $ip_num},
        'netmask'         => $params{"DATA_MASK" . $ip_num},
    );

    my $routing_group = $lif->get_one_state_attribute('routing-group');
    my @route         = NACL::C::Route->find(
        command_interface => $node,
        filter            => {
            'vserver'       => $vserver,
            'routing-group' => $routing_group,
        },
        'allow_empty' => 1,
    );

    # Setup the gateway if needed
    if (!@route) {
        $Test->step("Create network route for routing group " . $routing_group);
        NACL::C::Route->create(
            command_interface => $node,
            'vserver'         => $vserver,
            'routing-group'   => $routing_group,
            'destination'     => '0.0.0.0/0',
            'gateway'         => $params{"DATA_GATEWAY" . $ip_num},
        );
    }

    return $TCD::PASS;
} ## end sub configure_data_interface

#################
# Name: _vserver_nfs_modify
# Description: Apply the NFS vserver modifications.
# Input: Set of parameters to modify
# Output: Pass or failure indicator
#################
sub _vserver_nfs_modify {
    my %parameters = @_;

    $parameters{'vserver'} = $Options->get(option => 'VSERVER')
        unless (exists $parameters{'vserver'});

    {
        try {
            $Apiset->vserver_nfs_modify(%parameters);
        }
        catch NATE::BaseException with {
            my $err_msg = shift()->text();
            return $TCD::FAIL
                if $err_msg !~ m/invalid argument "-mount-rootonly"/;
            $Log->warn('"-mount-rootonly" is not a valid NFS server option');
            return $TCD::PASS;
        }
    }

    return $TCD::PASS;
} ## end sub _vserver_nfs_modify

#################
# Name: delete_vserver
# Description: Helper function to delete vserver
# Input:
#   vserver - The vserver to configure
# Output:
#################
sub delete_vserver {
    my $vserver = shift();
    my $vs_name = $vserver->get_one_state_attribute('vserver');

    # Clean up CIFS server configure
    $Test->step("Delete CIFS settings for $vs_name");
    my @cifs = NACL::C::VserverCifs->find(
        command_interface => $Node,
        'filter'          => {'vserver' => $vs_name,},
        allow_empty       => 1,
    );

    foreach my $cifs (@cifs) {
        $cifs->delete(
            'username' => $Options->get('option' => 'AD_SERVER_USER'),
            'password' => $Options->get('option' => 'AD_SERVER_PASSWORD'),
        );
    }

    $Test->step("Delete NFS settings for $vs_name");
    my @nfs = NACL::C::VserverNfs->find(
        command_interface => $Node,
        'filter'          => {'vserver' => $vs_name,},
        allow_empty       => 1,
    );

    foreach my $nfs (@nfs) {
        $nfs->delete();
    }

    my @volumes = NACL::STask::Volume->find(
        command_interface => $Node,
        filter            => {vserver => $vs_name},
        allow_empty       => 1
    );

    my $root_vol = undef;
    foreach my $volume (@volumes) {
        my $volname;

        try {
            $volname = $volume->get_one_state_attribute('volume');
        }
        catch NACL::Exceptions::NoElementsFound with {};
        next if (!$volname);

        if ($volume->get_one_state_attribute('type') ne "DC") {
            NACL::C::VolumeFlexcache->delete(
                command_interface => $Node,
                vserver           => $vs_name,
                'origin-volume'   => $volname
            );

            if ($vserver->get_one_state_attribute('rootvolume') eq $volname) {
                $root_vol = $volume;
            } else {
                $Test->step("Unmount, offline and delete volume: $volname");
                $volume->unmount()
                    if ($volume->get_one_state_attribute('junction-active')
                    =~ m/true/i);
                $volume->offline()
                    if (
                    $volume->get_one_state_attribute('state') !~ m/offline/i);
                $volume->delete(
                    "method-timeout" => $Options->get(option => "TIMEOUT"));
            }
        }
    }

    if (defined $root_vol) {
        $Test->step("Deleting root volume: "
                . $root_vol->get_one_state_attribute('volume'));
        $root_vol->offline()
            if ($root_vol->get_one_state_attribute('state') !~ m/offline/i);
        $root_vol->delete(
            "method-timeout" => $Options->get(option => "TIMEOUT"));
    }

    $Test->step("Delete export-policies for $vs_name");
    my @policies = NACL::C::ExportPolicy->find(
        command_interface => $Node,
        'filter'          => {'vserver' => $vs_name,},
        allow_empty       => 1,
    );

    foreach my $policy (@policies) {
        my $policyname = $policy->get_one_state_attribute('policyname');

        $Test->step("Delete export-policy rules in $policyname for $vs_name");
        my @rules = NACL::C::ExportPolicyRule->find(
            command_interface => $Node,
            'filter'          => {
                'vserver'    => $vs_name,
                'policyname' => $policyname,
            },
            allow_empty => 1,
        );

        foreach my $rule (@rules) {
            $rule->delete();
        }

        if ($policyname ne 'default') {
            $Test->nlog("Delete export-policy " . $policyname);
            $policy->delete();
        }
    }

    $Test->step("Stopping vserver " . $vs_name . " if it's not stopped");
    $vserver->stop()
        if ($vserver->get_one_state_attribute('admin-state') ne 'stopped');

    $Test->step("Deleting vserver: " . $vs_name);
    {    # Code to work around issue in burt 483819
        try {
            $vserver->delete(
                "method-timeout" => $Options->get(option => "TIMEOUT"));
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            $vserver->start();
            $vserver->delete(
                "method-timeout" => $Options->get(option => "TIMEOUT"));
            $Log->warn("Apply work around for BURT 483819 to delete vserver "
                    . $vs_name);
        }
    };
} ## end sub delete_vserver

#################
# Name: clean_cluster
# Description:
#   This test will try to remove everything in a cluster.
#   This only trigger when RESETCLUSTER=1 is set.
# Input:
# Output: Pass or failure indicator
#################
sub clean_cluster {
    $Test->description(
        "Delete all existing vservers and volumes.",
        "This test is enable by setting RESETCLUSTER=1",
    );

    my @cmdinfs = NACL::C::Node->find();
    if ($#cmdinfs != 1) {
        $Test->nlog('There need to be two nodes!');
        return $TCD::FAIL;
    }

    my $node1 = $cmdinfs[0];
    my $node2 = $cmdinfs[1];

    my $node = $node1;

    if ($Options->get(option => "RESETCLUSTER") == 1) {
        $Test->step("Make sure all aggregates are online");
        my @aggrs = NACL::C::Aggregate->find('command_interface' => $node);
        foreach (@aggrs) {
            $_->online()
                if (
                $_->get_one_state_attribute('state') =~ m/offline|restrict/);
        }

        $Test->step('Make sure all the volumes are online');
        my @volumes = NACL::STask::Volume->find('command_interface' => $node);
        foreach my $vol (@volumes) {
            $vol->online()
                if ($vol->get_one_state_attribute('state') !~ /online/);
        }

        #
        # Here we delete all data lifs.
        #
        my @lifs = NACL::C::Interface->find(
            command_interface => $node,
            filter            => {'role' => 'data',},
            allow_empty       => 1,
        );

        foreach my $lif (@lifs) {
            $lif->delete(
                "method-timeout" => $Options->get(option => "TIMEOUT"));
        }

        my @vservers = NACL::C::Vserver->find(
            command_interface => $node,
            filter            => {type => 'data'},
            allow_empty       => 1
        );

        foreach my $vserver (@vservers) {
            delete_vserver($vserver);
        }

        return $TCD::PASS;
    }

    return $NA;
} ## end sub clean_cluster

#################
# Name: _bg_job_helper
# Description:
#   This function will wait for the background job to complete
# Input:
#   job      - The NACL::C::Job object
#   wait_msg - The message we want to display as steps while we are waiting.
#   warn_msg - The warning message we want to display if the job complete alerady.
# Output:
#################
sub _bg_job_helper {
    my %opt = validate_with(
        params => \@_,
        spec   => {
            'job'      => {type => HASHREF},
            'wait_msg' => {type => SCALAR | UNDEF, optional => 1},
            'warn_msg' => {type => SCALAR | UNDEF, optional => 1},
            'erro_msg' => {type => SCALAR | UNDEF, optional => 1},
        },
    );

    my $job      = $opt{'job'};
    my $wait_msg = $opt{'wait_msg'};
    my $warn_msg = $opt{'warn_msg'};
    my $erro_msg = $opt{'erro_msg'};
    my $job_state;

    $Test->step($wait_msg) if ($wait_msg);
    {
        try {
            $job_state = $job->get_one_state_attribute('state');
        }
        catch NACL::Exceptions::NoElementsFound with {
            $Log->warn(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $Test->nlog(
                "Job disappear from job list, may be it's changing state, retry again"
            );
            $job_state = $job->get_one_state_attribute('state');
        };
        sleep(5) if ($job_state =~ m/queued/i);
    };

    if ($job_state !~ m/(running|queued)/i && $warn_msg) {
        $Log->warn($warn_msg);
        $Test->nlog($warn_msg);
    }

    $job->watch_progress(
        'method-timeout' => $Options->get(option => "TIMEOUT"));
    if ($job->get_one_state_attribute('state') !~ m/success/i && $erro_msg) {
        NACL::APISet::Exceptions::ResponseException->throw(
            text            => $erro_msg,
            response_object => $job
        );
    }
} ## end sub _bg_job_helper

1;
