# -- this is perl --
##
## Commonly used utilities for nfs thpls and PMs.

##
## Synopsis
## require nfs::CommonUtilities;
## our qw{$Node_cmd, $Client_cmd};
##
## client_execute("ls whatever");
## create_vol(volume=>$volume);
##
##

package nfs::CommonUtilities;
## Prerequisite to use any of these functions:
## The $Client_cmd, $Node_cmd must be defined in the user program, or this will fail.
use strict;
use warnings;

use base qw{Exporter};
use vars qw{
    $FILER_AGGR
    $VSERVER_GX_NAME
};

use Tharn qw{param $Log};
use Carp qw{croak};

use Params::Validate qw{validate SCALAR OBJECT};
param('FILER_AGGR',      '-default', 'testaggr');
param('EXPORT_GX_POL',   '-default', 'test_pol');
param('SUDO_PATH',       '-default', '/usr/software/bin/sudo');
param('VSERVER_GX_NAME', '-default', 'testvserver');

our @EXPORT_OK = qw{client_execute
    create_vol
    makefile
    $Node_cmd
    $Client_cmd
    check_client_and_sudo
    return_client
    check_client_object
    client_enter_sudo
    client_exit_sudo
    set_client_object
    reset_client_object};
our @EXPORT = @EXPORT_OK;

# These should be set before any of the functions are called
our ($Client_cmd, $Node_cmd);

##
## Execute the given command in the client.
## The optional second argument is to turn on sudo.
sub client_execute($;$$) {
    my $cmd        = shift;
    my $sudo       = shift;
    my $client_cmd = shift || $Client_cmd;
    $Log->debug(qq{Run the command "$cmd" in the client});
    if ($sudo) {
        $client_cmd->execute(command => "$::SUDO_PATH su");
    }
    my $output = $client_cmd->execute(command => qq{$cmd && echo _OK_});

    #$Log->comment(qq{output:$output, match:$1});
    if ($sudo) {
        $client_cmd->execute(command => "exit");
    }

    if ($output =~ /(.*)_OK_/s) {
        return $1;
    } else {
        croak(qq{Command failed in client:command=$cmd, output=$output});
    }
} ## end sub client_execute($;$$)

sub check_client_and_sudo($) {
    my $client = shift;
    $Log->enter();
    check_client_object($client);
    set_client_object($client);
    client_enter_sudo();
    $Log->exit();
} ## end sub check_client_and_sudo($)

sub return_client() {
    $Log->enter();
    client_exit_sudo();
    reset_client_object();
    $Log->exit();
} ## end sub return_client()

sub check_client_object( $ ) {
    my $client = shift;
    $Log->enter();

    if ($client eq '') {
        croak("Null value was passed, when NACL object was expected");
    }

    #Verify the object passed is a NACL::C::Client object or not
    if (!$client->isa("NACL::C::Client")) {
        croak("The object passed into the subroutine is not a NACL::C::Client object");
    }

    $Log->exit();
    return;
} ## end sub check_client_object( $ )

# Create a volume using the given params.
sub create_vol(%) {
    my %opts = validate(
        @_,
        {   volume   => {type => SCALAR},
            junction => {type => SCALAR, optional => 1},
            type     => {type => SCALAR, optional => 1, regex => qr/DP|RW|DC/, default => 'RW'},
            aggregate => {type => SCALAR, optional => 1, default => $::FILER_AGGR},
            node_cmd  => {type => OBJECT, optional => 1, default => $Node_cmd},
            vserver   => {type => SCALAR, optional => 1, default => $::VSERVER_GX_NAME},
        }
    );
    my $node_cmd = $opts{node_cmd};
    delete($opts{node_cmd});
    $Log->comment(qq{Create the volume:"$opts{volume}", of type: "$opts{type}"});
    if ($opts{type} eq "RW") {
        $opts{'junction-path'} = $opts{junction} ? $opts{junction} : "/$opts{volume}";
        $opts{'unix-permissions'} = "0777";
    }
    if ($opts{junction}) { delete $opts{junction} }
    NACL::STask::Volume->create(
        command_interface  => $node_cmd,
        nacltask_if_exists => 'purge',
        policy             => $::EXPORT_GX_POL,
        aggregate          => $opts{aggregate},
        %opts
    );

    my $volobj = NACL::STask::Volume->find(
        command_interface => $node_cmd,
        filter            => {%opts}
    );
    return $volobj;
} ## end sub create_vol(%)

# Simple command to create a file and fill it with some data.
# Arguments: Filename, Content(optional), Sudo flag(optional).
sub makefile($;$$) {
    my $file    = shift;
    my $sudo    = shift;
    my $content = shift || "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    $Log->comment(qq{Create (and/or) fill a file "$file" in the client});
    client_execute("echo $content > $file", $sudo);
} ## end sub makefile($;$$)

my %SUDO_ENABLED;

sub client_enter_sudo(;$$) {
    my $cli  = shift || $Client_cmd;
    my $sudo = shift || $::SUDO_PATH;
    my $id;

    if ($SUDO_ENABLED{$cli}) {
        $id = client_execute("id", "", $cli);
        if ($id =~ /uid=0/) {
            $Log->comment(qq{The client is already in sudo mode});
            $SUDO_ENABLED{$cli}++;
            return;
        }
    }

    $SUDO_ENABLED{$cli} = 1;
    $cli->execute(command => "$sudo su");

    $id = client_execute("id", "", $cli);
    if ($id !~ /uid=0/) {
        croak(qq{Expected root after doing sudo, but did not get it});
    }
} ## end sub client_enter_sudo(;$$)

sub client_exit_sudo(;$) {
    my $cli = shift || $Client_cmd;

    if (!$SUDO_ENABLED{$cli}) {
        croak(    qq{Programming error, we are attempting to exit sudo in a client,}
                . qq{ but we have not even entered the sudo(did not call  client_enter_sudo for this client).});
    }
    $SUDO_ENABLED{$cli}--;
    if ($SUDO_ENABLED{$cli}) {
        return;
    }

    my $output = client_execute("id", $cli);
    if ($output =~ /uid=0/) {
        $cli->execute(command => "exit");
    } else {
        $Log->comment(qq{Trying to exit from sudo, but already not the root});
    }
} ## end sub client_exit_sudo(;$)

my @CLIENT_OBJ_ARRAY;

sub set_client_object($) {
    my $new_client_obj = shift;
    if ($#CLIENT_OBJ_ARRAY == 10) {
        croak(
            qq{Programming error: You are trying to store more than 10 client objects. Please optimize your code.
        This is possibly due to recursively calling the same function several times without properly returning from that function.
        This should not be done. Best practice is to always set_client_object should be followed by reset_client_object()
            before being called again with yet another client object.}
        );
    }
    push @CLIENT_OBJ_ARRAY, $Client_cmd;
    $Client_cmd = $new_client_obj;
} ## end sub set_client_object($)

sub reset_client_object() {
    if ($#CLIENT_OBJ_ARRAY < 0) {
        croak(qq{reset_client_object must be called only after set_client_object});
    }

    $Client_cmd = pop(@CLIENT_OBJ_ARRAY);
} ## end sub reset_client_object()

1;
