#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/CompoundRequests.pm#6 $
#
## @summary CompoundRequests - Library for the NFSv4 compound request tests
## @author	dl-nfsqa-all
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

CompoundRequests - Library for the NFSv4 compound request tests

=head1 SYNOPSIS

  use nfs::CompoundRequests;

  # Setup environment and create filehandle points
  set_env(setup_h => \%setup_h, make_files => 1);

  # Cleanup environment
  clean_env(setup_h => \%setup_h);

  # Create volumes and the filehandle points
  create_filehandle_points(fc => $fc, fh_points => \%fh_points);

  # Remove the filehandle points by destroying the volumes
  remove_filehandle_points(fc => $fc, fh_points => \%fh_points);

  # Restore the 'dir_writable' and 'file_readonly_acl' filehandle points
  restore_filehandle_points(
      fc        => $fc,
      fh_points => \%fh_points,
      files     => ['dir_writable', 'file_readonly_acl'],
  );

  # Send a large compound request (all idempotent operations)
  $out = compound_basic(file => $file, ipaddr => $ipaddr);

  # Send a large compound request (include 5 non-idempotent operations)
  $out = compound_basic(file => $file, ipaddr => $ipaddr, non_idempotent => 5);

  # Send a null request (zero operations)
  $out = compound_basic(file => $file, ipaddr => $ipaddr, null => 1);

  # Send a series of compound requests which transition from the
  # filehandle specified by the 'file' parameter to the 'nfile'
  # parameter via PUTROOTFH, PUTPUBFH, PUTFH, RESTOREFH, LOOKUP and LOOKUPP.
  $returned = compound_transition(
      file   => '/vol/volx',
      nfile  => '/vol/volx/dir1',
      ipaddr => $ipaddr,
      lookup => 1,
  );

  # Send all operations with the current filehandle set to the filehandle
  # given by the 'file' parameter (directory)
  $returned = compound_allops(file => '/vol/volx', ipaddr => $ipaddr);

  # Send all operations with the current filehandle set to the filehandle
  # given by the 'file' parameter (regular file)
  $returned = compound_allops(file => '/vol/volx/file', ipaddr => $ipaddr);

  # Send all operations with the current filehandle set to the filehandle
  # given by the 'file' parameter (named attribute directory of file)
  $returned = compound_allops(
      file      => '/vol/volx/file',
      ipaddr    => $ipaddr,
      nattr_dir => 1,
  );

  # Send all operations with the current filehandle set to the filehandle
  # given by the 'file' parameter (named attribute of file)
  $returned = compound_allops(
      file   => '/vol/volx/file/stream',
      ipaddr => $ipaddr,
      nattr  => 1,
  );

  # Send all operations with the current filehandle set to the filehandle
  # given by the 'file' parameter (current filehandle is not set)
  $returned = compound_allops(
      file   => '/vol/volx/file/stream',
      ipaddr => $ipaddr,
      no_fh  => 1,
  );

  # Send a compound request with an invalid operation (opcode: 0)
  $returned = compound_malformed(
      file      => '/vol/volx',
      ipaddr    => $ipaddr,
      operation => 'invalid_op',
      opcode    => 0,
  );

  # Send a RENAME (file1 => file2) operation with no saved filehandle,
  $returned = compound_malformed(
      file      => '/vol/volx/file1',
      nfile     => '/vol/volx/file2',
      ipaddr    => $ipaddr,
      operation => 'renamef',
  );

  # Send a RENAME operation with the current filehandle set to the
  # filehandle given by the 'file' parameter and the saved filehandle
  # set to the filehandle given by the 'nfile' parameter
  $returned = compound_fh(
      file      => '/vol/volx/dir1',
      nfile     => '/vol/volx/dir2',
      ipaddr    => $ipaddr
      operation => 'renamef',
  );

  # Send a LINK (-> nfile) operation with the current filehandle set to the
  # filehandle given by the 'file' parameter and the saved filehandle
  # set to the filehandle given by the 'nfile' parameter
  $returned = compound_fh(
      file      => '/vol/volx/dir1',
      nfile     => '/vol/volx/file1',
      ipaddr    => $ipaddr
      operation => 'linkf',
  );

  # Check results returned from the compound_* functions to a list
  # of expected results
  $results = check_results(returned => $returned,
                           expected => \%expected);

  # Send the compound request specified by the list of operations
  $out = send_compound(nfs => $nfs_obj, ops => \@ops);

  # Return array which contains minor versions which will be tested. If any value
  # other than 0, 1 or both are specified, an error will be returned.
  my @minor_ver = validate_minor_ver (minor_ver => $MINOR_VERSION);


=head1 DESCRIPTION

Provides procedures that are common to compound request tests.
These include subroutines to setup and cleanup the compound request test
environment, subroutines to create different file system objects with
various attributes and subroutines to send compound requests. The
procedures to send compound requests can send malformed compound
requests, invalid operation codes, very large requests with or without
non-idempotent operations, all valid operation codes with different
object types as the current filehandle or saved filehandle.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-all
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/CompoundRequests.pm#6 $
  Last revised $Date: 2013/09/05 $

=head1 COPYRIGHT

  Copyright 2007, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::CompoundRequests;

use strict;
use warnings;
use base qw(Exporter);
use Tharn;
use Carp qw(croak);
use FS::Util::Setup qw(create_vol);
use Volume qw(vol_exists vol_purge);
use nfs::Setup qw(nfs_setup nfs_cleanup);
use SPARSE::Util::Logger qw(logStep logWarn logDbg);
use nfs::GenericExport qw(export_path unexport_path);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT);
use nfs::NfsClient;
use nfs::NfsDef qw(:NFSCONSTANTS bitmap4_convert nfs_error);
use nfs::NfsXdr qw(xdr_uint xdr_uint64 xdr_opaque xdr_fhandle xdr_bitmap4
    xdr_fattr4 xdr_cookie xdr_cookieverf xdr_sattr xdr_stateid4 xdr_compound);

our @EXPORT_OK = qw(
  set_env
  clean_env
  create_filehandle_points
  remove_filehandle_points
  restore_filehandle_points
  check_results
  compound_basic
  compound_transition
  compound_allops
  compound_malformed
  compound_fh
  send_compound
  validate_minor_ver
);

my $Hostrec = host($main::FILER);
if($Hostrec->hosttype()=~/filer-ng/) {
	require Server;
	require nfs_setup_gx;
	nfs_setup_gx->import("get_data_ip_cmode","create_gx_export_rule");
}

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub set_env(%);
sub clean_env(%);
sub create_filehandle_points(%);
sub remove_filehandle_points(%);
sub restore_filehandle_points(%);
sub check_results(%);
sub compound_basic(%);
sub compound_transition(%);
sub compound_allops(%);
sub compound_malformed(%);
sub compound_fh(%);
sub send_compound(%);
sub validate_minor_ver (%);
sub _create_file($$$;%);
sub _create_dir($$$;%);
sub _remove_file($$$;);
sub _get_nfs_error($;);
sub _nfs_mount(%);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '0.01';

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
# For burt 236344 the testing of unexported file handles needs to
# be disabled until the tests are fixed to correctly test the
# psuedofs.
my $do_unexported = 0;

# Base file name
my $base = "NFSv4CompoundReq.$$";
# Common parameter for all compound functions
my %common_parameters = (
    file   => { type => SCALAR },
    ipaddr => { type => SCALAR, optional => 1 },
);
# Bitmap of all attributes for getattr
my $all_attrs = bitmap4_convert(
    attrlist => [
        FATTR4_SUPPORTED_ATTRS,
        FATTR4_TYPE,
        FATTR4_FH_EXPIRE_TYPE,
        FATTR4_CHANGE,
        FATTR4_SIZE,
        FATTR4_LINK_SUPPORT,
        FATTR4_SYMLINK_SUPPORT,
        FATTR4_NAMED_ATTR,
        FATTR4_FSID,
        FATTR4_UNIQUE_HANDLES,
        FATTR4_LEASE_TIME,
        FATTR4_RDATTR_ERROR,
        FATTR4_ACL,
        FATTR4_ACLSUPPORT,
        FATTR4_ARCHIVE,
        FATTR4_CANSETTIME,
        FATTR4_CASE_INSENSITIVE,
        FATTR4_CASE_PRESERVING,
        FATTR4_CHOWN_RESTRICTED,
        FATTR4_FILEHANDLE,
        FATTR4_FILEID,
        FATTR4_FILES_AVAIL,
        FATTR4_FILES_FREE,
        FATTR4_FILES_TOTAL,
        FATTR4_FS_LOCATIONS,
        FATTR4_HIDDEN,
        FATTR4_HOMOGENEOUS,
        FATTR4_MAXFILESIZE,
        FATTR4_MAXLINK,
        FATTR4_MAXNAME,
        FATTR4_MAXREAD,
        FATTR4_MAXWRITE,
        FATTR4_MIMETYPE,
        FATTR4_MODE,
        FATTR4_NO_TRUNC,
        FATTR4_NUMLINKS,
        FATTR4_OWNER,
        FATTR4_OWNER_GROUP,
        FATTR4_QUOTA_AVAIL_HARD,
        FATTR4_QUOTA_AVAIL_SOFT,
        FATTR4_QUOTA_USED,
        FATTR4_RAWDEV,
        FATTR4_SPACE_AVAIL,
        FATTR4_SPACE_FREE,
        FATTR4_SPACE_TOTAL,
        FATTR4_SPACE_USED,
        FATTR4_SYSTEM,
        FATTR4_TIME_ACCESS,
        FATTR4_TIME_BACKUP,
        FATTR4_TIME_CREATE,
        FATTR4_TIME_DELTA,
        FATTR4_TIME_METADATA,
    ],
);

# Hash of all op codes
my %op_codes = (
    ACCESS              => 3,
    CLOSE               => 4,
    COMMIT              => 5,
    CREATE              => 6,
    DELEGPURGE          => 7,
    DELEGRETURN         => 8,
    GETATTR             => 9,
    GETFH               => 10,
    LINK                => 11,
    LOCK                => 12,
    LOCKT               => 13,
    LOCKU               => 14,
    LOOKUP              => 15,
    LOOKUPP             => 16,
    NVERIFY             => 17,
    OPEN                => 18,
    OPENATTR            => 19,
    OPEN_CONFIRM        => 20,
    OPEN_DOWNGRADE      => 21,
    PUTFH               => 22,
    PUTPUBFH            => 23,
    PUTROOTFH           => 24,
    READ                => 25,
    READDIR             => 26,
    READLINK            => 27,
    REMOVE              => 28,
    RENAME              => 29,
    RENEW               => 30,
    RESTOREFH           => 31,
    SAVEFH              => 32,
    SECINFO             => 33,
    SETATTR             => 34,
    SETCLIENTID         => 35,
    SETCLIENTID_CONFIRM => 36,
    VERIFY              => 37,
    WRITE               => 38,
    RELEASE_LOCKOWNER   => 39,
    ILLEGAL             => 10044
);

# Hash to map opcode to name
my %op_names = reverse %op_codes;

#---------------------------------------------------------------------
# ACEs used to initialize an ACL for READ-ONLY Access
#---------------------------------------------------------------------
my @ro_aces = (
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flag  => 0,
        mask  => ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_READ_ACL | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER | ACE4_EXECUTE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flag  => 0,
        mask  => ACE4_APPEND_DATA | ACE4_WRITE_DATA | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_WRITE_ATTRIBUTES | ACE4_SYNCHRONIZE,
        who   => 'OWNER@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flag  => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_READ_ACL | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER | ACE4_EXECUTE,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flag  => ACE4_IDENTIFIER_GROUP,
        mask  => ACE4_APPEND_DATA | ACE4_WRITE_DATA | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_WRITE_ATTRIBUTES | ACE4_SYNCHRONIZE,
        who   => 'GROUP@'
    },
    {
        type  => ACE4_ACCESS_ALLOWED_ACE_TYPE,
        flag  => 0,
        mask  => ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS |
                 ACE4_READ_ATTRIBUTES | ACE4_READ_ACL | ACE4_WRITE_ACL |
                 ACE4_WRITE_OWNER | ACE4_EXECUTE,
        who   => 'EVERYONE@'
    },
    {
        type  => ACE4_ACCESS_DENIED_ACE_TYPE,
        flag  => 0,
        mask  => ACE4_APPEND_DATA | ACE4_WRITE_DATA | ACE4_WRITE_NAMED_ATTRS |
                 ACE4_WRITE_ATTRIBUTES | ACE4_SYNCHRONIZE,
        who   => 'EVERYONE@'
    }
);

#---------------------------------------------------------------------

=head2 set_env

  set_env(setup_h => \%setup_h
          [, make_files => 0|1 ]);

Set up environment for NFSv4 compound request tests. It enables NFS
version 4 on the filer. All setup state is passed back to the test
on the 'setup_h' named option.
This hash is populated with the following:

  {
    fc          => Filer connection object,
    volume      => Full path of created volume,
    tfname      => Test file name,
    filehandles => A reference to a hash of file paths created on the
                   filer for use by the test. Empty if make_files is not set.
                   Hash keys reflect the type of file created:
                       source_file,
                       readwrite_export,
                       readonly_export,
                       unexported_vol,
                       dir_writable,
                       dir_readonly_export,
                       dir_readonly_modebits,
                       dir_readonly_acl,
                       file_writable,
                       file_readonly_export,
                       file_readonly_modebits,
                       file_readonly_acl,
                       symlink_writable_export,
                       symlink_readonly_export,
                       char_special_writable_export,
                       char_special_readonly_export,
                       block_special_writable_export,
                       block_special_readonly_export,
                       socket_special_writable_export,
                       socket_special_readonly_export,
                       fifo_special_writable_export,
                       fifo_special_readonly_export,
                       named_attr_dir_writable_export,
                       named_attr_writable_export,
                       named_attr_dir_readonly_export,
                       named_attr_readonly_export
  }

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item make_files

(Optional) : If set to non-zero, causes export directories and files to be
created on the filer for use by the test.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_env(%) {
    my (%opts) = validate (
        @_,
        {
            setup_h    => { type => HASHREF },
            make_files => { type => SCALAR, default => 0 },
        }
    );
    logStep('set_env: setup environment for test');

	my $f = host ($main::FILER);
    if ($f->hosttype =~ /filer-ng/){
		if($main::SETUP_DONE) {
			# Create or Modify NFS parameters as required by test
			my $server_obj= _get_server_node();
			eval {
				$server_obj->vserver_nfs_show(vserver=>$main::VSERVER_GX_NAME);
			};
			if ($@ && $@ =~ /no entries/){
				eval {
					$server_obj->vserver_nfs_create(vserver=>$main::VSERVER_GX_NAME, 'v3'=>'enabled', 'v4.0'=>'enabled', 'tcp'=>'enabled', 'udp'=>'enabled', 'v4.1'=>'enabled','v4.0-acl'=>'enabled', 'v4.1-acl'=>'enabled');
				};
				if($@) {
					croak("NFS create failed - $@");
				}
			}
			else {
				eval {
					$server_obj->vserver_nfs_modify(vserver=>$main::VSERVER_GX_NAME, 'v3'=>'enabled', 'v4.0'=>'enabled', 'tcp'=>'enabled', 'udp'=>'enabled', 'v4.1'=>'enabled','v4.0-acl'=>'enabled', 'v4.1-acl'=>'enabled');
				};
				if($@) {
					croak("NFS modify failed - $@");
				}
			}
		}
	}

    # Common setup -- creates and exports a volume depending
    # on the $main::SETUP_DONE parameter
    nfs_setup(
        setup_h      => $opts{setup_h},
        nfs_version  => 4,
        acl_enable   => 'on',
        create_files => 1,
    );

    if ( $opts{make_files} ) {
        # Create a series of test exports and files
        $opts{setup_h}{filehandles} = {};
        create_filehandle_points(
            fc        => $opts{setup_h}{fc},
            fh_points => $opts{setup_h}{filehandles},
        );
    } else {
        # Set tfname to the name of the file created on nfs_setup
        $opts{setup_h}{tfname} = $opts{setup_h}{files}[0]{file};
    }

    logStep('set_env: done');
    return;
}

#---------------------------------------------------------------------

=head2 clean_env

  clean_env(setup_h => \%setup_h);

Cleanup after a test run if necessary by removing the files
created on the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed.
These hash is initialized on setup.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub clean_env(%) {
    my (%opts) = validate (
        @_,
        {
            setup_h => { type => HASHREF },
        }
    );
    logStep('clean_env: cleanup environment for test');

    my $f = host ($main::FILER);
    if (!($f->hosttype =~ /filer-ng/)){

        # Catch any errors to properly run common_cleanup
        if ( $main::CLEANUP ne 'none' && $opts{setup_h}{volume} ) {
            if ( $opts{setup_h}{filehandles} ) {
                # Remove the created volumes (and their contents)
                eval {
                    remove_filehandle_points(
                        fc        => $opts{setup_h}{fc},
                        fh_points => $opts{setup_h}{filehandles},
                    );
                };
                logWarn("clean_env: $@") if $@;
            }
        }

        # Common cleanup -- cleans volume depending
        # on the $main::CLEANUP parameter
        nfs_cleanup(setup_h => $opts{setup_h});

        logStep('clean_env: done');
    } else {
		if ( $main::CLEANUP ne 'none') {
			# Delete export_rw and export_ro volumes
			my $server_obj= _get_server_node();
            foreach my $vol (qw{export_rw export_ro}) {
			    eval {
				    $server_obj->volume_unmount(vserver=>$main::VSERVER_GX_NAME, volume=>$vol);
				    $server_obj->volume_offline(vserver=>$main::VSERVER_GX_NAME, volume=>$vol);
				    $server_obj->volume_delete(vserver=>$main::VSERVER_GX_NAME, volume=>$vol);
                }
			};

            nfs_setup_gx::clean_env_gx();
		}
	}

    return;
}

#---------------------------------------------------------------------

=head2 create_filehandle_points

  create_filehandle_points(fc => $fc, fh_points => \%fh_points);

Creates a variety of filehandles including regular files, directories,
named pipes, named sockets, character and block devices, symlinks
and named attributes. All file system objects are created on both
read-only and writable volumes. In addition, regular files and directories
are also created with read-only permissions, as well as read-only ACL.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of filer.

=item fh_points

(Required) : Hash reference where filehandle paths (paths to directories
and files) created in this function are returned to the caller.

=back

=item Returns

Nothing.

=item Notes

The exports are created as new volumes, and all filehandle points
are created on the new volumes. Cleanup can be done by calling
L<C<remove_filehandle_points>|/remove_filehandle_points>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create_filehandle_points(%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            fh_points => { type => HASHREF },
        }
    );
    my ($nfs, $vol_name, $out, $filename, $filename2, $dir, $dout, $ng_check);
    my ($rw_mnt_fh, $ro_mnt_fh, $mnt_fh, $export, $mode, $dhandle);
    my ($unexported, $rw_export, $ro_export, %cr_files, $ro_vol, $rw_vol);
    my %attr_args = ( FATTR4_MODE() => 0777 );
    my $source_file = 'source';
    my $fc = $opts{fc};
    my $fh_points = $opts{fh_points};
    # Different types of file system objects to create
    my $f = host ($main::FILER);
    if ($f->hosttype =~ /filer-ng/){$ng_check = 1;}
    my @files = qw(
        readwrite_export
        readonly_export
        dir_writable
        dir_readonly_export
        dir_readonly_modebits
        dir_readonly_acl
        file_writable
        file_readonly_export
        file_readonly_modebits
        file_readonly_acl
        symlink_writable_export
        symlink_readonly_export
        char_special_writable_export
        char_special_readonly_export
        block_special_writable_export
        block_special_readonly_export
        socket_special_writable_export
        socket_special_readonly_export
        fifo_special_writable_export
        fifo_special_readonly_export
    );
    if (!$ng_check) {
        #7mode supports streams, which is not done in cmode yet.
        push @files, qw(
           named_attr_dir_writable_export
           named_attr_writable_export
           named_attr_dir_readonly_export
           named_attr_readonly_export
        );
    }

    # Generic name for source file to create
    $fh_points->{source_file} = $source_file;

    #------------------------
    # Read-write export
    #------------------------
    my $hostname1;
	if (!$ng_check) {
        $rw_vol = "export_rw";
        $rw_export = "/vol/$rw_vol";
        if ( ! $fh_points->{readwrite_export} && ! vol_exists($fc, $rw_vol) ) {
            # Create volume
            create_vol(
                filer_vol    => $rw_vol,
                extern_setup => 0,
                guarantee    => 'none',
            );
        }
        croak $@ if $@;

        # Export volume read-writable to create files
        export_path($fc, $rw_export, OPTIONS => "rw,anon=0") or croak $@;
		$hostname1 = $fc->connhost;
	if ( $main::ADDR_TYPE eq 'inet6' ) {
             my $h = host($main::FILER);
             $hostname1 = $h->net_best_address(sort_preference=>{family=>'inet6'});
             logcomment("IPv6 address used for synthetic client: $hostname1");
        }

	}
    else {
		$rw_vol = "export_rw";
        $rw_export='/export_rw';

		# Create export_rw volume if not present
		my $server_obj= _get_server_node();
        eval {
            $server_obj->volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$rw_vol);
        };
        if ($@ && $@ =~ /no entries/) {
            eval {
				$server_obj->volume_create(vserver=>$main::VSERVER_GX_NAME,
                                           volume=>$rw_vol,
                                           aggregate=>$main::FILER_GX_AGGR,
                                           'unix-permissions' => 0777,
                                           size=> '1GB',
                                           'junction-path'=> $rw_export);
			};
			if($@) {
				croak("Volume Create of $rw_vol failed - $@");
			}
		}

		$hostname1 = get_data_ip_cmode($main::FILER);
		$server_obj->volume_modify(vserver=>$main::VSERVER_GX_NAME, volume=>'export_rw', policy=>$::EXPORT_GX_POL);
	}

    # Create NFS object
    $nfs = nfs::NfsClient->new(
        hostname => $hostname1,
        version  => 4,
        proto    => 'tcp',
    );

    # Mount volume
    $nfs->mount(path => $rw_export);
	logcomment("Mount Success");
    $rw_mnt_fh = $nfs->mnt_fh;
	
    # Change permission to 777 of mount point as root
    $nfs->rpc->auth_sys_set(uid => 0, gid => 0);
    $nfs->setattr(attr => {FATTR4_MODE() => 0777});
    $nfs->rpc->auth_sys_set();

    # Create source file on top level directory of writable export
    eval { _create_file($nfs, $rw_mnt_fh, $source_file) };
    croak "Could not create file $source_file: $@" if $@;

    #------------------------
    # Read-only export
    #------------------------
    if (!$ng_check) {
        $ro_vol = "export_ro";
        $ro_export = "/vol/$ro_vol";
        if ( ! $fh_points->{readonly_export} && ! vol_exists($fc, $ro_vol) ) {
            # Create volume
            create_vol(
                filer_vol    => $ro_vol,
                extern_setup => 0,
                guarantee    => 'none',
            );
        }
        croak $@ if $@;

        # Export volume read-writable to create files
        export_path($fc, $ro_export, OPTIONS => "rw,anon=0") or croak $@;
    }
    else {
		$ro_export = '/export_ro';
		$ro_vol = "export_ro";

		# Create export_ro volume if not present
		my $server_obj= _get_server_node();
        eval {
            $server_obj->volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$ro_vol);
        };
        if ($@ && $@ =~ /no entries/){
            eval {
				$server_obj->volume_create(vserver=>$main::VSERVER_GX_NAME, volume=>$ro_vol, aggregate=>$main::FILER_GX_AGGR,
																	size=> '1GB', 'junction-path'=> $ro_export);
			};
			if($@) {
				croak("Volume Create of $ro_vol failed - $@");
			}
		}

		# Create a read-only export-policy for export_ro volume
		create_gx_export_rule( {'vserver_name'=>$main::VSERVER_GX_NAME,
							   'export_pol'=>'ro_pol',
							   'export_rule_ro'=>'any',
							   'export_rule_rw'=>'none',}
							 );
		$server_obj->volume_modify(vserver=>$main::VSERVER_GX_NAME, volume=>'export_ro', policy=>$::EXPORT_GX_POL);
	}

        # Mount volume
    $nfs->mount(path => $ro_export);
    $ro_mnt_fh = $nfs->mnt_fh;

    # Change permission to 777 of mount point as root
    $nfs->rpc->auth_sys_set(uid => 0, gid => 0);
    $nfs->setattr(attr => {FATTR4_MODE() => 0777});
    $nfs->rpc->auth_sys_set();

    # Create source file on top level directory of read-only export
    eval { _create_file($nfs, $ro_mnt_fh, $source_file) };
    croak "Could not create file $source_file: $@" if $@;

    #------------------------
    # Unexported point
    #------------------------
    if ($do_unexported && !$ng_check) {
        $vol_name = "unexported_vol";
        $unexported = "/vol/$vol_name";
        if ( ! $fh_points->{$vol_name} && ! vol_exists($fc, $vol_name) ) {
            # Create volume
            create_vol(
                filer_vol    => $vol_name,
                extern_setup => 0,
                guarantee    =>'none',
            ) or croak $@;
        }

        # Unexport the volume
        unexport_path($fc, $unexported);

        # Add unexported volume to list of filehandle points
        $fh_points->{$vol_name} = $unexported;
    }

    # Hash used to create only one file on the top level directory
    # of each exported type (read-only and writable exports)
    %cr_files = (
        $ro_export => 1,
        $rw_export => 1,
        named_attr_dir_writable_export => 1,
        named_attr_dir_readonly_export => 1,
    );

    #------------------------
    # Create all file system objects
    #------------------------
    foreach $filename ( @files ) {
        # Use the read-only or writable export according to the name
        # of the file system object to create
        if ( $filename =~ /readonly_export/ ) {
            $mnt_fh = $ro_mnt_fh;
            $export = $ro_export;
        } else {
            $mnt_fh = $rw_mnt_fh;
            $export = $rw_export;
        }

        # Filehandle point already exists
        next if $fh_points->{$filename};

        # The named attributes are created when creating the
        # named attribute directories
        if ( $filename eq 'named_attr_writable_export' ) {
            $filename2 = 'named_attr_dir_writable_export';
            if ( $cr_files{$filename2} ) {
                $filename = $filename2;
                $cr_files{$filename2} = 0;
            } else {
                next;
            }
        }
        if ( $filename eq 'named_attr_readonly_export' ) {
            $filename2 = 'named_attr_dir_readonly_export';
            if ( $cr_files{$filename2} ) {
                $filename = $filename2;
                $cr_files{$filename2} = 0;
            } else {
                next;
            }
        }

        if ( $filename =~ /^read(only|write)_export$/ ) {
            # Save full path of file system object
            $fh_points->{$filename} = $export;
        } else {
            # Save full path of file system object
            $fh_points->{$filename} = "$export/$filename";

            # Remove filehandle if it exists
            _remove_file($nfs, $mnt_fh, $filename);
        }

        $dir = 0;
        $dhandle = undef;
        if ( $filename =~ /^dir/ ) {
            logDbg("Create directory $export/$filename");
            # Create a directory
            $out = eval { _create_dir($nfs, $mnt_fh, $filename, mode => 0777) };
            croak "Could not create directory $export/$filename: $@" if $@;
            $dhandle = $out->{fhandle};

            logDbg("Create file on directory $export/$filename");
            # Create a regular file on directory just created
            eval { _create_file($nfs, $out->{fhandle}, $source_file) };
            croak "Could not create file " .
                  "$export/$filename/$source_file: $@" if $@;
            # Set mode to read-only for directory if this directory is
            # a read-only directory based on the mode bits
            $mode = 0555;
            $dir = 1;
        } elsif ( $filename =~ /^symlink/ ) {
            logDbg("Create symbolic link $export/$filename");
            # Create a symbolic link
            $out = $nfs->create(
                dhandle => $mnt_fh,
                name    => $filename,
                type    => NFLNK,
                data    => $fh_points->{file_writable},
                attr    => \%attr_args,
            );
        } elsif ( $filename =~ /^char_special/ ) {
            logDbg("Create character special device $export/$filename");
            # Create a character special device as root
            $nfs->rpc->auth_sys_set(uid => 0, gid => 0);
            $out = $nfs->create(
                dhandle => $mnt_fh,
                name    => $filename,
                type    => NFCHR,
                attr    => \%attr_args,
                major   => 13,
                minor   => 2,
            );
            $nfs->rpc->auth_sys_set();
        } elsif ( $filename =~ /^block_special/ ) {
            logDbg("Create block special device $export/$filename");
            # Create a block special device as root
            $nfs->rpc->auth_sys_set(uid => 0, gid => 0);
            $out = $nfs->create(
                dhandle => $mnt_fh,
                name    => $filename,
                type    => NFBLK,
                attr    => \%attr_args,
                major   => 13,
                minor   => 2,
            );
            $nfs->rpc->auth_sys_set();
        } elsif ( $filename =~ /^socket_special/ ) {
            logDbg("Create named socket $export/$filename");
            # Create a named socket
            $out = $nfs->create(
                dhandle => $mnt_fh,
                name    => $filename,
                type    => NFSOCK,
                attr    => \%attr_args,
            );
        } elsif ( $filename =~ /^fifo_special/ ) {
            logDbg("Create named pipe $export/$filename");
            # Create a named pipe
            $out = $nfs->create(
                dhandle => $mnt_fh,
                name    => $filename,
                type    => NFFIFO,
                attr    => \%attr_args,
            );
        } elsif ( $filename !~ /^read(only|write)_export$/ ) {
            logDbg("Create regular file $export/$filename");
            # Create a regular file
            $out = eval { _create_file($nfs, $mnt_fh, $filename) };
            croak "Could not create file $export/$filename: $@" if $@;

            # Set mode to read-only for file if this file is
            # a read-only file based on the mode bits
            $mode = 0444;
        } else {
            $out = { fhandle => $mnt_fh };
        }

        # Open named attribute directory (create it if necessary)
        $dout = $out; # Save the create file/dir results
        if (!$ng_check) {
            if ( $out && $filename !~ /^(symlink|char|block|socket|fifo)/ ) {
                logDbg("Create named attribute directory on $export/$filename");
				$out = $nfs->openattr(fhandle => $out->{fhandle}, createdir => 1);
            }
        }

        if ( $filename =~ /^named_attr/ ) {
            # Create a named attribute
            $filename2 = $filename;
            $filename2 =~ s/dir_//;
            $fh_points->{$filename2} = "$export/$filename/$filename2";

            # Set the filehandle to the named attribute directory just opened
            $dhandle = $out->{fhandle};

            logDbg("Create named attribute $filename2 on $export/$filename");
            # Create the named attribute
            eval { _create_file($nfs, $out->{fhandle}, $filename2) };
            croak "Could not create named attribute " .
                  "$export/$filename/$filename2: $@" if $@;
            $dir = 1;
            $cr_files{$filename} = 0;
        }

        # Create files needed for various operations which require a
        # file to be present. Create if filehandle point is a directory
        # (including a named attribute directory) or the top level
        # directory of the export
        if ( $cr_files{$export} || defined $dhandle ) {
            # Get proper dhandle
            if ( ! defined $dhandle ) {
                # The directory filehandle is not set so this is the
                # top level directory of the export -- set flag so
                # not to create these files again on this export
                $dhandle = $mnt_fh;
                $cr_files{$export} = 0;
                $dir = 1;
            }

            logDbg("Create other files");
            # Create source file for LINK
            _create_file($nfs, $dhandle, "$base.source");

            if ( $dir ) {
                # Create file used for LOOKUP
                _create_file($nfs, $dhandle, "$base.lookup");
                # Create file used for REMOVE
                _create_file($nfs, $dhandle, "$base.remove");
                # Create file used for RENAME
                _remove_file($nfs, $dhandle, "$base.newname");
                _create_file($nfs, $dhandle, "$base.rename");
                # Create file used for SECINFO
                _create_file($nfs, $dhandle, "$base.secinfo");
            }
        }

        # Change the file/directory to read-only
        # This change must be done last to be able to create the other files
        if ( $filename =~ /readonly_modebits/ ) {
            # Set readonly mode bits on directory/file
            logDbg("Set readonly mode bits on $export/$filename");
            eval { $nfs->setattr(
                fhandle => $dout->{fhandle},
                attr    => {FATTR4_MODE() => $mode},
            ) };
            croak "Could not make $export/$filename readonly: $@" if $@;
        } elsif ( $filename =~ /readonly_acl/ ) {
            # Set readonly ACL on directory/file
            logDbg("Set readonly ACL on $export/$filename");
            eval { $nfs->setattr(
                fhandle => $dout->{fhandle},
                attr    => { FATTR4_ACL() => \@ro_aces },
            ) };
            croak "Could not set ACL on $export/$filename: $@" if $@;
        }
    }

    #------------------------
    # Change the read-only export status
    #------------------------
    if (!$ng_check) {
        export_path($fc, $ro_export, OPTIONS => "ro,anon=0") or croak $@;
    }
    else {
       my $server_obj= _get_server_node();
	   $server_obj->volume_modify(vserver=>$main::VSERVER_GX_NAME, volume=>'export_ro', policy=>'ro_pol');
    }

    return;
}

#---------------------------------------------------------------------

=head2 remove_filehandle_points

  remove_filehandle_points(fc => $fc, fh_points => \%fh_points);

Removes the various filehandles (exports/files) by destroying the volumes
created by L<C<create_filehandle_points>|/create_filehandle_points>.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of filer.

=item fh_points

(Required) : Hash reference of filehandle paths (paths to directories
and files).

=back

=item Returns

Nothing, on error logs a warning.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub remove_filehandle_points(%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            fh_points => { type => HASHREF },
        }
    );
    my $fc = $opts{fc};
    my $fh_points  = $opts{fh_points};
    my $rw_vol     = $fh_points->{readwrite_export};
    my $ro_vol     = $fh_points->{readonly_export};
    # Get the volume names
    $rw_vol     =~ s{.*/}{};
    $ro_vol     =~ s{.*/}{};

    # Remove created volumes (and their files/dirs)
    vol_purge($fc, $rw_vol)     or logWarn("remove_filehandle_points: $@");
    vol_purge($fc, $ro_vol)     or logWarn("remove_filehandle_points: $@");

    if ($do_unexported) {
        my $unexported = $fh_points->{unexported_vol};
        $unexported =~ s{.*/}{};
        vol_purge($fc, $unexported) or logWarn("remove_filehandle_points: $@");
    }

    return;
}

#---------------------------------------------------------------------

=head2 restore_filehandle_points

  restore_filehandle_points(fc => $fc,
                            fh_points => \%fh_points,
                            files => \@files);

Restores the given filehandle points as necessary. RENAME testing may
cause one or more filehandles to be changed or deleted. This function
restores the specified filehandles to their original state. It removes
the entries in the filehandle hash, then calls create_filehandle_points
to rebuild them.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Connectrec object of filer.

=item fh_points

(Required) : Hash reference of filehandle paths (paths to directories
and files).

=item files

(Required) : An array reference of filehandle names to be restored.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub restore_filehandle_points(%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            fh_points => { type => HASHREF },
            files     => { type => ARRAYREF },
        }
    );
    my $file;

    foreach $file ( @{$opts{files}} ) {
        # Delete the file from the filehandle points hash
        # so it will be re-created on create_filehandle_points
        delete $opts{fh_points}{$file};
    }

    # Re-create missing filehandle points
    create_filehandle_points(fc => $opts{fc}, fh_points => $opts{fh_points});

    return;
}

#---------------------------------------------------------------------

=head2 check_results

  $results = check_results(returned => \%returned,
                           expected => \%expected,
                           [, check_last => 0|1 ]);

Compare compound requests results received from the server to a list
of expected results for each operation. The keys for both the returned
and expected hashes is either the name of the operation or its opcode.
The value for each of the entries in the returned hash is the reply of
the compound request as returned from L<C<send_compound>|/send_compound>,
while the value for the expected hash is the status/error code
(NFS4_OK or NFS4ERR_?) or a string for the expected error. The
status/error code is compared with the status code on the NFSv4 reply,
while the string is compared to the error where there is no NFSv4 reply.
See L<C<send_compound>|/send_compound> subroutine for more information
on the output and error processing.

The value for the expected hash is either a single value (scalar) or a list
of values (array reference) for comparing with multiple expected results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item returned

(Required) : Hash of NFSv4 operations and the actual result received from
performing that operation.

=item expected

(Required) : Hash of NFSv4 operations and the expected results when that
operation is to be performed.

=item check_last

(Required) : If true, it checks if the error code on the NFSv4 reply
is from the expected operation. Default is 1.

=back

=item Returns

A reference to a hash of pass and fail results.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub check_results(%) {
    my (%opts) = validate (
        @_,
        {
            returned   => { type => HASHREF },
            expected   => { type => HASHREF },
            check_last => { type => SCALAR, default => 1 },
        }
    );
    my ($key, @expected, $expect, $op_num, $result, %result_map);
    my (@fail, @pass, $exp_res, @exp_msgs, @exp_codes);

    foreach $key ( sort keys %{$opts{expected}} ) {
        # Create expected results into an array
        @exp_msgs  = ();
        @exp_codes = ();
        @expected  = (ref $opts{expected}{$key} eq 'ARRAY') ?
                     @{$opts{expected}{$key}} : ( $opts{expected}{$key} );

        # Split expected results into status code and plain string results
        foreach $expect ( @expected ) {
            if ( $expect =~ /^\d+$/ ) {
                # This is a number: status code
                push @exp_codes, $expect;
                $result_map{$expect} = _get_nfs_error({ status => $expect })
                  unless $result_map{$expect};
            } else {
                # Plain string result
                push @exp_msgs, $expect;
            }
        }
        # The expected results in string format, it converts all the status
        # codes into their respective status string
        $exp_res = join('|', (map { $result_map{$_}{status_code} } @exp_codes),
                             @exp_msgs);

        if ( !$opts{returned}{$key} ) {
            push @fail, "$key: Unexpected failure, missing results";
            next;
        }

        # Get result for this operation
        $result = $opts{returned}{$key};

        if ( !ref $result ) {
            # This result is a string
            if ( @exp_msgs ) {
                # Check result against expected strings
                if ( grep { $result =~ /$_/ } @exp_msgs ) {
                    push @pass, "$key: Operation returned $result as expected";
                } else {
                    push @pass, "$key: Operation returned $result, " .
                                "expected $exp_res";
                }
            } else {
                push @fail, "$key: Unexpected failure, $result";
            }
            next;
        }

        # Get operation number to check if error is from expected
        # operation or it just failed in any other operation within
        # the compound request
        if ( $opts{check_last} ) {
            $op_num = ($key =~ /^\d+$/) ? $key : eval "OP_$key()";
        }
        if ( $opts{check_last} && defined $op_num
          && $op_num != $result->{results}[-1]{op} ) {
            push @fail, "$key: No results received from operation. " .
                        $op_names{$result->{results}[-1]{op}} . " returned " .
                        $result->{status_code};
        } elsif ( ! grep { $result->{status} == $_ } @exp_codes ) {
            push @fail, "$key: Operation returned $result->{status_code}, " .
              "expected $exp_res";
        } else {
            push @pass, "$key: Operation returned $result->{status_code} " .
              "as expected";
        }
    }

    # Returned all pass and fail checks
    return { pass => \@pass, fail => \@fail };
}

#---------------------------------------------------------------------

=head2 compound_basic

  $out = compound_basic(file => $file,
                        ipaddr => $ipaddr
                        [, null => 0|1 ]
                        [, non_idempotent => $number ]);

Issues a basic compound request; a null request or a large compound
request (lots of idempotent operations) with an optional number of
non-idempotent operations.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Full path of file to perform operation on.

=item ipaddr

(Required) : IP address of filer.

=item null

(Required) : If true send a null compound request (zero ops), else send
a very large compound. Default is 0.

=item non_idempotent

(Required) : Number on non-idempotent operations to include in the request.
Default is 0.

=item minor_ver

(Optional) : Minor version of NFSv4 to be tested.
Default is 0.

=back

=item Returns

The result of the compound request as returned from
L<C<send_compound>|/send_compound>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound_basic(%) {
    my (%opts) = validate (
        @_,
        {
            null           => { type => SCALAR, default => 0 },
            non_idempotent => { type => SCALAR, default => 0,
                                                  regex => qr/^\d+$/ },
            minor_ver      => {type => SCALAR, default => 0},
            %common_parameters
        }
    );
    my ($nfs, @all_ops, @ops, $xdr_mnt_fh, $xdr_access, $out, $idx);
    my ($args, $seqid, $tag);
    my $nidem_cnt = $opts{non_idempotent};
    my $minor_ver = $opts{minor_ver};
    my $count = $nidem_cnt > 5 ? $nidem_cnt : 5;
    my $name = "compound_basic.$$";
    my ($path, $file) = ($opts{file} =~ m{(.+)/(.+)});
	my $f = host($main::FILER);
    if ( $main::ADDR_TYPE eq 'inet6' ) {
	$opts{ipaddr} = $f->net_best_address(sort_preference=>{family=>'inet6'});
        logcomment("IPv6 address used for synthetic client: $opts{ipaddr}");
    }
    if ($f->hosttype =~ /filer-ng/) {
	$opts{ipaddr} = get_data_ip_cmode($main::FILER);
    }
    # Create NFS object
    $nfs = nfs::NfsClient->new(
        hostname => $opts{ipaddr},
        version  => 4,
        proto    => 'tcp',
        minor_version  => $minor_ver
    );

    # Mount file
    $nfs->mount(path => $path);

    my $clientid;

    if ($minor_ver == 1){
        $clientid = $nfs->{object}->{clientid4};
    }
    elsif ($minor_ver == 0){
        $clientid = $nfs->{object}->{clientid};
    }


    if ( $opts{null} ) {
        # Send null compound (zero ops)
        $tag = 'null compound';
    } else {
        $tag = 'large compound';
        # Convert the topdir filehandle to xdr format
        $xdr_mnt_fh = xdr_fhandle(encode => $nfs->mnt_fh, version => 4);
        # Convert the access bitmask to xdr format
        $xdr_access = xdr_uint(encode => ACCESS_READ|ACCESS_MODIFY);

        # Operations to send to the server
        @ops = (
            { op => OP_PUTFH, args => $xdr_mnt_fh },
            { op => OP_SAVEFH },
            { op => OP_LOOKUP, args => xdr_opaque(encode => $file) },
            { op => OP_GETFH },
            { op => OP_ACCESS, args => $xdr_access },
            { op => OP_GETATTR, args => xdr_bitmap4(encode => $all_attrs) },
            { op => OP_NVERIFY,
              args => xdr_fattr4(encode => { FATTR4_FILEID() => 0 }) },
            { op => OP_RESTOREFH },
            { op => OP_LOOKUPP },
            { op => OP_LOOKUPP },
        );

        # Add the same operations a few times to create an even larger
        # compound request
        push(@all_ops, @ops) foreach (1..$count);

        # Add non-idempotent operations
        $idx = 0;
        $seqid = 0;
        while ( $nidem_cnt && $idx < @all_ops ) {
            if ( $all_ops[$idx]{op} == OP_RESTOREFH ) {
                # Arguments for open
                $args = xdr_uint(encode => $seqid++) .
                        xdr_uint(encode => OPEN4_SHARE_ACCESS_READ) .
                        xdr_uint(encode => OPEN4_SHARE_DENY_NONE) .
                        $clientid .
                        xdr_opaque(encode => $$) .
                        xdr_uint(encode => 1) .
                        xdr_uint(encode => GUARDED) .
                        xdr_fattr4(encode => { FATTR4_MODE() => 0644 }) .
                        xdr_uint(encode => 0) .
                        xdr_opaque(encode => "$name.$nidem_cnt");
                # Insert non-idempotent operation before the RESTOREFH
                splice @all_ops, $idx, 0, (
                    { op => OP_RESTOREFH },
                    { op => OP_OPEN, args => $args },
                );
                $idx += 2;
                $nidem_cnt--;
            }
            $idx++;
        }
    }

    # Send the large compound request
    $out = send_compound(
        nfs => $nfs,
        ops => \@all_ops,
        tag => $tag,
    );

    # Cleanup files created with non-idempotent operations
    foreach $idx ( 1..$opts{non_idempotent} ) {
        eval { $nfs->remove(dhandle => $nfs->mnt_fh, name => "$name.$idx") };
    }

    return $out;
}

#---------------------------------------------------------------------

=head2 compound_transition

  $returned = compound_transition(file => $file,
                                  nfile => $nfile,
                                  ipaddr => $ipaddr
                                  [, lookup => 0|1]
                                  [, minor_ver => 0|1]);

Issues a compound request for export regime transition.
It transitions from the filehandle of the 'file' parameter to another
filehandle via PUTROOTFH, PUTPUBFH, PUTFH, RESTOREFH, LOOKUP and LOOKUPP.
For PUTFH and RESTOREFH, it transitions to the filehandle of the 'nfile'
parameter. For LOOKUPP, the transitions if from filehandle of the 'nfile'
parameter to the filehandle of the 'file' parameter.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Full path of file to set the current filehandle right before
the transition operation.

=item nfile

(Required) : Full path of file to transition to with PUTFH and RESTOREFH.

=item ipaddr

(Required) : IP address of filer.

=item lookup

(Required) : If true, also perform LOOKUP and LOOKUPP requests,
option 'nfile' must be a directory and 'file' must be the parent
directory of 'nfile'. Default is 0.

=item minor_ver

(Optional) : Minor version of NFSv4 to be tested.
Default is 0.

=back

=item Returns

A reference to a hash of the results of each operation performed.
The key is the operation name and the value is the result of the
compound request as returned from L<C<send_compound>|/send_compound>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound_transition(%) {
    my (%opts) = validate (
        @_,
        {
            nfile  => { type => SCALAR },
            lookup => { type => SCALAR, default => 0 },
            minor_ver => {type => SCALAR, default => 0},
            %common_parameters
        }
    );
    my ($nfs, @op_list, $xdr_mnt_fh, $out, $item, $minor_ver);
    my (@op_file, @op_nfile, %results);

	my $f = host($main::FILER);
	if ($f->hosttype =~ /filer-ng/) {
		$opts{ipaddr} = get_data_ip_cmode($main::FILER);
	}

    $nfs = nfs::NfsClient->new(
        hostname => $opts{ipaddr},
        version  => 4,
        proto    => 'tcp',
        minor_version  => $opts{minor_ver},
    );

    # Get filehandle for the nfile
    $nfs->mount(path => $opts{nfile});

    # Convert the filehandle to xdr format
    $xdr_mnt_fh = xdr_fhandle(encode => $nfs->mnt_fh, version => 4);

    # Remove leading '/'
    $opts{file}  =~ s{^/}{};
    $opts{nfile} =~ s{^/}{};

    # Lookup operations for file
    @op_file = (
        { op => OP_PUTROOTFH },
        map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
          split(/\//, $opts{file}),
    );

    # A PUTROOTFH and LOOKUP's for the nfile parameter
    # This is used for the transition via RESTOREFH, LOOKUP and LOOKUPP
    @op_nfile = (
        { op => OP_PUTROOTFH },
        map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
          split(/\//, $opts{nfile}),
    );

    # Add each of the following operations to the compound request
    @op_list = (
        {
            # Transition via PUTROOTFH
            name => 'PUTROOTFH',
            ops  => [ @op_file, { op => OP_PUTROOTFH } ],
        },
        {
            # Transition via PUTPUBFH
            name => 'PUTPUBFH',
            ops  => [ @op_file, { op => OP_PUTPUBFH } ],
        },
        {
            # Transition via PUTFH
            name => 'PUTFH',
            ops  => [ @op_file, { op => OP_PUTFH, args => $xdr_mnt_fh } ],
        },
        {
            # Transition via RESTOREFH
            name => 'RESTOREFH',
            ops  => [
                @op_nfile,
                { op => OP_SAVEFH },
                @op_file,
                { op => OP_RESTOREFH },
            ],
        },
        {
            # Transition via LOOKUP
            name => 'LOOKUP',
            ops  => \@op_nfile,
        },
        {
            # Transition via LOOKUPP
            name => 'LOOKUPP',
            ops  => [ @op_nfile, { op => OP_LOOKUPP } ],
        },
    );

    # Create NFS object
    # Send each compound request for the transitions via
    # PUTROOTFH, PUTPUBFH, GETFH, PUTFH and RESTOREFH
    foreach $item ( @op_list ) {
        if ( $item->{name} =~ /^LOOKUPP?$/ ) {
            next unless $opts{lookup};
        }

        # Send compound request
        $out = send_compound(
            nfs => $nfs,
            ops => $item->{ops},
            tag => 'trans compound',
        );
        $results{$item->{name}} = $out;
    }

    return \%results;
}

#---------------------------------------------------------------------

=head2 compound_allops

  $returned = compound_allops(file => $file,
                              ipaddr => $ipaddr
                              [, no_fh => 0|1 ]
                              [, nattr_dir => 0|1 ]
                              [, nattr => 0|1 ]
                              [, inval_type => 0|1 ]
                              [, inval_fsid => 0|1 ]
                              [, inval_inode => 0|1 ]
                              [, inval_gen => 0|1 ]
                              [, root => 0|1 ]
                              [, root_setattr => 0|1 ]
                              [, file_name => $name ]
                              [, minor_ver => 0|1 ]);

Issues a compound request for each supported NFSv4 operation when the
current filehandle is set to the filehandle of the 'file' parameter.
The following NFSv4 operations are supported:

    ACCESS
    CLOSE
    COMMIT
    CREATE
    GETATTR
    GETFH
    LINK
    LOCK
    LOCKU
    LOCKT
    LOOKUP
    LOOKUPP
    NVERIFY
    OPEN
    OPENATTR
    OPEN_CONFIRM
    OPEN_DOWNGRADE
    READ
    READDIR
    READLINK
    REMOVE
    RENAME
    SAVEFH
    SECINFO
    SETATTR
    VERIFY
    WRITE

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Full path of file to set the current filehandle.

=item ipaddr

(Required) : IP address of filer.

=item no_fh

(Optional) : Run all operations with no current file handle set.
Default is 0.

=item nattr_dir

(Optional) : Option file is a named attribute directory.
Default is 0.

=item nattr

(Optional) : Option file is a named attribute.
Default is 0.

=item inval_type

(Optional) : Use filehandle with invalid type field.
Default is 0.

=item inval_fsid

(Optional) : Use filehandle with invalid fsid.
Default is 0.

=item inval_inode

(Optional) : Use filehandle with invalid inode.
Default is 0.

=item inval_gen

(Optional) : Use filehandle with invalid generation number.
Default is 0.

=item root

(Optional) : Perform all operations as root.
Default is 0.

=item root_setattr

(Optional) : Perform the SETATTR operation as root.
Default is 0.

=item file_name

(Optional) : Use this file name for the following operations:
LOOKUP, REMOVE, RENAME, SECINFO.
Default is to use specific files for each operation created by
L<C<create_filehandle_points>|/create_filehandle_points>.
This is useful when 'file' is a read-only directory in which
the required files can not be created, e.g., '/' or '/vol'.

=item minor_ver

(Optional) : Minor version of NFSv4 to be tested.
Default is 0.

=back

=item Returns

A reference to a hash of the results of each operation performed.
The key is the operation name and the value is the result of the
compound request as returned from L<C<send_compound>|/send_compound>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound_allops(%) {
    my (%opts) = validate (
        @_,
        {
            no_fh        => { type => SCALAR,  default => 0 },
            nattr_dir    => { type => SCALAR,  default => 0 },
            nattr        => { type => SCALAR,  default => 0 },
            inval_type   => { type => SCALAR,  default => 0 },
            inval_fsid   => { type => SCALAR,  default => 0 },
            inval_inode  => { type => SCALAR,  default => 0 },
            inval_gen    => { type => SCALAR,  default => 0 },
            root         => { type => SCALAR,  default => 0 },
            root_setattr => { type => SCALAR,  default => 0 },
            file_name    => { type => SCALAR, optional => 1 },
            minor_ver    => { type => SCALAR,   default => 0 },
            %common_parameters
        }
    );
    my ($nfs, @ops, @op_list, %op_args, @op_path, @op_file, $item);
    my (%results, $dhandle, $out, $xdr_source_fh, $dir);
    my ($open, $lock, $seqid, $lock_seqid, @names, $named_attr);
    my ($fhandle, $status, $can_open, %open_args, $write_stateid);
    my ($sourcef, $lookupf, $removef, $renamef, $secinfof);
    my $path = $opts{file};
    my $minor_ver = $opts{minor_ver};
    my $setattr = { FATTR4_TIME_MODIFY_SET() => SET_TO_SERVER_TIME };
    my ($clientid, $ng_check);

    if ( defined $opts{file_name} ) {
        # Use the file provided for the following operations
        $sourcef  = $opts{file_name};
        $lookupf  = $opts{file_name};
        $removef  = $opts{file_name};
        $renamef  = $opts{file_name};
        $secinfof = $opts{file_name};
    } else {
        # Use specific files for the following operations
        $sourcef  = "$base.source";
        $lookupf  = "$base.lookup";
        $removef  = "$base.remove";
        $renamef  = "$base.rename";
        $secinfof = "$base.secinfo";
    }

    my $f = host ($main::FILER);
    if ($f->hosttype =~ /filer-ng/){
		$ng_check = 1;
		$opts{ipaddr} = get_data_ip_cmode($main::FILER);
	}
    # Remove leading '/'
    $opts{file} =~ s{^/}{};
    @names = split /\//, $opts{file};

    if ( $opts{nattr} ) {
        # Last entry is the named attribute
        $named_attr = pop @names;
        $path = '/' . join('/', @names);
    }

    if ( $main::ADDR_TYPE eq 'inet6' ) {
	 $opts{ipaddr} = $f->net_best_address(sort_preference=>{family=>'inet6'});
         logcomment("IPv6 address used for synthetic client: $opts{ipaddr}");
    }
    # Create NFS object
    $nfs = nfs::NfsClient->new(
        hostname => $opts{ipaddr},
        version  => 4,
        proto    => 'tcp',
        minor_version  => $minor_ver
    );


    # Get filehandle for the file
    $nfs->mount(path => $path);

    # Perform all operations as root
    $nfs->rpc->auth_sys_set(uid => 0, gid => 0) if $opts{root};

    # Create compound request:
    #   PUTROOTFH to get the filehandle of the root
    @ops = ( { op => OP_PUTROOTFH } );
    #   traverse the path using LOOKUP
    @op_file = map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
      @names;

    # Get type of file
    $can_open = 0;
    $out = $nfs->getattr;
    if ( $opts{nattr} ) {
        # Parent directory for the named attribute
        $can_open = 1;
        $dir = 0;
        $nfs->openattr(fhandle => $nfs->mnt_fh);
        $dhandle = $nfs->curr_fh;
    } elsif ( $opts{nattr_dir} ) {
        $dir = 1;
        $nfs->openattr(fhandle => $nfs->mnt_fh);
        $dhandle = $nfs->curr_fh;
    } elsif ( $out->{fattr}{FATTR4_TYPE()} == NFDIR ) {
        $dir = 1;
        $dhandle = $nfs->mnt_fh;
    } else {
        # Get filehandle of parent directory
        $can_open = 1 if $out->{fattr}{FATTR4_TYPE()} == NFREG;
        $dir = 0;
        @op_path = @op_file;
        pop @op_path;
        $out = $nfs->compound(ops => [@ops, @op_path, { op => OP_GETFH }]);
        $dhandle = $out->{results}[-1]{result}{fhandle};
    }

    $out = eval { $nfs->lookup(dhandle => $dhandle, name => $sourcef) };
    $out = { fhandle => $nfs->mnt_fh } if $@;
    $xdr_source_fh = xdr_fhandle(encode => $out->{fhandle}, version => 4);

    if ( $dir || !$can_open ) {
        $open = { stateid => STATEID0 };
        $lock = { lock_stateid => STATEID0 };
        $write_stateid = STATEID0;
    } else {
        # Open file for READ, WRITE and CLOSE
        %open_args = (
            dhandle => $dhandle,
            name    => $named_attr ? $named_attr : $names[-1],
            access  => OPEN4_SHARE_ACCESS_BOTH,
        );
        # Open file for both read and write
        $open = eval { $nfs->open(%open_args) };
        if ( $@ =~ /NFS4ERR_(ROFS|PERM|ACCES)/ ) {
            # This must be a read-only export or file
            # Try to open the file again just for read
            $open_args{access} = OPEN4_SHARE_ACCESS_READ;
            $open = $nfs->open(%open_args);
        } elsif ( $@ ) {
            croak $@;
        }

        # Get proper stateid for the write operation
        $write_stateid = $open_args{access} == OPEN4_SHARE_ACCESS_READ ?
                         STATEID0 : $open->{stateid};

        # Lock the file to get the lock stateid
        $lock = $nfs->lock(
            offset  => 0,
            count   => 10,
            stateid => $open->{stateid},
        );
    }

    # Get sequence id for OPEN
    $seqid = $nfs->seqid();
    # Get lock sequence id for LOCK
    $lock_seqid = $nfs->lock_seqid();

    push @ops, @op_file;
    push @ops, { op => OP_OPENATTR, args => xdr_uint(encode => 0) }
      if $opts{nattr_dir} || $opts{nattr};
    push @ops, { op => OP_LOOKUP, args => xdr_opaque(encode => $named_attr) }
      if $opts{nattr};

    # Set invalid filehandles
    if ( $opts{inval_type} || $opts{inval_fsid}
      || $opts{inval_inode} || $opts{inval_gen} ) {
        # Get filehandle
        $out = $nfs->compound(ops => [@ops, { op => OP_GETFH }]);
        $fhandle = $out->{results}[-1]{result}{fhandle};
    }
    if ( $opts{inval_type} ) {
        if (!$ng_check) {
            substr($fhandle, 24, 4, pack('N', 0xFFFFFFFF));
        }
        else {
            substr($fhandle, 28, 4, pack('N', 0xFFFFFFFF));
        }
    }
    if ( $opts{inval_fsid} ) {
        if (!$ng_check) {
            substr($fhandle, 20, 4, pack('N', 0xFFFFFFFF));
        }
        else {
            substr($fhandle, 20, 8, pack('N', 0xFFFFFFFF));
        }
    }
    if ( $opts{inval_inode} ) {
        substr($fhandle, 12, 4, pack('N', 0xFFFFFFFF));
    }
    if ( $opts{inval_gen} ) {
        substr($fhandle, 16, 4, pack('N', 0xFFFFFFFF));
    }
    if ( $opts{inval_type} || $opts{inval_fsid}
      || $opts{inval_inode} || $opts{inval_gen} ) {
        @ops = ( { op => OP_PUTFH,
                   args => xdr_fhandle(encode => $fhandle, version => 4) } );
        @op_file = @ops;
    }
    if ( $opts{no_fh} ) {
        # Send operations with no current filehandle
        @ops = ();
    }

    if ($minor_ver == 1){
        $clientid = $nfs->{object}->{clientid4};
    }
    elsif ($minor_ver == 0){
        $clientid = $nfs->{object}->{clientid};
    }

    # Arguments for the operations under test
    %op_args = (
        WRITE   => xdr_stateid4(encode => $write_stateid) .
                   xdr_uint64(encode => 0) .
                   xdr_uint(encode => FILE_SYNC) .
                   xdr_opaque(encode => "Write data"),
        READ    => xdr_stateid4(encode => $open->{stateid}) .
                   xdr_uint64(encode => 0) .
                   xdr_uint(encode => 1000),
        COMMIT  => xdr_uint64(encode => 0) .
                   xdr_uint(encode => 1000),
        LOCKT   => xdr_uint(encode => READ_LT) .  # Type
                   xdr_uint64(encode => 0) .      # Offset
                   xdr_uint64(encode => 1000) .   # Length
                   $clientid .
                   xdr_opaque(encode => $$),      # Owner
        LOCK    => xdr_uint(encode => READ_LT) .  # Type
                   xdr_uint(encode => 0) .        # Reclaim
                   xdr_uint64(encode => 0) .      # Offset
                   xdr_uint64(encode => 1000) .   # Length
                   xdr_uint(encode => 1) .        # Encode new lock owner
                   xdr_uint(encode => $seqid++) . # Sequence id
                   xdr_stateid4(encode => $open->{stateid}) .
                   xdr_uint(encode => $lock_seqid++) . # Lock sequence id
                   $clientid .
                   xdr_opaque(encode => $$),      # Owner
        LOCKU   => xdr_uint(encode => READ_LT) .  # Type
                   xdr_uint(encode => $lock_seqid++) . # Lock sequence id
                   xdr_stateid4(encode => $lock->{lock_stateid}) .
                   xdr_uint64(encode => 0) .      # Offset
                   xdr_uint64(encode => 1000),    # Length
        OPEN_DOWNGRADE => xdr_stateid4(encode => $open->{stateid}) .
                   xdr_uint(encode => $seqid++) .
                   xdr_uint(encode => OPEN4_SHARE_ACCESS_READ) .
                   xdr_uint(encode => OPEN4_SHARE_DENY_NONE),
        OPEN    => xdr_uint(encode => $seqid+1) .
                   xdr_uint(encode => OPEN4_SHARE_ACCESS_READ) .
                   xdr_uint(encode => OPEN4_SHARE_DENY_NONE) .
                   $clientid .
                   xdr_opaque(encode => $$) .
                   xdr_uint(encode => 0) .
                   xdr_uint(encode => 0) .
                   xdr_opaque(encode => $sourcef),
        ACCESS  => xdr_uint(encode => ACCESS_READ|ACCESS_MODIFY),
        CREATE  => xdr_uint(encode => NFDIR) .
                   xdr_opaque(encode => "$base.create") .
                   xdr_sattr(encode => { FATTR4_MODE() => 0777 }, version => 4),
        GETATTR => xdr_bitmap4(encode => $all_attrs),
        LINK    => xdr_opaque(encode => "$base.link"),
        LOOKUP  => xdr_opaque(encode => $lookupf),
        VERIFY  => xdr_fattr4(encode => {}),
        OPENATTR => xdr_uint(encode => 0),
        READDIR => xdr_cookie(encode => 0, version => 4) .
                   xdr_cookieverf(encode => '') .
                   xdr_uint(encode => 8*1024*1024) .   # dircount
                   xdr_uint(encode => 512*1024*1024) . # maxcount
                   xdr_bitmap4(encode => $all_attrs),
        REMOVE  => xdr_opaque(encode => $removef),
        RENAME  => xdr_opaque(encode => $renamef) .
                   xdr_opaque(encode => "$base.newname"),
        SECINFO => xdr_opaque(encode => $secinfof),
        SETATTR => xdr_sattr(encode => $setattr, version => 4),
    );

    # Operations to send in the given order, do the I/O operations first
    # so the file under test is closed when all other operations are
    # performed
    @op_list = (
        {
            name => 'WRITE',
            ops  => [ { op => OP_WRITE, args => $op_args{WRITE} } ],
        },
        {
            name => 'READ',
            ops  => [ { op => OP_READ, args => $op_args{READ} } ],
        },
        {
            name => 'COMMIT',
            ops  => [ { op => OP_COMMIT, args => $op_args{COMMIT} } ],
        },
        {
            name => 'LOCKT',
            ops  => [ { op => OP_LOCKT, args => $op_args{LOCKT} } ],
        },
        {
            name => 'LOCK',
            ops  => [ { op => OP_LOCK, args => $op_args{LOCK} } ],
        },
        {
            name => 'LOCKU',
            ops  => [ { op => OP_LOCKU, args => $op_args{LOCKU} } ],
        },
        {
            name => 'OPEN_DOWNGRADE',
            ops  => [
                { op => OP_OPEN_DOWNGRADE, args => $op_args{OPEN_DOWNGRADE} }
            ],
        },
        {
            name => 'CLOSE',
            # The args are dynamically set since it requires the new stateid
            # returned from the open_downgrade
            ops  => [ { op => OP_CLOSE } ],
        },
        {
            name => 'OPEN',
            ops  => [ { op => OP_OPEN, args => $op_args{OPEN} } ],
        },
        {
            name => 'ACCESS',
            ops  => [ { op => OP_ACCESS, args => $op_args{ACCESS} } ],
        },
        {
            name => 'CREATE',
            ops  => [ { op => OP_CREATE, args => $op_args{CREATE} } ],
        },
        {
            name => 'GETATTR',
            ops  => [ { op => OP_GETATTR, args => $op_args{GETATTR} } ],
        },
        {
            name => 'GETFH',
            ops  => [ { op => OP_GETFH } ],
        },
        {
            name => 'LINK',
            ops  => [
                { op => OP_PUTFH, args => $xdr_source_fh },
                { op => OP_SAVEFH },
                { op => OP_PUTROOTFH },
                @op_file,
                { op => OP_LINK, args => $op_args{LINK} },
            ],
        },
        {
            name => 'LOOKUP',
            ops  => [ { op => OP_LOOKUP, args => $op_args{LOOKUP} } ],
        },
        {
            name => 'LOOKUPP',
            ops  => [ { op => OP_LOOKUPP } ],
        },
        {
            name => 'NVERIFY',
            ops  => [ { op => OP_NVERIFY, args => $op_args{VERIFY} } ],
        },
        {
            name => 'OPENATTR',
            ops  => [ { op => OP_OPENATTR, args => $op_args{OPENATTR} } ],
        },
        {
            name => 'READDIR',
            ops  => [ { op => OP_READDIR, args => $op_args{READDIR} } ],
        },
        {
            name => 'READLINK',
            ops  => [ { op => OP_READLINK } ],
        },
        {
            name => 'REMOVE',
            ops  => [ { op => OP_REMOVE, args => $op_args{REMOVE} } ],
        },
        {
            name => 'RENAME',
            ops  => [
                # Rename on same directory (sfh) == (cfh)
                { op => OP_SAVEFH },
                { op => OP_RENAME, args => $op_args{RENAME} },
            ],
        },
        {
            name => 'SAVEFH',
            ops  => [ { op => OP_SAVEFH } ],
        },
        {
            name => 'SECINFO',
            ops  => [ { op => OP_SECINFO, args => $op_args{SECINFO} } ],
        },
        {
            name => 'VERIFY',
            ops  => [ { op => OP_VERIFY, args => $op_args{VERIFY} } ],
        },
        {
            name => 'SETATTR',
            ops  => [ { op => OP_SETATTR, args => $op_args{SETATTR} } ],
        },
        {
            name => 'OPEN_CONFIRM',
            # The args are dynamically set since it requires the stateid
            # from an open and the open is done immediately before this call
            ops  => [ { op => OP_OPEN_CONFIRM } ],
        },
    );

    # Send all compound requests
    foreach $item ( @op_list ) {
        if ( $item->{name} eq 'CLOSE' ) {
			$item->{ops}[0]{args} = xdr_uint(encode => $seqid++) .
                                    xdr_stateid4(encode => $open->{stateid});
        }
        if ( $item->{name} eq 'OPEN_CONFIRM' ) {
			if ( $can_open ) {
                # seqid subroutine differs a bit for v4.1 and v4.0
				if ($minor_ver == 1){
					$nfs->seqid($seqid+1);
				}
				else {
					$nfs->seqid(seqid => $seqid+1);
				}
				# Need to send the OPEN_CONFIRM immediately after an OPEN
                # use a different open owner
                $open = $nfs->open(%open_args, owner => 'owner');
                $seqid = $nfs->seqid();
            }
            $item->{ops}[0]{args} = xdr_stateid4(encode => $open->{stateid}) .
                                    xdr_uint(encode => $seqid++);
        }

        if ( $opts{no_fh} && $item->{name} =~ /^(RENAME|LINK)$/ ) {
            # Just send the RENAME/LINK operation
            $item->{ops} = [ $item->{ops}[-1] ];
        }

        # Perform the SETATTR operation as root
        $nfs->rpc->auth_sys_set(uid => 0, gid => 0)
          if $opts{root_setattr} && $item->{name} eq 'SETATTR';

        # Send compound request
        $out = send_compound(
            nfs => $nfs,
            ops => [ @ops, @{$item->{ops}} ],
            tag => 'allops compound',
        );
        $results{$item->{name}} = $out;

        # Reset authentication
        $nfs->rpc->auth_sys_set
          if $opts{root_setattr} && !$opts{root} && $item->{name} eq 'SETATTR';

        if ( $item->{name} eq 'OPEN_DOWNGRADE'
          && ref $out && !$out->{status} ) {
            $open->{stateid} = $out->{results}[-1]{result}{stateid};
        }
    }

    # Set the sequence id
    if ($minor_ver == 1){
		$nfs->seqid($seqid);
	}
	else {
		$nfs->seqid(seqid => $seqid);
	}

    # Set the lock sequence id
    $nfs->lock_seqid($lock_seqid);

    return \%results;
}

#---------------------------------------------------------------------

=head2 compound_malformed

  $returned = compound_malformed(file => $file,
                                 ipaddr => $ipaddr,
                                 operation => 'invalid_op'|'renamef'|'linkf'|
                                              'restorefh'|'truncated_op'
                                 [, nfile => $nfile ]
                                 [, opcode => $opcode ]
                                 [, whence => $opcode ]
                                 [, minor_ver => $minor_ver]);

Issues a malformed compound request according to the 'operation' parameter.

Issues a compound request for export regime transition.
It transitions from the filehandle of the 'file' parameter to another
filehandle via PUTROOTFH, PUTPUBFH, PUTFH, RESTOREFH, LOOKUP and LOOKUPP.
For PUTFH and RESTOREFH, it transitions to the filehandle of the 'nfile'
parameter. For LOOKUPP, the transitions if from filehandle of the 'nfile'
parameter to the filehandle of the 'file' parameter.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Full path of file to set the current filehandle.

=item ipaddr

(Required) : IP address of filer.

=item operation

(Optional) : Operation to issue:

    invalid_op   => Send a request with invalid op number,
    renamef      => Performs a RENAME with no saved filehandle,
    linkf        => Performs a LINK with no saved filehandle,
    restorefh    => Performs a RESTOREFH with no saved filehandle,
    truncated_op => Send a request that has been truncated.
                    The LOOKUP operation is truncated.

=item nfile

(Optional) : Target file for operations: 'renamef', 'linkf' or 'restorefh'.

=item opcode

(Optional) : Opcode to send when operation is 'invalid_op'.
This is required when operation is 'invalid_op'.

=item whence

(Optional) : Specifies where the truncation takes place within the opcode:
'start', 'middle' or 'end'.
This is required when operation is 'truncated_op'.

=item minor_ver

(Optional) : Minor version of NFSv4 to be tested.
Default is 0.

=back

=item Returns

A reference to a hash of the results of each operation performed.
The key is the operation name and the value is the result of the
compound request as returned from L<C<send_compound>|/send_compound>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound_malformed(%) {
    my (%opts) = validate (
        @_,
        {
            operation => { type => SCALAR, regex =>
                qr/^(invalid_op|renamef|linkf|restorefh|truncated_op)$/ },
            nfile     => { type => SCALAR, optional => 1 },
            opcode    => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
            whence    => { type => SCALAR, optional => 1,
                          regex => qr/^(start|middle|end)$/ },
            minor_ver => { type => SCALAR, default => 0 },
            %common_parameters
        }
    );
    my ($nfs, @ops, @op_list, @names, $item, $out, %args, $arg, %results);
    my ($file, $nfile);
    my $tag = 'malform compound';
    my $minor_ver = $opts{minor_ver};

    # Check required parameters depending on type of operation
    if ( $opts{operation} eq 'invalid_op' ) {
        croak "Need parameter 'opcode' for operation 'invalid_op'"
          unless defined $opts{opcode};
    }
    if ( $opts{operation} eq 'truncated_op' ) {
        croak "Need parameter 'whence' for operation 'truncated_op'"
          unless $opts{whence};
    }
    if ( $opts{operation} =~ /^(renamef|linkf|restorefh)$/ ) {
        croak "Need parameter 'nfile' for operation '$opts{operation}'"
          unless $opts{nfile};
    }

    my $f = host($main::FILER);
    if ($f->hosttype =~ /filer-ng/) {
	$opts{ipaddr} = get_data_ip_cmode($main::FILER);
    }
    if ( $main::ADDR_TYPE eq 'inet6' ) {
	$opts{ipaddr} = $f->net_best_address(sort_preference=>{family=>'inet6'});
        logcomment("IPv6 address used for synthetic client: $opts{ipaddr}");
    }
    # Create NFS object
    $nfs = nfs::NfsClient->new(
        hostname => $opts{ipaddr},
        version  => 4,
        proto    => 'tcp',
        minor_version  => $minor_ver,
    );

    # Get filehandle for the file
    $nfs->mount(path => $opts{file});

    # Remove leading '/'
    $opts{file} =~ s{^/}{};
    @names = split /\//, $opts{file};

    # Get filenames
    if ( $opts{operation} =~ /^(renamef|linkf|restorefh)$/ ) {
        $file = pop @names if $opts{operation} eq 'renamef';
        $nfile = $opts{nfile};
        $nfile =~ s{.*/}{};
    }

    # Create compound request:
    #   PUTROOTFH to get the filehandle of the root
    @ops = ( { op => OP_PUTROOTFH } );
    #   traverse the path using LOOKUP
    push @ops, map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
      @names;

    # Create list of operations to send according to requested operation
    if ( $opts{operation} eq 'invalid_op' ) {
        # List of invalid operation numbers
        @op_list = (
            {
                name => $opts{opcode},
                ops  => [ { op => $opts{opcode} } ],
            },
        );
    } elsif ( $opts{operation} eq 'renamef' ) {
        # RENAME with no saved fh
        @op_list = (
            {
                name => 'RENAME',
                ops  => [
                   {
                       op  => OP_RENAME,
                       arg => xdr_opaque(encode => $file) .
                              xdr_opaque(encode => $nfile),
                   }
                ],
            },
        );
    } elsif ( $opts{operation} eq 'linkf' ) {
        # LINK with no saved fh
        @op_list = (
            {
                name => 'LINK',
                ops  => [
                   {
                       op  => OP_LINK,
                       arg => xdr_opaque(encode => $nfile),
                   }
                ],
            },
        );
    } elsif ( $opts{operation} eq 'restorefh' ) {
        # RESTOREFH with no saved fh
        @op_list = (
            {
                name => 'RESTOREFH',
                ops  => [ { op  => OP_RESTOREFH } ],
            },
        );
    } elsif ( $opts{operation} eq 'truncated_op' ) {
        # Set the number of operations to at least 3 more than
        # the actual number of operations to be sent
        %args = (num => @ops + 4);
        if ( $opts{whence} eq 'start' ) {
            @op_list = (
                {
                    name => 'LOOKUP',
                    ops  => [],
                },
            );
        } elsif ( $opts{whence} eq 'middle' ) {
            @op_list = (
                {
                    name => 'LOOKUP',
                    ops  => [ { op => OP_LOOKUP } ],
                },
            );
            # Encode the compound request arguments to get size to truncate
            $arg = xdr_compound(
                tag => $tag,
                ops => [ @ops, @{$op_list[0]{ops}} ],
            );
            # Truncate request just in the middle of the operation code
            $args{size} = length($arg) - 2;
        } elsif ( $opts{whence} eq 'end' ) {
            # Send opcode but no arguments
            @op_list = (
                {
                    name => 'LOOKUP',
                    ops  => [ { op => OP_LOOKUP } ],
                },
            );
        }
    }

    # Send all compound requests
    foreach $item ( @op_list ) {
        # Send compound request
        $out = send_compound(
            nfs => $nfs,
            ops => [ @ops, @{$item->{ops}} ],
            tag => $tag,
            %args,
        );
        $results{$item->{name}} = $out;
    }

    return \%results;
}

#---------------------------------------------------------------------

=head2 compound_fh

  $returned = compound_fh(file => $file,
                          nfile => $nfile,
                          ipaddr => $ipaddr
                          operation => 'renamef'|'linkf'|'restorefh'
                          [, file_attr_dir => 0|1 ]
                          [, file_attr => 0|1 ]
                          [, nfile_attr_dir => 0|1 ]
                          [, nfile_attr => 0|1 ]
                          [, minor_ver => 0|1 ]);

Issues a compound request with current and saved file handles given
by the 'file' and 'nfile' parameters respectively.

The following named arguments are accepted:

=over

=item Arguments

=over

=item file

(Required) : Full path of file to set the current filehandle.

=item nfile

(Required) : Full path of file to set the saved filehandle.

=item ipaddr

(Required) : IP address of filer.

=item operation

(Required) : Operation to issue:

    renamef      => Performs a RENAME,
    linkf        => Performs a LINK,
    restorefh    => Performs a RESTOREFH.

=item file_attr_dir

(Optional) : If true, 'file' is a named attribute directory.
Default is 0.

=item file_attr

(Optional) : If true, 'file' is a named attribute.
Default is 0.

=item nfile_attr_dir

(Optional) : If true, 'nfile' is a named attribute directory.
Default is 0.

=item nfile_attr

(Optional) : If true, 'nfile' is a named attribute.
Default is 0.

=item link_name

(Optional) : The name to use for the link destination.
Default is $base.link.

=item optag

(Optional) : The tag name to use for the compound request so that it
is easy to tie packet traces back to the testcase.
Default is 'fh compound'.

=item minor_ver

(Optional) : Minor version of NFSv4 to be tested.
Default is 0.

=back

=item Returns

A reference to a hash of the results of each operation performed.
The key is the operation name and the value is the result of the
compound request as returned from L<C<send_compound>|/send_compound>.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound_fh(%) {
    my (%opts) = validate (
        @_,
        {
            nfile          => { type => SCALAR },
            operation      => { type => SCALAR,
                               regex => qr/^(renamef|linkf|restorefh)$/ },
            file_attr_dir  => { type => SCALAR,  default => 0 },
            file_attr      => { type => SCALAR,  default => 0 },
            nfile_attr_dir => { type => SCALAR,  default => 0 },
            nfile_attr     => { type => SCALAR,  default => 0 },
            link_name      => { type => SCALAR,  default => "$base.link" },
            optag          => { type => SCALAR,  default => 'fh compound' },
            minor_ver      => { type => SCALAR,  default => 0 },
            %common_parameters
        }
    );
    my ($nfs, @names, @names_n, $named_attr, $named_attr_n, $out);
    my (@ops, @op_file, @op_nfile);
    my $path   = $opts{file};
    my $path_n = $opts{nfile};
    my %map_h = (
        renamef   => 'RENAME',
        linkf     => 'LINK',
        restorefh => 'RESTOREFH',
    );

    # Remove leading '/'
    $opts{file} =~ s{^/}{};
    @names = split /\//, $opts{file};

    # Remove leading '/'
    $opts{nfile} =~ s{^/}{};
    @names_n = split /\//, $opts{nfile};

    if ( $opts{file_attr} ) {
        # Last entry is the named attribute
        $named_attr = pop @names;
        $path = '/' . join('/', @names);
    }

    if ( $opts{nfile_attr} ) {
        # Last entry is the named attribute
        $named_attr_n = pop @names_n;
        $path_n = '/' . join('/', @names_n);
    }

    my $f = host($main::FILER);
    if ( $main::ADDR_TYPE eq 'inet6' ) {
        $opts{ipaddr} = $f->net_best_address(sort_preference=>{family=>'inet6'});
        logcomment("IPv6 address used for synthetic client: $opts{ipaddr}");
    }
    if ($f->hosttype =~ /filer-ng/) {
	$opts{ipaddr} = get_data_ip_cmode($main::FILER);
    }

    # Create NFS object
    $nfs = nfs::NfsClient->new(
        hostname => $opts{ipaddr},
        version  => 4,
        proto    => 'tcp',
        minor_version  => $opts{minor_ver},
    );

    # Get filehandle for the file
    $nfs->mount(path => $path);

    # Traverse the path using LOOKUP for both files
    @op_file = map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
      @names;
    @op_nfile = map { { op => OP_LOOKUP, args => xdr_opaque(encode=>$_) } }
      @names_n;

    # Get to the named attribute directory/file
    push @op_file, { op => OP_OPENATTR, args => xdr_uint(encode => 0) }
      if $opts{file_attr_dir} || $opts{file_attr};
    push @op_file, { op => OP_LOOKUP,
                     args => xdr_opaque(encode => $named_attr) }
      if $opts{file_attr};

    push @op_nfile, { op => OP_OPENATTR, args => xdr_uint(encode => 0) }
      if $opts{nfile_attr_dir} || $opts{nfile_attr};
    push @op_nfile, { op => OP_LOOKUP,
                      args => xdr_opaque(encode => $named_attr_n) }
      if $opts{nfile_attr};

    # Create compound request:
    #   PUTROOTFH to get the filehandle of the root
    @ops = (
        { op => OP_PUTROOTFH },
        @op_nfile,
        { op => OP_SAVEFH },
        { op => OP_PUTROOTFH },
        @op_file,
    );

    if ( $opts{operation} eq 'restorefh' ) {
        push @ops, { op => OP_RESTOREFH };
    } elsif ( $opts{operation} eq 'renamef' ) {
        push @ops, {
            op   => OP_RENAME,
            args => xdr_opaque(encode => "$base.rename") .
                    xdr_opaque(encode => "$base.newname"),
        };
    } elsif ( $opts{operation} eq 'linkf' ) {
        push @ops, {
            op   => OP_LINK,
            args => xdr_opaque(encode => $opts{link_name}),
        };
    }

    # Send compound request
    $out = send_compound(
        nfs => $nfs,
        ops => \@ops,
        tag => $opts{optag},
    );

    return { $map_h{$opts{operation}} => $out };
}

#---------------------------------------------------------------------

=head2 send_compound

  $out = send_compound(nfs => $nfs_obj,
                       ops => \@ops
                       [, tag => $tag ]
                       [, num => $num ]
                       [, size => $size ]);

Sends a compound request on NFS version 4. The compound request is made
of all operations listed on the ops argument. Each operation is a hash
having the following:

  {
      op   => Operation number,
      args => XDR encoded arguments,
  }

The following named arguments are accepted:

=over

=item Arguments

=over

=item nfs

(Required) : A L<C<nfs::NfsClient>|engineering::ontap_tests::agnostic::docs::lib-nfs-NfsClient-pm/> object.

=item ops

(Required) : Reference to an array of operations to be included in the
compound request.

=item tag

(Optional) : Tag for the compound request. Default is 'compound'.

=item num

(Optional) : Use this as the number of operations to encode instead
of the actual number of operations encoded.

=item size

(Optional) : Truncate the arguments to the compound procedure to
this size.

=back

=item Returns

A reference to a hash of the results of the compound request.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub send_compound(%) {
    my (%opts) = validate (
        @_,
        {
            nfs  => { type => OBJECT },
            ops  => { type => ARRAYREF },
            tag  => { type => SCALAR,  default => 'compound' },
            num  => { type => SCALAR, optional => 1 },
            size => { type => SCALAR, optional => 1 },
        }
    );
    my ($out, $status, $result, $err);
    my $nfs = $opts{nfs};
    delete $opts{nfs};

    $out = eval {
        $nfs->compound(%opts);
    };
    $err = $@;

    if ( !$err || $err =~ /^COMPOUND/ ) {
        # Get failed results from the NFS object if error from server
        $result = $err ? $nfs->result : $out;
        $status = _get_nfs_error($result);
        $result->{$_} = $status->{$_} foreach keys %$status;
    } elsif ( $err ) {
        # Unknown error
        $result = $err;
    }

    return $result;
}

#---------------------------------------------------------------------
# Internal function to create a file
#---------------------------------------------------------------------
sub _create_file($$$;%) {
    my ($nfs, $dhandle, $file, %opts) = @_;
    my $out;
    $opts{mode} = 0644 if !exists $opts{mode};

    $out = $nfs->open(
        dhandle => $dhandle,
        name    => $file,
        access  => OPEN4_SHARE_ACCESS_WRITE,
        cr_mode => UNCHECKED,
        attr    => { FATTR4_MODE() => $opts{mode} },
    );
    $nfs->close(stateid => $out->{stateid}) unless $opts{noclose};
    return $out;
}

#---------------------------------------------------------------------
# Internal function to create a directory
#---------------------------------------------------------------------
sub _create_dir($$$;%) {
    my ($nfs, $dhandle, $dir, %opts) = @_;
    my ($out, %attr_args);
    $opts{mode} = 0755 if !exists $opts{mode};
    %attr_args = ( FATTR4_MODE() => $opts{mode} );
    %attr_args = ( FATTR4_ACL()  => $opts{acl} ) if $opts{acl};

    $out = $nfs->create(
        dhandle => $dhandle,
        name    => $dir,
        type    => NFDIR,
        attr    => \%attr_args,
    );

    return $out;
}

#---------------------------------------------------------------------
# Internal function to remove a directory tree
#---------------------------------------------------------------------
sub _remove_file($$$;) {
    my ($nfs, $dhandle, $name) = @_;
    my ($out, $fhandle, $item);

    # Check if file/directory exists
    eval { $nfs->lookup(dhandle => $dhandle, name => $name) };
    return if $@ =~ /NFS4ERR_NOENT/;
    croak $@ if $@;

    # Save filehandle of object to remove
    $fhandle = $nfs->curr_fh;

    # Get the file type of object
    $out = $nfs->getattr(attr => [FATTR4_TYPE]);

    if ( $out->{fattr}{FATTR4_TYPE()} == NFDIR ) {
        # This is a directory -- make it writable
        $nfs->setattr(attr => { FATTR4_MODE() => 0777 });

        # Get directory entries
        $out = $nfs->readdir(attr => [FATTR4_TYPE]);

        foreach $item ( @{$out->{list}} ) {
            if ( $item->{fattr}{attr_vals}{FATTR4_TYPE()} == NFDIR ) {
                # Recursively remove this directory
                _remove_file($nfs, $fhandle, $item->{name});
            } else {
                # Remove this file on current directory
                $nfs->remove(dhandle => $fhandle, name => $item->{name});
            }
        }
    }

    # Remove file/directory
    $nfs->remove(dhandle => $dhandle, name => $name);

    return 1;
}

#---------------------------------------------------------------------
# Internal function to convert the NFS status code to readable strings
#---------------------------------------------------------------------
sub _get_nfs_error($;) {
    my $result = shift;
    my (%ret, $out);

    $out = nfs_error(error => $result->{status}, version => 4);
    $ret{status_fullmsg} = $out;
    $out =~ /([^:]+):\s+(.*)/;
    $ret{status_code} = $1;
    $ret{status_mesg} = $2;

    return \%ret;
}

#---------------------------------------------------------------------
# Internal function to create an nfs::NfsClient object and mount
# the path given and change the permissions to 777 as root
#---------------------------------------------------------------------
sub _nfs_mount(%) {
    my %opts = @_;

    # Create NFS object
    my $nfs = nfs::NfsClient->new(
        hostname => $opts{host},
        version  => 4,
        proto    => 'tcp',
    );

    # Mount volume
    $nfs->mount(path => $opts{path});

    # Change permission to 777 of mount point as root
    $nfs->rpc->auth_sys_set(uid => 0, gid => 0);
    $nfs->setattr(attr => {FATTR4_MODE() => 0777});
    $nfs->rpc->auth_sys_set();

    return $nfs;
}

#--------------------------------------------------------------------------------------------------------------
# Function to return array of minor versions of NFSv4 to be tested depending upon value passed to NFS4_MINOR_VER
#---------------------------------------------------------------------------------------------------------------

sub validate_minor_ver (%) {

    my %opts = @_;

    my $ng_check;
    my $f = host ($main::FILER);
    if ($f->hosttype =~ /filer-ng/){
        $ng_check = 1;
    }
    if ($opts{minor_ver} eq 'both'){

        if ($ng_check) {return (0, 1);}
        return;
    }
    elsif ($opts{minor_ver} == 0){
        return (0);
    }
    elsif ($opts{minor_ver} == 1){
        if ($ng_check) {return (1);}
        return;
    }
    else {
        return;
    }
}

my $SERVER_OBJ;
sub _get_server_node() {
    if (!$SERVER_OBJ) {
        $SERVER_OBJ = new Server(node=>$::FILER);
    }
    return $SERVER_OBJ;
}

1;
