# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Contents.pm#1 $
#
## @summary     Contents - part of the Flg library
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Contents - library functions specifically for use by the Flg library

=head1 SYNOPSIS

  use nfs::Contents;

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

This package is intended for use by the nfs::Flg.pm library only. Do not
call nfs::Contents.pm directly.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Contents.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Contents;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Distheap;

use Carp qw(croak);
use Carp::Assert;
use Params::Validate qw(validate SCALAR HASHREF SCALARREF);
use Time::HiRes qw(gettimeofday);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
use constant NAMESET_ENTRY_MAX_FHSIZE => 64;
use constant CKSUM_LEN => 32;

our @EXPORT_OK = qw(
    contents_alloc
    contents_setfh
    contents_lookup_unsafe
    contents_lookup
    contents_flush_all
    contents_new_write
    contents_fill_write
    contents_new_read
    contents_verify_read
    NAMESET_ENTRY_MAX_FHSIZE
);
#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------

sub contents_init(%;);
sub contents_alloc(%;);
sub contents_setfh(%;);
sub contents_lookup_unsafe(%;);
sub contents_sanity(%;);
sub _contents_sanity(%;);
sub contents_lookup(%;);
sub contents_flush_all(%;);
sub contents_new_write(%;);
sub clock_now();
sub contents_generate_datablock(%;);
sub contents_cksum(%;);
sub _contents_pre_update_range(%;);
sub contents_range_lookup(%;);
sub contents_range_split(%;);
sub contents_range_alloc(%;);
sub contents_range_new(%;);
sub contents_fill_write(%;);
sub contents_new_read(%;);
sub contents_verify_read(%;);
sub contents_verify_one(%;);
sub testme();

#--------------------------------------------------------------------------------
# Package local variables
#--------------------------------------------------------------------------------

my $dh_contents;
my $dh_contents_ranges;

#--------------------------------------------------------------------------------

=head2 contents_init

  $contents = nfs::Contents->contents_init( maxfiles => $maxfiles );

Initializes a contents object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item maxfiles

(Required) : Maximum number of files represented by the contents heap

=back

=item Returns

The contents object

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_init(%;) {
    #   /*
    #    * Initialize the contents heap, allocating twelve
    #    *   contents ranges per file to the heap.
    #    * XXO: size upwards as necessary?
    #    */
    #   int
    #   contents_init(int maxfiles)
    #   {
    #       if ((dh_contents =
    #           distheap_init(maxfiles, sizeof (struct contents))) == NULL) {
    #               report_error(FATAL, "distheap_init error");
    #               return -1;
    #       }
    #       if ((dh_contents_ranges =
    #           distheap_init(12*maxfiles,
    #           sizeof (struct contents_range))) == NULL) {
    #               report_error(FATAL, "distheap_init error");
    #               return -1;
    #       }
    #       return 0;
    #   }
    #

    my $class  = shift;
    my (%opts) = validate (
        @_,
        {
            maxfiles => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );
    my $self = bless { %opts }, $class;

    $self->{dh_contents} =
            nfs::Distheap->distheap_init( maxsize => $self->{maxfiles},
                                          sizeof  => 'contents' );
    $self->{dh_contents_ranges} =
            nfs::Distheap->distheap_init( maxsize => $self->{maxfiles} * 12,
                                          sizeof  => 'contents_range' );

    $dh_contents        = $self->{dh_contents};
    $dh_contents_ranges = $self->{dh_contents_ranges};

    return $self;
} # End contents_init

#--------------------------------------------------------------------------------

=head2 contents_alloc

  $ct = contents_alloc( nse => $nse );

contents_alloc function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry for which to allocate contents

=back

=item Returns

The allocated contents entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

{
    my $next_id = 1;

    sub contents_alloc(%;) {
        #   contents_t
        #   contents_alloc(nameset_entry_t nse)
        #   {
        #       static u_int64_t next_id = 1;
        #       contents_t ct = distheap_alloc(dh_contents, 1);
        #
        #       if (ct == NULL) {
        #           report_error(FATAL, "distheap_alloc error");
        #           return NULL;
        #       }
        #
        #       memset(ct, 0x88, sizeof(*ct));
        #       Q_INIT(&ct->ct_crlist);
        #       assert(nse->fhlen <= NAMESET_ENTRY_MAX_FHSIZE);
        #       memcpy(ct->ct_fhdata, nse->fhdata, nse->fhlen);
        #       ct->ct_fhlen = nse->fhlen;
        #       ct->ct_type = nse->type;
        #       ct->ct_id = next_id++;
        #       nse->ct_idx = distheap_toidx(dh_contents, ct);
        #       return ct;
        #   }
        #

        my (%opts) = validate (
                @_,
                {
                    nse => { type => HASHREF },
                }
        );

        # $next_id is static
        my ( $ct, $index ) = $dh_contents->distheap_alloc( weight => 1 );

        # $ct would have been a pointer to the u.data segment of the
        # contents structure, so reassign it so that it is
        $ct = $dh_contents->{entries}{$index}{data};

        # Q_INIT( $ct->{ct_crlist} );
        $ct->{ct_crlist_head} = 0;
        $ct->{ct_crlist_tail} = 0;

        $opts{nse}{fhlen} = 0 if ( !defined $opts{nse}{fhlen} );
        assert( $opts{nse}{fhlen} <= NAMESET_ENTRY_MAX_FHSIZE );

        $ct->{ct_fhlen} = $opts{nse}->{fhlen};
        $ct->{ct_type}  = $opts{nse}->{type};
        $ct->{ct_id}    = $next_id++;

        $opts{nse}->{ct_idx} = $index;

        return $ct;
    } # End contents_alloc
}

#--------------------------------------------------------------------------------

=head2 contents_setfh

  contents_setfh( nse => $opts{nse} );

contents_setfh function taken from the original C code

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry for which the contents filehandle will be set

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_setfh(%;) {
    #   void
    #   contents_setfh(nameset_entry_t nse)
    #   {
    #       contents_t ct = contents_lookup_unsafe(nse);
    #       assert(nse->fhlen <= NAMESET_ENTRY_MAX_FHSIZE);
    #       memcpy(ct->ct_fhdata, nse->fhdata, nse->fhlen);
    #       ct->ct_fhlen = nse->fhlen;
    #   }
    #

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );

    my $ct = contents_lookup_unsafe( nse => $opts{nse} );

    assert( $opts{nse}->{fhlen} <= NAMESET_ENTRY_MAX_FHSIZE );

    $ct->{ct_fhdata} = $opts{nse}->{fhdata};
    $ct->{ct_fhlen}  = $opts{nse}->{fhlen};

} # End contents_setfh

#--------------------------------------------------------------------------------

=head2 contents_lookup_unsafe

  $ct = contents_lookup_unsafe( nse => $nse );

contents_lookup_unsafe function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry for which contents will be looked up

=back

=item Returns

A contents entry for the specified nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_lookup_unsafe(%;) {
    #  contents_t
    #  contents_lookup_unsafe(nameset_entry_t nse)
    #  {
    #      contents_t ct = distheap_fromidx(dh_contents, nse->ct_idx);
    #      assert(ct != NULL);
    #      contents_sanity(ct);
    #      return ct;
    #  }
    #
    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );

    my $idx = $opts{nse}->{ct_idx};
    my $ct  = $dh_contents->{entries}{$idx}{data};

    assert( defined $ct );

    contents_sanity( ct => $ct );

    return $ct;

} # End contents_lookup_unsafe

#--------------------------------------------------------------------------------

=head2 contents_sanity

  contents_sanity( ct => $ct );

contents_sanity function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_sanity(%;) {
    #   static void
    #   contents_sanity(contents_t ct)
    #   {
    #       _contents_sanity(ct, 0);
    #   }

    my (%opts) = validate (
        @_,
        {
            ct => { type => HASHREF },
        }
    );

    _contents_sanity( ct        => $opts{ct},
                      allow_new => 0 );

} # End contents_sanity

#--------------------------------------------------------------------------------

=head2 _contents_sanity

  _contents_sanity( ct        => $ct,
                    allow_new => 0 );

_contents_sanity function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry

=item allow_new

(Required) : Allow new flag (C code always sets 0)

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub _contents_sanity(%;) {
    #   #define Q_DH_FOREACH(X, Q, F, H)                     \
    #       for (X = Q_DH_FIRST(Q, H); X; X = distheap_fromidx(H, X->F.next))
    #
    #   #define Q_DH_FIRST(Q, H)                             \
    #       distheap_fromidx(H, (Q)->head)
    #

    #   static void
    #   _contents_sanity(contents_t ct, int allow_new)
    #   {
    #       contents_range_t cr;
    #       u_int64_t last_end = 0;
    #
    #       Q_DH_FOREACH(cr, &ct->ct_crlist, cr_crlink, dh_contents_ranges) {
    #           /* In order and non-overlapping */
    #           assert(last_end <= cr->cr_begin);
    #
    #           /* Non-trivial ranges. */
    #           assert(cr->cr_begin < cr->cr_end);
    #
    #           if (!allow_new) {
    #               /* Someone dropped the ball. */
    #               assert(!cr->cr_new);
    #           } else if (!cr->cr_new) {
    #               /* Pattern covers the entire range. */
    #               assert(cr->cr_fill_begin <= cr->cr_begin);
    #           }
    #
    #           last_end = cr->cr_end;
    #       }
    #   }
    #
    #   static void
    #   contents_sanity(contents_t ct)
    #   {
    #       _contents_sanity(ct, 0);
    #   }
    #

    my (%opts) = validate (
        @_,
        {
            ct        => { type => HASHREF },
            allow_new => { type => SCALAR, default => 0 },
        }
    );
    my $cr;
    my $last_end = 0;

    $opts{ct}->{ct_crlist_head} = 0 if ( !defined $opts{ct}->{ct_crlist_head} );
    my $ct_crlist_head = $opts{ct}->{ct_crlist_head};
    for ( $cr = $dh_contents_ranges->{entries}{$ct_crlist_head}{data};
            defined $cr;
            $cr = $dh_contents_ranges->{entries}{$cr->{cr_crlink_next}}{data} ) {
        assert( $last_end <= $cr->{cr_begin} );

        assert( $cr->{cr_begin} < $cr->{cr_end} );

        if ( !$opts{allow_new} ) {
            assert( !$cr->{cr_new} );
        } elsif ( !$cr->{cr_new} ) {
            assert( $cr->{cr_fill_begin} <= $cr->{cr_begin} );
        }

        $cr->{cr_end} = 0 if ( !defined $cr->{cr_end} );
        $last_end     = $cr->{cr_end};
    }

} # End _contents_sanity

#--------------------------------------------------------------------------------

=head2 contents_lookup

  $ct = contents_lookup( nse => $sibling );

contents_lookup function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=back

=item Returns

A contents entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_lookup(%;) {
    #   contents_t
    #   contents_lookup(nameset_entry_t nse)
    #   {
    #       contents_t ct;
    #
    #       ct = contents_lookup_unsafe(nse);
    #       assert(ct->ct_fhlen == nse->fhlen);
    #       assert(memcmp(ct->ct_fhdata, nse->fhdata, nse->fhlen) == 0);
    #       return ct;
    #   }
    #

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF }
        }
    );
    my $nse = $opts{nse};
    my $ct  = contents_lookup_unsafe( nse => $nse );

    assert( $ct->{ct_fhlen}  == $nse->{fhlen} );
    $ct->{ct_fhdata}         =  '' if ( !defined $ct->{ct_fhdata} );
    $nse->{fhdata}           =  '' if ( !defined $nse->{fhdata} );
    assert( $ct->{ct_fhdata} eq $nse->{fhdata} );

    return $ct;
} # End contents_lookup

#--------------------------------------------------------------------------------

=head2 contents_flush_all

  contents_flush_all( nse => $nse );

contents_flush_all function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_flush_all(%;) {
    #   void
    #   contents_flush_all(nameset_entry_t nse)
    #   {
    #       contents_t ct;
    #       contents_range_t cr;
    #
    #       if (!(ct = contents_lookup_unsafe(nse)))
    #           return;
    #
    #       CT_DBG(ct, ("flush all\n"));
    #
    #       while ((cr = Q_DH_FIRST(&ct->ct_crlist, dh_contents_ranges)) != NULL) {
    #           Q_DH_REMOVE(&ct->ct_crlist, cr, cr_crlink, dh_contents_ranges, contents_range_t);
    #              memset(cr, 0xaa, sizeof(*cr));
    #              distheap_dele(dh_contents_ranges, cr);
    #       }
    #       memset(ct, 0x99, sizeof(*ct));
    #       distheap_dele(dh_contents, ct);
    #   }

    #   #define Q_DH_FIRST(Q, H)                             \
    #       distheap_fromidx(H, (Q)->head)

    #   #define Q_DH_LAST(Q, H)                              \
    #       distheap_fromidx(H, (Q)->tail)

    #   #define Q_DH_REMOVE(Q, X, F, H, Tp) {                \
    #       if (Q_DH_FIRST(Q, H) == (X)) {               \
    #           (Q)->head = (X)->F.next;             \
    #       }                                            \
    #       if (Q_DH_LAST(Q, H) == (X)) {                \
    #           (Q)->tail = (X)->F.prev;             \
    #       }                                            \
    #       if ((X)->F.prev) {                           \
    #           Tp X_prev;                   \
    #           X_prev = distheap_fromidx(H, X->F.prev);\
    #           X_prev->F.next = (X)->F.next;        \
    #       }                                            \
    #       if ((X)->F.next) {                           \
    #           Tp X_next;                   \
    #           X_next = distheap_fromidx(H, X->F.next);\
    #           X_next->F.prev = (X)->F.prev;        \
    #       }                                            \
    #   }

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF }
        }
    );

    my $ct  = contents_lookup_unsafe( nse => $opts{nse} );

    return if ( !defined $ct );

    # CT_DBG defined as null op in original application

    my $cr =
      $dh_contents_ranges->distheap_fromidx( idx => $ct->{ct_crlist_head} );

    while ( defined $cr ) {
        # Q_DH_REMOVE
        if ( $dh_contents_ranges->distheap_fromidx(
                idx => ( $ct->{ct_crlist_head} == $cr ) ) ) {
            $ct->{ct_crlist_head} = $cr->{cr_crlink_next};
        }
        if ( $dh_contents_ranges->distheap_fromidx(
                idx => ( $ct->{ct_crlist_tail} == $cr ) ) ) {
            $ct->{ct_crlist_tail} = $cr->{cr_crlink_prev};
        }
        if ( $cr->{cr_crlink_prev} ) {
            my $X_prev = $dh_contents_ranges->distheap_fromidx(
                    idx => $cr->{cr_crlink_prev} );
            $X_prev->{cr_crlink_next} = $cr->{cr_crlink_next};
        }
        if ( $cr->{cr_crlink_next} ) {
            my $X_next = $dh_contents_ranges->distheap_fromidx(
                    idx => $cr->{cr_crlink_next} );
            $X_next->{cr_crlink_prev} = $cr->{cr_crlink_prev};
        }

        # memset(cr, 0xaa, sizeof(*cr));
        # Set the memory to a junk value handled by distheap_dele
        $dh_contents_ranges->distheap_dele( hd => $cr );

        # Get the next $cr
        $cr =
          $dh_contents_ranges->distheap_fromidx( idx => $ct->{ct_crlist_head} );
    }

    # memset(ct, 0x99, sizeof(*ct));
    # Set the memory for $ct to a junk value?
    # handled by distheap_dele
    $dh_contents->distheap_dele( hd => $ct );

} # End contents_flush_all

#--------------------------------------------------------------------------------

=head2 contents_new_write

  $data = contents_new_write( nse              => $nse,
                              cookie           => \$contents_cookie,
                              off              => $offset,
                              cnt              => $count,
                              max_payload_size => MAX_PAYLOAD_SIZE );

contents_new_write function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=item cookie

(Required) : The contents cookie

=item off

(Required) : The offset at which to start the write

=item cnt

(Required) : The number of bytes to be written

=item max_payload_size

(Required) : The maximum size of the write data

=back

=item Returns

The data written

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_new_write(%;) {
    #   void
    #   contents_new_write(nameset_entry_t nse, contents_write_cookie_t *cookie,
    #           char **data, u_int64_t off, u_int32_t cnt)
    #   {
    #       int ndblocks = cnt / sizeof(contents_datablock_t);
    #       ssize_t rem = cnt - ndblocks * sizeof(contents_datablock_t);
    #       u_int64_t begin = off;
    #       u_int64_t end = off + cnt;
    #       u_int64_t last_end;
    #       contents_datablock_t db;
    #       contents_datablock_t *dst;
    #       contents_range_t cr;
    #       contents_t ct;
    #       int ii;
    #
    #       cookie->when_issued = clock_now();
    #
    #       /* Generate pattern. */
    #       cookie->fill_generation = cookie->when_issued;
    #       contents_generate_datablock(&db, nse, off, cnt,
    #               cookie->fill_generation);
    #       cookie->fill_cksum = db.cksum;
    #
    #       /* Fill data buffer. */
    #       assert(cnt <= MAX_PAYLOAD_SIZE);
    #       *data = malloc(cnt);
    #       dst = (contents_datablock_t*) *data;
    #       for (ii = 0; ii < ndblocks; ii++, dst++) {
    #           memcpy(dst, &db, sizeof(contents_datablock_t));
    #       }
    #       if (rem) {
    #           assert(rem > 0);
    #           memcpy(dst, &db, rem);
    #       }
    #
    #       /*
    #        * Accept some slop, in case op generator wants to generate
    #        * zero-length reads/writes.
    #        */
    #       if ( cnt == 0 )
    #           return;
    #  #    if ( nse->dele_when_free || nse->rename ) {
    #           /* File removed! Do NOT create a new contents obj. */
    #           // assert( !contents_lookup( nse ) ); // see contents_fill_write
    #           assert( nse->dele_when_free || !nse->go_away );
    #           return;
    #       }
    #
    #       ct = contents_lookup( nse );
    #       assert( ct != NULL );
    #
    #       CT_DBG( ct, ( "new_write %llu -> %llu gen %llu\n", begin, end,
    #              cookie->fill_generation ) );
    #
    #       /* Notify that a write is pending for the write range. */
    #       cr = _contents_pre_update_range( ct, begin, end );
    #       assert( cr->cr_begin == begin );
    #       last_end = begin;
    #       for ( ; cr && cr->cr_end <= end;
    #               cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next ) ) {
    #           assert( last_end == cr->cr_begin );
    #
    #           if ( cr->cr_busycnt == 0 ) {
    #               cr->cr_fill_begin = off;
    #               cr->cr_fill_generation = cookie->fill_generation;
    #               cr->cr_fill_cksum = cookie->fill_cksum;
    #               assert( !cr->cr_new || !cr->cr_invalid );
    #               cr->cr_new = 0;
    #               cr->cr_invalid = 0;
    #           } else {
    #               assert( !cr->cr_new );
    #           }
    #
    #           if ( ++cr->cr_busycnt > 1 ) {
    #               /*
    #                * Overlapping writes.  Cannot do data
    #                * validation, since we do not know which one
    #                * will be committed to the filesystem.
    #               */
    #               cr->cr_invalid = 1;
    #           }
    #
    #           last_end = cr->cr_end;
    #       }
    #       assert( last_end == end );
    #
    #       contents_sanity( ct );
    #   }

    my (%opts) = validate (
       @_,
       {
           nse              => { type => HASHREF },
           cookie           => { type => SCALARREF },
           off              => { type => SCALAR },
           cnt              => { type => SCALAR },
           max_payload_size => { type => SCALAR }
       }
    );
    my $nse = $opts{nse};
    my $cookie = $opts{cookie};
    my $off = $opts{off};
    my $cnt = $opts{cnt};

    my $begin = $off;
    my $end   = $off + $cnt;

    my $data  = '';

    $$cookie->{when_issued} = clock_now();

    $$cookie->{fill_generation} = $$cookie->{when_issued};
    my $db = contents_generate_datablock( nse => $nse,
                                          off => $off,
                                          cnt => $cnt,
                                          fill_generation =>
                                            $$cookie->{fill_generation} );
    $$cookie->{fill_cksum} = $db->{cksum};

    # Use only fill_generation, fhdata, and cksum for compatibility
    # with contents_verify_read
    my $data_string = "$db->{fill_generation}" .
                      "$db->{fhdata}" . "$db->{cksum}";
    my $data_length = do { use bytes; length $data_string };

    my $ndblocks = int ( $cnt / $data_length );
    my $rem      = $cnt - ( $ndblocks * $data_length );

    # Fill data buffer
    assert( $cnt <= $opts{max_payload_size} );

    for ( my $ii = 0; $ii < $ndblocks; $ii++ ) {
        $data .= $data_string;
    }
    if ( $rem ) {
        assert( $rem > 0 );
        $data .= substr( $data_string, 0, $rem );
    }

    if ( $cnt == 0 ) { return $data };

    if ( $nse->{dele_when_free} || $nse->{rename} ) {
        assert( $nse->{dele_when_free} || !$nse->{go_away} );
        return $data;
    }

    my $ct = contents_lookup( nse => $nse );
    assert( defined $ct );

    # CT_DBG defined as null op in original application

    my $cr = _contents_pre_update_range( ct    => $ct,
                                         begin => $begin,
                                         end   => $end );
    assert( $cr->{cr_begin} == $begin );

    my $last_end = $begin;
    for ( ; defined $cr && $cr->{cr_end} <= $end;
            $cr = $dh_contents_ranges->distheap_fromidx(
                    idx => $cr->{cr_crlink_next} ) ) {
        assert( $last_end == $cr->{cr_begin} );

        if ( $cr->{cr_busycnt} == 0 ) {
            $cr->{cr_fill_begin}      = $off;
            $cr->{cr_fill_generation} = $$cookie->{fill_generation};
            $cr->{cr_fill_cksum}      = $$cookie->{fill_cksum};
            assert( !$cr->{cr_new} || !$cr->{cr_invalid} );
            $cr->{cr_new}     = 0;
            $cr->{cr_invalid} = 0;
        } else {
            assert( !$cr->{cr_new} );
        }

        if ( ++$cr->{cr_busycnt} > 1 ) {
            $cr->{cr_invalid} = 1;
        }

        $last_end = $cr->{cr_end};
    }

    assert( $last_end == $end );

    contents_sanity( ct => $ct );

    return $data;

} # End contents_new_write

#--------------------------------------------------------------------------------

=head2 clock_now

  $now = clock_now();

clock_now function taken from the original C code. Keeps a running
total of number of times called.

Arguments are not accepted.

=over

=item Returns

The current value of number of times called

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

# clock_now function from clock.c
{
    #   u_int64_t
    #   clock_now( void )
    #   {
    #       static u_int64_t next = 1;
    #       return next++;
    #   }

    my $next = 1;
    sub clock_now() { return $next++ }
} # End clock_now

#--------------------------------------------------------------------------------

=head2 contents_generate_datablock

  $db = contents_generate_datablock( nse => $nse,
                                     off => $off,
                                     cnt => $cnt,
                                     fill_generation => $fill_generation );

contents_generate_datablock function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=item off

(Required) : The offset at which to start filling the datablock

=item cnt

(Required) : The number of bytes to add to the datablock

=item fill_generation

(Required) : The fill generation cookie

=back

=item Returns

The filled datablock

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_generate_datablock(%;) {
    #   static void
    #   contents_generate_datablock( contents_datablock_t *db, nameset_entry_t nse,
    #           u_int64_t off, u_int32_t cnt, u_int64_t fill_generation ) {
    #       /* Clear db.cksum and any structure padding. */
    #       memset( db, 0, sizeof( *db ) );
    #
    #       db->fill_generation = fill_generation;
    #       memcpy( db->fhdata, nse->fhdata, NAMESET_ENTRY_MAX_FHSIZE );
    #       gettimeofday( &db->timestamp, NULL );
    #       db->off = off;
    #       db->cnt = cnt;
    #
    #       db->cksum = contents_cksum( db, sizeof( *db ) );
    #   }

    my (%opts) = validate (
        @_,
        {
            nse             => { type => HASHREF },
            off             => { type => SCALAR },
            cnt             => { type => SCALAR },
            fill_generation => { type => SCALAR }
        }
    );

    my $db;
    $db->{fill_generation} = $opts{fill_generation};
    $db->{fhdata}          = $opts{nse}->{fhdata};
    $db->{timestamp}       = gettimeofday;
    $db->{off}             = $opts{off};
    $db->{cnt}             = $opts{cnt};

    # db->cksum = contents_cksum( db, sizeof( *db ) );
    # Not going to do cksum right now, added to data to be written
    $db->{cksum} = contents_cksum( data => "$db->{fill_generation}" .
                                           "$db->{fhdata}" );

    return $db;
} # End contents_generate_datablock

#--------------------------------------------------------------------------------

=head2 contents_cksum

  $cksum = contents_cksum( data => $fill_generation .$fhdata );

contents_cksum function taken from original C code. Generates a
check sum based on the data specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item data

(Required) : Data for which the checksum is to be generated

=back

=item Returns

The generated checksum

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_cksum(%;) {
    #   static inline u_int32_t
    #   contents_cksum( const void *data, size_t len ) {
    #       const unsigned char *tmp;
    #       u_int32_t cksum = 0;
    #       int ii;
    #
    #       tmp = ( const unsigned char* ) data;
    #       for ( ii = 0; ii < len; ii++ )
    #           cksum += *tmp++;
    #
    #       return cksum;
    #   }

    my (%opts) = validate (
        @_,
        {
            data => { type => SCALAR }
        }
    );
    my $data = $opts{data};

    my $value = 0;
    while ( $data ne '' ) {
        my $char = chop $data;
        $value  += ord( $char );
    }

    # After calculating the cksum value, pad it with 0's to 32 chars
    my $cksum_len = CKSUM_LEN;
    my $cksum  = sprintf( "%0${cksum_len}d", $value );
    return $cksum;

} # End contents_cksum

#--------------------------------------------------------------------------------

=head2 _contents_pre_update_range

  $cr = _contents_pre_update_range( ct    => $ct,
                                    begin => $begin,
                                    end   => $end );

_contents_pre_update_range function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry

=item begin

(Required) : The begin point of the update range

=item end

(Required) : The end point of the update range

=back

=item Returns

A contenst range structure

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub _contents_pre_update_range(%;) {
    #   static contents_range_t
    #   _contents_pre_update_range( contents_t ct, u_int64_t begin, u_int64_t end ) {
    #       contents_range_t cr;
    #       contents_range_t firstcr;
    #       u_int64_t tbegin;
    #
    #       assert( begin < end );
    #
    #       cr = contents_range_lookup( ct, begin );
    #       if ( cr ) {
    #           assert( cr->cr_begin <= begin );
    #           if ( cr->cr_begin < begin ) {
    #               if ( begin < cr->cr_end ) {
    #                   contents_range_t cr_next;
    #                   /*
    #                    * Update range overlaps current
    #                    * contents obj.  Split the range at
    #                    * 'begin', but fall through to do the
    #                    * rest.
    #                   */
    #                   contents_range_split( ct, cr, begin );
    #                   assert( cr->cr_end == begin );
    #                   cr_next = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #                   assert( cr_next->cr_begin == begin );
    #               }
    #               cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #           } else {
    #               assert( cr->cr_begin == begin );
    #           }
    #        } else {
    #           cr = Q_DH_FIRST( &ct->ct_crlist, dh_contents_ranges );
    #        }
    #
    #        /*
    #         * At this point, everything we have left to scan is either in
    #         * or beyond the update range.  Allocate any missing objs
    #         * within the update range.
    #         */
    #        assert( !cr || begin <= cr->cr_begin );
    #        tbegin = begin;
    #        firstcr = NULL;
    #        for ( ;; ) {
    #            if ( !cr ) {
    #                /* Past all existing objects. */
    #                cr = contents_range_new( tbegin, end );
    #                Q_DH_INSERT_TAIL( &ct->ct_crlist, cr, cr_crlink, dh_contents_ranges, contents_range_t );
    #            }
    #
    #            assert( tbegin <= cr->cr_begin );
    #            if ( tbegin < cr->cr_begin ) {
    #                /* Fill a gap between objects. */
    #                contents_range_t ncr = contents_range_new( tbegin,
    #                MIN( cr->cr_begin, end ) );
    #                Q_DH_INSERT_BEFORE( &ct->ct_crlist, ncr, cr, cr_crlink, dh_contents_ranges, contents_range_t );
    #                cr = ncr;
    #            }
    #
    #            if ( !firstcr ) {
    #                assert( cr );
    #                firstcr = cr;
    #            }
    #
    #            if ( end <= cr->cr_end )
    #                break;
    #
    #                tbegin = cr->cr_end;
    #                cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #       }
    #
    #       /*
    #        * If last contents obj extends beyond the update range, then
    #        * split it.
    #        */
    #       if ( end < cr->cr_end )
    #           contents_range_split( ct, cr, end );
    #
    #       assert( cr->cr_end == end );
    #       _contents_sanity( ct, 1 );
    #       assert( firstcr->cr_begin == begin );
    #
    #       return firstcr;
    #  }

    my (%opts) = validate (
        @_,
        {
            ct    => { type => HASHREF },
            begin => { type => SCALAR },
            end   => { type => SCALAR }
        }
    );
    my $ct = $opts{ct};
    my $begin = $opts{begin};
    my $end = $opts{end};

    assert( $begin < $end );

    my $cr = contents_range_lookup( ct => $ct, off => $begin );

    if ( defined $cr ) {
        assert( $cr->{cr_begin} <= $begin );
        if ( $cr->{cr_begin} < $begin ) {
            if ( $begin < $cr->{cr_end} ) {
                contents_range_split( ct     => $ct,
                                      cr     => $cr,
                                      offset => $begin );
                assert( $cr->{cr_end} == $begin );
                my $cr_next = $dh_contents_ranges->distheap_fromidx(
                        idx => $cr->{cr_crlink_next} );
                assert( $cr_next->{cr_begin} == $begin );
            }
            $cr = $dh_contents_ranges->distheap_fromidx(
                        idx => $cr->{cr_crlink_next} );
        } else {
            assert( $cr->{cr_begin} == $begin );
        }
    } else {
        # $cr = Q_DH_FIRST( $ct->{ct_crlist}, $dh_contents_ranges );
        $cr = $dh_contents_ranges->distheap_fromidx(
                        idx => $ct->{ct_crlist_head} );
    }

    assert( !defined $cr || $begin <= $cr->{cr_begin} );

    my $tbegin  = $begin;
    my $firstcr = undef;
    while ( 1 ) {
        if ( !defined $cr ) {
            $cr = contents_range_new( begin => $tbegin, end => $end );

            # Q_DH_INSERT_TAIL( $ct->{ct_crlist}, $cr, cr_crlink,
            #         $dh_contents_ranges, contents_range_t );
            my $X_idx = $cr->{my_idx};
            $cr->{cr_crlink_prev} = $ct->{ct_crlist_tail};
            $cr->{cr_crlink_next} = 0;
            if ( $ct->{ct_crlist_tail} ) {
                my $Q_tail = $dh_contents_ranges->distheap_fromidx(
                        idx => $ct->{ct_crlist_tail} );
                $Q_tail->{cr_crlink_next} = $X_idx;
            } else {
                $ct->{ct_crlist_head} = $X_idx;
            }
            $ct->{ct_crlist_tail} = $X_idx;
        }

        assert( $tbegin <= $cr->{cr_begin} );
        if ( $tbegin < $cr->{cr_begin} ) {
            # Fill a gap between objects
            my $tend = ( $cr->{cr_begin} < $end ) ? $cr->{cr_begin} : $end;
            my $ncr  = contents_range_new( begin => $tbegin, end => $tend );

            # Q_DH_INSERT_BEFORE( $ct->{ct_crlist}, $ncr, $cr, cr_crlink,
            #        $dh_contents_ranges, contents_range_t );
            my $X_idx = $ncr->{my_idx};
            my $Y_idx = $cr->{my_idx};
            $ncr->{cr_crlink_prev} = $cr->{cr_crlink_prev};
            $ncr->{cr_crlink_next} = $Y_idx;
            $cr->{cr_crlink_prev}  = $X_idx;

            if ( $ncr->{cr_crlink_prev} == 0 ) {
                assert( $ct->{ct_crlist_head} == $Y_idx );
                $ct->{ct_crlist_head} = $X_idx;
            } else {
                my $X_prev = $dh_contents_ranges->distheap_fromidx(
                        idx => $ncr->{cr_crlink_prev} );
                $X_prev->{cr_crlink_next} = $X_idx;
            }

            $cr = $ncr;
        }

        if ( !$firstcr ) {
            assert( defined $cr );
            $firstcr = $cr;
        }

        if ( $end <= $cr->{cr_end} ) {
            last;
        }

        $tbegin = $cr->{cr_end};
        $cr = $dh_contents_ranges->distheap_fromidx(
                        idx => $cr->{cr_crlink_next} );
    }

    if ( $end < $cr->{cr_end} ) {
        contents_range_split( ct     => $ct,
                              cr     => $cr,
                              offset => $end );
    }

    assert( $cr->{cr_end} == $end );
    _contents_sanity( ct => $ct, allow_new => 1 );
    assert( $firstcr->{cr_begin} == $begin );

    return $firstcr;

} # End _contents_pre_update_range

#--------------------------------------------------------------------------------

=head2 contents_range_lookup

  $cr = contents_range_lookup( ct => $ct, off => $begin );

contents_range_lookup function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry

=item off

(Required) : The offset at which lookup should start

=back

=item Returns

A contents range structure

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_range_lookup(%;) {
    #   static contents_range_t
    #   contents_range_lookup( contents_t ct, u_int64_t off ) {
    #       contents_range_t cr;
    #       contents_range_t last = NULL;
    #
    #       Q_DH_FOREACH( cr, &ct->ct_crlist, cr_crlink, dh_contents_ranges ) {
    #           if ( off < cr->cr_begin )
    #               break;
    #           last = cr;
    #       }
    #       return last;
    #   }

    #   #define Q_DH_FOREACH( X, Q, F, H )
    #       for ( X = Q_DH_FIRST( Q, H ); X; X = distheap_fromidx( H, X->F.next ) )

    #   #define Q_DH_FIRST( Q, H )
    #       distheap_fromidx( H, (Q)->head )

    my (%opts) = validate (
        @_,
        {
            ct  => { type => HASHREF },
            off => { type => SCALAR }
        }
    );

     my $last = undef;

    my $start = $dh_contents_ranges->distheap_fromidx(
                              idx => $opts{ct}->{ct_crlist_head} );
    for ( my $cr = $start; defined $cr;
            $cr = $dh_contents_ranges->distheap_fromidx(
                    idx => $cr->{cr_crlink_next} ) ) {
        if ( $opts{off} < $cr->{cr_begin} ) {
            last;
        }
        $last = $cr;
    }

    return $last;

} # End contents_range_lookup

#--------------------------------------------------------------------------------

=head2 contents_range_split

  contents_range_split( ct     => $ct,
                        cr     => $cr,
                        offset => $begin );

contents_range_split function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry

=item cr

(Required) : A contents range

=item offset

(Required) : The offset

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_range_split(%;) {
    #   void
    #   contents_range_split( contents_t ct, contents_range_t cr, u_int64_t offset ) {
    #       contents_range_t ncr;
    #
    #       assert( cr->cr_begin < offset );
    #       assert( offset < cr->cr_end );
    #       assert( !cr->cr_new );
    #
    #       ncr = contents_range_alloc( offset, cr->cr_end,
    #               cr->cr_fill_begin, cr->cr_fill_generation,
    #               cr->cr_fill_cksum, cr->cr_fill_when,
    #               cr->cr_busycnt, cr->cr_invalid, cr->cr_trunc );
    #       Q_DH_INSERT_AFTER( &ct->ct_crlist, ncr, cr, cr_crlink, dh_contents_ranges, contents_range_t );
    #       cr->cr_end = offset;
    #   }

    #   #define Q_DH_INSERT_AFTER( Q, X, Y, F, H, Tp ) {
    #       int X_idx = distheap_toidx( H, X );
    #       int Y_idx = distheap_toidx( H, Y );
    #       (X)->F.prev = Y_idx;
    #       (X)->F.next = (Y)->F.next;
    #       (Y)->F.next = X_idx;
    #       if ( (X)->F.next == 0 ) {
    #           assert( (Q)->tail == Y_idx );
    #           (Q)->tail = X_idx;
    #       } else {
    #           Tp X_next;
    #           X_next = distheap_fromidx( H, X->F.next );
    #           X_next->F.prev = X_idx;
    #       }
    #   }

    my (%opts) = validate (
        @_,
        {
            ct     => { type => HASHREF },
            cr     => { type => HASHREF },
            offset => { type => SCALAR }
        }
    );
    my $ct = $opts{ct};
    my $cr = $opts{cr};
    my $offset = $opts{offset};

    assert( $cr->{cr_begin} < $offset );
    assert( $offset < $cr->{cr_end} );
    assert( !$cr->{cr_new} );

    my $ncr = contents_range_alloc( begin           => $offset,
                                    end             => $cr->{cr_end},
                                    fill_begin      => $cr->{cr_fill_begin},
                                    fill_generation => $cr->{cr_fill_generation},
                                    fill_cksum      => $cr->{cr_fill_cksum},
                                    fill_when       => $cr->{cr_fill_when},
                                    busycnt         => $cr->{cr_busycnt},
                                    invalid         => $cr->{cr_invalid},
                                    trunc           => $cr->{cr_trunc} );
    # Q_DH_INSERT_AFTER( $ct->{ct_crlist}, $ncr, $cr, cr_crlink,
    #  $dh_contents_ranges, contents_range_t );
    my $X_idx = $ncr->{my_idx};
    my $Y_idx = $cr->{my_idx};
    $ncr->{cr_crlink_prev} = $Y_idx;
    $ncr->{cr_crlink_next} = $cr->{cr_crlink_next};
    $cr->{cr_crlink_next}  = $X_idx;
    if ( $ncr->{cr_crlink_next} == 0 ) {
        assert( $ct->{ct_crlist_tail} == $Y_idx );
        $ct->{ct_crlist_tail} = $X_idx;
    } else {
        my $X_next = $dh_contents_ranges->distheap_fromidx(
                idx => $ncr->{cr_crlink_next} );
        $X_next->{cr_crlink_prev} = $X_idx;
    }

    $cr->{cr_end} = $offset;

} # End contents_range_split

#--------------------------------------------------------------------------------

=head2 contents_range_alloc

  $ncr = contents_range_alloc( begin           => $offset,
                               end             => $cr->{cr_end},
                               fill_begin      => $cr->{cr_fill_begin},
                               fill_generation => $cr->{cr_fill_generation},
                               fill_cksum      => $cr->{cr_fill_cksum},
                               fill_when       => $cr->{cr_fill_when},
                               busycnt         => $cr->{cr_busycnt},
                               invalid         => $cr->{cr_invalid},
                               trunc           => $cr->{cr_trunc} );

contents_range_lookup function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item begin

(Required) : begin parameter

=item end

(Required) : end parameter

=item fill_begin

(Required) : fill_begin parameter

=item fill_generation

(Required) : fill_generation parameter

=item fill_cksum

(Required) : fill_cksum parameter

=item fill_when

(Required) : fill_when parameter

=item busycnt

(Required) : busycnt parameter

=item invalid

(Required) : 'invalid' parameter

=item trunc

(Required) : trunc parameter

=back

=item Returns

A contents range structure

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_range_alloc(%;) {
    #   static contents_range_t
    #   contents_range_alloc( u_int64_t begin, u_int64_t end, u_int64_t fill_begin,
    #           u_int64_t fill_generation, u_int32_t fill_cksum, u_int64_t fill_when,
    #           int busycnt, int invalid, int trunc ) {
    #       contents_range_t cr = distheap_alloc( dh_contents_ranges, 1 );
    #       memset( cr, 0xbb, sizeof( *cr ) );
    #
    #       assert( begin < end );
    #       assert( fill_begin <= begin );
    #
    #       cr->cr_begin = begin;
    #       cr->cr_end = end;
    #       cr->cr_new = 0;
    #       cr->cr_busycnt = busycnt;
    #       cr->cr_invalid = invalid;
    #       cr->cr_trunc = trunc;
    #
    #       cr->cr_fill_begin = fill_begin;
    #       cr->cr_fill_generation = fill_generation;
    #       cr->cr_fill_cksum = fill_cksum;
    #       cr->cr_fill_when = fill_when;
    #
    #       return cr;
    #    }

    my (%opts) = validate (
        @_,
        {
            begin           => { type => SCALAR },
            end             => { type => SCALAR },
            fill_begin      => { type => SCALAR },
            fill_generation => { type => SCALAR },
            fill_cksum      => { type => SCALAR },
            fill_when       => { type => SCALAR },
            busycnt         => { type => SCALAR },
            invalid         => { type => SCALAR },
            trunc           => { type => SCALAR },
        }
    );

    assert( $opts{begin} < $opts{end} );
    assert( $opts{fill_begin} <= $opts{begin} );

    my ( $cr, undef ) = $dh_contents_ranges->distheap_alloc( weight => 1 );

    $cr->{cr_begin}   = $opts{begin};
    $cr->{cr_end}     = $opts{end};
    $cr->{cr_new}     = 0;
    $cr->{cr_busycnt} = $opts{busycnt};
    $cr->{cr_invalid} = $opts{invalid};
    $cr->{cr_trunc}   = $opts{trunc};

    $cr->{cr_fill_begin}      = $opts{fill_begin};
    $cr->{cr_fill_generation} = $opts{fill_generation};
    $cr->{cr_fill_cksum}      = $opts{fill_cksum};
    $cr->{cr_fill_when}       = $opts{fill_when};

    return $cr;

} # End contents_range_alloc

#--------------------------------------------------------------------------------

=head2 contents_range_new

  $cr = contents_range_new( begin => $tbegin, end => $end );

contents_range_new function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item begin

(Required) : Begin value of new contents range

=item end

(Required) : End value of new contents range

=back

=item Returns

The new contents range structure

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_range_new(%;) {
    #   static contents_range_t
    #   contents_range_new( u_int64_t begin, u_int64_t end ) {
    #       contents_range_t cr = distheap_alloc( dh_contents_ranges, 1 );
    #       memset( cr, 0xcc, sizeof( *cr ) );
    #
    #       assert(begin < end);
    #
    #       cr->cr_begin = begin;
    #       cr->cr_end = end;
    #       cr->cr_new = 1;
    #       cr->cr_busycnt = 0;
    #       cr->cr_invalid = 0;
    #       cr->cr_trunc = 0;
    #
    #       return cr;
    #   }

    my (%opts) = validate (
        @_,
        {
            begin => { type => SCALAR },
            end   => { type => SCALAR }
        }
    );

    my ( $cr, undef ) = $dh_contents_ranges->distheap_alloc( weight => 1 );

    assert( $opts{begin} < $opts{end} );

    $cr->{cr_begin}   = $opts{begin};
    $cr->{cr_end}     = $opts{end};
    $cr->{cr_new}     = 1;
    $cr->{cr_busycnt} = 0;
    $cr->{cr_invalid} = 0;
    $cr->{cr_trunc}   = 0;

    return $cr;

} # End contents_range_new

#--------------------------------------------------------------------------------

=head2 contents_fill_write

  contents_fill_write( nse    => $nse,
                       off    => $offset,
                       cnt    => $count,
                       error  => $error,
                       cookie => $contents_cookie );

contents_fill_write function from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : The nameset entry

=item off

(Required) : The offset to begin the write

=item cnt

(Required) : The number of bytes to be written

=item error

(Required) : Contents range should be invalidated

=item cookie

(Required) : Contents write cookie

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_fill_write(%;) {
    #   void
    #   contents_fill_write( nameset_entry_t nse, u_int64_t off, u_int32_t cnt,
    #           int error, const contents_write_cookie_t *cookie ) {
    #       contents_t ct;
    #       contents_range_t cr;
    #       u_int64_t begin = off;
    #       u_int64_t end = off + cnt;
    #       u_int64_t last_end;
    #
    #       /*
    #        * Accept some slop, in case op generator wants to generate
    #        * zero-length reads/writes.
    #        */
    #       if ( cnt == 0 )
    #           return;
    #
    #       ct = contents_lookup( nse );
    #       assert( ct );
    #       if ( !ct )
    #           return; /* File removed. */
    #
    #       /*
    #        * An nse can be in the "dele_when_free" state while in the middle
    #        * of a rename; in this case we want file i/o to the filehandle to
    #        * continue uninterrupted, so return only if the nse is not also
    #        * in the "rename" state.
    #        */
    #       if ( nse->dele_when_free && !nse->rename )
    #           return; /* File removed. */
    #
    #       /*
    #        * No one must have flushed the write range, otherwise we
    #        * can't ensure correct behavior.  We must assume that all
    #        * range objects in the update range were touched by this
    #        * write.  (Though this write was not necessarily the _first_
    #        * to touch the range objects, but in that case they MUST
    #        * already be invalidated.)
    #        */
    #
    #       CT_DBG( ct, ( "fill_write %llu -> %llu gen %llu\n", begin, end, cookie->fill_generation ) );
    #
    #       /* Update patterns in write range. */
    #       cr = _contents_pre_update_range( ct, begin, end );
    #       assert( cr->cr_begin == begin );
    #       last_end = begin;
    #       while ( cr && cr->cr_end <= end ) {
    #           /* No gaps. */
    #           assert( last_end == cr->cr_begin );
    #
    #           assert( !cr->cr_new );
    #           assert( cr->cr_busycnt > 0 );
    #           cr->cr_busycnt--;
    #
    #           /*
    #            * We only fill if there were no overlapping writes.
    #            * Overlapping writes are detected when new writes are
    #            * issued.  If no one has invalidated this range, then
    #            * we can safely assume that the write succeeded and
    #            * the range in the filesystem currently contains our
    #            * pattern.
    #            */
    #           if ( !cr->cr_invalid ) {
    #               /* It must be our write. */
    #               assert( cr->cr_busycnt == 0 );
    #               assert( cr->cr_fill_begin == off );
    #               assert( cr->cr_fill_generation == cookie->fill_generation );
    #               assert( cr->cr_fill_cksum == cookie->fill_cksum );
    #
    #               /*
    #                * XXO: We should roll back to the previous
    #                * pattern on error.
    #                */
    #               if ( error )
    #                   cr->cr_invalid = 1;
    #
    #               cr->cr_fill_when = clock_now();
    #           }
    #
    #           last_end = cr->cr_end;
    #
    #           /*
    #            * If the range was part of a trunc, discard it.
    #            */
    #           if ( cr->cr_trunc && cr->cr_busycnt == 0 ) {
    #               contents_range_t del = cr;
    #               cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #               Q_DH_REMOVE( &ct->ct_crlist, del, cr_crlink, dh_contents_ranges, contents_range_t );
    #               memset( del, 0xee, sizeof( *del ) );
    #               distheap_dele( dh_contents_ranges, del );
    #           } else {
    #               cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #           }
    #      }
    #      assert( last_end == end );
    #
    #      contents_sanity( ct );
    #  }

    my (%opts) = validate (
        @_,
        {
            nse    => { type => HASHREF },
            off    => { type => SCALAR },
            cnt    => { type => SCALAR },
            error  => { type => SCALAR },
            cookie => { type => HASHREF }
        }
    );
    my $nse = $opts{nse};
    my $off = $opts{off};
    my $cnt = $opts{cnt};
    my $error = $opts{error};
    my $cookie = $opts{cookie};

    my $begin = $off;
    my $end   = $off + $cnt;

    if ( $cnt == 0 ) {
        return;
    }

    my $ct = contents_lookup( nse => $nse );

    assert( defined $ct );
    if ( !defined $ct ) {
        # This doesn't make sense. $ct will die on assert if not defined
        return;
    }

    if ( $nse->{dele_when_free} && !$nse->{rename} ) {
        return; # File removed
    }

    # CT_DBG defined as null op in original application

    my $cr = _contents_pre_update_range( ct    => $ct,
                                         begin => $begin,
                                         end   => $end );
    assert( $cr->{cr_begin} == $begin );
    my $last_end = $begin;

    while ( defined $cr && $cr->{cr_end} <= $end ) {
        assert( $last_end == $cr->{cr_begin} );

        assert( !$cr->{cr_new} );
        assert( $cr->{cr_busycnt} > 0 );
        $cr->{cr_busycnt}--;

        if ( !$cr->{cr_invalid} ) {
            assert( $cr->{cr_busycnt}         == 0 );
            assert( $cr->{cr_fill_begin}      == $off );
            assert( $cr->{cr_fill_generation} == $cookie->{fill_generation} );
            assert( $cr->{cr_fill_cksum}      == $cookie->{fill_cksum} );

            if ( $error ) {
                $cr->{cr_invalid} = 1;
            }
            $cr->{cr_fill_when} = clock_now();
        }

        $last_end = $cr->{cr_end};

        if ( $cr->{cr_trunc} && $cr->{cr_busycnt} == 0 ) {
            my $del = $cr;
            $cr = $dh_contents_ranges->distheap_fromidx(
                               idx => $cr->{cr_crlink_next} );
            # Q_DH_REMOVE( $ct->{ct_crlist}, $del, cr_crlink, $dh_contents_ranges, contents_range_t );
            my $dhfirst = $dh_contents_ranges->distheap_fromidx(
                    idx => $ct->{cr_crlist_head} );
            if ( $dhfirst == $del ) {
                $ct->{ct_crlist_head} = $del->{cr_crlink_next};
            }
            my $dhlast = $dh_contents_ranges->distheap_fromidx(
                    idx => $ct->{cr_crlist_tail} );
            if ( $dhlast == $del ) {
                $ct->{ct_crlist_tail} = $del->{cr_crlink_prev};
            }
            if ( $del->{cr_crlink_prev} ) {
                my $X_prev = $dh_contents_ranges->distheap_fromidx(
                        idx => $del->{cr_crlink_prev} );
                $X_prev->{cr_crlink_next} = $del->{cr_crlink_next};
            }
            if ( $del->{cr_crlink_next} ) {
                my $X_next = $dh_contents_ranges->distheap_fromidx(
                        idx => $del->{cr_crlink_next} );
                $X_next->{cr_crlist_prev} = $del->{cr_crlink_prev};
            }

            $dh_contents_ranges->distheap_dele( hd => $del );
        } else {
            $cr = $dh_contents_ranges->distheap_fromidx(
                        idx => $cr->{cr_crlink_next} );
        }
    }
    assert( $last_end == $end );

    contents_sanity( ct => $ct );

} # End contents_fill_write

#--------------------------------------------------------------------------------

=head2 contents_new_read

  contents_new_read( cookie => \%contents_cookie );

contents_new_read function taken from original C code.

The nse parameter from the original C code is removed. It is not used
in the function itself, and introduces errors on Perl 5.6.1 installations.
The cookie parameter is passed as a HASHREF rather than as a SCALARREF that
points to a HASH.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cookie

(Required) : Reference to the contents cookie hash

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_new_read(%;) {
    #   void
    #   contents_new_read( nameset_entry_t nse, contents_read_cookie_t *cookie ) {
    #       cookie->when_issued = clock_now();
    #   }

    my (%opts) = validate (
        @_,
        {
            cookie => { type => HASHREF }
        }
    );
    my $cookie = $opts{cookie};

    $cookie->{when_issued} = clock_now();

} # End contents_new_read

#--------------------------------------------------------------------------------

=head2 contents_verify_read

  contents_verify_read( nse    => $nse,
                        off    => $offset,
                        cnt    => $count,
                        data   => $res->{data},
                        cookie => \%contents_cookie );

contents_verify_read function taken from original C code.

The cookie parameter is passed in as a HASHREF rather than a SCALARREF
that points to the HASH.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=item off

(Required) : The offset at which to begin read verification

=item cnt

(Required) : The number of bytes to verify

=item data

(Required) : Data to verify

=item cookie

(Required) : The contents cookie hash reference

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_verify_read(%;) {
    #   void
    #   contents_verify_read( nameset_entry_t nse, u_int64_t off, u_int32_t cnt,
    #           const char *data, const contents_read_cookie_t *cookie ) {
    #       contents_t ct;
    #       contents_range_t cr;
    #       u_int64_t begin = off;
    #       u_int64_t end = off + cnt;
    #
    #       /*
    #        * Accept some slop, in case op generator wants to generate
    #        * zero-length reads/writes.
    #        */
    #       if ( cnt == 0 )
    #           return;
    #
    #       ct = contents_lookup( nse );
    #       if ( !ct )
    #           return; /* Nothing known about this file yet. */
    #
    #       cr = contents_range_lookup( ct, begin );
    #       /* cr is before, contains, or overlaps the verify range */
    #       if ( !cr )
    #           cr = Q_DH_FIRST( &ct->ct_crlist, dh_contents_ranges );
    #       else {
    #           if ( cr->cr_end <= begin )
    #               cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #           else
    #               assert( cr->cr_begin <= begin );
    #       }
    #       /* cr is beyond, contains, or overlaps the verify range */
    #
    #       while ( cr && cr->cr_begin < end ) {
    #           u_int64_t rbegin = MAX( cr->cr_begin, begin );
    #           u_int64_t rend = MIN( end, cr->cr_end );
    #
    #           /*
    #            * Reads can only be validated against fills which
    #            * completed before the read was issued.  With
    #            * overlapping operations, there's no guarantee of
    #            * when requests were executed.
    #            */
    #           if ( cr->cr_fill_when < cookie->when_issued ) {
    #               contents_verify_one( cr, nse, rbegin, rend, data + ( rbegin - begin ) );
    #           }
    #
    #           cr = distheap_fromidx( dh_contents_ranges, cr->cr_crlink.next );
    #       }
    #   }

    my (%opts) = validate (
        @_,
        {
            nse    => { type => HASHREF },
            off    => { type => SCALAR },
            cnt    => { type => SCALAR },
            data   => { type => SCALAR },
            cookie => { type => HASHREF }
        }
    );
    my $nse = $opts{nse};
    my $cookie = $opts{cookie};
    my $data = $opts{data};

    my $begin = $opts{off};
    my $end   = $opts{off} + $opts{cnt};

    if ( $opts{cnt} == 0 ) {
        return;
    }

    my $ct = contents_lookup( nse => $nse );
    if ( !defined $ct ) {
        return;    # Nothing known about this file yet.
    }

    my $cr = contents_range_lookup( ct => $ct, off => $begin );

    if ( !defined $cr ) {
        # cr = Q_DH_FIRST( &ct->ct_crlist, dh_contents_ranges )
        $cr = $dh_contents_ranges->distheap_fromidx(
                  idx => $ct->{ct_crlist_head} );
    } else {
        if ( $cr->{cr_end} <= $begin ) {
            $cr = $dh_contents_ranges->distheap_fromidx(
                      idx => $cr->{cr_crlink_next} );
        } else {
            assert( $cr->{cr_begin} <= $begin );
        }
    }

    while ( defined $cr && $cr->{cr_begin} < $end ) {
        my $rbegin = ( $cr->{cr_begin} > $begin ) ? $cr->{cr_begin} : $begin;
        my $rend   = ( $end < $cr->{cr_end} ) ? $end : $cr->{cr_end};

        if ( $cr->{cr_fill_when} < $cookie->{when_issued} ) {
            contents_verify_one( cr    => $cr,
                                 nse   => $nse,
                                 begin => $rbegin,
                                 end   => $rend,
                                 data  => substr( $data, $rbegin - $begin ) );
        }

        $cr = $dh_contents_ranges->distheap_fromidx(
                   idx => $cr->{cr_crlink_next} );
    }

} # End contents_verify_read

#--------------------------------------------------------------------------------

=head2 contents_verify_one

  contents_verify_one( cr    => $cr,
                       nse   => $nse,
                       begin => $rbegin,
                       end   => $rend,
                       data  => substr( $data, $rbegin - $begin ) );

contents_verify_one function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item cr

(Required) : A contents range structure

=item nse

(Required) : A nameset entry

=item begin

(Required) : The range offset at which to begin

=item end

(Required) : The range offset at which to end

=item data

(Required) : The data to verify

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub contents_verify_one(%;) {
    #   static void
    #   contents_verify_one( contents_range_t cr, nameset_entry_t nse,
    #           u_int64_t begin, u_int64_t end, const char *data )  {
    #       contents_datablock_t db;
    #       int ndblocks;
    #       size_t rem;
    #       int ii;
    #
    #       /* Cannot validate a read when a write is in progress. */
    #       if ( cr->cr_busycnt > 0 || cr->cr_invalid )
    #           return;
    #
    #       assert( begin < end );
    #       assert( cr->cr_begin < cr->cr_end );
    #       assert( cr->cr_begin <= begin );
    #       assert( end <= cr->cr_end );
    #       assert( cr->cr_fill_begin <= cr->cr_begin );
    #
    #       db.fill_generation = cr->cr_fill_generation;
    #       db.cksum = cr->cr_fill_cksum;
    #       memcpy( db.fhdata, nse->fhdata, NAMESET_ENTRY_MAX_FHSIZE );
    #
    #       /*
    #        * N.B. We do not verify against partial copies of the
    #        * pattern at the beginning or end of the verify range.
    #        */
    #       assert( begin >= cr->cr_fill_begin );
    #       rem = ( begin - cr->cr_fill_begin ) % sizeof( contents_datablock_t );
    #       if ( rem ) {
    #           int skipahead = sizeof( contents_datablock_t ) - rem;
    #           data += skipahead;
    #           begin += skipahead;
    #           if ( end <= begin )
    #               return;
    #       }
    #
    #       assert( end > begin );
    #       ndblocks = ( end - begin ) / sizeof( contents_datablock_t );
    #
    #       for ( ii = 0; ii < ndblocks; ii++ ) {
    #           contents_datablock_t src;
    #           u_int32_t data_cksum;
    #
    #           memcpy( &src, data, sizeof( src ) );
    #           data_cksum = src.cksum;
    #           src.cksum = 0;
    #           src.cksum = contents_cksum( &src, sizeof( src ) );
    #
    #           if ( db.fill_generation != src.fill_generation ||
    #                   db.cksum != src.cksum ||
    #                   db.cksum != data_cksum ||
    #                   memcmp( db.fhdata, src.fhdata, nse->fhlen ) ) {
    #               /* DATA CORRUPTION! */
    #               report_corruption( nse, begin + ii *
    #                       sizeof( contents_datablock_t ),
    #                       sizeof( contents_datablock_t ),
    #                       &db, &src );
    #           }
    #           data += sizeof( contents_datablock_t );
    #       }
    #   }

    my (%opts) = validate (
        @_,
        {
            cr    => { type => HASHREF },
            nse   => { type => HASHREF },
            begin => { type => SCALAR },
            end   => { type => SCALAR },
            data  => { type => SCALAR }
        }
    );
    my $cr = $opts{cr};
    my $nse = $opts{nse};
    my $begin = $opts{begin};
    my $end = $opts{end};
    my $data = $opts{data};

    if ( $cr->{cr_busycnt} > 0 || $cr->{cr_invalid} ) {
        return;
    }

    assert( $begin < $end );
    assert( $cr->{cr_begin}      <  $cr->{cr_end} );
    assert( $cr->{cr_begin}      <= $begin );
    assert( $end                 <= $cr->{cr_end} );
    assert( $cr->{cr_fill_begin} <= $cr->{cr_begin} );

    my $db;
    $db->{fill_generation} = $cr->{cr_fill_generation};
    $db->{cksum}           = $cr->{cr_fill_cksum};
    $db->{fhdata}          = $nse->{fhdata};

    # Convert the $db struct to a single data string without
    # the cksum and save the length
    my $db_str        = "$db->{fill_generation}" . "$db->{fhdata}";
    my $sizeof_c_db_t = do { use bytes; length $db_str };

    assert( $begin >= $cr->{cr_fill_begin} );
    my $rem = ( $begin - $cr->{cr_fill_begin} ) % ( $sizeof_c_db_t + CKSUM_LEN );

    if ( $rem ) {
        my $skipahead = $sizeof_c_db_t + CKSUM_LEN - $rem;
        # The skipahead may put the pointer past the end of the data. The
        #  subroutine then returns because $end <= $begin. Perl substr function
        #  will produce a warning that cannot be caught. Let's just check the
        #  length beforehand, and return if the skipahead is too large
        #  for the data.
        my $data_len = do { use bytes; length $data };
        return if ( $data_len <= $skipahead );

        $data = eval { substr( $data, $skipahead ) };
        if ( $@ && $@ !~ /substr outside of string/ ) {
            croak "contents_verify_one: $@";
        }
        $begin += $skipahead;
        if ( $end <= $begin ) {
            return;
        }
    }

    assert( $end > $begin );
    my $ndblocks = int( ( $end - $begin ) / ( $sizeof_c_db_t + CKSUM_LEN ) );

    for ( my $ii = 0; $ii < $ndblocks; $ii++ ) {
        my $src = substr( $data, 0, $sizeof_c_db_t );

        # $data_cksum is the cksum as found in
        #   the data from the read request.
        # $src_cksum is the cksum as calculated from the
        #   fill_generation and fhdata from the read request.
        my $data_cksum = substr( $data, $sizeof_c_db_t, CKSUM_LEN );
        my $src_cksum  = contents_cksum( data => $src );

        # Separate out the comparisons, so we can print a more
        #   specific error message about which component failed
        if ( $db_str ne $src ) {
            # report_corruption();
            croak "contents_verify_one: *** DATA CORRUPTION ***: " .
                    "fill_generation and fhdata do not match";
        }
        if ( $db->{cksum} ne $data_cksum ) {
            # report_corruption();
            croak "contents_verify_one: *** DATA CORRUPTION ***: " .
                    "cksum from read does not match";
        }
        if ( $db->{cksum} ne $src_cksum ) {
            # report_corruption();
            croak "contents_verify_one: *** DATA CORRUPTION ***: " .
                    "cksum from src does not match";
        }

        $data = substr( $data, $sizeof_c_db_t + CKSUM_LEN );
    }

} # End contents_verify_one

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Contents SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    # Call clock_now a random number of times
    my $num = int(rand 10) + 1;
    my $time = 0;
    for (1..$num) {
        $time = clock_now();
    }

    # Time should be $num
    if ($time != $num) {
        Tharn::logresult('FAIL', "clock_now function returned incorrect value");
    } else {
        Tharn::logresult('PASS', "SELF TEST of nfs/Contents.pm");
    }
}

1;
