#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Delegations.pm#3 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::Delegations;

use strict;
use base qw(Exporter);

use Tharn;
use nfs::CommonParams;
#use Data::Dumper;

use Carp qw(croak);
use SPARSE::Util::Logger;
use GenericMount qw(mount umount);
use FileManip qw(read_stream write_stream);
use nfs::RemoteFileIO qw(client_read_file start_io stop_io);
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF HASHREF CODEREF);
use nfs::Utils qw{get_client_addr};

#([{

our @EXPORT_OK = qw(
  set_env
  clean_env
  check_delegation
  check_oplock
  get_delegation_and_check_recall
  get_lock_owner
  file_lock_send
  get_file_lock
  check_oplock_nfs_interaction
  get_number_of_delegs
);

our $VERSION = '0.10';

# These values were needed for subtests.
# Earlier the subtests can just open a file and wait for signals.
# Now it is not the case, the new clients will not keep the delegations
# if there is no continuous read or write. So instead of waiting for
# a signal from the parent process, we also try to execute an IO in the subtest.
my %GH;
$GH{read_wait} = <<'EOT';
logcomment(qq($lock_no: Waiting to close file));
while (1) {
    sleep 1;
    my $dummy;
    read($FH, $dummy, 10, 0);
    logcomment ("$lock_no: Read done");
    if ($socket->can_read()) {
        my $mesg = $socket->read_message();
        if ($mesg eq "close") {
            logcomment(qq{$lock_no: Got the signal to exit});
            last;
        } elsif ($mesg eq "flock") {
            logcomment(qq{$lock_no: Got the signal to lock, but I have alread locked, continuing.});
        } else {
            die "Some stupid IPC problem, exiting";
        }
    }
}
EOT

$GH{write_wait} = <<'EOT';
logcomment(qq($lock_no: Waiting to close file));
while (1) {
    sleep 1;
    my $dummy;
    print $FH "Write to deleg file by get_file_lock, for delegation tests\n";
    logcomment ("$lock_no: Write done");
    if ($socket->can_read()) {
        my $mesg = $socket->read_message();
        if ($mesg eq "close") {
            logcomment(qq{$lock_no: Got the signal to exit});
            last;
        } elsif ($mesg eq "flock") {
            logcomment(qq{$lock_no: Got the signal to lock, but I have alread locked, continuing.});
        } else {
            die "Some stupid IPC problem, exiting";
        }
    }
}
EOT

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $usermap = '/etc/usermap.cfg';
my $test_file   = "Delegations.$$";
my $nfs_v4_enable;
my $nfs_v4_acl;
my $nfs_v4_rdel;
my $nfs_v4_wdel;
my $nfs_mount_rootonly;
my $admin_added;
my @file_locks;
my $lock_pid;
my $lock_uid;
my $lockid;
my $lock_no = 0;

#---------------------------------------------------------------------
## @name ::get_delegation_and_check_recall
## @description
##  Get a delegation on a file or stream and check whether the
##  delegation is recalled by the operation on the file/stream
##  from another client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = get_delegation_and_check_recall( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg file
##  Required: Get delegation on this file.
##
## @arg inode
##  Required: Inode of file/stream to get delegation for.
##
## @arg client
##  Required: Connection handle for NFSv4 client to get delegation on.
##
## @arg mountpoint
##  Required: The directory to mount the volume onto.
##
## @arg delegation
##  Optional: Type of delegation to get: 'read' or 'write',
##  (default is 'read').
##
## @arg stream
##  Optional: Get delegation on this stream instead of the file.
##
## @arg oper_client
##  Optional: Connection handle for client where operation is performed.
##  (default is the same as option 'client').
##
## @arg oper_mtpoint
##  Optional: The directory on the operation client to mount the volume onto.
##  (default is the same as option 'mountpoint').
##
## @arg operation
##  Optional: Reference to a function to execute on the NFS client.
##
## @arg oper_args
##  Optional: Reference to a list of arguments to pass to the operation
##  function.
##
## @arg operation_after_ck
##  Optional: Reference to a function to execute on the NFS client
##  after the delegation has been checked for recalled.
##  This is used to revert any destructive changes made by 'operation'.
##
## @arg oper_args_after_ck
##  Optional: Reference to a list of arguments to pass to the
##  operation_after_ck function.
##
## @arg NFSversion
##  Optional: NFS version to use on client, (default is 4, allowed:4, 4.1).
##
## @arg delegation_expected
##  Optional: Expect to get a delegation (default is 1).
##
## @arg delegation_recalled
##  Optional: Expect a delegation to be recalled (default is 0).
##
## @arg message
##  Optional: Message to log as a step before performing operation.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub get_delegation_and_check_recall(%) {
    my (%opts) = validate (
        @_,
        {
            set_h               => { type => HASHREF },
            file                => { type => SCALAR },
            inode               => { type => SCALAR },
            client              => { type => OBJECT },
            mountpoint          => { type => SCALAR },
            delegation          => { type => SCALAR,    default => 'read',
                                    regex => qr/^(read|write)$/ },
            stream              => { type => SCALAR,   optional => 1 },
            oper_client         => { type => OBJECT,   optional => 1 },
            oper_mtpoint        => { type => SCALAR,   optional => 1 },
            operation           => { type => CODEREF,  optional => 1 },
            oper_args           => { type => ARRAYREF, optional => 1,
                                  depends => ['operation'] },
            operation_after_ck  => { type => CODEREF,  optional => 1 },
            oper_args_after_ck  => { type => ARRAYREF, optional => 1,
                                  depends => ['operation_after_ck'] },
            minorversion    => { type => SCALAR, default => $::NFS4_MINOR_VERSION, regex => qr/^\d$/ },
            NFSversion  => { type => SCALAR,   default => '4' },
            delegation_expected => { type => BOOLEAN,   default => 1 },
            delegation_recalled => { type => BOOLEAN,   default => 0 },
            message             => { type => SCALAR,
                                  default => q(Perform operation) },
            message_after_ck    => { type => SCALAR,
                                  default => q(Perform operation after check) },
            sec                 => { type => SCALAR, default => $::AUTH,
                                    regex => qr/(sys|krb5|krb5i|krb5p)$/ },
        }
    );
    my ($mountpoint, $unmount_flag, $file, $stream, $err, $not, $op);
    my $fc         = ($opts{set_h}{fc} ? $opts{set_h}{fc} : $opts{set_h}{hostname});
    my $hostname= $opts{set_h}{hostname};
    my $path     = $opts{set_h}{volume};
    my $vol_name     = $opts{set_h}{vol_name};
    my $tfcont     = $opts{set_h}{tfcontents};
    my $cc         = $opts{client};
    my $conn       = defined($opts{oper_client}) ? $opts{oper_client} : $cc;
    my $NFSversion = $opts{NFSversion};
    my $recalled   = $opts{delegation_recalled};
    my $mtpoint    = defined($opts{oper_mtpoint}) ? $opts{oper_mtpoint}
                                                  : $opts{mountpoint};
    #logStep('BEGIN: get_delegation_and_check_recall vars:' . Dumper(\%opts));

    # Check if it is a CIFS client
    if ( $conn->hostp()->hosttype() =~ /^windows/ ) {
        $NFSversion = 0;
    }

    $opts{nfs_version_under_test} = $opts{nfs_version_under_test} ? $opts{nfs_version_under_test} :
        "4" . ($opts{minorversion} ? ".$opts{minorversion}" : "");

    # Catch any errors to properly unexport the volume
    my $subid;
    eval {
        logStep('Mount the volume on the client');
        $mountpoint = mount($hostname,
                            $path,
                            mntopts    => 'vers=4' . ($opts{minorversion} eq "0" ? "" : ",minorversion=$opts{minorversion}") . ",sec=$opts{sec}",
                            sharing    => 'exclusive',
                            mountpoint => $opts{mountpoint},
                            client     => $cc) or croak $@;
        my $client_addr = get_client_addr($cc, $hostname);

        # Catch any errors to properly unmount the volume
        eval {
            # File as seen by the client
            $file = "$mountpoint/$opts{file}";
            # Get a read/write delegation on a file/stream
            my $msg = "Get a $opts{delegation} delegation on a";
            if ( defined($opts{stream}) ) {
                logStep("$msg stream");
                $stream = $opts{stream};
                # Read/Write stream to get delegation
                if ( $opts{delegation} eq 'read' ) {
                    read_stream(client => $cc,
                                stream => $stream,
                                file   => $file);
                } else {
                    write_stream(client   => $cc,
                                 stream   => $stream,
                                 contents => $tfcont,
                                 file     => $file);
                }
            } else {
                logStep("$msg file");
                # Read/Write file to get delegation
                $subid = start_io( client => $cc->hostp(),
                                   type  => $opts{delegation},
                                    file => $file);
            }

            logStep('Verify proper granting of delegation');
            my %delegation_h = (fc     => $fc,
                                inode  => $opts{inode},
                                protocol => "nfsv" . $opts{nfs_version_under_test},
                                ipaddr => $client_addr,
                                volume => $path,
                                file   => $opts{file},
                                type   => $opts{delegation});


            $op = eval { check_delegation(%delegation_h) };
            my $firsterr = $@;
            if ($firsterr) {
                logStep (qq{check_delegation failed. Checking if it is really expected or not});
            } else {
                logStep(qq{check_delegation success});
            }

            if ( $op xor $opts{delegation_expected} ) {
                $not = " not" if not $opts{delegation_expected};
                $file = defined($opts{stream}) ? "stream" : "file";
                croak "$opts{delegation} delegation should$not " .
                      "have been granted on $file: $@";
            }

            if (not defined($opts{operation})) {
                logStep(qq{No other operation is needed, returning});
                stop_io(sub_id => $subid) if $subid;
                return 1;
            }

            logStep(qq{Performing the next operation});

            # Mount volume on unix client only since the windows
            # client has already been mapped at setup
            if ( $NFSversion == 4 and $mtpoint eq $mountpoint
              and $cc->connhost() eq $conn->connhost() ) {
                # No need to mount volume, it is the same client that
                # holds the delegation and the same NFS version, so
                # it is already mounted.
                # Also, make sure the volume is not unmounted
                $unmount_flag = 0;
            } elsif ( $NFSversion ) {
                logStep('Mount the volume on client B');
                my $opts_string = "vers=$NFSversion";
                if ($NFSversion =~ /(\d+)\.(\d+)/) {
                    $opts_string = "vers=$1,minorversion=$2";
                }
                $opts_string .= ",sec=$opts{sec}";
                $mtpoint = mount($hostname,
                                 $path,
                                 mntopts    => $opts_string,
                                 sharing    => 'exclusive',
                                 mountpoint => $mtpoint,
                                 client     => $conn) or croak $@;
                # Make sure volume is unmounted
                $unmount_flag = 1;
            }
            logStep(qq{ClientB mounted, now doing the operation:});

            # Perform operation
            logStep($opts{message});
            #logStep(qq{operation:$opts{operation}, and args:} . Dumper(@{$opts{oper_args}}));
            &{$opts{operation}}( @{$opts{oper_args}} );
            croak $@ if $@;

            logStep(qq{Operation done, check if deleg is recalled or not});
            # Verify that the delegation is properly recalled
            $not = ' not' if not $recalled;
            logStep("Verify that the $opts{delegation} " .
                    "delegation is$not recalled");
            eval { check_delegation(%delegation_h) };

            if ( $recalled and $@ =~ qr/Could not find .* delegation/ ) {
                # Expecting a recall
                undef $@;
            } elsif ( $recalled and ! $@ ) {
                croak "Delegation($opts{delegation}) should have been recalled";
            }
            if ($subid) {
                stop_io(sub_id => $subid);
                undef $subid;
            }
            logStep(qq{Do the after check operation now});
            if ( $opts{operation_after_ck} ) {
                # Perform operation after check
                logStep($opts{message_after_ck});
                &{$opts{operation_after_ck}}(@{$opts{oper_args_after_ck}});
                croak $@ if $@;
            }
        };

        $err = $@;
        logStep(qq{the first err: $err}) if $err;

        if ( $mtpoint and $unmount_flag ) {
            logStep('Unmount the volume on client B');
            umount($mtpoint, client=>$conn, timeout=>5);
            #umount($mtpoint, client=>$conn) if $@;
            $err .= " $@" if $@;
            logStep(qq{the second err: $err}) if $@;
        }

        if ($err) { croak $err; };

    }; # First eval
    $err = $@;

    logStep('Stop traffic and unmount the volume from the primary client');
    stop_io(sub_id => $subid) if $subid;
    umount($mountpoint, client=>$cc, timeout=>5) if ($mountpoint);
    $mountpoint = "";
    croak "Number1:$err\n#######################################################\nand\n########################################################\nerror1:$@" if ($err or $@);

    logStep('END: get_delegation_and_check_recall');

    return 1;

}

#---------------------------------------------------------------------
## @name ::get_lock_owner
## @description
##  Return the process id and the effective user id of the current
##  lock.
##
## @synopsis
##  $lock_owner = get_lock_owner();
##
## @return
##  Hash reference with the following keys: 'pid' and 'uid'.
#---------------------------------------------------------------------
sub get_lock_owner() {
    return { pid => $lock_pid, uid => $lock_uid };
}

#---------------------------------------------------------------------
## @name ::file_lock_send
## @description
##  Send message to the file open by get_file_lock().
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = file_lock_send( %options );
##
## @arg lockid
##  Optional: Lock id to send message,
##  Default is current id set by file_lock_send.
##
## @arg message
##  Required: Message to send to file: 'flock' or 'close',
##  (default is 'close').
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub file_lock_send(%) {
    my (%opts) = validate (
        @_,
        {
            lockid  => { type => OBJECT, default => $lockid },
            message => { type => SCALAR, default => 'close',
                                           regex => qr/^(flock|close)$/ },
            timeout => { type => SCALAR, default => $main::CMDTIMEOUT },
        }
    );

    if ( $opts{lockid}->is_alive() ) {
        # Send message to subtest to terminate
        $opts{lockid}->ipc()->write_message($opts{message});

        if ( $opts{message} eq 'close' ) {
            # Wait for the subtest to finish
            $opts{lockid}->wait_finish(timeout=>$opts{timeout});
        }
    }

    return 1;
}


#---------------------------------------------------------------------
## @name ::get_file_lock
## @description
##  Open a file from a client, lock the file and hold it open until
##  instructed to release it using the 'file_lock_send' function.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = get_file_lock( %options );
##
## @arg client
##  Required: Hostspec, Hostrec or Connectrec of client.
##
## @arg file
##  Required: File to open.
##
## @arg user
##  Optional: Open the file as this user.
##
## @arg gid
##  Optional: Open the file with this group id.
##
## @arg open_type
##  Optional: Type of open: 'read' or 'write', (default is 'read').
##
## @arg lock_type
##  Optional: Type of lock to get: 'flock', 'excl' or 'lvl2', (default 'none').
##
## @arg start_byte
##  Optional: Starting byte of range lock, (default is 0).
##
## @arg byte_length
##  Optional: Number of bytes of range lock, (default is 0).
##
## @arg xstart_byte
##  Optional: Starting byte of extended range lock.
##  If option is passed process waits for a signal to extend the range lock.
##
## @arg xbyte_length
##  Optional: Number of bytes of extended range lock.
##  If option is passed process waits for a signal to extend the range lock.
##
## @arg lock_wait
##  Optional: Wait for file_lock_send to signal to get the actual lock.
##  Must use this option with 'flock' lock type, (default is 0).
##
## @arg upgrade_lock
##  Optional: Upgrade a read lock to a write lock, (flock only, default is 0).
##
## @arg xopen_type
##  Optional: Open the file with the proper type (default is write)
##
## @arg upgrade_lock_sig
##  Optional: Wait for file_lock_send signal to upgrade a read lock to a write lock, (flock only, default is 0).
##
## @arg downgrade_lock_sig
##  Optional: Wait for file_lock_send signal to downgrade a write lock to a read lock, (flock only, default is 0).
##
## @arg lvl2_info
##  Optional: NFSv4 client information hash. To get a level 2 oplock
##  on a CIFS client, first get a delegation on the file and then
##  the open of the file will give the level 2 oplock instead of an
##  exclusive oplock.  The information hash contains:
##    fc     => Connection handle for filer,
##    cc     => Connection handle for NFSv4 client,
##    volume => Volume on filer where file resides.
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out,
##  (default is 0, wait forever).
##
## @return
##  lock id.
#---------------------------------------------------------------------
sub get_file_lock(%) {
    my (%opts) = validate (
        @_,
        {
            client       => { type => SCALAR|OBJECT },
            file         => { type => SCALAR },
            user         => { type => SCALAR,  optional => 1 },
            gid          => { type => SCALAR,  optional => 1 },
            lvl2_info    => { type => HASHREF, optional => 1 },
            lock_wait    => { type => BOOLEAN,  default => 0 },
            timeout      => { type => SCALAR,   default => 0 },
            start_byte   => { type => SCALAR,   default => 0,
                             regex => qr/^\d+$/ },
            xopen_type   => { type => SCALAR,   optional => 1,
                              regex => qr/^(read|write)$/ },
            byte_length  => { type => SCALAR,   default => 0,
                             regex => qr/^\d+$/ },
            upgrade_lock_sig => { type => BOOLEAN,  default => 0 },
            downgrade_lock_sig => { type => BOOLEAN,  default => 0 },
            minor_version => { type => SCALAR, default => $::NFS4_MINOR_VERSION, regex => qr/^\d$/ },
            xstart_byte  => { type => SCALAR,  optional => 1,
                             regex => qr/^\d+$/ },
            xbyte_length => { type => SCALAR,  optional => 1,
                             regex => qr/^\d+$/ },
            upgrade_lock => { type => BOOLEAN,  default => 0 },
            keep_io         => { type => SCALAR, optional => 1,
                                regex => qr/^\d$/, default => 1 },
            open_type    => { type => SCALAR,   default => 'read',
                             regex => qr/^(read|write)$/ },
            lock_type    => { type => SCALAR,   default => 'none',
                             regex => qr/^(none|excl|lvl2|flock)$/ },
            sec         => { type => SCALAR, default => $::AUTH,
                             regex => qr/^(sys|krb5|krb5i|krb5p)$/ },
        }
    );
    my (%extra_options, $mountpoint, $msgs, $cc, $op);
    my $file = $opts{file};

    logStep('opts =' . Dumper(\%opts));

    # Get Hostrec object from the client option
    if ( UNIVERSAL::isa($opts{client}, "Connectrec") ) {
        # Get hostrec object from a connectrec object
        $opts{client} = $opts{client}->hostp();
    } else {
        # Get hostrec object from a hostspec
        $opts{client} = host($opts{client});
    }

    logStep(qq{Crossed the host processing});
    # Validate arguments
    if ( ($opts{lock_type} eq 'lvl2' or $opts{lock_type} eq 'excl')
      and $opts{client}->hosttype() !~ /^windows/ ) {
        croak "Exclusive or level 2 oplocks requires a CIFS client";
    }
    if ( $opts{lock_wait} and $opts{lock_type} ne 'flock' ) {
        croak "Option 'lock_wait' must be used with 'flock' lock type";
    }
    if ( $opts{upgrade_lock} and $opts{lock_type} ne 'flock' ) {
        croak "Option 'upgrade_lock' must be used with 'flock' lock type";
    }

    logStep(qq{Entering the level 2 check processing});
    if ( $opts{lock_type} eq 'lvl2' ) {
        if ( ! defined($opts{lvl2_info}) ) {
            croak "Level 2 oplock requires option 'lvl2_info'";
        }
        $cc = $opts{lvl2_info}->{cc};
        # To get a level 2 oplock, first get
        # read delegation and then open the file on CIFS
        logcomment(qq{hostname = } . Dumper($opts{lvl2_info}));
        my @params = ($opts{lvl2_info}->{hostname},
                            $opts{lvl2_info}->{volume},
                            client  => $cc,
                            mntopts => "vers=4" . ($opts{minor_version} ? ",minorversion=$opts{minor_version}" : "") . ",sec=$opts{sec}"
                            );
        #logStep(qq{mounting : params:} . Dumper(@params));
        $mountpoint = mount($opts{lvl2_info}->{hostname},
                            $opts{lvl2_info}->{volume},
                            client  => $cc,
                            mntopts => "vers=4" . ($opts{minor_version} ? ",minorversion=$opts{minor_version}" : "") . ",sec=$opts{sec}"
                            ) or croak $@;
        logStep(qq{Done with mount});

        # Remove drive off the filename
        my $afile = $file;
        $afile =~ s/^.://;
        $afile =~ s{\\}{/}g;

        if ( $afile =~ /:/ ) {
            # Get a read delegation on the stream
            my ($bfile, $stream) = ($afile =~ /(.*):(.*)/);
            logStep(qq{Reading from stream});
            read_stream(client => $cc,
                        stream => $stream,
                        file   => "$mountpoint/$bfile");
        } else {
            # Get a read delegation on the file
            logStep(qq{Getting read deleg from file});
            client_read_file($cc, "$mountpoint/$afile");
        }
    }
    logStep(qq{Exiting the level 2 check processing});

    my $flock_cmd   = '';
    my $upgrade_cmd = '';
    my $xflock_cmd  = '';
    my $gid_cmd     = '';
    my $user = defined($opts{user}) ? $opts{user} : '';
    my $xopen;

    # Setup open and lock type
    my $otype  = $opts{open_type} eq 'read' ? 'O_RDONLY' : 'O_WRONLY';
    my $ltype  = $opts{open_type} eq 'read' ? 'F_RDLCK'  : 'F_WRLCK';
    my $ctype  = $opts{open_type} eq 'read' ? 1 : 2;
    my $start  = defined($opts{start_byte})  ? $opts{start_byte}  : 0;
    my $len    = defined($opts{byte_length}) ? $opts{byte_length} : 0;
    my $xstart = $opts{xstart_byte};
    my $xlen   = $opts{xbyte_length};

    # by default, wait for signal, while continuously reading from file.
    my $wait_for_signal = qq{\$socket->read_message();\n};
    if ($opts{keep_io}) {
        if ($opts{open_type} eq "read") {
            $wait_for_signal = $GH{read_wait};
        } else {
            $wait_for_signal = $GH{write_wait};
        }
    }

    logStep(qq{open_type = $opts{open_type}});

    if ( defined($opts{xopen_type})) {
       $xopen  = $opts{xopen_type} eq 'read' ? 'F_RDLCK'  : 'F_WRLCK';
    }

    logStep(qq{Preparing the write_message});
    my $command = qq{\$socket->write_message(qq(--open_ok--\$\$--\$>));};
    # Process the 'upgrade_lock_sig' option
    if ( $opts{upgrade_lock_sig} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        $wait_for_signal = $GH{write_wait} if ($opts{keep_io});
        # start with a read lock
        $ltype = 'F_RDLCK';
        # upgrade to a write lock
        $upgrade_cmd = qq{
            $command
            logcomment(qq(\$lock_no: Wait to extend lock on \$file));
            \$socket->read_message();
            \$lock = pack(\$flock_st, F_WRLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
        };
    }

    # Process the 'upgrade_lock' option
    if ( $opts{upgrade_lock} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        $wait_for_signal = $GH{write_wait} if ($opts{keep_io});
        # start with a read lock
        $ltype = 'F_RDLCK';
        # upgrade to a write lock
        $upgrade_cmd = qq{
            \$lock = pack(\$flock_st, F_WRLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
        };
    }

    # Process the 'downgrade_lock_sig' option
    if ( $opts{downgrade_lock_sig} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        $wait_for_signal = $GH{write_wait} if ($opts{keep_io});
        # start with a write lock
        $ltype = 'F_WRLCK';
        # downgrade to a read lock
        $upgrade_cmd = qq{
            logcomment(qq(\$lock_no: Wait to extend lock on \$file));
            $command
            \$socket->read_message();
            \$lock = pack(\$flock_st, F_RDLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
        };
    }


    if ( defined($xstart) ) {
        # Check if open type is specified 
        if ( defined($opts{xopen_type})) {
                $ltype = $xopen;
                $otype = 'O_RDWR';
                $wait_for_signal = $GH{write_wait} if ($opts{keep_io});
        }
        # Wait for signal to extend the range of lock
        $xflock_cmd = qq{
            logcomment(qq(\$lock_no: Wait to extend lock on \$file));
            $command
            \$socket->read_message();
            \$lock = pack(\$flock_st, $ltype, SEEK_SET, \@xrange);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
        };
    } else {
        # Make sure these variables are defined
        $xstart = 0;
        $xlen   = 0;
    }

    my $fop;
    if ( $opts{lock_wait} ) {
        $fop = 'F_SETLKW';
    } else {
        $fop = 'F_SETLK';
    }
    # Process the 'lock_type' option
    if ( $opts{lock_type} eq 'flock' ) {
        $flock_cmd = qq{
            \$lock = pack(\$flock_st, $ltype, SEEK_SET, \@range);
            fcntl(\$FH, $fop, \$lock) or die \$!;
            $xflock_cmd
            $upgrade_cmd
        };
    }

        # Process the 'lock_wait' option
    if ( $opts{lock_wait} ) {
        $command = qq{
            logcomment(qq(\$lock_no: Wait for lock on \$file));
            $command
            $flock_cmd
        };
    } else {
        $command = qq{
            $flock_cmd
            $command
        };
    }

    # Process the 'gid' option
    if ( $opts{gid} ) {
        # Change the credentials (gid) to only list
        # the gid specified and the change the uid
        # to the current uid since subtest will be
        # run as root
        $gid_cmd = qq{
            \$) = q($opts\{gid\} $opts\{gid\});
            \$> = $>;
            logcomment(qq\{\$lock_no: Changed credentials to <\$>,\$)>\});
        };
        # Option gid implies user => 'root' since the
        # credentials of the process must be changed
        # and only root can do it
        $opts{user} = 'root';
    }

    logDbg("get_file_lock: file <$file> user <$user> " .
           "open <$opts{open_type}> lock <$opts{lock_type}>");
    ++$lock_no;
    # Function to open the file
    my $mod_file = __FILE__ . ": get_file_lock at line " . (__LINE__ + 1);
    my $get_file_lock = qq{ \{ eval \{
        use POSIX;
        use Tharn;
        my (\$FH, \$lock, \$fc, \$flock_st, \@range, \@xrange,\$fop);
        my \$file = q($file);
        my \$lock_no = $lock_no;
        my (\$system, \$mtype) = (POSIX::uname)[0,4];
        if (\$system =~ /^Linux\$/ && \$mtype =~ /^x86_64\$/) \{
            \$flock_st = 'ssLLLL';
            \@range = (0, $start, 0, $len, 0);
            \@xrange = (0, $xstart, 0, $xlen, 0);
        \} elsif (\$mtype =~ /^[ix]\\d?86/) \{
            \$flock_st = 'ssLLLL';
            \@range = ($start, 0, $len, 0);
            \@xrange = ($xstart, 0, $xlen, 0);
        \} elsif (\$mtype eq 'sun4u') \{
            \$flock_st = 'ssLLLLL';
            \@range = (0, 0, $start, 0, $len);
            \@xrange = (0, 0, $xstart, 0, $xlen);
        \} elsif (\$system eq 'AIX') \{
            \$flock_st = 'ssLLLLLLL';
            \@range = (0, 0, 0, 0, $start, 0, $len);
            \@xrange = (0, 0, 0, 0, $xstart, 0, $xlen);
        \} else \{
            die "Could not set flock structure, " .
                "unknown client architecture: \$system \$mtype";
        \}
        $gid_cmd
        my \$socket = Subtest::ipc();
        sysopen(\$FH, \$file, $otype) or die \$!;
        close \$FH;
        sysopen(\$FH, \$file, $otype) or die \$!;
        logcomment(qq(\$lock_no: File \$file opened));
        $command
        $wait_for_signal
        logcomment(qq(\$lock_no: Closing file \$file));
        close(\$FH);
    \}; die "\$@ ($mod_file)\\n" if \$@; \} };

    if ( $opts{lock_type} eq 'flock'
      and $opts{client}->hosttype() =~ /^windows/ ) {
        # For windows, use flock to lock the whole file
        $get_file_lock =~ s/fcntl.*?F_SETLK.*?;/flock(\$FH,$ctype);/g;
        $get_file_lock =~ s/F_(RD|WR)LCK/0/g;
    }

    # Only append the username if the option is given
    $extra_options{'-username'} = $opts{user} if defined($opts{user});

    logStep(qq{Subtesting: $get_file_lock });

    # Open the file on the client
    $lockid = subtest( $get_file_lock,
                       '-bg', '-ipc', '-runid', "${lock_no}_get_file_lock",
                       '-timeout', $opts{timeout},
                       '-client',  $opts{client},
                       %extra_options,
                       '-onexit', sub {} );

    # Wait for client to open the file
    $op = eval { $lockid->ipc()->read_message() };
    if ( $@ || ! (($lock_pid, $lock_uid) = ($op =~ /--open_ok--(\d+)--(\d+)/)) ) {
        # There was an error so release, i.e., terminate subtest.
        eval { file_lock_send(lockid => $lockid) };

        $msgs = join(' ', $lockid->results());
    }

    # Unmount the volume on the NFSv4 clients
    if ( $opts{lock_type} eq 'lvl2' ) {
        umount($mountpoint, client=>$cc, force=>1);
    }

    croak "Error opening the file: $msgs" if $msgs;

    return $lockid;
}

#---------------------------------------------------------------------
## @name ::check_oplock_nfs_interaction
## @description
##  Get an oplock on a file and check its interaction with NFS
##  operations on the same file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_oplock_nfs_interaction( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg file
##  Required: File/stream to get oplock for.
##
## @arg inode
##  Required: Inode of file/stream to get oplock for.
##
## @arg client
##  Required: Hostspec of CIFS client.
##
## @arg nfscc
##  Optional: Connection handle for NFS client,
##  (default is the primary client).
##
## @arg operation
##  Optional: Reference to a function to execute on the NFS client.
##
## @arg oper_args
##  Optional: Reference to a list of arguments to pass to the operation
##  function.
##
## @arg oper_fail
##  Optional: Expect the operation on the file to fail, (default is 0).
##
## @arg mountpoint
##  Optional: The directory to mount the volume onto.
##
## @arg NFSversion
##  Optional: NFS version to use on client, (default is 4).
##
## @arg oplock
##  Optional: Type of oplock to get: 'excl' or 'lvl2', (default is 'excl').
##
## @arg expect_oplock
##  Optional: Type of oplock to expect after operation: 'excl' or 'lvl2'
##  (default is the same as option 'oplock').
##
## @arg oplock_revoked
##  Optional: Expect the oplock to be revoked after the operation,
##  (default is 0).
##
## @arg message
##  Optional: Message to log as a step before performing operation.
##
## @return
## 1 if succeeds.
#---------------------------------------------------------------------
sub check_oplock_nfs_interaction(%) {
    my (%opts) = validate (
        @_,
        {
            set_h          => { type => HASHREF },
            client         => { type => SCALAR },
            file           => { type => SCALAR },
            inode          => { type => SCALAR },
            nfscc          => { type => OBJECT,   optional => 1 },
            operation      => { type => CODEREF,  optional => 1 },
            oper_args      => { type => ARRAYREF, optional => 1,
                             depends => ['operation'] },
            oper_fail      => { type => SCALAR,    default => 0 },
            mountpoint     => { type => SCALAR,   optional => 1 },
            NFSversion     => { type => SCALAR,    default => 4 },
            minorversion   => { type => SCALAR, default => $::NFS4_MINOR_VERSION },
            oplock         => { type => SCALAR,    default => 'excl',
                               regex => qr/^(excl|lvl2)$/ },
            expect_oplock  => { type => SCALAR,   optional => 1,
                               regex => qr/^(excl|lvl2)$/ },
            oplock_revoked => { type => SCALAR,    default => 0 },
            message        => { type => SCALAR,
                             default => q(Perform operation) },
            sec            => { type => SCALAR, default => $::AUTH,
                             regex => qr/(sys|krb5|krb5i|krb5p)$/ },
        }
    );
    my ($mountpoint, $oplockid, $err, $op);
    my $fc         = ($opts{set_h}{fc} ? $opts{set_h}{fc} : $opts{set_h}{hostname});
    my $hostname=$opts{set_h}{hostname};
    my $data_server = $hostname;
    my $volume = $opts{set_h}{volume};
    my $nfsver = $opts{NFSversion};
    my $oplock = $opts{oplock};
    my $cc     = defined($opts{nfscc}) ? $opts{nfscc} : $opts{set_h}{cc}[0];

    # Expected oplock after operation
    my $expect_oplock = defined($opts{expect_oplock}) ?
                        $opts{expect_oplock} : $oplock;

    # Oplock information
    my %oplock_info = ( inode  => $opts{inode},
                        ipaddr => $opts{client},
                        type   => $oplock );

    # Catch any errors to properly unexport the volume
    eval {
        # Information needed for a level 2 oplock
        my %extra = ( hostname => $data_server, fc => $fc, cc => $cc, volume => $volume, data_server => $data_server );
        # Get an oplock on the file/stream
        $op = $opts{file} =~ /:[^\\]/ ? 'stream' : 'file';
        logStep("Get an oplock on the $op");
        $oplockid = get_file_lock( client    => $opts{client},
                                   file      => $opts{file},
                                   lock_type => $oplock,
                                   lvl2_info => \%extra );

        # Verify the oplock is granted
        logStep('Verify the oplock is granted');
        $oplock_info{ipaddr} = get_client_addr($opts{client}, $hostname);
        check_oplock(%oplock_info);

        # No operation is requested
        return if ( ! defined($opts{operation}) );

        # Change the oplock info to expect for given oplock
        $oplock_info{type} = $expect_oplock;

        # Catch any errors to properly release the oplock
        eval {
            # Mount the volume on the NFS client
            logStep('Mount the volume on the NFS client');
            $mountpoint = mount($hostname,
                                $volume,
                                mntopts    => "vers=4" . ($opts{minor_version} ? ",minorversion=$opts{minor_version}" : "") . ",sec=$opts{sec}",
                                mountpoint => $opts{mountpoint},
                                sharing    => 'exclusive',
                                client     => $cc) or croak $@;

            # Catch any errors to properly unmount the volume
            eval {
                logStep($opts{message});
                &{$opts{operation}}( @{$opts{oper_args}} );
            };

            if ( $@ xor $opts{oper_fail} ) {
                my $failed = $opts{oper_fail} ? 'failed' :
                                                'been successful';
                $err = "Operation should have $failed: $@";
            } else {
                # Check the status of the oplock
                logStep('Check the status of the oplock');
                eval { check_oplock(%oplock_info) };
                $op = ( $@ =~ qr/Could not find .* oplock/ );
                if ( ( ! $@ and $opts{oplock_revoked} )
                  or ( $op and ! $opts{oplock_revoked} ) ) {
                    my $not = " not" if not $opts{oplock_revoked};
                    $err = "Oplock should$not have been revoked: $@";
                } elsif ( $@ and ! $op ) {
                    $err = $@;
                }
            }

            # Unmount the volume on the NFS client
            logStep('Unmount the volume on the NFS client');
            umount($mountpoint, client => $cc, timeout=>5);
            umount($mountpoint, client => $cc) if $@;
            croak "$err $@" if ($err or $@);
        };

        $err = $@;

        # Release the oplock on the file
        logStep('Release the oplock on the file');
        eval {file_lock_send(lockid => $oplockid) };
        croak "$err $@" if ($err or $@);
    };

    $err = $@;
    if ($oplockid) {
        eval {
            file_lock_send(lockid => $oplockid);
        };
    };
    croak "main error:$err\n###################################and secondary error:$@" if ($err or $@);

    return 1;
}

use vars qw{$FILER};
param('FILER', '-mesg', 'Hostname of the filer');
my $f = host($main::FILER);
if ($f->hosttype =~ /filer-ng/m)
{
    require nfs::Delegations_cMode;
} else {
    require nfs::Delegations_7Mode;
}

#}])
1;
