#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/Delegations.pm#21 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

## @summary Library for the NFSv4 delegations tests
##
## @description
##   Common routines for NFSv4 delegations testing in 7Mode only. Called by nfs::Delegations
##
## @usage
##   use nfs::Delegations;
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @owner	dl-nfsqa
##

use strict;

use Tharn;
use Tharn qw(host connect);
use ExportFS qw(share_add share_delete);
use nfs::CommonParams;
my $f = host($main::FILER);

use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::Utils qw(validate_nfs_version);
use GenericMount qw(mount umount);
use Network::Utils qw(run_getXXbyYY);
use Module::Common qw(validate_mixed);
use TestEnv::MinWin qw(minwin_compatible);
use nfs::Utils qw(validate_nfsv4_clients);
use FileManip qw(read_stream write_stream);
use MapShare qw(map_cifs_share unmap_cifs_share);
use nfs::GenericExport qw(export_path unexport_path);
use Users qw(addPasswdEntry addGroupEntry delPasswdEntry delGroupEntry);
use FS::Util::Setup qw(create_vol common_cleanup create_and_map_cifs_share);
use nfs::RemoteFileIO qw(client_read_file client_write_file client_del_file);
use filer qw(filer_options_get filer_options_set filer_lock_status_file_get filer_qtree_create filer_qtree_security_set);
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF HASHREF CODEREF);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $usermap = '/etc/usermap.cfg';
my $test_file   = "Delegations.$$";
my $nfs_v4_enable;
my $nfs_v4_acl;
my $nfs_v4_rdel;
my $nfs_v4_wdel;
my $nfs_mount_rootonly;
my $admin_added;
my @file_locks;
my $lock_pid;
my $lock_uid;
my $lockid;
my $lock_no = 0;

#---------------------------------------------------------------------

=head2 nfslocks_setup

  $out = nfslocks_setup(setup_h => \%setup_h,
                        [, clients => \@clients ]
                        [, cifs_clients => \@cifs_clients ]
                        [, nfs_allowed => \@allowed ]
                        [, nfs_version => 2|3|4|'all'|'not2'|'not3'|'not4' ]
                        [, lock_type => 'read'|'write'|'both' ]
                        [, mt_root_only => 'on'|'off' ]
                        [, local_host => 0|1 ]
                        [, mount => 0|1 ]);

Set up environment for NFS locks tests. It creates a volume and
a test file on that volume. All setup state is passed back to
the test on the 'setup_h' named option. This hash is populated
with the following:

  { fc            => Filer connection object,
    cc            => List of client connection objects,
    files         => List of files created by setup,
    volume        => Full path of created volume,
    local_ipaddr  => Local ip address as seen by the filer,
    lock_types    => List of lock types to test,
    cifs          => List of CIFS hashes having drive and host as keys,
    nfs_versions  => List of valid NFS versions,
    setup_mtpoint => Setup mountpoint }

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item clients

(Optional) : Array reference of NFS client hostspecs.

=item cifs_clients

(Optional) : Array reference of CIFS client hostspecs.

=item nfs_allowed

(Optional) : An array reference indicating what NFS versions are
allowed for this test. An error occurs if the nfs_version parameter
does not match one of these values. Default is to allow all versions.

=item nfs_version

(Optional) : NFS version(s) that will be tested. Acceptable values are
2, 3, 4, 'not2', 'not3', 'not4', and 'all'. Default is 'all' which will
test versions 2, 3 and 4.

=item local_host

(Optional) : If true names in client lists can be 'localhost': 0 or 1.
Default is 1.

=item lock_type

(Optional) : Lock type(s) that will be tested. Acceptable values are
'read', 'write', 'both'. Default is 'both'.

=item create_files

(Optional) : Create this number of files on test volume.
Default is to create one file.

=item mount

(Optional) : Mount the volume if true: 0 or 1. Mounting the volume
using the highest NFS version provided by the nfs_version parameter.
Default is to not mount the volume.

=item qtree_sec

(Optional) : Specify the qtree security style to use for FILER_VOL.
Default is to use 'unix'.

=item mt_root_only

(Optional) : Allow only privileged NFS mounts, i.e. only the root
user can mount the volume. Valid values are 'on' or 'off'.

=back

=item Returns
1 if succeeds else croaks with error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub set_env(\%;%);
#sub clean_env(\%;%);
sub check_delegation(%;);
sub check_oplock(%;);
sub get_delegation_and_check_recall(%;);
sub get_lock_owner();
sub file_lock_send(%;);
sub get_file_lock(%;);
sub check_oplock_nfs_interaction(%;);

#---------------------------------------------------------------------
## @name ::set_env
## @description
##  Set up environment for NFSv4 delegations tests and initialize all
##  common variables.  The first argument is the structure to be
##  initialized and the rest of the arguments are options passed
##  in "key => value" style.
##
## @synopsis
##  $op = set_env( %setup_h [, %options ] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables will get initialized in this function.
##
## @arg nfsv4_clients
##  Optional: Reference to list of NFSv4 client hostspecs.
##
## @arg cifs_clients
##  Optional: Reference to list of CIFS client hostspecs.
##
## @arg deleg_clients
##  Optional: Reference to a list of client hostspecs which support
##  delegations. This list must be a subset of list in 'nfsv4_clients'.
##
## @arg stream_clients
##  Optional: Reference to a list of client hostspecs which support
##  named attributes. This list must be a subset of list in 'nfsv4_clients'.
##
## @arg stream
##  Optional: Boolean, create a stream (named attribute) on the test file,
##  (default => 0).
##
## @arg qtree
##  Optional: Mode of qtree to create (unix, mixed, ntfs).
##
## @arg users
##  Optional: Reference to list of user names to check and get their uid's.
##
## @arg groups
##  Optional: Reference to list of group names to check and get their gid's.
##
## @arg client_conn
##  Optional: Clients connection spec for NFSv4 clients, (default => '#rsh').
##
## @arg enable_acls
##  Optional: Flag to enable acls on the filer, (default => 0).
##
## @arg enable_read_delegations
##  Optional: Flag to enable read delegations on the filer, (default => 1).
##
## @arg enable_write_delegations
##  Optional: Flag to enable write delegations on the filer, (default => 1).
##
## @arg setup
##  Optional: Reference to a function to run at the end of setup.
##
## @return
##  Nothing.
##
## @notes
##  Structure used by the NFSv4 delegations tests:
##
##  %setup_h = (
##
##      fc            => Filer's connection object.
##
##      cc            => Reference to a list of all client connection objects.
##
##      scc           => Reference to a list of all client connection objects
##                       except for the primary client, cc[1...N].
##
##      volume        => Absolute path of volume.
##
##      vol_name      => Volume name.
##
##      tfname        => Test file name.
##
##      tfinode       => Test file inode.
##
##      tfcontents    => Test file contents.
##
##      tsname        => Test stream name.
##
##      tsinode       => Test stream inode.
##
##      tscontents    => Test stream contents.
##
##      qfinode       => Inode for test file on qtree.
##
##      qsinode       => Inode for stream on qtree file.
##
##      qtree         => Absolute path of qtree.
##
##      qtree_name    => Name of qtree.
##
##      qtree_mode    => Mode of qtree (unix, mixed, or ntfs).
##
##      nfsv4cc       => Reference to a list of NFSv4 client connection objects.
##
##      cifscc        => Reference to a list of CIFS client connection objects.
##
##      cifs_drive    => Reference to a list of CIFS maps (mountpoints)
##
##      cifs_host     => Reference to a list of CIFS hostspecs
##
##      users         => Reference to a hash of users: { username => uid }.
##
##      groups        => Reference to a hash of groups: { groupname => gid }.
##
##      mountpoint    => Tentative mountpoint for tests.
##
##      setup_mtpoint => The mount point of volume on the primary client
##                       for setup and cleanup purposes.
##  );
#---------------------------------------------------------------------
sub set_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            nfsv4_clients  => { type => ARRAYREF, optional => 1 },
            cifs_clients   => { type => ARRAYREF, optional => 1 },
	    nfs_version  => { type => SCALAR,    default => 'all' },
            deleg_clients  => { type => ARRAYREF, optional => 1 },
            stream_clients => { type => ARRAYREF, optional => 1 },
            client_conn    => { type => SCALAR,    default => '#rsh' },
            stream         => { type => BOOLEAN,   default => 0 },
            users          => { type => ARRAYREF, optional => 1 },
            groups         => { type => ARRAYREF, optional => 1 },
            qtree          => { type => SCALAR,   optional => 1,
                               regex => qr/^(unix|mixed|ntfs)$/ },
            setup          => { type => CODEREF,  optional => 1 },
            enable_acls              => { type => BOOLEAN, default => 0 },
            enable_read_delegations  => { type => BOOLEAN, default => 1 },
            enable_write_delegations => { type => BOOLEAN, default => 1 },
        }
    );
    my ($mountpoint, $client, $nfsv4, $file, $cc, $h, $op);
    my ($cifs_passwd, $cifs_user, $fc, $fh, $fname, $vol_name, $mode,
        $mountpoint, @nfsver, %nfsver_h, $new_text, $op, %qtree_out);

    logStep('BEGIN: set_env: setup environment for test');

    # Connect to filer
    my $f = host($main::FILER);

    # Make sure at least one NFSv4 client is specified
    if ( ! $opts{nfsv4_clients} ) {
        croak "No NFSv4 clients are specified";
    }

    # Validate clients and create connection handles for all clients,
    # start with the NFSv4 clients and when it reaches the separator '-'
    # process the CIFS clients.
    logStep('Validate and create connection handles for clients');

    # Validate that nfsv4 clients support NFSv4
    # and validate clients that support named attributes
    validate_nfsv4_clients( clients        => $opts{nfsv4_clients},
                            deleg_clients  => $opts{deleg_clients},
                            stream_clients => $opts{stream_clients} );

    # Create connection handles for all clients
    $nfsv4 = 1;
    foreach $client ( @{$opts{nfsv4_clients}}, '-', @{$opts{cifs_clients}} ) {
        if ( $client eq '-' ) {
            $nfsv4 = 0;
            next;
        }
        # Validate that clients can not be 'localhost'
        if ( $client =~ /^localhost\b/ ) {
            croak "Clients can not be 'localhost'";
        }

        # Verify client
        $h = host($client);
        if ( $nfsv4 and $h->hosttype() !~ /^unix/ ) {
            croak "Client '$client' must be unix";
        } elsif ( ! $nfsv4 and $h->hosttype() !~ /^windows/ ) {
            croak "Client '$client' must be windows";
        }
        # Create connection for client
        $cc = connect($opts{client_conn}, '-timeout', $main::TIMEOUT);
        $cc->timeout($main::CMDTIMEOUT);

        # Save the connection handle of secondary clients
        push @{$set_h->{scc}}, $cc if defined($set_h->{cc});

        # Save the connection handle
        push @{$set_h->{cc}}, $cc;
        if ( $nfsv4 ) {
            push @{$set_h->{nfsv4cc}}, $cc;
        } else {
            push @{$set_h->{cifscc}}, $cc;
            push @{$set_h->{cifs_host}}, $client;
        }
    }

    logStep('7G/7M specific setup');
	
    # Common setup -- creates a volume depending
    # on the $main::SETUP_DONE parameter
    logStep('Create test volume');
    ($set_h->{fc}, $set_h->{vol_name}) = create_vol() or croak $@;
	#Workaround alert.
	#{fc} is not being set by the above function!
	host($main::FILER);
	$set_h->{fc} = connect('#rsh');
	$set_h->{hostname} = $set_h->{fc}->connhost();

    # Full path for volume
    $set_h->{volume} = "/vol/$set_h->{vol_name}";

    # Mountpoint for tests
    $set_h->{mountpoint} = '/t/' . $set_h->{hostname} .
                           '/' . $set_h->{vol_name} . '/delegations';

    if ( $opts{users} ) {
        my $msg;
        # Verify the existence of specified users
        # on the filer and get their uid's
        logStep("Check existence of specified users " .
                "on the filer and get their uid's");
        foreach my $user ( @{$opts{users}} ) {
            $op   = run_getXXbyYY($set_h->{fc}, 'getpwbyname_r', $user);
            ($op) = ( $op =~ /pw_uid = (\d+)/ );
            if ( ! $op ) {
                $msg .= "User '$user' must exist on the filer. ";
            }
            $set_h->{users}{$user} = $op;
        }
        croak $msg if $msg;
    }

    if ( $opts{groups} ) {
        my $msg;
        # Verify the existence of specified groups
        # on the filer and get their gid's
        logStep("Check existence of specified groups " .
                "on the filer and get their gid's");
        foreach my $group ( @{$opts{groups}} ) {
            $op   = run_getXXbyYY($set_h->{fc}, 'getgrbyname', $group);
            ($op) = ( $op =~ /gid = (\d+)/ );
            if ( ! $op ) {
                $msg .= "Group '$group' must exist on the filer. ";
            }
            $set_h->{groups}{$group} = $op;
        }
        croak $msg if $msg;
    }

    if ( @{$opts{cifs_clients}} ) {
        # Check existence of the administrator account in unix
        $op = run_getXXbyYY($set_h->{fc}, 'getpwbyname_r', 'administrator');
        if ( $op !~ /pw_name = / ) {
            # Create administrator account on /etc/passwd on the filer
            addPasswdEntry($set_h->{fc}, 'administrator', 0, 0);
            croak $@ if $@;
            $admin_added = 1;
            # Create root group from /etc/group on the filer
            addGroupEntry($set_h->{fc}, 'root', 0);
            croak $@ if $@;
        }

        # If CIFS clients are specified, setup CIFS on the filer
        # and share (export) the volume and map it on the client
        logStep('Setup CIFS on the filer and map the volume on the client');
        $op = create_and_map_cifs_share(cifs_client=>$opts{cifs_clients}[0])
          or croak $@;
        $set_h->{cifs_drive}[0] = $op;

        # Validate windows clients
        my @cifs_clients = @{$opts{cifs_clients}};
        # The first CIFS client has already been validated
        # so remove it from the list
        shift @cifs_clients;
        foreach $client ( @cifs_clients ) {
            # Make sure windows client has proper environment
            minwin_compatible($client) or croak $@;

            # Map volume on client
            $op = map_cifs_share(server => $set_h->{hostname},
                                 client => $client,
                                 share  => $main::CIFS_SHARE,
                                 drive  => '*') or croak $@;
            push @{$set_h->{cifs_drive}}, $op;
        }
    }

    # Save NFSv4 state
    logStep('Save NFSv4 state');
    $nfs_v4_enable = filer_options_get($set_h->{fc}, 'nfs.v4.enable');
    $nfs_mount_rootonly = filer_options_get($set_h->{fc}, 'nfs.mount_rootonly');
    $nfs_v4_rdel   = filer_options_get($set_h->{fc}, 'nfs.v4.read_delegation')
      if $opts{enable_read_delegations};
    $nfs_v4_wdel   = filer_options_get($set_h->{fc}, 'nfs.v4.write_delegation')
      if $opts{enable_write_delegations};
    $nfs_v4_acl    = filer_options_get($set_h->{fc}, 'nfs.v4.acl.enable')
      if $opts{enable_acls};

    logStep('Enable NFSv4 read and write delegations');
    if ( $nfs_v4_enable eq 'off' ) {
        filer_options_set($set_h->{fc}, 'nfs.v4.enable', 'on');
    }
    if ( $nfs_v4_rdel eq 'off' ) {
        filer_options_set($set_h->{fc}, 'nfs.v4.read_delegation', 'on');
    }
    if ( $nfs_v4_wdel eq 'off' ) {
        filer_options_set($set_h->{fc}, 'nfs.v4.write_delegation', 'on');
    }
    if ( $nfs_v4_acl eq 'off' ) {
        filer_options_set($set_h->{fc}, 'nfs.v4.acl.enable', 'on');
    }
    if ($nfs_mount_rootonly eq 'on') {
        filer_options_set($set_h->{fc}, 'nfs.mount_rootonly', 'off');
    }

    # With the license v2 and nfsv4, the nfs server must be turned
    # on explicitly.
    say($set_h->{fc},"nfs on");

    logStep('Mount the test volume');
    $mountpoint = mount($set_h->{hostname},
                        $set_h->{volume},
                        mntopts => 'vers=4',
                        client  => $set_h->{cc}[0],
                        'chmod' => '0777') or croak $@;

    logStep('Create test file for read delegations');
    $set_h->{tfname}     = "NFSv4Delegations.$$";
    $set_h->{tfcontents} = "Test file for NFSv4 delegations";
    $file = "$mountpoint/$set_h->{tfname}";
    ($set_h->{tfinode}) = ( client_write_file($set_h->{cc}[0],
                                              $file,
                                              $set_h->{tfcontents},
                                              mode=>0666) )[1];

    # Save mountpoint so if setup should fail,
    # it will get properly unmounted
    $set_h->{setup_mtpoint} = $mountpoint;

    if ( $opts{stream} ) {
        logStep('Create stream on test file');
        $set_h->{tsname}     = "NFSv4stream.$$";
        $set_h->{tscontents} = "Test stream for NFSv4 delegations";
        ($set_h->{tsinode}) = ( write_stream(client   => $set_h->{cc}[0],
                                             stream   => $set_h->{tsname},
                                             contents => $set_h->{tscontents},
                                             file     => $file,
                                             sstat    => 1) )[1];
    }

    if ( defined($opts{qtree}) ) {
        logStep("Create a $opts{qtree} mode qtree");
        $set_h->{qtree_name} = "$set_h->{tfname}.qtree";
        $set_h->{qtree_mode} = $opts{qtree};
        $set_h->{qtree} = "$set_h->{volume}/$set_h->{qtree_name}";
        filer_qtree_create($set_h->{fc}, $set_h->{qtree});

        # Change the security mode of the qtree
        filer_qtree_security_set($set_h->{fc},
                                 $set_h->{qtree},
                                 $set_h->{qtree_mode});

        logStep('Create a file on the qtree');
        $file = "$mountpoint/$set_h->{qtree_name}/$set_h->{tfname}";
        ($set_h->{qfinode}) = ( client_write_file($set_h->{cc}[0],
                                                  $file,
                                                  $set_h->{tfcontents},
                                                  mode=>0666) )[1];

        if ( $opts{stream} ) {
            logStep('Create stream on qtree file');
            ($set_h->{qsinode}) =
                ( write_stream(client   => $set_h->{cc}[0],
                               stream   => $set_h->{tsname},
                               contents => $set_h->{tscontents},
                               file     => $file,
                               sstat    => 1) )[1];
        }
    }

    if ( $opts{setup} ) {
        # Run specific setup for test
        &{$opts{setup}}( $set_h );
        # Just in case the specific setup function returns eval
        croak $@ if $@;
    }

    logStep('Unmount test volume');
    umount($mountpoint, client=>$set_h->{cc}[0]) or croak $@;

    # Already unmounted, no need to unmount it anymore at
    # clean_env if setup should fail
    undef $set_h->{setup_mtpoint};

    #logStep('Unexport the test volume');
    #unexport_path($set_h->{fc}, $set_h->{volume}) or croak $@;

    logStep('End of 7G/7M specific set_env');
    logStep('END: set_env');
    return 1;
}

#---------------------------------------------------------------------

#---------------------------------------------------------------------
## @name ::clean_env
## @description
##  Cleanup after a test run if necessary by removing the files
##  created on the filer.
##
## @synopsis
##  $op = clean_env( %setup_h [, cleanup=>\&cleanup_f] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables are initialized on setup.
##
## @arg cleanup
##  Optional: Reference to a function to run after mounting volume on
##  client and before calling common cleanup.
##
## @return
##  Nothing, on error logs a warning.
#---------------------------------------------------------------------
sub clean_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            cleanup => { type => CODEREF, optional => 1 }
        }
    );
    my $mountpoint = $set_h->{setup_mtpoint};

    logStep("CLEANUP == $main::CLEANUP");

    # Catch errors to log a warning
    eval {
        if ( ! $set_h->{fc} ) {
            # Filer connection is not defined, so re-connect
            logStep("Re-connect to the filer");
            host($main::FILER);
            $set_h->{fc} = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
            $set_h->{fc}->timeout($main::CMDTIMEOUT);
        }

        if ( $main::CLEANUP ne 'none' and $set_h->{cc}[0]
          and $set_h->{volume} ) {
            if ( $set_h->{cifs_drive} ) {
                logStep('Unmap volume on CIFS clients');
                my $N = @{$set_h->{cifs_drive}} - 1;
                foreach my $i ( 0..$N ) {
                    unmap_cifs_share(mappath => $set_h->{cifs_drive}[$i],
                                     client  => $set_h->{cifs_host}[$i]);
                    logWarn($@) if $@;
                }
                if ( $admin_added ) {
                    # Remove administrator account from /etc/passwd on the filer
                    delPasswdEntry($set_h->{fc}, 'administrator');
                    logWarn($@) if $@;
                    # Remove root group from /etc/group on the filer
                    delGroupEntry($set_h->{fc}, 'root');
                    logWarn($@) if $@;
                }
            }

            eval {
                if ( ! $mountpoint ) {
                    my $h = host ($main::FILER);
                    logStep("Export test volume");
                    # Export volume as read-write with root access for 7-mode
                    export_path($set_h->{fc}, $set_h->{volume},
                                     OPTIONS=>"rw,anon=0") or croak $@;

                    logStep("Mount volume on client");
                    $mountpoint = mount($set_h->{hostname},
                                        $set_h->{volume},
                                        mntopts=>'vers=4',
                                        client=>$set_h->{cc}[0],
                                        mountretry=>2) or croak $@;
                } else {
                    logStep("Volume already mounted on client");
                }

                if ( $opts{cleanup} ) {
                    # Run specific cleanup for test
                    # Pass the mountpoint to the specific cleanup function
                    $set_h->{setup_mtpoint} = $mountpoint;

                    # Eval the specific cleanup function just in
                    # case the function does not return eval, so
                    # the error can be properly log as a warning.
                    eval {
                        &{$opts{cleanup}}( $set_h );
                        # Just in case the function does returns eval
                        croak $@ if $@;
                    };
                    logWarn("clean_env: $@") if $@;
                }

                if ( $set_h->{tfname} ) {
                    logStep("Delete all files created by this test");
                    eval { client_del_file($set_h->{cc}[0],
                                           "$mountpoint/$set_h->{tfname}*",
                                           recursive=>1) };
                }

                logStep("Unmount volume on client");
                umount($mountpoint, client=>$set_h->{cc}[0]);
                logWarn($@) if $@;

                # Make sure all clients unmounts the exclusive mounts
                if ( $set_h->{mountpoint} ) {
                    foreach my $conn ( @{$set_h->{nfsv4cc}} ) {
                        umount($set_h->{mountpoint}, client=>$conn, force=>1);
                    }
                }
                # Do not try to unmount in common_cleanup
                undef $mountpoint;
            };
            logWarn($@) if $@;

            eval {
                logStep('Restore NFSv4 state');
                filer_options_set($set_h->{fc}, 'nfs.v4.enable',
                                  $nfs_v4_enable) if $nfs_v4_enable;

                filer_options_set($set_h->{fc}, 'nfs.v4.acl.enable',
                                  $nfs_v4_acl) if $nfs_v4_acl;

                filer_options_set($set_h->{fc}, 'nfs.v4.read_delegation',
                                  $nfs_v4_rdel) if $nfs_v4_rdel;

                filer_options_set($set_h->{fc}, 'nfs.v4.write_delegation',
                                  $nfs_v4_wdel) if $nfs_v4_wdel;
                
                filer_options_set($set_h->{fc}, 'nfs.mount_rootonly',
                                  $nfs_mount_rootonly) if $nfs_mount_rootonly ;
                 
            };
            logWarn($@) if $@;
            logStep("Unexport volume");

        }

    };

    logWarn($@) if $@;
    logStep("END: CLEANUP");

    return;
}

#---------------------------------------------------------------------
## @name ::check_delegation
## @description
##  Checks if delegation exists for a given file and client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_delegation( %options );
##
## @arg fc
##  Required: Filer connection handle.
##
## @arg inode
##  Required: Inode of file.
##
## @arg ipaddr
##  Required: IP address of owner of delegation.
##
## @arg type
##  Required: Check for specific type of delegation ('read' or 'write').
##
## @arg buffer
##  Optional: Do not get delegations from the filer, use the buffer instead.
##  This is useful to check for a series of delegations without having
##  to fetch all delegations for each check, (default is 0).
##
## @arg partner
##  Optional: Check for delegation on partner in a cluster pair,
##  (default is 0).
##
## @return
##  1 if delegation exists
#---------------------------------------------------------------------
sub check_delegation(%;) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            inode   => { type => SCALAR,    regex => qr/^\d+$/ },
            ipaddr  => { type => SCALAR,    regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
            type    => { type => SCALAR,    regex => qr/^(read|write)$/ },
            buffer  => { type => BOOLEAN, default => 0 },
            partner => { type => BOOLEAN, default => 0 },
            file    => { type => SCALAR,  optional => 1 },
        }
    );

    my ($lock_h, $inode, $owner, $out);

    # Hash to map delegation names on filer to input types
    my %delegation = ( 'Deleg-Read'=>'read', 'Deleg-Wrt'=>'write' );
    
     my $h = host ($main::FILER);
    # Get all delegations on filer -- grouped by file inode
    if ( ! $opts{buffer} ) {
        @file_locks = filer_lock_status_file_get( $opts{fc},
                                                  partner=>$opts{partner} );
    }

    foreach my $lock_h ( @file_locks ) {
        if ( $delegation{$lock_h->{mode}} eq $opts{type}
          and $lock_h->{inode} == $opts{inode}
          and $lock_h->{ipaddr} eq $opts{ipaddr} ) {
            # Found correct delegation for inode on IP address
            logTrc("Delegation granted <$delegation{$lock_h->{mode}}>");
            return 1;
        }
    }

    # Delegation not found
    my $inode = sprintf('0x%x', $opts{inode});
    croak "Could not find $opts{type} delegation for " .
          "inode $opts{inode} <$inode> on client $opts{ipaddr}";

}

#---------------------------------------------------------------------
## @name ::check_oplock
## @description
##  Checks if an oplock exists for a given file and client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_oplock( %options );
##
## @arg fc
##  Required: Filer connection handle.
##
## @arg inode
##  Required: Inode of file.
##
## @arg ipaddr
##  Required: IP address of owner of oplock.
##
## @arg type
##  Required: Check for specific type of oplock ('excl', 'lvl2', 'none', 'byte').
##  Checking for 'byte' verifies that the given client has a byte range
##  lock on the file.
##
## @arg buffer
##  Optional: Do not get oplocks from the filer, use the buffer instead.
##  This is useful to check for a series of oplocks without having
##  to fetch all oplocks for each check, (default is 0).
##
## @arg partner
##  Optional: Check for delegation on partner in a cluster pair.
##
## @return
##  1 if oplock exists
#---------------------------------------------------------------------
sub check_oplock(%;) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            inode   => { type => SCALAR,  regex => qr/^\d+$/ },
            ipaddr  => { type => SCALAR,  regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
            type    => { type => SCALAR,  regex => qr/^(excl|lvl2|none|byte)$/ },
            buffer  => { type => BOOLEAN, default => 0 },
            partner => { type => BOOLEAN, default => 0 }
        }
    );
    # Hash to map oplock names on filer to input types
    my %oplock = ( 'Oplock-Excl'=>'excl', Excl=>'excl', 'Oplock-Lvl2'=>'lvl2', Lvl2=>'lvl2', None=>'none' );

    ## Check if filer is c-mode or 7 mode
    my $h = host ($main::FILER);
    # Get all locks on filer -- grouped by file inode
    	if ( ! $opts{buffer} ) {
       		@file_locks = filer_lock_status_file_get( $opts{fc},
                                                  partner=>$opts{partner} );
    	}
    
    if ( $opts{type} ne 'byte' ) {
        # Check for specified oplock type
        foreach my $lock_h ( @file_locks ) {
            if ( $oplock{$lock_h->{oplock}} eq $opts{type}
                 and $lock_h->{inode} == $opts{inode}
                 and $lock_h->{ipaddr} eq $opts{ipaddr} ) {
                # Found correct oplock for inode on IP address
                logTrc("Oplock granted <$oplock{$lock_h->{oplock}}>");
                return 1;
            }
        }

        # Oplock not found
        my $inode = sprintf('0x%x', $opts{inode});
        croak "Could not find $opts{type} oplock for " .
          "inode $opts{inode} <$inode> on client $opts{ipaddr}";
    } else {
        # Check for byte range lock
        foreach my $lock_h ( @file_locks ) {
            if ( defined $lock_h->{sbyte}
                 and $lock_h->{inode} == $opts{inode}
                 and $lock_h->{ipaddr} eq $opts{ipaddr} ) {
                # Found a byte lock for inode on IP address
                logTrc("Bytelock granted");
                return 1;
            }
        }

        # Bytelock not found
        my $inode = sprintf('0x%x', $opts{inode});
        croak "Could not find $opts{type} oplock for " .
          "inode $opts{inode} <$inode> on client $opts{ipaddr}";
    }
   #end c-mode
}

#---------------------------------------------------------------------
## @name ::get_delegation_and_check_recall
## @description
##  Get a delegation on a file or stream and check whether the
##  delegation is recalled by the operation on the file/stream
##  from another client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = get_delegation_and_check_recall( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg file
##  Required: Get delegation on this file.
##
## @arg inode
##  Required: Inode of file/stream to get delegation for.
##
## @arg client
##  Required: Connection handle for NFSv4 client to get delegation on.
##
## @arg mountpoint
##  Required: The directory to mount the volume onto.
##
## @arg delegation
##  Optional: Type of delegation to get: 'read' or 'write',
##  (default is 'read').
##
## @arg stream
##  Optional: Get delegation on this stream instead of the file.
##
## @arg oper_client
##  Optional: Connection handle for client where operation is performed.
##  (default is the same as option 'client').
##
## @arg oper_mtpoint
##  Optional: The directory on the operation client to mount the volume onto.
##  (default is the same as option 'mountpoint').
##
## @arg operation
##  Optional: Reference to a function to execute on the NFS client.
##
## @arg oper_args
##  Optional: Reference to a list of arguments to pass to the operation
##  function.
##
## @arg operation_after_ck
##  Optional: Reference to a function to execute on the NFS client
##  after the delegation has been checked for recalled.
##  This is used to revert any destructive changes made by 'operation'.
##
## @arg oper_args_after_ck
##  Optional: Reference to a list of arguments to pass to the
##  operation_after_ck function.
##
## @arg NFSversion
##  Optional: NFS version to use on client, (default is 4).
##
## @arg delegation_expected
##  Optional: Expect to get a delegation (default is 1).
##
## @arg delegation_recalled
##  Optional: Expect a delegation to be recalled (default is 0).
##
## @arg message
##  Optional: Message to log as a step before performing operation.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub get_delegation_and_check_recall(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h               => { type => HASHREF },
            file                => { type => SCALAR },
            inode               => { type => SCALAR },
            client              => { type => OBJECT },
            mountpoint          => { type => SCALAR },
            delegation          => { type => SCALAR,    default => 'read',
                                    regex => qr/^(read|write)$/ },
            stream              => { type => SCALAR,   optional => 1 },
            oper_client         => { type => OBJECT,   optional => 1 },
            oper_mtpoint        => { type => SCALAR,   optional => 1 },
            operation           => { type => CODEREF,  optional => 1 },
            oper_args           => { type => ARRAYREF, optional => 1,
                                  depends => ['operation'] },
            operation_after_ck  => { type => CODEREF,  optional => 1 },
            oper_args_after_ck  => { type => ARRAYREF, optional => 1,
                                  depends => ['operation_after_ck'] },
	    NFSversion  => { type => SCALAR,   default => '4' },
            delegation_expected => { type => BOOLEAN,   default => 1 },
            delegation_recalled => { type => BOOLEAN,   default => 0 },
            message             => { type => SCALAR,
                                  default => q(Perform operation) },
            message_after_ck    => { type => SCALAR,
                                  default => q(Perform operation after check) },
        }
    );
    my ($mountpoint, $unmount_flag, $file, $stream, $err, $not, $op);
    my $fc         = $opts{set_h}{fc};
	my $hostname= $opts{set_h}{hostname};
    my $volume     = $opts{set_h}{volume};
    my $tfcont     = $opts{set_h}{tfcontents};
    my $cc         = $opts{client};
    my $conn       = defined($opts{oper_client}) ? $opts{oper_client} : $cc;
    my $NFSversion = $opts{NFSversion};
    my $recalled   = $opts{delegation_recalled};
    my $mtpoint    = defined($opts{oper_mtpoint}) ? $opts{oper_mtpoint}
                                                  : $opts{mountpoint};
    logStep('BEGIN: get_delegation_and_check_recall');

    # Check if it is a CIFS client
    if ( $conn->hostp()->hosttype() =~ /^windows/ ) {
        $NFSversion = 0;
    }

    # Connect to filer
    my $f = host($main::FILER);

    logStep('Export the volume');
    export_path($fc, $volume, OPTIONS=>'rw,anon=0') or croak $@;

    # Catch any errors to properly unexport the volume
    eval {
        logStep('Mount the volume on the client');
        $mountpoint = mount($hostname,
                            $volume,
                            mntopts    => 'vers=4',
                            sharing    => 'exclusive',
                            mountpoint => $opts{mountpoint},
                            client     => $cc) or croak $@;

		# WORK AROUND ALERT: from Archana Ramani. Need this to 
		# ensure the subsequent tests get a delegation when they open a file.
		logStep('Do dummy OPEN and then Sleep');
		my $file1 = "$mountpoint/delegfile";
		client_write_file($cc, $file1, $tfcont);
		sleep(10);

        # Catch any errors to properly unmount the volume
        eval {
            # File as seen by the client
            $file = "$mountpoint/$opts{file}";
            # Get a read/write delegation on a file/stream
            my $msg = "Get a $opts{delegation} delegation on a";
            if ( defined($opts{stream}) ) {
                logStep("$msg stream");
                $stream = $opts{stream};
                # Read/Write stream to get delegation
                if ( $opts{delegation} eq 'read' ) {
                    read_stream(client => $cc,
                                stream => $stream,
                                file   => $file);
                } else {
                    write_stream(client   => $cc,
                                 stream   => $stream,
                                 contents => $tfcont,
                                 file     => $file);
                }
            } else {
                logStep("$msg file");
                # Read/Write file to get delegation
                if ( $opts{delegation} eq 'read' ) {
                    client_read_file($cc, $file);
                } else {
                    client_write_file($cc, $file, $tfcont);
                }
            }

            logStep('Verify proper granting of delegation');
            my %delegation_h = (fc     => $fc,
                                inode  => $opts{inode},
                                ipaddr => $cc->connhost(),
				file   => $opts{file},
                                type   => $opts{delegation});
            $op = eval { check_delegation(%delegation_h) };

            if ( $op xor $opts{delegation_expected} ) {
                $not = " not" if not $opts{delegation_expected};
                $file = defined($opts{stream}) ? "stream" : "file";
                croak "$opts{delegation} delegation should$not " .
                      "have been granted on $file: $@";
            }

            return 1 if not defined($opts{operation});

            # Mount volume on unix client only since the windows
            # client has already been mapped at setup
            if ( $NFSversion == 4 and $mtpoint eq $mountpoint
              and $cc->connhost() eq $conn->connhost() ) {
                # No need to mount volume, it is the same client that
                # holds the delegation and the same NFS version, so
                # it is already mounted.
                # Also, make sure the volume is not unmounted
                $unmount_flag = 0;
            } elsif ( $NFSversion ) {
                logStep('Mount the volume on client B');
                $mtpoint = mount($hostname,
                                 $volume,
                                 mntopts    => "vers=$NFSversion",
                                 sharing    => 'exclusive',
                                 mountpoint => $mtpoint,
                                 client     => $conn) or croak $@;
                # Make sure volume is unmounted
                $unmount_flag = 1;
            }

            # Perform operation
            logStep($opts{message});
            &{$opts{operation}}( @{$opts{oper_args}} );
            croak $@ if $@;

            # Verify that the delegation is properly recalled
            $not = ' not' if not $recalled;
            logStep("Verify that the $opts{delegation} " .
                    "delegation is$not recalled");
            eval { check_delegation(%delegation_h) };

            if ( $recalled and $@ =~ qr/Could not find .* delegation/ ) {
                # Expecting a recalled
                undef $@;
            } elsif ( $recalled and ! $@ ) {
                croak "Delegation should have been recalled";
            }
            croak $@ if $@;

            if ( $opts{operation_after_ck} ) {
                # Perform operation after check
                logStep($opts{message_after_ck});
                &{$opts{operation_after_ck}}(@{$opts{oper_args_after_ck}});
                croak $@ if $@;
            }
        };

        $err = $@;

        if ( $mtpoint and $unmount_flag ) {
            logStep('Unmount the volume on client B');
            umount($mtpoint, client=>$conn, timeout=>5);
            umount($mtpoint, client=>$conn) if $@;
            $err .= " $@" if $@;
        }

        logStep('Unmount the volume');
        umount($mountpoint, client=>$cc, timeout=>5);
        umount($mountpoint, client=>$cc) if $@;
        croak "$err $@" if ($err or $@);
    };

    $err = $@;
       logStep('Unexport the volume');
       unexport_path($fc, $volume);

       croak "$err $@" if ($err or $@);

    logStep('END: get_delegation_and_check_recall');

    return 1;
}

#---------------------------------------------------------------------
## @name ::get_lock_owner
## @description
##  Return the process id and the effective user id of the current
##  lock.
##
## @synopsis
##  $lock_owner = get_lock_owner();
##
## @return
##  Hash reference with the following keys: 'pid' and 'uid'.
#---------------------------------------------------------------------
sub get_lock_owner() {
    return { pid => $lock_pid, uid => $lock_uid };
}

#---------------------------------------------------------------------
## @name ::file_lock_send
## @description
##  Send message to the file open by get_file_lock().
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = file_lock_send( %options );
##
## @arg lockid
##  Optional: Lock id to send message,
##  Default is current id set by file_lock_send.
##
## @arg message
##  Required: Message to send to file: 'flock' or 'close',
##  (default is 'close').
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub file_lock_send(%;) {
    my (%opts) = validate (
        @_,
        {
            lockid  => { type => OBJECT, default => $lockid },
            message => { type => SCALAR, default => 'close',
                                           regex => qr/^(flock|close)$/ },
            timeout => { type => SCALAR, default => $main::CMDTIMEOUT },
        }
    );

    if ( $opts{lockid}->is_alive() ) {
        # Send message to subtest to terminate
        $opts{lockid}->ipc()->write_message($opts{message});

        if ( $opts{message} eq 'close' ) {
            # Wait for the subtest to finish
            $opts{lockid}->wait_finish(timeout=>$opts{timeout});
        }
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::get_file_lock
## @description
##  Open a file from a client, lock the file and hold it open until
##  instructed to release it using the 'file_lock_send' function.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = get_file_lock( %options );
##
## @arg client
##  Required: Hostspec, Hostrec or Connectrec of client.
##
## @arg file
##  Required: File to open.
##
## @arg user
##  Optional: Open the file as this user.
##
## @arg gid
##  Optional: Open the file with this group id.
##
## @arg open_type
##  Optional: Type of open: 'read' or 'write', (default is 'read').
##
## @arg lock_type
##  Optional: Type of lock to get: 'flock', 'excl' or 'lvl2', (default 'none').
##
## @arg start_byte
##  Optional: Starting byte of range lock, (default is 0).
##
## @arg byte_length
##  Optional: Number of bytes of range lock, (default is 0).
##
## @arg xstart_byte
##  Optional: Starting byte of extended range lock.
##  If option is passed process waits for a signal to extend the range lock.
##
## @arg xbyte_length
##  Optional: Number of bytes of extended range lock.
##  If option is passed process waits for a signal to extend the range lock.
##
## @arg lock_wait
##  Optional: Wait for file_lock_send to signal to get the actual lock.
##  Must use this option with 'flock' lock type, (default is 0).
##
## @arg upgrade_lock
##  Optional: Upgrade a read lock to a write lock, (flock only, default is 0).
##
## @arg xopen_type
##  Optional: Open the file with the proper type (default is write)
##
## @arg upgrade_lock_sig
##  Optional: Wait for file_lock_send signal to upgrade a read lock to a write lock, (flock only, default is 0).
##
## @arg downgrade_lock_sig
##  Optional: Wait for file_lock_send signal to downgrade a write lock to a read lock, (flock only, default is 0).
##
## @arg lvl2_info
##  Optional: NFSv4 client information hash. To get a level 2 oplock
##  on a CIFS client, first get a delegation on the file and then
##  the open of the file will give the level 2 oplock instead of an
##  exclusive oplock.  The information hash contains:
##    fc     => Connection handle for filer,
##    cc     => Connection handle for NFSv4 client,
##    volume => Volume on filer where file resides.
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out,
##  (default is 0, wait forever).
##
## @return
##  lock id.
#---------------------------------------------------------------------
sub get_file_lock(%;) {
    my (%opts) = validate (
        @_,
        {
            client       => { type => SCALAR|OBJECT },
            file         => { type => SCALAR },
            user         => { type => SCALAR,  optional => 1 },
            gid          => { type => SCALAR,  optional => 1 },
            lvl2_info    => { type => HASHREF, optional => 1 },
            lock_wait    => { type => BOOLEAN,  default => 0 },
            timeout      => { type => SCALAR,   default => 0 },
            start_byte   => { type => SCALAR,   default => 0,
                             regex => qr/^\d+$/ },
            xopen_type   => { type => SCALAR,   optional => 1,
                              regex => qr/^(read|write)$/ },
            byte_length  => { type => SCALAR,   default => 0,
                             regex => qr/^\d+$/ },
            upgrade_lock_sig => { type => BOOLEAN,  default => 0 },
            downgrade_lock_sig => { type => BOOLEAN,  default => 0 },
            xstart_byte  => { type => SCALAR,  optional => 1,
                             regex => qr/^\d+$/ },
            xbyte_length => { type => SCALAR,  optional => 1,
                             regex => qr/^\d+$/ },
            upgrade_lock => { type => BOOLEAN,  default => 0 },
            open_type    => { type => SCALAR,   default => 'read',
                             regex => qr/^(read|write)$/ },
            lock_type    => { type => SCALAR,   default => 'none',
                             regex => qr/^(none|excl|lvl2|flock)$/ },
        }
    );
    my (%extra_options, $mountpoint, $msgs, $cc, $op);
    my $file = $opts{file};

   my $h = host ($main::FILER);

    # Get Hostrec object from the client option
    if ( UNIVERSAL::isa($opts{client}, "Connectrec") ) {
        # Get hostrec object from a connectrec object
        $opts{client} = $opts{client}->hostp();
    } else {
        # Get hostrec object from a hostspec
        $opts{client} = host($opts{client});
    }

    # Validate arguments
    if ( ($opts{lock_type} eq 'lvl2' or $opts{lock_type} eq 'excl')
      and $opts{client}->hosttype() !~ /^windows/ ) {
        croak "Exclusive or level 2 oplocks requires a CIFS client";
    }
    if ( $opts{lock_wait} and $opts{lock_type} ne 'flock' ) {
        croak "Option 'lock_wait' must be used with 'flock' lock type";
    }
    if ( $opts{upgrade_lock} and $opts{lock_type} ne 'flock' ) {
        croak "Option 'upgrade_lock' must be used with 'flock' lock type";
    }

    if ( $opts{lock_type} eq 'lvl2' ) {
        if ( ! defined($opts{lvl2_info}) ) {
            croak "Level 2 oplock requires option 'lvl2_info'";
        }
        $cc = $opts{lvl2_info}{cc};
        # To get a level 2 oplock, first get
        # read delegation and then open the file on CIFS
        $mountpoint = mount($opts{lvl2_info}->{data_server},
                            $opts{lvl2_info}{volume},
                            client  => $cc,
                            mntopts => "vers=4") or croak $@;

        # Remove drive off the filename
        my $afile = $file;
        $afile =~ s/^.://;
        $afile =~ s{\\}{/}g;

        if ( $afile =~ /:/ ) {
            # Get a read delegation on the stream
            my ($bfile, $stream) = ($afile =~ /(.*):(.*)/);
            read_stream(client => $cc,
                        stream => $stream,
                        file   => "$mountpoint/$bfile");
        } else {
            # Get a read delegation on the file
            client_read_file($cc, "$mountpoint/$afile");
        }
    }

    my $flock_cmd   = '';
    my $upgrade_cmd = '';
    my $xflock_cmd  = '';
    my $gid_cmd     = '';
    my $user = defined($opts{user}) ? $opts{user} : '';
    my $xopen;

    # Setup open and lock type
    my $otype  = $opts{open_type} eq 'read' ? 'O_RDONLY' : 'O_WRONLY';
    my $ltype  = $opts{open_type} eq 'read' ? 'F_RDLCK'  : 'F_WRLCK';
    my $ctype  = $opts{open_type} eq 'read' ? 1 : 2;
    my $start  = defined($opts{start_byte})  ? $opts{start_byte}  : 0;
    my $len    = defined($opts{byte_length}) ? $opts{byte_length} : 0;
    my $xstart = $opts{xstart_byte};
    my $xlen   = $opts{xbyte_length};

    if ( defined($opts{xopen_type})) {
       $xopen  = $opts{xopen_type} eq 'read' ? 'F_RDLCK'  : 'F_WRLCK';
      }

    my $command = qq{\$socket->write_message(qq(--open_ok--\$\$--\$>));};
    # Process the 'upgrade_lock_sig' option
    if ( $opts{upgrade_lock_sig} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        # start with a read lock
        $ltype = 'F_RDLCK';
        # upgrade to a write lock
        $upgrade_cmd = qq{
            logTrc(qq(\$lock_no: Wait to extend lock on \$file));
            $command
            \$socket->read_message();
            \$lock = pack(\$flock_st, F_WRLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
            if ( logCond('dbg') ) \{
                eval \{ filer_lock_status_file_get(\$fc) \};
            \}
        };
    }

    # Process the 'upgrade_lock' option
    if ( $opts{upgrade_lock} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        # start with a read lock
        $ltype = 'F_RDLCK';
        # upgrade to a write lock
        $upgrade_cmd = qq{
            \$lock = pack(\$flock_st, F_WRLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
            if ( logCond('dbg') ) \{
                eval \{ filer_lock_status_file_get(\$fc) \};
            \}
        };
    }

    # Process the 'downgrade_lock_sig' option
    if ( $opts{downgrade_lock_sig} ) {
        # Ignore setting of open_type and open file as rdrw
        $otype = 'O_RDWR';
        # start with a read lock
        $ltype = 'F_WRLCK';
        # upgrade to a write lock
        $upgrade_cmd = qq{
            logTrc(qq(\$lock_no: Wait to extend lock on \$file));
            $command
            \$socket->read_message();
            \$lock = pack(\$flock_st, F_RDLCK, SEEK_SET, \@range);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
            if ( logCond('dbg') ) \{
                eval \{ filer_lock_status_file_get(\$fc) \};
            \}
        };
    }


    if ( defined($xstart) ) {
        # Check if open type is specified 
        if ( defined($opts{xopen_type})) {
                $ltype = $xopen;
                $otype = 'O_RDWR';
        }
        # Wait for signal to extend the range of lock
        $xflock_cmd = qq{
            logTrc(qq(\$lock_no: Wait to extend lock on \$file));
            $command
            \$socket->read_message();
            \$lock = pack(\$flock_st, $ltype, SEEK_SET, \@xrange);
            fcntl(\$FH, F_SETLKW, \$lock) or die \$!;
            if ( logCond('dbg') ) \{
                eval \{ filer_lock_status_file_get(\$fc) \};
            \}
        };
    } else {
        # Make sure these variables are defined
        $xstart = 0;
        $xlen   = 0;
    }

    my $fop;
	if ( $opts{lock_wait} ) {
		$fop = 'F_SETLKW';
	}
	else {
		$fop = 'F_SETLK';
	}
	# Process the 'lock_type' option
    if ( $opts{lock_type} eq 'flock' ) {
        $flock_cmd = qq{
            \$lock = pack(\$flock_st, $ltype, SEEK_SET, \@range);
            fcntl(\$FH, $fop, \$lock) or die \$!;
            if ( logCond('dbg') ) \{
                host(\$main::FILER);
                \$fc = connect(\$main::FILER_CONN);
                eval \{ filer_lock_status_file_get(\$fc) \};
            \}
            $xflock_cmd
            $upgrade_cmd
        };
    }

    	# Process the 'lock_wait' option
    if ( $opts{lock_wait} ) {
		$command = qq{
			logTrc(qq(\$lock_no: Wait for lock on \$file));
            $command
            \$socket->read_message();
            $flock_cmd
        };
    } 
	else {
		$command = qq{
        	$flock_cmd
            $command
        };
    }

    # Process the 'gid' option
    if ( $opts{gid} ) {
        # Change the credentials (gid) to only list
        # the gid specified and the change the uid
        # to the current uid since subtest will be
        # run as root
        $gid_cmd = qq{
            \$) = q($opts\{gid\} $opts\{gid\});
            \$> = $>;
            logTrc(qq\{\$lock_no: Changed credentials to <\$>,\$)>\});
        };
        # Option gid implies user => 'root' since the
        # credentials of the process must be changed
        # and only root can do it
        $opts{user} = 'root';
    }

    logDbg("get_file_lock: file <$file> user <$user> " .
           "open <$opts{open_type}> lock <$opts{lock_type}>");
    ++$lock_no;
    # Function to open the file
    my $mod_file = __FILE__ . ": get_file_lock at line " . (__LINE__ + 1);
    my $get_file_lock = qq{ \{ eval \{
        use POSIX;
        use SPARSE::Util::Logger;
        use filer qw(filer_lock_status_file_get);
        my (\$FH, \$lock, \$fc, \$flock_st, \@range, \@xrange,\$fop);
        my \$file = q($file);
        my \$lock_no = $lock_no;
        my (\$system, \$mtype) = (POSIX::uname)[0,4];
        if (\$mtype =~ /^[ix]\\d?86/) \{
            \$flock_st = 'ssLLLL';
            \@range = ($start, 0, $len, 0);
            \@xrange = ($xstart, 0, $xlen, 0);
        \} elsif (\$mtype eq 'sun4u') \{
            \$flock_st = 'ssLLLLL';
            \@range = (0, 0, $start, 0, $len);
            \@xrange = (0, 0, $xstart, 0, $xlen);
        \} elsif (\$system eq 'AIX') \{
            \$flock_st = 'ssLLLLLLL';
            \@range = (0, 0, 0, 0, $start, 0, $len);
            \@xrange = (0, 0, 0, 0, $xstart, 0, $xlen);
        \} else \{
            die "Could not set flock structure, " .
                "unknown client architecture: \$system \$mtype";
        \}
        $gid_cmd
        my \$socket = Subtest::ipc();
        sysopen(\$FH, \$file, $otype) or die \$!;
        logTrc(qq(\$lock_no: File \$file opened));
        $command
        logTrc(qq(\$lock_no: Waiting to close file));
        \$socket->read_message();
        logTrc(qq(\$lock_no: Closing file \$file));
        close(\$FH);
    \}; die "\$@ ($mod_file)\\n" if \$@; \} };

    if ( $opts{lock_type} eq 'flock'
      and $opts{client}->hosttype() =~ /^windows/ ) {
        # For windows, use flock to lock the whole file
        $get_file_lock =~ s/fcntl.*?F_SETLK.*?;/flock(\$FH,$ctype);/g;
        $get_file_lock =~ s/F_(RD|WR)LCK/0/g;
    }

    # Only append the username if the option is given
    $extra_options{'-username'} = $opts{user} if defined($opts{user});

    # Open the file on the client
    $lockid = subtest( $get_file_lock,
                       '-bg', '-ipc', '-runid', "01_get_file_lock",
                       '-timeout', $opts{timeout},
                       '-client',  $opts{client},
                       %extra_options,
                       '-onexit', sub {} );

    # Wait for client to open the file
    $op = eval { $lockid->ipc()->read_message() };
    if ( $@ || ! (($lock_pid, $lock_uid) = ($op =~ /--open_ok--(\d+)--(\d+)/)) ) {
        # There was an error so release, i.e., terminate subtest.
        eval { file_lock_send(lockid => $lockid) };

        $msgs = join(' ', $lockid->results());
    }

    # Unmount the volume on the NFSv4 clients
    if ( $opts{lock_type} eq 'lvl2' ) {
        umount($mountpoint, client=>$cc, force=>1);
    }

    croak "Error opening the file: $msgs" if $msgs;

    return $lockid;
}

#---------------------------------------------------------------------
## @name ::check_oplock_nfs_interaction
## @description
##  Get an oplock on a file and check its interaction with NFS
##  operations on the same file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_oplock_nfs_interaction( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg file
##  Required: File/stream to get oplock for.
##
## @arg inode
##  Required: Inode of file/stream to get oplock for.
##
## @arg client
##  Required: Hostspec of CIFS client.
##
## @arg nfscc
##  Optional: Connection handle for NFS client,
##  (default is the primary client).
##
## @arg operation
##  Optional: Reference to a function to execute on the NFS client.
##
## @arg oper_args
##  Optional: Reference to a list of arguments to pass to the operation
##  function.
##
## @arg oper_fail
##  Optional: Expect the operation on the file to fail, (default is 0).
##
## @arg mountpoint
##  Optional: The directory to mount the volume onto.
##
## @arg NFSversion
##  Optional: NFS version to use on client, (default is 4).
##
## @arg oplock
##  Optional: Type of oplock to get: 'excl' or 'lvl2', (default is 'excl').
##
## @arg expect_oplock
##  Optional: Type of oplock to expect after operation: 'excl' or 'lvl2'
##  (default is the same as option 'oplock').
##
## @arg oplock_revoked
##  Optional: Expect the oplock to be revoked after the operation,
##  (default is 0).
##
## @arg message
##  Optional: Message to log as a step before performing operation.
##
## @return
## 1 if succeeds.
#---------------------------------------------------------------------
sub check_oplock_nfs_interaction(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h          => { type => HASHREF },
            client         => { type => SCALAR },
            file           => { type => SCALAR },
            inode          => { type => SCALAR },
            nfscc          => { type => OBJECT,   optional => 1 },
            operation      => { type => CODEREF,  optional => 1 },
            oper_args      => { type => ARRAYREF, optional => 1,
                             depends => ['operation'] },
            oper_fail      => { type => SCALAR,    default => 0 },
            mountpoint     => { type => SCALAR,   optional => 1 },
            NFSversion     => { type => SCALAR,    default => 4 },
            oplock         => { type => SCALAR,    default => 'excl',
                               regex => qr/^(excl|lvl2)$/ },
            expect_oplock  => { type => SCALAR,   optional => 1,
                               regex => qr/^(excl|lvl2)$/ },
            oplock_revoked => { type => SCALAR,    default => 0 },
            message        => { type => SCALAR,
                             default => q(Perform operation) },
        }
    );
    my ($mountpoint, $oplockid, $err, $op);
    my $fc     = $opts{set_h}{fc};
    my $hostname=$opts{set_h}{hostname};
    my $data_server=$opts{set_h}{data_server};
    my $volume = $opts{set_h}{volume};
    my $nfsver = $opts{NFSversion};
    my $oplock = $opts{oplock};
    my $cc     = defined($opts{nfscc}) ? $opts{nfscc} : $opts{set_h}{cc}[0];

    # Expected oplock after operation
    my $expect_oplock = defined($opts{expect_oplock}) ?
                        $opts{expect_oplock} : $oplock;

    # Oplock information
    my $h = host($opts{client});
    my %oplock_info = ( fc     => $fc,
                        inode  => $opts{inode},
                        ipaddr => $h->default_ip(),
                        type   => $oplock );

    # Export the volume
    #logStep('Export the volume');
    #export_path($fc, $volume, OPTIONS=>'rw,anon=0') or croak $@;

    # Catch any errors to properly unexport the volume
    eval {
        # Information needed for a level 2 oplock
        my %extra = ( fc => $fc, cc => $cc, volume => $volume, data_server => $data_server );
        # Get an oplock on the file/stream
        $op = $opts{file} =~ /:[^\\]/ ? 'stream' : 'file';
        logStep("Get an oplock on the $op");
        $oplockid = get_file_lock( client    => $opts{client},
                                   file      => $opts{file},
                                   lock_type => $oplock,
                                   lvl2_info => \%extra );

        # Verify the oplock is granted
        logStep('Verify the oplock is granted');
        check_oplock(%oplock_info);

        # No operation is requested
        return if ( ! defined($opts{operation}) );

        # Change the oplock info to expect for given oplock
        $oplock_info{type} = $expect_oplock;

        # Catch any errors to properly release the oplock
        eval {
            # Mount the volume on the NFS client
            logStep('Mount the volume on the NFS client');
            $mountpoint = mount($hostname,
                                $volume,
                                mntopts    => "vers=$nfsver",
                                mountpoint => $opts{mountpoint},
                                sharing    => 'exclusive',
                                client     => $cc) or croak $@;

            # Catch any errors to properly unmount the volume
            eval {
                logStep($opts{message});
                &{$opts{operation}}( @{$opts{oper_args}} );
            };

            if ( $@ xor $opts{oper_fail} ) {
                my $failed = $opts{oper_fail} ? 'failed' :
                                                'been successful';
                $err = "Operation should have $failed: $@";
            } else {
                # Check the status of the oplock
                logStep('Check the status of the oplock');
                eval { check_oplock(%oplock_info) };
                $op = ( $@ =~ qr/Could not find .* oplock/ );
                if ( ( ! $@ and $opts{oplock_revoked} )
                  or ( $op and ! $opts{oplock_revoked} ) ) {
                    my $not = " not" if not $opts{oplock_revoked};
                    $err = "Oplock should$not have been revoked: $@";
                } elsif ( $@ and ! $op ) {
                    $err = $@;
                }
            }

            # Unmount the volume on the NFS client
            logStep('Unmount the volume on the NFS client');
            umount($mountpoint, client => $cc, timeout=>5);
            umount($mountpoint, client => $cc) if $@;
            croak "$err $@" if ($err or $@);
        };

        $err = $@;

        # Release the oplock on the file
        logStep('Release the oplock on the file');
        eval {file_lock_send(lockid => $oplockid) };
        croak "$err $@" if ($err or $@);
    };

    $err = $@;

    # Unexport the volume
      #logStep('Unexport the volume');
      #unexport_path($fc, $volume);
      croak "$err $@" if ($err or $@);

    return 1;
}


1;
