#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/Delegations.pm#18 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

## @summary Library for the NFSv4 delegations tests
##
## @description
##   Common routines for NFSv4 delegations testing in cMode only. Called by nfs::Delegations.
##   This file doesnot provide the complete package, "use nfs::Delegations;" to get the full package.
##
## @usage
##   use nfs::Delegations;
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by rraghav.
## @owner	dl-nfsqa

use strict;

use Tharn;
use Tharn qw(host connect);
use CIFS::Setup_GX qw(cifs_setup);
use ExportFS qw(share_add share_delete);
use nfs::CommonParams;

use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::Utils qw(validate_nfs_version);
use GenericMount qw(mount umount);
use Network::Utils qw(run_getXXbyYY);
use Module::Common qw(validate_mixed);
use TestEnv::MinWin qw(minwin_compatible);
use nfs::Utils qw(validate_nfsv4_clients);
use FileManip qw(read_stream write_stream);
use MapShare qw(map_cifs_share unmap_cifs_share);
use Users qw(addPasswdEntry addGroupEntry delPasswdEntry delGroupEntry);
use FS::Util::Setup qw(create_vol common_cleanup);
use nfs::RemoteFileIO qw(client_read_file client_write_file client_del_file);
use filer qw(filer_options_get filer_options_set filer_lock_status_file_get filer_qtree_create filer_qtree_security_set);
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF HASHREF CODEREF UNDEF);
use Data::Dumper;
use nfs_setup_gx;

use NACL::C::DiagSecdAuthentication;
use NACL::C::Node;
use NACL::C::SystemNode;
use NACL::C::VserverNfs;
use NACL::C::VserverCifs;
use NACL::C::VserverCifsShare;
use NACL::C::VserverLocks;
use NACL::C::VserverLocksNfsv4;

use NACL::CS::NetworkInterface;

use NACL::Exceptions::NoElementsFound qw{:try};
use NACL::APISet::Exceptions::NoMatchingEntriesException qw{:try};

our @EXPORT_OK;
push @EXPORT_OK, qw{get_number_of_locks enable_nfs_versions_in_filer};

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $usermap = '/etc/usermap.cfg';
my $test_file   = "Delegations.$$";
my $nfs_v4_enable;
my $nfs_v4_acl;
my $nfs_v4_rdel;
my $nfs_v4_wdel;
my $nfs_mount_rootonly;
my @file_locks;
my $lock_pid;
my $lock_uid;
my $lockid;
my $lock_no = 0;
my %GH;
my @ALL_NODE_OBJS;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub set_env(\%;%);
sub clean_env(\%;%);
sub check_delegation(%;);
sub check_oplock(%;);
sub get_number_of_delegs(%;);

#---------------------------------------------------------------------
## @name ::set_env
## @description
##  Set up environment for NFSv4 delegations tests and initialize all
##  common variables.  The first argument is the structure to be
##  initialized and the rest of the arguments are options passed
##  in "key => value" style.
##
## @synopsis
##  $op = set_env( %setup_h [, %options ] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables will get initialized in this function.
##
## @arg nfsv4_clients
##  Optional: Reference to list of NFSv4 client hostspecs.
##
## @arg cifs_clients
##  Optional: Reference to list of CIFS client hostspecs.
##
## @arg deleg_clients
##  Optional: Reference to a list of client hostspecs which support
##  delegations. This list must be a subset of list in 'nfsv4_clients'.
##
## @arg stream_clients
##  Optional: Reference to a list of client hostspecs which support
##  named attributes. This list must be a subset of list in 'nfsv4_clients'.
##
## @arg stream
##  Optional: Boolean, create a stream (named attribute) on the test file,
##  (default => 0).
##
## @arg qtree
##  Optional: Mode of qtree to create (unix, mixed, ntfs).
##
## @arg users
##  Optional: Reference to list of user names to check and get their uid's.
##
## @arg groups
##  Optional: Reference to list of group names to check and get their gid's.
##
## @arg client_conn
##  Optional: Clients connection spec for NFSv4 clients, (default => '#rsh').
##
## @arg enable_acls
##  Optional: Flag to enable acls on the filer, (default => 0).
##
## @arg enable_read_delegations
##  Optional: Flag to enable read delegations on the filer, (default => 1).
##
## @arg enable_write_delegations
##  Optional: Flag to enable write delegations on the filer, (default => 1).
##
## @arg setup
##  Optional: Reference to a function to run at the end of setup.
##
## @return
##  Nothing.
##
## @notes
##  Structure used by the NFSv4 delegations tests:
##
##  %setup_h = (
##
##      cc            => Reference to a list of all client connection objects.
##
##      scc           => Reference to a list of all client connection objects
##                       except for the primary client, cc[1...N].
##
##      volume        => Absolute path of volume.
##
##      vol_name      => Volume name.
##
##      tfname        => Test file name.
##
##      tfinode       => Test file inode.
##
##      tfcontents    => Test file contents.
##
##      tsname        => Test stream name.
##
##      tsinode       => Test stream inode.
##
##      tscontents    => Test stream contents.
##
##      qfinode       => Inode for test file on qtree.
##
##      qsinode       => Inode for stream on qtree file.
##
##      qtree         => Absolute path of qtree.
##
##      qtree_name    => Name of qtree.
##
##      qtree_mode    => Mode of qtree (unix, mixed, or ntfs).
##
##      nfsv4cc       => Reference to a list of NFSv4 client connection objects.
##
##      cifscc        => Reference to a list of CIFS client connection objects.
##
##      cifs_drive    => Reference to a list of CIFS maps (mountpoints)
##
##      cifs_host     => Reference to a list of CIFS hostspecs
##
##      users         => Reference to a hash of users: { username => uid }.
##
##      groups        => Reference to a hash of groups: { groupname => gid }.
##
##      mountpoint    => Tentative mountpoint for tests.
##
##      setup_mtpoint => The mount point of volume on the primary client
##                       for setup and cleanup purposes.
##  );
#---------------------------------------------------------------------
sub set_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            nfsv4_clients  => { type => ARRAYREF, optional => 1 },
            cifs_clients   => { type => ARRAYREF, optional => 1 },
			nfs_version  => { type => SCALAR,    default => 'all' },
			minor_version  => { type => SCALAR, optional =>1, default => $::NFS4_MINOR_VERSION, regex => qr/^\d+$/ },
            deleg_clients  => { type => ARRAYREF, optional => 1 },
            stream_clients => { type => ARRAYREF, optional => 1 },
            create_files    => { type => SCALAR, default => 1, },
            client_conn    => { type => SCALAR,    default => '#rsh' },
            stream         => { type => BOOLEAN,   default => 0 },
            users          => { type => ARRAYREF, optional => 1 },
            groups         => { type => ARRAYREF, optional => 1 },
            qtree          => { type => SCALAR,   optional => 1,
                               regex => qr/^(unix|mixed|ntfs)$/ },
            setup          => { type => CODEREF,  optional => 1 },
            enable_acls              => { type => BOOLEAN, default => 0 },
            enable_read_delegations  => { type => BOOLEAN, default => 1 },
            enable_write_delegations => { type => BOOLEAN, default => 1 },
			address_family	=> { type => SCALAR, optional => 1, default => "inet4",
								regex => qr/^(inet4|inet6|ipv4|ipv6)$/},
        }
    );
    my ($mountpoint, $client, $nfsv4, $file, $cc, $h, $op);
    my ($cifs_passwd, $cifs_user, $fh, $fname, $mode,
        $mountpoint, @nfsver, %nfsver_h, $new_text, $op, %qtree_out);

    logStep('BEGIN: set_env: setup environment for test');

    # Connect to filer
    my $f = host($main::FILER);
	get_node_obj();

    # Make sure at least one NFSv4 client is specified
    if ( ! $opts{nfsv4_clients} ) {
        croak "No NFSv4 clients are specified";
    }

    # Validate clients and create connection handles for all clients,
    # start with the NFSv4 clients and when it reaches the separator '-'
    # process the CIFS clients.
    logStep('Validate and create connection handles for clients');

    # Validate that nfsv4 clients support NFSv4
    # and validate clients that support named attributes
    validate_nfsv4_clients( clients        => $opts{nfsv4_clients},
                            deleg_clients  => $opts{deleg_clients},
                            stream_clients => $opts{stream_clients} );

    # Create connection handles for all clients
    $nfsv4 = 1;
    foreach $client ( @{$opts{nfsv4_clients}}, '-', @{$opts{cifs_clients}} ) {
        if ( $client eq '-' ) {
            $nfsv4 = 0;
            next;
        }
        # Validate that clients can not be 'localhost'
        if ( $client =~ /^localhost\b/ ) {
            croak "Clients can not be 'localhost'";
        }

        # Verify client
        $h = host($client);
        if ( $nfsv4 and $h->hosttype() !~ /^unix/ ) {
            croak "Client '$client' must be unix";
        } elsif ( ! $nfsv4 and $h->hosttype() !~ /^windows/ ) {
            croak "Client '$client' must be windows";
        }
        # Create connection for client
        $cc = connect($opts{client_conn}, '-timeout', $main::TIMEOUT);
        $cc->timeout($main::CMDTIMEOUT);

        # Save the connection handle of secondary clients
        push @{$set_h->{scc}}, $cc if defined($set_h->{cc});

        # Save the connection handle
        push @{$set_h->{cc}}, $cc;
        if ( $nfsv4 ) {
            push @{$set_h->{nfsv4cc}}, $cc;
        } else {
            push @{$set_h->{cifscc}}, $cc;
            push @{$set_h->{cifs_host}}, $client;
        }
    }

    ##################################################################
    # Call 10 Mode specific setup modules to make the node test-worthy
    ##################################################################
    logStep ("Starting 10 Mode setup");

	my $h = host($main::FILER);
	my $server_obj = get_server_obj();

    eval {
		nfs_setup_gx::set_env_gx( setup_h  => $set_h,
                 allowed  => [ 3, 4 ],
                 version  => "not2",
				 address_family => $opts{address_family},
                 protocol => 'tcp' );
	};
	if ($@)
	{
		logStep ("Unable to setup filer, set_env_gx failed.");
		croak "set_env: $@";
	}

	my $hostname = $set_h->{hostname};

	if ($opts{minor_version} == 1) {
		$GH{filer_nfs_version}= "4.1";
		$GH{mntopts} = "vers=4,minorversion=1";
	} else {
		$GH{filer_nfs_version}= "4.0";
		$GH{mntopts} = "vers=4";
	}

    $set_h->{mountpoint} = '/t/' . $set_h->{hostname} .
                '/' . $set_h->{volume};

	# Delegations setup!
	my %nfs_options = (vserver=>$::VSERVER_GX_NAME,
						"v$GH{filer_nfs_version}" => "enabled", 
						"v$GH{filer_nfs_version}-acl" => "disabled",
						"v$GH{filer_nfs_version}-read-delegation" => $opts{enable_read_delegations} ? "enabled" : "disabled",
						"v$GH{filer_nfs_version}-write-delegation" => $opts{enable_write_delegations} ? "enabled" : "disabled");

	# Should ACLs be enabled ?
	$nfs_options{"v$GH{filer_nfs_version}-acl"} = 'enabled'
		if $opts{enable_acls};

	$server_obj->vserver_nfs_modify(%nfs_options);

    if ( $opts{users} ) {
        my $msg;
        # Verify the existence of specified users
        # on the filer and get their uid's
        logStep("Check existence of specified users " .
                "on the filer and get their uid's");
        foreach my $user ( @{$opts{users}} ) {
	        my @op;
		    try {
		        $op = NACL::C::DiagSecdAuthentication->translate(command_interface=>get_node_obj(),
                    vserver => $::VSERVER_GX_NAME, 'unix-user-name' => $user, node => 'local');
               	$op = $server_obj -> diag_secd_authentication_translate ( vserver => $::VSERVER_GX_NAME,
                                             'unix-user-name' => $user, node => 'local');
		        $set_h->{users}{$user} = $op;
			} catch NACL::Exceptions::NoElementsFound with {
		        $msg .= "User '$user' must exist on the filer, for vserver: $::VSERVER_GX_NAME. ";
		    };
        }
        croak $msg if $msg;
    }

    if ( $opts{groups} ) {
        my $msg;
        # Verify the existence of specified groups
        # on the filer and get their gid's
        logStep("Check existence of specified groups " .
                "on the filer and get their gid's");
        foreach my $group ( @{$opts{groups}} ) {
	    my @op;
            try {
		        $op = NACL::C::DiagSecdAuthentication->translate(command_interface=>get_node_obj(),
                    vserver => $::VSERVER_GX_NAME, 'unix-group-name' => $group, node => 'local');
               	$op = $server_obj -> diag_secd_authentication_translate ( vserver => $::VSERVER_GX_NAME,
                                             'unix-group-name' => $group, node => 'local');
                $set_h->{groups}{$group} = $op;
			} catch NACL::Exceptions::NoElementsFound with {
		        $msg .= "User '$group' must exist on the filer, for vserver: $::VSERVER_GX_NAME. ";
		    };
        }
        croak $msg if $msg;
    }

    if ( @{$opts{cifs_clients}} ) {
        # Create username to use for cifs based on whether AD is setup or not
        # Default is workgroup
        $fname =~ tr/a-z/A-Z/;
        $cifs_user = "$fname\\$main::CIFS_USER";
        $cifs_passwd = "$main::CIFS_PASSWD";
        if ($main::CIFS_DOMAIN) {
            my $tmp = "$main::CIFS_DOMAIN";
            $tmp =~ s/\s+//g;
            if ($tmp ne '') {
                # Using Domains
                $cifs_user = "$main::CIFS_DOMAIN_ADMIN";
                $cifs_passwd = "$main::CIFS_DOMAIN_ADMIN_PASSWD";
            }
        }

		if ($main::SETUP_DONE == 0){

			# if the cifs server already exists, delete it.
			try {
				my $cifs_server = NACL::C::VserverCifs->delete(command_interface=>get_node_obj(),
										"password" => $::AD_PASSWORD,
										"username" => $::AD_USERNAME,
										vserver => $::VSERVER_GX_NAME);
			} catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
				logStep(qq{Ignoring warning about cifs server not being present});
			};

			sleep 3;
			logStep('Setup CIFS on the filer');
			my $ret_val = cifs_setup(node => $main::FILER,
					cifs_server => $main::CIFS_SERVER,
					cifs_domain_fullname => $main::CIFS_DOMAIN_FULLNAME,
					cifs_domain_shortname=> $main::CIFS_SHORT_DOMAIN,
					ad_server => [$main::AD_SERVER],
					ad_server_ip => [$main::AD_SERVER_IP],
					vserver => $::VSERVER_GX_NAME,
					ad_server_user => $main::AD_USERNAME,
					ad_server_password => $main::AD_PASSWORD,
					rootshare => $main::CIFS_SHARE,
				);

			if ($ret_val == 0) {
				croak(qq{Setting up CIFS has failed. Please search the log for more info on why cifs setup failed.});
			}

		}

		# If cifs is present, then we need security-style = mixed in all the concerned volumes.
		$GH{root_vol} = $server_obj->vserver_show(vserver=>$::VSERVER_GX_NAME)->[0]->{"rootvolume"};
		foreach my $vol ($GH{root_vol}, $::FILER_GX_VOL) {
			NACL::C::Volume->modify(command_interface=>get_node_obj(),
				volume => $vol, 'security-style' => 'mixed', vserver => $::VSERVER_GX_NAME);
		}

    }

     # Validate windows clients
    my @cifs_clients = @{$opts{cifs_clients}};
    foreach my $client ( @cifs_clients ) {
	    # Make sure windows client has proper environment
	    minwin_compatible($client) or croak "nfslocks_setup: $@";

	    # Map volume on client
		logStep ("cifs domain is $main::CIFS_SHORT_DOMAIN");
		# The current share is mapped to the root junction path, not the actual volume junction path.
		# So we need to set the correct path for share name below.
		eval {
			unmap_cifs_share(mappath => $::CIFS_DRIVE, client => $client);
		};
	    $op = map_cifs_share(server => $::CIFS_SERVER,
	                         client => $client,
	                         username => $::CIFS_DOMAIN_ADMIN,
	                         password => $::CIFS_DOMAIN_ADMIN_PASSWD,
	                         share  => $main::CIFS_SHARE . convert_to_cifs($set_h->{volume}),
	                         drive  => $main::CIFS_DRIVE)
			or croak "nfslocks_setup: $@";

		# @#$%!!!! some places use cifs_drive and some others use drive, grrr...
	   	push @{$set_h->{cifs}}, { drive => $op, cifs_drive => $op, cifs_host => $client };
	}

	# Export the test volume correctly.
	# Rather, apply the correct policy to the testvolume.
	#$server_obj->volume_modify(volume=>$main::FILER_GX_VOL, vserver=>$::VSERVER_GX_NAME, policy=>$GH{deleg_policy});

    logStep('Mount test volume');
    $mountpoint = mount($hostname,
                        $set_h->{volume},
                        client => $set_h->{cc}[0],
                        mntopts => $GH{mntopts},
                        chmod  => '0777') or croak "set_env: $@";

    logStep('Create test file for read delegations');
    $set_h->{tfname}     = "NFSv4Delegations.$$";
    $set_h->{tfcontents} = "Test file for NFSv4 delegations";
    $file = "$mountpoint/$set_h->{tfname}";
    ($set_h->{tfinode}) = ( client_write_file($set_h->{cc}[0],
                                              $file,
                                              $set_h->{tfcontents},
                                              mode=>0666) )[1];

    # Save mountpoint so if setup should fail,
    # it will get properly unmounted
    $set_h->{setup_mtpoint} = $mountpoint;

    logStep('Create test files') if $opts{create_files};
    foreach my $index ( 1..$opts{create_files} ) {
        my $tfile = "$test_file.$index";
        my $inode = ( client_write_file($set_h->{cc}[0], "$mountpoint/$tfile", join('', ('A'..'Z', "\n"))x9) )[1];
        push @{$set_h->{files}}, { file => $tfile, inode => $inode };
    }

    if ( $opts{stream} ) {
        logStep('Create stream on test file');
        $set_h->{tsname}     = "NFSv4stream.$$";
        $set_h->{tscontents} = "Test stream for NFSv4 delegations";
        ($set_h->{tsinode}) = ( write_stream(client   => $set_h->{cc}[0],
                                             stream   => $set_h->{tsname},
                                             contents => $set_h->{tscontents},
                                             file     => $file,
                                             sstat    => 1) )[1];
    }
    
    if ( defined($opts{qtree}) ) {
        logStep("Create a $opts{qtree} mode qtree");
        $set_h->{qtree_name} = "$set_h->{tfname}.qtree";
        $set_h->{qtree_mode} = $opts{qtree};
        $set_h->{qtree} = "$set_h->{volume}/$set_h->{qtree_name}";
	
	eval {
	$server_obj->volume_qtree_create( vserver => $::VSERVER_GX_NAME,
					  volume => $set_h->{vol_name},
					  qtree   => $set_h->{qtree_name},
					  );
	};

        # Change the security mode of the qtree
	eval {
	$server_obj->volume_qtree_security( vserver => $::VSERVER_GX_NAME,
					  volume => $set_h->{vol_name},
					  qtree   => $set_h->{qtree_name},
					  'security-style' => $set_h->{qtree_mode}
					  );
	};

        logStep('Create a file on the qtree');
        $file = "$mountpoint/$set_h->{qtree_name}/$set_h->{tfname}";
        ($set_h->{qfinode}) = ( client_write_file($set_h->{cc}[0],
                                                  $file,
                                                  $set_h->{tfcontents},
                                                  mode=>0666) )[1];

        if ( $opts{stream} ) {
            logStep('Create stream on qtree file');
            ($set_h->{qsinode}) =
                ( write_stream(client   => $set_h->{cc}[0],
                               stream   => $set_h->{tsname},
                               contents => $set_h->{tscontents},
                               file     => $file,
                               sstat    => 1) )[1];
        }
    }

    if ( $opts{setup} ) {
        # Run specific setup for test
        &{$opts{setup}}( $set_h );
        # Just in case the specific setup function returns eval
        croak $@ if $@;
    }


    if ( ! $opts{mount} ) {
        logStep('Unmount test volume');
        umount($mountpoint, client=>$set_h->{cc}[0])
          or croak "set_env: $@";

        # Already unmounted, no need to unmount it
        # anymore at nfslocks_cleanup if setup should fail
        delete $set_h->{setup_mtpoint};
    }

    #logStep('END: set_env: hash:' . Dumper($set_h));
    logStep('END: set_env');
    return 1;
}

# Convert the given argument into a cifs format.
# convert all // to /. After that, convert all / to \\
sub convert_to_cifs($) {
	my $string = shift;
	$string =~ s@//@/@g;
	$string =~ s@/@\\@g;
	logStep(qq{convert_to_cifs: returning $string});
	return $string;
}

# Remove the domain name from the given name.
# If no domain is present, it simply returns the given string as it is.
sub remove_domain($) {
	my $string = shift;
	$string =~ s@\..*@@g;
	return $string;
}

#---------------------------------------------------------------------

#---------------------------------------------------------------------
## @name ::clean_env
## @description
##  Cleanup after a test run if necessary by removing the files
##  created on the filer.
##
## @synopsis
##  $op = clean_env( %setup_h [, cleanup=>\&cleanup_f] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables are initialized on setup.
##
## @arg cleanup
##  Optional: Reference to a function to run after mounting volume on
##  client and before calling common cleanup.
##
## @return
##  Nothing, on error logs a warning.
#---------------------------------------------------------------------
sub clean_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            cleanup => { type => CODEREF, optional => 1 }
        }
    );
    my $mountpoint = $set_h->{setup_mtpoint};

    logStep("CLEANUP == $main::CLEANUP");

    # Catch errors to log a warning
    eval {
        if ( $main::CLEANUP ne 'none' and $set_h->{cc}[0]
          and $set_h->{volume} ) {
            if ( $set_h->{cifs} ) {
                logStep('Unmap volume on CIFS clients');
                my $N = @{$set_h->{cifs}} - 1;
                foreach my $i ( 0..$N ) {
                    unmap_cifs_share(mappath => $set_h->{cifs}[$i]->{cifs_drive},
                                     client  => $set_h->{cifs}[$i]->{cifs_host});
                    logWarn($@) if $@;
                }
            }

            eval {
                if ( ! $mountpoint ) {
                    logStep("Mount volume on client");
                    $mountpoint = mount($set_h->{hostname},
                                        $set_h->{volume},
                                        mntopts=>$GH{mntopts},
                                        client=>$set_h->{cc}[0],
                                        mountretry=>2) or croak $@;
                } else {
                    logStep("Volume already mounted on client");
                }

                if ( $opts{cleanup} ) {
                    # Run specific cleanup for test
                    # Pass the mountpoint to the specific cleanup function
                    $set_h->{setup_mtpoint} = $mountpoint;

                    # Eval the specific cleanup function just in
                    # case the function does not return eval, so
                    # the error can be properly log as a warning.
                    eval {
                        &{$opts{cleanup}}( $set_h );
                        # Just in case the function does returns eval
                        croak $@ if $@;
                    };
                    logWarn("clean_env: $@") if $@;
                }

                if ( $set_h->{tfname} ) {
                    logStep("Delete all files created by this test");
                    eval { client_del_file($set_h->{cc}[0],
                                           "$mountpoint/$set_h->{tfname}*",
                                           recursive=>1) };
                }

                logStep("Unmount volume on client");
                umount($mountpoint, client=>$set_h->{cc}[0]);
                logWarn($@) if $@;

                # Do not try to unmount in common_cleanup
                undef $mountpoint;
            };
            logWarn($@) if $@;

        }
    };
    logWarn($@) if $@;

    eval {
        nfs_setup_gx::clean_env_gx();
    };
    logWarn($@) if $@;

    logStep("END: CLEANUP");

    return;
}

#---------------------------------------------------------------------
## @name ::check_oplock
## @description
##  Checks if an oplock exists for a given file and client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_oplock( %options );
##
## @arg ipaddr
##  Required: IP address of owner of oplock.
##
## @arg type
##  Required: Check for specific type of oplock ('excl', 'lvl2', 'none', 'byte').
##  Checking for 'byte' verifies that the given client has a byte range
##  lock on the file.
##
## @arg buffer
##  Optional: Do not get oplocks from the filer, use the buffer instead.
##  This is useful to check for a series of oplocks without having
##  to fetch all oplocks for each check, (default is 0).
##
## @arg partner
##  Optional: Check for delegation on partner in a cluster pair.
##
## @return
##  1 if oplock exists
#---------------------------------------------------------------------
sub check_oplock(%;) {
    my (%opts) = validate (
        @_,
        {
			# fc is only for 7-mode compatibility, ignored here.
			fc => { type => OBJECT, optional => 1 },
			inode => { type => SCALAR, optional => 1},
            ipaddr  => { type => SCALAR,  regex => qr/[a-f\:\.0-9]/ },
            type    => { type => SCALAR,  regex => qr/^(excl|lvl2|none|byte)$/ },
			volume  => { type => SCALAR, optional => 1, default => $::JUNCTION_PATH },
        }
    );
    # Hash to map oplock names on filer to input types
    my %oplock = ( 'Oplock-Excl'=>'excl', Excl=>'excl', 'Oplock-Lvl2'=>'lvl2', Lvl2=>'lvl2', None=>'none' );

    ## Check if filer is c-mode or 7 mode

	my ($lock_file, $file_n);

	my $server_obj = get_server_obj();

	$file_n = "NFSv4Delegations.$$";
	$lock_file = $opts{volume} .  '/'.$file_n;

	my @locks;
	@locks = $server_obj->vserver_locks_show( vserver => $::VSERVER_GX_NAME,
												volume => $::FILER_GX_VOL,
												path => $lock_file,
												type => 'op-lock',
												'client-address' => $opts{ipaddr},
											);

	foreach my $loks (@locks) {
		#if (($loks->{type} eq 'op-lock') && ($loks->{'client-address'} eq $opts{ipaddr}) &&
		if  ($loks->{'oplock-level'} ne 'null') {
			logStep ("There is a granted op-lock ");
			return 1;
		}
	}
	logStep ("There is no granted op-lock");
	croak "Could not find $opts{type} oplock for " .
          	"$lock_file on client $opts{ipaddr}";

}

#---------------------------------------------------------------------
## @name ::check_delegation
## @description
##  Checks if delegation exists for a given file and client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_delegation( %options );
##
## @arg ipaddr
##  Required: IP address of owner of delegation.
##
## @arg type
##  Required: Check for specific type of delegation ('read' or 'write').
##
## @arg buffer
##  Optional: Do not get delegations from the filer, use the buffer instead.
##  This is useful to check for a series of delegations without having
##  to fetch all delegations for each check, (default is 0).
##
## @arg partner
##  Optional: Check for delegation on partner in a cluster pair,
##  (default is 0).
##
## @return
##  1 if delegation exists
#---------------------------------------------------------------------
sub check_delegation(%;) {
	$Log->enter();
    my (%opts) = validate (
        @_,
        {
			# fc, buffer and partner are kept for backward compatibility for 7mode, but not used here. $#@@#!!!!
            fc      => { type => OBJECT|UNDEF, optional => 1, },
            inode   => { type => SCALAR,    regex => qr/^\d+$/ },
            ipaddr  => { type => SCALAR, regex => qr/^[0-9a-fA-F\:\.]/ },
            type    => { type => SCALAR, regex => qr/^(read|write)$/ },
            buffer  => { type => BOOLEAN, default => 0 },
            partner => { type => BOOLEAN, default => 0 },
            file    => { type => SCALAR, optional => 1 },

			# Extra cmode stuff
			vserver => { type => SCALAR, optional => 1, default => $::VSERVER_GX_NAME },
			protocol=> { type => SCALAR, default => 'nfsv4' . ($::NFS4_MINOR_VERSION ? ".$::NFS4_MINOR_VERSION" : ""), optional => 1, regex => qr/nfsv4|nfsv4.1/ },
			volume  => { type => SCALAR, optional => 1, default => $::JUNCTION_PATH },
			vol_name  => { type => SCALAR, optional => 1, default => $::FILER_GX_VOL },

        }
    );

	my ($lock_file, $file_n);

	if ( $opts{file} ) {
		$file_n = $opts{file};
	} else {
		$file_n = "NFSv4Delegations.$$";
	}
	$lock_file = $opts{volume} . '/'.$file_n;

	my @locks;
	# Get the Lock UUID from vserver locks show.
	# then use it for nfsv4 locks show.
	my @locks;
	my $node = get_node_obj();
	my $return_value = undef;
    my %opt;
    $opt{volume} = $opts{vol_name};
    try {
        NACL::C::Volume->find(command_interface=>$node,
            filter => { vserver => $opts{vserver}, volume => $opts{vol_name}, 'volume-style' => 'infinitevol' });
        delete $opt{volume};
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment(qq{$opts{vol_name} is not an infinite vol});
    };
	try {
		@locks = NACL::CS::VserverLocks->fetch(command_interface=>$node,
			filter => { vserver=> $opts{vserver},
						%opt,
						path => $lock_file,
						protocol => $opts{protocol},
						type => "delegation",
						'delegation-type' => $opts{type}});
	} catch NACL::Exceptions::NoElementsFound with {
		croak "Could not find $opts{type} delegation for file:" .
			"$opts{file} on client $opts{ipaddr}: $@";
	};

	if ($#locks != 0) {
		croak (qq{There are multiple locks for the same file, can't figure out which one to use, this is possibly an ONTAP burt});
	}
	my $uuid = $locks[0]->{'lockid'};
	if (!$uuid) {
		croak (qq{Couldnot get an uuid from the vserver locks show command. locks show dump:} . Dumper(@locks));
	}

    $Log->comment(qq{Number of nodes to try:$#ALL_NODE_OBJS});
    foreach my $current_node_obj (@ALL_NODE_OBJS) {
		try {
			@locks = NACL::CS::VserverLocksNfsv4->fetch(command_interface=>$current_node_obj,
					filter => { 'lock-uuid' => $uuid } );
			last;
		} catch NACL::Exceptions::NoElementsFound with {
            # Simply try next node.
		};
    }

	if ($#locks > 0) {
		croak (qq{There are multiple locks for the same lock uuid, can't figure out which one to use, this is a bug in ontap.});
	}

	foreach my $loks (@locks) {
		$loks->{'command_interface'} = undef;
		if (($loks->{lock_type} eq 'delegation') && ($loks->{'delegation_type'} eq $opts{type}) &&
			($loks->{'lease_state'} ne 'expired') && ($loks->{'lease_state'} ne 'reviving')) {
			# This is the only condition this function will return pass!
			logStep ("There is a granted $opts{type} delegation " . Dumper($loks));
			return 1;
		}
	}
	logStep ("There is no granted $opts{type} delegation. Looking for: $opts{type}, nfsv$GH{filer_nfs_version}. Got:" . Dumper(@locks));
	$Log->exit();
	croak "Could not find $opts{type} delegation for file:" .
		"$opts{file} on client $opts{ipaddr}: $@";
}

sub get_number_of_delegs(%;) {
	return get_number_of_locks(@_, lock_type => 'delegation');
}

sub get_number_of_share_locks(%;) {
	return get_number_of_locks(@_, lock_type => 'share-level');
}

sub get_number_of_locks(%;) {
    my (%opts) = validate (
        @_,
        {
            type    => { type => SCALAR, regex => qr/^(read|write)$/ },
			deny	=> { type => SCALAR, optional => 1, regex => qr/^(none|read|read-write|write|all)$/, default => 'none' },
			lock_type => { type => SCALAR, optional => 1, regex => qr/^(byte-range|delegation|share-level|op-lock)$/, default => 'delegation' },
            file    => { type => SCALAR,  optional => 1 },
			protocol=> { type => SCALAR, optional => 0, regex => qr/nfsv4|nfsv4.1/ },
			volume  => { type => SCALAR, optional => 1, default => $::JUNCTION_PATH },
			vol_name  => { type => SCALAR, optional => 1, default => $::FILER_GX_VOL },
        }
    );

	my ($lock_file, $file_n, %filter_args);

	if ( $opts{file} ) {
		$file_n = $opts{file};
	} else {
		$file_n = "NFSv4Delegations.$$";
	}
	$lock_file = $opts{volume} . '/' . $file_n;

	my $node = get_node_obj();

	if ($opts{lock_type} eq "delegation") {
		$filter_args{'delegation-type'} = $opts{type};
	} if ($opts{lock_type} eq "share-level") {
		
	}

	my @locks;
	try {
		@locks = NACL::CS::VserverLocks->fetch(command_interface=>$node,
			filter => { volume => $opts{vol_name}, path => $lock_file, protocol => $opts{protocol}, type => $opts{lock_type}, %filter_args });
	} catch NACL::Exceptions::NoElementsFound with {
		logWarn(qq{Couldnot find any matching lock for the given parameters: $opts{type}, $opts{lock_type}, $opts{file}, $opts{protocol}, $opts{volume}, and $opts{vol_name}});
	};

	return ($#locks+1);
}

sub get_node_obj() {
	if (! $GH{node_obj}) {
		$GH{node_obj} = NACL::C::Node->find(); 
        @ALL_NODE_OBJS = NACL::C::SystemNode->find(command_interface=>$GH{node_obj});
	}
    foreach (@ALL_NODE_OBJS) {
        $_->run(command=>'hostname');
    }
	return $GH{node_obj};
}

sub get_server_obj() {
	if (! $GH{server_obj}) {
		logStep(qq{Found it:get_server_obj: Find server});
		$GH{server_obj} = Server->new(node => $main::FILER);
	}
	return $GH{server_obj};
}

# A crude function to enable all the given versions of nfs in a filer.
sub enable_nfs_versions_in_filer(@) {
    my (@versions) = @_;
    my @required_versions = ();
    foreach (@versions) {
        if ($_ =~ /^3$/) {
            push @required_versions, "v3";
        } elsif ($_ =~ /^4$|^4.0$/){
            push @required_versions, "v4.0";
        } elsif ($_ =~ /4.1/) {
            push @required_versions, "v4.1";
        } elsif ($_ =~ /^2$/) {
            croak (qq{NFS V2 was discontinued in cmode, please remove this from the script that is calling this function});
        } else {
            croak (qq{Unrecognized version:$_});
        }
    }

    my $node = get_node_obj();
    my $vservernfs;
    try {
        $vservernfs = NACL::C::VserverNfs->find(command_interface => $node, filter => {vserver => $::VSERVER_GX_NAME});
    } catch NACL::Exceptions::NoElementsFound with {
        croak (qq{Unable to find the required vserver:$::VSERVER_GX_NAME});
    };
    foreach (@required_versions) {
        $vservernfs->modify($_ => "enabled");
    }
}

1;
