# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Distheap.pm#1 $
#
## @summary     Distheap - part of the Flg library
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Distheap - library functions specifically for use by the Flg library

=head1 SYNOPSIS

  use nfs::Distheap;

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

This package is intended for use by the nfs::Flg.pm library only. Do not
call nfs::Distheap.pm directly.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Distheap.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Distheap;

use strict;
use warnings;
use base qw(Exporter);

use Carp qw(croak);
use Carp::Assert;
use Params::Validate qw(validate SCALAR HASHREF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------
sub distheap_init($%;);
sub distheap_alloc($%;);
sub distheap_parent(%;);
sub distheap_select();
sub distheap_fromidx($%;);
sub distheap_dele($%;);
sub testme();

#--------------------------------------------------------------------------------

=head2 distheap_init

  $dh = nfs::Distheap->distheap_init( maxsize => $setsize,
                                      sizeof  => 'int' );

Initialize the distheap object

The following named arguments are accepted:

=over

=item Arguments

=over

=item maxsize

(Required) : Maximum number of records in the distheap

=item sizeof

(Required) : Size of an entry in the distheap

=back

=item Returns

The distheap object.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_init($%;) {
    #  distheap_t
    #  distheap_init(int maxsize, int elemdatasize)
    #  {
    #          distheap_entry_t he;
    #          distheap_t heap;
    #          int i;
    #
    #          assert(maxsize >= 0);
    #          elemdatasize = (elemdatasize + 3) & ~3; /*keep four byte aligned*/
    #
    #          if ((heap = malloc(sizeof(struct distheap))) == NULL) {
    #                  report_perror(FATAL, "malloc");
    #                  return NULL;
    #          }
    #          bzero(heap, sizeof(struct distheap));
    #          heap->entrysize = (sizeof(struct distheap_entry)-4) +elemdatasize;
    #          heap->maxsize = maxsize;
    #
    #          if ((heap->entries =
    #               malloc(heap->entrysize * (maxsize + 1))) == NULL) {
    #                  report_perror(FATAL, "malloc");
    #                  free(heap);
    #                  return NULL;
    #          }
    #          bzero(heap->entries, heap->entrysize * (maxsize + 1));
    #
    #          heap->freestart = 1;
    #          for (i=1 ; i<=maxsize ; i++) {
    #                  he = distheap_idx2entry(heap, i);
    #                  he->weight = he->treeweight = 0; /* redundant with bzero*/
    #                  he->u.freelink = i < maxsize ? i + 1 : 0;
    #          }
    #
    #          return heap;
    #  }
    #

    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            maxsize => { type => SCALAR, regex => qr/^\d+$/ },
            sizeof  => { type => SCALAR,
                    regex => qr/^(nameset_entry|contents|contents_range|int)$/ },
        }
    );
    my $self = bless { %opts }, $class;

    # $self->{maxsize} = $self->{maxsize};
    $self->{freestart} = 1;

    # for ( 1 .. $self->{maxsize} ) {
    for ( 1 ) {
        $self->{entries}{$_}{weight} = 0;
        $self->{entries}{$_}{treeweight} = 0;
        $self->{entries}{$_}{freelink} =
                ( $_ < $self->{maxsize} ) ? $_ + 1 : 0;

        # Create the distheap differently for nameset_entries and
        # for contents heaps. We check the sizeof parameter to determine
        # the type of initialization needed. They may look the same, but
        # different calls used for each just to make things clear, and
        # provides easy modification if future plans require different
        # structures for each type.
        if ( $self->{sizeof} eq 'nameset_entry' ) {
            # We're going to initialize the u.data component here, so it is
            # in effect in all nameset_entry distheaps.
            $self->{entries}{$_}{data} = { # Store the index for easy
                                           #  dereferencing
                                           my_idx         => $_,
                                           # We'll just store the constant here
                                           NAMESET_ENTRY_MAX_FHSIZE => 64 };
        } elsif ( $self->{sizeof} eq 'contents' ) {
            $self->{entries}{$_}{data} = { # Store the index for easy
                                           #  dereferencing
                                           my_idx         => $_, };
        } elsif ( $self->{sizeof} eq 'contents_range' ) {
            $self->{entries}{$_}{data} = { # Store the index for easy
                                           #  dereferencing
                                           my_idx         => $_, };
        } elsif ( $self->{sizeof} eq 'int' ) {
            $self->{entries}{$_}{data} = { value => 0,
                                           # Store the index for easy
                                           #  dereferencing
                                           my_idx => $_, };
        }

        $self->{entries}{$_}{valid} = 0;
    }

    return $self;
} # End distheap_init

#--------------------------------------------------------------------------------

=head2 distheap_alloc

  ( $data, $index ) = $dh->distheap_alloc( weight => $op->{weight} );

distheap_alloc function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item weight

(Required) : Initial weight for distheap allocation

=back

=item Returns

An array of the data and index from the heap entry.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_alloc($%;) {
    #  distheap_data_t
    #  distheap_alloc(distheap_t heap, int weight)
    #  {
    #          distheap_entry_t he, i;
    #
    #          assert(weight >= 0);
    #
    #          if (heap->freestart == 0) {
    #                  report_error(FATAL, "distheap_alloc: full");
    #                  return NULL;
    #          }
    #
    #          /*
    #           * grab a free entry and fill it in.
    #           */
    #          he = distheap_idx2entry(heap, heap->freestart);
    #          heap->freestart = he->u.freelink;
    #
    #          assert(he->weight == 0);
    #          he->weight = weight;
    #          /* he->treeweight includes lower entries, it may be non-zero! */
    #
    #          /*
    #           * propagate the weight gain up the tree, starting here.
    #           */
    #          for (i = he ; i ; i = distheap_parent(heap, i)) {
    #                  assert(i->treeweight + he->weight >= i->treeweight);/*ovfl*/
    #                  i->treeweight += he->weight;
    #          }
    #
    #          heap->size++;
    #
    #          assert(he->valid == 0);
    #          he->valid = 1;
    #
    #          return distheap_entry2data(heap, he);
    #  }
    #

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            weight => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    if ( $self->{freestart} == 0 ) {
        croak "distheap_alloc: full";
    }

    my $index = $self->{freestart};
    $self->{entries}{$index}{data}{my_idx} = $index;
    my $he = $self->{entries}{$index};

    if ( !defined $he->{freelink} ) {
        $he->{freelink} = $index + 1;
    }
    $self->{freestart} = $he->{freelink};

    $he->{weight} = 0 if ( !defined $he->{weight} );
    assert( $he->{weight} == 0 );
    $he->{weight} = $opts{weight};

    # propagate the weight gain up the tree
    for ( my $i = $index; $i; $i = distheap_parent( idx => $i ) ) {
        $self->{entries}{$i}{treeweight} = 0 if
                ( !defined $self->{entries}{$i}{treeweight} );
        if ( $self->{entries}{$i}{treeweight} + $he->{weight} <
                $self->{entries}{$i}{treeweight} ) {
            croak "distheap_alloc: treeweight overflow";
        }
        $self->{entries}{$i}{treeweight} += $he->{weight};
    }

    $self->{size}++;

    $he->{valid} = 0 if ( !defined $he->{valid} );
    assert( $he->{valid} == 0 );
    $he->{valid} = 1;

    # Here we would return a pointer to the u.data component of the
    #  distheap_entry struct. So return the data pointer if it exists,
    #  otherwise return undef.

    # We're also going to return the index of the specified distheap_entry, as
    #  that information might be more useful. If the caller simply wants to know
    #  if either u.freelink or u.data have value, then the $ref will determine
    #  that. If the caller wants to use u.data or wants to find the structure
    #  that contains the u.data pointer, then the index will help locate that
    #  information.

    my $ref = ( $he->{data} ) ? $he->{data} : undef;
    return ( $ref, $index );

} # End distheap_alloc

#--------------------------------------------------------------------------------

=head2 distheap_parent

  $i = distheap_parent( idx => $i );

distheap_parent function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item idx

(Required) : Index of current distheap entry

=back

=item Returns

Parent distheap entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_parent(%;) {
    #  static __inline distheap_entry_t
    #  distheap_parent(const distheap_t heap, const distheap_entry_t he)
    #  {
    #          register int idx = distheap_entry2idx(heap, he), ridx = idx / 2;
    #          return (ridx > 0 ? distheap_idx2entry(heap, ridx) : 0);
    #  }
    #
    my (%opts) = validate (
        @_,
        {
            idx => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    return int ($opts{idx} / 2);
} # End distheap_parent

#--------------------------------------------------------------------------------

=head2 distheap_select

  $selection = $rsize_distdh->distheap_select();

distheap_select function taken from the original C code.

The following named arguments are accepted:

=over

=item Returns

A selected entry from the current distheap.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_select() {
    #  distheap_data_t
    #  distheap_select(distheap_t heap)
    #  {
    #          distheap_entry_t he = distheap_idx2entry(heap, 1) /* root */, left;
    #          distheap_data_t r;
    #          int w, leftw;
    #
    #          if (he->treeweight == 0) {
    #  #if 0
    #                  /* XXO: This should be possible for runs w/o nameset file */
    #                  report_error(NONFATAL, "distheap_select: treeweight==0");
    #  #endif
    #                  return NULL; /* empty tree */
    #          }
    #          w = random() % he->treeweight;
    #
    #          while (1) {
    #                  /*
    #                   * if the weight falls in the left subtree, recurse left.
    #                   */
    #                  if ((left = distheap_left(heap, he)) != NULL) {
    #                          leftw = left->treeweight;
    #                  } else {
    #                          leftw = 0;
    #                  }
    #                  if (w < leftw) {
    #                          he = left;
    #                          continue;
    #                  }
    #
    #                  /*
    #                   * if remaining weight falls in the current node, use it.
    #                   */
    #                  w -= leftw;
    #                  if (w < he->weight) {
    #                          r = distheap_entry2data(heap, he);
    #                          assert(r != NULL);
    #                          return r;
    #                  }
    #
    #                  /*
    #                   * recurse right with remaining weight.
    #                   */
    #                  w -= he->weight;
    #                  he = distheap_right(heap, he);
    #          }
    #          return NULL; /* unreachable */
    #  }
    #

    my $self  = shift;
    my $index = 1;
    my $leftw;

    my $he = $self->{entries}{$index}; # root

    if ( $he->{treeweight} == 0 ) {
        return undef; # Empty tree
    }

    my $w = int (rand( $he->{treeweight} ));

    while ( 1 ) {
        # Set the initial weight if not defined
        $he->{weight} = 0 if ( !defined $he->{weight} );

        # Recurse left
        my $ridx = $index * 2;

        # Don't change the index yet
        my $l_index = ( $ridx <= $self->{maxsize} ) ? $ridx : 0;
        my $left    = $self->{entries}{$l_index};

        if ( $left ) {
            $leftw = $left->{treeweight};
        } else {
            $leftw = 0;
        }

        if ( $w < $leftw ) {
            $he    = $left;
            $index = $l_index; # Save the current index before proceeding
            next;  # Return to top of while loop
        }

        # Use current node
        $w -= $leftw;
        if ( $w < $he->{weight} ) {
            my $r = $he->{data};
            assert( defined $r );
            return $r;
        }

        # Recurse right
        $w    -= $he->{weight};
        $ridx  = ( $index * 2 ) + 1;
        $index = ( $ridx <= $self->{maxsize} ) ? $ridx : 0;
        $he    = $self->{entries}{$index};
    } # Continue while loop until it gets out by itself or croaks

} # End distheap_select

#--------------------------------------------------------------------------------

=head2 distheap_fromidx

  $cr = $dh_contents_ranges->distheap_fromidx( idx => $index )

distheap_fromidx function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item idx

(Required) : Index for which a distheap entry is desired.

=back

=item Returns

Distheap entry corresponding to the index

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_fromidx($%;) {
    #  distheap_data_t
    #  distheap_fromidx(distheap_t heap, int idx)
    #  {
    #          if (idx == 0)
    #                  return NULL;
    #          assert(idx > 0);
    #          if (idx > heap->maxsize) {
    #                  return NULL; /* out of range */
    #          }
    #          return distheap_entry2data(heap, distheap_idx2entry(heap, idx));
    #  }

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            idx => { type => SCALAR },
        }
    );
    my $idx = $opts{idx};

    return undef if ( $idx == 0 );
    assert( $idx > 0 );

    return undef if ( $idx > $self->{maxsize} ); # out of range

    return $self->{entries}{$idx}{data};

} # End distheap_fromidx

#--------------------------------------------------------------------------------

=head2 distheap_dele

  $dh_contents_ranges->distheap_dele( hd => $cr );

distheap_dele function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hd

(Required) : Distheap entry data to be deleted

=back

=item Returns

0.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub distheap_dele($%;) {
    #  int
    #  distheap_dele(distheap_t heap, distheap_data_t hd)
    #  {
    #          distheap_entry_t he = distheap_data2entry(heap, hd), i;
    #
    #          assert(he && he->weight >= 0);
    #          assert(he->valid == 1);
    #
    #          /*
    #           * propagate the weight loss up the tree, starting here.
    #           */
    #          for (i = he ; i ; i = distheap_parent(heap, i)) {
    #                  i->treeweight -= he->weight;
    #                  assert(i->treeweight >= 0);
    #          }
    #
    #          /*
    #           * clear this record and add it to the freelist.
    #           */
    #          he->weight = 0;
    #
    #          he->valid = 0;
    #          he->u.freelink = heap->freestart;
    #          heap->freestart = distheap_entry2idx(heap, he);
    #
    #          heap->size--;
    #
    #          return 0;
    #  }

    my $self   = shift;
    my (%opts) = validate (
        @_,
        {
            hd => { type => HASHREF },
        }
    );

    my $index = $opts{hd}->{my_idx};
    my $he    = $self->{entries}{$index};

    assert( defined $he && $he->{treeweight} >= 0 );
    assert( $he->{valid} == 1 );

    for ( my $i = $index; $i; $i = distheap_parent( idx => $i ) ) {
        $self->{entries}{$i}{treeweight} -= $he->{weight};
        assert( $self->{entries}{$i}{treeweight} >= 0 );
    }

    $he->{weight} = 0;
    $he->{valid}  = 0;

    # Remove the data component and establish the freelink component
    delete $he->{data};
    $he->{freelink} = $self->{freestart};
    $self->{freestart} = $index;

    $self->{size}--;

    return 0;

} # End distheap_dele

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Distheap SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    my $maxsize = 500;
    my $sizeof = 'nameset_entry';

    # Call distheap_init
    my $dh = nfs::Distheap->distheap_init( maxsize => $maxsize,
                                           sizeof  => $sizeof );

    # Validate structure initialized correctly
    if ( $dh->{sizeof} ne $sizeof ||
         $dh->{maxsize} != $maxsize ||
         $dh->{entries}{1}{treeweight} != 0 ) {
        Tharn::logresult('FAIL', 'distheap_init did not initialize correctly');
    } else {
        Tharn::logresult('PASS', 'SELF TEST of nfs/Distheap.pm');
    }
}

1;
