# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Export/Aix.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::Export::Aix;

## @summary
##  Class to handle OS dependent exports on AIX.
##
## @description
##  This class handles the OS specific portion of the NFS export
##  functions. It overrides the default behavior of the base
##  export class.
##
## @usage
##  use nfs::Export::Aix;
##
## @status  public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.

use strict;
use Carp qw(croak);
use nfs::Export::ExportBase;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF);
use vars qw( @ISA );

#---------------------------------------------------------------------
# Inherit all functionality from ExportBase class.
#---------------------------------------------------------------------
@ISA = qw(nfs::Export::ExportBase);

#---------------------------------------------------------------------
## @name ::new
##
## @description
##  Aix specific class constructor. It overrides default
##  behavior of the base exports class.
##
## @return
##  Aix specific export object.
#---------------------------------------------------------------------
sub new($$;) {
    my ($class, $conn) = @_;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class ( $conn,
        # For re-exporting all in the exports file,
        # need to unexport all first
        unexport_r_cmd => '/usr/sbin/exportfs -au',
        export_r_flags => '-a',
    );
}

#---------------------------------------------------------------------
## @name ::export_path
##
## @description
##  Overrides the default export_path function from the base exports
##  class.
##
##  In Aix, the options 'ro' and 'rw' are mutually exclusive
##  with or without list.  If 'ro=list' is specified, then all clients
##  in the list are granted read-only and all others are granted read-
##  write (ro=list,rw).  In order to grant read-only to the clients
##  in the list and denied all others access then the options must be
##  ro=list,access=list.
##
##  This function first changes the 'ro' and 'rw' options specified
##  to the proper Aix options.  Then, it calls the function inherited
##  from the base class to perform the actual export.
##
## @synopsis
##  $op = $obj->export_path($path, $access, $persist);
##
## @arg $path
##  Required: Path to be exported.
##
## @arg $access
##  Required: Comma separated access options.
##
## @arg $persist
##  Required: True for a persistent export.
##  False for a non-persistent export.
##
## @notes
##  Option "rw=list" is mapped to "rw=list,access=list"
##
##  Option "ro=list" is mapped to "ro=list,access=list"
##
##  Option "ro,rw=list" is mapped to "rw=list"
##
##  Option "rw,ro=list" is mapped to "ro=list"
##
##  Option "ro=rlist,rw=wlist" is mapped to "access=rlist,rw=wlist"
##
## @return
##  1 if succeeds.
##
## @note
##  This function should not be called directly by the user, so
##  no argument validation is performed.
#---------------------------------------------------------------------
sub export_path($$$$;) {
    my ($this, $path, $access, $persistent) = @_;
    my ($list, $op);

    if ( $access ) {
        # Validate host entries, aix does not accept
        # negated hosts.
        my ($hosts) = ( $access =~ /\br[ow]=([^,])/ );
        my ($op)    = ( $access =~ /\broot=([^,])/ );
        $hosts = ":$hosts:$op";
        if ( $hosts =~ /:-/ ) {
            croak "Aix nfs does not support negated hosts";
        }
    }

    if ( $access =~ /\br[ow](?!=)\b/ and $access =~ /\br[ow]=/ ) {
        # If any of 'ro' or 'rw' exist, remove it
        #   Option "ro,rw=list" is mapped to "rw=list"
        #   Option "rw,ro=list" is mapped to "ro=list"
        $access =~ s/r[ow](?!=),?\b//;
        $access =~ s/,+$//;	# Remove any trailing ','s left behind
    } elsif ( $access =~ /\bro=/ and $access =~ /\brw=/ ) {
        # Both ro=list and rw=list exist,
        # change 'ro=' to 'access='
        #   Option "ro=rlist,rw=wlist" is mapped
        #       to "access=rlist,rw=wlist"
        $access =~ s/ro=/access=/;
    } elsif ( $access =~ /\br[ow]=/ ) {
        # Either ro=list or rw=list exist,
        # add 'access='
        #   Option "rw=list" is mapped to "rw=list,access=list"
        #   Option "ro=list" is mapped to "ro=list,access=list"
        ($list) = ($access =~ /\br[ow]=([^,]+)\b/);
        $access .= ",access=$list";
    }

    # Call inherited export_path
    $op = $this->SUPER::export_path($path, $access, $persistent);

    return $op;
}


#---------------------------------------------------------------------
## @name ::export_paths
##
## @description
##  This function is calling export_path in the loop
## 
## @author gvg@netapp.com
##
#---------------------------------------------------------------------
sub export_paths() {
    my ($this, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
	    paths          => { type => HASHREF },
            OPTIONS        => { type => SCALAR },
	    persistent     => { type => BOOLEAN },
            retry          => { type => SCALAR },
	    export_all     => { type => BOOLEAN, default => 0, },
        },
    );

    # $export_all is not implemented on this plateform for now. It
    # just loops through paths one by one

    my $paths_hashref = $opts{paths};

    foreach my $path (keys %$paths_hashref) {
	my $access = $paths_hashref->{$path}->{OPTIONS} ? $paths_hashref->{$path}->{OPTIONS} : $opts{OPTIONS};
	export_path($this, $path, $access, $opts{persistent});
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::list_exports
##
## @description
##  Overrides the default list_exports function from the base
##  export class.
##
##  It calls the function inherited from the base class to get the
##  export options off the client.  Then, reformats the option back
##  to the standard format.
##
## @synopsis
##  $list_h = $obj->list_exports();
##
## @notes
##  Option "rw=list,access=list" is mapped back to "rw=list"
##
##  Option "ro=list,access=list" is mapped back to "ro=list"
##
##  Option "rw=list" is mapped back to "ro,rw=list"
##
##  Option "ro=list" is mapped back to "rw,ro=list"
##
##  Option "access=rlist,rw=wlist" is mapped back to "ro=rlist,rw=wlist" 
##
##  It only gets the clients that are not in the r[ow]=list to
##  populate the r[wo]=list, so that no client will be in both lists.
##
## @return
##  A reference to a hash, where the keys are the exported paths
##  and the values are the access options.
##
## @note
##  This function should not be called directly by the user, so
##  no argument validation is performed.
#---------------------------------------------------------------------
sub list_exports($;) {
    my $this = shift;
    my ($access, $list_h, $vol, $alist, $rlist, @acc, @row);

    # Get list of exports from inherited function
    $list_h = $this->SUPER::list_exports();

    # Modify access options for all exported paths
    foreach $vol (keys %$list_h) {
        # Get access for this volume
        $access = $list_h->{$vol};

        if ( $access =~ /\baccess=/ and $access =~ /\br[ow]=/ ) {
            #  Option "rw=list,access=list" is mapped to "rw=list"
            #  Option "ro=list,access=list" is mapped to "ro=list"
            #  Option "access=rlist,rw=wlist" is mapped
            #      to "ro=rlist,rw=wlist" 
            ($rlist) = ($access =~ /\br[ow]=([^,]+)\b/);  # Get list of [rw|ro]
            ($alist) = ($access =~ /\baccess=([^,]+)\b/); # Get list of access
            @row = split /:/, $rlist;
            foreach my $host ( split /:/, $alist ) {
                next if grep { /^$host$/ } @row;
                # Only add to list, clients which
                # are not in the [rw|ro] list
                push @acc, $host;
            }
            if ( @acc ) {
                # The [rw|ro] list and the access list are different
                # so need to add its counterpart [ro|rw]
                $alist = ($access =~ /\bro=/ ? 'rw=' : 'ro=') . join(':', @acc);
            } else {
                # Both the [rw|ro] list and the access list are the
                # same so no need to add its counterpart [ro|rw]
                $alist = "";
            }
            # Replace 'access=list' by [ro|rw]=newlist
            $access =~ s/\baccess=[^,]+\b/$alist/;
            $access =~ s/,+/,/g;	# Get rid off duplicate ','s
            $access =~ s/^,//;		# Get rid off leading ','
            $access =~ s/,$//;		# Get rid off trailing ','
        } elsif ( $access =~ /\br[ow]=/ ) {
            #  Option "ro=list" is mapped to "rw,ro=list"
            $access =~ s/\b(ro=)/rw,$1/;
            #  Option "rw=list" is mapped to "ro,rw=list"
            $access =~ s/\b(rw=)/ro,$1/;
        }
        # Replace access options
        $list_h->{$vol} = $access;
    }

    return $list_h;
}

1;

