# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Export/ExportBase.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::Export::ExportBase;

## @summary
##  Base class to the OS dependent exports classes
##
## @description
##  Provides a default behavior for the NFS exports command.
##  Generic functions for exporting, unexporting and listing
##  of exports are included in this class.  This should be 
##  the parent class for an OS-dependent class.
##
## @usage
##  use nfs::Export::ExportBase;
##
## @status  public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.

use strict;
use Carp qw(croak);
use SayWrap qw(sayEOF sayPerl);
use nfs::Subnets qw(fix_subnets);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF);

#---------------------------------------------------------------------
# Define Constants
#---------------------------------------------------------------------
use constant NO_CLOSE => 0;
use constant CLOSE    => 1;
use constant P_START  => 0;
use constant P_END    => 1;

#---------------------------------------------------------------------
## @name ::new
##
## @description
##  Exports base class constructor which initializes default behavior.
##  Class behavior is controlled by the blessed hash of the constructor.
##  This list, which can be overrode by the constructor of a class
##  inheriting this base class, includes the export commands, flags
##  and format of the exports file.
##
## @synopsis
##  $export_obj = nfs::Export::ExportBase::->new($conn, %override_defaults);
##
## @arg $conn
##  Required: The Connectrec object for the server.
##
## @arg export_cmd
##  Optional: Full path of export command.
##
## @arg unexport_cmd
##  Optional: Full path of unexport command.
##
## @arg export_r_cmd
##  Optional: Full path of export all command.
##
## @arg unexport_r_cmd
##  Optional: Full path of unexport all command.
##
## @arg sudo
##  Optional: Full path of sudo command.
##
## @arg export_flags
##  Optional: Flags needed by the export command.
##
## @arg unexport_flags
##  Optional: Flags needed by the unexport command.
##
## @arg export_r_flags
##  Optional: Flags needed by the export all command.
##
## @arg export_file
##  Optional: Location of the exports file.
##
## @arg export_list_s
##  Optional: First character expected after getting the list
##  of exports off a system.  Used for error checking.
##
## @arg export_format
##  Optional: Format of the exports file.
##
## @arg path_origin
##  Optional: Where the exported path is located in each line
##  in the exports file.
##  This should reflect what the export_format has.
##  Use P_START if export path is located at the
##  beginning of the line. Use P_END if the export
##  path is located at the end of the line.
##
## @arg netgrp_atsign
##  Optional: True if OS supports the '@' on a netgroup.
##
## @return
##  Exports base object.
##
## @example
##  $obj = nfs::Export::ExportBase::->new($conn, export_cmd=>'/bin/exportfs');
#---------------------------------------------------------------------
sub new($$;%) {
    my $class = shift;
    my $conn  = shift;

    bless {
        # Connect rec object for server
        conn           => $conn,
        # Defaults
        export_cmd     => '/usr/sbin/exportfs',
        unexport_cmd   => '/usr/sbin/exportfs',
        export_r_cmd   => '/usr/sbin/exportfs',
        unexport_r_cmd => '',
        sudo           => '/usr/software/bin/sudo',
        export_flags   => '-i -o',
        unexport_flags => '-u',
        export_r_flags => '-r',
        export_file    => '/etc/exports',
        export_list_s  => '/',
        export_format  => '"$path -$access"',
        path_origin    => P_START,
        netgrp_atsign  => 0,
 
        # Override defaults
        @_
    }, $class;
}

#---------------------------------------------------------------------
## @name ::del_persistent_cmd
##
## @description
##  Generates the perl code to remove a persistent export in the
##  machine where the exports file is to be modified.
##  Reads the entire contents of the exports file and re-writes
##  it without the export specified by the path argument.
##  It correctly handles multi-line exports, i.e. those
##  exports broken down into lines with a '\' at the end
##  of each line.
##  The exported path to be removed must be located at the
##  beginning (or end) of the export line according to the
##  the contents of 'path_origin'.
##
## @synopsis
##  $cmd = $obj->del_persistent_cmd( $path, CLOSE|NO_CLOSE );
##
## @arg $path
##  Required: Export path to be removed from the exports file.
##
## @arg $close
##  Required: CLOSE: Closes the exports file.
##  NO_CLOSE: Leaves the exports file open if further
##  processing on the file is required.
##
## @return
##  The perl code to remove the exported path off the exports file.
#---------------------------------------------------------------------
sub del_persistent_cmd($$$;) {
    my ($this, $path, $close) = @_;
    my $export_file = $this->{export_file};

    # Convert close to 0 if NO_CLOSE
    # Just in case NO_CLOSE is not defined to 0
    $close = ( $close == $this->CLOSE ) ? 1 : 0;

    # Open the exports file, read all the contents at once,
    # and remove all lines which contain '$path'.
    # If the exports file does not exists, it gets created only
    # if it is to be left open, i.e., NO_CLOSE option is given.
    my $cmd = qq{
        \$exp_file = "$export_file";
        \$op = open(IN, "<\$exp_file");
        if ( not \$op and $close ) \{
            print \$err ? \$err : q\{--persist_ok--\};
            exit;
        \}
        \@cont = <IN> if \$op;
        close(IN);
        open(OUT, ">\$exp_file") or die "\$exp_file: \$OS_ERROR";
        foreach \$line (\@cont) \{
            \$line =~ s/\\s\*\$//;
            if (\$line =~ /\\\\\$/) \{
                push \@t, \$line;
                next;
            \} elsif (\@t) \{
                push \@t, \$line;
                \$line = join("", \@t);
                \$line =~ s/\\\\//g;
            \}
            if (\$line =~ m%^$path\\s%) \{
                undef \@t;
                next;
            \}
            if (\@t) \{
                print OUT "\$_\\n" foreach \@t;
            \} else \{
                print OUT "\$line\\n";
            \}
            undef \@t;
        \}
    };

    if ( $this->{path_origin} == $this->P_END ) {
        # Path is found at the end of the line:
        # modify match criteria
        $cmd =~ s{m%\^($path)\\s%}{m%\^[^#].*\\s$1\$%};
    }

    # Close exports file
    if ( $close ) {
        $cmd .= qq{ close(OUT); };
    }

    return $cmd;
}

#---------------------------------------------------------------------
## @name ::export_path
##
## @description
##  Exports the specified path on the machine provided by the
##  class object. Validates all subnet specifications and
##  converts them into a standard format.
##
## @synopsis
##  $op = $obj->export_path($path, $access, $persist);
##
## @arg $path
##  Required: Path to be exported.
##
## @arg $access
##  Required: Comma separated access options.
##
## @arg $persist
##  Required: True for a persistent export.
##  False for a non-persistent export.
##
## @return
##  1 if succeeds.
##
## @example
##  $op = $obj->export_path('/vol/testvol', 'rw=client1,ro', 0);
#---------------------------------------------------------------------
sub export_path($$$$;) {
    my ($this, $path, $access, $persistent) = @_;
    my ($cmd, $op);
    my $export_cmd  = $this->{export_cmd};
    my $export_file = $this->{export_file};
    my $flags       = $this->{export_flags};
    my $sudo        = $this->{sudo};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    if ( $access ) {
        # Prefix '@' on all subnets
        # Need to escape the '@' so perl will
        # not interpret it when $access is used later
        $access = fix_subnets($access, prefix=>'\@');

        # Move 'sec=sec_flavor' to the beginning of
        # list. Some servers need this option before
        # 'ro' or 'rw'.
        $access =~ s/(.*)(sec=[^,]+,?)(.*)/$2,$1$3/;
        $access =~ s/,+/,/;	# Remove duplicate ','
        $access =~ s/,$//;	# Remove trailing ','
    } else {
        # No access options, get rid off '-o' flag
        $flags =~ s/\s+-o\b//;
    }

    # Export command
    $cmd = qq{
        \$err = qx($export_cmd $flags $access $path);
        \$err .= \$CHILD_ERROR if \$CHILD_ERROR;
    };

    if ( $persistent ) {
        # Delete export from exports file; don't close file
        $cmd .= $this->del_persistent_cmd($path, $this->NO_CLOSE);
        # Add new export to exports file
        $op = eval $this->{export_format};
        $cmd .= qq{
            print OUT "$op\\n";
            close(OUT);
        };
    }

    # Display error or ok message
    $cmd .= qq{
        print \$err ? \$err : q\{--export_ok--\};
    };

    # Do the actual export on the server
    (undef, undef, $op) = sayPerl($this->{conn}, $cmd, $sudo, @say_opts);
    croak $op if ( $op !~ /(--export_ok--|--persist_ok--)/ );

    return 1;
}

#---------------------------------------------------------------------
## @name ::export_paths
##
## @description
##  This function is calling export_path in the loop
## 
## @author gvg@netapp.com
##
#---------------------------------------------------------------------
sub export_paths() {
    my ($this, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
	    paths          => { type => HASHREF },
            OPTIONS        => { type => SCALAR },
	    persistent     => { type => BOOLEAN },
            retry          => { type => SCALAR },
	    export_all     => { type => BOOLEAN, default => 0, },
        },
    );

    # $export_all is not implemented by default. It
    # just loops through paths one by one

    my $paths_hashref = $opts{paths};

    foreach my $path (keys %$paths_hashref) {
	my $access = $paths_hashref->{$path}->{OPTIONS} ? $paths_hashref->{$path}->{OPTIONS} : $opts{OPTIONS};
	export_path($this, $path, $access, $opts{persistent});
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::unexport_path
##
## @description
##  Unexports the specified path on the machine provided by the
##  class object.
##
## @synopsis
##  $op = $obj->unexport_path($path, $persist, $ponly);
##
## @arg $path
##  Required: Path to be unexported.
##
## @arg $persist
##  Required: True for a persistent unexport.
##  False for a non-persistent unexport.
##
## @arg $ponly
##  Required: Remove persistent export only.
##
## @return
##  1 if succeeds.
##
## @example
##  $op = $obj->unexport_path('/vol/testvol', 0, 0);
#---------------------------------------------------------------------
sub unexport_path($$$$;) {
    my ($this, $path, $persistent, $persist_only) = @_;
    my ($cmd, $op);
    my $unexport_cmd   = $this->{unexport_cmd};
    my $unexport_flags = $this->{unexport_flags};
    my $export_file    = $this->{unexport_file};
    my $sudo           = $this->{sudo};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Unexport command
    if ( ! $persist_only ) {
        $cmd = qq{
            \$err = qx($unexport_cmd $unexport_flags $path);
            \$err .= \$CHILD_ERROR if \$CHILD_ERROR;
        };
    }

    if ( $persistent ) {
        # Delete export from exports file; close file
        $cmd .= $this->del_persistent_cmd($path, $this->CLOSE);
    }

    # Display error or ok message
    $cmd .= qq{
        print \$err ? \$err : q\{--unexport_ok--\};
    };

    # Do the actual unexport
    (undef, undef, $op) = sayPerl($this->{conn}, $cmd, $sudo, @say_opts);
    croak $op if ( $op !~ /(--unexport_ok--|--persist_ok--)/ );

    return 1;
}

#---------------------------------------------------------------------
## @name ::list_exports
##
## @description
##  List all exports on the machine provided by the class object.
##
## @synopsis
##  $list_h = $obj->list_exports();
##
## @return
##  A reference to a hash, where the keys are the exported paths
##  and the values are the access options.
#---------------------------------------------------------------------
sub list_exports($;) {
    my $this = shift;
    my ($cmd, $vol, $access, %access_h, $op);
    my $export_cmd    = $this->{export_cmd};
    my $export_list_s = $this->{export_list_s};
    my $sudo          = $this->{sudo};

    # Get list of exports from server
    $cmd = "$sudo $export_cmd";
    $op = sayEOF($this->{conn}, $cmd);

    # Server machine is really verbose,
    # do not croak on that account.
    $op =~ s/^.*nothing exported//;

    # Croak if it is not what was expected
    croak $op if ($op and $op !~ /^$export_list_s/);

    # Create return hash
    foreach my $line (split /\n/, $op) {
        $line =~ s/^-\s+//;
        ($vol, $access) = split /\s+/, $line;
        # Get rid off leading '-'
        $access =~ s/^-//;
        # Get rid off '@' in subnets
        $access =~ s/([=:]-?)\@(\d*\.\d*\.\d*\.\d*\/\d*)/$1$2/g;
        $access_h{$vol} = $access;
    }

    return \%access_h;
}

#---------------------------------------------------------------------
## @name ::export_reset_persistent
##
## @description
##  Resets all persistent exports by re-exporting all persistent
##  exports only, all others are unexported.
##
## @synopsis
##  $op = $obj->export_reset_persistent();
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub export_reset_persistent($;) {
    my $this = shift;
    my ($cmd, $op);
    my $export_r_cmd   = $this->{export_r_cmd};
    my $unexport_r_cmd = $this->{unexport_r_cmd};
    my $export_r_flags = $this->{export_r_flags};
    my $sudo           = $this->{sudo};

    if ( $unexport_r_cmd ) {
        # Need to unexport all exports first.
        # Need to run a sub-shell in order to add proper path
        $cmd = "$sudo sh -c 'PATH=/usr/bin:/usr/sbin; export PATH; " .
               "$unexport_r_cmd; $export_r_cmd $export_r_flags'";
    } else {
        # Re-export all persistent exports
        $cmd = "$sudo $export_r_cmd $export_r_flags";
    }

    # Do the actual command
    $op = sayEOF($this->{conn}, $cmd);
    croak $op if $op;

    return 1;
}

1;

