# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Export/Filer.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::Export::Filer;

## @summary
##  Class to handle OS dependent exports on Filer.
##
## @description
##  This class handles the OS specific portion of the NFS export
##  functions. It overrides the default behavior of the base
##  export class.
##
## @usage
##  use nfs::Export::Filer;
##
## @status  public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.

use strict;
use Carp qw(croak);
use SayWrap qw(sayEOF);
use nfs::Export::ExportBase;
use CLI::FileIO qw(rdfile wrfile);
use Filer::Version qw(get_wafl_version WV_ANCHORSTEAM_MIN);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF);
use Data::Dumper;

#---------------------------------------------------------------------
# Inherit all functionality from ExportBase class.
#---------------------------------------------------------------------
our @ISA = qw(nfs::Export::ExportBase);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($$;);
sub unexport_path($$$$%;);
sub list_exports($%;);
sub export_reset_persistent($%;);

#---------------------------------------------------------------------
# Private variables
#---------------------------------------------------------------------
my %wafl_vers;
my $export_cmd  = 'exportfs';
my $export_file = '/etc/exports';
my $re_retry    = qr/exportfs already in progress/;

#---------------------------------------------------------------------
## @name ::new
##
## @description
##  Filer specific class constructor. It overrides default
##  behavior of the base exports class.
##
## @return
##  Filer specific export object.
#---------------------------------------------------------------------
sub new($$;) {
    my ($class, $conn) = @_;

    # Get version of the filer and save it so
    # subsequent calls for the same filer the
    # version is just retrieved from memory.
    my $conn_ip = $conn->connhost();
    $wafl_vers{$conn_ip} ||= get_wafl_version($conn->hostp());

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class (
        $conn,
        # No full path specified on filer
        export_cmd   => 'exportfs',
        export_r_cmd => 'exportfs',
        # No sudo command on filer, and besides all filer
        # connections are done as root.
        sudo => '',
        # OS supports the '@' on netgroups since anchorsteam
        netgrp_atsign => ($wafl_vers{$conn_ip} < WV_ANCHORSTEAM_MIN) ? 0 : 1
    );
}

#---------------------------------------------------------------------
# _del_exports_from_file
#
# Removes an export(s) line from the exports file.
# Reads the entire contents of the exports file and removes any
# lines associated with the exports specified by the paths argument.
# It correctly handles multi-line exports, i.e. those
# exports broken down into lines with a '\' at the end
# of each line.
#
# It takes two arguments: the connection object of the server
# and the export paths to be removed from the exports file.
#
# It returns the contents of the exports file minus the export
# specified by the paths argument.
#
# This is an internal function which should only be used within
# this module.
#---------------------------------------------------------------------
sub _del_exports_from_file {
    my ($conn, @paths) = @_;
    my (@cont, @tmp, $op);

    # Read exports file and break it into lines.
    # Don't croak if file does not exist.
    @cont = split /\r?\n/, rdfile($conn, $export_file);

    foreach my $line (@cont) {
        $line =~ s/\s*$//;      # Get rid off trailing spaces
	$line =~ s/^\s*//;      # Get rid off leading spaces
        if ( $line =~ /\\$/ ) {
            # Found '\' at the end of line, multi-line continues.
            push @tmp, $line;
            next;
        } elsif (@tmp) {
            # This is the last line of a multi-line exports.
            push @tmp, $line;
            $line = join("", @tmp);     # Merge all lines in multi-line
            $line =~ s/\\//g;           # Get rid off all '\' characters.
        }

	foreach my $path (@paths) {
	    if ( $line =~ /^$path\s/) {
		# Deleting export path
		undef @tmp;                 # Clear multi-line buffer
		undef $line;                # Clear single-line buffer
		next;
	    }
	}

        if ( @tmp ) {
            # Multi-line exports which must be kept
            $op .= "$_\n" foreach @tmp;
        } elsif ($line) {
            # Single-line exports which must be kept
            $op .= "$line\n";
        }
    }
    return $op;
}

#---------------------------------------------------------------------
## @name ::export_paths
##
## @description
##  Exports the specified paths on the filer provided by the
##  class object. Validates all subnet specifications and
##  converts them into a standard format.
##
## @synopsis
##  $op = $obj->export_paths( paths => $paths_hashref, 
##                            OPTIONS => $access, 
##                            persistent => $persist, 
##                            retry => $retry
##                            );
##
## @arg $paths
##  Required: Path to be exported.
##
## @arg $access
##  Required: Comma separated access options.
##
## @arg $persistent
##  Required: 0 for a non-persistent export.
##            1 for a persistent export modifying the exports file.
##            2 for a persistent export using '-p' option of the filer.
##
## @arg retry
##  Required: Number of times to retry the command if the exportfs
##  command is busy.
##
## @return
##  1 if succeeds.
##
## @example
##  $op = $obj->export_paths( paths => { '/vol/testvol' => { }, },
##                            OPTIONS => 'rw=client1,ro', 
##                            cmd_execute => 1,
##                            persistent => 0,);
#---------------------------------------------------------------------
sub export_paths() {
    my ($this, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
	    paths          => { type => HASHREF },
            OPTIONS        => { type => SCALAR },
	    persistent     => { type => BOOLEAN },
            retry          => { type => SCALAR },
	    export_all     => { type => BOOLEAN, default => 0, },
        },
    );

    my $paths_hashref = $opts{paths};
    my @paths = keys %$paths_hashref;

    my ($cmd, $cont, $op);
    my $conn    = $this->{conn};
    my $flags   = $this->{export_flags};
    my $retries = $opts{retry};
    my $stime   = 1;

    if(!$opts{export_all}) {
	foreach my $path (keys %$paths_hashref) {
	    my $access = $paths_hashref->{$path}->{OPTIONS} ? $paths_hashref->{$path}->{OPTIONS} : $opts{OPTIONS};

	    # No access options, get rid off '-o' flag
	    $flags =~ s/\s+-o\b// if !$access;
	
	    # Filer specific persistent option
	    $flags = '-p' if $opts{persistent} == 2;
	
	    # Export command
	    $cmd = "$export_cmd $flags $access $path";
	    # Redo block
	    {
		$op = sayEOF($conn, $cmd);
		if ( $retries-- and $op =~ /$re_retry/ ) {
		    sleep $stime++;
		    redo;
		}
		# Croak on error
		if ( $op and $op !~ qr/Making this ro/
		     and $op !~ qr/Entry.*multiple times, using first found/ ) {
		    croak $op;
		}
	    }
	}
    }

    if ( $opts{persistent} == 1 ) {
        # Remove previous export from the exports file
        $cont = _del_exports_from_file($conn, @paths);
        # Add exports to exports file
	foreach my $path (keys %$paths_hashref) {
	    my $access = $paths_hashref->{$path}->{OPTIONS} ? $paths_hashref->{$path}->{OPTIONS} : $opts{OPTIONS};

	    $cont .= "$path\t-$access\n";
	}
        wrfile($conn, $export_file, $cont);
        croak $@ if $@;
    }

    $retries = $opts{retry};
    if($opts{export_all}) {	
	# Export command
	$cmd = "$export_cmd -a";
	# Redo block
	{
	    $op = sayEOF($conn, $cmd);
	    if ( $retries-- and $op =~ /$re_retry/ ) {
		sleep $stime++;
		redo;
	    }
	    # Croak on error
	    if ( $op and $op !~ qr/Making this ro/
		 and $op !~ qr/Entry.*multiple times, using first found/ ) {
		croak $op;
	    }
	}
    }

    # Croak on warning
    croak $op if $op;

    return 1;
}


#---------------------------------------------------------------------
## @name ::unexport_path
##
## @description
##  Unexports the specified path on the filer provided by the
##  class object.
##
## @synopsis
##  $op = $obj->unexport_path($path, $persist, $ponly, retry => $retry);
##
## @arg $path
##  Required: Path to be unexported.
##
## @arg $persist
##  Required: True for a persistent unexport.
##  False for a non-persistent unexport.
##
## @arg $ponly
##  Required: Remove persistent export only.
##
## @arg retry
##  Required: Number of times to retry the command if the exportfs
##  command is busy.
##
## @return
##  1 if succeeds.
##
## @example
##  $op = $obj->unexport_path('/vol/testvol', 0, 0);
#---------------------------------------------------------------------
sub unexport_path($$$$%;) {
    my ($this, $path, $persistent, $persist_only, %opts) = @_;
    my ($cmd, $op);
    my $conn    = $this->{conn};
    my $retries = $opts{retry};
    my $stime   = 1;

    # Unexport command
    if ( ! $persist_only ) {
        $cmd = "$export_cmd -u $path";
        # Redo block
        {
            $op = sayEOF($conn, $cmd);
            if ( $retries-- and $op =~ /$re_retry/ ) {
                sleep $stime++;
                redo;
            }
            croak $op if $op;
        }
    }

    if ( $persistent ) {
        # Remove export from the exports file
        $op = _del_exports_from_file($conn, $path);
        wrfile($conn, $export_file, $op);
        croak $@ if $@;
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::list_exports
##
## @description
##  List all exports on the filer provided by the class object.
##
## @synopsis
##  $list_h = $obj->list_exports(retry => $retry);
##
## @arg retry
##  Required: Number of times to retry the command if the exportfs
##  command is busy.
##
## @return
##  A reference to a hash, where the keys are the exported paths
##  and the values are the access options.
#---------------------------------------------------------------------
sub list_exports($%;) {
    my ($this, %opts) = @_;
    my $retries = $opts{retry};
    my $stime   = 1;
    my $list_h;

    # Redo block
    {
        # Call inherited list_exports
        $list_h = eval { $this->SUPER::list_exports() };
        if ( $retries-- and $@ =~ /$re_retry/ ) {
            sleep $stime++;
            redo;
        }
        croak $@ if $@;
    }

    return $list_h;
}

#---------------------------------------------------------------------
## @name ::export_reset_persistent
##
## @description
##  Resets all persistent exports by re-exporting all persistent
##  exports only, all others are unexported.
##
## @synopsis
##  $op = $obj->export_reset_persistent(retry => $retry);
##
## @arg retry
##  Required: Number of times to retry the command if the exportfs
##  command is busy.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub export_reset_persistent($%;) {
    my ($this, %opts) = @_;
    my $retries = $opts{retry};
    my $stime   = 1;
    my $ret;

    # Redo block
    {
        # Call inherited export_reset_persistent
        $ret = eval { $this->SUPER::export_reset_persistent() };
        if ( $retries-- and $@ =~ /$re_retry/ ) {
            sleep $stime++;
            redo;
        }
        croak $@ if $@;
    }

    return $ret;
}

1;

