# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Export/Linux.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::Export::Linux;

## @summary
##  Class to handle OS dependent exports on Linux.
##
## @description
##  This class handles the OS specific portion of the NFS export
##  functions. It overrides the default behavior of the base
##  export class.
##
## @usage
##  use nfs::Export::Linux;
##
## @status  public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.

use strict;
use Carp qw(croak);
use SayWrap qw(sayEOF sayPerl);
use nfs::Subnets qw(fix_subnets);
use nfs::Export::ExportBase;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF);
use vars qw( @ISA );

#---------------------------------------------------------------------
# Inherit all functionality from ExportBase class.
#---------------------------------------------------------------------
@ISA = qw(nfs::Export::ExportBase);

#---------------------------------------------------------------------
## @name ::new
##
## @description
##  Linux specific class constructor. It overrides default
##  behavior of the base exports class.
##
## @return
##  Linux specific export object.
#---------------------------------------------------------------------
sub new($$;) {
    my ($class, $conn) = @_;

    # Calls constructor of parent class with
    # the following overrides:
    SUPER::new $class ( $conn,
        # OS supports the '@' on netgroups
        netgrp_atsign => 1
    );
}

#---------------------------------------------------------------------
# _break_options_per_client
# Breaks access options on a per client basis.
# Returns a hash reference with client as a key and access as the value,
# and an array reference of common options for all clients.
#---------------------------------------------------------------------
sub _break_options_per_client($;) {
    my $access = shift;
    my (%opt_h, @common_opt, $acc);

    # Change anon option to corresponding Linux option
    $access =~ s/\banon=/anonuid=/g;

    # Loop through each access option
    foreach my $acc_opt ( split /,/, $access ) {
        if ( $acc_opt !~ /^(ro|rw|root)\b/ ) {
            # Save common options for all clients
            push @common_opt, $acc_opt;
            next;
        }
        ($acc, $acc_opt) = split /=/, $acc_opt;
        $acc_opt =~ s/$/:/;
        # Change root option to corresponding Linux option
        $acc =~ s/^root$/no_root_squash/;
        foreach my $client ( $acc_opt =~ /(.*?):/g ) {
            $client =~ s/^$/\*/;    # Client is empty, export to everyone '*'
            # Save specific access options per client
            push @{$opt_h{$client}}, $acc;
        }
    }

    if ( $opt_h{'*'} ) {
        # Add r[ow] to clients that don't have a specific access option
        foreach my $client (keys %opt_h) {
            push(@{$opt_h{$client}}, $opt_h{'*'}->[0])
              if ! grep { /^r[ow]$/ } @{$opt_h{$client}}
        }
    }

    return (\%opt_h, \@common_opt);
}

#---------------------------------------------------------------------
## @name ::export_path
##
## @description
##  Overrides the default export_path function from the base
##  exports class.
##  Linux uses an export command on a per client basis; i.e.:
##  exporting a path to a group of clients is done by exporting
##  the same path to each client individually, e.g.:
##  exporting the following:
##
##    exportfs -i -o ro=client1:client2,rw /vol/testvol
##
##  requires the following exports in Linux:
##
##    exportfs -i -o ro client1:/vol/testvol
##
##    exportfs -i -o ro client2:/vol/testvol
##
##    exportfs -i -o rw *:/vol/testvol
##
## @synopsis
##  $op = $obj->export_path($path, $access, $persist);
##
## @arg $path
##  Required: Path to be exported.
##
## @arg $access
##  Required: Comma separated access options.
##
## @arg $persist
##  Required: True for a persistent export.
##  False for a non-persistent export.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub export_path($$$$;) {
    my ($this, $path, $access, $persistent) = @_;
    my ($cmd, @exp_c, $exps, $opt_h, $common_opt, $acc, $op);
    my $export_cmd   = $this->{export_cmd};
    my $export_flags = $this->{export_flags};
    my $sudo         = $this->{sudo};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    if ( $access ) {
        # Validate host entries, linux does not accept
        # negated hosts.
        my ($hosts) = ( $access =~ /\br[ow]=([^,])/ );
        my ($op)    = ( $access =~ /\broot=([^,])/ );
        $hosts = ":$hosts:$op";
        if ( $hosts =~ /:-/ ) {
            croak "Linux nfs does not support negated hosts";
        }

        # Validate and fix all subnet specifications provided
        # in the access list.
        $access = fix_subnets($access);

        # Fix all domains, prefix a '*'
        $access =~ s/([=:]-?)\./$1\*\./g;
 
        ($opt_h, $common_opt) = _break_options_per_client($access);

        # Add sync option, so Linux does not complain:
        @exp_c = grep { /^a?sync$/ } @$common_opt;
        if ( ! %$opt_h ) {
            # No ro, rw, or root specified, export to
            # the world with only the common options.
            push @{$opt_h->{'*'}}, @exp_c;
            @$common_opt = grep { !/^a?sync$/ } @$common_opt;
        } else {
            push(@$common_opt, 'sync') if !@exp_c;
        }

        # Create a list of export commands,
        # one for each client.
        @exp_c = map {
            "q{$export_cmd $export_flags " .
            join(',',  @$common_opt, @{$opt_h->{$_}}) .
            " $_:$path}";
        } keys %$opt_h;
    } else {
        # Export to the world with default option
        @exp_c = ("q{$export_cmd -i *:$path}");
    }

    # List of export commands
    $exps = join(', ', @exp_c);

    # Perl code to go through the list of export
    # commands and execute them one by one.
    # First, it unexports all client:/path
    # combinations which are not on the list of
    # exports, since all others will be replaced.
    $cmd = qq{
        \@argv = ($exps);
        foreach (qx($export_cmd)) \{
            (\$vol, \$cl) = split /\\s+/;
            \$cl = "*" if \$cl eq "<world>";
            next if \$vol ne "$path";
            \$err .= qx($export_cmd -u \$cl:\$vol);
            \$err .= \$CHILD_ERROR if \$CHILD_ERROR;
            next if grep \{ /\\s\$cl:\$vol/ \} \@argv;
        \}
        foreach \$exp_cmd (\@argv) \{
            \$err .= qx(\$exp_cmd);
            \$err .= \$CHILD_ERROR if \$CHILD_ERROR;
        \}
    };

    if ( $persistent ) {
        # Delete export from exports file; don't close file
        $cmd .= $this->del_persistent_cmd($path, $this->NO_CLOSE);
        # Add new export to exports file
        $cmd .= qq{
            \$first = 1;
            foreach (\@argv) \{
                (\$opt, \$cl, \$vol) = /^.*-i -o (.*)\\s+(.*):(.*)/;
                if (\$first) \{
                    print OUT "\$vol";
                    \$first = 0;
                \}
                print OUT " \$cl(\$opt)";
            \}
            print OUT "\\n";
            close(OUT);
        };
    }

    # Display error or ok message
    $cmd .= qq{
        print \$err ? \$err : q\{--export_ok--\};
    };

    # Execute the perl code on the server.
    (undef, undef, $op) = sayPerl($this->{conn}, $cmd, $sudo, @say_opts);
    croak $op if ( $op !~ /^(--export_ok--|--persist_ok--)$/ );

    return 1;
}


#---------------------------------------------------------------------
## @name ::export_paths
##
## @description
##  This function is calling export_path in the loop
## 
## @author gvg@netapp.com
##
#---------------------------------------------------------------------
sub export_paths() {
    my ($this, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
	    paths          => { type => HASHREF },
            OPTIONS        => { type => SCALAR },
	    persistent     => { type => BOOLEAN },
            retry          => { type => SCALAR },
	    export_all     => { type => BOOLEAN, default => 0, },
        },
    );

    # $export_all is not implemented on this plateform for now. It
    # just loops through paths one by one

    my $paths_hashref = $opts{paths};

    foreach my $path (keys %$paths_hashref) {
	my $access = $paths_hashref->{$path}->{OPTIONS} ? $paths_hashref->{$path}->{OPTIONS} : $opts{OPTIONS};
	export_path($this, $path, $access, $opts{persistent});
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::unexport_path
##
## @description
##  Overrides the default unexport_path function from the base
##  export class.
##  It gets a list of exported paths, if the path to be unexported
##  is found, then it issues a command to unexport that path for
##  each client.
##
## @synopsis
##  $op = $obj->unexport_path($path, $persist, $ponly);
##
## @arg $path
##  Required: Path to be unexported.
##
## @arg $persist
##  Required: True for a persistent unexport.
##  False for a non-persistent unexport.
##
## @arg $ponly
##  Required: Remove persistent export only.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub unexport_path($$$$;) {
    my ($this, $path, $persistent, $persist_only) = @_;
    my ($cmd, @exp_c, $opt_h, $common_opt, $acc, $op);
    my $export_cmd = $this->{export_cmd};
    my $sudo       = $this->{sudo};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Perl code to go through the list of exported
    # paths and unexport those that match the path.
    if ( ! $persist_only ) {
        $cmd = qq{
            my \$cline;
            foreach \$export (qx($export_cmd)) \{
                \$exp_line = \$export;
                chomp \$exp_line;
                \$exp_line =~ s/^\\s+//;
                if ( ! \$cline ) \{
                    (\$vol, \$exp_line) = split /\\s+/, \$exp_line;
                    if ( ! \$exp_line ) \{
                        \$cline = 1;
                        next;
                    \}
                \}
                \$cl = \$exp_line;
                \$cline = 0;
                next if \$vol ne "$path";
                \$cl = "*" if \$cl eq q\{<world>\};
                \$err .= qx($export_cmd -u \$cl:\$vol);
                \$err .= \$CHILD_ERROR if \$CHILD_ERROR;
            \}
        };
    }

    if ( $persistent ) {
        # Delete export from exports file; close file
        $cmd .= $this->del_persistent_cmd($path, $this->CLOSE);
    }

    # Display error or ok message
    $cmd .= qq{
        print \$err ? \$err : q\{--unexport_ok--\};
    };

    # Execute the perl code on the server.
    (undef, undef, $op) = sayPerl($this->{conn}, $cmd, $sudo, @say_opts);
    croak $op if ( $op !~ /(--unexport_ok--|--persist_ok--)/ );

    return 1;
}

#---------------------------------------------------------------------
## @name ::list_exports
##
## @description
##  Overrides the default list_exports function from the base
##  export class.
##
## @synopsis
##  $list_h = $obj->list_exports();
##
## @return
##  A reference to a hash, where the keys are the exported paths
##  and the values are the access options.
##  Since linux can have any options per client for the same path,
##  this function only returns the access options for the last occurrence
##  of path.
#---------------------------------------------------------------------
sub list_exports($;) {
    my $this = shift;
    my ($cmd, $vol, $client, $access, %list_h, %access_h, $cline, $op);
    my $export_cmd = $this->{export_cmd};
    my $sudo       = $this->{sudo};

    # Command to list all exports
    $cmd = "$sudo $export_cmd -v";

    # Run the command on the server
    $op = sayEOF($this->{conn}, $cmd);
    croak $op if ($op and $op !~ /^\//);

    # Create a hash where the exported path is the key
    # and the value is a hash to the list of clients
    # for each ro, rw, root.
    foreach my $exp_line (split /\n/, $op) {
        $exp_line =~ s/^\s+//;	# Get rid off leading spaces
        # Get exported path
        if ( ! $cline ) {
            ($vol, $exp_line) = split /\s+/, $exp_line;
            if ( ! $exp_line ) {
                # Line continues on next line
                $cline = 1;
                next;
            }
        }
        $cline = 0;
        # Get client and its access
        ($client, $exp_line) = split /[()]/, $exp_line;
        # Change back to standard anon
        $exp_line =~ s/\banonuid=/anon=/;
        if ( $client ne '<world>' ) {
            # Populate the list for root=
            $op = ($exp_line =~ s/\bno_root_squash\b//);
            push(@{$list_h{$vol}{root}}, $client) if $op;

            # Populate the list for ro=
            $op = ($exp_line =~ s/\bro\b//);
            push(@{$list_h{$vol}{ro}}, $client) if $op;

            # Populate the list for rw=
            $op = ($exp_line =~ s/\brw\b//);
            push(@{$list_h{$vol}{rw}}, $client) if $op;
        } else {
            $exp_line =~ s/\bno_root_squash\b/root/;
        }
        # Get rid off default non-root option
        $exp_line =~ s/\broot_squash\b//g;
        $exp_line =~ s/,+/,/g;	# Get rid off multiple ','s
        $exp_line =~ s/^,//;	# Get rid off leading ','s
        $exp_line =~ s/,$//;	# Get rid off tailing ','
        $list_h{$vol}{opts} = $exp_line;
    }

    # Create return hash of accesses
    foreach $vol (keys %list_h) {
        # Get common options for all clients
        # for this exported path
        $access = $list_h{$vol}{opts};
        foreach my $acc (keys %{$list_h{$vol}}) {
            # Skip over the common options
            next if $acc =~ /^opts$/;
            # Remove r[ow]=list if r[ow] is available
            next if $list_h{$vol}{opts} =~ /\b$acc\b/;
            # Add [ro|rw|root]=list to access
            # The order must be reversed since the first on the list
            # was the last to be exported.
            $access .= ",$acc=" . join(':', reverse @{$list_h{$vol}{$acc}});
        }
        $access =~ s/^,+//g;	# Get rid off leading ','s
        # Change all dns domains
        $access =~ s/([=:])(-?)\*\./$1$2\./g;
        $access_h{$vol} = $access;
    }

    return \%access_h;
}

1;

