#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Export/NFSexports.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::Export::NFSexports;

## @summary Library for the NFS exports functional tests
##
## @description
##   Common routines for NFS exports functional testing.
##
## @usage
##   use nfs::Export::NFSexports;
##
## @status	public
##
## @author	corbin@netapp.com
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;
use base qw(Exporter);

use Tharn;
use Carp qw(croak);
use SayWrap qw(sayPerl);
use SPARSE::Util::Logger;
use GenericMount qw(mount umount);
use Module::Common qw(validate_mixed);
use Network::Utils qw(run_getXXbyYY get_subnet);
use nfs::Subnets qw(validate_subnet fix_subnets);
use FS::Util::Setup qw(create_vol common_cleanup);
use nfs::GenericExport qw(export_path unexport_path get_export_options list_exports);
use Filer::Version qw(get_wafl_version WV_ANCHORSTEAM_MIN UNSUPPORTED_FEATURE_EXCEPTION);
use Params::Validate qw(validate validate_pos SCALAR BOOLEAN ARRAYREF HASHREF CODEREF OBJECT UNDEF);
use nfs::RemoteFileIO qw(client_read_file client_write_file client_del_file client_exec_file client_mknod);
use filer qw(filer_options_get filer_options_set filer_netgroup_add filer_access_cache_flush filer_access_cache_get filer_client_cache filer_qtree_get filer_qtree_security_set);

our @EXPORT_OK = qw(
  set_env
  clean_env
  create_host_list
  check_export_access
  check_export_access_list
  check_host_in_netgroup
  add_result
  compare_access_cache
  check_access_cache
  basic_access_test
  export_path_mount_and_read
  access_cache_test
  anon_properties_test
  nosuid_option_test
  is_atnetgroup_supported_on_filer
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '0.01';

#---------------------------------------------------------------------
# Package local constants
#---------------------------------------------------------------------
use constant SUDO => '/usr/software/bin/sudo';

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $export_list;
my $nfs_v4_enable;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub set_env(\%;%);
sub clean_env(\%;%);
sub validate_clients(\%;);
sub create_host_list($$;);
sub check_error(\$$$;);
sub client_read_write($$$$;);
sub check_export_access(\%$$;%);
sub check_export_access_list(\%\@;);
sub check_host_in_netgroup($$$;);
sub get_dns_domain($$;);
sub add_result($$;);
sub compare_access_cache($\%;%);
sub check_access_cache(\%\@;%);
sub basic_access_test(%;);
sub export_path_mount_and_read(%;);
sub access_cache_test(%;);
sub anon_properties_test(%;);
sub nosuid_option_test(%;);
sub is_atnetgroup_supported_on_filer(%;);

#---------------------------------------------------------------------
## @name ::set_env
## @description
##  Set up environment for NFS export tests and initialize all common
##  variables.  The first argument is the structure to be initialized
##  and the rest of the arguments are options passed in "key => value"
##  style.
##
## @synopsis
##  $op = set_env( %setup_h [, %options ] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables will get initialized in this function.
##
## @arg pclients
##  Optional: Reference to list of client hostspecs used for positive testing.
##  These clients must be in the same subnet and dns domain.
##  None of the clients can be 'localhost'.
##
## @arg nclients
##  Optional: Reference to list of client hostspecs used for negative testing.
##  These clients must not be in the same subnet or dns domain as the
##  positive testing clients.
##  None of the clients can be 'localhost'.
##
## @arg users
##  Optional: Reference to list of user names to check and get their uid's.
##
## @arg client_conn
##  Optional: Clients connection spec, (default => '#rsh').
##
## @arg setup
##  Optional: Reference to a function to run at the end of setup.
##
## @arg nfs_timeouts
##  Optional: Value to set the nfs export timeouts. If set to 0, it does
##  not set the timeouts but verifies proper nfs export timeout values.
##
## @arg root_file
##  Optional: Boolean, true to create a root-owned file, (default is 0).
##
## @return
##  Nothing.
##
## @notes
##  Structure used by the NFS export tests:
##
##  %setup_h = (
##
##      fc              => Filer's connection object.
##
##      cc              => Reference to a list of client connection objects.
##
##      ncc             => Reference to a list of negative testing client
##                         connection objects.
##
##      pcc             => Reference to a list of positive testing client
##                         connection objects.
##
##      npos            => Number of positive testing clients.
##
##      tfname          => The file name used for testing proper access on
##                         clients.
##
##      tfcontents      => The contents of the test file.
##
##      vol_name        => The volume name created on the filer.
##
##      volume          => The full path to the volume on the filer.
##
##      access_options  => Reference to a list of access options to test for,
##                         e.g., ro, rw, root.
##
##      host_entries    => Reference to a list of host entry types to test for,
##                         e.g., hostname, ip, netgroup, dns domain and subnet.
##
##      client_ip       => Reference to a list of client ip addresses as seen
##                         by the filer.
##
##      pclient_ip      => Reference to a list of positive testing client
##                         ip addresses as seen by the filer.
##
##      nclient_ip      => Reference to a list of negative testing client
##                         ip addresses as seen by the filer.
##
##      client_names    => Reference to a list of client hostnames.
##
##      pclient_names   => Reference to a list of positive testing client
##                         hostnames.
##
##      nclient_names   => Reference to a list of negative testing client
##                         hostnames.
##
##      users           => Reference to a hash of users: { username => uid }.
##
##      clients_h       => Reference to a hash where the key is either
##                         the hostname or ip address of the primary
##                         client and the value is the concatenated list
##                         of the rest of the positive testing client
##                         hostnames or ip addresses respectively.
##
##      client_ip_h     => Reference to a hash where the key is either
##                         the client's hostname or ip address
##                         (as returned from Hostrec::connhost())
##                         and the value is the ip address of the client
##                         as seen by the filer.
##
##      setup_mtpoint   => The mount point of volume on the primary client
##                         for setup and cleanup purposes.
##
##      dns_domain      => The DNS domain of the positive testing clients.
##
##      dns_netgroup    => The netgroup that looks like a dns domain.
##
##      tst_netgroup    => The netgroup which the positive testing clients
##                         belong to.
##
##      neg_timeout     => Initial value of nfs.export.neg.timeout option.
##
##      pos_timeout     => Initial value of nfs.export.pos.timeout option.
##
##      harvest_timeout => Initial value of nfs.export.harvest.timeout option.
##
##  );
#---------------------------------------------------------------------
sub set_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            pclients     => { type => ARRAYREF, optional => 0 },
            nclients     => { type => ARRAYREF, optional => 1 },
            users        => { type => ARRAYREF, optional => 1 },
            client_conn  => { type => SCALAR,    default => '#rsh' },
            root_file    => { type => BOOLEAN,   default => 0 },
            nfs_timeouts => { type => BOOLEAN,  optional => 1 },
            setup        => { type => CODEREF,  optional => 1 },
        }
    );
    my ($mountpoint, $subnet, $n, $pos, $fail, $op, %qtree_out, $t_time);

    logStep("BEGIN: set_env: setup environment for test");

    # Connect to filer
    host($main::FILER);
    $set_h->{fc} = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $set_h->{fc}->timeout($main::CMDTIMEOUT);

    # Save the current export options for volume
    $export_list = list_exports($set_h->{fc});

    # Initialize variables to return
    $set_h->{tfname}       = "nfs_exports.$$";
    $set_h->{tfcontents}   = "This file is used to test access";
    $set_h->{tst_netgroup} = "export_functionality_test";
    $set_h->{dns_netgroup} = ".dns.export.test.lab.netapp.com";

    # Validate NFS_VERSION
    if ( $main::NFS_VERSION !~ /^[234]$/ ) {
        croak "set_env: Invalid NFS version <$main::NFS_VERSION>, " .
              "NFS_VERSION must be 2, 3 or 4";
    }

    # Validate that clients can not be 'localhost'
    # and must be fully qualified
    foreach my $client ( @{$opts{pclients}}, @{$opts{nclients}} ) {
        if ( $client =~ /^localhost\b/ ) {
            croak "set_env: Clients can not be 'localhost'";
        }
        # Get only the host part of the hostspec
        # and check for a fully qualified hostname
        my ($host) = ($client =~ /([^;]*)/);
        if ( $host !~ /\./ ) {
            croak "set_env: Clients must be fully qualified";
        }
    }

    # Common setup -- creates a volume depending
    # on the $main::SETUP_DONE parameter
    logStep("Create test volume if necessary");
    ($set_h->{fc}, $set_h->{vol_name}) = create_vol() or croak $@;

    # Full volume path
    $set_h->{volume} = "/vol/$set_h->{vol_name}";

    %qtree_out = filer_qtree_get($set_h->{fc}, volume => $set_h->{vol_name});
    if ($qtree_out{$set_h->{volume}}{Style} !~ /(mixed|unix)/) {
        $set_h->{qtree_style_org} = $qtree_out{$set_h->{volume}}{Style};

        # Set the qtree security mode on the default volume
        logStep("Setting volume qtree security mode setting to unix");
        if (filer_qtree_security_set($set_h->{fc}, $set_h->{volume},
                                     'unix') == 0) {
            croak("qtree set failed on $set_h->{volume}");
        }
    }

    logStep("Create connection handles for clients");
    $n = 0;
    $pos = 1;
    $fail = 0;
    # Create connection handles for all clients, start with the
    # positive testing clients and when it reaches the separator '-'
    # process the negative testing clients.
    # Do not croak on the first occurrence of an error, find as many
    # errors as possible before croaking.
    foreach my $client ( @{$opts{pclients}}, '-', @{$opts{nclients}} ) {
        if ( $client eq '-' ) {
            # Separator found, now process negative testing clients
            $set_h->{npos} = $n;
            $pos = 0;
            next;
        }
        # Create connection for client
        my $h = host($client);
        my $c = connect ( $opts{client_conn}, -timeout, $main::TIMEOUT );
        $c->timeout( $main::CMDTIMEOUT );
        my $hostname = $h->{hostname};

        # For exports testing the client name used must be
        # resolvable to an IP addr on the filer for the mount
        # to succeed.
        $op = run_getXXbyYY($set_h->{fc}, "gethostbyname_r", $hostname);
        if ( $op =~ /not found/ ) {
            logcomment("set_env: $hostname not resolvable on filer");
            undef $op;
            ++$fail;
        } else {
            # Save the filer's concept of the hostname to IP address
            # mapping because for exports all IP address and Name
            # resolution is checked on the filer.
            ($op) = ($op =~ /addresses:\s+(\d+.\d+.\d+.\d+)/);
        }
        # Save the connection handle, hostname,
        # and ip address for all clients
        push @{$set_h->{cc}}, $c;
        push @{$set_h->{client_names}}, $hostname;
        push @{$set_h->{client_ip}}, $op;
        # Save hash of clients: key hostname/connectrec connhost
        # and the ip of client as seen by the filer is the value.
        $set_h->{client_ip_h}{$hostname} = $op;
        $set_h->{client_ip_h}{$c->connhost()} = $op;
        if ( $pos ) {
            # Save the connection handle, hostname,
            # and ip address for positive testing clients
            push @{$set_h->{pcc}}, $c;
            push @{$set_h->{pclient_names}}, $hostname;
            push @{$set_h->{pclient_ip}}, $op;
        } else {
            # Save the connection handle, hostname,
            # and ip address for negative testing clients
            push @{$set_h->{ncc}}, $c;
            push @{$set_h->{nclient_names}}, $hostname;
            push @{$set_h->{nclient_ip}}, $op;
        }
        ++$n;
    }
    if ( $fail != 0 ) {
        croak("set_env: Client hostnames not resolvable on filer");
    }

    if ( $main::NFS_VERSION == 4 ) {
        # Save current NFSv4 state
        $nfs_v4_enable = filer_options_get($set_h->{fc}, 'nfs.v4.enable');

        # Enable NFSv4 on filer
        if ( $nfs_v4_enable eq 'off' ) {
            logStep("Enable NFSv4");
            filer_options_set($set_h->{fc}, 'nfs.v4.enable', 'on');
        } else {
            # No change was made
            $nfs_v4_enable = undef;
        }
    }

    logStep('Mount the test volume');
    $mountpoint = mount($set_h->{fc}->connhost(),
                        $set_h->{volume},
                        client  => $set_h->{cc}[0],
                        mntopts => "vers=$main::NFS_VERSION",
                        'chmod' => '0777') or croak $@;

    # Save mountpoint so if setup should fail,
    # it will get properly unmounted.
    $set_h->{setup_mtpoint} = $mountpoint;

    logStep("Create test file");
    eval {
        client_write_file($set_h->{cc}[0],
                          "$mountpoint/$set_h->{tfname}",
                          $set_h->{tfcontents});
    };
    croak("set_env: Couldn't create test file: $@") if $@;

    # Create root-owned file if specified
    if ( $opts{root_file} ) {
        logStep("Create root-owned file");
        eval {
            client_write_file($set_h->{cc}[0],
                              "$mountpoint/$set_h->{tfname}.root",
                              $set_h->{tfcontents},
                              root => 1);
        };
        croak("set_env: Couldn't create root-owned file: $@") if $@;
    }

    # Save and check proper nfs export timeouts
    if ( defined($opts{nfs_timeouts}) ) {
        logStep("Save current nfs export timeouts");
        $set_h->{neg_timeout} =
            filer_options_get($set_h->{fc}, 'nfs.export.neg.timeout');
        $set_h->{pos_timeout} =
            filer_options_get($set_h->{fc}, 'nfs.export.pos.timeout');
        $set_h->{harvest_timeout} =
            filer_options_get($set_h->{fc}, 'nfs.export.harvest.timeout');

        if ( $opts{nfs_timeouts} == 0 ) {
            logStep("Check proper timeouts");
            if ( $set_h->{neg_timeout} < 50
              or $set_h->{pos_timeout} < 50
              or $set_h->{harvest_timeout} < 50 ) {
                croak "nfs.export.*.timeout are too small";
            }
        } elsif ( $opts{nfs_timeouts} > 0 ) {
            $op = $opts{nfs_timeouts};
            # Save the option to properly set values back
            # to original values on clean_env
            $set_h->{nfs_timeouts} = $op;
            logStep("Set nfs export timeouts to $op sec");
            filer_options_set($set_h->{fc}, 'nfs.export.neg.timeout',     $op);
            filer_options_set($set_h->{fc}, 'nfs.export.pos.timeout',     $op);
            filer_options_set($set_h->{fc}, 'nfs.export.harvest.timeout', $op);
        } else {
            croak "Option nfs_timeouts must be >= 0";
        }
    }

    # Initialize access list used in all tests
    @{$set_h->{access_options}} = qw(ro rw root);

    # Get subnet of primary client and validate it
    $subnet = get_subnet($set_h->{fc}, $set_h->{client_ip}[0]);
    eval { $subnet = validate_subnet($subnet) };
    croak("set_env: Couldn't validate subnet: $@") if ($@ or !$subnet);

    # Get dns domain of primary client
    $set_h->{dns_domain} = get_dns_domain($set_h->{fc}, $set_h->{client_ip}[0]);

    # Contains an array of different ways to identify the
    # primary client - name, ip, netgroup, dns domain, and subnet
    @{$set_h->{host_entries}} = ( $set_h->{cc}[0]->hostp()->hostname(),
                                  $set_h->{client_ip}[0],
                                  '@' . $set_h->{tst_netgroup},
                                  '.' . $set_h->{dns_domain},
                                  $subnet );

    logStep("Validate clients");
    $t_time = time();
    $n = 1;
    # Validate all positive and negative testing clients.
    # This will attempt to create the netgroup and wait for
    # it to become online for a maximum of CMDTIMEOUT seconds.
    while ( time() - $t_time < $main::CMDTIMEOUT ) {
        $op = validate_clients( %$set_h ) and last;
        logStep("Clients do not meet the requirements, " .
                "attempting to add the proper netgroup.");
        filer_netgroup_add($set_h->{fc}, $set_h->{tst_netgroup},
                         @{$set_h->{pclient_names}}) if $n;
        # Wait a few seconds for the filer to read the netgroup
        # into memory.
        sleep(10);
        $n = 0;
    }
    croak "set_env: Timeout trying to validate clients" if ! $op;

    # Create a hash of positive testing clients excluding the
    # the primary client.  The keys are the name and ip address
    # of the primary client, and the values are the concatenated
    # strings of hostnames and ip addresses, respectively.
    #    hostname1 => ':hostname2:hostname3:...:hostnameN'
    #    ipaddr1   => ':ipaddr2:ipaddr3:...:ipaddrN'
    # where N is the number of positive testing clients.
    logStep("Create a list of clients for positive testing");
    $n = $set_h->{npos} - 1;
    %{$set_h->{clients_h}} = (
                  $set_h->{host_entries}[0] =>
                     ':' . join(':', @{$set_h->{pclient_names}}[1..$n]),
                  $set_h->{host_entries}[1] =>
                     ':' . join(':', @{$set_h->{pclient_ip}}[1..$n])
                 );

    if ( $opts{setup} ) {
        # Run specific setup for test
        &{$opts{setup}}( $set_h );
        # Just in case the specific setup function returns eval
        croak $@ if $@;
    }

    logStep("Unmount volume on primary client");
    umount($mountpoint, client=>$set_h->{cc}[0]) or croak $@;

    # Already unmounted, no need to unmount it anymore at clean_env
    # if setup should fail
    undef $set_h->{setup_mtpoint};

    logStep("Unexport the volume");
    unexport_path($set_h->{fc}, $set_h->{volume}) or croak $@;

    logStep("END: set_env");
    return;
}

#---------------------------------------------------------------------
## @name ::clean_env
## @description
##  Cleanup after a test run if necessary by removing the files
##  created on the filer.
##
## @synopsis
##  $op = clean_env( %setup_h [, cleanup=>\&cleanup_f] );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables are initialized on setup.
##
## @arg cleanup
##  Optional: Reference to a function to run after mounting volume on
##  client and before calling common cleanup.
##
## @return
##  Nothing, on error logs a warning.
#---------------------------------------------------------------------
sub clean_env(\%;%) {
    my ($set_h, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF }
        ],
        {
            cleanup => { type => CODEREF, optional => 1 }
        }
    );
    my $mountpoint = $set_h->{setup_mtpoint};

    logStep("CLEANUP == $main::CLEANUP");

    # Catch errors trying to re-connect to the filer
    eval {
        if ( ! $set_h->{fc} ) {
            # Filer connection is not defined, so re-connect
            logStep("Re-connect to the filer");
            host($main::FILER);
            $set_h->{fc} = connect($main::FILER_CONN, -timeout, $main::TIMEOUT);
            $set_h->{fc}->timeout($main::CMDTIMEOUT);
        }

        # Catch any errors to properly run common_cleanup
        if ( $main::CLEANUP ne 'none' and $set_h->{cc}[0] ) {
            eval {
                logStep("Export test volume");
                # Export volume as read-write with root access
                export_path($set_h->{fc}, $set_h->{volume},
                            OPTIONS=>"rw,anon=0") or croak $@;

                if ( ! $mountpoint ) {
                    logStep("Mount volume on client");
                    $mountpoint = mount($set_h->{fc}->connhost(),
                                        $set_h->{volume},
                                        mntopts=>"vers=$main::NFS_VERSION",
                                        client=>$set_h->{cc}[0],
                                        mountretry=>2) or croak $@;
                } else {
                    logStep("Volume already mounted on client");
                }

                if ( $opts{cleanup} ) {
                    # Run specific cleanup for test
                    # Pass the mountpoint to the specific cleanup function
                    $set_h->{setup_mtpoint} = $mountpoint;

                    # Eval the specific cleanup function just in
                    # case the function does not return eval, so
                    # the error can be properly log as a warning.
                    eval {
                        &{$opts{cleanup}}( $set_h );
                        # Just in case the function does returns eval
                        croak $@ if $@;
                    };
                    logWarn("clean_env: $@") if $@;
                }

                if ( $set_h->{tfname} ) {
                    logStep("Delete all files created by this test");
                    eval { client_del_file($set_h->{cc}[0],
                                           "$mountpoint/$set_h->{tfname}*") };
                };

                logStep("Unmount volume on client");
                umount($mountpoint, client=>$set_h->{cc}[0]);
                # Do not try to unmount in common_cleanup
                undef $mountpoint;
            };

            logWarn("clean_env: $@") if $@;

            if ( $set_h->{nfs_timeouts} > 0 ) {
                logStep("Reset nfs export timeouts to original values");
                eval { filer_options_set($set_h->{fc},
                                         'nfs.export.neg.timeout',
                                         $set_h->{neg_timeout}) };
                logWarn("clean_env: $@") if $@;
                eval { filer_options_set($set_h->{fc},
                                         'nfs.export.pos.timeout',
                                         $set_h->{pos_timeout}) };
                logWarn("clean_env: $@") if $@;
                eval { filer_options_set($set_h->{fc},
                                         'nfs.export.harvest.timeout',
                                         $set_h->{harvest_timeout}) };
                logWarn("clean_env: $@") if $@;
            }

            if ( defined($nfs_v4_enable) ) {
                logStep("Restore NFSv4 state");
                eval { filer_options_set($set_h->{fc},
                                         'nfs.v4.enable',
                                         $nfs_v4_enable) };
                logWarn("clean_env: $@") if $@;
            }

            if ( defined($set_h->{qtree_style_org}) ) {
                logStep("Restore volume qtree security mode setting");
                eval { filer_qtree_security_set($set_h->{fc}, $set_h->{volume},
                                                $set_h->{qtree_style_org}) };
            }
        }

        # Run common cleanup and log a warning in order
        # to continue and unexport the volume
        logStep("Run common cleanup");
        common_cleanup($set_h->{fc}, client=>$set_h->{cc}[0], mnt=>$mountpoint)
          or logWarn("clean_env: $@");

        if ( $main::CLEANUP ne 'none' and $set_h->{volume} and $main::CLEANUP ne 'all') {
            my $access;
            if ( defined($export_list) and $export_list->{$set_h->{volume}} ) {
                $access = $export_list->{$set_h->{volume}};
            } else {
                $access = 'sec=sys,rw,nosuid';
            }
            logStep('Restore export options for volume');
            export_path($set_h->{fc}, $set_h->{volume}, OPTIONS=>$access)
              or logWarn("clean_env: $@");
        }
    };

    logWarn("clean_env: $@") if $@;
    logStep("END: CLEANUP");

    return;
}

#---------------------------------------------------------------------
## @name ::validate_clients
## @description
##  Validates the clients used for positive testing, i.e, clients
##  which should have proper access to a specific nfs exported path.
##  Also validates clients used for negative testing.
##
## @synopsis
##  $op = validate_clients( %setup_h );
##
## @arg %setup_h
##  Required: Structure of common variables used by setup, cleanup and tests.
##  These variables are initialized on setup.
##
## @return
##  1 if all clients are valid.
##  0 if anyone client is invalid.
#---------------------------------------------------------------------
sub validate_clients(\%;) {
    my ($setup_h) = validate_pos(
        @_,
        { type => HASHREF }
    );
    my ($n, $cl, $op, $msg, $neg, $fmsg, $host, $ipaddr);
    my ($prim_client, $prim_client_subnet, $prim_client_dnsdomain);

    logStep("BEGIN: validate_clients: validating clients");

    $n = 0;
    # Validate all clients -- starting with the positive
    # testing clients
    logStep("Validate clients for positive testing");
    foreach $cl ( @{$setup_h->{cc}} ) {
        if ( $n == $setup_h->{npos} ) {
            # Already processed all the positive testing clients
            # start processing the negative testing clients
            logStep("Validate clients for negative testing");
            # Negate the check criteria
            $neg = 1;
            # Used to properly display error message
            $fmsg = " not";
        }
        $host   = $setup_h->{client_names}[$n];
        $ipaddr = $setup_h->{client_ip}[$n];
        $prim_client = $host if ! $prim_client; # Save primary client's name
        # Check if host belongs to netgroup
        $op = check_host_in_netgroup($setup_h->{fc},
                                     $setup_h->{tst_netgroup}, $host);
        if ( ! $op xor $neg ) {
            $msg = "$host should$fmsg be a member of " .
                   "$setup_h->{tst_netgroup} netgroup";
            logcomment("validate_clients: $msg");
        }
        # Catch errors from get_subnet in order to process
        # all clients before dying.
        eval {
            # Get the subnet that the host belongs to
            $op = get_subnet($setup_h->{fc}, $ipaddr);
            logDbg("validate_clients: $host subnet($op)");
        };
        if ( ! $prim_client_subnet and $op ) {
            # Save primary client's subnet
            $prim_client_subnet = $op;
        } elsif ( ! $op ) {
            # Error getting the subnet
            $msg = "Could not get $host subnet";
            logcomment("validate_clients: $msg: $@");
        } elsif ( $op ne $prim_client_subnet xor $neg ) {
            # Host does not have the required subnet
            $msg = "$host subnet($op) should$fmsg be on same " .
                   "subnet as $prim_client subnet($prim_client_subnet)";
            logcomment("validate_clients: $msg");
        }

        # Catch errors from get_dns_domain in order to process
        # all clients before dying.
        eval {
            # Get the dns domain name for client
            $op = get_dns_domain($setup_h->{fc}, $ipaddr);
            logDbg("validate_clients: $host dns domain($op)");
        };
        if ( ! $prim_client_dnsdomain and $op ) {
            # Save primary client's dns domain
            $prim_client_dnsdomain = $op;
        } elsif ( ! $op ) {
            # Error getting the dns domain
            $msg = "Could not get $host dns domain";
            logcomment("validate_clients: $msg: $@");
        } elsif ( $op ne $prim_client_dnsdomain xor $neg ) {
            # Host does not have the required dns domain
            $msg = "$host dns domain($op) should$fmsg be on same dns domain " .
                   "as $prim_client dns domain($prim_client_dnsdomain)";
            logcomment("validate_clients: $msg");
        }

        ++$n;
    }

    if ( $msg ) {
        # At least one of the clients failed validation
        return 0;
    }

    logStep("END: validate_clients:");
    return 1;
}

#---------------------------------------------------------------------
## @name ::create_host_list
## @description
##  Create list of host entries
##
## @synopsis
##  $hosts = create_host_list( $host, $count );
##
## @arg $host
##  Required: host entry definition, e.g.(hostname, ip, dns, ...)
##
## @arg $count
##  Required: Total number of hosts to create. The count is restricted
##  to less than: 254*(254 - third octet) + (255 - fourth octet) for
##  ip address.
##
## @return
##  Array of host entries
#---------------------------------------------------------------------
sub create_host_list($$;) {
    my ($host, $count) = validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR }
    );
    my (@hosts, $dnsdomain);

    if ( $host =~ /^\.(.*)/ ) {         # DNS?  e.g. ".domain.name"
        $dnsdomain = $1;                # and save main domain.name
        # Create list:
        #   .sub0.domain.name:.sub1.domain.name:...:.sub$count.domain.name
        @hosts = map { ".sub$_.$dnsdomain" } (0..$count);

    } elsif ( (my @octets = split(/\./, $host)) == 4  and $host !~ /\// ) {
        # Create list of IP's
        my $st = $octets[3] + 1; # starting ip
        foreach (0..int(($st+$count)/254)) {
            # only fit at most 254 ip's per octet
            my $end = ($st+$count) > 254 ? 254 : ($st+$count);
            @hosts = (@hosts, (map { join('.', (@octets[0..2], $_)) }
                            ($st..$end)));
            # Reduce total count by number of ip's created this iteration
            $count -= $end - $st + 1;
            ++$octets[2];  # Increment third octet
            $st = 1;       # Set starting point for next iteration
        }
    } elsif ( (my @octets = split(/\:/, $host)) == 8 and $host !~ /\// ) {
		my $v6_address;
		my $index = 7;
		my ($decval, $hexval, $four_hex);
		my $new_address;
		
		$host =~ /\[(.*)\]/;
		$v6_address = $1;
		@octets = split(/\:/, $v6_address);
		$decval = hex(@octets[$index]);
		foreach (1..int($count % 254)) {
		    $decval = $decval + 1;
			$hexval = sprintf("%x", $decval);
			$hexval =~ /(....)$/;
			$four_hex = $1;
			
			$new_address = join(':', (@octets[0..6], $four_hex));
			$new_address = "[$new_address]";
			
			@hosts = (@hosts, $new_address);
		}	
	} else {
        # Create list: host0:hosts1:...:host$count
        @hosts = map { $host.$_ } (0..$count);
    }

    @hosts;
}

#---------------------------------------------------------------------
# ::check_error
#
#  This is an internal function so it is not available for external use.
#
#  Appends the right message according to the error and the expected
#  failure flag. Returns 1 if expected behavior is found.
#---------------------------------------------------------------------
sub check_error(\$$$;) {
    my ($msg, $error, $fail) = @_;

    if ( $error ) {
        $$msg .= ' failed';
        $$msg .= ' as expected' if $fail;
        $$msg .= ": $error";
    } elsif ( $fail ) {
        $$msg .= ' should have failed';
    } else {
        $$msg .= ' succeeded as expected';
    }

    return ( $error xor $fail ) ? 0 : 1;
}

#---------------------------------------------------------------------
# ::client_read_write
#
#  This is an internal function so it is not available for external use.
#
#  Reads the given file on the client as the user (or root if specified).
#  It also creates a file on the client as the same user.
#  This is used to check for proper access on a client,
#  it uses only one trip to the client for both read and write.
#  This routine is used only by check_export_access() and it is
#  a separate function to improve readability.
#
#  $op = client_read_write($cc, $testfile", $cont, 0);
#    $op will be "--read_ok--" if ONLY read was allowed.
#    $op will be "--read_ok----write_ok--" if read/write was allowed.
#---------------------------------------------------------------------
sub client_read_write($$$$;) {
    my ($cc, $file, $contents, $root) = @_;
    my ($cmd, $sudo, $op);
    my $client_name = $cc->hostp()->hostname();

    # Get hostname only, this is appended to the
    # name of the file to be created
    $client_name =~ s/\..*//;

    # The following is a script to read and create a file on the client.
    # First it sets the creation mask for access by owner ONLY.
    # Reads a file then creates a file, checks ownership of file
    # and returns the uid.
    # The file being created is derived from the read filename,
    # the client's name and a random number.  If the file already
    # exists, another filename is generated.
    $cmd = qq{
        umask 077;
        \$cont = "$contents";
        \$file = "$file";
        if ( \$op = open(IN, "<\$file") ) \{
            \$op = <IN>;
            print qq(--read_ok--\$op\\n) if \$op =~ /\$cont/;
            close(IN);
        \}
        print qq(--read_err--\$OS_ERROR\\n) unless \$op;
        do \{
           \$rno += int(rand(100000));
           \$wfile = "\$file.$client_name.\$rno";
        \} while ( -e \$wfile );
        if ( \$op = open(OUT, ">\$wfile") ) \{
            \$op = print OUT \$cont;
            close(OUT);
            if ( \$op ) \{
                (\$uid,\$gid,\$size) = (stat("\$wfile"))[4,5,7];
                print qq(--write_ok--uid\$uid--gid\$gid--size\$size--\\n)
                  if \$size == length(\$cont);
            \}
        \}
        print qq(--write_err--\$OS_ERROR\\n) unless \$op;
        print qq(--wfile\$wfile--);
    };

    $sudo = SUDO if $root;
    # Execute script to create the file on the client
    (undef, undef, $op) = sayPerl ($cc, $cmd, $sudo, '-swallow');

    # Do not check for errors, error checking
    # will be done in check_export_access
    return $op;
}

#---------------------------------------------------------------------
## @name ::check_export_access
## @description
##  Checks given access on client. First, it mounts the volume
##  on client, then it reads and writes a file. If the root
##  option is provided, the read/write will be done as root.
##  If access is 'ro', client should ONLY be able to read; if
##  access is 'rw', client should be able to read and write.
##  If neither of 'expect_uid' or 'expect_gid' is provided,
##  the ownership of the file will be checked only if the 'root'
##  option is set. Unmounts the volume when done.
##  The read and write is not performed when the access option
##  is one of the following options: 'modify', 'remove' or 'execute';
##  instead, the proper operation is performed.
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_export_access(%setup_h, $cc, $access [, %options]);
##
## @arg %setup_h
##  Required: Hash of common variables initialized at setup.
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $access
##  Required: Check client for this access. Must be either 'ro', 'rw',
##  'modify', 'remove', or 'execute'.
##
## @arg root
##  Optional: Boolean, for reading/writing file as the user 'root'
##  (default is 0).
##
## @arg fail
##  Optional: Boolean, mount is expected to fail.
##  (default is 0).
##
## @arg read_fail
##  Optional: Boolean, read is expected to fail.
##  (default is 0).
##
## @arg write_fail
##  Optional: Boolean, write is expected to fail.
##  (default is 0).
##
## @arg root_cfail
##  Optional: Boolean, create as root is expected to fail.
##  (default is 0).
##
## @arg exec_fail
##  Optional: Boolean, execute is expected to fail.
##  (default is 0).
##
## @arg rm_fail
##  Optional: Boolean, remove is expected to fail.
##  (default is 0).
##
## @arg mount_opts
##  Optional: Options pass to mount.
##
## @arg expect_uid
##  Optional: Expected uid of file to be created.
##
## @arg expect_gid
##  Optional: Expected gid of file to be created.
##
## @arg check_ruid
##  Optional: Boolean, check that the file is owned by root
##  (default is 1). This check is only done if the 'root'
##  option is set and 'access' is set to 'modify'.
##
## @return
##  1 if client has proper access 0 if not.
##
## @example
##  $op = check_export_access(%setup_h, $cc, "rw");
##   Will return 1 if client is able to read and write.
##   Will return 0 if client is NOT able to read and write.
#---------------------------------------------------------------------
sub check_export_access(\%$$;%) {
    my ($setup_h, $cc, $access, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF },
            { type => OBJECT },
            { type => SCALAR, regex => qr/^(ro|rw|modify|remove|execute)$/ }
        ],
        {
            root       => { type => BOOLEAN,  default => 0 },
            fail       => { type => BOOLEAN,  default => 0 },
            read_fail  => { type => BOOLEAN,  default => 0 },
            write_fail => { type => BOOLEAN,  default => 0 },
            root_cfail => { type => BOOLEAN,  default => 0 },
            exec_fail  => { type => BOOLEAN,  default => 0 },
            rm_fail    => { type => BOOLEAN,  default => 0 },
            mount_opts => { type => SCALAR,  optional => 1 },
            expect_uid => { type => SCALAR,  optional => 1 },
            expect_gid => { type => SCALAR,  optional => 1 },
            check_ruid => { type => BOOLEAN,  default => 1 },
        }
    );
    my ($mountpoint, $mntopts, $op, $err, $uid, $gid, $size, $msg, $rslt);
    my ($wfile, $read_err, $write_err, $quote_mtpoint, $write_fail_cond, $no);
    my $fc         = $setup_h->{fc};
    my $volume     = $setup_h->{volume};
    my $tfname     = $setup_h->{tfname};
    my $tfcontents = $setup_h->{tfcontents};
    my $mount_opts = $opts{mount_opts};
    my $expect_uid = $opts{expect_uid};
    my $expect_gid = $opts{expect_gid};
    my $server     = $fc->hostp()->hostname();
    my $client     = $cc->connhost();
    my $user       = $opts{root} ? ' as root' : '';

    # Check setup variables
    if ( !$volume ) {
        croak "check_export_access: NFSexports was not properly initialized."
    }

    $rslt = eval {
        # Include mount options
        $mntopts = "vers=$main::NFS_VERSION";
        $mntopts .= ",$mount_opts" if defined($mount_opts);

        # Log correct access message
        if ( $access eq 'modify' ) {
            $no = $opts{write_fail} ? ' no' : '';
        } elsif ( $access eq 'remove' ) {
            $no = $opts{rm_fail} ? ' no' : '';
        } elsif ( $access eq 'execute' ) {
            $no = $opts{exec_fail} ? ' no' : '';
        } else {
            # This is for ro and rw
            $no = ( $opts{fail} or $opts{read_fail} or $opts{write_fail}
                    or ($opts{root} and $opts{root_cfail}) ) ?  ' no' : '';
        }

        # Log the status of all operations
        $msg = "\nVerifying that client $client has$no <$access> access" .
               "$user to $server:$volume\n";

        # Mount volume on client
        $mountpoint = mount($fc->connhost(), $volume, client=>$cc,
                            mountretry=>0, mntopts=>$mntopts);

        $no = $mountpoint ? '' : ' no';
        $msg .= "MOUNT $client has$no access$user to $server:$volume";

        if ( $mountpoint ) {
            # Check if mount should have failed
            if ( $opts{fail} ) {
                $msg .= ': mount succeeded where it should have failed';
                return 0;
            } else {
                $msg .= " as expected\n";
            }

            if ( $access eq 'modify' ) {
                $msg .= "MODIFY $tfname (open write only, create, truncate)";
                eval {
                    client_write_file($cc, "$mountpoint/$tfname", $tfcontents,
                                      root       => $opts{root},
                                      check_ruid => $opts{check_ruid});
                };
                return check_error($msg, $@, $opts{write_fail});
            } elsif ( $access eq 'remove' ) {
                $msg .= "REMOVE $tfname";
                eval {
                    client_del_file($cc,
                                    "$mountpoint/$tfname",
                                    root=>1, force=>0, recursive=>0);
                };
                return check_error($msg, $@, $opts{rm_fail});
            } elsif ( $access eq 'execute' ) {
                $msg .= "EXECUTE $tfname";
                eval {
                    client_exec_file($cc, "$mountpoint/$tfname",
                                     root=>$opts{root});
                };
                return check_error($msg, $@, $opts{exec_fail});
            }

            # Attempt to read and write a file
            $op = client_read_write($cc, "$mountpoint/$tfname", $tfcontents,
                                                                $opts{root});

            # Check for failures
            $op =~ s/[\r\n]//g;
            ($read_err) = ( $op =~ /--read_err--(.*)--write_(?:ok|err)--/ );
            ($write_err) = ( $op =~ /--write_err--(.*)--wfile/ );

            # Get write file
            $quote_mtpoint = quotemeta("$mountpoint/");
            ($wfile) = ( $op =~ /--wfile$quote_mtpoint(.*)--/ );

            $msg .= "READ  $tfname (open read only, no create)";
            check_error($msg, $read_err, $opts{read_fail});

            $msg .= "\nWRITE $wfile (open write only, create, truncate)";
            $write_fail_cond = ( $access eq 'ro' or $opts{write_fail} );
            check_error($msg, $write_err, $write_fail_cond);

            # Check proper read and write access
            if ( ($read_err xor $opts{read_fail})
              or ($write_err xor $write_fail_cond) ) {
                return 0; # Unsuccessful access to volume
            }
            if ( $access eq 'rw' ) {
                # Get uid and gid of the file just created
                ($uid, $gid, $size) =
                  ($op =~ /--uid(\d+)--gid(\d+)--size(\d+)--/);
                if ( not defined($expect_uid) and not defined($expect_gid) ) {
                    # Check proper root access
                    if ( $opts{root} and $uid != 0 ) {
                        $op = 0;
                        $msg .= "\nCould not create file as root";
                        if ( $opts{root_cfail} ) {
                            $msg .= ' as expected';
                            $op = 1;
                        }
                        $msg .= ": file uid is <$uid>";
                        return $op;
                    }
                } elsif ( (defined($expect_uid) and $expect_uid != $uid)
                       or (defined($expect_gid) and $expect_gid != $gid) ) {
                    # Check proper file creation
                    $msg .= "\nExpected credentials <$expect_uid, $expect_gid>".
                            " are not equal to file credentials <$uid, $gid>";
                    return 0;
                }
            }
            return 1;     # Successful access to volume
        } else {
            # Check if mount is expected to fail
            $op = 0;
            if ( $opts{fail} ) {
                $msg .= ' as expected';
                $op = 1;
            }

            $msg .= ": $@" if $@;
            return $op;
        }
    };

    # Save error
    $err = $@;

    if ( $rslt ) {
        $msg .= "\nSuccessful access to volume";
    } else {
        $msg .= "\nUnsuccessful access to volume";
    }

    logcomment("check_export_access: $msg") if $msg;

    # Unmount volume
    undef $@;
    umount($mountpoint, client=>$cc) if $mountpoint;
    croak("$err $@") if ($err or $@);
    return $rslt;
}

#---------------------------------------------------------------------
## @name ::check_export_access_list
## @description
##  Checks clients for proper access to an export.
##  Iterates over a list of clients and checks for
##  proper access using check_export_access().
##
## @synopsis
##  $op = check_export_access_list(%setup_h, @clnts);
##
## @arg %setup_h
##  Required: Hash of common variables initialized at setup.
##
## @arg @clnts
##  Required: List of client hashes to check access for.
##
## @return
##  Success if all clients have proper access.
##
## @example @clnts = ( { client=>$cc1, access=>'rw' },
##                     { client=>$cc2, access=>'ro', fail=>1 } );
##  $op = check_export_access_list(%setup_h, @clnts);
##  Returns success if client ($cc1) is able to read and write,
##  and client ($cc2) is denied read.
##
## @notes
##  Structure of each item in @clnts includes
##
##  { client     => Client's connect object,
##
##    access     => Access to check for this client,
##
##    %options }
##
##  Where %options are optional named parameters for check_export_access().
#---------------------------------------------------------------------
sub check_export_access_list(\%\@;) {
    my ($setup_h, $accs) = validate_pos(
        @_,
        { type => HASHREF },
        { type => ARRAYREF }
    );
    my ($acc, $op);

    # Validate arguments:
    foreach $acc (@$accs) {
        if ( !defined($acc->{client}) or !defined($acc->{access}) ) {
            croak "check_export_access_list: validating arguments";
        }
    }

    # For every item in the list check for proper access,
    foreach $acc (@$accs) {
        # Create hash of options to pass to check_export_access,
        # only the options that are defined are added to the hash.
        my %options = %$acc;
        delete $options{client};
        delete $options{access};
        delete $options{$_}
          foreach ( grep { !defined($options{$_}) } keys %options );

        # Check client access on the export
        $op = check_export_access( %$setup_h,
                                   $acc->{client},
                                   $acc->{access},
                                   %options );

        # Fail on first occurrence of a failure
        last if not $op;
    }

    return $op;
}

#---------------------------------------------------------------------
## @name ::check_host_in_netgroup
## @description
##  Check to see if the given host is in the the given netgroup
##  on the filer.
##
## @synopsis
##  $op = check_host_in_netgroup($fc, $netgroup, $hostname);
##
## @arg $fc
##  Required: Filer's connection object.
##
## @arg $netgroup
##  Required: Netgroup to check against.
##
## @arg $hostname
##  Required: Hostname to check.
##
## @return
##  1 if client in netgroup 0 if not.
#---------------------------------------------------------------------
sub check_host_in_netgroup($$$;) {
    my ($fc, $netgroup, $hostname) = validate_pos(
        @_,
        { type => OBJECT },
        { type => SCALAR },
        { type => SCALAR }
    );
    my $output;

    # Check if host belongs to netgroup
    $output = run_getXXbyYY($fc, "netgrp $netgroup", $hostname);
    if ( $output !~ /not in netgroup|returned failure/ ) {
        return 1;
    } else {
        return 0;
    }
}

#---------------------------------------------------------------------
## @name ::get_dns_domain
## @description
##  Gets dns domain name for given client.
##
## @synopsis
##  $domain = get_dns_domain($fc, $ipaddr);
##
## @arg $fc
##  Required: Filer's connection object.
##
## @arg $ipaddr
##  Required: Client's ip address to get dns domain for.
##
## @return
##  Client's dns domain.
#---------------------------------------------------------------------
sub get_dns_domain($$;) {
    my ($fc, $ipaddr) = validate_pos(
        @_,
        { type => OBJECT },
        { type => SCALAR },
    );
    my ($fqdn, $name, $domain, $op);

    # Get the FQDN of ip address
    $op = run_getXXbyYY($fc, "gethostbyaddr_r", $ipaddr);
    croak $op if ( $op =~ /Host not found/ );

    # Get the FQDN
    ($fqdn) = ($op =~ /^name:\s+(.*)\n/);

    # Get the domain name
    ($name, $domain) = ($fqdn =~ /^([^.]+)\.(.*)/);
    croak "DNS domain name not found for $ipaddr" if ! $domain;

    return $domain;
}

#---------------------------------------------------------------------
## @name ::add_result
## @description
##  Adds a result to a list of results.
##
## @synopsis
##  $rslt_obj = add_result($robj, $msg);
##
## @arg $robj
##  Result object to append to.  This could be either a NATE::Result
##  object, an array reference of messages, or a single message.
##
## @arg $msg
##  Result object to append to the list.  This could be either a
##  NATE::Result object, an array reference of messages,
##  or a single message.
##
## @return
##  Returns a NATE::Result object
##
## @example
##  $new_rslt_obj = add_result($rslt_obj, "New message");
##  $new_rslt_obj = add_result($rslt_obj, [$msg1, $msg2]);
##  $new_rslt_obj = add_result($rslt_obj, $rslt_obj1);
##  $new_rslt_obj = add_result($msg, $rslt_obj1);
#---------------------------------------------------------------------
sub add_result($$;) {
    my ($rslt_obj, $msg) = validate_pos(
        @_,
        { type => SCALAR | ARRAYREF | OBJECT | UNDEF },
        { type => SCALAR | ARRAYREF | OBJECT | UNDEF },
    );
    my @msgs;

    # Return undef if both arguments are empty
    return if (!$rslt_obj and !$msg);

    if ( ref($msg) and UNIVERSAL::isa($msg, 'NATE::Result') ) {
        # NATE::Result object, get messages from object
        @msgs = $msg->messages();
    } elsif ( ref($msg) eq 'ARRAY' ) {
        # Array reference, save messages
        @msgs = @$msg;
    } elsif ( $msg ) {
        # Single message, save it
        @msgs = ($msg);
    }

    if ( ref($rslt_obj) and UNIVERSAL::isa($rslt_obj, 'NATE::Result') ) {
        # NATE::Result object, Append messages to object
        return $rslt_obj if ! @msgs;
        $rslt_obj->messages([$rslt_obj->messages(), @msgs]);
    } elsif ( ref($rslt_obj) eq 'ARRAY' ) {
        # Array Reference, append messages and create object
        @msgs = (@$rslt_obj, @msgs) if $rslt_obj;
        $rslt_obj = fatal_result(\@msgs);
    } else {
        # Single message, append messages and create object
        @msgs = ($rslt_obj, @msgs) if $rslt_obj;
        $rslt_obj = fatal_result(\@msgs);
    }

    return $rslt_obj;
}

#---------------------------------------------------------------------
## @name ::compare_access_cache
## @description
##  Compares the access cache with a previous reading.
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = compare_access_cache($fc, %cache [, %options]);
##
## @arg $fc
##  Required: Filer's connection object.
##
## @arg $cache
##  Required: Hash of previous access cache reading.
##
## @arg hit
##  Optional: True if expecting a hit/miss, defaults to 0.
##
## @arg cold
##  Optional: True if checking a cold access cache, defaults to 0.
##
## @return
##  Current access cache, else croaks if expected behavior is not met.
#---------------------------------------------------------------------
sub compare_access_cache($\%;%) {
    my ($fc, $old_cache, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT  },
            { type => HASHREF },
        ],
        {
            hit  => { type => BOOLEAN, default => 0 },
            cold => { type => BOOLEAN, default => 0 },
        }
    );
    my (%new_cache, $hits, $miss);
    my $flag = $opts{hit};
    my $cold = $opts{cold};

    # Get current access cache
    %new_cache = filer_access_cache_get($fc);

    # Get hits and misses for current operation
    $hits =  $new_cache{hits} - $old_cache->{hits};
    $miss =  $new_cache{misses} - $old_cache->{misses};

    # True if expected behavior is met for hits.
    $hits =   !($hits xor $flag);
    # True if expected behavior is met for miss.
    $miss = ((!($miss xor $flag) and $cold) or (!$miss and !$cold));

    if ( !( $hits and $miss ) ) {
        # Return correct error messages
        my ($msg, $not);
        if ( ! $hits ) {
            $not = " not" if $flag;
            $msg = "There's$not been a cache hit.  ";
        }
        if ( ! $miss ) {
            $not  = " not" if ($flag and $cold);
            $msg .= "There's$not been a cache miss.";
        }
        croak $msg;
    }

    return %new_cache;
}

#---------------------------------------------------------------------
## @name ::check_access_cache
## @description
##  Function to check for specific access cache behavior.
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = check_access_cache(%setup_h, @accs [, %options]);
##
## @arg %setup_h
##  Required: Structure of common variables initialized by setup.
##
## @arg @accs
##  Required: List of client hashes to check cache access for.
##
## @arg cold
##  Optional: Check a cold access cache, defaults to 0.
##
## @arg timeout
##  Optional: Time in seconds to wait for the access cache to become cold.
##  If this is 0, there is no wait.
##  Otherwise, it caches the client then waits before checking the
##  access cache, defaults to 0.
##
## @return
##  1 on success
##
## @notes
##  Structure of each item in @accs
##
##  { cc         => Client's connect object,
##
##    access     => Access to check for this client,
##
##    mountd     => Expecting a hit/miss in a mountd request,
##
##    nfsdu      => Expecting a hit/miss in a nfsd user request,
##
##    nfsdr      => Expecting a hit/miss in a nfsd root request,
##
##    cache      => Expecting a hit/miss in all requests.
##                     This sets mountd, nfsdu, and nfsdr to this value,
##
##    mount_fail => Access should be denied
##  }
#---------------------------------------------------------------------
sub check_access_cache(\%\@;%) {
    my ($set_h, $accs, %opts) = validate_mixed (
        @_,
        [
            { type => HASHREF },
            { type => ARRAYREF },
        ],
        {
            cold    => { type => BOOLEAN,        default => 0 },
            timeout => { type => SCALAR | UNDEF, default => 0 },
        }
    );
    my (%access_h, %old_cache, $mountpoint, $acc, $op, $err);
    my ($client_name, $client_ip, $read_file, $write_file);
    my $fc         = $set_h->{fc};
    my $volume     = $set_h->{volume};
    my $tfname     = $set_h->{tfname};
    my $tfcontents = $set_h->{tfcontents};
    my $cold       = $opts{cold};
    my $timeout    = $opts{timeout};

    foreach $acc (@$accs) {
        # Get client's hostname and ip address as seen by the filer
        $client_name = $acc->{cc}->hostp()->hostname();
        $client_ip   = $set_h->{client_ip_h}{$client_name};

        # Process arguments:
        # 'cache' overrides (mountd, nfsdu, nfsdr).
        if ( $acc->{cache} ) {
            $acc->{mountd} = $acc->{cache};
            $acc->{nfsdu}  = $acc->{cache};
            $acc->{nfsdr}  = $acc->{cache};
        }

        # Flush access cache for mountd request
        filer_access_cache_flush($fc) if $cold;

        # Get access cache
        %old_cache = filer_access_cache_get($fc) if not %old_cache;

        if ( $timeout > 0 ) {
            # Cache client for mountd
            filer_client_cache($fc, $client_ip, $volume, 'mount');
            # Wait until cache has expired
            sleep($timeout);
            # Check as cache being cold.
            $cold = 1;
        }

        # Mount volume
        $mountpoint = mount($fc->connhost(), $volume, client=>$acc->{cc},
                            mntopts=>"vers=$main::NFS_VERSION");
        if ( ! $mountpoint and ! $acc->{mount_fail} ) {
            # Mount was not expected to fail
            croak "mount: Couldn't mount volume on $client_name: $@";
        }

        # Catch any errors to properly unmount volume on client
        eval {
            if ( $mountpoint and $acc->{mount_fail} ) {
                # Volume mounted where it should have failed
                croak "Access should have been denied for $client_name";
            }

            # Mountd request
            %old_cache = compare_access_cache( $fc, %old_cache,
                                               hit  => $acc->{mountd},
                                               cold => $cold );

            # Access should have been denied, check is done
            return if $acc->{mount_fail};

            # Initialize variables for this iteration
            $read_file   = "$mountpoint/$tfname";
            $write_file  = "$mountpoint/$tfname.$client_name";

            # Flush access cache for nfsd request as normal user
            filer_access_cache_flush($fc) if $cold;

            if ( $timeout > 0 ) {
                # Cache client for ro/rw
                filer_client_cache($fc, $client_ip, $volume, 'rw');
                # Wait until cache has expired
                sleep($timeout);
            }

            # Read a file as a normal user
            $op = client_read_file($acc->{cc}, $read_file);
            if ( $op !~ /$tfcontents/ ) {
                croak "Couldn't read file: $read_file";
            }

            # Write a file as a normal user
            if ( $acc->{access} eq 'rw' ) {
                client_write_file($acc->{cc}, $write_file, $tfcontents);
            }

            # Nfsd request for normal user
            %old_cache = compare_access_cache( $fc, %old_cache,
                                               hit  => $acc->{nfsdu},
                                               cold => $cold );

            # Flush access cache for nfsd request as root
            filer_access_cache_flush($fc) if $cold;

            if ( $timeout > 0 ) {
                # Cache client for root
                filer_client_cache($fc, $client_ip, $volume, 'root');
                # Wait until cache has expired
                sleep($timeout);
            }

            # Ignore errors from read/write, the access cache hit or miss
            # is the only thing of relevance
            eval {
                # Read a file as root
                $op = client_read_file($acc->{cc}, "$read_file.root", root=>1);

                # Write a file as root
                if ( $acc->{access} eq 'rw' ) {
                    client_write_file( $acc->{cc},
                                       "$write_file.root",
                                       $tfcontents,
                                       root=>1 );
                }
            };

            # Nfsd request for root
            %old_cache = compare_access_cache( $fc, %old_cache,
                                               hit  => $acc->{nfsdr},
                                               cold => $cold );
        };
        # Save error message from eval
        $err = $@;

        # Umount volume
        undef $@;
        umount($mountpoint, client=>$acc->{cc}) if $mountpoint;
        croak "$err $@" if ($err or $@);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::basic_access_test
## @description
##  Basic exports test for read-only, read-write and root with or
##  without an access lists.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = basic_access_test( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg format
##  Required: Access option format, where:
##  $h1 is the primary client;
##  $h2 is the list of other clients used for positive testing,
##    this always starts with a ':';
##  $h3 is the list of clients used for negative testing.
##  (e.g. format => 'ro=$h1$h2')
##
## @arg accs
##  Required: Reference to a list of hashes of expected behavior
##  as needed by the check_export_access_list() function.
##
## @arg negate
##  Optional: If true, all host entries in $h2 will be negated.
##  (default is 0).
##
## @return
##  1 if succeeded.
#---------------------------------------------------------------------
sub basic_access_test(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h  => { type => HASHREF },
            accs   => { type => ARRAYREF },
            format => { type => SCALAR },
            negate => { type => BOOLEAN, default => 0 },
        }
    );
    my ($access, %access_h, $err, $op, $h1, $h2, $h3);
    my $fc            =   $opts{set_h}->{fc};
    my $volume        =   $opts{set_h}->{volume};
    my %clients_h     = %{$opts{set_h}->{clients_h}};
    my @host_entries  = @{$opts{set_h}->{host_entries}};
    my @nclient_names = @{$opts{set_h}->{nclient_names}};

    logStep("BEGIN: basic_access_test");

    foreach $h1 (@host_entries) {
        logStep("Add clients to access list");
        $h2 = $clients_h{$h1};
        $h3 = join(':', @nclient_names);
        # Negate rest of clients for positive testing
        $h2 =~ s/:/:-/g if $opts{negate};

        # Eval the format to get the actual access list to export
        $access = eval qq("$opts{format}");

        logStep("Export volume: $volume -$access");
        export_path($fc, $volume, OPTIONS=>$access) or croak($@);

        # Catch any errors to properly unexport the path
        eval {
            logStep("Get exports table");
            %access_h = get_export_options($fc, $volume);

            logStep("Validate subnets and convert them to a standard format");
            $access = fix_subnets($access);
            %access_h = map { fix_subnets($_) } %access_h;

            logStep("Validate exports table");
            # Put the hash of access into a string format similar
            # to the one sent to export_path. Keep the same order
            # of options to make the comparison a lot simpler.
            $op = join(',', map { $access_h{(split /=/)[0]} }
                                  split(/,/, $access));
            croak("Unable to export path correctly: $volume -$access")
              if ( $op ne $access );

            logStep("Check proper access on clients");
            $op = check_export_access_list(%{$opts{set_h}}, @{$opts{accs}});
            croak("Clients do not have proper access") if !$op;
        };

        $err = $@;

        logStep("Unexport volume");
        unexport_path($fc, $volume, EXISTENT=>0);
        croak "$err $@" if ($err or $@);

        # If the access doesn't have a list of hosts
        # do not loop through all host entries
        last if $access =~ qr/^r[ow]$/;
    }

    logStep('END: basic_access_test');

    return 1;
}

#---------------------------------------------------------------------
## @name ::export_path_mount_and_read
## @description
##  Export the volume, mount it on client, and if specified read a file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = export_path_mount_and_read( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg path
##  Required: Path to export.
##
## @arg file
##  Optional: File name to read, (default is $set_h->{tfname}).
##
## @arg client
##  Optional: Connection handle for client to mount and read file,
##  (default is the primary client).
##
## @arg do_read
##  Optional: Boolean, true to read the file, (default is 1).
##
## @arg access
##  Optional: Access option of path to export, (default is 'ro').
##
## @return
##  Contents of the file read or 1 if the 'do_read' option is false.
#---------------------------------------------------------------------
sub export_path_mount_and_read(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h   => { type => HASHREF },
            path    => { type => SCALAR },
            file    => { type => SCALAR,  optional => 1 },
            client  => { type => OBJECT,  optional => 1 },
            do_read => { type => BOOLEAN,  default => 1 },
            access  => { type => SCALAR,   default => 'ro' },
        }
    );
    my (%access_h, $mountpoint, $contents, $err, $op);
    my $fc        = $opts{set_h}->{fc};
    my $vol_name  = $opts{set_h}->{vol_name};
    my $test_path = $opts{path};
    my $access    = $opts{access};
    my $cc        = defined($opts{client}) ? $opts{client}
                                           : $opts{set_h}->{cc}[0];
    my $file      = defined($opts{file})   ? $opts{file}
                                           : $opts{set_h}->{tfname};
    my $filer_ip  = $fc->connhost();
    my $mt_point  = "/t/$filer_ip/$vol_name";

    logStep("BEGIN: export_path_mount_and_read");

    logStep("Export test path");
    export_path($fc, $test_path, OPTIONS=>$access) or croak($@);

    # Catch any errors to properly unexport the path
    eval {
        logStep("Get exports table");
        %access_h = get_export_options($fc, $test_path);

        logStep("Validate exports table");
        # Put the hash of access into a string format similar
        # to the one sent to export_path. Keep the same order
        # of options to make the comparison a lot simpler.
        $op = join(',', map { $access_h{(split /=/)[0]} }
                              split(/,/, $access));
        croak "Unable to export path correctly: $test_path -$access"
          if ( $op ne $access );

        logStep("Mount test path on client");
        $mountpoint = mount( $filer_ip,
                             $test_path,
                             client     => $cc,
                             mntopts    => "vers=$main::NFS_VERSION",
                             sharing    => 'exclusive',
                             mountpoint => $mt_point )
                             or croak "Unable to mount path on client: $@";

        # Catch any errors to properly unmount the path
        eval {
            if ( $opts{do_read} ) {
                logStep("Read file on test path");
                $contents = client_read_file($cc, "$mountpoint/$file");
            }
        };

        $err = $@;

        logStep("Unmount test path");
        umount($mountpoint, client=>$cc);
        croak "$err $@" if ($err or $@);
    };

    $err = $@;

    logStep("Unexport test path");
    unexport_path($fc, $test_path, EXISTENT=>0);
    croak "$err $@" if ($err or $@);

    logStep("END: export_path_mount_and_read");

    return $opts{do_read} ? $contents : 1;
}

#---------------------------------------------------------------------
## @name ::anon_properties_test
## @description
##  It exports the volume with the access options provided
##  and checks whether or not the export should have failed.
##  Then it verifies that the read and write (if 'rw' access)
##  operations are performed with the correct credentials.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = anon_properties_test( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg access
##  Required: Access options to export volume.
##
## @arg file
##  Optional: File for testing read access,
##  (default is the file created on setup).
##
## @arg root
##  Optional: True if operation is to be performed by root, (default is 0).
##
## @arg check_access
##  Optional: Check for specific access, could be either 'ro', 'rw',
##  'modify', 'remove', 'execute'.  If this option is not specified,
##  it is derived from the 'access' option.
##
## @arg mount_opts
##  Optional: Mount options.
##
## @arg expect_uid
##  Optional: Expected user id of file to be created,
##  (default is the uid of user creating the file).
##
## @arg expect_gid
##  Optional: Expected group id of file to be created,
##  (default is the gid of user creating the file).
##
## @arg check_ruid
##  Optional: Boolean, check that the file is owned by root
##  (default is 1). This check is only done if the 'root'
##  option is set and 'check_access' is set to 'modify'.
##
## @arg expect_export_fail
##  Optional: True if export command should fail, (default is 0).
##
## @arg expect_access_fail
##  Optional: True if access should be denied, (default is 0).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub anon_properties_test(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h              => { type => HASHREF },
            access             => { type => SCALAR },
            file               => { type => SCALAR, optional => 1 },
            root               => { type => SCALAR,  default => 0 },
            mount_opts         => { type => SCALAR, optional => 1 },
            expect_uid         => { type => SCALAR, optional => 1 },
            expect_gid         => { type => SCALAR, optional => 1 },
            check_access       => { type => SCALAR, optional => 1 },
            check_ruid         => { type => SCALAR,  default => 1 },
            expect_export_fail => { type => SCALAR,  default => 0 },
            expect_access_fail => { type => SCALAR,  default => 0 },
        }
    );
    my (@accs, %set_h, %access_h, $err, $msg, $not, $op, %options);
    my $access      = $opts{access};
    my $export_fail = $opts{expect_export_fail} ? 1 : 0;
    my $access_fail = $opts{expect_access_fail} ? 1 : 0;
    my $fc          = $opts{set_h}{fc};
    my $volume      = $opts{set_h}{volume};

    logStep('BEGIN: anon_properties_test');

    # Catch any errors to properly unexport the volume
    eval {
        logStep('Export volume');
        $op = export_path($fc, $volume, OPTIONS=>$access);
        if ( !($op xor $export_fail) ) {
            $not = ' not' if not $export_fail;
            $msg = "Export should$not have failed";
            if ( $export_fail ) {
                # Export was expected to fail, but somehow succeeded
                croak $msg;
            } else {
                # Export failed where it should not have failed
                croak "$msg: $@";
            }
        }
        if ( $export_fail ) {
            # Export is expected to fail
            return;
        }

        logStep("Get exports table");
        %access_h = get_export_options($fc, $volume);

        logStep("Validate exports table");
        # Remove anon=65534 since this is the default and it is
        # not listed on the access options for the volume
        $access =~ s/,anon=65534//;
        # Put the hash of access into a string format similar
        # to the one sent to export_path. Keep the same order
        # of options to make the comparison a lot simpler.
        $op = join(',', map { $access_h{(split /=/)[0]} }
                              split(/,/, $access));
        croak "Unable to export path correctly: $volume -$access"
          if ( $op ne $access );

        logStep('Verify proper credentials');
        # Check proper access to volume: defaults to 'rw'
        # unless access has 'ro' flag.
        # Also, check the access is done with
        # proper credentials
        $access = ( $opts{check_access} ? $opts{check_access}
                                        : ($access =~ /\bro\b/) ? 'ro' : 'rw' );
        if ( $access =~ /^(r[ow]|modify)$/ ) {
            %options = (read_fail => $access_fail, write_fail => $access_fail);
        } elsif ( $access eq 'execute' ) {
            %options = (exec_fail => $access_fail);
        } elsif ( $access eq 'remove' ) {
            %options = (rm_fail => $access_fail);
        }
        @accs = map {
                      { client     => $_,
                        access     => $access,
                        root       => $opts{root} ? 1 : 0,
                        check_ruid => $opts{check_ruid},
                        mount_opts => $opts{mount_opts},
                        expect_uid => $opts{expect_uid},
                        expect_gid => $opts{expect_gid},
                        %options
                      }
                    } @{$opts{set_h}{cc}};
        # Copy main hash to modify file to test with
        %set_h = %{$opts{set_h}};
        $set_h{tfname} = $opts{file} if $opts{file};
        if ( ! check_export_access_list(%set_h, @accs) ) {
            my $root = ' root' if $opts{root};
            croak "The$root user was not properly mapped to the expected uid";
        }
    };

    $err = $@;

    logStep('Unexport volume');
    unexport_path($fc, $volume);
    croak "$err $@" if ($err or $@);

    logStep('END: anon_properties_test');

    return 1;
}

#---------------------------------------------------------------------
## @name ::nosuid_option_test
## @description
##  This function exports the test volume with access option given
##  by the 'access' parameter, and mounts it on the specified client.
##  Then, it creates a special device file or set the special bits
##  of a file.
##
##  Options device_type and set_mode are mutually exclusive, one is to
##  test the creation of a special device file and the other is to test
##  setting of special bits for the file, respectively.
##  Also, major_number and minor_number need only to be defined for
##  creating block (b) or character (c) special devices.
##
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = nosuid_option_test( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg access
##  Required: Access options to export volume.
##
## @arg client
##  Optional: Connection handle for client. (default is the primary client).
##
## @arg device_type
##  Optional: Device type to create: 'b' for block, 'c' for character,
##  'p' for fifo or 's' for named socket.
##
## @arg major_number
##  Optional: Major number for block or character special device.
##
## @arg minor_number
##  Optional: Minor number for block or character special device.
##
## @arg set_mode
##  Optional: Permissions set on file to test setting of setuid,
##  setguid and sticky bit on file
##
## @arg expect_fail
##  Optional: True if operation should fail, (default is 0).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub nosuid_option_test(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h        => { type => HASHREF },
            access       => { type => SCALAR },
            client       => { type => OBJECT,  optional => 1 },
            device_type  => { type => SCALAR,  optional => 1 },
            major_number => { type => SCALAR,  optional => 1 },
            minor_number => { type => SCALAR,  optional => 1 },
            expect_fail  => { type => BOOLEAN,  default => 0 },
            set_mode     => { type => SCALAR,  optional => 1 },
        }
    );
    my ($mountpoint, %access_h, %major_minor, $mode, $err, $op);
    my $rno        = int(rand(100000));
    my $access     = $opts{access};
    my $fail       = $opts{expect_fail};
    my $dtype      = $opts{device_type};
    my $set_mode   = $opts{set_mode};
    my $not        = ' not' if !$fail;
    my $fc         = $opts{set_h}{fc};
    my $tfname     = $opts{set_h}{tfname};
    my $tfcontents = $opts{set_h}{tfcontents};
    my $volume     = $opts{set_h}{volume};
    my $cc         = defined($opts{client}) ? $opts{client}
                                            : $opts{set_h}->{cc}[0];

    logStep('BEGIN: nosuid_option_test');

    # Check input parameters
    if ( $dtype and $set_mode ) {
        croak "Options 'device_type' and 'set_mode' are mutually exclusive";
    } elsif ( ! $dtype and ! $set_mode ) {
        croak "Either 'device_type' or 'set_mode' must be defined";
    }

    logStep('Export volume');
    export_path($fc, $volume, OPTIONS=>$access) or croak $@;

    # Catch any errors to properly unexport the volume
    eval {
        logStep("Get exports table");
        %access_h = get_export_options($fc, $volume);

        logStep("Validate exports table");
        # Put the hash of access into a string format similar
        # to the one sent to export_path. Keep the same order
        # of options to make the comparison a lot simpler.
        $op = join(',', map { $access_h{(split /=/)[0]} }
                              split(/,/, $access));
        croak "Unable to export path correctly: $volume -$access"
          if ( $op ne $access );

        logStep('Mount volume on client');
        $mountpoint = mount($fc->connhost(), $volume, client=>$cc,
                            mntopts=>"vers=$main::NFS_VERSION") or croak $@;

        # Catch any errors to properly unmount the volume
        eval {
            if ( $dtype ) {
                logStep('Creation of special files');
                my $special_file;
                my $dfile = "$mountpoint/$tfname.${dtype}dev.$rno";

                # Set correct type of file for error message
                if ( $dtype eq 'b' ) {
                    $special_file = 'block special device';
                } elsif ( $dtype eq 'c' ) {
                    $special_file = 'character special device';
                } elsif ( $dtype eq 'p' ) {
                    $special_file = 'fifo special file';
                }

                # Add major and minor numbers options only if defined
                $major_minor{major} = $opts{major_number}
                  if defined($opts{major_number});
                $major_minor{minor} = $opts{minor_number}
                  if defined($opts{minor_number});

                # Make special device file on client
                $op = eval {
                    client_mknod($cc, $dfile, type => $dtype, %major_minor,
                                              root => 1);
                };
                if ( ! $op xor $fail ) {
                    croak "Creation of a $special_file should" .
                          "$not have failed: $@";
                }
            } elsif ( $set_mode ) {
                logStep('Setting setuid, setgid and sticky bit');
                my $special_op;

                # Set correct operation for error message
                if ( $set_mode & 04000 ) {
                    $special_op = 'setuid';
                } elsif ( $set_mode & 02000 ) {
                    $special_op = 'setgid';
                } elsif ( $set_mode & 01000 ) {
                    $special_op = 'sticky bit';
                }

                # Create file and change mode to set special bit
                ($mode) = ( client_write_file($cc,
                                              "$mountpoint/$tfname.w.$rno",
                                              $tfcontents,
                                              root       => 1,
                                              mode       => $set_mode,
                                              check_mode => 0) )[2];
                if ( ($mode & 07777) != $set_mode xor $fail ) {
                    croak "An exclusive $special_op operation " .
                          "should$not have failed";
                }
            }
        };

        $err = $@;

        logStep('Unmount volume on client');
        umount($mountpoint, client=>$cc);
        croak "$err $@" if ($err or $@);
    };

    $err = $@;

    logStep('Unexport volume');
    unexport_path($fc, $volume);
    croak "$err $@" if ($err or $@);

    logStep('END: nosuid_option_test');

    return 1;
}

#---------------------------------------------------------------------
## @name ::access_cache_test
## @description
##  This function exports the volume with the access option given
##  as an argument and checks for proper access cache behavior.
##  It tests both cold and hot access cache.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = access_cache_test( %options );
##
## @arg set_h
##  Required: Structure of common variables initialized by setup.
##
## @arg access
##  Required: Access option to export volume.
##
## @arg accs
##  Required: List of client hashes to check cache access for.
##
## @arg expire_timeout
##  Optional: Time in seconds to wait for the access cache to become cold.
##  If this is 0, there is no wait.
##  Otherwise, it caches the client then waits before checking the
##  access cache, (default is 0).
##
## @return
##  1 if succeeds
#---------------------------------------------------------------------
sub access_cache_test(%;) {
    my (%opts) = validate (
        @_,
        {
            set_h          => { type => HASHREF },
            access         => { type => SCALAR },
            accs           => { type => ARRAYREF },
            expire_timeout => { type => SCALAR, default => 0 }
        }
    );
    my (%access_h, $err, $op);
    my $fc        =   $opts{set_h}{fc};
    my $volume    =   $opts{set_h}{volume};
    my $access    =   $opts{access};
    my @accs      = @{$opts{accs}};
    my @client_ip = @{$opts{set_h}{client_ip}};
    my $timeout   =   $opts{expire_timeout};

    logStep('BEGIN access_cache_test');

    logStep("Export volume");
    export_path($fc, $volume, OPTIONS=>$access) or croak $@;

    # Catch any errors to properly unexport the volume
    eval {
        logStep("Get exports table");
        %access_h = get_export_options($fc, $volume);

        logStep("Validate exports table");
        # Put the hash of access into a string format similar
        # to the one sent to export_path. Keep the same order
        # of options to make the comparison a lot simpler.
        $op = join(',', map { $access_h{(split /=/)[0]} } split(/,/, $access));
        if ( $op ne $access ) {
            croak "Unable to export path correctly: $volume -$access";
        }

        logStep("Test cold access cache");
        check_access_cache(%{$opts{set_h}}, @accs, cold=>1, timeout=>$timeout);

        if ( ! $timeout ) {
            # Cache all clients
            foreach my $clnt_ip (@client_ip) {
                logStep("Cache client for mountd, ro, and rw");
                filer_client_cache($fc, $clnt_ip, $volume, 'mount');

                logStep("Cache client for root");
                filer_client_cache($fc, $clnt_ip, $volume, 'root');
            }

            logStep("Test hot access cache");
            check_access_cache(%{$opts{set_h}}, @accs);
        }
    };

    $err = $@;

    logStep("Unexport volume");
    unexport_path($fc, $volume, EXISTENT=>0);
    croak "$err $@" if ($err or $@);

    logStep("END access_cache_test");

    return 1;
}

#---------------------------------------------------------------------
## @name ::is_atnetgroup_supported_on_filer
## @description
##  This function checks if the specified filer supports the '@'
##  character on netgroups.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = is_atnetgroup_supported_on_filer( %options );
##
## @arg filer
##  Required: Hostspec of filer.
##
## @return
##  1 if '@' is supported on filer else croaks with error message.
#---------------------------------------------------------------------
sub is_atnetgroup_supported_on_filer(%;) {
    my (%opts) = validate (
        @_,
        {
            filer => { type => SCALAR },
        }
    );

    # Check correct version of filer
    my $fh = host($opts{filer});
    if ( get_wafl_version($fh) < WV_ANCHORSTEAM_MIN ) {
        # Pre-Anchorsteam releases do not support the @ sign on netgroups
        croak(UNSUPPORTED_FEATURE_EXCEPTION .
              ": Pre-Anchorsteam releases do not " .
              "support the '\@' on netgroups");
    }

    return 1;
}

1;

