# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Flg.pm#1 $
#
## @summary     Flg - Filesystem load generator
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Flg - Filesystem load generator, based on original C code

=head1 SYNOPSIS

  use nfs::Flg;

  flg( server => 'my_server.at.domain.com',
       path   => '/vol/testvol_flex',
       rate => 2000,
       duration => 300,
       ns_prefix => 'my_flg',
       version => 3,
       protocol => 'udp',
       timeout => 30 )

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Flg.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance, Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Flg;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Nameset;
use nfs::Contents qw(contents_lookup contents_new_write contents_fill_write
                     contents_new_read contents_verify_read);
use nfs::Operation;
use nfs::Timer qw(global_timer);

use Carp qw(croak);
use Carp::Assert;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF SCALARREF);
use nfs::NfsDef qw(:NFSCONSTANTS);
use nfs::NfsClient;
use Sys::Hostname;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
our @EXPORT_OK = qw(
  flg
);

use constant MAX_PAYLOAD_SIZE => 8192;
use constant LINGER_READ => 0;
use constant LINGER_SEQREAD => 1;
use constant LINGER_WRITE => 2;
use constant LINGER_SEQWRITE => 3;
use constant LINGER_APPENDWRITE => 4;
use constant NUM_LINGER_TYPES => 5;

my %stats = ( call => 0,
              rexmit => 0,
              reply_success => 0,
              reply_error => 0,
              reply_cancel => 0, );
my $op_distdh;
my $rsize_distdh;
my $wsize_distdh;
my $nameset;
my $contents;
my $nfs_obj;
my @linger_list;
my $linger_maxdepth = 16;
my $linger_maxinuse = 8192;
my $linger_inuse = 0;
my $mount_retry = 10;

#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------
sub flg(%;);
sub print_statline(%;);
sub gen_func();
sub rsize_dist_func();
sub wsize_dist_func();
sub dist_init(%;);
sub gen_op(%;);
sub random_ftype();
sub random_ftype_prefer(%;);
sub nameset_ref(%;);
sub nameset_deref(%;);
sub nameset_getfname(%;);
sub nameset_getlinkname(%;);
sub do_create(%;);
sub do_lookup(%;);
sub do_mkdir(%;);
sub do_link(%;);
sub do_symlink(%;);
sub do_getattr(%;);
sub do_readdir(%;);
sub do_readdirplus(%;);
sub do_access(%;);
sub do_fsstat(%;);
sub do_readlink(%;);
sub do_rename(%;);
sub do_remove(%;);
sub do_rmdir(%;);
sub do_setattr(%;);
sub do_write(%;);
sub do_read(%;);
sub do_commit(%;);
sub linger_rename(%;);
sub linger_dele_notify(%;);
sub linger_select(%;);
sub linger_iodone(%;);
sub measure_op_call();
sub measure_op_reply(%;);
sub statrec_add(%;);
sub statrec_reset(%;);
sub measure_op_printstats();
sub statrec_avg(%;);
sub statrec_stddev(%;);
sub measure_op_contrib(%;);
sub measure_op_global_avg();
sub measure_op_called();
sub measure_op_achieved();
sub testme();

#--------------------------------------------------------------------------------

=head2 flg

  flg( server => 'my_server.at.domain.com',
       path   => '/vol/testvol_flex',
       rate => 2000,
       duration => 300,
       ns_prefix => 'my_flg',
       version => 3,
       protocol => 'udp',
       timeout => 30 )

Runs the flg filesystem load generator, based on the original flg C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Filer to run flg against

=item path

(Required) : Path on filer where files will be maniuplated

=item rate

(Optional) : Requested number of ops/sec. Defaults to 1000.

=item duration

(Optional) : Length of execution. Defaults to 600.

=item ns_prefix

(Optional) : Prefix for file/directory creation. Defaults to process pid.

=item version

(Optional) : NFS version to use for operations. Acceptable values are
2, 3, and 4. Defaults to 3.

=item protocol

(Optional) : Connection protocol to use for NFS requests. Acceptable
values are 'tcp' or 'udp'. Defaults to 'udp'.

=item retry

(Optional) : Number of times to retry each RPC call if response is not
received within the timeout period (default is 10).

=item timeout

(Optional) : Timeout for requests. Defaults to 30.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub flg(%;) {
    #  int
    #  main(int argc, char *argv[])
    #  {
    #   struct in_addr addr;
    #   unsigned char fhdata[64];
    #   struct fh rootfh = {0, fhdata};
    #   int fd, retval = 0;
    #
    #   srandom(time(NULL));
    #
    #   report_start(argc, argv);
    #   parse_args(argc, argv);
    #
    #  #ifdef IPPROTO_TUF
    #   if (strcmp(transp, "tuf") == 0) {
    #           tuf_client = 1;
    #           transp = "tcp";
    #   }
    #  #endif
    #
    #   if (dns_name2addr(server, &addr) != 0) {
    #           report_error(FATAL, "dns_name2addr(\"%s\", ...) failed",
    #                        server);
    #           return -1;
    #   }
    #   if (mount_getroot(addr, path, &rootfh) != 0) {
    #           report_error(FATAL, "mount_getroot failed");
    #           return -1;
    #   }
    #
    #   if (cont_file) {
    #           /*
    #            * Continuing from a previous run.  Nameset file
    #            * provided.  Load and initialize distributions.
    #            */
    #           if ((fd = open(cont_file, O_RDONLY)) < 0) {
    #                   report_perror(FATAL, "open(\"%s\")", cont_file);
    #                   return -1;
    #           }
    #           if (nameset_load(fd) < 0) {
    #                   report_error(FATAL, "nameset_load failed");
    #                   return -1;
    #           }
    #           close(fd);
    #
    #           if (invalfh) {
    #                   nameset_invalidate_fh();
    #                   nameset_setroot(rootfh.data, rootfh.len);
    #           }
    #   } else {
    #           unsigned long maxfiles;
    #           unsigned long maxdirs;
    #           unsigned long maxsymlinks;
    #           nameset_entry_t rootnse;
    #
    #           /*
    #            * No given nameset file.  Initialize fresh nameset
    #            * from distributions.
    #            */
    #
    #           /* XXO: Copied from flg_init -- are these reasonable?? */
    #           maxfiles = 100 + ((40 * rate) * 10);
    #           maxdirs = maxfiles / 5;
    #           maxsymlinks = maxfiles / 50;
    #
    #           if (nameset_init(ns_prefix, maxfiles, maxdirs,
    #                           maxsymlinks) < 0) {
    #                   report_error(FATAL, "nameset_init failed");
    #                   return -1;
    #           }
    #
    #           if (contents_init(maxfiles) < 0) {
    #                   report_error(FATAL, "contents_init failed");
    #                   return -1;
    #           }
    #
    #           /*
    #            * create a root name entry.
    #            */
    #           /* XXO: weight == 1 ok? */
    #           if ((rootnse = nameset_alloc(NULL, NFDIR, 1)) == NULL) {
    #                   report_error(FATAL, "nameset_alloc error");
    #                   return -1;
    #           }
    #           nameset_setfh(rootnse, rootfh.data, rootfh.len);
    #   }
    #
    #   if (op_init(addr, strcmp(transp, "udp") == 0 ?
    #               SOCK_DGRAM : SOCK_STREAM) < 0) {
    #           report_error(FATAL, "op_init failed");
    #           return -1;
    #   }
    #
    #   measure_op_resetstats();
    #   printf("generating load...\n");
    #   if (op_metronome(rate, duration, gen_func, NULL) < 0) {
    #           report_error(FATAL, "op_metronome failed");
    #           return -1;
    #   }
    #   if (op_barrier(0) < 0) {
    #           report_error(FATAL, "op_barrier failed");
    #           return -1;
    #   }
    #   report_flush();
    #   printf("done generating load.\n");
    #
    #   measure_op_printstats();
    #   print_statline("SUMMARY", duration);
    #
    #   if (measure_op_global_avg()/1000 >= latency_cutoff) {
    #           printf("latency (%0.1f msecs/op) exceeded maxlat (%d), dying.\n",
    #                  (float)measure_op_global_avg()/1000.0, latency_cutoff);
    #           retval = 1;
    #   }
    #
    #   if (op_uninit() < 0) {
    #           report_error(FATAL, "op_uninit failed");
    #           return -1;
    #   }
    #
    #   if (save_file) {
    #           if ((fd = open(save_file, O_RDWR|O_CREAT|O_TRUNC, 0644)) < 0) {
    #                   report_perror(FATAL, "open(\"%s\")", save_file);
    #                   return -1;
    #           }
    #           if (nameset_save(fd) < 0) {
    #                   report_error(FATAL, "nameset_save failed");
    #                   return -1;
    #           }
    #           close(fd);
    #           if (nameset_uninit() < 0) {
    #                   report_error(FATAL, "nameset_uninit failed");
    #                   return -1;
    #           }
    #   }
    #
    #   report_stop(rusage);
    #
    #   if (retval != 0) {
    #           report_error(FATAL, "main returning with error status %d",
    #                        retval);
    #           /* don't return here */
    #   }
    #   return retval;
    #  }
    #

    # Seed random number gen
    srand( time );

    # parse args
    my (%opts) = validate (
        @_,
        {
            server    => { type => SCALAR },
            path      => { type => SCALAR },
            # Default rate is 1000 ops/sec
            rate      => { type => SCALAR, default => 1000,
                                           regex   => qr/^\d+$/ },
            # Set duration to 10 minutes
            duration  => { type => SCALAR, default => 600,
                                           regex   => qr/^\d+$/ },
            ns_prefix => { type => SCALAR, default => hostname . "-" . $$ },

            version   => { type => SCALAR, default => 3,
                                           regex   => qr/^(2|3|4)$/ },
            protocol  => { type => SCALAR, default => 'tcp',
                                           regex   => qr/^(tcp|udp)$/ },
            retry     => { type => SCALAR, default => 10,
                                           regex   => qr/^\d+$/ },
            timeout   => { type => SCALAR, default => 30,
                                           regex   => qr/^\d+$/ },
            localaddr => { type => SCALAR, optional => 1 },
        }
    );

    # report start
    my ($msg, $retry);
    my $options = '';
    foreach ( keys %opts ) {
        $options .= "$_:$opts{$_} ";
    }
    my @time = localtime();
    my $time = sprintf( "%02d:%02d:%02d", $time[2], $time[1], $time[0] );
    print "[$time] START host '" . hostname . "' command line '$options'\n";

    my %nfs_opts;
    if (defined ($opts{localaddr})) {
        %nfs_opts = ( localaddr => $opts{localaddr} );
    }

    # Create the NFS client object
    $retry = $opts{retry};
    do {
        eval { $nfs_obj = nfs::NfsClient->new( hostname => $opts{server},
                                               version  => $opts{version},
                                               proto    => $opts{protocol},
                                               retry    => $opts{retry},
                                               timeout  => $opts{timeout},
                                               %nfs_opts ) };
        $msg = $@;
        sleep(1) if ($msg =~ /Operation now in progress/);
    } while ( --$retry > 0 && $msg =~ /Operation now in progress/ );
    croak $msg if $msg;

    $retry = $mount_retry;
    do {
        # Mount the path
        eval { $nfs_obj->mount(path => $opts{path}) };
        $msg = $@;
        sleep(1) if ($msg =~ /Operation now in progress/);
    } while ( --$retry > 0 && $msg =~ /Operation now in progress/ );
    croak $msg if $msg;

    # mount_getroot
    # Save the mount rootfh
    my $rootfh = { data => $nfs_obj->{mnt_fh},
                   len  => do { use bytes; length ($nfs_obj->{mnt_fh}) }, };

    # We'll establish the linger_list here, since there is not
    # really a constructor for it. It is simply created (in the C app)
    # when the linger component is included.

    #  struct linger_state {
    #          int iocnt;
    #
    #          contents_t ct;
    #          nameset_entry_t nse;
    #          u_int64_t off;
    #          u_int32_t cnt;
    #
    #          Q_ENTRY(linger_state) link;
    #  };
    #  Q_HEAD(NULL, linger_state) linger_list[NUM_LINGER_TYPES];
    #
    for ( my $t = 0; $t < NUM_LINGER_TYPES; $t++ ) {
        $linger_list[$t] = { head => undef,
                             tail => undef };
    }

    # Initialize nameset from distribution
    my $maxfiles    = 100 + ((40 * $opts{rate}) * 10);
    my $maxdirs     = $maxfiles / 5;
    my $maxsymlinks = $maxfiles / 50;

    # nameset_init
    $nameset = nfs::Nameset->nameset_init( prefix => $opts{ns_prefix},
                                           regmax => $maxfiles,
                                           dirmax => $maxdirs,
                                           lnkmax => $maxsymlinks );

    # contents_init
    $contents = nfs::Contents->contents_init( maxfiles => $maxfiles );

    # nameset_alloc
    my $rootnse = $nameset->nameset_alloc( parent => undef,
                                           type   => NFDIR(),
                                           weight => 1 );

    # nameset_setfh
    $nameset->nameset_setfh( nse    => $rootnse,
                             fhdata => $rootfh->{data},
                             fhlen  => $rootfh->{len} );

    # op_init
    my $op_obj = nfs::Operation->op_init( client_obj => $nfs_obj );

    # measure_op_resetstats
    # Stats have not been used, no need to reset

    print "\ngenerating load ...\n";

    # op_metronome
    $op_obj->op_metronome( rate     => $opts{rate},
                           duration => $opts{duration},
                           func     => \&gen_func,
                           arg      => undef );

    # op_barrier
    #     op_barrier waits for outstanding requests to finish
    #     All requests have finished, by nfs_client implementation

    print "done generating load.\n";

    # measure_op_printstats
    measure_op_printstats();

    # print_statline
    print_statline( desc     => 'SUMMARY',
                    seconds  => $opts{duration},
                    rate     => $opts{rate},
                    nclients => 0 );

    # report_stop
    @time = localtime();
    $time = sprintf( "%02d:%02d:%02d", $time[2], $time[1], $time[0] );
    print "[$time] STOP\n";

} # End flg

#--------------------------------------------------------------------------------

=head2 print_statline

  print_statline( desc     => 'SUMMARY',
                  seconds  => $duration,
                  rate     => $rate,
                  nclients => 0 );

Prints a summary line of desired nfsops, called ops, performed ops,
average op time, and number of clients.

The following named arguments are accepted:

=over

=item Arguments

=over

=item desc

(Required) : Description string

=item seconds

(Required) : Length of execution time

=item rate

(Required) : Desired nfs ops rate

=item nclients

(Required) : Number of clients used

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub print_statline(%;) {
    #  static void
    #  print_statline(char *desc, int seconds)
    #  {
    #          printf("%s: ", desc);
    #          printf("wanted %d nfsops, ", rate);
    #          printf("called %d nfsops, ", measure_op_called() / seconds);
    #          printf("got %d nfsops, ", measure_op_achieved() / seconds);
    #          printf("avg %0.2f ms ", (float)measure_op_global_avg()/1000.0);
    #          printf("nclients %d", nclients);
    #          printf("\n");
    #  }

    my (%opts) = validate (
        @_,
        {
            desc     => { type => SCALAR },
            seconds  => { type => SCALAR },
            rate     => { type => SCALAR },
            nclients => { type => SCALAR }
         }
    );

    printf("%s: ", $opts{desc});
    printf("wanted %d nfsops, ", $opts{rate});
    printf("called %d nfsops, ", measure_op_called() / $opts{seconds});
    printf("got %d nfsops, ", measure_op_achieved() / $opts{seconds});
    printf("avg %0.2f ms ", measure_op_global_avg() / 1000.0);
    printf("nclients %d", $opts{nclients});
    printf("\n");

} # End print_statline

#--------------------------------------------------------------------------------

=head2 gen_func

  gen_func();

Generates a function distribution

=over

=item Returns

Nothing.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub gen_func() {
    #  static void
    #  gen_func(void *arg)
    #  {
    #          if (gen_op((*op_dist_func)()) < 0) {
    #                  fprintf(stderr, "gen_op failed\n");
    #                  return;
    #          }
    #  }

    # op_dist_func is op_dist by default
    # op_dist is created by CREATE_DIST:

    #  #define CREATE_DIST( _name ) \
    #  static distheap_t _name ## dh; \
    #  static struct distitem _name ## mix[]; \
    #  int _name (void) { \
    #          if ( _name ## dh == NULL ) { \
    #                  _name ## dh = dist_init( _name ## mix ); \
    #          } \
    #          return *(int *)distheap_select( _name ## dh ); \
    #  } \
    #  static struct distitem _name ## mix[] =
    #

    # Replacing _name with op_dist and merging tokens (##):

    #  #define CREATE_DIST( op_dist ) \
    #  static distheap_t op_distdh; \
    #  static struct distitem op_distmix[]; \
    #  int op_dist (void) { \
    #          if ( op_distdh == NULL ) { \
    #                  op_distdh = dist_init(op_distmix); \
    #          } \
    #          return *(int *)distheap_select( op_distdh ); \
    #  } \
    #  static struct distitem op_distmix[] =
    #

    # Here's the op_dist distribution:

    #  CREATE_DIST(op_dist) {
    #          { NFSPROC_LOOKUP, 27 },
    #          { NFSPROC_READ, 18 },
    #          { NFSPROC_WRITE, 9 },
    #          { NFSPROC_GETATTR, 11 },
    #          { NFSPROC_READLINK, 7 },
    #          { NFSPROC_READDIR, 2 },
    #          { NFSPROC_CREATE, 1 },
    #          { NFSPROC_MKDIR, 1 },
    #          { NFSPROC_RMDIR, 1 },
    #          { NFSPROC_REMOVE, 1 },
    #          { NFSPROC_RENAME, 1 },
    #          { NFSPROC_LINK, 1 },
    #          { NFSPROC_SYMLINK, 1},
    #          { NFSPROC_FSSTAT, 1 },
    #          { NFSPROC_SETATTR, 1 },
    #          { NFSPROC_READDIRPLUS, 9 },
    #          { NFSPROC_ACCESS, 7 },
    #          { NFSPROC_COMMIT, 5 },
    #          { PROC_SEQREAD, 18 },
    #          { PROC_SEQWRITE, 9 },
    #          { PROC_APPENDWRITE, 9 },
    #          { -1, -1}
    #  };

    # Call op_dist functionality

    if ( ! $op_distdh ) {
        my @op_dist = ( { value => 'NFSPROC_LOOKUP',      weight => 27 },
                        { value => 'NFSPROC_READ',        weight => 18 },
                        { value => 'NFSPROC_WRITE',       weight => 9 },
                        { value => 'NFSPROC_GETATTR',     weight => 11 },
                        { value => 'NFSPROC_READLINK',    weight => 7 },
                        { value => 'NFSPROC_READDIR',     weight => 2 },
                        { value => 'NFSPROC_CREATE',      weight => 1 },
                        { value => 'NFSPROC_MKDIR',       weight => 1 },
                        { value => 'NFSPROC_RMDIR',       weight => 1 },
                        { value => 'NFSPROC_REMOVE',      weight => 1 },
                        { value => 'NFSPROC_RENAME',      weight => 1 },
                        { value => 'NFSPROC_LINK',        weight => 1 },
                        { value => 'NFSPROC_SYMLINK',     weight => 1},
                        { value => 'NFSPROC_FSSTAT',      weight => 1 },
                        { value => 'NFSPROC_SETATTR',     weight => 1 },
                        { value => 'NFSPROC_READDIRPLUS', weight => 9 },
                        { value => 'NFSPROC_ACCESS',      weight => 7 },
                        { value => 'NFSPROC_COMMIT',      weight => 5 },
                        { value => 'PROC_SEQREAD',        weight => 18 },
                        { value => 'PROC_SEQWRITE',       weight => 9 },
                        { value => 'PROC_APPENDWRITE',    weight => 9 }, );
        $op_distdh = dist_init( mix => \@op_dist );
    }
    my $op = $op_distdh->distheap_select();

    gen_op( proc => $op );
} # End gen_func

#--------------------------------------------------------------------------------

=head2 rsize_dist_func

  rsize_dist_func();

Selects a read size from a distribution set

=over

=item Returns

Size of read operation to be used

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub rsize_dist_func() {
    # rsize_dist_func is rsize_dist by default
    # rsize_dist is created by CREATE_DIST:

    #  #define CREATE_DIST( _name ) \
    #  static distheap_t _name ## dh; \
    #  static struct distitem _name ## mix[]; \
    #  int _name (void) { \
    #          if ( _name ## dh == NULL ) { \
    #                  _name ## dh = dist_init( _name ## mix ); \
    #          } \
    #          return *(int *)distheap_select( _name ## dh ); \
    #  } \
    #  static struct distitem _name ## mix[] =
    #

    # Here's the rsize_dist distribution:

    #  CREATE_DIST(rsize_dist) {
    #          { 8192 * 1, 85 },
    #          { 8192 * 2, 8 },
    #          { 8192 * 4, 4 },
    #          { 8192 * 8, 2 },
    #          { 8192 * 16, 1 },
    #          { -1, -1 }
    #  };

    if ( ! $rsize_distdh ) {
        my @rsize_dist = ( { value => 8192 * 1,  weight => 85 },
                           { value => 8192 * 2,  weight => 8 },
                           { value => 8192 * 4,  weight => 4 },
                           { value => 8192 * 8,  weight => 2 },
                           { value => 8192 * 16, weight => 1 }, );
        $rsize_distdh = dist_init( mix => \@rsize_dist );
    }

    my $selection = $rsize_distdh->distheap_select();
    return $selection->{value};

} # End rsize_dist_func

#--------------------------------------------------------------------------------

=head2 wsize_dist_func

  $size = wsize_dist_func();

Selects a write size from a distribution set.

=over

=item Returns

Size of write operation to use

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub wsize_dist_func() {
    # wsize_dist_func is wsize_dist by default
    # wsize_dist is created by CREATE_DIST:

    #  #define CREATE_DIST( _name ) \
    #  static distheap_t _name ## dh; \
    #  static struct distitem _name ## mix[]; \
    #  int _name (void) { \
    #          if ( _name ## dh == NULL ) { \
    #                  _name ## dh = dist_init( _name ## mix ); \
    #          } \
    #          return *(int *)distheap_select( _name ## dh ); \
    #  } \
    #  static struct distitem _name ## mix[] =
    #

    # Here's the wsize_dist distribution:

    #  CREATE_DIST(wsize_dist) {
    #          { 8192 * 0, 49 },
    #          { 8192 * 1, 36 },
    #          { 8192 * 2, 8 },
    #          { 8192 * 4, 4 },
    #          { 8192 * 8, 2 },
    #          { 8192 * 16, 1 },
    #          { -1, -1 }
    #  };

    if ( ! $wsize_distdh ) {
        my @wsize_dist = ( { value => 8192 * 0, weight  => 49 },
                           { value => 8192 * 1, weight  => 36 },
                           { value => 8192 * 2, weight  => 8 },
                           { value => 8192 * 4, weight  => 4 },
                           { value => 8192 * 8, weight  => 2 },
                           { value => 8192 * 16, weight => 1 }, );
        $wsize_distdh = dist_init( mix => \@wsize_dist );
    }

    my $selection = $wsize_distdh->distheap_select();
    return $selection->{value};

} # End wsize_dist_func

#--------------------------------------------------------------------------------

=head2 dist_init

  $dh = dist_init( mix => \@mix );

Initializes a distribution heap.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mix

(Required) : Reference to an array containing the distribution
to be initialized.

=back

=item Returns

A distheap object containing the distribution.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub dist_init(%;) {
    #  static distheap_t
    #  dist_init(struct distitem *mix)
    #  {
    #          int i, *data, setsize = 0;
    #          distheap_t dh;
    #
    #          while (mix[setsize].weight != -1) {
    #                  setsize++;
    #          }
    #          if ((dh = distheap_init(setsize, sizeof(int))) == NULL) {
    #                  report_error(FATAL, "distheap_init failed");
    #                  return NULL;
    #          }
    #          for (i=0 ; i<setsize ; i++) {
    #                  if ((data = distheap_alloc(dh, mix[i].weight)) == NULL) {
    #                          report_error(FATAL, "distheap_alloc failed");
    #                          return NULL;
    #                  }
    #                  *data = mix[i].value;
    #          }
    #          return dh;
    #  }

    my (%opts) = validate (
        @_,
        {
            mix => { type => ARRAYREF },
         }
    );

    # Get the size of the distribution array
    my $setsize = scalar( @{$opts{mix}} );

    my $dh = nfs::Distheap->distheap_init( maxsize => $setsize,
                                           sizeof  => 'int' );

    for ( 0 .. ($setsize-1) ) {
        my $op = @{$opts{mix}}[$_];
        my ( $data, $index ) = $dh->distheap_alloc( weight => $op->{weight} );
        $dh->{entries}{$index}{data}{value} = $op->{value};
    }

    return $dh;

} # End dist_init

#--------------------------------------------------------------------------------

=head2 gen_op

  gen_op( proc => \%proc );

Calls the appropriate setup and nfs operation for a given
procedure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item proc

(Required) : Hashref to the procedure type to be called.

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub gen_op(%;) {
    #  int
    #  gen_op(int proc)
    #  {
    #          nameset_entry_t nse, pnse, onse;
    #          u_int64_t offset;
    #          u_int32_t count;
    #          int ftype;
    #
    #          /*
    #           * XXO: Need to implement
    #           *              MKNOD
    #           */
    #
    #          switch (proc) {
    #          case NFSPROC_NULL:
    #                  do_null();
    #                  break;
    #          case NFSPROC_LOOKUP:
    #                  nse = nameset_select_safe(random_ftype());
    #                  if (nse && nameset_has_parent(nse))
    #                          do_lookup(nse);
    #                  break;
    #          case NFSPROC_READ:
    #                  nse = linger_select(LINGER_READ, &offset, &count);
    #                  if (nse && count) {
    #                          do_read(nse, offset, count);
    #                  }
    #                  break;
    #          case PROC_SEQREAD:
    #                  nse = linger_select(LINGER_SEQREAD, &offset, &count);
    #                  if (nse && count) {
    #                          do_read(nse, offset, count);
    #                  }
    #                  break;
    #          case NFSPROC_WRITE:
    #                  nse = linger_select(LINGER_WRITE, &offset, &count);
    #                  if (nse && count) {
    #                          do_write(nse, offset, count, NFSV3WRITE_UNSTABLE); /*XXO*/
    #                  }
    #                  break;
    #          case PROC_SEQWRITE:
    #                  nse = linger_select(LINGER_SEQWRITE, &offset, &count);
    #                  if (nse && count) {
    #                          do_write(nse, offset, count, NFSV3WRITE_UNSTABLE); /*XXO*/
    #                  }
    #                  break;
    #          case PROC_APPENDWRITE:
    #                  nse = linger_select(LINGER_APPENDWRITE, &offset, &count);
    #                  if (nse && count) {
    #                          do_write(nse, offset, count, NFSV3WRITE_UNSTABLE); /*XXO*/
    #                  }
    #                  break;
    #          case NFSPROC_GETATTR:
    #                  nse = nameset_select_safe(random_ftype());
    #                  if (nse)
    #                          do_getattr(nse);
    #                  break;
    #          case NFSPROC_READLINK:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFLNK));
    #                  if (nse)
    #                          do_readlink(nse);
    #                  break;
    #          case NFSPROC_READDIR:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFDIR));
    #                  if (nse)
    #                          do_readdir(nse);
    #                  break;
    #          case NFSPROC_CREATE:
    #                  /*
    #                   * XXO: This should be hacked to test the error path
    #                   * where 'dir' is really a file or symlink.
    #                   */
    #                  pnse = nameset_select_safe(NFDIR);
    #                  if (pnse) {
    #                          nse = nameset_alloc(pnse, NFREG, 1); /*XXO*/
    #                          if (nse)
    #                                  do_create(nse);
    #                  }
    #                  break;
    #          case NFSPROC_MKDIR:
    #                  /*
    #                   * XXO: This should be hacked to test the error path
    #                   * where 'dir' is really a file or symlink.
    #                   */
    #                  pnse = nameset_select_safe(NFDIR);
    #                  if (pnse) {
    #                          nse = nameset_alloc(pnse, NFDIR, 1); /*XXO*/
    #                          if (nse)
    #                                  do_mkdir(nse);
    #                  }
    #                  break;
    #          case NFSPROC_RENAME:
    #                  /*
    #                   * XXO: This should be hacked to test the error paths
    #                   * for rename: renaming a directory to an existing
    #                   * non-directory, and so on and so forth.
    #                   */
    #                  ftype = NFREG; /*random_ftype_prefer(NFREG);*/
    #                  if ((pnse = nameset_select_safe(NFDIR)) != NULL
    #                   && (onse = nameset_select_safe(ftype)) != NULL
    #                   && nameset_has_parent(onse)
    #                   && (nse = nameset_link(pnse, onse, 1)) != NULL) {
    #                          do_rename(onse, nse);
    #                  }
    #                  break;
    #          case NFSPROC_LINK:
    #                  if ((pnse = nameset_select_safe(NFDIR)) != NULL
    #                   && (onse = nameset_select_safe(NFREG)) != NULL
    #                   && nameset_has_parent(onse)
    #                   && (nse = nameset_link(pnse, onse, 1)) != NULL) {
    #                          do_link(nse); /* filehandle set in nameset_link() */
    #                  }
    #                  break;
    #          case NFSPROC_SYMLINK:
    #                  if ((pnse = nameset_select_safe(NFDIR)) != NULL
    #                   && (nse = nameset_alloc(pnse, NFLNK, 1)) != NULL) {
    #                          do_symlink(nse);
    #                  }
    #                  break;
    #          case NFSPROC_REMOVE:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFREG));
    #                  if (nse && nameset_has_parent(nse)) {
    #                          contents_t ct = contents_lookup(nse);
    #                          assert(ct != NULL);
    #                          linger_dele_notify(ct);
    #                          do_remove(nse);
    #                  }
    #                  break;
    #          case NFSPROC_RMDIR:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFDIR));
    #                  if (nse && nameset_has_parent(nse))
    #                          do_rmdir(nse);
    #                  break;
    #          case NFSPROC_FSSTAT:
    #                  nse = nameset_select_safe(random_ftype()); /*root? XXO*/
    #                  if (nse)
    #                          do_fsstat(nse);
    #                  break;
    #          case NFSPROC_SETATTR:
    #                  nse = nameset_select_safe(random_ftype());
    #                  if (nse)
    #                          do_setattr(nse);
    #                  break;
    #          case NFSPROC_READDIRPLUS:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFDIR));
    #                  if (nse)
    #                          do_readdirplus(nse);
    #                  break;
    #          case NFSPROC_ACCESS:
    #                  nse = nameset_select_safe(random_ftype());
    #                  if (nse)
    #                          do_access(nse);
    #                  break;
    #          case NFSPROC_COMMIT:
    #                  nse = nameset_select_safe(random_ftype_prefer(NFREG));
    #                  if (nse)
    #                          do_commit(nse);
    #                  break;
    #          default:
    #                  report_error(FATAL, "gen_op bad proc %d", proc);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #

    my (%opts) = validate (
        @_,
        {
            proc => { type => HASHREF },
         }
    );
    my $proc = $opts{proc};

    for ( $proc->{value} ) {

        /^NFSPROC_NULL$/ && do {
                # NULL op not selected by default distribution list
                last;
            };
        /^NFSPROC_LOOKUP$/ && do {
                # OK for all NFS versions
                my $nse = $nameset->nameset_select_safe( type => random_ftype() );
                if ( defined $nse && ( $nse->{parentdir_idx} != -1 ) ) {
                    do_lookup( nse => $nse );
                }
                last;
            };
        /^NFSPROC_READ$/ && do {
                # OK for all NFS versions
                my ( $offset, $count ) = ( 0, 0 );
                my $nse = linger_select( type => LINGER_READ,
                                         off  => \$offset,
                                         cnt  => \$count );
                if ( defined $nse && $count ) {
                    do_read( nse    => $nse,
                             offset => $offset,
                             count  => $count );
                }
                last;
            };
        /^PROC_SEQREAD$/ && do {
                # OK for all NFS versions
                my ( $offset, $count ) = ( 0, 0 );
                my $nse = linger_select( type => LINGER_SEQREAD,
                                         off  => \$offset,
                                         cnt  => \$count );
                if ( defined $nse && $count ) {
                    do_read( nse    => $nse,
                             offset => $offset,
                             count  => $count );
                }
                last;
            };
        /^NFSPROC_WRITE$/ && do {
                # OK for all NFS versions
                my ( $offset, $count ) = ( 0, 0 );
                my $nse = linger_select( type => LINGER_WRITE,
                                         off  => \$offset,
                                         cnt  => \$count );
                if ( defined $nse && $count ) {
                    do_write( nse    => $nse,
                              offset => $offset,
                              count  => $count,
                              stable => UNSTABLE );
                }
                last;
            };
        /^PROC_SEQWRITE$/ && do {
                # OK for all NFS versions
                my ( $offset, $count ) = ( 0, 0 );
                my $nse = linger_select( type => LINGER_SEQWRITE,
                                         off  => \$offset,
                                         cnt  => \$count );
                if ( defined $nse && $count ) {
                    do_write( nse    => $nse,
                              offset => $offset,
                              count  => $count,
                              stable => UNSTABLE );
                }
                last;
            };
        /^PROC_APPENDWRITE$/ && do {
                # OK for all NFS versions
                my ( $offset, $count ) = ( 0, 0 );
                my $nse = linger_select( type => LINGER_APPENDWRITE,
                                         off  => \$offset,
                                         cnt  => \$count );
                if ( defined $nse && $count ) {
                    do_write( nse    => $nse,
                              offset => $offset,
                              count  => $count,
                              stable => UNSTABLE );
                }
                last;
            };
        /^NFSPROC_GETATTR$/ && do {
                # OK for all NFS versions
                my $nse = $nameset->nameset_select_safe( type => random_ftype() );
                if ( defined $nse ) {
                    do_getattr( nse => $nse );
                }
                last;
            };
        /^NFSPROC_READLINK$/ && do {
                # OK for all NFS versions
                my $ftype = random_ftype_prefer( ftype => NFLNK );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse ) {
                    do_readlink( nse => $nse );
                }
                last;
            };
        /^NFSPROC_READDIR$/ && do {
                # OK for all NFS versions
                my $ftype = random_ftype_prefer( ftype => NFDIR );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse ) {
                    do_readdir( nse => $nse );
                }
                last;
            };
        /^NFSPROC_CREATE$/ && do {
                # OK for all NFS versions
                my $pnse = $nameset->nameset_select_safe( type => NFDIR );
                if ( defined $pnse ) {
                    my $nse = $nameset->nameset_alloc( parent => $pnse,
                                                       type   => NFREG,
                                                       weight => 1 );
                    if ( defined $nse ) {
                        do_create( nse         => $nse,
                                   create_type => NFREG );
                    }
                }
                last;
            };
        /^NFSPROC_MKDIR$/ && do {
                # OK for NFS versions 2 and 3
                # NFS version 4 should call CREATE instead
                my $pnse = $nameset->nameset_select_safe( type => NFDIR );
                if ( defined $pnse ) {
                    my $nse = $nameset->nameset_alloc( parent => $pnse,
                                                       type   => NFDIR,
                                                       weight => 1 );
                    if ( defined $nse ) {
                        if ( $nfs_obj->{version} != 4 ) {
                            do_mkdir( nse => $nse );
                        } else {
                            do_create( nse         => $nse,
                                       create_type => NFDIR );
                        }
                    }
                }
                last;
            };
        /^NFSPROC_RENAME$/ && do {
                # OK for all NFS versions
                my $ftype = NFREG;
                my $pnse = $nameset->nameset_select_safe( type => NFDIR );
                last if ( !defined $pnse );
                my $onse = $nameset->nameset_select_safe( type => $ftype );
                last if ( !defined $onse );
                last if ( $onse->{parentdir_idx} == -1 );
                my $nse = $nameset->nameset_link( parent  => $pnse,
                                                  sibling => $onse,
                                                  weight  => 1 );
                if ( defined $nse ) {
                    do_rename( from => $onse,
                               to   => $nse );
                }
                last;
            };
        /^NFSPROC_LINK$/ && do {
                # OK for all NFS versions
                my $pnse = $nameset->nameset_select_safe( type => NFDIR );
                last if ( !defined $pnse );
                my $onse = $nameset->nameset_select_safe( type => NFREG );
                last if ( !defined $onse );
                last if ( $onse->{parentdir_idx} == -1 );
                my $nse = $nameset->nameset_link( parent  => $pnse,
                                                  sibling => $onse,
                                                  weight  => 1 );
                if ( defined $nse ) {
                    do_link( nse => $nse );
                }
                last;
            };
        /^NFSPROC_SYMLINK$/ && do {
                # OK for NFS versions 2 and 3
                # NFS version 4 should call CREATE instead
                my $pnse = $nameset->nameset_select_safe( type => NFDIR );
                if ( defined $pnse ) {
                    my $nse = $nameset->nameset_alloc( parent => $pnse,
                                                       type   => NFLNK,
                                                       weight => 1 );
                    if ( defined $nse ) {
                        if ( $nfs_obj->{version} != 4 ) {
                            do_symlink( nse => $nse );
                        } else {
                            do_create( nse         => $nse,
                                       create_type => NFLNK );
                        }
                    }
                }
                last;
            };
        /^NFSPROC_REMOVE$/ && do {
                # OK for all NFS versions
                my $ftype = random_ftype_prefer( ftype => NFREG );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse && $nse->{parentdir_idx} != -1 ) {
                    my $ct = contents_lookup( nse => $nse );
                    assert( defined $ct );
                    linger_dele_notify( ct => $ct );
                    do_remove( nse => $nse, remove_type => 'REMOVE' );
                }
                last;
            };
        /^NFSPROC_RMDIR$/ && do {
                # OK for NFS versions 2 and 3
                # NFS version 4 should call REMOVE instead
                my $ftype = random_ftype_prefer( ftype => NFDIR );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse && $nse->{parentdir_idx} != -1 ) {
                    if ( $nfs_obj->{version} != 4 ) {
                        do_rmdir( nse => $nse );
                    } else {
                        do_remove( nse => $nse, remove_type => 'RMDIR' );
                    }
                }
                last;
            };
        /^NFSPROC_FSSTAT$/ && do {
                # OK for NFS versions 2 and 3
                # Not supported on NFS version 4
                last if ( $nfs_obj->{version} == 4 );
                my $nse = $nameset->nameset_select_safe( type => random_ftype() );
                if ( defined $nse ) {
                    do_fsstat( nse => $nse );
                }
                last;
            };
        /^NFSPROC_SETATTR$/ && do {
                # OK for all NFS versions
                my $nse = $nameset->nameset_select_safe( type => random_ftype() );
                if ( defined $nse ) {
                    do_setattr( nse => $nse );
                }
                last;
            };
        /^NFSPROC_READDIRPLUS$/ && do {
                # OK for NFS version 3 only
                # Not supported on NFS versions 2 and 4
                last if ( $nfs_obj->{version} != 3 );
                my $ftype = random_ftype_prefer( ftype => NFDIR );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse ) {
                    do_readdirplus( nse => $nse );
                }
                last;
            };
        /^NFSPROC_ACCESS$/ && do {
                # OK for NFS versions 3 and 4
                # Not supported on NFS version 2
                last if ( $nfs_obj->{version} == 2 );
                my $nse = $nameset->nameset_select_safe( type => random_ftype() );
                if ( defined $nse ) {
                    do_access( nse => $nse );
                }
                last;
            };
        /^NFSPROC_COMMIT$/ && do {
                # OK for NFS versions 3 and 4
                # Not supported on NFS version 2
                last if ( $nfs_obj->{version} == 2 );
                my $ftype = random_ftype_prefer( ftype => NFREG );
                my $nse = $nameset->nameset_select_safe( type => $ftype );
                if ( defined $nse ) {
                    do_commit( nse => $nse );
                }
                last;
            };

        # Croak if the procedure is not found
        croak "gen_op: bad procedure: $proc";

    }

} # End gen_op

#--------------------------------------------------------------------------------

=head2 random_ftype

  $type = random_ftype();

Selects a random file type from NFDIR, NFREG, and NFLNK.

=over

=item Returns

The selected file type

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub random_ftype() {
    #  int
    #  random_ftype(void)
    #  {
    #          /* XXO: Currently nameset only supports 3 types: reg,dir,lnk */
    #          switch (random() % 3) {
    #          case 0: return NFREG;
    #          case 1: return NFDIR;
    #          case 2: return NFLNK;
    #          default:
    #                  report_error(FATAL, "random_ftype: bad type");
    #                  return NFREG;
    #          }
    #  }

    my $switch = int( rand( 3 ) );

    for ( $switch ) {
        /0/ && do { return NFREG; last };
        /1/ && do { return NFDIR; last };
        /2/ && do { return NFLNK; last };
        croak "random_ftype: bad type: $switch";
    }

} # End random_ftype

#--------------------------------------------------------------------------------

=head2 random_ftype_prefer

  $type = random_ftype_prefer( ftype => NFREG );

Returns the passed in file type 90% of the time, and a random
file type in the remaining 10%. File type is chosen from
NFREG, NFDIR, and NFLNK.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ftype

(Required) : An NFS file type from NFREG, NFDIR, or NFLNK.

=back

=item Returns

The selected file type.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub random_ftype_prefer(%;) {
    #  int
    #  random_ftype_prefer(int ftype)
    #  {
    #          if (random() % 10 == 0)
    #                  return random_ftype();  /* random type 10% of the time */
    #          else
    #                  return ftype;           /* preferred type 90% of the time */
    #  }

    my (%opts) = validate (
        @_,
        {
            ftype => { type => SCALAR },
         }
    );

    if ( int(rand( 10 )) == 0 ) {
        return random_ftype();
    } else {
        return $opts{ftype};
    }

} # End random_ftype_prefer

#--------------------------------------------------------------------------------

=head2 nameset_ref

  nameset_ref( nse => \%nse );

Increments the reference count to the Nameset object entry given.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : The Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub nameset_ref(%;) {
    #  void
    #  nameset_ref(nameset_entry_t nse)
    #  {
    #          nse->refcnt++;
    #          if (nse->refcnt <= 0) {
    #                  report_error(FATAL, "nameset entry refcnt overflow");
    #          }
    #  }

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
         }
    );

    $opts{nse}->{refcnt}++;
    if ( $opts{nse}->{refcnt} <= 0 ) {
        croak "nameset_ref: nameset entry refcnt overflow";
    }
} # End nameset_ref

#--------------------------------------------------------------------------------

=head2 nameset_deref

  nameset_deref( nse => \%nse );

Decrements the reference count to the specified Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub nameset_deref(%;) {
    #  void
    #  nameset_deref(nameset_entry_t nse)
    #  {
    #          if (nse->refcnt > 0) {
    #                  nse->refcnt--;
    #          } else {
    #                  report_error(FATAL, "nameset entry refcnt underflow");
    #          }
    #          if (nse->refcnt == 0 && nse->dele_when_free) {
    #                  nameset_dele(nse);
    #          }
    #  }

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
         }
    );

    my $nse = $opts{nse};

    if ( $nse->{refcnt} > 0 ) {
        $nse->{refcnt}--;
    } else {
        croak "nameset_deref: nameset entry refcnt underflow";
    }
    if ( $nse->{refcnt} == 0 && $nse->{dele_when_free} ) {
        $nameset->nameset_dele( nse => $nse );
    }

} # End nameset_deref

#--------------------------------------------------------------------------------

=head2 nameset_getfname

  $name = nameset_getfname( nse => \%nse );

Returns a filename corresponding to the named Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

The filename for that Nameset object entry.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub nameset_getfname(%;) {
    #  int
    #  nameset_getfname(nameset_entry_t nse, char *fname, int maxlen)
    #  {
    #          int idx = distheap_toidx(type2dh(nse->type), nse);
    #          char type;
    #
    #          switch (nse->type) {
    #          case NFREG:
    #                  type = 'f';
    #                  break;
    #          case NFDIR:
    #                  type = 'd';
    #                  break;
    #          case NFLNK:
    #                  type = 'l';
    #                  break;
    #          default:
    #                  report_error(FATAL, "nameset_getfname: bad type %d",
    #                          nse->type);
    #                  return -1;
    #          }
    #          if (snprintf(fname, maxlen, "%s%c-%d", name_prefix, type, idx) >= maxlen) {
    #                  report_error(FATAL, "nameset_getfname: buffer too small (%d)",
    #                          maxlen);
    #                  return -1; /* string truncated -- wanted to overrun */
    #          }
    #          return 0;
    #  }
    #

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
         }
    );
    my $idx;
    my $type;

    for ( $opts{nse}->{type} ) {
        $_ == NFREG && do { $idx = $opts{nse}->{my_idx}; $type = 'f'; last };
        $_ == NFDIR && do { $idx = $opts{nse}->{my_idx}; $type = 'd'; last };
        $_ == NFLNK && do { $idx = $opts{nse}->{my_idx}; $type = 'l'; last };
        croak "nameset_getfname: bad type $opts{nse}->{type}";
    }

    return "$nameset->{prefix}$type-$idx";

} # End nameset_getfname

#--------------------------------------------------------------------------------

=head2 nameset_getlinkname

  $name = nameset_getlinkname( nse => \%nse );

Returns a link name for the named Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

The link name for the named Nameset object entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub nameset_getlinkname(%;) {
    #  char *
    #  nameset_getlinkname(nameset_entry_t nse)
    #  {
    #      int idx = distheap_toidx(type2dh(nse->type), nse);
    #      static char random_pathname[] =
    #              "/Now/is/the/time/for/all/good/men/to/come/to/the/aid/of/their/country";
    #
    #      /*
    #       * Return a non-empty string.
    #       */
    #      return &random_pathname[idx % (sizeof random_pathname - 1)];
    #  }

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
         }
    );

    my $idx = $opts{nse}->{my_idx};
    my $random_pathname =
      "/Now/is/the/time/for/all/good/men/to/come/to/the/aid/of/their/country";

    return substr( $random_pathname, ($idx % length($random_pathname)) - 1 );

} # End nameset_getlinkname

#--------------------------------------------------------------------------------

=head2 do_create

  do_create( nse => \%nse, create_type => $create_type );

Makes a create request for the passed in Nameset object entry of the
specified type.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=item create_type

(Required) : The type of file to create

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub do_create(%;) {
    #  int
    #  do_create(nameset_entry_t nse)
    #  {
    #          contents_t ct = contents_lookup(nse);
    #          struct nfsmsg *nm = NULL;
    #          struct create_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #          assert(nse->type == NFREG);
    #
    #          if ((nm = op_alloc(NFSPROC_CREATE)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.create_arg;
    #
    #          /* Remember who created this file. */
    #          assert(ct != NULL);
    #          ct->ct_uid = nm->uid;
    #          ct->ct_gid = nm->gid;
    #
    #          arg->where.dir.len = nameset_parent(nse)->fhlen;
    #          arg->where.dir.data = nameset_parent(nse)->fhdata;
    #          arg->where.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->where.name.data, NFS_MAXNAMLEN);
    #          arg->where.name.len = strlen(arg->where.name.data);
    #          arg->createmode = NFSV3CREATE_UNCHECKED;
    #          arg->sattr.sa_mode.present = 1;
    #          arg->sattr.sa_mode.sa_mode = 00644; /*rw-r--r--*/
    #          arg->sattr.sa_uid.present = 0;
    #          arg->sattr.sa_gid.present = 0;
    #          arg->sattr.sa_size.present = 0;
    #          arg->sattr.sa_atime.present = 0;
    #          arg->sattr.sa_mtime.present = 0;
    #
    #          if (op_send_nse(nm, do_create_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }

    #  static void
    #  do_create_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct create_res *res = &reply->u.create_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          contents_t ct = contents_lookup(nse);
    #
    #          assert(ct != NULL); /* XXO: deref? */
    #
    #          nse->go_away = 0;
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "create cancelled");
    #                  }
    #                  linger_dele_notify(ct);
    #                  nameset_dele(nse); /* can't keep.. need fhandle! */
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "create reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  assert(res->optfh.present);
    #                  nameset_setfh(nse, res->optfh.fh.data, res->optfh.fh.len);
    #          } else {
    #                  linger_dele_notify(ct);
    #                  nameset_dele(nse);
    #          }
    #  }

    # do_create
    my (%opts) = validate (
        @_,
        {
            nse         => { type => HASHREF },
            create_type => { type => SCALAR },
        }
    );
    my $nse = $opts{nse};
    my $create_type = $opts{create_type};

    my $status;
    my $ct = contents_lookup( nse => $nse );

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    assert( $nse->{type} == $create_type );

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $mode_key = ( $nfs_obj->{version} == 4 ) ? FATTR4_MODE() : 'mode';
    my $mode_val = ( $create_type == NFDIR ) ? 0755 :
                   ( $create_type == NFLNK ) ? 0777 : 0644;
    my $res;
    eval {
        if ( $nfs_obj->{version} == 4 && $create_type == NFREG ) {
            $res = $nfs_obj->open( dhandle => $nspnt->{fhdata},
                                   name    => nameset_getfname( nse => $nse ),
                                   cr_mode => UNCHECKED,
                                   attr    => { $mode_key => $mode_val } );
        } elsif ( $nfs_obj->{version} == 4 && $create_type == NFLNK ) {
            $res = $nfs_obj->create( dhandle => $nspnt->{fhdata},
                                     type    => $nse->{type},
                                     name    => nameset_getfname( nse => $nse ),
                                     data    => nameset_getlinkname( nse => $nse ),
                                     cr_mode => UNCHECKED,
                                     attr    => { $mode_key => $mode_val } );
        } else {
            $res = $nfs_obj->create( dhandle => $nspnt->{fhdata},
                                     type    => $nse->{type},
                                     name    => nameset_getfname( nse => $nse ),
                                     cr_mode => UNCHECKED,
                                     attr    => { $mode_key => $mode_val } );
        }
    };

    my $proc = 'NFSPROC_CREATE';
    # NFS version 4 uses CREATE instead of MKDIR and SYMLINK,
    # but record the operation under the v2,v3 type for consistency
    if ( $nfs_obj->{version} == 4 ) {
        $proc = 'NFSPROC_MKDIR' if ( $create_type == NFDIR );
        $proc = 'NFSPROC_SYMLINK' if ( $create_type == NFLNK );
    }
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_create_callback
    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        linger_dele_notify( ct => $nse );
        $nameset->nameset_dele( nse => $nse );
        return;
    }

    if ( $status == NFS_OK ) {
        if ( $nfs_obj->{version} == 4 ) {
            # Find the result with the filehandle
            my @results_arr = @{$res->{results}};
            for ( reverse 0 .. $#results_arr ) {
                if ( $results_arr[$_]->{op} == OP_GETFH ) {
                    $res = $results_arr[$_]->{result};
                    last;
                }
            }
        }
        assert( defined $res->{fhandle} && $res->{fhandle} ne '' );
        my $fhlen = do { use bytes; length $res->{fhandle} };
        $nameset->nameset_setfh( nse    => $nse,
                                 fhdata => $res->{fhandle},
                                 fhlen  => $fhlen, );
    } else {
        linger_dele_notify( ct => $ct );
        $nameset->nameset_dele( nse => $nse );
    }

} # End do_create

#--------------------------------------------------------------------------------

=head2 do_mkdir

  do_mkdir( nse => \%nse );

Makes a mkdir request for the specified Nameset

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub do_mkdir(%;) {
    #  int
    #  do_mkdir(nameset_entry_t nse)
    #  {
    #          contents_t ct = contents_lookup(nse);
    #          struct nfsmsg *nm = NULL;
    #          struct mkdir_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #          assert(nse->type == NFDIR);
    #
    #          if ((nm = op_alloc(NFSPROC_MKDIR)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.mkdir_arg;
    #
    #          /* Remember who created this directory. */
    #          assert(ct != NULL);
    #          ct->ct_uid = nm->uid;
    #          ct->ct_gid = nm->gid;
    #
    #          arg->where.dir.len = nameset_parent(nse)->fhlen;
    #          arg->where.dir.data = nameset_parent(nse)->fhdata;
    #          arg->where.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->where.name.data, NFS_MAXNAMLEN);
    #          arg->where.name.len = strlen(arg->where.name.data);
    #          arg->sattr.sa_mode.present = 1;
    #          arg->sattr.sa_mode.sa_mode = 00755; /*rwxr-xr-x*/
    #          arg->sattr.sa_uid.present = 0;
    #          arg->sattr.sa_gid.present = 0;
    #          arg->sattr.sa_size.present = 0;
    #          arg->sattr.sa_atime.present = 0;
    #          arg->sattr.sa_mtime.present = 0;
    #
    #          if (op_send_nse(nm, do_mkdir_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }

    #  static void
    #  do_mkdir_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct mkdir_res *res = &reply->u.mkdir_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nse->go_away = 0;
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "mkdir cancelled");
    #                  }
    #                  nameset_dele(nse); /* can't keep.. need fhandle! */
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "mkdir reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  assert(res->optfh.present);
    #                  nameset_setfh(nse, res->optfh.fh.data, res->optfh.fh.len);
    #          } else {
    #                  nameset_dele(nse);
    #          }
    #  }

    # do_mkdir

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    my $ct = contents_lookup( nse => $nse );

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    assert( $nse->{type} == NFDIR );

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->mkdir( dhandle => $nspnt->{fhdata},
                                name    => nameset_getfname( nse => $nse ),
                                attr    => { mode => 0755 }, );
    };

    my $proc = 'NFSPROC_MKDIR';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_mkdir_callback
    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        $nameset->nameset_dele( nse => $nse );
        return;
    }

    if ( $status == NFS_OK ) {
        assert( $res->{fhandle} ne '' );
        my $fhlen = do { use bytes; length $res->{fhandle} };
        $nameset->nameset_setfh( nse    => $nse,
                                 fhdata => $res->{fhandle},
                                 fhlen  => $fhlen, );
    } else {
        $nameset->nameset_dele( nse => $nse );
    }

} # End do_mkdir

#--------------------------------------------------------------------------------

=head2 do_lookup

  do_lookup( nse => \%nse );

Makes a lookup request for the specified Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub do_lookup(%;) {
    #  int
    #  do_lookup(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct lookup_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_LOOKUP)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.lookup_arg;
    #
    #          arg->what.dir.data = nameset_parent(nse)->fhdata;
    #          arg->what.dir.len = nameset_parent(nse)->fhlen;
    #          arg->what.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->what.name.data, NFS_MAXNAMLEN);
    #          arg->what.name.len = strlen(arg->what.name.data);
    #
    #          if (op_send_nse(nm, do_lookup_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }

    #  static void
    #  do_lookup_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct lookup_res *res = &reply->u.lookup_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "lookup cancelled");
    #                  }
    #                  nameset_deref(nse);
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "lookup reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  assert(res->fh.len == nse->fhlen || *((u_int64_t *)nse->fhdata) == 0);
    #                  assert(*((u_int64_t *)res->fh.data) != 0);
    #                  if (*((u_int64_t *)nse->fhdata) == 0) {
    #                          /* a create operation didn't fill in the fh? */
    #                          /* fill it in now. XXO */
    #                          nameset_setfh(nse, res->fh.data, res->fh.len);
    #                  }
    #          }
    #
    #          nameset_deref(nse);
    #  }
    #

    # do_lookup

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->lookup( dhandle => $nspnt->{fhdata},
                                 name    => nameset_getfname( nse => $nse ), );
    };

    my $proc = 'NFSPROC_LOOKUP';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_lookup_callback
    if ( $status == -1 ) {
        nameset_deref( nse => $nse );
        return;
    }

    # Not concerned with NFS failures, only report erroneous data
    if ( $status == NFS_OK ) {
        if ( $nfs_obj->{version} == 4 ) {
            # Find the result with the filehandle
            my @results_arr = @{$res->{results}};
            for ( reverse 0 .. $#results_arr ) {
                if ( $results_arr[$_]->{op} == OP_GETFH ) {
                    $res = $results_arr[$_]->{result};
                    last;
                }
            }
        }
        my $res_fhlen = do { use bytes; length ( $res->{fhandle} ) };
        assert( $res_fhlen == $nse->{fhlen} || $nse->{fhdata} eq '' );
        assert( $res->{fhandle} ne '' );
        if ( $nse->{fhdata} eq '' ) {
            $nameset->nameset_setfh( nse    => $nse,
                                     fhdata => $res->{fhandle},
                                     fhlen  => $res_fhlen, );
        }
    }

    nameset_deref( nse => $nse );

} # End do_lookup

#--------------------------------------------------------------------------------

=head2 do_link

  do_link( nse => \%nse );

Makes a link request for the specified Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub do_link(%;) {
    #  int
    #  do_link(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct link_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #
    #          if ((nm = op_alloc(NFSPROC_LINK)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          arg = &nm->u.link_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #
    #          arg->link.dir.len = nameset_parent(nse)->fhlen;
    #          arg->link.dir.data = nameset_parent(nse)->fhdata;
    #          arg->link.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->link.name.data, NFS_MAXNAMLEN);
    #          arg->link.name.len = strlen(arg->link.name.data);
    #
    #          if (op_send_nse(nm, do_link_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }
    #
    #  static void
    #  do_link_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct link_res *res = &reply->u.link_res;
    #          nameset_entry_t nse = callback_arg;
    #
    #          nse->go_away = 0;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "link cancelled");
    #                  }
    #                  nameset_dele(nse);
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "rename reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  /* XXO: increment "hard" link count */
    #          } else {
    #                  nameset_dele(nse);
    #          }
    #  }

    # do_link

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->link( todir   => $nspnt->{fhdata},
                               fhandle => $nse->{fhdata},
                               name    => nameset_getfname( nse => $nse ), );
    };

    my $proc = 'NFSPROC_LINK';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_link_callback
    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        $nameset->nameset_dele( nse => $nse );
        return;
    }

    if ( $res->{status} == NFS_OK ) {
        # increment *hard* link count
    } else {
        $nameset->nameset_dele( nse => $nse );
    }

} # End do_link

#--------------------------------------------------------------------------------

=head2 do_symlink

  do_symlink( nse => \%nse );

Makes a symlink request for the specified nameset object

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------
sub do_symlink(%;) {
    #  int
    #  do_symlink(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm;
    #          struct symlink_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          if ((nm = op_alloc(NFSPROC_SYMLINK)) == NULL) {
    #                  report_error(FATAL, "op_alloc error");
    #                  return -1;
    #          }
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #
    #          arg = &nm->u.symlink_arg;
    #          arg->where.dir.len = nameset_parent(nse)->fhlen;
    #          arg->where.dir.data = nameset_parent(nse)->fhdata;
    #          arg->where.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->where.name.data, NFS_MAXNAMLEN);
    #          arg->where.name.len = strlen(arg->where.name.data);
    #          arg->symlinkdata.sattr.sa_mode.present = 1;
    #          arg->symlinkdata.sattr.sa_mode.sa_mode = 00777; /*rwxrwxrwx*/
    #          arg->symlinkdata.sattr.sa_uid.present = 0;
    #          arg->symlinkdata.sattr.sa_gid.present = 0;
    #          arg->symlinkdata.sattr.sa_size.present = 0;
    #          arg->symlinkdata.sattr.sa_atime.present = 0;
    #          arg->symlinkdata.sattr.sa_mtime.present = 0;
    #          arg->symlinkdata.path.data = nameset_getlinkname(nse);
    #          arg->symlinkdata.path.len = strlen(arg->symlinkdata.path.data);
    #
    #          if (op_send(nm, do_symlink_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0; /* resume via callback */
    #  }
    #
    #  static void
    #  do_symlink_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #
    #          struct symlink_res *res = &reply->u.symlink_res;
    #          nameset_entry_t nse = callback_arg;
    #
    #          nse->go_away = 0;
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "symlink cancelled");
    #                  }
    #                  nameset_dele(nse);
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "rename reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status != NFS_OK) {
    #                  nameset_dele(nse);
    #          } else {
    #                  nameset_setfh(nse, res->optfh.fh.data, res->optfh.fh.len);
    #          }
    #  }

    # do_symlink

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    my $data = nameset_getlinkname( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->symlink(
                dhandle => $nspnt->{fhdata},
                name    => nameset_getfname( nse => $nse ),
                data    => nameset_getlinkname( nse => $nse ), );
    };

    my $proc = 'NFSPROC_SYMLINK';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    if ( $nfs_obj->{version} == 2 && $status == NFS_OK ) {
        # NFS version 2 does not return a filehandle, so do a lookup
        eval {
            $res = $nfs_obj->lookup(
                    dhandle => $nspnt->{fhdata},
                    name    => nameset_getfname( nse => $nse ) );
        };
        if ( !$@ ) {
            $status = $res->{status};
        } elsif ( $nfs_obj->{result}{status} ) {
            $status = $nfs_obj->{result}{status};
        } else {
            $status = -1;       # Operation cancelled
        }
    }

    # do_symlink_callback
    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        $nameset->nameset_dele( nse => $nse );
        return;
    }

    if ( $status != NFS_OK ) {
        $nameset->nameset_dele( nse => $nse );
    } else {
        my $fhlen = do { use bytes; length $res->{fhandle} };
        $nameset->nameset_setfh( nse    => $nse,
                                 fhdata => $res->{fhandle},
                                 fhlen  => $fhlen, );
    }

} # End do_symlink

#--------------------------------------------------------------------------------

=head2 do_getattr

  do_getattr( nse => \%nse );

Makes a getattr request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub do_getattr(%;) {
    #  int
    #  do_getattr(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct getattr_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_GETATTR)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.getattr_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #
    #          if (op_send_nse(nm, do_getattr_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_getattr_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct getattr_res *res = &reply->u.getattr_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "getattr cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "getattr reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #  }

    # do_getattr

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->getattr( fhandle => $nse->{fhdata} );
    };

    my $proc = 'NFSPROC_GETATTR';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_getattr_callback
    nameset_deref( nse => $nse );

} # End do_getattr

#--------------------------------------------------------------------------------

=head2 do_readdir

  do_readdir( nse => \%nse );

Makes a readdir request for the specified Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_readdir(%;) {
    #  int
    #  do_readdir(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct readdir_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_READDIR)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.readdir_arg;
    #
    #          arg->dir.len = nse->fhlen;
    #          arg->dir.data = nse->fhdata;
    #          arg->cookie = 0;
    #          arg->cookieverf = 0;
    #          arg->count = MAX_PAYLOAD_SIZE;
    #
    #          if (op_send_nse(nm, do_readdir_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_readdir_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct readdir_res *res = &reply->u.readdir_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "readdir cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "readdir reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #
    #          /* keep reading until no more cookies? XXO */
    #  }

    # do_readdir

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->readdir( dhandle    => $nse->{fhdata},
                                  cookie     => 0,
                                  cookieverf => 0,
                                  maxcount   => MAX_PAYLOAD_SIZE );
    };

    my $proc = 'NFSPROC_READDIR';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_readdir_callback
    nameset_deref( nse => $nse );

} # End do_readdir

#--------------------------------------------------------------------------------

=head2 do_readdirplus

  do_readdirplus( nse => \%nse );

Makes a readdirplus request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_readdirplus(%;) {
    #  int
    #  do_readdirplus(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct readdirplus_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_READDIRPLUS)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.readdirplus_arg;
    #
    #          arg->dir.len = nse->fhlen;
    #          arg->dir.data = nse->fhdata;
    #          arg->cookie = 0;
    #          arg->cookieverf = 0;
    #          arg->dircount = MAX_PAYLOAD_SIZE;
    #          arg->maxcount = MAX_PAYLOAD_SIZE;
    #
    #          if (op_send_nse(nm, do_readdirplus_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_readdirplus_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct readdirplus_res *res = &reply->u.readdirplus_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "readdirplus cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "readdirplus reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #
    #          /* keep reading until no more cookies? XXO */
    #  }

    # do_readdirplus

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->readdirplus( dhandle    => $nse->{fhdata},
                                      cookie     => 0,
                                      cookieverf => 0,
                                      dircount   => MAX_PAYLOAD_SIZE,
                                      maxcount   => MAX_PAYLOAD_SIZE );
    };

    my $proc = 'NFSPROC_READDIRPLUS';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_readdirplus_callback
    nameset_deref( nse => $nse );

} # End do_readdirplus

#--------------------------------------------------------------------------------

=head2 do_access

  do_access( nse => \%nse );

Makes an access request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_access(%;) {
    #  int
    #  do_access(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct access_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_ACCESS)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.access_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #          arg->access = NFSV3ACCESS_READ;
    #
    #          if (op_send_nse(nm, do_access_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_access_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct access_res *res = &reply->u.access_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "access cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "access reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #  }

    # do_access

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->access( fhandle => $nse->{fhdata},
                                 access => ACCESS_READ );
    };

    my $proc = 'NFSPROC_ACCESS';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_access_callback
    nameset_deref( nse => $nse );

} # End do_access

#--------------------------------------------------------------------------------

=head2 do_fsstat

  do_fsstat( nse => \%nse );

Makes an fsstat request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

s(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_fsstat(%;) {
    #  int
    #  do_fsstat(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct fsstat_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_FSSTAT)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.fsstat_arg;
    #
    #          arg->fsroot.len = nse->fhlen;
    #          arg->fsroot.data = nse->fhdata;
    #
    #          if (op_send_nse(nm, do_fsstat_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_fsstat_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct fsstat_res *res = &reply->u.fsstat_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "fsstat cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "fsstat reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #  }

    # do_fsstat

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->fsstat( fhandle => $nse->{fhdata} );
    };

    my $proc = 'NFSPROC_FSSTAT';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }

    measure_op_reply( proc => $proc, status => $status );

    # do_fsstat_callback
    nameset_deref( nse => $nse );

} # End do_fsstat

#--------------------------------------------------------------------------------

=head2 do_readlink

  do_readlink( nse => \%nse );

Makes a readlink request for the specified Nameset object entry.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_readlink(%;) {
    #  int
    #  do_readlink(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct readlink_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_READLINK)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.readlink_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #
    #          if (op_send_nse(nm, do_readlink_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_readlink_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct readlink_res *res = &reply->u.readlink_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "readlink cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "readlink reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK
    #           && strncmp(nameset_getlinkname(nse), res->path.data, res->path.len) != 0) {
    #                  report_error(FATAL, "symlink path mismatch");
    #          }
    #  }

    # do_readlink

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->readlink( fhandle => $nse->{fhdata} );
    };

    my $proc = 'NFSPROC_READLINK';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_readlink_callback
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        return;
    }

    if ( $status == NFS_OK ) {
        # Compare the link information with what's expected
        # only for error reporting, irrelevant to load generation
    }

} # End do_readlink

#--------------------------------------------------------------------------------

=head2 do_rename

  do_rename( from => \%onse, to => \%nse );

Makes a rename request from the old Nameset object entry to the
new Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item from

(Required) : A Nameset object entry, to be renamed

=item to

(Required) : A Nameset object entry, to which the rename will occur

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub do_rename(%;) {
    #  int
    #  do_rename(nameset_entry_t from, nameset_entry_t to)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct rename_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(from);
    #          nameset_ref(to);
    #          from->go_away = 1;
    #          from->rename = 1;
    #          to->go_away = 1;
    #          assert(from->type == to->type);
    #
    #          if ((nm = op_alloc(NFSPROC_RENAME)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  from->rename = 0;
    #                  from->go_away = 0;
    #                  to->go_away = 0;
    #                  nameset_deref(from);
    #                  nameset_deref(to);
    #                  return -1;
    #          }
    #
    #          arg = &nm->u.rename_arg;
    #
    #          arg->from.dir.len = nameset_parent(from)->fhlen;
    #          arg->from.dir.data = nameset_parent(from)->fhdata;
    #          arg->from.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(from, arg->from.name.data, NFS_MAXNAMLEN);
    #          arg->from.name.len = strlen(arg->from.name.data);
    #          arg->from_nse = from;
    #
    #          arg->to.dir.len = nameset_parent(to)->fhlen;
    #          arg->to.dir.data = nameset_parent(to)->fhdata;
    #          arg->to.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(to, arg->to.name.data, NFS_MAXNAMLEN);
    #          arg->to.name.len = strlen(arg->to.name.data);
    #          arg->to_nse = to;
    #
    #          if (op_send_nse(nm, do_rename_callback, from, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  from->rename = 0;
    #                  from->go_away = 0;
    #                  to->go_away = 0;
    #                  nameset_deref(from);
    #                  nameset_deref(to);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }
    #
    #  static void
    #  do_rename_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          const struct rename_arg *arg = &call->u.rename_arg;
    #          struct rename_res *res = &reply->u.rename_res;
    #          nameset_entry_t from, to;
    #
    #          from = arg->from_nse;
    #          to = arg->to_nse;
    #
    #          from->go_away = 0;
    #          to->go_away = 0;
    #
    #          nameset_deref(from);
    #          nameset_deref(to);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "rename cancelled");
    #                  }
    #                  from->rename = 0;
    #                  nameset_dele(to); /* can't keep.. need fhandle! */
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "rename reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  from->go_away = 1; /* allow go_away to persist in order to stifle i/o on the from nse */
    #                  nameset_dele(from);
    #                  linger_rename(from, to);
    #          } else {
    #                  from->rename = 0;
    #                  nameset_dele(to);
    #          }
    #  }

    # do_rename

    my (%opts) = validate (
        @_,
        {
            from => { type => HASHREF },
            to   => { type => HASHREF }
        }
    );
    my $from = $opts{from};
    my $to   = $opts{to};

    my $status;

    nameset_ref( nse => $from );
    nameset_ref( nse => $to );
    $from->{go_away} = 1;
    $from->{rename}  = 1;
    $to->{go_away}   = 1;
    assert( $from->{type} == $to->{type} );

    my $from_nspnt = $nameset->nameset_parent( nse => $from );
    my $to_nspnt   = $nameset->nameset_parent( nse => $to );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->rename( fromdir  => $from_nspnt->{fhdata},
                                 fromname => nameset_getfname( nse => $from ),
                                 todir    => $to_nspnt->{fhdata},
                                 toname   => nameset_getfname( nse => $to ) );
    };

    my $proc = 'NFSPROC_RENAME';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }

    measure_op_reply( proc => $proc, status => $status );

    # do_rename_callback
    nameset_deref( nse => $from );
    nameset_deref( nse => $to );

    if ( $status == -1 ) {
        $from->{rename} = 0;
        $nameset->nameset_dele( nse => $to );
        return;
    }

    if ( $status == NFS_OK ) {
        $from->{go_away} = 1;
        $nameset->nameset_dele( nse => $from );
        linger_rename( from => $from, to => $to );
    } else {
        $from->{rename} = 0;
        $nameset->nameset_dele( nse => $to );
    }

} # End do_rename

#--------------------------------------------------------------------------------

=head2 do_remove

  do_remove( nse => \%nse, remove_type => $remove_type );

Makes a remove request for the specified Nameset object entry on the
specified type of file

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=item remove_type

(Required) : The type of object to remove

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_remove(%;) {
    #  int
    #  do_remove(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct remove_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #
    #          if ((nm = op_alloc(NFSPROC_REMOVE)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.remove_arg;
    #
    #          arg->what.dir.data = nameset_parent(nse)->fhdata;
    #          arg->what.dir.len = nameset_parent(nse)->fhlen;
    #          arg->what.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->what.name.data, NFS_MAXNAMLEN);
    #          arg->what.name.len = strlen(arg->what.name.data);
    #
    #          if (op_send_nse(nm, do_remove_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }
    #
    #  static void
    #  do_remove_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          struct remove_res *res = &reply->u.remove_res;
    #          contents_t ct = contents_lookup(nse);
    #
    #          assert(ct != NULL); /* XXO: deref? */
    #
    #          nse->go_away = 0;
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "remove cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "remove reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK || res->status == NFSERR_NOENT) {
    #                  linger_dele_notify(ct);
    #                  nameset_dele(nse);
    #          }
    #  }

    # do_remove

    my (%opts) = validate (
        @_,
        {
            nse         => { type => HASHREF },
            remove_type => { type => SCALAR },
        }
    );
    my $nse = $opts{nse};
    my $remove_type = $opts{remove_type};

    my $status;

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->remove( dhandle => $nspnt->{fhdata},
                                 name    => nameset_getfname( nse => $nse ) );
    };

    my $proc = 'NFSPROC_REMOVE';
    # NFS version 4 uses REMOVE instead of RMDIR,
    # but record the operation under the v2,v3 type for consistency
    $proc = 'NFSPROC_RMDIR' if ( $nfs_obj->{version} == 4 &&
                                 $remove_type eq 'RMDIR' );
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }

    measure_op_reply( proc => $proc, status => $status );

    # do_remove_callback
    my $ct = contents_lookup( nse => $nse );
    assert( defined $ct );

    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    if ( $status == -1 ) {
        return;
    }

    if ( $status == NFS_OK || $status == NFSERR_NOENT ) {
        linger_dele_notify( ct => $ct );
        $nameset->nameset_dele( nse => $nse );
    }

} # End do_remove

#--------------------------------------------------------------------------------

=head2 do_rmdir

  do_rmdir( nse => \%nse );

Makes a rmdir request on the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_rmdir(%;) {
    #  int
    #  do_rmdir(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct rmdir_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #          nse->go_away = 1;
    #
    #          if ((nm = op_alloc(NFSPROC_RMDIR)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.rmdir_arg;
    #
    #          arg->what.dir.data = nameset_parent(nse)->fhdata;
    #          arg->what.dir.len = nameset_parent(nse)->fhlen;
    #          arg->what.name.data = nfsmsg_alloc_data(nm, NFS_MAXNAMLEN);
    #          nameset_getfname(nse, arg->what.name.data, NFS_MAXNAMLEN);
    #          arg->what.name.len = strlen(arg->what.name.data);
    #
    #          if (op_send_nse(nm, do_rmdir_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nse->go_away = 0;
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #
    #          return 0;
    #  }
    #
    #  static void
    #  do_rmdir_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          struct rmdir_res *res = &reply->u.rmdir_res;
    #
    #          nse->go_away = 0;
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "rmdir cancelled");
    #                  }
    #                  /* nameset_dele(nse); */ /*XXO*/
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "rmdir reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->status == NFS_OK) {
    #                  nameset_dele(nse);
    #          }
    #  }

    # do_rmdir

    my (%opts) = validate (
        @_,
        {
            nse         => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );
    $nse->{go_away} = 1;

    my $nspnt = $nameset->nameset_parent( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->rmdir( dhandle => $nspnt->{fhdata},
                                name    => nameset_getfname( nse => $nse ) );
    };

    my $proc = 'NFSPROC_RMDIR';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_rmdir_callback
    $nse->{go_away} = 0;
    nameset_deref( nse => $nse );

    # Most rmdir calls should fail because NFSERR_NOTEMPTY, occasionally
    # NFSERR_NOTDIR. Must check for result as well as the status, so
    # we don't create problems.
    if ( $status == NFS_OK ) {
        $nameset->nameset_dele( nse => $nse );
    }

} # End do_rmdir

#--------------------------------------------------------------------------------

=head2 do_setattr

  do_setattr( nse => \%nse );

Makes a setattr request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_setattr(%;) {
    #  int
    #  do_setattr(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct setattr_arg *arg;
    #          u_int32_t xid = 0;
    #          contents_t ct = contents_lookup(nse);
    #
    #          assert(ct != NULL);
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_SETATTR)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.setattr_arg;
    #
    #          /*
    #           * XXO: Add random size changes.  Must flush contents cache if
    #           * the file is truncated.
    #           */
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #          arg->sattr.sa_mode.present = 0;
    #          arg->sattr.sa_uid.present = 0;
    #          arg->sattr.sa_gid.present = 0;
    #          arg->sattr.sa_atime.present = 0;
    #          arg->sattr.sa_mtime.present = 0;
    #          arg->guard_ctime.present = 0;
    #
    #          /*
    #           * Disable random truncation of files for now -- this feature hasn't
    #           * had all the bugs worked out of it and it will fail on occasion
    #           * with inconsistencies in contents.c
    #           */

# Removed if (0) section here

    #
    #          if (op_send_nse(nm, do_setattr_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }

    #  static void
    #  do_setattr_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          struct setattr_res *res = &reply->u.setattr_res;
    #          const struct setattr_arg *arg = &call->u.setattr_arg;
    #          int error = (reply == NULL || res->status != NFS_OK);
    #
    #          if (arg->sattr.sa_size.present) {
    #                  u_int32_t cnt = arg->ct_size - arg->sattr.sa_size.sa_size;
    #                  contents_fill_trunc(nse, arg->sattr.sa_size.sa_size, cnt, error);
    #          }
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "setattr cancelled");
    #                  }
    #                  nameset_deref(nse);
    #                  return;
    #          }
    #          if (error && report_nfs_errors) {
    #                  report_error(NONFATAL, "setattr reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #
    #          if (!error && arg->sattr.sa_size.present) {
    #                  assert(res->wcc_data.after.present);
    #                  assert(res->wcc_data.after.fattr.fa_size == arg->sattr.sa_size.sa_size);
    #          }
    #
    #          nameset_deref(nse);
    #  }

    # do_setattr

    my (%opts) = validate (
        @_,
        {
            nse         => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    my $ct = contents_lookup( nse => $nse );
    assert( defined $ct );

    nameset_ref( nse => $nse );

    measure_op_call();

    # Original code doesn't seem to set any attributes on the call
    my $res;
    eval {
        $res = $nfs_obj->setattr( fhandle => $nse->{fhdata},
                                  attr => { } );
    };

    my $proc = 'NFSPROC_SETATTR';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_setattr_callback
    if ( $status == -1 ) {
        nameset_deref( nse => $nse );
        return;
    }

    # Not handling error reporting right now, just deref the nameset entry
    nameset_deref( nse => $nse );

} # End do_setattr

#--------------------------------------------------------------------------------

=head2 do_write

  do_write( nse    => \%nse,
            offset => $offset,
            count  => $count,
            stable => $stable );

Makes a write request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=item offset

(Required) : Offset at which to start the write

=item count

(Required) : Number of bytes to write

=item stable

(Required) : Type of write to perform

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_write(%;) {
    #  int
    #  do_write(nameset_entry_t nse, u_int64_t offset, int count, int stable)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct write_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          assert(!nse->go_away && !nse->rename && !nse->dele_when_free);
    #
    #          assert(0 < count && count <= MAX_PAYLOAD_SIZE);
    #
    #          if (stable == NFSV3WRITE_UNSTABLE) {
    #                  nse->needs_commit = 1;
    #          }
    #
    #          if ((nm = op_alloc(NFSPROC_WRITE)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.write_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #          arg->offset = offset;
    #          arg->count = count;
    #          arg->stable = stable;
    #          arg->data.len = count;
    #
    #          contents_new_write(nse, &arg->contents_cookie,
    #                  &arg->data.data, offset, count);
    #
    #          if (op_send_nse(nm, do_write_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }

    #  static void
    #  do_write_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          const struct write_arg *arg = &call->u.write_arg;
    #          struct write_res *res = &reply->u.write_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          contents_t ct = contents_lookup(nse);
    #
    #          if (arg->data.data)
    #                  free(arg->data.data);
    #
    #          /* Inform data validation that a write completed. */
    #          contents_fill_write(nse, arg->offset, arg->count,
    #                  (!reply || res->status != NFS_OK), /* error */
    #                  &arg->contents_cookie);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "write cancelled");
    #                  }
    #                  linger_iodone(ct);
    #                  nameset_deref(nse);
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "write reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #          if (res->wcc_data.after.present) {
    #                  if (res->wcc_data.after.fattr.fa_size != ct->ct_size) {
    #                          ct->ct_size = res->wcc_data.after.fattr.fa_size;
    #                  }
    #          }
    #
    #          linger_iodone(ct);
    #          nameset_deref(nse);
    #  }

    # do_write

    my (%opts) = validate (
        @_,
        {
            nse    => { type => HASHREF },
            offset => { type => SCALAR },
            count  => { type => SCALAR },
            stable => { type => SCALAR },
        }
    );
    my $nse    = $opts{nse};
    my $offset = $opts{offset};
    my $count  = $opts{count};
    my $stable = $opts{stable};

    my $status;

    nameset_ref( nse => $nse );

    assert( !$nse->{go_away} && !$nse->{rename} && !$nse->{dele_when_free} );

    assert( 0 < $count && $count <= MAX_PAYLOAD_SIZE );

    if ( $stable == UNSTABLE ) {
        $nse->{needs_commit} = 1;
    } else {
        # Establish the needs_commit component, in case it is not there
        $nse->{needs_commit} = 0;
    }

    my $contents_cookie; # arg->contents_cookie in original code
    my $data = contents_new_write( nse              => $nse,
                                   cookie           => \$contents_cookie,
                                   off              => $offset,
                                   cnt              => $count,
                                   max_payload_size => MAX_PAYLOAD_SIZE );

    measure_op_call();

    my %stateid = ( $nfs_obj->{version} == 4 ) ? ( stateid => 0 ) : ();
    my $res;
    eval {
        $res = $nfs_obj->write( fhandle => $nse->{fhdata},
                                offset  => $offset,
                                data    => $data,
                                stable  => $stable,
                                datalen => $count, %stateid );
    };

    my $proc = 'NFSPROC_WRITE';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_write_callback
    my $ct = contents_lookup( nse => $nse );

    my $error = ( !defined $res || $res->{status} != NFS_OK );
    contents_fill_write( nse    => $nse,
                         off    => $offset,
                         cnt    => $count,
                         error  => $error,
                         cookie => $contents_cookie );

    if ( $status == -1 ) {
        linger_iodone( ct => $ct );
        nameset_deref( nse => $nse );
        return;
    }

    if ( defined $res->{file_wcc}{after}{size} ) {
        if ( $res->{file_wcc}{after}{size} != $ct->{ct_size} ) {
            $ct->{ct_size} = $res->{file_wcc}{after}{size};
        }
    }

    linger_iodone( ct => $ct );
    nameset_deref( nse => $nse );

} # End do_write

#--------------------------------------------------------------------------------

=head2 do_read

  do_read( nse    => \%nse,
           offset => $offset,
           count  => $count );

Makes a read request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=item offset

(Required) : Offset at which to start the read

=item count

(Required) : Number of bytes to request in the read

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_read(%;) {
    #  int
    #  do_read(nameset_entry_t nse, u_int64_t offset, int count)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct read_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_READ)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.read_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #          arg->offset = offset;
    #          arg->count = count;
    #
    #          contents_new_read(nse, &arg->contents_cookie);
    #
    #          if (op_send_nse(nm, do_read_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_read_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          const struct read_arg *arg = &call->u.read_arg;
    #          struct read_res *res = &reply->u.read_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #          contents_t ct = contents_lookup(nse);
    #
    #          /* Verify data against latest logged writes. */
    #          if (reply && res->status == NFS_OK) {
    #                  contents_verify_read(nse, arg->offset, res->count,
    #                          res->data.data, &arg->contents_cookie);
    #          }
    #
    #          assert(ct != NULL); /* XXO: deref? */
    #          linger_iodone(ct);
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "read cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "read reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #  }

    # do_read

    my (%opts) = validate (
        @_,
        {
            nse    => { type => HASHREF },
            offset => { type => SCALAR },
            count  => { type => SCALAR },
        }
    );
    my $nse    = $opts{nse};
    my $offset = $opts{offset};
    my $count  = $opts{count};

    my $status;

    nameset_ref( nse => $nse );

    # Contents_cookie, used for data validation?
    # Not really needed unless we do validation, but subroutine is
    # simple, so implemented just in case
    my %contents_cookie;
    contents_new_read( cookie => \%contents_cookie );

    measure_op_call();

    my %stateid = ( $nfs_obj->{version} == 4 ) ? ( stateid => 0 ) : ();
    my $res;
    eval {
        $res = $nfs_obj->read( fhandle => $nse->{fhdata},
                               offset  => $offset,
                               count   => $count, %stateid );
    };

    my $proc = 'NFSPROC_READ';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_read_callback
    my $ct = contents_lookup( nse => $nse );
    assert( defined $ct );

    if ( $status == NFS_OK ) {
        contents_verify_read( nse    => $nse,
                              off    => $offset,
                              cnt    => $count,
                              data   => $res->{data},
                              cookie => \%contents_cookie );
    }

    linger_iodone( ct  => $ct );
    nameset_deref( nse => $nse );

} # End do_read

#--------------------------------------------------------------------------------

=head2 do_commit

  do_commit( nse => \%nse );

Makes a commit request for the specified Nameset object entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub do_commit(%;) {
    #  int
    #  do_commit(nameset_entry_t nse)
    #  {
    #          struct nfsmsg *nm = NULL;
    #          struct commit_arg *arg;
    #          u_int32_t xid = 0;
    #
    #          nameset_ref(nse);
    #
    #          if ((nm = op_alloc(NFSPROC_COMMIT)) == NULL) {
    #                  report_error(NONFATAL, "op_alloc error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          arg = &nm->u.commit_arg;
    #
    #          arg->fh.len = nse->fhlen;
    #          arg->fh.data = nse->fhdata;
    #          arg->offset = 0;
    #          arg->count = 0; /* commit whole file */
    #
    #          if (op_send_nse(nm, do_commit_callback, nse, &xid) < 0) {
    #                  report_error(FATAL, "op_send error");
    #                  nameset_deref(nse);
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #
    #  static void
    #  do_commit_callback(void *callback_arg, const struct nfsmsg *call,
    #          struct nfsmsg *reply, u_int32_t xid)
    #  {
    #          struct commit_res *res = &reply->u.commit_res;
    #          nameset_entry_t nse = (nameset_entry_t)callback_arg;
    #
    #          nameset_deref(nse);
    #
    #          if (reply == NULL) {
    #                  if (report_nfs_errors) {
    #                          report_error(NONFATAL, "commit cancelled");
    #                  }
    #                  return;
    #          }
    #          if (res->status != NFS_OK && report_nfs_errors) {
    #                  report_error(NONFATAL, "commit reply error %d: %s",
    #                               res->status, nfs_errstr(res->status));
    #          }
    #  }

    # do_commit

    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    my $status;

    nameset_ref( nse => $nse );

    measure_op_call();

    my $res;
    eval {
        $res = $nfs_obj->commit( fhandle => $nse->{fhdata},
                                 offset  => 0,
                                 count   => 0 );
    };

    my $proc = 'NFSPROC_COMMIT';
    if ( !$@ ) {
        $status = $res->{status};
    } elsif ( $nfs_obj->{result}{status} ) {
        $status = $nfs_obj->{result}{status};
    } else {
        $status = -1; # Operation cancelled
    }
    measure_op_reply( proc => $proc, status => $status );

    # do_commit_callback
    nameset_deref( nse => $nse );

} # End do_commit

#--------------------------------------------------------------------------------

=head2 linger_rename

  linger_rename( from => \%nse, to => \%to );

linger_rename function taken from original C code

The following named arguments are accepted:

=over

=item Arguments

=over

=item from

(Required) : A Nameset object entry

=item to

(Required) : A Nameset object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub linger_rename(%;) {
    #  void
    #  linger_rename(nameset_entry_t from, nameset_entry_t to)
    #  {
    #          struct linger_state *ls;
    #          int t;
    #
    #          for (t=0; t < NUM_LINGER_TYPES ; t++) {
    #                  Q_FOREACH(ls, &linger_list[t], link) {
    #                          if (ls->nse == from)
    #                                  ls->nse = to;
    #                  }
    #          }
    #  }

    my (%opts) = validate (
        @_,
        {
            from => { type => HASHREF },
            to   => { type => HASHREF },
        }
    );
    my $from = $opts{from};
    my $to   = $opts{to};

    for ( my $t = 0; $t < NUM_LINGER_TYPES; $t++ ) {
        for ( my $ls = $linger_list[$t]->{head};
                defined $ls; $ls = $ls->{link_next} ) {
            if ( defined $ls->{nse} && $ls->{nse} == $from ) {
                $ls->{nse} = $to;
            }
        }
    }
} # End linger_rename

#--------------------------------------------------------------------------------

=head2 linger_dele_notify

  linger_dele_notify( ct => \%ct );

linger_dele_notify function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents entry object

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------

sub linger_dele_notify(%;) {
    #  void
    #  linger_dele_notify(contents_t ct)
    #  {
    #      struct linger_state *ls;
    #      int type, loopcount = 0;
    #
    #      /* linear scan.  big enough to make a real hash table? */
    #      for (type=0 ; type < NUM_LINGER_TYPES ; type++) {
    #      again:
    #          Q_FOREACH(ls, &linger_list[type], link) {
    #              if (ls->ct == ct) {
    #                  Q_REMOVE(&linger_list[type], ls, link);
    #                  my_free(ls);
    #                  linger_inuse--;
    #                  /*
    #                   * Q_FOREACH can't tolerate removal, and
    #                   * it's possible that the name nameset
    #                   * entry would appear in two or more linger
    #                   * state entries of the same type.  so, if
    #                   * we found one, rescan the list to see if
    #                   * there is another.
    #                   */
    #                  assert(loopcount++ < linger_maxinuse);
    #                  goto again;
    #              }
    #          }
    #      }
    #  }

    #  #define Q_REMOVE(Q, X, F) {                          \
    #      if ((Q)->head == (X)) {                      \
    #          (Q)->head = (X)->F.next;             \
    #      }                                            \
    #      if ((Q)->tail == (X)) {                      \
    #          (Q)->tail = (X)->F.prev;             \
    #      }                                            \
    #      if ((X)->F.prev) {                           \
    #          (X)->F.prev->F.next = (X)->F.next;   \
    #      }                                            \
    #      if ((X)->F.next) {                           \
    #          (X)->F.next->F.prev = (X)->F.prev;   \
    #      }                                            \
    #  }

    my (%opts) = validate (
        @_,
        {
            ct => { type => HASHREF },
        }
    );
    my $ct = $opts{ct};
    my $loopcount = 0;

  AGAIN:
    for ( my $type = 0; $type < NUM_LINGER_TYPES; $type++ ) {
        for ( my $ls = $linger_list[$type]->{head};
                defined $ls; $ls = $ls->{link_next} ) {
            if ( defined $ls->{ct} && $ls->{ct} == $ct ) {
                # Q_REMOVE
                if (($linger_list[$type])->{head} == ($ls)) {
                    ($linger_list[$type])->{head} = ($ls)->{link_next};
                }
                if (($linger_list[$type])->{tail} == ($ls)) {
                    ($linger_list[$type])->{tail} = ($ls)->{link_prev};
                }
                if (($ls)->{link_prev}) {
                    ($ls)->{link_prev}->{link_next} = ($ls)->{link_next};
                }
                if (($ls)->{link_next}) {
                    ($ls)->{link_next}->{link_prev} = ($ls)->{link_prev};
                }

                $linger_inuse--;
                assert( $loopcount++ < $linger_maxinuse );
                goto AGAIN;
            }
        }
    }

} # End linger_dele_notify

#--------------------------------------------------------------------------------

=head2 linger_select

  $nse = linger_select( type => $type,
                        off  => \$off,
                        cnt  => \$cnt );

linger_select function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Required) : SCALAR value indicating type

=item off

(Required) : SCALARREF value indicating offset

=item cnt

(Required) : SCALARREF value indicating count

=back

=item Returns

A Nameset object entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub linger_select(%;) {
    #  nameset_entry_t
    #  linger_select(int type, u_int64_t *off, u_int32_t *cnt)
    #  {
    #      struct linger_state *ls;
    #      nameset_entry_t nse;
    #      contents_t ct;
    #
    #      /*
    #       * search for an "idle-enough" active entry.
    #       *
    #       * XXO: a bunch of checks are thrown in right now to toss nse's
    #       * which may be in some kind of rename/remove state. This can
    #       * be fixed by doing linger select on inodes (contents_t) rather
    #       * than filenames (nse).
    #       */
    #   retry:
    #      Q_FOREACH(ls, &linger_list[type], link) {
    #          if (ls->iocnt < linger_maxdepth
    #          && !ls->nse->go_away
    #          && !ls->nse->dele_when_free
    #          && !ls->nse->rename) {
    #              break;
    #          }
    #      }
    #      if (ls != NULL && ls->ct->ct_type != NFREG) {
    #          report_error(NONFATAL, "linger_select found a stale sequential I/O entry, retrying");
    #          Q_REMOVE(&linger_list[type], ls, link);
    #          my_free(ls);
    #          linger_inuse--;
    #          goto retry;
    #      }
    #
    #      /*
    #       * if we didn't find one, allocate a new one.
    #       */
    #      if (ls == NULL)    {
    #          int t;
    #
    #          if (linger_inuse >= linger_maxinuse) {
    #              report_error(NONFATAL,
    #                       "linger_select: maxinuse (%d) exceeded",
    #                       linger_maxinuse);
    #              *off = 0;
    #              *cnt = 0;
    #              return NULL;
    #          }
    #
    #          /*
    #            * Get a new nse; if its contents pointer is already
    #           * cached in a linger list (hard links, or rename)
    #           * then return NULL. XXO: could we be more aggressive
    #           * and keep looking for an unused nse? I don't want
    #           * to fall into an infinite loop.
    #           */
    #          nse = nameset_select_safe(NFREG);
    #          if (nse == NULL)
    #              return NULL; /* All files busy */
    #          ct = contents_lookup(nse);
    #          assert(ct != NULL);
    #
    #          for (t=0; t < NUM_LINGER_TYPES ; t++) {
    #              Q_FOREACH(ls, &linger_list[t], link) {
    #                  if (ls->ct == ct)
    #                      return NULL;
    #              }
    #          }
    #
    #          linger_inuse++;
    #
    #          ls = my_malloc(sizeof(struct linger_state));
    #          bzero(ls, sizeof(struct linger_state));
    #          ls->iocnt = 0;
    #          ls->ct = ct;
    #          ls->nse = nse;
    #          switch(type) {
    #          case LINGER_READ:
    #              ls->off = ct->ct_size ? (random() % ct->ct_size) : 0;
    #              ls->cnt = (*rsize_dist_func)();
    #              /* Occasionally read beyond end of file */
    #              break;
    #          case LINGER_SEQREAD:
    #              ls->off = 0;
    #              ls->cnt = ct->ct_size;
    #              break;
    #          case LINGER_WRITE:
    #              ls->off = ct->ct_size ? (random() % ct->ct_size) : 0;
    #              ls->cnt = (*wsize_dist_func)();
    #              /* Occasionally write beyond end of file */
    #              break;
    #          case LINGER_SEQWRITE:
    #              ls->off = 0;
    #              ls->cnt = (*wsize_dist_func)();
    #              ls->cnt = MAX(ls->cnt, ct->ct_size);
    #              break;
    #          case LINGER_APPENDWRITE:
    #              ls->off = ct->ct_size;
    #              ls->cnt = (*wsize_dist_func)();
    #              break;
    #          }
    #          Q_INSERT_TAIL(&linger_list[type], ls, link);
    #      }
    #
    #      /*
    #       * "ls" is an active entry with a low-enough I/O count.
    #       */
    #      nse = ls->nse;
    #      *off = ls->off;
    #      *cnt = MIN(MAX_PAYLOAD_SIZE, ls->cnt);
    #      ls->off += *cnt;
    #      ls->cnt -= *cnt;
    #
    #      ls->iocnt += 1;
    #
    #      if (ls->cnt == 0) {
    #          Q_REMOVE(&linger_list[type], ls, link);
    #          my_free(ls);
    #          linger_inuse--;
    #      }
    #
    #      assert(nse->type == NFREG);
    #      return nse;
    #  }

    #  #define Q_FOREACH(X, Q, F)                           \
    #      for (X = (Q)->head ; X ; X = X->F.next)

    #  #define Q_REMOVE(Q, X, F) {                          \
    #          if ((Q)->head == (X)) {                      \
    #                  (Q)->head = (X)->F.next;             \
    #          }                                            \
    #          if ((Q)->tail == (X)) {                      \
    #                  (Q)->tail = (X)->F.prev;             \
    #          }                                            \
    #          if ((X)->F.prev) {                           \
    #                  (X)->F.prev->F.next = (X)->F.next;   \
    #          }                                            \
    #          if ((X)->F.next) {                           \
    #                  (X)->F.next->F.prev = (X)->F.prev;   \
    #          }                                            \
    #  }

    my (%opts) = validate (
        @_,
        {
            type => { type => SCALAR },
            off  => { type => SCALARREF },
            cnt  => { type => SCALARREF },
        }
    );
    my $type = $opts{type};
    my $off  = $opts{off};
    my $cnt  = $opts{cnt};

    my $nse;
    my $ct;
    my $ls;

  RETRY:
    for ( $ls = $linger_list[$type]->{head};
            defined $ls; $ls = $ls->{link_next} ) {
        if ( $ls->{iocnt} < $linger_maxdepth &&
                !$ls->{nse}{go_away} &&
                !$ls->{nse}{dele_when_free} &&
                !$ls->{nse}{rename} ) {
            last;
        }
    }

    if ( defined $ls && $ls->{ct}{ct_type} != NFREG ) {
        # Q_REMOVE
        if ( $linger_list[$type]->{head} == $ls ) {
            $linger_list[$type]->{head} = $ls->{link_next};
        }
        if ( $linger_list[$type]->{tail} == $ls ) {
            $linger_list[$type]->{tail} = $ls->{link_prev};
        }
        if ( $ls->{link_prev} ) {
            $ls->{link_prev}{link_next} = $ls->{link_next};
        }
        if ( $ls->{link_next} ) {
            $ls->{link_next}{link_prev} = $ls->{link_prev};
        }
        # Perl will release the memory when out of scope
        # my_free( $ls );
        $linger_inuse--;
        goto RETRY;
    }

    if ( !defined $ls ) {
        if ( $linger_inuse >= $linger_maxinuse ) {
            print "NONFATAL: linger_select: " .
                  "maxinuse ($linger_maxinuse) exceeded\n";
            $$off = 0;
            $$cnt = 0;
            return undef;
        }

        $nse = $nameset->nameset_select_safe( type => NFREG );
        if ( !defined $nse ) {
            return undef; # All files busy
        }
        $ct = contents_lookup( nse => $nse );
        assert( defined $ct );

        # Need to define ct->{ct_size} if not there
        $ct->{ct_size} = 0 if ( !defined $ct->{ct_size} );

        for ( my $t = 0; $t < NUM_LINGER_TYPES; $t++ ) {
            for ( $ls = $linger_list[$t]->{head};
                    defined $ls; $ls = $ls->{link_next} ) {
                if ( $ls->{ct} == $ct ) {
                    return undef;
                }
            }
        }

        $linger_inuse++;

        # Add our own assertion - $ls should be undef at this point
        assert( !defined $ls );

        $ls->{iocnt} = 0;
        $ls->{ct} = $ct;
        $ls->{nse} = $nse;

        for ( $type ) {
            ( $_ == LINGER_READ ) && do {
                    $ls->{off} = ( $ct->{ct_size} ) ?
                            int( rand($ct->{ct_size}) ) : 0;
                    $ls->{cnt} = rsize_dist_func();
                    last;
                };
            ( $_ == LINGER_SEQREAD ) && do {
                    $ls->{off} = 0;
                    $ls->{cnt} = $ct->{ct_size};
                    last;
                };
            ( $_ == LINGER_WRITE ) && do {
                    $ls->{off} = ( $ct->{ct_size} ) ?
                            int( rand($ct->{ct_size}) ) : 0;
                    $ls->{cnt} = wsize_dist_func();
                    last;
                };
            ( $_ == LINGER_SEQWRITE ) && do {
                    $ls->{off} = 0;
                    $ls->{cnt} = wsize_dist_func();
                    $ls->{cnt} = ( $ls->{cnt} > $ct->{ct_size} ) ?
                            $ls->{cnt} : $ct->{ct_size};
                    last;
                };
            ( $_ == LINGER_APPENDWRITE ) && do {
                    $ls->{off} = $ct->{ct_size};
                    $ls->{cnt} = wsize_dist_func();
                    last;
                };
            croak "linger_select: invalid type selected: $type";
        }

        # Q_INSERT_TAIL
        $ls->{link_prev} = $linger_list[$type]->{tail};
        $ls->{link_next} = undef;
        if ( defined $linger_list[$type]->{tail} ) {
            $linger_list[$type]->{tail}{link_next} = $ls;
        } else {
            $linger_list[$type]->{head} = $ls;
        }
        $linger_list[$type]->{tail} = $ls;
    } # End  if undef ( $ls )

    $nse = $ls->{nse};
    $$off = $ls->{off};
    $ls->{cnt} = 0 if ( !defined $ls->{cnt} );
    $$cnt = ( MAX_PAYLOAD_SIZE < $ls->{cnt} ) ? MAX_PAYLOAD_SIZE : $ls->{cnt};
    $ls->{off} += $$cnt;
    $ls->{cnt} -= $$cnt;

    $ls->{iocnt} += 1;

    if ( $ls->{cnt} == 0 ) {
        # Q_REMOVE
        if ( defined $linger_list[$type]->{head} &&
                $linger_list[$type]->{head} == $ls ) {
            $linger_list[$type]->{head} = $ls->{link_next};
        }
        if ( defined $linger_list[$type]->{tail} &&
                $linger_list[$type]->{tail} == $ls ) {
            $linger_list[$type]->{tail} = $ls->{link_prev};
        }
        if ( $ls->{link_prev} ) {
            $ls->{link_prev}{link_next} = $ls->{link_next};
        }
        if ( $ls->{link_next} ) {
            $ls->{link_next}{link_prev} = $ls->{link_prev};
        }

        # Perl will release the memory when out of scope
        # my_free( $ls );
        $linger_inuse--;
    }

    assert( $nse->{type} == NFREG );

    return $nse;

} # End linger_select

#--------------------------------------------------------------------------------

=head2 linger_iodone

    linger_iodone( ct => \%ct );

linger_iodone function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item ct

(Required) : A contents object entry

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub linger_iodone(%;) {
    #  void
    #  linger_iodone(contents_t ct)
    #  {
    #          struct linger_state *ls;
    #          int type;
    #
    #          /* linear scan.  big enough to make a real hash table? */
    #          for (type=0 ; type < NUM_LINGER_TYPES ; type++) {
    #                  Q_FOREACH(ls, &linger_list[type], link) {
    #                          if (ls->ct == ct) {
    #                                  /* weak record of outstanding I/O count */
    #                                  if (ls->iocnt > 0) {
    #                                          ls->iocnt--;
    #                                  }
    #                          }
    #                  }
    #          }
    #  }

    my (%opts) = validate (
        @_,
        {
            ct => { type => HASHREF },
        }
    );
    my $ct = $opts{ct};

    for ( my $type = 0; $type < NUM_LINGER_TYPES; $type++ ) {
        for ( my $ls = $linger_list[$type]->{head};
                defined $ls; $ls = $ls->{link_next} ) {
            if ( $ls->{ct} == $ct ) {
                if ( $ls->{iocnt} > 0 ) {
                    $ls->{iocnt}--;
                }
            }
        }
    }
} # End linger_iodone

#--------------------------------------------------------------------------------

# The next block covers the op_measure functions
#
# The op_measure functions operate on a list of outstanding ops
# which is needed by the C implementation for the sending and
# receiving of the NFS messages. The perl implementation does
# not keep this list, since the op sends and receives are all
# handled by the nfs_client library. We reduce the functionality
# to those parts that are necessary for statistic recording and reporting.
{
    my $stats; # Ref to statistics hash
    my $oop; # Current op data
    my %nfs_srs; # Hash of stats on a per op basis

    my %nfs_opcodes = (
                       0  => 'NFSPROC_NULL',
                       1  => 'NFSPROC_GETATTR',
                       2  => 'NFSPROC_SETATTR',
                       3  => 'NFSPROC_LOOKUP',
                       4  => 'NFSPROC_ACCESS',
                       5  => 'NFSPROC_READLINK',
                       6  => 'NFSPROC_READ',
                       7  => 'NFSPROC_WRITE',
                       8  => 'NFSPROC_CREATE',
                       9  => 'NFSPROC_MKDIR',
                       10 => 'NFSPROC_SYMLINK',
                       11 => 'NFSPROC_MKNOD',
                       12 => 'NFSPROC_REMOVE',
                       13 => 'NFSPROC_RMDIR',
                       14 => 'NFSPROC_RENAME',
                       15 => 'NFSPROC_LINK',
                       16 => 'NFSPROC_READDIR',
                       17 => 'NFSPROC_READDIRPLUS',
                       18 => 'NFSPROC_FSSTAT',
                       19 => 'NFSPROC_FSINFO',
                       20 => 'NFSPROC_PATHCONF',
                       21 => 'NFSPROC_COMMIT',
                       22 => 'NQNFSPROC_GETLEASE',
                       23 => 'NQNFSPROC_VACATED',
                       24 => 'NQNFSPROC_EVICTED',
                       25 => 'NFSPROC_NOOP',
                       26 => 'NFS_NPROCS',
                       NFS_NPROCS => 26,
                      );

#    for ( my $i = 0; $i < $nfs_opcodes{NFS_NPROCS}; $i++ ) {
#        $nfs_srs{"$nfs_opcodes{$i}"} = { good => 0,
#                                       error => 0,
#                                       rexmit => 0,
#                                       cancel => 0 };
#    }

#--------------------------------------------------------------------------------

=head2 measure_op_call

  measure_op_call();

measure_op_call function taken from original C code. Measures statistics
for an operation request.

=over

=item Returns

Nothing.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_call() {
        #  void
        #  measure_op_call(struct outstanding_op *oop, struct nfsmsg *call)
        #  {
        #          stats.call++;
        #
        #          oop->measure = 1;
        #          oop->start = global_timer();
        #  }

        # We are beginning a new op, destroy any previous $oop information
        $oop = undef;

        $stats->{call}++;
        $oop->{measure} = 1;
        $oop->{start} = global_timer();
    } # End measure_op_call

#--------------------------------------------------------------------------------

=head2 measure_op_reply

    measure_op_reply( proc   => \%proc,
                      status => \%status );

measure_op_reply function taken from original C code. Measures statistics
for an operation reply.

The following named arguments are accepted:

=over

=item Arguments

=over

=item proc

(Required) : The procedure which is being measured

=item status

(Required) : The status returned in the reply to the procedure request

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_reply(%;) {
        #  void
        #  measure_op_reply(struct outstanding_op *oop, struct nfsmsg *call,
        #                   int status)
        #  {
        #          u_int64_t now = global_timer();
        #
        #          switch (status) {
        #          case NFS_OK:
        #                  stats.reply_success++;
        #                  break;
        #          case -1:
        #                  stats.reply_cancel++;
        #                  break;
        #          default:
        #                  stats.reply_error++;
        #                  break;
        #          }
        #
        #          if (oop->measure == 0) {
        #                  return; /* do not measure */
        #          }
        #
        #          switch (status) {
        #          case NFS_OK:
        #                  statrec_add(&nfs_srs[call->proc], now - oop->start, 1);
        #                  nfs_srs[call->proc].good++;
        #                  break;
        #          case -1:
        #                  nfs_srs[call->proc].cancel++;
        #                  break;
        #          default:
        #                  statrec_add(&nfs_srs[call->proc], now - oop->start, 0);
        #                  nfs_srs[call->proc].error++;
        #                  break;
        #          }
        #
        #  #ifdef INCLUDE_SLICE_THRESHOLD_MEASURES
        #          if (call->proc == NFSPROC_READ || call->proc == NFSPROC_WRITE) {
        #                  int proc2 = -1;
        #                  if (call->proc == NFSPROC_READ) {
        #                          if (call->u.read_arg.offset +
        #                              call->u.read_arg.count <
        #                              STRIPING_ZONE_OFFSET) {
        #                                  proc2 = NFSPROC_READ_small;
        #                          } else {
        #                                  proc2 = NFSPROC_READ_large;
        #                          }
        #                  }
        #                  if (call->proc == NFSPROC_WRITE) {
        #                          if (call->u.write_arg.offset +
        #                              call->u.write_arg.count <
        #                              STRIPING_ZONE_OFFSET) {
        #                                  proc2 = NFSPROC_WRITE_small;
        #                          } else {
        #                                  proc2 = NFSPROC_WRITE_large;
        #                          }
        #                  }
        #                  assert(proc2 != -1);
        #                  switch (status) {
        #                  case NFS_OK:
        #                          statrec_add(&nfs_srs[proc2], now - oop->start, 1);
        #                          nfs_srs[proc2].good++;
        #                          break;
        #                  case -1:
        #                          nfs_srs[proc2].cancel++;
        #                          break;
        #                  default:
        #                          statrec_add(&nfs_srs[proc2], now - oop->start, 0);
        #                          nfs_srs[proc2].error++;
        #                          break;
        #                  }
        #          }
        #  #endif
        #  }

    my (%opts) = validate (
        @_,
        {
            proc   => { type => SCALAR },
            status => { type => SCALAR },
        }
    );
    my $proc = $opts{proc};
    my $status = $opts{status};

        my $now = global_timer();

        for ( $status ) {
            $_ == NFS_OK && do {
                    $stats->{reply_success}++;
                    last;
                };
            $_ == -1 && do {
                    $stats->{reply_cancel}++;
                    last;
                };
            do { # default
                    $stats->{reply_error}++;
                    last;
                };
        }

        # if (oop->measure == 0)
        #   return; /* do not measure */

        for ( $status ) {
            $_ == /NFS_OK/ && do {
                    statrec_add( sr        => \%{$nfs_srs{$proc}},
                                 n         => $now - $oop->{start},
                                 histogram => 1 );
                    $nfs_srs{$proc}{good}++;
                    last;
                };
            $_ == -1 && do {
                    $nfs_srs{$proc}{cancel}++;
                    last;
                };
            do { # default
                    statrec_add( sr        => \%{$nfs_srs{$proc}},
                                 n         => $now - $oop->{start},
                                 histogram => 0 );
                    $nfs_srs{$proc}{error}++;
                    last;
                };
        }

    } # End measure_op_reply

#--------------------------------------------------------------------------------

=head2 statrec_add

    statrec_add( sr        => \%sr,
                 n         => \%n,
                 histogram => \%histogram );

statrec_add function from original C code. Adds a statistics record for
execution output.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sr

(Required) : Statistics record object

=item n

(Required) : Value to add to statrec

=item histogram

(Required) : Histogram creation flag

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub statrec_add(%;) {
        #  static void
        #  statrec_add(struct statrec *sr, u_int32_t n, int histogram)
        #  {
        #          if (sr->magic != 0xcafebabe) {
        #                  statrec_reset(sr);
        #          }
        #          assert (n >= 0);
        #
        #          mpf_add_ui(sr->sum, sr->sum, n);   /* sr->sum += n */
        #
        #          mpf_set_ui(sr->tmp, n);
        #          mpf_mul_ui(sr->tmp, sr->tmp, n);
        #          mpf_add(sr->ssq, sr->ssq, sr->tmp); /* sr->ssq += n*n */
        #
        #          sr->cnt += 1;
        #
        #          if (histogram) {
        #                  /* only record in highest resolution bucket */
        #
        #                  int decimsec = (n + 50) / 100;
        #                  int msec = (decimsec + 5) / 10;
        #                  int decamsec = (msec + 5) / 10;
        #                  int hectomsec = (decamsec + 5) / 10;
        #
        #                  if (decimsec < 100) {
        #                          sr->hist_decimsec[decimsec]++;
        #                  } else if (msec < 100) {
        #                          sr->hist_msec[msec]++;
        #                  } else if (decamsec < 100) {
        #                          sr->hist_decamsec[decamsec]++;
        #                  } else if (hectomsec > 100) {
        #                          sr->hist_hectomsec[hectomsec]++;
        #                  } else {
        #                          sr->hist_other++;
        #                  }
        #          }
        #  }

        my (%opts) = validate (
            @_,
            {
                sr        => { type => HASHREF },
                n         => { type => SCALAR },
                histogram => { type => SCALAR },
            }
        );
        my $sr        = $opts{sr};
        my $n         = $opts{n};
        my $histogram = $opts{histogram};

        if ( !defined $sr->{magic} || $sr->{magic} != 0xcafebabe ) {
            statrec_reset( sr => $sr );
        }
        assert( $n >= 0 );

        $sr->{sum} = $sr->{sum} + $n;

        $sr->{tmp} = $n;
        $sr->{tmp} = $sr->{tmp} * $n;
        $sr->{ssq} = $sr->{ssq} + $sr->{tmp};

        $sr->{cnt} += 1;

        if ( $histogram ) {
            my $decimsec  = int( ( $n + 50 ) / 100 );
            my $msec      = int( ( $decimsec + 5 ) / 10 );
            my $decamsec  = int( ( $msec + 5 ) / 10 );
            my $hectomsec = int( ( $decamsec + 5 ) / 10 );

            if ( $decimsec < 100 ) {
                $sr->{hist_decimsec}{$decimsec}++;
                } elsif ( $msec < 100 ) {
                    $sr->{hist_msec}{$msec}++;
                } elsif ( $decamsec < 100 ) {
                    $sr->{hist_decamsec}{$decamsec}++;
                } elsif ( $hectomsec > 100 ) {
                    $sr->{hist_hectomsec}{$hectomsec}++;
                } else {
                    $sr->{hist_other}++;
                }
        }

    } # End statrec_add

#--------------------------------------------------------------------------------

=head2 statrec_reset

    statrec_reset( sr => \%sr );

statrec_reset function taken from original C code. Initializes a statrec
structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sr

(Required) : The statrec structure to initialize

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub statrec_reset(%;) {
        #  static void
        #  statrec_reset(struct statrec *sr)
        #  {
        #                  bzero(sr, sizeof *sr);
        #                  sr->magic = 0xcafebabe;
        #
        #                  mpf_init(sr->sum);
        #                  mpf_init(sr->ssq);
        #                  mpf_init(sr->tmp);
        #                  sr->cnt = 0;
        #  }

        my (%opts) = validate (
            @_,
            {
                sr => { type => HASHREF },
            }
        );

        my $sr = $opts{sr};

        $sr->{magic} = 0xcafebabe;

        $sr->{sum} = 0;
        $sr->{ssq} = 0;
        $sr->{tmp} = 0;
        $sr->{cnt} = 0;

    } # End statrec_reset

#--------------------------------------------------------------------------------

=head2 measure_op_printstats

  measure_op_printstats();

Prints the op measurement statistics

=over

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_printstats() {
        #  void
        #  measure_op_printstats(void)
        #  {
        #          int i, j;
        #
        #          printf("good\t");
        #          printf("error\t");
        #          printf("rexmit\t");
        #          printf("cancel\t");
        #          printf("avg\t");
        #          printf("stddev\t");
        #          printf("contrib\t");
        #          printf("name\n");
        #
        #          for (i=0 ; i<NFS_NPROCS ; i++) {
        #                  printf("%6d\t", nfs_srs[i].good);
        #                  printf("%6d\t", nfs_srs[i].error);
        #                  printf("%6d\t", nfs_srs[i].rexmit);
        #                  printf("%6d\t", nfs_srs[i].cancel);
        #                  printf("%6.2f\t", (float)statrec_avg(&nfs_srs[i])/1000.0);
        #                  printf("%6.2f\t", (float)statrec_stddev(&nfs_srs[i])/1000.0);
        #                  printf("%6.2f\t", (float)measure_op_contrib(i));
        #                  printf("%s\n", nfs_procstr(i));
        #          }
        #
        #          printf("\t");
        #          printf("\t");
        #          printf("\t");
        #          printf("\t");
        #          printf("%6.2f\t", (float)measure_op_global_avg()/1000.0);
        #          printf("\t");
        #          printf("\t");
        #          printf("global average\n");
        #
        #  #ifdef INCLUDE_SLICE_THRESHOLD_MEASURES
        #          for (i=NFS_NPROCS ; i<NUM_NFS_SRS ; i++) {
        #                  printf("%6d\t", nfs_srs[i].good);
        #                  printf("%6d\t", nfs_srs[i].error);
        #                  printf("%6d\t", nfs_srs[i].rexmit);
        #                  printf("%6d\t", nfs_srs[i].cancel);
        #                  printf("%6.2f\t", (float)statrec_avg(&nfs_srs[i])/1000.0);
        #                  printf("%6.2f\t", (float)statrec_stddev(&nfs_srs[i])/1000.0);
        #                  printf("%6.2f\t", (float)measure_op_contrib(i));
        #                  printf("%s %dK\n",
        #                         i == NFSPROC_READ_small ? "read <" :
        #                         i == NFSPROC_READ_large ? "read >" :
        #                         i == NFSPROC_WRITE_small ? "write <" :
        #                         i == NFSPROC_WRITE_large ? "write >" :
        #                         "(error)", STRIPING_ZONE_OFFSET/1024);
        #          }
        #  #endif
        #
        #  #if 1
        #          printf("histograms (msecs:count ... msecs+:count)\n");
        #          for (i=0 ; i<NFS_NPROCS ; i++) {
        #                  printf("HIST %s ", nfs_procstr(i));
        #                  /* [0 to 10) ms */
        #                  for (j=0 ; j<100 ; j++) {
        #                          if (nfs_srs[i].hist_decimsec[j]) {
        #                                  printf(" %0.1f:%d", (float)j / 10.0,
        #                                         nfs_srs[i].hist_decimsec[j]);
        #                          }
        #                  }
        #                  /* [10 to 100) ms */
        #                  for (j=10 ; j<100 ; j++) {
        #                          if (nfs_srs[i].hist_msec[j]) {
        #                                  printf(" %d:%d", j,
        #                                         nfs_srs[i].hist_msec[j]);
        #                          }
        #                  }
        #                  /* [100 to 1000) ms */
        #                  for (j=10 ; j<100 ; j++) {
        #                          if (nfs_srs[i].hist_decamsec[j]) {
        #                                  printf(" %d:%d", j * 10,
        #                                         nfs_srs[i].hist_decamsec[j]);
        #                          }
        #                  }
        #                  /* [1000 to 10000) ms */
        #                  for (j=10 ; j<100 ; j++) {
        #                          if (nfs_srs[i].hist_hectomsec[j]) {
        #                                  printf(" %d:%d", j * 100,
        #                                         nfs_srs[i].hist_hectomsec[j]);
        #                          }
        #                  }
        #                  /* [10000+) ms */
        #                  if (nfs_srs[i].hist_other) {
        #                          printf(" %d:%d", 10000,
        #                                 nfs_srs[i].hist_other);
        #                  }
        #                  printf("\n");
        #          }
        #  #endif
        #  }

        print "good\t" . "error\t" . "rexmit\t" . "cancel\t" . "avg\t" .
          "stddev\t" . "contrib\t" . "name\n";
        for ( my $i = 0; $i < $nfs_opcodes{NFS_NPROCS}; $i++ ) {
            # Convert the index to the op code name
            my $i_name = $nfs_opcodes{$i};
            $nfs_srs{$i_name}->{cnt} = 0 if ( !defined $nfs_srs{$i_name} );
            next if ( !defined $nfs_srs{$i_name} );
            printf("%6d\t", ( $nfs_srs{$i_name}->{good} ) ?
                    $nfs_srs{$i_name}->{good} : 0);
            printf("%6d\t", ( $nfs_srs{$i_name}->{error} ) ?
                    $nfs_srs{$i_name}->{error} : 0);
            printf("%6d\t", ( $nfs_srs{$i_name}->{rexmit} ) ?
                    $nfs_srs{$i_name}->{rexmit} : 0);
            printf("%6d\t", ( $nfs_srs{$i_name}->{cancel} ) ?
                    $nfs_srs{$i_name}->{cancel} : 0);
            printf("%6.2f\t", statrec_avg( sr => \%{$nfs_srs{$i_name}} )
                    / 1000.0);
            printf("%6.2f\t", statrec_stddev( sr => \%{$nfs_srs{$i_name}} )
                    / 1000.0);
            printf("%6.2f\t", measure_op_contrib( proc => $i_name ));
            printf("%s\n", $i_name);
        }

        print "\t" . "\t" . "\t" . "\t";
        printf("%6.2f\t", measure_op_global_avg() / 1000.0);
        print "\t" . "\t" . "global average\n";

        print "histograms (msecs:count ... msecs+:count)\n";
        for (my $i = 0 ; $i < $nfs_opcodes{NFS_NPROCS} ; $i++) {
            # Convert the index to the op code name
            my $i_name = $nfs_opcodes{$i};
            my $j;
            printf("HIST %s ", $i_name);
		# [0 to 10) ms #
            for ($j = 0 ; $j < 100 ; $j++) {
                if ($nfs_srs{$i_name}->{hist_decimsec}{$j}) {
                    printf(" %0.1f:%d", $j / 10.0,
                           $nfs_srs{$i_name}->{hist_decimsec}{$j});
                }
            }
            # [10 to 100) ms #
            for ($j = 10 ; $j < 100 ; $j++) {
                if ($nfs_srs{$i_name}->{hist_msec}{$j}) {
                    printf(" %d:%d", $j,
                           $nfs_srs{$i_name}->{hist_msec}{$j});
                }
            }
            # [100 to 1000) ms #
            for ($j = 10 ; $j < 100 ; $j++) {
                if ($nfs_srs{$i_name}->{hist_decamsec}{$j}) {
                    printf(" %d:%d", $j * 10,
                           $nfs_srs{$i_name}->{hist_decamsec}{$j});
                }
            }
            # [1000 to 10000) ms #
            for ($j = 10 ; $j < 100 ; $j++) {
                if ($nfs_srs{$i_name}->{hist_hectomsec}{$j}) {
                    printf(" %d:%d", $j * 100,
                           $nfs_srs{$i_name}->{hist_hectomsec}{$j});
                }
            }
            # [10000+) ms #
            if ($nfs_srs{$i_name}->{hist_other}) {
                printf(" %d:%d", 10000,
                       $nfs_srs{$i_name}->{hist_other});
            }
            print "\n";
        }

    } # End measure_op_printstats

#--------------------------------------------------------------------------------

=head2 statrec_avg

    $avg = statrec_avg( sr => \%sr );

statrec_avg function from original C code. Calculates an average
for the specified statistics record.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sr

(Required) : A statistics record object

=back

=item Returns

The average of the requests from this record.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub statrec_avg(%;) {
        #  static double
        #  statrec_avg(struct statrec *sr)
        #  {
        #          if (sr->cnt == 0) {
        #                  return 0;
        #          }
        #          mpf_div_ui(sr->tmp, sr->sum, sr->cnt);
        #          return mpf_get_d(sr->tmp); /* sr->sum / sr->cnt */
        #  }

        my (%opts) = validate (
            @_,
            {
                sr => { type => HASHREF },
            }
        );

        my $sr = $opts{sr};

        if ( $sr->{cnt} == 0 ) {
            return 0;
        }
        $sr->{tmp} = $sr->{sum} / $sr->{cnt};
        return $sr->{tmp};

    } # End statrec_avg

#--------------------------------------------------------------------------------

=head2 statrec_stddev

    $std_dev = statrec_stddev( sr => \%sr );

statrec_stddev function taken from original C code. Calculates the
standard deviation for the specified statistics record.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sr

(Required) : A statistics record object

=back

=item Returns

The standard deviation of the requests from this record

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub statrec_stddev(%;) {
        #  static double
        #  statrec_stddev(struct statrec *sr)
        #  {
        #          if (sr->cnt <= 1) {
        #                  return 0;
        #          }
        #          mpf_mul(sr->tmp, sr->sum, sr->sum);       /* tmp = sum*sum */
        #          mpf_div_ui(sr->tmp, sr->tmp, sr->cnt);    /* tmp = sum*sum/cnt */
        #          mpf_sub(sr->tmp, sr->ssq, sr->tmp);       /* tmp = ssq - sum*sum/cnt */
        #          mpf_div_ui(sr->tmp, sr->tmp, sr->cnt - 1);/* tmp = [3] / (cnt-1) */
        #          mpf_sqrt(sr->tmp, sr->tmp);               /* tmp = sqrt([3]/(cnt-1)) */
        #          return mpf_get_d(sr->tmp); /* sqrt((ssq - sum*sum/cnt) / (cnt - 1)) */
        #  }

        my (%opts) = validate (
            @_,
            {
                sr => { type => HASHREF },
            }
        );

        my $sr = $opts{sr};

        if ( $sr->{cnt} <= 1 ) {
            return 0;
        }
        $sr->{tmp} = $sr->{sum} * $sr->{sum};
        $sr->{tmp} = $sr->{tmp} / $sr->{cnt};
        $sr->{tmp} = $sr->{ssq} - $sr->{tmp};
        $sr->{tmp} = $sr->{tmp} / ( $sr->{cnt} - 1 );
        $sr->{tmp} = sqrt( $sr->{tmp} );
        return $sr->{tmp};

    } # End statrec_stddev

#--------------------------------------------------------------------------------

=head2 measure_op_contrib

    $contrib = measure_op_contrib( proc => \%proc );

measure_op_contrib function taken from original C code. Measures a
procedures overall contribution to the execution statistics.

The following named arguments are accepted:

=over

=item Arguments

=over

=item proc

(Required) : Procedure to be measured

=back

=item Returns

The contribution made by the specified procedure

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_contrib(%;) {
        #  static double
        #  measure_op_contrib(int proc)
        #  {
        #          double sum = 0;
        #          int i;
        #
        #          for (i=0 ; i<NFS_NPROCS ; i++) {
        #                  sum += statrec_avg(&nfs_srs[i]) * nfs_srs[i].cnt;
        #          }
        #          return (double)
        #                  (statrec_avg(&nfs_srs[proc]) * nfs_srs[proc].cnt) * 100.0 /
        #                  sum;
        #  }

        my (%opts) = validate (
            @_,
            {
                proc => { type => SCALAR },
            }
        );

        my $proc = $opts{proc};

        my $sum = 0;

        for ( my $i = 0; $i < $nfs_opcodes{NFS_NPROCS}; $i++ ) {
            # Convert the index to an op name
            my $i_name = $nfs_opcodes{$i};
            next if ( !defined $nfs_srs{$i_name} );
            $sum += statrec_avg( sr => \%{$nfs_srs{$i_name}} )
                    * $nfs_srs{$i_name}->{cnt};
        }

        return statrec_avg( sr => \%{$nfs_srs{$proc}} )
                * $nfs_srs{$proc}->{cnt} * 100 / $sum;

    } # End measure_op_contrib

#--------------------------------------------------------------------------------

=head2 measure_op_global_avg

    measure_op_global_avg();

measure_op_global_avg function taken from original C code. Measures
the average of all procedure requests.

=over

=item Returns

The average of all requests.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_global_avg() {
        #  double
        #  measure_op_global_avg(void)
        #  {
        #          double sum = 0;
        #          int i, cnt = 0;
        #
        #          for (i=0 ; i<NFS_NPROCS ; i++) {
        #                  sum += statrec_avg(&nfs_srs[i]) * nfs_srs[i].cnt;
        #                  cnt += nfs_srs[i].cnt;
        #          }
        #
        #          return sum/cnt;
        #  }

        my $sum = 0;
        my $cnt = 0;

        for ( my $i = 0; $i < $nfs_opcodes{NFS_NPROCS}; $i++ ) {
            # Convert the index to an op name
            my $i_name = $nfs_opcodes{$i};
            $sum += statrec_avg( sr => $nfs_srs{$i_name} ) *
                    $nfs_srs{$i_name}->{cnt};
            $cnt += $nfs_srs{$i_name}->{cnt};
        }

        return $sum / $cnt;
    } # End measure_op_global_avg

#--------------------------------------------------------------------------------

=head2 measure_op_called

    $op = measure_op_called();

measure_op_called function from original C code.

=over

=item Returns

The statistics for the currently executing request.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_called() {
        #  int
        #  measure_op_called(void)
        #  {
        #          return stats.call;
        #  }

        return $stats->{call};

    } # End measure_op_called

#--------------------------------------------------------------------------------

=head2 measure_op_achieved

    measure_op_achieved();

measure_op_achieved function from original C code.

=over

=item Returns

The total number of operations requested

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

    sub measure_op_achieved() {
        #  int
        #  measure_op_achieved(void)
        #  {
        #          return stats.reply_success + stats.reply_error;
        #  }

        return $stats->{reply_success} + $stats->{reply_error};

    } # End measure_op_achieved

} # End op_measure function block

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Flg SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    # Call flg with some dummy values
    eval {
        flg( server    => 'my_server.at.domain.com',
             path      => '/vol/testvol_flex',
             rate      => 2000,
             duration  => 300,
             ns_prefix => 'my_flg',
             version   => 3,
             protocol  => 'udp',
             timeout   => 30 )
    };

    # Validate that the correct error was returned
    if ($@ !~ /Can't bind to socket: Invalid argument/) {
        Tharn::logresult('FAIL', "Flg did not report socket error: $@");
    } else {
        Tharn::logresult('PASS', "SELF TEST of nfs/Flg.pm");
    }

}

1;
