# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/GenericAcl.pm#2 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::GenericAcl;

## @summary
##   Provide an OS independent interface for working with NFS V4 Access
##   Control Lists (ACL).
##
## @description
##   Provide subroutines that are called directly by tests when
##   working with ACLs. The test should treat an ACL as an opaque
##   object that is manipulated by the routines in this module.
##   This module will handle the call if it can, if not then it will
##   call the OS dependent ACL routines through the ACL class mechanism.
##
## @usage
##   use nfs::GenericAcl;
##
## @status	review
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.
##
## @notes
##  Most NFS V4 clients do not expose the full capabilities of ACLs to
##  the users yet which creates a strong temptation to use the synthetic
##  client everywhere. The approach taken is to access the ACL through
##  the client wherever possible and use the synthetic client if there
##  is no choice. By taking this approach it should be easier to modify
##  the OS dependent class to use the client more as it becomes more
##  capable without having to change any thpl scripts that use this library.
#
#   ACL Format
#
#   The ACL is a reference to a hash array. The contents are divided into
#   two catagories: ACL Class Options and ACL Contents. The ACL Class
#   Options follow:
#
#	get_acl_cmd	Command to get an ACL.
#	set_acl_cmd	Command to set an ACL.
#	sudo		Full path to the sudo command.
#	fc		Filer connection object.
#	filer_vol	Name of volume used on filer.
#	filer_path	Path of exported volume from filer.
#	mountpoint	Mountpoint on current client.
#	v4t_path	Path to v4t synthetic NFS V4 client commands.
#	cifs_passwd	CIFS admin password to use.
#	filename	File or directory name that ACL is associated with.
#	isdir		Set to 1 if filename refers to a directory.
#	use_synthetic	Set to 1 to always use synthetic client for get_acl
#			and set_acl operations.
#       retry_conn      Set to 1 to force synthetic client to retry failed
#                       socket connection attempts
#	owner		SID string (NT ACL only)
#	owner_sid	SID (NT ACL only)
#	group		SID string (NT ACL only)
#	group_sid	SID (NT ACL only)
#
#   The ACL Contents follow:
#	ace_cnt		Number of ACEs
#	ace		Array of hashes that contain ACEs. The format of an
#			ACE follows the NFS V4 specification. The reason this
#			format was chosen is because this class is used for
#			NFS V4 ACL testing. Bitmask values are defined
#			in nfs::ACL::AclNFSV4.pm. Each ACE contains the
#			following fields:
#
#			type	- Possible values are:
#				  'Allow', 'Deny', 'Audit', 'Alarm'
#			flags	- Bitmask
#			mask	- Bitmask
#			who	- name associated with ACE
#			who_sid	- SID associated with ACE (NT ACL only)

use strict;
use Carp qw(croak);
use Tharn;
use Params::Validate qw(validate ARRAYREF BOOLEAN OBJECT SCALAR);
use SPARSE::Util::Logger;
use nfs::ACL::AclNFSV4 qw(:ACL4_SUPP_ATTR :ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);
use nfs::ACL::AclClass;
use Storable qw(dclone);
use NACL::C::Node;

#-------------------------------------------------------
# Global variables
#-------------------------------------------------------
use vars qw(
  $VERSION
  @ISA
  @EXPORT
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  print_acl
  print_aces
  create_with_acl
  get_acl
  get_acl_array
  set_acl
  get_acl_support_attr
  get_acl_ace
  get_acl_filename
  set_acl_ace
  get_acl_ace_cnt
  set_acl_ace_cnt
  validate_acl_mode_bits
  compare_acl
  clear_acl
  check_aces_acl
  get_nfs_client_obj
  acl_to_synthetic_client_acl
  synthetic_client_acl_to_acl
);

#-------------------------------------------------------
# Package global variables
#-------------------------------------------------------
my $VERSION = '0.01';

#-------------------------------------------------------
# Local subroutine prototypes
#-------------------------------------------------------
sub set_acl_ace_internal ( % );

## @name ::print_acl
## @description
##  Contains the generic front end to print out an ACL. This is typically
##  used for debugging.
##
## @synopsis
##  print_acl(acl => $acl);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl			Required: ACL to set for file.
##
## @return
##  Nothing

sub print_acl ( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT }
        }
    );
    my ($ipaddr);
    my $acl = $opts{acl};

    print "ACL Object Contents: $acl\n";

    print "  ACL Class Options:\n";
    print "\tget_acl_cmd: $acl->{get_acl_cmd}\n";
    print "\tset_acl_cmd: $acl->{set_acl_cmd}\n";
    print "\tsudo: $acl->{sudo}\n";

    if ($acl->{fc}) {
        $ipaddr = $acl->{fc}->connhost();
    } else {
        $ipaddr = 0;
    }
    print "\tfiler IP: $ipaddr\n";
    print "\tfiler_vol: $acl->{filer_vol}\n";
    print "\tfiler_path: $acl->{filer_path}\n";
    print "\tmountpoint: $acl->{mountpoint}\n";
    print "\tv4t_path: $acl->{v4t_path}\n";
    print "\tcifs_passwd: $acl->{cifs_passwd}\n";
    print "\tfilename: $acl->{filename}\n";
    print "\tisdir: $acl->{isdir}\n";
    print "\tuse_synthetic: $acl->{use_synthetic}\n";
    print "\tretry_conn: $acl->{retry_conn}\n";
    print "\towner: $acl->{owner}\n";
    print "\towner_sid: $acl->{owner_sid}\n";
    printf "\towner_sid_binary: %*vb\n", "", $acl->{owner_sid_binary};
    print "\tgroup: $acl->{group}\n";
    print "\tgroup_sid: $acl->{group_sid}\n";

    print "  ACL Contents:\n";
    print "\tace_cnt: $acl->{ace_cnt}\n";
    if (defined($acl->{ace})) {
        print_aces(aces => $acl->{ace}, isdir => $acl->{isdir});
    }
    return;
}

## @name ::print_aces
## @description
##  Contains the generic front end to print out an ACE array. This
##  is typically used for debugging.
##
## @synopsis
##  print_aces(aces => $aces);
##
## @arg %opts	Required: A hash of options that include:
##
##  aces	Required: ACEs to print out
##
## @return
##  Nothing

sub print_aces ( % ) {
    my %opts = validate(
        @_,
        {
            aces  => { type => ARRAYREF },
            isdir => { type => BOOLEAN, default => 0 },
        }
    );
    my ($flags, $mask, $type);

    for (my $i = 0; $i <= $#{$opts{aces}}; $i++) {
        $type = $opts{aces}[$i]{type};
        $flags = $opts{aces}[$i]{flags};
        $mask = $opts{aces}[$i]{mask};

        # Use the naming conventions used by the v4t routines.
        print "\tACE[$i]: Type: ";
        if ($type == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
            print "Allow\n";
        } elsif ($type == ACE4_ACCESS_DENIED_ACE_TYPE) {
            print "Deny\n";
        } elsif ($type == ACE4_SYSTEM_AUDIT_ACE_TYPE) {
            print "Audit\n";
        } elsif ($type == ACE4_SYSTEM_ALARM_ACE_TYPE) {
            print "Alarm\n";
        } else {
            print "Unknown: $type\n";
        }

        print "\tACE[$i]: Flags: ";
        if ($flags & ACE4_FILE_INHERIT_ACE) {
            print "FI ";
        }
        if ($flags & ACE4_DIRECTORY_INHERIT_ACE) {
            print "DI ";
        }
        if ($flags & ACE4_NO_PROPAGATE_INHERIT_ACE) {
            print "NPI ";
        }
        if ($flags & ACE4_INHERIT_ONLY_ACE) {
            print "IONLY ";
        }
        if ($flags & ACE4_SUCCESSFUL_ACCESS_ACE_FLAG) {
            print "S ";
        }
        if ($flags & ACE4_FAILED_ACCESS_ACE_FLAG) {
            print "F ";
        }
        if ($flags & ACE4_IDENTIFIER_GROUP) {
            print "GR ";
        }
        print "\n";

        print "\tACE[$i]: Mask: ";
        if ($mask & ACE4_READ_DATA) {
            if ($opts{isdir}) {
                print "Ldir ";
            } else {
                print "Rdata ";
            }
        }
        if ($mask & ACE4_WRITE_DATA) {
            if ($opts{isdir}) {
                print "Afile ";
            } else {
                print "Wdata ";
            }
        }
        if ($mask & ACE4_APPEND_DATA) {
            if ($opts{isdir}) {
                print "Asubd ";
            } else {
                print "Adata ";
            }
        }
        if ($mask & ACE4_READ_NAMED_ATTRS) {
            print "Rna ";
        }
        if ($mask & ACE4_WRITE_NAMED_ATTRS) {
            print "Wna ";
        }
        if ($mask & ACE4_EXECUTE) {
            print "Ex ";
        }
        if ($mask & ACE4_DELETE_CHILD) {
            print "Dch ";
        }
        if ($mask & ACE4_READ_ATTRIBUTES) {
            print "Rattr ";
        }
        if ($mask & ACE4_WRITE_ATTRIBUTES) {
            print "Wattr ";
        }
        if ($mask & ACE4_DELETE) {
            print "Del ";
        }
        if ($mask & ACE4_READ_ACL) {
            print "Racl ";
        }
        if ($mask & ACE4_WRITE_ACL) {
            print "Wacl ";
        }
        if ($mask & ACE4_WRITE_OWNER) {
            print "Wown ";
        }
        if ($mask & ACE4_SYNCHRONIZE) {
            print "Sync ";
        }
        print "\n";
        print "\tACE[$i]: Who: $opts{aces}[$i]{who}\n";
        if ($opts{aces}[$i]{who_sid}) {
            print "\tACE[$i]: Who_sid: $opts{aces}[$i]{who_sid}\n";
        }
    }
    return;
}

## @name ::create_with_acl
## @description
##  Create a file or directory with the specified ACEs. It will return a
##  new and complete ACL for the file/directory just created.
##
## @synopsis
##  $acl = create_with_acl(fc => $fc, filename => $fname,
##		   filer_path => $filer_path, filer_vol => $filer_vol,
##		   mountpoint => $mountpoint);
##
## @arg %opts	Required: A hash of options that include:
##
##  ace			Required: An array of one or more ACEs.
##
##  fc			Required: Connection to filer
##
##  filename		Required: Name of file or dir to create.
##
##  filetype		Optional: "file" or "dir". Default is file.
##
##  filer_path		Required: Full pathname of volume on filer.
##
##  filer_vol		Required: Volume name to use on filer.
##
##  mode		Optional: mode to compare to.
##			Octal number or string, e.g., 0644, '644'
##			Windows does not support this option.
##
##  mountpoint		Required: Mountpoint to mount on.
##
##  retry_conn		Set to 1 to force synthetic client to retry failed
##			socket connection attempts. If you are rebooting
##			the filer under ACL load or doing givebacks/takeover
##			then you should turn this on. Default is 0 (off).
##
##  use_synthetic	Optional: If set to 1 the synthetic NFS V4
##			client is used. Default is 0. This value is
##			stored in the ACL so that when set, the synthetic
##			client is always used when working with this ACL.
##
## @return
##  the ACL object

sub create_with_acl ( % ) {
    my %opts = validate(
        @_,
        {
            ace           => { type => ARRAYREF },
            fc            => { type => OBJECT },
            filename      => { type => SCALAR },
            filetype      => { type => SCALAR,
                               regex => qr/^(file|dir)$/, default => 'file' },
            filer_path    => { type => SCALAR },
            filer_vol     => { type => SCALAR },
            mode          => { type => SCALAR, optional => 1,
                               regex => qr{^\d+$} },
            mountpoint    => { type => SCALAR },
            retry_conn    => { type => SCALAR, default => 0 },
            use_synthetic => { type => SCALAR, default => 0 },
        }
    );
    # Create the AclClass object
    # Note that the ACEs are set exactly as they are passed in as a
    # named parameter.
    my $acl = nfs::ACL::AclClass->new(%opts);

    $acl->{ace_cnt} = $#{$opts{ace}} + 1;

    # Call the OS dependent create with ACL routine
    $acl->create_with_acl(%opts);

    # Call the OS dependent get ACL routine
    $acl->get_acl();

    # Print the ACL for debugging purposes
    if (logCond("dbg")) {
        print_acl(acl => $acl);
    }

    return ($acl);
}

## @name ::get_acl
## @description
##  Contains the generic front end to the OS dependent get_acl command.
##
## @synopsis
##  $acl = get_acl(fc => $fc, filename => $fname, filer_path => $filer_path,
##                 filer_vol => $filer_vol, mountpoint => $mountpoint,
##                 use_synthetic => 1);
##
## @arg %opts	Required: A hash of options that include:
##
##  fc			Required: Connection to filer
##
##  filename		Required: Name of current file or dir to get the
##			ACL for.
##
##  filer_path		Required: Full pathname of volume on filer.
##
##  filer_vol		Required: Volume name to use on filer.
##
##  mountpoint		Required: Mountpoint to mount on.
##
##  retry_conn		Set to 1 to force synthetic client to retry failed
##			socket connection attempts. If you are rebooting
##			the filer under ACL load or doing givebacks/takeover
##			then you should turn this on. Default is 0 (off).
##
##  use_synthetic	Optional: If set to 1 the synthetic NFS V4
##			client is used. Default is 0. This value is
##			stored in the ACL so that when set, the synthetic
##			client is always used when working with this ACL.
##
## @return
##  the ACL object

sub get_acl ( % ) {
    my %opts = validate(
        @_,
        {
            fc            => { type => OBJECT },
            filename      => { type => SCALAR },
            filer_path    => { type => SCALAR },
            filer_vol     => { type => SCALAR },
            mountpoint    => { type => SCALAR },
            retry_conn    => { type => SCALAR, default => 0 },
            use_synthetic => { type => SCALAR, default => 0 },
        }
    );
    my ($acl);

    # Create the AclClass object
    $acl = nfs::ACL::AclClass->new(%opts);

    # Call the OS dependent get ACL routine
    $acl->get_acl();

    return ($acl);
}

## @name ::get_acl_array
## @description
##  Contains the generic front end to the OS dependent get_acl_array command.
##
## @synopsis
##  $acl = get_acl_array(fc => $fc, filename => $fname, filer_path => $path,
##                       filer_vol => $vol, mountpoint => $mountpoint,
##                       use_synthetic => 1);
##
## @arg %opts	Required: A hash of options that include:
##
##  fc			Required: Connection to filer
##
##  filename		Required: Name of current dir to get the ACLs for.
##
##  filer_path		Required: Full pathname of volume on filer.
##
##  filer_vol		Required: Volume name to use on filer.
##
##  mountpoint		Required: Mountpoint to mount on.
##
##  readdir_len		Optional: Reference to an array where to store the
##			reply length of each readdir request made by the
##			synthetic client.
##
##  retry_conn		Set to 1 to force synthetic client to retry failed
##			socket connection attempts. If you are rebooting
##			the filer under ACL load or doing givebacks/takeover
##			then you should turn this on. Default is 0 (off).
##
##  total_max		Optional: Specifies the maximum size to be returned
##			by the READDIR request.
##
##  use_synthetic	Optional: If set to 1 the synthetic NFS V4
##			client is used. Default is 0. This value is
##			stored in the ACL so that when set, the synthetic
##			client is always used when working with this ACL.
##
## @return
##  An array of ACL objects

sub get_acl_array ( % ) {
    my %opts = validate(
        @_,
        {
            fc            => { type => OBJECT },
            filename      => { type => SCALAR },
            filer_path    => { type => SCALAR },
            filer_vol     => { type => SCALAR },
            mountpoint    => { type => SCALAR },
            readdir_len   => { type => ARRAYREF, optional => 1 },
            total_max     => { type => SCALAR, optional => 1 },
            use_synthetic => { type => SCALAR, default => 0 },
            add_attribute => { type => SCALAR, optional => 1 },
            retry_conn    => { type => SCALAR, default => 0 },
        }
    );
    my ($acl, @acls);

    # Create an AclClass object to be used as a base for executing
    # the synthetic client
    $acl = nfs::ACL::AclClass->new(%opts);

    # Put the add_attribute option onto the acl
    $acl->{add_attribute} = $opts{add_attribute};

    # Call the OS dependent get ACL array routine
    @acls = $acl->get_acl_array();

    # return the array
    return (@acls);
}

## @name ::set_acl
## @description
##  Contains the generic front end to the OS dependent set_acl command.
##  This method writes the ACL to the server.
##
## @synopsis
##  $status = set_acl(acl => $acl);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
## @return
##  NFS4_OK on success or NFS V4 error string

sub set_acl ( % ) {
    my %opts = validate(
        @_,
        {
            acl      => { type => OBJECT },
        }
    );

    # Print the ACL for debugging purposes
    if (logCond("dbg")) {
        print_acl(acl => $opts{acl});
    }

    # Call the classes set_acl routine
    return ($opts{acl}->set_acl());
}

## @name ::get_acl_ace
## @description
##  Gets the requested ACE from an ACL. The requested ACE can be specified
##  by index with valid idx values from 0 to ace_cnt - 1, *or* the ACE can be
##  specified by the type and who parameters.
##
## @synopsis
##  ($type, $flags, $mask, $who) = get_acl_ace(acl => $acl, idx => $idx);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
##  idx		Optional: Index of ACE to return.
##
##  type	Optional: ACE Type
##
##  who		Optional: ACE Who
##
## @return
##
##  type - ACE Type or undef if ACE not found
##
##  flags - ACE Flags or undef if ACE not found
##
##  mask - ACE Mask or undef if ACE not found
##
##  who	 - ACE Who or undef if ACE not found

sub get_acl_ace ( % ) {
    my %opts = validate(
        @_,
        {
            acl  => { type => OBJECT },
            idx  => { type => SCALAR, optional => 1 },
            type => { type => SCALAR, depends => 'who', optional => 1 },
            who  => { type => SCALAR, depends => 'type', optional => 1 }
        }
    );
    my ($idx, $type, $who);
    my $acl = $opts{acl};

    if (defined($opts{idx})) {
        $idx = $opts{idx};

        # Make sure the idx is valid
        if ($idx >= 0 && $idx < $acl->{ace_cnt}) {
            return ($acl->{ace}[$idx]{type},
                    $acl->{ace}[$idx]{flags},
                    $acl->{ace}[$idx]{mask},
                    $acl->{ace}[$idx]{who});
        }
    } elsif (defined($opts{type}) && defined($opts{who})) {
        $type = $opts{type};
        $who = $opts{who};
        # Step Through the ACEs looking for a match
        for ($idx = 0; $idx < $acl->{ace_cnt}; $idx++) {
            if ($who eq $acl->{ace}[$idx]{who} &&
                $type == $acl->{ace}[$idx]{type}) {
                return ($acl->{ace}[$idx]{type},
                        $acl->{ace}[$idx]{flags},
                        $acl->{ace}[$idx]{mask},
                        $acl->{ace}[$idx]{who});
            }
        }
    } else {
        croak "get_acl_ace: need to pass in idx or type and who";
    }
    return (undef, undef, undef, undef);
}

## @name ::set_acl_ace
## @description
##  Sets an ACE in a ACL and make sure that the allowed and
##  denied fields are consistent with respect to each other if the
##  consistent flag is set.
##  This method does not write the ACL back to the server.
##
## @synopsis
##  set_acl_ace(acl => $acl, type => ACE4_ACCESS_ALLOWED_ACE_TYPE,
##              mask => ACE4_READ_DATA | ACE4_WRITE_DATA | ACE4_EXECUTE,
##              mask_how => 'SET',
##              who => 'OWNER@');
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
##  consistent	Optional: If set to yes then keep the allow and deny
##		ACEs consistent with respect to one another. If set
##		to no then do only the operation asked. Default is yes.
##
##  flags	Optional: ACE Flags
##
##  flags_how	Optional: What to do with the flags. Valid values
##		are REPLACE, SET, or CLEAR. REPLACE will replace
##		the old flag value with the new one. SET will or
##		in the bits from flags. CLEAR will unset the bits
##		from flags. Default is CLEAR.
##
##  mask	Optional: ACE Mask
##
##  mask_how	Optional: What to do with the mask. Valid values
##		are REPLACE, SET, or CLEAR. REPLACE will replace
##		the old mask value with the new one. SET will or
##		in the bits from mask. CLEAR will unset the bits
##		from mask. Default is CLEAR.
##
##  merge	Optional: "yes" indicates to merge this ACE into
##		the first ACE in the ACL that matches $who and $type.
##		Setting to "no" indicates to add the ACE at the end
##		of the ACE list. Default is "yes".
##
##  type	Required: ACE Type
##
##  who		Required: ACE Who
##
## @return
##  Nothing

sub set_acl_ace ( % ) {
    my %opts = validate(
        @_,
        {
            acl        => { type => OBJECT },
            consistent => { type => SCALAR,
                            regex => qr/^(yes|no)$/,
                            default => 'yes' },
            flags      => { type => SCALAR, default => 0 },
            flags_how  => { type => SCALAR,
                           regex => qr/^(REPLACE|SET|CLEAR)$/,
                           default => 'CLEAR' },
            mask       => { type => SCALAR, default => 0 },
            mask_how   => { type => SCALAR,
                           regex => qr/^(REPLACE|SET|CLEAR)$/,
                           default => 'CLEAR' },
            merge      => { type => SCALAR,
                            regex => qr/^(yes|no)$/,
                            default => "yes" },
            type       => { type => SCALAR },
            who        => { type => SCALAR }
        }
    );
    my ($how);

    # Perform the operation requested
    set_acl_ace_internal(%opts);

    if ($opts{consistent} eq 'yes') {
        # Perform the operation to the opposite ACE type requested
        # Whatever we do to the flags for one case (ex: allow) would
        # also be done for the opposite case (ex: deny). The mask
        # is different and must change if it was set.
        if (defined($opts{mask})) {
            if ($opts{mask_how} eq 'REPLACE') {
                # For this case, generate the complimentary mask
                # and limit it to 32 bits
                $opts{mask} = (~$opts{mask} & 0xFFFFFFFF);
            } elsif ($opts{mask_how} eq 'SET') {
                # reverse the how field
                $opts{mask_how} = 'CLEAR';
            } else {
                # reverse the how field
                $opts{mask_how} = 'SET';
            }
        }

        # For the specified type allow or deny, get the opposite type
        if ($opts{type} == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
            $opts{type} = ACE4_ACCESS_DENIED_ACE_TYPE;
        } else {
            $opts{type} = ACE4_ACCESS_ALLOWED_ACE_TYPE;
        }
        set_acl_ace_internal(%opts);
    }
    return;
}

# @name ::set_acl_ace_internal
# @description
#  Does the main work of the set_acl_ace subroutine.
#  Sets an ACE in a ACL and make sure that the allowed and
#  denied fields are consistent with respect to each other.
#  This method does not write the ACL back to the server.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options that include:
#
#  acl		Required: Valid ACL object.
#
#  consistent	Required: If set to yes then keep the allow and deny
#		ACEs consistent with respect to one another. If set
#		to no then do only the operation asked.
##
#  flags	Required: ACE Flags
#
#  flags_how	Required: What to do with the flags. Valid values
#		are REPLACE, SET, or CLEAR. REPLACE will replace
#		the old flag value with the new one. SET will or
#		in the bits from flags. CLEAR will unset the bits
#		from flags.
#
#  mask		Required: ACE Mask
#
#  mask_how	Required: What to do with the mask. Valid values
#		are REPLACE, SET, or CLEAR. REPLACE will replace
#		the old mask value with the new one. SET will or
#		in the bits from mask. CLEAR will unset the bits
#		from mask.
#
#  merge	Required: "yes" indicates to merge this ACE into
#		the first ACE in the ACL that matches $who and $type.
#		Setting to "no" indicates to add the ACE at the end
#		of the ACE list.
#
#  type		Required: ACE Type
#
#  who		Required: ACE Who
#
# @return
#  Nothing

sub set_acl_ace_internal ( % ) {
    my %opts = validate(
        @_,
        {
            acl        => { type => OBJECT },
            consistent => { type => SCALAR },
            flags      => { type => SCALAR },
            flags_how  => { type => SCALAR },
            mask       => { type => SCALAR },
            mask_how   => { type => SCALAR },
            merge      => { type => SCALAR },
            type       => { type => SCALAR },
            who        => { type => SCALAR }
        }
    );
    my ($i);
    my ($acl, $type, $flags, $flags_how, $mask, $mask_how, $merge, $who) =
        @opts{qw(acl type flags flags_how mask mask_how merge who)};

    if ($merge ne "no") {
        # See if the ACE already is on the list.
        for ($i = 0; $i < $acl->{ace_cnt}; $i++) {
            if (($acl->{ace}[$i]{who} eq $who) &&
                ($acl->{ace}[$i]{type} eq $type)) {
                    last;
            }
        }
    } else {
        # Add ACE to end of the list.
        $i = $acl->{ace_cnt};
    }

    # If this is a new ACE then copy in type and who and
    # increment the ace_cnt.
    if ($i == $acl->{ace_cnt}) {
        $acl->{ace}[$i]{type} = $type;
        $acl->{ace}[$i]{who} = $who;
        $acl->{ace_cnt}++;
    }

    # Change the flags field if specified
    if (defined($flags)) {
        if ($flags_how eq 'REPLACE') {
            $acl->{ace}[$i]{flags} = $flags;
        } elsif ($flags_how eq 'SET') {
            $acl->{ace}[$i]{flags} |= $flags;
        } else {
            # Clear the bits
            $acl->{ace}[$i]{flags} &= ~($flags);
        }
    }

    # Change the mask field if specified
    if (defined($mask)) {
        if ($mask_how eq 'REPLACE') {
            $acl->{ace}[$i]{mask} = $mask;
        } elsif ($mask_how eq 'SET') {
            $acl->{ace}[$i]{mask} |= $mask;
        } else {
            # Clear the bits
            $acl->{ace}[$i]{mask} &= ~($mask);
        }
    }

    # Delete the ACE if it has empty mask and flags
    # decrement the ace_cnt.
    if ($acl->{ace}[$i]{mask} == 0 && $acl->{ace}[$i]{flags} == 0) {
        splice @{$acl->{ace}}, $i, 1;
        $acl->{ace_cnt}--;
    }

    return;
}

## @name ::get_acl_support_attr
## @description
##  Return the ACL support attribute from the server where the
##  ACL came from.
##
## @synopsis
##  $support = get_acl_support_attr(acl => $acl);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
## @return
##  The ACL support value

sub get_acl_support_attr ( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT }
        }
    );
    return ($opts{acl}->get_acl_support_attr());
}

## @name ::get_acl_ace_cnt
## @description
##  Return the ACE count from an ACL.
##
## @synopsis
##  $ace_cnt = get_acl_ace_cnt(acl => $acl);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
## @return
##  The ACL ACE count

sub get_acl_ace_cnt ( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT }
        }
    );
    return ($opts{acl}->{ace_cnt});
}

## @name ::get_acl_filename
## @description
##  Return the filename associated with the ACL.
##
## @synopsis
##  $name = get_acl_filename(acl => $acl);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
## @return
##  The ACL ACE count

sub get_acl_filename ( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT }
        }
    );
    return ($opts{acl}->{filename});
}

## @name ::set_acl_ace_cnt
## @description
##  Set the ACE count for the ACL.
##  This method does not write the ACL back to the server.
##
## @synopsis
##  set_acl_ace_cnt(acl => $acl, ace_cnt => $ace_cnt);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl		Required: Valid ACL object.
##
##  ace_cnt	Required:  ACE count for the ACL. The new value must
##		be less than or equal to the current value.
##
## @return
##  Nothing

sub set_acl_ace_cnt ( % ) {
    my %opts = validate(
        @_,
        {
            acl     => { type => OBJECT },
            ace_cnt => { type => SCALAR }
        }
    );
    my $acl = $opts{acl};
    my $ace_cnt = $opts{ace_cnt};

    if ($ace_cnt <= $acl->{ace_cnt}) {
        $acl->{ace_cnt} = $ace_cnt;
    } else {
        croak "set_acl_ace_cnt: New ace_cnt: $ace_cnt > ace_cnt: " .
              "$acl->{ace_cnt}";
    }
    return;
}

## @name ::validate_acl_mode_bits
## @description
##  Verify that the Unix-style mode matches the given ACL.
##  If the mode is not passed in then use the Perl stat
##  routine which should work on most platforms.
##
## @synopsis
##  $status = validate_acl_mode_bits(acl => $acl, mode => $mode);
##
## @arg %opts	Required: A hash of options that include:
##
##  acl			Required: ACL to compare to.
##
##  compare		Optional: Which mode bits to compare. Valid
##			values are 'owner', 'group', 'everyone' or
##			any combination. Default is 'owner,group,everyone'.
##
##  inherit		Optional: Set to 'file' to only check ACEs
##			with the ACE4_FILE_INHERIT_ACE flag set. Set
##			to 'dir' to only check ACES with
##			ACE4_DIRECTORY_INHERIT_ACE set.
##
##  mode		Optional: mode to compare to.
##			Octal number or string, e.g., 0644, '644'
##
## @notes
##  PERL stat() call doesn't work well on Windows so always
##  return 2 which means not supported on this platform.
##
## @return
##  0 if they are the same
##  1 if they are not the same
##  2 if this platform doesn't allow the comparison.
##  croaks with error in $@.

sub validate_acl_mode_bits ( % ) {
    my %opts = validate(
        @_,
        {
            acl     => { type => OBJECT },
            compare => { type => SCALAR,
                         regex => qr/owner|group|everyone/,
                         default => 'owner,group,everyone' },
            inherit => { type => SCALAR, regex => qr/^(file|dir)$/,
                         optional => 1 },
            mode    => { type => SCALAR, optional => 1, regex => qr{^\d+$} }
        }
    );
    my (%bits, %checked, $flag, $key, $mask, $mode, $status, $tmp);
    my $acl = $opts{acl};
    my $rc = 0;
    my $h = host('localhost');

    # Windows is not supported
    if ($h->hosttype() =~ /windows/i) {
        return 2;
    }

    if (defined($opts{inherit})) {
        if ($opts{inherit} eq 'file') {
            $flag = ACE4_FILE_INHERIT_ACE;
        } else {
            $flag = ACE4_DIRECTORY_INHERIT_ACE;
        }
    } else {
        $flag = 0;
    }

    if (defined($opts{mode})) {
        $mode = $opts{mode};
    } else {
        $mode = (stat("$acl->{filename}"))[2] or
            croak "validate_acl_mode_bits: stat failed: $acl->{filename}";
    }

    $bits{owner} = ($mode & 0700)  >> 6;
    $bits{group} = ($mode & 0070)  >> 3;
    $bits{every} = ($mode & 0007);

    # Step Through the ACEs unsetting the bits in the mode that match the ACE
    for (my $i = 0; $i < $acl->{ace_cnt}; $i++) {
        if ($flag != 0 && !($flag & $acl->{ace}[$i]{flags})) {
            # If we care about the flag setting then skip ACEs that don't
            # have the right bit set.
            next;
        }

        $key = undef;
        SWITCH: {
            $acl->{ace}[$i]{who} =~ /OWNER@/ && do {
                if ($opts{compare} =~ /owner/) {
                    $key = 'owner';
                }
                last SWITCH
            };
            $acl->{ace}[$i]{who} =~ /GROUP@/ && do {
                if ($opts{compare} =~ /group/) {
                    $key = 'group';
                    if (($acl->{ace}[$i]{flags} & ACE4_IDENTIFIER_GROUP) == 0) {
                        print_acl(acl => $acl);
                        croak "validate_acl_mode_bits: ACE has group name " .
                              "GROUP@ but doesn't have ACE4_IDENTIFIER_GROUP " .
                              "flag set";
                    }
                }
                last SWITCH
            };
            $acl->{ace}[$i]{who} =~ /EVERYONE@/ && do {
                if ($opts{compare} =~ /everyone/) {
                    $key = 'every';
                }
                last SWITCH
            };
        }
        # For mode bit checking only interested in the ones above so
        # skip this ACE.
        if (!defined($key)) {
            next;
        }
        $checked{$key} = 1;
        $tmp = 0;
        if ($acl->{ace}[$i]{type} == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
            # Truth table for checking allowed ACEs. For status 0 => PASS
            # 1 => FAIL. Fail if XOR is true.
            #
            # bit mask status
            #  0   0     0
            #  0   1     1
            #  1   0     1
            #  1   1     0
            if ((($bits{$key} & 4) != 0) ^
                (($acl->{ace}[$i]{mask} & ACE4_READ_DATA)) != 0) {
                logDbg("validate_acl_mode_bits: $key bits XOR mask " .
                       "ACE4_READ_DATA is true");
                $rc = 1;
            } else {
                $tmp |= 4;
            }
            if ((($bits{$key} & 2) != 0) ^
                (($acl->{ace}[$i]{mask} & ACE4_WRITE_DATA)) != 0) {
                logDbg("validate_acl_mode_bits: $key bits XOR mask " .
                       "ACE4_WRITE_DATA is true");
                $rc = 1;
            } else {
                $tmp |= 2;
            }
            if ((($bits{$key} & 1) != 0) ^
                (($acl->{ace}[$i]{mask} & ACE4_EXECUTE)) != 0) {
                logDbg("validate_acl_mode_bits: $key bits XOR mask " .
                       "ACE4_EXECUTE is true");
                $rc = 1;
            } else {
                $tmp |= 1;
            }
            $bits{$key} &= ~$tmp;
        } elsif ($acl->{ace}[$i]{type} == ACE4_ACCESS_DENIED_ACE_TYPE) {
            # Truth table for checking denied ACEs. For status 0 => PASS
            # 1 => FAIL. Fail if AND is true.
            #
            # bit mask status
            #  0   0     0
            #  0   1     0
            #  1   0     0
            #  1   1     1
            if (($bits{$key} & 4) && ($acl->{ace}[$i]{mask} & ACE4_READ_DATA)) {
                logDbg("validate_acl_mode_bits: $key read mode set and ACE " .
                       "deny has ACE4_READ_DATA");
                $rc = 1;
            }
            if (($bits{$key} & 2) &&($acl->{ace}[$i]{mask} & ACE4_WRITE_DATA)) {
                logDbg("validate_acl_mode_bits: $key write set and ACE deny " .
                       "has ACE4_WRITE_DATA");
                $rc = 1;
            }
            if (($bits{$key} & 1) && ($acl->{ace}[$i]{mask} & ACE4_EXECUTE)) {
                logDbg("validate_acl_mode_bits: $key execute mode set and " .
                       "ACE deny has ACE4_EXECUTE");
                $rc = 1;
            }
        }
    }
    if (($checked{owner} && $bits{owner} != 0) ||
        ($checked{group} && $bits{group} != 0) ||
        ($checked{every} && $bits{every} != 0)) {
        logDbg("validate_acl_mode_bits: remainder owner: $bits{owner} group: " .
               "$bits{group} every: $bits{every} do not agree with ACEs");
        if (logCond("dbg")) {
            printf "validate_acl_mode_bits: mode: %o\n", $mode;
            print_acl(acl => $acl);
        }
        $rc = 1;
    }
    return ($rc);
}

## @name ::compare_acl
## @description
##  Compare 2 or more ACLs to make sure they are the same. The
##  routine does not assume that the ACEs are in the same order.
##
## @arg
##
##  acls	Required: Array of ACLs to compare. There must be
##		at least 2 ACLs in the array.
##
## @return
##  0 if the ACLs are all the same, 1 otherwise

sub compare_acl ( % ) {
    my %opts = validate(
        @_,
        {
            acls => { type => ARRAYREF }
        }
    );
    my ($ace_cnt, $acl, $acl1);
    my $acls_ref = $opts{acls};
    my $rc = 0;

    # Make sure at least 2 ACLs were passed in
    if (@$acls_ref < 2) {
        croak "compare_acl: At least 2 ACLs required only @$acls_ref passed in";
    }

    # Compare all the ACLs
    foreach $acl (@{$acls_ref}) {
        # Use the first ACL to compare with the rest
        if (!defined($acl1)) {
            $acl1 = $acl;
            next;
        }
        # Make sure the ace_cnts are the same
        if ($acl1->{ace_cnt} == $acl->{ace_cnt}) {
            # Step Through the ACEs in acl1 and make sure they are in acl.
            # RFC3530 states that each ACE entry is processed in order
            # by the server, therefore compare the ACEs in order to make
            # sure they are equal.
            for (my $i = 0; $i < $acl1->{ace_cnt}; $i++) {
                if ($acl->{ace}[$i]{type} != $acl1->{ace}[$i]{type} ||
                    $acl->{ace}[$i]{flags} != $acl1->{ace}[$i]{flags} ||
                    $acl->{ace}[$i]{mask} != $acl1->{ace}[$i]{mask} ||
                    $acl->{ace}[$i]{who} ne $acl1->{ace}[$i]{who}) {
                    $rc = 1;
                    last;
                }
            }
        } else {
            $rc = 1;
        }
        if ($rc != 0) {
            last;
        }
    }
    return ($rc);
}

## @name ::clear_acl
## @description
##  Clear out an ACL by changing the file mode permissions on it.
##  This doesn't clear it all out but sets the ACL to the default
##  values that reflect the current mode.
##
## @arg
##
##  acl		Required: ACL to clear out.
##
## @return
##  nothing

sub clear_acl ( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT }
        }
    );
    my ($flags, $h, $mask, $status, $type, $v4_id_domain, $username, $who);
    my $acl = $opts{acl};

    # Get NFS V4 ID domain from filer and create V4 username
    my $Node = NACL::C::Node->find();
	if($Node->mode() =~ /CMode/) {
		require nfs::ACL::CModeAcl;
		$v4_id_domain = nfs::ACL::CModeAcl::acl_get_v4_id_domain(fc => $acl->{fc});
	} else {
		require nfs::ACL::Acl;
		$v4_id_domain = nfs::ACL::Acl::acl_get_v4_id_domain(fc => $acl->{fc});
	}
    $h = host('localhost');
    if ($h->hosttype() =~ /windows/i) {
        # Windows way to get current username. Use a tag that the
        # ACL library will use to map to the current user.
        $username = 'WIN_MAP_TO_CURRENT';
    } else {
        # Unix way to get current username
        ($username) = getpwuid($>);
        $username .= "\@$v4_id_domain";
    }

    # Add an ACE for current user so that it is easy to see if the ACL gets
    # dropped because the ACE count will go to 6.

    # Add an ACE for the current user to the ACL
    set_acl_ace(acl => $acl, consistent => 'no',
                type => ACE4_ACCESS_ALLOWED_ACE_TYPE,
                mask => ACE4_READ_DATA | ACE4_WRITE_DATA | ACE4_EXECUTE,
                mask_how => 'SET', who => $username);
    $status = $acl->set_acl();
    if ($status ne 'NFS4_OK') {
        print_acl(acl => $acl);
        croak "clear_acl: set_acl failed ret: $status";
    }

    # Verify the current user ACE was added. Since Windows will
    # do a mapping on the username and we don't know what the real
    # username is, therefore we can only verify on non-windows
    # platforms.
    if ($h->hosttype() !~ /windows/i) {
        $acl->get_acl();
        ($type, $flags, $mask, $who) = get_acl_ace(acl => $acl,
                                        type => ACE4_ACCESS_ALLOWED_ACE_TYPE,
                                        who => $username);
        if (!defined($type)) {
            print_acl(acl => $acl);
            croak "clear_acl: newly added ACE not found in ACL";
        }
    }

    # Change the permission bits on the files mode to drop ACL
    # Value used for mode doesn't really matter
    chmod 0755, $acl->{filename} or
        croak "clear_acl: chmod failed: $acl->{filename} 0755: $!";

    # Verify the previous ACL was dropped
    # When an ACL is dropped, the ACE count should go down to 6 (for C-mode the default ACE count will be 3) and
    # the ACL should match the permission bits on the file. This only
    # works on NFS V4 and isn't true for windows.
	if ($h->hosttype() !~ /windows/i) {
        $acl->get_acl();
		if($Node->mode() =~ /CMode/) {
			if ($acl->{ace_cnt} != 3) {
				print_acl(acl => $acl);
				croak "clear_acl: ACL count exp 3 got: $acl->{ace_cnt}";
			}
		}
		else {
			if ($acl->{ace_cnt} != 6) {
				print_acl(acl => $acl);
				croak "clear_acl: ACL count exp 6 got: $acl->{ace_cnt}";
			}
		}	
        $status = validate_acl_mode_bits(acl => $acl);
        if ($status != 0) {
            print_acl(acl => $acl);
            croak "clear_acl: Failed validating ACL and mode bits";
        }
    }
    return;
}

## @name ::check_aces_acl
## @description
##  Verify that all the passed in ACEs are in the passed in ACL.
##  The flag field is only compared when flag_cmp is set to 'yes'
##  and then the flag fields are only anded to see if a subset of
##  bits match. This is useful when looking at directory inherited
##  ACEs. The mask field is always checked for an exact match.
##
## @arg
##
##  aces	Required: An array of one or more ACEs.
##
##  acl		Required: ACL to verify against.
##
##  flag_cmp	Optional: If set to 'yes' indicates that the
##			  flag field of the ACEs should have
##			  a subset of bits that match. If set
##			  to 'no', then no check is done on the
##			  flags. This is useful when comparing
##			  inherited file ACEs. The default is 'yes'.
##
## @return
##  0 if it passed the check, 1 otherwise

sub check_aces_acl ( % ) {
    my %opts = validate(
        @_,
        {
            aces        => { type => ARRAYREF },
            acl         => { type => OBJECT },
            flag_cmp    => { type => SCALAR, default => 'yes' },
            group_cmp   => { type => SCALAR, default => 'yes' },
        }
    );
    my ($flags, $found, $i, $j, $mask, $num_aces, $type, $who);

    # Step through the passed in ACEs and makes sure that each one is in
    # the ACL. Note that the debugging statements are commented out because
    # they generate a lot of output which is only needed when debugging
    # this routine.
    $num_aces = scalar(@{$opts{aces}});
    logDbg("check_aces_acl: num_aces:$num_aces ace_cnt:$opts{acl}->{ace_cnt}");
    for ($i = 0; $i < $num_aces; $i++) {
        # Skip if we are looking at a group ACE and group_cmp is not set
        next if (${opts{aces}}[$i]{who} eq 'GROUP@' &&
                 $opts{group_cmp} ne 'yes');

        if (logCond("dbg")) {
            my $x_f = sprintf "0x%lx", ${opts{aces}}[$i]{flags};
            my $x_m = sprintf "0x%lx", ${opts{aces}}[$i]{mask};
#            logDbg("check_aces_acl: ACE idx: $i type: " .
#                   ${opts{aces}}[$i]{type} .  " flags: $x_f mask: $x_m who: " .
#                   ${opts{aces}}[$i]{who});
        }
        # Step through all the ACEs in the ACL instead of just searching
        # on type and who because there can be more than one types and
        # who can be repeated and get_acl_ace() just returns the first one.
        $found = 0;
        for ($j = 0; $j < $opts{acl}->{ace_cnt}; $j++) {
            ($type, $flags, $mask, $who) = get_acl_ace(acl => $opts{acl},
                                                       idx => $j);
            if (logCond("dbg")) {
                my $x_f = sprintf "0x%lx", $flags;
                my $x_m = sprintf "0x%lx", $mask;
#                logDbg("check_aces_acl: idx: $j type: $type flags: $x_f" .
#                       " mask: $x_m who: $who");
            }
            if ($type == ${opts{aces}}[$i]{type} &&
                $who eq ${opts{aces}}[$i]{who}) {
                # If the flag_cmp is set to yes then and the flags.
                # Check for an exact match on the mask.
                if (($opts{flag_cmp} eq 'no' ||
                    $flags & ${opts{aces}}[$i]{flags} ||
                    $flags == ${opts{aces}}[$i]{flags}) &&
                    $mask == ${opts{aces}}[$i]{mask}) {
                    $found = 1;
#                    logDbg("check_aces_acl: found idx: $j");
                    last;
                }
            }
        }
        if ($found == 0) {
            return (1);
        }
    }
    return (0);
}

## @name ::get_nfs_client_obj
## @description
##  Extracts the nfs::NfsClient object contained within the nfs::ACL::AclClass
##  object when the AclClass is based on the nfs::ACL::PerlSyntheticClient class.
##  The NfsClient object is a reference in the AclClass object. That reference
##  is returned if it exists.
##
## @arg
##
##  acl         Required: ACL that contains the desired NfsClient object.
##
## @return
##  Reference to NfsClient object, croaks on failure

sub get_nfs_client_obj( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT },
        }
    );

    # Verify that the AclClass object is using nfs::ACL::PerlSyntheticClient
    if (!$opts{acl}->isa("nfs::ACL::PerlSyntheticClient")) {
        croak "get_nfs_client_obj: ACL object is not using PerlSyntheticClient";
    }

    # Make sure an NfsClient object has been defined before trying to return one
    if (!defined $opts{acl}->{nfs_obj}) {
        croak "get_nfs_client_obj: ACL object does not have an nfs_obj definition";
    }

    return $opts{acl}->{nfs_obj};
}

## @name ::acl_to_synthetic_client_acl
## @description
##  Takes an nfs::ACL::AclClass object and converts it to a format that is
##  usable by nfs::NfsClient routines. The NfsClient procedures use 'flag'
##  for the flag bits on each ACE, where the default AclClass format is
##  to use 'flags' for the bits.
##
##  This procedure creates a clone of the AclClass object and returns the
##  reference to the clone. The Storable::dclone procedure cannot clone
##  the NfsClient portion of the AclClass object due to a CODE type inclusion.
##  To accomplish the clone, the NfsClient reference is temporarily removed
##  from the AclClass then added to both the AclClass and the clone after
##  the clone process is complete.
##
##  This process does not make sense unless the AclClass object is a
##  nfs::ACL::PerlSyntheticClient object as well. The function will
##  croak if it is not.
##
##  If the original AclClass object is already in NfsClient format or does not
##  contain any ACEs with the 'flags' key, an exact duplicate will be returned
##  without an error.
##
## @arg
##
##  acl         Required: ACL that is to be converted.
##
## @return
##  Reference to AclClass object with ACEs modified for NfsClient use,
##  croaks on failure

sub acl_to_synthetic_client_acl( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT },
        }
    );
    my ($clone, $nfs_clnt);

    # Verify that the AclClass object is using nfs::ACL::PerlSyntheticClient
    if (!$opts{acl}->isa("nfs::ACL::PerlSyntheticClient")) {
        croak "acl_to_synthetic_client_acl: ACL object is not using PerlSyntheticClient";
    }

    # Remove the nfs::NfsClient object from the ACL object, if it exists
    if (defined $opts{acl}->{nfs_obj}) {
        $nfs_clnt = get_nfs_client_obj(acl => $opts{acl});
        delete $opts{acl}->{nfs_obj};
    }

    # Deep clone the ACL object
    $clone = dclone($opts{acl});

    # Convert the 'flags' key in each ACE to 'flag'
    foreach (@{$clone->{ace}}) {
        next if (!exists $_->{flags});
        $_->{flag} = $_->{flags};
        delete $_->{flags};
    }

    # Put the NfsClient object back into both if it previously existed
    if (defined $nfs_clnt) {
        $opts{acl}->{nfs_obj} = $nfs_clnt;
        $clone->{nfs_obj} = $nfs_clnt;
    }

    return $clone;
}

## @name ::synthetic_client_acl_to_acl
## @description
##  Takes an nfs::ACL::AclClass object with ACEs in nfs::NfsClient format
##  and converts it to the default format of nfs::ACL::AclClass. The NfsClient
##  procedures use 'flag' for the flag bits on each ACE, where the default
##  AclClass format is to use 'flags' for the bits.
##
##  This procedure creates a clone of the NfsClient format ACL object and returns
##  the reference to the clone. The Storable::dclone procedure cannot clone
##  the NfsClient portion of the AclClass object due to a CODE type inclusion.
##  To accomplish the clone, the NfsClient reference is temporarily removed
##  from the AclClass then added to both the AclClass and the clone after
##  the clone process is complete.
##
##  This process does not make sense unless the AclClass object is a
##  nfs::ACL::PerlSyntheticClient object as well. The function will
##  croak if it is not.
##
##  If the original AclClass object is already in default format or does not
##  contain any ACEs with the 'flag' key, an exact duplicate will be returned
##  without an error.
##
## @arg
##
##  acl         Required: ACL that is to be converted.
##
## @return
##  Reference to AclClass object with ACEs modified for default AclClass use,
##  croaks on failure

sub synthetic_client_acl_to_acl( % ) {
    my %opts = validate(
        @_,
        {
            acl => { type => OBJECT },
        }
    );
    my ($clone, $nfs_clnt);

    # Verify that the AclClass object is using nfs::ACL::PerlSyntheticClient
    if (!$opts{acl}->isa("nfs::ACL::PerlSyntheticClient")) {
        croak "synthetic_client_acl_to_acl: ACL object is not using PerlSyntheticClient";
    }

    # Remove the nfs::NfsClient object from the ACL object, if it exists
    if (defined $opts{acl}->{nfs_obj}) {
        $nfs_clnt = get_nfs_client_obj(acl => $opts{acl});
        delete $opts{acl}->{nfs_obj};
    }

    # Deep clone the ACL object
    $clone = dclone($opts{acl});

    # Convert the 'flag' key in each ACE to 'flags'
    foreach (@{$clone->{ace}}) {
        next if (!exists $_->{flag});
        $_->{flags} = $_->{flag};
        delete $_->{flag};
    }

    # Put the NfsClient object back into both if it previously existed
    if (defined $nfs_clnt) {
        $opts{acl}->{nfs_obj} = $nfs_clnt;
        $clone->{nfs_obj} = $nfs_clnt;
    }

    return $clone;
}

1;
