# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/GenericExport.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::GenericExport;

## @summary
##  Provides an OS independent interface to exportfs.
##
## @description
##  This module handles an OS independent interface to the exportfs
##  command.  Creates an export object, according to the connection
##  object provided, and calls the appropriate OS dependent exportfs
##  function.
##
## @usage
##  use nfs::GenericExport;
##
## @status  public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes   Created by EP Network Storage Performance Lab, Inc.
## @modifications
## Aug 11, 2009 IST
##  Modified export_paths sub routine to export a path using anon=0 incase the access options doesn't include either -root option or anon option. (Burt #371060)
##  Modified by: Meharpraveen.Maddineni@netapp.com

## Aug 28, 2009 IST
##  Modified export_paths sub routine to export a path using anon=0 such that it preceeds all the options (Burt #375258)
##  Modified by: Meharpraveen.Maddineni@netapp.com

use strict;
use base qw(Exporter);

BEGIN {
    our @EXPORT_OK = qw(
                      &export_path
                      &export_paths
                      &unexport_path
                      &list_exports
                      &export_reset_persistent
                      &get_export_options);
};

use Carp qw(croak);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF);
use nfs::Export::ExportClass;
use Data::Dumper;
use SPARSE::Util::Logger;

our $VERSION = "1.0";

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub export_path($$;%);
sub unexport_path($$;%);
sub list_exports($;%);
sub export_reset_persistent($;%);
sub get_export_options($$;%);

#---------------------------------------------------------------------
# Netgroup state variable:
# This hash is used to keep track of the netgroups passed to exports.
# Some machines require an '@' as the first character of a netgroup,
# others the '@' is optional, and yet others cannot have '@'.
# In order to have a call to list_exports return the same access
# options given to export_path, it is required to keep track of the
# netgroups passed to export_path since, depending on the OS, the '@'
# will be removed running the export command. Therefore, list_exports
# must add the '@' character back to all netgroups specified.
#
# Note: This will keep the state of netgroups within a test only
# (while this module is in scope). If a test is run and exports a
# path to netgroup and another test is run to get the exports list,
# all netgroups specified in the first test could lose the '@'
# depending on the host type.
#---------------------------------------------------------------------
my %netgroup_h;

#---------------------------------------------------------------------
## @name ::export_path
##
## @description
##  Contains the generic front end to the OS dependent export command.
##  It handles the following:
##    1) Process options that are not passed to the OS dependent export.
##    2) Add default options to the list of options but do not override
##       user settings of these parameters.
##    3) Construct export class that has the OS specific info.
##    4) Call the OS specific export with generic options.
##
## @synopsis
##  $op = export_path($conn, $path [, %options]);
##
## @arg $conn
##  Required: Server's connection object.
##
## @arg $path
##  Required: Path to be exported.
##
## @arg OPTIONS
##  Optional: A string that contains a comma separated list of
##            access options, see NFS_EXPORT_OPTIONS.
##            Defaults to 'rw,anon=0'.
##
## @arg WRITE_ETC
##  Optional: Write changes to the exports file (default=0).
##
## @arg PERSISTENT
##  Optional: Same as WRITE_ETC (default=0).
##
## @arg PERSISTENT_OPT
##  Optional: Write changes to the exports file (default=0)
##            using the '-p' option (filer only). Same as
##            WRITE_ETC for all other platforms.
##
## @arg retry
##  Optional: Number of times to retry the command if the exportfs
##  command is busy (default = 5).
##
## @arg NFS_EXPORT_OPTIONS:
##
## @arg anon=uid
##  Set uid to be the effective user ID of unknown users.
##  By default, unknown users are given the effective user ID UID_NOBODY.
##  If uid is set to -1, access is denied.
##
## @arg nosuid
##  By default, clients are allowed to create files on the shared file
##  system with the setuid or setgid mode enabled.  Specifying nosuid
##  causes the server file system to silently ignore any attempt to
##  enable the setuid or setgid mode bits.
##
## @arg ro
##  Sharing will be read-only to all clients.
##
## @arg ro=list
##  Sharing will be read-only to the clients listed in 'list';
##  overrides the rw suboption for the clients specified.
##
## @arg root=list
##  Only root users from the hosts specified in 'list' will have
##  root access.  By default, no host has root access, so root users
##  are mapped to an anonymous user ID (see the anon=uid option
##  described above).  Netgroups can be used if the file system shared
##  is using UNIX authentication (AUTH_SYS).
##
## @arg rw
##  Sharing will be read-write to all clients.
##
## @arg rw=list
##  Sharing will be read-write to the clients listed in 'list';
##  overrides the ro suboption for the clients specified.
##
## @arg sec=mode
##  Sharing will use one or more of the specified security modes.
##  The mode in the sec=mode option must be a node name supported
##  on the client.  If the sec= option is not specified, the default
##  security mode used is AUTH_SYS.  Multiple sec= options can be
##  specified on the command line, although each mode can appear only
##  once.  The security modes are defined in nfssec(5).
##  This option must be specified before 'ro' or 'rw' for some versions
##  of the OS, so it will be moved to the beginning of list.
##
## @return
## 1 if succeeds else croaks with error in $@.
##
## @examples
##  $op = export_path($conn, '/vol/testvol', OPTIONS=>"rw,ro=client1");
##
##  $op = export_path($conn, '/vol/testvol', OPTIONS=>"rw", WRITE_ETC=>1);
##
## @status	review
##
## @author      corbin@netapp.com
## @notes       Created by EP Network Storage Performance Lab, Inc.
##
## @notes	Differences between GenericExport.pm and ExportFS.pm
##
## 1) If a warning is thrown by the export command it goes through in its
##    entirety; so if exporting a persistent path, the exports file is
##    modified and then the warning is returned.
## 2) This function returns eval for backwards compatibility with
##    ExportsFS.pm.
#---------------------------------------------------------------------
sub export_path($$;%) {
    my ($conn, $path, %opts) = @_;
    export_paths($conn, 
		 paths => { $path => {} },
		 %opts);
}

#---------------------------------------------------------------------
## @name ::export_paths
##
## @description
##  Contains the generic front end to the OS dependent export command.
##  It handles the following:
##    1) Process options that are not passed to the OS dependent export.
##    2) Add default options to the list of options but do not override
##       user settings of these parameters.
##    3) Construct export class that has the OS specific info.
##    4) Call the OS specific export with generic options.
##  
##  See export_path() for detailed doc on the args.
##
## @synopsis
##  $op = export_paths($conn, paths => $path_options_hashref,
##                          [ WRITE_ETC | PERSISTENT | PERSISTENT_OPT => 0|1,],
##                          [retry => 5],
##                     );
##
## @example
##  $op = export_paths($fc, paths => { "/vol/vol0" => {
##                                     },
##                                     "/vol/testvol" => {
##                                            OPTIONS => 'rw=@netgrp',
##                                     },
##                                   },
##                          OPTIONS => "rw,anon=0",
##                          WRITE_ETC => 1,
##                          retry => 5,
##                     );
##
## @author   gvg@netapp.com
##  
##
sub export_paths($$;%) {
    my ($conn, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
	    paths          => { type => HASHREF },
            OPTIONS        => { type => SCALAR,  default => "rw,anon=0" },
            WRITE_ETC      => { type => BOOLEAN, default => 0 },
            PERSISTENT     => { type => BOOLEAN, default => 0 },
            PERSISTENT_OPT => { type => BOOLEAN, default => 0 },
            retry          => { type => SCALAR,  default => 10 },
        }
    );
    my ($obj, $op, $persistent);

    # Get if this export should be persistent
    $persistent = $opts{WRITE_ETC} || $opts{PERSISTENT} || 
	( $opts{PERSISTENT_OPT} ? 2:0 );

    # Extra options
    my %nopts = ( retry => $opts{retry} );
    my $paths = $opts{paths};

    ## The following three lines were added to resolve burt 371060
    if (($opts{OPTIONS}!~/root=/) && ($opts{OPTIONS}!~/anon=/)){
    	$opts{OPTIONS}="anon=0,".$opts{OPTIONS};
    }

   ##Has to be removed
    $opts{OPTIONS} =~ s/\s//g;

    return eval {
        # Create export object
        $obj = nfs::Export::ExportClass::->new($conn);
	
	foreach my $path (keys %$paths) {
	    # Access the options reference whereever it was set (global|local)
	    my $options = $paths->{$path}->{OPTIONS} ? \$paths->{$path}->{OPTIONS} : \$opts{OPTIONS};
	    if ( $$options =~ /@/ and ! $obj->{netgrp_atsign} ) {
		# Save the access for this hostname:path
		# if a netgroup is in the list
		my $k = $conn->hostp()->hostname() . $path;
		$netgroup_h{$k} = $$options;
		
		# Get rid off '@' in netgroups
		$$options =~ s/([=:]-?)\@/$1/g;
	    }	    
	}
	
	my $export_all = (scalar keys %$paths > 1) ? 1 : 0;
	# When export_all is set, force persistent to 1
	$persistent = $export_all ? 1 : $persistent;

	# Export the path
	$op = $obj->export_paths( paths => $opts{paths},
				  OPTIONS => $opts{OPTIONS},
				  persistent => $persistent,
				  retry => $opts{retry},
				  export_all => $export_all,
				  );

        return $op;
    };
}


#---------------------------------------------------------------------
## @name ::unexport_path
##
## @description
##  Contains the generic front end to the OS dependent unexport command.
##  It handles the following:
##    1) Process options that are not passed to the OS dependent unexport.
##    2) Add default options to the list of options but do not override
##       user settings of these parameters.
##    3) Construct export class that has the OS specific info.
##    4) Call the OS specific unexport with generic options.
##
## @synopsis
##  $op = unexport_path($conn, $path [, %options]);
##
## @arg $conn
##  Required: Server's connection object.
##
## @arg $path
##  Required: Path to be unexported.
##
## @arg WRITE_ETC
##  Optional: Write changes to the exports file (default=0).
##
## @arg PERSISTENT
##  Optional: Same as WRITE_ETC (default=0).
##
## @arg EXISTENT
##  Optional: Only unexport the path if it is currently exported,
##  if the path is not exported return success (default=1).
##
## @arg retry
##  Optional: Number of times to retry the command if the exportfs
##  command is busy (default = 10).
##
## @return
## 1 if succeeds else croaks with error in $@.
##
## @examples
##  $op = unexport_path($conn, '/vol/testvol');
##
##  $op = unexport_path($conn, '/vol/testvol', WRITE_ETC=>1)
##
## @status	review
##
## @author      corbin@netapp.com
## @notes       Created by EP Network Storage Performance Lab, Inc.
##
## @notes	Differences between GenericExport.pm and ExportFS.pm
##
## 1) If a warning is thrown by the unexport command it goes through in its
##    entirety; so if unexporting a persistent path, the exports file is
##    modified and then the warning is returned.
## 2) This function returns eval for backwards compatibility with
##    ExportsFS.pm.
#---------------------------------------------------------------------
sub unexport_path($$;%) {
    my ($conn, $path, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR },
        ],
        {
            WRITE_ETC  => { type => BOOLEAN, default => 0 },
            PERSISTENT => { type => BOOLEAN, default => 0 },
            EXISTENT   => { type => BOOLEAN, default => 1 },
            retry      => { type => SCALAR,  default => 5 },
        }
    );
    my ($obj, $op, $persistent, $persist_only);

    return eval {
        # Get if the unexport should be persistent
        $persistent = $opts{WRITE_ETC} || $opts{PERSISTENT};

        # Create export object
        $obj = nfs::Export::ExportClass::->new($conn);

        if ( $opts{EXISTENT} ) {
            # Return successful if the path is not already exported
            my $exp_h = $obj->list_exports();
            return 1 unless ( $exp_h->{$path} or $persistent );
            # Flag to remove the persistent export only
            # Since the export is not in the exports table
            $persist_only = $exp_h->{$path} ? 0 : $persistent;
        }

        # Extra options
        my %nopts = ( retry => $opts{retry} );

        # Unexport the path
        $op = $obj->unexport_path($path, $persistent, $persist_only, %nopts);

        return $op;
    };
}

#---------------------------------------------------------------------
## @name ::list_exports
##
## @description
##  Contains the generic front end to the OS dependent list exports command.
##  It handles the following:
##    1) Construct export class that has the OS specific info.
##    2) Call the OS specific list exports.
##
## @synopsis
##  $access_h = list_exports($conn [, retry => $retry]);
##
## @arg $conn
##  Required: Server's connection object.
##
## @arg retry
##  Optional: Number of times to retry the command if the exportfs
##  command is busy (default = 5).
##
## @return
## A reference to a hash if succeeds else croaks with error in $@.
## The hash has the exported paths as the key and the access option
## for that path as the value.
##
## @examples
##  $access_h = list_exports($conn);
##
## @status	review
##
## @author      corbin@netapp.com
## @notes       Created by EP Network Storage Performance Lab, Inc.
##
## 1) This function returns eval for backwards compatibility with
##    ExportsFS.pm.
#---------------------------------------------------------------------
sub list_exports($;%) {
    my ($conn, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
            retry => { type => SCALAR, default => 5 },
        }
    );
    my ($obj, $access_h, $hostname, $op);

    return eval {
        # Create export object
        $obj = nfs::Export::ExportClass::->new($conn);

        # Get list of exports
        $access_h = $obj->list_exports(%opts);

        # Get name of server
        $hostname = $conn->hostp()->hostname();

        # Change all netgroups back to include '@'
        # Go through each path found on list
        foreach my $p ( keys %$access_h ) {
            my ($naccess, %net, %acc, $client);
            if ( $netgroup_h{"$hostname$p"} ) {
                # Path for this hostname is found on the
                # netgroup state hash
                $naccess = $netgroup_h{"$hostname$p"};

                # Break up access options for the netgroup
                # state hash, include only options with
                # netgroups in them to minimize search time.
                %net = map { my @a = split /=/;
                             # Break up hosts for access
                             # e.g.,  ro => [ c1, c2 ]
                             $a[0], [split /:/, $a[1]]
                             # Break up access, select only
                             # ones with netgroups in them
                           } grep { /\@/ } split /,/, $naccess;

                # Break up access options for the access
                # string for this path
                %acc = map { my @a = split /=/;
                             # Break up hosts for access
                             # e.g.,  ro => [ c1, c2 ]
                             $a[0], [split /:/, $a[1]]
                             # Break up access
                           } split /,/, $access_h->{$p};

                # Go through each access which has a netgroup
                foreach my $ac (keys %net) {
                    my $n = @{$net{$ac}} - 1;
                    # Go through each host in access
                    foreach my $i (0..$n) {
                        # Get client off the netgroup state hash
                        $client = $net{$ac}->[$i];
                        if ( $client =~ s/^(-?)@(.*)/$1$2/ ) {
                            # This is a netgroup
                            if ( $acc{$ac}->[$i] eq $client ) {
                                # There is a match, add '@' to netgroup
                                $acc{$ac}->[$i] =~ s/^(-?)(.*)/$1\@$2/;
                            } else {
                                # No match, not the right netgroup state
                                return $access_h;
                            }
                        }
                    }
                }

                # Put options back together
                undef $naccess;
                foreach my $ac (keys %acc) {
                    # Add access option
                    $naccess .= ",$ac";
                    # Add clients for this option if any
                    $naccess .= '=' . join(':', @{$acc{$ac}}) if @{$acc{$ac}};
                }
                # Get rid off leading ','
                $naccess =~ s/^,//;
                # Replace access option for path
                $access_h->{$p} = $naccess;
            }
        }

        return $access_h;
    };
}

#---------------------------------------------------------------------
## @name ::export_reset_persistent
##
## @description
##  Resets all persistent exports by re-exporting all persistent
##  exports only, all others are unexported, i.e., synchronizes
##  the in-memory exports table with the exports file.
##
## @synopsis
##  $op = export_reset_persistent($conn [, retry => $retry]);
##
## @arg $conn
##  Required: Server's connection object.
##
## @arg retry
##  Optional: Number of times to retry the command if the exportfs
##  command is busy (default = 5).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub export_reset_persistent($;%) {
    my ($conn, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
            retry => { type => SCALAR, default => 5 },
        }
    );
    my ($obj, $op);

    # Create export object
    $obj = nfs::Export::ExportClass::->new($conn);

    # Reset persistent exports
    $op = $obj->export_reset_persistent(%opts);

    return $op;
}

#---------------------------------------------------------------------
## @name ::get_export_options
##
## @description
##  Gets export options for the given path. First it gets the
##  options for a given path, then breaks and saves them into a hash.
##
## @synopsis
##   %opts = get_export_options($conn, $path [, retry => $retry]);
##
## @arg $conn
##  Required: Server's connection handle.
##
## @arg $path
##  Required: Path to get options for.
##
## @arg retry
##  Optional: Number of times to retry the command if the exportfs
##  command is busy (default = 5).
##
## @return
##  Hash of all options for path: access option is the key.
##
## @example
##  %opts = get_export_options($conn, '/vol/tempvol');
##
##     %opts will be { ro=>"ro=c1:c2:c3", rw=>"rw", root=>"root=c5" }
#---------------------------------------------------------------------
sub get_export_options($$;%) {
    my ($conn, $path, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR },
        ],
        {
            retry => { type => SCALAR, default => 5 },
        }
    );
    my (%opt, $exports);

    # Check exports table
    $exports = list_exports($conn, %opts);
    croak $@ if $@;
    # Create hash of options for path
    %opt = map { (split /=/)[0], $_ }       # Break up options
           split(/,/, $exports->{$path});   # Get options for path

    return %opt;
}

1;

