## @summary     Large_Reads - Routines for NFSv3 Large Reads
## @author      kartikr@netapp.com
## @status      public

package nfs::Large_Reads;

### Import Modules
use strict;
use warnings;

use base qw(Exporter);
use Tharn;
use Params::Validate qw(validate SCALAR HASHREF SCALARREF ARRAYREF);
use Carp qw(croak);
use nfs::NfsClient;
use nfs::NfsDef qw(FATTR4_MAXREAD FATTR4_MAXWRITE OPEN4_SHARE_ACCESS_WRITE);
use GenericMount qw(mount umount);

### Import NACL Modules

use NACL::C::Node;
use NACL::C::Cluster;
use NACL::C::VserverNfs;
use NACL::C::Volume;
use NACL::C::Aggregate;
use NACL::C::VolumeSnapshot;
use NACL::C::NetworkInterface;
use NACL::C::ClusterHa;
use NACL::C::StorageFailover;
use NACL::C::SystemNode;
use NACL::Transit;
use NACL::STask::Fiji;
use nfs::CommonParams;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
} ## end BEGIN

our @EXPORT_OK = qw(
    create_node_object
    get_nodes_in_cluster
    nfs_create
    nfs_delete
    nfs_modify
    nfs_show
    create_nfs_clients
    construct_vers_proto_hash 
    send_fsinfo 
    hash_validate
    vol_show
    aggr_show
    snapshot_create
    snapshot_restore
    snapshot_delete
    create_file
    send_read
    read_entire_file
    set_fiji
    clear_all_fiji
    send_write
    ha_show
    ha_modify
    sfo_modify
    interface_revert
    interface_show
    interface_migrate
    reboot_node
);

#---------------------------------------------------------------------
# Package global variables
#---------------------------------------------------------------------

# Fiji filters
my %fiji = ('mdatamodtime' =>  "nfs3.read.change.mdatamodtime(nil)(return 1)",
            'spinnp'       =>  "nfs3.read.spinnp.error(nil)(return 1)",
            'postopattrs'  =>  "nfs3.read.postopattrs.missing(nil)(return 1)");

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------

sub create_node_object(%);
sub get_nodes_in_cluster(%);
sub nfs_create(%);
sub nfs_delete(%);
sub nfs_modify(%);
sub nfs_show(%);
sub create_nfs_clients(%);
sub construct_vers_proto_hash(%); 
sub send_fsinfo(%); 
sub hash_validate(%);
sub vol_show(%);
sub aggr_show(%);
sub snapshot_create(%);
sub snapshot_restore(%);
sub snapshot_delete(%);
sub create_file(%);
sub send_read(%);
sub read_entire_file(%);
sub set_fiji(%);
sub clear_all_fiji(%);
sub send_write(%);
sub ha_show(%);
sub ha_modify(%);
sub sfo_modify(%);
sub interface_revert(%);
sub interface_show(%);
sub reboot_node(%);
sub interface_migrate(%);

#---------------------------------------------------------------------

sub create_node_object(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    filer   => { type => SCALAR }
                                }
                        );
    
    my($hostrec, $node_obj, $filer);
    
    $filer = $input_hash{filer};
    
    # Create a NACL Node Component Object
    
    $hostrec = host($filer);             
    $node_obj =  NACL::C::Node->new(node => $filer,
                                    hostrec => $hostrec); 
    $node_obj->apiset_must({ set => 'CMode' });
    
    return $node_obj 
}

sub get_nodes_in_cluster(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF }
                                }
                        );
    
    my ($nodeobj, @cluster_object, @cluster_nodes);
    
    $nodeobj = $input_hash{node_object};
    
    # Find out if cluster is proper and get all the nodes in the cluster
    
    @cluster_object = NACL::C::Cluster->find(command_interface => $nodeobj);
   
    foreach(@cluster_object) {
        push(@cluster_nodes,$_->node());
        if ($_->state()->health() ne 'true' || $_->state()->eligibility() ne 'true') {
            croak("Given Nodes are part of same cluster but node ".$_->node()."'s health or eligibility is false.");
        }
    }
    
    return (@cluster_nodes);
}

sub nfs_create (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR },
                                    fields        => { type => HASHREF, optional => 1}
                                }
                        );

    my($nodeobj, $vserver, $fields);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    $fields = $input_hash{fields};
    
    # Create nfs for the given vserver
    try {
        NACL::C::VserverNfs->create(
                                    command_interface => $nodeobj,
                                    vserver           => $vserver,
                                    %{$fields}
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("NFS create failed  - ".$ex->text);
    };    
}

sub nfs_delete (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR }
                                }
                        );

    my($nodeobj, $vserver);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    
    # Delete nfs for given vserver
    try {
        NACL::C::VserverNfs->delete(
                                    command_interface => $nodeobj,
                                    vserver           => $vserver
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("NFS delete failed  - ".$ex->text);
    };    
}

sub nfs_modify (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR },
                                    fields        => { type => HASHREF}
                                }
                        );

    my($nodeobj, $vserver, $fields);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    $fields = $input_hash{fields};
    
    # Modify nfs for the given vserver
    try {
        NACL::C::VserverNfs->modify(
                                    command_interface => $nodeobj,
                                    vserver           => $vserver,
                                    %{$fields}
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("NFS modify failed  - ".$ex->text);
    };    
}

sub nfs_show (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR },
                                    fields        => { type => HASHREF, optional => 1}
                                }
                        );
    
    my($nodeobj, $vserver, $fields, $nfs_object);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    $fields = $input_hash{fields};
    
    try {
        if($fields) {
            $nfs_object = NACL::C::VserverNfs->find(
                                                    command_interface => $nodeobj,
                                                    filter => { 
                                                                vserver    => $vserver,
                                                                %{$fields}
                                                              }    
                                                );
        } 
        else {
            $nfs_object = NACL::C::VserverNfs->find(
                                                    command_interface => $nodeobj,
                                                    filter => { 
                                                                vserver           => $vserver
                                                              }    
                                                );
        }
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("NFS show failed  - ".$ex->text);
    };    
    
    # Check if function is called and assigned to scalar. If yes, then return the nfs CLI parameters.
    if((wantarray) == 0){
        return $nfs_object->state();
    }
    
}

sub create_nfs_clients(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    hostname   => {type => SCALAR},
                                    version    => {type => ARRAYREF, default => [3,4,'4.1']},
                                    v3_proto   => {type => ARRAYREF, default => ['tcp','udp']}
                                }
                        );
    my ($nfs_clients, $version_ref, $v3_proto_ref, $hostname, $minvers);
    
    $hostname = $input_hash{hostname};
    $version_ref = $input_hash{version};
    $v3_proto_ref = $input_hash{v3_proto};
    
    foreach(@$version_ref) {
        if ($_ == 3) {
            foreach(@$v3_proto_ref) {
                $nfs_clients->{3}->{$_} = nfs::NfsClient->new(hostname => $hostname, version => 3, proto => $_);
            }
        }
        else {
            $minvers = 0;
            if($_ eq '4.1') {
                $minvers = 1;    
            }
            $nfs_clients->{$_}->{'tcp'} = nfs::NfsClient->new(hostname => $hostname, version =>4, proto =>'tcp', minor_version => $minvers);
        }
    }
    return $nfs_clients;
}

sub construct_vers_proto_hash(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    nfs_params   => {type => HASHREF},
                                    nfs_clients  => {type => HASHREF}
                                }
                        );

    my ($nfs_params, $nfs_clients, $version, $proto, $new_hash);
    
    $nfs_params = $input_hash{nfs_params};
    $nfs_clients = $input_hash{nfs_clients};

    foreach $version (3,4,'4.1') {
        if($version == 3) {
            foreach $proto ('tcp','udp') {
                if($proto =~ /tcp/) {
                    if(defined($nfs_clients->{$version}->{$proto})) {
                        $new_hash->{$version}->{$proto}->{rtmax} = $nfs_params->{v3_tcp_max_read_size};
                        $new_hash->{$version}->{$proto}->{wtmax} = $nfs_params->{v3_tcp_max_write_size};
                    }
                }
                else {
                    if(defined($nfs_clients->{$version}->{$proto})) {
                        $new_hash->{$version}->{$proto}->{rtmax} = $nfs_params->{udp_max_xfer_size}; 
                        $new_hash->{$version}->{$proto}->{wtmax} = $nfs_params->{udp_max_xfer_size}; 
                    }
                }
            }
        }
        else {
            if(defined($nfs_clients->{$version}->{'tcp'})) {
                $new_hash->{$version}->{'tcp'}->{rtmax} = $nfs_params->{tcp_max_xfer_size}; 
                $new_hash->{$version}->{'tcp'}->{wtmax} = $nfs_params->{tcp_max_xfer_size};
            }
        }
    }

    return $new_hash;
}

sub send_fsinfo(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    nfs_clients  => {type => HASHREF},
                                    path         => {type => SCALAR}
                                }
                        );
    
    my ($nfs_clients, $path, $version, $proto, $fsinfo_hash, $out);
    
    $nfs_clients = $input_hash{nfs_clients};
    $path = $input_hash{path};
    
    foreach $version (3,4,'4.1') {
        if($version == 3) {
            foreach $proto ('tcp','udp') {
                if(defined($nfs_clients->{$version}->{$proto})) {
                    $nfs_clients->{$version}->{$proto}->mount(path => $path);
                    $out = $nfs_clients->{$version}->{$proto}->fsinfo();
                    $nfs_clients->{$version}->{$proto}->umount(path => $path);
                    $fsinfo_hash->{$version}->{$proto}->{rtmax} = $out->{rtmax};
                    $fsinfo_hash->{$version}->{$proto}->{wtmax} = $out->{wtmax};
                }                
            }
        }
        else {
            if(defined($nfs_clients->{$version}->{'tcp'})) {
                $nfs_clients->{$version}->{'tcp'}->mount(path => $path);
                $out = $nfs_clients->{$version}->{'tcp'}->getattr(attr => [FATTR4_MAXREAD(),FATTR4_MAXWRITE()]);
                $nfs_clients->{$version}->{'tcp'}->umount(path => $path);
                $fsinfo_hash->{$version}->{'tcp'}->{rtmax} = $out->{fattr}{FATTR4_MAXREAD()};
                $fsinfo_hash->{$version}->{'tcp'}->{wtmax} = $out->{fattr}{FATTR4_MAXWRITE()};
            }
        }
    }
    
    return $fsinfo_hash;
}

sub hash_validate(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    hash1  => {type => HASHREF},
                                    hash2  => {type => HASHREF}
                                }
                        );

    my ($hash1, $hash2, $key, $value);                        

    $hash1 = $input_hash{hash1};
    $hash2 = $input_hash{hash2};
    
    while (($key,$value) = each %$hash1) {
        if (ref($value) ne 'HASH') {
            if($value != $hash2->{$key}) {
                return 0;
            }
            next;
        }
        if(defined($hash2->{$key})){
            my $validation = hash_validate(hash1 => $hash1->{$key},hash2 => $hash2->{$key});
            if($validation == 0) {
                return 0;
            }
        }
        else {
            return 0;
        }
    }
    return 1;
}

sub vol_show(%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF},
                                    fields        => { type => HASHREF, optional => 1}
                                }
                        );
    
    my($nodeobj, $fields, @vol_objects, @vol_state_objects);
    
    $nodeobj = $input_hash{node_object};
    $fields = $input_hash{fields};
    
    try {
        if($fields) {
            @vol_objects = NACL::C::Volume->find(
                                                    command_interface => $nodeobj,
                                                    filter => { 
                                                                %{$fields}
                                                              }    
                                                );
        } 
        else {
            @vol_objects = NACL::C::Volume->find(
                                                    command_interface => $nodeobj,
                                                );
        }
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("vol show failed  - ".$ex->text);
    };    
    
    # Check if function is called and assigned to array. If yes, then return the nfs CLI parameters.
    if((wantarray) == 1){
        foreach(@vol_objects){
            push(@vol_state_objects,$_->state());
        }
        return @vol_state_objects;
    }

}

sub aggr_show(%) {
    my %input_hash = validate(
                                @_,
                                    {
                                        node_object   => { type => HASHREF},
                                        fields        => { type => HASHREF, optional => 1}
                                    }
                            );
    
    my($nodeobj, $fields, @aggr_objects, @aggr_state_objects);
    
    $nodeobj = $input_hash{node_object};
    $fields = $input_hash{fields};
    
    try {
        if($fields) {
            @aggr_objects = NACL::C::Aggregate->find(
                                                    command_interface => $nodeobj,
                                                    filter => { 
                                                                %{$fields}
                                                              }    
                                                );
        } 
        else {
            @aggr_objects = NACL::C::Aggregate->find(
                                                    command_interface => $nodeobj,
                                                );
        }
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("aggr show failed  - ".$ex->text);
    };    
    
    # Check if function is called and assigned to array. If yes, then return the nfs CLI parameters.
    if((wantarray) == 1){
        foreach(@aggr_objects){
            push(@aggr_state_objects,$_->state());
        }
        return @aggr_state_objects;
    }
}

sub snapshot_create (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    snapshot      => { type => SCALAR },
                                    volume        => { type => SCALAR },
                                    vserver       => { type => SCALAR }
                                }
                        );

    my($nodeobj, $snapshot, $volume, $vserver);
    
    $nodeobj = $input_hash{node_object};
    $snapshot = $input_hash{snapshot};
    $volume = $input_hash{volume};
    $vserver = $input_hash{vserver};
    
    # Create snapshot for the given volume
    try {
        NACL::C::VolumeSnapshot->create(
                                    command_interface => $nodeobj,
                                    snapshot          => $snapshot,
                                    volume            => $volume,
                                    vserver           => $vserver
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Snapshot create failed  - ".$ex->text);
    };    
}

sub snapshot_restore (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    snapshot      => { type => SCALAR },
                                    volume        => { type => SCALAR },
                                    vserver       => { type => SCALAR },
                                    'method-timeout' =>
                                        { type => SCALAR, optional => 1},
                                }
                        );

    my $nodeobj = $input_hash{node_object};
    delete $input_hash{node_object};

    # Restore snapshot for the given volume
    try {
        NACL::C::VolumeSnapshot->restore(
                                    command_interface => $nodeobj,
                                    %input_hash,
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Snapshot restore failed  - ".$ex->text);
    };
}

sub snapshot_delete (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    snapshot      => { type => SCALAR },
                                    volume        => { type => SCALAR },
                                    vserver       => { type => SCALAR }
                                }
                        );

    my($nodeobj, $snapshot, $volume, $vserver);
    
    $nodeobj = $input_hash{node_object};
    $snapshot = $input_hash{snapshot};
    $volume = $input_hash{volume};
    $vserver = $input_hash{vserver};
    
    # Delete snapshot for the given volume
    try {
        NACL::C::VolumeSnapshot->delete(
                                    command_interface => $nodeobj,
                                    snapshot          => $snapshot,
                                    volume            => $volume,
                                    vserver           => $vserver
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Snapshot delete failed  - ".$ex->text);
    };    
}

sub create_file(%) {
    my %input_hash = validate(
                              @_,
                                {
                                    junction_path   => { type => SCALAR },
                                    data_ip         => { type => SCALAR },
                                    file            => { type => SCALAR },
                                    file_size       => { type => SCALAR, default => '1048576'}
                                }
                        );
    
    my($junction_path, $contents, $data_ip, $file, $mountpoint, $fh, $file_size, @letters);                    
    
    $junction_path = $input_hash{junction_path};
    $data_ip = $input_hash{data_ip};
    $file = $input_hash{file};
    $file_size = $input_hash{file_size};
    @letters = ('a'..'z');
    foreach(1..$file_size) {
        $contents=$contents.$letters[int rand @letters];
    }
    
    # Mount the test volume on localhost to create the file.
    $mountpoint = mount($data_ip,
                        $junction_path,
                        mntopts => "vers=3",
                        chmod  => '0777') or croak "Mount error: $@";
    
    open($fh, '>', "$mountpoint/$file") or die "error opening $mountpoint/$file - $!";

    print $fh $contents;
    close($fh);
    umount($mountpoint);
    
    return $contents;
}

sub send_read (%) {
    my %input_hash = validate(
                              @_,
                                {
                                    nfs_clients     => { type => SCALARREF },
                                    offset          => { type => SCALAR },
                                    count           => { type => SCALAR },
                                    file              => { type => SCALAR },
                                    stateid         => { type => SCALAR, default => 0}
                                }
                        );
                        
    my ($nfs_clients, $offset, $count, $full_path, $junction_path, $file, $response, $stateid, $open);
    
    $nfs_clients = $input_hash{nfs_clients};
    $offset = $input_hash{offset};
    $count = $input_hash{count};
    $full_path = $input_hash{file};
    $stateid = $input_hash{stateid};
    
    ($junction_path,$file) = ($full_path =~ m!/(\w+)/(\w+)!);
    $junction_path = "/$junction_path";
    
    repeat:
    eval {
        $$nfs_clients->mount(path => $junction_path);
        if($$nfs_clients->{object}->{version} == 4) {
            $open = $$nfs_clients->open(name => $file,attr => {33 => 0777});
            $stateid = $open->{stateid};
        }
        else {
            $$nfs_clients->lookup(name => $file);
        }
        $response = $$nfs_clients->read(offset => $offset, count => $count, stateid => $stateid);
        if($$nfs_clients->{object}->{version} == 4) {
            $$nfs_clients->close(stateid => $stateid);
        }
        $$nfs_clients->umount(path => $junction_path);
    };
    
    # Large Reads lead to TCP window full. Hence reestablish the connection back.
    if($@) {
        if($@ =~ /Connection reset/) {
            $$nfs_clients->{object}->{rpc}->{sock}->close();
            $$nfs_clients = nfs::NfsClient->new(hostname => $$nfs_clients->{object}->{hostname}, version =>$$nfs_clients->{object}->{version}, proto =>$$nfs_clients->{object}->{proto});
            goto repeat;
        }
        else {
            croak("$@");
        }
    }
    
    return $response;
}    

sub read_entire_file (%) {
    my %input_hash = validate(
                              @_,
                                {
                                    nfs_clients     => { type => SCALARREF },
                                    max_read_size   => { type => SCALAR },
                                    file            => { type => SCALAR }
                                }
                        );

    my ($nfs_clients, $max_read_size, $file, $out, $full_contents, $offset, $returned_count);
                        
    $nfs_clients = $input_hash{nfs_clients};
    $max_read_size = $input_hash{max_read_size};
    $file = $input_hash{file};                    
    $offset = 0;    
    
    # Read entire file till eof is received
    
    do {
        $out = send_read(nfs_clients => $nfs_clients, offset => $offset, count => $max_read_size, file => $file);
        $full_contents = $full_contents.$out->{data};
        
        # READ response in NFSv3 returns count but NFSv4/NFSv4.1 does not
        if($$nfs_clients->{object}->{version} == 3) {
            $returned_count = $out->{count};
        }
        else {
            $returned_count = length($out->{data});
        }
        
        if ($returned_count != $max_read_size && $out->{eof} != 1) {
            croak("Read returned unexpected count $returned_count instead of $max_read_size");
        }
        $offset = $offset + $returned_count;
    } while ($out->{eof} != 1);
    
    # Remove extra padding bytes if any
    $full_contents = ($full_contents =~ /(\w+)/)? $1 : $full_contents;
    return $full_contents;
}

sub set_fiji (%) {
    my %input_hash = validate(
                              @_,
                                {
                                    node_object     => { type => HASHREF },
                                    repeat          => { type => SCALAR },
                                    filter          => { type => SCALAR },
                                    tag             => { type => SCALAR, default => 'default'}
                                }
                        );
    
    my ($node_object, $repeat, $filter, $tag);
                        
    $node_object = $input_hash{node_object};
    $repeat = $input_hash{repeat};
    $filter = $input_hash{filter};
    $tag = $input_hash{tag};
    
    # Set the given fiji filter
    try {
        NACL::STask::Fiji->set(
            command_interface => $node_object,
            tag => $tag,
            program => "\'(repeat $repeat ($fiji{$filter}))\'"
        );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Setting Fiji failed - ".$ex->text);
    };    
                        
}

sub clear_all_fiji (%) {
    my %input_hash = validate(
                              @_,
                                {
                                    node_object     => { type => HASHREF },
                                    slot            => { type => SCALAR, default => 'all'}
                                }
                        );
    
    my ($node_object, $slot);
                        
    $node_object = $input_hash{node_object};
    $slot = $input_hash{slot};
    
    # Clear the given slot of Fiji filter or all slots if no slot is specified. 
    try {
        NACL::STask::Fiji->clear(
            command_interface => $node_object,
            slot => $slot
        );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Clearing $slot slot Fiji failed - ".$ex->text);
    };    
}

sub send_write (%) {
    my %input_hash = validate(
                              @_,
                                {
                                    nfs_clients     => { type => HASHREF },
                                    offset          => { type => SCALAR },
                                    data            => { type => SCALAR },
                                    file              => { type => SCALAR },
                                    stateid         => { type => SCALAR, default => 0}
                                }
                        );
                        
    my ($nfs_clients, $offset, $data, $full_path, $junction_path, $file, $response, $stateid, $open, $fragment, $error);
    
    $nfs_clients = $input_hash{nfs_clients};
    $offset = $input_hash{offset};
    $data = $input_hash{data};
    $full_path = $input_hash{file};
    $stateid = $input_hash{stateid};
    
    ($junction_path,$file) = ($full_path =~ m!/(\w+)/(\w+)!);
    $junction_path = "/$junction_path";
    
    $fragment = $nfs_clients->{object}->{rpc}->{tcp_frg_size};
    $nfs_clients->{object}->{rpc}->{tcp_frg_size} = 100000;
    
    eval {
        if($nfs_clients->{object}->{minor_version} == 1) {
            $nfs_clients->mount(path => $junction_path, maxrequestsize => 100000, maxresponsesize => 100000);
        }
        else {
            $nfs_clients->mount(path => $junction_path);
        }
        if($nfs_clients->{object}->{version} == 4) {
            $open = $nfs_clients->open(name => $file, access=> OPEN4_SHARE_ACCESS_WRITE, attr => {33 => 0777});
            $stateid = $open->{stateid};
        }
        else {
            $nfs_clients->lookup(name => $file);
        }
        $response = $nfs_clients->write(offset => $offset, data => $data, stateid => $stateid);
        if($nfs_clients->{object}->{version} == 4) {
            $nfs_clients->close(stateid => $stateid);
        }
        $nfs_clients->umount(path => $junction_path);
    };
    if ($@) {
        $error = $@;
        $nfs_clients->{object}->{rpc}->{tcp_frg_size} = $fragment;
        croak("$error");
    }
    
    $nfs_clients->{object}->{rpc}->{tcp_frg_size} = $fragment;
    
    return $response; 
}

sub interface_show (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF},
                                    fields        => { type => HASHREF, optional => 1}
                                }
                        );
    
    my($nodeobj, $fields, @interface_objects, @interface_state_objects);
    
    $nodeobj = $input_hash{node_object};
    $fields = $input_hash{fields};
    
    try {
        if($fields) {
            @interface_objects = NACL::C::NetworkInterface->find(
                                                    command_interface => $nodeobj,
                                                    filter => { 
                                                                %{$fields}
                                                              }    
                                                );
        } 
        else {
            @interface_objects = NACL::C::NetworkInterface->find(
                                                    command_interface => $nodeobj,
                                                );
        }
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("interface show failed  - ".$ex->text);
    };    
    
    # Check if function is called and assigned to array. If yes, then return the interface CLI parameters.
    if((wantarray) == 1){
        foreach(@interface_objects){
            push(@interface_state_objects,$_->state());
        }
        return @interface_state_objects;
    }

}

sub interface_revert (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR },
                                    lif           => { type => SCALAR }
                                }
                        );

    my($nodeobj, $vserver, $lif);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    $lif = $input_hash{lif};
    
    # Revert the given lif
    try {
        NACL::C::NetworkInterface->revert(
                                    command_interface => $nodeobj,
                                    vserver           => $vserver,
                                    lif               => $lif
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Interface revert failed  - ".$ex->text);
    };    
}

sub ha_show (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF}
                                }
                        );
    
    my($nodeobj, $ha_object);
    
    $nodeobj = $input_hash{node_object};
    
    try {
        $ha_object = NACL::C::ClusterHa->find(
                                                command_interface => $nodeobj
                                              );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("ha show failed  - ".$ex->text);
    };    
    
    # Check if function is called and assigned to scalar. If yes, then return the ha CLI parameters.
    if((wantarray) == 0){
        return $ha_object->state();
    }
}

sub interface_migrate (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    vserver       => { type => SCALAR },
                                    lif           => { type => SCALAR },
                                    dest_node     => { type => SCALAR },
                                    fields        => { type => HASHREF, optional => 1}
                                }
                        );
                        
    my($nodeobj, $vserver, $lif, $dest_node, $fields);
    
    $nodeobj = $input_hash{node_object};
    $vserver = $input_hash{vserver};
    $lif = $input_hash{lif};
    $dest_node = $input_hash{dest_node};
    $fields = $input_hash{fields};
    
    # Revert the given lif
    try {
        if($fields) {
            NACL::C::NetworkInterface->migrate(
                                        command_interface => $nodeobj,
                                        vserver           => $vserver,
                                        lif               => $lif,
                                        'dest-node'       => $dest_node,
                                        %{$fields}
                                    );
        }
        else {
            NACL::C::NetworkInterface->migrate(
                                        command_interface => $nodeobj,
                                        vserver           => $vserver,
                                        lif               => $lif,
                                        'dest-node'       => $dest_node
                                    );
        }
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("Interface migrate failed  - ".$ex->text);
    };    
}

sub ha_modify (%) {
    my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    configured    => { type => SCALAR }
                                }
                        );

    my($nodeobj, $configured);
    
    $nodeobj = $input_hash{node_object};
    $configured = $input_hash{configured};
    
    # Modify HA
    try {
        NACL::C::ClusterHa->modify(
                                   command_interface    => $nodeobj,
                                   configured           => $configured
                                  );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("HA modify failed  - ".$ex->text);
    };    

}

sub sfo_modify (%) {
        my %input_hash = validate(
                            @_,
                                {
                                    node_object   => { type => HASHREF },
                                    node          => { type => SCALAR },
                                    fields        => { type => HASHREF}
                                }
                        );

    my($nodeobj, $node, $fields);
    
    $nodeobj = $input_hash{node_object};
    $node = $input_hash{node};
    $fields = $input_hash{fields};
    
    # Modify failover capability for the given node
    try {
        NACL::C::StorageFailover->modify(
                                    command_interface => $nodeobj,
                                    node           => $node,
                                    %{$fields}
                                );
    } catch NATE::BaseException with {
        my $ex = shift;
        croak("SFO modify failed  - ".$ex->text);
    };    
}

sub reboot_node (%) {
    my %input_hash = validate(
                        @_,
                            {
                                node_object   => { type => HASHREF },
                                node          => { type => SCALAR }
                            }
                    );
    
    my($nodeobj, $node, $transit_obj);
    
    $nodeobj = $input_hash{node_object};
    $node = $input_hash{node};
    
    # Reboot the given node
    try {
        NACL::C::SystemNode->reboot(
                                command_interface => $nodeobj,
                                node           => $node,
                                apiset_should => { connid => 'console' },
                                'method-timeout' => 5
                            );
    }catch NATE::BaseException with {
        my $ex = shift;
        if($ex->text =~ /Timeout/) {
            $transit_obj = NACL::Transit->new(name => $nodeobj->hostrec());
            $transit_obj->change_state(to => 'CLI');
            return;
        }
        croak("Reboot failed  - ".$ex->text);
    }; 
}

1;
