#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Locks.pm#3 $
#
## @summary Locks - Common functionality for testing NFS locks
## @author    dl-maryv (FSS Protocols QA)
## @status    public
## @notes    Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Locks - Common functionality for testing NFS locks

=head1 SYNOPSIS

  use nfs::Locks;

  my %setup_h;
  $out = nfslocks_setup(setup_h => \%setup_h,
                        clients => @clients,
                        mount => 1);
  $out = lock_check(fc => $fc, owner => { pid=>$$, uid=>1234 },
                    ipaddr => '10.0.0.25', type => 'read',
                    file   => '/vol/testvol/lock_file');
  $out = nsm_unreachable_make(fc => $fc,
                                clients => [{ ipaddr=>'10.0.0.4',
                                            status => 'noresp' },
                                            { ipaddr=>'10.0.0.6',
                                            status => 'unreach' }]);
  $out = name_services_unreachable_make(fc => $fc);
  $out = nfslocks_filer_state_restore(fc => $fc);
  nfslocks_cleanup(%setup_h);

=head1 DESCRIPTION

Provides a common functionality for testing NFS locks.
It includes functions for common setup and cleanup.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Locks.pm#3 $
  Last revised $Date: 2013/06/18 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Locks;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Rpc;
use Carp qw(croak);
use POSIX qw(uname);
use Tharn qw(host connect);
use CLI::FileIO qw(rdfile wrfile);
use GenericMount qw(mount umount);
use Network::Nis qw(get_nis_info);
use Network::Utils qw(run_getXXbyYY);
use nfs::Utils qw(validate_nfs_version);
use FileManip qw(write_file remove_file);
use TestEnv::MinWin qw(minwin_compatible);
use SPARSE::Util::Logger qw(logStep logWarn);
use nfs::RemoteFileIO qw(client_write_file);
use MapShare qw(map_cifs_share unmap_cifs_share);
use nfs::GenericExport qw(export_path list_exports);
use Params::Validate qw(validate UNDEF SCALAR OBJECT ARRAYREF HASHREF);
use FS::Util::Setup qw(create_vol common_cleanup);
use filer qw(filer_options_get filer_options_set filer_qtree_get
            filer_qtree_security_set filer_lock_status_file_get);
use CIFS::Setup_GX qw(cifs_setup);
use ExportFS qw(share_add share_delete);
use nfs::CommonParams;
use nfs_setup_gx;
use Server;
import Tharn qw(host connect param logcomment logresult);

my $f = host($main::FILER);

if ($f->hosttype !~ /filer-ng/m) 
{
    require CIFS::Ntfs;
}

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                        chomp(@msgs);
                        Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

our @EXPORT_OK = qw(
  nfslocks_setup
  nfslocks_cleanup
  lock_check
  nsm_unreachable_make
  name_services_unreachable_make
  nfslocks_filer_state_restore
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub nfslocks_setup(%);
sub nfslocks_cleanup(%);
sub lock_check(%);
sub nsm_unreachable_make(%);
sub name_services_unreachable_make(%);
sub nfslocks_filer_state_restore(%);
sub testme();

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my @file_locks;
my $export_list;
my $nfs_v4_id_domain;
my $nfs_v4_enable;
my $nfs_v4_rdel;
my $nfs_v4_wdel;
my $mt_root_only;
my %host_ipaddr_h;
my %nis_servers_h;
my %rc_contents_h;
my %hosts_contents_h;
my $hosts_file  = '/etc/hosts';
my $rc_file    = '/etc/rc';
my $dns_file    = '/etc/resolv.conf';
my $nsswitch    = '/etc/nsswitch.conf';
my $test_file   = "NFSlocks.$$";
my $nsm_program = 100024;
my $usermap = '/etc/usermap.cfg';
my $filer_usermap_text;
my $cifs_share_completed = 0;
my $qtree_sec_saved;

#---------------------------------------------------------------------

=head2 nfslocks_setup

  $out = nfslocks_setup(setup_h => \%setup_h,
                        [, clients => \@clients ]
                        [, cifs_clients => \@cifs_clients ]
                        [, nfs_allowed => \@allowed ]
                        [, nfs_version => 2|3|4|'all'|'not2'|'not3'|'not4' ]
                        [, lock_type => 'read'|'write'|'both' ]
                        [, mt_root_only => 'on'|'off' ]
                        [, local_host => 0|1 ]
                        [, mount => 0|1 ]);

Set up environment for NFS locks tests. It creates a volume and
a test file on that volume. All setup state is passed back to
the test on the 'setup_h' named option. This hash is populated
with the following:

  { fc            => Filer connection object,
    cc            => List of client connection objects,
    files        => List of files created by setup,
    volume        => Full path of created volume,
    local_ipaddr  => Local ip address as seen by the filer,
    lock_types    => List of lock types to test,
    cifs            => List of CIFS hashes having drive and host as keys,
    nfs_versions  => List of valid NFS versions,
    setup_mtpoint => Setup mountpoint }

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item clients

(Optional) : Array reference of NFS client hostspecs.

=item cifs_clients

(Optional) : Array reference of CIFS client hostspecs.

=item nfs_allowed

(Optional) : An array reference indicating what NFS versions are
allowed for this test. An error occurs if the nfs_version parameter
does not match one of these values. Default is to allow all versions.

=item nfs_version

(Optional) : NFS version(s) that will be tested. Acceptable values are
2, 3, 4, 'not2', 'not3', 'not4', and 'all'. Default is 'all' which will
test versions 2, 3 and 4.

=item local_host

(Optional) : If true names in client lists can be 'localhost': 0 or 1.
Default is 1.

=item lock_type

(Optional) : Lock type(s) that will be tested. Acceptable values are
'read', 'write', 'both'. Default is 'both'.

=item create_files

(Optional) : Create this number of files on test volume.
Default is to create one file.

=item mount

(Optional) : Mount the volume if true: 0 or 1. Mounting the volume
using the highest NFS version provided by the nfs_version parameter.
Default is to not mount the volume.

=item qtree_sec

(Optional) : Specify the qtree security style to use for FILER_VOL.
Default is to use 'unix'.

=item mt_root_only

(Optional) : Allow only privileged NFS mounts, i.e. only the root
user can mount the volume. Valid values are 'on' or 'off'.

=back

=item Returns

1 if succeeds else croaks with error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nfslocks_setup(%) {
    my (%opts) = validate (
        @_,
        {
            setup_h        => { type => HASHREF },
            clients        => { type => ARRAYREF,  default => ['localhost'] },
            cifs_clients => { type => ARRAYREF, optional => 1 },
            nfs_allowed  => { type => ARRAYREF,  default => [2, 3, 4] },
            nfs_version  => { type => SCALAR,    default => 'all' },
            local_host   => { type => SCALAR,    default => 1,
                            regex => qr/^[01]$/ },
            lock_type    => { type => SCALAR,    default => 'both',
                            regex => qr/^(read|write|both)$/ },
            create_files => { type => SCALAR,    default => 1,
                            regex => qr/^\d+$/ },
            mount        => { type => SCALAR,    default => 0,
                            regex => qr/^[01]$/ },
            qtree_sec    => { type => SCALAR, default => 'unix',
                            regex => qr/^(ntfs|mixed|unix)$/ },
            mt_root_only => { type => SCALAR,   optional => 1,
                            regex => qr/^(on|off)$/ },
        }
    );
    my ($cifs_passwd, $cifs_user, $fc, $fh, $fname, $vol_name, $mode,
        $mountpoint, @nfsver, %nfsver_h, $new_text, $op, %qtree_out);
    my $ng_check = 0;
    my $set_h = $opts{setup_h};

    logStep('BEGIN: nfslocks_setup: Setup environment for test');

    # Connect to filer
    $fh = host($main::FILER);
    $fname = $fh->hostname();
    $set_h->{fc} = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $set_h->{fc}->timeout($main::CMDTIMEOUT);

    if ($fh->hosttype =~ /filer-ng/m) {
        logStep ("In C-Mode setup");
        $ng_check = 1;
    }

    if ( defined($opts{nfs_version}) ) {
        logStep('Validate NFS version');
        @nfsver = validate_nfs_version(version => $opts{nfs_version},
                                        allowed => $opts{nfs_allowed});
        $set_h->{nfs_versions} = \@nfsver;
    }
    # Array of lock types to test
    $set_h->{lock_types} = $opts{lock_type} eq 'both' ? ['read', 'write']
                                                        : [ $opts{lock_type} ];

    # To simplify check for specific versions
    %nfsver_h = map { $_ => 1 } @nfsver;
    if ($ng_check == 0) {
    logStep('7G/7M specific setup');
    logStep('Save the current export options for volume');
    $export_list = list_exports($set_h->{fc});

    # Get localhost ip address as seen by the filer
    my ($local_name) = (uname())[1];
    $op = run_getXXbyYY($set_h->{fc}, 'gethostbyname_r', $local_name);
    ($set_h->{local_ipaddr}) = ($op =~ /addresses:\s+(\d+\.\d+\.\d+\.\d+)/);
    }
    logStep('Get client connection objects');
    my $cifs_cl = 0;
    foreach my $client ( @{$opts{clients}}, '--', @{$opts{cifs_clients}} ) {
        if ( $client eq '--' ) {
            # Now connecting to CIFS clients
            $cifs_cl = 1;
            next;
        }

        # Validate that clients can not be 'localhost'
        if ( ! $opts{local_host} and $client =~ /^localhost\b/ ) {
            croak "Clients can not be 'localhost'";
        }

        # Validate type of client
        my $fh = host($client);
        if ( $cifs_cl and $fh->hosttype() !~ /^window/ ) {
            croak "nfslocks_setup: $client is not a CIFS client";
        } elsif ( !$cifs_cl and $fh->hosttype() !~ /^unix/ ) {
            croak "nfslocks_setup: $client is not an NFS client";
        }

        # Get connection object for client
        push @{$set_h->{cc}}, connect('#rsh', '-timeout', $main::TIMEOUT);
        $set_h->{cc}[-1]->timeout($main::CMDTIMEOUT);
    }

    my ($data_server);
    if ($ng_check == 0) {
        logStep('7G/7M specific setup');
        # Common setup -- creates a volume depending
        # on the $main::SETUP_DONE parameter
        # The name of the volume to be created
        # is defined in $main::FILER_VOL
        logStep('Create test volume');
        ($set_h->{fc}, $vol_name) = create_vol() or croak "nfslocks_setup: $@";
        $fc = $set_h->{fc};

        # Full path for volume
        $set_h->{volume} = "/vol/$vol_name";

        # At this point the test has asked us to create the vol and the
        # vol already existed or did not exist and was created. Make sure
        # the qtree security style for FILER_VOL matches the requested
        # qtree security style.
        %qtree_out = filer_qtree_get($fc, vols => 1);

        if ($qtree_out{$main::FILER_VOL}{''}{Style} ne $opts{qtree_sec}) {
            logStep("Set qtree security mode $main::FILER_VOL to $opts{qtree_sec}");
            $qtree_sec_saved = $qtree_out{$main::FILER_VOL}{''}{Style};
            if (filer_qtree_security_set($fc, $set_h->{volume}, $opts{qtree_sec}) == 0) {
                croak("qtree set mode $opts{qtree_sec} failed on $set_h->{volume}");
            }
        }

        if ( @{$opts{cifs_clients}} ) {
            # Create username to use for cifs based on whether AD is setup or not
            # Default is workgroup
            $fname =~ tr/a-z/A-Z/;
            $cifs_user = "$fname\\$main::CIFS_USER";
            $cifs_passwd = "$main::CIFS_PASSWD";
            if ($main::CIFS_DOMAIN) {
                my $tmp = "$main::CIFS_DOMAIN";
                $tmp =~ s/\s+//g;
                if ($tmp ne '') {
                    # Using Domains
                    $cifs_user = "$main::CIFS_DOMAIN_ADMIN";
                    $cifs_passwd = "$main::CIFS_DOMAIN_ADMIN_PASSWD";
                }
            }

            logStep("nfslocks_setup: Setup $usermap file on $fname");
            $filer_usermap_text = rdfile($fc, $usermap);
            $new_text = "\n$cifs_user == root\n";
            wrfile($fc, $usermap, "$filer_usermap_text$new_text")
                or croak("nfslocks_setup: can't write $usermap: $@");

            logStep('Setup CIFS on the filer');
            if ($opts{qtree_sec} eq 'ntfs') {
                $mode = 'ntfs';
            } else {
                $mode = 'multi';
                # For an unknown reason, running a 'mixed' qtree test after
                # an 'ntfs' qtree test causes a mount problem on the chmod.
                # Running ntfs_cifs_setup twice seems to cure the problem,
                # so run a 'pre' ntfs_cifs_setup if qtree type is 'mixed'.
                ntfs_cifs_setup(filer => $main::FILER, mode => $mode)
                    if $opts{qtree_sec} eq 'mixed';
            }
            ntfs_cifs_setup(filer => $main::FILER,
            mode  => $mode);

            logStep('Create the share on the filer');
            if ($main::CIFS_SHARE !~  /^(etc\$|c\$|home)$/i) {
                # If $main::CIFS_SHARE is ETC$, C$, or HOME don't add it,
                # it already exists
                share_add($fc, $main::CIFS_SHARE, "/vol/$main::FILER_VOL", verify => 1, purge_existing => 1, umask => 0)
                    or croak("nfslocks_setup: failed to share $main::CIFS_SHARE");
                $cifs_share_completed = 1;
            }

            # Validate windows clients
            my @cifs_clients = @{$opts{cifs_clients}};
            foreach my $client ( @cifs_clients ) {
                # Make sure windows client has proper environment
                minwin_compatible($client) or croak "nfslocks_setup: $@";

                # Map volume on client
                $op = map_cifs_share(server => $fc->connhost(),
                        client => $client,
                        username => $cifs_user,
                        password => $cifs_passwd,
                        share  => $main::CIFS_SHARE,
                        drive  => $main::CIFS_DRIVE)
                    or croak "nfslocks_setup: $@";
                push @{$set_h->{cifs}}, { drive => $op, host => $client };
            }
        }

        if ( $nfsver_h{4} ) {
            # Verify NFS V4 ID domain is set on Filer
            logStep("Verify NFS V4 ID domain is set on Filer");
            $nfs_v4_id_domain = filer_options_get($fc, 'nfs.v4.id.domain');
            if (length($nfs_v4_id_domain) < 1) {
                croak('Filer does not have NFS V4 ID domain set');
            }
            # Save current NFSv4 state
            logStep('Save current NFSv4 state');
            $nfs_v4_enable = filer_options_get($fc, 'nfs.v4.enable');
            $nfs_v4_rdel   = filer_options_get($fc, 'nfs.v4.read_delegation');
            $nfs_v4_wdel   = filer_options_get($fc, 'nfs.v4.write_delegation');

            logStep('Enable NFSv4 on server');
            if ( $nfs_v4_enable eq 'off' ) {
                filer_options_set($fc, 'nfs.v4.enable', 'on');
            }
            if ( $nfs_v4_rdel eq 'off' ) {
                filer_options_set($fc, 'nfs.v4.read_delegation', 'on');
            }
            if ( $nfs_v4_wdel eq 'off' ) {
                filer_options_set($fc, 'nfs.v4.write_delegation', 'on');
            }
        }

        if ( defined($opts{mt_root_only}) ) {
            $mt_root_only = filer_options_get($fc, 'nfs.mount_rootonly');
            if ( $opts{mt_root_only} ne $mt_root_only ) {
                filer_options_set($fc, 'nfs.mount_rootonly', $opts{mt_root_only});
            }
        }

        logStep('Export test volume');
        export_path($fc, $set_h->{volume}, OPTIONS=>'rw,anon=0')
            or croak "nfslocks_setup: $@";

        $data_server = $fc->connhost();
    } else {
    ##################################################################
    # Call Cmode specific setup modules to make the node test-worthy
    ##################################################################
    logStep ("Starting Cmode setup");
            my %set_h;
            my $ref1;

    eval {
        $ref1 = nfs_setup_gx::set_env_gx( setup_h  => \%set_h,
                allowed  => $opts{nfs_allowed},
                version  => $opts{nfs_version},
                protocol => 'both' );
    };

    if ($@) {
        logStep ("Unable to setup filer");
        croak "nfslocks_setup: $@";
    }


    # The main nfs/cifs server to be addressed is
    $data_server = $ref1->{hostname};
     host($main::FILER);

    $set_h->{fc} = connect($main::FILER_CONN);
    #$set_h->{fc} = $ref1->{fc1};
    $fc = $set_h->{fc};
    # Full path for volume
    $set_h->{volume} = $ref1->{volume};
     $set_h->{data_server} = $ref1->{data_server};

    if ( @{$opts{cifs_clients}} ) {
        # Create username to use for cifs based on whether AD is setup or not
        # Default is workgroup
        $fname =~ tr/a-z/A-Z/;
        $cifs_user = "$fname\\$main::CIFS_USER";
        $cifs_passwd = "$main::CIFS_PASSWD";
        if ($main::CIFS_DOMAIN) {
            my $tmp = "$main::CIFS_DOMAIN";
            $tmp =~ s/\s+//g;
            if ($tmp ne '') {
                # Using Domains
                $cifs_user = "$main::CIFS_DOMAIN_ADMIN";
                $cifs_passwd = "$main::CIFS_DOMAIN_ADMIN_PASSWD";
            }
        }

        if ($main::SETUP_DONE == 0){
        logStep("nfslocks_setup: Setup $usermap file on $fname");

        logStep('Setup CIFS on the filer');
        cifs_setup(node => $main::FILER,
                    cifs_server => $main::CIFS_SERVER,
                    cifs_domain_fullname => $main::CIFS_DOMAIN_FULLNAME,
                    cifs_domain_shortname=> $main::CIFS_SHORT_DOMAIN,
                    ad_server => $main::AD_SERVER,
                    ad_server_ip => $main::AD_SERVER_IP,
                    vserver => $::VSERVER_GX_NAME,
                    ad_server_user => $main::AD_USERNAME,
                    ad_server_password => $main::AD_PASSWORD,
                    rootshare => $main::CIFS_SHARE,
                    );
        }
    }
    # Validate windows clients
    my @cifs_clients = @{$opts{cifs_clients}};
    foreach my $client ( @cifs_clients ) {
    # Make sure windows client has proper environment
    minwin_compatible($client) or croak "nfslocks_setup: $@";

    # Map volume on client
        print "cifs domain is $main::CIFS_SHORT_DOMAIN\n";
    $cifs_user = "$main::CIFS_SHORT_DOMAIN\\$main::CIFS_USER";
    $cifs_passwd = "$main::CIFS_PASSWORD";print "passwd is $cifs_passwd\n";

    $op = map_cifs_share(server => $data_server,
                        client => $client,
                        username => $cifs_user,
                        password => $cifs_passwd,
                        share  => $main::CIFS_SHARE,
                        drive  => $main::CIFS_DRIVE)
    or croak "nfslocks_setup: $@";
    push @{$set_h->{cifs}}, { drive => $op, host => $client };
        }
         $set_h->{data_server}=$data_server;
    }

    logStep('Mount test volume');
    $mountpoint = mount($data_server,
                        $set_h->{volume},
                        client => $set_h->{cc}[0],
                        mntopts => "vers=$nfsver[-1]",
                        chmod  => '0777') or croak "nfslocks_setup: $@";

    # Save mountpoint so if setup should fail
    # it will get properly unmounted
    $set_h->{setup_mtpoint} = $mountpoint;

    logStep('Create test files') if $opts{create_files};
    foreach my $index ( 1..$opts{create_files} ) {
        my $tfile = "$test_file.$index";
        my $inode = ( write_file(client    => $set_h->{cc}[0],
                                file    => "$mountpoint/$tfile",
                                fstat    => 1,
                                contents=> join('', ('A'..'Z', "\n"))x9) )[1]; 
        push @{$set_h->{files}}, { file => $tfile, inode => $inode };
    }

    if ( ! $opts{mount} ) {
        logStep('Unmount test volume');
        umount($mountpoint, client=>$set_h->{cc}[0])
            or croak "nfslocks_setup: $@";

        # Already unmounted, no need to unmount it
        # anymore at nfslocks_cleanup if setup should fail
        delete $set_h->{setup_mtpoint};
    }

    logStep('END: nfslocks_setup');
    return 1;
}

#---------------------------------------------------------------------

=head2 nfslocks_cleanup

  nfslocks_cleanup([ fc => $fc ]
                    [ cc => \@cc ]
                    [, volume => $volume ]
                    [, local_ipaddr => $ipaddr ]
                    [, files => \@files ]
                    [, lock_types => \@lock_types ]
                    [, nfs_versions => \@nfsvers ]
                    [, cifs => \@cifs_drives ]
                    [, setup_mtpoint => $mtpoint]);

Clean up environment for NFS locks tests. It destroys the volume and
removes any files given in list. It restores the NFSv4 state.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Optional) : Filer's connection object.

=item cc

(Optional) : Array reference of client connection objects.

=item volume

(Optional) : Absolute path of volume.

=item files

(Optional) : Array reference of files to remove.
Each entry can either be a file name or a hash reference having
the file name on the 'file' key.
Files names are relative to the 'volume' option.

=item lock_types

(Optional) : Array reference of lock types,

=item nfs_versions

(Optional) : Array reference of NFS versions tested.

=item cifs

(Optional) : Array reference of CIFS drives to unmapped.
Each element is a hash having 'drive' and 'host' as keys.

=item setup_mtpoint

(Optional) : The mount point of volume while in setup.

=item local_ipaddr

(Optional) : Local ip address as seen by the filer,

=back

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub nfslocks_cleanup(%) {
    my (%opts) = validate (
        @_,
        {
            fc            => { type => UNDEF|OBJECT,   optional => 1 },
            cc            => { type => UNDEF|ARRAYREF, optional => 1 },
            files        => { type => UNDEF|ARRAYREF, optional => 1 },
            lock_types    => { type => UNDEF|ARRAYREF, optional => 1 },
            nfs_versions  => { type => UNDEF|ARRAYREF, optional => 1 },
            cifs            => { type => UNDEF|ARRAYREF, optional => 1 },
            volume        => { type => UNDEF|SCALAR,   optional => 1 },
            local_ipaddr  => { type => UNDEF|SCALAR,   optional => 1 },
            setup_mtpoint => { type => UNDEF|SCALAR,   optional => 1 },
        }
    );
    my $mountpoint = $opts{setup_mtpoint};

    logStep("CLEANUP == $main::CLEANUP");

    eval {
        logStep('Re-connect to filer');
        host($main::FILER);
        $opts{fc} = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
        $opts{fc}->timeout($main::CMDTIMEOUT);

        if ( $main::CLEANUP ne 'none' ) {
            eval {
                logStep('Restore filer boot up state');
                eval { nfslocks_filer_state_restore(fc=>$opts{fc}) };
                logWarn("nfslocks_cleanup: $@") if $@;

                if ( $opts{volume} and $opts{cc} ) {
                    if ( ! $mountpoint ) {
                        logStep('Export test volume');
                        export_path($opts{fc}, $opts{volume},
                                    OPTIONS=>'rw,anon=0') or croak $@;

                        logStep('Mount test volume');
                        $mountpoint = mount($opts{fc}->connhost(),
                                            $opts{volume},
                                            client => $opts{cc}[0],
                                            mountretry => 2) or croak $@;
                    } else {
                        logStep('Volume already mounted');
                    }

                    if ( $opts{files} ) {
                        my $file;
                        logStep('Remove created files');
                        foreach ( @{$opts{files}} ) {
                            $file = ref($_) eq 'HASH' ? $_->{file} : $_;
                            eval { remove_file(client => $opts{cc}[0],
                                                file   => "$mountpoint/$file") };
                            logWarn("nfslocks_cleanup: $@") if $@;
                        }
                    }
                }
            };

            logWarn("nfslocks_cleanup: $@") if $@;

            if ($opts{fc} && $filer_usermap_text) {
                logStep("nfslocks_cleanup: Restore $usermap file on filer");
                wrfile($opts{fc}, $usermap, $filer_usermap_text)
                    or logWarn("nfslocks_cleanup: can't write $usermap: $@");
            }

            if ( $opts{cifs} ) {
                logStep('Unmap volume on CIFS clients');
                foreach my $cifs ( @{$opts{cifs}} ) {
                    unmap_cifs_share(mappath => $cifs->{drive},
                                    client  => $cifs->{host},
                                    force => 1);
                    logWarn("nfslocks_cleanup: $@") if $@;
                }
            }

            if ($opts{fc} && $cifs_share_completed != 0) {
                logStep("nfslocks_cleanup: Unsharing the CIFS volume");
                eval {
                    share_delete($opts{fc}, $main::CIFS_SHARE);
                };
                logWarn("nfslocks_cleanup: $@") if $@;
            }

            # Restore filer options
            if ( defined($nfs_v4_enable) ) {
                logStep('Restore NFSv4 state');
                eval { filer_options_set($opts{fc}, 'nfs.v4.enable',
                                        $nfs_v4_enable) };
                logWarn("nfslocks_cleanup: $@") if $@;
            }
            if ( defined($nfs_v4_rdel) ) {
                eval { filer_options_set($opts{fc}, 'nfs.v4.read_delegation',
                                        $nfs_v4_rdel) };
                logWarn("nfslocks_cleanup: $@") if $@;
            }
            if ( defined($nfs_v4_wdel) ) {
                eval { filer_options_set($opts{fc}, 'nfs.v4.write_delegation',
                                        $nfs_v4_wdel) };
                logWarn("nfslocks_cleanup: $@") if $@;
            }
            if ( defined($mt_root_only) ) {
                eval { filer_options_set($opts{fc}, 'nfs.mount_rootonly',
                                        $mt_root_only) };
                logWarn("nfslocks_cleanup: $@") if $@;
            }

            if ($opts{fc} && defined($qtree_sec_saved)) {
                logStep("Restore qtree security mode $main::FILER_VOL to " .
                        "$qtree_sec_saved");
                if (filer_qtree_security_set($opts{fc}, $opts{volume},
                    $qtree_sec_saved) == 0) {
                    logWarn("nfslocks_cleanup: qtree set mode " .
                            "$qtree_sec_saved failed on $opts{volume}");
                }
            }
        }

        # Run common cleanup and log a warning in order
        # to continue and restore the export options for volume
        logStep('Run common cleanup');
        common_cleanup($opts{fc}, client=>$opts{cc}[0], mnt=>$mountpoint)
            or logWarn("nfslocks_cleanup: $@");

        if ( $main::CLEANUP ne 'none' and $opts{volume} ) {
            my $access;
            if ( defined($export_list) and $export_list->{$opts{volume}} ) {
                $access = $export_list->{$opts{volume}};
            } else {
                $access = 'sec=sys,rw,nosuid';
            }
            logStep('Restore export options for volume');
            export_path($opts{fc}, $opts{volume}, OPTIONS=>$access)
                or logWarn("nfslocks_cleanup: $@");
        }
    };

    logWarn("nfslocks_cleanup: $@") if $@;
    logStep('END: CLEANUP');
    return;
}

#---------------------------------------------------------------------

=head2 lock_check

  $out = lock_check(fc => $fc,
                    owner => { pid=>$pid, uid=>$uid },
                    ipaddr => $ipaddr,
                    type => 'read'|'write'
                    [, file   => $file ]
                    [, inode   => $inode ]
                    [, sbyte  => $sbyte ]
                    [, blength => $blength ]
                    [, buffer => 0|1 ]
                    [, partner => 0|1 ]
                    [, verify_owner => 0|1 ]);

Checks if specific lock state exists on the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Filer's connection object.

=item owner

(Required) : Hash reference of lock owner having the following keys:
'pid' and 'uid'.

=item ipaddr

(Required) : IP address of owner of lock.

=item type

(Required) : Check for specific type of lock: 'read' or 'write'.

=item file

(Optional) : Absolute path name of file.

=item inode

(Optional) : Inode of file.

=item sbyte

(Optional) : Starting byte of range lock. Default is 0.

=item blength

(Optional) : Number of bytes of range lock.
Default is 0 (to the end of file).

=item buffer

(Optional) : Do not get locks from the filer, use the buffer instead.
This is useful to check for a series of locks without having to fetch
all locks for each check. Default is 0.

=item partner

(Optional) : Check for lock on partner in a cluster pair. Default is 0.

=item verify_owner

(Optional) : Check lock owner id on filer. Default is 1.

=back

=item Returns

1 if lock is found else croaks with error.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub lock_check(%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            owner     => { type => HASHREF },
            file      => { type => SCALAR, default => '' },
            client    => { type => OBJECT, default => 0 },
            ipaddr    => { type => SCALAR },
            type      => { type => SCALAR, regex => qr/^(read|write)$/ },
            inode     => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
            sbyte     => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            blength   => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            buffer    => { type => SCALAR, regex => qr/^[01]$/, default => 0 },
            partner   => { type => SCALAR, regex => qr/^[01]$/, default => 0 },
            lockstate => { type => SCALAR, default => "granted" },
            'verify_owner'    => { type => SCALAR, default => 1 }
        }
    );
    my ($lock_h, $inode, $owner, $out);
    my $ng_check = 0;

    # Hash to map lock type on filer to input types
    my %lock_type = ( 0=>'read', 1=>'write' );
    my %cifs_type = ( no=>'read', yes=>'write' );

    my $h = host ($main::FILER);
    if ($h->hosttype =~ /filer-ng/m) {
        $ng_check = 1;
    }

    if ($ng_check == 0) {

        # Get all locks on filer -- grouped by file inode
        if ( ! $opts{buffer} ) {
            @file_locks = filer_lock_status_file_get($opts{fc},
                                                    file    => $opts{file},
                                                    partner => $opts{partner});
        }
        # Set the correct inode if only the file parameter was passed
        if ( @file_locks and $opts{file} ne '' and !defined($opts{inode}) ) {
            $opts{inode} = $file_locks[0]{inode};
        }

        foreach $lock_h ( @file_locks ) {
            # Only process file locks
            next unless exists($lock_h->{sbyte});

            # Convert hostname to ipaddr for NLM locks
            # from the point of view of the filer
            if ( !exists($lock_h->{ipaddr}) ) {
                # Get the ipaddr of host from save copy
                if ( !($lock_h->{ipaddr} = $host_ipaddr_h{$lock_h->{host}}) ) {
                    # This is a new host, no ip address is saved,
                    # get the ipaddress from the filer
                    $out = run_getXXbyYY($opts{fc}, 'gethostbyname_r',
                                        $lock_h->{host});
                    ($out) = ($out =~ /addresses:\s+(\d+\.\d+\.\d+\.\d+)/);
                    $lock_h->{ipaddr} = $out;
                    # Save host and ipaddr for faster subsequent lookups
                    $host_ipaddr_h{$lock_h->{host}} = $out;
                }
            }

            # Get correct lock owner according to protocol
            $owner = $opts{owner}{($lock_h->{proto} eq 'NLM' ? 'pid' : 'uid')};
            $owner = '' unless defined($owner);

            # On CIFS ignore blength (whole file is locked with flock)
            my $ltype_test = $lock_h->{proto} eq 'CIFS' ?
                ( $cifs_type{$lock_h->{excl}} eq $opts{type} ) :
                ( $lock_type{$lock_h->{ltype}} eq $opts{type}
                and $lock_h->{blength} == $opts{blength} );
            # Check if this is the lock
            if ( $ltype_test
                and $lock_h->{state} eq 'GRANTED'
                and $lock_h->{inode} == $opts{inode}
                and $lock_h->{sbyte} == $opts{sbyte}
                and $lock_h->{ipaddr} eq $opts{ipaddr} ) { 
                    #Added additional condition to verify lock owner only if verify_owner field is set to 1.
                    if($opts{verify_owner}){
                            if($lock_h->{owner} eq $owner) {
                                return 1;
                            }
                        }else{
                            return 1;
                        }
            }

        }

        # Lock not found
        if(! defined wantarray) {
            $inode = defined($opts{inode}) ? sprintf(' %d <0x%x>', $opts{inode}, $opts{inode}) :croak "Could not find $opts{type} lock for " ."inode$inode on client $opts{ipaddr}";
            croak "Could not find $opts{type} lock for " ."inode$inode on client $opts{ipaddr}";
        } else {
            return 0;
        }

    } else {
        logStep ("Check locks in Cmode context");
        my $ret2 = gx_lock_check(%opts);

        if ($ret2 == 1) {
            return 1;
        } else {
            return 0;
        }
    }
}

#---------------------------------------------------------------------

=head2 nsm_unreachable_make

  $out = nsm_unreachable_make(fc => $fc, clients => \@clients);

Modifies the contents of the initialization file to make all clients
listed unreachable/non-responsive/delay-response by NSM the next time
the filer is rebooted.

It saves the current state of the files it modifies if this function
is the first to modify those specific files, so this function
is compatible with name_services_unreachable_make. In order to restore
the original state, use function nfslocks_filer_state_restore.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Filer's connection object.

=item clients

(Required) : Array reference of clients and their respective status
to set. Each entry in the array is a hash having the following keys:
ipaddr and status. The status key can be any of the following values:
'unreach', 'noresp', 'delresp'.

=back

=item Returns

1 if succeeds else croaks with error.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub nsm_unreachable_make(%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            clients => { type => ARRAYREF },
        }
    );
    my ($new_rc, $client, $ipaddr, $udp, $tcp, $err);
    my $fc_ipaddr = $opts{fc}->connhost();
    my $rule = 100;

    # Read current contents of the /etc/rc file
    if ( !defined($rc_contents_h{$fc_ipaddr}) ) {
        $rc_contents_h{$fc_ipaddr} = rdfile($opts{fc}, $rc_file)
            or croak "Could not read $rc_file file on filer";
    }

    # Enable ipfw (firewall)
    $new_rc = $rc_contents_h{$fc_ipaddr} . "options ipfw.enable on\n" .
                                            "priv set -q test\n";

    # Make each client unreachable, no response, delay response
    foreach $client ( @{$opts{clients}} ) {
        $ipaddr = $client->{ipaddr};
        # Get the udp port number of the status monitor
        $udp = eval { nfs::Rpc::pmap_getport(program => $nsm_program,
                                            host    => $ipaddr,
                                            version => 1,
                                            proto   => 'udp') };
        $err = $@;
        # Get the tcp port number of the status monitor
        $tcp = eval { nfs::Rpc::pmap_getport(program => $nsm_program,
                                            host    => $ipaddr,
                                            version => 1,
                                            proto   => 'tcp') };
        $err .= $@;
        croak "Status monitor is not enabled on $ipaddr: $err"
            unless ( $udp or $tcp );

        if ( $client->{status} eq 'unreach' ) {
            $new_rc .= "ipfw add $rule deny udp from any to $ipaddr $udp\n"
                if $udp;
            $new_rc .= "ipfw add $rule deny tcp from any to $ipaddr $tcp\n"
                if $tcp;
        } elsif ( $client->{status} eq 'noresp' ) {
            $new_rc .= "ipfw add $rule deny udp from $ipaddr $udp to any\n"
                if $udp;
            $new_rc .= "ipfw add $rule deny tcp from $ipaddr $tcp to any\n"
                if $tcp;
        } else {
            #XXX delay response is not implemented yet
        }
        $rule += 10;
    }

    # Write the new rc file
    wrfile($opts{fc}, $rc_file, $new_rc)
    or croak "Could not write $rc_file file on filer";

    return 1;
}

#---------------------------------------------------------------------

=head2 name_services_unreachable_make

  $out = name_services_unreachable_make(fc => $fc);

Modifies the contents of the initialization file to make all name
services unreachable the next time the filer is rebooted.

It saves the current state of the files it modifies if this function
is the first to modify those specific files, so this function
is compatible with nsm_unreachable_make. In order to restore the
original state, use function nfslocks_filer_state_restore.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Filer's connection object.

=back

=item Returns

1 if succeeds else croaks with error.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub name_services_unreachable_make(%) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    my ($ns_contents, $ns_search, $ns_type, @nis_servers, $out);
    my (@server_list, $server, $new_rc, $local_name, $local_ipaddr);
    my $fc_ipaddr = $opts{fc}->connhost();
    my $rule = 100;

    # Get list and order of name services
    $ns_contents = rdfile($opts{fc}, $nsswitch)
        or croak "Could not read $nsswitch file on the filer";
    ($ns_search) = ($ns_contents =~ /\bhosts:\s+(.*)/);

    # Go through list of name services for hosts
    foreach $ns_type ( split ' ', $ns_search ) {
        if ( $ns_type eq 'nis' ) {
            # Get list of NIS servers
            my $nis_info_h = get_nis_info($opts{fc});
            push @nis_servers, keys %{$nis_info_h};
            push @server_list, @nis_servers;
            # Save nis.servers option in case it is a '*'
            $nis_servers_h{$fc_ipaddr} = filer_options_get($opts{fc},
                                                            'nis.servers');
            # Replace option by the list of nis servers obtain with
            # nis info so only these servers are used by the filer
            filer_options_set($opts{fc}, 'nis.servers',
                                join(',', @nis_servers));
        } elsif ( $ns_type eq 'dns' ) {
            # Get list of DNS servers
            my $dns_info = rdfile($opts{fc}, $dns_file)
                or croak "Could not read $dns_file file on the filer";
            foreach ( split /\n/, $dns_info ) {
                ($server) = (/^nameserver\s+(.*)/);
                push @server_list, $server if $server;
            }
        }
    }

    # Get local host name
    ($local_name) = (uname())[1];
    push @server_list, $local_name;

    # Change all hostnames on list to ip addresses
    foreach $server ( @server_list ) {
        if ( $server !~ /^\d+\.\d+\.\d+\.\d+$/ ) {
            # Find ip address of server as seen by the filer
            $out = run_getXXbyYY($opts{fc}, 'gethostbyname_r', $server);
            ($out) = ($out =~ /addresses:\s+(\d+\.\d+\.\d+\.\d+)/);
            croak "Could not find ip address for server $server" unless $out;
            # Replace server name by its ip address
            $server = $out;
        }
    }

    # Get the local ipaddr from the list
    $local_ipaddr = pop @server_list;
    if ( !defined($hosts_contents_h{$fc_ipaddr}) ) {
        $hosts_contents_h{$fc_ipaddr} = rdfile($opts{fc}, $hosts_file)
            or croak "Could not read $hosts_file file on filer";
    }

    # Add the local host to the filer's hosts file
    # This is done since all name services will be unreachable
    # so any connections from the local host which is running
    # the test will be denied and the test won't be able to
    # talk to the filer even if it is running.
    wrfile($opts{fc}, $hosts_file, $hosts_contents_h{$fc_ipaddr} .
                                    "$local_ipaddr $local_name\n")
    or croak "Could not write $hosts_file file on filer";

    # Read current contents of the /etc/rc file
    if ( !defined($rc_contents_h{$fc_ipaddr}) ) {
        $rc_contents_h{$fc_ipaddr} = rdfile($opts{fc}, $rc_file)
            or croak "Could not read $rc_file file on filer";
    }

    # Enable ipfw (firewall)
    $new_rc = $rc_contents_h{$fc_ipaddr} . "options ipfw.enable on\n" .
                                            "priv set -q test\n";

    # Make each server unreachable
    foreach $server ( @server_list ) {
        $new_rc .= "ipfw add $rule deny ip from any to $server\n";
        $rule += 10;
    }

    # Write the new rc file
    wrfile($opts{fc}, $rc_file, $new_rc)
    or croak "Could not write $rc_file file on filer";

    return 1;
}

#---------------------------------------------------------------------

=head2 nfslocks_filer_state_restore

  $out = nfslocks_filer_state_restore(fc => $fc);

Restores the filer state changed by functions nsm_unreachable_make
and name_services_unreachable_make.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : Filer's connection object.

=back

=item Returns

1 if restore succeeds else croaks with error.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub nfslocks_filer_state_restore(%) {
    my (%opts) = validate (
        @_,
        {
            fc => { type => OBJECT },
        }
    );
    # Get the ip address of filer, so the
    # restore is done for that specific filer
    my $fc_ipaddr = $opts{fc}->connhost();

    if ($opts{fc}->{hosttype} =~ /ng/) {
        return;
    }

    # Disable firewall so all requests are completed after
    # the hosts file is restore to its original contents
    filer_options_set($opts{fc}, 'ipfw.enable', 'off');

    # Check if the nis.servers options has been modified
    if ( defined($nis_servers_h{$fc_ipaddr}) ) {
        filer_options_set($opts{fc}, 'nis.servers', $nis_servers_h{$fc_ipaddr});
        undef $nis_servers_h{$fc_ipaddr};
    }

    # Check if the rc file has been modified
    if ( defined($rc_contents_h{$fc_ipaddr}) ) {
        # Restore the contents of the file
        wrfile($opts{fc}, $rc_file, $rc_contents_h{$fc_ipaddr})
            or croak "Could not restore the contents of the $rc_file file";
        undef $rc_contents_h{$fc_ipaddr};
    }

    # Check if the hosts file has been modified
    if ( defined($hosts_contents_h{$fc_ipaddr}) ) {
        # Restore the contents of the file
        wrfile($opts{fc}, $hosts_file, $hosts_contents_h{$fc_ipaddr})
            or croak "Could not restore the contents of the $hosts_file file";
        undef $hosts_contents_h{$fc_ipaddr};
    }

    return 1;
}

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Locks SUBNAME=testme FILER=myfiler \
    util/test_pkg.thpl

=over

=item Arguments

=over

=item FILER

(Required) : Filer to test.

=back

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    require nfs::CommonParams;
    require Tharn;
    import Tharn qw(host connect param logcomment logresult);
    use vars qw($FILER);
    logcomment("Performing NFS Locks self test");

    param('FILER', '-mesg', 'Filer to test');

    my (%setup_h, $fc, $out, $mtpoint, $file, $inode, $err);

    # Setup test
    eval { nfslocks_setup(setup_h => \%setup_h, mount => 1) };
    $err = $@;
    logresult('FAIL', "nfslocks_setup failed: $err") if $err;
    if ( !$err ) {
        # Get a lock on a file
        $mtpoint = $setup_h{setup_mtpoint};
        $file    = $setup_h{files}[0]{file};
        $inode   = $setup_h{files}[0]{inode};
        if ( open(FH, ">$mtpoint/$file") ) {
            if ( flock(FH, 2) ) {
                # Check for lock status on filer
                $out = eval { lock_check(fc    => $setup_h{fc},
                                        owner  => { uid=>$> },
                                        ipaddr => $setup_h{local_ipaddr},
                                        inode  => $inode,
                                        type   => 'write') };
                logresult('FAIL', "lock_check failed: $@") if !$out;
            }
            close(FH);
        }
    }

    # Connect to filer
    host($main::FILER);
    $fc = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $fc->timeout($main::CMDTIMEOUT);

    # Make list of clients unreachable by NSM
    my @clients = (
        { ipaddr=>'127.0.0.1', status => 'unreach' },
        { ipaddr=>'127.0.0.1', status => 'noresp' },
    );
    $out = eval { nsm_unreachable_make(fc => $fc, clients => \@clients) };
    logresult('FAIL', "nsm_unreachable_make failed: $@") if !$out;

    # Make name services unreachable on the filer
    $out = eval { name_services_unreachable_make(fc => $fc) };
    logresult('FAIL', "name_services_unreachable_make failed: $@") if !$out;

    # Restore changes done by nsm_unreachable_make
    # and name_services_unreachable_make
    $out = eval { nfslocks_filer_state_restore(fc => $fc) };
    logresult('FAIL', "nfslocks_filer_state_restore failed: $@") if !$out;

    # Cleanup test
    eval { nfslocks_cleanup(%setup_h) };
    logresult('FAIL', "nfslocks_cleanup failed: $@") if $@;

    # Do not log a PASS result if there has been a problem
    foreach my $result ( Tharn::own_results() ) {
        return if ( $result->severity() > 2 );
    }

    logresult('PASS', "SELF TEST of nfs/Locks.pm");
}

sub gx_lock_check (%) {
    my (%opts) = validate (
        @_,
        {
            fc        => { type => OBJECT },
            owner     => { type => HASHREF },
            file      => { type => SCALAR, default => '' },
            client    => { type => OBJECT, default => 0 },
            ipaddr    => { type => SCALAR },
            type      => { type => SCALAR, regex => qr/^(read|write)$/ },
            inode     => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
            sbyte     => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            blength   => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            buffer    => { type => SCALAR, regex => qr/^[01]$/, default => 0 },
            partner   => { type => SCALAR, regex => qr/^[01]$/, default => 0 },
            lockstate => { type => SCALAR, default => "granted" },
            'verify_owner' => { type => SCALAR, default => 1 }
        }
    );

    if ($opts{type} eq 'read') {
        $opts{type} = "false";
    } else {
        $opts{type} = "true";
    }
    my $granted_lock = 0;

    my @temp = split (/\//, $opts{file});
    my $lock_file = $opts{file};
    my $server_obj = Server->new(node => $main::FILER);

    if ($main::NFS_VERSION != 4 ) { 
        my $pid = sprintf("%x",$opts{owner} ->{pid});
        my @locks;
        my $owner_id="*";    #Added new variable $owner_id to decide owner-id based on value set in field 'verify_owner'
        if ($opts{verify_owner}){
            $owner_id="$pid*";
        }

        my $client_addr = $opts{ipaddr};

        # Get client address from the client option
        if ( $opts{client} ) {
            if (UNIVERSAL::isa($opts{client}, "Connectrec")) {
                my $hostrec = $opts{client}->hostp();
                if ($hostrec->IP_1()) {
                    my $client_name = $hostrec->hostname();
                    $client_addr = $hostrec->IP_1();
                    logStep("gx_lock_check(): client name $client_name");
                }
            }
        };

	logStep("IP Address used by gx_lock_check(): $client_addr");
        eval {

	    $server_obj->cli( cmd => "vserver locks show -vserver $::VSERVER_GX_NAME -volume $::FILER_GX_VOL -instance");
            @locks = $server_obj->vserver_locks_show ( vserver => $::VSERVER_GX_NAME,
                                                       volume => $::FILER_GX_VOL,
                                                       path => $lock_file,
                                                       protocol => "nlm",
                                                       #"owner-id" => $owner_id,  # Passing owner id value based on $owner_id
                                                       "client-address" => $client_addr,
                                                       "lock-state" => "$opts{lockstate}",
                                                       "bytelock-offset" => $opts{sbyte},
                                                       "bytelock-length" => $opts{blength},
                                                       "bytelock-exclusive" => $opts{type}
                                                      );
        };
        my $num_locks = @locks;

        if ($@ || $num_locks != 1) {
            logStep ("There is no granted lock");
        } else {
            logStep ("There is a granted lock for file $lock_file");
            $granted_lock = 1;
        }
    } else {
        my @locks;
        eval {
            @locks = $server_obj->vserver_locks_show ( vserver => $::VSERVER_GX_NAME,
                                                       volume => $::FILER_GX_VOL,
                                                       path => $lock_file,
                                                       'lock-state'=> "granted",
                                                       'bytelock-offset'=>$opts{sbyte},
                                                       'bytelock-length'=>$opts{blength},
                                                       'bytelock-exclusive'=>$opts{type},
                                                      );
        };
        if ($@) {
            logStep("Could not get the lock info");
        }
        foreach my $loks (@locks) {
            if ((($loks->{'lock-state'}) =~ m/granted/) &&
                (($loks->{'bytelock-offset'}) == $opts{sbyte}) &&
                (($loks->{'bytelock-length'}) == $opts{blength}) &&
                (($loks->{'bytelock-exclusive'}) eq $opts{type})) {
                logStep ("There is a granted lock for file $lock_file");
                $granted_lock = 1;
            }
        }
    }
    return $granted_lock;
}

1;

