# this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/Migration.pm#1 $
#
## @summary	Migration routines - NFS client v4.x
## @author	dl-svl-nfsqa (NFS QA Sunnyvale)
## @status	public
## @notes	khanhn@netapp.com

package nfs::Migration;

use warnings;
use SPARSE::Util::Logger;
use Data::Dumper;
use Tharn qw(host $Log param);

use base qw {nfs::NfsClient};
use Params::Validate qw( validate SCALAR ARRAYREF );
use nfs::NfsXdr qw( xdr_bitmap4 xdr_compound xdr_compound_res 
    xdr_fhandle xdr_getattr xdr_opaque xdr_uint
);
use nfs::NfsDef qw( bitmap4_convert
    FATTR4_TYPE FATTR4_SIZE FATTR4_FSID FATTR4_FILEID
    FATTR4_MODE FATTR4_NUMLINKS FATTR4_OWNER FATTR4_OWNER_GROUP
    FATTR4_RAWDEV FATTR4_TIME_ACCESS FATTR4_TIME_METADATA FATTR4_TIME_MODIFY
    OP_GETATTR OP_GETFH OP_LOOKUP OP_PUTFH OP_PUTROOTFH OP_SEQUENCE
);


#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
#
#---------------------
# homebrew khanhn 11/11/10
#---------------------
#
sub ref_getattr($%);
sub mig_getfh($%);
sub mig_putfh($%);

# code for getattr
sub ref_getattr($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
	    fhandle => { type => SCALAR,   default => $self->curr_fh() },
	    name    => {type => SCALAR, optional => 1},
	    attr    => { type => ARRAYREF, default => [ FATTR4_TYPE,
                                                        FATTR4_SIZE,
                                                        FATTR4_FSID,
                                                        FATTR4_FILEID,
                                                        FATTR4_MODE,
                                                        FATTR4_NUMLINKS,
                                                        FATTR4_NO_TRUNC,
                                                        FATTR4_OWNER,
                                                        FATTR4_OWNER_GROUP,
                                                        FATTR4_RAWDEV,
                                                        FATTR4_TIME_ACCESS,
                                                        FATTR4_TIME_METADATA,
                                                        FATTR4_TIME_MODIFY,
                                                        #FATTR4_LAYOUT_BLKSIZE
                                                      ] },
        }
    );

    my $out;
    my $rpc = $self->rpc();
    my $ver = $self->version();
    my $min_ver = 0;

    # Convert filehandle of object into XDR format
    $Log->debug(qq{calling xdr_fhandle with: $opts{fhandle}\n});
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    if ( $ver == 4 || $ver == 1 ) {
        # Convert the list of attributes
        my $attr = bitmap4_convert(attrlist => $opts{attr});
	my @ops;

    	if (!$opts{name})
    	{
            if ($self->minor_version() == 0) {
                @ops = ( { op => OP_PUTFH,   args => $fhandle },
                         { op => OP_GETATTR, args => xdr_bitmap4(encode=>$attr) } );
            } else {
                $min_ver = 1;
                my $seq_arg = $self->_sequence();
                @ops = ( { op => OP_SEQUENCE, args => $seq_arg },
                         { op => OP_PUTFH,   args => $fhandle },
                         { op => OP_GETATTR, args => xdr_bitmap4(encode=>$attr) } );
	    }
	} else { 
		if ($self->minor_version() == 1) {
                    $min_ver = 1;
                    my $seq_arg = $self->_sequence();
                    @ops = ( { op => OP_SEQUENCE, args => $seq_arg } );
                }

		# process name agrs for getattr
		my @namelist = split('/',$opts{name});

		my %first_op;
		my $first_element;

		if ($namelist[0] eq "") {
			push @ops, {op => OP_PUTROOTFH };
			$first_element = 1;
		} else {
			push @ops, {op => OP_PUTFH, args => $fhandle};
			$first_element = 0;
		}

		my $elem;
		foreach $elem (@namelist[$first_element .. $#namelist]) {
			push @ops, {op => OP_LOOKUP, args => xdr_opaque(encode=>$elem)};
		}
		push @ops, {op => OP_GETATTR, args => xdr_bitmap4(encode => $attr)};
	}

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
				arg => xdr_compound(	tag => 'getattr', 
							minor => $min_ver,
							ops => \@ops),

                               out => \&xdr_compound_res);
    } else {

        # Call correct procedure on server
        $out = $rpc->clnt_call(proc => NFSPROC_GETATTR,
                               arg  => $fhandle,
                               out  => sub { xdr_getattr(version => $ver, @_) });
    }

    # Check for errors
    $self->error_check(name => 'GETATTR', result => $out);

    # Create a copy of the fattr4 reference for ease of use
    if ( $ver == 4 || $ver == 1 ) {
        if ($min_ver  == 0) {
            $out->{fattr} = $out->{results}[1]{result}{fattr}{attr_vals};
        } else {
            $out->{fattr} = $out->{results}[2]{result}{fattr}{attr_vals};
        }
    }

    return $out;

} #end sub ref_getattr

sub mig_getfh($%) {
    my $self = shift;
    my (%opts) = validate ( @_,
        {
            fhandle => { type => SCALAR, default => $self->curr_fh() },
        }
    );
    my $out;
    my $rpc = $self->rpc();
    my $ver = $self->version();
    my $min_ver = 0;

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode=>$opts{fhandle}, version=>$ver);

    if ( $ver == 4 || $ver == 1 ) {
        # Create compound request
	my @ops;
	if ($self->minor_version() == 0) {
                @ops = ( { op => OP_PUTFH,   args => $fhandle },
                         { op => OP_GETFH } );
	} else {
		$min_ver = 1;
		my $seq_arg = $self->_sequence();
		@ops = ( { op => OP_SEQUENCE, args => $seq_arg },
			{ op => OP_PUTFH,   args => $fhandle },
                        { op => OP_GETFH } );
	}
        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg  => xdr_compound(tag => 'mig_getfh',
						    minor => $min_ver,
                                                    ops => \@ops),
                               out  => \&xdr_compound_res);
    }

    # Check for errors
    $self->error_check(name=>'GETFH', result=>$out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[-1]{result}{fhandle};

    return $out;
} #end sub mig_getfh

sub mig_putfh($%) {
    my $self = shift;
    my (%opts) = validate ( @_,
        {
            fhandle => { type => SCALAR, default => $self->curr_fh() },
	    # migrated fhanfle
            mhandle => { type => SCALAR, },
        }
    );
    my $out;
    my $rpc = $self->rpc();
    my $ver = $self->version();
    my $min_ver = 0;

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode=>$opts{fhandle}, version=>$ver);
    my $mhandle = xdr_fhandle(encode=>$opts{mhandle}, version=>$ver);

    if ( $ver == 4 || $ver == 1 ) {
        # Create compound request
        my @ops;
        if ($self->minor_version() == 0) {
	   @ops = ( { op => OP_PUTFH, args => $mhandle },
                    { op => OP_PUTFH, args => $fhandle } );
        } else {
	   $min_ver = 1;
	   my $seq_arg = $self->_sequence();
	   @ops = ( { op => OP_SEQUENCE, args => $seq_arg },
		    { op => OP_PUTFH, args => $mhandle },
                    { op => OP_PUTFH, args => $fhandle } );
        }
        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg  => xdr_compound(tag => 'mig_putfh',
                                                    minor => $min_ver,
                                                    ops => \@ops),
                               out  => \&xdr_compound_res);
    }

    # Check for errors
    $self->error_check(name=>'PUTFH', result=>$out);

    return $out;
} #end sub mig_putfh

1;
