## -*-perl-*x
##
## $Id: //depot/prod/DOT/dev/test/lib/nfs/NFSv4xCommonLib.pm $
##
## Copyright (c) 2013 Network Appliance, Inc.
## All rights reserved.
##
## @summary
##     Library functions for NFS Synthetic client scenarios
##
## @description
##     main                - Create Test Case Driver object.
##     init                - Common method to setup client and filer for thpl files
##                           execution.
##     uninit              - Method to undo the client and filer changes done in init().
##     cleanup             - Method to cleanup all the files created by the test case.
##     byte_range_lock     - Method to get a byte range lock.
##     byte_range_unlock   - Method to unlock the byte range lock taken by the test case.
##     open_file_mode      - Method to open a file in the access and deny modes as
##                           specified by the user.
##     create_client_object- Creates a client object, mounts the volume on the client
##                           and gets the current file handle.
##     check_lock          - Checks for byte range/share locks on the filer side.
##     verify_error_type   - Method to check whether the error code has the particular error message.
##     get_filer_object    - Method to create NACL filer object.
##     cleanup_file_objects- Unlocks and closes a file for a particular client object.
##     get_delegations_status - Method to get the delegations of v4.1 of vserver.
##     disable_delgations - Method to disable the delegations of v4.1 of vserver.
##     reset_delegations - Method to reset the delegations of v4.1 to its original values in vserver.
##
## @usage
##     use NFSv4xCommonLib;
##
## @author srivatsk@netapp.com,majidm@netapp.com
##
## @status private


use strict;
use TCD;
use vars qw($Test);
use Tharn qw($Log);
use Params::Validate qw(validate SCALAR);
use nfs::NfsDef qw(NFS4_OK UNCHECKED WRITE_LT FATTR4_MODE READ_LT );
use NATE::ParamSet;
use Tharn;
use Params::Validate qw (validate validate_pos SCALAR SCALARREF OBJECT ARRAYREF HASHREF);
use nfs_setup_gx qw{set_env_gx clean_env_gx};
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::VserverLocks;
use NACL::CS::VserverNFS;
use NACL::C::VserverNfs;
our @ISA = qw(TCD);
our %setup;

#---------------------------------------------------------------------
# Initialize and validate parameters
#---------------------------------------------------------------------
BEGIN {
    use vars qw(
    $FILER
    $FILER_GX_VOL
    $TESTCASES
    $NFS4_MINOR_VERSION
    $VSERVER_GX_NAME
    $NFS_VERSION
    $ADDR_TYPE
    );
};
#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
# These are needed for TCD
sub main();
sub init();
sub uninit();
sub cleanup();
sub open_file_mode();
sub byte_range_lock();
sub create_client_object();
sub check_lock();
sub verify_error_type();
sub get_filer_object();
sub cleanup_file_objects();
sub get_delegations_status();
sub disable_delgations();
sub reset_delegations();

#---------------------------------------------------------------------
# Run Tests
#---------------------------------------------------------------------

#---------------------------------------------------------------------
# main
#---------------------------------------------------------------------
our @Testcases;
#our @EXPORT = qw(main,$Test);

# Import the rest of the common parameters and their default values
use nfs::CommonParams;
use Carp qw(croak);
use SPARSE::Util::Logger;
use nfs::NfsClient;
use nfs::NfsDef;
use nfs::GenericExport qw(export_path unexport_path);


# Some common variables.

our $Test;
our $Curr_fh;
our $Client_object;
our $State_id;
our $Lock_result;
our $Vserver_data_ip;
our $Vol_junc_path;
my $Filename = "nfsv4_test_file";
my $Mount_result;
my $Length = 50;
my $Offset = 0;
our $Node;
our $read_deleg_4_1_status;
our $write_deleg_4_1_status;


#---------------------------------------------------------------------
# test_main
#---------------------------------------------------------------------
sub main() {
    # Debug break point
    $DB::single=2;

    #  Create Test Case Driver object
    $Test = new TCD(-testcases=>[@Testcases]);
    if ($Test->error){
            $Test->log($Test->errmsg);
            return $TCD::FAIL;
    }
    push (@{$Test->{options}}, '-rsc');

    # Performs method callbacks
    $Test->run_test();
    if ($Test->error){
            $Test->log($Test->errmsg);
            return $TCD::FAIL;
    }
    exit(0);
}

#--------------------------------------------------------------------------------
#
#=head2 init
#
#    init()
#
#=item Arguments
#
#This function doesnot take any arguments. It picks them from the global array.
#
#=back
#
#=item Returns
#
#This function aborts if any of its steps fails. otherwise it returns PASS value.
#
#=back
#
#--------------------------------------------------------------------------------

sub init() {

    $Test->description(  "Retrieve variables from the test harness and "
                       . "instantiate objects needed for test execution");
    my %setup_h;
    my $ret_h;

    #Set test specific defaults
    param('FILER', '-mesg', 'HostSpec of Cmode Filer');
    param('TESTCASES', '-default', 'all');
    param('FILER_GX_VOL','-default','testvol_flex');
    param('NFS4_MINOR_VERSION', '-default',    '0' );
    param('NFS_VERSION', '-default',    '4' );
    param('ADDR_TYPE', '-default', 'inet4' );

    eval {
        $ret_h = set_env_gx (setup_h => \%setup_h);
    };
    if ($@){
        logStep ("Setup failed: $@");
        return $TCD::CONFIG;
    }
     $Node = get_filer_object();

     if ($NFS4_MINOR_VERSION) {
              get_delegations_status();
              disable_delgations();
    }
    #Establish connection to filer
    host($FILER);

    $Vserver_data_ip = $ret_h->{hostname};
    $Vol_junc_path = $ret_h->{volume};

    $Test->log("The NFS Server Address selected is : $Vserver_data_ip");

    $Test->log("Creating the Client object");
    ($Client_object, $Mount_result, $Curr_fh) = create_client_object();
    return $TCD::PASS;
}

#--------------------------------------------------------------------------------
#
#=head2 cleanup
#
#    cleanup()
#
#The cleanup function as given in TCD format. It removes all the unnecessary files
#created and marked for deletion by the previous test.
#
#=item Arguments
#
#This function does not take any argument
#
#=back
#
#=item Returns
#
#This function returns a TCD::PASS.
#
#=back
#
#--------------------------------------------------------------------------------

sub cleanup () {

    $Test->description("Cleaning up the files created by testcase");
    my $close_status;
    my $remove_status;
    #Check if the hash returned from above statemetnt is a valid lock_state_id hash
    if(defined scalar(%$Lock_result->{lock_stateid})) {
        my $lock_state_id = $Lock_result->{lock_stateid};
        my $unlock_status = $Client_object->locku(
                                            offset => $Offset,
                                            count => $Length,
                                            stateid => $lock_state_id );
        $Test->nlog("Unlocked the file successfully");
    }
    #checking if the file is opened from client1 and has a valid stateid
    if(defined scalar(%$State_id)) {
        $close_status = $Client_object->close(stateid => $State_id);
        $Test->nlog("Closed the file successfully");
        $remove_status = $Client_object->remove(name => $Filename,
                                          dhandle => $Curr_fh );
        $Test->nlog("Deleted the file successfully");
    }
    $Lock_result = undef;
    $State_id = undef;
    return $TCD::PASS;
}

#--------------------------------------------------------------------------------
#
#=head2 uninit
#
#    uninit()
#
#=item Arguments
#
#This function doesnot take any argument
#
#=back
#
#=item Returns
#
#This function returns $TCD::PASS on completion.
#
#=back
#
#--------------------------------------------------------------------------------

sub uninit() {
    $Test->description("Uninit started");
    my $unmount;
    if( defined scalar(%$Mount_result)) {
        $unmount = $Client_object->umount(path => $Vol_junc_path);
        $Test->nlog("volume unmount was successfully");
    }
    if($NFS4_MINOR_VERSION) {
        reset_delegations();
    }

    clean_env_gx();
    $Test->log("Clean up successfully completed");

    return $TCD::PASS;
}

################################################################################
#                         User defined functions.
################################################################################
#--------------------------------------------------------------------------------
#
#=head2 get_delegations_status
#
#  get_delegations_status()
#
#=item Arguments
#
#This function takes no arguments
#
#=back
#
#=item Returns
#
#This function gets the status of the current delegations of v4.1 that are set on vserver.
#
#=back
#
#--------------------------------------------------------------------------------
sub get_delegations_status() {
    $Log->comment("Getting the delegations of v4.1");
    my $status;
    try {
        $status = NACL::CS::VserverNFS->fetch(  command_interface=>$Node,
                                    filter => {
                                    vserver=> $::VSERVER_GX_NAME, }
                                    );

    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->error("Unable to fetch the requested information from vserver");
    };
    $write_deleg_4_1_status = $status->v4_1_write_delegation;
    $read_deleg_4_1_status = $status->v4_1_read_delegation;
    $Log->comment("Getting the delegations of v4.1 was successfull");
}
#--------------------------------------------------------------------------------
#
#=head2 disable_delgations
#
#  disable_delgations()
#
#=item Arguments
#
#This function takes no arguments
#
#=back
#
#=item Returns
#
#This function disables the delegations of v4.1 on vserver.
#
#=back
#
#--------------------------------------------------------------------------------
sub disable_delgations() {
    $Log->comment('Disabling the delegations of v4.1');
    try {
        NACL::C::VserverNfs->modify( command_interface=>$Node,
                             vserver=>$::VSERVER_GX_NAME,
                            'v4.1-read-delegation' => 'disable',
                            'v4.1-write-delegation' => 'disable',);
    }
    catch NACL::Exceptions::NoElementsFound with  {
        $Log->error("Unable to modify the information in vserver");
    };
    $Log->comment("Disabling the delegations of v4.1 was successfull");
}
#--------------------------------------------------------------------------------
#
#=head2 reset_delegations
#
#  reset_delegations()
#
#=item Arguments
#
#This function takes no arguments
#
#=back
#
#=item Returns
#
#This function resets the delegations of v4.1 back to its orginal state in vserver.
#
#=back
#
#--------------------------------------------------------------------------------
sub reset_delegations() {
        $Log->comment('Resetting the delegations of v4.1');
        try {
            NACL::C::VserverNfs->modify( command_interface=>$Node,
                             vserver=> $main::VSERVER_GX_NAME,
                            'v4.1-read-delegation' => $read_deleg_4_1_status,
                            'v4.1-write-delegation' => $write_deleg_4_1_status,);
        }
        catch NACL::Exceptions::NoElementsFound with  {
            $Log->error("Unable to modify the information in vserver");
        };
        $Log->comment("Resetting the delegations of v4.1 was successfull");
    }
#--------------------------------------------------------------------------------
#
#=head2 open_file_mode
#
#  open_file_mode()
#
#=item Arguments
#
#This function takes synthetic client object, current file handler,
#                    access and deny modes as arguments.
#=back
#
#=item Returns
#
#This function returns open status object on completion.
#
#=back
#
#--------------------------------------------------------------------------------

sub open_file_mode($%;) {
    my (%opts) = validate (
        @_,
        {
            client_object => { type => HASHREF},
            curr_fh => { type => SCALAR},
            owner => { type => SCALAR,   default => $$ },
            access_mode => { type => SCALAR, regex => qr/^\d+$/ },
            deny_mode => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );
    my ($access_type_msg,$deny_type_msg) = get_access_deny_codes( access_type => $opts{access_mode},
                                              deny_type =>  $opts{deny_mode} );
    my $client_object = $opts{client_object};
    $Test->step("Opening the file \'$Filename\' with Share Modes :
    Share Access = \'$access_type_msg\' and Share Deny = \'$deny_type_msg\'");

    my $open_status = $client_object->open(
                                     dhandle => $opts{curr_fh},
                                     access => $opts{access_mode},
                                     name => $Filename,
                                     owner => $opts{owner},
                                     attr => {FATTR4_MODE() => 0777},
                                     cr_mode => UNCHECKED,
                                     deny => $opts{deny_mode});
    $Test->nlog("Open command executed successfully, status of the command is NFS4_OK");
    return $open_status;
}


#--------------------------------------------------------------------------------
#
#=head2 get_access_deny_codes
#
#  get_access_deny_codes()
#
#=item Arguments
#
#This function takes access and deny codes as arguments.
#
#=back
#
#=item Returns
#
#This function returns the strings of the corresponding access and deny codes as
#a tuple.
#
#=back
#
#--------------------------------------------------------------------------------
sub get_access_deny_codes($%;) {

    my $access_code_hash = { 1 =>'OPEN4_SHARE_ACCESS_READ',
                             2 =>'OPEN4_SHARE_ACCESS_WRITE',
                             3 =>'OPEN4_SHARE_ACCESS_BOTH'};

    my $deny_code_hash =  { 0 => 'OPEN4_SHARE_DENY_NONE',
                            1 => 'OPEN4_SHARE_DENY_READ',
                            2 => 'OPEN4_SHARE_DENY_WRITE',
                            3 => 'OPEN4_SHARE_DENY_BOTH'};
    my (%opts) = validate (
        @_,
        {
            access_type =>{ type => SCALAR, regex => qr/^\d+$/ },
            deny_type => {type => SCALAR, regex => qr/^\d+$/}
        }
    );

    return ($access_code_hash->{$opts{access_type}},$deny_code_hash->{$opts{deny_type}});

}

#--------------------------------------------------------------------------------
#
#=head2 byte_range_lock
#
#    byte_range_lock()
#
#=item Arguments
#
#This function takes synthetic client object, state id, offset , type,
#            owner name(e.g: $owner =>'O1') as arguments.
#
#=back
#
#=item Returns
#
#This function returns a lock result object.
#
#=back
#
#--------------------------------------------------------------------------------

sub byte_range_lock($%;){
    my (%opts) = validate (
        @_,
        {
            client_object => { type => HASHREF },
            offset        => { type => SCALAR,   regex => qr/^\d+$/ },
            type          => { type => SCALAR,   regex => qr/^\d+$/ },
            owner         => { type => SCALAR,   default => $$ },
            state_id      => { type => HASHREF },
            byte_length   => { type => SCALAR , default => $Length},
        }
    );
    my $lock_type_hash = {
                           1 =>'READ_LT',
                           2 =>'WRITE_LT',
                           3 =>'READW_LT',
                           4 =>'WRITEW_LT',
                         };
    $Test->step("Issue a \'$lock_type_hash->{$opts{type}}\' on the file \'$Filename\'");
    my $client_object = $opts{client_object};
    $Test->nlog("Applying lock\n offset : $opts{offset}\n No of bytes: $Length");
    my $lock_result = $client_object->lock(
                                        offset => $opts{offset},
                                        count => $opts{byte_length},
                                        type => $opts{type},
                                        stateid => $opts{state_id},
                                        owner => $opts{owner} );
    return $lock_result;
}

#--------------------------------------------------------------------------------
#
#=head2 create_client_object
#
#    create_client_object()
#
#=item Arguments
#
#This function takes no arguments
#
#=back
#
#=item Returns
#
#This function returns  client object , mount object and current file handel.
#
#=back
#
#--------------------------------------------------------------------------------
sub create_client_object(){
    my $mnt_result;
    my $client_obj = nfs::NfsClient->new(
                                        hostname=>$Vserver_data_ip,
                                        proto=>'tcp',
                                        addr_type => $ADDR_TYPE,
                                        version=>$NFS_VERSION,
                                        minor_version=> $NFS4_MINOR_VERSION );

    $mnt_result = $client_obj->mount(path => $Vol_junc_path);
    $Test->nlog("Volume mounted Successfull");
    my $cur_fh = $mnt_result->{fhandle };

    return ($client_obj, $mnt_result, $cur_fh);
}

#--------------------------------------------------------------------------------
#
#=head2 check_lock
#
#    check_lock()
#
#=item Arguments
#
#This function takes lock type(share-level/byte range) strings as an argument
#
#=back
#
#=item Returns
#
#This function returns 1 if there is lock or 0 is there is no lock.
#
#=back
#
#--------------------------------------------------------------------------------
sub check_lock(){
    my (%opts) = validate (
        @_,
        {
          lock_type => {type => SCALAR, regex => qr/^\D+$/, default => 'share-level'},
        }
    );
    my $lock_found = 1;
    my $file_path = "$Vol_junc_path/$Filename";
    try{
        NACL::C::VserverLocks->find( command_interface=> $Node,
                                    filter => { type => $opts{lock_type},
                                               path => $file_path,
                                               vserver => $VSERVER_GX_NAME,
                                               volume => $FILER_GX_VOL,});
    } catch NACL::Exceptions::NoElementsFound with {
        $Test->log("There is no lock matching these parameters");
        $lock_found = 0;
    };
    return $lock_found;
}

#--------------------------------------------------------------------------------
#
#=head2 verify_error_type
#
#    verify_error_type()
#
#=item Arguments
#
#This function takes error code obtained and error message to be mached in the
#  error code.
#
#=back
#
#=item Returns
#
#This function returns 1 is the error message got maches in the error code
#   and returns 0 its not mached
#
#=back
#
#--------------------------------------------------------------------------------
sub verify_error_type($%) {
    my (%opts) = validate (
    @_,
    {
        error_code => { type => SCALAR,},
        error_msg  => { type => SCALAR,},
    }
    );
    if($opts{error_code}){
        if($opts{error_code} =~ /$opts{error_msg}/){
            $Test->nlog("Op failed as expected, ERROR: $opts{error_msg}");
            return 1;
        } else {
            $Test->nlog("Expected error code: $opts{error_msg}, instead got: $opts{error_code}");
            return 0;
        }
    } else {
        $Test->nlog("Did n't find any error code: Op executed successfully! ");
        return 0;
    }
}

#--------------------------------------------------------------------------------
#
#=head2 get_filer_object
#
#    get_filer_object()
#
#=item Arguments
#
#This function takes no arguments.
#
#=back
#
#=item Returns
#
#This function returns NACL filer object
#
#=back
#
#--------------------------------------------------------------------------------
sub get_filer_object() {
    my $filer_obj = NACL::C::Node->new(node => $FILER);
    return $filer_obj;
}

#--------------------------------------------------------------------------------
#
#=head2 cleanup_file_objects
#
#    cleanup_file_objects()
#
#The function unlocks and closes the files from respective clients.
#
#=item Arguments
#
#This function takes client object, lock result, state id, offset and length
#  as arguments
#
#=back
#
#=item Returns
#
#This function returns nothing.
#
#=back
#
#--------------------------------------------------------------------------------
sub cleanup_file_objects($%){
    my (%opts) = validate (
        @_,
        {
            client_object => { type => HASHREF },
            offset        => { type => SCALAR, regex => qr/^\d+$/ },
            length        => { type => SCALAR, regex => qr/^\d+$/ },
            lock_result   => {  },
            state_id      => {  },
        }
    );
    my $lock_state_id;
    my $unlock_status;
    my $close_status;
    my $Client_object = $opts{client_object};
    my $Lock_result = $opts{lock_result};
    my $State_id = $opts{state_id};
    if(defined scalar(%$Lock_result->{lock_stateid})) {
        $lock_state_id = $Lock_result->{lock_stateid};
        $unlock_status = $Client_object->locku(
                                            offset => $opts{offset},
                                            count => $opts{length},
                                            stateid => $lock_state_id );
    }
    if(defined scalar(%$State_id)){
        $close_status = $Client_object->close(stateid => $State_id);
    }
}
#--------------------------------------------------------------------------------
#
#=head2 open_downgrade_util
#
#    open_downgrade_util()
#
#The function downgrades the access and deny modes of a file.
#
#=item Arguments
#
#This function takes client object, state id, owner, access and deny modes
#  as arguments
#
#=back
#
#=item Returns
#
#This function returns output hash after downgrading the file.
#
#=back
#
#--------------------------------------------------------------------------------
sub open_downgrade_util($%){
    my (%opts) = validate (
        @_,
        {
            client_object => { type => HASHREF },
            owner =>         { type => SCALAR, optional => 1, default => $$ },
            access_mode =>   { type => SCALAR, regex => qr/^\d+$/ },
            deny_mode =>     { type => SCALAR, regex => qr/^\d+$/ },
            state_id      => { type => HASHREF},
        }
    );
    my $Output;
    if($NFS4_MINOR_VERSION){
        $Output = $opts{client_object}->open_downgrade( stateid => $opts{state_id},
                                                  access => $opts{access_mode},
                                                  deny => $opts{deny_mode} );
    } else {
        $Output = $opts{client_object}->open_downgrade( stateid => $opts{state_id},
                                                      owner  => $opts{owner},
                                                      access => $opts{access_mode},
                                                      deny => $opts{deny_mode} );
    }
    return $Output;
}
1;