#! -- this is perl --

use strict;
use nfs::NfsTCD qw{cleanup init uninit main add_to_cleanup %Setup_hash $Node_cmd $Client_cmd @CLEANUP};
use Params::Validate qw{validate SCALAR};
use vars qw{$NFS_VERSION $FILER_GX_VOL $ADDR_TYPE $NFS4_MINOR_VERSION};
use nfs::CommonUtilities;

param('NFS4_MINOR_VERSION', '-default', 0);
param('NFS_VERSION', '-default', 4);
param('FILER_GX_VOL', '-default', 'testvol_flex');
param('ADDR_TYPE', '-default', 'ipv4');

use NACL::CS::ClusterIdentity;
use NACL::CS::Vserver;
use NACL::C::Snapmirror;
use NACL::C::ExportPolicy;

use NACL::STask::Volume;
use NACL::STask::Client::Mount;

use Carp qw{croak};

use NACL::Exceptions::NoElementsFound qw{:try};

my @SM_LIST;

sub cleanup() {
    client_execute("cd /");
    my $obj;
    foreach $obj (@SM_LIST) {
        try {
            wait_for_transfer_complete($obj,1);
        } catch NACL::Exceptions::NoElementsFound with {
            $Log->comment(qq{The object is referencing an non-existing element, ignoring it.});
        };
    }
    push @CLEANUP, @SM_LIST;
    $#SM_LIST = -1;
    try {
        NACL::C::Snapmirror->release(command_interface=>$Node_cmd, 'destination-path' => '*', 'source-path'=>'*');
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        $Log->comment(qq{There is no snapmirror to release, ignoring it.});
    };
    nfs::NfsTCD::cleanup();
}

sub init() {
    nfs::NfsTCD::init();
    disable_err_moved();
    return $TCD::PASS;
}

sub uninit() {
    nfs::NfsTCD::uninit();
}

sub main() {
    nfs::NfsTCD::main();
}

sub create_dp_vol($) {
    create_vol(volume=>shift, type=>"DP");
}

sub create_vol(%) {
    my $vol = (nfs::CommonUtilities::create_vol(@_));
    add_to_cleanup($vol);
    return $vol;
}

sub wait_for_transfer_complete($;$) {
    my $sm_obj = shift;
    my $uninit_ok = shift;
    my $count = 30;
    $Log->comment("Verify that the snapmirror data transfer is completed");
    while ($count) {

        my $state = $sm_obj->state();
        if (!$uninit_ok) {
            if ($state->{state} eq "Uninitialized") {
                croak (qq{The snapmirror is not initialized yet, we can't expect any data transfer to wait for, this is a program error});
            }
        }
        if ($state->{status} eq "Idle") {
            return;
        }
        sleep 30;
        $count--;
    }
    croak (qq{The snapmirror status didnot become Idle after 900 seconds, aborting\n});
}

my $CLUSTER_ID;
sub get_cluster_name() {
    if (!$CLUSTER_ID) {
        $Log->comment(qq{Get the cluster name});
        my $cluster = NACL::CS::ClusterIdentity->fetch(command_interface=>$Node_cmd);
        $CLUSTER_ID = $cluster->{name};
    }
    return $CLUSTER_ID;
}

sub get_rootvol() {
    $Log->comment(qq{Get the root volume of the vserver});
    my $vserver = NACL::CS::Vserver->fetch(command_interface=>$Node_cmd,
        filter=> {vserver => $VSERVER_GX_NAME}, requested_fields => ['rootvolume']);

    return $vserver->{rootvolume};
}

sub create_snapmirror($$$) {
    my ($source, $dest, $type) = @_;
    $Log->comment(qq{Creating the snapmirror between "$source" and "$dest" of type $type});

    my $snapmirror_obj = NACL::C::Snapmirror->create(command_interface=>$Node_cmd,
        #'nacltask_if_exists' => 'purge',
        'source-path' => $source,
        'destination-path' => $dest,
        'type' => $type);

    push @SM_LIST, ($snapmirror_obj);
    return $snapmirror_obj;
}

sub sm_initialize($$) {
    $Log->comment(qq{Initializing the snapmirror});
    NACL::C::Snapmirror->initialize(command_interface=>$Node_cmd,
        'source-path' => shift, "destination-path"=>shift);
}

sub initialize_ls_set($) {
    $Log->comment(qq{Initializing LS set for the snapmirror});
    NACL::C::Snapmirror->initialize_ls_set(command_interface=>$Node_cmd,
        'source-path' => shift);
}

sub update_ls_set($) {
    $Log->comment(qq{Updating LS set for the snapmirror});
    NACL::C::Snapmirror->update_ls_set(command_interface=>$Node_cmd,
        'source-path' => shift);
}

sub namespace_mount($) {
    my ($path) = shift;
    $Log->comment(qq{Mounting the path "$path" of the filer in the client});

    my %opts;
    $opts{mountopts} = "vers=$NFS_VERSION";
    if ($NFS4_MINOR_VERSION) { $opts{mountopts} .= ",minorversion=$NFS4_MINOR_VERSION"; }
    $opts{mountopts} .= ",sec=$::AUTH";

    my $mount_obj = NACL::STask::Client::Mount->create(command_interface=>$Client_cmd,
        nacltask_if_exists => 'purge',
        server => $Setup_hash{hostname},
        path => "$path",
        %opts);

    add_to_cleanup($mount_obj);
    return $mount_obj->mount_point();
}

sub namespace_unmount($) {
    my ($path) = shift;
    client_execute("cd /");
    $Log->comment(qq{Unmounting the path "$path" in the client});
    my $index = 0;
    foreach my $obj (@CLEANUP) {
        if (ref($obj) eq "NACL::STask::Client::Mount") {
            if ($obj->mount_point() eq $path) {
                $obj->purge();
                delete($CLEANUP[$index]);
                last;
            }
        }
        $index++;
    }
}

sub modify_fsid($) {
    my $enable = shift;
    my $vserver_nfs = NACL::STask::VserverNfs->find(command_interface=>$Node_cmd,
        filter => { vserver => $VSERVER_GX_NAME });
    $Log->comment(qq{Changing the "v4-fsid-change" option in the vserver to $enable});

    $vserver_nfs->modify('v4.1-pnfs' => 'disabled');
    if ($enable =~ /^dis/) {
        #These should be disabled before disabling fsid.
        $vserver_nfs->modify("v4.0-migration" => "disabled", "v4.1-migration" => "disabled");
        $vserver_nfs->modify("v4.1-referrals" => "disabled", "v4.0-referrals" => "disabled");
    }
    $vserver_nfs->modify('v4-fsid-change' => $enable);
}

sub get_vol_attribute($;$) {
    my $attribute = shift;
    my $vol = shift || $FILER_GX_VOL;
    $Log->comment(qq{Retrieve the attribute "$attribute" for the volume "$vol"});
    my $volobj = NACL::CS::Volume->fetch(command_interface=>$Node_cmd,
        filte => { vserver => $VSERVER_GX_NAME, volume => $vol });

    return $volobj->{$attribute};
}

sub sm_path($;$) {
    my ($volume, $vserver) = (shift, shift);
    $vserver || ($vserver = $VSERVER_GX_NAME);
    my $cluster = get_cluster_name();
    return "$cluster://$vserver/$volume";
}

sub create_snapshot($;$) {
    my ($snapshot, $volume) = @_;
    $volume = $volume ? $volume : $FILER_GX_VOL;
    $Log->comment(qq{Create a snapshot "$snapshot" for the volume "$volume"});

    my $obj = NACL::C::Snapshot->create(command_interface=>$Node_cmd,
        volume => $volume, snapshot => $snapshot, vserver => $VSERVER_GX_NAME);
    add_to_cleanup($obj);

    return $obj;
}

sub get_client_net_mask(;$) {
    my $intf = shift || "eth0";
    $Log->comment(qq{Get the netmask for the client's IP address});

    my $regex;
    use NetAddr::IP::Lite;
    my $ip_obj;

    my $regex = "inet addr[ \t]*:[ \t]*([0-9\.]+)[ \t]*Bcast[ \t]*:[ \t]*[0-9\.]+[ \t]*Mask[ \t]*:[ \t]*([0-9\.]+)";
    if ($ADDR_TYPE eq "ipv6" or $ADDR_TYPE eq "inet6") {
        $regex = "inet6 addr[ \t]*:[ \t]*([0-9a-f\:]+)\/([0-9]+)\s+";
    }

    my $output = $Client_cmd->execute(command=>"ifconfig $intf");
    if ($output =~ /$regex/) {
        $ip_obj = NetAddr::IP::Lite->new("$1/$2");
        return $ip_obj->network();
    } else {
        croak (qq{Can't analyze the ifconfig output, got:$output});
    }
}

sub populate_data_on_volume($) {
    my $volume = shift;
    my $junction = $volume->state()->{junction_path};
    my $mnt_path = namespace_mount($junction);
    $Log->comment(qq{Fill some data on the mountpath "$mnt_path"});
    client_execute("cp /etc/hosts $mnt_path/vol_data.txt", "sudo");
    client_execute("cp /etc/issue $mnt_path/vol_data1.txt", "sudo");

    client_execute("cat $mnt_path/vol_data.txt > /dev/null", "sudo");
    client_execute("cat $mnt_path/vol_data1.txt > /dev/null", "sudo");
    namespace_unmount($mnt_path);
}

sub setup_export_policy($%) {
    my ($volume, %opts) = @_;
    $Log->comment(qq{Setup an export-policy "$opts{policyname}", its rule and apply it on a volume "$volume"});

    try {
        NACL::C::ExportPolicy->find(command_interface=>$Node_cmd,
            filter => {vserver => $VSERVER_GX_NAME, policyname => $opts{policyname}});
    } catch NACL::Exceptions::NoElementsFound with {
        push @CLEANUP, NACL::C::ExportPolicy->create(command_interface=>$Node_cmd,
            vserver => $VSERVER_GX_NAME, policyname => $opts{policyname});
    };

    if (!$opts{clientmatch}) {
        $opts{clientmatch} = ($ADDR_TYPE eq "inet6" or $ADDR_TYPE eq "ipv6") ? "0::/0" : "0.0.0.0/0";
    }

    push @CLEANUP, NACL::C::ExportPolicyRule->create(command_interface=>$Node_cmd,
        vserver => $VSERVER_GX_NAME, %opts);

    $volume->modify(policy=>$opts{policyname});
}

sub disable_err_moved($) {
    my $vserver = shift;
    $vserver = $vserver ? $vserver : $::VSERVER_GX_NAME;

    NACL::C::VserverNfs->modify(command_interface=>$Node_cmd,
        vserver => $vserver,
        'v4.0-migration' => 'disabled',
        'v4.1-migration' => 'disabled',
        'v4.0-referrals' => 'disabled',
        'v4.1-referrals' => 'disabled',
    );
}

1;
