# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Nameset.pm#1 $
#
## @summary     Nameset - part of the Flg library
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Nameset - library functions specifically for use by the Flg library

=head1 SYNOPSIS

  use nfs::Nameset;

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

This package is intended for use by the nfs::Flg.pm library only. Do not
call nfs::Nameset.pm directly.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Nameset.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Nameset;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Distheap;
use nfs::Contents qw(contents_alloc contents_setfh contents_lookup_unsafe
                     contents_lookup contents_flush_all
                     NAMESET_ENTRY_MAX_FHSIZE);

use Carp qw(croak);
use Carp::Assert;
use Params::Validate qw(validate SCALAR HASHREF UNDEF);
use nfs::NfsDef qw(:NFSFTYPE);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------
sub nameset_init($%;);
sub nameset_alloc($%;);
sub nameset_setfh($%;);
sub type2dh(%);
sub nameset_select_safe($%;);
sub nameset_select($%;);
sub nameset_parent($%;);
sub nameset_dele($%;);
sub nameset_link($%;);
sub testme();

#--------------------------------------------------------------------------------

=head2 nameset_init

    $nameset = nfs::Nameset->nameset_init( prefix => $ns_prefix,
                                      regmax => $maxfiles,
                                      dirmax => $maxdirs,
                                      lnkmax => $maxsymlinks );

Initializes a Nameset object

The following named arguments are accepted:

=over

=item Arguments

=over

=item prefix

(Required) : Test specific string to use for file naming operations

=item regmax

(Optional) : Maximum number of regular files that will be created.
Defaults to 400100.

=item dirmax

(Optional) : Maximum number of directories that will be created
Defaults to 80020.

=item lnkmax

(Optional) : Maximum number of links that will be created
Defaults to 8002.

=back

=item Returns

The Nameset object

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_init($%;) {
    #  int
    #  nameset_init(char *prefix, int regmax, int dirmax, int lnkmax)
    #  {
    #          int elemdatasize = sizeof(struct nameset_entry);
    #
    #          if (strlen(prefix) > sizeof(name_prefix) - 1) {
    #                  report_error(FATAL, "nameset_init prefix too long (%d > %d)",
    #                               strlen(prefix), sizeof(name_prefix) - 1);
    #                  return -1;
    #          }
    #          strncpy(name_prefix, prefix, sizeof(name_prefix));
    #
    #          if ((dh_reg = distheap_init(regmax, elemdatasize)) == NULL) {
    #                  report_error(FATAL, "distheap_init error");
    #                  return -1;
    #          }
    #          if ((dh_dir = distheap_init(dirmax, elemdatasize)) == NULL) {
    #                  report_error(FATAL, "distheap_init error");
    #                  return -1;
    #          }
    #          if ((dh_lnk = distheap_init(lnkmax, elemdatasize)) == NULL) {
    #                  report_error(FATAL, "distheap_init error");
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #

    my $class  = shift;
    my (%opts) = validate (
        @_,
        {
            prefix => { type => SCALAR },
            # Base default options on values used in C code
            regmax => { type => SCALAR, default => 400100, regex => qr/^\d+$/ },
            dirmax => { type => SCALAR, default => 80020,  regex => qr/^\d+$/ },
            lnkmax => { type => SCALAR, default => 8002,   regex => qr/^\d+$/ },
        }
    );
    my $self = bless { %opts }, $class;

    $self->{dh_reg} = nfs::Distheap->distheap_init( maxsize => $self->{regmax},
                                                    sizeof  => 'nameset_entry' );
    $self->{dh_dir} = nfs::Distheap->distheap_init( maxsize => $self->{dirmax},
                                                    sizeof  => 'nameset_entry' );
    $self->{dh_lnk} = nfs::Distheap->distheap_init( maxsize => $self->{lnkmax},
                                                    sizeof  => 'nameset_entry' );

    return $self;
} # End nameset_init

#--------------------------------------------------------------------------------

=head2 nameset_alloc

    $nse = $nameset->nameset_alloc( parent => undef,
                                    type => NFDIR,
                                    weight => 1 );

Allocates a nameset entry

The following named arguments are accepted:

=over

=item Arguments

=over

=item parent

(Required) : Parent of nameset entry

=item type

(Required) : Type of entry to allocate

=item weight

(Required) : Weight. (C code always uses 1)

=back

=item Returns

The newly allocated nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_alloc($%;) {
    #  nameset_entry_t
    #  nameset_alloc(nameset_entry_t parent, int type, int weight)
    #  {
    #          distheap_t dh = type2dh(type);
    #          nameset_entry_t nse;
    #          contents_t ct;
    #
    #          assert(parent == NULL || parent->type == NFDIR);
    #          assert(weight >= 0);
    #
    #          if ((nse = distheap_alloc(dh, weight)) == NULL) {
    #                  report_error(FATAL, "distheap_alloc error");
    #                  return NULL;
    #          }
    #
    #          bzero(nse, sizeof(struct nameset_entry));
    #          nse->parentdir_idx = parent ? distheap_toidx(dh_dir, parent) : -1;
    #          nse->type = type;
    #
    #          ct = contents_alloc(nse);
    #
    #          ct->ct_uid = -1; /* unknown */
    #          ct->ct_gid = -1; /* unknown */
    #          ct->ct_linkcount = 1;
    #
    #          return nse;
    #  }
    #
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            parent => { type => SCALAR|HASHREF|UNDEF },
            type   => { type => SCALAR, regex => qr/^\d+$/ },
            weight => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    my $parentdir_idx;

    my $dh;
    eval { $dh = type2dh( type => $opts{type} ) };
    croak "nameset_alloc: $@" if $@;

    if ( ref( $opts{parent} ) ) {
        # Parent is a HASHREF, validate type and get index
        assert( $opts{parent}->{type} == NFDIR );
        $parentdir_idx = $opts{parent}->{my_idx};
    } else {
        # Parent is a value, use the value as the index
        $parentdir_idx = ( defined $opts{parent} ) ? $opts{parent} : -1;
    }

    # distheap_alloc now returns an array with the ref to u.freelink
    #  and the index to the distheap_entry
    my ( $nse, $index ) = $self->{$dh}->distheap_alloc(
            weight => $opts{weight} );

    # $nse would have been a pointer to the u.data segment,
    #  so reassign it so that is the case
    $nse = $self->{$dh}{entries}{$index}{data};
    croak "nameset_alloc: distheap_alloc returned undef reference"
            if !defined $nse;

    # Remove the freelink component and establish the data component,
    #  simulating bzero of the address
    delete $self->{$dh}{entries}{$index}{freelink};

    # Record the parentdir_idx and type
    $self->{$dh}{entries}{$index}{data}{parentdir_idx} = $parentdir_idx;
    $self->{$dh}{entries}{$index}{data}{type}          = $opts{type};

    my $ct = contents_alloc( nse => $nse );

    $ct->{ct_uid}       = -1; # unknown
    $ct->{ct_gid}       = -1; # unknown
    $ct->{ct_linkcount} = 1;

    return $nse;

} # End nameset_alloc

#--------------------------------------------------------------------------------

=head2 nameset_setfh

    $nameset->nameset_setfh( nse => $nse,
                             fhdata => $fh->{data},
                             fhlen => $fh->{len} );

nameset_setfh function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : A nameset entry

=item fhdata

(Required) : Filehandle applied to nse

=item fhlen

(Required) : Length of filehandle being applied

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_setfh($%;) {
    #  void
    #  nameset_setfh(nameset_entry_t nse, char *fhdata, int fhlen)
    #  {
    #          assert(fhlen <= NAMESET_ENTRY_MAX_FHSIZE);
    #          assert(*((u_int64_t *)fhdata) != 0); /* data in first 8 bytes */
    #          bcopy(fhdata, nse->fhdata, fhlen);
    #          nse->fhlen = fhlen;
    #          contents_setfh(nse);
    #  }
    #

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            nse    => { type => HASHREF },
            fhdata => { type => SCALAR },
            fhlen  => { type => SCALAR },
        }
    );

    assert( $opts{fhlen} <= NAMESET_ENTRY_MAX_FHSIZE );

    assert( $opts{fhdata} ne '' );

    $opts{nse}->{fhdata} = $opts{fhdata};
    $opts{nse}->{fhlen}  = $opts{fhlen};
    contents_setfh( nse => $opts{nse} );

} # End nameset_setfh

#--------------------------------------------------------------------------------

=head2 type2dh

    $dh = type2dh( type => $type );

Get a distheap pointer for the specified type.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Required) : File type for which to get a distheap pointer.

=back

=item Returns

Pointer to distheap entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub type2dh(%) {
    #  static __inline distheap_t
    #  type2dh(int type)
    #  {
    #          switch(type) {
    #          case NFREG:
    #                  return dh_reg;
    #          case NFDIR:
    #                  return dh_dir;
    #          case NFLNK:
    #                  return dh_lnk;
    #          }
    #          report_error(FATAL, "type2dh: bad type %d", type);
    #          return NULL;
    #  }
    #
    my (%opts) = validate (
        @_,
        {
            type    => { type => SCALAR },
        }
    );
    my $type = $opts{type};

    if ( $type == NFREG() ) { return 'dh_reg' }
    elsif ( $type == NFDIR() ) { return 'dh_dir' }
    elsif ( $type == NFLNK() ) { return 'dh_lnk' }
    else { croak "type2dh: bad type $type" };

} # End type2dh

#--------------------------------------------------------------------------------

=head2 nameset_select_safe

  $nse = $nameset->nameset_select_safe( type => NFREG );

nameset_select_safe function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Required) : Type of nameset entry to select

=back

=item Returns

A nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_select_safe($%;) {
    #  nameset_entry_t
    #  nameset_select_safe(int type)
    #  {
    #          nameset_entry_t nse;
    #          nameset_entry_t parent;
    #          int retries = 0;
    #
    #          do {
    #                  nse = nameset_select(type);
    #                  if (!nse)
    #                          break;
    #          } while ((nse->go_away || nse->dele_when_free) && ++retries < 100);
    #
    #          if (!nse || nse->go_away || nse->dele_when_free || nse->rename) {
    #                  nse = NULL;
    #                  /* XXO: stats */
    #                  return NULL;
    #          }
    #
    #          /*
    #           * ROD?
    #           */
    #          parent = nameset_parent(nse);
    #          if (parent != NULL && parent->fhlen == 0)
    #                  nse = parent;
    #
    #          /*
    #           * ROD?
    #           */
    #          if (nse->fhlen == 0) {
    #                  contents_t ct;
    #
    #                  /*
    #                   * For linkcounts > 1 another lookup from
    #                   * a different nse may have already set the
    #                   * new filehandle. In that case, "back copy"
    #                   * it into this nse.
    #                   *
    #                   * XXO: ultimately nse's should not really
    #                   * encode filehandle data; they should all be
    #                   * in the contents_t's.
    #                   */
    #                  ct = contents_lookup_unsafe(nse);
    #                  if (ct->ct_fhlen > 0) {
    #                          assert(ct->ct_linkcount > 1);
    #                          nse->fhlen = ct->ct_fhlen;
    #                          memcpy(nse->fhdata, ct->ct_fhdata, nse->fhlen);
    #                          return nse;
    #                  }
    #
    #                  /*
    #                   * Find the nearest valid parent.
    #                   */
    #                  for (;;) {
    #                          parent = nameset_parent(nse);
    #                          assert(parent != NULL);
    #                          assert(parent != nse);
    #
    #                          if (parent->fhlen != 0)
    #                                  break;
    #
    #                          nse = parent;
    #                  }
    #
    #                  /*
    #                   * Issue a lookup on the nse, and abort
    #                   * whatever op was looking for a valid nse.
    #                   *
    #                   * NB: zero the first dword of fhdata so
    #                   * do_lookup_callback performs a nameset_setfh().
    #                   */
    #                  *((u_int64_t *)nse->fhdata) = 0;
    #                  do_lookup(nse);
    #                  return NULL;
    #          }
    #
    #          assert(nse->type == type);
    #          return nse;
    #  }

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            type   => { type => SCALAR },
        }
    );
    my $type = $opts{type};

    my $nse;
    my $parent;
    my $retries = 0;

    $nse = $self->nameset_select( type => $type );
    if ( defined $nse ) {
        $nse->{go_away}        = 0 if ( !defined $nse->{go_away} );
        $nse->{dele_when_free} = 0 if ( !defined $nse->{dele_when_free} );
        $nse->{rename}         = 0 if ( !defined $nse->{rename} );
    }

    while ( defined $nse && ($nse->{go_away} || $nse->{dele_when_free}) &&
                ++$retries < 100 ) {
        $nse = $self->nameset_select( type => $type );
        last if ( ! defined $nse );
        $nse->{go_away}        = 0 if ( !defined $nse->{go_away} );
        $nse->{dele_when_free} = 0 if ( !defined $nse->{dele_when_free} );
        $nse->{rename}         = 0 if ( !defined $nse->{rename} );
    }

    if ( !defined $nse || $nse->{go_away} ||
                   $nse->{dele_when_free} || $nse->{rename} ) {
        $nse = undef;
        return undef;
    }

    $parent = $self->nameset_parent( nse => $nse );

    if ( defined $parent && $parent->{fhlen} == 0 ) {
        $nse = $parent;
    }

    if ( $nse->{fhlen} == 0 ) {
        my $ct = contents_lookup_unsafe( nse => $nse );

        if ( $ct->{ct_fhlen} > 0 ) {
            assert( $ct->{ct_linkcount} > 1 );
            $nse->{fhlen}  = $ct->{ct_fhlen};
            $nse->{fhdata} = $ct->{ct_fhdata};
            return $nse;
        }

        while ( 1 ) {
            $parent = $self->nameset_parent( nse => $nse );
            assert( defined $parent );
            assert( $parent != $nse );

            last if ( $parent->{fhlen} != 0 );

            $nse = $parent;
        }

        $nse->{fh_data} = '';
        flg::do_lookup( nse => $nse );

        return undef;
    }

    assert( $nse->{type} == $type );
    return $nse;

} # End nameset_select_safe

#--------------------------------------------------------------------------------

=head2 nameset_select

  $nse = $self->nameset_select( type => $type );

nameset_select function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Required) : Type of nameset entry to select

=back

=item Returns

A nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_select($%;) {
    #  nameset_entry_t
    #  nameset_select(int type)
    #  {
    #          distheap_t dh = type2dh(type);
    #          nameset_entry_t nse;
    #
    #          nse = distheap_select(dh);
    #          assert(nse == NULL || nse->type == type);
    #
    #          return nse;
    #  }

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            type   => { type => SCALAR },
        }
    );
    my $dh;
    my $nse;

    for ( $opts{type} ) {
        $_ == NFREG && do { $dh = 'dh_reg'; last };
        $_ == NFDIR && do { $dh = 'dh_dir'; last };
        $_ == NFLNK && do { $dh = 'dh_lnk'; last };
        croak "nameset_select: bad type: $opts{type}";
    }

    $nse = $self->{$dh}->distheap_select();
    assert( !defined $nse || $nse->{type} == $opts{type} );

    return $nse;

} # End nameset_select

#--------------------------------------------------------------------------------

=head2 nameset_parent

  $parent = $self->nameset_parent( nse => $nse );

nameset_parent function taken from original C code

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : The nameset entry for which the parent will be found

=back

=item Returns

The parent nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_parent($%;) {
    #  nameset_entry_t
    #  nameset_parent(nameset_entry_t nse)
    #  {
    #          if (nse->parentdir_idx == -1) {
    #                  return NULL;
    #          }
    #          return (nameset_entry_t)distheap_fromidx(type2dh(NFDIR),
    #                                                   nse->parentdir_idx);
    #  }

    #  distheap_data_t
    #  distheap_fromidx(distheap_t heap, int idx)
    #  {
    #          if (idx == 0)
    #                  return NULL;
    #          assert(idx > 0);
    #          if (idx > heap->maxsize) {
    #                  return NULL; /* out of range */
    #          }
    #          return distheap_entry2data(heap, distheap_idx2entry(heap, idx));
    #  }
    #

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    $nse->{parentdir_idx} = 0 if ( !defined $nse->{parentdir_idx} );
    if ( $nse->{parentdir_idx} == -1 ) {
        return undef;
    }

    my $idx = $nse->{parentdir_idx};
    return undef if ( $idx == 0 );
    if ( $idx > $self->{dh_dir}{maxsize} ) {
        return undef; # out of range
    }
    return $self->{dh_dir}{entries}{$idx}{data};
} # End nameset_parent

#--------------------------------------------------------------------------------

=head2 nameset_dele

  $nameset->nameset_dele( nse => $nse );

nameset_dele function taken from original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item nse

(Required) : The nameset entry to be deleted

=back

=item Returns

0 if success

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_dele($%;) {
    #  int
    #  nameset_dele(nameset_entry_t nse)
    #  {
    #          if (nse->type == NFNON) {
    #                  /*
    #                   * if an nse is marked for deletion ("dele_when_free") and
    #                   * would be deleted anyway (remove, rmdir), then both
    #                   * nameset_deref and an explicit nameset_dele will attempt
    #                   * to delete it.  this is not a race.  a sort-of-cleaner
    #                   * fix would be to delay the unref until after the dele in
    #                   * remove and rmdir.
    #                   */
    #                  return 0; /* already deleted */
    #          }
    #          nse->dele_when_free = 1;
    #          if (nse->refcnt == 0) {
    #                  distheap_t dh = type2dh(nse->type);
    #                  contents_t ct = contents_lookup(nse);
    #                  nse->type = NFNON; /* further type2dh calls will fail */
    #
    #                  assert(ct->ct_linkcount > 0);
    #                  if (--ct->ct_linkcount == 0)
    #                          contents_flush_all(nse);
    #                  nse->ct_idx = 0;
    #                  nse->rename = 0;
    #                  nse->go_away = 0;
    #                  return distheap_dele(dh, nse);
    #          }
    #          return 0;
    #  }

    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            nse => { type => HASHREF },
        }
    );
    my $nse = $opts{nse};

    if ( $nse->{type} == NFNON ) {
        return 0;
    }

    $nse->{dele_when_free} = 1;
    if ( $nse->{refcnt} == 0 ) {
        my $dh = type2dh( type => $nse->{type} );
        my $ct = contents_lookup( nse => $nse );
        $nse->{type} = NFNON;

        assert( $ct->{ct_linkcount} > 0 );
        if ( --$ct->{ct_linkcount} == 0 ) {
            contents_flush_all( nse => $nse );
        }
        $nse->{ct_idx}  = 0;
        $nse->{rename}  = 0;
        $nse->{go_away} = 0;
        return $self->{$dh}->distheap_dele( hd => $nse );
    }

    return 0;

} # End nameset_dele

#--------------------------------------------------------------------------------

=head2 nameset_link

  $nse = $nameset->nameset_link( parent  => $pnse,
                                 sibling => $onse,
                                 weight  => 1 );

nameset_link function taken from original C code

The following named arguments are accepted:

=over

=item Arguments

=over

=item parent

(Required) : Nameset entry pointing to the parent

=item sibling

(Required) : Nameset entry pointing to a sibling

=item weight

(Required) : Weight

=back

=item Returns

The linked nameset entry

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub nameset_link($%;) {
    #  nameset_entry_t
    #  nameset_link(nameset_entry_t parent, nameset_entry_t sibling, int weight)
    #  {
    #          distheap_t dh = type2dh(sibling->type);
    #          contents_t ct = contents_lookup(sibling);
    #          nameset_entry_t nse;
    #
    #          assert(parent == NULL || parent->type == NFDIR);
    #          assert(weight >= 0);
    #
    #          if ((nse = distheap_alloc(dh, weight)) == NULL) {
    #                  report_error(FATAL, "distheap_alloc error");
    #                  return NULL;
    #          }
    #
    #          *nse = *sibling;
    #          nse->refcnt = 0;
    #          assert(!nse->rename);
    #          assert(!nse->dele_when_free);
    #          nse->parentdir_idx = parent ? distheap_toidx(dh_dir, parent) : -1;
    #          ct->ct_linkcount++;
    #          // assert(ct->ct_linkcount <= 2);
    #
    #          return nse;
    #  }

    my $self = shift;

    my (%opts) = validate (
        @_,
        {
            parent  => { type => HASHREF|UNDEF },
            sibling => { type => HASHREF },
            weight  => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    assert( !defined $opts{parent} || $opts{parent}->{type} == NFDIR );

    my $dh = type2dh( type => $opts{sibling}->{type} );
    my $ct = contents_lookup( nse => $opts{sibling} );

    my ( $nse, $index ) = $self->{$dh}->distheap_alloc(
            weight => $opts{weight} );

    # $nse would have been a pointer to the u.data segment,
    #  so reassign it so that is the case
    $nse = $self->{$dh}{entries}{$index}{data};

    if ( !defined $nse ) {
        croak "nameset_link: error in distheap_alloc call";
    }

    # Assign sibling values to nse
    $nse->{fhdata}         = $opts{sibling}->{fhdata};
    $nse->{fhlen}          = $opts{sibling}->{fhlen};
    $nse->{type}           = $opts{sibling}->{type};
    $nse->{ct_idx}         = $opts{sibling}->{ct_idx};
    $nse->{rename}         = $opts{sibling}->{rename};
    $nse->{dele_when_free} = $opts{sibling}->{dele_when_free};
    $nse->{go_away}        = $opts{sibling}->{go_away};

    $nse->{refcnt} = 0;
    assert( !defined $nse->{rename} || $nse->{rename} == 0 );
    assert( !defined $nse->{dele_when_free} || $nse->{dele_when_free} == 0 );

    $nse->{parentdir_idx} = ( defined $opts{parent} ) ?
                              $opts{parent}->{my_idx} : -1;
    $ct->{ct_linkcount}++;

    return $nse;

} # End nameset_link

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Nameset SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    # Call type2dh on NFREG
    my $type = type2dh( type => NFREG() );

    # Should return 'dh_reg'
    if ($type ne 'dh_reg') {
        Tharn::logresult('FAIL', "type2dh returned incorrect dh string");
    } else {
        Tharn::logresult('PASS', "SELF TEST of nfs/Nameset.pm");
    }
}

1;
