# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsClient.pm#2 $
#
## @summary nfs::NfsClient - NFS client Perl Synthetic Class
## @author	dl-ssqa
## @status	public
## @pod here

=head1 NAME

nfs::NfsClient - NFS client protocol class

=head1 SYNOPSIS

  use nfs::NfsClient;

  $nfs_obj = nfs::NfsClient->new(hostname => $host, 
                                 version  => 4,
                                 minor_version => 0,
                                 proto => 'tcp');

=head1 DESCRIPTION

Provides an object interface for the NFS client protocol version
2, 3 and 4 and 4.1. All procedures for NFS version 2 and 3 are implemented
including some MOUNT protocol procedures to mount, unmount and dump
a list of mounts and exports on the NFS server. On NFSv4, only those
operations needed to have similar functionality with NFSv2 and v3
are implemented as of now.

Most methods are called with the same parameters for all three
versions of NFS. NFSv3 and v4 may require additional input parameters.
This should be noted in the documentation of each particular method.
In addition, some methods are not supported by all versions of NFS.
If this is the case the method croaks with an unsupported error message.

NFS versions 2,3 and 4 are implemented in nfs::NfsClient::4_0.pm file
where as NFS version 4.1 is going to be in nfs::NfsClient::4_1.pm.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-ssqa (SSQA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsClient.pm#2 $
  Last revised $Date: 2013/06/05 $

=head1 COPYRIGHT

  Copyright 2010, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsClient;

use strict;
use warnings;
use Params::Validate qw(validate);
use Data::Dumper;

use Tharn qw($Log config_result script_result);

#---------------------------------------------------------------------

=head2 new

  $nfs_obj = nfs::NfsClient->new(hostname => $host,
                                 version => 2|3|4
                                 proto => 'udp'|'tcp'
                                 [, minor_version => 0|1 ]
                                 [, setclientid_on_stale => 0|1 ]
                                 [, proto => 'udp'|'tcp' ]
                                 [, addr_type => 'inet4'|'inet6' ]
                                 [, mnt_ver => 1|2|3 ]
                                 [, mnt_proto => 'udp'|'tcp' ]
                                 [, nlm => 0|1 ]
                                 [, nlm_ver => 1|2|3|4 ]
                                 [, nlm_proto => 'udp'|'tcp' ]
                                 [, nsm => 0|1 ]
                                 [, nsm_proto => 'udp'|'tcp' ]
                                 [, port => $port ]
                                 [, resvport => 0|1 ]
                                 [, xid => $xid ]
                                 [, retry => $retry ]
                                 [, nfs_retry => $nfs_retry ]
                                 [, nfs_delay => $nfs_delay ]
                                 [, reconnect => 0|1 ]
                                 [, timeout => $timeout ]
                                 [, qsize => $qsize ]
                                 [, localaddr => $localaddr ]);

NFS object constructor. It creates the RPC connection object to the
server and initializes the class variables.  For NFSv4, the sequence
id's for both open and lock are initialized to zero, the callback
queue and the list of open delegations are defined and the client
identifier and verifier are set to a random number so every instance
of this object is a completely different client.

See subclass perldoc for details about arguments.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%) {
    my ($class, %opts) = @_;
    $Log->enter(Dumper(\%opts));

    my $self = bless {minor_version        => $opts{minor_version}        || 0,
                      version              => $opts{version},
                      }, $class;

    my ($object, $nfsclass);
    if (   $self->{version} == 2
        || $self->{version} == 3) {
        $nfsclass = 'nfs::NfsClient::3_0';
    } elsif ($self->{version} eq "4.0" || ($self->{version} == 4 && $self->{minor_version} == 0)) {
		$opts{version} = "4";
		$opts{minor_version} = "0";
        $nfsclass = 'nfs::NfsClient::4_0';
    } elsif ($self->{version} eq "4.1" || ($self->{version} == 4 && $self->{minor_version} == 1)) {
		$opts{version} = "4";
		$opts{minor_version} = "1";
        $nfsclass = 'nfs::NfsClient::4_1';
    } else {
        die script_result(
            "nfs version $self->{version}.$self->{minor_version} is not supported",
            stacktrace => 0);
    }

    $Log->comment("Loading $nfsclass");
    eval "require $nfsclass";
    die config_result("'$nfsclass' failed: $@", stacktrace => 0) if ($@);
    $object = $nfsclass->new(%opts);
    $self->{object} = $object;

    $Log->exit(Dumper($self));
    return $self;
} ## end sub new($%)

#---------------------------------------------------------------------

=head2 protocol

  $nfs_version = $nfs_obj->protocol();

Returns the NFS Protocol (tcp/udp)

=for html <hr />

=cut

#---------------------------------------------------------------------
sub protocol { return shift->{object}->{proto} }

#---------------------------------------------------------------------

=head2 version

  $nfs_version = $nfs_obj->version();

Returns the NFS Version

=for html <hr />

=cut

#---------------------------------------------------------------------
sub version { return shift->{object}->{version} }

#---------------------------------------------------------------------

=head2 minor_version

  $nfs_minorversion = $nfs_obj->minor_version();

Returns the NFS Minor Version

=for html <hr />

=cut

#---------------------------------------------------------------------
sub minor_version { return shift->{object}->{minor_version} }

#---------------------------------------------------------------------

#---------------------------------------------------------------------

=head2 rpc

  $rpc_obj = $nfs_obj->rpc();

Returns the RPC object of the connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub rpc { return shift->{object}->{rpc} }

#---------------------------------------------------------------------

=head2 mnt_fh

  $mnt_fh = $nfs_obj->mnt_fh();

Returns the filehandle of the root of the mount point.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mnt_fh { return shift->{object}->{mnt_fh} }

#---------------------------------------------------------------------

=head2 hostname

  $mnt_fh = $nfs_obj->hostname();

Returns the hostname that is connected to.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub hostname { return shift->{object}->{hostname} }

#---------------------------------------------------------------------

=head2 curr_fh

  $curr_fh = $nfs_obj->curr_fh();

Returns the current filehandle of the object. Most methods which accept
a filehandle have as a default this filehandle. It is set internally
with operations like LOOKUP, CREATE, OPEN, etc.

=over

=item Arguments

=over

=item <curr_fh>

(Optional) : Set the nfs curr_fh to this.

=back

=item Returns

The nfs curr_fh value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub curr_fh { 
    my ($self, $new_value) = @_;
    my $nfs = $self->{object};
    my $old = $nfs->{curr_fh};

    if (defined($new_value)) {
        $nfs->{curr_fh} = $new_value;
    }

    return $old;
}

#---------------------------------------------------------------------

=head2 mnt_path

  $mnt_path = $nfs_obj->mnt_path();

Returns the pathname of the mount.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mnt_path { return shift->{object}->{mnt_path} }

#---------------------------------------------------------------------

=head2 result

  $result = $nfs_obj->result();

Returns the last failed result. All methods croak when status is not
NFS?_OK and the results are saved so they can be retrieved using this
method.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub result { return shift->{object}->{result} }

#---------------------------------------------------------------------

=head2 nfs_retry

  $nfs_retry = $nfs_obj->nfs_retry([ $nfs_retry_new ]);

Gets or sets the current nfs retry for the NFS object.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the nfs retry to this.

=back

=item Returns

The nfs retry value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nfs_retry($$;) {
    my ($self, $new_value) = @_;
    my $nfs = $self->{object};
    my $old = $nfs->{nfs_retry};

    if (defined($new_value)) {
        $nfs->{nfs_retry} = $new_value;
    }

    return $old;
} ## end sub nfs_retry($$;)

#---------------------------------------------------------------------

=head2 nfs_delay

  $nfs_delay = $nfs_obj->nfs_delay([ $nfs_delay_new ]);

Gets or sets the current nfs delay for the NFS object.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the nfs delay to this.

=back

=item Returns

The nfs delay value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nfs_delay($$;) {
    my ($self, $new_value) = @_;
    my $nfs = $self->{object};
    my $old = $nfs->{nfs_delay};

    if (defined($new_value)) {
        $nfs->{nfs_delay} = $new_value;
    }

    return $old;
} ## end sub nfs_delay($$;)

#---------------------------------------------------------------------

sub AUTOLOAD {
    my $method = our $AUTOLOAD;
    $Log->enter("method = $method");
    if ($method =~ /::DESTROY/) {
        $Log->exit();
        return;
    }

    no strict 'refs';
    my $self   = shift @_;
    my %params = (@_,);
    $method =~ s/.*:://;    # trim the package name

    my $object = $self->{object};
    $Log->trace("nfs::NfsClient::AUTOLOAD calling $method in $object");

    if (!$object->can($method)) {
        die "$method is not supported by $object";
    }

    my $out = $object->$method(%params);

    $Log->exit(Dumper($out));
    return $out;
} ## end sub AUTOLOAD

#---------------------------------------------------------------------
# Destructor
# There is none for this class.
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;

}

#---------------------------------------------------------------------
# END block
# There is none to be done for this class
#---------------------------------------------------------------------
END {
}

1;
