# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsClient/3_0.pm#2 $
#
## @summary NfsClient - NFS client protocol class
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

NfsClient - NFS client protocol class

=head1 SYNOPSIS

  use nfs::NfsClient;

  $nfs_obj = nfs::NfsClient->new(hostname => $host, version  => 3,
                                 proto => 'tcp');
  $out = $nfs_obj->mount(path => '/vol/testvol');
  $nfs_obj->null();
  $out = $nfs_obj->getattr();
  $out = $nfs_obj->setattr(attr => {mode => 0777});
  $out = $nfs_obj->lookup(name => 'file1');
  $out = $nfs_obj->access(access => ACCESS_READ);
  $out = $nfs_obj->read(offset => 0, count => 4096);
  $out = $nfs_obj->readlink(fhandle => $fhandle);
  $out = $nfs_obj->mkdir(name => 'test_dir', attr => {mode => 0755});
  $out = $nfs_obj->symlink(name => 'sym_link', data => 'test_dir/file2');
  $out = $nfs_obj->mknod(name => 'dev_char', type => NFCHR, major=>123,
                         minor => 12, attr => {mode => 0666});
  $out = $nfs_obj->rename(fromdir => $f_fhandle, fromname => 'file3',
                          dodir   => $f_fhandle, toname   => 'file4');
  $out = $nfs_obj->remove(name => 'file4');
  $out = $nfs_obj->link(todir => $t_fhandle, name => 'hard_link1');
  $out = $nfs_obj->readdir();
  $out = $nfs_obj->readdirplus();
  $out = $nfs_obj->fsstat();
  $out = $nfs_obj->write(offset => $offset, data => $data);
  $out = $nfs_obj->commit();
  $out = $nfs_obj->create(name => 'fifo', type => NFFIFO);
  $out = $nfs_obj->rmdir(name => 'fifo');
  $out = $nfs_obj->fsinfo();
  $out = $nfs_obj->pathconf();
  $nfs_obj->umount(path => '/vol/testvol');
  $out = $nfs_obj->mount_dump();
  $out = $nfs_obj->mount_export();
  $nfs_obj->current_fh_set(fhandle => $fhandle);
  $nfs_obj->sm_notify();
  nfs::NfsClient::sm_notify(host => $nsm_server,
                            mon_name => $mon_name,
                            state => $sm_state);
  $nfs_obj->sm_simu_crash(host => $nsm_server);
  nfs::NfsClient::sm_simu_crash(host => $nsm_server);

  # To lock a file on NFSv2 or NFSv3
  $lock = $nfs_obj->lock(offset => 32, count => 64);

  # To unlock the file on NFSv2 or NFSv3
  $out = $nfs_obj->locku(offset => 32, count => 64);

  # or simply use
  $out = $nfs_obj->locku(lockid => $lock->{lockid});
  # To unlock only a subregion of the lock
  $out = $nfs_obj->locku(lockid => $lock->{lockid}, offset => 50, count=>10);

  # To cancel a blocked lock on NFSv2 or NFSv3
  $out = $nfs_obj->lock_cancel(lockid => \%lockid);

  # To check whether a lock is available to the client on NFSv2 or NFSv3.
  $out = $nfs_obj->lock_test(offset => 32, count => 64, exclusive => 1);
  # To check whether a lock is available to the client
  # For NFSv2 and NFSv3 it calls method lock_test with exclusive set
  # to 1 if type is WRITE(W)_LT and set to 0 if type is READ(W)_LT
  $out = $nfs_obj->lockt(offset => 32, count => 64, type => WRITE_LT);

  # NLM and NSM callback functions (timeout is in milliseconds)
  $nfs_obj->nlm_callback_process(timeout => 10000);
  $nfs_obj->nsm_callback_process(timeout => 10000);

  # Object description
  $rpc        = $nfs_obj->rpc();         # RPC object
  $mnt_fh     = $nfs_obj->mnt_fh();      # Filehandle of root of mount
  $curr_fh    = $nfs_obj->curr_fh();     # Current filehandle use as default
  $mnt_path   = $nfs_obj->mnt_path();    # Mount path
  $result     = $nfs_obj->result();      # Last failed results
  $nfs_retry  = $nfs_obj->nfs_retry();   # Number of times to retry a request
  $nfs_delay  = $nfs_obj->nfs_delay();   # Seconds to delay before a retry

  # Set the number of times to retry a request if NFS3ERR_JUKEBOX
  # or NFS4ERR_DELAY is returned
  $nfs_retry_prev = $nfs_obj->nfs_retry($nfs_retry_new);

  # Set the number of seconds to delay before a retry
  $nfs_delay_prev = $nfs_obj->nfs_delay($nfs_delay_new);

=head1 DESCRIPTION

Provides an object interface for the NFS client protocol version
2 and 3. All procedures for NFS version 2 and 3 are implemented
including some MOUNT protocol procedures to mount, unmount and dump
a list of mounts and exports on the NFS server.

Most methods are called with the same parameters for v2 and v3
versions of NFS. NFSv3 may require additional input parameters.
This should be noted in the documentation of each particular method.
In addition, some methods are not supported by all versions of NFS.
If this is the case the method croaks with an unsupported error message.

The format of the 'attr' option for setting the attributes of a
file system object is as follows. For NFSv2 and v3, the valid
options are: mode, size, uid, gid, atime and mtime.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsClient/3_0.pm#2 $
  Last revised $Date: 2013/06/05 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsClient::3_0;

use strict;
use warnings;

use nfs::Rpc;
use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF CODEREF UNDEF);
use nfs::NfsXdr qw(xdr_access xdr_cb_compound xdr_cb_exit
  xdr_cb_info xdr_cb_nlm_granted xdr_cb_nsm_notify xdr_cb_pid xdr_cb_port
  xdr_cb_register xdr_commit xdr_compound xdr_compound_res xdr_cookie
  xdr_cookieverf xdr_create xdr_daemon_info xdr_dump xdr_export
  xdr_fhandle xdr_fsinfo xdr_fsstat xdr_getattr xdr_link xdr_lookup xdr_mkdir
  xdr_mknod xdr_mount xdr_nfstime xdr_nlm_lock xdr_nlm_stat xdr_nlm_test
  xdr_opaque xdr_pathconf xdr_read xdr_readdir xdr_readdirplus xdr_readlink
  xdr_remove xdr_rename xdr_sattr xdr_setattr xdr_specdata
  xdr_symlink xdr_uint xdr_uint64 xdr_write);
use nfs::NfsDef qw(nfs_error nlm_error bitmap4_convert
  EXCLUSIVE FILE_SYNC
  MOUNTPROC_DUMP MOUNTPROC_EXPORT MOUNTPROC_MNT MOUNTPROC_PATHCONF
  MOUNTPROC_UMNT NFBLK NFCHR NFDIR NFFIFO NFLNK NFREG NFSOCK NFSPROC3_ACCESS NFSPROC3_COMMIT
  NFSPROC3_CREATE NFSPROC3_FSINFO NFSPROC3_FSSTAT NFSPROC3_LINK
  NFSPROC3_LOOKUP NFSPROC3_MKDIR NFSPROC3_MKNOD NFSPROC3_PATHCONF
  NFSPROC3_READDIRPLUS NFSPROC3_REMOVE NFSPROC3_RENAME NFSPROC3_RMDIR
  NFSPROC3_SYMLINK NFSPROC3_WRITE NFSPROC_CREATE NFSPROC_GETATTR NFSPROC_LINK
  NFSPROC_LOOKUP NFSPROC_MKDIR NFSPROC_READ NFSPROC_READDIR NFSPROC_READLINK
  NFSPROC_REMOVE NFSPROC_RENAME NFSPROC_RMDIR NFSPROC_SETATTR NFSPROC_STATFS
  NFSPROC_SYMLINK NFSPROC_WRITE NLM_CANCEL NLM_GRANTED NLM_LOCK NLM_TEST
  NLM_UNLOCK READ_LT READW_LT UNCHECKED WEBNFS2_FHANDLE
  WEBNFS3_FHANDLE WRITE_LT WRITEW_LT NFS3ERR_JUKEBOX);

use Data::Dumper;
use Tharn qw($Log);

#
# To handle different addr_type family, and avoid sending this addr information across
# callbacks, define a local variable and set the value when object is initiated (new)
#
my $loopback;

BEGIN {

    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
} ## end BEGIN

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($%);
sub _daemon(%);
sub _start_daemon($%);
sub _stop_daemon($$$;);
sub _nlm_start($;);
sub _nsm_start($;);
sub _nlm_daemon(%);
sub _nsm_daemon(%);
sub _nlm_process_queue($$;);
sub _nsm_process_queue($$;);
sub nlm_callback_process($;%);
sub nsm_callback_process($;%);
sub null($;);
sub getattr($%);
sub setattr($%);
sub lookup($%);
sub access($%);
sub readlink($%);
sub mkdir($%);
sub symlink($%);
sub mknod($%);
sub remove($%);
sub rename($%);
sub link($%);
sub readdir($%);
sub readdirplus($%);
sub fsstat($%);
sub read($%);
sub write($%);
sub commit($%);
sub lock($%);
sub locku($%);
sub lockt($%);
sub lock_cancel($%);
sub lock_test($%);
sub create($%);
sub rmdir($%);
sub fsinfo($%);
sub pathconf($%);
sub mount_rpc_get($;);
sub nlm_rpc_get($;);
sub mount($%);
sub umount($%);
sub mount_dump($;);
sub mount_export($;);
sub sm_notify(%);
sub sm_simu_crash(%);
sub current_fh_set($%);
sub error_check($%);
sub nfs_call($%);
sub _cleanup_daemons();
sub compound($%);
sub rpc;
sub mnt_fh;
sub curr_fh;
sub mnt_path;
sub result;
sub nfs_retry($$;);
sub nfs_delay($$;);
sub DESTROY;

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $sudo = '/usr/software/bin/sudo';
my $nlm_pid;
my $nsm_pid;
my $nlm_lock_file = '/tmp/.nlm_lock_file';
my $nsm_lock_file = '/tmp/.nsm_lock_file';
my $nlm_cb_queue  = \@nfs::NfsXdr::nlm_cb_queue;
my $nsm_cb_queue  = \@nfs::NfsXdr::nsm_cb_queue;
my $NLM_PROGRAM   = 100021;
my $NSM_PROGRAM   = 100024;
my $daemon_info;
my $daemon_clients;
my $daemon_cb_queue;
my $daemon_reg_ports;

# Map the callback procedures to its
# corresponding callback function
my %call_back = (0 => sub { return },
                 1 => \&xdr_cb_compound,);
my %common_call_back = (1000 => \&xdr_cb_info,
                        1001 => \&xdr_cb_pid,
                        1002 => \&xdr_cb_port,
                        1003 => \&xdr_cb_exit,
                        2000 => \&xdr_cb_register,);
my %nlm_call_back = (0 => sub { return },
                     5 => \&xdr_cb_nlm_granted,
                     %common_call_back);
my %nsm_call_back = (0 => sub { return },
                     6 => \&xdr_cb_nsm_notify,
                     %common_call_back);

# Retry the NFS request for the following errors
# The NFS3ERR_JUKEBOX is included here to make it clear the error is expected
# a retry of the NFS request
my %retry_errors = (NFS3ERR_JUKEBOX() => 1,);

#---------------------------------------------------------------------

=head2 new

  $nfs_obj = nfs::NfsClient->new(hostname => $host,
                                 version => 2|3
                                 [, proto => 'udp'|'tcp' ]
                                 [, addr_type => 'inet4'|'inet6' ]
                                 [, mnt_ver => 1|2|3 ]
                                 [, mnt_proto => 'udp'|'tcp' ]
                                 [, nlm => 0|1 ]
                                 [, nlm_ver => 1|2|3|4 ]
                                 [, nlm_proto => 'udp'|'tcp' ]
                                 [, nsm => 0|1 ]
                                 [, nsm_proto => 'udp'|'tcp' ]
                                 [, port => $port ]
                                 [, resvport => 0|1 ]
                                 [, xid => $xid ]
                                 [, retry => $retry ]
                                 [, nfs_retry => $nfs_retry ]
                                 [, nfs_delay => $nfs_delay ]
                                 [, reconnect => 0|1 ]
                                 [, timeout => $timeout ]
                                 [, qsize => $qsize ]
                                 [, localaddr => $localaddr ]
                                 [, sec => $authentication ]);

NFS object constructor. It creates the RPC connection object to the
server and initializes the class variables.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : Hostname or ip address of server to connect.

=item version

(Required) : NFS version: 2 or 3.

=item proto

(Optional) : Protocol of connection: 'udp', 'tcp' (default is 'udp').

=item mnt_ver

(Optional) : Mount version to use: 1, 2, or 3.
(default is the same as NFS version).
It is only supported on NFSv2 and NFSv3.

=item mnt_proto

(Optional) : Protocol for mount connection: 'udp', 'tcp'
(default is the same as proto).

=item nlm

(Optional) : Start the NLM daemon (default is 0).

=item nlm_ver

(Optional) : NLM version to use: 1, 2, 3 or 4.
(default is the same as NFS version).
It is only supported on NFSv2 and NFSv3.

=item nlm_proto

(Optional) : Protocol for NLM connection: 'udp', 'tcp'
(default is the same as proto).

=item nsm

(Optional) : Start the NSM daemon (default is 0).

=item nsm_proto

(Optional) : Protocol for NSM connection: 'udp', 'tcp' (default is udp).

=item port

(Optional) : Port number of server to connect
(default is 0 which attempts to connect to the portmapper to
get the port number).

=item resvport

(Optional) : Bind to a reserved port number on the local machine.
A reserved port number is less than or equal to 1024. If this
option is true an attempt to bind to a reserved port number
between 600 and 1023 is made. On most unix machines, root
is the only user able to bind to a reserved port.

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item retry

(Optional) : Number of times to retry each RPC call if response is not
received within the timeout period (default is 0).

=item nfs_retry

(Optional) : Number of times to retry the request if NFS3ERR_JUKEBOX
 is returned (default is 5). It will wait nfs_delay seconds before each retry.

=item nfs_delay

(Optional) : Number of seconds to wait before each retry (default is 1).

=item reconnect

(Optional) : Reconnect if TCP connection is dropped. This only works for
TCP connections and when the retry parameter is non-zero (default is 0)

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item qsize

(Optional) : Size of RPC call queue.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=item addr_type

(Optional) : The type of address, inet4/inet6. If not specified, "ADDR_TYPE" will be used.
This is used only when "hostname" is given but "dstaddr" is not given, hence uses NATE hostobject
to choose the IP address. See nfs::Rpc->name2ip().

=back

=item sec

(Optional) : The type of security to be used. If not specified, it uses "sys" authentication by default.
The valid values are none, sys, krb5, krb5i, krb5p. Please note as of writing this code, krb5p and krb5i 
are not yet supported in c-Mode.

=item Returns

NfsClient object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%) {
    my $class = shift;
    my (%opts) =
      validate(@_,
               {hostname => {type => SCALAR},
                version  => {type => SCALAR, regex => qr/^[23]$/},
                resvport => {type    => SCALAR,
                             regex   => qr/^\d+$/,
                             default => 0},
                proto => {type    => SCALAR,
                          regex   => qr/^(udp|tcp)$/,
                          default => 'udp'},
                mnt_ver => {type     => SCALAR,
                            regex    => qr/^[123]$/,
                            optional => 1},
                mnt_proto => {type     => SCALAR,
                              regex    => qr/^(udp|tcp)$/,
                              optional => 1},
                nlm_ver => {type     => SCALAR,
                            regex    => qr/^[1234]$/,
                            optional => 1},
                nlm_proto => {type     => SCALAR,
                              regex    => qr/^(udp|tcp)$/,
                              optional => 1},
                port => {type    => SCALAR,
                         regex   => qr/^\d+$/,
                         default => 0},
                xid => {type    => SCALAR,
                        regex   => qr/^\d+$/,
                        default => int(rand(1000000000))},
                retry => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                nfs_retry => {type    => SCALAR,
                              regex   => qr/^\d+$/,
                              default => 5},
                nfs_delay => {type    => SCALAR,
                              regex   => qr/^\d+$/,
                              default => 1},
                reconnect => {type    => SCALAR,
                              regex   => qr/^[01]$/,
                              default => 0},
                timeout => {type    => SCALAR,
                            regex   => qr/^\d+$/,
                            default => 30},
                qsize => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                localaddr => {type => SCALAR, optional => 1},
                nlm       => {type => SCALAR, default  => 0},
                nsm       => {type => SCALAR, default  => 0},
                nsm_proto => {type    => SCALAR,
                              regex   => qr/^(udp|tcp)$/,
                              default => 'udp'},
                addr_type => {type => SCALAR, default => nfs::Rpc->addr_type()},
				sec      => { type => SCALAR,
						default => nfs::Rpc->auth_type(),
						regex   => qr/^(sys|krb5|krb5i|krb5p|none)$/,},
               });
    my ($self, %rpc_opts_h);

    # Mount version defaults to current version for NFS
    $opts{mnt_ver} = $opts{version} unless defined($opts{mnt_ver});

    # Mount protocol defaults to current protocol for NFS
    $opts{mnt_proto} = $opts{proto} unless defined($opts{mnt_proto});

    # NLM version defaults 3 for NFSv3 and 1 for NFSv2
    if (!defined($opts{nlm_ver})) {
        $opts{nlm_ver} = ($opts{version} == 2 ? 1 : 4);
    }

    # NLM protocol defaults to current protocol for NFS
    $opts{nlm_proto} = $opts{proto} unless defined($opts{nlm_proto});

    # Pass local address if it is given
    $rpc_opts_h{localaddr} = $opts{localaddr} if defined($opts{localaddr});

	$opts{port} = $opts{port} ? $opts{port} : 2049;

    # Create an rpc object for NFS (100003)
    my $rpc = nfs::Rpc->clnt_create(hostname  => $opts{hostname},
                                    version   => $opts{version},
                                    resvport  => $opts{resvport},
                                    proto     => $opts{proto},
                                    port      => $opts{port},
                                    xid       => $opts{xid},
                                    retry     => $opts{retry},
                                    timeout   => $opts{timeout},
                                    qsize     => $opts{qsize},
                                    reconnect => $opts{reconnect},
                                    addr_type => $opts{addr_type},
                                    program   => 100003,
                                    sec		  => $opts{sec},
                                    %rpc_opts_h);
    $self->{loopback} = $rpc->loopback();
    $loopback = $self->{loopback};

    # Make sure connection is established
    # Burt 178530 null procedure does not return void
    # it returns a null string of 4 bytes
    eval { $rpc->clnt_call(proc => 0) };
    croak $@ if ($@ and $@ !~ /4 extra bytes/);

    # Get local host name
    $opts{localname} = (POSIX::uname())[1];

    $self = bless {%opts,
                   rpc      => $rpc,
                   mnt_port => 0,
                   nlm_port => 0,
                   nsm_port => 0,
                   nlm_lprt => 0,
                   nsm_lprt => 0,
                   sm_state => 1}, $class;

    if ($self->{nlm} || $self->{nsm}) {

        # Trap the INT signal to cleanup the daemons
        $SIG{INT} = sub { exit };
    }

    if ($self->{nlm}) {

        # Register the NLM service
        $self->{nlm_id}   = time + int(rand(1000000000));
        $self->{callback} = $self->_nlm_start();
    }

    if ($self->{nsm}) {

        # Register the NSM service
        $self->{nsm_id}       = time + int(rand(1000000000));
        $self->{nsm_callback} = $self->_nsm_start();
    }

    if ($self->{nlm} || $self->{nsm}) {

        # Set the default monitor host for NLM and NSM
        $self->{mon_host} =
          defined($self->{localaddr}) ? $self->{localaddr} : $self->{localname};
    }

    # Return NFS object
    return $self;
} ## end sub new($%)

#---------------------------------------------------------------------
# Internal function
# This function is a daemon which serves as a proxy for requests
#---------------------------------------------------------------------
sub _daemon(%) {
    my %opts = @_;
    my (@srpc, $rpc,      $version, $proto,     $port,  $out,  $item);
    my ($read, $read_set, $write,   $write_set, $error, $sock, $pid);
    $Log->enter();

    # Make this process a daemon: change to the root directory,
    # close open descriptors, fork to detach from parent and
    # become the session leader
    umask 0;
    chdir '/' or croak "Can't chdir to /: $!";
    CORE::open STDIN,  '</dev/null' or croak "Can't read /dev/null: $!";
    CORE::open STDOUT, '>/dev/null' or croak "Can't write to /dev/null: $!";
    CORE::open STDERR, '>/dev/null' or croak "Can't write to /dev/null: $!";
    defined($pid = fork) or croak "Can't fork: $!";

    if (!$pid) {

        # Open the lock file with O_WRONLY|O_TRUNC
        # to write the real process id of the daemon
        sysopen(FH, $opts{lockfile}, 513)
          or croak "Unable to create lock file $opts{lockfile}: $!";
        print FH $$;
        CORE::close(FH);
    } else {

        # This is the parent of the daemon sleep for a while
        # to prevent others from starting another daemon
        sleep 10;
        exit(0);
    }

    # Become the session leader
    POSIX::setsid();

    # Catch the SIGHUP(1) signal for cleanup
    $SIG{HUP} = sub { exit };

    # Renice process so daemon has
    # higher priority than other processes
    $Log->debug("$sudo renice -10 $$");
    qx($sudo renice -10 $$);

    # Instantiate read and write select objects
    require IO::Select;
    $read_set  = IO::Select->new();
    $write_set = IO::Select->new();

    # Get list of registered ports
    $out               = nfs::Rpc::pmap_dump();
    @$daemon_reg_ports = (@$out);

    # Get port number of daemon
    $port = eval {
        nfs::Rpc::pmap_getport(host    => "$loopback",
                               program => $opts{program},
                               version => 0,
                               proto   => 'udp');
    };

    die "nfs::Rpc::pmap_getport failed: $@\n" if $@;

    if ($port) {

        # Unregister current program
        $out = nfs::Rpc::pmap_unset(program => $opts{program},
                                    version => 0,
                                    proto   => 'udp',
                                    port    => $port);
    }

    foreach $proto ('udp', 'tcp') {
        push @srpc,
          nfs::Rpc->svc_create(program  => $opts{program},
                               timeout  => $opts{timeout},
                               version  => $proto eq 'udp' ? 0 : 1,
                               register => $proto eq 'udp' ? 1 : 0,
                               proto    => $proto,
                               versions => [1, 2, 3, 4],
                               dispatch => $opts{dispatch});

        # Add service socket to select
        $read_set->add($srpc[-1]->{lsock});

        # Add port number to info structure
        $daemon_info->{$proto} = $srpc[-1]->{port};
    } ## end foreach $proto ('udp', 'tcp')

    while (1) {
        eval {

            # Add client sockets to select
            foreach $item (keys %{$daemon_clients}) {
                if (!exists($daemon_clients->{$item}{rpc})) {
                    $port    = $daemon_clients->{$item}{port};
                    $proto   = $daemon_clients->{$item}{proto};
                    $version = $daemon_clients->{$item}{version};

                    # Create RPC connection to proxy clients
                    $rpc =
                      nfs::Rpc->clnt_create(hostname => "$loopback",
                                            program  => $opts{program},
                                            version  => $version,
                                            addr_type=> ($opts{addr_type} ? $opts{addr_type} : nfs::Rpc->addr_type()),
                                            proto    => $proto,
                                            port     => $port,
                                            timeout  => $opts{timeout});
                    $daemon_clients->{$item}{rpc} = $rpc;
                    $write_set->add($rpc->{sock});
                } ## end if (!exists($daemon_clients...
            } ## end foreach $item (keys %{$daemon_clients...

            # Wait until a socket is ready
            ($read, $write, $error) =
              IO::Select->select($read_set, $write_set, undef, undef);

            foreach $sock (@$read) {

                # A callback is ready
                foreach $rpc (@srpc) {
                    if ($sock == $rpc->{lsock}) {
                        $out = eval { $rpc->svc_getreqset() };
                        last;
                    }
                }
            } ## end foreach $sock (@$read)

            if (@$write) {

                # Process callback queue
                &{$opts{qprocess}}($write, $write_set);
            }
        };
    } ## end while (1)

    $Log->exit();
    return;
} ## end sub _daemon(%)

#---------------------------------------------------------------------
# Internal function
# This function starts a daemon
#---------------------------------------------------------------------
sub _start_daemon($%) {
    my ($self, %opts) = @_;
    my ($rpc, $srpc, $info, $item, $out, $port, $pid);
    my $retry     = 10;
    my $register  = 1;
    my $service   = $opts{service};
    my $daemon    = $opts{daemon};
    my $dispatch  = $opts{dispatch};
    my $spid      = $opts{spid};
    my $program   = $opts{program};
    my $proto     = $opts{proto};
    my $version   = $opts{version};
    my $lock_file = $opts{lock_file};
    my $save_port = $opts{save_port};
    my $timeout   = $self->{timeout};
    $Log->enter();

    while (!defined($info)) {

        # Check for current process registered on local machine
        eval {

            # Create an rpc object for program on local machine
            $rpc = nfs::Rpc->clnt_create(hostname => "$loopback",
                                         program  => $program,
                                         version  => 0,
                                         addr_type=> $opts{addr_type} ? $opts{addr_type} : $self->{addr_type},
                                         proto    => 'udp',
                                         timeout  => $timeout);

            # Call the null procedure to check if service is alive
            $info = $rpc->clnt_call(proc => 1000, out => \&xdr_daemon_info);
        };
        if (   $@ =~ /Could not get port/
            or $@ =~ /PROG_MISMATCH/
            or $@ =~ /Connection refused/
            or $@ =~ /PROC_UNAVAIL/
            or $@ =~ /AUTH_BADCRED/) {
            croak "Could not start $service daemon: $@" if $retry < 0;
            $Log->debug("Service daemon is not running yet, error rcvd: $@");

            # Check if daemon is being started by another process
            if (-e $lock_file) {
                $Log->debug(
                     "The lock file exists, check if process is still running");
                CORE::open(FH, "<$lock_file")
                  or croak "Unable of open $service lock file: $!";
                $pid = <FH>;
                CORE::close(FH);
                chomp($pid);
                if (!kill 0, $pid) {
                    unlink $lock_file;
                    undef $pid;
                }
            } ## end if (-e $lock_file)

            if (!defined($pid)) {
                $Log->debug("Daemon is not running: create the lock file");

                # O_WRONLY|O_CREAT|O_EXCL so only one process starts it
                sysopen(FH, $lock_file, 193)
                  or croak
                  "Unable to create $service lock file ($lock_file): $!";
                print FH $$;
                CORE::close(FH);

                $Log->debug("Fork to start daemon's parent process");
                defined($pid = fork) or croak "Can't fork: $!";

                if (!$pid) {

                    # This is the daemon's parent process
                    # Reset global variables so this process when it
                    # exits does not kill the previous daemon started
                    # by this process
                    undef $nlm_pid;
                    undef $nsm_pid;

                    $Log->debug("Calling function to start daemon");

                    # Call function to start correct daemon
                    eval "&$daemon(timeout => $timeout)";

                    # Make sure daemon exits
                    exit(0);
                } ## end if (!$pid)

                # Wait for daemon's parent to finish
                waitpid($pid, 0);
            } ## end if (!defined($pid))

            # Wait for service to start
            sleep 2;
            $retry--;
        } elsif ($@) {
            croak "Unexpected error received: $@";
        }
    } ## end while (!defined($info))

    $Log->debug("Check if there is a need to register with the portmapper");
    foreach $item (@{$info->{ports}}) {
        if ($item->{version} == $version and $item->{proto} eq $proto) {
            $Log->debug("No need to register for current version and protocol");
            $register = 0;
            last;
        }
    }

    if ($register) {

        # Get list of registered ports
        $out = nfs::Rpc::pmap_dump();

        # Check if current program needs to be unregistered
        foreach $item (@$out) {
            if (    $item->{program} == $program
                and $item->{version} == $version
                and $item->{proto} eq $proto) {

                # Unregister current program
                $out =
                  nfs::Rpc::pmap_unset(program => $program,
                                       version => $version,
                                       proto   => $proto,
                                       port    => $item->{port});
                last;
            } ## end if ($item->{program} ==...
        } ## end foreach $item (@$out)

        # Port number of server
        $port = $info->{$proto};

        # Register program
        $out = nfs::Rpc::pmap_set(program => $program,
                                  version => $version,
                                  proto   => $proto,
                                  port    => $port);

        # Create an rpc object for program
        $rpc = nfs::Rpc->clnt_create(hostname => "$loopback",
                                     program  => $program,
                                     version  => $version,
                                     addr_type=> $opts{addr_type} ? $opts{addr_type} : $self->{addr_type},
                                     proto    => $proto,
                                     timeout  => $timeout);

        # Save port number so it can be unregistered
        $rpc->clnt_call(proc => 1002, arg => xdr_uint(encode => $port));

        # Save service port number
        $self->{$save_port} = $rpc->{port};
    } ## end if ($register)

    if (!$$spid) {

        # Re-connect to service again for tcp
        if ($proto eq 'tcp') {

            # Create an rpc object for program
            $rpc = nfs::Rpc->clnt_create(hostname => "$loopback",
                                         program  => $program,
                                         version  => $version,
                                         addr_type=> $opts{addr_type} ? $opts{addr_type} : $self->{addr_type},
                                         proto    => $proto,
                                         port     => $self->{$save_port},
                                         timeout  => $timeout);

            # Save service port number
            $self->{$save_port} = $rpc->{port} unless $self->{$save_port};
        } ## end if ($proto eq 'tcp')

        # Register this process with the daemon
        $$spid = $rpc->clnt_call(proc => 1001, out => \&xdr_uint);
    } ## end if (!$$spid)

    # Create service port but without registering with the portmapper
    $srpc = nfs::Rpc->svc_create(program  => $program,
                                 version  => $version,
                                 proto    => $proto,
                                 timeout  => $timeout,
                                 dispatch => $dispatch,
                                 register => 0);
    $Log->exit();

    return $srpc;
} ## end sub _start_daemon($%)

#---------------------------------------------------------------------
# Internal function
# This function stops a daemon if this object is the last one to use it
#---------------------------------------------------------------------
sub _stop_daemon($$$;) {
    my ($program, $pid, $lock_file) = @_;
    my ($rpc, $item, $out);
    $Log->enter();

    eval {

        # Create an rpc object for program
        $rpc = nfs::Rpc->clnt_create(hostname => "$loopback",
                                     program  => $program,
                                     version  => 0,
                                     proto    => 'udp',
                                     timeout  => 30);

        # Check if daemon is actually running
        $out = $rpc->clnt_call(proc => 1003, out => \&xdr_daemon_info);
    };

    # Check if this is the last process using the daemon
    if (defined($out) && $out->{used} == 0) {

        # Unregister main program
        push @{$out->{ports}},
          { proto   => 'udp',
            port    => $rpc->{port},
            version => 0,
            restore => 0};

        # Unregister all programs first
        foreach $item (@{$out->{ports}}) {
            if (!$item->{restore}) {
                eval {
                    nfs::Rpc::pmap_unset(program => $program,
                                         version => $item->{version},
                                         proto   => $item->{proto},
                                         port    => $item->{port});
                };
            }
        } ## end foreach $item (@{$out->{ports...

        # Register original programs
        foreach $item (@{$out->{ports}}) {
            if ($item->{restore}) {
                eval {
                    nfs::Rpc::pmap_set(program => $program,
                                       version => $item->{version},
                                       proto   => $item->{proto},
                                       port    => $item->{port});
                };
            }
        } ## end foreach $item (@{$out->{ports...

        # Send signal to terminate child process
        kill 1, $pid;

        # Remove lock file
        unlink $lock_file;
    } ## end if (defined($out) && $out...

    $Log->exit();
    return;
} ## end sub _stop_daemon($$$;)

#---------------------------------------------------------------------
# Internal function
# This function starts the NLM daemon
#---------------------------------------------------------------------
sub _nlm_start($;) {
    my $self = shift;

    return
      $self->_start_daemon(service   => 'NLM',
                           daemon    => '_nlm_daemon',
                           dispatch  => \%nlm_call_back,
                           spid      => \$nlm_pid,
                           program   => $NLM_PROGRAM,
                           proto     => $self->{nlm_proto},
                           version   => $self->{nlm_ver},
                           lock_file => $nlm_lock_file,
                           save_port => 'nlm_lprt');

} ## end sub _nlm_start($;)

#---------------------------------------------------------------------
# Internal function
# This function starts the NSM daemon
#---------------------------------------------------------------------
sub _nsm_start($;) {
    my $self = shift;

    return
      $self->_start_daemon(service   => 'NSM',
                           daemon    => '_nsm_daemon',
                           dispatch  => \%nsm_call_back,
                           spid      => \$nsm_pid,
                           program   => $NSM_PROGRAM,
                           proto     => $self->{nsm_proto},
                           version   => 1,
                           lock_file => $nsm_lock_file,
                           save_port => 'nsm_lprt');

} ## end sub _nsm_start($;)

#---------------------------------------------------------------------
# Internal function
# This function is the NLM daemon
#---------------------------------------------------------------------
sub _nlm_daemon(%) {
    my %opts = @_;
    $Log->enter();

    $opts{program}  = $NLM_PROGRAM;
    $opts{dispatch} = \%nlm_call_back;
    $opts{qprocess} = \&_nlm_process_queue;
    $opts{lockfile} = $nlm_lock_file;

    # Set globals for NLM
    $daemon_info      = \%nfs::NfsXdr::nlm_info;
    $daemon_clients   = \%nfs::NfsXdr::nlm_clients;
    $daemon_cb_queue  = \@nfs::NfsXdr::nlm_cb_queue;
    $daemon_reg_ports = \@nfs::NfsXdr::nlm_reg_ports;

    # Set globals for NLM on nfs::NfsXdr package
    $nfs::NfsXdr::daemon_info      = \%nfs::NfsXdr::nlm_info;
    $nfs::NfsXdr::daemon_clients   = \%nfs::NfsXdr::nlm_clients;
    $nfs::NfsXdr::daemon_program   = $NLM_PROGRAM;
    $nfs::NfsXdr::daemon_reg_ports = \@nfs::NfsXdr::nlm_reg_ports;

    $Log->exit();
    return _daemon(%opts);
} ## end sub _nlm_daemon(%)

#---------------------------------------------------------------------
# Internal function
# This function is the NSM daemon
#---------------------------------------------------------------------
sub _nsm_daemon(%) {
    my %opts = @_;
    $Log->enter();

    $opts{program}  = $NSM_PROGRAM;
    $opts{dispatch} = \%nsm_call_back;
    $opts{qprocess} = \&_nsm_process_queue;
    $opts{lockfile} = $nsm_lock_file;

    # Set globals for NSM
    $daemon_info      = \%nfs::NfsXdr::nsm_info;
    $daemon_clients   = \%nfs::NfsXdr::nsm_clients;
    $daemon_cb_queue  = \@nfs::NfsXdr::nsm_cb_queue;
    $daemon_reg_ports = \@nfs::NfsXdr::nsm_reg_ports;

    # Set globals for NSM on nfs::NfsXdr package
    $nfs::NfsXdr::daemon_info      = \%nfs::NfsXdr::nsm_info;
    $nfs::NfsXdr::daemon_clients   = \%nfs::NfsXdr::nsm_clients;
    $nfs::NfsXdr::daemon_program   = $NSM_PROGRAM;
    $nfs::NfsXdr::daemon_reg_ports = \@nfs::NfsXdr::nsm_reg_ports;

    $Log->exit();
    return _daemon(%opts);
} ## end sub _nsm_daemon(%)

#---------------------------------------------------------------------
# Internal function
# This function processes the NLM callback queue
#---------------------------------------------------------------------
sub _nlm_process_queue($$;) {
    my ($write, $write_set) = @_;
    my ($item, $index, $idx, $svid, $rpc, $args, $out);
    $Log->enter();

    # Process callback queue
    $index = 0;
    while ($index < @{$daemon_cb_queue}) {
        $item = $daemon_cb_queue->[$index];
        $svid = $item->{nlm_lock}{svid};

        $idx = 0;

        # Find if client is ready for this callback
        while ($idx < @$write
               && (!exists($daemon_clients->{$svid})
                   || $daemon_clients->{$svid}{rpc}{sock} != $write->[$idx])
          ) {
            $idx++;
        }

        if ($idx == @$write) {

            # Could not find proper client for given
            # queue item, check the next item
            $index++;
            next;
        }

        # Client is ready to process the callback
        $rpc = $daemon_clients->{$svid}{rpc};

        eval {
            $args =
                xdr_opaque(encode => $item->{cookie})
              . xdr_uint(encode => $item->{exclusive})
              . xdr_nlm_lock(encode  => $item->{nlm_lock},
                             version => $rpc->{version});

            $out = $rpc->clnt_call(proc => NLM_GRANTED,
                                   arg  => $args,
                                   out  => \&xdr_nlm_stat);

            # Remove callback from list
            splice @{$daemon_cb_queue}, $index, 1;
        };

        # Remove socket from select list
        $write_set->remove($rpc->{sock});
        splice @{$write}, $idx, 1;
        undef $rpc;
        delete $daemon_clients->{$svid};
    } ## end while ($index < @{$daemon_cb_queue...

    $Log->exit();
    return;
} ## end sub _nlm_process_queue($$;)

#---------------------------------------------------------------------
# Internal function
# This function processes the NSM callback queue
#---------------------------------------------------------------------
sub _nsm_process_queue($$;) {
    my ($write, $write_set) = @_;
    my ($item, $index, $idx, $mon_name, $svid, $rpc, $args, $out);
    $Log->enter();

    # Process callback queue
    $index = 0;
    while ($index < @{$daemon_cb_queue}) {
        $item     = $daemon_cb_queue->[$index];
        $mon_name = $item->{mon_name};
        $svid     = (keys %$daemon_clients)[0];

        $idx = 0;

        # Find if client is ready for this callback
        while (   $idx < @$write
               && $daemon_clients->{$svid}{rpc}{sock} != $write->[$idx]) {
            $idx++;
        }

        if ($idx == @$write) {

            # Could not find proper client for given
            # queue item, check the next item
            $index++;
            next;
        }

        # Client is ready to process the callback
        $rpc = $daemon_clients->{$svid}{rpc};

        eval {
            $args =
                xdr_opaque(encode => $item->{mon_name})
              . xdr_uint(encode => $item->{state});

            $out = $rpc->clnt_call(proc => 6,
                                   arg  => $args);

            # Remove callback from list
            splice @{$daemon_cb_queue}, $index, 1;
        };

        # Remove socket from select list
        $write_set->remove($rpc->{sock});
        splice @{$write}, $idx, 1;
        undef $rpc;
        delete $daemon_clients->{$svid};
    } ## end while ($index < @{$daemon_cb_queue...

    $Log->exit();
    return;
} ## end sub _nsm_process_queue($$;)

#---------------------------------------------------------------------

=head2 nlm_callback_process

  $nfs_obj->nlm_callback_process([ timeout => $time_in_mseconds ]);

Process a pending NLM callback. It waits until a callback is
received.

=over

=item Arguments

=over

=item timeout

(Optional) : Timeout value in milliseconds to check if any
callback is pending. A value of 0 returns immediately if no
callback is currently pending. If argument is not specified
it waits until a callback is ready with no time limit.

=back

=item Returns

True if callback is received and processed correctly.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nlm_callback_process($;%) {
    my $self = shift;
    $Log->enter();

    my (%opts) =
      validate(@_,
               {timeout => {type => SCALAR, regex => qr/^\d+$/, optional => 1},}
               );
    my $ver = $self->{version};
    my ($rpc, $item, $args, $out, $index, %options);

    if (!defined($self->{callback})) {
        croak "Must create object with NLM enabled "
          . "in order to use this function";
    }

    # Add the timeout argument if it is defined
    $options{timeout} = $opts{timeout} if defined($opts{timeout});

    # Connect to NLM service to register the service
    $rpc = nfs::Rpc->clnt_create(hostname => "$self->{loopback}",
                                 program  => $NLM_PROGRAM,
                                 addr_type=> $self->{addr_type},
                                 version  => $self->{nlm_ver},
                                 proto    => $self->{nlm_proto},
                                 port     => $self->{nlm_lprt});

    # Save NLM port number
    $self->{nlm_lprt} = $rpc->{port} unless $self->{nlm_lprt};

    # Register this object with the NLM daemon
    $args =
        xdr_opaque(encode => $self->{nlm_id})
      . xdr_uint(encode => $self->{callback}{port});
    $out = $rpc->clnt_call(proc => 2000, arg => $args, out => \&xdr_uint);

    # Wait until a callback is received
    $out = $self->{callback}->svc_getreqset(%options);

    # Process callback queue
    $index = 0;
    while ($index < @{$nlm_cb_queue}) {
        $item = $nlm_cb_queue->[$index];

        if ($item->{nlm_lock}{svid} == $self->{nlm_id}) {

            # Remove callback from list
            splice @{$nlm_cb_queue}, $index, 1;
            return $item;
        }

        $index++;
    } ## end while ($index < @{$nlm_cb_queue...

    $Log->exit();
    return;
} ## end sub nlm_callback_process($;%)

#---------------------------------------------------------------------

=head2 nsm_callback_process

  $nfs_obj->nsm_callback_process([ timeout => $time_in_mseconds ]);

Process a pending NSM callback. It waits until a callback is
received.

=over

=item Arguments

=over

=item timeout

(Optional) : Timeout value in milliseconds to check if any
callback is pending. A value of 0 returns immediately if no
callback is currently pending. If argument is not specified
it waits until a callback is ready with no time limit.

=back

=item Returns

A Reference to a list of callback results and undef if a callback
is not received.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nsm_callback_process($;%) {
    my $self = shift;
    $Log->enter();

    my (%opts) =
      validate(@_,
               {timeout => {type => SCALAR, regex => qr/^\d+$/, optional => 1},}
               );
    my $ver = $self->{version};
    my ($rpc, $args, $item, $itm, $index, $idx, @res, $out, %options);

    if (!defined($self->{nsm_callback})) {
        croak "Must create object with NSM enabled "
          . "in order to use this function";
    }

    # Add the timeout argument if it is defined
    $options{timeout} = $opts{timeout} if defined($opts{timeout});

    # Connect to NSM service to register the service
    $rpc = nfs::Rpc->clnt_create(hostname => "$self->{loopback}",
                                 program  => $NSM_PROGRAM,
                                 version  => 1,
                                 addr_type=> $self->{addr_type},
                                 proto    => $self->{nsm_proto},
                                 port     => $self->{nsm_lprt});

    # Save NSM port number
    $self->{nsm_lprt} = $rpc->{port} unless $self->{nsm_lprt};

    # Register this object with the NSM daemon
    $args =
        xdr_opaque(encode => $self->{nsm_id})
      . xdr_uint(encode => $self->{nsm_callback}{port});
    $out = $rpc->clnt_call(proc => 2000, arg => $args, out => \&xdr_uint);

    # Wait until a callback is received
    $out = $self->{nsm_callback}->svc_getreqset(%options);
    return if !$out;

    $index = 0;
    while ($index < @{$nsm_cb_queue}) {
        $item = $nsm_cb_queue->[$index];
        $idx  = $index + 1;
        $index++;
        while ($idx < @{$nsm_cb_queue}) {
            $itm = $nsm_cb_queue->[$idx];

            # Compare callbacks
            if (   $item->{mon_name} eq $itm->{mon_name}
                && $item->{state} == $itm->{state}) {

                # Remove duplicate callback
                splice(@{$nsm_cb_queue}, $idx, 1);
            } else {
                $idx++;
            }
        } ## end while ($idx < @{$nsm_cb_queue...

        # Check if this item on the queue has been processed by this client
        if (exists($item->{list}{$self->{nsm_id}})
            && $item->{list}{$self->{nsm_id}}) {
            next;
        }

        # Add queue item to the results and mark it as read
        push @res, {mon_name => $item->{mon_name}, state => $item->{state}};
        $item->{list}{$self->{nsm_id}} = 1;
    } ## end while ($index < @{$nsm_cb_queue...

    $Log->exit();
    return \@res;
} ## end sub nsm_callback_process($;%)

#---------------------------------------------------------------------

=head2 null

  $nfs_obj->null();

Calls the NULL procedure on the server.  This procedure takes
no arguments and returns nothing.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub null($;) {
    my $self = shift;
    my $rpc  = $self->{rpc};

    # Call null procedure on server
    return $rpc->clnt_call(proc => 0);
}

#---------------------------------------------------------------------

=head2 getattr

  $out = $nfs_obj->getattr([ fhandle => $fhandle ]
                           [, attr => \@fattr ]);

Calls the GETATTR procedure on the server.
Retrieves the attributes of the specified file system object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the object whose attributes
are to be retrieved (default is the current filehandle).

=item attr

(Optional) : Placeholder for look of NFSv4. No use of this option for NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub getattr($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                attr    => {
                         type    => ARRAYREF,
                         optional => 1,
                  },});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC_GETATTR,
                           arg  => $fhandle,
                           out  => sub { xdr_getattr(version => $ver, @_) }
        );

    # Check for errors
    $self->error_check(name => 'GETATTR', result => $out);

    return $out;
} ## end sub getattr($%)

#---------------------------------------------------------------------

=head2 setattr

  $out = $nfs_obj->setattr(attr => \%sattr
                           [, fhandle => $fhandle ]
                           [, ctime => \%nfstime3 ]);

Calls the SETATTR procedure on the server.
Changes one or more attributes of a file system object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the object whose attributes
are to be changed (default is the current filehandle).

=item attr

(Required) : A reference to a hash of attributes to be changed.

=item ctime

(Optional) : A reference to an nfstime3 hash. If provided, the
server is to verify that this matches the ctime for the object.
If there is no match, the server preserves the object attributes
and returns an error. Not used in NFSv2.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setattr($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR,  default  => $self->{curr_fh}},
                attr    => {type => HASHREF},
                ctime   => {type => HASHREF, optional => 1},});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object and
    # the sattr structure into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $sattr = xdr_sattr(encode => $opts{attr}, version => $ver);

    if ($ver == 3) {
        if (defined($opts{ctime})) {

            # Verify ctime
            $sattr .=
                xdr_uint(encode => 1)
              . xdr_nfstime(encode => $opts{ctime}, version => $ver);
        } else {

            # Do not verify ctime
            $sattr .= xdr_uint(encode => 0);
        }
    } ## end if ($ver == 3)

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC_SETATTR,
                           arg  => $fhandle . $sattr,
                           out  => sub { xdr_setattr(version => $ver, @_) }
        );

    # Check for errors
    $self->error_check(name => 'SETATTR', result => $out);

    return $out;
} ## end sub setattr($%)

#---------------------------------------------------------------------

=head2 lookup

  $out = $nfs_obj->lookup(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]);

Calls the LOOKUP procedure on the server.
Searches a directory for a specific name and returns the
filehandle for the corresponding file system object.
It sets the current filehandle.

If the name is '.' and the dhandle is WEBNFS2_FHANDLE for NFSv2
or WEBNFS3_FHANDLE for NFSv3, it effectively mounts the WebNFS
path so it replaces the object's mnt_fh and mnt_path as if the
mount method was called.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to search
(default is the current filehandle).

=item name

(Required) : The name of the object to find.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lookup($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $fhandle);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory and
    # the name of object into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_LOOKUP : NFSPROC3_LOOKUP;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $dhandle . $name,
                           out  => sub { xdr_lookup(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'LOOKUP', result => $out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[2]{result}{fhandle} if $ver == 4;

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    # LOOKUP with a filehandle of all 0's (v2) or null (v3)
    # returns the filehandle of the specified file relative
    # to the WebNFS path -- this is a WebNFS mount if file is '.'
    if ($opts{name} eq '.'
        && (   ($ver == 2 && $opts{dhandle} eq WEBNFS2_FHANDLE)
            || ($ver == 3 && $opts{dhandle} eq WEBNFS3_FHANDLE))
      ) {

        # Save mount path
        $self->{mnt_path} = $opts{path};

        # Save the root filehandle on the NfsClient object
        $self->{mnt_fh} = $out->{fhandle};
    } ## end if ($opts{name} eq '.'...

    return $out;
} ## end sub lookup($%)

#---------------------------------------------------------------------

=head2 access

  $out = $nfs_obj->access(access => $access
                          [, fhandle => $fhandle ]);

Calls the ACCESS procedure on the server.
Determines the access rights that a user, as identified by the
credentials in the request, has with respect to a file system object.
This function is not supported on NFSv2.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file system object to which access
is to be checked (default is the current filehandle).

=item access

(Required) : Bit mask of access permissions to check.
Valid access mode bits: ACCESS_READ, ACCESS_LOOKUP, ACCESS_MODIFY,
ACCESS_EXTEND, ACCESS_DELETE and ACCESS_EXECUTE.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub access($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                access  => {type => SCALAR, regex   => qr/^\d+$/},});
    my $out;
    my $ver = $self->{version};

    if ($ver == 2) {

        # ACCESS is only supported on NFSv3 and v4
        croak "ACCESS is not supported on NFSv2";
    }

    # Convert filehandle of object and
    # the requested access bit mask into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $access = xdr_uint(encode => $opts{access});

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC3_ACCESS,
                           arg  => $fhandle . $access,
                           out  => sub { xdr_access(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'ACCESS', result => $out);

    return $out;
} ## end sub access($%)

#---------------------------------------------------------------------

=head2 readlink

  $out = $nfs_obj->readlink([ fhandle => $fhandle ]);

Calls the READLINK procedure on the server.
Reads the pathname stored in a symbolic link.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of symbolic link
(default is the current filehandle).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub readlink($%) {
    my $self = shift;
    my (%opts) =
      validate(@_, {fhandle => {type => SCALAR, default => $self->{curr_fh}},});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC_READLINK,
                           arg  => $fhandle,
                           out  => sub { xdr_readlink(version => $ver, @_) }
        );

    # Check for errors
    $self->error_check(name => 'READLINK', result => $out);

    return $out;
} ## end sub readlink($%)

#---------------------------------------------------------------------

=head2 mkdir

  $out = $nfs_obj->mkdir(name => $name [, namelen => $namelen ],
                         attr => \%sattr
                         [, dhandle => $fhandle ]);

Calls the MKDIR procedure on the server.
Creates a new directory.
It sets the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the new directory
is to be created (default is the current filehandle).

=item name

(Required) : The name of the new directory.

=item attr

(Required) : The initial attributes for the new directory.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mkdir($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                attr    => {type => HASHREF},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    $Log->enter(Dumper(\%opts));
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory, name of directory
    # to create and initial attributes into XDR format
    my $args =
        xdr_fhandle(encode => $opts{dhandle}, version => $ver)
      . xdr_opaque(encode => $opts{name}, %len)
      . xdr_sattr(encode => $opts{attr}, version => $ver);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_MKDIR : NFSPROC3_MKDIR;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $args,
                           out  => sub { xdr_mkdir(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'MKDIR', result => $out);

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    $Log->exit(Dumper($out));
    return $out;
} ## end sub mkdir($%)

#---------------------------------------------------------------------

=head2 symlink

  $out = $nfs_obj->symlink(name => $name [, namelen => $namelen ],
                           data => $linktext [, datalen => $datalen ]
                           [, dhandle => $fhandle ]);

Calls the SYMLINK procedure on the server.
Creates a new symbolic link.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the new symbolic
link is to be created (default is the current filehandle).

=item name

(Required) : The name of the new symbolic link.

=item data

(Required) : The string containing the symbolic link data.

=item namelen

(Optional) : Force the length of name to this number.

=item datalen

(Optional) : Force the length of data to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub symlink($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                data    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                datalen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver  = $self->{version};
    my %nlen = (len => $opts{namelen}) if defined($opts{namelen});
    my %dlen = (len => $opts{datalen}) if defined($opts{datalen});

    # Convert filehandle of directory and
    # name of symbolic link into XDR format
    my $args =
        xdr_fhandle(encode => $opts{dhandle}, version => $ver)
      . xdr_opaque(encode => $opts{name}, %nlen);

    # Convert the link data and initial attributes into XDR format
    my $data = xdr_opaque(encode => $opts{data}, %dlen);
    my $sattr = xdr_sattr(encode => {}, version => $ver);

    # Append the link data and the initial
    # attributes to the arguments
    if ($ver == 2) {
        $args .= $data . $sattr;
    } elsif ($ver == 3) {
        $args .= $sattr . $data;
    }

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_SYMLINK : NFSPROC3_SYMLINK;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $args,
                           out  => sub { xdr_symlink(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'SYMLINK', result => $out);

    return $out;
} ## end sub symlink($%)

#---------------------------------------------------------------------

=head2 mknod

  $out = $nfs_obj->mknod(name => $name [, namelen => $namelen ],
                         type => $type,
                         attr => \%sattr
                         [, dhandle => $fhandle ]
                         [, major => $major, minor => $min ]);

Calls the MKNOD procedure on the server.
Creates a new special file: device, sockets or named pipes.
This function is not supported on NFSv2 and v4;
use create instead.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the new special
file is to be created (default is the current filehandle).

=item name

(Required) : The name of the new special file.

=item type

(Required) : The type of special file to be created: NFBLK, NFCHR,
NFSOCK or NFFIFO.

=item attr

(Required) : The initial attributes for the new special file.

=item major

(Optional) : The major device number, only for type NFBLK or NFCHR.

=item minor

(Optional) : The minor device number, only for type NFBLK or NFCHR.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mknod($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                type    => {type => SCALAR, regex   => qr/^\d+$/},
                attr    => {type => HASHREF},
                major   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                minor   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    if ($ver != 3) {

        # MKNOD is not supported on NFSv2 and v4,
        # use CREATE instead
        croak "MKNOD is not supported on NFSv$ver";
    }

    # Convert filehandle of directory, name of object, type of file
    # to create and the initial attributes into XDR format
    my $args =
        xdr_fhandle(encode => $opts{dhandle}, version => $ver)
      . xdr_opaque(encode => $opts{name}, %len)
      . xdr_uint(encode => $opts{type})
      . xdr_sattr(encode => $opts{attr}, version => $ver);

    if ($opts{type} == NFBLK or $opts{type} == NFCHR) {

        # Append the major and minor numbers to
        # the arguments if creating a device file
        $args .= xdr_specdata(encode => {specdata1 => $opts{major},
                                         specdata2 => $opts{minor}});
    }

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC3_MKNOD,
                           arg  => $args,
                           out  => \&xdr_mknod);

    # Check for errors
    $self->error_check(name => 'MKNOD', result => $out);

    return $out;
} ## end sub mknod($%)

#---------------------------------------------------------------------

=head2 remove

  $out = $nfs_obj->remove(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]);

Calls the REMOVE procedure on the server.
Removes an entry from a directory. If the entry in the directory
was the last reference to the corresponding file system object,
the object may be destroyed.
Removing a directory object is only supported on NFSv4.
For NFSv2 and v3 use rmdir instead.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the entry
is to be removed (default is the current filehandle).

=item name

(Required) : The name of the entry to be removed.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub remove($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory and
    # name of object into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_REMOVE : NFSPROC3_REMOVE;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $dhandle . $name,
                           out  => sub { xdr_remove(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'REMOVE', result => $out);

    return $out;
} ## end sub remove($%)

#---------------------------------------------------------------------

=head2 rename

  $out = $nfs_obj->rename(fromdir => $f_fhandle,
                          fromname => $o_name [, fromlen => $fromlen ],
                          todir => $t_handle,
                          toname => $n_name [, tolen => $tolen ]);

Calls the RENAME procedure on the server.
Renames the directory entry identified by 'fromname' in the
directory 'fromdir' to 'toname' in the directory 'todir'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fromdir

(Required) : Filehandle of the directory from which the entry
is to be renamed.

=item fromname

(Required) : The name of the entry to be renamed.

=item todir

(Required) : Filehandle of the directory to which the entry
is to be renamed.

=item toname

(Required) : The new name for the object.

=item fromlen

(Optional) : Force the length of fromname to this number.

=item tolen

(Optional) : Force the length of toname to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub rename($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fromdir  => {type => SCALAR},
                fromname => {type => SCALAR},
                todir    => {type => SCALAR},
                toname   => {type => SCALAR},
                fromlen  => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                tolen    => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
               });
    my $out;
    my $ver  = $self->{version};
    my %flen = (len => $opts{fromlen}) if defined($opts{fromlen});
    my %tlen = (len => $opts{tolen}) if defined($opts{tolen});

    # Convert filehandle of the from directory, the original object name,
    # the filehandle of the to directory and the new name of object
    # into XDR format
    my $fromdir = xdr_fhandle(encode => $opts{fromdir}, version => $ver);
    my $fromname = xdr_opaque(encode => $opts{fromname}, %flen);
    my $todir = xdr_fhandle(encode => $opts{todir}, version => $ver);
    my $toname = xdr_opaque(encode => $opts{toname}, %tlen);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_RENAME : NFSPROC3_RENAME;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $fromdir . $fromname . $todir . $toname,
                           out  => sub { xdr_rename(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'RENAME', result => $out);

    return $out;
} ## end sub rename($%)

#---------------------------------------------------------------------

=head2 link

  $out = $nfs_obj->link(todir => $t_fhandle,
                        name => $name [, namelen => $namelen ]
                        [, fhandle => $fhandle ]);

Calls the LINK procedure on the server.
Creates a hard link 'name' in the directory 'todir' for object
specified by 'fhandle'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the existing file system object
(default is the current filehandle).

=item todir

(Required) : Filehandle of the directory in which the link
is to be created.

=item name

(Required) : The name that is to be associated with the created link.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub link($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                todir   => {type => SCALAR},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of the original object,
    # the filehandle of the to directory and the
    # name of hard link into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $todir   = xdr_fhandle(encode => $opts{todir},   version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_LINK : NFSPROC3_LINK;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $fhandle . $todir . $name,
                           out  => sub { xdr_link(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'LINK', result => $out);

    return $out;
} ## end sub link($%)

#---------------------------------------------------------------------

=head2 readdir

  $out = $nfs_obj->readdir([ dhandle => $fhandle ]
                           [, dircount => $dcount ]
                           [, maxcount => $mcount ]
                           [, cookie => $cookie ]
                           [, cookieverf => $cookieverf ]
                           [, attr => \@attr ]);

Calls the READDIR procedure on the server.
Retrieves a variable number of entries, in sequence, from a
directory and returns the name and file identifier for each,
with information to allow the client to request additional
directory entries in a subsequent READDIR request.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to be read
(default is the current filehandle).

=item maxcount

(Optional) : Maximum size of the reply in bytes (default is 4096).

=item cookie

(Optional) : This should be set to 0 in the first request to read the
directory. On subsequent requests, it should be a cookie as returned
by the server (default is 0).

=item cookieverf

(Optional) : This should be set to a null string in the first request
to read the directory. On subsequent requests, it should be a
cookieverf as returned by the server (default is the null string).

=item attr

(Optional) : A reference to an array of attributes to be retrieved,
only used by NFSv4 (default is not to retrieve any attributes).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub readdir($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle  => {type => SCALAR, default => $self->{curr_fh}},
                maxcount => {type    => SCALAR,
                             default => 4096,
                             regex   => qr/^\d+$/},
                cookie => {type    => SCALAR,
                           default => 0,
                           regex   => qr/^\d+$/},
                cookieverf => {type => SCALAR,   default => ''},
                attr       => {type => ARRAYREF, default => []},});
    my ($args, $out, $cookieverf);
    my $ver = $self->{version};

    # Convert filehandle of directory to be read,
    # the cookie and the count into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $cookie = xdr_cookie(encode => $opts{cookie}, version => $ver);
    my $count = xdr_uint(encode => $opts{maxcount});

    if ($ver != 2) {

        # NFSv3 and v4 require a cookieverf structure
        $cookieverf = xdr_cookieverf(encode => $opts{cookieverf});
    }

    # Arrange arguments for NFSv2 and v3
    if ($ver == 2) {
        $args = $dhandle . $cookie . $count;
    } elsif ($ver == 3) {
        $args = $dhandle . $cookie . $cookieverf . $count;
    }

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC_READDIR,
                           arg  => $args,
                           out  => sub { xdr_readdir(version => $ver, @_) }
        );

    # Check for errors
    $self->error_check(name => 'READDIR', result => $out);

    return $out;
} ## end sub readdir($%)

#---------------------------------------------------------------------

=head2 readdirplus

  $out = $nfs_obj->readdirplus([ dhandle => $fhandle ]
                               [, dircount => $dcount ]
                               [, maxcount => $mcount ]
                               [, cookie => $cookie ]
                               [, cookieverf => $cookieverf ]);

Calls the READDIRPLUS procedure on the server.
Retrieves a variable number of entries, in sequence, from a
directory and returns complete information about each entry along
with information to allow the client to request additional
directory entries in a subsequent READDIRPLUS request.
This function is only supported on NFSv3. 

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to be read
(default is the current filehandle).

=item dircount

(Optional) : Maximum number of bytes of directory information returned
(default is 4096).

=item maxcount

(Optional) : Maximum size of the reply in bytes (default is 4096).

=item cookie

(Optional) : This should be set to 0 in the first request to read the
directory. On subsequent requests, it should be a cookie as returned
by the server (default is 0).

=item cookieverf

(Optional) : This should be set to a null string in the first request
to read the directory. On subsequent requests, it should be a
cookieverf as returned by the server (default is the null string).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub readdirplus($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle  => {type => SCALAR, default => $self->{curr_fh}},
                dircount => {type    => SCALAR,
                             default => 4096,
                             regex   => qr/^\d+$/},
                maxcount => {type    => SCALAR,
                             default => 4096,
                             regex   => qr/^\d+$/},
                cookie => {type    => SCALAR,
                           default => 0,
                           regex   => qr/^\d+$/},
                cookieverf => {type => SCALAR, default => 0},});
    my $out;
    my $ver = $self->{version};

    if ($ver != 3) {

        # READDIRPLUS is only supported on NFSv3
        # Use READDIR on NFSv4 to get similar results
        croak "READDIRPLUS is not supported on NFSv$ver";
    }

    # Convert filehandle of directory to be read, the cookie, the cookieverf,
    # the max dircount and the max count into XDR format
    my $args =
        xdr_fhandle(encode => $opts{dhandle}, version => $ver)
      . xdr_cookie(encode => $opts{cookie}, version => $ver)
      . xdr_cookieverf(encode => $opts{cookieverf})
      . xdr_uint(encode => $opts{dircount})
      . xdr_uint(encode => $opts{maxcount});

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC3_READDIRPLUS,
                           arg  => $args,
                           out  => \&xdr_readdirplus);

    # Check for errors
    $self->error_check(name => 'READDIRPLUS', result => $out);

    return $out;
} ## end sub readdirplus($%)

#---------------------------------------------------------------------

=head2 fsstat

  $out = $nfs_obj->fsstat([ fhandle => $fhandle ]);

Calls the FSSTAT(v3)/STATFS(v2) procedure on the server.
Retrieves volatile file system state information.
This function is only supported on NFSv2 and v3.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle identifying an object in the file system
(default is the current filehandle).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub fsstat($%) {
    my $self = shift;
    my (%opts) =
      validate(@_, {fhandle => {type => SCALAR, default => $self->{curr_fh}},});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_STATFS : NFSPROC3_FSSTAT;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $fhandle,
                           out  => sub { xdr_fsstat(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'FSSTAT', result => $out);

    return $out;
} ## end sub fsstat($%)

#---------------------------------------------------------------------

=head2 read

  $out = $nfs_obj->read(offset => $offset,
                        count => $count
                        [, fhandle => $fhandle ]);

Calls the READ procedure on the server.
Reads the data from a given offset in a file returning the data.
For NFSv2 and v3, the file attributes are returned as well.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be read
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the read is to begin.

=item count

(Required) : The number of bytes of data that are to be read.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub read($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                count   => {type => SCALAR, regex   => qr/^\d+$/},
                stateid => {type => HASHREF | SCALAR, optional => 1},});
    my ($out, $offset, $args);
    my $ver = $self->{version};

    # Convert filehandle of file
    # and the count into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $count = xdr_uint(encode => $opts{count});

    # Convert offset into XDR format
    if ($ver == 2) {
        $offset = xdr_uint(encode => $opts{offset});
    } else {
        $offset = xdr_uint64(encode => $opts{offset});
    }
    # Arguments common to both NFSv2 and v3
    $args = $fhandle . $offset . $count;
    
    if ($ver == 2) {
        
            # Add unused argument 'totalcount'
        $args .= xdr_uint(encode => 0);
    }
    
    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC_READ,
                           arg  => $args,
                           out  => sub { xdr_read(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'READ', result => $out);

    return $out;
} ## end sub read($%)

#---------------------------------------------------------------------

=head2 write

  $out = $nfs_obj->write(offset => $offset,
                         data => $data [, datalen => $datalen ]
                         [, fhandle => $fhandle ]
                         [, stable => $stable_how ]);

Calls the WRITE procedure on the server.
Writes the data to a given offset in a file.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be written
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the write is to begin.

=item data

(Required) : The data to be written.

=item stable

(Optional) : How the write is to be performed: UNSTABLE, DATA_SYNC
or FILE_SYNC (default). Only used in NFSv3 and v4.

=item datalen

(Optional) : Force the length of data to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub write($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                data    => {type => SCALAR},
                stable  => {
                           type    => SCALAR,
                           default => FILE_SYNC,
                           regex   => qr/^\d+$/},
                stateid => {type => HASHREF | SCALAR, optional => 1},
                datalen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $offset, $stable, $args);
    my $ver = $self->{version};
    my %len = (len => $opts{datalen}) if defined($opts{datalen});

    # Convert filehandle of file into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Convert offset into XDR format
    if ($ver == 2) {
        $offset = xdr_uint(encode => $opts{offset});
    } else {
        $offset = xdr_uint64(encode => $opts{offset});
        $stable = xdr_uint(encode => $opts{stable});
    }

    # Concatenate all arguments for both NFSv2 and v3
    if ($ver == 2) {
        
        # There are two unused fields in the arguments
        # so just encode it once
        my $unused = xdr_uint(encode => 0);
        $args =
            $fhandle 
            . $unused 
            . $offset 
            . $unused
            . xdr_opaque(encode => $opts{data}, %len);
    } else {

        # On NFSv3, the count for the data is included
        # even though the data is encoded as a variable length
        # opaque which is encoded as the length followed by
            # the data
        $args =
            $fhandle 
              . $offset
              . xdr_uint(encode => length($opts{data}))
              . $stable
              . xdr_opaque(encode => $opts{data}, %len);
    } ## end else [ if ($ver == 2)
    
    # Call correct procedure on server
    my $proc = ($ver == 2) ? NFSPROC_WRITE : NFSPROC3_WRITE;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $args,
                           out  => sub { xdr_write(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'WRITE', result => $out);

    return $out;
} ## end sub write($%)

#---------------------------------------------------------------------

=head2 commit

  $out = $nfs_obj->commit([ fhandle => $fhandle ]
                          [, offset => $offset ]
                          [, count => $count ]);

Calls the COMMIT procedure on the server.
Forces or flushes data to stable storage that was previously written
with a WRITE procedure call with the stable field set to UNSTABLE.
COMMIT is not supported on NFSv2 since all writes are committed.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be committed
(default is the current filehandle).

=item offset

(Optional) : The position within the file at which the commit is to begin
(default is 0).

=item count

(Optional) : The number of bytes of data to flush
(default is 0, whole file).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub commit($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, default => 0, regex => qr/^\d+$/},
                count   => {type => SCALAR, default => 0, regex => qr/^\d+$/},}
               );
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    if ($ver == 2) {

        # COMMIT is not supported on NFSv2
        croak "COMMIT is not supported on NFSv$ver";
    }

    # Convert filehandle of file, offset
    # and count into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $offset = xdr_uint64(encode => $opts{offset});
    my $count = xdr_uint(encode => $opts{count});

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC3_COMMIT,
                           arg  => $fhandle . $offset . $count,
                           out  => sub { xdr_commit(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'COMMIT', result => $out);

    return $out;
} ## end sub commit($%)

#---------------------------------------------------------------------

=head2 lock

  $out = $nfs_obj->lock(offset => $offset,
                        count => $count
                        [, owner => $owner [, ownrlen => $ownrlen ] ] ]
                        [, fhandle => $fhandle ]
                        [, type => $lock_type ]
                        [, reclaim => 0|1 ]
                        [, caller_name => $caller_name ]
                        [, sm_state => $sm_state ]
                        [, cookie => $cookie ]
                        [, svid => $svid ]);

Calls the NLM_LOCK (NFSv2,NFSv3) procedure on the server.
This function requests a record lock for the byte range specified
by the offset and count parameters.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be locked
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the lock is to begin.

=item count

(Required) : The number of bytes of data that are to be locked.

=item type

(Optional) : The type of lock to get: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item reclaim

(Optional) : True for a reclaim request (default is 0).

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item ownrlen

(Optional) : Force the length of owner to this number.

=item caller_name

(Optional) : Identifies the client making the request
(default is the client's local address or host name).
It is only supported on NFSv2 and NFSv3.

=item sm_state

(Optional) : The current value of the client's status monitor state number
(default is the status monitor state number of the object).
It is only supported on NFSv2 and NFSv3.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=item svid

(Optional) : Identifier of the process making the request
(default is the current process id).
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure
as well as the lock_stateid and the lockid to be used in locku()
and lock_cancel().

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lock($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle     => {type => SCALAR, default => $self->{curr_fh}},
                caller_name => {type => SCALAR, default => $self->{mon_host}},
                sm_state    => {type => SCALAR, default => $self->{sm_state}},
                cookie      => {type => SCALAR, default => ''},
                offset      => {type => SCALAR, regex   => qr/^\d+$/},
                count       => {type => SCALAR, regex   => qr/^\d+$/},
                type        => {
                         type    => SCALAR,
                         default => READ_LT,
                         regex   => qr/^\d+$/},
                reclaim => {type    => SCALAR,
                            default => 0,
                            regex   => qr/^[01]$/},
                owner        => {type => SCALAR,  default  => $$},
                ownrlen      => {
                            type     => SCALAR,
                            optional => 1,
                            regex    => qr/^\d+$/},
                svid => {type    => SCALAR,
                         default => ($self->{nlm_id} ? $self->{nlm_id} : $$),
                         regex   => qr/^\d+$/},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $retry = $self->{nfs_retry};
    my %len   = (len => $opts{ownrlen}) if defined($opts{ownrlen});
    my $owner = $opts{owner};

        # Use NLM version
        my $nver = $self->{nlm_ver};

        # Connect to NLM server
        $rpc = $self->nlm_rpc_get();

        # Block and exclusive flags:
        my $block =
          ($opts{type} == READW_LT or $opts{type} == WRITEW_LT) ? 1 : 0;
        my $excl =
          ($opts{type} == WRITE_LT or $opts{type} == WRITEW_LT) ? 1 : 0;

        # NLM lock structure
        my $nlm_lock_h = {caller_name => $opts{caller_name},
                          fh          => $opts{fhandle},
                          l_offset    => $opts{offset},
                          l_len       => $opts{count},
                          oh          => $opts{owner},
                          svid        => $opts{svid}};

        # Encode the cookie,
        my $args = xdr_opaque(encode => $opts{cookie}) .

          # the block flag,
          xdr_uint(encode => $block) .

          # the exclusive flag,
          xdr_uint(encode => $excl) .

          # the nlm_lock structure,
          xdr_nlm_lock(encode => $nlm_lock_h, version => $nver) .

          # the reclaim flag,
          xdr_uint(encode => $opts{reclaim}) .

          # the state flag,
          xdr_uint(encode => $opts{sm_state});

        # Call correct procedure on server
        $out = $rpc->clnt_call(proc => NLM_LOCK,
                               arg  => $args,
                               out  => \&xdr_nlm_stat);
        $self->{rpc}{xid} = $rpc->{xid};

        # Check for errors
        $self->error_check(name => 'NLM_LOCK', result => $out, type => 'nlm');

        # Save lockid to be used on locku() and lock_cancel()
        $out->{lockid} = {nlm_lock  => $nlm_lock_h,
                          block     => $block,
                          exclusive => $excl};

        return $out;
} ## end sub lock($%)

#---------------------------------------------------------------------

=head2 locku

  $out = $nfs_obj->locku([ lockid => \%lockid ]
                         [, offset => $offset ]
                         [, count => $count ]
                         [, fhandle => $fhandle ]
                         [, type => $lock_type ]
                         [, caller_name => $caller_name ]
                         [, cookie => $cookie ]
                         [, owner => $owner ]
                         [, svid => $svid ]);

Calls the NLM_UNLOCK (NFSv2,NFSv3) procedure on the server.
This function unlocks the record lock specified by the parameters.

The following named arguments are accepted:

=over

=item Arguments

=over

=item lockid

(Optional) : Reference to a hash of the lock to be unlocked as returned by
the lock() function. All other options can be used to overwrite options
in lockid.
If this argument is not specified the following options are required:
offset, count and stateid(NFSv4).

=item fhandle

(Optional) : Filehandle of the file to be unlocked
(default is the current filehandle).

=item offset

(Optional) : The position within the file at which the unlock is to begin.

=item count

(Optional) : The number of bytes of data that are to be unlocked.

=item type

(Optional) : The type of lock to release: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item caller_name

(Optional) : Identifies the client making the request
(default is the client's local address or host name).
It is only supported on NFSv2 and NFSv3.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=item owner

(Optional) : Owner of file locking state (default is the current process id).
It is only supported on NFSv2 and NFSv3.

=item svid

(Optional) : Identifier of the process making the request
(default is the current process id).
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub locku($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle     => {type => SCALAR,  optional => 1},
                caller_name => {type => SCALAR,  optional => 1},
                cookie      => {type => SCALAR,  default  => ''},
                lockid      => {type => HASHREF, optional => 1},
                owner       => {type => SCALAR,  optional => 1},
                offset      => {
                           type     => SCALAR,
                           optional => 1,
                           regex    => qr/^\d+$/},
                count => {type     => SCALAR,
                          optional => 1,
                          regex    => qr/^\d+$/},
                type => {type     => SCALAR,
                         optional => 1,
                         regex    => qr/^\d+$/},
                svid => {type     => SCALAR,
                         optional => 1,
                         regex    => qr/^\d+$/},});
    my ($out, $args, $fhandle, @ops, $nver, $lockid, $nlm_lock);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

        # Use NLM version
        $nver = $self->{nlm_ver};

        # Use the nlm_lock structure if defined
        $nlm_lock = defined($opts{lockid}) ? $opts{lockid}{nlm_lock} : {};

        # Explicit options have precedence
        $nlm_lock->{caller_name} = $opts{caller_name}
          if defined($opts{caller_name});
        $nlm_lock->{fh}       = $opts{fhandle} if defined($opts{fhandle});
        $nlm_lock->{l_offset} = $opts{offset}  if defined($opts{offset});
        $nlm_lock->{l_len}    = $opts{count}   if defined($opts{count});
        $nlm_lock->{oh}       = $opts{owner}   if defined($opts{owner});
        $nlm_lock->{svid}     = $opts{svid}    if defined($opts{svid});

        # Default values for unspecified options
        $nlm_lock->{caller_name} = $self->{mon_host}
          if not exists($nlm_lock->{caller_name});
        $nlm_lock->{fh}   = $self->{curr_fh} if not exists($nlm_lock->{fh});
        $nlm_lock->{oh}   = $$               if not exists($nlm_lock->{oh});
        $nlm_lock->{svid} = $$               if not exists($nlm_lock->{svid});

        # Connect to NLM server
        $rpc = $self->nlm_rpc_get();

        # Encode the cookie,
        $args = xdr_opaque(encode => $opts{cookie}) .

          # the nlm_lock structure,
          xdr_nlm_lock(encode => $nlm_lock, version => $nver);

        # Call correct procedure on server
        $out = $rpc->clnt_call(proc => NLM_UNLOCK,
                               arg  => $args,
                               out  => \&xdr_nlm_stat);
        $self->{rpc}{xid} = $rpc->{xid};

        # Check for errors
        $self->error_check(name => 'NLM_UNLOCK', result => $out, type => 'nlm');

        return $out;
} ## end sub locku($%)

#---------------------------------------------------------------------

=head2 lockt

  $out = $nfs_obj->lockt(offset => $offset,
                         count => $count
                         [, owner => $owner [, ownrlen => $ownrlen ] ] ]
                         [, fhandle => $fhandle ]
                         [, type => $lock_type ]
                         [, cookie => $cookie ]);

Calls the NLM_TEST (NFSv2,NFSv3) procedure on the server.
This function checks whether a lock is available to the client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be tested for lock
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the lock is to begin.

=item count

(Required) : The number of bytes of data that are to be locked.

=item type

(Optional) : The type of lock to test: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item ownrlen

(Optional) : Force the length of owner to this number.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lockt($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                cookie  => {type => SCALAR, default => ''},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                count   => {type => SCALAR, regex   => qr/^\d+$/},
                type    => {
                         type    => SCALAR,
                         default => READ_LT,
                         regex   => qr/^\d+$/},
                owner   => {type => SCALAR, default => $$},
                ownrlen => {type     => SCALAR,
                            optional => 1,
                            regex    => qr/^\d+$/},});
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{ownrlen}) if defined($opts{ownrlen});

        my $excl =
          ($opts{type} == WRITE_LT || $opts{type} == WRITEW_LT) ? 1 : 0;

        return
          $self->lock_test(fhandle   => $opts{fhandle},
                           offset    => $opts{offset},
                           count     => $opts{count},
                           exclusive => $excl,
                           cookie    => $opts{cookie},);
} ## end sub lockt($%)

#---------------------------------------------------------------------

=head2 lock_cancel

  $out = $nfs_obj->lock_cancel(lockid => \%lockid);

Cancels a blocked lock request.

The following named arguments are accepted:

=over

=item Arguments

=over

=item lockid

(Required) : Reference to a hash of the lock to be canceled as returned by
the lock() function.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').

=back

=item Returns

It returns a reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lock_cancel($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {lockid => {type => HASHREF},
                cookie => {type => SCALAR, default => ''},});
    my $out;
    my $ver    = $self->{nlm_ver};
    my $rpc    = $self->nlm_rpc_get();
    my $lockid = $opts{lockid};

    # Encode the cookie,
    my $args = xdr_opaque(encode => $opts{cookie}) .

      # the block flag,
      xdr_uint(encode => $lockid->{block}) .

      # the exclusive flag,
      xdr_uint(encode => $lockid->{exclusive}) .


      # the nlm_lock structure,
      xdr_nlm_lock(encode => $lockid->{nlm_lock}, version => $ver);

    # Call correct procedure on server
    $out = $rpc->clnt_call(proc => NLM_CANCEL,
                           arg  => $args,
                           out  => \&xdr_nlm_stat);
    $self->{rpc}{xid} = $rpc->{xid};

    # Check for errors
    $self->error_check(name => 'NLM_CANCEL', result => $out, type => 'nlm');

    return $out;
} ## end sub lock_cancel($%)

#---------------------------------------------------------------------

=head2 lock_test

  $out = $nfs_obj->lock_test(offset => $offset,
                             count => $count,
                             exclusive => 0|1
                             [, cookie => $cookie ]
                             [, fhandle => $fhandle ]);

Checks whether a lock is available to the client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be unlocked
(default is the current filehandle).

=item offset

(Required) : Offset of locked region to test.

=item count

(Required) : The number of bytes of data in locked region to test.

=item exclusive

(Required) : Set to 1 if testing whether the client could get access
to the lock.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').

=back

=item Returns

It returns a reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lock_test($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle   => {type => SCALAR, default => $self->{curr_fh}},
                cookie    => {type => SCALAR, default => ''},
                offset    => {type => SCALAR, regex   => qr/^\d+$/},
                count     => {type => SCALAR, regex   => qr/^\d+$/},
                exclusive => {type => SCALAR, regex   => qr/^[01]$/},});
    my $out;
    my $ver = $self->{nlm_ver};
    my $rpc = $self->nlm_rpc_get();

    # NLM lock structure
    my $nlm_lock_h = {caller_name => $self->{mon_host},
                      fh          => $opts{fhandle},
                      l_offset    => $opts{offset},
                      l_len       => $opts{count},
                      oh          => $rpc->{sock}->sockhost(),
                      svid        => $$,};

    # Encode the cookie (0 synchronous),
    my $args = xdr_opaque(encode => $opts{cookie}) .

      # the exclusive flag,
      xdr_uint(encode => $opts{exclusive}) .

      # the nlm_lock structure,
      xdr_nlm_lock(encode => $nlm_lock_h, version => $ver);

    # Call correct procedure on server
    $out = $rpc->clnt_call(proc => NLM_TEST,
                           arg  => $args,
                           out  => sub { xdr_nlm_test(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'NLM_TEST', result => $out, type => 'nlm');

    return $out;
} ## end sub lock_test($%)

#---------------------------------------------------------------------

=head2 create

  $out = $nfs_obj->create(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]
                          [, type => $type ]
                          [, attr => \%sattr ]
                          [, cr_mode => $mode ]
                          [, major => $major, minor => $minor ]);

Calls the CREATE procedure on the server.
This function creates different types of file system objects.
For NFSv2, it creates regular files, special device files,
named sockets or named pipes.
For NFSv3, it only creates regular files and accepts a creation
mode flag.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the new file
is to be created (default is the current filehandle).

=item name

(Required) : The name that is to be associated with the created file.

=item type

(Optional) : The type of file to be created: NFREG (default), NFDIR,
NFBLK, NFCHR, NFLNK, NFSOCK or NFFIFO.

=item attr

(Optional) : The initial attributes for the new file
(default is no initial attributes).

=item cr_mode

(Optional) : Create mode: UNCHECKED (default), GUARDED, EXCLUSIVE.
Used only in NFSv3.

=item major

(Optional) : The major device number, only for type NFBLK and NFCHR.
Used in NFSv2 and v4.

=item minor

(Optional) : The minor device number, only for type NFBLK and NFCHR.
Used in NFSv2 and v4.

=item verf

(Optional) : Create verifier. Only used if cr_mode is EXCLUSIVE.
Default is the current time.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                type    => {
                         type    => SCALAR,
                         default => NFREG,
                         regex   => qr/^\d+$/},
                attr    => {type => HASHREF, default => {}},
                cr_mode => {type    => SCALAR,
                            default => UNCHECKED,
                            regex   => qr/^\d+$/},
                major   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                minor   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                verf    => {type => SCALAR, default  => time},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $args, $spec, $attr, $fhandle);
    my $ver  = $self->{version};
    my %nlen = (len => $opts{namelen}) if defined($opts{namelen});

    if ($ver == 2) {

        # Create list of file types
        my %ftype = (NFDIR()  => 0040000,
                     NFCHR()  => 0020000,
                     NFBLK()  => 0060000,
                     NFREG()  => 0100000,
                     NFLNK()  => 0120000,
                     NFSOCK() => 0140000,
                     NFFIFO() => 0020000,);

        # CREATE in NFSv2 is overloaded so the file type
        # is encoded in the mode of the initial sattr structure
        $opts{attr}{mode} |=
          exists($ftype{$opts{type}}) ? $ftype{$opts{type}} : $opts{type} << 12;
    } ## end if ($ver == 2)

    # Convert the major and minor numbers into XDR format
    # if this is a special device file
    if ($opts{type} == NFBLK or $opts{type} == NFCHR) {
        if ($ver == 2) {

            # CREATE in NFSv2 is overloaded so the major
            # and minor numbers go into the size of the
            # initial sattr structure
            $opts{attr}{size} = ($opts{major} << 8) | $opts{minor};
        }
    } ## end if ($opts{type} == NFBLK...

    # Convert filehandle of directory and the name
    # of the new file into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %nlen);

    # Encode initial attributes of file
    $attr = xdr_sattr(encode => $opts{attr}, version => $ver);

    if ($ver == 2) {

        # Complete arguments for NFSv2
        $args = $dhandle . $name . $attr;
    } elsif ($ver == 3) {

        # For NFSv3, need to add the creation mode flag and its
        # parameters
        $args = $dhandle . $name . xdr_uint(encode => $opts{cr_mode});
        if ($opts{cr_mode} == EXCLUSIVE) {

            # Add createverf3 if exclusive create is requested
            # No initial attributes are set with option
            $args .= xdr_opaque(encode => xdr_uint(encode => $opts{verf}),
                                fixed  => 1,
                                maxlen => 8);
        } else {

            # UNCHECKED and GUARDED require initial attributes
            $args .= $attr;
        }
    } ## end elsif ($ver == 3)

        # Call correct procedure on server
        my $proc = $ver == 2 ? NFSPROC_CREATE : NFSPROC3_CREATE;
        $out = $self->nfs_call(proc => $proc,
                               arg  => $args,
                               out  => sub { xdr_create(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'CREATE', result => $out);

    # Set the current filehandle
    if ($opts{type} == NFDIR or $opts{type} == NFREG) {
        $self->current_fh_set(fhandle => $out->{fhandle});
    }

    return $out;
} ## end sub create($%)

#---------------------------------------------------------------------

=head2 rmdir

  $out = $nfs_obj->rmdir(name => $name [, namelen => $namelen ]
                         [, dhandle => $fhandle ]);

Calls the RMDIR procedure on the server.
Removes a subdirectory from a directory.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory from which the subdirectory
is to be removed (default is the current filehandle).

=item name

(Required) : The name of the subdirectory to be removed.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub rmdir($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory and name of
    # directory to be removed into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

    # Call correct procedure on server
    my $proc = $ver == 2 ? NFSPROC_RMDIR : NFSPROC3_RMDIR;
    $out = $self->nfs_call(proc => $proc,
                           arg  => $dhandle . $name,
                           out  => sub { xdr_remove(version => $ver, @_) });

    # Check for errors
    $self->error_check(name => 'RMDIR', result => $out);

    return $out;
} ## end sub rmdir($%)

#---------------------------------------------------------------------

=head2 fsinfo

  $out = $nfs_obj->fsinfo([ fhandle => $fhandle ]);

Calls the FSINFO procedure on the server.
Retrieves nonvolatile file system state information and general
information about the NFSv3 protocol server implementation.
This function is only supported on NFSv3.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle identifying the file object
(default is the current filehandle).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub fsinfo($%) {
    my $self = shift;
    my (%opts) =
      validate(@_, {fhandle => {type => SCALAR, default => $self->{curr_fh}},});
    my $out;
    my $ver = $self->{version};

    if ($ver != 3) {

        # FSINFO is not supported on NFSv2 and v4,
        croak "FSINFO is not supported on NFSv$ver";
    }

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Call correct procedure on server
    $out = $self->nfs_call(proc => NFSPROC3_FSINFO,
                           arg  => $fhandle,
                           out  => \&xdr_fsinfo);

    # Check for errors
    $self->error_check(name => 'FSINFO', result => $out);

    return $out;
} ## end sub fsinfo($%)

#---------------------------------------------------------------------

=head2 pathconf

  $out = $nfs_obj->pathconf([ fhandle => $fhandle |
                              path => $path [, pathlen => $pathlen ] ]);

Calls the PATHCONF procedure on the server.
Retrieves the pathconf information for a file or directory.
This function is not supported on NFSv4, to get similar
information use the getattr function instead.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle identifying the file object
(default is the current filehandle).
Use only for NFSv3.

=item path

(Optional) : Path of object
(default is the mounted path).
Use only for NFSv2.

=item pathlen

(Optional) : Force the length of path to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub pathconf($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default  => $self->{curr_fh}},
                path    => {type => SCALAR, default  => $self->{mnt_path}},
                pathlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($rpc, $args, $out);
    my $ver = $self->{version};
    my %len = (len => $opts{pathlen}) if defined($opts{pathlen});

    if ($ver == 2) {

        # PATHCONF on NFSv2 is implemented on the MOUNT protocol
        # so create a connection to it
        $rpc = $self->mount_rpc_get();

        # Convert path of object into XDR format
        $args = xdr_opaque(encode => $opts{path}, %len);
    } else {

        # PATHCONF on NFSv3 is implemented on the NFS protocol
        # so use the current Rpc object
        $rpc = $self->{rpc};

        # Convert filehandle of object into XDR format
        $args = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    }

    # Call correct procedure on server
    my $proc = $ver == 2 ? MOUNTPROC_PATHCONF : NFSPROC3_PATHCONF;
    $out = $rpc->clnt_call(proc => $proc,
                           arg  => $args,
                           out  => sub { xdr_pathconf(version => $ver, @_) });
    $self->{rpc}{xid} = $rpc->{xid} if ($ver == 2);

    # Check for errors
    $self->error_check(name => 'PATHCONF', result => $out);

    if ($ver == 2) {

        # Convert pc_mask into similar fields as in NFSv3
        $out->{chown_restricted} = ($out->{pc_mask1} >> 7) & 0x1;
        $out->{no_trunc}         = ($out->{pc_mask1} >> 8) & 0x1;
    }

    return $out;
} ## end sub pathconf($%)

#---------------------------------------------------------------------

#---------------------------------------------------------------------
# Internal function.
# Creates and returns an Rpc object for the MOUNT protocol used
# for NFS version 2 and 3 which requires a different protocol
# for the mount and umount operations as well as the PATHCONF
# procedure on NFSv2.
#---------------------------------------------------------------------
sub mount_rpc_get($;) {
    my $self = shift;
    my %rpc_opts_h;

    # Pass local address if it is given
    $rpc_opts_h{localaddr} = $self->{localaddr} if defined($self->{localaddr});

    # Create an rpc object for mountd (100005)
    my $rpc = nfs::Rpc->clnt_create(hostname => $self->{hostname},
                                    version  => $self->{mnt_ver},
                                    proto    => $self->{mnt_proto},
                                    port     => $self->{mnt_port},
                                    xid      => $self->{rpc}{xid},
                                    retry    => $self->{retry},
                                    resvport => $self->{resvport},
                                    timeout  => $self->{timeout},
                                    program  => 100005,
                                    addr_type=> $self->{addr_type},
                                    %rpc_opts_h);

    # Save the mount port number so next time
    # the portmapper is not called to get it
    $self->{mnt_port} = $rpc->{port} if !$self->{mnt_port};

    return $rpc;
} ## end sub mount_rpc_get($;)

#---------------------------------------------------------------------
# Internal function.
# Creates and returns an Rpc object for the NLM protocol used
# for NFS version 2 and 3 which requires a different protocol
# for lock operations.
#---------------------------------------------------------------------
sub nlm_rpc_get($;) {
    my $self = shift;
    my %rpc_opts_h;

    # Pass local address if it is given
    $rpc_opts_h{localaddr} = $self->{localaddr} if defined($self->{localaddr});

	# Work around for 527689 and 531296.
	$self->{nlm_port} = 4045;

    # Create an rpc object for NLM (100021)
    my $rpc = nfs::Rpc->clnt_create(hostname => $self->{hostname},
                                    version  => $self->{nlm_ver},
                                    proto    => $self->{nlm_proto},
                                    port     => $self->{nlm_port},
                                    xid      => $self->{rpc}{xid},
                                    retry    => $self->{retry},
                                    resvport => $self->{resvport},
                                    timeout  => $self->{timeout},
                                    program  => $NLM_PROGRAM,
                                    addr_type=> $self->{addr_type},
                                    %rpc_opts_h);

    # Save the mount port number so next time
    # the portmapper is not called to get it
    $self->{nlm_port} = $rpc->{port} if !$self->{nlm_port};

    return $rpc;
} ## end sub nlm_rpc_get($;)

#---------------------------------------------------------------------

=head2 mount

  $out = $nfs_obj->mount(path => $path [, pathlen => $pathlen ]);

Returns the root filehandle of the path to be mounted.
For NFSv2 and v3, creates an Rpc object to the
MOUNTD program to get the initial filehandle of path using the
MOUNTPROC_MNT procedure on the server.

It saves the mountpoint filehandle on object's blessed hash
under the 'mnt_fh' key, as well as setting the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : The pathname to be mounted.

=item pathlen

(Optional) : Force the length of path to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mount($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {path    => {type => SCALAR},
                pathlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($rpc, $out);
    my $ver = $self->{version};
    my %len = (len => $opts{pathlen}) if defined($opts{pathlen});

        # Mount on NFSv2 and v3 is implemented on the MOUNT protocol
        # so create a connection to it
        $rpc = $self->mount_rpc_get();

        # Use the mount version on RPC call
        my $mver = $self->{mnt_ver};

        # Call correct procedure on server
        $out = $rpc->clnt_call(proc => MOUNTPROC_MNT,
                               arg  => xdr_opaque(encode => $opts{path}, %len),
                               out  => sub { xdr_mount(version => $mver, @_) });
        $self->{rpc}{xid} = $rpc->{xid};

        # Check for errors
        $self->error_check(name => 'MOUNT', result => $out);

    # Set current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    # Save the root filehandle on the NfsClient object
    $self->{mnt_fh} = $out->{fhandle};

    # Save mount path
    $self->{mnt_path} = $opts{path};

    return $out;
} ## end sub mount($%)

#---------------------------------------------------------------------

=head2 umount

  $nfs_obj->umount(path => $path [, pathlen => $pathlen ]);

For NFSv2 and v3, removes a MOUNT entry from the server's MOUNT list.
This procedure is an advisory call to the server as notification
that the client has unmounted the file system and will no longer
be using it. It uses the MOUNTPROC_UMNT procedure of the MOUNTD
protocol.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : The pathname to be unmounted.

=item pathlen

(Optional) : Force the length of path to this number.

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub umount($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {path    => {type => SCALAR},
                pathlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $rpc;
    my $ver = $self->{version};
    my %len = (len => $opts{pathlen}) if defined($opts{pathlen});

        # Unmount on NFSv2 and v3 is implemented on the MOUNT protocol
        # so create a connection to it
        $rpc = $self->mount_rpc_get();

    # Call correct procedure on server
    my $ret = $rpc->clnt_call(proc => MOUNTPROC_UMNT,
                              arg  => xdr_opaque(encode => $opts{path}, %len));
    $self->{rpc}{xid} = $rpc->{xid};

    return $ret;
} ## end sub umount($%)

#---------------------------------------------------------------------

=head2 mount_dump

  $out = $nfs_obj->mount_dump();

Calls the MOUNTPROC_DUMP procedure of the MOUNTD protocol.
Returns a list of clients and the exported file systems they
have mounted. Since the number of clients and mounted
file systems can be very long, clients should use this
call via a TCP connection.

=over

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mount_dump($;) {
    my $self = shift;
    my $rpc  = $self->mount_rpc_get();

    # Call correct procedure on server
    my $ret = $rpc->clnt_call(proc => MOUNTPROC_DUMP, out => \&xdr_dump);
    $self->{rpc}{xid} = $rpc->{xid};

    return $ret;
} ## end sub mount_dump($;)

#---------------------------------------------------------------------

=head2 mount_export

  $out = $nfs_obj->mount_export();

Calls the MOUNTPROC_EXPORT procedure of the MOUNTD protocol.
Returns a list of exported file systems and their export information.
Since the export list can be large, a TCP connection should be used
to retrieve this data.

=over

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mount_export($;) {
    my $self = shift;
    my $rpc  = $self->mount_rpc_get();

    # Call correct procedure on server
    my $ret = $rpc->clnt_call(proc => MOUNTPROC_EXPORT, out => \&xdr_export);
    $self->{rpc}{xid} = $rpc->{xid};

    return $ret;
} ## end sub mount_export($;)

#---------------------------------------------------------------------

=head2 sm_notify

  $nfs_obj->sm_notify([ host => $nsm_server ]
                      [, mon_name => $mon_name ]
                      [, state => $sm_state ]
                      [, resvport => 0|1 ]
                      [, proto => 'udp'|'tcp' ]
                      [, xid => $xid ]);

  nfs::NfsClient::sm_notify(host => $nsm_server,
                            mon_name => $mon_name,
                            state => $sm_state
                            [, resvport => 0|1 ]
                            [, proto => 'udp'|'tcp' ]
                            [, xid => $xid ]);

Notify of state change.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item host

(Optional) : Hostname/IP address to connect to.
It is optional when called as an instance method with a default being
the current host of the object.
When called as a class method, this argument is required.

=item mon_name

(Optional) : The name of the host from which the request is sent.
It is optional when called as an instance method with a default being
the current local host of the object.
When called as a class method, this argument is required.

=item state

(Optional) : The new state number of the host.
It is optional when called as an instance method with a default being
the current state incremented by two.
When called as a class method, this argument is required
(must be an odd number).

=item resvport

(Optional) : Bind to a reserved port number on the local machine.
A reserved port number is less than or equal to 1024. If this
option is true an attempt to bind to a reserved port number
between 600 and 1023 is made. On most unix machines, root
is the only user able to bind to a reserved port.

=item proto

(Optional) : Protocol to connect to status monitor: 'udp', 'tcp'
(default is 'udp').

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item timeout

(Optional) : Timeout value for connection
(default is the timeout value for the object or 30).

=item localaddr

(Optional) : Hostname or ip address of local machine
(default is the localaddr for the object).

=back

=item Returns

Output of SM_NOTIFY (procedure 6).

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub sm_notify(%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_,
               {host     => {type => SCALAR, optional => 1},
                mon_name => {type => SCALAR, optional => 1},
                state    => {type => SCALAR, optional => 1},
                resvport => {type    => SCALAR,
                             regex   => qr/^[01]$/,
                             default => 0},
                proto => {type    => SCALAR,
                          regex   => qr/^(udp|tcp)$/,
                          default => 'udp'},
                xid => {type     => SCALAR,
                        regex    => qr/^\d+$/,
                        optional => 1},
                timeout => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                localaddr => {type => SCALAR, optional => 1},});
    my (%rpc_opts_h, $args);

    if (!defined($self)) {

        # Function is called as a class method: nfs::Rpc::sm_notify()
        # The host, mon_name and state arguments are required
        foreach my $arg (qw(host mon_name state)) {
            croak "sm_notify: The $arg argument is required "
              . "when called as a class method"
              if !defined($opts{$arg});
        }
    } ## end if (!defined($self))

    # Get NSM server to connect to
    my $host = defined($opts{host}) ? $opts{host} : $self->{hostname};

    # Get the correct name of the host being monitored
    my $mon_name =
      defined($opts{mon_name})
      ? $opts{mon_name}
      : $self->{mon_host};

    # Get the correct state number of client, when using the object's
    # state number increment it by two first
    my $state = defined($opts{state}) ? $opts{state} : ($self->{sm_state} += 2);

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{rpc}{xid}
      :                       int(rand(1000000000));

    # Get the correct local address
    my $localaddr =
        defined($opts{localaddr}) ? $opts{localaddr}
      : defined($self)            ? $self->{localaddr}
      :                             undef;
    $rpc_opts_h{localaddr} = $localaddr if defined($localaddr);

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;
    $rpc_opts_h{timeout} = $timeout if defined($timeout);

    # Call the status monitor on 'udp' to simulate a crash on the client
    my $rpc = nfs::Rpc->clnt_create(hostname => $host,
                                    proto    => $opts{proto},
                                    xid      => $xid,
                                    resvport => $opts{resvport},
                                    program  => 100024,
                                    version  => 1,
                                    addr_type=> $self->{addr_type},
                                    %rpc_opts_h);

    # Encode arguments
    $args = xdr_opaque(encode => $mon_name) . xdr_uint(encode => $state);

    # Call SM_NOTIFY (procedure 6)
    my $ret = $rpc->clnt_call(proc => 6, arg => $args);
    $self->{rpc}{xid} = $rpc->{xid} if defined($self);

    return $ret;
} ## end sub sm_notify(%)

#---------------------------------------------------------------------

=head2 sm_simu_crash

  $nfs_obj->sm_simu_crash([ host => $nsm_server ]
                          [, resvport => 0|1 ]
                          [, proto => 'udp'|'tcp' ]
                          [, xid => $xid ]);

  nfs::NfsClient::sm_simu_crash(host => $nsm_server
                                [, resvport => 0|1 ]
                                [, proto => 'udp'|'tcp' ]
                                [, xid => $xid ]);

Simulates a server or client crash by calling the simulate crash procedure
of the NSM (network status monitor) on the specified host.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item host

(Optional) : Hostname/IP address to connect to.
It is optional when called as an instance method with a default being
the current host of the object.
When called as a class method, this argument is required.

=item resvport

(Optional) : Bind to a reserved port number on the local machine.
A reserved port number is less than or equal to 1024. If this
option is true an attempt to bind to a reserved port number
between 600 and 1023 is made. On most unix machines, root
is the only user able to bind to a reserved port.

=item proto

(Optional) : Protocol to connect to status monitor: 'udp', 'tcp'
(default is 'udp').

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item timeout

(Optional) : Timeout value for connection
(default is the timeout value for the object or 30).

=item localaddr

(Optional) : Hostname or ip address of local machine
(default is the localaddr for the object).

=back

=item Returns

Output of SM_SIMU_CRASH (procedure 5).

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub sm_simu_crash(%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_,
               {host     => {type => SCALAR, optional => 1},
                resvport => {type    => SCALAR,
                             regex   => qr/^[01]$/,
                             default => 0},
                proto => {type    => SCALAR,
                          regex   => qr/^(udp|tcp)$/,
                          default => 'udp'},
                xid => {type     => SCALAR,
                        regex    => qr/^\d+$/,
                        optional => 1},
                timeout => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                localaddr => {type => SCALAR, optional => 1},});
    my %rpc_opts_h;

    if (!defined($self) and !defined($opts{host})) {

        # Function is called as a class method: nfs::Rpc::sm_simu_crash()
        # The host argument is required with this type of call
        croak "sm_simu_crash: The host argument is required "
          . "when called as a class method";
    }

    # Get NSM server to connect to
    my $host = defined($opts{host}) ? $opts{host} : $self->{hostname};

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{rpc}{xid}
      :                       int(rand(1000000000));

    # Get the correct local address
    my $localaddr =
        defined($opts{localaddr}) ? $opts{localaddr}
      : defined($self)            ? $self->{localaddr}
      :                             undef;
    $rpc_opts_h{localaddr} = $localaddr if defined($localaddr);

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;
    $rpc_opts_h{timeout} = $timeout if defined($timeout);

    # Call the status monitor on 'udp' to simulate a crash on the client
    my $rpc = nfs::Rpc->clnt_create(hostname => $host,
                                    proto    => $opts{proto},
                                    xid      => $xid,
                                    resvport => $opts{resvport},
                                    program  => 100024,
                                    addr_type=> ($self->{addr_type} ? $self->{addr_type} : nfs::Rpc->addr_type()),
                                    version  => 1,
                                    %rpc_opts_h);

    # Call SM_SIMU_CRASH (procedure 5)
    my $ret = $rpc->clnt_call(proc => 5);
    $self->{rpc}{xid} = $rpc->{xid} if defined($self);

    return $ret;
} ## end sub sm_simu_crash(%)

#---------------------------------------------------------------------

=head2 current_fh_set

  $nfs_obj->current_fh_set(fhandle => $fhandle);

Sets the current filehandle of the object.
All functions which require a filehandle uses this current
filehandle as the default.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Required) : Filehandle to set.

=back

=item Returns

The filehandle before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub current_fh_set($%) {
    my $self = shift;
    my (%opts) = validate(@_, {fhandle => {type => SCALAR},});

    # Save the current filehandle on prev_fh and set
    # the current filehandle to the one provided
    $self->{prev_fh} = $self->{curr_fh};
    $self->{curr_fh} = $opts{fhandle};

    return $self->{prev_fh};
} ## end sub current_fh_set($%)

#---------------------------------------------------------------------
# Internal function.
# This function checks for errors on the results hash.
# If the status is not NFS_OK, the function croaks with
# the corresponding error.
# The error message has the following format:
#   'name' NFS?ERR_error_name: error_message
# The 'name' argument is just a string to identify which function
# has failed, usually is set to the procedure name,
# e.g., LOOKUP, CREATE, etc.
#---------------------------------------------------------------------
sub error_check($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {name   => {type => SCALAR},
                result => {type => HASHREF | UNDEF},
                type   => {type => SCALAR, default => 'nfs'},});
    $Log->enter("opts: ".Dumper(\%opts));

    # Clear previous error result
    $self->{result} = undef;

    my (undef, $file, $line, undef) = caller(0);
    my (undef, undef, undef, $subr) = caller(1);
    my $callermsg = " ($subr at $file:$line) ";

    # Check if result is defined
    croak "No results received from server: $callermsg"
      if !defined $opts{result};

    if ($opts{result}{status}) {

        # Procedure returned an error therefore save the
        # result on the object so it can be retrieved by
        # the calling process if needed
        $self->{result} = $opts{result};

        if ($opts{type} eq 'nfs') {

            # Croak with the proper NFS error message
            croak $opts{name} . " "
              . nfs_error(error   => $opts{result}{status},
                          version => $self->{version}) . ": $callermsg";
        } elsif ($opts{type} eq 'nlm') {

            # Croak with the proper NLM error message
            croak $opts{name} . " "
              . nlm_error(error => $opts{result}{status})
              . ": $callermsg";
        }
    } ## end if ($opts{result}{status...

    $Log->exit();
    return;
} ## end sub error_check($%)

#---------------------------------------------------------------------
# Internal function
# Calls a procedure on the server with the given arguments, it retries
# the request if any of the errors defined in %retry_errors is returned
#---------------------------------------------------------------------
sub nfs_call($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {proc  => {type => SCALAR},
                arg   => {type => SCALAR},
                out   => {type => CODEREF},
                retry => {type => SCALAR, default => $self->{nfs_retry}},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $retry = $opts{retry};

    # Retry block
    {

        # Call the NFS procedure
        $out = $rpc->clnt_call(proc => $opts{proc},
                               arg  => $opts{arg},
                               out  => $opts{out});

        # Retry call after nfs_delay seconds if specified
        # Do not retry for NFS version 2
        if (   $ver != 2
            && $out
            && $retry-- > 0
            && $retry_errors{$out->{status}}) {
            sleep $self->{nfs_delay};
            redo;
        }
    }

    return $out;
} ## end sub nfs_call($%)

#---------------------------------------------------------------------
# Register all original port number with the portmapper
# and kill NLM and NSM daemons
#---------------------------------------------------------------------
sub _cleanup_daemons() {
    my ($current, $new, $item, $out);

    if ($nlm_pid) {

        # Stop the NLM daemon
        _stop_daemon($NLM_PROGRAM, $nlm_pid, $nlm_lock_file);
    }

    if ($nsm_pid) {

        # Stop the NSM daemon
        _stop_daemon($NSM_PROGRAM, $nsm_pid, $nsm_lock_file);
    }

    if (defined($daemon_reg_ports)) {

        # Get current list of ports registered
        $current = nfs::Rpc::pmap_dump();

        # Unregister current ports for program
        foreach $item (@$current) {
            $out = 0;

            # Only process current program numbers
            next if $item->{program} != $nfs::NfsXdr::daemon_program;
            foreach $new (@$daemon_reg_ports) {
                if ($item->{version} == 0
                    || (   $new->{program} == $item->{program}
                        && $new->{version} == $item->{version}
                        && $new->{proto} eq $item->{proto}
                        && $new->{port} != $item->{port})
                  ) {
                    $out = 1;
                    last;
                }
            } ## end foreach $new (@$daemon_reg_ports)
            if ($out) {

                # Unregister port so the original port can be restored
                nfs::Rpc::pmap_unset(program => $item->{program},
                                     version => $item->{version},
                                     proto   => $item->{proto},
                                     port    => $item->{port});
            }
        } ## end foreach $item (@$current)

        # Register saved ports for program
        foreach $new (@$daemon_reg_ports) {
            $out = 1;

            # Only process current program numbers
            next if $new->{program} != $nfs::NfsXdr::daemon_program;
            foreach $item (@$current) {
                if (   $new->{program} == $item->{program}
                    && $new->{version} == $item->{version}
                    && $new->{proto} eq $item->{proto}
                    && $new->{port} == $item->{port}) {

                    # Port is register -- do nothing
                    $out = 0;
                    last;
                }
            } ## end foreach $item (@$current)
            if ($out) {

                # Port is no longer registered -- register saved port
                eval {
                    nfs::Rpc::pmap_set(program => $new->{program},
                                       version => $new->{version},
                                       proto   => $new->{proto},
                                       port    => $new->{port});
                };
            } ## end if ($out)
        } ## end foreach $new (@$daemon_reg_ports)
    } ## end if (defined($daemon_reg_ports...
} ## end sub _cleanup_daemons()

#---------------------------------------------------------------------
# Destructor
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;
} ## end sub DESTROY

#---------------------------------------------------------------------
# END block to cleanup the daemons
#---------------------------------------------------------------------
END {
    _cleanup_daemons();
}

1;

