# this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/NfsClient/4_0.pm#13 $
#
## @summary NfsClient - NFS client protocol class
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

NfsClient - NFS client protocol class

=head1 SYNOPSIS

  use nfs::NfsClient;

  $nfs_obj = nfs::NfsClient->new(hostname => $host, version  => 4,
                                 proto => 'tcp');
  $nfs_obj->callback_register(version => 1);
  $nfs_obj->callback_process();
  $out = $nfs_obj->mount(path => '/vol/testvol');
  $nfs_obj->null();
  $out = $nfs_obj->getattr();
  $out = $nfs_obj->setattr(attr => {FATTR4_MODE() => 0777});
  $out = $nfs_obj->lookup(name => 'file1');
  $out = $nfs_obj->access(access => ACCESS_READ);
  $opn = $nfs_obj->open(name => 'file1', access => OPEN4_SHARE_ACCESS_READ,
                                         deny   => OPEN4_SHARE_DENY_NONE);
  $out = $nfs_obj->read(offset => 0, count => 1024, stateid => $opn->{stateid});
  $out = $nfs_obj->open_downgrade(stateid => $opn->{stateid},
                                  access => OPEN4_SHARE_ACCESS_WRITE,
                                  deny => OPEN4_SHARE_DENY_NONE);
  $out = $nfs_obj->open_confirm(stateid => $opn->{stateid});
  $out = $nfs_obj->openattr();
  $out = $nfs_obj->close(stateid => $opn->{stateid});
  $out = $nfs_obj->lookupp();
  $out = $nfs_obj->readlink(fhandle => $fhandle);
  $out = $nfs_obj->mkdir(name => 'test_dir', attr => {FATTR4_MODE() => 0755});
  $out = $nfs_obj->rename(fromdir => $f_fhandle, fromname => 'file3',
                          dodir   => $f_fhandle, toname   => 'file4');
  $out = $nfs_obj->remove(name => 'file4');
  $out = $nfs_obj->readdir();
  $out = $nfs_obj->write(offset => $offset, data => $data);
  $out = $nfs_obj->commit();
  $out = $nfs_obj->delegreturn(stateid => $deleg_stateid);
  $out = $nfs_obj->delegpurge();
  $out = $nfs_obj->renew();
  $out = $nfs_obj->create(name => 'fifo', type => NFFIFO);
  $out = $nfs_obj->rmdir(name => 'fifo');
  $out = $nfs_obj->fsinfo();
  $out = $nfs_obj->secinfo(name => '.');
  $out = $nfs_obj->setclientid();
  $out = $nfs_obj->setclientid_confirm();
  $nfs_obj->umount(path => '/vol/testvol');
  $nfs_obj->current_fh_set(fhandle => $fhandle);
  $nfs_obj->putpubfh();
  $nfs_obj->putrootfh();

  # To lock a file on NFSv4
  $lock = $nfs_obj->lock(offset => 32, count => 64, stateid => $opn->{stateid});

  # To unlock the file on NFSv4
  $out = $nfs_obj->locku(offset => 32, count => 64,
                         stateid => $lock->{lock_stateid});

  # or simply use
  $out = $nfs_obj->locku(lockid => $lock->{lockid});
  # To unlock only a subregion of the lock
  $out = $nfs_obj->locku(lockid => $lock->{lockid}, offset => 50, count=>10);

  # Release lock owner on NFSv4
  $out = $nfs_obj->release_lockowner();

  # Call a compound request
  @ops = ( { op => OP_PUTFH,   args => $fhandle },
           { op => OP_GETATTR, args => xdr_bitmap4(encode=>$attr) } );
  $out = $nfs_obj->compound( ops => \@ops, tag => 'getattr' );

  # Object description
  $rpc        = $nfs_obj->rpc();         # RPC object
  $mnt_fh     = $nfs_obj->mnt_fh();      # Filehandle of root of mount
  $curr_fh    = $nfs_obj->curr_fh();     # Current filehandle use as default
  $mnt_path   = $nfs_obj->mnt_path();    # Mount path
  $result     = $nfs_obj->result();      # Last failed results
  $clientid   = $nfs_obj->clientid();    # NFSv4 client id
  $seqid      = $nfs_obj->seqid();       # NFSv4 open sequence id
  $lock_seqid = $nfs_obj->lock_seqid();  # NFSv4 lock sequence id
  $nfs_retry  = $nfs_obj->nfs_retry();   # Number of times to retry a request
  $nfs_delay  = $nfs_obj->nfs_delay();   # Seconds to delay before a retry

  # Set NFSv4 open sequence id and return previous value
  $seqid_prev = $nfs_obj->seqid( seqid => $seqid_new);

  # Set NFSv4 lock sequence id and return previous value
  $lock_seqid_prev = $nfs_obj->lock_seqid($lock_seqid_new);

  # Set the number of times to retry a request if
  # NFS4ERR_DELAY is returned
  $nfs_retry_prev = $nfs_obj->nfs_retry($nfs_retry_new);

  # Set the number of seconds to delay before a retry
  $nfs_delay_prev = $nfs_obj->nfs_delay($nfs_delay_new);

=head1 DESCRIPTION

Provides an object interface for the NFS client protocol version
 4. On NFSv4, only those
operations needed to have similar functionality with NFSv2 and v3
are implemented as of now.

Most methods are called with the same parameters for all three
versions of NFS. NFSv3 and v4 may require additional input parameters.
This should be noted in the documentation of each particular method.
In addition, some methods are not supported by all versions of NFS.
If this is the case the method croaks with an unsupported error message.

The format of the 'attr' option for setting the attributes of a
file system object is as follows. For NFSv2 and v3, the valid
options are: mode, size, uid, gid, atime and mtime. For NFSv4,
the valid options are any of the FATTR4_* attribute constants which
can be set. Note that there is a difference in the time structure:
NFSv2 is { seconds=>x, useconds=>y } while NFsv3 and v4 is
{ seconds=>x, nseconds=>y }.

When the 'attr' option is used to retrieve a list of attributes
for NFSv4, the option is a reference to an array of FATTR4_*
attribute constants. The best way to test this library is by running
the tests in nfs/ops.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/main/lib/nfs/NfsClient/4_0.pm#13 $
  Last revised $Date: 2012/07/11 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsClient::4_0;

use strict;
use warnings;

use nfs::Rpc;
use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF CODEREF UNDEF);
use nfs::NfsXdr qw(xdr_bitmap4 xdr_cb_compound xdr_compound xdr_compound_res
  xdr_cookie xdr_cookieverf xdr_fattr4 xdr_fhandle xdr_opaque xdr_sattr
  xdr_specdata xdr_stateid4 xdr_uint xdr_uint64);
use nfs::NfsDef qw(nfs_error bitmap4_convert
  EXCLUSIVE FATTR4_FILEID FATTR4_FSID FATTR4_MODE FATTR4_NUMLINKS FATTR4_OWNER
  FATTR4_OWNER_GROUP FATTR4_RAWDEV FATTR4_SIZE FATTR4_TIME_ACCESS
  FATTR4_TIME_METADATA FATTR4_TIME_MODIFY FATTR4_TYPE FILE_SYNC
  NFBLK NFCHR NFDIR NFFIFO NFLNK NFREG NFS4ERR_DELAY
  NFS4ERR_DENIED NFS4ERR_GRACE NFSOCK
  OP_ACCESS OP_CB_RECALL OP_CLOSE OP_COMMIT OP_CREATE OP_DELEGPURGE
  OP_DELEGRETURN OP_GETATTR OP_GETFH OP_LINK OP_LOCK OP_LOCKT OP_LOCKU
  OP_LOOKUP OP_NVERIFY OP_LOOKUPP OP_OPEN OP_OPEN_CONFIRM OP_OPEN_DOWNGRADE OP_OPENATTR
  OP_PUTFH OP_PUTROOTFH OP_PUTPUBFH OP_READ OP_READDIR OP_READLINK OP_RELEASE_LOCKOWNER
  OP_REMOVE OP_RENAME OP_RENEW OP_SAVEFH OP_SECINFO OP_SETATTR OP_SETCLIENTID
  OP_SETCLIENTID_CONFIRM OP_VERIFY OP_WRITE OPEN4_SHARE_ACCESS_READ
  OPEN4_SHARE_DENY_NONE READ_LT READW_LT UNCHECKED WRITE_LT WRITEW_LT :NFS4DCLAIMS
  OPEN_DELEGATE_NONE OPEN_DELEGATE_READ OPEN_DELEGATE_WRITE
  OPEN4_SHARE_ACCESS_WRITE OPEN4_SHARE_ACCESS_BOTH);

use Tharn qw($Log);
use Data::Dumper qw{Dumper};

#
# To handle different addr_type family, and avoid sending this addr information across
# callbacks, define a local variable and set the value when object is initiated (new)
#
my $loopback;

BEGIN {

    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
} ## end BEGIN

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($);
sub callback_register($);
sub callback_process($;%);
sub null($;);
sub getattr($);
sub setattr($);
sub lookup($);
sub lookupp($);
sub access($);
sub mkdir($);
sub symlink($);
sub remove($);
sub rename($);
sub readdir($);
sub read($);
sub write($);
sub commit($);
sub open($);
sub open_downgrade($);
sub open_confirm($);
sub openattr($);
sub close($);
sub lock($);
sub locku($);
sub lockt($);
sub release_lockowner($);
sub delegreturn($);
sub delegpurge($);
sub renew($;);
sub create($);
sub secinfo($);
sub setclientid($;);
sub setclientid_confirm($;);
sub mount($);
sub umount($);
sub current_fh_set($);
sub error_check($);
sub nfs_call($);
sub deleg_remove($$;);
sub compound($);
sub rpc;
sub mnt_fh;
sub curr_fh;
sub mnt_path;
sub result;
sub clientid;
sub seqid($);
sub lock_seqid($$;);
sub nfs_retry($$;);
sub nfs_delay($$;);
sub DESTROY;
sub savefh($);
sub secinfo_no_name($);
sub nverify($);
sub verify($);


#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
# Map the callback procedures to its
# corresponding callback function
my %call_back = (0 => sub { return },
                 1 => \&xdr_cb_compound,);

# Retry the NFS request for the following errors
# The NFS4ERR_DELAY -  a retry of the NFS request
my %retry_errors = (NFS4ERR_DELAY()   => 1,);

#---------------------------------------------------------------------

=head2 new

  $nfs_obj = nfs::NfsClient->new(hostname => $host,
                                 version => 4
                                 [, proto => 'tcp' ]
                                 [, addr_type => 'inet4'|'inet6' ]
                                 [, port => $port ]
                                 [, resvport => 0|1 ]
                                 [, xid => $xid ]
                                 [, retry => $retry ]
                                 [, nfs_retry => $nfs_retry ]
                                 [, nfs_delay => $nfs_delay ]
                                 [, reconnect => 0|1 ]
                                 [, timeout => $timeout ]
                                 [, qsize => $qsize ]
                                 [, localaddr => $localaddr ]);

NFS object constructor. For NFSv4, the sequence
id's for both open and lock are initialized to zero, the callback
queue and the list of open delegations are defined and the client
identifier and verifier are set to a random number so every instance
of this object is a completely different client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : Hostname or ip address of server to connect.

=item version

(Required) : NFS version: 4.

=item proto

(Optional) : Protocol of connection: 'tcp' (default is 'tcp').

=item port

(Optional) : Port number of server to connect
(default is 0 which attempts to connect to the portmapper to
get the port number).

=item resvport

(Optional) : Bind to a reserved port number on the local machine.
A reserved port number is less than or equal to 1024. If this
option is true an attempt to bind to a reserved port number
between 600 and 1023 is made. On most unix machines, root
is the only user able to bind to a reserved port.

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item retry

(Optional) : Number of times to retry each RPC call if response is not
received within the timeout period (default is 0).

=item nfs_retry

(Optional) : Number of times to retry the request if
NFS4ERR_DELAY is returned (default is 5). It will wait nfs_delay
seconds before each retry.

=item nfs_delay

(Optional) : Number of seconds to wait before each retry (default is 1).

=item reconnect

(Optional) : Reconnect if TCP connection is dropped. This only works for
TCP connections and when the retry parameter is non-zero (default is 0)

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item qsize

(Optional) : Size of RPC call queue.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=item addr_type

(Optional) : The type of address, inet4/inet6. If not specified, "ADDR_TYPE" will be used.
This is used only when "hostname" is given but "dstaddr" is not given, hence uses NATE hostobject
to choose the IP address. See nfs::Rpc->name2ip().

=item sec

(Optional) : The type of security to be used. If not specified, it uses "sys" authentication by default.
The valid values are none, sys, krb5, krb5i, krb5p. Please note as of writing this code, krb5p and krb5i
are not yet supported in c-Mode.

=back

=item Returns

NfsClient object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($) {
    my $class = shift;
    my (%opts) =
      validate(@_,
               {hostname => {type => SCALAR},
                version  => {type => SCALAR, regex => qr/^[4]$/},
                resvport => {type    => SCALAR,
                             regex   => qr/^\d+$/,
                             default => 0},
                proto => {type    => SCALAR,
                          regex   => qr/^tcp$/,
                          default => 'tcp'},
                port => {type    => SCALAR,
                         regex   => qr/^\d+$/,
                         default => 0},
                xid => {type    => SCALAR,
                        regex   => qr/^\d+$/,
                        default => int(rand(1000000000))},
                retry => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                nfs_retry => {type    => SCALAR,
                              regex   => qr/^\d+$/,
                              default => 5},
                nfs_delay => {type    => SCALAR,
                              regex   => qr/^\d+$/,
                              default => 1},
                reconnect => {type    => SCALAR,
                              regex   => qr/^[01]$/,
                              default => 0},
                timeout => {type    => SCALAR,
                            regex   => qr/^\d+$/,
                            default => 30},
                qsize => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                localaddr => {type => SCALAR, optional => 1},
                addr_type => {type => SCALAR, default => nfs::Rpc->addr_type()},
                minor_version   => {type => SCALAR,  optional => 1, default => 0 },
				sec      => { type => SCALAR,
					default => nfs::Rpc->auth_type(),
					regex   => qr/^(sys|krb5|krb5i|krb5p|none)$/,},	

               });
    my ($self, %rpc_opts_h);

    # Pass local address if it is given
    $rpc_opts_h{localaddr} = $opts{localaddr} if defined($opts{localaddr});

	$opts{port} = $opts{port} ? $opts{port} : 2049;

    # Create an rpc object for NFS (100003)
    my $rpc = nfs::Rpc->clnt_create(hostname  => $opts{hostname},
                                    version   => $opts{version},
                                    resvport  => $opts{resvport},
                                    proto     => $opts{proto},
                                    port      => $opts{port},
                                    xid       => $opts{xid},
                                    retry     => $opts{retry},
                                    timeout   => $opts{timeout},
                                    qsize     => $opts{qsize},
									sec		  => $opts{sec},
                                    reconnect => $opts{reconnect},
                                    addr_type => $opts{addr_type},
                                    program   => 100003,
                                    %rpc_opts_h);
    $self->{loopback} = $rpc->loopback();
    $loopback = $self->{loopback};

    # Make sure connection is established
    # Burt 178530 null procedure does not return void
    # it returns a null string of 4 bytes
    eval { $rpc->clnt_call(proc => 0, auth_type=>0) };
    croak $@ if ($@ and $@ !~ /4 extra bytes/);

    my %v4options = (lock_seqid  => 0,
                      cb_queue    => \@nfs::NfsXdr::cb_queue,
                      open_delegs => \@nfs::NfsXdr::open_delegs,
                      cl_verifier => time + int(rand(1000000000)),
                      cl_id       => time + int(rand(1000000000)),);

    # Get local host name
    $opts{localname} = (POSIX::uname())[1];

    $self = bless {%opts, %v4options,
                   rpc      => $rpc,}, $class;

	$Log->exit("new client: " . Dumper($self));
    # Return NFS object
    return $self;
} ## end sub new($)

#---------------------------------------------------------------------

=head2 callback_register

  $nfs_obj->callback_register([ version => $version ]
                              [, timeout => $timeout ]);

Registers the callback service for NFSv4 using the given
version, the same protocol transport as the object
and using a program number from the transient range.
This function is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Optional) : Version number of callback service to register
(default is version 1).

=item timeout

(Optional) : Timeout value for connection (default is 30).

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub callback_register($) {
    my $self = shift;
    $Log->enter(qq{self dump} . Dumper($self));

    my (%opts) =
		validate(@_,
		{
			version => { type => SCALAR, regex => qr/^\d+$/, default => 1 },
			addr_type => { type => SCALAR, regex => qr/inet|inet4|inet6/, default => $self->{addr_type}},
			timeout => { type => SCALAR, regex => qr/^\d+$/, default => 30 },
		}
	);

    # Create the callback service using the given version,
    # the same protocol transport as the NfsClient object,
    # and use a program number from the transient range
    $self->{callback} =
		nfs::Rpc->svc_create(version  => $opts{version},
							proto    => $self->{proto},
							addr_type=> $opts{addr_type},
							timeout  => $opts{timeout},
							dispatch => \%call_back);

    $Log->exit(qq{callback is:} . Dumper($self->{callback}));

    return;
} ## end sub callback_register($)

#---------------------------------------------------------------------

=head2 callback_process

  $nfs_obj->callback_process([ timeout => $time_in_mseconds ]);

Process a pending callback. It waits until a callback is
received. It only supports the CB_RECALL callback.
This function is only supported on NFSv4.

=over

=item Arguments

=over

=item timeout

(Optional) : Timeout value in milliseconds to check if any
callback is pending. A value of 0 returns immediately if no
callback is currently pending. If argument is not specified
it waits until a callback is ready with no time limit.

=back

=item Returns

True if callback is received and processed correctly.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub callback_process($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            timeout => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
        }
    );
    my $ver = $self->{version};
    my ($item, $out, %options);

    if ( $ver != 4 ) {
        croak "This function is not supported on NFSv$ver";
    }

    if ( ! defined($self->{callback}) ) {
        croak "Must use callback_register before calling this function";
    }

   	$Log->debug("callback_process: begin" . Dumper($self->{callback}));
    # Add the timeout argument if it is defined
    $options{timeout} = $opts{timeout} if defined($opts{timeout});

    # Wait until a callback is received
    $out = $self->{callback}->svc_getreqset(%options);

    # Process callback queue
    while ( @{$self->{cb_queue}} ) {
    	$Log->debug("We are in the callback queue \n");
        $item = shift @{$self->{cb_queue}};
        if ( $item->{op} == OP_CB_RECALL ) {
            # Recall open delegation
            $self->delegreturn(fhandle => $item->{args}{fhandle},
                               stateid => $item->{args}{stateid});
        }
    }

    return $out;
} ## end sub callback_process($;%)

#---------------------------------------------------------------------

=head2 null

  $nfs_obj->null();

Calls the NULL procedure on the server.  This procedure takes
no arguments and returns nothing.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub null($;) {
    my $self = shift;
    my $rpc  = $self->{rpc};

    # Call null procedure on server
    return $rpc->clnt_call(proc => 0);
}

#---------------------------------------------------------------------

=head2 getattr

  $out = $nfs_obj->getattr([ fhandle => $fhandle ]
                           [, attr => \@fattr ]);

Calls the GETATTR procedure on the server.
Retrieves the attributes of the specified file system object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the object whose attributes
are to be retrieved (default is the current filehandle).

=item attr

(Optional) : A reference to an array of attributes to be retrieved.
Only used by NFSv4 (default is to get the same attributes as NFSv3).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub getattr($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                attr    => {
                         type    => ARRAYREF,
                         default => [FATTR4_TYPE,          FATTR4_SIZE,
                                     FATTR4_FSID,          FATTR4_FILEID,
                                     FATTR4_MODE,          FATTR4_NUMLINKS,
                                     FATTR4_OWNER,         FATTR4_OWNER_GROUP,
                                     FATTR4_RAWDEV,        FATTR4_TIME_ACCESS,
                                     FATTR4_TIME_METADATA, FATTR4_TIME_MODIFY,]
                  },});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

        # Convert the list of attributes
        my $attr = bitmap4_convert(attrlist => $opts{attr});

        # Create compound request
        my @ops = ({op => OP_PUTFH,   args => $fhandle},
                   {op => OP_GETATTR, args => xdr_bitmap4(encode => $attr)});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'getattr',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'GETATTR', result => $out);

    # Create a copy of the fattr4 reference for ease of use
    $out->{fattr} = $out->{results}[1]{result}{fattr}{attr_vals};

    return $out;
} ## end sub getattr($)

#---------------------------------------------------------------------

=head2 setattr

  $out = $nfs_obj->setattr(attr => \%sattr
                           [, fhandle => $fhandle ]
                           [, ctime => \%nfstime3 ]);

Calls the SETATTR procedure on the server.
Changes one or more attributes of a file system object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the object whose attributes
are to be changed (default is the current filehandle).

=item attr

(Required) : A reference to a hash of attributes to be changed.

=item ctime

(Optional) : A reference to an nfstime3 hash. If provided, the
server is to verify that this matches the ctime for the object.
If there is no match, the server preserves the object attributes
and returns an error. Not used in NFSv2.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setattr($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR,  default  => $self->{curr_fh}},
                retry   => {type => SCALAR, default  => $self->{nfs_retry}, regex => qr/^\d+$/},
                delay   => {type => SCALAR, default  => $self->{nfs_delay}, regex => qr/^\d+$/},
                attr    => {type => HASHREF},
                ctime   => {type => HASHREF, optional => 1},});
    my $out;
    my $ver = $self->{version};
    my $retry = $opts{retry};
    my $retry_delay = $opts{delay};

    # Convert filehandle of object and
    # the sattr structure into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $sattr = xdr_sattr(encode => $opts{attr}, version => $ver);

        # Create compound request
        my @ops = ({op => OP_PUTFH, args => $fhandle});
        if (defined($opts{ctime})) {

            # Verify ctime
            my $ctime = {FATTR4_TIME_METADATA() => $opts{ctime}};
            push @ops, {op => OP_VERIFY, args => xdr_fattr4(encode => $ctime)};
        }
        push @ops, {op => OP_SETATTR, args => $sattr};

        #retry block(Ref :Burt 503767)
        {
                $out = $self->nfs_call(proc => 1,
                                arg =>
                                  xdr_compound(tag => 'setattr',
                                               ops => \@ops),
                                out => \&xdr_compound_res);

            # Retry call
            if ($retry-- and $out and $out->{status} == NFS4ERR_DELAY) {
               sleep $retry_delay;
               redo;
            }
        }
    # Check for errors
    $self->error_check(name => 'SETATTR', result => $out);

    # Create a copy of the attrsset reference for ease of use
    $out->{attrsset} = $out->{results}[1]{result}{attrsset};

    return $out;
} ## end sub setattr($)

#---------------------------------------------------------------------

=head2 lookup

  $out = $nfs_obj->lookup(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]);

Calls the LOOKUP procedure on the server.
Searches a directory for a specific name and returns the
filehandle for the corresponding file system object.
It sets the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to search
(default is the current filehandle).

=item name

(Required) : The name of the object to find.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lookup($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $fhandle);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory and
    # the name of object into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

        # Create compound request
        my @ops = ({op => OP_PUTFH,  args => $dhandle},
                   {op => OP_LOOKUP, args => $name},
                   {op => OP_GETFH});

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'lookup',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'LOOKUP', result => $out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[2]{result}{fhandle};

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    return $out;
} ## end sub lookup($)

#---------------------------------------------------------------------

=head2 lookupp

  $out = $nfs_obj->lookupp([ dhandle => $fhandle ]);

Calls the LOOKUPP procedure on the server.
Gets the filehandle for the parent directory of the directory
entry specified by the given filehandle.
It is not supported on NFSv2 and v3; use lookup with '..' instead.
It sets the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to search
(default is the current filehandle).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lookupp($) {
    my $self = shift;
    my (%opts) =
      validate(@_, {dhandle => {type => SCALAR, default => $self->{curr_fh}},});
    my ($out, $fhandle);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Convert filehandle of directory into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);

    # Create compound request
    my @ops =
      ({op => OP_PUTFH, args => $dhandle}, {op => OP_LOOKUPP},
        {op => OP_GETFH});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg  => xdr_compound(tag => 'lookupp', ops => \@ops),
                           out  => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'LOOKUPP', result => $out);

    # Get the filehandle of object to save it
    $out->{fhandle} = $out->{results}[2]{result}{fhandle};

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    return $out;
} ## end sub lookupp($)

#---------------------------------------------------------------------

=head2 access

  $out = $nfs_obj->access(access => $access
                          [, fhandle => $fhandle ]);

Calls the ACCESS procedure on the server.
Determines the access rights that a user, as identified by the
credentials in the request, has with respect to a file system object.
This function is not supported on NFSv2.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file system object to which access
is to be checked (default is the current filehandle).

=item access

(Required) : Bit mask of access permissions to check.
Valid access mode bits: ACCESS_READ, ACCESS_LOOKUP, ACCESS_MODIFY,
ACCESS_EXTEND, ACCESS_DELETE and ACCESS_EXECUTE.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub access($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                access  => {type => SCALAR, regex   => qr/^\d+$/},});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object and
    # the requested access bit mask into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $access = xdr_uint(encode => $opts{access});

        # Create compound request
        my @ops = ({op => OP_PUTFH,  args => $fhandle},
                   {op => OP_ACCESS, args => $access});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'access',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'ACCESS', result => $out);

    return $out;
} ## end sub access($)

#---------------------------------------------------------------------

=head2 readlink

  $out = $nfs_obj->readlink([ fhandle => $fhandle ]);

Calls the READLINK procedure on the server.
Reads the pathname stored in a symbolic link.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of symbolic link
(default is the current filehandle).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub readlink($) {
    my $self = shift;
    my (%opts) =
      validate(@_, {fhandle => {type => SCALAR, default => $self->{curr_fh}},});
    my $out;
    my $ver = $self->{version};

    # Convert filehandle of object into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

        # Create compound request
        my @ops = ({op => OP_PUTFH, args => $fhandle}, {op => OP_READLINK});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'readlink',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'READLINK', result => $out);

    # Create a copy of the link data reference for ease of use
    $out->{data} = $out->{results}[1]{result}{data};

    return $out;
} ## end sub readlink($)

#---------------------------------------------------------------------

=head2 remove

  $out = $nfs_obj->remove(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]);

Calls the REMOVE procedure on the server.
Removes an entry from a directory. If the entry in the directory
was the last reference to the corresponding file system object,
the object may be destroyed.
Removing a directory object is only supported on NFSv4.
For NFSv2 and v3 use rmdir instead.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the entry
is to be removed (default is the current filehandle).

=item name

(Required) : The name of the entry to be removed.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub remove($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                retry   => {type => SCALAR, default  => $self->{nfs_retry}, regex => qr/^\d+$/},
                delay   => {type => SCALAR, default  => $self->{nfs_delay}, regex => qr/^\d+$/},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});
    my $retry = $opts{retry};
    my $retry_delay = $opts{delay};


    # Convert filehandle of directory and
    # name of object into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

        # Create compound request
        my @ops =
          ({op => OP_PUTFH, args => $dhandle},
            {op => OP_REMOVE, args => $name});

     #retry block
    {
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'remove',
                                              ops => \@ops),
                               out => \&xdr_compound_res);
         # Retry call
         if ($retry-- and $out and $out->{status} == NFS4ERR_DELAY) {
                       sleep $retry_delay;
                       redo;
                 }
    }
    # Check for errors
    $self->error_check(name => 'REMOVE', result => $out);

    return $out;
} ## end sub remove($)

#---------------------------------------------------------------------

=head2 rename

  $out = $nfs_obj->rename(fromdir => $f_fhandle,
                          fromname => $o_name [, fromlen => $fromlen ],
                          todir => $t_handle,
                          toname => $n_name [, tolen => $tolen ]);

Calls the RENAME procedure on the server.
Renames the directory entry identified by 'fromname' in the
directory 'fromdir' to 'toname' in the directory 'todir'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fromdir

(Required) : Filehandle of the directory from which the entry
is to be renamed.

=item fromname

(Required) : The name of the entry to be renamed.

=item todir

(Required) : Filehandle of the directory to which the entry
is to be renamed.

=item toname

(Required) : The new name for the object.

=item fromlen

(Optional) : Force the length of fromname to this number.

=item tolen

(Optional) : Force the length of toname to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub rename($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fromdir  => {type => SCALAR},
                fromname => {type => SCALAR},
                todir    => {type => SCALAR},
                toname   => {type => SCALAR},
                fromlen  => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                retry   => {type => SCALAR, default  => $self->{nfs_retry}, regex => qr/^\d+$/},
                delay   => {type => SCALAR, default  => $self->{nfs_delay}, regex => qr/^\d+$/},
                tolen    => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
               });
    my $out;
    my $ver  = $self->{version};
    my %flen = (len => $opts{fromlen}) if defined($opts{fromlen});
    my %tlen = (len => $opts{tolen}) if defined($opts{tolen});
    my $retry = $opts{retry};
    my $retry_delay = $opts{delay};


    # Convert filehandle of the from directory, the original object name,
    # the filehandle of the to directory and the new name of object
    # into XDR format
    my $fromdir = xdr_fhandle(encode => $opts{fromdir}, version => $ver);
    my $fromname = xdr_opaque(encode => $opts{fromname}, %flen);
    my $todir = xdr_fhandle(encode => $opts{todir}, version => $ver);
    my $toname = xdr_opaque(encode => $opts{toname}, %tlen);

        # Create compound request
        # The rename operation must use the saved
        # filehandle as the from directory and the
        # current filehandle as the to directory
        my @ops = ({op => OP_PUTFH,  args => $fromdir},
                   {op => OP_SAVEFH},
                   {op => OP_PUTFH,  args => $todir},
                   {op => OP_RENAME, args => $fromname . $toname});

    #retry block
    {
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'rename',
                                              ops => \@ops),
                               out => \&xdr_compound_res);
        # Retry call
        if ($retry-- and $out and $out->{status} == NFS4ERR_DELAY) {
                       sleep $retry_delay;
                       redo;
        }
    }

    # Check for errors
    $self->error_check(name => 'RENAME', result => $out);

    return $out;
} ## end sub rename($)

#---------------------------------------------------------------------

=head2 link

  $out = $nfs_obj->link(todir => $t_fhandle,
                        name => $name [, namelen => $namelen ]
                        [, fhandle => $fhandle ]);

Calls the LINK procedure on the server.
Creates a hard link 'name' in the directory 'todir' for object
specified by 'fhandle'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the existing file system object
(default is the current filehandle).

=item todir

(Required) : Filehandle of the directory in which the link
is to be created.

=item name

(Required) : The name that is to be associated with the created link.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub link($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                todir   => {type => SCALAR},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of the original object,
    # the filehandle of the to directory and the
    # name of hard link into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $todir   = xdr_fhandle(encode => $opts{todir},   version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

        # Create compound request
        # The link operation must use the saved filehandle
        # as the existing file system object and the
        # current filehandle as the to directory
        # where the link is to be created
        my @ops = ({op => OP_PUTFH, args => $fhandle},
                   {op => OP_SAVEFH},
                   {op => OP_PUTFH, args => $todir},
                   {op => OP_LINK,  args => $name});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'link',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'LINK', result => $out);

    return $out;
} ## end sub link($)

#---------------------------------------------------------------------

=head2 readdir

  $out = $nfs_obj->readdir([ dhandle => $fhandle ]
                           [, dircount => $dcount ]
                           [, maxcount => $mcount ]
                           [, cookie => $cookie ]
                           [, cookieverf => $cookieverf ]
                           [, attr => \@attr ]);

Calls the READDIR procedure on the server.
Retrieves a variable number of entries, in sequence, from a
directory and returns the name and file identifier for each,
with information to allow the client to request additional
directory entries in a subsequent READDIR request.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to be read
(default is the current filehandle).

=item dircount

(Optional) : Maximum number of bytes of directory information returned
(default is 4096). Only used in NFSv4.

=item maxcount

(Optional) : Maximum size of the reply in bytes (default is 4096).

=item cookie

(Optional) : This should be set to 0 in the first request to read the
directory. On subsequent requests, it should be a cookie as returned
by the server (default is 0).

=item cookieverf

(Optional) : This should be set to a null string in the first request
to read the directory. On subsequent requests, it should be a
cookieverf as returned by the server (default is the null string).

=item attr

(Optional) : A reference to an array of attributes to be retrieved,
only used by NFSv4 (default is not to retrieve any attributes).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub readdir($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle  => {type => SCALAR, default => $self->{curr_fh}},
                maxcount => {type    => SCALAR,
                             default => 4096,
                             regex   => qr/^\d+$/},
                dircount => {type    => SCALAR,
                             default => 4096,
                             regex   => qr/^\d+$/},
                cookie => {type    => SCALAR,
                           default => 0,
                           regex   => qr/^\d+$/},
                cookieverf => {type => SCALAR,   default => ''},
                attr       => {type => ARRAYREF, default => []},});
    my ($args, $out, $cookieverf);
    my $ver = $self->{version};

    # Convert filehandle of directory to be read,
    # the cookie and the count into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $cookie = xdr_cookie(encode => $opts{cookie}, version => $ver);
    my $count = xdr_uint(encode => $opts{maxcount});

    # NFSv3 and v4 require a cookieverf structure
    $cookieverf = xdr_cookieverf(encode => $opts{cookieverf});
 
        # Convert the list of attributes
        my $attr = bitmap4_convert(attrlist => $opts{attr});

        # Create arguments for compound request
        $args =
            $cookie
          . $cookieverf
          . xdr_uint(encode => $opts{dircount})
          . $count
          . xdr_bitmap4(encode => $attr);

        # Create compound request
        my @ops = ({op => OP_PUTFH,   args => $dhandle},
                   {op => OP_READDIR, args => $args});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'readdir',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'READDIR', result => $out);

    # Create a copy of the directory list reference for ease of use
        $out->{eof}        = $out->{results}[1]{result}{eof};
        $out->{cookieverf} = $out->{results}[1]{result}{cookieverf};
        $out->{list}       = $out->{results}[1]{result}{list};

    return $out;
} ## end sub readdir($)

#---------------------------------------------------------------------

=head2 read

  $out = $nfs_obj->read(offset => $offset,
                        count => $count
                        [, fhandle => $fhandle ]
                        [, stateid => \%stateid|STATEID0|STATEID1 ]);

Calls the READ procedure on the server.
Reads the data from a given offset in a file returning the data.
For NFSv2 and v3, the file attributes are returned as well.
Must use open before calling this function for NFSv4,
unless the stateid is either STATEID0 or STATEID1.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be read
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the read is to begin.

=item count

(Required) : The number of bytes of data that are to be read.

=item stateid

(Optional) : Reference to a hash containing the stateid of the opened file.
The constants STATEID0 or STATEID1 can be used instead of the hash
for the special stateid of all 0's or all 1's respectively.
Only required for NFSv4.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=item retry 

The number of times to reattempt the READ op on a retriable error. Errors
such as NFS4ERR_DELAY are considered retriable.

=back

=item delay 

The number of seconds to wait between retry operations.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub read($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                count   => {type => SCALAR, regex   => qr/^\d+$/},
                retry   => {type => SCALAR, default  => $self->{nfs_retry}, regex => qr/^\d+$/},
                delay   => {type => SCALAR, default  => $self->{nfs_delay}, regex => qr/^\d+$/},
                verf  => {type => SCALAR, default => time},
                stateid => {type => HASHREF | SCALAR, optional => 1},});
    my ($out, $offset, $args);
    my $ver = $self->{version};
    my $retry = $opts{retry};
    my $retry_delay = $opts{delay};

    # Convert filehandle of file
    # and the count into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $count = xdr_uint(encode => $opts{count});

    # Convert offset into XDR format
    $offset = xdr_uint64(encode => $opts{offset});

    # Convert the stateid into XDR format and
    # append all other arguments to operation
    $args = xdr_stateid4(encode => $opts{stateid}) . $offset . $count;

    # Create compound request
    my @ops = ({op => OP_PUTFH, args => $fhandle},
               {op => OP_READ, args => $args});

    # retry block
    {
        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'read',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

        # Retry call
        if ($retry-- and $out and $out->{status} == NFS4ERR_DELAY) {
            sleep $retry_delay;
            redo;
        }
    }

    # Check for errors
    $self->error_check(name => 'READ', result => $out);

    # Create a copy of the read data reference for ease of use
        $out->{eof}                      = $out->{results}[1]{result}{eof};
        $out->{data}                     = $out->{results}[1]{result}{data};
        $out->{results}[1]{result}{data} = \$out->{data};

    return $out;
} ## end sub read($)

#---------------------------------------------------------------------

=head2 write

  $out = $nfs_obj->write(offset => $offset,
                         data => $data [, datalen => $datalen ]
                         [, fhandle => $fhandle ]
                         [, stable => $stable_how ]
                         [, stateid => \%stateid|STATEID0|STATEID1 ]);

Calls the WRITE procedure on the server.
Writes the data to a given offset in a file.
Must use open before calling this function for NFSv4,
unless the stateid is either STATEID0 or STATEID1.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be written
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the write is to begin.

=item data

(Required) : The data to be written.

=item stable

(Optional) : How the write is to be performed: UNSTABLE, DATA_SYNC
or FILE_SYNC (default). Only used in NFSv3 and v4.

=item stateid

(Optional) : Reference to a hash containing the stateid of the opened file.
The constants STATEID0 or STATEID1 can be used instead of the hash
for the special stateid of all 0's or all 1's respectively.
Only required for NFSv4.

=item datalen

(Optional) : Force the length of data to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub write($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                data    => {type => SCALAR},
                stable  => {
                           type    => SCALAR,
                           default => FILE_SYNC,
                           regex   => qr/^\d+$/},
                stateid => {type => HASHREF | SCALAR, optional => 1},
                datalen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $offset, $stable, $args);
    my $ver = $self->{version};
    my %len = (len => $opts{datalen}) if defined($opts{datalen});

		# use stateid(0) if not given.
		$opts{stateid} = $opts{stateid} ? $opts{stateid} : 0x0 ;
    # Convert filehandle of file into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Convert offset into XDR format
        $offset = xdr_uint64(encode => $opts{offset});
        $stable = xdr_uint(encode => $opts{stable});

        # Convert the stateid into XDR format and
        # append all other arguments to operation
        $args =
            xdr_stateid4(encode => $opts{stateid}) 
          . $offset 
          . $stable
          . xdr_opaque(encode => $opts{data}, %len);

        # Create compound request
        my @ops =
          ({op => OP_PUTFH, args => $fhandle}, {op => OP_WRITE, args => $args});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'write',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'WRITE', result => $out);

    return $out;
} ## end sub write($)

#---------------------------------------------------------------------

=head2 commit

  $out = $nfs_obj->commit([ fhandle => $fhandle ]
                          [, offset => $offset ]
                          [, count => $count ]);

Calls the COMMIT procedure on the server.
Forces or flushes data to stable storage that was previously written
with a WRITE procedure call with the stable field set to UNSTABLE.
COMMIT is not supported on NFSv2 since all writes are committed.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be committed
(default is the current filehandle).

=item offset

(Optional) : The position within the file at which the commit is to begin
(default is 0).

=item count

(Optional) : The number of bytes of data to flush
(default is 0, whole file).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub commit($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                offset  => {type => SCALAR, default => 0, regex => qr/^\d+$/},
                count   => {type => SCALAR, default => 0, regex => qr/^\d+$/},}
               );
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Convert filehandle of file, offset
    # and count into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);
    my $offset = xdr_uint64(encode => $opts{offset});
    my $count = xdr_uint(encode => $opts{count});

        # Create compound request
        my @ops = ({op => OP_PUTFH,  args => $fhandle},
                   {op => OP_COMMIT, args => $offset . $count});

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'commit',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'COMMIT', result => $out);

    return $out;
} ## end sub commit($)

#---------------------------------------------------------------------

=head2 open

  $out = $nfs_obj->open(name => $name [, namelen => $namelen ]
                        [, dhandle => $fhandle ]
                        [, access => $access ]
                        [, deny => $deny ]
                        [, cr_mode => $cr_mode
                           [, attr => \%sattr ] [, verf => $verf ] ]
                        [, owner => $owner [, ownrlen => $ownrlen ] ]
                        [, retry => $retry ]
                        [, deleg => $deleg ]);

Calls the OPEN procedure on the server.
This function opens and/or creates regular files.
OPEN resembles LOOKUP in that it generates a filehandle for the
client to use.  Unlike LOOKUP though, OPEN creates server state on
the filehandle.  In normal circumstances, the client can only release
this state with a CLOSE operation.
It is not supported on NFSv2 and v3; use create instead.
It sets the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the file is located
or the new file is to be created (default is the current filehandle).

=item name

(Required) : The name of the file to be opened or created.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item cr_mode

(Optional) : Create mode: UNCHECKED, GUARDED or EXCLUSIVE
(default is to just open the file without creating it).

=item attr

(Optional) : The initial attributes for the new file.
It is only used when create mode is either UNCHECKED or GUARDED
(default is no initial attributes).

=item access

(Optional) : Share access mode of open: OPEN4_SHARE_ACCESS_READ (default),
OPEN4_SHARE_ACCESS_WRITE or OPEN4_SHARE_ACCESS_BOTH.

=item deny

(Optional) : Share deny mode of open: OPEN4_SHARE_DENY_NONE (default),
OPEN4_SHARE_DENY_READ, OPEN4_SHARE_DENY_WRITE or OPEN4_SHARE_DENY_BOTH.

=item retry

(Optional) : Number of times to retry the request if NFS4ERR_GRACE
is returned (default is 0). Each retry takes 10 seconds.

=item deleg

(Optional) : Number of times to retry the request if NFS4ERR_DELAY
is returned (default is the nfs_retry of the object).
Each retry takes nfs_delay seconds. A callback is processed returning
the delegation before retrying for the first time.

=item verf

(Optional) : Create verifier. Only used if cr_mode is EXCLUSIVE.
Default is the current time.

=item namelen

(Optional) : Force the length of name to this number.

=item ownrlen

(Optional) : Force the length of owner to this number.

=item deleg_type

(Optional) : CLAIM_NULL (No special rights to file. Ordinary OPEN of the specified file.),
             CLAIM_PREVIOUS (Right to the file established by an open previous to serve reboot.  File identified by filehandle obtained at that time rather than by name.),
             CLAIM_DELEGATE_CUR (Right to file based on a delegation granted by the server. File is specified by name.) or
             CLAIM_DELEGATE_PREV (Right to file based on a delegation granted to a previous boot instance of the client.  File is specified by name.)

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub open($) {
    my $self = shift;
    $Log->enter();
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR,  default  => $self->{curr_fh}},
                name    => {type => SCALAR},
                owner   => {type => SCALAR,  default  => $$},
                attr    => {type => HASHREF, optional => 1},
                access  => {
                           type    => SCALAR,
                           default => OPEN4_SHARE_ACCESS_READ,
                           regex   => qr/^\d+$/},
                deny => {type    => SCALAR,
                         default => OPEN4_SHARE_DENY_NONE,
                         regex   => qr/^\d+$/},
                cr_mode => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                retry   => {type => SCALAR, default  => 0, regex => qr/^\d+$/},
                verf  => {type => SCALAR, default => time},
                deleg => {type    => SCALAR,
                          default => $self->{nfs_retry},
                          regex   => qr/^\d+$/},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                ownrlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                deleg_type =>
                  {type => SCALAR, default => CLAIM_NULL, regex => qr/^\d+$/},}
               );
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $retry = $opts{retry};
    my $deleg = $opts{deleg};
    my $owner = $opts{owner};

    # return_delegation is expected only if callback is registered,
    # otherwise this doesn't make sense.
    my $return_deleg = (defined($self->{callback})) ? 1 : 0;
    my %nlen = (len => $opts{namelen}) if defined($opts{namelen});
    my %olen = (len => $opts{ownrlen}) if defined($opts{ownrlen});

    # If the owner passed is new, then initialize the seqid for the
    # new owner
    if (!defined($self->{seq_id}{$owner})) {
        $self->{seq_id}{$owner} = 0;
    }

    # Convert filehandle of directory and the name
    # of the new file into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);

    # For CLAIM_PREVIOUS, the given filename is assumed to be the filehandle itself.
    # This will become the "current filehandle" during the open operation.
    my $name = "";
    if ($opts{deleg_type} == CLAIM_PREVIOUS) {
        $name = xdr_fhandle(encode => $opts{name}, version => $ver);
    } else {
        $name = xdr_opaque(encode => $opts{name}, %nlen);
    }

    # Encode the sequence id,
    my $args = xdr_uint(encode => $self->{seq_id}{$owner}++) .

      # Share access,
      xdr_uint(encode => $opts{access}) .

      # Share deny,
      xdr_uint(encode => $opts{deny}) .

      # Owner: client id
      $self->{clientid} .

      #        owner
      xdr_opaque(encode => $opts{owner}, %olen);

    # Open type
    if (!defined($opts{cr_mode})) {

        # OPEN4_NOCREATE
        $args .= xdr_uint(encode => 0);
    } else {

        # OPEN4_CREATE includes the create mode
        # and initial attributes if not EXCLUSIVE
        $args .= xdr_uint(encode => 1) . xdr_uint(encode => $opts{cr_mode});
        if ($opts{cr_mode} == EXCLUSIVE) {

            # Add verifier4 if exclusive create is requested
            # No initial attributes are set with option
            $args .= xdr_opaque(encode => xdr_uint(encode => $opts{verf}),
                                fixed  => 1,
                                maxlen => 8);
        } else {

            # UNCHECKED and GUARDED require initial attributes
            $args .= xdr_fattr4(encode => $opts{attr});
        }
    } ## end else [ if (!defined($opts{cr_mode...

    # only deleg_types implemented and tested are CLAIM_NULL and CLAIM_DELEGATE_CUR
    # if you need support for others, you have to implement and/or test the same

    # set CLAIM type
    $args .= xdr_uint(encode => $opts{deleg_type});

    if (   ($opts{deleg_type} == CLAIM_DELEGATE_CUR)
        || ($opts{deleg_type} == CLAIM_PREVIOUS)) {

        if ($opts{deleg_type} == CLAIM_DELEGATE_CUR) {
            # Find deleg id
            my $found = 0;
            my $open_deleg;
            my $saved_arg;
            foreach $open_deleg (@{$self->{open_delegs}}) {
                if ($opts{name} eq $open_deleg->{name}) {
                    $saved_arg = xdr_stateid4(encode => ($open_deleg->{stateid}));
                    $found = 1;
                }
            }
            croak(
                "OPEN with CLAIM_DELEGATE_CUR: delegation-id is not found for file $opts{name}"
                ) if (!$found);
            $args .= $saved_arg;
            $args .= $name;
        } else {

            # We assume that the name is actually the filehandle itself!
            # For the next few operands that are coming up, dhandle is the current filehandle.
            # But in case of CLAIM_PREVIOUS, the current filehandle has to be set to the actual
            # handle of the file to be opened. So for further processing, we need to set the
            # dhandle to the actual filehandle ($name). Remaining operations are same as the
            # case of CLAIM_DELEGATE_CUR.
            $Log->debug("changing the dhandle to the name itself: $name\n");
            $dhandle = $name;
            if ($opts{access} == OPEN4_SHARE_ACCESS_READ) {
                $args .= xdr_uint(encode => OPEN_DELEGATE_READ);
            } elsif ($opts{access} == OPEN4_SHARE_ACCESS_WRITE) {
                $args .= xdr_uint(encode => OPEN_DELEGATE_WRITE);
            }
        } ## end else [ if ($opts{deleg_type} ...
    } else {
        $args .= $name;
    }

    # Create compound request
    my @ops = ({op => OP_PUTFH, args => $dhandle},
               {op => OP_OPEN,  args => $args},
               {op => OP_GETFH});

    # Retry block
    {

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'open',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

        # Retry call after 10 seconds if specified
        if ($out && $retry-- > 0 and $out->{status} == NFS4ERR_GRACE) {

            # Use a different sequence id for retry
            substr($ops[1]{args}, 0, 4) =
              xdr_uint(encode => $self->{seq_id}{$owner}++);
            sleep 10;
            redo;
        }

        # Retry call after returning delegation
        if ($out && $deleg-- > 0 and $out->{status} == NFS4ERR_DELAY) {
            if ($return_deleg) {

                # Answer callback and return delegation
                $self->callback_process();
                $return_deleg = 0;
            }

            # Use a different sequence id for retry
            substr($ops[1]{args}, 0, 4) =
              xdr_uint(encode => $self->{seq_id}{$owner}++);
            sleep $self->{nfs_delay};
            redo;
        } ## end if ($out && $deleg-- >...
    }

    # Check for errors
    $self->error_check(name => 'OPEN', result => $out);

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{results}[2]{result}{fhandle});

    # Create a copy of the stateid reference for ease of use
    $out->{stateid} = $out->{results}[1]{result}{stateid};

    # Save the filehandle for this open
    $out->{fhandle} = $self->{curr_fh};

    if (exists($out->{results}[1]{result}{deleg})) {

        # Create a copy of the delegation stateid reference for ease of use
        $out->{deleg_stateid} = $out->{results}[1]{result}{deleg}{stateid};

        # Save it in a list of open delegations
        push @{$self->{open_delegs}},
          { fhandle => $self->{curr_fh},
            stateid => $out->{deleg_stateid},
            name    => $opts{name},};
    } ## end if (exists($out->{results...

    $Log->exit();
    return $out;
} ## end sub open($)

#---------------------------------------------------------------------

=head2 open_downgrade

  $out = $nfs_obj->open_downgrade(stateid => $stateid
                                  [, owner => $owner ]
                                  [, access => $access ]
                                  [, deny => $deny ]
                                  [, fhandle => $fhandle ]);

Calls the OPEN_DOWNGRADE procedure on the server.
This function is used to adjust the share access and share deny
bits for a given open. This is necessary when a given open owner
opens the same file multiple times with different share access and
share deny flags. In this situation, a close of one of the opens may
change the appropriate share access and share deny flags to remove
bits associated with opens no longer in effect.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be closed
(default is the current filehandle).

=item stateid

(Required) : Reference to a hash containing the stateid of the opened file.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item access

(Optional) : Share access mode to set: OPEN4_SHARE_ACCESS_READ (default),
OPEN4_SHARE_ACCESS_WRITE or OPEN4_SHARE_ACCESS_BOTH.

=item deny

(Optional) : Share deny mode to set: OPEN4_SHARE_DENY_NONE (default),
OPEN4_SHARE_DENY_READ, OPEN4_SHARE_DENY_WRITE or OPEN4_SHARE_DENY_BOTH.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub open_downgrade($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                stateid => {type => HASHREF},
                owner   => {type => SCALAR, default => $$},
                access  => {
                           type    => SCALAR,
                           default => OPEN4_SHARE_ACCESS_READ,
                           regex   => qr/^\d+$/},
                deny => {type    => SCALAR,
                         default => OPEN4_SHARE_DENY_NONE,
                         regex   => qr/^\d+$/},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $owner = $opts{owner};

    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the Stateid of open,
    my $args = xdr_stateid4(encode => $opts{stateid}) .

      # Sequence id,
      xdr_uint(encode => $self->{seq_id}{$owner}++) .

      # Share access,
      xdr_uint(encode => $opts{access}) .

      # Share deny
      xdr_uint(encode => $opts{deny});

    # Create compound request
    my @ops = ({op => OP_PUTFH,          args => $fhandle},
               {op => OP_OPEN_DOWNGRADE, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'open dgrade',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'OPEN_DOWNGRADE', result => $out);

    # Create a copy of the stateid reference for ease of use
    $out->{stateid} = $out->{results}[1]{result}{stateid};

    return $out;
} ## end sub open_downgrade($)

#---------------------------------------------------------------------

=head2 open_confirm

  $out = $nfs_obj->open_confirm(stateid => $stateid
                                [, owner   => $owner ]
                                [, fhandle => $fhandle ]);

Calls the OPEN_CONFIRM procedure on the server.
This function is used to confirm the sequence id usage for the first
time that an open owner is used by a client.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the opened file to be confirmed
(default is the current filehandle).

=item stateid

(Required) : Reference to a hash containing the stateid of the opened file.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub open_confirm($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                stateid => {type => HASHREF},
                owner   => {type => SCALAR, default => $$},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $owner = $opts{owner};

    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the Stateid of open,
    my $args = xdr_stateid4(encode => $opts{stateid}) .

      # Sequence id,
      xdr_uint(encode => $self->{seq_id}{$owner}++);

    # Create compound request
    my @ops = ({op => OP_PUTFH,        args => $fhandle},
               {op => OP_OPEN_CONFIRM, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'open conf',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'OPEN_CONFIRM', result => $out);

    # Create a copy of the stateid reference for ease of use
    $out->{stateid} = $out->{results}[1]{result}{stateid};

    return $out;
} ## end sub open_confirm($)

#---------------------------------------------------------------------

=head2 openattr

  $out = $nfs_obj->openattr([ fhandle => $fhandle ]
                            [, createdir => 0|1 ]);

Calls the OPENATTR procedure on the server.
This function is used to obtain the filehandle of the named attribute
directory associated with the specified filehandle.
It is only supported on NFSv4.
It sets the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the object whose named attribute directory
is to be retrieved (default is the current filehandle).

=item createdir

(Required) : Reference to a hash containing the stateid of the opened file.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub openattr($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                createdir =>
                  {type => SCALAR, default => 0, regex => qr/^[01]$/},});
    my $out;
    my $ver = $self->{version};

    # Encode the filehandle
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the createdir flag
    my $args = xdr_uint(encode => $opts{createdir});

    # Create compound request
    my @ops = ({op => OP_PUTFH,    args => $fhandle},
               {op => OP_OPENATTR, args => $args},
               {op => OP_GETFH});

    # Call the compound procedure
    $out = $self->nfs_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'openattr',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'OPENATTR', result => $out);

    # Set the current filehandle
    $self->current_fh_set(fhandle => $out->{results}[2]{result}{fhandle});

    # Save the filehandle for this openattr
    $out->{fhandle} = $self->{curr_fh};

    return $out;
} ## end sub openattr($)

#---------------------------------------------------------------------

=head2 close

  $out = $nfs_obj->close(stateid => $stateid
                         [, owner => $owner ]
                         [, fhandle => $fhandle ]);

Calls the CLOSE procedure on the server.
This function closes a regular file.
The CLOSE operation releases share reservations for the regular or
named attribute file as specified by the current filehandle.  The
share reservations and other state information released at the server
as a result of this CLOSE is only associated with the supplied
stateid.  The sequence id provides for the correct ordering. State
associated with other OPENs is not affected.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be closed
(default is the current filehandle).

=item stateid

(Required) : Reference to a hash containing the stateid of the opened file.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub close($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                owner   => {type => SCALAR, default => $$},
                stateid => {type => HASHREF},});
    my $out;
    my $rpc     = $self->{rpc};
    my $ver     = $self->{version};
    my $retry   = $self->{nfs_retry};
    my $stateid = $opts{stateid};
    my $owner   = $opts{owner};

    # Convert filehandle of file
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Retry block, not using nfs_call since the seqid needs to be
    # incremented for every retry
    {

        # Encode the sequence id,
        my $args = xdr_uint(encode => $self->{seq_id}{$owner}++) .

          # Stateid of open
          xdr_stateid4(encode => $stateid);

        # Create compound request
        my @ops =
          ({op => OP_PUTFH, args => $fhandle}, {op => OP_CLOSE, args => $args});

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'close',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

        # Retry call after 10 seconds if specified
        if ($out && $retry-- > 0 && $retry_errors{$out->{status}}) {
            sleep $self->{nfs_delay};
            redo;
        }
    }

    # Check for errors
    $self->error_check(name => 'CLOSE', result => $out);

    return $out;
} ## end sub close($)

#---------------------------------------------------------------------

=head2 lock

  $out = $nfs_obj->lock(offset => $offset,
                        count => $count
                        [, lock_stateid => \%lock_stateid |
                           stateid => \%stateid
                           [, owner => $owner [, ownrlen => $ownrlen ] ] ]
                        [, fhandle => $fhandle ]
                        [, type => $lock_type ]
                        [, reclaim => 0|1 ]
                        [, caller_name => $caller_name ]
                        [, cookie => $cookie ]
                        [, svid => $svid ]);

Calls the LOCK (NFSv4) procedure on the server.
This function requests a record lock for the byte range specified
by the offset and count parameters.

On NFSv4 only, to get a lock using an existing
lock stateid (from a previous lock, using the same owner) use the
lock_stateid parameter. On the other hand, to get a lock using a new owner,
the stateid of the opened file and the owner must be specified.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be locked
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the lock is to begin.

=item count

(Required) : The number of bytes of data that are to be locked.

=item type

(Optional) : The type of lock to get: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item reclaim

(Optional) : True for a reclaim request (default is 0).

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item stateid

(Optional) : Reference to a hash containing the stateid of the opened file.
It is only supported on NFSv4.

=item lock_stateid

(Optional) : Reference to a hash containing the stateid of a lock owner.
It is only supported on NFSv4.

=item ownrlen

(Optional) : Force the length of owner to this number.

=item caller_name

(Optional) : Identifies the client making the request
(default is the client's local address or host name).
It is only supported on NFSv2 and NFSv3.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=item svid

(Optional) : Identifier of the process making the request
(default is the current process id).
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure
as well as the lock_stateid and the lockid to be used in locku()
and lock_cancel().

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lock($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle     => {type => SCALAR, default => $self->{curr_fh}},
                caller_name => {type => SCALAR, default => $self->{mon_host}},
                cookie      => {type => SCALAR, default => ''},
                offset      => {type => SCALAR, regex   => qr/^\d+$/},
                count       => {type => SCALAR, regex   => qr/^\d+$/},
                type        => {
                         type    => SCALAR,
                         default => READ_LT,
                         regex   => qr/^\d+$/},
                reclaim => {type    => SCALAR,
                            default => 0,
                            regex   => qr/^[01]$/},
                owner        => {type => SCALAR,  default  => $$},
                stateid      => {type => HASHREF, optional => 1},
                lock_stateid => {type => HASHREF, optional => 1},
                ownrlen      => {
                            type     => SCALAR,
                            optional => 1,
                            regex    => qr/^\d+$/},
                svid => {type    => SCALAR,
                         default => $$,
                         regex   => qr/^\d+$/},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $retry = $self->{nfs_retry};
    my %len   = (len => $opts{ownrlen}) if defined($opts{ownrlen});
    my $owner = $opts{owner};

    # Convert filehandle of file into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Retry block, not using nfs_call since the seqid and lock_seqid
    # need to be incremented for every retry
    {

        # Encode the locktype,
        my $args = xdr_uint(encode => $opts{type}) .

          # Reclaim,
          xdr_uint(encode => $opts{reclaim}) .

          # Offset,
          xdr_uint64(encode => $opts{offset}) .

          # Length,
          xdr_uint64(encode => $opts{count});

        # Encode the locker:
        if (defined($opts{lock_stateid})) {

            # Encode known locker
            $args .=
                xdr_uint(encode => 0)
              . xdr_stateid4(encode => $opts{lock_stateid})
              . xdr_uint(encode => $self->{lock_seqid}++);
        } else {

            # Encode new locker
            $args .=
                xdr_uint(encode => 1)
              . xdr_uint(encode => $self->{seq_id}{$owner}++)
              . xdr_stateid4(encode => $opts{stateid})
              . xdr_uint(encode => $self->{lock_seqid}++)
              .

              # Owner: client id
              $self->{clientid} .

              #        owner
              xdr_opaque(encode => $opts{owner}, %len);
        } ## end else [ if (defined($opts{lock_stateid...

        # Create compound request
        my @ops =
          ({op => OP_PUTFH, args => $fhandle}, {op => OP_LOCK, args => $args});

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'lock',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

        # Retry call after 10 seconds if specified
        if ($out && $retry-- > 0 && $retry_errors{$out->{status}}) {
            sleep $self->{nfs_delay};
            redo;
        }
    }

    # Check for errors
    $self->error_check(name => 'LOCK', result => $out);

    # Create a copy of the lock stateid reference for ease of use
    $out->{lock_stateid} = $out->{results}[1]{result}{stateid};

    # Save lockid to be used on locku()
    $out->{lockid} = {fhandle => $opts{fhandle},
                      stateid => $out->{lock_stateid},
                      offset  => $opts{offset},
                      count   => $opts{count},
                      type    => $opts{type},};

    return $out;
} ## end sub lock($)

#---------------------------------------------------------------------

=head2 locku

  $out = $nfs_obj->locku([ lockid => \%lockid ]
                         [, stateid => $lock_stateid ]
                         [, offset => $offset ]
                         [, count => $count ]
                         [, fhandle => $fhandle ]
                         [, type => $lock_type ]
                         [, caller_name => $caller_name ]
                         [, cookie => $cookie ]
                         [, owner => $owner ]
                         [, svid => $svid ]);

Calls the LOCKU (NFSv4) procedure on the server.
This function unlocks the record lock specified by the parameters.

The following named arguments are accepted:

=over

=item Arguments

=over

=item lockid

(Optional) : Reference to a hash of the lock to be unlocked as returned by
the lock() function. All other options can be used to overwrite options
in lockid.
If this argument is not specified the following options are required:
offset, count and stateid(NFSv4).

=item fhandle

(Optional) : Filehandle of the file to be unlocked
(default is the current filehandle).

=item stateid

(Optional) : Reference to a hash containing the stateid of a lock owner.
It is only supported on NFSv4.

=item offset

(Optional) : The position within the file at which the unlock is to begin.

=item count

(Optional) : The number of bytes of data that are to be unlocked.

=item type

(Optional) : The type of lock to release: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item caller_name

(Optional) : Identifies the client making the request
(default is the client's local address or host name).
It is only supported on NFSv2 and NFSv3.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=item owner

(Optional) : Owner of file locking state (default is the current process id).
It is only supported on NFSv2 and NFSv3.

=item svid

(Optional) : Identifier of the process making the request
(default is the current process id).
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub locku($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle     => {type => SCALAR,  optional => 1},
                caller_name => {type => SCALAR,  optional => 1},
                cookie      => {type => SCALAR,  default  => ''},
                lockid      => {type => HASHREF, optional => 1},
                stateid     => {type => HASHREF, optional => 1},
                owner       => {type => SCALAR,  optional => 1},
                offset      => {
                           type     => SCALAR,
                           optional => 1,
                           regex    => qr/^\d+$/},
                count => {type     => SCALAR,
                          optional => 1,
                          regex    => qr/^\d+$/},
                type => {type     => SCALAR,
                         optional => 1,
                         regex    => qr/^\d+$/},
                svid => {type     => SCALAR,
                         optional => 1,
                         regex    => qr/^\d+$/},});
    my ($out, $args, $fhandle, @ops, $nver, $lockid);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Get lockid from options
    $lockid = $opts{lockid};

    if (defined($lockid)) {

        # Explicit options have precedence
        $opts{fhandle} = $lockid->{fhandle} if not defined($opts{fhandle});
        $opts{stateid} = $lockid->{stateid} if not defined($opts{stateid});
        $opts{offset}  = $lockid->{offset}  if not defined($opts{offset});
        $opts{count}   = $lockid->{count}   if not defined($opts{count});
        $opts{type}    = $lockid->{type}    if not defined($opts{type});
    } else {

        # Default values for unspecified options
        $opts{fhandle} = $self->{curr_fh} if not defined($opts{fhandle});
        $opts{type}    = READ_LT          if not defined($opts{type});
    }

    # Convert filehandle of file into XDR format
    $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the locktype,
    $args = xdr_uint(encode => $opts{type}) .

      # Lock sequence id,
      xdr_uint(encode => $self->{lock_seqid}++) .

      # Lock stateid,
      xdr_stateid4(encode => $opts{stateid}) .

      # Offset,
      xdr_uint64(encode => $opts{offset}) .

      # Length,
      xdr_uint64(encode => $opts{count});

    # Create compound request
    @ops =
      ({op => OP_PUTFH, args => $fhandle}, {op => OP_LOCKU, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'locku',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'LOCKU', result => $out);

    return $out;
} ## end sub locku($)

#---------------------------------------------------------------------

=head2 lockt

  $out = $nfs_obj->lockt(offset => $offset,
                         count => $count
                         [, owner => $owner [, ownrlen => $ownrlen ] ] ]
                         [, fhandle => $fhandle ]
                         [, type => $lock_type ]
                         [, cookie => $cookie ]);

Calls the LOCKT (NFSv4) procedure on the server.
This function checks whether a lock is available to the client.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of the file to be tested for lock
(default is the current filehandle).

=item offset

(Required) : The position within the file at which the lock is to begin.

=item count

(Required) : The number of bytes of data that are to be locked.

=item type

(Optional) : The type of lock to test: READ_LT (default), WRITE_LT,
READW_LT, WRITEW_LT.

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item ownrlen

(Optional) : Force the length of owner to this number.

=item cookie

(Optional) : Cookie used to match asynchronous response with a request
(default is '').
It is only supported on NFSv2 and NFSv3.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lockt($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                cookie  => {type => SCALAR, default => ''},
                offset  => {type => SCALAR, regex   => qr/^\d+$/},
                count   => {type => SCALAR, regex   => qr/^\d+$/},
                type    => {
                         type    => SCALAR,
                         default => READ_LT,
                         regex   => qr/^\d+$/},
                owner   => {type => SCALAR, default => $$},
                ownrlen => {type     => SCALAR,
                            optional => 1,
                            regex    => qr/^\d+$/},});
    my $out;
    my $ver = $self->{version};
    my %len = (len => $opts{ownrlen}) if defined($opts{ownrlen});

    # Convert filehandle of file into XDR format
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the locktype,
    my $args = xdr_uint(encode => $opts{type}) .

      # Offset,
      xdr_uint64(encode => $opts{offset}) .

      # Length,
      xdr_uint64(encode => $opts{count}) .

      # Lock owner: client id
      $self->{clientid} .

      #             owner
      xdr_opaque(encode => $opts{owner}, %len);

    # Create compound request
    my @ops =
      ({op => OP_PUTFH, args => $fhandle}, {op => OP_LOCKT, args => $args});

    # Call the compound procedure
    $out = $self->nfs_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'lockt',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Create a copy of the lock4denied structure reference for ease of use
    $out->{lock_denied} = $out->{results}[1]{result}
      if $out && $out->{status} == NFS4ERR_DENIED;

    # Check for errors
    $self->error_check(name => 'LOCKT', result => $out);

    return $out;
} ## end sub lockt($)

#---------------------------------------------------------------------

=head2 release_lockowner

  $out = $nfs_obj->release_lockowner([, owner => $owner
                                        [, ownrlen => $ownrlen ] ]);

Calls the RELEASE_LOCKOWNER procedure on the server.
This function is used to notify the server that the lock owner is no
longer in use by the client. This allows the server to release cached
state related to the specified lock owner.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item owner

(Optional) : Owner of file locking state (default is the current process id).

=item ownrlen

(Optional) : Force the length of owner to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub release_lockowner($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {owner   => {type => SCALAR, default => $$},
                ownrlen => {type     => SCALAR,
                            optional => 1,
                            regex    => qr/^\d+$/},});
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};
    my %len = (len => $opts{ownrlen}) if defined($opts{ownrlen});

    # Encode the Lock owner: client id
    my $args = $self->{clientid} .

      #             owner
      xdr_opaque(encode => $opts{owner}, %len);

    # Create compound request
    my @ops = ({op => OP_RELEASE_LOCKOWNER, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'rlse lockowner',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'RELEASE_LOCKOWNER', result => $out);

    return $out;
} ## end sub release_lockowner($)

#---------------------------------------------------------------------

=head2 delegreturn

  $out = $nfs_obj->delegreturn(stateid => $deleg_stateid
                               [, fhandle => $fhandle ]);

Calls the DELEGRETURN procedure on the server.
This function returns an open delegation represented by the
current filehandle and stateid.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Optional) : Filehandle of opened file (default is the current filehandle).

=item stateid

(Required) : Reference to a hash containing the stateid of the opened file.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub delegreturn($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {fhandle => {type => SCALAR, default => $self->{curr_fh}},
                stateid => {type => HASHREF},});
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Convert filehandle of file
    my $fhandle = xdr_fhandle(encode => $opts{fhandle}, version => $ver);

    # Encode the stateid of open delegation
    my $args = xdr_stateid4(encode => $opts{stateid});

    # Create compound request
    my @ops = ({op => OP_PUTFH,       args => $fhandle},
               {op => OP_DELEGRETURN, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'delegreturn',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'DELEGRETURN', result => $out);

    # Remove open delegation from list
    $self->deleg_remove(\%opts);

    return $out;
} ## end sub delegreturn($)

#---------------------------------------------------------------------

=head2 delegpurge

  $out = $nfs_obj->delegpurge();

Calls the DELEGPURGE procedure on the server.
This function purges all of the delegations awaiting recovery for a
given client. This is useful for clients which do not commit delegation
information to stable storage to indicate that conflicting requests need
not be delayed by the server awaiting recovery of delegation information.
It is only supported on NFSv4.

=over

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub delegpurge($) {
    my $self = shift;
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Create compound request
    my @ops = ({op => OP_DELEGPURGE, args => $self->{clientid}});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'delegpurge',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'DELEGPURGE', result => $out);

    return $out;
} ## end sub delegpurge($)

#---------------------------------------------------------------------

=head2 renew

  $out = $nfs_obj->renew();

Calls the RENEW procedure on the server.
This function renews all leases which are currently held
by this client at the server.
It is only supported on NFSv4.

It returns a reference to a hash containing the status and output of procedure.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub renew($;) {
    my $self = shift;
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Create compound request
    my @ops = ({op => OP_RENEW, args => $self->{clientid}});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'renew',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'RENEW', result => $out);

    return $out;
} ## end sub renew($;)

#---------------------------------------------------------------------

=head2 create

  $out = $nfs_obj->create(name => $name [, namelen => $namelen ]
                          [, dhandle => $fhandle ]
                          [, type => $type ]
                          [, attr => \%sattr ]
                          [, cr_mode => $mode ]
                          [, data => $data [, datalen => $datalen ] ]
                          [, major => $major, minor => $minor ]);

Calls the CREATE procedure on the server.
This function creates different types of file system objects.
For NFSv2, it creates regular files, special device files,
named sockets or named pipes.
For NFSv3, it only creates regular files and accepts a creation
mode flag.
For NFSv4, it creates everything but a regular file including
directories, symbolic links, special device files, named
sockets and named pipes.
To have similar behavior with mkdir and open, it sets the current
filehandle if creating a regular file on NFSv2 and v3
and if creating a directory on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory in which the new file
is to be created (default is the current filehandle).

=item name

(Required) : The name that is to be associated with the created file.

=item type

(Optional) : The type of file to be created: NFREG (default), NFDIR,
NFBLK, NFCHR, NFLNK, NFSOCK or NFFIFO.

=item attr

(Optional) : The initial attributes for the new file
(default is no initial attributes).

=item data

(Optional) : The string containing the symbolic link data.
Used only in NFSv4.

=item major

(Optional) : The major device number, only for type NFBLK and NFCHR.
Used in NFSv2 and v4.

=item minor

(Optional) : The minor device number, only for type NFBLK and NFCHR.
Used in NFSv2 and v4.

=item verf

(Optional) : Create verifier. Only used if cr_mode is EXCLUSIVE.
Default is the current time.

=item namelen

(Optional) : Force the length of name to this number.

=item datalen

(Optional) : Force the length of data to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub create($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                type    => {
                         type    => SCALAR,
                         default => NFREG,
                         regex   => qr/^\d+$/},
                attr    => {type => HASHREF, default => {}},
                cr_mode => {type    => SCALAR,
                            default => UNCHECKED,
                            regex   => qr/^\d+$/},
                data    => {type => SCALAR, optional => 1},
                major   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                minor   => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                verf    => {type => SCALAR, default  => time},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},
                datalen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $args, $spec, $attr, $fhandle);
    my $ver  = $self->{version};
    my %nlen = (len => $opts{namelen}) if defined($opts{namelen});
    my %dlen = (len => $opts{datalen}) if defined($opts{datalen});

    # Convert the major and minor numbers into XDR format
    # if this is a special device file
    if ($opts{type} == NFBLK or $opts{type} == NFCHR) {
            # For NFSv4 use the specdata structure
            $spec = xdr_specdata(encode => {specdata1 => $opts{major},
                                            specdata2 => $opts{minor}});
    } ## end if ($opts{type} == NFBLK...

    # Convert filehandle of directory and the name
    # of the new file into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %nlen);

        # Use the fattr4 structure for the initial sattr
        # instead of using the sattr structure
        $attr = xdr_fattr4(encode => $opts{attr});

        # Encode the object type
        $args = xdr_uint(encode => $opts{type});
        if ($opts{type} == NFLNK) {

            # Add the linkdata if symbolic link is requested
            $args .= xdr_opaque(encode => $opts{data}, %dlen);
        } elsif ($opts{type} == NFBLK or $opts{type} == NFCHR) {

            # Otherwise, add the specdata structure
            # for block or character special devices
            $args .= $spec;
        }

        # Finally, append the name of the new object
        # and the initial attributes
        $args .= $name . $attr;

        # Create compound request
        my @ops = ({op => OP_PUTFH,  args => $dhandle},
                   {op => OP_CREATE, args => $args},
                   {op => OP_GETFH});

        # Call the compound procedure
        $out = $self->nfs_call(proc => 1,
                               arg =>
                                 xdr_compound(tag => 'create',
                                              ops => \@ops),
                               out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'CREATE', result => $out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[2]{result}{fhandle};

    # Set the current filehandle
    if ($opts{type} == NFDIR or $opts{type} == NFREG) {
        $self->current_fh_set(fhandle => $out->{fhandle});
    }

    return $out;
} ## end sub create($)

#---------------------------------------------------------------------

=head2 secinfo

  $out = $nfs_obj->secinfo(name => $name [, namelen => $namelen ]
                          [, dhandle => $dhandle ]);

Calls the SECINFO procedure on the server.
Searches a directory for a specific name and returns a list of valid
RPC authentication flavors for the specific directory filehandle,
file name pair.
It is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dhandle

(Optional) : Filehandle of the directory to search
(default is the current filehandle).

=item name

(Required) : The name of the object to get the SECINFO.

=item namelen

(Optional) : Force the length of name to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub secinfo($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {dhandle => {type => SCALAR, default => $self->{curr_fh}},
                name    => {type => SCALAR},
                namelen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($out, $fhandle);
    my $rpc = $self->{rpc};
    my $ver = $self->{version};
    my %len = (len => $opts{namelen}) if defined($opts{namelen});

    # Convert filehandle of directory and
    # the name of object into XDR format
    my $dhandle = xdr_fhandle(encode => $opts{dhandle}, version => $ver);
    my $name = xdr_opaque(encode => $opts{name}, %len);

    # Create compound request
    my @ops =
      ({op => OP_PUTFH, args => $dhandle}, {op => OP_SECINFO, args => $name});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'secinfo',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'SECINFO', result => $out);

    # Create a copy of the secinfo reference for ease of use
    $out->{secinfo} = $out->{results}[1]{result}{secinfo};

    return $out;
} ## end sub secinfo($)

#---------------------------------------------------------------------

=head2 setclientid

  $out = $nfs_obj->setclientid([ id => $client_id ] [, idlen => $idlen ]
                               [, verifier => $verifier ]);

Calls the SETCLIENTID procedure on the server.
The client uses the SETCLIENTID operation to notify the server of its
intention to use a particular client identifier, callback and
callback_ident for subsequent requests that entail creating lock,
share reservation and delegation state on the server.  Upon
successful completion, the server will return a shorthand clientid
which, if confirmed via a separate step, will be used in subsequent
file locking and file open requests. Confirmation of the clientid
must be done via the SETCLIENTID_CONFIRM operation to return the
clientid and setclientid_confirm values, as verifiers, to the server.

It saves the clientid and the setclientid_confirm verifier on the
object for later use.
This function is only supported on NFSv4.

The following named arguments are accepted:

=over

=item Arguments

=over

=item id

(Optional) : This number uniquely identifies the client.
The default is the client identifier defined at object creation.

=item verifier

(Optional) : Client incarnation verifier that is used to detect client
reboots.  Only if the verifier is different from that which the
server has previously recorded the client does the server
start the process of canceling the client's leased state.
The default is the client verifier defined at object creation.

=item idlen

(Optional) : Force the length of id to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setclientid($;) {
    my $self = shift;
    $Log->enter();
    my (%opts) =
      validate(@_,
               {id => {type    => SCALAR,
                       default => $self->{cl_id},
                       regex   => qr/^\d+$/},
                verifier => {type    => SCALAR,
                             default => $self->{cl_verifier},
                             regex   => qr/^\d+$/},
                idlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},});
    my $out;
    my $hport   = 0;
    my $lport   = 0;
    my $program = 0;
    my $rpc     = $self->{rpc};
    my $ver     = $self->{version};
    my %len     = (len => $opts{idlen}) if defined($opts{idlen});

    my $tcp_type = 'tcp'; # default ipv4
    # ipv6 => tcp6
    if ( $self->{addr_type} eq 'inet6' ) {
	$tcp_type = 'tcp6';
    }

    # Convert client verifier into a string
    # and create the id string
    my $verifier = xdr_uint(encode => $opts{verifier});

    # The id string is a constant and unique value for the client
    # it is created by concatenating both the ip addresses of
    # the client and server
    my $id =
        $rpc->{sock}->peeraddr()
      . $rpc->{sock}->sockaddr
      . xdr_uint(encode => $opts{id});

    if (defined($self->{callback})) {

        # Change the default values with
        # the real callback information
        $program = $self->{callback}{program};
        $hport   = ($self->{callback}{port} >> 8) & 0x0FF;
        $lport   = $self->{callback}{port} & 0x0FF;
    }

    # Encode the verifier, id, callback program, callback location
    # and callback identifier
    my $args =
        xdr_opaque(encode => $verifier, fixed => 1, maxlen => 8)
      . xdr_opaque(encode => $id, %len)
      . xdr_uint(encode => $program)
      . xdr_opaque(encode => $tcp_type)
      . xdr_opaque(encode => $rpc->{sock}->sockhost() . ".$hport.$lport")
      . xdr_uint(encode => $program);

    # Create compound request
    my @ops = ({op => OP_SETCLIENTID, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'setclientid',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'SETCLIENTID', result => $out);

    # Save the clientid and setclientid_confirm verifier
    $self->{clientid} = $out->{results}[0]{result}{clientid};
    $self->{confirm}  = $out->{results}[0]{result}{confirm};

    $Log->exit();
    return $out;
} ## end sub setclientid($;)

#---------------------------------------------------------------------

=head2 setclientid_confirm

  $out = $nfs_obj->setclientid_confirm();

Calls the SETCLIENTID_CONFIRM procedure on the server.
This operation is used by the client to confirm the results from a
previous call to SETCLIENTID.  The client provides the server
supplied (from a SETCLIENTID response) clientid.  The server responds
with a simple status of success or failure.
This function is only supported on NFSv4.

It returns a reference to a hash containing the status and output of procedure.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub setclientid_confirm($;) {
    my $self = shift;
    $Log->enter();
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};

    # Add the clientid and verifier to the arguments
    my $args = $self->{clientid} . $self->{confirm};

    # Create compound request
    my @ops = ({op => OP_SETCLIENTID_CONFIRM, args => $args});

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'setclientid_confirm',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'SETCLIENTID_CONFIRM', result => $out);

    $Log->exit();
    return $out;
} ## end sub setclientid_confirm($;)

#---------------------------------------------------------------------

=head2 mount

  $out = $nfs_obj->mount(path => $path [, pathlen => $pathlen ]);

Returns the root filehandle of the path to be mounted.
In NFSv4, the operation PUTROOTFH is issue to set the current
filehandle to the root of the server's file tree, then using LOOKUP
to traverse the given path and finally GETFH to get the filehandle for
the path to be mounted.

It saves the mountpoint filehandle on object's blessed hash
under the 'mnt_fh' key, as well as setting the current filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : The pathname to be mounted.

=item pathlen

(Optional) : Force the length of path to this number.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub mount($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {path    => {type => SCALAR},
                pathlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my ($rpc, $out);
    my $ver = $self->{version};
    my %len = (len => $opts{pathlen}) if defined($opts{pathlen});

        # Mount on NFSv4 is implemented on the NFS protocol
        # so use the current Rpc object
        $rpc = $self->{rpc};

        # Create a list of directories for the path to be mounted
        $opts{path} =~ s{/}{};
        my @paths = split /\//, $opts{path};

        # Create compound request:
        #   PUTROOTFH to get the filehandle of the pseudo-fs root
        my @ops = ({op => OP_PUTROOTFH});

        #   traverse the path using LOOKUP
        push @ops,
          map { {op => OP_LOOKUP, args => xdr_opaque(encode => $_, %len)} }
          @paths;

        #   and finally get the filehandle of the last LOOKUP operation
        push @ops, {op => OP_GETFH};

        # Call the compound procedure
        #$out = $rpc->clnt_call(proc => 1,
        #                       arg =>
        #                         xdr_compound(tag => 'mount',
        #                                      ops => \@ops),
        #                       out => \&xdr_compound_res);

        $out = $self->nfs_call(proc => 1,
            arg => xdr_compound(tag => 'mount', ops => \@ops),
            out => \&xdr_compound_res);

        # Check for errors
        $self->error_check(name => 'MOUNT', result => $out);

        # Create a copy of the filehandle for ease of use
        $out->{fhandle} = $out->{results}[-1]{result}{fhandle};

        # Set the client id and confirm it
        $self->setclientid();
        $self->setclientid_confirm();
        ++$self->{cl_verifier};

    # Set current filehandle
    $self->current_fh_set(fhandle => $out->{fhandle});

    # Save the root filehandle on the NfsClient object
    $self->{mnt_fh} = $out->{fhandle};

    # Save mount path
    $self->{mnt_path} = $opts{path};

    return $out;
} ## end sub mount($)

#---------------------------------------------------------------------

=head2 umount

  $nfs_obj->umount(path => $path [, pathlen => $pathlen ]);

For NFSv4, this function does nothing since there
is no state associated with the root filehandle on the server for
any given path.

The following named arguments are accepted:

=over

=item Arguments

=over

=item path

(Required) : The pathname to be unmounted.

=item pathlen

(Optional) : Force the length of path to this number.

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub umount($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {path    => {type => SCALAR},
                pathlen => {type => SCALAR, optional => 1, regex => qr/^\d+$/},}
               );
    my $rpc;
    my $ver = $self->{version};
    my %len = (len => $opts{pathlen}) if defined($opts{pathlen});

        # Return open delegations
        foreach my $deleg (@{$self->{open_delegs}}) {
            $self->delegreturn(fhandle => $deleg->{fhandle},
                               stateid => $deleg->{stateid});
        }

        # No umount on NFSv4 so just return
        return;
} ## end sub umount($)

#---------------------------------------------------------------------

=head2 putpubfh

  $nfs_obj->putpubfh();

Calls OP_PUTPUBFH and returns the filehandle returned by the server.

=item Returns

The pub filehandle.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub putpubfh {
    my $self = shift;
    my $rpc  = $self->{rpc};
    my $out;
    my $vers = $self->{version};

    # Create compound request:
    #   PUTPUBFH to get the filehandle of the public filehandle from the server
    my @ops = ({op => OP_PUTPUBFH});

    #   and finally get the filehandle that should represent the last putpubfh
    push @ops, {op => OP_GETFH};

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'putpubfh',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'PUTPUBFH', result => $out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[-1]{result}{fhandle};

    # Save the current filehandle on prev_fh and set
    # the current filehandle to the one provided
    $self->{prev_fh} = $self->{curr_fh};
    $self->{curr_fh} = $out->{fhandle};
    $self->{pub_fh}  = $out->{fhandle};

    return $self->{pub_fh};
} ## end sub putpubfh

#---------------------------------------------------------------------

=head2 putrootfh

  $nfs_obj->putrootfh();

Calls OP_PUTROOTFH and returns the filehandle returned by the server.

=item Returns

The root filehandle.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub putrootfh {
    my $self = shift;
    my $rpc  = $self->{rpc};
    my $out;
    my $vers = $self->{version};

    # Create compound request:
    #   PUTROOTFH to get the filehandle of the public filehandle from the server
    my @ops = ({op => OP_PUTROOTFH});

    #   and finally get the filehandle that should represent the last putrootfh
    push @ops, {op => OP_GETFH};

    # Call the compound procedure
    $out = $rpc->clnt_call(proc => 1,
                           arg =>
                             xdr_compound(tag => 'putrootfh',
                                          ops => \@ops),
                           out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'PUTROOTFH', result => $out);

    # Create a copy of the filehandle for ease of use
    $out->{fhandle} = $out->{results}[-1]{result}{fhandle};

    # Save the current filehandle on prev_fh and set
    # the current filehandle to the one provided
    $self->{prev_fh} = $self->{curr_fh};
    $self->{curr_fh} = $out->{fhandle};
    $self->{root_fh} = $out->{fhandle};

    return $self->{root_fh};
} ## end sub putrootfh

#---------------------------------------------------------------------

=head2 current_fh_set

  $nfs_obj->current_fh_set(fhandle => $fhandle);

Sets the current filehandle of the object.
All functions which require a filehandle uses this current
filehandle as the default.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fhandle

(Required) : Filehandle to set.

=back

=item Returns

The filehandle before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub current_fh_set($) {
    my $self = shift;
    my (%opts) = validate(@_, {fhandle => {type => SCALAR},});

    # Save the current filehandle on prev_fh and set
    # the current filehandle to the one provided
    $self->{prev_fh} = $self->{curr_fh};
    $self->{curr_fh} = $opts{fhandle};

    return $self->{prev_fh};
} ## end sub current_fh_set($)

#---------------------------------------------------------------------
# Internal function.
# This function checks for errors on the results hash.
# If the status is not NFS_OK, the function croaks with
# the corresponding error.
# The error message has the following format:
#   'name' NFS?ERR_error_name: error_message
# The 'name' argument is just a string to identify which function
# has failed, usually is set to the procedure name,
# e.g., LOOKUP, CREATE, etc.
#---------------------------------------------------------------------
sub error_check($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {name   => {type => SCALAR},
                result => {type => HASHREF | UNDEF},
                type   => {type => SCALAR, default => 'nfs'},});

    # Clear previous error result
    $self->{result} = undef;

    my (undef, $file, $line, undef) = caller(0);
    my (undef, undef, undef, $subr) = caller(1);
    my $callermsg = " ($subr at $file:$line) ";

    # Check if result is defined
    croak "No results received from server: $callermsg"
      if !defined $opts{result};

    if ($opts{result}{status}) {

        # Procedure returned an error therefore save the
        # result on the object so it can be retrieved by
        # the calling process if needed
        $self->{result} = $opts{result};

        if ($opts{type} eq 'nfs') {

            # Croak with the proper NFS error message
            croak $opts{name} . " "
              . nfs_error(error   => $opts{result}{status},
                          version => $self->{version}) . ": $callermsg";
        }
    } ## end if ($opts{result}{status...

    return;
} ## end sub error_check($)

#---------------------------------------------------------------------
# Internal function
# Calls a procedure on the server with the given arguments, it retries
# the request if any of the errors defined in %retry_errors is returned
#---------------------------------------------------------------------
sub nfs_call($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {proc  => {type => SCALAR},
                arg   => {type => SCALAR},
                out   => {type => CODEREF},
                retry => {type => SCALAR, default => $self->{nfs_retry}},});
    my $out;
    my $rpc   = $self->{rpc};
    my $ver   = $self->{version};
    my $retry = $opts{retry};

    # Retry block
    {

        # Call the NFS procedure
        $out = $rpc->clnt_call(proc => $opts{proc},
                               arg  => $opts{arg},
                               out  => $opts{out});

        # Retry call after nfs_delay seconds if specified
        # Do not retry for NFS version 2
        if (   $ver != 2
            && $out
            && $retry-- > 0
            && $retry_errors{$out->{status}}) {
            $Log->warn(qq{Retrying the operation, due to a transient error});
            sleep $self->{nfs_delay};
            redo;
        }
    }

    return $out;
} ## end sub nfs_call($)

#---------------------------------------------------------------------
# Internal function
# Remove the given delegation from the list of open delegations.
#---------------------------------------------------------------------
sub deleg_remove($$;) {
    my $self  = shift;
    my $deleg = shift;

    my $index = 0;

    # Search the specified delegation in list
    foreach my $open_deleg (@{$self->{open_delegs}}) {
        if ($deleg->{fhandle} eq $open_deleg->{fhandle}) {

            # Remove delegation from list
            splice @{$self->{open_delegs}}, $index, 1;
            last;
        }
        ++$index;
    }

    return;
} ## end sub deleg_remove($$;)

#---------------------------------------------------------------------

=head2 compound

  $out = $nfs_obj->compound(ops => \@op_list
                            [, tag => $tag ]
                            [, num => $num ]
                            [, size => $size ]);

Calls a compound request on NFS version 4. The compound request is made
of all operations listed on the ops argument. Each operation is a hash
having the following:

  {
      op   => Operation number,
      args => XDR encoded arguments,
  }

The following named arguments are accepted:

=over

=item Arguments

=over

=item ops

(Required) : Reference to an array of operations to be included in the
compound request.

=item tag

(Optional) : Tag for the compound request. Default is 'compound'.

=item num

(Optional) : Use this as the number of operations to encode instead
of the actual number of operations encoded.

=item size

(Optional) : Truncate the arguments to the compound procedure to
this size.

=back

=item Returns

Reference to a hash containing the status and output of procedure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compound($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {ops  => {type => ARRAYREF},
                tag  => {type => SCALAR, default => 'compound'},
                num  => {type => SCALAR, optional => 1},
                size => {type => SCALAR, optional => 1},});
    my ($out, $arg);
    my $ver = $self->{version};
    my %args = (num => $opts{num}) if defined $opts{num};

    # Encode compound arguments
    $arg = xdr_compound(tag => $opts{tag}, ops => $opts{ops}, %args);
    #$arg = xdr_compound(tag => "kerb-perl synth client", ops => $opts{ops}, %args);
    substr($arg, $opts{size}) = '' if defined $opts{size};

    # Call the compound procedure
    $out = $self->nfs_call(proc => 1, arg => $arg, out => \&xdr_compound_res);

    # Check for errors
    $self->error_check(name => 'COMPOUND', result => $out);

    return $out;
} ## end sub compound($)

#---------------------------------------------------------------------

=head2 clientid

  $clientid = $nfs_obj->clientid();

Returns the NFSv4 client id.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub clientid { return shift->{clientid} }

#---------------------------------------------------------------------

=head2 seqid

  $seqid = $nfs_obj->seqid([ seqid => $seqid_new ],
                           [ owner => $ownername ]);

Gets or sets the current open sequence id for the NFS object.
Only useful for NFS version 4.

=over

=item Arguments

=over

=item seqid

(Optional) : Set the current open sequence id to this.

=item owner

(Optional) : The interested seqid of the owner. Default: $$

=back

=item Returns

The open sequence id value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub seqid($) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {owner => {type => SCALAR, default  => $$},
                seqid => {type => SCALAR, optional => 1},});

    my $new_value = $opts{seqid};
    my $owner     = $opts{owner};
    my $old       = $self->{seq_id}{$owner};

    if (defined($new_value)) {
        $self->{seq_id}{$owner} = $new_value;
    }

    return $old;
} ## end sub seqid($)

#---------------------------------------------------------------------

=head2 lock_seqid

  $lock_seqid = $nfs_obj->lock_seqid([ $lock_seqid_new ]);

Gets or sets the current lock sequence id for the NFS object.
Only useful for NFS version 4.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the current lock sequence id to this.

=back

=item Returns

The lock sequence id value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub lock_seqid($$;) {
    my ($self, $new_value) = @_;
    my $old = $self->{lock_seqid};

    if (defined($new_value)) {
        $self->{lock_seqid} = $new_value;
    }

    return $old;
} ## end sub lock_seqid($$;)

#---------------------------------------------------------------------
# Destructor
# It unregisters the callback program for NFSv4
#---------------------------------------------------------------------
sub DESTROY {
    my $self = shift;

    if (defined($self->{callback})) {
        eval {
            nfs::Rpc::pmap_unset(program => $self->{callback}{program},
                                 version => $self->{callback}{version},
                                 proto   => $self->{callback}{proto},
                                 port    => $self->{callback}{port});
        };
    }
} ## end sub DESTROY

sub savefh($) {
	my $self = shift;
	my (%opts) = validate ( @_,
		{
			fhandle => { type => SCALAR, default => $self->{curr_fh} },
		}
	);
	my $out;
	my $rpc = $self->{rpc};
	my $ver = $self->{version};
	my $minor_version = 0;

	# Convert filehandle of object into XDR format
	my $fhandle = xdr_fhandle(encode=>$opts{fhandle}, version=>$ver);

	if ( $ver == 4 ) {
		# Create compound request
		my @ops;
		if ($self->{minor_version} == 0) {
			@ops = ( { op => OP_PUTFH, args => $fhandle },
			{ op => OP_SAVEFH } );
		}
		# Call the compound procedure
		$out = $rpc->clnt_call(proc => 1,
			arg  => xdr_compound(tag => 'savefh',
			minor => $minor_version,
			ops => \@ops),
			out  => \&xdr_compound_res);
	}

	# Check for errors
	$self->error_check(name=>'SAVEFH', result=>$out);

	return $out;
} #end sub savefh

sub nverify($) {
	 my $self = shift;
	 my (%opts) = validate ( @_,
		 {
			fhandle => { type => SCALAR, default => $self->{curr_fh} },
			attr    => { type => HASHREF },
		 }
	 );
	 my $out;
	 my $rpc = $self->{rpc};
	 my $ver = $self->{version};
	 my $minor_version = 0;

	 # Convert filehandle of object and
	 # the sattr structure into XDR format
	 my $fhandle = xdr_fhandle(encode=>$opts{fhandle}, version=>$ver);
	 my $sattr   = xdr_fattr4(encode=>$opts{attr});

	 if ( $ver == 4 ) {
		 # Create compound request
		 my @ops;
		 if ($self->{minor_version} == 0) {
			@ops = ( { op => OP_PUTFH, args => $fhandle },
			{ op => OP_NVERIFY, args => $sattr } );
		 }
		 # Call the compound procedure
		 $out = $rpc->clnt_call(proc => 1,
			arg  => xdr_compound(tag => 'nverify',
			minor => $minor_version,
			ops => \@ops),
			out  => \&xdr_compound_res);
	 }

	 # Check for errors
	 $self->error_check(name=>'NVERIFY', result=>$out);

	 return $out;
} #end sub nverify

sub verify($) {
    my $self = shift;
    my (%opts) = validate ( @_,
        {
            fhandle => { type => SCALAR, default => $self->{curr_fh} },
            attr    => { type => HASHREF },
        }
    );
    my $out;
    my $rpc = $self->{rpc};
    my $ver = $self->{version};
    my $minor_version = 0;

    # Convert filehandle of object and
    # the sattr structure into XDR format
    my $fhandle = xdr_fhandle(encode=>$opts{fhandle}, version=>$ver);
    my $sattr   = xdr_fattr4(encode=>$opts{attr});

    if ( $ver == 4 ) {
        # Create compound request
        my @ops;
	if ($self->{minor_version} == 0) {
           @ops = ( { op => OP_PUTFH,  args => $fhandle },
                    { op => OP_VERIFY, args => $sattr } );
        }
        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
				arg  => xdr_compound(tag => 'verify',
			    	minor => $minor_version,
				ops => \@ops),
				out  => \&xdr_compound_res);
    }

    # Check for errors
    $self->error_check(name=>'VERIFY', result=>$out);

    return $out;
} #end sub verify


1;

