# this is basically -*- perl -*-
# $Id: $
#
## @summary nfs::NfsClient::Linux - NFS Linux Client Class
## @author	gvg@netapp.com
## @status	shared
## @pod here

=head1 NAME

nfs::NfsClient::Linux - NFS Linux client protocol class

=head1 SYNOPSIS

  use nfs::NfsClient::Linux;

  $nfs_obj = nfs::NfsClient::Linux->new(hostname => $host, version  => 4,
                                 proto => 'tcp');
  @entries = $nfs_obj->ls( handle => $fhandle );
  $nfs_obj->cat( fhandle => $handle );

=head1 DESCRIPTION

Provides an object interface for the NFS Linux client. This is a subclass
of nfs::NfsClient and hence inherits all the functions defined in the
parent class.

=head1 STATUS

  Private (Not yet ready for others to use)

=head1 AUTHOR

  GV Govindasamy (gvg@netapp.com)
  $Id:  $
  Last revised $Date:  $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 BUGS

  * Need to improve error handling
  * Need to add more timeouts/retries specific to Linux
  * Functions starting with __ are not implemented yet.

=head1 ROUTINES

=cut

package nfs::NfsClient::Linux;
use strict;
use base qw(nfs::NfsClient);
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use nfs::Rpc;
use Carp qw(croak);
use nfs::NfsXdr qw(xdr_access xdr_bitmap4 xdr_cb_compound xdr_cb_exit
  xdr_cb_info xdr_cb_nlm_granted xdr_cb_nsm_notify xdr_cb_pid xdr_cb_port
  xdr_cb_register xdr_commit xdr_compound xdr_compound_res xdr_cookie
  xdr_cookieverf xdr_create xdr_daemon_info xdr_dump xdr_export xdr_fattr4
  xdr_fhandle xdr_fsinfo xdr_fsstat xdr_getattr xdr_link xdr_lookup xdr_mkdir
  xdr_mknod xdr_mount xdr_nfstime xdr_nlm_lock xdr_nlm_stat xdr_nlm_test
  xdr_opaque xdr_pathconf xdr_read xdr_readdir xdr_readdirplus xdr_readlink
  xdr_remove xdr_rename xdr_sattr xdr_setattr xdr_specdata xdr_stateid4
  xdr_symlink xdr_uint xdr_uint64 xdr_write);
use nfs::NfsDef qw(nfs_error nlm_error bitmap4_convert
  EXCLUSIVE FATTR4_FILEID FATTR4_FSID FATTR4_MODE FATTR4_NUMLINKS FATTR4_OWNER
  FATTR4_OWNER_GROUP FATTR4_RAWDEV FATTR4_SIZE FATTR4_TIME_ACCESS
  FATTR4_TIME_METADATA FATTR4_TIME_MODIFY FATTR4_TYPE FILE_SYNC
  MOUNTPROC_DUMP MOUNTPROC_EXPORT MOUNTPROC_MNT MOUNTPROC_PATHCONF
  MOUNTPROC_UMNT NFBLK NFCHR NFDIR NFFIFO NFLNK NFREG NFS4ERR_DELAY
  NFS4ERR_GRACE NFSOCK NFSPROC3_ACCESS NFSPROC3_COMMIT NFSPROC3_CREATE
  NFSPROC3_FSINFO NFSPROC3_FSSTAT NFSPROC3_LINK NFSPROC3_LOOKUP NFSPROC3_MKDIR
  NFSPROC3_MKNOD NFSPROC3_PATHCONF NFSPROC3_READDIRPLUS NFSPROC3_REMOVE
  NFSPROC3_RENAME NFSPROC3_RMDIR NFSPROC3_SYMLINK NFSPROC3_WRITE NFSPROC_CREATE
  NFSPROC_GETATTR NFSPROC_LINK NFSPROC_LOOKUP NFSPROC_MKDIR NFSPROC_READ
  NFSPROC_READDIR NFSPROC_READLINK NFSPROC_REMOVE NFSPROC_RENAME NFSPROC_RMDIR
  NFSPROC_SETATTR NFSPROC_STATFS NFSPROC_SYMLINK NFSPROC_WRITE NLM_CANCEL
  NLM_GRANTED NLM_LOCK NLM_TEST NLM_UNLOCK OP_ACCESS OP_CB_RECALL OP_CLOSE
  OP_COMMIT OP_CREATE OP_DELEGRETURN OP_GETATTR OP_GETFH OP_LINK OP_LOCK
  OP_LOCKU OP_LOOKUP OP_LOOKUPP OP_OPEN OP_OPEN_DOWNGRADE OP_PUTFH OP_PUTROOTFH
  OP_READ OP_READDIR OP_READLINK OP_REMOVE OP_RENAME OP_RENEW OP_SAVEFH
  OP_SETATTR OP_SETCLIENTID OP_SETCLIENTID_CONFIRM OP_VERIFY OP_WRITE
  OPEN4_SHARE_ACCESS_READ OPEN4_SHARE_DENY_NONE READ_LT READW_LT UNCHECKED
  WRITE_LT WRITEW_LT ACCESS_READ ACCESS_LOOKUP);
use TracePkg;
use Options;

my $Trace = new TracePkg;
my $Options = new Options;

$Options->parameters(definitions=>
		     [			  
		      {
			  name => "NFS_RETRIES",			      
			  spec => "s",
			  description => "Number of nfs retries by a real NFS Client",			      
			  default => 3,
		      },
		      {
			  name => "MNT_TCP_RETRIES",
			  spec => "s",
			  description => "Number of mount TCP retries by a real Linux Client",			      
			  default => 0,
		      },
		      {
			  name => "MNT_UDP_RETRIES",			      
			  spec => "s",
			  description => "Number of MNT UDP retries by a real NFS Client",			      
			  default => 6,
		      },
		      {
			  name => "MNT_TCP_TIMEOUT",			      
			  spec => "s",
			  description => "Number of MNT TCP timeout by a real NFS Client",			      
			  default => 20,
		      },
		      {
			  name => "MNT_UDP_TIMEOUT",			      
			  spec => "s",
			  description => "Number of MNT UDP timeout by a real NFS Client",			      
			  default => 3,
		      },
		      {
			  name => "RPC_RETRIES",			      
			  spec => "s",
			  description => "Number of RPC retries by a real NFS Client",			      
			  default => 6,
		      },
		      {
			  name => "RPC_TIMEOUT",			      
			  spec => "s",
			  description => "Number of RPC timeout by a real NFS Client",			      
			  default => 3,
		      },
		      ]
		     );    

sub new {
    my $class = shift;
    my %params = @_;
    $Options->process();
    my $opts = $Options->get_hash();

    $params{timeout} = $params{timeout} ? $params{timeout} : $opts->{RPC_TIMEOUT};
    $params{retry} = $params{retry} ? $params{retry} : $opts->{RPC_RETRIES};

    my $self = new nfs::NfsClient(%params);

    # Return NfsClient::Linux
    return bless $self, $class;
}


sub __mount($%;) {
    my ($self, %opts) = 
	validate_mixed(
		       @_,
		       [ { type => OBJECT } ],
		       {
			   path => { type => SCALAR },
			   pathlen => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
		       });
    $Trace->entry();

    my ($rpc, $out);
    my $ver = $self->{version};
    my %len = ( len => $opts{pathlen} ) if defined($opts{pathlen});
    my $opts = $Options->get_hash();

    if($opts->{PROTOCOL} =~ /tcp/) {
	$self->{retry} = $opts->{MNT_TCP_RETRIES};
	$self->{timeout} = $opts->{MNT_TCP_TIMEOUT};       
    } else {
	$self->{retry} = $opts->{MNT_UDP_RETRIES};
	$self->{timeout} = $opts->{MNT_UDP_TIMEOUT};
    }
    $self->{mnt_proto} = $opts->{PROTOCOL};
    $self->{mnt_ver} = $opts->{NFS_VERSION};

    if ( $ver == 4 ) {
        # Mount on NFSv4 is implemented on the NFS protocol
        # so use the current Rpc object
        $rpc = $self->{rpc};

        # Create a list of directories for the path to be mounted
        $opts{path} =~ s{/}{};
        my @paths = split /\//, $opts{path};

        # Create compound request:
        #   PUTROOTFH to get the filehandle of the pseudo-fs root
        my @ops = ( { op => OP_PUTROOTFH } );
        #   traverse the path using LOOKUP
        push @ops, map { { op   => OP_LOOKUP,
                           args => xdr_opaque(encode=>$_, %len) }
                       } @paths;
        #   and finally get the filehandle of the last LOOKUP operation
        push @ops, { op => OP_GETFH };

        # Call the compound procedure
        $out = $rpc->clnt_call(proc => 1,
                               arg  => xdr_compound(tag => 'mount',
                                                    ops => \@ops),
                               out  => \&xdr_compound_res);

        # Check for errors
        $self->error_check(name=>'MOUNT', result=>$out);

        # Save the root filehandle on the NfsClient object
        $self->{mnt_fh} = $out->{results}[-1]{result}{fhandle};

        # Set the client id and confirm it
        $self->setclientid();
        $self->setclientid_confirm();
        ++$self->{cl_verifier};
    } else {
	# Get Port for NFS
	# ??? -GV

	# Send nfs null request
	$self->nfs_null();

	# Mount on NFSv2 and v3 is implemented on the MOUNT protocol
	# so create a connection to it
	$self->{mnt_port} = 0; # Make Getport call prior to call other calls
	$self->{mount_rpc} = $self->mount_rpc_get();
		
        # Send Mount NULL request (Try all 3 mount versions starting from NFS_VERSION)
	my $mnt_versions = 3;
	my $error;
	while($mnt_versions) {
	    if ( ! $self->{mnt_port} ) {
		# Port is not defined, call portmapper to get the port number
		$self->{mnt_port} = 
		    $self->{mount_rpc}->pmap_getport(program => 100005,
					version => $self->{mnt_ver},
					proto   => $self->{mnt_proto});
		croak "Could not get port number for program 100005"
		    unless $self->{mnt_port};
	    }

	    eval { $self->mount_null(); };
	    $error = $@;
#	    last if ! $error;

	    # Error occured. Reset port number so GetPort will be called again
	    $self->{mnt_port} = 0;
	    if($self->{mnt_ver} == 3) {
		$self->{mnt_ver} = 1;
	    } else {
		$self->{mnt_ver} += 1;
	    }
	    $mnt_versions--;
	}
	croak $error if $error; # Report mount null failure, if any
		    
	$opts{mnt_ver} = $self->{mnt_ver};
	$out = $self->_mount(%opts);
    }

    # Set current filehandle
    $self->current_fh_set(fhandle=>$self->{mnt_fh});

    # Save mount path
    $self->{mnt_path} = $opts{path};

    $Trace->exit();

    return $out;
}

sub nfs_null {
    my $self = shift;
    my $rpc = $self->{rpc};
    $Trace->entry();
    # Burt 178530 null procedure does not return void
    # it returns a null string of 4 bytes
    eval { $rpc->clnt_call(proc => 0) };
    croak $@ if ($@ and $@ !~ /4 extra bytes/);
    $Trace->exit();
}

sub mount_null {
    my $self = shift;
    my $rpc = $self->{mount_rpc};

    $Trace->entry();
    # Burt 178530 null procedure does not return void
    # it returns a null string of 4 bytes
    eval { $rpc->clnt_call(proc => 0) };
    croak $@ if ($@ and $@ !~ /4 extra bytes/);
    $Trace->exit();
}

sub ___mount {
    my ($self, %opts) = 
	validate_mixed(
		       @_,
		       [ { type => OBJECT } ],
		       {
			   path => { type => SCALAR },
			   pathlen => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
			   mnt_ver => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
		       });
    my %len = ( len => $opts{pathlen} ) if defined($opts{pathlen});
    my $rpc = $self->{mount_rpc};
    my $out;

    $Trace->entry();

    # Call correct procedure on server
    $out = $rpc->clnt_call(proc => MOUNTPROC_MNT,
			   arg  => xdr_opaque(encode => $opts{path}, %len),
			   out  => sub { xdr_mount(version=>$opts{mnt_ver}, @_) });
    $self->{rpc}{xid} = $rpc->{xid};
    
    # Check for errors
    $self->error_check(name=>'MOUNT', result=>$out);
    
    # Save the root filehandle on the NfsClient object
    $self->{mnt_fh} = $out->{fhandle};
    
    return $out;
}


#-------------------------------------------------------------------

=head2 ls
    
    @entries = $nfs->ls( [handle => $handle] );

Lists and returns all directory entries on the filehandle. By
default, default root filehandle is listed.

The following named arguments are accepted:

=over

=item Arguments

=over

=item handle

(Optional) : Dir handle.

=back

=item Returns

All the file entries on the directory

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub ls {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
				handle => {type => SCALAR, default => $self->{curr_fh} },
			    }
			    );
    $Trace->entry();

    # GetAttr
    $self->getattr(fhandle => $params{handle});

    # Access
    my $results = $self->access(fhandle => $params{handle}, 
		  access => ACCESS_READ | ACCESS_LOOKUP );

    # Readdirplus
    my $dirs = $self->readdirplus(dhandle => $params{handle});

    my $entries = $dirs->{list};
    # Keep directoy handle in each element
    foreach (@{$entries}) {
        $_->{dhandle} = $params{handle};
    }

    $Trace->exit();
    return $entries;
}

#-------------------------------------------------------------------

=head2 cat

    $nfs->cat( name => "myfile", dhandle => $dhandle );

Read the entire file like unix 'cat <file> > /dev/null' does

The following named arguments are accepted:

=over

=item Arguments

=over

=item name

(Required) : Name of the file to be read.

=item dhandle

(Optional) : Dir handle where file exists.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub cat {
    my $self = shift;
    my (%params) = validate(
	      @_,
	      {
                  dhandle => {type => SCALAR, default => $self->{curr_fh}},
                  name    => {type => SCALAR},
                  offset  => {type => SCALAR, default => 0},
                  count   => {type => SCALAR, default => 1024},
              });
    $Trace->entry();

    my ($out, $fsize, $count, $fhandle);

    eval {
        # Lookup
        $out = $self->lookup(name    => $params{name},
                             dhandle => $params{dhandle});
        $fhandle = $out->{fhandle};
        
        # GetAttr
        $out = $self->getattr(fhandle => $fhandle);
        $fsize = $out->{fattr}->{size};
        
        # Access
        $out = $self->access(fhandle => $fhandle, 
                             access  => ACCESS_READ | ACCESS_LOOKUP );
        
        # Read
        my $offset = 0;
        my $length = ($fsize > 8192) ? 8192 : $fsize ;
        while ($fsize > 0) {
            $self->read(fhandle => $fhandle,
		    offset  => $offset,
                        count   => $length);
            $fsize =- $length;
            $offset =+ $length;
        }
    };
    if($@) {
        croak "nfs::NfsClient::Linux::cat() failed with \"$@\" for args\n".Dumper(\%params);
    }
    
    $Trace->exit();
}

#-------------------------------------------------------------------

=head2 fpolicy_read

    $nfs->fpolicy_read( name => "myfile", dhandle => $dhandle );

Read the first 1024 byts to make sure the file is not a stub file
but real file. FPolicy server creates the stubfile when real file
is migrated to secondary filer as part of NetApp Transparent 
Migration solution. If it is found a stub file, this function
warns with appropriate error message.

The following named arguments are accepted:

=over

=item Arguments

=over

=item name

(Required) : File name

=item dhandle

(Optional) : Directory handle where file exists.

=back

=item Returns

Nothing if successful. Else it warns with error message.

=back

=for html <hr />

=cut

#-------------------------------------------------------------------

sub fpolicy_read {
    my $self = shift;
    my (%params) = validate(
			    @_,
			    {
                                dhandle => {type => SCALAR, default => $self->{curr_fh}},
                                name    => {type => SCALAR},
                                offset  => {type => SCALAR, default => 0},
                                count   => {type => SCALAR, default => 1024},
			    }
			    );
    $Trace->entry();

    my ($out, $fsize, $count, $fhandle);

    # Lookup
    $out = $self->lookup(name    => $params{name},
                         dhandle => $params{dhandle});
    $fhandle = $out->{fhandle};

    # GetAttr
    $out = $self->getattr(fhandle => $fhandle);
    $fsize = $out->{fattr}->{size};
    
    # Access
    $out = $self->access(fhandle => $fhandle, 
                         access => ACCESS_READ | ACCESS_LOOKUP );
    
    # Read
    $count = ($fsize < $params{count}) ? $fsize : $params{count};
    $out = $self->read(fhandle => $fhandle,
                       offset  => $params{offset},
                       count   => $count);

    if($out->{data} =~ /migrtime=/i) {
        $Trace->warn(msg=>"Looks like stub file ".$out->{name}." was given by server. Data found was:\n ".$out->{data});
    }
    
    $Trace->exit();
}

1;
