# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsDef.pm#3 $
#
## @summary NfsDef - NFS protocol constant definitions
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

NfsDef - NFS protocol constant definitions

=head1 SYNOPSIS

  use nfs::NfsDef;

  $err_mesg = nfs_error(error => NFS4ERR_NOENT, version => 4);
  @attrlist = bitmap4_convert(attrmask => \@attrmask);
  @attrmask = bitmap4_convert(attrlist => \@attrlist);

=head1 DESCRIPTION

Provides NFS protocol constant definitions.
The constants are grouped under the following tags:

  NFSCONSTANTS
  NFSFTYPE
  NFSFSINFO
  NFSACCESS
  NFSCREATE
  NFSSTABLEHOW
  NFSSIZES
  NFSSATTR
  NFS4OPEN
  NFS4STATEID
  NFS4DELEG
  NFS4DCLAIMS
  NFS4LOCK
  FATTR4
  NFS2STAT
  NFS3STAT
  NFS4STAT
  NFS4OPS
  NFS4CBOPS
  NFS2PROCS
  NFS3PROCS
  MOUNTPROCS
  RPC_GSS
  ACLCONSTANTS
  WEBNFS
  NFSKERB
  WND4_TAGS

The best way to test this library is by running the tests in nfs/ops.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsDef.pm#3 $
  Last revised $Date: 2013/05/30 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsDef;

use strict;
use warnings;
use base qw(Exporter);

use Carp qw(croak);
use Params::Validate qw(validate SCALAR ARRAYREF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

# Variables used by nfs_error
my %nfs_error_h;
my %nlm_error_h;
my %nfs2_status_h;
my %nfs3_status_h;
my %nfs4_status_h;
#my %nfs4_1_status_h;
my %nlm_status_h;

# Variables used to save the constant names
# to be exported and grouped by tags
my @nfs_ftype;
my @nfs_fsinfo;
my @nfs_access;
my @nfs_create;
my @nfs_stable_how;
my @nfs_sizes;
my @nfs_sattr;
my @nfs4_open;
my @nfs4_stateid;
my @nfs4_deleg;
my @nfs4_deleg_claims;
my @nfs4_lock;
my @nfs4_1_exchange_id;
my @nfs4_1_state_protect;
my @nfs4_1_open_want_deleg;
my @nfs2_procs;
my @nfs2_stat;
my @nfs3_stat;
my @nfs4_stat;
my @nlm_stat;
my @nlm_procs;
my @nfs3_procs;
my @nfs4_ops;
my @nfs4_1_ops;
my @nfs4_cb_ops;
my @fattr4;
my @mount_procs;
my @rpc_gss;
my @acl_support;
my @acetype4;
my @aceflag4;
my @acemask4;
my @nfskerb;
my @webnfs;
my @wnd4_tags;

# Functions to be exported
my @functions = qw(
    nfs_error
    nlm_error
    bitmap4_convert
);

# ACL constants
my @acl = (
    @acl_support,
    @acetype4,
    @aceflag4,
    @acemask4,
);

# All constants
my @constants = (
    @nfs_ftype,
    @nfs_fsinfo,
    @nfs_access,
    @nfs_create,
    @nfs_stable_how,
    @nfs_sizes,
    @nfs_sattr,
    @nfs4_open,
    @nfs4_stateid,
    @nfs4_deleg,
    @nfs4_deleg_claims,
    @nfs4_lock,
    @nfs4_1_exchange_id,
    @nfs4_1_state_protect,
	@nfs4_1_open_want_deleg,
    @nfs2_procs,
    @nfs2_stat,
    @nfs3_stat,
    @nfs4_stat,
    #@nfs4_1_stat,
    @nlm_stat,
    @nlm_procs,
    @nfs3_procs,
    @nfs4_ops,
    @nfs4_1_ops,
    @nfs4_cb_ops,
    @fattr4,
	@nfskerb,
    @mount_procs,
    @rpc_gss,
    @webnfs,
    @acl,
    @wnd4_tags,
);

our @EXPORT_OK = (
    @functions,
    @constants
);

our %EXPORT_TAGS = (
    NFSCONSTANTS => \@constants,
    NFSFTYPE     => \@nfs_ftype,
    NFSFSINFO    => \@nfs_fsinfo,
    NFSACCESS    => \@nfs_access,
    NFSCREATE    => \@nfs_create,
    NFSSTABLEHOW => \@nfs_stable_how,
    NFSSIZES     => \@nfs_sizes,
    NFSSATTR     => \@nfs_sattr,
    NFS4OPEN     => \@nfs4_open,
    NFS4STATEID  => \@nfs4_stateid,
    NFS4DELEG    => \@nfs4_deleg,
    NFS4DCLAIMS  => \@nfs4_deleg_claims,
    NFS4LOCK     => \@nfs4_lock,
    NFS4EXCHANGE => \@nfs4_1_exchange_id,
    NFS4STATEPROTECT => \@nfs4_1_state_protect,
    FATTR4       => \@fattr4,
    NFS2STAT     => \@nfs2_stat,
    NFS3STAT     => \@nfs3_stat,
    NFS4STAT     => \@nfs4_stat,
    NLMSTAT      => \@nlm_stat,
    NFS4OPS      => \@nfs4_ops,
    NFS4_1OPS    => \@nfs4_1_ops,
    NFS4CBOPS    => \@nfs4_cb_ops,
    NFS2PROCS    => \@nfs2_procs,
    NFS3PROCS    => \@nfs3_procs,
    NLMPROCS     => \@nlm_procs,
    MOUNTPROCS   => \@mount_procs,
    NFSKERB		 => \@nfskerb,
    RPC_GSS      => \@rpc_gss,
    ACLCONSTANTS => \@acl,
    WEBNFS       => \@webnfs,
    WND4_TAGS    => \@wnd4_tags,
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub nfs_error(%);
sub nlm_error(%);
sub bitmap4_convert(%);

#---------------------------------------------------------------------

=head2 nfs_error

  $err_mesg = nfs_error(error => $error, version => $version);

Returns the NFS error code and message for a given error number.

The following named arguments are accepted:

=over

=item Arguments

=over

=item error

(Required) : Error number.

=item version

(Required) : NFS version.

=back

=item Returns

NFS error message

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nfs_error(%) {
    my (%opts) = validate (
        @_,
        {
            error   => { type => SCALAR, regex => qr/^\d+$/ },
            version => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );
    my $nfs_status_h;

    # Get correct hash of error codes
    # according to NFS version
    if ( $opts{version} == 2 ) {
        $nfs_status_h = \%nfs2_status_h;
    } elsif ( $opts{version} == 3 ) {
        $nfs_status_h = \%nfs3_status_h;
    } elsif ( $opts{version} == 4 ) {
        $nfs_status_h = \%nfs4_status_h;
    }

    # Return error message
    if ( !exists $nfs_status_h->{$opts{error}} ) {
        return $opts{error} . ': Unknown error code';
    }
    return $nfs_status_h->{$opts{error}} . ': ' . $nfs_error_h{$opts{error}};
}

#---------------------------------------------------------------------

=head2 nlm_error

  $err_mesg = nlm_error(error => $error);

Returns the NLM error code and message for a given error number.

The following named arguments are accepted:

=over

=item Arguments

=over

=item error

(Required) : Error number.

=back

=item Returns

NLM error message

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nlm_error(%) {
    my (%opts) = validate (
        @_,
        {
            error => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );

    # Return error message
    if ( !exists $nlm_status_h{$opts{error}} ) {
        return $opts{error} . ': Unknown error code';
    }
    return $nlm_status_h{$opts{error}} . ': ' . $nlm_error_h{$opts{error}};
}

#---------------------------------------------------------------------

=head2 bitmap4_convert

  @attrlist|@attrmask = bitmap4_convert(attrmask|attrlist => \@attrs);

Convert a NFSv4 attribute bitmap4 mask to and from a list of
attributes.
Both arguments are mutually exclusive.

The following named arguments are accepted:

=over

=item Arguments

=over

=item attrmask

(Optional) : Reference to an array of bitmap4 mask to be converted
into an array of attributes.

=item attrlist

(Optional) : Reference to an array of attributes to be converted
into an array of bitmap4 mask.

=back

=item Returns

A reference to an array of either a bitmap4 mask or attributes.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub bitmap4_convert(%) {
    my (%opts) = validate (
        @_,
        {
            attrmask => { type => ARRAYREF, optional => 1 },
            attrlist => { type => ARRAYREF, optional => 1 },
        }
    );
    if ( ! ($opts{attrmask} xor $opts{attrlist}) ) {
        croak "Options 'attrmask' and 'attrlist' are mutually exclusive " .
              "but at least one must be provided";
    }

    if ( $opts{attrmask} ) {
        # Convert from bitmap4 to list of attributes
        my @attrlist;
        my $n = 0;
        # Process all bitmap4 mask integers
        foreach ( @{$opts{attrmask}} ) {
            # Copy iterator since bitmap gets changed
            # and iterator is a reference to the actual
            # bitmap mask which must be kept intact
            my $bitmap = $_;
            if ( $bitmap ) {
                # For each integer bitmap mask,
                # go though each bit
                foreach my $i ( 0..31 ) {
                    if ( $bitmap & 1 ) {
                        # Save actual bit position
                        push @attrlist, $n + $i;
                    }
                    # Get next bit
                    $bitmap >>= 1;
                }
            }
            # Increment to next bitmap mask integer
            $n += 32;
        }

        # Return list of attributes
        return \@attrlist;
    } else {
        # Convert from list of attributes to bitmap4
        my (@attrmask, $index, $nshift);
        # Start with lowest bit first
        foreach my $bit ( sort {$a <=> $b} @{$opts{attrlist}} ) {
            $index  = int($bit / 32);
            $nshift = $bit % 32;
            $attrmask[$index] |= (1<<$nshift);
        }
        # Zero out all undefines on @attrmask
        (!defined($_)?$_=0:undef) foreach @attrmask;

        # Return bitmap4
        return \@attrmask;
    }
}

#---------------------------------------------------------------------
# Use the BEGIN block to create and define all constants
# so the names are available to be exported
#---------------------------------------------------------------------
BEGIN {
    # Create constants from the input hash
    # It only works for scalar values
    sub create_constants(\%) {
        my $const_h = shift;
        my @const_exp;
        @const_exp = keys %{$const_h};
        foreach my $name ( @const_exp ) {
            # A constant is nothing more than a subroutine definition
            # which returns a constant value -- perl inline this type
            # of subroutines.
            # It is done this way in order not to duplicate writing
            # constants' names all over the place: constant definition,
            # EXPORT_OK, and EXPORT_TAGS.
            eval "sub $name () {$const_h->{$name}}";
        }
        # Return array of constants' names
        return @const_exp;
    }

    # Constants for NFS file type
    my %nfs_ftype_h = (
        NFNON       => 0x00,
        NFREG       => 0x01,
        NFDIR       => 0x02,
        NFBLK       => 0x03,
        NFCHR       => 0x04,
        NFLNK       => 0x05,
        NFSOCK      => 0x06,
        NFFIFO      => 0x07,
        NFATTRDIR   => 0x08,
        NFNAMEDATTR => 0x09,
    );

    @nfs_ftype = create_constants(%nfs_ftype_h);

    # Constants for NFSv3 FSINFO
    my %nfs_fsinfo_h = (
        FSF3_LINK        => 0x01,
        FSF3_SYMLINK     => 0x02,
        FSF3_HOMOGENEOUS => 0x08,
        FSF3_CANSETTIME  => 0x10,
    );

    @nfs_fsinfo = create_constants(%nfs_fsinfo_h);

    # Constants for NFS ACCESS permission bit masks
    my %nfs_access_h = (
        ACCESS_READ    => 0x001,
        ACCESS_LOOKUP  => 0x002,
        ACCESS_MODIFY  => 0x004,
        ACCESS_EXTEND  => 0x008,
        ACCESS_DELETE  => 0x010,
        ACCESS_EXECUTE => 0x020,
    );

    @nfs_access = create_constants(%nfs_access_h);

    # Constants for NFS create flags (NFSv3 CREATE, NFSv4 OPEN)
    my %nfs_create_h = (
        UNCHECKED => 0,
        GUARDED   => 1,
        EXCLUSIVE => 2,
    );

    @nfs_create = create_constants(%nfs_create_h);

    # Constants for NFS stable_how flags
    my %nfs_stable_how_h = (
        UNSTABLE  => 0,
        DATA_SYNC => 1,
        FILE_SYNC => 2,
    );

    @nfs_stable_how = create_constants(%nfs_stable_how_h);

    # Constants for NFS sizes
    my %nfs_sizes_h = (
        MAXDATA     => 8192,
        MAXPATHLEN  => 1024,
        MAXNAMLEN   => 255,
    );

    @nfs_sizes = create_constants(%nfs_sizes_h);

    # Constants for NFSv3 and v4 sattr
    my %nfs_sattr_h = (
        SET_TO_SERVER_TIME => 1,
    );

    @nfs_sattr = create_constants(%nfs_sattr_h);

    # Constants for NFSv4 OPEN share access
    my %nfs4_open_share_h = (
        OPEN4_SHARE_ACCESS_READ  => 1,
        OPEN4_SHARE_ACCESS_WRITE => 2,
        OPEN4_SHARE_ACCESS_BOTH  => 3,
    );

    @nfs4_open = create_constants(%nfs4_open_share_h);

    # Constants for NFSv4 OPEN share deny
    my %nfs4_open_deny_h = (
        OPEN4_SHARE_DENY_NONE  => 0,
        OPEN4_SHARE_DENY_READ  => 1,
        OPEN4_SHARE_DENY_WRITE => 2,
        OPEN4_SHARE_DENY_BOTH  => 3,
    );

    push @nfs4_open, create_constants(%nfs4_open_deny_h);

    # Constants for NFSv4 special stateid of all 0's and all 1's
    my %nfs4_stateid_h = (
        STATEID0 => 0,
        STATEID1 => 1,
    );

    @nfs4_stateid = create_constants(%nfs4_stateid_h);

    # Constants for NFSv4 delegations
    my %nfs4_deleg_h = (
        OPEN_DELEGATE_NONE  => 0,
        OPEN_DELEGATE_READ  => 1,
        OPEN_DELEGATE_WRITE => 2,
        OPEN_DELEGATE_NONE_EXT => 3,
    );

    @nfs4_deleg = create_constants(%nfs4_deleg_h);

    # Constants for claim type of NFSv4 delegations
    my %nfs4_deleg_claims_h = (
        CLAIM_NULL          => 0,
        CLAIM_PREVIOUS      => 1,
        CLAIM_DELEGATE_CUR  => 2,
        CLAIM_DELEGATE_PREV => 3,
    );

    @nfs4_deleg_claims = create_constants(%nfs4_deleg_claims_h);    

    # Constants for NFSv4 lock types
    my %nfs4_lock_h = (
        READ_LT   => 1,
        WRITE_LT  => 2,
        READW_LT  => 3,  # Blocking read
        WRITEW_LT => 4,  # Blocking write
    );

    @nfs4_lock = create_constants(%nfs4_lock_h);

    # Constants for NFSv4.1 exchange id
    my %nfs4_1_exchange_id_h = (
	EXCHGID4_FLAG_SUPP_MOVED_REFER    => 0x00000001,
	EXCHGID4_FLAG_SUPP_MOVED_MIGR     => 0x00000002,
	EXCHGID4_FLAG_BIND_PRINC_STATEID  => 0x00000100,
	EXCHGID4_FLAG_USE_NON_PNFS        => 0x00010000,
	EXCHGID4_FLAG_USE_PNFS_MDS        => 0x00020000,
	EXCHGID4_FLAG_USE_PNFS_DS         => 0x00040000,
	EXCHGID4_FLAG_MASK_PNFS           => 0x00070000,
	EXCHGID4_FLAG_UPD_CONFIRMED_REC_A => 0x40000000,
	EXCHGID4_FLAG_CONFIRMED_R         => 0x8000000
    );

    @nfs4_1_exchange_id = create_constants(%nfs4_1_exchange_id_h);

    my %nfs4_1_state_protect_h = (
	SP4_NONE => 0,
	SP4_MACH_CRED => 1,
	SP4_SSV => 2
    );

    @nfs4_1_state_protect = create_constants(%nfs4_1_state_protect_h);

	# Constants for nfs4.1 open want delegation flags
	my %nfs4_1_open_want_deleg_h = (
		OPEN4_SHARE_ACCESS_WANT_DELEG_MASK		=> 0xFF00,
		OPEN4_SHARE_ACCESS_WANT_NO_PREFERENCE	=> 0x0000,
		OPEN4_SHARE_ACCESS_WANT_READ_DELEG		=> 0x0100,
		OPEN4_SHARE_ACCESS_WANT_WRITE_DELEG		=> 0x0200,
		OPEN4_SHARE_ACCESS_WANT_ANY_DELEG		=> 0x0300,
		OPEN4_SHARE_ACCESS_WANT_NO_DELEG		=> 0x0400,
		OPEN4_SHARE_ACCESS_WANT_CANCEL			=> 0x0500,
	);
    @nfs4_1_open_want_deleg = create_constants(%nfs4_1_open_want_deleg_h);

    # Constants for NFSv2 procedures
    my %nfs2_procs_h = (
        NFSPROC_NULL     => 0,
        NFSPROC_GETATTR  => 1,
        NFSPROC_SETATTR  => 2,
        NFSPROC_LOOKUP   => 4,
        NFSPROC_READLINK => 5,
        NFSPROC_READ     => 6,
        NFSPROC_WRITE    => 8,
        NFSPROC_CREATE   => 9,
        NFSPROC_REMOVE   => 10,
        NFSPROC_RENAME   => 11,
        NFSPROC_LINK     => 12,
        NFSPROC_SYMLINK  => 13,
        NFSPROC_MKDIR    => 14,
        NFSPROC_RMDIR    => 15,
        NFSPROC_READDIR  => 16,
        NFSPROC_STATFS   => 17,
    );

    @nfs2_procs = create_constants(%nfs2_procs_h);

    # Constants for NFSv2 error codes
    my %nfs2_stat_h = (
        NFS_OK             => 0,
        NFSERR_PERM        => 1,
        NFSERR_NOENT       => 2,
        NFSERR_IO          => 5,
        NFSERR_NXIO        => 6,
        NFSERR_ACCES       => 13,
        NFSERR_EXIST       => 17,
        NFSERR_NODEV       => 19,
        NFSERR_NOTDIR      => 20,
        NFSERR_ISDIR       => 21,
        NFSERR_FBIG        => 27,
        NFSERR_NOSPC       => 28,
        NFSERR_ROFS        => 30,
        NFSERR_NAMETOOLONG => 63,
        NFSERR_NOTEMPTY    => 66,
        NFSERR_DQUOT       => 69,
        NFSERR_STALE       => 70,
        NFSERR_WFLUSH      => 99,
    );

    %nfs2_status_h = reverse %nfs2_stat_h;
    @nfs2_stat = create_constants(%nfs2_stat_h);

    # Constants for NFSv3 error codes
    my %nfs3_stat_h = (
        NFS3_OK             => 0,
        NFS3ERR_PERM        => 1,
        NFS3ERR_NOENT       => 2,
        NFS3ERR_IO          => 5,
        NFS3ERR_NXIO        => 6,
        NFS3ERR_ACCES       => 13,
        NFS3ERR_EXIST       => 17,
        NFS3ERR_XDEV        => 18,
        NFS3ERR_NODEV       => 19,
        NFS3ERR_NOTDIR      => 20,
        NFS3ERR_ISDIR       => 21,
        NFS3ERR_INVAL       => 22,
        NFS3ERR_FBIG        => 27,
        NFS3ERR_NOSPC       => 28,
        NFS3ERR_ROFS        => 30,
        NFS3ERR_MLINK       => 31,
        NFS3ERR_NAMETOOLONG => 63,
        NFS3ERR_NOTEMPTY    => 66,
        NFS3ERR_DQUOT       => 69,
        NFS3ERR_STALE       => 70,
        NFS3ERR_REMOTE      => 71,
        NFS3ERR_BADHANDLE   => 10001,
        NFS3ERR_NOT_SYNC    => 10002,
        NFS3ERR_BAD_COOKIE  => 10003,
        NFS3ERR_NOTSUPP     => 10004,
        NFS3ERR_TOOSMALL    => 10005,
        NFS3ERR_SERVERFAULT => 10006,
        NFS3ERR_BADTYPE     => 10007,
        NFS3ERR_JUKEBOX     => 10008,
    );

    %nfs3_status_h = reverse %nfs3_stat_h;
    @nfs3_stat = create_constants(%nfs3_stat_h);

    # Constants for NFSv4 error codes
    my %nfs4_stat_h = (
        NFS4_OK                     => 0,
        NFS4ERR_PERM                => 1,
        NFS4ERR_NOENT               => 2,
        NFS4ERR_IO                  => 5,
        NFS4ERR_NXIO                => 6,
        NFS4ERR_ACCESS              => 13,
        NFS4ERR_EXIST               => 17,
        NFS4ERR_XDEV                => 18,
        NFS4ERR_NOTDIR              => 20,
        NFS4ERR_ISDIR               => 21,
        NFS4ERR_INVAL               => 22,
        NFS4ERR_FBIG                => 27,
        NFS4ERR_NOSPC               => 28,
        NFS4ERR_ROFS                => 30,
        NFS4ERR_MLINK               => 31,
        NFS4ERR_NAMETOOLONG         => 63,
        NFS4ERR_NOTEMPTY            => 66,
        NFS4ERR_DQUOT               => 69,
        NFS4ERR_STALE               => 70,
        NFS4ERR_BADHANDLE           => 10001,
        NFS4ERR_BAD_COOKIE          => 10003,
        NFS4ERR_NOTSUPP             => 10004,
        NFS4ERR_TOOSMALL            => 10005,
        NFS4ERR_SERVERFAULT         => 10006,
        NFS4ERR_BADTYPE             => 10007,
        NFS4ERR_DELAY               => 10008,
        NFS4ERR_SAME                => 10009,
        NFS4ERR_DENIED              => 10010,
        NFS4ERR_EXPIRED             => 10011,
        NFS4ERR_LOCKED              => 10012,
        NFS4ERR_GRACE               => 10013,
        NFS4ERR_FHEXPIRED           => 10014,
        NFS4ERR_SHARE_DENIED        => 10015,
        NFS4ERR_WRONGSEC            => 10016,
        NFS4ERR_CLID_INUSE          => 10017,
        NFS4ERR_RESOURCE            => 10018,
        NFS4ERR_MOVED               => 10019,
        NFS4ERR_NOFILEHANDLE        => 10020,
        NFS4ERR_MINOR_VERS_MISMATCH => 10021,
        NFS4ERR_STALE_CLIENTID      => 10022,
        NFS4ERR_STALE_STATEID       => 10023,
        NFS4ERR_OLD_STATEID         => 10024,
        NFS4ERR_BAD_STATEID         => 10025,
        NFS4ERR_BAD_SEQID           => 10026,
        NFS4ERR_NOT_SAME            => 10027,
        NFS4ERR_LOCK_RANGE          => 10028,
        NFS4ERR_SYMLINK             => 10029,
        NFS4ERR_RESTOREFH           => 10030,
        NFS4ERR_LEASE_MOVED         => 10031,
        NFS4ERR_ATTRNOTSUPP         => 10032,
        NFS4ERR_NO_GRACE            => 10033,
        NFS4ERR_RECLAIM_BAD         => 10034,
        NFS4ERR_RECLAIM_CONFLICT    => 10035,
        NFS4ERR_BADXDR              => 10036,
        NFS4ERR_LOCKS_HELD          => 10037,
        NFS4ERR_OPENMODE            => 10038,
        NFS4ERR_BADOWNER            => 10039,
        NFS4ERR_BADCHAR             => 10040,
        NFS4ERR_BADNAME             => 10041,
        NFS4ERR_BAD_RANGE           => 10042,
        NFS4ERR_LOCK_NOTSUPP        => 10043,
        NFS4ERR_OP_ILLEGAL          => 10044,
        NFS4ERR_DEADLOCK            => 10045,
        NFS4ERR_FILE_OPEN           => 10046,
        NFS4ERR_ADMIN_REVOKED       => 10047,
        NFS4ERR_CB_PATH_DOWN        => 10048,
        NFS4ERR_BACK_CHAN_BUSY      => 10057,
        NFS4ERR_BADIOMODE           => 10049,
        NFS4ERR_BADLAYOUT           => 10050,
        NFS4ERR_BADSESSION          => 10052,
        NFS4ERR_BADSLOT             => 10053,
        NFS4ERR_BAD_HIGH_SLOT       => 10077,
        NFS4ERR_BAD_SESSION_DIGEST  => 10051,
        NFS4ERR_CLIENTID_BUSY       => 10074,
        NFS4ERR_COMPLETE_ALREADY    => 10054,
        NFS4ERR_CONN_NOT_BOUND_TO_SESSION  => 10055,
        NFS4ERR_DEADSESSION         => 10078,
        NFS4ERR_DELEG_ALREADY_WANTED       => 10056,
        NFS4ERR_DELEG_REVOKED       => 10087,
        NFS4ERR_DIRDELEG_UNAVAIL    => 10084,
        NFS4ERR_ENCR_ALG_UNSUPP     => 10079,
        NFS4ERR_HASH_ALG_UNSUPP     => 10072,
        NFS4ERR_LAYOUTTRYLATER      => 10058,
        NFS4ERR_LAYOUTUNAVAILABLE   => 10059,
        NFS4ERR_NOMATCHING_LAYOUT   => 10060,
        NFS4ERR_NOT_ONLY_OP         => 10081,
        NFS4ERR_OP_NOT_IN_SESSION   => 10071,
        NFS4ERR_PNFS_IO_HOLE        => 10075,
        NFS4ERR_PNFS_NO_LAYOUT      => 10080,
        NFS4ERR_RECALLCONFLICT      => 10061,
        NFS4ERR_REJECT_DELEG        => 10085,
        NFS4ERR_REP_TOO_BIG         => 10066,
        NFS4ERR_REP_TOO_BIG_TO_CACHE => 10067,
        NFS4ERR_REQ_TOO_BIG         => 10065,
        NFS4ERR_RESOURCE            => 10018,
        NFS4ERR_RETRY_UNCACHED_REP  => 10068,
        NFS4ERR_RETURNCONFLICT      => 10086,
        NFS4ERR_SEQUENCE_POS        => 10064,
        NFS4ERR_SEQ_FALSE_RETRY     => 10076,
        NFS4ERR_SEQ_MISORDERED      => 10063,
        NFS4ERR_TOO_MANY_OPS        => 10070,
        NFS4ERR_UNKNOWN_LAYOUTTYPE  => 10062,
        NFS4ERR_UNSAFE_COMPOUND     => 10069,
        NFS4ERR_WRONG_CRED          => 10082,
        NFS4ERR_WRONG_TYPE          => 10083,


    );

     #%nfs4_1_status_h = (
     #   NFS4ERR_BACK_CHAN_BUSY              =>10057
     #   NFS4ERR_BADIOMODE                   =>10049
     #   NFS4ERR_BADLAYOUT                   =>10050
     #   NFS4ERR_BADSESSION                  =>10052
     #   NFS4ERR_BADSLOT                     =>10053
     #   NFS4ERR_BAD_HIGH_SLOT               =>10077
     #   NFS4ERR_BAD_SESSION_DIGEST          =>10051
     #   NFS4ERR_CLIENTID_BUSY               =>10074
     #   NFS4ERR_COMPLETE_ALREADY            =>10054
     #   NFS4ERR_CONN_NOT_BOUND_TO_SESSION   =>10055
     #   NFS4ERR_DEADSESSION                 =>10078
     #   NFS4ERR_DELEG_ALREADY_WANTED        =>10056
     #   NFS4ERR_DELEG_REVOKED               =>10087
     #   NFS4ERR_DIRDELEG_UNAVAIL            =>10084
     #   NFS4ERR_ENCR_ALG_UNSUPP             =>10079
     #   NFS4ERR_HASH_ALG_UNSUPP             =>10072
     #   NFS4ERR_LAYOUTTRYLATER              =>10058
     #   NFS4ERR_LAYOUTUNAVAILABLE           =>10059
     #   NFS4ERR_NOMATCHING_LAYOUT           =>10060
     #   NFS4ERR_NOT_ONLY_OP                 =>10081
     #   NFS4ERR_OP_NOT_IN_SESSION           =>10071
     #   NFS4ERR_PNFS_IO_HOLE                =>10075
     #   NFS4ERR_PNFS_NO_LAYOUT              =>10080
     #   NFS4ERR_RECALLCONFLICT              =>10061
     #   NFS4ERR_REJECT_DELEG                =>10085
     #   NFS4ERR_REP_TOO_BIG                 =>10066
     #   NFS4ERR_REP_TOO_BIG_TO_CACHE        =>10067
     #   NFS4ERR_REQ_TOO_BIG                 =>10065
     #   NFS4ERR_RESOURCE                    =>10018
     #   NFS4ERR_RETRY_UNCACHED_REP          =>10068
     #   NFS4ERR_RETURNCONFLICT              =>10086
     #   NFS4ERR_SEQUENCE_POS                =>10064
     #   NFS4ERR_SEQ_FALSE_RETRY             =>10076
     #   NFS4ERR_SEQ_MISORDERED              =>10063
     #   NFS4ERR_TOO_MANY_OPS                =>10070
     #   NFS4ERR_UNKNOWN_LAYOUTTYPE          =>10062
     #   NFS4ERR_UNSAFE_COMPOUND             =>10069
     #   NFS4ERR_WRONG_CRED                  =>10082
     #   NFS4ERR_WRONG_TYPE                  =>10083
    #`);
#
    %nfs4_status_h = reverse %nfs4_stat_h;
    @nfs4_stat = create_constants(%nfs4_stat_h);
    #%nfs4_1_status_h = reverse %nfs4_1_stat_h;
    #@nfs4_1_stat = create_constants(%nfs4_1_stat_h);


    # Constants for NFSv4 and NFSv4.1 error codes
    %nfs_error_h = (
            0 => 'everything is okay',
            1 => 'caller not privileged',
            2 => 'no such file/directory',
            5 => 'hard I/O error',
            6 => 'no such device',
           13 => 'access denied',
           17 => 'file already exists',
           18 => 'different filesystems',
           19 => 'no such device',
           20 => 'should be a directory',
           21 => 'should not be directory',
           22 => 'invalid argument',
           27 => 'file exceeds server max',
           28 => 'no space on filesystem',
           30 => 'read-only filesystem',
           31 => 'too many hard links',
           63 => 'name exceeds server max',
           66 => 'directory not empty',
           69 => 'hard quota limit reached',
           70 => 'file no longer exists',
           71 => 'filehandle not server local',
	    10001 => 'illegal filehandle',
        10002 => 'setattr update problem',
        10003 => 'READDIR cookie is stale',
        10004 => 'operation not supported',
        10005 => 'response limit exceeded',
        10006 => 'undefined server error',
        10007 => 'invalid filetype',
        10008 => 'file "busy" - retry',
        10009 => 'nverify says attrs same',
        10010 => 'lock unavailable',
        10011 => 'lock lease expired',
        10012 => 'I/O failed due to lock',
        10013 => 'in grace period',
        10014 => 'filehandle expired',
        10015 => 'share reserve denied',
        10016 => 'wrong security flavor',
        10017 => 'clientid in use',
        10018 => 'resource exhaustion',
        10019 => 'filesystem relocated',
        10020 => 'current FH is not set',
        10021 => 'minor vers not supp',
        10022 => 'server has rebooted',
        10023 => 'server has rebooted',
        10024 => 'state is out of sync',
        10025 => 'incorrect stateid',
        10026 => 'request is out of seq.',
        10027 => 'verify - attrs not same',
        10028 => 'lock range not supported',
        10029 => 'should be file/directory',
        10030 => 'no saved filehandle',
        10031 => 'some filesystem moved',
        10032 => 'recommended attr not sup',
        10033 => 'reclaim outside of grace',
        10034 => 'reclaim error at server',
        10035 => 'conflict on reclaim',
        10036 => 'XDR decode failed',
        10037 => 'file locks held at CLOSE',
        10038 => 'conflict in OPEN and I/O',
        10039 => 'owner translation bad',
        10040 => 'utf-8 char not supported',
        10041 => 'name not supported',
        10042 => 'lock range not supported',
        10043 => 'no atomic up/downgrade',
        10044 => 'undefined operation',
        10045 => 'file locking deadlock',
        10046 => 'open file blocks op.',
        10047 => 'lockowner state revoked',
        #  and NFSv4.1
        10048 => 'callback path down',
        10049 => 'An invalid or inappropriate layout iomode was specified',
        10050 => 'The layout specified is invalid in some way',
        10051 => 'The digest used in a SET_SSV request is not valid.',
        10052 => 'The specified session ID is unknown to the server',
        10053 => 'It is possible the slot may have been retired.',
        10054 => 'An additional RECLAIM_COMPLETE operation is not necessary',
        10055 => 'Sequence operation was sent on a connection that has not been associated with the specified session',
        10056 => 'The client has requested a delegation when it had already registered that it wants that same delegation',
        10057 => 'backchannel is busy, server has callback requests outstanding',
        10058 => 'Layouts are temporarily unavailable for the file',
        10059 => 'layouts are not available for the current file system or the particular specified file',
        10060 => 'the client has no layouts matching the specification of the layouts being recalled.',
        10061 => 'a conflicting recall operation for an object, is currently in progress.',
        10062 => 'The client has specified a layout type which is not supported by the server.',
        10063 => 'The requester sent a Sequence operation with an invalid sequence ID.',
        10064 => 'A Sequence operation appeared in a position other than the first operation of a Compound request.',
        10065 => 'The Compound request exceeds the channel\'s negotiated maximum size for requests',
        10067 => 'maximum size for replies cached in the reply cache has been exceeded',
        10068 => 'The requester has attempted a retry of a Compound which it previously requested not be placed in the reply cache.',
        10069 => 'The client has sent a COMPOUND request with an unsafe mix of operations',
        10070 => 'The Compound request has too many operations',
        10071 => 'The operation is outside the context of session',
        10072 => 'the set of hashing algorithms presented by the client did not include any supported by the server',
        10074 => 'there are sessions and/or unexpired state associated with the client ID ',
        10075 => 'attempted to read from or write to an illegal hole of a file',
        10076 => 'retried request is not the same as the original request',
        10077 => 'The highest_slot argument in a Sequence operation exceeds the enforced highest_slotid.',
        10078 => 'specified session is a persistent session which is dead and does not accept new requests',
        10079 => 'the set of encryption algorithms presented by the client didnot include any supported by the server',
        10080 => 'attempt to read from or write to a file without holding a valid layout.',
        10081 => 'The required OP is required to be the only op in the entire compound request',
        10082 => 'principal that is not allowed to modify the given state',
        10083 => 'the object is of an invalid type for the current operation',
        10084 => 'the server is unable or unwilling to provide a requested directory delegation.',
        10085 => 'The callback operation invoked to deal with a new delegation has rejected it.',
        10086 => 'layout is unavailable due to an attempt to perform the LAYOUTGET',
        10087 => 'the stateid has been revoked due to the failure of the client to return the lock when it was recalled',

    );

    # Constants for NLM error codes
    my %nlm_stat_h = (
        NLM_OK                  => 0,
        NLM_DENIED              => 1,
        NLM_DENIED_NOLOCKS      => 2,
        NLM_BLOCKED             => 3,
        NLM_DENIED_GRACE_PERIOD => 4,
        NLM_DEADLCK             => 5,
        NLM_ROFS                => 6,
        NLM_STALE_FH            => 7,
        NLM_FBIG                => 8,
        NLM_FAILED              => 9,
    );

    %nlm_status_h = reverse %nlm_stat_h;
    @nlm_stat = create_constants(%nlm_stat_h);

    # Constants for NLM error codes
    %nlm_error_h = (
        0 => 'lock is granted',
        1 => 'lock is denied',
        2 => 'lack of resources',
        3 => 'lock is blocked',
        4 => 'in grace period',
        5 => 'deadlock',
        6 => 'read-only filesystem',
        7 => 'file no longer exists',
        8 => 'offset or length exceeds server max',
        9 => 'undefined server error',
    );

    # Constants for NFSv3 procedures
    my %nfs3_procs_h = (
        NFSPROC3_NULL        => 0,
        NFSPROC3_GETATTR     => 1,
        NFSPROC3_SETATTR     => 2,
        NFSPROC3_LOOKUP      => 3,
        NFSPROC3_ACCESS      => 4,
        NFSPROC3_READLINK    => 5,
        NFSPROC3_READ        => 6,
        NFSPROC3_WRITE       => 7,
        NFSPROC3_CREATE      => 8,
        NFSPROC3_MKDIR       => 9,
        NFSPROC3_SYMLINK     => 10,
        NFSPROC3_MKNOD       => 11,
        NFSPROC3_REMOVE      => 12,
        NFSPROC3_RMDIR       => 13,
        NFSPROC3_RENAME      => 14,
        NFSPROC3_LINK        => 15,
        NFSPROC3_READDIR     => 16,
        NFSPROC3_READDIRPLUS => 17,
        NFSPROC3_FSSTAT      => 18,
        NFSPROC3_FSINFO      => 19,
        NFSPROC3_PATHCONF    => 20,
        NFSPROC3_COMMIT      => 21,
    );

    @nfs3_procs = create_constants(%nfs3_procs_h);

    # Constants for NLM procedures
    my %nlm_procs_h = (
        NLM_TEST    => 1,
        NLM_LOCK    => 2,
        NLM_CANCEL  => 3,
        NLM_UNLOCK  => 4,
        NLM_GRANTED => 5,
    );

    @nlm_procs = create_constants(%nlm_procs_h);

    # Constants for NFSv4 operations
    my %nfs4_ops_h = (
        OP_ACCESS              => 3,
        OP_CLOSE               => 4,
        OP_COMMIT              => 5,
        OP_CREATE              => 6,
        OP_DELEGPURGE          => 7,
        OP_DELEGRETURN         => 8,
        OP_GETATTR             => 9,
        OP_GETFH               => 10,
        OP_LINK                => 11,
        OP_LOCK                => 12,
        OP_LOCKT               => 13,
        OP_LOCKU               => 14,
        OP_LOOKUP              => 15,
        OP_LOOKUPP             => 16,
        OP_NVERIFY             => 17,
        OP_OPEN                => 18,
        OP_OPENATTR            => 19,
        OP_OPEN_CONFIRM        => 20,
        OP_OPEN_DOWNGRADE      => 21,
        OP_PUTFH               => 22,
        OP_PUTPUBFH            => 23,
        OP_PUTROOTFH           => 24,
        OP_READ                => 25,
        OP_READDIR             => 26,
        OP_READLINK            => 27,
        OP_REMOVE              => 28,
        OP_RENAME              => 29,
        OP_RENEW               => 30,
        OP_RESTOREFH           => 31,
        OP_SAVEFH              => 32,
        OP_SECINFO             => 33,
        OP_SETATTR             => 34,
        OP_SETCLIENTID         => 35,
        OP_SETCLIENTID_CONFIRM => 36,
        OP_VERIFY              => 37,
        OP_WRITE               => 38,
        OP_RELEASE_LOCKOWNER   => 39,
        OP_COPY_NOTIFY         => 59,
        OP_COPY_REVOKE         => 60,
        OP_COPY                => 61,
        OP_COPY_ABORT          => 62,
        OP_COPY_STATUS         => 63,
        OP_ILLEGAL             => 10044
    );

    my %nfs4_1_ops_h = (
        OP_BACKCHANNEL_CTL     => 40,
        OP_BIND_CONN_TO_SESSION=> 41,
        OP_EXCHANGE_ID         => 42,
        OP_CREATE_SESSION      => 43,
        OP_DESTROY_SESSION     => 44,
        OP_FREE_STATEID        => 45,
        OP_GET_DIR_DELEGATION  => 46,
        OP_GETDEVICEINFO       => 47,
        OP_GETDEVICELIST       => 48,
        OP_LAYOUTCOMMIT        => 49,
        OP_LAYOUTGET           => 50,
        OP_LAYOUTRETURN        => 51,
        OP_SECINFO_NO_NAME     => 52,
        OP_SEQUENCE            => 53,
        OP_SET_SSV             => 54,
        OP_TEST_STATEID        => 55,
        OP_WANT_DELEGATION     => 56,
        OP_DESTROY_CLIENTID    => 57,
        OP_RECLAIM_COMPLETE    => 58,
    );


    @nfs4_ops = create_constants(%nfs4_ops_h);
    @nfs4_1_ops = create_constants(%nfs4_1_ops_h);

    # Constants for NFSv4 callback operations
    my %nfs4_cb_ops_h = (
        OP_CB_GETATTR => 3,
        OP_CB_RECALL  => 4,
        OP_CB_LAYOUTRECALL => 5,
        OP_CB_NOTIFY => 6,
        OP_CB_PUSH_DELEG => 7,
        OP_CB_RECALL_ANY => 8,
        OP_CB_RECALLABLE_OBJ_AVAIL => 9,
        OP_CB_RECALL_SLOT => 10,
        OP_CB_SEQUENCE  => 11,
        OP_CB_WANTS_CANCELLED => 12,
        OP_CB_NOTIFY_LOCK => 13,
        OP_CB_NOTIFY_DEVICEID => 14,
    );

    @nfs4_cb_ops = create_constants(%nfs4_cb_ops_h);

    # Constants for NFSv4 FATTR4 bit mapping
    my %fattr4_h = (
        FATTR4_SUPPORTED_ATTRS    => 0,
        FATTR4_TYPE               => 1,
        FATTR4_FH_EXPIRE_TYPE     => 2,
        FATTR4_CHANGE             => 3,
        FATTR4_SIZE               => 4,
        FATTR4_LINK_SUPPORT       => 5,
        FATTR4_SYMLINK_SUPPORT    => 6,
        FATTR4_NAMED_ATTR         => 7,
        FATTR4_FSID               => 8,
        FATTR4_UNIQUE_HANDLES     => 9,
        FATTR4_LEASE_TIME         => 10,
        FATTR4_RDATTR_ERROR       => 11,
        FATTR4_ACL                => 12,
        FATTR4_ACLSUPPORT         => 13,
        FATTR4_ARCHIVE            => 14,
        FATTR4_CANSETTIME         => 15,
        FATTR4_CASE_INSENSITIVE   => 16,
        FATTR4_CASE_PRESERVING    => 17,
        FATTR4_CHOWN_RESTRICTED   => 18,
        FATTR4_FILEHANDLE         => 19,
        FATTR4_FILEID             => 20,
        FATTR4_FILES_AVAIL        => 21,
        FATTR4_FILES_FREE         => 22,
        FATTR4_FILES_TOTAL        => 23,
        FATTR4_FS_LOCATIONS       => 24,
        FATTR4_HIDDEN             => 25,
        FATTR4_HOMOGENEOUS        => 26,
        FATTR4_MAXFILESIZE        => 27,
        FATTR4_MAXLINK            => 28,
        FATTR4_MAXNAME            => 29,
        FATTR4_MAXREAD            => 30,
        FATTR4_MAXWRITE           => 31,
        FATTR4_MIMETYPE           => 32,
        FATTR4_MODE               => 33,
        FATTR4_NO_TRUNC           => 34,
        FATTR4_NUMLINKS           => 35,
        FATTR4_OWNER              => 36,
        FATTR4_OWNER_GROUP        => 37,
        FATTR4_QUOTA_AVAIL_HARD   => 38,
        FATTR4_QUOTA_AVAIL_SOFT   => 39,
        FATTR4_QUOTA_USED         => 40,
        FATTR4_RAWDEV             => 41,
        FATTR4_SPACE_AVAIL        => 42,
        FATTR4_SPACE_FREE         => 43,
        FATTR4_SPACE_TOTAL        => 44,
        FATTR4_SPACE_USED         => 45,
        FATTR4_SYSTEM             => 46,
        FATTR4_TIME_ACCESS        => 47,
        FATTR4_TIME_ACCESS_SET    => 48,
        FATTR4_TIME_BACKUP        => 49,
        FATTR4_TIME_CREATE        => 50,
        FATTR4_TIME_DELTA         => 51,
        FATTR4_TIME_METADATA      => 52,
        FATTR4_TIME_MODIFY        => 53,
        FATTR4_TIME_MODIFY_SET    => 54,
        FATTR4_MOUNTED_ON_FILEID  => 55,
	FATTR4_FS_STATUS          => 61,
        FATTR4_LAYOUT_HINT        => 63,
	FATTR4_LAYOUT_TYPE        => 64,
	FATTR4_LAYOUT_BLKSIZE     => 65,
        FATTR4_LAYOUT_ALIGNMENT   => 66,
	FATTR4_FS_LOCATIONS_INFO  => 67,
	FATTR4_LAYOUT_MDSTHRESHOLD=> 68,
        FATTR4_SUPPATTR_EXCLCREAT => 75,
	# Keep this for now, change later
        FATTR4_SPACE_RESERVE      => 77,
        FATTR4_SPACE_FREED        => 78,
    );

    @fattr4 = create_constants(%fattr4_h);

    # Constants for MOUNT procedures
    my %mount_procs_h = (
        MOUNTPROC_NULL     => 0,
        MOUNTPROC_MNT      => 1,
        MOUNTPROC_DUMP     => 2,
        MOUNTPROC_UMNT     => 3,
        MOUNTPROC_EXPORT   => 5,
        MOUNTPROC_PATHCONF => 7,
    );

    @mount_procs = create_constants(%mount_procs_h);

    # Constants that indicate which Access Control Entries are supported
    my %acl_support_h = (
        ACL4_SUPPORT_ALLOW_ACL => 0x00000001,
        ACL4_SUPPORT_DENY_ACL  => 0x00000002,
        ACL4_SUPPORT_AUDIT_ACL => 0x00000004,
        ACL4_SUPPORT_ALARM_ACL => 0x00000008,
    );

    @acl_support = create_constants(%acl_support_h);

    # Constants for acetype4
    my %acetype4_h = (
        ACE4_ACCESS_ALLOWED_ACE_TYPE => 0x00000000,
        ACE4_ACCESS_DENIED_ACE_TYPE  => 0x00000001,
        ACE4_SYSTEM_AUDIT_ACE_TYPE   => 0x00000002,
        ACE4_SYSTEM_ALARM_ACE_TYPE   => 0x00000003,
    );

    @acetype4 = create_constants(%acetype4_h);

    # Constants for ACE flag
    my %aceflag4_h = (
        ACE4_FILE_INHERIT_ACE           => 0x00000001,
        ACE4_DIRECTORY_INHERIT_ACE      => 0x00000002,
        ACE4_NO_PROPAGATE_INHERIT_ACE   => 0x00000004,
        ACE4_INHERIT_ONLY_ACE           => 0x00000008,
        ACE4_SUCCESSFUL_ACCESS_ACE_FLAG => 0x00000010,
        ACE4_FAILED_ACCESS_ACE_FLAG     => 0x00000020,
        ACE4_IDENTIFIER_GROUP           => 0x00000040,
    );

    @aceflag4 = create_constants(%aceflag4_h);

    # Constants for ACE mask
    my %acemask4_h = (
        ACE4_READ_DATA         => 0x00000001,
        ACE4_LIST_DIRECTORY    => 0x00000001,
        ACE4_WRITE_DATA        => 0x00000002,
        ACE4_ADD_FILE          => 0x00000002,
        ACE4_APPEND_DATA       => 0x00000004,
        ACE4_ADD_SUBDIRECTORY  => 0x00000004,
        ACE4_READ_NAMED_ATTRS  => 0x00000008,
        ACE4_WRITE_NAMED_ATTRS => 0x00000010,
        ACE4_EXECUTE           => 0x00000020,
        ACE4_DELETE_CHILD      => 0x00000040,
        ACE4_READ_ATTRIBUTES   => 0x00000080,
        ACE4_WRITE_ATTRIBUTES  => 0x00000100,
        ACE4_DELETE            => 0x00010000,
        ACE4_READ_ACL          => 0x00020000,
        ACE4_WRITE_ACL         => 0x00040000,
        ACE4_WRITE_OWNER       => 0x00080000,
        ACE4_SYNCHRONIZE       => 0x00100000,

        # ACE4_GENERIC_READ -- defined as combination of
        #      ACE4_READ_ACL |
        #      ACE4_READ_DATA |
        #      ACE4_READ_ATTRIBUTES |
        #      ACE4_SYNCHRONIZE
        ACE4_GENERIC_READ => 0x00120081,

        # ACE4_GENERIC_WRITE -- defined as combination of
        #      ACE4_READ_ACL |
        #      ACE4_WRITE_DATA |
        #      ACE4_WRITE_ATTRIBUTES |
        #      ACE4_WRITE_ACL |
        #      ACE4_APPEND_DATA |
        #      ACE4_SYNCHRONIZE
        ACE4_GENERIC_WRITE => 0x00160106,

        # ACE4_GENERIC_EXECUTE -- defined as combination of
        #      ACE4_READ_ACL
        #      ACE4_READ_ATTRIBUTES
        #      ACE4_EXECUTE
        #      ACE4_SYNCHRONIZE
        ACE4_GENERIC_EXECUTE => 0x001200A0,
    );

    @acemask4 = create_constants(%acemask4_h);

    # Constants for NFSv2 and v3 WebNFS filehandles
    my $fhandle = pack('N', 0)x8;
    my %webnfs_h = (
        WEBNFS2_FHANDLE => "'$fhandle'",
        WEBNFS3_FHANDLE => "''",
    );

    @webnfs = create_constants(%webnfs_h);

	# Constants for RPC_GSS (Generic Security Services)
	my %rpc_gss_h = (
		RPC_GSS_SVC_NONE		=> 1,
		RPC_GSS_SVC_INTEGRITY	=> 2,
		RPC_GSS_SVC_PRIVACY		=> 3,
		RPCSEC_GSS				=> 6,
		RPCSEC_GSS_DATA			=> 0,
		RPCSEC_GSS_INIT			=> 1,
		RPCSEC_GSS_CONTINUE_INIT=> 2,
		RPCSEC_GSS_DESTROY		=> 3,
	);
	@nfskerb = create_constants(%rpc_gss_h);

    my %wndtags_h = (
        WND4_NOT_WANTED         => 0,
        WND4_CONTENTION         => 1,
        WND4_RESOURCE           => 2,
        WND4_NOT_SUPP_FTYPE     => 3,
        WND4_WRITE_DELEG_NOT_SUPP_FTYPE => 4,
        WND4_NOT_SUPP_UPGRADE   => 5,
        WND4_NOT_SUPP_DOWNGRADE => 6,
        WND4_CANCELLED          => 7,
        WND4_IS_DIR             => 8
    );

    @wnd4_tags = create_constants(%wndtags_h);
};

1;

