#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsOps.pm#2 $
#
## @summary    nfsops_v2_v3_v4 overlap test routines
## @author    dl-maryv (FSS Protocols QA)
## @status    public
## @notes    Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

NfsOps - nfsops_v2_v3_v4 overlap test routines

=head1 SYNOPSIS

 use nfs::NfsOps;

 $out = set_env( setup_h  => \%setup_h,
                allowed  => [ 3, 4 ],
                version  => 4,
                protocol => 'tcp' );

 clean_env( %setup_h );

 ($fc, $volume) = nfsops_vol_create( vol_name => 'temp_volume',
                                    vol_size => '20m' );

 nfsops_vol_purge( fc    => $fc,
                    volume => $volume );

 object_remove( hostname => $fc->hostp()->hostname(),
                volume   => $volume,
                path    => $files[0] );

 @files = file_make( hostname => $fc->hostp()->hostname(),
                    volume   => $volume,
                    name    => $file_name,
                    count    => 1,
                    fill    => 'A' );

 @dirs = dir_make( hostname => $fc->hostp()->hostname(),
                    volume   => $volume,
                    name    => $dir_name,
                    count    => 4 );

 @links = link_make( hostname => $fc->hostp()->hostname(),
                    volume   => $volume,
                    name    => $link_name,
                    target   => $target_name,
                    count    => 1 );

 @symlinks = symlink_make( hostname => $fc->hostp()->hostname(),
                            volume   => $volume,
                            name    => $symlink_name,
                            target   => $target_name,
                            count    => 1 );

 @devices = device_make( hostname => $fc->hostp()->hostname(),
                        volume   => $volume,
                        name    => $device_name,
                        type    => $dev,
                        count    => 1 ) );

 file_write( hostname => $fc->hostp()->hostname(),
            volume   => $volume,
            path    => $files[0],
            offset   => 0,
            data    => $data,
            stable   => FILE_SYNC,
            stateid  => $stateid );

 $valid = scalar_validate( op        => 'LOOKUP',
                            version  => $ver,
                            protocol => $proto,
                            expected => $expected,
                            returned => $returned );

 $valid = hash_validate( op        => $op,
                        version  => $ver,
                        protocol => $proto,
                        expected => \%expected,
                        returned => \%returned );

 $valid = hash_validate_change( op        => 'REMOVE',
                                version  => $ver,
                                protocol => $proto,
                                initial  => \%initial,
                                current  => \%current );

 nfsops_lookup_path( client => $client_obj,
                    path   => $files[0] );

 %attrs = nfsops_common_attrs( version  => $ver,
                                returned => \%returned );

 %previous = nfsops_save_common_attrs( attr_h => \%attrs );

 my $exp = nfsops_access_bits( bitmask => $expected );

 %stats = nfsops_fsstats( fc    => $fc,
                        volume => $volume );

 %devices = nfsops_set_devices ( device => $device );

 @ids = nfsops_get_new_id( fc   => $fc,
                            type => 'uid' );

=head1 DESCRIPTION

Provides generic functionality for use in nfsops_v2_v3_v4 overlap testing.

A setup routine is provided which establishes the workspace on the filer
and creates the NFS client objects which will be used during the test. A
corresponding cleanup routine restores the filer workspace and umounts the
NFS client objects.

A set of routines is provided for creating filesystem objects that may be
needed during the test. Validation routines compare results of an NFS call
with a set of expected results and report any discrepancies. Utility
routines provide common operations that may be shared among several tests,
such as looking up a path, converting results to a more readable common
format, or gathering information directly from the filer.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/NfsOps.pm#2 $
  Last revised $Date: 2013/07/30 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsOps;

use strict;
use warnings;
use base qw(Exporter);
use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use Cwd qw(getcwd);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use filer qw(filer_options_get filer_options_set);
use FS::Util::Setup qw(create_vol common_cleanup);
use Volume qw(vol_purge vol_destroy);
use Network::Utils qw(run_getXXbyYY);
use FileManip qw(remove_file);
use SayWrap qw(sayEOF);
use GenericMount qw(mount umount);
use nfs::GenericExport qw(unexport_path);
use nfs::Utils qw(validate_nfs_version get_good_nfs_version get_good_auth);
use nfs::NfsClient;
use nfs::NfsDef qw(FATTR4_FILEID FATTR4_FSID FATTR4_MODE FATTR4_NUMLINKS
    FATTR4_OWNER FATTR4_OWNER_GROUP FATTR4_RAWDEV FATTR4_SIZE
    FATTR4_TIME_ACCESS FATTR4_TIME_MODIFY FATTR4_TYPE FATTR4_TIME_METADATA
    FILE_SYNC NFBLK NFCHR NFDIR NFLNK UNCHECKED ACCESS_READ ACCESS_LOOKUP
    ACCESS_MODIFY ACCESS_EXTEND ACCESS_DELETE ACCESS_EXECUTE FATTR4_TIME_ACCESS_SET 
    FATTR4_TIME_MODIFY_SET);
use nfs::CompoundRequests qw(validate_minor_ver);
use Data::Dumper;

my $ng_check = 0;
my $f = host ($main::FILER);
if ($f->hosttype() =~ /filer-ng/) {
  use nfs_setup_gx qw (get_data_ip_cmode);
  $ng_check = 1;
}

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                        chomp(@msgs);
                        Carp::confess( Tharn::script_result( @msgs ) ) }
    );
};

our @EXPORT_OK = qw( set_env
                    clean_env
                    nfsops_vol_create
                    nfsops_vol_purge
                    object_remove
                    file_make
                    dir_make
                    link_make
                    symlink_make
                    device_make
                    file_write
                    scalar_validate
                    hash_validate
                    hash_validate_change
                    nfsops_lookup_path
                    nfsops_common_attrs
                    nfsops_save_common_attrs
                    nfsops_access_bits
                    nfsops_fsstats
                    nfsops_set_devices
                    get_gx_v4_id_domain
                    diskFree_gx
                    nfsops_get_new_id
                    testme
                    create_nfs_client);

#------------------------------------------------------------------------
# Version for package
#------------------------------------------------------------------------
our $VERSION = '0.01';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
my $nfs_v4_enable;
my $nfs_v4_acl;
my $nfs_v4_rd;
my $nfs_v4_wd;
my $nfs_mount_rootonly;
my $create_vol_called = 0;
my $map = {FATTR4_MODE() => 'mode',
           FATTR4_TIME_ACCESS_SET() => 'atime',
           FATTR4_TIME_MODIFY_SET() => 'mtime'};

#------------------------------------------------------------------------
# Subroutine prototypes
#------------------------------------------------------------------------
sub set_env(%);
sub clean_env(%);
sub nfsops_vol_create(%);
sub nfsops_vol_purge(%);
sub object_remove(%);
sub file_make(%);
sub dir_make(%);
sub link_make(%);
sub symlink_make(%);
sub device_make(%);
sub file_write(%);
sub scalar_validate(%);
sub hash_validate(%);
sub hash_validate_change(%);
sub nfsops_lookup_path(%);
sub nfsops_common_attrs(%);
sub nfsops_save_common_attrs(%);
sub nfsops_access_bits(%);
sub nfsops_fsstats(%);
sub nfsops_set_devices(%);
sub nfsops_get_new_id(%);
sub diskFree_gx(%);
sub testme();
sub create_nfs_client(%);

#------------------------------------------------------------------------

=head2 set_env

  $out = set_env( setup_h    => \%setup_h
                [, allowed  => \@allowed ]
                [, version  => $version ]
                [, protocol => 'udp'|'tcp'|'both' ]
                [, timeout  => $timeout ]
                [, minor_version => $minor_version]);

NfsOps overlap test setup routine. This routine establishes the environment
for the nfsops tests. It creates and exports the test volume if indicated.
It saves NFSv4 state from the filer and enables it if it is not. An array
of NFS client objects is created, one for each of the tested version/protocol
pairs, and the objects are mounted from the filer.

All setup state is passed back to the test on the 'setup_h' named parameter.
This hash is populated with the following information:
    volume        => the test volume that will be used
    fc            => the connection object to the filer
    nfs_client_objs => an array of client objects that are created
                        ( one for each version/protocol combination )

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item allowed

(Optional) : An array reference indicating what NFS version options are
allowed for this test. An error occurs if the version parameter does not
match one of these values. Default is 'all' versions.

=item version

(Optional) : NFS version(s) that will be tested. Acceptable values are
2, 3, 4, 'not2', 'not3', 'not4', and 'all'. Default is 'all' which will
test versions 2, 3, and 4.

=item protocol

(Optional) : The NFS connection protocol to be tested: 'udp', 'tcp', or
'both'. Default is 'both'.

=item timeout

(Optional) : Timeout value for connection (default is 30).

=back

=item Returns

1 if success, otherwise croaks with an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub set_env ( % ) {
    my %opts = validate(
        @_,
        {
            setup_h  => { type => HASHREF },
            allowed  => { type => ARRAYREF, default => [ 2, 3, 4 ] },
            version  => { type => SCALAR,   default => 'all' },
            protocol => { type => SCALAR,   default => 'both',
                                            regex   => qr/^(both|tcp|udp)$/ },
            timeout  => { type => SCALAR,   default => 30 },
            minor_ver   => { type => SCALAR, optional => 1, default => 0, regex => qr/^(both|[01])$/ },
        }
    );
    my $set_h = $opts{setup_h};
    my ( $vol_name, $hostname, $ver, @nfs_client_objs, @vers, @proto, @minor_ver, %mode );

    logStep('BEGIN: set_env: setup environment for test');

    logStep('set_env: Validate that host is unix');
    my $h = host( 'localhost' );
    croak "Client must be a unix host" if ( $h->hosttype() !~ /^unix/ );

    logStep('set_env: Validate allowed NFS versions');
    @vers = validate_nfs_version( version => $opts{version},
                                allowed => $opts{allowed} );
    @proto = ($opts{protocol} eq 'both') ? ('tcp', 'udp') : ($opts{protocol});
    @minor_ver= validate_minor_ver (minor_ver => $opts{minor_ver});
    logStep("minor_versions opt = " . (Dumper($opts{minor_ver})));
    logStep("minor_versions = " . (Dumper(@minor_ver)));

    logStep('set_env: Validate NFS version/protocol compatibility');
    croak "Invalid NFS version/protocol combination: " .
            "v$opts{version}, $opts{protocol}"
            if ( $opts{version} eq 4 && $opts{protocol} eq 'udp' );

    if ($ng_check == 0) {
        # Common setup -- connects to the filer, then creates and exports
        # a volume depending on the $main::SETUP_DONE parameter
        logStep('set_env: Create test volume');
        $create_vol_called = 1;
        ($set_h->{fc}, $vol_name) = create_vol() or croak $@;
        $hostname = $set_h->{fc}->hostp()->hostname();
    
        # Full path for volume
        $set_h->{volume} = "/vol/$vol_name";
    
        # Save NFSv4 state
        logStep('set_env: Save NFSv4 state');
        $nfs_v4_enable = filer_options_get( $set_h->{fc},
                                            'nfs.v4.enable' );
        $nfs_v4_acl    = filer_options_get( $set_h->{fc},
                                            'nfs.v4.acl.enable' );
        $nfs_v4_rd    = filer_options_get( $set_h->{fc},
                                            'nfs.v4.read_delegation' );
        $nfs_v4_wd    = filer_options_get( $set_h->{fc},
                                            'nfs.v4.write_delegation' );
    
        logStep('set_env: Enable NFSv4 functionality');
        filer_options_set( $set_h->{fc}, 'nfs.v4.enable', 'on' )
                if ( $nfs_v4_enable eq 'off' );
        filer_options_set( $set_h->{fc}, 'nfs.v4.acl.enable', 'on' )
                if ( $nfs_v4_acl eq 'off' );
        filer_options_set( $set_h->{fc}, 'nfs.v4.read_delegation', 'on' )
                if ( $nfs_v4_rd eq 'off' );
        filer_options_set( $set_h->{fc}, 'nfs.v4.write_delegation', 'on' )
                if ( $nfs_v4_wd eq 'off' );
    
        # Save reserved port state
        logStep('set_env: Save root only mount state');
        $nfs_mount_rootonly = filer_options_get( $set_h->{fc},
                                                'nfs.mount_rootonly' );
    
        logStep('set_env: Enable mounting from non-reserved ports');
        filer_options_set( $set_h->{fc}, 'nfs.mount_rootonly', 'off' )
                if ( $nfs_mount_rootonly eq 'on' );

    } else {
    
        logStep ('set_env: C-Mode specific setup');
        my $ref1 = nfs_setup_gx::set_env_gx( setup_h  => $set_h,
                allowed  => [ 2, 3, 4 ],
                version  => $opts{version},
                minor_version => $opts{minor_ver},
                protocol => $opts{protocol} );
        #creates user and group with uid=0 and gid=0 this is needed by getattr scripts
        my $server_obj = new Server (node=> $main::FILER);
    
        eval {
            $server_obj->vserver_services_unix_group_create(
                            vserver => $main::VSERVER_GX_NAME,
                            name => "root",
                            id    => "0"
                        );
        };
        if ($@ && $@ !~ /duplicate entry/)
        {
            logStep("unable to create user and group root with uid 0 and gid 0");
        }
    
        eval {
            $server_obj->vserver_services_unix_user_create(
                        vserver => $main::VSERVER_GX_NAME,
                        user    => "root",
                        id    => "0",
                        'primary-gid' => "0",
                        );
        };
        if ($@ && $@ !~ /duplicate entry/)
        {
            logStep("unable to create user and group root with uid 0 and gid 0");
        }
        $hostname = $ref1->{hostname};
        $set_h->{hostname} = $hostname;
        host($main::FILER);
    
        $set_h->{fc} = connect($main::FILER_CONN);
        # Full path for volume
        $set_h->{volume} = $ref1->{volume};
    
    }

    # Create array of NFS clients
    logStep('set_env: Create and mount NFS clients');
    foreach my $proto ( @proto ) {
        foreach my $vers ( @vers ) {
            next if ( $vers == 4 && $proto eq 'udp' );
            if ( $vers != 4) {
                push @nfs_client_objs,
                    nfs::NfsClient->new( hostname => $hostname,
                                        version  => $vers,
                                        proto    => $proto,
										sec		=> $main::AUTH,
                                        timeout  => $opts{timeout} );
            } elsif ( $vers == 4) {
                foreach my $minor_ver (@minor_ver) {
                    push @nfs_client_objs,
                        nfs::NfsClient->new( hostname => $hostname,
                                            version  => $vers,
                                            proto    => $proto,
											sec		=> $main::AUTH,
                                            minor_version  => $minor_ver,
                                            timeout  => $opts{timeout} );
                }

            }

        }
    }
    # Do not continue if list of NFS client objects is empty
    croak 'NFS client(s) not created'
            if ( $#nfs_client_objs == -1 );
    $set_h->{nfs_client_objs} = \@nfs_client_objs;

    # Mount the NFS objects
    foreach my $nfs_obj (@nfs_client_objs) {
        $nfs_obj->mount( path => $set_h->{volume} );
    }

    # Ensure volume is writable (for object creation)
    logStep('set_env: Setattr mode=0777 on test volume');
    if ( $nfs_client_objs[0]->version() == 4 ) {
        $mode{FATTR4_MODE()} = 0777;
    } else {
        $mode{mode} = 0777;
    }
    $nfs_client_objs[0]->rpc()->auth_sys_set( uid => 0, gid => 0 );
    $nfs_client_objs[0]->setattr( fhandle => $nfs_client_objs[0]->mnt_fh(),
                                attr    => \%mode );
    $nfs_client_objs[0]->rpc()->auth_correct_set();

    logStep('END: set_env');
    return 1;
}

#------------------------------------------------------------------------

=head2 clean_env

  clean_env( [ volume        => $volume ],
            [ fc            => $fc ],
            [ nfs_client_objs => \@nfs_client_objs ],
            [ files            => \@files ],
            [ dirs            => \@dirs ] );

nfsops overlap test cleanup routine. This routine restores the state of
the filer to the pre-test condition. Filer options for NFSv4 and for
root mount state are restored. Files and directories created by the test
are removed. NFS client objects that were mounted are unmounted.

The following named arguments are accepted:

=over

=item Arguments

=over

=item volume

(Optional) : The volume on the filer where the nfsops tests are performed.

=item fc

(Optional) : The filer connection object.

=item nfs_client_objs

(Optional) : An array of NFS client objects that have been mounted.

=item files

(Optional) : An array of files that have been created in the volume
directory. These files will be removed.

=item dirs

(Optional) : An array of directories that have been created in the volume
directory. These directories and any subdirectories/files will be removed.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub clean_env ( % ) {
    my %opts = validate(
        @_,
        {
            hostname        => { type => SCALAR | UNDEF, optional => 1},
            volume        => { type => SCALAR | UNDEF,   optional => 1 },
            vol_name    => { type => SCALAR | UNDEF,   optional => 1 },
            fc            => { type => OBJECT | UNDEF,   optional => 1 },
            nfs_client_objs => { type => ARRAYREF | UNDEF, optional => 1 },
            files            => { type => ARRAYREF | UNDEF, optional => 1 },
            dirs            => { type => ARRAYREF | UNDEF, optional => 1 },
  vol_name                => { type => SCALAR | UNDEF, optional => 1 },
        }
    );
    my ( $h, $curr_dir, $mountpoint );

    logStep("CLEANUP == $main::CLEANUP");

    $h = host($main::FILER);

    eval {
        eval {
            # Don't depend on fc parameter received
            logStep('clean_env: Re-connect to the filer');
            $h = host( $main::FILER );
            $opts{fc} = connect( $main::FILER_CONN, -timeout, $main::TIMEOUT );
            $opts{fc}->timeout( $main::CMDTIMEOUT );
        };
        if ($@) {
            logWarn("clean_env: $@");
            undef $opts{fc};
        }

        if ( $main::CLEANUP ne 'none' ) {
            logStep('clean_env: Umount the NFS client objects');
            if ( defined( $opts{nfs_client_objs} ) ) {
                foreach my $nfs_obj ( @{$opts{nfs_client_objs}} ) {
                    eval { $nfs_obj->umount( path => $opts{volume} ) };
                }
            }

            if ( defined( $opts{fc} ) && defined( $opts{volume} ) &&
                    ( ( defined( $opts{files} ) && $#{$opts{files}} > -1 ) ||
                    ( defined( $opts{dirs} ) && $#{$opts{dirs}} > -1 ) ) ) {
                logStep('clean_env: Mounting Volume');
#TODO:Use GenericMount.pm when available

my $mounthost1;
my $mounthost;
if ( $main::ADDR_TYPE eq 'inet6' ){				
$mounthost1 = ($h->hosttype !~ /filer-ng/) ? $opts{fc}->connhost() : get_data_ip_cmode($main::FILER);
                                $mounthost = "["."$mounthost1"."]";
} else {
 $mounthost = ($h->hosttype !~ /filer-ng/) ? $opts{fc}->connhost() : get_data_ip_cmode($main::FILER); 

}
                my $good_version = get_good_nfs_version();


                $mountpoint = mount( $mounthost, $opts{volume}, mntopts => "vers=$good_version" ) or
                        logWarn("clean_env: mount failed: $@");
            }

            if ( defined( $mountpoint ) ) {
                $curr_dir = getcwd();
                if ( chdir $mountpoint == 0 ) {
                    logWarn("clean_env: chdir failed: $!");
                    umount( $mountpoint ) or
                            logWarn("clean_env: umount failed: $@");
                    undef $mountpoint;
                }
            }

            # Remove files that were created
            if ( defined( $mountpoint ) && defined( $opts{files} ) &&
                    $#{$opts{files}} > -1 ) {

                # The unlink would fail if somebody accidentally stuck
                # "/" as a file because it is a directory so there is
                # no need to look for it on the list.
                logStep('clean_env: Removing files');
                unlink @{$opts{files}} or
                    logWarn("clean_env: unlink failed: $!");
                @{$opts{files}} = ();
            }

            # Remove directories that were created
            if ( defined( $mountpoint ) && defined( $opts{dirs} ) &&
                $#{$opts{dirs}} > -1 ) {

                logStep('clean_env: Removing directories');
                # Traverse array in reverse order so that the directories
                # are removed newest to oldest.
                my $dir;
                while ( $dir = pop @{$opts{dirs}} ) {
                    # Remove entire directory tree
                    chmod 0777, $dir;
                    remove_file( file    => "$mountpoint/$dir",
                                recursive => 1,
                                client    => 'localhost' );
                }
                @{$opts{dirs}} = ();
            }

            if ( defined( $mountpoint ) ) {
                chdir $curr_dir or
                    logWarn("clean_env: chdir $curr_dir failed: $!");
                umount( $mountpoint ) or
                    logWarn("clean_env: umount failed: $@");
            }

            if ($h->hosttype !~ /filer-ng/) {
            logStep('clean_env: Restore NFSv4 and root only mount states');
            if ( defined( $opts{fc} ) ) {
                eval {
                    filer_options_set($opts{fc}, 'nfs.v4.enable',
                                        $nfs_v4_enable) if $nfs_v4_enable;
                };
                logWarn("clean_env: $@") if $@;
                eval {
                    filer_options_set( $opts{fc}, 'nfs.v4.acl.enable',
                                        $nfs_v4_acl ) if $nfs_v4_acl;
                };
                logWarn("clean_env: $@") if $@;
                eval {
                    filer_options_set( $opts{fc}, 'nfs.v4.read_delegation',
                                        $nfs_v4_rd ) if $nfs_v4_rd;
                };
                logWarn("clean_env: $@") if $@;
                eval {
                    filer_options_set( $opts{fc}, 'nfs.v4.write_delegation',
                                        $nfs_v4_wd ) if $nfs_v4_wd;
                };
                logWarn("clean_env: $@") if $@;
                eval {
                    filer_options_set( $opts{fc}, 'nfs.mount_rootonly',
                                        $nfs_mount_rootonly )
                                        if $nfs_mount_rootonly;
                };
                logWarn("clean_env: $@") if $@;
            }
    }
        }

        # Call common cleanup if create_vol was called
        if ($h->hosttype !~ /filer-ng/) {
        if ( defined( $opts{fc} ) && $create_vol_called != 0 ) {
            logStep("Run common cleanup");
            common_cleanup( $opts{fc} );
            logWarn("clean_env: $@") if $@;
        }

        if ( $main::CLEANUP ne 'none' && $opts{volume} ) {
            # Unexport the volume
            logStep("Unexport volume");
            unexport_path( $opts{fc}, $opts{volume} ) or croak $@;
        }
    }
    };
    logWarn("clean_env: $@") if $@;

    logStep('END: clean_env');
    return;
}

#------------------------------------------------------------------------

=head2 nfsops_vol_create

  ($fc, $volume) = nfsops_vol_create( vol_name => $vol_name,
                                    vol_size => $vol_size );

Creates a volume with the name and size indicated by the input parameters.
All other volume creation parameters are the default values for the
create_vol function.

Permission on the volume is set to 0777 to allow all operations on the
volume.

The following named arguments are accepted:

=over

=item Arguments

=over

=item vol_name

(Required) : The name of the volume to be created.

=item vol_size

(Required) : The size of the volume to be created. Must conform to
'vol create' limits on the filer.

=back

=item Returns

An array containing the filer connection object and the volume created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_vol_create ( % ) {
    my %opts = validate(
        @_,
        {
            vol_name => { type => SCALAR },
            vol_size => { type => SCALAR },
        }
    );
    my ( $fc, $volume, $nfs_obj, $fh_mnt, $hostname1 );

    logDbg('Create the temporary volume');
    my $f=host($main::FILER);
    if ($f->hosttype()!~/filer-ng/) {

    ($fc, $volume) = create_vol( filer_vol    => $opts{vol_name},
                                vol_size    => $opts{vol_size},
                                extern_setup => 0 ) or croak $@;

    $volume = "/vol/$volume";
    $hostname1 = $fc->hostp()->hostname();
    } else {
        my $server_obj=new Server(node=>$main::FILER);
        eval {
            $server_obj->volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$opts{vol_name});
        };
        if ($@ && $@ =~ /no entries/){
            $server_obj->volume_create(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{vol_name}, aggregate=>$main::FILER_GX_AGGR,
                                size=> $opts{vol_size}, policy=>$main::EXPORT_GX_POL, 'junction-path'=> "/$opts{vol_name}");
        };
        $volume="/$opts{vol_name}";
        $hostname1 = get_data_ip_cmode($main::FILER);
        host($main::FILER);
        $fc = connect($main::FILER_CONN);
    }


    logDbg('Set the permission on the temporary volume to allow writing');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname =>$hostname1,
                                    version  => $version,
                                    proto   => 'tcp' );
    $fh_mnt = $nfs_obj->mount( path => $volume );
    $nfs_obj->rpc()->auth_sys_set( uid => 0,
                                    gid => 0 );
    if($version =~ /^4/) {
        $nfs_obj->setattr( fhandle => $nfs_obj->mnt_fh(),
                            attr    => { FATTR4_MODE() => 0777 } );
    } else {
        $nfs_obj->setattr( fhandle => $nfs_obj->mnt_fh(),
                            attr    => { mode => 0777 } );
    }
    $nfs_obj->rpc()->auth_correct_set();

    return ($fc, $volume);
}

#------------------------------------------------------------------------

=head2 nfsops_vol_purge

  nfsops_vol_purge( fc    => $conn,
                    volume => $junction_path
                    vol_name => $volume_name );
A simple wrapper to the Volume::vol_purge routine.

For some reason, the vol_purge was not destroying the volume in tests,
so Volume::vol_destroy is called as well.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : The filer connection object.

=item volume

(Required) : The volume name that is to be destroyed.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_vol_purge ( % ) {
    my %opts = validate(
                @_,
                {
                        fc      => { type => OBJECT },
                        volume => { type => SCALAR },
                        vol_name => { type => SCALAR | UNDEF, optional => 1}
                }
            );

	if (!$opts{vol_name}) {
		$opts{vol_name} = $opts{volume};
	}

    my $f=host($main::FILER);
    if ($f->hosttype() =~ /filer-ng/){
            if($opts{volume} =~ m!/(.*)!) {
                    $opts{volume} = $1;
            }
            my $server_obj=new Server(node=>$main::FILER);
            eval {
                    $server_obj->volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
            };
            if(!$@)
            {
                    $server_obj->volume_unmount(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
                    $server_obj->volume_offline(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
                    $server_obj->volume_destroy(vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
            }
    } else {
        vol_purge( $opts{fc}, $opts{volume} );

        # For some reason, vol_purge does not destroy the volume. Do it now.
        vol_destroy( $opts{fc}, $opts{volume} );
    }
    return;
}


#------------------------------------------------------------------------

=head2 object_remove

  object_remove( hostname => $hostname,
                volume   => $volume,
                path    => $path );

Removes an object from the specified host on the specified volume,
Uses NFS v4 protocol for removal. Establishes a new client object
mounts the volume, looks up the directory path, and calls REMOVE
on the object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer from which the object
will be removed.

=item volume

(Required) : The volume on the filer where the object resides.

=item path

(Required) : The path including the object name where the object
resides on the filer, relative to the mount point.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub object_remove ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            path    => { type => SCALAR },
        }
    );
    my ( $nfs_obj, $fh_mnt, $object, $name, @path, $rmv_failed );
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    logDbg('BEGIN: object_remove');

    # Get rid of contiguous '/'s and split path into segments
    $opts{path} =~ s{/+}{/};
    @path = split /\//, $opts{path};
    # Get the name of the object and remove it from the path
    $object = pop( @path );

    logDbg('object_remove: Create and mount v4 nfs_obj for object removal');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                    version  => $version,
                                         sec => get_good_auth());

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    logDbg('object_remove: Get the filehandle of the parent directory');
    foreach $name (@path) {
        $nfs_obj->lookup( name => $name );
    }

    eval {
        logDbg('object_remove: Remove the object');
        $nfs_obj->remove( name => $object );
    };

    logDbg('object_remove: Get the filehandle of the parent directory');
    foreach $name (@path) {
        $nfs_obj->lookup( name => $name );
    }

    eval {
        $nfs_obj->rmdir( name => $object );
    };


    logDbg('object_remove: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: object_remove');
    croak "object_remove: $rmv_failed" if $rmv_failed;
    return;
}

#------------------------------------------------------------------------

=head2 file_make

  @files = file_make( hostname => $hostname,
                    volume   => $volume
                    [, path  => $path ]
                    [, name  => $name ]
                    [, count => $count ]
                    [, attr  => \%attr ]
                    [, fill  => 'A' ] );

Creates one or more files on the specified host and volume.
Uses NFS v4 protocol for creating the files. Establishes and
mounts a new NFS client object, looks up the filehandle of the
directory in which the new files will be created, and calls
OPEN to make the new file objects.

Files will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number beginning with 1. Defaults to 'file'.

=item count

(Optional) : The number of objects to be created. Default is 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item fill

(Optional) : A single character to be used to fill the created file(s).
A GETATTR call will return the size of the file, and the character will
be written repetitively to fill the file to the size specified in the GETATTR.

=back

=item Returns

An array containing the file names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            path    => { type => SCALAR,  optional => 1 },
            name    => { type => SCALAR,  default  => 'file' },
            count    => { type => SCALAR,  default  => 1,
                                            regex    => qr/^[1-9]\d*$/ },
            attr    => { type => HASHREF, default  => {} },
            fill    => { type => SCALAR,  optional => 1, regex => qr/^.$/ },
        }
    );
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    my ( $nfs_obj, $timeout, $fh_mnt, $curr_fh, @files );

    logDbg('BEGIN: file_make');

    # Allow a longer timeout when using client object to fill a file
    $timeout = ( $opts{fill} ) ? 300 : 30;

    logDbg('file_make: Create and mount v4 nfs_obj for file creation');

    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                    version  => $version,
                                    proto   => 'tcp' );
    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('file_make: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $curr_fh = $nfs_obj->curr_fh();

    logDbg("file_make: Create $opts{count} files and set the attributes");
    # Call SETATTR separately to allow setting atime, mtime at creation
    if($version =~ /^4/) {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $file_name = "$opts{name}\.$iter";
            my $open = $nfs_obj->open( dhandle => $curr_fh,
                                        name    => $file_name,
                                        cr_mode => UNCHECKED,
                                        attr    => { FATTR4_MODE() => 0777 },
                                        retry   => 6 );


            push @files, $file_name;
            $nfs_obj->setattr( attr => $opts{attr} ) if ( $opts{attr} );

            if ( $opts{fill} ) {
                my $getattr = $nfs_obj->getattr();
                my $size = $getattr->{fattr}{FATTR4_SIZE()};
                my $offset = 0;
                while ( $offset < $size - 8192 ) {
                    my $written = $nfs_obj->write( offset  => $offset,
                                                    data    => $opts{fill}x8192,
                                                    stable  => FILE_SYNC,
                                                    stateid => 0 );
                    $offset += $written->{results}[-1]{result}{count};
                }
                $nfs_obj->write( offset  => $offset,
                                data    => $opts{fill}x( $size-$offset ),
                                stable  => FILE_SYNC,
                                stateid => 0 );
            }
            $nfs_obj->close( stateid => $open->{stateid} );
        }
    } else {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $file_name = "$opts{name}\.$iter";
            my $open = $nfs_obj->create( dhandle => $curr_fh,
                                        name    => $file_name,
                                        cr_mode => UNCHECKED,
                                        attr    => { mode => 0777 },
                                        );

            push @files, $file_name;
            
            # Map the corresponding NFSv2,3 attr in place of NFSv4 attr
            foreach my $key (keys(%{$opts{attr}})) {
                $opts{attr}->{$map->{$key}} = $opts{attr}->{$key};
            }
            
            $nfs_obj->setattr( attr => $opts{attr} ) if ( $opts{attr} );

            if ( $opts{fill} ) {
                my $getattr = $nfs_obj->getattr();
                my $size = $getattr->{fattr}{size};
                my $offset = 0;
                while ( $offset < $size - 8192 ) {
                    my $written = $nfs_obj->write( offset  => $offset,
                                                    data    => $opts{fill}x8192,
                                                    stable  => FILE_SYNC,
                                                    stateid => 0 );
                    $offset += $written->{results}[-1]{result}{count};
                }
                $nfs_obj->write( offset  => $offset,
                                data    => $opts{fill}x( $size-$offset ),
                                stable  => FILE_SYNC,
                                stateid => 0 );
            }
        }
    }
    logDbg('file_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: file_make');
    return @files;
}

#------------------------------------------------------------------------

=head2 dir_make

  @dirs = dir_make( hostname => $hostname,
                    volume   => $volume
                    [, path  => $path ]
                    [, name  => $name ]
                    [, count => $count ]
                    [, attr  => \%attr ] );

Creates one or more directories on the specified host and volume.
Uses NFS v4 protocol for creating the directories. Establishes and
mounts a new NFS client object, looks up the filehandle of the
directory in which the new directories will be created, and calls
CREATE to make the new directory objects.

Directories will be created with mode 0777 by default, and changed to
the attributes specified in the attr parameter if present.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'dir'.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=back

=item Returns

An array containing the directory names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub dir_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            path    => { type => SCALAR,  optional => 1 },
            name    => { type => SCALAR,  default  => 'dir' },
            count    => { type => SCALAR,  default  => 1,
                                            regex    => qr/^[1-9]\d*$/ },
            attr    => { type => HASHREF, default  => {} },
        }
    );
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    my ( $nfs_obj, $fh_mnt, $curr_fh, @dirs );

    logDbg('BEGIN: dir_make');

    logDbg('dir_make: Create and mount v4 nfs_obj for directory creation');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                    version  => $version,
                                     proto   => 'tcp' );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('dir_make: Get directory fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $curr_fh = $nfs_obj->curr_fh();

    logDbg("dir_make: Create $opts{count} directories and set the attributes");
    if($version =~ /^4/) {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $dir_name = "$opts{name}\.$iter";
            $nfs_obj->create( dhandle => $curr_fh,
                            name    => $dir_name,
                            type    => NFDIR,
                            attr    => { FATTR4_MODE() => 0777 } );
            push @dirs, $dir_name;

            $nfs_obj->setattr( attr => $opts{attr} ) if ( $opts{attr} );
        }
    } else {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $dir_name = "$opts{name}\.$iter";
            $nfs_obj->mkdir( dhandle => $curr_fh,
                            name    => $dir_name,
                            attr    => { mode => 0777 } );
            push @dirs, $dir_name;
            # Map the corresponding NFSv2,3 attr in place of NFSv4 attr
            foreach my $key (keys(%{$opts{attr}})) {
                $opts{attr}->{$map->{$key}} = $opts{attr}->{$key};
            }

            $nfs_obj->setattr( attr => $opts{attr} ) if ( $opts{attr} );
        }
    }
    logDbg('dir_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: dir_make');
    return @dirs;
}

#------------------------------------------------------------------------

=head2 link_make

  @links = link_make( hostname => $hostname,
                    volume   => $volume,
                    target   => $target
                    [, path  => $path ]
                    [, name  => $name ]
                    [, count => $count ] );

Creates one or more hard links on the specified host and volume.
Uses NFS v4 protocol for creating the links. Establishes and
mounts a new NFS client object, looks up the filehandle of the
directory in which the new links will be created, and calls
CREATE to make the new link objects.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects
will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item target

(Required) : The path/name of the target for the links. All created
objects will point to the same target object

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'link'.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=back

=item Returns

An array containing the link names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub link_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            target   => { type => SCALAR },
            path    => { type => SCALAR, optional => 1 },
            name    => { type => SCALAR, default  => 'link' },
            count    => { type => SCALAR, default  => 1,
                                        regex    => qr/^[1-9]\d*$/ },
        }
    );
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    my ( $nfs_obj, $fh_mnt, $dir_fh, @links );

    logDbg('BEGIN: link_make');

    logDbg('link_make: Create and mount v4 nfs_obj for link creation');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                    version  => $version,
                                     proto   => 'tcp' );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('link_make: Get fh of target directory from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $dir_fh = $nfs_obj->curr_fh();

    logDbg('link_make: Lookup the target filehandle');
    nfsops_lookup_path( client => $nfs_obj,
                        path   => $opts{target} );

    logDbg("link_make: Create $opts{count} links");
    foreach my $iter ( 1 .. $opts{count} ) {
        my $link_name = "$opts{name}\.$iter";
        $nfs_obj->link( fhandle => $nfs_obj->curr_fh(),
                        todir   => $dir_fh,
                        name    => $link_name );
        push @links, $link_name;

        $nfs_obj->lookup( dhandle => $dir_fh,
                        name    => $link_name );
        $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
    }

    logDbg('link_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: link_make');
    return @links;
}

#------------------------------------------------------------------------

=head2 symlink_make

  @symlinks = symlink_make( hostname => $hostname,
                            volume   => $volume,
                            target   => $target
                            [, path  => $path ]
                            [, name  => $name ]
                            [, attr  => \%attr ]
                            [, count => $count ] );

Creates one or more symbolic links on the specified host and volume.
Uses NFS v4 protocol for creating the symlinks. Establishes and
mounts a new NFS client object, looks up the filehandle of the
directory in which the new symlinks will be created, and calls
CREATE to make the new symlink objects.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects
will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item target

(Required) : The path/name of the target for the symlinks. All created
objects will point to the same target object

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'symlink'.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=back

=item Returns

An array containing the symlink names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub symlink_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            target   => { type => SCALAR },
            path    => { type => SCALAR,  optional => 1 },
            name    => { type => SCALAR,  default  => 'symlink' },
            attr    => { type => HASHREF, default  => {} },
            count    => { type => SCALAR,  default  => 1,
                                            regex    => qr/^[1-9]\d*$/ },
        }
    );
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    my ( $nfs_obj, $fh_mnt, $curr_fh, @links );

    logDbg('BEGIN: symlink_make');

    logDbg('symlink_make: Create and mount v4 nfs_obj for symlink creation');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                    version  => $version,
                                     proto   => 'tcp' );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('symlink_make: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $curr_fh = $nfs_obj->curr_fh();

    logDbg("symlink_make: Create $opts{count} symlinks");
    if($version =~ /^4/) {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $symlink_name = "$opts{name}\.$iter";
            $nfs_obj->create( dhandle => $curr_fh,
                            name    => $symlink_name,
                            type    => NFLNK,
                            data    => $opts{target} );
            push @links, $symlink_name;

            $nfs_obj->lookup( dhandle => $curr_fh,
                            name    => $symlink_name );
            $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
        }
    } else {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $symlink_name = "$opts{name}\.$iter";
            $nfs_obj->symlink( dhandle => $curr_fh,
                                name    => $symlink_name,
                                data    => $opts{target} );
            push @links, $symlink_name;

            $nfs_obj->lookup( dhandle => $curr_fh,
                            name    => $symlink_name );
            $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
        }
    }

    logDbg('symlink_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    logDbg('END: symlink_make');
    return @links;
}

#------------------------------------------------------------------------

=head2 device_make

  @devs = device_make( hostname => $hostname,
                        volume   => $volume,
                        type    => $type
                        [, path  => $path ]
                        [, name  => $name ]
                        [, attr  => \%attr ]
                        [, count => $count ] );

Creates one or more device files on the specified host and volume.
Uses NFS v4 protocol for creating the devices. Establishes and
mounts a new NFS client object, looks up the filehandle of the
directory in which the new devices will be created, and calls
CREATE to make the new symlink objects.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the objects will be created.

=item volume

(Required) : The volume on the filer where the objects will be created..

=item type

(Required) : The type of device to be created. Can be one of NFCHR,
NFBLK, NFSOCK, or NFFIFO.

=item path

(Optional) : The path on the filer where the objects will be created,
relative to the mount point. Defaults to the volume specified.

=item name

(Optional) : The base name to be used for the objects which are
created. Each object will be appended with a dot and an incremental
number. Defaults to 'dev'.

=item attr

(Optional) : A hash of attributes to be applied to the objects when
each is created.

=item count

(Optional) : The number of objects to be created. Defaults to 1.

=back

=item Returns

An array containing the device names that were created.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub device_make ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type    => SCALAR },
            volume   => { type    => SCALAR },
            type    => { type    => SCALAR },
            path    => { type    => SCALAR, optional => 1 },
            name    => { type    => SCALAR,  default => 'dev' },
            attr    => { type    => HASHREF,
                        default => { FATTR4_MODE() => 0777 } },
            count    => { type    => SCALAR,
                        default => 1,
                        regex   => qr/^[1-9]\d*$/ },
        }
    );
    my $f = host("$main::FILER");
        
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    my ( $fh_mnt, $nfs_obj, $h, $rdev, $curr_fh, @devices );
    my ( $major, $minor ) = ( 0,0 );

    logDbg('device_make: Create and mount v4 nfs_obj for device creation');
    my $version = get_good_nfs_version();
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                     version => $version,
                                       proto => 'tcp' );

    $fh_mnt = $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('device_make: Get directory fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }
    $curr_fh = $nfs_obj->curr_fh();

    logDbg("device_make: Create $opts{count} devices");
    # Device major, minor numbers are taken from the localhost
    if ( $opts{type} == NFCHR || $opts{type} == NFBLK ) {
        $h = host( 'localhost' );
        my $out = `df -k /`;
        my @dfrslt = split /\n/, $out;
        my @dfinfo = split / +/, $dfrslt[1];
        $rdev = ( stat( "/dev/zero" ) )[6];
        $rdev = ( stat( $dfinfo[0] ) )[6] if $opts{type} == NFBLK;
        if ( $h->hosttype() =~ /solaris/ ) {
            # Format of solaris rdev result is funky: handle it here
            $minor = $rdev & 0xFFFF;
            $major = ( $rdev>>18 ) & 0xFFFF;
        } else {
            $minor = $rdev & 0xFF;
            $major = ( $rdev>>8 ) & 0xFF;
        }
    }
    $nfs_obj->rpc()->auth_sys_set( uid => 0, gid => 0 );

    if("$main::NFS_VERSION" eq "4" || $main::NFS_VERSION =~ /not[23]/ || "$main::NFS_VERSION" =~ /^all$/) {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $dev_name = "$opts{name}\.$iter";
            my $result;
            eval {
                $result = $nfs_obj->create( dhandle => $curr_fh,
                                        name    => $dev_name,
                                        type    => $opts{type},
                                        major   => $major,
                                        minor   => $minor,
                                        attr    => { FATTR4_MODE() => 0777 } );
            };
            if ($@ && $@ !~ /ERR_EXIST/) {
                croak("device make fails:$@");
            }
            push @devices, $dev_name;

            $nfs_obj->lookup( dhandle => $curr_fh,
                            name    => $dev_name );
            $result = $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
        }
    }
    else {
        foreach my $iter ( 1 .. $opts{count} ) {
            my $dev_name = "$opts{name}\.$iter";
            my $result;
            eval {
                $result = $nfs_obj->mknod( dhandle => $curr_fh,
                                        name    => $dev_name,
                                        type    => $opts{type},
                                        major   => $major,
                                        minor   => $minor,
                                        attr    => { mode => 0777 } );
            };
            if ($@ && $@ !~ /ERR_EXIST/) {
                croak (qq{device make fails:$@});
            }
            push @devices, $dev_name;
            
            # Map the corresponding NFSv2,3 attr in place of NFSv4 attr
            foreach my $key (keys(%{$opts{attr}})) {
                $opts{attr}->{$map->{$key}} = $opts{attr}->{$key};
            }
            
            $nfs_obj->lookup( dhandle => $curr_fh,
                            name    => $dev_name );
            $result = $nfs_obj->setattr( attr => $opts{attr} ) if $opts{attr};
        }
    }
    $nfs_obj->rpc()->auth_correct_set();

    logDbg('device_make: Umount the nfs_obj');
    $fh_mnt = $nfs_obj->umount( path => $opts{volume} );

    return @devices
}

#------------------------------------------------------------------------

=head2 file_write

  file_write( hostname  => $hostname,
            volume    => $volume,
            path    => $path,
            data    => $data,
            stable    => $stable,
            stateid   => $stateid
            [, offset => $offset ] );

Writes data to a file on the hostname specified. Uses NFS v4 to establish
an NFS client object and write to the file.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : The hostname of the filer on which the file to be written
is residing.

=item volume

(Required) : The volume on the filer where file to be written is located..

=item path

(Required) : The volume relative path of the file to be written.

=item data

(Required) : The data to be written to the file.

=item stable

(Required) : A specification of how the NFS v4 write is to be performed.
Can be FILE_SYNC, DATA_SYNC, or UNSTABLE.

=item stateid

(Required) : The NFS v4 stateid of the file to be written, if one is
known. Defaults to 0.

=item offset

(Optional) : The starting byte offset in the file for the data to be
written. Defaults to 0.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub file_write ( % ) {
    my %opts = validate(
        @_,
        {
            hostname => { type => SCALAR },
            volume   => { type => SCALAR },
            path    => { type => SCALAR },
            data    => { type => SCALAR },
            stable   => { type => SCALAR },
            stateid  => { type => SCALAR },
            offset   => { type => SCALAR, default => 0 },
        }
    );
    my ($nfs_obj);
    my $f = host("$main::FILER");
    if ($f->hosttype() =~ /filer-ng/){
        $opts{hostname} = get_data_ip_cmode($main::FILER);
    }

    logDbg('BEGIN: file_write');
    my $version = get_good_nfs_version();
    logDbg("file_write: Create and mount v$version nfs_obj for file writing");
    $nfs_obj = nfs::NfsClient->new( hostname => $opts{hostname},
                                        version  => $version,
                                        proto    => 'tcp' );

    $nfs_obj->mount( path => $opts{volume} );

    if ( $opts{path} ) {
        logDbg('file_write: Get file fh from path');
        nfsops_lookup_path( client => $nfs_obj,
                            path   => $opts{path} );
    }

    logDbg('file_write: Write data to file');
    $nfs_obj->write( offset  => $opts{offset},
                    data    => $opts{data},
                    stable  => $opts{stable},
                    stateid => $opts{stateid} );

    logDbg('END: file_write');
    return;
}

#------------------------------------------------------------------------

=head2 scalar_validate

  $validation = scalar_validate( op        => $op,
                                version    => $version,
                                protocol    => $protocol,
                                expected    => $expected,
                                returned    => $returned
                                [, operator => 'eq'|'ne'|'lt'|'gt' ] );

Validates a scalar result from an NFS operation against an expected
result. Compares the expected and returned values passed in using the
operator given and reports an error if the comparison fails. The error
message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item expected

(Optional) : The expected value from the NFS operation.

=item returned

(Required) : The returned value from the NFS operation.

=item operator

(Optional) : The comparison operator to be used between the two values.
May be one of 'eq' to validate the two values are equal, 'ne' to validate
that the two values are not equal, 'gt' to validate that the expected
value is greater than the returned value, or 'lt' to validate that the
expected value is less than the returned value. Defaults to 'eq'.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub scalar_validate ( % ) {
    my %opts = validate(
        @_,
        {
            op        => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            expected => { type => SCALAR },
            returned => { type => SCALAR },
            operator => { type => SCALAR, default => 'eq',
                        regex => qr/eq|ne|lt|gt/ },
        }
    );
    my $op    = $opts{op};
    my $ver   = $opts{version};
    my $proto = $opts{protocol};

    logDbg('BEGIN: scalar_validate');

    logDbg('scalar_validate: Check that reply exists');
    return "scalar_validate: $op result undefined"
            if ( !defined( $opts{returned} ) );

    if ( $opts{operator} eq 'eq' &&
            !( $opts{expected} eq $opts{returned} ) ) {
        return "scalar_validate: $op results do not match: " .
                "\'$opts{expected}\' expected, \'$opts{returned}\' returned: " .
                "NFS v$ver, $proto";
    }
    if ( $opts{operator} eq 'ne' &&
            !( $opts{expected} ne $opts{returned} ) ) {
        return "scalar_validate: $op results match: " .
                "\'$opts{expected}\' given, \'$opts{returned}\' returned: " .
                "NFS v$ver, $proto";
    }
    if ( $opts{operator} eq 'gt' &&
            !( $opts{expected} gt $opts{returned} ) ) {
        return "scalar_validate: $op results do not validate: " .
                "\'$opts{expected}\' is not greater than \'$opts{returned}\': " .
                "NFS v$ver, $proto";
    }
    if ( $opts{operator} eq 'lt' &&
            !( $opts{expected} lt $opts{returned} ) ) {
        return "scalar_validate: $op results do not match: " .
                "\'$opts{expected}\' is not less than \'$opts{returned}\': " .
                "NFS v$ver, $proto";
    }

    logDbg('END: scalar_validate');
    return 'scalar_validate: validation succeeded';
}

#------------------------------------------------------------------------

=head2 hash_validate

  $validation = hash_validate( op        => $op,
                                version  => $version,
                                protocol => $protocol,
                                expected => \%expected,
                                returned => \%returned
                                [ key    => $key ] );

Validates a hash result from an NFS operation against an expected
result. Compares the expected and returned values passed in and reports
an error if they differ. Only those keys present in the expected hash
are validated, additional keys in the returned hash are ignored.
Nested hashes are compared through recursive calls to hash_validate.
The error message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item expected

(Required) : The hash of expected values from the NFS operation.

=item returned

(Required) : The hash of returned values from the NFS operation.

=item key

(Optional) : A string describing the parent key for reporting purposes
when recursive calls are made to hash_validate. Defaults to ''.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub hash_validate ( % ) {
    my %opts = validate(
        @_,
        {
            op        => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            expected => { type => HASHREF },
            returned => { type => HASHREF },
            key    => { type => SCALAR, default => '' },
        }
    );
    my $pnt_key;
    my $ver    = $opts{version};
    my $proto    = $opts{protocol};
    my %expected = %{$opts{expected}};
    my %returned = %{$opts{returned}};

    if ( $opts{key} ) {
        $pnt_key = "$opts{key}:";
    } else {
        $pnt_key = '';
    }

    logDbg('BEGIN: hash_validate');

    logDbg('hash_validate: Check that result exists');
    return "hash_validate: $opts{op} result undefined"
            if ( !defined( $opts{returned} ) );

    foreach my $key ( keys %expected ) {
        if ( ref( $expected{$key} ) eq 'HASH' ) {
            my $validation = hash_validate( op        => $opts{op},
                                            version  => $ver,
                                            protocol => $proto,
                                            expected => $expected{$key},
                                            returned => $returned{$key},
                                            key    => "$pnt_key$key" );
            if ( $validation !~ /validation succeeded/ ) {
                return $validation;
            }
        } else {
            if ( $expected{$key} ne $returned{$key} ) {


                return "hash_validate: $pnt_key$key mismatch - " .
                            "\'$expected{$key}\' expected, " .
                            "\'$returned{$key}\' returned: NFS v$ver, $proto";

            }
        }
    }

    logDbg('END: hash_validate');
    return 'hash_validate: validation succeeded';
}

#------------------------------------------------------------------------

=head2 hash_validate_change

  $validation = hash_validate_change( op        => $op,
                                    version  => $version,
                                    protocol => $protocol,
                                    initial  => \%initial,
                                    current  => \%current
                                    [ key    => $key ] );

Validates that current hash values have changed from initial hash values.
Compares the initial and current values passed in and reports an error
if they do not differ. Only those keys present in the initial hash
are validated, additional keys in the current hash are ignored.
Nested hashes are compared through recursive calls to hash_validate_change.
The error message is formatted with op, version, and protocol information.

The following named arguments are accepted:

=over

=item Arguments

=over

=item op

(Required) : A string describing the operation/results being validated.
Used in error messages to make them more readable.

=item version

(Required) : The NFS version used for the results being validated.
Used in error messages to make them more readable.

=item protocol

(Required) : The NFS communication protocol used for the results being
validated. Used in error messages to make them more readable.

=item initial

(Required) : The hash of initial values to be validated.

=item current

(Required) : The hash of current values to be validated.

=item key

(Optional) : A string describing the parent key for reporting purposes
when recursive calls are made to hash_validate_change. Defaults to ''.

=back

=item Returns

A string indicating successful validation, or an error.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub hash_validate_change ( % ) {
    my %opts = validate(
        @_,
        {
            op        => { type => SCALAR },
            version  => { type => SCALAR },
            protocol => { type => SCALAR },
            initial  => { type => HASHREF },
            current  => { type => HASHREF },
            key    => { type => SCALAR, default => '' },
        }
    );
    my $pnt_key;
    my $ver = $opts{version};
    my $proto = $opts{protocol};
    my %initial = %{$opts{initial}};
    my %current = %{$opts{current}};

    if ( $opts{key} ) {
        $pnt_key = "$opts{key}:";
    } else {
        $pnt_key = '';
    }

    logDbg('BEGIN: hash_validate_change');

    logDbg('hash_validate_change: Check that result exists');
    return "hash_validate_change: $opts{op} result undefined"
            if ( !defined( $opts{current} ) );

    foreach my $key ( keys %initial ) {
        if ( ref( $initial{$key} ) eq 'HASH' ) {
            my $validation = hash_validate_change( op        => $opts{op},
                                                    version  => $ver,
                                                    protocol => $proto,
                                                    initial  => $initial{$key},
                                                    current  => $current{$key},
                                                    key    => "$pnt_key$key" );
            if ( $validation !~ /validation succeeded/ ) {
                return $validation;
            }
        } else {
            if ( $initial{$key} eq $current{$key} ) {
                return "hash_validate_change: $pnt_key$key did not change - " .
                        "\'$initial{$key}\' initial, " .
                        "\'$current{$key}\' current: NFS v$ver, $proto";
            }
        }
    }

    logDbg('END: hash_validate_change');
    return 'hash_validate_change: validation succeeded';
}

#------------------------------------------------------------------------

=head2 nfsops_lookup_path

  nfsops_lookup_path( client => $client,
                    path   => $path );

Calls a series of LOOKUP operations on a path (relative to
mountpoint) and sets the current filehandle of the NFS object
to the last element of the path.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client

(Required) : An NFS object that has previously called the mount function.

=item path

(Required) : The path, relative to the mount point, to be looked up.

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_lookup_path ( % ) {
    my %opts = validate(
        @_,
        {
            client => { type => OBJECT },
            path   => { type => SCALAR },
        }
    );
    my ($name, @path);

    logDbg('BEGIN: nfsops_lookup_path');

    logDbg('nfsops_lookup_path: Verify mount and reset current filehandle');

    # Verify that client has been mounted
    croak 'NFS client object has no mount filehandle (not mounted)'
            if ( !$opts{client}->mnt_fh() );

    # Set curr_fh to mnt_fh
    $opts{client}->curr_fh($opts{client}->mnt_fh());

    logDbg('nfsops_lookup_path: Split path and call lookups');

    # Get rid of contiguous '/'s and split path into segments
    $opts{path} =~ s{/+}{/};
    $opts{path} =~ s{^/}{};
    @path = split /\//, $opts{path};

    foreach $name (@path) {
        eval { $opts{client}->lookup( name => $name ) };
        croak "LOOKUP error on $name in path '$opts{path}'\n$@" if $@;
    }

    logDbg('END: nfsops_lookup_path');
    return;
}

#------------------------------------------------------------------------

=head2 nfsops_common_attrs

  %attrs = nfs_common_attrs( version  => $version,
                            returned => \%returned );

Converts a hash of v2, v3, v4 attributes to a common format for comparison
between different NFS versions. Certain hash keys are removed, and NFS v4
operation codes are converted to hash keys matching those of v2 and v3.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : The NFS version which generated the attribute keys and values
in the current format.

=item returned

(Required) : A hash of attribute keys matching the format returned by
the version indicated.

=back

=item Returns

A hash containing the attributes with common format key names.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_common_attrs ( % ) {
    my %opts = validate(
        @_,
        {
            version  => { type => SCALAR },
            returned => { type => HASHREF },
        }
    );
    my %common = %{$opts{returned}};

    logDbg('BEGIN: nfsops_common_attrs');

    if ( $opts{version} == 2 ) {
        $common{acc_seconds}   = $common{atime}{seconds};
        $common{acc_usec_nsec} = $common{atime}{useconds};
        delete $common{atime};
        $common{mod_seconds}   = $common{mtime}{seconds};
        $common{mod_usec_nsec} = $common{mtime}{useconds};
        delete $common{mtime};
        my $rdev = $common{rdev};
        delete $common{rdev};
        $common{rdev}{specdata2} = $rdev & 0xFF;
        $common{rdev}{specdata1} = ($rdev>>8) & 0xFF;
    } elsif ( $opts{version} == 3 ) {
        $common{acc_seconds}   = $common{atime}{seconds};
        $common{acc_usec_nsec} = $common{atime}{nseconds};
        delete $common{atime};
        $common{mod_seconds}   = $common{mtime}{seconds};
        $common{mod_usec_nsec} = $common{mtime}{nseconds};
        delete $common{mtime};
    } else {
        $common{nlink}  = $common{FATTR4_NUMLINKS()};
        delete $common{FATTR4_NUMLINKS()};
        $common{fsid}   = $common{FATTR4_FSID()}{major};
        delete $common{FATTR4_FSID()};
        $common{fileid} = $common{FATTR4_FILEID()};
        delete $common{FATTR4_FILEID()};
        $common{size}   = $common{FATTR4_SIZE()};
        delete $common{FATTR4_SIZE()};
        $common{ctime}  = $common{FATTR4_TIME_METADATA()};
        delete $common{FATTR4_TIME_METADATA()};
        $common{mode}   = $common{FATTR4_MODE()};
        delete $common{FATTR4_MODE()};
        $common{type}   = $common{FATTR4_TYPE()};
        delete $common{FATTR4_TYPE()};
        $common{acc_seconds}   = $common{FATTR4_TIME_ACCESS()}{seconds};
        $common{acc_usec_nsec} = $common{FATTR4_TIME_ACCESS()}{nseconds};
        delete $common{FATTR4_TIME_ACCESS()};
        $common{mod_seconds}   = $common{FATTR4_TIME_MODIFY()}{seconds};
        $common{mod_usec_nsec} = $common{FATTR4_TIME_MODIFY()}{nseconds};
        delete $common{FATTR4_TIME_MODIFY()};
        $common{rdev} = $common{FATTR4_RAWDEV()};
        delete $common{FATTR4_RAWDEV()};
        $common{uid} = ( split '@', $common{FATTR4_OWNER()} )[0];
        $common{uid} = getpwnam( $common{uid} );
        delete $common{FATTR4_OWNER()};
        $common{gid} = ( split '@', $common{FATTR4_OWNER_GROUP()} )[0];
        $common{gid} = getgrnam( $common{gid} );
        delete $common{FATTR4_OWNER_GROUP()};
    }

    logDbg('END: nfsops_common_attrs');
    return %common;
}

#------------------------------------------------------------------------

=head2 nfsops_save_common_attrs

  %saved_attrs = nfsops_save_common_attrs( attr_h => \%attr_h );

Generates a hash consisting of a subset of attributes that are common
to all three NFS versions. The subset does not keep attributes that
are not common to all three versions or that can be determined by
other means from the filer. The main purpose of this subroutine is to
select and keep a set of attributes for comparison between several
NFS versions and calls.

The following named arguments are accepted:

=over

=item Arguments

=over

=item attr_h

(Required) : A hash of attributes in nfsops_common_attrs format.

=back

=item Returns

The hash of common attribute keys and values.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_save_common_attrs ( % ) {
    my %opts = validate(
        @_,
        {
            attr_h => { type => HASHREF },
        }
    );
    my %previous;

    logDbg('BEGIN: nfsops_save_common_attrs');

    $previous{nlink}  = $opts{attr_h}{nlink};
    $previous{fsid}   = $opts{attr_h}{fsid};
    $previous{fileid} = $opts{attr_h}{fileid};
    $previous{size}   = $opts{attr_h}{size};
    $previous{ctime}  = $opts{attr_h}{ctime};
    delete $previous{ctime}{nseconds};
    delete $previous{ctime}{useconds};

    logDbg('END: nfsops_save_common_attrs');
    return %previous;
}

#------------------------------------------------------------------------

=head2 nfsops_access_bits

  $access_bits = nfsops_access_bits( bitmask => $bitmask );

Generates a string of access permissions that are returned as a
bitmask from the NFS v3 and v4 ACCESS operation. Indicates which
access bits are set in the given bitmask.

The following named arguments are accepted:

=over

=item Arguments

=over

=item bitmask

(Required) : A bitmask of access rights as returned from an NFS v3 or
v4 ACCESS operation.

=back

=item Returns

A string indicating the set access types, or 'none' if the bitmask is 0.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_access_bits ( % ) {
    my %opts = validate(
        @_,
        {
            bitmask => { type => SCALAR },
        }
    );
    my $bitmask = $opts{bitmask};
    my $acc_bits;

    logDbg('BEGIN: nfsops_access_bits');

    my %access = ( ACCESS_READ    => ACCESS_READ,
                    ACCESS_LOOKUP  => ACCESS_LOOKUP,
                    ACCESS_MODIFY  => ACCESS_MODIFY,
                    ACCESS_EXTEND  => ACCESS_EXTEND,
                    ACCESS_DELETE  => ACCESS_DELETE,
                    ACCESS_EXECUTE => ACCESS_EXECUTE );

    foreach my $key ( sort { $access{$b} <=> $access{$a} } keys %access ) {
        if ( $bitmask >= $access{$key} ) {
            $bitmask  -= $access{$key};
            $acc_bits .= "$key | ";
        }
    }

    logDbg('END: nfsops_access_bits');
    if ( $acc_bits ) {
        return substr( $acc_bits, 0, -3 );
    } else {
        return 'none';
    }
}

#------------------------------------------------------------------------

=head2 nfsops_fsstats

  %fsstats = nfsops_fsstats( fc    => $fc,
                            volume => $volume );

Gathers fsstat information directly from the filer using the df command.
Parses the output for space information (total bytes, free bytes, available
bytes) and file information (total inodes, free inodes, available
inodes).

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : The filer connection object.

=item volume

(Required) : The volume on the filer where the nfsops tests are performed.

=back

=item Returns

A hash of the fsstat information.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_fsstats ( % ) {
    my %opts = validate(
        @_,
        {
            fc    => { type => OBJECT },
            volume => { type => SCALAR },
        }
    );
    my ( $cmd, $results, @dfrslt, @volrslt, $ufiles, %fsstat );

    logDbg('BEGIN: nfsops_fsstats');
    my $h=host($main::FILER);
    if ($h->hosttype() =~ /filer-ng/){
        my @results_out;
        eval{
            my $server_obj = new Server(node=>$main::FILER);
            $opts{volume}=~s/\///;
            @results_out = $server_obj -> volume_show (vserver=>$main::VSERVER_GX_NAME, volume=>$opts{volume});
        };
        $fsstat{fbytes}=$results_out[0]->{'available'};
        $fsstat{abytes}=$results_out[0]->{'available'};
        $fsstat{tbytes}=$results_out[0]->{'total'};
        $fsstat{tfiles}=$results_out[0]->{'files'};
        my $ufiles=$results_out[0]->{'files-used'};
        $fsstat{afiles}=$fsstat{tfiles}-$ufiles;
        $fsstat{ffiles}=$fsstat{tfiles}-$ufiles;
        return %fsstat;
    }

    logDbg('nfsops_fsstats: Call df on volume');
    $cmd    = "df $opts{volume}";
    $results = sayEOF( $opts{fc}, $cmd );

    # Get space information
    @dfrslt  = split /\n/, $results;
    @volrslt = split / +/, $dfrslt[1];
    $fsstat{tbytes} = int( $volrslt[1] ) * 1024;
    $fsstat{fbytes} = int( $volrslt[3] ) * 1024;
    $fsstat{abytes} = $fsstat{fbytes};

    logDbg('nfsops_fsstats: Call df -i on volume');
    $cmd = "df -i $opts{volume}";
    $results = sayEOF( $opts{fc}, $cmd );

    # Get inode information
    @dfrslt  = split /\n/, $results;
    @volrslt = split / +/, $dfrslt[1];
    $ufiles  = int( $volrslt[1] );
    $fsstat{ffiles} = int( $volrslt[2] );
    $fsstat{afiles} = $fsstat{ffiles};
    $fsstat{tfiles} = $fsstat{ffiles} + $ufiles;

    logDbg('END: nfsops_fsstats');
    return %fsstat;
}

#------------------------------------------------------------------------

=head2 nfsops_set_devices

  %devs = nfsops_set_devices( device => 'char'|'block'|'both' );

Creates a hash of device types and ids based on the device value passed in.
The hash indicates a character device type, a block device type, or both.
The returned hash is useful in tests involving devices for stepping through
the types of devices to be created and tested.

The following named arguments are accepted:

=over

=item Arguments

=over

=item device

(Required) : The type(s) of device(s) to be included. May be 'char',
'block', or 'both'.

=back

=item Returns

A hash, in which the keys are the type of device (block and/or
char) and the values are the NFS numeric value of that type.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_set_devices ( % ) {
    my %opts = validate(
        @_,
        {
            device => { type => SCALAR, regex => qr/char|block|both/ },
        }
    );
    my ( $dev, %devices );
    $dev = $opts{device};

    if ( $dev eq 'char' ) {
        %devices = ( char  => NFCHR )
    } elsif ( $dev eq 'block' ) {
        %devices = ( block => NFBLK )
    } elsif ( $dev eq 'both' ) {
        %devices = ( char  => NFCHR, block => NFBLK )
    } else {
        croak "Invalid device specification: $dev";
    }

    return %devices;
}

#------------------------------------------------------------------------

=head2 nfsops_get_new_id

  @ids = nfsops_get_new_id( fc   => $fc,
                            type => 'uid'|'gid' );

Selects a new uid or gid from 'testu1', 'testu2', 'testg1', or 'testg2'.
Type parameter determines if a uid or gid is selected. Tries to assign a new
uid or gid that is different from the current user and that exists on the
filer. If both uids or gids are unusable, an error is returned.

The following named arguments are accepted:

=over

=item Arguments

=over

=item fc

(Required) : A filer connection object to the filer to be used.

=item type

(Required) : A string that identifies the type of id to be selected. Must
be either 'uid' or 'gid'.

=back

=item Returns

An array containing the user or group name and the user or group
id number, or an error if a new id cannot be assigned.

=back

=for html <hr />

=cut

#------------------------------------------------------------------------
sub nfsops_get_new_id ( % ) {
    my %opts = validate(
        @_,
        {
            fc   => { type => HASHREF },
            type => { type => SCALAR, regex => qr/uid|gid/ },
        }
    );

    if ( $opts{type} eq 'uid' ) {
        # Assign 'testu1' if it exists and is not current user
        my $h = host($main::FILER);
        my $testu1;
        if ($h->hosttype()=~/filer-ng/){
            $testu1=run_getXXbyYY_gx();
        }
        else{
        $testu1 = run_getXXbyYY( $opts{fc}, 'getpwbyname_r', 'testu1' );
    }
        ($testu1)  = ( $testu1 =~ /pw_uid = (\d+)/ );
        return ( 'testu1', $testu1 ) if ( $testu1 && $testu1 != $> );

        # Assign 'testu2' if it exists and is not current user
        my $testu2;
        if ($h->hosttype()=~/filer-ng/){
            $testu2=run_getXXbyYY_gx();
        }
        else{
        $testu2 = run_getXXbyYY( $opts{fc}, 'getpwbyname_r', 'testu2' );
    }
        ($testu2)  = ( $testu2 =~ /pw_uid = (\d+)/ );
        return ( 'testu2', $testu2 ) if ( $testu2 && $testu2 != $> );

        # Croak because a new uid could not be assigned
        croak 'Could not assign testu1 or testu2: does not exist on ' .
                'filer or is the current user';
    } else {
        # Assign 'testg1' if it exists and is not current user's primary group
        my $h = host($main::FILER);
        my $testg1;
        if ($h->hosttype()=~/filer-ng/){
            $testg1=run_getXXbyYY_gx();
        }
        else{
        $testg1 = run_getXXbyYY( $opts{fc}, 'getgrbyname', 'testg1' );
    }
        ($testg1)  = ( $testg1 =~ /gid = (\d+)/ );
        return ( 'testg1', $testg1 ) if ( $testg1 != (split " ", $))[0] );

        # Assign 'testg2' if it exists and is not current user's primary group
        my $testg2;
        if ($h->hosttype()=~/filer-ng/){
            $testg2=run_getXXbyYY_gx();
        }
        else{
        $testg2 = run_getXXbyYY( $opts{fc}, 'getgrbyname', 'testg2' );
    }
        ($testg2)  = ( $testg2 =~ /gid = (\d+)/ );
        return ( 'testg2', $testg2 ) if ( $testg2 != (split " ", $))[0] );

        # Croak because a new gid could not be assigned
        croak "Could not assign testg1 or testg2: does not exist on " .
                "filer or is the current user's group";
    }
}

sub testme () {

    logcomment("Performing NfsOps self test");

    require nfs::CommonParams;

    my ( %set_h, %hash1, %hash2 );
    my $ver = 3;
    my $proto = 'tcp';

    # Test setup routine
    eval {
        set_env( setup_h  => \%set_h,
                allowed  => [ 2, 3, 4 ],
                version  => $ver,
                protocol => $proto );
        };
    logresult('FAIL', "set_env failed: $@") if $@;

    # Test cleanup routine
    eval {
        clean_env( %set_h );
    };
    logresult('FAIL', "clean_env failed: $@") if $@;

    # Test scalar validation for equality
    my $out = scalar_validate(
                        op        => 'TESTME: scalar validation, equality',
                        version  => $ver,
                        protocol => $proto,
                        expected => $proto,
                        operator => 'eq',
                        returned => $proto );
    logresult('FAIL', "scalar_validate equality failed: $out")
            if $out !~ /validation succeeded/;

    # Test scalar validation for inequality
    $out = scalar_validate(
                        op        => 'TESTME: scalar validation, inequality',
                        version  => $ver,
                        protocol => $proto,
                        expected => $proto,
                        operator => 'ne',
                        returned => $ver );
    logresult('FAIL', "scalar_validate inequality failed: $out")
            if $out !~ /validation succeeded/;

    # Test hash validation
    for my $index (1..10) {
        $hash1{$index}    = $index;
        $hash1{sub}{$index} = $index;
        $hash2{$index}    = $index;
        $hash2{sub}{$index} = $index;
    }
    $out = hash_validate( op        => 'TESTME: hash validation',
                        version  => $ver,
                        protocol => $proto,
                        expected => \%hash1,
                        returned => \%hash2 );
    logresult('FAIL', "hash_validate failed: $out")
            if $out !~ /validation succeeded/;

    # Test hash change validation
    for my $index (1..10) {
        $hash1{$index} = $index;
        $hash1{sub}{$index} = $index;
        $hash2{$index} = $index + 1;
        $hash2{sub}{$index} = $index + 1;
    }
    $out = hash_validate_change(
                        op        => 'TESTME: hash change validation',
                        version  => $ver,
                        protocol => $proto,
                        initial  => \%hash1,
                        current  => \%hash2 );
    logresult('FAIL', "hash_validate_change failed: $out")
            if $out !~ /validation succeeded/;

    logresult('PASS', "SELF TEST of NfsOps.pm");
}

sub get_gx_v4_id_domain (@) {
    my $filer = shift;
    my $vserver = shift;
    my $server_obj = Server->new(node => ($filer) ? $filer : $main::FILER);
    my @results_out = $server_obj->vserver_nfs_show(vserver => ($vserver) ? $vserver : $main::VSERVER_GX_NAME);
    return $results_out[0]->{'v4-id-domain'};
}

sub run_getXXbyYY_gx () {
    my $server_obj=Server->new(node => $main::FILER);
    my $output;
    eval {
        $server_obj->vserver_services_unix_user_show(
                        vserver => $::VSERVER_GX_NAME,
                        user        => "testu1");
    };
    if ($@) {
        eval {
            $server_obj->vserver_services_unix_user_create(
                        vserver => $::VSERVER_GX_NAME,
                        user    => "testu1",
                        id    => "3196",
                        'primary-gid' => "3196",
                        );
        };
        if ($@) {
        logWarn("$@");
        }
        else {
        $output='pw_uid = 3196';
        }
    }
    else {
        $output='pw_uid = 3196';
    }
    eval {
        $server_obj->vserver_services_unix_group_show(
                            vserver => $::VSERVER_GX_NAME,
                            name  => "testg1");
    };
    if ($@) {
        eval {
            $server_obj->vserver_services_unix_group_create(
                            vserver => $::VSERVER_GX_NAME,
                            name => "testg1",
                            id   => "3196"
                        );
        };
        if ($@) {
                logWarn("$@");
        }
        else {
            $output.=' gid = 3196';
        }
    }
    else {
        $output.=' gid = 3196';
    }
    return $output;

}


sub diskFree_gx (%) {
    my %opts = validate(
                @_,
                {
                    fc   => { type => HASHREF },
                    path => { type => SCALAR },
                }
            );
    my $h = host($main::FILER);
    my $path = $opts{path};
    my $available_space;
    if ($h->hosttype()=~/filer-ng/) {
        $path=~s/\///;
        my @results_out;
        eval{
            my $server_obj=new Server(node=>$main::FILER);
            @results_out=$server_obj->volume_show(vserver=>$main::VSERVER_GX_NAME, volume=> $path);
        };
        $available_space=$results_out[0]->{'available'};
        my $total_space=$results_out[0]->{'size'};
        $total_space/=1024;
        $available_space/=1024;
    }
    return $available_space;
}

#------------------------------------------------------------------------

=head2 create_nfs_client

  $nfs_client = create_nfs_client( hostname => $host
                , version  => $version
                [, protocol => 'udp'|'tcp' ]
                [, timeout  => $timeout ]
                , minor_version => $minor_version);

This routine creates an NFS client object based on the provided NFS
version. If the version is NFSv4, then minor version will be supplied
to the NfsClient constructor. Otherwise, minor version will be omitted.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : Hostname or ip address of server to connect.

=item version

(Required) : NFS version(s) that will be tested. Acceptable values are
3, 4.

=item proto

(Optional) : The NFS connection protocol to be tested: 'udp', 'tcp'.
Default is 'tcp'.

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item minor_version

(Optional) : Minor version of NFSv4.
Default is 0.

=back

=item Returns

An NFS client object

=back

=for html <hr />

=cut

sub create_nfs_client(%) {
    my %opts = validate(
          @_,
          {
             hostname       => { type => SCALAR },
             version        => { type => SCALAR },
             proto          => { type => SCALAR, regex => qr/^(udp|tcp)$/,
                                                 default => 'tcp'},
             timeout        => { type => SCALAR, default => 30 },
             minor_version  => { type => SCALAR, default => 0 }
          }
    );

    my %nfs_opts = %opts;

    my @nfs_ver = validate_nfs_version(version => $nfs_opts{version},
                                        allowed => [ 3, 4 ] );

    my @minor_ver= validate_minor_ver (minor_ver => $nfs_opts{minor_version});
    if ($#minor_ver == -1) {
        croak("Invalid minor version of $nfs_opts{minor_version} specified");
    }

    # Remove minor version from the hash as we only want to pass minor version
    # to the constructor for NFSv4
    if ($nfs_opts{version} == 3) {
        delete $nfs_opts{minor_version};
    }

    my $nfs_client = nfs::NfsClient->new(%nfs_opts);

    return $nfs_client;
}


1;
