# -*- perl -*-
#
## @summary
## This file is a wrapper of sorts for TCD, so that in the future, the
## nfs script developers dont have to keep writing the 5 TCD functions unless
## they need to. The thpls, if they don't need anything else, can simply assume
## the following has been setup for them, and do their testing.
## 1. The vserver is running.
## 2. The vserver nfs is running, and has enabled the required nfs versions.
## 3. The volume "testvol_flex" or whatever is passed for FILER_GX_VOL is setup.
## 4. The volume is mounted at $Mountpoint in the primary client CLIENT.
## 
## @synopsis

=head1 NAME

 nfs::NfsTCD - the base library for all things that are nfs, based on TCD.

=head1 Synopsis

    use nfs::NfsTCD qw{setup cleanup init uninit main $Node_cmd $Mountpoint @UNINIT @CLEANUP $Node_cmd $Client_cmd %Setup_hash};
    our @Testcases = (t_004_SFSR_samevol_difffile_acl => "Perform SFSR on a file
        same volume, different file. Do full restore. Make sure that acls are persistent.",);
    our (%Setup_hash, $Mountpoint, $snapshot_name, $testfile_name, $Client_cmd, $Node_cmd);
    &main;
    sub t_004_SFSR_samevol_difffile_acl() {
        # Do all the testing, push any newly created NACL object into @CLEANUP array as follows.
        # add_to_cleanup is a function from this package !
        add_to_cleanup (NACL::C::Whatver->create());
        # Do any test you care. Don't worry about catching errors, let them just blow up.
        return $TCD::PASS;
    }

=head1 Description

This file provides a wrapper for NFS test thpls around the Tcd framework. It
provides a generic init, setup, cleanup and uninit code. init and uninit will
be run only once per thpl run, and setup and cleanup will be run for every
testcase in the thpl run. 

=head2 If no setup is needed

If the thpl does not desire a setup for every testcase, it just does not import
the setup function into itself so:

    use nfs::NfsTCD qw{cleanup init uninit main $Node_cmd $Mountpoint @UNINIT @CLEANUP $Node_cmd $Client_cmd %Setup_hash};
    our @Testcases = (t_004_SFSR_samevol_difffile_acl => "Perform SFSR on a
        file same volume, different file. Do full restore. Make sure that acls are persistent.",);
    our (%Setup_hash, $Mountpoint, $snapshot_name, $testfile_name, $Client_cmd, $Node_cmd);
    &main;
    sub t_004_SFSR_samevol_difffile_acl() {
        # Do all the testing, push any newly created NACL object into @CLEANUP array.
        return $TCD::PASS;
    }

=head2 If no cleanup is needed.

    Actually you need it. The things created by a testcase must be cleanedup
    after every runs. Or you are going to create a mess later.

=head2 Uninit

    If you have created NACL objects in the init() itself, then push them to
    @UNINIT array. The uninit() function will take care of removing them.

=head2 Overriding.

    Yes, you are always welcome to override any of the functions provided. The
    thpls themselves can provide a init/setup/cleanup/uninit functions, which
    will be called in lieu of the functions provided by this PM file.

=head1 Author

    rraghav@netapp.com

=head1 Maintainer

    dl-svl-nosauto@netapp.com
    dl-nos-codereview@netapp.com

=cut

package nfs::NfsTCD;

use strict;
use nfs::Delegations qw{file_lock_send};
use nfs_setup_gx;
use NACL::C::Node;
use NACL::C::Client;
use NACL::C::VolumeSnapshot;

use NACL::STask::Client::Mount;
use NACL::STask::VolumeSnapshot;

use NACL::Exceptions::NoElementsFound qw{:try};
use NACL::APISet::Exceptions::TimeoutException qw{:try};
use NACL::C::Exceptions::Snapmirror::DoesNotExist qw{:try};
use NACL::C::Exceptions::Volume::DoesNotExist qw{:try};

use Params::Validate qw{validate SCALAR};
use Tharn;
use Data::Dumper;
use TCD;
use nfs::CommonParams;
use nfs::CommonUtilities qw{$Node_cmd $Client_cmd};

use base qw{Exporter};
our @EXPORT_OK = qw{setup cleanup init uninit main add_to_cleanup add_to_uninit
                $Client_cmd $Node_cmd $Mountpoint %Setup_hash @UNINIT @CLEANUP};
our @EXPORT = @EXPORT_OK;

#--------------------------------------------------------------------------------
# Initialize and validate parameters
#--------------------------------------------------------------------------------
param('NFS_VERSION', '-default', 4);
param('AUTH', '-default', 'sys');

our ($Test, %Setup_hash, $Node_cmd, $Client_cmd, $old_nfs_obj, $subid, $lock_id,
    $Mountpoint, @CLEANUP, @UNINIT);
my ($mount_obj);

#-------------------------------------------------------------------------------
# Function prototypes
#--------------------------------------------------------------------------------
# These are needed for TCD
sub main ();
sub init ();
sub setup ();
sub cleanup ();
sub uninit ();

#-------------------------------------------------------------------------------
# test_main
#-------------------------------------------------------------------------------
sub main() {

    #  Create Test Case Driver object
    $Test = new TCD(-testcases => [@::Testcases]);
    if ($Test->error) {
        $Test->log($Test->errmsg);
        return $TCD::FAIL;
    }

    # We want setup and cleanup for every test case. so add this option:
    push(@{$Test->{options}}, '-rsc');

    # Performs method callbacks
    $Test->run_test();
    if ($Test->error) {
        $Test->log($Test->errmsg);
        return $TCD::FAIL;
    }

    exit(0);
} ## end sub main()

#--------------------------------------------------------------------------------
#
#=head2 init
#
#    init ()
#
#The initialization function for nfs_rep. This function
#1. connects to the filer using the default connection type,
#2. sets up the clients.
#
#=item Arguments
#
#This function doesnot take any arguments. It picks them from the global array.
#
#=back
#
#=item Returns
#
#This function aborts if any of its steps fails. otherwise it returns PASS value.
#
#=back
#
#--------------------------------------------------------------------------------
sub init() {
    nfs_setup_gx::set_env_gx(setup_h=>\%Setup_hash,
        allowed => [ 4 ]);

    $Node_cmd = NACL::C::Node->find();
    $Client_cmd = NACL::C::Client->find();
    my $minor_vers = $::NFS4_MINOR_VERSION;

    my %params = (command_interface=>$Client_cmd,
                  server_command_interface => $Node_cmd,
                  server => $Setup_hash{hostname},
                  nacltask_if_exists => 'purge',
                  volume => $Setup_hash{vol_name},
                  vserver => $::VSERVER_GX_NAME,
                  vers => $::NFS_VERSION);

    $params{mountopts} = "sec=$::AUTH";
    if ($minor_vers) {
        $params{mountopts} .= ",minorversion=$minor_vers";
    }

    my $retry = 3;
    while ($retry) {
        try {
            $mount_obj = NACL::STask::Client::Mount->create(%params);
            $retry=0;
        } catch NACL::APISet::Exceptions::TimeoutException with {
            my $excp = shift;
            if ($retry) {
                $Log->warn(qq{Possibly hitting burt:740665});
                $retry--;
            } else {
                $excp->throw();
            }
        };
    }

    $Mountpoint = $mount_obj->mount_point;

    add_to_uninit($mount_obj);

    return $TCD::PASS;
} ## end sub init()

#-------------------------------------------------------------------------------
#=head2 setup()
# As of now this does nothing. If a common setup code pattern emerges among the
# nfs scripts that might be included here.
#-------------------------------------------------------------------------------
sub setup() {
    return $TCD::PASS;
}

#--------------------------------------------------------------------------------
#
#=head2 cleanup
#
#    cleanup()
#
#The cleanup function as given in TCD format. It removes all the unnecessary files
#created and marked for deletion by the previous test.
#
#=item Arguments
#
#This function does not take any argument
#
#=back
#
#=item Returns
#
#This function returns a TCD::PASS.
#
#=back
#
#--------------------------------------------------------------------------------
sub cleanup() {

    if ($subid) {
        stop_io(sub_id=>$subid);
        undef $subid;
    }

    if ($lock_id) {
        file_lock_send(lockid=>$lock_id);
        undef $lock_id;
    }

    clean_obj(\@CLEANUP);

    if ($old_nfs_obj) {

        my %options;
        my $minor = $::NFS4_MINOR_VERSION;
        foreach my $io ("read", "write") {
            $options{"v4.$minor-$io-delegation"} = $old_nfs_obj->{"v4_${minor}_${io}_delegation"};
        }

        NACL::C::VserverNfs->modify(command_interface => $Node_cmd,
                            "vserver" => $::VSERVER_GX_NAME, %options);
        undef $old_nfs_obj;
    }

    return $TCD::PASS;
} ## end sub cleanup ()

# ------------------------------------------------------------------------------
# Function     : uninit()
# Input        : None
# Return       : $TCD::PASS
# summary      : Uninit automatically called
#                after executing tests
# ------------------------------------------------------------------------------
sub uninit() {

    if ($mount_obj) {
        $mount_obj->purge();
    }

    clean_obj(\@UNINIT);
    nfs_setup_gx::clean_env_gx();
    return $TCD::PASS;
}
## end sub uninit()


sub clean_obj($) {
    my $array_ref = shift;
    my @array = @$array_ref;
    $Log->comment(qq{CLEANUP obj list has : } . $#array+1);
    my ($obj, @SECONDARY_CLEANUP);
    while ($obj = pop(@array)) {
        try {
            if ($obj->can('purge')) {
                $obj->purge();
            } elsif ($obj->can('cleanup')) {
                $obj->cleanup();
            } else {
                $obj->delete();
            }
        } catch NACL::Exceptions::NoElementsFound with {
            $Log->comment(qq{The object is referencing an non-existing element, ignoring it.});
        } catch NACL::C::Exceptions::Snapmirror::DoesNotExist with {
            $Log->comment(qq{The object is referencing an non-existing Snapmirror, ignoring it.});
        } catch NACL::C::Exceptions::Volume::DoesNotExist with {
            $Log->comment(qq{The object is referencing an non-existing volume, ignoring it.});
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            push @SECONDARY_CLEANUP, $obj;
        };
    }

    while ($obj = pop(@SECONDARY_CLEANUP)) {
        try {
            if ($obj->can('purge')) {
                $obj->purge();
            } else {
                $obj->delete();
            }
        } catch NATE::BaseException with {
            # Whatever.
            $Log->comment(qq{Can't remove an object, for some reason, ignoring});
        }
    }
}

sub add_to_cleanup($) {
    my $obj = shift;
    push @CLEANUP, $obj;
}

sub add_to_uninit($) {
    push @UNINIT, shift;
}

1;

