# -*-perl-*x
#
# $Id: //depot/prod/test/main/lib/nfs/NfsWrapperv4.pm $
#
# Copyright (c) 2011-2013 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for NFS Wrapper scenarios
##
## @description
##     main                 - Create Test Case Driver object.
##     init                 - Common method to setup client and filer for python script execution.
##     uninit               - Method to undo the client and filer changes done in init().
##     _myhandler           - Method to check the python log file on exit of the python script execution.
##     perform_testcase     - Common method to execute the python wrapper scripts.
##
## @usage
##     use NfsWrapperv4;
##
## @author sanjay.dakshinamurthy@netapp.com
##
## @status private

use strict;
use TCD;
use vars qw($Test);
use Params::Validate qw(validate SCALAR);
use Hostrec;
use NATE::ParamSet;
use Tharn;
use Client;
use Params::Validate qw (validate validate_pos SCALAR SCALARREF OBJECT ARRAYREF HASHREF);
use nfs_setup_gx qw{set_env_gx clean_env_gx};

use NACL::C::Client;
use NACL::C::Interface;
use NACL::C::NetworkInterface;
use NACL::C::Node;
use NACL::C::SecurityLogin;
use NACL::C::SecurityLoginPublickey;
use NACL::C::Volume;
use NACL::C::VserverNfs;
use NACL::C::Vserver;

use NACL::CS::ClusterIdentity;
use NACL::CS::VserverNfs;

use NACL::MTask::SCP;
use NACL::STask::SecurityLogin;
use NACL::STask::Volume;

use NACL::APISet::Exceptions::TimeoutException qw{:try};
use NACL::Exceptions::NoElementsFound qw{:try};
use NATE::BaseException qw(:try);
use Data::Dumper;

our @ISA=qw(TCD);
our @EXPORT=qw (main perform_testcase);
our ($Testcases);
my %setup;
#---------------------------------------------------------------------
# Initialize and validate parameters
#---------------------------------------------------------------------
BEGIN {
    use vars qw($CLIENT             $FILER
                $VSERVER_GX_NAME
                $PYTHON_EXE
                $RSA_ID_PATH        $SCRIPT_PATH
                $PYNFS_VERSION      $Python_run_status
                %GH                 $SED
                $Test
                $TIMEOUT            $PASSWD
                $FILER_VOL          $DATA_LIF_IP
                $FILER_GX_AGGR      $KEY_ID_FILE);

    # Set test specific defaults
    param('FILER',           '-mesg',    'Name of filer');
    param('FILER_GX_AGGR',      '-mesg',    'Name of filer aggregate');
    param('VSERVER_GX_NAME',         '-mesg',    'vserver name');
    param('SCRIPT_PATH',     '-mesg',    'Script user P4 path');
    param('FILER_VOL',       '-default', 'testvol_flex');
    param('CLIENT',          '-default', 'localhost');
    param('PYTHON_EXE',      '-default', '/usr/software/bin/python2.6');
    param('PYNFS_VERSION',   '-default', '40');
    param('RSA_ID_PATH',     '-default', '/tmp');
    param('TIMEOUT',         '-default', '200');
    param('PASSWD',          '-default', 'netapp1!');
    param('KEY_ID_FILE',     '-default', time);
    param('SSH_KEYGEN',      '-default', '/usr/bin/ssh-keygen');
    param('SED',             '-default', '/bin/sed');
} ## end BEGIN

$Python_run_status = 'PASS';
my $response;
my @interfaces;
my $nfs_out;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
# These are needed for TCD
sub main();
sub init ();
sub uninit ();
sub _myhandler($);
sub perform_testcase($$);

#---------------------------------------------------------------------
# Run Tests
#---------------------------------------------------------------------

#---------------------------------------------------------------------
# main
#---------------------------------------------------------------------
our @Testcases;
sub main() {
    #  Create Test Case Driver object
    $Test = new TCD(-testcases => [@Testcases]);
    if ($Test->error) {
        $Test->log($Test->errmsg);
        return $TCD::FAIL;
    }

    # Performs method callbacks
    $Test->run_test();
    if ($Test->error) {
        $Test->log($Test->errmsg);
        return $TCD::FAIL;
    }
    exit(0);
} ## end sub main()

sub create_rsa_key() {
    my $file = "$::RSA_ID_PATH/$::KEY_ID_FILE";
    if (! -e $file) {
        $GH{cc_cmd}->execute(command=>qq{$::SSH_KEYGEN -f $file -N ""});
    }
    return $file;
}

#--------------------------------------------------------------------------------
#
#=head2 init
#
#    init ()
#
#This function
#1.Check the script exists at the specified path or not.
#2.This function copies the public key into the client .ssh file.
#3. This function enable the NFS version 4.0 and 4.1.
#4.This function copies the public key into the filer.
#5.This function validate the Data ip .
#
#=item Arguments
#
#This function does not take any arguments.
#It picks them from the global array.
#
#=back
#
#=item Returns
#
#This function aborts if any of its steps fails.
#otherwise it returns PASS value.
#
#=back
#
#--------------------------------------------------------------------------------
sub init() {
    $Test->description("Performs python setup and does data ip verification");

    $Log->step("Script checking for existence");
    unless ((-e $SCRIPT_PATH)) {
        $Log->comment("CONFIG:Please check the python script path: SCRIPT_PATH:$SCRIPT_PATH");
        return $TCD::CONFIG;
    }

    $GH{cmdinf} = NACL::C::Node->find();
    $GH{cc_cmd} = NACL::C::Client->find();

    my $cluster_obj = NACL::CS::ClusterIdentity->fetch(command_interface=>$GH{cmdinf});
    $GH{cluster_name} = $cluster_obj->{'name'};
    $Log->comment(qq{cluster_name = $GH{cluster_name}});

    my %setup;
    set_env_gx(setup_h => \%setup, minor_version => ($PYNFS_VERSION eq "40" ? "0" : "1"));
    $GH{data_lif_ip} = $setup{hostname};

    my $address = $GH{cmdinf}->{hostrec}->default_ip();
    my $dumb = $GH{cc_cmd}->execute(command=>"host $address");
    my @dumber = split(/ /, $dumb);
    $dumber[-1] =~ s/\.$//g;
    $GH{FILER_MGMT_NAME} = $dumber[-1];
    while (chomp($GH{FILER_MGMT_NAME})) {};
    $Log->comment(qq{FILER_MGMT_NAME=...$GH{FILER_MGMT_NAME}...});

    my $retry = 3;
    do {
        create_rsa_key();

        $Log->step("performing SCP operation on filer");
        my $scp = NACL::MTask::SCP->new(target_command_interface => $GH{cmdinf},
                                        source_command_interface => $GH{cc_cmd},
                                        set_directory_on_target  => '/mroot',
                                        set_directory_on_source  => $::RSA_ID_PATH);
        my $files=["${KEY_ID_FILE}.pub"];
        $scp->copy_from_source_to_target('files_to_copy'=>$files);

        $GH{security_login} = NACL::STask::SecurityLogin->create(command_interface => $GH{cmdinf},
                                 username    => 'admin',
                                 application => 'ssh',
                                 vserver     => $GH{cluster_name},
                                 authmethod  => 'publickey',
                                 nacltask_if_exists => "purge",);

        my $result;
        try {
            NACL::C::SecurityLoginPublickey->load_from_uri(command_interface => $GH{cmdinf},
                                            username => 'admin',
                                            vserver  => $GH{cluster_name},
                                            uri => "file://localhost/mroot/${KEY_ID_FILE}.pub");
        } catch NACL::Exceptions::NoElementsFound with {
            my $excp = shift;
            $Log->comment ("Not able to load public key from uri : " . $excp->text());
            $result = $TCD::FAIL;
        } otherwise {
            my $excp = shift;
            $Log->comment("Failed message :" . Dumper($excp));
            $result = $TCD::FAIL;
        };
        return $result if (defined($result));

        $Log->step("Configuring the client");
        my $temp_file = "/tmp/known_hosts." . time();
        my $ssh_file = "~/.ssh/known_hosts";
        my $output = $GH{cc_cmd}->execute(command => "ls -l $ssh_file");
        if ($output =~ /No such file or directory/) {
            $GH{cc_cmd}->execute(command => "mkdir ~/.ssh");
            $GH{cc_cmd}->execute(command => "chmod 700 ~/.ssh");
            $GH{cc_cmd}->execute(command => "touch $ssh_file");
            $GH{cc_cmd}->execute(command => "chmod 644 $ssh_file");
        } else {
            $GH{cc_cmd}->execute(command => "grep -v $GH{FILER_MGMT_NAME} $ssh_file > $temp_file; mv $temp_file $ssh_file");
        }
        $GH{cc_cmd}->execute(command => "cp $RSA_ID_PATH/$KEY_ID_FILE ~/.ssh/",
                         timeout => "$TIMEOUT");
        $GH{cc_cmd}->execute(command => "chmod 600 ~/.ssh/$KEY_ID_FILE");
        $GH{cc_cmd}->execute(command => "eval `ssh-agent -s`");
        $GH{cc_cmd}->execute(command => "ssh-add ~/.ssh/$KEY_ID_FILE",
                             timeout => "$TIMEOUT",
                          must_match => "Identity");

        sleep(5);
        try {
            $output = $GH{cc_cmd}->execute( command => "ssh -o StrictHostKeyChecking=no admin\@$GH{FILER_MGMT_NAME} cluster identity show");
        } catch NACL::APISet::Exceptions::TimeoutException with {
            $retry--;
            if ($retry) {
                $Log->comment(qq{Got a password prompt. We will retry to configure the password-less config.});
                $GH{cc_cmd}->execute(command=>"rm -rf $RSA_ID_PATH/$KEY_ID_FILE");
            } else {
                croak (qq{Configuring password-less login to the test filer/vsim failed, aborting this run});
            }
        };

        if ($output) {
            if ($output !~ /Cluster Name/) {
                $Log->comment(qq{Didnot get the expected reply for ssh command, fail});
                return $TCD::CONFIG;
            } else {
                $Log->comment(qq{Got the expected output, proceding...});
                $retry = 0;
            }
        }
    } while ($retry);

    if ($main::PYNFS_VERSION eq '40') {
        $GH{cc_cmd}->execute(command => "export PYNFS_VERSION=40",
                     timeout => $TIMEOUT);
    } elsif ($main::PYNFS_VERSION eq '41') {
        $GH{cc_cmd}->execute(command => "export PYNFS_VERSION=41",
                     timeout => $TIMEOUT);
    } else {
        $Log->comment("The $PYNFS_VERSION provided is not proper. It should be either '40' or '41'");
        return $TCD::CONFIG;
    }

    $GH{cc_cmd}->execute(command => "env | grep PYNFS_VERSION",
                 timeout => $TIMEOUT);

    my $junction_path = $setup{volume};

    # We need to disable delegations since it messes up the python code's expectations.
    $GH{vserver_nfs_state} = NACL::CS::VserverNfs->fetch(command_interface=>$GH{cmdinf},
                filter => { vserver => $VSERVER_GX_NAME });
    NACL::C::VserverNfs->modify(command_interface=>$GH{cmdinf},
                vserver => $VSERVER_GX_NAME,
                'v4.0-read-delegation'  => 'disabled',
                'v4.0-write-delegation'  => 'disabled',
                'v4.1-read-delegation'  => 'disabled',
                'v4.1-write-delegation'  => 'disabled');


    return $TCD::PASS;
} ## end sub init()

#--------------------------------------------------------------------------------
#
#=head2 uninit
#
#    uninit ()
#
#1. This function removes the public key from filer so that it can be avavilable for next test.
#2. This function removes the public key and known_host from client SSH directory.
#3. This function disable the NFS version if modified by the script.
#=item Arguments
#
#This function does not take any argument
#
#=back
#
#=item Returns
#
#This function returns $TCD::PASS on completion.
#
#=back
#
#--------------------------------------------------------------------------------
sub uninit() {
    $Test->description("########## UNINIT ##########\n");
    $Test->log(" Removing the security login for username admin");

    clean_env_gx();

    if ($GH{vserver_nfs_state}) {
        foreach (qw{v4.0-read-delegation v4.0-write-delegation v4.1-read-delegation v4.1-write-delegation}) {
            NACL::C::VserverNfs->modify(command_interface=>$GH{cmdinf},
                vserver => $VSERVER_GX_NAME, $_ => $GH{vserver_nfs_state}->$_);
        }
    }

    my $result;
    try {
        NACL::C::SecurityLogin->delete(
                    command_interface => $GH{cmdinf},
                     username    => 'admin',
                     application => 'ssh',
                     vserver     => $GH{cluster_name},
                     authmethod  => 'publickey',
                    );
    } catch NACL::Exceptions::NoElementsFound with {
        $Log->comment ("The publickey login was not even added");
    } otherwise {
        $Log->comment("Failed message : $@");
        $result = $TCD::FAIL;
    };
    return $result if (defined($result));

    $Test->log(" Removing the copied public key from client");
    $GH{cc_cmd}->execute(command => "rm -f ~/.ssh/$::KEY_ID_FILE", timeout => 60);
    $GH{cc_cmd}->execute(command => "rm -f ${RSA_ID_PATH}/$::KEY_ID_FILE", timeout => 60);

    return $TCD::PASS;
} ## end sub uninit()

########################################################################
#     P R I V A T E   M E T H O D S
########################################################################

#--------------------------------------------------------------------------------
#
#=head2 _myhandler
#
#    _myhandler ()
#
#=item Arguments
#
#This function takes the subtest reference object as an argument.
#It Parses the python log file for the run status and sets the
#variable $Python_run_status as 'PASS|FAIL'.
#
#=back
#
#=item Returns
#
#This function returns '0' on completion.
#
#=back
#
#--------------------------------------------------------------------------------

sub _myhandler($) {
    my ($subtestobj) = shift;
    my $match;
    my $status;
    $Python_run_status = ();

    my $Logfile = $subtestobj->{'logfile'};
    $Log->comment("--- python run log file is :: $Logfile");
    open(FH, "$Logfile");
    my $recordSeparator = $/;
    undef $/;
    my $file_content = <FH>;
    $/ = $recordSeparator;
    if ($file_content =~ /# @@ 0 META command.*# @@ 0 META status=(\w*)/s) {
        $match = $1;
    } else {
        $Log->comment("--- python script did not return the expected output");
        return $TCD::CONFIG;
    }
    $Python_run_status = ($match eq "PASS" ? 'PASS' : 'FAIL');
    return 0;
} ## end sub _myhandler()

sub analyze_python_log($) {
    my $file_content = shift;
    my $match;
    if ($file_content =~ /# @@ 0 META command.*# @@ 0 META status=(\w*)/s) {
        $match = $1;
    } else {
        $Log->comment("--- python script did not return the expected output");
        return $TCD::CONFIG;
    }
    return ($match eq "PASS" ? $TCD::PASS : $TCD::FAIL);
}

#--------------------------------------------------------------------------------
#
#=head2 perform_testcase
#
# perform_testcase ()
#
#1. This function takes the testcase name ,runid and description as argument.
#2.This function returns the result to the concerned called functions.
#=item Arguments
#
#This function take testcasename,runid and description as parameter.
#
#=back
#
#=item Returns
#
#
#=back
#
#--------------------------------------------------------------------------------
sub perform_testcase($$) {
    my ($testcasename, $runid) = @_;
    $Test->description($runid);
    $Log->comment("$testcasename: Executing the python script and timeout value is $TIMEOUT");
    my $handle;
    my @subtest_args = ($PYTHON_EXE, '-onexit', \&_myhandler,
                        '-runid', "$runid",
                        '-iffatal', '-fatal', "Subtest failure");

    if ($CLIENT ne "localhost") {
        push @subtest_args, "-client", $CLIENT;
    }

    use File::Basename;
    my $dirname = dirname($SCRIPT_PATH);
    $GH{cc_cmd}->execute(command=>"cd $dirname");

    my $retry = 2;
    while ($retry) {
        my $output = $GH{cc_cmd}->execute(command=>"$PYTHON_EXE -B $SCRIPT_PATH $testcasename $GH{data_lif_ip} $GH{FILER_MGMT_NAME} $VSERVER_GX_NAME", timeout=>$TIMEOUT);

        if (analyze_python_log($output) != $TCD::PASS) {
            $retry--;
            $Log->warn(qq{Python code failed in this attempt.});
        } else {
            return $TCD::PASS;
        }
    }
} ## end sub perform_testcase($$$)
1;
