# this is basically -*- perl -*-
# $Id$
#
## @summary    NfsXdr - XDR encoding/decoding routines for NFS
## @author    dl-nfsqa-only (NFS QA Sunnyvale)
## @status    public
## @notes    Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

NfsXdr - XDR encoding/decoding routines for NFS

=head1 SYNOPSIS

  use nfs::NfsXdr;

  # Encoding/decoding functions for basic types
  $out = xdr_uint      ( encode  => $value );  # Encode
  $out = xdr_uint64    ( xdr     => $xdr_obj ); # Decode
  $out = xdr_opaque    ( encode  => $string );
  $out = xdr_name      ( version => 4, encode => $name );
  $out = xdr_path      ( version => 4, encode => $path );
  $out = xdr_fhandle   ( version => 3, encode => $fhandle );
  $out = xdr_cookie    ( version => 4, encode => $cookie );
  $out = xdr_cookieverf( encode  => $cookieverf );

  # Encoding/decoding functions for data structures
  $out = xdr_nfstime      ( version => 3, encode => \%nfstime );
  $out = xdr_fsid4        ( encode => \%fsid4 );
  $out = xdr_specdata     ( encode => \%specdata );
  $out = xdr_change_info4 ( encode => \%change_info4 );
  $out = xdr_nfsace4      ( encode => \%nfsace4 );
  $out = xdr_fattr4_acl   ( encode => \@fattr4_acl );
  $out = xdr_bitmap4      ( encode => \@bitmap4 );
  $out = xdr_stateid4     ( encode => \%stateid );
  $out = xdr_fs_locations4( encode => \@fs_locations4 );
  $out = xdr_fs_locations_info4( encode => \@fs_locations_info ); #kvn
  $out = xdr_fs_status( encode => \@fs_status ); #kvn

  $out = xdr_fattr4       ( encode => \%fattr4 );

  # Encoding functions
  $out = xdr_settime( encode => \%nfstime );
  $out = xdr_sattr  ( encode => {size=>20, mode => 0644}, version => 3 );

  # Decoding functions
  $pre_op_attr  = xdr_pre_op_attr ( xdr => $xdr_obj );
  $post_op_attr = xdr_post_op_attr( xdr => $xdr_obj );
  $post_op_fh3  = xdr_post_op_fh3 ( xdr => $xdr_obj );
  $wcc_data     = xdr_wcc_data    ( xdr => $xdr_obj );
  $out = xdr_fattr      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_lock4_denied( xdr => $xdr_obj );
  $out = xdr_status     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_access     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_getfh      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_getattr    ( xdr => $xdr_obj, version => 4 );
  $out = xdr_setattr    ( xdr => $xdr_obj, version => 4 );
  $out = xdr_lookup     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_readdir    ( xdr => $xdr_obj, version => 4 );
  $out = xdr_readdirplus( xdr => $xdr_obj );
  $out = xdr_readlink   ( xdr => $xdr_obj, version => 4 );
  $out = xdr_mkdir      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_mknod      ( xdr => $xdr_obj );
  $out = xdr_symlink    ( xdr => $xdr_obj, version => 4 );
  $out = xdr_create     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_remove     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_rename     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_link       ( xdr => $xdr_obj, version => 4 );
  $out = xdr_fsstat     ( xdr => $xdr_obj, version => 3 );
  $out = xdr_fsinfo     ( xdr => $xdr_obj );
  $out = xdr_pathconf   ( xdr => $xdr_obj, version => 3 );
  $out = xdr_read       ( xdr => $xdr_obj, version => 4 );
  $out = xdr_write      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_commit     ( xdr => $xdr_obj, version => 4 );
  $out = xdr_open       ( xdr => $xdr_obj, version => 4 );
  $out = xdr_open_downgrade( xdr => $xdr_obj, version => 4 );
  $out = xdr_close      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_lock       ( xdr => $xdr_obj, version => 4 );
  $out = xdr_locku      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_lockt      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_secinfo    ( xdr => $xdr_obj, version => 4 );
  $out = xdr_setclientid( xdr => $xdr_obj, version => 4 );
  $out = xdr_mount      ( xdr => $xdr_obj, version => 4 );
  $out = xdr_dump       ( xdr => $xdr_obj );
  $out = xdr_export     ( xdr => $xdr_obj );

  # NFSv4 compound requests and callbacks
  $cmpd_args = xdr_compound    ( tag => 'lookup', ops => \@operations );
  $results   = xdr_compound_res( xdr => $xdr_obj );
  $cmpd_args = xdr_cb_compound ( xdr => $xdr_obj );
  $out       = xdr_cb_recall   ( xdr => $xdr_obj );

  # NLM and NSM for NFSv2 and v3
  $out = xdr_nlm_stat( xdr => $xdr_obj );
  $out = xdr_nlm_test( xdr => $xdr_obj, version => $version );
  $out = xdr_nlm_lock( encode => \%nlm_lock, version => $version );
  $res = xdr_cb_nlm_granted( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $res = xdr_cb_nsm_notify( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

  # Generic daemon functions (used to create daemons for NLM and NSM)
  $res = xdr_cb_port( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $res = xdr_cb_register( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $res = xdr_cb_pid( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $res = xdr_cb_exit( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $res = xdr_cb_info( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );
  $out = xdr_daemon_info( xdr => $xdr_obj );

=head1 DESCRIPTION

A collection of functions for the encoding/decoding to/from XDR
format for common types and structures used in NFSv2, v3 and v4.

All functions used for decoding XDR, the optional argument 'xdr'
must be provided.
When encoding the 'xdr' argument is optional, if it is provided
the encoded string is appended to the Xdr object instead of
creating a new Xdr object. This is useful when creating a
function which encodes a structure of other types.
Therefore, if a series of objects need to be encoded, it is
best to create a new Xdr object and pass it to each conversion
function instead of each function creating a new Xdr object:

    # Encoding a 32bit unsigned integer
    $encoded_int = xdr_uint( encode => 24 );

    # Encoding a 32bit unsigned integer followed by a 64bit unsigned integer
    my $xdr = nfs::Xdr->new();               # Create a XDR object
    xdr_uint  ( encode => 24, xdr => $xdr ); # Encode 32bit int
    xdr_uint64( encode => 33, xdr => $xdr ); # Append 64bit int to encoded string

    # Decoding the above encoded string
    $int32 = xdr_uint  ( xdr => $xdr );
    $int64 = xdr_uint64( xdr => $xdr );

The best way to test this library is by running the tests in nfs/ops.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-nfsqa-only (NFS QA Sunnyvale)
  $Id$
  Last revised $Date$

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::NfsXdr;

use strict;
use warnings;
use base qw(Exporter);
use Tharn qw($Log);
use Data::Dumper;

use nfs::Xdr;
use Carp qw(croak);
use Params::Validate qw(validate SCALAR OBJECT ARRAYREF HASHREF UNDEF);
use Switch;
use nfs::NfsDef qw(bitmap4_convert
  FATTR4_ACL FATTR4_ACLSUPPORT FATTR4_ARCHIVE FATTR4_CANSETTIME
  FATTR4_CASE_INSENSITIVE FATTR4_CASE_PRESERVING FATTR4_CHANGE
  FATTR4_CHOWN_RESTRICTED FATTR4_FH_EXPIRE_TYPE FATTR4_FILEHANDLE FATTR4_FILEID
  FATTR4_FILES_AVAIL FATTR4_FILES_FREE FATTR4_FILES_TOTAL FATTR4_FSID
  FATTR4_FS_LOCATIONS FATTR4_FS_LOCATIONS_INFO FATTR4_FS_STATUS
  FATTR4_HIDDEN FATTR4_HOMOGENEOUS FATTR4_LEASE_TIME
  FATTR4_LINK_SUPPORT FATTR4_MAXFILESIZE FATTR4_MAXLINK FATTR4_MAXNAME
  FATTR4_MAXREAD FATTR4_MAXWRITE FATTR4_MIMETYPE FATTR4_MODE
  FATTR4_MOUNTED_ON_FILEID FATTR4_NAMED_ATTR FATTR4_NO_TRUNC FATTR4_NUMLINKS
  FATTR4_OWNER FATTR4_OWNER_GROUP FATTR4_QUOTA_AVAIL_HARD
  FATTR4_QUOTA_AVAIL_SOFT FATTR4_QUOTA_USED FATTR4_RAWDEV FATTR4_RDATTR_ERROR
  FATTR4_SIZE FATTR4_SPACE_AVAIL FATTR4_SPACE_FREE FATTR4_SPACE_TOTAL
  FATTR4_SPACE_USED FATTR4_SUPPORTED_ATTRS FATTR4_SYMLINK_SUPPORT FATTR4_SYSTEM
  FATTR4_TIME_ACCESS FATTR4_TIME_ACCESS_SET FATTR4_TIME_BACKUP
  FATTR4_TIME_CREATE FATTR4_TIME_DELTA FATTR4_TIME_METADATA FATTR4_TIME_MODIFY FATTR4_LAYOUT_TYPE FATTR4_LAYOUT_ALIGNMENT
  FATTR4_TIME_MODIFY_SET FATTR4_TYPE FATTR4_UNIQUE_HANDLES FATTR4_SPACE_RESERVE FATTR4_LAYOUT_BLKSIZE
  FATTR4_LAYOUT_ALIGNMENT FATTR4_SUPPATTR_EXCLCREAT FATTR4_LAYOUT_MDSTHRESHOLD FATTR4_LAYOUT_HINT
  MAXDATA MAXNAMLEN
  MAXPATHLEN NFS4ERR_BAD_STATEID NFS4ERR_BADHANDLE NFS4ERR_CLID_INUSE
  NFS4ERR_DENIED NFS4_OK NLM_DENIED OP_ACCESS OP_CB_RECALL OP_CLOSE OP_COMMIT
  OP_CREATE OP_DELEGRETURN OP_DELEGPURGE OP_GETATTR OP_GETFH OP_LINK OP_LOCK
  OP_LOCKT OP_LOCKU OP_LOOKUP OP_LOOKUPP OP_NVERIFY OP_OPEN OP_OPEN_CONFIRM
  OP_OPEN_DOWNGRADE OP_OPENATTR OP_PUTPUBFH OP_PUTFH OP_PUTROOTFH OP_READ
  OP_READDIR OP_READLINK OP_RELEASE_LOCKOWNER OP_REMOVE OP_RENAME OP_RENEW
  OP_RESTOREFH OP_SAVEFH OP_SECINFO OP_SECINFO_NO_NAME OP_SETATTR OP_SETCLIENTID
  OP_SETCLIENTID_CONFIRM OP_VERIFY OP_WRITE OP_EXCHANGE_ID OP_CREATE_SESSION OPEN_DELEGATE_WRITE OPEN_DELEGATE_NONE_EXT
  RPCSEC_GSS SET_TO_SERVER_TIME STATEID0 STATEID1 OP_DESTROY_SESSION OP_SEQUENCE OP_GETDEVICELIST
  OP_GETDEVICEINFO OP_LAYOUTGET
  OP_COPY_NOTIFY OP_COPY :WND4_TAGS SP4_MACH_CRED SP4_SSV);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
our @EXPORT_OK = qw(
  xdr_uint
  xdr_uint64
  xdr_opaque
  xdr_name
  xdr_path
  xdr_fhandle
  xdr_cookie
  xdr_cookieverf
  xdr_nfstime
  xdr_settime
  xdr_fsid4
  xdr_specdata
  xdr_change_info4
  xdr_nfsace4
  xdr_bitmap4
  xdr_stateid4
  xdr_fs_locations4
  xdr_fs_locations_info4
  xdr_fs_status
  xdr_fattr4
  xdr_sattr
  xdr_fattr
  xdr_lock4_denied
  xdr_pre_op_attr
  xdr_post_op_attr
  xdr_post_op_fh3
  xdr_wcc_data
  xdr_nlm_lock
  xdr_compound
  xdr_compound_res
  xdr_cb_compound
  xdr_cb_recall
  xdr_cb_nlm_granted
  xdr_cb_nsm_notify
  xdr_cb_port
  xdr_cb_register
  xdr_cb_pid
  xdr_cb_exit
  xdr_cb_info
  xdr_daemon_info
  xdr_status
  xdr_access
  xdr_getfh
  xdr_getattr
  xdr_setattr
  xdr_lookup
  xdr_readdir
  xdr_readdirplus
  xdr_readlink
  xdr_mkdir
  xdr_mknod
  xdr_symlink
  xdr_create
  xdr_remove
  xdr_rename
  xdr_link
  xdr_fsstat
  xdr_fsinfo
  xdr_pathconf
  xdr_read
  xdr_write
  xdr_commit
  xdr_open
  xdr_open_downgrade
  xdr_close
  xdr_lock
  xdr_locku
  xdr_lockt
  xdr_nlm_stat
  xdr_nlm_test
  xdr_secinfo
  xdr_setclientid
  xdr_mount
  xdr_dump
  xdr_export
  xdr_layouthint
  xdr_layouthint4
);

#---------------------------------------------------------------------
# Subroutine prototypes
#---------------------------------------------------------------------
sub xdr_uint(%);
sub xdr_uint64(%);
sub xdr_opaque(%);
sub xdr_name(%);
sub xdr_path(%);
sub xdr_fhandle(%);
sub xdr_cookie(%);
sub xdr_cookieverf(%);
sub xdr_nfstime(%);
sub xdr_settime(%);
sub xdr_fsid4(%);
sub xdr_specdata(%);
sub xdr_change_info4(%);
sub xdr_nfsace4(%);
sub xdr_fattr4_acl(%);
sub xdr_bitmap4(%);
sub xdr_stateid4(%);
sub xdr_fs_locations4(%);
sub xdr_fs_locations_info4(%);
sub xdr_fs_status(%);
sub xdr_fattr4(%);
sub xdr_sattr(%);
sub xdr_fattr(%);
sub xdr_lock4_denied(%);
sub xdr_pre_op_attr(%);
sub xdr_post_op_attr(%);
sub xdr_post_op_fh3(%);
sub xdr_wcc_data(%);
sub xdr_nlm_lock(%);
sub xdr_compound(%);
sub xdr_compound_res(%);
sub xdr_cb_compound(%);
sub xdr_cb_recall(%);
sub xdr_cb_nlm_granted(%);
sub xdr_cb_nsm_notify(%);
sub xdr_cb_port(%);
sub xdr_cb_register(%);
sub xdr_cb_pid(%);
sub xdr_cb_exit(%);
sub xdr_cb_info(%);
sub xdr_daemon_info(%);
sub xdr_status(%);
sub xdr_access(%);
sub xdr_getfh(%);
sub xdr_getattr(%);
sub xdr_setattr(%);
sub xdr_lookup(%);
sub xdr_readdir(%);
sub xdr_readdirplus(%);
sub xdr_readlink(%);
sub xdr_mkdir(%);
sub xdr_mknod(%);
sub xdr_symlink(%);
sub xdr_create(%);
sub xdr_remove(%);
sub xdr_rename(%);
sub xdr_link(%);
sub xdr_fsstat(%);
sub xdr_fsinfo(%);
sub xdr_pathconf(%);
sub xdr_read(%);
sub xdr_write(%);
sub xdr_commit(%);
sub xdr_open(%);
sub xdr_open_downgrade(%);
sub xdr_close(%);
sub xdr_lock(%);
sub xdr_locku(%);
sub xdr_lockt(%);
sub xdr_nlm_stat(%);
sub xdr_nlm_test(%);
sub xdr_secinfo(%);
sub xdr_setclientid(%);
sub xdr_mount(%);
sub xdr_dump(%);
sub xdr_export(%);

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our @cb_queue;
our @open_delegs;
our $daemon_info;
our $daemon_clients;
our $daemon_program;
our $daemon_reg_ports;
our %nlm_clients;
our %nsm_clients;
our @nlm_cb_queue;
our @nsm_cb_queue;
our @nlm_reg_ports;
our @nsm_reg_ports;
our %nlm_info = ( used => 0, udp => 0, tcp => 0, ports => [] );
our %nsm_info = ( used => 0, udp => 0, tcp => 0, ports => [] );
our %cb_info = ( used => 0, udp => 0, tcp => 0, ports => [] );

#---------------------------------------------------------------------
# Package local constants
#---------------------------------------------------------------------
# Map NFSv4 operation to its
# corresponding XDR decoding function
my %nfsv4_res = (
    OP_ACCESS()              => \&xdr_access,
    OP_READ()                => \&xdr_read,
    OP_WRITE()               => \&xdr_write,
    OP_COMMIT()              => \&xdr_commit,
    OP_OPEN()                => \&xdr_open,
    OP_OPEN_DOWNGRADE()      => \&xdr_open_downgrade,
    OP_OPEN_CONFIRM()        => \&xdr_open_downgrade,
    OP_OPENATTR()            => \&xdr_status,
    OP_CLOSE()               => \&xdr_close,
    OP_LOCK()                => \&xdr_lock,
    OP_LOCKU()               => \&xdr_locku,
    OP_LOCKT()               => \&xdr_lockt,
    OP_RELEASE_LOCKOWNER()   => \&xdr_status,
    OP_CREATE()              => \&xdr_create,
    OP_GETATTR()             => \&xdr_getattr,
    OP_GETFH()               => \&xdr_getfh,
    OP_LINK()                => \&xdr_link,
    OP_LOOKUP()              => \&xdr_status,
    OP_LOOKUPP()             => \&xdr_status,
    OP_PUTFH()               => \&xdr_status,
    OP_PUTROOTFH()           => \&xdr_status,
    OP_READDIR()             => \&xdr_readdir,
    OP_READLINK()            => \&xdr_readlink,
    OP_REMOVE()              => \&xdr_remove,
    OP_RENAME()              => \&xdr_rename,
    OP_SAVEFH()              => \&xdr_status,
    OP_SETATTR()             => \&xdr_setattr,
    OP_VERIFY()              => \&xdr_status,
    OP_NVERIFY()             => \&xdr_status,
    OP_RENEW()               => \&xdr_status,
    OP_DELEGRETURN()         => \&xdr_status,
    OP_DELEGPURGE()          => \&xdr_status,
    OP_SETCLIENTID()         => \&xdr_setclientid,
    OP_SETCLIENTID_CONFIRM() => \&xdr_status,
    OP_SECINFO()             => \&xdr_secinfo,
    OP_SECINFO_NO_NAME()     => \&xdr_secinfo,
    OP_PUTPUBFH()            => \&xdr_status,
    OP_RESTOREFH()           => \&xdr_status,
    OP_EXCHANGE_ID()         => \&xdr_exchange_id,
    OP_CREATE_SESSION()      => \&xdr_create_session,
    OP_SEQUENCE()            => \&xdr_sequence,
    OP_GETDEVICELIST()       => \&xdr_getdevicelist,
    OP_GETDEVICEINFO()       => \&xdr_getdeviceinfo,
    OP_LAYOUTGET()           => \&xdr_layoutget,
    OP_COPY_NOTIFY()         => \&xdr_copy_notify,
    OP_COPY()                => \&xdr_copy,
);

# Map NFSv4 callback operation to its corresponding XDR processing function
my %nfsv4_cb = (
    OP_CB_RECALL() => \&xdr_cb_recall,
);

# Map FATTR4 bits to its corresponding
# XDR encoding/decoding function
my %nfsv4_attr = (
    FATTR4_SUPPORTED_ATTRS()   => \&xdr_bitmap4,
    FATTR4_TYPE()              => \&xdr_uint,
    FATTR4_FH_EXPIRE_TYPE()    => \&xdr_uint,
    FATTR4_CHANGE()            => \&xdr_uint64,
    FATTR4_SIZE()              => \&xdr_uint64,
    FATTR4_LINK_SUPPORT()      => \&xdr_uint,
    FATTR4_SYMLINK_SUPPORT()   => \&xdr_uint,
    FATTR4_NAMED_ATTR()        => \&xdr_uint,
    FATTR4_FSID()              => \&xdr_fsid4,
    FATTR4_UNIQUE_HANDLES()    => \&xdr_uint,
    FATTR4_LEASE_TIME()        => \&xdr_uint,
    FATTR4_RDATTR_ERROR()      => \&xdr_uint,
    FATTR4_ACL()               => \&xdr_fattr4_acl,
    FATTR4_ACLSUPPORT()        => \&xdr_uint,
    FATTR4_ARCHIVE()           => \&xdr_uint,
    FATTR4_CANSETTIME()        => \&xdr_uint,
    FATTR4_CASE_INSENSITIVE()  => \&xdr_uint,
    FATTR4_CASE_PRESERVING()   => \&xdr_uint,
    FATTR4_CHOWN_RESTRICTED()  => \&xdr_uint,
    FATTR4_FILEHANDLE()        => \&xdr_fhandle,
    FATTR4_FILEID()            => \&xdr_uint64,
    FATTR4_FILES_AVAIL()       => \&xdr_uint64,
    FATTR4_FILES_FREE()        => \&xdr_uint64,
    FATTR4_FILES_TOTAL()       => \&xdr_uint64,
    FATTR4_FS_LOCATIONS()      => \&xdr_fs_locations4,
    FATTR4_FS_LOCATIONS_INFO() => \&xdr_fs_locations_info4,
    FATTR4_FS_STATUS()         => \&xdr_fs_status,
    FATTR4_HIDDEN()            => \&xdr_uint,
    FATTR4_HOMOGENEOUS()       => \&xdr_uint,
    FATTR4_MAXFILESIZE()       => \&xdr_uint64,
    FATTR4_MAXLINK()           => \&xdr_uint,
    FATTR4_MAXNAME()           => \&xdr_uint,
    FATTR4_MAXREAD()           => \&xdr_uint64,
    FATTR4_MAXWRITE()          => \&xdr_uint64,
    FATTR4_MIMETYPE()          => \&xdr_opaque,
    FATTR4_MODE()              => \&xdr_uint,
    FATTR4_NO_TRUNC()          => \&xdr_uint,
    FATTR4_NUMLINKS()          => \&xdr_uint,
    FATTR4_OWNER()             => \&xdr_opaque,
    FATTR4_OWNER_GROUP()       => \&xdr_opaque,
    FATTR4_QUOTA_AVAIL_HARD()  => \&xdr_uint64,
    FATTR4_QUOTA_AVAIL_SOFT()  => \&xdr_uint64,
    FATTR4_QUOTA_USED()        => \&xdr_uint64,
    FATTR4_RAWDEV()            => \&xdr_specdata,
    FATTR4_SPACE_AVAIL()       => \&xdr_uint64,
    FATTR4_SPACE_FREE()        => \&xdr_uint64,
    FATTR4_SPACE_TOTAL()       => \&xdr_uint64,
    FATTR4_SPACE_USED()        => \&xdr_uint64,
    FATTR4_SYSTEM()            => \&xdr_uint,
    FATTR4_TIME_ACCESS()       => \&xdr_nfstime,
    FATTR4_TIME_BACKUP()       => \&xdr_nfstime,
    FATTR4_TIME_CREATE()       => \&xdr_nfstime,
    FATTR4_TIME_DELTA()        => \&xdr_nfstime,
    FATTR4_TIME_METADATA()     => \&xdr_nfstime,
    FATTR4_TIME_MODIFY()       => \&xdr_nfstime,
    FATTR4_TIME_ACCESS_SET()   => \&xdr_settime,
    FATTR4_TIME_MODIFY_SET()   => \&xdr_settime,
    FATTR4_MOUNTED_ON_FILEID() => \&xdr_uint64,
    FATTR4_LAYOUT_TYPE()       => \&xdr_layouttype,
    FATTR4_SPACE_RESERVE()     => \&xdr_uint,
    FATTR4_LAYOUT_BLKSIZE()    => \&xdr_uint,
    FATTR4_LAYOUT_ALIGNMENT()  => \&xdr_uint,
    FATTR4_LAYOUT_MDSTHRESHOLD()  => \&xdr_mdsthreshold,
    FATTR4_SUPPATTR_EXCLCREAT()=> \&xdr_bitmap4,
    FATTR4_LAYOUT_HINT()       => \&xdr_layouthint
);

#---------------------------------------------------------------------

=head2 xdr_uint

  $out = xdr_uint( [ encode => $value ]
                   [, xdr   => $xdr_obj ] );

Encodes/decodes a 32bit unsigned integer 'uint32'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and a 32bit unsigned integer on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_uint(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            #encode => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
            encode => { type => SCALAR, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode 32bit unsigned integer
        $xdr->uint(encode=>$opts{encode});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return decoded 32bit unsigned integer
        return $opts{xdr}->uint();
    }
}

#---------------------------------------------------------------------

=head2 xdr_uint64

  $out = xdr_uint64( [ encode => $value ]
                     [, xdr   => $xdr_obj ] );

Encodes/decodes a 64bit unsigned integer 'uint64'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and a 64bit unsigned integer on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_uint64(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode 64bit unsigned integer
        $xdr->uint64(encode=>$opts{encode});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return decoded 64bit unsigned integer
        return $opts{xdr}->uint64();
    }
}

#---------------------------------------------------------------------

=head2 xdr_opaque

  $out = xdr_opaque( [ encode  => $string ]
                     [, xdr    => $xdr_obj ]
                     [, fixed  => 0|1 ]
                     [, maxlen => $maxlen ]
                     [, len    => $len ] );

Encodes/decodes a variable length opaque 'opaque<maxlen>'
or a fixed length opaque 'opaque[maxlen]'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=item fixed

(Optional) : Encode/decode a fixed length opaque: 0 or 1
(default is 0: a variable length opaque).

=item maxlen

(Optional) : Maximum number of bytes for a variable length opaque
or the number of bytes for a fixed length opaque
(default is 0: unlimited).

=item len

(Optional) : Force the length of the variable length opaque to this
number.

=back

=item Returns

The full XDR encoded message on encoding
and a string of bytes on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_opaque(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => SCALAR, optional => 1 },
            fixed  => { type => SCALAR,  default => 0, regex => qr/^[01]$/ },
            maxlen => { type => SCALAR,  default => 0, regex => qr/^\d+$/ },
            len    => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    # Set the type of opaque to encode/decode
    my $type = $opts{fixed} ? 'fixed' : 'variable';

    # Force the length for a variable length opaque
    my %len = ( varlen => $opts{len} ) if defined($opts{len});

    if ( defined($opts{encode}) ) {
        # Encode the opaque data
        $xdr->opaque(encode=>$opts{encode}, $type=>$opts{maxlen}, %len);
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded opaque data
        return $opts{xdr}->opaque($type=>$opts{maxlen});
    }
}

#---------------------------------------------------------------------

=head2 xdr_name

  $out = xdr_name( version   => $version
                   [, encode => $name ]
                   [, xdr    => $xdr_obj ] );

Encodes/decodes a NFS name.
For NFS version 2 the data type is 'opaque<MAXLEN>'
and for NFS version 3 and 4 it is 'opaque<>'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NFS version.

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS name on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_name(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT, optional => 1 },
            encode  => { type => SCALAR, optional => 1 },
            version => { type => SCALAR, regex => qr/^[234]$/ },
        }
    );
    if ( $opts{version} == 2 ) {
        # Set maxlen for NFSv2
        $opts{maxlen} = MAXNAMLEN;
    }

    # Delete the 'version' key and call xdr_opaque
    delete $opts{version};
    return xdr_opaque(%opts);
}

#---------------------------------------------------------------------

=head2 xdr_path

  $out = xdr_path( version   => $version
                   [, encode => $path ]
                   [, xdr    => $xdr_obj ] );

Encodes/decodes a NFS pathname.
For NFS version 2 the data type is 'opaque<MAXPATHLEN>'
and for NFS version 3 and 4 it is 'opaque<>'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NFS version.

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS pathname on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_path(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT, optional => 1 },
            encode  => { type => SCALAR, optional => 1 },
            version => { type => SCALAR, regex    => qr/^[234]$/ },
        }
    );
    if ( $opts{version} == 2 ) {
        # Set maxlen for NFSv2
        $opts{maxlen} = MAXPATHLEN;
    }

    # Delete the 'version' key and call xdr_opaque
    delete $opts{version};
    return xdr_opaque(%opts);
}

#---------------------------------------------------------------------

=head2 xdr_fhandle

  $out = xdr_fhandle( [ version => $version ]
                      [, encode => $fhandle ]
                      [, xdr    => $xdr_obj ] );

Encodes/decodes a NFS filehandle.
For NFS version 2 the data type is 'opaque[32]',
for NFS version 3 is 'opaque<64>'
and for NFS version 4 it is 'opaque<128>'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Optional) : NFS version (default is 4).

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS filehandle on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fhandle(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT, optional => 1 },
            encode  => { type => SCALAR, optional => 1 },
            version => { type => SCALAR, default  => 4, regex => qr/^[1234]$/ },
        }
    );
    my %options;
    my $ver = $opts{version};
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    # Set type and length of opaque
    if ( $ver == 2 ) {
        %options = (fixed=>32);
    } elsif ( $ver == 3 ) {
        %options = (variable=>64);
    } else {
        %options = (variable=>128);
    }

    if ( defined($opts{encode}) ) {
        # Encode the filehandle
        $xdr->opaque(encode=>$opts{encode}, %options);
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded filehandle
        return $xdr->opaque(%options);
    }
}

#---------------------------------------------------------------------

=head2 xdr_cookie

  $out = xdr_cookie( version   => $version
                     [, encode => $cookie ]
                     [, xdr    => $xdr_obj ] );

Encodes/decodes a NFS READDIR cookie.
For NFS version 2 the data type is 'uint32'
and for NFS version 3 and 4 it is 'uint64'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NFS version.

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS READDIR cookie on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cookie(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT, optional => 1 },
            encode  => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
            version => { type => SCALAR, regex    => qr/^[1234]$/ },
        }
    );
    my $ver = $opts{version};
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the NFS READDIR cookie
        ( $ver == 2 ) ? $xdr->uint(encode=>$opts{encode})
                      : $xdr->uint64(encode=>$opts{encode});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded NFS READDIR cookie
        return ( $ver == 2 ) ? $xdr->uint() : $xdr->uint64();
    }
}

#---------------------------------------------------------------------

=head2 xdr_cookieverf

  $out = xdr_cookieverf( [ encode => $cookieverf ]
                         [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS READDIR cookieverf.
The data type is 'opaque[8]'.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS READDIR cookieverf on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cookieverf(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => SCALAR, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the NFS READDIR cookieverf
        $xdr->opaque(encode=>$opts{encode}, fixed=>8);
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded NFS READDIR cookieverf
        return $xdr->opaque(fixed=>8);
    }
}

#---------------------------------------------------------------------

=head2 xdr_nfstime

  $out = xdr_nfstime( [ version => $version ]
                      [, encode => \%nfstime ]
                      [, xdr    => $xdr_obj ] );

Encodes/decodes a NFS time structure.
For NFS version 2 the data type is:

  { seconds  => uint32,
    useconds => uint32 }

for NFS version 3 it is:

  { seconds  => uint32,
    nseconds => uint32 }

and for NFS version 4 it is:

  { seconds  => uint64,
    nseconds => uint32 }

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Optional) : NFS version (default is 4).

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS time structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_nfstime(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT,  optional => 1 },
            encode  => { type => HASHREF, optional => 1 },
            version => { type => SCALAR,  default  => 4,
                                          regex    => qr/^[1234]$/ },
        }
    );
    my $ver = $opts{version};
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    # NFSv2 uses microseconds while NFSv3 and v4 use nanoseconds
    my $fseconds = ( $ver == 2 ) ? 'useconds' : 'nseconds';

    if ( defined($opts{encode}) ) {
        # Encode the seconds
        ( $ver == 4 ) ? $xdr->uint64(encode=>$opts{encode}{seconds})
                      : $xdr->uint(encode=>$opts{encode}{seconds});
        # Encode the micro|nano seconds
        $xdr->uint(encode=>$opts{encode}{$fseconds});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Decode the seconds
        my $seconds = ( $ver == 4 ) ? $xdr->uint64() : $xdr->uint();
        # Decode the micro|nano seconds and return time structure
        return { seconds => $seconds, $fseconds => $xdr->uint() };
    }
}

#---------------------------------------------------------------------

=head2 xdr_settime

  $out = xdr_settime( encode     => \%nfstime|SET_TO_SERVER_TIME
                      [, version => $version ]
                      [, xdr     => $xdr_obj ] );

Encodes a NFS time structure for NFSv3 and v4.
It includes the time_how flag to set the time to either
the server's time or the client's time, i.e. the passed
time structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NFS version.

=item xdr

(Optional) : Xdr object to encode to
(default is to create a new Xdr object).

=item encode

(Required) : Encode argument into XDR format.
If argument is undefined, the time is not modified.
If argument is the constant value SET_TO_SERVER_TIME, the
time is set to the server's time.
If argument is a time hash, SET_TO_CLIENT_TIME is used and
the time is set to the passed.

=back

=item Returns

The full XDR encoded message.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_settime(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT, optional => 1 },
            encode  => { type => HASHREF|SCALAR|UNDEF },
            version => { type => SCALAR,  default => 4, regex => qr/^[1234]$/ },
        }
    );
    my $ver = $opts{version};
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        if ( $opts{encode} == SET_TO_SERVER_TIME ) {
            # Set to server's time
            $xdr->uint(encode=>($ver==3)?1:0);
        } else {
            # Set to client's time using the passed time hash
            $xdr->uint(encode=>($ver==3)?2:1);
            xdr_nfstime(xdr=>$xdr, encode=>$opts{encode}, version=>$ver);
        }
    } else {
        # Do not change the time
        # On NFSv3, the time_how flag is 0
        $xdr->uint(encode=>0) if ($ver == 3);
    }

    # Return the encoded message
    return $xdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_fsid4

  $out = xdr_fsid4( [ encode => \%fsid4 ]
                    [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS fsid4 structure. The data type is:

  { major => uint64,
    minor => uint64 }

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS fsid4 structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fsid4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the fsid4 structure
        $xdr->uint64(encode=>$opts{encode}{major});
        $xdr->uint64(encode=>$opts{encode}{minor});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded fsid4 structure
        return { major => $xdr->uint64(), minor => $xdr->uint64() };
    }
}

#---------------------------------------------------------------------

=head2 xdr_specdata

  $out = xdr_specdata( [ encode => \%specdata ]
                       [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS specdata structure. The data type is:

  { specdata1 => uint32,
    specdata2 => uint32 }

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS specdata structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_specdata(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the specdata structure
        $xdr->uint(encode=>$opts{encode}{specdata1});
        $xdr->uint(encode=>$opts{encode}{specdata2});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded specdata structure
        return { specdata1 => $xdr->uint(), specdata2 => $xdr->uint() };
    }
}

#---------------------------------------------------------------------

=head2 xdr_change_info4

  $out = xdr_change_info4( [ encode => \%change_info4 ]
                           [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS change_info4 structure. The data type is:

  { atomic => uint32,
    before => uint64,
    after  => uint64 }

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS change_info4 structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_change_info4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the change_info4 structure
        $xdr->uint(encode=>$opts{encode}{atomic});
        $xdr->uint64(encode=>$opts{encode}{before});
        $xdr->uint64(encode=>$opts{encode}{after});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded change_info4 structure
        return { atomic => $xdr->uint(),
                 before => $xdr->uint64(),
                 after  => $xdr->uint64() };
    }
}

#---------------------------------------------------------------------

=head2 xdr_nfsace4

  $out = xdr_nfsace4( [ encode => \%nfsace4 ]
                      [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS nfsace4 structure. The data type is:

  { type => uint32,
    flag => uint32,
    mask => uint32,
    who  => opaque<> }

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS nfsace4 structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_nfsace4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode nfsace4 structure
        $xdr->uint(encode=>$opts{encode}->{type});
        $xdr->uint(encode=>$opts{encode}->{flag});
        $xdr->uint(encode=>$opts{encode}->{mask});
        $xdr->opaque(encode=>$opts{encode}->{who}, variable=>0);
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return decoded nfsace4 structure
        return {
            type => $xdr->uint(),
            flag => $xdr->uint(),
            mask => $xdr->uint(),
            who  => $xdr->opaque(variable=>0),
        };
    }
}

#---------------------------------------------------------------------

=head2 xdr_fattr4_acl

  $out = xdr_fattr4_acl( [ encode => \@fattr4_acl ]
                         [, xdr   => $xdr_obj ] );

Encodes/decodes a variable length array of NFS nfsace4 structures.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and a reference to an array of NFS nfsace4 structures on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fattr4_acl(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,   optional => 1 },
            encode => { type => ARRAYREF, optional => 1 },
        }
    );
    my @aces;
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode number of aces
        my $count = @{$opts{encode}};
        $Log->debug(qq{count = $count, aces = } . Dumper($opts{encode}));
        $xdr->uint(encode=>$count);

        # Encode each nfsace4 structure
        foreach my $ace ( @{$opts{encode}} ) {
            if ($ace) {
                xdr_nfsace4(encode=>$ace, xdr=>$xdr);
            }
        }
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Decode number of aces
        foreach ( 1..$xdr->uint() ) {
            # Decode each nfsace4 structure
            push @aces, xdr_nfsace4(xdr=>$xdr);
        }
        # Return decoded array of nfsace4 structures
        return \@aces;
    }
}

#---------------------------------------------------------------------

=head2 xdr_bitmap4

  $out = xdr_bitmap4( [ encode => \@bitmap4 ]
                      [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS bitmap4 data type.
The data type is 'uint32<>' (variable length array of uint32).

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS bitmap4 data type on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_bitmap4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,   optional => 1 },
            encode => { type => ARRAYREF, optional => 1 },
            count  => { type => SCALAR, optional => 1 },
        }
    );
    my @bitmap;
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode number of bitmap4 integers
        @bitmap = @{$opts{encode}};
        $xdr->uint(encode=>scalar(@bitmap));
        # Encode each bitmap4 integers
        foreach my $mask ( @bitmap ) {
            $xdr->uint(encode=>$mask);
        }
        # Return the encoded message
        return $xdr->{message};
    } else {
    #my $mesg = $xdr->hex_dump();
        # Decode number of bitmap4 integers
    if (defined($opts{count})) {
    foreach ( 1..$opts{count} ) {
            push @bitmap, $xdr->uint();
           }
    } else {
        foreach ( 1..$xdr->uint() ) {
            # Decode each bitmap4 integer
            push @bitmap, $xdr->uint();
        }
     }
        # Return the decoded bitmap4 data type
        return \@bitmap;
    }
}

#---------------------------------------------------------------------

=head2 xdr_stateid4

  $out = xdr_stateid4( [ encode => \%stateid|STATEID0|STATEID1 ]
                       [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS stateid4 structure.
The stateid4 structure is a hash containing the following:

  { seqid => uint32,
    other => opaque[12] }

In order to encode the special stateid of all 0's or all 1's,
the constants STATEID0 or STATEID1 can be used respectively
instead of the hash.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS stateid4 structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_stateid4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => HASHREF|SCALAR, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        if ( $opts{encode} == STATEID0 ) {
            # Special stateid of all 0's
            $opts{encode} = {seqid=>0, other=>''};
        } elsif ( $opts{encode} == STATEID1 ) {
            # Special stateid of all 1's
            $opts{encode} = {seqid=>0xFFFFFFFF, other=>"\xff"x12};
        }
        # Encode the stateid structure
        $xdr->uint(encode=>$opts{encode}{seqid});
        $xdr->opaque(encode=>$opts{encode}{other}, fixed=>12);
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded stateid structure
        return { seqid => $xdr->uint(), other => $xdr->opaque(fixed=>12) };
    }
}

#---------------------------------------------------------------------

=head2 xdr_fs_locations4

  $out = xdr_fs_locations4( [ encode => \@fs_locations4 ]
                            [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS fs_locations4 structure. The data type is:

  { fs_root   => opaque<>,
    locations => fs_location4<> }

where fs_location4 data type is:

  { server   => opaque<>,
    rootpath => opaque<> }

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS fs_locations4 structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fs_locations4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,   optional => 1 },
            encode => { type => ARRAYREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode fs_root
        $xdr->opaque(encode=>$opts{encode}{fs_root}, variable=>0);
        # Encode number of locations
        $xdr->uint(encode=>scalar @{$opts{encode}{locations}});
        foreach my $loc ( @{$opts{encode}{locations}} ) {
            # Encode each fs_location4 structure
            $xdr->opaque(encode=>$loc->{server},   variable=>0);
            $xdr->opaque(encode=>$loc->{rootpath}, variable=>0);
        }
        # Return the encoded message
        return $xdr->{message};
    } else {
        my %fs_locations;
    # Migration fs_locations variables

    @{$fs_locations{fs_root}} = ();
    @{$fs_locations{servers}} = ();

    # Decode fs_root
    my $number_of_fs_rootpaths = $xdr->uint();
    $Log->debug("\n-----------number of fs_rootpaths: $number_of_fs_rootpaths \n");

    foreach ( 1..$number_of_fs_rootpaths ) {
        push (@{$fs_locations{fs_root}},xdr_opaque(xdr=>$xdr, fixed=>0));
    }
        $Log->debug("fs_root path: @{$fs_locations{fs_root}} \n");

    my $number_of_locations = xdr_uint(xdr=>$xdr);
    $Log->debug("num of locations: $number_of_locations \n");
    foreach ( 1..$number_of_locations ) {
    
        my $number_of_servers = xdr_uint(xdr=>$xdr);
        $Log->debug("num of servers: $number_of_servers \n");
        foreach ( 1..$number_of_servers ) {
            push (@{$fs_locations{servers}},xdr_opaque(xdr=>$xdr, fixed=>0));
        }
        $Log->debug("servers: @{$fs_locations{servers}} \n");
    
        # Decode rootpath
        my $number_of_root_path_components = $xdr->uint();
            $Log->debug("   num of rootpath components $number_of_root_path_components \n");
        foreach ( 1..$number_of_root_path_components ) {
            push (@{$fs_locations{rootpath}},xdr_opaque(xdr=>$xdr, fixed=>0));
        }
            $Log->debug("   rootpath: @{$fs_locations{rootpath}} \n");
    }
    ############## END MODIFIED CODE ##############


        # Return the decoded fs_locations4 structure
        return \%fs_locations;
    }
}

#---------------------------------------------------------------------

=head2 xdr_fattr4

  $out = xdr_fattr4( [ encode => \%fattr4 ]
                     [, xdr   => $xdr_obj ] );

Encodes/decodes a NFS fattr4 structure. The data type is:

  { attrmask  => bitmap4,
    attr_vals => attrlist4 }

where attrlist4 data type is:

  { <attribute_number> => <attribute_value> }

Since the attrmask can be created using attrlist4,
on encoding it is only required to pass the attr_vals structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and an NFS fattr4 structure on decoding.

=item Examples

  # Encode a fattr4 to modify the size and the mode
  # Need to pass only the attr_vals structure
  $enc_mesg = xdr_fattr4(encode=>{FATTR4_SIZE()=>20,
                                  FATTR4_MODE()=>0644});

  # Decode a fattr4 after getting only the size and mode attributes
  # Decoding returns both the attrmask and attr_vals
  # The output is $fattr = { attrmask => [16,2],
  #                          attr_vals => { FATTR4_SIZE()=>20,
  #                                         FATTR4_MODE()=>0644 } }
  $fattr = xdr_fattr4(xdr=>$xdr);

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fattr4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    my $att = $opts{encode};

    if ( defined($att) ) {
        # Create list of bits in ascending order
        my @bits = sort {$a <=> $b} keys %{$att};

        # Convert list of attributes to bitmap4
        my $attrmask = bitmap4_convert(attrlist=>\@bits);

        # Add the bitmap4
        xdr_bitmap4(xdr=>$xdr, encode=>$attrmask);

        # Create an Xdr object to encode all attributes first
        # since all attributes must be encoded as a variable opaque
        my $axdr = nfs::Xdr->new();

        foreach my $bit ( @bits ) {
            if ( ! $nfsv4_attr{$bit} ) {
                croak "xdr_fattr4: attribute <$bit> not supported";
            }
            # Call the proper encoding function for given bit
            &{$nfsv4_attr{$bit}}(xdr=>$axdr, encode=>$att->{$bit});
        }

        # Add the attributes to be set
        $xdr->opaque(encode=>$axdr->{message}, variable=>0);

        # Return the encoded message
        return $xdr->{message};
    } else {
        my %fattr;
        # Decoding fattr4
        $fattr{attrmask} = xdr_bitmap4(xdr=>$xdr);

        # Decode the size of all attributes
        $xdr->uint();

        # Convert bitmap4 to list of attributes
        my $attrlist = bitmap4_convert(attrmask=>$fattr{attrmask});

        # Decode all attributes
        foreach my $bit ( @{$attrlist} ) {
            if ( ! $nfsv4_attr{$bit} ) {
                croak "xdr_fattr4: attribute <$bit> not supported";
            }
            # Call the proper decoding function for given bit
            $fattr{attr_vals}{$bit} = &{$nfsv4_attr{$bit}}(xdr=>$xdr);
        }

        # Return the decoded attributes
        return \%fattr;
    }
}

#---------------------------------------------------------------------

=head2 xdr_sattr

  $out = xdr_sattr( encode => \%sattr, version => $version );

Encodes a NFS sattr structure.
The data type is { <attribute> => <value> }.
For NFSv2 and v3, the list of valid attributes is:
'mode', 'uid', 'gid', 'size', 'atime' and 'mtime'.
For NFSv4, the list of valid attributes is any FATTR4_*
attribute which can be set, e.g., FATTR4_SIZE, FATTR4_MODE, etc.

On NFSv4, the sattr structure requires a stateid4 structure to
be encoded with the fattr4 structure, the stateid4 structure
must be included within the fattr4 structure. If the stateid
key is not specified, the special stateid of all 0's is used.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NFS version.

=item encode

(Required) : Encode argument into XDR format.

=back

=item Returns

The full XDR encoded message.

=item Examples

  # Encode sattr on NFSv3 to change the size and the mode:
  $sattr = xdr_sattr(encode=>{size=>20, mode=>0644}, version=>3);

  # Encode sattr on NFSv4 to change the size, using default stateid:
  $sattr = xdr_sattr(encode=>{FATTR4_SIZE()=>20}, version=>4);

  # Encode sattr on NFSv4 to change the size, using a valid stateid:
  $sattr = xdr_sattr(encode=>{FATTR4_SIZE()=>20, stateid=>$stateid},
                     version=>4);

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_sattr(%) {
    my (%opts) = validate (
        @_,
        {
            encode  => { type => HASHREF },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my $xdr = nfs::Xdr->new();
    my $att = $opts{encode};
    my $ver = $opts{version};
    my $neg_one = 0xFFFFFFFF;

    if ( $ver == 2 ) {
        # For NFSv2 a value of -1 means that
        # the attribute is not to be changed
        my $time_no_change = { seconds => $neg_one, useconds => $neg_one };
        # Encode the mode
        $xdr->uint(encode=>defined($att->{mode}) ? $att->{mode} : $neg_one);
        # Encode the uid
        $xdr->uint(encode=>defined($att->{uid})  ? $att->{uid}  : $neg_one);
        # Encode the gid
        $xdr->uint(encode=>defined($att->{gid})  ? $att->{gid}  : $neg_one);
        # Encode the size
        $xdr->uint(encode=>defined($att->{size}) ? $att->{size} : $neg_one);
        # Encode the atime structure
        my $atime = $att->{atime} ? $att->{atime} : $time_no_change;
        xdr_nfstime(xdr=>$xdr, encode=>$atime, version=>$ver);
        # Encode the mtime structure
        my $mtime = $att->{mtime} ? $att->{mtime} : $time_no_change;
        xdr_nfstime(xdr=>$xdr, encode=>$mtime, version=>$ver);
    } elsif ( $ver == 3 ) {
        # For NFSv3 encode a boolean before the attribute.
        # If the boolean is true, the attribute is modified
        # by the value that follows, otherwise the attribute
        # is left intact
        # Encode the mode
        if ( defined($att->{mode}) ) {
            $xdr->uint(encode=>1);
            $xdr->uint(encode=>$att->{mode});
        } else {
            $xdr->uint(encode=>0);
        }
        # Encode the uid
        if ( defined($att->{uid}) ) {
            $xdr->uint(encode=>1);
            $xdr->uint(encode=>$att->{uid});
        } else {
            $xdr->uint(encode=>0);
        }
        # Encode the gid
        if ( defined($att->{gid}) ) {
            $xdr->uint(encode=>1);
            $xdr->uint(encode=>$att->{gid});
        } else {
            $xdr->uint(encode=>0);
        }
        # Encode the size
        if ( defined($att->{size}) ) {
            $xdr->uint(encode=>1);
            $xdr->uint64(encode=>$att->{size});
        } else {
            $xdr->uint(encode=>0);
        }
        # Encode the atime structure
        xdr_settime(xdr=>$xdr, encode=>$att->{atime}, version=>$ver);
        # Encode the mtime structure
        xdr_settime(xdr=>$xdr, encode=>$att->{mtime}, version=>$ver);
    } else {
        # For NFSv4 encode the stateid, followed by
        # the list of attributes to be modified
        # Encode the NFSv4 stateid which is passed as
        # part of the attribute list or use the default:
        my $stateid = STATEID0;
        if ( exists($att->{stateid}) ) {
            $stateid = $att->{stateid} if defined($att->{stateid});
            delete $att->{stateid};
        }
        # Encode the stateid4 structure
        xdr_stateid4(xdr=>$xdr, encode=>$stateid);
        # Encode the fattr4 list
        xdr_fattr4(xdr=>$xdr, encode=>$att);
    }

    # Return the encoded message
    return $xdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_fattr

  $out = xdr_fattr( xdr => $xdr_obj, version => $version );

Decodes a NFS fattr structure.
For NFS version 2 the data type is:

  { type      => uint32,
    mode      => uint32,
    nlink     => uint32,
    uid       => uint32,
    gid       => uint32,
    size      => uint32,
    blocksize => uint32,
    rdev      => uint32,
    blocks    => uint32,
    fsid      => uint32,
    fileid    => uint32,
    atime     => nfstime,
    mtime     => nfstime,
    ctime     => nfstime }

for NFS version 3 is:

  { type   => uint32,
    mode   => uint32,
    nlink  => uint32,
    uid    => uint32,
    gid    => uint32,
    size   => uint64,
    used   => uint64,
    rdev   => specdata,
    fsid   => uint64,
    fileid => uint64,
    atime  => nfstime,
    mtime  => nfstime,
    ctime  => nfstime }

and for NFS version 4 it is the NFS fattr4 structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS fattr structure

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fattr(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %fattr;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    if ( $ver == 2 ) {
        # Decode the fattr structure
        %fattr = (
            type      => $xdr->uint(),
            mode      => $xdr->uint(),
            nlink     => $xdr->uint(),
            uid       => $xdr->uint(),
            gid       => $xdr->uint(),
            size      => $xdr->uint(),
            blocksize => $xdr->uint(),
            rdev      => $xdr->uint(),
            blocks    => $xdr->uint(),
            fsid      => $xdr->uint(),
            fileid    => $xdr->uint(),
            atime     => xdr_nfstime(xdr=>$xdr, version=>$ver),
            mtime     => xdr_nfstime(xdr=>$xdr, version=>$ver),
            ctime     => xdr_nfstime(xdr=>$xdr, version=>$ver)
        );
    } elsif ( $ver == 3 ) {
        # Decode the fattr3 structure
        %fattr = (
            type   => $xdr->uint(),
            mode   => $xdr->uint(),
            nlink  => $xdr->uint(),
            uid    => $xdr->uint(),
            gid    => $xdr->uint(),
            size   => $xdr->uint64(),
            used   => $xdr->uint64(),
            rdev   => xdr_specdata(xdr=>$xdr),
            fsid   => $xdr->uint64(),
            fileid => $xdr->uint64(),
            atime  => xdr_nfstime(xdr=>$xdr, version=>$ver),
            mtime  => xdr_nfstime(xdr=>$xdr, version=>$ver),
            ctime  => xdr_nfstime(xdr=>$xdr, version=>$ver)
        );
    } elsif ( $ver == 4 ) {
        # Return the decoded fattr4 structure
        return xdr_fattr4(xdr=>$xdr);
    }

    # Return the fattr structure (NFSv2 and v3)
    return \%fattr;
}

#---------------------------------------------------------------------

=head2 xdr_lock4_denied

  $out = xdr_lock4_denied( xdr => $xdr_obj );

Decodes the NFSv4 LOCK4denied structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv4 LOCK4denied structure.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_lock4_denied(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    my %lock4denied = (
        offset => $xdr->uint64(),
        count  => $xdr->uint64(),
        type   => $xdr->uint(),
        owner  => { clientid => $xdr->opaque(fixed=>8),
                    owner    => $xdr->opaque(variable=>1024) },
    );

    # Return the LOCK4denied structure
    return \%lock4denied;
}

#---------------------------------------------------------------------

=head2 xdr_pre_op_attr

  $pre_op_attr = xdr_pre_op_attr( xdr => $xdr_obj );

Decodes the NFSv3 pre_op_attr structure.
This is the subset of preoperation attributes needed to better
support the weak cache consistency semantics.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 pre_op_attr structure if available.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_pre_op_attr(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Check boolean for existence of pre_op_attr
    if ( $xdr->uint() ) {
        # Return the decoded pre_op_attr structure
        return {
            size  => $xdr->uint64(),
            mtime => xdr_nfstime(xdr=>$xdr, version=>3),
            ctime => xdr_nfstime(xdr=>$xdr, version=>3),
        };
    }

    return {};
}

#---------------------------------------------------------------------

=head2 xdr_post_op_attr

  $post_op_attr = xdr_post_op_attr( xdr => $xdr_obj );

Decodes the NFSv3 post_op_attr structure.
This structure is used for returning attributes in operations
that are not directly involved with manipulating attributes.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 post_op_attr structure if available.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_post_op_attr(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Check boolean for existence of post_op_attr
    if ( $xdr->uint() ) {
        # Decode the fattr structure
        return xdr_fattr(xdr=>$xdr, version=>3);
    }

    return {};
}

#---------------------------------------------------------------------

=head2 xdr_post_op_fh3

  $post_op_fh3 = xdr_post_op_fh3( xdr => $xdr_obj );

Decodes the NFSv3 post_op_fh3 structure.
This structure is designed to allow the server to recover
from errors encountered while constructing a filehandle.
It is used to return a filehandle from the CREATE, MKDIR,
SYMLINK, MKNOD and READDIRPLUS requests.
The filehandle is returned as an optimization so that the
client is not forced to immediately issue a LOOKUP request
to get the filehandle.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 post_op_fh3 structure if available.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_post_op_fh3(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Check boolean for existence of post_op_fh3
    if ( $xdr->uint() ) {
        # Decode the filehandle
        return xdr_fhandle(xdr=>$xdr, version=>3);
    }

    return;
}

#---------------------------------------------------------------------

=head2 xdr_wcc_data

  $wcc_data = xdr_wcc_data( xdr => $xdr_obj );

Decodes the NFSv3 wcc_data structure.
This structure consists of certain key fields from the object
attributes before the operation, together with the object
attributes after the operation.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 wcc_data structure

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_wcc_data(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Return the decoded wcc_data structure
    return {
        # Decode pre_op_attr
        before => xdr_pre_op_attr(xdr=>$xdr),
        # Decode post_op_attr
        after  => xdr_post_op_attr(xdr=>$xdr),
    };
}

#---------------------------------------------------------------------

=head2 xdr_nlm_lock

  $out = xdr_nlm_lock( version   => $version
                       [, xdr    => $xdr_obj ]
                       [, encode => \%nlm_lock ] );

Encodes the nlm_lock structure.
This structure identifies a particular lock. It is used by all five
basic NLM locking procedures: TEST, LOCK, CANCEL, UNLOCK and GRANTED.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : NLM version.

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Required) : Encode argument into XDR format.

=back

=item Returns

The full XDR encoded message or
the NLM lock structure on decode.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_nlm_lock(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT,  optional => 1 },
            encode  => { type => HASHREF, optional => 1 },
            version => { type => SCALAR,     regex => qr/^[1234]$/ },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    my $nlm = $opts{encode};
    my $ver = $opts{version};
    my %ret;

    if ( defined($opts{encode}) ) {
        # Encode the caller's name
        $xdr->opaque(encode=>$nlm->{caller_name}, variable=>1024);
        # Encode the file handle
        $xdr->opaque(encode=>$nlm->{fh}, variable=>1024);
        # Encode the lock owner
        $xdr->opaque(encode=>$nlm->{oh}, variable=>1024);
        # Encode the unique process id
        $xdr->uint(encode=>$nlm->{svid});

        # Encode byte range
        if ( $ver == 4 ) {
            # Encode offset
            $xdr->uint64(encode=>$nlm->{l_offset});
            # Encode extent
            $xdr->uint64(encode=>$nlm->{l_len});
        } else {
            # Encode offset
            $xdr->uint(encode=>$nlm->{l_offset});
            # Encode extent
            $xdr->uint(encode=>$nlm->{l_len});
        }

        # Return the encoded message
        return $xdr->{message};
    } else {
        # Decode:
        %ret = ( caller_name => $xdr->opaque(), # caller's name
                 fh          => $xdr->opaque(), # file handle
                 oh          => $xdr->opaque(), # lock owner
                 svid        => $xdr->uint() ); # unique process id
        if ( $ver == 4 ) {
            # Decode offset
            $ret{l_offset} = $xdr->uint64();
            # Decode extent
            $ret{l_len}    = $xdr->uint64();
        } else {
            # Decode offset
            $ret{l_offset} = $xdr->uint();
            # Decode extent
            $ret{l_len}    = $xdr->uint();
        }

        # Return the nlm_lock structure
        return \%ret;
    }
}

#---------------------------------------------------------------------

=head2 xdr_compound

  $cmpd_args = xdr_compound( tag      => $tag,
                             ops      => \@ops
                             [, num   => $num ]
                             [, minor => $minor ] );

Encodes the arguments for COMPOUND procedure on NFSv4.
This procedure is used to combine one or more of the NFSv4
operations into a single RPC request.

The following named arguments are accepted:

=over

=item Arguments

=over

=item tag

(Required) : Tag given to the compound request.

=item ops

(Required) : A reference to an array of hashes of NFSv4 operations
with its arguments: { op=><operation_number>, args=><arguments> }.

=item num

(Optional) : Use this as the number of operations to encode instead
of the actual number of operations encoded.

=item minor

(Optional) : Minor version. Default is 0.

=back

=item Returns

The encoded compound request arguments.

=item Examples

  # Compound request for LOOKUP operation
  # using the following operations:
  my @operations = ( { op => OP_PUTFH,  args => $fh },
                     { op => OP_LOOKUP, args => $name },
                     { op => OP_GETFH } );
  $cmpd_args = xdr_compound(tag=>'lookup', ops=>\@operations);

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_compound(%) {
    my (%opts) = validate (
        @_,
        {
            tag   => { type => SCALAR },
            ops   => { type => ARRAYREF },
            num   => { type => SCALAR, optional => 1 },
            minor => { type => SCALAR,  default => 0 },
        }
    );
    $opts{tag} = "PSC:$opts{tag}";
    my $xdr = nfs::Xdr->new();
    my @ops = @{$opts{ops}};
    my $num = defined $opts{num} ? $opts{num} : scalar(@ops);

    # Encode the tag for this compound request
    $xdr->opaque(encode=>$opts{tag}, variable=>0);
    $xdr->uint(encode=>$opts{minor}); # Minor version
    $xdr->uint(encode=>$num);         # Number of operations

    # Append each operation with its arguments
    foreach my $op ( @ops ) {
        # Encode the operation number
        $xdr->uint(encode=>$op->{op});
        # Append the already encoded arguments for this operation
        $xdr->{message} .= $op->{args}
          if (defined($op->{args}) && length($op->{args}));
    }

    # Return the encoded message
    return $xdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_compound_res

  $results = xdr_compound_res( xdr => $xdr_obj );

Decodes the results for COMPOUND procedure on NFSv4.
The operation functions must take the xdr and version options
as arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

The compound request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_compound_res(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};
    my ($op, $len, $result);

    # Decode the compound request status,
    # the tag and the number of results
    my %ret = (
        status => $xdr->uint(),
        tag    => $xdr->opaque(variable=>0),
        nres   => $xdr->uint(),
    );

    # Decode all expected results
    foreach ( 1..$ret{nres} ) {
        # Get number of bytes remaining in the result
        $len = $xdr->remain();
        # Decode operation number for this result
        $op = $xdr->uint();
        # Decode operation results
        $result = $nfsv4_res{$op} ? &{$nfsv4_res{$op}}(xdr=>$xdr, version=>4)
                                  : xdr_status(xdr=>$xdr, version=>4);
        # Append the next result
        push @{$ret{results}}, { op => $op, result => $result };
        # Get total number of bytes received for this operation
        $len -= $xdr->remain();
        push @{$ret{result_sizes}}, { op => $op, size => $len };
    }

    # Return compound request results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_cb_compound

  $cmpd_args = xdr_cb_compound( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Decodes the arguments for the callback compound procedure
on NFSv4. Then calls the corresponding function for the
operation specified in the arguments. The operation
function results are then encoded into XDR and returned.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments and return a reference to a hash
containing the following keys: status => status of operation
and result => the XDR encoded results of operation.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback compound request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_compound(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    my @ret;
    my $xdr = $opts{xdr};

    # Decode arguments to callback compound procedure
    my $tag = $xdr->opaque(variable=>0);
    # Decode minor and callback identifier
    $xdr->uint();
    $xdr->uint();
    # Decode the number of operations in compound request
    my $nops = $xdr->uint();

    # Decode all arguments
    foreach ( 1..$nops ) {
        # Decode operation number for this argument
        my $op = $xdr->uint();
        # Call the corresponding callback
        # operation and save the output
        push @ret, {
            op  => $op,
            out => &{$nfsv4_cb{$op}}(xdr=>$xdr),
        };
    }

    # Encode operation results
    my $exdr = nfs::Xdr->new();
    # Encode status of the last operation,
    # the same tag as the call and the
    # number of results
    $exdr->uint(encode=>$ret[-1]{out}{status});
    $exdr->opaque(encode=>$tag, variable=>0);
    $exdr->uint(encode=>scalar(@ret));

    # Encode each operation result
    foreach my $out ( @ret ) {
        # Encode the operation number,
        # its status and results
        my $op = $exdr->uint(encode=>$out->{op});

        $Log->debug("In NfsXdr::xdr_cb_compound: op is $op");

        $exdr->uint(encode=>$out->{out}{status});
        my $result = $out->{out}{result};
        if ( defined($result) and length($result) ) {
            $exdr->opaque(encode=>$result, fixed=>length($result));
        }
    }

    return $exdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_cb_recall

  $out = xdr_cb_recall( xdr => $xdr_obj );

Process the callback recall operation on NFSv4.
Decodes the arguments and checks if the filehandle and stateid
of delegation to be recalled is one of the currently opened
delegations. It does not initiate a delegation recall, it only
acknowledges the recall message from the server.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

Reference to a hash of results. Only the status is returned.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_recall(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Decode arguments for CB_RECALL
    my %args = (
        stateid  => xdr_stateid4(xdr=>$xdr),
        truncate => $xdr->uint(),
        fhandle  => xdr_fhandle(xdr=>$xdr, version=>4),
    );

    my $index = 0;
    my $found = 0;
    # Check for existence of delegation to be recalled
    foreach my $deleg ( @open_delegs ) {
        if ( $args{fhandle} eq $deleg->{fhandle} ) {
            # Filehandle is found on
            # list of open delegations
            $found = 1;
            last;
        }
        ++$index;
    }

    if ( ! $found ) {
        # It is not a valid open delegation filehandle
        return { status => NFS4ERR_BADHANDLE };
    }

    if ( $args{stateid}{other} ne $open_delegs[$index]{stateid}{other}
      or $args{stateid}{seqid} ne $open_delegs[$index]{stateid}{seqid} ) {
        # Not a valid open delegation stateid
        return { status => NFS4ERR_BAD_STATEID };
    }

    # Queue delegation recall
    push @cb_queue, {
        op   => OP_CB_RECALL,
        args => \%args
    };

    return { status => NFS4_OK };
}

#---------------------------------------------------------------------

=head2 xdr_cb_nlm_granted

  $res = xdr_cb_nlm_granted( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Decodes the arguments for the NLM callback on NFSv2 and v3.
Then adds the information into a callback queue to be processed
by the client.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_nlm_granted(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    my $xdr = $opts{xdr};
    my $ver = $opts{rpc_hdr}{version};

    # Decode the cookie
    my $cookie = $xdr->opaque();
    # Decode the exclusive flag
    my $exclusive = $xdr->uint();
    # Decode the nlm_lock structure
    my $nlm_lock = xdr_nlm_lock(xdr=>$xdr, version=>$ver);

    # Queue NLM granted
    push @nlm_cb_queue, {
        cookie    => $cookie,
        exclusive => $exclusive,
        nlm_lock  => $nlm_lock,
    };

    # Encode operation results
    my $exdr = nfs::Xdr->new();

    # Encode the cookie
    $exdr->opaque(encode=>$cookie);
    # Encode the status
    $exdr->uint(encode=>0);

    return $exdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_cb_nsm_notify

  $res = xdr_cb_nsm_notify( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Decodes the arguments for the NSM callback on NFSv2 and v3.
Then adds the information into a callback queue to be processed
by the client.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_nsm_notify(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    my $xdr = $opts{xdr};

    # Decode name of the host from which the request is sent
    my $mon_name = $xdr->opaque();
    # Decode the new state number of the host
    my $state = $xdr->uint();

    # Queue NSM notify
    push @nsm_cb_queue, {
        mon_name => $mon_name,
        state    => $state,
        list     => {},
    };

    # Return void result
    return;
}

#---------------------------------------------------------------------

=head2 xdr_cb_port

  $res = xdr_cb_port( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Save the port number so it can be unregistered later if needed.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_port(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    my $xdr     = $opts{xdr};
    my $proto   = $opts{rpc_hdr}{proto};
    my $version = $opts{rpc_hdr}{version};

    # Decode port number
    my $port = $xdr->uint();

    # Save on info structure
    push @{$daemon_info->{ports}}, { proto   => $proto,
                                     version => $version,
                                     port    => $port,
                                     restore => 0 };

    # Check if original program needs to be restored
    foreach my $item ( @$daemon_reg_ports ) {
        if ( $item->{program} == $daemon_program
          and $item->{version} == $version and $item->{proto} eq $proto ) {
            push @{$daemon_info->{ports}}, { proto   => $proto,
                                             version => $version,
                                             port    => $item->{port},
                                             restore => 1 };
            last;
        }
    }

    return;
}

#---------------------------------------------------------------------

=head2 xdr_cb_register

  $res = xdr_cb_register( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Register port with the daemon.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_register(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    my $res = 0;
    my $xdr = $opts{xdr};

    # Decode the client's id
    my $id = $xdr->opaque();
    # Decode the client's port number
    my $port = $xdr->uint();

    # Add client
    if ( defined($daemon_clients->{$id}) ) {
        $res = 1;
    } else {
        $daemon_clients->{$id} = { port    => $port,
                                   proto   => $opts{rpc_hdr}{proto},
                                   version => $opts{rpc_hdr}{version} };
    }

    # Encode operation results
    my $exdr = nfs::Xdr->new();

    # Encode the status
    $exdr->uint(encode=>$res);

    return $exdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_cb_pid

  $res = xdr_cb_pid( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Returns the process id of the daemon.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_pid(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    # Add client
    $daemon_info->{used}++;

    # Encode operation results
    my $exdr = nfs::Xdr->new();

    # Encode the pid
    $exdr->uint(encode=>$$);

    return $exdr->{message};
}

#---------------------------------------------------------------------

=head2 xdr_cb_exit

  $res = xdr_cb_exit( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

The client is no longer using the daemon.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_exit(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );
    # Remove client
    $daemon_info->{used}--;

    return xdr_daemon_info(encode=>$daemon_info);
}

#---------------------------------------------------------------------

=head2 xdr_cb_info

  $res = xdr_cb_info( xdr => $xdr_obj, rpc_hdr => \%rpc_hdr );

Returns the daemon info structure.

The callback operation functions must take the xdr and rpc_hdr
options as the only arguments.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item rpc_hdr

(Required) : RPC header.

=back

=item Returns

The encoded callback request results.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_cb_info(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            rpc_hdr => { type => HASHREF },
        }
    );

    return xdr_daemon_info(encode=>$daemon_info);
}

#---------------------------------------------------------------------

=head2 xdr_daemon_info

  $out = xdr_daemon_info( [ encode => \%daemon_info ]
                          [, xdr => $xdr_obj ] );

Encodes/decodes a daemon info structure.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Optional) : Xdr object to decode/encode from/to
(default is to create a new Xdr object).

=item encode

(Optional) : Encode argument into XDR format
(default is to decode from the Xdr object provided).

=back

=item Returns

The full XDR encoded message on encoding
and a daemon info structure on decoding.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_daemon_info(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,  optional => 1 },
            encode => { type => HASHREF, optional => 1 },
        }
    );
    my ($used, $udp, $tcp, @ports, $count, $item);
    my $xdr = defined($opts{xdr}) ? $opts{xdr} : nfs::Xdr->new();
    my $prt = $opts{encode};

    if ( defined($prt) ) {
        # Encode the in-use flag
        $xdr->uint(encode=>$prt->{used});
        # Encode the udp port
        $xdr->uint(encode=>$prt->{udp});
        # Encode the tcp port
        $xdr->uint(encode=>$prt->{tcp});

        # Encode list of ports
        $count = scalar @{$prt->{ports}};
        $xdr->uint(encode=>$count);
        foreach $item ( @{$prt->{ports}} ) {
            $xdr->opaque(encode=>$item->{proto});
            $xdr->uint(encode=>$item->{version});
            $xdr->uint(encode=>$item->{restore});
            $xdr->uint(encode=>$item->{port});
        }

        return $xdr->{message};
    } else {
        # Decode the in-use flag
        $used = $xdr->uint();
        # Decode the udp port
        $udp = $xdr->uint();
        # Decode the tcp port
        $tcp = $xdr->uint();

        # Decode list of ports
        $count = $xdr->uint();
        foreach ( 1..$count ) {
            push @ports, { proto   => $xdr->opaque(),
                           version => $xdr->uint(),
                           restore => $xdr->uint(),
                           port    => $xdr->uint() };
        }

        return { used  => $used,
                 udp   => $udp,
                 tcp   => $tcp,
                 ports => \@ports };
    }
}

#---------------------------------------------------------------------

=head2 xdr_status

  $out = xdr_status( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 status.
Used for NFSv4 operation which returns only the status.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

A hash reference having only the NFSv4 status.

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_status(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    # Return the decoded NFS status
    return {status => $opts{xdr}->uint()};
}

#---------------------------------------------------------------------

=head2 xdr_access

  $out = xdr_access( xdr => $xdr_obj, version => $version );

Decodes the NFS ACCESS results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS ACCESS results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_access(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $opts{version} == 3 ) {
        # Decode the post_op_attr and the accessbits for NFSv3
        $ret{fattr}      = xdr_post_op_attr(xdr=>$xdr);
        $ret{accessbits} = $xdr->uint() unless $ret{status};
    } elsif ( ! $ret{status} )  {
        # Decode the supported and access bitmasks for NFSv4
        $ret{supported} = $xdr->uint();
        $ret{access}    = $xdr->uint();
    }

    # Return the ACCESS results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_getfh

  $out = xdr_getfh( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 GETFH results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 GETFH results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_getfh(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    # Decode the filehandle
    $ret{fhandle} = xdr_fhandle(xdr=>$xdr, version=>$opts{version})
      unless $ret{status};

    # Return the GETFH results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_getattr

  $out = xdr_getattr( xdr => $xdr_obj, version => $version );

Decodes the NFS GETATTR results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS GETATTR results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_getattr(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    # Decode the fattr structure
    $ret{fattr} = xdr_fattr(xdr=>$xdr, version=>$opts{version})
      unless $ret{status};

    # Return the GETATTR results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_setattr

  $out = xdr_setattr( xdr => $xdr_obj, version => $version );

Decodes the NFS SETATTR results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS SETATTR results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_setattr(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 2 ) {
        # Decode the fattr structure for NFSv2
        $ret{fattr} = xdr_fattr(xdr=>$xdr, version=>$ver)
          unless $ret{status};
    } elsif ( $ver == 3 ) {
        # Decode the wcc_data structure for NFSv3
        $ret{wcc_data} = xdr_wcc_data(xdr=>$xdr);
    } else {
        # Decode the bitmap4 structure for NFSv4
        $ret{attrsset} = xdr_bitmap4(xdr=>$xdr);
    }

    # Return the SETATTR results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_lookup

  $out = xdr_lookup( xdr => $xdr_obj, version => $version );

Decodes the NFS LOOKUP results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS LOOKUP results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_lookup(%) {
    $Log->enter();
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status

    $ret{status} = $xdr->uint();

    if ( ! $ret{status} ) {
        # Decode the filehandle
        $ret{fhandle} = xdr_fhandle(xdr=>$xdr, version=>$ver);

        if ( $ver == 2 ) {
            # Decode the fattr structure for NFSv2
            $ret{fattr} = xdr_fattr(xdr=>$xdr, version=>$ver);
        } elsif ( $ver == 3 ) {
            # Decode the object post_op_attr for NFSv3
            $ret{fattr} = xdr_post_op_attr(xdr=>$xdr);
        }
    }

    if ( $ver == 3 ) {
        # Decode the directory post_op_attr for NFSv3
        $ret{dattr} = xdr_post_op_attr(xdr=>$xdr);
    }

    # Return the LOOKUP results
    $Log->exit();
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_readdir

  $out = xdr_readdir( xdr => $xdr_obj, version => $version );

Decodes the NFS READDIR results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS READDIR results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_readdir(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the directory post_op_attr for NFSv3
        $ret{dattr} = xdr_post_op_attr(xdr=>$xdr);
    }

    if ( ! $ret{status} ) {
        if ( $ver != 2 ) {
            # Decode the cookie verifier for NFSv3 and v4
            $ret{cookieverf} = xdr_cookieverf(xdr=>$xdr);
        }
        if ( $ver == 4 ) {
            # Decode NFSv4 list of directory entries
            while ( $xdr->uint() ) {
                push @{$ret{list}},
                      { cookie => xdr_cookie(xdr=>$xdr, version=>4),
                        name   => xdr_name(xdr=>$xdr, version=>4),
                        fattr  => xdr_fattr(xdr=>$xdr, version=>4) };
            }
        } else {
            # Decode NFSv2 and v3 list of directory entries
            while ( $xdr->uint() ) {
                push @{$ret{list}},
                      { fileid => $ver == 2 ? $xdr->uint() : $xdr->uint64(),
                        name   => xdr_name(xdr=>$xdr, version=>$ver),
                        cookie => xdr_cookie(xdr=>$xdr, version=>$ver) };
            }
        }
        # Decode the eof boolean
        $ret{eof} = $xdr->uint();
    }

    # Return the READDIR results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_readdirplus

  $out = xdr_readdirplus( xdr => $xdr_obj );

Decodes the NFSv3 READDIRPLUS results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 READDIRPLUS results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_readdirplus(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    # Decode the directory post_op_attr
    $ret{dattr} = xdr_post_op_attr(xdr=>$xdr);

    if ( ! $ret{status} ) {
        # Decode the cookie verifier
        $ret{cookieverf} = xdr_cookieverf(xdr=>$xdr);

        # Decode NFSv3 list of directory entries
        # with the post_op_attr for each entry
        while ( $xdr->uint() ) {
            push @{$ret{list}},
                  { fileid    => $xdr->uint64(),
                    name      => xdr_name(xdr=>$xdr, version=>3),
                    cookie    => xdr_cookie(xdr=>$xdr, version=>3),
                    name_attr => xdr_post_op_attr(xdr=>$xdr) };

            # Decode the filehandle if given
            $ret{list}[-1]{fhandle} = xdr_fhandle(xdr=>$xdr, version=>3)
              if $xdr->uint();
        }
        # Decode the eof boolean
        $ret{eof} = $xdr->uint();
    }

    # Return the READDIRPLUS results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_readlink

  $out = xdr_readlink( xdr => $xdr_obj, version => $version );

Decodes the NFS READLINK results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS READLINK results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_readlink(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the symbolic link post_op_attr
        $ret{sym_attr} = xdr_post_op_attr(xdr=>$xdr);
    }

    # Decode the linkdata
    $ret{data} = xdr_path(xdr=>$xdr, version=>$ver) unless $ret{status};

    # Return the READLINK results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_mkdir

  $out = xdr_mkdir( xdr => $xdr_obj, version => $version );

Decodes the NFS MKDIR results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS MKDIR results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_mkdir(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( ! $ret{status} ) {
        if ( $ver == 2 ) {
            # Decode the filehandle and fattr for NFSv2
            $ret{fhandle} = xdr_fhandle(xdr=>$xdr, version=>$ver);
            $ret{fattr}   = xdr_fattr(xdr=>$xdr, version=>$ver);
        } elsif ( $ver == 3 ) {
            # Decode the post_op_fh3 and post_op_attr for NFSv3
            $ret{fhandle} = xdr_post_op_fh3(xdr=>$xdr);
            $ret{fattr}   = xdr_post_op_attr(xdr=>$xdr);
        }
    }

    if ( $ver == 3 ) {
        $ret{dir_wcc} = xdr_wcc_data(xdr=>$xdr);
    }

    # Return the MKDIR results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_mknod

  $out = xdr_mknod( xdr => $xdr_obj );

Decodes the NFSv3 MKNOD results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 MKNOD results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_mknod(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( ! $ret{status} ) {
        # Decode the post_op_fh3 and post_op_attr
        $ret{fhandle} = xdr_post_op_fh3(xdr=>$xdr);
        $ret{fattr}   = xdr_post_op_attr(xdr=>$xdr);
    }

    # Decode the wcc_data structure
    $ret{dir_wcc} = xdr_wcc_data(xdr=>$xdr);

    # Return the MKNOD results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_symlink

  $out = xdr_symlink( xdr => $xdr_obj, version => $version );

Decodes the NFS SYMLINK results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS SYMLINK results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_symlink(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        if ( ! $ret{status} ) {
            # Decode the post_op_fh3 and post_op_attr
            $ret{fhandle} = xdr_post_op_fh3(xdr=>$xdr);
            $ret{fattr}   = xdr_post_op_attr(xdr=>$xdr);
        }
        # Decode the wcc_data structure
        $ret{dir_wcc} = xdr_wcc_data(xdr=>$xdr);
    }

    # Return the SYMLINK results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_create

  $out = xdr_create( xdr => $xdr_obj, version => $version );

Decodes the NFS CREATE results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS CREATE results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_create(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( !$ret{status} ) {
        if ( $ver == 2 ) {
            # Decode the filehandle and fattr for NFSv2
            $ret{fhandle} = xdr_fhandle(xdr=>$xdr, version=>$ver);
            $ret{fattr}   = xdr_fattr(xdr=>$xdr, version=>$ver);
        } elsif ( $ver == 3 ) {
            # Decode the post_op_fh3 and post_op_attr for NFSv3
            $ret{fhandle} = xdr_post_op_fh3(xdr=>$xdr);
            $ret{fattr}   = xdr_post_op_attr(xdr=>$xdr);
        } else {
            # Decode the change_info4 and bitmap4 for NFSv4
            $ret{cinfo}   = xdr_change_info4(xdr=>$xdr);
            $ret{attrset} = xdr_bitmap4(xdr=>$xdr);
        }
    }

    if ( $ver == 3 ) {
        # Decode the wcc_data structure for NFSv3
        $ret{dir_wcc} = xdr_wcc_data(xdr=>$xdr);
    }

    # Return the CREATE results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_remove

  $out = xdr_remove( xdr => $xdr_obj, version => $version );

Decodes the NFS REMOVE results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS REMOVE results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_remove(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the wcc_data structure for NFSv3
        $ret{dir_wcc} = xdr_wcc_data(xdr=>$xdr);
    } elsif ( !$ret{status} and $ver == 4 ) {
        # Decode the change_info4 for NFSv4
        $ret{cinfo} = xdr_change_info4(xdr=>$xdr);
    }

    # Return the REMOVE results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_rename

  $out = xdr_rename( xdr => $xdr_obj, version => $version );

Decodes the NFS RENAME results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS RENAME results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_rename(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the wcc_data for both fromdir and todir for NFSv3
        $ret{fromdir_wcc} = xdr_wcc_data(xdr=>$xdr);
        $ret{todir_wcc}   = xdr_wcc_data(xdr=>$xdr);
    } elsif ( !$ret{status} and $ver == 4 ) {
        # Decode the change_info4 for both source and target for NFSv4
        $ret{source_cinfo} = xdr_change_info4(xdr=>$xdr);
        $ret{target_cinfo} = xdr_change_info4(xdr=>$xdr);
    }

    # Return the RENAME results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_link

  $out = xdr_link( xdr => $xdr_obj, version => $version );

Decodes the NFS LINK results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS LINK results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_link(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the fromdir post_op_attr and linkdir wcc_data for NFSv3
        $ret{fromdir_wcc} = xdr_post_op_attr(xdr=>$xdr);
        $ret{linkdir_wcc} = xdr_wcc_data(xdr=>$xdr);
    } elsif ( !$ret{status} and $ver == 4 ) {
        # Decode the change_info4 for NFSv4
        $ret{cinfo} = xdr_change_info4(xdr=>$xdr);
    }

    # Return the LINK results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_fsstat

  $out = xdr_fsstat( xdr => $xdr_obj, version => $version );

Decodes the NFSv2 STATFS or NFSv3 FSSTAT results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS FSSTAT/STATFS results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fsstat(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();
    # Decode the post_op_attr for NFSv3
    $ret{fattr}  = xdr_post_op_attr(xdr=>$xdr) if ($ver == 3);

    if ( ! $ret{status} ) {
        if ( $ver == 2 ) {
            # Decode the STATFS results
            %ret = ( %ret,
                     tsize  => $xdr->uint(),
                     bsize  => $xdr->uint(),
                     blocks => $xdr->uint(),
                     bfree  => $xdr->uint(),
                     bavail => $xdr->uint() );
        } else {
            # Decode the FSSTAT results
            %ret = ( %ret,
                     tbytes   => $xdr->uint64(),
                     fbytes   => $xdr->uint64(),
                     abytes   => $xdr->uint64(),
                     tfiles   => $xdr->uint64(),
                     ffiles   => $xdr->uint64(),
                     afiles   => $xdr->uint64(),
                     invarsec => $xdr->uint() );
        }
    }

    # Return the FSSTAT/STATFS results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_fsinfo

  $out = xdr_fsinfo( xdr => $xdr_obj );

Decodes the NFSv3 FSINFO results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NFSv3 FSINFO results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_fsinfo(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();
    # Decode the post_op_attr for NFSv3
    $ret{fattr}  = xdr_post_op_attr(xdr=>$xdr);

    if ( ! $ret{status} ) {
        # Decode the FSINFO results
        %ret = ( %ret,
                 rtmax       => $xdr->uint(),
                 rtperf      => $xdr->uint(),
                 rtmult      => $xdr->uint(),
                 wtmax       => $xdr->uint(),
                 wtperf      => $xdr->uint(),
                 wtmult      => $xdr->uint(),
                 dtperf      => $xdr->uint(),
                 maxfilesize => $xdr->uint64(),
                 time_delta  => xdr_nfstime(xdr=>$xdr, version=>3),
                 properties  => $xdr->uint() );
    }

    # Return the FSINFO results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_pathconf

  $out = xdr_pathconf( xdr => $xdr_obj, version => $version );

Decodes the NFS PATHCONF results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS PATHCONF results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_pathconf(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    if ( $ver == 2 ) {
        # Decode the PATHCONF result on NFSv2
        %ret = ( linkmax      => $xdr->uint(),
                 pc_max_canon => $xdr->uint(),
                 pc_max_input => $xdr->uint(),
                 name_max     => $xdr->uint(),
                 pc_path_max  => $xdr->uint(),
                 pc_pipe_buf  => $xdr->uint(),
                 pc_vdisable  => $xdr->uint(),
                 pc_xxx       => $xdr->uint(),
                 pc_mask1     => $xdr->uint(),
                 pc_mask2     => $xdr->uint() );
    } else {
        # Decode the NFSv3 status
        $ret{status} = $xdr->uint();
        # Decode the post_op_attr for NFSv3
        $ret{fattr}  = xdr_post_op_attr(xdr=>$xdr);

        # Decode the PATHCONF result on NFSv3
        if ( ! $ret{status} ) {
            %ret = ( %ret,
                     linkmax          => $xdr->uint(),
                     name_max         => $xdr->uint(),
                     no_trunc         => $xdr->uint(),
                     chown_restricted => $xdr->uint(),
                     case_insensitive => $xdr->uint(),
                     case_preserving  => $xdr->uint() );
        }
    }

    # Return the PATHCONF results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_read

  $out = xdr_read( xdr => $xdr_obj, version => $version );

Decodes the NFS READ results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS READ results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_read(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the post_op_attr for NFSv3
        $ret{file_attr} = xdr_post_op_attr(xdr=>$xdr);
    }

    if ( ! $ret{status} ) {
        if ( $ver == 3 ) {
            # Decode the number of bytes actually read
            $ret{count} = $xdr->uint();
        }

        if ( $ver == 2 ) {
            # Decode the fattr and the data read for NFSv2
            $ret{fattr} = xdr_fattr(xdr=>$xdr, version=>$ver);
            $ret{data}  = $xdr->opaque(variable=>MAXDATA);
        } else {
            # Decode the eof boolean and the
            # data read for NFSv3 and v4
            $ret{eof}  = $xdr->uint();
            $ret{data} = $xdr->opaque();
        }
    }

    # Return the READ results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_write

  $out = xdr_write( xdr => $xdr_obj, version => $version );

Decodes the NFS WRITE results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS WRITE results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_write(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the wcc_data structure for NFSv3
        $ret{file_wcc} = xdr_wcc_data(xdr=>$xdr);
    }

    if ( ! $ret{status} ) {
        if ( $ver == 2 ) {
            # Decode the fattr for NFSv2
            $ret{fattr} = xdr_fattr(xdr=>$xdr, version=>$ver);
        } else {
            # Decode the number of bytes of data actually written,
            # the level of commitment of the data and metadata,
            # and the write verifier
            $ret{count}     = $xdr->uint();
            $ret{committed} = $xdr->uint();
            $ret{writeverf} = $xdr->opaque(fixed=>8);
        }
    }

    # Return the WRITE results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_commit

  $out = xdr_commit( xdr => $xdr_obj, version => $version );

Decodes the NFS COMMIT results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS COMMIT results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_commit(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ver == 3 ) {
        # Decode the wcc_data structure for NFSv3
        $ret{file_wcc} = xdr_wcc_data(xdr=>$xdr);
    }

    if ( ! $ret{status} ) {
        # Decode the write verifier
        $ret{writeverf} = $xdr->opaque(fixed=>8);
    }

    # Return the COMMIT results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_open

  $out = xdr_open( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 OPEN results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 OPEN results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_open(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the OPEN results
        %ret = ( %ret,
                 stateid    => xdr_stateid4(xdr=>$xdr),
                 cinfo      => xdr_change_info4(xdr=>$xdr),
                 rflags     => $xdr->uint(),
                 attrset    => xdr_bitmap4(xdr=>$xdr) );
        my $deleg_type = $xdr->uint();
        if ( $deleg_type ) {
            if ($deleg_type == OPEN_DELEGATE_NONE_EXT) {
                my $ond_why = $xdr->uint();
                if ($ond_why == WND4_CONTENTION) {
                    $ret{ond_server_will_push_deleg} = $xdr->uint();
                } elsif ($ond_why == WND4_RESOURCE) {
                    $ret{ond_server_will_signal_avail} = $xdr->uint();
                }
                $ret{ond_why} = $ond_why;
            } else {
                $ret{deleg} = {
                    stateid => xdr_stateid4(xdr=>$xdr),
                    recall  => xdr_uint(xdr=>$xdr),
                };
                if ( $deleg_type == OPEN_DELEGATE_WRITE ) {
                    $ret{deleg}{space_limit} = $xdr->uint() == 1 ?
                                           { filesize => $xdr->uint64() } :
                                           { num_blocks      => $xdr->uint(),
                                             bytes_per_block => $xdr->uint() };
                } 
                $ret{deleg}{permissions} = xdr_nfsace4(xdr=>$xdr);
            }
        }
    }

    # Return the OPEN results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_open_downgrade

  $out = xdr_open_downgrade( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 OPEN_DOWNGRADE results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 OPEN_DOWNGRADE results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_open_downgrade(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the OPEN_DOWNGRADE results
        %ret = ( %ret, stateid => xdr_stateid4(xdr=>$xdr) );
    }

    # Return the OPEN results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_close

  $out = xdr_close( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 CLOSE results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 CLOSE results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_close(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the stateid4
        $ret{stateid} = xdr_stateid4(xdr=>$xdr);
    }

    # Return the CLOSE results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_lock

  $out = xdr_lock( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 LOCK results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 LOCK results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_lock(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the stateid4
        $ret{stateid} = xdr_stateid4(xdr=>$xdr);
    } elsif ( $ret{status} == NFS4ERR_DENIED ) {
        %ret = ( %ret, %{ xdr_lock4_denied(xdr => $xdr) } );
    }

    # Return the LOCK results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_locku

  $out = xdr_locku( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 LOCKU results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 LOCKU results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_locku(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the stateid4
        $ret{stateid} = xdr_stateid4(xdr=>$xdr);
    }

    # Return the LOCKU results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_lockt

  $out = xdr_lockt( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 LOCKT results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 LOCKT results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_lockt(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4ERR_DENIED ) {
        %ret = ( %ret, %{ xdr_lock4_denied(xdr => $xdr) } );
    }

    # Return the LOCKT results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_nlm_stat

  $out = xdr_nlm_stat( xdr => $xdr_obj );

Decodes the NLM status results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

NLM status results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_nlm_stat(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my $xdr = $opts{xdr};

    # Decode the cookie and the status
    # and return the results
    return { cookie => $xdr->opaque(),
             status => $xdr->uint() };
}

#---------------------------------------------------------------------

=head2 xdr_nlm_test

  $out = xdr_nlm_test( xdr => $xdr_obj, version => $version );

Decodes the NLM_TEST results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NLM version.

=back

=item Returns

NLM_TEST results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_nlm_test(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the cookie
    $ret{cookie} = $xdr->opaque();

    # Decode the status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NLM_DENIED ) {
        %ret = (%ret,
            exlusive => $xdr->uint(),
            svid     => $xdr->uint(),
            oh       => $xdr->opaque());
        if ( $ver == 4 ) {
            %ret = (%ret,
                l_offset => $xdr->uint64(),
                l_len    => $xdr->uint64());
        } else {
            %ret = (%ret,
                l_offset => $xdr->uint(),
                l_len    => $xdr->uint());
        }
    }

    # Return the NLM_TEST results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_secinfo

  $out = xdr_secinfo( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 SECINFO results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 SECINFO results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_secinfo(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode number of secinfo4 structures
        foreach ( 1..$xdr->uint() ) {
            # Decode each secinfo4 structure
            push @{$ret{secinfo}}, { flavor => $xdr->uint() };
            if ( $ret{secinfo}[-1]{flavor} == RPCSEC_GSS ) {
                $ret{secinfo}[-1]{info} = {
                    oid     => $xdr->opaque(),
                    qop     => $xdr->uint(),
                    service => $xdr->uint(),
                };
            }
        }
    }

    # Return the SECINFO results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_setclientid

  $out = xdr_setclientid( xdr => $xdr_obj, version => $version );

Decodes the NFSv4 SETCLIENTID results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 SETCLIENTID results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_setclientid(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the clientid and the clientid_confirm
        $ret{clientid} = $xdr->opaque(fixed=>8);
        $ret{confirm}  = $xdr->opaque(fixed=>8);
    } elsif ( $ret{status} == NFS4ERR_CLID_INUSE ) {
        # NFS4ERR_CLID_INUSE
        $ret{client_using} = {
            r_netid => $xdr->opaque(),
            r_addr  => $xdr->opaque(),
        };
    }

    # Return the SETCLIENTID results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_mount

  $out = xdr_mount( xdr => $xdr_obj, version => $version );

Decodes the NFS MOUNT results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version.

=back

=item Returns

NFS MOUNT results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_mount(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my (%ret, @auth);
    my $xdr = $opts{xdr};
    my $ver = $opts{version};

    # Decode the NFS status
    $ret{status} = $xdr->uint();
    return \%ret if $ret{status};

    # Decode the filehandle of the mounted path
    $ret{fhandle} = xdr_fhandle(xdr=>$xdr, version=>$ver);

    if ( $ver == 3 ) {
        # Decode the number of authentication flavors
        # the server supports
        foreach ( 1..$xdr->uint() ) {
            # Decode each supported authentication flavor
            push @auth, $xdr->uint();
        }
        $ret{auth_flavors} = \@auth;
    }

    # Return the MOUNT results
    return \%ret;
}

#---------------------------------------------------------------------

=head2 xdr_dump

  $out = xdr_dump( xdr => $xdr_obj );

Decodes the MOUNT DUMP results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

MOUNT DUMP results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_dump(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my @ret;
    my $xdr = $opts{xdr};

    # Get all elements in linked list
    while ( $xdr->uint() ) {
        # Decode the hostname and path
        push @ret, { hostname => $xdr->opaque(variable=>255),
                     pathname => $xdr->opaque(variable=>1024) };
    }

    # Return the MOUNT DUMP results
    return \@ret;
}

#---------------------------------------------------------------------

=head2 xdr_export

  $out = xdr_export( xdr => $xdr_obj );

Decodes the MOUNT EXPORT results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=back

=item Returns

MOUNT EXPORT results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub xdr_export(%) {
    my (%opts) = validate (
        @_,
        {
            xdr => { type => OBJECT },
        }
    );
    my (@ret, $dir, @name);
    my $xdr = $opts{xdr};

    # Get all elements in linked list
    while ( $xdr->uint() ) {
        # Decode the directory name of export
        $dir = $xdr->opaque(variable=>1024);

        # Get all elements in linked list
        while ( $xdr->uint() ) {
            # Decode the names with access to export
            push @name, $xdr->opaque(variable=>255);
        }
        # Add entry for this export
        push @ret, {directory => $dir, name => \@name };
    }

    # Return the MOUNT EXPORT results
    return \@ret;
}

#---------------------------------------------------------------------

=head2 xdr_exchange_id

  $out = xdr_exchange_idn( xdr => $xdr_obj, version => $version );

Decodes the NFSv4.1 EXCHANGE_ID results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 EXCHANGE_ID results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------


sub xdr_exchange_id(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();
    if ( $ret{status} == NFS4_OK ) {
        # Decode the EXCHANGE_ID results
        %ret = ( %ret,
         clientid   => $xdr->opaque (fixed => 8),
         seq_id     => xdr_uint (xdr=>$xdr),
         eir_flag   => xdr_uint (xdr => $xdr),
         state_pro  => xdr_uint (xdr => $xdr)
	 );
	 if($ret{state_pro} == SP4_MACH_CRED) {
	   my $attr1 = xdr_bitmap4(xdr=>$xdr);
	   $ret{spo_must_enforce} = bitmap4_convert(attrmask=>$attr1);
	   my $attr2 = xdr_bitmap4(xdr=>$xdr);
	   $ret{spo_must_allow} = bitmap4_convert(attrmask=>$attr2);
	 }
     }
    # Return the EXCHANGE_ID results
    return \%ret;
}


#---------------------------------------------------------------------

=head2 xdr_create_session

  $out = xdr_create_session( xdr => $xdr_obj, version => $version );

Decodes the NFSv4.1 CREATE_SESSION results.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object to decode from.

=item version

(Required) : NFS version. This option is not used but
it is required since this function can be called by
xdr_compound_res which passes this argument.

=back

=item Returns

NFSv4 CREATE_SESSION results

=back

=for html <fr />

=cut

#---------------------------------------------------------------------


sub xdr_create_session(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the CREATE_SESSION results
        %ret = ( %ret,
                 sessionid   => $xdr->opaque (fixed => 16),
                 seq_id     => xdr_uint (xdr=>$xdr),
                 csr_flag   => xdr_uint (xdr => $xdr),
         headerpadsize       => xdr_uint(xdr => $xdr),
         maxrequestsize      => xdr_uint(xdr => $xdr),
         maxresponsesize     => xdr_uint(xdr => $xdr),
         maxresponsesize_c   => xdr_uint(xdr => $xdr),
         maxoperations       => xdr_uint(xdr => $xdr),
         maxrequests         => xdr_uint(xdr => $xdr),
         rdma_ird            => xdr_uint(xdr => $xdr),
                 headerpadsize_b     => xdr_uint(xdr => $xdr),
                 maxrequestsize_b    => xdr_uint(xdr => $xdr),
                 maxresponsesize_b   => xdr_uint(xdr => $xdr),
                 maxresponsesize_c_b => xdr_uint(xdr => $xdr),
                 maxoperations_b     => xdr_uint(xdr => $xdr),
                 maxrequests_b       => xdr_uint(xdr => $xdr),
                 rdma_ird_b          => xdr_uint(xdr => $xdr),

        );
        }

    # Return the CREATE_SESSION results
    return \%ret;
}


sub xdr_sequence(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the SEQUENCE results
        %ret = ( %ret,
                 sessionid_s   => $xdr->opaque (fixed => 16),
                 seq_id_s      => xdr_uint (xdr=>$xdr),
                 sr_slotid    => xdr_uint (xdr => $xdr),
         high_slotid  => xdr_uint (xdr => $xdr),
              target_slotid => xdr_uint (xdr => $xdr),
             status_flags => xdr_uint (xdr => $xdr),    
                );
        }

    # Return the SEQUENCE results
    return \%ret;
}



sub xdr_getdevicelist(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the SEQUENCE results
        %ret = ( %ret,
                 devlist_cookie   => $xdr->uint (),
                 devlist_verf     => xdr_opaque (xdr=>$xdr, fixed => 1, maxlen => 16),
                 deviceid         => xdr_opaque (xdr => $xdr, fixed => 1, maxlen => 16),
                 devlist_bool     => xdr_uint (xdr => $xdr),
                );
        }

    # Return the GETDEVICELIST results
    return \%ret;
}

sub xdr_getdeviceinfo(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[1234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};

    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
       $ret{dev_layout} = $xdr->uint();
       $ret{response_length} =  $xdr->uint ();
       $ret{number_index} =  $xdr->uint ();
       $ret{dev_index} = xdr_bitmap4 (xdr=>$xdr, count => $ret{number_index});
       $ret{no_lifs} = $xdr->uint();

        #%ret = ( %ret,
        #         dev_layout       => xdr_uint (xdr=>$xdr),
        #         response_length  => $xdr->uint (),
        #         number_index     => $xdr->uint (),
        #         dev_index        => xdr_bitmap4 (xdr=>$xdr, count => 4096),
        #         no_lifs          => $xdr->uint(),
        #        );

    my ($i, $k);
    my (@arr1, @arr2, @arr3);
    $k = $ret{no_lifs};

    $Log->debug("No of lifs is $k");



if ($main::ADDR_TYPE eq "inet6") {
    for ($i=0; $i < $k; $i++) {
		$arr1[$i] = xdr_uint(xdr=> $xdr);;
		$arr2[$i] = xdr_opaque(xdr=> $xdr, fixed => 1, maxlen => 8);
		$arr3[$i] = xdr_opaque(xdr=> $xdr, fixed => 0, len => 128);
	}
} elsif ($main::ADDR_TYPE eq "inet4") {
    for ($i=0; $i < $k; $i++) {
		$arr1[$i] = xdr_uint(xdr=> $xdr);;
		$arr2[$i] = xdr_opaque(xdr=> $xdr, fixed => 1, maxlen => 8);
		$arr3[$i] = xdr_opaque(xdr=> $xdr, fixed => 1, maxlen => 24);
	}
}

	$ret{lifs_paths} = \@arr1;
	$ret{proto} = \@arr2;
	$ret{lif_ip} = \@arr3;
   print "Dumper Lif IP is" . Dumper(\%ret);   

    

        }

    # Return the GETDEVICEINFO results
    return \%ret;
}


sub xdr_layoutget(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[234]$/ },
        }
    );
    my %ret;
    my $xdr = $opts{xdr};
    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the LAYOUTGET results
        %ret = ( %ret,
                 ret_on_close     => xdr_uint (xdr=>$xdr),
                 stateid          => xdr_stateid4 (xdr=>$xdr),
         layout_info      => xdr_layout4 (xdr=>$xdr),

                );
        }

    # Return the LAYOUTGET results
    return \%ret;
}


sub xdr_layout4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => HASHREF|SCALAR, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    if ( defined($opts{encode}) ) {
        # Encode the stateid structure
        $xdr->uint64(encode=>$opts{encode}{layout_offset});
        $xdr->uint64(encode=>$opts{encode}{layout_length});
        $xdr->uint(encode=>$opts{encode}{layout_iomode});
        $xdr->uint(encode=>$opts{encode}{layouttype});
        $xdr->opaque(encode=>$opts{encode}{layoutbody});
        # Return the encoded message
        return $xdr->{message};
    } else {
        # Return the decoded stateid structure
    #return { layout_offset => $xdr->uint64(), layout_length => $xdr->uint64(), io_mode => $xdr->uint(), layouttype => $xdr->uint(), layoutbody => $xdr->opaque() };
    my %ret = ( segmebt => $xdr->uint(), 
                    layout_offset => $xdr->uint64(),
                    layout_length => $xdr->uint64(),
                    io_mode => $xdr->uint(),
                    layouttype => $xdr->uint(),
            unknown    => $xdr->uint(),
                    deviceid_l => xdr_opaque(xdr => $xdr, fixed => 1, maxlen => 16),
                    nfsl_util => $xdr->uint(),
                    stripe_to_use => $xdr->uint(), 
                    offset_l => $xdr->uint64(),
            fh_count => $xdr->uint(),
                    fh_layout => xdr_fhandle(xdr=>$xdr, version => 4)
                 );
    return \%ret;
    }
}


sub xdr_layouttype(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    my @layout;

        # Return the decoded stateid structure

        foreach ( 1..$xdr->uint() ) {
            # Decode each nfsace4 structure
        @layout = $xdr->uint();
    }
        return \@layout;
}

sub xdr_mdsthreshold(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    my ($out, %out);
    my $i = 0;

        # Return the decoded stateid structure
    $out{layouttype} = $xdr->uint();
        foreach ( 1..$xdr->uint() ) {
            # Decode each nfsace4 structure
                $out{layout} = $xdr->uint();
        $i++;
        }
   my $ol = $xdr->uint();
   $out{hintlist} = $xdr->opaque(fixed => $ol); 
 return %out;
}

sub xdr_layouthint (%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT, optional => 1 },
            encode => { type => OBJECT, optional => 1 },

        }
    );

    my ($out, %out);
    my $xdr = $opts{xdr} ? $opts{encode} : nfs::Xdr->new();
    $Log->debug("XDR object:" .  Dumper($xdr));
    $Log->trace("There is a xdr object");
    $out{layouttype} = $xdr->uint();
        $out{nflh_care} = $xdr->uint();
        $out{nflh_util} = $xdr->uint();
    $out{count} = $xdr->uint();
    return %out;

}


sub xdr_layouthint4 (%) {
    my (%opts) = validate (
        @_,
        {
            encode => { type => HASHREF|SCALAR, optional => 1 },

        }
    );

    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();
    my ($out, %out);
        $xdr->uint(encode => $opts{encode}{layouttype});
        $xdr->uint(encode => $opts{encode}{nflh_care});
        $xdr->uint(encode => $opts{encode}{nflh_util});
        $xdr->uint(encode => $opts{encode}{count});
    return $xdr;

}

sub xdr_copy_notify(%) {
    $Log->enter();
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[234]$/ },
        }
    );
    my %ret;
    my ($i, $op_leng);
    my $xdr = $opts{xdr};
    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
        # Decode the COPY_NOFITY results
        %ret = ( %ret,
                 seconds     => xdr_uint64 (xdr=>$xdr),
                 nseconds    => xdr_uint (xdr=>$xdr),
                 arr_l       => xdr_uint (xdr => $xdr),
                );
        
        for ($i = 0; $i < $ret{arr_l} ; $i++) {
            $ret{nloc_type}[$i] = xdr_uint(xdr=> $xdr);
            if ($ret{nloc_type}[$i] == 0) {
                $ret{host_name}[$i] = xdr_opaque(xdr=>$xdr, fixed =>1, len => $op_leng);
            } elsif ($ret{nloc_type}[$i] == 2) {
                $ret{t_proto}[$i] = xdr_opaque(xdr=>$xdr, len => 3);
                $ret{t_addr}[$i] = xdr_opaque(xdr=>$xdr,  len => 15);
            } elsif ($ret{nloc_type}[$i] == 1) { 
                $ret{url}[$i] = xdr_opaque(xdr=>$xdr);
            }
        }
    }
    $Log->trace("Copy notify returned - $ret{status}");

    # Return the COPY_NOTIFY results
    $Log->exit();
    return \%ret;
}


sub xdr_copy(%) {
    my (%opts) = validate (
        @_,
        {
            xdr     => { type => OBJECT },
            version => { type => SCALAR, regex => qr/^[234]$/ },
        }
    );
    my (%ret, $i);
    my $xdr = $opts{xdr};
    # Decode the NFS status
    $ret{status} = $xdr->uint();

    if ( $ret{status} == NFS4_OK ) {
    $ret{arr_size} = $xdr->uint();
    for ($i = 0; $i < $ret{arr_size} ; $i++) {
        $ret{stateid} = xdr_stateid4 (xdr=>$xdr);
        }
        }

    return \%ret;
}

sub xdr_fs_locations_info4(%) {
    my (%opts) = validate (
        @_,
        {
            xdr    => { type => OBJECT,   optional => 1 },
            encode => { type => ARRAYREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    $Log->debug("\n nfsXdr : xdr_fs_locations_info4 : xdr = \n" . Dumper($xdr)."\n");

    my %fs_locations_info;
    my $fli_flags = $xdr->uint();
    my $fli_valid_for = $xdr->uint();

    # Decode fli_fs_root
    my $number_of_components = $xdr->uint();
    foreach ( 1..$number_of_components ) {
        push (@{$fs_locations_info{fs_root}},xdr_opaque(xdr=>$xdr, fixed=>0));
    }

    # Decode fli_items
     my $number_of_fli_items = xdr_uint(xdr=>$xdr);

      foreach ( 1..$number_of_fli_items ) {
    
        # Decode fli_entries
        my $number_of_fli_entries = xdr_uint(xdr=>$xdr);
        foreach ( 1..$number_of_fli_entries ) {
    
            push (@{$fs_locations_info{fls_currency}},xdr_uint(xdr=>$xdr));
    
            my $number_of_fli_infos = xdr_uint(xdr=>$xdr);
            foreach ( 1..$number_of_fli_infos ) {
                push (@{$fs_locations_info{fls_info}},xdr_opaque(xdr=>$xdr, fixed=>0));
            }
    
            push (@{$fs_locations_info{servers}},xdr_opaque(xdr=>$xdr, fixed=>0));
        }
    
        # Decode rootpath
        my $number_of_root_path_components = $xdr->uint();
        foreach ( 1..$number_of_root_path_components ) {
            push (@{$fs_locations_info{rootpath}},xdr_opaque(xdr=>$xdr, fixed=>0));
        }
        $Log->debug("   rootpath: @{$fs_locations_info{rootpath}} \n");
    }

    # Return the decoded fs_locations4 structure
    return \%fs_locations_info;
} #end sub xdr_fs_locations_info4

sub xdr_fs_status(%) {
    my (%opts) = validate (
        @_,
        {
        xdr    => { type => OBJECT,   optional => 1 },
        encode => { type => ARRAYREF, optional => 1 },
        }
    );
    my $xdr = $opts{xdr} ? $opts{xdr} : nfs::Xdr->new();

    $Log->debug ("\n nfsXdr : xdr_fs_status : xdr = \n" . Dumper($xdr)."\n");

    my %fs_status;
    $fs_status{fss_absent}  = xdr_uint(xdr=>$xdr);
    $fs_status{fss_type}    = xdr_uint(xdr=>$xdr);
    $fs_status{fss_source}  = xdr_opaque(xdr=>$xdr, fixed=>0);
    $fs_status{fss_current} = xdr_opaque(xdr=>$xdr, fixed=>0);
    $fs_status{fss_age}     = xdr_uint(xdr=>$xdr);
    $fs_status{fss_version} = xdr_nfstime(xdr=>$xdr, version=>4),

    # Return the decoded fs_status structure
    return \%fs_status;
}


1;

