# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Operation.pm#1 $
#
## @summary     Operation - part of the Flg library
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Operation - library functions specifically for use by the Flg library

=head1 SYNOPSIS

  use nfs::Operation;

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

This package is intended for use by the nfs::Flg.pm library only. Do not
call nfs::Operation.pm directly.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Operation.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Operation;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Timer qw(global_timer);

use Carp qw(croak);
use Params::Validate qw(validate SCALAR HASHREF CODEREF OBJECT UNDEF);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
use constant NUM_OP_RECS => 16384;
use constant HASHSIZE => 1024;

my @op_recs;
my %ops_free   = ( head => 0, tail => 0 );
my @ops_hash;
my $op_metronome_callout;
my $op_metronome_callout_arg;

#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------
sub op_init($%;);
sub op_metronome($%;);
sub op_metronome_func();
sub metronome_start(%);
sub metronome_active(%);
sub metronome_tick(%);
sub testme();

#--------------------------------------------------------------------------------

=head2 op_init

  $op_obj = nfs::Operation->op_init( client_obj => $nfs_obj );

Initialize an Operation object.

The following named arguments are accepted:

=over

=item Arguments

=over

=item client_obj

(Required) : An NfsClient object

=back

=item Returns

The Operation object.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub op_init($%;) {
    #  int
    #  op_init(struct in_addr addr, int socktype)
    #  {
    #          int i, sock, sport;
    #
    #          if ((op_recs =
    #                  malloc(sizeof(struct operation_rec) * NUM_OP_RECS)) == NULL) {
    #                  report_perror(FATAL, "malloc");
    #                  return -1;
    #          }
    #          bzero(op_recs, sizeof(struct operation_rec) * NUM_OP_RECS);
    #
    #          sport = (getuid() == 0 ? 1 : 0); /* ask for privileged port if root */
    #          if ((sock =
    #               rpc_client(addr, NFS_PROG, NFS_VER3, socktype, sport)) < 0) {
    #                  report_error(FATAL, "rpc_client error");
    #                  return -1;
    #          }
    #
    #          op_sock = sock;
    #          op_socktype = socktype;
    #          op_outstanding = 0;
    #
    #          Q_INIT(&ops_active);
    #          Q_INIT(&ops_free);
    #          for (i=0 ; i<HASHSIZE ; i++) {
    #                  Q_INIT(&ops_hash[i]);
    #          }
    #          for (i=0 ; i<NUM_OP_RECS ; i++) {
    #                  Q_INSERT_HEAD(&ops_free, &op_recs[i], link);
    #          }
    #          return 0;
    #  }
    #

    # Rather than passing in the addr and socktype and establishing a client
    #  connection, just pass in the existing NFS client object and reuse it.
    my $class  = shift;
    my (%opts) = validate (
        @_,
        {
            client_obj => { type => OBJECT },
        }
    );
    my $self = bless { %opts }, $class;

    $self->{client_obj} = $opts{client_obj};

    # Q_INIT ops_active, ops_free just sets head and tail to 0
    #  So the index of the head and tail are both 0

    foreach ( 1 .. HASHSIZE ) {
        $ops_hash[$_] = { head => 0, tail => 0 };
    }
    foreach ( 1 .. NUM_OP_RECS ) {
        $op_recs[$_] = { lastsend      => 0,
                          retransmits  => 0,
                          call         => 0,
                          xid          => 0,
                          callback     => 0,
                          callback_arg => 0,
                          oop          => { measure => 0, start => 0 },
                          link_next    => 0,
                          link_prev    => 0,
                          hash_next    => 0,
                          hash_prev    => 0, };
        $op_recs[$_]{link_prev} = undef;
        $op_recs[$_]{link_next} = $ops_free{head};
        if ( $ops_free{head} ) {
            $op_recs[$ops_free{head}]{link_prev} = $_;
        } else {
            $ops_free{tail} = $_;
        }
        $ops_free{head} = $_;
    }

    return $self;
} # End op_init

#--------------------------------------------------------------------------------

=head2 op_metronome

  $op_obj->op_metronome( rate     => $rate,
                         duration => $duration,
                         func     => \&gen_func,
                         arg      => undef );

op_metronome function taken from the original C code.

The following named arguments are accepted:

=over

=item Arguments

=over

=item rate

(Required) : The desired ops/sec rate for Flg

=item duration

(Required) : The execution time for the load generation

=item func

(Required) : Reference to the operation generation function

=item arg

(Required) : Additional arguments to func

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub op_metronome($%;) {
    #  int
    #  op_metronome(int rate, int duration, void (*func)(void *), void *arg)
    #  {
    #          struct metronome mn;
    #          int waitusecs = 1000;
    #
    #  #define noSPIN_INSTEAD_OF_BLOCK
    #  #ifdef SPIN_INSTEAD_OF_BLOCK
    #  #warning "SPIN INSTEAD OF BLOCK"
    #          waitusecs = 0;
    #  #endif
    #
    #          op_metronome_callout = func;
    #          op_metronome_callout_arg = arg;
    #
    #          metronome_start(&mn, rate, duration, op_metronome_func, NULL);
    #
    #          while (metronome_active(&mn)) {
    #                  /*
    #                   * read and process replies
    #                   */
    #                  if (op_poll(waitusecs) < 0) {
    #                          report_error(FATAL, "op_poll error");
    #                          return -1;
    #                  }
    #                  /*
    #                   * send new messages
    #                   */
    #                  if (metronome_tick(&mn) < 0) {
    #                          report_error(FATAL, "metronome_tick error");
    #                          return -1;
    #                  }
    #                  /*
    #                   * retransmit stale messages
    #                   */
    #                  if (op_retransmit() < 0) {
    #                          report_error(FATAL, "op_retransmit error");
    #                          return -1;
    #                  }
    #          }
    #          return 0;
    #  }
    #

    my $self   = shift;
    my (%opts) = validate (
        @_,
        {
            rate     => { type => SCALAR },
            duration => { type => SCALAR },
            func     => { type => CODEREF },
            arg      => { type => SCALAR|UNDEF },
        }
    );

    my %mn;

    $op_metronome_callout     = $opts{func};
    $op_metronome_callout_arg = $opts{arg};

    metronome_start( mn                 => \%mn,
                     target_ops_per_sec => $opts{rate},
                     lifetime           => $opts{duration},
                     func               => \&op_metronome_func,
                     func_arg           => undef, );

    while ( metronome_active( mn => \%mn ) ) {
        # /*
        #  * read and process replies
        #  */
        #     Read and process replies
        #     This is all done in the send messages section
        #     nfs_client library sends/receives in one function call

        # /*
        #  * send new messages
        #  */
        if ( metronome_tick( mn => \%mn ) < 0 ) {
            croak "op_metronome: metronome_tick error";
        }

        # /*
        #  * retransmit stale messages
        #  */
        #     Retransmit stale messages
        #     Retransmission is handled by the send messages section
        #     nfs_client handles retransmissions
        #     This makes the report a little less accurate, since
        #     a retransmit success will not show as a retransmit on
        #     the report. The nfs_client doesn't report retransmission
        #     information. The timing for a given call will also
        #     include any retransmissions in addition to the original
        #     request, so timing might be off if there are many
        #     retransmissions required.
    }

} # End op_metronome

#--------------------------------------------------------------------------------

=head2 op_metronome_func

  op_metronome_func();

op_metronome_func function taken from the original C code.

=over

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub op_metronome_func() {
    #  static int
    #  op_metronome_func(char *arg)
    #  {
    #          op_retransmit(); /* retransmit any stale sends */
    #          (*op_metronome_callout)(op_metronome_callout_arg);
    #          return 0;
    #  }
    #

    #  op_retransmit();
    &$op_metronome_callout( $op_metronome_callout_arg );

} # End op_metronome_func

#--------------------------------------------------------------------------------

=head2 metronome_start

  metronome_start( mn                 => \%mn,
                   target_ops_per_sec => $rate,
                   lifetime           => $duration,
                   func               => \&op_metronome_func,
                   func_arg           => undef, );

metronome_start function taken from the original C code. Initializes
the metronome for Flg operation generation.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mn

(Required) : A metronome object

=item target_ops_per_sec

(Required) : Desire ops/sec rate

=item lifetime

(Required) : Length of execution

=item func

(Required) : Pointer to op_metronome_func, to generate the operations

=item func_arg

(Required) : Additional arguments to func

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub metronome_start(%) {
    #  int
    #  metronome_start(struct metronome *mn, int target_ops_per_sec, int lifetime,
    #                  int (*func)(char *), char *func_arg)
    #  {
    #          if (target_ops_per_sec <= 0) {
    #                  report_error(FATAL, "invalid ops/sec %d", target_ops_per_sec);
    #                  return -1;
    #          }
    #          mn->start = global_timer();
    #          mn->target_ops_per_sec = target_ops_per_sec;
    #          mn->lifetime = lifetime;
    #          mn->func = func;
    #          mn->func_arg = func_arg;
    #          mn->ops = 0;
    #          mn->msecs = 0;
    #          return 0;
    #  }
    #

    my (%opts) = validate (
        @_,
        {
            mn                 => { type => HASHREF },
            target_ops_per_sec => { type => SCALAR },
            lifetime           => { type => SCALAR },
            func               => { type => CODEREF },
            func_arg           => { type => SCALAR|UNDEF },
        }
    );

    if ( $opts{target_ops_per_sec} <= 0 ) {
        croak "metronome_start: invalid ops/sec $opts{target_ops_per_sec}";
    }
    $opts{mn}->{start}              = global_timer();
    $opts{mn}->{target_ops_per_sec} = $opts{target_ops_per_sec};
    $opts{mn}->{lifetime}           = $opts{lifetime};
    $opts{mn}->{func}               = $opts{func};
    $opts{mn}->{func_arg}           = $opts{func_arg};
    $opts{mn}->{ops}                = 0;
    $opts{mn}->{msecs}              = 0;

} # End metronome_start

#--------------------------------------------------------------------------------

=head2 metronome_active

  metronome_active( mn => \%mn );

metronome_active function taken from the original C code. Returns
true (1) if the metronome is still running.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mn

(Required) : The metronome object

=back

=item Returns

True if time is remaining on the metronome

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub metronome_active(%) {
    #  int
    #  metronome_active(struct metronome *mn)
    #  {
    #          return (mn->lifetime == 0 ||
    #                  (global_timer() - mn->start) / 1000000 < mn->lifetime);
    #  }

    my (%opts) = validate (
        @_,
        {
            mn => { type => HASHREF },
        }
    );

    # Use 2 lines for easier reading
    my $time_remains =      # Elapsed time less than desired duration
      ( global_timer() - $opts{mn}->{start} ) / 1000000 < $opts{mn}->{lifetime};
    return ( $opts{mn}->{lifetime} == 0 || $time_remains );

} # End metronome_active

#--------------------------------------------------------------------------------

=head2 metronome_tick

  metronome_tick( mn => \%mn )

metronome_tick function taken from the original C code. Generates the
next operation from the metronome.

The following named arguments are accepted:

=over

=item Arguments

=over

=item mn

(Required) : The metronome object

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub metronome_tick(%) {
    #  int
    #  metronome_tick(struct metronome *mn)
    #  {
    #          u_int64_t usecs = global_timer() - mn->start;
    #          long target_ops;
    #          int local_ops = 0, r;
    #
    #          mn->msecs = usecs / (u_int64_t)1000;
    #          target_ops = ((u_int64_t)mn->target_ops_per_sec * usecs) /
    #                  (u_int64_t)1000000;
    #
    #          while (mn->ops < target_ops) {
    #                  if ((r = (*mn->func)(mn->func_arg)) < 0) {
    #                          report_error(FATAL, "metronome func error %d", r);
    #                          return r;
    #                  }
    #                  mn->ops++;
    #                  local_ops++;
    #          }
    #
    #          return local_ops;
    #  }

    my (%opts) = validate (
        @_,
        {
            mn => { type => HASHREF },
        }
    );

    my $usecs     = global_timer() - $opts{mn}->{start};
    my $local_ops = 0;

    $opts{mn}->{msecs} = int ( $usecs / 1000 );
    my $target_ops = int ( $opts{mn}->{target_ops_per_sec} * $usecs / 1000000 );

    while ( $opts{mn}->{ops} < $target_ops ) {
        my $r = &{$opts{mn}->{func}}( $opts{mn}->{func_arg} );
        # Set $r to something, so code_check will not complain
        # $r is normally used to report an error, but the functions
        # themselves will do that.
        $r = 0;
        $opts{mn}->{ops}++;
        $local_ops++;
    }

    return 1;
} # End metronome_tick

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Operation SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    # Create a dummy nfs_obj object
    my %dhash = ( key1 => 1,
                  key2 => 2,
                  key3 => 3 );

    my $class = 'nfs::NfsClient';
    my $obj = bless { %dhash }, $class;

    # Validate that op_recs is not yet defined
    if (@op_recs) {
        Tharn::logresult('FAIL', abort => 1, "op_recs defined prematurely");
    }

    # Call initialization
    my $op_obj = nfs::Operation->op_init( client_obj => $obj );

    # Verify op_recs is now defined
    if (!@op_recs) {
        Tharn::logresult('FAIL', "op_recs not defined correctly");
    } elsif (!defined($op_obj)) {
        Tharn::logresult('FAIL', "op_obj not initialized correctly");
    } else {
        Tharn::logresult('PASS', "SELF TEST of nfs/Operation.pm");
    }
}

1;
