#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/PopulateTreeAcl.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All Rights reserved
#

package nfs::PopulateTreeAcl;

## @summary
##  Populate, set ACLs, or validate ACLs in directory hierarchy
##
## @description
##  Routines used to populate directory trees with files/dirs with ACLs
##  set, to change the ACLs of a populated directory tree, or to validate
##  the ACLs and files of a populated directory tree.
##
## @usage
## use nfs::PopulateTreeAcl;
##
## @status	review
##
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;
use Tharn;
use Cwd;
use Carp qw(croak);
use Params::Validate qw(validate BOOLEAN OBJECT SCALAR);
use SPARSE::Util::Logger;
use nfs::ACL::Acl qw(get_ace_array wait_for_file);
use nfs::GenericAcl qw(get_acl get_acl_array set_acl set_acl_ace print_acl print_aces create_with_acl check_aces_acl get_acl_filename);
use nfs::ACL::AclNFSV4 qw(:ACE4_TYPES :ACE4_FLAGS :ACE4_MASK);

use vars qw(
  $VERSION
  @ISA
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  populate_with_acl
  populate_cleanup
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
$VERSION = '0.01';

#-------------------------------------------------------
# Package local variables
#-------------------------------------------------------
my $params_file;
my $partial_file = 'partial';
my $verify_ack_file;
my $acl_shift = 0;

# Array of ACLs used to set on files. The ACLs are pre-built and
# constant so we know what value should be set on a file based
# on the ACL number used on the file and the current $acl_shift.
my @dir_acls;
my @file_acls;

# Need a fairly long timeout to wait for validation to finish
my $file_wait_timeout = 15 * 60;

# The ACL number part of the file is always the last N characters
# and starts with an _ followed by 4 digits
my $acl_num_len = 5;

# The maximum number of ACEs supported in one ACL by the
# synthetic client. The actual number varies since it puts it into
# a 4K buffer and the size of the who field varies.
my $max_aces = 160;

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub write_params_file( %; );
sub read_params_file( \%; );
sub write_partial_file( $$$; );
sub read_partial_file( $; );
sub gen_file_name( $$; );
sub get_acl_from_filename( $; );
sub build_acls( $$$; );
sub chk_verify_in_progress( %; );
sub populate( %; );
sub setacl_curdir_recur( $%; );
sub setacl( %; );
sub verify_curdir_recur( $%; );
sub verify_curdir_synth_recur( $%; );
sub verify( %; );
sub populate_with_acl( %; );
sub populate_cleanup( %; );

# @name ::write_params_file
# @description
#  Write out the parameters needed for validating this tree
#  later to the parameters file.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing, croaks if file cannot be created

sub write_params_file( %; ) {
    my %opts = @_;

    logDbg("write_params_file: called");

    open PARAM, "+> $params_file" or
        croak "write_params_file: create failed: $params_file: $!";
    print PARAM "DIRS_PER_DIR=$opts{dirs_per_dir}\n";
    print PARAM "FILES_PER_DIR=$opts{files_per_dir}\n";
    print PARAM "INHERIT=$opts{inherit}\n";
    print PARAM "NUMFILES=$opts{numfiles}\n";
    print PARAM "ACL_SHIFT=$acl_shift\n";
    close PARAM;
}

# @name ::read_params_file
# @description
#  Read the parameters file and set the global parameters
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing, croaks if file cannot be opened

sub read_params_file( \%; ) {
    my ($optsref) = @_;
    my ($var, $value);

    logDbg("read_params_file: called");

    open PARAM, "< $params_file" or
        croak "read_params_file: open failed: $params_file: $!";
    while(<PARAM>) {
        chomp;
        ($var, $value) = split('=', $_);
        if ($var eq 'DIRS_PER_DIR') {
            $$optsref{dirs_per_dir} = $value;
        } elsif ($var eq 'FILES_PER_DIR') {
            $$optsref{files_per_dir} = $value;
        } elsif ($var eq 'INHERIT') {
            $$optsref{inherit} = $value;
        } elsif ($var eq 'NUMFILES') {
            $$optsref{numfiles} = $value;
        } elsif ($var eq 'ACL_SHIFT') {
            $acl_shift = $value;
        }
    }
    close PARAM;
}

# @name ::write_partial_file
# @description
#  Write the partial file contents which includes the number of files
#  and directories that should be in the current directory. This file
#  only exists in a partially populated directory. The path to the partial
#  file can get long and PERL seems to only handle paths that are 1024. For
#  that reason change to the directory where the partial file is being
#  created and change back.
#  INTERNAL SUBROUTINE
#
# @arg
#  curdir		Required: Directory to create file in.
#
#  file_cnt		Required: Number of files in directory.
#
#  dir_cnt		Required: Number of dirs in directory.
#
# @return
#  Nothing, croaks if file cannot be created

sub write_partial_file( $$$; ) {
    my ($curdir, $file_cnt, $dir_cnt) = @_;
    my ($cwd, $status);

    $cwd = getcwd();
    chdir $curdir;
    $status = open PART, "+> $partial_file";
    chdir $cwd;
    if (!$status) {
        croak "write_partial_file: create failed: $curdir/$partial_file: $!";
    }
    print PART "curdir_file_cnt=$file_cnt\n";
    print PART "curdir_dir_cnt=$dir_cnt\n";
    close PART;
}

# @name ::read_partial_file
# @description
#  Read the partial file contents and return the values. The path to the
#  partial file can get long and PERL seems to only handle paths that are
#  1024. For that reason change to the directory where the partial file
#  is being read and change back.
#  INTERNAL SUBROUTINE
#
# @arg
#  curdir		Required: Directory to create file in.
#
# @return
#  ($file_cnt, $dir_cnt) Number of files and directories in directory.
#   croaks if the file cannot be opened

sub read_partial_file( $; ) {
    my ($curdir) = @_;
    my ($cwd, $file_cnt, $dir_cnt, $status, $var, $value);

    $cwd = getcwd();
    chdir $curdir;
    $status = open PART, "< $partial_file";
    chdir $cwd;
    if (!$status) {
        croak "read_partial_file: open failed: $curdir/$partial_file: $!";
    }
    while(<PART>) {
        ($var, $value) = split('=', $_);
        if ($var eq 'curdir_file_cnt') {
            $file_cnt = $value;
        } elsif ($var eq 'curdir_dir_cnt') {
            $dir_cnt = $value;
        }
    }
    close PART;
    return ($file_cnt, $dir_cnt);
}

# @name ::gen_file_name
# @description
#  Generate a random file name with a length between the minimum and
#  maximum file length. The passed in file number is used to ensure that
#  the file name is unique within the directory.
#  INTERNAL SUBROUTINE
#
# @arg
#  name		Required: File number of the file.
#
#  maxlen	Required: Maximum length of the file name.
#
# @return
#  Generated file name

sub gen_file_name( $$; ) {
    my ($name, $maxlen) = @_;
    my $comp = "Zz";
    my $comp_len = length($comp);
    my $len = int(rand ($maxlen + 1));

    # Subtract the length of the file number from the string len
    # Make sure that the name length has room for the ACL number
    $len -= length($name);
    $len = $acl_num_len if ($len < $acl_num_len);

    if ($len > 0) {
        # Generate the name
        $name .= $comp x int($len/$comp_len);

        # Subtract the length just added to the name. The length left is
        # the remainder which is less than $comp_len
        $len -= int($len/$comp_len) * $comp_len;

        # Pad the name to get it to the requested len and add the ACL marker
        while($len > 0) {
            $name .= 'Q';
            $len--;
        }
    }
    return $name;
}

# @name ::get_acl_from_filename
# @description
#  Return ACL given the file name. The ACL number should always be at
#  the end of the filename after the '_'.
#  INTERNAL SUBROUTINE
#
# @arg
#  name		Required: Name of the file.
#
# @return
#  ACL object or croaks if the ACL number is not valid.

sub get_acl_from_filename( $; ) {
    my ($file) = @_;
    my ($acl, $acl_num);

    # Get the ACL number from the file name
    ($acl_num) = $file =~ m/_(\d+)$/;
    if (!defined($acl_num)) {
        croak "get_acl_from_filename: $file doesn't have ACL number extension";
    }

    if (-d $file) {
        # Validate the ACL number.
        if ($acl_num >= 0 && $acl_num <= $#dir_acls) {
            # Add in the current ACL number shift
            $acl_num += $acl_shift;
            $acl_num -= ($#dir_acls + 1) if ($acl_num > $#dir_acls);
            $acl = $dir_acls[$acl_num];
        } else {
            croak "get_acl_from_filename: $file mapped to $acl_num which " .
                  "is invalid dir_acls: $#dir_acls";
        }
    } else {
        # Validate the ACL number.
        if ($acl_num >= 0 && $acl_num <= $#file_acls) {
            # Add in the current ACL number shift
            $acl_num += $acl_shift;
            $acl_num -= ($#dir_acls + 1) if ($acl_num > $#dir_acls);
            $acl = $file_acls[$acl_num];
        } else {
            croak "get_acl_from_filename: $file mapped to $acl_num which " .
                  "is invalid file_acls: $#file_acls";
        }
    }
    return ($acl);
}

# @name ::build_acls
# @description
#  Build the directory and file ACLs used for this test. It will vary
#  11 mask bits over the GROUP@ and EVERYONE@ ACEs creating a total
#  of 121 file ACLs and 121 dir ACLs. Note that directory and file ACLs
#  usually use a different base ACL. Does not modify the settings for OWNER@
#  to avoid cleanup issues.
#  INTERNAL SUBROUTINE
#
# @arg
#  aref			Required: Reference to the array of ACLs to populate.
#
#  bref			Required: Reference to the base ACL used to create
#			the populated ACLs.
#
#  large_num_aces	Required: When set build one ACL with a very large
#			number of ACEs.
#
# @return
#  Croaks if it cannot find GROUP@ or EVERYONE@ in the $bref ACL

sub build_acls( $$$; ) {
    my ($aref, $bref, $large_num_aces) = @_;
    my ($ace, $acecnt, $aclcnt, $didx, $ebit, $emask, $gbit, $gmask, $sidx,
        $gallow_idx, $gdeny_idx, $eallow_idx, $edeny_idx);
    my @mask_bits = ( ACE4_LIST_DIRECTORY, ACE4_ADD_FILE, ACE4_APPEND_DATA,
                      ACE4_EXECUTE, ACE4_READ_NAMED_ATTRS,
                      ACE4_WRITE_NAMED_ATTRS, ACE4_READ_ATTRIBUTES,
                      ACE4_WRITE_ATTRIBUTES, ACE4_READ_ACL, ACE4_WRITE_ACL,
                      ACE4_SYNCHRONIZE
                    );
    $gallow_idx = $gdeny_idx = $eallow_idx = $edeny_idx = -1;

    $sidx = 0;
    foreach $ace (@{$bref}) {
        if ($ace->{who} eq 'GROUP@') {
            if ($ace->{type} == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
                $gallow_idx = $sidx;
            } elsif ($ace->{type} == ACE4_ACCESS_DENIED_ACE_TYPE) {
                $gdeny_idx = $sidx;
            }
        } elsif ($ace->{who} eq 'EVERYONE@') {
            if ($ace->{type} == ACE4_ACCESS_ALLOWED_ACE_TYPE) {
                $eallow_idx = $sidx;
            } elsif ($ace->{type} == ACE4_ACCESS_DENIED_ACE_TYPE) {
                $edeny_idx = $sidx;
            }
        }
        $sidx++;
    }
    if ($gallow_idx == -1 || $gdeny_idx == -1 ||
        $eallow_idx == -1 || $edeny_idx == -1) {
        croak "build_acls: Could not determine group/everyone and/or deny " .
              "ACE idx";
    }

    # Create the directory ACEs
    # Use default settings for OWNER@ and don't vary them to avoid
    # cleanup issues. Only vary GROUP@ and EVERYONE@ and start by
    # clearing out these masks.
    $$bref[$gallow_idx]{mask} = 0;
    $$bref[$gdeny_idx]{mask} = 0;
    $$bref[$eallow_idx]{mask} = 0;
    $$bref[$edeny_idx]{mask} = 0;

    $acecnt = 0;
    $didx = 0;
    $aclcnt = 0;
    $gmask = 0;
    foreach $gbit (@mask_bits) {
        # For the group ACEs take the template and make sure the gmask
        # bits are set on the allow and not set on the deny.
        $gmask |= $gbit;
        $$bref[$gallow_idx]{mask} |= $gmask;
        $$bref[$gdeny_idx]{mask} &= ~$gmask;

        $emask = 0;
        foreach $ebit (@mask_bits) {
            # For the everyone ACEs take the template and make sure the emask
            # bits are set on the allow and not set on the deny.
            $emask |= $ebit;
            $$bref[$eallow_idx]{mask} |= $emask;
            $$bref[$edeny_idx]{mask} &= ~$emask;
            $sidx = 0;
            # If the large number of ACEs flag is not set then increment
            # the destination ACE count otherwise all the ACEs should go
            # into the same ACL
            if ($large_num_aces == 0) {
                $didx = 0;
                $acecnt = 0;
            }
            foreach $ace (@{$bref}) {
                # If the mask is 0 then skip it
                if ($$bref[$sidx]->{mask} == 0) {
                    $sidx++;
                    next;
                }
                $$aref[$aclcnt][$didx]->{type} = $$bref[$sidx]->{type};
                $$aref[$aclcnt][$didx]->{flags} = $$bref[$sidx]->{flags};
                $$aref[$aclcnt][$didx]->{mask} = $$bref[$sidx]->{mask};
                $$aref[$aclcnt][$didx]->{who} = $$bref[$sidx]->{who};
                $sidx++;
                $didx++;
                $acecnt++;
                if ($acecnt >= $max_aces) {
                    return;
                }
            }
            # If the large number of ACEs flag is not set then increment
            # the ACL count otherwise all the ACEs go into the same ACL
            if ($large_num_aces == 0) {
                $aclcnt++;
            }
        }
    }
}

# @name ::chk_verify_in_progress
# @description
#  Check if the verify file is there which indicates that a
#  verification process is running. If the file is there then create
#  the verify_ack file and wait until it is removed.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub chk_verify_in_progress( %; ) {
    my %opts = @_;
    my ($status);

    if (-f "$opts{dirname}_verify") {
        open MODE, "+> $verify_ack_file" or
            croak "chk_verify_in_progress: create failed: " .
                  "$verify_ack_file: $!";
        close MODE;

        # Wait for verify to remove the ack file before continuing
        logDbg("chk_verify_in_progress: wait for removal of $verify_ack_file");
        $status = wait_for_file(filename => "$verify_ack_file",
                                op => 'remove',
                                timeout => $file_wait_timeout);
        if ($status != 0) {
            croak "chk_verify_in_progress: timeout waiting for " .
                  "removal of verify_ack look at verify process";
        }
    }
}

# @name ::populate
# @description
#  Populate a directory tree with files and directories setting the ACLs
#  along the way.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub populate( %; ) {
    my %opts = @_;
    my ($aces, $acl, $curdir, $curdir_file_cnt, $curdir_dir_cnt, $curdir_ace,
        $curfile_ace, @dirs_to_populate, $filetype, $fname, $nlen, $status);
    my $num_files = 0;

    $curdir = $opts{dirname};
    logDbg("populate: called: $curdir");

    write_params_file(%opts);

    $nlen = $acl_num_len - 1;
    $curdir_ace = $curfile_ace = 0;
    $curdir_file_cnt = $curdir_dir_cnt = 0;
    while ($num_files < $opts{numfiles} && time < $opts{end_time}) {
        # Check if the verification is in progress and if so wait until
        # it is finished.
        chk_verify_in_progress(%opts);

        $fname = gen_file_name($curdir_dir_cnt + $curdir_file_cnt,
                               $opts{namelen_max});
        if ($curdir_dir_cnt < $opts{dirs_per_dir}) {
            # Create another directory
            my $t = sprintf "%0" . $nlen . "d", $curdir_ace;
            $fname =~ s/.....$/_$t/;
            $filetype = 'dir';
            $curdir_dir_cnt++;

            $aces = $dir_acls[$curdir_ace];
            if ($curdir_ace >= $#dir_acls) {
                 $curdir_ace = 0;
            } else {
                 $curdir_ace++;
            }
            push @dirs_to_populate, "$curdir/$fname";
        } elsif ($curdir_file_cnt < $opts{files_per_dir}) {
            # Create another file in the current directory.
            my $t = sprintf "%0" . $nlen . "d", $curfile_ace;
            $fname =~ s/.....$/_$t/;
            $filetype = 'file';
            $curdir_file_cnt++;

            $aces = $file_acls[$curfile_ace];
            if ($curfile_ace >= $#file_acls) {
                 $curfile_ace = 0;
            } else {
                 $curfile_ace++;
            }
        }
        $acl = create_with_acl(ace => $aces, fc => $opts{fc},
                               filename => "$curdir/$fname",
                               filetype => $filetype,
                               filer_path => $opts{filer_path},
                               filer_vol => $opts{filer_vol},
                               mountpoint => $opts{mountpoint},
                               use_synthetic => $opts{use_synthetic},
                               retry_conn => $opts{retry_conn});

        $num_files++;
        # Update the partial file
        write_partial_file($curdir, $curdir_file_cnt, $curdir_dir_cnt);

        # Verify that the file/dir was created with the correct ACL
        $status = check_aces_acl(acl => $acl, aces => $aces);
        if ($status != 0) {
            print_aces(aces => $aces);
            print_acl(acl => $acl);
            croak "populate: Returned ACL not same as passed in ACEs\n";
        }

        if ($filetype eq 'dir') {
            # If we just created a directory create a partial file for it
            # to show that it has not been populated yet. This is done
            # so the verification process can tell the difference between
            # a deliberate empty directory or one the server zapped.
            write_partial_file("$curdir/$fname", 0, 0);
        }

        # If this directory is full then move on to the next
        # directory in the list to populate.
        if ($curdir_dir_cnt >= $opts{dirs_per_dir} &&
            $curdir_file_cnt >= $opts{files_per_dir}) {

            # Remove the partial file
            unlink "$curdir/$partial_file" or
                croak "populate: remove of partial file failed: $!";

            $curdir_dir_cnt = 0;
            $curdir_file_cnt = 0;
            $curdir = shift @dirs_to_populate;
            if (!$curdir && $num_files < $opts{numfiles}) {
                croak "populate: No more directories to populate";
            }
        }
    }
}

# @name ::setacl_curdir_recur
# @description
#  Continuously go through a directory tree changing the ACLs
#  along the way using the synthetic client. Uses a depth first descent.
#  INTERNAL SUBROUTINE
#
# @arg
#  curdir	Required: Current directory name.
#
#  opts		Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub setacl_curdir_recur( $%; ) {
    my ($curdir, %opts) = @_;
    my ($aces, $acl, @allfiles, $curdir_dir_cnt, $curdir_file_cnt, $dir_handle,
        $exp_dir_cnt, $exp_file_cnt, $file, $status);

    logDbg("setacl_curdir_recur: called: $curdir");

    # Read the entire current directory
    opendir($dir_handle, $curdir) or
        croak "setacl_curdir_recur: Could not open current directory";
    @allfiles = readdir $dir_handle or
        croak "setacl_curdir_recur: Could not read from dir $!";
    closedir $dir_handle;

    #pop off . and .. from array
    shift(@allfiles);
    shift(@allfiles);

    $exp_file_cnt = $opts{files_per_dir};
    $exp_dir_cnt = $opts{dirs_per_dir};
    $curdir_file_cnt = 0;
    $curdir_dir_cnt = 0;
    foreach $file (@allfiles) {
        # If there is a partial file in this directory then get the
        # number of files and dirs that were created.
        if ($file eq $partial_file) {
            ($exp_file_cnt, $exp_dir_cnt) = read_partial_file($curdir);
            next;
        }

        # Get the new ACE
        $aces = get_acl_from_filename("$curdir/$file");
        # Get the current ACL on the file
        $acl = get_acl(fc => $opts{fc}, filename => "$curdir/$file",
                       filer_path => $opts{filer_path},
                       filer_vol => $opts{filer_vol},
                       mountpoint => $opts{mountpoint},
                       use_synthetic => $opts{use_synthetic},
                       retry_conn => $opts{retry_conn});
        # Set the new ACEs on the ACL object
        for (my $i = 0; $i <= $#{$aces}; $i++) {
            set_acl_ace(acl => $acl, consistent => 'no',
                        type => $$aces[$i]{type},
                        flags => $$aces[$i]{flags},
                        flags_how => 'SET',
                        mask => $$aces[$i]{mask},
                        mask_how => 'SET',
                        who => $$aces[$i]{who});
        }
        $status = set_acl(acl => $acl);
        if ($status ne 'NFS4_OK') {
            print_acl(acl => $acl);
            croak "setacl_curdir_recur: Could not set ACL";
        }

        if (-d "$curdir/$file") {
            $curdir_dir_cnt++;
            setacl_curdir_recur("$curdir/$file", %opts);
        } else {
            $curdir_file_cnt++;
        }
    }

    # Verify that  the file and dir counts match exactly.
    if ($curdir_file_cnt != $exp_file_cnt) {
        croak "setacl_curdir_recur: Expected file count did not match " .
              "actual. exp file cnt: $exp_file_cnt file cnt: $curdir_file_cnt";
    }
    if ($curdir_dir_cnt != $exp_dir_cnt) {
        croak "setacl_curdir_reg: Expected dir count did not match " .
              "actual. exp dir cnt: $exp_dir_cnt dir cnt: $curdir_dir_cnt";
    }
}

# @name ::setacl
# @description
#  Increment the ACL shift value and call the recursive setacl routine
#  to change the ACLs on an entire tree
#
#  An ACL shift parameter is defined in the ACL parameter file that gives
#  an integer to add to the files ACL number to determine the ACL that
#  it currently maps to this is done for the entire directory tree. With
#  this approach is important to rewrite all the ACLs for the directory
#  tree before exiting so that validation will work properly.
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub setacl( %; ) {
    my %opts = @_;

    read_params_file(%opts);

    do {
        # Check if the verification is in progress and if so wait until
        # it is finished.
        chk_verify_in_progress(%opts);

        # Increment the ACL shift counter and save the parameters.
        $acl_shift++;
        write_params_file(%opts);
        setacl_curdir_recur($opts{dirname}, %opts);
    } while (time < $opts{end_time});
}

# @name ::verify_curdir_recur
# @description
#  Recursively descend a directory tree using local directory routines.
#  Validate the ACLs along the way using either local or synthetic client
#  depending on the settings. Uses a depth first descent.
#  INTERNAL SUBROUTINE
#
# @arg
#  curdir	Required: Current directory name.
#
#  opts		Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub verify_curdir_recur( $%; ) {
    my ($curdir, %opts) = @_;
    my ($aces, $acl, @allfiles, $curdir_file_cnt, $curdir_dir_cnt, $dir_handle,
        $exp_dir_cnt, $exp_file_cnt, $file, $status);

    logDbg("verify_curdir_recur: called: $curdir");

    # Read the entire current directory
    opendir($dir_handle, $curdir) or
        croak "verify_curdir_recur: could not open current directory";
    @allfiles = readdir $dir_handle or
        croak "verify_curdir_recur: Could not read from dir $!";
    closedir $dir_handle;

    #pop off . and .. from array
    shift(@allfiles);
    shift(@allfiles);

    $exp_file_cnt = $opts{files_per_dir};
    $exp_dir_cnt = $opts{dirs_per_dir};
    $curdir_file_cnt = 0;
    $curdir_dir_cnt = 0;
    foreach $file (@allfiles) {
        # If there is a partial file in this directory then get the
        # number of files and dirs that were created.
        if ($file eq $partial_file) {
            ($exp_file_cnt, $exp_dir_cnt) = read_partial_file($curdir);
            next;
        }

        # Get the expected ACL ACEs
        $aces = get_acl_from_filename("$curdir/$file");
        # Get the actual ACL
        $acl = get_acl(fc => $opts{fc}, filename => "$curdir/$file",
                       filer_path => $opts{filer_path},
                       filer_vol => $opts{filer_vol},
                       mountpoint => $opts{mountpoint},
                       use_synthetic => $opts{use_synthetic},
                       retry_conn => $opts{retry_conn});

        # Verify that the file/dir was created with the correct ACEs
        logDbg("verify_curdir_recur: verify ACEs for $file");
        $status = check_aces_acl(acl => $acl, aces => $aces);
        if ($status != 0) {
            print_aces(aces => $aces);
            print_acl(acl => $acl);
            croak "verify_curdir_recur: ACL not same as expected ACEs\n";
        }

        if (-d "$curdir/$file") {
            $curdir_dir_cnt++;
            verify_curdir_recur("$curdir/$file", %opts);
        } else {
            $curdir_file_cnt++;
        }
    }

    # Verify that  the file and dir counts match exactly.
    if ($curdir_file_cnt != $exp_file_cnt) {
        croak "verify_curdir_recur: Expected file count did not match " .
              "actual. exp file cnt: $exp_file_cnt file cnt: $curdir_file_cnt";
    }
    if ($curdir_dir_cnt != $exp_dir_cnt) {
        croak "verify_curdir_recur: Expected dir count did not match " .
              "actual. exp dir cnt: $exp_dir_cnt dir cnt: $curdir_dir_cnt";
    }
}

# @name ::verify_curdir_synth_recur
# @description
#  Recursively descend a directory tree verifying the ACLs
#  along the way using the synthetic client READDIR routine
#  to get both directory information and ACLs in one call.
#  Uses a depth first descent.
#  INTERNAL SUBROUTINE
#
# @arg
#  curdir	Required: Current directory name.
#
#  opts		Required: A hash of options.
#
# @return
#  Nothing, croaks if an error is detected

sub verify_curdir_synth_recur( $%; ) {
    my ($curdir, %opts) = @_;
    my ($aces, $acl, @acls, $curdir_file_cnt, $curdir_dir_cnt, $exp_dir_cnt,
        $exp_file_cnt, $file, $status);

    logDbg("verify_curdir_synth_recur: called: $curdir");

    ## @step Read directory and get all ACLs for files
    logStep("Read directory and get all ACLs for files");
    @acls = get_acl_array(fc => $opts{fc}, filename => $curdir,
                          filer_path => $opts{filer_path},
                          filer_vol => $opts{filer_vol},
                          mountpoint => $opts{mountpoint},
                          retry_conn => $opts{retry_conn});

    $exp_file_cnt = $opts{files_per_dir};
    $exp_dir_cnt = $opts{dirs_per_dir};
    $curdir_file_cnt = 0;
    $curdir_dir_cnt = 0;
    foreach $acl (@acls) {
        # Note that the filename returned will have the full pathname
        $file = get_acl_filename(acl => $acl);

        # If there is a partial file in this directory then get the
        # number of files and dirs that were created.
        if ($file eq "$curdir/$partial_file") {
            ($exp_file_cnt, $exp_dir_cnt) = read_partial_file($curdir);
            next;
        }

        # Get the expected ACL ACEs
        $aces = get_acl_from_filename($file);

        # Verify that the file/dir was created with the correct ACEs
        logDbg("verify_curdir_synth_recur: verify ACEs for $file");
        $status = check_aces_acl(acl => $acl, aces => $aces);
        if ($status != 0) {
            print_aces(aces => $aces);
            print_acl(acl => $acl);
            croak "verify_curdir_synth_recur: ACL not same as expected ACEs\n";
        }

        if (-d $file) {
            $curdir_dir_cnt++;
            verify_curdir_synth_recur("$file", %opts);
        } else {
            $curdir_file_cnt++;
        }
    }

    # Verify that  the file and dir counts match exactly.
    if ($curdir_file_cnt != $exp_file_cnt) {
        croak "verify_curdir_synth_recur: Expected file count did not match " .
              "actual. exp file cnt: $exp_file_cnt file cnt: $curdir_file_cnt";
    }
    if ($curdir_dir_cnt != $exp_dir_cnt) {
        croak "verify_curdir_synth_recur: Expected dir count did not match " .
              "actual. exp dir cnt: $exp_dir_cnt dir cnt: $curdir_dir_cnt";
    }
}

# @name ::verify
# @description
#  Call either the client-based verification routine or the
#  one that uses the synthetic client READDIR
#  INTERNAL SUBROUTINE
#
# @arg %opts	Required: A hash of options.
#
# @return
#  Nothing

sub verify( %; ) {
    my %opts = @_;

    # If another test is running, wait for the verify_ack file
    # to be created by another process to indicate that they
    # see verification wants to run and they are paused waiting
    # for verification to finish (i.e. until the ack file is removed).
    if (-f "$opts{dirname}_populate" || -f "$opts{dirname}_setacl") {
        logDbg("verify: wait for creation of $verify_ack_file");
        my $status = wait_for_file(filename => "$verify_ack_file",
                                   op => 'create',
                                   timeout => $file_wait_timeout);
        if ($status != 0) {
            croak "verify: timeout waiting for creation of verify_ack look " .
                  "at other processes that called populate_with_acl";
        }
    }

    read_params_file(%opts);

    if ($opts{readdir}) {
        verify_curdir_synth_recur($opts{dirname}, %opts);
    } else {
        verify_curdir_recur($opts{dirname}, %opts);
    }
}

## @name ::populate_with_acl
##
## @description
##  This library has 3 different modes of operation. The population mode
##  is used to to populate a directory hierarchy with files/dirs that have
##  known ACLs set on them. The setacl mode is used to change the ACLs to
##  known values on the files/dirs in a previously populated directory
##  hierarchy. The verify mode is used to verify that all the files/dirs that
##  were previously populated exist and have the correct ACL set on them.
##
##  The population or setacl mode can be run continuously in the background
##  by specifying the background option and by not specifying the numfiles
##  nor the runtime option.
##
##  When a directory is being populated a file named 'partial' is created
##  to indicate that the directory has not yet been fully populated. This
##  file contains the current number of files and directories in the current
##  directory. If a directory does not have a partial file then it is fully
##  populated. When a new directory is created it will have a partial file
##  indicating there are 0 files and directories in the new directory.
##
##  While this test is running the following files can be created
##  in $opts{dirname} to control operation of the population process:
##
##	params		Contains the parameters used to populate the tree.
##			This file is created by the population mode, updated
##			by the setacl mode, and read by the verification mode.
##
##	population	A test is running that is populating this tree.
##
##	setacl		A test is running that is changing ACLs in the tree.
##
##	verify		A test is running that is verifying the ACLs in
##			the tree. Tests that are doing population or setacl
##			will watch for this file and create the verify_ack
##			file to acknowledge they see a request for verification
##			and to pause what they are doing. Population will
##			pause when getting ready to move to another directory.
##                      Setacl will only pause after having modified the
##                      entire tree. When the verification is complete it
##			will remove the verify_ack file to indicate to
##			population or setacl tests that they can continue.
##
##	verify_ack	File created by population and setacl tests to
##			acknowledge that they have paused waiting for
##			verification to complete. The paused test will
##			watch for this file to be removed before resuming
##			operation.
##
## @arg %opts	Required: A hash of options that include:
##
##  background		Optional: (boolean) When set indicates that this
##			subroutine should be sub-tested and run in the
##			background. A subtest object will be returned.
##			Only the populate and setacl mode can be run in
##			the background. Default is not set.
##
##  dirname		Required: Directory where to start performing
##			the operations.
##
##  dirs_per_dir	Optional: Number of directories to put in a directory.
##			When set to 0, only files will be created and not
##			directories. Default is 20.
##
##  fc			Required: Connection to filer
##
##  filer_path		Required: Full pathname of volume on filer.
##
##  filer_vol		Required: Volume name to use on filer.
##
##  files_per_dir	Optional: Number of files to put in a directory.
##			When set to 0, only directories will be created
##			and not files. Default is 50.
##
##  inherit		Optional: (boolean) When set indicates that the
##			inheritable ACE should be set on DIRNAME
##			and that the verification routines should verify
##			the ACEs were inherited. This flag is ignored by
##			the 'setacl' mode. Default is not set.
##
##  large_num_aces	Optional: (boolean) When set indicates that the
##			built ACLs should have a large number of ACEs.
##			Instead of creating a lot of different ACLs just
##			create one with a large number of ACEs.
##			Default is not set.
##
##  mode		Required: Defines the mode of operation. The
##			possible values are: 'populate', 'setacl', or
##			'verify'. 'populate' will create files/dir in
##			the directory tree passing in the ACL on
##			creation. 'setacl' will change the ACLs on
##			existing files/dirs in the tree. 'verify'
##			will verify the ACLs on the files/dir in
##			the tree.
##
##  mountpoint		Required: The mountpoint where the filer volume
##			is mounted.
##
##  namelen_max		Optional: Set the maximum for the file name
##			length. The length will vary randomly from
##			8 characters up to this length. The default
##			is 64.
##
##  numfiles		Optional: Set to the number of files/dirs
##			to create in population mode. Defaults to
##			2 billion.
##
##  readdir		Optional: (boolean) When set indicates that the
##			synthetic client READDIR call should be used
##			to get the directory contents and ACLs. Default is
##			not set.
##
##  runtime		Optional: Set to the number of seconds to
##			run before stopping. Defaults to 4 billion.
##
##  use_synthetic	Optional: (boolean) When set use the synthetic
##			client otherwise use the CLIENT OS to create files
##			and get ACLs. Default is not set.
##
## @return
##  subtest object if background flag was set, otherwise nothing is returned
##  croaks if an error is detected

sub populate_with_acl( %; ) {
    my %opts = validate(
        @_,
        {
            background      => { type => BOOLEAN, default => 0 },
            dirname         => { type => SCALAR },
            dirs_per_dir    => { type => SCALAR, default => 20 },
            fc              => { type => OBJECT },
            filer_path      => { type => SCALAR },
            filer_vol       => { type => SCALAR },
            files_per_dir   => { type => SCALAR, default => 50 },
            inherit         => { type => BOOLEAN, default => 0 },
            large_num_aces  => { type => BOOLEAN, default => 0 },
            mode            => { type => SCALAR,
                                 regex => qr/^(populate|verify|setacl)$/ },
            mountpoint      => { type => SCALAR },
            namelen_max     => { type => SCALAR, default => 64 },
            numfiles        => { type => SCALAR, default => 0x7FFFFFFF },
            readdir         => { type => BOOLEAN, default => 0 },
            runtime         => { type => SCALAR, default => 0 },
            use_synthetic   => { type => BOOLEAN, default => 0 },
            retry_conn      => { type => SCALAR, default => 0 },
        }
    );
    my (@aces, @dir_inherit_aces, @dir_init_aces, @file_inherit_aces,
        @file_init_aces, $mode_file, $sobj);

    if ($opts{dirs_per_dir} <= 0 && $opts{files_per_dir} <= 0) {
        croak "populate_with_acl: One or both of dirs_per_dir and " .
              "files_per_dir must be greater than 0";
    }
    if ($opts{dirs_per_dir} == 0 && $opts{files_per_dir} < $opts{numfiles}) {
        # If no directories are being created then the maximum
        # number of files being created is files_per_dir
        $opts{numfiles} = $opts{files_per_dir};
        logcomment('Set numfiles to files_per_dir since dirs_per_dir was 0');
    }

    # Run this routine in the background if requested
    if ($opts{background}) {
        if ($opts{mode} eq 'verify') {
            croak "populate_with_acl: Cannot do verification in the background";
        }

        $opts{background} = 0;
        $sobj = subtest( sub {populate_with_acl(%opts);}, -bg, '-onexit',
                         sub {populate_cleanup(dirname => $opts{dirname},
                                               mode => $opts{mode})} );
        croak $@ if ($@);
        return ($sobj);
    }

    # Store the ending time in the options hash
    if ($opts{runtime} == 0) {
        $opts{end_time} = 0xFFFFFFFF;
    } else {
        $opts{end_time} = time + $opts{runtime};
    }

    # Create the state file names based on the directory to populate
    $mode_file = "$opts{dirname}_$opts{mode}";
    $params_file = "$opts{dirname}_params";
    $verify_ack_file = "$opts{dirname}_verify_ack";

    # Seed the random number generator with a constant so we get the
    # same random numbers each time the program is run for reproducibility
    # purposes.
    srand 12345678;

    # Get the default dir and file ACE arrays.
    @dir_init_aces = get_ace_array(type => 'dir_aces');
    @file_init_aces = get_ace_array(type => 'file_aces');

    # Build the ACLs to use for the test
    build_acls(\@dir_acls, \@dir_init_aces, $opts{large_num_aces});
    build_acls(\@file_acls, \@file_init_aces, $opts{large_num_aces});

    if ($opts{mode} eq 'populate') {
        if ($opts{inherit}) {
            @dir_init_aces = get_ace_array(type => 'dir_aces');
            @file_inherit_aces = get_ace_array(type => 'file_inherit_aces');
            @dir_inherit_aces = get_ace_array(type => 'dir_inherit_aces');
            @aces = (
                @dir_init_aces,
                @file_inherit_aces,
                @dir_inherit_aces
            );

            # Create directory with inherit ACEs for everyone
            create_with_acl(ace => \@aces, fc => $opts{fc},
                            filename => $opts{dirname},
                            filetype => 'dir',
                            filer_path => $opts{filer_path},
                            filer_vol => $opts{filer_vol},
                            mountpoint => $opts{mountpoint},
                            use_synthetic => $opts{use_synthetic},
                            retry_conn => $opts{retry_conn});
        } else {
            # Create directory with default ACEs
            mkdir("$opts{dirname}") or
                croak "populate_with_acl: mkdir failed: $opts{dirname}: $!";
        }
    }

    # Create mode file indicator
    open MODE, "+> $mode_file" or
        croak "populate_with_acl: mode state create failed: $mode_file: $!";
    close MODE;

    if ($opts{mode} eq 'populate') {
        # Populate directory
        populate(%opts);
    } elsif ($opts{mode} eq 'setacl') {
        # Set ACLs in directory tree
        setacl(%opts);
    } else {
        # Verify ACLs in directory tree
        verify(%opts);
        # Delete the verify_ack file
        unlink $verify_ack_file;
    }
    populate_cleanup(dirname => $opts{dirname}, mode => $opts{mode});
    return;
}

## @name ::populate_cleanup
## @description
##  Cleanup state files that a call to populate_with_acl() may have
##  left around if it was abnormally terminated. If mode is specified
##  then only remove the files associated with the specified mode.
##  If mode is not specified then cleanup all the files.
##
## @arg %opts	Required: A hash of options.
##
##  dirname		Required: Directory where the state files may
##			exist.
##
##  mode		Optional: Defines the mode of operation. The
##			possible values are: 'populate', 'setacl', or
##			'verify'.
##
## @return
##  Nothing, croak if an error is detected

sub populate_cleanup( %; ) {
    my %opts = validate(
        @_,
        {
            dirname         => { type => SCALAR },
            mode            => { type => SCALAR, optional => 1,
                                 regex => qr/^(populate|verify|setacl)$/ },
        }
    );
    my ($file, @list);

    if ($opts{mode}) {
        push @list, "$opts{dirname}_$opts{mode}";
    } else {
        @list = ( "$opts{dirname}_populate",
                  "$opts{dirname}_params",
                  "$opts{dirname}_verify",
                  "$opts{dirname}_setacl"
                );
    }

    # If mode is not defined or this is verify mode then delete the ACK file
    if (!defined($opts{mode}) || $opts{mode} eq 'verify') {
        push @list, "$opts{dirname}_verify_ack";
    }

    foreach $file (@list) {
        if (-e $file) {
            unlink $file or
                croak "populate_cleanup: unlink failed: $file: $!";
        }
    }
    return;
}

1;
