# -*- perl -*-
#
## @summary
##    This perl module has the subroutine for qtree exports cli
##
## @author    tmenaka@netapp.com
##
##

use strict;

use Tharn qw{param $Log};
use NACL::C::Volume;
use NACL::C::Node;
use NACL::C::Client;
use Data::Dumper;
use NATE::BaseException qw(:try);
use NACL::STask::Volume;
use NACL::Exceptions::NoElementsFound qw{:try};
use NACL::STask::Qtree;
use NACL::STask::VserverExportPolicy;
use NACL::C::ExportPolicyRule;

# Parameters
use Params::Common;
use Params::NFS;
use NACL::C::Node;
use NACL::C::VolumeSnapshot;
use NACL::STask::Client::Mount;
use NACL::STask::VolumeSnapshot;

use NACL::C::Volume;
use Params::Validate qw{validate SCALAR BOOLEAN OBJECT};

#--------------------------------------------------------------------------------
# Initialize and validate parameters
#--------------------------------------------------------------------------------

use vars qw(
    $FILER_QTREE
    $EXPORT_ANON
    $VSERVER_GX_NAME
    $EXPORT_POLICY_INDEX
    $NFS_VERSION
    $SUDO_PATH
    $EXPORT_GX_POL
    $RORULE
    $RWRULE
    $VOLUMES
    $CLIENT_MATCH
    $SUPERUSER
    $PROTOCOL
    $POLICY
    $FILER_APISET_SHOULD_INTERFACE
);

BEGIN {
    param('FILER_QTREE', '-default', "subtree1,subtree2,subtree3,subtree4");
    param('EXPORT_ANON', '-default', '65534');
    param("EXPORT_POLICY_INDEX", "-default", time);
    param('SUDO_PATH',           '-default', '/usr/software/bin/sudo');
    param('EXPORT_GX_POL',       '-default', 'test_pol');
    param('VOLUMES',             -default,   "qtreevol1,qtreevol2");
    param('POLICY',       -default, "policy1,policy2,policy3,policy4");
    param('CLIENT_MATCH', -default, "0.0.0.0/0");
    param('RORULE',       -default, 'any');
    param('RWRULE',       -default, 'any');
    param('SUPERUSER',    -default, 'any');
    param('PROTOCOL',     -default, 'tcp');
}

use nfs::CommonParams;
use nfs::CommonUtilities;

use nfs::NfsTCD
    qw{setup cleanup init uninit main $Node_cmd $Mountpoint @UNINIT @CLEANUP $Client_cmd %Setup_hash };

our (
    $Node_cmd, $Client_cmd,    $Mountpoint,
    @Vol_Obj,  @Export_Policy, @Qtree_Name
);

######################################################################
# Description: Automatically called before executing tests
#              Gets/creates the Node, Vserver and Volume objects
##             Initializes Global variables.
######################################################################
sub init() {
    my ($aggr,     @export_policy, $volume, @vol,
        $policy,   $volume_obj,    $rorule, $rwrule,
        $protocol, $super_user,    $client_match,
    );

    nfs::NfsTCD::init();

    $rorule        = $RORULE;
    $rwrule        = $RWRULE;
    $super_user    = $SUPERUSER;
    $client_match  = $CLIENT_MATCH;
    $protocol      = $PROTOCOL;
    @Export_Policy = split(',', $POLICY);
    @Qtree_Name    = split(',', $FILER_QTREE);

    #$Log->comment(qq{Export Policy =} . Dumper(\@Export_Policy));

    $aggr = get_aggregate();
    $Log->comment("Aggregate - $aggr \n");

    @vol = split(',', $VOLUMES);

    #$Log->comment(qq{Volumes =} . Dumper(\@vol));

    my @tmp_uninit_objs;
    foreach $volume (@vol) {
        $volume_obj = volume_create(volume => $volume, aggregate => $aggr);
        push @Vol_Obj, $volume_obj;
        push @tmp_uninit_objs,  $volume_obj;
    }

    foreach $policy (@Export_Policy) {
        create_export_policy($policy);
        push @UNINIT, @tmp_uninit_objs;
        export_policy_rule($policy, $rorule, $rwrule, $super_user,
            $client_match);
    }

    return $TCD::PASS;
} ## end sub init()

sub create_export_policy_rule($$$$;$) {
    my $ruleindex    = shift;
    my $rorule       = shift;
    my $rwrule       = shift;
    my $suser        = shift;
    my $client_match = shift || "0.0.0.0/0";
    my $policy       = 'access' . $ruleindex;

    create_export_policy($policy);
    my $export_policy = export_policy_rule($policy, $rorule, $rwrule, $suser,
        $client_match);

    return $export_policy;
} ## end sub create_export_policy_rule($$$$;$)

#################################################################
#Name: get_aggregate()
#Description: Get aggregate for VSERVER_GX_NAME
################################################################
sub get_aggregate {
    my ($vserver_obj, $aggr);

    # Initialize Vserver Object
    $Log->comment("Vserver Name: $VSERVER_GX_NAME");
    $vserver_obj = NACL::C::Vserver->find(
        command_interface => $Node_cmd,
        filter            => {
            'is-repository' => 'false',
            'vserver'       => $VSERVER_GX_NAME
        },
    );
    $aggr = $vserver_obj->get_one_state_attribute('aggregate');
    $Log->comment("Aggregate - $aggr \n");

    return $aggr;
} ## end sub get_aggregate

##################################################################
# Name          : volume_create()
# Description   : Creates volume
##################################################################
sub volume_create(%) {
    my $issue  = 1;
    my $volobj = 0;
    my %opts   = validate(
        @_,
        {   volume   => {type => SCALAR},
            junction => {type => SCALAR, optional => 1},
            type => {
                type     => SCALAR,
                optional => 1,
                regex    => qr/DP|RW|DC/,
                default  => 'RW'
            },
            aggregate =>
                {type => SCALAR, optional => 1, default => $FILER_AGGR},
            node_cmd => {type => OBJECT, optional => 1, default => $Node_cmd},
            vserver =>
                {type => SCALAR, optional => 1, default => $VSERVER_GX_NAME},
            policy =>
                {type => SCALAR, optional => 1, default => $EXPORT_GX_POL},
        }
    );
    my $node_cmd = $opts{node_cmd};
    delete($opts{node_cmd});
    $Log->comment(
        qq{Create the volume:"$opts{volume}", of type: "$opts{type}"});
    if ($opts{type} eq "RW") {
        $opts{'junction-path'}
            = $opts{junction} ? $opts{junction} : "/$opts{volume}";
        $opts{'unix-permissions'} = "0777";
    }
    if ($opts{junction}) { delete $opts{junction} }

    #Because of burt714844, NACL::C APIs are used
    try {

        #zapi does not support "disable-offline" option
        NACL::C::Volume->offline(
            command_interface => $node_cmd,
            vserver           => $opts{vserver},
            volume            => $opts{volume},
        );

        NACL::C::Volume->delete(
            command_interface       => $node_cmd,
            vserver                 => $opts{vserver},
            volume                  => $opts{volume},
            foreground              => 'true',
            'disable-offline-check' => 'true',

        );
    } catch NACL::C::Exceptions::Volume::DoesNotExist with {
        $Log->comment("Volume, $opts{volume}, does not exist!!!");
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        $Log->comment("Volume, $opts{volume}, does not exist!!!");
    } catch NACL::APISet::Exceptions::CommandFailedException with {
        $Log->comment("Volume, $opts{volume}, does not exist!!!");
    };

    NACL::C::Volume->create(
        command_interface => $node_cmd,
        policy            => $opts{policy},
        aggregate         => $opts{aggregate},
        vserver           => $opts{vserver},
        %opts
    );

    my $volobj = NACL::STask::Volume->find(
        command_interface => $node_cmd,
        filter            => {
            vserver => $opts{vserver},
            volume  => $opts{volume},
        }
    );

    #$Log->comment(qq{Vol Obj =} . Dumper(\$volobj));

    return $volobj;

} ## end sub volume_create(%)

##################################################################
# Name          : create_export_policy()
# Description   : Configure export-policy for the given vserver
##################################################################
sub create_export_policy($;) {
    my ($export_policy_obj, $was_export_policy_created, $export_policy);
    $export_policy = shift;

    $export_policy_obj = NACL::STask::VserverExportPolicy->create(
        'command_interface'  => $Node_cmd,
        'policyname'         => $export_policy,
        'vserver'            => $VSERVER_GX_NAME,
        'nacltask_if_exists' => 'reuse',
        '_was_created'       => \$was_export_policy_created,
    );
    if ($was_export_policy_created) {
        push @UNINIT, $export_policy_obj;
    }

    return $export_policy_obj;
} ## end sub create_export_policy($;)

##################################################################
# Name          : export_policy_rule()
# Description   : Configure export-policy rule for the given vserver
##################################################################
sub export_policy_rule($$$$$;$$) {
    my $export_policy = shift;
    my $rorule        = shift;
    my $rwrule        = shift;
    my $super_user    = shift;
    my $client_match  = shift;
    my $suid          = shift || 'true';
    my $dev           = shift || 'true';
    my $protocol      = "any";

    try {
        NACL::C::VserverExportPolicyRule->delete(
            'command_interface' => $Node_cmd,
            'policyname'        => $export_policy,
            'vserver'           => $VSERVER_GX_NAME,
            'ruleindex'         => $EXPORT_POLICY_INDEX,
        );
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
        my $e = shift;
        $Log->comment(
            "Export policy rule index, $EXPORT_POLICY_INDEX, does not exist on vserver $VSERVER_GX_NAME"
                . $e->text());
    };

    my $export_policy_rule = NACL::C::ExportPolicyRule->create(
        'command_interface' => $Node_cmd,
        'policyname'        => $export_policy,
        'clientmatch'       => $client_match,
        'rorule'            => [$rorule],
        'rwrule'            => [$rwrule],
        'protocol'          => [$protocol],
        'vserver'           => $VSERVER_GX_NAME,
        'anon'              => $EXPORT_ANON,
        'superuser'         => [$super_user],
        'ruleindex'         => $EXPORT_POLICY_INDEX,
        'allow-suid'        => $suid,
        'allow-dev'         => $dev,
    );

    push @UNINIT, $export_policy_rule;

    return $export_policy_rule;
} ## end sub export_policy_rule($$$$$;$$)

##################################################################
# Name          : qtree_create()
# Description   : Creates qtree on given volume
##################################################################
sub qtree_create(%) {
    $Log->comment("Creates qtree on given volume");
    my ($created, $qtree_obj, $cmd, $node_obj, $count);

    my %opts = validate(
        @_,
        {   'qtree'         => {type => SCALAR},
            'volume'        => {type => SCALAR},
            'export-policy' => {type => SCALAR, optional => 1},
        }
    );

    $Log->comment(
        qq(Creating the qtree: "$opts{qtree}" on volume: "$opts{volume}, export-policy: $opts{'export-policy'}")
    );

    $count = 5;
    while ($count > 0) {
        try {
            $qtree_obj = NACL::STask::Qtree->create(
                'command_interface'  => $Node_cmd,
                'vserver'            => $VSERVER_GX_NAME,
                'nacltask_if_exists' => 'purge',
                '_was_created'       => \$created,
                %opts,
            );
            $count = 0;
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            my $e = shift;
            $Log->comment("Exception : " . $e . text());

            sleep(5);
            $qtree_obj = NACL::STask::Qtree->create(
                'command_interface'  => $Node_cmd,
                'vserver'            => $VSERVER_GX_NAME,
                'nacltask_if_exists' => 'purge',
                '_was_created'       => \$created,
                %opts,
            );
            if ($created) {
                $count = 0;
            } else {
                $count--;
            }
        };
    }

    #$Log->comment(qq{Qtree Obj =} . Dumper(\$qtree_obj));

    if ($created) {
        push @CLEANUP, $qtree_obj;
    }

    return $qtree_obj;
} ## end sub qtree_create(%)

##################################################################
# Name          : is_parent_policy()
# Description   : Parse "qtree show -export" to check if qtree
#                 has inherited parent policy
##################################################################
sub is_parent_policy($) {
    my $obj = shift;

    if (is_zapi_enable()) {
        return parse_qtree_show_exports_for_zapi($obj,
            'is_export_policy_inherited');
    }
    return parse_qtree_show_exports($obj, 'Is Export Policy Inherited');
} ## end sub is_parent_policy($)

##################################################################
# Name          : get_qtree_policy()
# Description   : Returns qtree export policy
##################################################################
sub get_qtree_policy ($) {
    my $obj = shift;

    if (is_zapi_enable()) {
        return parse_qtree_show_exports_for_zapi($obj, 'export_policy');
    }
    return parse_qtree_show_exports($obj, 'Policy Name');
} ## end sub get_qtree_policy ($)

######################################################################
# Name          : is_zapi_enable()
# Description   : Returns true if FILER_APISET_SHOULD_INTERFACE variable
#                 is set to ZAPI.
##################################################################
sub is_zapi_enable {
    if ($FILER_APISET_SHOULD_INTERFACE eq 'ZAPI') {
        return 1;
    }
    return 0;
} ## end sub is_zapi_enable

##################################################################
# Name          : parse_qtree_show_exports()
# Description   : Parse qtree show exports command and return hash value
##################################################################
sub parse_qtree_show_exports($;$) {
    my $qtree_obj = shift;
    my $key       = shift;
    my ($cmd,  $show_output, $qtree_show, @lines, $node_obj,
        %hash, @values,      @keys,       $i
    );

    ## Using apiset to test -exports option in qtree show command.
    $cmd
        = "qtree show -exports -vserver "
        . $qtree_obj->vserver()
        . " -volume "
        . $qtree_obj->volume()
        . "  -qtree "
        . $qtree_obj->qtree();

    $node_obj = $Node_cmd->apiset();
    $show_output = $node_obj->execute_raw_command('command' => $cmd);

    @lines = split("\n", $show_output);

    @keys   = split(/\s{2,}/, $lines[1]);
    @values = split(/\s{1,}/, $lines[3]);

    #$Log->comment(qq{Keys =} . Dumper(\@keys));

    for ($i = 0; $i <= $#keys; $i++) {
        if ($keys[$i] eq 'Policy Inherited') {
            $keys[$i] = 'Is Export' . " " . $keys[$i];
        }
        $hash{$keys[$i]} = $values[$i];
    }
    #$Log->comment(qq{show out  =} . Dumper(\%hash));

    if (!$key) {
        return %hash;
    }

    return $hash{$key};
} ## end sub parse_qtree_show_exports($;$)

####################################################################################
# Name          : parse_qtree_show_exports_for_zapi()
# Description   : Returns the value of 'is-export-policy-inherited'/'export-policy'
####################################################################################
sub parse_qtree_show_exports_for_zapi($;$) {
    my $qtree_obj = shift;
    my $key       = shift;
    my $temp;
    my $vol_obj;

    $vol_obj = NACL::CS::VolumeQtree->fetch(
        command_interface => $Node_cmd,
        requested_fields  => ['is-export-policy-inherited', 'export-policy'],
        filter            => {
            'vserver' => $VSERVER_GX_NAME,
            'volume'  => $qtree_obj->volume,
            'qtree'   => $qtree_obj->qtree,
        }
    );

    #$Log->comment(qq{show out  =} . Dumper(\$vol_obj));
    if (!$key) {
        return $vol_obj;
    }
    return $vol_obj->$key();
} ## end sub parse_qtree_show_exports_for_zapi($;$)

##################################################################
# Name          : get_volume_policy()
# Description   : Returns volume policy
##################################################################
sub get_volume_policy($) {
    my $qtree_obj = shift;

    $Log->comment($qtree_obj->{volume});
    my $vol_obj = NACL::CS::Volume->fetch(
        command_interface => $Node_cmd,
        filter            => {
            'vserver' => $VSERVER_GX_NAME,
            'volume'  => $qtree_obj->volume,
        }
    );

    return $vol_obj->policy();
} ## end sub get_volume_policy($)

##################################################################
# Name          : find_node_interface
# Description   : This returns node interface
##################################################################
sub find_node_interface {
    my $interface;

    $interface = NACL::CS::Interface->fetch(
        command_interface => $Node_cmd,
        requested_fields  => ['address'],
        filter            => {
            'role'          => 'data',
            'status-admin'  => 'up',
            'data-protocol' => ['nfs'],
            'vserver'       => $VSERVER_GX_NAME,
        }
    );

    return $interface->address();
} ## end sub find_node_interface

########################################################################
# Name          : mount_volume()
# Description   : This mounts the volume on nfsv4 and returns mount point
#########################################################################
sub mount_volume($) {
    my ($lif_ip, $mount_point, $volume, $status, $cmd, $error);
    $volume      = shift;
    $mount_point = 0;
    $error       = 0;

    $lif_ip = find_node_interface;

    $mount_point = NACL::STask::Client::Mount->create(
        command_interface => $Client_cmd,
        path              => "/$volume",
        server            => $lif_ip,
        mountopts         => "vers=$NFS_VERSION",
    );
    push @CLEANUP, $mount_point;

    return $mount_point;
} ## end sub mount_volume($)

########################################################################
# Name          : create_snapshot()
# Description   : Create Snapshot
#########################################################################
sub create_snapshot($$) {
    my $vol_name      = shift;
    my $snapshot_name = shift;
    my $snapshot_obj;

    $Log->comment(qq{Creating the snapshot:$snapshot_name });

    $snapshot_obj = NACL::STask::VolumeSnapshot->create(
        command_interface  => $Node_cmd,
        vserver            => $VSERVER_GX_NAME,
        nacltask_if_exists => 'purge',
        volume             => $vol_name,
        snapshot           => $snapshot_name,
        foreground         => 'true'
    );

    NACL::C::VolumeSnapshot->find(
        command_interface => $Node_cmd,
        filter            => {
            vserver  => $VSERVER_GX_NAME,
            volume   => $vol_name,
            snapshot => $snapshot_name
        }
    );

    push @CLEANUP, $snapshot_obj;

    return $snapshot_obj;
} ## end sub create_snapshot($$)

########################################################################
# Name          : parse_qtree_show_fields()
# Description   : Parse qtree show fields.
#########################################################################
sub parse_qtree_show_fields($) {
    my $qtree_obj = shift;
    my ($cmd,  $show_output, $qtree_show, @lines, $node_obj,
        %show, @values,      @keys,       $i
    );

    ## Using apiset to test -exports option in qtree show command.
    $cmd
        = "qtree show -vserver "
        . $qtree_obj->vserver()
        . " -volume "
        . $qtree_obj->volume()
        . "  -qtree "
        . $qtree_obj->qtree()
        . " -fields export-policy,is-export-policy-inherited";

    $node_obj = $Node_cmd->apiset();
    $show_output = $node_obj->execute_raw_command('command' => $cmd);

    @lines = split("\n", $show_output);

    @keys   = split('X', $lines[0]);
    @values = split('X', $lines[2]);

    for ($i = 0; $i <= $#keys; $i++) {
        if ($keys[$i] !~ /\w/) {
            next;
        }
        $show{$keys[$i]} = $values[$i];
    }
    #$Log->comment(qq{show out =} . Dumper(\%show));

    return %show;
} ## end sub parse_qtree_show_fields($)

########################################################################
# Name          : parse_qtree_show_instance()
# Description   : Parse qtree show instance.
#########################################################################
sub parse_qtree_show_instance($) {
    my $qtree_obj = shift;
    my ($cmd, $show_output, @lines, $node_obj, %show, $value, $key);

    %show = 0;

    ## Using apiset to test -exports option in qtree show command.
    $cmd
        = "qtree show -vserver "
        . $qtree_obj->vserver()
        . " -volume "
        . $qtree_obj->volume()
        . "  -qtree "
        . $qtree_obj->qtree()
        . " -instance";

    $node_obj = $Node_cmd->apiset();
    $show_output = $node_obj->execute_raw_command('command' => $cmd);

    if ($show_output =~ /There are no entries matching your query./) {
        return 0;
    }

    @lines = split("\n", $show_output);

    foreach (@lines) {
        ($key, $value) = split(':', $_);
        if ($key !~ /\w/) {
            next;
        }
        $key   =~ s/^\s+//;
        $value =~ s/^\s+//;
        $show{$key} = $value;
    }

    #$Log->comment(qq{show out =} . Dumper(\%show));

    return %show;
} ## end sub parse_qtree_show_instance($)

########################################################################
# Name          : create_files()
# Description   : Create files.
#########################################################################
sub create_files ($;$) {
    my $mount_point = shift;
    my $file_name   = shift;
    my ($cmd, $out);

    $Client_cmd->execute(command => "$SUDO_PATH su");

    $cmd = "echo \"hello nfs\"  >> $mount_point->{mount_point}/$file_name";
    $out = $Client_cmd->execute('command' => "$cmd && echo _OK_");

    if ($out !~ /_OK_/) {
        $Log->comment("Command failed: $cmd");
        return $out;
    }

    $cmd = "ls -al $mount_point->{mount_point}/ | grep $file_name";
    $out = $Client_cmd->execute('command' => "$cmd ");

    if ($out !~ /$file_name/) {
        $Log->comment("Command failed: $cmd");
        return $out;
    }

    $Client_cmd->execute(command => "exit");

    return $out;
} ## end sub create_files ($;$)

########################################################################
# Name          : list_files()
# Description   : List files.
#########################################################################
sub list_files ($;$) {
    my $mount_point = shift;
    my $file_name   = shift;
    my $cmd;

    $cmd = "ls -al $mount_point->{mount_point}/ ";
    client_execute($cmd, 'su');

    return 1;
} ## end sub list_files ($;$)

########################################################################
# Name          : open_file()
# Description   : open file on mnt point.
#########################################################################
sub open_file($$) {
    my $mount_point = shift;
    my $file_name   = shift;
    my $result      = 1;
    my ($cmd, $out);

    $cmd = "cat $mount_point->{mount_point}/$file_name";
    client_execute($cmd, 'su');

    return 1;
} ## end sub open_file($$)

########################################################################
# Name          : create_dir()
# Description   : Create dir on mnt point.
########################################################################
sub create_dir ($;$) {
    my $mount_point = shift;
    my $dir         = shift;
    my $cmd;

    if (!$dir) {
        $dir = "exportstest_$$";
    }

    $cmd = "mkdir $mount_point->{mount_point}/$dir";
    client_execute($cmd, 'su');

    $cmd = "ls $mount_point->{mount_point} | grep $dir ";
    client_execute($cmd, 'su');

    return 1;
} ## end sub create_dir ($;$)

########################################################################
# Name          : delete_dir_or_file()
# Description   : Delete file/dir on mnt point.
#########################################################################
sub delete_dir_or_file ($$) {
    my $mount_point = shift;
    my $to_delete   = shift;
    my $result      = 1;
    my ($cmd, $out);

    $Client_cmd->execute(command => "$SUDO_PATH su");

    $cmd = "rm -rf $mount_point->{mount_point}/$to_delete";
    client_execute($cmd);

    $cmd = "ls $mount_point->{mount_point} | grep $to_delete ";
    $out = $Client_cmd->execute('command' => "$cmd ");

    if ($out =~ /$to_delete/) {
        $Log->comment("Command failed: $cmd");
        $result = 0;
    }

    $Client_cmd->execute(command => "exit");

    return $result;

} ## end sub delete_dir_or_file ($$)

########################################################################
# Name          : verify_qtree_show_instance()
# Description   : Verify show instance output.
#########################################################################
sub verify_qtree_show_instance($) {
    my $obj = shift;
    my ($status_show_instance, %show_out, $qtree_obj);

    $status_show_instance = 0;
    %show_out             = parse_qtree_show_instance($obj);
    $qtree_obj            = NACL::CS::VolumeQtree->fetch(
        command_interface => $Node_cmd,
        filter            => {
            'vserver' => $VSERVER_GX_NAME,
            'volume'  => $obj->{volume},
            'qtree'   => $obj->{qtree},
        }
    );

    if ($show_out{'Vserver Name'} eq $qtree_obj->{vserver}) {
        $Log->comment("$show_out{'Vserver Name'} eq $qtree_obj->{vserver}");
        if ($show_out{'Volume Name'} eq $qtree_obj->{volume}) {
            $Log->comment("$show_out{'Volume Name'} eq $qtree_obj->{volume}");
            if ($show_out{'Qtree Name'} eq $qtree_obj->{qtree}) {
                $Log->comment(
                    "$show_out{'Qtree Name'} eq $qtree_obj->{qtree}");
                if ($show_out{'Export Policy'} eq
                    $qtree_obj->{'export_policy'}) {
                    $Log->comment(
                        "$show_out{'Export Policy'} eq $qtree_obj->{'export_policy'}"
                    );
                    if ($show_out{'Is Export Policy Inherited'} eq
                        $qtree_obj->{'is_export_policy_inherited'}) {
                        $Log->comment(
                            "$show_out{'Is Export Policy Inherited'} eq $qtree_obj->{'is_export_policy_inherited'}"
                        );
                        $status_show_instance = 1;
                    }
                }
            }
        }
    }

    return $status_show_instance;
} ## end sub verify_qtree_show_instance($)

########################################################################
# Name          : verify_qtree_show_fields()
# Description   : Verify show instance output.
#########################################################################
sub verify_qtree_show_fields ($) {
    my $obj = shift;
    my ($status_show_fields, %show_out, $qtree_obj);

    $status_show_fields = 0;
    %show_out           = parse_qtree_show_fields($obj);
    $qtree_obj          = NACL::CS::VolumeQtree->fetch(
        command_interface => $Node_cmd,
        filter            => {
            'vserver' => $VSERVER_GX_NAME,
            'volume'  => $obj->{volume},
            'qtree'   => $obj->{qtree},
        }
    );

    if ($show_out{'vserver'} eq $qtree_obj->{vserver}) {
        if ($show_out{'volume'} eq $qtree_obj->{volume}) {
            if ($show_out{'qtree'} eq $qtree_obj->{qtree}) {
                if ($show_out{'export-policy'} eq
                    $qtree_obj->{'export_policy'}) {
                    if ($show_out{'is-export-policy-inherited'} eq
                        $qtree_obj->{'is_export_policy_inherited'}) {
                        $status_show_fields = 1;
                    }
                }
            }
        }
    }

    return $status_show_fields;
} ## end sub verify_qtree_show_fields ($)

########################################################################
# Name          : verify_clnt_access()
# Description   : Verify client access permissions on mount point
#########################################################################
sub verify_clnt_access($) {
    my $mnt_point = shift;

    if (create_files($mnt_point, "nfstest.txt")
        =~ /No such file or directory/) {
        $Log->comment("Create_file is failed !! \n");
        return 0;
    }

    if (!open_file($mnt_point, "nfstest.txt")) {
        $Log->comment("Error in file open operation !! \n");
        return 0;
    }

    if (create_dir($mnt_point, "export") =~ /No such file or directory/) {
        $Log->comment("Error in directory creation !! \n");
        return 0;
    }

    if (!delete_dir_or_file($mnt_point, "nfstest.txt")) {
        $Log->comment("Delete file failed !! \n");
        return 0;
    }

    if (!delete_dir_or_file($mnt_point, "export")) {
        $Log->comment("Dir delte is failed !! \n");
        return 0;
    }

    return 1;
} ## end sub verify_clnt_access($)

sub main() {
    nfs::NfsTCD::main();
}

1;
