#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/RemoteFileIO.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::RemoteFileIO;

## @summary Library for basic remote file I/O
##
## @description
##  Basic fast file I/O on remote clients.
##  This module includes functions to read, write, remove and execute
##  a file on a remote client. As well as functions to create and
##  remove directories and create special device files. If you
##  do not have performance critical requirements then consider
##  using FileManip.pm.
##
## @usage
##  use nfs::RemoteFileIO;
##
## @status	public
##
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @notes
##  Only performance critical routines should be placed in this
##  library. This library will use sayPerl() to send the I/O
##  request to the client which typically takes 100 msec
##  instead of subtesting a PERL script to the client which
##  can take several seconds.

use strict;
use base qw(Exporter);
use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use FileManip qw(write_file);
use Module::Common qw(validate_mixed);
use Params::Validate qw( SCALAR BOOLEAN OBJECT );
use SayWrap qw(sayPerl sayEOF);

our @EXPORT_OK = qw(
  client_read_file
  client_write_file
  client_exec_file
  client_del_file
  client_readdir
  client_mkdir
  client_rmdir
  client_mknod
  start_io
  stop_io
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '0.01';

#---------------------------------------------------------------------
# Package local constants
#---------------------------------------------------------------------
use constant SUDO  => '/usr/software/bin/sudo';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub client_read_file($$;%);
sub client_write_file($$$;%);
sub client_exec_file($$;%);
sub client_del_file($$;%);
sub client_readdir($$;%);
sub client_mkdir($$;%);
sub client_rmdir($$;%);
sub client_mknod($$;%);

#---------------------------------------------------------------------
## @name ::client_read_file
## @description
##  Reads the contents of a file from a client as the user
##  (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_read_file($cc, $file [, %options]);
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $file
##  Required: Filename (absolute path).
##
## @arg root
##  Optional: Boolean, for reading file as the 'root' user
##  (default is 0).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  Contents of file.
#---------------------------------------------------------------------
sub client_read_file($$;%) {
    my ($cc, $file, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            root    => { type => BOOLEAN, default => 0 },
            timeout => { type => SCALAR,  default => 0 },
        }
    );
    my ($cmd, $op);

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # The following is a script to read a file on the client.
    $cmd = qq{
        open(IN, q(<$file)) or die \$OS_ERROR;
        \@cont = <IN>;
        close(IN);
        print q(--read_ok--), \@cont;
    };

    # Run the command as root if the root flag is set
    my $sudo = SUDO if $opts{root};

    # Execute script to read the file on the client
    (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);

    # Remove carriage returns from output
    $op =~ s/\r//g;
    logDbg("client_read_file: $file <$sudo> <<$op>>");

    # Get rid off the error message line number since the
    # perl script gets converted into only one line
    $op =~ s/ at -e line.*/\./;

    # Remove the 'read_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($op =~ s/^--read_ok--//) ) {
        croak "Could not read file '$file': $op";
    }

    return $op;
}

#---------------------------------------------------------------------
## @name ::client_write_file
## @description
##  Creates a file on the client as the user (or root if specified)
##  and checks the size of the file just created.
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  ($uid, $gid, $size, $mode) =
##     client_write_file($cc, $file, $contents [, %options]);
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $file
##  Required: File to be created (absolute path).
##
## @arg $contents
##  Required: Contents of file.
##
## @arg root
##  Optional: Boolean, for creating file as the 'root' user
##  (default is 0).
##
## @arg mode
##  Optional: Permission mode to set for file.
##  Octal number or string, e.g., 0644, '644'
##  (default is 0600).
##
## @arg uid
##  Optional: Change ownership of file to this uid
##  (default is the effective user id of process).
##  This option implies 'root=>1'.
##
## @arg gid
##  Optional: Change group of file to this gid
##  (default is the effective group id of process).
##
## @arg check_ruid
##  Optional: Boolean, check that the file is owned by root
##  (default is 1). This check is only done if the 'root'
##  option is set.
##
## @arg check_mode
##  Optional: Boolean, check that the file's permission mode
##  was set correctly (default is 1). This check is only done
##  if the 'mode' option is passed.
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  List of attributes for file.
#---------------------------------------------------------------------
sub client_write_file($$$;%) {
    my ($cc, $file, $contents, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR },
            { type => SCALAR }
        ],
        {
            root       => { type => BOOLEAN,  default => 0 },
            mode       => { type => SCALAR,  optional => 1, regex => qr{^\d+$}},
            uid        => { type => SCALAR,  optional => 1 },
            gid        => { type => SCALAR,  optional => 1 },
            check_ruid => { type => BOOLEAN,  default => 1 },
            check_mode => { type => BOOLEAN,  default => 1 },
            timeout    => { type => SCALAR,   default => 0 },
        }
    );
    my ($cmd, @attrs, $uid, $gid, $size, $mode, $inode);
    my ($setmode, $ownership, $euid, $egid, $op);

    # Option uid implies root
    $opts{root} = 1 if defined($opts{uid});

    # Doing a write with sayPerl() over a telnet connection is
    # problematic due to an OS dependency on the command line
    # length. If the connection type is telnet then just use
    # the FileManip write_file method.
    if ( $cc->conntype() =~ /telnet/i ) {
        my @options;

        # Map options, all options are mapped directly except
        # for the root option which maps to the user option
        push @options, user => 'root'      if $opts{root};
        push @options, mode => $opts{mode} if defined($opts{mode});
        push @options, uid  => $opts{uid}  if defined($opts{uid});
        push @options, gid  => $opts{gid}  if defined($opts{gid});

        @attrs = write_file( client   => $cc,
                             file     => $file,
                             contents => $contents,
                             fstat    => 1,
                             @options );
        goto ERRCHK;
    }

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Process 'mode' option
    if ( defined($opts{mode}) ) {
        if ( (~$opts{mode} & $opts{mode}) ne '0' ) {
            # Mode was passed as an octal string,
            # convert it to an octal number.
            $opts{mode} = oct($opts{mode});
        }
        # Do not die if chmod fails, it is better to
        # compare the permission mode from file with
        # the expected mode. Also, if option 'check_mode'
        # is set to 0, purposely ignored chmod errors
        $setmode = "chmod($opts{mode}, \$file);";
    }

    # Process 'uid' and 'gid' options
    if ( defined($opts{uid}) or defined($opts{gid}) ) {
        # Get the effective group id of process (first gid)
        # since $) is a space separated list of group ids
        ($op) = ( $) =~ /^(\d+)/ );
        # If not specified used the effective uid/gid
        # of the process or <0,0> if 'root' is defined
        $euid = defined($opts{uid}) ? $opts{uid} : $opts{root} ? 0 : $>;
        $egid = defined($opts{gid}) ? $opts{gid} : $opts{root} ? 0 : $op;
        # Do not die if chown fails, it is better to
        # compare the <uid,gid> from file with the
        # expected values
        $ownership = "chown($euid, $egid, \$file);";
    }

    # Convert all lines of contents into a comma separated
    # words of text lines. Escape all '\', '(' and ')'.
    $contents =~ s/\\/\\\\/g;
    $contents =~ s/\(/\\\(/g;
    $contents =~ s/\)/\\\)/g;
    $contents =~ s/\n/\), q\(/g;
    $contents =~ s/^/q\(/;
    $contents =~ s/$/\)/;

    # The following is a script to create a file on the client.
    # First it sets the creation mask for access by owner ONLY.
    # Creates file, checks ownership of file and returns
    # the uid and size of the file just created.
    $cmd = qq{
        umask 077;
        \$file = q($file);
        open(OUT, qq(>\$file)) or die qq(Could not create file: \$OS_ERROR);
        \$fsize = 0;
        \@lines = ($contents);
        foreach (\@lines) \{
            \$line = \$fsize ? qq(\\n\$_) : \$_;
            print OUT \$line;
            \$fsize += length(\$line);
        \}
        close(OUT);
        $setmode
        $ownership
        \@attrs = stat \$file or die \$OS_ERROR;
        print q(--write_ok--), join(q(--), \@attrs) if \$attrs[7] == \$fsize;
    };

    # Run the command as root if the root flag is set
    my $sudo = SUDO if $opts{root};

    # Execute script to create the file on the client
    (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);
    logDbg("client_write_file: $file <$sudo> <<$op>>");

    # Get rid off the error message line number since the
    # perl script gets converted into only one line
    $op =~ s/ at -e line.*/\./;

    # Remove the 'write_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($op =~ s/^--write_ok--//) ) {
        croak "Could not write file: '$file': $op";
    }

    # Get return values
    @attrs = split(/--/, $op);
ERRCHK:
    ($inode, $mode, $uid, $gid, $size) = @attrs[1,2,4,5,7];

    # Check for errors
    if ( $op =~ /Could not create/ ) {
        # Unable to create file
        croak $op;
    } elsif ( $opts{root} and $opts{check_ruid}
      and ! defined($opts{uid}) and $uid != 0 ) {
        # Couldn't create file as root,
        # returned uid is not '0' (root)
        croak "Could not create '$file' as root";
    } elsif ( ( defined($opts{uid}) and $opts{uid} != $uid )
      or ( defined($opts{gid}) and $opts{gid} != $gid ) ) {
        # Couldn't change ownership of file
        croak "Could not change ownership of '$file'";
    } elsif ( defined($opts{mode}) and $opts{check_mode}
      and $opts{mode} != ($mode & 07777) ) {
        # Couldn't change permission mode of file
        croak "Could not change permission mode of '$file'";
    } elsif ( defined($uid) and defined($gid)
      and defined($size) and defined($mode) ) {
        # Everything is fine
        return @attrs;
    } else {
        # It should never get to this point, but just in
        # case all the other checks failed
        croak "Could not write '$file': $op";
    }
}

#---------------------------------------------------------------------
## @name ::client_exec_file
## @description
##  Executes a file on the client as the user (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_exec_file($cc, $file [, %options]);
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $file
##  Required: File to be executed.
##
## @arg root
##  Optional: Boolean, for executing file as the 'root' user
##  (default is 0).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  Returns the output of the file executed.
#---------------------------------------------------------------------
sub client_exec_file($$;%) {
    my ($cc, $file, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR },
        ],
        {
            root    => { type => BOOLEAN, default => 0 },
            timeout => { type => SCALAR,  default => 0 },
        }
    );
    my (@say_opts, $op);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Execute the file as root if specified
    my $sudo = SUDO if $opts{root};

    # Execute file on the client
    $op = sayEOF($cc, "$sudo $file && echo --exec_ok--", @say_opts);

    # Remove the 'exec_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($op =~ s/--exec_ok--//) ) {
        croak "Could not execute file '$file': $op";
    }

    return $op;
}

#---------------------------------------------------------------------
## @name ::client_del_file
## @description
##  Removes a file on the client as the user (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_del_file($cc, $file [, %options]);
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $file
##  Required: File to be removed.
##
## @arg root
##  Optional: Boolean, for removing file as the 'root' user
##  (default is 0).
##
## @arg force
##  Optional: Deprecated, no longer in use.
##
## @arg recursive
##  Optional: Boolean, recursively removes directories and subdirectories.
##  This option allows the use of wild cards (e.g., '*', '?') on the file name,
##  (default is 1).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub client_del_file($$;%) {
    my ($cc, $file, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            root      => { type => BOOLEAN, default => 0 },
            force     => { type => BOOLEAN, default => 1 },
            recursive => { type => BOOLEAN, default => 1 },
            timeout   => { type => SCALAR,  default => 0 },
        }
    );
    my (@say_opts, $cmd, $op);

    # Do not attempt to remove root directory
    if ( $file =~ /^.:\\?\*?$/ or $file =~ m{^/\*?$} ) {
        croak "Not removing root directory";
    }

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    if ( $opts{recursive} ) {
        # Perl script to remove a directory tree.
        # It also accepts wild cards ('*', '?', ...)
        # to get a list of top level files and
        # directories and removes them recursively.
        $cmd = qq{
            use File::Find;
            sub rmelem() \{
                return if /^\\.\$/;
                ((-d) ? rmdir : unlink) or die \$OS_ERROR;
            \}
            \@list = glob(q($file));
            if ( \@list < 2 and not -e \$list[0] ) \{
                die q(No such file or directory);
            \}
            \$op = finddepth(\\&rmelem, \@list);
            (rmdir or die \$OS_ERROR) foreach grep \{-d\} \@list;
            print q(--del_ok--);
        };
    } else {
        # Perl script to remove a single file
        $cmd = qq{
            unlink q($file) or die \$OS_ERROR;
            print q(--del_ok--);
        };
    }

    # Remove the file as root if specified
    my $sudo = SUDO if $opts{root};

    # Get rid off multiple '/'s
    $file =~ s{/+}{/};
    # Delete file if it is not in the root directory in unix,
    # i.e., contains more than one '/'.
    if ( $file =~ /^.:/ or ($file =~ tr{/}{/}) > 1 ) {
        # Remove the file on the client
        (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);
        logDbg("client_del_file: $file <$sudo> <<$op>>");

        # Check for errors
        if ( $op !~ /--del_ok--/ ) {
            croak "Could not remove file $file: $op";
        }
    } else {
        croak "Not removing file located in root directory";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::client_readdir
## @description
##  Reads the contents of a directory from a client as the user
##  (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_readdir($cc, $dir [, %options]);
##
## @arg $cc
##  Required: Connection handle for client.
##
## @arg $dir
##  Required: Directory (absolute path).
##
## @arg root
##  Optional: Boolean, for reading directory as the 'root' user
##  (default is 0).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  List of entries in the directory.
#---------------------------------------------------------------------
sub client_readdir($$;%) {
    my ($cc, $dir, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            root    => { type => BOOLEAN, default => 0 },
            timeout => { type => SCALAR,  default => 0 },
        }
    );
    my ($cmd, $op);

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # The following is a script to read a directory on the client.
    $cmd = qq{
        opendir(DIR, q($dir)) or die \$OS_ERROR;
        \@allfiles = readdir DIR;
        closedir DIR;
        print q(--readdir_ok--);
        foreach \$file (\@allfiles) \{
            print qq(\$file\\n);
        \}
    };

    # Run the command as root if the root flag is set
    my $sudo = SUDO if $opts{root};

    # Execute script to read the file on the client
    (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);

    # Get rid off the error message line number since the
    # perl script gets converted into only one line
    $op =~ s/ at -e line.*/\./;

    # Remove the 'readdir_ok' flag from the output
    # and if it does not exist croak the error
    if ( ! ($op =~ s/^--readdir_ok--//) ) {
        croak "Could not read directory '$dir': $op";
    }

    return split(/\r?\n/, $op);
}

#---------------------------------------------------------------------
## @name ::client_mkdir
## @description
##  Creates a directory structure on the client as the user
##  (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_mkdir($cc, $dir [, %options ]);
##
## @arg $cc
##  Required: Client's connection object.
##
## @arg $dir
##  Required: Directory to be created.
##
## @arg mode
##  Optional: Permission mode to set for directory.
##  Octal number or string, e.g., 0755, '755'
##  (default is the current setting of umask, usually 0755).
##
## @arg root
##  Optional: Boolean, for creating directory as the 'root' user.
##  If this option is specified, the 'dirbase' option must also
##  be specified (default is 0).
##
## @arg dirbase
##  Optional: Directory base where $dir is located. If this option
##  is specified it will create one directory at a time starting at
##  'dirbase', otherwise the '-p' option of mkdir will be used to
##  create the directory structure at once.
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub client_mkdir($$;%) {
    my ($cc, $dir, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            root    => { type => BOOLEAN,  default => 0,
                      depends => ['dirbase'] },
            mode    => { type => SCALAR,  optional => 1, regex => qr{^\d+$} },
            dirbase => { type => SCALAR,  optional => 1 },
            timeout => { type => SCALAR,   default => 0 },
        }
    );
    my ($cmd, $mode, $setup_dir, $op);
    my $dirbase = $opts{dirbase};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Create directory as root if specified
    my $sudo = SUDO if $opts{root};

    # Process 'mode' option
    if ( defined($opts{mode}) ) {
        # Convert mode into an octal string
        if ( (~$opts{mode} & $opts{mode}) eq '0' ) {
            # Mode was passed as a octal number,
            # convert it to a string.
            $opts{mode} = sprintf '%lo', $opts{mode};
        }
        $mode = "-m $opts{mode}";
    }

    # Make dir relative to dirbase if necessary
    if ( $dirbase ) {
        $dirbase =~ s/\/+$//;           # Get rid off trailing '/'s
        $dir =~ s/^$dirbase\///;        # Relative directory
        $dir =~ s/^\/+//;               # Get rid off trailing '/'s
    }

    # Break up $dir since csh limits command line to 1024 chars
    if ( length($dir) > 700 ) {
        # Length too long to fit in a single command
        my $subdir;
        my $offset = 0;
        while ( $subdir = substr($dir, $offset, 700) ) {
            $setup_dir .= qq{\$dir .= q($subdir);\n};
            $offset += length($subdir);
        }
    } else {
        # Length is fine, all in one command
        $setup_dir = qq{\$dir = q($dir);};
    }

    # The following is a script to create a directory on the client.
    # If dirbase is set, the directory structure will be created
    # recursively one directory at a time, otherwise a '-p' option
    # is used to create all directories at once. This is done since
    # rsh as root with 'mkdir -p' gives an error. Therefore, when
    # creating a directory as root, the dirbase option must also
    # be specified.
    $cmd = qq{
        my \$dir;
        $setup_dir
        \$mode = q($mode);
        \$dir_base = q($dirbase);
        if ( \$dir_base ) \{
            \@dirs = split(/\\//, \$dir);
            \$cd = q(cd);
        \} else \{
            \@dirs = (\$dir);
        \}
        \$curr = \$dir_base;
        foreach \$dd (\@dirs) \{
            \$cd = qq(cd \$curr;) if \$cd;
            \$op = qx(\$cd mkdir -p \$mode \$dd);
            die if \$?;
            die \$op if \$op;
            \$curr .= qq(/\$dd);
        \}
        if ( \$mode ) \{
            \$mode =~ s/-m //;
            \$dir = qq(\$dir_base/\$dir) if \$dir_base;
            \$dmod = 07777 & (stat(\$dir))[2];
            die q(Failed to change permission mode) if (\$dmod ne oct(\$mode));
        \}
        print q(--mkdir_ok--);
    };

    # Create directory structure on the client
    (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);

    # Get rid off the error message line number since the
    # perl script gets converted into only one line
    $op =~ s/ at -e line.*/\./;

    # Check for errors
    if ( $op !~ /--mkdir_ok--/ ) {
        $dir = "$dirbase/$dir" if $dirbase;
        croak "Could not create directory '$dir': $op";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::client_rmdir
## @description
##  Removes an empty directory on the client as the user
##  (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_rmdir($cc, $dir [, %options ]);
##
## @arg $cc
##  Required: Client's connection object.
##
## @arg $dir
##  Required: Directory to be removed.
##
## @arg root
##  Optional: Boolean, for removing directory as the 'root' user.
##  If this option is specified, the 'dirbase' option must also
##  be specified (default is 0).
##
## @arg dirbase
##  Optional: Directory base where $dir is located. If this option
##  is specified it will remove one directory at a time starting at
##  'dirbase', otherwise only one directory will be removed.
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub client_rmdir($$;%) {
    my ($cc, $dir, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            root    => { type => BOOLEAN,  default => 0,
                      depends => ['dirbase'] },
            dirbase => { type => SCALAR,  optional => 1 },
            timeout => { type => SCALAR,   default => 0 },
        }
    );
    my ($cmd, $setup_dir, $op);
    my $dirbase = $opts{dirbase};

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Remove directory as root if specified
    my $sudo = SUDO if $opts{root};

    # Make dir relative to dirbase if necessary
    if ( $dirbase ) {
        $dirbase =~ s/\/+$//;           # Get rid off trailing '/'s
        $dir =~ s/^$dirbase\///;        # Relative directory
        $dir =~ s/^\/+//;               # Get rid off trailing '/'s
    }

    # Break up $dir since csh limits command line to 1024 chars
    if ( length($dir) > 700 ) {
        # Length too long to fit in a single command
        my $subdir;
        my $offset = 0;
        while ( $subdir = substr($dir, $offset, 700) ) {
            $setup_dir .= qq{\$dir .= q($subdir);\n};
            $offset += length($subdir);
        }
    } else {
        # Length is fine, all in one command
        $setup_dir = qq{\$dir = q($dir);};
    }

    # The following is a script to remove a directory on the client.
    # If dirbase is set, the directory structure will be removed
    # recursively one directory at a time, otherwise only one
    # directory will be removed.
    $cmd = qq{
        my \$dir;
        $setup_dir
        \$dir_base = q($dirbase);
        if ( \$dir_base ) \{
            \@dirs = split(/\\//, \$dir);
            \$cd = q(cd);
        \} else \{
            \@dirs = (\$dir);
        \}
        \$curr = qq(\$dir_base/\$dir);
        foreach \$dd (reverse \@dirs) \{
            \$curr =~ s/\\/\$dd\$//;
            \$cd = qq(cd \$curr;) if \$cd;
            \$op = qx(\$cd rmdir \$dd);
            die if \$?;
            die \$op if \$op;
        \}
        print q(--rmdir_ok--);
    };

    # Remove directory structure on the client
    (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);

    # Check for errors
    if ( $op != /--rmdir_ok--/ ) {
        $dir = "$dirbase/$dir" if $dirbase;
        croak "Could not remove directory '$dir': $op";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::client_mknod
## @description
##  Creates a special device file on the client as the user
##  (or root if specified).
##  Optional arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = client_mknod($cc, $file [, %options ]);
##
## @arg $cc
##  Required: Client's connection object.
##
## @arg $file
##  Required: Special device file to be created.
##
## @arg type
##  Optional: Special device type: 'b' for block, 'c' for character,
##  'p' for fifo(default) or 's' for named socket.
##
## @arg major
##  Optional: The major device number.
##
## @arg minor
##  Optional: The minor device number.
##
## @arg root
##  Optional: Boolean, for creating device file as the 'root' user
##  (default is 0).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub client_mknod($$;%) {
    my ($cc, $file, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            type    => { type => SCALAR,   default => 'p',
                        regex => qr{^[bcps]$} },
            major   => { type => SCALAR,  optional => 1,
                      depends => ['type', 'minor'] },
            minor   => { type => SCALAR,  optional => 1,
                      depends => ['type', 'major'] },
            root    => { type => BOOLEAN,  default => 0 },
            timeout => { type => SCALAR,   default => 0 },
        }
    );
    my (@say_opts, $mknod, $cmd, $op);

    # For fifo and socket the major and minor should be undef
    # Only check for major since either both are defined or not
    if ( $opts{type} =~ /[ps]/ and $opts{major} ) {
        croak "Options major and minor not valid with type <$opts{type}>";
    }

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Execute the file as root if specified
    my $sudo = SUDO if $opts{root};

    if ( $opts{type} eq 's' ) {
        # Create unix domain (named) socket
        $cmd = qq{
            use Socket;
            socket(SERVER, PF_UNIX, SOCK_STREAM, 0);
            bind(SERVER, sockaddr_un(q($file)))
              or die qq(Can not create named socket: \$OS_ERROR);
            print q(--mknod_ok--);
        };

        # Add options to sayPerl
        push @say_opts, '-swallow';

        # Execute script to create a named socket on the client
        (undef, undef, $op) = sayPerl($cc, $cmd, $sudo, @say_opts);
    } else {
        # Set proper path for mknod command
        if ( $cc->hostp()->hosttype() eq 'unix-linux' ) {
            $mknod = '/bin/mknod';
        } else {
            $mknod = '/usr/sbin/mknod';
        }

        # Create special device file on the client
        $op = sayEOF($cc, "$sudo $mknod $file $opts{type} $opts{major} " .
                          "$opts{minor} && echo --mknod_ok--", @say_opts);
    }

    # Check for errors
    if ( $op !~ /--mknod_ok--/ ) {
        croak "Could not create special file '$file': $op";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::start_io
## @description
##  Open a file from a client, continuously keep doing the required IO to that file until
##  instructed to release it using the 'stop_io' function.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = start_io ( %options );
##
## @arg client
##  Required: Hostspec, Hostrec or Connectrec of client.
##
## @arg file
##  Required: File to open.
##
## @arg user
##  Optional: Open the file as this user.
##
## @arg gid
##  Optional: Open the file with this group id.
#---------------------------------------------------------------------
sub start_io(%;) {
    my (%opts) = validate_mixed (
        @_,
		[],
        {
            client	=> { type => SCALAR|OBJECT },
            file	=> { type => SCALAR },
            user	=> { type => SCALAR, optional => 1, default => $< },
			gid		=> { type => SCALAR, optional => 1 },
			type	=> { type => SCALAR, optional => 1, regex => qr/read|write/, default => 'read' },
        }
    );

    my ($file, $err, $op, %extra_options);
	my $file   = $opts{file};

	my $gid_cmd = "";
    if ( $opts{gid} ) {
        # Change the credentials (gid) to only list
        # the gid specified and the change the uid
        # to the current uid since subtest will be
        # run as root
        $gid_cmd = <<'EOT';
		{
            $) = q($opts{gid} $opts{gid});
            $> = $>;
        };
EOT
        # Option gid implies user => 'root' since the
        # credentials of the process must be changed
        # and only root can do it
        $opts{user} = 'root';
    }
    $extra_options{'-username'} = $opts{user} if defined($opts{user});

	my $io_cmd;
	if ($opts{type} eq "write") {
		$file = ">" . $file;
		$io_cmd = q{print FD "Write to deleg file, by delegation tests\n";};
	} elsif ($opts{type} eq "read") {
		$file = "<" . $file;
		$io_cmd = q{read(FD, $dummy, 10, 0);};
	} else {
		croak (qq{type can be only either of read or write. Please contact the script owner about this.});
	}

	#Actual command that is going to run in the subtest.
	my $open_and_run_cmd = <<'EOT';
	{
		my ($file, $gid_cmd, $io_cmd) = @_;
		my $dummy;
		my $socket =  Subtest::ipc();
		eval "$gid_cmd";
		open (FD, $file) || die "Couldnot open file:$!";
		close FD;
		open (FD, $file) || die "Couldnot reopen file:$!";
		$socket->write_message(qq(--open_ok--$$--$>));
		while (1) {
			sleep 1;
			eval "$io_cmd";
			if ($socket->can_read()) {
				my $msg = $socket->read_message();
				chomp($msg);
				if ($msg eq "stop") {
					exit 0;
				} else {
					# As of now nothing else.
					die (qq{Stupid IPC, some problem, got a message:$msg});
				}
			}
		}
	}
EOT

	logStep(qq{Going to do subtest with: $open_and_run_cmd});
	logStep(qq{Args for subtest: file:$file, gid:$gid_cmd, io_cmd:$io_cmd});

	my $sub_id = subtest ( $open_and_run_cmd, '-bg',
						'-ipc', '-runid', '01_continuous_write',
						'-client', (ref($opts{client}) eq 'Connectrec' ? $opts{client}->hostp() : $opts{client}),
						%extra_options,
						'-onexit', sub {},
						'--', $file, $gid_cmd, $io_cmd);

    $op = eval { $sub_id->ipc()->read_message() };
	my ($subid, $io_uid);
    if ( $@ || ! (($subid, $io_uid) = ($op =~ /--open_ok--(\d+)--(\d+)/)) ) {
        # There was an error so release, i.e., terminate subtest.
        eval { stop_io(sub_id => $sub_id) };

        $err = join(' ', $sub_id->results());
	}

	croak $err if ($err);

	return $sub_id;
}

sub stop_io(%;) {
    my (%opts) = validate_mixed (
        @_,
		[],
        {
            sub_id  => { type => OBJECT },
            timeout => { type => SCALAR, default => $main::CMDTIMEOUT },
        }
    );

	logStep(qq{Stopping the continuous io});
    if ( $opts{sub_id}->is_alive() ) {
        # Send message to subtest to terminate
        $opts{sub_id}->ipc()->write_message(qq{stop\n});
        #$opts{sub_id}->kill();
        $opts{sub_id}->wait_finish(timeout=>$opts{timeout});
    }

    return 1;
}


1;

