# this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/Rpc.pm#28 $
#
## @summary Rpc - Rpc class for basic ONC RPC functionality
## @author      dl-ssqa, dl-nb-nfs-qa, dl-nfsqa
## @status      public
## @notes       Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Rpc - Rpc class for basic ONC RPC functionality

=head1 SYNOPSIS

  use nfs::Rpc;

  # RPC client constructor and functions
  $rpc_obj = nfs::Rpc->clnt_create( hostname => $host,
                                    program  => $program,
                                    version  => 4,
                                    proto    => 'tcp' );

  # Call the null procedure
  $out = $rpc_obj->clnt_call( proc => 0 );

  # Call procedure N with arguments and output processing function
  $out = $rpc_obj->clnt_call( proc => N,
                              arg  => $xdr_args,
                              out  => \&xdr_out );

  # Communication functions
  $rpc_obj->message_send( xdr => $xdr_obj );
  $xdr_obj = $rpc_obj->message_receive();

  # Reconnect TCP connection
  $rpc_obj->reconnect();

  # Re-transmit last call
  $out = $rpc_obj->call_retransmit();

  # Process a reply
  $out = $rpc_obj->reply_process( xdr => $reply_xdr_obj,
                                  xid => $xid,
                                  out => \&xdr_out );
  # RPC authentication
  $out = $rpc_obj->auth_sys_set();
  $out = $rpc_obj->auth_none_set();

  # RPC functions to connect to portmapper called as class methods:
  $port = nfs::Rpc::pmap_getport( program => $program,
                                  version => 4,
                                  proto   => 'udp',
                                  host    => 'client_host',
                                  pmproto => 'tcp' );

  $out = nfs::Rpc::pmap_set( program => $program,
                             version => 4,
                             proto   => 'udp',
                             port    => 2000,
                             pmproto => 'tcp' );

  $out = nfs::Rpc::pmap_unset( program => $program,
                               version => 4,
                               proto   => 'udp',
                               port    => 2000,
                               pmproto => 'tcp' );

  $plist = nfs::Rpc::pmap_dump( host => $hostname );

  # RPC functions to connect to portmapper called as instance methods:
  $port = $rpc_obj->pmap_getport( program => $program,
                                  version => 4,
                                  proto   => 'udp' );

  $out = $rpc_obj->pmap_set( program => $program,
                             version => 4,
                             proto   => 'udp',
                             port    => 2000 );

  $out = $rpc_obj->pmap_unset( program => $program,
                               version => 4,
                               proto   => 'udp',
                               port    => 2000 );

  $plist = $rpc_obj->pmap_dump( host => $hostname );

  # RPC server constructor and functions
  $rpc_obj = nfs::Rpc->svc_create( dispatch => \%funcs,
                                   version  => 4 );
  $rpc_obj->svc_run();
  $rpc_obj->svc_getreqset();
  $rpc_obj->svc_register_transient( version => 4,
                                    proto   => 'udp',
                                    port    => 2000 );

  # Object description
  $xid       = $rpc_obj->xid();             # Xid to use on next call
  $retry     = $rpc_obj->retry();           # Retry count for object
  $qsize     = $rpc_obj->qsize();           # Size of saved call queue
  $timeout   = $rpc_obj->timeout();         # Timeout for object
  $program   = $rpc_obj->program();         # Program number
  $version   = $rpc_obj->version();         # Version number
  $proto     = $rpc_obj->proto();           # Protocol
  $port      = $rpc_obj->port();            # Port number of connection
  $localaddr = $rpc_obj->localaddr();       # Local ip address
  $localport = $rpc_obj->localport();       # Local port number
  $last_call = $rpc_obj->last_call();       # Last call info
  $callq     = $rpc_obj->callq();           # Call queue info
  $flag      = $rpc_obj->wait_for_reply();  # Flag for waiting for reply
                                            #   after each call

  # Change object behavior, all return the value before being set
  $xid_prev     = $rpc_obj->xid($xid_new);
  $retry_prev   = $rpc_obj->retry($retry_new);
  $qsize_prev   = $rpc_obj->qsize($qsize_new);
  $timeout_prev = $rpc_obj->timeout($timeout_new);
  $flag_prev    = $rpc_obj->wait_for_reply($flag_new);

=head1 DESCRIPTION

Provides an object for basic ONC RPC (Remote Procedure Call)
operations. Basic functionality is included for creating an
RPC connection object to a server and calling procedures
on such a server, as well as creating an RPC server and
accept calls from an RPC client.

The best way to test this library is by running the tests in nfs/ops.

=head1 STATUS

   Shared

=head1 AUTHOR

  dl-ssqa
  $Id: //depot/prod/test/main/lib/nfs/Rpc.pm#28 $
  Last revised $Date: 2012/07/11 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Rpc;

use strict;
use warnings;

use nfs::Xdr;
use Carp qw( croak );
use IO::Socket::INET;
use IO::Socket::INET6;
use POSIX qw(uname F_SETFL O_NONBLOCK EINTR EAGAIN);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF CODEREF OBJECT UNDEF);
use Tharn qw(host $Log param);
use Net::IP qw(ip_is_ipv4 ip_is_ipv6);
use Data::Dumper;
use GSSAPI;

# Kerberos flavors
use constant RPC_GSS_SVC_NONE         => 1;
use constant RPC_GSS_SVC_INTEGRITY    => 2;
use constant RPC_GSS_SVC_PRIVACY      => 3;

# Authentication types.
use constant AUTH_NONE                => 0;
use constant AUTH_SYS                 => 1;
use constant RPCSEC_GSS               => 6;

# Kerberos constants.
use constant RPCSEC_GSS_DATA          => 0;
use constant RPCSEC_GSS_INIT          => 1;
use constant RPCSEC_GSS_CONTINUE_INIT => 2;
use constant RPCSEC_GSS_DESTROY       => 3;
use constant GSS_S_COMPLETE           => 0;
use constant GSS_C_QOP_DEFAULT        => 0;

BEGIN {

    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
}

#------------------------------------------------------------------------
# Subroutine prototypes
#------------------------------------------------------------------------
sub name2ip(%);
sub addr_type(%);
sub loopback(%);
sub localaddr(%);
sub pmversion(%);
sub clnt_create($%);
sub svc_create($%);
sub svc_run($;);
sub svc_getreqset($;%);
sub clnt_call($%);
sub call_retransmit($;$$);
sub reply_process($%);
sub message_send($%);
sub message_receive($;);
sub reconnect($;%);
sub xid($;$);
sub retry($;$);
sub qsize($;$);
sub timeout($;$);
sub wait_for_reply($;$);
sub program;
sub version;
sub proto;
sub port;
sub localport;
sub last_call;
sub callq;
sub auth_sys_set($%);
sub auth_none_set($;);
sub auth_gss_set($%);
sub pmap_getport(%);
sub pmap_set(%);
sub pmap_unset(%);
sub pmap_dump(%);
sub svc_register_transient($;%);
sub init_cred($$$;);
sub _xdr_pmap(%);
sub _xdr_pmap6(%);
sub _xdr_uint(%);
sub _xdr_uint6(%);
sub _xdr_pmap_dump(%);
sub _xdr_auth_sys(%);
sub _xdr_auth_gss($%);
sub _rpc_sec_flavor($);
sub _sock_bind($;);
sub _sysread($$;);
sub _syswrite($$;);
sub _set_obj_var($$$;);

#------------------------------------------------------------------------
# Package local variables
#------------------------------------------------------------------------
my %ipproto      = (tcp => 6, udp => 17);
my $LAST_FRG     = (1 << 31);
my $TCP_FRG_SIZE = 32768;
my $UDP_MAX_SIZE = 65535;
my %auth_error = (1 => 'AUTH_BADCRED',
                  2 => 'AUTH_REJECTEDCRED',
                  3 => 'AUTH_BADVERF',
                  4 => 'AUTH_REJECTEDVERF',
                  5 => 'AUTH_TOOWEAK',
                  6 => 'AUTH_INVALIDRESP',
                  7 => 'AUTH_FAILED',);
my %auth_error_msg = (1 => 'Bad credential',
                      2 => 'Client must begin new session',
                      3 => 'Bad verifier',
                      4 => 'Verifier expired or replayed',
                      5 => 'Rejected for security reasons',
                      6 => 'Bogus response verifier',
                      7 => 'Unknown error',);
my %accepted_error = (1 => 'PROG_UNAVAIL',
                      2 => 'PROG_MISMATCH',
                      3 => 'PROC_UNAVAIL',
                      4 => 'GARBAGE_ARGS',
                      5 => 'SYSTEM_ERR',);
my %accepted_error_msg = (1 => 'Program number unavailable',
                          2 => 'Program mismatch',
                          3 => 'Procedure number unavailable',
                          4 => 'Arguments cannot be decoded',
                          5 => 'Unknown error',);
my @retry_errors = ('Timeout waiting',
                    'Operation now in progress',
                    'protocol failure in circuit setup',
                    'all ports in use');

#------------------------------------------------------------------------

=head2 name2ip

    my $ip = nfs::Rpc->name2ip(hostname => $hostname,
                               [, addr_type => 'inet4|inet6' ]);

This is a temporary (hopefully) utility function to resolve to desired IP family
from hostname. It is used by NfsClient.pm mostly.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : Hostname or ip address to be resolved to an address

=item addr_type

(Optional) : The address family type. inet4 or inet6.

=back

=back

=cut

sub name2ip(%) {
	$Log->enter();
    my $self;
    $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_,
               {hostname => {type => SCALAR},
                addr_type => {type => SCALAR, optional => 1},});

	if (!$opts{addr_type}) {
		if ($self) {
			if (UNIVERSAL::isa($self, "nfs::Rpc") && $self->{addr_type}) {
				$opts{addr_type} = $self->{addr_type};
			} else {
				$opts{addr_type} = $self->addr_type();
			}
		} else {
			$opts{addr_type} = addr_type();
		}
	}

    my $ipaddr;
    if (ip_is_ipv4($opts{hostname}) || ip_is_ipv6($opts{hostname})) {

        # IP address was sent, send the address as it is.
        $ipaddr = $opts{hostname};
    } else {
        my $host = host($opts{hostname});
        if ($host->can('net_best_address')) {
            $ipaddr = $host->net_best_address(
                                     include => {family => "$opts{addr_type}"});
        } else {
            $ipaddr = $host->default_ip();
        }
    }

    croak "addr_type is $opts{addr_type} but $opts{hostname} was given"
      if ($opts{addr_type} ne "inet" . Net::IP::ip_get_version($ipaddr));

	$Log->exit("$ipaddr");
    return $ipaddr;
}

#------------------------------------------------------------------------

=head2 addr_type

    my $addr_type = nfs::Rpc->addr_type();
    or
    my $addr_type = $rpc_obj->addr_type();

Returns addr_type of the class object, if set. Otherwise returns the desired addr_type.

=cut

sub addr_type(%) {
	$Log->enter();
    my $self = shift;
	if (ref($self) && UNIVERSAL::isa($self, "nfs::Rpc")) {
    	if (defined($self->{addr_type})) {
        	return $self->{addr_type};
		} else {
			croak (qq{Something is weird. addr_type is not available inside myself ???});
		}
    } else {

        # Get it from NATE env, since we got the package name itself.
        my $addr_type = param('ADDR_TYPE', "-default", "inet4");
        return $addr_type;
    }
}

=head2 auth_type

    my $auth_type = nfs::Rpc->auth_type();
    or
    my $auth_type = $rpc_obj->auth_type();

Returns auth_type of the class object, if set. Otherwise returns the desired auth_type.

=cut

sub auth_type() {
	$Log->enter();
    my $self = shift;
    if (!$self or !ref($self)) {
        # Get it from NATE env, since we got the package name itself or nothing.
        param('AUTH', "-default", "sys");
        return $::AUTH;
	} elsif (UNIVERSAL::isa($self, "nfs::Rpc")) {
    	if (defined($self->{auth_type})) {
        	return $self->{auth_type};
		} else {
			croak (qq{Something is weird. auth_type is not available inside myself ???});
		}
    }
}

#------------------------------------------------------------------------

=head2 loopback
    
    my $loopback = nfs::Rpc->loopback();

Returns loopback address for given address family.

=cut

sub loopback(%) {
    my $self;
    $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_, {addr_type => {type => SCALAR, optional => 1},});

	if ($opts{addr_type}) {
        return ($opts{addr_type} eq "inet6") ? "::1" : "127.0.0.1";
	}

	# If no argument was given, then use the self stuff.
    if (ref($self) && UNIVERSAL::isa($self, "nfs::Rpc") && defined($self->{addr_type})) {
		return loopback(addr_type=>$self->{addr_type});
    } else {
		# Return the global variable.
		return loopback(addr_type=>addr_type());
	}
}

#------------------------------------------------------------------------

=head2 localaddr

    my $localaddr = nfs::Rpc->localaddr();
    or
    my $localaddr = $rpc_obj->localaddr();

Returns localaddr address for given address family, if object did not have localaddr set.

=cut

sub localaddr(%) {
    my $self;
    $self = shift if ((@_ % 2) == 1);
	$Log->enter();
    my (%opts) =
      validate(@_, {addr_type => {type => SCALAR, optional => 1},});

	if ($opts{addr_type}) {
        return ($opts{addr_type} eq "inet6") ? "::0" : "0.0.0.0";
	}

	# If no argument was given, then use the self stuff.
    if (ref($self) && UNIVERSAL::isa($self, "nfs::Rpc") && defined($self->{localaddr})) {
        return $self->{localaddr};
    } elsif (UNIVERSAL::isa($self, "nfs::Rpc") && defined($self->{addr_type})) {
		return localaddr(addr_type=>$self->{addr_type});
    } else {
		# Return the global variable.
		return localaddr(addr_type=>addr_type());
	}
}

#------------------------------------------------------------------------

=head2 pmversion

    my $pmversion = nfs::Rpc->pmversion();
    or
    my $pm_version = $rpc_obj->pmversion();

Returns pmversion of the object, if exists. Otherwise returns the desired versin of pmversion.

=cut

sub pmversion(%) {
    my $self;
    $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_, {addr_type => {type => SCALAR, optional => 1},});

	if ($opts{addr_type}) {
		return ($opts{addr_type} eq "inet6") ? 4 : 2;
	}

	# If no argument was given, use the self stuff
    if (ref($self) && UNIVERSAL::isa($self, "nfs::Rpc")) {
		if (defined($self->{pmversion})) {
			return $self->{pmversion};
		} elsif (defined($self->{addr_type})) {
			return pmversion(addr_type=>$self->{addr_type});
    	}
    } else {
		# No object, and no addr_type, so just return the global (local to this Rpc.pm) value.
		return pmversion(addr_type=>addr_type());
	}
}

#------------------------------------------------------------------------

=head2 clnt_create

  $rpc_obj = nfs::Rpc->clnt_create(hostname => $host,
                                   program => $program,
                                   version => $version
                                   [, addr_type => 'inet4'|'inet6' ]
                                   [, proto => 'udp'|'tcp' ]
                                   [, port => $port ]
                                   [, resvport => 0|1 ]
                                   [, xid => $xid ]
                                   [, retry => $retry ]
                                   [, reconnect => 0|1 ]
                                   [, qsize => $qsize ]
                                   [, timeout => $timeout ]
                                   [, localaddr => $localaddr ]
                                   [, sec   => 'none'|'sys'|'krb'|'krb5i'|'krb5p']);

Rpc class constructor which initializes communications for the
given program, protocol and version with an RPC server.
Creates an INET/INET6 socket connection for the given protocol and
port number. If the port number is not provided, a connection
to the portmapper is established to get the port number for
specified program/version/protocol on the given server.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostname

(Required) : Hostname or ip address of server to connect.

=item program

(Required) : Program number to connect.

=item version

(Required) : Version number of program.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item proto

(Optional) : Protocol of connection: 'udp', 'tcp' (default is 'udp').

=item port

(Optional) : Port number of program/version/proto (default is 0,
which attempts to connect to the portmapper to get the port number).

=item resvport

(Optional) : Bind to a reserved port number on the local machine.
A reserved port number is less than or equal to 1024. If this
option is true, an attempt to bind to a reserved port number
between 600 and 1023 is made. On most unix machines, root
is the only user able to bind to a reserved port (default is 0).

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period (default is 0).

=item reconnect

(Optional) : Reconnect if TCP connection is dropped. This only works for
TCP connections and when the retry parameter is non-zero (default is 0)

=item qsize

(Optional) : Size of saved call queue.

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item localaddr

(Optional) : Hostname or ip address of local machine.

=item sec

(Optional) : The type of security to be used. If not specified, it uses "sys" authentication by default.
The valid values are none, sys, krb5, krb5i, krb5p. Please note as of writing this code, krb5p and krb5i
are not yet supported in c-Mode.

Internal details:
"sec" is used for the CLI based options : none, krb5, krb5i, krb5p, and sys.

"auth" and "auth_type" are used for the internal representations: AUTH_SYS(1), AUTH_NONE(0), and RPCSEC_GSS(6).
$self->{auth} is again one of the above. It just represents the type. auth_type is used only in a few functions.

$self->{auth_sys}, $self->{auth_gss} represent the actual data structures for the respective auth types. There is no such structure for auth_none.

"sec" is used in NfsClient and the clnt_call. Everywhere else in Rpc.pm it is auth.

= item tcp_frg_size

(Optional) : TCP Fragmentation size. Tunable using Rpc Object (default is 32768).

=back

=item Returns

RPC Object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub clnt_create($%) {
    my $class = shift;
    my (%opts) = validate(
        @_,
        {hostname => {type => SCALAR},
         addr_type => {type => SCALAR, optional => 1},
         program   => {type => SCALAR, regex   => qr/^\d+$/},
         version   => {type => SCALAR, regex   => qr/^\d+$/},
         port      => {
                  type    => SCALAR,
                  regex   => qr/^\d+$/,
                  default => 0},
         resvport => {type    => SCALAR,
                      regex   => qr/^[01]$/,
                      default => 0},
         proto => {type    => SCALAR,
                   regex   => qr/^(udp|tcp|tcp6|udp6)$/,
                   default => 'udp'},
         xid => {type    => SCALAR,
                 regex   => qr/^\d+$/,
                 default => int(rand(1000000000))},
         retry => {type    => SCALAR,
                   regex   => qr/^\d+$/,
                   default => 0},
         reconnect => {type    => SCALAR,
                       regex   => qr/^[01]$/,
                       default => 0},
         qsize => {type    => SCALAR,
                   regex   => qr/^\d+$/,
                   default => 5},
         timeout => {type    => SCALAR,
                     regex   => qr/^\d+$/,
                     default => 30},
         localaddr => {type => SCALAR, optional => 1,},
         auth_type => {type => SCALAR, optional => 1, default => 1},
         cb        => {
                type     => SCALAR,
                optional => 1,},
         sec        => { type => SCALAR,
                         default => 'sys',
                         regex   => qr/^(sys|krb5|krb5i|krb5p|none)$/,},
         rpc_vers => {type    => SCALAR,
                      regex   => qr/^\d+$/,
                      default => 2},
    	 tcp_frg_size => {type => SCALAR,
			              default => $TCP_FRG_SIZE}
        });

	if (!$opts{addr_type}) {
		$opts{addr_type} = $class->addr_type();
	}
	if (!$opts{localaddr}) {
		$opts{localaddr} = localaddr(addr_type=>$opts{addr_type});
	}
    $opts{dstaddr} = $class->name2ip(hostname  => $opts{hostname},
                                     addr_type => $opts{addr_type});
    $Log->enter(Dumper(\%opts));

    # Encode static RPC header information
    # and save it as part of the blessed hash
    my $xdr = nfs::Xdr->new();
    $xdr->uint(encode => 0);                  # Call
    $xdr->uint(encode => $opts{rpc_vers});    # RPC version (2)
    $xdr->uint(encode => $opts{program});     # Program number
    $xdr->uint(encode => $opts{version});     # Version number
    my $call_hdr = $xdr->{message};

    # Bless the hash to an RPC class
    my $self = bless {%opts,
                      auth     => 1,
                      call_hdr => $call_hdr,}, $class;

    # Create AUTH_SYS structure for current user
    $self->auth_sys_set();

    if (!$opts{port}) {

        # Port is not defined, call portmapper to get the port number
        $opts{port} =
          $self->pmap_getport(program => $opts{program},
                              version => $opts{version},
                              proto   => $opts{proto});
        croak "Could not get port number for program <$opts{program}>"
                  unless $opts{port};
    }

    # Replace port number on object
    $self->{port} = $opts{port};

    # Set default value of flag to wait for the reply
    $self->{wait_for_reply} = 1;

    # Create connection and save socket object
    $self->{sock} = $self->_sock_bind();

    $self->auth_correct_set(sec => $opts{sec},hostname=>$opts{hostname});

    $Log->exit(Dumper($self));

    # Return the Rpc object
    return $self;
}

#---------------------------------------------------------------------

=head2 svc_create

  $rpc_obj = nfs::Rpc->svc_create( dispatch     => \%funcs,
                                   version      => $version
                                   [, addr_type => 'inet4'|'inet6' ]
                                   [, program   => $program ]
                                   [, proto     => 'udp'|'tcp' ]
                                   [, xid       => $xid ]
                                   [, retry     => $retry ]
                                   [, qsize     => $qsize ]
                                   [, register  => 0|1 ]
                                   [, versions  => \@versions ]
                                   [, timeout   => $timeout ]
                                   [, localaddr => $localaddr ] );

Rpc class constructor which initializes communications for the
given program, protocol and version as an RPC server.
Creates an INET socket connection for the given protocol.

The following named arguments are accepted:

=over

=item Arguments

=over

=item dispatch

(Required) : Reference to a hash of procedure functions. The key being
the procedure number and the value is a reference to the procedure
function.

=item version

(Required) : Version number of program.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item program

(Optional) : Program number of RPC server. If this option is not
given, a program number in the transient range is used.

=item proto

(Optional) : Protocol of connection: 'udp', 'tcp' (default is 'udp').

=item xid

(Optional) : Initial transaction id (default is a random-generated number).

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period (default is 0).

=item qsize

(Optional) : Size of call queue.

=item register

(Optional) : Register program with portmapper (default is 1).

=item versions

(Optional) : Reference to a list of versions to be supported in
addition to the 'version' option.

=item timeout

(Optional) : Timeout value for connection (default is 30).

=item localaddr

(Optional) : Hostname or ip address of local machine.

=back

=item Returns

RPC Object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub svc_create($%) {
    my $class = shift;
    my (%opts) =
      validate(@_,
               {dispatch => {type => HASHREF},
                version   => {type => SCALAR, regex   => qr/^\d+$/},
                addr_type => {type => SCALAR, default => $class->addr_type()},
                proto     => {
                          type    => SCALAR,
                          regex   => qr/^(udp|tcp)$/,
                          default => 'udp'},
                program => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                xid => {type    => SCALAR,
                        regex   => qr/^\d+$/,
                        default => int(rand(1000000000))},
                retry => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                qsize => {type    => SCALAR,
                          regex   => qr/^\d+$/,
                          default => 0},
                timeout => {type    => SCALAR,
                            regex   => qr/^\d+$/,
                            default => 30},
                register    => {type => SCALAR,   default  => 1},
                versions    => {type => ARRAYREF, optional => 1},
                localaddr   => {type => SCALAR,   optional => 1},
                port        => {type => SCALAR,   optional => 1},
                create_sock => {type     => SCALAR,
                                default  => 1,
                                optional => 1},});
    $Log->enter(Dumper(\%opts));

    my %sock_opts_h;
    my $retries = 20;

    # Check correct options
    if (!$opts{register} && !defined($opts{program})) {
        croak "A program option is required if it is not to be registered"
          . " with the portmapper";
    }

    # Add all versions to be supported
    $opts{versions} = {map { $_ => 1 } (@{$opts{versions}}, $opts{version})};

    # Bless the hash to an RPC class
    my $self = bless {%opts, server => 1}, $class;

    # BAD, BAD, BAD....
    $self->{cb_auth} = 1;

    if ($opts{create_sock}) {
        # Use given local address or self's localaddr
        $sock_opts_h{LocalAddr} = $opts{localaddr} ? $opts{localaddr} : $self->localaddr(addr_type=>$opts{addr_type});
        if ($self->{proto} eq 'udp') {
            $sock_opts_h{LocalPort} = 45000 + int(rand(10000));
        } else {
            $sock_opts_h{Listen} = 5;
        }

		{
	        $sock_opts_h{Proto}    = $self->{proto};
	        $sock_opts_h{Timeout}  = $self->{timeout};
	        $sock_opts_h{Blocking} = 1;

	        # Connect socket for listening to incoming connections
	        if ($self->addr_type() eq "inet6") {
	        	$self->{lsock} = IO::Socket::INET6->new(%sock_opts_h);
			} else {
		            $self->{lsock} = IO::Socket::INET->new(%sock_opts_h);
		    }
		    if (   $retries--
		            && $self->{proto} eq 'udp'
		            && !$self->{lsock}
		            && $! =~ /Address already in use/) {
		            $sock_opts_h{LocalPort}++;
		            redo;
		    }
		}
        croak "Socket could not be created: $!" unless $self->{lsock};

        # Add port number on object
        $self->{port} = $self->{lsock}->sockport();

        # Save local ip address
        $self->{localaddr} = $self->{lsock}->sockhost();

        # make sure localaddr matches with addr_type, otherwise fail - Could be a configuration
        # issue or this library issue
        if ($self->addr_type() ne "inet"
            . Net::IP::ip_get_version($self->{localaddr})) {
            croak "$self->{localaddr} is not desired addr family : "
              . $self->addr_type();
        }
    } else {
        $self->{lsock} = $self->{rpc} ? $self->{rpc}->{sock} : undef;
    }

    # Return the Rpc object
    return $self unless $self->{register};

	$self->{dispatch} = $opts{dispatch};

    if (defined($self->{program})) {
        # Register the service for the given program number
        my $out = $self->pmap_set(program => $self->{program},
                                  version => $self->{version},
                                  proto   => $self->{proto},
                                  auth_type=> 0,
                                  port    => $self->{port});
        $self->{program} = undef unless $out;

    } else {

        # Get a program number from the transient range
        # and register the service
        $self->{program} =
          $self->svc_register_transient(version => $self->{version},
                                        proto   => $self->{proto},
                                        port    => $self->{port});
    }
    croak "Unable to register service with the portmapper"
      unless $self->{program};

    # BAD, BAD, BAD....
    $self->{cb_auth} = 1;

	$Log->exit(qq{returning self:} . Dumper($self));
    # Return the Rpc object
    return $self;
}

#---------------------------------------------------------------------

=head2 svc_run

  $rpc_obj->svc_run();

This function process incoming requests using the dispatcher.
It never returns control to the calling function unless
a fatal error is detected.

=over

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub svc_run($;) {
    my $self = shift;

    while (1) {
        $self->svc_getreqset();
    }

    # Should never get to this point
    return;
}

#---------------------------------------------------------------------

=head2 svc_getreqset

  $rpc_obj->svc_getreqset([ timeout => $time_in_mseconds ]);

This function is the RPC remote procedure dispatcher.
Waits for a request, dispatches request to the appropriate
procedure then sends the reply to the client.

=over

=item Arguments

=over

=item timeout

(Optional) : Timeout value in milliseconds to check if any
request is pending. A value of 0 returns immediately if no
request is currently pending. If argument is not specified
it waits until a request is ready with no time limit.

=back

=item Returns

True if request is received and processed correctly.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub svc_getreqset($;%) {
    my $self = shift;
	$Log->enter(qq{self: } . Dumper($self));
    my (%opts) =
      validate(@_,
               {timeout => {type => SCALAR, regex => qr/^\d+$/, optional => 1},
                accept  => {
                           type     => SCALAR,
                           regex    => qr/^\d+$/,
                           default  => 1,
                           optional => 1},});
    my ($rout, $out, %rpc_hdr);
    my $rin             = '';
    my $reply_status    = 0;
    my $accepted_status = 0;
    my $rejected_status = 0;

    if (defined($opts{timeout})) {

        # Check if there is a reply waiting
        vec($rin, fileno($self->{lsock}), 1) = 1;
        if (!select($rout = $rin, undef, undef, $opts{timeout} / 1000)) {
            return 0;
        }
    }

    if ($self->{proto} eq 'udp') {
        $self->{sock} = $self->{lsock};
    } elsif ($opts{accept}) {

		if ($self->{lsock}) {
        # Wait for a connection
        	$self->{sock} = $self->{lsock}->accept()
				or croak "svc_getreqset accept: $! and $@";
		} else {
			croak (qq{Some function is asking to do accept, but there is no lsocket});
		}
    }
    # Make socket non-blocking
    fcntl($self->{sock}, F_SETFL, O_NONBLOCK);

    # Get incoming call message
    my $xdr = $self->message_receive();

    # Decode call message
    my $xid  = $xdr->uint();
    my $call = $xdr->uint();

    # This is not a call, just ignore it
    return if $call;

    my $rpcver  = $xdr->uint();
    my $program = $xdr->uint();
    my $version = $xdr->uint();
    my $proc    = $xdr->uint();
    my $credential = {flavor => $xdr->uint(),
                      auth   => $xdr->opaque(variable => 400)};
    my $verifier = {flavor => $xdr->uint(),
                    auth   => $xdr->opaque(variable => 400)};

    # RPC call header
    %rpc_hdr = (proto      => $self->{proto},
                xid        => $xid,
                rpcver     => $rpcver,
                program    => $program,
                version    => $version,
                proc       => $proc,
                credential => $credential,
                verifier   => $verifier,);

    # Encode reply header
    my $reply_xdr = nfs::Xdr->new();
    $reply_xdr->uint(encode => $xid);
    $reply_xdr->uint(encode => 1);

    # Check correct version of RPC
    if ($rpcver != 2) {
        $reply_status    = 1;
        $rejected_status = 0;    # RPC MISMATCH
    } elsif ($credential->{flavor} > 1 || $verifier->{flavor} > 0) {
        $reply_status    = 1;
        $rejected_status = 1;    # AUTH ERROR
    }

    # Encode the reply status
    $reply_xdr->uint(encode => $reply_status);

    if ($reply_status) {

        # MSG_DENIED
        $reply_xdr->uint(encode => $rejected_status);
        if ($rejected_status) {
            if ($credential->{flavor} > 1) {
                # AUTH ERROR: AUTH_TOOWEAK
                $reply_xdr->uint(encode => 5);

            } else {
                # AUTH ERROR: AUTH_BADVERF
                $reply_xdr->uint(encode => 3);

            }
        } else {
            # RPC MISMATCH: (low, high)
            $reply_xdr->uint(encode => 2);
            $reply_xdr->uint(encode => 2);
        }
    } else {
        # Check proper program, procedure and version
        if ($program != $self->{program}) {
            $accepted_status = 1;    # PROG_UNAVAIL
        } elsif (not $self->{versions}{$version}) {
            $accepted_status = 2;    # PROG_MISMATCH
        } elsif (not $self->{dispatch}{$proc}) {
            $accepted_status = 3;    # PROC_UNAVAIL
        }

        # Encode the verifier
        $reply_xdr->uint(encode => 0);
        $reply_xdr->uint(encode => 0);

        if (not $accepted_status) {
            # So far everything is fine
            # Call appropriate procedure
            $out = eval {
                &{$self->{dispatch}{$proc}}(xdr     => $xdr,
                                            rpc_hdr => \%rpc_hdr);
            };

            # Checking for XDR decoding errors and extra bytes in call
            if (   $@ =~ /^XDR: Not enough bytes/
                || $@ =~ /^XDR Decode:/
                || $xdr->remain()) {

                # XDR problem decoding arguments
                $accepted_status = 4;    # GARBAGE ARGS
            } elsif ($@) {

                # Unknown error, reply with system error and save it
                $accepted_status = 5;       # SYSTEM ERR
                $self->{system_err} = $@;
            }
        }

        # Encode the accepted status
        $reply_xdr->uint(encode => $accepted_status);

        if (!$accepted_status) {

            # Append procedure results to reply message
            my $len = defined($out) ? length($out) : 0;
            $reply_xdr->opaque(encode => $out, fixed => $len) if $len;
        } elsif ($accepted_status == 2) {

            # PROG_MISMATCH
            $reply_xdr->uint(encode => $self->{version});
            $reply_xdr->uint(encode => $self->{version});
        }
    }

    # Send the reply to the client
    $self->message_send(xdr => $reply_xdr);

	#This check is invalid for NFSv4.1, since the callback socket is same as the client's nfs socket.
	#If we close this socket, we end up with a broken socket for further communication.
    #if (($opts{accept}) and ($self->{proto} eq 'tcp')) {
    #    $self->{sock}->close();
    #}

    return 1;
}

#---------------------------------------------------------------------

=head2 clnt_call

  $out = $rpc_obj->clnt_call(proc => $proc
                             [, arg => $xdr_args ]
                             [, out => \&xdr_out ]);

Calls a procedure on the server with the given arguments.
=item auth_type

(Optional) : The type of the authentication to be used. Valid values are
0(AUTH_NONE), 1(AUTH_SYS) and 6(RPCSEC_GSS).


The following named arguments are accepted:

=over

=item Arguments

=over

=item proc

(Required) : Procedure number to call.

=item arg

(Optional) : Arguments to remote procedure in XDR format
(the XDR encoded string of all arguments to procedure).

=item out

(Optional) : Reference to a function to decode the XDR format output
of remote procedure.

=back

=item Returns

The output of remote procedure, once it is processed by
the XDR decoding function which is given on the 'out' argument.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub clnt_call($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {proc => {type => SCALAR,  regex    => qr/^\d+$/},
                arg  => {type => SCALAR,  optional => 1},
                auth_type => {type => SCALAR,  optional => 1, default => $self->auth_type()},
                out  => {type => CODEREF, optional => 1},});
    $Log->enter(qq{self:} . Dumper($self));
    my ($out, $retry, $retry_error, $err);

    # Create an XDR object to encode arguments
    my $xdr = nfs::Xdr->new();

    # Save current transaction id
    my $xid = $self->{xid};

    # Add the transaction id
    $xdr->uint(encode => $self->{xid}++);

    # Append static call header which has already
    # been XDR encoded, so just copy the buffer
    # using a fixed length opaque
    $xdr->opaque(encode => $self->{call_hdr},
                 fixed  => length($self->{call_hdr}));

    # Add the procedure number
    $xdr->uint(encode => $opts{proc});

    my $used_auth_type = (defined $opts{auth_type}) ? $opts{auth_type} : $self->auth_type();
    # Add the authentication structure
    if ($used_auth_type == AUTH_SYS) {
            # AUTH_SYS authentication credential
            my $auth_sys = _xdr_auth_sys(%{$self->{auth_sys}});
            $xdr->uint(encode => 1);
            $xdr->opaque(encode => $auth_sys, variable => 400);
    } elsif ($used_auth_type == RPCSEC_GSS) {
            # RPCSEC GSS authentication credential
            my $cred_gss = $self->_xdr_auth_gss(%{$self->{auth_gss}});
            $xdr->uint(encode=>6);
            my $len = length $cred_gss;
            $xdr->opaque(encode=>$cred_gss, variable=>$len);
    } else {
            # AUTH_NONE authentication credential
            $xdr->uint(encode => 0);
            $xdr->uint(encode => 0);
    }
    if ($used_auth_type == RPCSEC_GSS) {
        # Add RPCSEC GSS verifier
        if($self->{auth_gss}{gss_proc} != (RPCSEC_GSS_INIT or RPCSEC_GSS_CONTINUE_INIT))
        {
            my $verf;
            my $status;
            $status = GSSAPI::Context::get_mic($self->{gss_context},GSS_C_QOP_DEFAULT,$xdr->{message},$verf);
            if ($status->major != GSS_S_COMPLETE)
            {
                croak " $status";
            }
            my $verf_len = length $verf;
            $xdr->uint(encode=>RPCSEC_GSS);
            $xdr->opaque(encode=>$verf,variable => $verf_len);
        } else {
            # Add the null verifier
            $xdr->uint(encode => 0);
            $xdr->uint(encode => 0);
        }

    } else {
        # Add NULL verifier
        $xdr->uint(encode=>0);
        $xdr->uint(encode=>0);
    }

    # Add the NFS data according to the security
    if (($used_auth_type == RPCSEC_GSS) and
        (($self->{auth_gss}{gss_proc}) != (RPCSEC_GSS_INIT or RPCSEC_GSS_CONTINUE_INIT))) {
        if ($self->{auth_gss}{service} == RPC_GSS_SVC_INTEGRITY) {
            # data = opaque[gss_seq_num+data] + opaque[checksum]
            my $xdrnew = nfs::Xdr->new();
            my $checksum;
            $xdrnew->uint(encode=>$self->{gss_seq_num});
            if ( !defined($opts{arg}) )
            {
                $opts{arg} = '';
            }

            $xdrnew->{message}.= $opts{arg};
            GSSAPI::Context::get_mic($self->{gss_context},0,$xdrnew->{message},$checksum);
            my $data = $xdrnew->{message};
            $xdrnew->{message}= '';
            $xdrnew->opaque(encode=>$data, variable => length $data);
            $xdrnew->opaque(encode=>$checksum, variable=>length $checksum);
            $opts{arg} = $xdrnew->{message};
        } elsif ($self->{auth_gss}{service} == RPC_GSS_SVC_PRIVACY)  {
            # data = opaque[wrap([gss_seq_num+data])]
            my $xdrnew = nfs::Xdr->new();
            $xdrnew->uint(encode=>$self->{gss_seq_num});
            if ( !defined($opts{arg}) )
            {
                $opts{arg} = '';
            }
            $xdrnew->{message}.= $opts{arg};
            my $data = '';
            my $status = GSSAPI::Context::wrap($self->{gss_context},1,GSS_C_QOP_DEFAULT,$xdrnew->{message}, my $outstate, $data);
            $xdrnew->{message} = '';
            $xdrnew->opaque(encode=>$data, variable =>length $data);
            $opts{arg} = $xdrnew->{message};
        }
    }

    # Add the already encoded arguments
    if (defined($opts{arg})) {
        $xdr->{message} .= $opts{arg};
    }

    # Save XDR object of call and decoding function on the call queue
    if ($used_auth_type == RPCSEC_GSS) {
        $self->{last_call} = {
            xid => $xid,
            xdr => $xdr,
            out => $opts{out},
            auth_gss => $self->{auth_gss},
            seq_num => $self->{gss_seq_num},
            gss_context => $self->{gss_context} };

    } else {
        $self->{last_call} = {xid => $xid, xdr => $xdr, out => $opts{out}};
    }
    if ($self->{qsize} > 0) {
        if ($self->{callq} && @{$self->{callq}} >= $self->{qsize}) {

            # Queue is full so remove first element to make room for this call
            shift @{$self->{callq}};
        }
        push @{$self->{callq}}, $self->{last_call};
    }

    # Initialize number of retries
    $retry = $self->{retry};

    do {
        # Send call to server and wait for reply
        $out = eval { $self->call_retransmit($self->{last_call}, $used_auth_type) };
        $err = $@;
		chomp($err) if $err;    # Remove newline char if any
        $retry_error = 0;
        if ($err && $retry) {

            # Check if call needs to be retransmitted
            foreach (@retry_errors) {
                if ($err =~ /$_/) {
                    $retry_error = 1;
                    $Log->trace("retry error \"$err\" occured for version "
                                . $self->{version});
                    last;
                }
            }

            # Check if reconnect is required
            if ($self->{reconnect}
                && (($err =~ /Connection\s+reset/i || $err =~ /Broken\s+pipe/i)
                    || $self->{version} == 4)
              ) {
                $retry++;
                do {
                    $Log->trace(
                            "reconnect required due to \"$err\", reconnecting");
                    eval { $self->reconnect(timeout => $self->{timeout}) };
                } while ($@ && $retry--);
                croak "reconnect failed: $err ($@)" if $@;
                $retry_error = 1;
            }
        }
    } while ($retry_error && $retry--);
    croak "Rpc::clnt_call failed: $err" if $err;

    # Return results
    $Log->exit();
    return $out;
}

#---------------------------------------------------------------------

=head2 call_retransmit

  $out = $rpc_obj->call_retransmit([ \%call_msg ]);

Re-transmits the previous call message to the server.

=over

=item Arguments

=over

=item <hash reference>

(Optional) : Re-transmit this call message instead.

=back

=item Returns

The output of remote procedure, once it is processed by
the XDR decoding function used by the previous call.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub call_retransmit($;$$) {
    my ($self, $call, $auth_type) = @_;
    my ($xdr, $out);
    my $last_call = $call ? $call : $self->{last_call};
    $Log->enter(qq{self:} . Dumper($self));

    # Retransmit RPC call
    $self->message_send(xdr => $last_call->{xdr});
    return unless $self->{wait_for_reply};

    # Discard replies with a different transaction id than the call
    do {

        # Wait for reply
        $xdr = $self->message_receive();
		my $new_xdr = $xdr;
        my $xid     = $new_xdr->uint();
        my $call_o  = $new_xdr->uint();

        if (!$call && $call_o == 0 && $self->{cb} == 1) {
            test_for_cb($self, xdr => $xdr, xid => $xid);

        } elsif ( $self->{qsize} > 0 && ! $call  && $call_o == 0) { 
            # save current reply
            $self->{callq}[-1]{reply} = $xdr;
            $self->{unknown_call}[0] = $xdr;
        }

        # Process reply
        my $used_auth_type = $auth_type ? $auth_type : $self->auth_type();
        if ($used_auth_type == RPCSEC_GSS)
        {
            $out = eval { $self->reply_process(xdr => $xdr,
                xid => $last_call->{xid},
                out => $last_call->{out},
                auth_gss => $last_call->{auth_gss},
                seq_num => $last_call->{seq_num} ? $last_call->{seq_num} : 0,
                gss_context => $last_call->{gss_context}) };

        } else {
            $out = eval {
                $self->reply_process(xdr => $xdr,
                                xid => $last_call->{xid},
                                out => $last_call->{out});
            };
        }
    } while ($@ =~ /RPC:\s+Transaction\s+numbers\s+are\s+not\s+the\s+same/i);

    # Propagate error from reply_process
    croak "Rpc::call_retransmit failed: $@" if $@;

    $Log->exit();

    # Return results
    return $out;
}

#---------------------------------------------------------------------

=head2 reply_process

  $out = $rpc_obj->reply_process(xdr => $xdr_obj,
                                 xid => $xid
                                 [, out => \&out ]);

Decode the specified reply with the given XDR decoding function.
=item auth_gss

(Optional):	The actual rpc_gss structure for kerberos decoding functions. This strucuture is a must if we are expecting
a kerberized reply. If the reply has a flavour of RPCSEC_GSS, and this arugment is null, we croak.

Used to get the results from a saved reply.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : XDR object of reply to be processed.

=item xid

(Required) : Transaction id of reply to be processed.

=item out

(Optional) : Reference to a function to decode the XDR format output
    my (%init_res, $xdrnew, $status);
of remote procedure.

=back

=item Returns

The output of remote procedure, once it is processed by
the XDR decoding function which is given on the 'out' argument.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub reply_process($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {xdr => {type => OBJECT},
                xid => {type => SCALAR,          regex    => qr/^\d+$/},
                out => {type => CODEREF | UNDEF, optional => 1},
                auth_gss => { optional => 1},
                seq_num => {type => SCALAR, optional => 1, default=>0},
                gss_context => {optional => 1}},
            );
    $Log->enter();
    my (%init_res, $xdrnew, $status);
    my $out;
    my $flag = 0; #to send the correct xdr object to nfs decoding function
    my $xdr = $opts{xdr};

    # Reset offset on reply
    $xdr->{offset} = 0;

    # Get header of reply
    my $rxid    = $xdr->uint();
	my $reply   = $xdr->uint();
	# Check the reply has the same transaction number as the call
    if ($rxid != $opts{xid}) {
        croak "RPC: Transaction numbers are not the same";
    }

    my $rstatus = $xdr->uint();
    # Make sure this is an RPC reply
    if ($reply != 1) {
        croak "RPC: Expecting a reply";
    }
    if ($rstatus) {

        # Message denied
        my $rej_status = $xdr->uint();

        if ($rej_status) {
            my $err = $xdr->uint();
            croak "RPC ERROR $auth_error{$err}: $auth_error_msg{$err}";
        } else {
            my $low  = $xdr->uint();
            my $high = $xdr->uint();

            #croak "RPC MISMATCH (low = $low, high = $high)";
        }
    } else {


        # Message accepted:
        # Get verifier
        my $flavor = $xdr->uint();
		# Verifier is needed in case of rpcsec_gss. This is used further in the code,
		# to verify MIC. Also the verifier should be null in the middle of a reply.
		# This is also checked in the same function.
        my $verf = $xdr->opaque();
        my $errmsg  = '';
        # Accepted status
        my $astatus = $xdr->uint();
        if ($astatus == 2) {
	        # Program mismatch, get programs supported
            my $low  = $xdr->uint();
            my $high = $xdr->uint();
            $errmsg = "(low = $low, high = $high)";
        }

        if ($astatus) {
            # Accept error
			croak "NFS $accepted_error{$astatus}: "
              . "$accepted_error_msg{$astatus} $errmsg";
        }

        if ($flavor == RPCSEC_GSS) {

            croak (qq{The reply contains gss flavour, but this function argument doesnot have auth_gss structure})
                unless $opts{auth_gss};
            if ( $opts{auth_gss}{gss_proc} == (RPCSEC_GSS_INIT or RPCSEC_GSS_CONTINUE_INIT )) {
                $init_res{handle} = $xdr->opaque();
                $init_res{gss_major} = $xdr->uint();
                $init_res{gss_minor} = $xdr->uint();
                $init_res{seq_window} = $xdr->uint();
                $init_res{gss_token} = $xdr->opaque();
                my $junk = \%init_res;
                if ($verf eq '' )
                {
                    if($init_res{gss_major} == GSS_S_COMPLETE)
                    {
                        croak qq{Expected seq window(non-null verifier)at end of the encrypted reply.\nThere seems to be a problem in filer response.\n Please get a packet trace for analysis.};
                    }
                } else {
                    if($init_res{gss_major} != GSS_S_COMPLETE)
                    {
                        croak qq{expected null verifier in the middle of an encrypted reply.\nThere seems to be a problem in filer response.\nPlease get a packet trace for analysis. the obtained verifier:"} . unpack("h[400]", $verf) . qq{"};
                    }
                }
                return \%init_res;
            } else {

			    $xdrnew = nfs::Xdr->new();
                $xdrnew->uint(encode=>$opts{seq_num});
                $status = GSSAPI::Context::verify_mic($opts{gss_context},$xdrnew->{message},$verf , GSS_C_QOP_DEFAULT);

                if ($status->major != GSS_S_COMPLETE)
                {
                    croak qq{Verify mic error $status. There seems to be a problem in verifying something called "MIC" in the reply from the filer.};
                }
            }
        }

	 if ($astatus == 0 ) {

    	  if($flavor == RPCSEC_GSS)
              {
    		  # Decrypt the NFS Data
              if ( $opts{auth_gss}{service} == RPC_GSS_SVC_INTEGRITY)
              {
    		    # data = opaque[seq_num+data] + opaque[checksum]
                        my $data = $xdr->opaque();
    		    my $checksum = $xdr->opaque();
    		    $status = GSSAPI::Context::verify_mic($opts{gss_context}, $data, $checksum , 0);
    		    if ($status->major != GSS_S_COMPLETE)
                        {
                          croak " verify mic failed $status ";
                        }
                $xdrnew = nfs::Xdr->new(message=>$data);
    	    my $seq_num = $xdrnew->uint();
                if ( $seq_num != $opts{seq_num})
                {
                    croak "There is a mismatch between the seq number expected from the encrypted reply and the actual one.";
                }
                $flag = 1;
              }
              elsif ( $opts{auth_gss}{service} == RPC_GSS_SVC_PRIVACY)
              {
                my $data = $xdr->opaque();
                $status = GSSAPI::Context::unwrap($opts{gss_context},$data,my $nfsdata, 1,0);
                if ($status -> major != GSS_S_COMPLETE)
                {
                    croak "unwrap error";
                }
                $xdrnew = nfs::Xdr->new(message=>$nfsdata);
                my $seq_num = $xdrnew->uint();
                if ( $seq_num != $opts{seq_num})
                {
                    croak " error seq num mismatch";
                }
                $flag = 1;
              }
         }
    	 if( $flag == 1)
             {
                 $xdr = $xdrnew; #copy the xdr obj
             }
    }



        if ($self->{qsize} > 0) {

            # Save current reply size
            $self->{callq}[-1]{reply_size} = $xdr->remain();
        }

        # Call the results decoding function
        if (defined($opts{out})) {
			$out = &{$opts{out}}(xdr => $xdr);
		}
		 # Check if any bytes remain on the reply buffer
        my $left = $xdr->remain();

        #croak "RPC reply contains $left extra bytes" if $left;
    }

    $Log->exit();
    return $out;
}

#---------------------------------------------------------------------

=head2 message_send

  $rpc_obj->message_send(xdr => $xdr_obj);

Sends the message in the Xdr object to the peer client.
It handles multi-fragment records on TCP.

The following named arguments are accepted:

=over

=item Arguments

=over

=item xdr

(Required) : Xdr object of message to send.

=back

=item Returns

Nothing

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub message_send($%) {
    my $self = shift;
    my (%opts) = validate(@_, {xdr => {type => OBJECT}});
    $Log->enter();
    my $xdr = $opts{xdr};

    # Get the XDR encoded message
    my $message = $xdr->{message};

    # Get message length
    my $len = length($message);

    if ($self->{proto} eq 'tcp') {

        # We support only one RPC header in a request. So we always say
        # this is our last rpc fragment.
        # The header is <Last frag bit> <length of the data>
        $message = pack('N', $len+$LAST_FRG) . $message;

        # Break the RPC message into a sequence
        # of variable-length fragments for TCP
        my ($mesg, $frg_len);
        my $offset   = 0;
        if(!$self->{tcp_frg_size}) {
		    $self->{tcp_frg_size} = $TCP_FRG_SIZE;
	    }
	    my $frg_size = $self->{tcp_frg_size} - 4;
        while ($offset < $len) {

            # Get message fragment
            $mesg = substr($message, $offset, $frg_size);
            $offset += $frg_size;

            # Send the message fragment to the RPC server
            my $debug_len = int(length($mesg) - 4);
            if ($debug_len > 10000) {
                $Log->trace(
                       "writing message (total size: $len) size of $debug_len");
            } else {
                $Log->debug(
                       "writing message (total size: $len) size of $debug_len");
            }

            $self->_syswrite($mesg);
        }
    } else {

        # Check size of complete message
        if ($len > $UDP_MAX_SIZE) {
            croak
              "Rpc::message_send: Size of message is greater than the maximum UDP packet";
        }

        # Send message to peer
        $self->_syswrite($message);
    }

    $Log->exit();
    return;
}

#---------------------------------------------------------------------

=head2 message_receive

  $xdr_obj = $rpc_obj->message_receive();

Receives a message from the peer client.
It handles multi-fragment records on TCP.

=over

=item Returns

Xdr object of received message.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub message_receive($;) {
    my $self = shift;
    my $message;
    $Log->enter();

    if ($self->{proto} eq 'tcp') {
        my ($mesg, $last_frg);
        my $frg_len = 0;

        # Read each fragment at a time
        while (!$last_frg) {

            # Get the fragment size and last fragment bit from message
            $mesg = $self->_sysread(4);
            $frg_len = unpack('N', $mesg);
            croak "Unable to get TCP fragment length" unless $frg_len;
            $last_frg = $frg_len & $LAST_FRG;
            $frg_len &= (~$LAST_FRG);

            # Read the complete fragment from stream
            while ($frg_len) {

                # Try to read the complete fragment at once
                $mesg = $self->_sysread($frg_len);

                # Append current read message to overall message
                $message .= $mesg;

                # Adjust current read size to get missing
                # bytes for current fragment
                $frg_len -= length($mesg);
            }
        }
    } else {

        # Get the complete message at once
        $message = $self->_sysread($UDP_MAX_SIZE);
    }

    $Log->exit();
	# Return XDR object of received message
    return nfs::Xdr->new(message => $message);
}

#---------------------------------------------------------------------

=head2 reconnect

  $rpc_obj->reconnect([ timeout => $timeout ]);

Reconnect TCP connection.

The following named arguments are accepted:

=over

=item Arguments

=over

=item timeout

(Optional) : Timeout waiting to reconnect (default is 30).

=back

=item Returns

Nothing.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub reconnect($;%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {timeout => {type => SCALAR, regex => qr/^\d+$/, default => 30},}
               );
    $Log->enter();
    my ($sock, $err);
    my $endtime = time + $opts{timeout};
    return if $self->{proto} ne 'tcp';

    # Close connection
    $self->{sock}->close;
    $self->{dstaddr} = name2ip(hostname => $self->{hostname})
      if (!$self->{dstaddr});

    do {
        sleep 5;
        eval {

            # Try reconnecting to server
            my %sock_opts;
            $sock_opts{PeerAddr}  = $self->{dstaddr};
            $sock_opts{PeerPort}  = $self->{port};
            $sock_opts{Proto}     = $self->{proto};
            $sock_opts{Timeout}   = $self->{timeout};
            $sock_opts{Blocking}  = 1;
            $sock_opts{LocalPort} = $self->localport;

            if ($self->addr_type() eq "inet6") {
                $sock = IO::Socket::INET6->new(%sock_opts)
                  or croak "Could not reconnect to server using inet6: $!";
            } else {
                $sock = IO::Socket::INET->new(%sock_opts)
                  or croak "Could not reconnect to server using inet4: $!";
            }
        };
        $err = $@;
    } while ($err && time < $endtime);

    croak "Rpc::reconnect failed: $err" if $err;

    # Make socket non-blocking
    fcntl($sock, F_SETFL, O_NONBLOCK);

    # Replace socket object
    $self->{sock} = $sock;

    $Log->exit();
    return;
}

#---------------------------------------------------------------------

=head2 xid

  $xid = $rpc_obj->xid([ $xid_new ]);

Gets or sets the current transaction id (xid) for the RPC connection.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the current transaction id to this.

=back

=item Returns

The xid before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub xid($;$) {
    my ($self, $value) = @_;
    return $self->_set_obj_var('xid', $value);
}

#---------------------------------------------------------------------

=head2 retry

  $retry = $rpc_obj->retry([ $retry_new ]);

Gets or sets the current retry count for the RPC connection.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the current retry count to this.

=back

=item Returns

The retry value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub retry($;$) {
    my ($self, $value) = @_;
    return $self->_set_obj_var('retry', $value);
}

#---------------------------------------------------------------------

=head2 qsize

  $qsize = $rpc_obj->qsize([ $qsize_new ]);

Gets or sets the current qsize for the RPC connection.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the current qsize to this.

=back

=item Returns

The qsize value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub qsize($;$) {
    my ($self, $value) = @_;
    return $self->_set_obj_var('qsize', $value);
}

#---------------------------------------------------------------------

=head2 timeout

  $timeout = $rpc_obj->timeout([ $timeout_new ]);

Gets or sets the current timeout for the RPC connection.

=over

=item Arguments

=over

=item <number>

(Optional) : Set the current timeout to this.

=back

=item Returns

The timeout value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub timeout($;$) {
    my ($self, $value) = @_;
    return $self->_set_obj_var('timeout', $value);
}

#---------------------------------------------------------------------

=head2 wait_for_reply

  $value = $rpc_obj->wait_for_reply([ $new_value ]);

Gets or sets the current value of flag for the RPC connection.
If this flag is set to 1 (default), after sending the call to
the server it waits for the reply. On the other hand, if the
flag is set to 0, it just sends the call to the server and
returns immediately.

=over

=item Arguments

=over

=item <0|1>

(Optional) : Set the current value of flag to this.

=back

=item Returns

The current value before being set.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub wait_for_reply($;$) {
    my ($self, $value) = @_;
    return $self->_set_obj_var('wait_for_reply', $value);
}

#---------------------------------------------------------------------

=head2 program

  $program = $rpc_obj->program();

Returns the program number of the RPC connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub program { return shift->{program} }

#---------------------------------------------------------------------

=head2 version

  $version = $rpc_obj->version();

Returns the version of the RPC connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub version { return shift->{version} }

#---------------------------------------------------------------------

=head2 proto

  $proto = $rpc_obj->proto();

Returns the protocol of the RPC connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub proto { return shift->{proto} }

#---------------------------------------------------------------------

=head2 port

  $port = $rpc_obj->port();

Returns the port number of the RPC connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub port { return shift->{port} }

#---------------------------------------------------------------------

=head2 localport

  $port = $rpc_obj->localport();

Returns the local port number of the RPC connection.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub localport { return shift->{sock}->sockport }

#---------------------------------------------------------------------

=head2 last_call

  $last_call = $rpc_obj->last_call();

Returns the last call.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub last_call { return shift->{last_call} }

#---------------------------------------------------------------------

=head2 callq

  $callq = $rpc_obj->callq();

Returns the call queue.

=for html <hr />

=cut

#---------------------------------------------------------------------
sub callq { return shift->{callq} }

#---------------------------------------------------------------------

=head2 auth_sys_set

  $out = $rpc_obj->auth_sys_set([ uid => $uid ] [, gid => $gid ]);

Sets unix authentication for the RPC connection.

The following named arguments are accepted:

=over

=item Arguments

=over

=item uid

(Optional) : Set authentication to this uid
(default: effective uid of current process).

=item gid

(Optional) : Set authentication to this gid
(default: effective gid of current process).

=back

=item Returns

1 if it succeeds.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub auth_sys_set($%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {uid => {type => SCALAR, default  => $>, regex => qr/^\d+$/},
                gid => {type => SCALAR, optional => 1,  regex => qr/^\d+$/},});
    if (defined($opts{gid})) {

        # Create AUTH_SYS structure for the specified user and group id
        $self->{auth_sys} = {hostname => (uname())[1],
                             uid      => $opts{uid},
                             gid      => $opts{gid},
                             gids     => [$opts{gid}],};
    } else {

        # Create AUTH_SYS structure for the specified user
        # and the current group ids
        my ($gid, @gids) = split(" ", $));
        $self->{auth_sys} = {hostname => (uname())[1],
                             uid      => $opts{uid},
                             gid      => $gid,
                             gids     => \@gids,};
    }

    # Set authentication flavor to sys
    $self->{auth_type} = AUTH_SYS;

    return 1;
}

#---------------------------------------------------------------------

=head2 auth_none_set

  $out = $rpc_obj->auth_none_set();

Sets null authentication for the RPC connection.

It returns 1 if it succeeds.

=cut

#---------------------------------------------------------------------
sub auth_none_set($;) {
    my $self = shift;

    # Set authentication flavor to none
    $self->{auth_type} = AUTH_NONE;

    return 1;
}

sub auth_correct_set(%) {
    my $self = shift;
    my %opts = validate (
        @_,
        {sec    => {type=> SCALAR, optional => 1 },
         hostname=>{type=> SCALAR, optional => 1 },
         uid    => {type=> SCALAR, optional => 1 },
         gid    => {type=> SCALAR, optional => 1 },
        });
    if (!$opts{sec}) {
        $opts{sec} = auth_type();
    }
    my $hostname = $opts{hostname} ? $opts{hostname} : $self->{hostname};
    if ($opts{sec} =~ /^krb5/ && !$hostname) {
        croak (qq{Cannot use auth:RPCSEC_GSS without hostname, wrong programming});
    }

    if ($opts{sec} =~ /^krb5/) {
        $self->{auth_type} = RPCSEC_GSS;
        if ($self->{auth_gss}) { return; }
        # Create a context with the server
        my $krb_nfs_host = "nfs@".$opts{hostname};
        $self->init_cred($krb_nfs_host, _rpc_sec_flavor($opts{sec}));
    } elsif ($opts{sec} eq 'sys') {
        my %sys_arg;
        $sys_arg{uid} = $opts{uid} if $opts{uid};
        $sys_arg{gid} = $opts{gid} if $opts{gid};
        # Create AUTH_SYS structure for current user
        $self->auth_sys_set(%sys_arg);
    } else {
        $self->auth_none_set();
    }

    return 1;
}
#---------------------------------------------------------------------

=head2 pmap_getport

  $port = $rpc_obj->pmap_getport(program => $program,
                                  version => $version,
                                  proto => 'udp'|'tcp'
                                  [, addr_type => 'inet4'|'inet6' ]
                                  [, host => $host ]
                                  [, pmversion => $pmversion ]
                                  [, pmproto => 'udp'|'tcp' ]
                                  [, retry => $retry ]
                                  [, timeout => $timeout ]
                                  [, xid => $xid ]
                                  [, localaddr => $localaddr ]);

  $port = nfs::Rpc::pmap_getport(program => $program,
                                 version => $version,
                                 proto => 'udp'|'tcp',
                                 host => $host
                                 [, addr_type => 'inet4'|'inet6' ]
                                 [, pmversion => $pmversion ]
                                 [, pmproto => 'udp'|'tcp' ]
                                 [, retry => $retry ]
                                 [, timeout => $timeout ]
                                 [, xid => $xid ]
                                 [, localaddr => $localaddr ]);

Get the port number for the program/version/proto specified
using the portmapper.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item program

(Required) : Program number of service.

=item version

(Required) : Version number of program.

=item proto

(Required) : Protocol of service: 'udp', 'tcp'.

=item host

=item sec

(Optional) : Authentication type to be used. Valid values are none, sys, krb5, krb5i, and krb5p.

(Optional) : Hostname/IP address to connect to.
It is optional when called as an instance method with a default being
the current host of the object.
When called as a class method, this argument is required.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item pmversion

(Optional) : Version to connect to portmapper (default is 2).

=item pmproto

(Optional) : Protocol to connect to portmapper: 'udp', 'tcp'.
When called as an instance method the default is the protocol of the object,
else is 'udp'.

=item xid

(Optional) : Initial transaction id.
When called as an instance method the default is the xid of the object,
else is a random-generated number.

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period.
When called as an instance method the default is the retry of the object,
else is 0.

=item timeout

(Optional) : Timeout value for connection.
When called as an instance method the default is the timeout of the object,
else is 30.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=item sec

(Optional) : Security flavour to use. Defaults to none. Accepts any of: 
                     none, sys, krb5, krb5i, krbp

=back

=item Returns

The port number of service.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub pmap_getport(%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) = validate(
        @_,
        {program   => {type => SCALAR, regex   => qr/^\d+$/},
         version   => {type => SCALAR, regex   => qr/^\d+$/},
         proto     => {type => SCALAR, regex   => qr/^(udp|tcp)$/},
         addr_type => {type => SCALAR, default => $self->addr_type()},
         xid       => {
                 type     => SCALAR,
                 regex    => qr/^\d+$/,
                 optional => 1},
         retry => {type     => SCALAR,
                   regex    => qr/^\d+$/,
                   optional => 1},
         timeout => {type     => SCALAR,
                     regex    => qr/^\d+$/,
                     optional => 1},
         pmversion => {type     => SCALAR,
                       regex    => qr/^\d+$/,
                       optional => 1},
         pmproto => {type     => SCALAR,
                     regex    => qr/^(udp|tcp)$/,
                     optional => 1},
         host      => {type => SCALAR, optional => 1},
         localaddr => {type => SCALAR, optional => 1},
         rpc_vers  => {type => SCALAR, optional => 1},
         sec => { type => SCALAR, optional => 1, default => "none",
            regex => qr/^(none|sys|krb5|krb5i|krb5p)$/,},
        });
    my %rpc_opts_h;
	$Log->enter(Dumper(\%opts));

    if (!defined($self) && !defined($opts{host})) {

        # Function is called as a class method: nfs::Rpc::pmap_getport()
        # The host argument is required with this type of call
        croak "pmap_getport: The host argument is required "
          . "when called as a class method";
    }

    # Get server host name or ip address of the server
    my $host = defined($opts{host}) ? $opts{host} : $self->{hostname};

    # get pmversion
    if (!defined($opts{pmversion})) {
        $opts{pmversion} =
          ($self)
          ? $self->pmversion()
          : pmversion(addr_type => $opts{addr_type});
    }

    # Get protocol to connect to portmapper
    my $proto =
        defined($opts{pmproto}) ? $opts{pmproto}
      : defined($self)          ? $self->{proto}
      :                           'udp';

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{xid}
      :                       int(rand(1000000000));

    # Get the correct number of retries
    my $retry =
        defined($opts{retry}) ? $opts{retry}
      : defined($self)        ? $self->{retry}
      :                         0;

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;

    my $rpc_vers =
      defined($opts{rpc_vers})
      ? $opts{rpc_vers}
      : 2;

    # Get the correct local address
    my $localaddr =
        defined($opts{localaddr}) ? $opts{localaddr}
      : defined($self)            ? $self->{localaddr}
      :                             undef;
    $rpc_opts_h{localaddr} = $localaddr if defined($localaddr);

    # Call the portmapper to get port number of requested service
    my $rpc = nfs::Rpc->clnt_create(hostname => $host,
                                    proto    => $proto,
                                    xid      => $xid,
                                    retry    => $retry,
                                    timeout  => $timeout,
                                    sec      => $opts{sec},
                                    program  => 100000,
                                    rpc_vers => $rpc_vers,
                                    version  => $opts{pmversion},
                                    port     => 111,
                                    addr_type=> $opts{addr_type},
                                    %rpc_opts_h);

    # Return output of portmapper procedure 3 (PMAPPROC_GETPORT)
    my $ret;
    if ($rpc->addr_type() eq "inet6") {
        $opts{port}  = 111;
        $opts{proto} = $proto;
        $ret = $rpc->clnt_call(proc => 3,
                               arg  => _xdr_pmap6(%opts),
                               out  => \&_xdr_uint6);
    } else {
        delete $opts{host};
        $ret = $rpc->clnt_call(proc => 3,
                               arg  => _xdr_pmap(%opts),
                               out  => \&_xdr_uint);
    }
    $self->{xid} = $rpc->{xid} if defined($self);

    return $ret;
}

#---------------------------------------------------------------------

=head2 pmap_set

  $out = $rpc_obj->pmap_set(program => $program,
                            version => $version,
                            proto => 'udp'|'tcp',
                            port => $port
                            [, addr_type => 'inet4'|'inet6' ]
                            [, pmversion => $pmversion ]
                            [, pmproto => 'udp'|'tcp' ]
                            [, retry => $retry ]
                            [, timeout => $timeout ]
                            [, xid => $xid ]
                            [, localaddr => $localaddr ]);

  $out = nfs::Rpc::pmap_set(program => $program,
                            version => $version,
                            proto => 'udp'|'tcp',
                            port => $port
                            [, addr_type => 'inet4'|'inet6' ]
                            [, pmversion => $pmversion ]
                            [, pmproto => 'udp'|'tcp' ]
                            [, retry => $retry ]
                            [, timeout => $timeout ]
                            [, xid => $xid ]
                            [, localaddr => $localaddr ]);

Registers the RPC service provided by program/version/proto/port
with the portmapper.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item program

(Required) : Program number of service to register.

=item version

(Required) : Version number of program.

=item proto

(Required) : Protocol of service: 'udp', 'tcp'

=item port

(Required) : Port number of service.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item pmversion

(Optional) : Version to connect to portmapper (default is 2).

=item pmproto

(Optional) : Protocol to connect to portmapper: 'udp', 'tcp'.
When called as an instance method the default is the protocol of the object,
else is 'udp'.

=item xid

(Optional) : Initial transaction id.
When called as an instance method the default is the xid of the object,
else is a random-generated number.

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period.
When called as an instance method the default is the retry of the object,
else is 0.

=item timeout

(Optional) : Timeout value for connection.
When called as an instance method the default is the timeout of the object,
else is 30.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=back

=item Returns

1 if service was successfully registered, 0 otherwise.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub pmap_set(%) {
    $Log->enter();
    my $self = shift if ((@_ % 2) == 1);
    $Log->enter(Dumper($self)) if $self;
    my (%opts) =
      validate(@_,
               {program   => {type => SCALAR, regex   => qr/^\d+$/},
                version   => {type => SCALAR, regex   => qr/^\d+$/},
                port      => {type => SCALAR, regex   => qr/^\d+$/},
                addr_type => {type => SCALAR, default => $self ? $self->addr_type() : addr_type()},
                proto     => {type => SCALAR, regex   => qr/^(udp|tcp)$/},
                xid       => {
                        type     => SCALAR,
                        regex    => qr/^\d+$/,
                        optional => 1},
                retry => {type     => SCALAR,
                          regex    => qr/^\d+$/,
                          optional => 1},
                timeout => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                pmversion => {type     => SCALAR,
                              regex    => qr/^\d+$/,
                              optional => 1,},
                pmproto => {type     => SCALAR,
                            regex    => qr/^(udp|tcp)$/,
                            optional => 1},
                localaddr => {type => SCALAR, optional => 1},
                auth_type => {type => SCALAR, optional => 1, default => AUTH_SYS},
        });
                #Leaving the default to AUTH_SYS, since kerberized callback is not yet in vogue.


    my %rpc_opts_h;

    $Log->enter("pmap_set input:" . Dumper(\%opts));

    # get pmversion
    if (!defined($opts{pmversion})) {
        $opts{pmversion} =
          ($self)
          ? $self->pmversion()
          : pmversion(addr_type => $opts{addr_type});
    }

    # Get protocol to connect to portmapper
    my $proto =
        defined($opts{pmproto}) ? $opts{pmproto}
      : defined($self)          ? $self->{proto}
      :                           'udp';

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{xid}
      :                       int(rand(1000000000));

    # Get the correct number of retries
    my $retry =
        defined($opts{retry}) ? $opts{retry}
      : defined($self)        ? $self->{retry}
      :                         0;

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;

    # Get the correct local address
    my $localaddr =
        defined($opts{localaddr}) ? $opts{localaddr}
      : defined($self)            ? $self->{localaddr}
      :                             undef;
    $rpc_opts_h{localaddr} = $localaddr if defined($localaddr);

    $opts{host} = ($self) ? $self->loopback() : nfs::Rpc->loopback();

    # Call the portmapper on 'udp' to register the program number
    my $rpc = nfs::Rpc->clnt_create(hostname => $opts{host},
                                    proto    => $proto,
									addr_type=> $opts{addr_type},
                                    xid      => $xid,
                                    retry    => $retry,
                                    timeout  => $timeout,
                                    program  => 100000,
                                    version  => $opts{pmversion},
                                    port     => 111,
                                    auth_type=> $opts{auth_type},
                                    %rpc_opts_h);

	$rpc->{dispatch} = $self->{callback}->{dispatch} if ($self->{callback} && $self->{callback}->{dispatch});
	$Log->enter("Rpc has dispatch" . Dumper($rpc->{dispatch})) if ($rpc->{dispatch});

    # Return output of portmapper procedure 1 (PMAPPROC_SET)
    my $ret;
    if ($rpc->addr_type() eq "inet6") {
		$Log->enter(qq{setting up for inet6});
        $opts{port}  = 111;
        $opts{proto} = $proto;
        $ret = $rpc->clnt_call(proc => 1,
                               arg  => _xdr_pmap6(%opts),
                               out  => \&_xdr_uint);
    } else {
		$Log->enter(qq{setting up for inet4});
        delete $opts{host};

        $ret = $rpc->clnt_call(proc => 1,
                               arg  => _xdr_pmap(%opts),
                               out  => \&_xdr_uint,
                               auth_type => $opts{auth_type});
    }
    $self->{xid} = $rpc->{xid} if defined($self);

    return $ret;
}

#---------------------------------------------------------------------

=head2 pmap_unset

  $out = $rpc_obj->pmap_unset(program => $program,
                              version => $version,
                              proto => 'udp'|'tcp',
                              port => $port
                              [, addr_type => 'inet4'|'inet6' ]
                              [, pmversion => $pmversion ]
                              [, pmproto => 'udp'|'tcp' ]
                              [, retry => $retry ]
                              [, timeout => $timeout ]
                              [, xid => $xid ]
                              [, localaddr => $localaddr ]);

  $out = nfs::Rpc::pmap_unset(program => $program,
                              version => $version,
                              proto => 'udp'|'tcp',
                              port => $port
                              [, addr_type => 'inet4'|'inet6' ]
                              [, pmversion => $pmversion ]
                              [, pmproto => 'udp'|'tcp' ]
                              [, retry => $retry ]
                              [, timeout => $timeout ]
                              [, xid => $xid ]
                              [, localaddr => $localaddr ]);

Unregisters the RPC service provided by program/version/proto/port
with the portmapper.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item program

(Required) : Program number of service to unregister.

=item version

(Required) : Version number of program.

=item proto

(Required) : Protocol of service: 'udp', 'tcp'

=item port

(Required) : Port number of service.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item pmversion

(Optional) : Version to connect to portmapper (default is 2).

=item pmproto

(Optional) : Protocol to connect to portmapper: 'udp', 'tcp'.
When called as an instance method the default is the protocol of the object,
else is 'udp'.

=item xid

(Optional) : Initial transaction id.
When called as an instance method the default is the xid of the object,
else is a random-generated number.

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period.
When called as an instance method the default is the retry of the object,
else is 0.

=item timeout

(Optional) : Timeout value for connection.
When called as an instance method the default is the timeout of the object,
else is 30.

=item localaddr

(Optional) : Hostname or ip address of local machine.

=back

=item Returns

1 if service was successfully unregistered, 0 otherwise.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub pmap_unset(%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_,
               {program   => {type => SCALAR, regex   => qr/^\d+$/},
                version   => {type => SCALAR, regex   => qr/^\d+$/},
                port      => {type => SCALAR, regex   => qr/^\d+$/},
                addr_type => {type => SCALAR, default => addr_type()},
                proto     => {type => SCALAR, regex   => qr/^(udp|tcp)$/},
                xid       => {
                        type     => SCALAR,
                        regex    => qr/^\d+$/,
                        optional => 1},
                retry => {type     => SCALAR,
                          regex    => qr/^\d+$/,
                          optional => 1},
                timeout => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                pmversion => {type     => SCALAR,
                              regex    => qr/^\d+$/,
                              optional => 1,},
                pmproto => {type     => SCALAR,
                            regex    => qr/^(udp|tcp)$/,
                            optional => 1},
                localaddr => {type => SCALAR, optional => 1},});
    my %rpc_opts_h;

    # get pmversion
    if (!defined($opts{pmversion})) {
        $opts{pmversion} =
          ($self)
          ? $self->pmversion()
          : pmversion(addr_type => $opts{addr_type});
    }

    # Get protocol to connect to portmapper
    my $proto =
        defined($opts{pmproto}) ? $opts{pmproto}
      : defined($self)          ? $self->{proto}
      :                           'udp';

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{xid}
      :                       int(rand(1000000000));

    # Get the correct number of retries
    my $retry =
        defined($opts{retry}) ? $opts{retry}
      : defined($self)        ? $self->{retry}
      :                         0;

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;

    # Get the correct local address
    my $localaddr =
        defined($opts{localaddr}) ? $opts{localaddr}
      : defined($self)            ? $self->{localaddr}
      :                             undef;
    $rpc_opts_h{localaddr} = $localaddr if defined($localaddr);

    my $host = ($self) ? $self->loopback() : nfs::Rpc->loopback();

    # Call the portmapper on 'udp' to unregister the program number
    my $rpc = nfs::Rpc->clnt_create(hostname => "$host",
                                    proto    => $proto,
                                    xid      => $xid,
                                    retry    => $retry,
                                    timeout  => $timeout,
                                    program  => 100000,
                                    version  => $opts{pmversion},
                                    port     => 111,
                                    %rpc_opts_h);

    # Call portmapper procedure 2 (PMAPPROC_UNSET)
    my $ret;
    if ($rpc->addr_type() eq "inet6") {
        $opts{port}  = 111;
        $opts{proto} = $proto;
        $ret = $rpc->clnt_call(proc => 2,
                               arg  => _xdr_pmap6(%opts),
                               out  => \&_xdr_uint);
    } else {
        delete $opts{host};
        $ret = $rpc->clnt_call(proc => 2,
                               arg  => _xdr_pmap(%opts),
                               out  => \&_xdr_uint);
    }
    $self->{xid} = $rpc->{xid} if defined($self);

    return $ret;
}

#---------------------------------------------------------------------

=head2 pmap_dump

  $plist = $rpc_obj->pmap_dump([ host => $hostname ]
                               [, addr_type => 'inet4'|'inet6' ]
                               [, xid => $xid ]
                               [, retry => $retry ]
                               [, timeout => $timeout ]
                               [, pmversion => $pmversion ]
                               [, pmproto => 'udp'|'tcp' ]);

  $plist = nfs::Rpc::pmap_dump([ host => $hostname ]
                               [, addr_type => 'inet4'|'inet6' ]
                               [, xid => $xid ]
                               [, retry => $retry ]
                               [, timeout => $timeout ]
                               [, pmversion => $pmversion ]
                               [, pmproto => 'udp'|'tcp' ]);

Dump all programs currently register on the host.
This method may be called as a class or instance method.

The following named arguments are accepted:

=over

=item Arguments

=over

=item host

(Optional) : Hostname/IP address to connect to.
When called as an instance method the default is the current host of the
object, else is localhost.

=item addr_type

(Optional) : IP Address family type 'inet4' or 'inet6'.

=item xid

(Optional) : Initial transaction id.
When called as an instance method the default is the xid of the object,
else is a random-generated number.

=item retry

(Optional) : Number of times to retry each call if response is not
received within the timeout period.
When called as an instance method the default is the retry of the object,
else is 0.

=item timeout

(Optional) : Timeout value for connection.
When called as an instance method the default is the timeout of the object,
else is 30.

=item pmversion

(Optional) : Version to connect to portmapper (default is 2).

=item pmproto

(Optional) : Protocol to connect to portmapper: 'udp', 'tcp'.
When called as an instance method the default is the protocol of the object,
else is 'udp'.

=back

=item Returns

A reference to a list of services currently registered with the portmapper.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub pmap_dump(%) {
    my $self = shift if ((@_ % 2) == 1);
    my (%opts) =
      validate(@_,
               {host      => {type => SCALAR, optional => 1},
                addr_type => {type => SCALAR, default  => addr_type()},
                xid       => {
                        type     => SCALAR,
                        regex    => qr/^\d+$/,
                        optional => 1},
                retry => {type     => SCALAR,
                          regex    => qr/^\d+$/,
                          optional => 1},
                timeout => {type     => SCALAR,
                            regex    => qr/^\d+$/,
                            optional => 1},
                pmversion => {type     => SCALAR,
                              regex    => qr/^\d+$/,
                              optional => 1},
                pmproto => {type     => SCALAR,
                            regex    => qr/^(udp|tcp)$/,
                            optional => 1},});

    # Get server host name or ip address of the server
    my $host =
        defined($opts{host}) ? $opts{host}
      : defined($self)       ? $self->{hostname}
      :                        $self->loopback();

    # get pmversion
    if (!defined($opts{pmversion})) {
        $opts{pmversion} =
          ($self)
          ? $self->pmversion()
          : pmversion(addr_type => $opts{addr_type});
    }

    # Get protocol to connect to portmapper
    my $proto =
        defined($opts{pmproto}) ? $opts{pmproto}
      : defined($self)          ? $self->{proto}
      :                           'udp';

    # Get the correct RPC transaction id
    my $xid =
        defined($opts{xid}) ? $opts{xid}
      : defined($self)      ? $self->{xid}
      :                       int(rand(1000000000));

    # Get the correct number of retries
    my $retry =
        defined($opts{retry}) ? $opts{retry}
      : defined($self)        ? $self->{retry}
      :                         0;

    # Get the correct timeout
    my $timeout =
        defined($opts{timeout}) ? $opts{timeout}
      : defined($self)          ? $self->{timeout}
      :                           30;

    # Call the portmapper on 'udp'
    my $rpc = nfs::Rpc->clnt_create(hostname => "$host",
                                    proto    => $proto,
                                    xid      => $xid,
                                    retry    => $retry,
                                    timeout  => $timeout,
                                    program  => 100000,
                                    version  => $opts{pmversion},
                                    port     => 111);

    # Call portmapper procedure 4 (PMAPPROC_DUMP)
    my $ret = $rpc->clnt_call(proc => 4, out => \&_xdr_pmap_dump);
    $self->{xid} = $rpc->{xid} if defined($self);

    return $ret;
}

#---------------------------------------------------------------------

=head2 svc_register_transient

  $rpc_obj->svc_register_transient(version => $version,
                                   proto => 'udp'|'tcp',
                                   port => $port);

Register a service with a program number in the transient range.

The following named arguments are accepted:

=over

=item Arguments

=over

=item version

(Required) : Version number of program.

=item proto

(Required) : Protocol of service: 'udp', 'tcp'

=item port

(Required) : Port number of service.

=back

=item Returns

The registered program number if successful
or undef if unable to register the service.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub svc_register_transient($;%) {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {version => {type => SCALAR, regex => qr/^\d+$/},
                port    => {type => SCALAR, regex => qr/^\d+$/},
                proto   => {type => SCALAR, regex => qr/^(udp|tcp)$/},});

    # The transient range
    my $program = 0x40000000;
    my $maxprog = 0x50000000;

    # Try to register the service with the portmapper
    # starting at the beginning of the transient range
    # until the registration is successful
    while (   $self->pmap_set(%opts, program => $program) == 0
           && $program < $maxprog) {
        ++$program;
    }

    if ($program < $maxprog) {

        # Registration was successful so
        # return the program number
        return $program;
    }

    return;
}

#---------------------------------------------------------------------
# Internal function
# Encodes the arguments for portmapper procedures into XDR format.
# It takes the program, version, proto and port number as
# arguments and returns the XDR encoded message.
#---------------------------------------------------------------------
sub _xdr_pmap(%) {
    my (%opts) = @_;
	$Log->enter(Dumper(\%opts));
    my $port = defined($opts{port}) ? $opts{port} : 0;

    my $xdr = nfs::Xdr->new();
    $xdr->uint(encode => $opts{program});
    $xdr->uint(encode => $opts{version});
    $xdr->uint(encode => $ipproto{$opts{proto}});
    $xdr->uint(encode => $port);

    return $xdr->{message};
}

sub _xdr_pmap6(%) {
    my (%opts) = @_;
    $Log->enter(Dumper(\%opts));
    my $port = defined($opts{port}) ? $opts{port} : 0;

    my $xdr = nfs::Xdr->new();
    $xdr->uint(encode => $opts{program});
    $xdr->uint(encode => $opts{version});
    $xdr->opaque(encode => "$opts{proto}6");

    my ($hport) = ($opts{port} >> 8) & 0x0FF;
    my ($lport) = $opts{port} & 0x0FF;

	if (!$opts{host}) {
		$Log->enter("Need to know if this is localaddr actually");
		$opts{host} = "::1";
	}
    $xdr->opaque(encode => "$opts{host}.$hport.$lport");
    $xdr->opaque(encode => "");

    return $xdr->{message};
}

#---------------------------------------------------------------------
# Internal function
# Decodes a 32bit unsigned integer.
# It takes the Xdr object as an argument and returns the decoded
# integer.
#---------------------------------------------------------------------
sub _xdr_uint(%) {
    my (%opts) = @_;

    # Decode and return port number
    return $opts{xdr}->uint();
}

sub _xdr_uint6(%) {
    my (%opts) = @_;

    # Decode and return port number
    my ($full_string) = $opts{xdr}->opaque();

    if ($full_string !~ /([0-9a-f:]+).([0-9]+).([0-9]+)/) {
        croak
          "Portmapper V4 GETADDR Call received a bad reply address: $full_string";
    } else {
        return ($2 * 256 + $3);
    }
}

if (0) {
	my (%opts) = @_;

	# Decode and return true or croak on failure.
	my $result = $opts{xdr}->uint();

	return $result;
}

#---------------------------------------------------------------------
# Internal function
# Decodes the list of from pmap_dump.
# It takes the Xdr object as an argument and returns the decoded
# list.
#---------------------------------------------------------------------
sub _xdr_pmap_dump(%) {
    my (%opts) = @_;
    my @list;
    my %protos = reverse(%ipproto);
    my $xdr    = $opts{xdr};
    while ($xdr->uint()) {
        push @list,
          { program => $xdr->uint(),
            version => $xdr->uint(),
            proto   => $protos{$xdr->uint()},
            port    => $xdr->uint()};
    }

    return \@list;
}

#---------------------------------------------------------------------
# Internal function
# XDR encoding of AUTH_SYS authentication flavor
# Even though this is an internal function, the auth_sys hash can
# be set externally so validate arguments
#---------------------------------------------------------------------
sub _xdr_auth_sys(%) {
    my (%opts) =
      validate(@_,
               {hostname => {type => SCALAR},
                uid      => {type => SCALAR, regex => qr/^\d+$/},
                gid      => {type => SCALAR, regex => qr/^\d+$/},
                gids     => {type => ARRAYREF},});
    my $xdr = nfs::Xdr->new();

    $xdr->uint(encode => time);    # Time stamp
    $xdr->opaque(encode => $opts{hostname}, variable => 255);
    $xdr->uint(encode => $opts{uid});
    $xdr->uint(encode => $opts{gid});

    # Encode array of gids
    $xdr->uint(encode => scalar(@{$opts{gids}}));
    foreach my $gid (@{$opts{gids}}) {
        $xdr->uint(encode => $gid);
    }

    return $xdr->{message};
}

#---------------------------------------------------------------------
# Internal function
# Creates a socket and binds to the local port if specified
# or to any local port if not specified.  If the resvport
# is defined then it tries to bind to a reserved port.
# It returns the created socket object.
#---------------------------------------------------------------------
sub _sock_bind($;) {
    my $self = shift;
    $Log->enter(Dumper($self));
    my (%sock_opts_h, $sock);
    my $port       = -1;
    my $start_port = 600;
    my $end_port   = 1023;
    my $nports     = $end_port - $start_port + 1;

    if ($self->{resvport}) {

        # Start with a random port number within
        # the valid reserved port range
        $port = int(rand($nports)) + $start_port;
    }

    $self->{dstaddr} = $self->name2ip(hostname => $self->{hostname})
      if (!$self->{dstaddr});

    if (defined($self->{localaddr})) {
        # Use given local address
        $sock_opts_h{LocalAddr} = $self->{localaddr};
    }

    # Loop through the valid reserved port range
    # If the resvport option is false, then port is now -1
    # which is incremented by one to zero so binding to any
    # local port number is attempted
    foreach (1 .. $nports) {

        # Go back to the start of the valid reserved port range
        # if at the end of range
        $port = $start_port if (++$port > $end_port);
        if ($port) {
            # Connect to the specified port number
            $sock_opts_h{LocalPort} = $port;
        }

        # Connect socket
        $sock_opts_h{PeerAddr}  = $self->{dstaddr};
        $sock_opts_h{PeerPort}  = $self->{port};
        $sock_opts_h{Proto}     = $self->{proto};
        $sock_opts_h{Timeout}   = $self->{timeout};
        $sock_opts_h{Blocking}  = 1;
        $sock_opts_h{qsize}     = 4;
        $sock_opts_h{ReuseAddr} = 1;

        $Log->debug("Args for socket : " . Dumper(\%sock_opts_h));
        if ($self->addr_type() eq "inet6") {
            $sock = IO::Socket::INET6->new(%sock_opts_h);
        } else {
            $sock = IO::Socket::INET->new(%sock_opts_h);
        }

        # Try another port number if it is in use,
        # otherwise croak with the error
        last if ($sock || $! !~ /Address already in use/);
    }

    croak qq{Can't bind to socket using the arguments:} . Dumper(\%sock_opts_h) . qq{. and self } . Dumper($self) . qq{. The error: $!}
      unless $sock;

    # Save local address, replace original definition
    # (local name or ip address) with the local ip address
    $self->{localaddr} = $sock->sockhost();

    if ($self->addr_type() ne "inet"
        . Net::IP::ip_get_version($self->{localaddr})) {
        croak "$self->{localaddr} is not desired addr family : "
          . $self->addr_type() . ". It is " . Net::IP::ip_get_version($self->{localaddr});
    }

    # Make socket non-blocking
    fcntl($sock, F_SETFL, O_NONBLOCK);

    $Log->exit();
    return $sock;
}

#---------------------------------------------------------------------
# Internal function.
# Reads data from the socket using select to timeout the request
#---------------------------------------------------------------------
sub _sysread($$;) {
    my ($self, $num) = @_;
    my ($rout, $stime, $nread, $n);
    my $rin         = '';
    my $buf         = '';
    my $sel_timeout = $self->{timeout} ? 0.25 : 0;
    $Log->enter();

    # Just check file descriptor for current read socket
    vec($rin, fileno($self->{sock}), 1) = 1;

    $stime = time();
    while (time() - $stime < $self->{timeout}) {

        # Wait until the socket is ready for 250ms if timeout is set
        if (select($rout = $rin, undef, undef, $sel_timeout)) {
            $nread = length($buf);

            # Read data and append it to buffer
            if ($self->{proto} eq 'tcp' || !$self->{server}) {
                $n = sysread($self->{sock}, $buf, $num - $nread, $nread);
            } else {
                $n = recv($self->{sock}, $buf, $num, 0);
                if (defined($n)) {
                    $self->{srcpaddr} = $n;
                    $n = $num;
                }
            }

            # Try again if system call was interrupted
            # or if read would block
            if (!defined($n)) {
                if ($! == EINTR || $! == EAGAIN) {
                    next;
                } else {

                    # Error on read
                    croak "_sysread: $!";
                }
            } elsif ($n == 0) {

                # TCP connection was reset
                croak '_sysread: Connection reset by server';
            }

            $Log->exit();

            # Return read data
            return $buf;
        }
    }
    croak "_sysread: Timeout waiting for data to be received";
}

#---------------------------------------------------------------------
# Internal function.
# Writes data to the socket using select to timeout the request
#---------------------------------------------------------------------
sub _syswrite($$;) {
    my ($self, $buf) = @_;
    my ($wout, $stime, $n);
    my $win         = '';
    my $total       = 0;
    my $sel_timeout = $self->{timeout} ? 0.25 : 0;

    # Just check file descriptor for current write socket
    vec($win, fileno($self->{sock}), 1) = 1;

    $stime = time();
    while (time() - $stime < $self->{timeout}) {

        # Wait until the socket is ready for 250ms if timeout is set
        if (select(undef, $wout = $win, undef, $sel_timeout)) {

            # Write data from buffer
            if ($self->{proto} eq 'tcp'
                || !$self->{server}
                || !defined($self->{srcpaddr})) {
                $n =
                  syswrite($self->{sock}, $buf, length($buf) - $total, $total);
            } else {
                $n = send($self->{sock}, $buf, 0, $self->{srcpaddr});
                $n = length($buf) if defined($n);
            }
            if (defined($n)) {
                $total += $n;
                next if (length($buf) - $total);
            } else {

                # Try again if system call was interrupted
                # or if write would block
                if ($! == EINTR || $! == EAGAIN) {
                    next;
                } else {

                    # Error on write
                    croak "_syswrite: $!";
                }
            }

            # Return number of bytes actually written
            return $total;
        }
    }
    croak "_syswrite: Timeout waiting for data to be sent";
}

#---------------------------------------------------------------------
# Internal function
# Get or set an object variable
#---------------------------------------------------------------------
sub _set_obj_var($$$;) {
    my ($self, $var, $new_value) = @_;
    my $old = $self->{$var};

    if (defined($new_value)) {
        $self->{$var} = $new_value;
    }

    return $old;
}

sub test_for_cb($%) {
    my $self = shift;
	$Log->enter(qq{self:} . Dumper($self));
    my (%opts) = validate(
        @_,
        {xid => {type => SCALAR},
         xdr => {type => HASHREF},

        });
    my $xdr = $opts{xdr};
    my $xid = $opts{xid};
    my ($rpcver, $program, $version, $proc, $credential, $verifier);
    my ($rout, $out, %rpc_hdr);
    my $reply_status    = 0;
    my $accepted_status = 0;
    my $rejected_status = 0;

    eval {
        $rpcver  = $xdr->uint();
        $program = $xdr->uint();
        $version = $xdr->uint();
        $proc    = $xdr->uint();
        $credential = {flavor => $xdr->uint(),
                       auth   => $xdr->opaque(variable => 400)};
        $verifier = {flavor => $xdr->uint(),
                     auth   => $xdr->opaque(variable => 400)};

    };

    # RPC call header
    %rpc_hdr = (proto      => 'tcp',
                xid        => $xid,
                rpcver     => $rpcver,
                program    => $program,
                version    => $version,
                proc       => $proc,
                credential => $credential,
                verifier   => $verifier,);
    if ($proc == 0) {
        $Log->trace("found CB_NULL.");
        $Log->debug("xdr data: " . Dumper($xdr));
        process_cb_null($self, $xdr, %rpc_hdr);

    } elsif ($proc == 1) {
        $Log->trace("found CB_COMPOUND.");
        $Log->debug("xdr data: " . Dumper($xdr));
        process_cb_compound($self, $xdr, %rpc_hdr);
    }

	$Log->exit();
    return 1;
}

sub process_cb_null($$%) {

    my ($self, $xdr, %rpc_hdr) = @_;
    my $reply_status    = 0;
    my $accepted_status = 0;
    my $rejected_status = 0;
    my $out;

    my $proto      = 'tcp';
    my $xid        = $rpc_hdr{xid};
    my $rpcver     = $rpc_hdr{rpcver};
    my $program    = $rpc_hdr{program};
    my $version    = $rpc_hdr{version};
    my $proc       = $rpc_hdr{proc};
    my $credential = $rpc_hdr{credential};
    my $verifier   = $rpc_hdr{verifier};

    # Encode reply header
    my $reply_xdr = nfs::Xdr->new();
    $reply_xdr->uint(encode => $xid);
    $reply_xdr->uint(encode => 1);

    $Log->trace("Creating the reply xdr");

    # Check correct version of RPC
    if ($rpcver != 2) {
        $reply_status    = 1;
        $rejected_status = 0;    # RPC MISMATCH
    } elsif ($credential->{flavor} > 1 || $verifier->{flavor} > 0) {
        $reply_status    = 1;
        $rejected_status = 1;    # AUTH ERROR
    }

    # Encode the reply status
    $reply_xdr->uint(encode => $reply_status);

    if ($reply_status) {

        # MSG_DENIED
        $reply_xdr->uint(encode => $rejected_status);
        if ($rejected_status) {
            if ($credential->{flavor} > 1) {

                # AUTH ERROR: AUTH_TOOWEAK
                $reply_xdr->uint(encode => 5);
            } else {

                # AUTH ERROR: AUTH_BADVERF
                $reply_xdr->uint(encode => 3);
            }
        } else {

            # RPC MISMATCH: (low, high)
            $reply_xdr->uint(encode => 2);
            $reply_xdr->uint(encode => 2);
        }
    } else {

        # Check proper program, procedure and version

        # Encode the verifier
        $reply_xdr->uint(encode => 0);
        $reply_xdr->uint(encode => 0);
        $accepted_status = 0;

        $Log->debug("Accepted status is $accepted_status");

        # So far everything is fine
        # Call appropriate procedure
        $out = eval {
			$Log->enter(qq{self =} . Dumper($self));
			$Log->enter(qq{xdr =} . Dumper($xdr));
			$Log->enter(qq{rpc_hdr =} . Dumper(\%rpc_hdr));
			if ($self->{dispatch}) {
            	&{$self->{dispatch}{$proc}}(xdr     => $xdr,
                                        rpc_hdr => \%rpc_hdr);
			} else {
				$Log->enter(qq{There is no dispatching given for this callback. Ignoring it});
			}
        };
		if ($@) {
			croak (qq{dispatching request failed. Reason:$@});
		}

        # Encode the accepted status
        $reply_xdr->uint(encode => $accepted_status);
        $Log->debug("accepted status is $accepted_status");

        if (!$accepted_status) {

            # Append procedure results to reply message
            my $len = defined($out) ? length($out) : 0;

            $Log->debug("len is $len and out is $out if this matches in here")
              if $len;

            $reply_xdr->opaque(encode => $out, fixed => $len) if $len;

        } elsif ($accepted_status == 2) {

            # PROG_MISMATCH
            $reply_xdr->uint(encode => $self->{version});
            $reply_xdr->uint(encode => $self->{version});
        }
    }

    $Log->debug("Reply xdr is " . Dumper($reply_xdr));

    # Send the reply to the client
    $self->message_send(xdr => $reply_xdr);
    $self->{cb_null} = 1;
    $self->{hidden_callback_flag} = 1;

    return 1;
}

sub process_cb_compound ($$%) {
    my ($self, $xdr, %rpc_hdr) = @_;
    my $reply_status    = 0;
    my $accepted_status = 0;
    my $rejected_status = 0;
    my $out;

    my $proto      = 'tcp';
    my $xid        = $rpc_hdr{xid};
    my $rpcver     = $rpc_hdr{rpcver};
    my $program    = $rpc_hdr{program};
    my $version    = $rpc_hdr{version};
    my $proc       = $rpc_hdr{proc};
    my $credential = $rpc_hdr{credential};
    my $verifier   = $rpc_hdr{verifier};

    my $tag_l    = $xdr->uint();
    my $tag      = $xdr->opaque(fixed => $tag_l);
    my $min_ver  = $xdr->uint();
    my $cb_ident = $xdr->uint();
    my $number   = $xdr->uint();

    my (@opcode, @r);
    my $i = 0;
    while ($xdr->remain() != 0) {
        $opcode[$i] = $xdr->uint();
        $r[$i] = parse_cb_ops($xdr, $opcode[$i]);
        $i++;
    }
    my $r        = $r[0];
    my $r1       = $r[1];
    my $opcode   = $opcode[0];
    my $opcode_1 = $opcode[1];

    $Log->debug(
        "in process_cb_compound: opcode is $opcode and number of ops is $number"
        );

    $Log->trace("Creating the reply xdr");

    my $reply_xdr = nfs::Xdr->new();
    $reply_xdr->uint(encode => $xid);
    $reply_xdr->uint(encode => 1);

    # Check correct version of RPC
    if ($rpcver != 2) {
        $reply_status    = 1;
        $rejected_status = 0;    # RPC MISMATCH

    } elsif ($credential->{flavor} > 1 || $verifier->{flavor} > 0) {
        $reply_status    = 1;
        $rejected_status = 1;    # AUTH ERROR
    }

    # Encode the reply status
    $reply_xdr->uint(encode => $reply_status);

    # Remove this just hardcoding to check something

    if ($reply_status) {

        # MSG_DENIED
        $reply_xdr->uint(encode => $rejected_status);
        if ($rejected_status) {
            if ($credential->{flavor} > 1) {

                # AUTH ERROR: AUTH_TOOWEAK
                $reply_xdr->uint(encode => 5);

            } else {

                # AUTH ERROR: AUTH_BADVERF
                $reply_xdr->uint(encode => 3);
            }

        } else {

            # RPC MISMATCH: (low, high)
            $reply_xdr->uint(encode => 2);
            $reply_xdr->uint(encode => 2);
        }

    } else {

        # Check proper program, procedure and version
        # Encode the verifier
        $reply_xdr->uint(encode => 0);
        $reply_xdr->uint(encode => 0);
        $accepted_status = 0;

        # So far everything is fine
        # Call appropriate procedure
        $out = eval {
            &{$self->{dispatch}{$proc}}(xdr     => $xdr,
                                        rpc_hdr => \%rpc_hdr);
        };
    }

    # Encode the accepted status
    $reply_xdr->uint(encode => $accepted_status);

    if (!$accepted_status) {

        # Append procedure results to reply message
        my $len = defined($out) ? length($out) : 0;
        $reply_xdr->opaque(encode => $out, fixed => $len) if $len;
    } elsif ($accepted_status == 2) {

        # PROG_MISMATCH
        $reply_xdr->uint(encode => $self->{version});
        $reply_xdr->uint(encode => $self->{version});
    }

    #Encode the CB_COMPOUND output
    $reply_xdr->uint(encode => 0);
    $reply_xdr->opaque(encode => 'ONTAP');
    $reply_xdr->uint(encode => 2);
    $reply_xdr->uint(encode => 11);
    $reply_xdr->uint(encode => 0);
    $reply_xdr->opaque(encode => $r->{sessionid}, fixed => 16);
    $reply_xdr->uint(encode => $r->{seq_id4});
    $reply_xdr->uint(encode => $r->{sid1});
    $reply_xdr->uint(encode => $r->{sid2});
    $reply_xdr->uint(encode => 10000);
    $reply_xdr->uint(encode => 5);
    $reply_xdr->uint(encode => 0);

    if ($opcode_1 == 5) {
        $self->{cb_layoutreturn} = 1;
    }

    $Log->debug("Reply xdr is " . Dumper($reply_xdr));

    # Send the reply to the client
    $self->message_send(xdr => $reply_xdr);

    return 1;
}

sub parse_cb_ops($$;) {
    my ($xdr, $op) = @_;
    my $ret;

    if ($op == 11) {
        $ret->{sessionid} = $xdr->opaque(fixed => 16);
        $ret->{seq_id4}   = $xdr->uint();
        $ret->{sid1}      = $xdr->uint();
        $ret->{sid2}      = $xdr->uint();
        $ret->{cache}     = $xdr->uint();
        $ret->{refer}     = $xdr->uint();

    } elsif ($op == 5) {
        $ret->{layout} = $xdr->uint();
        $ret->{iomode} = $xdr->uint();
        $ret->{clora}  = $xdr->uint();
        $ret->{type}   = $xdr->uint();

        if ($ret->{type} == 1) {

            $ret->{nf_fh4} = $xdr->opaque(fixed => 128);
            $ret->{offset} = $xdr->uint64();
            $ret->{length} = $xdr->uint64();
            $ret->{seqid}  = $xdr->uint();
            $ret->{other}  = $xdr->opaque(fixed => 12);

        } elsif ($ret->{type} == 2) {
            $ret->{f_major} = $xdr->uint64();
            $ret->{f_minor} = $xdr->uint64();
        }
    }
    return $ret;
}

## _rpc_sec_flavour
## 
## Internal RPC function to determine the type of kerberos to use.
##
sub _rpc_sec_flavor($) {
    my $string = shift;
    if ($string !~ /^(krb5|krb5i|krb5p)$/) {
        croak ("$string is an unknown kerberos flavour. Supported values:krb5, krb5i, krb5p");
    }
    my %dummy = (
        krb5    => RPC_GSS_SVC_NONE,
        krb5i    => RPC_GSS_SVC_INTEGRITY, 
        krb5p    => RPC_GSS_SVC_PRIVACY,
    );

    return $dummy{$string};
}

## init_cred
## newly added for RPCSEC GSS support
## Initializing the credentials using RPCSEC GSS, and GSSAPI.
## Internal function used by RPC.
## 
## Gets the token for the kerberos communication, establishes a context for 
## further communications.
##
sub init_cred($$$;)
{
    # GSSAPI::Name->import produces $gss_server_name
    # which is then passed in to GSSAPI::Context::init
    # $gss_server_name represents the principcal name
    # of the app server to which we are authenticating
    
    my $self = shift;
    my $server_name = shift;
    my $service = shift;
    my $xdr = nfs::Xdr->new();
    my $token_len='';
    my $init_res='';
    my $handle = '';
    my $gss_input_token = '';
    my $gss_auth_flags = 0;
    my $error = 0;
    my $proc = RPCSEC_GSS_INIT;
    my $token='';    
    my $out_flags='';
    my $out_time='';
    my $gss_server='';
    my $gss_output_token='';
    my $ctx = GSSAPI::Context->new();
    my $gss_server_name = GSSAPI::Name->new();
    $self->{auth_type}=RPCSEC_GSS;
    
    require GSSAPI;
    import GSSAPI qw{:all};
	require GSSAPI::OID;
	import GSSAPI::OID qw{:all};
    my $status = $gss_server_name->import($gss_server_name,$server_name, GSSAPI::OID::gss_nt_hostbased_service);
    $status || gss_exit("CLIENT::Unable to import server name: $server_name", $status);

    # The main purpose of GSSAPI::Context::init is to produce
    # an authentication token ($gss_output_token) which will
    # be sent by the app client to app server. Note the output
    # is binary data.
    $self->{gss_context}= $ctx;
    TRY:
    {
        do {
            $status = $ctx->init(GSS_C_NO_CREDENTIAL,
                            $gss_server_name,          # authenticate to this name
                            GSS_C_NO_OID,              # use default mechanism (krb5)
                            $gss_auth_flags,           # input flags
                            0,                         # input time
                            GSS_C_NO_CHANNEL_BINDINGS, # no channel binding
                            $gss_input_token,          # input token
                            undef,
                            $gss_output_token,
                            $out_flags,
                            $out_time);
                            
            $status || gss_exit("CLIENT::Unable to initialize security context", $status);
            $token = $gss_output_token;
            $self->{gss_context}= $ctx;
            $self->auth_gss_set(context => $handle, service => $service, gss_proc => $proc, qop => 0); 
            $proc = RPCSEC_GSS_CONTINUE_INIT;
            
            # xdr encoding of the token.....
            $token_len = length $token;
            $xdr->opaque(encode=>$token, variable=>$token_len);
            
            # Send the token to the server
			eval {
            	$init_res = $self->clnt_call(proc => 0, arg => $xdr->{message});
			};
			if ($@) {
				croak(qq{Seeing the symptoms of 476834. Please ensure the user who is running this test exists on the filer as a valid user. Otherwise, atleast run this test as root. Error: $@}) if ($@ =~ /_sysread: Timeout waiting for data/);
				croak ("Unknown error in init_cred:$@");
			};
            if ($init_res->{gss_major}!=GSS_S_COMPLETE)
            {
                croak "Bad reply from server!!" . $init_res->{gss_major};
            }
            
            # get the handle back from the server.....
            $handle = $init_res->{handle};
                        # The GSS protocol can do mutual authentication. If this is requested, the token
                # that we generate in the first pass will indicate this to the server. The major
                # status will have the GSS_S_CONTINUE_NEEDED bit set to indicate that we are
                # expecting a reply with a server identity token. This loop will continue until
                # that bit is no longer set. It should go through only once (non-mutual) or twice
                # (mutual).

                #The following code can be used when a mutual authetication from server is required
            #if ($status->major & GSS_S_CONTINUE_NEEDED) {
                #print "CLIENT::Mutual auth requested ...\n";
                #if ($gss_input_token) {
                    #print "CLIENT::got mutual auth token from server\n";
                #} else {
                    #print "CLIENT::server did not send needed continue token back\n";
                    #$error = 1;
                #}
            #}

        } while ($status->major == GSS_S_CONTINUE_NEEDED);

    } unless($status->major == GSS_S_COMPLETE  ) {
        croak (qq{\nGSS errors: $status});
    }
    $proc = RPCSEC_GSS_DATA;
    $self->auth_gss_set(context => $handle, service => $service, gss_proc => $proc, qop => 0);
    return 0;
}

##
## gss_exit
## Internal function for Rpc.pm, used to come out of the GSS context.
## If there are errors in establishing the GSS context, this function prints the relevant error codes.
## 
sub gss_exit {
    my $errmsg = shift;
    my $status = shift;

    my @major_errors = $status->generic_message();
    my @minor_errors = $status->specific_message();

    use Sys::Hostname;
    my $hostname = hostname;
    $Log->comment("Client was launched from the host:$hostname. $errmsg:\n");
    foreach my $s (@major_errors) {
        $Log->comment(qq{GSS error: MAJOR::$s.});
    }
    foreach my $s (@minor_errors) {
        $Log->comment(qq{GSS error: Minor::$s. "});
    }
    if ($#major_errors != -1 || $#minor_errors != -1) {
        croak ("#"x80 . qq{\nThere were GSSAPI errors. Please search this same logfile for "GSS error" somewhere above.}
		.  qq{Once you know the error, please see this link for explanations and possible solutions: http://wikid.netapp.com/w/NOSBU_QA/SVL-NFSQA/Kerberos_scripts_Issues \n}
		. "#"x80 . "\n");
    }
    exit(1);
}

#---------------------------------------------------------------------

=head2 auth_gss_set

  $out = $rpc_obj->auth_gss_set(context	=> $context,
  								service	=> $service,
								qop		=> $qop,
								gss_proc=> $gss_proc,
							);

Sets Kerberos authentication for the RPC connection.

The following named arguments are accepted:

=over

=item Arguments

=over

=item context

(Required) : Kerberos context to use.

=item service

(Required) : Which Kerberos service to use, valid values include
RPC_GSS_SVC_NONE(krb5), RPC_GSS_SVC_INTEGRITY(krb5i), RPC_GSS_SVC_PRIVACY(krb5p).

=item gss_proc

(Required) : The actual GSS work that needs to be done. The valid values are RPCSEC_GSS_DATA,
RPCSEC_GSS_INIT, and RPCSEC_GSS_CONTINUE_INIT.

=item qop

(Optional) : Quality of Protection. As of now, only 0 is valid for this. We may support 1 and 2 later.
These are again matching the services mentioned above - auth only, auth+integrety and auth+protection.

=back

=item Returns

1 if it succeeds.

=back

=for html <hr />

=cut
sub auth_gss_set($%)
{
    my $self = shift;
    my %opts = validate(@_,
			{
				context => {optional => 0},
                service => {optional => 0, regex => qr{^[123]$}}, 
                qop 	=> {optional => 0, regex => qr{^0$}}, 
                gss_proc => {optional => 0, regex => qr{^[0123]$}}
                });

    $self->{auth_gss} = {
            context    => $opts{context},
            service    => $opts{service},
            qop        => $opts{qop},
            gss_proc    => $opts{gss_proc}
        };
    $self->{auth_type}=RPCSEC_GSS;
    return 1;
}

####################################################################
## Internal function: _xdr_auth_gss
## encode gss cred
## This function encodes the given context, service qop and gss_proc
## into XDR format.
####################################################################
sub _xdr_auth_gss($%)
{
    my $self = shift;
    my %opts = validate(@_,{context => {optional=> 0},
                            service =>{ optional => 0}, 
                            qop => {optional => 0}, 
                            gss_proc => {optional => 0},
                            });

    my $seq_num = $self->{gss_seq_num};
    if($opts{gss_proc}==RPCSEC_GSS_INIT)
    {
        $seq_num=0;
        $self->{gss_seq_num} = 0;
        $opts{context}='';
    } else {
        $seq_num = $self->{gss_seq_num} + 1;
        $self->{gss_seq_num} += 1;
    }
    my $xdr = nfs::Xdr->new();
    my $rpcsec_ver = 1;
    my $handle_len;
    $xdr->uint(encode=>$rpcsec_ver);  # RPCSEC_GSS_VERS_1
    $xdr->uint(encode=>$opts{gss_proc});
    $xdr->uint(encode=>$seq_num);
    $xdr->uint(encode=>$opts{service});
    $handle_len = length($opts{context});
    $xdr->opaque(encode=>$opts{context}, variable =>$handle_len);
    return $xdr->{message};
}

1;

