# -*- perl -*-
#
## @summary
## Single file snap restore is the ability to revert a file in the active
## file system to the state contained in the snapshot.  After a file has
## been snap-restored, all user-visible information (data and file attributes)
## for that file in the active file system are identical to the state
## contained in the snapshot.
## All these testcases are for C-mode setup. Works for real clients, NFSV4 only
##
## @param   FILER        Required: hostspec of filer1.
##
## @param   VSERVER_GX_NAME      Required: The vserver to use.
##
## @author              rraghav@netapp.com
##
## @Maintainer dl-svl-nosauto@netapp.com
##
##

use strict;
use Carp qw{croak};
use nfs::RemoteFileIO qw(client_read_file);
use nfs::Delegations qw{file_lock_send};
use FileManip qw(write_file);
use nfs_setup_gx;

use NACL::C::Node;
use NACL::C::Client;
use NACL::C::VolumeSnapshot;
use NACL::STask::Client::Mount;
use NACL::STask::Qtree;
use NACL::STask::VolumeSnapshot;

use Params::Validate qw{validate SCALAR};
use nfs::NfsTCD qw{cleanup init uninit main $Node_cmd $Mountpoint @UNINIT @CLEANUP $Client_cmd %Setup_hash};

#--------------------------------------------------------------------------------
# Initialize and validate parameters
#--------------------------------------------------------------------------------
use vars qw($DD);
param('DD',          '-default', '/bin/dd');

our (%Setup_hash, $Node_cmd, $Client_cmd, $Mountpoint, $subid);
my ($snapshot_obj, $snapshot_name);

sub create_file($) {
    my $filename = shift;
    $Log->comment("Creating the test file, and setting all-permissions on it");
    $Client_cmd->execute(command => "$DD bs=1000 count=2000 if=/dev/urandom of=$Mountpoint/$filename");
    $Client_cmd->execute(command => "chmod 777 $Mountpoint/$filename");

    return "$Mountpoint/$filename";
}

sub perform_file_restore(%) {
    my %opts = validate (@_,
        {
            origin_file => { type => SCALAR },
            destination_file => { type => SCALAR },
            origin_volume => { type => SCALAR, default => $Setup_hash{vol_name} },
            destination_volume => { type => SCALAR => default => $Setup_hash{vol_name} },
            partial => { type => SCALAR, optional => 1, regex => qr/^[01]$/, },
            start => { type => SCALAR, optional => 1, regex => qr/^[\d]+$/, },
            bytes => { type => SCALAR, optional => 1, regex => qr/^[\d]+$/, },
        });

    my ($string, $function, %my_opts) = ();
    if ($opts{partial}) {
        $string = "partial";
        $function = "partial_restore_file";
        if ($opts{start} && $opts{bytes}) {
            $my_opts{'start-byte'} = $opts{start};
            $my_opts{'byte-count'} = $opts{bytes};
        } else {
            croak (qq{for partial restore, start & bytes are mandatory arguments, this is a programming error.});
        }
    } else {
        $string = "full";
        $function = "restore_file";
    }
    $Log->comment("Performing $string singlefile snapshot restore on the file");
    NACL::C::VolumeSnapshot->$function(command_interface=>$Node_cmd,
                                           vserver  => $::VSERVER_GX_NAME,
                                           snapshot => $snapshot_name,
                                           volume  => $Setup_hash{vol_name},
                                           path => "/vol/$Setup_hash{vol_name}/$opts{origin_file}",
                                           'restore-path' => "/vol/$opts{destination_volume}/$opts{destination_file}",
                                           %my_opts,
                                         );
}

sub create_snapshot($) {
    $snapshot_name = shift;
    $Log->comment(qq{Creating the snapshot:$snapshot_name});
    $snapshot_obj = NACL::STask::VolumeSnapshot->create(command_interface=>$Node_cmd,
                                    vserver    => $::VSERVER_GX_NAME,
                                    nacltask_if_exists => 'purge',
                                    volume     => $Setup_hash{vol_name},
                                    snapshot   => $snapshot_name,
                                    foreground => 'true');
    NACL::C::VolumeSnapshot->find(command_interface=>$Node_cmd,
                            filter => { vserver  => $::VSERVER_GX_NAME,
                                        volume   => $Setup_hash{vol_name},
                                        snapshot => $snapshot_name});
}

# nfs4_getfacl:
#20130612 144958 [01_stest/01_SFSR] <<proto-howler.rtp.netapp.com#login_0|A:fd:testu2@netapp.com:rw\r\n
#20130612 144958 [01_stest/01_SFSR] <<proto-howler.rtp.netapp.com#login_0|A::OWNER@:rwaxtTnNcCy\r\n
#20130612 144958 [01_stest/01_SFSR] <<proto-howler.rtp.netapp.com#login_0|A:g:GROUP@:rwaxtTnNcy\r\n
#20130612 144958 [01_stest/01_SFSR] <<proto-howler.rtp.netapp.com#login_0|A::EVERYONE@:rwaxtTnNcy\r\n

#getfacl:

sub get_file_acl_stuff($) {
    my $file = shift;
    my %output;

    my $out_cmd = client_execute("getfacl $file");
    $out_cmd =~ /\#\s+owner\:\s+(\S+)/;
    $output{owner} = $1;
    $out_cmd =~ /\#\s+group\:\s+(\S+)/;
    $output{group} = $1;
    $out_cmd =~ /user::(\S+)/;
    $output{user_permission} = $1;
    $out_cmd =~ /group::(\S+)/;
    $output{group_permission} = $1;
    $out_cmd =~ /mask::(\S+)/;
    $output{mask} = $1;
    $out_cmd =~ /other::(\S+)/;
    $output{other_permission} = $1;

    my $get_acl_out = client_execute("nfs4_getfacl $file");
    $get_acl_out =~ /.+OWNER\@:(\S+)/;
    $output{owner_acl} = $1;
    $get_acl_out =~ /.+GROUP\@:(\S+)/;
    $output{group_acl} = $1;
    $get_acl_out =~ /.+EVERYONE\@:(\S+)/;
    $output{everyone_acl} = $1;
    $Log->comment("Attributes of the file:\n" . Dumper(%output) );

    return \%output;
}

sub compare_entries($$) {
    my ($first_hash, $second_hash) = @_;

    my @entries = qw{owner
                     group
                     user_permission
                     group_permission
                     mask
                     other_permission
                     owner_acl
                     group_acl
                     everyone_acl};
    foreach my $entry (@entries) {
        if ($second_hash->{$entry} ne $first_hash->{$entry}) {
            croak (qq{Attributes are not same: $entry: values: $first_hash->{$entry}, $second_hash->{$entry}});
        }
    }
}

sub cleanup() {
    $Log->comment(qq{SFSR cleanup});
    if ($subid) {
        stop_io(sub_id=>$subid);
    }
    #Delete the snapshot created
    if ($snapshot_obj) {
        $snapshot_obj->purge();
    }
    undef $snapshot_obj;

    nfs::NfsTCD::cleanup();
}

sub main() {
    nfs::NfsTCD::main();
}

1;
