######################################################################
## nfs/SpinnpReplay.pm: All the filer-side spinnp replay related stuff.
######################################################################
package nfs::SpinnpReplay;

use strict;
use warnings;

use Data::Dumper;
use Carp qw{croak};
use Params::Validate qw(validate SCALAR OBJECT ARRAYREF);
use Tharn qw{$Log};

use NACL::C::SystemNode;
use NACL::C::Node;

use base qw(Exporter);
our @EXPORT = qw{
    check_for_dsid
    check_in_prog_ops
    check_replayed_ops
    establish_cmdintf
    get_and_parse_replay_details
    get_bins
    get_buckets
    get_bucket_details
    get_details_bin
    get_details_last_bin
    get_last_bin
    get_lru_number
    get_lrud_buckets
    get_spinnp_replay_stats
    mem_stats
    parse_dump_bin_output
    parse_spinnp_replay_stats
    purge_cache
    set_default_spinnp_node
    run_dump_bin
    run_dump_bins
    turn_verbose_on
    verify_empty_cache
    verify_full_bin
    verify_no_cache_for_volume_dsid
    verify_same_cache
};

our @EXPORT_OK = @EXPORT;
my $cmd_intf = NACL::C::Node->find();

sub establish_cmdintf() {
    $Log->comment(qq{SpinnpReplay:Establishing a new connection to the filer});
    my %opts = (
        no_version_change      => 1,
        max_reconnect          => 10,
        max_reconnect_timewait => 60
    );
    $cmd_intf->refresh_command_interface(%opts);

}
establish_cmdintf();
my $spinnp_node_obj;

# All spinnp_replay related functions below.

sub set_default_spinnp_node() {
    my $node = NACL::CS::Volume->fetch(command_interface=>$cmd_intf,
        filter => {volume => $::FILER_GX_VOL, vserver => $::VSERVER_GX_NAME}, 'requested_fields'=>['node'])->{node};
    $spinnp_node_obj = NACL::C::SystemNode->find(command_interface=>$cmd_intf, filter => {node => $node});
}

sub turn_verbose_on(;%) {

    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay verbose on"});
}

sub purge_cache(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay purge_cache"});
    verify_empty_cache(%opts);
}

sub verify_empty_cache(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $out = run_dump_bins(%opts);
    if ($out =~ "Bin:") {
        croak (qq{Bins is not empty, please check spurious traffic from some other source to the filer});
    }
}

sub run_dump_bins(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });

    return $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay dump_bins"});
}

sub run_dump_bin(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR },
                    });
    return $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay dump_bin $opts{binid}"});
}

# Returns a list of bin ids in the filer.
sub get_bins(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $output = run_dump_bins(cmd_intf=>$opts{cmd_intf});
    my @ret;
    foreach my $line (split(/\n/, $output)) {
        if ($line =~ /Bin: (.*)/) {
            push @ret, $1;
        }
    }
    return @ret;
}

sub get_and_parse_replay_details(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR },
                    });

    my $output = run_dump_bin(%opts);
    return parse_dump_bin_output(output=>$output);
}


# 1. the number of buckets in the given bin is now maximum.
# 2. The LRU count is non zero.
sub verify_full_bin(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR, optional => 1 },
                    });

    $opts{binid} = $opts{binid} ? $opts{binid} : get_last_bin();
    my $buckets = get_buckets(cmd_intf=>$opts{cmd_intf}, binid => $opts{binid});
    my @buckets = @$buckets;
    $Log->comment(qq{buckets = } . Dumper($buckets));
    if ($#buckets+1 < $::MAX_BUCKETS) {
        croak (qq{The max buckets has not been reached. The number of buckets:} . ($#buckets+1) . qq{, expected $::MAX_BUCKETS.});
    } else {
        $Log->comment(qq{The number of buckets has hit the max limit, pass});
    }
    my $replay_stats = get_and_parse_replay_details(cmd_intf=>$opts{cmd_intf}, binid=>$opts{binid});
    $Log->debug(qq{replay_stats = } . Dumper($replay_stats));
    if ($replay_stats->{'LRU Buckets'} == 0) {
        croak (qq{LRU length is zero, some problem});
    }
}

# Get the list of buckets for the given bin id.
sub get_buckets(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR, optional => 1 },
                    });

    if (! $opts{binid}) {
        $opts{binid} = get_last_bin(cmd_intf=>$opts{cmd_intf});
    }
    if (! $opts{binid}) {
        croak (qq{Hmm, we are expecting a bin id, but seems there is no bin in the filer output :(});
    }

    my $output = run_dump_bin(cmd_intf=>$opts{cmd_intf}, binid=>$opts{binid});
    return (parse_dump_bin_output(output=>$output))->{buckets};
}

sub parse_dump_bin_output(%) {

    my (%opts) = validate (
                    @_,
                    {
                        output => { type => SCALAR }
                    });
    my $ret;
    @{$ret->{buckets}} = ();
    foreach my $line (split(/\n/, $opts{output})) {
        if ($line =~ /LRU Buckets: (\d+)/) {
            $Log->debug("matched LRU Buckets");
            $ret->{'LRU Buckets'} = $1;
        } elsif ($line =~ /Label: (.+)/) {
            $Log->debug("matched Label");
            $ret->{'Label'} = $1;
        } elsif ($line =~ /Flags: (.+)/) {
            $Log->debug("matched Flags");
            $ret->{'Flags'} = $1;
        } elsif ($line =~ /Type: (.+)/) {
            $Log->debug("matched Type");
            $ret->{'Type'} = $1;
        } elsif ($line =~ /Policy: (.+)/) {
            $Log->debug("matched Policy");
            $ret->{'Policy'} = $1;
        } elsif ($line =~ /Bucket tag: (.*)/) {
            $Log->debug("matched Buckets");
            push @{$ret->{buckets}}, $1;
        }
    }
    return $ret;
}

#rraghav-vsim2*> spinnp_replay dump_bucket 0x501390730000001a 0x1bd09d47001c0001
#Bucket: 0x0
#        Tag: 0x1bd09d47001c0001
#        Handle: 0xffffffff
#        Data: 0xffffff0167315578
#        Data Length: 188
#        DSID: 1061
#        AGGR UUID: d741a9c8-d81c-11e1-a90e-123478563412
#        LRU time: 1343459461
#rraghav-vsim2*>
sub get_bucket_details(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR },
                        bucketid    => { type => SCALAR },
                    });
    my $output = run_dump_bucket(%opts);

    my $ret;
    foreach my $line (split(/\n/, $output)) {
        if ($line =~ /\s+(.+):\s*(.+)/) {
            $ret->{$1} = $2;
            chomp($ret->{$1});
        }
    }
    return $ret;
}

sub run_dump_bucket(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR },
                        bucketid    => { type => SCALAR },
                    });
    my $output = $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay dump_bucket $opts{binid} $opts{bucketid}"});
    return $output;
}

# Get the last bin id, dump it, find each file handle and dsid. and return a data structure with this data.
sub get_details_last_bin(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $binid = get_last_bin(%opts);
    return get_details_bin(cmd_intf=>$opts{cmd_intf}, binid=>$opts{binid});
}

sub get_details_bin(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR },
                    });
    my @buckets = get_buckets(%opts);
    my $ret;

    foreach my $bucket (@buckets) {
        $ret->{$bucket} = get_bucket_details(%opts, bucketid => $bucket);
    }
    return $ret;
}

sub get_last_bin(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    return (get_bins(%opts))[-1];
}

# Check if the given dsid is present in the given get_details_last_bin() output.
sub check_for_dsid(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        dsid     => { type => SCALAR,  optional => 0 },
                    });

    $Log->comment(qq{check_for_dsid: looking for $opts{dsid}});

    my @bins = get_bins(cmd_intf=>$opts{cmd_intf});

    foreach my $bin (@bins) {
        my $buckets = get_buckets( cmd_intf=>$cmd_intf, binid => $bin);
        foreach my $bucket (@$buckets) {
            my $bucket_detail = get_bucket_details( cmd_intf=>$cmd_intf, bucketid => $bucket, binid => $bin);
            $Log->comment(qq{bucket dsid:"} . $bucket_detail->{DSID} . qq{"});
            if ($bucket_detail->{DSID} eq $opts{dsid}) {
                return 1;
            }
        }
    }
    croak (qq{No matching dsid found in the replay cache output.});
}

sub verify_no_cache_for_volume_dsid(%) {
    eval {
        check_for_dsid(@_);
    };
    if ($@) {
        if ($@ =~ /No matching dsid/) {
            $Log->comment(qq{There is no matching volume dsid, as expected});
        } else {
            croak ($@);
        }
    } else {
        croak (qq{Found an unexpected matching volume dsid, error});
    }
}

sub get_spinnp_replay_stats(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $output = $opts{cmd_intf}->run(command=>qq{"priv set test; spinnp_replay_stats"});
    return (parse_spinnp_replay_stats(output=>$output));
}

sub parse_spinnp_replay_stats(%) {
    my (%opts) = validate (
                    @_,
                    {
                        output => { type => SCALAR },
                    });
    my $ret;

    foreach my $line (split(/[\r\n]/, $opts{output})) {
        chomp($line);
        if ($line =~ /(.+):\s*([^\s]+)/) {
            $ret->{$1} = $2;
        }
    }

    return $ret;
}

sub check_in_prog_ops(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $reply = get_spinnp_replay_stats(%opts);
    return ($reply->{'In progress buckets found'});
}

sub check_replayed_ops(;%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                    });
    my $reply = get_spinnp_replay_stats(%opts);
    return ($reply->{'Replayed ops'});
}

sub verify_same_cache($$) {
    my ($obj1, $obj2) = @_;
    my @arr1 = sort(@$obj1);
    my @arr2 = sort(@$obj2);
    if (@arr1 eq @arr2) {
        return 1;
    }
    croak (qq{The 2 lists are different(please check if there is some other traffic in the cluster):} . Dumper($obj1) . " and " . Dumper($obj2));
}

sub get_lru_number() {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR, optional => 1 },
                    });

    if (! $opts{binid}) {
        $opts{binid} = get_last_bin(cmd_intf=>$opts{cmd_intf});
    }

    my $output = run_dump_bin(cmd_intf=>$opts{cmd_intf}, binid=>$opts{binid});
    return (parse_dump_bin_output(output=>$output))->{'LRU Buckets'};
}    

# What are the buckets that were lru'd out of the given list of buckets?
sub get_lrud_buckets(%) {
    my (%opts) = validate (
                    @_,
                    {
                        cmd_intf    => { type => OBJECT,  optional => 1, default => $spinnp_node_obj },
                        binid       => { type => SCALAR, optional => 1 },
                        buckets     => { type => ARRAYREF },
                    });

    if (! $opts{binid}) {
        $opts{binid} = get_last_bin(cmd_intf=>$opts{cmd_intf});
    }

    my $output = run_dump_bin(cmd_intf=>$opts{cmd_intf}, binid=>$opts{binid});
    my @old_buckets = @{$opts{buckets}};
    my @current_buckets = (parse_dump_bin_output(output=>$output))->{buckets};

    my $string = " @current_buckets ";
    my @ret;
    foreach (@old_buckets) {
        if (" $_ " !~ / $string /) {
            push @ret, $_;
        }
    }
    return @ret;
}

# Same as nfs/reply_cache/OctetReplyCache.pm, sub mem_stats, for c-mode.
sub mem_stats(%) {
    my (%opts) = validate (
        @_,
        {
            cmd_intf   => { type => OBJECT, default => $spinnp_node_obj },
            mode => { type => SCALAR, default => 'get',
                                        regex => qr/^(get|clean|diff|save)$/ },
        }
    );
    my (@ret, $line, $allocs, $bytes, $addr, $out);
    # Table to map mode into its corresponding command flag
    my %flag = (
        get   => '',
        clean => '-c',
        save  => '-b',
        diff  => '-e',
    );

    # Send the mem_stats command to the filer
    $out = $opts{cmd_intf}->run(command=>qq{"priv set -q diag; mem_stats $flag{$opts{mode}}"});

    # Process output of command
    foreach $line ( split /\n/, $out ) {
        ($allocs, $bytes, $addr) = ($line =~ /^\s*(\d+)\s+(\d+)\s+(\S+)/);
        if ($allocs) {
            push @ret, {
                allocs => $allocs,
                bytes  => $bytes,
                addr   => $addr,
            };
        }
    }

    return \@ret;

}

1;
