#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Subnets.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::Subnets;

## @summary Library for subnet validation
##
## @description
##   Subnet validation functions
##
## @usage
##   use nfs::Subnets;
##
## @status      public
##
## @author  dl-maryv (FSS Protocols QA)
## @notes       Created by EP Network Storage Performance Lab, Inc.

use strict;
use Socket;
use Module::Common qw(validate_mixed);
use Params::Validate qw( validate_pos SCALAR );

use vars qw(
  $VERSION
  @ISA
  @EXPORT
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  &fix_subnets
  &validate_subnet
);

#---------------------------------------------------------------------
# Package global variables
#---------------------------------------------------------------------
$VERSION      = '0.01';

#---------------------------------------------------------------------
## @name ::validate_subnet
## @description
##  Validate subnet of the form 'xxx.xxx.xxx.xxx/nn'
##
## @synopsis
##  $new_subnet|@subnet_l = validate_subnet($subnet);
##
## @arg $subnet
##  Required: Subnet specification in any form:
##            192.168.1.0/24,
##            192.168.1/24,
##            "192.168.1.0 255.255.255.0",
##            "network 192.168.1.0 netmask 255.255.255.0".
##
## @return
## Scalar context:
##   Subnet of the form "192.168.1.0/24".
##
## List context:
##   ($subnet, $subnet_mask, $subnetno, $subnet_len)
##   $subnet		(e.g. 192.168.1.0)
##   $subnet_mask	(e.g. 255.255.255.0)
##   $subnetno		(e.g. 192.168.1)
##   $subnet_len	(e.g. 24)
##
## @example
##   $subnet = validate_subnet("192.168.1.0 255.255.255.0");
##   ($sub, $sub_mask, $sub_no, $sub_len) = validate_subnet("192.168.1/24");
##   (@sub) = validate_subnet("192.168.1.0/24");
#
#  Note: this function uses a local copy of '$_' for ease of use since
#  a lot of pattern matching is required.
#---------------------------------------------------------------------
sub validate_subnet($;) {
    local ($_) = validate_pos(@_,
        { type => SCALAR }
    );
    my $negated = '';

    # Save negated status
    $negated = '-' if /^-/;
    s/^-//;
    # Get rid off quotes
    s/"//g;

    # Determine type of subnet and change it into
    # the form "xxx.xxx.xxx.xxx/nn"
    if ( m#/# ) {	# Subnet type: "xxx.xxx.xxx.xxx./nn"
        # Add missing octets
        foreach my $n (tr/././..2) {
            # Append '.0' to subnet part
            substr($_, index($_, '/'), 0) = '.0';
        }
    } else {
        # Subnet type: "[network x.x.x.x [netmask] y.y.y.y"
        s/(network|netmask)//g;	# Get rid off words
        my ($sub, $mask) = split;
        return unless $mask;	# Not a valid subnet
        # Convert netmask to binary form "1111...10000"
        $_ = unpack('B32', inet_aton($mask));
        # Append subnet bits
        $_ = "$sub/" . tr/1/1/;
    }

    # At this point '$_' has a subnet in the form "xxx.xxx.xxx.xxx/nn"
    my ($subnet, $subnet_len) = split('/');

    my $host_len = (32 - $subnet_len);    # Number of host bits

    # Convert subnet bits to binary representation
    my $subnet_bin = pack('B32', '1' x $subnet_len . '0' x $host_len);

    # Apply subnet mask to subnet
    my $subnet_addr = inet_aton($subnet) & $subnet_bin;

    # Validate subnet
    if ($subnet_len > 32  or  $subnet ne inet_ntoa($subnet_addr)) {
        return;  # Empty list
    }

    if ( wantarray ) {
        # List context
        my $subnet_mask = inet_ntoa($subnet_bin);

        # Keep subnet part only
        my $subnetno =
           join('.', (split(/\./, $subnet))[0..(3-int($host_len/8))]);

        ($subnet, $subnet_mask, $subnetno, $subnet_len);
    } else {
        # Scalar context
        "$negated$_";
    }
}

#---------------------------------------------------------------------
## @name ::fix_subnets
## @description
##  Validate all subnets in access list.
##
## @synopsis
##  $new_access_list = fix_subnets($acc_list [, %opts]);
##
## @arg $acc_list
##  Required: Access list to be validated.
##
## @arg prefix
##  Optional: Prefix to add to subnet definitions.
##
## @arg postfix
##  Optional: Postfix to add to subnet definitions.
##
## @return
##  Modified access list, all subnets are validated
##  and modified to the 'xxx.xxx.xxx.xxx/nn' format.
##
## @example
##   $access_list = 'ro=10.0/16:client:"192.168.2.0 255.255.255.0":client2'
##   $new_access_list = fix_subnets($access_list);
##
##     $new_access_list is 'ro=10.0.0.0/16:client:192.168.2.0/24:client2'
##
##   $access = fix_subnets("ro=912.168/16", prefix => '"', postfix => '"'
##   $access is 'ro="192.168.0.0/16"'
#---------------------------------------------------------------------
sub fix_subnets($;%) {
    my ($access, %opts) = validate_mixed (
        @_,
        [ { type => SCALAR } ],
        {
            prefix  => { type => SCALAR, optional => 1 },
            postfix => { type => SCALAR, optional => 1 },
        }
    );
    my (@hlist, $acc, $new_access, $neg, $op);
    my $pre  = $opts{prefix};
    my $post = $opts{postfix};

    # Split options
    foreach (split /,/, $access) {
        ($acc, $_) = split /=/;
        # Validate subnets, all others are unmodified
        @hlist = map {  ($neg, $_) = /^(-?)(.*)/;
                        if ( $op = validate_subnet($_) ) {
                            "$neg$pre$op$post";
                        } else {
                            "$neg$_";
                        }
                     } split /:/;
        # Reconstruct access options
        $new_access .= ",$acc";
        $new_access .= "=" . join(':', @hlist) if @hlist;
    }
    # Clear leading ','
    $new_access =~ s/^,//;

    return $new_access;
}

1;

