# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Timer.pm#1 $
#
## @summary     Timer - part of the Flg library
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Timer - library functions specifically for use by the Flg library

=head1 SYNOPSIS

  use nfs::Timer;

=head1 DESCRIPTION

This is a Perl port of the flg filesystem load generator originally
written in C. Functionality is taken directly from the C source.

This package is intended for use by the nfs::Flg.pm library only. Do not
call nfs::Timer.pm directly.

Original code is inserted as comments for reference where applicable.
Triple X comments in the original code have been changed to XXO (oh, not zero).

The porting to Perl is based solely on functionality, without an
examination of the procedure used for the flg load generator.
An attempt was made if at all possible to keep the structure and
form of the original C code. Modifications were needed to implement
the C struct as a Perl hash, and several code changes were necessary
to implement multiple NFS version support. Otherwise, the code is
intended to be as close as possible to the original C form.
Parameters to functions are taken from the C code, and may not be
well explained in the Perl module. Please refer to the original C code
for functionality and procedural clarification.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Timer.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Timer;

use strict;
use warnings;
use base qw(Exporter);

use Time::HiRes qw(gettimeofday tv_interval);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#--------------------------------------------------------------------------------
# Version for package
#--------------------------------------------------------------------------------
our $VERSION = '1.0';

#------------------------------------------------------------------------
# Package global variables
#------------------------------------------------------------------------
our @EXPORT_OK = qw(
  global_timer
);

my $ready = 0;
my $tv;
my $last_timer = 0;

#--------------------------------------------------------------------------------
# Subroutine prototypes
#--------------------------------------------------------------------------------
sub global_timer();
sub timer_start();
sub timer_read();
sub testme();

#--------------------------------------------------------------------------------

=head2 global_timer

  $time = global_timer();

global_timer function taken from original C code. Returns the elapsed
time since the global timer began.

Arguments are not accepted:

=over

=item Returns

Total elapsed time since timer_start

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub global_timer() {
    #  u_int64_t
    #  global_timer(void)
    #  {
    #          static int ready = 0;
    #          static struct timeval tv;
    #          u_int64_t ret;
    #          static u_int64_t last_timer;
    #
    #  #ifdef USE_CYCLE_COUNTER
    #          static u_int32_t last_cc, cycles_per_usec = 0;
    #          u_int32_t this_cc = cc_read();
    #
    #          if (cycles_per_usec == 0) {
    #                  cycles_per_usec = cc_freq();
    #          } else {
    #                  if ((this_cc > last_cc) &&
    #                      (this_cc - last_cc) < (cycles_per_usec * 1000)) {
    #                          return last_timer +
    #                                  ((this_cc - last_cc) / cycles_per_usec);
    #                  }
    #          }
    #  #endif
    #
    #          if (ready == 0) {
    #                  if (timer_start(&tv) < 0) {
    #                          report_error(FATAL, "timer_start error");
    #                          return 0;
    #                  }
    #                  ready = 1;
    #          }
    #
    #          ret = timer_read(&tv);
    #          if (ret < last_timer) {
    #                  ret = last_timer;
    #          }
    #          last_timer = ret;
    #
    #  #ifdef USE_CYCLE_COUNTER
    #          last_cc = cc_read();
    #  #endif
    #          return ret;
    #  }
    #

    if ( $ready == 0 ) {
        timer_start();
        $ready = 1;
    }

    my $ret = timer_read();
    if ( $ret < $last_timer ) {
        $ret = $last_timer;
    }
    $last_timer = $ret;

    return $ret;

} # End global_timer
#--------------------------------------------------------------------------------

=head2 timer_start

  timer_start();

timer_start function taken from the original C code. Initializes the
timing functionality for Flg.

Arguments are not accepted:

=over

=item Returns

Nothing

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub timer_start() {
    #  static int
    #  timer_start(struct timeval *timer)
    #  {
    #          if (gettimeofday(timer, NULL) < 0) {
    #                  report_perror(FATAL, "gettimeofday");
    #                  return -1;
    #          }
    #          return 0;
    #  }
    #

    $tv = [gettimeofday];

} # End timer_start

#--------------------------------------------------------------------------------

=head2 timer_read

  $etime = timer_read();

timer_read function taken from original C code. Calculates the elapsed
time since the timer was started.

Arguments are not accepted:

=over

=item Returns

Elapsed time since global timer was started.

=back

=for html <hr />

=cut

#--------------------------------------------------------------------------------

sub timer_read() {
    #  static u_int64_t
    #  timer_read(struct timeval *timer)
    #  {
    #          struct timeval now, delta;
    #
    #          if (gettimeofday(&now, NULL) < 0) {
    #                  report_perror(FATAL, "gettimeofday");
    #                  return -1;
    #          }
    #          delta.tv_sec = now.tv_sec - timer->tv_sec;
    #          delta.tv_usec = now.tv_usec - timer->tv_usec;
    #          if (delta.tv_usec < 0) {
    #                  delta.tv_sec -= 1;
    #                  delta.tv_usec += 1000000;
    #          }
    #          return (((u_int64_t)delta.tv_sec * (u_int64_t)1000000) +
    #                  (u_int64_t)delta.tv_usec);
    #  }

    # Get the number of elapsed seconds since start
    my $delta = tv_interval( $tv, [gettimeofday] );

    # Convert to microseconds and return
    return $delta * 1000000;

} # End timer_read

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Timer SUBNAME=testme util/test_pkg.thpl

=over

=item Arguments

None

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    # Start the timer
    my $time = global_timer();

    # Sleep a while
    sleep 2;

    # Check the timer
    $time = global_timer();

    # Validate timer returned between 2 and 3 seconds
    if ( int($time/1000000) < 2 || int($time/1000000) > 3 ) {
        Tharn::logresult('FAIL', 'Timer not operating correctly');
    } else {
        Tharn::logresult('PASS', "SELF TEST of nfs/Timer.pm");
    }
}

1;
