#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Utils.pm#2 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.

package nfs::Utils;

## @summary Library of basic NFS utilities
##
## @description
##  Basic utilities for NFS tests. Includes client validation
##  functions for proper NFS testing.
##
## @usage
##  use nfs::Utils;
##
## @status	public
##
## @author	dl-nfsqa-only (NFS QA Sunnyvale)
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;
use warnings;
use base qw(Exporter);

use Carp qw(croak);
use SayWrap qw(sayPerl sayEOF);
use Tharn qw(host connect param);
use SPARSE::Util::Logger qw(logDbg logStep);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT UNDEF);
use cluster qw(get_cf_state cf_giveback wait_state cf_enable);
use config_base qw(host_is_alive);
use boot qw(transit_to reboot);
use filer qw(filer_cluster_license filer_cluster_is_licensed);
use NACL::C::Client;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our @EXPORT_OK = qw(
  validate_nfsv4_clients
  setup_cluster
  validate_nfs_version
  range_ipaddrs
  ping_ipaddrs
  available_ipaddrs
  check_for_panic
  flush_reply_cache
  get_reply_cache
  get_reply_cache_entry
  get_reply_cache_flow
  get_client_addr
  get_good_nfs_version
  get_good_auth
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '0.01';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub validate_nfsv4_clients(%);
sub setup_cluster(;%);
sub validate_nfs_version(%);
sub range_ipaddrs(%);
sub ping_ipaddrs(%);
sub available_ipaddrs(%);
sub check_for_panic(%);
sub flush_reply_cache(%);
sub get_reply_cache(%);
sub get_reply_cache_entry(%);
sub get_reply_cache_flow(%);
sub get_good_nfs_version(;$);
sub _change_keys($;);

#---------------------------------------------------------------------
## @name ::validate_nfsv4_clients
## @description
##  Validate that a list of clients are NFS V4 hosts.
##  Arguments are passed in "key => value" style.
##  It will croak on the first client that does not
##  support NFS V4. Note that Solaris 10 or later has
##  NFS V4. AIX 5.3 or later has NFS V4. Linux 2.6.18 or
##  later is assumed to have NFS V4 capabilities although
##  that may not be the case depending on how the kernel
##  was built. All other OSes are assumed to not have NFS V4 support.
##
## @synopsis
##  $op = validate_nfsv4_clients( %options );
##
## @arg clients
##  Required: Reference to a list of client hostspecs.
##
## @arg deleg_clients
##  Optional: Reference to a list of client hostspecs which support
##  delegations. This list must be a subset of list in 'clients'.
##
## @arg stream_clients
##  Optional: Reference to a list of client hostspecs which support
##  named attributes. This list must be a subset of list in 'clients'.
##
## @arg pseudofs_clients
##  Optional: Reference to a list of client hostspecs which support
##  Pseudo-FS. This list must be a subset of list in 'clients'.
##
## @arg client_conn
##  Optional: Clients connection spec, (default => '#rsh').
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  (default is $CMDTIMEOUT for the connection).
##
## @example
##  $op = validate_nfsv4_clients( clients=>[qw(localhost solaris_1)] );
##
## @return
##  1 if all clients are NFS V4 hosts or else croaks on first client
##  that does not support NFS V4
#---------------------------------------------------------------------
sub validate_nfsv4_clients(%) {
    my (%opts) = validate (
        @_,
        {
            clients          => { type => ARRAYREF },
            deleg_clients    => { type => ARRAYREF|UNDEF, optional => 1 },
            stream_clients   => { type => ARRAYREF|UNDEF, optional => 1 },
            pseudofs_clients => { type => ARRAYREF|UNDEF, optional => 1 },
            client_conn      => { type => SCALAR, default => '#rsh' },
            timeout          => { type => SCALAR, default => 0 }
        }
    );
    my ($release, $version, $rel, $subrel, $os);
    my (%dclient_h, %sclient_h, %pclient_h, $client, $cmd, $cc, $out, $h);

    # Default options passed to sayPerl
    my @say_opts = qw(-swallow);

    # Add timeout to the list of options
    push @say_opts, '-timeout', $opts{timeout} if $opts{timeout};

    # Hash of NFSv4 clients to ease comparison of lists
    my %nfsv4_h = map { $_ => 1 } @{$opts{clients}};

    # Make sure all clients in 'deleg_clients' exists in 'clients'
    if ( defined($opts{deleg_clients}) ) {
        if ( grep { ! $nfsv4_h{$_} } @{$opts{deleg_clients}} ) {
            croak "Clients specified in option 'deleg_clients' " .
                  "must exist in the 'clients' list";
        }
        # Create hash of clients to validate delegations support
        %dclient_h = map { $_ => 1 } @{$opts{deleg_clients}};
    }

    # Make sure all clients in 'stream_clients' exists in 'clients'
    if ( defined($opts{stream_clients}) ) {
        if ( grep { ! $nfsv4_h{$_} } @{$opts{stream_clients}} ) {
            croak "Clients specified in option 'stream_clients' " .
                  "must exist in the 'clients' list";
        }
        # Create hash of clients to validate named attributes support
        %sclient_h = map { $_ => 1 } @{$opts{stream_clients}};
    }

    # Make sure all clients in 'pseudofs_clients' exists in 'clients'
    if ( defined($opts{pseudofs_clients}) ) {
        if ( grep { ! $nfsv4_h{$_} } @{$opts{pseudofs_clients}} ) {
            croak "Clients specified in option 'pseudofs_clients' " .
                  "must exist in the 'clients' list";
        }
        # Create hash of clients to validate named attributes support
        %pclient_h = map { $_ => 1 } @{$opts{pseudofs_clients}};
    }

    # Validate list of clients
    foreach $client ( @{$opts{clients}} )  {
        $h = host($client);

        if ( $client =~ /^localhost\b/ ) {
            # For localhost, no need to connect
            # but put output in same format as
            # the remote clients
            $out = join(q(--), POSIX::uname());
        } else {
            # Connect to remote client with default timeouts
            $cc = connect ( $opts{client_conn}, -timeout, $main::TIMEOUT );
            $cc->timeout( $main::CMDTIMEOUT );

            # Script to get version of OS on remote client
            $cmd = qq{
                use POSIX qw(uname);
                \@list = uname();
                print join(q(--), \@list);
            };
            # Execute script on remote client as the current user
            (undef, undef, $out) = sayPerl($cc, $cmd, "", @say_opts);
        }

        # Break up options of interest
        ($os, $release, $version) = (split(/--/, $out))[0,2,3];

        if ( $h->hosttype() eq "unix-solaris" ) {
            # Solaris requires 5.10 or later for NFS V4
            # Also, all Solaris NFSv4 clients support
            # delegations and named attributes
            ($rel, $subrel) = split(/\./, $release);
            if ( $rel < 5 or ($rel == 5 and $subrel < 10) ) {
                croak "Solaris $release does not support NFS V4. " .
                      "Must use 5.10 or later";
            }
            # Solaris 5.10 does not support Pseudo-FS
            if ( $pclient_h{$client} ) {
                # As of now, Solaris does not support pseudofs
                croak "Burt 216041: Solaris does not support Pseudo-FS";
            }
        } elsif ( $h->hosttype() eq "unix-aix" ) {
            # AIX requires 5.3 or later for NFS V4
            if ( $version < 5 or ($version == 5 and $release < 3) ) {
                croak "AIX $version.$release does not support NFS V4. " .
                      "Must use 5.3 or later";
            }
            if ( $dclient_h{$client} ) {
                # As of now, no version of AIX supports delegations
                croak "AIX does not support delegations";
            }
            if ( $sclient_h{$client} ) {
                # As of now, no version of AIX supports named attributes
                croak "AIX does not support named attributes";
            }
        } elsif ( $h->hosttype() =~ /unix-linux/ ) {
            my (@rel_sup, @osrel, $i);
            # Linux releases post 2.6.18 (RHEL5) support NFSv4
            @rel_sup = (2, 6, 18);
            # Note that a release does not have to have 3 numbers, we will check
            # up to first 3 numbers. It does have to have at least 2 numbers.
            @osrel = $release =~ /^(\d+)\.(\d+)(?:\.(\d+))?.*/;
            if (@osrel == 0) {
                croak "validate_nfsv4_clients: Could not handle Linux OS Release format";
            }
            for ($i = 0; $i < @rel_sup; $i++) {
                if ($osrel[$i] < $rel_sup[$i]) {
                    croak "validate_nfsv4_clients: Only Linux releases post " .
                          "2.6.18 support NFSv4. Prior releases such as " .
                          "$release do not!!!\n";
                }
                next if ($osrel[$i] == $rel_sup[$i]);
                last if ($osrel[$i] > $rel_sup[$i]);
            }
            if ( $sclient_h{$client} ) {
                # As of now, no version of Linux supports named attributes
                croak "Linux does not support named attributes";
        }
        } else {
            # Any other client does not support NFS V4
            croak "$os does not support NFS V4";
        }
    }
    return 1;
}

#---------------------------------------------------------------------
## @name ::setup_cluster
##
## @description
##  Make sure both filers are in the 'UP' state and the cluster is enabled.
##  It checks both filers are in a valid state and if one has taken over
##  the other, it attempts a giveback and waits for both filers to become
##  online.
##
## @synopsis
##  $op = setup_cluster([%options]);
##
## @arg filerA
##  Optional: Hostspec of filerA on a cluster, (default is $main::FILERA).
##
## @arg filerB
##  Optional: Hostspec of filerB on a cluster, (default is $main::FILERB).
##
## @arg timeout
##  Optional: Timeout in seconds for the cluster pair to become online,
##  (default is 120sec).
##
## @return
##  1 if succeeds.
##
## @notes
##  This function is used to get around burt 160146 when using a cluster
##  with filer simulators.
#---------------------------------------------------------------------
sub setup_cluster(;%) {
    my (%opts) = validate (
        @_,
        {
            filerA  => { type => SCALAR, default => $main::FILERA },
            filerB  => { type => SCALAR, default => $main::FILERB },
            timeout => { type => SCALAR, default => 120 },
        }
    );
    my ($filerA, $connA);
    my ($filerB, $connB);
    my $stateA = '';
    my $stateB = '';

    # Valid filer states
    my @valid_states = qw(UP TAKEOVER DEAD);

    # Create connection handles for both filers
    $filerA = host($opts{filerA});
    $connA  = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $connA->timeout($main::CMDTIMEOUT);

    $filerB = host($opts{filerB});
    $connB  = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
    $connB->timeout($main::CMDTIMEOUT);

    # Make sure both filers are licensed for cluster
    if (!filer_cluster_is_licensed($connA)) {
        # Ignore licensing errors since the filer must be rebooted
        eval { filer_cluster_license($connA) };
        eval { reboot($connA->hostp()) };
        croak "Could not reboot after licensing cluster on FILERA: $@" if $@;
        if (!filer_cluster_is_licensed($connA)) {
            croak "Could not license clustering on FILERA: $@" if $@;
        }
    }
    if (!filer_cluster_is_licensed($connB)) {
        # Ignore licensing errors since the filer must be rebooted
        eval { filer_cluster_license($connB) };
        eval { reboot($connB->hostp()) };
        croak "Could not reboot after licensing cluster on FILERB: $@" if $@;
        if (!filer_cluster_is_licensed($connB)) {
            croak "Could not license clustering on FILERB: $@" if $@;
        }
    }

    # Make sure both filers are in a valid running state
    my $stime = time();
    while ( time() - $stime < $opts{timeout}
      and ( ! grep { /^$stateA$/ } @valid_states
         or ! grep { /^$stateB$/ } @valid_states ) ) {
        sleep 10;
        $stateA = get_cf_state($connA);
        $stateB = get_cf_state($connB);
        logDbg("filerA state <$stateA>, filerB state <$stateB>");

        # Enable cluster on both filers
        cf_enable($filerA) if ($stateA =~ /disabled/i);
        cf_enable($filerB) if ($stateB =~ /disabled/i);
    }

    # Make sure both filers are in a valid state
    if ( ! ( grep { /^$stateA$/ } @valid_states )
      or ! ( grep { /^$stateB$/ } @valid_states ) ) {
        croak "Filer cluster pair is not running properly";
    }

    # Giveback to the other filer if
    # one of them is in takeover mode
    if ( $stateA eq 'TAKEOVER' ) {
        cf_giveback($filerA, partner=>$filerB) or croak $@;
    } elsif ( $stateB eq 'TAKEOVER' ) {
        cf_giveback($filerB, partner=>$filerA) or croak $@;
    }

    # Wait for both filers to be in normal state
    wait_state($filerA, "UP", $opts{timeout}) or croak $@;
    wait_state($filerB, "UP", $opts{timeout}) or croak $@;

    # Enable cluster on both filers
    cf_enable($filerA) or croak $@;
    cf_enable($filerB) or croak $@;

    return 1;
}

#---------------------------------------------------------------------
## @name ::validate_nfs_version
## @description
##  Validate NFS version with list of allowed versions.
##
## @synopsis
##  $out = validate_nfs_version(version => 2|3|4|'all'|'not2'|'not3'|'not4',
##                              allowed => \@allowed_versions);
##
## @arg version
##  Required: NFS version to validate. In addition to the version number to
##  validate, it could be one of the following strings: 'all' (2, 3, 4),
##  'not2' (3, 4), 'not3' (2, 4) or 'not4' (2, 3).
##
## @arg allowed
##  Required: Reference to an array of allowed NFS versions.
##
## @return
##  Returns the array of validated versions else croaks with error.
#---------------------------------------------------------------------
sub validate_nfs_version(%) {
    my (%opts) = validate (
        @_,
        {
            version => { type => SCALAR,
                        regex => qr/^([234]|all|not2|not3|not4)$/ },
            allowed => { type => ARRAYREF },
        }
    );
    my (@vers, $ver, %valid_vers);

    # Convert the version string to an
    # array of versions to validate
    @vers = ( $opts{version} eq 'all' )  ? (2, 3, 4) :
            ( $opts{version} eq 'not2' ) ? (3, 4) :
            ( $opts{version} eq 'not3' ) ? (2, 4) :
            ( $opts{version} eq 'not4' ) ? (2, 3) : ( $opts{version} );

    # Hash of valid NFS versions to ease comparison of lists
    %valid_vers = map { $_ => 1 } @{$opts{allowed}};

    # Validate versions
    foreach $ver ( @vers ) {
        croak "Invalid NFS version specified: $opts{version} " .
              "(v$ver not supported)" if ( !$valid_vers{$ver} );
    }

    return @vers;
}

#---------------------------------------------------------------------
## @name ::range_ipaddrs
## @description
##  Get a list of ip addresses given by the initial ip address and
##  the number of addresses in the range.
##
## @synopsis
##  $iplist = range_ipaddrs(ipaddr => $initial_ipaddr,
##                          number => $number);
##
## @arg ipaddr
##  Required: Initial ip address of range.
##
## @arg number
##  Required: Number of ip addresses in range.
##
## @return
##  A reference to an array of ip addresses in the specified range.
#---------------------------------------------------------------------
sub range_ipaddrs(%) {
    my (%opts) = validate (
        @_,
        {
            ipaddr => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
            number => { type => SCALAR, regex => qr/^\d+$/ },
        }
    );
    my (@ipaddrs, $count, $octet);
    my $ip = 0;

    # Convert the initial virtual ip address to a number
    ( $ip = ($ip<<8)|$_ ) foreach split(/\./, $opts{ipaddr});

    $count = 0;
    while ( $count < $opts{number} ) {
        $octet = $ip&0xFF;
        # Skip the '0' and '255' addresses
        if ( $octet > 0 and $octet < 255 ) {
            # Convert the number back to an ip address string
            push @ipaddrs, join('.', map {($ip>>$_)&0xFF} (24, 16, 8, 0));
            ++$count;
        }
        # Increment ip address
        ++$ip;
    }

    return \@ipaddrs;
}

#---------------------------------------------------------------------
## @name ::ping_ipaddrs
## @description
##  Ping all ip address given in list and group them into alive
##  and dead lists. This function must be run as root since the
##  ping uses the icmp protocol.
##
## @synopsis
##  $alive|($alive, $dead) = ping_ipaddrs(ipaddrs => \@ipaddrs
##                                        [, timeout => $ping_timeout ]
##                                        [, pcount  => $pcount ]);
##
## @arg ipaddrs
##  Required: Reference to an array of ip addresses to ping.
##
## @arg timeout
##  Ping timeout to decide if ip address is alive or dead (default is 5).
##
## @arg pcount
##  Number of pings to execute in parallel (default is 100).
##
## @return
##  In scalar context: a reference to an array of alive ip addresses.
##  In array context: both array references of alive and dead ip addresses.
#---------------------------------------------------------------------
sub ping_ipaddrs(%) {
    my (%opts) = validate (
        @_,
        {
            ipaddrs => { type => ARRAYREF },
            timeout => { type => SCALAR, regex => qr/^\d+$/, default => 5 },
            pcount  => { type => SCALAR, regex => qr/^\d+$/, default => 100 },
        }
    );
    my (@ipaddrs, @alive, @dead, @fhs, $fh, $ipaddr, $index, $end, $cmd, $out);
    my $timeout = $opts{timeout};
    my $count = @{$opts{ipaddrs}};
    my $idx = 0;

    # Process pings in parallel in batches of 100 (default)
    while ( $idx < $count ) {
        $end = $idx + $opts{pcount};
        $end = $count if ($end > $count);
        @ipaddrs = @{$opts{ipaddrs}}[$idx..$end-1];
        $idx = $end;
        @fhs = ();
        $index = 0;
        # Ping all ip addresses in parallel
        foreach $ipaddr ( @ipaddrs ) {
            # Program to ping the client
            $cmd = qq{
                use Net::Ping;
                my \$p = Net::Ping->new(q(icmp), $timeout);
                print q(--alive--) if \$p->ping(q($ipaddr));
                print q($ipaddr);
            };
            # Fork and exec program and pipe the output
            open($fhs[$index++], qq{perl -e '$cmd'|});
        }

        # Wait for all pings to finish
        foreach $fh ( @fhs ) {
            # Read from pipe
            $out = <$fh>;
            # Add ip address to correct list
            if ( $out =~ s/--alive--// ) {
                push @alive, $out;
            } else {
                push @dead, $out;
            }
            # Close pipe and wait for child to finish
            close($fh);
        }
    }

    # Return results
    if ( wantarray ) {
        return (\@alive, \@dead);
    } else {
        return \@alive;
    }
}

#---------------------------------------------------------------------
## @name ::available_ipaddrs
## @description
##  Get a list of available ip addresses given by the initial ip
##  address. Options 'number' and 'max_num' are mutually exclusive.
##
## @synopsis
##  $iplist = available_ipaddrs(ipaddr => $initial_ipaddr,
##                              [, number => $number | max_num => $max_num ],
##                              [, timeout => $timeout ],
##                              [, pcount => $pcount ]);
##
## @arg ipaddr
##  Required: Initial ip address to test for available ip addresses.
##
## @arg number
##  Total number of available ip addresses to return.
##
## @arg max_num
##  Maximum number of ip addresses to check.
##
## @arg timeout
##  Ping timeout to decide if ip address is alive or dead (default is 5).
##
## @arg pcount
##  Number of pings to execute in parallel (default is 100).
##
## @return
##  A reference to an array of available ip addresses in the specified range.
#---------------------------------------------------------------------
sub available_ipaddrs(%) {
    my (%opts) = validate (
        @_,
        {
            ipaddr  => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
            number  => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
            max_num => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
            timeout => { type => SCALAR, regex => qr/^\d+$/,  default => 5 },
            pcount  => { type => SCALAR, regex => qr/^\d+$/,  default => 100 },
        }
    );
    my (@ipaddrs, $ips, $available);
    my $ipaddr  = $opts{ipaddr};
    my $count   = $opts{number};
    my $max_num = $opts{max_num};
    my %ping_opts = ( timeout => $opts{timeout}, pcount => $opts{pcount} );

    # Check mutually exclusive options
    if ( ! (defined($count) xor defined($max_num)) ) {
        croak "Options 'number' and 'max_num' are mutually exclusive";
    }

    # Check exactly max_num
    $count = $max_num if defined($max_num);

    while ( $count > 0 ) {
        if ( defined($opts{number}) ) {
            # Get 10% more than requested
            $count = 1.1*$count;
            # Ping at least 10 clients
            $count = 10 if $count < 10;
        }
        # Get a list of ip addresses (10% more than requested)
        $ips = range_ipaddrs(ipaddr=>$ipaddr, number=>int($count));
        # Ping list of ip addresses in parallel
        (undef, $available) = ping_ipaddrs(ipaddrs=>$ips, %ping_opts);

        # The maximum number of ip's have been checked
        # so return all available addresses
        return $available if defined($max_num);

        # Save list of ip addresses which did not respond in time
        push @ipaddrs, @$available;

        # Get the number of ip addresses missing
        $count = $opts{number} - @ipaddrs;
        # Make the last ip address in range the new
        # initial address for next iteration
        $ipaddr = pop @$ips;

        if ( @ipaddrs and $ipaddr eq $ipaddrs[-1] ) {
            pop @ipaddrs;
            ++$count;
        }
    }

    # Make sure list is exactly the number of ip addresses requested
    splice @ipaddrs, $opts{number} if $count < 0;

    return \@ipaddrs;
}

#---------------------------------------------------------------------
## @name ::check_for_panic
## @description
##  Determine if the named filer has panicked by attempting to ping
##  the filer. If the filer does not respond to the ping, assume a
##  panic has occurred, and reboot the filer using transit_to UP.
##
## @synopsis
##  $panic = check_for_panic( filer   => $FILER,
##                            timeout => 30 );
##
## @arg filer
##  Hostspec of filer to test. Defaults to $main::FILER.
##
## @arg timeout
##  Timeout for ping operation. Defaults to 15.
##
## @return
##  Returns 1 if filer has panicked, 0 otherwise..
#---------------------------------------------------------------------
sub check_for_panic(%) {
    my (%opts) = validate (
        @_,
        {
            filer   => { type => SCALAR, default => $main::FILER },
            timeout => { type => SCALAR, default => 15, regex => qr/^\d+$/ },
        }
    );

    my $panic;
    my $hostrec = host( $opts{filer} );

    $panic = ! host_is_alive( $opts{filer}, $opts{timeout} );

    if ( $panic ) {
        eval { transit_to( $hostrec, 'UP' ) };
        croak "Host is down, could not reboot: $@" if $@;
    }

    return $panic;
}

#---------------------------------------------------------------------
## @name ::flush_reply_cache
## @description
##  Flush the NFS reply cache on the filer.
##
## @synopsis
##  $out = flush_reply_cache( conn => $fc [, ipaddr => $ipaddr] );
##
## @arg conn
##  Required: Connectrec object of filer.
##
## @arg ipaddr
##  Optional: flush cache entries for this ip address only.
##
## @return
##  1 if succeeds.
#---------------------------------------------------------------------
sub flush_reply_cache(%) {
    my (%opts) = validate (
        @_,
        {
            conn   => { type => OBJECT },
            ipaddr => { type => SCALAR, default => '',
                       regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
        }
    );
    my $out;

    # Flush reply cache on the filer
    $out = sayEOF($opts{conn}, "priv set -q diag;" .
                  "nfs diag flush duplicate_request_cache $opts{ipaddr}");

    # Remove all flush messages
    $out =~ s/\d+ flows were flushed from the reply cache.\s*//mg;
    croak $out if $out;

    return 1;
}

#---------------------------------------------------------------------
## @name ::get_reply_cache
## @description
##  Get the NFS reply cache from the filer.
##
## @synopsis
##  $cache = get_reply_cache( conn => $fc [, ipaddr => $ipaddr] );
##
## @arg conn
##  Required: Connectrec object of filer.
##
## @arg ipaddr
##  Optional: get cache entries for this ip address only.
##
## @return
##  Returns a hash reference of the reply cache
#---------------------------------------------------------------------
sub get_reply_cache(%) {
    my (%opts) = validate (
        @_,
        {
            conn   => { type => OBJECT },
            ipaddr => { type => SCALAR, default => '',
                       regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
        }
    );
    my ($head, @flist, $text, @lines, @fkeys, %hash, $key);
    my ($fhead, $alloc_fail, $conn_drop, $flow, %ret, $out, @hms);
    my @hconv = qw(XID CKSUM FILEID GEN FSID);

    # Get reply cache from the filer
    $out = sayEOF($opts{conn}, "priv set -q diag;" .
                  "nfs diag show duplicate_request_cache $opts{ipaddr}");

    # Remove '\r' if any and split output into header and flows
    $out =~ s/\r//g;
    ($head, @flist) = split /\n{2,}/, $out;

    # Split header into general info, alloc failures and connection drops
    ($fhead, $alloc_fail, $conn_drop) = ($head =~ /NFS Reply Cache\s+(.*)\s+Alloc failures:\s+(.*)\s+Connection drops because of in progress hits:\s+(.*)/s);

    croak $out unless defined $fhead;

    # Process general info
    %hash = ($fhead =~ /([^=]+)=(\d+),?\s*/sg);
    %ret = %{_change_keys(\%hash)};

    # Process alloc failures
    %hash = ($alloc_fail =~ /([^=]+)=(\d+)\s*/sg);
    $ret{Alloc_Fail} = _change_keys(\%hash);

    # Process connection drops
    %hash =
      ($conn_drop =~ /([^=]+) conn dropped(?:, no reconnect)?=(\d+)\s*/sg);
    $ret{Conn_Drop} = _change_keys(\%hash);

    # Process all flows
    $ret{Flows} = [];
    foreach $head ( @flist ) {
        # Split flow into header and cache entries
        ($fhead, $text) = ($head =~ /(.*)\s+(STATE.*)/s);

        # Process flow header
        %hash = ($fhead =~ /([^=]+)=([^,\n]+),?\s*/sg);
        $flow = _change_keys(\%hash);

        # Change TOLR to seconds
        @hms = ($flow->{TOLR} =~ /(\d+)/g);
        $flow->{TOLR} = 3600*$hms[0] + 60*$hms[1] + $hms[2];

        # Split cache entries and get column names
        @lines = split /\n/, $text;
        $fhead = shift @lines;
        @fkeys = split ' ', $fhead;

        # Process all cache entries
        %hash = ();
        $flow->{Cache} = [];
        foreach $text ( @lines ) {
            # Create cache entry
            @hash{@fkeys} = split ' ', $text;
            push @{$flow->{Cache}}, _change_keys(\%hash);

            # Convert hex numbers to decimal
            foreach $key ( @hconv ) {
                $flow->{Cache}[-1]{$key} = hex($flow->{Cache}[-1]{$key});
            }
        }

        # Add flow information to return hash
        push @{$ret{Flows}}, $flow;
    }

    return \%ret;
}

#---------------------------------------------------------------------
## @name ::get_reply_cache_entry
## @description
##  Get the entry from the specified reply cache hash which matches
##  the specified reply information.
##
## @synopsis
##  $out = get_reply_cache_entry( rc => $rc, xid => $xid, version => 2|3|4,
##                                proto => 'udp'|'tcp', port => $port,
##                                ipaddr => $ipaddr );
##
## @arg rc
##  Required: Reply cache hash as returned from get_reply_cache.
##
## @arg xid
##  Required: Transaction id (xid) of reply.
##
## @arg version
##  Optional: NFS version of reply: 2, 3 or 4.
##
## @arg proto
##  Optional: Protocol of reply: 'udp' or 'tcp'.
##
## @arg port
##  Optional: Port number of reply.
##
## @arg ipaddr
##  Optional: Client IP address of reply.
##
## @return
##  Returns a hash reference of the reply cache entry
##  or undef if entry not found
#---------------------------------------------------------------------
sub get_reply_cache_entry(%) {
    my (%opts) = validate (
        @_,
        {
            rc      => { type => HASHREF },
            xid     => { type => SCALAR, regex => qr/^\d+$/ },
            version => { type => SCALAR, regex => qr/^[234]$/ },
            proto   => { type => SCALAR, regex => qr/^(udp|tcp)$/ },
            port    => { type => SCALAR, regex => qr/^\d+$/ },
            ipaddr  => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
        }
    );
    my ($flow, $item);
    my $proto = uc($opts{proto});

    # Find reply on all flows
    foreach $flow ( @{$opts{rc}{Flows}} ) {
        # Check for correct flow
        next unless $flow->{Client_IP} eq $opts{ipaddr}
          && $flow->{Protocol} eq $proto && $flow->{Source_Port} eq $opts{port};

        # Find reply cache entry on flow
        foreach $item ( @{$flow->{Cache}} ) {
            if ( $item->{XID} == $opts{xid}
              && $item->{PROC} =~ /^v$opts{version}_/ ) {
                # Found correct entry
                return $item;
            }
        }
    }

    # Entry not found
    return;
}

#---------------------------------------------------------------------
## @name ::get_reply_cache_flow
## @description
##  Get the flow from the specified reply cache hash which matches
##  the specified flow information.
##
## @synopsis
##  $flow = get_reply_cache_flow( rc => $rc, proto => 'udp'|'tcp',
##                                port => $port, ipaddr => $ipaddr );
##
## @arg rc
##  Required: Reply cache hash as returned from get_reply_cache.
##
## @arg proto
##  Optional: Protocol of flow: 'udp' or 'tcp'.
##
## @arg port
##  Optional: Port number of flow.
##
## @arg ipaddr
##  Optional: Client IP address of flow.
##
## @return
##  Returns a hash reference of the reply cache flow
##  or undef if flow not found
#---------------------------------------------------------------------
sub get_reply_cache_flow(%) {
    my (%opts) = validate (
        @_,
        {
            rc      => { type => HASHREF },
            proto   => { type => SCALAR, regex => qr/^(udp|tcp)$/ },
            port    => { type => SCALAR, regex => qr/^\d+$/ },
            ipaddr  => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
        }
    );
    my $flow;
    my $proto = uc($opts{proto});

    # Find flow
    foreach $flow ( @{$opts{rc}{Flows}} ) {
        # Check for correct flow
        if ( $flow->{Client_IP} eq $opts{ipaddr}
          && $flow->{Protocol} eq $proto
          && $flow->{Source_Port} eq $opts{port} ) {
            return $flow;
        }
    }

    # Flow not found
    return;
}

#--------------------------------------------------------------------
## @name ::get_good_nfs_version
## @description
##  Get a good enough version of nfs from the given string.
##
## @synopsis
##  $version = get_good_nfs_version( $nfs_version )
## 
## @arg nfs_version
##  Optional: Any valid string for nfs version: 2, 3, 4, 4.1, not2, not3, not4, not4.1, all
##
## @returns
##  The highest usable value for NFS version based on the give argument.
#--------------------------------------------------------------------
sub get_good_nfs_version(;$) {
	my $given = shift;

	if (!$given) {
        param('NFS_VERSION', '-default', 'all');
		$given = $main::NFS_VERSION;
	}
	param('NFS4_MINOR_VERSION', '-default', '0');
        my $nfs4_minor_version = $main::NFS4_MINOR_VERSION;

        if ($given eq "all") {
            return "4";
        } elsif ($given eq "2") {
            return "2";
        } elsif ($given eq "3") {
            return "3";
        } elsif ($given eq "4") {
            if ($nfs4_minor_version eq "1"){
                return "4.1";
            }else{
                return "4";
            }
        } elsif ($given eq "4.1") {
            return "4.1";
        } elsif ($given eq "not2") {
            return "4";
        } elsif ($given eq "not3") {
            return "4";
        } elsif ($given eq "not4") {
            return "3";
        } elsif ($given eq "not4.1") {
            return "4";
        } else {
            croak (qq{Unrecognized version given: $given});
    }
}

#--------------------------------------------------------------------
## 
## @name ::get_good_auth
## @description
##  Get a good enough authentication. This is to be used for configuring filer
##  before test, and mounting for cleanup after the test.
##
## @synopsis
##  $auth = get_good_auth([$auth_string]);
##
## @auth_string
## (Optional): Any string representing nfs security types: any, none, sys, krb5, krb5i, and krb5p.
##				If not provided, the function uses the value from $main::AUTH.
##
## @out
##  Returns a valid authentication for the given string. (sys if "any" is given, and verbatim for all other input values).
##
sub get_good_auth() {
	my $given = shift;

	if (!$given) {
		$given = $main::AUTH;
	}

	if ($given eq "all") {
		return "sys";
	} else {
		return $given;
	}
}

#---------------------------------------------------------------------
# Internal function to change spaces on the keys of the specified hash
#---------------------------------------------------------------------
sub _change_keys($;) {
    my $hash = shift;
    my (%ret, $key);

    # Process all keys in the hash
    foreach ( keys %$hash ) {
        $key = $_;
        # Remove single quotes
        $key =~ s/\'//g;
        # Change spaces to underscore
        $key =~ s/ /_/g;
        $ret{$key} = $hash->{$_};
    }

    return \%ret;
}

#--------------------------------------------------------------------
# What is the address used by the client to talk to the given filer ?
#--------------------------------------------------------------------
sub get_client_addr($$) {
	my ($client, $dest) = @_;
	#logStep(qq{Need to look for $dest});
    my %opt;
    if (ref($client)) {
        if (UNIVERSAL::isa($client, "Connectrec")) {
            $opt{hostrec} = $client->hostp();
        } elsif (UNIVERSAL::isa($client, "Hostrec")) {
            $opt{hostrec} = $client;
        }
    } else {
        $opt{client} = $client;
    }
	my $client_obj = NACL::C::Client->new(%opt);
    my @cmd = ("netstat -naT | grep $dest | grep ESTABLISHED", "netstat -naW | grep $dest | grep ESTABLISHED");
    if ($client_obj->hostrec->{hosttype} =~ /win/) {
        @cmd = ("netstat -na");
    }
	chomp($dest);

    my ($line, $output);
    foreach my $cmd (@cmd) {
	    $output = $client_obj->execute(command=>$cmd);
        if ($output) {
            logStep(qq{Found a output(cmd:$cmd):\n$output});
        }
        if (!$output) {
		    croak (qq{Netstat didnot work for(cmd:$cmd):address($dest). netstat output:\n..$output..\n});
        }

        foreach (split("\n", $output)) {
            if (/$dest/) {
                $line = $_;
                last;
            }
        }
        if ($line) {
            last;
        }
    }

	if (!$line) {
		croak (qq{Couldnot find the required ip address($dest). netstat output:$output\n});
	}
	#logStep(qq{Got the line: $line});
	#tcp        0      0 fd20:8b1e:b255:832f:214:5eff:fe55:c2f7:938 fd20:8b1e:b255:832f::3c:2049 ESTABLISHED 

	my $pattern = q{(\S*):\d+\s+\[?} . $dest . q{\]?.+ESTABLISHED};
	if ($line =~ /$pattern/) {
	    my $result = $1;
	    $result =~ s/[\[\]]//g;
	    #print(qq{The client is using this address:...$result...\n});
		return $result;
	} else {
	    die (qq{Couldnot find the required ip $dest. line:$line\n});
	}

}

1;

