#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/V4t.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#

package nfs::V4t;

## @summary Library to run the v4t synthetic NFS V4 client
##
## @description
##  Wrapper functions for v4t_* executables which are found
##  in /usr/local/test/bin.
##
## @usage
##  use nfs::V4t;
##
## @status	review
##
## @author	dl-maryv (FSS Protocols QA)
## @notes	Created by EP Network Storage Performance Lab, Inc.

use strict;

use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF OBJECT UNDEF);

use vars qw(
  $VERSION
  @ISA
  @EXPORT_OK
);

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
  v4t_supported
  v4t_write_stateid
  v4t_open_deny
  v4t_byte_range_lock
  v4t_clear_locks
  v4t_op_create
  v4t_list_attr
  v4t_set_acl
  v4t_compound_basic
  v4t_compound_malformed
  v4t_compound_fh
  v4t_compound_transition
  v4t_compound_allops
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
$VERSION = '0.01';

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $sudo     = '/usr/software/bin/sudo';
my $v4t_path = "/usr/software/test/bin/v4t";

# Common parameter for all functions running v4t executables
my %common_parameters = (
    client_id => { type => SCALAR, default => $$ },
    ipaddr    => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
    user      => { type => SCALAR, optional => 1 },
    timeout   => { type => SCALAR, default => $main::CMDTIMEOUT ?
                                              $main::CMDTIMEOUT : 0 },
    client    => { type => SCALAR|OBJECT, default => $main::CLIENT ?
                                                     $main::CLIENT :
                                                     'localhost' },
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub v4t_get_errors($;);
sub v4t_run(%;);
sub v4t_supported(%;);
sub v4t_write_stateid(%;);
sub v4t_open_deny(%;);
sub v4t_byte_range_lock(%;);
sub v4t_clear_locks(%;);
sub v4t_op_create(%;);
sub v4t_list_attr(%;);
sub v4t_set_acl(%;);
sub v4t_compound_basic(%;);
sub v4t_compound_malformed(%;);
sub v4t_compound_fh(%;);
sub v4t_compound_transition(%;);
sub v4t_compound_allops(%;);

#---------------------------------------------------------------------
# This is an internal function.
# Get all errors found in the output of the v4t synthetic client
#---------------------------------------------------------------------
sub v4t_get_errors($;) {
    my $out = shift;

    # Get errors of type: '*** some error ***'
    my @err = grep { /^\*\*\* .* \*\*\*$/ } split(/\r?\n/, $out);
    # Get errors of type: 'Step 1.3 operation -- failed'
    push @err, grep { /^Step .* -- failed/ } split(/\r?\n/, $out);
    # Get all NFS4ERR* errors
    push @err, grep { /NFS4ERR/ } split(/\r?\n/, $out);

    # Remove the 'Step X.X' from all errors
    s/Step [^:]+:\s+// foreach @err;

    return @err;
}

#---------------------------------------------------------------------
# This is an internal function.
# Get the hostrec object from the client hostspec or connectrec object.
# If the argument is already a hostrec object, it is simply returned.
#---------------------------------------------------------------------
sub get_hostp($;) {
    my $client = shift;

    if ( UNIVERSAL::isa($client, "Connectrec") ) {
        # Get hostrec object from a connectrec object
        $client = $client->hostp();
    } else {
        # Get hostrec object from a hostspec
        $client = host($client);
    }

    return $client;
}

#---------------------------------------------------------------------
# This is an internal function.
# Run the specified v4t command on a local or remote client.
# Arguments are passed in a "key => value" style.
#
# v4t_cmd  Required: Command to run on client.
# runid    Optional: Runid given to subtest when running remotely.
# client   Optional: Hostspec of client where the v4t synthetic client runs,
#                    (default is 'localhost').
# user     Optional: Run command as this user.
# timeout  Optional: Time in seconds to wait before timing out
#                    This option is only used when running remotely.
#                    (default is 0; wait forever).
#
# return:  Output of v4t command or croak on failure.
#---------------------------------------------------------------------
sub v4t_run(%;) {
    my (%opts) = validate (
        @_,
        {
            v4t_cmd => { type => SCALAR },
            runid   => { type => SCALAR,        default => 'v4t_run' },
            user    => { type => SCALAR|UNDEF, optional => 1 },
            client  => { type => SCALAR|OBJECT, default => 'localhost' },
            timeout => { type => SCALAR,        default => 0 },
        }
    );
    my (%extra_options, $sub_obj, $out);

    # Get hostspec from the client option
    my $client = get_hostp($opts{client});
    my $client_name = $client->hostname();

    my $v4t_cmd = $opts{v4t_cmd};
    logTrc("$client_name: $v4t_cmd");

    # Only append the username if the option is given
    $extra_options{'-username'} = $opts{user} if defined($opts{user});

    if ( $client_name eq 'localhost' and not defined($opts{user}) ) {
        # Run the synthetic client locally
        $out = qx($v4t_cmd);
        logDbg($out);
    } else {
        # Script to run the synthetic client remotely
        my $cmd = qq{ \{
            logcomment(q($v4t_cmd));
            my \$out = qx($v4t_cmd);
            if ( \$out ) \{
                logresult(q(PASS), \$out);
            \} else \{
                logresult(q(FAIL), q(v4t synthetic client failed to run));
            \}
        \} };

        # Run the synthetic client remotely
        $sub_obj = subtest( $cmd,
                            '-runid',   $opts{runid},
                            '-timeout', $opts{timeout},
                            '-client',  $client,
                            %extra_options,
                            '-onexit', sub {} );

        # Get all results from subtest
        my @results = $sub_obj->results();
        my $worst = NATE::Result::worst(@results);
        if ( $worst->severity() > 2 ) {
            croak $worst;
        }

        # Get the output of the synthetic client
        # which is in the PASS result
        foreach my $result ( @results ) {
            if ( $result->type() eq 'PASS' ) {
                ($out) = $result->messages();
                last;
            }
        }
    }

    return $out;
}

#---------------------------------------------------------------------
## @name ::v4t_supported
## @description
##  Checks if client supports v4t synthetic client. In specific,
##  it checks whether or not the right v4t executables are installed
##  on the client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_supported( %options );
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec of client to check support for v4t synthetic client.
##  This could be either a hostspec or a reference to a list of hostspecs,
##  (default is $main::CLIENT).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if v4t synthetic client is supported or croaks if not supported
#---------------------------------------------------------------------
sub v4t_supported(%;) {
    my (%opts) = validate (
        @_,
        {
            user    => { type => SCALAR, optional => 1 },
            client  => { type => SCALAR|ARRAYREF, default => $main::CLIENT ?
                                                             $main::CLIENT :
                                                             'localhost' },
            timeout => { type => SCALAR, default => $main::CMDTIMEOUT ?
                                                    $main::CMDTIMEOUT : 0 },
        }
    );
    my (@clients, $cmd, $out);

    # Create a list of clients to check
    if ( ref($opts{client}) ) {
        # This is a reference to a list
        @clients = @{$opts{client}};
    } else {
        # This is just a single hostspec
        @clients = ($opts{client});
    }

    # Check existence of one v4t_ executable
    $cmd = "if [ -f $v4t_path/v4t_onelk ]; then echo --v4t_ok--; fi;";

    foreach my $client ( @clients ) {
        # Run check on client
        $out = v4t_run( v4t_cmd => $cmd,
                        runid   => 'v4t_supported',
                        client  => $client,
                        timeout => $opts{timeout},
                        user    => $opts{user} );
    }

    # Check output of command
    if ( $out !~ /--v4t_ok--/ ) {
        croak "Client does not support v4t synthetic client";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_write_stateid
## @description
##  Issue a write with the special stateid of all zeroes or all ones.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_write_stateid( %options );
##
## @arg stateid
##  Required: If '1' issue a write with the special stateid of all ones,
##  if '0' all zeroes are used for the stateid.
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_write_stateid(%;) {
    my (%opts) = validate (
        @_,
        {
            file    => { type => SCALAR },
            topdir  => { type => SCALAR, optional => 1 },
            stateid => { type => SCALAR,    regex => qr/^[01]$/ },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options, $regex);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}";

    if ( $opts{stateid} ) {
        # Send a write with the special stateid of all ones
        $v4t_cmd = "$v4t_path/v4t_stid $options -w -1";

        # Create regex expression to validate operation
        $regex = qr/write\s+\S+\s+\S+\s+F{8}\s+F{8}\s+F{8}/;
    } else {
        # Send a write with the special stateid of all zeroes
        $v4t_cmd = "$v4t_path/v4t_stid $options -w";

        # Create regex expression to validate operation
        $regex = qr/write\s+\S+\s+\S+\s+0\s+0\s+0/;
    }

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_write_stateid',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( $out !~ /$regex/ ) {
        my @err = v4t_get_errors($out);
        croak "Write with the special stateid of all $opts{stateid} failed: ",
          join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_open_deny
## @description
##  Open a file for reading or writing denying read or write access.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_open_deny( %options );
##
## @arg open_type
##  Required: Type of access to open file: 'read' or 'write'.
##
## @arg deny_type
##  Required: Type of access to deny: 'read', 'write' or 'all'.
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_open_deny(%;) {
    my (%opts) = validate (
        @_,
        {
            file      => { type => SCALAR },
            topdir    => { type => SCALAR, optional => 1 },
            open_type => { type => SCALAR, regex => qr/^(read|write)$/ },
            deny_type => { type => SCALAR, regex => qr/^(read|write|all)$/ },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}";

    # Create v4t synthetic client command
    my %deny_h = ( 'read' => 1, 'write' => 2, all => 3 );
    my $ltype = $opts{open_type} eq 'read' ? '-a 1' : '-a 2';
    my $lock  = "$ltype -x $deny_h{$opts{deny_type}}";
    $v4t_cmd  = "$v4t_path/v4t_onelk $options $lock";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_open_deny',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( $out !~ /\slock\[2:0:-1\]\(0\)\s--\swent\sOK/ ) {
        croak "Could not open file with denying $opts{deny_type}";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_byte_range_lock
## @description
##  Get a byte range lock on a file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_byte_range_lock( %options );
##
## @arg lock_type
##  Required: Type of lock to get: 'read' or 'write'.
##
## @arg start_byte
##  Required: Starting byte of range lock.
##
## @arg byte_length
##  Required: Number of bytes of range lock.
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_byte_range_lock(%;) {
    my (%opts) = validate (
        @_,
        {
            file        => { type => SCALAR },
            topdir      => { type => SCALAR, optional => 1 },
            lock_type   => { type => SCALAR, regex => qr/^(read|write)$/ },
            start_byte  => { type => SCALAR, regex => qr/^\d+$/ },
            byte_length => { type => SCALAR, regex => qr/^\d+$/ },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}";

    # Create v4t synthetic client command
    my $ltype = $opts{lock_type} eq 'read' ? '-a 1 -s' : '-a 2 -e';
    my $lock  = "$ltype $opts{start_byte}:$opts{byte_length}";
    $v4t_cmd  = "$v4t_path/v4t_onelk $options $lock";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_byte_range_lock',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Create regex expression to validate operation
    my $lk = $opts{lock_type} eq 'read' ? 1 : 2;
    my $sb = $opts{start_byte};
    my $bl = $opts{byte_length};

    # Check output of v4t synthetic client
    if ( $out !~ /\slock\[$lk:$sb:$bl\]\(0\)\s--\swent\sOK/ ) {
        croak "Could not get a byte range lock";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_clear_locks
## @description
##  Clear or release all locks for v4t client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_clear_locks( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_clear_locks(%;) {
    my (%opts) = validate (
        @_,
        {
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";

    # Clear or release all locks from this client id
    $v4t_cmd = "$v4t_path/v4t_stid $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_clear_locks',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( $out !~ /client confirm\S+\s--\swent OK/ ) {
        croak "Could not clear locks for client";
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_op_create
## @description
##  Create a file using 'op-create'.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_op_create( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg directory
##  Optional: If true, option 'file' is a directory, (default is 0).
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg acl_file
##  Optional: Specifies file to get ACL(s) from.
##
## @arg mode
##  Optional: Specifies mode bits to set.
##  Octal number or string, e.g., 0644, '644'.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_op_create(%;) {
    my (%opts) = validate (
        @_,
        {
            file      => { type => SCALAR },
            topdir    => { type => SCALAR,  optional => 1 },
            directory => { type => BOOLEAN,  default => 0 },
            acl_file  => { type => SCALAR,  optional => 1 },
            mode      => { type => SCALAR,  optional => 1,
                                               regex => qr{^\d+$}},
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options, $regex);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}";

    # Process optional arguments
    if ( defined($opts{mode}) ) {
        # Convert mode into an octal string
        if ( (~$opts{mode} & $opts{mode}) eq '0' ) {
            # Mode was passed as a octal number,
            # convert it to a string.
            $opts{mode} = sprintf '0%lo', $opts{mode};
        }
    }

    # Create v4t synthetic client command
    my $flags = '-r ' if $opts{directory};
    $flags   .= "-b $opts{mode}" if $opts{mode};
    $flags   .= "-a $opts{acl_file}" if $opts{acl_file};
    $v4t_cmd  = "$v4t_path/v4t_crf $options $flags";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_op_create',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Create regex expression to validate operation
    if ( $opts{directory} ) {
        $regex = qr/create-dir\S+\(0\) -- went OK/;
    } else {
        $regex = qr/open-cr\S+\(0\) -- went OK/;
    }

    # Check output of v4t synthetic client
    if ( $out !~ /$regex/ ) {
        my @err = v4t_get_errors($out);
        croak "Could not create file: ", join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_list_attr
## @description
##  Get list of attributes for file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_list_attr( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Optional: File to perform operation on.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg all
##  Optional: Fetch all file attributes, (default is 0).
##
## @arg acl
##  Optional: Include ACLs on the output, (default is 0).
##
## @arg rand_attr
##  Optional: Causes a random set of options to be fetched, (default is 0).
##
## @arg readdir
##  Optional: Do a READDIR on the passed in directory, (default is 0).
##  If this option is set then 'topdir' is used and the 'file' option
##  is ignored.
##
## @arg total_max
##  Optional: Specifies the maximum output size.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  Output of v4t synthetic client or croaks if it fails
#---------------------------------------------------------------------
sub v4t_list_attr(%;) {
    my (%opts) = validate (
        @_,
        {
            file      => { type => SCALAR, optional => 1 },
            topdir    => { type => SCALAR, optional => 1 },
            all       => { type => BOOLEAN, default => 0 },
            acl       => { type => BOOLEAN, default => 0 },
            rand_attr => { type => BOOLEAN, default => 0 },
            readdir   => { type => BOOLEAN, default => 0 },
            total_max => { type => SCALAR, optional => 1 },
            %common_parameters
        }
    );
    my ($out, $options, $v4t_cmd);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    if ($opts{readdir}) {
        $options .= " -D $opts{topdir}" if $opts{topdir};
        $options .= " -t $opts{total_max}" if $opts{total_max};
    } else {
        $options .= " -C $opts{topdir}" if $opts{topdir};
        $options .= " -F $opts{file}";
    }
    $options .= ' -e' if $opts{all};
    $options .= ' -a' if ($opts{acl} and !$opts{all});
    $options .= ' -r' if $opts{rand_attr};

    # Create v4t synthetic client command
    $v4t_cmd  = "$v4t_path/v4t_ls $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_list_attr',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( !$opts{readdir} && $out !~ /Getting attributes -- went OK/ ) {
        my @err = v4t_get_errors($out);
        croak "Could not get attributes for file: ", join("\n", @err);
    }

    return $out;
}

#---------------------------------------------------------------------
## @name ::v4t_set_acl
## @description
##  Set the ACL on a file.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_set_acl( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg topdir
##  Required: Directory where file is located.
##
## @arg acl_file
##  Required: Specifies file to get acl(s) from.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  Output of v4t synthetic client or croaks if it fails
#---------------------------------------------------------------------
sub v4t_set_acl(%;) {
    my (%opts) = validate (
        @_,
        {
            file     => { type => SCALAR },
            topdir   => { type => SCALAR },
            acl_file => { type => SCALAR },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file} -f $opts{acl_file} -c 1";

    # Create v4t synthetic client command
    $v4t_cmd  = "$v4t_path/v4t_setacl $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_set_acl',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( $out !~ /set attr -- went OK/ ) {
        my @err = v4t_get_errors($out);
        croak "Could not set ACL attributes on file: ", join("\n", @err);
    }

    return $out;
}

#---------------------------------------------------------------------
## @name ::v4t_compound_basic
## @description
##  Issues a basic compound request.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_compound_basic( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg null
##  Optional: If true send a null compound, else send a very large
##  compound, (default is 0).
##
## @arg non_idempotent
##  Optional: Number on non-idempotent operations, (default is 0).
##
## @arg file
##  Optional: File to perform operation on.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_compound_basic(%;) {
    my (%opts) = validate (
        @_,
        {
            file           => { type => SCALAR, optional => 1 },
            topdir         => { type => SCALAR, optional => 1 },
            null           => { type => BOOLEAN, default => 0 },
            non_idempotent => { type => SCALAR,  default => 0,
                                                   regex => qr/^\d+$/ },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options, $regex);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}"   if $opts{file};

    # Check proper arguments
    if ( !$opts{null} and !defined($opts{non_idempotent}) ) {
        croak "Need parameter 'non_idempotent' for " .
              "sending a very large compound request";
    }

    if ( $opts{null} ) {
        # Sending a null compound request
        $options .= " -t 1";
        $regex = qr/Null compound -- went OK/;
    } else {
        # Sending a very large compound request
        $options .= " -t 2 -n $opts{non_idempotent}";
        $regex = qr/Very large compound, \d+ ops\n\t\t\d+ non-idempotent -- went OK/;
    }

    # Create v4t synthetic client command
    $v4t_cmd = "$v4t_path/v4t_cmpd_basic $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_compound_basic',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    if ( $out !~ /$regex/ ) {
        my @err = v4t_get_errors($out);
        croak "Basic limits: ", join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_compound_malformed
## @description
##  Issues a malformed compound request.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_compound_malformed( %options );
##
## @arg operation
##  Required: Operation to issue:
##    invalid_op   => Send a request with invalid op numbers,
##    renamef      => Performs a RENAME,
##    linkf        => Performs a LINK,
##    restorefh    => Performs a RESTOREFH,
##    truncated_op => Send requests that have been truncated.
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Optional: File to perform operation on.
##
## @arg nfile
##  Optional: Target file for operations: 'renamef', 'linkf', or 'restorefh'.
##
## @arg whence
##  Optional: Specifies where the truncation takes place within the opcode:
##  'start', 'middle' or 'end'.
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_compound_malformed(%;) {
    my (%opts) = validate (
        @_,
        {
            file      => { type => SCALAR, optional => 1 },
            topdir    => { type => SCALAR, optional => 1 },
            operation => { type => SCALAR },
            nfile     => { type => SCALAR, optional => 1 },
            whence    => { type => SCALAR, optional => 1 },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};

    # Hash of valid operation flags
    my %valid_ops = ( invalid_op   => ' -i',
                      renamef      => ' -r',
                      linkf        => ' -l',
                      restorefh    => ' -t',
                      truncated_op => ' -s' );

    # Check for valid operation
    if ( ! $valid_ops{$opts{operation}} ) {
        croak "Operation '$opts{operation}' is not supported";
    }

    # Append correct flag for operation
    $options .= $valid_ops{$opts{operation}};

    if ( $opts{operation} eq 'truncated_op' ) {
        # Check whence parameter is provided for this operation
        croak "Need parameter 'whence' for operation 'truncated_op'"
          if ( ! $opts{whence} );
        my %trunc_opts = ( start=>1, middle=>2, end=>3 );

        # Check validity of whence parameter
        croak "Parameter 'whence' must be 'start', 'middle', or 'end'"
          if ( ! $trunc_opts{$opts{whence}} );

        # Append whence option
        $options .= " -p $trunc_opts{$opts{whence}}";
    } elsif ( $opts{operation} =~ /^(renamef|linkf|restorefh)$/ ) {
        # Check first file is provided for these operations
        croak "Need parameter 'file' for operation '$opts{operation}'"
           if ( ! $opts{file} );

        # Check second file is provided for these operations
        croak "Need parameter 'nfile' for operation '$opts{operation}'"
           if ( ! $opts{nfile} );

        # Append the first file
        $options .= " -F $opts{file}";

        # Append the second file
        $options .= " -f $opts{nfile}";
    }

    # Create v4t synthetic client command
    $v4t_cmd = "$v4t_path/v4t_cmpd_mal $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_compound_malformed',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Hash of regular expressions to check for errors according to operation
    my %regex_h = ( invalid_op   => qr/Invalid operation: -?\d+ -- went OK/,
                    renamef      => qr/RENAME operation -- went OK/,
                    linkf        => qr/LINK operation -- went OK/,
                    restorefh    => qr/RESTOREFH operation -- went OK/,
                    truncated_op => qr/Short request[^\n]* -- went OK/ );

    # Check output of v4t synthetic client
    if ( $out !~ /$regex_h{$opts{operation}}/ ) {
        my @err = v4t_get_errors($out);
        croak "Malformed compound request: " . join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_compound_fh
## @description
##  Issues a compound request with current and saved file handles.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_compound_fh( %options );
##
## @arg operation
##  Required: Operation to issue:
##    renamef      => Performs a RENAME,
##    linkf        => Performs a LINK,
##    restorefh    => Performs a RESTOREFH.
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg nfile
##  Required: Target file for operation.
##
## @arg file_attr_dir
##  Optional: Boolean, 'file' is a named attribute directory
##
## @arg file_attr
##  Optional: Boolean, 'file' is a named attribute
##
## @arg nfile_attr_dir
##  Optional: Boolean, 'nfile' is a named attribute directory
##
## @arg nfile_attr
##  Optional: Boolean, 'nfile' is a named attribute
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_compound_fh(%;) {
    my (%opts) = validate (
        @_,
        {
            file      => { type => SCALAR },
            topdir    => { type => SCALAR, optional => 1 },
            operation => { type => SCALAR },
            nfile     => { type => SCALAR },
            file_attr_dir  => { type => SCALAR,  default => 0 },
            file_attr      => { type => SCALAR,  default => 0 },
            nfile_attr_dir => { type => SCALAR,  default => 0 },
            nfile_attr     => { type => SCALAR,  default => 0 },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file} -f $opts{nfile}";
    if ($opts{file_attr_dir}) {
        $options .= " -a 1";
    } elsif ($opts{file_attr}) {
        $options .= " -a 2";
    }
    if ($opts{nfile_attr_dir}) {
        $options .= " -b 1";
    } elsif ($opts{nfile_attr}) {
        $options .= " -b 2";
    }

    # Hash of valid operation flags
    my %valid_ops = ( renamef   => ' -r',
                      linkf     => ' -l',
                      restorefh => ' -t' );

    # Check for valid operation
    if ( ! $valid_ops{$opts{operation}} ) {
        croak "Operation '$opts{operation}' is not supported";
    }

    # Append correct flag for operation
    $options .= $valid_ops{$opts{operation}};

    # Create v4t synthetic client command
    $v4t_cmd = "$v4t_path/v4t_cmpd_fh $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_compound_fh',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Hash of regular expressions to check for errors according to operation
    my %regex_h = ( renamef   => qr/RENAME operation -- went OK/,
                    linkf     => qr/LINK operation -- went OK/,
                    restorefh => qr/RESTOREFH operation -- went OK/ );

    # Check output of v4t synthetic client
    if ( $out !~ /$regex_h{$opts{operation}}/ ) {
        my @err = v4t_get_errors($out);
        croak "Compound request failed: " . join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_compound_transition
## @description
##  Issues a compound request for export regime transition.
##  Performs PUTROOTFH, PUTPUBFH, PUTFH, and RESTOREFH.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_compound_transition( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg nfile
##  Required: Second file.
##
## @arg lookup
##  Optional: Boolean, also perform LOOKUP and LOOKUPP requests,
##  option 'nfile' must be a directory and 'file' must be the parent
##  directory of 'nfile', (default is 0).
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  1 if succeeds or croaks if it fails
#---------------------------------------------------------------------
sub v4t_compound_transition(%;) {
    my (%opts) = validate (
        @_,
        {
            file   => { type => SCALAR },
            topdir => { type => SCALAR, optional => 1 },
            nfile  => { type => SCALAR },
            lookup => { type => BOOLEAN, default => 0 },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);
    # Number of operations including one for PUTFH setup
    my $total_ops = 5;

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file} -f $opts{nfile}";
    if ( $opts{lookup} ) {
        $options .= " -l";
        $total_ops += 2;
    }

    # Create v4t synthetic client command
    $v4t_cmd = "$v4t_path/v4t_cmpd_trans $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_compound_transition',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    # Check output of v4t synthetic client
    my $npass = grep { /operation -- went OK/ } split(/\r?\n/, $out);
    if ( $npass != $total_ops ) {
        my @err = v4t_get_errors($out);
        croak "Compound request failed: " . join("\n", @err);
    }

    return 1;
}

#---------------------------------------------------------------------
## @name ::v4t_compound_allops
## @description
##  Performs all ops on client.
##  Arguments are passed in a "key => value" style.
##
## @synopsis
##  $op = v4t_compound_allops( %options );
##
## @arg ipaddr
##  Required: IP address of filer.
##
## @arg file
##  Required: File to perform operation on.
##
## @arg no_fh
##  Optional: Run all ops with no current file handle,
##  (default is 0).
##
## @arg nattr_dir
##  Optional: Option file is a named attribute directory,
##  (default is 0).
##
## @arg nattr
##  Optional: Option file is a named attribute,
##  (default is 0).
##
## @arg inval_type
##  Optional: Use filehandle with invalid type field,
##  (default is 0).
##
## @arg inval_fsid
##  Optional: Use filehandle with invalid fsid,
##  (default is 0).
##
## @arg inval_inode
##  Optional: Use filehandle with invalid inode,
##  (default is 0).
##
## @arg inval_gen
##  Optional: Use filehandle with invalid generation number,
##  (default is 0).
##
## @arg topdir
##  Optional: Directory where file is located.
##
## @arg user
##  Optional: Perform operation as this user.
##
## @arg client
##  Optional: Hostspec, Hostrec or Connectrec of client.
##  The connection type of the Connectrec object is not used
##  since subtest uses its own connection type.
##  (default is $main::CLIENT).
##
## @arg client_id
##  Optional: Client id given to the v4t synthetic client,
##  (default is the current process id).
##
## @arg timeout
##  Optional: Time in seconds to wait before timing out
##  This option is only used when running remotely.
##  (default is $main::CMDTIMEOUT).
##
## @return
##  Hash of operation error codes or croaks if it fails
#---------------------------------------------------------------------
sub v4t_compound_allops(%;) {
    my (%opts) = validate (
        @_,
        {
            file        => { type => SCALAR },
            topdir      => { type => SCALAR, optional => 1 },
            no_fh       => { type => BOOLEAN, default => 0 },
            nattr_dir   => { type => BOOLEAN, default => 0 },
            nattr       => { type => BOOLEAN, default => 0 },
            inval_type  => { type => BOOLEAN, default => 0 },
            inval_fsid  => { type => BOOLEAN, default => 0 },
            inval_inode => { type => BOOLEAN, default => 0 },
            inval_gen   => { type => BOOLEAN, default => 0 },
            %common_parameters
        }
    );
    my ($v4t_cmd, $out, $options);

    # Create list of options to v4t command
    $options  = "-I $opts{ipaddr} -N $opts{client_id}";
    $options .= " -C $opts{topdir}" if $opts{topdir};
    $options .= " -F $opts{file}";
    $options .= " -n" if $opts{no_fh};
    $options .= " -d" if $opts{nattr_dir};
    $options .= " -a" if $opts{nattr};
    $options .= " -t" if $opts{inval_type};
    $options .= " -f" if $opts{inval_fsid};
    $options .= " -i" if $opts{inval_inode};
    $options .= " -g" if $opts{inval_gen};

    # Create v4t synthetic client command
    $v4t_cmd  = "$v4t_path/v4t_cmpd_allops $options";

    # Run the v4t synthetic client
    $out = v4t_run( v4t_cmd => $v4t_cmd,
                    runid   => 'v4t_compound_allops',
                    client  => $opts{client},
                    timeout => $opts{timeout},
                    user    => $opts{user} );

    my (%ret, $oper, $err);
    my $op_done = 1;
    foreach my $line ( split(/\r?\n/, $out) ) {
        if ( $line =~ /^Step / ) {
            if ( !$op_done and $oper ) {
                # Operation was unsuccessful but no error code found
                $ret{$oper} = "ERROR";
            }
            # Get operation
            ($oper) = ($line =~ /(\S+) operation/);
            $op_done = 0;
        } elsif ( $line =~ /Current fh: .* -- succeeded/ ) {
            # Operation succeeded
            $ret{$oper} = 'NFS4_OK';
            $op_done = 1;
        } elsif ( $line =~ /NFS4ERR/ ) {
            # Found error code
            ($err) = ($line =~ /(NFS4ERR_\S+)/);
            $ret{$oper} = $err;
            $op_done = 1;
        } elsif ( $op_done ) {
            # Ignore rest of operation output
            next;
        }
    }

    return %ret;
}

1;

