#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/nfs/Xdr.pm#2 $
#
## @summary Xdr - Xdr class for encoding/decoding basic data types
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Xdr - Xdr class for encoding/decoding basic data types

=head1 SYNOPSIS

  use nfs::Xdr;

  # Encoding into XDR
  $xdr_obj = nfs::Xdr->new();
  $xdr_obj->uint(encode => 21);
  $xdr_obj->uint64(encode => 87);
  $xdr_obj->opaque(encode => $string, variable => 12);
  # Get encoded message
  $buffer = $xdr_obj->{message};

  # Decoding from XDR
  $xdr_obj = nfs::Xdr->new(message => $buffer);
  $value   = $xdr_obj->uint();
  $value   = $xdr_obj->uint64();
  $string  = $xdr_obj->opaque(variable => 12);

  # Check how many bytes remain on decoding buffer
  $count = $xdr_obj->remain();
  # Get the next four bytes out of the decoding buffer
  $bytes = $xdr_obj->fetch(count => 4);
  # Create a hex dump of current data
  $mesg = $xdr_obj->hex_dump();

=head1 DESCRIPTION

Provides an object for the conversion of basic data types
to and from XDR (eXternal Data Representation).
The basic data types supported by this object are a 32bit
unsigned integer, 64bit unsigned integer and fixed/variable
length opaque (uninterpreted sequence of bytes).

The best way to test this library is by running the tests in nfs/ops.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/Rsierranevada/lib/nfs/Xdr.pm#2 $
  Last revised $Date: 2013/05/30 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package nfs::Xdr;

use strict;
use warnings;

use Carp qw(croak);
use Params::Validate qw(validate SCALAR);

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub new($%);
sub remain($;);
sub fetch($;%);
sub hex_dump($;);
sub uint($;%);
sub uint64($;%);
sub opaque($%);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $uint_max = 4294967296;     # 2^32 for 64bit conversions

#---------------------------------------------------------------------

=head2 new

  $xdr_obj = nfs::Xdr->new([ message => $buffer ]);

Xdr class constructor which initializes the conversion buffer
and current offset.

The following named arguments are accepted:

=over

=item Arguments

=over

=item message

(Optional) : Set conversion buffer for decoding.
Default is an empty buffer for encoding.

=back

=item Returns

XDR Object

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub new($%) {
    my $class = shift;
    my (%opts) = validate (
        @_,
        {
            message => { type => SCALAR, default => '' },
        }
    );

    bless { %opts, offset=>0 }, $class;
}

#---------------------------------------------------------------------

=head2 remain

  $count = $xdr_obj->remain();

Returns the number of bytes remaining on the conversion buffer.

=over

=item Returns

The number of bytes left for decoding.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub remain($;) {
    my $self = shift;
    return (length($self->{message}) - $self->{offset});
}

#---------------------------------------------------------------------

=head2 fetch

  $bytes = $xdr_obj->fetch([ count => $nbytes ]);

Returns a string of bytes from the conversion buffer starting
at the current offset.

The following named arguments are accepted:

=over

=item Arguments

=over

=item count

(Optional) : Number of bytes to fetch (default is to fetch all remaining bytes).

=back

=item Returns

Bytes requested

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub fetch($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            count => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
        }
    );
    my $buf;
    my $count = $opts{count};

    # Check if there is enough bytes left on conversion buffer
    croak "XDR: Not enough bytes left in message"
      if ( $count > $self->remain() );

    # Fetch data
    if ( defined($count) ) {
        # Get count bytes starting at current offset
        $buf = substr($self->{message}, $self->{offset}, $count);
    } else {
        # Get all remaining bytes
        $buf = substr($self->{message}, $self->{offset});
        $count = length($buf);
    }
    # Set message pointer
    $self->{offset} += $count;

    return $buf;
}

#---------------------------------------------------------------------

=head2 hex_dump

  $mesg = $xdr_obj->hex_dump();

Create a hex dump of the current encoding/decoding buffer.
It inserts a space for every four bytes of data.

=over

=item Returns

The hex dump of the current encoding/decoding buffer.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub hex_dump($;) {
    my $self = shift;

    # Create a hex dump of current buffer.
    my $mesg = unpack('H*', $self->{message});
    # Insert a space for every four bytes (8 hex characters) of data.
    $mesg =~ s/(.{8})/$1 /g;

    return $mesg;
}

#---------------------------------------------------------------------

=head2 uint

  [$value] = $xdr_obj->uint([ encode => $value ]);

Decodes/encodes a 32bit unsigned integer. If the optional
'encode' parameter is given, the value is encoded into
a four byte network ordered string and appended to the conversion
buffer. Decoding takes no argument and fetches four bytes out
of the conversion buffer and decodes them into a machine dependent
unsigned integer. Then the offset is incremented by four.

The following named arguments are accepted:

=over

=item Arguments

=over

=item encode

(Optional) : Encode this value into the conversion buffer
(default is to decode a 32bit unsigned integer from the
conversion buffer starting at the current offset).

=back

=item Returns

Nothing on encoding, and a 32bit unsigned integer on decoding.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub uint($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            encode => { type => SCALAR, optional => 1},
        }
    );
    if ( defined($opts{encode}) ) {
        # Convert to a network ordered four byte string
        $self->{message} .= pack('N', $opts{encode});
        return;
    } else {
        # Fetch four bytes and convert them to a machine
        # dependent unsigned integer
        return unpack('N', $self->fetch(count=>4));
    }
}

#---------------------------------------------------------------------

=head2 uint64

  [$value] = $xdr_obj->uint64([ encode => $value ]);

Decodes/encodes a 64bit unsigned integer. If the optional
'encode' parameter is given, the value is encoded into
an eight byte network ordered string and appended to the conversion
buffer. Decoding takes no argument and fetches eight bytes out
of the conversion buffer and decodes them into a machine dependent
64bit unsigned integer. Then the offset is incremented by eight.

The following named arguments are accepted:

=over

=item Arguments

=over

=item encode

(Optional) : Encode this value into the conversion buffer
(default is to decode a 64bit unsigned integer from the
conversion buffer starting at the current offset).

=back

=item Returns

Nothing on encoding, and a 64bit unsigned integer on decoding.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub uint64($;%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            encode => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
        }
    );
    # Use of '>>' does not work for integers >= 2^32
    # so using brute force approach for encoding/decoding
    if ( defined($opts{encode}) ) {
        # Convert to a network ordered eight byte string
        my $low  = $opts{encode} % $uint_max;
        my $high = int($opts{encode} / $uint_max);
        $self->{message} .= pack('NN', $high, $low);
        return;
    } else {
        # Fetch eight bytes and convert them to a machine
        # dependent 64bit unsigned integer
        my ($high, $low) = unpack('NN', $self->fetch(count=>8));
        return ($high*$uint_max + $low);
    }
}

#---------------------------------------------------------------------

=head2 opaque

  [$string] = $xdr_obj->opaque([ encode => $string ]
                               [, fixed|variable => $nbytes ]
                               [, varlen => $varlen ]);

Decodes/encodes a fixed/variable length opaque.
An opaque is an uninterpreted sequence of bytes which could have
either a fixed or a variable length. A fixed length opaque is
encoded by appending exactly 'fixed' number of bytes to the
conversion buffer.  Conversely, a variable length opaque is
encoded by appending a four byte count followed by the string.
Both types are padded into a four byte boundary.

The following named arguments are accepted:

=over

=item Arguments

=over

=item encode

(Optional) : Encode a fixed/variable length opaque
(default is to decode a fixed/variable opaque from the
conversion buffer starting at the current offset).

=item fixed

(Optional) : Number of bytes to encode/decode for a fixed length opaque.

=item variable

(Optional) : Maximum number of bytes to encode/decode
for a variable length opaque. A value of zero means that
there is no limit (default).

=item varlen

(Optional) : Replace the length of a variable length opaque with
this number.

=back

=item Returns

Nothing on encoding, and an opaque string on decoding.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub opaque($%) {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            encode   => { type => SCALAR, optional => 1 },
            fixed    => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
            variable => { type => SCALAR,  default => 0, regex => qr/^\d+$/ },
            varlen   => { type => SCALAR, optional => 1, regex => qr/^\d+$/ },
        }
    );
    my $pad;
    my $nbytes = $opts{fixed};

    if ( defined($opts{fixed}) and $opts{fixed} < 1 ) {
        croak "XDR: Fixed length opaque must have a length greater than zero";
    }

    if ( defined($opts{encode}) ) {
        # Encode opaque
        my $len = length($opts{encode});
        if ( $opts{fixed} ) {
            croak "XDR Encode: $len exceeds fixed size $opts{fixed}"
              if ( $len > $opts{fixed} );
            # Pad up to the fixed length
            # if string is not long enough
            $pad = $opts{fixed} - $len;
        } else {
            croak "XDR Encode: $len exceeds maximum size $opts{variable}"
              if ( $opts{variable} and $len > $opts{variable} );
            # Encode length of opaque
            $nbytes = defined($opts{varlen}) ? $opts{varlen} : $len;
            $self->uint(encode=>$nbytes);
        }

        # Append string to conversion buffer as is
        $self->{message} .= $opts{encode};

        # Find if string is not a multiple of four bytes
        my $pad4 = $nbytes & 3;
        if ( $pad or $pad4 ) {
            # Padding is needed
            $pad += (4 - $pad4) if $pad4;
            $self->{message} .= pack("a$pad", "");
        }
        return;
    }

    # Decode opaque
    if ( ! $nbytes ) {
        # Get size for variable length opaque
        $nbytes = $self->uint();

        croak "XDR Decode: $nbytes exceeds maximum size $opts{variable}"
          if ( $opts{variable} and $nbytes > $opts{variable} );
    }

    # Fetch data
    my $buf = $self->fetch(count=>$nbytes);

    # Strip padded zeroes
    $pad = $nbytes & 3;
    if ( $pad ) {
        $self->fetch(count=>4 - $pad);
    }

    return $buf;
}

1;

