## @summary NFS /etc/exports management library
## @status review
## @author  dl-maryv (FSS Protocols QA)
package nfs::exports;

use strict;
use File::Basename;

use Tharn;
use nfs::rshlib;
use nfs::say;

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&exports_checkvol
	     &exports_add &exports_del &exports_mod &exports_exportfs
	     &exports_check &exports_volstatus
	     &exports_mksecs);

sub exports_volstatus($$;) {
    my $filer = shift();
    my $vol = File::Basename::basename(shift());
    my %status = ();

    my $output = join('', rsh($filer, "vol status $vol"));
    my ($vol_stat) = ($output =~ /($vol?.+\S+)/g);

    logcomment("vol_stat=$vol_stat");

    if (! $vol_stat) {
	xlogresult('WARN', 0, "no such volume $vol");
	return ();
    }

    my @values = split(/\s+/, $vol_stat);
    if ($values[1] =~ /TRAD|FLEX/) {
	($status{volume}, $status{type}, $status{state},
	 $status{status}) = @values;
    } else {
	($status{volume},$status{state},$status{status}) = @values;
    }
    if (! $status{volume}) {
	xlogresult('WARN', 0, "invalid status for volume $vol");
	return ();
    }

    return %status;
}

## @name ::exports_checkvol
## @summary Check to see if a volume exists and is online
## @description See above
## @synopsis
##     use nfs::exports;
##
##     exports_checkvol($server, $vol);
##
## @arg $filer Required:  hostname of filer.
##
## @arg $vol Required:  name of volume, i.e. vol1.
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    param('FILER','-mesg',"hostspec for a filer");
##    param('FILER_VOL', 'vol0');
##    my $h = host($FILER);
##
##    if (! exports_checkvol($h->{hostname}, $FILER_VOL)) {
##       logcomment("no such volume $FILER_VOL");
##    }
##
## @dependencies
##   rsh access to filer
##
sub exports_checkvol($$;) {
    my($filer, $filer_vol) = @_;
    my %status = exports_volstatus($filer, $filer_vol);
    foreach my $k (keys(%status)) {
	xlogresult('WARN', 0, "k=$k, v=$status{$k}");
    }
    xlogresult('WARN', 0, "vol status for $filer_vol=$status{status}");
    if (! defined($status{state}) || ! defined($status{status})) {
	xlogresult('WARN', 0, "couldn't get status for $filer_vol");
	return 0;
    } elsif ($status{state} eq 'online' &&
 	     ($status{status} eq 'normal' ||	      
 	      $status{status} =~ /raid[0-9]+|/)) {
	return 1;
    }
    xlogresult('WARN', 0, "volume $filer_vol doesn't exist");
    return 0;
}

## @name ::exports_add
## @summary Add or replace a filer's export.
## @description See above.
## @synopsis
##    use nfs::exports;
##
##    if (! exports_add('u89', "/vol/vol1 -anon=0")) {
##       logcomment("couldn't add export");
##    }
##
## @arg $filer Required:  hostname of filer.
## @arg $export Required:  full export entry.
## @arg %opts, Required:  an associative array of options.
##             append=>[0|1] just append, don't replace. False, usually.
##             children=>[0|1], non-zero means "override all children." False.
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    param('FILER','-mesg',"hostspec for a filer");
##    param('FILER_VOL', '-default', 'vol0');
##    param('FILER_VOL_MNT', '-default', "/vol/$FILER_VOL");
##    my $h = host($FILER);
##
##    if (! exports_add($h->{hostname}, "$FILER_VOL_MNT -sec=krb5")) {
##       logcomment("couldn't add export");
##    }
##
## @dependencies
##   rsh access to filer
##
sub exports_add($$@;) {
    my ($filer, $exp, @opts) = @_;
    my (@settings) = split(/\s+/, $exp);
    my $vol = $settings[0];
    my (%opts) = @opts;
    my (@cur) = rsh($filer, "rdfile /etc/exports");
    my (@repl) = ();

    if (! $opts{append}) {
	my $replaced = 0;
	foreach my $line (@cur) {
	    next if ($opts{children} && $line =~ m{$exp});
	    my(@a) = split(/\s+/, $line);
	    if ($a[0] eq $vol) {
		push(@repl, join(' ', @settings) . "\n");
		$replaced++;
	    } else {
		push(@repl, $line);
	    }
	}
	push(@repl, "$exp\n") if not $replaced;
    } else {
	push(@repl, "$exp\n");
    }

    if (! rsh_write($filer, "wrfile /etc/exports", @repl)) {
	xlogcomment("rsh_write() of new exports failed to $filer");
	return 0;
    } elsif (! rsh($filer, "exportfs -v $vol")) {
	xlogcomment("rsh() to reset exports failed to $filer");
	return 0;
    }

    return 1;
}

## @name ::exports_del
## @summary Delete a particular export (and optionally all of its children)
## @description See above.
## @synopsis
##    use nfs::exports;
##
##    if (! exports_add('u89', "/vol/vol1 -anon=0")) {
##       logcomment("couldn't add export");
##    }
##
## @arg $filer Required:  hostname of filer.
## @arg $export Required:  full export entry. Can't be a regexp.
## @arg %opts, Required:  an associative array of options.
##             children=>[0|1], non-zero means "override all children."
##             Default: false.
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    param('FILER','-mesg',"hostspec for a filer");
##    param('FILER_VOL', '-default', 'vol0');
##    param('FILER_VOL_MNT', '-default', "/vol/$FILER_VOL");
##    my $h = host($FILER);
##
##    if (! exports_del($h->{hostname}, $FILER_VOL_MNT)) {
##       logcomment("couldn't remove export");
##    }
##
## @dependencies
##   rsh access to filer
##
sub exports_del($$@;) {
    my ($filer, $exp, @opts) = @_;
    my (%opts) = @opts;
    my (@cur) = rsh($filer, "rdfile /etc/exports");
    my (@repl) = ();

    foreach my $line (@cur) {
	next if (index($line, $exp, 0) == 0 && $opts{children});
	my (@tokens) = split(/\s+/, $line);
	next if ($tokens[0] eq $exp);
	push(@repl, $line);
    }

    rsh_write($filer, "wrfile /etc/exports", @repl);

    rsh($filer, "exportfs -vu $exp");
}

## @name ::exports_mod
## @summary Modify a given export.
## @synopsis
##    use nfs::exports;
##
##    if (! exports_mod('u89', "/vol/vol1", "/vol/vol1 -anon=0")) {
##       logcomment("couldn't modify export");
##    }
##
## @arg $filer Required:  hostname of filer.
## @arg $export Required: volume that's currently exported (or to
##      add). Can be regexp.
## @arg $change Required:  new export, fully specified.
##
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    param('FILER','-mesg',"hostspec for a filer");
##    param('FILER_VOL', '-default', 'vol0');
##    param('FILER_VOL_MNT', '-default', "/vol/$FILER_VOL");
##    my $h = host($FILER);
##
##    if (! exports_mod($h->{hostname}, "$FILER_VOL_MNT",
##                      "$FILER_VOL_MNT -sec=krb5")) {
##       logcomment("couldn't modify export");
##    }
##
## @dependencies
##   rsh access to filer
##
sub exports_mod($$$@;) {
    my ($filer, $exp, $change) = @_;
    my (@cur) = rsh($filer, "rdfile /etc/exports");
    my (@repl) = ();

    foreach my $line (@cur) {
	my(@a) = split(/\s+/, $line);
	if ($a[0] eq $exp) {
	    $line = sprintf("%s\n", $change);
	}
	push(@repl, $line);
    }

    rsh_write($filer, "wrfile /etc/exports", @repl);
}


## @name ::exports_exportfs
## @summary Export a given volume path.
## @synopsis
##    use nfs::exports;
##
##    if (! exports_exportfs('u89', "/vol/vol1")) {
##       logcomment("couldn't export volume");
##    }
##
## @arg $filer Required:  hostname of filer.
## @arg $export Required:  volume to export.
## @arg $opts Required:  export options such as -sec=krb5, overriding what exists.
##
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    param('FILER','-mesg',"hostspec for a filer");
##    param('FILER_VOL', '-default', 'vol0');
##    param('FILER_VOL_MNT', '-default', "/vol/$FILER_VOL");
##    my $h = host($FILER);
##
##    if (! exports_exportfs($h->{hostname}, "$FILER_VOL_MNT", "-sec=krb5")) {
##       logcomment("couldn't export volume");
##    }
##
## @dependencies
##   rsh access to filer
##
sub exports_exportfs($@;) {
    my ($filer, $exp, $opts) = @_;

    if (! $exp) { $exp = ""; }
    if (! $opts) { $opts = ""; }

    return join('', rsh($filer, "exportfs -v $opts $exp"));
}

## @name ::exports_check
## @summary Check a given export for some expected values.
## @synopsis
##    use nfs::exports;
##
##    if (! exports_check('u89', "/vol/vol1", "sec=krb5")) {
##       logcomment("volume isn't exported as expected");
##    }
##
## @arg $filer Required:  hostname of filer.
## @arg $export Required:  volume to export.
## @arg $xvals Required:  expected values.
##
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @dependencies
##   rsh access to filer
##
sub exports_check($$$;) {
    my ($filer, $vol, $xval) = @_;

    return grep(m{$vol\s+.*$xval}, rsh($filer, "rdfile /etc/exports")) > 0;
}

## @name ::exports_mksecs
## @summary Provide a combined export value.
## @synopsis
##    use nfs::exports;
##
##    my $var = exports_mksecs($val);
##
## @arg $client Required:  how the client is specifying sec=.
## @arg $filer Required:  filer's required sec= combination.
##
## @usage
##    Upon failure, 0 is returned. Success returns non-zero.
##
## @example
##    use nfs::exports;
##
##    my $secopt = exports_mksecs('krb5p');
##
## @dependencies
##   KDC (possibly).
##
sub exports_mksecs($;) {
    my $client = shift();
    my $filer = 'sec=none';

    if ($client eq 'krb5')  { $filer = 'sec=sys:krb5'; }
    if ($client eq 'krb5i') { $filer = 'sec=sys:krb5:krb5i'; }
    if ($client eq 'krb5p') { $filer = 'sec=sys:krb5:krb5i:krb5p'; }
    if ($client eq 'sys')   { $filer = 'sec=sys'; }
    if ($client eq 'none')  { $filer = 'sec=none'; }

    xlogcomment("FILER_X = $filer");

    return $filer;
}

1;
