## @summary NFS Kerberos V library.
## @status review
## @description
## Copy a krb5.keytab and krb5.conf file to a filer and configure it
## for MIT Kerberos. TODO: Active Directory. By Stephen Wynne November 2002.
## @author  dl-maryv (FSS Protocols QA)
package nfs::krb5;

use strict;
use nfs::say;

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&krb5_fqn &krb5_mit_filer_setup &krb5_kinit &krb5_kdestroy);

use Tharn;
use mount;
use nfs::say;

# Fully qualified hostnames.
sub krb5_fqn($;) {
    my $name = shift;
    my $n = undef;

    foreach my $ext ("", ".sim") {
	chomp($n = `ypmatch ${name}${ext} hosts.byname 2> /dev/null`);
	last if (length($n));
    }

    $n =~ s/^[^\s]+\s+//;
    return $n;
}

# From Lincon's standby
sub make_conn (@;) {
    my ($spec, $conntype) = @_;
    if ($conntype) {
	return connect($spec->{id} . $conntype);
    } elsif (ref($spec) =~ /Connectrec/) {
	return $spec;
    } elsif (ref($spec) =~ /Hostrec/) {
	return connect($spec);
    }
}

## @name ::krb5_mit_filer_setup
## @summary Sets up an MIT Kerberos filer.
##
## @description
##   Uses keys in /u/wynne/nfsv4/k
##   See options:
##     nfs.kerberos.file_keytab.enable off
##     nfs.kerberos.principal
##     nfs.kerberos.realm
##
## @synopsis
##
## use nfs::krb5;
##
## krb5_mit_filer_setup($filer, $client, $realm, $keydir);
##
## @arg $filer Required:  filer hostname
## @arg $client Required:  client hostname
## @arg $realm Required:  Kerberos realm.
## @arg $keydir Required:  Where the keys are stored.
##
## @usage Returns undef.
## 
## @example
##
## use nfs::krb5;
##
## krb5_mit_filer_setup('u89', 'c37', 'NFSV4.LAB.NETAPP.COM',
##                      '/u/wynne/nfsv4/k');
##
## @dependencies rsh to filer, rlogin to client.
##
sub krb5_mit_filer_setup($$$$;) {
    my ($filer, $client, $realm, $keydir) = @_;
    my $fhost = host($filer);
    my $fhostname = $fhost->{hostname}; $fhostname =~ s/\..*//;
    my $princ = krb5_fqn($fhostname);
    my $ktab = "krb5.keytab-$realm";
    my $kconf = "krb5.conf-$realm";
    my $chost = host($client);
    my $fconn = make_conn($fhost, '#rsh') or die("$@");
    my $cconn = make_conn($chost, '#rlogin') or die("$@");

    if (! -d $keydir) {
	die("no such key directory $keydir\n");
    } elsif (! -d "$keydir/$fhostname") {
	die("no such host-specific key directory $keydir/$fhostname\n");
    } elsif (! -f "$keydir/$kconf") {
	die("no such realm-specific conf file  $keydir/$kconf\n");
    } elsif (! -f "$keydir/$fhostname/mit/$ktab") {
	die("no such host-specific key file $keydir/$fhostname/mit/$ktab\n");
    } else {
	$ktab = "$keydir/$fhostname/mit/$ktab";
	$kconf = "$keydir/$kconf";

	foreach my $tab ($kconf, $ktab) {
	    say($cconn, "test -f $tab && echo \$0", '-wait', "0",
		'-timeout', 40, '-fatal', "Couldn't read $tab on client!");
	}
    }

    my $mnt = mount($fhost->{default_ip},
		    '/', client => $client, vers => 3) or die("$@");

    foreach my $tab ($kconf, $ktab) {
	my $targ = $tab =~ /keytab/ ? 'krb5.keytab' : 'krb5.conf';
	say($cconn, "sudo cp $tab $mnt/etc/$targ && echo \$?",
	    '-wait', "0", '-timeout', 40,
	    '-fatal', "Couldn't cp $tab to $mnt/etc on client!");
    }

    umount($mnt, client => $client) or die("$@");

    say($fconn, "priv set admin");
    say($fconn, "options nfs.kerberos.file_keytab.enable on");
    say($fconn, "options nfs.kerberos.principal $princ");
    say($fconn, "options nfs.kerberos.realm $realm");
}

## @name ::krb5_kinit
## @summary acquire Kerberos tokens on a UNIX client.
## @synopsis
##
## use nfs::krb5;
##
## krb5_kinit($client, $kuser, $kpass, $krealm);
##
## @arg $client Required:  client to use for obtaining tokens
## @arg $kuser Required:  kerberos principal
## @arg $kpass Required:  kerberos password
## @arg $krealm Required:  kerberos realm
##
## @usage Returns true on success.
##
## @example
##
## use nfs::krb5;
##
## krb5_kinit('c37', 'testu1', 'testp1', 'NFSV4.LAB.NETAPP.COM') or
##     logcomment("couldn't kinit in realm");
##
## @dependencies rsh to filer, rlogin to client.
##
sub krb5_kinit($$$$;) {
    my $client = shift;
    my $kuser = shift;
    my $kpass = shift;
    my $krealm = shift || "";

    my $ch = host($client);
    my $cc = make_conn($ch, '#rlogin') or die("$@");

    my (@o) = say($cc, "kinit $kuser\@$krealm",
		     '-prompt', "Password for $kuser\@.*: ");
    (@o) = say($cc, "$kpass");

    if ($o[2] ne "\r\n") {
	map(s/[\r\n]+//g, @o);
	xlogresult('WARN', 0, "couldn't kinit: >>>>@o<<<<");
	return 0;
    }

    return 1;
}

## @name ::krb5_kdestroy
## @summary destroy Kerberos tokens
## @synopsis
##
## use nfs::krb5;
##
## krb5_destroy($client);
##
## @arg $client Required:  client to use for obtaining tokens
## @arg $kuser Required:  kerberos principal
## @arg $kpass Required:  kerberos password
## @arg $krealm Required:  kerberos realm
##
## @usage Returns true on success.
##
## @example
##
## use nfs::krb5;
##
## krb5_kdestory('c37') or logcomment("couldn't destroy any tokens");
##
## @dependencies rsh to filer, rlogin to client.
##
## @bugs needs to specify a user for which to destroy tokens.
##
sub krb5_kdestroy($;) {
    my $client = shift;

    my $ch = host($client);
    my $cc = make_conn($ch, '#rlogin') or die("$@");

    if (! unix_say($cc, "kdestroy -q")) {
	xlogcomment("couldn't destroy tokens");
	return undef;
    }

    return 1;
}
