## @summary OOP Adapter for the lib/mount.pm library.
## @description Provides an object-oriented interface to NAS mounts.
## @dependencies variant of lib/mount.pm
## @author  dl-maryv (FSS Protocols QA)
package nfs::mnt;

use strict;
use Tharn;
use POSIX;
use nfs::say;
use mount;
use nfs::rshlib;
use nfs::krb5;

use vars qw(@ISA @EXPORT);
use vars qw(%mounts);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&mnt_rrmdir &mnt_mkmntpt);

use constant baseopts => 'intr,hard,noquota,bg';

%mounts = ();

##############################################################################

# Make a short random number.
sub mktag() {
    return sprintf("%s.%3.3d", 'mnt', rand(10000));
}

# Make a stamp
sub mkstamp() {
    return time();
}

# Set either our base options, or our general and extra mount options.
sub setopt($$;) {
    my $self = shift;
    my $opt = shift;
    my $val = shift;
    my (@mntopts);
    my $done = 0;

    return if (! $opt);		# usage: nothing to do

    # Make sure we replace anything in baseopts that's changed.
    # Note: can also set baseopts, but user might not know which is what
    # type of option.
    foreach my $k (split(/,/, $self->{baseopts})) {
	my($k, $v) = split(/=/, $k);
	if ($opt eq $k or $opt =~ /.+$k$/ or $k =~ /.+$opt$/) {
	    # replace base options, possibly <no>opt or <un>opt, etc...
	    if ($val) {
		push(@mntopts, "$opt=$val");
	    } else {
		push(@mntopts, "$opt");
	    }
	    $done++;		# only one should ever be found
	} else {
	    push(@mntopts, $k);
	}
    }

    if ($done) {
	$self->{baseopts} = join(',', @mntopts);
	return;
    }

    # Handle general options that match $opt, or extra mount options
    foreach my $hash ('mntopts', 'genopts') {
	foreach my $k (keys(%{$self->{$hash}})) {
	    if ($opt eq $k or $opt =~ /.+$k$/ or $k =~ /.+$opt$/) {
		delete($self->{$hash}->{$k});
		$self->{$hash}->{$opt} = $val;
		$done++;
		last;		# only one should ever be found
	    }
	}
    }

    if ($opt eq 'path' or $opt eq 'server') {
	$self->{$opt} = $val;	# a couple of hacks
	$done++;
    }

    if (! $done) {
	xlogcomment("didn't find any valid options to change for $opt" .
		   ($val ? "=$val" : ""))
    }
}

# Remove from either our base options, or our general and extra mount options.
# This time we don't touch values set on the instance itself, i.e. server/path.
sub rmopt($$;) {
    my $self = shift;
    my $opt = shift;
    my (@mntopts);
    my $done = 0;

    return if (! $opt);		# usage: nothing to do

    # Make sure we replace anything in baseopts that's changed.
    # Note: can also set baseopts, but user might not know which is what
    # type of option.
    foreach my $k (split(/,/, $self->{baseopts})) {
	my($k, $v) = split(/=/, $k);
	if ($opt eq $k or $opt =~ /.+$k$/ or $k =~ /.+$opt$/) {
	    $done++;		# Skip. Only one should be found.
	} else {
	    push(@mntopts, $k);
	}
    }

    if ($done) {
	$self->{baseopts} = join(',', @mntopts);
	return;
    }

    # Handle general options that match $opt, or extra mount options
    foreach my $hash ('mntopts', 'genopts') {
	foreach my $k (keys(%{$self->{$hash}})) {
	    if ($opt eq $k or $opt =~ /.+$k$/ or $k =~ /.+$opt$/) {
		delete($self->{$hash}->{$k});
		$done++;
		last;		# only one should ever be found
	    }
	}
    }

    if (! $done) {
	xlogcomment("didn't find any valid options to change for $opt");
    }
}

sub mopts() {
    my $self = shift;
    my(@mntopts);

    foreach my $k (split(/,/, $self->{baseopts})) {
	# Let mntopts override baseopts.
	next if grep(/^$k$/, (keys(%{$self->{mntopts}})));
	push(@mntopts, $k);
    }

    foreach my $k (keys(%{$self->{mntopts}})) {
	if (defined($self->{mntopts}->{$k})) {
	    push(@mntopts, "$k=" . $self->{mntopts}->{$k});
	} else {
	    push(@mntopts, $k);
	}
    }

    return join(',', @mntopts);
}

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};

    my ($server, $path, $genopts, $mntopts) = @_;

    if (! $genopts) { $genopts = {}; }
    if (! $mntopts) { $mntopts = {}; }

    # Internal
    $self->{pid} = $$;		# can compare in DESTROY to avoid fork damage
    $self->{hostname} = (POSIX::uname())[1]; # check locality in mnt_rrmdir

    # Filer resources
    $self->{server} = simname($server);	# NFS server to mount
    $self->{path} = $path;	# export path on the server

    %{$self->{genopts}} = %{$genopts}; # random settings, not mnt-related
    %{$self->{mntopts}} = %{$mntopts}; # explicit settings.
    # our base mount options, also set in mount.pm
    $self->{baseopts} =	$genopts->{baseopts} || baseopts();

    # Actions
    $self->{do_rmdir} = 1;	# yes, do remove mount directory afterwards.
    $self->{do_umount} = 1;	# yes, do unmount afterwards.
    $self->{do_newstamps} = 1;	# yes, create new stamps/tags each mount

    # Internal state
    $self->{mount} = undef;	# the mount point on the client
    $self->{ismounted} = 0;	# isn't it mounted yet/still?

    bless ($self, $class);

    return $self;
}

sub str() {
    my $self = shift;

    my $server = $self->{server} ? $self->{server} : "(nohost)";
    my $path = $self->{path} ? $self->{path} : "(/no/such/vol)";
    my $client = $self->{genopts}->{client} ? $self->{genopts}->{client} :
      "(noclient)";
    my $mount = $self->{mount} ? $self->{mount} : "(/unknown/mnt)";

    return "$server:$path @ $client:$mount";
}

sub mnt {
    my $self = shift;

    return eval {

	$self->{genopts}->{tag} = mktag() if not $self->{genopts}->{tag};
	$self->{genopts}->{stamp} = mkstamp() if $self->{genopts}->{stamp};
	$self->{genopts}->{mountopts} = $self->mopts();

	# Possible resettings.
	foreach my $k ('path', 'server') {
	    if ($self->{genopts}->{$k} &&
		$self->{genopts}->{$k} ne $self->{$k}) {
		if ($k eq 'path') {
		    $self->{genopts}->{mountponint} = undef;
		}
		$self->{$k} = $self->{genopts}->{$k};
	    }
	}

	if (! $self->{genopts}->{mountpoint}) {
	    $self->{genopts}->{mountpoint} =
	      mnt_mkmntpt(sprintf("/t/%s", $self->{server}),
			  {
			   server => $self->{server},
			   xpoint => $self->{path},
			   proto => $self->{mntopts}->{proto},
			   vers => $self->{mntopts}->{vers},
			   sec => $self->{mntopts}->{sec},
			   stamp => $self->{genopts}->{stamp},
			   tag => $self->{genopts}->{tag},
			  });
	}

	if (! ($self->{mount} = mount($self->{server}, $self->{path},
				       %{$self->{genopts}}))) {
	    die("mount didn't succeed: $@");
	}

	xlogcomment("mntopts=" . $self->mopts());

	$self->{ismounted}++;

	return $self->{mount};
    };
}

sub toggle_umount() {
    my $self = shift;
    $self->{do_umount} =! $self->{do_umount};
}

sub toggle_rmdir() {
    my $self = shift;
    $self->{do_rmdir} =! $self->{do_rmdir};
}

sub toggle_newstamps() {
    my $self = shift;
    $self->{do_newstamps} =! $self->{do_newstamps};
}

sub mountpoint() {
    my $self = shift;

    return $self->{genopts}->{mountpoint};
}

sub umnt {
    my $self = shift;

    return eval {
	if ($self->{hostname} ne (POSIX::uname())[1]) {
	    xlogcomment("skipping umount on non-initiating node");
	    return 1;
	} elsif ($self->{mount}) {
	    if (umount($self->{mount}, client=>$self->{genopts}->{client})) {
		if ($self->{do_rmdir}) {
		    if (mnt_rrmdir($self->{mount}, $self->{genopts}->{client},
				   $self->{hostname})) {
			$self->{mountpoint} = $self->{mount} = undef;
		    }
		}
		$self->{ismounted} = undef;
		return 1;
	    } else {
		die("unmount didn't succeed: $@");
	    }
	} else {
	    return 1;		# quiet default when nothing to unmount
	}
    };
}

sub remnt($;) {
    my $self = shift;

    $self->umnt();
    return $self->mnt();
}

sub DESTROY {
    my $self = shift;

    if ($self->{pid} != $$) {
	return;			# this is a forked child
    }

    if ($self->{do_umount} && $self->{ismounted} && $self->{mount}) {
	$self->umnt() or
	  xlogcomment("unmount of mountpoint $self->{mount} failed on " .
		      "$self->{genopts}->{client}");
    }
}

sub mnt_rrmdir($$$;) {
    return eval {
	my $dir = shift;
	my $client = shift;
	my $host = shift;

	if ($host && $host ne (POSIX::uname())[1]) {
	    xlogcomment("skipping removal on non-initiating node XXX");
	    return 1;
	}

	if ($client) {
	    # XXX umount -f not universal
	    my $sobj = Tharn::subtest("{ system(\"sudo rmdir $dir\") == 0
                                    or tstlog('-fatal', \$@) }",
				      '-client', $client,
				      '-onexit', sub {}, '--', client => "");
	    my $result = Rslt::worst($sobj->rslts());
	    if ($result && $result->severity() >
                           NATE::Result::Warn->severity()) {
		die join("\n", $result->msgs());
	    }
	    return 1;
	} else {
	    my $cmd = "sudo rmdir $dir";
	    xlogcomment("cmd=$cmd");
	    my $text = `$cmd 2>&1`;
	    $text =~ s/^\s+//;
	    $text =~ s/\s+$//;
	    die "$text\n" if $?;
	    return 1;
	}
    };
}

# Flatten out /path/separators/into underbars, and remove leading slash.
sub flatten($;) {
    my $x = shift;
    $x =~ s,/,_,g;
    $x =~ s/^_//;
    return $x;
}

sub mnt_mkmntpt {
    my $base = shift || "/t/unknown";
    my $o = shift || {};

    # Some random defaults, which without specifics, will produce
    # something like: /t/nohost/no_vol-noproto-0-sys-1046852667-mnt.7205

    if (! $$o{where})    { $$o{where}    = '/t/%s/%s';    }
    if (! $$o{server})   { $$o{server}   = 'nohost';      }
    if (! $$o{xpoint})   { $$o{xpoint}   = '/no/vol';     }
    if (! $$o{proto})    { $$o{proto}    = 'noproto';     }
    if (! $$o{vers})     { $$o{vers}     = 0;             }
    if (! $$o{sec})      { $$o{sec}      = 'sys';         }
    if (! $$o{stamp})    { $$o{stamp}    = mkstamp(),     }
    if (! $$o{tag})      { $$o{tag}      = mktag();       }

    # making sure to escape any slashes

    my $extra = flatten(sprintf("%s-%s-%d-%s-%d-%s",
				flatten($$o{xpoint}),
				$$o{proto},
				$$o{vers},
				$$o{sec},
				$$o{stamp},
				$$o{tag}));

    # Example: /t/u57/vol_vol1-3-sys-udp-intr,noquota-1046426598.101001
    return sprintf($$o{where}, $$o{server}, $extra);
}

1;
