## @summary Complements for lib/mount.pm
## @status review
## @author  dl-maryv (FSS Protocols QA)
package Xnfs::mntlibX;

use strict;
use Tharn;
use nfs::say;
use mount;

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&mntlib_rrmdir &mntlib_mkmntpt);

## @name ::mntlib_rrmdir
## @summary remove a (possibly) remote directory
## @synopsis
##   use nfs::mntlib;
##   mntlib_rrmdir($dir, $client);
## @arg $dir Required: (possibly) remote directory.
## @arg $client Client.
## Borrowed remote scripting from Lincoln's mount.pm
sub mntlib_rrmdir($$;) {
    return eval {
	my $dir = shift;
	my $client = shift;
	if ($client) {
	    my $sobj = Tharn::subtest("{ system(\"sudo rmdir $dir\") == 0  \
                                    or tstlog('-fatal', \$@) }",
				      '-client',$client,
				      '-onexit',sub {},'--',client=>"");
	    my $result = Rslt::worst($sobj->rslts());
	    if ($result && $result->severity() >
                           NATE::Result::Warn->severity()) {
		die join("\n", $result->msgs());
	    }
	    return 1;
	} else {
	    my $cmd = "sudo rmdir $dir";
	    my $text = `$cmd 2>&1`;
	    $text =~ s/^\s+//;
	    $text =~ s/\s+$//;
	    die "$text\n" if $?;
	    return 1;
	}
    };
}

## @name ::mntlib_mkmntpt
## @summary construct a name for a mount point
## @description
##   Produce a name from the following components:
##       <ol>
##        <li> where - template for mount point name: /t/%s/%s
##        <li> xpoint - export point, such as /vol/vol0
##        <li> proto - protocol, such as udp or tcp
##        <li> vers - NFS version number, such as 4.
##        <li> mountopts - comma-separated options, i.e. intr,hard,sec=sys.
##        <li> tag - a base name of the user's choosing.
##       </ol>
## A uniquifier, normally the date in seconds since 1970
## will be added. Reasonable choices will also be provided
## for you in their absence.
## @synopsis
##   use nfs::mntlib;
##   mntlib_mkmntpt($base, \%opts);
## @arg $base Required: your base mount point.
## @arg \%opts Required: your options.
## @example
##   use nfs::mntlib;
##   %opts = (xpoint => $FILER_VOL_MNT, proto=>$NFSPROTO, vers=>$NFSVERS,
##            mountopts=>$MNTOPTS);
##   $mnt = mntlib_mkmntpt($base, \%opts); # /
## @dependencies
## @bugs
sub flatten($;) { my $x = shift; $x =~ s,/,_,g; $x =~ s/^_//; return $x; }
sub mktag() { return sprintf("%s.%3.3d", 'mntlib', rand(10000)); }

sub mntlib_mkmntpt {
    my $base = shift;
    my %opts;
    my $o = shift || \%opts;		# options

    if (! $$o{where})    { $$o{where}    = '/t/%s/%s'; }
    if (! $$o{server})   { $$o{server}   = 'nohost'; }
    if (! $$o{xpoint})   { $$o{xpoint}   = '/no/vol'; }
    if (! $$o{proto})    { $$o{proto}    = 'noproto'; }
    if (! $$o{vers})     { $$o{vers}     = 0; }
    if (! $$o{mntopts})  { $$o{mntopts}  = 'no,mnt,opts'; }
    if (! $$o{tag})      { $$o{tag}      = mktag(); }

    ## want: vol_vol1-3-udp-intr,noquota-1046426598.101001
    my $extra = flatten(sprintf("%s-%s-%d-%s-%d-%s",
			flatten($$o{xpoint}),
			$$o{proto},
			$$o{vers},
			$$o{mntopts},
			time(),
			$$o{tag}));

    ## Returning /t/u57/vol_vol1-3-udp-intr,noquota-1046426598.101001
    return sprintf($$o{where}, $$o{server}, $extra);
}

1;
