# -*- perl -*-
#
## @summary
## Remote quota is the ability to query the quota of the file system.
## All these testcases are for C-mode setup. Works with rQuotaClient.pm(synthetic client),
## This feature does not depend on nfs version. You can run the tests on all versions of NFS,
## and all tests will be same.
##
## @param   FILER               Required: hostspec of filer1.
##
## @param   VSERVER_GX_NAME     Required: The vserver to use.
##
## @param   SECURITY_STYLE      Optional: The security style(unix/mixed/ntfs) for the qtree in test. Default: mixed.
##
## @param   ID                  Optional: The numeric user id to use. Default: the id of the user running the thpl.
##
## @param   QTREE               Optional: The name of the qtree to use. Default: rquota_qtree_$PID.
##
## @param   QUOTA_POLICY        Optional: The name of the policy for the quota. Default: nfs_rquota_policy
##
## @param   UNIX_USER           Optional: The username to use for tests. Default: username of the user running the thpl.
##
## @param   UNIX_GROUP_NAME     Optional: The groupname to use for the tests. Default: group name of the user running the thpl.
##
## @param   DISK_LIMIT          Optional: The disk size limit to use for the tests. Default: 40000 KB.
##
## @param   DISK_LIMT_RESIZE    Optional: The disk size limit to use for "resize" tests. Default: 80000 KB.
##
## @param   FILE_LIMIT          Optional: The limit for number of files to use for the tests. Default: 200.
##
## @param   FILE_LIMIT_RESIZE   Optional: The limit for number of files to use for "resize" tests. Default: 400.
##
## The rest of the NFS parameters are defined in nfs/CommonParams.pm.
##
## @author  rraghav@netapp.com
##
## @Maintainer  dl-svl-nosauto@netapp.com
##

use strict;

use NACL::STask::VolumeQuota;
use NACL::STask::VserverServicesUnixUser;
use NACL::STask::VserverServicesUnixGroup;
use NACL::STask::Qtree;
use NACL::MTask::PacketCapture;

use NACL::C::VserverNfs;
use NACL::C::Vserver;
use NACL::C::VolumeQuotaPolicy;
use NACL::C::VolumeQuotaPolicyRule;
use NACL::C::Client;
use NACL::C::Job;

use NACL::CS::VolumeQuota;

use NACL::Exceptions::NoElementsFound qw{:try};
use nfs::rQuotaClient;

use nfs::CommonUtilities;
use nfs::NfsTCD qw{cleanup init uninit main $Node_cmd @UNINIT @CLEANUP $Client_cmd %Setup_hash};
use Carp qw{croak};
use Params::Validate qw{validate SCALAR};
use vars qw{$CLIENT};

#--------------------------------------------------------------------------------
# Initialize and validate parameters
#--------------------------------------------------------------------------------
use vars qw($ID $QUOTA_POLICY $QTREE $PRIMARY_GID $VSERVER_GX_NAME $SUDO $FILER_AGGR $FILER_AGGRB
            $SECURITY_STYLE $UNIX_USER $UNIX_GROUP_NAME $DISK_LIMIT_RESIZE $DISK_LIMIT $FILE_LIMIT $FILE_LIMIT_RESIZE $rQuotaClient $Vserver_nfs);
my $mygid = (split(/ /, $())[0];
param('VSERVER_GX_NAME', '-mesg', 'Name of the vserver to use');
param('SECURITY_STYLE',  '-default', 'mixed');
param('ID',              '-default', $<);
param('PRIMARY_GID',     '-default', $mygid);
param('QTREE',           '-default', 'rquota_qtree_' . $$);
param('QUOTA_POLICY',    '-default', 'nfs_rquota_policy');
param('UNIX_USER',       '-default', (getpwuid($<))[0]);
param('UNIX_GROUP_NAME', '-default', (getgrgid($mygid))[0]);
param('DISK_LIMIT', '-default', 40000); #KB
param('DISK_LIMIT_RESIZE', '-default', 80000); #KB
param('FILE_LIMIT', '-default', 200); #no of files
param('FILE_LIMIT_RESIZE', '-default', 400); #no of files

#--------------------------------------------------------------------------------
#
#=head2 init
#
#    init ()
#
#The initialization function. This function
#1. finding the nodes using the Find function,
#2. sets up the clients.
#
#=item Arguments
#
#This function does not take any arguments. It picks them from the global array.
#
#=back
#
#=item Returns
#
#This function aborts if any of its steps fails. otherwise it returns PASS value.
#
#=back
#
#--------------------------------------------------------------------------------
sub init() {

    nfs::NfsTCD::init();



    $Vserver_nfs = NACL::C::VserverNfs->find(command_interface => $Node_cmd,
                                 filter => {'vserver' => $VSERVER_GX_NAME});
    enable_rquota();

    $rQuotaClient = nfs::rQuotaClient->new(hostname=>$Setup_hash{hostname});

    create_qtree();
    create_quota_policy();

    NACL::C::Vserver->modify(command_interface => $Node_cmd,
                             'vserver'         => $VSERVER_GX_NAME,
                             'quota-policy'    => $QUOTA_POLICY);

    create_group();
    create_user();

    return $TCD::PASS;
}

#--------------------------------------------------------------------------------
#
#=head2 cleanup
#
#    cleanup()
#
#The cleanup function as given in TCD format. It removes all the unnecessary files
#created and marked for deletion by the previous test.
#
#=item Arguments
#
#This function does not take any argument
#
#=back
#
#=item Returns
#
#This function returns a TCD::PASS.
#
#=back
#
#--------------------------------------------------------------------------------
sub cleanup () {
    $Vserver_nfs->modify(rquota=>'enabled');
    nfs::NfsTCD::cleanup();
    return $TCD::PASS;
}

# ------------------------------------------------------------------------------
# Function     : uninit()
# Input        : None
# Return       : $TCD::PASS
# summary      : Uninit automatically called
#                after executing tests
# ------------------------------------------------------------------------------
sub uninit() {
    $Log->comment("########## UNINIT ##########\n");

    NACL::C::Vserver->modify(command_interface => $Node_cmd,
                             'vserver'         => $VSERVER_GX_NAME,
                             'quota-policy'    => 'default');

    nfs::NfsTCD::uninit();
    return $TCD::PASS;
} ## end sub uninit()



#--------------------------------------------------------------------------------
# create_quota_policy
#   Create a new quota policy using the given arguments.
#   Takes the same arguments as the volume quota policy create.
#--------------------------------------------------------------------------------
sub create_quota_policy() {
    $Log->comment("Create quota policy");
    try {
        NACL::C::VolumeQuotaPolicy->find(command_interface=>$Node_cmd,
            filter => { vserver => $VSERVER_GX_NAME, 'policy-name' => $QUOTA_POLICY });
    } catch NACL::Exceptions::NoElementsFound with {
        push @UNINIT, NACL::C::VolumeQuotaPolicy->create(command_interface=>$Node_cmd,
            vserver => $VSERVER_GX_NAME, 'policy-name' => $QUOTA_POLICY);
    };
}

#---------------------------------------------------------------------------------
# _VolumeQuotaPolicyRule
#   Internal function to create or modify a quota policy rule.
#   Used by two other functions in this module. The arguments are almost same as
#   volume quota policy rule create, only difference is
#   disk-limit is disk
#   file-limit is file
#   The rest are same arguments as the command.
#---------------------------------------------------------------------------------
sub _VolumeQuotaPolicyRule($%) {

    $Log->comment("Create/Modify quota policy rule matching the given conditions.");
    my $func = shift;
    my %opts = validate (@_,
        {
            disk => { type => SCALAR, optional => 1, default => ($func eq "modify" ? $DISK_LIMIT_RESIZE : $DISK_LIMIT), regex => qr/^\d+$/ },
            file => { type => SCALAR, optional => 1, default => ($func eq "modify" ? $FILE_LIMIT_RESIZE : $FILE_LIMIT), regex => qr/^\d+$/ },
            mapping => { type => SCALAR, optional=>1, regex=> qr/^(on|off)$/, default => "off" },
            type => { type => SCALAR, optional=>1, regex=> qr/^(user|disk|qtree)$/, default => "user" },
            target => { type => SCALAR, optional=>1, default=> $ID },
            qtree => { type => SCALAR, optional=>1, default=>$QTREE },
            volume => { type => SCALAR, optional=>1, default=>$Setup_hash{vol_name}},
            'policy-name' => { type => SCALAR, optional=>1, default=>$QUOTA_POLICY},

        });

    $Log->comment("Create user quota policy rule for qtree with target user name");
    return NACL::C::VolumeQuotaPolicyRule->$func(command_interface => $Node_cmd,
                                           vserver       => $VSERVER_GX_NAME,
                                           'policy-name'  => $QUOTA_POLICY,
                                           volume         => $opts{volume},
                                           qtree          => $opts{qtree},
                                           target         => $opts{target},
                                           type           => $opts{type},
                                           'user-mapping' => $opts{mapping},
                                           'disk-limit'   => $opts{disk},
                                           'file-limit'   => $opts{file});
}

#---------------------------------------------------------------------------------
# create_quota_policy_rule
#   Create a new quota policy rule.
#   Same arguments as volume quota policy rule.
#---------------------------------------------------------------------------------
sub create_quota_policy_rule(%) {

    my (%opts) = validate(@_,
        {
            target => { type=>SCALAR, optional=>1, default=>$ID },
            qtree  => { type=>SCALAR, optional=>1, default=>$QTREE },
            volume => { type=>SCALAR, optional=>1, default=>$Setup_hash{vol_name} },
            'policy-name' => { type=>SCALAR, optional=>1, default=>$QUOTA_POLICY },
        });
    $Log->comment(qq{Check for the quota policy rule});
    try {
        NACL::C::VolumeQuotaPolicyRule->find(command_interface=>$Node_cmd,
            filter => { 'vserver'     => $VSERVER_GX_NAME,
                        'policy-name' => $opts{'policy-name'},
                        volume        => $opts{volume},
                        'target'      => $opts{target},
                        qtree         => $opts{qtree}, });

    } catch NACL::Exceptions::NoElementsFound with {
        push @CLEANUP, _VolumeQuotaPolicyRule('create', %opts);
    };

}

#---------------------------------------------------------------------------------
# resize_quota
# Modifying the quota parameters in a already created policy rule.
#---------------------------------------------------------------------------------
sub resize_quota(%) {
    _VolumeQuotaPolicyRule('modify', @_);
}

#---------------------------------------------------------------------------------
# volume_quota
#   Modify the volume quota state.
#   Same parameters as volume quota command.
#---------------------------------------------------------------------------------
sub volume_quota($;$) {
    my $state = shift;
    my $volume = shift || $Setup_hash{vol_name};
    $Log->comment(qq{Apply the quota policy on the volume});
    NACL::STask::VolumeQuota->modify(command_interface => $Node_cmd,
                                'vserver'          => $VSERVER_GX_NAME,
                                'volume'           => $volume,
                                'foreground'       => 'true',
                                'state'            => $state);

    verify_volume_quota($state,$volume);
}

#---------------------------------------------------------------------------------
# volume_quota_on
#   Turn on the quota on the given volume.
#---------------------------------------------------------------------------------
sub volume_quota_on(;$) {
    my $vol = shift || $Setup_hash{vol_name};
    $Log->comment("Turn on volume quota state on $vol");
    volume_quota('on',$vol);
}

#---------------------------------------------------------------------------------
# volume_quota_off
#   Turn off the quota on the given volume.
#---------------------------------------------------------------------------------
sub volume_quota_off(;$) {
    my $vol = shift || $Setup_hash{vol_name};
    $Log->comment("Turn off volume quota state on $vol");
    volume_quota('off', $vol);
}

#---------------------------------------------------------------------------------
# verify_volume_quota_on
#   Confirm that the quota is turned on the given volume.
#---------------------------------------------------------------------------------
sub verify_volume_quota_on() {
    verify_volume_quota('on');
}

#---------------------------------------------------------------------------------
# verify_volume_quota_off
#   Confirm that the quota is turned off in the given volume.
#---------------------------------------------------------------------------------
sub verify_volume_quota_off() {
    verify_volume_quota('off');
}
 
#---------------------------------------------------------------------------------
# verify_volume_quota
#   Verify that the status of the volume quota is correct.
#---------------------------------------------------------------------------------
sub verify_volume_quota($;$) {
    my $state = shift;
    my $vol = shift || $Setup_hash{vol_name};
    $Log->comment(qq{Verify that the quota is $state in the volume});
    my $obj = NACL::CS::VolumeQuota->fetch(command_interface => $Node_cmd,
                            filter => {'vserver' => $VSERVER_GX_NAME,
                                        'volume' => $vol});

    croak (qq{Volume quota should be $state, but found to be " . $obj->{state}}) if ($state ne $obj->{state});
}

#---------------------------------------------------------------------------------
# volume_quota_resize
#   Resize the volume quota.
#---------------------------------------------------------------------------------
sub volume_quota_resize() {
    $Log->comment("RESIZE volume quota");
    NACL::STask::VolumeQuota->resize(command_interface => $Node_cmd,
                                'vserver'          => $VSERVER_GX_NAME,
                                'volume'           => $Setup_hash{vol_name},
                                'foreground'       => 'true');
}

#---------------------------------------------------------------------------------
# create_qtree
#   Function to create a qtree in the given volume. Reuse if it exists.
#---------------------------------------------------------------------------------
sub create_qtree(;$) {
    $Log->comment("Create Qtree inside the volume");
    my $volume = shift || $Setup_hash{vol_name};
    my $created;
    my $obj = NACL::STask::Qtree->create(command_interface  => $Node_cmd,
                                           'nacltask_if_exists' => 'reuse',
                                           '_was_created'      => \$created,
                                           'vserver'          => $VSERVER_GX_NAME,
                                           'security-style'   => $SECURITY_STYLE,
                                           volume             => $volume,
                                           qtree              => $QTREE,
                                           'oplock-mode'      => 'enable',
                                           'unix-permissions' => '777');

    if ($created) { push @UNINIT, $obj; }
}

#---------------------------------------------------------------------------------
# get_quota_from_client
#   Function to get the quota from the synthetic client.
#   Send "GET_QUOTA" procedure to the rQuota server and decodes and returns
#   numbers sent by the rQuota server. If there is a problem it croaks.
#   If successful, it returns the disk limit in KB and file limit.
#---------------------------------------------------------------------------------
sub get_quota_from_client(;$$) {
    my $path = shift || "$Setup_hash{volume}/$QTREE";
    my $id   = shift || $ID;
    $Log->comment(qq{Get the quota using the client for $path for id:$id});
    my $quota = $rQuotaClient->get_quota(path=>$path, id=>$id);

    $Log->comment("get_quota output : " . Dumper($quota));
    # disk limit in kB.
    my $disk_limit = $quota->{bsize} * $quota->{disk_hard_limit} / 1024.0;
    return ($disk_limit, $quota->{file_hard_limit});
}

#---------------------------------------------------------------------------------
# verify_limits_are_same
#   Verify that the expected value is same as the obtained value.
#   Args: obtained disk limit, obtained file limit,
#         expected disk limit and expected file limit.
#---------------------------------------------------------------------------------
sub verify_limits_are_same($$;$$) {
    $Log->comment(qq{Verify the expected and actual quota results are same});
    my ($disk, $file) = (shift, shift);
    my $exp_disk = shift || $DISK_LIMIT;
    my $exp_file = shift || $FILE_LIMIT;
    if ( $disk == $exp_disk && $file == $exp_file ) {
        $Log->comment("disk-limit and file-limit are same as expected");
    } else {
        croak ("disk-limit and file-limit are not same as expected:got:$disk, $file. Expected:$exp_disk, $exp_file");
    }
}

#---------------------------------------------------------------------------------
# verify_limits_are_same_after_mod
#   Confirm that the limits are same after a quota resize.
#---------------------------------------------------------------------------------
sub verify_limits_are_same_after_mod($$) {
    my @args = @_;
    verify_limits_are_same($args[0], $args[1], $DISK_LIMIT_RESIZE, $FILE_LIMIT_RESIZE);
}

#---------------------------------------------------------------------------------
# move_test_vol
#   Volume move function.
#---------------------------------------------------------------------------------
sub move_test_vol(;$$) {
    my $volume = shift || $Setup_hash{vol_name};
    my $dest = shift || $FILER_AGGRB;
    $Log->comment("Move the volume to another aggregate");

    my $move_obj = NACL::STask::VolumeMove->start(command_interface=>$Node_cmd,
        vserver=>$::VSERVER_GX_NAME,
        volume => $volume, 'destination-aggregate'=>$dest);

    $move_obj->wait_for_move();

}

#---------------------------------------------------------------------------------
# move_test_vol_back
#   Move the volume back to the original volume. If the volume was never moved,
#   this may error out.
#---------------------------------------------------------------------------------
sub move_test_vol_back() {
    $Log->comment("Move the volume back to original aggregate");
    move_test_vol($Setup_hash{vol_name}, $FILER_AGGR);
}

#---------------------------------------------------------------------------------
# enable_rquota
#   Enable the rquota on the vserver nfs.
#---------------------------------------------------------------------------------
sub enable_rquota() {
    $Log->comment("Enable rquota in the vserver");
    $Vserver_nfs->modify(rquota=>'enabled');

}

#---------------------------------------------------------------------------------
# disable_rquota
#   Disable the rquota on vserver nfs.
#---------------------------------------------------------------------------------
sub disable_rquota() {
    $Log->comment("Disable rquota in the vserver");
    $Vserver_nfs->modify(rquota=>'disabled');
}

#---------------------------------------------------------------------------------
# create_group
#   Create the given unix group in the filer.
#---------------------------------------------------------------------------------
sub create_group(;$$) {

    my ($gid, $gname) = (shift || $PRIMARY_GID, shift || $UNIX_GROUP_NAME);
    my $created;
    my $unix_group_obj = NACL::STask::VserverServicesUnixGroup->create(
                                                 command_interface => $Node_cmd,
                                                nacltask_if_exists => "reuse",
                                                         'vserver' => $VSERVER_GX_NAME,
                                                              'id' => $gid,
                                                            'name' => $gname,
                                                      _was_created => \$created);

    if ($created) { push @UNINIT, $unix_group_obj };
    $UNIX_GROUP_NAME = $unix_group_obj->state()->name();

}

#---------------------------------------------------------------------------------
# create_user
#   Create the unix user in the filer.
#---------------------------------------------------------------------------------
sub create_user(;$$$) {
    my ($id, $name, $gid) = (shift||$ID, shift||$UNIX_USER, shift || $PRIMARY_GID);
    #Create a unix-user
    $Log->comment("Create a unix user on filer");
    my $created;
    my $unix_user_obj = NACL::STask::VserverServicesUnixUser->create(command_interface => $Node_cmd,
            nacltask_if_exists => "reuse",
            vserver => $VSERVER_GX_NAME,
            user => $name,
            id   => $id,
            'primary-gid' => $gid,
            'full-name'   => $name,
            _was_created => \$created);
    if ($created) { push @UNINIT, $unix_user_obj };
    my $state_user = $unix_user_obj->state();
    $UNIX_USER   = $state_user->user();
    $ID          = $state_user->id();
    $PRIMARY_GID = $state_user->primary_gid();
}

#---------------------------------------------------------------------------------
# add_users
#   Add the given number of users in the filer.
#---------------------------------------------------------------------------------
sub add_users(;$) {
    my $count = shift || 1;

    if ($count <= 0) { 
        croak (qq{Programming error: add_users: Some function is asking for $count users !});
    }
    my @ret;
    while ($count) {
        create_user($ID+$count, "${UNIX_USER}_$count", $PRIMARY_GID+$count);
        push @ret, $ID+$count;
        $count--;
    }
    return @ret;
}

#---------------------------------------------------------------------------------
# add_groups
#   Add the given number of groups in the filer.
#---------------------------------------------------------------------------------
sub add_groups(;$) {
    my $count = shift || 1;

    if ($count <= 0) { 
        croak (qq{Programming error: add_groups: Some function is asking for $count groups !});
    }
    while ($count) {
        create_group($PRIMARY_GID+$count, "$UNIX_GROUP_NAME" . "_$count");
        $count--;
    }
}

#---------------------------------------------------------------------------------
# create_vol
#   Wrapper to add volume. Push into the cleanup array too.
#---------------------------------------------------------------------------------
sub create_vol(%) {
    push @CLEANUP, nfs::CommonUtilities::create_vol(@_);
}

#---------------------------------------------------------------------------------
# start_packet_capture
#   Function to start the traffic capture "tcpdump" in the client.
#---------------------------------------------------------------------------------
sub start_packet_capture(;$) {
    my $intf = shift || "eth0";
    my @arguments = [{
                target_machine => $Client_cmd,
                interface_filter => [$intf],
                capture_filter => "host $Setup_hash{hostname}",
            }];
    $Log->comment(qq{Starting the capture of packets in the client});
    my $packetcapture = NACL::MTask::PacketCapture->new(packet_capture_details => \@arguments);
    $packetcapture->start();
    push @CLEANUP, $packetcapture;
    return $packetcapture;
}

#---------------------------------------------------------------------------------
# stop_packet_capture
#   Function to stop the packet capture in the client.
#---------------------------------------------------------------------------------
sub stop_packet_capture($) {
    my $packetcapture = shift;
    $Log->comment(qq{Stopping the capture of packets in the client});
    $packetcapture->stop();
}

#---------------------------------------------------------------------------------
# Re-init function to create a second vserver, and configure it accordingly.
# This function is untested as of now. The only thpl that is supposed to use this
# is nfs/rquota/c-mode/rquota_disp_diff_vservers.thpl
# If we manually create 2 vservers in the setup, then this thpl will work without
# issues. This function simply modifies the nfs parameters, and re-runs the init()
# routine.
#---------------------------------------------------------------------------------
sub re_init() {
    use vars qw($VSERVER_B $UNIX_USER_B $UNIX_GROUP_NAME_B $ID_B $PRIMARY_GID_B $FILERB $DATA_LIF_IP_B $FILER_AGGR_B $DATA_PORT_B);

    param('FILER_AGGR_B',      '-mesg',    'aggr name of FILERB');
    param('DATA_LIF_IP_B',     '-default', undef);
    param('VSERVER_B',         '-mesg',    'vserver name of FILERB');
    param('FILERB',            '-mesg',    'Name of filer B');
    param('PRIMARY_GID_B',     '-default', '4452');
    param('ID_B',              '-default', '46458');
    param('UNIX_USER_B',       '-default', 'rquota_testuser');
    param('UNIX_GROUP_NAME_B', '-default', 'rquota_testgroup');
    param('DATA_PORT_B', '-default', 'e0c');
    
    $VSERVER_GX_NAME = $VSERVER_B;
    $FILER = $FILERB;
    $ID = $::ID_B;
    $UNIX_USER = $UNIX_USER_B;
    $UNIX_GROUP_NAME = $UNIX_GROUP_NAME_B;
    $PRIMARY_GID = $PRIMARY_GID_B;
    if ($DATA_PORT_B) {
        $::DATA_PORT = $DATA_PORT_B;
    }
    if ($DATA_LIF_IP_B) {
        $::DATA_LIF_IP = $DATA_LIF_IP_B;
    }

    init();
}

1;
