#! - this is perl -
## Simple rQuota Client, based on Rpc.pm, similar to NfsClient.
## Implements get_quota, but only get_quota is
## completely tested against our filers. The set_quota is not
## supported in our filer, and so is not implemented in this PM.
##

=head1

=Synopsis

    use nfs::rQuotaClient;
    my $rq_client = nfs::rQuotaClient->new(hostname=>$filer_data_ip, ...);

    my $quota = $rq_client->get_quota(path=>"/vol/vol1/qtree1", id=>$UID);
    print $quota->{disk_hard_limit};
    print $quota->{disk_soft_limit};
    print $quota->{size}; etc...

=cut


package nfs::rQuotaClient;

use strict;
use warnings;
use Data::Dumper;
use Tharn qw{$Log};
use Carp qw{croak};

## Please refer to this file:
## /usr/include/rpcsvc/rquota.x
## We used this as a template to write this rQuota Client.
## Please run nfs/rquota/c-mode/rQuotaClient.thpl as unit test for this pm.

use nfs::Rpc;
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF OBJECT);

# Rquota constants.
use constant RQUOTA => 100011;
use constant RQUOTA_VERSION => 1;

# Rquota procedures.
use constant RQUOTAPROC_GETQUOTA => 1;
use constant RQUOTAPROC_GETACTIVEQUOTA => 2;
use constant RQUOTAPROC_SETQUOTA => 3;
use constant RQUOTAPROC_SETACTIVEQUOTA => 4;

# Quota types
use constant RQUOTA_USRQUOTA => 0x00;
use constant RQUOTA_GRPQUOTA => 0x01;
use constant RQUOTA_MAXQUOTAS => 0x02;

# Rquota results
use constant Q_OK => 1;
use constant Q_NOQUOTA => 2;
use constant Q_EPERM => 3;

sub new() {
    my $class = shift;
    my (%opts) = validate (@_,
        {
            hostname => { type=>SCALAR },
            proto => { type=>SCALAR, optional=>1, default=>"udp" },
            version => { type=>SCALAR, optional=>1, default=>RQUOTA_VERSION },
        });
    eval {
        $opts{rpc} = nfs::Rpc->clnt_create(program => RQUOTA, %opts);
        $opts{rpc}->clnt_call(proc => 0);
    };
    if ($@) {
        if ($@ =~ /Timeout/i) {
            croak (qq{Please don't run the rquota tests across WAN, it will fail with timeout. The full error:$@});
        } else {
            croak $@;
        }
    }

    return (bless { %opts }, $class);
}

sub xdr_quota(%) {
    my %opts = validate(@_,
        {
            encode => { type=>HASHREF, optional => 1 },
            xdr => { type=>OBJECT, optional=>1 },
         });

    if ($opts{encode}) {
        my $xdr = nfs::Xdr->new();
        $xdr->uint(encode=>$opts{encode}->{size});
        $xdr->uint(encode=>$opts{encode}->{active});
        $xdr->uint(encode=>$opts{encode}->{disk_hard_limit});
        $xdr->uint(encode=>$opts{encode}->{disk_soft_limit});
        $xdr->uint(encode=>$opts{encode}->{curr_blk_count});
        $xdr->uint(encode=>$opts{encode}->{file_hard_limit});
        $xdr->uint(encode=>$opts{encode}->{file_soft_limit});
        $xdr->uint(encode=>$opts{encode}->{curr_files_count});
        $xdr->uint(encode=>$opts{encode}->{disk_time_left});
        $xdr->uint(encode=>$opts{encode}->{file_time_left});
        return $xdr->{message};
    }

    # decode the quota stuff.
    my $xdr = $opts{xdr} || croak (qq{Programming error, either encode or xdr must be present});
    my %ret = ();
    $ret{bsize} = $xdr->uint();
    $ret{active} = $xdr->uint();
    $ret{disk_hard_limit} = $xdr->uint();
    $ret{disk_soft_limit} = $xdr->uint();
    $ret{curr_blk_count} = $xdr->uint();
    $ret{file_hard_limit} = $xdr->uint();
    $ret{file_soft_limit} = $xdr->uint();
    $ret{curr_files_count} = $xdr->uint();
    $ret{disk_time_left} = $xdr->uint(); #Time left for excessive disk usage.
    $ret{file_time_left} = $xdr->uint(); #Time left for excessive files usage.
    return \%ret;
}

sub xdr_get_quota(%) {
    my %opts = validate(@_,
        {
            xdr    => { type=>OBJECT },
        });

    rquota_result(xdr=>$opts{xdr});
    return (xdr_quota(xdr=>$opts{xdr}));
}

sub rquota_result(%) {
    my %opts = validate(@_,
        {
            xdr => { type => OBJECT },
        });

    my $xdr = $opts{xdr};
    my $status = $xdr->uint();
    if ($status != Q_OK) {
        if ($status == Q_NOQUOTA) {
            croak (qq{No quota is available in the given path});
        } elsif ($status == Q_EPERM) {
            croak (qq{The user does not have the permissions to read quota from the path});
        } else {
            croak (qq{Unexpected xdr result, please check programming on client or the rQuota server itself});
        }
    }

    return $status;
}

sub get_quota($%) {
    my $self = shift;
    my %opts = validate (@_,
        {
            path => { type=>SCALAR },
            id   => { type=>SCALAR, regex=>qr/^\d+$/ },
        });

    my $xdr = nfs::Xdr->new();
    $xdr->opaque(encode=>$opts{path});
    $xdr->uint(encode=>$opts{id});
    my $get_quota_result = $self->{rpc}->clnt_call(proc=>RQUOTAPROC_GETQUOTA, arg=>$xdr->{message}, out => sub { xdr_get_quota(@_) });
    return $get_quota_result;
}

1;
