## @summary NFS reference directory utilities
## @status review
## @author  dl-maryv (FSS Protocols QA)
package nfs::refdir;

use strict;
use nfs::say;

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&refdir_makefiles &refdir_mkrefdir &refdir_setup &refdir_cleanup);

use Tharn;
use mount;
use nfs::say;
use nfs::rshlib;
use nfs::exports;
use nfs::RemoteFileIO qw(client_del_file);
use constant SUDO => '/usr/software/bin/sudo';

sub refdir_makefiles($$$;) {
    my $ccon = shift();
    my $refscript = shift();
    my $target = shift();

    if (! unix_say($ccon, SUDO . " touch $target/yow")) { 
	xlogresult('FATAL', 0, "could not create $target/yow");
	return 0;
    }

    return 1;			# XXX

    if (! unix_say($ccon, SUDO . " $refscript $target")) { 
	return 0;
    }

    return 1;
}

## @name ::refdir_mkrefdir
## @summary Create a reference directory.
## @description Either use a supplied script or an internal one.
## @synopsis
##   use nfs::refdir;
##   refdir_mkrefdir($client_conn, $ontapdir, $target);
## @arg $client_conn Required: client connection
## @arg $ontapdir Required: location of ONTAPDIR for default mkrefdir script.
## @arg $target Required: location to build reference directory
## @usage returns true on success
## @example
##   use nfs::refdir;
##   refdir_mkrefdir($client_conn, $ontapdir, $target) or logcomment("failed");
## @dependencies an existing client connection.
sub refdir_mkrefdir($$$;) {
    my ($ccon, $ontap, $target) = @_;
    my $refscript = "$ontap/test/NFS/mkrefdir.pl";

    if (! -f $refscript) {
	xlogresult('FATAL', 1, "no $refscript script!");
    }

    client_del_file($ccon, $target, root => 1) or xlogresult('FATAL', 1, "couldn't remove $target");

    if (! unix_say($ccon, SUDO ." mkdir $target")) {
	xlogresult('FATAL', 1, "couldn't create target directory $target");
    } elsif (! unix_say($ccon, SUDO . " chmod a+rwx $target")) {
	xlogresult('FATAL', 1, "couldn't make target directory $target " .
		   "world writable");
    } elsif (! refdir_makefiles($ccon, $refscript, $target)) {
	xlogresult('FATAL', 1, "couldn't populate refdir $target");
    }

    return 1;
}


## @name ::refdir_setup
## @summary setup a reference directory
## @description
## 1. check volume exists.
## 2. export volume with anon=0
## 3. modify exports
## @synopsis
##   use nfs::refdir;
##   refdir_setup($client, $ccontype, $filer, $vol, $volmnt, $mntopts,
##                  $vers, $refdir, $ontap, $timeo);
##
## @arg $client Required: client hostname on which to cleanup.
## @arg $ccontype Required: standard client conntype.
## @arg $filer Required: filer hostname.
## @arg $vol Required: volume to cleanup
## @arg $volmnt Required: filer volume mount point
## @arg $mntopts Required: generic mount options suitible for sprintf.
## @arg $vers Required: nfs version to use in mount
## @arg $refdir Required: location to build reference directory
## @arg $ontapdir Required: location of ONTAPDIR for default mkrefdir script.
## @arg $timeo Required: recommended timeout for operation.
##
## @usage returns true on success
## @example
##   use nfs::refdir;
##   refdir_setup($client, $ccontype, $filer, $vol, $volmnt, $mntopts,
##                $vers, $refdir, $ontap, $timeo) or logcomment("failed");
## @dependencies an existing client connection.
sub refdir_setup($$$$$$$$$$;) {
    my($client, $ccontype, $filer, $vol, $volmnt, $mntopts,
       $vers, $refdir, $ontap, $timeo) = @_;
    my $chost = host($client);
    my $ccon = connect($ccontype);
    my $mnt;

    if (! exports_checkvol($filer, $volmnt)) {
	xlogresult('FATAL', 1, "$vol is unusable");
    } elsif (! exports_add($filer, "$volmnt -anon=0")) {
	xlogresult('FATAL', 1, "exporting $vol failed (so do it already)");
    } elsif (! exports_check($filer, $volmnt, 'anon=0')) {
	xlogresult('FATAL', 1, "$volmnt is not exported (so do it already)");
    }

    exports_exportfs($filer, $volmnt);

    # Do a quick check to see that we can mount the volume, and
    # create a temporary working directory.
    if (! ($mnt = mount($filer, $volmnt, client=>$client, type=>'nfs',
			   vers=>$vers, mountopts=>$mntopts))) {
	xlogresult('FATAL', 1, "couldn't mount $filer:$volmnt: $@");
    } elsif (! unix_say($ccon, "test -d $mnt")) {
	xlogresult('FATAL', 1, "nfs mount of $filer:$volmnt at $mnt" .
		   " with $mntopts,vers=$vers: no directory $mnt exists");
    } else {
	logcomment( "mounted $filer:$volmnt at $mnt: " .
	       `/bin/ls -ld $mnt` . "\nAnd df: " . `df -k $mnt`);
    }

    if (! refdir_mkrefdir($ccon, $ontap, "$mnt/$refdir")) {
	xlogresult('FATAL', 1, "couldn't create refdir $mnt/$refdir");
    } elsif (! umount($mnt, client=>$client)) {
	xlogresult('FATAL', 1, "couldn't unmount $mnt: $@");
    }
}

## @name ::refdir_cleanup
## @summary cleanup a reference directory
## @synopsis
##   use nfs::refdir;
##   refdir_cleanup($client, $ccontype, $filer, $vol, $volmnt, $mntopts,
##                  $vers, $refdir, $ontap, $timeo);
##
## @arg $client Required: client hostname on which to cleanup.
## @arg $ccontype Required: standard client conntype.
## @arg $filer Required: filer hostname.
## @arg $vol Required: volume to cleanup
## @arg $volmnt Required: filer volume mount point
## @arg $mntopts Required: generic mount options suitible for sprintf.
## @arg $vers Required: nfs version to use in mount
## @arg $refdir Required: location to build reference directory
## @arg $ontapdir Required: location of ONTAPDIR for default mkrefdir script.
## @arg $timeo Required: recommended timeout for operation.
##
## @usage returns true on success
## @example
##   use nfs::refdir;
##   refdir_cleanup($client_conn, $ontapdir, $target) or logcomment("failed");
## @dependencies an existing client connection.
sub refdir_cleanup($$$$$$$$$$;) {
    my($client, $ccontype, $filer, $vol, $volmnt, $mntopts,
       $vers, $refdir, $ontap, $timeo) = @_;
    my $export = "$volmnt -sec=sys,anon=0";
    my $mnt = undef;

    my $chost = host($client);
    my $ccon = connect($ccontype);

    if (! exports_add($filer, $export)) {
	xlogresult('FATAL', 1, "readding $export to $filer");
    } elsif (! exports_exportfs($filer, $volmnt)) {
	xlogresult('FATAL', 1, "reexporting $filer:$volmnt for last time");
    } elsif (! ($mnt = mount(simname($filer), $volmnt, client=>$client,
			     type=>'nfs', vers=>$vers,
			     mountopts=>$mntopts))) {
	xlogresult('FATAL', 1, "couldn't remount $filer:$volmnt: $@");
    } elsif (! client_del_file($ccon, "$mnt/$refdir", root => 1)) {
	xlogresult('FATAL', 1, "couldn't remove refdir $mnt/$refdir.");
    } elsif (! umount($mnt, client=>$client)) {
	xlogresult('FATAL', 1, "finally couldn't unmount $mnt: $@");
    }
}

1;
