## @summary NFS related interface to the rsh(1) command.
## @status review
## @author  dl-maryv (FSS Protocols QA)
package nfs::rshlib;

use strict;
use Tharn;
use nfs::say;

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&rsh &rsh_write &simname &fqname);

# utility for returning qualified hostname, if a sim
sub simname($) {
    my $host = shift;

    if (length(`ypmatch $host.sim hosts 2> /dev/null`) > 0) {
	return "$host.sim";
    }
    return $host;
}

# fully qualified hostname
sub fqname($) {
    my $host = shift;

    my $name = simname($host);

    my @a = split(/\s+/, `ypmatch $name hosts 2> /dev/null`);

    return $a[1];
}

## @name ::rsh
## @summary interface to the rsh(1) command
## @synopsis
##   use nfs::rshlib;
##   rsh($filer, $cmd);
## @arg $filer Required: filer hostname
## @arg $cmd Required: command to write file
##
## @usage returns undef, or the output from the remote filer.
## @example
##   use nfs::rshlib;
##   my $cmd = "rdfile /etc/exports";
##   rsh($filer, $cmd) > 0 or logcomment("failed");
## @dependencies rsh
## @bugs
sub rsh($$;) {
    my ($filer, $cmd) = @_;
    my(@results) = ();
    $filer = simname($filer);

    tstlog('-comment', "Sending \"$cmd\" to $filer");
    if (! open(RSH, "rsh $filer -l root: \"$cmd\"|")) {
        xlogresult('WARN', 0, "couldn't open rsh socket to $filer for $cmd");
	return undef;
    }

    while (my $line = <RSH>) {
	push(@results, $line);
    }

    if (! close(RSH)) {
        xlogresult('WARN', 0, "couldn't close rsh socket to $filer for $cmd");
	return undef;
    }

    return @results;
}

## @name ::rsh_write
## @summary use the ONTAP wrfile command via an rsh pipe to write a file.
## @synopsis
##   use nfs::rshlib;
##   rsh_write($filer, $cmd, @lines);
## @arg $filer Required: filer hostname
## @arg $cmd Required: command to write file
## @arg @lines Required: lines of text to write.
## @usage returns undef, or the number of lines written.
## @example
##   use nfs::rshlib;
##   my $cmd = "wrfile /etc/exports";
##   my (@lines) = ("/vol/vol0 -sec=krb5,anon=-1\n");
##   rsh_write($filer, $cmd, @lines) == 1 or logcomment("failed");
## @dependencies rsh
## @bugs ONTAP shell is probably not preferable to an NFS mount.
sub rsh_write($$@;) {
    my ($filer, $cmd, @lines) = @_;
    my(@results) = ();
    my $written = 0;
    $filer = simname($filer);

    if (! open(RSH, "|rsh $filer -l root: $cmd")) {
        xlogresult('WARN', 0, "couldn't open rsh socket to $filer for $cmd");
	return undef;
    }

    foreach my $line (@lines) {
	if (! print(RSH $line)) {
	    xlogresult('WARN', 0, "couldn't print through rsh socket to" .
		       " $filer for $cmd");
	    close(RSH);		# ignore any additional errors
	    return undef;
	}
	xlogcomment("WROTE $line");
	$written++;
    }

    if (! close(RSH)) {
        xlogresult('WARN', 0, "couldn't close rsh socket to $filer for $cmd");
	return undef;
    }

    return $written;
}

1;
