## @summary NFS communications and logging library.
## @status review
## @author  dl-maryv (FSS Protocols QA)
package nfs::say;

use strict;
use Tharn;

use vars qw(@ISA @EXPORT $TIMEOUT);

param('TIMEOUT', '-default', 200);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&unix_say &unix_say2 &xlogresult &xlogcomment &funclog);

## @name ::unix_say
## @summary execute a remote unix command and assert that exit status was 0
## @synopsis
##   use nfs::say;
##   unix_say($conn, $cmd);
## @arg $conn Required: standard unix client connection
## @arg $cmd Required: command to execute
##
## @usage returns true if command returned 0.
## @example
##   use nfs::say;
##   param('CLIENT', '-mesg', "client spec");
##   param('CLIENT_CONN', '-default', '#rlogin');
##   my $ch = host($CLIENT);
##   my $ccon = connect($CLIENT_CONN);
##   my $cmd = '/bin/pwd';
##   unix_say($ccon, $cmd) or logcomment("failed");
## @dependencies UNIX rlogin semantics
sub unix_say($$@;) {
    my ($conn, $cmd) = @_;
    my $xtoken = '__XstatusX';
    my $timeout = $TIMEOUT;

    if (! length($TIMEOUT) || $TIMEOUT !~ /^[0-9]+$/) {
	$timeout = 200;
    }

    my $sayout = say($conn, "$cmd; echo $xtoken=\$?",
		     "$xtoken=[0-9]+", '-timeout', $timeout);

    if ($sayout !~ /$xtoken=0/) {
	(my $errval = $sayout) =~ /$xtoken=([0-9]+)/;
	logcomment("$cmd returned status=$1");
    }

    return $sayout =~ /$xtoken=0/;
}

sub unix_say2($$$;) {
    my ($conn, $cmd, $opts) = @_;
    my $xtoken = '__XstatusX';
    if (! $$opts{timeout}) { $$opts{timeout} = $TIMEOUT; }

    my $sayout = say($conn, "$cmd; echo $xtoken=\$?",
		     "$xtoken=[0-9]+", '-timeout', $$opts{timeout});

    if ($sayout !~ /$xtoken=0/) {
	(my $errval = $sayout) =~ /$xtoken=([0-9]+)/;
	logcomment("$cmd returned status=$1");
    }

    return $sayout =~ /$xtoken=0/;
}

## @name ::xlogresult
## @summary interface to logresult
## @description prints package, function, and line number. 
## @synopsis
##   use nfs::say;
##   xlogresult($type, $abort, $message);
## @arg $type Required: standard logresult log type.
## @arg $abort Required: boolean to abort or not
## @arg $message Required: message to go with result
##
## @example
##   use nfs::say;
##   xlogresult('FAIL', 1, "aborting NOW");
sub xlogresult($$$;) {
    my $type = shift();
    my $abort = shift();
    my $comment = shift();

    my (undef, $file, $line, $routine) = caller(1);
    logresult(type=>$type, abort=>$abort,
	      $comment, msg =>("in $routine() at $file:$line"));
}

## @name ::xlogcomment
## @summary interface to logcomment
## @description prints package, function, and line number.
## @synopsis
##   use nfs::say;
##   xlogcomment($message);
## @arg $message Required: message to go with result
## @example
##   use nfs::say;
##   xlogcomment("hello");
sub xlogcomment($;) {
    my $comment = shift();

    my ($p, $f, $l, $routine) = caller(1);
    logcomment("$comment in $routine() at $f:$l");
}

## @name ::funclog
## @summary special function-printing interface to logcomment
## @description prints an announcement that we entered a given function.
## @synopsis
##   use nfs::say;
##   sub foo { funclog(); }
sub funclog() {
    my $func = (caller(1))[3];
    logcomment("$func()");
}

1;
