# -*- perl -*-

## ---------------------------------------------------
## Copyright (c) 2008-2013 NetApp, Inc., All Rights Reserved
## Permission is granted to use this script on a NetApp Product.
## Any other use, modification, or distribution is prohibited
## without prior written consent from NetApp, Inc.
## ---------------------------------------------------
#{[(
package nfs_setup_gx;

## ---------------------------------------------------
## @summary
##  Routines for setting up 10 mode filer for GX testing
##
## @description
##  This routine will be used to set up a C-Mode node ready for NFS testing. We will use this
##  to setup a node in similar way the nfssetup function does for 7G system.
##
## @param FILER_GX_VOL      Required: Volume for GX filer
##
## @param FILER_GX_AGGR     Required: Aggregate for the GX filer
##
## @param AGGR_GX_TYPE      Required: Type of the aggregate
##
## @param VSERVER_GX_NAME   Required: Name of the vserver
##
## @param EXPORT_GX_POL     Required: Name of the export policy
##
## @usage
##  use nfs_setup_gx;
##
## @status      new
##
## @author      talukder@netapp.com, rraghav@netapp.com
##
## @contact     dl-nfsqa@netapp.com
##
##
##

use strict;
use Tharn;
use Carp;
use SPARSE::Util::Logger;
BEGIN {
    $Log->comment(qq{Entering begin block});
    use vars qw( $FILER_GX_VOL $FILER_VOL
            $FILER_GX_AGGR
            $AGGR_GX_TYPE
            $VSERVER_GX_NAME
            $EXPORT_GX_POL
            $EXPORT_GX_RULE_CLIENT_MATCH
            $EXPORT_GX_RULE_RO
            $EXPORT_GX_RULE_RW
            $EXPORT_PROTO
            $JUNCTION_PATH
            $NFS4_MINOR_VERSION
            $V4_ID_DOMAIN
            $VOL_GX_SIZE );

    # This function is used already in the BEGIN block, so needs to be defined.
    param('ADDR_TYPE',          "-default", 'inet4');
    sub correct_address_family(;$) {
        my $given = shift;

        if (!$given) {
            $given = $::ADDR_TYPE;
        }

        if ($given eq "inet6" or $given eq "ipv6") {
            return "ipv6";
        } else {
            return "ipv4";
        }
    }

    # Keep this alphabetical !
    param("AGGR_GX_TYPE",       "-default", 'flex');
    param("AGGR_SIZE",          "-default", 3);
    param("DNS_DOMAIN",         "-default", 'lab.netapp.com');
    param('DNS_SERVER',         "-default", '172.19.2.30');
    param('DATA_LIF_IP',        "-default", undef);
    param('DATA_MASK',        "-default", undef);
    param('DATA_PORT',        "-default", undef);
    param("EXPORT_GX_POL",      "-default", 'test_pol');
    param("EXPORT_GX_RULE",     "-default", (correct_address_family() eq "ipv4" ? "0.0.0.0/0" : "::0/0"));
    param("EXPORT_GX_RULE_RO",  "-default", 'any');
    param("EXPORT_GX_RULE_RW",  "-default", 'any');
    param("EXPORT_PROTO",       "-default", 'any');
    param("EXPORT_ANON",        "-default", '65534');
    param("FILER_GX_VOL",       "-default", 'testvol_flex');
    param("FILER_GX_AGGR",      "-default", 'testaggr');
    param("FILER_VOL",          "-set",     $::FILER_GX_VOL);
    param("INTERFACE_UNDER_TEST","-default",'');
    param("JUNCTION_PATH",      "-default", '/'.$::FILER_GX_VOL);
    param("NFS4_MINOR_VERSION", "-default", 0);
    param("NIS_DOMAIN",         "-default", 'lab.netapp.com');
    param('NIS_SERVER',         "-default", '172.19.3.11');
    param("NOBODY_GROUP",       "-default", 'nogroup');
    param("NOBODY_GROUP_ID",    "-default", '65535');
    param("NOBODY_USER_ID",     "-default", $::EXPORT_ANON);
    param("SUPERUSER",          "-default", 'any');
    param("V4_ID_DOMAIN",       "-default", 'netapp.com');
    param("VSERVER_GX_NAME",    "-default", 'testvserver');
    param('VOL_GX_SIZE',        "-default", '200m');
    $Log->comment(qq{Exiting begin block});
};

use Params::Validate qw(validate SCALAR ARRAYREF HASHREF UNDEF OBJECT);
use Server;
use nfs::RemoteFileIO qw{stop_io start_io};
use vars qw(
        $VERSION
        @ISA
        @EXPORT_OK
       );

use NACL::CS::Interface;
use NACL::CS::Volume;
use NACL::CS::NetworkInterface;

use NACL::STask::VserverExportPolicy;
use NACL::C::ExportPolicyRule;
use NACL::STask::Node;
use NACL::STask::NetworkInterface;
use NACL::STask::VserverNfs;
use NACL::STask::VserverServicesDns;
use NACL::STask::VserverServicesNisDomain;
use NACL::STask::VserverServicesUnixGroup;
use NACL::STask::VserverServicesUnixUser;

use NACL::STask::NetworkInterface;
use NACL::STask::Aggregate;
use NACL::STask::Vserver;
use NACL::STask::Volume;

use NACL::Exceptions::NoElementsFound qw{:try};
use NACL::APISet::Exceptions::NoMatchingEntriesException qw{:try};

use Data::Dumper;

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(
    &set_env_gx
    &clean_env_gx
    &create_gx_aggr
    &create_nfs_server
    &create_gx_volume
    &create_gx_vserver
    &create_gx_export_rule
    &create_gx_users_and_group
    &get_data_ip_cmode
    get_number_of_delegs
);

my $server_obj;
sub _get_server() {
    if ($server_obj) { return; }
    $server_obj = Server->new(node => $main::FILER);
}

my $node_obj;
sub _get_node() {
    if ($node_obj) { return; }
    $node_obj = NACL::STask::Node->find();
}
my ($opts1, @NFS_SETUP_CLEANUP);

##Define a hash with variables set to parameters defined in CommonParams

sub init_opts() {
    $opts1->{vol_name} = $::FILER_GX_VOL;
    $opts1->{aggr_name} = $::FILER_GX_AGGR;
    $opts1->{aggr_type} = $::AGGR_GX_TYPE;
    $opts1->{vserver_name} = $::VSERVER_GX_NAME;
    $opts1->{export_pol} =  $::EXPORT_GX_POL;
    $opts1->{junc_path} = $::JUNCTION_PATH;
    $opts1->{sec_style} = 'unix';
    $opts1->{export_rule} = $::EXPORT_GX_RULE;
    $opts1->{export_rule_ro} = $::EXPORT_GX_RULE_RO;
    $opts1->{export_rule_rw} = $::EXPORT_GX_RULE_RW;
    $opts1->{export_proto} = $::EXPORT_PROTO;
    $opts1->{export_anon} = $::EXPORT_ANON;
    $opts1->{superuser} = $::SUPERUSER;
    $opts1->{id_domain} = $::V4_ID_DOMAIN;
    $opts1->{large_rule_index} = 500000;
    $opts1->{data_ip_cmode} = undef;
}

sub add_to_cleanup($) {
    push @NFS_SETUP_CLEANUP, shift;
}

##################################################
#set_env_gx() is called
#from NfsOps.pm, CompoundRequests.pm and Locks.pm
##################################################

sub set_env_gx ( % ) {
    my %opts = validate(
            @_,
            {
            setup_h  => { type => HASHREF },
            allowed  => { type => ARRAYREF, default => [ 2, 3, 4 ] },
            version  => { type => SCALAR,   default => 'all' },
            minor_version => { type => SCALAR, default => $::NFS4_MINOR_VERSION },
            protocol => { type => SCALAR,   default => 'both',
            regex    => qr/^(both|tcp|udp)6?$/ },
            timeout  => { type => SCALAR,   default => 30 },
            address_family    => { type => SCALAR, optional => 1, default => correct_address_family(),
                                regex => qr/^(inet4|inet6|ipv4|ipv6)$/},
            }
        );

    init_opts();
    ## Check if setup needs to be done
    if ($main::SETUP_DONE == 0)
    {
        $Log->comment ("Setting up the GX filer as requested");
        gx_filer_setup(%opts);
    }

    # Full path for volume
    my $vol;
    _get_node();

    try {
        my $volobj = NACL::CS::Volume->fetch(command_interface=>$node_obj,
            filter => { vserver => $::VSERVER_GX_NAME, volume => $opts1->{vol_name}, }, requested_fields => [ 'junction-path' ]);

        $vol = $volobj->{'junction_path'};
    } catch NACL::Exceptions::NoElementsFound with {
        croak (qq{Couldnot find the required volume: } . $opts1->{vol_name});
    };

    if (!$vol) {
        croak (qq{Junction path couldnot be found for } . $opts1->{vol_name});
    }

    $opts{setup_h}->{fc} = $node_obj->{version_manager}->{cli_connection};

    # Actual data ip address to use.
    $opts{setup_h}->{hostname} = get_data_ip_cmode($main::FILER, $main::VSERVER_GX_NAME, $opts{address_family});

    # Actual volume is available in vol_name
    $opts{setup_h}->{vol_name} = $opts1->{vol_name};

    # Actual junctionpath is named volume.
    $opts{setup_h}->{volume} = $vol;

    #logStep(qq{set_env:} . Dumper($opts{setup_h}));
    return $opts{setup_h};
}

#########################
#This is the main function
#that will call other's to
#complete the setup.
#########################

sub gx_filer_setup (%) {
    logStep(qq{Entering gx_filer_setup});

    my %opts = @_;
    create_gx_aggr(\%opts);
    create_gx_vserver (\%opts);
    create_gx_net_int(\%opts);
    create_nfs_server (\%opts);
    create_nis();
    create_dns();
    create_gx_export_rule(\%opts);
    create_gx_users_and_group(\%opts);
    create_gx_volume(\%opts);

    logStep(qq{Exiting gx_filer_setup});
}

sub create_gx_aggr($) {

    my $opts= shift;
    _get_server();
    _get_node();

    ## Populate the required parameters needed for creating the aggregate

    my $aggr_type = ($opts->{aggr_type}) ? $opts->{aggr_type} : $opts1->{aggr_type};
    my $aggr_name = ($opts->{aggr_name}) ? $opts->{aggr_name} : $opts1->{aggr_name};

    eval {
        if ($server_obj->storage_aggregate_exists(aggregate=>$aggr_name)) {
            return;
        }

        $Log->comment ("Creating an aggregate");
        add_to_cleanup (NACL::STask::Aggregate->create(command_interface=>$node_obj,
                    aggregate => "$aggr_name",
                    diskcount => $::AGGR_SIZE,
                    nodes => ["$main::FILER"],
                    raidtype => 'raid4',
                    nacltask_if_exists => 'reuse'));
    };

    if ($@) {
        if ($@ =~ /There are no entries matching your query./) {
            # need to wait for the aggr to finish initialization. Wait for maximum 600 seconds, check every 10 seconds.
            my $wait_time = 600;
            my $check_period = 10;
            while ($check_period <= $wait_time) {
                if ($server_obj->storage_aggregate_exists(aggregate=>$aggr_name)) {
                    return 0;
                }
                $check_period+=10;
                sleep 10;
            }

            die config_result ("Could not create the aggregate, may be you should wait for the aggregate to get created.");
        } else {
            die config_result ("Could not create the aggregate, $@");
        }
    }

}

sub create_gx_vserver($) {
    my $opts = shift;
    _get_server();
    _get_node();

    ## Populate the required parameters needed for creating the vserver

    my $vserver_name =  ($opts->{vserver_name}) ? $opts->{vserver_name} : $opts1->{vserver_name};
    my $root_vol = $vserver_name . '_' . 'root';
    my $sec_style = $opts->{sec_style} || 'unix';
    my $aggr_type = ($opts->{aggr_type}) ? $opts->{aggr_type} : $opts1->{aggr_type};
    my $aggr_name = ($opts->{aggr_name}) ? $opts->{aggr_name} : $opts1->{aggr_name};

    if ($server_obj->vserver_exists(vserver=>$vserver_name)) {
        return;
    }

    ## Create the vserver
    $Log->comment ("Creating vserver");
    add_to_cleanup (NACL::STask::Vserver->create(command_interface=>$node_obj,
                vserver => "$vserver_name",
                rootvolume => "$root_vol",
                aggregate => "$aggr_name",
                'ns-switch' => ['file'],
                'rootvolume-security-style' => "$sec_style",
                language => 'C',
                'snapshot-policy' => 'default',
                'antivirus-on-access-policy' => 'default'));


    # Enable IPv6 if it is requested.
    $server_obj->network_options_ipv6_modify("enabled"=>"true") if (correct_address_family() eq "ipv6");
}

#----------------------------------------------------------------------------------------
##1. If the user gives INTERFACE_UNDER_TEST, and it has the correct ADDR_TYPE address, use it.
##2. If INTERFACE_UNDER_TEST is not given, try to pick the first best possible address from the filer,
##   using network interface show -vserver $::VSERVER_GX_NAME -address-family $::ADDR_TYPE -home-node $::FILERA -role data
##3. If there is none available, check if there is any DATA_IP defined in tharnhosts for $::FILERA for correct ADDR_TYPE.
##   Configure that ip in the filer as a new interface, and use that address.
##4. If that is also unavailable, throw up config error.
#----------------------------------------------------------------------------------------
sub create_gx_net_int ($) {

    my $opt = shift;
    _get_server();
    _get_node();

    my $address = get_data_ip_cmode(undef, undef, 1);
    if ($address) { return; }

    # No INTERFACE_UNDER_TEST, and no interface in filer :( let us see if tharnhosts helps.
    ## Get the data port and data ip from the tharnhost file of the node
    ## TBD: For multi lif we need to address how to handle
    my ($dport, $hostrec, $dmask);
    $hostrec = $node_obj->{hostrec};

    $dport = $::DATA_PORT || $hostrec->{DATA_PORT1};
    $dport || ($dport = $hostrec->{DATA_PORT});
    $dmask = $::DATA_MASK || $hostrec->{DATA_MASK1};
    $dmask || ($dmask = $hostrec->{DATA_MASK});
    $opt->{hostname} = $::DATA_LIF_IP || $hostrec->net_best_address(include=>({nettype=>'data', family=>correct_addr_type()}));

    if (!$dport or !$dmask or !$opt->{hostname}) {
        $Log->warn(qq{Tharnhost file for $::FILER doesnot contain one of these values: DATA_PORT($dport), DATA_IP(} . $opt->{hostname} . qq{) or DATA_MASK($dmask)});
        croak (qq{Couldnot get a proper IP address or port for data interface after all attempts. Please do any of the following:
        1. Create a data vif in vserver:$::VSERVER_GX_NAME, and pass it as:INTERFACE_UNDER_TEST (or)
        2. Create a data vif for vserver:$::VSERVER_GX_NAME and rerun this script (or)
        3. Update the tharnhost file with proper values for the fields: DATA_PORT, DATA_IP and DATA_MASK (or)
        4. Provide the parameters: DATA_PORT, DATA_LIF_IP, and DATA_MASK to the script});
    }

    # We have all the data, but no interface name. We will create our own interface :)
    my $intf_name = $::INTERFACE_UNDER_TEST ? $::INTERFACE_UNDER_TEST : "intf_" . $opt->{hostname};

    $Log->comment ("Creating network interface");
    add_to_cleanup (NACL::STask::NetworkInterface->create(command_interface=>$node_obj,
                server    => $::VSERVER_GX_NAME,
                role    => 'data',
                lif        => $intf_name,
                'home-node' => "$::FILER",
                'home-port' => "$dport",
                address => $opt->{hostname},
                netmask => "$dmask"));

}

sub create_gx_volume ( $ ) {
    my $opt = shift;
    _get_server();
    _get_node();

    ## Get all the variables either from the calling function or take the default value

    my $vol_name =  ($opt->{vol_name}) ? $opt->{vol_name} : $opts1->{vol_name};
    my $aggr_name =  ($opt->{aggr_name}) ? $opt->{aggr_name} : $opts1->{aggr_name};
    my $vserver_name =  ($opt->{vserver_name}) ? $opt->{vserver_name} : $opts1->{vserver_name};
    my $junction_path =  ($opt->{junc_path}) ? $opt->{junc_path} : $opts1->{junc_path};
    my $export_pol =  ($opt->{export_pol}) ? $opt->{export_pol} : $opts1->{export_pol};

    my $root_vol = ($server_obj->vserver_show(vserver=>$vserver_name, fields=>"rootvolume"))[0]->{rootvolume};
    if ($root_vol) {

        NACL::STask::Volume->modify(command_interface=>$node_obj,
                    volume => "$root_vol",
                    vserver => "$vserver_name",
                    policy => "$export_pol",
                    'unix-permissions' => '---rwxrwxrwx');
    } else {
        croak(qq{Couldnot find the root volume for $vserver_name});
    }

    my @output;
    eval {
        @output = $server_obj->volume_show(volume=>$vol_name, vserver=>$vserver_name);
    };
    if ($#output == 0) {

        if ($output[0]->{'policy'} ne $export_pol) {
            NACL::STask::Volume->modify(command_interface=>$node_obj, volume=>$vol_name, vserver=>$vserver_name, 'policy'=>$export_pol);
        }

        my $actual = $output[0]->{'junction-path'};
        if ($actual ne $junction_path) {
            NACL::STask::Volume->unmount(command_interface=>$node_obj, volume=>$vol_name, vserver=>$vserver_name);
            NACL::STask::Volume->online(command_interface=>$node_obj, volume=>$vol_name, vserver=>$vserver_name);
            NACL::STask::Volume->mount(command_interface=>$node_obj, volume=>$vol_name, vserver=>$vserver_name, 'junction-path'=>$junction_path);
        }
        return $junction_path;

    }


    ## Call cdef to create the volume
    $Log->comment ("Creating Volume ");
    add_to_cleanup(NACL::STask::Volume->create(command_interface=>$node_obj,
                    aggregate => "$aggr_name",
                    volume => "$vol_name",
                    vserver => "$vserver_name",
                    size => $::VOL_GX_SIZE,
                    'junction-path' => "$junction_path",
                    policy => "$export_pol",
                    'junction-active' => 'true',
                    type => 'RW',
                    'unix-permissions' => '---rwxrwxrwx'));

}

sub create_gx_export_rule($) {
    my $opt = shift;
    _get_server();
    _get_node();

    ## Get all the variables either from the calling function or take the default value

    my $vserver_name =  ($opt->{vserver_name}) ? $opt->{vserver_name} : $opts1->{vserver_name};
    my $export_pol =  ($opt->{export_pol}) ? $opt->{export_pol} : $opts1->{export_pol};
    my $clnt_match = ($opt->{export_rule}) ? $opt->{export_rule} : $opts1->{export_rule};
    my $ro_rule = ($opt->{export_rule_ro}) ? $opt->{export_rule_ro} : $opts1->{export_rule_ro};
    my $rw_rule = ($opt->{export_rule_rw}) ? $opt->{export_rule_rw} : $opts1->{export_rule_rw};
    my $prot = ($opt->{export_proto}) ? $opt->{export_proto} : $opts1->{export_proto};
    my $anon = ($opt->{export_anon}) ? $opt->{export_anon} : $opts1->{export_anon};
    my $superuser = ($opt->{superuser}) ? $opt->{superuser} : $opts1->{superuser};

    # Save the current state of the filer:
    eval {
        $opts1->{vserver_nfs_old_state} = ($server_obj->vserver_nfs_show(vserver=>$::VSERVER_GX_NAME))[0];
    };
    if ($@) {
        if ($@ !~ /This table is currently empty|There are no entries matching your query|is an invalid value for field/) {
            croak $@;
        }
    } else {
        $opts1->{root_vol} = $server_obj->vserver_show(vserver=>$::VSERVER_GX_NAME)->[0]->{"rootvolume"};
        $Log->comment(qq{vserver_show_out: } . Dumper($server_obj->vserver_show(vserver=>$::VSERVER_GX_NAME)));
        $opts1->{root_original_policy} = $server_obj->volume_show(vserver=>$::VSERVER_GX_NAME, volume=>$opts1->{root_vol})->[0]->{policy};
        if ($server_obj->volume_exists(vserver=>$::VSERVER_GX_NAME, volume=>$::FILER_GX_VOL)) {
            $opts1->{original_policy} = $server_obj->volume_show(vserver=>$::VSERVER_GX_NAME, volume=>$::FILER_GX_VOL)->[0]->{policy};
        }
    }

    if (!$server_obj->vserver_export_policy_exists(vserver=>$vserver_name, policyname=>$export_pol)) {
        $Log->comment ("Creating a export rule");
        ## Create the export policy
        add_to_cleanup (NACL::STask::VserverExportPolicy->create(command_interface=>$node_obj,
            vserver => "$vserver_name",
            policyname => "$export_pol",
            nacltask_if_exists => 'reuse'));
    }

    eval {
        NACL::C::ExportPolicyRule->delete(command_interface=>$node_obj,
                vserver => "$vserver_name",
                policyname => "$export_pol",
                ruleindex => '*');
    };

    ## Create rules for the policy
    #add_to_cleanup (NACL::C::ExportPolicyRule->create(command_interface=>$node_obj,
    NACL::C::ExportPolicyRule->create(command_interface=>$node_obj,
                vserver => "$vserver_name",
                policyname => "$export_pol",
                clientmatch => "$clnt_match",
                superuser => [$superuser],
                rorule => ["$ro_rule"],
                rwrule => ["$rw_rule"],
                anon => "$anon",
                ruleindex => $opts1->{large_rule_index},
                protocol => ["$prot"]);
}

sub create_gx_users_and_group($) {

    my $obj;
    try {
        NACL::STask::VserverServicesUnixGroup->find(command_interface=>$node_obj,
            filter => { vserver => $opts1->{vserver_name}, id => $::NOBODY_GROUP_ID });
    } catch NACL::Exceptions::NoElementsFound with {
        $obj = NACL::STask::VserverServicesUnixGroup->create(command_interface=>$node_obj,
            vserver => $::VSERVER_GX_NAME,
            id => $::NOBODY_GROUP_ID,
            'name' => $::NOBODY_GROUP);
        add_to_cleanup ($obj);
    };

    try {
        NACL::STask::VserverServicesUnixUser->find(command_interface=>$node_obj,
            filter => { vserver => $opts1->{vserver_name}, id => $::EXPORT_ANON });
    } catch NACL::Exceptions::NoElementsFound with {
        $obj = NACL::STask::VserverServicesUnixUser->create(command_interface=>$node_obj,
            vserver => $::VSERVER_GX_NAME,
            id => $::NOBODY_USER_ID,
            'primary-gid' => $::NOBODY_GROUP_ID,
            'user'    => "nobody");
        add_to_cleanup($obj);
    };
}

sub create_nfs_server($) {
    my $opt = shift;
    my $on = 'enabled';
    my $off = 'disabled';
    my %nfs_opts;

    my $vserver_name =  ($opt->{vserver_name}) ? $opt->{vserver_name} : $opts1->{vserver_name};
    my $id_domain =  ($opt->{id_domain}) ? $opt->{id_domain} : $opts1->{id_domain};
    foreach ($opt->{allowed}) {
        $nfs_opts{$_} = $on;
    }

    $Log->comment(qq{nfs_opts =} . Dumper(\%nfs_opts));

    %nfs_opts = (
        access => "true",
        'rpcsec-ctx-high' => 0,
        'rpcsec-ctx-idle' => 0,
        v3 => ($opt->{version} =~ /^(3|all|not[24])$/ ) ? $on : $off,
        v4 => ($opt->{version} =~ /^(4|all|not[23])$/ ) ? $on : $off,
        "v4.1" => ($opt->{minor_version} =~ /^(both|1)$/) ? $on : $off,
        tcp => $on,
        udp => $on,
        "v4." . $opt->{minor_version} . '-read-delegation' => "$on",
        "v4." . $opt->{minor_version} . '-write-delegation' => "$on",
        "v4." . $opt->{minor_version} . '-acl' => "$on",
        'v4-id-domain' => "$id_domain",
        'mount-rootonly' => 'disabled',
        'nfs-rootonly' => 'disabled',
    );

    _get_server();
    _get_node();
    if ($server_obj->vserver_nfs_exists(vserver=>$vserver_name)) {
        NACL::STask::VserverNfs->modify(command_interface=>$node_obj, vserver=>$vserver_name, %nfs_opts);
    } else {
        ## Create NFS server for this vserver
        logStep("Configuring nfs server");
        add_to_cleanup (NACL::STask::VserverNfs->create(command_interface=>$node_obj,
                    vserver => "$vserver_name", %nfs_opts));
    }

}

sub correct_addr_type(;$) {
    my $given = shift;

    if (!$given) {
        $given = $::ADDR_TYPE;
    }

    if ($given eq "inet6" or $given eq "ipv6") {
        return "inet6";
    } else {
        return "inet4";
    }
}

sub get_data_ip_cmode (@) {
    my ($cmode_filer, $vserver, $setup_mode) = @_;
    _get_node();

    $vserver = $vserver ? $vserver : ($opts1->{data_ip_cmode} ? return $opts1->{data_ip_cmode} : $::VSERVER_GX_NAME);
    $Log->comment(qq{Entering the block to find data ip from filer itself});

    ## Populate the required parameters needed for creating the network interface
    my %find_opts = (
        server => $vserver,
        'address-family'=> correct_address_family(),
        role   => 'data',
        );

    if ($::INTERFACE_UNDER_TEST ne "") {
        try {
            my $if_obj = NACL::STask::NetworkInterface->find(command_interface=>$node_obj,
                                filter => { lif => $::INTERFACE_UNDER_TEST, %find_opts });
            $opts1->{data_ip_cmode} = $if_obj->state()->{address};
        } catch NACL::Exceptions::NoElementsFound with {
            $Log->warn(qq{The interface INTERFACE_UNDER_TEST=$::INTERFACE_UNDER_TEST doesn't exist. So the script will try to find the correct data address by other means});
        };
    }

    if ($opts1->{data_ip_cmode}) { return $opts1->{data_ip_cmode}; }

    my $try_now = 2;
    while ($try_now > 0) {
        try {
            my $if_obj = NACL::CS::NetworkInterface->fetch(command_interface=>$node_obj,
                                filter => { %find_opts,
                                    'home-node' => $node_obj->name(),
                                    'curr-node' => $node_obj->name(),
                                    'status-admin'  => 'up',
                                },
                                requested_fields => ['address']
                                );
            $opts1->{data_ip_cmode} = $if_obj->address();
            $try_now = -1;
        } catch NACL::Exceptions::NoElementsFound with {
            $Log->comment(qq{trying a revert and checking again...});
        };
        try {
            NACL::STask::NetworkInterface->revert(command_interface=>$node_obj, lif=>"*",
                vserver=>$vserver);
        } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            $Log->comment(qq{The vserver has no network interface configured. We have to check the tharnhost files, and configure new!});
            $try_now=-1;
        };
        $try_now--;
    }

    if ($opts1->{data_ip_cmode}) { return $opts1->{data_ip_cmode}; }

    die config_result ("Couldn't get the Data address from the filer:$main::FILER., please run with SETUP_DONE=0 or make sure all the required config is done in the filer.")
        if (!$setup_mode);
}

sub create_dns() {
    _get_node();
    try {
        NACL::STask::VserverServicesDns->find(command_interface=>$node_obj,
            filter => { vserver => $::VSERVER_GX_NAME } );
    } catch NACL::Exceptions::NoElementsFound with {
        add_to_cleanup (NACL::STask::VserverServicesDns->create(command_interface=>$node_obj,
            vserver=>$::VSERVER_GX_NAME,
            domains=>[$::DNS_DOMAIN],
            'name-servers'=>[$::DNS_SERVER]));
    };
}

sub create_nis() {
    _get_node();
    try {
        NACL::STask::VserverServicesNisDomain->find(command_interface=>$node_obj,
                    filter => {vserver => $::VSERVER_GX_NAME} );
    } catch NACL::Exceptions::NoElementsFound with {
        add_to_cleanup (NACL::STask::VserverServicesNisDomain->create(command_interface=>$node_obj,
                    vserver=>$::VSERVER_GX_NAME,
                    domain=>$::NIS_DOMAIN,
                    active=>"true",
                    servers=>[$::NIS_SERVER]));

    };
}

sub clean_env_gx() {
    $Log->comment("Entering the generic cleanup routine");

    _get_server();
    _get_node();
    if ($opts1->{vserver_nfs_old_state}) {
        $Log->comment("Restoring vserver old state");
        $server_obj->vserver_nfs_modify(vserver => $main::VSERVER_GX_NAME,
                'v4.0' => $opts1->{vserver_nfs_old_state}->{'v4.0'},
                'v4.1' => $opts1->{vserver_nfs_old_state}->{'v4.1'},
                'v4.0-read-delegation' => $opts1->{vserver_nfs_old_state}->{'v4.0-read-delegation'},
                'v4.1-read-delegation' => $opts1->{vserver_nfs_old_state}->{'v4.1-read-delegation'},
                'v4.0-write-delegation' => $opts1->{vserver_nfs_old_state}->{'v4.0-write-delegation'},
                'v4.1-write-delegation' => $opts1->{vserver_nfs_old_state}->{'v4.1-write-delegation'},
                'v4.0-acl' => $opts1->{vserver_nfs_old_state}->{'v4.0-acl'},
                'v4.1-acl' => $opts1->{vserver_nfs_old_state}->{'v4.1-acl'});
    }

    if ($opts1->{root_vol} && $opts1->{root_original_policy}) {
        $server_obj->volume_modify(volume=>$opts1->{root_vol}, vserver=>$::VSERVER_GX_NAME, policy=>$opts1->{root_original_policy});
    }
    if ($opts1->{original_policy}) {
        $Log->comment("Restoring original export policy");
        $server_obj->volume_modify(volume=>$main::FILER_GX_VOL, vserver=>$::VSERVER_GX_NAME, policy=>$opts1->{original_policy});
    }

    if ($::SETUP_DONE) {
        # If the setup was not done by the script, don't do any cleanup either.
        return;
    }

    if (!$::CLEANUP) {
        # The user has asked not to cleanup. So don't.
        return;
    }

    $Log->comment(qq{NFS_SETUP_CLEANUP has : } . (1+$#NFS_SETUP_CLEANUP) . qq{ objects});
    while (my $obj = pop(@NFS_SETUP_CLEANUP)) {
        if ($obj) {
            if ($obj->can("purge")) {
                $obj->purge();
            } else {
                $obj->delete();
            }
        }
    }

    $Log->exit();
}


#)]}
1;
