# $Id: //depot/prod/test/Rsierranevada/lib/plat_pcycle.pm#1 $
## @summary Library to power cycle filer/cache, and/or device(s)

## @description
##   This library telnets to a remote power box, and then performs
##   the requested operation which typically is power off/on a
##   designated set of plugs. At test completion, an e-mail is
##   sent indicating results.
##
##   The typical configuration is a filer/cache, and/or a set of
##   shelves plugged into a network-attached remote power device.
##     
##   For now, the only supported remote power device is the WTI NPS-style
##   unit.
##
##   Various options are available to set sleep delays between power cycle
##   iterations, specify the number of iterations, specify a particular
##   message, and so forth.
##
##   NOTE 1: This library can be invoked via NATE thpl.
##
##   NOTE 2: This library is testbed compliant.
##
##   NOTE 3: If any of the *REQUIRED* parameters are supplied in the respective 
##           host's tharnhost file, then they don't need to be supplied when 
##           running the script via command line. 
## 
##   NOTE 4: If any of the *REQUIRED* parameters are supplied in the respective
##           host's testbed.info file, then they don't need to be supplied when
##           running the script via the testbed environment.
 
## @param FILER Required: hostname of filer or cache

## @param DEVICE Optional: name of the device 

## @param FILER_RPB Required: hostname of the remote power box the filer/cache is plugged into

## @param DEVICE_RPB Optional: hostname of the remote power box the device is plugged into
##        Required if DEVICE is specified.

## @param FILER_RPB_TYPE Required: numeric value which identifies the remote power box the
##        filer/cache is plugged into. The remote power boxes are identified as follows:
##        1 = WTI NPS-style 
##        2 = Sentry; (currently not supported)

## @param DEVICE_RPB_TYPE Optional: numeric value which identifies the remote power box the
##        device is plugged into.  Refer to the FILER_RPB_TYPE param for complete information.
##        Required if DEVICE is specified.

## @param FILER_PLUGS Required: plug outlet number(s) on the remote power box the filer
##        is plugged into.  NOTE:  Multiple plugs are separated using "+" symbol 

## @param DEVICE_PLUGS Optional: plug outlet number(s) on the remote power box the device
##        is plugged into.  Refer to the FILER_PLUGS param for complete information.
##        Required if DEVICE is specified.

## @param ITERATION Optional: positive integer indicating the number of times the power cycle
##        process will be performed. Default = -1 (indefinite)

## @param SLEEP Optional: positive integer indicating the number of seconds to sleep between
##        iterations. Default = 300    NOTE: The word "RANDOM" can also be specified which
##        generates a random value between 0 and 120.

## @param ACTION Optional: operation performed by the remote power box.  Typically these 
##        devices can perform three functions: off, on, and boot (off then on). 
##        Default = boot

## @param MESSAGE_STRING Optional: text string used to determine the success of the power cycle
##        script. Default = "System rebooted after power-on".  Because this parameter is
##        critical to test functionality, the value specified must be any text or partial text
##        displayed on the console.  The test will also check for the following text, and if 
##        detected will display a message and terminate:
##          - PANIC
##          - Replayed "n" checksum blocks 

## @example
## ntest -hw FILER=plat-fas980-5 ITERATION=5 SLEEP=25 platform/plat_pcycle.thpl
## ntest -hw FILER=plat-fas270 DEVICE=shelf-1 DEVICE_RPB=plat-rpb3 DEVICE_RPB_TYPE=1 DEVICE_PLUGS=5+6+7+8 platform/plat_pcycle.thpl
## ntest -hw FILER=plat-fas250 SLEEP=RANDOM platform/plat_pcycle.thpl

## @author dl-sar-qa@netapp.com, dl-step@netapp.com

use Tharn;

use strict;
use warnings;
use Net::Telnet;
use IO::File;
use Mail::Sendmail qw(sendmail); 

use base qw(Exporter);
BEGIN {
    our %EXPORT_TAGS = (
			all => [ qw(
				    process_parameters
				    do_power_cycle_tc
				    do_power_cycle_evt 
				    )
				 ] );
    Exporter::export_ok_tags('all');
}   


#---------------------------------
# the table is used for APT's menu
#---------------------------------
our %powercycle_table = (
			 0    => \&demo_ts_cfg, 	  # '_cfg' is key
			 1    => \&do_power_cycle_tc, # power cycle test case
			 2    => \&do_power_cycle_evt # power cycle event
			 );

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = (%powercycle_table);


#Following variables are relative to the filer and/or device, and other power cycle specific functions
use vars qw($FILER $filer $DEVICE $device $FILER_RPB $filer_rpb $DEVICE_RPB $device_rpb $FILER_RPB_TYPE $filer_rpb_type $DEVICE_RPB_TYPE $device_rpb_type $FILER_PLUGS $filer_plugs $DEVICE_PLUGS $device_plugs $ITERATION $iteration $SLEEP $sleep $ACTION $action $MESSAGE_STRING $message_string $LOGDIR $logdir $CONSOLE_LOG $console_log $plugs);


#Following variables are relative to the telnet process used for the remote power box
use vars qw($telnet $telnet_loop $telnet1 $telnet1_loop $telnet2 $telnet2_loop);


#Following variables are relative to either the send_mail or test_failure subroutines
use vars qw($error_message $text $result $to $LOGNAME $logname $failed);


#Following variables are general flags,loop counters, etc.  
use vars qw($failed $counter $pass $current_time $random $indefinite $reboot_reason $replay_checksum $event);

param('FILER',-default,'Filer');
my $fh = host($FILER);

param('DEVICE',-default,$fh->{DEVICE});
param('FILER_RPB',-default,$fh->{FILER_RPB});
param('DEVICE_RPB',-default,$fh->{DEVICE_RPB});
param('FILER_RPB_TYPE',-default,$fh->{FILER_RPB_TYPE});
param('DEVICE_RPB_TYPE',-default,$fh->{DEVICE_RPB_TYPE});
param('FILER_PLUGS',-default,$fh->{FILER_PLUGS});
param('DEVICE_PLUGS',-default,$fh->{DEVICE_PLUGS});
param('ITERATION',-default,"-1");
param('SLEEP',-default,"300");
param('ACTION',-default,"boot");
param('MESSAGE_STRING',-default,"System rebooted after power-on.");
param('CONSOLE_LOG',-default,"$LOGDIR/console-$FILER.txt");
param('LOGNAME',-default,"LOGNAME");


#Misc. variables
my $pcycle_counter=2;
my $from=qw(PLAT_PCYCLE.THPL@netapp.com);

 
sub demo_ts_cfg {
}


#This subroutine sends mail to the user.  The mail body contains the test results.
sub send_mail() {
      
    if ($logname) {
	$to = join "@",$logname,"netapp.com";
    }
    else {
        $to = join "@",getlogin(),"netapp.com";
    }
    
    my %text_message = (
			filer_fail => "The PLAT_PCYCLE test failed.  Performed $counter out of $pass power cycles successfully on filer $filer prior to the failure.  Review the log files located at $logdir for further information",
			filer_device_fail => "The PLAT_PCYCLE test failed.  Performed $counter out of $pass power cycles successfully on filer $filer and device $device prior to the failure.  Review the log files located at $logdir for further information",
			filer_pass => "The PLAT_PCYCLE test passed.  Performed $counter out of $pass power cycles successfully on filer $filer.  If desired, review the log files located at $logdir for further information.",
			filer_device_pass => "The PLAT_PCYCLE test passed.  Performed $counter out of $pass power cycles successfully on filer $filer and device $device.  If desired, review the log files located at $logdir for further information.",
			);           
    
    sendmail(To => $to,
             From => $from,
             Subject => "PLAT-PCYCLE Test $result",
             Body => $text_message{$text})
	or die "Error trying to send mail to $to";
}     


#This subroutine handles test failures.
sub test_failure() {
    print $error_message; 
    --$counter;
    
    $result = "Failed";
    if ($device) {     
        $text = "filer_device_fail";
        send_mail();
    }
    else {
        $text = "filer_fail";
        send_mail();
    }		 
}


#This subroutine verifies the parameter list, as good as realistically possible.  If any required parameters aren't supplied, an appropriate message is written to the NATE log file, and the test aborts.
sub process_parameters() {

    
    print "\nParameter list is as follows:\n";

    print "FILER = $filer, DEVICE = $device, FILER_RPB = $filer_rpb, DEVICE_RPB = $device_rpb\n";
    print "FILER_RPB_TYPE = $filer_rpb_type, DEVICE_RPB_TYPE = $device_rpb_type, FILER_PLUGS = $filer_plugs, DEVICE_PLUGS = $device_plugs\n";
    print "ITERATION = $iteration, ACTION = $action, SLEEP = $sleep, MESSAGE_STRING = $message_string\n";
    print "LOGDIR = $logdir\n";
    print "CONSOLE_LOG = $console_log\n";
    print "LOGNAME = $logname\n"; 

#Verify filer parameters; fail if any aren't provided
    if (!$filer) {
	die "A filername has not been provided";
    }
    
    if (!$filer_rpb) {
        die "The filer remote power box name has not been provided";
    }   
    
    if (!$filer_rpb_type) {
        die "The filer remote power box type has not been provided.  Valid entries are: 1 = WTI/NPS, and 2 = Sentry";
    }
    
    if (!$filer_plugs) {
        die "The filer remote power box plug(s) has/have not been provided";
    }
       
#Verify device parameters (if applicable); either all are present or fail accordingly
    if (($device) && ((!$device_rpb) || (!$device_rpb_type) || (!$device_plugs))) {
	die "Incomplete list of parameters supplied for device: $device.  Require DEVICE, DEVICE_RPB, DEVICE_RPB_TYPE, and DEVICE_PLUGS values.";
    }  

#Setup random flag; sleep random seconds between power cycles
    $random = 0;
    if ($sleep eq "RANDOM") {
	$random = 1;
    }

#Setup number of iterations; -1 = forever
    $pass = "";
    if ($iteration > 0) {
	$pass = $iteration;
    }
    else {
	$pass = "indefinite"; 
    } 

#Verify remote power box type is supported.  For now, just supporting the WTI NPS-style box
    if ($filer_rpb_type != 1) {
	die "Remote power box type is not supported. Currently only the WTI NPS-style box is supported.";
    }    

    if ($device) {
	if ($device_rpb != 1) {
	    die "Remote power box type is not supported. Currently only the WTI NPS-style box is supported.";
	}
    }
}  


#Check results of power cycle
sub check_results() {

#Check the console log(s) for errors, NVRAM log replay, and to ensure the requested message(s) is/are displayed.    
    my $console = IO::File->new($console_log, O_RDONLY) or 
	die "Unable to open console file: $console_log!";
      
    $reboot_reason = 0;
    $replay_checksum = 0;
    $failed = 0;
  
    while (my $line = <$console>) {
	if ($line =~ /Replayed (\d|\d\d|\d\d\d|\d\d\d\d) checksum blocks/) {
	    $replay_checksum++;
	}
	
	if ($line =~ /$message_string/) {
	    $reboot_reason++;
	}             
	
	if ($line =~ /PANIC/) {
	    $error_message = "Test Failure!!! System paniced. Review the console log located at $console_log for further information.\n";
	    test_failure();
	    $failed = 1;               
	}
        
	if ($line =~ /DUMPCORE: START/) {
	    $error_message = "Test Failure!!! System took a coredump. Review the console log located at $console_log for further information.\n";
	    test_failure();
	    $failed = 1;
	}
    }

#Determine whether test passed or not by comparing: 1) the number of NVRAM log replays against the number of passes, and 2) the number of appropriate messages against the number of passes.  If okay, write an appropriate message to the log file, and continue.  If not, generate an appropriate message, send e-mail, and abort the test.
    if ($counter > $replay_checksum) {
	$error_message = "\nTest Failure!!! System rebooted after power cycling without replaying the NVRAM log.  Review the log files located at $logdir for further information.\n"; 
	test_failure();
	$failed = 1;
    }
    
    if ($counter > $reboot_reason) {
	$error_message = "\nTest Failure!!! System rebooted after power cycling without displaying the message: $message_string.  Review the log files located at $logdir for further information.\n";
	test_failure();
	$failed = 1;
    }
}


#The main power cycle routine.  
sub do_power_cycle {
  
#Start the power cycle loop.
    $failed = 0;
    $counter = 1;
    while (($counter <= $iteration) || ($pass)) {

#If $random, generate a random number between 0 and 120
	if ($random) {
	    $sleep = int(rand 120);
	}   
      
#Generate and print a timestamp      
	$current_time = time();
	print "\nBegin pass $counter: ",scalar localtime($current_time),"\n";
                
#Power cycle filer and/or device, sleep "n" seconds, then power cycle filer and/or device again.
	for (my $i = 1; $i <= $pcycle_counter; $i++) {
	    if ((!$device_rpb) || ($filer_rpb eq $device_rpb)) {   
	        if ($device_plugs) {
                    $plugs = join '+',$filer_plugs,$device_plugs;   
		    
                } 
                else {
                    $plugs = $filer_plugs;
                }
                
#If necessary, loop until either the telnet connection is established or we've made 50 attempts
#Sleep for 5 seconds between each telnet connection retry
                $telnet_loop = 0;
                do {        
                    $telnet = new Net::Telnet (Host=>$filer_rpb,
					       Timeout=>60,
                                               Errmode=>"return"
					       );
		    if (!$telnet ) {
                        $telnet_loop++;
		        sleep 5;
                    }
	        } until (($telnet) || ($telnet_loop > 50));
		
                if ($telnet_loop > 50) {
		    die "Unable to establish telnet connection to $filer_rpb";
                } 

#If necessary use the following syntax to display all telnet connection and command output.
#		$telnet->input_log(\*STDOUT);

                $telnet->waitfor('/> $/');
                $telnet->print("/$action $plugs $/");  
                $telnet->waitfor('/> $/');
                $telnet->print("/x");
                $telnet->waitfor('/Disconnected.$/');
                $telnet->close; 
	    }
	    elsif ($device_rpb) { 
         	$telnet1_loop = 0;
                $telnet2_loop = 0;
                do {        
                    $telnet1 = new Net::Telnet (Host=>$filer_rpb,
                                                Timeout=>60,
		                                Errmode=>"return"
                                                );           
                    $telnet2 = new Net::Telnet (Host=>$device_rpb,
                                                Timeout=>60,
		         		        Errmode=>"return"
                                                );
		    if (!$telnet1) {
                        $telnet1_loop++;
			sleep 5;
                        if ($telnet2) {
                            $telnet2->close;  
		        }
                    }
                    elsif (!$telnet2) {                        
			$telnet2_loop++;                        			
			sleep 5;
                        if ($telnet1) {
                            $telnet1->close;   
                        }
                    } 
		} until (($telnet1) && ($telnet2)) || (($telnet1_loop > 50) || ($telnet2_loop > 50));
		
		if ($telnet1_loop > 50) {
		    die "Unable to establish telnet connection to $filer_rpb.";
		}
		elsif ($telnet2_loop > 50) {
		    die "Unable to establish telnet connection to $device_rpb.";
		}  
		
#If necessary use the following syntax to display all telnet connection and command output. 
#$telnet1->input_log(\*STDOUT);
#$telnet2->input_log(\*STDOUT);

		$telnet1->waitfor('/> $/');
		$telnet2->waitfor('/> $/');
		$telnet1->print("/$action $filer_plugs $/");  
		$telnet2->print("/$action $device_plugs $/");
		$telnet2->waitfor('/> $/');
		$telnet2->print("/x"); 
		$telnet2->waitfor('/Disconnected.$/');
		$telnet1->waitfor('/> $/');
		$telnet1->print("/x");  
		$telnet1->waitfor('/Disconnected.$/'); 
		$telnet2->close;
		$telnet1->close;               
	    }                                  
	    if ($i < $pcycle_counter) { 
		sleep $sleep;
	    }
	}  
	
	sleep 360;
	
	$current_time = time();
	print "End pass $counter: ",scalar localtime($current_time),"\n";
	print "Slept $sleep seconds between power cycles.\n";
	
	if (!$event) {
	    check_results();
	    if ($failed) {
		return 1;
	    }
	}   
	
#Check to see if we're done or not.  If yes, write an appropriate message to the log file, send an e-mail, and terminate the test.
	if ($counter == $iteration) {                      
	    if (!$event) {
		$result = "Passed";
		$text = "filer_pass"; 
		if (!$device) {                  
		    send_mail();
		}
		else {  
		    $text = "filer_device_pass";
		    send_mail;
		}
	    }
	    print "Completed $counter out of $pass iterations successfully.\n";
	    print "\nTest passed.  Performed $counter power cycles successfully.\n";
	    return;
	} 

#Not done. Bump counter and continue . . . 
	print "Completed $counter out of $pass iterations successfully.\n";          
	$counter++;
    }
}


#The main entry point when running via command line, testX, or when running as a "testcase" under APT.
sub do_power_cycle_tc($$$$$$$$$$$$$$$) {
    
    my ($flr,$dev,$flr_rpb,$dev_rpb,$flr_rpb_type,$dev_rpb_type,$flr_plugs,$dev_plugs,$iter,$slp,$act,$mess_string,$lgdir,$con_log,$lgname) = @_;
    
    $filer = $FILER || $flr || "";
    $device = $DEVICE || $dev || "";
    $filer_rpb = $FILER_RPB || $flr_rpb || "";
    $device_rpb = $DEVICE_RPB || $dev_rpb || "";
    $filer_rpb_type = $FILER_RPB_TYPE || $flr_rpb_type || "";
    $device_rpb_type = $DEVICE_RPB_TYPE || $dev_rpb_type || "";
    $filer_plugs = $FILER_PLUGS || $flr_plugs || "";
    $device_plugs = $DEVICE_PLUGS || $dev_plugs || "";
    $iteration = $ITERATION || $iter || -1;
    $sleep = $SLEEP || $slp || 300;
    $action = $ACTION || $act || "boot";
    $message_string = $MESSAGE_STRING || $mess_string || "System rebooted after power-on.";
    $logdir = $LOGDIR || $lgdir || "";
    $console_log = $CONSOLE_LOG || $con_log || "$LOGDIR/console-$FILER.txt";
    $logname = $LOGNAME || $lgname || "";

    process_parameters();
    
    do_power_cycle();
    
    if ($failed) {
	return 1;
    }
}
1;

#The main entry point when running as an "event" under APT
sub do_power_cycle_evt($$$$$$$$$$$$$$$) {
    
    my ($flr,$dev,$flr_rpb,$dev_rpb,$flr_rpb_type,$dev_rpb_type,$flr_plugs,$dev_plugs,$iter,$slp,$act,$mess_string,$lgdir,$con_log,$lgname) = @_;
    
    $filer = $FILER || $flr || "";
    $device = $DEVICE || $dev || "";
    $filer_rpb = $FILER_RPB || $flr_rpb || "";
    $device_rpb = $DEVICE_RPB || $dev_rpb || "";
    $filer_rpb_type = $FILER_RPB_TYPE || $flr_rpb_type || "";
    $device_rpb_type = $DEVICE_RPB_TYPE || $dev_rpb_type || "";
    $filer_plugs = $FILER_PLUGS || $flr_plugs || "";
    $device_plugs = $DEVICE_PLUGS || $dev_plugs || "";
    $iteration = $ITERATION || $iter || -1;
    $sleep = $SLEEP || $slp || 300;
    $action = $ACTION || $act || "boot";
    $message_string = $MESSAGE_STRING || $mess_string || "System rebooted after power-on.";
    $logdir = $LOGDIR || $lgdir || "";
    $console_log = $CONSOLE_LOG || $con_log || "$LOGDIR/console-$FILER.txt";
    $logname = $LOGNAME || $lgname || "";
    
    process_parameters();
    
    $event = 1;
    do_power_cycle();
    return 0;     
}    
1;
