
package platval::Interconnect::HAiosetup;
use HApackage;
our @ISA = qw( HApackage );
use strict;
use warnings;

use Tharn qw($Log subtest host);
use Params::Validate qw(:all);
use NACL::APISet;
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::C::Vserver;
use NACL::C::License;
use NACL::Exceptions::NoElementsFound qw(:try);
use NACL::STask::NetworkInterface;
use NACL::STask::Volume;
use NACL::C::ExportPolicy;
use NACL::C::ExportPolicyRule;
use NACL::STask::VserverNFS;
use NACL::STask::Client::Mount;
use NACL::C::Client;
use NACL::STask::Cluster;
use NACL::C::NetworkRoutingGroupsRoute;
use Data::Dumper;
use Switch;
use HAconstants;
use NATE::ProcManager;
use NATE::STAF;
use HAutils qw(get_client);
use NDST::NDST qw(randelem do_shell);

## Verify VS exists. If not, create it
## Verify VIF exists and is bound to VS. If not, create it
## Verify exportruleset RULESETNAME exists. If not, create it
## Verify VOLUME exists on VS. If not, create it
## Verify open-access rule in RULESETNAME. If not, create it
## Verify failover rules for VIF exist to each remaining data port in the SFO
##      pair. If not, create all FO rules needed
## Create mountpoint directory on CLIENT
## Mount VOLUME onto CLIENT using VIF
## Change permissions of mountpoint on CLIENT to 777
## Start IO on back ground
## Usage:
##    my $vserver_io_setup = HAiosetup->new($node)
##    $vserver_io_setup->setup();
sub setup {

    my $self  = shift;
    my %opts = validate(
	@_, 
	{
                vserver => { type => SCALAR , optional=> 1 , default => SFO_VSERVER },
                aggregate => { type => SCALAR , optional=> 1 , default => SFO_AGGR }, 
                volume => {type => SCALAR , optional=> 1 , default => SFO_VOL },
                data_lif => { type => SCALAR , optional=> 1 , default => SFO_DATALIF },
		'ns-switch' => { type => SCALAR , optional=> 1 , default => "file" }, ## Need to change to ARRAYREF	
		security => { type => SCALAR,optional=> 1,  default => "unix" },
		number_of_data_aggrs => { type => SCALAR,optional=> 1,  default => NUM_AGGR },  ## One aggr on each node
                number_of_data_disks => { type => SCALAR,optional=> 1,  default => DISKCOUNT},
                number_of_data_vols =>  { type => SCALAR,optional=> 1,  default =>  NUM_VOL }, ## Two volumes on ecah aggregate
                vol_size => {type => SCALAR,optional=> 1, default => "300G" }, 
                io_method => {type => SCALAR,optional=> 1, default => "hammer" },
                io_duration => {type => SCALAR,optional=> 1, default => 80000 },
	},
    );	

    my $node = $self->{_local};
    my $partner = $self->{_partner};
    my $sfo_vserver_aggr;
    my @sfo_data_aggrs;
    my @sfo_data_vols;
    my $sfo_vserver;
    my @sfo_data_lifs;
    my $sfo_vserver_nfs;
    my $server_gateway;
    my @sfo_export_rule;
    my @sfo_export_policy_rule;
    my @filers;
    my $aggr_count = delete $opts{number_of_data_aggrs};
    my $vol_count = delete $opts{number_of_data_vols};
    my $vol_size = delete $opts{vol_size};
    push ( @filers, $node);
    push ( @filers, $partner);
    
    ########################################################################
    # Ensure aggregate  exists. Create the aggregate if needed
    ########################################################################
    my $root_aggr_name = $opts{aggregate} ."_vserver";
    $sfo_vserver_aggr = NACL::STask::Aggregate->create(
            command_interface => $node,
            aggregate         => $root_aggr_name,  
            if_exists         => 'reuse',
            diskcount         => DISKCOUNT,
        );

    ########################################################################
    # Ensure VS exists. Create the vserver if needed
    ########################################################################
    try {
        $sfo_vserver = NACL::C::Vserver->find(command_interface => $node, 
		filter => { vserver => $opts{vserver} },
	);
    }catch NACL::Exceptions::NoElementsFound with {
        my $vserver_root_volume = $opts{volume} ."_vs_root";
	$sfo_vserver = NACL::C::Vserver->create(
                command_interface           => $node,
                vserver                     => $opts{vserver},
                 rootvolume                 => $vserver_root_volume,
                aggregate                   => $sfo_vserver_aggr,
                "ns-switch"                 => $opts{"ns-switch"},
                "rootvolume-security-style" => $opts{security},
            );
    };

   ########################################################################
   # Ensure at least one data LIF exists from each Filer is bound to this
   # VS. Create the LIF if needed using optionally passed in params. Let
   # ngsh error out if some of the required params to create LIF are not
   # specified
   ########################################################################\
   foreach my $filer ( @filers ) {

       my $lifinfo =  NACL::STask::Node->get_hostrec_lif_info(
    	    command_interface    => $filer,
    	    default_mtu          => DEFAULT_MTU,
    	    required_fields      => [qw( NAME IP MASK GATEWAY PORT )],
    	    verbose_checks       => 1,
        );

        my $data_lif = $lifinfo->{DATA};
        for (my $i = 0 , my $j = 1 ; $i < (scalar (@$data_lif)) ; $i++, $j++) {
           # Create new Lif on the vserver if not exists earlier.
           my $info = $data_lif->[$i];
           $server_gateway= $info->{GATEWAY} ;
           my $data_lif_name = $opts{data_lif} . $j;
           my $node_name = $filer->node();
           $Log->comment("Creating Lif $data_lif_name on node $node_name");
           try {
	       my $sfo_data_lif = NACL::STask::NetworkInterface->create(
                    command_interface => $filer,
                    server            => $sfo_vserver,
                    address           => $info->{IP},
                    lif               => $data_lif_name,
                    role              => 'data',
                    'home-node'       => $filer,
                    'home-port'       => $info->{PORT},
                    netmask           => $info->{MASK}, 
                    nacltask_if_exists => 'reuse',
    	      );
              push(@sfo_data_lifs, $sfo_data_lif);
          }catch NACL::APISet::Exceptions::ResponseException with {
             my $exception_object = shift;
             $Log->comment( "Caught ResponseException in partner with"
              . $exception_object->text() );
          };

        }
    }
    if ( scalar ( @sfo_data_lifs ) == 0 ) {
        $Log->exit();
        NATE::BaseException->throw("Failed to create data lifs");
    }
    ########################################################################
    ## Ensure NFS is created for our VS. Create NFS if needed
    ########################################################################
    $sfo_vserver_nfs = NACL::STask::VserverNFS->create(
             vserver => $sfo_vserver,
             command_interface => $node, 
             if_exists	       => 'reuse',
             access => "true",
              );

    # Find Routing Group for data lif
    my $route;
    my $default_routing_group = $sfo_data_lifs[0]->state()->routing_group();

    ## Create New route for Vserver
    try {
	$route = NACL::C::NetworkRoutingGroupsRoute->find(
        command_interface => $node,
        filter            => {
            server          => $sfo_vserver,
	    'routing-group' => $default_routing_group, }
  	);
    }catch NACL::Exceptions::NoElementsFound with {	
	NACL::C::NetworkRoutingGroupsRoute->create(
        	command_interface => $node,
        	server            => $sfo_vserver,
        	gateway           => $server_gateway,
        	'routing-group'   => $default_routing_group,
        	destination       => '0.0.0.0/0'
    	);
    };

    ########################################################################
    # Ensure ruleset exists for VS. Create exportruleset is needed
    ########################################################################
    try {
	 @sfo_export_rule = NACL::C::ExportPolicy->find(command_interface => $node,
                 filter            => {
                    vserver           => $sfo_vserver,
	            policyname        => 'default',
 		 }
 	);   
    }catch NACL::Exceptions::NoElementsFound with {
    	$sfo_export_rule[0] =
      		NACL::C::ExportPolicy->create(command_interface => $node,
                 vserver           => $sfo_vserver, 
                 policyname        => 'default',
               );
    };
    ########################################################################
    # Ensure export rule within the ruleset. Create new export rule if needed
    ########################################################################
    try {
    	@sfo_export_policy_rule = NACL::C::ExportPolicyRule->find(command_interface => $node,
		filter            => {
                   vserver           => $sfo_vserver,
                   policyname        => 'default',
                   rwrule            => ["any"],
                   rorule            => ["any"], }
	);
    }catch NACL::Exceptions::NoElementsFound with {
	
       $sfo_export_policy_rule[0] = NACL::C::ExportPolicyRule->create(command_interface => $node,
                vserver           => $sfo_vserver,
                policyname        => 'default',
                clientmatch       => "0.0.0.0/0",
                rorule            => ["any"],
                rwrule            => ["any"],
		protocol          => ["any"],
		anon		  => '0',
		superuser	  => ["any"],
		'chown-mode'	  => 'unrestricted',
       );
    };
    ########################################################################
    # Ensure Aggregates for data population  exists on each node, otherwise 
    # Create given number of aggregates on each node
    ########################################################################
    my $count = 0 ;
    while ($aggr_count != 0 ) {
        $Log->comment("Creating Data Aggregate..."); 
        $sfo_data_aggrs[$count] =  NACL::STask::Aggregate->create(
                command_interface => $node,
                aggregate         => $opts{aggregate} . $count,
                if_exists         => 'reuse',
                diskcount         => $opts{number_of_data_disks},
            );
         $count++;
         $sfo_data_aggrs[$count] =  NACL::STask::Aggregate->create(
                command_interface => $partner,
                aggregate         =>  $opts{aggregate} . $count,,
                if_exists         => 'reuse',
                diskcount         => $opts{number_of_data_disks},
            );
         $count++;
         $aggr_count--;
    }  
    
    #########################################################################
    # Ensure that volumes exists on each node, otherwise create given number 
    # of volumes on ecah aggregate.
    ##########################################################################
    ## If exists will be chaged to purge the volume
    my $vol_index = 0;
    foreach my $aggr ( @sfo_data_aggrs ) { 
        my $count = $vol_count;
        while ($count != 0 ) {
            my $vol_name = $opts{volume} . $vol_index;
            my $exception_object;
            LOOP: {
                try {
                    $sfo_data_vols[$vol_index] = NACL::STask::Volume->create(
                    command_interface => $node,
                    volume            => $vol_name,
                    aggregate         => $aggr,
                    vserver           => $sfo_vserver,
                    size              => $vol_size,
		    if_exists         => "purge",
	            'junction-path'   => "/$vol_name",
	            'unix-permissions' => "777",
                );
                } catch NACL::APISet::Exceptions::ResponseException with {
                    $exception_object = shift;
                    if ($exception_object->text() =~ 
                /Request to create volume failed because there is not enough space in the given aggregate/ ) 
                    {
                        ###  Add another aggregate and recreate the volume
                        try {
                            $aggr->add_disks(
                               diskcount          => "1",
                            );
                            redo LOOP;
                        }catch NACL::APISet::Exceptions::ResponseException with {
                             $exception_object = shift;
                             $Log->comment("Failed to add extra disk to aggregate, Cannot create new volume" . $exception_object->text() );
                             $vol_index--;
                        };
                    } 
                    else { 
                       $Log->exit();
                       $exception_object->throw(); 
                    }
                };
            }
            $count--;
            $vol_index++;
        }
    }

    my $clients = get_client();
    my $num_client = scalar(@{$clients});
    my $num_vols = ( scalar ( @sfo_data_vols ));
    my $client_iterator = 0;
    my $lif_iterator = 0;
    my @mount_points;
    my $mount_status = 1;
    my $io_status = 0;
    ## Using round robin method to mount volumes on data lifs and multiple clients
    for(my $index = 0 ; $index < $num_vols ; $index++) {

        $lif_iterator = 0 if ( $lif_iterator == scalar @sfo_data_lifs )  ;
        $client_iterator = 0 if ($client_iterator == $num_client );

        ## Get the IP adress of the lif. 
        my $interface_state = $sfo_data_lifs[$lif_iterator]->state();
        my $data_ip = $interface_state->address();
        my $mount; 
        try {
         $mount = NACL::STask::Client::Mount->create(
                         command_interface=>$clients->[$client_iterator], 
                         server=>$data_ip,
                         vserver => $sfo_vserver, 
                         volume=> $sfo_data_vols[$index] , 
         );
        }catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            my $response         = $exception_object->get_response_object();
            $Log->comment( "Caught ResponseException in partner with"
                     . $exception_object->text() );
            $mount_status = 0;
        };
        if ( $mount_status ) { 
            push(@mount_points , $mount);  
            my $mount_point = $mount->{mount_point};
            my $lif_mnt=
               $sfo_data_lifs[$lif_iterator].":"."/".$sfo_data_vols[$index];
            my $hostname = 
               $clients->[$client_iterator]->hostrec()->{hostname};
            if ( $opts{io_method} eq "hammer"){
            ###### ProcManager controls the execution of IO scripts. 
            ###### independent argument decides whether sub-units are independent
            ###### This module starts IO on all the volumes and exits the execution
            ###### Default IO duration is set to 80000, 1200 is used only for testing.
                my @proc_hammer;
				my $cnt;
				for($cnt=0; $cnt<=5; $cnt++){
					my $mnt_point="$mount_point/mount_point"."$cnt";
					do_shell("/usr/software/bin/sudo mkdir -p $mnt_point");
					do_shell("/usr/software/bin/sudo chmod 0777 $mnt_point");
					try {	
						$proc_hammer[$index] = 
						NATE::ProcManager->new( proc_info => [ { codespec => HAMMER , 
									params => { CLIENT => $hostname , MOUNT_POINT => $mnt_point , RUNTIME => $opts{io_duration}} }], 
									independent => 1 ,client => $hostname );
						$proc_hammer[$index]->start();
						$io_status++;
					} catch NACL::APISet::Exceptions::ResponseException with {
						my $exception_object = shift;
						$Log->comment("Error is: " . $exception_object->text());
						my $resp = $exception_object->get_response_object();
						$Log->comment("Raw output was:\n" . $resp->get_raw_output());
					};
				}
            } ## End of hammer
            elsif($opts{io_method} eq "filersio") {
               $Log->comment("Not yet implemented");
            }## End of filersio
            elsif ($opts{io_method} eq "dt") {
               $Log->comment("Net yet implemented");
            }### End of dt
            elsif ($opts{io_method} eq "populate"){
               $Log->comment("Not yet implemented");
           }### End of populate
           else {
                $Log->exit();
                NACL::Exceptions::InvalidChoice->throw("Invalid Takeover type");
           }### Invalid choice

        } ### end of if mount status
        $client_iterator++;
        $lif_iterator++;
    }## end of for

    if ( $io_status ) {
       $Log->debug("Sucessfully started IO on $io_status volume");
    } else {
       $Log->exit();
       NATE::BaseException->throw("Failed to start IO on any of the volume");
   }
}

## Usage:
##    my $vserver_io_setup = HAiosetup->new($node)
##    $vserver_io_setup->vserver_cleanup();

sub vserver_cleanup() {

    my $self  = shift;
    my %opts = validate(
        @_,
        {
                vserver => { type => SCALAR , optional=> 1 , default => SFO_VSERVER },
                aggregate => { type => SCALAR , optional=> 1 , default => SFO_AGGR },
        }
    ); 

    
    my $sfo_vserver;
    my $sfo_vserver_root_vol;
    my @sfo_data_vols;
    my @sfo_aggrs;
    my @sfo_data_lifs;

    my $node = $self->{_local};
    my $clients = get_client();

    my $staf = NATE::STAF->new(global=>1);

    ### Get The vserver object
    $sfo_vserver = NACL::C::Vserver->find(command_interface => $node,
            filter => { vserver => $opts{vserver} },
    );
    my $volumes_exists = 1;
    ### Get the list of volumes in vserver
    try {
        @sfo_data_vols = NACL::STask::Volume->find(
            command_interface   => $node,
            filter              => {vsroot => 'false' ,vserver => $sfo_vserver},
        );
    }catch NACL::Exceptions::NoElementsFound with {
            $volumes_exists = 0;
            $Log->comment(
"Caught ResponseException in cleanup_funciton when getting volumes : No elements found \n "
            );
    };

    if ( $volumes_exists ) {
        foreach my $client ( @{$clients} ) {
            ## check if IO is running on given clients
            my $endpoint = $clients->[0]->hostrec()->{default_ip};
            my $client_apiset = $client->apiset( category  => "Host", interface => "CLI",  
                                set       => "Linux", );

            my @IO_handles = @{$staf->handle_list(endpoint => $endpoint )}; 
            print Dumper(\@IO_handles);
            foreach my $handle (@IO_handles) { 
                if ($handle->{name} =~ /PM\d+\/\d+\_\s*/ ) {
                    my $io_name = $handle->{name};
                    $Log->comment("$io_name IO is still on progress, stop the IO");
                    my $pid = $handle->{pid};
                    $Log->comment("Killing the pid $pid ...........");
                    my $cmdstring = "/usr/software/bin/sudo kill -9 $pid";
                    $client_apiset->execute_raw_command(command=>$cmdstring);
                }
            } ## End for loop IO_handles
           ## Get the list of mount points mounted on the client.    
           ## Unmount all the volumes mounted on different clients
           my @mount_objs = NACL::STask::Client::Mount->find(command_interface => $client ); 
           foreach my $mount_obj (@mount_objs) {
               foreach my $vol ( @sfo_data_vols ) {
                   my $volume = $vol->volume();
	           if ( $mount_obj->mount_point() =~ /$volume/ ) {
	               $Log->comment("Unmounting the mount point" . $mount_obj->mount_point());
                       try {
                           $mount_obj->purge( nacltask_verify => 1 );	
                       } catch NACL::APISet::Exceptions::ResponseException with {
                             my $exception_object = shift;
            $Log->comment(
"Caught ResponseException in cleanup_funciton when deleting mount points on client: \n "
                  . $exception_object->text() );
    };
                   } ## End of If
               }## End of for each @sfo_data_vols
           }## End of for each @mount_objects
        } ## End of for each @clients
      
        ## Delete all data volumes      
        foreach my $vol ( @sfo_data_vols ) {
            my $name = $vol->volume();
            $Log->comment("Deleting volume $name");
            NACL::STask::Volume->purge(
               command_interface => $node,
               volume            => $vol,
               vserver           => $sfo_vserver,
            );
        } ## End of for loop, deleting SFO volumes
    } ## End if $volume_exists

    ## Get the list  of DATA lifs created for vserver
    my $data_lifs_exists = 1;
    try {
        @sfo_data_lifs = NACL::STask::NetworkInterface->find(
            command_interface => $node,
            filter => {vserver => $sfo_vserver, },
        );
    }catch NACL::Exceptions::NoElementsFound with {
            my $exception_object = shift;
            $data_lifs_exists = 0;
            $Log->comment(
"Caught ResponseException in cleanup_funciton when identyfying data lifs on filer: \n "
                  . $exception_object->text() );
    };

    if ( $data_lifs_exists ) {
        ## Get the default routing group created.
        my $routing_group = $sfo_data_lifs[0]->state()->routing_group();
        my $routing_group_obj;
        my $route_exists=1 ;
        $Log->comment("Deleting the default routing group of the interfaces $routing_group " );
        try {
            $routing_group_obj = NACL::C::NetworkRoutingGroupsRoute->find(
            command_interface => $node,
            filter            => {
            server          => $sfo_vserver,
            'routing-group' => $routing_group, }
          );
        }catch NACL::Exceptions::NoElementsFound with {
            $route_exists = 0;
            $Log->comment(" Routing group $routing_group doesnot exists ");
        };
 
        $routing_group_obj->delete() if ( $route_exists );

        foreach my $lif (@sfo_data_lifs) {
           my $interface = $lif->lif(); 
           $Log->comment("Deleting the interface $interface");
           NACL::STask::NetworkInterface->purge(
                command_interface=>$node ,
                lif       => $lif,
                vserver    => $sfo_vserver,
                nacltask_verify => 1, # default is 0
            );
        } ## End of for each
    }

    #my $root_volume_name = $sfo_vserver_root_vol->volume();
    #$Log->comment("All the data volumes are deleted, Now deleting the vserver root volume $root_volume_name ");
   
    $sfo_vserver_root_vol  = NACL::STask::Volume->find(
            command_interface   => $node,
            filter              => {vsroot => 'true' ,vserver => $sfo_vserver},
        ); 
 
    NACL::STask::Volume->modify(
        command_interface=>$node, 
        volume  => $sfo_vserver_root_vol,
        vserver => $sfo_vserver,
        state => "offline",
    );

    NACL::STask::Volume->delete(
        command_interface=>$node,
        volume  => $sfo_vserver_root_vol,
        vserver => $sfo_vserver,
    );

    my $vserver = $sfo_vserver->vserver();
    $Log->comment("Deleting the vserver $vserver");
     NACL::C::Vserver->delete(
          command_interface => $node,
          vserver           => $sfo_vserver,
     );


   ## Get the lists of  aggregate present on the filer 
   my $aggrs_exists = 1;
   try {
       @sfo_aggrs = NACL::C::Aggregate->find(
           command_interface => $node, 
           filter => { aggregate => $opts{aggregate} . "*" } , 
       );
    }catch NACL::Exceptions::NoElementsFound with {
        $aggrs_exists = 0;
        $Log->comment("No aggregate starting with name $opts{aggregate} found");
    };

    if ( $aggrs_exists ) {
        foreach my $aggr (@sfo_aggrs ) {
            my $aggr_name = $aggr->aggregate();
            $Log->comment(" Deleting Aggregate $aggr_name ..... ");
            try {
                NACL::STask::Aggregate->purge(
                    command_interface => $node,
                    aggregate         => $aggr,
                );
            }catch NACL::APISet::Exceptions::ResponseException with {
                my $exception_object = shift;
            $Log->comment(
"Caught ResponseException in cleanup_funciton when deleting aggregate $aggr_name on filer: \n "
                  . $exception_object->text() );
            };
        }
    }

}

1;	
