#Perl module for performing Takeover.Giveback operations and validation of IC connection throughout the process.
#by Weishi Huang, 04/10/2011
###############################################################################
=head1 NAME

TOGB: Perl module for C-mode takeover and giveback operations under verious triggers

=head1 SYNOPSIS

use IConnect::lib::SFO;
$sfo_obj = SFO->new();

=head1 REQUIRES

None

=head1 EXPORTS

None

=head1 DESCRIPTION

This module is to handle TOGB due to user-issued command

=head1 METHODS

=head2 sfo_on_cmd

=over 4

$sfo_obj->cmd_on_cmd(
          con_obj_a => $con_a,
          con_obj_b => $con_b,
          tran_a => $tran_a,
          tran_b => $tran_b,
          );

required: con_obj_a, connection object for filer that is doing TOGB
          con_obj_b, connection object for filer that is being taken over (rebooted) 
          tran_obj_a, transit object for filer a (TOGB filer)
          tran_obj_b, transit object for filer b (reboot/taken over filer)

optional: graceful, is either 1 or 0 for yes and no. Default is 0 for ungraceful takeover.
          timeout, timeout following takeover command is issued. Default is 180 seconds.

=back

=head1 AUTHOR

Weishi Huang weishi@netapp.com, 4/10/2011

=cut
############################################################################

package platval::Interconnect::SFO;
use Carp;
use Params::Validate qw(:all);
use strict;

#############################################
sub new {
  my $class = shift;
  my $this = {};
  bless $this, $class;
  $this->{err} = '';
  return $this;
}

#=============================================================
## method: takeover on command (either forceful or unforceful)
#=============================================================
sub sfo_on_cmd {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});

  my ($cmd, $out, $ct);
  $cmd='storage failover takeover -bynode local -option force';
  if ($in{forceful} == 0) {
     $cmd = 'storage failover takeover -bynode local -option normal';
  }
  print "Perform takeover by issing '$cmd'\n" unless !$in{debug};
  $out=$in{con_obj_a}->execute(command=>$cmd);
  print "pausing 360 sec after '$cmd'\n" unless !$in{debug};
  Tharn::snooze(360);
  $ct=0;
  while ($out =~ /^error:/i) {
    $ct++;
    print "pausing 20 sec before retry again. retry count=$ct\n";
    Tharn::snooze(20);
    $out=$in{con_obj_a}->execute(command=>$cmd);
    confess ("'$cmd' failed even after $ct retries") if $ct > 5;
  }
#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 100) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_cmd()

#=============================================================
## method: takeover on halt
#=============================================================
sub sfo_on_halt {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});

  my ($cmd, $out, $ct, $err);
  $cmd='system node halt -node local';
  
  print "Perform takeover by issing '$cmd'\n" unless !$in{debug};
  $out=$in{con_obj_b}->execute_start(command=>$cmd);
  print "pausing 60 sec after '$cmd'\n" unless !$in{debug};
  Tharn::snooze(60);
 my $state=$in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
 
 while($state ne 'FIRMWARE'){
    print ("checking for FIRMWARE state\n");
	$ct++;
	if($ct >= 10){
		$err="did not reach FIRMWARE state\n";
		$class->{err}=$err;
		confess($err);
	}
 }
	print "Reached FIRMWARE state\n";
    $in{tran_obj_b}->change_state(to=>'TAKEN_OVER_WAITING_ONTAP');

#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 60) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_halt()

#=============================================================
## method: takeover on reboot
#=============================================================
sub sfo_on_reboot {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});

  my ($cmd, $out, $ct);
  $cmd='system node reboot -node local';
  
  print "Perform takeover by issing '$cmd'\n" unless !$in{debug};
  $out=$in{con_obj_b}->execute_start(command=>$cmd);
  print "pausing 60 sec after '$cmd'\n" unless !$in{debug};
  Tharn::snooze(60);

#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 60) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_reboot()

#=============================================================
## method: takeover on panic
#=============================================================
sub sfo_on_panic {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});

  my ($cmd, $out, $ct);
  $cmd='set test;system node run -node local -command panic';
  
  print "Perform takeover by issing '$cmd'\n" unless !$in{debug};
  $out=$in{con_obj_b}->execute_start(command=>$cmd);
  print "pausing 10 sec after '$cmd'\n" unless !$in{debug};
  Tharn::snooze(10);
  $out=$in{tran_obj_b}->{conn}->execute(command=>' ', timeout=>100, prompts=>['DUMPCORE: START','\\.\\.\\.\\.\\.','DUMPCORE: END']);
  print "waiting for DUMPCORE:END\n";
  $out=$in{tran_obj_b}->{conn}->execute(command=>' ', timeout=>1800, prompts=>['DUMPCORE: END','Press Ctrl-C for Boot Menu']);
  
#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 60) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_panic()

#=============================================================
## method: takeover on power cycle
#=============================================================
sub sfo_on_powercmd {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
		  rlm_con_a=>{type=>OBJECT},
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});
  my ($cmd, $out, $ct);
  print "Starting to perform power cycle\n" unless !$in{debug};
  $in{rlm_con_a}->execute(command=>'system power cycle', timeout=>900, prompts=>['This will cause a dirty shutdown of your appliance']);
  $in{rlm_con_a}->execute(command=>'y');
  print "power cycle is done\n" unless !$in{debug};
  print "pausing 60 sec after 'power cycle'\n" unless !$in{debug};
  Tharn::snooze(60);

#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 60) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_powercmd()

#=============================================================
## method: takeover on reset
#=============================================================
sub sfo_on_reset {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj_a=>{type=>OBJECT},
          tran_obj_a=>{type=>OBJECT},
          con_obj_b=>{type=>OBJECT},
          tran_obj_b=>{type=>OBJECT},
          forceful=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
		  rlm_con_a=>{type=>OBJECT},
            },
          );

  $class->_turnoff_confirm($in{con_obj_a});
  $class->_turnoff_mpage($in{con_obj_a});

  $class->_turnoff_confirm($in{con_obj_b});
  $class->_turnoff_mpage($in{con_obj_b});
  my ($cmd, $out, $ct);
  print "Starting to perform reset\n" unless !$in{debug};
  $in{rlm_con_a}->execute(command=>'system reset', timeout=>900, prompts=>['This will cause a dirty shutdown of your appliance']);
  $in{rlm_con_a}->execute(command=>'y');
  print "reset is done\n" unless !$in{debug};
  print "pausing 60 sec after 'reset'\n" unless !$in{debug};
  Tharn::snooze(60);

#for console connect, no need to expect any output.
# $in{con_obj_a}->execute_expect();
  my ($ct, $state_a, $state_b, $err);
  $ct = 0;
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after TO=$state_a\n" unless !$in{debug};
    print "STATE_B after TO=$state_b\n" unless !$in{debug};
    if ($ct >= 600) {
      $err="Error: Giveback was not ready after waiting $ct sec";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /TAKEN_OVER_WAITING_ONTAP/) {
      last;
    } else {
      print "showing takeover status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>"storage failover show");
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless $in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

# print "checking ic status after TO\n" unless !$in{debug};
# $class->check_ic_status(con_obj=>$in{con_obj_a}, stage=>'after TO');

  $cmd='storage failover giveback -fromnode local -require-partner-waiting false';
  if ($in{forceful}) {
    $out=$in{con_obj_a}->execute(command=>$cmd);
    print "waiting 180 sec for giveback to complete..\n";;
    Tharn::snooze(180);
    $ct=0;
    while ($out =~ /^error:/i) {
      $ct++;
      print "pausing 20 sec before retry again. retry count=$ct\n";
      Tharn::snooze(20);
      $out=$in{con_obj_a}->execute(command=>$cmd);
      confess ("'$cmd' failed even after $ct retries") if $ct > 5;
    }
  } else {
   print "waiting 450 sec for auto giveback to complete..\n" unless !$in{debug};
   Tharn::snooze(450);
  }
  $ct = 0;
# $in{con_obj_a}->execute_expect();
  while (1) {
    $state_a = $in{tran_obj_a}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    $state_b = $in{tran_obj_b}->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
    print "STATE_A after GB=$state_a\n" unless !$in{debug};
    print "STATE_B after GB=$state_b\n" unless !$in{debug};
    if ($ct > 60) {
      $err="Error: Giveback was not complete after waiting $ct sec: STAET_A=$state_a; STATE_B=$state_b";
      $class->{err}=$err;
      confess($err);
    }
    if ($state_a =~ /CLI/ and $state_b =~ /PASSWORD|USERNAME/) {
      last;
    } else {
      print "showing giveback status...\n" unless !$in{debug};
      $out=$in{con_obj_a}->execute(command=>'storage failover show-giveback');
      print "Pausing 10 sec before checking again. Elapsed time=$ct...\n" unless !$in{debug};
      Tharn::snooze(10);
      $ct += 10;
    }
  }

  $class->_restore_2_cli($in{con_obj_a}, $in{tran_obj_b});
  return 0;
} ## end of sfo_on_reset()

#=============================================================
## method: return latest err msg saved on object
#=============================================================
sub show_err {
  my $class = shift;
  return $class->{err};
}

#######################################################
#=============================================================
## method: check_ic_status
#=============================================================
sub check_ic_status {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj=>{type=>OBJECT},
          stage=>{type=>SCALAR,
                  regexp=>/^(before TO|after TO)$/,
                 },
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
             },
          );
  my ($out, $exp);
  if ($in{stage} =~ /before/) {
    $exp='Link 0:.+?up.+IC.+connection.+?up.+NVLOG.+?up';
  } else {
    $exp='Link 0:.+?up.+IC.+connection.+?up.+NVLOG.+?down';
  }

  $out=$in{con_obj}->execute(command=>'priv set -q test');
  $out = $in{con_obj}->execute(command=>'ic status');
  if ($out !~ /$exp/igms) {
    return "Error: ic statu is not expected \($exp\): $out.";
  } else {
    print "Note: ic status is expected \($exp\): $out.\n" unless !$in{debug};
    return 0;
  }
}

#=============================================================
## method: check_sf_status
#=============================================================
sub check_sf_status {
  my $class = shift;
  my %in=validate(
        @_, {
          con_obj=>{type=>OBJECT},
          debug=>{type=>SCALAR,
                optional=>1,
                default=>1,
                regex=>qr/^(1)|(0)$/,
               },
             },
          );
  my ($out, $exp);
  $exp='true\s+Connected.+true\s+Connected';
  $out = $in{con_obj}->execute(command=>'storage failover show; set test -c off; echo -node local __prompt__', timeout=>300, prompts=>['SecD returned __prompt__']);
  if ($out !~ /$exp/igms) {
    return "Error: storage failover status is not expected \($exp\):\n$out.";
  } else {
    print "Note: storage failover status is expected \($exp\):\n$out.\n" unless !$in{debug};
    return 0;
  }
}
#=============================================
sub _restore_2_cli {
  my ($class, $con_obj, $tran_obj) = @_;
  my ($state_b, $out);
  $state_b = $tran_obj->change_state(to=>'CLI');
  $state_b = $tran_obj->get_state(timeout=>300,maxidle=>360,probe_timeout=>[5,400]);
  if ($state_b ne 'CLI') {
    confess("unable to get back to CLI state");
  }
  print "STATE_B after change_state=$state_b is verified.\n";

  print "Check sf status to ensure cf is enabled...";
  $out =  $class->check_sf_status(con_obj=>$con_obj);
  foreach (1..3) {
    if ($out) {
      print "pausing 10 sec before checking cf status again...";
      sleep 10;
      $out =  $class->check_sf_status(con_obj=>$con_obj);
    } else {
      last;
    }
  }
  if ($out) {
    confess("unable to get back to cf enabled status after TOGB and 3 trials.");
  }
} ## end of _restore_2_cli
#=============================================
sub _turnoff_confirm {
  my ($class, $con_obj) = @_;
  my $out=$con_obj->execute(command=>'set -confirmation off');
  return undef;
}
#=============================================
sub _turnoff_mpage {
  my ($class, $con_obj) = @_;
  my $out=$con_obj->execute(command=>'set -rows 0');
  return undef;
}

1,
