# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/platval/platsw/MachineErrors.pm#1 $

#   Copyright (c) 2001-2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
# lib/platval/platsw/MachineErrors.pm

## @summary Base class for all Machine Errors
## @author jkonjeti@netapp.com, dl-pvt-platsw@netapp.com
## @status shared
## @pod here


package MachineErrors;

#
# Compiler Directives
#
use strict;
use warnings;

# Other module imports
use Tharn qw($Log);
use Data::Dumper;

###############################################################################
# Method:    new
# Objective: Constructor to create new MachineErrors object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
	my $class = shift;
	my $machine_errors = {};

        # Bless the hash returned which will make it a Perl object
	bless $machine_errors, $class;
}

###############################################################################
# Method:    mc_extract_machine_errors_csv_file 
# Objective: Provides hash table by extracting data from Machine Errors CSV file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub mc_extract_machine_errors_csv_file ($) {
	my ($machine_errors, $csv_file_name) = @_;
	my ($line, $flag, $sts_reg_flag, $misc_reg_flag, @sts_hdr) ;

	open (FILE, "$csv_file_name") || logresult ('SCRIPT', "Invalid file or File not exist: $csv_file_name");
	while ($line=<FILE>) {
		next if ($line =~ /^#/);		# Skipping to next line, if the line is commented
		$line =~ s/\,+$//;                      # Deleting comma(,) at every end of line
		if ($line =~ /^\s*$/) {                 # Go to next line, if it is a blank line
			$flag = 0;
			$sts_reg_flag = 0;
			$misc_reg_flag = 0;
			next;
		}
		chomp($line);				# Removing newline characters
		$line =~ s/\,\s+|\s+\,/\,/g;		# Removing spaces in every cell

		if ($line =~ /CAP|Error|STATUS/i) {
			$flag = 1;
			$sts_reg_flag = 1 if ($line =~ /STATUS/i);
			$line = <FILE>;				# Header row - Capability register: Platform, address, value
			chomp ($line);
			@sts_hdr = split(/\,/, $line) if ($sts_reg_flag);
			next;
		}
		
		if ($flag) {
			my @d = split(/\,/, $line);
			if ($line =~ /^[a-z]+/i) {
				$machine_errors->{$d[0]} = {'addr' => $d[1], 'banks' => $d[2], 'poll_cli' => $d[3]};
			} elsif ($sts_reg_flag) {
				my @sts_banks = ($d[0]);
				if ($d[0] =~ /(\d+)\-(\d+)/) {
					@sts_banks = ($1..$2);
				} 
				foreach my $b (@sts_banks) {
					my @sts_errs;
					for (my $i=1; $i < scalar(@d); $i++) {
						push @sts_errs, ($sts_hdr[$i] => $d[$i]);
					}
					my $bank_name = $machine_errors->{'bank'}->{$b};
					$machine_errors->{'status'}->{$bank_name} = { @sts_errs };
				}		
			} elsif ($line =~ /^\d+/) {
				$machine_errors->{'bank'}->{$d[0]} = $d[1];
				$machine_errors->{'mc'}->{$d[1]} = {'status' => $d[2], 'addr' => $d[3], 'misc' => $d[4]};
			}
		}

		if ($line =~ /MISC/i) {
			$misc_reg_flag = 1;
			next ;
		}

		if ($misc_reg_flag) {
			my $blank_line;		
			do {
				chomp ($line);
				if ($line =~ /^\s*$/) {
					last if ($blank_line);
					$blank_line = 1;
				} elsif ($line =~ /^bank/) {
					$blank_line = 0;			
					my @misc_hdr = split(/\,/, $line);

					$line = <FILE>;
					chomp ($line);
					my @misc_data = split(/\,/, $line);

					my @misc_banks = ($misc_data[0]);
					if ($misc_data[0] =~ /(\d+)\-(\d+)/) {
						@misc_banks = ($1..$2);
					}

					foreach my $b (@misc_banks) {
						my @misc_errs;
						for (my $i=1; $i < scalar(@misc_data); $i++) {
							push @misc_errs, ($misc_hdr[$i] => $misc_data[$i]);
						}
						my $bank_name = $machine_errors->{'bank'}->{$b};
						$machine_errors->{'misc'}->{$bank_name} = { @misc_errs };
					}
				} else {
					last;
				}
			} while ($line = <FILE>);
		}
	}
}

###############################################################################
# Method:    mc_prog_machine_errors
# Objective: Programs and Validates any one Machine Check Bank Errors
# Details:   See POD documentation at the end of this file.
###############################################################################
sub mc_prog_machine_errors ($$) {
  my ($self, $APISetSS, $values) = @_;
  my (@enbl_excp_engine, @reg_errs, @clear_errs, @prog_errs, @run_test);
  my ($MODEL, $BANK, $response);
  my $Test = $values->{'tcd_obj'};
  $Log->enter();
 
  $MODEL = $values->{'model'};
  $BANK = $values->{'bank'};

  if($values->{'bank'} =~ /^$/) {
    my $no_of_banks = $self->{$MODEL}->{'banks'};
    $no_of_banks =~ /(\d{2})$/;
    $no_of_banks = hex($1);
    $BANK = int rand $no_of_banks;
  }

  if ($BANK =~ /^\d/) {
     $BANK = $values->{'bank'} = $self->{'bank'}->{$BANK};
  } elsif ($BANK =~ /^\w/) {
     foreach my $v (values $self->{'bank'}) {
       if ($v =~ /$BANK/i) {
          $BANK = $values->{'bank'} = $v;
          last;
       }
     }
  }
  $Log->comment("mc_prog_machine_errors: Bank = $BANK");
  $values = $self->_mc_get_reg_values($values);

  @reg_errs = (
      'debug.eeng.addr', $self->{$MODEL}->{'addr'},
      'debug.eeng.value', $self->{$MODEL}->{'banks'},
      'debug.eeng.flag', 0,
      'debug.eeng.add', 1,
      'debug.eeng.addr', $self->{'mc'}->{$BANK}->{'status'},
      'debug.eeng.value', $values->{'status'},
      'debug.eeng.flag', 2,
      'debug.eeng.add', 1,
   );
   
   unless ($self->{'mc'}->{$BANK}->{'misc'} =~ /na/i) {
     push @reg_errs, (
      'debug.eeng.addr', $self->{'mc'}->{$BANK}->{'addr'},
      'debug.eeng.value', $values->{'addr'},
      'debug.eeng.flag', 1,
      'debug.eeng.add', 1,
      'debug.eeng.addr', $self->{'mc'}->{$BANK}->{'misc'},
      'debug.eeng.value', $values->{'misc'},
      'debug.eeng.flag', 1,
      'debug.eeng.add', 1,
      'debug.eeng.list', 1,
     );
   }

  @enbl_excp_engine = (
      $self->{$MODEL}->{'poll_cli'}, 0,
      'machdep.cecc_enabled', 0,
      'debug.eeng.mode', 1,
      'debug.eeng.test', 0,
   );

  @clear_errs = ('debug.eeng.delete', 1 );

  @prog_errs = (@clear_errs, @enbl_excp_engine, @reg_errs);
					 
  $Test->step("Programming values into $BANK registers");
  for (my $i=0; $i<(scalar @prog_errs); $i+=2) {
     $APISetSS->sysctl(
                   "name" => $prog_errs[$i], 
                   "value" => $prog_errs[$i+1],
                   "privilege-level" => 'root'
                );
  }
	
  $Test->step("Executing the test");
  $response = $APISetSS->sysctl("name" => 'debug.eeng.trigger', "value" => 1, "privilege-level" => 'root');
  my $mc_err_output = $response->get_raw_output();
  $Log->debug("\nraw output = $mc_err_output");

  $Test->step("Validating Machine Error Output");
  if ($mc_err_output =~ /correctable/i) {
      $self->_mc_validate_merr($mc_err_output, $values);
  } elsif ( !(exists $values->{'sbits'}->{'63'}) ) {
      $Log->comment("\"Valid\" field \(bit:63\) is not set, No output is Expected: $mc_err_output");
      return 1;
  } else {
      $Log->comment("INVALID output: Not sure what could be the issue: $mc_err_output \n");
      return 0;
  }

  $Test->step("Executing SRAM output of ONTAP region");
  $response = $APISetSS->ssram( "region" => 1, "region-type" => 2, 'privilege-level' => 'root', );
  my $mc_sram_output = $response->get_raw_output();

  $Test->step("Verifying SRAM output of ONTAP region");
  unless ($mc_sram_output =~ /Rec.+\s+node.+/) {
     unless ( exists ($values->{'sbits'}->{'61'}) || ($values->{'bank'} =~ /ha/i) ) {
       $Log->log("No entry in SRAM output as Expected");
     } 
   } elsif ( exists ($values->{'sbits'}->{'61'}) ) {
      print "calling sram validate function \n";
      $self->_mc_validate_sram($mc_sram_output, $mc_err_output, $values);
   } else {
      $Log->warn("There should NOT be any entry in SRAM");
      return 0;
   }

  $Log->exit();
  return 1;
}

########################################################################
#     P R I V A T E   M E T H O D S
########################################################################

###############################################################################
# Method:    _mc_get_reg_values
# Objective: Programs and Validates any one Machine Check Bank Errors
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_get_reg_values ($) {
  my ($self, $values) = @_;
  my (@status_reg_bits, $status_reg_bits_to_prog, @misc_reg_bits, $misc_reg_bits_to_prog);
  my $BANK = $values->{'bank'};
  $Log->enter();

### Scenarios ###
  my @scenarios = (63, 61, 59, 58, 57);
### END ###

  if($values->{'status'} =~ /^$|^0$/) {
    $Log->comment("\nSTATUS\n");
    @status_reg_bits = sort keys %{$self->{'status'}->{$BANK}};
    $status_reg_bits_to_prog  = $self->_mc_get_random_errors(\@status_reg_bits);
    my @sbits_prg = sort keys %{$status_reg_bits_to_prog};
    $Log->comment("Final status reg bits to prog: @sbits_prg\n");

### Adding scenarios ###
    unless (exists $status_reg_bits_to_prog->{$scenarios[0]}) {
        $status_reg_bits_to_prog->{$scenarios[0]} = '';
    }
### END ###

    foreach my $t (@sbits_prg) {
      if (exists $status_reg_bits_to_prog->{$t}) {
         $status_reg_bits_to_prog->{$t} = $self->{'status'}->{$BANK}->{$t};
      }
    }

    $values->{'sbits'} = $status_reg_bits_to_prog;
    $values->{'status'} = $self->_mc_get_hex_value($status_reg_bits_to_prog);
  } else {
    $Log->comment("Status Reg Value: $values->{'status'} ");
    $values->{'sbits'} = $self->_mc_decode_reg_value('status', $values);
  }

  if($values->{'misc'} =~ /^$|^0$/) {
    $Log->comment("\nMISC\n");
    @misc_reg_bits = sort keys %{$self->{'misc'}->{$BANK}};
    $misc_reg_bits_to_prog  = $self->_mc_get_random_errors(\@misc_reg_bits);
    my @mbits_prg = sort keys %{$misc_reg_bits_to_prog};
    $Log->comment("Final misc reg bits to prog: @mbits_prg");

    foreach my $t (@mbits_prg) {
      if (exists $misc_reg_bits_to_prog->{$t}) {
         $misc_reg_bits_to_prog->{$t} = $self->{'misc'}->{$BANK}->{$t};
      }
    }
    $values->{'mbits'} = $misc_reg_bits_to_prog;
    $values->{'misc'} = $self->_mc_get_hex_value($misc_reg_bits_to_prog);

  } else {
    $Log->comment("Misc Reg Value: $values->{'misc'} ");
    $values->{'mbits'} = $self->_mc_decode_reg_value('misc', $values);
  }

  $Log->exit();
  return $values;
}

###############################################################################
# Method:    _mc_get_random_errors
# Objective: Select Random BITs/Fields for Status and Misc Bank registers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_get_random_errors ($;$) {
    my (%combinations_hash, @sub_reg_bits);
    my $self = shift;
    my $reg_bits = shift;
    my $bits_limit = ($_[0] ? shift : 6);
    my $no_of_bits = int rand (scalar @{$reg_bits});
    $Log->enter();
    $Log->comment("$no_of_bits reg bits selected out of "  . scalar @{$reg_bits} . ". Range: @{$reg_bits}");

#    $no_of_bits = int ((scalar @{$reg_bits})/2) if ($no_of_bits < $bits_limit);
    if ($no_of_bits < $bits_limit) {
       $no_of_bits = (scalar @{$reg_bits} <= $bits_limit) ? scalar @{$reg_bits} : int ((scalar @{$reg_bits})/2);
    }
    $Log->comment("No of reg bits corrected to prog.  = $no_of_bits");

    until (scalar (keys %combinations_hash) == $no_of_bits) {
       my $index = int rand (scalar @{$reg_bits});
       next if (exists $combinations_hash{$reg_bits->[$index]});
       $combinations_hash{$reg_bits->[$index]} = '';
    }

    $Log->exit();
    return \%combinations_hash;
}

###############################################################################
# Method:    _mc_get_hex_value
# Objective: Generates Hexadecimal value for Status and Misc registers
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_get_hex_value ($) {
    my ($self, $reg_bits_to_prog) = @_;
    my @bin_arr = (0)x64;
    $Log->enter();

    foreach my $reg_bit (sort keys %{$reg_bits_to_prog}) {
        if ($reg_bit =~ /(\d+)[-:](\d+)/) {
            my @sub_arr = ($2..$1);
            if ( (scalar @sub_arr) < 6 ) {
               @bin_arr[@sub_arr] = (1)x100;
            } else {
               @bin_arr[keys %{$self->_mc_get_random_errors(\@sub_arr)}] = (1)x100;
            }
        } else {
            $bin_arr[$reg_bit] = (1);
        }
    }
    @bin_arr = reverse @bin_arr;

    my $hbin_value = join ("", @bin_arr[0..31]);
    my $lbin_value = join ("", @bin_arr[32..63]);
    my $hex_value = unpack ("H*", pack("B*", $hbin_value)) . unpack ("H*", pack("B*", $lbin_value));
    $hex_value = '0x' . $hex_value;
    $Log->comment("binary value     = $hbin_value$lbin_value");
    $Log->comment("hex value        = $hex_value");
    $Log->exit();
    return $hex_value;
}

###############################################################################
# Method:    _mc_decode_reg_value
# Objective: This function decodes hex value given by user, which helps in validating output.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_decode_reg_value($$$) {
   my ($self, $reg_type, $values) = @_;
   my $BANK = $values->{'bank'};
   my ($hhvalue, $lhvalue) = $values->{$reg_type} =~ /(?:0x)?(\w{8})(\w{8})/i;
   my @barr = split( //, unpack("B*", pack("H*", $hhvalue)) . unpack("B*", pack("H*", $lhvalue)) );
   my $reg_bits;

   @barr = reverse @barr;

   $Log->comment("Finding the bits and fields for the given \U$reg_type\E Hex value");
   for (my $i=0; $i<scalar(@barr); $i++) {
     if ($barr[$i] == "1") {
       foreach my $k (sort keys %{$self->{$reg_type}->{$BANK}} ) {
         if ($k =~ /(\d+)[-:](\d+)/) {
            my $f = 0;
            for (my $j=$2; $j<=$1; $j++) {
              if ($j == $i) {
                $reg_bits->{$k} =  $self->{$reg_type}->{$BANK}->{$k};
                $i=$1;
                $f=1;
                last;
              }
            }
          last if ($f);
         } elsif ( exists ($self->{$reg_type}->{$BANK}->{$i}) ) {
            $reg_bits->{$i} =  $self->{$reg_type}->{$BANK}->{$i};
            last;
         }
       }
     }
   }
   return $reg_bits;
}

###############################################################################
# Method:    _mc_validate_merr
# Objective: Validates Machine Check Bank Error output w.r.t input Values/Fields 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_validate_merr ($$) {
   my ($self, $merr, $values) = @_;
   my (@sreg, @mreg, @areg, @s_err_fields, @m_err_fields);
   my $Test = $values->{'tcd_obj'};
   my $BANK = $values->{'bank'};
   no warnings;
   $Log->enter();
   
   @sreg = $merr =~ /((?:Un)?correctable).+(CPU\d+)\.\s+(\w+).+STATUS\<(\w+)\>\(([\w\,\(\)]+)\)(?:MISC|ADDR)?/i;
   @mreg = $merr =~ /MISC\<(\w+)\>\(([\w\,\(\)]+)\)(?:ADDR)?/;
   @areg = $merr =~ /ADDR\<(\w+)\>\(\((\w+)\)/;
   
   @s_err_fields = split (/\,/, $sreg[4]);  
   @m_err_fields = split (/\,/, $mreg[1]);

   $Test->step("Validating Bank & Error type:");
   if ($sreg[2] =~ /$values->{'bank'}/i) {
      $Log->comment("Bank reported by ONTAP is valid: \"$sreg[2]\" ");
   } else {
      $Log->warn("ONTAP reported invalid Bank details: sreg = $sreg[2], programmed bank: " . $values->{'bank'});
   }

   if ( ( exists $values->{'sbits'}->{'61'} || exists $values->{'sbits'}->{'57'} ) && ($sreg[0] eq "Uncorrectable") ) {
      $Log->comment("Error type reported by ONTAP : \"$sreg[0]\" is valid");
   } elsif ($sreg[0] eq "Correctable") {
      $Log->comment("Error type reported by ONTAP : \"$sreg[0]\" is valid");
   } else {
      $Log->warn("ONTAP reported invalid error type: $sreg[0]");
   }

   $Test->step("Validating Hex values of Status, Misc and Addr registers:");
   if ($sreg[3] eq $values->{'status'}) {
      $Log->comment("Status register value is valid: $sreg[3]");
   } else {
      $Log->warn("Status register value is invalid: $sreg[3]");
   }

   if ( ($mreg[0] eq $values->{'misc'}) && (exists $values->{'sbits'}->{'59'}) ) {
      $Log->comment("Misc register fields are valid: $mreg[0]");
   } elsif ( !(exists $values->{'sbits'}->{'59'}) ) {
      $Log->comment("Misc register NOT preset, which is valid");
   } elsif ($self->{'mc'}->{$values->{'bank'}}->{'misc'} =~ /na/i) {
      $Log->comment("Misc register NOT reported, which is valid");
   } else {
      $Log->warn("Misc register value is invalid: $mreg[0]");
   }   
 
   if ( ($areg[0] eq $values->{'addr'}) && (exists $values->{'sbits'}->{'58'}) ) {
      $Log->comment("Addr register value is valid: $areg[0]");
   } elsif ( !(exists $values->{'sbits'}->{'58'}) ) {
      $Log->comment("Addr register NOT preset, which is valid");
   } elsif ($self->{'mc'}->{$values->{'bank'}}->{'addr'} =~ /na/i) {
      $Log->comment("Addr register NOT reported, which is valid");
   } else {
      $Log->warn("Addr register value is invalid: $areg[0]");
   }
   
   $Test->step("Validating Status Register Fields:");
   my (@s_matched, @s_unmatched);
   foreach my $s (@s_err_fields) {
     my $m=0;
     foreach my $ps (values %{$values->{'sbits'}}) {
       if ($s =~ /$ps/i) {
         $m=1;
         last;
       }
     }
     if ($m) {
       push @s_matched, $s;
     } else {
       push @s_unmatched, $s;
     }
   }
	
   if (scalar(@s_unmatched) == 0) {
      $Log->comment("All Status register fields are Valid: @s_matched");
   } else {
      $Log->warn("Few Status register fields are NOT Valid: @s_unmatched");
   }
   my @sprog_fields = values %{$values->{'sbits'}};
   $Log->comment("Programmed fields: @sprog_fields");

   $Test->step("Validating Misc Register Fields:");
   my (@m_matched, @m_unmatched);
   foreach my $s (@m_err_fields) {
     my $m=0;
     foreach my $ps (values %{$values->{'mbits'}}) {
       if ($s =~ /$ps/i) {
         $m=1;
         last;
       }
     }
     if ($m) {
       push @m_matched, $s;
     } else {
       push @m_unmatched, $s;
     }
   }

   #my $r = NATE::Result::Pass->new("everything cool");
	
   if ( (scalar(@m_matched) == 0) && (scalar(@m_unmatched) == 0) ) {
      $Log->log("Misc register fields are NOT present, which is Valid");
   } elsif (scalar(@m_unmatched) == 0) {
      $Log->log("All Misc register fields are Valid: @m_matched");
   } else {
      $Log->warn("Few Misc register fields are NOT Valid: @m_unmatched");
   }
   my @mprog_fields = values %{$values->{'mbits'}};
   $Log->comment("Programmed fields: @mprog_fields");
#   $r->message("Programmed fields: @mprog_fields");
#   $r->log;
#use NATE::Log qw(log_global);
#my $log = log_global();
#$log->pass("Actual output: \n$merr");
   $Log->comment("\nActual output: \n$merr \n");
   use warnings;
   $Log->exit();
}

###############################################################################
# Method:    _mc_validate_sram
# Objective: Validates SSRAM output of ONTAP region to make sure Machine errors are getting logged properly
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _mc_validate_sram ($$;$) {
   my ($self, $sram_out, $merr, $values) = @_;
   my (@a, $m_cpu_no, $s_cpu_no);
   @a = $sram_out =~ /Rec.+\s+node\((\d+)\)\s+core\((\d+)\)/;
   
   ($m_cpu_no) = $merr =~ /(?:Un)?correctable.+CPU(\d+)/i;
   if ($a[1] == $m_cpu_no) {
      $Log->log("Core number: $a[1] is valid w.r.t Machine error output: $m_cpu_no");
   } else {
      $Log->warn("Core number is NOT Valid: $a[1] w.r.t Machine error output: $m_cpu_no");
   }
}

1;

##############################################################################
#                                    POD here
##############################################################################

=head1 NAME

MachineErrors - Machine Errors class

=head1 SYNOPSIS

 use MachineErrors;

=head1 DESCRIPTION

A C<MachineErrors> contains data and methods common to all Machine
Check Errors. It extracts details from the machine errors CSV file 
and validates of every platform/processor.

=head1 CONSTRUCTOR

=head2 new

Returns a reference to a newly created C<MachineErrors> module.  
No input parameters are required.

=over

=item Synopsis

 $Merr_Obj = new MachineErrors;

=item Arguments

 None.

=item Returns

A reference to a new MachineError module.

=back

=head1 METHODS

=head2 mc_extract_machine_errors_csv_file

This method is used to extract all details from F<Machine Check Bank Errors CSV file> 
based up on the Model I<eg: FAS8040> and updates the blessed hash table which was created 
using constructor L<new|new>.

=over

=item Synopsis

 $Merr_Obj->mc_extract_machine_errors_csv_file($csv_file_name);

=item Arguments

 $csv_file_name

=item Returns

 Nothing.

=back

=head2 mc_prog_machine_errors

This method is used to program all the machine check bank error registers of a 
particular bank and executes test case. Also validates the machine check error 
reported by ONTAP & SRAM outputs of I<ONTAP region>. 
It takes random values for Bank number, Status and Misc registers if
no inputs provided to the script.

=over

=item Synopsis

 $result = $Merr_Obj->mc_prog_machine_errors($APISetSS, $mc_reg_values);

=item Arguments

B<$APISetSS>
 
 API System Shell Object

B<$mc_reg_values>

 Input values (hash) to the script

 $mc_reg_values = {
      'bank'    => $BANK,
      'status'  => $STATUS_VALUE,
      'addr'    => $ADDR_VALUE,
      'misc',   => $MISC_VALUE,
      'model'   => $MODEL,
      'tcd_obj' => $Test,
  };

=item Returns

Returns 0 I<FAIL> or 1 I<PASS>.

=back

=head2 _mc_get_reg_values

This method is to be used to get the Status, Misc and Address register
values if no inputs provided to the script.

Register values will be calculated by selecting random bits of Bank 
specific register value and fields.

=over

=item Synopsis

$values = $self->_mc_get_reg_values($values);

=item Arguments

$values - Input parameters

=item Returns

Returns the hex values and updates the same member

=back

=head2 _mc_get_random_errors

This method is used to select Random BITs/Fields for Status and Misc Bank registers
to help L<_mc_get_reg_values|_mc_get_reg_values> method.

=over

=item Synopsis

 $status_reg_bits_to_prog  = $self->_mc_get_random_errors(\@status_reg_bits);
 $misc_reg_bits_to_prog    = $self->_mc_get_random_errors(\@misc_reg_bits);

=item Arguments

=over

=item @status_reg_bits

 Status register all bits (0 to 63)

=item @misc_reg_bits

 Misc register all bits (0..63)

=back

=item Returns

Randomly selected bits & fields in a hash table

=back

=head2 _mc_get_hex_value

This method Generates Hexadecimal value for randomly selected bits of Status and Misc registers

=over

=item Synopsis

 $values->{'status'} = $self->_mc_get_hex_value($status_reg_bits_to_prog);
 $values->{'misc'} = $self->_mc_get_hex_value($misc_reg_bits_to_prog);

=item Arguments

 $status_reg_bits_to_prog
 $misc_reg_bits_to_prog

=item Returns

 64 bit Hexadecimal value

=back

=head2 _mc_decode_reg_value

This method decodes the hex values given by user and finds bit & field combinations of respective register (Status or Misc)

=over

=item Synposis

 $values->{'sbits'} = $self->_mc_decode_reg_value('status', $values);
 $values->{'mbits'} = $self->_mc_decode_reg_value('misc', $values);

=item Arguments

  'status', $values
  'misc', $values

=item Returns

  Hash reference with valid bit & field combinations based on Hex value given by user

=back

=head2 _mc_validate_merr

This method Validates Machine Check Bank Error output w.r.t input Values/Fields.

=over

=item Synopsis

 $self->_mc_validate_merr($mc_err_output, $values);

=item Arguments

=over

=item $mc_err_output

 Machine Check Error reported by ONTAP once after test case execution

=item $values
 
 Input values used to program registers

=back

=item Returns

Nothing.

=back

=head2 _mc_validate_sram

Validates SSRAM output of ONTAP region to make sure Machine errors are getting logged properly

=over

=item Synopsis

 $self->_mc_validate_sram($mc_sram_output, $mc_err_output, $values);

=item Arguments

=over

=item $mc_sram_output

 SRAM output returned by the command C<sudo ssram -a -g 2>

=item $mc_err_output

 Machine Check Error reported by ONTAP once after test case execution

=item $values
 
 Input values used to program registers

=back

=item Returns

Nothing.

=back

=head1 AUTHOR/MAINTAINER

=over

=item jkonjeti

=item dl-pvt-platsw@netapp.com

=back

=cut


