#!/usr/local/bin/perl
#*********************************************************************************************************************************************************
## @summary	Modules developed to support NVM feature
## 
## @description
## "nvmem.pm" is intended to have support that would verify the nvlogs, take the filer to known state, start stress and perform power cycle
## 
## @synopsis
## goto_systemshell
## setup 
## verify_nvm_log
## wdr_l1_l2
## dirty_shutdown
## start_stress
##
## @author              kakshay@netapp.com
#********************************************************************************************************************************************************** 
use NACL::APISet;
use NACL::Transit; 				# Used to get/set states of the filer
use platval::service_processor::rlm_ipv6;  	# This is used to perform dirty shutdown from SP

use constant SUCCESS	=>	1;
use constant FAILURE	=> 	0;
use constant TIMEOUT	=>	800;

use strict;


#*********************************************************************************************
#Function Name: goto_systemshell
#Parameters   : NA
#Return Type  : "SYSTEMSHELL" if Success
#               "void" if Failure
#Description  : Modules intended to take the filer to Systemshell
#*********************************************************************************************
sub goto_systemshell
{
  my $obj = NACL::Transit->new( name => $FILER );
  my $out;

  my $filer_state = $obj->get_state();
  if($filer_state !~ /SYSTEMSHELL/)
  {
    tstlog -comment, "Going from $filer_state to SYSTEMSHELL";
    $obj->change_state(to => 'SYSTEMSHELL', username => $FREEBSD_USER, password => $FREEBSD_PASS, timeout => TIMEOUT, append_output_to => \$out);
  }
  return SUCCESS if($obj->get_state() =~ /SYSTEMSHELL/);
}
#*********************************************************************************************
#Function Name: setup
#Parameters   : NA
#Return Type  : "CLI" if Success
#               "void" if Failure
#Description  : Modules intended to takes the filer to Ontap. 
#*********************************************************************************************
sub setup
{
  my $obj = NACL::Transit->new( name => $FILER );
  my $out;

  my $filer_state = $obj->get_state();
  if($filer_state =~ /FIRMWARE/ || $filer_state =~ /SYSTEMSHELL/ || $filer_state =~ /BOOT_MENU/ || $filer_state =~ /MAINT/ || $filer_state =~ /PASSWORD/)
  {
    tstlog -comment, "Going from $filer_state to ONTAP";
    $obj->change_state(to => 'CLI', username => $USER, password => $PASS, timeout => TIMEOUT, append_output_to => \$out);
  }
  return SUCCESS if($obj->get_state() =~ /CLI/);
}
#*********************************************************************************************
#Function Name: verify_nvm_log
#Parameters   : NA
#Return Type  : SUCCESS -> if the logs get replayed during boot
#               FAILURE -> If NVM failed to replay the logs
#Description  : Modules intended and verify that the logs are replayed 
#*********************************************************************************************
sub verify_nvm_log
{
  my $conn = connect("#console");
  my @prompts = ('LOADER(-\w+)?\*?>','CFE(-\w+)?\*?>'); #akshay-> What can we do about this, we need to keep it outside of the script

  my (undef,undef,$out,undef) = say($conn,"?",-prompt,@prompts,-swallow,-timeout,TIMEOUT);

  if($out =~ /\++/)
  {
    tstlog -comment,"NVM is Armed";

    my $obj = NACL::Transit->new( name => $FILER );
    $obj->change_state(to => 'CLI', username => $USER, password => $PASS, timeout => TIMEOUT, append_output_to => \$out);

    #($out =~ /Replaying WAFL log/) ? tstlog -pass,"NVM Logs got Replayed" : tstlog -fail, "NVM Logs Did Not Got Replayed";
    return ($out =~ /Replaying WAFL log/) ? SUCCESS : FAILURE;

   }
   else
   {
     #tstlog -fail, "NVM is Not Armed";
     return FAILURE;
   }
}
#*********************************************************************************************
#Function Name: start_stress
#Parameters   : NA
#Return Type  : NA
#Description  : Modules developed to make sure the load(hammer,disk scrub,filersio) is started
#*********************************************************************************************
sub start_stress
{
  my $conn = connect("#console");
  my $host = host($FILER);

  say($conn,"priv set test",-prompt,$host->{'prompt_norm'},-swallow);
  say($conn,"disk scrub start",-prompt,$host->{'prompt_norm'},-swallow);
  say($conn,"hammer -f -1 /vol/vol0/filename 1234",-prompt,$host->{'prompt_norm'},-swallow);
  say($conn,"filersio asyncio_active 00 -r 100 4k 0 7g 84000 5 /vol/vol0/file.90g -create",-prompt,$host->{'prompt_norm'},-swallow);
  sleep 10;

  my(undef,undef,$load,undef) = say($conn,"sysstat -c 1 1",-prompt,$host->{'prompt_norm'},-swallow);
  tstlog -comment,"Load is:$load";
}
#*********************************************************************************************
#Function Name: wdr_l1_l2
#Parameters   : $reset_type -> Takes "L1" or "L2"
#Return Type  : NA
#Description  : Trigges L1 or L2 watchdog 
#*********************************************************************************************
sub wdr_l1_l2
{
  my($reset_type) = @_;
  my $host = host($FILER);
  my $raw_output;
  my $Api_Set_Obj = NACL::APISet->new(
                                 hostobj   => $host,
                                 category  => "Node",
                                 interface => "CLI",
                                 set       => "Systemshell",
                                 );
                                 #shell_type => "7Mode",
  my $system_shell_prompt = $Api_Set_Obj->get_connection()->hostp()->id().'__prompt__>\s';
  if($reset_type =~ /L1/i)
  {
    $raw_output = $Api_Set_Obj->execute_raw_command(command => 'sudo sysctl debug.watchdog.ignore_nmi=0','connectrec-prompts' => [$system_shell_prompt] );
  }
  if($reset_type =~ /L2/i)
  {
    $raw_output = $Api_Set_Obj->execute_raw_command(command => 'sudo sysctl debug.watchdog.ignore_nmi=1','connectrec-prompts' => [$system_shell_prompt] );
    $raw_output = $Api_Set_Obj->execute_raw_command(command => 'sudo sysctl debug.watchdog.trigger_l2_timeout=1','connectrec-prompts' => [$system_shell_prompt] );
  }
  $raw_output = $Api_Set_Obj->execute_raw_command(command => 'sudo sysctl debug.watchdog.hang_cpu=1','connectrec-prompts' => [$system_shell_prompt] );
}
#********************************************************************************************************************
#Function Name: dirty_shutdown
#Parameters   : $sp_reset_type -> Can be "system power cycle", "system reset", "system core", "system power off/on"
#Return Type  : NA
#Description  : Performs dirty shutdown from SP
#********************************************************************************************************************
sub dirty_shutdown
{
  my($sp_reset_type,$sp_ssh_ip,$sp_user,$sp_passwd) = @_;
  rlm_find_rlm_sp_prompt($FILER);
  my $sp_conn = rlm_make_conn_to_rlm($FILER,$sp_ssh_ip,$sp_user,$sp_passwd);

  say($sp_conn,$sp_reset_type,-prompt,'Continue',-timeout,100,-swallow);
  say($sp_conn, "y",-noprompt,-timeout,50,-swallow);
  say($sp_conn,"exit",-noprompt,-timeout,50,-swallow);
}
1;
