#!/usr/local/bin/perl
# $Id: $
#
#**********************************************************************************************************************************************
##
## @description
## "nvram8.pm" helps to "program FPGA", "validating fpga", "getting fpga version", "destage and restage information", 
##                       "CECC and UECC info", "WAFL nvlog info"  and start hammer traffic
##
## @synopsis
## fpga_program($conn,$path);
## fpga_validate($conn,$image);
## fpga_version($conn,$str);
## destage_restage_done($conn);
## restage_error($conn);
## wafl_nvlog_replay($conn);
## ecc_count($conn,$field);
## cmp_files($conn,$file1,$file2);
## hammer($conn);
##
## @dependencies     test/main/lib/platval/platsw/util.pm
##
## @status      public
##
## @author      vasant.patil@netapp.com
##
## @notes       Created by Platform Validation.
##
#************************************************************************************************************************************************

use Tharn;
use strict;
use warnings;
use Switch;

use constant SUCCESS                    => 1;
use constant FAILURE                    => 0;

#*********************************************************************************************
## Function Name: fpga_program
##
## Parameters   : $conn           -> Filer connection
##              : $path           -> PATH to the FPGA image to be programmed
##
## Return Type  : SUCCESS         -> 1, if fpga image is valid and programmed
##                FAILURE         -> 0, if fpga image is invalid and not programmed
##
## Description  : Verify FPGA image and program the FPGA 
##
#*********************************************************************************************

sub fpga_program
{
 my($conn,$path) = @_;
 my $host = host($FILER);
 my $filer_state = get_filer_state($conn);
 $filer_state = goto_systemshell($conn) if($filer_state !~ "IN_SYSTEM_SHELL");

 my $str = "ok";
 tstlog -comment, "Verifying FPGA image";
 my ($match,$before,$out,$prompt) = say($conn,"sudo mriflashtool verify $path",-prompt,$host->{'prompt_norm'},-timeout,360,-swallow);
  if($out =~ m/$str.*/)
  {
  tstlog -comment, "Programming the FPGA image";
  say($conn,"sudo mriflashtool program $path",-prompt,$host->{'prompt_norm'},-timeout,360,-swallow);
  return SUCCESS;
  }else
  {
  tstlog -comment, "FPGA image is not valid";
  return FAILURE;
  }

} #End of function

#*********************************************************************************************
## Function Name: fpga_validate
##
## Parameters   : $conn           -> Filer connection
##              : $image          -> 0 or 1 :FPGA image to be validated
##
## Return Type  : SUCCESS         -> 1, if fpga image is valid 
##                FAILURE         -> 0, if fpga image is invalid
##
## Description  : Validating the FPGA image(0 or 1) in NVRAM8 
##
#*********************************************************************************************

sub fpga_validate
{
 my($conn,$image) = @_;
 my $host = host($FILER);

 my $filer_state = get_filer_state($conn);
 $filer_state = goto_systemshell($conn) if($filer_state !~ "IN_SYSTEM_SHELL");

 my $str = "ERROR";
 tstlog -comment, " Validating the FPGA image";
 my ($match,$before,$out,$prompt) = say($conn,"sudo mriflashtool validate $image",-prompt,$host->{'prompt_norm'},-timeout,360,-swallow);

 if($out =~ m/$str.*/)
 {
 tstlog -comment, "FPGA image $image is not valid";
 return FAILURE;
 }else
 {
 tstlog -comment, "FPGA image $image is valid";
 return SUCCESS;
 }

} #End of function

#*********************************************************************************************
## Function Name: fpga_version
##
## Parameters   : $conn           -> Filer connection
##              : $str            -> string to be matched for version info
##
## Return Type  : $version        -> FPGA version
##                FAILURE         -> 0, if fails to match the version
##
## Description  : Returns the requested (loaded,selected etc)FPGA version
##
#*********************************************************************************************

sub fpga_version
{
my($conn,$str) = @_;
my $host = host($FILER);

my $filer_state = get_filer_state($conn);
$filer_state = goto_systemshell($conn) if($filer_state !~ "IN_SYSTEM_SHELL");

tstlog -comment, " checking the current FPGA version";

my ($match,$before,$out,$prompt) = say($conn,"sudo mriflashtool info",-prompt,$host->{'prompt_norm'},-timeout,360,-swallow);

if($out =~ m/$str.*/)
{
my (undef,$version) = split(/:/,$&);
return $version ;
}
else
{
tstlog -comment, "mriflashtool fails to return required information";
return FAILURE;
}

}

#*********************************************************************************************
## Function Name: destage_restage_done
##
## Parameters   : $conn           -> Filer connection
##
## Return Type  : $restage        -> 1, if destage and restage happed
##                                -> 0, if destage and restage happed did not happen
##
## Description  : Verify wheather destage and restage happened or not
##
#*********************************************************************************************

sub destage_restage_done
{
my($conn) = @_;
my $host = host($FILER);

my $filer_state = get_filer_state($conn);
$filer_state = goto_systemshell($conn) if($filer_state !~ "IN_SYSTEM_SHELL");
tstlog -comment, "Verifying Destage/restage \n";
my ($match,$before,$out,$prompt) = say($conn,"sysctl dev.nvram8.0.restage_done",-prompt,$host->{'prompt_norm'},-timeout,80,-swallow);

my $str = "restage_done:";
	if($out =~ m/$str.*/)
	{
	my (undef,$restage) = split(/:/,$&);
	return $restage;
	}

}#End of destage_restage_done function

#*********************************************************************************************
## Function Name: restage_error
##
## Parameters   : $conn           -> Filer connection
##
## Return Type  : $restage_fail   -> 1, if restage fails
##                                -> 0, if restage complete successfully
##
## Description  : Verify wheather restage completed successfully or not
##
#*********************************************************************************************

sub restage_error
{
my($conn) = @_;
my $host = host($FILER);

my $filer_state = get_filer_state($conn);
$filer_state = goto_systemshell($conn) if($filer_state !~ "IN_SYSTEM_SHELL");
tstlog -comment, "Verifying any failure during restage \n";
my ($match,$before,$out,$prompt) = say($conn,"sysctl dev.nvram8.0.restage_fail",-prompt,$host->{'prompt_norm'},-timeout,80,-swallow);

my $str = "restage_fail:";
	if($out =~ m/$str.*/)
	{
	my (undef,$restage_fail) = split(/:/,$&);
	return $restage_fail;
	}

}#End of restage_error function

#*********************************************************************************************
## Function Name: wafl_nvlog_replay
##
## Parameters   : $conn           -> Filer connection
##
## Return Type  : SUCCESS         -> 1, if WAFL nvlog reply happens and complete successfully
##                FAILURE         -> 0, if WAFL nvlog reply did not happen
##
## Description  : Verify wheather WAFL nvlog replay happened or not
##
#*********************************************************************************************

sub wafl_nvlog_replay
{
my($conn) = @_;
my $host = host($FILER);

## @step Verifying that WAFL nvlog Replay happened
tstlog -comment, "Verify the WAFL nvlog reply log messages\n";
my ($match,$before,$after,$prompt) = say($conn,"boot_ontap",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);

my $str = "Replaying WAFL log";
my $str1 = "raid.stripe.replay.summary:info";

if($after =~ m/$str1.+/)
{
my $tmp = $& ;
print "\n output: $tmp \n" ;
my (undef,undef,$num) = split(/ /, $&);
print " NUM = $num \n ";

if($num > 0)
{
tstlog -comment, "\n $num RAID Stripe Replay happened\n";
}else
{
tstlog -comment, "\n NO RAID stripe Replay happened \n";
}
}

if($after =~ m/$str.*/)
{
tstlog -comment, "WAFL nvlog reply happened" ;
return SUCCESS;
}
else
{
tstlog -comment, "WAFL nvlog reply did not happen" ;
return FAILURE;
}

}#End of wafl_nvlog_replay function

#*********************************************************************************************
## Function Name: ecc_count
## Parameters   : $conn           -> Filer connection
##                $field          -> cecc or uecc count parameter 
##
## Return Type  : $count          -> cecc or uecc number of errors
##
## Description  : This function returns the number of CECC or UECC errors
##
#*********************************************************************************************

sub ecc_count
{
 my($conn,$field) = @_;
 my $host = host($FILER);

 my ($match,$before,$after,$prompt) = say($conn,"sudo sysctl dev.mri_ddr.0",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);
 if($after =~ m/$field.*/)
 {
 my (undef,$count) = split(/:/, $&);
 return $count ;
 }else
 {
 tstlog -comment, "Fail to get the parameter details \n"; 
 return -1 ;
 }

}#End of ecc_count function

#*********************************************************************************************
## Function Name: cmp_files
##
## Parameters   : $conn           -> Filer connection
##                $file1          
##                $file2        
##
## Return Type  : SUCCESS         -> 1, if both the files are same
##                FAILURE         -> 0, if both files differ
##
## Description  : Compare the files and return the result
##
#*********************************************************************************************

sub cmp_files
{
 my($conn,$file1,$file2) = @_;
 my $host = host($FILER);
 my $str = "differ";
 my ($match,$before,$after,$prompt) = say($conn,"diff $file1 $file2",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);
 if($after =~ m/$str.*/)
 {
 tstlog -comment, "both the files are same\n";
 return SUCCESS;
 }
 else
 {
 tstlog -comment, "Both files differ\n";
 return FAILURE;
 }

}# End of cmp_files function

#*********************************************************************************************
## Function Name: hammer
##
## Parameters   : $conn           -> Filer connection
##
## Return Type  : NA
##
## Description  : Start the hammer traffic if hammer is not running already
##
#*********************************************************************************************
sub hammer
{
 my($conn) = @_;
 my $host = host($FILER);
 my $str = "blacksmith";
 my $vol = "/vol/vol0/ham1"; 

 say($conn,"priv set test",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);
 my ($match,$before,$after,$prompt) = say($conn,"hammer status",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);

 if($after =~ m/blacksmith.*/)
 {
 tstlog -comment, "Hammer traffic is aleady running";
 }else
 {
 tstlog -comment, "running hammer traffic";
 our @mat = ('Hammer?');
 say($conn,"hammer -1 $vol 100000",-prompt,@mat,-timeout,300,-swallow);
 say($conn, "y",-prompt,$host->{'prompt_norm'},-timeout,300,-swallow);
 sleep 10;
 }

}#End of hammer function

1;

