# -*- Perl -*-
# $Id:
##
##	Copyright (c) 2007 Network Appliance, Inc.
##	All Rights reserved
##
## @summary
## BTS_Config.pm is a common module for all the BIOS tests.
##
## @description
## This module provides common functions required for BIOS tests.
## Data structures used:
## 1. Version table (Ver_Table)
##  - This table contains version information when 'version' or 'version -b'
##    command issued from the ONTAP prompt.
##    Ex: It can contain the primary, backup ONTAP version present in CF.
## Functions:
## 1. Halt
##  - This function halts the system when the control is at ONTAP prompt.
## 2. Pre_Cond_Halt
##  - Checks for the prompt and if it is not halted then halts the system.
## 3. Boot_Diags
##  - Boots the diags firmware.
## 4. Boot_Backup
##  - Boots the backup ONTAP located in CF.
## 5. Ver_Ontap
##  - Issues the 'version' command on ONTAP and stores the result in version
##    table.
## 6. Ver_B_Ontap
##  - Issues the 'version -b' command on ONTAP and stores the result in version
##    table.
## 7. Env_Get
##  - Gets the environmental variable.
## 8. ChopData
##  - Removes the extra characters at the end of line.
## 9. Boot_Drive
##	- To identify the boot drive of the filer
## 10. Set_Priv
##	-  To set privilege level on the filer
## 11. Boot
##	-  To execute the command which involves boot cycle of the system
## 12. Pre_Boot
##	-  To boot the filer to ONTAP
## 13. Bios_Version
##	-  To extract version of BIOS running on the system.
## 14. Update_Flash
##	-  Update firmware image from CF to SF
## 15. Ifconfig
##	-  To configure the ethernet port
## 16. Ping
##	-  Ping to the specified IP
## 17. Set_Env
##	-  Set bios environment variables
## 18. Get_Sysmodel
##  -  Get system model value
## 19. Env_Set_Variable
##  -  To set the environment variable at the loader prompt.
## 20. Env_Set_Default_Variable
##  -  Set environment variables to default values and boot ONTAP
## 21. Filer_Prompt
##  -  Get filer prompt
## 22. Booted_From
##  -  Verify the source from which ONTAP booted
## 23. Get_Version_Info
##  -  Get verion info of GX kernel
## 24. Get_Gx_Mode
##  -  Identify whether GX kernel running 7mode or 10mode
## 25. Version_Type
##  -  Identify whether build is debug or optimised
## 26. Loader_Version
##	-  To extract version of Loader running on the system.
## 27. Install_Ontap
##	-  Install ONTAP build
## 28. Update_Flash_1
##	-  Execute update flash and reset bios
## 29. Extract_Bios_Ver_While_Update_Flash
##	-  Extract BIOS version from the strings displayed while running update_flash command
## 30. Extract_Loader_Ver_While_Update_Flash
##	-  Extract LOADER version from the strings displayed while running update_flash command
## 31. Cleanup_Prompt
##	-  Function to enter login name and password if filer is not found at ONTAP and Loader prompt
## 32. Flash_Image
##	-  Function to flash BIOS image using flash command at the Loader prompt
## 33. System_Reset
##	-  To reset the system BIOS through RLM
## 34. Get_Bios_After_Download
##	-  Function to retrieve BIOS version string from "version -b" output
## 35. Search_File
##  -  Search for the given file in the specified directory
## 36. File_Rename
##  -  Rename a specified file name to a desired file name.
## 37. File_Append
##	-  To add a line entry to the specified file
## 38. Get_Vol_Space_Info
##	-  To get Total, Used and Free spaces in the specified volume and its snapshot area.
## 39. Fill_Volume
##	-  To fill the given volume with specified size of file
## 40. Free_Volume
##	-  Delete the files in given volume, to make free space
## 41. Delete_File
##	-  To delete a file from filer's root directory
## 42. Boot_Backup_10mode
##  -  Boot the OS from backup CF in 10 mode
## 43. Non_Active_Image_10mode
##  -  To find the non active image in 10 Mode
## 44. Install_Ontap_10mode
##  -  To update image in 10 Mode
## 45. Get_Sysmodel_10mode
##  -  To find the system model in 10 Mode
## 46. Cleanup_Prompt_10mode
##  -  Function to enter login name and password if filer is not found at
##     ONTAP and Loader prompt
## 47. Env_Set_Default_Variable_10mode
##  -  Set Env var to default values
## 48. Halt_10mode
##  -  To bring down the 10-mode filer to LOADER prompt
## 49. Pre_Cond_Halt_10mode
##  -  PreCondition function checks for the prompt, if the prompt
##     is ONTAP then it halts the system to LOADER
## 50. Reboot_10mode
##  -  Rebooting 10-mode filer
## 51. Check_Filer_Mode
##	-  To get if the filer is a 7mode or 10mode filer
## 52. ONTAP_Login
##	-  Verify the Filer Prompt and Login to the FILER.
## 53. Add_Prompt_10mode
##  -  Add 10mode prompt to filer object
## 54. Dblade_10mode
##  -  To enter and exit from dblade engine from 10mode shell
## 55. Check_7G_Mode
##  -  Confirms if filer is running 7G release
## 56. save_filer_logs
##  -  To save and start fresh messages and ems logs
## 57. mount_filer_vol
##  -  NFS Mount the volume specified on client
## 58. inspect_logs
##  -  To inspect /etc/messages log for errors
## 59. cleanup_filer_save_logs
##  -  To save logs during cleanup
## 60. Filer_Mode_Check
##	-
## @status private
##
## @author iferoz@netapp.com,dl-platform-qa@netapp.com,sandeepk@netapp.com
##
###############################################################################

package platval::service_processor::BTS_Config;

use strict;
use warnings;
use Shell;
use Tharn;
use filer;
use Filer::Version qw(get_wafl_version);
use Filer::Version qw(:DEFAULT :releases);
use LogReader;
use Params::Validate qw(validate OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use Server;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our (%Ver_Table, $tfc, $fc, $rlm_flag);

our @EXPORT = qw(
	%Ver_Table
	Halt
	Pre_Cond_Halt
	Boot_Diags
	Boot_Ontap
	Boot_Backup
	Ver_Ontap
	Ver_B_Ontap
	Env_Set
	Env_Get
	ChopData
	Boot_Drive
	Set_Priv
	Boot
	Pre_Boot
	Bios_Version
	Update_Flash
	Ifconfig
	Ping
	Set_Env
	Get_Sysmodel
	Env_Set_Variable
	Env_Set_Default_Variable
	Filer_Prompt
	Booted_From
	Get_Version_Info
	Get_Gx_Mode
	Version_Type
	Loader_Version
	Install_Ontap
	Update_Flash_1
	Extract_Bios_Ver_While_Update_Flash
	Extract_Loader_Ver_While_Update_Flash
	Cleanup_Prompt
	Flash_Image
	System_Reset
	Get_Bios_After_Download
	Search_File
	File_Rename
	File_Append
	Get_Vol_Space_Info
	Fill_Volume
	Free_Volume
	Delete_File
	Boot_Backup_10mode
	Non_Active_Image_10mode
	Install_Ontap_10mode
	Get_Sysmodel_10mode
	Cleanup_Prompt_10mode
	Env_Set_Default_Variable_10mode
	Halt_10mode
	Pre_Cond_Halt_10mode
	Reboot_10mode
	Check_Filer_Mode
	ONTAP_Login
	Add_Prompt_10mode
	Dblade_10mode
	Check_7G_Mode
	save_filer_logs
	mount_filer_vol
	inspect_logs
	cleanup_filer_save_logs
	Filer_Mode_Check
	Start_GCOV
	Save_GCOV_Logs
);

our $VERSION = '0.01';
my $gLOADER = "LOADER.*>";
my $gCFE = "CFE.*>";
my $BMC_PROMPT    = "^bmc.*->";

##############################     FUNCTIONS     ##############################
#--------------------------------------------------------------------------
# FUNCTION:	To bring down the filer to LOADER prompt.
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Halt
{
 	say($_[0],"halt -t 0",-prompt,"$gLOADER","$gCFE",-swallow,-timeout,300);
 	snooze(15);
}

#--------------------------------------------------------------------------
# FUNCTION: PreCondition function checks for the prompt, if the prompt is
# ontap then it halts the system to LOADER.
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Pre_Cond_Halt
{
	our (undef, undef, undef, $Prompt)= say($_[0],"version","-prompt","$gLOADER","$gCFE",-swallow,-timeout,300);
	if($Prompt !~ /$gLOADER|$gCFE/) {
 		Halt($_[0]);
	}
}

#--------------------------------------------------------------------------
# FUNCTION: Boot the DIAGS firmware.
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Boot_Diags
{
	say($_[0],"boot_diags","-prompt","Enter Diag, Command or Option:",-timeout,300);
	snooze(5);
}

#--------------------------------------------------------------------------
# FUNCTIN: Boot the OS from backup CF.
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Boot_Backup
{
	say($_[0],"boot_backup","-prompt","Selection (1-5)?",-timeout,300);
	say($_[0],"1",-swallow,-timeout,300);
	snooze(10);
	say($_[0],"\n",-timeout,300);
	snooze(5);
}

#--------------------------------------------------------------------------
# FUNCTION: Issue the command 'version' and place the value in hash table.
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Ver_Ontap
{
	snooze(5);
	say($_[0],"\n",-timeout,300);
	our (undef, undef, $output, undef)= say($_[0],"version","-wait","Release ",-swallow,-timeout,300);
	our @local_vars=split(/:/,$output);
	$Ver_Table{VERSION_ONTAP} = $local_vars[0];
}

#--------------------------------------------------------------------------
# FUNCTION: Issue the command 'version -b' and place the values of primary
# ONTAP version, secodary ONTAP version, DIAGS version, BIOS version in hash
# table (Ver_Table).
# INPUT   : $_[0]=Filer Connection Handle.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Ver_B_Ontap
{
	snooze(5);
	my $version_ref;
	my $VersionStr;
	my $Prompt;
	my @ver_string;

	say($_[0],"echo ONTAP prompt",-swallow);
	$version_ref = Get_Version_Info ($_[0]);
	if ($$version_ref{buildtype} eq "debug") {
		say ($_[0], "priv set diag;dbg to off", -swallow, -timeout, 60);
		(undef, undef, $VersionStr, $Prompt)= say($_[0],"version -b",-swallow,-timeout,300);
		say ($_[0], "dbg to console;priv set", -swallow, -timeout, 60);
	}
	else {
		(undef, undef, $VersionStr, $Prompt)= say($_[0],"version -b",-swallow,-timeout,300);
	}
	@ver_string = split(/\n/, $VersionStr);

	for(our $i = 0; $i <= $#ver_string; $i++)
	{
		if($ver_string[$i] =~ /1:\/x86_.*\/kernel\/primary.krn: .. /)
		{
			$Ver_Table{VERSION_ONTAP_PRI} = $';
			ChopData($Ver_Table{VERSION_ONTAP_PRI});
 		}
 		elsif($ver_string[$i] =~ /backup.*: .. /)
 		{
 			$Ver_Table{VERSION_ONTAP_BACKUP} = $';
			ChopData($Ver_Table{VERSION_ONTAP_BACKUP});
 		}
 		elsif($ver_string[$i] =~ /diag.*:  /)
 		{
 			$Ver_Table{VERSION_DIAG_CF} = $';
			ChopData($Ver_Table{VERSION_DIAG_CF});
	 	}
 		elsif($ver_string[$i] =~ /excelsio.*: Firmware /)
 		{
 			$Ver_Table{VERSION_BIOS_EXCELSIOR} = $';
			ChopData($Ver_Table{VERSION_BIOS_EXCELSIOR});
 		}
 		elsif($ver_string[$i] =~ /DrWho.*: Firmware /)
 		{
 			$Ver_Table{VERSION_BIOS_DRWHO} = $';
			ChopData($Ver_Table{VERSION_BIOS_DRWHO});
 		}
 		elsif($ver_string[$i] =~ /SB_XV\/firmware.* Firmware /)
		{
			$Ver_Table{VERSION_BIOS_SPECTRE} = $';
			ChopData($Ver_Table{VERSION_BIOS_SPECTRE});
		}
		elsif($ver_string[$i] =~ /SB_XVI\/firmware.* Firmware /)
		{
			$Ver_Table{VERSION_BIOS_CARNEGIE} = $';
			ChopData($Ver_Table{VERSION_BIOS_CARNEGIE});
		}
		elsif($ver_string[$i] =~ /SB_XIV\/firmware.* Firmware /)
		{
			$Ver_Table{VERSION_BIOS_SHUAR} = $';
			ChopData($Ver_Table{VERSION_BIOS_SHUAR});
		}
		elsif($ver_string[$i] =~ /SB_XVII\/firmware.* Firmware /)
		{
			$Ver_Table{VERSION_BIOS_SHUAR_XL} = $';
			ChopData($Ver_Table{VERSION_BIOS_SHUAR_XL});
		}

	}
}

#--------------------------------------------------------------------------
# FUNCTION: To clear the trailing characters like \n and \r.
# INPUT   : $_[0]= Variable name.
# RETURN  : Chopped value of the variable.
#--------------------------------------------------------------------------
sub ChopData
{
	chomp($_[0]);
	chop($_[0]);
	chop($_[0]);
}

#--------------------------------------------------------------------------
# FUNCTION: To print the environment variable at the loader prompt.
# INPUT   : $_[0]=Filer Connection Handle, $_[1]=LOADER environment variable.
# RETURN  : NONE
#--------------------------------------------------------------------------
sub Env_Get
{
 	my (undef, undef, $response, undef) = say ($_[0], "printenv $_[1]", '-swallow',-prompt,"$gLOADER","$gCFE",-timeout,300);
 	return $response;
}

#################### New Entries ###################
#-------------------------------------------------------------------------------------------
# FUNCTION: To identify the boot drive of the filer
# INPUT   : $_[0]=filer connection handle
# RETURN  : Boot Drive
#------------------------------------------------------------------------------------------
sub Boot_Drive
{
	my (undef, undef, $drive, undef)=say($_[0],"bootfs help dir","-wait","boot drive is",-swallow,-timeout,300);
	ChopData($drive);
	return $drive;
}

#-------------------------------------------------------------------------------------------
# FUNCTION: To set privilege level on the filer
# INPUT   : $_[0]=Filer Connection Handle, $_[1]=Privilege
# RETURN  : 0/1
#------------------------------------------------------------------------------------------
sub Set_Priv
{
	my (undef, undef, $set_priv, undef) = say($_[0],"priv set $_[1]",-swallow,-prompt,"$gLOADER","$gCFE",-timeout,300);
	my (undef, undef, $priv_set, undef) = say($_[0],"priv",-swallow,-prompt,"$gLOADER","$gCFE",-timeout,300);
    if($set_priv =~ /unrecognized privilege/) {
		logcomment("$_[1] privilege not available with this build");
		return 0;
	} elsif(($priv_set =~ /$_[1]/))	{
		logcomment("$_[1] privilege is set succesfully");
		return 1;
	}
}



#---------------------------------------------------------------------
# FUNCTION: To execute the command which involves boot cycle of the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Login, $_[2]= Password,
#			$_[3]= Command, $_[4]= Filer
# RETURN: 	NONE
#---------------------------------------------------------------------
sub Boot
{
	 my (undef, undef, undef, $Prompt)= say($_[0],$_[3], "-prompt", "Password:", "login:","$gLOADER","$gCFE",-swallow,-timeout,900);
         sleep 300;
	 if($Prompt !~ /$_[4]/ && $Prompt !~ /Password:/ && $Prompt !~ /$gLOADER|$gCFE/) {
	     if(($_[1]) && (!$_[2])) {
			 	 say($_[0], $_[1], "-prompt", "Password:", -swallow, -timeout, 500);
			 	 say($_[0], "\r", -swallow, -timeout, 300);
		 } elsif($_[1] && $_[2]) {
			 say($_[0],$_[1],"-prompt","Password:",-swallow,-timeout,500);
			 (undef, undef, undef, $Prompt)= say($_[0],$_[2],"-prompt",-swallow,-timeout,500);
	 	 }
	} elsif($Prompt =~ /Password:/) {
		(undef, undef, undef, $Prompt)= say($_[0],$_[2],"-prompt",-swallow,-timeout,500);
	}
	snooze(5);
}


#---------------------------------------------------------------------
# FUNCTION: To boot the filer to ONTAP
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Login, $_[2]= Password,
#			$_[3]= Command, $_[4]= Filer
# RETURN: 	NONE
#---------------------------------------------------------------------
sub Pre_Boot
{
 	my (undef, undef, undef, $Prompt)= say($_[0],"version","-prompt","$gLOADER","$gCFE",-timeout,300);
 	if($Prompt =~ /$gLOADER|$gCFE/) {
  		Boot($_[0], $_[1], $_[2], $_[3], $_[4]);
 	}
}

#-------------------------------------------------------------------------------------------
# FUNCTION: To extract version of BIOS running on the system.
# INPUT   : $_[0]=filer connection handle
# RETURN  : Version number
#-------------------------------------------------------------------------------------------
sub Bios_Version
{
	my $bios_version_sf_pri;
	my (undef, undef, $out, $Prompt)= say($_[0],"version","-prompt", $gLOADER,$gCFE,-swallow,-timeout,300);
	if($out =~ /(.*BIOS_VERSION.*)/) {
		$bios_version_sf_pri = (split(" ",$1))[-1];
		logcomment("BIOS version in SF : $bios_version_sf_pri");
	}
	return $bios_version_sf_pri;
}

#-------------------------------------------------------------------------------------------
# FUNCTION: Update firmware image from CF to SF
# INPUT   : $_[0]=filer connection handle
# RETURN:	Status
#-------------------------------------------------------------------------------------------
sub Update_Flash
{
	my (undef, undef, $flash_result, undef) = say($_[0], "update_flash -noversion", "-prompt", "$gLOADER","$gCFE",-swallow,-timeout,300);
	if (($flash_result =~ /Saving Primary/) && ($flash_result =~ /Updating Primary/)) {
		logcomment ("Bios image update is  Successful");
		return 1;
	} else {
		logcomment ("Bios image update is Not Successful");
		return 0;
	}
}

#-------------------------------------------------------------------------------------------
# FUNCTION: To configure the ethernet port
# INPUT: 	$_[0]=Filer connection Handle, $_[1]=Ethernet port, $_[2]=Filer_IP,
#			$_[3]=Net_Mask, $_[4]=Filer_Gateway_IP
# RETURN:	Status
#-------------------------------------------------------------------------------------------
sub Ifconfig
{
	my $command = "ifconfig $_[1] -addr=$_[2] -mask=$_[3] -gw=$_[4]";
	my (undef, undef, $result, undef) = say ($_[0] , $command , -swallow , -prompt , "$gLOADER","$gCFE",-timeout,300);

	if($result !~ /command status/i) {
		logcomment ("Network Port $_[1] Is Configured Successfully");
		return 1;
	} else {
		logcomment ("Network Port $_[1] Is Not Configured");
		return 0;
	}
}

#-------------------------------------------------------------------------------------------
# FUNCTION: Ping to the specified IP
# INPUT: 	$[1]=Filer connection Handle, $_[1]=Destination IP address
# RETURN:	Status
#-------------------------------------------------------------------------------------------
sub Ping
{
	my (undef, undef, $ping, undef) = say ($_[0] , "ping $_[1]" , -swallow , -prompt , "$gLOADER","$gCFE",-timeout,300);
	if ($ping =~ /alive/) {
		logcomment ("Ping to $_[1] is Successful");
		return 1;
	} else {
		logcomment ("Ping to $_[1] is NOT Successful");
		return 0;
	}
}

#---------------------------------------------------------------------
# FUNCTION: Set bios environment variables
# INPUT: 	$_[0]=Filer connection Handle, $_[1]=LOADER environment variable,
#			$_[2]=value to be set
# RETURN: 	NONE
#---------------------------------------------------------------------
sub Set_Env
{
	say($_[0],"setenv $_[1] $_[2]",-prompt,"$gLOADER","$gCFE",-swallow,-timeout,300);
}

#---------------------------------------------------------------------
# FUNCTION: Get system model value
# INPUT: 	$_[0]=Filer connection Handle
# RETURN:   sys_model
#---------------------------------------------------------------------
sub Get_Sysmodel
{
	Pre_Cond_Halt($_[0]);
    our $sys_model = Env_Get($_[0],"SYS_MODEL");
    logcomment ("System Model : $sys_model");

    $sys_model =~ /SYS_MODEL.*(FAS.*|IBM.*)/;
    $sys_model = $1;
    chop($sys_model);
    return $sys_model;
}


#---------------------------------------------------------------------
# FUNCTION: Set Env var to default values
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Login, $_[2]= Password,
#			$_[3]= Command, $_[4]= Filer
# RETURN:   sys_model
#---------------------------------------------------------------------
sub Env_Set_Default_Variable
{
	say($_[0],"set-defaults",-prompt,"$gLOADER","$gCFE",-swallow,-timeout,300);
	Boot($_[0], $_[1], $_[2], $_[3], $_[4]);
	Halt($_[0]);
}

#---------------------------------------------------------------------
# FUNCTION: To check the filer prompt
# INPUT: 	$_[0] = Filer connection handle
# RETURN: 	Current Filer Prompt
#---------------------------------------------------------------------
sub Filer_Prompt
{
    my $diags_prompt = "Enter Diag, Command or Option:";
	my (undef, undef, undef, $Prompt)= say($_[0],"\r",-swallow,"-prompt","$gLOADER","$gCFE","$BMC_PROMPT","Password:","login:", $diags_prompt,-timeout,300);
	return $Prompt;
}

#-------------------------------------------------------------------------------------------
# FUNCTION: check bootarg for variable BOOTED_FROM
# INPUT: 	$_[0] = Filer Connection Handle,
#			$_[1] = name of boot source e.g. PRIMARY or NETWORK
# RETURN: 	on Success return 1 else 0
#-------------------------------------------------------------------------------------------
sub Booted_From
{
	my $response;
	snooze(10);
	if (!(Set_Priv ($_[0], "test")))
	{
		return 0;
	}
	my $command = "bootargs print BOOTED_FROM";
	(undef, undef, $response, undef)= say($_[0],$command,-swallow,-timeout,300);
	if($response =~ /$_[1]/)
	{
		logcomment("Filer is booted from $_[1]");
		return 1;
	}
	else
	{
		logcomment("Filer is NOT booted from $_[1]");
		return 0;
	}
}

#---------------------------------------------------------------------
# FUNCTION: To get Build info about version, type, running mode
# INPUT: 	$_[0]=Filer connection handle.
# RETURN:   Reference to hash and keys are {version} {buildtype} {mode} {versionindex}
#---------------------------------------------------------------------
sub Get_Version_Info
{
	my %buildinfo = (
		version => ,
		buildtype => ,
		mode => ,
		versionindex => ,
	);

	# Get version running on filer
	$buildinfo{version} = filer_version_get($_[0]);

	# Get type of version, whether optimised or debug
	$buildinfo{buildtype} = Version_Type ($_[0]);

	# Get in if loop if it is BR release
	if ($buildinfo{version} =~ /main|8.0|Rboiler|rr/) {
		my $mode;
		$buildinfo{versionindex} = "8";
		logcomment ("GX release");
		# Get GX mode, 7mode or 10mode
		$mode = Get_Gx_Mode ($_[0]);
		if ($mode eq "7mode") {
	        $buildinfo{mode} = "7mode";
	    }
	    elsif ($mode eq "10mode"){
	    	$buildinfo{mode} = "10mode";
    	}
	}
	# Get in else loop if it is 7G release
	# <other releses can be added as elsif e.g rollingrock>
	else {
		logcomment ("7 G release");
		# Set mode to 7G
		$buildinfo{mode} ="7G";
		$buildinfo{versionindex} = "7";
	}
	# return reference to hash
	return (\%buildinfo);
}

#---------------------------------------------------------------------
# FUNCTION: To check if the filer is running 7Mode or 10Mode.
# INPUT: 	$_[0]=Filer connection handle.
# RETURN:   "7mode" = If filer is running 7Mode,
#			"10mode" = If filer is running 10Mode.
#---------------------------------------------------------------------
sub Get_Gx_Mode
{

	my $mode_ret;
	my (undef, undef, $out, undef) = say ($_[0], "?", -swallow);
	if ($out =~ /Go up one directory/i) {
		$mode_ret = "10mode";
	}
	else {
		$mode_ret = "7mode";
	}
    return $mode_ret;
}

#---------------------------------------------------------------------
# FUNCTION: To check the type of build that the filer is running on.
# INPUT: 	$_[0]=Filer connection handle.
# RETURN:   "debug" = If filer is running a debug build,
#			"optimised" = If filer is running a optimised build.
#			"undefined" = If filer is running a undefined build,
#---------------------------------------------------------------------
sub Version_Type
{
	my $version  = filer_version_get($_[0]);
	my $versiontype;
	if ($version =~ /\.*d\>/) {
		$versiontype = "debug";
	}
	elsif ($version =~ /\.*O\>/) {
		$versiontype = "optimised";
	}
	else {
		$versiontype = "undefined";
	}
	return $versiontype;
}

#-------------------------------------------------------------------------------------------
# FUNCTION: To extract version of Loader running on the system.
# INPUT   : $_[0]=filer connection handle
# RETURN  : Version number
#-------------------------------------------------------------------------------------------
sub Loader_Version
{
        my $loader_version_sf_pri;
        my (undef, undef, $out, $Prompt)= say($_[0],"version","-prompt", $gLOADER,-swallow,-timeout,300);
        if($out =~ /(.*LOADER_VERSION.*)/)
        {
                $loader_version_sf_pri = (split(" ",$1))[-1];
                logcomment("Loader version in SF : $loader_version_sf_pri");
        }
        return $loader_version_sf_pri;
}

#---------------------------------------------------------------------
# FUNCTION:     Install ONTAP build
# INPUT:        $_[0] = 'install or update' depending upon whether
#				'software install/software update' command
#               $_[1] = '-r' or null option,
#				$_[2] = "completed" word or "Please type",
#				$_[3] =  URL of ONTAP build,
#               $_[4] = Filer connection handle
# RETURN:       0 = Fail, 1 = Pass
#---------------------------------------------------------------------
sub Install_Ontap
{
    my $version_install;
    say($_[4],"\n");
    # Install ONTAP build using either "software update" or "software install" command
    (undef,undef,$version_install,undef)=say($_[4],"software $_[0] $_[3] -f $_[1]",-swallow,-timeout,600);

    # Check for the required matchstring that appeared during the install process,if does'nt appaear return 0
    if($version_install!~/$_[2]/) {
        logcomment("software $_[0] test failed");
        return 0;
    }
    # Download the image onto CF only when installing using "software install" command
    if($_[0] eq "install") {
        say($_[4], "download","download completed",-noprompt,-timeout,420,-fail,"software $_[0] failed since download was not successfull");
    }
    return 1;
}


#---------------------------------------------------------------------
# FUNCTION:     Execute update flash and reset bios
# INPUT:        $_[0] = BIOS version after download,
#				$_[1] = Filer connection handle, $_[2] = Login name,
#				$_[3] = password, $_[4] = filer name
# RETURN:       0 = Fail, 1 = Pass
#---------------------------------------------------------------------
sub Update_Flash_1
{
    # Local variables;
    my $version_str;
    my($bios_at_ontap)=@_;
    chomp($bios_at_ontap);

    # flash the image from CF to SF using update_flash command
    (undef,undef,$version_str, undef)=say($_[1],"update_flash -noversion",-swallow,-prompt,"$gLOADER","$gCFE",-timeout,300);
    if ($version_str!~/Saving Primary Image/) {
        return 0;
    }
    Boot($_[1],$_[2],$_[3],'bye',$_[4]);
    # Extract BIOS version from the LOADER after install of the image
    my $bios_after_install=&Bios_Version($_[1]);
    # Extract Loader version from the LOADER after install of the image
    my $loader_after_install=&Loader_Version($_[1]);
    # Extract BIOS version while update_flash
    my $bios_while_upd=&Extract_Bios_Ver_While_Update_Flash($version_str);
    # Extract Loader version while update_flash
    my $loader_while_upd=&Extract_Loader_Ver_While_Update_Flash($version_str);

    # Compare the new versions with the versions extracted after "download" and
    # versions extracted while update_flash, return 1 if version matches else
    # return 0
    logcomment("BIOS_AT_ONTAP:$bios_at_ontap:\n");
    logcomment("BIOS_while_update:$bios_while_upd:Loader while update:$loader_while_upd:\n");
    logcomment("BIOS_after_update:$bios_after_install:Loader_after_update:$loader_after_install:\n" );

    if($bios_at_ontap eq "$bios_while_upd") {
        logcomment("BIOS in CF matched with BIOS in SF");
        if($bios_after_install eq "$bios_while_upd") {
            logcomment("BIOS while update matched with BIOS in SF");
            if($loader_after_install eq "$loader_while_upd") {
                logcomment("Loader in SF matched with Loader while update");
                return 1;
            }
            else {
                 logcomment("Loader in SF did not match Loader while update");
            }

        }
        else {
            logcomment("BIOS in SF  did not match with BIOS while update");
        }
    }
    else {
        logcomment("BIOS in CF did not match with BIOS while update");
        return 0;
    }
}

#-----------------------------------------------------------------------------------------------------
# FUNCTION:     Extract BIOS version from the strings displayed while running
#				update_flash command
# INPUT:        $_[0] = strings displayed while running update_flash command
# RETURN:       BIOS version string
#--------------------------------------------------------------------------------------------------------
sub Extract_Bios_Ver_While_Update_Flash
{
    logcomment("Extract BIOS version while flash update");
    # Local variable
    my $bios_version_while_update;
    if($_[0] =~ /(.*New BIOS Version:.*)/) {
         $bios_version_while_update = (split(" ",$1))[-1];
         logcomment("Version in SF:$bios_version_while_update");
    }
    return $bios_version_while_update;
}

#-----------------------------------------------------------------------------------------------------
# FUNCTION:     Extract LOADER version from the strings displayed while running
#				update_flash command
# INPUT:        $_[0] = strings displayed while running update_flash command
# RETURN:       LOADER version string
#--------------------------------------------------------------------------------------------------------
sub Extract_Loader_Ver_While_Update_Flash
{
    logcomment("Extract Loader version while flash update");
    # Local variable
    my $loader_version_while_update;
    if($_[0] =~ /(.*New Loader Version:.*)/) {
        $loader_version_while_update = (split(" ",$1))[-1];
        logcomment("Version in SF:$loader_version_while_update");
    }
    return $loader_version_while_update;

}

#-----------------------------------------------------------------------------------------------------
# FUNCTION:     Function to enter login name and password if filer is not found
#				at ONTAP and Loader prompt
# INPUT:        $_[0] = Filer connection handle, $_[1] = Login ID,
#				$_[2] = Password
# RETURN:       0 = Fail, 1 = Pass
#--------------------------------------------------------------------------------------------------------
sub Cleanup_Prompt
{
    # Check filer state
    my $prompt = Filer_Prompt($_[0]);
    # Check if filer is at Loader prompt or Login prompt.
    # and Boot the Ontap.
    if ($prompt !~ /$gLOADER|$gCFE/) {
        if(!$prompt) {
            return 0;
        }
        if($prompt =~ /login:/) {
            (undef, undef, undef, $prompt)=say($_[0],$_[1],"-prompt","Password:",-swallow,-timeout,300);
        }
        if ($prompt =~ /Password:/) {
             say($_[0],$_[2],"-prompt",-swallow,-timeout,300);
        }
        Halt($_[0]);
     }
    return 1;
}

#-----------------------------------------------------------------------------------------------------
# FUNCTION:     Function to flash BIOS image using flash command at Loader
#				prompt
# INPUT:        $_[0] = Filer connection handle, $_[1] = flash0a or flash0b,
#				$_[2] = HTTP or TFTP URL of BIOS image, $_[3] = Login ID,
#				$_[4] = Password, $_[5] = RLM connection handle
# RETURN:       0 = Fail, 1 = Pass
#--------------------------------------------------------------------------------------------------------
sub Flash_Image
{
   # Local variable
   my $version_str;

   (undef,undef,$version_str,undef)=say($_[0],"flash $_[2] $_[1] -noversion",-swallow,-prompt,"$gLOADER",-timeout,300);
   if($version_str!~/written/) {
       logcomment("BIOS image was not successfully flashed");
       return 0;
   }
   Boot($_[0],$_[3],$_[4],'bye',$gLOADER);
   my $bios_while_flash=&Extract_Bios_Ver_While_Update_Flash($version_str);
   my $loader_while_flash=&Extract_Loader_Ver_While_Update_Flash($version_str);
   logcomment("BIOS_VERSION_WHILE_FLASH:$bios_while_flash:");
   logcomment("LOADER_VERSION_WHILE_FLASH:$loader_while_flash:");
   # Get the bios and loader versions after flashing
   my $bios_version_after_flash=&Bios_Version($_[0]);
   my $loader_version_after_flash=&Loader_Version($_[0]);
   logcomment("BIOS_VERSION_AFTER_FLASH:$bios_version_after_flash:");
   logcomment("LOADER_VERSION_AFTER_FLASH:$loader_version_after_flash:");
   # Compare the above versions and return 1 if match is successfull else return 0;
   if(("$bios_while_flash" eq "$bios_version_after_flash")&&("$loader_version_after_flash" eq "$loader_while_flash")) {
       return 1;
    }
    else {
        return 0;
    }
}

#-------------------------------------------------------------------------------
# FUNCTION: To reset the system BIOS through RLM
# INPUT:	$_[0]=rlm connection handle, $_[1]='primary' or 'backup'
# Return: 	NONE
#-------------------------------------------------------------------------------
sub System_Reset
{
    logcomment("Reset the BIOS to $_[1] through RLM");
    # Set the privilege to advanced mode
    say($_[0],"priv set advanced", -noprompt,'-timeout',100);
    # Execute 'system reset' command
    say($_[0],"system reset $_[1]",-prompt,"Continue", -swallow, '-timeout', 360);
    say($_[0], "y", -noprompt, -swallow,'-timeout', 360);
    snooze(60);
}
#----------------------------------------------------------------------------------
# FUNCTION:     Function to retrieve BIOS version string from "version -b" output
# INPUT:        $_[0] = Filer connection handle, $_[1] = SYSMODEL of the filer
# RETURN:       BIOS version string
#---------------------------------------------------------------------------------
sub Get_Bios_After_Download
{
    # Local variables
    my $sys_model=$_[1];

    my $bios_at_ontap;
    Ver_B_Ontap($_[0]);
    if ($sys_model =~ /FAS60.*/) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_EXCELSIOR};
    }
    elsif ($sys_model =~ /FAS30.*/) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_DRWHO};
    }
    elsif ($sys_model =~ /FAS31.*/) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_SPECTRE};
    }
    elsif (($sys_model =~ /FAS2020/) || ($sys_model =~ /FAS2050/)) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_SHUAR};
    }
    elsif ($sys_model =~ /FAS2040/) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_SHUAR_XL};
    }
    elsif ($sys_model =~ /FAS32.*/) {
        $bios_at_ontap=$Ver_Table{VERSION_BIOS_CARNEGIE};
    }
    else {
        logcomment("Filer need to be a BIOS based platform");
    }
    return $bios_at_ontap;
}

###############################################################################

#-------------------------------------------------------------------------------
# FUNCTION: To set the environment variable at the loader prompt.
# INPUT   : $_[0] = Filer connection handle, $_[1] = Variable name,
#			$_[2] = value.
# RETURN  : $response
#-------------------------------------------------------------------------------
sub Env_Set_Variable
{
    my (undef,undef,$response,undef) = say($_[0],"setenv $_[1] $_[2]",-prompt,$gLOADER,$gCFE,-swallow,-timeout,360);
    return $response;
}

#-------------------------------------------------------------------------------
# FUNCTION: To search a given file in root directory of filer's root
#			directory.
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = File name with absolute path from root directory.
#						eg: "/etc/syslog.conf"
# RETURN  : 0 = File Exists, 1 = File doesn't Exist
#-------------------------------------------------------------------------------
sub Search_File
{
	my (undef, undef, $response) = say($_[0],"rdfile $_[1]",-prompt,
										-swallow,-timeout,300);
	if ($response =~ /No such file/){
		return 1;
	}else{
		return 0;
	}
}

#-------------------------------------------------------------------------------
# FUNCTION: Rename a specified file name to a desired file name.
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = Source File name with absolute path from root directory,
#						eg: "/etc/syslog.conf"
#			$_[2] = Destination File name with absolute path from root directory
#						eg: "/etc/syslog_backup.conf"
# RETURN  : 0 = Success, 1 = Failure
#-------------------------------------------------------------------------------
sub File_Rename
{
	eval{
		say($_[0], "mv $_[1] $_[2]", -prompt, -swallow, -timeout, 100);
	};

	if ($@) {
         my $err = $@;
         logcomment("Failed to change Filename - $err");
         return 1;
	}else{
		return 0;
	}
}

#-------------------------------------------------------------------------------
# FUNCTION: To add a line entry to the specified file.
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = File name with absolute path from root directory,
#						eg: "/etc/syslog.conf"
#			$_[2] = 1 line data to be added.
#						eg: "*.info    /dev/console"
# RETURN  : 0 = Success, 1 = Failure
#-------------------------------------------------------------------------------
sub File_Append
{
	eval{
		my (undef,undef,$response) = say($_[0],"wrfile -a $_[1] '$_[2]'",-prompt,
										-swallow,-timeout,300);
	};
	if ($@) {
         my $err = $@;
         logcomment("Failed to Edit File $_[1] becuase of $err");
         return 1;
	}else{
		return 0;
	}
}

#-------------------------------------------------------------------------------
# FUNCTION: To get Total, Used and Free spaces in the specified volume and its
#			snapshot area.
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = unit(b=bytes |k=KB |m=MB |g=Gb |t=TB)
#			$_[2] = Volume name.
# RETURN  : %dfoutput
#-------------------------------------------------------------------------------
sub Get_Vol_Space_Info
{
   my $fc  = $_[0];
   say ($fc, "\r", -swallow, -timeout, 30);
   my (undef,undef,$result) = say($fc,"df -$_[1] $_[2]", -swallow, -prompt, -timeout,50);
   my @tmprslt = split(/\n/,$result);

   my @volout = split(/ +/,$tmprslt[2]);
   my @volsnapout = split(/ +/,$tmprslt[3]);
   my %dfoutput = (TOTAL => $volout[1], # total space on volume
           USED => $volout[2],  # Used space on the volume
           FREE => $volout[3],  # Free space on the volume
           SNAPTOTAL => $volsnapout[1], # total space in snapshots area on volume
           SNAPUSED => $volsnapout[2],  # used space in snapshots area on volume
           SNAPFREE => $volsnapout[3]); # free space in snapshots area on volume
   return (%dfoutput);
}

#-------------------------------------------------------------------------------------------
# FUNCTION: To fill the given volume with specified size of file
# INPUT   : $_[0] = Filer connection handle, $_[1] = Size of file e.g 100
#			$_[2] = units of size m->megabyte, t->terabyte etc
#           $_[3] = Name of volune in which file to be written e.g vol0
#			$_[4] = Name of file e.g test_file
# RETURN  : 1 = Success, 0=Failure
#-------------------------------------------------------------------------------------------
sub Fill_Volume
{
    use vars qw ( $fc $size $scale $vol_name $test_file_name %vol_status $used_space_before_vol_fill $used_space_after_vol_fill $cmd $wait_time);
    $fc = shift;
    $size = shift;
    $scale = shift;
    $vol_name = shift;
    $test_file_name = shift;

    # command sent on filer to fill the volume

    if($scale ne "b"){
    	chop ($size);
    	chop ($size);
	}

    if ($scale =~/t/i) {$wait_time = (($size * 1024)        / 2)* 60;}
    if ($scale =~/g/i) {$wait_time = ( $size 		    / 2)* 60;}
    if ($scale =~/m/i) {$wait_time = (($size / (1024 * 1))  / 2)* 60;}
    if ($scale =~/k/i) {$wait_time = (($size * (1024 * 2))  / 2)* 60;}

    # Swith to diags mode
    if (!(Set_Priv ($fc, 'diag'))) {
        return 0;
    }

    # To check $test_file_name file already exists, if yes then delete it first
    my (undef, undef, $out, undef) = say ($fc, "ls /vol/$vol_name", -swallow, -timeout, 30);
    if ($out =~ /$test_file_name/) {
        logcomment ("file named $test_file_name already exist");
        logcomment ("Deleting the old file $test_file_name");
        say ($fc, "rm /vol/$vol_name/$test_file_name", -swallow, -timeout, 30);
        # Snoozing for 30 seconds, after removing file using rm command,
        # it takes around 30 seconds to reflect the volume size in the output
        # of df command
        snooze(30);
    }

    # Get space info before filling the volume
    %vol_status = Get_Vol_Space_Info($fc,$scale,$vol_name);
    $used_space_before_vol_fill = $vol_status{USED};

    # Fill the volume with specified size
    eval {
	$cmd = "mkfile $size$scale /vol/$vol_name/$test_file_name";
	say ($fc, $cmd, -swallow, -timeout, "$wait_time");
    };
    if ($@) {
        my $err = $@;
        logcomment("Filling the volume is taking longer time than expected");
        return 0;
    }

    # Get space info after filling the volume
    say ($fc, "\r", -swallow, -timeout, 30);
    %vol_status = Get_Vol_Space_Info($fc,$scale,$vol_name);
    $used_space_after_vol_fill = $vol_status{USED};

    if($scale ne "b"){
    	chop ($used_space_after_vol_fill);
    	chop ($used_space_after_vol_fill);
    	chop ($used_space_before_vol_fill);
    	chop ($used_space_before_vol_fill);
	}
    snooze (15);
    # Confirm whether write process is successful or not
    if ($used_space_after_vol_fill >= $used_space_before_vol_fill + ($size - 1)) {
        logcomment ("$vol_name is filled with $size $scale"."b");
        return 1;
    } else {
        logcomment "Filling $vol_name volume is not successful";
        return 0;
    }
}

#-------------------------------------------------------------------------------------------
# FUNCTION: Delete the files in given volume, to make free space
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = name of volume e.g vol0
#           $_[2] = file name to be deleted e.g test_file.
# RETURN  : On success return 1, else 0, if file not present return 2
#-------------------------------------------------------------------------------------------
sub Free_Volume {
    use vars qw ( $fc $vol_name $test_file_name %vol_status $used_space_before $used_space_after $cmd);
    $fc = shift;
    $vol_name = shift;
    $test_file_name = shift;

    # Command sent on the filer
    $cmd = "rm /vol/$vol_name/$test_file_name";

    # Swith to diags mode
    if (!(Set_Priv ($fc, 'diag'))) {
        return 0;
    }

    # Get space info of volume before deletion of file
    %vol_status = Get_Vol_Space_Info($fc,"m", "$vol_name");
    $used_space_before = $vol_status{USED};

    # Confirm file exists which is to be deleted, if exist delete it
    my (undef, undef, $out, undef) = say ($fc, "ls /vol/$vol_name", -swallow, -timeout, 30);
    if ($out =~ /$test_file_name/) {
        say ($fc, "$cmd", -swallow, -timeout, 30);
        # Snoozing for 30 seconds, after removing file using rm command,
        # it takes around 30 seconds to reflect the size in the output
        # of df command.
        snooze(30);
    } else {
        logcomment ("File $test_file_name does not exist at /vol/$vol_name");
        return 2;
    }

    # Get space info of volume after deletion of file
    %vol_status = Get_Vol_Space_Info($fc,"m", "$vol_name");
    $used_space_after = $vol_status{USED};

    chop ($used_space_before);
	chop ($used_space_before);
	chop ($used_space_after);
	chop ($used_space_after);

    # Confire whether delete process is successful or not
    if ($used_space_after <= $used_space_before + ($size - 1)) {
        logcomment "Space free in $vol_name by size : $size mb";
        return 1;
    } else {
        logcomment "Not able to free the space in volume $vol_name";
        return 0;
    }
}

#-------------------------------------------------------------------------------
# FUNCTION: To delete a file from filer's root directory.
# INPUT   : $_[0] = Filer connection handle,
#			$_[1] = From root directory onwards, provide the absolute path of
#					 the parent directory of the file to be deleted
#						eg: 1. etc
#							2. etc/messages etc
#			$_[2] = File Name
#						eg: syslog.conf
# RETURN  : 0 = Success, 1 = Failure
#-------------------------------------------------------------------------------
sub Delete_File
{
	if (!(Set_Priv ($_[0],"diag"))) {
        return 1; # Failure
    }
    my (undef, undef, $response) = say($_[0],"ls /$_[1]",-prompt,
										-swallow,-timeout,300);
	if ($response =~ /$_[2]/){
		say($_[0],"rm /$_[1]/$_[2]", -prompt, -swallow, -timeout,300);
	}else{
		logcomment("Could not delete since, $_[2] file does not exist at all");
		return 2; # Failure or Success depending on the usage
	}
	(undef, undef, $response) = say($_[0],"ls /$_[1]",-prompt,
										-swallow,-timeout,300);
	if ($response =~ /$_[2]/){
		return 0; # Success
	}else{
		return 1; # Failure
	}
}
################################################################################
#-------------------------- 10-Mode Functions ----------------------------------
################################################################################


#-------------------------------------------------------------------------------
# FUNCTIN : Boot the OS from backup CF in 10 mode.
# INPUT   : $_[0] = Filer connection handle.
# RETURN  : NONE
#-------------------------------------------------------------------------------
sub Boot_Backup_10mode
{
    say($_[0],"boot_backup",-prompt,"How would you like to continue booting?",
                                                        -swallow,-timeout,260);
	say ($_[0], "normal", -noprompt, -swallow);
	snooze (10);
	Boot($_[0],$_[1],$_[2],$_[3],$_[4]);
}

#-------------------------------------------------------------------------------
# FUNCTION : To find the non active image in 10 Mode
# INPUT    : $_[0] = Filer connection handle, $_[1] = Filer name
# RETURN   : Name of non active image
#-------------------------------------------------------------------------------
sub Non_Active_Image_10mode
{
    my $fc = shift;
    my $filer = shift;

    my (undef, undef, $image_entry, undef) = say($fc,"system image show",
                            "-wait", "$filer",-swallow,-timeout,30);
    my @arry = split ("\n", $image_entry);

    if ($arry[1] =~ /true/i) {
        logcomment "Non active image is : image2";
        return "image2";
    } else {
        logcomment "Non -active image is : image1";
        return "image1";
    }
}

#-------------------------------------------------------------------------------
# FUNCTION : To update image in 10 Mode
# INPUT    : $_[0] = Filer connection handle, $_[1] = Filer name
#            $_[2] = Package url, $_[3] = Image name to be replaced
#            $_[4] = username, $_[5] = password
# RETURN   : Pass = 0, Fail = 1
#-------------------------------------------------------------------------------
sub Install_Ontap_10mode
{
    my $fc = shift;
    my $filer = shift;
    my $package_url = shift;
    my $replace_image = shift;
    my $username = shift;
    my $password = shift;

    my $cmd = "system image update -node $filer -package $package_url ".
                                    "-replace $replace_image -setdefault true";

    say($fc,"$cmd",-prompt,"Enter User:",-swallow,-timeout,30);
    say($fc,"$username",-prompt,"Enter Password:",-swallow,-timeout,30);
	my (undef, undef, $result, undef) = say ($fc, "$password",
	                                            -swallow,-timeout, 600);
	if ($result =~ /Installation complete/i) {
    	logcomment "system image update is successful";
    	return 0;
	} else {
    	logcomment "system image update is NOT successful";
    	return 1;
	}
}



#-------------------------------------------------------------------------------
# FUNCTION : To find the system model in 10 Mode
# INPUT    : $_[0] = Filer connection handle, $_[1] = Filer name
# RETURN   : System Model
#-------------------------------------------------------------------------------
sub Get_Sysmodel_10mode
{
	Pre_Cond_Halt_10mode($_[0], $_[1]);
    my $sys_model = Env_Get($_[0],"SYS_MODEL");
    logcomment ("System Model : $sys_model");

    $sys_model =~ /SYS_MODEL.*(FAS.*|IBM.*)/;
    $sys_model = $1;
    chop($sys_model);
    return $sys_model;
}



#-------------------------------------------------------------------------------
# FUNCTION : Function to enter login name and password if filer is not found at
#            ONTAP and Loader prompt
# INPUT    : $_[0]=Filer connection handle, $_[1]= Login, $_[2]= Password,
#            $_[3] = Filer name
# RETURN   : 1 if Pass 0 if Fail
#-------------------------------------------------------------------------------
sub Cleanup_Prompt_10mode
{
    # Check filer state
    my $prompt = Filer_Prompt($_[0]);
    # Check if filer is at Loader prompt or Login prompt.
    # and Boot the Ontap.
    if ($prompt !~ /$gLOADER|$gCFE/) {
        if(!$prompt) {
            return 0;
        }
        if($prompt =~ /login:/) {
            (undef, undef, undef, $prompt)=say($_[0],$_[1],"-prompt",
                                        "Password:",-swallow,-timeout,300);
        }
        if ($prompt =~ /Password:/) {
             say($_[0],$_[2],"-prompt",-swallow,-timeout,300);
        }
        Halt_10mode($_[0], $_[3]);
     }
    return 1;
}


#-------------------------------------------------------------------------------
# FUNCTION : Set Env var to default values
# INPUT    : $_[0]=Filer connection handle, $_[1]= Login, $_[2]= Password,
#			 $_[3]= Filer name
# RETURN   : NONE
#-------------------------------------------------------------------------------
sub Env_Set_Default_Variable_10mode
{
	say($_[0], "set-defaults", -prompt,$gLOADER,$gCFE,-swallow,-timeout,20);
	say($_[0],"setenv bootarg.init.boot_clustered TRUE",-prompt,$gLOADER,
	                                         $gCFE,-swallow,-timeout,20);
	Boot($_[0], $_[1], $_[2], 'boot_ontap', $_[3]);
	Halt_10mode($_[0], $_[3]);
}


#---------------------------------------------------------------------
# FUNCTION: To bring down the 10-mode filer to LOADER prompt.
# INPUT   : $_[0]=Filer console handle, $_[1]= filer name
# RETURN  : NONE
#---------------------------------------------------------------------
sub Halt_10mode
{
    say($_[0],"system node halt $_[1]", -prompt,
        "Are you sure you want to halt the node", -swallow,-timeout,20);
    my (undef,undef,undef,$prompt) = say($_[0],"y",-prompt,$gLOADER,
                                        $gCFE,-swallow,-timeout,300);
    if($prompt =~ /$gLOADER|$gCFE/) {
        logcomment('Filer halted successfully');
    }
}


#---------------------------------------------------------------------
# FUNCTION: PreCondition function checks for the prompt, if the prompt
# is ONTAP then it halts the system to LOADER.
# INPUT   : $_[0]=Filer console handle, $_[1] = Filer name
# RETURN  : NONE
#---------------------------------------------------------------------
sub Pre_Cond_Halt_10mode
{
    my (undef,undef,undef,$prompt) = say($_[0],"\r",-prompt,$gLOADER,
                                           $gCFE,-swallow,-timeout,30);
    if($prompt !~ /$gLOADER|$gCFE/) {
        Halt_10mode($_[0], $_[1]);
    }
}


#---------------------------------------------------------------------
# FUNCTION: Rebooting 10-mode filer.
# INPUT   : $_[0]=Filer console handle, $_[1]=Filer Login,
#           $_[2]=Filer login Password, $_[3] = Filer name
# RETURN  : NONE
#---------------------------------------------------------------------
sub Reboot_10mode
{
    say($_[0],"system node reboot $_[3]","System rebooting",-noprompt,-timeout,60);
    snooze(60);
    my (undef,undef,undef,$prompt) = say($_[0],"\r",-prompt,"login:",-timeout,300);
    if($prompt =~ /login:/) {
        (undef,undef,undef,$prompt) = say($_[0],$_[1],-prompt,"Password:",
                                                                -timeout,10);
        if ($prompt =~ /Password:/) {
            say($_[0],$_[2],-prompt,-timeout,10);
            logcomment('Login to the filer successfull');
        }
    }
}

#---------------------------------------------------------------------
#
# THIS FUNCTION WILL BE DEPRICATED IN FUTURE. USE 'Filer_Mode_Check()' INSTEAD
#
# FUNCTION	: Determines the filer mode and add 10 mode prompt if filer
#			  is in 10mode
# INPUT		: $_[0]= filer host object, $_[1]=Filer connection object,
#			  $_[2] = filer name
# RETURN	: Mode of the filer
#---------------------------------------------------------------------
sub Check_Filer_Mode
{
    my $fh = shift;
	my $fc = shift;
    my $filer = shift;
    my $mode;
	my $filertype = $fh -> hosttype();
    if ($filertype eq "filer-cfe") {
	    logcomment ("7-Mode/7G Filer");
	    $mode = "7mode";
    } elsif ($filertype eq "filer-ng-cfe") {
        $mode = "10mode";
        logcomment("10-Mode Filer");
		#Adding the Default 10 mode filer prompt to filer object
		Add_Prompt_10mode ($fc, $filer);
    }
	return $mode;
}

#---------------------------------------------------------------------
# FUNCTION	: Determines the filer mode and add 10 mode prompt if filer
#			  is in 10mode
# INPUT         : $_[0]= filer host object, $_[1]=Filer connection object,
#			  $_[2] = filer name, $_[3] = Expected Mode.
# RETURN	: Mode of the filer
#---------------------------------------------------------------------
sub Filer_Mode_Check
{
    my $fh = shift;
	my $fc = shift;
    my $filer = shift;
    my $mod = shift;
    my $mode;
	my $filertype = $fh -> hosttype();
    if (($filertype eq "filer-cfe") && ($mod eq "7mode")) {
	    logcomment ("7-Mode/7G Filer");
	    $mode = "7mode";
	    if (get_wafl_version($fh) <= WV_BOILERMAKER_MIN) {
	    	$mode = "7G";
	    }
    } elsif (($filertype eq "filer-ng-cfe") && ($mod eq "10mode")) {
        $mode = "10mode";
        logcomment("10-Mode Filer");
		#Adding the Default 10 mode filer prompt to filer object
		Add_Prompt_10mode ($fc, $filer);
    }
	return $mode;
}

#---------------------------------------------------------------------
# FUNCTION: Verify the Filer Prompt and Login to the FILER.
# INPUT   : $_[0]= Filer console handle, $_[1]=Login, $_[2]=Password
#			$_[3]= Filer Name
# RETURN  : 0 = PASS, 1 = FAIL
#---------------------------------------------------------------------
sub ONTAP_Login
{
	my $prompt;
	my $login_prompt = "login:";
	my $passwd_prompt = "Password:";
	$prompt = Filer_Prompt($_[0]);
print "\$prompt = $prompt\n";
	if($prompt =~ /$_[3]/) {
		logcomment("Filer is at ONTAP prompt");
		return 0;
	}

    if (($prompt =~ /$gLOADER/)||($prompt =~ /$gCFE/)){
    	Boot($_[0],$_[1],$_[2],"boot_ontap",$_[3]);
    	return 0;
	}
	else {
		if ($prompt =~ /$login_prompt/) {
			say($_[0],$_[1],-prompt,$passwd_prompt,-swallow,-timeout,300);
	    	(undef, undef, undef, $prompt) = say($_[0],$_[2],-prompt,
	    												-swallow,-timeout,300);
		} elsif ($prompt =~ /$passwd_prompt/) {
	    	(undef, undef, undef, $prompt)=say($_[0],$_[2],-prompt,-swallow,
	    														-timeout,300);
		}
    	if($prompt =~ /$_[3]/) {
	    	logcomment("Filer is at ONTAP prompt");
	    	return 0;
		}
		else {
	    	logcomment("Failed to Login to the filer");
	    	return 1;
    	}
    }
}

#---------------------------------------------------------------------
# FUNCTION: Add 10mode prompt to filer object
# INPUT:    $_[0]=Filer Connection object, $_[1]=filer name
# RETURN:   NONE
#---------------------------------------------------------------------
sub Add_Prompt_10mode {
    my $fc = shift;
    my $FILER = shift;
    my $prompt_10mode = $FILER."::.*>";

    my @prompts = $fc->prompts();

    #Add 10Mode filer prompt.
    push(@prompts, $prompt_10mode);
    $fc->prompts(@prompts);
}

#--------------------------------------------------------------------------
# FUNCTION: To enter and exit from dblade engine from 10mode shell
# INPUT:    $_[0] = Filer Connection Handle,  $_[1] = 'enter' or 'exit'
# RETURN:   None
#--------------------------------------------------------------------------
sub Dblade_10mode {
    my($fc, $state) = @_;
    if($state eq 'enter') {
        say ($fc, "run -node local", '-swallow');
    }
    elsif($state eq 'exit') {
        say ($fc, "exit", '-swallow');
    }
}

#---------------------------------------------------------------------
# FUNCTION: Confirms if filer is running 7G release
# INPUT:    $_[0]=Filer Host object, $_[1]=filer name
# RETURN:   1 =7G if filer running 7G releasse else 0
#---------------------------------------------------------------------
sub Check_7G_Mode {
    if (get_wafl_version($_[0]) <= WV_BOILERMAKER_MIN) {
        logcomment ("Filer is running 7G release");
    	return "7G";
    } else {
        return 0;
    }
}

#-------------------------------------------------------------------------------
# SUB     :  save_filer_logs -> Save and start fresh messages and ems logs.
# INPUT   :  filer console, FILER_VOL
# RETURN  :  time_stamp
#-------------------------------------------------------------------------------
sub save_filer_logs {

    my (%opts) = validate (
        @_,
        {
            cc   	      => { type => OBJECT },
            filer 	      => { type => SCALAR },
            vol_name 	  => { type => SCALAR },
            mode		  => { type => SCALAR },

        }
    );

    my (%saved_logs, $mount_point, $cmd, $ems, $message);
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);
    if ($opts{mode} eq "7mode") {
		$ems = "/etc/log/ems";
		$message = "/etc/messages";
	}
	elsif ($opts{mode} eq "10mode") {
		$ems = "/mroot/etc/log/ems";
		$message = "/mroot/etc/messages";
	}

    ## Mount root volume of filer
    $mount_point = mount_filer_vol (
        filer           => $opts{filer},
        cc              => $opts{cc},
        vol_name        => $opts{vol_name},
        mode		    => $opts{mode},
    );

    ##Get current timestamp
    ($sec,$min,$hour,$mday,$mon,$year,$wday, $yday,$isdst)=localtime(time);
    $year = $year+1900; $mon = sprintf('%02d', $mon+1);
    $mday = sprintf('%02d', $mday); $hour = sprintf('%02d', $hour);
    $min = sprintf('%02d', $min); $sec = sprintf('%02d', $sec);
    $saved_logs{time_stamp} = $year.$mon.$mday."_".$hour.$min.$sec;

    ##Saving messages file
    $saved_logs{orig_messages} = $mount_point. $message;
    $saved_logs{saved_messages} = $mount_point. $message;
    $saved_logs{saved_messages} .= $saved_logs{time_stamp};
    $cmd = "r mv $saved_logs{orig_messages} $saved_logs{saved_messages}";
    say ($opts{cc}, $cmd);

    ##Saving ems file
    $saved_logs{orig_ems} = $mount_point. $ems;
    $saved_logs{saved_ems} = $mount_point. $ems;
    $saved_logs{saved_ems} .= $saved_logs{time_stamp};
    $cmd = "r mv $saved_logs{orig_ems} $saved_logs{saved_ems}";
    say ($opts{cc}, $cmd);

    return (\%saved_logs);

} #ends sub save_logs

#-------------------------------------------------------------------------------
# SUB     :  mount_filer_vol -> NFS Mount the volume specified on client
# INPUT   :  filer name and client connnection
# RETURN  :  NONE
#-------------------------------------------------------------------------------
sub mount_filer_vol {

    my (%opts) = validate (
        @_,
        {
            filer	=> { type => SCALAR },
            cc          => { type => OBJECT },
            vol_name    => { type => SCALAR },
            mode    => { type => SCALAR },
        }
    );

    my  ($mount_point, $cmd, $volname);

    # Mount point
    $mount_point = "/t/".$opts{filer}."/root";
    $cmd = "r mkdir ". $mount_point;
    say ($opts{cc}, $cmd);

    # Volume
    $volname = "/vol/";
    $volname .= "$opts{vol_name}";

    if ($opts{mode} eq "10mode") {
    	$cmd = "r mount -t nfs ".$opts{filer}.":".$volname. " ".$mount_point;
	} else {
		$cmd = "r qa_mount $opts{filer}";
	}
    say ($opts{cc}, $cmd);

    return $mount_point;

} #ends sub mount_filer_vol()


#-------------------------------------------------------------------------------
# SUB     :  inspect_syslog -> Inspect /etc/messages log for errors
# INPUT   :  Log ref pointer, and messages
# RETURN  :  0 if pass, 1 if fail
#-------------------------------------------------------------------------------
sub inspect_logs {
	my (%opts) = validate (
		@_,
		{
		    log  			=> { type => OBJECT },
		    syslog_message	=> { type => SCALAR },
		    ems_event		=> { type => SCALAR },
		    check_for_panic	=> { type => SCALAR },
		}
	);

    my ($log, @ems_events, $ems, $status);
    $log = $opts{log};
    $status = 0;

    # Inspect Logs for messages FILER Panic
    if ($opts{check_for_panic} eq 'yes') {
		if ($log->panicked()) {
			logStpe ("Filer panicked");
	        $status = 1;
	    }
    }

 	# Inspect Logs for messages. Return 1 if fails
    if ($log->{'logs'}{'messages'}->to_string =~ /$opts{syslog_message}/gi) {
		logcomment ("Console message found => $opts{syslog_message}");
	} else {
		$status = 1;
	}

    # Inspect Logs for ems
	@ems_events = split ( /,/, $opts{ems_event} );
    $ems = $log->{'logs'}{'ems'}->find_events(@ems_events);
    if (defined(${$ems}[0])){
	    logcomment ("EMS event found => @{$ems}");

	} else {
		$status = 1;
	}
    return $status;
}


#-------------------------------------------------------------------------------
# SUB     :  cleanup_filer_save_logs -> Save logs during cleanup
# INPUT   :  filer name and client connnection
# RETURN  :  NONE
#-------------------------------------------------------------------------------

sub cleanup_filer_save_logs {

	    my (%opts) = validate (
        @_,
        {
            filer       	=> { type => SCALAR },
            cc          	=> { type => OBJECT },
            testname		=> { type => SCALAR },
            LOGDIR			=> { type => SCALAR },
            saved_logs 		=> { type => SCALAR},
            mode    		=> { type => SCALAR },
        }
    );

    my  (%saved_logs, $mount_point, $testname_messages, $testname_ems, $cmd);

    tstlog(-comment, "Mount filer vol");
    $mount_point = mount_filer_vol (
					filer		=> $opts{filer},
					cc			=> $opts{cc},
					vol_name	=> 'vol0',
					mode	    => $opts{mode},
    			);

    ##Saving messages file
    logcomment("Save Messages and EMS logs ");
    %saved_logs = %{$opts{saved_logs}};
    $testname_messages = $saved_logs{orig_messages}."_". $opts{testname};
    $testname_messages .= "_". $saved_logs{time_stamp}."\.log";
    $cmd = "r mv $saved_logs{orig_messages} $testname_messages";
    say ($opts{cc}, $cmd);
    $cmd = "r cp $testname_messages $opts{LOGDIR}";
    say ($opts{cc}, $cmd);
    $cmd = "r mv $saved_logs{saved_messages} $saved_logs{orig_messages}";
    say ($opts{cc}, $cmd);

    ##Saving ems file
    $testname_ems = $saved_logs{orig_ems}."_". $opts{testname};
    $testname_ems .= "_". $saved_logs{time_stamp}."\.log";
    $cmd = "r mv $saved_logs{orig_ems} $testname_ems";
    say ($opts{cc}, $cmd);
    $cmd = "r cp $testname_ems $opts{LOGDIR}";
    say ($opts{cc}, $cmd);
    $cmd = "r mv $saved_logs{saved_ems} $saved_logs{orig_ems}";
    say ($opts{cc}, $cmd);

}#ends sub cleaup_save_logs

1;
