
package platval::service_processor::scratchpad_forensics;

use strict;
use warnings;

use TCD;
use Tharn;
use NACL::APISet;
use NACL::Transit;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );
our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
    initialize_region
    clear_sel_logs
    verify_system_forensics_log_collection
    system_scratchpad_read
    verify_file_content
    get_privileges
);

#----------------------------------------------------------------------------
# FUNCTION:     To initialize bios region
# INPUT   :     SP System Shell API object
# RETURN  :     None
#----------------------------------------------------------------------------

sub initialize_region($) {

    my $APISetSystemShell = shift;
    $APISetSystemShell->execute_raw_command("command" => "sudo /bin/csh");
    $APISetSystemShell->ssram('initialize-region' => '1','region-type' => '1');
    $APISetSystemShell->ssram('initialize-region' => '1','region-type' => '2');
}

#----------------------------------------------------------------------------
# FUNCTION:     To clear SEL logs
# INPUT   :     SP API object
# RETURN  :     None
#----------------------------------------------------------------------------

sub clear_sel_logs($) {
 
    my $APISetSP = shift;
    $APISetSP->execute_raw_command("command" => "priv set test");
    $APISetSP->system_forensics_log("clear" => 1);
}

#----------------------------------------------------------------------------
# FUNCTION:     To verify if system forensic log collection happens
# INPUT   :     SP Root Shell API object, SP System Shell API object,
#               Transit object, SP SP API object
# RETURN  :     0 = pass; 1 = fail
#----------------------------------------------------------------------------

sub verify_system_forensics_log_collection($$$$$) {

    my ($APISetSPRoot,$APISetSP,$Transit_obj,$APISetSystemShell,$Test) = @_;
    my ($response,$response1,$response2,$input_file,$output_file,$timestamp_file,
        $Original_State,$tar_file,$parsed_output);
    my $forensic_time = "120";
    #Step1:Check if forensic_logs directory is present at "/mnt/logs/forenscis/"
    $Test->step("Check if \"forensic_logs\" directory is present at \"/mnt/logs/forenscis/\" directory");
    $response = $APISetSPRoot->execute_raw_command("command" => "ls /mnt/logs/forensics");
    if ($response =~ /forensics_logs/) {
        tstlog -pass, "\"forensics_logs\" directory is present at \"/mnt/logs/forenscis/\" directory";
        $Test->step("Verify whether the scratchpad related files are present in \"/mnt/logs/forenscis/forensics_logs\" directory");
        $response = $APISetSPRoot->execute_raw_command("command" => "ls /mnt/logs/forensics/forensics_logs");
        if (($response =~ /sram_bios_log/) && ($response =~ /sram_ontap_log/)) {
            tstlog -pass, "The scratchpad related files \"sram_bios_log\" and \"sram_ontap_log\" are present in \"/mnt/logs/forenscis/forensics_logs\" directory";
        return 0;
        }   
    } else {
        tstlog -comment, "\"forensics_logs\" directory is not present at \"/mnt/logs/forenscis/\" directory";
        $Test->step("Executing the below steps to make sure that automatic system forensics log collection happens");

        $APISetSPRoot->execute_raw_command("command" => "cd /mnt/logs/");
        $APISetSPRoot->execute_raw_command("command" => "rm -rf forensics");

        #Step2: Clearing all the messages from /var/log/messages directory
        $Test->step("Clearing all the messages from /var/log/messages directory");
        $APISetSPRoot->execute_raw_command("command" => "echo > /var/log/messages");

        #Step3: Enabling system forensics
        $Test->step("Enabling system forensics");
        $APISetSP->execute_raw_command("command" => "priv set test");
        $APISetSP->system_forensics_enable();
        $response = $APISetSP->system_forensics_show();
        $parsed_output = $response->get_parsed_output();
        if($$parsed_output[0]{'forensics_enabled'} =~ /true/) {
            tstlog -pass, "System Forensics enabled successfully";
        }else {
            tstlog -fail, "System Forensics could not be enabled";
            return 1;
        }

        #Step4: Set forensics time limit
        $Test->step("Set forensics time limit");
        $APISetSP->system_forensics_set_time_limit("time-limit" => "$forensic_time");
        $response = $APISetSP->system_forensics_show();
        $parsed_output = $response->get_parsed_output();
        if($$parsed_output[0]{'forensics_time_limit'} =~ /$forensic_time/) {
            tstlog -pass, "The specified time limit is set successfully";
        } else {
            tstlog -fail, "The specified time limit could not be set";
            return 1;
        }

        #Step5: Hang the controller to let system forensic data collection happens
        $Test->step("Triggering L2 timeout by ignoring NMI and hanging controller to let system forensic data collection happens");
        $APISetSystemShell->execute_raw_command("command" => "sudo sysctl debug.watchdog.disable=0");
        $APISetSystemShell->execute_raw_command("command" => "sudo sysctl debug.watchdog.ignore_nmi=1");
        $APISetSystemShell->execute_raw_command("command" => "sudo sysctl debug.watchdog.trigger_l2_timeout=1");
        $APISetSystemShell->execute_raw_command("command" => "sudo sysctl debug.watchdog.hang_cpu=1");

        tstlog -comment, "Sleeping for $forensic_time seconds";
        sleep($forensic_time);

        #Step6: Check for the message "Forensics data collection complete"
        $Test->step("Check for the message \"Forensics data collection complete\"");
        $response1 = $APISetSPRoot->execute_raw_command("command" =>"cat /var/log/messages | grep \"Forensics collection aborted due to timeout\"");
        print "response: $response1\n";
        if ($response1 =~ /Forensics collection aborted due to timeout/) {
            tstlog -fail, "Forensics aborted due to timeout. Increase the time to collect full forensics logs";
            return 1;
        }

        $response2 = $APISetSPRoot->execute_raw_command("command" => "cat /var/log/messages | grep \"Forensics data collection complete\"");
        if ($response2 =~ /Forensics data collection complete/) {
            tstlog -pass, "Expected message is displayed";
        } else {
            tstlog -fail, "Expected message is not displayed";
            return 1;
        }
        #Wait for the filer to come up
        tstlog -comment, "Waiting for the filer to come up";
        sleep(300);
        $Original_State = $Transit_obj->get_state(timeout => 600);
        tstlog -comment, "Original state is: $Original_State";
        $Transit_obj->change_state(to => 'CLI') if($Original_State !~ /CLI/);

        $APISetSP->api_reconnect_session() if !$APISetSP->api_is_session_alive();
        $APISetSPRoot->api_reconnect_session() if !$APISetSPRoot->api_is_session_alive();
        $APISetSystemShell->api_reconnect_session() if !$APISetSystemShell->api_is_session_alive();
        #Step7: Check if Forensics log directory is created or not
        $Test->step("Check if Forensics log directory is created or not");
        $response = $APISetSPRoot->execute_raw_command("command" => "ls /mnt/logs/forensics/");
        if ($response =~ /forensics_logs/) {
            tstlog -pass, "The Forensics log directory \"forensics_logs\" is created successfully";
        } else {
            tstlog -fail, "The Forensics log directory \"forensics_logs\" is not created";
            return 1;
        }

        #Step8: Check whether the number of output xml files created are same as
        #        that of input xml files and also whether a timestamp file is created
        $Test->step("Check whether the number of output xml files created are same as
                    that of input xml files and also whether a timestamp file is created");
        tstlog -comment, "Listing the input xml files";
        $APISetSPRoot->execute_raw_command("command" => "cd /etc/forensics/");
        $input_file = $APISetSPRoot->execute_raw_command("command" => "ls *.xml");
        tstlog -comment, "Listing the output xml files";
        $APISetSPRoot->execute_raw_command("command" => "cd /mnt/logs/forensics/forensics_logs/");
        $output_file = $APISetSPRoot->execute_raw_command("command" => "ls *.xml");
        if ($input_file eq $output_file) {
            tstlog -pass, "The output xml file corresponding to the input xml file is successfully created";
        } else {
            tstlog -fail, "The output xml file is not created";
            return 1;
        }
        $timestamp_file = $APISetSPRoot->execute_raw_command("command" => "ls");
        if ($timestamp_file =~ /timestamp/) {
            tstlog -pass, "timestamp file is created successfully in the output xml file directory";
        } else {
            tstlog -fail, "timestamp file is not created in the output xml file directory";
            return 1;
        }
        $APISetSPRoot->execute_raw_command("command" => "cd /mnt/logs/forensics/");
        $tar_file = $APISetSPRoot->execute_raw_command("command" => "ls *.tar");
        if ($tar_file =~ /.tar/) {
            tstlog -pass, "The .tar file \"forensics_logs.tar\" which contains the output xml files is created successfully";
        }else {
            tstlog -fail, "The .tar file \"forensics_logs.tar\" which contains the output xml files is not created";
            return 1;
        }
        tstlog -pass, "System forensics log collection happened successfully";
    } #end of else loop
    return 0;
}

#----------------------------------------------------------------------------
# FUNCTION:     To execute "system scratchpad read" command
# INPUT   :     SP API object
# RETURN  :     Contents of the "system scratchpad read" command
#----------------------------------------------------------------------------

sub system_scratchpad_read($) {
 
    my $APISetSP = shift;
    my ($response,@arr,@arr1,@arr2,$arr1,$arr2);
    $APISetSP->api_reconnect_session() if !$APISetSP->api_is_session_alive();
    $APISetSP->execute_raw_command("command" => "priv set test");
    $response = $APISetSP->execute_raw_command("command" => "system scratchpad read");
    @arr = split("region 2 header",$response);
    @arr1 = split("\n",$arr[0]);
    @arr2 = split("\n",$arr[1]);
    splice(@arr2,0,1,"region 2 header");
    $arr1 = join(" ",@arr1);
    $arr2 = join(" ",@arr2);
    if(($response =~ /region 1 header/) && ($response =~ /region 2 header/)) {
        tstlog -pass, "The \"system scratchpad read\" command works fine";
    } else {
        tstlog -fatal, "The \"system scratchpad read\" command does not work properly";
    }
    return ($arr1,$arr2);
}

#----------------------------------------------------------------------------
# FUNCTION:     To verify whether scratchpad related files are present
#               at "/mnt/logs/forenscis/forensics_logs" directory
# INPUT   :     SP Root Shell API object
# RETURN  :     Contents of the scratchpad related files
#----------------------------------------------------------------------------

sub verify_file_content($) {

    my $APISetSPRoot = shift;
    my ($response,@ontap_log_arr,$ontap_log_arr,@bios_log_arr,$bios_log_arr);
    $APISetSPRoot->api_reconnect_session() if !$APISetSPRoot->api_is_session_alive();
    $response = $APISetSPRoot->execute_raw_command("command" => "ls /mnt/logs/forensics/forensics_logs");
    if (($response =~ /sram_bios_log/) && ($response =~ /sram_ontap_log/)) {
        tstlog -pass, "The scratchpad related files \"sram_bios_log\" and \"sram_ontap_log\" are present in \"/mnt/logs/forenscis/forensics_logs\" directory";
        $APISetSPRoot->execute_raw_command("command" => "cd /mnt/logs/forensics/forensics_logs");
        $response = $APISetSPRoot->execute_raw_command("command" => "cat sram_bios_log");
        @ontap_log_arr = split("\n",$response);
        $ontap_log_arr = join(" ",@ontap_log_arr);
        $response = $APISetSPRoot->execute_raw_command("command" => "cat sram_ontap_log");
        @bios_log_arr = split("\n",$response);
        $bios_log_arr = join(" ",@bios_log_arr);
    }else {
        tstlog -fatal, "The scratchpad related files are not present in \"/mnt/logs/forenscis/forensics_logs\" directory";
    }
    return ($ontap_log_arr,$bios_log_arr);
}

#----------------------------------------------------------------------------
# FUNCTION:     To get the available privileges in SP CLI
# INPUT   :     SP API object
# RETURN  :     Array containing the available privileges in SP CLI
#----------------------------------------------------------------------------

sub get_privileges($) {

    my $APISetSP = shift;
    my (@privileges,@line,$response);
    $response = $APISetSP->execute_raw_command(command => "priv set \?");
    @line = split('\n',$response);
    foreach (@line) {
        if ($_ =~ /\w+\s+\w+\s+(\w+)\s+\-\s+\w+\s+\w+\s+\w+\s+\w+\s+\w+/) {
            push (@privileges,$1);
        } else {
            next;
        }
    }
    return @privileges;
}

1;
