# -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/raidDiskError.pm#1 $
#
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for injecting unrecoverable RAID disk errors
##
## @description
##     The RAID disk error library consists of the following functions:
##
##     raid_diskerror_inject - Inject the requested RAID error
##
## @usage
##     use raidDiskError;
##
## @author parags@netapp.com
##
## @status private
## There is a issue where FILER_CONN is overridden by some of the dependencies to use 
## #console, use FILER_CONN=#rsh as arg to ntest to avoid  

use strict;
use warnings;
use Tharn qw($Log host connect);
use raidv2_lib;
use raidqa_lib;
use Carp qw(croak);
use Volume qw(vol_exists vol_status vol_container);
use SayWrap qw(sayEOF); 
use Params::Validate qw(validate validate_pos SCALAR OBJECT);
use Module::Common qw(validate_mixed);
use Storage::HDDQualUtils qw(disk_list);

use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(raid_diskerror_inject);

# Set up standard params with defaults
require "FS/commonParams.pl";


## @name ::raid_diskerror_inject(vol => $vol, fileid => $fileid[, %opts])
##
## @summary
##    Inject real unrecoverable disk error
##
## @description
##     raid_diskerror_inject - Inject the requested RAID error
##
## @arg $filer Required: 
##	filer => filer name
##
## @arg $vol Required: 
##      vol => name of aggregate, traditional volume or flexible volume
##
## @arg $fileid Required:
##      fileid => fileid of the file to cause corruption on
##
## @arg %opts Optional: options can be set:
##
##      fbn    => Specific FBN to cause corruption on
##      tofbn  => Specific FBN to cause corruption till
##
## @return value
##      Returns 0 if command successful
##                          
## @example my $result = raid_diskerror_inject(filer=>$FILER, vol=>$vol fileid=>$inum) # -> corrupts fbn 0 by default
##          my $result = raid_diskerror_inject(filer=>$FILER, vol=>$vol, fileid=>$inum, fbn=>$fbn) # -> corrupts given fbn 
##	    my $result = raid_diskerror_inject(filer=>$FILER, vol=>$vol, fileid=>$inum, fbn=>$fbn, tofbn=>$tofbn) # -> corrupts fbns in the given range
## 
## 


sub raid_diskerror_inject(%)
{
	my %opts = validate(
		@_,
		{
			filer =>{ type => SCALAR}, 
			vol  => { type => SCALAR}, 
			fileid => { type => SCALAR},
			fbn  => { type => SCALAR, optional => 1, default => 0}, 
			tofbn => { type => SCALAR, optional => 1, default => 0}, 
		}
	);
	
	my $filer = $opts{filer};
	my $vol = $opts{vol};
	my $fileid = $opts{fileid};
	my $fbn = $opts{fbn};
	my $tofbn = $opts{tofbn};
	my ($host, $conn, $bn, $dbn, $factor, $entry, $sector);
	my $dlist;

	if($tofbn < $fbn) {
		$tofbn =  $fbn;
	}
	
	
	$host = host($filer);
	$conn = connect("#rsh");

	my $output = sayEOF($conn,"priv set test; disk_mung");
	if($output =~ m/simulator/) {
		croak("Filer $filer is a simulator, disk_mung not supported");
	}
	else {
		if($output =~ m/not found/) {
			croak("Filer $filer does not support disk_mung");
		}
	}

	#--------------
	# turn off media scrubbing, verify that it gets turned off
	#--------------
	sayEOF($conn, "options raid.media_scrub.enable off");
	#sleep(60);
	my $out = sayEOF($conn, "vol media_scrub status $vol");
	my $out2 = sayEOF($conn, "aggr media_scrub status $vol");
	if(($out !~ /Media scrubbing is currently disabled on this system/) && 
	   ($out2 !~ /Media scrubbing is currently disabled on this system/)) {
		croak("Unable to stop media scrub");
	}

	#--------------------------------
	# Get disk_list to check the BPS value of disks
	#-------------------------------
	$dlist = disk_list($conn, 'diag');

	for ($bn = $fbn; $bn <= $tofbn; $bn++) { 
		#-----------------
		# get diskid and dbn from dumpblock
		#-------------------
		my $output = sayEOF($conn, "priv set -q diag; dumpblock file -V $vol  $fileid $bn");
		my @block_info = split(/\n/, $output);
		my $diskid=0;
		my $dbn=0;
		foreach $entry (@block_info) {
			if ($entry =~ /disk (.*), dbn (.*), vbn (.*)/ ){
				$diskid=$1;
				$Log->log("diskid is $diskid \n");
				$dbn=$2;
				$Log->log("dbn is $dbn \n");
				next;
			}
		}

		if ($diskid eq "0" || $dbn == 0) {
			croak("Didn't get the diskid or dbn of the allocated block info\n");
		}
	
		#-------------------------
		# sector = factor*dbn
		#-------------------------
		if ($dlist->{$diskid}{'BPS'} == 520) {
			$factor = 8;
		} else {
			$factor = 9;
		}
		$sector = $dbn*$factor;

		$Log->log("The factor is $factor and bps value is $dlist->{$diskid}{'BPS'}\n");
		$Log->log("Injecting raid real disk error on fbn:$fbn disk:$diskid sector:$sector\n");

		#-------------------------
		# disk-mung the sector
		#------------------------	
		my @disks = _get_other_disks($host,$conn,$vol,$diskid);
		push(@disks, $diskid);
		foreach $entry (@disks) {
			if($entry ne "") {
				_disk_mung($conn, $entry, $sector);
			}
		}
	}
	#------------------------
	# enable media scrub back to on.
	#-------------------------
	sayEOF($conn, "options raid.media_scrub.enable on");
	sleep 5;

	#---------------
	# flush buffers to disk
	#----------------
	sayEOF($conn,"priv set -q diag; wafl flush cache");

	return 0;

}


sub _disk_mung {

	my ($conn, $diskid, $sector) = @_;
	$Log->log("\n disk $diskid sector $sector\n");
	my $out = sayEOF($conn, "priv set -q test;disk_mung $diskid $sector");
	if($out !~ /scsi_write_long succeeded on block $sector/) {
	        croak("disk_mung fail on disk:$diskid sector:$sector\n");
	}

}

sub _get_other_disks {
	my ($host, $conn,$vol,$diskid) = @_;
	my (@disks, $data, $plex_1);

	#-----------------------
	#find out the diskid belong to which raidgroup, and get the parity disks of that rg
	#-----------------------
	my $group = get_group_by_diskname($diskid);
	$Log->log( "group is $group\n");
	my @parity = raidv2_get_disks_by_group_type("$group","PARITY");
	$Log->log( "The parity disk of $group is $parity[0] \n");
	push(@disks, $parity[0]);
	eval {
		@parity = raidv2_get_disks_by_group_type("$group","DPARITY");
		$Log->log( "The dparity disk of $group is $parity[0] \n");
		push(@disks, $parity[0]);
	};
	if($@) {
		$Log->log("Dual parity disk does not exist for group $group");
	}


	#### get rg id (rg0)
	my $rg;
	if ($group =~ /\/(.*)\/(.*)\/(.*)/ ) {
		$rg = $3;
	}

	#-------------------------
	# find the corresponding disks from the other plex of the same rg
	#---------------------------
	my @plexes = raidv2_get_plexes_by_vol($vol);
	my $plex = get_plex_by_diskname($diskid);
	my $plex_other;
	foreach $plex_other (@plexes) {
		if ($plex_other ne $plex ) {
			$Log->log("The other plex is $plex_other\n");
			$plex_1 = $plex_other;
		}
	}

	if($plex_1 ne "") {
		$Log->log("plex: $plex plex1: $plex_1\n");
		my @data_1 = raidv2_get_disks_by_group_type("$plex_1/$rg","DATA");
		$Log->log("data of $plex_1/$rg is @data_1\n");
		push(@disks, @data_1);

		#-------------------------
		# find the parity disks from the other plex
		#-------------------------
		@parity = raidv2_get_disks_by_group_type("$plex_1/$rg","PARITY");
		$Log->log("The parity disk of $plex_1/$rg is $parity[0] \n");
		push(@disks, $parity[0]);
		eval {
			@parity = raidv2_get_disks_by_group_type("$plex_1/$rg","DPARITY");
			$Log->log("The dparity disk of $plex_1/$rg is $parity[0] \n");
			push(@disks, $parity[0]);
		};
		if($@) {
			$Log->log("Dual parity disk does not exist for group $plex_1/$rg\n");
		}
	}

	@disks;
}

1;
