# -*-perl-*x
#
# $Id: //depot/prod/test/Rsierranevada/lib/raidFijiError.pm#1 $
#
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for simulating different kind of RAID errors
##
## @description
##     The RAID error library consists of the following functions:
##
##     raid_fijierror_clear - Clear any Fiji error points
##     raid_fijierror_inject - Inject the requested RAID error
##     raid_fijierror_verify_userdata - Verify EMS error on user data read
##
## @usage
##     use raid_fijierror_inject;
##
## @author pankajg@netapp.com 
##
## @status private

package raidFijiError;

use strict;
use warnings;

use Carp qw(croak);
use Tharn qw($Log);
use Volume qw(vol_exists vol_status vol_container);
use SayWrap qw(sayEOF);
use LogReader::EMS;
use Params::Validate qw (validate validate_pos SCALAR OBJECT);
use Module::Common qw(validate_mixed);
use Utils qw(make_conn_and_complain);
use Filer::Version qw(:releases get_wafl_version);
use FS::Util::Utilities qw(access_file_blocks);
use Wafliron qw(wafliron_start);

use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(raid_fijierror_clear raid_fijierror_inject raid_fijierror_verify_userdata);


## @name ::raid_fijierror_clear(conn => $conn)
##
## @summary
##     Library function for clearing any Fiji error points
##
## @description
##     raid_fijierror_clear - Clear any Fiji error points
##
## @arg $conn Required:
##      conn => filer connection object.
##
## @return value
##      Returns nothing
##                          

sub raid_fijierror_clear(%)
{
    my %opts = validate(
	@_,
	{
		conn => { type => OBJECT },
	}
    );

    my $conn = $opts{conn};

    ## Clear any fiji error injection points
    my $fijicmd = "fiji clear all";

    my $output = sayEOF($conn, $fijicmd);
    if (!($output =~ /filter cleared/i)) {
	croak("fiji clear: $output");
    }

    return 0; 
}

## @name ::raid_fijierror_inject(conn => $conn, vol => $vol [, %opts])
##
## @summary
##     Simulate Fiji error point as specified
##
## @description
##     raid_fijierror_inject - Inject the requested RAID error
##
## @arg $conn Required:
##      conn => filer connection object.
##
## @arg $vol Required: 
##      vol => name of aggregate, traditional volume or flexible volume
##
## @arg %opts Optional: options can be set:
##
##      fileid => fileid of the file to cause corruption on
##      fileop => fileid operator "!=", "<", "<=", "==", ">=" or ">"
##      path   => file path if corruption is desired on a given file
##                This is ignored if a fileid is specified. 
##      snapid => snapid of snapshot on which corruption is requested. 
##                Default is 0 (Active FS)
##      level  => block level for which error has to be injected, default is 0
##      fbn    => Specific FBN to cause corruption on
##      vbn    => Specific VBN to cause corruption on
##      count  => Number of errors to inject, default is 999
##      is_priv=> Inject private/public inode space.Applicable to dual inode concept.
##	iron   => 1 implies start wafliron after injecting error.
##      unverified => run fiji with unverified flag  

## @return value
##      Returns 0 if command successful
##                          
## @example my $result = raid_fijierror_inject(conn => $conn, vol => $vol, 
##						level=>0);
##          my $result = raid_fijierror_inject(conn => $conn, vol => $vol, 
##                                 		path=>"/vol/$vol/f1", 
##						snapid=>1); 
##
sub raid_fijierror_inject (%) 
{
    my %opts = validate( 
	@_,
	{
		conn    => { type => OBJECT},
                vol     => { type => SCALAR},
		fileid	=> { type => SCALAR, optional => 1 },
		fileop	=> { type => SCALAR, optional => 1, default => "==" },
		path	=> { type => SCALAR, optional => 1 },
		snapid	=> { type => SCALAR, optional => 1 },
		level	=> { type => SCALAR, optional => 1 },
		fbn	=> { type => SCALAR, optional => 1 },
		vbn	=> { type => SCALAR, optional => 1 },
		count	=> { type => SCALAR, optional => 1, default => 999 },
		is_priv => { type => SCALAR, optional => 1, default =>0 },
		iron => { type => SCALAR, optional => 1, default =>0 },
                unverified => { type => SCALAR, optional => 1, default =>0 },
	}
    );

    my $conn = $opts{conn};
    my $vol = $opts{vol};

    my ($fileid, $path, $snapid, $level, $fbn, $vbn, $count, $is_priv,$unverified);

    my $fileidstr = "";
    my $snapidstr = "";
    my $levelstr = "";
    my $fbnstr = "";
    my $vbnstr = "";
    my $countstr = "";
    my $privstr = "";
    
    
    my $fh = $conn->{'hostp'};
    my $version= get_wafl_version($fh);
   
    if (($version >= WV_BOILERMAKER_MIN && $version <= WV_BOILERMAKER_MAX) ||
       ($version >= WV_ROLLINGROCK_MIN && $version <= WV_ROLLINGROCK_MAX)) {

       #Version is BR.x or RR 
       $is_priv=$opts{is_priv};
       $privstr = "(== \@is_priv $is_priv)";
    }
    if (defined($opts{fileid})) {
        $fileid = $opts{fileid};
        $fileidstr = "($opts{fileop} \@fileid $fileid)";
    } 

    if (defined($opts{path})) {
        $path = $opts{path};
    }

    # Use path only if fileid is not defined 
    if (!defined($opts{fileid}) && defined($opts{path})) {
	my $output = sayEOF($conn, "priv set -q diag; showfh $path");
    	if (!($output =~ /fileid/i)) {
		croak("showfh $path: $output");
    	}
        $output =~ /fileid=0x0*([0-9,a-f,A-F]*)/; # leading 0x and 0's 
						  # confuse hex()
	my $fileid = hex($1); # yes, hex() actually returns base 10
        $fileidstr = "($opts{fileop} \@fileid $fileid)";
    }

    if (defined($opts{snapid})) {
        $snapid = $opts{snapid};
        $snapidstr = "(== \@snapid $snapid)";
    } 

    if (defined($opts{level})) {
        $level = $opts{level};
        $levelstr = "(== \@level $level)";
    }

    if (defined($opts{fbn})) {
        $fbn = $opts{fbn};
        $fbnstr = "(== \@fbn $fbn)";
    }

    if (defined($opts{vbn})) {
        $vbn = $opts{vbn};
        $vbnstr = "(== \@vbn $vbn)";
    }
    
    $unverified = $opts{unverified};
    $count = $opts{count};
    $countstr = "repeat $count";

    ## Introduce data corruption for the volume
    $Log->trace("Introduce data corruption on volume");
    my $fijicmd="";
    if($unverified==0)
    {
        $fijicmd = "fiji filter \"( $countstr (raid.return.pseudo.bad.error (&& (eq \@vol $vol) $levelstr $fileidstr $privstr $snapidstr $fbnstr $vbnstr)  (return (true))))\"";
    }
    else
    {
        $fijicmd = "fiji filter \"( $countstr (raid.return.unverified.blk.error (&& (eq \@vol $vol) $levelstr $fileidstr $privstr $snapidstr $fbnstr $vbnstr)(return (true))))\"\\";
    }

    my $output = sayEOF($conn, $fijicmd);
    if (!($output =~ /filter set/i)) {
	croak("fiji filter: $output");
    }

	if ($opts{iron} =~ /1/) {
		my ($found, $error) = wafliron_start(
											$conn,
											$vol,
											wait => 0,
											scan_speed => 10);
	}
    return 0; 
}

## @name ::raid_fijierror_verify_userdata(conn => $conn, vol => $vol, 
##					path => $path, pattern => $pattern)
##
## @summary
##     Verify EMS error on user data read
##
## @description
##     raid_fijierror_verify_userdata - Verify EMS error on user data read
##
## @arg $conn Required:
##      conn => filer connection object.
##
## @arg $vol Required: 
##      vol => name of aggregate, traditional volume or flexible volume
##
## @arg $path Required: 
##      path   => path for the file to verify corruption for
##
## @arg $pattern Required: 
##      pattern => The EMS event pattern to look for
##
## @arg $debug Optional:
##      debug => 1 implies , the output of 'ems event status' command will 
##      be shown in the log for the debugging purpose while verifying the ems.
##
## @return value
##      Returns reference to an array containing the strings that
##      matched the specified patterns or an empty array if no
##      matches were found.
##                          
## @example my $result = raid_fijierror_verify_userdata(conn => $conn, 
##					vol => $vol, 
##                                      path => "/etc/messages", 
##					pattern => "testpattern");
##

sub raid_fijierror_verify_userdata (%)
{
    my %opts = validate( 
	@_,
	{
		conn    => { type => OBJECT},
                vol     => { type => SCALAR},
		path	=> { type => SCALAR},
		pattern	=> { type => SCALAR},
                debug   => { type => SCALAR, optional => 1, default => 0},
	}
    );

    my $conn = $opts{conn};
    my $vol = $opts{vol};
    my $path = $opts{path};
    my $pattern = $opts{pattern};
    my $debug = $opts{debug};

    ## Start capturing the EMS
    my $emsLog = LogReader::EMS->new($conn);
    if ($debug == 1) {
       $emsLog->start_tail(pattern => $pattern,
                           debug  => $debug);
    }
    else {
       $emsLog->start_tail(pattern => $pattern);
    }

    ## Flush the WAFL cache
    sayEOF($conn, "priv set -q diag; wafl flush cache");

    ## Read the corrupted file 
	access_file_blocks(conn => $conn, path => $path, msgtype => 'abortable');

    ## Stop capturing the EMS
    $emsLog->stop_tail([$pattern]);
  
    ## Search the EMS log for desired EMS event
    my $err_msgs = $emsLog->find_events($pattern);

    my $num_match = 0;
    foreach (@{$err_msgs}) {
       $num_match++;
       $Log->trace("Matched $_");
    }

    $Log->trace("Num matched elements $num_match");

    if ($num_match > 0) {
        $Log->trace("Lib Detected data corruption");
    }
    else {
        $Log->trace("Lib Did not detect data corruption");
    }

    return $err_msgs;
}

1;
