# -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/raidv2_util.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary RAIDv2 util library
## 
## @description 
##  The library defines common routines to be use for RAID
##  testing.
## 
## @usage
##  Here is the list of the defined modules and functions which
##  can be used for both filers and simulators.
##   
##  ==========================================================
##  Function Prototype
##  
##  raidv2_ata_fiji_inject_sim($) - inject error to a sim disk
##  raidv2_ata_fiji_inject_hw($) - inject error to a filer disk
##  raidv2_ata_cleanup_test($$$) - cleanup ATA before exit
##  raidv2_fail_disks($) - fail the specified num of disks
##  raidv2_disk_unfail($) - unfail the specified list of disks
##  raidv2_disk_verify_opt_recons($) - verify disk was optimized
##  raidv2_get_largest_spare() - return the largest spare
##  raidv2_get_spare_disks() - get a list of spares
##  raidv2_init_test() - initialize a test 
##  raidv2_mount_root_vol($) - mount the specified filer root vol
##  raidv2_start_hammer($$) - start hammer with the specified size
##  raidv2_pause_hammer() - suspend all hammer
##  raidv2_restart_hammer() - restart all hammer
##  raidv2_pull_push_disks($) - pull and push the specified disks
##  raidv2_set_flags($) - set flag options 
##  raidv2_umount_root_vol($) - umount the root volume 
##  raidv2_vol_create_verify($$) - create specified volume 
##  raidv2_vol_is_degraded($) - verify the vol is degraded
##  raidv2_vol_recons_wait($) - wait until the recons is done
##  raidv2_vol_is_resync($) - wait until resync is started
##  raidv2_vol_verify($) - run verify on the mirror volume 
## 
##  ===========================================================
## @status public
## @author dl-step@netapp.com, dl-sar-qa@netapp.com
## 

#require "tharnit_lib.pl"; # needed for the saycmd function
use LogReader;
use mount;

# ------------------------------
# define global status variables
# ------------------------------
$PASSED = 1; 
$FAILED = 0; 
$NOOP   = 2; # no operation is required

#@ISA = qw(Exporter);
#@EXPORT = (raidv2_init_test, 
#	   raidv2_ata_fiji_inject_hw,
#	   raidv2_ata_fiji_inject_sim,
#	   raidv2_vol_create_verify, 
#	   raidv2_set_flags
#	   );
#use strict;

# ============================ BEGIN =============================

# --------------------------------------------
## @name ::raidv2_ata_fiji_inject_hw()
## @summary inject fiji points to the appliance disk 
## @description 
##  This function injects the fiji to the specified disk
##  of the specified appliance. This should only be tested
##  against filer (not simulator). 
## @synopsis
##  raidv2_ata_fiji_inject_hw(disk); 
## @arg disk required: disk to be injected with fiji points
# --------------------------------------------
sub raidv2_ata_fiji_inject_hw($) {
    my $disk = $_[0];

    raidv2_fiji_clear_all();
    for my $ec (split(' ', $FIJI_TYPE_HW)) {

	if ($VERBOSE) {
	    tstlog -comment, "ata_fiji_inject_hw: inject error code <$ec>"; 
	}

	raidv2_saycmd($FILER, "priv set -q test; scsi set_error -e $ec -l 1 $disk"); 
	raidv2_saycmd($FILER, "priv set -q test; scsi force_error -e $disk"); 
    }
} #raidv2_ata_fiji_inject_hw

# ---------------------------------------------------------
## @name ::raidv2_ata_fiji_inject_sim()
## @summary inject fiji points to the disk on simulator
## @description 
##  The function injects one or multiple fiji points to the
##  specified disk of the current simulator. For simulator
##  only. It should work for only one disk. 
## @synopsis
##  raidv2_ata_fiji_inject_sim(disk)
## @arg disk required: disk to be injected with fiji points
# ---------------------------------------------------------
sub raidv2_ata_fiji_inject_sim($) {
    my $disk = $_[0];

    #raidv2_fiji_clear_all();
    foreach my $fiji (split(' ', $FIJI_TYPE_SIM)) {

	if ($VERBOSE) {
	    tstlog -comment, "ata_fiji_inject_sim: inject point <$fiji>"; 
	}
	    
        if (uc($fiji) eq "DISK.IO.AGRSV.TIMEOUT.FAILURE") {
            raidv2_fiji_add_filter("(disk.io.agrsv.timeout.failure (eq \@disk $disk) return 1)");
            next;
        }
	
        if (uc($fiji) eq "RAID.DISK.READ.DONE") {
            raidv2_fiji_add_filter("(raid.disk.read.done (eq \@disk $disk)
 return buf_Errno E_DISK_AGRSV_TIMEOUT)");
            next;
        }

        if (uc($fiji) eq "RAID.LABEL.WRITE.DONE") {
            raidv2_fiji_add_filter("(raid.label.write.done (eq \@disk $disk) return buf_Errno E_DISK_AGRSV_TIMEOUT))");
            next;
        }

    }#foreach
} #raidv2_ata_fiji_inject_sim

# -------------------------------------------------------------
## @name ::raidv2_ata_cleanup_test
## @summary stop the hammers, cleanup the test environment, and
##  print the write the result message. 
## @description
##   The function will stop the hammers, destroy
##   the test volume if specified, online all disks, stop all 
##   injection points, and zero all disks.
##   The function should be called before exiting the test.
## @synopsis
##  raidv2_ata_cleanup_test($status, $message, $rmvol)
## @arg status Required - pass, fatal, config, warning. 
## @arg message Required - the result message
## @arg rmvol Required - 1 remove the test volume or 0 - save the volume
## @returns 1 -- success
# -------------------------------------------------------------
sub raidv2_ata_cleanup_test($$$) {
    my ($status, $message, $rmvol) = @_;

    raidv2_stop_hammer();
    if ($rmvol && uc($VOL_SAVE) eq "NO") {
        raidv2_remove_volume($FILER_VOL);
    }

    #raidv2_fiji_clear_all(); 
    raidv2_disk_online_all();
    raidv2_unfail_disks();
    raidv2_saycmd($FILER, "priv set -q diag; disk zero spares");

    #skip waiting for now
    #raidv2_wait_for_zeroing();

    tstlog $status, $message;
    return 1;
}#raidv2_ata_cleanup_test

# -------------------------------------------------------------
## @name ::raidv2_mount_root_vol
## @summary mounts the root volume on to client
## @description
##  The routines mounts the specified root volume to the current
##  client so the test program can have access to EMS message log
##  and other system files. 
## 
##  The caller must invoke raidv2_umount_root_vol to dismount
##  the mount point when there is no more access to the root
##  volume. 
## @arg $filername - name of the filer which root volume to be mounted
# -------------------------------------------------------------
sub raidv2_mount_root_vol($)
{
    my $filername = $_[0];
    
    # - - - - - - - - - - - - - - - - - - - -
    # License NFS 
    # - - - - - - - - - - - - - - - - - - - -
    saycmd($FILER, "priv set -q test; license add TZBGCKE");

    # - - - - - - - - - - - - - - - - - - - -
    # Unmount to clear stale mounts
    # - - - - - - - - - - - - - - - - - - - -
    raidv2_client_saycmd($CLIENT, "/usr/software/bin/sudo umount -f /t/$filername/root");

    #MTK_EDIT:24/09/08:: Added the Code to mount the root volume         
    # - - - - - - - - - - - - - - - - - - - -
    # Mount the Filer
    # - - - - - - - - - - - - - - - - - - - -
    saycmd($FILER, "exportfs -p sec=sys,rw,anon=0,nosuid /vol/vol0");
    saycmd($FILER, "exportfs -a");
    
    raidv2_client_saycmd($CLIENT, "/usr/software/bin/sudo sim_mount $filername");
} #raidv2_mount_root_vol


# -----------------------------------------------
## @name ::raidv2_init_test
## @summary initialize the test environment
## @description
##  initialization taking place by removing the 
##  existing volume, zeroing all spares, and 
##  wait until zeroing is done. 
## @synopsis 
##  raidv2_init_test() 
# -----------------------------------------------
sub raidv2_init_test() {
    raidv2_remove_volume($FILER_VOL); #just in case
    if ($FILER_VOL_SPLIT != "") {
      raidv2_remove_volume($FILER_VOL_SPLIT); #just in case
    }
    raidv2_saycmd($FILER, "priv set -q diag; disk zero spares");
    raidv2_wait_for_zeroing();
    raidv2_info_load(); 
}#raidv2_init_test


#---------------------------------------------------
## @name ::raidv2_set_flags
## @summary set specific flags ON/OFF before testing
## @description and flags are defined in parameter FLAG_OPTIONS
## @synopsis 
##  raidv2_set_flags($cmd)  
## @arg cmd Required 
##  cmd - 1 to set flags
##  cmd - 0 to unset flags
#---------------------------------------------------
sub raidv2_set_flags($) {
    my $cmd = $_[0];
    foreach $flag (split(' ', lc($FLAG_OPTIONS))) {
	if ($VERBOSE) {
	    tstlog -comment, "set_flags: flag to be set <$flag>"; 
	}
	
        my $fl = lc($flag);
        raidv2_saycmd($FILER, "priv set test; setflag $fl $cmd");
    }#foreach
} #raidv2_set_flags

# -----------------------------------------------
## @name ::raidv2_vol_create_verify
## @summary create the specified vol with the specified disks
## @description
##  create the specified vol with the specified of disks.
##  If the list_disks is no passed, the number of disks
##  will be determined based on the $VOL_CREATE_COUNT 
##  value. In this case, the function will verify to
##  make sure there are sufficient spares. 
## @synopsis
##   raidv2_vol_create_verify($vol, $list_disks)
## @arg vol Required -- the specified volume name
## @arg list_disks Optional - list of disks 
## @return $PASSED or $FAILED
# -----------------------------------------------
sub raidv2_vol_create_verify($@) {
    my ($vol, @list_disks) = @_;

    if ($VERBOSE) {
	tstlog -comment, "vol=<$vol>; list_disks=<@list_disks>"; 
    }
    my @spare_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool0"}); 
    my @spare1_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool1"}); 
    push(@spare_disks, @spare1_disks);

    my $spare_count = $#spare_disks + 1;
    if ($spare_count < $VOL_CREATE_COUNT) {
        tstlog -comment, "Need $VOL_CREATE_COUNT spare disks to run this test.";
        return $FAILED;
    }

    my $VOL_CREATE_ARGS = ""; 
    if (length(@list_disks) > 2) {
	# create a volume with the specified disks
	#
	$VOL_CREATE_ARGS = "$vol -t $RAID_TYPE -d @list_disks";
    } else {
	$VOL_CREATE_ARGS = "$vol -t $RAID_TYPE $VOL_CREATE_COUNT";
    }
    if (!raidv2_vol_create($VOL_CREATE_ARGS)) {
        tstlog -comment, "Unable to create a test volume <$vol>";
        return $FAILED;
    }

    tstlog -comment, "Successfully created a test volume <$vol>";
    return $PASSED;
}#raidv2_vol_create_verify


# -------------------------------------------------------------
## @name ::raidv2_get_largest_spare
## @summary find and return the largest spare 
## @description 
##  identifies the largest spare that is available and
##  returns the disk. If all spares have the same size, then it
##  will pick the first on the spare list.  
## @synopsis raidv2_get_largest_spare()
## @arg 
##  None
# ------------------------------------
sub raidv2_get_largest_spare() {
    raidv2_info_disk_load();

    # search for spares from both Pool0 and Pool1
    my @sdisks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool0"});
    my @sdisk2s = split(",", $raidv2_info_tree_disk_list{"POOL:Pool1"});

    push(@sdisks, @sdisk2s);
    my $largest_size = $raidv2_info_disk_by_name{$sdisks[0]}{avl_blks};
    my $largest_disk = $sdisks[0];

    # ---------------------------
    # search for the largest disk
    # ---------------------------
    for (my $j=1; $j <= $#sdisks; $j++ ) {
        if ($raidv2_info_disk_by_name{$sdisks[$j]}{avl_blks} > $largest_size) {
            $largest_disk = $sdisks[$j];
            $largest_size = $raidv2_info_disk_by_name{$sdisks[$j]}{avl_blks};
        }
        tstlog -comment, "disk size= $sdisks[$j]";
    }

    tstlog -comment, "Large disk = $largest_disk";
    return $largest_disk;
}#raidv2_get_largest_spare

# -----------------------------------------------
## @name ::raidv2_get_spare_disks
## @summary get a list of spares from both pools. 
## @description
##  Returns a list of spares from both Pool0 and Pool1
## @synopsis 
##  raidv2_get_spare_disks()
## @arg 
##  None
# -----------------------------------------------
sub raidv2_get_spare_disks() {
    my @spare_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool0"}); 
    my @spare1_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool1"}); 
    push(@spare_disks, @spare1_disks);

    my $spare_count = $#spare_disks + 1;
    if ($spare_count < 1) {
        tstlog -fatal, "There is no spare available"; 
        return $FAILED;
    }
    return @spare_disks; 
}#raidv2_get_spare_disks

# ------------------------------------------------
## @name ::raidv2_fail_disks
## @summary fail the number of spares by creating 
## a test volume, then fails all disks in the vol, 
## and then destroys the volume.
##
## @description
##  The function will interact with the filer/sim
##  to create a temp volume with the specified number
##  of the disks and then try to fail all of disks.
## @synopsis
##  raidv2_fail_disks($disk_cnt)
## @arg disk_cnt Required - number of disks to be created
##  with the test volume. 
## return -- a list of failed disks. 
# ------------------------------------------------
sub raidv2_fail_disks($) {
    my $disk_cnt = $_[0]; 

    my $vol_foo = "vfoo";
    $disk_cnt++; # increase by 1 
    raidv2_vol_create("$vol_foo -t 4 $disk_cnt"); 

    # -------------------------------------------
    # Let fail a bunch of parity and data disks
    # -------------------------------------------
    my @disks = raidv2_get_disks_by_type($vol_foo, "DATA"); 

    if ($VERBOSE) {
	tstlog -comment, "list_disks: <@disks>"; 
    }

    # fail data and parity disks
    # 
    for($i=0; $i <= $#disks; $i++) {
	if ($VERBOSE) {
	    tstlog -comment, "disk fail <$disks[$i]>"; 
	}
	raidv2_saycmd($FILER, "priv set -q diag; disk fail -f $disks[$i]"); 
    }
    
    # destroy the volume
    raidv2_destroy_volume($vol_foo);

    # let see all of the failed disks
    raidv2_saycmd($FILER, "sysconfig -r"); 

    return @disks; 
} #raidv2_fail_disks

#-------------------------------------------------------------
## @name ::raidv2_disk_unfail 
## @summary unfails a list of specified failed disks
## @description - issue the disk unfail command for each of
##  the specified disk. Zero all spares and wait for zeoring
##  to be completed. 
## @synopsis 
##  raidv2_disk_unfail(disks) 
## @arg 
##  disks Required -- specify a list of disks to be unfailed. 
## @return -- PASSED if everything is going well. 
#-------------------------------------------------------------
sub raidv2_disk_unfail($) {
    my @disks = @_; 

    if ($VERBOSE) {
	tstlog -comment, "disks to be unfailed: <@disks>"; 
    }

    # unfail the specified disks
    # 
    for($i=0; $i <= $#disks; $i++) {
	if ($VERBOSE) {
	    tstlog -comment, "disk to be unfailed: <@disks[$i]>"; 
	}
	raidv2_saycmd($FILER, "priv set -q diag; disk unfail @disks[$i]"); 
    }
    
    raidv2_saycmd($FILER, "priv set -q diag; disk zero spares");
    # check out volume and disks
    raidv2_saycmd($FILER, "sysconfig -r"); 

    #skip waiting for now
    raidv2_wait_for_zeroing();

    return $PASSED; 
} #raidv2_disk_unfail

# -----------------------------------------------------------------
## @name ::raidv2_disk_verify_opt_recons
## @summary verifies optimize recons was completed, is going or just
## started on the specified disk. 
## @description routine verifies whether or not the specified disk got
## optimize reconstructed or it could be in the normal state.
## 
## First, verifies if the disk is being optimize reconstructed or
## it is being onlined and it is done with either one.  
## 
## If not checks and verifies the parent group could be in RECONS state,
## NORMAL state or even DEGRADED -> RECONS state (under simulators). 
## The disk could be in optimized or ONLINING. Since optimize
## recons could happen very quickly, the routine shall check for
## the NORMAL state or even DEGRADED (which should occur briefly). 
## @arg $disk_name Required a valid disk name
## @returns 0 -- failed; 1 -- success
# -----------------------------------------------------------------
sub raidv2_disk_verify_opt_recons($) {
    my $disk = $_[0]; 

    if ($VERBOSE) {
      tstlog(-comment, "disk_verify_opt_recons: with disk <$disk>"); 
    }

    # -----------------------------------------------------------    
    # !!! DO NOT reserve the order of getting disk and group here
    # it would fail on simulator. TD 09/15/05
    # -----------------------------------------------------------
    raidv2_info_disk_load($disk);
    raidv2_info_group_load(); 

    # ------------------------------------------------------------------
    # if is very likely that the disk is in optimize reconstructed state
    # ------------------------------------------------------------------
    if ($raidv2_info_disk_by_name{$disk}{optimized_recons} == 1) {
	tstlog -comment, "disk <$disk> is being optimized"; 
	return $PASSED; 
    }

    if ($raidv2_info_disk_by_name{$disk}{fsm_state} =~ /ONLINING/) {
	tstlog -comment, "disk <$disk> is being onlined"; 
	return $PASSED; 
    }

    my $parent_id = $raidv2_info_disk_by_name{$disk}{parent_id};
    my $group_state = $raidv2_info_group_by_id{$parent_id}{fsm_state};
    tstlog -comment, "the raid group state is <$group_state>"; 

    # -----------------------------------------------
    # It could be in the DEGRADED state very briefly 
    # especially on simulators. So let get out of the
    # DEGRADED mode 
    # -----------------------------------------------
    my $cnt = 0; 
    my $max_cnt = 20; 
    while ($cnt++ < $max_cnt && $group_state eq "DEGRADED") {
	tstlog -comment, "raid group is in <$group_state> state"; 
	raidv2_info_group_load(); 
	$group_state = $raidv2_info_group_by_id{$parent_id}{fsm_state};
	sleep(1);
    }

    if ($cnt > $max_cnt) {
	tstlog -fail, "the raid group of disk <$disk> is in DEGRADED for over <$max_cnt>"; 
	return $FAILED; 
    }


    # -------------------------------------------------------------------
    # check for raid group RECONS state and the disk could be in one of the 
    # following situation: 
    #  1. the disk is being optimized so everything works 
    #  2. the disk is being ONLINING so optimize done and everything works
    #     (this should be seen a lot under simulator)
    #  3. If the disk does not belong to either 1 or 2 then this is a problem
    # --------------------------------------------------------------------
    if ($group_state eq "RECONS") {
	if ($raidv2_info_disk_by_name{$disk}{optimized_recons} == 1) {
	    tstlog -comment, "disk <$disk> has been optimizing"; 
	    return $PASSED; 
	} elsif ($raidv2_info_disk_by_name{$disk}{fsm_state} =~ /ONLINING/) {
	    tstlog -comment, "The disk <$disk> is in the transition state of ONLINING"; 
	    return $PASSED; 
	} else {
	    tstlog -fail, "The raid group was RECONS but the disk <$disk> was not in neither optimize nor ONLINING";
	    return $FAILED; 
	}
    }

    # --------------------------------------------------------------
    # if not RECONS it must be in NORMAL state otherwise there is a
    # problem.
    # --------------------------------------------------------------
    raidv2_info_group_load(); 
    $group_state = $raidv2_info_group_by_id{$parent_id}{fsm_state};
    if ($group_state ne "NORMAL") {
	tstlog -fail, "disk <$disk> vol state was <$group_state> which was not in the expected NORMAL state";
	return $FAILED; 
    } else {
	tstlog -comment, "the raid group of disk <$disk> was in <$group_state> state; probably optimize reconstruct took place very briefly"; 
    }
    return $PASSED; 
} #raidv2_disk_verify_opt_recons

# ----------------------------------------------------------
## @name ::raidv2_pause_hammer
## @summary pauses the hammer
## @description the routine will suspend all hammer.
##  It waits for the hammer status to go blank then returns 
##  a success (1). If the hammer failed to be suspended then 
##  returns a fail (0).
##  raidv2_restart_hammer is called to undo this. 
## @arg none
## @return 1 - success 
##         0 - fail 
# -----------------------------------------------------------
sub raidv2_pause_hammer {
    my($wait_interval) = 5;
    my($time_waited) = 0;
    my($max_wait) = 60; # wait 1 minute for hammer to pause

    # ------------------------------
    # only pause if hammer is active
    # ------------------------------
    my $hammer_status = saycmd_real($FILER, "priv -level -q diag; hammer status", $RAIDV2_TIMEOUT);    
    if ($hammer_status eq "") {
	tstlog(-comment, "no pause since there is no hammer activities."); 
	return FAILED; 
    }
    
    saycmd_real($FILER, "priv -level -q diag; hammer pause", $RAIDV2_TIMEOUT);
    my $hammer_status = saycmd_real($FILER, "priv -level -q diag; hammer status", $RAIDV2_TIMEOUT);
    
    while ($hammer_status ne "" && $time_waited < $max_wait) {
	$hammer_status = saycmd_real($FILER, "priv -level -q diag; hammer status", $RAIDV2_TIMEOUT);
	print("hammer status=<$hammer_status>"); 
	$time_waited += $wait_interval;
	sleep $wait_interval;
    }
    
    if ($time_waited >= $max_wait) {
	tstlog -fail, "Unable to pause hammer in <$max_wait> seconds";
	return $FAILED; 
    } else {
	tstlog -comment, "Successfully paused hammer";
    }
    return $PASSED; 
} #raidv2_pause_hammer

# -----------------------------------------------------------
## @name ::raidv2_start_hammer_size
## @summary the routine starts all hammers
## @description routine starts hammer activities against the
## specified volume based on the specified hammer size. 
## Params - 
##    volume or aggr name 
##    size of the hammer in block
## 
## @return none
# -----------------------------------------------------------
sub raidv2_start_hammer_size($$) {
    my $filervol = $_[0]; 
    my $blk_size = $_[1]; 
    my $iterations = -1; 
    
    my($sim) = raidv2_is_simulator(); 
    my($filename) = "/vol/$filvervol/testfile";
    
    my($o) = saycmd_real($FILER, "priv -level -q diag; hammer -f $iterations $filename $blk_size", $RAIDV2_TIMEOUT);
} 

# -----------------------------------------------------------
## @name ::raidv2_restart_hammer
## @summary the routine restarts all hammers
## @description routine restarts hammer activities across all
## volumes.  It is impossible to tell whether or not there is any
## suspend hammer or none. Therefore the routine will restart 
## and then wait for a few seconds to determine whether or not
## hammer was restarted successfully. 
## @return none
# -----------------------------------------------------------
sub raidv2_restart_hammer {
    my($wait_interval) = 4;
    my($time_waited) = 0;
    my($max_wait) = 20; # wait 1 minute for hammer to be restarted
    
    saycmd_real($FILER, "priv -level -q diag; hammer restart", $RAIDV2_TIMEOUT);
    my $hammer_status = saycmd_real($FILER, "priv -level -q diag; hammer status", $RAIDV2_TIMEOUT);
    
    # -----------------------------------------------------    
    # if there is no hammer then restarting will do nothing
    # -----------------------------------------------------
    while ($hammer_status eq "" && $time_waited < $max_wait) {
	my $hammer_status = saycmd_real($FILER, "priv -level -q diag; hammer status", $RAIDV2_TIMEOUT);
	print("hammer status of restart=<$hammer_status>\n");
	$time_waited += $wait_interval;
	sleep $wait_interval;
    }

    # ----------------------------------------    
    # the hammer could have not running at all.
    # ----------------------------------------
    if ($time_waited >= $max_wait) {
	tstlog -comment, "Unable to restart hammer in <$max_wait> seconds or there is no hammer activities to be restared."; 
    } else {
	tstlog -comment, "Successfully restarted hammer";
    }
} #raidv2_restart_hammer

# ----------------------------------------------------
## @name ::raidv2_pull_push_disks 
## @description pull and push the specified number of 
##  disk. For simulator only. 
## @synopsis 
##  raidv2_pull_push_disks(disk_cnt)
##  
## @arg 
##  disk_cnt Required - Number of disks to be pulled/pushed
## 
## @return - a list of pull/push disks
# ----------------------------------------------------
sub raidv2_pull_push_disks($) {
    my $disk_cnt = $_[0]; 

    my @list_disks; 
    # get a list of spares 
    # 
    my @spare_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool0"}); 
    my @spare1_disks = split(",", $raidv2_info_tree_disk_list{"POOL:Pool1"});
    push(@spare_disks, @spare1_disks);
    
    if ($#spare_disks < $disk_cnt) {
	tstlog -comment, "there is not enough spares to run"; 
	return ""; 
    }

    # pull out the number of spares
    for (my $i=0; $i<$disk_cnt; $i++) {
	raidv2_saycmd($FILER, "priv set -q diag; disk simpull $spare_disks[$i]"); 
	push(@list_disks, @spare_disks[$i]); 
    }

    # let see all of the failed disks
    raidv2_saycmd($FILER, "sysconfig -r"); 

    # push all back in 
    raidv2_disk_simpush_all(); 

    # let see all of the failed disks
    raidv2_saycmd($FILER, "sysconfig -r"); 
    
    return @list_disks; 
} #raidv2_pull_push_disks


# ------------------------------------------------------------
## @name ::raidv2_umount_root_vol 
## @summary routine umounts the root volume
## @description the routine umounts the filer root volume
## disregard whether or not the volume has been mounted. 
## @arg 
##  $filer_name name of the filer to be umounted its 
##  root volume. 
## @return none
# ------------------------------------------------------------
sub raidv2_umount_root_vol($)
{
    my ($filer_name) = $_[0];
    
    # - - - - - - - - - - - -
    # unmount the root volume
    # - - - - - - - - - - - -
    raidv2_client_saycmd($CLIENT, "/usr/software/bin/sudo umount -f /t/$filer_name/root"); #MTK_EDIT::19/06/08:r to SUDO
} #raidv2_umount_root_vol

# --------------------------------------------------------------
## @name ::raidv2_vol_is_degraded
## @summary routine returns success (1) if the volume is in DEGRADED state
## @description the routine checks and verifies whether or not the specified
## volume and its plexes are in the DEGRADED state.  
## If it is in the DEGRADED mode, returns 1 else return 0. 
## @arg - $vol - specify the volume name
# ---------------------------------------------------------------
sub raidv2_vol_is_degraded($) {
    my ($vol) = $_[0];

    # ------------------------------------------------
    # retrieve all plexes of the specified volume/aggr
    # ------------------------------------------------
    raidv2_info_load();

    # ----------------------------------------------------
    # get a list of raid_names from the specified vol/aggr
    # ----------------------------------------------------
    my @groups = split(',', $raidv2_info_tree_group_list{$vol});    

    # ---------------------------------------------------------------
    # verify each group list of the specified vol/aggr is in DEGRADED 
    # ---------------------------------------------------------------
    foreach my $group (@groups) {
	if ($raidv2_info_group_by_name{$group}{fsm_state} ne "DEGRADED") {
	    tstlog(-comment, "raid name <$group> state is <$raidv2_info_group_by_name{$group}{fsm_state}> not DEGRADED as expected."); 
	    return $FAILED; 
	}
	tstlog(-comment, "raid name <$group> is in DEGRADED state."); 
    }
    return $PASSED;
} #raidv2_vol_is_degraded

# --------------------------------------------------------------
## @name ::raidv2_vol_is_resync
## @summary waits for the specified vol to start a resync
## @description the routine checks and wait until a resync is
## started for the specified volume. 
##   Return $PASSED if resync is started. 
##   Return $FAILED is resync has not been started within 30 minutes
## 
## @arg - $vol - specify the volume name
# ---------------------------------------------------------------
sub raidv2_vol_is_resync($) {
    my ($vol) = $_[0];

    # ------------------------------------------------
    # retrieve all plexes of the specified volume/aggr
    # ------------------------------------------------
    raidv2_info_plex_load();
    my @plexes = split(',', $raidv2_info_tree_plex_list{$vol});    
    my $plex1 = $plexes[0]; 
    my $plex2 = $plexes[1]; 

    my $cnt =0; 
    my $max_cnt = 60; 
    my $sleep_time = 30; 
    while ($cnt++ <= $max_cnt) {
      if ($raidv2_info_plex_by_name{$plex1}{resyncing} eq "1" ||
	 $raidv2_info_plex_by_name{$plex2}{resyncing} eq "1") {
	return $PASSED; 
      }
      sleep(30); 
      raidv2_info_plex_load();
    }

    my $tot_time = $sleep_time * $max_cnt; 
    tstlog(-fail, "volume <$vol> resync did not occur within <$tot_time> seconds"); 
    return $FAILED;
} #raidv2_vol_is_resync


# --------------------------------------------------------------
## @name ::raidv2_vol_recons_wait
## @summary routine checks and waits for RECONS to be completed
## @description the routine checks and verifies whether or not the specified
## volume and its raid groups are in the RECONS or NORMAL state.
## The routine shall wait until RECONS is completed. The routine
## will return success (1) when the state goes to normal. 
## @arg $vol - name of the volume
## @returns 1 if RECONS completed in time or NORMAL is the current state. 
## @returns 0 if the state is not RECONS nor NORMAL. 
# ---------------------------------------------------------------
sub raidv2_vol_recons_wait($) {
    my ($vol) = $_[0]; 

    # ------------------------------------------------
    # retrieve all plexes of the specified volume/aggr
    # ------------------------------------------------
    raidv2_info_plex_load();

    # ----------------------------------------------------
    # get a list of raid_names from the specified vol/aggr
    # ----------------------------------------------------
    my @groups = split(',', $raidv2_info_tree_group_list{$vol});    

    # -------------------------------------------------------------
    # verify each group list of the specified vol/aggr is in NORMAL 
    # -------------------------------------------------------------
    my $normal_cnt = -1; 
    foreach my $group (@groups) {
	if ($raidv2_info_group_by_name{$group}{fsm_state} ne "NORMAL") {
	    $normal_cnt++; 
	}
    }

    # -------------------------------------    
    # done if all raid groups are in NORMAL
    # -------------------------------------
    if ($normal_cnt == $#groups) {
	return $PASSED; 
    }

    # ---------------------------------------------------------------	
    # wait until RECONS is completed
    # i believe recons should start on both raid group about the same
    # time therefore it does not matter much which group should be 
    # verified before the other. The wait time here is the total 
    # wait time for all raid group. 
    # ---------------------------------------------------------------
    my $cnt = 0; 
    my $max_cnt = 50; 
    my $simulator = raidv2_is_simulator(); 

    while ($cnt++ < $max_cnt) {
	raidv2_info_group_load(); 
	foreach my $group (@groups) {
	    if ($raidv2_info_group_by_name{$group}{fsm_state} eq "RECONS") {
		if ($simulator) {
		    sleep(5); 
		} else {
		    sleep(10); #sleep longer for filers 
		}
	    }
	} #foreach 
    } #while 
    return $PASSED;
} #raidv2_vol_recons_wait

# --------------------------------------------------------------
## @name ::raidv2_vol_verify
## @summary routine to run vol verification
## @description the routine runs "vol verify" on the specified 
## volume and looks for any mismatch or errors. The volume must
## be online and must have mirror. 
## The routine checks the log messages to determine when vol
## verification is started or ended; it looks for any 
## mismatch messages which indicating there are issues;  
## it also checks for other types of abnormal messages 
## such as abort, suspend, and resume which indicating
## some sort of problems. 
## 
## The routine returns a FAILED if there is a mismatch message
## or any of the abnormal messages. 
## The option -n should be invoked to preserve any mismatch content. 
##
## @arg $vol - specify the volume or aggregate needs to be verified
## @arg $opt - specify either blank or '-n'. When option -n is 
## specified, all verification information shall be retained.    
## 
## Mar 12, 2009 - Replaced NFS mount with LogReader since the CLIENT
##                NFS mount has caused some problems. Added code to
##                check for PI, media, and check sum error. Tom Doan
## ---------------------------------------------------------------
sub raidv2_vol_verify($;$) {
    my ($vol, $opt) = @_; 
    my $max_cnt = 50; 
    my $cnt, $output; 
    my($cmd) = ($AGGR|$VVOLS) ? "aggr" : "vol"; 

    my $rcon = connect("#rsh"); 
    my $f = LogReader->new($rcon); 

    # Setup up
    $f->start(['messages']); 

    my $rsh = connect("#rsh"); 
    my $logreader = LogReader::Messages->new($rsh);
    $LogReader::Messages::Path = '/etc/messages'; 
    $logreader->start; 

    # ------------------------------------------------------------
    # verify that the vol/aggr is in NORMAL state and mirror is on
    # ------------------------------------------------------------
    $cnt = 0; 
    while ($cnt++ < $max_cnt) {
	raidv2_info_vol_load($vol);
	
	if ($raidv2_info_vol_by_name{$vol}{fsm_state} eq "NORMAL" && 
	    $raidv2_info_vol_by_name{$vol}{mirror_state_name} eq "mirrored") {
	    tstlog(-comment, "<$vol> is in normal state"); 
	    last; 
	} 
	sleep(5); 
    } 

    if ($raidv2_info_vol_by_name{$vol}{fsm_state} ne "NORMAL" || 
	$raidv2_info_vol_by_name{$vol}{mirror_state_name} ne "mirrored") {
	tstlog(-fail, "<$vol> is NOT NORMAL state for starting mirror verifying"); 
	return $FAILED; 
    }
    
    # --------------------------------------------------------
    # pause the hammer so can run vol/aggr mirror verification
    # --------------------------------------------------------
    raidv2_pause_hammer($vol); 

    # ---------------------------
    # start vol/aggr verification 
    # ---------------------------
    $output = saycmd_real($FILER, "priv set -q diag; $cmd verify start $vol $opt", $RAIDV2_TIMEOUT); 
    # ------------------------------------
    # looking for the mirror start message
    # ------------------------------------
    #my $start_msg = "raid.mirror.verify.start:notice.*/$vol"; 
    my $start_msg = "$vol: created mirror verification"; 
 
    $cnt = 0;
    $max_cnt = 5;
    # --------------------------------------------------
    # wait until vol/aggr verify is done or be suspended
    # --------------------------------------------------
    while ($cnt++ < $max_cnt) {
        $output = saycmd_real($FILER, "priv set -q diag; $cmd verify status", $RAIDV2_TIMEOUT);
        if ($output =~ /no verifies in progress/ || $output =~ /suspended/) {
            tstlog(-comment, "Found $cmd verify is done on <$vol>"); 
            last; 
	}
	sleep(10); 
      }
    # ----------------------------------------------------    
    # looking for verification ended or any error messages
    # ----------------------------------------------------
    my $done_msg =     "raid.mirror.verify.done:notice.*/$vol"; 
    my $mismatch_msg = "raid.mirror.verify.mismatch.*/$vol"; 
    my $unexp_msg =     "raid.mirror.verify.aborted.*/$vol|raid.mirror.verify.suspend.*/$vol|raid.mirror.verify.resume.*/$vol"; 

    $logreader->stop(); 
    $buf = $logreader->to_string; 

    print "Output of the logreader"; 
    print $buf; 

    if ($buf =~ /$start_msg/) {
        tstlog(-comment, "Found $cmd verify start message"); 
    } else {
        tstlog(-fail, "Could not find $cmd verify start message"); 
    }

    if ($buf =~ /$done_msg/) {
        tstlog(-comment, "Found $cmd verify done message"); 
    } else {
        tstlog(-fail, "Could not find $cmd verify stop message"); 
    }

    # --------------------------------------------------
    # Looking for any potential verify problem
    # --------------------------------------------------
    if ($buf =~ /$mismatch_msg/) {
        tstlog(-fail, "Found a mismatch message $mismatch_msg"); 
    } else {
        tstlog(-comment, "Found no mismatch message"); 
    }
        
    if ($buf =~ /$unexp_msg/) {
        tstlog(-fail, "Found an unexpected message $unexp_msg"); 
    } else {
        tstlog(-comment, "Found no unexpected message"); 
    }

    # -----------------------------------------------
    # Looking for PI, media error, and checksum error
    # -----------------------------------------------
    tstlog(-comment, "Checking for PI, media error, and checksum problem"); 

    raidv2_info_group_load();
    my @groups = split(',', $raidv2_info_tree_group_list{$vol});

    foreach $group (@groups) {
	if ($raidv2_info_group_by_name{$group}{scrub_pi} != 0) {
	    tstlog -fail, "scrub ERROR found pi in <$group>";
	} else {
	    tstlog -comment, "found NO PI error in group <$group>"; 
	}
	
	if ($raidv2_info_group_by_name{$group}{scrub_cksum_err} != 0) {
	    tstlog -fail, "scrub ERROR found checksum error in <$group>";
	} else {
	    tstlog -comment, "found NO checksum error in group <$group>"; 
	}
    } #foreach 
    
    return $PASSED;
} #raidv2_vol_verify


# ============================ END =============================
1






