package rsh;

## @summary  Wrapper around 'say' over rsh connection
## @author dl-systemtest@netapp.com
## @status private (for torture/crusher use only)
## @pod here

use strict;
use vars ('@ISA','@EXPORT_OK');
use English qw();

require Exporter;
@ISA = ('Exporter');
@EXPORT_OK = ('rsh','ssh');

# This is a wrapper around 'say' over an rsh connection, with
# enhancements that make it useful for the torture tests:

# - one can pass in a host object instead of a connection object.

# - rsh requests are retried on error unless '-noretry' is passed in.
#   (such errors are probably timeouts, but for better or worse we
#   retry for other kinds of errors as well)

# - even with -noretry, rsh requests are retried when we think the
#   filer didn't even get a chance to consider the request, probably
#   because the filer says there are too many rshs being served
#   already ("Try again") or because rsh indicates the filer isn't
#   even beginning to answer ("Connection timed out" or "No route to
#   host").

# - The output of the command is returned when there is no error, the
#   empty string is returned when there is no output but no error
#   occurred, and the undefined value (undef) is returned when there's
#   an error.  say by itself doesn't make it easy for its caller to
#   distinguish the last two cases (one must use a complainspec)

use constant    RSH_ERRORS  => join('|',
                                    'Try again',
                                    'Connection timed out',
                                    'Connection refused',
                                    'No route to host',
                                    'All ports in use',
                                    '^try again later$',
                                );
sub rsh ( $$@ ) {
    my($conn, $cmd, @options) = @_;
    my @sayoptions;
    my $retry=1;
    my $conntype = 'rsh';
    my $saved_conn;
    my $opt;
    while ($opt = shift @options) {
	    if ($opt eq '-timeout') {
	        my $optarg = shift @options;
	        push @sayoptions, ('-timeout', $optarg);
	    } elsif ($opt eq '-retry') {
	        $retry=1;
	    } elsif ($opt eq '-noretry') {
	        $retry=0;
	    } elsif ($opt eq '-ssh') {
            $conntype = 'ssh';
        } elsif ($opt eq '-noprompt') {
            push @sayoptions, $opt;
        } elsif ($opt eq '-prompt') {
            my $optarg = shift @options;
            push @sayoptions, ('-prompt', $optarg);
        }
        else {
	        &Tharn::tstlog('-script',"Unknown cluster::rsh option \"$opt\"");
	    }
    }
    $saved_conn = $conn;
    if (ref($conn) =~ /Connectrec/) {
        # $conn is already a good connection to use
        
    } elsif (ref($conn) =~ /Hostrec/) {
        $conn = Tharn::connect(Tharn::host($conn)->{id} . "#$conntype");
        if( (!defined $conn) && ($conntype =~ /ssh/) )
        { $conn = ssh_reconnect(Tharn::host($saved_conn)->{id} . "#$conntype") }
    } else {
        $conn = Tharn::connect($conn . "#$conntype");
        if( (!defined $conn) && ($conntype =~ /ssh/) )
        { $conn = ssh_reconnect($saved_conn . "#$conntype") }
    }
    
    my ($after,$complained);
    my $id = $conn->hostp()->id();

    while (1) {
	    $complained=0;
	    (undef, undef, $after) = Tharn::say($conn, $cmd, @sayoptions, '-eval_error');

        # If we evaled above or got any one of RSH_ERRORS
        # set $complained and retry in 10 secs only if asked to.
        # abir: Should we have an SSH_ERRORS?
        if($English::EVAL_ERROR ||
                Tharn::matchstrings($after, RSH_ERRORS)) {
            my $msg = $English::EVAL_ERROR || $after;
            chomp($msg);
            Tharn::tstlog('-commline',"$conntype failed: $msg");
            $complained = 1;
            last if(!$retry);
            if($conntype =~ /ssh/) { $conn = ssh_reconnect( $conn->{id} ); next; }
            sleep(30);
	    }
        
        # We are a success otherwise.
        else {
            last;
        }
    }

    if (!$complained && $::rshcounter) {
	    # bump up global counter of successful rsh requests
        if(defined($::rshcounter->can("log"))) {
            $::rshcounter->log(level => 'info', message => "SUCCESS:$id\n");
        } else {
            $::rshcounter->inc();
        }
    }
    return ($complained) ? undef : ($after ? $after : 1);
}

# ssh is rsh except for the name.
# While using ssh be sure to pass in the 
# connection object in ur tests. 
sub ssh { push( @_, '-ssh' ); &rsh; };

# Reestablish ssh connection.
# A must have for 10 mode reboot testing.

sub ssh_reconnect
{
  my $connspec = shift;
  my ($try, $conn);

  while(sleep(60), ++$try)
  {
    Tharn::disconnect("$connspec", '-ok'); # This line is a must if you hope to ever reconnect
    $conn = Tharn::connect("$connspec", '-eval_error');
    if(defined $conn)
    {
      Tharn::tstlog('-commline',"Connect attempt $try succeeded!\n\n");
#      Tharn::say($conn, "ngsh", '-eval_error');
      last; 
    }
    else 
    { Tharn::tstlog('-commline',"Connect attempt $try failed.\n\tRetry in 60 secs. $English::EVAL_ERROR\n") }
  }
  return $conn;

}

use IO::Socket;
use IO::Select;

use constant READ_SIZE => 16384;

sub do_one_rsh ( $$$$$ ) {
    my ($ipaddr,$username,$password,$cmd,$timeout) = @_;
    my $out='';
    return eval {
	my ($msg, $rshstatus, $error);

	my $inout  = new IO::Socket::INET(PeerAddr => $ipaddr,
					  PeerPort => 514,
					  Proto => 'tcp',
					  Timeout => $timeout)
	    or die "Unable to create rsh i/o socket: $@\n";

	my $errlisten = new IO::Socket::INET(Listen => 1,
					     Proto => 'tcp',
					     Timeout => $timeout)
	    or die "Unable to create rsh err listen socket: $@\n";

	# send our stderr port number
	$msg = $errlisten->sockport() . "\0";
	syswrite($inout,$msg,length($msg)) or die "Writing rsh port info: $!\n";

	# accept rshd's connection to that stderr port
	my $errsock = $errlisten->accept() or die "Accepting rsh err port: $!\n";
	$errsock->timeout($timeout);

	# send local user name, remote user name, command
	$msg = "root" . "\0" . "$username:$password" . "\0" . $cmd . "\0";
	syswrite($inout,$msg,length($msg)) or die "Writing rsh user info: $!\n";

	# read rsh status byte (decides whether we return the rest of
	# the output as an error or normally)
	sysread($inout,$rshstatus,1) or die "Reading status byte: $!\n";

	# read rsh command stdout/stderr or error
	my $sel = new IO::Select;
	$sel->add($inout);
	$sel->add($errsock);

	while ($sel->handles > 0) {
	    my @avail = $sel->can_read($timeout);
	    if (!@avail) {
		die "rsh timed out\n";
	    }
	    my $h;
	    foreach $h (@avail) {
		my $name = ($h eq $inout) ? "stdout" : "stderr";
		my $res = sysread($h,$out,READ_SIZE,length($out));
		if (!defined($res)) {
		    die "error reading rsh $name: $!\n";
		} elsif ($res == 0) {
		    # END of file
		    $sel->remove($h);
		}
	    }
	}
	if ($rshstatus ne "\0") {
	    die "rsh server returned error: $out";
	}
	return $out;
    };
}


sub rsh_old ( $$@ ) {
    my($spec, $cmd, @options) = @_;

    my $retry=1;
    my $opt;
    my $timeout=60;
    while ($opt = shift @options) {
	if ($opt eq '-timeout') {
	    $timeout = shift @options;
	} elsif ($opt eq '-retry') {
	    $retry=1;
	} elsif ($opt eq '-noretry') {
	    $retry=0;
	} else {
	  Tharn::tstlog('-script',"Unknown lib::rsh option \"$opt\"");
	}
    }

    # make the timeout unambiguously relative
    $timeout=Tharn::timeout2time($timeout) - time();

    my $id;
    my $ipaddr;
    my $username;
    my $password;
    if (ref($spec) =~ /Hostrec/) {
	$id=$spec->{id};
	$ipaddr=$spec->{default_ip};
	$username="root";
	$password=$spec->{default_password};
    } else {
	if (ref($spec) !~ /Connectrec/) {
	    $spec=Tharn::connect($spec);
	}
	$id=$spec->{hostp}->{id};
	$ipaddr=$spec->{connhost};
	$username=$spec->{username};
	$password=$spec->{password};
    }
    
    while (1) {
    Tharn::tstlog('-commline',">>$id#rsh\|'$cmd'" . "\\n");
	my $out = do_one_rsh($ipaddr,$username,$password,$cmd,$timeout);
	if (!defined($out)) {
	    my $msg = "..$id#rsh: $@";
	    chomp $msg;
	    &Tharn::tstlog('-commline',$msg);
	} else {
	    my @outlines = split(/\n/,$out);
	    @outlines=('') if (!@outlines);
	    # (we have to put the result in [] because raw
	    # complainspecs are arbitrarily limited to 999 emax in
	    # _parsecmds())
	    &Tharn::tstlog('-commline',
			   [map { 
			       "<<$id#rsh\|" . Tharn::_string_cookctl_nl( $_ ) . "\\n"
			   } @outlines]); 
	}

	if (!defined($out)) {
	    # some kind of problem (probably timeout)
	    return undef if (!$retry);
	    sleep(10);
	} elsif ($out =~ /^.*Try again/) {
	    # retry immediately
	} elsif ($out =~ (/^.*(Connection timed out|Connection refused|No route to host|All ports in use|try again later)/)) {
	    # retry in N seconds
	    sleep(30);
	} else {
	    # success
	    if ($::rshcounter) {
		# bump up global counter of successful rsh requests
		$::rshcounter->inc();
	    }
	    return $out;
	}
    }
}

1;
