# Hey emacs, this is basically -*- perl -*-
#
#Copyright (c) 2004 Network Appliance, Inc.
#All rights reserved.
#
## @summary  Snaplock Compliance Clock library
##
## @description
##
## $Id: //depot/prod/test/Rsierranevada/lib/snaplock.pm#5 $
##
## These test functions use various commands to set and test the compliance clock
##
##
## @usage
##   use snaplock;
##   most routines are passed a  valid connection $conn which
##   is created using $conn=FS:Util:Setup::create_vol_and_mount()
##
## @keywords snaplock compliance clock
## @dependencies
##
## @author dl-johny (FSS WAFL QA)
##
## @status public
##
package snaplock;

use POSIX ('uname');
use strict;
use vars ('@ISA','@EXPORT_OK');
use Data::Dumper;

BEGIN {
    require Exporter;
    @ISA = ('Exporter');
    @EXPORT_OK = qw(
             exec_zapi
             get_vol_expiry_date
             snaplock_license
             snaplock_unlicense
             snaplock_enterprise_license
             compliance_clock_initialize
             compliance_clock_ensure_initialized
             volume_expiry_date
             compliance_clock_extern_setup_warning
             dates_are_close
             dates_delta_in_secs
             system_time_get
             system_time_set
             test_java_enabled
             compliance_clock_licenses_store
             compliance_clock_licenses_restore
             timed_options_store
             timed_options_restore
             set_default_retention_period
             set_retention_periods
             set_retention_or_die
             compliance_clock_status
             SNAPLOCK_UNLICENSED_EXCEPTION
             COMPLIANCE_CLOCK_ALREADY_SET
             snaplock_remove_file
             snaplock_remove_dir
             RETENTION_TOO_SHORT
             wait_for_expiry
             wait_file_expiry
             retention_expired
             parse_retention
             client_set_atime
             client_get_atime
             client_get_ctime
             client_commit_file
             snaplock_scan_start
             snaplock_scan_wait
             snaplock_scan_status
             snaplock_get_compliance_clock
             file_set_snaplock_retention_time
             file_get_snaplock_retention_time
             file_get_snaplock_retention_time_list_info_max
             compliance_clock_status_verbose
             corrupt_volinfo_secure_clock_metadata
             check_worm_flag_secure_clock
             snaplock_get_system_compliance_clock
			 snaplock_clock_status_verbose
             );
}

use Carp qw(croak);
use Tharn;
use Admin::License qw(license unlicense);
use SPARSE::Util::Logger;
use Date::Manip;
use Filer::Version qw(
    get_wafl_version
    UNSUPPORTED_FEATURE_EXCEPTION
    WV_SCRIMSHAW_MAX
);
use Volume qw(vol_options volume_options_list_info);
use FS::Util::Utilities qw(uSystem util_chmod);
use Utils qw( make_conn );
use TestEnv::MinWin qw(:all);
use CIFS::Setup qw(cifs_setup);
use ExportFS qw(share_add);
use MapShare qw(map_cifs_share);
use SayWrap;
use Zapi;
use ZapiErrno;
use NaElement;
use XML::Simple;
use Params::Validate qw(validate BOOLEAN SCALAR HASHREF ARRAYREF SCALARREF CODEREF OBJECT UNDEF);
use Filer::Version qw(get_wafl_version WV_ROLLINGROCK_MIN WV_BOILERQAN_MIN);
use Wreck;
use Volume qw(vol_online);
use FS::Util::Utilities qw(dumpblock);

use constant SNAPLOCK_UNLICENSED_EXCEPTION => 'Snaplock licensing is not installed';
use constant COMPLIANCE_CLOCK_ALREADY_SET => "The compliance clock has already been initialized and 'strict' is set";

use constant SECONDS_PER_DAY   => 86400;
use constant SECONDS_PER_MONTH => int(SECONDS_PER_DAY * (365.2422/12)); # roughly
use constant SECONDS_PER_YEAR  => int(SECONDS_PER_DAY * 365.2422); # close enough

use constant MAX_WAIT_SECONDS => 600; # seconds


# the time it takes to initialize and break the snapmirror is not known in
# advance - however, for this test to work, the retention time must be longer
# than the snapmirroring time - this error occurs if the retention is too short
use constant RETENTION_TOO_SHORT => 'The WORM_RETENTION time was too short to complete the test - please specify a longer retention period and try again.';


my %say_hash = (
    # the clock is already inited, return the time
    'Compliance Clock: '
        => sub { $_[0]; },
    # RR output looks like this
    'Volume Compliance Clock'
        => sub {$_[0]; },
    # scrimshaw doesn't support this command - die
    'date: Invalid date specified.'
        => sub { $_[0]; },
# even scrimshaw needs to run... all callers seem to catch it ok...
#           # looks like scrimshaw
#           die na_result(UNSUPPORTED_FEATURE_EXCEPTION);
#       },
    # die if snaplock is not licensed
    'date: The -c option requires a snaplock or snaplock_enterprise license'
        => sub { die SNAPLOCK_UNLICENSED_EXCEPTION; },
    ## Required to initialize
    'Is the current local system time correct?',
        => 'y' ,

);

## @name ::snaplock_license
##
## @description
##  Give the filer a snaplock license
##
## @synopsis
##  snaplock_license ($conn)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @example
##  snaplock_license($conn);
##

sub snaplock_license($)
{
        my ($conn) = @_;
        license($conn,"snaplock")
                or logError("Can not set snaplock license");
        my ($match, undef, $after) = say($conn, "license", "SnapLock");
}

## @name ::snaplock_unlicense
##
## @description
##  Take from the filer all snaplock licenses
##
## @synopsis
##  snaplock_unlicense ($conn)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @example
##  snaplock_unlicense($conn);
##

sub snaplock_unlicense($)
{
        my ($conn) = @_;
        my ($match, undef, $after) =
                say($conn, "license");
        my ($snaplock_licensed, $snaplock_enterprise_licensed) = 0;
        if ($after =~ /SnapLock/) {
                $snaplock_licensed=1;
        } else {
                $snaplock_licensed=0;
        }
        if ($after =~ /SnapLock_Enterprise/) {
                $snaplock_enterprise_licensed=1;
        } else {
                $snaplock_enterprise_licensed=0;
        }
        if ($snaplock_licensed==1) {
                unlicense($conn,"snaplock")
                        or logError("Can not remove snaplock license");
        }
        ($match, undef, $after) =
                say($conn, "license", qr/snaplock/i,'-pass',"In SN onwards if there is no license, then the product name won't be displayed in license");
        if ($snaplock_enterprise_licensed==1) {
                unlicense($conn,"snaplock_enterprise")
                      or logError("Can not remove snaplock enterprise license");
        }
        ($match, undef, $after) =
                say($conn, "license", qr/SnapLock_Enterprise/i,'-pass',"In SN onwards if there is no license, then the product name won't be displayed in license");
}


## @name ::snaplock_enterprise_license
##
## @description
##  Give the filer an enterprise snaplock license
##
## @synopsis
##  snaplock_enterprise_license ($conn)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @example
##  snaplock_enterprise_license($conn);
##

sub snaplock_enterprise_license($)
{
        my ($conn) = @_;
        license($conn,"snaplock_enterprise")
                or logError("Can not set snaplock enterprise license");
        my ($match, undef, $after) =
                say($conn, "license", "SnapLock_Enterprise");
}


## @name ::snaplock_no_license
##
## @description
##  Check that the filer has no snaplock license i
##  (neither snaplock or snaplock enterprise)
##
## @synopsis
##  snaplock_no_license ($conn)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @example
##  snaplock_no_license($conn);
##

sub snaplock_no_license($)
{
        my ($conn) = @_;
        my ($match, undef, $after) =
                say($conn, "license", qr/snaplock/i,'-pass',"In SN onwards if there is no license, then the product name won't be displayed in license");
        ($match, undef, $after) =
                say($conn, "license", qr/snaplock/i,'-pass',"In SN onwards if there is no license, then the product name won't be displayed in license");
}


## @name ::compliance_clock_initialize
##
## @description
##  Initialize the compliance clock. Returns the compliance clock time and a
##  flag indicating whether the clock was actually initialized.
##
## @synopsis
##  ($cc_time, $initialized) = compliance_clock_initialize($conn, strict => 1)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg strict Optional: If true, the call will die if the compliance clock
## cannot be set. Otherwise, this condition just issues a warning and returns
## undef.
##

sub compliance_clock_initialize {
    my ($conn, %opts) = @_;

    # strict can be set in %opts, or as $main::SLCC_STRICT. otherwise, it
    # defaults to 1/true
    my $strict = (exists $opts{strict}) ? $opts{strict} : (defined($main::SLCC_STRICT)) ? $main::SLCC_STRICT : 1;

    # reinit_ok can be set in %opts, or as $main::REINIT_OK. Otherwise, it
    # defaults to 0/false
    my $reinit_ok = (exists $opts{reinit_ok}) ? $opts{reinit_ok} : (defined($main::REINIT_OK)) ? $main::REINIT_OK : 0;

    logTrc("get the version and decide what to do...");
    my $ver = get_wafl_version($conn->{hostp});

    logDbg("ver is '$ver', strict is '$strict'");

    if ($ver <= WV_SCRIMSHAW_MAX) {
        my $message = "current version of OnTap does not support setting the compliance clock.";
        if ($strict) {
            logTrc("strict is set and the " . $message);
            die na_result(UNSUPPORTED_FEATURE_EXCEPTION);
        }
        # just warn if not strict
        logWarn(ucfirst($message));
    } else {
        my $did_init = 0;

        # see if the clock is already set...
        my $time = compliance_clock_status($conn);

        if (defined($time) and $time =~ /^\d+/) {
            die COMPLIANCE_CLOCK_ALREADY_SET unless ($reinit_ok);
        } else {
            # not initialized yet, do it
            if ( $ver < WV_ROLLINGROCK_MIN ) {
                $time = _do_say($conn, 'priv set -q diag; date -f -c initialize; priv set', \%say_hash);
                $did_init = 1;
            } else {
                my (undef , undef , $time_rr)  = say($conn, 'priv set -q diag; snaplock clock initialize -f');
                if ( !$time_rr ) {
                        ## Verify compliance clock is initialized by executing "snaplock clock status command"
                        my $new_time = compliance_clock_status($conn);
                        if (defined($new_time) and $new_time =~ /^\d+/) {
                            $did_init = 1;
                        }
                } else {
                        die "Failed to initialize compliance clock";
                }
            }
        }
        return ($time, $did_init);
    }
    return undef;
}

## @name ::compliance_clock_extern_setup_warning
##
## @description
##  Warn that the compliance clock cannot be uninitialzed
##
## @synopsis
##  compliance_clock_extern_setup_warning ()
##
## @example
##  compliance_clock_extern_setup_warning();
##

sub  compliance_clock_extern_setup_warning()
{
        logStep("This test can not clean up after itself. The compliance clock"
                ." can NOT be reinitialized easily. To do this you need to "
                ." reboot filer and re-zero disks. To help minimize this please"
                ." see the param SLCC_SETONCEONLY in clock_reinit.thpl");
}


## @name ::dates_are_close
##
## @description
##  Compare two times
##
## @synopsis
##  dates_are_close (date1, date2, closesecs, simdbg,
#                               expectedrealresult)
##
## @example
##  dates_are_close($date1, $date2, $closecs, $SLCC_SIMDBG,
##                                      "EXPECTTRUE");
##

sub  dates_are_close($$$$$)
{
        my ($date1, $date2, $closesecs, $simdbg, $expectedrealresult) = @_;
        my $delta = &DateCalc($date1, $date2);
        logDbg("delta is $delta");
        my $deltasecs = &Delta_Format($delta,undef,"%st");
        logDbg("delta in seconds is $deltasecs");
        if (abs($deltasecs) < $closesecs) {   #closesecs defaults to 5 mins
                if ( $expectedrealresult eq "EXPECTTRUE") {
                        logDbg("dates are close enough, $date1 $date2");
                } else {
                        if ($simdbg ne "TRUE") {
                           logWarn("dates are too close, $date1 $date2");
                        } else {
                           logDbg("SIMDBGLIE dates close enough,$date1 $date2");
                        }
                }
                return "true";
        } else {
                if ( $expectedrealresult eq "EXPECTFALSE") {
                        logDbg("dates are far apart, $date1 $date2");
                } else {
                        if ($simdbg ne "TRUE") {
                           logWarn("dates are too far apart, $date1 $date2");
                        } else {
                           logDbg("SIMDBGLIE dates far apart, $date1 $date2");
                        }
                }
                return "false";
        }
}

## @name ::dates_delta_in_secs
##
## @description
##  get the difference between two times in seconds
##
## @synopsis
##  dates_delta_in_secs (date1, date2)
##
## @example
##  dates_delta_in_secs($date1, $date2)
##

sub  dates_delta_in_secs($$)
{
        my ($date1, $date2) = @_;
        my $delta = &DateCalc($date1, $date2);
        logDbg("delta is $delta");
        my $deltasecs = &Delta_Format($delta,undef,"%st");
        logDbg("delta in seconds is $deltasecs");
        return $deltasecs;
}

## @name ::system_time_get
##
## @description
##  Get the system time
##
## @synopsis
##  system_time_get ($)
##
## @example
##  system_time_get($conn);
##

sub  system_time_get($)
{
        my ($conn) = @_;
        my ($match, undef, $sysdate) = say($conn, "date");
        return $sysdate;
}


## @name ::system_time_set
##
## @description
##  Set the system time
##
## @synopsis
##  system_time_set ($$$)
##
## @example
##  system_time_set($conn,$newtime,$simdbg);
##

sub  system_time_set($$$)
{
        my ($conn, $settodate, $simdbg) = @_;
        logDbg("about to set system time, settodate, to $settodate");
        if ($simdbg eq "TRUE") {
                my ($match, undef, $after) =
                say($conn, "date $settodate","Simulator\.\*change that time");
                logDbg("system time NOT changed to $after");
        } else {
                my ($match, undef, $after) =
                        say($conn, "date $settodate");
                logDbg("system time set to $after");
        }
}

## @name ::compliance_clock_licenses_store
##
## @description
##  Store the current values for the snaplock licenses
##
## @synopsis
##  compliance_clock_licenses_store ($$)
##
## @example
##  compliance_clock_licenses_store($$);
##

sub  compliance_clock_licenses_store($$)
{
        my ($conn,$slcclicensesref) = @_;
        my ($match, undef, $after) =
                say($conn, "license");
        if ($after =~ /SnapLock/) {
                ${$slcclicensesref}{'snaplock'}=1;
        } else {
                ${$slcclicensesref}{'snaplock'}=0;
        }
        if ($after =~ /SnapLock_Enterprise/) {
                ${$slcclicensesref}{'snaplock_enterprise'}=1;
        } else {
                ${$slcclicensesref}{'snaplock_enterprise'}=0;
        }
}


## @name ::compliance_clock_licenses_restore
##
## @description
##  Restore the values for the snaplock licenses
##
## @synopsis
##  compliance_clock_licenses_restore ($$)
##
## @example
##  compliance_clock_licenses_restore($$);
##

sub  compliance_clock_licenses_restore($$)
{
        my ($conn,$slcclicensesref) = @_;
        #use Data::Dumper;
        #print &Dumper($slcclicensesref);
        if (${$slcclicensesref}{'snaplock'} == 1) {
                snaplock_license($conn)
        }
        if (${$slcclicensesref}{'snaplock_enterprise'} == 1) {
                snaplock_enterprise_license($conn)
        }
}

## @name ::timed_options_store
##
## @description
##  Store the timed options
##
## @synopsis
##  timed_options_store ($$)
##
## @example
##  timed_options_store($$);
##

sub  timed_options_store($$)
{
        my ($conn,$timedoptionsref) = @_;

        my ($match, undef, $after) =
                say($conn, "options timed");
        my @lines = split(/\n/,$after);
        my $line="";
        foreach $line (@lines) {
                my ($option, $setting) = split (/\s+/,$line);
                ${$timedoptionsref}{$option}=$setting;
        }
}

## @name ::timed_options_restore
##
## @description
##  Restore the timed options
##
## @synopsis
##  timed_options_restore ($$)
##
## @example
##  timed_options_restore($$);
##

sub  timed_options_restore($$)
{
        my ($conn, $timedoptionsref) = @_;
        my ($option, $setting) = "";
        #use Data::Dumper;
        #print &Dumper($timedoptionsref);
        foreach $option (keys %{$timedoptionsref}) {
                $setting = ${$timedoptionsref}{$option};
                $setting = "\' \'" if $setting eq "";
                my ($match, undef, $after) =
                        say($conn, "options $option $setting","-noprompt");
        }
}

## @name ::test_java_enabled
##
## @description
##
## @synopsis
##  test_java_enabled ($)
##
## @example
##  test_java_enabled($);
##

sub test_java_enabled
{
        my ($conn) = @_;
        my %optionsorig = ();
        timed_options_store($conn, \%optionsorig);
        # first reset the timed.servers  so that if we do manage to enable
        # timed then the system clock won't get changed too early in the test
        my ($match, undef, $after) =
                say($conn, "priv set advanced; options timed.servers \'\'");
        logDbg("turn_timed_off set server to \'\' ");

        # if already on reply is: timed is already on
        # if on the message is timed: time daemon started (to stderr only)
        # if no java present the message is:  timed requires java which is not
        #       enabled on this system
        #
        # NB we have to be careful because say does not catch stderr such as
        # Thu Apr 29 20:10:26 GMT [rshd_0:ALERT]: timed: time daemon started
        # to guard against this we issues some commands twice!!
        # places in code are labeled *TWICE*
        # the sayWAIT wrapper did  not appear to solve the problem
        # *TWICE*
        ($match, undef, $after) =
                say($conn, "priv set advanced; options timed.enable on");
        #($match, undef, $after) =
        #        say($conn, "priv set advanced; options timed.enable on","timed");
        logDbg("test_java_enabled match is $match after is $after");
        timed_options_restore($conn, \%optionsorig);
        if ($after =~ /requires java/ ) {
                logError("A full software install (PCdist.zip or similar) needs"
                        ." to be performed - see snaplock.pm. This test "
                        ." REQUIRES it. We test here by  checking that java is"
                        ." present.");
                # Steps are: 1. bring up simulator or filesystem
                #               e.g. (as root) fgo -f sim maytag.Ld
                #            2. add aggr or volumes
                #               e.g. aggr add vol0 2
                #            3. install software  (takes about 15 minutes)
                #               e.g. software install http://web.netapp.com/
                #               engineering/rlse/Rancorsteam/releases/pc/dist
                #               /PCdist.zip
                #            4. reboot simulator or filer
        }
}

sub set_default_retention_period
{
    my ($conn, $period, $target_vol) = @_;

    return set_retention_periods($conn, $target_vol, default=>$period);
}

sub set_retention_period_to_default
{

    my ($conn, $target_vol) = @_;
    my $volopts = vol_options($conn, $target_vol);
    my $def = $volopts->{snaplock_default_period};
    vol_options($conn, $target_vol, snaplock_default_period=>'max');
    vol_options($conn, $target_vol, snaplock_minimum_period=>'0y');
}

## @name ::set_retention_periods
##
## @description
##  Set the retention periods for the volume
## @synopsis
##  set_retention_periods ($$%)
##
## @example
##  set_retention_periods($conn, $volume, %periods);
##

##Modify the order of setting the retention periods
## First Set the default, then Min and Max

sub set_retention_periods
{
    my ($conn, $target_vol, %options) = @_;

    # default this to period
    my $default = $options{default};
    my $minimum = $options{minimum};
    my $maximum = $options{maximum};

    $default = $main::WORM_RETENTION if (!defined($default));

    logDbg("setting retention to min=$minimum, def=$default, max=$maximum");
    # default this to WORM_VOL
    $target_vol = $main::WORM_VOL if (!defined($target_vol));

    if ($main::SKIP_SNAPLOCK_RETENTION_SETUP) {
        return 'SKIP_SNAPLOCK_RETENTION_SETUP is set';
    } else {
        logTrc("get the version and decide whether to set the retention period...");
        my $ver = get_wafl_version($conn->{hostp});

        logDbg("ver is '$ver', SKIP_SNAPLOCK_RETENTION_SETUP is '$main::SKIP_SNAPLOCK_RETENTION_SETUP'");

        if ($ver <= WV_SCRIMSHAW_MAX) {
            logTrc("cannot set the snaplock default retention period and SKIP_SNAPLOCK_RETENTION_SETUP is NOT set!!!");
            die na_result(UNSUPPORTED_FEATURE_EXCEPTION);
        } else {
            my @times = ();
            # set the retention periods
            logTrc("set the snaplock retention periods");
            logDbg("set to min=$minimum, def=$default, max=$maximum");
            my %volopts;
            set_retention_period_to_default($conn,$target_vol);
            if (defined($maximum)) {
                %volopts = vol_options(
                                       $conn,
                                       $target_vol,
                                       snaplock_maximum_period => $maximum,
                                      ) or die $@;
            }

            if (defined($minimum)) {
                %volopts = vol_options(
                                       $conn,
                                       $target_vol,
                                       snaplock_minimum_period => $minimum,
                                      ) or die $@;
            }
            if (defined($default)) {
               %volopts = vol_options(
                                       $conn,
                                       $target_vol,
                                       snaplock_default_period => $default,
                                      ) or die $@;
            }

            my $retention_default = $volopts{snaplock_default_period};
            my $retention_minimum = $volopts{snaplock_minimum_period};
            my $retention_maximum = $volopts{snaplock_maximum_period};
            logTrc("SnapLock minimum retention time is '$retention_minimum'");
            if ($minimum and ($retention_minimum ne $minimum)) {
                die("SnapLock minimum retention mismatch: set $minimum, is $retention_minimum");
            }
            logTrc("SnapLock default retention time is '$retention_default'");
            if ($default) {
                if ($retention_default ne $default) {
                    die("SnapLock default retention mismatch: set $default, is $retention_default");
                } else {
                    $main::WORM_RETENTION = $retention_default;
                }
            }
            logTrc("SnapLock maximum retention time is '$retention_maximum'");
            if ($maximum and ($retention_maximum ne $maximum)) {
                die("SnapLock maximum retention mismatch: set $maximum, is $retention_maximum");
            }
            return wantarray ? ($retention_minimum, $retention_default,
                                $retention_maximum) : $retention_default;
        }
    }
    return undef;
}

sub set_retention_or_die {
    my @args = @_;

    my $retention = eval {
        return set_default_retention_period(@args);
    };
    if ($@) {
        if (index($@, UNSUPPORTED_FEATURE_EXCEPTION) != -1) {
            die na_result($@);
        }
        die $@;
    }
    return $retention;
}


## @name ::compliance_clock_status
##
## @description
##  Check whether the compliance clock has been initialized. Returns the
##  compliance clock time or undef.
##
##  Dies with UNSUPPORTED_FEATURE_EXCEPTION if the underlying OnTap version
##  does not supoort compliance clock initialization (scrimshaw).
##
##  Dies with SNAPLOCK_UNLICENSED_EXCEPTION if no snaplock license is
##  installed.
##
## @synopsis
##  compliance_clock_status ($conn)
##
## @arg $conn required: valid connection
##
## @example
##  $time = compliance_clock_status($conn);
##

sub compliance_clock_status($;$) {
    my ($conn, $vol) = @_;

    ##my $volume = defined($vol) ? $vol : "vol0";
    my $volume = $vol;

    # add an entry to the generic say hash
    $say_hash{'Error: Compliance Clock has not been initialized.'} = sub { return undef; };
    my $ver = get_wafl_version($conn->{hostp}) ;
    logDbg("Wafl version is $ver");
    my $cc_time;
    my $ret;
    if ( $ver < WV_ROLLINGROCK_MIN ) {
        logDbg("Releases before RR");
        $ret = _do_say($conn, 'date -c', \%say_hash);
        chomp $ret;
        $ret =~ /^(Compliance Clock: )?(.*)$/;
        my $date = &UnixDate($2, "%s");
        logDbg("compliance clock: date($ret) -> timestamp($date)");
        return ($date and $date =~ /^\d+/) ? $date : $ret;
    }else {
        logDbg("New command from rolling rock");
        my $cmd = 'snaplock clock status '.$volume;
        eval {
                (undef, undef, $ret)  = say($conn, $cmd);
        };
        if ( $ret !~ /Error/  ) {
            if ($ret !~ /snaplock clock status requires snaplock or snaplock enterprise license/  ) {
                 if ( ! defined $volume) {
                     logDbg("Volume Name is not defined. So get SCC from here");
                     my @line = split(/\n/, $ret) ;
                     $ret = $line[0];
                     chomp $ret;
                     $ret =~ /^(System Compliance Clock: )?(.*)$/;
                     my $date = &UnixDate($2, "%s");
                     logDbg("compliance clock: date($ret) -> timestamp($date)");
                     return ($date and $date =~ /^\d+/) ? $date : $ret;
                  } else {
                        logDbg("Command Succeded : remove the header info");
                        die "Error stripping OutPut header: $ret\n"
                        unless ( $ret =~ s/.*?\s*\-+\s*(.*)$/$1/s );
                        chomp $ret;
                        foreach my $line (split(/\n/, $ret)) {
                            if ( $line =~ /^\s*(\S+)\s*(.*)$/) {
                                $cc_time = $2;
                            }
                        }
                        $ret = $cc_time;
                  }
                 logDbg("In case of Failure, return the error message");
                 my $date = &UnixDate($ret, "%s");
                 logDbg("compliance clock: date($ret) -> timestamp($date)");
                 return ($date and $date =~ /^\d+/) ? $date : $ret;
           }
       }
        return $ret;
    }
}

## @name :: compliance_clock_status_verbose
##
## @description
##  Returns the hashref of compliance clock details from RR release onwards.
##
## Dies with UNSUPPORTED_FEATURE_EXCEPTION if the underlying OnTap version
## does not support this command.
##
## @synopsis
##  compliance_clock_status($conn,$volume)
##
## @arg $conn required: valid connection
##
## @arg $volume optional: If not specified, returns the status
##      of 'vol0' compliance clock ( i.e system compliance clock )
##
## @example
## $data = compliance_clock_status($conn, $vol_name);
##

sub compliance_clock_status_verbose {

    my ($conn, $volume) = @_;
    my $ver = get_wafl_version($conn->{hostp}) ;
    logDbg("Wafl version is $ver");
    if ( $ver < WV_ROLLINGROCK_MIN ) {
        croak 'Command is not available on pre-RR releases';
    }
    my $cmd = 'priv set diag; snaplock clock status -v '.$volume.' ;priv set';
    my (undef, undef,$ret) = say($conn, $cmd);

    if ( $ret =~ /Error/ ) {
        return $ret;
    }
    ## Remove the header information.
    die "Error stripping OutPut header: $ret\n"
        unless ( $ret =~ s/.*?\s*\-+\s*(.*)$/$1/s );
    chomp $ret;

    my %info;
############################################## Sample output format #######################################################
##System Compliance Clock: 2010-04-23:06:41:48                                                                            #
##System Clock: 2010-04-23:06:41:48                                                                                       #
##System Compliance Clock Skew: 0                                                                                         #
##Node ID: 118069060                                                                                                      #
##SCC ID: 0x128286b7500                                                                                                   #
##                     In-Core Volume      On-disk Volume      On-disk System                                             #
##Volume             Compliance Clock    Compliance Clock    Compliance Clock          SCC id   Node id      VCC skew     #
##--------------- ------------------- ------------------- ------------------- --------------- --------- -------------     #
##vl03125_733f_3cd2595f 2010-04-23:06:41:48 2010-04-23:06:20:45 2010-04-23:06:20:45 0x128286b7500 118069060             0 #
###########################################################################################################################

    foreach my $line (split(/\n/, $ret)) {
        if ( $line =~ /Warning/ ) {last;}
        if ( $line =~ /^\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)\s*(.*)$/ ) {
            $info{volume} = $1;
            $info{in_core_vcc} = $2;
            $info{on_disk_vcc} = $3;
            $info{on_disk_scc} = $4;
            $info{scc_id}      = $5;
            $info{node_id}     = $6;
            $info{skew}        = $7;
         }
    }

    ## Convert the Date strings in the hash to UNIX time stamp.
    ## Execute 'date' command and get the time zone info.
    my ( undef, undef, $date)  =  say($conn, "date");
    #my @date_split = split(/ / , $date);
    my $timezone;
    if ($date =~ /^(\w+)\s*((\w+)\s*(\d+))\s*((..):(..):(..))\s*(\w+)\s*(\d+)/)  {
        $timezone = $9;
    } else {
        die "date format is not proper";
    }
    $info{in_core_vcc} =~ s/-|://g;
    $info{in_core_vcc} = $info{in_core_vcc}.$timezone;
    $info{in_core_vcc} = &UnixDate($info{in_core_vcc}, "%s");
    print ("incore vcc $info{in_core_vcc} \n");
    $info{on_disk_vcc} =~ s/-|://g;
    $info{on_disk_vcc} = $info{on_disk_vcc}.$timezone;
    $info{on_disk_vcc} = &UnixDate($info{on_disk_vcc} , "%s");
    $info{on_disk_scc} =~ s/-|://g;
    $info{on_disk_scc} = $info{on_disk_scc}.$timezone;
    $info{on_disk_scc} =  &UnixDate($info{on_disk_scc}, "%s");

    return (\%info);
}

#
# Examples of how filers respond to compliance clock initialization
# commands.
#
# mcrae1> version; date -c
# NetApp Release RscrimshawN_050208_2300: Tue Feb  8 23:18:51 PST 2005  <td>
# date: Invalid date specified.
# usage: date [-u] [[[[CC]yy]mm]dd]hhmm[.ss]]
#
# mcrae2> version; date -c
# NetApp Release RanchorsteamN_050208_2215: Tue Feb  8 22:42:42 PST 2005  <td>
# date: The -c option requires a snaplock or snaplock_enterprise license.
#
# mcrae2> version; date -c initialize
# NetApp Release RanchorsteamN_050208_2215: Tue Feb  8 22:42:42 PST 2005  <td>
# date: The -c option requires a snaplock or snaplock_enterprise license.
#
# -> licensed snaplock
# mcrae2> version; date -c; date -c initialize
# NetApp Release RanchorsteamN_050208_2215: Tue Feb  8 22:42:42 PST 2005  <td>
# Error: Compliance Clock has not been initialized.
#
# mcrae2> date -f -c initialize
# Wed Feb  9 16:35:51 GMT [rc:info]: Compliance Clock initialized (via "date") to
# Wed Feb  9 16:35:51 GMT 2005
# Compliance Clock: Wed Feb  9 16:35:51 GMT 2005
# mcrae2> date -c
# Compliance Clock: Wed Feb  9 16:36:48 GMT 2005
#


# this is an internal utility function which 'says' the given command and
# watches the output for a match. any action associated with a matched
# string is then executed and the result returned.
sub _do_say($$$) {
    my ($conn, $cmd, $say_hash) = @_;

    my ($match, $before, $after) = say(
        $conn,
        $cmd,
        keys %{$say_hash},
    );

    die "unexpected result: $after" if (!exists $say_hash->{$match});

    return &{$say_hash->{$match}}($after);
}

## @name ::compliance_clock_ensure_initialized
##
## @description
##  Backwards-compatibility wrapper for compliance_clock_initialize().
##  This function is currently used only by:
##       test/dense/integ/snaplock/dense_snaplock.thpl
##       test/dense/integ/snaplock/dense_snaplock_enterprise.thpl
##
## @synopsis
##  compliance_clock_ensure_initialized ($conn)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @example
##  compliance_clock_ensure_initialized($conn);
##

sub compliance_clock_ensure_initialized($)
{
    compliance_clock_initialize(@_, reinit_ok => 1);
}

## @name ::snaplock_remove_file
##
## @description
##    Remove a (snaplock) file if the underlying version supports it
##
## @synopsis
##  snaplock_remove_file($conn, $filepath)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg $filepath required: the file to remove
##
## @example
##  snaplock_remove_file($conn, $filepath);
##

sub snaplock_remove_file {
    my ($conn, $filename) = @_;

    logTrc("get the version and decide what to do...");
    my $ver = get_wafl_version($conn->{hostp});
    logDbg("ver is '$ver'");

    if ($ver <= WV_SCRIMSHAW_MAX) {
        # just warn
        logWarn('pre-anchorsteam: not able to remove files on snaplock volumes');
    } else {
        my (undef, undef, $after) = say($conn, "priv set -q diag; rm $filename");
        return $after;
    }
    return undef;
}

## @name ::snaplock_remove_dir
##
## @description
##  Backwards-compatibility wrapper for compliance_clock_initialize().
##  This function is currently used only by:
##       test/dense/integ/snaplock/dense_snaplock.thpl
##       test/dense/integ/snaplock/dense_snaplock_enterprise.thpl
##
## @synopsis
##  snaplock_remove_dir($conn, $dir)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg $dir required: the directory to remove
##
## @example
##  compliance_clock_ensure_initialized($conn);
##

sub snaplock_remove_dir {
    my ($conn, $dir) = @_;

    logTrc("get the version and decide what to do...");
    my $ver = get_wafl_version($conn->{hostp});
    logDbg("ver is '$ver'");

    if ($ver <= WV_SCRIMSHAW_MAX) {
        # just warn
        logWarn('pre-anchorsteam: not able to remove directories on snaplock volumes');
    } else {
        my $sudo = NATE::ParamSet::get_global()->get('SUDO_PATH');
	return uSystem("/usr/local/bin/r rmdir $dir");
    }
    return undef;
}

## @name ::wait_for_expiry
##
## @description
##  This routine waits for the expiry of the file locked at the given
##  $lock_time, as long as the time is not longer than $max_wait. The
##  expiry time is calculated based on $retention_spec.
##
## @synopsis
##  wait_for_expiry($conn, $lock_time, $retention, $max_wait)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg $lock_time required: the compliance clock time (in epoch seconds) at which the target object was locked. use the compliance_clock_status() function at file lock time to get this value, and Date::Manip::UnixDate() to convert it to epoch-seconds.
##
## @arg $retention_spec optional: the target's retention time - uses the same format as the WORM_RETENTION param. defaults to $main::WORM_RETENTION.
##
## @arg $max_wait optional: the maximum time to wait in seconds. defaults to MAX_WAIT_SECONDS.
##
## @example
##
##   wait_for_expiry($conn, &UnixDate($lock_time, '%s'));
##

sub wait_for_expiry {
    my ($conn, $lock_time, $retention_spec, $max_wait, $vol_name) = @_;

    die "conn is a required parameter" if (!defined($conn));
    die "lock time is a required parameter" if (!defined($lock_time));
    $vol_name = "vol0" if ( !defined($vol_name));

    # set defaults?
    $retention_spec = $main::WORM_RETENTION if (!defined($retention_spec));
    $max_wait = MAX_WAIT_SECONDS if (!defined($max_wait));

    # get the retention in seconds
    my $retention = parse_retention($retention_spec);

    my $curr_time = compliance_clock_status($conn , $vol_name) or die "failed to get current compliance clock time";
    if ($curr_time !~ /^\d+/) {
        logTrc("'date -c' returns non-numeric, assume pre-7.0 and don't bother waiting: ($curr_time)");
        return;
    }
    my $elapsed = &UnixDate($curr_time,'%s') - $lock_time;
    my $sleep_time = $retention - $elapsed + 2;

    logDbg("\nretention is '$retention' seconds.\nelapsed time is '$elapsed' seconds.\nsleep time is '$sleep_time' seconds");

    # sleep for a bit (wait for file locks to expire)
    if ($sleep_time < $max_wait) {
        if ($sleep_time > 0) {
            logStep("sleep for $sleep_time seconds");
            sleep($sleep_time);
        }
    } else {
        logStep("sleep time is too long: $sleep_time seconds");
    }
}

## @name ::wait_file_expiry
##
## @description
##  This routine waits for the expiry of the file (defined as ability to
##  chmod +w it), as long as the time is not longer than $max_wait.
##
## @synopsis
##  wait_file_expiry($conn, $filepath, $max_wait)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg $filepath required: relative path to file
##
## @arg $max_wait optional: the maximum time to wait in seconds.
##                defaults to MAX_WAIT_SECONDS.
##
## @arg client_conn optional: client connectobj for access to filepath.
##
## @example
##
##   wait_file_expiry($conn, $filepath [, $max_wait ], [client_conn=>$cc ]);
##

sub wait_file_expiry {
    my ($conn, $filepath, $max_wait, %opts) = @_;

    die "conn is a required parameter" if (!defined($conn));
    die "filepath is a required parameter" if (!defined($filepath));

    # set defaults?
    $max_wait = MAX_WAIT_SECONDS if (!defined($max_wait));
    my $cc = defined($opts{client_conn}) ? $opts{client_conn} : undef;
    my $endit = time() + $max_wait;
    logDbg("max wait time $endit (".gmtime($endit).")");

    my $atime;
    eval { $atime = client_get_atime($cc, $filepath); };
    return if $@;               # assume file expired
    chomp($atime);
    logDbg("waiting for expiry atime $atime (".gmtime($atime).")");

    my @path_split = split(/\//, $filepath, 5);
    my $vol_name = $path_split[3];

    ## Volume name is required to get CC from RR release
    my $curr_time = compliance_clock_status($conn, $vol_name)
        or die "failed to get current compliance clock time";
    if ($curr_time !~ /^\d+/) {
        logTrc("'date -c' returns non-numeric, assume pre-7.0 and don't bother waiting: ($curr_time)");
        return;
    }

    while (($curr_time < $atime) and (time() < $endit)) {
        sleep($::WORM_RETENTION/2 + 10);
        $curr_time = compliance_clock_status($conn , $vol_name )
            or die "failed to get current compliance clock time";
    }

    if (time() > $endit) {
        logDbg("Maximum wait time expired while waiting for expiry");
    }
}

## @name ::retention_expired
##
## @description
##  Returns true if $lock_time is at least $retention seconds behind the current
##  compliance clock time.
##
## @synopsis
##  retention_expired($conn, $lock_time, $retention)
##
## @arg $conn required: valid connection (see snaplock usage description)
##
## @arg $lock_time required: the compliance clock time (in epoch seconds) at which the target object was locked. use the compliance_clock_status() function at file lock time to get this value.
##
## @arg $retention_spec optional: the target's retention time - uses the same format as the WORM_RETENTION param. defaults to $main::WORM_RETENTION.
##
## @example
##
##   retention_expired($conn, &UnixDate($lock_time, '%s'));
##

sub retention_expired {
    my ($conn, $lock_time, $retention_spec, $vol_name) = @_;

    die "conn is a required parameter" if (!defined($conn));
    die "lock_time is a required parameter" if (!defined($lock_time));

    $vol_name = "vol0" if ( !defined($vol_name ) ) ;

    # set defaults?
    $retention_spec = $main::WORM_RETENTION if (!defined($retention_spec));

    my $retention = parse_retention($retention_spec);

    #### Volume name is required to get CC from RR release
    my $curr_time = compliance_clock_status($conn , $vol_name) or die "failed to get current compliance clock time";
    if ($curr_time !~ /^\d+/) {
        logTrc("'date -c' returns non-numeric, assume pre-7.0 and doesn't expire: ($curr_time)");
        return 0;
    }

    my $elapsed = $curr_time - $lock_time;

    logDbg("\nretention is '$retention' seconds.\ntime elapsed since lock is '$elapsed' seconds");

    return($retention <= $elapsed);
}


## @name ::parse_retention
##
## @description
##  Returns the value of given retention spec in seconds.
##
## @synopsis
##  parse_retention($retention)
##
## @arg $retention_spec optional: the target retention time - uses the same
## format as the WORM_RETENTION param. defaults to $main::WORM_RETENTION.
##
## @example
##
##   $retention_seconds = parse_retention();
##

sub parse_retention {
    my ($retention_spec) = @_;

    $retention_spec = $main::WORM_RETENTION if (!defined($retention_spec));

    my ($magnitude, $units) = ($retention_spec =~ m/([0-9]+)(.*)/);
    if ($units eq 'd') {
        $magnitude *= SECONDS_PER_DAY;
    } elsif ($units eq 'm') {
        $magnitude *= SECONDS_PER_MONTH;
    } elsif ($units eq 'y') {
        $magnitude *= SECONDS_PER_YEAR;
    } else {
        # nothing to do (assume units is 'seconds')
    }

    return $magnitude;
}

## @name ::client_set_atime
##
## @description
##  Set the atime of a file from a client. Returns empty if success,
##  error message if failure.
##
## @synopsis
##  client_set_atime($$$)
##
## @arg client_conn   Connection object to a client.
##
## @arg filename      Client-relative path to file
##
## @arg time          atime to set: 'touch -t' format. (GMT)
##
## @arg protocol      named parameter: cifs|nfs; default = nfs
##
## @example
##
##   $result = client_set_atime($cl_conn, $filename, $date);
##

sub client_set_atime ($$$;@)
{
    my ($cc, $file, $date, %opts) = @_;

    my $ret;
    my $cmd = "env TZ=GMT touch -a -t $date $file";
    if (defined($cc)) {
        if (defined($opts{protocol}) and $opts{protocol} eq 'cifs') {
            $cc = make_conn($cc) unless ref $cc;
            (undef, undef, $ret) = say($cc, "cmd /c x:\\snaplock.exe extend $file $date");
        } else {                # protocol == nfs
            (undef, undef, $ret) = say($cc, $cmd);
        }
    } else {
        $ret = uSystem($cmd);
    }
    chomp($ret);
    return $ret;
}

## @name ::client_commit_file
##
## @description
##  Commit the file from a client. Returns empty if success,
##  die if failure
##
## @synopsis
##  client_commit_file($$$)
##
## @arg client_conn   Connection object to a client.
##
## @arg filename      Client-relative path to file
##
## @arg protocol      named parameter: cifs|nfs; default = nfs
##
## @example
##
##   $result = client_commit_file($cl_conn, $filename);
##

sub client_commit_file ($$;@)
{
    my ($cc, $filename, %opts) = @_;

    my $ret;
    my $cmd = "test -w $filename && chmod -w $filename && test \! -w $filename && echo 'file was committed' || echo 'wrong state'";

    if (defined($cc)) {
        if ($opts{protocol} and $opts{protocol} eq 'cifs') {
            $ret = execute_cmd($cc, "cmd /c x:\\snaplock.exe commit $filename");
            if ($ret) {
                die "client_commit_file: $ret";
            } else {
                return;
            }
        } else {                # protocol == nfs
            (undef, undef, $ret) = say($cc, $cmd);
        }
    } else {
        $ret = uSystem($cmd);
    }
    chomp($ret);
    if ($ret =~ /wrong state/) {
        die "client_commit_file: $filename was not writable on entry";
    } elsif ($ret =~ /file was committed/) {
        return;
    } else {
        die "client_commit_file: $ret";
    }
}

## @name ::client_get_atime
##
## @description
##  Get the atime of a file via a client. Returns unix timestamp (int) GMT
##  die()s with a failure.
##
## @synopsis
##  client_get_atime($$$)
##
## @arg client_conn   Connection object to a client.
##
## @arg filename      Client-relative path to file
##
## @arg protocol      named parameter: cifs|nfs; default = nfs
##      NOTE: currently does not work via CIFS, use NFS
##
## @example
##
##   $atime = client_get_atime($cl_conn, $filename);
##

sub client_get_atime ($$;@)
{
    my ($cc, $filename, %opts) = @_;

    my ($ret, $perl_path);

    if (defined($cc)) {
        if ($opts{protocol} and $opts{protocol} eq 'cifs') {
            $cc = make_conn($cc) unless ref $cc;
            (undef, undef, $ret) = say($cc, qq(cmd /c x:\\perl\\bin\\perl -le "\$a=(stat(qq($filename)))[8]; print \$a"));
        } else {                # protocol == nfs
        		
        	(undef, undef, $perl_path) = say($cc, "which perl");
        	chomp($perl_path);
        	my @perlp = split("\r", $perl_path);
        	print "Perl path on client is $perlp[0]";
        	#(undef, undef, $ret) = say($cc, qq(env TZ=GMT /usr/local/bin/perl -le '\$a=(stat(qq($filename)))[8]; print \$a'));
        
        	(undef, undef, $ret) = say($cc, qq(env TZ=GMT $perlp[0] -le '\$a=(stat(qq($filename)))[8]; print \$a'));
        }
    } else {
        $ret = (stat("$filename"))[8];
    }
    if ($ret =~ /^\r*0*\r*$/) {
        return "0";
    } else {
        chomp($ret);
    }
    my $gmt = gmtime($ret);
    logDbg("returning atime for $filename as $ret (".chomp($gmt).")");
    if ($ret =~ /\d+/) {
        return $ret;
    } else {
        die($ret);
    }
}


## @name ::client_get_ctime
##
## @description
##  Get the ctime of a file via a client. Returns unix timestamp (int) GMT
##  die()s with a failure.
##
## @synopsis
##  client_get_ctime($$$)
##
## @arg client_conn   Connection object to a client.
##
## @arg filename      Client-relative path to file
##
## @arg protocol      named parameter: cifs|nfs; default = nfs
##      NOTE: currently does not work via CIFS, use NFS
##
## @example
##
##   $ctime = client_get_ctime($cl_conn, $filename);
##

sub client_get_ctime ($$;@)
{
    my ($cc, $filename, %opts) = @_;

    my ($ret, $perl_path);

    if (defined($cc)) {
        if ($opts{protocol} and $opts{protocol} eq 'cifs') {
            $cc = make_conn($cc) unless ref $cc;
            (undef, undef, $ret) = say($cc, qq(cmd /c x:\\perl\\bin\\perl -le "\$a=(stat(qq($filename)))[10]; print \$a"));
        } else {                # protocol == nfs
        	(undef, undef, $perl_path) = say($cc, "which perl");
        	chomp($perl_path);
        	my @perlp = split("\r", $perl_path);
        	print "Perl path on client is $perlp[0]";
        	#(undef, undef, $ret) = say($cc, qq(env TZ=GMT /usr/local/bin/perl -le '\$a=(stat(qq($filename)))[10]; print \$a'));
        
        	(undef, undef, $ret) = say($cc, qq(env TZ=GMT $perlp[0] -le '\$a=(stat(qq($filename)))[10]; print \$a'));
        }
    } else {
        $ret = (stat("$filename"))[10];
    }
    if ($ret =~ /^\r*0*\r*$/) {
        return "0";
    } else {
        chomp($ret);
    }
    my $gmt = gmtime($ret);
    chomp($gmt);
    logDbg("returning ctime for $filename as $ret (".chomp($gmt).")");
    if ($ret =~ /\d+/) {
        return $ret;
    } else {
        die($ret);
    }
}

## @name ::snaplock_scan_start($conn, $vol [,%opts])
##
## @summary
##      Start the wafl snaplock scanner
##
## @description
##      Start the snaplock scanner for the specified volume. Returns 1 for
##      success; croaks otherwise
##
## @arg $conn  Required: filer connection object
##
## @arg $vol   Required: volume name
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg wait Optional: 0 or 1. Defaults to 0.
##                              Wait for the scanner to complete ?
##
## @arg timeout Optional: if wait is set, how long to wait. Defaults to 600.
##
sub snaplock_scan_start($$;%)
{
        my ($conn, $vol, %opts) = @_;
        my $cmd = "wafl scan snaplock $vol";
        my $text = say($conn, "priv set -q diag; $cmd");
        $text eq "" or croak("$cmd failed: $text");
        sleep(30);

        if (defined($opts{wait}) && $opts{wait} == 1) {
                my $timeout = $opts{timeout} || 600;
                return(snaplock_scan_wait($conn, $vol, timeout => $timeout));

        }
        return 1;
}

## @name ::snaplock_scan_wait($fc, $vol [,%opts])
##
## @summary
##      Wait for the wafl snaplock scanner to complete.
##
## @description
##      Wait for the snaplock scanner for the specified volume to complete.
##      Returns 1 for success; croaks on timeout
##
## @arg $conn  Required: filer connection object
##
## @arg $vol   Required: volume name
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg timeout Optional: how long to wait. Defaults to 600.
##
sub snaplock_scan_wait($$;%)
{
        my ($conn, $vol, %opts) = @_;
        my $timeout = $opts{timeout} || 600;
        my $etime = time() + $timeout;

        while(snaplock_scan_status($conn, $vol)) {
                my $t = time();
                print ("Time = $t and etime = $etime");
                if (time() > $etime) {
                        croak("Timed out waiting for snaplock scanner to complete");
                        return 0;
                }
                sleep(30);
        }
        return 1;
}

## @name ::snaplock_scan_status($fc, $vol)
##
## @summary
##     Query snaplock scanner status for the specified volume.
##
## @description
##     Queries the snaplock scanner status for the specified volume.
##
##     If the snaplock scanner is not currently running for the specified
##      volume, the function returns 0.  For any unexpected behavior, the
##      function croaks.
##
## @arg $conn Required: filer connection object.
##
## @arg $vol Required: volume name.
##
## @arg %opts Optional: name/value pairs of the following
##
## @arg retries Optional: how many times to try and read the snaplock scan
##                      status. Defaults to 5.
##

sub snaplock_scan_status($$)
{
    my ($conn, $vol, %opts) = @_;
    my $retries = $opts{retries} || 5;
    my $tries = 0;
    while ($tries <= $retries) {
        my $rc;
        eval { $rc = _snaplock_scan_status($conn, $vol); };
        if ($@) {
            if ($@ =~ /Unexpected output/) {
                ++$tries;
                sleep(5);
                next;
            } else {
                croak($@);
            }
        } else {
            return($rc);
        }
    }
}

sub _snaplock_scan_status ($$)
{
    my ($conn, $vol) = @_;

    my $idle_re = qr(No\s+WAFL\s+scans\s+in\s+progress);
    my $vol_re = qr(Volume\s$vol:);
    my $hdr_re = qr(Scan\s+id\s+Type\s+of\s+scan\s+progress);

    # There are two expected cases:
    #
    # 1. No scanners in progress: return 0
    # 2. Scanner status

    my $cmd = "wafl scan status $vol";
    my ($match, undef, $after) = say($conn, "priv set -q advanced; $cmd");
    say($conn, "priv set");

    # No scans in progress?
    return 0 if ($after =~ /$idle_re/ || $after !~ /snaplock/);

    return 1;
}

## @name ::snaplock_get_compliance_clock
##
## @Description
##       Gets the Current compliance clock value of a filer.
##       the format is as follows
##        set_log_volume_retention($Filer_conn, $default_retention)
##
## @arg $filer_ip Required. Required to execute API call to the filer
##
## @arg $filer_passwd Required. Required for authentication
##
## @arg $result Required. To decide pass fail of API call
##

sub snaplock_get_compliance_clock {
        #shift @_;

        ### Get the required values.
        my ($filer_ip,$filer_passwd,$result) = @_;


        my $zapi = new ZapiTest($filer_ip,"root", $filer_passwd);

        my $zcmd = new NaElement("snaplock-get-compliance-clock");
        my $xo = $zapi->invoke_elem($zcmd, "any");
        my $test = $xo->toString();
        logcomment("$test");

        if ($test =~ /results status=\"$result\"/i) {
                logcomment("Zapi expected to be $result and it $result");
                return 1;
        } else {
                logcomment ("Zapi call expected to be $result and it is failed");
                return 0;
        }
}

## @name ::snaplock_get_compliance_clock
##
## @Description
##      Gets the Current compliance clock value of a filer.
##      the format is as follows
##              set_log_volume_retention($Filer_conn, $default_retention)
##
## @arg $filer_ip Required. Required to execute API call to the filer
##
## @arg $filer_passwd Required. Required for authentication
##
## @arg $result Required. To decide pass fail of API call
##

sub file_set_snaplock_retention_time  {
        my ($filer_ip,$filer_passwd,$file_path,$retention_time,$result) = @_;
        my $zapi = new ZapiTest($filer_ip,"root", $filer_passwd);
        my $zcmd = new NaElement("file-set-snaplock-retention-time");
        $zcmd->child_add(new NaElement("path", $file_path));
        $zcmd->child_add(new NaElement("retention-time", $retention_time));
        my $xo = $zapi->invoke_elem($zcmd, "any");
        my $test = $xo->toString();
        logcomment("$test");
        if ($test =~ /results status=\"$result\"/i) {
                logcomment("Zapi expected to be $result and it $result");
                return 1
        } else {
                logcomment ("Zapi call expected to be $result and it is failed");
                return 0;
        }

}

## @name ::snaplock_get_compliance_clock
##
## @Description.
##      Gets the Current compliance clock value of a filer.
##      the format is as follows
##              set_log_volume_retention($Filer_conn, $default_retention)
## @arg $filer_ip Required. Required to execute API call to the filer
##
## @arg $filer_passwd Required. Required for authentication
##
## @arg $result Required. To decide pass fail of API call
##

sub file_get_snaplock_retention_time {
        my ($filer_ip,$filer_passwd,$file_path,$result) = @_;
        my $zapi = new ZapiTest($filer_ip,"root", $filer_passwd);
        my $zcmd = new NaElement("file-get-snaplock-retention-time");
        $zcmd->child_add(new NaElement("path", $file_path));
         my $xo = $zapi->invoke_elem($zcmd, "any");
        my $test = $xo->toString();
        logcomment("$test");
        if ($test =~ /results status=\"$result\"/i) {
                logcomment("Zapi expected to be $result and it $result");
                 return $xo->child_get("retention-time")->content;
        } else {
                logcomment ("Zapi call expected to be $result and it is failed");
                return 0;
        }

}

## @name ::file_get_snaplock_retention_time_list_info_max
##
## @Description
##      Gets the Current compliance clock value of a filer.
##      the format is as follows
##              file_get_snaplock_retention_time_list_info_max ($filer_ip,$filer_passwd,"passed")
## @arg $filer_ip Required. Required to execute API call to the filer
##
## @arg $filer_passwd Required. Required for authentication
##
## @arg $result Required. To decide pass fail of API call
##
sub file_get_snaplock_retention_time_list_info_max {
        #shift @_;

        ### Get the required values.
        my ($filer_ip,$filer_passwd,$result) = @_;


        my $zapi = new ZapiTest($filer_ip,"root", $filer_passwd);

        my $zcmd = new NaElement("file-get-snaplock-retention-time-list-info-max");
        my $xo = $zapi->invoke_elem($zcmd, "any");
        my $test = $xo->toString();
        logcomment("$test");

        if ($test =~ /results status=\"$result\"/i) {
                logcomment("Zapi expected to be $result and it $result");
                return $xo->child_get("max-list-entries")->content;
        } else {
                logcomment ("Zapi call expected to be $result and it is failed");
                return 0;
        }
}

## @name ::check_worm_flag_secure_clock
##
## @description
##  Verify worm flag secure clock is set or not on the given volume.
##  Returns true if the flag is set, otherwise nothing will be returned.
##
## @synopsis
##  check_worm_flag_secure_clock($vol_obj)
##
## @arg $vol_obj required: valid NOTE volume object
##
## @example
##  $status = check_worm_flag_secure_clock($vol_obj);
##


sub check_worm_flag_secure_clock($) {
    my ($vol) = @_;
    sleep 2;
    my $dump_hash = dumpblock(conn => $vol->Filer->conn,  block_type=>"vbn",  volume=> $vol->name , vbn=>1, );
    my $line_cnt = 0;
    my $flag_line_no = 0;

    ### Worm flag offset = 0x00e0
    while ($dump_hash->{block_contents}[$line_cnt]) {
        if ($dump_hash->{block_contents}[$line_cnt] =~ m/0x00e0\:/i){
            $flag_line_no = $line_cnt;
        }
            $line_cnt++;
    }
    my $worm_flag_dump = $dump_hash->{block_contents}[$flag_line_no];
    logcomment("secure clock flag obtained from volume $worm_flag_dump");
    my @worm_flag_values = split(' ',$worm_flag_dump);
    my $worm_flag = sprintf("%b" , hex($worm_flag_values[1]));
    logcomment("Binary format of secure clock flag $worm_flag ");
    my @eighth_bit_to_comp = split(//,$worm_flag);
    if ( $#eighth_bit_to_comp > 4 ) {
        logcomment("Obtained value is >= 0x0000080, Check of WORM_VIFLAG_VOL_SECURE_CLOCK_SET is set or not");
        my $Secure_flag_bit = $#eighth_bit_to_comp - 7;
        if($eighth_bit_to_comp[$Secure_flag_bit] == 1 ) {
            logcomment("WORM_VIFLAG_VOL_SECURE_CLOCK_SET is set on the volume\n");
            return 1;
        } else {
            return undef;
        }
    }else {
        logcomment("Obtained value is < 0x0000080, So WORM_VIFLAG_VOL_SECURE_CLOCK_SET is not set on the volume");
        return undef;
    }

}

## @name ::corrupt_volinfo_secure_clock_metadata
##
## @description
## Using wreck function call, corrupt the volinfo
## of secure clock metadata.
##
## @synopsis
## corrupt_volinfo_secure_clock_metadata(%);
##
## @arg conn=> $conn required: valid conn object
## @arg vcc_time => $vcc_time required: New Vcc time to change.
## @arg scc_time => $scc_time optional: New Scc time to change .Default set to 0.
## @arg scc_id => $scc_id optional: New Scc ID to change. Default set to 0.
## @arg node_id => $node_id. New node id to change. Default set to 0.
## @arg vol_name => $volume_name. Valid volume name.
## @arg vol_type => $volume_type. trad/flex.
##
## @example
##  eval {
##      corrupt_volinfo_secure_clock_metadata(conn => $conn
##                                             vcc_time => $vcc_time
##                                             vol_name => $vol_name
##                                              vol_type => "flex");
##

sub corrupt_volinfo_secure_clock_metadata (%) {
    my %opts = validate(@_, {
                        conn       => {type => OBJECT},
                        vcc_time   => {type  => SCALAR},
                        scc_time   => { type => SCALAR, optional => 1,
                                        default => 0 },
                        scc_id     => { type => SCALAR, optional => 1,
                                        default => 0 },
                        node_id    => { type => SCALAR, optional => 1,
                                        default => 0 },
                        vol_name   => {type  => SCALAR},
                        vol_type => {type  => SCALAR}});


     my $inject_script = "volinfo_secure_clock_metadata vcc_time=$opts{vcc_time} scc_time=$opts{scc_time} scc_id=$opts{scc_id} node_id=$opts{node_id}";
     my $conn = $opts{conn};
     my $vol_name = $opts{vol_name};
     my $vol_type = $opts{vol_type};
     print ("My volume = $vol_name, my type = $vol_type\n");
     wreck( $conn , $vol_name , "type" => $opts{vol_type} , "inject_script"=>$inject_script  )
                or die "Wreck failed: $@" ;

    my $cmd = "priv set test; wafl flush cache";
    sleep 20;
    my ( undef, undef , $out ) = say($conn , $cmd);
    sleep 20;
    ## Online the volume after wreck;
    vol_online($opts{conn}, $opts{vol_name});

}

## @name ::volume_expiry_date
##
## @description
## Returns the exipray date of a given volume
##
## @synopsis
## volume_expiry_date($filer_conn->conn,$sle_vol->name);
## $filer_conn->conn : Filer connection
## $sle_vol->name: Name of the volume for which expiry date is required.
##

sub volume_expiry_date($;$) {
    my ($conn, $volume) = @_;
    my $return_val;
    my $cmd;

    logcomment("I am at step1");
    #$say_hash {'No such volume'} = sub { return undef; };
    $cmd = "vol status $volume -w";
    logcomment("Command constructed is: $cmd");
    #$return_val = _do_say($conn, $cmd ,\%say_hash);

    (undef,undef,$return_val) =  say($conn,$cmd);
    logcomment("#### Return Value obtained is $return_val");

    chomp $return_val;
    #### verify the return value for the correctness.
    if ($return_val  =~ /vol status: No volume named \'$volume\' exists./i) {
        logcomment("No such volume $volume exist on the system");
   }

   if ($return_val  =~ /$volume Unknown\.\.\.volume offline/i) {
        logcomment("Volume $volume is offlined");
   }

   if ($return_val =~ /$volume none/i) {
        logcomment ("$volume, expiry date is none");
   }

   ### Final choice left is valid expiry date, that should be returned with the unix date format.
   #$return_val =~ /^($volume )?(.*)$/;
  my %extract_val;
  my $ver = get_wafl_version($conn->{hostp}) ;
  if ( $ver < WV_ROLLINGROCK_MIN ) {
        logcomment ("Running Pre-RR version");
        $return_val =~ /$volume +(.*)$/;
         logcomment ("Expiry date is: $1");
         my $extract_val = $1;
         %extract_val= (
                expiry_date => $1,
        );
  } else {
        logcomment ("Running RR Version");
        $return_val =~ /$volume +(.*)   (.*)$/;
        logcomment ("Value obtained is : $return_val");
        logcomment ("Expiry date is: $1");
        %extract_val= (
                expiry_date => $1,
                vcc     => $2,
        );
 }
 return (\%extract_val);
 #my $date = &UnixDate($final_val, "%s");
 #return ($date and $date =~ /^\d+/) ? $date : $return_val;
}

## @name :: snaplock_get_system_compliance_clock
##
## @description
## ZAPI execution snaplock-get-system-compliance-clock
##
## @synopsis
## snaplock_get_system_compliance_clock($filer_conn)
##
## @arg $conn required: valid conn object
## @example
##  eval {
##      snaplock_get_system_compliance_clock($conn ) };
##

sub snaplock_get_system_compliance_clock($) {
    my ($filer_conn) = @_;

    my $zcmd = new NaElement("snaplock-get-system-compliance-clock");
    my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
                                "root",
                                $filer_conn->hostp()->default_password());
    $zapi->{fail} = \&_fail_default;
    my $xo = $zapi->invoke_elem($zcmd,"any");
    if ( ! $xo->results_ok() ) {
            my $error  = "ZAPI snaplock-get-system-compliance-clock". $xo->toString();
            return $error;
    }
    my $v1 = $xo->child_get("system-compliance-clock");
    my $v2 = $v1->child_get("compliance-clock-info");

    my %zapi_out = (
                       "scc" => $v2->child_get("snaplock-compliance-clock")->{content},
                       "fscc" => $v2->child_get("formatted-snaplock-compliance-clock")->{content},
    );
    return (\%zapi_out);
 }

## @name :: snaplock_get_volume_compliance_clock
##
## @description
## ZAPI execution snaplock-get-volume-compliance-clock
##
## @synopsis
## snaplock_get_volume_compliance_clock($filer_conn)
##
## @arg $conn required: valid conn object
## @arg $volume required: Name of the volume.
## @example
##  eval {
##      snaplock_get_system_compliance_clock($conn , $volume ); };
##

sub snaplock_get_volume_compliance_clock($$) {
    my ( $filer_conn , $volume ) = @_;

    my $zcmd = new NaElement("snaplock-get-volume-compliance-clock");
    $zcmd->child_add(new NaElement("volume", $volume));
    my $zapi = Zapi->new($filer_conn->hostp()->default_ip(),
                                "root",
                                $filer_conn->hostp()->default_password());
    $zapi->{fail} = \&_fail_default;
    my $xo = $zapi->invoke_elem($zcmd,"any");
    if ( ! $xo->results_ok() ) {
        my $error  = "ZAPI snaplock-get-volume-compliance-clock". $xo->toString();
        return $error;
    }

    my $v1 = $xo->child_get("volume-compliance-clock");
    my $v2 = $v1->child_get("compliance-clock-info");
    my $v3 = $xo->child_get("system-compliance-clock");
    my $v4 = $v3->child_get("compliance-clock-info");

    my %zapi_out = (
           "vcc" => $v2->child_get("snaplock-compliance-clock")->{content},
           "fvcc" => $v2->child_get("formatted-snaplock-compliance-clock")->{content},
           "scc" => $v4->child_get("snaplock-compliance-clock")->{content},
           "fscc" => $v4->child_get("formatted-snaplock-compliance-clock")->{content},
    );
    return (\%zapi_out);
}

=head2 ::get_vol_expiry_date

Get volume expiry time using ZAPI

The format is as follows
    get_vol_expiry_date ($vol_obj );

=item Arguments

B<Volume_obj> => I<Volume object>

I<Required>
A Volume NOTE object.

=item Examples

=back

    get_vol_expiry_date ( $vol_obj )

=head2 ::status

Returns epoc time of volume expiry time.

=over

=cut

sub get_vol_expiry_date {

        my ( $vol_obj ) = @_;
        my $FILER_OBJ = $vol_obj->Filer;
        my $vol_name = $vol_obj->name;

        my $zapi_exec = exec_zapi ($FILER_OBJ , "ZAPI|volume-list-info*verbose|true*volume|$vol_name");
        my $xml = XML::Simple->new;
        my $zapi_hash_ref = $xml->XMLin($zapi_exec);
        print Dumper $zapi_hash_ref;
        my $expiry_date = $zapi_hash_ref->{'volumes'}{'volume-info'}{'expiry-date'};
        if ( ! defined $expiry_date ) {
            ## expiry date field will not be available if the retention is infinite
            ## or if scanner is in progress, So in that case take formatted expiry date.
            if ( !defined ($zapi_hash_ref->{'volumes'}{'volume-info'}{'formatted-expiry-date'}) ) {
                croak "Did not get neither expiry_date or formatted-expiry-date";
            } else {
                $expiry_date = $zapi_hash_ref->{'volumes'}{'volume-info'}{'formatted-expiry-date'};
                return $expiry_date;
            }
        }
        return $expiry_date;
}


=head2 ::exec_zapi

Execute Zapi Command

The format is as follows
    exec_zapi ($FILER_OBJ,$zapi_param);

=item Arguments

B<Filer_obj> => I<Filer object>

I<Required>
A Filer NOTE object.

B<zapi_param> => I<string>

I<Required>
A string .. Example of Param : "ZAPI|volume-list-info*verbose|true*volume|$vol_name"

=item Examples

=back

    exec_zapi ($FILER_OBJ,$zapi_param)

=head2 ::status

Returns xml formated output Or returns Error message on Failure.

=over

=cut

sub exec_zapi{

        my ($FILER_OBJ,$zapi_param) = @_;

        my $conn = $FILER_OBJ->conn;
        my $ip =$conn->hostp()->default_ip();
        my $user = $conn->hostp()->username;
        my $passwd = $conn->hostp()->default_password();
        $user = 'root';

        my $zapi =  Zapi->new(
                                     $conn->hostp()->default_ip(),
                                     $user ,
                                     $conn->hostp()->default_password(),
                             );

        my @zapi_params = split '\*',$zapi_param;
        my @get_zapi = split '\|',$zapi_params[0];


        my $zapi_name = $get_zapi[1];
        my $zcmd = new NaElement("$zapi_name");


        my $n_params = @zapi_params ;
        my @t_zapi_params =  @zapi_params;
        shift @t_zapi_params;
        foreach my $pairs (@t_zapi_params){

                @zapi_params = split '\|',$pairs;

                my $name = $zapi_params[0];
                my $value = $zapi_params[1];
                $zcmd->child_add(new NaElement($name,$value));
        }
         $zapi->{fail} = \&_fail_default;
         my $zapi_exec = $zapi->invoke_elem($zcmd, "any");
         my $zapi_op  = $zapi_exec->toString();

         return $zapi_op;
}

=head2 ::snaplock_clock_status_verbose

get snaplock clock status of volume in verbose

The format is as follows
    snaplock_clock_status_verbose ($vol_obj);

=item Arguments

B<vol_obj> => I< NOTE volume object>

I<Required>
A Volume  NOTE object.

=item Examples

=back

   snaplock_clock_status_verbose($vol_obj)

=head2 ::status

Returns ref to hash having verose details .

=over

=cut


sub snaplock_clock_status_verbose {
                my ($vol_obj) = @_;
                my $vol_name = $vol_obj->name;
                my $response = $vol_obj->Filer->say ("priv set diag;snaplock clock status -v $vol_name");
                my @lines = split '\n', $response;
                my $line = $lines[@lines - 1];
                my @elements = split /\s+/, $line;
                my @keys = ('vol-name' , 'in-core-vcc-v' , 'on-disk-vcc-v' , 'on-disk-scc-v' , 'scc-id' , 'node-id' , 'vcc-skew');
                my $i = 0;
                my %cc_verbose_hash;
                foreach my $element (@elements){
                        $cc_verbose_hash{$keys[$i]} = $element;
                        $i++;
                }
                my $in_core_vcc = $cc_verbose_hash{'in-core-vcc-v'};$in_core_vcc =~ s/^\s+//;$in_core_vcc =~ s/\s+$//;
                my $on_disk_vcc = $cc_verbose_hash{'on-disk-vcc-v'};$on_disk_vcc =~ s/^\s+//;$on_disk_vcc =~ s/\s+$//;
                my $on_disk_scc = $cc_verbose_hash{'on-disk-scc-v'};$on_disk_scc =~ s/^\s+//;$on_disk_scc =~ s/\s+$//;
                $cc_verbose_hash{'in-core-vcc'} = UnixDate(ParseDate($in_core_vcc), "%s");
                $cc_verbose_hash{'on-disk-vcc'} = UnixDate(ParseDate($on_disk_vcc), "%s");
                $cc_verbose_hash{'on-disk-scc'} = UnixDate(ParseDate($on_disk_scc), "%s");
                print Dumper(\%cc_verbose_hash);
                return \%cc_verbose_hash;
}


