# $Id: //depot/prod/test/main/lib/uACL.pm# $
#
## @summary  uACL class
## @status   public
## @pod      here
## @author   dl-nb-cr-qa@netapp.com

=head1 NAME

uACL

=head1 DESCRIPTION

This is the class for uACL Objects.

=head1 METHODS

=cut

package uACL;

use strict;
use warnings;

use Tharn;

use nfs::NfsClient;
use Carp qw(croak);
use nfs::NfsDef;
use nfs::NfsXdr;
use Data::Dumper;
use NACL::C::Node;
use File::Basename;
use Params::Validate qw[validate OBJECT SCALAR HASHREF BOOLEAN ARRAYREF];
Params::Validate::validation_options( allow_extra => 1 );

use uACL::LinuxUtils;
use uACL::smbUtils;
use uACL::uACLSetup;
use uACL::SolarisUtils;

#List of valid client versions accepted
our $CLIENTS = {
    smb    => 'smb',
    lnx_v3 => 'linux_v3',
    lnx_v4 => 'linux_v4',
    sol_v3 => 'sol_v3',
    sol_v4 => 'sol_v4',
};

BEGIN {
    our @EXPORT_OK = qw(new
        create_files
        get_acl
        set_acl
        destroy
        create_basic_setup
    );
} ## end BEGIN

=head2 new

  my uACL_object = uACL->new( smb => $smb_hash_ref, nfsv3 => $v3_hash_ref );

  Constructor for uACL class.

  Following parameters are accepted:

=over

=item Arguments

=over

=item NODES

  Required: Cluster nodes against which tests have to be run

=item smb

  Optional: Hash reference to smb parameters needed to create and test over
  smb1.0 protocol

  Hash should contain the following parameters (All Mandatory):

    client   => HostName of the Windows Client
    server   => Name of the CIFS Server
    share    => CIFS Share to be used
    domain   => Domain to which filers & Clients are member of
    user     => CIFS User to be used for Mapping
    passwd   => Password of CIFS user for Mapping

=item nfsv3

  Optional: Hash reference to Linux NFSv3 parameters needed to create and test over
  NFSv3 protocol

  Hash should contain the following parameters (All Mandatory):

    data_lif      => Data LIF on filer to be used for NFSv3 Operations
    junction_path => Junction path to be mounted, defaults to "/NS"
    proto         => Protocol to be used, defaults to "tcp"

=item nfsv4

  Optional: Hash reference to Linux NFSv4 parameters needed to create and test over
  NFSv3 protocol

  Hash should contain the following parameters (All Mandatory):

    data_lif      => Data LIF on filer to be used for NFSv4 Operations
    junction_path => Junction path to be mounted, defaults to "/NS"
    proto         => Protocol to be used, defaults to "tcp"

=item solv4

  Optional: Hash reference to Solaris parameters needed to create and test over
  NFSv4 protocol

  Hash should contain the following parameters (All Mandatory):

    data_lif      => Data LIF on filer to be used for NFSv3 Operations
    junction_path => Junction path to be mounted, defaults to "/NS"
    proto         => Protocol to be used, defaults to "tcp"

=item solv3

  Optional: Hash reference to Solaris parameters needed to create and test over
  NFSv3 protocol

  Hash should contain the following parameters (All Mandatory):

    data_lif      => Data LIF on filer to be used for NFSv3 Operations
    junction_path => Junction path to be mounted, defaults to "/NS"
    proto         => Protocol to be used, defaults to "tcp"

=back

=item Returns:

  Success - uACL object for further operations
  Failure - undef

=back

=cut

sub new {
    my $class = shift;

    my %args = validate(
        @_,
        {   smb     => { type => HASHREF, optional => 1 },
            nfsv3   => { type => HASHREF, optional => 1 },
            nfsv4   => { type => HASHREF, optional => 1 },
            solv4   => { type => HASHREF, optional => 1 },
            solv3   => { type => HASHREF, optional => 1 },
            verbose => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
                regex    => qr/0|1/
            },
        }
    );

    my $self = {};
    $self->{_owner}  = $$;
    $self->{verbose} = $args{verbose};

    my @nd = NACL::C::Node->find();
    $self->{NODES} = \@nd;
    my $node_obj = host( $self->{NODES}->[0]->node );

    my $node_info = NACL::APISet::Version->new(
        hostobj    => $node_obj,
        category   => "Node",
        shell_type => "CMode"
    );

    $Log->comment( "ONTAP Build Release: "
            . $node_info->get_version_attribute( attribute => "node_version" )
    );

    ## CIFS
    $self->{smbclient}   = undef;
    $self->{test_on_smb} = 0;

    ## NFSv3
    $self->{v3_data_lif}      = $args{nfsv3}->{data_lif};
    $self->{v3_junction_path} = $args{nfsv3}->{junction_path} || "/NS";
    $self->{v3_proto}         = $args{nfsv3}->{proto} || "tcp";
    $self->{_v3_client}       = undef;
    $self->{_v3_mountpath}    = undef;
    $self->{test_on_v3}       = 0;

    ## NFSv4
    $self->{v4_data_lif}      = $args{nfsv4}->{data_lif};
    $self->{v4_junction_path} = $args{nfsv4}->{junction_path} || "/NS";
    $self->{v4_proto}         = $args{nfsv4}->{proto} || "tcp";
    $self->{_v4_client}       = undef;
    $self->{_v4_mountpath}    = undef;
    $self->{test_on_v4}       = 0;

    ## Solaris_v4
    $self->{solv4_data_lif}      = $args{solv4}->{data_lif};
    $self->{solv4_junction_path} = $args{solv4}->{junction_path} || '/NS';
    $self->{solv4_proto}         = $args{solv4}->{proto} || 'tcp';
    $self->{_sol_v4_client}      = undef;
    $self->{test_on_solv4}       = 0;

    ## Solaris_v3
    $self->{solv3_data_lif}      = $args{solv3}->{data_lif};
    $self->{solv3_junction_path} = $args{solv3}->{junction_path} || '/NS';
    $self->{solv3_proto}         = $args{solv3}->{proto} || 'tcp';
    $self->{_sol_v3_client}      = undef;
    $self->{test_on_solv3}       = 0;

    ## Get test platforms

    $self->{test_on_smb}   = 1 if ( defined( $args{smb}->{client} ) );
    $self->{test_on_v3}    = 1 if ( defined( $self->{v3_data_lif} ) );
    $self->{test_on_v4}    = 1 if ( defined( $self->{v4_data_lif} ) );
    $self->{test_on_solv4} = 1 if ( defined( $self->{solv4_data_lif} ) );
    $self->{test_on_solv3} = 1 if ( defined( $self->{solv3_data_lif} ) );

    bless( $self, $class );
    croak $@ if $@;

    ## Create CIFS object
    if ( $self->{test_on_smb} ) {
        $self->{smbclient} = uACL::smbUtils->new(
            smb   => $args{smb},
            debug => $self->{verbose},
            NODES => $self->{NODES}
        );

        unless ( defined( $self->{smbclient} ) ) {
            $Log->comment("Unable to create smb Object");
            return undef;
        }
    } ## end if ( $self->{test_on_smb...

    ## Create NFSv3 object
    if ( $self->{test_on_v3} ) {

        $self->{_v3_client} = uACL::LinuxUtils->new();
        if ( defined( $self->{_v3_client} ) ) {
            $self->{_v3_client}->mount(
                data_lif      => $self->{v3_data_lif},
                junction_path => $self->{v3_junction_path},
                version       => '3'
            );
        } else {
            return undef;
        }

    } ## end if ( $self->{test_on_v3...

    ## Create NFSv4 object
    if ( $self->{test_on_v4} ) {

        $self->{_v4_client} = uACL::LinuxUtils->new();
        if ( defined( $self->{_v4_client} ) ) {
            $self->{_v4_client}->mount(
                data_lif      => $self->{v4_data_lif},
                junction_path => $self->{v4_junction_path},
                version       => '4'
            );
        } else {
            return undef;
        }

    } ## end if ( $self->{test_on_v4...

    ## Create Solaris object for v4
    if ( $self->{test_on_solv4} ) {
        $self->{_sol_v4_client} = uACL::SolarisUtils->new();
        if ( defined( $self->{_sol_v4_client} ) ) {
            $self->{_sol_v4_client}->mount(
                data_lif      => $self->{solv4_data_lif},
                junction_path => $self->{solv4_junction_path},
                mount_version => '4'
            );
        } else {
            return undef;
        }
    } ## end if ( $self->{test_on_solv4...

    ## Create Solaris object for v3
    if ( $self->{test_on_solv3} ) {
        $self->{_sol_v3_client} = uACL::SolarisUtils->new();
        if ( defined( $self->{_sol_v3_client} ) ) {
            $self->{_sol_v3_client}->mount(
                data_lif      => $self->{solv3_data_lif},
                junction_path => $self->{solv3_junction_path},
                mount_version => '3'
            );
        } else {
            return undef;
        }
    } ## end if ( $self->{test_on_solv3...

    return $self;
} ## end sub new

sub destroy {
    my $self = shift;

    if ( defined( $self->{smbclient} ) ) {
        $self->{smbclient}->destroy();
        $Log->comment("Destroying the smb object");
        $self->{smbclient} = undef;
    }

    if ( $self->{test_on_v3} ) {
        $self->{_v3_client}->umount();
        $Log->comment("Destroying the NFSv3 Synthetic Client");
        $self->{_v3_client} = undef;
    }

    if ( $self->{test_on_v4} ) {
        $self->{_v4_client}->umount();
        $Log->comment("Destroying the NFSv4 Synthetic Client");
        $self->{_v4_client} = undef;
    }

    if ( defined $self->{_sol_v4_client} ) {
        $self->{_sol_v4_client}->umount();
        $Log->comment("Destroying the Solaris Client for v4");
        $self->{_sol_v4_client} = undef;
    }

    if ( defined $self->{_sol_v3_client} ) {
        $self->{_sol_v3_client}->umount();
        $Log->comment("Destroying the Solaris Client for v3");
        $self->{_sol_v3_client} = undef;
    }

    return 1;
} ## end sub destroy

=head2 create_files

    $result = $uacl_obj->create_files( version => "smb", count => 10 );

    $result = $obj->create_files(
            version  => "smb",
            path     => "folder\\win7",
            count    => $num,
            preserve => 1
        );

  Method to create files over different protocols through uACL class.

  Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Requried: Protocol version on which file has to be created

  Valid values are smb, linux_v3, linux_v4, sol_v3, sol_v4

=item path

  Optional: if file is to be create in additional path relative to
  mount/share. Share drive/mount point will be appended during file
  create.

=item count

  Optional: Number of files to be created, defaults to 1.

=item preserve

  Optional: If files created are not to be deleted at the end of the test.
  Default behavior is to delete them.

  Valid values are 0 or 1

=item oplock 

  Optional: Oplock files are created instead of normal files if this option is set
  Default : 0
  Returns : File handle which can be used to perform I/O operations

  Valid values are 0 or 1

=back

=item Returns:

  Success - scalar file name if single file is created,
            Array ref if more than one file is created
  Failure - undef

=back

=cut

sub create_files {
    my $self = shift;
    my %args = validate(
        @_,
        {   count    => { type => SCALAR, optional => 1, default => 1 },
            path     => { type => SCALAR, optional => 1 },
            preserve => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
                regex    => qr/0|1/
            },
            oplock   => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
                regex    => qr/0|1/
            },
            version => { type => SCALAR, optional => 0 },
        }
    );

    my $count    = $args{count};
    my $version  = $args{version};
    my $path     = $args{path};
    my $preserve = $args{preserve};
    my $oplock   = $args{oplock};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
       
        return $self->{smbclient}->CreateFiles_Lock(%args) if ($oplock);

        return $self->{smbclient}->CreateFiles(
            count    => $count,
            path     => $path,
            preserve => $preserve
        );
    } elsif ( ( $version eq $CLIENTS->{sol_v4} )
        && ( $self->{test_on_solv4} ) )
    {
        return $self->{_sol_v4_client}->create_file(%args);
    } elsif ( ( $version eq $CLIENTS->{sol_v3} )
        && ( $self->{test_on_solv3} ) )
    {
        return $self->{_sol_v3_client}->create_file(%args);
    } elsif ( ( $version eq $CLIENTS->{lnx_v3} ) && ( $self->{test_on_v3} ) )
    {
        return $self->{_v3_client}->create_file(%args);

    } elsif ( ( $version eq $CLIENTS->{lnx_v4} ) && ( $self->{test_on_v4} ) )
    {
        return $self->{_v4_client}->create_file(%args);

    } else {
        return $self->_client_version_error($version);
    }
} ## end sub create_files

sub create_dirs {
    my $self = shift;
    my %args = validate(
        @_,
        {   count    => { type => SCALAR,  default  => 1 },
            nested   => { type => BOOLEAN, default  => 'false' },
            path     => { type => SCALAR,  optional => 1 },
            version  => { type => SCALAR,  optional => 0 },
            preserve => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
                regex    => qr/0|1/
            },
        }
    );
    my $count    = $args{count};
    my $version  = $args{version};
    my $path     = $args{path};
    my $preserve = $args{preserve};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
        return $self->{smbclient}->CreateDirs(
            count    => $count,
            path     => $path,
            preserve => $preserve
        );
    } elsif ( ( $version eq $CLIENTS->{sol_v4} )
        && ( $self->{test_on_solv4} ) )
    {
        return $self->{_sol_v4_client}->create_dir(%args);
    } elsif ( ( $version eq $CLIENTS->{sol_v3} )
        && ( $self->{test_on_solv3} ) )
    {
        return $self->{_sol_v3_client}->create_dir(%args);
    } elsif ( ( $version eq $CLIENTS->{lnx_v3} ) && ( $self->{test_on_v3} ) )
    {
        return $self->{_v3_client}->create_dir(%args);

    } elsif ( ( $version eq $CLIENTS->{lnx_v4} ) && ( $self->{test_on_v4} ) )
    {
        return $self->{_v4_client}->create_dir(%args);

    } else {
        return $self->_client_version_error($version);
    }

} ## end sub create_dirs

sub rename() {

    #  use File::Basename;
    #takes path relative from mount point.

    my $self = shift;

    my %args = validate(
        @_,
        {   src_file_path => { type => ARRAYREF, optional => 0 },
            dst_file_path => { type => ARRAYREF, optional => 0 },

            #path  => { type => SCALAR, default => $self->mnt_path()} ,
            version => { type => SCALAR, optional => 0 },
        }
    );

    my $src_file_path = $args{src_file_path};
    my $dst_file_path = $args{dst_file_path};
    my $version       = $args{version};

    if ( ( $version eq $CLIENTS->{sol_v4} ) && ( $self->{test_on_solv4} ) ) {
        return $self->{_sol_v4_client}->move(%args);
    } elsif ( ( $version eq $CLIENTS->{sol_v3} )
        && ( $self->{test_on_solv3} ) )
    {
        return $self->{_sol_v3_client}->move(%args);
    } elsif ( defined( $self->{_v3_client} ) ) {

        return $self->{_v3_client}->move(%args);

    } elsif ( defined( $self->{_v4_client} ) ) {

        return $self->{_v4_client}->move(%args);

    } else {
        $Log->comment('No client type is available');
        return 0;
    }

} ## end sub rename()

sub move() {

    my $self = shift;

    my %args = validate(
        @_,
        {   src_file_path => { type => ARRAYREF, optional => 0 },
            dst_file_path => { type => ARRAYREF, optional => 0 },
            version       => { type => SCALAR,   optional => 0 },
        }
    );

    my $src_file_path = $args{src_file_path};
    my $dst_file_path = $args{dst_file_path};
    my $version       = $args{version};

    if ( ( $version eq $CLIENTS->{sol_v4} ) && ( $self->{test_on_solv4} ) ) {
        return $self->{_sol_v4_client}->move(%args);
    } elsif ( ( $version eq $CLIENTS->{sol_v3} )
        && ( $self->{test_on_solv3} ) )
    {
        return $self->{_sol_v3_client}->move(%args);
    } elsif ( defined( $self->{_v3_client} ) ) {

        return $self->{_v3_client}->move(%args);

    } elsif ( defined( $self->{_v4_client} ) ) {

        return $self->{_v4_client}->move(%args);

    } else {
        $Log->comment('No client type is available');
        return 0;
    }

} ## end sub move()

sub copy() {

    my $self = shift;

    my %args = validate(
        @_,
        {   src_file_path => { type => ARRAYREF, optional => 0 },
            dst_file_path => { type => ARRAYREF, optional => 0 },
            version       => { type => SCALAR,   optional => 0 },
        }
    );

    my $src_file_path = $args{src_file_path};
    my $dst_file_path = $args{dst_file_path};
    my $version       = $args{version};

    if ( ( $version eq $CLIENTS->{sol_v4} ) && ( $self->{test_on_solv4} ) ) {
        return $self->{_sol_v4_client}->copy(%args);
    } elsif ( ( $version eq $CLIENTS->{sol_v3} )
        && ( $self->{test_on_solv3} ) )
    {
        return $self->{_sol_v3_client}->copy(%args);
    } elsif ( defined( $self->{_v3_client} ) ) {

        return $self->{_v3_client}->copy(%args);

    } elsif ( defined( $self->{_v4_client} ) ) {

        return $self->{_v4_client}->copy(%args);

    } else {
        $Log->comment('No client type is available');
        return 0;
    }

} ## end sub copy()

sub delete {
    my ( $self, %args ) = @_;
    if ( defined( $self->{_sol_v3_client} ) ) {
        return $self->{_sol_v3_client}->delete(%args);
    } elsif ( defined( $self->{_sol_v4_client} ) ) {

        return $self->{_sol_v4_client}->delete(%args);

    } elsif ( defined( $self->{_v3_client} ) ) {

        return $self->{_v3_client}->delete(%args);

    } elsif ( defined( $self->{_v4_client} ) ) {

        return $self->{_v4_client}->delete(%args);

    } else {
        $Log->comment('No client type is available');
        return 0;
    }
} ## end sub delete

=head2 get_acl

    $result = $uacl_obj->get_acl( version => "smb", path => "z:\\testfile" );

    Method to get ACLs on a file over different protocols through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Requried: Protocol version on which file has to be created

  Valid values are 'smb','linux_v3','linux_v4','sol_v3','sol_v4'.

=item path

  Required: Absolute path to the file

=back

=item Returns:

  Success - hash ref to the ACL structure for the file
  Failure - undef

=back

=cut

sub get_acl() {

    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 0 }, } );
    my $version = $args{version};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
        return $self->{smbclient}->getACL(%args);

    } elsif ( ( $self->{test_on_solv4} )
        && ( $version eq $CLIENTS->{sol_v4} ) )
    {
        return $self->{_sol_v4_client}->getacl(%args);
    } elsif ( ( $self->{test_on_solv3} )
        && ( $version eq $CLIENTS->{sol_v3} ) )
    {
        return $self->{_sol_v3_client}->getacl(%args);
    } elsif ( ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->getacl(%args);
    } elsif ( ( $self->{test_on_v4} )
        && ( $version eq $CLIENTS->{lnx_v4} ) )
    {
        return $self->{_v4_client}->getacl(%args);
    } else {
        return $self->_client_version_error($version);
    }
} ## end sub get_acl()

=head2 verify_acl

    $result = $uacl_obj->verify_acl( version => "smb", path => "z:\\testfile", identity  => 'Everyone', has_allow => 'Delete');

    Method to get ACLs on a file over different protocols through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Requried: Protocol version on which file has to be created

  Valid values are 'smb','linux_v3','linux_v4','sol_v3','sol_v4'.

=item path

  Required: Absolute path to the file

=item identity

  Optional: Identity i.e. username or groupname which has to be verified

=item has_allow

  Optional: Allow right that have to be checked for

=item has_deny

  Optional: Allow right that have to be checked for

=item has_owner

  Optional: Owner of the file

=item has_group

  Optional: Group of the file

  Although all the fields are optional, at least one of the has_ field has to be specified
  for verfication

=back

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub verify_acl() {

    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 1 }, } );
    my $version = $args{version};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
        return $self->{smbclient}->VerifyACL(%args);

    } elsif ( ( $self->{test_on_solv4} )
        && ( $version eq $CLIENTS->{sol_v4} ) )
    {
        return $self->{_sol_v4_client}->verifyacls(%args);
    } elsif ( ( $self->{test_on_solv3} )
        && ( $version eq $CLIENTS->{sol_v3} ) )
    {
        return $self->{_sol_v3_client}->verifyacls(%args);
    } elsif ( ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->verifyacls(%args);
    } elsif ( ( $self->{test_on_v4} )
        && ( $version eq $CLIENTS->{lnx_v4} ) )
    {
        return $self->{_v4_client}->verifyacls(%args);
    } else {
        return $self->_client_version_error($version);
    }
} ## end sub verify_acl()

=head2 set_acl

    $result = $uacl_obj->verify_acl( version => "smb", 
                                     path => "z:\\testfile", 
                                     identity    => 'crgdl2k8\saurabhg',
                                     deny_rights => 'write_owner', 
                                     verify      => 1, );

    Method to set ACLs on a file over different protocols through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Requried: Protocol version on which file has to be created

  Valid values are 'smb','linux_v3','linux_v4','sol_v3','sol_v4'.

=item path

  Required: Absolute path to the file

=item identity

  Optional: Identity i.e. username or groupname which has to be verified

=item allow_rights

  Optional: Rights to be granted on the file

=item deny_rights

  Optional:  Rights to be denied on the file
  
=item inheritance

  Optional: To set the inheritance flags
  
            Valid Values are:  
                 so: sub-objects
                 sc: sub-containers
                 np: no propagation
                 io: inherit only

            Default value: np
              
=item owner

  Optional: Owner to be set on the file

=item group

  Optional: Group to be set on the file

=item replace

  Optional:  To replace the existing ACLs on the file. Default behavior
             is to add to existing ACLs

             Valid values are 0 or 1

=item verify

  Optional:  To verify the ACLs after setting them. Default behavior is
             is to perform a verfication by doing a getACL on the file

             Valid values are 0 or 1

=back

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub set_acl() {

    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 1 }, } );
    my $version = $args{version};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
        return $self->{smbclient}->SetACL(%args);

    } elsif ( ( $self->{test_on_solv4} )
        && ( $version eq $CLIENTS->{sol_v4} ) )
    {
        return $self->{_sol_v4_client}->setacl(%args);
    } elsif ( ( $self->{test_on_solv3} )
        && ( $version eq $CLIENTS->{sol_v3} ) )
    {
        return $self->{_sol_v3_client}->setacl(%args);
    } elsif ( ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->setacl(%args);
    } elsif ( ( $self->{test_on_v4} )
        && ( $version eq $CLIENTS->{lnx_v4} ) )
    {
        return $self->{_v4_client}->setacl(%args);
    } else {
        return $self->_client_version_error($version);
    }

} ## end sub set_acl()

##### Private Functions ###

sub v3_client {
    return shift->{_v3_client};
}

sub v4_client {
    return shift->{_v4_client};
}

sub smbdrive {
    my $self = shift;
    if ( defined( $self->{smbclient} ) ) {
        return $self->{smbclient}->{mapdrive};
    } else {
        return undef;
    }
} ## end sub smbdrive

sub sol_v4_client {
    return shift->{_sol_v4_client};
}

sub sol_v3_client {
    return shift->{_sol_v3_client};
}

sub v3_mountpath {
    return shift->{_v3_mountpath};
}

sub v4_mountpath {
    return shift->{_v4_mountpath};
}

# sub _mount_nfs_client {

# my ( $self, %args ) = @_;
# my $nfs_obj;
# if ( defined( $self->{_v3_client} ) ) {

# return $self->{_v3_client}->mount_nfs(%args) ? 1 : 0;
# } else {
# return $self->{_v4_client}->mount_nfs(%args) ? 1 : 0;
# }

# } ## end sub _mount_nfs_client

sub _client_version_error {
    my ( $self, $ver ) = @_;
    $Log->error("Unexpected combination of client version $ver");
    $Log->comment("Valid list of combinations is:\n");
    $Log->comment($_) foreach ( values(%$CLIENTS) );
    return undef;
} ## end sub _client_version_error

sub add_user {
    my $self = shift;

    my %args = validate(
        @_,
        {   version => { type => SCALAR,   optional => 0 },
            user    => { type => ARRAYREF, optional => 0 },
        }
    );
    my $version = $args{version};
    my $user    = $args{user};

    if ( ( $self->{test_on_solv4} ) && ( $version eq $CLIENTS->{sol_v4} ) ) {
        return $self->{_sol_v4_client}->useradd( user => $user ) ? 1 : 0;
    } elsif ( ( $self->{test_on_solv3} )
        && ( $version eq $CLIENTS->{sol_v3} ) )
    {
        return $self->{_sol_v3_client}->useradd( user => $user ) ? 1 : 0;
    } elsif ( ( $self->{test_on_v4} ) && ( $version eq $CLIENTS->{lnx_v4} ) ) {
        return $self->{_v4_client}->useradd( user => $user ) ? 1 : 0;
    } elsif ( ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->useradd( user => $user ) ? 1 : 0;
    }
} ## end sub add_sol_user

sub add_group {
    my $self = shift;

    my %args = validate(
        @_,
        {   version => { type => SCALAR,   optional => 0 },
            group   => { type => ARRAYREF, optional => 0 },
        }
    );
    my $version = $args{version};
    my $group   = $args{group};

    if ( ( $self->{test_on_solv4} ) && ( $version eq $CLIENTS->{sol_v4} ) ) {
        return $self->{_sol_v4_client}->groupadd( group => $group ) ? 1 : 0;
    } elsif ( ( $self->{test_on_solv3} )
        && ( $version eq $CLIENTS->{sol_v3} ) )
    {
        return $self->{_sol_v3_client}->groupadd( group => $group ) ? 1 : 0;
    } elsif ( ( $self->{test_on_v4} ) && ( $version eq $CLIENTS->{lnx_v4} ) ) {
        return $self->{_v4_client}->groupadd( group => $group ) ? 1 : 0;
    } elsif ( ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->groupadd( group => $group ) ? 1 : 0;
    }
	
} ## end sub add_group

sub max_nfs_aces() {

    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 1 }, } );
    my $version = $args{version};

   if ( ( $self->{test_on_v4} )
        && ( $version eq $CLIENTS->{lnx_v4} ) )
    {
        return $self->{_v4_client}->max_nfs_aces(%args);
    } else {
        return $self->_client_version_error($version);
    }
} ## end sub verify_acl()

=head2 create_basic_setup

    Method to create name-mapping/group-mapping for testing through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item vserver

  Optional: Vserver for which these mappings need to be created

  Defaults to "vs0"

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub create_basic_setup() {
    my $self = shift;
    my %args = validate( @_,
        { vserver => { type => SCALAR, optional => 1, default => 'vs0' }, } );
    my $vserver      = $args{vserver};
    my $return_value = 1;
    $self->uACLSetup::useUACLSetup();
    unless ( $self->uACLSetup::create_unix_groups(%args) ) {
        $return_value = 0;
    }
    if ( $return_value == 1 ) {
        unless ( $self->uACLSetup::add_unix_users(%args) ) {
            $return_value = 0;
        }
    }
    if ( $return_value == 1 ) {
        unless ( $self->uACLSetup::create_local_cifs_groups(%args) ) {
            $return_value = 0;
        }
    }
    if ( $return_value == 1 ) {
         unless ( $self->uACLSetup::create_local_cifs_users(%args) ) {
            $return_value = 0;
        }
    }
    if ( $return_value == 1 ) {
        unless ( $self->uACLSetup::verver_group_mapping_create(%args) ) {
            $return_value = 0;
        }
    }
    if ( $return_value == 1 ) {
        unless ( $self->uACLSetup::vserver_name_mapping_create(%args) ) {
            $return_value = 0;
        }
    }
    return $return_value;
} ## end sub create_basic_setup()

=head2 chmod

    $result = $uacl_obj->chmod( version => "linux_v4", mode => 644 );

    Method to change mode bits over NFS through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Requried: Protocol version on which file has to be created

  Valid values are 'smb','linux_v3','linux_v4','sol_v3','sol_v4'.

=item mode

  Required: Mode bits to be set. Valid formats are:

            0644 or 777 (Sticky bit is optional)

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub chmod {
    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 0 }, } );
    my $version = $args{version};

    if (   ( $self->{test_on_v3} )
        && ( $version eq $CLIENTS->{lnx_v3} ) )
    {
        return $self->{_v3_client}->chmod(%args);
    } elsif ( ( $self->{test_on_v4} )
        && ( $version eq $CLIENTS->{lnx_v4} ) )
    {
        return $self->{_v4_client}->chmod(%args);
    } else {
        return $self->_client_version_error($version);
    }

} ## end sub chmod

=head2 get_nfs_sid

    $sid = $uacl_obj->get_nfs_sid( username => "uaclnfsuser10" );

    Method to fetch the corresponding SID for a given NFS user.

    Following parameters are accepted:

=over

=item Arguments

=over

=item username

  Optional: NFS username

=item groupname

  Optional: NFS groupname
  
 For this method either user or group must be supplied.
 
=item Returns:

  Success - Corresponding SID
  Failure - undef

=back

=cut

sub get_nfs_sid {
    my $self = shift;
    my %args = validate(
        @_,
        {   username  => { type => SCALAR, optional => 1 },
            groupname => { type => SCALAR, optional => 1 },
        }
    );

    if ( defined( $args{username} ) ) {

        return $self->{smbclient}->_get_nfs_sid( $args{username}, 0 );
    } elsif ( defined( $args{groupname} ) ) {
        return $self->{smbclient}->_get_nfs_sid( $args{groupname}, 1 );

    } else {
        $Log->comment(
            "uACL::get_nfs_sid: Neither NFS user or NFS group is specified to fetch SID"
        );
        return undef;
    } ## end else [ if ( defined( $args{username...
} ## end sub get_nfs_sid

=head2 create_localuser_setup

    $uacl_obj->create_localuser_setup( vserver => "vs0", count => 300 );

    Method to create local NFS users, CIFS users & corresponding mapping

    Following parameters are accepted:

=over

=item Arguments

=over

=item vserver

  Required: Vserver for which mapping needs to be created

=item count

  Optional: Number of local users to be created
  
   By default method will create 500 NFS & 500 CIFS users respectively
   
=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub create_localuser_setup() {
    my $self = shift;
    my %args = validate(
        @_,
        {   vserver => { type => SCALAR, optional => 1, default => 'vs0' },
            count   => {
                type     => SCALAR,
                optional => 1,
                default  => 10,
                regex    => qr/\d+/
            },
        }
    );

    $self->uACLSetup::useUACLSetup();

    return 0 unless ( $self->uACLSetup::create_local_cifs_users(%args) );

    return 0 unless ( $self->uACLSetup::create_unix_groups(%args) );

    return 0 unless ( $self->uACLSetup::add_unix_users(%args) );

    #return 0 unless ( $self->uACLSetup::verver_group_mapping_create(%args) );

    return 0 unless ( $self->uACLSetup::vserver_name_mapping_create(%args) );

    return 1;
} ## end sub create_localuser_setup()

=head2 copy_acl

    $result = $uacl_obj->copy_acl( version => "smb", from => "z:\\sourcefile", to => "y:\desitnation" );

    Method to copy ACLs from one file to another through uACL class.

    Following parameters are accepted:

=over

=item Arguments

=over

=item version

  Optional: Protocol version on which file has to be created

  Valid values are 'smb'. 
  (This option is being given for future enhancements)

=item from

  Required: File or directory whose ACLs are to be copied

=item to

  Required: File or directory on which copied ACLs have to be applied
  
=item verify

  Optional: Verify the copied ACL on destination file
    
=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub copy_acl {
    my $self = shift;

    my %args = validate(
        @_,
        {   version => { type => SCALAR, optional => 1, default => "smb" },
            from    => { type => SCALAR, optional => 0 },
            to      => { type => SCALAR, optional => 0 },
        }
    );

    return $self->{smbclient}->CopyACL(%args);

} ## end sub copy_acl

=head2 add_wellknown_sids

    $uacl_obj->add_wellknown_sids( path => "z:\\somefile" );
    
    $uacl_obj->add_wellknown_sids( path => "z:\\somefile", nfs_only => 1 );

    Method to create ACEs for well known SIDs

    Following parameters are accepted:

=over

=item Arguments

=over

=item path

  Required: CIFS path of the file where ACEs are to be set

=over

=item nfs_only

  Optional: Add only NFS well known ACEs (Full Control to OWNER@, GROUP@ & EVERYONE@)

=over  
  
=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub add_wellknown_sids {
    my $self = shift;
    my %args = validate( @_, { path => { type => SCALAR, optional => 0 }, } );

    return $self->{smbclient}->_add_wellknown_sids(%args);
} ## end sub add_wellknown_sids

=head2 is_present

    $uacl_obj->is_present( path => "z:\\somefile" );
    
    Method to check if a file exists

    Following parameters are accepted:

=over

=item Arguments

=over

=item path

  Required: CIFS path of the file where ACEs are to be set

=over

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub is_present {
    my $self = shift;
    my %args = validate( @_, { path => { type => SCALAR, optional => 0 }, } );

    return $self->{smbclient}->_is_present(%args);
} ## end sub is_present

=head2 raw_command_interface

    $uacl_obj->raw_command_interface( command => "dir z:\\somefile",
                                      version => smb );

    Method to execute the raw command by ci or powershell connection

    Following parameters are accepted:

=over

=item Arguments

=over

=item command

  Required: Raw command to be executed in the client

=item version

  Required: Protocol version on which file has to be created

  Valid values are 'smb'.
  (This option is being given for future enhancements)

=item timeout

  Optional: Timeout value to be set for the connection established
  Default : 600 for powershell & 60 for ci conn

=item verify

  Optional: This flag is to mention whether need to validate the output by this function itself or not
  Default : 0

=item match_string

  Optional: Pattern to be matched from the output when verify flag is set

=item Returns:

  Success - 1
  Failure - 0

=back

=cut

sub raw_command_interface {
    my $self = shift;

    my %args
        = validate( @_, { version => { type => SCALAR, optional => 0 }, } );
    my $version = $args{version};

    if ( ( $version eq $CLIENTS->{smb} ) && ( $self->{test_on_smb} ) ) {
        return $self->{smbclient}->execute_cmd(%args);
    }
} ## end sub raw_command_interface

1;
