# $Id://depot/prod/test/main/lib/uACL/LinuxUtils.pm#1 $
#
# Copyright (c) 2001-2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary LinuxUtils - Class for Unified ACLs on Linux Synthetic client
## @author yadag@netapp.com
## @status private
## @pod here

=head1 NAME

uACL::LinuxUtils Linux ACL Class

=head1 DESCRIPTION

C<uACL::LinuxUtils> provides methods for ACL tests on Linux platform.

=head1 METHODS

=cut

package uACL::LinuxUtils;

use strict;
use warnings;

use Tharn qw($Log);
use Params::Validate qw( validate_with ARRAYREF BOOLEAN HASHREF SCALAR );

use uACL::uACLCommonParams;
use Data::Dumper;
use NACL::C::Client;
use NACL::STask::Client::Mount;
use Options;

#constants

use constant SETFACL => '/usr/bin/nfs4_setfacl';
use constant GETFACL => '/usr/bin/nfs4_getfacl';

# use NACL::APISet::Exceptions::CommandFailedException qw( :try );
use nfs::ACL::V4tClient qw(get_acl create_with_acl set_acl);
Params::Validate::validation_options(allow_extra => 1);
use NACL::CS::VserverNFS;

=head2 new

  my $uacl = uACL::LinuxUtils->new();

  Creates an object of type uACL::LinuxUtils. No arguments required.

=cut

our ($REGEX, $BOOLEAN_VALUES, $NUMERICS, $ELEMENTS, $SETACL_RIGHTS,
    $GETACL_RIGHTS, $FLAGS, $Options);

$Options = new Options();

sub new {
    $Log->enter();

    my $class = shift;
    my $self  = {};

    _init_globals();
    $self->{linux} = _create_nfs_client();
    bless($self, $class);
    $Log->exit();

    return $self;
} ## end sub new

=head2 mount

  my %args = (
      'data-lif' => 'nfs_data_lif',
      'junction-path' => '/NS',
      'mount-point' => '/t/mnt',
      'mount-version' => 4,
  );
  $uacl->mount( %args );

  Mounts the juction path of an infinite volume the specified mount point.

  Following parameters are accepted:

=over

=item Arguments

=over

=item data-lif

  Required: IP address of a Data LIF associated with repository vserver.

=item junction-path

  Required: Junction path of the infinite volume

=item mount-point

  Optional: Mount point to which junction path of an infinite volume will be mount.
            If not passed then /t/<ClientHostname>/NS_version will be used.

=item version

  Optional: Version 4 is Default.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub mount {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            data_lif      => {type => SCALAR},
            junction_path => {type => SCALAR},
            version       => {type => SCALAR},
        },
    );
    my $apiset        = $self->{linux};
    my $data_lif      = $args{data_lif};
    my $junction_path = $args{junction_path};
    my $version       = $args{version};
    my $mount_point
        = '/t/' . $apiset->{hostobj}->{hostname} . '/NS_V' . $version . $$;
    my $result = 1;
    my ($output, $type, $options);

    ## Make mount point directory
    $Log->step(
        "uACL::LinuxUtils::mount: Creating mount directory [$mount_point]");
    unless ($apiset->mkdir('make-parents' => 1, 'paths' => $mount_point)) {
        $Log->error(
            "uACL::LinuxUtils::mount: ERROR: Unable to create the mount directory"
        );
        return 0;
    } ## end unless ( $apiset->mkdir( 'make-parents'...
    ## Mount Infinite Volume on the linux client
    $Log->step("Mount Infinite Volume junction path [$junction_path]"
            . " to mount point [$mount_point]");

    ($version == 4)
        ? ($type = 'nfs' . $version, $options = 'noac,rw,minorversion=1')
        : ($type = 'nfs', $options = 'noac,rw,vers=3');
    eval {
        $output = $apiset->mount(
            'file-system-type' => $type,
            'source-path'      => $data_lif . ':' . $junction_path,
            'mount-point'      => $mount_point,
            'options'          => $options,
        );
    };
    unless (defined($output)) {
        $Log->error(
            "uACL::LinuxUtils::mount: ERROR: Could not mount Infinite Volume's junction path"
                . "\nReason: "
                . $@
                . "\n\n");
        $result = 0;
    } ## end unless ( defined($output) )

    ## Change permission
    $Log->step(
        "uACL::LinuxUtils::mount:  Change permission [0777] of the directory [$mount_point]"
    );
    unless ($apiset->chmod(mode => '0777', paths => $mount_point)) {
        $Log->error(
            "uACL::LinuxUtils::mount: ERROR: Unable to change the DIR permission to [0777]"
        );
        $result = 0;
    } ## end unless ( $apiset->chmod( mode...
    ## Store the mountpoint
    $self->{'mount_point'} = $mount_point;

    $Log->exit();
    return $result;
} ## end sub mount

=head2 umount

  $uacl->umount( 'mount-point' => '/t/mnt' );

  Unmounts the juction path of an infinite volume from the specified mount point.

  Following parameters are accepted:

=over

=item Arguments

=over

=item mount-point

  Optional: Mount point to which juction path of an infinite volume was mounted.

=item umount

  Optional: Specify umount is required or not. Default value is 'true'.

=item delete

  Optional: Specify mount point needs to be deleted or not.
            Default value is 'true'.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub umount {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            'mount-point' =>
                {type => SCALAR, default => $self->{'mount_point'}},
            umount => {type => SCALAR, default => 'true'},
            delete => {type => SCALAR, default => 'true'},
        }
    );
    my $apiset        = $self->{linux};
    my $mount_point   = $args{'mount-point'};
    my $umount        = $args{umount};
    my $delete        = $args{delete};
    my $result        = 1;
    my $result_umount = 1;
    my ($umount_res, $mount_dir);

    ## Unmount Infinite Volume from the linux client
    if ($umount eq 'true') {

        $Log->step(
            "uACL::LinuxUtils::umount: Unmount Infinite Volume from mount point [$mount_point]"
        );
        eval { $umount_res = $apiset->umount('paths' => $mount_point,); };
        unless (defined($umount_res)) {
            $Log->error(
                "\n uACL::LinuxUtils::umount: ERROR: Could not unmount Infinite Volume's junction path"
                    . "\nReason:"
                    . $@
                    . "\n\n");
            $result        = 0;
            $result_umount = 0;
        } ## end unless ( defined($umount_res...
    } ## end if ( $umount eq 'true')

    ## Remove the mount point directory
    if ($delete eq 'true' && $result_umount == 1) {
        $Log->step(
            "uACL::LinuxUtils::umount: Delete mount point directory [$mount_point]"
        );
        eval { $mount_dir = $apiset->rmdir(paths => $mount_point); };
        unless (defined($mount_dir)) {
            $Log->error(
                "\n uACL::LinuxUtils::umount: ERROR: Could not Delete mount point"
                    . "\nReason: "
                    . $@
                    . "\n\n");
            $result = 0;
        } ## end unless ( defined($mount_dir...
    } ## end if ( $delete eq 'true'...
    $Log->exit();
    return $result;
} ## end sub umount

=head2 create_file

my %args = ( path => $dir, count => 5 );
$uacl->create_file( %args ) 


Create file(s) under a specified path.
Following parameters are accepted:

=over

=item Arguments

=over

=item count

Optional: Number of files to create. If not passed only one file will be  created by default.
Log
=item path

Optional: Path under which file(s) need to be created. If not passed file(s)
            will be created under the mount point. Assumes that path that mentoined
			already existed.

=back

=item Return:

  Success - Array of file names or a single file name.
  Failure - 0

=back

=cut

sub create_file() {

    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            count => {type => SCALAR, optional => 1, default => $NUMERICS},
            path  => {type => SCALAR, optional => 1},
        },
    );
    my $count      = $args{count};
    my $path       = $args{path};
    my $apiset     = $self->{linux};
    my $mountpoint = $self->{'mount_point'};
    my $result     = 1;
    my ($file, @files, $filename);

    for (my $i = 0; $i < $count; $i++) {
        if (defined($path)) {
            $file     = "nfs_file_$i" . time() . ".txt";
            $filename = $mountpoint . "/" . $path . "/" . $file;

            # $filename =~ $REGEX;    # removing "if multiple slashes"
        } else {
            $file     = "nfs_file_$i" . time() . ".txt";
            $filename = $mountpoint . "/" . $file;
        }
        push @files, $file;
        ## Create file
        $Log->step("uACL::LinuxUtils::create_file: Created file   $filename");
        unless ($apiset->touch('path' => $filename)) {
            $Log->error(
                "\n uACL::LinuxUtils::create_file: ERROR: Could not create file [$filename] \n\n"
            );
            return 0;
        } ## end unless ( $apiset->touch( 'path'...
        $path = $args{path};
    } ## end for ( my $i = 0; $i < $count...

    $Log->exit();
    return ($count == 1) ? $files[0] : \@files;
} ## end sub create_file()

=head2 create_dir
my %args = ( path => $dir, count => 5 );
$uacl->create_dir( %args );

 Create directories under the mount point.  

  Following parameters are accepted:    
  =over
  =item Arguments
  =over

=item count
  Optional: Number of directories to create. If not passed only one directory
            will be created by default.
			
=item nested  

Optional: Default value is 'false'. If passed as 'true' nested directories    will be created (e.g. /t/mnt/dir1/dir2/dir3/)

=back

=item Return:

 Success - Array of directory names or a directory name.
  Failure - 0

=back

=cut

sub create_dir {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            count  => {type => SCALAR,  default  => $NUMERICS, optional => 1},
            nested => {type => BOOLEAN, default  => 'false',   optional => 1},
            path   => {type => SCALAR,  optional => 1},
        },
    );
    my $count      = $args{count};
    my $nested     = $args{nested};
    my $path       = $args{path};
    my $apiset     = $self->{linux};
    my $mountpoint = $self->{'mount_point'};
    my ($dir, @dirs, $dirname) = undef;

    if ($nested eq 'false') {
        ## Create directories under the root mount point
        for (my $i = 1; $i <= $count; $i++) {
            if (defined($path)) {
                $dir     = 'uacl_dir' . time() . $i;
                $dirname = $mountpoint . "/" . $path . "/" . $dir;

                # $dirname =~ $REGEX;    # removing "if multiple slashes"
            } else {

                $dir     = 'uacl_dir' . time() . $i;
                $dirname = $mountpoint . "/" . $dir;
            }
            push @dirs, $dir;
            $Log->step(
                "uACL::LinuxUtils::create_dir: Created DIR   $dirname");
            unless (
                $apiset->mkdir(
                    'make-parents' => 1,
                    'paths'        => "$dirname"
                )
                ) {
                $Log->error(
                    "\n uACL::LinuxUtils::create_dir: ERROR: Could not create DIR [$dirname] \n\n"
                );
                return 0;
            } ## end unless ( $apiset->mkdir( 'make-parents'...
            $path = $args{path};
        } ## end for ( my $i = 1; $i <= ...
    } else {
        ## Create nested directories
        for (my $i = 1; $i <= $count; $i++) {
            $dir .= '/dir' . time() . $i;
        }
        $dir =~ s/^\///g;
        push @dirs, $dir;
        $Log->step(
            "uACL::LinuxUtils::create_dir: Created DIR $mountpoint/$dir");
        unless (
            $apiset->mkdir(
                'make-parents' => 1,
                'paths'        => "$mountpoint/$dir"
            )
            ) {
            $Log->error(
                "\n uACL::LinuxUtils::create_dir: ERROR: Could not create DIR [$mountpoint/$dir] \n\n"
            );
            return 0;
        } ## end unless ( $apiset->mkdir( 'make-parents'...

    } ## end else [ if ( $nested eq 'false')

    $Log->exit();
    return ($count == 1) ? $dirs[0] : \@dirs;
} ## end sub create_dir

=head2 move

  Move a file to a different directory:
  $uacl->move( source => [ /t/mnt/file1 ], destination => [ /t/mnt/dir1/ ] );

  Rename a file:
  $uacl->move( source => [ /t/mnt/file1 ], destination => [ /t/mnt/file2 ] );

  Move or rename multiple file(s)/directory(s):
  my %args = (
      source => [ /t/mnt/file1, /t/mnt/file3 ],
      destination => [ /t/mnt/file2, /t/mnt/dir2 ]
  );
  $uacl->move( %args );

  This method moves or renames the file named by the source argument to the
  destination specified by the destination argument. The destination argument
  can be a file or directory. If the destination file does not exist then this
  method creates it.

  Following parameters are accepted:

=over

=item Arguments

=over

=item source

  Required: ArrayRef of file names that need to be moved to a new location
            or renamed.

=item destination

  Required: ArrayRef of file or directory names to which the source file needs
            be moved or renamed.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub move {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            src_file_path => {type => ARRAYREF},
            dst_file_path => {type => ARRAYREF},
        },
    );
    my $source      = $args{src_file_path};
    my $destination = $args{dst_file_path};
    my $apiset      = $self->{linux};
    my $result      = 1;
    my $move;

    ## Move file/directory
    $Log->step("uACL::LinuxUtils::move: Move files from SRC to DST");
    for (my $i = 0; $i < @$source; $i++) {
        eval {
            $move = $apiset->mv(
                source => $source->[$i],
                target => $destination->[$i]
            );
        };
        unless (defined($move)) {
            $Log->error(
                "\n uACL::LinuxUtils::move: ERROR: Could not Move files from SRC to DST \n\n"
                    . $@);
            return 0;
        } ## end unless ( defined($move) )
    } ## end for ( my $i = 0; $i < @$source...

    $Log->exit();
    return $result;
} ## end sub move

=head2 copy

  Copy a file to a different directory:
  $uacl->copy( src_file_path => [ /t/mnt/file1 ], dest_file_path => [ /t/mnt/dir1/ ] );

  Copy a file to a different file:
  $uacl->copy( src_file_path => [ /t/mnt/file1 ], dest_file_path => [ /t/mnt/file2 ] );

  Copy a dir to a different dir:
  $uacl->copy( src_file_path => [ /t/mnt/dir1 ], dest_file_path => [ /t/mnt/dir2 ] );

  $uacl->copy( %args );

  This method copies the file named by the source argument to the destination
  specified by the destination argument.If the destination file does not exist
  then this method creates it. If destination file exists then it overwrites it.
  If destination is a directory then it should exists.

  Following parameters are accepted:

=over

=item Arguments

=over

=item source

  Required: ArrayRef of file names that need to be moved to a new location
            or renamed.

=item destination

  Required: ArrayRef of file or directory names to which the source file needs
            be moved or renamed.

=item recursive

  Optional: Scalar value which allows to copy directory and its file to other
            directory.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub copy {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            src_file_path => {type => ARRAYREF},
            dst_file_path => {type => ARRAYREF},
        },
    );
    my $source      = $args{src_file_path};
    my $destination = $args{dst_file_path};
    my $apiset      = $self->{linux};
    my $result      = 1;
    my $copy;
    ## Copy file/directory
    $Log->step("uACL::LinuxUtils::copy: Copy files from SRC to DST");
    for (my $i = 0; $i < @$source; $i++) {
        eval {
            $copy = $apiset->cp(
                source => $source->[$i],
                dest   => $destination->[$i]
            );
        };
        unless (defined($copy)) {
            $Log->error(
                "\n uACL::LinuxUtils::copy: ERROR: Could not Copy files from SRC to DST \n\n"
                    . $@);
            return 0;
        } ## end unless ( defined($copy) )
    } ## end for ( my $i = 0; $i < @$source...

    $Log->exit();
    return $result;
} ## end sub copy

=head2 delete

  Delete a file:
  $uacl->delete( path => [ /t/mnt/file1 ] );

  Delete multiple files:
  $uacl->delete( path => [ /t/mnt/file1, /t/mnt/file2 ] );

  Delete a directory:
  $uacl->delete( path => [ /t/mnt/dir1 ] );

  This method removes the file(s) or directory(ies) passed by the paths argument.
  By default this method delete file(s) or directory(ies) forcefully.

  Following parameters are accepted:

=over

=item Arguments

=over

=item paths

  Required: ArrayRef of file or directory names that need to be removed.

=item type

  Optional: Type of file to be deleted whether directory or file.
            Default value is file.

=item force

  Optional: Enables forceful deletion. Default value is 'true'.

=item recursive

  Optional: Recursive deletion. Must pass a true value while deleting a
            directory. Default value is 'false'.

=item parent

  Optional: Removes dir and its parent directories also.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub delete {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            paths  => {type => ARRAYREF},
            type   => {type => SCALAR, default => 'file'},
            parent => {type => SCALAR, default => $BOOLEAN_VALUES},
        },
    );
    my $paths  = $args{paths};
    my $type   = $args{type};
    my $apiset = $self->{linux};
    my $result = 1;
    my $count;

    # Delete file/directory
    my $delete;
    if ($type eq 'file') {
        $Log->step(
            "uACL::LinuxUtils::delete: Delete the files that we created");
        for (my $count = 0; $count < @$paths; $count++) {
            eval { $delete = $apiset->rm(paths => $paths->[$count], force => 1); };
            unless ( defined($delete) ) {
                $Log->comment(
                    "***** fail *****" .
                    "uACL::LinuxUtils::Delete: ERROR: Could not delete the files \n\n"
                );
                return 0;
            } ## end unless ( $apiset->rm( paths...

        } ## end for ( my $count = 0; $count...
    } else {
        $Log->step(
            "uACL::LinuxUtils::delete: Delete the DIRs that we created");
        for (my $count = 0; $count < @$paths; $count++) {
            #duplicate entries will return failure, we don't want to fail the test..
            eval { $delete = $apiset->rmdir(paths => $paths->[$count]); };
            unless ( defined ($delete) ) {
                $Log->comment(
                    "***** fail *****" .
                    "uACL::LinuxUtils::Delete: ERROR: Could not delete the DIRS \n\n"
                );
                return 0;
            } ## end unless ( $apiset->rmdir( paths...
        } ## end for ( my $count = 0; $count...
    } ## end else [ if ( $type eq 'file' )
    $Log->exit();
    return $result;
} ## end sub delete

=head7  get_acl
my %args = ( filename => $path1,   version => 4 );
$uacl->get_acl( %args );

 Get the acl correcponding to the the mentioned path
  Following parameters are accepted:   
  =over
  =item Arguments
  =over
=item filename

   required: The relative path, from the mount point , the name of the file/dir

=item  version

  Optional: Which nfs version.This is 4 by default

=back

=item Return:

=cut

sub getacl() {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {filename => {type => SCALAR, optional => 0},}
    );
    my $apiset     = $self->{linux};
    my $getfacl_op = $apiset->execute_raw_command(
        command => GETFACL . " $args{filename}");
    my ($line, @get_split, @ace_split, @mapped_flag, $flag_req, @mapped,
        $mode, $result);
    my %hash;

    # Parse the output from the get_acl command and put into the acl
    @get_split = split(/\r?\n/, $getfacl_op);

    foreach (@get_split) {
        @ace_split   = split(':', $_);
        @mapped_flag = ();
        @mapped      = ();
        @mapped      = map { $GETACL_RIGHTS->{$_} } split('', $ace_split[3]);
        @mapped_flag = map { $FLAGS->{$_} } split('', $ace_split[1])
            if ($ace_split[1]);
        ($ace_split[0] =~ /A/)
            ? ($mode = $GETACL_RIGHTS->{A})
            : ($mode = $GETACL_RIGHTS->{D});
		next if(exists($hash{$ace_split[2]}{$mode}));
        $hash{$ace_split[2]}{$mode} = join(' ', @mapped);
        $hash{$ace_split[2]}{'flags'} = join(' ', @mapped_flag) 
		if (join(' ', @mapped_flag));
		
    }
    $result = \%hash;
    print Dumper($result);
    return $result;
} ## end sub getacl()

=head7  set_acl
my %args = ( filename => $path1, ace_array=> $aces _list  version => 4 );
$uacl->set_acl( %args );

 Get the acl correcponding to the the mentioned path
  Following parameters are accepted
  =over
  =item Arguments
  =over
=item filename

   required: The relative path, from the mount point , the name of the file/dir. 

=item  version

  Optional: Which nfs version.This is 4 by default. 
=item  ace_array

   required: The reference to an  array of hashes containing the acls to be set
=back

=item Return:

=cut

sub setacl() {
    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            filename            => {type => SCALAR,   optional => 0},
            edit                => {type => SCALAR,   default  => 1},
            aclfile             => {type => ARRAYREF, optional => 0},
            del_file_spec       => {type => SCALAR,   optional => 1},
            overwrite_file_spec => {type => SCALAR,   optional => 1},
            verify              => {type => SCALAR,   default => 1},
            recursive           => {type => SCALAR,   optional => 1},
        },
        allow_extra => 1,
    );
    my $command = undef;
    my $apiset  = $self->{linux};
    my ($set_facl, $aclfile);
    (defined($args{recursive}))
        ? ($command = SETFACL . ' -R')
        : ($command = SETFACL);
    (defined($args{edit}))          ? ($command = $command . ' -A') : undef;
    (defined($args{del_file_spec})) ? ($command = $command . ' -X') : undef;
    (defined($args{overwrite_file_spec}))
        ? ($command = $command . ' -S')
        : undef;

    $aclfile = _mapped_perms($args{aclfile}, $self);
    $apiset->execute_raw_command(command => "cat  $aclfile ");
    $set_facl = $apiset->execute_raw_command(
        command => "$command  $aclfile  $args{filename}");
	&getacl($self,filename => $args{filename}, version  => "linux_v4") if($args{verify});
    $apiset->rm(paths => $aclfile, force => 1);
    ($set_facl) ? return 0 : return 1;

} ## end sub setacl()

=head2 verifyacls

  my %args = ( path => 'file-path', identity => 'user or group' );
  $uacl->verify_acls( %args );

  Verifies ACLs. This method takes verifies whether specified acls are present
  on the file.

  Following parameters are accepted:

=over

=item Arguments

=over

=item path

  Required: File name to check acls.

=item identity

  Required: User or group to whom verify acls.

=item Allow

  Optional: Allow permissions to search for user.

=item Deny

  Optional: Deny permissions to search for user.

=item Flags

  Optional: Flags to search for the user.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub verifyacls {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            path       => {type => SCALAR,  optional => 0},
            identity   => {type => SCALAR,  optional => 0},
            type       => {type => SCALAR,  optional => 1},
            Allow      => {type => SCALAR,  optional => 1},
            Deny       => {type => SCALAR,  optional => 1},
            Flags      => {type => SCALAR,  optional => 1},
            get_acl_op => {type => HASHREF, optional => 0},
            version    => {type => SCALAR,  optional => 1},
        },
    );

    my $path     = $args{path};
    my $identity = $args{identity};
    my $type     = exists $args{type} ? $args{type} : 'file';
    my $Flags    = exists $args{Flags} ? $args{Flags} : undef;
    my $Allow    = exists $args{Allow} ? $args{Allow} : undef;
    my $Deny     = exists $args{Deny} ? $args{Deny} : undef;
    my $attr;

    unless ((defined($args{has_owner}))
        || (defined($args{has_group}))
        || (defined($args{Allow}))
        || (defined($args{Deny}))
        || (defined($args{Flags}))) {
        $Log->comment(
            "Nothing specified to verify in ACL in any \'has_*\' parameter, exiting."
        );
        return 0;
    } ## end unless ( ( defined( $args{...

    my $current_acl = $args{get_acl_op};
    $Log->comment("Current ACL on the file are:" . Dumper($current_acl));

    if (defined($identity)) {
        if (exists($current_acl->{$identity})) {
            $Log->comment("The specified user[$identity] has acls on file ");
        } else {
            $Log->comment("User[$identity] does not has acls on file ");
            return 0;
        }
    } ## end if ( defined($identity...

    if ((defined($Allow) && !exists($current_acl->{$identity}->{Allow}))
        || (defined($Deny)
            && !exists($current_acl->{$identity}->{Deny}))
        || (defined($Flags)
            && !exists($current_acl->{$identity}->{Flags}))
        ) {
        $Log->comment(
            "[Invalid Entry]ACL your are checking not exists on file");
        return 0;
    } ## end if ( ( defined($Allow)...

    if (defined($Allow)) {
        foreach $attr (split(/,/, $Allow)) {
            if ($current_acl->{$identity}->{Allow} =~ m/$attr/i) {
                $Log->comment(
                    "Allow rights on the file for $identity matches with "
                        . $attr
                        . "\n");
            } else {
                $Log->comment(
                    "Allow rights on the file for $identity DO NOT match with "
                        . $attr
                        . "\n");
                $Log->comment("Allow rights on the file are "
                        . $current_acl->{$identity}->{Allow});
                return 0;
            } ## end else [ if ( $current_acl->{$identity...
        } ## end foreach $attr ( split( /,/,...
    } ## end if ( defined($Allow) )

    if (defined($Deny)) {
        foreach $attr (split(/,/, $Deny)) {
            if ($current_acl->{$identity}->{Deny} =~ m/$attr/i) {
                $Log->comment(
                    "Deny rights on the file for $identity matches with "
                        . $attr
                        . "\n");
            } else {
                $Log->comment(
                    "Deny rights on the file for $identity DO NOT match with "
                        . $attr
                        . "\n");
                $Log->comment("Deny rights on the file are "
                        . $current_acl->{$identity}->{Deny});
                return 0;
            } ## end else [ if ( $current_acl->{$identity...
        } ## end foreach $attr ( split( /,/,...
    } ## end if ( defined($Deny) )

    if (defined($Flags)) {
        foreach $attr (split(/,/, $Flags)) {
            if ($current_acl->{$identity}->{Flags} =~ m/$attr/i) {
                $Log->comment("Flags on the file for $identity matches with " 
                        . $attr
                        . "\n");
            } else {
                $Log->comment(
                          "Flags on the file for $identity DO NOT match with "
                        . $attr
                        . "\n");
                $Log->comment("Flags on the file are "
                        . $current_acl->{$identity}->{Flags});
                return 0;
            } ## end else [ if ( $current_acl->{$identity...
        } ## end foreach $attr ( split( /,/,...
    } ## end if ( defined($Flags) )

    $Log->exit();
    return 1;
} ## end sub verifyacls

sub chmod {
    my $self = shift;

    my %args = validate_with(
        params => \@_,
        spec   => {
            version => {type => SCALAR, optional => 0},
            mode    => {type => SCALAR, optional => 0},
            name    => {type => SCALAR, optional => 1},
        }
    );
    my $apiset = $self->{linux};
    my $result = 1;
    unless ($apiset->chmod(mode => $args{mode}, paths => $args{name})) {
        $Log->error(
            " uACL::LinuxUtils::chmod:: ERROR: Could not change the mode bits of file/DIR"
        );
        return 0;
    } ## end unless ( $apiset->chmod( mode...
    $Log->exit();
    return $result;

} ## end sub chmod

sub useradd {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {user => {type => ARRAYREF},},
    );

    my $user   = $args{user};
    my $apiset = $self->{linux};
    my $command;
    my $result = 1;

    for (my $i = 0; $i < @$user; $i++) {
        $command = "useradd $user->[$i]";
        $Log->comment("command: $command");
        my $output = $apiset->execute_raw_command(command => $command);
        $result = 0
            if ($output =~ /\w+/
            && $output !~ /already in use|name too long/);
    } ## end for ( my $i = 0; $i < @$user...

    $Log->exit();
    return $result;
} ## end sub useradd

sub groupadd {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {group => {type => ARRAYREF},},
    );

    my $group  = $args{group};
    my $apiset = $self->{linux};
    my $command;
    my $result = 1;

    for (my $i = 0; $i < @$group; $i++) {
        $command = "groupadd $group->[$i]";
        $Log->comment("command: $command");
        my $output = $apiset->execute_raw_command(command => $command);
        $result = 0
            if ($output =~ /\w+/
            && $output !~ /already in use|name too long/);
    } ## end for ( my $i = 0; $i < @$group...

    $Log->exit();
    return $result;
} ## end sub groupadd

sub max_nfs_aces {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            command_interface => {type => HASHREF, optional => 0},
            vserver           => {type => SCALAR,  optional => 0},
            ace_count         => {type => SCALAR,  optional => 0},
        },
    );

    my $ci        = $args{command_interface};
    my $vserver   = $args{vserver};
    my $ace_count = $args{ace_count};
    my $result    = 1;

    NACL::C::VserverNFS->modify(
        command_interface => $ci,
        'vserver'         => $vserver,
        'v4-acl-max-aces' => $ace_count,
    );

    $Log->exit();
    return $result;
} ## end sub max_nfs_aces
##--------------------- Private Methods ---------------------##
sub _create_nfs_client {
    ## Construct linux client apiset
    my $linux_client = NACL::C::Client->find();
    my $client_hostrec = $linux_client->hostrec();
    my $login = $client_hostrec->{connection_fields}{ssh}{username};
    my $password = $client_hostrec->{connection_fields}{ssh}{password};
    my $apiset       = $linux_client->apiset(
        category  => 'Host',
        interface => 'CLI',
        username  => $login,
        password  => $password,
        connid    => 'ssh-persist',
        set       => 'Linux',
    );

    ## Set root privilege
    $apiset->set_privilege(
        'privilege-level' => 'root',
        'answer-password' => 1
    );

    ## Set client connection timeout
    $apiset->set_timeout('connectrec-timeout' => 1500);

    ## Return cleint APISet object
    return $apiset;
} ## end sub _create_nfs_client

sub _mapped_perms {
    my ($ele, $self) = @_;
    my $apiset = $self->{linux};

    my $hash_count = @{$ele};
    my ($counter, @ace_hash, @mapped_array, @mapped_mask, @temp, $final_form,
        $mask, $aclfile);
    @mapped_array = ();
    @mapped_mask  = ();
    @temp         = ();
    my $i = 0;
    while ($hash_count--) {
        push @ace_hash, values(%{$ele->[$i++]});
    }
    $aclfile = '/t/setacl' . $$;

    unless ($apiset->touch('path' => $aclfile)) {
        $Log->error(
            "\n uACL::LinuxUtils::setacl: ERROR: Could not create file [$aclfile] \n\n"
        );
    }
    $apiset->execute_raw_command(
        command => " /usr/software/bin/sudo chmod 777 $aclfile");
    foreach (@ace_hash) {
        $counter++;
        ($_ =~ /\s+/) ? (push(@mapped_mask, split('\s+', $_)))
            : (($_ =~ /@/) ? push(@mapped_array, $_)
            : push(@mapped_array, $SETACL_RIGHTS->{$_}));
        if ($counter == 4) {
            @temp = map { $SETACL_RIGHTS->{$_} } @mapped_mask;
            push(@mapped_array, join('', @temp));
			$final_form = join(':', @mapped_array);
            $apiset->execute_raw_command(
                command => " echo $final_form >> $aclfile");

            $counter      = 0;
            @mapped_mask  = ();
            @mapped_array = ();
            @temp         = ();
        }
    }

    return $aclfile;
} ## end sub _mapped_perms

sub _init_globals() {

    my $self = shift;

    $Options->process();

    #Used compiled regex for frequent ones
    $NUMERICS       = qr/\d+/;
    $BOOLEAN_VALUES = qr/true|false/i;

    #SetACL Rights
    $SETACL_RIGHTS->{Rdata} = 'r';
    $SETACL_RIGHTS->{Wdata} = 'w';
    $SETACL_RIGHTS->{Adata} = 'a';
    $SETACL_RIGHTS->{Rna}   = 'n';
    $SETACL_RIGHTS->{Wna}   = 'N';
    $SETACL_RIGHTS->{Rattr} = 't';
    $SETACL_RIGHTS->{Wattr} = 'T';
    $SETACL_RIGHTS->{Racl}  = 'c';
    $SETACL_RIGHTS->{Wacl}  = 'C';
    $SETACL_RIGHTS->{Sync}  = 'y';
    $SETACL_RIGHTS->{Del}   = 'd';
    $SETACL_RIGHTS->{Wown}  = 'o';
    $SETACL_RIGHTS->{Ex}    = 'x';
    $SETACL_RIGHTS->{Allow} = 'A';
    $SETACL_RIGHTS->{Deny}  = 'D';
    $SETACL_RIGHTS->{GR}    = 'g';
    $SETACL_RIGHTS->{FI}    = 'f';
    $SETACL_RIGHTS->{DI}    = 'd';
    $SETACL_RIGHTS->{NPI}   = 'n';
    $SETACL_RIGHTS->{IONLY} = 'i';

    #GetAcl Rights
    $GETACL_RIGHTS->{r} = 'Rdata';
    $GETACL_RIGHTS->{w} = 'Wdata';
    $GETACL_RIGHTS->{a} = 'Adata';
    $GETACL_RIGHTS->{n} = 'Rna';
    $GETACL_RIGHTS->{N} = 'Wna';
    $GETACL_RIGHTS->{t} = 'Rattr';
    $GETACL_RIGHTS->{T} = 'Wattr';
    $GETACL_RIGHTS->{c} = 'Racl';
    $GETACL_RIGHTS->{C} = 'Wacl';
    $GETACL_RIGHTS->{y} = 'Sync';
    $GETACL_RIGHTS->{d} = 'Del';
    $GETACL_RIGHTS->{o} = 'Wown';
    $GETACL_RIGHTS->{x} = 'Ex';
    $GETACL_RIGHTS->{A} = 'Allow';
    $GETACL_RIGHTS->{D} = 'Deny';
	
	$FLAGS->{d} = 'DI';
	$FLAGS->{n} = 'NPI';
	$FLAGS->{g} = 'GR';
	$FLAGS->{f} = 'FI';
	$FLAGS->{i} = 'IONLY';
    return 1;
} ## end sub _init_globals()

1;
