# $Id: $
#
# Copyright (c) 2001-2010 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
## @summary SolarisUtils - Class for Unified ACLs on Solaris Platform
## @author natha@netapp.com, veerr@netapp.com
## @status private
## @pod here

=head1 NAME

uACL::SolarisUtils - Solaris ACL Class

=head1 DESCRIPTION

C<uACL::SolarisUtils> provides methods for ACL tests on Solaris platform.

=head1 METHODS

=cut

package uACL::SolarisUtils;

use strict;
use warnings;

use Tharn qw($Log);
use Params::Validate qw( validate_with ARRAYREF BOOLEAN HASHREF SCALAR );

use Data::Dumper;
use NACL::C::Client;
use NACL::STask::Client::Mount;
use NACL::APISet::Exceptions::CommandFailedException qw( :try );

## Command paths
use constant CD      => '/usr/bin/cd';
use constant MV      => '/usr/bin/mv';
use constant RM      => '/usr/bin/rm';
use constant RMDIR   => '/usr/bin/rmdir';
use constant TOUCH   => '/usr/bin/touch';
use constant GETFACL => '/usr/bin/getfacl';
use constant CHMOD   => '/usr/bin/chmod';
use constant LS      => '/usr/bin/ls';
use constant CP      => '/usr/bin/cp';

=head2 new

  my $uacl = uACL::SolarisUtils->new();

  Creates an object of type uACL::SolarisUtils. No arguments required.

=cut

sub new {
    $Log->enter();

    my $class = shift;
    my $self  = {};

    $self->{solaris} = _create_solaris_client();
    bless( $self, $class );

    $Log->exit();

    return $self;
} ## end sub new

=head2 mount

  my %args = (
      'data-lif' => NACL::C::Node->find()->{hostrec}->{DATA_IP_1},
      'junction-path' => '/NS',
      'mount-point' => '/t/mnt',
      'mount-version' => 4,
  );
  $uacl->mount( %args );

  Mounts the juction path of an infinite volume the specified mount point.

  Following parameters are accepted:

=over

=item Arguments

=over

=item data-lif

  Required: IP address of a Data LIF associated with repository vserver.

=item junction-path

  Required: Junction path of the infinite volume

=item mount-point

  Optional: Mount point to which juction path of an infinite volume will be mount.
            If not passed then /t/<ClientHostname>/repo_vol will be used.

=item mount-version

  Optional: NFS version for mounting.
            Allowed values are 3 or 4
            If not passed then 4 will be used by default

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub mount {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            data_lif      => { type => SCALAR },
            junction_path => { type => SCALAR },
            mount_version => { type => SCALAR },
        },
        allow_extra => 1
    );
    my $apiset        = $self->{solaris};
    my $data_lif      = $args{data_lif};
    my $junction_path = $args{junction_path};
    my $mount_version = $args{mount_version};
    my $mount_point   = '/t/' . $apiset->{hostobj}->{hostname} . '/repo_vol' . $mount_version;
    my $result        = 1;

    ## Make mount point directory
    $Log->step("Make mount point directory [$mount_point]");
    try {
        $apiset->mkdir( 'make-parents' => 1, 'paths' => $mount_point );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        $Log->comment( "\nERROR: Could not create mount point directory\n"
                . "Reason: "
                . $exception->get_error_message()
                . "\n\n" );
        $result = 0;
    };

    ## Mount Infinite Volume on the solaris client
    $Log->step( "Mount Infinite Volume junction path [$junction_path]"
            . " to mount point [$mount_point]" );
    try {
        my $output = $apiset->mount(
            'file-system-type' => 'nfs',
            'source-path'      => $data_lif . ':' . $junction_path,
            'mount-point'      => $mount_point,
            'options'          => 'vers=' . $mount_version,
        );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        $Log->comment(
                  "\nERROR: Could not mount Infinite Volume's junction path"
                . "\nReason: "
                . $exception->get_error_message()
                . "\n\n" );
        $result = 0;
    };

    ## Change permission
    $Log->step("Change permission [0777] of the directory [$mount_point]");
    try {
        $apiset->chmod( mode => '0777', paths => $mount_point );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        $Log->comment( "\nERROR: Could not change permission of mount point "
                . "directory\nReason: "
                . $exception->get_error_message()
                . "\n\n" );
        $result = 0;
    };

    ## Store the mountpoint
    $self->{'sol-mount-point'} = $mount_point;

    $Log->exit();
    return $result;
} ## end sub mount

=head2 umount

  $uacl->umount( 'mount-point' => '/t/mnt' );

  Unmounts the juction path of an infinite volume from the specified mount point.

  Following parameters are accepted:

=over

=item Arguments

=over

=item mount-point

  Optional: Mount point to which juction path of an infinite volume was mounted.

=item umount

  Optional: Specify umount is required or not. Default value is 'true'.

=item delete

  Optional: Specify mount point needs to be deleted or not.
            Default value is 'true'.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub umount {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            'mount-point' =>
                { type => SCALAR, default => $self->{'sol-mount-point'} },
            umount => { type => SCALAR, default => 'true' },
            delete => { type => SCALAR, default => 'true' },
        }
    );
    my $apiset      = $self->{solaris};
    my $mount_point = $args{'mount-point'};
    my $umount      = $args{umount};
    my $delete      = $args{delete};
    my $result      = 1;

    ## Unmount Infinite Volume from the solaris client
    if ( $umount eq 'true' ) {
        $Log->step("Unmount Infinite Volume from mount point [$mount_point]");
        try {
            $apiset->umount(
                'paths'          => $mount_point,
                'forcily-umount' => 1,
            );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment(
                "\nERROR: Could not unmount Infinite Volume's junction path"
                    . "\nReason: "
                    . $exception->get_error_message()
                    . "\n\n" );
            $result = 0;
        };
    } else {
        $Log->comment('Skipping umount operation...');
    }

    ## Remove the mount point directory
    if ( $delete eq 'true' ) {
        $Log->step("Delete mount point directory [$mount_point]");
        try {
            $apiset->rmdir( paths => $mount_point );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment( "\nERROR: Could not Delete mount point"
                    . "\nReason: "
                    . $exception->get_error_message()
                    . "\n\n" );
            $result = 0;
        };
    } else {
        $Log->comment('Retaining mount point for debugging');
    }

    $Log->exit();
    return $result;
} ## end sub umount

=head2 create_file

  my %args = ( path => $dir, count => 5 );
  $uacl->create_file( %args );

  Create file(s) under a specified path.

  Following parameters are accepted:

=over

=item Arguments

=over

=item count

  Optional: Number of files to create. If not passed only one file will be
            created by default.

=item path

  Optional: Path under which file(s) need to be created. If not passed file(s)
            will be created under the mount point.

=back

=item Return:

  Success - Array of file names or a single file name.
  Failure - 0

=back

=cut

sub create_file {

    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            count => { type => SCALAR, optional => 1, default => 1 },
            path  => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1,
    );
    my $count      = $args{count};
    my $path       = $args{path};
    my $apiset     = $self->{solaris};
    my $mountpoint = $self->{'sol-mount-point'};
    my $result     = 1;
    my ( $file, @files );

    for ( my $i = 0; $i < $count; $i++ ) {
        if ( defined($path) ) {
            $file = "uACL_Test_$i".time().".txt";
            $path = "$mountpoint/$path/$file";
            $path =~ s|//|/|;
        } else {
            $file = "uACL_Test_$i".time().".txt";
            $path = "$mountpoint/$file";
        }
        push @files, $file;

        ## Create file
        try {
            $Log->step("Create file: $path");
            my $command = TOUCH;
            $apiset->execute_raw_command( command => "$command $path" );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment( "\nERROR: Could not create file [$path]"
                    . "\nReason: "
                    . $exception->get_error_message()
                    . "\n\n" );
            return 0;
        };
        $path = $args{path};
    }

    $Log->exit();
    return $files[0] if ( $count == 1 );
    return \@files;
} ## end sub create_file

=head2 create_dir

  my %args = ( nested => 'true', count => 5 );
  $uacl->create_dir( %args );

  Create directories under the mount point.

  Following parameters are accepted:

=over

=item Arguments

=over

=item count

  Optional: Number of directories to create. If not passed only one directory
            will be created by default.

=item nested

  Optional: Default value is 'false'. If passed as 'true' nested directories
            will be created (e.g. /t/mnt/dir1/dir2/dir3/)

=back

=item Return:

  Success - Array of directory names or a directory name.
  Failure - 0

=back

=cut

sub create_dir {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            count  => { type => SCALAR,  default => 1,       optional => 1 },
            nested => { type => BOOLEAN, default => 'false', optional => 1 },
            path   => { type => SCALAR, optional => 1 },
        },
        allow_extra => 1
    );
    my $count      = $args{count};
    my $nested     = $args{nested};
    my $path       = $args{path};
    my $apiset     = $self->{solaris};
    my $mountpoint = $self->{'sol-mount-point'};
    my ( $dir, @dirs ) = undef;

    if ( $nested eq 'false' ) {
        ## Create directories under the base mount point
        for ( my $i = 1; $i <= $count; $i++ ) {
            if ( defined($path) ) {
                $dir = 'dir' .time(). $i;
                $path = "$mountpoint/$path/$dir";
                $path =~ s|//|/|;
            } else {
                $dir = 'dir' .time(). $i;
                $path = "$mountpoint/$dir";
            }
            push @dirs, $dir;
            $Log->step("Create directory $path");
            try {
                $apiset->mkdir(
                    'make-parents' => 1,
                    'paths'        => "$path"
                );
            }
            catch NACL::APISet::Exceptions::CommandFailedException with {
                my $exception = shift;
                $Log->comment(
                    "\nERROR: Could not create directory [$mountpoint/$dir] "
                        . "\nReason: "
                        . $exception->get_error_message()
                        . "\n\n" );
                return 0;
            };
            $path = $args{path};
        }
    } else {
        ## Create nested directories
        for ( my $i = 1; $i <= $count; $i++ ) {
            $dir .= '/dir' .time(). $i;
        }
        $dir =~ s/^\///g;
        push @dirs, $dir;
        $Log->step("Create directory $mountpoint/$dir");
        try {
            $apiset->mkdir(
                'make-parents' => 1,
                'paths'        => "$mountpoint/$dir"
            );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment( "\nERROR: Could not create directory ["
                    . $mountpoint
                    . $dir
                    . "]\nReason: "
                    . $exception->get_error_message()
                    . "\n\n" );
            return 0;
        };
    }

    $Log->exit();
    return $dirs[0] if ( $count == 1 );
    return \@dirs;
} ## end sub create_dir

=head2 move

  Move a file to a different directory:
  $uacl->move( source => [ /t/mnt/file1 ], destination => [ /t/mnt/dir1/ ] );

  Rename a file:
  $uacl->move( source => [ /t/mnt/file1 ], destination => [ /t/mnt/file2 ] );

  Move or rename multiple file(s)/directory(s):
  my %args = (
      source => [ /t/mnt/file1, /t/mnt/file3 ],
      destination => [ /t/mnt/file2, /t/mnt/dir2 ]
  );
  $uacl->move( %args );

  This method moves or renames the file named by the source argument to the
  destination specified by the destination argument. The destination argument
  can be a file or directory. If the destination file does not exist then this
  method creates it.

  Following parameters are accepted:

=over

=item Arguments

=over

=item source

  Required: ArrayRef of file names that need to be moved to a new location
            or renamed.

=item destination

  Required: ArrayRef of file or directory names to which the source file needs
            be moved or renamed.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub move {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            src_file_path => { type => ARRAYREF },
            dst_file_path => { type => ARRAYREF },
        },
        allow_extra => 1
    );
    my $source      = $args{src_file_path};
    my $destination = $args{dst_file_path};
    my $apiset      = $self->{solaris};
    my $result      = 1;
    my $command     = undef;

    ## Move file/directory
    for ( my $i = 0; $i < @$source; $i++ ) {
        $command = MV . " -f $source->[$i] $destination->[$i]";
        $Log->comment("command: $command");
        try {
            $apiset->execute_raw_command( command => $command );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment(
                "ERROR: Could not move [$source] to [$destination]");
            $result = 0;
        };
    }

    $Log->exit();
    return $result;
} ## end sub move

=head2 copy

  Copy a file to a different directory:
  $uacl->copy( src_file_path => [ /t/mnt/file1 ], dest_file_path => [ /t/mnt/dir1/ ] );

  Copy a file to a different file:
  $uacl->copy( src_file_path => [ /t/mnt/file1 ], dest_file_path => [ /t/mnt/file2 ] );

  Copy a dir to a different dir:
  $uacl->copy( src_file_path => [ /t/mnt/dir1 ], dest_file_path => [ /t/mnt/dir2 ] );

  $uacl->copy( %args );

  This method copies the file named by the source argument to the destination
  specified by the destination argument.If the destination file does not exist
  then this method creates it. If destination file exists then it overwrites it.
  If destination is a directory then it should exists.

  Following parameters are accepted:

=over

=item Arguments

=over

=item source

  Required: ArrayRef of file names that need to be moved to a new location
            or renamed.

=item destination

  Required: ArrayRef of file or directory names to which the source file needs
            be moved or renamed.

=item recursive

  Optional: Scalar value which allows to copy directory and its file to other
            directory.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub copy {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            src_file_path => { type => ARRAYREF },
            dst_file_path => { type => ARRAYREF },
            recursive     => { type => SCALAR, default => 'false' },
        },
        allow_extra => 1
    );
    my $source      = $args{src_file_path};
    my $destination = $args{dst_file_path};
    my $apiset      = $self->{solaris};
    my $recursive   = $args{recursive};
    my $result      = 1;
    my $command     = undef;

    $recursive = ( $recursive eq 'true' ) ? ' -r' : '';
    ## Copy file/directory
    for ( my $i = 0; $i < @$source; $i++ ) {
        $command = CP . "$recursive $source->[$i] $destination->[$i]";
        $Log->comment("command: $command");
        try {
            $apiset->execute_raw_command( command => $command );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment(
                "ERROR: Could not copy [$source] to [$destination]");
            $result = 0;
        };
    }

    $Log->exit();
    return $result;
} ## end sub copy

=head2 delete

  Delete a file:
  $uacl->delete( path => [ /t/mnt/file1 ] );

  Delete multiple files:
  $uacl->delete( path => [ /t/mnt/file1, /t/mnt/file2 ] );

  Delete a directory:
  $uacl->delete( path => [ /t/mnt/dir1 ] );

  This method removes the file(s) or directory(ies) passed by the paths argument.
  By default this method delete file(s) or directory(ies) forcefully.

  Following parameters are accepted:

=over

=item Arguments

=over

=item paths

  Required: ArrayRef of file or directory names that need to be removed.

=item type

  Optional: Type of file to be deleted whether directory or file.
            Default value is file.

=item force

  Optional: Enables forceful deletion. Default value is 'true'.

=item recursive

  Optional: Recursive deletion. Must pass a true value while deleting a
            directory. Default value is 'false'.

=item parent

  Optional: Removes dir and its parent directories also.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub delete {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            paths     => { type => ARRAYREF },
            type      => { type => SCALAR, default => 'file' },
            force     => { type => SCALAR, default => 'true' },
            recursive => { type => SCALAR, default => 'false' },
            parent    => { type => SCALAR, default => 'false' },
        },
        allow_extra => 1
    );
    my $paths     = $args{paths};
    my $type      = $args{type};
    my $force     = $args{force};
    my $recursive = $args{recursive};
    my $parent    = $args{parent};
    my $apiset    = $self->{solaris};
    my $result    = 1;
    my $command;
    $type eq 'file' ? ( $command = RM ) : ( $command = RMDIR );
    $command .= ( $force     eq 'true' && $type eq 'file' ) ? ' -f' : '';
    $command .= ( $recursive eq 'true' && $type eq 'file' ) ? ' -r' : '';
    $command .= ( $parent    eq 'true' && $type eq 'dir' )  ? ' -p' : '';

    ## Delete file/directory
    for ( my $i = 0; $i < @$paths; $i++ ) {
        try {
            $apiset->execute_raw_command(
                command => "$command $paths->[$i]" );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception = shift;
            $Log->comment("ERROR: Could not delete [$paths->[$i]]");
            $result = 0;
        };
    }

    $Log->exit();
    return $result;
} ## end sub delete

=head2 setacl

  my %args = ( acl_entries => 'A+user:veerr:read_data:file_inherit:allow',
               file => 'uACL.txt' );
  $uacl->setacl( %args );

  The chmod command is used to modify Access Control Lists (ACLs)
  on files and directories.

  Following parameters are accepted:

=over

=item Arguments

=over

=item file

  Required: Name of the file which ACLs need to be modified.

=item acl_entries

  Required: acl_entries are one or more comma-separated ACL entries.

    chmod [options] A+acl_specification file ...
    chmod [options] A-acl_specification file ...

    Example: A+user:read_data:allow,A+user:tom:read_data/write_data:deny
             A-group:delete:deny

=item force

    Optional :  Force. chmod does not complain if it fails  to  change
                the mode of a file.

=item recursive

    Optional :  Recursively descends through directory arguments, set-
                ting  the  mode for each file as described above. When
                symbolic links are encountered, the mode of the target
                file is changed, but no recursion takes place.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub setacl {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            acl_entries       => { type => ARRAYREF, optional => 0 },
            file              => { type => SCALAR,   optional => 0 },
            force             => { type => SCALAR,   optional => 1 },
            recursive         => { type => SCALAR,   optional => 1 },
        },
        allow_extra => 1
    );

    my $file        = $args{file};
    my $force       = exists $args{force} ? $args{force} : undef;
    my $recursive   = exists $args{recursive} ? $args{recursive} : undef;
    my @acl_entries = join( ',', @{ $args{acl_entries} } );
    my $apiset  = $self->{solaris};
    my $command = CHMOD;
    my $options = undef;
    my $result = 1;

    ## Get options passed for CHMOD
    if ( defined($force) ) {
        $options = '-f' if ( $force eq 'true' );
    } elsif ( defined($recursive) ) {
        $options = '-R' if ( $recursive eq 'true' );
    }

    if ( defined($options)) {
        $command .= " $options @acl_entries $file";
    } else {
        $command .= " @acl_entries $file";
    } ## end if ( defined($options..

    ## Execute chmod command from Solaris client
    $Log->comment("command: $command");
    try {
        $apiset->execute_raw_command( command => $command );
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        $Log->comment("ERROR: Could not run chmod command");
        $result = 0;
    };

    $Log->exit();
    return $result;
} ## end sub setacl

=head2 getacl

  my %args = ( file => 'uACL.txt', type => 'dir', ver => '4' );
  $uacl->getacl( %args );

  For each file that is a directory, ls lists the contents  of the  directory.
  For  each file that is an ordinary file, ls repeats its name and any other
  information  requested.  The output is sorted alphabetically by default.

  With -v options, ls displays verbose acl information.

  Following parameters are accepted:

=over

=item Arguments

=over

=item file

  Required: The path name of a regular file, special file, or named pipe.

=item type

  Required: The type whether it is file or directory. Default value is 'file'.

=item version

  Optional: The version to use to do getacl on file. Default value is '4'.

=back

=item Return:

  Success - Returns a HASHREF containing the ACL information.
  Failure - 0

=back

=cut

sub getacl {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            file   => { type => SCALAR,   optional => 0 },
            type   => { type => SCALAR,   optional => 1 },
            ver   =>  { type => SCALAR,   optional => 1, default => 4 },
        },
        allow_extra => 1
    );

    my $file    = exists $args{file} ? $args{file} : undef;
    my $type    = exists $args{type} ? $args{type} : 'file';
    my $ver     = $args{ver};
    my $apiset  = $self->{solaris};
    my $command = LS;
    my %hash;

    $type eq 'dir' ? ($command .= " -dv $file") : ($command .= " -lv $file");
    $Log->comment("command: $command");

    ## Execute ls command from Solaris client
    $apiset->execute_raw_command( command => GETFACL ." $file" );
    my $output = $apiset->execute_raw_command( command => $command );
    my $result = $output !~ /No such file or directory/i ? 1 : 0;

    ## Extract the acl entries from the command output and
    ## construct a HASH for acls
    if ($result) {
        my @array = split(/\n\s+\d+:/,$output);
        my $bits = shift(@array);
        if( ($bits =~ /(d|-)([rwx-]+).*/) ) {
            $hash{ 'user-modebits' }    = substr( $2, 0, 3 );
            $hash{ 'group-modebits' }   = substr( $2, 3, 3 );
            $hash{ 'others-modebits' }  = substr( $2, 6, 3 );
        }

        if ( $ver eq '4') {
        foreach my $out (@array) {
            my ( @perms, @flags, $modes, $access, $user );
            $out =~ s/\s+//g;
            my @values = split(/:/,$out);
            if ( $values[0] eq 'user' || $values[0] eq 'group') {
                $user = $values[1];
                @perms = split(/\//,$values[2]);
                if ($#values == 4) {
                    @flags = split(/\//,$values[3]);
                    $modes = $values[4];
                } else {
                    $modes = $values[3];
                }
            } else {
                $user = $values[0];
                @perms = split(/\//,$values[1]);
                if ($#values == 3) {
                    @flags = split(/\//,$values[2]) ;
                    $modes = $values[3];
                } else {
                    $modes = $values[2];
                }
            }

            $modes eq 'allow' ? ($access = 'deny') : ($access = 'allow');
            foreach my $val ( @perms ) {
                if ( defined $hash{$user}{$access} ) {
                    next if( $hash{$user}{$access} =~ /$val/);
                }
                if ( defined $hash{$user}{$modes} ) {
                    next if ( $hash{$user}{$modes} =~ /$val/ );
                }
                $hash{$user}{$modes} .= "$val ";
            }
            foreach my $flg ( @flags ) {
                if ( defined $hash{$user}{'flags'}) {
                    next if( $hash{$user}{'flags'} =~ /$flg/);
                }
                $hash{$user}{'flags'} .= "$flg ";
            }
        }
        }
        $result = \%hash;
    }  ## end if ($result ....

    $Log->exit();
    return $result;
} ## end sub getacl

=head2 verifyacls

  my %args = ( path => 'file-path', identity => 'user' );
  $uacl->verify_acls( %args );

  Verifies ACLs. This method takes verifies whether specified acls are present
  on the file.

  Following parameters are accepted:

=over

=item Arguments

=over

=item path

  Required: File name to check acls.

=item identity

  Required: User to search the acls for.

=item has_allow

  Optional: Allow permissions to search for user.

=item has_deny

  Optional: Deny permissions to search for user.

=item has_flags

  Optional: Flags to search for the user.

=back

=item Return:

  Success - 1
  Failure - 0

=back

=cut

sub verifyacls {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   =>
        {   path           => { type => SCALAR, optional => 0 },
            identity       => { type => SCALAR, optional => 0 },
            type           => { type => SCALAR, optional => 1 },
            has_allow      => { type => SCALAR, optional => 1 },
            has_deny       => { type => SCALAR, optional => 1 },
            has_flag       => { type => SCALAR, optional => 1 },
            version        => { type => SCALAR, optional => 1 },
        },
     );

    my $path       = $args{path};
    my $identity   = $args{identity};
    my $type       = exists $args{type} ? $args{type} : 'file';
    my $has_flag   = exists $args{has_flag}  ? $args{has_flag} : undef;
    my $has_allow  = exists $args{has_allow} ? $args{has_allow}: undef;
    my $has_deny   = exists $args{has_deny}  ? $args{has_deny} : undef;

    unless ( ( defined( $args{has_owner} ) )|| ( defined( $args{has_group} ) )
           || ( defined( $args{has_allow} ) )|| ( defined( $args{has_deny} ) )
           || ( defined( $args{has_flag} ) ))
    {
        $Log->comment(
            "Nothing specified to verify in ACL in any \'has_*\' parameter, exiting."
        );
        return 0;
    } ## end unless ( ( defined( $args{...

    my $current_acl = $self->getacl( file => $path, type => $type );
    $Log->comment("Current ACL on the file are:" . Dumper($current_acl) );

    if ( defined( $identity ) ) {
        if ( exists( $current_acl->{$identity} ) ) {
            $Log->comment("The specified user[$identity] has acls on file ");
        } else {
            $Log->comment("User[$identity] does not has acls on file ");
            return 0;
        } ## end else [ if exists( $current_acl...
    } ## end if ( defined( $identity...

    if (  (defined($has_allow) && !exists( $current_acl->{ $identity }->{allow}))
       || (defined($has_deny) && !exists( $current_acl->{ $identity }->{deny}))
       || (defined($has_flag) && !exists( $current_acl->{ $identity }->{flags})) )
    {
        $Log->comment("uACL::SolarisUtils::verifyacls => ACL your are checking not exists on file");
        return 0;
    } ## end if ( defined( $has_....

    if ( defined( $has_allow ) ) {
        foreach my $attr (split(/,/,$has_allow)) {
            if ( $current_acl->{ $identity }->{allow} =~ m/$attr/i ) {
            $Log->comment("Allow rights on the file for $identity matches with "
                        . $attr . "\n" );
            } else {
            $Log->comment( "Allow rights on the file for $identity DO NOT match with "
                          . $attr. "\n" );
            $Log->comment( "Allow rights on the file are "
                          . $current_acl->{ $identity }->{allow} );
            return 0;
            } ## end else [ if ( $current_acl->{ $has_allow...
        } ## end foreach
    } ## end if ( defined( $has_allow...

    if ( defined( $has_deny ) ) {
        foreach my $attr (split(/,/,$has_deny)) {
            if ($current_acl->{ $identity }->{deny} =~ m/$attr/i ) {
            $Log->comment( "Deny rights on the file for $identity matches with "
                         . $attr . "\n" );
            } else {
            $Log->comment( "Deny rights on the file for $identity DO NOT match with "
                         . $attr . "\n" );
            $Log->comment( "Deny rights on the file are "
                         . $current_acl->{ $identity }->{deny} );
            return 0;
            } ## end else [ if ( $current_acl->{ $has...
        } ## end foreach
    } ## end if ( defined( $has_deny...

    if ( defined( $has_flag ) ) {
        foreach my $attr (split(/,/,$has_flag)) {
            if ($current_acl->{ $identity }->{flags} =~ m/$attr/i ) {
            $Log->comment( "Flags on the file for $identity matches with "
                          . $attr . "\n" );
            } else {
            $Log->comment( "Flags on the file for $identity DO NOT match with "
                          . $attr . "\n" );
            $Log->comment( "Flags on the file are "
                          . $current_acl->{ $identity }->{flags} );
            return 0;
            } ## end else [ if ( $current_acl->{ $has...
        } ## end foreach
    } ## end if ( defined( $has_flag...

    $Log->exit();
    return 1;
} ## end sub verifyacls

sub useradd {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            user => { type => ARRAYREF },
        },
        allow_extra => 1
    );

    my $user = $args{user};
    my $apiset  = $self->{solaris};
    my $command;
    my $result = 1;

    for ( my $i = 0; $i < @$user; $i++ ) {
        $command = "useradd $user->[$i]";
        $Log->comment("command: $command");
        my $output = $apiset->execute_raw_command( command => $command );
        $result = 0 if ( $output =~ /\w+/ &&
                         $output !~ /already in use|name too long/);
    }

    $Log->exit();
    return $result;
}

sub groupadd {
    $Log->enter();

    my $self = shift;
    my %args = validate_with(
        params => \@_,
        spec   => {
            group => { type => ARRAYREF },
        },
        allow_extra => 1
    );

    my $group = $args{group};
    my $apiset  = $self->{solaris};
    my $command;
    my $result = 1;

    for ( my $i = 0; $i < @$group; $i++ ) {
        $command = "groupadd $group->[$i]";
        $Log->comment("command: $command");
        my $output = $apiset->execute_raw_command( command => $command );
        $result = 0 if ( $output =~ /\w+/ &&
                         $output !~ /already in use|name too long/);
    }

    $Log->exit();
    return $result;
}

##--------------------- Private Methods ---------------------##
sub _create_solaris_client {
    ## Construct solaris client apiset
    my $solaris_client = NACL::C::Client->find();
    my $apiset         = $solaris_client->apiset(
        category  => 'Host',
        interface => 'CLI',
        username  => 'root',
        password  => 'netapp',
        connid    => 'ssh-persist',
        set       => 'Solaris',
    );

    ## Set root privilegde
    $apiset->set_privilege(
        'privilege-level' => 'root',
        'answer-password' => 1
    );

    ## Set client connection timeout
    $apiset->set_timeout( 'connectrec-timeout' => 1500 );

    ## Return cleint APISet object
    return $apiset;
} ## end sub _create_solaris_client

1;
