# $Id: //depot/prod/test/main/lib/uACL/smbUtils.pm# $
#
## @summary  uACL smb utility class
## @status   public
## @pod      here
## @author   saurabhg@netapp.com, dl-nb-cr-qa@netapp.com

package uACL::smbUtils;

use Tharn;
use strict;
use warnings;
use uACL::uACLCommonParams;
use Data::Dumper;
use Options;
use NACL::C::Client;
use NACL::C::VserverCifs;
use NACL::C::Client::File;
use List::Util qw(shuffle);
use NACL::C::Client::Mount;
use MapShare qw(map_cifs_share);
use NACL::C::Client::Remote::File;
use NACL::CS::VserverServicesUnixUser;
use NACL::CS::VserverServicesUnixGroup;
use NACL::APISet::Exceptions::TimeoutException;
use NACL::APISet::Exceptions::ResponseException;
use NACL::GeneralUtils qw(random_name_generator);
use TestEnv::MinWin qw(execute_cmd minwin_compatible);
use NACL::APISet::Exceptions::ConnectionFailedException;
use Params::Validate qw[validate OBJECT SCALAR HASHREF UNDEF ARRAYREF];
Params::Validate::validation_options( allow_extra => 1 );

our (
    $ELEMENTS, $SETACL,           $NACACL,
    $PSGETSID, $SETACL_RIGHTS,    $WELLKNOWN_SIDS,
    $FC_SIDS,  $CIFS_BUFFER_SIZE, $REGEX,
    $Options,
);

sub new {

    my $class = shift;
    my %args  = validate(
        @_,
        {   smb   => { type => HASHREF,  optional => 0 },
            debug => { type => SCALAR,   optional => 0 },
            NODES => { type => ARRAYREF, optional => 1 },
        }
    );

    my $self = {};
    $self->{_owner} = $$;

    $self->{vserver}    = $args{smb}->{vserver};
    $self->{client}     = $args{smb}->{client};
    $self->{server}     = $args{smb}->{server};
    $self->{share}      = $args{smb}->{share};
    $self->{domain}     = $args{smb}->{domain};
    $self->{user}       = $args{smb}->{user};
    $self->{passwd}     = $args{smb}->{passwd};
    $self->{dbg}        = $args{debug};
    $self->{NODES}      = $args{NODES};
    $self->{mapdrive}   = undef;
    $self->{ci}         = undef;
    $self->{powershell} = undef;
    $self->{filelist}   = undef;
$self->{filehandlelist} = undef;

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::new::Contents passed to smbUtils constructor are: \n"
            . Dumper( $args{smb} ) );
    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::new::Contents of smbUtils Object are: \n"
            . Dumper($self) );
    bless( $self, $class );
    croak $@ if $@;

    $self->_init_globals();

    unless ( defined( $self->{vserver} ) ) {
        my $fetch_output = NACL::CS::VserverCifs->fetch(
            command_interface => $self->{NODES}->[0],
            filter            => { 'cifs-server' => $self->{server}, },
            requested_fields  => ['vserver'],
        );
        $self->{vserver} = $fetch_output->vserver();

        $Log->comment(
            "uACL::smbUtils::new:: Vserver for CIFS Server $self->{server} is found to be: "
                . $self->{vserver} );
    } ## end unless ( defined( $self->{...

    $self->_set_cifs_buffer($CIFS_BUFFER_SIZE);
    $self->{ci} = NACL::C::Client->new( name => $self->{client} );
    $self->_create_ps_apiset();

    $self->_map_cifs();
    defined( $self->{mapdrive} ) ? return $self : return undef;
} ## end sub new

sub getACL {

    my $self = shift;
    my %args = validate( @_, { path => { type => SCALAR, optional => 0 }, } );

    my $cmd
        = "get-acl $args{path}|format-list -Property Owner,Group;(get-acl -Path $args{path}).access|format-list";

    $self->{dbg} && $Log->comment( "Constructed Command is: " . $cmd );
    my $output;

    $output = $self->{powershell}->execute_command(
        command                => $cmd,
        return_response_object => 1
    );

    $self->{dbg} && $Log->comment( "Output:" . $output->{output} );

    return $self->_parse_getacls( $output->{output} );
} ## end sub getACL

sub VerifyACL {
    my $self = shift;
    my %args = validate(
        @_,
        {   path     => { type => SCALAR, optional => 0 },
            identity => { type => SCALAR, optional => 1 },
            has_allow =>
                { type => SCALAR, optional => 1, depends => ['identity'], },
            has_deny =>
                { type => SCALAR, optional => 1, depends => ['identity'], },
            has_owner => { type => SCALAR, optional => 1 },
            has_group => { type => SCALAR, optional => 1 },
        }
    );

    unless ( ( defined( $args{has_owner} ) )
        || ( defined( $args{has_group} ) )
        || ( defined( $args{has_allow} ) )
        || ( defined( $args{has_deny} ) ) )
    {
        $Log->comment(
            "uACL::smbUtils::VerifyACL:: Nothing specified to verify in CIFS ACL in any \'has_*\' parameter, exiting."
        );
        return 0;
    } ## end unless ( ( defined( $args{...

    my $current_acl = $self->getACL( path => $args{path} );

    unless ( length($current_acl) > 1 ) {
        $Log->comment(
            "uACL::smbUtils::VerifyACL => Unable to fetch current ACL on the file for verification"
        );
        return 0;
    } ## end unless ( length($current_acl...

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::VerifyACL => Current ACL on the file are:"
            . Dumper($current_acl) );

    if ( defined( $args{has_owner} ) ) {
        if ($args{has_owner} =~ m!\Q$current_acl->{ $ELEMENTS->{owner} }\E!i )
        {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Owner on the file matches "
                    . $args{has_owner} );
        } else {
            $Log->comment( "uACL::smbUtils::VerifyACL:: Owner on the file "
                    . $current_acl->{ $ELEMENTS->{owner} }
                    . " DOES NOT match with "
                    . $args{has_owner} );
            return 0;
        } ## end else [ if ( $args{has_owner} ...
    } ## end if ( defined( $args{has_owner...

    if ( defined( $args{has_group} ) ) {
        if ($current_acl->{ $ELEMENTS->{group} } =~ m!\Q$args{has_group}\E!i )
        {
            $Log->comment( "Group on the file matches " . $args{has_group} );
        } else {
            $Log->comment( "Group on the file \""
                    . $current_acl->{ $ELEMENTS->{group} }
                    . "\" DOES NOT match with "
                    . $args{has_group} );
            return 0;
        } ## end else [ if ( $current_acl->{ $ELEMENTS...
    } ## end if ( defined( $args{has_group...

    my $access_key = undef;
    my $id;

    if ( defined( $args{identity} ) ) {
        my $id_present = 0;
        $id = $args{identity};
        $id = $self->_sid2user($id) if ( $id =~ $REGEX->{sid} );
        $self->{dbg}
            && $Log->comment(
            'uACL::smbUtils::VerifyACL:: Identity being checked for Current ACL on the file is: '
                . $id );

        foreach my $key ( keys(%$current_acl) ) {
            if ( $key =~ m/^\Q$id\E$/i ) {
                $Log->comment(
                    "uACL::smbUtils::VerifyACL:: We have some ACEs present for $key"
                );
                $access_key = $key;
                $id_present = 1;
                last;
            } ## end if ( $key =~ m/^\Q$id\E$/i)
        } ## end foreach my $key ( keys(%$current_acl...
        unless ($id_present) {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: There are no ACEs for $id on file $args{path}"
            );
            return 0;
        } ## end unless ($id_present)
    } ## end if ( defined( $args{identity...

    if ( defined( $args{has_allow} ) ) {
        $self->{dbg}
            && $Log->comment( Dumper( $current_acl->{$access_key} ) );
        if (!exists(
                $current_acl->{$access_key}->{ $ELEMENTS->{rights} }->{Allow}
            )
            )
        {
            $self->{dbg}
                && $Log->comment(
                $current_acl->{$access_key}->{ $ELEMENTS->{rights} }
                    ->{Allow} );
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: File $args{path} does not have any Allow rights for $id"
            );
            return 0;
        } elsif (
            $self->_match_rights(
                $args{has_allow},
                $current_acl->{$access_key}->{ $ELEMENTS->{rights} }->{Allow}
            )
            )
        {

            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Allow rights on the file for $id matches "
                    . $args{has_allow} );
        } else {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Allow rights on the file for $id DO NOT match with "
                    . $args{has_allow} );
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Allow rights on the file are "
                    . $current_acl->{$access_key}->{ $ELEMENTS->{rights} }
                    ->{Allow} );

            return 0;
        } ## end else [ if ( !exists( $current_acl...
    } ## end if ( defined( $args{has_allow...

    if ( defined( $args{has_deny} ) ) {
        if (!exists(
                $current_acl->{$access_key}->{ $ELEMENTS->{rights} }->{Deny}
            )
            )
        {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: File $args{path} does not have any Deny rights for $args{identity}"
            );
            return 0;
        } elsif (
            $self->_match_rights(
                $args{has_deny},
                $current_acl->{$access_key}->{ $ELEMENTS->{rights} }->{Deny}
            )
            )
        {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Deny rights on the file for $args{identity} matches "
                    . $args{has_deny} );
        } else {
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Deny rights on the file for $args{identity} DO NOT match with "
                    . $args{has_deny} );
            $Log->comment(
                "uACL::smbUtils::VerifyACL:: Deny rights on the file are "
                    . $current_acl->{$access_key}->{ $ELEMENTS->{rights} }
                    ->{Deny} );

            return 0;
        } ## end else [ if ( !exists( $current_acl...
    } ## end if ( defined( $args{has_deny...

    return 1;
} ## end sub VerifyACL

sub SetACL {
    my $self = shift;
    my %args = validate(
        @_,
        {   path     => { type => SCALAR, optional => 0 },
            identity => { type => SCALAR, optional => 1 },
            allow_rights =>
                { type => SCALAR, optional => 1, depends => ['identity'], },
            deny_rights =>
                { type => SCALAR, optional => 1, depends => ['identity'], },
            owner       => { type => SCALAR, optional => 1 },
            group       => { type => SCALAR, optional => 1 },
            inheritance => {
                type     => SCALAR,
                optional => 1,
                regex    => $REGEX->{iflags},
                default  => 'np'
            },
            remove => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
                regex    => $REGEX->{boolean},
                depends  => ['identity'],
            },
            verify => {
                type     => SCALAR,
                optional => 1,
                default  => 1,
                regex    => $REGEX->{boolean}
            },
        }
    );

    unless ( ( defined( $args{allow_rights} ) )
        || ( defined( $args{deny_rights} ) )
        || ( defined( $args{owner} ) )
        || ( defined( $args{remove} ) )
        || ( defined( $args{group} ) ) )
    {
        $Log->comment(
            "uACL::smbUtils::SetACL: Nothing specified to set in CIFS ACL in any parameter, exiting."
        );
        return 0;
    } ## end unless ( ( defined( $args{...

    if ( defined( $args{owner} ) ) {
        my $sid;
        if ( _is_sid( $args{owner} ) eq 'y' ) {
            $sid = $args{owner};
        } else {
            $sid = $self->_get_sid( $args{owner} );

            if ( defined($sid) ) {
                $Log->comment(
                    "SID Obtained for \"$args{owner}\" is: " . $sid );
            } else {
                $Log->comment(
                    "Unable to get SID for account: " . $args{owner} );
                return 0;
            }
        } ## end else [ if ( _is_sid( $args{owner...
        my $cmd = $NACACL . " /set /o=" . $sid . " " . $args{path};
        (undef) = $self->{powershell}->execute_command(
            command                => $cmd,
            return_response_object => 1
        );
        $cmd = undef;

        if ( $args{verify} ) {
            if ($self->VerifyACL(
                    version   => "smb",
                    path      => $args{path},
                    has_owner => $args{owner}
                )
                )
            {
                $Log->comment(
                    "Successfully set the owner of file $args{path} to $args{owner}"
                );
            } else {
                return 0;
            }
        } ## end if ( $args{verify} )
    } ## end if ( defined( $args{owner...
    my $sid_flag;

    if ( defined( $args{group} ) ) {
        $sid_flag = _is_sid( $args{group} );

        my $cmd
            = $SETACL . " -on "
            . $args{path}
            . " -ot file -actn setgroup -grp \"n:"
            . $args{group} . ";s:"
            . $sid_flag . "\"";

        (undef) = $self->{powershell}->execute_command(
            command                => $cmd,
            return_response_object => 1
        );
        $cmd = undef;

        if ( $args{verify} ) {
            if ($self->VerifyACL(
                    version   => "smb",
                    path      => $args{path},
                    has_group => $args{group}
                )
                )
            {
                $Log->comment(
                    "Successfully set the group of file $args{path} to $args{group}"
                );
            } else {
                return 0;
            }
        } ## end if ( $args{verify} )
    } ## end if ( defined( $args{group...

    $sid_flag = _is_sid( $args{identity} )
        if ( defined( $args{identity} ) );

    my $error = 0;
    try {
        if ( defined( $args{allow_rights} ) ) {
            return 0 unless ( _is_valid_right( $args{allow_rights} ) );

            my $cmd
                = $SETACL . " -on "
                . $args{path}
                . " -ot file -actn ace -rec no -ace \"n:"
                . $args{identity} . ";p:"
                . $args{allow_rights} . ";i:"
                . $args{inheritance} . ";s:"
                . $sid_flag
                . "\" -ignoreerr";
            (undef) = $self->{powershell}->execute_command(
                command                => $cmd,
                return_response_object => 1
            );
            $cmd = undef;

            if ( $args{verify} ) {
                if ($self->VerifyACL(
                        version   => "smb",
                        path      => $args{path},
                        identity  => $args{identity},
                        has_allow => $args{allow_rights}
                    )
                    )
                {
                    $Log->comment(
                        'uACL::smbUtils::SetACL: Successfully set the allow rights '
                            . $args{allow_rights}
                            . ' on the file '
                            . $args{path}
                            . ' for user '
                            . $args{identity} );
                } else {
                    return 0;
                }
            } ## end if ( $args{verify} )
        } ## end if ( defined( $args{allow_rights...

        if ( defined( $args{deny_rights} ) ) {
            return 0 unless ( _is_valid_right( $args{deny_rights} ) );

            my $cmd
                = $SETACL . " -on "
                . $args{path}
                . " -ot file -actn ace -rec no -ace \"n:"
                . $args{identity} . ";p:"
                . $args{deny_rights}
                . ";m:deny;i:"
                . $args{inheritance} . ";s:"
                . $sid_flag . "\"";

            (undef) = $self->{powershell}->execute_command(
                command                => $cmd,
                return_response_object => 1
            );

            $cmd = undef;

            if ( $args{verify} ) {
                if ($self->VerifyACL(
                        version  => "smb",
                        path     => $args{path},
                        identity => $args{identity},
                        has_deny => $args{deny_rights}
                    )
                    )
                {
                    $Log->comment(
                        "uACL::smbUtils::SetACL: Successfully set the deny rights "
                            . $args{deny_rights}
                            . " on the file "
                            . $args{path}
                            . " for user "
                            . $args{identity} );
                } else {
                    return 0;
                }
            } ## end if ( $args{verify} )
        } ## end if ( defined( $args{deny_rights...

    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $err_msg = shift;
        my $msg     = $err_msg->get_response_object()->{output};

        $Log->comment(
            "uACL::smbUtils::SetACL: setacl.exe Response Exception:\n");
        $Log->comment(
            "uACL::smbUtils::SetACL: Setting ACLs failed due to Error:\n"
                . $msg );
        $error = 1;
    };
    if ( $args{remove} ) {
        my $sid_flag = _is_sid( $args{identity} );
        my $cmd
            = $SETACL . " -on "
            . $args{path}
            . " -ot file -actn trustee -trst \"n1:"
            . $args{identity} . ";s1:"
            . $sid_flag
            . ";ta:remtrst;w:dacl\" -rec no -ignoreerr";

        (undef) = $self->{powershell}->execute_command(
            command                => $cmd,
            return_response_object => 1
        );

        $cmd = undef;

        if ( $args{verify} ) {
            unless (
                $self->VerifyACL(
                    version   => 'smb',
                    path      => $args{path},
                    identity  => $args{identity},
                    has_allow => 'read',
                    has_deny  => 'write',
                )
                )
            {
                $Log->comment(
                    "Successfully removed ACE for $args{identity} on the file $args{path} "
                );
            } else {
                return 0;
            }
        } ## end if ( $args{verify} )
    } ## end if ( $args{remove} )

    return ($error) ? 0 : 1;
} ## end sub SetACL

sub CreateFiles {
    my $self = shift;
    my %args = validate(
        @_,
        {   count => { type => SCALAR,         optional => 0, default => 1 },
            path  => { type => SCALAR | UNDEF, optional => 1 },
            preserve => { type => SCALAR, optional => 0 },
        }
    );

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::create_files: Parameters obtained are count= $args{count}, preserve as $args{preserve}"
        );

    my ( $filename, $created, @namearray );

    for ( my $i = 0; $i < $args{count}; $i++ ) {
        $filename = "uACL_automation_" . $self->_getname();

        if ( defined( $args{path} ) ) {
            $self->{dbg}
                && $Log->comment(
                "uACL::smbUtils::create_files: Path received for file creation is $args{path}"
                );
            $filename
                = $self->{mapdrive} . "\\" . $args{path} . "\\" . $filename;
        } else {
            $filename = $self->{mapdrive} . "\\" . $filename;
        }
        $self->{dbg}
            && $Log->comment(
            "uACL::smbUtils::create_files: Generated Filename is: "
                . $filename );

        my $content = time();
        $self->{ci}->execute( command => "echo $content > $filename" );
        $Log->step(
            "uACL::smbUtils::create_files: Created file $filename on CIFS client"
        );

        if ( $self->_is_present( path => $filename ) ) {
            $created++;
            push( @namearray, $filename );
        } else {
            $Log->error(
                "uACL::smbUtils::create_files: Failed to create file $filename on CIFS client"
            );
            return undef;
        } ## end else [ if ( $self->_is_present...
    } ## end for ( my $i = 0; $i < $args...

    $Log->comment(
        "uACL::smbUtils::create_files: Successfully created $created files out of requested $args{count} files"
    );

    $args{preserve} || $self->_addto_filelist( \@namearray );

    return ( $args{count} == 1 ) ? $namearray[0] : \@namearray;
} ## end sub CreateFiles

sub CreateFiles_Lock {
    my $self = shift;
    my %args = validate(
        @_,
        {   count       => { type => SCALAR, optional => 0, default => 1 },
            access_mode => {
                type     => SCALAR,
                optional => 1,
                default  => 'read-write'
            },
            open_flags => {
                type     => SCALAR,
                optional => 1,
                default  => 'create-new'
            },
            share_mode => {
                type     => SCALAR,
                optional => 1,
            },
            attributes => {
                type     => SCALAR,
                optional => 1,
            },
            path     => { type => SCALAR | UNDEF, optional => 1 },
            filename => { type => SCALAR | UNDEF, optional => 1 },
            preserve => { type => SCALAR,         optional => 0 },
        }
    );

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::CreateFiles_Lock: Parameters obtained are count= $args{count}, preserve as $args{preserve}"
        );

    my ($file_handle, @fh_array, @namearray, $filename,
        $created,     %options,  @filenames
    );
    my $doc_flag = 0;

    for ( my $i = 0; $i < $args{count}; $i++ ) {
        $filename
            = ( defined( $args{filename} ) )
            ? $args{filename}
            : "uACL_automation_testfile" . '_' . $self->_getname();

        $filename = $filename . '_' . $self->_getname()
            if ( defined( $args{filename} ) && $i > 0 );
        if ( defined( $args{path} ) ) {
            $self->{dbg}
                && $Log->comment(
                "path received for file creation is $args{path}");
            $filename
                = $self->{mapdrive} . "\\" . $args{path} . "\\" . $filename;
        } else {
            $filename = $self->{mapdrive} . "\\" . $filename;
        }
        $self->{dbg}
            && $Log->comment( "Generated Filename is: " . $filename );

        $self->{dbg}
            && $Log->comment(
            "uACL::smbUtils::CreateFiles_Lock: Parameters obtained for file creation are access_mode= $args{access_mode}, open_flags= $args{open_flags}"
            );

        $options{share_mode} = $args{share_mode}
            if ( defined( $args{share_mode} ) );

        if ( defined( $args{attributes} ) ) {
            $options{attributes} = $args{attributes};

            if ( $options{attributes} eq 'FILE_FLAG_DELETE_ON_CLOSE' ) {
                $doc_flag++;
            }
        } ## end if ( defined( $args{attributes...

        $file_handle = NACL::C::Client::Remote::File->open(
            command_interface => $self->{ci},
            file_name         => $filename,
            arguments         => {
                access_mode => $args{access_mode},
                open_flags  => $args{open_flags},
		%options
            },
        );

        if ( defined($file_handle) ) {
            $Log->step(
                "uACL::smbUtils::CreateFiles_Lock: Created file $filename on CIFS client"
            );
            $created++;
            push( @fh_array,  $file_handle );
            push( @namearray, $file_handle->file_name) if(!$doc_flag);
        } else {
            $Log->error(
                "uACL::smbUtils::CreateFiles_Lock: Failed to create file $filename on CIFS client"
            );
            return undef;
        } ## end else [ if ( defined($file_handle...
    } ## end for ( my $i = 0; $i < $args...

    $Log->comment(
        "uACL::smbUtils::CreateFiles_Lock: Successfully created $created files out of requested $args{count} files"
    );

    $self->_filehandle_list( \@fh_array );
    (@filenames) = map { ( $_ !~ /.*\\.*:/g ) ? $_ : () } @namearray;
    $args{preserve} || $self->_addto_filelist( \@filenames );

    return ( $args{count} == 1 ) ? $fh_array[0] : \@fh_array;

} ## end sub CreateFiles_Lock

sub execute_cmd {

    my $self = shift;
    my %args = validate(
        @_,
        {   command => { type => SCALAR, optional => 0, default => 1 },
            powershell   => { type => SCALAR | UNDEF, optional => 1 },
            timeout      => { type => SCALAR | UNDEF, optional => 1 },
            match_string => { type => SCALAR | UNDEF, optional => 1 },
            verify       => { type => SCALAR | UNDEF, optional => 1 },
        }
    );

    my ( $default_timeout, $timeout, $obj, $method, $output, %param );

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::execute_cmd: Parameters obtained are command= $args{command}, match string as $args{match_string}"
        );

    if ( defined( $args{powershell} ) ) {
        $default_timeout = 600;
        $obj             = $self->{powershell};
        $method          = 'execute_command';
    } else {
        $default_timeout = 60;
        $obj             = $self->{ci};
        $method          = 'execute';
    }

    $timeout = defined( $args{timeout} ) ? $args{timeout} : $default_timeout;
    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::execute_cmd: timeout value: $timeout");

    ( defined( $args{powershell} ) )
        ? $obj->set_timeout( 'connectrec-timeout' => $timeout )
        : $param{'timeout'} = $timeout;
    $param{'command'} = $args{command};

    $output = $obj->$method(%param);
    return $output if ( !defined( $args{verify} ) );

    unless ( defined( $args{match_string} ) ) {
        $Log->error(
            "uACL::smbUtils::execute_cmd: Param match string is not passed to verify the output"
        );
        return 0;
    } ## end unless ( defined( $args{match_string...

    $args{match_string} =~ s/\(/\\(/g;
    $args{match_string} =~ s/\)/\\)/g;

    if ( $output =~ /$args{match_string}/g ) {
        $Log->comment(
            "uACL::smbUtils::execute_cmd: Output and match string passed are verified successfully");
    } else {
        $Log->error(
            "uACL::smbUtils::execute_cmd: Verification failed for the match string passed\n"
                . "Actual output : $output" );
        return 0;
    } ## end else [ if ( $output =~ /$args{match_string}/g)

    return 1;
} ## end sub execute_cmd

sub destroy {
    my $self = shift;

    $Log->step("Closing the file handle opened before deleting the file");
    $self->_close_filehandle();

    $Log->step("Deleting the files that were created with preserve set to 0");
    $self->_delete_filelist();

    $Log->step(
        "Unmapping the drive $self->{mapdrive} on Windows client $self->{client}"
    );

    try {
        NACL::C::Client::Mount->delete(
            command_interface => $self->{ci},
            mount_point       => $self->{mapdrive}
        );
    } ## end try
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $err_msg = shift;
        $Log->comment(
            'uACL::smbUtils::destroy:: Exception when unmapping on Windows Client'
        );
        $Log->comment( "Error Message is:\n" . $err_msg->text() );
        return 0;
    };

    return 1;
} ## end sub destroy

sub _map_cifs {
    my $self = shift;
    eval {
        $self->{mapdrive} = map_cifs_share(
            server   => $self->{server},
            client   => $self->{client},
            domain   => $self->{domain},
            username => $self->{user},
            password => $self->{passwd},
            share    => $self->{share},
            drive    => '*'
        );
    };

    $Log->error($@) if ($@);

    unless ( defined( $self->{mapdrive} ) ) {
        $Log->error('uACL::smbUtils::_map_cifs: Unable to Map CIFS Share');
        return 0;
    } else {
        $Log->comment(
            'uACL::smbUtils::_map_cifs: Successfully mapped the drive to '
                . $self->{mapdrive}
                . ' on client' );
    } ## end else
    return 1;
} ## end sub _map_cifs

sub _filehandle_list {
    my ( $self, $names ) = @_;
    my @fh_list;

    $self->{dbg}
        && $Log->comment(
        "File handles need to be closed before deletion are:");
    $self->{dbg} && $Log->comment($_) for (@$names);

    if ( defined( $self->{filehandlelist} ) ) {
        @fh_list = @{ $self->{filehandlelist} };
    }

    @fh_list = ( @fh_list, @$names );
    $self->{filehandlelist} = \@fh_list;
    $self->{dbg}
        && $Log->comment(
        "List of file handles to be deleted stored in object are: "
            . Dumper( $self->{filehandlelist} ) );
    return 1;
} ## end sub _filehandle_list

sub _addto_filelist {
    my ( $self, $names ) = @_;
    my @filearray;

    $self->{dbg} && $Log->comment("Names received for deletion later are:");
    $self->{dbg} && $Log->comment($_) for (@$names);

    if ( defined( $self->{filelist} ) ) {
        @filearray = @{ $self->{filelist} };
    }

    @filearray = ( @filearray, @$names );
    $self->{filelist} = \@filearray;
    $self->{dbg}
        && $Log->comment(
        "Contents of file delete list stored in object are: "
            . Dumper( $self->{filelist} ) );

    return 1;
} ## end sub _addto_filelist

sub _close_filehandle {
    my $self = shift;

    return 0 unless ( defined( $self->{filehandlelist} ) );

    foreach ( @{ $self->{filehandlelist} } ) {
        # Handled the fh close function call with eval/$@ temporarily. Since
        # NACL:C:Client:File.pm not returning any flag when success or
        # failure which will be fixed in future sprint. Burt id#694327
        # Need to revisit this part once the burt is fixed
        eval { defined($_) ? $_->close : (); };
        if ( $@ =~ /The handle is invalid/ ) {
            $Log->comment('File handle is already closed');
        }
        elsif ($@) {
            $Log->warn( 'Error:' . $@ );
        }
        else {
            $self->{dbg}
              && $Log->comment('Closed the file handle successfully!!!!!');
        }
    } ## end foreach ( @{ $self->{filehandlelist...

    #    (undef) = map {defined($_)? $_->close:()} @{ $self->{filehandlelist} };
    $self->{filehandlelist} = undef;

    return 1;
} ## end sub _close_filehandle

sub _delete_filelist {
    my $self = shift;

    return 0 unless ( defined( $self->{filelist} ) );

    (undef) = map {
        $self->{ci}->execute(
            command => ( $self->_is_dir( path => $_ ) )
            ? "rmdir /S /Q $_"
            : "del $_"
        );
    } @{ $self->{filelist} };
    $self->{filelist} = undef;

    return 1;
} ## end sub _delete_filelist

sub _getname {
    my $self = shift;
    my @arr = split( //, random_name_generator( size => 1 ) );
    return join( '', shuffle(@arr) );

} ## end sub _getname

sub _parse_getacls {
    my ( $self, $output ) = @_;

    my $acl;
    $acl->{Owner} = ["None"];
    $acl->{Group} = ["None"];

    my @lines = split( /\n/, $output );
    my ( $ace, $sid, $rights, $acetype );

    foreach my $line (@lines) {
        if ( $line =~ m/(.+)/ ) {

            if ( $line =~ m/($ELEMENTS->{rights})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment(
                    'uACL::smbUtils::_parse_getacls:: Rights Obtained for line "'
                        . $line . '" is '
                        . $3 );
                $rights = $3;
            } elsif ( $line =~ m/($ELEMENTS->{type})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Type Obtained for line \""
                        . $line
                        . "\" is "
                        . $3 );
                $acetype = $3;
                $ace->{ $ELEMENTS->{rights} }->{$acetype} = $rights;
            } elsif ( $line =~ m/($ELEMENTS->{sid})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: SID Obtained for line \""
                        . $line
                        . "\" is "
                        . $3 );
                $sid = $3;
                $sid = $self->_sid2user($sid) if ( $sid =~ $REGEX->{sid} );
            } elsif ( $line =~ m/($ELEMENTS->{inherited})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment( "\n Inherited Obtained for line \"" 
                        . $line 
                        . "\" is "
                        . $3 );
                $ace->{ $ELEMENTS->{inherited} } = $3;
            } elsif ( $line =~ m/($ELEMENTS->{inheritflag})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Inheritflag Obtained for line \""
                        . $line
                        . "\" is "
                        . $3 );
                $ace->{ $ELEMENTS->{inheritflag} } = $3;
            } elsif ( $line =~ m/($ELEMENTS->{propagate})(\s+:\s+)(.+)/ ) {

                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Propogate Obtained for line \""
                        . $line
                        . "\" is "
                        . $3 );
                $ace->{ $ELEMENTS->{propagate} } = $3;

                #This indicates the end of this ace, now constructing the ace
                if ( exists( $acl->{$sid} ) ) {
                    $self->{dbg}
                        && $Log->comment(
                        "uACL::smbUtils::_parse_getacls:: Acetype is $acetype"
                        );
                    $self->{dbg}
                        && $Log->comment(
                        "uACL::smbUtils::_parse_getacls:: Pre-existing value is: "
                            . Dumper( $acl->{sid} ) );
                    $acl->{$sid}->{ $ELEMENTS->{rights} }->{$acetype}
                        = $rights;
                } else {
                    $acl->{$sid} = $ace;
                }
                $ace = undef;
            } elsif ( $line =~ m/($ELEMENTS->{owner})(\s+:\s+)(.+)/ ) {
                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Owner Obtained for: "
                        . $line . " is "
                        . $3 );
                $acl->{Owner} = $3;
            } elsif ( $line =~ m/($ELEMENTS->{group})(\s+:\s+)(.+)/ ) {
                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Group Obtained for: "
                        . $line . " is "
                        . $3 );
                $acl->{Group} = $3;
            } else {

                #this is a spill over line, can be possible only in case of
                #Filesystem rights, so append it to last parsed rights value
                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: In spill over, value obtained is: "
                        . $line );
                $line =~ s/\s*//g;
                $rights .= $line;
                $self->{dbg}
                    && $Log->comment(
                    "uACL::smbUtils::_parse_getacls:: Now value of rights is $rights"
                    );
            } ## end else [ if ( $line =~ m/($ELEMENTS->{rights})(\s+:\s+)(.+)/)
        } ## end if ( $line =~ m/(.+)/ )
    } ## end foreach my $line (@lines)

    #Assumption: we get ordered output every time
    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::_parse_getacls:: Parsed ACL structure is: "
            . Dumper($acl) );
    return $acl;
} ## end sub _parse_getacls

sub _sid2user {
    my $self = shift;
    my ($sid) = @_;

    $self->{dbg}
        && $Log->comment(
        'uACL::smbUtils::_sid2user:: SID obtained for translation ' . $sid );
    return $FC_SIDS->{$sid} if ( exists( $FC_SIDS->{$sid} ) );

    return $WELLKNOWN_SIDS->{$sid} if ( exists( $WELLKNOWN_SIDS->{$sid} ) );

    my $node = $self->{NODES}->[0];
    $node->privilege_level('diag')
        unless ( $node->privilege_level() =~ $REGEX->{priv} );

    my $cmd
        = 'diag secd authentication sid-to-unix-name -node '
        . $node->node()
        . ' -vserver '
        . $self->{vserver}
        . ' -sid '
        . $sid;
    my $output = $node->apiset->execute_raw_command( command => $cmd );
    $output =~ s/\s+//g;

    if ( $output =~ $REGEX->{winname} ) {
        my $username = $5;
        my $domain   = $3;

        unless ( $username =~ $REGEX->{nfsuser} ) {
            $username = $domain . '\\' . $username;
        } else {
            $self->{dbg}
                && $Log->comment(
                "uACL::smbUtils::_sid2user:: User $username is a NFS user, hence skipping appending domain $domain to it"
                );
        } ## end else
        $self->{dbg}
            && $Log->comment(
            "uACL::smbUtils::_sid2user:: Username corresponding to SID $sid is found to be $username"
            );

        return $username;
    } ## end if ( $output =~ $REGEX...

    $Log->comment(
        "uACL::smbUtils::VerifyACL::_sid2user:: Unable to find username corresponding to SID $sid"
    );
    return $sid;

} ## end sub _sid2user

sub _init_globals() {

    $Options = new Options();;

    $Options->process();
    my $acl_unix_user = $Options->get(option => 'ACL_UNIX_USER');
    my $acl_cifs_user = $Options->get(option => 'ACL_CIFS_USER');

    #Intialize the ACL Elements
    $ELEMENTS->{owner}       = "Owner";
    $ELEMENTS->{group}       = "Group";
    $ELEMENTS->{sid}         = "IdentityReference";
    $ELEMENTS->{rights}      = "FileSystemRights";
    $ELEMENTS->{type}        = "AccessControlType";
    $ELEMENTS->{inherited}   = "IsInherited";
    $ELEMENTS->{inheritflag} = "InheritanceFlags";
    $ELEMENTS->{propagate}   = "PropagationFlags";

    #Used compiled regex for frequent ones
    $REGEX->{boolean}  = qr/0|1/;
    $REGEX->{sid}      = qr/(^S\-1\-(\d|\-)+)/;
    $REGEX->{winname}  = qr/(.+)(DomainName:)(.+)(WindowsName:)(.+)/;
    $REGEX->{priv}     = qr/test|diag/i;
    $REGEX->{nfsuser}  = qr/$acl_unix_user\d{1,3}/i;
    $REGEX->{cifsuser} = qr/$acl_cifs_user\d{1,3}/i;
    $REGEX->{anyuser}  = qr/u?acl(n|ci)fsuser\d{1,3}/i;
    $REGEX->{anything} = qr/.+?/;
    $REGEX->{iflags}   = qr/(so|sc|io|np|,)+?/;

    #Increase CIFS Buffer size on filer to 64K
    $CIFS_BUFFER_SIZE = 65535;

    #Path to Utilities
    $SETACL   = 'X:\bin\SetACL.exe';
    $NACACL   = 'X:\bin\cifs\security\NACacls.exe';
    $PSGETSID = 'X:\pstools\psgetsid.exe';

    #SetACL Rights
    $SETACL_RIGHTS->{traverse}    = 'ExecuteFile';
    $SETACL_RIGHTS->{list_dir}    = 'ReadData';
    $SETACL_RIGHTS->{read_attr}   = 'ReadAttributes';
    $SETACL_RIGHTS->{read_ea}     = 'ReadExtendedAttributes';
    $SETACL_RIGHTS->{add_file}    = 'CreateFiles';
    $SETACL_RIGHTS->{add_subdir}  = 'AppendData';
    $SETACL_RIGHTS->{write_attr}  = 'WriteAttributes';
    $SETACL_RIGHTS->{write_ea}    = 'WriteExtendedAttributes';
    $SETACL_RIGHTS->{del_child}   = 'DeleteSubdirectoriesAndFiles';
    $SETACL_RIGHTS->{delete}      = 'delete';
    $SETACL_RIGHTS->{read_dacl}   = 'ReadPermissions';
    $SETACL_RIGHTS->{write_dacl}  = 'ChangePermissions';
    $SETACL_RIGHTS->{write_owner} = 'TakeOwnership';
    $SETACL_RIGHTS->{read}        = 'read';
    $SETACL_RIGHTS->{write}       = 'write';
    $SETACL_RIGHTS->{list_folder} = 'ReadAndExecute';
    $SETACL_RIGHTS->{read_ex}     = 'ReadAndExecute';
    $SETACL_RIGHTS->{change}      = 'Modify';
    $SETACL_RIGHTS->{full}        = 'FullControl';
    $SETACL_RIGHTS->{profile}     = 'ChangePermissions';

    _init_wellknown_sids();

    return 1;
} ## end sub _init_globals()

sub _get_sid {
    my $self    = shift;
    my $account = shift;

    my $cmd  = $PSGETSID . " " . $account;
    my $resp = $self->{powershell}->execute_command(
        command                => $cmd,
        return_response_object => 1
    );

    my @output = split( /\s+/i, $resp->{output} );

    foreach my $line (@output) {
        return $line if ( $line =~ $REGEX->{sid} );
    }

    return undef;
} ## end sub _get_sid

sub _is_sid {
    return ( shift =~ $REGEX->{sid} ) ? 'y' : 'n';
}

sub _is_valid_right {
    my $rights = shift;
    my $valid  = 1;

    my $end = sub {
        my $right = shift;
        $Log->comment(
            "uACL::smbUtils::_is_valid_right:Invalid right $right specified!\nOnly following rights are allowed:\n"
        );
        map { $Log->comment("\t$_\n") } keys(%$SETACL_RIGHTS);
        $valid = 0;
    };

    my @given_rights = _get_tokens($rights);
    map { \&$end($_) unless ( exists( $SETACL_RIGHTS->{$_} ) ) }
        @given_rights;

    return $valid;
} ## end sub _is_valid_right

sub CreateDirs {
    my $self = shift;

    my %args = validate(
        @_,
        {   count => { type => SCALAR,         optional => 0, default => 1 },
            path  => { type => SCALAR | UNDEF, optional => 1 },
            preserve => { type => SCALAR, optional => 0 },
        }
    );

    $self->{dbg}
        && $Log->comment(
        "uACL::smbUtils::CreateDirs:: We have count= $args{count}, preserve as $args{preserve}"
        );

    my ( $dirname, $created, @namearray );

    for ( my $i = 0; $i < $args{count}; $i++ ) {
        $dirname = $self->{mapdrive} . '\\';

        if ( defined( $args{path} ) ) {
            $self->{dbg}
                && $Log->comment(
                'uACL::smbUtils::CreateDirs:: path received for directory creation is '
                    . $args{path} );
            $dirname .= $args{path} . '\\';
        } ## end if ( defined( $args{path...

        $dirname .= 'uACL_automation_' . $self->_getname();
        $self->{dbg}
            && $Log->comment(
            'uACL::smbUtils::CreateDirs:: Generated directory path is: '
                . $dirname );

        if ( $self->_mkdir($dirname) ) {
            $Log->step(
                "uACL::smbUtils::CreateDirs:: Created Directory $dirname on CIFS client"
            );
            $created++;
            push( @namearray, $dirname );
        } else {
            $Log->error(
                "uACL::smbUtils::CreateDirs:: Failed to create directory $dirname on CIFS client"
            );
            return undef;
        } ## end else [ if ( $self->_mkdir($dirname...
    } ## end for ( my $i = 0; $i < $args...

    $Log->comment(
        "uACL::smbUtils::CreateDirs:: Successfully created $created directories out of requested $args{count} directories"
    );

    $args{preserve}
        || $self->_addto_filelist( \@namearray );

    return ( $args{count} == 1 ) ? $namearray[0] : \@namearray;
} ## end sub CreateDirs

sub _is_present {

    my $self = shift;
    my %args = validate( @_, { path => { type => SCALAR, optional => 0 }, } );

    my $response = $self->{powershell}->execute_command(
        command                => "Test-Path $args{path}",
        return_response_object => 1
    );

    return ( $response->{output} =~ m/true/i ) ? 1 : 0;
} ## end sub _is_present

sub _mkdir {
    my $self = shift;
    my ($path) = @_;

    my $output = $self->{ci}->execute( command => "mkdir $path" );

    if ( $output =~ m/\w+/ ) {
        $Log->comment(
            "uACL::smbUtils::_mkdir:: Failed to Create Directory $path\nError: $output"
        );
        return 0;
    } ## end if ( $output =~ m/\w+/)
    return 1;
} ## end sub _mkdir

sub _match_rights {
    my $self = shift;
    my ( $to_match, $all_rights ) = @_;
    my $present_rights;
    my $match = 1;

    (undef)
        = map { $present_rights->{ lc($_) } = 1; } _get_tokens($all_rights);

    $self->{dbg}
        && $Log->comment(
        'uACL::smbUtils::_match_rights:: User rights are: ' . $all_rights );
    $self->{dbg}
        && $Log->comment( 'uACL::smbUtils::_match_rights:: All rights are: '
            . Dumper($present_rights) );

    (undef) = map {
        $match = 0
            if ( $match
            && !
            exists( $present_rights->{ lc( $SETACL_RIGHTS->{ lc($_) } ) } )
            && !exists( $present_rights->{ lc($_) } ) )
    } _get_tokens($to_match);

    return $match;
} ## end sub _match_rights

sub _get_tokens {
    my $line = shift;

    $line =~ s/\s+//g;
    return split( /,/, $line );
} ## end sub _get_tokens

sub _get_nfs_sid {

    #subroutine to generate the SID for given NFS username

    my $self = shift;
    my ( $nfsname, $is_grp ) = @_;

    unless ($is_grp) {
        my $output = NACL::CS::VserverServicesUnixUser->fetch(
            command_interface => $self->{NODES}->[0],
            filter            => {
                'vserver' => $self->{vserver},
                'user'    => $nfsname,
            },
            requested_fields => ['id'],
        );

        my $uid = $output->id();
        $self->{dbg}
            && $Log->comment(
            "uACL::smbUtils::_get_nfs_sid: ID for NFS user \"$nfsname\" is $uid"
            );

        return 'S-1-8-' . $uid;
    } else {
        my $output = NACL::CS::VserverServicesUnixGroup->fetch(
            command_interface => $self->{NODES}->[0],
            filter            => {
                'vserver' => $self->{vserver},
                'name'    => $nfsname,
            },
            requested_fields => ['id'],
        );

        my $gid = $output->id();
        $self->{dbg}
            && $Log->comment(
            "uACL::smbUtils::_get_nfs_sid: ID for NFS group \"$nfsname\" is $gid"
            );

        return 'S-1-264-' . $gid;
    } ## end else

    return undef;
} ## end sub _get_nfs_sid

sub CopyACL {
    my $self = shift;
    my %args = validate(
        @_,
        {   from    => { type => SCALAR, optional => 0 },
            to      => { type => SCALAR, optional => 0 },
            version => { type => SCALAR, optional => 1 },
            verify  => { type => SCALAR, optional => 1, default => 0 },
        }
    );

    my $cmd = "get-acl $args{from}|set-acl -Path $args{to}";
    my ( $response, $xcption );

    try {
        $response = $self->{powershell}->execute_command(
            command                => $cmd,
            return_response_object => 1
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $err_msg = shift;
        my $msg     = $err_msg->get_response_object()->{output};

        if ( $msg
            =~ /'Set-Acl : Some or all identity references could not be translated.'/
            )
        {
            $xcption = 0;
        } else {
            $Log->comment(
                " uACL::smbUtils::CopyACL: set-acl Response Exception:\n");
            $Log->comment( "Setting ACLs failed due to Error:\n" . $msg );
            $xcption = 1;
        } ## end else [ if ( $msg =~ ...
    } ## end with
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $err_msg = shift;
        $Log->comment(
            " uACL::smbUtils::CopyACL: set-acl ConnectionFailed Exception:\n"
        );
        $Log->comment( "Setting ACLs failed due to Connetcion failure:\n "
                . $err_msg->get_response_object()->{output} );
        $xcption = 1;
    } ## end with
    catch NACL::APISet::Exceptions::TimeoutException with {
        my $err_msg = shift;
        $Log->comment(
            "uACL::smbUtils::CopyACL: set-acl Timeout Exception:\n");
        $Log->comment( "Setting ACLs failed due to Windows Client Timeout:\n"
                . $err_msg );
        $xcption = 1;
    };

    $self->_create_ps_apiset() if ($xcption);

    if ( $args{verify} ) {
        my $reffile_aces
            = $self->_count_aces( path => $args{from}, usertype => 'any' );
        my $trgtfile_aces
            = $self->_count_aces( path => $args{to}, usertype => 'any' );

        if ( $reffile_aces == $trgtfile_aces ) {
            $Log->comment(
                "uACL::smbUtils::CopyACL: Explicit Verification Passed, $args{to} has exactly same number of users ACEs set as that on refrence file"
            );
            return 1;
        } else {
            $Log->comment(
                "uACL::smbUtils::CopyACL: Aces set on reference file $args{from} are $reffile_aces"
            );
            $Log->comment(
                "uACL::smbUtils::CopyACL: Aces set on destination file $args{to} are $trgtfile_aces"
            );
            return 0;
        } ## end else [ if ( $reffile_aces == ...

    } else {
        return ( $xcption || $response->{output} =~ $REGEX->{anything} )
            ? 0
            : 1;
    }
} ## end sub CopyACL

sub _init_wellknown_sids {

    #Well known CIFS SIDs
    $WELLKNOWN_SIDS->{'S-1-0-0'}      = 'NULL SID';
    $WELLKNOWN_SIDS->{'S-1-2-0'}      = 'Local';
    $WELLKNOWN_SIDS->{'S-1-2-1'}      = 'Console logon';
    $WELLKNOWN_SIDS->{'S-1-5-32'}     = 'Builtin SID String';
    $WELLKNOWN_SIDS->{'S-1-5-2'}      = 'NT AUTHORITY\Network';
    $WELLKNOWN_SIDS->{'S-1-5-7'}      = 'NT AUTHORITY\ANONYMOUS LOGON';
    $WELLKNOWN_SIDS->{'S-1-5-11'}     = 'NT AUTHORITY\Authenticated Users';
    $WELLKNOWN_SIDS->{'S-1-5-32-544'} = 'BUILTIN\Administrators';
    $WELLKNOWN_SIDS->{'S-1-5-32-545'} = 'BUILTIN\Users';
    $WELLKNOWN_SIDS->{'S-1-5-32-546'} = 'BUILTIN\Guests';
    $WELLKNOWN_SIDS->{'S-1-5-32-547'} = 'BUILTIN\Power users';
    $WELLKNOWN_SIDS->{'S-1-5-32-548'} = 'Account Operators';
    $WELLKNOWN_SIDS->{'S-1-5-32-549'} = 'Server Operators';
    $WELLKNOWN_SIDS->{'S-1-5-32-550'} = 'BUILTIN\Print Operators';
    $WELLKNOWN_SIDS->{'S-1-5-32-551'} = 'BUILTIN\Backup Operators';
    $WELLKNOWN_SIDS->{'S-1-5-32-552'} = 'BUILTIN\Replicators';
    $WELLKNOWN_SIDS->{'S-1-5-64-10'}  = 'NT AUTHORITY\NTLM Authentication';
    $WELLKNOWN_SIDS->{'S-1-5-64-14'} = 'NT AUTHORITY\SChannel Authentication';
    $WELLKNOWN_SIDS->{'S-1-5-64-21'} = 'NT AUTHORITY\Digest Authentication';
    $WELLKNOWN_SIDS->{'S-1-5-80'}    = 'NT Service';
    $WELLKNOWN_SIDS->{'S-1-16-0'}    = 'Untrusted Mandatory level';
    $WELLKNOWN_SIDS->{'S-1-5-32-552'} = 'BUILTIN\Replicator';

    #SIDs that need Full Control Access
    $FC_SIDS->{'S-1-520-0'} = 'OWNER@';
    $FC_SIDS->{'S-1-520-1'} = 'GROUP@';
    $FC_SIDS->{'S-1-520-2'} = 'EVERYONE@';

    #$FC_SIDS->{'S-1-1-0'}   = 'EVERYONE';

    #Well known SIDS that cannot be added
    #$WELLKNOWN_SIDS->{'S-1-520-3'} = 'INTERACTIVE@';
    #$WELLKNOWN_SIDS->{'S-1-520-4'} = 'NETWORK@';
    #$WELLKNOWN_SIDS->{'S-1-520-5'} = 'DIALUP@';
    #$WELLKNOWN_SIDS->{'S-1-520-6'} = 'ANONYMOUS@';
    #$WELLKNOWN_SIDS->{'S-1-520-7'} = 'AUTHENTICATED@';
    #$WELLKNOWN_SIDS->{'S-1-520-8'} = 'SERVICE@';
    #$WELLKNOWN_SIDS->{'S-1-0'}        = 'Null Authority';
    #$WELLKNOWN_SIDS->{'S-1-1'}        = 'World Authority';
    #$WELLKNOWN_SIDS->{'S-1-2'}        = 'Local Authority';

    return 1;
} ## end sub _init_wellknown_sids

sub _add_wellknown_sids {

    my $self = shift;
    my %args = validate(
        @_,
        {   path     => { type => SCALAR, optional => 0 },
            nfs_only => { type => SCALAR, optional => 1, default => 0 },
        }
    );

    unless ( $args{nfs_only} ) {
        my $addsid = sub {
            shift;
            my $result;
            try {
                $result = $self->SetACL(
                    path         => $args{path},
                    identity     => $_,
                    allow_rights => 'write',
                    inheritance => 'so,sc',
                    verify       => 1
                );
            } ## end try
            catch NACL::APISet::Exceptions::ResponseException with {
                $Log->comment(
                    "uACL::smbUtils::add_wellknown_sids: SID $_ cannot be added"
                );
            };
            $Log->warn( 'Verification failed for SID' . $_ )
                unless ($result);
        };

        (undef) = map { \&$addsid($_) } keys( %{$WELLKNOWN_SIDS} );
    } else {

        #This is needed as inherited ACEs account for the child objects
        (undef) = map {
            $self->SetACL(
                path         => $args{path},
                identity     => $_,
                allow_rights => 'full',
                inheritance => 'so,sc',
                verify       => 0
                )
        } keys( %{$FC_SIDS} );
    } ## end else
    return 1;
} ## end sub _add_wellknown_sids

sub _create_ps_apiset {
    my $self = shift;

    my $winhost = host( $self->{client} );
    unless ( $winhost->hosttype() =~ /windows/im
        && minwin_compatible( $self->{client} ) )
    {
        $Log->error( "uACL::smbUtils::_create_ps_apiset: CONFIG: The host "
                . $winhost->hostname()
                . " is not MinWin compatible, exiting." );
        return 0;
    } ## end unless ( $winhost->hosttype...

    try {
        $self->{powershell} = NACL::APISet->new(
            hostobj   => host( $self->{client} ),
            category  => "Host",
            interface => "CLI",
            set       => "PowerShell",
            connid    => 'login',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ConnectionFailedException with {
        my $err_msg = shift;
        $Log->comment(
            'uACL::smbUtils::_create_ps_apiset:: Unable to login to Powershell on Windows Client'
        );
        $Log->comment(
            'Make sure that default_username & default_password are set in corresponding Windows Client Tharnhost file'
        );
        $Log->comment( "Error Message is:\n" . $err_msg->text() );
        return 0;
    };

    #$self->{powershell}->timeout_scaling_factor('4');
    $self->{powershell}->set_timeout( 'connectrec-timeout' => 600 );

    return ( defined( $self->{powershell} ) ) ? 1 : 0;
} ## end sub _create_ps_apiset

sub _set_cifs_buffer {
    my $self = shift;
    my $regex_setenv
        = qr/sysvar.nblade.debug.cifs.max_buffer_size_server: 65535/m;
    my $display_warning
        = 'uACL::smbUtils::_set_cifs_buffer:: Unable to set CIFS server buffer size to 64K';
    $display_warning
        .= "\nTests Involving ACL Size > 32K (approximately 900 ACEs) will fail, unless this is set manually";

    my $setvar = sub {
        shift;
        my $api;
        my $created = 1;
        try {
            $api = $_->apiset(
                category  => 'Node',
                interface => 'CLI',
                set       => 'Systemshell'
            );
        } ## end try
        catch NACL::APISet::Exceptions::ConnectionFailedException with {
            my $err_msg = shift;
            $Log->warn(
                'uACL::smbUtils::_set_cifs_buffer:: Unable to login to Node '
                    . $_->name() . "\n"
                    . $display_warning );

            #$Log->comment( "Error Message is:\n" . $err_msg->text() );
            $created = 0;
        };

        if ($created) {
            my $response
                = $api->execute_raw_command( command =>
                    'sudo sysctl sysvar.nblade.debug.cifs.max_buffer_size_server=65535'
                );

            $response
                = $api->execute_raw_command( command =>
                    'sudo sysctl sysvar.nblade.debug.cifs.max_buffer_size_server'
                );

            unless ( $response =~ $regex_setenv ) {
                $Log->warn($display_warning);
                $Log->comment( "\nError Message:\n" . $response );
            }
        } ## end if ($created)
    };

    (undef) = map { \&$setvar($_) } @{ $self->{NODES} };
    return 1;
} ## end sub _set_cifs_buffer

sub _is_dir {

    my $self = shift;
    my %args = validate( @_, { path => { type => SCALAR, optional => 0 }, } );

    my $response = $self->{powershell}->execute_command(
        command => "Test-Path -path $args{path} -pathtype Container",
        return_response_object => 1
    );

    return ( $response->{output} =~ m/true/i ) ? 1 : 0;
} ## end sub _is_dir

sub _count_aces {

    my $self = shift;
    my %args = validate(
        @_,
        {   path     => { type => SCALAR, optional => 0 },
            usertype => {
                type     => SCALAR,
                optional => 1,
                default  => 'any',
                regex    => qr/smb|nfs|any/im,
            },
        }
    );

    my $acls = $self->getACL( version => 'smb', path => $args{path} );

    my $userpattern;

    if ( $args{usertype} =~ /any/im ) {
        $userpattern = $REGEX->{anyuser};
    } elsif ( $args{usertype} =~ /nfs/im ) {
        $userpattern = $REGEX->{nfsuser};
    } else {
        $userpattern = $REGEX->{cifsuser};
    }

    my $ace_count = 2 * scalar( grep { $_ =~ $userpattern } keys(%$acls) );

    return $ace_count;
} ## end sub _count_aces

1;
