# $Id:$ 
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
#
## @summary Library of functions to run useradmin commands
## @author sujathas@netapp.com
## @status public
## @pod here



=head1 NAME

useradmin.pm - Functions to execute useradmin commands on filer.

=head1 SYNOPSIS

Note: C<$CLIENT> B<MUST> be a UNIX machine.

use useradmin;

To set verbose variable 

use useradmin;
$useradmin::verbose = 1;

=head1 DESCRIPTION

Note: C<$CLIENT> B<MUST> be a UNIX machine.

The module can add/modify/list/delete filer objects like users, groups and roles.
This module also gets the details of the above objects.

B<Usage>

To add a user to the filer with default verbose mode.

    if ( user_add($fc, "user1", "Password" => "netapp123", 
                                "Groups" => "Adminsitrators") ) {

        $Log->comment("User added");
    } else {
        $Log->comment("User not added");
    }


To add a user to the filer with verbose set to 1.

    if (user_add($fc, "user1", "Password" => "netapp123", 
                                "Groups" => "Adminsitrators") =~ /user.*added/i ) {

        $Log->comment("User added");
    } else {
        $Log->comment("User not added");
    }


=head1 FUNCTIONS

All functions take filer connection object and the argument list for the various 
useradmin subcommands as input arguments and run the command over the given  
Connection object . If the command execution is success then it will return 1 else 
return undef.

If the user wants to handle the error and want the functions to return the 
output as it is , then they need to set the verbose variable to 1 in their file.


=cut

use strict;

package useradmin;
use misc;
use Tharn;
use Data::Dumper;
require "tharnit_lib.pl";
require Exporter;

use vars qw(@ISA @EXPORT );

@ISA = ('Exporter');

our $verbose = 0;

our @EXPORT = qw(

    verbose
    domainuser_add
    domainuser_delete
    domainuser_list

    get_group_info
    get_nonroot_users
    get_option_value
    get_sid
    get_role_info
    get_user_info
    

    group_add
    group_delete
    group_list
    group_list_with_uoption
    group_modify
    group_modify_with_fswitch

    role_add
    role_delete
    role_list
    role_modify
    role_modify_with_fswitch

    user_add 
    user_delete 
    user_list
    user_list_with_xoption
    user_modify
    user_modify_with_fswitch

    );

############## domainuser_add() ###############################################


=head2 domainuser_add()

This function will add domainuser to the filer. The function accepts
filer connection object, domainuser name and groupname as input and return 1 
for success and undef for failure.

If verbose is set to 1 then return useradmin domainuser add command`s output.

B<Input> 
Filer connection Object
domainusername [ can be networkloginname, SID ]
groupname

B<Returns> 
Returns useradmin domainuser add command`s output.

B<Usage>


To add a domainuser to the filer with default verbose mode.

    if ( domainuser_add($fc, "testuser" , "Administrators" ) ) {

        $Log->comment("User added");
    } else {
        $Log->comment("User not added");
    }


To add a domainuser to the filer with verbose set to 1.

    if (domainuser_add($fc, "SMAD\\testuser", "Adminsitrators" ) =~ /user.*added/i ) {

        $Log->comment("User added");
    } else {
        $Log->comment("User not added");
    }

This will add testuser to administrators group.


=cut

sub domainuser_add ( $$$ )
{
    my ($fc, $name, $group) = @_;

    return eval {

    my $cmd = "useradmin domainuser add '$name' -g '$group'";

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /user.*added/i ) ? return 1 : return undef;
    } else {
        return $output;
    }

    };
}

############## domainuser_delete() ############################################

=head2 domainuser_delete()

This function will delete domainuser from the filer. The function accepts
filer connection object, domainusername and groupname as input and return 1 
for success and undef for failure.

If verbose is set to 1 then returns useradmin domainuser delete command`s output.

B<Input> 
Filer connection Object
domainusername [ can be networkloginname, SID ]
groupname

B<Returns> 
Returns useradmin domainuser delete command`s output.

B<Usage>

To delete a domainuser from the filer with default verbose mode.

    if ( domainuser_delete($fc, "S-1-5-21-506817408-802123488-1230116082-131999", 
                                    "Administrators") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To delete a domainuser from the filer with verbose set to 1.

    if (domainuser_delete($fc, "SMAD\\testuser", "Adminsitrators" ) =~ /user.*deleted/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will delete the user from administrators group.


=cut

sub domainuser_delete ( $$$ )
{
    my ($fc, $name, $group) = @_;

    return eval {

    my $cmd = "useradmin domainuser delete '$name' -g '$group'";

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /user.*deleted/i ) ? return 1 : return undef;
    } else {
        return $output;
    }

    };
}

############## domainuser_list() ##############################################

=head2 domainuser_list()

This function will list domainusers. The function accepts
filer connection object and groupname as input and return the commands output 
for success and undef for failure.

If verbose is set to 1 then returns useradmin domainuser list command`s output.

B<Input> 
Filer connection Object
groupname

B<Returns> 
Returns useradmin domainuser list command`s output.

B<Usage>

To delete a domainuser from the filer with default verbose mode.

    if ( domainuser_list($fc, "Adminsitrators" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To delete a domainuser from the filer with verbose set to 1.

    if (domainuser_list($fc, "Adminsitrators" ) =~ /could not list user/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }


This will list SIDs of users in administrators group.


=cut

sub domainuser_list ( $$ )
{
    my ($fc, $group) = @_;

    return eval {

    my $cmd = "useradmin domainuser list -g '$group'";

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /could not list user/i ) ? return undef : return $output;
    } else {
        return $output;
    }

    };
}

############## get_group_info() ###############################################

=head2 get_group_info()

This function will get the groups information from the filer and return it 
in a  hash. This function takes filer connection object and groupname as input. 
This funciton will return hash of group information if group exists on filer else
return undef.

B<Input> 
Filer connection Object
groupname

B<Returns> Returns hash of groups information or undef.

B<Usage>

    %group_info = get_group_info($fc,"mygroup");

    %group_info will have the following fields
    $group_info->{"Name"}
    $group_info->{"Info"}
    $group_info->{"Allowed Capabilities"}
    $group_info->{"Rid"}
    $group_info->{"Roles"}

=cut

sub get_group_info ( $$ ) {
    my ($fc, $name) = @_;
    my %info = ();
    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin group list '$name'");

    if ( $output =~ /Could not list group/ ) {
         return undef;
    } else {
        my @lines = split(/\r\n/, $output);
        chomp(@lines);
        foreach my $line (@lines) {
            chomp($line);
            my ( $key, $value ) = split(/:\s*/,$line);
            chomp($value);
            $info{$key} = $value;
        }
        return (%info);
    }
    };
}

############## get_nonroot_users() ############################################

=head2 get_nonroot_users()

This function will return the list of all non root users in the filer.

B<Input> 
Filer connection Object

B<Returns> 
Returns an array of names of all non root users exist on the filer.

B<Usage>

    @non_root_users_list = get_nonroot_users($fc);

=cut

sub get_nonroot_users ( $ ) {
    my ($fc) = @_;

    return eval {
    my @users;

    my ( undef, undef,$output ) = say($fc, "useradmin user list");
    my @lines = split(/[\n\r]+/, $output);
    shift(@lines); 
    # throw out the header "List of all non-root users:"
    chomp(@lines);
    foreach my $line (@lines) {
       my ($type,$info) = split(/\s+/,$line);
       if ($type =~ /Name:/) {
            if ( $info !~ /root/ ) {
                push(@users,$info);
            }
       }
    }
    return(@users);
    }
}

############## get_option_value() #############################################

=head2 get_option_value()

This function will get the value for the given option.
The funciton accepts filer connection object and option key and returns
value for that key..

B<Input> 
Filer connection Object
Option Key

B<Returns> 
Returns value for the given Option Key.

B<Usage>

    $output = get_option_value($fc, "security.passwd.rules.enable" );

=cut

sub get_option_value ( $$ )
{
    my ($fc, $key ) = @_;

    return eval {
        my ( undef, undef, $out ) = say( $fc, "options $key");
        chomp( $out );
        $out =~ s/\r//;
        my ( $key, $value ) = split(/\s+/ , $out );
        return $value;
    }
}


############## get_role_info() ################################################

=head2 get_role_info()

This function will get the given roles information from the filer and return it 
in a  hash. This function takes filer connection object and role name as input. 
This funciton will return hash of role information.

B<Input> 
Filer connection Object
role name

B<Returns> returns hash of role information.

B<Usage>

    %role_info = get_role_info($fc,"myrole");

    %role_info will have the following fields
    $role_info->{"Name"}
    $role_info->{"Info"}
    $role_info->{"Allowed Capabilities"}

=cut

sub get_role_info ( $$ ) {
    my ($fc, $name) = @_;
    my %info = ();
    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin role list '$name'");

    if ( $output =~ /Could not list role/ ) {
        return undef;
    } else {
        my @lines = split(/\r\n/, $output);
        chomp(@lines);

        foreach my $line (@lines) {
            $line =~ s/\r//;
            my ( $key, $value ) = split(/:\s/,$line);
            $info{$key} = $value;
        }
        return (%info);
    }
    };
}

############## get_sid() ######################################################

=head2 get_sid()

This function will get the SID for the given domainuser by doing cifs lookup.
The funciton accepts filer connection object and domainuser as input and returns
SID of the given user.

Note: If SID is given as input then this will return the domainuser name.

B<Input> 
Filer connection Object
domainusername [ can be networkloginname, SID ]

B<Returns> 
Returns SID/Name for the given domainusername/SID.

B<Usage>

    $output = get_sid($fc, "testuser" );
    $output = get_sid($fc, "SMAD\testuser");
    $output = get_sid($fc, "S-1-5-21-506817408-802123488-1230116082-131999");

=cut

sub get_sid ( $$ )
{
    my ($fc, $name ) = @_;

    return eval {

    if ( $name =~ /^S\-*/) {
        $Log->comment("Returning $name");
        return $name;
    } else {
        my $cmd = "cifs lookup $name";

        my ( undef, undef,$output ) = say($fc, $cmd);
        if ( $output =~ /lookup failed/ ) {
            return undef;
        } else {
            chomp($output);
            $output =~ s/\r//;
            my ( undef, $sid ) = split(/ = /, $output);
            $Log->comment("Returning $sid");
            return $sid;
        }
    }
    };
}

############## get_user_info() ################################################

=head2 get_user_info()

This function will get the users information from the filer and return it as 
in a  hash. This function takes filer connection object and username as input. 
This funciton will return hash of user`s information if the user exists on the 
filer else return undef.

B<Input> 
Filer connection Object
Username

B<Returns> returns hash of users information or undef.

B<Usage>

    %user_info = get_user_info($fc,"user1");

    %user_info will have the following fields
    $user_info->{"Name"}
    $user_info->{"Info"}
    $user_info->{"Full Name"}
    $user_info->{"Allowed Capabilities"}
    $user_info->{"Minage"}
    $user_info->{"Maxage"}
    $user_info->{"Status"}

=cut

sub get_user_info ( $$ ) {
    my ($fc, $name) = @_;
    my %info = ();
    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin user list '$name'");

    if ( $output =~ /Could not list user/) {
        return undef;
    } else {
        my @lines = split(/\r\n/, $output);

        chomp(@lines);

        foreach my $line (@lines) {
            $line =~ s/\r//;
            my ( $key, $value ) = split(/:\s/,$line);

            if ($key eq "Password min/max age in days") {
                my ($min, $max) = split(/\//, $value);
                    $info{"Minage"} = $min;
                    $info{"Maxage"} = $max;
            } else {
                $info{$key} = $value;
            }
        }
        return (%info);
    }
    };
}


############## group_add() ####################################################

=head2 group_add()

This function will add a custom defined group to the filer. The function accepts
filer connection object, groupname and hash of group info as input and return 1 
for success and undef for failure.

If verbose is set to 1 then returns useradmin group add command`s output.

B<Input> 
Filer connection Object
groupname
hash of group fields

B<Returns> 
Returns useradmin group add command`s output.

B<Usage>


To add a group to the filer with default verbose mode.

    if ( group_add($fc, "mygroup" , "Info" => "my group",
                                                "Roles" => "admin" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To add a group to the filer with verbose set to 1.

    if (group_add($fc, "mygroup" , "Roles" => "admin" ) =~ /group.*added/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

=cut

sub group_add ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {

    my $cmd = "useradmin group add '$name'";
    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -r '$args{Roles}'" if (defined($args{Roles}));

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /group.*added/i ) ? return 1 : return undef;
    } else {
        return $output;
    }

    };
}

############## group_delete() #################################################

=head2 group_delete()

This function will delete the given group from the filer. The function accepts 
filer connection object and a groupname to delete as input and return 1 
for success and undef for failure.

If verbose is set to 1 then returns useradmin group delete command`s output.

B<Input> 
Filer connection Object
groupname

B<Returns> 
Returns useradmin group delete command`s output.

B<Usage>

To delete a group from the filer with default verbose mode.

    if ( group_delete($fc, "mygroup" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To delete a group from the filer with verbose set to 1.

    if (group_delete($fc, "mygroup" ) =~ /group.*deleted/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

=cut

sub group_delete ( $$ )
{
    my ($fc, $name) = @_;

    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin group delete '$name'");

    if ( $verbose == 0 ) {
        ( $output =~ /group.*deleted/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}


############## group_list() ###################################################

=head2 group_list()

This function will list a group in the filer. This function takes filer 
connection object, groupname as input. 
This funciton will return useradmin group list command`s output in case of 
success and undef for failure.

If verbose is set to 1 then returns useradmin group list command's output.

B<Input> 
Filer connection Object
groupname

B<Returns> useradmin group list command`s output.

B<Usage>

To list a group with default verbose mode.

    if ( group_list($fc, "Group" => "mygroup" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To list a group with verbose set to 1.

    if (group_list($fc, "Group" => "mygroup") =~ /could not list group/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }

=cut

sub group_list ( $% ) {

    my ($fc, %args) = @_;
    return eval {
    my $cmd = "useradmin group list";
    $cmd .= " '$args{Group}'" if (defined( $args{Group} ));

    my ( undef, undef, $output) = say ($fc, $cmd );

    if ( $verbose == 0 ) {
        ( $output =~ /could not list group/i ) ? return undef : return $output;
    } else {
        return $output;
    }
    };
}

############## group_list_with_uoption() ######################################

=head2 group_list_with_uoption()

This function will list users in a group in the filer. This function takes filer 
connection object, groupname as input. 
This function will return useradmin group list -u command`s output inca se of 
success and undef for failure.

If verbose is set to 1 then returns the commands output as such.

B<Input> 
Filer connection Object
groupname

B<Returns> useradmin group list -u command`s output.

B<Usage>

To list users in a group with default verbose mode.

    if ( group_list_with_uoption($fc, "Group" => "Administrators") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To list users in a group with verbose set to 1.

    if (group_list($fc, "Group" => "mygroup") =~ /could not list/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }

=cut

sub group_list_with_uoption ( $% ) {

    my ($fc, %args) = @_;
    return eval {
    my $cmd = "useradmin group list";
    $cmd .= " '$args{Group}'" if (defined( $args{Group} ));
    $cmd .= " '-u'";

    my ( undef, undef, $output) = say ($fc, $cmd );

    if ( $verbose == 0 ) {
        ( $output =~ /could not list/i ) ? return undef : return $output;
    } else {
        return $output;
    }
    };
}

############## group_modify() #################################################

=head2 group_modify()

This function will modify the given group from the filer. The function accepts 
filer connection object, groupname and group details to modify as input and 
return 1 for success and undef for failure.

If verbose is set to 1 then returns useradmin group modify command`s output.

B<Input> 
Filer connection Object
groupname
hash of group fields

B<Returns> 
Returns useradmin group modify command`s output.

B<Usage>


To modify a group with default verbose mode.

    if ( group_modify($fc, "mygroup" , "Info" => "Modified",
                                                "Roles" => "audit",
                                                "Group" => "myauditgroup") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To modify a group with verbose set to 1.

    if (group_modify($fc, "mygroup" , "Info" => "Modified",
                            "Roles" => "audit",
                            "Group" => "myauditgroup") =~ /group.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify group with the given values and rename the group to 
'myauditgroup'.

=cut

sub group_modify ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {
    my $cmd = "useradmin group modify '$name'" ;
    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -g '$args{Group}'" if ( defined( $args{Group} ));
    $cmd .= " -r '$args{Roles}'" if ( defined( $args{Roles} ));

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /group.*modified/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}


############## group_modify_with_fswitch() ####################################

=head2 group_modify_with_fswitch()

This function will modify the given group from the filer. The function accepts 
filer connection object, groupname and group details to modify as input and 
return 1 for success and undef for failure.

If verbose is set to 1 then returns useradmin group modify command`s output.

B<Input> 
Filer connection Object
groupname
hash of group fields

B<Returns> 
Returns useradmin group modify -f command`s output.

B<Usage>

To modify a group with default verbose mode.

    if ( group_modify_with_fswitch($fc, "mygroup" , "Info" => "Modified",
                                        "Roles" => "audit",
                                        "Group" => "myauditgroup") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To modify a group with verbose set to 1.

    if (group_modify_with_fswitch($fc, "mygroup" , "Info" => "Modified",
                                "Roles" => "audit",
                                "Group" => "myauditgroup") =~ /group.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify group with the given values and rename the group to 
'myauditgroup'.

=cut

sub group_modify_with_fswitch ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {
    my $cmd = "useradmin group modify '$name' -f" ;
    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -g '$args{Group}'" if ( defined( $args{Group} ));
    $cmd .= " -r '$args{Roles}'" if ( defined( $args{Roles} ));

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /group.*modified/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}


############## role_add() #####################################################

=head2 role_add()

This function will add a custom defined role to the filer. The function accepts
filer connection object, role name and hash of role fields as input and return 1 
for success and undef for failure.

If verbose is set to 1 then returns useradmin role add command`s output.

B<Input> 
Filer connection Object
role name
hash of role fields

B<Returns> 
Returns useradmin role add command`s output.

B<Usage>

To add a role with default verbose mode.

    if ( role_add($fc, "myrole" , "Info" => "my role",
                    "Allowed Capabilities" => "login-*,cli-*" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To add a role with verbose set to 1.

    if (role_add($fc, "myrole" , "Info" => "my role",
            "Allowed Capabilities" => "login-*,cli-*" )=~ /role.*added/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will add "myrole" with "login-*,cli-*" capabilities.
=cut

sub role_add ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {

    my $cmd = "useradmin role add '$name'";
    my $caps = defined($args{"Allowed Capabilities"}) ? $args{"Allowed Capabilities"} : "";

    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -a '$caps'";

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /role.*added/i ) ? return 1 : return undef;
    } else {
        return $output;
    }

    };
}

############## role_delete() ##################################################

=head2 role_delete()

This function will delete the given role from the filer. The function accepts 
filer connection object and a role name to delete as input and return 1 
for success and undef for failure.

If verbose is set to 1 then returns useradmin role delete command`s output.

B<Input> 
Filer connection Object
role name

B<Returns> 
Returns useradmin role delete command`s output.

B<Usage>

To delete a role with default verbose mode.

    if ( role_delete($fc, "myrole" )) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

To delete a role with verbose set to 1.

    if (role_delete($fc, "myrole" )=~ /role.*deleted/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will delete myrole from filer.

=cut


sub role_delete ( $$ )
{
    my ($fc, $name) = @_;

    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin role delete '$name'");

    if ( $verbose == 0 ) {
        ( $output =~ /role.*deleted/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}


############## role_list() ####################################################

=head2 role_list()

This function will list a role in the filer. This function takes filer 
connection object, role name as input. 
This funciton will return useradmin role list command`s output in case of 
success and undef for failure.

If verbose is set to 1 then returns the command's output as such.

B<Input> 
Filer connection Object
role name

B<Returns> useradmin role list command`s output.

B<Usage>

To delete a role with default verbose mode.

    if ( role_list($fc, "Role" => "myrole" ) ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To delete a role with verbose set to 1.

    if (role_list($fc, "Role" => "myrole" )=~ /could not list role/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }

=cut

sub role_list ( $% ) {

    my ($fc, %args) = @_;
    return eval {
    my $cmd = "useradmin role list";
    $cmd .= " '$args{Role}'" if (defined( $args{Role} ));

    my ( undef, undef, $output) = say ($fc, $cmd );

    if ( $verbose == 0 ) {
        ( $output =~ /could not list role/i ) ? return undef : return $output;
    } else {
        return $output;
    }
    };
}

############## role_modify() ##################################################

=head2 role_modify()

This function will modify the given role from the filer. The function accepts 
filer connection object, role name and role details to modify as input and 
return 1 for success and undef for failure.

If verbose is set to 1 then returns useradmin role modify command`s output.

B<Input> 
Filer connection Object
role name
hash of role fields

B<Returns> 
Returns useradmin role modify command`s output.

B<Usage>

To modify a role with default verbose mode.

    if ( role_modify($fc, "myrole" , "Info" => "Modified",
                    "Allowed Capabilities" => "*") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To modify a role with verbose set to 1.

    if (role_modify($fc, "myrole" , "Info" => "Modified",
                "Allowed Capabilities" => "*") =~ /role.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify role "myrole" with the given values 

=cut

sub role_modify ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {
    my $cmd = "useradmin role modify '$name'" ;
    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -a '$args{'Allowed Capabilities'}'" if (defined($args{'Allowed Capabilities'}));

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /role.*modified/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}

############## role_modify_with_fswitch() #####################################

=head2 role_modify_with_fswitch()

This function will modify the given role from the filer. The function accepts 
filer connection object, role name and role details to modify as input and 
return 1 for success and undef for failure.

If verbose is set to 1 then returns useradmin role modify -f command`s output.

B<Input> 
Filer connection Object
role name
hash of role fields

B<Returns> 
Returns useradmin role modify command`s output.

B<Usage>

To modify a role with default verbose mode.

    if (role_modify_with_fswitch($fc, "myrole" , "Info" => "Modified",
                        "Allowed Capabilities" => "*") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To modify a role with verbose set to 1.

    if (role_modify_with_fswitch($fc, "myrole" , "Info" => "Modified",
                "Allowed Capabilities" => "*") =~ /role.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify role "myrole" with the given values 

=cut


sub role_modify_with_fswitch ( $$% )
{
    my ($fc, $name, %args) = @_;

    return eval {
    my $cmd = "useradmin role modify '$name' -f" ;
    $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
    $cmd .= " -a '$args{'Allowed Capabilities'}'" if (defined($args{'Allowed Capabilities'}));

    my ( undef, undef,$output ) = say($fc, $cmd);

    if ( $verbose == 0 ) {
        ( $output =~ /role.*modified/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}


############## user_add() #####################################################
=head2 user_add()

This function will add a user to the filer. This function takes filer 
connection object, username and a hash of arguments required for adding a user. 
This funciton will return 1 for success and undef for failure.

If verbose is set to 1 then return command`s output.

B<Input> 
Filer connection Object
Username
Arguments Hash

B<Returns> useradmin user add command`s output.

B<Usage>

To add a user with default verbose mode.

    if ( user_add($fc, "user1", "Password" => "netapp123", 
                                "Groups" => "Adminsitrators") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To add a user with verbose set to 1.

    if ( user_add($fc, "user1", "Password" => "netapp123", 
                                "Info" => "Comment", 
                                "Full Name" => "Full name for the user",
                                "Groups" => "Adminsitrators",
                                "Maxage" = "5",
                                "Minage" => "1" ) =~ /user.*added/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


=cut

sub user_add ( $$% ) {
    my ($fc, $name, %args) = @_;

    return eval {

    if ( $fc->conntype() =~ /rsh|ssh-persay/ ) {
        my $passwd = defined($args{Password}) ? $args{Password} : "";
        my $group  = defined($args{Groups}) ? $args{Groups} : "";

        my $cmd = "useradmin user add '$name' -p '$passwd'";
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'}));
        $cmd .= " -g '$group'";
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        my ( undef, undef,$output ) = say($fc, $cmd);

        if ( $verbose == 0 ) {
            ( $output =~ /user.*added/i ) ? return 1 : return undef;
        } else {
            return $output;
        }

    } else {
        my $passwd = defined($args{Password}) ? $args{Password} : "";
        my $group  = defined($args{Groups}) ? $args{Groups} : "";

        my $cmd = "useradmin user add '$name'";
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'}));
        $cmd .= " -g '$group'";
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        say($fc, "useradmin user add '$name' -g '$group'", -blind);
        say($fc, $passwd, -blind);
        my $output = say($fc, $passwd, "User.*added");

        if ( $verbose == 0 ) {
            ( $output =~ /user.*added/i ) ? return 1 : return undef;
        } else {
            return $output;
        }
    }

    };
}

############## user_delete() ##################################################

=head2 user_delete()

This function will delete the given user from the filer. 
This function takes filer connection object and username as input. 
This funciton will return 1 for success and undef for failure.

If verbose is set to 1 thenreturn useradmin user delete command`s output.

B<Input> 
Filer connection Object
Username

B<Returns> useradmin user delete command`s output.

B<Usage>

To delete a user with default verbose mode.

    if ( user_delete($fc,"user1") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To delete a user with verbose set to 1.

    if ( user_delete($fc,"user1") =~ /user.*deleted/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will delete the user user1 from the filer.

=cut

sub user_delete ( $$ ) {
    my ($fc, $name) = @_;

    return eval {
    my ( undef, undef,$output ) = say($fc, "useradmin user delete '$name'");

    if ( $verbose == 0 ) {
        ( $output =~ /user.*deleted/i ) ? return 1 : return undef;
    } else {
        return $output;
    }
    };
}

############## user_list() ####################################################

=head2 user_list()

This function will list a user in the filer. This function takes filer 
connection object, username and groupname as input. 
This funciton will return useradmin user list command`s output in case of 
success and undef for failure.

If verbose is set to 1 , then returns the commands output as such.

B<Input> 
Filer connection Object
Username
groupname

B<Returns> useradmin user list command`s output.

B<Usage>

To list all user in administrators group with default verbose mode.

    if ( user_list($fc, "Group" => "Administrators") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To list a user with verbose set to 1.

    if ( user_list($fc, "Loginname" => "user1" ) =~ /Could not list user/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }

=cut

sub user_list ( $% ) {

    my ($fc, %args) = @_;
    return eval {
    my $cmd = "useradmin user list";
    $cmd .= " '$args{Loginname}'" if (defined( $args{Loginname} ));
    $cmd .= " -g '$args{Group}'" if (defined( $args{Group} ));

    my ( undef, undef, $output) = say ($fc, $cmd );

    if ( $verbose == 0 ) {
        ( $output =~ /could not list user/i ) ? return undef : return $output;
    } else {
        return $output;
    }
    };
}

############## user_list_with_xoption() #######################################

=head2 user_list_with_xoption()

This function will list the extended information for the users in the given 
group in the filer. This function takes filer connection object and groupname 
as input. This funciton will return useradmin user list -x command`s output for success and undef for failure.

If verbose is set , then returns the commands output as such.

B<Input> 
Filer connection Object
Username

B<Returns> useradmin user list -x command`s output.

B<Usage>

To list all user in administrators group with default verbose mode.

    if ( user_list_with_xoption($fc,"Administrators") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


To list all user in administrators group with verbose set to 1.

    if ( user_list_with_xoption($fc,"Administrators") =~ /Could not list user/i ) {

        $Log->comment("FAIL");
    } else {
        $Log->comment("PASS");
    }

    This will list the extended information of all the users in Administrators 
    group in the filer.

=cut

sub user_list_with_xoption ( $$ ) {

    my ($fc, $group) = @_;
    return eval {
    my $cmd = "useradmin user list -x -g '$group'";
    my ( undef, undef, $output) = say ($fc, $cmd );

    if ( $verbose == 0 ) {
        ( $output =~ /could not list user/i ) ? return undef : return $output;
    } else {
        return $output;
    }
    };
}

############## user_modify() ##################################################

=head2 user_modify()

This function will modify the user with the given user information.
This function takes filer connection object, username and hash of user 
information as input. This funciton will return 1 
for success and undef for failure.

If verbose is set to 1 then return useradmin user modify command`s 
output.

B<Input> 
Filer connection Object
Username
Hash of user info

B<Returns> 
Returns useradmin user modify command`s output.

B<Usage>


To modify a user with default verbose mode.

    if ( user_modify($fc, "user1")) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify the user with null values.


To modify a user with verbose set to 1.

    if ( user_modify($fc, "user1", "Info" => "New comment",
                                    "Full Name" => "New Full name",
                                    "Groups" => "Guests,Users",
                                    "Maxage" => "2",
                                    "Minage" => "1",
                                    "Password" => "newpassword") =~ /user.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


This will modify the user user1 with the given values.
Password will be modified if the filer connection object is either rsh 
or ssh-persay. Else password field will be ignored.


=cut

sub user_modify ( $$% ) {
    my ($fc, $name, %args) = @_;

    return eval {

    if ( $fc->conntype() =~ /rsh|ssh-persay/ ) {

        my $cmd = "useradmin user modify '$name'";

        $cmd .= " -p '$args{Password}'" if ( defined( $args{Password} ));
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'} ));
        $cmd .= " -g '$args{Groups}'" if ( defined( $args{Groups} ));
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        my ( undef, undef,$output ) = say($fc, $cmd);

        if ( $verbose == 0 ) {
            ( $output =~ /user.*modified/i ) ? return 1 : return undef;
        } else {
            return $output;
        }

    } else {
        my $cmd = "useradmin user modify '$name'";
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'} ));
        $cmd .= " -g '$args{Groups}'" if ( defined( $args{Groups} ));
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        my ( undef, undef,$output ) = say($fc, $cmd);

        if ( $verbose == 0 ) {
            ( $output =~ /user.*modified/i ) ? return 1 : return undef;
        } else {
            return $output;
        }
    }
    };
}

############## user_modify_with_fswitch() #####################################

=head2 user_modify_with_fswitch()

This function will modify the user using -f switch with the given user information.
This function takes filer connection object, username and hash of user 
information as input. This funciton will return 1 
for success and undef for failure.

If verbose is set to 1 then return useradmin user modify -f command`s 
output.

B<Input> 
Filer connection Object
Username
Hash of user info

B<Returns> 
Returns useradmin user modify command`s output.

B<Usage>

To modify a user with default verbose mode.

    if ( user_modify_with_fswitch($fc, "user1") ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }

This will modify the user with null values.


To modify a user with verbose set to 1.

    if ( user_modify_with_fswitch($fc, "user1", "Info" => "New comment",
                                "Full Name" => "New Full nane",
                                "Groups" => "Guests,Users",
                                "Maxage" => "2",
                                "Minage" => "1",
                                "Password" => "newpassword") =~ /user.*modified/i ) {

        $Log->comment("PASS");
    } else {
        $Log->comment("FAIL");
    }


This will modify the user user1 with the given values.
Password will be modified if the filer connection object is either rsh 
or ssh-persay. Else password field will be ignored.

=cut

sub user_modify_with_fswitch ( $$% ) {
    my ($fc, $name, %args) = @_;

    return eval {

    if ( $fc->conntype() =~ /rsh|ssh-persay/ ) {

        my $cmd = "useradmin user modify '$name' -f";

        $cmd .= " -p '$args{Password}'" if ( defined( $args{Password} ));
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'} ));
        $cmd .= " -g '$args{Groups}'" if ( defined( $args{Groups} ));
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        my ( undef, undef,$output ) = say($fc, $cmd);

        if ( $verbose == 0 ) {
            ( $output =~ /user.*modified/i ) ? return 1 : return undef;
        } else {
            return $output;
        }

    } else {
        my $cmd = "useradmin user modify '$name' -f";
        $cmd .= " -c '$args{Info}'" if ( defined( $args{Info} ));
        $cmd .= " -n '$args{'Full Name'}'" if (defined($args{'Full Name'} ));
        $cmd .= " -g '$args{Groups}'" if ( defined( $args{Groups} ));
        $cmd .= " -M '$args{Maxage}'" if ( defined( $args{Maxage} ));
        $cmd .= " -m '$args{Minage}'" if ( defined( $args{Minage} ));

        my ( undef, undef,$output ) = say($fc, $cmd);

        if ( $verbose == 0 ) {
            ( $output =~ /user.*modified/i ) ? return 1 : return undef;
        } else {
            return $output;
        }
    }
    };
}

1;

############## End of File ####################################################
