# -*-perl-*x
#
# $Id: //depot/prod/test/main/lib/vbncheck.pm
#
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for parsing `vol vbncheck` output
##
## @description
##
##     vol_vbncheck - Calls vbncheck on an array of VBNs.
##     vol_vbncheck_read_output - Parses the output of vbncheck
##
## @usage
##     use vbncheck;
##
## @author Karthik Ravichandra (karth@netapp.com) 
##
## @status private

package vbncheck;

use Carp qw(croak);
use Tharn qw($Log);
use misc qw(cli);
use Params::Validate qw (validate validate_pos SCALAR OBJECT ARRAYREF);
use SayWrap;

use vars ('@ISA','@EXPORT');
@ISA = qw(Exporter);
@EXPORT = qw(vol_vbncheck vol_vbncheck_read_output);

## @name ::vol_vbncheck($PATH)
##
## @summary
#i     Calls vbncheck on an array of VBNs
## @description
##     
##
## @arg $fc Required: filer connection object.
## @arg $vol Required: Name of the volume.
## @arg $@vbn Required: Array  of VBNs used a input for vbncheck.

##
## @return value
##      Returns errcode = 0 - Success
##                      < 0 - Failure
##              result = On Success, path of file, where output of vbncheck is stored.
##                     = On Failure, error message.
##                          
sub vol_vbncheck(%)
{
     my %opts = validate(
	@_,
	{
		conn => { type => OBJECT },
		vol => {type => SCALAR },
		vbn => {type => ARRAYREF , optional => 1},
                mode => {type => SCALAR, optional => 1},
                maps => {type => SCALAR, optional => 1},
	}
    );

    my $fc = $opts{conn};
    my $vol = $opts{vol};
    my $vbn = $opts{vbn};
    my $mode = $opts{mode};
    my $maps = $opts{maps};
	
    my $result;
    my $msg = "Starting scan on ".$vol.", logfile ";
    my $errcode = 0;
    my $err_msg = "vol vbncheck: ";
	
    if(!defined($opts{mode})){
        $mode = "-q test"; 
    }
     
    my $output;
    if(!defined($opts{maps}))
     {
          $output = sayEOF($fc,"priv set $mode; vol vbncheck $vol @$vbn");
     }
    else
     {
          $output = sayEOF($fc,"priv set $mode; vol vbncheck $vol -maps @$vbn");
     }
    my @lines = split /\n/, $output;

    foreach (@lines) {

       if($_ =~ /$err_msg/) {
           $errcode = -1;
           $result = $_;
       }
       if($_ =~ m/$msg(.*)/){
           $errcode = 0;
           $result = $1;
       }
    }

    return ($errcode, $result);
}

## @name ::vol_vbncheck_read_ouput($fc, $PATH)
##
## @summary
##     Parses output of `vbncheck`
## @description
##     Returns the parsed output of vbncheck log.

## ##
## @arg $fc Required: filer connection object.
## @arg $path Required: path of the log, returned when vol_vbncheck() is called
## @arg $maps Optional: to enable -maps option 
##
## @return value
##      Returns  
##               %result = A hash where VBNs are the key, and the pathname is a value.
##                                pathname is set to 'in active file system' if present in afs.
##                                pathname is set to 'in snapshot' if present in snapshot.
##                                pathname is set to 'not in use' if VBN is not used.                 
##               @snapshot_list = Reference to Array of snapshots that possibly contain this VBN.

sub vol_vbncheck_read_output(%)
{
    my %opts = validate(
        @_,
        {
                conn => { type => OBJECT },
                path => {type => SCALAR },
                maps => {type => SCALAR, optional => 1},
        }
    );

    my $fc = $opts{conn};
    my $path = $opts{path};
    my $maps = $opts{maps};

    my $in_afs = "in active file system";
    my $notinuse_msg = "not in use";
    my $pathname_msg = "pathname = ";
    my $in_snapshot =  "in snapshot";
    my $snapshot_msg = "The following snapshots contain any of the listed blocks:";
    my $snapshots_exist;

    my %result = ();
    my $vvbn;

    my @snapshot_list;

    my $output = sayEOF($fc, ,"rdfile ".$path);
    my @lines = split /\n/, $output;
    
    my $new_format = 0;
    foreach (@lines) {

      # Store the corrupt blocks in a list.
      if($_ =~ /VVBN\s(\d+)/) {
        $vvbn = $1;
        $result{$1} = $notinuse_msg;
        $new_format = 0;
      } elsif ($_ =~ /VVBN:\s(\d+)/) {

        # New format in IC.3
        # VVBN: 4150, Disk: 0a.18, Volume: testvol_flex, Inode: 100
        # Inode 100 has 1 name(s).
        # Volume: testvol_flex, Inode: 100, Pathname: /vol/testvol_flex/file1

        $vvbn = $1;
        $result{$1} = $notinuse_msg;
        $new_format = 1;
      } 
     
         # New format in IC.3 'vbncheck with -maps option'
         # Whenever the vvbn is in Active file system the output will be
         # VVBN 4150  in active file system
      if($_ =~ /VVBN\s(\d+)\s+$in_afs/)
      {
        
        $vvbn = $1;
        $result{$1} = $in_afs;
      }
      
      if($_ =~ /VVBN\s(\d+)\s+$in_snapshot/)
      {
        $vvbn = $1;
        $result{$1} = $in_snapshot;
      }

      if($new_format) {
        if($_ =~ /Pathname:\s(.*)/) {
           $result{$vvbn} = $1;
        }
      } else {
        if($_ =~ /$pathname_msg(.*)/) {
           $result{$vvbn} = $1;
        }
      }

      # Determine if there are any corrupt blocks in the snapshots.
      if($_ =~ /$snapshot_msg/) {
        $snapshots_exist = 1;
      }

      # If there are snapshots, then add them to a list.
      if(($_ =~ /\d+:\d+:\d+/) && ($snapshots_exist == 1)) {
        my @line = split / /, $_;
        push @snapshot_list, @line[0];
      }

    }
    
    if(!defined($opts{maps}))
    {
          return (%result, \@snapshot_list);
    }
    else
    {
          return (\%result, \@snapshot_list);
    }
    
}

1;



