#-*-perl-*-
#$Id: //depot/prod/test/Rsierranevada/lib/vol.pm#1 $

## @summary Routines for issuing volume commands to the Filer
##
## @description
##      The vol.pm functions are wrappers around NATE's say() function that
##      send a 'vol xxx' command to the filer and return 1 for success or
##      undef for failure.  In the case of failure, an error string
##      is also returned in $EVAL_ERROR ($@)
##
## @synopsis
##      use vol;
##      $success = vol_create( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_destroy( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_add( &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_rename( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_mirror( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_split( &nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_online( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_offline( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vol_restrict( $fullconnspec, $vol_name [, <i>optional_params</i>] );
##
## @usage
##      <tt>$success = <i>volfunction</i>( $fullconnspec, $vol_name [, options => ""] [, timeout => -1]);</tt>
##
##      where <i><tt>volfunction</tt></i></b> is one of the following:
##      <pre><ul>vol_create
##      vol_destroy
##      vol_add
##      vol_rename
##      vol_mirror
##      vol_split
##      vol_online
##      vol_offline
##	vol_options
##      vol_restrict
##	vol_status_r
##	vol_status_s
##      </ul></pre>
##
##      <tt>$fullconnspec</tt> identifies the host and/or connection to be
##      used to submit the appropriate 'vol xxx' command.  It can be a host
##      object, connection object, hostspec, connspec or hostspec#connspec. 
##      If a connection is not specified, #rsh is used.
##
##      <tt>$vol_name</tt> is the name of the volume on which <i>volfunction</i>
##      will be performed.
##
##      The <tt><i>optional_params</i></tt> argument is a string containing
##      any desired options for the particular vol command.  It is null by
##      default.  If supplied, it must conform to that vol command's syntax.
##
##      The <tt>timeout</tt> argument specifies the maximum amount of time,
##      in seconds, for the function to return.  If that maximum is
##      exceeded, a timeout error is returned.  The default timeout is -1
##      (infinite).
##
## @examples
##      To create a volume with two disks:
##      <blockquote>
##      use vol;
##      my $h = host($FILER);
##      vol_create($h, "testvol", options=>2, timeout=>300)
##      &nbsp;&nbsp;&nbsp; or tstlog('-script', $@);
##      # vol_create succeeded: continue with test
##      </blockquote>
## 
## @author dl-johny@netapp.com (FSS WAFL Quality)
## @status deprecated - use Volume.pm instead

package vol;

use strict;
use vars ('@ISA','@EXPORT');
use Tharn;
require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(vol_create vol_destroy vol_rename vol_mirror vol_join vol_split vol_online vol_offline
	   vol_options vol_scrub vol_restrict vol_add vol_status vol_status_s vol_status_r vol_get_disks_by_volume vol_get_disks_by_type vol_get_disks_by_group wait_for_vol_status vol_get_rg_stat wait_for_vol_zeroing vol_is_there);


# Creates a volume with the disks list specified in options
sub vol_create ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'create',$args{timeout},$args{options});
	if ($after =~ /has\s+been\s+initiated/ || $after =~ /has\s+completed/) {
		#return wait_for_vol_status($spec,$vol,$args{timeout},"online","normal|mirrored|raid4");
		sleep(5);
		return wait_for_vol_zeroing($spec,$vol,$args{timeout});
	} else {
	    die $after;
	}
    };
}

# Renames a volume to the new volume specified in options.
sub vol_rename ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'rename',$args{timeout},$args{options});
	if ($after =~ /$vol\s+renamed\s+to\s+$args{options}/){
	    my %status = vol_status($spec,$args{options},timeout=>$args{timeout});
	    return 1 if ($status{State} =~ /online/);
	}else{
		die $after;
	}
    };
}

# Split a mirrores volume to a new volume specified in options
sub vol_split ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'split',$args{timeout},$args{options});
	if ($after =~ /new\s+volume\s+$args{options}\s+created/g) {
		return 1;
	}else{
		die $after;
	}
    };
}

# Destroys the volume.  Caller must include '-f' in options in order
# to make filer not issue an "are you sure?" type of prompt, which
# this routine cannot handle.
sub vol_destroy ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	vol_offline($spec,$vol,%args);
	my $after = say_cmd($spec,$vol,'destroy',$args{timeout},$args{options});
	if ($after =~ /destroyed/){
	    return 1;
	}else{
	    die $after;
	}
    };
}

# Takes the volume online.
sub vol_online ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'online',$args{timeout},$args{options});
	my %status = vol_status($spec,$vol,timeout=>$args{timeout});
	if ($status{State} =~ /online/){
	    return 1; 
	}else{
	    die "$after";
	}
    };
}

# Takes the volume offline
sub vol_offline ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'offline',$args{timeout});
	my %status = vol_status($spec,$vol,timeout=>$args{timeout});
	if ($status{State} =~ /offline/ || $status{State} =~ /failed/){
	    return 1; 
	}else{
	    die $after;
	}
    };
}

# Restrict the volume 
sub vol_restrict ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'restrict',$args{timeout});
	my %status = vol_status($spec,$vol,timeout=>$args{timeout});
	if ($status{State} =~ /restricted/){
		return 1;
	}else{
		die $after;
	}
    };
}

# Mirrors a volume with disk list as specified in options.
sub vol_mirror ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'mirror',$args{timeout},$args{options});
	if ($after =~ /has\s+completed/ || $after =~ /has\s+been\s+initiated/) {
	    return wait_for_vol_status($spec,$vol,$args{timeout},"online","mirrored");
	} else {
	    die $after;
        }
    };
}

# Joins the plexes of a split mirrored volume 
# Specify the timeout argument when using this routine, to break out of the while loop, if vol join fails.
sub vol_join ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'mirror',$args{timeout},$args{options});
	if ($after =~ /Invalid\s+operation/ || $after =~ /join/) {
	    tstlog ('-comment', "Cannot join the plexes");
	    return $after;
	} else {
	    return wait_for_vol_status($spec,$vol,$args{timeout},"online","mirrored");
        }
    };
}


# Start,suspend,resume or stop scrubbing, as specified in option.
sub vol_scrub ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	if ($args{options} eq 'start') {
	my $after = say_cmd($spec,$vol,'scrub',$args{timeout},$args{options},1);
	    if ($after =~ /vol\s+scrub:\s+started\s+scrub\s+of\s+\/$vol*/g) {
		return 1;
	    }else{
		die  $after;
	    }
	}elsif($args{options} eq 'stop') {
	my $after = say_cmd($spec,$vol,'scrub',$args{timeout},$args{options},1);
	    if ($after =~ /vol\s+scrub:\s+stopped\s+scrub\s+of\s+\/$vol*/g) {
		return 1;
	    }else{
		die "$after";
	    }
	}elsif($args{options} eq 'suspend') {
	my $after = say_cmd ($spec,$vol,'scrub',$args{timeout},$args{options},1);
	    if ($after =~ /vol\s+scrub:\s+suspended\s+scrub\s+of\s+\/$vol*/g) {
		return 1;
	    }else {
		die $after;
	    }
	}elsif($args{options} eq 'resume') {
	my $after = say_cmd($spec,$vol,'scrub',$args{timeout},$args{options},1);
	    if ($after =~ /vol\s+scrub:\s+resumed\s+scrub\s+of\s+\/$vol*/g) {
		return 1;
	    }else{
		die "$after";
	    }
	}
    };
}

# Adds disks as specified in options to the volume
sub vol_add ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
	my $after = say_cmd($spec,$vol,'add',$args{timeout},$args{options});
	if ($after =~ /has\s+been\s+initiated/ || $after =~ /has\s+completed/) {
		#return wait_for_vol_status($spec,$vol,$args{timeout},"online","normal|mirrored|raid4");
	    sleep(5);
	    return wait_for_vol_zeroing($spec,$vol,$args{timeout});
	} else {
	    die $after;
	}
    };
}

# Returns volume name,state and status in a hash.
sub vol_status ($$%) {
    my ($spec,$vol,%args) = @_;
    return eval {
        my $after = say_cmd($spec,$vol,'status',$args{timeout},$args{options});
	if ($after !~ /vol\s+status:/) {
	    my %status;

	    my @lines = split(/\n/, $after);
	    my $begin = 0;
	    foreach my $line (@lines) {
		    $begin = 1 if ($line =~ /Volume\s+State\s+Status\s+Options/);

		    if ($line =~ /($vol?.+\S)/ && $begin) {
			    ($status{Vol},$status{State},$status{Status}) = split(/\s+/,$1);
			    return %status;
		    }
	    }
	}
	else{
	    return $after;
	}
    }
}

# loop, calling 'vol status $vol' until the state and status match
# $state and $status.
sub wait_for_vol_status {
    my ($spec, $vol, $timeout, $state, $status) = @_;
    my $endtime = timeout($timeout);
    while (time() < $endtime) {
	my %status = vol_status($spec,$vol,timeout=>$endtime);
	logcomment("state = $status{State}\tstatus = $status{Status}");
	if ($status{State} =~ /$state/ && $status{Status} =~ /$status/){
	    return 1;
	} else {
	    sleep 10;
	}
    }
    die "Timed out waiting for volume '$vol' to become ($state,$status)\n";
}


# return spare disks information
sub vol_status_s {
	my ($spec) = @_;

	my $conn = make_connection($spec);
	my (undef, undef, $after) = 
		say($conn, "priv set -q advanced; vol status -s", 
			#-prompt, "eNdOfOuTpUt",
			-swallow,
			-timeout, 60);

	
	my %pools = ();
	my $cur_pool = 0;
	my @lines = split(/\n/, $after);
	for (my $i = 0; $i <= $#lines; $i++) {
		my $line = $lines[$i];

		if ($line =~ /Pool(\d+) spare disks/) {
			$cur_pool = $1;
			$pools{$cur_pool} = {};
		}
		#elsif ($line =~ /^spare\s+(\d+\w*\.\d+)\s+(\d+\w*)\s+(\d+)\s+(\d+)\s+(.*)\s+(\d+)\/(\d+)\s+(\d+)\/(\d+)/) {
		elsif ($line =~ /^spare\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\/(\S+)\s+(\S+)\/(\S+)/) {
			my $name = $1;
			my $ha = $2;
			my $shelf = $3;
			my $bay = $4;
			my $chan = $5;
			my $used_mb = int($6);
			my $used_blks = int($7);
			my $phys_mb = int($8);
			my $phys_blks = int($9);
			my $state = "ZEROED";
			$state = "ZEROING" if ($line =~ /zeroing/i);
			$state = "NOTZEROED" if ($line =~ /not zeroed/i);

			${$pools{$cur_pool}}{$name} = {
				"pool"=>$cur_pool, 
				"ha"=>$ha, 
				"shelf"=>$shelf, 
				"bay"=>$bay, 
				"chan"=>$chan, 
				"used_mb"=>$used_mb, 
				"used_blks"=>$used_blks,
				"phys_mb"=>$phys_mb,
				"phys_blks"=>$phys_blks,
				"state"=>$state,
			};
		}
	}

	return %pools;
}


# return raid group info of volume
sub vol_status_r {
	my ($spec, $vol) = @_;

	my $conn = make_connection($spec);
	my (undef, undef, $after) =
		say($conn, "priv set -q advanced; vol status -r $vol",
			-swallow,
			-timeout, 60);

	my %info;
	my $plex = "";
	my $rg = "";
	my ($vol_n, $plex_n, $rg_n) = (0, 0, 0);

	my @lines = split(/\n/, $after);
	for (my $i = 0; $i <= $#lines; $i++) {
		my $line = $lines[$i];
		if ($line =~ /Volume $vol \((.*)\) \((.*)\)/) {
			$info{'status'} = $1;
			$info{'cksum_type'} = $2;
			$info{'n_disks'} = 0;
			$info{'n_plex'} = 0;
			$info{'n_rg'} = 0;
			$info{'plex_names'} = "";
		}
		elsif ($line =~ /Plex (.*) \((.*)\)/) {
			$plex = $1;

			$info{$plex} = {
				'status' => $2, 
				'n_disks' => 0, 
				'n_rg' => 0, 
				'rg_names' => "",
			};
			$rg = "";
			$info{'n_plex'}++;
			$info{'plex_names'} .= "$1 ";
		}
		elsif ($line =~ /RAID group (.*) \((.*)\)/) {
			$rg = $1; 

			${$info{$plex}}{$rg} = {
				'status'=>$2, 
				'n_disks'=>0,
				'disk_names'=>"",
			};
			$info{'n_rg'}++;
			${$info{$plex}}{'n_rg'}++;
			${$info{$plex}}{'rg_names'} .= "$1 ";
		}
		elsif ($line =~ /(dparity|parity|data)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\d+)\/(\d+)\s+(\d+)\/(\d+)/) {	
			${${$info{$plex}}{$rg}}{$2} = {
				'type' => $1,
				'name' => $2,
				'ha'   => $3,
				'shelf' => $4,
				'bay'  => $5,
				'chan' => $6,
				'used_mb' => $7,
				'used_blks' => $8,
				'phys_mb' => $9,
				'phys_mb' => $10,
			};

			# record number of disks in vol, plex and rg
			$info{'n_disks'}++;
			${$info{$plex}}{'n_disks'}++;
			${${$info{$plex}}{$rg}}{'n_disks'}++;
			${${$info{$plex}}{$rg}}{'disk_names'} .= "$2 ";
		}
		elsif ($line =~ /(pending)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\d+)\/(\d+)\s+(\d+)\/(\d+)/) {
			$info{'pending'} = {} if (!defined($info{'pending'}));
			my $diskname = $2;

			${$info{'pending'}}{$diskname} = {
				'type' => $1,
				'name' => $2,
				'ha'   => $3,
				'shelf' => $4,
				'bay'  => $5,
				'chan' => $6,
				'used_mb' => $7,
				'used_blks' => $8,
				'phys_mb' => $9,
				'phys_mb' => $10,
			};

			if ($line =~ /(\d+)\%/) {
				${${$info{'pending'}}{$diskname}}{'percent'} = $1;
			}
		}
	}

	return %info;
}


# wait for all disks zeroing to complete in the specified volume
sub wait_for_vol_zeroing($$$) {
	my ($spec, $vol, $timeout) = @_;

	$timeout = 999999 if (!defined($timeout));
	my $endtime = timeout($timeout);

	my %info = vol_status_r($spec, $vol);
	return 1 if ($info{'status'} !~ /initializing|growing/i); # must be either state 
	return 1 if (!defined($info{'pending'})); # just to be safe

	my $min = find_min_percent($info{'pending'});
	my $sleeptime = 6 * (100 - $min); # assume 1% finished per 6 seconds
	my $done = 0;
	while (time() < $endtime) {
		logcomment("Wait for $sleeptime seconds.");
		sleep($sleeptime);

		%info = vol_status_r($spec, $vol);

		return 1 if ($info{'status'} !~ /initializing|growing/i || (!defined($info{'pending'}))); 
		
		my $min2 = find_min_percent($info{'pending'});

		$sleeptime = int(1.2 * (100 - $min2) / (($min2 - $min)/$sleeptime)); # add 20% more time to wait for
		$min = $min2;
	}

	return $done;

	sub find_min_percent {
		my ($p_pending) = @_;
		my $min = 100;
		
		foreach my $name (keys %$p_pending) {
			my $disk = $$p_pending{$name};
			next if (!defined($$disk{'percent'}));
			$min = $$disk{'percent'} if ($$disk{'percent'} < $min);
		}

		return $min;
	}
}


# set or display volume options
sub vol_options {
	my ($spec, $vol, $argstr) = @_;

	#my @args = split(/\s+/, $argstr);
	#my $conn = make_connection($spec);

	if (defined $argstr) {
		my @args = split(/\s+/, $argstr);
		# ie. setting an option
		# simply execute the command
		return say_cmd($spec, $vol, "options", options=>$argstr);
	}
	else {
		# need to return hash of option values
		my %options = ();
		
		my $after = say_cmd($spec, $vol, "options");
		my @lines = split(/\n/, $after);
		
		# ignore the first
		shift @lines;

		#logcomment("\@lines is @lines");
		foreach my $line (@lines) {
			my @opts = split(/,\s*/, $line);
			
			foreach my $opt (@opts) {
				if ($opt =~ /(.*)=(.*)/) {
					$options{$1} = $2;
				}
				else {
					$options{$opt} = 1;
				}
			}
		}

		return %options;
	}
}


#Create A good connection to use.
sub make_connection($) {
	my ($spec) = @_;	
    return eval {
	if (ref($spec) =~ /Connectrec/) {
		return $spec;
	}elsif(ref($spec) =~ /Hostrec/) {
		return connect($spec->{id} . "#rsh");
	}else{
	        my $hspec=host($spec);
		return connect($hspec->{id} . "#rsh");
	}
    };
}

# Create an absolute time from the timeout option.Timeout2time is defined in
# tharn.

sub timeout ($) {
	my ($timeout) = @_;
	$timeout = -1 if !defined($timeout);
	return timeout2time($timeout);
}



#Create a Say command to issue with every routine
sub say_cmd ($$$$;$$) {
    my ($spec,$vol,$call,$timeout,$options,$flag) = @_;
    $options = $options ? $options : "";
    $flag = $flag ? $flag : "";
    #Create a connection to use;
    my $conn = make_connection($spec);
    #Create absolute timeout;
    my $timetouse = timeout($timeout);
    my ($after,@complaint);
    if ($flag) {
        (undef,undef,$after) = say ($conn,"priv set -q advanced;vol $call $options $vol",'-timeout',$timetouse,'-eval',sub { @complaint = @_;1; });
    }else{
        (undef,undef,$after) = say ($conn,"priv set -q advanced;vol $call $vol $options",'-timeout',$timetouse,'-eval',sub { @complaint = @_;1; });
    }
    if (@complaint){
	die join ('\n',@complaint);
    }else{
    	return $after;
    }
}




# utility functions to better use the vol_ functions to extract information

# returns string of disk names in a raid group
sub vol_get_disks_by_group($$) {
	my ($p_info, $grpname) = @_;

	my ($vol, $plex);
	# check raid name format
	if ($grpname =~ /\/(\S+)\/(\S+)\/\S+/) {
		($vol, $plex) = ($1, "/$1/$2");
	}
	else {
		return undef;
	}

	my %info = %$p_info;
	if (!defined(${$info{$plex}}{$grpname})) {
		logcomment "info plex has " . keys %{$info{$plex}};
	}

	return ${${$info{$plex}}{$grpname}}{'disk_names'};
}


# returns string of disk names in a volume
sub vol_get_disks_by_volume($) {
	my ($p_info) = @_;

	my %info = %$p_info;
	#my @disks = ();
	my $disks = "";
	foreach my $plex_name (split(/\s+/, $info{'plex_names'})) {
		my $plex = $info{$plex_name};
		foreach my $rg_name (split(/\s+/, $$plex{'rg_names'})) {
			my $rg = $$plex{$rg_name};
			logcomment("Adding disks in $rg: $$rg{'disk_names'}");
			#push @disks, split(/\s+/, $$rg{'disk_names'});
			$disks .= "$$rg{'disk_names'} ";
		}
	}
	#return join(" ", @disks);
	return $disks;
}


# $type can be "data", "parity", "dparity"
# return string of disk name(s) in a volume
sub vol_get_disks_by_type($$$) {
	my ($p_info, $vol, $type) = @_;

	my %info = %$p_info;
	#my @disks = ();
	my $disks = "";
	foreach my $plex_name (split(/\s+/, $info{'plex_names'})) {
		my $plex = $info{$plex_name};
		foreach my $rg_name (split(/\s+/, $$plex{'rg_names'})) {
			my $rg = $$plex{$rg_name};
			foreach my $disk_name (split(/\s+/, $$rg{'disk_names'})) {
				my $disk = $$rg{$disk_name};
				if ($$disk{'type'} =~ /$type/i) {
					#push @disks, $disk_name;
					$disks .= "$disk_name ";
					logcomment("$disk_name is of type $type");
				}
			}
		}
	}
	#return join(" ", @disks);
	return $disks;
}



# return the named raid group hash in a status_r hash
sub vol_get_rg_stat($$) {
	my ($p_info, $rg) = @_;

	my $volname;
	my $plex;
	if ($rg =~ /\/(\S+)\/(\S+)\/\S+/) {
		$volname = $1;
		$plex = "/$1/$2";
	}
	else {
		return undef;
	}
	my %info = %$p_info;

	return undef if (!defined($info{$plex}));
	#return undef if (!defined(${$info{$plex}}{$rg}));

	return ${$info{$plex}}{$rg};
}


# return true or false for volume existence
sub vol_is_there ($$) {
	my ($spec, $vol) = @_;

	my $after = say_cmd($spec, $vol, "status", 60);
	return ($after !~ /No volume named.*exists/i);
}
