
# $Id: //depot/prod/test/Rsierranevada/lib/vol_db.pm#1 $

package vol_db;

use 5.008008;
use strict;
use warnings;

require Exporter;

use Params::Validate qw(validate SCALAR HASHREF ARRAYREF);
use Carp qw(croak);

use Tharn;
use Server;
use SayWrap qw(sayEOF);

BEGIN {
    Params::Validate::validation_options(
        # allow presence of options other than those named in the validation specification
        allow_extra => 1,
   );
}

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use vol_db ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = (
    'all' => [ qw() ]
);

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw();

our $VERSION = '0.01';

sub new {
    my $class = shift;
    $class = ref($class) || $class;

    my (%opts) = validate (
	@_,
	{
	    host => {
		type => SCALAR | HASHREF,
	    },
	}
    );
    my $self;
    #Get all options
    foreach my $ele (keys(%opts)) {
       $self->{$ele} = $opts{$ele};
    }

    # In case the filer is in 7 mode dont bother with Server
    $self->{host_obj} = host($self->{host}) or return "$@";
    $self->{host_type} = $self->{host_obj}->hosttype();
    if($self->{host_type} =~/filer-cfe/) {
        $self->{host_conn} = connect('#rsh') or return "$@";
    }
    bless($self, $class);
    return $self;
}

# run the dump command and return a vol_db::dump reference
sub dump {
    return shift->_diag_say("vol_db dump");
}

# run the iter command and return a list of vol_db handle strings
sub iter {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    cmd_args => {
		type => SCALAR,
		default => '',
	    },
	}
    );

    my $cmd = "vol_db iter $args{cmd_args}";
    my $response = $self->_diag_say($cmd);

    my @matches;
    foreach my $line (split(/\n/, $response)) {
        my $match = $self->_extract_handle($line);
	if ($match) {
	    push(@matches, $match);
	} else {
	    die "unrecognized vol_db iter output: '$line'";
	}
    }

    return @matches
}

# run the lookup command and return a vol_db::handle reference
sub lookup {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    cmd_args => {
		type => SCALAR,
	    },
	}
    );

    # first, get the list of valid handles so we can use it to
    # recognize the lookup response
    my @valid_handles = $self->iter();

    # now, do the lookup
    my $cmd = "vol_db lookup $args{cmd_args}";
    my $response = $self->_diag_say($cmd);

    my ($handle, $error);

    # check the result
    if ($response eq 'null, err = not found') {
	$Log->trace("vol_db lookup found no entries matching $args{cmd_args}.");
    } elsif ($response eq 'null, err = duplicate') {
	$Log->trace("vol_db lookup found more than one entry matching $args{cmd_args}.");
	$error = $response;
    } else {
	# see if we have a valid handle...
	$handle = $self->_extract_handle($response);
	if (!defined($handle) || (!grep $handle eq $_, @valid_handles)) {
	    $error = $response;
	}
    }

    # no match, just return whatever other response was received
    return ($handle, $error);
}

# run the get command. return the requested value or throw an error.
sub get {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    attr => {
		type => SCALAR | ARRAYREF,
	    },
	    handle => {
		type => SCALAR,
		optional => 1,
	    },
	    flags => {
		type => SCALAR,
		optional => 1,
		default => '',
	    },
	}
    );
    my $attr_list = (ref($args{attr}) ne 'ARRAY') ? [ $args{attr} ] : $args{attr};
    my $handle    = $args{handle};
    my $name      = $args{name};
    my $flags     = $args{flags};

    if (!defined($handle)) {
	if (!defined($name)) {
	    die "missing parameter: need one of 'handle' or 'name'" if (!defined($handle) && !defined($name));
	} else {
	    # do a vol_db lookup to get the handle
	    my $error;
	    my $cmd_args = "$flags name '$name'";
	    ($handle, $error) = $self->lookup(cmd_args => $cmd_args);

	    if (defined($handle)) {
		# double-check that we've got the right volume/aggr
		my ($fetched_name) = $self->get(
		    attr => 'name',
		    handle => $handle
		);

		if ($fetched_name ne $name) {
		    die "internal problem detected with vol_db lookup, failed to match the right volume/aggr! ('$fetched_name' != '$name')";
		}
	    } else {
		die "handle not specified, and vol_db lookup failed to find a volume of the given name! ($name)";
	    }
	}
    }

    my @values;
    foreach my $attr (@$attr_list) {
	my $cmd = "vol_db get $attr $handle";
	my $response = $self->_diag_say($cmd);

	# try to extract the target value from the response
	if ($response =~ /^\s*${attr}:\s+'?([^']*)'?\s*$/) {
	    # capture the value
	    push(@values, $1);
	} else {
	    # must be an error
	    die $response;
	}
    }

    return @values;
}

sub snap {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    handle => {
		type => SCALAR,
	    },
	}
    );
    my $handle = $args{handle};

# WIP: here, check that the modeness is correct...snap translations only exist in 10-mode (clustered) OnTap for BR.0

    my $cmd = "vol_db snap $handle";
    my $response = $self->_diag_say($cmd);

    my @lines = split(/\n/, $response);

    # the first line in the output should contain the target handle,
    # use this as a means to check whether the response contains valid
    # data (as opposed to some error).
    my $h = $self->_extract_handle($lines[0]);
    if (!defined($h)) {
	# must be an error
	die $response;
    } elsif ($h ne $handle) {
	# failed sanity check
	die "ERROR: handle mismatch in vol_db snap output. expected: '$handle', actual: '$h' ->\n$response";
    }

    # now, discard the handle
    pop @lines;

    # parse the snap translations
    my @translations;
    foreach my $line (@lines) {
        if ($line =~ /^\s*(\d+):\s+lsnapid = (\d+)\s*$/) {
	    push(@translations, [ $1, $2 ]);
	}
    }
    return @translations;
}

#function to acquire admin lock
sub alock_acquire {

    my $self = shift;

    my (%args) = validate(
        @_,
        {
            vdb_handle => {
                type => SCALAR,
            },
            admin_lock_op => {
                type => SCALAR,
            },
            wait => {
                type => SCALAR,
                optional => 1,
                default => 'no-wait',
            },
            force => {
                type => SCALAR,
                optional => 1,
                default => '',
            },

        }
    );
    my $vdb_handle=$args{vdb_handle};
    my $admin_lock_op=$args{admin_lock_op};
    my $wait=$args{wait};
    my $force=$args{force};

    my $cmd="vol_db alock acquire $vdb_handle $admin_lock_op $wait $force";
    my $response=$self->_test_say($cmd);
    return $response;
}

#function to release admin lock
sub alock_release {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            vdb_handle => {
                type => SCALAR,
            },
            admin_lock_op => {
                type => SCALAR,
            },

        }
    );
    my $vdb_handle=$args{vdb_handle};
    my $admin_lock_op=$args{admin_lock_op};

    my $cmd="vol_db alock release $vdb_handle $admin_lock_op ";
    my $response=$self->_diag_say($cmd);
    return $response;
}

#function to rename admin lock
sub alock_rename {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            vdb_handle => {
                type => SCALAR,
            },
            curr_lock_handle => {
                type => SCALAR,
            },
            new_alock_op => {
                type => SCALAR,
            },

        }
    );
    my $vdb_handle=$args{vdb_handle};
    my $curr_lock_handle=$args{curr_lock_handle};
    my $new_alock_op=$args{new_alock_op};

    my $cmd="vol_db alock rename $vdb_handle $curr_lock_handle $new_alock_op ";
    my $response=$self->_test_say($cmd);
    return $response;
}

#function to dump admin lock data
sub alock_dump {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            locked => {
                type => SCALAR,
                optional => 1,
                default => '',
            },

        }
    );
    my $locked=$args{locked};

    my $cmd="vol_db alock dump $locked";
    my $response=$self->_diag_say($cmd);
    return $response;
}

#function to set/display alock timeout 
sub alock_timeout {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            seconds => {
                type => SCALAR,
                optional => 1,
                default => '',
            },

        }
    );
    my $seconds=$args{seconds};

    my $cmd="vol_db alock timeout $seconds";
    my $response=$self->_test_say($cmd);
    return $response;
}


#function to get_curr_lockers 
sub alock_get_curr_lockers {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            vdb_handle => {
                type => SCALAR,

            },
        }
    );
    my $vdb_handle=$args{vdb_handle};

    my $cmd="vol_db alock get_curr_lockers $vdb_handle";
    my $response=$self->_diag_say($cmd);
    return $response;
}

#function to print stats 
sub alock_print {
    my $self = shift;

    my (%args) = validate(
        @_,
        {
            vdb_handle => {
                type => SCALAR,

            },
         }
    );
    my $vdb_handle=$args{vdb_handle};

    my $cmd="vol_db alock print $vdb_handle";
    my $response=$self->_diag_say($cmd);
    return $response;
}



sub _test_say {
    #return shift->_say(level => 'test', cmd => @_);
    return shift->_say(level => 'test', cmd => "priv set -q test; @_");
}

sub _diag_say {
    #return shift->_say(level => 'diag', cmd => @_);
    return shift->_say(level => 'test', cmd => "priv set -q diag; @_");
}

sub _say {
    my $self = shift;

    my (%args) = validate(
	@_,
	{
	    level => {
		type => SCALAR,
	    },
	    cmd => {
		type => SCALAR,
	    },
	}
    );
    # In case the filer is in 7 mode dont bother with Server
    my $level = $args{level};
    my $cmd   = $args{cmd};
    if($self->{host_type} =~/filer-cfe/) {
        return sayEOF($self->{host_conn}, "priv set -q $level;$cmd");
    }

    # make sure we have a Server
    if ((!exists $self->{_server_obj}) || (!defined $self->{_server_obj})) {
	$self->{_server_obj} = Server->new(nodes => [ $self->{host} ]);
    }

#$Log->trace("level is '$level'\ncmd is '$cmd'");

    $self->{_server_obj}->dblade_set_privilege(
	level => $level,
	quiet => 'on',
    );

    return $self->{_server_obj}->dblade(cmd => $cmd);
}

sub _extract_handle {
    my $self = shift;

    my $handle;

    if ($_[0] =~ /([aft]:\d+:\d+)\[/) {
        $handle = $1;
    }

    return $handle;
}

1;

__END__

## @summary     vol_db - NATE library for testing vol_db feature of Data OnTap
##
## @author      mcrae@netapp.com
##
## @pod here

=head1 NAME

  vol_db - NATE library for testing vol_db feature of Data OnTap

=head1 SYNOPSIS

  use vol_db;

  # create a new vol_db object here
  my $vdb = vol_db->new(host => $fh);

  # get the vol_db handle for $FILER_VOL using lookup by name
  my ($vdb_handle, $error) = $vdb->lookup(cmd_args => "name '$FILER_VOL'");

  # retrieve the name of the volume indicated by the given vol_db handle
  my $name = $vdb->get(attr => 'name', handle => $vdb_handle);

=head1 STATUS

  Private, in development.

=head1 DESCRIPTION

  This module provides various utility methods for testing the vol_db facility in Data OnTap.

=head2 EXPORT

  None by default.

=head1 METHODS

=over

=item new()

Create a new instance of this class, tied to the given host.

=over

=item host

[Required] A NATE 'hostspec' for the target storage controller (filer). See http://web.netapp.com/engineering/nate/releases/latest/html/natedata.html#hostspec.

=back

=item dump()

Execute the vol_db 'dump' command on the target controller. This method takes no parameters. The contents of the volume database are returned as a string.

=item iter()

Execute the vol_db 'iter' command with the given arguments. Returns a list of vol_db handles corresponding to the matching volumes.

=over

=item cmd_args

[Optional] this string is appended to the iter command to restrict the scope of the iteration. See the vol_db documentation for valid syntax.

=back

=item lookup()

Execute the vol_db 'lookup' command with the given arguments.

This method returns a list of two values: $handle and $error. If the lookup succeeds, the matching volume's vol_db handle is returned in $handle and $error will be undefined. Otherwise, $handle will be undefined and $error will contain the error string.

=over

=item cmd_args

[Required] this string is appended to the lookup command to define the lookup criteria. See the vol_db documentation for valid syntax.

=back

=item get()

Execute the vol_db 'get' command with the given arguments. If the fetches succeed, the specified attribute value(s) are returned in the order specified. Otherwise, an error is thrown (use the PERL eval function to trap such errors).

=over

=item attr

[Required] the name of the desired attribute value(s), can be either a single attribute name or a list of such.

=item handle

[Required] vol_db handle for the target volume.

=back

=item snap()

Execute the vol_db 'snap' command with the given handle. A list of lists is returned, where each sub-list represents a single snapshot translation. Each translation is a list of two numbers indicating the logical and physical snap id's, respectively.

=over

=item handle

[Required] vol_db handle for the target volume.

=back

=item alock_acquire()

Execute the vol_db alock acquire command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle 

[Required] vol_db handle on which the lock is to e taken.

=item admin_lock_op 

[Required] lock to be acquired.

=item wait 

[Optional] wait or no-wait.

=item force 

[Optional] force or null.

=back

=item alock_release()

Execute the vol_db alock release command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle

[Required] vol_db handle on which the lock is to be released.

=item admin_lock_op

[Required] lock to be released.

=back

=item alock_rename()

Execute the vol_db alock rename command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle

[Required] vol_db handle on which the lock is to be renamed.

=item curr_lock_handle 

[Required] lock to be changed.

=item new_alock_op 

[Required] new lock.

=back

=item alock_dump()

Execute the vol_db alock dump command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle

[Optional] vol_db handle on which the lock info is dumped. 

=back

=item alock_get_curr_lockers()

Execute the vol_db alock get_curr_lockers command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle

[Required] vol_db handle on which the lock info is required.

=back

=item alock_timeout()

Execute the vol_db alock timeout command with the given arguments. Return the output. User is expected to parse the output.

=over

=item seconds 

[Optional] Change/display timeout for which the lock aquisition waits.

=back

=item alock_print()

Execute the vol_db alock print command with the given arguments. Return the output. User is expected to parse the output.

=over

=item vdb_handle 

[Required] vol_db handle on whichs stats are required.

=back

=back

=head1 SEE ALSO

  NATE documentation
  Server.pm

=head1 AUTHOR

  Chris McRae (mcrae@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/vol_db.pm#1 $
  Last Revised $Date: 2013/01/22 $

=head1 COPYRIGHT AND LICENSE

  Copyright (c) 2008 Network Appliance, Inc.
  All rights reserved.

=cut

