#-*-perl-*-

## @summary Routines for issuing volume commands to the Filer
##
## @description
##      The vol.pm functions are wrappers around NATE's say() function that
##      send a 'vol xxx' command to the filer and return 1 for success or
##      undef for failure.  In the case of failure, an error string
##      is also returned in $EVAL_ERROR ($@)
##
## @synopsis
##      use vvol;
##      $success = vvol_add( &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_container( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_copy( &nbsp;&nbsp;$fullconnspec, [, <i>optional_params</i>] );
##      $success = vvol_create( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = create_volume( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_destroy( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_join( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_lang( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_mirror( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_offline( &nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_online( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_rename( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_restrict( $fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_scrub( &nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_size( &nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_snaprestore( &nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_split( &nbsp;&nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_status( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##      $success = vvol_verify( &nbsp;&nbsp;$fullconnspec, $vol_name [, <i>optional_params</i>] );
##
## @usage
##      <tt>$success = <i>volfunction</i>( $fullconnspec, $vol_name [, options => ""] [, timeout => -1]);</tt>
##
##      where <i><tt>volfunction</tt></i></b> is one of the following:
##      <pre><ul>
##      vvol_add
##      vvol_container
##      vvol_copy
##      vvol_create
##      create_volume
##      vvol_destroy
##      vvol_join 
##      vvol_lang
##      vvol_mirror
##      vvol_offline
##      vvol_online
##      vvol_options
##      vvol_rename
##      vvol_restrict
##      vvol_scrub
##      vvol_size
##      vvol_snaprestore
##      vvol_split
##      vvol_status
##      vvol_verify
##      </ul></pre>
##
##      <tt>$fullconnspec</tt> identifies the host and/or connection to be
##      used to submit the appropriate 'vol xxx' command.  It can be a host
##      object, connection object, hostspec, connspec or hostspec#connspec. 
##      If a connection is not specified, #rsh is used.
##
##      <tt>$vol_name</tt> is the name of the volume on which <i>volfunction</i>
##      will be performed.
##
##      The <tt><i>optional_params</i></tt> argument is a string containing
##      any desired options for the particular vol command.  It is null by
##      default.  If supplied, it must conform to that vol command's syntax.
##
##      The <tt>timeout</tt> argument specifies the maximum amount of time,
##      in seconds, for the function to return.  If that maximum is
##      exceeded, a timeout error is returned.  The default timeout is -1
##      (infinite).
##
## @examples
##      To create a volume with two disks:
##      <blockquote>
##      use vvol;
##      my $h = host($FILER);
##      vvol_create($h, "testvol", options=>2, timeout=>300)
##      &nbsp;&nbsp;&nbsp; or logresult('SCRIPT', $@);
##      # vvol_create succeeded: continue with test
##      </blockquote>
##
## @status deprecated - use Volume.pm instead
## 
## @author dl-johny@netapp.com (FSS WAFL Quality)
##
## @keywords vvol

package vvol;

use strict;
use Module::Common;
our ($Validation, $VERSION);

BEGIN
{
    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @vol_subs = qw(
        vvol_add 
        vvol_container
        vvol_copy
        vvol_create
        create_volume
        vvol_destroy 
        vvol_join 
        vvol_mirror 
        vvol_lang
        vvol_online 
        vvol_offline
        vvol_options
        vvol_rename 
        vvol_restrict 
        vvol_scrub 
        vvol_size
        vvol_snaprestore
        vvol_split 
        vvol_status
        vvol_verify
        vvol_exists
        vvol_purge
        get_root_vol
        vvol_size_info
    );

    Module::Common::setup_export_tags(
        vvol => [ @vol_subs ]
    );
}

use Tharn;

sub vvol_offline ( $$% ) ;
sub vvol_container ( $$% ) ;
sub vvol_restrict ( $$% ) ;

use Utils qw(make_conn);
use Utils::Units qw(convert_to_bytes convert_to_units);
use aggr qw(get_raid_create_opts wait_for_aggr);
use Volume qw(vol_status vol_exists vol_purge);


# Conventions followed
# Indent width - 4
# no tabs - spaces only
# connection type - rsh only
# dont use logresult/tstlog ... to exit... use die with eval
# functions have to return 1 on success 



# ----------------------------------
# Exported sub-routines begin here
# ----------------------------------


## @name ::vvol_add()

## @summary Adds disks as specified in options to the volume

## @synopsis
##   vvol_add($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=> 
##          timeout=>

## @status in-progress

sub vvol_add ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'add',$vol,$args{timeout},$args{options});
    if ($after =~ /has\s+been\s+initiated/ || $after =~ /has\s+completed/) {
        return wait_for_vvol_add($spec,$vol,$args{timeout},"online","growing");
    } else {
        die($after);
    }
    };
}


## @name ::vvol_container()

## @summary Get the aggregate that the volume is associated with.

## @synopsis
##   vvol_container($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=> 
##          timeout=>

## @status public

sub vvol_container ( $$% ) {
    my ($spec, $vol, %args) = @_;

    return eval {
    my $after = say_cmd($spec,'container',$vol,$args{timeout},$args{options});
    if ($after =~ /is hosted by aggregate '(\w+)'/) {
        return $1;
    } else {
        die($after);
    }
    };

}

## @name ::vvol_copy()

## @summary Copy a volume.
## vol copy has subcommands start, status, abort, throttle

## @synopsis
##   vvol_copy($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg args Optional: additional arguments like
##          options=>  specify entire command here
##                     <sub-command> <its options>
##                     e.g abort all
##                     e.g start -s snap1 src_vol dst_vol
##                     e.g status 2
##                     e.g throttle 2 10
##          timeout=>  time to wait for the copy to complete

## @status in-complete, can use for 'vol copy start'

## @notes have to improve this function - but that will mean changing
## say_cmd()

# vol copy abort [ all | operation_number]
# vol copy start [-S | -s snapshot_name] src_vol dest_vol
# vol copy status [operation_number]
# vol copy throttle [operation_number] throttle_value

sub vvol_copy ( $% ) {
    my ($spec,%args) = @_;

    return eval {
    my $expected;
    if ($args{options} =~ /start/) {
        $expected = "vol copy restore .* begun";
    } elsif ($args{options} =~ /stop/) {
        $expected = '';
    } elsif ($args{options} =~ /status/) {
        $expected = '';
    } elsif ($args{options} =~ /help/) {
        $expected = '';
    } else {
        die("Invalid option: ".$args{options});
    }

    my $after = say_cmd($spec,'copy','',$args{timeout},$args{options});
    if ($after =~ /$expected/gm) { 
        return(1);
    } else {
        die("Expected $expected. Got $after");
    }
    };
}

## @name ::vvol_create()

## @summary create a volume 

## @synopsis
##   vvol_create($spec, $vol [,options=>] [,timeout=>])

## @notes To create volumes of various types use create_volume. 
##        create_volume has an option to use an existing volume. 
##        It also to creates paired (old style) volumes.
##        If the test uses paired volumes, reusing the volumes
##        will save time when disks are not zeroed.

## @arg spec    Required: hostspec of the filer
## @arg vol     Required: name of volume
## @arg options Optional: options to be sent down to vol create
## @arg timeout Optional: time in sec to wait for the volume creation 
##              to complete

## @see-also lib/vvol.pm::create_volume

## @status public

sub vvol_create ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'create',$vol,$args{timeout},$args{options});
    if ($after =~ /has\s+completed/) {
        return 1;
    } elsif ($after =~ /has\s+been\s+initiated/) {
        return(wait_for_vvol_status($spec,$vol,$args{timeout},"online","normal|mirrored|raid4"));
    } else {
        die($after);
    }
    };
}


## @name ::create_volume()

## @synopsis create_volume($filer, $vol, 
##                       [, aggr=>0]
##                       [, purge_existing=>0]
##                       [, type=>'flex']
##                       [, vol_size=>'100m']
##                       [, create_opts=>'']
##                       [, disks=>2]
##                       [, raid_mirror=>'normal']
##                       [, cksum_style=>'']
##                       [, raid_type=>'raid4']
##                       [, lang=>'']
##                       [, spcres=>''] );

## @arg filer Required: Hostspec of filer
## @arg aggr  Optional: Name of aggregate for the volume to live in.
##            Default: $vol (same name as volume)
## @arg vol   Required: Name of volume to be created
## @arg type  Optional: Type of volume to be created.
##            Possible values: 'flex' / 'trad'
##            Default: 'flex'
##            vol_type will also work ; since this was the old name
## @arg vol_size  Optional: Size of volume to be created.
##            Will be used iff the type is 'flex'
##            In 'trad' size of the volume is set to size of hosting aggregate
##            Default: 100m
## @arg purge_existing Optional: Flag to denote if existing vol
##            must be purged and a new one created in its place.
##            Default: 0
## @arg create_opts Optional: Options to create the volume.
##            Refer create_raid_obj for default
## @arg disks Optional: Number of disks to use for volume / aggregate creation
##            Will be used iff the type is 'trad'
##            Refer create_raid_obj for default
## @arg raid_mirror Optional: To create a normal or mirrored volume / aggregate.
##            Will be used iff the type is 'trad'
##            Refer create_raid_obj for default / possible values
## @arg cksum_style Optional: Style of checksum to use for volume / aggregate.
##            Will be used iff the type is 'trad'
##            Refer create_raid_obj for default / possible values
## @arg raid_type Optional: Type of raid group to use for volume / aggregate creation.
##            Will be used iff the type is 'trad'
##            Refer create_raid_obj for default / possible values
## @arg lang Optional: Type of language code to use while creating
##            the volume.
##            Default: ''
## @arg spcres Optional: Type of space reservation to use while creating
##            the volume. Possible values : none / file / volume
##            Default: ''
## @arg timeout Optional: Time (in sec) to wait for vol to be created.
##            Default: 1000 (sec)

## @returns 1 if the vol was created by this routine
##          0 if vol was not created in this routine
##          will die with $@ if the routine could not complete for some reason

## @status testing

sub create_volume ( $$% ) {

    my ($spec, $vol, %args) = @_;


    eval {
    my $aggr = defined($args{aggr}) ? $args{aggr} : $vol;
    my $create_opts = defined($args{create_opts}) ?  $args{create_opts} : '';

    # XXX/tapl Update the type of volume
    # if we're using the new convention (trad->pvol, flex->vvol), convert it
    # to the appropriate type used in this function.
    $args{vol_type} = $args{type}; # XXX/Archana backwards compatible
    $args{vol_type} = ($args{vol_type} eq "trad") ? 'pvol' : $args{vol_type};
    $args{vol_type} = ($args{vol_type} eq "flex") ? 'vvol' : $args{vol_type};
    my $vol_type = defined($args{vol_type}) ?  $args{vol_type} : 'vvol';

    my $vol_size = defined($args{vol_size}) ? $args{vol_size} : '100m';
    my $purge_existing = defined($args{purge_existing}) ? $args{purge_existing} : 0;
    my $timeout = defined($args{timeout}) ? $args{timeout} : 1000;

    my $exists = vol_exists($spec, $vol);
    if ($exists) {
        if ($purge_existing) { 
            vol_purge($spec, $vol) or die("Purge of existing vol '$vol' failed :$@\n");
            $exists=0;
            # TBD if its a pvol and if the aggr did not get purged because
            # it contained other volumes etc... then we will not be able to create
            # the pvol again since aggr of same name exists
        } else {
            # check if volume is online
            my $status_obj = vol_status($spec, $vol);
            die($@) if ($@);
            die("Volume '$vol' exists but is not online. Will not be able to use it.")
                if ($status_obj->state() ne 'online');

            # verify if existing volume / aggregate is what we want
            # if its pvol or vvol as we required.
            my $ri = new RaidInfo($spec);
            my $volume = $ri->volume($vol); 
            die("Could not get volume '$vol' information from raid_config")
                if (!$volume);

            die("Existing vol '$vol' is a flexible volume. We require it to be a traditional volume")
                if (($volume->{store_type} eq 'FLEX') && $vol_type eq 'pvol');
            die("Existing vol '$vol' is a traditional volume. We require it to be a flexible volume")
                if (($volume->{store_type} eq 'TRAD') && $vol_type eq 'vvol');

            logcomment("DBG: Will use existing volume '$vol'");
        }
    }

    my $we_created = 0;
    if (!$exists) {
        if ($vol_type eq 'vvol') {
            # Create virtual volume
            #vol create flex_volname
            #         [ -l language_code ]
            #         [ -s none | file | volume ]
            #         aggrname size
            if (!$create_opts) {
                $create_opts .= "-l $args{lang} " if ($args{lang});
                $create_opts .= "-s $args{spcres} " if ($args{spcres});
                $create_opts .= "$aggr $vol_size";
            }
        } elsif ($vol_type eq 'pvol') {
            # Create paired volume
            if (!$create_opts) {
                # get raid create options
                $create_opts = get_raid_create_opts($spec, 
                                disks=>$args{disks},
                                cksum_style=>$args{cksum_style},
                                raid_type=>$args{raid_type},
                                raid_mirror=>$args{raid_mirror},
                                timeout=>$timeout);
                die($@) if ($@);
            }
        } else {
            die("Invalid vol type $vol_type in \$vol_type\n");
        }
        $we_created = vvol_create($spec, $vol, 
                                options=>$create_opts,
                                timeout=>$timeout) or
            die("Volume '$vol' could not be created :$@");
    }

    if ($vol_type eq 'pvol') {
        wait_for_aggr($spec, $vol, $timeout);
    }

    return($we_created);
    };

}


## @name ::vvol_destroy()

## @summary Destroys the volume. (will forcefully destroy the volume)

## @synopsis
##   vvol_destroy($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status public

sub vvol_destroy ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'destroy',$vol,$args{timeout},$args{options});
    if ($after =~ /^Volume \'$vol\' destroyed/){
        return(1);
    } else {
        die($after);
    }
    };
}

## @name ::vvol_join()

## @summary Joins the plexes of a split mirrored volume 
## Specify the timeout argument when using this routine, 
## to break out of the while loop, if vol join fails.

## @synopsis
##   vvol_join($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress


sub vvol_join ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'mirror',$vol,$args{timeout},$args{options});
    if ($after =~ /Invalid\s+operation/ || $after =~ /join/) {
        logcomment ("Cannot join the plexes");
        return $after;
    } else {
        return wait_for_vvol_status($spec,$vol,$args{timeout},"online","mirrored");
    }
    };
}

## @name ::vvol_lang()

## @summary Displays / Sets language on a volume

## @synopsis
##   vvol_lang($spec, $vol, %args)
##
##   # sets the filer language code to simplified chinese
##   vvol_lang($fc, $vol_name, options=>'zh');
##
##   # get the filer language code and character set
##   my %lang = vvol_lang($fc, $vol_name);
##
##   # display language code
##   tstlog('-comment', "Language code is $lang{code}");
##
##   # display character set
##   tstlog('-comment', "Character set is $lang{charset}");
##
## @arg spec    
##   Required: hostspec of the filer
## @arg vol     
##   Required: name of volume
## @arg timeout 
##   Optional: Time (in sec) to wait for the say(). 
##             Default: -1 (forever)
## @arg options
##   Optional: Options to pass to the 'vol lang' command
##             Default: empty string
##
## @returns
##   - If volume name is specified, returns a hash containing the language
##     code and corresponding character set. e.g $lang{code} and $lang{charset}
##
##   - If language code is specified in the options, returns a 1 if filer
##     successful set the language code, otherwise it will die with the error
##     message.
##
## @bug XXX/tapl I have not tested this function wrt setting the language code.
##               I'm only using it to display the lang code and charset on 
##               filer
## @status in-progress

sub vvol_lang ( $$% ) {
    my ($spec,$vol,%args) = @_;

    return eval {
    my $after = say_cmd($spec,'lang',$vol,$args{timeout},$args{options});

    my %lang;

    # set the language on volume
    if (defined($args{options})) {
        if ($after =~ /^\s*The\s+new\s+language\s+mappings\s+will\s+be\s+available\s+after\s+reboot/) {
            return 1;
        }

    } elsif (length($vol)) {
        # display language code and character set used on volume
        if ($after =~ /^\s*Volume\s+language\s+is\s+(.*)\s+\((.*)\)/) {
            $lang{code}    = $1;
            $lang{charset} = $2;
            return %lang;
        }
    }

    die($after);
    };
}


## @name ::vvol_mirror()

## @summary Mirrors a volume with disk list as specified in options.

## @synopsis
##   vvol_mirror($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_mirror ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'mirror',$vol,$args{timeout},$args{options});
    if ($after =~ /has\s+completed/ || $after =~ /has\s+been\s+initiated/) {
        return wait_for_vvol_status($spec,$vol,$args{timeout},"online","mirrored");
    } else {
        die($after);
    }
    };
}

## @name ::vvol_offline()

## @summary Takes the volume offline

## @synopsis
##   vvol_offline($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_offline ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'offline',$vol,$args{timeout});
    my %status = vvol_status($spec,$vol,timeout=>$args{timeout})
        or die($after);
    if (defined($status{State}) && $status{State} =~ /offline/){
        return(1); 
    } else {
        die($after);
    }
    };
}

## @name ::vvol_online()

## @summary Takes the volume online.

## @synopsis
##   vvol_online($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_online ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'online',$vol,$args{timeout},$args{options});
    my %status = vvol_status($spec,$vol,timeout=>$args{timeout})
        or die($after);
    if (defined($status{State}) && $status{State} =~ /online/){
        return(1); 
    } else {
        die($after);
    }
    };
}

## @name ::vvol_options()

## @summary Displays / Sets options on a volume

## @synopsis
##   vvol_options($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_options ( $$% ) {
}

## @name ::vvol_rename()

## @summary Renames a volume to the new volume specified in options.

## @synopsis
##   vvol_rename($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_rename ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'rename',$vol,$args{timeout},$args{options});
    if ($after =~ /'$vol'\s+renamed\s+to\s+'$args{options}'/){
        my %status = vvol_status($spec,$args{options},timeout=>$args{timeout});
        if ($status{State} !~ /online|offline|restrict/) {
            die("Cannot rename volume from \'$vol\' to \'$args{options}\'. Current state is $status{State}");
        } else {
            return(1);
        }
    } else {
        die($after);
    }
    };
}

## @name ::vvol_restrict()

## @summary Restricts the volume

## @synopsis
##   vvol_restrict($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status public

sub vvol_restrict ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'restrict',$vol,$args{timeout});
    my %status = vvol_status($spec,$vol,timeout=>$args{timeout})
        or die($after);
    if (defined($status{State}) && $status{State} =~ /restricted/){
        return(1);
    } else {
        die($after);
    }
    };
}

## @name ::vvol_scrub()

## @summary Start,suspend,resume or stop scrubbing, as specified in option.

## @synopsis
##   vvol_scrub($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_scrub ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    if ($args{options} eq 'start') {
        my $after = say_cmd($spec,'scrub',$vol,$args{timeout},$args{options},1);
        if ($after =~ /vol\s+scrub:\s+started\s+scrub\s+of\s+\/$vol*/g) {
            return(1);
        } else {
            die($after);
        }
    } elsif($args{options} eq 'stop') {
        my $after = say_cmd($spec,'scrub',$vol,$args{timeout},$args{options},1);
        if ($after =~ /vol\s+scrub:\s+stopped\s+scrub\s+of\s+\/$vol*/g) {
            return(1);
        } else {
            die($after);
        }
    } elsif($args{options} eq 'suspend') {
        my $after = say_cmd ($spec,'scrub',$vol,$args{timeout},$args{options},1);
        if ($after =~ /vol\s+scrub:\s+suspended\s+scrub\s+of\s+\/$vol*/g) {
            return(1);
        } else {
            die($after);
        }
    } elsif($args{options} eq 'resume') {
        my $after = say_cmd($spec,'scrub',$vol,$args{timeout},$args{options},1);
        if ($after =~ /vol\s+scrub:\s+resumed\s+scrub\s+of\s+\/$vol*/g) {
            return(1);
        } else {
            die($after);
        }
    }
    };
}

## @name ::vvol_size()

## @summary Sets size of a volume and checks if it has really 
##          grown or shrunk the volume according to what you supply in options.
## @synopsis
##   vvol_size($spec, $vol, %args)
## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>'can be specified as +100m to grow the volume by 100m',
##                   'can be specifid as -100k to shrink the volume by 100k',
##                   'can be specified as 100m to set the vol size to 100m'.
##          timeout=>
##          units=> b|k|m|g|t (defaults to 'k')
## @example vvol_size($filer, $filer_vol ,units=>'m', options=>'+100k')
## @return returns the size of the vol by default in 'k', if units is specified
##         returns the size in that unit.
## @status TESTME

sub vvol_size ( $$% ) {
    
    my ($spec,$vol,%args) = @_;
    my ($orig_size,$new_size, $this_size, $diff_size);
    my $volre = quotemeta($vol);
    my $diff_flag = "";
    my $units= defined($args{units}) ? $args{units} : 'k';
    my $timeout= defined($args{timeout}) ? $args{timeout} : '500';
    my $after ="";
    return eval {
	$after = say_cmd($spec, 'size', $vol, $timeout);
	if ($after =~ /\s+volume\s+\'$volre\'\s+has\s+size\s+(\w+).*/) {
	  $orig_size = $1;
	}else{
		die("Could not get original size of volume :$after");
	}
	if (defined($args{options}) && $args{options} =~ /^(\+|\-)(\w+)$/){
	    $diff_size = $2;
	    if($1 eq '+'){
		$diff_flag = "grow";
	    }elsif($1 eq '-') {
		$diff_flag = "shrink";
	    }
	}elsif(!defined($args{options})) {
	    # no options given. probably just want to get the size of volume
	    $this_size = convert_to_bytes($orig_size);
	    return(convert_to_units($this_size,units=>$units));
	}
	if($diff_flag eq "grow"){
	    $new_size = convert_to_bytes($orig_size) + convert_to_bytes($diff_size);
	}elsif($diff_flag eq "shrink"){
	    $new_size = convert_to_bytes($orig_size) - convert_to_bytes($diff_size);
	}else{
	    $new_size = $args{options};
	    $new_size = convert_to_bytes($new_size);
	}
	$after = say_cmd($spec, 'size', $vol, $timeout, $args{options});
	if ($after =~/vol\s+size:\s+No\s+volume\s+named\s+.*/){
	    die($after);
	}elsif ($after =~ /\s+volume\s+\'$volre\'\s+size\s+set\s+to\s+.*/){
	    my $response= say_cmd($spec, 'size', $vol, $timeout);
	    if ($response =~ /\s+volume\s+\'$volre\'\s+has\s+size\s+(\w+).*/){
		$this_size = convert_to_bytes($1);
		if($this_size == $new_size){
		    return(convert_to_units($this_size,units=>$units));
		}else{
		    die("Did not set the volume to the right size :$after");
		}
	    }else { 
		    die("vol size without size (options) did not ".
                    "respond with the right message : $after");
	    }
	}else{
	    die($after);
	}
    };   
}


## @name ::vvol_size_info()

## @summary Returns the vvol size info

## @synopsis
##   vvol_size_info($spec, $vol, %args)
##
##   # return size of volume only (type=> default to 'volume')
##   my %vol_size = vvol_size_info($spec, $vol);
##
##   # same as above
##   my %vol_size = vvol_size_info($spec, $vol, type=>'volume');
##
##   # return size of snapshot usage only
##   my %sshot_size = vvol_size_info($spec, $vol, type=>'snapshot');
##
##   # return size of volume and snapshot usage
##   my %total_size = vvol_size_info($spec, $vol, $type=>'all');
##
## @arg spec    
##   Required: hostspec of the filer
## @arg vol     
##   Required: name of volume
## @arg timeout 
##   Optional: Time (in sec) to wait for the say(). 
##             Default: -1 (forever)
## @arg type    
##   Optional: Specify what type of info to return. 'volume' will return only
##             the size (total, used, avail) of the volume. 'snapshot' will
##             return the same info of the snapshot only. 'all' will return
##             both info from 'volume' and 'snapshot'. 
##             Default: volume
##
## @returns returns the total ,used , avail size of the vol in kb
##          die's if df fails
## @example %all_sizes=vvol_size_info($filer,$vol);
## @status TESTME

sub vvol_size_info( $$% ) {

    my ($spec,$vol,%options) = @_;

    my $timeout = defined($options{timeout}) ? $options{timeout} : '-1';
    my $type    = defined($options{type})    ? $options{type}    : 'volume';

    return eval {
	my %args;
        my %snapshot_args;
	#Create a connection to use;
	my $conn = make_conn($spec);
	my(undef,undef,$after)=say($conn,"df /vol/$vol",'-timeout', $timeout);
	if($after =~ m#.*/vol/$vol/\s+(\d+)\s+(\d+)\s+(\d+).*#s) {
	   $args{total}=$1;
	   $args{used}=$2;
	   $args{avail}=$3;

           # return size of vol only
           if ($type eq "volume") {
               return %args;

           } else {
               if($after =~ m#.*/vol/$vol/.snapshot\s+(\d+)\s+(\d+)\s+(\d+).*#s) {
                  $snapshot_args{total} = $1;
                  $snapshot_args{used}  = $2;
                  $snapshot_args{avail} = $3;
              }

               # return size of snapshot usage only
               if ($type eq "snapshot") {
                   return %snapshot_args;

               } elsif ($type eq "all") {
                   # return the size of vol and snapshot usage
                   foreach my $key (keys %args) {
                       $args{$key} += $snapshot_args{$key};
                   }
                   return %args;
               } else {
                   die("Unknown \'type\' argument: $type");
               }
           }

       }else{
	   die("df returned unexpected results $after");
       }
    };
}


## @name ::vvol_snaprestore()

## @summary Reverts volume to a previous snapshot

## @synopsis
##   vvol_snaprestore($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_snaprestore ( $$% ) {
}

## @name ::vvol_split()

## @summary Split a mirrores volume to a new volume specified in options

## @synopsis
##   vvol_split($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_split ( $$% ) {
    my ($spec,$vol,%args) = @_;
    return eval {
    my $after = say_cmd($spec,'split',$vol,$args{timeout},$args{options});
    if ($after =~ /new\s+volume\s+$args{options}\s+created/g) {
        return(1);
    } else {
        die($after);
    }
    };
}


## @name ::vvol_status()

## @summary Returns volume name,state and status in a hash.

## @synopsis
##   vvol_status($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status deprecated, use Volume::vol_status instead

sub vvol_status ( $$% ) {
    my ($spec,$vol,%args) = @_;

    return eval {

    logresult('warn', "vvol::vvol_status is deprecated. Migrate to Volume::vol_status");

    my $after = say_cmd($spec,'status',$vol,$args{timeout},$args{options});

    # if the volume doesn't exist, just die and return.
    # have to do this as this function returns either a hash or a scalar, and
    # some caller expects this function to return a hash. perl will issue a
    # warning if this function returns a scalar when caller expected a hash.
    # by die'ing, the caller can trap the exception and avoid the perl warning.
    # see vvol_{offline|online|restrict}() for example.
    if ($after =~ /^\s*vol\s+status:\s+No\s+volume\s+named\s+\'$vol\'\s+exists/) {
        die($after);
    }

    # '-l' option to display language code
    if (defined($args{options}) &&
        ($args{options} =~ /\s*\-l\s*/)) {

        if ($after =~ /^\s*vol\s+status\:\s+No\s+volume\s+named\s+\'$vol\'\s+exists\./) {
            die($after);
        }

        my %status_lang;

        #tapl> vol status tap -l
        #     Volume Language
        #        tap ja (Japanese euc-j*)
        
        foreach my $line (split(/\n/, $after)) {
            if ($line =~ /Volume\s+Language/) {
                next;
            } elsif ($line =~ /^\s*$vol\s+(.*)\s+\((.*)\)/) {
                $status_lang{code}    = $1;
                $status_lang{charset} = $2;
                return %status_lang;
            } 
        }

        die("Unable to parse text: $after");

    } elsif ($after !~ /vol\s+status:/) {
        my %status;

        my @lines = split(/\n/, $after);
	#get rid of last \r on each line
	for (my $i=0; $i < $#lines; $i++) {
		$lines[$i] =~ s/\n$|\r$//g;
	}

        # Simple 'vol status' format as of 08/25/2003:
        # 
        #filer> vol status vol0
        #        Volume State      Status            Options <= line 1
        #          vol0 online     raid4, flex       root    <= line 2
        #               Containing aggregate: 'vol0'              <= line 3

        if ($lines[0] !~ /\s+Volume\s+State\s+Status\s+Options/) {
                logcomment("Warning: new vol status format ?\n");
        }

	# make sure the 2nd line exists before grepping. 
	# XXX/tapl contary to my earlier belief, when the 2nd line doesn't
	# exist, it usually means the the volume in question is config locked.
	# this is due to a bug in 'vol status' for not being able to list
	# the status of a locked volume. see burt 108336 for more details.
        if (defined($lines[1]) &&
            $lines[1] =~ /$vol\s+(\w+)\s+((\w,?\s?)+)\s+([\s\w,=]+)/) {
            $status{State}   = $1;
            $status{Status}  = $2;
            $status{Options} = $4;
            $status{Vol}     = $vol;
            $status{Type}    = ($status{State}=~/trad/) ? 'trad' : 'flex';
        } else {
	    # XXX/tapl volume is config locked. die() with a message saying
	    # that volume is config locked. this will be changed once
	    # burt 108336 is fixed.
	    die("Volume '$vol' is currently locked.");
        }

        if($#lines == 3) {
                $lines[2] =~ /^\s+([\s\w-]+)/;
                $status{Status} .= " $1";

                $lines[3] =~ /\s+Containing aggregate:\s+'(\w+)'/;
                $status{Aggr} = $1;

        } elsif ($#lines == 2) {
                if ($lines[2] =~ /\s+Containing aggregate:\s+'(\w+)'/) {
                        $status{Aggr} = $1;
                } elsif ($lines[2] =~ /\s+([\s\w-]+)/) {
                        $status{Status} .= " $1";
                }
        }
        return %status;

    } else {
        return $after;
    }
    }
}

## @name ::vvol_verify()

## @summary Verify a volume

## @synopsis
##   vvol_verify($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          options=>  
##          timeout=>

## @status in-progress

sub vvol_verify ( $$% ) {
}


## @name ::vvol_exists()

## @summary return 1 if specified vol exists, 0 otherwise

## @synopsis
##   vvol_exists($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##          timeout=>

## @status deprecated, use Volume::vol_exists instead

sub vvol_exists( $$% ) {
    my ($spec, $vol, %args) = @_;

    eval {
    logresult('warn', "vvol::vvol_exists is deprecated. Migrate to Volume::vol_exists");
    my $conn = make_conn($spec);

    my $after = say_cmd($conn,'status',$vol,$args{timeout});
    if ($after =~ /vol status: No volume named '$vol' exists/) {
        return(0);
    }
    return(1);
    }; 
}

## @name ::vvol_purge()

## @summary Will offline and destroy the volume if existing.
##          Will not fail if the volume does not exist.

## @returns returns 1 if purge of vvol succeeded, undef otherwise

## @synopsis
##   vvol_purge($spec, $vol, %args)

## @arg spec Required: hostspec of the filer
## @arg vol  Required: name of volume
## @arg args Optional: additional arguments like
##           timeout=>

## @status public

sub vvol_purge ( $$% ) {

    my ($spec, $vol, %args) = @_;

    eval {
    my $conn = make_conn($spec);

    if (vvol_exists($conn,$vol,timeout=>$args{timeout})) {
        vvol_offline($conn,$vol,timeout=>$args{timeout})
            or die("Unable to offline $vol: $@");
        vvol_destroy($conn,$vol,timeout=>$args{timeout},options=>"-f")
            or die("Unable to destroy $vol: $@");
    } else {
        print ("\nVolume $vol does not exist");
    }
    return(1);
    };
}

## @name ::get_root_vol ()

## @summary Will get the root volume on the filer

## @returns returns root volume if succeeds, undef otherwise

## @synopsis
##   get_root_vol($spec)

## @arg spec   Required: hostspec of the filer

## @status deprecated, use Volume::vol_root instead

sub get_root_vol( $ ) {
	my ($spec) = @_;

    eval {
    logresult('warn', "vvol::get_root_vol is deprecated. Migrate to Volume::vol_root");
	my $conn = make_conn($spec);

	my (undef, undef, $out) = say($conn, "vol status");
	my ($root) = ($out =~ /\s+(\S+).*root/); 
	die("get_root_vol: Could not get root vol") if (!$root);
	return($root);
    }

}


#-----------------------------
# Local subroutines begin here
#-----------------------------

# @name ::wait_for_vvol_add()

# loop, calling 'vol status $vol' until the state matches $state and 
# status does *NOT* match $status

sub wait_for_vvol_add {
    my ($spec, $vol, $timeout, $state, $status) = @_;

    my $endtime = timeout2time( ($timeout) ? $timeout : 1800);
    while (time() < $endtime) {

        # get the raid info via '-r' option in vol status
        my $after = say_cmd($spec, 'status', $vol, $timeout, "-r");

        # tapl> vol status tap -r
        # Aggregate tap (online, raid4, growing) (block checksums)
        #   Plex /tap/plex0 (online, normal, active)
        #     RAID group /tap/plex0/rg0 (normal)

        foreach my $line (split(/\n/, $after)) {
            # only grep for the first line
            if ($line =~ /^\s*Aggregate\s+.*\s+\((\w+),\s+(.*)\)\s+\(/) {
                my $add_state = $1;
                my $add_status_list = $2;

                if ($add_state =~ /$state/ &&
                    $add_status_list !~ /$status/) {
                    return(1);
                }
            } else {
                last;
            }
        }
        sleep 10;
    }
    die("Timed out waiting for volume '$vol' to become ($state,$status)\n");
}

# @name ::wait_for_vvol_status()

# loop, calling 'vol status $vol' until the state and status match
# $state and $status.

sub wait_for_vvol_status {
    my ($spec, $vol, $timeout, $state, $status) = @_;

    eval {
    my $endtime = timeout2time( ($timeout) ? $timeout : 1800);
    while (time() < $endtime) {
        my $status_obj = vol_status($spec, $vol);
        die($@) if ($@);
        if ($status_obj->state() =~ /$state/ && $status_obj->status() =~ /$status/){
            return(1);
        } else {
            sleep 10;
        }
    }
    die("Timed out waiting for volume '$vol' to become ($state,$status)\n");
    };
}

# @name ::say_cmd()

# wrapper to Tharn::say() with more stuff

sub say_cmd ( $$$$;$$ ) {

    my ($spec, $subcmd, $vol, $timeout, $options, $flag) = @_;

    $options = $options ? $options : "";
    $flag = $flag ? $flag : "";

    my $conn = make_conn($spec);
    $timeout = $timeout ? $timeout : $conn->{timeout};
    my $saytime = timeout2time($timeout);
    my $cmd;
    if ($flag) {
        $cmd = "vol $subcmd $options $vol";
    } else {
        $cmd = "vol $subcmd $vol $options";
    }

    for(my $t = 0; $t < 6; ++$t) {
	my @complaint;
	my (undef,undef,$after) = say ($conn,"$cmd",
                    '-timeout',$saytime,
                    '-eval', sub { @complaint = @_;1; });

	if ($after =~ /in use by other operations/) {
		sleep 10;
		next;
	}

	if (@complaint) {
		die(join('\n',@complaint));
	} else {
		return($after);
	}
    }
}

sub contains ( $@ ) {
    my $item = shift @_;
    return(scalar(grep { $_ eq $item } @_));
}


1;



