# @summary library to run wack on a filer
##
## @description
##   This library contains one function 'wack', which uses a filer's
##   console to (re)boot a filer to the boot menu, run wack
##   (WAFL_check), and wait for it to finish, answering any queries
##   wack may have along the way.
##
## @synopsis
##
##   use wack;
##
##   wack($host,...options...);
##
## @example
##
##   # boot into wack to confirm that $FILER_VOL is clean, then boot
##   # back into ontap
##
##   use wack;
##   use boot;
##   my $h = host($FILER);
##   my $res = wack($h,args=&gt;"$FILER_VOL");
##   tstlog('-fatal',$@) if (!$res);
##   tstlog('-fail',"vol $FILER_VOL is not supposed to be dirty") if ($res ne WACK_CLEAN);
##   boot($h) or tstlog('-script',$@);
##
##   #for ontap-ng filers/cluster-mode
##   use wack;
##   use Server;
##   my $Host=host($FILER);
##   $Server = new Server;
##   my $res = wack($Host,args=>$FILER_VOL,server=>$Server);
##   tstlog('-fatal',$@) if (!$res);
##   tstlog('-fail',"vol $FILER_VOL is not supposed to be dirty") if ($res ne WACK_CLEAN);
##   $Server->transit_to(state => $Server->CLI(),no_connections=>1,) or tstlog('-script',$@);
##
## @usage
##
## <HR><center><b>Constants</b></center>
##
## This library exports the following constants:
##
## <b>WACK_CLEAN</b>:
## (wack ran to completion and did not ask whether to commit any
## changes, so it must not have found filesystem errors)
##
## <b>WACK_DIRTY</b>:
## (wack ran to completion but it did ask whether to commit changes or
## whether to delete something from the filesystem, so it must have
## found filesystem errors of some kind)
##
## <b>WACK_NOTDONE</b>:
## (wack was not performed, if volume specified in args did
## not exist on filer)
##
## <b>WACK_INFONEEDED</b>:
## (wack changes were not committed, as developer needs information
## to fix the issue)
##
## <b>WACK_CONSOLENOTLOGGED</b>:
## (wack was not performed, as user has not logged through
## console)
##
## <HR><center><b>wack</b></center>
##
## <tt>wack($f, [args=&gt;"-all"], [timeout=&gt;-1], [commit=&gt;0],
## [output=&gt;...],
## [maxcycle=&gt;...], [powercycle=&gt;...], [bootimage=&gt;...],
## [haltcmd=&gt;...], [prefix=>"FILER"] );</tt>
##
## Reach the BOOT_MENU of $filer (see transit_to in lib/boot.pm), and
## issue "wack".  Answer "yes" to questions like "Check <i>volume
## name</i>?".  Answer yes or no (based on the value of the commit
## option) to "Delete?" or "Commit changes for volume <i>volume name</i>?".
## Finish (without typing anything) when we see "press any key to
## reboot system".
##
## On success, return one of two true values: either WACK_DIRTY, if
## ever asked whether to commit changes to a volume or whether to
## delete something, or WACK_CLEAN otherwise.  On error, return undef,
## and return the text of the error in $@.
##
## <b>$f</b> - host record or connection record or connspec
## (for host records, connect to that host's #console)
##
## <b>server</b> - mandatory if filertype is ontapNG. the value passed
## to this option should be a Server object.
##
## <b>args</b> - what arguments to pass to "wafl_check".  Default is
## "-all" which says to check all volumes; see the documentation for
## wack (the na_WAFL_check.1h manpage) for other options.
##
## <b>timeout</b> - if wack has not finished by this time, return with
## an error.  timeout is a usual NATE timeout, which can be a number
## of seconds, an absolute time by when to finish, or -1 to mean
## unlimited.  The default is -1 (unlimited).
##
## <b>commit</b> - if 0, which is the default, answer "no" to the
## question "Commit changes for volume <i>volume name</i>?"  and to the
## question "Delete? ".  if 1, then answer "yes" to such questions.
##
## <b>selection</b> - "c" to select inofile of previous cp, "r" for
## regular inofile, or snap number "N" in case snapshot number to revert
## is known. default reply to selection query is "c".
##
## <b>infoneeded</b> - if 0, which is the default, answer "no" to the
## question "information needed by developer?"
## if 1, then answer "yes" to such questions.
##
## <b>conslogged</b> - if 1, which is the default, answer "yes" to the
## question "connsole logged?"
## if 0, then answer "no" to such questions.
##
## <b>output</b> - if specified, the value passed to this option
## should be a reference to a scalar; the scalar will be rewritten to
## contain the entire output from wack.
##
## <b>force</b> - if specified will run IOC forcefully irrespective
## of the build.
##
## <b>prefix</b>, <b>maxcycle</b>, <b>powercycle</b>, <b>bootimage</b> and <b>haltcmd</b>
## - options to "transit_to" when attempting to reach the boot menu in
## order to type "wack".  See documentation on "transit_to" in lib/boot.pm
##
## @see-also lib/boot.pm Node/NG/Transit.pm
## @author lincoln@netapp.com
##
## @history
##    1. Added the CodePath to perform WAFLIRON if the below API is called
##       on ONTAP version > BR
##    2. Sanpaul :: Added support to 10-mode Runs

package wack;

use strict;
use vars ('@ISA', '@EXPORT');
require Exporter;

BEGIN {
    our @ISA = ('Exporter');
    our @EXPORT = ('wack',            'WACK_CLEAN',
                   'WACK_DIRTY',      'WACK_NOTDONE',
                   'WACK_INFONEEDED', 'WACK_CONSOLENOTLOGGED');
}

use Tharn;
use boot;
use Filer::Version qw(get_wafl_version
  WV_BOILERMAKER_MAX ontap_to_wafl_ver);
use Wafliron qw(wafliron_start wafliron_status wafliron_commit wafliron_reject wafliron_on_root); 
use Aggregate qw(aggr_status aggr_online aggr_list aggr_root);
use SFOTest qw(get_root_object);
use LogReader::MlogMessages;
use Volume qw(vol_root);
use constant WACK_CLEAN            => 'CLEAN';
use constant WACK_DIRTY            => 'DIRTY';
use constant WACK_NOTDONE          => 'NOTDONE';
use constant WACK_INFONEEDED       => 'NOTNEEDED';
use constant WACK_CONSOLENOTLOGGED => 'NOTLOGGED';

# convert $spec (which might be a hostspec or connspec) to a
# connection object (to console, if not specified)

sub make_conn ( $ ) {
    my ($spec) = @_;
    return eval {
        if (ref($spec) =~ /Connectrec/) {
            return $spec;
        } elsif (ref($spec) =~ /Hostrec/) {
            return connect($spec->{id} . "#console");
        } else {
            return connect($spec);
        }
    };
}    #END of sub make_conn

use constant RE_CONTINUE       => "Continue with boot\\?";
use constant RE_COMMIT         => "Commit changes";
use constant RE_DELETE         => "Delete\\? ";
use constant RE_CHECK          => "Check (\\w:\?)\\S*\\? ";
use constant RE_DONE           => "Press (any key|Enter) to reboot system";
use constant RE_BOOTING        => "Boot Loader version";
use constant RE_BOOTING2       => "CFE version";
use constant RE_CONSOLE_LOGGED => "Continue\\?\\?";
use constant RE_INFO_NEEDED    => "on these corrupt disks\\?";
use constant RE_SELECTION      => "Selection\\?";
use constant RE_NGS            => "NetApp Global Services\\? \\?";

sub wack ( $@ ) {
    my ($spec, %opts) = @_;

    $opts{infoneeded} = 0 unless defined $opts{infoneeded};
    $opts{conslogged} = 1 unless defined $opts{conslogged};
    $opts{force}  = 0 unless defined $opts{force};
    $opts{commit} = 1 unless defined $opts{commit};

    return eval {
        my $conn = make_conn($spec) or die $@;

        my $args    = defined($opts{args})    ? $opts{args}    : "-all";
        my $timeout = defined($opts{timeout}) ? $opts{timeout} : $::TIMEOUT;
        my $outputref = $opts{output};
        my $reviewref = $opts{review};
        my $endtime   = Tharn::timeout2time($timeout);
        my @aggregates;
        my @unstable_aggr_list;      # Store the Array Name where Corruption was found but changes rejected
        my $root_aggr;
        my $root_vol;
        my $commit = $opts{commit};

        #Placeholder Variables
        my $ret  = WACK_CLEAN;        # Stores the API run return value
        my $inst = $conn->hostp();    # Stores the Connection Object
        my $filer_mode;               # Store the Mode of the Filer
        my $wafl_version;             # Store the WAFl Version of the Filer
        my @Nodes;
        my $Node;
        my $fconn;
        my $temp;                     # Temporary Placeholder Variable
        my $ioc_status;               # PlaceHolder for IOC Status Msg
        my $run_ioc_on_root = 0;      # Placeholder Flag for Running IOC on Root Aggr
        my $Server;                   # Server Object
        my $ems_handle;               # EMS Handler
        my @aggregate_status;         # Array for Aggr Status
        my %aggr_hash;                # Hash for Aggr Status

        # Placeholder for Wafl Iron Test Runs
        my ($corruption_found, $messages);

        if (defined($outputref)) {
            $$outputref = '';
        } #END of if (defined($outputref))

        if (defined($reviewref)) {
            $$reviewref = '';
        } #END of if (defined($reviewref))

        # Check the $hosttype @ Tharnhost for Filer Mode
        if ($conn->hostp()->hosttype() =~ m/filer-ng/) {
            $filer_mode = 10;
        } else {
            $filer_mode = 7;
        } #END of if ($conn->hostp()->hosttype() =~ m/filer-ng/)
        logcomment("WACK : Identified the test ENV Mode = '$filer_mode' mode");
        
        # Get the WAFL version Of the Filer
        if ($filer_mode == 10) {
            $Server = $opts{server};
            unless (ref($Server) and $Server->isa('Server')) {
                    die "server object was not passed to subroutine";
            }
            @Nodes = $Server->servers;
            $Node  = $Nodes[0];

            $temp = $Server->system_node_run(
                                      node       => $Node,
                                      command    => "version",
                                      '-Timeout' => 300);
            #Regexp the WAFL Version
            $temp =~ /Release\s(.*?)\s/;
            $wafl_version = ontap_to_wafl_ver($1);
        } else {
            # Use the standard 7-mode version.pm
            $wafl_version = get_wafl_version(host($inst));
        } #END of if ($filer_mode == 10)

        # Handle $args = -all
        if ($args =~ /all/) {
            logcomment("WACK : Selected to run on All Aggrs - Getting aggregate list");
            # Since aggr_list() returns a arrya of Hash in 10-mode, convert it to array.
            if ($filer_mode == 10) {
                my @aggr = $Server->storage_aggregate_show;
                for ($temp = 0; $temp < scalar(@aggr); $temp++) {
                    # Omit Inclusion to Aggr List, if the aggr is of Vol Style = Striped
                    if ($aggr[$temp]{'volume-style'} eq "striped") {
                        logcomment("Since IOC nnot supported on Stripped Aggr, ".
                                   "Omiting the Aggr $aggr[$temp]{'aggregate'}");
                    } else {
                        push(@aggregates,$aggr[$temp]{'aggregate'});
                    }
                }
            } else {
                @aggregates = aggr_list($conn);
            }
        } else {
            #Handling single as well as multiple aggrs
            # Splitting given args in case of multiple aggrs
            @aggregates = split(/\s/, $args);
        } # End of if ($args eq "-all")

        logcomment ("WACK : Aggregates to test '@aggregates'");

        if ((( $wafl_version > WV_BOILERMAKER_MAX) || ($opts{force} == 1)) && ($filer_mode == 10)) {

            logcomment(
                    "\n\n\n=================================================="
                  . "\n\t\tWARNING & IMPORTANT NOTE"
                  . "\n=================================================="
                  . "\n Filer Running in '10' mode & 'POST-BR' Release"
                  . "\n WACK is not supported in this S/W release - Running IOC"
                  . "\n==================================================\n\n");

            # Get the Root Object Name
            $root_aggr = get_root_object(
                        serverobj    => $Server, 
                        filername    => $Node,
                        object_type  => "aggr");

            $root_vol = get_root_object(
                        serverobj    => $Server,
                        filername    => $Node,
                        object_type  => "vol");

            logcomment("INFO : Identified Root Vol & Aggr as '$root_vol' & '$root_aggr'");

            # Configure the Required Flags
            $Server->system_node_run(
                          node    => $Node,
                          command => "setflag wafl_inconsistent_panic_testing 1");

            $temp = $Server->system_node_run(
                                      node       => $Node,
                                      command    => "printflag wafl_inconsistent_panic_testing",
                                      '-Timeout' => 300);
            $temp =~ /(\d)/;
            die ("Unable to Set the Flag 'wafl_inconsistent_panic_testing'")
                if ($1 != 1);

            $Server->system_node_run(
                          node    => $Node,
                          command => "setflag wafl_iron_panic_on_corruption 0");

            $temp = $Server->system_node_run(
                                      node       => $Node,
                                      command    => "printflag wafl_iron_panic_on_corruption",
                                      '-Timeout' => 300);
            $temp =~ /(\d)/;
            die ("Unable to UnSet the Flag 'wafl_iron_panic_on_corruption'")
                if ($1 != 0);

            logcomment("INFO : Successfully set the required WAFL flag");

            # Set the Filer Priviledges to TEST
            $Server->set(privilege => "test");

            foreach my $aggr (@aggregates) {
                if ($aggr eq $root_aggr) {
                    logcomment(  "INFO : Starting IOC on the Root Aggr "
                               . "only after completing IOC on other aggregates"
                               . " to avoid failures");
                    $run_ioc_on_root = 1;
                } else {
                    # Start the EMS Logger
                    $ems_handle = &_log_start($Server);

                    # Offline the Aggregate
                    logcomment("INFO : Offline & Confirm the Non-Root aggr '$aggr'");
                    $Server->storage_aggregate_offline(
                                                aggregate         => $aggr,
                                                'unmount-volumes' => "TRUE");

                    my @temp_arr = $Server->storage_aggregate_show(
                                                aggregate => $aggr);
                    die ("Failed to Offline the Aggr '$aggr'")
                        if ($temp_arr[0]{'state'} ne "offline");

                    # Start IOC on the Aggregate
                    logcomment("INFO : Start & Confirm IOC Run on Non-Root aggr '$aggr'");
                    $Server->storage_aggregate_wafliron_start(
                                                aggregate        => $aggr,
                                                '-Timeout'       => $timeout,
                                                'optional-commit'=> "TRUE");
                    $temp = 0;
                    do {
                        $ioc_status = $Server->storage_aggregate_wafliron_show(
                                                                   aggregate  => $aggr);
                        $temp = $temp + 10;
                    } until (($ioc_status->[0]{'iron-status'} =~ /mounting|scanning/) ||
                             ($temp > $timeout));

                    die("IOC didn't start on the Aggr '$aggr' even after '$timeout' Seconds")
                        if ($ioc_status->[0]{'iron-last-start-error-info'} ne "Successful.");

                    # Confirm IOC Completed on the Aggr
                    logcomment("INFO : Confirm that IOC completed on Non-Root aggr '$aggr'");
                    $temp = 0;
                    do {
                        $ioc_status = $Server->storage_aggregate_wafliron_show(
                                                                   aggregate  => $aggr);
                        $temp = $temp + 10;
                    } until (($ioc_status->[0]{'iron-status'} =~ /not_running/) || 
                             ($temp > $timeout));

                    die ("IOC is reported running for a prolonged time [>$timeout Secs] - Could be Hung")
                        if (($ioc_status->[0]{'iron-status'} !~ /not_running/) &&
                            ($temp >= $timeout));
                    sleep 30; # avoid panic while aggr offlining

                    logcomment("INFO : Check the Aggr Status");
                    eval {
                        @aggregate_status = @{$Server->storage_aggregate_show("aggregate" => $aggr)};
                    };

                    %aggr_hash = %{$aggregate_status[0]};
                    logcomment("INFO : Aggr in $aggr_hash{'state'} State");
                        
                    logcomment("INFO : Review for IOC Status [Commit / Reject]");
                    eval {
                        # Check the IOC Review Status
                        my $cmd = "storage aggregate wafliron review -aggregate $aggr";
                        $Server->cli(cmd => $cmd, output => \$temp);
                    };
                    $$reviewref = $temp;

                    if (($aggr_hash{'state'} !~ m/restricted/) || ($temp !~ /clearing/i && $@)) {
                        logcomment("INFO : IOC Found No Corruption on the Aggr '$aggr'");
                    } else {    
                        logcomment("INFO : IOC Found some Corruption & is waiting for Action");
                        # 
                        $ret = WACK_DIRTY;
                        if ($commit == 1) {
                            logcomment("INFO : Commit the Changes");
                            $Server->storage_aggregate_wafliron_commit(
                                                                aggregate  => $aggr);
                        } else {
                            logcomment("INFO : Reject the Changes");
                            $Server->storage_aggregate_wafliron_reject(
                                                                aggregate  => $aggr);
                        } # END of if ($commit == 1)
                    } # END of if ($@)

                    # Stop the EMS Logger
                    my $messages = &_log_stop($ems_handle);

                    # Append the EMS Messages
                    $$outputref .= $messages;

                    if ($ret eq WACK_DIRTY) {
                        logcomment( "INFO : Found corruption on $aggr: Returning '$ret'");
                        return $ret;
                    }

                    # Online the Aggregate
                    $Server->storage_aggregate_online(
                                                aggregate => $aggr);

                    @temp_arr = $Server->storage_aggregate_show(
                                                aggregate => $aggr);
                    die ("Failed to Online the Aggr '$aggr'")
                        if ($temp_arr[0]{'state'} ne "online");

                } # END of if ($aggr =~ /^$root_aggr$/)
            } # END of FOREACH

            my $lines;

            # Run IOC on Root Aggr        
            if ($run_ioc_on_root) {
                $::IRON_ON_ROOT = "core";
                logcomment("INFO : Running IOC on root aggr");

                # Transiting to IBM (interactive boot menu) prompt,via  HALT,CFE prompts.
                # direct transition to IBM does'nt work properly.
                eval {
                    logcomment("INFO : Transit to BOOT MENU");
                    $Server->transit_to(state => $Server->HALT());

                    $Server->transit_to(state          => $Server->IBM(),
                                        no_connections => 1,
                                        state_timeout  => $timeout);
                };
                die ("Unable to Transit the filer to Boot Menu : $@")
                    if ($@);                

                $temp = $Server->state();
                die ("Unable to Transit filer to IBM State")
                    if ($temp !~ /IBM/);

                eval {
                    logcomment("INFO : Run IOC from the BOOT MENU");
                    # issue wafliron command and wait for it to complete
                    my ( $match, $before, $after, $prompt )
                        = say( $conn, 'wafliron -o', '-wait', 
                               'System rebooting...', '-noprompt',
                               '-Timeout', $timeout);

                    logcomment("INFO : Transit to CLI State");
                    $Server->transit_to(state => "CLI");

                    $temp = $Server->state();
                    die ("Unable to Transit filer to CLI State")
                        if ($temp !~ /CLI/);

                    $lines .= $match . $before . $after . $prompt;
                    $$outputref .= $lines;

                }; # END of Eval

                die ("Unable to Transit the filer to CLI Mode : $@")
                    if ($@);

                $ems_handle = &_log_start($Server);

                logcomment("INFO : Check the Root Aggr WaflIron Review.");
                $Server->set(privilege => "test");

                # Commit / Reject the Changes on IOC run on Root Aggr
                eval {
                    # Check the IOC Review Status
                    $temp = $Server->storage_aggregate_wafliron_review(
                                                               aggregate  => $root_aggr);
                };

                if ($@) {
                    logcomment("INFO : IOC Found No Corruption on the Aggr '$root_aggr'");
                    $corruption_found = 0;
                } elsif ($temp =~ /No wafliron changes/) {
                    logcomment("INFO : IOC Found No Corruption on the Aggr '$root_aggr'");
                    $corruption_found = 0;
                } else {    
                    logcomment("INFO : IOC Found some Corruption & is waiting for Action");
                    $corruption_found = 1;
                    if ($commit == 1) {
                        logcomment("INFO : Commit the Changes");
                        $Server->storage_aggregate_wafliron_commit(
                                                            aggregate  => $root_aggr);
                    } else {
                        logcomment("INFO : Reject the Changes");
                        $Server->storage_aggregate_wafliron_reject(
                                                            aggregate  => $root_aggr);
                    } # END of if ($commit == 1)
                } # END of if ($@)

                if ($corruption_found) {
                    logcomment( "IOC found some corruption on root aggr '$root_aggr'");
                    $ret = WACK_DIRTY;
                }    # End of if ($corruption_found)

                $::IRON_ON_ROOT = undef;
                my $messages = &_log_stop($ems_handle);
                # Append the EMS Messages
                $$outputref .= $messages;                

                logcomment("INFO : Online All Aggrs on the Filer.");
                # Bring all Aggrs Online
                foreach my $aggr (@aggregates) {
                    $Server->storage_aggregate_online(
                                                aggregate  => $aggr);
                } # END of FOR
            }    # End of if (defined($root_aggr))

        } elsif ((( $wafl_version > WV_BOILERMAKER_MAX) || ($opts{force} == 1)) && ($filer_mode == 7)) {

            logcomment(
                    "\n\n\n=================================================="
                  . "\n\t\tWARNING & IMPORTANT NOTE"
                  . "\n=================================================="
                  . "\n Filer Running in '7' mode & 'POST-BR' Release"
                  . "\n WACK is not supported in this S/W release - Running IOC"
                  . "\n==================================================\n\n");

            # Checking for root aggr
            $root_aggr = aggr_root($conn);
            logcomment("INFO : Root Aggregate is '$root_aggr'");

            foreach my $aggr (@aggregates) {
                if ($aggr eq $root_aggr) {
                    logcomment(  "INFO : Starting IOC on the Root Aggr "
                               . "after completing IOC on other aggregates"
                               . " to avoid failures");
                    $run_ioc_on_root = 1;
                } else {
                    logcomment("INFO : Starting IOC on normal aggr $aggr");
                    ($corruption_found, $messages) =
                         wafliron_start($conn, $aggr, oc => 1, timeout => $timeout);

                    # Append the WaflIron Run Messages.
                    $$outputref .= $messages;

                    if ($corruption_found) {
                        logcomment("INFO : IOC found some Corruption");
                        $ret = WACK_DIRTY;
                    } else {
                        logcomment("INFO : IOC did not find any Corruption");
                        $ret = WACK_CLEAN;
                    }    #End of $corruption_found

                    #Performs commit/reject if the aggregate is waiting
                    #for the same after IOC
                    _wafliron_commit_reject_changes($conn, $aggr, $commit);

                    if ($ret eq WACK_DIRTY) {
                        logcomment("INFO : Found corruption on $aggr - ".
                                   "Returning '$ret'");
                        return $ret;
                    }
                }    ## End of else
            }    # END of foreach

            if ($run_ioc_on_root) {

                $::IRON_ON_ROOT = "core"
                    unless (defined($::IRON_ON_ROOT));

                # Connect console
                my $conn_type = $conn->conntype();
                if ($conn_type !~ m/console|pm_cons/) {
                    $fconn = connect( $conn->hostp()->hostname() . "#console", -login );
                } else {
                    $fconn = $conn;
                }

                $root_vol  = vol_root($conn);
                logcomment("INFO : Running IOC on Root Aggr ".
                           "[$root_aggr & $root_vol]");
                ($corruption_found, $messages) =
                    wafliron_start($conn, $root_aggr,
                                   oc       => 1,
                                   root_vol => $root_vol,
                                   timeout  => $timeout);
                $$outputref .= $messages;

                if ($corruption_found) {
                    logcomment( "INFO : IOC found some corruption on root ".
                                "aggr '$root_aggr'");
                    $ret = WACK_DIRTY;
                } else {
                    logcomment("INFO : IOC did not find any corruption on ".
                               "the root aggr");
                    $ret = WACK_CLEAN;
                }    #End of if ($corruption_found)

                if ($corruption_found) {
                    sleep 30;
                    #Performs commit/reject if the aggregate is waiting
                    #for the same after IOC
                    _wafliron_commit_reject_changes($fconn, $root_aggr, $commit);
                }

                # re-connect
                transit_to( $fconn, 'UP' ) or die("failed to re-connect: $@");

                # Bring all Aggrs Online
                foreach my $aggr (@aggregates) {
                    aggr_online($fconn, $aggr) || die("Cannot bring aggr '$aggr' online: $@");
                } # END of FOR

                if($ret eq WACK_DIRTY) {
                    logcomment("INFO : Found corruption Root aggr [$root_aggr]".
                               " - Returning '$ret'");
                }
                return $ret;

            }    # End of if (defined($root_aggr))

        } elsif (($wafl_version <= WV_BOILERMAKER_MAX) || ($opts{force} == 1)) {

            logcomment(
                    "\n\n\n=================================================="
                  . "\n\t\tWARNING & IMPORTANT NOTE"
                  . "\n=================================================="
                  . "\n Filer Running in '$filer_mode' mode & 'PRE-RR' Release"
                  . "\n WACK is supported in this S/W release - Running WACK"
                  . "\n==================================================\n\n");

            # Regular Flow for Ontap Version <= BR
            if ($filer_mode == 10) {
                # Transiting to IBM (interactive boot menu) prompt,via  HALT,CFE prompts.
                # direct transition to IBM does'nt work properly.
                eval {
                    $Server->transit_to(state => $Server->HALT());
                }; # END of Eval

                die ("Unable to HALT the filer")
                    if ($@);

                $Server->transit_to(state => $Server->IBM(),
                                          no_connections => 1,
                                          state_timeout  => 300);
            } else {
                transit_to($conn,
                           BOOT_MENU,
                           map { exists($opts{$_}) ? ($_ => $opts{$_}) : () } (
                                                      'prefix',    'powercycle',
                                                      'bootimage', 'maxcycle',
                                                      'haltcmd')) or die $@;
            }    #END of if ($filer_mode == 10)

            say($conn, "WAFL_check $args", '-blind', '-eval_error');
            die $@ if $@;

            while (1) {
                my ($match, $before) =
                  say($conn, "", '-nocr',
                      '-noprompt',
                      '-Timeout',
                      $endtime, '-wait',
                      [RE_CONTINUE,    RE_COMMIT,
                       RE_CHECK,       RE_DELETE,
                       RE_DONE,        RE_BOOTING,
                       RE_BOOTING2,    RE_CONSOLE_LOGGED,
                       RE_INFO_NEEDED, RE_SELECTION,
                       RE_NGS],
                      '-maxidle',
                      90,
                      '-eval_error');
                if ($@) {
                    my $last_error = $@;
                    logcomment($last_error);
                    if ($last_error =~ /Err=IDLE/) {
                        # Send a carriage return to see if we missed a message
                        say($conn, "\r", '-nocr', '-blind');
                        next;
                    } else {
                        die $last_error;
                    }    #END of if ($last_error
                }    #END of if ($@)

                if ($outputref) {
                    $$outputref .= $before . $match;
                } #END of if ($outputref)

                if ($match =~ RE_CONTINUE || $match =~ RE_NGS) {
                    say($conn, "yes", '-blind', '-eval_error');
                    die $@ if $@;
                } elsif ($match =~ RE_CHECK) {

                    # if args contains '-all' this prompt will not be encountered
                    # will reach here if all vols specified in args are not found
                    $ret = WACK_NOTDONE;
                    say($conn, 'no', '-blind', '-eval_error');
                    die $@ if $@;
                } elsif ($match =~ RE_COMMIT || $match =~ RE_DELETE) {
                    $ret = WACK_DIRTY;
                    say($conn, $opts{commit} ? "yes" : "no",
                        '-blind', '-eval_error');
                    die $@ if $@;
                } elsif (   $match =~ RE_DONE
                         || $match =~ RE_BOOTING
                         || $match =~ RE_BOOTING2) {
                    last;
                } elsif ($match =~ RE_INFO_NEEDED) {
                    $ret = WACK_INFONEEDED if $opts{infoneeded};
                    say($conn, $opts{infoneeded} ? "yes" : "no",
                        '-blind', '-eval_error');
                    die $@ if $@;
                } elsif ($match =~ RE_CONSOLE_LOGGED) {
                    $ret = WACK_CONSOLENOTLOGGED unless $opts{conslogged};
                    say($conn, $opts{conslogged} ? "yes" : "no",
                        '-blind', '-eval_error');
                    die $@ if $@;
                } elsif ($match =~ RE_SELECTION) {

                    # for prev_cp the default is to revert to previous cp inofile
                    $opts{selection} = "c" unless defined $opts{selection};
                    say($conn, "$opts{selection}", '-blind', '-eval_error');
                    die $@ if $@;
                }    #END of IF $match
            }    #END of WHILE
        }    # END of IF version
        return $ret;
    };
} ## end sub wack ( $@ )


######################### PRIVATE METHOD ########################

#=================== _wafliron_commit_reject_changes ==========================
# Check the WaflIron Status & Commit / Rehect changes
#==============================================================================
sub _wafliron_commit_reject_changes {
    my ($conn, $aggr, $commit) = @_;
    my $wafliron_status;
    logcomment(  "INFO : Checking whether aggr '$aggr' is waiting "
               . "for commit or reject the changes");
    do {
        eval {
            $wafliron_status = wafliron_status($conn, $aggr);
        };

        die ("wack.pm - Filer PANICed while checking Wafliron Status $@")
            if ($@ =~ m/PANIC/);

        if ($wafliron_status eq "") {
            logcomment("INFO : Unable to determine the IOC status. Could be ".
                       "due to heavy change log.Retry after 180 Sec");
            sleep 180;
        }
    } while ($wafliron_status eq "");

    if ($wafliron_status == 2) {
        # Handle the Commit Flag
        if ($commit == 1) {
            logcomment("INFO : Commit the Changes");
            wafliron_commit($conn, $aggr)
              or die
              "IOC in Commit mode Failed - Unable to commit the changes : $@";
        } elsif ($commit == 0) {
            logcomment("INFO : Reject the Changes");
            die("Unable to perform wafliron reject on '$aggr'")
              if (wafliron_reject($conn, $aggr) !~
                m/Rejecting change log files for aggregate|Rejected changes for aggregate/
              );
        } # END of if ($commit == 1)
    } else {
        logcomment("INFO : No Wafliron Instance is currently Running - ".
                   "Possibly no Corruption found");
    }    # END of if (wafliron_status($conn, $aggr) == 2)

    logcomment("INFO : Bring the aggregate online ");
    eval {
        aggr_online($conn, $aggr);
    };
    if ($@) {
        my $croak = 1;
        if ($@ =~ /Unrecognized format for aggr status/i) {
            my $status_obj = aggr_status( $conn, $aggr );
            $croak = 0 if $status_obj->state() eq "online";
        }
        die("Cannot bring aggr online: $@") if $croak;
    }

} # END of sub _wafliron_commit_reject_changes


#============================== _log_start ====================================
# Start to monitor the messages logged by current Wafliron session.
#==============================================================================
sub _log_start {
    my $server = shift;
    my $logrdr = LogReader::MlogMessages->new($server);
    logcomment("INFO : Starting Logger");
    $logrdr->start;
    sleep 2;
    return $logrdr;
} # END of sub _log_start

#============================== _log_stop ====================================
# Stop  monitoring  the messages logged by current Wafliron session and return
# the set of wafl messages logged between start and stop.                     
#=============================================================================
sub _log_stop {
    my $handle = shift;
    my ($lines, $report, $logrdr);
    $logrdr = $handle;
    sleep 5;

    logcomment("INFO : Stoping Logger");        
    $logrdr->stop;
    $lines = $logrdr->matches('.');
    if (@$lines) {
        $report = join "\n", @$lines;
    }
    return $report;
} # End of sub _log_stop

1;
