# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/ontap/main/test/wafl/wafl_utils.pm#9 $
# 
# Copyright (c) 2003 Network Appliance, Inc.
# All rights reserved.
#
# {wafl/wafl_utils.pm}

## @summary Functions used in wafl test scripts
##  
## @description
##   Common functions used among wafl test scripts. Some of the APIs will be
##   eventually moved to the lib directory. They are here temporarily.
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
## @status public
# Submitted via burt98550
package wafl_utils;

use strict;
use vars qw(@ISA @EXPORT);
use Tharn;
require Exporter;

use Data::Dumper;

@ISA = qw(Exporter);
@EXPORT = qw(
             df snap_schedule snap_list snap_status
             poll_snap_status snap_create wafl_showvbns
             );


# ------------------ API for WAFL QA internal use --------------------------

## @name ::poll_snap_status
## @description Returns the status of a snapshot.
##              Function returns a 1 when the snapshot specified has attained
##              the required status of 'completed', 'deleting', or 'creating'
##              as desired by the caller.
##
## @synopsis
##   use wafl::wafl_utils;
##   poll_snap_status($filerspec, $filer_volume, $snapshot_name,
##                    status=>..., [timeout=>...]);
##
##   Timeout defaults to 300 secs.
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
##
sub poll_snap_status($$$%) {
    my ($fc, $vol, $snapname, %args) = @_;

    return eval {
    if (!defined($args{status}) || !$args{status}) {
        die("You need to specify a status to check for the snapshot");
    }

    # timeout defaults to 300 secs
    $args{timeout} = defined($args{timeout}) ? $args{timeout} : 300;

    my $sshot_found = 0;
    my $curr_time = time();
    my $fut_time  = $curr_time + $args{timeout};

    while ($curr_time < $fut_time) {
        my @status = snap_status($fc, $vol);
    
        # look for the snapshot's name and return its status
        foreach my $snap_status (@status) {
            if ($snap_status->{name} eq "$snapname") {
                if ($snap_status->{status} eq $args{status}) {
                    return 1;
                }
                $sshot_found = 1;
            }
        }

        if (!$sshot_found) {
            # wail, snapshot is not found in the snap status list
            die("Snapshot \'$snapname\' cannot be found!");
        }

        sleep 5;
        $curr_time = time();
        $sshot_found = 0;
    }

    # timed out
    die("Timed out. Aborting determination of the status of the snapshot");
    };
}


# ------------- Potential APIs to export to lib directory ------------------

## @name ::df
## @description Returns the complete df status of a particular volume. This
##              includes the .snaphost directory of the volume.
## @synopsis
##
##   use wafl::wafl_utils;
##
##   # General
##   my %diskfree = df($filerspec, $filer_vol, [options=>...]);
##
##   # Same as 'df $FILER_VOL'
##   my %diskfree = df($fc, $FILER_VOL);
##
##   # Same as 'df -i $FILER_VOL'
##   my %diskfree = df($fc, $FILER_VOL, options=>'-i');
##
##   # Same as 'df -(h|k|m|g|t) $FILER_VOL'
##   my %diskfree = df($fc, $FILER_VOL, options=>'-(h|k|m|g|t)');
##
## @usage
##   This section describes the data structure of the hash returned using
##   different options.
##
##   <pre>my %diskfree = df($fc, $FILER_VOL);</pre>
##
##   General data struture: $diskfree{type}->{field} where<br>
##      (1) type can be either 'volume' or 'snapshot'<br>
##      (2) field can be 'filesystem', 'total', 'used', 'avail',
##          'capacity', or 'mount'.<br>
##
##   <pre>
##      Filesystem          kbytes  used  avail capacity Mounted on
##      /vol/vol0/          501992 71682 494031       1% /vol/vol0/
##      /vol/vol0/.snapshot 122998 10673 124232       1% /vol/vol0/.snapshot
##   </pre>
##
##   Referring to the above 'df vol0' output, to get the disk usage on vol0,
##   you do the following.
##
##   <pre>
##   my %diskfree = df($fc, $FILER_VOL);
##   my $vol0_usage = $diskfree{volume}->{used};
##   </pre>
##
##   To get the total space availability on the .snapshot, you do the 
##   following.
##
##   <pre>my $snapshot_avail = $diskfree{snapshot}->{avail};</pre>
##
##   To print out the whole data structure, the following code can be used.
##
##   <pre>
##   foreach my $type (keys %diskfree) {
##   &nbsp;# note the stuff in parenthesis. that's how you extract the keys
##   &nbsp;# in the nested hash
##   &nbsp;&nbsp;foreach my $field ( keys %{$diskfree{$type}} ) {
##   &nbsp;&nbsp;&nbsp;print "For $type, $field: $diskfree{$type}->{$field} kbytes\n";
##   &nbsp;&nbsp;}
##   }
##   </pre>
##
##   <hr>
##
##   <pre>my %diskfree = df($fc, $FILER_VOL, options=>'-i');</pre>
##
##   General data struture: $diskfree{type}->{field} where<br>
##      (1) type can be 'volume' *only*<br>
##      (2) field can be 'filesystem', 'iused', 'ifree', '%iused',
##          'mount'.<br>
##
##   The usage to get the data from the hash is similar to the above. Note,
##   for field '%iused', make use you put the word in quotes otherwise Perl
##   will complain, e.g.
##
##   # remember to put '%iused' in quotes
##   <pre>my percent_used_inodes = $diskfree{volume}->{'%iused'};</pre>
##
##   <hr>
##
##   <pre>my %diskfree = df($fc, $FILER_VOL, options=>'-(h|k|m|g|t)');</pre>
##
##   General data struture: $diskfree{type}->{field} where<br>
##      (1) type can be ether 'volume' or 'snapshot'<br>
##      (2) field can be 'filesystem', 'total', 'used', 'avail',
##          'capacity', or 'mount'.<br>
##
##   Look above for usage. The different options for 'df' just return the
##   the values in kilobytes, megabytes, gigabytes, or terabytes.
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
##
sub df($$%) {
    my ($fc, $vol, %args) = @_;

    return eval {
    if (length($vol) == 0) {
        die("df(): Please specify a volume name");
    }

    # the trailing slash '/' is required for the grep to work properly below
    my $path = "/vol/$vol/";

    $args{options} = defined($args{options}) ? $args{options} : "";

    my (undef, undef, $df_output) = say($fc, "df $args{options} $path", -timeout, 120);

    if (!defined($args{options}) || 
        length($args{options}) == 0) {

        my $vol_status = {
            "filesystem" => undef,
            "total"      => undef,
            "used"       => undef,
            "avail"      => undef,
            "capacity"   => undef,
            "mount"      => undef,
        };

        my $sshot_status = {
            "filesystem" => undef,
            "total"      => undef,
            "used"       => undef,
            "avail"      => undef,
            "capacity"   => undef,
            "mount"      => undef,
        };

        my %df = (
            volume   => $vol_status,
            snapshot => $sshot_status,
        );

        # Filesystem          kbytes  used  avail capacity Mounted on
        # /vol/vol0/          501992 71682 494031       1% /vol/vol0/
        # /vol/vol0/.snapshot 122998 10673 124232       1% /vol/vol0/.snapshot
    
        foreach my $line (split(/\n/, $df_output)) {
            chomp($line);

            # added a space for grep so that it's less greedy
            if ($line =~ /$path\s+/) {

                if ($line =~ /^\s*(.*)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)%\s+(.*)/) {
                    $df{volume}->{filesystem} = $1;
                    $df{volume}->{total}      = $2;
                    $df{volume}->{used}       = $3;
                    $df{volume}->{avail}      = $4;
                    $df{volume}->{capacity}   = "$5%";
                    $df{volume}->{mount}      = $6;

                    # strip out \r or \n or \s since we did a greedy grep
                    $df{volume}->{filesystem} =~ s/\s//g;
                    $df{volume}->{mount}      =~ s/\s//g;
                }

            } elsif ($line =~ /$path\.snapshot\s+/) {

                if ($line =~ /^\s*(.*)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)%\s+(.*)/) {
                    $df{snapshot}->{filesystem} = $1;
                    $df{snapshot}->{total}      = $2;
                    $df{snapshot}->{used}       = $3;
                    $df{snapshot}->{avail}      = $4;
                    $df{snapshot}->{capacity}   = "$5%";
                    $df{snapshot}->{mount}      = $6;

                    $df{snapshot}->{filesystem} =~ s/\s//g;
                    $df{snapshot}->{mount}      =~ s/\s//g;
                }
            }
        }
        return %df;

    } elsif ($args{options} eq "-i") {

        my $vol_status = {
            "filesystem" => undef,
            "iused"      => undef,
            "ifree"      => undef,
            "%iused"     => undef,
            "mount"      => undef,
        };

        my %df_i = (
            volume => $vol_status,
        );

        # Filesystem               iused      ifree  %iused  Mounted on
        # /vol/testvol/              437    2066144     0%   /vol/testvol/
    
        foreach my $line (split(/\n/, $df_output)) {
            chomp($line);

            
            if ($line =~ /^\s*(.*)\s+(\d+)\s+(\d+)\s+(\d+)%\s+(.*)/) {
                $df_i{volume}->{filesystem} = $1;
                $df_i{volume}->{iused}      = $2;
                $df_i{volume}->{ifree}      = $3;
                $df_i{volume}->{"%iused"}   = "$4%";
                $df_i{volume}->{mount}      = $5;
            }                            
        }
        return %df_i;

    } elsif ( ($args{options} eq "-h") ||
              ($args{options} eq "-k") ||
              ($args{options} eq "-m") ||
              ($args{options} eq "-g") ||
              ($args{options} eq "-t") ) {

        my $vol_status = {
            "filesystem" => undef,
            "total"      => undef,
            "used"       => undef,
            "avail"      => undef,
            "capacity"   => undef,
            "mount"      => undef,
        };

        my $sshot_status = {
            "filesystem" => undef,
            "total"      => undef,
            "used"       => undef,
            "avail"      => undef,
            "capacity"   => undef,
            "mount"      => undef,
        };

        my %df_size = (
            volume  => $vol_status,
            snaphot => $sshot_status,
        );

        # u49> df -t /vol/vol0
        # Filesystem          total  used   avail capacity  Mounted on
        # /vol/vol0/          0TB     0TB     0TB       1%  /vol/vol0/
        # /vol/vol0/.snapshot 0TB     0TB     0TB       0%  /vol/vol0/.snapshot
    
        foreach my $line (split(/\n/, $df_output)) {
            chomp($line);

            # added a space for grep so that it's less greedy
            if ($line =~ /$path\s+/) {

                if ($line =~ /^\s*(.*)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\d+)%\s+(.*)/) {
                    $df_size{volume}->{filesystem} = $1;
                    $df_size{volume}->{total}      = $2;
                    $df_size{volume}->{used}       = $3;
                    $df_size{volume}->{avail}      = $4;
                    $df_size{volume}->{capacity}   = "$5%";
                    $df_size{volume}->{mount}      = $6;

                    # strip out \r or \n or \s since we did a greedy grep
                    $df_size{volume}->{filesystem} =~ s/\s//g;
                    $df_size{volume}->{mount}      =~ s/\s//g;
                }

            } elsif ($line =~ /$path\.snapshot\s+/) {

                if ($line =~ /^\s*(.*)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\d+)%\s+(.*)/) {
                    $df_size{snapshot}->{filesystem} = $1;
                    $df_size{snapshot}->{total}      = $2;
                    $df_size{snapshot}->{used}       = $3;
                    $df_size{snapshot}->{avail}      = $4;
                    $df_size{snapshot}->{capacity}   = "$5%";
                    $df_size{snapshot}->{mount}      = $6;

                    # strip out \r or \n or \s since we did a greedy grep
                    $df_size{snapshot}->{filesystem} =~ s/\s//g;
                    $df_size{snapshot}->{mount}      =~ s/\s//g;
                }
            }
        }
        return %df_size;

    } else {
        # for error testing
        return $df_output;
    }
    };
}


## @name ::snap_schedule
## @description Returns the snap schedule or set the snap schedule of a
##              particular volume.
## @synopsis
##   use wafl::wafl_utils;
##   my $sched = snap_schedule($fc, $FILER_VOL);
##   snap_schedule($fc, $FILER_VOL, "0 2 6@8,12,16,20");
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
##
sub snap_schedule($$;$) {
    my ($fc, $vol, $sched) = @_;

    return eval {
    if (defined($sched)) {
        # set the schedule
        my (undef, undef, $sched_output) = say($fc, "snap sched $vol $sched");

        foreach my $line (split(/\n/, $sched_output)) {
            if ($line =~ /^\s*snap\s+sched:\s+invalid\s+schedule/) {
                die("Invalide Schedule: snap sched $vol $sched");
            }
            last;
        }

    } else {
        # return the schedule
        my (undef, undef, $sched_output) = say($fc, "snap sched $vol");

        if ($sched_output =~ /^\s*Volume\s+$vol:\s+(.*)/) {
            $sched = $1;
            
            # strip out carriage returns and newlines
            $sched =~ s/\r|\n//g;
            return $sched;
        } else {
            die("Error parsing snap sched output: $sched_output");
        }
    }
    };
}


## @name ::snap_list
## @description Returns a list of snapshots in an array and other information
##              pertaining to each snapshot in that list for particular volume.
## @synopsis
##   Only the following two options have been implemented. Please add to this
##   API if you need other options that are not found here.
##
##   use wafl::wafl_utils;
##
##   # General
##   my @list = snap_list($filerspec, $filer_volume, [options=>...]);
##
##   # same as 'snap list $FILER_VOL'
##   my @list = snap_list($fc, $FILER_VOL);
##
##   # same as 'snap list -n $FILER_VOL'
##   my @list = snap_list($fc, $FILER_VOL, options=>'-n');
##
## @usage
##   This section describes the data structure of the array returned using
##   different options.
##
##   <pre>my @list = snap_list($fc, $FILER_VOL);</pre>
##
##   General data struture: $list[x]{field} where<br>
##      (1) X is any index to the array and<br>
##      (2) field can be 'used', 'total', 'date', 'name'<br>
##
##   To get the name of the filer for the first snapshot that snap_list()
##   captured.
##
##   <pre>my $snapshot_name = $list[0]{name};</pre>
##
##   Similarly, to get the date/time stamp of the third snapshot that
##   snap_list() captured.
##
##   <pre>my $snapshot_date = $list[2]{date};</pre>
##
##   To search through the array to find a particular snapshot name, you
##   can do the following
##
##   <pre>
##   for (my $i = 0; $i < @$list; ++$i) {
##   &nbsp;if ($list[$i]{name} eq "snapshot_name") {
##   &nbsp;&nbsp;## do something here
##   &nbsp;}
##   }
##   </pre>
##
##   <hr>
##
##   <pre>my @list = snap_list($fc, $FILER_VOL, options=>'-n');</pre>
##
##   General data struture: $list[X]{field} where<br>
##      (1) X is any index to the array and<br>
##      (2) field can be 'date', 'name'<br>
##
##   The '-n' option shows the data/time stamp and name of the snapshot
##   hence there are only two fields.
##
##   To access these two fields in the array, please see above for the
##   usage.
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
##
sub snap_list($$%) {
    my ($fc, $vol, %args) = @_;

    return eval {
    if (length($vol) == 0) {
        die("snap_list(); Please specify a volume name");
    }

    $args{options} = defined($args{options}) ? $args{options} : "";
    my (undef, undef, $snap_output) = say($fc, "snap list $args{options} $vol", -timeout, 120);

    # no snapshots found
    if ($snap_output =~ /^\s*No\s+snapshots\s+exist\./) {
        # return empty array
        return my @snap_list;
    }
        
    if (!defined($args{options}) ||
        length($args{options}) == 0) {

        my @snap_list = (
           {
            "used"  => undef,
            "total" => undef,
            "date"  => undef,
            "name"  => undef,
           },
        );

        # u49> snap list vol0   
        # Volume vol0
        #  working...
        # 
        # %/used      %/total     date          name
        # ----------  ----------  ------------  --------
        # 0% ( 0%)    0% ( 0%)    Jun 05 16:00  hourly.0       
        # 1% ( 1%)    0% ( 0%)    Jun 05 12:00  hourly.1        

        my $i = 0;
        foreach my $line (split(/\n/, $snap_output)) {
            chomp($line);

            print "$line\n";
            if ($line =~ /^\s*Volume\s+$vol/) {
                next;
            } elsif ($line =~ /^\s*working\.\.\./) {
                next;
            } elsif ((length($line) == 0) || ($line =~ /^\r|^\n/)) {
                next;
            } elsif ($line =~ /^\s*%\/used\s+%\/total\s+/) {
                next;
            } elsif ($line =~ /^\s*----/) {
                next;
            } elsif ($line =~ /^\s*No snapshots exist\./) {
                last;
            } else {
                if ($line =~ /^\s*(\d+)%\s+\(\s*(\d+)%\)\s+(\d+)%\s+\(\s*(\d+)%\)\s+(\w+)\s+(\d+)\s+(\d+)\:(\d+)\s+(.*)/) {

                    $snap_list[$i]{used}  = "$1% ($2%)";
                    $snap_list[$i]{total} = "$3% ($4%)";
                    $snap_list[$i]{date}  = "$5 $6 $7:$8";
                    $snap_list[$i]{name}  = "$9";

                    # remove [\s] in snapshot's name
                    $snap_list[$i]{name} =~ s/\s//g;
                    $i++;
                }
            }
        }
        return @snap_list;

    } elsif ($args{options} eq "-n") {

        my @snap_list = (
           {
            "date"  => undef,
            "name"  => undef,
           }
        );

        # u49> snap list -n vol0
        # Volume vol0
        # working...
        #
        # date          name
        # ------------  --------
        # Jun 05 16:00  hourly.0       
        # Jun 05 12:00  hourly.1  

        my $i = 0;
        foreach my $line (split(/\n/, $snap_output)) {
            chomp($line);

            print "$line\n";
            if ($line =~ /^\s*Volume\s+$vol/) {
                next;
            } elsif ($line =~ /^\s*working\.\.\./) {
                next;
            } elsif ((length($line) == 0) || ($line =~ /^\r|^\n/)) {
                next;
            } elsif ($line =~ /^\s*date\s+name\s*/) {
                next;
            } elsif ($line =~ /^\s*----/) {
                next;
            } else {
                if ($line =~ /^\s*(\w+)\s+(\d+)\s+(\d+)\:(\d+)\s+(.*)/) {
                    $snap_list[$i]{date} = "$1 $2 $3\:$4";
                    $snap_list[$i]{name} = "$5";

                    # remove [\s] in snapshot's name
                    $snap_list[$i]{name} =~ s/\s//g;
                    $i++;
                }
            }
        }
        return @snap_list;

    } else {
        die("snap_list(): Option \'$args{options}\' not implemented yet");
    }
    };
}

## @name ::snap_create
## @usage snap_create($fc, $vol, $snap, %args)
## @synopsis
##    Creates a snapshot of specified name in specified volume/aggregate
##
sub snap_create($$$%) {
    my ($fc, $vol, $snap, %args) = @_;
    my ($options) = defined($args{options}) ? $args{options} : "";

    return eval {
      my (undef, undef, $result) = say($fc, "snap create $options $vol $snap");
      die "$result\n" unless ($result =~ s/creating snapshot\.\.\.//);
      return 1;
    }
}

## @name ::snap_status
## @description Returns the status of all the snapshots in a particular volume.
##              The 'snap status' command is an advance hidden command.
##              This command should be run after any creation of any snapshot
##              as it will ensure that the snapshot metadata has been updated
##              properly before any I/O operations is performed on that
##              snapshot.
##
## @synopsis
##   use wafl::wafl_utils;
##   my @status = snap_status($filerspec, $filer_volume);
##
## @usage
##   This section describes the data structure of the array returned using
##   different options.
##
##   <pre>my @status = snap_status($fc, $FILER_VOL);</pre>
##
##   General data struture: $status[X]{field} where<br>
##      (1) X is any index to the array and<br>
##      (2) field can be 'snapid', 'status', 'date', 'ownblks', 'version', 'fsrev', 'name'<br>
##
##   To get the status of the filer for the first snapshot that snap_status()
##   captured.
##
##   <pre>my $snapshot_status = $status[0]{name};</pre>
##
##   Similarly, to get the date/time stamp of the third snapshot that
##   snap_status() captured.
##
##   <pre>my $snapshot_date = $status[2]{date};</pre>
##
##   To search through the array to find a particular snapshot's status, you
##   can do the following
##
##   <pre>
##   for (my $i = 0; $i < @$list; ++$i) {
##   &nbsp;if ($status[$i]{status} eq "complete") {
##   &nbsp;&nbsp;## do something here
##   &nbsp;}
##   }
##   </pre>
##
## @author Thanapong Lertpanyavit (tapl@netapp.com)
##
sub snap_status($$) {
    my ($fc, $vol) = @_;

    return eval {
    my (undef, undef, $output) = say($fc, "priv set -q advanced; snap status $vol");

    if ($output =~ /Volume\s+$vol\s+does not exist or is not online/) {
        die("$output");
    }

    my @snap_status = (
       {
         "snapid"  => undef,
         "status"  => undef,
         "date"    => undef,
         "ownblks" => undef,
         "release" => undef,
         "fsrev"   => undef,
         "name"    => undef,
       },
    );

    # separate each entry into each line
    #
    # u49*> snap status testvol
    # Volume testvol
    # snapid  status     date           ownblks release fsRev name
    #------  ------     ------------   ------- ------- ----- --------
    #  5     complete   Mar 08 16:00       470     7.0    51 hourly.0      
    #  3     complete   Mar 08 12:00        24     7.0    51 hourly.1      
    #  7     deleting   Mar 08 18:05         0   0.5          (46/447 remaining)

    my $i = 0;
    foreach my $line (split(/\n|\rname/, $output)) {
        if ($line =~ /^\s*Volume\s+$vol/) {
            next;
                
        } elsif ($line =~ /^snapid/) {
            next;
                
        } elsif ($line =~ /---/) {
            next;
            
        } elsif ((length($line) == 0) || ($line =~ /^\r|^\n/)) { # blank lines
            next;        

        } else {
            chomp($line);

            if ($line =~ /^\s*(\d+)\s+(\w+)\s+(\w+)\s+(\d+)\s+(\d+)\:(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(.*)/) {

                $snap_status[$i]{snapid}  = $1;
                $snap_status[$i]{status}  = $2;
                $snap_status[$i]{date}    = "$3 $4 $5\:$6";
                $snap_status[$i]{ownblks} = $7;
                $snap_status[$i]{release}    = $8;
                $snap_status[$i]{fsrev}    = $9;
                $snap_status[$i]{name}    = $10;

                # If $snap_status[$i]{name} looks like this...
                #
                #   base_snapshot  (510/534 remaining)
                #
                # We need to do some more filtering just to grab the
                # snapshot's name out. We can discard the status.
                if ($snap_status[$i]{name} =~ /(.*)\s+?\(/) {
                    $snap_status[$i]{name} = $1;
                }

                # remove [\s] in snapshot's name
                $snap_status[$i]{name} =~ s/\s//g;
                $i++;
            }
        }
    }
    return @snap_status;
    };
}

#
# This function returns the vbn and vvbn of a given fbn of a file
# This function works only with the syntax
#   wafl showvbns /vol/vol0/file1 0
#

sub wafl_showvbns($$$) {
    
    my( $fc, $file, $fbn) = @_;
    my( undef,undef,$op) = say($fc, "priv set test;wafl showvbns $file $fbn");
    #fbn 1 is vbn 69846311, vvbn 14408
    if( $op =~ m/fbn\s+\d+\s+is\s+vbn\s+(\d+),\s+vvbn\s+(\d+)/) {
        return($1, $2);    
    } else { 
        return(0);
    }
}
   
1;
