# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/platqa/user_interfaces/shell_commands/SysCmdDefs.pm#2 $
#
## @summary	SysCmdDefs.pm - Defines capabilities matrix for platform
## 				system commands tests.
## @author	dl-platform-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

SysCmdDefs - Defines capabilities matrix for platform system commands tests.

=head1 SYNOPSIS

  use platqa::user_interfaces::shell_commands::SysCmdDefs;

  # Get privilege hash for selected command:
  %privs = get_privileges(command=>$cmd, mode=>$mode);

=head1 DESCRIPTION

Provides a capabilities matrix for platform system commands tests for both
7 mode and 10 mode. The library only specifies if command is available in that
privilege level. It does not account for differing levels of ability depending
on the privilege level.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/platqa/user_interfaces/shell_commands/SysCmdDefs.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

  Copyright 2008 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package platqa::user_interfaces::shell_commands::SysCmdDefs;

use strict;
use warnings;
use Carp qw(croak);

use Params::Validate qw(validate SCALAR);

use base qw(Exporter);

our @EXPORT_OK = qw(
    get_privileges
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub get_privileges(%);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my %command_matrix;

#---------------------------------------------------------------------

=head2 get_privileges

  %privs = get_privileges(command => $cmd, mode => $mode);

This routine gets the  privilege hash for selected command.

The following named arguments are accepted:

=over

=item Arguments

=over

=item command

(Required) : Command under test.

=item mode

(Required) : Filer mode.

=back

=item Returns

Returns the desired hash element if exists otherwise returns undef with
an error in $@.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_privileges(%) {
    my (%opts) = validate (
        @_,
        {
            command => { type => SCALAR },
            mode    => { type => SCALAR },
        }
    );

    my $mode = $opts{mode};
    my $command = $opts{command};

    # Make sure the command_matrix has entries for the specified mode
    if (!defined $command_matrix{$mode}) {
        croak "Mode '$mode' not defined";
    }

    # Make sure the command_matrix has an entry for the specified command
    if (!defined $command_matrix{$mode}{$command}) {
        croak "Command '$command' for mode '$mode' not defined";
    }

    return $command_matrix{$mode}{$command};
}

#---------------------------------------------------------------------
# Hash of 7 mode and 10 mode commands. All privilege levels are included
# so that the command availability can be tested on each level.
#---------------------------------------------------------------------
%command_matrix = (

# ---------------------------------------------------------------------
# 7 mode commands
# ---------------------------------------------------------------------

    '7'  => {
        'agent_rw' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'availtime' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'bmc' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'bootargs' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'bootfs' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'com' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'date' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'download' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'exit' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'fru_led' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'hammer' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'i2c_stress' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'in' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'incfg' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'inb' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'inl' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'inw' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'java' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'log' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'logger' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'mdumpb' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'mdumpl' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'mdumpw' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'memdump' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'memerr' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'memio' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'mem_scrub_force' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'mem_scrub_stats' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'mem_stats' => {
            admin    => 0,
            advanced => 0,
            diag     => 1,
            test     => 1,
        },
        'options' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'out' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'outcfg' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'outb' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'outl' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'outw' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'pcicfgdump' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'perf' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'platform_error' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'printflag' => {
            admin    => 0,
            advanced => 0,
            diag     => 1,
            test     => 1,
        },
        'priv' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'ps' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'rdate' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'reg_dump' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'result' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'rlm' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'rmc' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'rtag' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'setflag' => {
            admin    => 0,
            advanced => 0,
            diag     => 1,
            test     => 1,
        },
        'setral' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'show_bios_log' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'smbus' => {
            admin    => 0,
            advanced => 0,
            diag     => 0,
            test     => 1,
        },
        'software' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'source' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'statit' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'sysconfig' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'syslog' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'test_lcd' => {
            admin    => 0,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'timezone' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'ups' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'uptime' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'version' => {
            admin    => 1,
            advanced => 1,
            diag     => 1,
            test     => 1,
        },
        'watchdog' => {
            admin    => 0,
            advanced => 0,
            diag     => 1,
            test     => 1,
        },
    },

# ---------------------------------------------------------------------
# 10 mode commands
# ---------------------------------------------------------------------

    '10' => {
        'exit' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'set'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'statistics periodic' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'statistics show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system environment' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system environment agent sensors show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system environment set-state' => {
            admin      => 0,
            advanced   => 0,
            diagnostic => 1,
            test       => 1,
        },
        'system firmware download' => {
            admin      => 0,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system hardware fru show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system hardware lcd modify' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system hardware lcd show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node date modify' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node date show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment agent sensors show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment alarms show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment fan show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment fan status show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment led show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment policies show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment power status show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment set-state' => {
            admin      => 0,
            advanced   => 0,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment thermometer show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node environment voltage show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node firmware download' => {
            admin      => 0,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node hardware fru show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node hardware lcd modify' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node hardware lcd show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node modify'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
        'system node show'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
            test       => 1,
        },
    },

# ---------------------------------------------------------------------
# Tricky mode commands
# ---------------------------------------------------------------------

    'Tricky' => {
        'exit' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'set'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'statistics periodic' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'statistics show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system date modify' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system date show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment agent sensors show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment alarms show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment set-state' => {
            admin      => 0,
            advanced   => 0,
            diagnostic => 1,
        },
        'system firmware download' => {
            admin      => 0,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment fan show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment fan status show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment led show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment policies show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment power status show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment set-state' => {
            admin      => 0,
            advanced   => 0,
            diagnostic => 1,
        },
        'system environment thermometer show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system environment voltage show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system firmware download' => {
            admin      => 0,
            advanced   => 1,
            diagnostic => 1,
        },
        'system hardware fru show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system hardware lcd modify' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system hardware lcd show' => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system modify'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
        'system show'      => {
            admin      => 1,
            advanced   => 1,
            diagnostic => 1,
        },
    },
);

# ---------------------------------------------------------------------
# Map other 7 mode names to the 7 mode table
# ---------------------------------------------------------------------

$command_matrix{'7g'} = $command_matrix{7};

# ---------------------------------------------------------------------
# Map other 10 mode names to the 10 mode table
# ---------------------------------------------------------------------

$command_matrix{gx} = $command_matrix{10};
$command_matrix{ng} = $command_matrix{10};
$command_matrix{Tricky} = $command_matrix{Tricky};

1;
