# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/platqa/user_interfaces/shell_commands/SysCmdUtils.pm#2 $
#
## @summary	SysCmdUtils.pm - Utility with routines for
##                               platform system command tests
## @author	dl-platform-qa@netapp.com
## @status	private
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

SysCmdUtils - Utility with routines for platform system command tests

=head1 SYNOPSIS

  use platqa::user_interfaces::shell_commands::SysCmdUtils;

  # Get the type of filer we are working with.
  $mode = filer_mode(server=>$server);

  # Converts a 10 mode command into the Autoload format for Server.pm objects.
  $command = convert_10mode_command(command=>$cmd);

  # Sets privilege level on filer to specified value
  set_priv_level(server=>$server, mode=>$mode, privilege=>$priv);

  # This is wrapper for SysCmdDefs::get_privileges.
  # It returns the desired hash reference element if exists.
  %privs= get_command_privs(command=>$cmd, mode=>$mode);

  # Test the command at privilege levels.
  $result = test_command_privs(server=>$server, command=>$cmd, mode=>$mode,
                               privs=>\%privs);

  # Gets the filer's model information
  $model = get_model_info(hostspec=>$hostspec);

=head1 DESCRIPTION

Provides routines for the platform system commands being tested. These are
for filers that are 10 mode or 7 mode.

=head1 STATUS

  Private

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/platqa/user_interfaces/shell_commands/SysCmdUtils.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

  Copyright 2008 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package platqa::user_interfaces::shell_commands::SysCmdUtils;

use strict;
use warnings;
use Carp qw(croak);
use Tharn;
use Sysconfig;

use Params::Validate qw(validate OBJECT SCALAR HASHREF);
use platqa::user_interfaces::shell_commands::SysCmdDefs qw(get_privileges);

use base qw(Exporter);

our @EXPORT_OK = qw(
    filer_mode
    set_priv_level
    convert_10mode_command
    get_command_privs
    test_command_privs
    get_model_info
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

BEGIN {

    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess( Tharn::script_result(@msgs) );
        }
    );
}

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub filer_mode(%);
sub convert_10mode_command(%);
sub set_priv_level(%);
sub get_command_privs(%);
sub test_command_privs(%);
sub get_model_info(%);

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
# This hash defines the modes that are acceptable input/output
# for the functions below. If a new mode is implemented, it must
# be added to this hash as a key (the value does not matter).
my %modes = (
    '7'      => 1,
    '10'     => 1,
    'Tricky' => 1,
);

# This converts the modes hash above to a regex for use in
# parameter validation
my $MODE_QR = '^(' . join( "|", keys %modes ) . ')$';

#---------------------------------------------------------------------

=head2 filer_mode

  $mode = filer_mode(server => $server);

This routine gets the type of filer we are working with.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=back

=item Returns

Returns the mode of the filer or croaks on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub filer_mode(%) {
    my (%opts) = validate( @_, { server => { type => OBJECT }, } );

    my ( $mode, $mode_list, $raw_output, $result );

    $mode = 0;
    $mode_list = join( ", ", keys %modes );

    eval {
        $result = $opts{server}->cli( cmd => '?', output => \$raw_output );
    };
    if ($@) {
        croak
            "Could not send '?' command to filer to test for filer mode: $@";
    }

    if ( $raw_output =~ /up\s+Go up one directory/i ) {
        $mode = 10;
    } elsif ( $raw_output =~ /version/ ) {
        $mode = 7;
    } else {
        croak "Cannot figure out the filer mode (known types are "
            . "$mode_list):\n$raw_output";
    }

    if ( $mode !~ /$MODE_QR/ ) {

        # If this message appears, it means that the filer_mode
        # function has been changed to allow another type of mode but
        # this mode is not defined in %modes above. The mode
        # value assigned here needs to be added above.
        croak "Filer type (e.g. 7 mode, 10 mode) specification '$mode' "
            . "not defined properly in library. Known modes are $mode_list.";
    }

    return $mode;
} ## end sub filer_mode(%)

#---------------------------------------------------------------------

=head2 convert_10mode_command

  $command = convert_10mode_command(command => $cmd);

This routine converts a 10 mode command into the Autoload format for Server.pm
objects. e.g. 'system node date show' => 'system_node_date_show'

The following named arguments are accepted:

=over

=item Arguments

=over

=item command

(Required) : Command under test.

=back

=item Returns

Returns the command in the Autoload format.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub convert_10mode_command(%) {
    my (%opts) = validate( @_, { command => { type => SCALAR }, } );

    my $command = $opts{command};

    $command =~ s/ |-/_/g;
    return $command;
} ## end sub convert_10mode_command(%;)

#---------------------------------------------------------------------

=head2 set_priv_level

  set_priv_level(server => $server, mode => $mode, privilege => $priv);

This routine sets privilege level on filer to specified value.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item mode

(Required) : Filer mode.

=item privilege

(Required) : Privilege level to set.

=back

=item Returns

Returns nothing or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub set_priv_level(%) {
    my (%opts) = validate(
        @_,
        {   server    => { type => OBJECT },
            mode      => { type => SCALAR },
            privilege => { type => SCALAR },
        }
    );

    my $server    = $opts{server};
    my $mode      = $opts{mode};
    my $privilege = $opts{privilege};

    if ( $mode eq "7" ) {
        my $cmd = "priv set -q $privilege";
        my $raw_output;

        # Set the privilege level
        eval { $server->cli( cmd => $cmd, output => \$raw_output ); };
        croak "Sending $cmd to Server failed: $@" if $@;
        if ($raw_output) {

            # The output should be empty
            croak
                "Setting privilege level to '$privilege' failed: $raw_output";
        }

    } elsif ( $mode eq "10" ) {
        my $result;

        # Set the privilege level
        eval { $server->set( privilege => $privilege ); };
        croak "Setting privilege to '$privilege' failed: $@" if $@;

        # Verify privilege level was set
        eval { $result = $server->set(); };
        croak "Setting privilege to '$privilege' failed: $@" if $@;
        if ( $result !~ /Privilege Level:\s+$privilege/i ) {
            croak "Privilege level not set to '$privilege'";
        }

    } else {
        croak "Filer mode '$mode' not implemented";
    }
} ## end sub set_priv_level(%;)

#---------------------------------------------------------------------

=head2 get_command_privs

  %privs= get_command_privs(command => $cmd, mode => $mode);

This is wrapper for SysCmdDefs::get_privileges.
It returns the desired hash reference element if exists.

The following named arguments are accepted:

=over

=item Arguments

=over

=item command

(Required) : Command under test.

=item mode

(Required) : Filer mode.

=back

=item Returns

Returns the desired hash element if exists otherwise returns undef with
an error in $@.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_command_privs(%) {
    my (%opts) = validate(
        @_,
        {   command => { type => SCALAR },
            mode    => { type => SCALAR },
        }
    );

    return get_privileges(%opts);
} ## end sub get_command_privs(%;)

#---------------------------------------------------------------------

=head2 test_command_privs

  $result = test_command_privs(server  => $server,
                               command => $cmd,
                               [, mode  => $mode ]
                               [, privs => \%privs ]);

This routine tests the command at privilege levels.

The following named arguments are accepted:

=over

=item Arguments

=over

=item server

(Required) : Name of the server object.

=item command

(Required) : Command under test.

=item mode

(Optional) : Filer mode.

=item privs

(Optional) : Hash reference of privilege levels to test.

=back

=item Returns

Returns 'SUCCESS' if command is found only in expected privilege levels,
returns 'FAILURE' and a message string if command is available where it
should not be or not available where it should be or croaks on error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub test_command_privs(%) {
    my (%opts) = validate(
        @_,
        {   server  => { type => OBJECT },
            command => { type => SCALAR },
            mode    => { type => SCALAR, regex => $MODE_QR, optional => 1 },
            privs => { type => HASHREF, optional => 1 },
        }
    );

    # Dereference for ease of use
    my $server  = $opts{server};
    my $command = $opts{command};
    my $mode    = $opts{mode};
    my $privs   = $opts{privs};
    my ( $priv, $result, $raw_output );

    if ( !$mode ) {
        $mode = filer_mode( server => $server );
    }

    if ( !$privs ) {
        eval {
            $privs = get_command_privs( command => $command, mode => $mode );
        };
        croak("Could not get priv hash for '$command' in mode '$mode': $@")
            if $@;
    }

    if ( $mode =~ /^7$/ ) {

        # 7 mode testing:
        #    if command is a single word, send 'help <command>'. Look for
        #        the command name in the output.
        #    if command is a string of words (e.g. wafl scan snaplock), find
        #        the first word and send '<word> help' (e.g. 'wafl help) then
        #        search the output for full command string.

        foreach $priv ( keys %$privs ) {
            my $send_cmd = $command;    # So we don't whack the command string
            my $expected = $privs->{$priv};    # For ease of reading

            # Prepare the command to send
            if ( $command !~ / / ) {

                # Single word. Set priv level and perpend 'help'
                $send_cmd = "priv set -q $priv; help " . $send_cmd;
            } else {

                # Multiple words. Pull out first word, set priv level and
                # append 'help'
                $send_cmd =~ s/ .*//;
                $send_cmd = "priv set -q $priv; " . $send_cmd . " help";
            }

            # Send the command
            eval {
                $result = $server->cli(
                    cmd    => $send_cmd,
                    output => \$raw_output
                );
            };
            croak "Sending $send_cmd to Server failed: $@" if $@;

            # Check the output for the expected results
            if ( $raw_output !~ /$command/ ) {
                if ($expected) {

                    # Command not found, but expected
                    return "FAILURE: expected to find '$command' in priv "
                        . "level '$priv', but did not";
                }
            } else {
                if ( !$expected && $raw_output !~ /unrecognized command/i ) {

                    # Command found, but not expected
                    return "FAILURE: found '$command' in priv level '$priv', "
                        . "but expected it to not be available";
                } elsif ( $expected
                    && $raw_output =~ /unrecognized command/i ) {

                    # Command not found, but expected
                    return "FAILURE: expected to find '$command' in priv "
                        . "level '$priv', but did not";
                }
            }
        }

    } elsif ( $mode =~ /^(10|Tricky)$/ ) {

        # 10 mode testing:
        #    send full command followed by '?'. If command is not available
        #    output should contain 'is not a recognized command'.

        foreach $priv ( keys %$privs ) {
            my $expected = $privs->{$priv};    # For ease of reading

            # Set the privilege level
            eval { $server->set( privilege => $priv ); };
            croak "Setting privilege to $priv failed: $@" if $@;

            # Verify privilege level was set
            eval { $result = $server->set(); };
            croak "Setting privilege to $priv failed: $@" if $@;
            if ( $result !~ /Privilege Level:\s+$priv/i ) {
                croak "Priv level not set correctly";
            }

            # Send the command
            eval {

                # No carriage return required for '?' help
                # But send a 'poke' to clear the command line afterward
                $result = $server->cli(
                    cmd  => "$command ?",
                    nocr => 1,
                    poke => 1
                );
            };

            # Check the output for the expected results.
            # 'poke => 1' from above may trigger an error so ignore it
            if (   $@
                && $@ =~ /not a recognized command/i
                && $@ !~ /.*ERROR:.*1/i ) {
                if ($expected) {

                    # Command not found, but expected
                    return "FAILURE: expected to find '$command' in priv "
                        . "level '$priv', but did not";
                }
            } elsif ( $@ && $@ !~ /.*ERROR:.*1/i && $@ !~ /invalid value/i ) {

                # Command failed unexpectedly, does not matter
                # if it was expected or not - croak
                croak "'$command' failed for unexpected reason: $@";
            } elsif ( !$expected ) {

                # Command found, but not expected
                return "FAILURE: found '$command' in priv level '$priv', "
                    . "but expected it to not be available";
            }
        }

    } else {
        croak "Filer mode '$mode' not implemented";
    }

    return "SUCCESS";
} ## end sub test_command_privs(%;)

#---------------------------------------------------------------------

=head2 get_model_info

  $model = get_model_info(hostspec => $hostspec);

This routine gets the filer's model information through an interface that gets
the sysconfig output of the filer.

The following named arguments are accepted:

=over

=item Arguments

=over

=item hostspec

(Required) : Hostspec of environment.

=back

=item Returns

A hash reference of the model information or croak on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_model_info(%) {
    my (%opts) = validate( @_, { hostspec => { type => SCALAR }, } );
    my ( $group, %model, $obj );

    $obj = Sysconfig->new( $opts{hostspec} );
    $model{name} = $obj->slot( '0', 'model_name' );

    # Get the platform group
    ($group) = ( $model{name} =~ /(\D+\d\d)\d*/g );
    if ( $group =~ 'FAS27' ) {
        $model{group} = 'FAS200';
    } elsif ( $group =~ 'FAS22' ) {
        $model{group} = 'FAS2200';
    } elsif ( $group =~ 'FAS20' ) {
        $model{group} = 'FAS2000';
    } elsif ( $group =~ 'FAS9' ) {
        $model{group} = 'FAS900';
    } elsif ( $group =~ 'FAS32' ) {
        $model{group} = 'FAS3200';
    } elsif ( $group =~ 'FAS31' ) {
        $model{group} = 'FAS3100';
    } elsif ( $group =~ 'FAS30' ) {
        $model{group} = 'FAS3000';
    } elsif ( $group =~ 'FAS62' ) {
        $model{group} = 'FAS6200';
    } elsif ( $group =~ 'FAS60' ) {
        $model{group} = 'FAS6000';
    } else {
        $model{group} = $model{name};
    }

    return \%model;
} ## end sub get_model_info(%;)

1;
