#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/BLAST/CDB.pm#1 $

## @summary Construct the CDB in a SCSI commmand
## @author evan@netapp.com
## @status private
## @pod here

=pod

=head1 NAME

CDB.pm

=head1 DESCRIPTION

This package can be used to create a CDB byte string which includes the SCSI
operation code, logical bus address (LBA), and transfer length.

=head1 METHODS

=cut

package CDB;

use strict;

use San::SanClient::ClientUtil;
use Error;
use bigint;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error);

# Global definitions of SCSI commands
$CDB::TUR        = 0x00;
$CDB::READ_6     = 0x08;
$CDB::READ_10    = 0x28;
$CDB::READ_12    = 0xA8;
$CDB::READ_16    = 0x88;
$CDB::XDREAD_10  = 0x52;
$CDB::WRITE_6    = 0x0A;
$CDB::WRITE_10   = 0x2A;
$CDB::WRITE_12   = 0xAA;
$CDB::WRITE_16   = 0x8A;
$CDB::XDWRITE_10 = 0x50;

# The following codes have the service action appended to make them unique
$CDB::XDREAD_32  = 0x7F0003;
$CDB::XDWRITE_32 = 0x7F0004;

##########################################################################

=head2 C<new()>

=over 4

=item * Description

Creates a new CDB object.

=item * Arguments

=over 4

=item 1 Command Length
 (Required)

The byte size of the CDB to create. Can be 6, 10, 12, or 16 bytes.
Variable sized CDBs are not currently supported.

=back

=item * Returns

CDB object. Caller should check for errors in the object.

=back

=cut

##########################################################################
sub new {
    my($class, $command_length) = @_;
    my $self = {};
    bless($self, $class);

    $self->{command_length} = $command_length;

    if ($self->{command_length} != 6 &&
        $self->{command_length} != 10 &&
        $self->{command_length} != 12 &&
        $self->{command_length} != 16 &&
        $self->{command_length} != 32) {
        $self->_set_err("CDB:new parameter \"command_length\" value of $self->{command_length} is not currently supported.");
        return $self;
    }

    $self->{cdb} = [];
    return $self;
}

##########################################################################

=head2 C<get_cdb_string()>

=over 4

=item * Description

Get the ASCII representation of the CDB hex byte string.

=item * Arguments

=over 4

=item 1 Spaces
 (Optional)

A boolean value that specifies whether to put a space between each byte
for readability.

=back

=item * Returns

The CDB hex byte string in ASCII.

=back

=cut

##########################################################################
sub get_cdb_string {
    my($self,$space) = @_;
    my $str;
    
    if (defined($space)) {
        $space = " ";
    }

    for (my $byte=0; $byte<$self->{command_length}; $byte++) {
        $str .= sprintf("%02x%s", $self->{cdb}[$byte], $space);
    }
    return $str;
}

##########################################################################

=head2 C<set_operation()>

=over 4

=item * Description

Set the SCSI op code.

=item * Arguments

=over 4

=item 1 Operation
 (Required)

The SCSI byte op code. This object has some global definitions to simplify
the specification of the op code. No check is made to the validity of the
op code.

=back

=item * Returns

$ClientUtil::TRUE

=back

=cut

##########################################################################
sub set_operation {
    my($self, $operation) = @_;

    $self->{cdb}[0] = $self->{operation} = $operation;

    # The following commands have Service Action codes appended to them 
    # in the global definition
    if ($self->{operation} == $CDB::XDREAD_32 ||
        $self->{operation} == $CDB::XDWRITE_32 ) {
        # Set the ADDITIONAL CDB LENGTH field
        $self->{cdb}[7] = 0x18;
        
        # Set the SERVICE ACTION field
        $self->{cdb}[8] = &_byte_shift_right( $self->{operation} & 0x00FF00, 2 );
        $self->{cdb}[9] = $self->{operation} & 0x0000FF;

        # Set the OPERATION CODE field
        $self->{cdb}[0] = &_byte_shift_right( $self->{operation} & 0xFF0000, 4 );
    }

    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_address()>

=over 4

=item * Description

Set the logical block address.

=item * Arguments

=over 4

=item 1 Address
 (Required)

The SCSI Logical Block Address (LBA). The address has a limitation based 
on the byte size of the CDB.

=back

=item * Returns

$ClientUtil::TRUE  If the address was successfully set.

$ClientUtil::FALSE If the address exceeds the CDB's allowable size.

=back

=cut

##########################################################################
sub set_address {
    my($self, $address) = @_;

    $self->{address} = $address;

    if ($self->{command_length} == 6 ) {
        if ($self->{address} > 0x1FFFF) {
            $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
            return $ClientUtil::FALSE;
        }
       $self->{cdb}[1] = &_byte_shift_right( $self->{address} & 0x10000, 4 );
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0x0FF00, 2 );
       $self->{cdb}[3] = $self->{address} & 0x000FF;

    } elsif ($self->{command_length} == 10 || $self->{command_length} == 12) {
        if ($self->{address} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
            return $ClientUtil::FALSE;
        }
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0xFF000000, 6 );
       $self->{cdb}[3] = &_byte_shift_right( $self->{address} & 0x00FF0000, 4 );
       $self->{cdb}[4] = &_byte_shift_right( $self->{address} & 0x0000FF00, 2 );
       $self->{cdb}[5] = $self->{address} & 0x000000FF;

    } elsif ($self->{command_length} == 16) {
       if ($self->{address} > 0xFFFFFFFFFFFFFFFF) {
           $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
           return $ClientUtil::FALSE;
       }
       $self->{cdb}[2] = &_byte_shift_right( $self->{address} & 0xFF00000000000000, 14 );
       $self->{cdb}[3] = &_byte_shift_right( $self->{address} & 0x00FF000000000000, 12 );
       $self->{cdb}[4] = &_byte_shift_right( $self->{address} & 0x0000FF0000000000, 10 );
       $self->{cdb}[5] = &_byte_shift_right( $self->{address} & 0x000000FF00000000, 8 );
       $self->{cdb}[6] = &_byte_shift_right( $self->{address} & 0x00000000FF000000, 6 );
       $self->{cdb}[7] = &_byte_shift_right( $self->{address} & 0x0000000000FF0000, 4 );
       $self->{cdb}[8] = &_byte_shift_right( $self->{address} & 0x000000000000FF00, 2 );
       $self->{cdb}[9] = $self->{address} & 0x00000000000000FF;

    } elsif ($self->{command_length} == 32) {
       if ($self->{address} > 0xFFFFFFFFFFFFFFFF) {
           $self->_set_err("CDB:set_address address value exceeds allowable range for operation");
           return $ClientUtil::FALSE;
       }
       $self->{cdb}[12] = &_byte_shift_right( $self->{address} & 0xFF00000000000000, 14 );
       $self->{cdb}[13] = &_byte_shift_right( $self->{address} & 0x00FF000000000000, 12 );
       $self->{cdb}[14] = &_byte_shift_right( $self->{address} & 0x0000FF0000000000, 10 );
       $self->{cdb}[15] = &_byte_shift_right( $self->{address} & 0x000000FF00000000, 8 );
       $self->{cdb}[16] = &_byte_shift_right( $self->{address} & 0x00000000FF000000, 6 );
       $self->{cdb}[17] = &_byte_shift_right( $self->{address} & 0x0000000000FF0000, 4 );
       $self->{cdb}[18] = &_byte_shift_right( $self->{address} & 0x000000000000FF00, 2 );
       $self->{cdb}[19] = $self->{address} & 0x00000000000000FF;
    } else {
        $self->_set_err("CDB:set_address unknown command length");
        return $ClientUtil::FALSE;

    }
    return $ClientUtil::TRUE;
}

##########################################################################

=head2 C<set_transfer_length()>

=over 4

=item * Description

Set the transfer byte length.

=item * Arguments

=over 4

=item 1 Length
 (Required)

The SCSI transfer length in bytes. The transfer length has a limitation 
based on the byte size of the CDB.

=back

=item * Returns

$ClientUtil::TRUE  If the transfer length was successfully set.

$ClientUtil::FALSE If the transfer length exceeds the CDB's allowable
size.

=back

=cut

##########################################################################
sub set_transfer_length {
    my($self, $transfer_length) = @_;

    $self->{transfer_length} = $transfer_length;

    if ($self->{transfer_length} % 512) {
        $self->_set_err("CDB::set_transfer_length value must be a multiple of 512");
        return $ClientUtil::FALSE;
    } else {
        $self->{transfer_length} = $self->{transfer_length} / 512;
    }

    if ($self->{command_length} == 6 ) {
        if ($self->{transfer_length} > 0xFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[4] = $self->{transfer_length};

    } elsif ($self->{command_length} == 10 ) {
        if ($self->{transfer_length} > 0xFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[7] = &_byte_shift_right( $self->{transfer_length} & 0xFF00, 2 );
        $self->{cdb}[8] = $self->{transfer_length} & 0x00FF;

    } elsif ($self->{command_length} == 12 ) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[6] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[7] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[8] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[9] = $self->{transfer_length} & 0x000000FF;

    } elsif ($self->{command_length} == 16) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[10] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[11] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[12] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[13] = $self->{transfer_length} & 0x000000FF;

    } elsif ($self->{command_length} == 32) {
        if ($self->{transfer_length} > 0xFFFFFFFF) {
            $self->_set_err("CDB:set_transfer_length value exceeds allowable range for command type");
            return $ClientUtil::FALSE;
        }
        $self->{cdb}[28] = &_byte_shift_right( $self->{transfer_length} & 0xFF000000, 6 );
        $self->{cdb}[29] = &_byte_shift_right( $self->{transfer_length} & 0x00FF0000, 4 );
        $self->{cdb}[30] = &_byte_shift_right( $self->{transfer_length} & 0x0000FF00, 2 );
        $self->{cdb}[31] = $self->{transfer_length} & 0x000000FF;

    } else {
        $self->_set_err("CDB:set_address unknown command length");
        return $ClientUtil::FALSE;

    }
    return $ClientUtil::TRUE;

}
#### P r i v a t e    M e t h o d s ####

=pod

=head1 PRIVATE METHODS

=cut

##########################################################################

=head2 C<_byte_shift_right()>

=over 4

=item * Description

Shifts an integer to the right by the specified byte count.

=item * Arguments

=over 4

=item 1 Value
 (Required)

The value to shift.

=item 2 Count
 (Required)

The number of bytes to shift to the right.

=back

=item * Returns

The shifted value.

=back

=cut

##########################################################################
sub _byte_shift_right {
    my ($value, $count) = @_;

    return ( $value >> ($count << 2) );
}

=pod

=head1 AUTHOR

Evan Chiu (evan@netapp.com)

=head1 COPYRIGHT

Copyright (c) 2008 Network Appliance, Inc. All rights reserved.

=cut
