#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/BLAST/Client.pm#1 $

## @summary Interface for a host machine
## @author evan@netapp.com
## @status private
## @pod here

=pod

=head1 NAME

Client.pm

=head1 DESCRIPTION

Wrapper for SanClient with additional methods to facillitate ONTAP SAN
unit tests.

=head1 METHODS

=cut

package Client;

use strict;

use Error;
use Utils;
use Tharn;

use BLAST::CDB;
use San::SanClient;
use San::SanClient::ClientUtil;
use Data::Dumper;

use vars qw(@ISA);
@ISA = qw(Error SanClient);

$Client::TRUE  = $ClientUtil::TRUE;
$Client::FALSE = $ClientUtil::FALSE;

##########################################################################

=head2 C<new()>

=over 4

=item * Description

Creates a new Client object.

=item * Arguments Hash

=over 4

=item 1 SanClient Arguments hash
 (Required)

The arguments hash required by SanClient.

=back

=item * Returns

The Client object.

=back

=cut

##########################################################################
sub new {
    my($class, %args) = @_;
    my $self = {};
    my (@targets);
    bless ($self,$class);
   
    foreach (keys %args) {
        if ($_ =~ /-?mp_type/) {
            if ($args{$_} =~ /auto/) {
                # Try to determine multipathing type. Default to 'none' if cannot be
                # determined. If not set to auto, pass through what caller set.
                my $conn = $self->_connect($args{-host},$args{-conn_type});
                if ($conn->{hostp}{hosttype} =~ /solaris/) {
                    # Check if MPXIO is enabled
                    my ($status, @output) = &ClientUtil::execute($conn,"cat /kernel/drv/fp.conf");
                    if ( grep(/^mpxio-disable="no"/i, @output) ) {
                        $args{-mp_type} = "mpxio";
                        $self->{lvm} = "svm";
                        next;
                    }

                    # Check if Veritas DMP is enabled
                    my ($status, @output) = &ClientUtil::execute($conn,"vxdctl mode");
                    if ( grep(/enabled/i, @output) ) {
                        $args{-mp_type} = "dmp";
                        $self->{lvm} = "vxvm";
                        next;
                    }
                    $args{-mp_type} = "none";
                    $self->{lvm} = "svm";
                } elsif ($conn->{hostp}{hosttype} =~ /aix/) {
                    # Check if Veritas DMP is enabled
                    my ($status, @output) = &ClientUtil::execute($conn,"vxdctl mode");
                    if ( grep(/enabled/i, @output) ) {
                        $args{-mp_type} = "dmp";
                        $self->{lvm} = "vxvm";
                        next;
                    }

                    # Check if SANpath is installed
                    my ($status, @output) = &ClientUtil::execute($conn,"lslpp -l DotHill*");
                    if ( grep(/SANpath/i, @output) ) {
                        $args{-mp_type} = "sanpath";
                        $self->{lvm} = "nativelvm";
                        next;
                    }

                    # Check if we have MPIO ODM definitions installed
                    my ($status, @output) = &ClientUtil::execute($conn,"lslpp -l *mpio_attach_kit*");
                    if ( grep(/pcmodm/i, @output) ) {
                        $args{-mp_type} = "mpio";
                        $self->{lvm} = "nativelvm";
                        next;
                    }
                    $args{-mp_type} = "none";
                    $self->{lvm} = "nativelvm";
                } else {
                    $args{-mp_type} = "none";
                    $self->{lvm} = "nativelvm";
                }
                $self->_disconnect($conn);
            }
	}
    }

    my $client = SanClient->new(%args);
    if ($client->error) {
        $self->_set_err("Client::new error creating SanClient object");
        return $self;
    }

    $self->{client} = $client; 
    $self->{hostname} = $client->hostname();

    # Set some OS specific executable locations and settings
    if ($self->{client}->hosttype =~ /windows/) {
        $self->{slash} = "\\";
        $self->{dt} = "dt.stable\\dt.exe";
        $self->{sio} = "sio.exe";
        $self->{spt} = "spt.exe";
    } else {
        $self->{slash} = "/";
        $self->{dt} = "dt.stable";
        $self->{sio} = "sio";
        $self->{spt} = "spt.0.27";
    }

    if ($self->{client}{mp_type} ne "none") {
        $self->{client}{mp}->load_path_info;
    }
    return $self;
}

##########################################################################

=head2 C<get_devices()>

=over 4

=item * Description

Gets the devices from LUNs mapped to the host.

=item * Arguments Hash

=over 4

=item 1 Host DeviceManager Arguments hash
 (Required)

The arguments hash required by the Device Manager object of the host.

=back

=item * Returns

An array of free devices.

=back

=cut

##########################################################################
sub get_devices {
    my($self, %args) = @_;
    my (@devices, @targets);

    print Dumper(%args);

    $self->{client}->configure(-cfo_mode => 'single_image');

    if ($self->{client}->hosttype =~ /solaris/) {
        $self->{client}->reboot;
    }
 
    if ($self->{client}->hosttype =~ /windows/) {
        $self->{devmgr} = $self->{client}->create_devmgr() if (!defined($self->{devmgr}));
        $self->{devmgr}->cleanup();
	
	$self->{client}->discover();

        @devices = $self->{devmgr}->get_free_devices( %args );
    } else {
        $self->{devmgr} = $self->{client}->create_devmgr() if (!defined($self->{devmgr}));
	if ($self->{client}->error) {
		$self->_set_err("Client:get_devices() error while creating device manager: " .
				$self->{client}->errmsg);
		print Dumper($self);
		return undef;
	}
        $self->{devmgr}->cleanup();
	
	$self->{client}->discover();

        @devices = $self->{devmgr}->get_free_devices( %args );
	if (!defined($self->{lvmmgr})) {
		$self->{client}->create_lvm(-lvm => $self->{lvm});
		$self->{lvmmgr} = $self->{client}->{obj}->{lvm};
	}
	$self->{lvmmgr}->rescan();
	@devices = $self->{lvmmgr}->get_free_devices( %args );
	if ($self->{client}->hosttype =~ /solaris/) {
		foreach my $device (@devices) {
			$device =~ s/d(\d+)/d${1}s2/;
			if ($args{-primary_path_only}) {
				$device = $self->_get_primary($device);
			}
		}
	}
    }
    if (!@devices) {
	print Dumper($self);
        $self->_set_err("Client::get_devices() no devices were found");
    }

    return @devices;
}

##########################################################################

=head2 C<get_device_size()>

=over 4

=item * Description

Gets the size of the LUN of the specified device in bytes.

=item * Arguments

=over 4

=item 1 Device
 (Required)

The device to get the size of.

=back

=item * Returns

The size of the LUN in bytes. If an error is encountered, a null value is 
returned. Caller should check for errors set in the object.

=back

=cut

##########################################################################
sub get_device_size {
    my ($self, $device) = @_;

    if (defined($self->{devmgr})) {
        if ($self->{client}->hosttype =~ /windows/) {
            $device =~ /PHYSICALDRIVE(\d+)/;
            $device = $1;
        }
        my $devobj = $self->{devmgr}->create_device(-name => $device);
        if ($devobj->error) {
            my $errmsg = join('\n', $devobj->errmsg());
            $self->_set_err("Client::get_device_size error while creating device object",
                            "Error message: $errmsg");
            return;
        }
        return $devobj->get_size();
    } else {
        $self->_set_err("Client::get_device_size device manager has not been previously created");
        return;
    }
    
}

sub set_primary {
    my ($self, $device) = @_;
    $self->{pathid} = $self->_get_path($device, 1);
    return $self->{pathid};
}

sub set_secondary {
    my ($self, $device) = @_;
    $self->{pathid} = $self->_get_path($device, 0);
    return $self->{pathid};
}

##########################################################################

=head2 C<get_lun()>

=over 4

=item * Description

Get the ONTAP LUN path for a given device.

=item * Arguments

=over 4

=item 1 Device
 (Required)

The device name on the host.

=back

=item * Returns

The string of the ONTAP LUN path.

=back

=cut

##########################################################################
sub get_lun {
    my ($self, $device) = @_;
    my $lun;
    if (!defined($self->{devmgr})) {
        $self->_set_err("Client::get_lun device manager has not been previously created");
        return;
    }
    my $dfmap = $self->{devmgr}{dfmap};

    if ($self->{client}->hosttype =~ /solaris/) {
        # Remove slice information
        $device =~ s/s2$//;
    }

    if (defined($dfmap)) {
        $lun = "/vol/" . $dfmap->{$device}{vol} . "/" . $dfmap->{$device}{lun};
    }
    return $lun;
}

sub destroy_filesystem {
    my ($self, %args) = @_;

    if ($self->{client}->hosttype =~ /windows/) {
        # Get the devices to populate the object with disk objects
        my @devices = $self->get_devices();

        # Get a disk object to get access to the diskpart subroutine
        my $disks = $self->{client}{WinClient}{disks};
        my $disk = $disks->{(keys %$disks)[0]};

        foreach my $drive (@{$self->{client}{WinClient}{usedLetters}}) {
            my (@output) = $self->{client}->execute("vol ${drive}:");
            if (grep(/blast/, @output)) {
                my ($status, @output) = $disk->_diskpartCmd(
                        commands => [("select volume $drive",
                                      "delete volume",
                                      "exit")],
                        comment  => "Deleting volume at ${drive}:");
                if ($status) {
                    $self->_set_err("Client::destroy_filesystem() error deleting volume");
                    return $Client::FALSE;
                }
            }
        }
    } else {
        my $lvm = $self->{client}->create_lvm(-lvm => $self->{lvm});
        if ($lvm->error()) {
            $self->_set_err("Client::destroy_filesystem() error while creating ($self->{lvm}) lvm object: ",
                    join('\n', $lvm->errmsg) );
            return $Client::FALSE;
        }

        my @vgs = $lvm->get_vgs;
        foreach my $vg (@vgs) {
            my @vols = $vg->get_vols;
            foreach my $vol (@vols) {
                $vol->destroy(-force => 1);
                if ($vol->error) {
                    $self->_set_err("Client::destroy_filesystem() failed to destroy volume: ",
                                     join('\n', $vol->errmsg) );
                }
            } #end foreach vol
            $vg->destroy;
            if ($vg->error) {
                $self->_set_err("Client::destroy_filesystem() failed to destroy volume group: ",
                                join('\n', $vg->errmsg) );
            }
        } #end foreach vg
    }

    return ($Client::TRUE);
}

sub get_mounts {
    my ($self) = @_;
    my @mounts;

    if ($self->{client}->hosttype =~ /windows/) {
        # Get the devices to populate the object with disk objects
        my @devices = $self->get_devices();
        
        # Get a disk object to get access to the diskpart subroutine
        my $disks = $self->{client}{WinClient}{disks};
        my $disk = $disks->{(keys %$disks)[0]};
        
        foreach my $drive (@{$self->{client}{WinClient}{usedLetters}}) {
            my (@output) = $self->{client}->execute("vol ${drive}:");
            if (grep(/blast/, @output)) {
                push (@mounts, "${drive}:\\");
            }   
        }   
    } else {
        my $lvm = $self->{client}->create_lvm(-lvm => $self->{lvm});
        if ($lvm->error()) {
            $self->_set_err("Client::get_mounts() error while creating lvm object: ",
                    join('\n', $lvm->errmsg) );
            return undef;
        }   
        
        my @vgs = $lvm->get_vgs;
        foreach my $vg (@vgs) {
            my @vols = $vg->get_vols;
            foreach my $vol (@vols) {
                if ($vol->is_mounted) {
                    push (@mounts, $vol->get_mount);
                    if ($vol->error) {
                        $self->_set_err("Client::get_mounts() failed to get mount point: ",
                                        join('\n', $vol->errmsg) );
                    }                   
                }   
            } #end foreach vol
        } #end foreach vg
    }   
    return @mounts;
}

sub create_filesystem {
    my ($self, %args) = @_;
    my ($filers);

    foreach (keys %args) {
        if (/-?filers/) {
            length($filers = $args{$_}) or
                do {
                    $self->_set_err("Client::create_filesystem invalid value for \"filers\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
    }   

    my @devices;
    if (defined($args{-filers})) {
        foreach my $filer (split(/,/,$args{-filers})) {
            my @devs = $self->get_devices(-filer => $filer);
            @devices = (@devices, @devs);
        }
    } else {
        @devices = $self->get_devices();
    }
    
    if ($self->error) {
	return $Client::FALSE;
    }

    if ($self->{client}->hosttype =~ /windows/) {
        my ($disk, @disks, $disks, $diskstr, $drive, $status, @output);
        $disks = $self->{client}{WinClient}{disks};
        foreach my $device (@devices) {
            my $dnum = $self->{client}{obj}{devmgr}{dfmap}{$device}{drive_number};
            # Save the disk number for later use
            push (@disks, $dnum);
            
            $disk = $disks->{$dnum};
            if (!defined($disk)) {
                $self->_set_err("Client:create_filesystem() not a valid device (dnum=$dnum)");
                return $Client::FALSE;
            }
            ($status, @output) = $disk->_diskpartCmd(
                    commands => [("select disk $dnum",
                                  "clean",
                                  "exit")],
                    comment  => "Converting disk $dnum to dynamic");
            if ($status) {
                $self->_set_err("Client::create_filesystem() error cleaning disk $dnum");
                return $Client::FALSE;
            }

            # Wait a few seconds to allow the clean to finish
            sleep(5);

            ($status, @output) = $disk->_diskpartCmd(
                    commands => [("select disk $dnum",
                                  "convert dynamic",
                                  "exit")],
                    comment  => "Converting disk $dnum to dynamic");
            if ($status) {
                $self->_set_err("Client::create_filesystem() error converting disk $dnum to dynamic");
                return $Client::FALSE;
            }
        }
        $drive = shift(@{$self->{client}{WinClient}{availLetters}});
        $diskstr = join(',', @disks);
        ($status, @output) = $disk->_diskpartCmd(
                commands => [('rem',
                              "create volume stripe disk=$diskstr",
                              "assign letter=$drive",
                              "exit")],
                comment  => "Creating striped volume");
        if ($status) {
            $self->_set_err("Client::create_filesystem() error creating striped volume");
            return $Client::FALSE;
        }
        
        $self->{client}->execute("format ${drive}: /V:blast /FS:ntfs /Q /y");
        if ($self->{client}->error) {
            $self->_set_err("Client::create_filesystem() error formatting volume: ",
                            join('\n', $self->{client}->errmsg) );
            return $Client::FALSE;
        }
        $self->{mount_point} = $drive . ":\\";
    } else {
        if ($self->{client}->hosttype =~ /solaris/ && $self->{lvm} eq "vxvm") {
            foreach my $d (@devices) {
                $d =~ s/s\d+//;
	    }
	}
        my $lvm = $self->{client}->create_lvm(-lvm => $self->{lvm});
        if ($lvm->error()) {
            $self->_set_err("Client::create_filesystem() error while creating lvm object: ",
                    join('\n', $lvm->errmsg) );
            return $Client::FALSE;
        }
    
        $lvm->initialize(-devices => \@devices);
        if ($lvm->error()) {
            $self->_set_err("Client::create_filesystem() error while initializing devices: ",
                    join('\n', $lvm->errmsg) );
            return $Client::FALSE;
        }

        my $vg = $lvm->create_vg(-devices => \@devices, -name => "blastvg");
        if ($vg->error()) {
            $self->_set_err("Client::create_filesystem() error while creating volume group: ",
                    join('\n', $vg->errmsg) );
            return $Client::FALSE;
        }

        my $devstr = join(' ', @devices);
        # SanClient SVM libraries partition it such that all data is on slice 0
        $devstr =~ s/d(\d+)/d${1}s0/g;
        my $devcnt = @devices;
        my $size;
        if ($self->{lvm} =~ /vxvm/i) {
	    # Shave off 50m for overhead
            $size = ( ($self->get_device_size($devices[0]) - 52428800) * $devcnt ) / 512;
        } else {
            $size = $self->get_device_size($devices[0]);
        }
        my $vol = $vg->create_vol(-layout => 'stripe',
                                  -name => 'vol',
                                  -size => $size,
                                  # The following parameters are needed for SVM only
                                  -num_dev_per_vol => $devcnt,
                                  -num_of_vols => 1,
                                  -dev_per_vol_list => $devstr);
        if ($vol->error()) {
            $self->_set_err("Client::create_filesystem() error while creating volume: ",
                    join('\n', $vol->errmsg) );
            return $Client::FALSE;
        }

        $vol->create_fs();
        if ($vol->error()) {
            $self->_set_err("Client::create_filesystem() error while creating filesystem: ",
                    join('\n', $vol->errmsg) );
            return $Client::FALSE;
        }

        $vol->mount if (!$vol->is_mounted);
        if ($vol->error()) {
            $self->_set_err("Client::create_filesystem() error while mounting filesystem: ",
                    join('\n', $vol->errmsg) );
            return $Client::FALSE;
        }

        $self->{mount_point} = $vol->{mount};
    }
    
    return $Client::TRUE;
}

sub cleanup {
    my ($self) = @_;

    $self->destroy_filesystem;

    if ($self->{client}->hosttype =~ /aix/) {
        $self->{client}->orphaned_device_cleanup;
    }

    if ($self->{client}->error) {
        $self->_set_err("Client::cleanup() error while cleaning up devices",
            join(/\n/, $self->{client}->errmsg));
        return ($Client::FALSE);
    }
    return ($Client::TRUE);
}

sub reset {
    my ($self) = @_;

    if ($self->{client}->hosttype =~ /aix/) {
        $self->{client}->reboot(1);
    } elsif ($self->{client}->hosttype =~ /solaris/) {
        $self->{client}->reboot;
    }

    if ($self->{client}->error) {
        $self->_set_err("Client::reset() error while cleaning up devices",
            join(/\n/, $self->{client}->errmsg));
        return ($Client::FALSE); 
    }
    return ($Client::TRUE); 
}

##########################################################################

=head2 C<read()>

=over 4

=item * Description

Issues a SCSI READ command on the host to the specified device.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to perform the read.

=item 2 -address
 (Required)

The address from which to start reading from.

=item 3 -length
 (Required)

The amount of data to read in bytes.

=back

=item * Returns

$Client::TRUE If the read was successful.
$Client::FALSE If the read failed.

=back

=cut

##########################################################################
sub read {
    my($self, %args) = @_;

    my ($device, $address, $length, $cmd, $cdb, $output);
    my ($path, $pathid, $primary_path, $secondary_path);

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::read invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?address/) {
            length($address = $args{$_}) or
                do {
                    $self->_set_err("Client::read invalid value for \"address\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?length/) {
            length($length = $args{$_}) or
                do {
                    $self->_set_err("Client::read invalid value for \"length\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?primary_path/) {
            length($primary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::read invalid value for \"primary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?secondary_path/) {
            length($secondary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::read invalid value for \"secondary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
	}
    }

    # Check for required values
    if (!defined($device) ||
        !defined($address) ||
        !defined($length) ) {
        $self->_set_err("Client::read mandatory parameter \"device\", \"address\", or \"length\" not specified");
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    # Build the cdb
    if ($address <= 0x1FFFF) { # 21-bit address
        $cdb = new CDB (6);
        $cdb->set_operation($CDB::READ_6);
        $cdb->set_address($address);
        $cdb->set_transfer_length($length);
    } elsif ($address <= 0xFFFFFFFF) { # 32-bit address
        $cdb = new CDB (10);
        $cdb->set_operation($CDB::READ_10);
        $cdb->set_address($address);
        $cdb->set_transfer_length($length);
    } else { # assume 64-bit address
        $cdb = new CDB (16);
        $cdb->set_operation($CDB::READ_16);
        $cdb->set_address($address);
        $cdb->set_transfer_length($length);
    }

    if ($primary_path) {
    	$pathid = $self->_get_path($device, 1);
    } elsif ($secondary_path) {
        $pathid = $self->_get_path($device, 0);
    }
    if (defined($pathid)) {
        $path = "path=$pathid";
    } else {
        $path = "";
    }

    # Execute read command using the spt utility
    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s cdb=\"%s\" dir=read length=%d %s", $self->{client}->{TEST_BIN}, $device, $cdb->get_cdb_string(1), $length, $path);
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

sub xdread {
    my($self, %args) = @_;

    my ($device, $address, $length, $cmd, $cdb, $output);

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::xdread invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        if (/-?address/) {
            length($address = $args{$_}) or
                do {
                    $self->_set_err("Client::xdread invalid value for \"address\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        if (/-?length/) {
            length($length = $args{$_}) or
                do {
                    $self->_set_err("Client::xdread invalid value for \"length\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
    }

    # Check for required values
    if (!defined($device) ||
        !defined($address) ||
        !defined($length) ) {
        $self->_set_err("Client::xdread mandatory parameter \"device\", \"address\", or \"length\" not specified");
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    # Build the cdb
    if ($address <= 0xFFFFFFFF) { # 32-bit address
        $cdb = new CDB (10);
        $cdb->set_operation($CDB::XDREAD_10);
        $cdb->set_address($address);
        $cdb->set_transfer_length($length);
    } else { # 64-bit address
        $cdb = new CDB (32);
        $cdb->set_operation($CDB::XDREAD_32);
        $cdb->set_address($address);
        $cdb->set_transfer_length($length);
    }

    # Execute read command using the spt utility
    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s cdb=\"%s\" dir=read length=%d onerr=stop", $self->{client}->{TEST_BIN}, $device, $cdb->get_cdb_string(1), $length);
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}
##########################################################################

=head2 C<write()>

=over 4

=item * Description

Issues a SCSI WRITE command on the host to the specified device. The data
written is a generated pattern from dt.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to perform the write.

=item 2 -address
 (Required)

The address from which to start writing to.

=item 3 -length
 (Required)

The amount of data to write in bytes.

=back

=item * Returns

$Client::TRUE If the write was successful.
$Client::FALSE If the write failed.

=back

=cut

##########################################################################
sub write {
    my($self, %args) = @_;

    my ($device, $address, $length, $cmd, $cdb, $output);
    my ($path, $pathid, $primary_path, $secondary_path);

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::write invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?address/) {
            length($address = $args{$_}) or
                do {
                    $self->_set_err("Client::write invalid value for \"address\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?length/) {
            length($length = $args{$_}) or
                do {
                    $self->_set_err("Client::write invalid value for \"length\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?primary_path/) {
            length($primary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::write invalid value for \"primary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?secondary_path/) {
            length($secondary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::write invalid value for \"secondary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
	}
    }

    # Check for required values
    if (!defined($device) ||
        !defined($address) ||
        !defined($length) ) {
        $self->_set_err("Client::write mandatory parameter \"device\", \"address\", or \"length\" not specified");
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    # Build the cdb
    if ($address <= 0x1FFFF) { # 21-bit address
        $cdb = new CDB (6);
        $cdb->set_operation($CDB::WRITE_6);
        $cdb->set_address($address);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB address: $errmsg");
            return $Client::FALSE;
        }
        $cdb->set_transfer_length($length);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB transfer length: $errmsg");
            return $Client::FALSE;
        }
    } elsif ($address <= 0xFFFFFFFF) { # 32-bit address
        $cdb = new CDB (10);
        $cdb->set_operation($CDB::WRITE_10);
        $cdb->set_address($address);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB address: $errmsg");
            return $Client::FALSE;
        }
        $cdb->set_transfer_length($length);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB transfer length: $errmsg");
            return $Client::FALSE;
        }
    } else { # assume 64-bit address
        $cdb = new CDB (16);
        $cdb->set_operation($CDB::WRITE_16);
        $cdb->set_address($address);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB address: $errmsg");
            return $Client::FALSE;
        }
        $cdb->set_transfer_length($length);
        if ($cdb->error) {
            my $errmsg = join('\n', $cdb->errmsg());
            $self->_set_err("Client:write error while setting CDB transfer length: $errmsg");
            return $Client::FALSE;
        }
    }   
    
    if ($primary_path) {
    	$pathid = $self->_get_path($device, 1);
    } elsif ($secondary_path) {
        $pathid = $self->_get_path($device, 0);
    }
    if (defined($pathid)) {
        $path = "path=$pathid";
    } else {
        $path = "";
    }

    # Execute write command using the spt utility using dt to generate a write pattern
    # Default shell has a command line limit of 254 characters. Break commmand up accordingly.
    $cmd = sprintf("%s$self->{slash}$self->{dt} of=- bs=%d count=1 pattern=iot disable=verify,stats | %s$self->{slash}$self->{spt} dsf=%s din=- cdb=\"%s\" dir=write length=%d %s",
    #$cmd = sprintf("%s$self->{slash}$self->{dt} of=- bs=%d count=1 pattern=iot disable=verify,stats | %s%s$self->{slash}$self->{spt} dsf=%s din=- cdb=\"%s\" dir=write length=%d onerr=stop",
                    $self->{client}->{TEST_BIN},
                    $length,
                    #($self->{client}->hosttype =~ /unix/) ? "\\\n" : undef,
                    $self->{client}->{TEST_BIN},
                    $device, 
                    $cdb->get_cdb_string(1), 
                    $length,
		    $path);

    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

##########################################################################

=head2 C<test_unit_ready()>

=over 4

=item * Description

Issues a SCSI TUR command on the host to the specified device. Because
this uses the SCSI pass through tool (spt) not all platforms are capable
of issuing a TUR with this method.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to perform the TUR.

=back

=item * Returns

$Client::TRUE If the TUR was successful.
$Client::FALSE If the TUR failed.

=back

=cut

##########################################################################
sub test_unit_ready {
    my($self, %args) = @_;
    
    my ($device, $cmd, $cdb, $output, $path, $pathid);
    
    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::test_unit_ready invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    if (defined($self->{pathid})) {
        $path = "path=$self->{pathid}";
    } else {
        $path = "";
    }
    
    # Build the cdb
    $cdb = new CDB (6);

    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s cdb=\"%s\" $path", $self->{client}->{TEST_BIN}, $device, $cdb->get_cdb_string(1));
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

##########################################################################

=head2 C<abort_task_set()>

=over 4

=item * Description

Issues an Abort Task Set TMF command on the host to the specified device. Because
this uses the SCSI pass through tool (spt) not all platforms are capable
of issuing an Abort Task Set TMF with this method.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to send the Abort Task Set TMF.

=back

=item * Returns

$Client::TRUE If the TMF was successful.
$Client::FALSE If the TMF failed.

=back

=cut

##########################################################################
sub abort_task_set {
    my($self, %args) = @_;

    my ($device, $primary_path, $secondary_path, $pathid, $path, $cmd, $output);

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::abort_task_set invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?primary_path/) {
            length($primary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::abort_task_set invalid value for \"primary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?secondary_path/) {
            length($secondary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::abort_task_set invalid value for \"secondary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
	}
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    if (defined($primary_path)) {
    	$pathid = $self->_get_path($device, 1);
    } elsif (defined($secondary_path)) {
        $pathid = $self->_get_path($device, 0);
    }
    if (defined($pathid)) {
        $path = "path=$pathid";
    } else {
        $path = "";
    }
	
    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s op=abort_task_set %s", $self->{client}->{TEST_BIN}, $device, $path);
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

##########################################################################

=head2 C<lun_reset()>

=over 4

=item * Description

Issues a LUN Reset TMF command on the host to the specified device. Because
this uses the SCSI pass through tool (spt) not all platforms are capable
of issuing a LUN Reset TMF with this method.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to send the LUN Reset TMF.

=item 2 -path
 (Optional)

The path (if multipathed) to issue the command over. Default is path 0.

=back

=item * Returns

$Client::TRUE If the TMF was successful.
$Client::FALSE If the TMF failed.

=back

=cut

##########################################################################
sub lun_reset {
    my($self, %args) = @_;

    my ($device, $primary_path, $secondary_path, $pathid, $path, $cmd, $output);

    # Initialize default values
    $path = 0;

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::lun_reset invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?primary_path/) {
            length($primary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::lun_reset invalid value for \"primary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?secondary_path/) {
            length($secondary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::lun_reset invalid value for \"secondary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    if (defined($primary_path)) {
    	$pathid = $self->_get_path($device, 1);
    } elsif (defined($secondary_path)) {
        $pathid = $self->_get_path($device, 0);
    }
    if (defined($pathid)) {
        $path = "path=$pathid";
    } else {
        $path = "";
    }

    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s op=lun_reset %s", $self->{client}->{TEST_BIN}, $device, $path);
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

##########################################################################

=head2 C<target_reset()>

=over 4

=item * Description

Issues a Target Reset TMF command on the host to the specified device. Because
this uses the SCSI pass through tool (spt) not all platforms are capable
of issuing a Target Reset TMF with this method.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to send the Target Reset TMF.

=back

=item * Returns

$Client::TRUE If the TMF was successful.
$Client::FALSE If the TMF failed.

=back

=cut

##########################################################################
sub target_reset {
    my($self, %args) = @_;

    my ($device, $primary_path, $secondary_path, $cmd, $path, $pathid, $output);

    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::target_reset invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?primary_path/) {
            length($primary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::target_reset invalid value for \"primary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        } elsif (/-?secondary_path/) {
            length($secondary_path = $args{$_}) or
                do {
                    $self->_set_err("Client::target_reset invalid value for \"secondary_path\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
	}
    }

    # Get the full path to the device handle
    $device = $self->_normalize_device($device);

    if (defined($primary_path)) {
    	$pathid = $self->_get_path($device, 1);
    } elsif (defined($secondary_path)) {
        $pathid = $self->_get_path($device, 0);
    }
    if (defined($pathid)) {
        $path = "path=$pathid";
    } else {
        $path = "";
    }
	
    $cmd = sprintf("%s$self->{slash}$self->{spt} dsf=%s op=target_reset %s", $self->{client}->{TEST_BIN}, $device, $path);
    $output = $self->{client}->execute($cmd);

    if ($self->{client}->error) {
        return ($Client::FALSE, $output);
    } else {
        return ($Client::TRUE, $output);
    }
}

##########################################################################

=head2 C<run_io()>

=over 4

=item * Description

Starts a continuous I/O process on the host to the specified device. Only
a single I/O process can occur on the host at a time. See stop_io() to stop
the I/O process.

=item * Arguments Hash

=over 4

=item 1 -device
 (Required)

The device on the host on which to perform I/O upon.

=item 2 -runtime
 (Optional)

The amount of time to run I/O. Default is 1 hour.

=item 3 -blocksize
 (Optional)

The block size to use. Default is 32k.

=item 4 -threads
 (Optional)

The number of threads to spawn. Default is 4.

=back

=item * Returns

$Client::TRUE If I/O was started successfully.
$Client::FALSE If I/O was not able to start.

=back

=cut

##########################################################################
sub run_io {
    my($self, %args) = @_;

    my ($device, $directory, $cmd, $size);
    my $runtime = "5m";
    my $blocksize = "32k";
    my $threads = 4;
    my $generator = "sio";
    my $sio_read_percent = 50;
    my $sio_rand_percent = 50;

    if (defined($self->{dt_subtest})) {
        $self->_set_err("Client::run_io I/O appears to already be running");
        return $Client::FALSE;
    }
    foreach (keys %args) {
        if (/-?device/) {
            length($device = $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"device\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?directory/) {
            length($directory= $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"directory\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?runtime/) {
            length($runtime = $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"runtime\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?blocksize/) {
            length($blocksize = $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"blocksize\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?threads/) {
            length($threads = $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"threads\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?generator/) {
            length($generator= $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"generator\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?sio_read_percent/) {
            length($sio_read_percent= $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"sio_read_percent\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
        elsif (/-?sio_rand_percent/) {
            length($sio_rand_percent= $args{$_}) or
                do {
                    $self->_set_err("Client::run_io invalid value for \"sio_rand_percent\":",
                            "$args{$_}");
                    return $Client::FALSE;
                };
        }
    }

    if (defined($device) && defined($directory)) {
        $self->_set_err("Client::run_io parameter\"device\" and \"directory\" are mutually exclusive");
        return $Client::FALSE;
    }
    
    if (defined($device)) {
        $size = $self->get_device_size($device);
        if ($self->error) {
            return $Client::FALSE;
        }
    } elsif (defined($directory)) {
        $size = "4g"; # temp value for now
    }
    
    # Create dt/sio invocation command. We will be manually stopping I/O.
    # If on Windows, use sio as dt appears to have issues with raw
    # disk devices (no device error).
    if (defined($device)) {
        # Run I/O to a single device/LUN
        if ($self->{client}->hosttype =~ /windows/ && $generator ne "sio") {
            $generator = "sio";
        }

        if ($generator eq "sio") {
            $cmd = sprintf("%s$self->{slash}$self->{sio} %d %d %s 1 %d %d %d %s -verify -nosysinfo",
                    $self->{client}{TEST_BIN},
                    $sio_read_percent,
                    $sio_rand_percent,
                    $blocksize,
                    (&ClientUtil::convert_to_bytes($size) - &ClientUtil::convert_to_bytes("128m"))/512,
                    &ClientUtil::convert_to_seconds($runtime),
                    $threads, 
                    $self->_normalize_device($device)
                    );
        } elsif ($generator eq "dt") {
            $cmd = sprintf("%s$self->{slash}$self->{dt} of=%s capacity=%s rlimit=%s bs=%s slices=%d oncerr=abort disable=pstats errors=1 pattern=iot runtime=%d",
                    $self->{client}{TEST_BIN},
                    $self->_normalize_device($device),
                    &ClientUtil::convert_to_bytes($size),
                    &ClientUtil::convert_to_bytes($size),
                    $blocksize,
                    $threads,
                    &ClientUtil::convert_to_seconds($runtime)
                    );
        } else {
            $self->_set_err("Client::run_io invalid generator specified");
        } 
    } else {
        my $file = $directory . $self->{slash} . "blastfile";
        # Run I/O to a mounted filesystem
        if ($generator eq "sio") {
            # First we need to zero out the file
            if ($self->{client}->hosttype =~ /windows/) {
                $self->{client}->execute("echo ' ' > $file");
            } else {
                $self->{client}->execute("touch $file");
            }
            $cmd = sprintf("%s$self->{slash}$self->{sio} 0 0 4k 0 %d 0 %d %s -fillonce",
                    $self->{client}{TEST_BIN},
                    &ClientUtil::convert_to_bytes($size)/512 - 1024,
                    $threads,
                    $file
                    );
            $self->{client}->execute($cmd);
    
            # Now we can run actual I/O 
            $cmd = sprintf("%s$self->{slash}$self->{sio} %d %d %s 1 %d %d %d %s -verify -nosysinfo",
                    $self->{client}{TEST_BIN},
                    $sio_read_percent,
                    $sio_rand_percent,
                    $blocksize,
                    &ClientUtil::convert_to_bytes($size)/512 - 1,
                    &ClientUtil::convert_to_seconds($runtime),
                    $threads,
                    $file
                    );
        } elsif ($generator eq "dt") {
	    # Remove any existing file
            if ($self->{client}->hosttype =~ /windows/) {
                $self->{client}->execute("del $file /y");
            } else {
                $self->{client}->execute("rm -f $file");
            }
            $cmd = sprintf("%s$self->{slash}$self->{dt} of=%s limit=%s %s procs=%s oncerr=abort disable=pstats errors=1 pattern=iot runtime=%d",
                    $self->{client}{TEST_BIN},
                    $file,
                    &ClientUtil::convert_to_bytes($size),
                    $blocksize eq "random" ? "incr=var min=16k max=256k" : "bs=$blocksize",
                    $threads,
                    &ClientUtil::convert_to_seconds($runtime)
                    );
        } else {
            $self->_set_err("Client::run_io invalid generator specified");
        }
    }

    $self->{dt_subtest} = $self->_execute_background(-command => $cmd,
                                                     -timeout => &ClientUtil::convert_to_seconds($runtime) + 300,
                                                     -subtest_id => $self->{client}->hostname);

    return $Client::TRUE;
}

##########################################################################

=head2 C<stop_io()>

=over 4

=item * Description

Stops any running I/O process that this object created. See run_io() to
start the I/O process.

=item * Returns

$Client::TRUE If I/O was stopped successfully.
$Client::FALSE If I/O could not be stopped.

=back

=cut

##########################################################################
sub stop_io {
    my ($self) = @_;

    if (defined($self->{dt_subtest}) && $self->{dt_subtest}->is_alive) {
        $self->{dt_subtest}->kill( complain=>["-ok"]);
        $self->{dt_subtest} = undef;
        return $Client::TRUE;
    } else {
        $self->_set_err("Client::stop_io no I/O is running that needs to be stopped");
        return $Client::FALSE;
    }

}

sub kill_io {
    my ($self) = @_;

    if ( &ClientUtil::kill_process_instances($self->{hostname}, "sio", undef, $self->{client}{conn}) != $ClientUtil::TRUE &&
         &ClientUtil::kill_process_instances($self->{hostname}, "dt", undef, $self->{client}{conn}) != $ClientUtil::TRUE) {
        $self->_set_err("Client::kill_io unable to kill running I/O process");
        return $Client::FALSE;
    }

    return $Client::TRUE;
}

##########################################################################

=head2 C<io_alive()>

=over 4

=item * Description

Check if the host currently has a running I/O process.

=item * Returns

$Client::TRUE If I/O process is still running.
$Client::FALSE If I/O process is not running.

=back

=cut

##########################################################################
sub io_alive {
    my ($self) = @_;

    if (defined($self->{dt_subtest}) && $self->{dt_subtest}->is_alive) {
        return $Client::TRUE;
    } else {
        if (defined($self->{dt_subtest})) {
            $self->{dt_subtest} = undef;
        }
        return $Client::FALSE;
    }
}

sub io_exit_status {
    my ($self) = @_;

    if (defined($self->{dt_subtest})) {
        return $self->{dt_subtest}->exit_status;
    } else {
        return undef;
    }
}

#### P r i v a t e    M e t h o d s ####

=pod

=head1 PRIVATE METHODS

=cut

##########################################################################

=head2 C<_normalize_device()>

=over 4

=item * Description

Converts a device string to the full device path on Unix hosts.

=item * Arguments

=over 4

=item 1 Device
(Required)

The device string to convert.

=back

=item * Returns

The full path string to the device.

=back

=cut

##########################################################################
sub _normalize_device {
    my ($self,$device) = @_;

    if ($self->{client}->hosttype =~ /solaris/i) {
        # Write to slice 6 as the partitions are repartitioned by sanclient
        $device =~ s/s\d$/s6/;
        $device = "/dev/rdsk/" . $device;
    } elsif ($self->{client}->hosttype =~ /windows/i) {
    } elsif ($self->{client}->hosttype =~ /aix/i) {
        $device = "/dev/" . $device;
    } elsif ($self->{client}->hosttype =~ /hpux/i) {
    } else {
        $self->_set_err("Client::_normalize_device unknown host type");
        return;
    }

    return $device;
}

##########################################################################

=head2 C<_execute_background()>

=over 4

=item * Description

Background a client command by creating a new subtest.

=item * Arguments Hash

=over 4

=item 1 -command
(Required)

The command to run on the host.

=item 2 -subtest_id

The ID to assign the subtest.

=back

=item * Returns

The subtest object.

=back

=cut

##########################################################################
sub _execute_background {
    my ($self,%args)=@_;
    my ( $command, $timeout, $subtest_id, $subtest);

    foreach (keys %args) {
        if (/-?command$/) {
            if (not length($command = $args{$_})) {
                $self->_set_err("Error: Client::_execute_background()",
                        "Invalid value for argument '-command'");
                return $self;
            }
            next;
        }
        elsif (/-?subtest_id$/) {
            if (not length($subtest_id = $args{$_})) {
                $self->_set_err("Error: Client::_execute_background()",
                        "Invalid value for argument '-subtest_id'");
                return $self;
            }
            next;
        }
        elsif (/-?timeout$/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Error: Client::_execute_background()",
                        "Invalid value for argument '-timeout'");
                return $self;
            }
            next;
        }
    }

    $subtest = Tharn::subtest(
            sub {
                my ($status, @output) = &ClientUtil::execute($self->{client}{conn}, $command, $timeout);
                if ($status) {
                    die( "Error: Client::_execute_background() Background process '$subtest_id' failed: " .
                         join(' ', $self->{client}->errmsg) );
                 }
            },
            -bg,
            -runid=>$subtest_id );

    return $subtest;
}

##########################################################################

=head2 C<_connect()>

=over 4

=item * Description

Create a connection object. This is mainly used to collect information
about a host to instantiate SanClient with. Normally, we would let
SanClient handle it.

=item * Arguments

=over 4

=item 1 Hostname
(Required)

The hostname of the host to connect to.

=item 2 Connection Type
(Required)

The type of connection to make. A value of THARNHOST would use the
connection entries in the tharnhost file.

=back

=item * Returns

A connection object.

=back

=cut

##########################################################################
    sub _connect {
        my ($self, $hostname, $conn_type) = @_;
        my ($host, $conn_str, $connect);

        $host = Utils::make_host($hostname);
        if ($conn_type eq "THARNHOST") {
            $conn_type = "";
# Check to see if any connection types are defined, if not return the default value
            if (!defined($host->{connection_fields})) {
# Set the $conn_type to the default of telnet
                $conn_type = "telnet";
            }
# Go throught each remainder field and set $conn_type
# IF rsh is defined, set that. This means that if the
# conn type of rsh is defined in the tharnhost file,
# we set $conn_type to that. To not use rsh, comment
# out the conn type of rsh from the tharnhost file.
# This will then check to see if the next available
# conn type is defined in the thatnhost file.
# The order of connection type is:
# 1) rsh
# 2) ssh
# 3) telnet
# 4) rlogin
# 5) console
            elsif (defined($host->{connection_fields}->{rsh})) {
                $conn_type = "rsh";
            }
            elsif (defined($host->{connection_fields}->{ssh})) {
                $conn_type = "ssh";
            }
            elsif (defined($host->{connection_fields}->{telnet})) {
                $conn_type = "telnet";
            }
            elsif (defined($host->{connection_fields}->{rlogin})) {
                $conn_type = "rlogin";
            }
            elsif (defined($host->{connection_fields}->{console})) {
                $conn_type = "console";
            }
        }
        elsif ($conn_type eq "") {
            $conn_type = "telnet";
            if ($host->hosttype() =~ /windows/) {
                $conn_type = "rsh-persay";
            }
        }
        else {
# Default $conn_type to the specified value user passed in
        }

        $conn_str=$hostname."#".$conn_type.";"   .
            "username=".$host->{connection_fields}{$conn_type}{username}.";" .
            "password=".$host->{connection_fields}{$conn_type}{password}.";" .
            "timeout=".$self->{timeout}.";inherit=0;";

        $self->sanlog(-msg_level => 3, -msg => "Connection string is $conn_str");
        $connect = Tharn::connect(
                $conn_str,
                -eval=>sub{
                if (@_) {
                my $errmsg = join("",@_);
                $errmsg =~ s/\n//g;
                $self->_set_err($errmsg);

# Display an error in case user doesn't include error
# handler on object instantiation.
                $self->sanlog(-msg => "ERROR: Client::_connect() " .
                    "unable to connect to client '" .
                    $hostname."': $errmsg");

                Tharn::logresult("CONFIG",abort=>0,"\n@_");

                return $self;
                }
                });

        return $connect;
    }

##########################################################################

=head2 C<_disconnect()>

=over 4

=item * Description

Disconnect a connection. Normally we use the SanClient object's methods
but in the case where we pulled additional information from the host, we
use this method to disconnect instead.

=item * Arguments

=over 4

=item 1 Connection Object
(Required)

The connection object to disconnect.

=back

=item * Returns

Nothing.

=back

=cut

##########################################################################
sub _disconnect {
	my ($self, $conn) = @_;

	Tharn::disconnect($conn);
}

sub _get_path {
	my ($self, $device, $primary) = @_;
	my ($status, @output);
	if ($self->{client}->hosttype =~ /aix/i) {
		@output = $self->{client}->execute("sanlun lun show -d $device");
		$output[1] =~ /\s*(\S+:)\s+(\S+)/;
		my $lunpath = $1 . $2;

		@output = $self->{client}->execute("sanlun lun show -p $lunpath");
		foreach my $line (@output) {
			if ($line =~ /primary\s+path(\d+)/) {
				return $1;
			}
		}
	} else {
		return undef;
	}
}

sub _get_primary {
	my ($self, $device) = @_;
	my ($status, @output);
	if ($self->{client}->hosttype =~ /solaris/i && $self->{lvm} =~ /vxvm/i) {
		@output = $self->{client}->execute("sanlun lun show -d /dev/rdsk/$device");
		$output[1] =~ /\s*(\S+:)\s+(\S+)/;
		my $lunpath = $1 . $2;

		@output = $self->{client}->execute("sanlun lun show -p $lunpath");
		foreach my $line (@output) {
			if ($line =~ /primary\s+\/dev\/rdsk\/(\S+)/) {
				return $1;
			}
		}
	} else {
		return $device;
	}
}

=pod

=head1 AUTHOR

Evan Chiu (evan@netapp.com)

=head1 COPYRIGHT

Copyright (c) 2008 Network Appliance, Inc. All rights reserved.

=cut

1;
