#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/BLAST/Controller.pm#1 $

## @summary Interface for an ONTAP controller
## @author evan@netapp.com
## @status private
## @pod here

=pod

=head1 NAME

Controller.pm

=head1 DESCRIPTION

Package to interface with an ONTAP controller. This library utilizes the
SanUtil library to issue ZAPI commands.

=head1 METHODS

=cut

package Controller;

use strict;

use Tharn;
use Error;
use San::SanFiler;
use San::SanUtil;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error);

$Controller::TRUE  = 1;
$Controller::FALSE = 0;

##########################################################################

=head2 C<new()>

=over 4

=item * Description

Creates a new Controller object.

=item * Arguments Hash

=over 4

=item 1 -hostname
 (Required)

The DNS hostname of the ONTAP controller.

=item 2 -priv
 (Optional)

The privilege level to use. Default is test.

=item 3 -conn_type
 (Optional)

The type of connection to use: ZAPI, telnet, rsh, etc. Default is ZAPI.

=item 4 -timeout

The timeout value to use when issuing commands to the controller. This only
applies to connection types that are not ZAPI.

=back

=item * Returns

Controller object. Caller should check for errors in the object.

=back

=cut

##########################################################################
sub new {
    my($class, %args) = @_;

    my $self = {}; 
    bless $self, $class;
    
    # Initialize default values
    $self->{priv} = "diag";
    $self->{conn_type} = "zapi";
    $self->{timeout} = 180;

    foreach (keys %args) {
        if (/-?hostname/) {
            length($self->{hostname} = $args{$_}) or
                do {
                    $self->_set_err("Controller::new invalid value for \"hostname\":",
                            "$args{$_}");
                    return $self;
                };
        }
        if (/-?priv/) {
            (length($self->{priv} = $args{$_}) && $args{$_} =~ /(admin|advanced|diag|test)/) or
                do {
                    $self->_set_err("Controller::new invalid value for \"priv\":",
                            "$args{$_}");
                    return $self;
                };
        }
        if (/-?conn_type/) {
            length($self->{conn_type} = $args{$_}) or
                do {
                    $self->_set_err("Controller::new invalid value for \"conn_type\":",
                            "$args{$_}");
                    return $self;
                };
        }
        if (/-?timeout/) {
            length($self->{timeout} = $args{$_}) or
                do {
                    $self->_set_err("Controller::new invalid value for \"timeout\":",
                            "$args{$_}");
                    return $self;
                };
        }
    }
    if (!defined($self->{hostname})) {
        $self->_set_err("Controller::new mandatory parameter \"hostname\" not specified");
    }

    # Store available rastrace modules
    my (undef, $status, @loutput) = $self->execute(-command => 'rastrace list -m 0');
    foreach (@loutput) {
        #    FCT_0a, id =  2, inst = 0,     ACTIVE
        if ($_ =~ /\s+(\S+),\s+id\s+=\s+(\d+),\s+inst\s+=\s+(\d+).\s+(\S+)/) {
            $self->{modules}{$1}{id} = $2;
            $self->{modules}{$1}{instance} = $3;
            $self->{modules}{$1}{state} = $4;
        }
    }

    # Private variable to determine if connections should persist when using the helper
    # subroutines console and telnet.
    $self->{persist} = 0;

    return $self;
}

##########################################################################

=head2 C<get_modinst()>

=over 4

=item * Description

Gets the module and/or instance identifiers for the matching name. If
more than one instance of a module is found, only the module is returned.

=item * Arguments

=over 4

=item 1 Name
 (Required)

The string name to match on.

=back

=item * Returns

A two item array with the module ID in the first element and the instance
ID in the second element.

=back

=cut

##########################################################################
sub get_modinst {
    my ($self, $name) = @_;
    my ($id, $inst);

    foreach my $module (keys %{$self->{modules}}) {
        if ($module =~ /$name/) {
            if (!defined($id)) {
                $id = $self->{modules}{$module}{id};
                $inst = $self->{modules}{$module}{instance};
            } elsif ($id == $self->{modules}{$module}{id}) {
                # There must be more than one instance of the same module
                $inst = undef;
            } else {
                $self->_set_err("Controller::get_modinst multiple module ids match given name");
                return undef;
            }
        }
    }

    return ($id, $inst);
}

##########################################################################

=head2 C<get_interfaces()>

=over 4

=item * Description

Gets all the target interfaces on the controller including their FCID,
WWPN, and WWNN.

=item * Returns

Address to a hash containing each interface as a key to the hash. Sub keys
include state, fcid, wwpn, and wwnn.

=back

=cut

##########################################################################
sub get_interfaces {
    my ($self) = @_;
    my ($status, @output);
    my $hash = {};

    (undef, $status, @output) = $self->execute(-command => 'fcp config');
    for (my $i=0; $i<@output; $i++) {
        # execute munges with the < and > symbols and replaces them with ( and ).
        if ($output[$i] =~ /(\S+):\s+(.*)\s+[\(<]/) {
            my $adapter = $1;
            $hash->{$adapter}{state} = $2;
            $hash->{$adapter}{fcid} = (split(/\s+/,$output[$i+1]))[3];
            $hash->{$adapter}{wwpn} = (split(/\s+/,$output[$i+2]))[2];
            $hash->{$adapter}{wwnn} = (split(/\s+/,$output[$i+2]))[4];

        }
    }

    return $hash;
}

sub get_interface {
    my ($self, $interface) = @_;
    my $hash = {};

    my (undef, $status, @output) = $self->execute(-command => "fcp config $interface");
    for (my $i=0; $i<@output; $i++) {
        # execute munges with the < and > symbols and replaces them with ( and ).
        if ($output[$i] =~ /\S+:\s+(.*)\s+[\(<]/) {
            $hash->{state} = $1;
            $hash->{fcid} = (split(/\s+/,$output[$i+1]))[3];
            $hash->{wwpn} = (split(/\s+/,$output[$i+2]))[2];
            $hash->{wwnn} = (split(/\s+/,$output[$i+2]))[4];

        }
    }

    return $hash;
}

##########################################################################

=head2 C<get_adapter_family()>

=over 4

=item * Description

Gets the target adapter's Qlogic family number.

=item * Arguments

=over 4

=item 1 Interface
 (Required)

The target interface name.

=back

=item * Returns

The adapter's Qlogic family number.

=back

=cut

##########################################################################
sub get_adapter_family {
    my ($self, $interface) = @_;
    my ($status, @output, $family);

    (undef, $status, @output) = $self->execute(-command => "fcp show adapter $interface");

    #Description:             Fibre Channel Target Adapter 0a (Dual-channel, QLogic 2432 (2462) rev. 2)
    foreach (@output) {
        if ($_ =~ /\((\d+)\)/) {
            $family = $1;  
            last;
        }
    }

    return $family;
}
##########################################################################

=head2 C<get_firmware_files()>

=over 4

=item * Description

Gets a list of the firmware files in ONTAP's /etc/firmware directory that
match the supplied name string.

=item * Arguments

=over 4

=item 1 Name
 (Required)

The string name to match on.

=back

=item * Returns

An array of matching files.

=back

=cut

##########################################################################
sub get_firmware_files {
    my ($self, $family) = @_;
    my ($status, @output, @files);

    (undef, $status, @output) = $self->execute(-command => 'ls /etc/firmware');
    $family = "2532" if ($family eq "2562");

    @files = grep(/$family/, @output);

    return @files;
}

sub is_cf_enabled {
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'cf status');
    if ($output =~ /Cluster enabled/) {
        return $Controller::TRUE;
    } else {
        return $Controller::FALSE;
    }
}

sub is_cf_takeover {
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'cf status');
    if ($output =~ /has taken over/) {
        return $Controller::TRUE;
    } else {
        return $Controller::FALSE;
    }
}

sub cf_enable{
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'cf enable');
    return $Controller::TRUE;
}

sub takeover {
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'cf takeover');
    if ($output =~ /takeover initiated/) {
        return $Controller::TRUE;
    } else {
        return $Controller::FALSE;
    }
}

sub giveback {
    my ($self, %args) = @_;
    my ($status, $output);
    my $force = "";

    foreach (keys %args) {
        if (/-?force/) {
            length($force = $args{$_}) or
                do {
                    $self->_set_err("Controller::giveback invalid value for \"force\":",
                            "$args{$_}");
                    return $self;
                };
        }
    }

    if ($force) {
	$force = "-f";
    }

    ($output, $status) = $self->execute(-command => "cf giveback $force");
    # output doesn't appear to come back through
    #if ($output =~ /giveback initiated/) {
        return $Controller::TRUE;
    #} else {
    #    return $Controller::FALSE;
    #}
}

sub is_fcp_running{
	my ($self) = @_;
	my ($status, $output);

	($output, $status) = $self->execute(-command => 'fcp status');
	if ($output =~ /FCP service is running/) {
		return $Controller::TRUE;
	} else {
		return $controller::FALSE;
	}
}

sub fcp_stop {
    my ($self) = @_;
    my ($status, $output);
    
    ($output, $status) = $self->execute(-command => 'fcp stop');
    if ($output =~ /FCP service stopped/) {
        return $Controller::TRUE;
    } else {
        return $Controller::FALSE;
    }
}

sub iscsi_stop {
    my ($self) = @_;
    my ($status, $output);
    
    ($output, $status) = $self->execute(-command => 'iscsi stop');
    if ($output =~ /FCP service stopped/) {
        return $Controller::TRUE;
    } else {
        return $Controller::FALSE;
    }
}

sub fcp_start {
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'fcp start');
    if ($output =~ /Partner WWNN must match local WWNN to start FCP/) {
        $self->_set_err("Controller::fcp_start() error while starting fcp service: $output");
        return $Controller::FALSE;
    }
    return $Controller::TRUE;
}

sub wwnn {
    my ($self, $wwnn) = @_;
    my ($status, $output);

    if (defined($wwnn)) {
        ($output, $status) = $self->execute(-command => "fcp nodename $wwnn");
    } else {
        ($output, $status) = $self->execute(-command => "fcp nodename");
        $output =~ /\((\w+)\)/;
        return $1;
    }
}

sub wwpn {
    my ($self, $adapter, $wwpn) = @_;
    my ($status, @output);

    if (!defined($adapter)) {
    	$self->_set_err("Controller::wwpn() adapter not specified");
	return "";
    }
	
    if (defined($wwpn)) {
    	# The fcp portname set command prompts for confirmation
	$self->{persist} = 1;
     	$self->console("fcp portname set $adapter $wwpn", -noprompt, -prompt, "Do you want to continue (y/n)?");
	$self->{persist} = 0;
	$self->console("y", -noprompt, -prompt, "Change will take effect when adapter is onlined.");
	return $wwpn;
    } else {
        (undef, $status, @output) = $self->execute(-command => "fcp portname show");
	foreach my $line (@output) {
		if ($line =~ /(\S+)\s+$adapter/) {
			$wwpn = $1;
			$wwpn =~ s/://g;
			return $wwpn;
		}
	}
	$self->_set_err("Controller::wwpn() unable to find matching adapter");
    }

    return "";
}

sub get_unused_wwpns {
	my ($self) = @_;
	my (@wwpns) = ();

	my (undef, $status, @output) = $self->execute(-command => "fcp portname show -v");
	#50:0a:09:83:8a:d2:ad:c2   unused
	foreach my $line (@output) {
		if ($line =~ /unused/i) {
			my @temp = split(/\s+/, $line);
			push (@wwpns, $temp[0]);
		}
	}

	return @wwpns;
}

sub wwpn_swap {
	my ($self, $adapter1, $adapter2) = @_;

	if (!defined($adapter1) || !defined($adapter2)) {
		$self->_set_err("Controller::wwpn_swap() two adapters must be specified");
		return $Controller::FALSE;
	}

	$self->execute(-command => "fcp portname swap $adapter1 $adapter2");
	if ($self->error) {
		return $Controller::FALSE;
	}
}

sub iscsi_start {
    my ($self) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command => 'iscsi start');
    return $Controller::TRUE;
}

sub options {
	my ($self, $option, $value) = @_;

	if (defined($value)) {
		$self->execute(-command => "options $option $value");
	} else {
		$value = $self->execute(-command => "options $option");
	}
	return $value;
}

sub dump_coverage_data {
	my ($self, $filename) = @_;

	my (undef, $status) = $self->execute(-command => "prof arcdump /$filename");
	if ($status) {
		return $Controller::FALSE;
	} else {
		return $Controller::TRUE;
	}
}

sub create_igroup {
    my ($self, %args) = @_;
    my ($name, $ostype, $initiators, $type);
    my ($status, $output);

    $type = "fcp";

    foreach (keys %args) {
        if (/-?name/) {
            length($name = $args{$_}) or
                do {
                    $self->_set_err("Controller::create_igroup invalid value for \"name\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };  
        } elsif (/-?ostype/) {
            length($ostype = $args{$_}) or
                do {
                    $self->_set_err("Controller::create_igroup invalid value for \"ostype\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?initiators/) {
            length($initiators = $args{$_}) or
                do {
                    $self->_set_err("Controller::create_igroup invalid value for \"initiators\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?type/) {
            length($type = $args{$_}) or
                do {
                    $self->_set_err("Controller::create_igroup invalid value for \"type\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        }
    }

    if ($type eq "fcp") {
        $type = "-f";
    } elsif ($type eq "iscsi") {
        $type = "-i";
    } else {
        $self->_set_err("Controller::create_igroup invalid type ($type) specified");
        return ($Controller::FALSE);
    }

    ($output, $status) = $self->execute(command => "igroup create $type -t $ostype $name $initiators");
    if ($status != 0) {
        $self->_set_err("Controller:create_igroup error while creating igroup: $output");
        return ($Controller::FALSE);
    }

    return $Controller::TRUE;
}

sub setup_luns {
    my ($self, %args) = @_;
    my ($volume_count, $volume_size, $lun_count, $lun_size, $volume_prefix, $lun_prefix, $lun_id, $automap);
    my ($igroup, $ostype, $aggr);
    my ($output, $status, @output);

    # Default values
    $volume_count = 1;
    $volume_size = "512m";
    $lun_count = 1;
    $lun_size = "256m";
    $volume_prefix = "vol";
    $lun_prefix = "lun";
    $lun_id = 0;
    $automap = 1;
    $ostype = "linux";

    foreach (keys %args) {
        if (/-?volume_count/) {
            length($volume_count = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"volume_count\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?volume_size/) {
            length($volume_size = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"volume_size\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?lun_count/) {
            length($lun_count = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"lun_count\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?lun_size/) {
            length($lun_size = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"lun_size\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?volume_prefix/) {
            length($volume_prefix = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"volume_prefix\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?lun_prefix/) {
            length($lun_prefix = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"lun_prefix\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?automap/) {
            length($automap = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"automap\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?lun_id/) {
            length($lun_id = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"lun_id\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?igroup/) {
            length($igroup = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"igroup\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?ostype/) {
            length($ostype = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"ostype\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } elsif (/-?aggr/) {
            length($aggr = $args{$_}) or
                do {
                    $self->_set_err("Controller::setup_luns invalid value for \"aggr\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        } 
    }

    if (!defined($igroup)) {
        $self->_set_err("Controller::setup_luns mandatory paramter \"igroup\" not set");
        return ($Controller::FALSE);
    }

    if (!defined($aggr)) {
        # Look for an available aggregate to create volumes on
        (undef, $status, @output) = $self->execute(-command => "aggr status");
        foreach my $line (@output) {
            if ($line =~ /root/) {
                next;
            }
            if ($line =~ /(\S+)\s+online/) {
                $aggr = $1;
            }
        }
        if (!defined($aggr)) {
            $self->_set_err("Controller::setup_luns unable to determine aggregate");
            return ($Controller::FALSE);
        }
    }

    if (!defined($ostype)) {
	($output, $status) = $self->execute(-command => "igroup show $igroup");
	if ($output =~ /\(ostype:\s+(\S+)\)/) {
	    $ostype = $1;
	} else {
	    $self->_set_err("Controller::setup_luns unable to determine igroup ostype");
	    return ($Controller::FALSE);
	}
    }

    for (my $v=0; $v<$volume_count; $v++) {
        my $volume = "${igroup}_${volume_prefix}${v}";
        ($output, $status) = $self->execute(-command => "vol create $volume $aggr $volume_size");
#        if ($status) {
        if (($status != 0) && ($status != 255)) {	# FIXME
            $self->_set_err("Controller::setup_luns error while creating volume $volume on $aggr with size $volume_size: $output");
            return ($Controller::FALSE);
        }
        for (my $l=0; $l<$lun_count; $l++) {
            my $lun = "/vol/${volume}/${igroup}_${lun_prefix}${l}";
            ($output, $status) = $self->execute(-command => "lun create -s $lun_size -t $ostype $lun");
            if ($status != 0) {
                $self->_set_err("Controller::setup_luns error while creating lun $lun in volume $volume with size $lun_size: $output");
                return ($Controller::FALSE);
            }

            my $cmd = "lun map $lun $igroup";
	    if ($automap == 0) {
                $cmd .= " $lun_id";
            }
            ($output, $status) = $self->execute(-command => $cmd);
            if ($status != 0) {
                $self->_set_err("Controller::setup_luns error while mapping lun $lun to igroup $igroup: $output");
                return ($Controller::FALSE);
            }
        }
    }
}
##########################################################################

=head2 C<get_interface_state()>

=over 4

=item * Description

Gets the state of the supplied interface.

=item * Arguments

=over 4

=item 1 Interface
 (Required)

The interface to get the state of.

=back

=item * Returns

A string with the interface status.

=back

=cut

##########################################################################
sub get_interface_state {
    my ($self, $port) = @_;
    my ($status, $output);

    ($output, $status) = $self->execute(-command=> "fcp config $port");

    if ($output =~ /$port:\s+(.*)\s+</) {
        return $1;
    } else {
        return undef;
    }
}

##########################################################################

=head2 C<set_interface()>

=over 4

=item * Description

Sets speed and/or port type of a specified target interface.

=item * Arguments Hash

=over 4

=item 1 -interface
 (Required)

The interface to operate on.

=item 2 -speed
 (Optional)

The speed to set the interface to. This must be a digit (1, 2, or 4) or 'auto'.

=item 3 -type
 (Optional)

The type of port to set the interface to. This can be ptp, loop, or auto.

=back

=item * Returns

$Controller::TRUE If the interface settings were successful.
$Controller::FALSE If he parameters provided were incorrect.

=back

=cut

##########################################################################
sub set_interface {
    my ($self, %args) = @_;
    my ($interface, $speed, $type, $state);

    foreach (keys %args) {
        if (/-?interface/) {
            length($interface = $args{$_}) or
                do {
                    $self->_set_err("Controller:set_interface invalid value for \"interface\":",
                        "$args{$_}");
                    return ($Controller::FALSE);
                };
        }
        if (/-?speed/) {
            if ( (not length($speed = $args{$_})) && ($args{$_} !~ /^\d$/ || $speed ne 'auto') ) {
                $self->_set_err("Controller:set_interface invalid value for \"speed\":",
                        "$args{$_}");
                return ($Controller::FALSE);
            }
        }
        if (/-?type/) {
            if ( (not length($type = $args{$_})) && $args{$_} !~ /^(ptp|auto|loop)$/ ) {
                $self->_set_err("Controller:set_interface invalid value for \"type\":",
                        "$args{$_}");
                return ($Controller::FALSE);
            }
        }
    }

    if (!defined($interface)) {
        $self->_set_err("Controller:set_interface mandatory paramter interface not provided");
        return ($Controller::FALSE);
    }

    # Bring the interface down first
    $self->execute(-command => "fcp config $interface down");
    
    # Set the speed to the desired value
    if (defined($speed)) {
        $self->execute(-command => "fcp config $interface speed $speed");
    }

    # Set the type to the desired value
    if (defined($type)) {
        $self->execute(-command => "fcp config $interface mediatype $type");
    }

    # Bring the interface back up
    $self->execute(-command => "fcp config $interface up");
    
    return ($Controller::TRUE);
}

##########################################################################

=head2 C<rastrace_clear()>

=over 4

=item * Description

Clears the RAS trace buffer for the specified module and/or instance.

=item * Arguments Hash

=over 4

=item 1 -module
 (Required)

The module ID to clear.

=item 2 -interface
 (Optional)

The interface ID to clear.

=back

=item * Returns

$Controller::TRUE If the clear was successful.
$Controller::FALSE If the clear was not successful.

=back

=cut

##########################################################################
sub rastrace_clear {
    my ($self, %args) = @_;
    my ($module, $instance, $clearcmd, $showcmd, $count);
    my ($output, $status);
    
    foreach (keys %args) {
        if (/-?module/) {
            length($module = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_clear invalid value for \"module\":",
                        "$args{$_}");
                    return $Controller::FALSE;
                };
        }
        if (/-?instance/) {
            length($instance = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_clear invalid value for \"instance\":",
                        "$args{$_}");
                    return $Controller::FALSE;
                };
        }
    }
    if (!defined($module)) {
        $self->_set_err("Controller::rastrace_clear mandatory parameter module not specified");
        return $Controller::FALSE;
    }
    if ($self->{priv} !~ /(diag|test)/) {
        $self->_set_err("Controller::rastrace_clear privilege needs to be set to at least diag");
        return $Controller::FALSE;
    }
    $clearcmd = "rastrace clear -m $module";
    $showcmd = "rastrace show -m $module";
    if (defined($instance)) {
        $clearcmd .= " -i $instance"; 
        $showcmd .= " -i $instance"; 
    }

    $count = 0;
    $output = "not empty";
    
    my $oldpriv = $self->{priv};
    $self->{priv} = "test";

    while ($output !~ /^\s*$/ && $count < 10) {
        ($output, $status) = $self->execute(-command => $clearcmd);
        ($output, $status) = $self->execute(-command => $clearcmd);
        ($output, $status) = $self->execute(-command => $showcmd);
    }

    $self->{priv} = $oldpriv;

    if ($count == 10) {
        return $Controller::FALSE;
    } else {
        return $Controller::TRUE;
    }
}

sub rastrace_dump {
    my ($self, %args) = @_;
    my ($module, $instance);
    foreach (keys %args) {
        if (/-?module/) {
            length($module = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_dump() invalid value for \"module\":",
                        "$args{$_}");
                    return $Controller::FALSE;
                };
        }
        if (/-?instance/) {
            length($instance = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_dump() invalid value for \"instance\":",
                        "$args{$_}");
                    return $Controller::FALSE;
                };
        }
    }

    if (!defined($module)) {
        $self->_set_err("Controller::rastrace_dump() mandatory parameter module not specified");
        return $Controller::FALSE;
    }
    if ($self->{priv} !~ /(diag|test)/) {
        $self->_set_err("Controller::rastrace_dump() privilege needs to be set to at least diag");
        return $Controller::FALSE;
    }
    
    my $cmd = "rastrace dump -m $module";
    if (defined($instance)) {
	$cmd .= " -i $instance";
    }

    $self->execute(-command => $cmd);
    if ($self->error) {
	return $Controller::FALSE;
    } else {
	return $Controller::TRUE;
    }
}

##########################################################################

=head2 C<get_rastrace_buffer()>

=over 4

=item * Description

Gets all entries in the RAS trace buffer for the specified module and/or instance.

=item * Arguments Hash

=over 4

=item 1 -module
 (Required)

The module ID to retrieve.

=item 2 -interface
 (Optional)

The interface ID to retrieve.

=back

=item * Returns

An array of RAS entries with each entry being a hash with keys number,
timestamp, type, label, and description.

=back

=cut

##########################################################################
sub get_rastrace_buffer {
    my ($self, %args) = @_;
    my ($module, $instance, $number, $cmd, $old_priv);
    my (@ras_array);
    
    foreach (keys %args) {
        if (/-?module/) {
            length($module = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_clear invalid value for \"module\":",
                        "$args{$_}");
                    return;
                };
        }
        if (/-?instance/) {
            length($instance = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_clear invalid value for \"instance\":",
                        "$args{$_}");
                    return;
                };
        }
        if (/-?number/) {
            length($number = $args{$_}) or
                do {
                    $self->_set_err("Controller:rastrace_clear invalid value for \"number\":",
                        "$args{$_}");
                    return;
                };
        }
    }

    if (!defined($module)) {
        $self->_set_err("Controller::get_rastrace_buffer mandatory parameter module not specified");
        return;
    }
    if ($self->{priv} !~ /test/) {
	$old_priv = $self->{priv};
	$self->{priv} = "test";
        #$self->_set_err("Controller::get_rastrace_buffer privilege needs to be set to test");
        #return;
    }

    # Build command and append optional arguments
    $cmd = "rastrace show -m $module";
    if (defined($instance)) {
        $cmd .= " -i $instance"; 
    }
    if (defined($number)) {
        $cmd .= " -n $number";
    }
    my $oldpriv = $self->{priv};
    $self->{priv} = "test";

    my (undef, $status, @loutput) = $self->execute(-command => $cmd);
   
    $self->{priv} = $oldpriv;

    if (!$self->error()) { 
        # Parse output and fill ras_entry hash structure then store it in an array
        my $component;
        for (my $i=0; $i<@loutput; $i++) {
            if ($loutput[$i] =~ /=====/) {
                $component = (split(/,/,$loutput[$i+1]))[0];
                $i += 2;
                next;
            }
            if ($loutput[$i] =~ /TRACE/) {
                my %ras_entry;
                my @temp = split(/\s+/, $loutput[$i]);
                $ras_entry{number} = $temp[1];
                $ras_entry{number} =~ s/#//;
                $ras_entry{timestamp} = $temp[2];
                $ras_entry{type} = $temp[3];
                $ras_entry{label} = $loutput[$i+1];
                $ras_entry{description} = $loutput[$i+2];
                push (@ras_array, \%ras_entry);
                $i+=3;
                next;
            }
        }   
    }

    if (defined($old_priv)) {
	$self->{priv} = $old_priv;
    }

    return @ras_array;

}

##########################################################################

=head2 C<get_root_volname()>

=over 4

=item * Description

Gets the root volume name on the controller

=item * Returns

The root volume name string. Caller should check for errors in the object
in the event the root volume name could not be determined.

=back

=cut

##########################################################################
sub get_root_volname {
    my ($self) = @_;
    my ($rootvol);
    my (undef, $status, @loutput) = $self->execute(-command => 'vol status');

    foreach (@loutput) {
        if ($_ =~ /root/ && $_ =~ /^\s+(\S+)/) {
            $rootvol = $1;
            last;
        }
    }

    if (!defined($rootvol)) {
        $self->_set_err("Controller::get_root_volname unable to determine the root volume name");
    }

    return $rootvol;
}

##########################################################################

=head2 C<execute()>

=over 4

=item * Description

Executes a command to the controller.

=item * Arguments Hash

=over 4

=item 1 -command
 (Required)

The command string to execute.

=back

=item * Returns

A three element array with the first element being the output from the
command as a concatenated string, the second element being the status of
the command, and the third element being the output as an array.

=back

=cut

##########################################################################
sub execute {
    my ($self, %args) = @_;
    my ($output, $status, @l_output);
    my ($command, $conn_str);
    foreach (keys %args) {

        if (/-?command/) {
            length($command = $args{$_}) or
                do {
                    $self->_set_err("Controller:execute invalid value for \"command\":",
                        "$args{$_}");
                    return;
                };
        }
    }
   
    $conn_str = $self->{hostname} . "#" . $self->{conn_type};

    if ($self->{conn_type} eq "zapi") {
        ($output, $status, @l_output) = &SanUtil::system_zapi($self->{hostname}, $command, $self->{priv});
        # Remap status to be more inline with ONTAP's CLI status
        # 0 = fail
        # -1 = zapi failure
        # 1 = pass
        #if ($status == -1) {
        #    # Arbitrarily assign a value of 2 for zapi failure
        #    $status = 2;
        #} elsif ($status == 0) {
        #    # Failures should be non zero
        #    $status = 1;
        #} elsif ($status == 1) {
        #    # Successes should be zero
        #    $status = 0;
        #}
    } else {
        my $connect = Tharn::connect($conn_str.";timeout=$self->{timeout}",
                -eval=>sub{
                if (@_) {
                my $errmsg = join("",@_);
                $errmsg =~ s/\n//g;
                $self->_set_err($errmsg);
                $Tharn::InEND=1; #Don't abort
                Tharn::logresult("CONFIG","\n@_");
                return;
                }
                });
        if (!defined($connect)) {
            $self->_set_err("Controller:execute unable to connect to controller $self->{hostname} using $self->{conn_type}");
            $status = 1;
        } else {
            $command = "priv set -q $self->{priv};" . $command . ";result";
            my (undef,undef,$buffer,undef) = Tharn::say($connect,$command,-timeout=>$self->{timeout},
                    -eval=>sub{
                    if (@_) {
                    my $errmsg = join("",@_);
                    $errmsg =~ s/\n//g;
                    print NTEST_STDOUT "$errmsg\n";
                    Tharn::logresult("CONFIG",abort=>0,"\n@_");
                    }
                    });

            Tharn::disconnect($connect);

            # Split the buffer into array     
            @l_output = split(/\n/,$buffer);

            # Remove result value from last line
            $status = pop(@l_output);
            $output = join(//,@l_output);
        }
    }

    return ($output, $status, @l_output);
}

sub clear_stats {
	my ($self, $adapter) = @_;
	$self->execute(-command => "fcp stats -z $adapter");
	if ($self->error) {
		$self->_set_err("Controller::clear_stats() error enounted while issuing clear stats command: ",
				$self->errmsg);
		return $Controller::FALSE;
	} else {
		return $Controller::TRUE;
	}
}

sub get_stats {
	my ($self, $adapter) = @_;
	my (@output, %hash);

	(undef, undef, @output) = $self->execute(-command => "fcp stats $adapter");
	$adapter = undef;
	foreach my $line (@output) {
		if ($line =~ /Message:\s+sent\s+completed\s+received/ ||
		    $line =~ /Statistics for SFP/ ||
		    $line =~ /Statistics for SFF/) {
			last;
		} elsif ($line =~ /Statistics for adapter (\S+)/) {
			$adapter = $1;
		} elsif ($line =~ /(.+):\s+(\S+)/) {
			$hash{$adapter}{$1} = $2;
		}
	}

	return \%hash;
}

sub console {
    my ($self, $cmd, @passthrough) = @_;
    my $conn_str = $self->{hostname} . "#console";
    my $connect = Tharn::connect($conn_str);

    my ($match, $before, $after, $prompt) = &say($connect, $cmd, -swallow, @passthrough);

    if (!$self->{persist}) {
        Tharn::disconnect($connect);
    }
    return ($match, $before, $after, $prompt);
}

sub telnet {
    my ($self, $cmd, @passthrough) = @_;
    my $conn_str = $self->{hostname} . "#telnet";
    my $connect = Tharn::connect($conn_str);

    my ($match, $before, $after, $prompt) = &say($connect, $cmd, -swallow, @passthrough);

    if (!$self->{persist}) {
        Tharn::disconnect($connect);
    }

    return ($match, $before, $after, $prompt);
}

sub test_point {
    my ($self, %args) = @_;

    my ($module, $tp);
    my $hitcount = 1;
    my $param = 1;

    foreach (keys %args) {
        if (/-?module/) {
            length($module = $args{$_}) or
                do {
                    $self->_set_err("Controller:execute invalid value for \"module\":",
                        "$args{$_}");
                    return;
                };
        } elsif (/-?tp/) {
            length($tp = $args{$_}) or
                do {
                    $self->_set_err("Controller:execute invalid value for \"tp\":",
                        "$args{$_}");
                    return;
                };
        } elsif (/-?hitcount/) {
            length($hitcount = $args{$_}) or
                do {
                    $self->_set_err("Controller:execute invalid value for \"hitcount\":",
                        "$args{$_}");
                    return;
                };
        } elsif (/-?param/) {
            length($param = $args{$_}) or
                do {
                    $self->_set_err("Controller:execute invalid value for \"param\":",
                        "$args{$_}");
                    return;
                };
        }
    }

    if (!defined($module) || !defined($tp)) {
	$self->_set_err("Controller::test_point() module and tp are required parameters");
	return $Controller::FALSE;
    }

    $self->execute(-command => "tp error $module $tp $hitcount $param");
    if ($self->error) {
	return $Controller::FALSE;
    } else {
	return $Controller::TRUE;
    }
}

=pod

=head1 AUTHOR

Evan Chiu (evan@netapp.com)

=head1 COPYRIGHT

Copyright (c) 2008 NetApp, Inc. All rights reserved.

=cut

1;
