#
# $Id: //depot/prod/test/Rsierranevada/lib/Compression/UI/CLI/Formats/Type2.pm#2 $
#

## @summary Utility module for parsing formatted CLI output of type 1.
##
## @description
## This module provides functionality for parsing of CLI command output
## that is expected to be in a 'list-like' formatted form.
## For example, part of the output of the 'vol compress status -l' command:
## Checkpoint info of volume vl20494_68c9_54cf28c0:
## Checkpoint magic : 0x434f4f50
## Checkpoint version : 1
## Compressing blocks shared with snapshots : yes
## Compressing de-duplicated blocks : yes
## Inode number to restore from : 96
## Fbn number to restore from : 262650
## Snapshot skipped data (KB) : 0
## Shared skipped data (KB) : 0
## Total data scanned (KB) : 1050600
## Checkpoint creation time : Wed Jan 20 12:42:15 PST 2010
##
## @synopsis
## N/A: To be only used internally by the Compression library.
## 
## @author voicu@netapp.com, dl-compression-qa@netapp.com
## @status private (in-progress)



package Compression::UI::CLI::Formats::Type2;

use strict;
use warnings;

use Tharn;
use Data::Dumper;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF BOOLEAN UNDEF OBJECT);
use Compression::Constants qw(:all);


BEGIN
{
	require Exporter;
	our @ISA = qw(Exporter);
}



# @arg version Required: Compression::Version object for the hosting filer.
# @arg boot_mode Optional: Boot mode of the hosting filer.
# Defaults to 7M.
# @arg debug Optional: Controls verbosity of this module.
# Defaults to FALSE.
#
# @author voicu@netapp.com, dl-compression-qa@netapp.com
# @status private (in-progress)
#

sub new
{
	my $class = shift;
        my %args = validate(@_,
	{
                version => {type => OBJECT,
                            isa  => "Compression::Version",
                            },
                
                boot_mode => {type => SCALAR,
                              optional => 1,
                              default  => C_ONTAP_BOOT_MODE,
                              },
                
                debug  => {  
			type 	 => SCALAR,
			optional => 1,
                        default  => FALSE,
			},
	} );
        
	my $self = {};
        
        $self->{_debug} = $args{debug};
        $self->{_version} = $args{version};
        $self->{_boot_mode} = $args{boot_mode};

	logcomment("Created new Formats Type2 object:\n" . Dumper($self)) if ($self->{_debug});
	bless $self, $class;

	return $self;
}


sub _process_output
{
        my $class = shift;
	my %args = validate(@_,
	{
                code => { type => SCALAR, },
                output => { type => SCALAR, },
                allow_multiple_formats => { type => BOOLEAN, },
                modifier => {type => SCALAR|UNDEF, optional => 1,},
                required_return_hash_key => {type => SCALAR|UNDEF, optional => 1,},
                disable_data_identifier_processing => {type => BOOLEAN, optional => 1, default => FALSE},
                variable_format => {type => SCALAR, optional => 1, default => FALSE},
	} );
        
        my $code = $args{code};
        my $output = $args{output};
        my $allow_multiple_formats = $args{allow_multiple_formats};
        # This is used for sanity checks. If passed in,
        # we will check that the return hash contains this key.
        # This check is useful for commands for a specific entity (for example, volume),
        # when we want to make sure the output was for that specific entity
        # For this format type, it is expected to be found in the header
        my $required_return_key = $args{required_return_hash_key};
        # Not used 
        #my $modifier = $args{modifier};
        #my $disable_data_identifier_processing = $args{disable_data_identifier_processing}
        my $found_required_return_key = 1;
        if (defined($required_return_key))
        {
                $found_required_return_key = 0;
        }

        my %return_hash = (rc => C_RC_FAILURE);
        
        if ($output eq '')
        {
                logresult('fatal', "Formats Type 2: Empty output found when formatted output expected");
        }
        my @lines = split(/\n/, $output);
        my @unprocessed_lines = @lines;
        logcomment("Formats Type 2: Starting processing output:\n" . Dumper(@lines)) if ($class->{_debug});
        
        # The logic is the following:
        # We will process the output in chunks of lines, as defined in the format hash for this type
        # (for example, for a specific code, if the entry hash has one header and 5 expected_info entries,
        # we will process series of 6 lines from the output).
        # We will allow the output to be for multiple entries (ex: volumes), as long as the header has entry-related info
        # (ex: 'Checkpoint info of volume vl20494_68c9_54cf28c0:')
        # If required_return_hash_key is defined, we will continue processing the entire output,
        # to ensure that any format for an extra entity stays within the definition.
        # But, at the end, we will only add that entry to the return hash (also checking that we have such an entry)
        my $allow_info_for_multiple_entities = FALSE;
        my $looped_at_least_once = 0;
        
        while (@lines)
        {
                my %process = $class->_process_output_by_set_of_lines(lines => \@lines,
                                                                      unprocessed_lines => \@unprocessed_lines,
                                                                      code => $code,
                                                                      );
                
                logcomment("Formats Type 2: _process_output:\nlines:\n" . Dumper(@lines) .
                           "\n\nunprocessed_lines:\n" . Dumper(@unprocessed_lines) .
                           "\n\nReturned hash:\n" . Dumper(%process)) if ($class->{_debug});
                
                # Delete internally-used hash entries
                my $rc = $process{rc};
                my $found_entry_info_in_header = $process{found_entry_info_in_header};
                delete($process{rc});
                delete($process{found_entry_info_in_header});
                
                if ($rc != C_RC_SUCCESS)
                {
                        logcomment("Formats Type 2: _process_output: Detected unrecognized output") if ($class->{_debug});
                        last;
                }
                
                my $add_processed_entry_to_return_hash = TRUE;
                
                if ($found_entry_info_in_header == TRUE)
                {
                        if (($allow_info_for_multiple_entities == FALSE) && ($looped_at_least_once == 1))
                        {
                                # We already looped at least once and did not find entry info before
                                # The output changed...
                                logresult('fatal', "Detected inconsistent format of type 2 for code $code: \n" .
                                                   "Header format changed from \"without entry info\" to \"with entry info\"\n" .
                                                   "Output:\n$output");
                        }
                        $allow_info_for_multiple_entities = TRUE;
                        
                        if (defined($required_return_key))
                        {
                                # In this case, the returned hash has all the info nested in only one hash entry
                                if (defined($process{$required_return_key}))
                                {
                                        $found_required_return_key = 1;
                                }
                                else
                                {
                                        # Legal format, but not the info for our required entity
                                        $add_processed_entry_to_return_hash = FALSE;
                                }
                        }
                }
                else
                {
                        if ($allow_info_for_multiple_entities == TRUE)
                        {
                                # We already looped at least once and found entry info every time
                                # The output changed...
                                logresult('fatal', "Detected inconsistent format of type 2 for code $code: \n" .
                                                   "Header format changed from \"with entry info\" to \"without entry info\"\n" .
                                                   "Output:\n$output");
                        }
                        
                        if (defined($required_return_key))
                        {
                                logresult('fatal', "Unexpected output format for code $code:\n" .
                                                   "Failed to detect entity-level info in the header, but it is required: $required_return_key\n" .
                                                   "Output:\n$output");
                        }
                }
                
                if ($add_processed_entry_to_return_hash == TRUE)
                {
                        foreach my $key (keys(%process))
                        {
                                $return_hash{$key} = $process{$key};
                        }
                }
                
                $looped_at_least_once = 1;
        }
        
        if ((scalar(@unprocessed_lines) > 0) && ($allow_multiple_formats == FALSE))
        {
                # Exit with failure, to allow detecting of other eventual error messages.
                logcomment("Formats Type 2: Unexpected output format for code $code: \n" .
                           "Multiple formats are not allowed. Expected format of type 2.\n" .
                           "Extra output remaining after processing:\n" . Dumper(@unprocessed_lines) . "\n\n" .
                           "Output:\n$output") if ($class->{_debug});
                return %return_hash;
        }

        if ($found_required_return_key == 0)
        {
                # Exit with failure, to allow detecting of other eventual error messages.
                logcomment("Formats Type 2: Unexpected output format for code $code: \n" .
                           "Required return key not found or the information was found incomplete.\n" .
                           "Extra output remaining after processing:\n" . Dumper(@unprocessed_lines) .
                           "Returning:\n" . Dumper(%return_hash));
                return %return_hash;
        }
        
        $return_hash{rc} = C_RC_SUCCESS;
        $return_hash{unprocessed_output} = join("\n", @unprocessed_lines); 

        logcomment("Formats Type 2: _process_output returning:\n" . Dumper(%return_hash)) if ($class->{_debug});
        return %return_hash;
}

sub _process_output_by_set_of_lines
{
        my $class = shift;
	my %args = validate(@_,
	{
                code => { type => SCALAR, },
                lines => { type => ARRAYREF, },
                unprocessed_lines => { type => ARRAYREF, },
	} );
        
        my $code = $args{code};
        my $lines = $args{lines};
        my $unprocessed_lines = $args{unprocessed_lines};
        
        my %return_hash = (rc => C_RC_FAILURE,
                           found_entry_info_in_header => FALSE);
        
        my $format = $class->_get_format_info(code => $code);
        logcomment("Formats Type 2:\n" .
                   "Code $code. Processing:\n" . Dumper($lines) .
                   "\nExpected format info\n:" . Dumper($format)) if ($class->{_debug});
        
        my $header = $format->{header};
        my $ignore_lines = (exists($format->{ignore_lines})) ? $format->{ignore_lines} : [];
        my $return_hash_key;
        
        # Look for header first, if asked
        if (defined($header))
        {
                my $header_line = shift(@$lines);
                my $header_suffix = $format->{header_suffix};
                logcomment("Formats Type 2: looking for header:$header\n" .
                           "Processing line: $header_line") if ($class->{_debug});
                
                # Skip pre-header lines, if they are in the 'ignore lines'
                my $check_for_skip_lines = 1;
                while (($check_for_skip_lines == 1) && (defined($header_line)) && ($header_line !~ /^$header/i))
                {
                        $check_for_skip_lines = 0;
                        foreach my $ignore_pattern (@$ignore_lines)
                        {
                                logcomment("Formats Type 2: code $code. Checking if (pre-header) line should be ignored:\n$header_line\n" .
                                           "Trying to match: $ignore_pattern") if ($class->{_debug});
                                        
                                if ($header_line =~ /$ignore_pattern/)
                                {
                                        shift(@$unprocessed_lines);
                                        $header_line = shift(@$lines);
                                        $check_for_skip_lines = 1;
                                        logcomment("Formats Type 2: code $code. Found (pre-header) line matching one of the ignore patterns: \"$ignore_pattern\"\n" .
                                                   "Line: $header_line\n" .
                                                   "Next line to be processed as a possible header: $header_line") if ($class->{_debug});
                                        last;
                                }
                        }
                        
                }
                
                if (!(defined($header_line)) || ($header_line !~ /^$header/i))
                {
                        logcomment("Formats Type 2: code $code. Did not find expected header: $header") if ($class->{_debug});
                        return %return_hash;
                }
                
                $header_line =~ s/^$header//;
                if (defined($header_suffix))
                {
                        $header_line =~ s/$header_suffix$//;
                }
                
                # If anything is left from the header line, it is the entry-related info (ex: name of the volume)
                $return_hash_key = $header_line;
                if (defined($return_hash_key) && ($return_hash_key ne ''))
                {
                        $return_hash{found_entry_info_in_header} = TRUE;
                }
                shift(@$unprocessed_lines);
        }
        
        my $expected_info = $format->{expected_info};
        if ((!defined($expected_info)) || (ref($expected_info) ne 'ARRAY'))
        {
                logresult('script', "Library problem found: format of type 2 expected, but expected_info not correctly defined for code $code");
        }
        
        foreach my $expected_line_info (@$expected_info)
        {
                my $expected_identification_string = $expected_line_info->{indentification_string};
                my $output_hash_key = $expected_line_info->{output_hash_key};
                my $extract_extra_info = $expected_line_info->{extract_information};
                my $line = shift(@$lines);
                
                if (!defined($expected_identification_string))
                {
                        logresult('script', "Library problem found: format of type 2 expected, but expected_info is missing indentification_string for code $code");
                }
                
                if ((!defined($line)) || ($line eq ''))
                {
                        logcomment("Formats Type 2: code $code. Failed to detect information: \"$expected_identification_string\"\n" .
                                           "No more output to process:\n" . Dumper($lines)) if ($class->{_debug});
                                # Exit with failure, to allow detecting of other eventual error messages.
                                return %return_hash;
                }
                
                logcomment("Formats Type 2:\n" .
                           "processing line: \"$line\"\n" .
                           "expecting info:\n" . Dumper($expected_line_info)) if ($class->{_debug});
                
                if ($line !~ /$expected_identification_string(.+)/)
                {
                        # Skip it, if it's one of the 'ignore lines'
                        my $line_should_be_ignored = FALSE;
                        foreach my $ignore_pattern (@$ignore_lines)
                        {
                                logcomment("Formats Type 2: code $code. Checking if line should be ignored:\n$line\n" .
                                           "Trying to match: $ignore_pattern") if ($class->{_debug});
                                
                                if ($line =~ /$ignore_pattern/)
                                {
                                        logcomment("Formats Type 2: code $code. Found line matching one of the ignore patterns: \"$ignore_pattern\"\n" .
                                                   "Line: $line") if ($class->{_debug});
                                      $line_should_be_ignored = TRUE;
                                      last;
                                }
                        }
                        
                        if ($line_should_be_ignored == TRUE)
                        {
                                logcomment("Formats Type 2: code $code. Skipping line marked as to-be-ignored:\n$line") if ($class->{_debug});
                                shift(@$unprocessed_lines);
                                # Put the $expected_line_info back as the first element of $expected_info:
                                # we will expect this to be the first not-to-ignore info
                                splice(@$expected_info, 0, 0, $expected_line_info);
                                next;
                        }
                        
                        logcomment("Formats Type 2: code $code. Failed to detect information: \"$expected_identification_string\"\n" .
                                   "Processing:\n" . Dumper($lines)) if ($class->{_debug});
                        # Exit with failure, to allow detecting of other eventual error messages.
                        return %return_hash;
                }
                
                my $return_data = $1;
                logcomment("Formats Type 2: code $code.\nIdentification string: $expected_identification_string\nInfo: \"$return_data\"") if ($class->{_debug});
                if (defined($output_hash_key))
                {   
                        if ((defined($return_hash_key)) && ($return_hash_key ne ''))
                        {
                                $return_hash{$return_hash_key}{$output_hash_key} = $return_data;
                        }
                        else
                        {
                                $return_hash{$output_hash_key} = $return_data;
                        }
                }
                
                if (defined($extract_extra_info))
                {
                        foreach my $info_to_extract (@$extract_extra_info)
                        {
                                my $regex = $info_to_extract->{regex};
                                logcomment("Formats Type 2: extra processing for \"$return_data\"\n" .
                                           "Attempting to extract info:\n" . Dumper($info_to_extract)) if ($class->{_debug});
                                if ($return_data !~ /^($regex)/)
                                {
                                        # Exit this loop to stop extracting extra info
                                        logcomment("Formats Type 2: Failed to match leading regex: $regex\n" .
                                                   "No more extra info will be extracted from:\n$return_data") if ($class->{_debug});
                                        last;
                                }
                                
                                my $extra_info_data = $1;
                                my $extra_info_output_hash_key = $info_to_extract->{output_hash_key};
                                if (defined($extra_info_output_hash_key))
                                {
                                        logcomment("Formats Type 2: Adding extra info to the return hash:\n" .
                                                   "$extra_info_output_hash_key => $extra_info_data") if ($class->{_debug});
                                        if ((defined($return_hash_key)) && ($return_hash_key ne ''))
                                        {
                                                $return_hash{$return_hash_key}{$extra_info_output_hash_key} = $extra_info_data;
                                        }
                                        else
                                        {
                                                $return_hash{$extra_info_output_hash_key} = $extra_info_data;
                                        }
                                }
                                
                                $return_data =~ s/^$extra_info_data//;
                        }
                }
                
                shift(@$unprocessed_lines);
        }
        
        # We successfully processed all the expected info,
        # but we may still have lines that we need to ignore (if they are at the end of the output)
        my $candidate_line = shift(@$lines);
        while (defined($candidate_line))
        {
                logcomment("Formats Type 2: code $code. Checking if extra line should be ignored:\n$candidate_line\n" .
                           "Lines to be ignored:\n" . Dumper(@$ignore_lines)) if ($class->{_debug});
                my $candidate_should_be_ignored = FALSE;
                foreach my $ignore_pattern (@$ignore_lines)
                {
                        if ($candidate_line =~ /$ignore_pattern/)
                        {
                                logcomment("Formats Type 2: code $code. Found extra line matching one of the ignore patterns: \"$ignore_pattern\"\n" .
                                           "Line: $candidate_line") if ($class->{_debug});
                                $candidate_should_be_ignored = TRUE;
                                last;
                        }
                }
                
                if ($candidate_should_be_ignored == TRUE)
                {
                        logcomment("Formats Type 2: code $code. Skipping extra line marked as to-be-ignored:\n$candidate_line") if ($class->{_debug});
                        shift(@$unprocessed_lines);
                        $candidate_line = shift(@$lines);
                }
                else
                {
                        # Put the line back in the array and return
                        logcomment("Formats Type 2: code $code. Extra line is not marked as to-be-ignored:\n$candidate_line") if ($class->{_debug});
                        splice(@$lines, 0, 0, $candidate_line);
                        last;
                }
        }
        
        $return_hash{rc} = C_RC_SUCCESS;
        return %return_hash;
}


sub _get_format_info
{
        my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
        my $ontap_ver = $class->{_version}->{ontap};
        my $code = $args{code};
        my $return_hash = {};
        
        if ($class->{_boot_mode} eq C_ONTAP_7MODE)
        {
                if ($ontap_ver eq '7.3')
                {
                        $return_hash = $class->_7mode_formats_7_3( code => $code );
                }
                elsif ($ontap_ver eq '8.0')
                {
                        $return_hash = $class->_7mode_formats_8_0( code => $code );
                }
                elsif ($ontap_ver eq '8.1')
                {
                        $return_hash = $class->_7mode_formats_8_1( code => $code );
                }
		elsif ($ontap_ver eq '8.2')
                {
                        $return_hash = $class->_7mode_formats_8_2( code => $code );
                }
                else
                {
                        logcomment("Formats Type 2: Specific support for sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
                }
                
                if ( !defined($return_hash) || !(scalar %$return_hash) )
                {
                        logcomment( "Formats Type 2: Could not find format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
                        $return_hash =  $class->_7mode_base_formats( code => $code );
                }
        }
        elsif ($class->{_boot_mode} eq C_ONTAP_CMODE)
        {
                if ($ontap_ver eq '8.1')
                {
                        $return_hash = $class->_Cmode_formats_8_1( code => $code );
                }
                else
                {
                        logcomment("Formats Type 2: Specific support for sw_version $ontap_ver not found. Will be using base formats") if ($class->{_debug});
                }
                
                if ( !defined($return_hash) || !(scalar %$return_hash) )
                {
                        logcomment( "Formats Type 2: Could not find format definition specific for $ontap_ver. Checking base formats.") if ($class->{_debug});
                        $return_hash =  $class->_Cmode_base_formats( code => $code );
                }
        }

        return $return_hash;
}



############# 7-mode ################
sub _7mode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats Type 2: _7mode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                GET_FILE_ID => {expected_info => [{indentification_string => 'flags=',
                                                   extract_information => [{regex => '\S+',
                                                                            output_hash_key => 'flags',
                                                                            },
                                                                           {regex => ' snapid=',
                                                                            },
                                                                           {regex => '\S+',
                                                                            output_hash_key => 'snapid',
                                                                            },
                                                                           {regex => ' fileid=',
                                                                            },
                                                                           {regex => '\S+',
                                                                            output_hash_key => 'fileid',
                                                                            },],},],
                               },
                GET_CONTAINMENT_VERSION => {ignore_lines => ['volume:\S+:', # Ignore all others besides expected_info
                                                             ],
                                            expected_info => [{indentification_string => 'volume:\S+:wv_fsinfo_containment_version_compression:',
                                                               output_hash_key => 'containment_version'},
                                                              {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_compression:',
                                                               output_hash_key => 'highest_containment_version'},
                                                              ],
                                            },
                GET_COMPRESSION_STATS => {expected_info => [{indentification_string => 'compression:\S+:num_blocks_saved:',
                                                             output_hash_key => 'num_blocks_saved'},
                                                            {indentification_string => 'compression:\S+:num_blocks_cdata:',
                                                             output_hash_key => 'num_blocks_cdata'},
                                                            {indentification_string => 'compression:\S+:num_cg_decompressed:',
                                                             output_hash_key => 'num_cg_decompressed'},
                                                            {indentification_string => 'compression:\S+:num_partial_write:',
                                                             output_hash_key => 'num_partial_write'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf:',
                                                             output_hash_key => 'num_valid_buf'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf_suspended:',
                                                             output_hash_key => 'num_valid_buf_suspended'},
                                                            {indentification_string => 'compression:\S+:num_blocking_io:',
                                                             output_hash_key => 'num_blocking_io'},
                                                            {indentification_string => 'compression:\S+:num_decompress_suspends:',
                                                             output_hash_key => 'num_decompress_suspends'},
                                                            {indentification_string => 'compression:\S+:avg_decompress_cg_time:',
                                                             output_hash_key => 'avg_decompress_cg_time'},
							    {indentification_string => 'compression:\S+:num_qcheck_fail:',
							     output_hash_key => 'num_qcheck_fail'},
							    {indentification_string => 'compression:\S+:num_icg_inline:',
							     output_hash_key => 'num_icg_inline'},
                                                            {indentification_string => 'compression:\S+:num_compress_attempts:',
                                                             output_hash_key => 'num_compress_attempts'},
                                                            {indentification_string => 'compression:\S+:num_compress_fail:',
                                                             output_hash_key => 'num_compress_fail'},
                                                            {indentification_string => 'compression:\S+:num_not_enough_savings:',
                                                             output_hash_key => 'num_not_enough_savings'},
                                                            {indentification_string => 'compression:\S+:num_no_mem:',
                                                             output_hash_key => 'num_no_mem'},
							    {indentification_string => 'compression:\S+:num_min_input:',
                                                             output_hash_key => 'num_min_input'},
                                                            {indentification_string => 'compression:\S+:num_cgs_with_holes:',
                                                             output_hash_key => 'num_cgs_with_holes'},
							    {indentification_string => 'compression:\S+:avg_compress_cg_time:',
							     output_hash_key => 'avg_compress_cg_time'},
							    {indentification_string => 'compression:\S+:avg_idd_qcheck_time:',
							     output_hash_key => 'avg_idd_qcheck_time'},
							    {indentification_string => 'compression:\S+:avg_cd_qcheck_time:',
							     output_hash_key => 'avg_cd_qcheck_time'},
                                                            {indentification_string => 'compression:\S+:num_extra_cp_reads:',
                                                             output_hash_key => 'num_extra_cp_reads'},
                                                            {indentification_string => 'compression:\S+:num_write_uncompressed:',
                                                             output_hash_key => 'num_write_uncompressed'},
                                                            {indentification_string => 'compression:\S+:num_eos_bufs:',
                                                             output_hash_key => 'num_eos_bufs'},
                                                            {indentification_string => 'compression:\S+:num_decompress_read_bad_blks:',
                                                             output_hash_key => 'num_decompress_read_bad_blks'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_hw_bad:',
                                                             output_hash_key => 'num_decompress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_sw_bad:',
                                                             output_hash_key => 'num_decompress_mark_sw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_hw_bad:',
                                                             output_hash_key => 'num_compress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_sw_bad:',
                                                             output_hash_key => 'num_compress_mark_sw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compressed_inline:',
                                                             output_hash_key => 'num_compressed_inline'},
                                                            {indentification_string => 'compression:\S+:num_compressed_bg:',
                                                             output_hash_key => 'num_compressed_bg'},
                                                            {indentification_string => 'compression:\S+:num_compressed_bg_cg:',
                                                             output_hash_key => 'num_compressed_bg_cg'},
                                                            {indentification_string => 'compression:\S+:num_partial_cg_modifies:',
                                                             output_hash_key => 'num_partial_cg_modifies'},
                                                            {indentification_string => 'compression:\S+:num_partial_cg_new:',
                                                             output_hash_key => 'num_partial_cg_new'},
							    {indentification_string => 'compression:\S+:num_vbn_zero_cg_skipped:',
                                                             output_hash_key => 'num_vbn_zero_cg_skipped'},
							    {indentification_string => 'compression:\S+:bce_received_in_exempt:',
							     output_hash_key => 'bce_received_in_exempt'},
							    {indentification_string => 'compression:\S+:bce_err_abort:',
							     output_hash_key => 'bce_err_abort'},
							    {indentification_string => 'compression:\S+:bce_err_stale_inode:',
							     output_hash_key => 'bce_err_stale_inode'},
							    {indentification_string => 'compression:\S+:bce_fbn_invalid:',
							     output_hash_key => 'bce_fbn_invalid'},
							    {indentification_string => 'compression:\S+:bce_policy_stage:',
							     output_hash_key => 'bce_policy_stage'},
							    {indentification_string => 'compression:\S+:bce_compress_stage:',
							     output_hash_key => 'bce_compress_stage'},
                                                            #burt635279
							    {indentification_string => 'compression:\S+:bce_compress_stage_overwrites:',
							     output_hash_key => 'bce_compress_stage_overwrites'},
							    {indentification_string => 'compression:\S+:bce_sending_to_exempt:',
							     output_hash_key => 'bce_sending_to_exempt'},
							    {indentification_string => 'compression:\S+:bce_post_processing_stage:',
							     output_hash_key => 'bce_post_processing_stage'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_abort:',
							     output_hash_key => 'bce_err_post_processing_abort'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_stale_inode:',
							     output_hash_key => 'bce_err_post_processing_stale_inode'},
                                                            #burt635279
							    {indentification_string => 'compression:\S+:bce_post_processing_overwrites:',
							     output_hash_key => 'bce_post_processing_overwrites'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_nospace:',
							     output_hash_key => 'bce_err_post_processing_nospace'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s1\sblock:',
							     output_hash_key => 'detailed_avg_compress_cg_time.to_1_block'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s2\sblocks:',
							      output_hash_key => 'detailed_avg_compress_cg_time.to_2_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s3\sblocks:',
							      output_hash_key => 'detailed_avg_compress_cg_time.to_3_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s4\sblocks:',
							      output_hash_key => 'detailed_avg_compress_cg_time.to_4_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s5\sblocks:',
							      output_hash_key => 'detailed_avg_compress_cg_time.to_5_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s6\sblocks:',
							      output_hash_key => 'detailed_avg_compress_cg_time.to_6_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.not\scompressed:',
							      output_hash_key => 'detailed_avg_compress_cg_time.not_compressed'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s1\sblock:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_1_block'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s2\sblocks:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_2_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s3\sblocks:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_3_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s4\sblocks:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_4_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s5\sblocks:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_5_blocks'},
							    {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s6\sblocks:',
							      output_hash_key => 'detailed_avg_decompress_cg_time.from_6_blocks'},
                                                            ],
                                             ignore_lines => ['StatisticsID: \S+',
							      'compression:\S+:instance_name:',
                                                              'compression:\S+:instance_uuid:',
                                                              'compression:\S+:node_name:',
							      'compression:\S+:node_uuid:',
							      'compression:\S+:process_name:',],
                                            },
                GET_WAFL_SCAN_SPEED => {expected_info => [{indentification_string => 'WAFL scan speed is\s+',
                                                           output_hash_key => 'current_value'},
                                                          ],
                                        },
                GET_SNAPLOCK_CONTAINMENT_BITS => {ignore_lines => ['volume:\S+:', # Ignore all others besides expected_info
                                                                   ],
                                                  expected_info => [{indentification_string => 'volume:\S+:wv_fsinfo_containment_version_slc:',
                                                                     output_hash_key => 'containment_version_slc'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_sle:',
                                                                     output_hash_key => 'containment_version_sle'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_slc:',
                                                                     output_hash_key => 'containment_version_highest_slc'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_sle:',
                                                                     output_hash_key => 'containment_version_highest_sle'},
                                                                    ],
                                                 },
		GET_VOLUME_STATS => {expected_info => [{indentification_string => 'volume:\S+:instance_uuid:',
							output_hash_key => 'instance_uuid'},
						       {indentification_string => 'volume:\S+:parent_aggr:',
							output_hash_key => 'parent_aggr'},
						       {indentification_string => 'volume:\S+:avg_latency:',
							output_hash_key => 'avg_latency'},
						       {indentification_string => 'volume:\S+:total_ops:',
							output_hash_key => 'total_ops'},
						       {indentification_string => 'volume:\S+:read_data:',
							output_hash_key => 'read_data'},
						       {indentification_string => 'volume:\S+:read_latency:',
							output_hash_key => 'read_latency'},
						       {indentification_string => 'volume:\S+:read_ops:',
							output_hash_key => 'read_ops'},
						       {indentification_string => 'volume:\S+:write_data:',
							output_hash_key => 'write_data'},
						       {indentification_string => 'volume:\S+:write_latency:',
							output_hash_key => 'write_latency'},
						       {indentification_string => 'volume:\S+:write_ops:',
							output_hash_key => 'write_ops'},
						       {indentification_string => 'volume:\S+:other_latency:',
							output_hash_key => 'other_latency'},
						       {indentification_string => 'volume:\S+:other_ops:',
							output_hash_key => 'other_ops'},
						       {indentification_string => 'volume:\S+:internal_msgs:',
							output_hash_key => 'internal_msgs'},
						       {indentification_string => 'volume:\S+:read_blocks:',
							output_hash_key => 'read_blocks'},
						       {indentification_string => 'volume:\S+:write_blocks:',
							output_hash_key => 'write_blocks'},
						       {indentification_string => 'volume:\S+:synchronous_frees:',
							output_hash_key => 'synchronous_frees'},
						       {indentification_string => 'volume:\S+:asynchronous_frees:',
							output_hash_key => 'asynchronous_frees'},
						       {indentification_string => 'volume:\S+:df_worker:',
							output_hash_key => 'df_worker'},
						       {indentification_string => 'volume:\S+:nfs_read_data:',
							output_hash_key => 'nfs_read_data'},
						       {indentification_string => 'volume:\S+:nfs_read_latency:',
							output_hash_key => 'nfs_read_latency'},
						       {indentification_string => 'volume:\S+:nfs_read_ops:',
							output_hash_key => 'nfs_read_ops'},
						       {indentification_string => 'volume:\S+:nfs_write_data:',
							output_hash_key => 'nfs_write_data'},
						       {indentification_string => 'volume:\S+:nfs_write_latency:',
							output_hash_key => 'nfs_write_latency'},
						       {indentification_string => 'volume:\S+:nfs_write_ops:',
							output_hash_key => 'nfs_write_ops'},
						       {indentification_string => 'volume:\S+:nfs_other_latency:',
							output_hash_key => 'nfs_other_latency'},
						       {indentification_string => 'volume:\S+:nfs_other_ops:',
							output_hash_key => 'nfs_other_ops'},
						       {indentification_string => 'volume:\S+:cifs_read_data:',
							output_hash_key => 'cifs_read_data'},
						       {indentification_string => 'volume:\S+:cifs_read_latency:',
							output_hash_key => 'cifs_read_latency'},
						       {indentification_string => 'volume:\S+:cifs_read_ops:',
							output_hash_key => 'cifs_read_ops'},
						       {indentification_string => 'volume:\S+:cifs_write_data:',
							output_hash_key => 'cifs_write_data'},
						       {indentification_string => 'volume:\S+:cifs_write_latency:',
							output_hash_key => 'cifs_write_latency'},
						       {indentification_string => 'volume:\S+:cifs_write_ops:',
							output_hash_key => 'cifs_write_ops'},
						       {indentification_string => 'volume:\S+:cifs_other_latency:',
							output_hash_key => 'cifs_other_latency'},
						       {indentification_string => 'volume:\S+:cifs_other_ops:',
							output_hash_key => 'cifs_other_ops'},
						       {indentification_string => 'volume:\S+:san_read_data:',
							output_hash_key => 'san_read_data'},
						       {indentification_string => 'volume:\S+:san_read_latency:',
							output_hash_key => 'san_read_latency'},
						       {indentification_string => 'volume:\S+:san_read_ops:',
							output_hash_key => 'san_read_ops'},
						       {indentification_string => 'volume:\S+:san_write_data:',
							output_hash_key => 'san_write_data'},
						       {indentification_string => 'volume:\S+:san_write_latency:',
							output_hash_key => 'san_write_latency'},
						       {indentification_string => 'volume:\S+:san_write_ops:',
							output_hash_key => 'san_write_ops'},
						       {indentification_string => 'volume:\S+:san_other_latency:',
							output_hash_key => 'san_other_latency'},
						       {indentification_string => 'volume:\S+:san_other_ops:',
							output_hash_key => 'san_other_ops'},
						       {indentification_string => 'volume:\S+:fcp_read_data:',
							output_hash_key => 'fcp_read_data'},
						       {indentification_string => 'volume:\S+:fcp_read_latency:',
							output_hash_key => 'fcp_read_latency'},
						       {indentification_string => 'volume:\S+:fcp_read_ops:',
							output_hash_key => 'fcp_read_ops'},
						       {indentification_string => 'volume:\S+:fcp_write_data:',
							output_hash_key => 'fcp_write_data'},
						       {indentification_string => 'volume:\S+:fcp_write_latency:',
							output_hash_key => 'fcp_write_latency'},
						       {indentification_string => 'volume:\S+:fcp_write_ops:',
							output_hash_key => 'fcp_write_ops'},
						       {indentification_string => 'volume:\S+:fcp_other_latency:',
							output_hash_key => 'fcp_other_latency'},
						       {indentification_string => 'volume:\S+:fcp_other_ops:',
							output_hash_key => 'fcp_other_ops'},
						       {indentification_string => 'volume:\S+:iscsi_read_data:',
							output_hash_key => 'iscsi_read_data'},
						       {indentification_string => 'volume:\S+:iscsi_read_latency:',
							output_hash_key => 'iscsi_read_latency'},
						       {indentification_string => 'volume:\S+:iscsi_read_ops:',
							output_hash_key => 'iscsi_read_ops'},
						       {indentification_string => 'volume:\S+:iscsi_write_data:',
							output_hash_key => 'iscsi_write_data'},
						       {indentification_string => 'volume:\S+:iscsi_write_latency:',
							output_hash_key => 'iscsi_write_latency'},
						       {indentification_string => 'volume:\S+:iscsi_write_ops:',
							output_hash_key => 'iscsi_write_ops'},
						       {indentification_string => 'volume:\S+:iscsi_other_latency:',
							output_hash_key => 'iscsi_other_latency'},
						       {indentification_string => 'volume:\S+:iscsi_other_ops:',
							output_hash_key => 'iscsi_other_ops'},
						       {indentification_string => 'volume:\S+:flexcache_read_data:',
							output_hash_key => 'flexcache_read_data'},
						       {indentification_string => 'volume:\S+:flexcache_read_ops:',
							output_hash_key => 'flexcache_read_ops'},
						       {indentification_string => 'volume:\S+:flexcache_write_data:',
							output_hash_key => 'flexcache_write_data'},
						       {indentification_string => 'volume:\S+:flexcache_write_ops:',
							output_hash_key => 'flexcache_write_ops'},
						       {indentification_string => 'volume:\S+:flexcache_other_ops:',
							output_hash_key => 'flexcache_other_ops'},
						       {indentification_string => 'volume:\S+:flexcache_send_data:',
							output_hash_key => 'flexcache_send_data'},
						       {indentification_string => 'volume:\S+:flexcache_receive_data:',
							output_hash_key => 'flexcache_receive_data'},
						       {indentification_string => 'volume:\S+:wv_fsid:',
							output_hash_key => 'wv_fsid'},
						       {indentification_string => 'volume:\S+:wv_vol_type:',
							output_hash_key => 'wv_vol_type'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_fs_version:',
							output_hash_key => 'wv_fsinfo_fs_version'},
						       {indentification_string => 'volume:\S+:wv_volinfo_fs_options:',
							output_hash_key => 'wv_volinfo_fs_options'},
						       {indentification_string => 'volume:\S+:wv_volinfo_fs_flags:',
							output_hash_key => 'wv_volinfo_fs_flags'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_total:',
							output_hash_key => 'wv_fsinfo_blks_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_reserve:',
							output_hash_key => 'wv_fsinfo_blks_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_used:',
							output_hash_key => 'wv_fsinfo_blks_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_used_by_plane0:',
							output_hash_key => 'wv_fsinfo_blks_used_by_plane0'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_holes_cifs:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_holes:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_overwrite:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_overwrite'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_rsrv_absents:',
							output_hash_key => 'wv_fsinfo_blks_rsrv_absents'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_snap_reserve_pct:',
							output_hash_key => 'wv_fsinfo_blks_snap_reserve_pct'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_res_state:',
							output_hash_key => 'wv_fsinfo_blks_res_state'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_overwrite_slider_pct:',
							output_hash_key => 'wv_fsinfo_blks_overwrite_slider_pct'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks64_blks_rsrv_holes_cifs:',
							output_hash_key => 'wv_fsinfo_blks64_blks_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks64_blks_rsrv_holes:',
							output_hash_key => 'wv_fsinfo_blks64_blks_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_total:',
							output_hash_key => 'wv_fsinfo_public_inos_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_reserve:',
							output_hash_key => 'wv_fsinfo_public_inos_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_used:',
							output_hash_key => 'wv_fsinfo_public_inos_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_total:',
							output_hash_key => 'wv_fsinfo_private_inos_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_reserve:',
							output_hash_key => 'wv_fsinfo_private_inos_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_used:',
							output_hash_key => 'wv_fsinfo_private_inos_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blkr_cp:',
							output_hash_key => 'wv_fsinfo_blkr_cp'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_vvol_dbys_df_cnt:',
							output_hash_key => 'wv_fsinfo_blks_vvol_dbys_df_cnt'},
						       {indentification_string => 'volume:\S+:wvblk_past_eof:',
							output_hash_key => 'wvblk_past_eof'},
						       {indentification_string => 'volume:\S+:wvblk_past_eof64:',
							output_hash_key => 'wvblk_past_eof64'},
						       {indentification_string => 'volume:\S+:wvblk_snap_reserve:',
							output_hash_key => 'wvblk_snap_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_zombie_blks:',
							output_hash_key => 'wvblk_zombie_blks'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_rsrv_parent:',
							output_hash_key => 'wv_fsinfo_blks_rsrv_parent'},
						       {indentification_string => 'volume:\S+:wvblk_spcres_in_parent:',
							output_hash_key => 'wvblk_spcres_in_parent'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_holes:',
							output_hash_key => 'wvblk_rsrv_parent_holes'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_overwrite:',
							output_hash_key => 'wvblk_rsrv_parent_overwrite'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_overwrite_always:',
							output_hash_key => 'wvblk_rsrv_parent_overwrite_always'},
						       {indentification_string => 'volume:\S+:wvbd_active_frees:',
							output_hash_key => 'wvbd_active_frees'},
						       {indentification_string => 'volume:\S+:wvbd_whole_frees:',
							output_hash_key => 'wvbd_whole_frees'},
						       {indentification_string => 'volume:\S+:wvbd_active_frees_y:',
							output_hash_key => 'wvbd_active_frees_y'},
						       {indentification_string => 'volume:\S+:wvbd_owner_changed_y:',
							output_hash_key => 'wvbd_owner_changed_y'},
						       {indentification_string => 'volume:\S+:wvbd_whole_frees_o:',
							output_hash_key => 'wvbd_whole_frees_o'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_total:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_total'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_reserve:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_used:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_used'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_total:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_total'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_reserve:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_used:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_used'},
						       {indentification_string => 'volume:\S+:wvblk_delalloc:',
							output_hash_key => 'wvblk_delalloc'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_delalloc:',
							output_hash_key => 'wvblk_rsrv_delalloc'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes_cifs:',
							output_hash_key => 'wvblk_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes:',
							output_hash_key => 'wvblk_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_overwrite:',
							output_hash_key => 'wvblk_rsrv_overwrite'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_absents:',
							output_hash_key => 'wvblk_rsrv_absents'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes_cifs64:',
							output_hash_key => 'wvblk_rsrv_holes_cifs64'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes64:',
							output_hash_key => 'wvblk_rsrv_holes64'},
						       {indentification_string => 'volume:\S+:extent_size:',
							output_hash_key => 'extent_size'},
						       {indentification_string => 'volume:\S+:wvdf_enabled:',
							output_hash_key => 'wvdf_enabled'},
						       {indentification_string => 'volume:\S+:wvdf_inconsistent_scores:',
							output_hash_key => 'wvdf_inconsistent_scores'},
						       {indentification_string => 'volume:\S+:wvdf_last_fbn:',
							output_hash_key => 'wvdf_last_fbn'},
						       {indentification_string => 'volume:\S+:wvdf_water_mark:',
							output_hash_key => 'wvdf_water_mark'},
						       {indentification_string => 'volume:\S+:wvdf_max_frees:',
							output_hash_key => 'wvdf_max_frees'},
						       {indentification_string => 'volume:\S+:wvdf_total_score:',
							output_hash_key => 'wvdf_total_score'},
						       {indentification_string => 'volume:\S+:wvblk_parent_to_be_reclaimed:',
							output_hash_key => 'wvblk_parent_to_be_reclaimed'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_wi_blk_cnt:',
							output_hash_key => 'wvip_vvol_container_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_wi_size:',
							output_hash_key => 'wvip_vvol_container_wi_size'},
						       {indentification_string => 'volume:\S+:wvip_public_inofile_wi_blk_cnt:',
							output_hash_key => 'wvip_public_inofile_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_private_inofile_wi_blk_cnt:',
							output_hash_key => 'wvip_private_inofile_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_indirects:',
							output_hash_key => 'wvip_vvol_container_indirects'},
						       {indentification_string => 'volume:\S+:wvol_number_suspended:',
							output_hash_key => 'wvol_number_suspended'},
						       {indentification_string => 'volume:\S+:clone_storage_blocks:',
							output_hash_key => 'clone_storage_blocks'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_start:',
							output_hash_key => 'wvblk_reclaim_time_start'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_reset:',
							output_hash_key => 'wvblk_reclaim_time_reset'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_done:',
							output_hash_key => 'wvblk_reclaim_time_done'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_abort:',
							output_hash_key => 'wvblk_reclaim_time_abort'},
						       {indentification_string => 'volume:\S+:wvzmb_num_zmsgs_inuse:',
							output_hash_key => 'wvzmb_num_zmsgs_inuse'},
						       {indentification_string => 'volume:\S+:wvblk_ind_delalloc:',
							output_hash_key => 'wvblk_ind_delalloc'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_last_fbn:',
							output_hash_key => 'wvsblk_vvrd_last_fbn'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_spcflags:',
							output_hash_key => 'wvsblk_vvrd_spcflags'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_spc_clone_inherited:',
							output_hash_key => 'wvsblk_vvrd_spc_clone_inherited'},
						       {indentification_string => 'volume:\S+:wvsblk_space_tax:',
							output_hash_key => 'wvsblk_space_tax'},
						       {indentification_string => 'volume:\S+:wvsblk_lev0_over_nominal:',
							output_hash_key => 'wvsblk_lev0_over_nominal'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_vol_size:',
							output_hash_key => 'wvsblk_vvrd_vol_size'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_flags:',
							output_hash_key => 'wvsblk_vvrd_flags'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_slc:',
							output_hash_key => 'wv_fsinfo_containment_version_slc'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_sle:',
							output_hash_key => 'wv_fsinfo_containment_version_sle'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_compression:',
							output_hash_key => 'wv_fsinfo_containment_version_compression'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_vmalign:',
							output_hash_key => 'wv_fsinfo_containment_version_vmalign'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_spare1:',
							output_hash_key => 'wv_fsinfo_containment_version_spare1'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_spare2:',
							output_hash_key => 'wv_fsinfo_containment_version_spare2'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_slc:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_slc'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_sle:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_sle'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_compression:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_compression'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_vmalign:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_vmalign'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_spare1:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_spare1'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_spare2:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_spare2'},
						       {indentification_string => 'volume:\S+:wvi2p_wip_wi_size:',
							output_hash_key => 'wvi2p_wip_wi_size'},
						       {indentification_string => 'volume:\S+:bad_user_blk_read_error_propagated:',
							output_hash_key => 'bad_user_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_user_blk_read_error_not_propagated:',
							output_hash_key => 'bad_user_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_non_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_non_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_non_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_non_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_metafile_blk_read_error_propagated:',
							output_hash_key => 'bad_fixable_metafile_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_metafile_blk_read_error_not_propagated:',
							output_hash_key => 'bad_fixable_metafile_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_zombie_ind_blk_read_err_propagated:',
							output_hash_key => 'bad_zombie_ind_blk_read_err_propagated'},
						       {indentification_string => 'volume:\S+:bad_zombie_ind_blk_read_err_not_propagated:',
							output_hash_key => 'bad_zombie_ind_blk_read_err_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_user_blk_read_error_propagated:',
							output_hash_key => 'bad_container_user_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_user_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_user_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_afs_blk_read_error_propagated:',
							output_hash_key => 'bad_container_fixable_afs_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_afs_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_fixable_afs_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_snap_blk_read_error_propagated:',
							output_hash_key => 'bad_container_fixable_snap_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_snap_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_fixable_snap_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_non_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_container_non_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_non_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_non_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:wv_playlist_hits:',
							output_hash_key => 'wv_playlist_hits'},
						       {indentification_string => 'volume:\S+:wv_playlist_misses:',
							output_hash_key => 'wv_playlist_misses'},
						       {indentification_string => 'volume:\S+:wv_playlist_not_present:',
							output_hash_key => 'wv_playlist_not_present'},
						       {indentification_string => 'volume:\S+:wv_playlist_entries:',
							output_hash_key => 'wv_playlist_entries'},
						       {indentification_string => 'volume:\S+:wv_playlist_reqs:',
							output_hash_key => 'wv_playlist_reqs'},
						       {indentification_string => 'volume:\S+:wv_playlist_no_raidbufs:',
							output_hash_key => 'wv_playlist_no_raidbufs'},
						       {indentification_string => 'volume:\S+:wv_playlist_getbuf_failures:',
							output_hash_key => 'wv_playlist_getbuf_failures'},
						       {indentification_string => 'volume:\S+:wv_playlist_cont_indirects:',
							output_hash_key => 'wv_playlist_cont_indirects'},
						       {indentification_string => 'volume:\S+:wv_playlist_vvbn_holes:',
							output_hash_key => 'wv_playlist_vvbn_holes'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_not_started:',
							output_hash_key => 'wv_playlist_prefetch_not_started'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_start_time:',
							output_hash_key => 'wv_playlist_prefetch_start_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_end_time:',
							output_hash_key => 'wv_playlist_prefetch_end_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_load_end_time:',
							output_hash_key => 'wv_playlist_load_end_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_used_slots:',
							output_hash_key => 'wv_playlist_apfi_used_slots'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_collision_inserts:',
							output_hash_key => 'wv_playlist_apfi_collision_inserts'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_collision_accesses:',
							output_hash_key => 'wv_playlist_apfi_collision_accesses'},
						       {indentification_string => 'volume:\S+:wvsnap_ondisk_count:',
							output_hash_key => 'wvsnap_ondisk_count'},
						       {indentification_string => 'volume:\S+:wvsnap_incore_count:',
							output_hash_key => 'wvsnap_incore_count'},
						       {indentification_string => 'volume:\S+:wvsnap_loaded_total:',
							output_hash_key => 'wvsnap_loaded_total'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_in_use:',
							output_hash_key => 'delete_log_dlog_in_use'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_ind_blk_cnt:',
							output_hash_key => 'delete_log_dlog_ind_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_blk_cnt:',
							output_hash_key => 'delete_log_dlog_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.purge_ind_blk_cnt:',
							output_hash_key => 'delete_log_purge_ind_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.purge_blk_cnt:',
							output_hash_key => 'delete_log_purge_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_hole_cnt:',
							output_hash_key => 'delete_log_dlog_hole_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_cap:',
							output_hash_key => 'delete_log_dlog_cap'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_in_use:',
							output_hash_key => 'delete_log_overflow_in_use'},
						       {indentification_string => 'volume:\S+:delete_log.overflow0_blk_cnt:',
							output_hash_key => 'delete_log_overflow0_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.overflow1_blk_cnt:',
							output_hash_key => 'delete_log_overflow1_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.del_log:',
							output_hash_key => 'delete_log_del_log'},
						       {indentification_string => 'volume:\S+:delete_log.purge_send:',
							output_hash_key => 'delete_log_purge_send'},
						       {indentification_string => 'volume:\S+:delete_log.purge_block:',
							output_hash_key => 'delete_log_purge_block'},
						       {indentification_string => 'volume:\S+:delete_log.purge_vbns:',
							output_hash_key => 'delete_log_purge_vbns'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_frees:',
							output_hash_key => 'delete_log_dlog_frees'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_vbn_clears:',
							output_hash_key => 'delete_log_dlog_vbn_clears'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_keys:',
							output_hash_key => 'delete_log_dlog_keys'},
						       {indentification_string => 'volume:\S+:delete_log.pre_split_failed:',
							output_hash_key => 'delete_log_pre_split_failed'},
						       {indentification_string => 'volume:\S+:delete_log.pre_bufs_loaded:',
							output_hash_key => 'delete_log_pre_bufs_loaded'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow:',
							output_hash_key => 'delete_log_purge_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow_susp:',
							output_hash_key => 'delete_log_purge_overflow_susp'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow_no_split:',
							output_hash_key => 'delete_log_purge_overflow_no_split'},
						       {indentification_string => 'volume:\S+:delete_log.trunc_overflow_of_append:',
							output_hash_key => 'delete_log_trunc_overflow_of_append'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_overuse_mem:',
							output_hash_key => 'delete_log_dlog_overuse_mem'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_overflow:',
							output_hash_key => 'delete_log_dlog_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_total:',
							output_hash_key => 'delete_log_overflow_total'},
						       {indentification_string => 'volume:\S+:delete_log.act_dirty:',
							output_hash_key => 'delete_log_act_dirty'},
						       {indentification_string => 'volume:\S+:delete_log.pre_overflow:',
							output_hash_key => 'delete_log_pre_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_vbns:',
							output_hash_key => 'delete_log_overflow_vbns'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_frees:',
							output_hash_key => 'delete_log_overflow_frees'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_frees_ind:',
							output_hash_key => 'delete_log_overflow_frees_ind'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_commits:',
							output_hash_key => 'delete_log_overflow_commits'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_send:',
							output_hash_key => 'delete_log_overflow_send'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_susp:',
							output_hash_key => 'delete_log_overflow_susp'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_throttle:',
							output_hash_key => 'delete_log_overflow_throttle'},
						       {indentification_string => 'volume:\S+:msgs_rejected_in_nvfailed_state:',
							output_hash_key => 'msgs_rejected_in_nvfailed_state'},
						       {indentification_string => 'volume:\S+:msgs_allowed_in_nvfailed_state:',
							output_hash_key => 'msgs_allowed_in_nvfailed_state'},
						       {indentification_string => 'volume:\S+:clone_inodes:',
							output_hash_key => 'clone_inodes'},
						       {indentification_string => 'volume:\S+:clone_lookups:',
							output_hash_key => 'clone_lookups'},
						       {indentification_string => 'volume:\S+:clone_lookup_hits:',
							output_hash_key => 'clone_lookup_hits'},
						       {indentification_string => 'volume:\S+:clone_blks_copied:',
							output_hash_key => 'clone_blks_copied'},
						       {indentification_string => 'volume:\S+:clone_eio_blks:',
							output_hash_key => 'clone_eio_blks'},
						       {indentification_string => 'volume:\S+:clone_num_entries:',
							output_hash_key => 'clone_num_entries'},
						       {indentification_string => 'volume:\S+:clone_max_entries:',
							output_hash_key => 'clone_max_entries'},
						       {indentification_string => 'volume:\S+:clone_max_hierarchy:',
							output_hash_key => 'clone_max_hierarchy'},
						       {indentification_string => 'volume:\S+:clone_unsplit_snap_entries:',
							output_hash_key => 'clone_unsplit_snap_entries'},
						       {indentification_string => 'volume:\S+:clone_num_share_stopped:',
							output_hash_key => 'clone_num_share_stopped'},
						       {indentification_string => 'volume:\S+:clone_read_redirected:',
							output_hash_key => 'clone_read_redirected'},
						       {indentification_string => 'volume:\S+:clone_split_ra:',
							output_hash_key => 'clone_split_ra'},
						       {indentification_string => 'volume:\S+:clone_split_ra_lag:',
							output_hash_key => 'clone_split_ra_lag'},
						       {indentification_string => 'volume:\S+:clone_inline_split_kireeti_in_progress:',
							output_hash_key => 'clone_inline_split_kireeti_in_progress'},
						       {indentification_string => 'volume:\S+:clone_inline_split_stage_size_limitation:',
							output_hash_key => 'clone_inline_split_stage_size_limitation'},
						       {indentification_string => 'volume:\S+:clone_inline_split_beyond_eof:',
							output_hash_key => 'clone_inline_split_beyond_eof'},
						       {indentification_string => 'volume:\S+:clone_inline_split_source_destination_dirty:',
							output_hash_key => 'clone_inline_split_source_destination_dirty'},
						       {indentification_string => 'volume:\S+:clone_inline_split_source_spec_vbn:',
							output_hash_key => 'clone_inline_split_source_spec_vbn'},
						       {indentification_string => 'volume:\S+:clone_inline_split_enospc:',
							output_hash_key => 'clone_inline_split_enospc'},
						       {indentification_string => 'volume:\S+:clone_inline_split_edquot:',
							output_hash_key => 'clone_inline_split_edquot'},
						       {indentification_string => 'volume:\S+:nonzero_dbys_cnt:',
							output_hash_key => 'nonzero_dbys_cnt'},
						       ],
				     ignore_lines => ['volume:\S+:instance_name:',
						      'volume:\S+:node_name:',
						      'volume:\S+:vserver_name:',
						      'volume:\S+:vserver_uuid:',
						      'volume:\S+:clone_sizes_hist',
						      ],
				     },
		GET_WAFL_STATS => {ignore_lines => ['.+', # Ignore all others besides expected_info
						    ],
				   expected_info => [{indentification_string => 'CGs resolved in KC\s*\t*=\s*\t*',
						      output_hash_key => 'cgs_resolved_in_KC'},
						     {indentification_string => 'CGs resolved in Bento\s*\t*=\s*\t*',
						      output_hash_key => 'cgs_resolved_in_Bento'},
						     {indentification_string => 'Truncate Suspends\s*\t*=\s*\t*',
						      output_hash_key => 'truncate_suspends'},
						     {indentification_string => 'PreCP overwritten\s*\t*=\s*\t*',
						      output_hash_key => 'precp_overwritten'},
						     {indentification_string => 'Flippity Suspends\s*\t*=\s*\t*',
						      output_hash_key => 'flippity_suspends'},
						     {indentification_string => 'Flippity Skipped\s*\t*=\s*\t*',
						      output_hash_key => 'flippity_skipped'},
						     {indentification_string => 'Decompress Nomem\s*\t*=\s*\t*',
						      output_hash_key => 'decompress_nomem'},
						     {indentification_string => 'Buffers cacheable in EC\s*\t*=\s*\t*',
						      output_hash_key => 'buffers_cacheable_in_EC'},
						     {indentification_string => 'Non-Stripe writes\s*\t*=\s*\t*',
						      output_hash_key => 'non_stripe_writes'},
						     {indentification_string => 'Partial CG throttled\s*\t*=\s*\t*',
						      output_hash_key => 'partial_cg_throttled'},
						     ],
				  },
                };

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logresult('script', "Library problem found: Processing required for format type 2, but no format info defined. Command code: $code");
        }

	return $format;
}

sub _7mode_formats_7_3
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_7_3 = {};
	my $code = $args{code};
        
        return $formats_hash_7_3->{$code}
}

sub _7mode_formats_8_0
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_8_0 = {
		COMPRESS_STATUS_CHECKPOINT => {header => 'Checkpoint info of volume ',
                                               header_suffix => ':',
                                               expected_info => [{indentification_string => 'Checkpoint magic : ',
                                                                  output_hash_key => 'checkpoint_magic'},
                                                                 {indentification_string => 'Checkpoint version : ',
                                                                  output_hash_key => 'checkpoint_version'},
                                                                 {indentification_string => 'Compressing blocks shared with snapshots : ',
                                                                  output_hash_key => 'snapshotted_blocks'},
                                                                 {indentification_string => 'Compressing de-duplicated blocks : ',
                                                                  output_hash_key => 'deduped_blocks'},
                                                                 {indentification_string => 'Inode number to restore from : ',
                                                                  output_hash_key => 'inode_to_restore_from'},
                                                                 {indentification_string => 'Fbn number to restore from : ',
                                                                  output_hash_key => 'fbn_to_restore_from'},
                                                                 {indentification_string => 'Snapshot skipped data \(KB\) : ',
                                                                  output_hash_key => 'snapshot_skipped_data'},
                                                                 {indentification_string => 'Shared skipped data \(KB\) : ',
                                                                  output_hash_key => 'deduped_skipped_data'},
                                                                 {indentification_string => 'Inconsistent skipped data \(KB\) : ',
                                                                  output_hash_key => 'inconsistent_skipped_data'},
                                                                 {indentification_string => 'Total data scanned \(KB\) : ',
                                                                  output_hash_key => 'total_data_scanned'},
                                                                 {indentification_string => 'Checkpoint creation time : ',
                                                                  output_hash_key => 'checkpoint_creation_time'},
                                                                 ],
                                               },
                
                GET_COMPRESSION_STATS => {expected_info => [{indentification_string => 'compression:\S+:num_blocks_saved:',
                                                             output_hash_key => 'num_blocks_saved'},
                                                            {indentification_string => 'compression:\S+:num_blocks_cdata:',
                                                             output_hash_key => 'num_blocks_cdata'},
                                                            {indentification_string => 'compression:\S+:num_blocks_uncompressed:',
                                                             output_hash_key => 'num_blocks_uncompressed'},
                                                            {indentification_string => 'compression:\S+:num_buf_decompressed:',
                                                             output_hash_key => 'num_buf_decompressed'},
                                                            {indentification_string => 'compression:\S+:num_cg_decompressed:',
                                                             output_hash_key => 'num_cg_decompressed'},
                                                            {indentification_string => 'compression:\S+:num_partial_write:',
                                                             output_hash_key => 'num_partial_write'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf:',
                                                             output_hash_key => 'num_valid_buf'},
                                                            {indentification_string => 'compression:\S+:num_invalid_raid_read:',
                                                             output_hash_key => 'num_invalid_raid_read'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf_suspended:',
                                                             output_hash_key => 'num_valid_buf_suspended'},
                                                            {indentification_string => 'compression:\S+:num_cow_cg:',
                                                             output_hash_key => 'num_cow_cg'},
                                                            {indentification_string => 'compression:\S+:num_blocking_io:',
                                                             output_hash_key => 'num_blocking_io'},
                                                            {indentification_string => 'compression:\S+:num_vol_decomp_ahead_busy:',
                                                             output_hash_key => 'num_vol_decomp_ahead_busy'},
                                                            {indentification_string => 'compression:\S+:num_vol_decomp_no_resource:',
                                                             output_hash_key => 'num_vol_decomp_no_resource'},
                                                            {indentification_string => 'compression:\S+:num_vol_decomp_zero_holes:',
                                                             output_hash_key => 'num_vol_decomp_zero_holes'},
                                                            {indentification_string => 'compression:\S+:num_decompress_suspends:',
                                                             output_hash_key => 'num_decompress_suspends'},
                                                            {indentification_string => 'compression:\S+:avg_vol_decompress_time:',
                                                             output_hash_key => 'avg_vol_decompress_time'},
                                                            {indentification_string => 'compression:\S+:avg_decompress_cg_time:',
                                                             output_hash_key => 'avg_decompress_cg_time'},
                                                            {indentification_string => 'compression:\S+:num_compress_attempts:',
                                                             output_hash_key => 'num_compress_attempts'},
                                                            {indentification_string => 'compression:\S+:num_compress_fail:',
                                                             output_hash_key => 'num_compress_fail'},
                                                            {indentification_string => 'compression:\S+:num_not_enough_savings:',
                                                             output_hash_key => 'num_not_enough_savings'},
                                                            {indentification_string => 'compression:\S+:num_no_mem:',
                                                             output_hash_key => 'num_no_mem'},
                                                            {indentification_string => 'compression:\S+:num_compress_suspends:',
                                                             output_hash_key => 'num_compress_suspends'},
                                                            {indentification_string => 'compression:\S+:num_cgs_frz_attempts:',
                                                             output_hash_key => 'num_cgs_frz_attempts'},
                                                            {indentification_string => 'compression:\S+:num_cgs_frz_failed:',
                                                             output_hash_key => 'num_cgs_frz_failed'},
                                                            {indentification_string => 'compression:\S+:num_buf_attempts:',
                                                             output_hash_key => 'num_buf_attempts'},
                                                            {indentification_string => 'compression:\S+:num_nocompress_files:',
                                                             output_hash_key => 'num_nocompress_files'},
                                                            {indentification_string => 'compression:\S+:num_min_input:',
                                                             output_hash_key => 'num_min_input'},
                                                            {indentification_string => 'compression:\S+:num_cgs_with_holes:',
                                                             output_hash_key => 'num_cgs_with_holes'},
                                                            {indentification_string => 'compression:\S+:num_next_cp:',
                                                             output_hash_key => 'num_next_cp'},
                                                            {indentification_string => 'compression:\S+:num_wait_for_compressor:',
                                                             output_hash_key => 'num_wait_for_compressor'},
                                                            {indentification_string => 'compression:\S+:avg_compress_cg_time:',
                                                             output_hash_key => 'avg_compress_cg_time'},
                                                            {indentification_string => 'compression:\S+:num_extra_cp_reads:',
                                                             output_hash_key => 'num_extra_cp_reads'},
                                                            {indentification_string => 'compression:\S+:num_compress_ahead:',
                                                             output_hash_key => 'num_compress_ahead'},
                                                            {indentification_string => 'compression:\S+:num_overload:',
                                                             output_hash_key => 'num_overload'},
                                                            {indentification_string => 'compression:\S+:hole_to_data:',
                                                             output_hash_key => 'hole_to_data'},
                                                            {indentification_string => 'compression:\S+:hole_to_compressed:',
                                                             output_hash_key => 'hole_to_compressed'},
                                                            {indentification_string => 'compression:\S+:data_to_hole:',
                                                             output_hash_key => 'data_to_hole'},
                                                            {indentification_string => 'compression:\S+:num_write_uncompressed:',
                                                             output_hash_key => 'num_write_uncompressed'},
                                                            {indentification_string => 'compression:\S+:num_eos_bufs:',
                                                             output_hash_key => 'num_eos_bufs'},
                                                            {indentification_string => 'compression:\S+:num_cgs_not_fully_dirty:',
                                                             output_hash_key => 'num_cgs_not_fully_dirty'},
                                                            {indentification_string => 'compression:\S+:num_wafl_decompress:',
                                                             output_hash_key => 'num_wafl_decompress'},
                                                            {indentification_string => 'compression:\S+:num_decompress_read_bad_blks:',
                                                             output_hash_key => 'num_decompress_read_bad_blks'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_hw_bad:',
                                                             output_hash_key => 'num_decompress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_sw_bad:',
                                                             output_hash_key => 'num_decompress_mark_sw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_hw_bad:',
                                                             output_hash_key => 'num_compress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_sw_bad:',
                                                             output_hash_key => 'num_compress_mark_sw_bad'},
                                                            ],
                                            },
	};
	my $code = $args{code};
        
        return $formats_hash_8_0->{$code}
}

sub _7mode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash_8_1 = {
                COMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                    # TODO: investigate: is this a bug?
                                    ignore_lines => ['Policy ID:\s*',
                                                     'Stale Fingerprints.*:\s+',
                                                     # TODO: move these to expected_info once burt 410442
                                                     # is merged to main
                                                     'Last Operation State:\s*',
                                                     'Last Successful Operation Begin:\s*',
                                                     'Last Successful Operation End:\s*',
                                                    ],
                                    expected_info => [{indentification_string => 'State:\s+',
                                                       output_hash_key => 'state'},
                                                      {indentification_string => 'Compression:\s+',
                                                       output_hash_key => 'compression'},
                                                      {indentification_string => 'Inline Compression:\s+',
                                                       output_hash_key => 'inline_compression'},
                                                      {indentification_string => 'Status:\s+',
                                                       output_hash_key => 'status'},
                                                      {indentification_string => 'Progress:\s+',
                                                       output_hash_key => 'progress',
                                                       extract_information => [{regex => '\d+',
                                                                                output_hash_key => 'data_scanned',
                                                                               },],},
                                                      {indentification_string => 'Type:\s+',
                                                       output_hash_key => 'type'},
                                                      {indentification_string => 'Schedule:\s+',
                                                       output_hash_key => 'schedule'},
                                                      {indentification_string => 'Minimum Blocks Shared:\s+',
                                                       output_hash_key => 'minimum_blocks_shared'},
                                                      {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                       output_hash_key => 'blocks_skipped_sharing'},
                                                      {indentification_string => 'Last Operation Begin:\s+',
                                                       output_hash_key => 'last_operation_begin'},
                                                      {indentification_string => 'Last Operation End:\s+',
                                                       output_hash_key => 'last_operation_end'},
                                                      {indentification_string => 'Last Operation Size:\s+',
                                                       output_hash_key => 'last_operation_size'},
                                                      {indentification_string => 'Last Operation Error:\s+',
                                                       output_hash_key => 'last_operation_error'},
                                                      {indentification_string => 'Change Log Usage:\s+',
                                                       output_hash_key => 'changelog_usage'},
                                                      {indentification_string => 'Logical Data:\s+',
                                                       output_hash_key => 'logical_data'},
                                                      {indentification_string => 'Queued Job:\s+',
                                                       output_hash_key => 'queued_jobs'},
                                                      {indentification_string => 'Stage:\s+',
                                                       output_hash_key => 'stage'},
                                                      {indentification_string => 'Checkpoint Time:\s+',
                                                       output_hash_key => 'checkpoint_time'},
                                                      {indentification_string => 'Checkpoint Operation Type:\s+',
                                                       output_hash_key => 'checkpoint_op_type'},
                                                      {indentification_string => 'Checkpoint Stage:\s+',
                                                       output_hash_key => 'checkpoint_stage'},
                                                      {indentification_string => 'Checkpoint Sub-stage:\s+',
                                                       output_hash_key => 'checkpoint_substage'},
                                                      {indentification_string => 'Checkpoint Progress:\s+',
                                                       output_hash_key => 'checkpoint_progress'},
                                                      {indentification_string => 'Fingerprints Gathered:\s+',
                                                       output_hash_key => 'fingerprints_gathered'},
                                                      {indentification_string => 'Gathering Begin:\s+',
                                                       output_hash_key => 'gathering_begin'},
                                                      {indentification_string => 'Fingerprints Sorted:\s+',
                                                       output_hash_key => 'fingerprints_sorted'},
                                                      {indentification_string => 'Duplicate Blocks Found:\s+',
                                                       output_hash_key => 'duplicated_blocks_found'},
                                                      {indentification_string => 'Sorting Begin:\s+',
                                                       output_hash_key => 'sorting_begin'},
                                                      {indentification_string => 'Blocks De-duplicated:\s+',
                                                       output_hash_key => 'blocks_deduplicated'},
                                                      {indentification_string => 'De-duping Begin:\s+',
                                                       output_hash_key => 'dedupe_begin'},
                                                      {indentification_string => 'Fingerprints Deleted:\s+',
                                                       output_hash_key => 'fingerprints_deleted'},
                                                      {indentification_string => 'Checking Begin:\s+',
                                                       output_hash_key => 'checking_begin'},
                                                      {indentification_string => 'Verify Trigger Rate:\s+',
                                                       output_hash_key => 'verify_trigger_rate'},
                                                      {indentification_string => 'Total Verify Time:\s+',
                                                       output_hash_key => 'total_verify_time'},
                                                      {indentification_string => 'Verify Suspend Count:\s+',
                                                       output_hash_key => 'verify_suspend_count'},
                                                      ],    
                                               },
                COMPRESS_STATUS_CHECKPOINT => {header => 'Path:\s+/vol/',
                                               ignore_lines => ['Policy ID:\s*',
                                                                # TODO: move these to expected_info once burt 410442
                                                                # is merged to main
                                                                'Last Operation State:\s*',
                                                                'Last Successful Operation Begin:\s*',
                                                                'Last Successful Operation End:\s*',
                                                                ],
                                               expected_info => [{indentification_string => 'State:\s+',
                                                                  output_hash_key => 'state'},
                                                                 {indentification_string => 'Compression:\s+',
                                                                  output_hash_key => 'compression'},
                                                                 {indentification_string => 'Inline Compression:\s+',
                                                                  output_hash_key => 'inline_compression'},
                                                                 {indentification_string => 'Status:\s+',
                                                                  output_hash_key => 'status'},
                                                                 {indentification_string => 'Progress:\s+',
                                                                  output_hash_key => 'progress',
                                                                  extract_information => [{regex => '\d+',
                                                                                           output_hash_key => 'total_data_scanned',
                                                                                           },],},
                                                                 {indentification_string => 'Type:\s+',
                                                                  output_hash_key => 'type'},
                                                                 {indentification_string => 'Schedule:\s+',
                                                                  output_hash_key => 'schedule'},
                                                                 {indentification_string => 'Minimum Blocks Shared:\s+',
                                                                  output_hash_key => 'minimum_blocks_shared'},
                                                                 {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                                  output_hash_key => 'blocks_skipped_sharing'},
                                                                 {indentification_string => 'Last Operation Begin:\s+',
                                                                  output_hash_key => 'last_operation_begin'},
                                                                 {indentification_string => 'Last Operation End:\s+',
                                                                  output_hash_key => 'last_operation_end'},
                                                                 {indentification_string => 'Last Operation Size:\s+',
                                                                  output_hash_key => 'last_operation_size'},
                                                                 {indentification_string => 'Last Operation Error:\s+',
                                                                  output_hash_key => 'last_operation_error'},
                                                                 {indentification_string => 'Change Log Usage:\s+',
                                                                  output_hash_key => 'changelog_usage'},
                                                                 {indentification_string => 'Logical Data:\s+',
                                                                  output_hash_key => 'logical_data'},
                                                                 {indentification_string => 'Queued Job:\s+',
                                                                  output_hash_key => 'queued_jobs'},
                                                                 {indentification_string => 'Stale Fingerprints.*:\s+',
                                                                  output_hash_key => 'stale_fingerprints'},
                                                                 {indentification_string => 'Stage:\s+',
                                                                  output_hash_key => 'stage'},
                                                                 {indentification_string => 'Checkpoint Time:\s+',
                                                                  output_hash_key => 'checkpoint_creation_time'},
                                                                 {indentification_string => 'Checkpoint Operation Type:\s+',
                                                                  output_hash_key => 'checkpoint_op_type'},
                                                                 {indentification_string => 'Checkpoint Stage:\s+',
                                                                  output_hash_key => 'checkpoint_stage'},
                                                                 {indentification_string => 'Checkpoint Sub-stage:\s+',
                                                                  output_hash_key => 'checkpoint_substage'},
                                                                 {indentification_string => 'Checkpoint Progress:\s+',
                                                                  output_hash_key => 'checkpoint_progress'},
                                                                 {indentification_string => 'Fingerprints Gathered:\s+',
                                                                  output_hash_key => 'fingerprints_gathered'},
                                                                 {indentification_string => 'Gathering Begin:\s+',
                                                                  output_hash_key => 'gathering_begin'},
                                                                 {indentification_string => 'Fingerprints Sorted:\s+',
                                                                  output_hash_key => 'fingerprints_sorted'},
                                                                 {indentification_string => 'Duplicate Blocks Found:\s+',
                                                                  output_hash_key => 'duplicated_blocks_found'},
                                                                 {indentification_string => 'Sorting Begin:\s+',
                                                                  output_hash_key => 'sorting_begin'},
                                                                 {indentification_string => 'Blocks De-duplicated:\s+',
                                                                  output_hash_key => 'blocks_deduplicated'},
                                                                 {indentification_string => 'De-duping Begin:\s+',
                                                                  output_hash_key => 'dedupe_begin'},
                                                                 {indentification_string => 'Fingerprints Deleted:\s+',
                                                                  output_hash_key => 'fingerprints_deleted'},
                                                                 {indentification_string => 'Checking Begin:\s+',
                                                                  output_hash_key => 'checking_begin'},
                                                                 {indentification_string => 'Verify Trigger Rate:\s+',
                                                                  output_hash_key => 'verify_trigger_rate'},
                                                                 {indentification_string => 'Total Verify Time:\s+',
                                                                  output_hash_key => 'total_verify_time'},
                                                                 {indentification_string => 'Verify Suspend Count:\s+',
                                                                  output_hash_key => 'verify_suspend_count'},
                                                                 ],
                                                
                                              },
                DECOMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                      ignore_lines => ['Policy ID:\s*',
                                                       # TODO: move these to expected_info once burt 410442
                                                       # is merged to main
                                                       'Last Operation State:\s*',
                                                       'Last Successful Operation Begin:\s*',
                                                       'Last Successful Operation End:\s*',
                                                       ],
                                      expected_info => [{indentification_string => 'State:\s+',
                                                         output_hash_key => 'state'},
                                                        {indentification_string => 'Compression:\s+',
                                                         output_hash_key => 'compression'},
                                                        {indentification_string => 'Inline Compression:\s+',
                                                         output_hash_key => 'inline_compression'},
                                                        {indentification_string => 'Status:\s+',
                                                         output_hash_key => 'status'},
                                                        {indentification_string => 'Progress:\s+',
                                                         output_hash_key => 'progress',
                                                         extract_information => [{regex => '\d+',
                                                                                  output_hash_key => 'data_processed',
                                                                                  },
                                                                                 {regex => '\s+KB\s+',
                                                                                 },
                                                                                 {regex => '\w+',
                                                                                  output_hash_key => 'processing_phase',
                                                                                 },],},
                                                        {indentification_string => 'Type:\s+',
                                                         output_hash_key => 'type'},
                                                        {indentification_string => 'Schedule:\s+',
                                                         output_hash_key => 'schedule'},
                                                        {indentification_string => 'Minimum Blocks Shared:\s+',
                                                         output_hash_key => 'minimum_blocks_shared'},
                                                        {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                         output_hash_key => 'blocks_skipped_sharing'},
                                                        {indentification_string => 'Last Operation Begin:\s+',
                                                         output_hash_key => 'last_operation_begin'},
                                                        {indentification_string => 'Last Operation End:\s+',
                                                         output_hash_key => 'last_operation_end'},
                                                        {indentification_string => 'Last Operation Size:\s+',
                                                         output_hash_key => 'last_operation_size'},
                                                        {indentification_string => 'Last Operation Error:\s+',
                                                         output_hash_key => 'last_operation_error'},
                                                        {indentification_string => 'Change Log Usage:\s+',
                                                         output_hash_key => 'changelog_usage'},
                                                        {indentification_string => 'Logical Data:\s+',
                                                         output_hash_key => 'logical_data'},
                                                        {indentification_string => 'Queued Job:\s+',
                                                         output_hash_key => 'queued_jobs'},
                                                        {indentification_string => 'Stale Fingerprints.*:\s+',
                                                         output_hash_key => 'stale_fingerprints'},
                                                        {indentification_string => 'Stage:\s+',
                                                         output_hash_key => 'stage'},
                                                        {indentification_string => 'Checkpoint Time:\s+',
                                                         output_hash_key => 'checkpoint_time'},
                                                        {indentification_string => 'Checkpoint Operation Type:\s+',
                                                         output_hash_key => 'checkpoint_op_type'},
                                                        {indentification_string => 'Checkpoint Stage:\s+',
                                                         output_hash_key => 'checkpoint_stage'},
                                                        {indentification_string => 'Checkpoint Sub-stage:\s+',
                                                         output_hash_key => 'checkpoint_substage'},
                                                        {indentification_string => 'Checkpoint Progress:\s+',
                                                         output_hash_key => 'checkpoint_progress'},
                                                        {indentification_string => 'Fingerprints Gathered:\s+',
                                                         output_hash_key => 'fingerprints_gathered'},
                                                        {indentification_string => 'Gathering Begin:\s+',
                                                         output_hash_key => 'gathering_begin'},
                                                        {indentification_string => 'Fingerprints Sorted:\s+',
                                                         output_hash_key => 'fingerprints_sorted'},
                                                        {indentification_string => 'Duplicate Blocks Found:\s+',
                                                         output_hash_key => 'duplicated_blocks_found'},
                                                        {indentification_string => 'Sorting Begin:\s+',
                                                         output_hash_key => 'sorting_begin'},
                                                        {indentification_string => 'Blocks De-duplicated:\s+',
                                                         output_hash_key => 'blocks_deduplicated'},
                                                        {indentification_string => 'De-duping Begin:\s+',
                                                         output_hash_key => 'dedupe_begin'},
                                                        {indentification_string => 'Fingerprints Deleted:\s+',
                                                         output_hash_key => 'fingerprints_deleted'},
                                                        {indentification_string => 'Checking Begin:\s+',
                                                         output_hash_key => 'checking_begin'},
                                                        {indentification_string => 'Verify Trigger Rate:\s+',
                                                         output_hash_key => 'verify_trigger_rate'},
                                                        {indentification_string => 'Total Verify Time:\s+',
                                                         output_hash_key => 'total_verify_time'},
                                                        {indentification_string => 'Verify Suspend Count:\s+',
                                                         output_hash_key => 'verify_suspend_count'},
                                                      ],
                                               },
                GET_STATS => {header => 'Path:\s+/vol/',
                              ignore_lines => ['^\s*<*(\d+|UNUSED)\s*<*(\d+|UNUSED)',
                                               'Block sharing histogram:',
                                               'Reference count histogram:',
                                               'Sort Stats:',
											   'No Op:\s*'
                                               ],
                              expected_info => [{indentification_string => 'Allocated:\s+',
                                                 output_hash_key => 'allocated'},
                                                {indentification_string => 'Shared:\s+',
                                                 output_hash_key => 'shared'},
                                                {indentification_string => 'Saving:\s+',
                                                 output_hash_key => 'saving'},
                                                {indentification_string => '%Saved:\s+',
                                                 output_hash_key => 'percentage_saved'},
                                                {indentification_string => 'Max Refcount:\s+',
                                                 output_hash_key => 'max_refcount'},
                                                {indentification_string => 'Total Processed:\s+',
                                                 output_hash_key => 'total_processed'},
                                                {indentification_string => 'Total Process Time:\s+',
                                                 output_hash_key => 'total_process_time'},
                                                {indentification_string => 'Total Verify Time:\s+',
                                                 output_hash_key => 'total_verify_time'},
                                                {indentification_string => 'SIS Files:\s+',
                                                 output_hash_key => 'sis_files'},
                                                {indentification_string => 'Succeeded Op:\s+',
                                                 output_hash_key => 'succeeded_op'},
                                                {indentification_string => 'Started Op:\s+',
                                                 output_hash_key => 'started_op'},
                                                {indentification_string => 'Failed Op:\s+',
                                                 output_hash_key => 'failed_op'},
                                                {indentification_string => 'Stopped Op:\s+',
                                                 output_hash_key => 'stopped_op'},
                                                {indentification_string => 'Deferred Op:\s+',
                                                 output_hash_key => 'deferred_op'},
                                                {indentification_string => 'Succeeded Check Op:\s+',
                                                 output_hash_key => 'succeeded_check_op'},
                                                {indentification_string => 'Failed Check Op:\s+',
                                                 output_hash_key => 'failed_check_op'},
                                                {indentification_string => 'Suspended Check Op:\s+',
                                                 output_hash_key => 'suspended_check_op'},
                                                {indentification_string => 'Total FP Deleted:\s+',
                                                 output_hash_key => 'total_fp_deleted'},
                                                {indentification_string => 'Total Sorted Blocks:\s+',
                                                 output_hash_key => 'total_sorted_blocks'},
                                                {indentification_string => 'Overlapped Blocks:\s+',
                                                 output_hash_key => 'overlapped_blocks'},
                                                {indentification_string => 'Same Fingerprint:\s+',
                                                 output_hash_key => 'same_fingerprint'},
                                                {indentification_string => 'Same FBN Location:\s+',
                                                 output_hash_key => 'same_fbn_location'},
                                                {indentification_string => 'Same Data:\s+',
                                                 output_hash_key => 'same_data'},
                                                {indentification_string => 'Same VBN:\s+',
                                                 output_hash_key => 'same_vbn'},
                                                {indentification_string => 'Mismatched Data:\s+',
                                                 output_hash_key => 'mismatched_data'},
                                                {indentification_string => 'Max Reference Hits:\s+',
                                                 output_hash_key => 'max_reference_hits'},
                                                {indentification_string => 'Staled Recipient:\s+',
                                                 output_hash_key => 'staled_recipient'},
                                                {indentification_string => 'Staled Donor:\s+',
                                                 output_hash_key => 'staled_donor'},
                                                {indentification_string => 'File Too Small:\s+',
                                                 output_hash_key => 'file_too_small'},
                                                {indentification_string => 'Out of Space:\s+',
                                                 output_hash_key => 'out_of_space'},
                                                {indentification_string => 'FP False Match:\s+',
                                                 output_hash_key => 'fp_false_match'},
                                                {indentification_string => 'Mismatch By Overwrites:\s+',
                                                 output_hash_key => 'mismatch_by_overwrite'},
                                                {indentification_string => 'Delino Records:\s+',
                                                 output_hash_key => 'delino_records'},
                                                {indentification_string => 'Unaligned Compression blocks:\s+',
                                                 output_hash_key => 'unaligned_compression_blocks'},
                                                {indentification_string => 'Additional Sharing Messages:\s+',
                                                 output_hash_key => 'additional_sharing_messages'},
                                                {indentification_string => 'Records Sorted:\s+',
                                                 output_hash_key => 'records_sorted'},
                                                {indentification_string => 'Used 0 Splits:\s+',
                                                 output_hash_key => 'used_0_splits'},
                                                {indentification_string => 'Used 1 Split:\s+',
                                                 output_hash_key => 'used_1_split'},
                                                {indentification_string => 'Used 2 Splits:\s+',
                                                 output_hash_key => 'used_2_splits'},
                                                {indentification_string => 'Used 1 Pass:\s+',
                                                 output_hash_key => 'used_1_pass'},
                                                {indentification_string => 'Used 2 Passes:\s+',
                                                 output_hash_key => 'used_2_passes'},
                                                {indentification_string => 'Used 3 Passes:\s+',
                                                 output_hash_key => 'used_3_passes'},
                                                {indentification_string => 'P1Merges:\s+',
                                                 output_hash_key => 'p1_merges'},
                                                {indentification_string => 'Merged Bins:\s+',
                                                 output_hash_key => 'merged_bins'},
                                                {indentification_string => 'Compression Saved:\s+',
                                                 output_hash_key => 'saved_blocks_by_compression'},
                                                {indentification_string => 'CGs Decompressed:\s+',
                                                 output_hash_key => 'groups_decompressed'},
                                                {indentification_string => 'Partial CG Modifies:\s+',
                                                 output_hash_key => 'partial_cg_modifies'},
                                                {indentification_string => 'Avg Decompress Time:\s+',
                                                 output_hash_key => 'average_cg_decompress_time'},
                                                {indentification_string => 'Avg Compress Time:\s+',
                                                 output_hash_key => 'average_compress_time'},
                                                {indentification_string => 'Extra CP Reads:\s+',
                                                 output_hash_key => 'extra_cp_reads'},
                                                {indentification_string => 'CGs Compressed Inline:\s+',
                                                 output_hash_key => 'inline_compressed'},
                                                {indentification_string => 'CGs Compressed:\s+',
                                                 output_hash_key => 'background_compressed'},
                                                {indentification_string => 'Compressed Blocks:\s+',
                                                 output_hash_key => 'compressed_blocks'},
                                                {indentification_string => 'Uncompressed Blocks:\s+',
                                                 output_hash_key => 'uncompressed_blocks'},
                                                {indentification_string => 'New Partial CG Writes:\s+',
                                                 output_hash_key => 'new_partial_cg_writes'},
                                                {indentification_string => 'Decompress Disk Bad:\s+',
                                                 output_hash_key => 'decompress_disk_bad'},
                                                {indentification_string => 'Decompress SW Bad:\s+',
                                                 output_hash_key => 'decompress_sw_bad'},
                                                {indentification_string => 'Compression Attempts:\s+',
                                                 output_hash_key => 'compress_attempts'},
                                                {indentification_string => 'Compression Failures:\s+',
                                                 output_hash_key => 'compress_failed'},
                                                {indentification_string => 'Poor Compression Ratio:\s+',
                                                 output_hash_key => 'poor_compression_ratio'},
                                                {indentification_string => 'Shared Blocks Skipped:\s+',
                                                 output_hash_key => 'shared_blocks_skipped'},
						{indentification_string => 'Snapshot Blocks Skipped:\s*',
                                                 output_hash_key => 'snapshot_blocks_skipped'},
                                                {indentification_string => 'Un-Flushed Change Logs:\s+',
                                                 output_hash_key => 'unflushed_changelogs'},
                                                ],
                              },
                VERIFY_COMPR_ATTEMPTED_BIT => {ignore_lines => ['Checking Semi Upgraded temporary metafiles',
                                                                'Checking Downgraded temporary metafiles'],
                                               expected_info => [{indentification_string => 'Checking clogdir\.0\s+\.+\s+'},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_0'},
                                                                 {indentification_string => 'Checking clogdir\.1\s+\.+\s+',},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_1'},],
                              },
		GET_COMPRESSION_STATS => {expected_info => [{indentification_string => 'compression:\S+:num_blocks_saved:',
							     output_hash_key => 'num_blocks_saved'},
							    {indentification_string => 'compression:\S+:num_blocks_cdata:',
							     output_hash_key => 'num_blocks_cdata'},
							    {indentification_string => 'compression:\S+:num_blocks_uncompressed:',
							     output_hash_key => 'num_blocks_uncompressed'},
							    {indentification_string => 'compression:\S+:num_buf_decompressed:',
							     output_hash_key => 'num_buf_decompressed'},
							    {indentification_string => 'compression:\S+:num_cg_decompressed:',
							     output_hash_key => 'num_cg_decompressed'},
							    {indentification_string => 'compression:\S+:num_partial_write:',
							     output_hash_key => 'num_partial_write'},
							    {indentification_string => 'compression:\S+:num_valid_buf:',
							     output_hash_key => 'num_valid_buf'},
							    {indentification_string => 'compression:\S+:num_invalid_raid_read:',
							     output_hash_key => 'num_invalid_raid_read'},
							    {indentification_string => 'compression:\S+:num_valid_buf_suspended:',
							     output_hash_key => 'num_valid_buf_suspended'},
							    {indentification_string => 'compression:\S+:num_cow_cg:',
							     output_hash_key => 'num_cow_cg'},
							    {indentification_string => 'compression:\S+:num_blocking_io:',
							     output_hash_key => 'num_blocking_io'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_ahead_busy:',
							     output_hash_key => 'num_vol_decomp_ahead_busy'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_no_resource:',
							     output_hash_key => 'num_vol_decomp_no_resource'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_zero_holes:',
							     output_hash_key => 'num_vol_decomp_zero_holes'},
							    {indentification_string => 'compression:\S+:num_decompress_suspends:',
							     output_hash_key => 'num_decompress_suspends'},
							    {indentification_string => 'compression:\S+:avg_vol_decompress_time:',
							     output_hash_key => 'avg_vol_decompress_time'},
							    {indentification_string => 'compression:\S+:avg_decompress_cg_time:',
							     output_hash_key => 'avg_decompress_cg_time'},
							    {indentification_string => 'compression:\S+:num_compress_attempts:',
							     output_hash_key => 'num_compress_attempts'},
							    {indentification_string => 'compression:\S+:num_compress_fail:',
							     output_hash_key => 'num_compress_fail'},
							    {indentification_string => 'compression:\S+:num_not_enough_savings:',
							     output_hash_key => 'num_not_enough_savings'},
							    {indentification_string => 'compression:\S+:num_no_mem:',
							     output_hash_key => 'num_no_mem'},
							    {indentification_string => 'compression:\S+:num_compress_suspends:',
							     output_hash_key => 'num_compress_suspends'},
							    {indentification_string => 'compression:\S+:num_cgs_frz_attempts:',
							     output_hash_key => 'num_cgs_frz_attempts'},
							    {indentification_string => 'compression:\S+:num_cgs_frz_failed:',
							     output_hash_key => 'num_cgs_frz_failed'},
							    {indentification_string => 'compression:\S+:num_buf_attempts:',
							     output_hash_key => 'num_buf_attempts'},
							    {indentification_string => 'compression:\S+:num_nocompress_files:',
							     output_hash_key => 'num_nocompress_files'},
							    {indentification_string => 'compression:\S+:num_min_input:',
							     output_hash_key => 'num_min_input'},
							    {indentification_string => 'compression:\S+:num_cgs_with_holes:',
							     output_hash_key => 'num_cgs_with_holes'},
							    {indentification_string => 'compression:\S+:num_next_cp:',
							     output_hash_key => 'num_next_cp'},
							    {indentification_string => 'compression:\S+:num_wait_for_compressor:',
							     output_hash_key => 'num_wait_for_compressor'},
							    {indentification_string => 'compression:\S+:avg_compress_cg_time:',
							     output_hash_key => 'avg_compress_cg_time'},
							    {indentification_string => 'compression:\S+:num_extra_cp_reads:',
							     output_hash_key => 'num_extra_cp_reads'},
							    {indentification_string => 'compression:\S+:num_compress_ahead:',
							     output_hash_key => 'num_compress_ahead'},
							    {indentification_string => 'compression:\S+:num_overload:',
							     output_hash_key => 'num_overload'},
							    {indentification_string => 'compression:\S+:hole_to_data:',
							     output_hash_key => 'hole_to_data'},
							    {indentification_string => 'compression:\S+:hole_to_compressed:',
							     output_hash_key => 'hole_to_compressed'},
							    {indentification_string => 'compression:\S+:data_to_hole:',
							     output_hash_key => 'data_to_hole'},
							    {indentification_string => 'compression:\S+:num_write_uncompressed:',
							     output_hash_key => 'num_write_uncompressed'},
							    {indentification_string => 'compression:\S+:num_eos_bufs:',
							     output_hash_key => 'num_eos_bufs'},
							    {indentification_string => 'compression:\S+:num_cgs_not_fully_dirty:',
							     output_hash_key => 'num_cgs_not_fully_dirty'},
							    {indentification_string => 'compression:\S+:num_wafl_decompress:',
							     output_hash_key => 'num_wafl_decompress'},
							    {indentification_string => 'compression:\S+:num_decompress_read_bad_blks:',
							     output_hash_key => 'num_decompress_read_bad_blks'},
							    {indentification_string => 'compression:\S+:num_decompress_mark_hw_bad:',
							     output_hash_key => 'num_decompress_mark_hw_bad'},
							    {indentification_string => 'compression:\S+:num_decompress_mark_sw_bad:',
							     output_hash_key => 'num_decompress_mark_sw_bad'},
							    {indentification_string => 'compression:\S+:num_compress_mark_hw_bad:',
							     output_hash_key => 'num_compress_mark_hw_bad'},
							    {indentification_string => 'compression:\S+:num_compress_mark_sw_bad:',
							     output_hash_key => 'num_compress_mark_sw_bad'},
							    {indentification_string => 'compression:\S+:num_compressed_inline:',
							     output_hash_key => 'num_compressed_inline'},
							    {indentification_string => 'compression:\S+:num_compressed_bg:',
							     output_hash_key => 'num_compressed_bg'},
							    {indentification_string => 'compression:\S+:num_partial_cg_modifies:',
							     output_hash_key => 'num_partial_cg_modifies'},
							    {indentification_string => 'compression:\S+:num_partial_cg_new:',
							     output_hash_key => 'num_partial_cg_new'},
							    ],
					 ignore_lines => ['StatisticsID: \S+',
							  'compression:\S+:instance_name:',
							  'compression:\S+:instance_uuid:',
							  'compression:\S+:node_name:'],
					 },
        };
	my $code = $args{code};
        
        return $formats_hash_8_1->{$code}
}


sub _7mode_formats_8_2
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash = {
                COMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                    ignore_lines => ['Policy ID:\s*',
                                                     'Stale Fingerprints.*:\s+',
                                                     # TODO: move these to expected_info once burt 410442
                                                     # is merged to main
                                                     'Last Operation State:\s*',
                                                     'Last Successful Operation Begin:\s*',
                                                     'Last Successful Operation End:\s*',
						     # The following 5 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
						     # If any of these fields become needed for testing, they will need to be moved to the expected_info hash
						     'Optimize:\s*',
						     'Change Log Size:\s*',
						     'Gathering Phase 2 Begin:\s*',
						     'Compression Quick Check File Size:\s*',
						     'Change Log Overflow:\s*',
						     'Change Log Size:\s*',
					 	     'Vault Transfer Log Size:\s*',
						     'Compression Change Log Size:\s*',
                                                     # burt613302
                                                     'Total Sorted Blocks:\s*',
                                                     'Same Fingerprint:\s*',
                                                     'Same FBN Location:\s*',
                                                     'Same Data:\s*',
													 'No Op:\s*',
                                                     'Same VBN:\s*',
                                                     'Mismatched Data:\s*',
                                                     'Same Sharing Records:\s*',
                                                     'Max Reference Hits:\s*',
                                                     'Staled Recipient:\s*',
                                                     'Staled Donor:\s*',
                                                     'Out of Space:\s*',
                                                     'Mismatch By Overwrites:\s*',
                                                     'Staled Aux Recipient:\s*',
                                                     'Staled Aux Recipient Blocks:\s*',
                                                     'Mismatched Aux Recipient:\s*',
                                                     'Skipped Aux Recipient Sharing:\s*',
                                                     'Skip Share Blocks Delta:\s*',
                                                     'Skip Share Blocks Upper:\s*',
                                                    ],
                                    expected_info => [{indentification_string => 'State:\s+',
                                                       output_hash_key => 'state'},
                                                      {indentification_string => 'Compression:\s+',
                                                       output_hash_key => 'compression'},
                                                      {indentification_string => 'Inline Compression:\s+',
                                                       output_hash_key => 'inline_compression'},
						      {indentification_string => 'Incompressible Data Detection:\s+',
						       output_hash_key => 'idd'},
						      #{indentification_string => 'Compression Quick Check File Size:\s+',
						      # output_hash_key => 'quick_check_file_size'},
                                                      {indentification_string => 'Status:\s+',
                                                       output_hash_key => 'status'},
                                                      {indentification_string => 'Progress:\s+',
                                                       output_hash_key => 'progress',
                                                       extract_information => [{regex => '\d+',
                                                                                output_hash_key => 'data_scanned',
                                                                               },],},
                                                      {indentification_string => 'Type:\s+',
                                                       output_hash_key => 'type'},
                                                      {indentification_string => 'Schedule:\s+',
                                                       output_hash_key => 'schedule'},
                                                      {indentification_string => 'Minimum Blocks Shared:\s+',
                                                       output_hash_key => 'minimum_blocks_shared'},
                                                      {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                       output_hash_key => 'blocks_skipped_sharing'},
                                                      {indentification_string => 'Last Operation Begin:\s+',
                                                       output_hash_key => 'last_operation_begin'},
                                                      {indentification_string => 'Last Operation End:\s+',
                                                       output_hash_key => 'last_operation_end'},
                                                      {indentification_string => 'Last Operation Size:\s+',
                                                       output_hash_key => 'last_operation_size'},
                                                      {indentification_string => 'Last Operation Error:\s+',
                                                       output_hash_key => 'last_operation_error'},
                                                      {indentification_string => 'Change Log Usage:\s+',
                                                       output_hash_key => 'changelog_usage'},
						      #{indentification_string => 'Change Log Overflow:\s+',
						      # output_hash_key => 'changelog_overflow'},
                                                      {indentification_string => 'Logical Data:\s+',
                                                       output_hash_key => 'logical_data'},
                                                      {indentification_string => 'Queued Job:\s+',
                                                       output_hash_key => 'queued_jobs'},
                                                      {indentification_string => 'Stage:\s+',
                                                       output_hash_key => 'stage'},
                                                      {indentification_string => 'Checkpoint Time:\s+',
                                                       output_hash_key => 'checkpoint_time'},
                                                      {indentification_string => 'Checkpoint Operation Type:\s+',
                                                       output_hash_key => 'checkpoint_op_type'},
                                                      {indentification_string => 'Checkpoint Stage:\s+',
                                                       output_hash_key => 'checkpoint_stage'},
                                                      {indentification_string => 'Checkpoint Sub-*stage:\s+',
                                                       output_hash_key => 'checkpoint_substage'},
                                                      {indentification_string => 'Checkpoint Progress:\s+',
                                                       output_hash_key => 'checkpoint_progress'},
                                                      {indentification_string => 'Fingerprints Gathered:\s+',
                                                       output_hash_key => 'fingerprints_gathered'},
                                                      {indentification_string => 'Gathering Begin:\s+',
                                                       output_hash_key => 'gathering_begin'},
                                                      {indentification_string => 'Fingerprints Sorted:\s+',
                                                       output_hash_key => 'fingerprints_sorted'},
                                                      {indentification_string => 'Duplicate Blocks Found:\s+',
                                                       output_hash_key => 'duplicated_blocks_found'},
                                                      {indentification_string => 'Sorting Begin:\s+',
                                                       output_hash_key => 'sorting_begin'},
                                                      {indentification_string => 'Blocks De-duplicated:\s+',
                                                       output_hash_key => 'blocks_deduplicated'},
                                                      {indentification_string => 'De-duping Begin:\s+',
                                                       output_hash_key => 'dedupe_begin'},
                                                      {indentification_string => 'Fingerprints Deleted:\s+',
                                                       output_hash_key => 'fingerprints_deleted'},
                                                      {indentification_string => 'Checking Begin:\s+',
                                                       output_hash_key => 'checking_begin'},
                                                      {indentification_string => 'Verify Trigger Rate:\s+',
                                                       output_hash_key => 'verify_trigger_rate'},
                                                      {indentification_string => 'Total Verify Time:\s+',
                                                       output_hash_key => 'total_verify_time'},
                                                      {indentification_string => 'Verify Suspend Count:\s+',
                                                       output_hash_key => 'verify_suspend_count'},
                                                      ],    
                                               },
                COMPRESS_STATUS_CHECKPOINT => {header => 'Path:\s+/vol/',
                                               ignore_lines => ['Policy ID:\s*',
                                                                # TODO: move these to expected_info once burt 410442
                                                                # is merged to main
                                                                'Last Operation State:\s*',
                                                                'Last Successful Operation Begin:\s*',
                                                                'Last Successful Operation End:\s*',
								# The following 5 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
								# If any of these fields become needed for testing, they will need to be moved to the expected_info hash
								'Optimize:\s*',
								'Change Log Size:\s*',
								'Gathering Phase 2 Begin:\s*',
								'Compression Quick Check File Size:\s*',
								'Change Log Overflow:\s*',
						     		'Change Log Size:\s*',
					 	     		'Vault Transfer Log Size:\s*',
						     		'Compression Change Log Size:\s*',
                                                     # burt613302
                                                     'Total Sorted Blocks:\s*',
                                                     'Same Fingerprint:\s*',
                                                     'Same FBN Location:\s*',
                                                     'Same Data:\s*',
													 'No Op:\s*',
                                                     'Same VBN:\s*',
                                                     'Mismatched Data:\s*',
                                                     'Same Sharing Records:\s*',
                                                     'Max Reference Hits:\s*',
                                                     'Staled Recipient:\s*',
                                                     'Staled Donor:\s*',
                                                     'Out of Space:\s*',
                                                     'Mismatch By Overwrites:\s*',
                                                     'Staled Aux Recipient:\s*',
                                                     'Staled Aux Recipient Blocks:\s*',
                                                     'Mismatched Aux Recipient:\s*',
                                                     'Skipped Aux Recipient Sharing:\s*',
                                                     'Skip Share Blocks Delta:\s*',
                                                     'Skip Share Blocks Upper:\s*',
                                                                ],
                                               expected_info => [{indentification_string => 'State:\s+',
                                                                  output_hash_key => 'state'},
                                                                 {indentification_string => 'Compression:\s+',
                                                                  output_hash_key => 'compression'},
                                                                 {indentification_string => 'Inline Compression:\s+',
                                                                  output_hash_key => 'inline_compression'},
								 {indentification_string => 'Incompressible Data Detection:\s+',
								  output_hash_key => 'idd'},
								 #{indentification_string => 'Compression Quick Check File Size:\s+',
								 # output_hash_key => 'quick_check_file_size'},
                                                                 {indentification_string => 'Status:\s+',
                                                                  output_hash_key => 'status'},
                                                                 {indentification_string => 'Progress:\s+',
                                                                  output_hash_key => 'progress',
                                                                  extract_information => [{regex => '\d+',
                                                                                           output_hash_key => 'total_data_scanned',
                                                                                           },],},
                                                                 {indentification_string => 'Type:\s+',
                                                                  output_hash_key => 'type'},
                                                                 {indentification_string => 'Schedule:\s+',
                                                                  output_hash_key => 'schedule'},
                                                                 {indentification_string => 'Minimum Blocks Shared:\s+',
                                                                  output_hash_key => 'minimum_blocks_shared'},
                                                                 {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                                  output_hash_key => 'blocks_skipped_sharing'},
                                                                 {indentification_string => 'Last Operation Begin:\s+',
                                                                  output_hash_key => 'last_operation_begin'},
                                                                 {indentification_string => 'Last Operation End:\s+',
                                                                  output_hash_key => 'last_operation_end'},
                                                                 {indentification_string => 'Last Operation Size:\s+',
                                                                  output_hash_key => 'last_operation_size'},
                                                                 {indentification_string => 'Last Operation Error:\s+',
                                                                  output_hash_key => 'last_operation_error'},
                                                                 {indentification_string => 'Change Log Usage:\s+',
                                                                  output_hash_key => 'changelog_usage'},
								 #{indentification_string => 'Change Log Overflow:\s+',
								 # output_hash_key => 'changelog_overflow'},
                                                                 {indentification_string => 'Logical Data:\s+',
                                                                  output_hash_key => 'logical_data'},
                                                                 {indentification_string => 'Queued Job:\s+',
                                                                  output_hash_key => 'queued_jobs'},
                                                                 {indentification_string => 'Stale Fingerprints.*:\s+',
                                                                  output_hash_key => 'stale_fingerprints'},
                                                                 {indentification_string => 'Stage:\s+',
                                                                  output_hash_key => 'stage'},
                                                                 {indentification_string => 'Checkpoint Time:\s+',
                                                                  output_hash_key => 'checkpoint_creation_time'},
                                                                 {indentification_string => 'Checkpoint Operation Type:\s+',
                                                                  output_hash_key => 'checkpoint_op_type'},
                                                                 {indentification_string => 'Checkpoint Stage:\s+',
                                                                  output_hash_key => 'checkpoint_stage'},
                                                                 {indentification_string => 'Checkpoint Sub-*stage:\s+',
                                                                  output_hash_key => 'checkpoint_substage'},
                                                                 {indentification_string => 'Checkpoint Progress:\s+',
                                                                  output_hash_key => 'checkpoint_progress'},
                                                                 {indentification_string => 'Fingerprints Gathered:\s+',
                                                                  output_hash_key => 'fingerprints_gathered'},
                                                                 {indentification_string => 'Gathering Begin:\s+',
                                                                  output_hash_key => 'gathering_begin'},
                                                                 {indentification_string => 'Fingerprints Sorted:\s+',
                                                                  output_hash_key => 'fingerprints_sorted'},
                                                                 {indentification_string => 'Duplicate Blocks Found:\s+',
                                                                  output_hash_key => 'duplicated_blocks_found'},
                                                                 {indentification_string => 'Sorting Begin:\s+',
                                                                  output_hash_key => 'sorting_begin'},
                                                                 {indentification_string => 'Blocks De-duplicated:\s+',
                                                                  output_hash_key => 'blocks_deduplicated'},
                                                                 {indentification_string => 'De-duping Begin:\s+',
                                                                  output_hash_key => 'dedupe_begin'},
                                                                 {indentification_string => 'Fingerprints Deleted:\s+',
                                                                  output_hash_key => 'fingerprints_deleted'},
                                                                 {indentification_string => 'Checking Begin:\s+',
                                                                  output_hash_key => 'checking_begin'},
                                                                 {indentification_string => 'Verify Trigger Rate:\s+',
                                                                  output_hash_key => 'verify_trigger_rate'},
                                                                 {indentification_string => 'Total Verify Time:\s+',
                                                                  output_hash_key => 'total_verify_time'},
                                                                 {indentification_string => 'Verify Suspend Count:\s+',
                                                                  output_hash_key => 'verify_suspend_count'},
                                                                 ],
                                                
                                              },
                DECOMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                      ignore_lines => ['Policy ID:\s*',
                                                       # TODO: move these to expected_info once burt 410442
                                                       # is merged to main
                                                       'Last Operation State:\s*',
                                                       'Last Successful Operation Begin:\s*',
                                                       'Last Successful Operation End:\s*',
						       # The following 5 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
						       # If any of these fields become needed for testing, they will need to be moved to the expected_info hash
						       'Optimize:\s*',
						       'Change Log Size:\s*',
						       'Gathering Phase 2 Begin:\s*',
						       'Compression Quick Check File Size:\s*',
						       'Change Log Overflow:\s*',
						       'Change Log Size:\s*',
					 	       'Vault Transfer Log Size:\s*',
						       'Compression Change Log Size:\s*',
                                                     # burt613302
                                                     'Total Sorted Blocks:\s*',
                                                     'Same Fingerprint:\s*',
                                                     'Same FBN Location:\s*',
                                                     'Same Data:\s*',
													 'No Op:\s*',
                                                     'Same VBN:\s*',
                                                     'Mismatched Data:\s*',
                                                     'Same Sharing Records:\s*',
                                                     'Max Reference Hits:\s*',
                                                     'Staled Recipient:\s*',
                                                     'Staled Donor:\s*',
                                                     'Out of Space:\s*',
                                                     'Mismatch By Overwrites:\s*',
                                                     'Staled Aux Recipient:\s*',
                                                     'Staled Aux Recipient Blocks:\s*',
                                                     'Mismatched Aux Recipient:\s*',
                                                     'Skipped Aux Recipient Sharing:\s*',
                                                     'Skip Share Blocks Delta:\s*',
                                                     'Skip Share Blocks Upper:\s*',
                                                       ],
                                      expected_info => [{indentification_string => 'State:\s+',
                                                         output_hash_key => 'state'},
                                                        {indentification_string => 'Compression:\s+',
                                                         output_hash_key => 'compression'},
                                                        {indentification_string => 'Inline Compression:\s+',
                                                         output_hash_key => 'inline_compression'},
							{indentification_string => 'Incompressible Data Detection:\s+',
							 output_hash_key => 'idd'},
							#{indentification_string => 'Compression Quick Check File Size:\s+',
							# output_hash_key => 'quick_check_file_size'},
                                                        {indentification_string => 'Status:\s+',
                                                         output_hash_key => 'status'},
                                                        {indentification_string => 'Progress:\s+',
                                                         output_hash_key => 'progress',
                                                         extract_information => [{regex => '\d+',
                                                                                  output_hash_key => 'data_processed',
                                                                                  },
                                                                                 {regex => '\s+KB\s+',
                                                                                 },
                                                                                 {regex => '\w+',
                                                                                  output_hash_key => 'processing_phase',
                                                                                 },],},
                                                        {indentification_string => 'Type:\s+',
                                                         output_hash_key => 'type'},
                                                        {indentification_string => 'Schedule:\s+',
                                                         output_hash_key => 'schedule'},
                                                        {indentification_string => 'Minimum Blocks Shared:\s+',
                                                         output_hash_key => 'minimum_blocks_shared'},
                                                        {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                         output_hash_key => 'blocks_skipped_sharing'},
                                                        {indentification_string => 'Last Operation Begin:\s+',
                                                         output_hash_key => 'last_operation_begin'},
                                                        {indentification_string => 'Last Operation End:\s+',
                                                         output_hash_key => 'last_operation_end'},
                                                        {indentification_string => 'Last Operation Size:\s+',
                                                         output_hash_key => 'last_operation_size'},
                                                        {indentification_string => 'Last Operation Error:\s+',
                                                         output_hash_key => 'last_operation_error'},
                                                        {indentification_string => 'Change Log Usage:\s+',
                                                         output_hash_key => 'changelog_usage'},
							#{indentification_string => 'Change Log Overflow:\s+',
							# output_hash_key => 'changelog_overflow'},
                                                        {indentification_string => 'Logical Data:\s+',
                                                         output_hash_key => 'logical_data'},
                                                        {indentification_string => 'Queued Job:\s+',
                                                         output_hash_key => 'queued_jobs'},
                                                        {indentification_string => 'Stale Fingerprints.*:\s+',
                                                         output_hash_key => 'stale_fingerprints'},
                                                        {indentification_string => 'Stage:\s+',
                                                         output_hash_key => 'stage'},
                                                        {indentification_string => 'Checkpoint Time:\s+',
                                                         output_hash_key => 'checkpoint_time'},
                                                        {indentification_string => 'Checkpoint Operation Type:\s+',
                                                         output_hash_key => 'checkpoint_op_type'},
                                                        {indentification_string => 'Checkpoint Stage:\s+',
                                                         output_hash_key => 'checkpoint_stage'},
                                                        {indentification_string => 'Checkpoint Sub-*stage:\s+',
                                                         output_hash_key => 'checkpoint_substage'},
                                                        {indentification_string => 'Checkpoint Progress:\s+',
                                                         output_hash_key => 'checkpoint_progress'},
                                                        {indentification_string => 'Fingerprints Gathered:\s+',
                                                         output_hash_key => 'fingerprints_gathered'},
                                                        {indentification_string => 'Gathering Begin:\s+',
                                                         output_hash_key => 'gathering_begin'},
                                                        {indentification_string => 'Fingerprints Sorted:\s+',
                                                         output_hash_key => 'fingerprints_sorted'},
                                                        {indentification_string => 'Duplicate Blocks Found:\s+',
                                                         output_hash_key => 'duplicated_blocks_found'},
                                                        {indentification_string => 'Sorting Begin:\s+',
                                                         output_hash_key => 'sorting_begin'},
                                                        {indentification_string => 'Blocks De-duplicated:\s+',
                                                         output_hash_key => 'blocks_deduplicated'},
                                                        {indentification_string => 'De-duping Begin:\s+',
                                                         output_hash_key => 'dedupe_begin'},
                                                        {indentification_string => 'Fingerprints Deleted:\s+',
                                                         output_hash_key => 'fingerprints_deleted'},
                                                        {indentification_string => 'Checking Begin:\s+',
                                                         output_hash_key => 'checking_begin'},
                                                        {indentification_string => 'Verify Trigger Rate:\s+',
                                                         output_hash_key => 'verify_trigger_rate'},
                                                        {indentification_string => 'Total Verify Time:\s+',
                                                         output_hash_key => 'total_verify_time'},
                                                        {indentification_string => 'Verify Suspend Count:\s+',
                                                         output_hash_key => 'verify_suspend_count'},
                                                      ],
                                               },
                GET_STATS => {header => 'Path:\s+/vol/',
                              ignore_lines => ['^\s*<*(\d+|UNUSED)\s*<*(\d+|UNUSED)',
                                               'Block sharing histogram:',
					       'Average:.*',
                                               'Reference count histogram:',
                                               'Sort Stats:',
											   'No Op:\s*'
                                               ],
                              expected_info => [{indentification_string => 'Allocated:\s+',
                                                 output_hash_key => 'allocated'},
                                                {indentification_string => 'Shared:\s+',
                                                 output_hash_key => 'shared'},
                                                {indentification_string => 'Saving:\s+',
                                                 output_hash_key => 'saving'},
                                                {indentification_string => '%Saved:\s+',
                                                 output_hash_key => 'percentage_saved'},
                                                {indentification_string => 'Max Refcount:\s+',
                                                 output_hash_key => 'max_refcount'},
                                                {indentification_string => 'Total Processed:\s+',
                                                 output_hash_key => 'total_processed'},
                                                {indentification_string => 'Total Process Time:\s+',
                                                 output_hash_key => 'total_process_time'},
                                                {indentification_string => 'Total Verify Time:\s+',
                                                 output_hash_key => 'total_verify_time'},
                                                {indentification_string => 'SIS Files:\s+',
                                                 output_hash_key => 'sis_files'},
                                                {indentification_string => 'Succeeded Op:\s+',
                                                 output_hash_key => 'succeeded_op'},
                                                {indentification_string => 'Started Op:\s+',
                                                 output_hash_key => 'started_op'},
                                                {indentification_string => 'Failed Op:\s+',
                                                 output_hash_key => 'failed_op'},
                                                {indentification_string => 'Stopped Op:\s+',
                                                 output_hash_key => 'stopped_op'},
                                                {indentification_string => 'Deferred Op:\s+',
                                                 output_hash_key => 'deferred_op'},
                                                {indentification_string => 'Succeeded Check Op:\s+',
                                                 output_hash_key => 'succeeded_check_op'},
                                                {indentification_string => 'Failed Check Op:\s+',
                                                 output_hash_key => 'failed_check_op'},
                                                {indentification_string => 'Suspended Check Op:\s+',
                                                 output_hash_key => 'suspended_check_op'},
                                                {indentification_string => 'Total FP Deleted:\s+',
                                                 output_hash_key => 'total_fp_deleted'},
                                                {indentification_string => 'Total Sorted Blocks:\s+',
                                                 output_hash_key => 'total_sorted_blocks'},
                                                {indentification_string => 'Overlapped Blocks:\s+',
                                                 output_hash_key => 'overlapped_blocks'},
                                                {indentification_string => 'Same Fingerprint:\s+',
                                                 output_hash_key => 'same_fingerprint'},
                                                {indentification_string => 'Same FBN Location:\s+',
                                                 output_hash_key => 'same_fbn_location'},
                                                {indentification_string => 'Same Data:\s+',
                                                 output_hash_key => 'same_data'},
                                                {indentification_string => 'Same VBN:\s+',
                                                 output_hash_key => 'same_vbn'},
                                                {indentification_string => 'Mismatched Data:\s+',
                                                 output_hash_key => 'mismatched_data'},
                                                {indentification_string => 'Max Reference Hits:\s+',
                                                 output_hash_key => 'max_reference_hits'},
                                                {indentification_string => 'Staled Recipient:\s+',
                                                 output_hash_key => 'staled_recipient'},
                                                {indentification_string => 'Staled Donor:\s+',
                                                 output_hash_key => 'staled_donor'},
						{indentification_string => 'Skip Share Blocks Delta:\s+',
						 output_hash_key => 'skip_share_blocks_delta'},
						{indentification_string => 'Skip Share Blocks Upper:\s+',
						 output_hash_key => 'skip_share_blocks_upper'},
                                                {indentification_string => 'File Too Small:\s+',
                                                 output_hash_key => 'file_too_small'},
                                                {indentification_string => 'Out of Space:\s+',
                                                 output_hash_key => 'out_of_space'},
                                                {indentification_string => 'FP False Match:\s+',
                                                 output_hash_key => 'fp_false_match'},
                                                {indentification_string => 'Mismatch By Overwrites:\s+',
                                                 output_hash_key => 'mismatch_by_overwrite'},
                                                {indentification_string => 'Delino Records:\s+',
                                                 output_hash_key => 'delino_records'},
                                                {indentification_string => 'Unaligned Compression Blocks:\s+',
                                                 output_hash_key => 'unaligned_compression_blocks'},
                                                {indentification_string => 'Additional Sharing Messages:\s+',
                                                 output_hash_key => 'additional_sharing_messages'},
                                                {indentification_string => 'Records Sorted:\s+',
                                                 output_hash_key => 'records_sorted'},
                                                {indentification_string => 'Used 0 Splits:\s+',
                                                 output_hash_key => 'used_0_splits'},
                                                {indentification_string => 'Used 1 Split:\s+',
                                                 output_hash_key => 'used_1_split'},
                                                {indentification_string => 'Used 2 Splits:\s+',
                                                 output_hash_key => 'used_2_splits'},
                                                {indentification_string => 'Used 1 Pass:\s+',
                                                 output_hash_key => 'used_1_pass'},
                                                {indentification_string => 'Used 2 Passes:\s+',
                                                 output_hash_key => 'used_2_passes'},
                                                {indentification_string => 'Used 3 Passes:\s+',
                                                 output_hash_key => 'used_3_passes'},
                                                {indentification_string => 'P1Merges:\s+',
                                                 output_hash_key => 'p1_merges'},
                                                {indentification_string => 'Merged Bins:\s+',
                                                 output_hash_key => 'merged_bins'},
                                                {indentification_string => 'Compression Saved:\s+',
                                                 output_hash_key => 'saved_blocks_by_compression'},
                                                {indentification_string => 'CGs Decompressed:\s+',
                                                 output_hash_key => 'groups_decompressed'},
                                                {indentification_string => 'Partial CG Modifies:\s+',
                                                 output_hash_key => 'partial_cg_modifies'},
                                                {indentification_string => 'Avg Decompress Time:\s+',
                                                 output_hash_key => 'average_cg_decompress_time'},
                                                {indentification_string => 'Avg Compress Time:\s+',
                                                 output_hash_key => 'average_compress_time'},
						{indentification_string => 'Avg Incompressible Data Quick Check Time:\s+',
						 output_hash_key => 'average_incompressible_data_quick_check_time'},
						{indentification_string => 'Avg Compressible Data Quick Check Time:\s+',
						 output_hash_key => 'average_compressible_data_quick_check_time'},
                                                {indentification_string => 'Extra CP reads:\s+',
                                                 output_hash_key => 'extra_cp_reads'},
                                                {indentification_string => 'Inline Compression Attempts:\s+',
                                                 output_hash_key => 'inline_compress_attempts'},
                                                {indentification_string => 'CGs Compressed:\s+',
                                                 output_hash_key => 'background_compressed'},
                                                {indentification_string => 'Compressed Blocks:\s+',
                                                 output_hash_key => 'compressed_blocks'},
                                                {indentification_string => 'Uncompressed Blocks:\s+',
                                                 output_hash_key => 'uncompressed_blocks'},
                                                {indentification_string => 'New Partial CG Writes:\s+',
                                                 output_hash_key => 'new_partial_cg_writes'},
                                                {indentification_string => 'Decompress Disk Bad:\s+',
                                                 output_hash_key => 'decompress_disk_bad'},
                                                {indentification_string => 'Decompress SW Bad:\s+',
                                                 output_hash_key => 'decompress_sw_bad'},
						{indentification_string => 'Inline Incompressible CGs:\s+',
                                                 output_hash_key => 'inline_incompressible_cgs'},
						{indentification_string => 'Incompressible CGs Found By Quick Check:\s+',
                                                 output_hash_key => 'quick_check_inline_incompressible_cgs'},
                                                {indentification_string => 'Compress Attempts:\s+',
                                                 output_hash_key => 'compress_attempts'},
                                                {indentification_string => 'Compression Failures:\s+',
                                                 output_hash_key => 'compress_failed'},
                                                {indentification_string => 'Poor Compression Ratio:\s+',
                                                 output_hash_key => 'poor_compression_ratio'},
                                                #burt635279
						{indentification_string => 'BCE Aborts Before Compress Stage:\s+',
						 output_hash_key => 'bce_aborts_before_compression_stage'},
						{indentification_string => 'BCE Aborts Due to Stale Inode Before Compress Stage:\s+',
						 output_hash_key => 'bce_aborts_due_to_stale_inode_before_compression_stage'},
						{indentification_string => 'BCE Aborts Due to Invalid FBN:\s+',
						 output_hash_key => 'bce_aborts_due_to_invalid_fbn'},
						{indentification_string => 'BCE Policy Stage Entries:\s+',
						 output_hash_key => 'bce_policy_stage_entries'},
						{indentification_string => 'BCE Compress Stage Entries:\s+',
						 output_hash_key => 'bce_compress_stage_entries'},
						{indentification_string => 'BCE CGs Skipped Due to Overwrites in Compress Stage:\s+',
						 output_hash_key => 'bce_cgs_skipped_due_to_overwrites_in_compress_stage'},
						{indentification_string => 'BCE Messages Sent to Exempt Domain:\s+',
						 output_hash_key => 'bce_messages_sent_to_exempt_domain'},
						{indentification_string => 'BCE Messages Received in Exempt Domain:\s+',
						 output_hash_key => 'bce_messages_received_in_extempt_domain'},
						{indentification_string => 'BCE SetFlag Stage Entries:\s+',
						 output_hash_key => 'bce_setflag_stage_entries'},
						{indentification_string => 'BCE Aborts Before Post Processing:\s+',
						 output_hash_key => 'bce_aborts_before_post_processing'},
						{indentification_string => 'BCE Aborts Due to Stale Inode Before Post Processing:\s+',
						 output_hash_key => 'bce_aborts_due_to_stale_inode_before_post_processing'},
						{indentification_string => 'BCE CGs Skipped Due to Overwrites in Post Processing:\s+',
						 output_hash_key => 'bce_cgs_skipped_due_to_overwrites_in_post_processing'},
						{indentification_string => 'BCE CGs Skipped Due to No Space in Post Processing:\s+',
						 output_hash_key => 'bce_messages_discarded_because_of_no_space'},
                                                {indentification_string => 'Shared Blocks Skipped:\s+',
                                                 output_hash_key => 'shared_blocks_skipped'},
						{indentification_string => 'Snapshot Blocks Skipped:\s*',
                                                 output_hash_key => 'snapshot_blocks_skipped'},
                                                {indentification_string => 'Un-Flushed Change Logs:\s+',
                                                 output_hash_key => 'unflushed_changelogs'},
                                                {indentification_string => 'Un-Flushed Data Logs:\s+',
                                                 output_hash_key => 'unflushed_changelogs'},
                                                {indentification_string => 'Un-Flushed Ref Logs:\s+',
                                                 output_hash_key => 'unflushed_reflogs'},
                                                {indentification_string => 'Skip Share Blocks Delta:\s+',
                                                 output_hash_key => 'skip_share_blocks_delta'},
                                                {indentification_string => 'Skip Share Blocks Upper:\s+',
                                                 output_hash_key => 'skip_share_blocks_upper'},
                                                ],
                              },
                VERIFY_COMPR_ATTEMPTED_BIT => {ignore_lines => ['Checking Semi Upgraded temporary metafiles',
                                                                'Checking Downgraded temporary metafiles'],
                                               expected_info => [{indentification_string => 'Checking clogdir\.0\s+\.+\s+'},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_0'},
                                                                 {indentification_string => 'Checking clogdir\.1\s+\.+\s+',},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_1'},],
                              },
        };
	my $code = $args{code};
        
        return $formats_hash->{$code}
}




################ C-mode #################
sub _Cmode_base_formats
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );

	logcomment("Formats Type 2: _Cmode_base_formats input:\n". Dumper(%args) ) if( $class->{_debug} );

	my $base_formats_hash = {
                COMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                    ignore_lines => ['Vserver:\s+',
						     'Auto State:',
                                                     'Policy ID:\s*',
                                                     'Efficiency Policy:\s*',
                                                     'Stale Fingerprints.*:\s*',
						     # TODO: move these to expected_info once burt 410442
						     # is merged to main
						     'Last Operation State:\s*',
						     'Last Successful Operation Begin:\s*',
						     'Last Successful Operation End:\s*',
						     # The following 8 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
						     # If any of these fields become needed for testing, they will need to be moved to the expected_info hash
						     'Optimize:\s*',
						     'Compression Quick Check File Size \(in bytes\):\s*',
						     'Change Log Size:\s*',
						     'Gathering Phase 2 Begin:\s*',
						     'Change Log Overflow:\s*',
						     'Operation Frequency:\s*',
						     'Verify Suspend Count:\s*',
						     'Savings Potential:\s*',
						     'Vault Transfer Log Size:\s*',
						     'Compression Change Log Size:\s*',
						     'Change Log Overflow:\s*',
                                                     # burt613302
                                                     'Total Sorted Blocks:\s*',
                                                     'Same Fingerprint:\s*',
                                                     'Same FBN Location:\s*',
                                                     'Same Data:\s*',
													 'No Op:\s*',
                                                     'Same VBN:\s*',
                                                     'Mismatched Data:\s*',
                                                     'Same Sharing Records:\s*',
                                                     'Max Reference Hits:\s*',
                                                     'Staled Recipient:\s*',
                                                     'Staled Donor:\s*',
                                                     'Out of Space:\s*',
                                                     'Mismatch By Overwrites:\s*',
                                                     'Staled Aux Recipient:\s*',
                                                     'Staled Aux Recipient Blocks:\s*',
                                                     'Mismatched Aux Recipient:\s*',
                                                     'Skipped Aux Recipient Sharing:\s*',
                                                     'Skip Share Blocks Delta:\s*',
                                                     'Skip Share Blocks Upper:\s*',
                                                    ],
                                    expected_info => [{indentification_string => 'State:\s+',
                                                       output_hash_key => 'state'},
                                                      {indentification_string => 'Compression:\s+',
                                                       output_hash_key => 'compression'},
                                                      {indentification_string => 'Inline Compression:\s+',
                                                       output_hash_key => 'inline_compression'},
						      {indentification_string => 'Incompressible Data Detection:*\s+',
						       output_hash_key => 'idd'},
						      #{indentification_string => 'Compression Quick Check File Size \(in bytes\):\s*',
						      # output_hash_key => 'quick_check_file_size'},
                                                      {indentification_string => 'Status:\s+',
                                                       output_hash_key => 'status'},
                                                      {indentification_string => 'Progress:\s+',
                                                       output_hash_key => 'progress',
                                                       extract_information => [{regex => '\d+',
                                                                                output_hash_key => 'data_scanned',
                                                                               },],},
                                                      {indentification_string => 'Type:\s+',
                                                       output_hash_key => 'type'},
                                                      {indentification_string => 'Schedule:\s+',
                                                       output_hash_key => 'schedule'},
                                                      {indentification_string => 'Minimum Blocks Shared:\s+',
                                                       output_hash_key => 'minimum_blocks_shared'},
                                                      {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                       output_hash_key => 'blocks_skipped_sharing'},
                                                      {indentification_string => 'Last Operation Begin:\s+',
                                                       output_hash_key => 'last_operation_begin'},
                                                      {indentification_string => 'Last Operation End:\s+',
                                                       output_hash_key => 'last_operation_end'},
                                                      {indentification_string => 'Last Operation Size:\s+',
                                                       output_hash_key => 'last_operation_size'},
                                                      {indentification_string => 'Last Operation Error:\s+',
                                                       output_hash_key => 'last_operation_error'},
						      #{indentification_string => 'Operation Frequency:\s+',
						      # output_hash_key => 'operation_frequency'},
                                                      {indentification_string => 'Change Log Usage:\s+',
                                                       output_hash_key => 'changelog_usage'},
						      #{indentification_string => 'Change Log Overflow:\s+',
						      # output_hash_key => 'changelog_overflow'},
                                                      {indentification_string => 'Logical Data:\s+',
                                                       output_hash_key => 'logical_data'},
                                                      {indentification_string => 'Queued Job:\s+',
                                                       output_hash_key => 'queued_jobs'},
                                                      {indentification_string => 'Stage:\s+',
                                                       output_hash_key => 'stage'},
						      {indentification_string => 'Checkpoint Time:\s+',
						       output_hash_key => 'checkpoint_time'},
						      {indentification_string => 'Checkpoint Operation Type:\s+',
						       output_hash_key => 'checkpoint_op_type'},
						      {indentification_string => 'Checkpoint Stage:\s+',
						       output_hash_key => 'checkpoint_stage'},
						      {indentification_string => 'Checkpoint Sub-*stage:\s+',
						       output_hash_key => 'checkpoint_substage'},
						      {indentification_string => 'Checkpoint Progress:\s+',
						       output_hash_key => 'checkpoint_progress'},
						      {indentification_string => 'Fingerprints Gathered:\s+',
                                                       output_hash_key => 'fingerprints_gathered'},
                                                      {indentification_string => 'Gathering Begin:\s+',
                                                       output_hash_key => 'gathering_begin'},
                                                      {indentification_string => 'Fingerprints Sorted:\s+',
                                                       output_hash_key => 'fingerprints_sorted'},
                                                      {indentification_string => 'Duplicate Blocks Found:\s+',
                                                       output_hash_key => 'duplicated_blocks_found'},
                                                      {indentification_string => 'Sorting Begin:\s+',
                                                       output_hash_key => 'sorting_begin'},
                                                      {indentification_string => 'Blocks De-duplicated:\s+',
                                                       output_hash_key => 'blocks_deduplicated'},
                                                      {indentification_string => 'De-duping Begin:\s+',
                                                       output_hash_key => 'dedupe_begin'},
                                                      {indentification_string => 'Fingerprints Deleted:\s+',
                                                       output_hash_key => 'fingerprints_deleted'},
                                                      {indentification_string => 'Checking Begin:\s+',
                                                       output_hash_key => 'checking_begin'},
                                                      {indentification_string => 'Verify Trigger Rate:\s+',
                                                       output_hash_key => 'verify_trigger_rate'},
                                                      {indentification_string => 'Total Verify Time:\s+',
                                                       output_hash_key => 'total_verify_time'},
						      #{indentification_string => 'Verify Suspend Count:\s+',
						      # output_hash_key => 'verify_suspend_count'},
						      #{indentification_string => 'Savings Potential:\s+',
						      # output_hash_key => 'savings_potential'},
                                                      ],
                                    },
                
                GET_FILE_ID => {expected_info => [{indentification_string => 'flags=',
                                                   extract_information => [{regex => '\S+',
                                                                            output_hash_key => 'flags',
                                                                            },
                                                                           {regex => ' snapid=',
                                                                            },
                                                                           {regex => '\S+',
                                                                            output_hash_key => 'snapid',
                                                                            },
                                                                           {regex => ' fileid=',
                                                                            },
                                                                           {regex => '\S+',
                                                                            output_hash_key => 'fileid',
                                                                            },],},],
                               },
                
                GET_STATS => {header => 'Path:\s+/vol/',
                              ignore_lines => ['Vserver:',
                                               '^\s*<*(\d+|UNUSED)\s*<*(\d+|UNUSED)',
                                               'Block sharing histogram:',
                                               'Reference count histogram:',
                                               'Sort Stats:',
											   'No Op:\s*',
					       'Compression Attempts:',
                                               ],
                              expected_info => [{indentification_string => 'Allocated:\s+',
                                                 output_hash_key => 'allocated'},
                                                {indentification_string => 'Shared:\s+',
                                                 output_hash_key => 'shared'},
                                                {indentification_string => 'Saving:\s+',
                                                 output_hash_key => 'saving'},
                                                {indentification_string => '%Saved:\s+',
                                                 output_hash_key => 'percentage_saved'},
                                                {indentification_string => 'Max Refcount:\s+',
                                                 output_hash_key => 'max_refcount'},
                                                {indentification_string => 'Total Processed:\s+',
                                                 output_hash_key => 'total_processed'},
                                                {indentification_string => 'Total Process Time:\s+',
                                                 output_hash_key => 'total_process_time'},
                                                {indentification_string => 'Total Verify Time:\s+',
                                                 output_hash_key => 'total_verify_time'},
                                                {indentification_string => 'Efficiency Files:\s+',
                                                 output_hash_key => 'sis_files'},
                                                {indentification_string => 'Succeeded Op:\s+',
                                                 output_hash_key => 'succeeded_op'},
                                                {indentification_string => 'Started Op:\s+',
                                                 output_hash_key => 'started_op'},
                                                {indentification_string => 'Failed Op:\s+',
                                                 output_hash_key => 'failed_op'},
                                                {indentification_string => 'Stopped Op:\s+',
                                                 output_hash_key => 'stopped_op'},
                                                {indentification_string => 'Deferred Op:\s+',
                                                 output_hash_key => 'deferred_op'},
                                                {indentification_string => 'Succeeded Check Op:\s+',
                                                 output_hash_key => 'succeeded_check_op'},
                                                {indentification_string => 'Failed Check Op:\s+',
                                                 output_hash_key => 'failed_check_op'},
                                                {indentification_string => 'Suspended Check Op:\s+',
                                                 output_hash_key => 'suspended_check_op'},
                                                {indentification_string => 'Total FP Deleted:\s+',
                                                 output_hash_key => 'total_fp_deleted'},
                                                {indentification_string => 'Total Sorted Blocks:\s+',
                                                 output_hash_key => 'total_sorted_blocks'},
                                                {indentification_string => 'Overlapped Blocks:\s+',
                                                 output_hash_key => 'overlapped_blocks'},
                                                {indentification_string => 'Same Fingerprint:\s+',
                                                 output_hash_key => 'same_fingerprint'},
                                                {indentification_string => 'Same FBN Location:\s+',
                                                 output_hash_key => 'same_fbn_location'},
                                                {indentification_string => 'Same Data:\s+',
                                                 output_hash_key => 'same_data'},
                                                {indentification_string => 'Same VBN:\s+',
                                                 output_hash_key => 'same_vbn'},
                                                {indentification_string => 'Mismatched Data:\s+',
                                                 output_hash_key => 'mismatched_data'},
                                                {indentification_string => 'Max Reference Hits:\s+',
                                                 output_hash_key => 'max_reference_hits'},
                                                {indentification_string => 'Staled Recipient:\s+',
                                                 output_hash_key => 'staled_recipient'},
                                                {indentification_string => 'Staled Donor:\s+',
                                                 output_hash_key => 'staled_donor'},
                                                {indentification_string => 'File Too Small:\s+',
                                                 output_hash_key => 'file_too_small'},
                                                {indentification_string => 'Out of Space:\s+',
                                                 output_hash_key => 'out_of_space'},
                                                {indentification_string => 'FP False Match:\s+',
                                                 output_hash_key => 'fp_false_match'},
                                                {indentification_string => 'Mismatch By Overwrites:\s+',
                                                 output_hash_key => 'mismatch_by_overwrite'},
                                                {indentification_string => 'Delino Records:\s+',
                                                 output_hash_key => 'delino_records'},
                                                {indentification_string => 'Unaligned Compression blocks:\s+',
                                                 output_hash_key => 'unaligned_compression_blocks'},
                                                {indentification_string => 'Additional Sharing Messages:\s+',
                                                 output_hash_key => 'additional_sharing_messages'},
                                                {indentification_string => 'Compression Saved:\s+',
                                                 output_hash_key => 'saving_compression'},
                                                {indentification_string => 'CGs Decompressed:\s+',
                                                 output_hash_key => 'groups_decompressed'},
                                                {indentification_string => 'Partial CG Modifies:\s+',
                                                 output_hash_key => 'partial_cg_modifies'},
                                                {indentification_string => 'Avg Decompress Time:\s+',
                                                 output_hash_key => 'average_decompress_time'},
                                                {indentification_string => 'Extra CP Reads:\s+',
                                                 output_hash_key => 'extra_cp_reads'},
                                                {indentification_string => 'CGs Compressed Inline:\s+',
                                                 output_hash_key => 'cgs_compressed_inline'},
                                                {indentification_string => 'CGs Compressed:\s+',
                                                 output_hash_key => 'cgs_compressed_in_bg'},
                                                {indentification_string => 'Compressed Blocks:\s+',
                                                 output_hash_key => 'compressed_blocks'},
                                                {indentification_string => 'Uncompressed Blocks:\s+',
                                                 output_hash_key => 'uncompressed_blocks'},
                                                {indentification_string => 'New Partial CG Writes:\s+',
                                                 output_hash_key => 'partial_cg_new_writes'},
                                                {indentification_string => 'Decompress Disk Bad:\s+',
                                                 output_hash_key => 'decompress_disk_bad'},
                                                {indentification_string => 'Decompress SW Bad:\s+',
                                                 output_hash_key => 'decompress_sw_bad'},
                                                {indentification_string => 'Avg Compress Time:\s+',
                                                 output_hash_key => 'average_compress_time'},
                                                {indentification_string => 'Compression Attempts:\s+',
                                                 output_hash_key => 'compression_attempts'},
                                                {indentification_string => 'Compression Failures:\s+',
                                                 output_hash_key => 'compression_failures'},
                                                {indentification_string => 'Poor Compression Ratio:\s+',
                                                 output_hash_key => 'poor_compression_ratio'},
                                                {indentification_string => 'Shared Blocks Skipped:\s+',
                                                 output_hash_key => 'shared_skipped_blocks'},
                                                {indentification_string => 'Snapshot Blocks Skipped:\s+',
                                                 output_hash_key => 'snapshot_skipped_blocks'},
                                                {indentification_string => 'Un-Flushed Change Logs:\s+',
                                                 output_hash_key => 'unflushed_changelogs'},],
                              },
                
                VERIFY_COMPR_ATTEMPTED_BIT => {ignore_lines => ['Checking Semi Upgraded temporary metafiles',
                                                                'Checking Downgraded temporary metafiles'],
                                               expected_info => [{indentification_string => 'Checking clogdir\.0\s+\.+\s+'},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_0'},
                                                                 {indentification_string => 'Checking clogdir\.1\s+\.+\s+',},
                                                                 {indentification_string => '\s*Number of changelog buckets:\s+',
                                                                  output_hash_key => 'changelog_buckets_1'},],
                                               },
                
                COMPRESS_STATUS_CHECKPOINT => {header => 'Path:\s+/vol/',
                                               ignore_lines => ['Vserver:\s+',
								'Auto State:',
                                                                'Policy ID:\s*',
                                                                'Efficiency Policy:\s*',
								# TODO: move these to expected_info once burt 410442
								# is merged to main
								'Last Operation State:\s*',
								'Last Successful Operation Begin:\s*',
								'Last Successful Operation End:\s*',
								# The following 8 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
								# If any of these fields become needed for testing, they will need to be moved to the expected_info hash
								'Optimize:\s*',
								'Compression Quick Check File Size \(in bytes\):\s*',
								'Change Log Size:\s*',
								'Gathering Phase 2 Begin:\s*',
								'Change Log Overflow:\s*',
								'Operation Frequency:\s*',
								'Verify Suspend Count:\s*',
								'Savings Potential:\s*',
						       		'Vault Transfer Log Size:\s*',
						       		'Compression Change Log Size:\s*',
						       		'Change Log Overflow:\s*',
                                                                # burt613302
                                                                'Total Sorted Blocks:\s*',
                                                                'Same Fingerprint:\s*',
                                                                'Same FBN Location:\s*',
                                                                'Same Data:\s*',
																'No Op:\s*',
                                                                'Same VBN:\s*',
                                                                'Mismatched Data:\s*',
                                                                'Same Sharing Records:\s*',
                                                                'Max Reference Hits:\s*',
                                                                'Staled Recipient:\s*',
                                                                'Staled Donor:\s*',
                                                                'Out of Space:\s*',
                                                                'Mismatch By Overwrites:\s*',
                                                                'Staled Aux Recipient:\s*',
                                                                'Staled Aux Recipient Blocks:\s*',
                                                                'Mismatched Aux Recipient:\s*',
                                                                'Skipped Aux Recipient Sharing:\s*',
                                                                'Skip Share Blocks Delta:\s*',
                                                                'Skip Share Blocks Upper:\s*',
                                                                ],
                                               expected_info => [{indentification_string => 'State:\s+',
                                                                  output_hash_key => 'state'},
                                                                 {indentification_string => 'Compression:\s+',
                                                                  output_hash_key => 'compression'},
                                                                 {indentification_string => 'Inline Compression:\s+',
                                                                  output_hash_key => 'inline_compression'},
								 {indentification_string => 'Incompressible Data Detection:*\s+',
								  output_hash_key => 'idd'},
								 #{indentification_string => 'Compression Quick Check File Size \(in bytes\):\s*',
								 # output_hash_key => 'quick_check_file_size'},
                                                                 {indentification_string => 'Status:\s+',
                                                                  output_hash_key => 'status'},
                                                                 {indentification_string => 'Progress:\s+',
                                                                  output_hash_key => 'progress',
                                                                  extract_information => [{regex => '\d+',
                                                                                           output_hash_key => 'total_data_scanned',
                                                                                           },],},
                                                                 {indentification_string => 'Type:\s+',
                                                                  output_hash_key => 'type'},
                                                                 {indentification_string => 'Schedule:\s+',
                                                                  output_hash_key => 'schedule'},
                                                                 {indentification_string => 'Minimum Blocks Shared:\s+',
                                                                  output_hash_key => 'minimum_blocks_shared'},
                                                                 {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                                  output_hash_key => 'blocks_skipped_sharing'},
                                                                 {indentification_string => 'Last Operation Begin:\s+',
                                                                  output_hash_key => 'last_operation_begin'},
                                                                 {indentification_string => 'Last Operation End:\s+',
                                                                  output_hash_key => 'last_operation_end'},
                                                                 {indentification_string => 'Last Operation Size:\s+',
                                                                  output_hash_key => 'last_operation_size'},
                                                                 {indentification_string => 'Last Operation Error:\s+',
                                                                  output_hash_key => 'last_operation_error'},
								 #{indentification_string => 'Operation Frequency:\s+',
								 # output_hash_key => 'operation_frequency'},
                                                                 {indentification_string => 'Change Log Usage:\s+',
                                                                  output_hash_key => 'changelog_usage'},
								 #{indentification_string => 'Change Log Overflow:\s+',
								 # output_hash_key => 'changelog_overflow'},
                                                                 {indentification_string => 'Logical Data:\s+',
                                                                  output_hash_key => 'logical_data'},
                                                                 {indentification_string => 'Queued Job:\s+',
                                                                  output_hash_key => 'queued_jobs'},
                                                                 {indentification_string => 'Stale Fingerprints.*:\s+',
                                                                  output_hash_key => 'stale_fingerprints'},
                                                                 {indentification_string => 'Stage:\s+',
                                                                  output_hash_key => 'stage'},
								 {indentification_string => 'Checkpoint Time:\s+',
								  output_hash_key => 'checkpoint_creation_time'},
								 {indentification_string => 'Checkpoint Operation Type:\s+',
								  output_hash_key => 'checkpoint_op_type'},
								 {indentification_string => 'Checkpoint Stage:\s+',
								  output_hash_key => 'checkpoint_stage'},
								 {indentification_string => 'Checkpoint Sub-*stage:\s+',
								  output_hash_key => 'checkpoint_substage'},
								 {indentification_string => 'Checkpoint Progress:\s+',
								  output_hash_key => 'checkpoint_progress'},
                                                                 {indentification_string => 'Fingerprints Gathered:\s+',
                                                                  output_hash_key => 'fingerprints_gathered'},
                                                                 {indentification_string => 'Gathering Begin:\s+',
                                                                  output_hash_key => 'gathering_begin'},
                                                                 {indentification_string => 'Fingerprints Sorted:\s+',
                                                                  output_hash_key => 'fingerprints_sorted'},
                                                                 {indentification_string => 'Duplicate Blocks Found:\s+',
                                                                  output_hash_key => 'duplicated_blocks_found'},
                                                                 {indentification_string => 'Sorting Begin:\s+',
                                                                  output_hash_key => 'sorting_begin'},
                                                                 {indentification_string => 'Blocks De-duplicated:\s+',
                                                                  output_hash_key => 'blocks_deduplicated'},
                                                                 {indentification_string => 'De-duping Begin:\s+',
                                                                  output_hash_key => 'dedupe_begin'},
                                                                 {indentification_string => 'Fingerprints Deleted:\s+',
                                                                  output_hash_key => 'fingerprints_deleted'},
                                                                 {indentification_string => 'Checking Begin:\s+',
                                                                  output_hash_key => 'checking_begin'},
                                                                 {indentification_string => 'Verify Trigger Rate:\s+',
                                                                  output_hash_key => 'verify_trigger_rate'},
                                                                 {indentification_string => 'Total Verify Time:\s+',
                                                                  output_hash_key => 'total_verify_time'},
								 #{indentification_string => 'Verify Suspend Count:\s+',
								 # output_hash_key => 'verify_suspend_count'},
								 #{indentification_string => 'Savings Potential:\s+',
								 # output_hash_key => 'savings_potential'},
                                                                 ],
                                                
                                              },
                
                DECOMPRESS_STATUS => {header => 'Path:\s+/vol/',
                                      ignore_lines => ['Vserver:\s+',
						       'Auto State:',
                                                       'Efficiency Policy:\s*',
                                                       'Policy ID:\s*',
                                                       'Stale Fingerprints.*:\s*',
						       # TODO: move these to expected_info once burt 410442
						       # is merged to main
						       'Last Operation State:\s*',
						       'Last Successful Operation Begin:\s*',
						       'Last Successful Operation End:\s*',
						       # The following 8 entries were added to the ignore list to allow the compression scripts to run on different SN product branches
						       # If any of these fields become needed for testing, they will need to be moved to the expected_info hash
						       'Optimize:\s*',
						       'Compression Quick Check File Size \(in bytes\):\s*',
						       'Change Log Size:\s*',
						       'Gathering Phase 2 Begin:\s*',
						       'Change Log Overflow:\s*',
						       'Operation Frequency:\s*',
						       'Verify Suspend Count:\s*',
						       'Savings Potential:\s*',
						       'Vault Transfer Log Size:\s*',
						       'Compression Change Log Size:\s*',
						       'Change Log Overflow:\s*',
                                                        # burt613302
                                                        'Total Sorted Blocks:\s*',
                                                        'Same Fingerprint:\s*',
                                                        'Same FBN Location:\s*',
                                                        'Same Data:\s*',
														'No Op:\s*',
                                                        'Same VBN:\s*',
                                                        'Mismatched Data:\s*',
                                                        'Same Sharing Records:\s*',
                                                        'Max Reference Hits:\s*',
                                                        'Staled Recipient:\s*',
                                                        'Staled Donor:\s*',
                                                        'Out of Space:\s*',
                                                        'Mismatch By Overwrites:\s*',
                                                        'Staled Aux Recipient:\s*',
                                                        'Staled Aux Recipient Blocks:\s*',
                                                        'Mismatched Aux Recipient:\s*',
                                                        'Skipped Aux Recipient Sharing:\s*',
                                                        'Skip Share Blocks Delta:\s*',
                                                        'Skip Share Blocks Upper:\s*',
                                                       ],
                                      expected_info => [{indentification_string => 'State:\s+',
                                                         output_hash_key => 'state'},
                                                        {indentification_string => 'Compression:\s+',
                                                         output_hash_key => 'compression'},
                                                        {indentification_string => 'Inline Compression:\s+',
                                                         output_hash_key => 'inline_compression'},
							{indentification_string => 'Incompressible Data Detection:*\s+',
							 output_hash_key => 'idd'},
							#{indentification_string => 'Compression Quick Check File Size \(in bytes\):\s*',
							# output_hash_key => 'quick_check_file_size'},
                                                        {indentification_string => 'Status:\s+',
                                                         output_hash_key => 'status'},
                                                        {indentification_string => 'Progress:\s+',
                                                         output_hash_key => 'progress',
                                                         extract_information => [{regex => '\d+',
                                                                                  output_hash_key => 'data_processed',
                                                                                  },
                                                                                 {regex => '\s+KB\s+',
                                                                                 },
                                                                                 {regex => '\w+',
                                                                                  output_hash_key => 'processing_phase',
                                                                                 },],},
                                                        {indentification_string => 'Type:\s+',
                                                         output_hash_key => 'type'},
                                                        {indentification_string => 'Schedule:\s+',
                                                         output_hash_key => 'schedule'},
                                                        {indentification_string => 'Minimum Blocks Shared:\s+',
                                                         output_hash_key => 'minimum_blocks_shared'},
                                                        {indentification_string => 'Blocks Skipped Sharing:\s+',
                                                         output_hash_key => 'blocks_skipped_sharing'},
                                                        {indentification_string => 'Last Operation Begin:\s+',
                                                         output_hash_key => 'last_operation_begin'},
                                                        {indentification_string => 'Last Operation End:\s+',
                                                         output_hash_key => 'last_operation_end'},
                                                        {indentification_string => 'Last Operation Size:\s+',
                                                         output_hash_key => 'last_operation_size'},
                                                        {indentification_string => 'Last Operation Error:\s+',
                                                         output_hash_key => 'last_operation_error'},
							#{indentification_string => 'Operation Frequency:\s+',
							# output_hash_key => 'operation_frequency'},
                                                        {indentification_string => 'Change Log Usage:\s+',
                                                         output_hash_key => 'changelog_usage'},
							#{indentification_string => 'Change Log Overflow:\s+',
							# output_hash_key => 'changelog_overflow'},
                                                        {indentification_string => 'Logical Data:\s+',
                                                         output_hash_key => 'logical_data'},
                                                        {indentification_string => 'Queued Job:\s+',
                                                         output_hash_key => 'queued_jobs'},
                                                        {indentification_string => 'Stage:\s+',
                                                         output_hash_key => 'stage'},
							{indentification_string => 'Checkpoint Time:\s+',
							 output_hash_key => 'checkpoint_time'},
							{indentification_string => 'Checkpoint Operation Type:\s+',
							 output_hash_key => 'checkpoint_op_type'},
							{indentification_string => 'Checkpoint Stage:\s+',
							 output_hash_key => 'checkpoint_stage'},
							{indentification_string => 'Checkpoint Sub-*stage:\s+',
							 output_hash_key => 'checkpoint_substage'},
							{indentification_string => 'Checkpoint Progress:\s+',
							 output_hash_key => 'checkpoint_progress'},
                                                        {indentification_string => 'Fingerprints Gathered:\s+',
                                                         output_hash_key => 'fingerprints_gathered'},
                                                        {indentification_string => 'Gathering Begin:\s+',
                                                         output_hash_key => 'gathering_begin'},
                                                        {indentification_string => 'Fingerprints Sorted:\s+',
                                                         output_hash_key => 'fingerprints_sorted'},
                                                        {indentification_string => 'Duplicate Blocks Found:\s+',
                                                         output_hash_key => 'duplicated_blocks_found'},
                                                        {indentification_string => 'Sorting Begin:\s+',
                                                         output_hash_key => 'sorting_begin'},
                                                        {indentification_string => 'Blocks De-duplicated:\s+',
                                                         output_hash_key => 'blocks_deduplicated'},
                                                        {indentification_string => 'De-duping Begin:\s+',
                                                         output_hash_key => 'dedupe_begin'},
                                                        {indentification_string => 'Fingerprints Deleted:\s+',
                                                         output_hash_key => 'fingerprints_deleted'},
                                                        {indentification_string => 'Checking Begin:\s+',
                                                         output_hash_key => 'checking_begin'},
                                                        {indentification_string => 'Verify Trigger Rate:\s+',
                                                         output_hash_key => 'verify_trigger_rate'},
                                                        {indentification_string => 'Total Verify Time:\s+',
                                                         output_hash_key => 'total_verify_time'},
							#{indentification_string => 'Verify Suspend Count:\s+',
							# output_hash_key => 'verify_suspend_count'},
							#{indentification_string => 'Savings Potential:\s+',
							# output_hash_key => 'savings_potential'},
                                                      ],
                                               },
                GET_SNAP_RECLAIMABLE => {header => 'Processing .*',
                                         expected_info => [{indentification_string => 'snap reclaimable: Approximately\s+',
                                                            extract_information => [{regex => '\d+',
                                                                                     output_hash_key => 'reclaimable_kb',
                                                                                     },],},
                                                           ],
                                        },
		DUMPBLOCK_INFO_L1 => {header => '^ok\s*$',
                                      ignore_lines => ['ckdbn\s+',
                                                       'cksum_dbn\s+',
                                                       ],
				      expected_info => [{indentification_string => 'disk\s+',
							 output_hash_key => 'disk'},
							{indentification_string => 'dbn\s+',
							 output_hash_key => 'dbn'},
							{indentification_string => 'pvbn\s+',
							 output_hash_key => 'pvbn'},
							{indentification_string => 'pvbn-from-parent\s+',
							 output_hash_key => 'pvbn-from-parent'},
							{indentification_string => 'pvbn-from-container\s+',
							 output_hash_key => 'pvbn-from-container'},
							{indentification_string => 'vvbn',
							 output_hash_key => 'vvbn'},
							{indentification_string => 'fbn-of-inofile\s+',
							 output_hash_key => 'fbn-of-inofile'},
							{indentification_string => 'pvbn-of-inofile\s+',
							 output_hash_key => 'pvbn-of-inofile'},
							{indentification_string => 'pvbn-of-inofile-from-parent\s+',
							 output_hash_key => 'pvbn-of-inofile-from-parent'},
							{indentification_string => 'pvbn-of-inofile-from-container\s+',
							 output_hash_key => 'pvbn-of-inofile-from-container'},
							{indentification_string => 'vvbn-of-inofile\s+',
							 output_hash_key => 'vvbn-of-inofile'},
							{indentification_string => 'fileid-of-container-l1\s+',
							 output_hash_key => 'fileid-of-container-l1'},
							{indentification_string => 'fileid-of_container-l1-is-private\s+',
							 output_hash_key => 'fileid-of_container-l1-is-private'},
							{indentification_string => 'fbn-of-container-l1\s+',
							 output_hash_key => 'fbn-of-container-l1'},
							{indentification_string => 'ckse-dbn\s+',
							 output_hash_key => 'ckse-dbn'},
							{indentification_string => 'ckse-checksum\s+',
							 output_hash_key => 'ckse-checksum'},
							{indentification_string => 'ckse-buftreeid\s+',
							 output_hash_key => 'ckse-buftreeid'},
							{indentification_string => 'ckse-fbn\s+',
							 output_hash_key => 'ckse-fbn'},
							{indentification_string => 'ckse-cp-count\s+',
							 output_hash_key => 'ckse-cp-count'},
							{indentification_string => 'ckse-notflexvol-flag\s+',
							 output_hash_key => 'ckse-notflexvol-flag'},
							],
				   },
		DUMPBLOCK_INFO => {header => '^ok\s*$',
				   expected_info => [{indentification_string => 'disk\s+',
						      output_hash_key => 'disk'},
						     {indentification_string => 'dbn\s+',
						      output_hash_key => 'dbn'},
						     {indentification_string => 'pvbn\s+',
						      output_hash_key => 'pvbn'},
						     {indentification_string => 'pvbn-from-parent\s+',
						      output_hash_key => 'pvbn-from-parent'},
						     {indentification_string => 'pvbn-from-container\s+',
						      output_hash_key => 'pvbn-from-container'},
						     {indentification_string => 'vvbn\s+',
						      output_hash_key => 'vvbn'},
						     {indentification_string => 'pvbn-of-parent\s+',
						      output_hash_key => 'pvbn-of-parent'},
						     {indentification_string => 'pvbn-of-parent-from-container\s+',
						      output_hash_key => 'pvbn-of-parent-from-container'},
						     {indentification_string => 'vvbn-of-parent\s+',
						      output_hash_key => 'vvbn-of-parent'},
						     ],
				   },
                GET_CONTAINMENT_VERSION => {ignore_lines => ['volume:\S+:', # Ignore all others besides expected_info
                                                             ],
                                            expected_info => [{indentification_string => 'volume:\S+:wv_fsinfo_containment_version_compression:',
                                                               output_hash_key => 'containment_version'},
                                                              {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_compression:',
                                                               output_hash_key => 'highest_containment_version'},
                                                              ],
                                            },
                GET_COMPRESSION_STATS => {expected_info => [{indentification_string => 'compression:\S+:num_blocks_saved:',
                                                             output_hash_key => 'num_blocks_saved'},
                                                            {indentification_string => 'compression:\S+:num_blocks_cdata:',
                                                             output_hash_key => 'num_blocks_cdata'},
                                                            {indentification_string => 'compression:\S+:num_cg_decompressed:',
                                                             output_hash_key => 'num_cg_decompressed'},
                                                            {indentification_string => 'compression:\S+:num_partial_write:',
                                                             output_hash_key => 'num_partial_write'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf:',
                                                             output_hash_key => 'num_valid_buf'},
                                                            {indentification_string => 'compression:\S+:num_valid_buf_suspended:',
                                                             output_hash_key => 'num_valid_buf_suspended'},
                                                            {indentification_string => 'compression:\S+:num_blocking_io:',
                                                             output_hash_key => 'num_blocking_io'},
                                                            {indentification_string => 'compression:\S+:num_decompress_suspends:',
                                                             output_hash_key => 'num_decompress_suspends'},
                                                            {indentification_string => 'compression:\S+:avg_decompress_cg_time:',
                                                             output_hash_key => 'avg_decompress_cg_time'},
							    {indentification_string => 'compression:\S+:num_qcheck_fail:',
							     output_hash_key => 'num_qcheck_fail'},
							    {indentification_string => 'compression:\S+:num_icg_inline:',
							     output_hash_key => 'num_icg_inline'},
                                                            {indentification_string => 'compression:\S+:num_compress_attempts:',
                                                             output_hash_key => 'num_compress_attempts'},
                                                            {indentification_string => 'compression:\S+:num_compress_fail:',
                                                             output_hash_key => 'num_compress_fail'},
                                                            {indentification_string => 'compression:\S+:num_not_enough_savings:',
                                                             output_hash_key => 'num_not_enough_savings'},
                                                            {indentification_string => 'compression:\S+:num_no_mem:',
                                                             output_hash_key => 'num_no_mem'},
                                                            {indentification_string => 'compression:\S+:num_min_input:',
                                                             output_hash_key => 'num_min_input'},
                                                            {indentification_string => 'compression:\S+:num_cgs_with_holes:',
                                                             output_hash_key => 'num_cgs_with_holes'},
                                                            {indentification_string => 'compression:\S+:avg_compress_cg_time:',
                                                             output_hash_key => 'avg_compress_cg_time'},
							    {indentification_string => 'compression:\S+:avg_idd_qcheck_time:',
							     output_hash_key => 'avg_idd_qcheck_time'},
							    {indentification_string => 'compression:\S+:avg_cd_qcheck_time:',
							     output_hash_key => 'avg_cd_qcheck_time'},
                                                            {indentification_string => 'compression:\S+:num_extra_cp_reads:',
                                                             output_hash_key => 'num_extra_cp_reads'},
                                                            {indentification_string => 'compression:\S+:num_write_uncompressed:',
                                                             output_hash_key => 'num_write_uncompressed'},
                                                            {indentification_string => 'compression:\S+:num_eos_bufs:',
                                                             output_hash_key => 'num_eos_bufs'},
                                                            {indentification_string => 'compression:\S+:num_decompress_read_bad_blks:',
                                                             output_hash_key => 'num_decompress_read_bad_blks'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_hw_bad:',
                                                             output_hash_key => 'num_decompress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_decompress_mark_sw_bad:',
                                                             output_hash_key => 'num_decompress_mark_sw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_hw_bad:',
                                                             output_hash_key => 'num_compress_mark_hw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compress_mark_sw_bad:',
                                                             output_hash_key => 'num_compress_mark_sw_bad'},
                                                            {indentification_string => 'compression:\S+:num_compressed_inline:',
                                                             output_hash_key => 'num_compressed_inline'},
                                                            {indentification_string => 'compression:\S+:num_compressed_bg:',
                                                             output_hash_key => 'num_compressed_bg'},
                                                            {indentification_string => 'compression:\S+:num_compressed_bg_cg:',
                                                             output_hash_key => 'num_compressed_bg_cg'},
                                                            {indentification_string => 'compression:\S+:num_partial_cg_modifies:',
                                                             output_hash_key => 'num_partial_cg_modifies'},
                                                            {indentification_string => 'compression:\S+:num_partial_cg_new:',
                                                             output_hash_key => 'num_partial_cg_new'},
                                                            {indentification_string => 'compression:\S+:num_vbn_zero_cg_skipped:',
                                                             output_hash_key => 'num_vbn_zero_cg_skipped'},
							    {indentification_string => 'compression:\S+:bce_received_in_exempt:',
							     output_hash_key => 'bce_received_in_exempt'},
							    {indentification_string => 'compression:\S+:bce_err_abort:',
							     output_hash_key => 'bce_err_abort'},
							    {indentification_string => 'compression:\S+:bce_err_stale_inode:',
							     output_hash_key => 'bce_err_stale_inode'},
							    {indentification_string => 'compression:\S+:bce_fbn_invalid:',
							     output_hash_key => 'bce_fbn_invalid'},
							    {indentification_string => 'compression:\S+:bce_policy_stage:',
							     output_hash_key => 'bce_policy_stage'},
							    {indentification_string => 'compression:\S+:bce_compress_stage:',
							     output_hash_key => 'bce_compress_stage'},
                                                            # burt613302
							    {indentification_string => 'compression:\S+:bce_compress_stage_overwrites:',
							     output_hash_key => 'bce_compress_stage_overwrites'},
							    {indentification_string => 'compression:\S+:bce_sending_to_exempt:',
							     output_hash_key => 'bce_sending_to_exempt'},
							    {indentification_string => 'compression:\S+:bce_post_processing_stage:',
							     output_hash_key => 'bce_post_processing_stage'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_abort:',
							     output_hash_key => 'bce_err_post_processing_abort'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_stale_inode:',
							     output_hash_key => 'bce_err_post_processing_stale_inode'},
                                                            # burt613302
							    {indentification_string => 'compression:\S+:bce_post_processing_overwrites:',
							     output_hash_key => 'bce_post_processing_overwrites'},
							    {indentification_string => 'compression:\S+:bce_err_post_processing_nospace:',
							     output_hash_key => 'bce_err_post_processing_nospace'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s1\sblock:',
                                                             output_hash_key => 'detailed_avg_compress_cg_time.to_1_block'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s2\sblocks:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.to_2_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s3\sblocks:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.to_3_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s4\sblocks:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.to_4_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s5\sblocks:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.to_5_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.to\s6\sblocks:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.to_6_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_compress_cg_time.not\scompressed:',
                                                              output_hash_key => 'detailed_avg_compress_cg_time.not_compressed'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s1\sblock:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_1_block'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s2\sblocks:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_2_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s3\sblocks:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_3_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s4\sblocks:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_4_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s5\sblocks:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_5_blocks'},
                                                            {indentification_string => 'compression:\S+:detailed_avg_decompress_cg_time.from\s6\sblocks:',
                                                              output_hash_key => 'detailed_avg_decompress_cg_time.from_6_blocks'},
                                                            ],
                                          ignore_lines => ['compression:\S+:instance_name:',
                                                           'compression:\S+:node_name:',
                                                           'compression:\S+:node_uuid:',
							   'compression:\S+:process_name:'],
                                         },
                GET_WAFL_SCAN_SPEED => {expected_info => [{indentification_string => 'WAFL scan speed is\s+',
                                                           output_hash_key => 'current_value'},
                                                          ],
                                        },
                GET_SNAPLOCK_CONTAINMENT_BITS => {ignore_lines => ['volume:\S+:', # Ignore all others besides expected_info
                                                                   ],
                                                  expected_info => [{indentification_string => 'volume:\S+:wv_fsinfo_containment_version_slc:',
                                                                     output_hash_key => 'containment_version_slc'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_sle:',
                                                                     output_hash_key => 'containment_version_sle'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_slc:',
                                                                     output_hash_key => 'containment_version_highest_slc'},
                                                                    {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_sle:',
                                                                     output_hash_key => 'containment_version_highest_sle'},
                                                                    ],
                                                 },
		GET_VOLUME_STATS => {expected_info => [{indentification_string => 'volume:\S+:instance_uuid:',
							output_hash_key => 'instance_uuid'},
						       {indentification_string => 'volume:\S+:parent_aggr:',
							output_hash_key => 'parent_aggr'},
						       {indentification_string => 'volume:\S+:avg_latency:',
							output_hash_key => 'avg_latency'},
						       {indentification_string => 'volume:\S+:total_ops:',
							output_hash_key => 'total_ops'},
						       {indentification_string => 'volume:\S+:read_data:',
							output_hash_key => 'read_data'},
						       {indentification_string => 'volume:\S+:read_latency:',
							output_hash_key => 'read_latency'},
						       {indentification_string => 'volume:\S+:read_ops:',
							output_hash_key => 'read_ops'},
						       {indentification_string => 'volume:\S+:write_data:',
							output_hash_key => 'write_data'},
						       {indentification_string => 'volume:\S+:write_latency:',
							output_hash_key => 'write_latency'},
						       {indentification_string => 'volume:\S+:write_ops:',
							output_hash_key => 'write_ops'},
						       {indentification_string => 'volume:\S+:other_latency:',
							output_hash_key => 'other_latency'},
						       {indentification_string => 'volume:\S+:other_ops:',
							output_hash_key => 'other_ops'},
						       {indentification_string => 'volume:\S+:internal_msgs:',
							output_hash_key => 'internal_msgs'},
						       {indentification_string => 'volume:\S+:read_blocks:',
							output_hash_key => 'read_blocks'},
						       {indentification_string => 'volume:\S+:write_blocks:',
							output_hash_key => 'write_blocks'},
						       {indentification_string => 'volume:\S+:synchronous_frees:',
							output_hash_key => 'synchronous_frees'},
						       {indentification_string => 'volume:\S+:asynchronous_frees:',
							output_hash_key => 'asynchronous_frees'},
						       {indentification_string => 'volume:\S+:df_worker:',
							output_hash_key => 'df_worker'},
						       {indentification_string => 'volume:\S+:nfs_read_data:',
							output_hash_key => 'nfs_read_data'},
						       {indentification_string => 'volume:\S+:nfs_read_latency:',
							output_hash_key => 'nfs_read_latency'},
						       {indentification_string => 'volume:\S+:nfs_read_ops:',
							output_hash_key => 'nfs_read_ops'},
						       {indentification_string => 'volume:\S+:nfs_write_data:',
							output_hash_key => 'nfs_write_data'},
						       {indentification_string => 'volume:\S+:nfs_write_latency:',
							output_hash_key => 'nfs_write_latency'},
						       {indentification_string => 'volume:\S+:nfs_write_ops:',
							output_hash_key => 'nfs_write_ops'},
						       {indentification_string => 'volume:\S+:nfs_other_latency:',
							output_hash_key => 'nfs_other_latency'},
						       {indentification_string => 'volume:\S+:nfs_other_ops:',
							output_hash_key => 'nfs_other_ops'},
						       {indentification_string => 'volume:\S+:cifs_read_data:',
							output_hash_key => 'cifs_read_data'},
						       {indentification_string => 'volume:\S+:cifs_read_latency:',
							output_hash_key => 'cifs_read_latency'},
						       {indentification_string => 'volume:\S+:cifs_read_ops:',
							output_hash_key => 'cifs_read_ops'},
						       {indentification_string => 'volume:\S+:cifs_write_data:',
							output_hash_key => 'cifs_write_data'},
						       {indentification_string => 'volume:\S+:cifs_write_latency:',
							output_hash_key => 'cifs_write_latency'},
						       {indentification_string => 'volume:\S+:cifs_write_ops:',
							output_hash_key => 'cifs_write_ops'},
						       {indentification_string => 'volume:\S+:cifs_other_latency:',
							output_hash_key => 'cifs_other_latency'},
						       {indentification_string => 'volume:\S+:cifs_other_ops:',
							output_hash_key => 'cifs_other_ops'},
						       {indentification_string => 'volume:\S+:san_read_data:',
							output_hash_key => 'san_read_data'},
						       {indentification_string => 'volume:\S+:san_read_latency:',
							output_hash_key => 'san_read_latency'},
						       {indentification_string => 'volume:\S+:san_read_ops:',
							output_hash_key => 'san_read_ops'},
						       {indentification_string => 'volume:\S+:san_write_data:',
							output_hash_key => 'san_write_data'},
						       {indentification_string => 'volume:\S+:san_write_latency:',
							output_hash_key => 'san_write_latency'},
						       {indentification_string => 'volume:\S+:san_write_ops:',
							output_hash_key => 'san_write_ops'},
						       {indentification_string => 'volume:\S+:san_other_latency:',
							output_hash_key => 'san_other_latency'},
						       {indentification_string => 'volume:\S+:san_other_ops:',
							output_hash_key => 'san_other_ops'},
						       {indentification_string => 'volume:\S+:fcp_read_data:',
							output_hash_key => 'fcp_read_data'},
						       {indentification_string => 'volume:\S+:fcp_read_latency:',
							output_hash_key => 'fcp_read_latency'},
						       {indentification_string => 'volume:\S+:fcp_read_ops:',
							output_hash_key => 'fcp_read_ops'},
						       {indentification_string => 'volume:\S+:fcp_write_data:',
							output_hash_key => 'fcp_write_data'},
						       {indentification_string => 'volume:\S+:fcp_write_latency:',
							output_hash_key => 'fcp_write_latency'},
						       {indentification_string => 'volume:\S+:fcp_write_ops:',
							output_hash_key => 'fcp_write_ops'},
						       {indentification_string => 'volume:\S+:fcp_other_latency:',
							output_hash_key => 'fcp_other_latency'},
						       {indentification_string => 'volume:\S+:fcp_other_ops:',
							output_hash_key => 'fcp_other_ops'},
						       {indentification_string => 'volume:\S+:iscsi_read_data:',
							output_hash_key => 'iscsi_read_data'},
						       {indentification_string => 'volume:\S+:iscsi_read_latency:',
							output_hash_key => 'iscsi_read_latency'},
						       {indentification_string => 'volume:\S+:iscsi_read_ops:',
							output_hash_key => 'iscsi_read_ops'},
						       {indentification_string => 'volume:\S+:iscsi_write_data:',
							output_hash_key => 'iscsi_write_data'},
						       {indentification_string => 'volume:\S+:iscsi_write_latency:',
							output_hash_key => 'iscsi_write_latency'},
						       {indentification_string => 'volume:\S+:iscsi_write_ops:',
							output_hash_key => 'iscsi_write_ops'},
						       {indentification_string => 'volume:\S+:iscsi_other_latency:',
							output_hash_key => 'iscsi_other_latency'},
						       {indentification_string => 'volume:\S+:iscsi_other_ops:',
							output_hash_key => 'iscsi_other_ops'},
						       {indentification_string => 'volume:\S+:flexcache_read_data:',
							output_hash_key => 'flexcache_read_data'},
						       {indentification_string => 'volume:\S+:flexcache_read_ops:',
							output_hash_key => 'flexcache_read_ops'},
						       {indentification_string => 'volume:\S+:flexcache_write_data:',
							output_hash_key => 'flexcache_write_data'},
						       {indentification_string => 'volume:\S+:flexcache_write_ops:',
							output_hash_key => 'flexcache_write_ops'},
						       {indentification_string => 'volume:\S+:flexcache_other_ops:',
							output_hash_key => 'flexcache_other_ops'},
						       {indentification_string => 'volume:\S+:flexcache_send_data:',
							output_hash_key => 'flexcache_send_data'},
						       {indentification_string => 'volume:\S+:flexcache_receive_data:',
							output_hash_key => 'flexcache_receive_data'},
						       {indentification_string => 'volume:\S+:wv_fsid:',
							output_hash_key => 'wv_fsid'},
						       {indentification_string => 'volume:\S+:wv_vol_type:',
							output_hash_key => 'wv_vol_type'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_fs_version:',
							output_hash_key => 'wv_fsinfo_fs_version'},
						       {indentification_string => 'volume:\S+:wv_volinfo_fs_options:',
							output_hash_key => 'wv_volinfo_fs_options'},
						       {indentification_string => 'volume:\S+:wv_volinfo_fs_flags:',
							output_hash_key => 'wv_volinfo_fs_flags'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_total:',
							output_hash_key => 'wv_fsinfo_blks_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_reserve:',
							output_hash_key => 'wv_fsinfo_blks_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_used:',
							output_hash_key => 'wv_fsinfo_blks_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_used_by_plane0:',
							output_hash_key => 'wv_fsinfo_blks_used_by_plane0'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_holes_cifs:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_holes:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_blks_rsrv_overwrite:',
							output_hash_key => 'wv_fsinfo_blks_blks_rsrv_overwrite'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_rsrv_absents:',
							output_hash_key => 'wv_fsinfo_blks_rsrv_absents'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_snap_reserve_pct:',
							output_hash_key => 'wv_fsinfo_blks_snap_reserve_pct'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_res_state:',
							output_hash_key => 'wv_fsinfo_blks_res_state'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_overwrite_slider_pct:',
							output_hash_key => 'wv_fsinfo_blks_overwrite_slider_pct'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks64_blks_rsrv_holes_cifs:',
							output_hash_key => 'wv_fsinfo_blks64_blks_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks64_blks_rsrv_holes:',
							output_hash_key => 'wv_fsinfo_blks64_blks_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_total:',
							output_hash_key => 'wv_fsinfo_public_inos_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_reserve:',
							output_hash_key => 'wv_fsinfo_public_inos_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_public_inos_used:',
							output_hash_key => 'wv_fsinfo_public_inos_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_total:',
							output_hash_key => 'wv_fsinfo_private_inos_total'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_reserve:',
							output_hash_key => 'wv_fsinfo_private_inos_reserve'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_private_inos_used:',
							output_hash_key => 'wv_fsinfo_private_inos_used'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blkr_cp:',
							output_hash_key => 'wv_fsinfo_blkr_cp'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_vvol_dbys_df_cnt:',
							output_hash_key => 'wv_fsinfo_blks_vvol_dbys_df_cnt'},
						       {indentification_string => 'volume:\S+:wvblk_past_eof:',
							output_hash_key => 'wvblk_past_eof'},
						       {indentification_string => 'volume:\S+:wvblk_past_eof64:',
							output_hash_key => 'wvblk_past_eof64'},
						       {indentification_string => 'volume:\S+:wvblk_snap_reserve:',
							output_hash_key => 'wvblk_snap_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_zombie_blks:',
							output_hash_key => 'wvblk_zombie_blks'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_blks_rsrv_parent:',
							output_hash_key => 'wv_fsinfo_blks_rsrv_parent'},
						       {indentification_string => 'volume:\S+:wvblk_spcres_in_parent:',
							output_hash_key => 'wvblk_spcres_in_parent'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_holes:',
							output_hash_key => 'wvblk_rsrv_parent_holes'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_overwrite:',
							output_hash_key => 'wvblk_rsrv_parent_overwrite'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_parent_overwrite_always:',
							output_hash_key => 'wvblk_rsrv_parent_overwrite_always'},
						       {indentification_string => 'volume:\S+:wvbd_active_frees:',
							output_hash_key => 'wvbd_active_frees'},
						       {indentification_string => 'volume:\S+:wvbd_whole_frees:',
							output_hash_key => 'wvbd_whole_frees'},
						       {indentification_string => 'volume:\S+:wvbd_active_frees_y:',
							output_hash_key => 'wvbd_active_frees_y'},
						       {indentification_string => 'volume:\S+:wvbd_owner_changed_y:',
							output_hash_key => 'wvbd_owner_changed_y'},
						       {indentification_string => 'volume:\S+:wvbd_whole_frees_o:',
							output_hash_key => 'wvbd_whole_frees_o'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_total:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_total'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_reserve:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_public_inos_used:',
							output_hash_key => 'wvblk_saved_fsinfo_public_inos_used'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_total:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_total'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_reserve:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_reserve'},
						       {indentification_string => 'volume:\S+:wvblk_saved_fsinfo_private_inos_used:',
							output_hash_key => 'wvblk_saved_fsinfo_private_inos_used'},
						       {indentification_string => 'volume:\S+:wvblk_delalloc:',
							output_hash_key => 'wvblk_delalloc'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_delalloc:',
							output_hash_key => 'wvblk_rsrv_delalloc'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes_cifs:',
							output_hash_key => 'wvblk_rsrv_holes_cifs'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes:',
							output_hash_key => 'wvblk_rsrv_holes'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_overwrite:',
							output_hash_key => 'wvblk_rsrv_overwrite'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_absents:',
							output_hash_key => 'wvblk_rsrv_absents'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes_cifs64:',
							output_hash_key => 'wvblk_rsrv_holes_cifs64'},
						       {indentification_string => 'volume:\S+:wvblk_rsrv_holes64:',
							output_hash_key => 'wvblk_rsrv_holes64'},
						       {indentification_string => 'volume:\S+:extent_size:',
							output_hash_key => 'extent_size'},
						       {indentification_string => 'volume:\S+:wvdf_enabled:',
							output_hash_key => 'wvdf_enabled'},
						       {indentification_string => 'volume:\S+:wvdf_inconsistent_scores:',
							output_hash_key => 'wvdf_inconsistent_scores'},
						       {indentification_string => 'volume:\S+:wvdf_last_fbn:',
							output_hash_key => 'wvdf_last_fbn'},
						       {indentification_string => 'volume:\S+:wvdf_water_mark:',
							output_hash_key => 'wvdf_water_mark'},
						       {indentification_string => 'volume:\S+:wvdf_max_frees:',
							output_hash_key => 'wvdf_max_frees'},
						       {indentification_string => 'volume:\S+:wvdf_total_score:',
							output_hash_key => 'wvdf_total_score'},
						       {indentification_string => 'volume:\S+:wvblk_parent_to_be_reclaimed:',
							output_hash_key => 'wvblk_parent_to_be_reclaimed'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_wi_blk_cnt:',
							output_hash_key => 'wvip_vvol_container_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_wi_size:',
							output_hash_key => 'wvip_vvol_container_wi_size'},
						       {indentification_string => 'volume:\S+:wvip_public_inofile_wi_blk_cnt:',
							output_hash_key => 'wvip_public_inofile_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_private_inofile_wi_blk_cnt:',
							output_hash_key => 'wvip_private_inofile_wi_blk_cnt'},
						       {indentification_string => 'volume:\S+:wvip_vvol_container_indirects:',
							output_hash_key => 'wvip_vvol_container_indirects'},
						       {indentification_string => 'volume:\S+:wvol_number_suspended:',
							output_hash_key => 'wvol_number_suspended'},
						       {indentification_string => 'volume:\S+:clone_storage_blocks:',
							output_hash_key => 'clone_storage_blocks'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_start:',
							output_hash_key => 'wvblk_reclaim_time_start'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_reset:',
							output_hash_key => 'wvblk_reclaim_time_reset'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_done:',
							output_hash_key => 'wvblk_reclaim_time_done'},
						       {indentification_string => 'volume:\S+:wvblk_reclaim_time_abort:',
							output_hash_key => 'wvblk_reclaim_time_abort'},
						       {indentification_string => 'volume:\S+:wvzmb_num_zmsgs_inuse:',
							output_hash_key => 'wvzmb_num_zmsgs_inuse'},
						       {indentification_string => 'volume:\S+:wvblk_ind_delalloc:',
							output_hash_key => 'wvblk_ind_delalloc'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_last_fbn:',
							output_hash_key => 'wvsblk_vvrd_last_fbn'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_spcflags:',
							output_hash_key => 'wvsblk_vvrd_spcflags'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_spc_clone_inherited:',
							output_hash_key => 'wvsblk_vvrd_spc_clone_inherited'},
						       {indentification_string => 'volume:\S+:wvsblk_space_tax:',
							output_hash_key => 'wvsblk_space_tax'},
						       {indentification_string => 'volume:\S+:wvsblk_lev0_over_nominal:',
							output_hash_key => 'wvsblk_lev0_over_nominal'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_vol_size:',
							output_hash_key => 'wvsblk_vvrd_vol_size'},
						       {indentification_string => 'volume:\S+:wvsblk_vvrd_flags:',
							output_hash_key => 'wvsblk_vvrd_flags'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_slc:',
							output_hash_key => 'wv_fsinfo_containment_version_slc'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_sle:',
							output_hash_key => 'wv_fsinfo_containment_version_sle'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_compression:',
							output_hash_key => 'wv_fsinfo_containment_version_compression'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_vmalign:',
							output_hash_key => 'wv_fsinfo_containment_version_vmalign'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_spare1:',
							output_hash_key => 'wv_fsinfo_containment_version_spare1'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_spare2:',
							output_hash_key => 'wv_fsinfo_containment_version_spare2'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_slc:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_slc'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_sle:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_sle'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_compression:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_compression'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_vmalign:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_vmalign'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_spare1:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_spare1'},
						       {indentification_string => 'volume:\S+:wv_fsinfo_containment_version_highest_spare2:',
							output_hash_key => 'wv_fsinfo_containment_version_highest_spare2'},
						       {indentification_string => 'volume:\S+:wvi2p_wip_wi_size:',
							output_hash_key => 'wvi2p_wip_wi_size'},
						       {indentification_string => 'volume:\S+:bad_user_blk_read_error_propagated:',
							output_hash_key => 'bad_user_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_user_blk_read_error_not_propagated:',
							output_hash_key => 'bad_user_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_non_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_non_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_non_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_non_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_metafile_blk_read_error_propagated:',
							output_hash_key => 'bad_fixable_metafile_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_fixable_metafile_blk_read_error_not_propagated:',
							output_hash_key => 'bad_fixable_metafile_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_zombie_ind_blk_read_err_propagated:',
							output_hash_key => 'bad_zombie_ind_blk_read_err_propagated'},
						       {indentification_string => 'volume:\S+:bad_zombie_ind_blk_read_err_not_propagated:',
							output_hash_key => 'bad_zombie_ind_blk_read_err_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_user_blk_read_error_propagated:',
							output_hash_key => 'bad_container_user_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_user_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_user_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_afs_blk_read_error_propagated:',
							output_hash_key => 'bad_container_fixable_afs_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_afs_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_fixable_afs_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_snap_blk_read_error_propagated:',
							output_hash_key => 'bad_container_fixable_snap_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_fixable_snap_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_fixable_snap_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_non_fixable_blk_read_error_propagated:',
							output_hash_key => 'bad_container_non_fixable_blk_read_error_propagated'},
						       {indentification_string => 'volume:\S+:bad_container_non_fixable_blk_read_error_not_propagated:',
							output_hash_key => 'bad_container_non_fixable_blk_read_error_not_propagated'},
						       {indentification_string => 'volume:\S+:wv_playlist_hits:',
							output_hash_key => 'wv_playlist_hits'},
						       {indentification_string => 'volume:\S+:wv_playlist_misses:',
							output_hash_key => 'wv_playlist_misses'},
						       {indentification_string => 'volume:\S+:wv_playlist_not_present:',
							output_hash_key => 'wv_playlist_not_present'},
						       {indentification_string => 'volume:\S+:wv_playlist_entries:',
							output_hash_key => 'wv_playlist_entries'},
						       {indentification_string => 'volume:\S+:wv_playlist_reqs:',
							output_hash_key => 'wv_playlist_reqs'},
						       {indentification_string => 'volume:\S+:wv_playlist_no_raidbufs:',
							output_hash_key => 'wv_playlist_no_raidbufs'},
						       {indentification_string => 'volume:\S+:wv_playlist_getbuf_failures:',
							output_hash_key => 'wv_playlist_getbuf_failures'},
						       {indentification_string => 'volume:\S+:wv_playlist_cont_indirects:',
							output_hash_key => 'wv_playlist_cont_indirects'},
						       {indentification_string => 'volume:\S+:wv_playlist_vvbn_holes:',
							output_hash_key => 'wv_playlist_vvbn_holes'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_not_started:',
							output_hash_key => 'wv_playlist_prefetch_not_started'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_start_time:',
							output_hash_key => 'wv_playlist_prefetch_start_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_prefetch_end_time:',
							output_hash_key => 'wv_playlist_prefetch_end_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_load_end_time:',
							output_hash_key => 'wv_playlist_load_end_time'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_used_slots:',
							output_hash_key => 'wv_playlist_apfi_used_slots'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_collision_inserts:',
							output_hash_key => 'wv_playlist_apfi_collision_inserts'},
						       {indentification_string => 'volume:\S+:wv_playlist_apfi_collision_accesses:',
							output_hash_key => 'wv_playlist_apfi_collision_accesses'},
						       {indentification_string => 'volume:\S+:wvsnap_ondisk_count:',
							output_hash_key => 'wvsnap_ondisk_count'},
						       {indentification_string => 'volume:\S+:wvsnap_incore_count:',
							output_hash_key => 'wvsnap_incore_count'},
						       {indentification_string => 'volume:\S+:wvsnap_loaded_total:',
							output_hash_key => 'wvsnap_loaded_total'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_in_use:',
							output_hash_key => 'delete_log_dlog_in_use'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_ind_blk_cnt:',
							output_hash_key => 'delete_log_dlog_ind_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_blk_cnt:',
							output_hash_key => 'delete_log_dlog_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.purge_ind_blk_cnt:',
							output_hash_key => 'delete_log_purge_ind_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.purge_blk_cnt:',
							output_hash_key => 'delete_log_purge_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_hole_cnt:',
							output_hash_key => 'delete_log_dlog_hole_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_cap:',
							output_hash_key => 'delete_log_dlog_cap'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_in_use:',
							output_hash_key => 'delete_log_overflow_in_use'},
						       {indentification_string => 'volume:\S+:delete_log.overflow0_blk_cnt:',
							output_hash_key => 'delete_log_overflow0_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.overflow1_blk_cnt:',
							output_hash_key => 'delete_log_overflow1_blk_cnt'},
						       {indentification_string => 'volume:\S+:delete_log.del_log:',
							output_hash_key => 'delete_log_del_log'},
						       {indentification_string => 'volume:\S+:delete_log.purge_send:',
							output_hash_key => 'delete_log_purge_send'},
						       {indentification_string => 'volume:\S+:delete_log.purge_block:',
							output_hash_key => 'delete_log_purge_block'},
						       {indentification_string => 'volume:\S+:delete_log.purge_vbns:',
							output_hash_key => 'delete_log_purge_vbns'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_frees:',
							output_hash_key => 'delete_log_dlog_frees'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_vbn_clears:',
							output_hash_key => 'delete_log_dlog_vbn_clears'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_keys:',
							output_hash_key => 'delete_log_dlog_keys'},
						       {indentification_string => 'volume:\S+:delete_log.pre_split_failed:',
							output_hash_key => 'delete_log_pre_split_failed'},
						       {indentification_string => 'volume:\S+:delete_log.pre_bufs_loaded:',
							output_hash_key => 'delete_log_pre_bufs_loaded'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow:',
							output_hash_key => 'delete_log_purge_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow_susp:',
							output_hash_key => 'delete_log_purge_overflow_susp'},
						       {indentification_string => 'volume:\S+:delete_log.purge_overflow_no_split:',
							output_hash_key => 'delete_log_purge_overflow_no_split'},
						       {indentification_string => 'volume:\S+:delete_log.trunc_overflow_of_append:',
							output_hash_key => 'delete_log_trunc_overflow_of_append'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_overuse_mem:',
							output_hash_key => 'delete_log_dlog_overuse_mem'},
						       {indentification_string => 'volume:\S+:delete_log.dlog_overflow:',
							output_hash_key => 'delete_log_dlog_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_total:',
							output_hash_key => 'delete_log_overflow_total'},
						       {indentification_string => 'volume:\S+:delete_log.act_dirty:',
							output_hash_key => 'delete_log_act_dirty'},
						       {indentification_string => 'volume:\S+:delete_log.pre_overflow:',
							output_hash_key => 'delete_log_pre_overflow'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_vbns:',
							output_hash_key => 'delete_log_overflow_vbns'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_frees:',
							output_hash_key => 'delete_log_overflow_frees'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_frees_ind:',
							output_hash_key => 'delete_log_overflow_frees_ind'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_commits:',
							output_hash_key => 'delete_log_overflow_commits'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_send:',
							output_hash_key => 'delete_log_overflow_send'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_susp:',
							output_hash_key => 'delete_log_overflow_susp'},
						       {indentification_string => 'volume:\S+:delete_log.overflow_throttle:',
							output_hash_key => 'delete_log_overflow_throttle'},
						       {indentification_string => 'volume:\S+:msgs_rejected_in_nvfailed_state:',
							output_hash_key => 'msgs_rejected_in_nvfailed_state'},
						       {indentification_string => 'volume:\S+:msgs_allowed_in_nvfailed_state:',
							output_hash_key => 'msgs_allowed_in_nvfailed_state'},
						       {indentification_string => 'volume:\S+:clone_inodes:',
							output_hash_key => 'clone_inodes'},
						       {indentification_string => 'volume:\S+:clone_lookups:',
							output_hash_key => 'clone_lookups'},
						       {indentification_string => 'volume:\S+:clone_lookup_hits:',
							output_hash_key => 'clone_lookup_hits'},
						       {indentification_string => 'volume:\S+:clone_blks_copied:',
							output_hash_key => 'clone_blks_copied'},
						       {indentification_string => 'volume:\S+:clone_eio_blks:',
							output_hash_key => 'clone_eio_blks'},
						       {indentification_string => 'volume:\S+:clone_num_entries:',
							output_hash_key => 'clone_num_entries'},
						       {indentification_string => 'volume:\S+:clone_max_entries:',
							output_hash_key => 'clone_max_entries'},
						       {indentification_string => 'volume:\S+:clone_max_hierarchy:',
							output_hash_key => 'clone_max_hierarchy'},
						       {indentification_string => 'volume:\S+:clone_unsplit_snap_entries:',
							output_hash_key => 'clone_unsplit_snap_entries'},
						       {indentification_string => 'volume:\S+:clone_num_share_stopped:',
							output_hash_key => 'clone_num_share_stopped'},
						       {indentification_string => 'volume:\S+:clone_read_redirected:',
							output_hash_key => 'clone_read_redirected'},
						       {indentification_string => 'volume:\S+:clone_split_ra:',
							output_hash_key => 'clone_split_ra'},
						       {indentification_string => 'volume:\S+:clone_split_ra_lag:',
							output_hash_key => 'clone_split_ra_lag'},
						       {indentification_string => 'volume:\S+:clone_inline_split_kireeti_in_progress:',
							output_hash_key => 'clone_inline_split_kireeti_in_progress'},
						       {indentification_string => 'volume:\S+:clone_inline_split_stage_size_limitation:',
							output_hash_key => 'clone_inline_split_stage_size_limitation'},
						       {indentification_string => 'volume:\S+:clone_inline_split_beyond_eof:',
							output_hash_key => 'clone_inline_split_beyond_eof'},
						       {indentification_string => 'volume:\S+:clone_inline_split_source_destination_dirty:',
							output_hash_key => 'clone_inline_split_source_destination_dirty'},
						       {indentification_string => 'volume:\S+:clone_inline_split_source_spec_vbn:',
							output_hash_key => 'clone_inline_split_source_spec_vbn'},
						       {indentification_string => 'volume:\S+:clone_inline_split_enospc:',
							output_hash_key => 'clone_inline_split_enospc'},
						       {indentification_string => 'volume:\S+:clone_inline_split_edquot:',
							output_hash_key => 'clone_inline_split_edquot'},
						       {indentification_string => 'volume:\S+:nonzero_dbys_cnt:',
							output_hash_key => 'nonzero_dbys_cnt'},
						       ],
				     ignore_lines => ['volume:\S+:instance_name:',
						      'volume:\S+:node_name:',
						      'volume:\S+:vserver_name:',
						      'volume:\S+:vserver_uuid:',
						      'volume:\S+:clone_sizes_hist',
						      ],
				     },
		GET_WAFL_STATS => {ignore_lines => ['.+', # Ignore all others besides expected_info (huge output)
						    ],
				   expected_info => [{indentification_string => 'CGs resolved in KC\s*\t*=\s*\t*',
						      output_hash_key => 'cgs_resolved_in_KC'},
						     {indentification_string => 'CGs resolved in Bento\s*\t*=\s*\t*',
						      output_hash_key => 'cgs_resolved_in_Bento'},
						     {indentification_string => 'Truncate Suspends\s*\t*=\s*\t*',
						      output_hash_key => 'truncate_suspends'},
						     {indentification_string => 'PreCP overwritten\s*\t*=\s*\t*',
						      output_hash_key => 'precp_overwritten'},
						     {indentification_string => 'Flippity Suspends\s*\t*=\s*\t*',
						      output_hash_key => 'flippity_suspends'},
						     {indentification_string => 'Flippity Skipped\s*\t*=\s*\t*',
						      output_hash_key => 'flippity_skipped'},
						     {indentification_string => 'Decompress Nomem\s*\t*=\s*\t*',
						      output_hash_key => 'decompress_nomem'},
						     {indentification_string => 'Buffers cacheable in EC\s*\t*=\s*\t*',
						      output_hash_key => 'buffers_cacheable_in_EC'},
						     {indentification_string => 'Non-Stripe writes\s*\t*=\s*\t*',
						      output_hash_key => 'non_stripe_writes'},
						     {indentification_string => 'Partial CG throttled\s*\t*=\s*\t*',
						      output_hash_key => 'partial_cg_throttled'},
						     ],
				  },
        };

	my $code = $args{code};
	my $format = $base_formats_hash->{$code};
        
        if (!defined($format))
        {
                $format = { format_expected => FALSE };
                logresult('script', "Library problem found: Processing required for format type 2, but no format info defined. Command code: $code");
        }

	return $format;
}

sub _Cmode_formats_8_1
{
	my $class = shift;
	my %args = validate(@_,
	{
                code  => { type => SCALAR, },
	} );
        
	my $formats_hash = {
		GET_COMPRESSION_STATS => {expected_info => [{indentification_string => 'compression:\S+:num_blocks_saved:',
							     output_hash_key => 'num_blocks_saved'},
							    {indentification_string => 'compression:\S+:num_blocks_cdata:',
							     output_hash_key => 'num_blocks_cdata'},
							    {indentification_string => 'compression:\S+:num_blocks_uncompressed:',
							     output_hash_key => 'num_blocks_uncompressed'},
							    {indentification_string => 'compression:\S+:num_buf_decompressed:',
							     output_hash_key => 'num_buf_decompressed'},
							    {indentification_string => 'compression:\S+:num_cg_decompressed:',
							     output_hash_key => 'num_cg_decompressed'},
							    {indentification_string => 'compression:\S+:num_partial_write:',
							     output_hash_key => 'num_partial_write'},
							    {indentification_string => 'compression:\S+:num_valid_buf:',
							     output_hash_key => 'num_valid_buf'},
							    {indentification_string => 'compression:\S+:num_invalid_raid_read:',
							     output_hash_key => 'num_invalid_raid_read'},
							    {indentification_string => 'compression:\S+:num_valid_buf_suspended:',
							     output_hash_key => 'num_valid_buf_suspended'},
							    {indentification_string => 'compression:\S+:num_cow_cg:',
							     output_hash_key => 'num_cow_cg'},
							    {indentification_string => 'compression:\S+:num_blocking_io:',
							     output_hash_key => 'num_blocking_io'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_ahead_busy:',
							     output_hash_key => 'num_vol_decomp_ahead_busy'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_no_resource:',
							     output_hash_key => 'num_vol_decomp_no_resource'},
							    {indentification_string => 'compression:\S+:num_vol_decomp_zero_holes:',
							     output_hash_key => 'num_vol_decomp_zero_holes'},
							    {indentification_string => 'compression:\S+:num_decompress_suspends:',
							     output_hash_key => 'num_decompress_suspends'},
							    {indentification_string => 'compression:\S+:avg_vol_decompress_time:',
							     output_hash_key => 'avg_vol_decompress_time'},
							    {indentification_string => 'compression:\S+:avg_decompress_cg_time:',
							     output_hash_key => 'avg_decompress_cg_time'},
							    {indentification_string => 'compression:\S+:num_compress_attempts:',
							     output_hash_key => 'num_compress_attempts'},
							    {indentification_string => 'compression:\S+:num_compress_fail:',
							     output_hash_key => 'num_compress_fail'},
							    {indentification_string => 'compression:\S+:num_not_enough_savings:',
							     output_hash_key => 'num_not_enough_savings'},
							    {indentification_string => 'compression:\S+:num_no_mem:',
							     output_hash_key => 'num_no_mem'},
							    {indentification_string => 'compression:\S+:num_compress_suspends:',
							     output_hash_key => 'num_compress_suspends'},
							    {indentification_string => 'compression:\S+:num_cgs_frz_attempts:',
							     output_hash_key => 'num_cgs_frz_attempts'},
							    {indentification_string => 'compression:\S+:num_cgs_frz_failed:',
							     output_hash_key => 'num_cgs_frz_failed'},
							    {indentification_string => 'compression:\S+:num_buf_attempts:',
							     output_hash_key => 'num_buf_attempts'},
							    {indentification_string => 'compression:\S+:num_nocompress_files:',
							     output_hash_key => 'num_nocompress_files'},
							    {indentification_string => 'compression:\S+:num_min_input:',
							     output_hash_key => 'num_min_input'},
							    {indentification_string => 'compression:\S+:num_cgs_with_holes:',
							     output_hash_key => 'num_cgs_with_holes'},
							    {indentification_string => 'compression:\S+:num_next_cp:',
							     output_hash_key => 'num_next_cp'},
							    {indentification_string => 'compression:\S+:num_wait_for_compressor:',
							     output_hash_key => 'num_wait_for_compressor'},
							    {indentification_string => 'compression:\S+:avg_compress_cg_time:',
							     output_hash_key => 'avg_compress_cg_time'},
							    {indentification_string => 'compression:\S+:num_extra_cp_reads:',
							     output_hash_key => 'num_extra_cp_reads'},
							    {indentification_string => 'compression:\S+:num_compress_ahead:',
							     output_hash_key => 'num_compress_ahead'},
							    {indentification_string => 'compression:\S+:num_overload:',
							     output_hash_key => 'num_overload'},
							    {indentification_string => 'compression:\S+:hole_to_data:',
							     output_hash_key => 'hole_to_data'},
							    {indentification_string => 'compression:\S+:hole_to_compressed:',
							     output_hash_key => 'hole_to_compressed'},
							    {indentification_string => 'compression:\S+:data_to_hole:',
							     output_hash_key => 'data_to_hole'},
							    {indentification_string => 'compression:\S+:num_write_uncompressed:',
							     output_hash_key => 'num_write_uncompressed'},
							    {indentification_string => 'compression:\S+:num_eos_bufs:',
							     output_hash_key => 'num_eos_bufs'},
							    {indentification_string => 'compression:\S+:num_cgs_not_fully_dirty:',
							     output_hash_key => 'num_cgs_not_fully_dirty'},
							    {indentification_string => 'compression:\S+:num_wafl_decompress:',
							     output_hash_key => 'num_wafl_decompress'},
							    {indentification_string => 'compression:\S+:num_decompress_read_bad_blks:',
							     output_hash_key => 'num_decompress_read_bad_blks'},
							    {indentification_string => 'compression:\S+:num_decompress_mark_hw_bad:',
							     output_hash_key => 'num_decompress_mark_hw_bad'},
							    {indentification_string => 'compression:\S+:num_decompress_mark_sw_bad:',
							     output_hash_key => 'num_decompress_mark_sw_bad'},
							    {indentification_string => 'compression:\S+:num_compress_mark_hw_bad:',
							     output_hash_key => 'num_compress_mark_hw_bad'},
							    {indentification_string => 'compression:\S+:num_compress_mark_sw_bad:',
							     output_hash_key => 'num_compress_mark_sw_bad'},
							    {indentification_string => 'compression:\S+:num_compressed_inline:',
							     output_hash_key => 'num_compressed_inline'},
							    {indentification_string => 'compression:\S+:num_compressed_bg:',
							     output_hash_key => 'num_compressed_bg'},
							    {indentification_string => 'compression:\S+:num_partial_cg_modifies:',
							     output_hash_key => 'num_partial_cg_modifies'},
							    {indentification_string => 'compression:\S+:num_partial_cg_new:',
							     output_hash_key => 'num_partial_cg_new'},
							    ],
					  ignore_lines => ['compression:\S+:instance_name:',
							   'compression:\S+:instance_uuid:',
							   'compression:\S+:node_name:'],
					  },
		COMPRESS_STATUS => {header => 'Path:\s+/vol/',
				    ignore_lines => ['Vserver:\s+',
						     'Policy ID:\s*',
						     'Efficiency Policy:\s*',
						     'Stale Fingerprints.*:\s*',
						     'Last Operation State:\s*',
						     'Last Successful Operation Begin:\s*',
						     'Last Successful Operation End:\s*',
                                                    ],
				    expected_info => [{indentification_string => 'State:\s+',
						       output_hash_key => 'state'},
						      {indentification_string => 'Compression:\s+',
						       output_hash_key => 'compression'},
						      {indentification_string => 'Inline Compression:\s+',
						       output_hash_key => 'inline_compression'},
						      {indentification_string => 'Status:\s+',
						       output_hash_key => 'status'},
						      {indentification_string => 'Progress:\s+',
						       output_hash_key => 'progress',
                                                       extract_information => [{regex => '\d+',
										output_hash_key => 'data_scanned',
										},],},
						      {indentification_string => 'Type:\s+',
						       output_hash_key => 'type'},
						      {indentification_string => 'Schedule:\s+',
						       output_hash_key => 'schedule'},
						      {indentification_string => 'Minimum Blocks Shared:\s+',
						       output_hash_key => 'minimum_blocks_shared'},
						      {indentification_string => 'Blocks Skipped Sharing:\s+',
						       output_hash_key => 'blocks_skipped_sharing'},
						      {indentification_string => 'Last Operation Begin:\s+',
						       output_hash_key => 'last_operation_begin'},
						      {indentification_string => 'Last Operation End:\s+',
						       output_hash_key => 'last_operation_end'},
						      {indentification_string => 'Last Operation Size:\s+',
						       output_hash_key => 'last_operation_size'},
						      {indentification_string => 'Last Operation Error:\s+',
						       output_hash_key => 'last_operation_error'},
						      {indentification_string => 'Change Log Usage:\s+',
						       output_hash_key => 'changelog_usage'},
						      {indentification_string => 'Logical Data:\s+',
						       output_hash_key => 'logical_data'},
						      {indentification_string => 'Queued Job:\s+',
						       output_hash_key => 'queued_jobs'},
						      {indentification_string => 'Stage:\s+',
						       output_hash_key => 'stage'},
						      {indentification_string => 'Checkpoint Time:\s+',
						       output_hash_key => 'checkpoint_time'},
						      {indentification_string => 'Checkpoint Operation Type:\s+',
						       output_hash_key => 'checkpoint_op_type'},
						      {indentification_string => 'Checkpoint Stage:\s+',
						       output_hash_key => 'checkpoint_stage'},
						      {indentification_string => 'Checkpoint Sub-stage:\s+',
						       output_hash_key => 'checkpoint_substage'},
						      {indentification_string => 'Checkpoint Progress:\s+',
						       output_hash_key => 'checkpoint_progress'},
						      {indentification_string => 'Fingerprints Gathered:\s+',
						       output_hash_key => 'fingerprints_gathered'},
						      {indentification_string => 'Gathering Begin:\s+',
						       output_hash_key => 'gathering_begin'},
						      {indentification_string => 'Fingerprints Sorted:\s+',
						       output_hash_key => 'fingerprints_sorted'},
						      {indentification_string => 'Duplicate Blocks Found:\s+',
						       output_hash_key => 'duplicated_blocks_found'},
						      {indentification_string => 'Sorting Begin:\s+',
						       output_hash_key => 'sorting_begin'},
						      {indentification_string => 'Blocks De-duplicated:\s+',
						       output_hash_key => 'blocks_deduplicated'},
						      {indentification_string => 'De-duping Begin:\s+',
						       output_hash_key => 'dedupe_begin'},
						      {indentification_string => 'Fingerprints Deleted:\s+',
						       output_hash_key => 'fingerprints_deleted'},
						      {indentification_string => 'Checking Begin:\s+',
						       output_hash_key => 'checking_begin'},
						      {indentification_string => 'Verify Trigger Rate:\s+',
						       output_hash_key => 'verify_trigger_rate'},
						      {indentification_string => 'Total Verify Time:\s+',
						       output_hash_key => 'total_verify_time'},
						      {indentification_string => 'Verify Suspend Count:\s+',
						       output_hash_key => 'verify_suspend_count'},
						      ],
				   },
		COMPRESS_STATUS_CHECKPOINT => {header => 'Path:\s+/vol/',
					       ignore_lines => ['Vserver:\s+',
								'Policy ID:\s*',
								'Efficiency Policy:\s*',
								'Last Operation State:\s*',
								'Last Successful Operation Begin:\s*',
								'Last Successful Operation End:\s*',
								],
					       expected_info => [{indentification_string => 'State:\s+',
								  output_hash_key => 'state'},
								 {indentification_string => 'Compression:\s+',
								  output_hash_key => 'compression'},
								 {indentification_string => 'Inline Compression:\s+',
								  output_hash_key => 'inline_compression'},
								 {indentification_string => 'Status:\s+',
								  output_hash_key => 'status'},
								 {indentification_string => 'Progress:\s+',
								  output_hash_key => 'progress',
								  extract_information => [{regex => '\d+',
											   output_hash_key => 'total_data_scanned',
											   },],},
								 {indentification_string => 'Type:\s+',
								  output_hash_key => 'type'},
								 {indentification_string => 'Schedule:\s+',
								  output_hash_key => 'schedule'},
								 {indentification_string => 'Minimum Blocks Shared:\s+',
								  output_hash_key => 'minimum_blocks_shared'},
								 {indentification_string => 'Blocks Skipped Sharing:\s+',
								  output_hash_key => 'blocks_skipped_sharing'},
								 {indentification_string => 'Last Operation Begin:\s+',
								  output_hash_key => 'last_operation_begin'},
								 {indentification_string => 'Last Operation End:\s+',
								  output_hash_key => 'last_operation_end'},
								 {indentification_string => 'Last Operation Size:\s+',
								  output_hash_key => 'last_operation_size'},
								 {indentification_string => 'Last Operation Error:\s+',
								  output_hash_key => 'last_operation_error'},
								 {indentification_string => 'Change Log Usage:\s+',
								  output_hash_key => 'changelog_usage'},
								 {indentification_string => 'Logical Data:\s+',
								  output_hash_key => 'logical_data'},
								 {indentification_string => 'Queued Job:\s+',
								  output_hash_key => 'queued_jobs'},
								 {indentification_string => 'Stale Fingerprints.*:\s+',
								  output_hash_key => 'stale_fingerprints'},
								 {indentification_string => 'Stage:\s+',
								  output_hash_key => 'stage'},
								 {indentification_string => 'Checkpoint Time:\s+',
								  output_hash_key => 'checkpoint_creation_time'},
								 {indentification_string => 'Checkpoint Operation Type:\s+',
								  output_hash_key => 'checkpoint_op_type'},
								 {indentification_string => 'Checkpoint Stage:\s+',
								  output_hash_key => 'checkpoint_stage'},
								 {indentification_string => 'Checkpoint Sub-stage:\s+',
								  output_hash_key => 'checkpoint_substage'},
								 {indentification_string => 'Checkpoint Progress:\s+',
								  output_hash_key => 'checkpoint_progress'},
								 {indentification_string => 'Fingerprints Gathered:\s+',
								  output_hash_key => 'fingerprints_gathered'},
								 {indentification_string => 'Gathering Begin:\s+',
								  output_hash_key => 'gathering_begin'},
								 {indentification_string => 'Fingerprints Sorted:\s+',
								  output_hash_key => 'fingerprints_sorted'},
								 {indentification_string => 'Duplicate Blocks Found:\s+',
								  output_hash_key => 'duplicated_blocks_found'},
								 {indentification_string => 'Sorting Begin:\s+',
								  output_hash_key => 'sorting_begin'},
								 {indentification_string => 'Blocks De-duplicated:\s+',
								  output_hash_key => 'blocks_deduplicated'},
								 {indentification_string => 'De-duping Begin:\s+',
								  output_hash_key => 'dedupe_begin'},
								 {indentification_string => 'Fingerprints Deleted:\s+',
								  output_hash_key => 'fingerprints_deleted'},
								 {indentification_string => 'Checking Begin:\s+',
								  output_hash_key => 'checking_begin'},
								 {indentification_string => 'Verify Trigger Rate:\s+',
								  output_hash_key => 'verify_trigger_rate'},
								 {indentification_string => 'Total Verify Time:\s+',
								  output_hash_key => 'total_verify_time'},
								 {indentification_string => 'Verify Suspend Count:\s+',
								  output_hash_key => 'verify_suspend_count'},
								 ],
					     },
		DECOMPRESS_STATUS => {header => 'Path:\s+/vol/',
				      ignore_lines => ['Vserver:\s+',
						       'Efficiency Policy:\s*',
						       'Policy ID:\s*',
						       'Stale Fingerprints.*:\s*',
						       'Last Operation State:\s*',
						       'Last Successful Operation Begin:\s*',
						       'Last Successful Operation End:\s*',
						       ],
				      expected_info => [{indentification_string => 'State:\s+',
							 output_hash_key => 'state'},
							{indentification_string => 'Compression:\s+',
							 output_hash_key => 'compression'},
							{indentification_string => 'Inline Compression:\s+',
							 output_hash_key => 'inline_compression'},
							{indentification_string => 'Status:\s+',
							 output_hash_key => 'status'},
							{indentification_string => 'Progress:\s+',
							 output_hash_key => 'progress',
							 extract_information => [{regex => '\d+',
										  output_hash_key => 'data_processed',},
										 {regex => '\s+KB\s+',},
										 {regex => '\w+',
										  output_hash_key => 'processing_phase',},],},
							{indentification_string => 'Type:\s+',
							 output_hash_key => 'type'},
							{indentification_string => 'Schedule:\s+',
							 output_hash_key => 'schedule'},
							{indentification_string => 'Minimum Blocks Shared:\s+',
							 output_hash_key => 'minimum_blocks_shared'},
							{indentification_string => 'Blocks Skipped Sharing:\s+',
							 output_hash_key => 'blocks_skipped_sharing'},
							{indentification_string => 'Last Operation Begin:\s+',
							 output_hash_key => 'last_operation_begin'},
							{indentification_string => 'Last Operation End:\s+',
							 output_hash_key => 'last_operation_end'},
							{indentification_string => 'Last Operation Size:\s+',
							 output_hash_key => 'last_operation_size'},
							{indentification_string => 'Last Operation Error:\s+',
							 output_hash_key => 'last_operation_error'},
							{indentification_string => 'Change Log Usage:\s+',
							 output_hash_key => 'changelog_usage'},
							{indentification_string => 'Logical Data:\s+',
							 output_hash_key => 'logical_data'},
							{indentification_string => 'Queued Job:\s+',
							 output_hash_key => 'queued_jobs'},
							{indentification_string => 'Stage:\s+',
							 output_hash_key => 'stage'},
							{indentification_string => 'Checkpoint Time:\s+',
							 output_hash_key => 'checkpoint_time'},
							{indentification_string => 'Checkpoint Operation Type:\s+',
							 output_hash_key => 'checkpoint_op_type'},
							{indentification_string => 'Checkpoint Stage:\s+',
							 output_hash_key => 'checkpoint_stage'},
							{indentification_string => 'Checkpoint Sub-stage:\s+',
							 output_hash_key => 'checkpoint_substage'},
							{indentification_string => 'Checkpoint Progress:\s+',
							 output_hash_key => 'checkpoint_progress'},
							{indentification_string => 'Fingerprints Gathered:\s+',
							 output_hash_key => 'fingerprints_gathered'},
							{indentification_string => 'Gathering Begin:\s+',
							 output_hash_key => 'gathering_begin'},
							{indentification_string => 'Fingerprints Sorted:\s+',
							 output_hash_key => 'fingerprints_sorted'},
							{indentification_string => 'Duplicate Blocks Found:\s+',
							 output_hash_key => 'duplicated_blocks_found'},
							{indentification_string => 'Sorting Begin:\s+',
							 output_hash_key => 'sorting_begin'},
							{indentification_string => 'Blocks De-duplicated:\s+',
							 output_hash_key => 'blocks_deduplicated'},
							{indentification_string => 'De-duping Begin:\s+',
							 output_hash_key => 'dedupe_begin'},
							{indentification_string => 'Fingerprints Deleted:\s+',
							 output_hash_key => 'fingerprints_deleted'},
							{indentification_string => 'Checking Begin:\s+',
							 output_hash_key => 'checking_begin'},
							{indentification_string => 'Verify Trigger Rate:\s+',
							 output_hash_key => 'verify_trigger_rate'},
							{indentification_string => 'Total Verify Time:\s+',
							 output_hash_key => 'total_verify_time'},
							{indentification_string => 'Verify Suspend Count:\s+',
							 output_hash_key => 'verify_suspend_count'},
							],
				     },
        };
	my $code = $args{code};
        
        return $formats_hash->{$code}
}




1;
