# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary An Event Module for CMode Storage Failover Events
## @author hrivera@netapp.com
## @status private
## @pod here

package EventMan::EventModules::ControllerEvents::CMode::SFOEvents;

use strict;
use warnings;
use Tie::IxHash;
use Data::Dumper;
use Tharn;
use San::SanUtil;
use ParamsMgr;
use NACL::C::Node;
use NACL::STask::Node;
use NACL::C::StorageFailover;
use NACL::CS::StorageFailover;
use NACL::C::SystemNodeAutosupport;
use NACL::APISet::Exceptions::TimeoutException;
use NACL::STask::Exceptions::UnexpectedSFOState;
use NATE::BaseException qw(:try);
use NACL::Transit;
use NATE::STAF::Peer qw ($Parent);
use NATE::Log qw(log_global);

# Eventman Exceptions
use EventMan::Exceptions::EncounteredSFOAbortState;
use EventMan::Exceptions::InvalidParameter;
use EventMan::Exceptions::TimedOutWaitingForSFOState;
use EventMan::Exceptions::UnexpectedSFOState;

use vars qw($Log);

$Log = log_global();

###############################################################################
#                               Internal Methods
#                     (WARNING: NEVER INVOKE THESE METHODS)
###############################################################################

###############################################################################
# Internal Method:    __register_event_module()
# Objective: Register this Event Module with the EventMan library
###############################################################################
sub __register_event_module {
    my ($self) = @_;

    my @event_modules = (__PACKAGE__,
                         "EventMan::EventModules::ControllerEvents::CMode",
                         "EventMan::EventModules::ControllerEvents",
                         "EventMan::EventModules");

    $self->EventUtil::register_event(-event_name => "sfo_takeoverCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -exec => sub { &sfo_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_forcedtakeoverCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -exec => sub { &sfo_forcedtakeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_givebackCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_givebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_forcedgivebackCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_forcedgivebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_takegiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover / giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_forcedtakegiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover / giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_forcedtakegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_takeforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover / forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_forcedtakeforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover / forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_forcedtakeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_panicgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic / giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { coredump_validation => 0 },
                                     -exec => sub { &sfo_panicgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_panicforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic / forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { coredump_validation => 0 },
                                     -exec => sub { &sfo_panicforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_panicCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic takeover",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_hardcoded => { coredump_validation => 0 },
                                     -exec => sub { &sfo_panicCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_haltgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt / giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { clam_validation_on_takeover => 0,
                                                        sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_haltgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_haltforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt / forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { clam_validation_on_takeover => 0,
                                                        sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_haltforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_haltCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt takeover",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { clam_validation_on_takeover => 0 },
                                     -exec => sub { &sfo_haltCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_rebootgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot / giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_allow_auto_giveback                 => 1,
                                                        clam_validation_on_takeover             => 0,
                                                        sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_rebootgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_rebootforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot / forced giveback",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_allow_auto_giveback                 => 1,
                                                        clam_validation_on_takeover             => 0,
                                                        sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_rebootforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_rebootCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot takeover",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { sfo_allow_auto_giveback => 1,
                                                        clam_validation_on_takeover => 0 },
                                     -exec => sub { &sfo_rebootCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_reset_takegiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover / giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_power_reset_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_off_on_takegiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover / power on / giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_power_off_on_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_reset_takeforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover / forced giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_power_reset_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_off_on_takeforcedgiveCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover / power on / forced giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_power_off_on_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_off_takeoverCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -exec => sub { &sfo_power_off_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_reset_takeoverCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -exec => sub { &sfo_power_reset_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_on_givebackCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power on / giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -exec => sub { &sfo_power_on_givebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "sfo_power_on_forcedgivebackCm_DEPRECATED",
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power on / forced giveback event",
                                     -required_resource_types => ["sanfilersCm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { sanfilersCm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -exec => sub { &sfo_power_on_forcedgivebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # NACL SFO Events
    $self->EventUtil::register_event(-event_name => "nacl_sfo_takeoverCm",
                                     -aliases => ["sfo_takeoverCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_forcedtakeoverCm",
                                     -aliases => ["sfo_forcedtakeoverCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_forcedtakeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_givebackCm",
                                     -aliases => ["sfo_givebackCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_givebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_forcedgivebackCm",
                                     -aliases => ["sfo_forcedgivebackCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_forcedgivebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_panicCm",
                                     -aliases => ["sfo_panicCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic takeover using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { clam_validation => 0, # must disable clam validation until NACL APIs are supported
                                                         coredump_validation => 0 },
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_panicCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_rebootCm",
                                     -aliases => ["sfo_rebootCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot takeover using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { sfo_allow_auto_giveback => 1,
                                                         clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_rebootCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_haltCm",
                                     -aliases => ["sfo_haltCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt takeover using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_defaults => { clam_validation_on_takeover => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_haltCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_takegiveCm",
                                     -aliases => ["sfo_takegiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover / giveback",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_takeforcedgiveCm",
                                     -aliases => ["sfo_takeforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO takeover " .
                                                     "/ forced giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_forcedtakegiveCm",
                                     -aliases => ["sfo_forcedtakegiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover " .
                                                     "/ giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_forcedtakegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_forcedtakeforcedgiveCm",
                                     -aliases => ["sfo_forcedtakeforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO forced takeover " .
                                                     " / forced giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_forcedtakeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_panicgiveCm",
                                     -aliases => ["sfo_panicgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic / giveback",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0, # must disable clam validation until NACL APIs are supported
                                                         coredump_validation => 0 },
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_panicgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_panicforcedgiveCm",
                                     -aliases => ["sfo_panicforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO panic / forced giveback",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0, # must disable clam validation until NACL APIs are supported
                                                         coredump_validation => 0 },
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_panicforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_rebootgiveCm",
                                     -aliases => ["sfo_rebootgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot / giveback",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_allow_auto_giveback                 => 1,
                                                        clam_validation_on_takeover             => 0,
                                                        sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_rebootgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_rebootforcedgiveCm",
                                     -aliases => ["sfo_rebootforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO reboot / forced giveback",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_allow_auto_giveback                 => 1,
                                                        clam_validation_on_takeover             => 0,
                                                        sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_rebootforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_haltgiveCm",
                                     -aliases => ["sfo_haltgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt / giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { clam_validation_on_takeover => 0,
                                                        sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_haltgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_haltforcedgiveCm",
                                     -aliases => ["sfo_haltforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO halt / forced giveback using NACL APIs",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { clam_validation_on_takeover => 0,
                                                        sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_haltforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_reset_takegiveCm",
                                     -aliases => ["sfo_power_reset_takegiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover / giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_reset_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_off_on_takegiveCm",
                                     -aliases => ["sfo_power_off_on_takegiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover / power on / giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_off_on_takegiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_reset_takeforcedgiveCm",
                                     -aliases => ["sfo_power_reset_takeforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover / forced giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_reset_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_off_on_takeforcedgiveCm",
                                     -aliases => ["sfo_power_off_on_takeforcedgiveCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover / power on / forced giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2+" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_off_on_takeforcedgiveCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_off_takeoverCm",
                                     -aliases => ["sfo_power_off_takeoverCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power off / takeover event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_off_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_reset_takeoverCm",
                                     -aliases => ["sfo_power_reset_takeoverCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power reset / takeover event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_reset_takeoverCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_on_givebackCm",
                                     -aliases => ["sfo_power_on_givebackCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power on / giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 1 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_on_givebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::register_event(-event_name => "nacl_sfo_power_on_forcedgivebackCm",
                                     -aliases => ["sfo_power_on_forcedgivebackCm"],
                                     -event_modules => \@event_modules,
                                     -description => "This is a Cmode event that does a SFO power on / forced giveback event",
                                     -required_resource_types => ["nacl_nodes_cm"],
                                     -optional_resource_types => ["sanclients"],
                                     -resource_counts => { nacl_nodes_cm => "2" },
                                     -arg_defaults => { sfo_disable_autosupport_before_giveback => 0 },
                                     -arg_hardcoded => { clam_validation => 0 }, # must disable clam validation until NACL APIs are supported
                                     -pre_validate => sub { &_convert_sanfilersCm_to_nacl_nodes_cm(@_); },
                                     -exec => sub { &nacl_sfo_power_on_forcedgivebackCm(@_) });
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # define the default values for Event parameters
    $self->EventUtil::register_event_module_defaults(-event_module => __PACKAGE__,
                                                     -arg_defaults =>
                                                         { -controller_down_time => 600,
                                                           -controller_down_time_after_takeover =>
                                                               sub {
                                                                   my ($self) = @_;
                                                                   return $self->EventUtil::get_arg(-arg => "controller_down_time");
                                                               },
                                                           -controller_down_time_after_giveback =>
                                                               sub {
                                                                   my ($self) = @_;
                                                                   return $self->EventUtil::get_arg(-arg => "controller_down_time");
                                                               },
                                                           -sfo_timeout          => 300,
                                                           -sfo_max_timeout      =>
                                                               sub {
                                                                   my ($self) = @_;
                                                                   return $self->EventUtil::get_arg(-arg => "sfo_timeout") * 4;
                                                               },
                                                           -symmetrical_events   => 1,
                                                           -collect_sfo_aggr_outage_times => 0,
                                                           -sfo_display_sfo_aggr_outage_ems_logs => 0,
                                                           -sfo_allow_auto_giveback => 0,
                                                           -sfo_auto_giveback_threshold => 30,
                                                           -sfo_failed_auto_giveback_sleep_time => 90,
                                                           -sfo_detect_waiting_for_giveback_timeout => 300,
                                                           -sfo_giveback_retry_count => 3,
                                                           -sfo_giveback_retry_sleep_time => 300,
                                                           -sfo_force_giveback_on_disk_failure => 1,
                                                           -sfo_disable_autosupport_before_giveback => 1,
                                                           -sfo_enable_autosupport_after_giveback => 1,
                                                           -sfo_terminate_savecore_process => 1,
                                                           -sfo_terminate_savecore_process_frequency => 30,
                                                           -sfo_terminate_savecore_process_init_sleep_time =>
                                                               sub {
                                                                   my ($self) = @_;
                                                                   return int($self->EventUtil::get_arg(-arg => "controller_down_time") / 4);
                                                               },
                                                           -sfo_parallel_delay_time => 0
                                                       }
                                                     );
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __init()
# Objective: Initialize this Event Module
###############################################################################
sub __init {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    __uninit()
# Objective: Uninitialize this Event Module
###############################################################################
sub __uninit {
    my ($self) = @_;

    return $EventUtil::TRUE;
}

###############################################################################
# Internal Method:    _generate_event_param_documentation()
# Objective: Generate Event Parameter documentation for this Utility Module
###############################################################################
sub __generate_event_param_documentation {
    my ($self) = @_;

    $self->EventUtil::document_event_param(-param => "sfo_timeout",
                                           -package => __PACKAGE__,
                                           -default => "300",
                                           -format => "integer (seconds)",
                                           -example => "-sfo_timeout => 180",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the maximum timeout to wait for the ".
                                                           "cluster to stabilize after a storage failover. " .
                                                           "If exceeded, it will be treated as a failure");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_max_timeout",
                                           -package => __PACKAGE__,
                                           -default => "sfo_timeout * 4",
                                           -format => "integer (seconds)",
                                           -example => "-sfo_max_timeout => 360",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the upper limit timeout to wait for the ".
                                                           "cluster to stabilize after a storage failover beyond ".
                                                           " the value of sfo_timeout.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "controller_down_time",
                                           -package => __PACKAGE__,
                                           -default => "600",
                                           -format => "integer (seconds)",
                                           -example => "-controller_down_time => 600",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the amount of time to sleep after " .
                                                           "SFO (takeover/giveback) operations");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "controller_down_time_after_takeover",
                                           -package => __PACKAGE__,
                                           -default => "default of -controller_down_time",
                                           -format => "integer (seconds)",
                                           -example => "-controller_down_time_after_takeover => 600",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the amount of time to sleep after " .
                                                           "SFO takeover operations");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "controller_down_time_after_giveback",
                                           -package => __PACKAGE__,
                                           -default => "default of -controller_down_time",
                                           -format => "integer (seconds)",
                                           -example => "-controller_down_time_after_giveback => 600",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies the amount of time to sleep after " .
                                                           "SFO giveback operations");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "symmetrical_events",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-symmetrical_events => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Indicates whether to perform symmetrically styled " .
                                                           "events (A -> B then B -> A)");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "collect_sfo_aggr_outage_times",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-collect_sfo_aggr_outage_times => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Indicates whether to collect sfo aggr outages after givebacks");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_display_sfo_aggr_outage_ems_logs",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sfo_display_sfo_aggr_outage_ems_logs => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Indicates whether to display the EMS logs while determining aggr outages");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_allow_auto_giveback",
                                           -package => __PACKAGE__,
                                           -default => "0",
                                           -format => "0 (off) | 1 (on)",
                                           -example => "-sfo_allow_auto_giveback => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Allow ONTAP to auto giveback after an SFO");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_giveback_retry_count",
                                           -package => __PACKAGE__,
                                           -default => "3",
                                           -format => "positive integer",
                                           -example => "-sfo_giveback_retry_count => 5",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Retry SFO givebacks if it fails for any particular reason");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_disable_autosupport_before_giveback",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0|1",
                                           -example => "-sfo_disable_autosupport_before_giveback => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to turn autosupport off before givebacks.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_enable_autosupport_after_giveback",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0|1",
                                           -example => "-sfo_enable_autosupport_after_giveback => 1",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to turn autosupport on after givebacks.");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_giveback_retry_sleep_time",
                                           -package => __PACKAGE__,
                                           -default => "300",
                                           -format => "positive integer",
                                           -example => "-sfo_giveback_retry_sleep_time => 300",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies how long to sleep before retrying an SFO giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_terminate_savecore_process",
                                           -package => __PACKAGE__,
                                           -default => "1",
                                           -format => "0 or 1",
                                           -example => "-sfo_terminate_savecore_process => 0",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Specifies whether to terminate the savecore process during panic SFO takeovers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_terminate_savecore_process_frequency",
                                           -package => __PACKAGE__,
                                           -default => "60",
                                           -format => "positive integer",
                                           -example => "-sfo_terminate_savecore_process_frequency => 60",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Indicates how often to run the terminate savecore process method");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::document_event_param(-param => "sfo_terminate_savecore_process_init_sleep_time",
                                           -package => __PACKAGE__,
                                           -default => "floor(controller_down_time / 4)",
                                           -format => "positive integer",
                                           -example => "-sfo_terminate_savecore_process_init_sleep_time => 40",
                                           -per_resource_capable => "0",
                                           -environment_capable => "1",
                                           -description => "Indicates how long to wait before starting the savecore termination process");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Event Methods
###############################################################################

###############################################################################
# Event Method:    sfo_takeoverCm()
# Objective: Perform a storage failover takeover on a single CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_takeover_abstracted(-forcetake => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_givebackCm()
# Objective: Perform a storage failover giveback on a single CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_givebackCm {
    my ($self, %args) = @_;

    $self->sfo_giveback_abstracted(-forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_forcedtakeoverCm()
# Objective: Perform a storage failover forced takeover on a single CMode
#            HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_forcedtakeoverCm {
    my ($self, %args) = @_;

    $self->sfo_takeover_abstracted(-forcetake => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_forcedgivebackCm()
# Objective: Perform a storage failover forced giveback on a single CMode
#            HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_forcedgivebackCm {
    my ($self, %args) = @_;

    $self->sfo_giveback_abstracted(-forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_takegiveCm()
# Objective: Perform a storage failover takeover/giveback scenario against
#            multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_takegiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::FALSE,
                                   -forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_forcedtakegive()
# Objective: Perform a storage failover forced takeover/giveback scenario
#            against  multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_forcedtakegiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::TRUE,
                                   -forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_takeforcedgiveCm()
# Objective: Perform a storage failover takeover / forced giveback scenario
#            against  multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_takeforcedgiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::FALSE,
                                   -forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_forcedtakeforcedgiveCm()
# Objective: Perform a storage failover forced takeover / forced giveback
#            scenario against multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_forcedtakeforcedgiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::TRUE,
                                   -forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_panicgiveCm()
# Objective: Perform a storage failover panic / giveback scenario against
#            multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_panicgiveCm {
    my ($self, %args) = @_;

    $self->sfo_panicgive_abstracted(-forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_panicforcedgiveCm()
# Objective: Perform a storage failover panic / forced giveback scenario
#            against multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_panicforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_panicgive_abstracted(-forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_panicCm()
# Objective: Perform a storage failover panic on a single Cmode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_panicCm {
    my ($self, %args) = @_;

    $self->sfo_panic_abstracted();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_haltgiveCm()
# Objective: Perform a storage failover halt / giveback scenario against
#            multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_haltgiveCm {
    my ($self, %args) = @_;

    $self->sfo_haltgive_abstracted(-forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_haltforcedgiveCm()
# Objective: Perform a storage failover halt / forced giveback scenario
#            against multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_haltforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_haltgive_abstracted(-forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_haltCm()
# Objective: Perform a storage failover halt takeover against a single
#            CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_haltCm {
    my ($self, %args) = @_;

    $self->sfo_halt_abstracted();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_rebootgiveCm()
# Objective: Perform a storage failover reboot / giveback scenario against
#            multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_rebootgiveCm {
    my ($self, %args) = @_;

    $self->sfo_rebootgive_abstracted(-forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_rebootforcedgiveCm()
# Objective: Perform a storage failover reboot / forced giveback scenario
#            against multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_rebootforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_rebootgive_abstracted(-forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_rebootCm()
# Objective: Perform a storage failover reboot/takeover against a single
#            CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_rebootCm {
    my ($self, %args) = @_;

    $self->sfo_reboot_abstracted();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_off_on_takegiveCm()
# Objective: Trigger a storage failover takeover / giveback scenario on
#            multiple CMode HA pairs by powering Nodes on and off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_off_on_takegiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::FALSE,
                                    -power_change_type => "off");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_reset_takegiveCm()
# Objective: Trigger a storage failover takeover / giveback scenario on
#            multiple CMode HA pairs by resetting power on Nodes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_reset_takegiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::FALSE,
                                    -power_change_type => "reset");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_off_on_takeforcedgive()
# Objective: Trigger a storage failover takeover / forced giveback scenario on
#            multiple CMode HA pairs by powering Nodes on and off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_off_on_takeforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::TRUE,
                                    -power_change_type => "off");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_reset_takeforcedgiveCm()
# Objective: Trigger a storage failover takeover / forced giveback scenario on
#            multiple CMode HA pairs by resetting power on Nodes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_reset_takeforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::TRUE,
                                    -power_change_type => "reset");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_off_takeoverCm()
# Objective: Trigger a storage failover takeover scenario on a single CMode
#            HA Pair by cutting the power of a Node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_off_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "off");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_reset_takeoverCm()
# Objective: Trigger a storage failover takeover scenario on a single CMode
#            HA Pair by resetting power on a Node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_reset_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "reset");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_on_givebackCm()
# Objective: Perform a storage failover giveback scenario on a single CMode
#            HA Pair where one of the Nodes is powered off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_on_givebackCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "on",
                                       -forcegive => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_power_on_forcedgivebackCm()
# Objective: Perform a storage failover forced giveback scenario on a single CMode
#            HA Pair where one of the Nodes is powered off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_power_on_forcedgivebackCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "on",
                                       -forcegive => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_takeoverCm()
# Objective: Perform a storage failover takeover on a single CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_takeover_abstracted(-forcetake => $EventUtil::FALSE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_givebackCm()
# Objective: Perform a storage failover giveback on a single CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_givebackCm {
    my ($self, %args) = @_;

    $self->sfo_giveback_abstracted(-forcegive => $EventUtil::FALSE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_forcedtakeoverCm()
# Objective: Perform a storage failover forced takeover on a single CMode
#            HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_forcedtakeoverCm {
    my ($self, %args) = @_;

    $self->sfo_takeover_abstracted(-forcetake => $EventUtil::TRUE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_forcedgivebackCm()
# Objective: Perform a storage failover forced giveback on a single CMode
#            HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_forcedgivebackCm {
    my ($self, %args) = @_;

    $self->sfo_giveback_abstracted(-forcegive => $EventUtil::TRUE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_panicCm()
# Objective: Perform a storage failover panic on a single Cmode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_panicCm {
    my ($self, %args) = @_;

    $self->sfo_panic_abstracted(-use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_rebootCm()
# Objective: Perform a storage failover reboot on a single Cmode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_rebootCm {
    my ($self, %args) = @_;

    $self->sfo_reboot_abstracted(-use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_haltCm()
# Objective: Perform a SFO halt on a single Cmode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_haltCm {
    my ($self, %args) = @_;

    $self->sfo_halt_abstracted(-use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_takegiveCm()
# Objective: Perform a storage failover takeover/giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_takegiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::FALSE,
                                   -forcegive => $EventUtil::FALSE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_forcedtakegiveCm()
# Objective: Perform a storage failover forced takeover/giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_forcedtakegiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::TRUE,
                                   -forcegive => $EventUtil::FALSE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_takeforcedgiveCm()
# Objective: Perform a storage failover takeover/forced giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_takeforcedgiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::FALSE,
                                   -forcegive => $EventUtil::TRUE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_forcedtakeforcedgiveCm()
# Objective: Perform a storage failover forced takeover/forced giveback
#            scenario against multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_forcedtakeforcedgiveCm {
    my ($self, %args) = @_;

    # invoke the takegive abstracted method that will abstract the
    # different variants of take give events
    $self->sfo_takegive_abstracted(-forcetake => $EventUtil::TRUE,
                                   -forcegive => $EventUtil::TRUE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_panicgiveCm()
# Objective: Perform a storage failover panic / giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_panicgiveCm {
    my ($self, %args) = @_;

    $self->sfo_panicgive_abstracted(-forcegive => $EventUtil::FALSE,
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_panicforcedgiveCm()
# Objective: Perform a storage failover panic / forced giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_panicforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_panicgive_abstracted(-forcegive => $EventUtil::TRUE,
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_rebootgiveCm()
# Objective: Perform a storage failover reboot / giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_rebootgiveCm {
    my ($self, %args) = @_;

    $self->sfo_rebootgive_abstracted(-forcegive => $EventUtil::FALSE,
                                     -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_rebootforcedgiveCm()
# Objective: Perform a storage failover reboot / forced giveback scenario against
#            multiple CMode HA pairs (using NACL apis)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_rebootforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_rebootgive_abstracted(-forcegive => $EventUtil::TRUE,
                                     -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_haltgiveCm()
# Objective: Perform a storage failover halt / giveback scenario against
#            multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_haltgiveCm {
    my ($self, %args) = @_;

    $self->sfo_haltgive_abstracted(-forcegive => $EventUtil::FALSE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_haltforcedgiveCm()
# Objective: Perform a storage failover halt / forced giveback scenario
#            against multiple CMode HA pairs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_haltforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_haltgive_abstracted(-forcegive => $EventUtil::TRUE,
                                   -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_off_on_takegiveCm()
# Objective: Trigger a storage failover takeover / giveback scenario on
#            multiple CMode HA pairs by powering Nodes on and off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_off_on_takegiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::FALSE,
                                    -power_change_type => "off",
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_reset_takegiveCm()
# Objective: Trigger a storage failover takeover / giveback scenario on
#            multiple CMode HA pairs by resetting power on Nodes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_reset_takegiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::FALSE,
                                    -power_change_type => "reset",
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_off_on_takeforcedgive()
# Objective: Trigger a storage failover takeover / forced giveback scenario on
#            multiple CMode HA pairs by powering Nodes on and off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_off_on_takeforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::TRUE,
                                    -power_change_type => "off",
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_reset_takeforcedgiveCm()
# Objective: Trigger a storage failover takeover / forced giveback scenario on
#            multiple CMode HA pairs by resetting power on Nodes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_reset_takeforcedgiveCm {
    my ($self, %args) = @_;

    $self->sfo_powergive_abstracted(-forcegive => $EventUtil::TRUE,
                                    -power_change_type => "reset",
                                    -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_off_takeoverCm()
# Objective: Trigger a storage failover takeover scenario on a single CMode
#            HA Pair by cutting the power of a Node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_off_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "off",
                                       -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_reset_takeoverCm()
# Objective: Trigger a storage failover takeover scenario on a single CMode
#            HA Pair by resetting power on a Node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_reset_takeoverCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "reset",
                                       -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_on_givebackCm()
# Objective: Perform a storage failover giveback scenario on a single CMode
#            HA Pair where one of the Nodes is powered off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_on_givebackCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "on",
                                       -forcegive => $EventUtil::FALSE,
                                       -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    nacl_sfo_power_on_forcedgivebackCm()
# Objective: Perform a storage failover forced giveback scenario on a single CMode
#            HA Pair where one of the Nodes is powered off
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nacl_sfo_power_on_forcedgivebackCm {
    my ($self, %args) = @_;

    $self->sfo_power_change_abstracted(-power_change_type => "on",
                                       -forcegive => $EventUtil::TRUE,
                                       -use_nacl_apis => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    sfo_takeover_abstracted()
# Objective: Issue a forced / non forced takeover on a single Node
###############################################################################
sub sfo_takeover_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $forcetake = &ParamsMgr::get_arg(-arg => "forcetake",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                         -faulted => $EventUtil::FALSE);
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);
    }

    # use only the first node on the list, this event cannot handle multiple node HA pairs
    # unless this event were run in parallel with different node lists
    my $primary_node = $node_names[0];

    # determine the partner name of the primary node
    my (undef, undef, $pairs) = $self->parse_node_pairs($primary_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $partner_node = $pairs->{$primary_node};

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_takeover_on_${primary_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # issue a takeover on this primary node
    $self->sfo_takeover_generic(-takeover_node => $primary_node,
                                -taken_over_node => $partner_node,
                                -forcetake => $forcetake,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "takeover_on_${primary_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "takeover_on_${primary_node}",
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $partner_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_giveback_abstracted()
# Objective: Issue a forced / non forced giveback on a single Node
###############################################################################
sub sfo_giveback_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # use only the first node on the list, this event cannot handle multiple node HA pairs
    # unless this event were run in parallel with different node lists
    my $primary_node = $node_names[0];

    # determine the partner name of the primary node
    my (undef, undef, $pairs) = $self->parse_node_pairs($primary_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $partner_node = $pairs->{$primary_node};

    # issue a giveback on this primary node
    $self->sfo_giveback_generic(-giveback_node => $primary_node,
                                -taken_over_node => $partner_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${primary_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${primary_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate paths after giveback occurred
    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_${primary_node}",
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram dump detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_panic_abstracted()
# Objective: Perform a storage failover panic on a single Cmode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_panic_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                         -faulted => $EventUtil::FALSE);
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # only perform the panic takeover on the first HA pair
    my $takeover_node = shift @{$primaries};
    my $panic_node = $pairs->{$takeover_node};

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_panic_on_${panic_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # perform a panic on this ha pair
    $self->sfo_panic_generic(-takeover_node => $takeover_node,
                             -panic_node => $panic_node,
                             -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::get_arg(-arg  => "controller_down_time_after_takeover");

    # check if user requested to terminate savecore process during panic takeovers
    my $terminate_savecore_process_subtest;
    if ($self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process")) {
        $self->terminate_savecore_process(-node => $takeover_node,
                                          -total_exec_time => $controller_down_time,
                                          -async => 0,
                                          -use_nacl_apis => $use_nacl_apis);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        # otherwise just sleep the time in the foreground
        $self->EventUtil::sleep_arg(-arg => "controller_down_time_after_takeover");
    }

    # synchronize before running path validation
    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "panic_on_${panic_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate paths after the takeover occurred
    $self->PathValidation::validate_paths_after_takeover(-id_extension => "panic_on_" . $panic_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $panic_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_reboot_abstracted()
# Objective: Perform a storage failover reboot/takeover against a single
#            CMode HA Pair
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_reboot_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                         -faulted => $EventUtil::FALSE);
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # only perform the reboot takeover on the first HA pair
    my $takeover_node = shift @{$primaries};
    my $reboot_node = $pairs->{$takeover_node};

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_reboot_on_${reboot_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # perform the reboot takeover on this HA pair
    $self->sfo_reboot_generic(-takeover_node => $takeover_node,
                              -reboot_node => $reboot_node,
                              -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "reboot_on_${reboot_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "reboot_on_" . $reboot_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $reboot_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Event Method:    sfo_halt_abstracted()
# Objective: Perform a storage failover halt takeover against a single
#            CMode node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_halt_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                         -faulted => $EventUtil::FALSE);
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # only perform the halt takeover on the first HA pair
    my $takeover_node = shift @{$primaries};
    my $halt_node = $pairs->{$takeover_node};

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_halt_on_${halt_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # perform the halt takeover on this HA pair
    my $halt_subtest = $self->sfo_halt_generic(-takeover_node => $takeover_node,
                                               -halt_node => $halt_node,
                                               -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    # reap the halt subtest
    $self->wait_for_subtests(-subtests => [$halt_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "halt_on_${halt_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "halt_on_" . $halt_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $halt_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_takegive_abstracted()
# Objective: Issue a forced / non forced takeover / giveback scenario against
#            multiple HA pairs, one pair at a time.
###############################################################################
sub sfo_takegive_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $allow_version_mismatch = &ParamsMgr::get_arg(-arg => "allow_version_mismatch",
                                                     -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # split up all the nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # for each HA pair, perform a takeover giveback
    for (my $i = 0; $i < scalar(@{$primaries}); $i++) {
        my $primary_node = $primaries->[$i];
        my $secondary_node = $pairs->{$primary_node};

        # delete the allow_version_mismatch option on the first takegive (if doing NDU)
        if ($allow_version_mismatch == $EventUtil::TRUE) {
            delete($args{-allow_version_mismatch});
        }

        # perform the takeover giveback on this HA pair
        $self->sfo_takegive_generic(-takeover_node => $primary_node,
                                    -taken_over_node => $secondary_node,
                                    %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if symmetrical events is enabled, then perform the event on the
        # secondaries as well.
        my $symmetrical_events = $self->EventUtil::get_arg(-arg => "symmetrical_events");
        if ($symmetrical_events == 0) {
            next;
        }

        # add the allow_version_mismatch option for the second takeover giveback (if doing NDU)
        if ($allow_version_mismatch == $EventUtil::TRUE) {
            $args{-allow_version_mismatch} = $allow_version_mismatch;
        }

        # perform the takeover giveback on this HA pair (reversed)
        $self->sfo_takegive_generic(-takeover_node => $secondary_node,
                                    -taken_over_node => $primary_node,
                                    %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_takegive_generic()
# Objective: Perform a takeover giveback scenario against a single HA Pair
###############################################################################
sub sfo_takegive_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    my $forcetake = &ParamsMgr::get_arg(-arg => "forcetake",
                                        -default => $EventUtil::FALSE);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $allow_version_mismatch = &ParamsMgr::get_arg(-arg => "allow_version_mismatch",
                                                     -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_takeover_on_${takeover_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # issue a takeover on the takeover node
    $self->sfo_takeover_generic(-takeover_node => $takeover_node,
                                -taken_over_node => $taken_over_node,
                                -forcetake => $forcetake,
                                -allow_version_mismatch => $allow_version_mismatch,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "takeover_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "takeover_on_${takeover_node}",
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $taken_over_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # issue a giveback on the takeover node
    $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                -taken_over_node => $taken_over_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_takeover_generic()
# Objective: Issue a takeover on a single Node
###############################################################################
sub sfo_takeover_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    my $forcetake = &ParamsMgr::get_arg(-arg => "forcetake",
                                        -default => $EventUtil::FALSE);
    my $allow_version_mismatch = &ParamsMgr::get_arg(-arg => "allow_version_mismatch",
                                                     -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my ($take_over_nacl_node_obj, $sanfiler_obj);

    if ($use_nacl_apis eq $EventUtil::FALSE) {
        # fetch the sanfiler object to use later
        $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $take_over_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $option;
    if ($allow_version_mismatch == $EventUtil::TRUE) {
        $option = "allow-version-mismatch";
    } elsif ($forcetake == $EventUtil::TRUE) {
        $option = "immediate";
    } else {
        $option = "normal";
    }

    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $sanfiler_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect stats
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $taken_over_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_start_sentinel = "takeover start " . time;

    # store this takeover start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        # only set it on sanfiler objects if we are not using nacl events
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    # always set the attrs for nacl resources
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # make sure auto giveback is disabled before issuing the takeover
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            # turn on auto giveback
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            # turn off auto giveback
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        # turn off auto giveback for Controller cluster
        $sanfiler_obj->disable_auto_giveback();
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing $option takeover on '$takeover_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing $option takeover on '$takeover_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # place the node that will be taken over in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $taken_over_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # takeover the partner node
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        try {
            NACL::C::StorageFailover->takeover(command_interface => $take_over_nacl_node_obj,
                                               option => $option,
                                               bynode => $takeover_node);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        # use sanfiler APIs to do the takeover
        $sanfiler_obj->sfo_takeover(-option => $option);
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # Store the exact time takeover call exited.
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -taken_over_node         => $taken_over_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate that the takeover occurred within -sfo_timeout
    my $takeover_time;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $takeover_time = $self->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                    -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                    -take_over_node => $takeover_node,
                                                    -taken_over_node => $taken_over_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $takeover_time = $sanfiler_obj->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                            -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"));
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # if the takeover time exeeded -sfo_timeout, then issue a warning
        my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
        if ($takeover_time > $sfo_timeout) {
            &SanUtil::nate_log('-warn', "SFO takeover completed but took ($takeover_time) seconds, " .
                                        "which is longer than sfo_timeout ($sfo_timeout)");
        }
    }

    # broadcast that the takeover has completed
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $taken_over_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_end_sentinel = "takeover end " . time;

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** $option takeover on '$takeover_node' has completed",
                                                   $sfo_takeover_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store this takeover end sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        # only set it for sanfiler objects if we are not using nacl APIs
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    detect_sfo_takeover()
# Objective: Detect an SFO takeover given a specified set of timeout values
###############################################################################
sub detect_sfo_takeover {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $sfo_timeout = &ParamsMgr::get_arg(-arg => "sfo_timeout",
                                          -required => 1);
    my $sfo_max_timeout = &ParamsMgr::get_arg(-arg => "sfo_max_timeout",
                                              -required => 1);
    my $take_over_node = &ParamsMgr::get_arg(-arg => "take_over_node",
                                             -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return -1;
    }

    my $take_over_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $take_over_node);
    if ($self->error()) {
        return -1;
    }
    my $taken_over_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $taken_over_node);
    if ($self->error()) {
        return -1;
    }

    # store the start time that the sfo will occur
    my $start_time = time();
    my $end_time;

    # default sfo max timeout to be the same value as sfo_timeout, if it is less than sfo_timeout
    if ($sfo_max_timeout < $sfo_timeout) {
        $sfo_max_timeout = $sfo_timeout;
    }

    # calculate the lower limit timeout time, this will be the time we will
    # report a warning if the sfo time exceeded
    my $sfo_timeout_time = $start_time + $sfo_timeout;

    # calculate the upper limit timeout time, this will be the time we will
    # report a failure if the sfo has not occurred
    my $sfo_max_timeout_time = $start_time + $sfo_max_timeout;

    my $timeout_hit = $EventUtil::FALSE;
    my $max_timeout_hit = $EventUtil::FALSE;

    # wait for the takeover to complete
    try {
        $self->wait_for_sfo_state(-command_interface => $take_over_node_obj,
                                  -sfo_type => "takeover",
                                  -sfo_state_hash =>
                                      {"$take_over_node" => { states => ["takeover"],
                                                              abort_states => ["takeover_failed"]}},
                                  -method_timeout => $sfo_max_timeout);

        $end_time = time();
        # check if we went over "sfo_timeout"
        if ($sfo_timeout_time < $end_time) {
            # we exceeded sfo_timeout
            $timeout_hit = $EventUtil::TRUE;
        }
        # check if we went over "sfo_max_timeout"
        if ($sfo_max_timeout_time < $end_time) {
            $max_timeout_hit = $EventUtil::TRUE;
        }
    } catch EventMan::Exceptions::TimedOutWaitingForSFOState with {
        # we exceeded both timeouts (max and warn)
        $end_time = time();
        $max_timeout_hit = $EventUtil::TRUE;
        $timeout_hit = $EventUtil::TRUE;
    } catch EventMan::Exceptions::EncounteredSFOAbortState with {
        # this node has hit one of the abort states
        my $e = shift;
        $self->_set_err("Encountered SFO Abort State: " . $e->text());
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error()) {
        return -1;
    }

    my $sfo_total_time = $end_time - $start_time;

    # check if any timeouts were hit
    if ($max_timeout_hit eq $EventUtil::TRUE) {
        # issue an error
        $self->_set_err("SFO takeover never occurred and exceeded SFO max " .
                        "timeout $sfo_max_timeout seconds. " .
                        "($sfo_total_time seconds)");
    } elsif ($timeout_hit eq $EventUtil::TRUE) {
        # issue a warning
        &SanUtil::nate_log('-warn', "SFO takeover completed but took ($sfo_total_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout) seconds.");
    } else {
        # the takeover successfully completed within our timeout windows
        &SanUtil::nate_log('-comment', "SFO takeover took $sfo_total_time " .
                           "seconds to complete (timeout = $sfo_timeout)");
    }

    return $sfo_total_time;
}

###############################################################################
# Method:    sfo_giveback_generic()
# Objective: Issue a giveback on a single Node
###############################################################################
sub sfo_giveback_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $giveback_node = &ParamsMgr::get_arg(-arg => "giveback_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $sfo_allow_auto_giveback = $self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback");
    my $perform_giveback = 1;

    my ($sanfiler_obj, $giveback_nacl_node_obj, $taken_over_nacl_node_obj);

    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $giveback_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $giveback_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $giveback_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $taken_over_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $taken_over_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $sanfiler_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # determine if we are going to do a forced or non forced giveback
    my $msg;
    my $override_vetoes;
    my $require_partner_waiting;
    if ($forcegive == $EventUtil::TRUE) {
        $override_vetoes = "true";
        $require_partner_waiting = "true";
        $msg = "forced ";
    } else {
        $override_vetoes = "false";
        $require_partner_waiting = "true";
        $msg = "";
    }

    # broadcast that a giveback is going to happen,
    # include data indicating which node is giving back
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_giveback",
                                          -data => {give_back_node => $giveback_node,
                                                    taken_over_node => $taken_over_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_giveback_start_sentinel = "giveback start " . time;

    # store this giveback start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $giveback_node,
                                              -type => "sanfilersCm",
                                              -attribute => "giveback_start_sentinel",
                                              -value => $sfo_giveback_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $giveback_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "giveback_start_sentinel",
                                          -value => $sfo_giveback_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing ${msg}giveback on '$giveback_node'",
                                                   $sfo_giveback_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing ${msg}giveback on '$giveback_node'",
                                                   $sfo_giveback_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # synchronize before performing the action
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # Check to see if caller allows auto giveback
    if ($sfo_allow_auto_giveback == 1) {
        # check SFO status
        if ($use_nacl_apis eq $EventUtil::TRUE) {
            try {
                $perform_giveback = $self->check_for_auto_sfo_giveback(-giveback_node => $giveback_node,
                                                                       -taken_over_node => $taken_over_node);
            } otherwise {
                my $e = shift;
                $self->_set_err("Unable to detect if an auto giveback occurred: Exception: " . $e->text());
            };
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            my ($output) = $sanfiler_obj->sfo_show(-fields => "state,possible,enabled,reason",
                                                   -node => $giveback_node);
            if ($output->{state}    eq "connected" and
                $output->{possible} eq "true" and
                $output->{enabled}  eq "true" and
                $output->{reason}   eq "-") {
                # We are in a giveback state already, don't need to do anything
                $perform_giveback = 0;
            } else {
                # We are not in a giveback state, we might be in a takeover state or
                # in the middle of a giveback, check takeover state
                # check the SFO status
                ($output) = $sanfiler_obj->sfo_show(-fields => "state,possible,enabled,time-until-takeover",
                                                    -node => $giveback_node);
                if ($output->{state}    eq "takeover" and
                    $output->{possible} eq "false" and
                    $output->{enabled}  eq "true") {
                    # We are in a takeover state and need to giveback
                    $perform_giveback = 1;
                } else {
                    # We are probably in the middle of a giveback, just detect giveback
                    $perform_giveback = 0;
                }
            }
        }
    }

    my ($sfo_giveback_start_time, $cv_subtest, $giveback_time, $rdata);

    my $sfo_giveback_retry_count = $self->EventUtil::get_arg(-arg => "sfo_giveback_retry_count");
    my $sfo_giveback_retry_sleep_time = $self->EventUtil::get_arg(-arg => "sfo_giveback_retry_sleep_time");
    my $disable_autosupport = 1;

    # keep attempting to perform the giveback for the number of retries specified
    while (1) {
        if ($perform_giveback == 1) {

            # disable autosupport before giveback if requested
            if ($self->EventUtil::get_arg(-arg => "sfo_disable_autosupport_before_giveback") and $disable_autosupport) {
                if ($use_nacl_apis eq $EventUtil::TRUE) {
                    try {
                        # ignore any exceptions thrown
                        # must issue commands seperately until NACL resolves burt 785858 
                        NACL::C::SystemNodeAutosupport->modify(command_interface => $giveback_nacl_node_obj,
                                                               node => "${giveback_node}",
                                                               support => "disable",
                                                               state => "disable",
                                                               'local-collection' => "false");

                        NACL::C::SystemNodeAutosupport->modify(command_interface => $giveback_nacl_node_obj,
                                                               node => "${taken_over_node}",
                                                               support => "disable",
                                                               state => "disable",
                                                               'local-collection' => "false");
                    } otherwise{
                        # Do nothing
                    };
                } else {
                    $sanfiler_obj->autosupport_modify(node => "${giveback_node}",
                                                      support => "disable",
                                                      state => "disable",
                                                      local_collection => "false");

                    $sanfiler_obj->autosupport_modify(node => "${taken_over_node}",
                                                      support => "disable",
                                                      state => "disable",
                                                      local_collection => "false");

                    # ignore any errors found
                    $sanfiler_obj->_clear_err();
                }
            }

            # make sure the taken over node is in the waiting for giveback state
            my $giveback_has_started = 0;
            if ($use_nacl_apis eq $EventUtil::TRUE) {
                my $rc = $self->detect_waiting_for_giveback(-take_over_node => $giveback_node,
                                                            -taken_over_node => $taken_over_node,
                                                            -timeout => $self->EventUtil::get_arg(-arg => "sfo_detect_waiting_for_giveback_timeout"));
                if ($self->error()) {
                    return $EventUtil::FALSE;
                }
                if ($rc == -2) {
                    # then the giveback started while we were looking to see if it was ready for giveback.
                    # allow the giveback to proceed automatically
                    $giveback_has_started = 1;
                }
            } else {
                $sanfiler_obj->detect_waiting_for_giveback(-node => $taken_over_node,
                                                           -take_over_node => $giveback_node,
                                                           -timeout => $self->EventUtil::get_arg(-arg => "sfo_detect_waiting_for_giveback_timeout"));
                if ($sanfiler_obj->error()) {
                    $self->_set_err($sanfiler_obj->errmsg());
                    return $EventUtil::FALSE;
                }
            }

            if ($use_nacl_apis eq $EventUtil::TRUE) {
                if ($giveback_has_started == 0) {
                    # perform the giveback (using NACL) only if the giveback has not started yet
                    my $retry = 0;
                    try {
                        NACL::C::StorageFailover->giveback(command_interface => $giveback_nacl_node_obj,
                                                           fromnode => $giveback_node,
                                                           'require-partner-waiting' => $require_partner_waiting,
                                                           'override-vetoes' => $override_vetoes,
                                                           'only-cfo-aggregates' => "false");
                    } otherwise {
                        my $e = shift;
                        $sfo_giveback_retry_count--;
                        if ($sfo_giveback_retry_count <= 0) {
                            # if we exhausted our retry counts, then error out
                            $self->_set_err("SFO giveback failed after retrying several times. Exception caught: " . $e->text);
                        } else {
                            # warn and retry the giveback
                            &SanUtil::nate_log('-warn', "SFO giveback failed, will retry in $sfo_giveback_retry_sleep_time " .
                                                   "seconds (retries left = $sfo_giveback_retry_count).\nException caught: " . $e->text);
                            &SanUtil::nate_sleep($sfo_giveback_retry_sleep_time);
                            $retry = 1;
                        }
                    };
                    if ($self->error()) {
                        return $EventUtil::FALSE;
                    }
                    # retry the giveback if we set the flag
                    next if ($retry);
                }
            } else {
                # perform the giveback (using controller cluster)
                $sanfiler_obj->sfo_giveback('-override-vetoes' => $override_vetoes,
                                            '-require-partner-waiting' => $require_partner_waiting);
                if ($sanfiler_obj->error()) {
                    # if we see "retry later" in the giveback error message, fall back to the retry logic
                    if (grep(/retry\s+later|Partner not ready/i, $sanfiler_obj->errmsg())) {
                        # decrease the sfo giveback retry count
                        $sfo_giveback_retry_count--;

                        if ($sfo_giveback_retry_count <= 0) {
                            # if we exhausted our retry counts, then error out
                            $self->_set_err("SFO giveback failed after retrying several times.");
                            return $EventUtil::FALSE;
                        } else {
                            # warn and retry the giveback
                            &SanUtil::nate_log('-warn', "SFO giveback failed, will retry in $sfo_giveback_retry_sleep_time " .
                                                        "seconds (retries left = $sfo_giveback_retry_count)");
                            &SanUtil::nate_sleep($sfo_giveback_retry_sleep_time);
                            next;
                        }
                    } elsif (grep(/giveback is already in progress/i, $sanfiler_obj->errmsg())) {
                        # looks like something has already issued the giveback,
                        &SanUtil::nate_log('-comment', "An auto-giveback appears to have just occurred, continuing...");
                    } else {
                        # otherwise the giveback failed, bail out now
                        $self->_set_err($sanfiler_obj->errmsg());
                        return $EventUtil::FALSE;
                    }
                }
            }
        }

        # Store the exact time giveback call exited.
        $sfo_giveback_start_time = time;
        $cv_subtest = $self->ClamValidationUtil::validate_clam_after_giveback(-taken_over_node         => $taken_over_node,
                                                                              -original_master         => $clam_master,
                                                                              -sfo_giveback_start_time => $sfo_giveback_start_time);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # detect how long it took for the giveback to occur
        if ($use_nacl_apis eq $EventUtil::TRUE) {
            $giveback_time = $self->detect_sfo_giveback(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                        -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                        -taken_over_node => $taken_over_node,
                                                        -giveback_node => $giveback_node);
            if ($self->error()) {
                $self->_set_err("Failed to detect SFO giveback");
                return $EventUtil::FALSE;
            }
        } else {
            $giveback_time = $sanfiler_obj->detect_sfo_giveback(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                                -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                                -wait_for_aggrs_to_sendhome => 1);
            if ($sanfiler_obj->error()) {
                $self->_set_err($sanfiler_obj->errmsg());
                return $EventUtil::FALSE;
            }
        }

        # if giveback time is -1, then the giveback failed from an internal error,
        # attempt to retry the giveback
        if ($giveback_time < 0) {
            # if cv_subtest is defined, kill it
            if (defined($cv_subtest) and ref($cv_subtest) =~ /Subtest/i) {
                $cv_subtest->kill(message => ["-na", "Stopping CLAM Validation because the giveback failed"]);
                $cv_subtest->wait_finish();
            }
            # decrease the sfo giveback retry count
            $sfo_giveback_retry_count--;

            if ($sfo_giveback_retry_count <= 0) {
                # if we exhausted our retry counts, then error out
                $self->_set_err("SFO giveback failed after retrying several times.");
                return $EventUtil::FALSE;
            } else {
                # warn and retry the giveback
                if ($perform_giveback eq 0) {
                    # if perform giveback is 0, then the auto giveback failed. Manually retry it
                    $Log->warn("SFO auto-giveback failed, will retry manually in $sfo_giveback_retry_sleep_time " .
                               "seconds (retries left = $sfo_giveback_retry_count)");
                    $perform_giveback = 1;
                    # we need to go through the disable autosupport routines.
                    $disable_autosupport = 1;
                } else {
                    $Log->warn("SFO manual giveback failed, will retry in $sfo_giveback_retry_sleep_time " .
                               "seconds (retries left = $sfo_giveback_retry_count)");
                    # dont disable autosupport (it should have been disabled before already)
                    $disable_autosupport = 0;
                }

                # if we failed the giveback because of a disk check error make sure the retried
                # giveback is forced
                if ($giveback_time == -3) {
                    $override_vetoes = "true";
                    $require_partner_waiting = "true";
                    $msg = "forced ";
                    # dont wait if we are trying to force the giveback after this
                    next;
                }

                # Sleep a some time before retrying the giveback
                &SanUtil::nate_sleep($sfo_giveback_retry_sleep_time);
                next;
            }
        } else {
            # break out of the giveback retry loop
            # because the giveback succeeded or it exhausted the completion timeout
            last;
        }
    }

    # mark that the takenover node is now in an unfaulted state
    $self->ResMgr::unset_faulted_resource(-name => $taken_over_node,
                                          -type => "nacl_nodes_cm",
                                          -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # if the giveback time exceeded -sfo_timeout, then issue a warning
    my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
    if ($giveback_time > $sfo_timeout) {
        &SanUtil::nate_log('-warn', "SFO giveback completed but took ($giveback_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout)");
    }

    # broadcast that the giveback process has completed
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_giveback",
                                          -data => {give_back_node => $giveback_node,
                                                    taken_over_node => $taken_over_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # refresh the connection to the taken over node (a version change may have occurred)
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->EventUtil::reconnect_all_sanfilers(-controller_names => [$taken_over_node]);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->EventUtil::reconnect_all_nacl_nodes(-controller_names => [$taken_over_node],
                                               -no_version_change => $EventUtil::FALSE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # toggle_scsit_fault_injection_panic_cdb on the node that just came back up
    $self->toggle_scsit_fault_injection_panic_cdb(-nodes => [$taken_over_node]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_giveback_end_sentinel = "giveback done " . time;

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** ${msg}giveback on '$giveback_node' has completed",
                                                   $sfo_giveback_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store this giveback start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $giveback_node,
                                              -type => "sanfilersCm",
                                              -attribute => "giveback_end_sentinel",
                                              -value => $sfo_giveback_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                              -type => "sanfilersCm",
                                              -attribute => "giveback_end_sentinel",
                                              -value => $sfo_giveback_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $giveback_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "giveback_end_sentinel",
                                          -value => $sfo_giveback_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $taken_over_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "giveback_end_sentinel",
                                          -value => $sfo_giveback_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # get the value of the takeover_start_sentinel from the ResMgr that exists for the giveback node
    my $start_sentinel;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $start_sentinel = $self->ResMgr::get_resource_attribute(-name => $giveback_node,
                                                                -type => "nacl_nodes_cm",
                                                                -attribute => "takeover_start_sentinel");
    } else {
        $start_sentinel = $self->ResMgr::get_resource_attribute(-name => $giveback_node,
                                                                -type => "sanfilersCm",
                                                                -attribute => "takeover_start_sentinel");
    }
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # dump the SFO Aggr outage information
    $self->get_sfo_aggr_outage_times(-giveback_node => $giveback_node,
                                     -taken_over_node => $taken_over_node,
                                     -start_sentinel => $start_sentinel,
                                     -end_sentinel => $sfo_giveback_end_sentinel,
                                     -display_label => "SFO Giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # enable autosupport after giveback if requested
    if ($self->EventUtil::get_arg(-arg => "sfo_enable_autosupport_after_giveback")) {
        if ($use_nacl_apis eq $EventUtil::TRUE) {
            try {
                # ignore any exceptions thrown
                NACL::C::SystemNodeAutosupport->modify(command_interface => $giveback_nacl_node_obj,
                                                       node => "${giveback_node}",
                                                       support => "enable",
                                                       state => "enable",
                                                       'local-collection' => "true");

                NACL::C::SystemNodeAutosupport->modify(command_interface => $giveback_nacl_node_obj,
                                                       node => "${taken_over_node}",
                                                       support => "enable",
                                                       state => "enable",
                                                       'local-collection' => "true");
            } otherwise {
                # Do nothing
            };
        } else {
            $sanfiler_obj->autosupport_modify(node => "${giveback_node}",
                                              support => "enable",
                                              state => "enable",
                                              local_collection => "true");

            $sanfiler_obj->autosupport_modify(node => "${taken_over_node}",
                                              support => "enable",
                                              state => "enable",
                                              local_collection => "true");

            # ignore any errors found
            $sanfiler_obj->_clear_err();
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    detect_waiting_for_giveback()
# Objective: Detect that an HA pair has entered the waiting for giveback state
###############################################################################
sub detect_waiting_for_giveback {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $timeout = &ParamsMgr::get_arg(-arg => "timeout",
                                      -required => 1);
    my $take_over_node = &ParamsMgr::get_arg(-arg => "take_over_node",
                                             -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return -1;
    }

    my $take_over_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $take_over_node);
    if ($self->error()) {
        return -1;
    }
    my $taken_over_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $taken_over_node);
    if ($self->error()) {
        return -1;
    }

    my $start_time = time();
    my $timeout_hit = $EventUtil::FALSE;
    my $giveback_already_in_progress = $EventUtil::FALSE;

    try {
        $self->wait_for_sfo_state(-command_interface => $take_over_node_obj,
                                  -sfo_state_hash =>
                                      {"$take_over_node"  => {states => ["takeover"], abort_states => ["giving_back", "giveback_partial_connected"]},
                                       "$taken_over_node" => {state_descriptions => [qr/Waiting for giveback \(HA Mailboxes\)/i]}},
                                  -method_timeout => $timeout);

    } catch EventMan::Exceptions::TimedOutWaitingForSFOState with {
        # we exceeded our timeout (max and warn)
        $timeout_hit = $EventUtil::TRUE;
    } catch EventMan::Exceptions::EncounteredSFOAbortState with {
        # this node has hit one of the abort states (a giveback was issued prior to entering this function)
        my $e = shift;
        # check to see if the giveback is in progress
        if ($e->get_node_sf_output()->state() =~ /giving_back|giveback_partial_connected/) {
            # if it is, then return back a special error code that indicates that the giveback is already
            # in progress
            $giveback_already_in_progress = $EventUtil::TRUE;
            $Log->comment("Detected that an autogiveback is in progress");

        } else {
            $self->_set_err("Exception caught while detecting waiting for giveback: " . $e->text());
        }
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught while detecting waiting for giveback: " . $e->text());
    };
    if ($self->error()) {
        return -1;
    }

    # return a special error code that the giveback has already started
    if ($giveback_already_in_progress eq $EventUtil::TRUE) {
        return -2;
    }

    my $end_time = time() - $start_time;

    if ($timeout_hit eq $EventUtil::TRUE) {
        $self->_set_err("Exceeded detect waiting for giveback timeout ($timeout seconds) on node '$taken_over_node'. " .
                        "Was expecting it to be in a waiting for giveback state, but it was not.");
    }

    return $end_time;
}

###############################################################################
# Method:    detect_sfo_giveback()
# Objective: Detect an SFO giveback given a specified set of timeout values
###############################################################################
sub detect_sfo_giveback {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $sfo_timeout = &ParamsMgr::get_arg(-arg => "sfo_timeout",
                                          -required => 1);
    my $sfo_max_timeout = &ParamsMgr::get_arg(-arg => "sfo_max_timeout",
                                              -required => 1);
    my $giveback_node = &ParamsMgr::get_arg(-arg => "giveback_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        # error state
        return -2;
    }

    my $giveback_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $giveback_node);
    if ($self->error()) {
        # error state
        return -2;
    }
    my $taken_over_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $taken_over_node);
    if ($self->error()) {
        # error state
        return -2;
    }

    # store the start time that the sfo will occur
    my $start_time = time();
    my $end_time;

    # default sfo max timeout to be the same value as sfo_timeout, if it is less than sfo_timeout
    if ($sfo_max_timeout < $sfo_timeout) {
        $sfo_max_timeout = $sfo_timeout;
    }

    # calculate the lower limit timeout time, this will be the time we will
    # report a warning if the sfo time exceeded
    my $sfo_timeout_time = $start_time + $sfo_timeout;

    # calculate the upper limit timeout time, this will be the time we will
    # report a failure if the sfo has not occurred
    my $sfo_max_timeout_time = $start_time + $sfo_max_timeout;

    my $timeout_hit = $EventUtil::FALSE;
    my $max_timeout_hit = $EventUtil::FALSE;
    my $retry_giveback = 0;
    my $force_giveback = 0;

    # wait for the giveback to complete
    try {
        $self->wait_for_sfo_state(-command_interface => $giveback_node_obj,
                                  -sfo_type => "giveback",
                                  -sfo_show_nodes => [$giveback_node],
                                  -sfo_state_hash =>
                                      {"$giveback_node"   => { state_descriptions => [qr/\"Connected to $taken_over_node\"/,
                                                                                      qr/\"Connected to $taken_over_node, Takeover is not possible: The version of software running on each node of the SFO pair is incompatible/i,
                                                                                      qr/\"Connected to $taken_over_node, Takeover is not possible: The size of NVRAM on each node of the SFO pair is different/i],
                                                               abort_state_descriptions => [qr/Previous giveback failed in module:/] },
                                       "$taken_over_node" => { state_descriptions => [qr/\"Connected to $giveback_node\"/,
                                                                                      qr/\"Connected to $giveback_node, Takeover is not possible: The version of software running on each node of the SFO pair is incompatible/i]}},
                                  -method_timeout => $sfo_max_timeout);

        $end_time = time();
        # check if we went over "sfo_timeout"
        if ($sfo_timeout_time < $end_time) {
            # we exceeded sfo_timeout
            $timeout_hit = $EventUtil::TRUE;
        }
        # check if we went over "sfo_max_timeout"
        if ($sfo_max_timeout_time < $end_time) {
            $max_timeout_hit = $EventUtil::TRUE;
        }
    } catch EventMan::Exceptions::TimedOutWaitingForSFOState with {
        # we exceeded our both timeouts (max and warn)
        $end_time = time();
        $max_timeout_hit = $EventUtil::TRUE;
        $timeout_hit = $EventUtil::TRUE;
    } catch EventMan::Exceptions::EncounteredSFOAbortState with {
        # this node has hit one of the abort states
        my $e = shift;

        # check if this abort state is retryable (Previous giveback failed in module: coredump|autosupport)
        if ($e->get_node_sf_output()->state_description() =~ /Previous giveback failed in module: (coredump|autosupport)/i) {
            $retry_giveback = 1;
        } elsif ($e->get_node_sf_output()->state_description() =~ /Previous giveback failed in module: disk check/i) {
            # if we failed for a disk check, set the retry flag only if we were told to force givebacks
            if ($self->EventUtil::get_arg(-arg => "sfo_force_giveback_on_disk_failure")) {
                $Log->warn("The giveback was vetoed due to a disk being failed, retrying but this time with a forced giveback");
                $retry_giveback = 1;
                $force_giveback = 1;
            } else {
                # otherwise we hit an error
                $self->_set_err("Encountered SFO Abort State: " . $e->text());
            }
        } else {
            # otherwise we hit an error
            $self->_set_err("Encountered SFO Abort State: " . $e->text());
        }
    } otherwise {
        my $e = shift;

        $self->_set_err("Exception caught: " . $e->text());
    };
    if ($self->error() or $retry_giveback) {
        # return -1 here indicates that we encountered an error while detecting if the giveback
        # completed, it may indicate that this may be a retryable error
        if ($force_giveback) {
            # -3 means that giveback should be retried but forced
            return -3;
        } else {
            # giveback should be retried
            return -1;
        }
    }

    my $sfo_total_time = $end_time - $start_time;

    # check if any timeouts were hit
    if ($max_timeout_hit eq $EventUtil::TRUE) {
        # issue an error
        $self->_set_err("SFO giveback never occurred and exceeded SFO max " .
                        "timeout $sfo_max_timeout seconds. " .
                        "($sfo_total_time seconds)");
    } elsif ($timeout_hit eq $EventUtil::TRUE) {
        # issue a warning
        &SanUtil::nate_log('-warn', "SFO giveback completed but took ($sfo_total_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout) seconds.");
    } else {
        # the giveback successfully completed within our timeout windows
        &SanUtil::nate_log('-comment', "SFO giveback took $sfo_total_time " .
                                       "seconds to complete (timeout = $sfo_timeout)");
    }

    return $sfo_total_time;
}

###############################################################################
# Method:    sfo_panicgive_abstracted()
# Objective: Issue a panic / giveback scenario against multiple HA Pairs,
#            one at a time
###############################################################################
sub sfo_panicgive_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@{$primaries}) {

        # get the partner name
        my $partner_name = $pairs->{$node_name};

        # issue a panic / giveback on this HA pair
        $self->sfo_panicgive_generic(-takeover_node => $node_name,
                                     -panic_node => $partner_name,
                                     %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if symmetrical events is enabled, then perform the event on the
        # ha partner as well
        my $symmetrical_events = $self->EventUtil::get_arg(-arg => "symmetrical_events");
        if ($symmetrical_events == 0) {
            next;
        }

        # issue a panic / giveback on this HA pair (reversed)
        $self->sfo_panicgive_generic(-takeover_node => $partner_name,
                                     -panic_node => $node_name,
                                     %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_panicgive_generic()
# Objective: Perform a panic / giveback scenario against a single HA Pair
###############################################################################
sub sfo_panicgive_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $panic_node = &ParamsMgr::get_arg(-arg => "panic_node",
                                         -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # issue a panic takeover on this ha pair
    $self->sfo_panic_generic(-takeover_node => $takeover_node,
                             -panic_node => $panic_node,
                             -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_panic_on_${panic_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    # check if user requested to terminate savecore process during panic takeovers
    my $terminate_savecore_process_subtest;
    if ($self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process")) {
        # run this in the background
        $terminate_savecore_process_subtest = $self->terminate_savecore_process(-node => $takeover_node,
                                                                                -total_exec_time => "forever",
                                                                                -async => 1,
                                                                                -use_nacl_apis => $use_nacl_apis);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    # sleep the controller down time
    my $controller_down_time = $self->EventUtil::sleep_arg(-arg => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "panic_on_${panic_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate paths
    $self->PathValidation::validate_paths_after_takeover(-id_extension => "panic_on_" . $panic_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $panic_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # issue a giveback on this HA pair
    $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                -taken_over_node => $panic_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # terminate the savecore process subtest if it exists since
    # the giveback has succeeded
    if (defined($terminate_savecore_process_subtest)) {
        if ($terminate_savecore_process_subtest->isAlive()) {
            $terminate_savecore_process_subtest->kill(message => ["-comment", "SFO Giveback has succeeded, stopping the Terminate Savecore Process subtest"]);
        }
        $terminate_savecore_process_subtest->wait_finish();
    }

    $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # validate paths once the giveback has settled
    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_" . $takeover_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_panic_generic()
# Objective: Perform a panic on a single Node
###############################################################################
sub sfo_panic_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $panic_node = &ParamsMgr::get_arg(-arg => "panic_node",
                                           -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my ($sanfiler_obj, $takeover_nacl_node_obj, $panic_nacl_node_obj);

    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $takeover_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $panic_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $panic_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $sanfiler_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a panic takeover is going to happen,
    # include data indicating which node is taking over
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_panic_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $panic_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_start_sentinel = "takeover start " . time;

    # store this takeover start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $panic_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $panic_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    if ($use_nacl_apis eq $EventUtil::TRUE) {
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            # turn on auto giveback
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            # turn off auto giveback
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        # use controller cluster
        # make sure auto giveback is turned off
        $sanfiler_obj->disable_auto_giveback();
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # make sure takeover onpanic is enabled
        $sanfiler_obj->sfo_modify(-onpanic => "true");
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # make sure takeover onpanic is enabled on the HA partner
        $sanfiler_obj->sfo_modify(-node => $panic_node,
                                  -onpanic => "true");
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Panicking '$panic_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Panicking '$panic_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # mark that the panic node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $panic_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # panic the takeover node's partner
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        try {
            &SanUtil::nate_log('-comment', "Issuing panic on node $panic_node 'sysctl debug.panic=1'");
            NACL::STask::Node->panic(command_interface => $panic_nacl_node_obj,
                                     node => $panic_node,
                                     nacltask_reboot => 0);
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $sanfiler_obj->panic_remote_node(-node => $panic_node);
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # Store the exact time panic was issued
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -taken_over_node         => $panic_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # detect that the takeover occurred
    my $takeover_time;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $takeover_time = $self->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                    -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                    -take_over_node => $takeover_node,
                                                    -taken_over_node => $panic_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $takeover_time = $sanfiler_obj->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                            -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"));
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # if the takeover time exeeded -sfo_timeout, then issue a warning
    my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
    if ($takeover_time > $sfo_timeout) {
        &SanUtil::nate_log('-warn', "SFO takeover completed but took ($takeover_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout)");
    }

    # broadcast that a panic takeover has completed
    # include data indicating which node took over
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_panic_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $panic_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_end_sentinel = "takeover end " . time;

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** panic takeover against '$panic_node' has completed",
                                                   $sfo_takeover_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store this takeover end sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $panic_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $panic_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_powergive_abstracted()
# Objective: Perform a power change / giveback scenario against multiple
#            HA pairs, one at a time.
###############################################################################
sub sfo_powergive_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@{$primaries}) {

        # get the partner name
        my $partner_name = $pairs->{$node_name};

        $self->sfo_powergive_generic(-takeover_node => $node_name,
                                     -power_change_node => $partner_name,
                                     %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if symmetrical events is enabled, then perform the event on the
        # ha partner as well
        my $symmetrical_events = $self->EventUtil::get_arg(-arg => "symmetrical_events");
        if ($symmetrical_events == 0) {
            next;
        }

        $self->sfo_powergive_generic(-takeover_node => $partner_name,
                                     -power_change_node => $node_name,
                                     %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_powergive_generic()
# Objective: Perform a power change / giveback scenario against a single
#            HA Pair
###############################################################################
sub sfo_powergive_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $power_change_node = &ParamsMgr::get_arg(-arg => "power_change_node",
                                                -required => 1);
    my $power_change_type = &ParamsMgr::get_arg(-arg => "power_change_type",
                                                -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_power_${power_change_type}_${power_change_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    $self->sfo_power_change_generic(-takeover_node => $takeover_node,
                                    -power_change_node => $power_change_node,
                                    -power_change_type => $power_change_type,
                                    -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # if power change type was 'off' then the node is still powered off,
    # it needs to be turned back on
    if ($power_change_type eq "off") {
        if (&SanUtil::power_control(-device => $power_change_node,
                                    -state => "on") == $SanUtil::FALSE) {
            $self->_set_err("Failed to change power state to 'on' for node '$power_change_node'");
            return $EventUtil::FALSE;
        }
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "power_${power_change_type}_${power_change_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "power_${power_change_type}_${power_change_node}",
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $power_change_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                -taken_over_node => $power_change_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_" . $takeover_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_power_change_abstracted()
# Objective: Perform a power change (on/off/reset) on a single Node
#            and issue givebacks if necessary.
###############################################################################
sub sfo_power_change_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $power_change_type = &ParamsMgr::get_arg(-arg => "power_change_type",
                                                -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # get all the nodes that in this event's scope
    # and that are not in a faulted state
    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode",
                                                         -faulted => $EventUtil::FALSE);
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode",
                                                        -faulted => $EventUtil::FALSE);
    }

    # use only the first node on the list, this event cannot handle multiple node HA pairs
    # unless this event were run in parallel with different node lists
    my $takeover_node = $node_names[0];

    # determine the partner name of the primary node
    my (undef, undef, $pairs) = $self->parse_node_pairs($takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $power_change_node = $pairs->{$takeover_node};

    # if doing an sfo power change then do that here
    if ($power_change_type =~ /^(off|reset)$/) {
        # validate controller before the takeover has occurred
        $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_power_${power_change_type}_on_${power_change_node}");
        if ($self->error()) {
            $self->_set_err("Failed to validate controller before takeover");
            return $EventUtil::FALSE;
        }

        # issue a power change (off/reset) on the power change node and allow the other
        # node to complete an sfo takeover
        $self->sfo_power_change_generic(-takeover_node => $takeover_node,
                                        -power_change_node => $power_change_node,
                                        -power_change_type => $power_change_type,
                                        -use_nacl_apis => $use_nacl_apis);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

        $self->EventUtil::synchronize();

        # validate controller after the takeover occurred
        $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "power_${power_change_type}_on_${power_change_node}",
                                                                        -last_timeout => $controller_down_time,
                                                                        -last_timeout_name => "controller_down_time_after_takeover");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        $self->PathValidation::validate_paths_after_takeover(-id_extension => "power_${power_change_type}_on_${power_change_node}",
                                                             -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                        reboot_node => $power_change_node},
                                                             -last_timeout => $controller_down_time,
                                                             -last_timeout_name => "controller_down_time_after_takeover");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } elsif ($power_change_type =~ /^on$/) {
        # power on this node and wait for it to come up
        $self->sfo_power_on_generic(-takeover_node => $takeover_node,
                                    -power_change_node => $power_change_node,
                                    -use_nacl_apis => $use_nacl_apis);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # now perform a giveback
        $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                    -taken_over_node => $power_change_node,
                                    -forcegive => $forcegive,
                                    -use_nacl_apis => $use_nacl_apis);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

        $self->EventUtil::synchronize();

        # validate controller after the giveback occurred
        $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                        -last_timeout => $controller_down_time,
                                                                        -last_timeout_name => "controller_down_time_after_giveback");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # Validate ports after giveback occurred
        $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                             -last_timeout      => $controller_down_time,
                                                             -last_timeout_name => "controller_down_time_after_giveback");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_" . $takeover_node,
                                                             -path_validation_rules => {validate_takeover_giveback => 1},
                                                             -last_timeout => $controller_down_time,
                                                             -last_timeout_name => "controller_down_time_after_giveback");
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # perform sram detection
        $self->SRAMDetection::check_for_new_sram_dumps();
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_power_change_generic()
# Objective: Issue a power on or power reset to trigger a takeover
#            against a single HA pair
###############################################################################
sub sfo_power_change_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $power_change_node = &ParamsMgr::get_arg(-arg => "power_change_node",
                                                -required => 1);
    my $power_change_type = &ParamsMgr::get_arg(-arg => "power_change_type",
                                                -required => 1);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # reconnect to all the nodes
    my ($takeover_node_sanfiler_obj, $takeover_nacl_node_obj);
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $takeover_node_sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $takeover_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect the clam master
    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $takeover_node_sanfiler_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect stats on nodes
    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a power off/reset takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_power_${power_change_type}_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $power_change_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_start_sentinel = "takeover start " . time;

    # store this takeover start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $power_change_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $power_change_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    if ($use_nacl_apis eq $EventUtil::TRUE) {
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            # turn on auto giveback
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            # turn off auto giveback
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        # make sure auto giveback is turned off
        $takeover_node_sanfiler_obj->disable_auto_giveback();
        if ($takeover_node_sanfiler_obj->error()) {
            $self->_set_err($takeover_node_sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # make sure takeover onreboot is enabled on both heads
        $takeover_node_sanfiler_obj->sfo_modify(-node => $power_change_node,
                                                -onreboot => "true");
        if ($takeover_node_sanfiler_obj->error()) {
            $self->_set_err($takeover_node_sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
        $takeover_node_sanfiler_obj->sfo_modify(-onreboot => "true");
        if ($takeover_node_sanfiler_obj->error()) {
            $self->_set_err($takeover_node_sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing 'power $power_change_type' on '$power_change_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing 'power $power_change_type' on '$power_change_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # synchronize before performing the action
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # mark that the power change node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $power_change_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # SanUtil::power_control expects the type to be 'reboot' not 'reset'
    my $type;
    if ($power_change_type eq "reset") {
        $type = "reboot";
    } else {
        $type = $power_change_type;
    }

    # perform the power change
    if (&SanUtil::power_control(-device => $power_change_node,
                                -state => $type) == $SanUtil::FALSE) {
        $self->_set_err("Failed to change power state to '$power_change_type' for node '$power_change_node'");
        return $EventUtil::FALSE;
    }

    # Store the exact time takeover call exited.
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -taken_over_node         => $power_change_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # detect that the takeover occurred
    my $takeover_time;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $takeover_time = $self->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                    -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                    -take_over_node => $takeover_node,
                                                    -taken_over_node => $power_change_node);
        if ($self->error()) {
            return {};
        }
    } else {
        $takeover_time = $takeover_node_sanfiler_obj->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                                          -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"));
        if ($takeover_node_sanfiler_obj->error()) {
            $self->_set_err($takeover_node_sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # if the takeover time exeeded -sfo_timeout, then issue a warning
    my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
    if ($takeover_time > $sfo_timeout) {
        &SanUtil::nate_log('-warn', "SFO takeover completed but took ($takeover_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout)");
    }

    # broadcast that a power off/reset takeover completed
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_power_${power_change_type}_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $power_change_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_end_sentinel = "takeover end " . time;

    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Completed 'power $power_change_type' on '$power_change_node'",
                                                   $sfo_takeover_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store this takeover end sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $power_change_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $power_change_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_power_on_generic()
# Objective: Power on a single Node (assuming it was previously power off)
#            and detect when it is ready for an SFO giveback.
###############################################################################
sub sfo_power_on_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $power_change_node = &ParamsMgr::get_arg(-arg => "power_change_node",
                                                -required => 1);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my ($takeover_node_sanfiler_obj, $takeover_nacl_node_obj);
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $takeover_node_sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $takeover_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a power on event is going to occur
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_power_on",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $power_change_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Issuing 'power on' on '$power_change_node'"],
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # perform the power change
    if (&SanUtil::power_control(-device => $power_change_node,
                                -state => "on") == $SanUtil::FALSE) {
        $self->_set_err("Failed to change power state to 'on' for node '$power_change_node'");
        return $EventUtil::FALSE;
    }

    # detect that the power change node is now communicating over the interconnect
    # and in the waiting for giveback screen
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $self->detect_waiting_for_giveback(-take_over_node => $takeover_node,
                                           -taken_over_node => $power_change_node,
                                           -timeout => $self->EventUtil::get_arg(-arg => "sfo_detect_waiting_for_giveback_timeout"));
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $takeover_node_sanfiler_obj->detect_waiting_for_giveback(-node => $power_change_node,
                                                                 -take_over_node => $takeover_node,
                                                                 -timeout => $self->EventUtil::get_arg(-arg => "sfo_detect_waiting_for_giveback_timeout"));
        if ($takeover_node_sanfiler_obj->error()) {
            $self->_set_err($takeover_node_sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # give the takenover node some time to get settled in
    &SanUtil::nate_sleep(30);

    # broadcast that the the power change node has been powered on
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_power_on",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $power_change_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_haltgive_abstracted()
# Objective: Issue a halt / giveback scenario against multiple HA pairs,
#            one at a time
###############################################################################
sub sfo_haltgive_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@{$primaries}) {

        # get the partner name
        my $partner_name = $pairs->{$node_name};

        $self->sfo_haltgive_generic(-takeover_node => $node_name,
                                    -halt_node => $partner_name,
                                    %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if symmetrical events is enabled, then perform the event on the
        # ha partner as well
        my $symmetrical_events = $self->EventUtil::get_arg(-arg => "symmetrical_events");
        if ($symmetrical_events == 0) {
            next;
        }

        $self->sfo_haltgive_generic(-takeover_node => $partner_name,
                                    -halt_node => $node_name,
                                    %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_haltgive_generic()
# Objective: Issue a halt / giveback scenario against a single HA pair
###############################################################################
sub sfo_haltgive_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $halt_node = &ParamsMgr::get_arg(-arg => "halt_node",
                                        -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_halt_on_${halt_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    my $halt_subtest = $self->sfo_halt_generic(-takeover_node => $takeover_node,
                                               -halt_node => $halt_node,
                                               -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    # reap the halt subtest
    $self->wait_for_subtests(-subtests => [$halt_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "halt_on_${halt_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "halt_on_" . $halt_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $halt_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                -taken_over_node => $halt_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_" . $takeover_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_halt_generic()
# Objective: Issue a halt takeover against a single HA Pair
###############################################################################
sub sfo_halt_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $halt_node = &ParamsMgr::get_arg(-arg => "halt_node",
                                           -required => 1);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    my ($sanfiler_obj, $halt_sanfiler_obj, $takeover_nacl_node_obj, $halt_nacl_node_obj);

    # reconnect to all the nodes
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return {};
        }

        $halt_sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $halt_node);
        if ($self->error()) {
            return {};
        }
    }
    $takeover_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return {};
    }
    $halt_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $halt_node);
    if ($self->error()) {
        return {};
    }

    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $halt_sanfiler_obj);
    if ($self->error()) {
        return {};
    }

    # broadcast that a halt takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_halt_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $halt_node});
    if ($self->error()) {
        return {};
    }

    my $sfo_takeover_start_sentinel = "takeover start " . time;

    # store this takeover start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return {};
        }
        $self->ResMgr::set_resource_attribute(-name => $halt_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return {};
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return {};
    }
    $self->ResMgr::set_resource_attribute(-name => $halt_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return {};
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return {};
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return {};
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return {};
    }

    # make sure auto giveback is disabled before issuing the reboot/takeover
    # and takeover on reboot is enabled
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            # turn on auto giveback
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            # turn off auto giveback
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        $sanfiler_obj->disable_auto_giveback();
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return {};
        }
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Halting '$halt_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Halting '$halt_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before halting the node
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # mark that the halted node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $halt_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return {};
    }

    my $event_obj = $self->EventUtil::get_current_event_obj();
    my $run_id_prefix = $event_obj->increment_subtest_runid_counter();

    # halt the node in a background subtest
    my $subtest = subtest(sub {
                              if ($use_nacl_apis eq $EventUtil::TRUE) {
                                  # Build a transit object for the halt node
                                  my $transit_obj;
                                  &SanUtil::nate_log(-comment, "Halting node $halt_node");
                                  try {
                                      $transit_obj = NACL::Transit->new(name => "$halt_node");

                                      # halt this node
                                      $transit_obj->change_state(to => "FIRMWARE",
                                                                 inhibit_takeover => $EventUtil::FALSE);
                                  } otherwise {
                                      my $e = shift;
                                      $self->_set_err("Failed to halt node $halt_node, " . $e->text());
                                  };
                                  if ($self->error()) {
                                      $self->SubtestComm::subtest_fail();
                                      return $EventUtil::FALSE;
                                  }
                                  # block on a read call to give the takeover node ample time to takeover.
                                  # the parent subtest will tell us to continue booting the halted node
                                  # once the takeover is complete.
                                  $Parent->message_get(type => "EventMan/Block");

                                  &SanUtil::nate_log(-comment, "Issuing 'boot_ontap' on $halt_node");
                                  try {
                                      $transit_obj->change_state(to => "UP_OR_TAKEN_OVER");
                                  } otherwise {
                                      my $e = shift;
                                      $self->_set_err("Failed to boot node $halt_node, " . $e->text());
                                  };
                                  if ($self->error()) {
                                      $self->SubtestComm::subtest_fail();
                                      return $EventUtil::FALSE;
                                  }
                              } else {
                                  # halt the takeover node's partner
                                  $halt_sanfiler_obj->halt_node();
                                  if ($halt_sanfiler_obj->error()) {
                                      $self->_set_err($halt_sanfiler_obj->errmsg());
                                      $self->SubtestComm::subtest_fail();
                                      return $EventUtil::FALSE;
                                  }

                                  # block on a read call to give the takeover node ample time to takeover.
                                  # the parent subtest will tell us to continue booting the halted node
                                  # once the takeover is complete.
                                  $Parent->message_get(type => "EventMan/Block");

                                  # boot the halted node back up
                                  $halt_sanfiler_obj->boot_node();
                                  if ($halt_sanfiler_obj->error()) {
                                      $self->_set_err($halt_sanfiler_obj->errmsg());
                                      $self->SubtestComm::subtest_fail();
                                      return $EventUtil::FALSE;
                                  }
                              }

                              # complete the halt subtest
                              $self->SubtestComm::subtest_complete();
                              return $EventUtil::TRUE;
                          }, -runid => "${run_id_prefix}_halting_node_${halt_node}", -bg, "-onexit" => sub{my ($self) = @_; $self->exithandler(1);});

    # sleep a bit to give the halt subtest time to connect via the console
    &SanUtil::nate_sleep(6);

    # Store the exact time halt was issued
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -taken_over_node         => $halt_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return {};
    }

    # detect that the takeover occurred while the halt is occuring
    my $takeover_time;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $takeover_time = $self->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                    -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                    -take_over_node => $takeover_node,
                                                    -taken_over_node => $halt_node);
        if ($self->error()) {
            return {};
        }
    } else {
        $takeover_time = $sanfiler_obj->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                            -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"));
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return {};
        }
    }

    # if the takeover time exeeded -sfo_timeout, then issue a warning
    my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
    if ($takeover_time > $sfo_timeout) {
        &SanUtil::nate_log('-warn', "SFO takeover completed but took ($takeover_time) seconds, " .
                                    "which is longer than sfo_timeout ($sfo_timeout)");
    }

    # broadcast that the halt takeover completed
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_halt_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $halt_node});
    if ($self->error()) {
        return {};
    }

    # tell the halt subtest to boot ontap
    $subtest->message_put(type => "EventMan/Block",
                          message => "boot_ontap");

    my $sfo_takeover_end_sentinel = "takeover end " . time;

    # log to all resources that the halt takeover completed
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Halt takeover against '$halt_node' by '$takeover_node' has completed",
                                                   $sfo_takeover_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return {};
    }

    # store this takeover end sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return {};
        }
        $self->ResMgr::set_resource_attribute(-name => $halt_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return {};
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return {};
    }
    $self->ResMgr::set_resource_attribute(-name => $halt_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return {};
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return {};
    }

    return $subtest;
}

###############################################################################
# Method:    sfo_rebootgive_abstracted()
# Objective: Issue a reboot / giveback scenario against multiple HA pairs,
#            one at a time
###############################################################################
sub sfo_rebootgive_abstracted {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my @node_names = ();
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        (@node_names) = $self->EventUtil::get_nacl_nodes(-type => "cmode");
    } else {
        # get all cmode nodes that are relevant to this event
        (@node_names) = $self->EventUtil::get_sanfilers(-type => "cmode");
    }

    # split up all the sanfiler nodes into HA pairs
    my ($primaries, $secondaries, $pairs) = $self->parse_node_pairs(@node_names);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    foreach my $node_name (@{$primaries}) {

        # get the partner name
        my $partner_name = $pairs->{$node_name};

        $self->sfo_rebootgive_generic(-takeover_node => $node_name,
                                      -reboot_node => $partner_name,
                                      %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }

        # if symmetrical events is enabled, then perform the event on the
        # ha partner as well
        my $symmetrical_events = $self->EventUtil::get_arg(-arg => "symmetrical_events");
        if ($symmetrical_events == 0) {
            next;
        }

        $self->sfo_rebootgive_generic(-takeover_node => $partner_name,
                                      -reboot_node => $node_name,
                                      %args);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_rebootgive_generic()
# Objective: Issue a reboot / giveback scenario against a single HA pair
###############################################################################
sub sfo_rebootgive_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $reboot_node = &ParamsMgr::get_arg(-arg => "reboot_node",
                                          -required => 1);
    my $forcegive = &ParamsMgr::get_arg(-arg => "forcegive",
                                        -default => $EventUtil::FALSE);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    # validate controller before the takeover has occurred
    $self->ControllerValidation::validate_controller_before_takeover(-id_extension => "before_reboot_on_${reboot_node}");
    if ($self->error()) {
        $self->_set_err("Failed to validate controller before takeover");
        return $EventUtil::FALSE;
    }

    $self->sfo_reboot_generic(-takeover_node => $takeover_node,
                              -reboot_node => $reboot_node,
                              -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_takeover");

    $self->EventUtil::synchronize();

    # validate controller after the takeover occurred
    $self->ControllerValidation::validate_controller_after_takeover(-id_extension => "reboot_on_${reboot_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_takeover");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_takeover(-id_extension => "reboot_on_" . $reboot_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1,
                                                                                    reboot_node => $reboot_node},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->sfo_giveback_generic(-giveback_node => $takeover_node,
                                -taken_over_node => $reboot_node,
                                -forcegive => $forcegive,
                                -use_nacl_apis => $use_nacl_apis);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $controller_down_time = $self->EventUtil::sleep_arg(-arg  => "controller_down_time_after_giveback");

    $self->EventUtil::synchronize();

    # validate controller after the giveback occurred
    $self->ControllerValidation::validate_controller_after_giveback(-id_extension => "giveback_on_${takeover_node}",
                                                                    -last_timeout => $controller_down_time,
                                                                    -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # Validate ports after giveback occurred
    $self->PortValidation::validate_ports_after_giveback(-id_extension      => "giveback_on_${takeover_node}",
                                                         -last_timeout      => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->PathValidation::validate_paths_after_giveback(-id_extension => "giveback_on_" . $takeover_node,
                                                         -path_validation_rules => {validate_takeover_giveback => 1},
                                                         -last_timeout => $controller_down_time,
                                                         -last_timeout_name => "controller_down_time_after_giveback");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # perform sram detection
    $self->SRAMDetection::check_for_new_sram_dumps();
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    sfo_reboot_generic()
# Objective: Issue a reboot / takeover scenario against a single HA Pair
###############################################################################
sub sfo_reboot_generic {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);

    my $takeover_node = &ParamsMgr::get_arg(-arg => "takeover_node",
                                            -required => 1);
    my $reboot_node = &ParamsMgr::get_arg(-arg => "reboot_node",
                                           -required => 1);
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my ($sanfiler_obj, $takeover_nacl_node_obj, $reboot_nacl_node_obj);

    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $takeover_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $takeover_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $takeover_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $reboot_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $reboot_node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $clam_master = $self->ClamValidationUtil::collect_clam_master(-sanfilerCm => $sanfiler_obj);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # broadcast that a reboot takeover is going to happen,
    $self->SubtestComm::log_event_message(-event_message => "starting_sfo_reboot_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $reboot_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_start_sentinel = "takeover start " . time;

    # store this takeover start sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $reboot_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_start_sentinel",
                                              -value => $sfo_takeover_start_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $reboot_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_start_sentinel",
                                          -value => $sfo_takeover_start_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    $self->SysStats::collect_sysstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->SysStats::collect_memstats_on_controllers(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # collect cluster statistics
    $self->SysStats::collect_cluster_statistics(-type => "cmode");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    if ($use_nacl_apis eq $EventUtil::TRUE) {
        if ($self->EventUtil::get_arg(-arg => "sfo_allow_auto_giveback") eq $EventUtil::TRUE) {
            # turn on auto giveback
            $self->toggle_auto_giveback(-state => "enable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        } else {
            # turn off auto giveback
            $self->toggle_auto_giveback(-state => "disable");
            if ($self->error()) {
                return $EventUtil::FALSE;
            }
        }
    } else {
        $sanfiler_obj->disable_auto_giveback();
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # make sure takeover onreboot is enabled on this node
        # make sure takeover onreboot is enabled on the HA partner node
        $sanfiler_obj->sfo_modify(-onreboot => "true",
                                  -node => "${takeover_node},${reboot_node}");
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # syslog to the clients asynchronously (in the background)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Rebooting '$reboot_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "clients",
                                      -quiet => $EventUtil::TRUE,
                                      -async => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syslog to the controllers (in the foreground)
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Rebooting '$reboot_node'",
                                                   $sfo_takeover_start_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # syncrhonize before performing the action.
    $self->EventUtil::synchronize_and_sleep(-arg => 'sfo_parallel_delay_time');

    # mark that the rebooted node is now in a faulted state
    $self->ResMgr::set_faulted_resource(-name => $reboot_node,
                                        -type => "nacl_nodes_cm",
                                        -propagate => $EventUtil::TRUE);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # reboot the takeover node's HA partner
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        try {
           NACL::C::Node->reboot(command_interface => $takeover_nacl_node_obj,
                                 node => $reboot_node,
                                 'inhibit-takeover' => "false");
        } otherwise {
            my $e = shift;
            $self->_set_err("Exception caught: " . $e->text());
        };
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $sanfiler_obj->system_node_reboot(-node => $reboot_node,
                                          '-inhibit-takeover' => "false");
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }
    }

    # Store the exact time panic was issued
    my $sfo_takeover_start_time = time;
    my $cv_subtest = $self->ClamValidationUtil::validate_clam_after_takeover(-original_master         => $clam_master,
                                                                             -taken_over_node         => $reboot_node,
                                                                             -sfo_takeover_start_time => $sfo_takeover_start_time);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # detect that the takeover occurred
    my $takeover_time;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $takeover_time = $self->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                    -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"),
                                                    -take_over_node => $takeover_node,
                                                    -taken_over_node => $reboot_node);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    } else {
        $takeover_time = $sanfiler_obj->detect_sfo_takeover(-sfo_timeout => $self->EventUtil::get_arg(-arg => "sfo_timeout"),
                                                            -sfo_max_timeout => $self->EventUtil::get_arg(-arg => "sfo_max_timeout"));
        if ($sanfiler_obj->error()) {
            $self->_set_err($sanfiler_obj->errmsg());
            return $EventUtil::FALSE;
        }

        # if the takeover time exeeded -sfo_timeout, then issue a warning
        my $sfo_timeout = $self->EventUtil::get_arg(-arg => "sfo_timeout");
        if ($takeover_time > $sfo_timeout) {
            &SanUtil::nate_log('-warn', "SFO takeover completed but took ($takeover_time) seconds, " .
                                   "which is longer than sfo_timeout ($sfo_timeout)");
        }
    }

    # broadcast that the reboot takeover completed
    $self->SubtestComm::log_event_message(-event_message => "completed_sfo_reboot_takeover",
                                          -data => {take_over_node => $takeover_node,
                                                    taken_over_node => $reboot_node});
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    my $sfo_takeover_end_sentinel = "takeover end " . time;

    # log to all resources that the reboot takeover completed
    $self->SysLogging::syslog_msg_all(-message => ["***EventMan*** Reboot takeover on '$reboot_node' has completed",
                                                   $sfo_takeover_end_sentinel],
                                      -type => "controllers");
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # store this takeover end sentinel on both of the node resources as an attribute
    if ($use_nacl_apis eq $EventUtil::FALSE) {
        $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
        $self->ResMgr::set_resource_attribute(-name => $reboot_node,
                                              -type => "sanfilersCm",
                                              -attribute => "takeover_end_sentinel",
                                              -value => $sfo_takeover_end_sentinel);
        if ($self->error()) {
            return $EventUtil::FALSE;
        }
    }
    $self->ResMgr::set_resource_attribute(-name => $takeover_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    $self->ResMgr::set_resource_attribute(-name => $reboot_node,
                                          -type => "nacl_nodes_cm",
                                          -attribute => "takeover_end_sentinel",
                                          -value => $sfo_takeover_end_sentinel);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    # wait for all the clam validation subtest to complete
    $self->ClamValidationUtil::reap_clamvalidation_subtest(-subtests => [$cv_subtest]);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }

    return $EventUtil::TRUE;
}

###############################################################################
# Method:    terminate_savecore_process
# Objective: terminate the savecore process on a node and its ha partner
###############################################################################
sub terminate_savecore_process {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node = &ParamsMgr::get_arg(-arg => "node",
                                   -required => 1);
    my $total_exec_time = &ParamsMgr::get_arg(-arg => "total_exec_time",
                                              -default => "forever");
    my $async = &ParamsMgr::get_arg(-arg => "async",
                                    -default => "0");
    my $use_nacl_apis = &ParamsMgr::get_arg(-arg => "use_nacl_apis",
                                            -default => $EventUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }
    my $subtest;
    if ($use_nacl_apis eq $EventUtil::TRUE) {
        $subtest = $self->EventUtil::background_execute(
            -runid => "terminate_savecore_process",
            -total_exec_time => $total_exec_time,
            -execute_frequency => $self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process_frequency"),
            -async => $async,
            -initial_sleep_time => $self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process_init_sleep_time"),
            -exec => sub {
                my $rc;
                try {
                    $rc = $self->_terminate_savecore_process_helper(%args);
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Exception caught: " . $e->text());
                    $rc = $EventUtil::FALSE;
                };
                return $rc;
            },
            -complain_exec =>
                sub {
                    if ($self->error()) {
                        return $EventUtil::FALSE;
                    }
                    return $EventUtil::TRUE;
                });
        if ($self->error()) {
            return {};
        }
    } else {
        # invoke savecore removal every 60 seconds in the background
        # while waiting controller down time as our total time
        my $sanfiler_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $node);
        if ($self->error()) {
            return {};
        }
        # run this in the background
        $subtest = $self->EventUtil::background_execute(
            -runid => "terminate_savecore_process",
            -total_exec_time => $total_exec_time,
            -execute_frequency => $self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process_frequency"),
            -async => $async,
            -initial_sleep_time => $self->EventUtil::get_arg(-arg => "sfo_terminate_savecore_process_init_sleep_time"),
            -init => sub { return $sanfiler_obj->reconnect(-disconnect_existing_conn => $EventUtil::FALSE); },
            -exec => sub { return $sanfiler_obj->terminate_savecore_process(); },
            -complain_exec =>
                sub {
                    if ($sanfiler_obj->error()) {
                        $self->_set_err($sanfiler_obj->errmsg());
                        return $EventUtil::FALSE;
                    }
                    return $EventUtil::TRUE;
                });
        if ($self->error()) {
            return {};
        }
    }

    return $subtest;
}

###############################################################################
# Method:    get_sfo_aggr_outage_times()
# Objective: Determine sfo aggr outages
###############################################################################
sub get_sfo_aggr_outage_times {
    my ($self, %args) = @_;

    # parse the arguments
    &ParamsMgr::process_args(%args);
    my $giveback_node = &ParamsMgr::get_arg(-arg => "giveback_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    my $start_sentinel = &ParamsMgr::get_arg(-arg => "start_sentinel",
                                             -required => 1);
    my $end_sentinel = &ParamsMgr::get_arg(-arg => "end_sentinel",
                                           -required => 1);
    my $display_label = &ParamsMgr::get_arg(-arg => "display_label",
                                            -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # dont run this collection if this is disabled
    my $collect_sfo_aggr_outage_times = $self->EventUtil::get_arg(-arg => "collect_sfo_aggr_outage_times");
    if ($collect_sfo_aggr_outage_times != 1) {
        &SanUtil::nate_log(-comment, "Not collecting SFO Aggr outages: ('collect_sfo_aggr_outage_times' is disabled)");
        return {};
    }

    my $giveback_node_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $giveback_node);
    if ($self->error()) {
        $self->_set_err("Failed to obtain SanFiler object for $giveback_node");
        return {};
    }
    my $taken_over_node_obj = $self->EventUtil::get_sanfiler_obj(-filer_name => $taken_over_node);
    if ($self->error()) {
        $self->_set_err("Failed to obtain SanFiler object for $giveback_node");
        return {};
    }

    # reconnect to the taken over node, just in case the caller did not perform a reconnect
    $taken_over_node_obj->reconnect();
    if ($taken_over_node_obj->error()) {
        $self->_set_err($taken_over_node_obj->errmsg());
        return $EventUtil::FALSE;
    }

    # drop into bsd to get the ems file
    $giveback_node_obj->set_shell_environment(-type => "bsd");
    if ($giveback_node_obj->error()) {
        $self->_set_err("Unable to switch to BSD to collect ems log files for sfo aggr outage collection:",
                        $giveback_node_obj->errmsg());
        return {};
    }

    my $verbose = $self->EventUtil::get_arg(-arg => "sfo_display_sfo_aggr_outage_ems_logs");
    my $no_verbose;
    if ($verbose eq "0") {
        $no_verbose = $EventUtil::TRUE;
    } else {
        $no_verbose = $EventUtil::FALSE;
    }

    # cat the ems file silently: don't output to logs, use awk to restrain the data between the
    # start and end sentinel, otherwise the command may take a long time to process
    my @giveback_node_output = $giveback_node_obj->execute(-cmd => 'cat /mroot/etc/log/ems | awk \'BEGIN {x=0} {' .
                                                                   'if ($0~"' . $start_sentinel . '") {x=1} ' .
                                                                   'if (x==1) {print $0} ' .
                                                                   'if ($0~"' . $end_sentinel. '") {x=0}}\'',
                                                           -no_verbose => $no_verbose,
                                                           -ignore_error_check => 1,
                                                           -raw_data => 1,
                                                           -timeout => 200);
    if ($giveback_node_obj->error()) {
        $self->_set_err("Unable to obtain the ems logs for sfo aggr outage collection:",
                        $giveback_node_obj->errmsg());
        return {};
    }

    # restore the shell environment back to ngsh
    $giveback_node_obj->set_shell_environment(-type => "ngsh");
    if ($giveback_node_obj->error()) {
        $self->_set_err("Unable to switch back to NGSH after collecting sfo aggr outage data:",
                        $giveback_node_obj->errmsg());
        return {};
    }

    # drop into bsd to get the ems file
    $taken_over_node_obj->set_shell_environment(-type => "bsd");
    if ($taken_over_node_obj->error()) {
        $self->_set_err("Unable to switch to BSD to collect ems log files for sfo aggr outage collection:",
                        $taken_over_node_obj->errmsg());
        return {};
    }

    # cat the ems file silently: don't output to logs, use awk to restrain the data between the
    # start and end sentinel, otherwise the command may take a long time to process
    my @taken_over_node_output = $taken_over_node_obj->execute(-cmd => 'cat /mroot/etc/log/ems | awk \'BEGIN {x=0} {' .
                                                                       'if ($0~"' . $start_sentinel . '") {x=1} ' .
                                                                       'if (x==1) {print $0} ' .
                                                                       'if ($0~"' . $end_sentinel. '") {x=0}}\'',
                                                           -no_verbose => $no_verbose,
                                                           -ignore_error_check => 1,
                                                           -raw_data => 1,
                                                           -timeout => 200);
    if ($taken_over_node_obj->error()) {
        $self->_set_err("Unable to obtain the ems logs for sfo aggr outage collection:",
                        $taken_over_node_obj->errmsg());
        return {};
    }

    # restore the shell environment back to ngsh
    $taken_over_node_obj->set_shell_environment(-type => "ngsh");
    if ($taken_over_node_obj->error()) {
        $self->_set_err("Unable to switch back to NGSH after collecting sfo aggr outage data:",
                        $taken_over_node_obj->errmsg());
        return {};
    }

    # parse the ems files on both nodes
    my @giveback_node_ems_file = split(/\n/, $giveback_node_output[0]);
    my @taken_over_node_ems_file = split(/\n/, $taken_over_node_output[0]);

    # Parse the ems file for aggr outages
    my $sfo_aggr_outages = {};
    tie(%{$sfo_aggr_outages}, "Tie::IxHash");
    my $prevline = "";
    my $nextline;
    my $nextnextline;
    my $nextnextnextline;

    # keep reading the ems file until it's empty
    while (scalar(@giveback_node_ems_file)) {
        my $line = shift(@giveback_node_ems_file);

        if ($line =~ /ha_giveback_aggr_time_1/) {
            # parse out the aggr name
            $nextline = shift(@giveback_node_ems_file);
            my ($aggrname) = ($nextline =~ /aggr=\"(.+?)\"/);
            if (not exists($sfo_aggr_outages->{$aggrname})) {
                $sfo_aggr_outages->{$aggrname} = {};
                tie(%{$sfo_aggr_outages->{$aggrname}}, "Tie::IxHash");
            }
            $nextnextline = shift(@giveback_node_ems_file);
            my ($sendhome_shutdown_end_timestamp) = ($prevline =~ /t=\"(.+?)\"/);
            push(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_end_timestamp'}}, $sendhome_shutdown_end_timestamp);
        }
        elsif ($line =~ /ha_giveback_sysCommit_1/) {
            # determine the sendhome shutdown duration
            $nextline = shift(@giveback_node_ems_file);
            # skip this if its not the wafl subsystem
            next if ($nextline !~ /subsys=\"wafl\"/i);
            $nextnextline =  shift(@giveback_node_ems_file);
            $nextnextnextline = shift(@giveback_node_ems_file);
            my ($aggrname) = ($nextnextnextline =~ /aggr=\"(.+?)\"/);
            if (not exists($sfo_aggr_outages->{$aggrname})) {
                $sfo_aggr_outages->{$aggrname} = {};
                tie(%{$sfo_aggr_outages->{$aggrname}}, "Tie::IxHash");
            }
            my ($sendhome_shutdown_start_timestamp) = ($prevline =~ /t=\"(.+?)\"/);
            my ($sendhome_wafl_duration_time) = ($nextnextline =~ /time=\"(.+?)\"/);
            $sendhome_wafl_duration_time /= 1000;
            push(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_start_timestamp'}}, $sendhome_shutdown_start_timestamp);
            push(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_wafl_duration'}}, $sendhome_wafl_duration_time);
        }
        elsif ($line =~ /ha_sfo_giveback_aggrStart_1/) {
            # parse when this aggr started the sendhome

            # parse out the aggr name
            $nextline = shift(@giveback_node_ems_file);
            my ($aggrname) = ($nextline =~ /aggrName=\"(.+?)\"/);
            if (not exists($sfo_aggr_outages->{$aggrname})) {
                $sfo_aggr_outages->{$aggrname} = {};
                tie(%{$sfo_aggr_outages->{$aggrname}}, "Tie::IxHash");
            }

            # parse out the time the sfo giveback start time for this aggr occurred
            my ($aggr_sfo_start_time_readable) = ($prevline     =~ /d=\"(.+?)\"/);
            my ($aggr_sfo_start_time_sec)      = ($prevline     =~ /t=\"(\d+)\"/);

            push(@{$sfo_aggr_outages->{$aggrname}->{'aggr_sfo_giveback_start_timestamp_readable'}}, $aggr_sfo_start_time_readable);
            push(@{$sfo_aggr_outages->{$aggrname}->{'aggr_sfo_giveback_start_timestamp'}}, $aggr_sfo_start_time_sec);
        } elsif ($line =~ /ha_sfo_giveback_aggrDone_1/) {
            # parse when this aggr finished the sendhome

            # parse out the aggr name
            $nextline = shift(@giveback_node_ems_file);
            my ($aggrname) = ($nextline =~ /aggrName=\"(.+?)\"/);
            if (not exists($sfo_aggr_outages->{$aggrname})) {
                $sfo_aggr_outages->{$aggrname} = {};
                tie(%{$sfo_aggr_outages->{$aggrname}}, "Tie::IxHash");
            }

            # parse out the time the sfo giveback endtime for this aggr occurred
            my ($aggr_sfo_end_time_readable) = ($prevline     =~ /d=\"(.+?)\"/);
            my ($aggr_sfo_end_time_sec)      = ($prevline     =~ /t=\"(\d+)\"/);

            push(@{$sfo_aggr_outages->{$aggrname}->{'aggr_sfo_giveback_end_timestamp_readable'}}, $aggr_sfo_end_time_readable);
            push(@{$sfo_aggr_outages->{$aggrname}->{'aggr_sfo_giveback_end_timestamp'}}, $aggr_sfo_end_time_sec);
        }

        $prevline = $line;
    }

    $prevline = "";

    # keep reading the ems file until it's empty
    while (scalar(@taken_over_node_ems_file)) {
        my $line = shift(@taken_over_node_ems_file);

        if ($line =~ /wafl_aggregate_online_phases_1/) {
            # parse out the aggr name
            $nextline = shift(@taken_over_node_ems_file);
            my ($aggrname) = ($nextline =~ /aggrName=\"(.+?)\"/);
            if (not exists($sfo_aggr_outages->{$aggrname})) {
                $sfo_aggr_outages->{$aggrname} = {};
                tie(%{$sfo_aggr_outages->{$aggrname}}, "Tie::IxHash");
            }

            my ($timestamp) = ($prevline =~ /t=\"(.+?)\"/);
            push(@{$sfo_aggr_outages->{$aggrname}->{'wafl_aggregate_online_phases_1_timestamp'}}, $timestamp);
        }

        $prevline = $line;
    }

    # calculate the sfo outages for all the aggrs we collected data for
    foreach my $aggrname (keys %{$sfo_aggr_outages}) {
        # if we don't have any of the values, then we cannot perform an accurate calculation for this aggr's outage
        if (ref($sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_start_timestamp'}) ne "ARRAY" or
            ref($sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_end_timestamp'}) ne "ARRAY" or
            ref($sfo_aggr_outages->{$aggrname}->{'sendhome_wafl_duration'}) ne "ARRAY" or
            ref($sfo_aggr_outages->{$aggrname}->{'wafl_aggregate_online_phases_1_timestamp'}) ne "ARRAY") {
            &SanUtil::nate_log('-comment', "Unable to determine SFO Aggregate outage for aggregate '$aggrname'");
            next;
        }

        for (my $i = 0; $i < scalar(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_start_timestamp'}}); $i++) {
            my $sendhome_start_timestamp = $sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_start_timestamp'}->[$i];
            my $sendhome_end_timestamp = $sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_end_timestamp'}->[$i];
            my $wafl_duration = $sfo_aggr_outages->{$aggrname}->{'sendhome_wafl_duration'}->[$i];
            my $wafl_aggr_online_timestamp = $sfo_aggr_outages->{$aggrname}->{'wafl_aggregate_online_phases_1_timestamp'}->[$i];

            # if we don't have any of the values, then we cannot perform an accurate calculation for this aggr's outage
            if (not length($sendhome_start_timestamp) or
                not length($sendhome_end_timestamp) or
                not length($wafl_duration) or
                not length($wafl_aggr_online_timestamp)) {
                &SanUtil::nate_log('-comment', "Unable to determine SFO Aggregate outage for aggregate '$aggrname'");
                next;
            }

            # calculate the outage for this aggr
            my $sendhome_shutdown_duration = $sendhome_end_timestamp - ($sendhome_start_timestamp - $wafl_duration);
            $sendhome_shutdown_duration =~ s/^(\d+\.\d{2}).*$/$1/;
            my $sendhome_online_duration = $wafl_aggr_online_timestamp - $sendhome_end_timestamp;
            my $total_outage = $sendhome_shutdown_duration + $sendhome_online_duration;
            push(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_shutdown_duration'}}, $sendhome_shutdown_duration);
            push(@{$sfo_aggr_outages->{$aggrname}->{'sendhome_online_duration'}}, $sendhome_online_duration);
            push(@{$sfo_aggr_outages->{$aggrname}->{'total_outage'}}, $total_outage);

            # log outages so CLV can parse it
            &SanUtil::nate_log('-comment', ":CLV: SFO Aggregate Outages [node=${taken_over_node}, aggr=$aggrname, sendhome_shutdown_duration=${sendhome_shutdown_duration}, sendhome_online_duration=${sendhome_online_duration}, total_outage=${total_outage}]");
        }
    }

    # Display the aggr outages
    &SanUtil::nate_log('-comment', "SFO Aggregate Outages '$display_label':\n" . Dumper($sfo_aggr_outages));

    # broadcast a message with the sfo aggr outage times
    $self->SubtestComm::log_event_message(-event_message => "sfo_aggr_outage_times",
                                          -data => $sfo_aggr_outages);

    return $sfo_aggr_outages;
}

###############################################################################
# Method:    check_for_auto_sfo_giveback()
# Objective: Return true or false, depending on whether a manual giveback
#            is required.
###############################################################################
sub check_for_auto_sfo_giveback {
    $Log->enter();

    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $giveback_node = &ParamsMgr::get_arg(-arg => "giveback_node",
                                            -required => 1);
    my $taken_over_node = &ParamsMgr::get_arg(-arg => "taken_over_node",
                                              -required => 1);
    if (&ParamsMgr::error()) {
        $Log->exit();
        EventMan::Exceptions::InvalidParameter->throw(&ParamsMgr::get_error);
        return $EventUtil::FALSE;
    }

    my $giveback_nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $giveback_node);
    if ($self->error()) {
        $Log->exit();
        NATE::BaseException->throw("Error obtaining NACL Node object for $giveback_node");
        return $EventUtil::FALSE;
    }

    my $perform_giveback = $EventUtil::TRUE;
    my $check_again = 0;

    while (1) {

        # if this is our second time checking for auto giveback, then disable the check again flag
        # to prevent infinite loops
        if ($check_again == 1) {
            $check_again = -1;
        }

        try {
            # if the giveback node sees that the auto giveback is within 30 seconds from occurring, then abort.
            $self->wait_for_sfo_state(-command_interface => $giveback_nacl_node_obj,
                                      -sfo_state_hash => { "$giveback_node" => { states => ["takeover"],
                                                                                 abort_state_descriptions => [qr/Auto giveback will be initiated in \d+ seconds/i,
                                                                                                              qr/Previous giveback failed in module/i,
                                                                                                              qr/Auto giveback deferred/i] },
                                                           "$taken_over_node" => { state_descriptions => [qr/Waiting for giveback/i] } } ,
                                      -check_once => 1);
            # No exception was thrown, this means the taken over node is
            # waiting for a giveback
            $Log->comment("No auto giveback has been detected, will perform a manual giveback.");
            $perform_giveback = $EventUtil::TRUE;
        } catch EventMan::Exceptions::UnexpectedSFOState with {
            my $e = shift;
            my @sf_states = $e->get_sf_states();

            # if we see Previous giveback failed in the sfo state, then we need to perform the giveback
            # else don't perform the giveback
            foreach my $sf_state (@sf_states) {
                # only look at the sf state for the giveback node
                next if ($sf_state->node() ne $giveback_node);

                if ($sf_state->state_description() =~ /Previous giveback failed/i) {
                    # the autogiveback may have been cancelled for some reason
                    # this means we have to perform a giveback
                    $Log->comment("Detected a failed auto giveback, will perform a manual giveback.");
                    $perform_giveback = $EventUtil::TRUE;
                    last;
                } else {
                    # we may be in the middle of an auto giveback state, don't perform the giveback
                    $Log->comment("Detected that an auto giveback occurred or is in process, will not perform a manual giveback.");
                    $perform_giveback = $EventUtil::FALSE;
                    last;
                }
            }
        } catch EventMan::Exceptions::EncounteredSFOAbortState with {
            my $e = shift;
            if ($e->get_abort_state() =~ /Auto giveback will be initiated/i) {
                # check if the autogiveback threshold has been met
                my $autogiveback_threshold = $self->EventUtil::get_arg(-arg => "sfo_auto_giveback_threshold");
                if ($e->get_node_sf_output()->state_description() =~ /Auto giveback will be initiated in (\d+) seconds/i) {
                    my $autogiveback_time = $1;
                    if ($autogiveback_time <= $autogiveback_threshold) {
                        # we are within the threshold for the autogiveback, dont perform the giveback
                        $Log->comment("Detected that an auto giveback will occur in the next " .
                                      "$autogiveback_time seconds. Will not perform a manual giveback.");
                        $perform_giveback = $EventUtil::FALSE;
                    } else {
                        # we are beyond the autogiveback threshold, do the manual giveback
                        $Log->comment("Detected an auto giveback that is not within the autogiveback threshold " .
                                      "of $autogiveback_threshold seconds, will perform a manual giveback.");
                        $perform_giveback = $EventUtil::TRUE;
                    }
                } else {
                    # code shouldn't get here, but if it does, manually perform the giveback
                    $Log->comment("Detected an imminent auto giveback, but could not determine when it will happen, ".
                                  "will perform a manual giveback.");
                    $perform_giveback = $EventUtil::TRUE;
                }
            } elsif ($e->get_abort_state() =~ /Previous giveback failed in module/) {
                # the autogiveback did run, but it failed. ONTAP will perform a couple auto givebacks
                # before giving up

                # if this is the first time we detect the auto giveback failure, we need to wait a bit
                # because the auto giveback might retry. Allow the retry to occur, if it is successful then
                # don't intervene, if it fails again then we need to step in and perform the manual giveback
                if ($check_again == 0) {
                    $Log->comment("Detected that an auto giveback failed, will give the system a chance to retry the auto giveback.");
                    $check_again = 1;
                    $self->EventUtil::sleep_arg(-arg => "sfo_failed_auto_giveback_sleep_time");
                } else {
                    $Log->comment("Detected that the auto giveback retry failed or never happened, will perform a manual giveback.");
                    $perform_giveback = $EventUtil::TRUE;
                }
            } elsif ($e->get_abort_state() =~ /Auto giveback deferred/) {
                # the autogiveback did run, but it failed , we need to perform the giveback
                $Log->comment("Detected that an auto giveback was deferred, will perform a manual giveback.");
                $perform_giveback = $EventUtil::TRUE;
            } else {
                # unknown abort state
                $Log->exit();
                $e->throw();
            }
        };

        # if the auto giveback failed then and we were told to check again, do so
        redo if ($check_again == 1);

        last;
    }

    $Log->exit();
    return $perform_giveback;
}

###############################################################################
# Method:    wait_for_sfo_state()
# Objective: Wait until the specified SFO state(s) have been hit.
###############################################################################
sub wait_for_sfo_state {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $ci = &ParamsMgr::get_arg(-arg => "command_interface",
                                 -required => $EventUtil::TRUE);
    my $sfo_state_hash = &ParamsMgr::get_arg(-arg => "sfo_state_hash",
                                             -required => 1,
                                             -ref_type => "HASH");
    my $method_timeout = &ParamsMgr::get_arg(-arg => "method_timeout",
                                             -default => 180);
    my $poll_interval = &ParamsMgr::get_arg(-arg => "poll_interval",
                                            -default => 1);
    my $check_once = &ParamsMgr::get_arg(-arg => "check_once",
                                         -default => $EventUtil::FALSE);
    my $abort_states =  &ParamsMgr::get_arg(-arg => "abort_states",
                                            -ref_type => "ARRAY",
                                            -default => []);
    my $sfo_type = &ParamsMgr::get_arg(-arg => "sfo_type",
                                       -default => "");
    my $sfo_show_nodes = &ParamsMgr::get_arg(-arg => "sfo_show_nodes",
                                             -default => [],
                                             -ref_type => "ARRAY");
    my $abort_state_descriptions =  &ParamsMgr::get_arg(-arg => "abort_state_descriptions",
                                                        -ref_type => "ARRAY",
                                                        -default => []);
    if (&ParamsMgr::error()) {
        EventMan::Exceptions::InvalidParameter->throw(&ParamsMgr::get_error);
        return $EventUtil::FALSE;
    }

    # verify parameters
    foreach my $nodename (keys %{$sfo_state_hash}) {
        my $node_states = exists($sfo_state_hash->{$nodename}->{states}) ? $sfo_state_hash->{$nodename}->{states} : [];
        my $node_state_descriptions = exists($sfo_state_hash->{$nodename}->{state_descriptions}) ? $sfo_state_hash->{$nodename}->{state_descriptions} : [];

        # verify that the node state arrays for this node are valid
        if (ref($node_states) ne "ARRAY" or ref($node_state_descriptions) ne "ARRAY") {
            EventMan::Exceptions::InvalidParameter->throw("$nodename does not have a valid node state or node state description array");
            return $EventUtil::FALSE;
        }
        if (not scalar(@{$node_states}) and not scalar(@{$node_state_descriptions})) {
            EventMan::Exceptions::InvalidParameter->throw("$nodename has empty node state and node state description arrays");
            return $EventUtil::FALSE;
        }

        # if this node has its own abort states, then ignore the global states
        unless ($sfo_state_hash->{$nodename}->{abort_states} and
            ref($sfo_state_hash->{$nodename}->{abort_states}) eq "ARRAY") {
            @{$sfo_state_hash->{$nodename}->{abort_states}} = @{$abort_states};
        }
        unless ($sfo_state_hash->{$nodename}->{abort_state_descriptions} and
            ref($sfo_state_hash->{$nodename}->{abort_state_descriptions}) eq "ARRAY") {
            @{$sfo_state_hash->{$nodename}->{abort_state_descriptions}} = @{$abort_state_descriptions};
        }
    }

    # calculate when this method will timeout
    my $timeout_time = time() + $method_timeout;
    my $method_has_timed_out = 0;

    # check the state of every node.
    my $total_nodes_reached_state = 0;
    my $abort_state_hit = 0;
    my $i = 0;

    my @sf_output = ();
    while (1) {
        # if all nodes that we are checking hit their states, then we are done
        if ($total_nodes_reached_state >= scalar(keys %{$sfo_state_hash})) {
            last;
        } elsif ($i != 0 and $check_once == 1) {
            # if we have gone through one iteration and not all nodes have hit their expected states
            # abort if told to check once.
            EventMan::Exceptions::UnexpectedSFOState->throw(text => "Not all nodes hit their expected SFO states.",
                                                            sf_output => \@sf_output);
            return $EventUtil::FALSE;
        }
        $i++;

        # if we timed out, bail out
        if (time() > $timeout_time) {
            EventMan::Exceptions::TimedOutWaitingForSFOState->throw("Timed out while waiting for SFO states");
            return $EventUtil::FALSE;
        }

        # check every state of each node
        @sf_output = NACL::CS::StorageFailover->fetch(command_interface => $ci,
                                                      apiset_must       => {interface => 'CLI'},
                                                      filter => { node => join (",", keys %{$sfo_state_hash}) },
                                                      requested_fields => ["node", "partner-name", "state",
                                                                           "reason", "partner-state", "partner-reason",
                                                                           "state-description"]);

        my %sfo_show_nodes_filter = ();
        if (scalar(@{$sfo_show_nodes})) {
            %sfo_show_nodes_filter = (filter => {node => join (",", @{$sfo_show_nodes})});
        } else {
            %sfo_show_nodes_filter = (filter => {node => join (",", keys %{$sfo_state_hash})});
        }

        if ($sfo_type =~ /takeover/) {
            try {
                NACL::C::StorageFailover->show_takeover(command_interface => $ci,
                                                        %sfo_show_nodes_filter);
            } otherwise {
                # ignore exceptions
            };
        }

        if ($sfo_type =~ /giveback/) {
            try {
                NACL::C::StorageFailover->show_giveback(command_interface => $ci,
                                                        %sfo_show_nodes_filter);
            } otherwise {
                # ignore exceptions
            };
        }

        # check the respective states of every node in the hash
        foreach my $nodename (keys %{$sfo_state_hash}) {
            my $node_states = exists($sfo_state_hash->{$nodename}->{states}) ? $sfo_state_hash->{$nodename}->{states} : [];
            my $node_state_descriptions = exists($sfo_state_hash->{$nodename}->{state_descriptions}) ? $sfo_state_hash->{$nodename}->{state_descriptions} : [];
            my $abort_states = exists($sfo_state_hash->{$nodename}->{abort_states}) ? $sfo_state_hash->{$nodename}->{abort_states} : [];
            my $abort_state_descriptions = exists($sfo_state_hash->{$nodename}->{abort_state_descriptions}) ? $sfo_state_hash->{$nodename}->{abort_state_descriptions} : [];

            # stop checking this node if we reached its state
            next if ($sfo_state_hash->{$nodename}->{state_reached});

            # if requested to not check this node, then indicate that we reached its state
            if (grep(/nocheck/, @{$node_states})) {
                $sfo_state_hash->{$nodename}->{state_reached} = 1;
                $total_nodes_reached_state++;
                next;
            }

            # find this node in the sf output
            my ($node_sf_output) = grep {
                $_->node() eq $nodename;
            } @sf_output;

            # scan the abort states for this node
            foreach my $abort_state (@{$abort_states}) {
                if ((ref($abort_state) eq "Regexp" and $node_sf_output->state() =~ $abort_state) or
                    $node_sf_output->state() eq $abort_state) {
                    # we need to abort, this node hit an abort state
                    EventMan::Exceptions::EncounteredSFOAbortState->throw(text => "$nodename has hit the abort state '$abort_state'",
                                                                          nodename => $nodename,
                                                                          node_sf_output => $node_sf_output,
                                                                          abort_state => $abort_state);
                }
            }

            # scan the abort state descriptions for this node
            foreach my $abort_state (@{$abort_state_descriptions}) {
                if ((ref($abort_state) eq "Regexp" and $node_sf_output->state_description() =~ $abort_state) or
                    $node_sf_output->state_description() eq $abort_state) {
                    # we need to abort, this node hit an abort state
                    EventMan::Exceptions::EncounteredSFOAbortState->throw(text => "$nodename has hit the abort state description '$abort_state'",
                                                                          nodename => $nodename,
                                                                          node_sf_output => $node_sf_output,
                                                                          abort_state => $abort_state);
                }
            }

            # scan the expected node states for this node
            foreach my $node_state (@{$node_states}) {
                if ((ref($node_state) eq "Regexp" and $node_sf_output->state() =~ $node_state) or
                    $node_sf_output->state() eq $node_state) {
                    # we've reached the state for this node
                    # stop checking this node
                    $sfo_state_hash->{$nodename}->{state_reached} = 1;
                    $total_nodes_reached_state++;
                    # stop checking for node states on this node if we have encountered one of the states
                    last;
                }
            }

            # scan the expected node state descriptions for this node
            foreach my $node_state (@{$node_state_descriptions}) {
                if ((ref($node_state) eq "Regexp" and
                     $node_sf_output->state_description() =~ $node_state) or
                     $node_sf_output->state_description() eq $node_state) {
                    # we've reached the state for this node
                    # stop checking this node
                    $sfo_state_hash->{$nodename}->{state_reached} = 1;
                    $total_nodes_reached_state++;
                    # stop checking for node states on this node if we have encountered one of the states
                    last;
                }
            }
        }

        # sleep the poll interval before scanning the list of nodes again
        sleep($poll_interval);
    }

    return $EventUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _terminate_savecore_process_helper
# Objective: terminate the savecore process on this ha pair
###############################################################################
sub _terminate_savecore_process_helper {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node = &ParamsMgr::get_arg(-arg => "node",
                                   -required => 1);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return $EventUtil::FALSE;
    }

    my $nacl_node_obj = $self->EventUtil::get_nacl_node_obj(-node_name => $node);
    if ($self->error()) {
        return $EventUtil::FALSE;
    }
    # obtain the apiset for the dblade cli
    my ($cmd, $output, $apiset);

    $apiset = $nacl_node_obj->apiset(category => "Node",
                                     interface => "CLI",
                                     set => "CMode");

    $cmd = "run -node $node -command \"priv set diag; cf status\"";
    $output = $apiset->execute_raw_command(command => $cmd);
    if ($output =~ /is up/) {
        &SanUtil::nate_log('-comment', "The controller HA pair is not in a failover state");
        return $EventUtil::TRUE;
    }

    # initialize the _core_name attribute if it's not defined
    if (not exists($self->{_core_name})) {
        $self->{_core_name} = "";
    }

    $cmd = "run -node $node -command \"priv set diag; partner savecore -s\"";
    $output = $apiset->execute_raw_command(command => $cmd);

    if ($output =~ /Saving\s(.+)\:\s/i && $self->{_core_name} eq "") {
        $self->{_core_name} = $1;
        &SanUtil::nate_log('-comment', "Core file found " . $self->{_core_name});
        return $EventUtil::TRUE;
    } elsif ($output =~ /Saving\s(.+)\:\s/i && $self->{_core_name} ne "") {
        $output =~ /Saving\s(.+)\:\s/;
        if ($self->{_core_name} eq $1) {
            &SanUtil::nate_log('-comment', "Found automation induced core that is still saving, will attempt to kill it.");

            # Have to kill the core, and go on
            $cmd = "run -node $node -command \"priv set diag; partner savecore -k\"";
            $output = $apiset->execute_raw_command(command => $cmd);

            $cmd = "run -node $node -command \"priv set diag; partner savecore -s\"";
            $output = $apiset->execute_raw_command(command => $cmd);

            if ($output =~ /Saving\s(.+):\s/i) {
                &SanUtil::nate_log('-warn', "Core deletion failed");
                &SanUtil::nate_sleep(10);
                return $EventUtil::TRUE;
            } else {
                &SanUtil::nate_log('-comment', "Core deletion completed");
            }

        } else {
            # in this case we found a new core that was not induced by the automation, this should be a failure
            $self->_set_err("Detected a new core that was not induced by the automation: '$1'");
            return $EventUtil::FALSE;
        }
    } elsif ($output =~ /No core need|No savecore in progress/i) {
        &SanUtil::nate_log('-comment', "Did not find a core file being saved");
        return $EventUtil::TRUE;
    } elsif ($output =~ /Generating mini-core/) {
        &SanUtil::nate_log('-comment', "Found mini-core: $output");
        # Sleep 60 seconds, it usually takes 40-50 seconds to complete
        # the mini-core generation
        &SanUtil::nate_sleep(60);
        return $EventUtil::TRUE;
    } elsif ($output =~ /Preparing to save core/i) {
        &SanUtil::nate_log('-comment', "Nodes are preparing to save core");
        # Sleep 15 seconds to allow the filers to start creating the core,
        # then return 0
        &SanUtil::nate_sleep(15);
        return $EventUtil::TRUE;
    } elsif ($output =~ /Waiting for partner dump to complete/i) {
        &SanUtil::nate_log('-comment', "Nodes are waiting for partner to dump core");
        # Sleep 60 seconds to allow the filers to start creating the core,
        # then return 0
        &SanUtil::nate_sleep(60);
        return $EventUtil::TRUE;
    } else {
        &SanUtil::nate_log('-comment', "Did not find a core file being saved");
    }

    return $EventUtil::TRUE;
}

sub _convert_sanfilersCm_to_nacl_nodes_cm {
    my ($self, %args) = @_;

    my $event_obj = $args{-event_obj};
    my $eventset_obj = $args{-event_set_obj};

    # case 1: a user specified nacl_nodes_cm on the event level
    # result: do nothing

    # case 2: a user specified a non array value for nacl_nodes_cm
    # result: do nothing

    # case 3: a user specified sanfilersCm on the event level but no nacl_nodes_cm on the event level
    # result: create a nacl_nodes_cm value and copy the contents from sanfilersCm

    # case 4: a user specified sanfilersCm on the eventset level but not nacl_nodes_cm on the event/eventset level
    # result: create a nacl_nodes_cm value and copy the contents from sanfilersCm

    # case 5: a user specified sanfilersCm and nacl_nodes_cm on the eventset level
    # result: do nothing

    # case 6: a user specified sanfilersCm on the eventset level and a non array on the eventset level
    # for nacl_ndoes_cm
    # result: do nothing

    # case 7: a user did not specify sanfilersCm at the eventset level or event level
    # result: do nothing

    my $nacl_nodes_cm = $self->EventUtil::get_arg(-arg => 'nacl_nodes_cm',
                                                  -objects => [$event_obj],
                                                  -default => []);
    if (ref($nacl_nodes_cm) eq "ARRAY" and scalar(@{$nacl_nodes_cm})) {
        # case 1
        return $EventUtil::TRUE;
    }
    if (ref($nacl_nodes_cm) ne "ARRAY") {
        # case 2
        # do nothing
        return $EventUtil::TRUE;
    }

    my $sanfilersCm = $self->EventUtil::get_arg(-arg => 'sanfilersCm',
                                                -objects => [$event_obj],
                                                -default => []);
    if (ref($sanfilersCm) eq "ARRAY" and scalar(@{$sanfilersCm})) {
        # case 3
        $event_obj->set_event_parameter('nacl_nodes_cm', [@{$sanfilersCm}]);
        return $EventUtil::TRUE;
    }

    $sanfilersCm = $self->EventUtil::get_arg(-arg => 'sanfilersCm',
                                                -objects => [$eventset_obj],
                                                -default => []);
    $nacl_nodes_cm = $self->EventUtil::get_arg(-arg => 'nacl_nodes_cm',
                                               -objects => [$eventset_obj],
                                               -default => []);

    if (ref($sanfilersCm) eq "ARRAY" and scalar(@{$sanfilersCm})) {
        if (ref($nacl_nodes_cm) eq "ARRAY") {
            if (not scalar(@{$nacl_nodes_cm})) {
                # case 4
                $event_obj->set_event_parameter('nacl_nodes_cm', [@{$sanfilersCm}]);
                return $EventUtil::TRUE;
            } else {
                # case 5
                # do nothing
                return $EventUtil::TRUE;
            }
        } else {
            # case 6 do nothing
            return $EventUtil::TRUE;
        }
    } else {
        # case 7
        return $EventUtil::TRUE;
    }

    return $EventUtil::TRUE;
}

1;

=pod

=head1 Package Name

SFOEvents.pm

=head1 Author

Robert Hunter

=head1 Description

Event Module for Controller based CMode Storage Failover event scenarios.

=head1 Default Event Module Inheritance

=over 4

=item EventMan::EventModules::ControllerModules::CMode::SFOEvents

=item EventMan::EventModules::ControllerModules::CMode

=item EventMan::EventModules::ControllerModules

=item EventMan::EventModules

=back

=head1 Utility Module Usage

=over 4

=item EventMan::UtilityModules::PathValidation

=item EventMan::UtilityModules::SRAMDetection

=item EventMan::UtilityModules::SysLogging

=item EventMan::UtilityModules::SysStats

=item EventMan::UtilityModules::ControllerValidation

=item EventMan::UtilityModules::PortValidation

=back

=head1 B<Event Parameters>

Event Parameters are special parameters that define environment
tunables for all Events/EventSets to use.

=head2 I<-controller_down_time>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after
SFO (takeover/giveback) operations.

Per Resource Capable: No

Default: 600 (seconds)

=back

=head2 I<-controller_down_time_after_takeover>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after
SFO takeover operations. Overrides -controller_down_time

Per Resource Capable: No

Default: value of -controller_down_time

=back

=head2 I<-controller_down_time_after_giveback>

=over 4

Description: Specifies the amount of time (in seconds) to sleep after
SFO giveback operations. Overrides -controller_down_time

Per Resource Capable: No

Default: value of -controller_down_time

=back

=head2 I<-sfo_timeout>

=over 4

Description: Specifies the maximum timeout to wait for the cluster to stabilize
after an SFO to/gb. If exceeded, it will be treated as a warning.

Per Resource Capable: No

Default: 300 (seconds)

=back

=head2 I<-sfo_max_timeout>

=over 4

Description: Specifies the upper limit timeout to wait for the cluster to
stabilize after a storage failover beyond the value of -sfo_timeout. If an SFO to/gb
does not occur before this timeout, it will be treated as a failure.

Per Resource Capable: No

Default: 4x the value of -sfo_timeout

=back

=head2 I<-sfo_giveback_retry_count>

=over 4

Description: Specifies the number of retries to perform when a giveback fails.

Per Resource Capable: No

Default: 3

=back

=head2 I<-sfo_giveback_retry_sleep_time>

=over 4

Description: Specifies the amount of time to sleep before retrying
an SFO giveback.

Per Resource Capable: No

Default: 300 (seconds)

=back

=head2 I<-sfo_terminate_savecore_process>

=over 4

Description: Specifies whether to terminate the savecore process during panic SFO takeovers

Per Resource Capable: No

Default: 1

=back

=head2 I<-sfo_terminate_savecore_process_frequency>

=over 4

Description: Indicates how often to run the terminate savecore process method

Per Resource Capable: No

Default: 60 (seconds)

=back

=head2 I<-sfo_terminate_savecore_process_init_sleep_time>

=over 4

Description: Indicates how long to wait before starting the savecore termination process

Per Resource Capable: No

Default: floor(controller_down_time / 4)

=back

=head2 I<-sfo_parallel_delay_time>

=over 4

Description: Indicates how long to delay after synchronizing events before starting SFO Events

Per Resource Capable: No

Default: 0

=back

=head2 I<-sfo_disable_autosupport_before_giveback>

=over 4

Description: Specifies whether to toggle autosupport off on all nodes before doing a giveback.

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-sfo_enable_autosupport_after_giveback>

=over 4

Description: Specifies whether to turn on autosupport on all nodes after doing a giveback.

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-symmetrical_events>

=over 4

Description: Indicates whether to perform symmetrically styled events (A -> B then B -> A)

Per Resource Capable: No

Default: 1 (on)

=back

=head2 I<-collect_sfo_aggr_outage_times>

=over 4

Description: Indicates whether to collect SFO aggr outages after givebacks.

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-sfo_display_sfo_aggr_outage_ems_logs>

=over 4

Description: Indicates whether to display EMS logs when determining aggr outages during SFOs

Per Resource Capable: No

Default: 0 (off)

=back

=head2 I<-sfo_allow_auto_giveback>

=over 4

Description: When performing a giveback, allow the logic to continue
if an auto giveback had already occurred.

Per Resource Capable: No

Default: $EventUtil::FALSE (will fail if a giveback had occurred automatically)
$EventUti::TRUE will allow case where ONTAP performed an auto giveback.

=back

=head2 I<-sfo_detect_waiting_for_giveback_timeout>

=over 4

Description: Specifies the timeout used when detecting whether a node is at the
waiting for giveback prompt prior to issuing a giveback.

Per Resource Capable: No

Default: 300 (seconds)

=back

=head2 I<-sfo_auto_giveback_threshold>

=over 4

Description: Prior to performing auto givebacks, ONTAP will display the amount
of seconds before the giveback will occur. Before performing manual givebacks,
if sfo_allow_auto_giveback is set to 1, the event will compare the auto giveback
timer with the threshold value of this parameter. If the giveback timer is greater
than the threshold value, then the event will intervene and perform a manual giveback.
If the giveback time is less than the threshold value, then the event will not intervene
and allow the auto giveback to proceed.

Per Resource Capable: No

Default: 30 (seconds)

=back

=head2 I<-sfo_failed_auto_giveback_sleep_time>

=over 4

Description: If an auto giveback is detected to have failed, this controls the timer that waits
for ONTAP to attempt to retry the auto giveback before intervening and performing a manual giveback.
Lowering this value may result in possible race conditions during the giveback process. One symptom
is that:
 1) an auto giveback failed,
 2) the event proceeded to intervene with a manual giveback,
 3) ONTAP retried the giveback and succeeded,
 4) the event attempted to manually giveback and fails because the system is already giving back

Per Resource Capable: No

Default: 90 (seconds)

=back

=head1 B<Event Keywords>

=head2 B<nacl_sfo_takeoverCm>

Performs an SFO takeover on a nacl_nodes_cm resource. To complete
this half event, use nacl_sfo_givebackCm or nacl_sfo_forcedgiveback.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_sfo_takeoverCm>

Required: An array reference of a cmode NACL node resources to
operate against, this event will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_takeoverCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_forcedtakeoverCm>

Performs an SFO forced takeover on a nacl_nodes_cm resource. To complete
this half event, use nacl_sfo_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_forcedtakeoverCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_givebackCm>

Performs an SFO giveback on a nacl_nodes_cm resource. Assumption
is that the node specified has currently taken over its HA partner.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_givebackCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_forcedgivebackCm>

Performs an SFO forced giveback on a nacl_nodes_cm resource. Assumption
is that the node specified has currently taken over its HA partner.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_forcedgivebackCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_panicCm>

Performs an SFO panic takeover on a nacl_nodes_cm resource. To complete
this half event, use nacl_sfo_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_panicCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_haltCm>

Performs an SFO halt takeover on a nacl_nodes_cm resource. To complete
this half event, use nacl_sfo_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_haltCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_rebootCm>

Performs an SFO reboot takeover on a nacl_nodes_cm resource. To complete
this half event, use nacl_sfo_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_rebootCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_reset_takeoverCm>

Performs an SFO power reset takeover on a nacl_nodes_cm resource pair. Power
will be reset on the specified nacl_nodes_cm resource and the HA partner should
issue a takeover. To complete this half event, use nacl_sfo_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_reset_takeoverCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_off_takeoverCm>

Performs an SFO power off takeover on a nacl_nodes_cm resource pair. Power
will be turned off on the specified nacl_nodes_cm resource and the HA partner will
issue a takeover. The first node in the list will remain powered off.
To complete this half event, use nacl_sfo_power_on_givebackCm.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_off_takeoverCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_on_givebackCm>

Performs an SFO power on giveback on a nacl_nodes_cm resource pair. Power
will be turned on on the specified nacl_nodes_cm resource and a giveback will
be issued on the HA partner once the booted node enters the waiting
for giveback state.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_on_givebackCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_on_forcedgivebackCm>

Performs an SFO power on forced giveback on a nacl_nodes_cm resource pair. Power
will be turned on on the specified nacl_nodes_cm resource and a forced giveback
will be issued on the HA partner once the booted node enters the waiting
for giveback state.

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against, this event
will only operate on the first node in the array.

Resource Count: 2

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_on_forcedgivebackCm",
                         -nacl_nodes_cm => ["fas3070-rtp40"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_takegiveCm>

Performs a SFO takeover giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_takegiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41", "fas3170-rtp1", "fas3170-rtp2"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_forcedtakegiveCm>

Performs a SFO forced takeover giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_forcedtakegiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_takeforcedgiveCm>

Performs a SFO takeover forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_takeforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_forcedtakeforcedgiveCm>

Performs a SFO forced takeover forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_forcedtakeforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_panicgiveCm>

Performs a SFO panic giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_panicgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_panicforcedgiveCm>

Performs a SFO panic forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_panicforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_haltgiveCm>

Performs a SFO halt giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_haltgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_haltforcedgiveCm>

Performs a SFO halt forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_haltforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_rebootgiveCm>

Performs a SFO reboot giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Event Parameters

=item *I<-allow_auto_nacl_sfo_giveback>

Optional: $EventUti::TRUE will handle case where ONTAP performs an auto giveback.
          $EventUtil::FALSE will perform giveback.

Default: $EventUtil::TRUE

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_rebootgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_rebootforcedgiveCm>

Performs a SFO reboot forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Event Parameters

=item *I<-allow_auto_nacl_sfo_giveback>

Optional: $EventUti::TRUE will handle case where ONTAP performs an auto giveback.
          $EventUtil::FALSE will perform giveback.

Default: $EventUtil::TRUE

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_rebootforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_reset_takegiveCm>

Performs a SFO power reset takeover giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_reset_takegiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_reset_takeforcedgiveCm>

Performs a SFO power reset takeover forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_reset_takeforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_off_on_takegiveCm>

Performs a SFO power off/on takeover giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_off_on_takegiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head2 B<nacl_sfo_power_off_on_takeforcedgiveCm>

Performs a SFO power off/on takeover forced giveback scenario on a series of
CMode HA pairs (in sequence if multiple pairs are specified).

=over 4

=item Event Module Inheritance

default

=item Resources

=item *I<-nacl_nodes_cm>

Required: An array reference of a cmode NACL node resources to operate against.

Resource Count: 2+ (any even number greater than or equal to 2)

Default: all cmode NACL nodes in the Event scope will be used

=item *I<-sanclients>

Optional: An array reference of SanClients to use when performing path validation

Default: all SanClients in the Event scope will be used

=item Examples

 # see EventMan::add_event()
 $EventManobj->add_event(-event_name => "nacl_sfo_power_off_on_takeforcedgiveCm",
                         -nacl_nodes_cm => ["fas3070-rtp40", "fas3070-rtp41"]);
 if ($EventManobj->error()) {
     ...
 }

=back

=head1 B<Public Methods>

=head2 B<sfo_takeover_abstracted(%args)>

Performs an sfo takeover on the first sanfilersCm resource available.

=over 4

=item Inputs

=item *I<-forcetake>

Optional: $EventUtil::TRUE will issue a forced takeover, $EventUtil::FALSE will issue a normal takeover
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue a fored sfo takeover on the first cmode controller in the resource list
 $EventManObj->sfo_takeover_abstracted(-forcetake => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_takeover_generic(%args)>

Performs an sfo takeover on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node to issue the takeover.

=item *I<-taken_over_node>

Required: The node that will be taken over.

=item *I<-forcetake>

Optional: $EventUtil::TRUE will issue a forced takeover, $EventUtil::FALSE will issue a normal takeover
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo takeover on the specified cmode HA pair
 $EventManObj->sfo_takeover_generic(-takeover_node => "fas3070-rtp50",
                                    -taken_over_node => "fas3070-rtp51",
                                    -forcetake => $EventUtil::FALSE,
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_giveback_abstracted(%args)>

Performs an sfo giveback on the first sanfilersCm resource available.

=over 4

=item Inputs

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an unforced sfo giveback on the first cmode controller in the resource list
 $EventManObj->sfo_giveback_abstracted(-forcegive => $EventUtil::FALSE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_giveback_generic(%args)>

Performs an sfo giveback on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-giveback_node>

Required: The node to issue the giveback.

=item *I<-taken_over_node>

Required: The node that is currently taken over.

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo giveback on the specified cmode HA pair
 $EventManObj->sfo_giveback_generic(-giveback_node => "fas3070-rtp51",
                                    -taken_over_node => "fas3070-rtp50",
                                    -forcegive => $EventUtil::FALSE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_takegive_abstracted(%args)>

Performs an sfo takeover / giveback on all sanfilersCm HA pairs that are available
in sequence. If symmetrical_events are enabled, then each HA pair will swap roles
each time.

=over 4

=item Inputs

=item *I<-forcetake>

Optional: $EventUtil::TRUE will issue a forced takeover, $EventUtil::FALSE will issue a normal takeover
Default: $EventUtil::FALSE

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo takeover / forced giveback on all cmode HA pairs in sequence
 $EventManObj->sfo_takegive_abstracted(-forcetake => $EventUtil::FALSE,
                                       -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_takegive_generic(%args)>

Performs an sfo takeover / giveback on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node to issue the takeover.

=item *I<-taken_over_node>

Required: The node that will be taken over.

=item *I<-forcetake>

Optional: $EventUtil::TRUE will issue a forced takeover, $EventUtil::FALSE will issue a normal takeover
Default: $EventUtil::FALSE

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo takeover / forced giveback on the specified cmode HA pair
 $EventManObj->sfo_takegive_generic(-takeover_node => "fas3070-rtp50",
                                    -taken_over_node => "fas3070-rtp51",
                                    -forcetake => $EventUtil::FALSE,
                                    -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_panicgive_abstracted(%args)>

Performs an sfo panic / giveback on all sanfilersCm HA pairs that are available
in sequence. If symmetrical_events are enabled, then each HA pair will swap roles
each time.

=over 4

=item Inputs

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo panic / forced giveback on all cmode HA pairs in sequence
 $EventManObj->sfo_takegive_abstracted(-forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_panicgive_generic(%args)>

Performs an sfo panic / giveback on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the panicked node.

=item *I<-panic_node>

Required: The node that will be panicked.

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo panic / forced giveback on the specified cmode HA pair
 $EventManObj->sfo_panicgive_generic(-takeover_node => "fas3070-rtp50",
                                     -panic_node => "fas3070-rtp51",
                                     -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_panic_abstracted(%args)>

Performs an sfo panic takeover on a single Cmode HA pair in scope

=over 4

=item Inputs

=item *I<-use_nacl_apis>

Optional: specifies whether to use NACL apis when performing the panic
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo panic takeover on the nodes in scope
 $EventManObj->sfo_panic_abstracted();
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_panic_generic(%args)>

Performs an sfo panic takeover on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the panicked node.

=item *I<-panic_node>

Required: The node that will be panicked.

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo panic takeover on the specified cmode HA pair
 $EventManObj->sfo_panic_generic(-takeover_node => "fas3070-rtp50",
                                 -panic_node => "fas3070-rtp51");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_rebootgive_abstracted(%args)>

Performs an sfo reboot / giveback on all sanfilersCm HA pairs that are available
in sequence. If symmetrical_events are enabled, then each HA pair will swap roles
each time.

=over 4

=item Inputs

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo reboot / forced giveback on all cmode HA pairs in sequence
 $EventManObj->sfo_rebootgive_abstracted(-forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_rebootgive_generic(%args)>

Performs an sfo reboot / giveback on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the rebooted node.

=item *I<-reboot_node>

Required: The node that will be rebooted.

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo reboot / forced giveback on the specified cmode HA pair
 $EventManObj->sfo_rebootgive_generic(-takeover_node => "fas3070-rtp50",
                                      -reboot_node => "fas3070-rtp51",
                                      -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_reboot_generic(%args)>

Performs an sfo reboot takeover on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the rebooted node.

=item *I<-reboot_node>

Required: The node that will be rebooted.

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo reboot takeover on the specified cmode HA pair
 $EventManObj->sfo_reboot_generic(-takeover_node => "fas3070-rtp50",
                                  -reboot_node => "fas3070-rtp51");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_haltgive_abstracted(%args)>

Performs an sfo halt / giveback on all sanfilersCm HA pairs that are available
in sequence. If symmetrical_events are enabled, then each HA pair will swap roles
each time.

=over 4

=item Inputs

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo halt / forced giveback on all cmode HA pairs in sequence
 $EventManObj->sfo_haltgive_abstracted(-forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_haltgive_generic(%args)>

Performs an sfo halt / giveback on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the halted node.

=item *I<-halt_node>

Required: The node that will be halted.

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo halt / forced giveback on the specified cmode HA pair
 $EventManObj->sfo_haltgive_generic(-takeover_node => "fas3070-rtp50",
                                    -halt_node => "fas3070-rtp51",
                                    -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_halt_generic(%args)>

Performs an sfo halt takeover on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the halted node.

=item *I<-halt_node>

Required: The node that will be halted.

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo halt takeover on the specified cmode HA pair
 $EventManObj->sfo_halt_generic(-takeover_node => "fas3070-rtp50",
                                -halt_node => "fas3070-rtp51");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_powergive_abstracted(%args)>

Performs an sfo takeover / giveback by on all sanfilersCm HA pairs that are available
in sequence. The takeover is triggered by a power loss fault.
If symmetrical_events are enabled, then each HA pair will swap roles
each time.

=over 4

=item Inputs

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an sfo takeover / forced giveback on all cmode HA pairs in sequence
 # triggered by power loss faults
 $EventManObj->sfo_powergive_abstracted(-forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_powergive_generic(%args)>

Performs an sfo takeover / giveback (triggered by power loss)
on a single CMode HA Pair specified by the caller.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the node that experiences a power loss.

=item *I<-power_change_node>

Required: The node that will experience a power loss.

=item *I<-power_change_type>

Required: The type of power loss to issue, 'off' or 'reset'. Power 'reset' behaves
more like a power blip. Power 'off' leaves the node down until the takeover is completed.
Once the takeover is completed, power will be restored.

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue a power off fault to trigger an sfo takeover / forced giveback
 # on the specified cmode HA pair
 $EventManObj->sfo_powergive_generic(-takeover_node => "fas3070-rtp50",
                                     -power_change_node => "fas3070-rtp51",
                                     -power_change_type => "reset",
                                     -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_power_change_abstracted(%args)>

Performs an a power change 'off|reboot' or 'on' against the first CMode sanfilersCm resource
available. Will issue a giveback if power change type is set to 'on'.

=over 4

=item Inputs

=item *I<-power_change_type>

Required: Specifies what type of power change to perform 'on', 'off', or 'reboot.'

=item *I<-forcegive>

Optional: $EventUtil::TRUE will issue a forced giveback, $EventUtil::FALSE will issue a normal giveback
Only applies to if -power_change_type => 'on'.
Default: $EventUtil::FALSE

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # power on the first CMode node resource available and perform a giveback on it's
 # HA partner.
 $EventManObj->sfo_power_change_abstracted(-power_change_type => "on",
                                           -forcegive => $EventUtil::TRUE);
 if ($EventManObj->error()) {
     ...
 }

 # power reset on the first CMode node resource available
 $EventManObj->sfo_power_change_abstracted(-power_change_type => "reset");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_power_change_generic(%args)>

Issue a power reset/off to trigger an SFO takeover against the
specified CMode HA pair.

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that will takeover the node that experiences a power loss.

=item *I<-power_change_node>

Required: The node that will experience a power loss.

=item *I<-power_change_type>

Required: The type of power loss to issue, 'off' or 'reset'. Power 'reset' behaves
more like a power blip. Power 'off' leaves the node down until the takeover is completed.
Once the takeover is completed, power will be restored.

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # issue an power reset on the power_change_node
 # to trigger a takeover to occur on the -takeover_node
 $EventManObj->sfo_power_change_generic(-takeover_node => "fas3070-rtp50",
                                        -power_change_node => "fas3070-rtp51",
                                        -power_change_type => "reset");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<sfo_power_on_generic(%args)>

Power on the specified node and wait for it to the
reach waiting for giveback state

=over 4

=item Inputs

=item *I<-takeover_node>

Required: The node that has taken over the node to be powered on

=item *I<-power_change_node>

Required: The node that is being powered on

=item *I<-use_nacl_apis>

Optional: Specifies whether to use NACL apis when performing the operation.
Default: $EventUtil::FALSE

=item Examples

 # poewr on the specified node and wait for it to reach waiting for giveback
 $EventManObj->sfo_power_on_generic(-takeover_node => "fas3070-rtp50",
                                    -power_change_node => "fas3070-rtp51");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<get_sfo_aggr_outage_times(%args)>

Obtain the sfo aggr outage times within a start and end sentinel via the EMS logs
of the controller.
Note: This functionality is controlled via the event parameter -collect_sfo_aggr_outages
is disabled by default.

=over 4

=item Inputs

=item *I<-giveback_node>

Required: The node that has taken over its ha partner

=item *I<-taken_over_node>

Required: The node that has been taken over by its ha partner

=item *I<-start_sentinel>

Required: The start sentinel string

=item *I<-end_sentinel>

Required: The end sentinel string

=item *I<-display_label>

Optional: A label to add to the data that is displayed
Default: ""

=item Examples

 # poewr on the specified node and wait for it to reach waiting for giveback
 my $outage_hash = $EventManObj->get_sfo_aggr_outage_times(-giveback_node => "fas3070-01",
                                                           -taken_over_node => "fas3070-02",
                                                           -start_sentinel => "some start sentinel",
                                                           -end_sentinel => "some end sentinel",
                                                           -display_label => "A sfo giveback");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<detect_sfo_takeover(%args)>

This will detect that an SFO takeover has completed successfully.

=over 4

=item Inputs

=item *I<-sfo_timeout>

Required: Timeout in seconds that specifies the lowerbound limit of
when to warn if the takeover takes this long.

=item *I<-sfo_max_timeout>

Required: Timeout in seconds that specifies the upperbound limit of
when to fail if the takeover does not complete by thie time.

=item *I<-take_over_node>

Required: The name of the node performing the take over

=item *I<-taken_over_node>

Required: The name of the node being taken over

=item *I<-take_over_state>

Optional: Specifies what state to expect the take over node to be in when it completes the takeover.

Valid Values: takeover, takeover_autogiveback_scheduled

Default: takeover

=item Examples

 # detect the takeover to complete
 $EventManObj->detect_sfo_takeover(-sfo_timeout => "300",
                                   -sfo_max_timeout => "1200",
                                   -take_over_node => "node_a",
                                   -taken_over_node => "node_b");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<detect_sfo_giveback(%args)>

This will detect that an SFO giveback has completed successfully.

=over 4

=item Inputs

=item *I<-sfo_timeout>

Required: Timeout in seconds that specifies the lowerbound limit of
when to warn if the giveback takes this long.

=item *I<-sfo_max_timeout>

Required: Timeout in seconds that specifies the upperbound limit of
when to fail if the giveback does not complete by thie time.

=item *I<-giveback_node>

Required: The name of the node performing the giveback

=item *I<-taken_over_node>

Required: The name of the node that is currently taken over

=item Examples

 # detect the giveback to complete
 $EventManObj->detect_sfo_giveback(-sfo_timeout => "300",
                                   -sfo_max_timeout => "600",
                                   -giveback_node => "node_a",
                                   -taken_over_node => "node_b");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<detect_waiting_for_giveback(%args)>

This will detect that an ha pair is ready for giveback.

=over 4

=item Inputs

=item *I<-timeout>

Required: Timeout in seconds that specifies when to fail if the ha pair
is not ready for giveback.

=item *I<-take_over_node>

Required: The name of the node that performed the take over

=item *I<-taken_over_node>

Required: The name of the node that is currently taken over

=item Examples

 # detect the giveback to complete
 $EventManObj->detect_waiting_for_giveback(-timeout => "300",
                                           -take_over_node => "node_a",
                                           -taken_over_node => "node_b");
 if ($EventManObj->error()) {
     ...
 }

=back

=head2 B<wait_for_sfo_state(%args)>

This will wait for the specified nodes to enter their respective SFO states. This function
will return successfully with no exception if all the respective states have been hit.
A EventMan::Exceptions::TimedOutWaitingForSFOState exception will be thrown if they
do not reach the states in the requested time. An EventMan::Exceptions::EncounteredSFOAbortState
exception will be thrown if any of the nodes encounter a specified abort states.

=over 4

=item Inputs

=item *I<-method_timeout>

Optional: Specifies when the method should timeout and throw the exception (EventMan::Exceptions::TimedOutWaitingForSFOState)

Default: 180

=item *I<-command_interface>

Required: The NACL Component command interface to use to issue commands, should be a derived class
of NACL::C::Node or NACL::C::Cluster.

=item *I<-poll_interval>

Optional: Specifies how often to send the command to check state

Default: 1 (second)

=item *I<-sfo_state_hash>

Required: A hash of 1 or more nodename keys whose values represent states each node should hit.

Format:
        { "mynode1" => {states => ["connected"],
                        state_descriptions => [qr/Connected/, qr/Up/]},
          "mynode2" => {states => ["giveback_partial_connected"],
                        state_descriptions => [qr/Unknown/i],
                        abort_states => ["connected"],
                        abort_state_descriptions => [qr/Failed/,"Encountered a Failure"]} }

=item *I<-check_once>

Optional: Specifies if the method should check only once. If the expected states are not encountered
then an exception EventMan::Exceptions::UnexpectedSFOState will be thrown.

Default: $EventUtil::FALSE

=item *I<-abort_states>

Optional: An array of global abort states (strings or regexes) for all nodes. If any of these abort
states are encountered, then an exception EventMan::Exceptions::EncounteredSFOAbortState will
be thrown

Default: []

=item *I<-abort_state_descriptions>

Optional: An array of global abort state descriptions (strings or regexes) for all nodes.
If any of these abort states are encountered, then an exception
EventMan::Exceptions::EncounteredSFOAbortState will be thrown.

Default: []

=item Examples

 # detect that $take_over_node_name has entered a takeover state, timeout after 1200 seconds
 $EventManObj->wait_for_sfo_state(-command_interface => $take_over_node_obj,
                                  -sfo_state_hash =>
                                      {"$take_over_node_name" => { states => ["takeover"] }},
                                  -method_timeout => 1200);

=back

=head2 B<check_for_auto_sfo_giveback(%args)>

Return true or false, depending on whether a manual giveback is required between the HA
pair specified.

=over 4

=item Inputs

=item *I<-giveback_node>

Required: The name of the node that will be performing the giveback

=item *I<-taken_over_node>

Required: The name of the node that may or may not be currently taken over.

=item Examples

 # detect whether a manual giveback is required
 my $should_perform_giveback = $EventManObj->check_for_auto_sfo_giveback(-giveback_node => "node_a",
                                                                         -taken_over_node => "node_b");
 if ($EventManObj->error()) {
     ...
 }

=back

=cut
