#
# $Id: //depot/prod/test/Rsierranevada/lib/GNOTE/GX/Volume.pm#2 $
#
## @summary  GNOTE GX Volume module
##
## @author   dl-dpgqa-automation@netapp.com
##
## @status   Public
##
## @pod      here

=head1 NAME

NOTE::GX::Volume

=head1 DESCRIPTION

The purpose of this module is to provide 10-mode specific methods for the volume object

=head1 METHODS

=cut

package NOTE::GX::Volume;

use strict;
use warnings;
no warnings 'redefine';

use GNOTE;
use GNOTE::GX;
use GNOTE::Volume;
use GNOTE::GX::Quota;
use GNOTE::GX::San::Lun;
use fileSystem;
use File::Basename;
use Getopt::Std;
use Carp qw(croak confess);
use Params::Validate qw(validate ARRAYREF HASHREF OBJECT SCALAR UNDEF);
use Data::Dumper;

our @ISA = qw(NOTE::GX::Quota NOTE::GX::San::Lun NOTE::Volume NOTE::GX);
use TracePkg;

my $Trace = TracePkg->new();

sub _connect {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $args = $self->{_args};

    $args->{Node} ||= $args->{Filer};
    if ( ref $args->{Node} eq "HASH" ) {
        $self->{_Node} = NOTE::GX::Node->connect( %{ $args->{Node} } );
    }
    elsif ( ref $args->{Node} eq "NOTE::GX::Cluster" ) {
        $self->{_Node} = $args->{Node}->Node;    # This is already an object
    }
    elsif ( ref $args->{Node} eq "NOTE::GX::Node" ) {
        $self->{_Node} = $args->{Node};          # This is already an object
    }
    elsif ( ref \$args->{Node} eq "SCALAR" ) {

        # Is just a hostspec
        $self->{_Node} = NOTE::GX::Node->connect( name => $args->{Node} );
    }
    $self->{_Cluster} = NOTE::GX::Cluster->connect( Node => $self->{_Node} );
    $self->{_Filer} = $self->{_Cluster};

    if ( !defined( $args->{Vserver} ) ) {

        # Expect vserver to be set if Vserver is not
        $self->{_Vserver} = NOTE::GX::Vserver->connect(
            Node => $self->{_Cluster},
            name => $args->{vserver}
        );
    }
    elsif ( ref $args->{Vserver} eq "HASH" ) {
        $self->{_Vserver} = NOTE::GX::Vserver->connect(
            Node => $self->{_Cluster},
            %{ $args->{Vserver} }
        );
    }
    elsif ( ref $args->{Vserver} eq "NOTE::GX::Vserver" ) {
        $self->{_Vserver} = $args->{Vserver};
        $self->{_Cluster} = $self->{_Vserver}->Cluster;
    }
    elsif ( ref \$args->{Vserver} eq "SCALAR" ) {
        $self->{_Vserver} = NOTE::GX::Vserver->connect(
            name => $args->{Vserver},
            Node => $self->{_Cluster},
        );
    }
    delete $args->{Filer};
    delete $args->{Node};
    delete $args->{Vserver};

    return $self;
}

sub new {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self = {};
    my $args;

    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    $self->{_args} = $args = \%args;

    exists $args->{type}  and $args->{type}  = NOTE::option( $args->{type} );
    exists $args->{name}  and $args->{name}  = NOTE::option( $args->{name} );
    exists $args->{disks} and $args->{disks} = NOTE::option( $args->{disks} );
    exists $args->{recycle}
      and $args->{recycle} = NOTE::option( $args->{recycle} );
    exists $args->{anonymous}
      and $args->{anonymous} = NOTE::option( $args->{anonymous} );
    exists $args->{snaplock}
      and $args->{snaplock} = NOTE::option( $args->{snaplock} );
    exists $args->{richdata}
      and $args->{richdata} = NOTE::option( $args->{richdata} );
    exists $args->{dataset}
      and $args->{dataset} = NOTE::option( $args->{dataset} );
    exists $args->{raidtype}
      and $args->{raidtype} = NOTE::option( $args->{raidtype} );
    exists $args->{rgsize}
      and $args->{rgsize} = NOTE::option( $args->{rgsize} );

    if ( $self->{_type} && exists $args->{ $self->{_type} } ) {
        %{$args} = %{
            NOTE::_hashOverlay(
                target    => $args,
                source    => $args->{ $self->{_type} },
                overwrite => 0
            )
          };
    }
    delete $args->{flex};
    delete $args->{striped};

    $args->{size} and $args->{size} = NOTE::option( $args->{size} );
    if ( !defined $args->{size} ) {
        if ( $args->{dataset} ) {

            # In this case, don't allow volume size to be undefined,
            # otherwise, the volume chosen might be bigger than the minimum.
            # If this is a source volume, the destination might then
            # end up being smaller, if it also was created without
            # specifying a size. Setting it here insures that it
            # will be exactly the minimum allowed size.
            $args->{size} =
              $self->{_type} eq 'striped'
              ? POSIX::ceil( 1.2 * 1024 * 1024 ) . 'k'
              : "20m";
        }
    }
    bless $self, $class;
    NOTE::reference($self);

    # This is needed because $self is already an object, but
    # hasn't been physically created yet.
    # This flag will be cleared when the volume is recycled
    # or created.
    $self->is_destroyed = 1;

    $self->_connect;
    $self->SUPER::new;

    # At this point, we know what the aggregate is.
    $args->{size} = NOTE::units(
        $args->{size}
          || (
            $self->{_type} eq 'striped'
            ? POSIX::ceil( 1.2 * 1024 * 1024 ) . 'k'
            : "20m"
          )
    );
    if ( ( !$args->{size} ) && $self->{_Aggregate}->voltype eq 'striped' ) {

        # Minimum size * number of nodes, in this case
        $args->{size} *= scalar $self->{_Aggregate}->nodes;
    }
    return $self;
}

sub create {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $args{type} eq 'trad' ) {
        confess "Tradvols are unsupported in GX";
    }

    my $self = __PACKAGE__->new(%args);
    if ( $self->exists ) {

        # Refresh junction and share for existing volume
        $self->_refresh_junction_and_share();
    }
    else {
        $self->SUPER::create;
    }
    $self->is_destroyed = 0;
    return $self;
}

sub attach {
    my ( $class, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $self = {};

    $self->{_args} = \%args;
    my $exportfs_opts = ( delete $args{exportfs} ) || "rw,anon=0";
    $self->{_debug} = $args{debug};
    $self->{_owner} = $$;

    exists $args{type} and $args{type} = NOTE::option( $args{type} );

    my $node_name = (
        ref( $args{Node} ) =~ /NOTE::GX::Node/
        ? $args{Node}->name
        : $args{Node}
    );
    NOTE::logcomment( "Attaching to " . ( exists $args{parent} ? "member " : "" ) . "volume " . $args{name} . " using node " . $node_name );

    if ( exists $args{parent} ) {
        confess "'parent' parameter must be a 'NOTE::Volume' object"
          unless ref( $args{parent} ) =~ /NOTE::.*Volume/;
        $self->{_parent}    = delete $args{parent};
        $self->{_type}      = "striped";
        $self->{_Aggregate} = delete $args{Aggregate};
    }

    bless $self, $class;
    NOTE::reference($self);
    $self->SUPER::attach(%args);

    $self->_refresh_junction_and_share;

    return $self;
}

# Refresh a volume's junction and share from ONTAP

sub _refresh_junction_and_share {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # remove existing junction or share
    delete $self->{_path};
    delete $self->{_share};

    # Setup object's junction and share from ONTAP
    my $s = $self->status("junction-path");
    if ( $s ne '-' ) {
        $self->{_path} = $s;
        my $output;

        # Check both R-Tree and RW-Tree for an existing share
        foreach my $extra ( "", "/.admin" ) {
            my $path = $extra . $self->{_path};
            eval { $output = $self->server->vserver_cifs_share_show( "path" => $path ); };
            if ($@) {
                confess "Failed to determine cifs share for " . $path . ". Error: $@"
                  unless $@ =~ /no entries/;
            }
            else {
                $self->{_share} = $output->[0]->{'share-name'};
                last;
            }
        }
    }
}

=head2 ::add_members

Increase the number of members in a striped volume

=over

=item Arguments

B<count> => I<integer>

I<Optional>

Number of members to add. Default: 1

B<background> => I<boolean>

Indicated whether to run operation in background. Default: 0

=item Examples

    $vol->add_members(); # Add a member to the striped volume

    $vol->add_members(background=>1); # Add a member in the background
    .
    .
    $vol->job_wait(); # Wait for add member to complete

=back

=cut

sub add_members {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Found job " . $self->{_job_id} . " Multiple concurrent volume operation not allowed"
      if exists $self->{_job_id};
    $args{count} = 1
      unless exists $args{count};

    my $background = NOTE::GX::background( \%args );

    my $text;
    if (! eval { $text = $self->server->volume_add_members( vserver => $self->Vserver->name, volume => $self->name, %args ); }) {
        $self->_processError($@);
    }

    $self->{_job_id} = $self->Node->job_id($text);
    if ( $args{background} ) {
        $self->Node->job_wait( id => $self->{_job_id}, state => "running" );
    }
    else {
        $self->job_wait(%args);
    }
}

=head2 ::Members

Return NOTE::Volume object(s) representing member volumes representing member volumes of a certain type.
For MDV/RDV volume a single NOTE::Volume object will be returned, otherwise an array.
If type is not specified all member volumes will be returned in an array

=over

=item Arguments

B<type> => I<mdv|rdv|dv>

I<Optional>

Returns NOTE::Volume object for member volumes of the specified type if specified.
'rdv' is specific acro member volume
'mdv' is specific coral member volume
'dv' member DV volumes for both acro and coral

=item Examples

=back

    my @vols = $striped_vol->Members; # Return all member volumes
    my $member = $striped_vol->Members(type=>"MDV"); # Return just one member volume

=cut

sub Members {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess $self->description . " is not a striped volume. Type " . $self->{_type}
      unless $self->{_type} =~ /striped/;

    if ( exists $args{type} ) {
        confess $self->description . " RDV member type only exists for acro volumes"
          if ( $args{type} =~ /rdv/i
            && !( $self->{_args}->{'stripe-format'} =~ /full/ ) );
        confess $self->description . " MDV member type only exists for coral volumes"
          if ( $args{type} =~ /mdv/i
            && !( $self->{_args}->{'stripe-format'} =~ /data/ ) );
    }

    if ( exists $self->{_parent} ) {
        return $self->{_parent}->Members(%args);
    }
    else {
        my $mrdv_name = $self->name;
        my $status    = $self->server->volume_member_show();    # Returns all
        my @aggr      = $self->Aggregate->Members();
        my %ma        = map { $_->name => $_ } @aggr;
        my ( @names, @list );
        foreach my $s ( @{$status} ) {
            next
              unless $s->{volume} =~ /^$mrdv_name/;             # discard members for other striped volumes
            if ( exists $args{type} ) {
                if ( $args{type} =~ /(m|r)dv/i ) {
                    next unless $s->{volume} eq $mrdv_name;
                }
                else {
                    next if $s->{volume} eq $mrdv_name;
                }
            }
            push @names, $s->{volume};
            push @list,
              NOTE::GX::Volume->attach(
                tag       => "member of " . $self->description,
                name      => $s->{volume},
                Node      => $ma{ $s->{aggregate} }->status->{node},
                Vserver   => $s->{vserver},
                Aggregate => $ma{ $s->{aggregate} },
                parent    => $self
              );
        }
        if ( $args{type} =~ /(m|r)dv/i ) {
            if (@list) {
                NOTE::logcomment( uc( $args{type} ) . " Volume for " . $self->description . ":\n    " . $list[0]->name );
                return $list[0];
            }
            else {
                confess( "Unable to locate the " . $args{type} . " Volume" );
            }
        }
        else {
            NOTE::logcomment( ( $args{type} || "All" ) . " members for " . $self->description . ":\n    " . ( join ", ", @names ) );
            return @list;
        }
    }
}

=head2 ::status

Return a HASHREF containing the status of this volume or a single column

=over

=item Arguments

B<element> = I<string>

I<Optional>

Field name of column to return. If not specified will return all columns in a HASHREF

=item Example

 my $status = $vol->status();
 my $column = $vol->status("node");

Output from C<print Dumper( $vol->status() )>, example of what fields can be used with this method.

 $VAR1 = {
   'read-realloc' => 'off',
   'files' => '622580',
   'dsid' => '1061',
   'junction-path' => '/vl03065_fell_Packards',
   'is-cluster-volume' => 'true',
   'autosize' => 'false',
   'convert-ucode' => 'true',
   'snapshot-clone-dependency' => 'off',
   'percent-snapshot-space' => '5%',
   'stripe-width' => '0',
   'junction-path-source' => 'RW_volume',
   'instance-uuid' => '00d2a570-3b90-11e0-ac61-00a098123984',
   'files-used' => '96',
   'max-autosize' => '25769410560',
   'volume-style' => 'flex',
   'block-type' => '64-bit',
   'index-dir-enabled' => '-',
   'extent-enabled' => 'off',
   'space-guarantee' => 'volume',
   'is-sis-volume' => 'false',
   'used' => '159744',
   'move-target' => 'false',
   'fractional-reserve' => '100%',
   'snapdir-access' => 'true',
   'max-write-alloc-blocks' => '0',
   'msid' => '2147484715',
   'is-copied-for-transition' => 'false',
   'properly-striped' => '-',
   'is-quiesced-in-memory' => 'false',
   'filesys-size-fixed' => 'false',
   'flexcache-connection-status' => '-',
   'vsroot' => 'false',
   'transition-state' => 'none',
   'size' => '21474836480',
   'flexcache-cache-policy' => '-',
   'group' => 'other',
   'create-time' => '"Wed Feb 16 13:51:41 2011"',
   'auto-snapshots-enabled' => 'false',
   'antivirus-on-access-policy' => 'default',
   'compression-space-saved-percent' => '0%',
   'language' => 'en_US',
   'create-ucode' => 'true',
   'autosize-increment' => '1073725440',
   'load-sharing-source' => '-',
   'security-style' => 'mixed',
   'ruleset' => 'default',
   'striping-epoch' => '-',
   'snapshot-policy' => 'none',
   'name-ordinal' => 'base',
   'overwrite-reserve' => '0',
   'stripe-count' => '-',
   'is-quiesced-on-disk' => 'false',
   'sched-snap-name' => 'create_time',
   'uuid' => 'cac7fbbc-39fd-11e0-b959-123478563412',
   'state' => 'online',
   'sis-space-saved-percent' => '0%',
   'maxdir-size' => '104857600',
   'policy' => 'default',
   'user' => 'root',
   'dedupe-space-saved-percent' => '0%',
   'space-mgmt-try-first' => 'volume_grow',
   'parent' => 'root_vs0',
   'is-sis-logging-enabled' => 'false',
   'total' => '20401094656',
   'sis-space-saved' => '0',
   'physical-replica' => 'false',
   'stripe-format' => '-',
   'dedupe-space-shared' => '0',
   'is-transitioned' => 'false',
   'type' => 'RW',
   'min-readahead' => 'false',
   'dedupe-space-saved' => '0',
   'nvfail' => 'off',
   'clone-volume' => 'false',
   'percent-used' => '5%',
   'snapshot-space-used' => '0%',
   'junction-active' => 'true',
   'flexcache-fill-policy' => '-',
   'aggregate' => 'AG15165_mutt_sinusoid',
   'available' => '20400934912',
   'space-guarantee-enabled' => 'true',
   'is-inconsistent' => 'false',
   'volume' => 'vl03065_fell_Packards',
   'vserver' => 'vs0',
   'atime-update' => 'true',
   'concurrency' => '-',
   'compression-space-saved' => '0',
   'provenance-uuid' => '00d2a570-3b90-11e0-ac61-00a098123984',
   'comment' => '""',
   'unix-permissions' => '---rwxrwxrwx',
   'optimize' => '-',
   'in-nvfailed-state' => 'false'
 };

=back

=cut

sub status {
    my ( $self, $element ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $status;

    my %args;
    if ( $self->Vserver ) {
        $args{vserver} = $self->Vserver->name;
    }
    else {
        $args{aggregate} = $self->Aggregate->name;
    }
    $args{volume} = $self->name;

    $status = $self->_status(%args);
    $self->_processError($@);

    if ( defined $element ) {
        return $status->[0]->{$element};
    }
    else {
        return $status->[0];
    }
}

# Convert path to 'full' path using PWD

sub _wex_path {
    my ( $self, $path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $path ||= "/";    # Defaults to root directory if path is undef

    #append pwd unless path contains private inode or snapshot id or inode number
    $path = $self->pwd . $path
      unless ( $path =~ /^\// || $path =~ /^\d+p$/ || $path =~ /^\d+/ || $path =~ /\~/ );

    if ( $path =~ /^(~\d+)\.\d+p$/ ) {    # if snapshot id is passed and private inode is present
        return $self->name . "$path";     # append volume name and return path
    }
    if ( $path =~ /^(~\d+)\.\d+$/ ) {     # if snapshot id is passed with just a inode number
        return $self->name . "$path";     # append volume name and return path
    }
    if ( $path =~ /^(~\d+)/ ) {           # if snapshot id is passed
        my $id = $1;                      # extract the snapid
        $path =~ s/$1/$id\./;             # replace it in path with <snapid>.<path>
        return $self->name . "$path";     # return path containing snapid
    }

    return $self->name . ".$path";
}

sub _stream_list {
    my ( $self, $inode, $snap_id ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $info;

    my $vol_name = $self->name;
    my $strm_dir = ".@$inode[0]->{sinode}";
    $strm_dir = "~$snap_id." . @$inode[0]->{sinode}
      if ($snap_id);
    eval { $info = $self->server->shell( cmd => "wex dir $vol_name$strm_dir" ); };    # use wex dir to get contents
    confess "wex dir for $vol_name$strm_dir failed. Error: $@"
      if $@;
    my %inode_hash;

    foreach my $line ( split( /\n/, $info ) ) {                                       # seperate the line of the wex dir output
        chomp $line;                                                                  # and get rid of newlines
        if ( $line =~ /entry/ && $line !~ /\".\"/ && $line !~ /\"..\"/ ) {            # this is a stream
            $line =~ / \"(.*)\" /;                                                    # get the stream name
            my $strm_name = $1;                                                       # ------^-------
            $line =~ /inum\s\d+/;                                                     # ----- v -------
            my $strm_inum = ( split( / /, $& ) )[1];                                  # get the inum of stream
            $inode_hash{$strm_name} = $strm_inum;
            $inode_hash{$strm_name} = "~$snap_id." . $strm_inum                       # append snap_id to stream inum
              if ($snap_id);                                                          # if the stream dir is in a snapshot
        }
    }
    return %inode_hash;
}

=head2 ::wex_info

Obtain wex information, such as inode number, generation number, etc,
on an object, such as file, directory, etc, from within this volume.
An Array containing a hash will be returned with the inode info. For striped
volumes each array member which is a hash will hold the inode info for each dsid
Optionally a named column can be passed to request the value of a
specific column.
If column name is specified a array containing the info will be returned

See http://wikid.netapp.com/w/WEX for full detail of parameters

=over

=item Arguments

B<format> = I<string>

I<Optional.>

Specifies which wex format to use.
Valid values are: volume, inode.

Default: inode

B<path> = I<object path>

I<Optional.>

A path to an 'object' within this volume that information is being requested.
If path does not have a leading '/'. it will be interpreted as related to the
current working directory as set by the previous 'cd' call to this object.

Default: / (i.e. Volume root)

B<column> => I<string>

I<Optional.>

Name of column to be return. If not specify returns all columns.

=item Examples

=back

    my @all_info = $vol->wex_info(path=>"/a_file");
    my @inode = $vol->wex_info(path=>"/a/file", column=>"inum"); # Returns the inode number of the file
    my @vol_info = $vol->wex_info(format=>"volume");

=cut

sub wex_info {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $format = delete $args{format} || "inode";

    # Process ACL and Stream inode request
    if ( $args{path} =~ /:/ ) {
        confess "only inode info supported for stream and ACL"
          unless $format =~ /inode/;

        my ( $item, $special ) = split /:/, $args{path};
	print "Will be calling with path: $item, generated from $args{path}\n";
        my @inode = $self->wex_info( path => $item );
        if ( $special =~ /\*/ ) {

            # Return inode of stream dir
            return $self->wex_info( path => $inode[0]->{sinode} );
        }
        elsif ( $special =~ /@/ ) {

            # Return ACL inode
            return $self->wex_info( path => $inode[0]->{xinode} );
        }
        else {

            # Return stream inode
            my %list = $self->_stream_list( \@inode );
            if ( exists $list{$special} ) {
                return $self->wex_info( path => $list{$special} );
            }
            else {
                confess "Unable to find stream " . $args{path} . " inode information";
            }
        }
    }

    my $path;
    if ( $format =~ /volume/ ) {
        $path = $self->name;
    }
    else {
        $path = $self->_wex_path( delete $args{path} );
    }

    my $info;
    $path =~ s/\.\//\//;    # BURT 589380

    eval { $info = $self->server->shell( cmd => "wex $format $path -dump all" ); };
    confess "wex $format info for " . $path . " failed. Error: $@"
      if $@;
    confess "wex $format info for " . $path . " unable to find file. Error: $info"
      if $info =~ /unable to find/;

    my @inod_info;
    my @tmp_ary;
    my $i = -1;

    foreach my $line ( split /\n/, $info ) {
        if ( $line =~ /=/ ) {
            my ( $key, $value ) = split /=/, $line;
            $i++ if ( $key eq 'dsid' );
            $inod_info[$i]->{$key} = $value;
        }
    }

    if ( $args{column} ) {
        foreach my $contents (@inod_info) {
            foreach ( keys(%$contents) ) {
                push( @tmp_ary, $contents->{$_} ) if ( $_ =~ $args{column} );
            }
        }
        return @tmp_ary;
    }
    else {
        return @inod_info;
    }
}

# private function, calls volume_explore( format=>str, dump=>"fild1,fld2..." );
# function returns array of hashes.
#
# data by files is multiline hash-string goupped in blocks which are separated by
# empty line.
#
#   dsid=1068
#   snapid=0
#   parent=64
#   name=.
#
#   dsid=1068
#   snapid=0
#   parent=64
#   name=..

sub _wex_volume_explore {   # WAFL Explorer
    my ($self,$cmd) = @_;               # args are format=>"" and dump=>""
    my $trace = $NOTE::TRACELOG->start_timer();     # log internal metrics
    my $raw = $self->server->shell(cmd => $cmd);    # getting raw output first
    my @wex;
    foreach(split(/\r*\n\r*\n/, $raw)) {        # split by empty line
        next unless /=/s;               # process actual data only
        my %elm = ( $_ =~ /\s*([^=]+)=([^\n\r]+)/gs );  # convert string to hash
        push @wex,\%elm;                # store hash ref in array
    }
    return @wex;                    # return array to user
}

=head2 ::wex_volume Get WEX info for volume.

Returns hash of WAFL parameters.

=over

=item Arguments

B<fields> = I<string>

I<Optional>

Comma separated list of parameters to return. Default value is 'all'. Acceptible values
are:

I<ressnap>, I<blockcount>, I<av-generation>, I<rec-policy>, I<c-totalfiles>, I<dsid>,
I<cptime>, I<p-curfiles>, I<spcres>, I<snapshot-schedule>, I<rec-percent>, I<vol_scc_id>,
I<maxfiles>, I<cifsflags>, I<magic>, I<totalfiles>, I<location>, I<mtime>, I<nfslanguage>,
I<name>, I<rec-trigger>, I<p-maxfiles>, I<vol_scc_time>, I<c-curfiles>, I<reserved>,
I<resfiles>, I<vol_scc_nodeid>, I<curfiles>, I<nvfail>, I<autogrow-max>, I<autogrow-inc>,
I<options>, I<crtime>, I<p-resfiles>, I<msid>, I<buftree-type>, I<maxdirsize>, I<raidflags>,
I<size>, I<slider>, I<overwrite>, I<fsid>, I<flags>, I<language>, I<version>, I<holes>,
I<cpcount>, I<snapid>, I<pclanguage>, I<rec-destroy>, I<vol_vcc_time>, I<av-cluster-version>

=item Note

If you are using name with parentheses, (same as parent=>1), then aggregate output is
returned instead however dsid stays the same.

=item Examples

    use strict;
    use Data::Dumper;
    use GNOTE;
    eval {
        my $vol = NOTE::SnapMirrorUtil->new()->create_volume(tag=>'src');
        my %wex = $vol->wex_volume();
        print "Vol WEX:\n". Dumper(\%wex);
    };
    logresult('FAIL',$@) if $@;
    NOTE::teardown;

=back

=cut

sub wex_volume {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();
    $args{name} ||= $self->name;
    $args{name} = "'($args{name})'" if $args{parent};
    return %{( $self->_wex_volume_explore("wex volume $args{name} -dump all") )[0]};
}

=head2 ::wex_inode Get WEX details for one or more inodes

Retuens array of hashes with WAFL info.

I<blockcount>, I<inum>, I<dsid>, I<junction>, I<state>, I<subtype>, I<fav>, I<i2p>,
I<inline>, I<epoch>, I<rdev>, I<uid>, I<location>, I<mtime>, I<name>, I<sinode>,
I<level>, I<used>, I<atime>, I<type>, I<gid>, I<vbn>, I<pvbn>, I<crtime>, I<av-gen-num>,
I<tid>, I<size>, I<nlink>, I<indircount>, I<payload>, I<mixed_buftree>, I<ctime>,
I<owner>, I<flags>, I<dvmask>, I<version>, I<xinode>, I<snapid>, I<generation>,
I<buftreeid>, I<umask>

=over

=item Arguments

B<inode> = I<string>

I<Optional>

Inode number or inode range in WEX format. Default is 64. Acceptible formats are:

 inode => '64'      # pull root inode for volume
 inode => '85p'     # pull metadata private inode
 inode => '64p-100p'    # pull range of inodes from 64 to 100

Refer to http://wikid.netapp.com/w/WEX for more information

B<parent> = I<boolean>

I<Optional>

Use parent container if parent is TRUE.

=item Examples

    use strict;
    use Data::Dumper;
    use GNOTE;
    eval {
        my $vol = NOTE::SnapMirrorUtil->new()->create_volume(tag=>'src');
        my @wex = $vol->wex_inode(inode=>'64p-100p');
        print "Inode WEX:\n". Dumper(\@wex);
    };
    logresult('FAIL',$@) if $@;
    NOTE::teardown;

=back

=cut

sub wex_inode {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();
    $args{inode}  ||= '64';
    my %volwex = $self->wex_volume();
    my $nodeinfo = "$volwex{dsid}.$args{inode}";
    $nodeinfo = "\'($volwex{dsid}).$args{inode}\'" if $args{parent};
    return $self->_wex_volume_explore("wex inode $nodeinfo -dump all" );
}

=head2 ::wex_dir Get WEX details for directory or file

Returns array of hashes with WAFL info. Following keys are returned:

I<dsid>, I<snapid>, I<parent>, I<vbn>, I<pvbn>, I<fbn>, I<location>, I<index>,
I<inum>, I<generation>, I<size>, I<flags>, I<version>, I<name>, I<shortname>,
I<altname>, I<epoch>, I<dvmask>, I<type>, I<used>,

=over

=item Arguments

B<path> = I<string>

I<Optional>

File or directory path relative to volume. Default is '/'. Path can also
be an inode number of directory or file.
Refer to http://wikid.netapp.com/w/WEX for more information

B<parent> = I<boolean>

I<Optional>

Use parent container if parent is TRUE.

=item Examples

    use strict;
    use Data::Dumper;
    use GNOTE;
    eval {
        my $vol = NOTE::SnapMirrorUtil->new()->create_volume(tag=>'src');
        my @wex = $vol->wex_dir();
        print "Dir WEX:\n". Dumper(\@wex);
    };
    logresult('FAIL',$@) if $@;
    NOTE::teardown;

=back

=cut

sub wex_dir {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $args{path}   ||= '/';
    my %volwex = $self->wex_volume();
    my $nodeinfo = "$volwex{dsid}.$args{path}";
    $nodeinfo = "\'($volwex{dsid}).$args{path}\'" if $args{parent};
    return $self->_wex_volume_explore("wex dir $nodeinfo -dump all" );
}


# Determine result of zsmcli cmd and parse and return requested field if specified

sub _zsmcli_result {
    my ( $text, $field ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "zsmcli failed. Error $text"
      unless $text =~ /results status="passed"/;

    if ($field) {
        if ( $text =~ /<$field>(.+)<\/$field>/m ) {
            return $1;
        }
        else {
            confess "Check for $field failed. Text: $text";
        }
    }
    else {
        return 1;
    }
}

# Returns extra clause in zsmcli calls for striped volumes

sub _tgt_set_clause {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return ( $self->{_type} =~ /striped/ ? " target-set=true" : "" );
}

# Internal function to cache the volume's uuid
sub _uuid {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_uuid} = $self->uuid
      unless exists $self->{_uuid};
    return $self->{_uuid};
}

=head2 ::data_protection

Returns the current data protection flag of this volume. An optional 'flag' parameter can be used to request a change.
In either case the old setting will be returned by this method.

=over

=item Arguments

B<flag> = I<true|false>

I<Optional>

Indicates to set or clear the data protection flag on this volume.

=item Examples

=back

    my $old_protect = $vol->data_protection("false"); # Turn off data protection and returns the old setting

=cut

sub data_protection {
    my ( $self, $flag ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    # Set from the RDV member for coral volumes
    return $self->Members( type => "mdv" )->data_protection($flag)
      if $self->{_args}->{'stripe-format'} =~ /data/;

    # Set from the RDV member for Acro volumes
    return $self->Members( type => "rdv" )->data_protection($flag)
      if $self->{_args}->{'stripe-format'} =~ /full/;

    my $uuid = $self->_uuid;
    my ( $output, $field ) ;
    my $mirror_type = $self->status->{type};
    if ( $mirror_type =~ /dp/i ) {
        ( $output, $field ) = ( "", "is-data-protection" );
    } else {     # if the mirror type is LS you need to change/get the is-load-sharing flag
        ( $output, $field ) = ( "", "is-load-sharing" );
    }
    my $vol_info = "id=$uuid desired-attrs=mirror-types-$field ";

    # Save old flag to return
    my $old_flag = _zsmcli_result( $self->server->shell( cmd => "zsmcli d-volume-list-info " . $vol_info ), $field );

    # Set new flag if different from current flag
    if ( $flag && $flag !~ /$old_flag/i ) {
        NOTE::logcomment( "For " . $self->description . ", set $field to $flag" );
        _zsmcli_result(
            $self->server->shell(
                cmd => "zsmcli d-volume-set-info " . $vol_info . "volume-attrs=\"[d-volume-info=[mirror-types=[d-mirror-types-info=[$field=$flag]]]]" . $self->_tgt_set_clause() . "\""
            )
        );
    }

    # Return old flag
    return $old_flag;
}

=head2 ::snapshot_id

Returns the current redirect-snapshot id for this volume. An optional 'id' parameter can be used to request that the redirect-snapshot id
be changed. In either case the old redirect-snapshot id will be returned by this method.

=over

=item Arguments

B<ssn_id> = I<integer>

I<Optional.>

New redirect-snapshot id to use.

=item Examples

=back

    my $old_ssn_id = $vol->snapshot_id(5); # Sets redirect-snapshot id to 5 and returns the old one.

=cut

sub snapshot_id {
    my ( $self, $ssn_id ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Set from the MDV member for Coral volumes
    return $self->Members( type => "mdv" )->snapshot_id($ssn_id)
      if $self->{_args}->{'stripe-format'} =~ /data/;

    # Set from the RDV member for Acro volumes
    return $self->Members( type => "rdv" )->snapshot_id($ssn_id)
      if $self->{_args}->{'stripe-format'} =~ /full/;

    my $uuid = $self->_uuid;
    my ( $output, $field ) = ( "", "redirect-snapshot-id" );
    my $vol_info = "id=$uuid desired-attrs=$field ";

    my $old_ssn_id = _zsmcli_result( $self->server->shell( cmd => "zsmcli d-volume-list-info " . $vol_info ), $field );

    # Set new snapshot_id if different to current snapshot id
    if ( $ssn_id != $old_ssn_id ) {
        NOTE::logcomment( "For " . $self->description . ", set $field to $ssn_id" );
        _zsmcli_result( $self->server->shell( cmd => "zsmcli d-volume-set-info " . $vol_info . "volume-attrs='[d-volume-info=[$field=$ssn_id]]'" . $self->_tgt_set_clause() ) );
    }

    # Return old redirect-snapshot id
    return $old_ssn_id;
}

=head2 ::inject_error

Inject error into volume's file system. Used to cause the dst of a snapmirror relationship to be
different from the src.

When calling this method for a snapmirrored volume, the script needs to perform a
'null' snapmirror update immediately upon return.

Four categories of error injection is supported:

inode - for inode entry change

data - for file data block corruption

hole - to punch a hole in a file. Only supported for flex volumes

dir - directory entry (For 'dir' entries, volume mounted r/o will cause an error)

=over

=item Arguments

B<inode> => I<change inode commands>

I<Optional>

I<path>[:<dsid>]  = path to file to change inode info, Optional dsid for a specific vol member in striped volume

A semi-comma separated list of commands, each consisting of a file path and a space separated list of
inode change commands to 'wex inode' interface. The method 'wex_info', above, can be use to obtain a
list of fields that can be use by this interface to change the inode.

B<data> => I<data change parameters>

I<Optional>

A semi-comma separated list of commands, each consisting of a file path followed by upto 3 integers:

I<b> => I<byte offset>

I<Optional>

Integer value to specify the offset within the file to begin. Default: 1

I<l> => I<length>

I<Optional>

Integer value to specify the number of bytes to change. Default: 1

I<v> => I<value>

I<optional>

Char value to be used to fill the range. Default: 97, character 'a'

B<hole> => I<punch hole commands>

I<Optional>

A semi-comma separated list of commands, each used to punch a hole in a file.
Each entry consists of:

I<path> = path to file to punch hole in

I<-o> => I<start fbn>  - Integer to indicate the starting block for hole punch. Value starts from 0.

I<-s> => I<fbn count> - Integer to indicate number of fbn blocks to punch holes for.

Example: hole => "/dir/hole_f -o 1 -s 2" - which means punch a 2 block hole in file /dir/hole_f starting from fbn 1

B<dir> => I<dir entry commands>

I<Optional>

A semi-comma separated list of commands, each consisting of specification to add or delete a directory entry.
Each entry consists of:

I<path> = directory entry path to affect

I<action> = <add> or <delete>

I<nfs command> = Optional NFS command to execute. Default: For 'add' is 'touch', for 'delete' is 'unlink'.

=item Examples

=back

    my $sm = NOTE::SnapMirror->initialize(src=>$src, dst=>$dst);
    $src->inject_error( inode => "/a/file uid=6 gid=7;"         # Change uid and gid
                               . "/second_file size=4096"       # Change size
                      , data  => "/data_file -b 5 -l 10 -v 120" # Set bytes offset 5 for 10 bytes for file /data_file to char 'x'
                      , hole  => "/dir/hole_f -o 0 -s 1"        # Punch a hole i block 0 of file hole_f (Only supported for flex volumes)
                      , dir   => "/a/file delete;"              # Remove directory entry for file /a/file
                               . "/a/new_file add"              # Add a new file new_file to directory /a
                      );
    $sm->update(); # Null snapmirror update is required

=cut

sub inject_error {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $flag   = $self->data_protection("false");
    my $ssn_id = $self->snapshot_id(0);

    my $was_mounted = $self->mounted;    # Save mount state
    foreach my $type ( sort keys %args ) {

        # Only mount volume if not mounted and running 'dir' commands as the workaround until dir add/delete is supported by wex
        $self->mount()
          if $type =~ /dir/ && !$was_mounted;
        foreach my $cmd ( split ";", $args{$type} ) {
            NOTE::logcomment("Execute Type $type, Cmd $cmd:");
            eval {
                my ( $path, @chg_list ) = split " ", $cmd;
                if ( $type =~ /inode/ ) {
                    $self->_mod_inode( $path, @chg_list );
                }
                elsif ( $type =~ /data/ ) {
                    $self->_block_cmd( $path, @chg_list );
                }
                elsif ( $type =~ /hole/ ) {
                    $self->_hole_cmd( $path, @chg_list );
                }
                elsif ( $type =~ /dir/ ) {
                    $self->_mod_dir( $path, @chg_list );
                }
                else {
                    confess "Invalid type $type";
                }
            };
            confess "Type $type, Cmd $cmd Failed. Error: $@"
              if $@;
        }
        $self->umount
          if $self->mounted && !$was_mounted;
    }

    $self->snapshot_id($ssn_id);
    $self->data_protection($flag);
}

# Modify inode data for list of files

sub _mod_inode {
    my ( $self, $path, @chg_list ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( $path =~ /:/ ) {
        my @file_part = split( ":", $path );
        my ($inode) = $self->wex_info( path => $file_part[0], column => "inum" );    # Returns the inode number of the file
        $path = "$file_part[1]." . "$inode";
    }
    else {
        $path = $self->_wex_path($path);
    }
    my $output = $self->server->shell( cmd => "wex inode $path -set " . ( join " ", @chg_list ) );
    confess $output
      unless $output =~ /inode(s*) changed/i;
}

# Sets a byte range in a path to a specific byte value

sub _block_cmd {
    my ( $self, $path, @cmd ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $path = $self->_wex_path($path);
    my %args = {
        'b' => 1    # Default start byte
        , 'l' => 1  # Default length
        , 'v' => 97 # Default to character 'a'
    };
    while ( my $opt = shift @cmd ) {
        if ( $opt =~ /^-(b|l|v)$/ ) {
            $args{$1} = shift @cmd;
        }
        else {
            confess "Invalid option $opt";
        }
    }
    my $output = $self->server->shell( cmd => "wex raw $path\@" . $args{b} . ":" . $args{l} . " -fill " . $args{v} );
    confess $output if $output =~ /error/;
    confess $output unless $output =~ /found .* to be inode/i;
}

# Obtain inode info (inum, generation #, fsid and type) for a given path in the volume

sub _get_key {
    my ( $self, $path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my @ssn_list = $self->snapshot_list();
    my %slist = map { $_ => 1 } @ssn_list;
    my ( $ssn, @rest ) = split "/", $path;
    if ( exists $slist{$ssn} ) {

        # Found snapshot name in path
        $path = join "/", @rest;
    }
    else {

        # Use latest snapshot name
        $ssn = pop @ssn_list;
    }
    $path = "/" . $path
      unless $path =~ /^\//;

    my $output = $self->server->dblade( cmd => "priv set test; get_inode /vol/" . $self->name . "/../$ssn$path" );
    confess "Failed to get inode info. Error $output"
      unless $output =~ /Success/;
    my %info =
      map { split(/=/) }
      split( ", ", join ", ", ( split( /\n/, $output ) )[ 3 .. 4 ] );    # The relevant information is on lines 4 and 5 only
    return ( $info{fileid}, $info{generation}, $info{fsid}, $info{wafl_type} );
}

# Modify data block or punch holes within files

sub _hole_cmd {
    my ( $self, $path, @chg_list ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "send_hole only supported for flex volumes"
      unless $self->{_type} =~ /flex/;

    my $paloma_enable_cmd = "priv set test; setflag wafl_paloma_enable 1;";
    my ( $inode, $gen, $fsid, $type ) = $self->_get_key($path);
    my $output = $self->server->cli( cmd => "run -node " . $self->Node->name . " -command \"$paloma_enable_cmd send_holes -i $inode -g $gen -f $fsid -t $type " . ( join " ", @chg_list ) . "\"" );
    confess $output
      unless $output =~ /success/i;
}

# add / delete directory entries. Expects the volume's RW-Tree to be mounted

sub _mod_dir {
    my ( $self, $path, $action, @cmd ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $nfs_cmd = join " ", @cmd;
    if ( $action =~ /add/ ) {
        $nfs_cmd = "touch " . $self->mountpoint . $path
          unless $nfs_cmd;
        my $rc = system( $nfs_cmd. " 2>&1" );
        confess "NFS Cmd: $nfs_cmd. Failed. RC: $rc"
          if $rc;
    }
    elsif ( $action =~ /delete/ ) {
        $self->nfs->unlink($path);
    }
    else {
        confess "Invalid dir action: $action. Only 'add' or 'delete' allowed";
    }
}

=head2 ::exportfs

This is placeholder for compatibility with 7-mode. Some 7-mode scripts
can be executed as is, therefore we must have this empty method. Actual
exporting is done automatically during mount and we do not need to make
explicit calls.

Method does nmothing and returns nothing.

=cut

sub exportfs {
    # no op
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

=head2 ::unexportfs

This is placeholder for compatibility with 7-mode. Some 7-mode scripts
can be executed as is, therefore we must have this empty method. Actual
clean up is done automatically nd we do not need to make explicit calls.

Method does nmothing and returns nothing.

=cut

sub unexportfs {
    # no op
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

=head2 ::size

Returns the size of the volume in KB.

=over

=item Arguments

I<size spec>

I<Optional.>
A new size to make the volume.

=item Examples

=back

    $size = $vol->size;
    $size = $vol->size( '500m' );   # 500MB returned as kilobytes
    $size = $vol->size( '+100m' );  # Increase volume size

=cut

sub size {
    my ( $self, $newsize ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $ret;

    if ($newsize) {

        # If there are no units specified, default to 'k'
        $newsize .= "k" unless ( $newsize =~ /[tbkmg]$/i );
    }

    if ($newsize) {
        if ( !$self->substant ) {
            confess "Can't change the size of a volume that doesn't yet exist";
        }
        eval {
        $self->modify_attrs( 'filesys-size-fixed' => 'false' );
        $self->modify_attrs( size => $newsize );
    };
        if ($@) {
            if ( $@ =~ /Insufficient space to enforce guarantee/ ) {
                confess "Not enough space";
            }
            confess $@;
        }
    }
    if ( !$self->substant ) {
        return $self->args->{'size'};
    }
    else {
        $ret           = $self->status->{'size'} / 1024;    # Convert to kilobytes
                                                            # Caching the size helps performance when sorting arrays of volumes by size
        $self->{_size} = $ret;
        return $ret;
    }
}

=head2 ::modify_attrs

modifies the attributes of the volume

=over

=item Arguments

I<hash containing the attributes to be modified>

=item Examples

=back

    %mod_options = (
        juntion-active => "true",
        user => "Obama",
        );

    $vol->modify_attrs(%mod_options);
    $vol->modify_attrs(size => "20g", user => "obama" );

=cut

sub modify_attrs {
    my ( $self, %changes ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $changes{vserver} = $self->Vserver->name;
    $changes{volume}  = $self->name;

    eval { $self->server->volume_modify(%changes); };
    confess "$@" if $@;
    eval { $self->server->volume_show( vserver => $self->Vserver->name, volume => $self->name, ); };
    confess "$@" if $@;
}

=head2 ::options

Translates 7-mode options command into volume_modify format and calls ::modify_attrs as is. Used
by old scripts which are written for 7-mode and being executed against GX filers.

=cut

sub options {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # fs_size_fixed => {on|off} is translated to filesys-size-fixed => {true|false}
    $args{'filesys-size-fixed'} = delete $args{fs_size_fixed} eq 'on' ? 'true' : 'false' if $args{fs_size_fixed};

    # add more compatibility checks here
    # ...

    return $self->modify_attrs( %args );
}

=head2 ::copy

Copies contents of given src volume to this volume. 'vol copy' is a replacement of the current
volume, so if the current volume exists, it will remove it first prior to invoking the copy
action

The sequence is to delete this volume first before doing a copy as ONTAP 'copy' command
creates the volume as part of the copy action. It is possible to avoid the need to delete the
volume by creating a GX::Volume object using the 'new' method, and using the resultant object
as the destination of the copy.

=over

=item Arguments

B<src> => I<NOTE::Volume>

A NOTE::Volume object representing the volume where data is to copy from.

B<background> => I<1|0>

Boolean to indicate whether action is to run in the background or not. Default: 0

=item Example

=back

    $dst_vol->copy(src=>$src_vol);

=cut

sub copy {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess 'Found job ' . $self->{_job_id} . ' Multiple concurrent volume operation not allowed'
      if exists $self->{_job_id};
    confess "Missing 'src' parameter"
      unless exists $args{src};

    NOTE::logcomment( 'Copy ' . $args{src}->description . ' to ' . $self->description . ': ' );
    my $text;
    eval {
        if ( $self->_vol_exists() ) {
            NOTE::logcomment("Remove existing volume as 'vol copy' creates a new volume:");
            $self->purge;
        }

        # volops cli changes during rollingrock. if the change is there
        # use the new calling sequence otherwise revert to the original.
        if ($self->server->has_change(change => 'volops-directory')) {
            $text = $self->server->volume_copy_start(
                vserver                 => $args{src}->Vserver->name,
                volume                  => $args{src}->name,
                'destination-volume'    => $self->name,
                'destination-aggregate' => $self->Aggregate->name,
                'foreground'            => 'false'
            );
        } else {
            $text = $self->server->volume_copy(
                vserver                 => $args{src}->Vserver->name,
                volume                  => $args{src}->name,
                'destination-volume'    => $self->name,
                'destination-aggregate' => $self->Aggregate->name,
                'foreground'            => 'false'
            );
        }
    };
    confess $@ if $@;
    $self->_processError($@);

    $self->{_job_id} = $self->Node->job_id($text);
    if ( $args{background} ) {
        $self->Node->job_wait( id => $self->{_job_id}, state => '(success|running)' );
    } else {
        $self->job_wait(%args);
        confess 'Unable to find volume ' . $self->description . "after 'vol copy'"
          unless $self->_vol_exists();
    }
}

=head2 ::move

Moves the current volume to the supplied aggregate

=over

=item Arguments

B<to_aggr> => I<NOTE::Aggregate>

A NOTE::Aggregate object to move the volume to.

B<background> => I<1|0>

I<Optional>

Boolean flag to indicate the move is to run in background. Default: 0

B<cutover-window> => I<seconds>

I<Optional>

Cutover time window in seconds.

B<cutover-attempts> => I<attempts>

I<Optional>

Number of cutover attempts.

B<cutover-action> => I<defer_on_failure|abort_on_failure|wait|force>

I<Optional>

Action to take when a volume move enters a deferral state.

B<force-cutover> => I<true|false>

I<Optional>

Indicates whether or not cutover should be forced if time guarantees cannot be
met.  If the build includes the change for cutover-action, specifying this option
will set the volume move cutover-action parameter to 'force' if force-cutover is
set to 'true'; if force-cutover is set to 'false', no value will be passed to
volume move for cutover-action and the default behavior (defer_on_failure)
will be used.

=item Examples

=back

    # Moves a volume to a new aggregate
    $vol->move(to_aggr=>$dst_aggr);

    # Moves volume member 0 to aggregate member 1
    my @member = $striped_vol->Members;
    my @aggr = $striped_vol->Aggregate->Members;
    $member[0]->move(to_aggr=>$aggr[1]);

=cut

sub move {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess 'Found job ' . $self->{_job_id} . ' Multiple concurrent volume operation not allowed'
      if exists $self->{_job_id};

    my $name;
    if ( exists $args{'destination-aggregate'} ) {
        NOTE::logcomment("WARNING: Please use 'to_aggr' instead of deprecated 'destination-aggregate' parameter");
        $name = delete $args{'destination-aggregate'};
        $self->{_to_aggr} = $name;
    } else {
        confess "Missing 'to_aggr' parameter"
          unless exists $args{to_aggr};
        $name = $args{to_aggr}->name;
        $self->{_to_aggr} = delete $args{to_aggr};
    }

    my %opts;
    $opts{vserver}                 = $self->Vserver->name;
    $opts{volume}                  = $self->name;
    $opts{'destination-aggregate'} = $name;
    my $background = NOTE::GX::background( \%args );
    $opts{foreground} = 'false';

    if ( exists $args{'cutover-window'} ) {
        $opts{'cutover-window'} = $args{'cutover-window'};
    }

    if ( exists $args{'cutover-attempts'} ) {
        $opts{'cutover-attempts'}  = $args{'cutover-attempts'};
    }

    if ( exists $args{'force-cutover'} ) {
        if ($self->server->has_change(change => 'volume-move-cutover-action')) {
            # If the build includes the change for cutover action, set the
            # cutover action to 'force' if force-cutover is set to true
            # If force-cutover is set to false, don't specify cutover-action
            # and use the default behavior (defer_on_failure).
            # This was included for backwards compatibility for existing
            # automation.
            $opts{'cutover-action'} = 'force' if ($args{'force-cutover'} =~ /true/i);
        } else {
            $opts{'force-cutover'} = $args{'force-cutover'};
        }
    }

    if ( exists $args{'cutover-action'} ) {

        confess "Invalid value '$args{'cutover-action'}' specified for argument 'cutover-action'"
            unless $args{'cutover-action'} =~ /defer_on_failure|abort_on_failure|wait|force/;

        if ($self->server->has_change(change => 'volume-move-cutover-action')) {
            $opts{'cutover-action'} = $args{'cutover-action'};
        }
    }

    NOTE::logcomment( $self->description . "\n  move from " . $self->Aggregate->name . ' to aggregate ' . $name );
    my $text;
    eval {
        if ( exists $self->{_parent} ) {
            # volops cli changes during rollingrock. if the change is there
            # use the new calling sequence otherwise revert to the original.
            if ($self->server->has_change(change => 'volops-directory')) {
                $text = $self->server->volume_member_move_start(%opts);
            } else {
                $text = $self->server->volume_member_move(%opts);
            }
        } else {
            # volops cli changes during rollingrock. if the change is there
            # use the new calling sequence otherwise revert to the original.
            if ($self->server->has_change(change => 'volops-directory')) {
                $text = $self->server->volume_move_start(%opts);
            } else {
                $text = $self->server->volume_move(%opts);
            }
        }
    };
    $self->_processError($@);

    $self->{_job_id} = $self->Node->job_id($text);
    if ($background) {
        $self->Node->job_wait( id => $self->{_job_id}, state => '(success|running)' );
    } else {
        $self->job_wait(%args);

        eval { $self->status(); };
    }
    confess "$@" if $@;
}

=head2 ::job_wait

Waits for existing background volume operation to complete.

=over

=item Arguments

I<state> => I<running|success|failure>

Which state to wait for. Default: (success|failure)

I<Optional>

B<timeout> => I<integer>

Timeout in seconds to wait for completion of move. Default: 1800 (0.5 hour)

I<Optional>

B<interval> => I<integer>

Interval in seconds to poll move job. Default: 5

B<column> => I<string>

I<Optional>

Returns the requested column. Default: state

=item Example

=back

    $vol->move(background=>1);
    $vol->job_wait();

=cut

sub job_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "No job id."
      unless exists $self->{_job_id};

    my $column   = delete $args{column}   || "state";
    my $timeout  = delete $args{timeout}  || 30 * 60;
    my $interval = delete $args{interval} || 5;

    my $exp_state = delete $args{state} || "(success|failure)";
    NOTE::logcomment( $self->description . ": Wait for volume operation to reach state: $exp_state. Job Id: " . $self->{_job_id} );
    my $actual;
    my $id = delete $self->{_job_id};    # Remove job id in case job fails
    eval { $actual = $self->Node->job_wait( id => $id, state => $exp_state, timeout => $timeout, interval => $interval, column => $column ); };
    $self->_processError($@);

    if ( $exp_state =~ /^running$/i ) {
        $self->{_job_id} = $id;          # return id as job is still running
    }
    else {
        NOTE::logcomment( $self->description . ": job " . $id . " completed" );
        my $to_aggr = delete $self->{_to_aggr};
        if ( ref($to_aggr) =~ /NOTE::.*Aggregate/ ) {
            $self->{_Aggregate} = $to_aggr;

            #            $self->{_Node}      = $to_aggr->Node;
        }
    }

    return $actual;
}

=head2 ::disk_free

Returns a list of items about the disk usage
of the volume.
The items returned,
and the order in which they are returned are:

B<total> -- the total disk space extant in the volume

B<used> -- the total disk space currently used in the volume

B<avail> -- the total disk space currently available in the volume

B<capacity> -- the percentage of the total disk space
in the volume that is currently being used

There are also methods by each of the item names
above that return only the value for that particular item.

=over

=item Arguments

None.

=item Example

=back

     ($total $used $avail $capacity) = $vol->disk_free;

=cut

sub disk_free {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $vol;

    exists $args{'type'} and $args{'space_type'} = delete $args{'type'};

    $vol = $self;
    my $status   = $self->status;
    my $total    = $status->{'total'} / 1024;
    my $used     = $status->{'used'} / 1024;
    my $avail    = $status->{'available'} / 1024;
    my $capacity = $status->{'percent-used'};
    confess $@ if $@;
    return wantarray ? ( $total, $used, $avail, $capacity ) : $total;
}

=head2 ::total

Returns the total disk space extant in the volume.

=over

=item Arguments

None.

=item Examples

=back

     $total = $vol->total;

=head2 ::used

Returns the total disk space currently used in the volume.

=over

=item Arguments

None.

=item Examples

=back

     $used = $vol->used;

=head2 ::avail

Returns the total disk space currently available in the volume.

=over

=item Arguments

None.

=item Examples

=back

     $total = $vol->avail;

=head2 ::capacity

Returns the percentage of the total disk space
in the volume that is currently being used.

=over

=item Arguments

None.

=item Examples

=back

     $capacity = $vol->capacity;

=cut

sub total {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[0];
}

sub used {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[1];
}

sub avail {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[2];
}

sub capacity {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->disk_free(@_) )[3];
}

=head2 ::disks

Returns a list of the disks
comprising the containing aggregate
of a flexible or striped volume.

=over

=item Arguments

None.

=item Example

=back

     @disks = $vol->disks;

=cut

sub disks {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->Aggregate->disks;
}

sub _junction_path {
    my ( $self, $path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $path || $self->{_path} || "/" . $self->name;
}

=head2 ::add_junction

Add a junction for this volume to the global name space of the vserver.
If volume already have a junction, it will remove it before adding the
new junction. And it also remount the nfs mount for the volume on the
client side.

=over

=item Arguments

I<path> => I<junction path>

Path on the global name space to add the junction. Default: /<volume name>

=back

=cut

sub add_junction {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $Trace->entry( msg => ">>entry" );

    my $path = $self->_junction_path( $args{path} );
    if ( $path ne $self->{_path} || !$self->{_path} ) {    # make sure the junction exists and is different to the new path
        my $was_mounted = $self->mounted;
        if ( exists $self->{_path} ) {
            $Trace->comment( msg => "Remove existing mount/junction for " . $self->name );
            $self->remove_junction();
        }

        if ( $self->{_mirror_type} ne 'LS' ) {
            NOTE::logcomment( "Add junction " . $path . " for " . $self->name );
            my %mnt_args;
            $mnt_args{active} = $args{active}
              || "true";                                   # Visible from vserver's namespace
            $mnt_args{'policy-override'} = $args{'policy-override'} || "false";
            $self->server->volume_mount(
                vserver         => $self->Vserver->name,
                volume          => $self->name,
                'junction-path' => $path,
                %mnt_args
            );
        }
        $self->{_path} = $path;

        $self->mount if (($was_mounted) && !($self->{_keepclient}));
    }
    $Trace->exit( msg => "<<exit" );
    return $path;
}

=head2 ::remove_junction

Removes the volume's junction if exists, and also unmounts the volume's
nfs mount from the client is exists.

=cut

sub remove_junction {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $Trace->entry( msg => ">>entry" );

    return unless exists $self->{_path};

    $self->umount(%args)
      if (($self->{_mntpoint}) && !($self->{_keepclient}));

    if ( $self->{_mirror_type} ne 'LS' ) {
        NOTE::logcomment( "Remove junction " . $self->{_path} . " from " . $self->name );
        eval { $self->server->volume_unmount( vserver => $self->Vserver->name, volume => $self->name ); };
        if ($@) {
            my $error    = $@;
            my $vol_type = $self->type;
            if (   ( $self->status->{type} =~ /ls/i )
                && ( $error =~ /This operation is not supported on a volume of type "LS"/ ) )
            {

                # Fixed for BURT 346957
                NOTE::logcomment("Ignoring expected LS umount error");
                return 0;
            }
            confess $self->description . ", remove junction error: $error";
        }
    }
    delete $self->{_path};
    $Trace->entry( msg => "<<exit" );

}

=head2 ::mount

NFS mounts the volume. Adds a junction to the volume on the global name space
before adding an nfs mount to this junction.

The default mount point is to the R-Tree of a volume. This can be overridden with
the RW parameter, which indicates that it should mount to the RW-Tree,
i.e. that is the path with a '.admin', which is automatically created for a
load-sharing source volume.

Actual client side mount is performed by GenericMount::mount call
Check GenericMount library for client side specific parameters. Following
arguments are passed to Generic::Mount as is:

    chmod
    fstype
    mntopts
    mountpoint
    sharing
    timeout

Following parameters are defaulted to different value before being passed to
GenericMount::mount()

    client  => 'localhost'
    mountretry  => 3
    mntopts => 'hard,intr,proto=tcp,vers=3'

Mount options C<mntopts> can be also changed by C<NOTE_NFSMOUNTOPTS> environment
variable or CLI parameter. Priority for C<mntopts> have following sequence:
If C<mntopts> is set then its used if not then CLI is checked. If CLI is empty
then defaults are used.

=over

=item Arguments

=over

=item B<mntops> => I<string>

I<Optional.>

A list of mount options, passed to the B<mount> command. For a list of
valid options, see the man pages for B<mount> and B<nfs>.

=item B<RW> => I<1|0>

I<Optional.>

Boolean to indicate mounting the RW-tree of a load sharing source. Default: 0

=item B<ip> => I<IP address>

An IP address to mount to. Default: the first working LIF for the vserver that
this volume belongs to.

=back

=item Examples

=back

    # Mount R-Tree by default
    $vol->mount( chmod => '777' );
    # Mount RW-Tree
    $vol->mount( chmod => '777', RW => 1 );

    # Enable NFSv4 on the Vserver
    $vol->Vserver->nfs( v4 => 'enable', 'v4-id-domain' => $domain );
    # Mount using NFSv4 over UDP with 30 seconds attributes cache
    $vol->mount( mntopts => 'proto=udp,vers=4,acregmin=30' );

=cut

sub mount {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $Trace->entry( msg => ">>entry" );
    my $address;
    my $path;

    $args{RW} = $args{rw}
      unless exists $args{RW};    # Convert to upper case in case of typo

    my %nfs_args = ( mountretry => 3, client => 'localhost' );    # Prepare parameters for GenericMount::mount

    $nfs_args{mntopts} = Tharn::param('NOTE_NFSMOUNTOPTS', -default, undef);
    if (!defined($nfs_args{mntopts})) {
        # default mount options as determined by burt 490749
        $nfs_args{mntopts} = 'hard,intr,proto=tcp,vers=3';
    }

    foreach (qw (chmod client fstype mntopts mountpoint mountretry sharing timeout)) {    # in separate hash %nfs_args because calls
        $nfs_args{$_} = delete $args{$_}
          if exists $args{$_};                                                            # like add_junction(%args) may have parameters
    }    # incompatible with GenericMount::mount()

    my $rw_tree = delete $args{RW} || 0;

    $self->{_RW} = {} unless $self->{_RW};

    # Unmount the client if requested to mount on the other tree
    $self->umount( client => $nfs_args{client} )
      if $self->mounted( client => $nfs_args{client} )
          && $self->{_RW}->{ $nfs_args{client} } != $rw_tree;

    unless ( $self->{_mntpoint}->{ $nfs_args{client} } ) {
        $self->add_junction(%args);

        $path = $self->{_path};    # set default to regular volumes
        $Trace->debug( msg => "Default GNS path is set to: $path" );

        # Mount the RW-Tree of a load sharing mirror is specified.
        # Does not need to add this junction as it is automatically added for a load sharing source
        $path = "/.admin" . $path if $rw_tree;

        # XXX Load sharing source is not needed as we provide ability to mount both RW-Tree as well as R-Tree
        # XXX Left for backward compatibility
        # For load sharing source, the path changes as we want rw access
        # As an example: mount dpg-gx27:/.admin/lsm_src
        if ( $self->{_loadsharing_src} ) {
            NOTE::logcomment("WARNING: '_loadsharing_src' is a deprecated feature. Please use 'RW' parameter to mount the RW-Tree");

            # Since we are specifying the mountpoint, we need to build the complete path
            #            $nfs_args{mountpoint} = "/t"."/".$self->Vserver->lif_address()."/".$self->name."/loadshare_source";
            #            $nfs_args{sharing} = "exclusive";
            $path = "/.admin" . "$path";
            $Trace->debug( msg => "load share source path: $path" );
        }

        # XXX New LS UI means that 'mirror-type' is never going to be LS, leaving this for backward compatibility
        if ( $self->{_mirror_type} eq 'LS' ) {
            NOTE::logcomment("WARNING: 'LS' mirror type is a deprecated feature, please use 'DP'");

            # If this is an LS mirror, then we need to find out the
            # the load sharing source and mount that volume.
            my $vol_status = $self->server->volume_show(
                volume  => $self->name,
                vserver => $self->Vserver->name,
            );

            $path = "/" . $vol_status->[0]{'load-sharing-source'};
            $Trace->debug( msg => "GNS path changed to $path" );
        }

        # BURT 446633 - Only select DATA LIFs with NFS protocol
        my $ip;
        if( exists $args{ip} ){
            $ip = delete $args{ip};
        } else {
            # First try with RR specific command
            my $eval_ret = eval {
                                    $ip = $self->Vserver->lif_address('data-protocol' => 'nfs');
                                    1;
                                 };
            # backward compatibility with BR builds
            if ( !$eval_ret ){
                $ip = $self->Vserver->lif_address();
            }
        }

        if ( $nfs_args{client} eq "localhost" ) {
            $nfs_args{sharing} = "exclusive"
              unless exists $nfs_args{sharing};

            # Specify special mountpoint for localhost to distinguish it from client that is the same as localhost
            $nfs_args{mountpoint} = "/t/localhost.$$/$ip/$path"
              unless exists $nfs_args{mountpoint};

            # Specifying mountpoint make it exclusive, i.e. not shareable
            # So explicitly umount in case mountpoint was left behind from a previous process - ignore any errors
            eval { GenericMount::umount( $nfs_args{mountpoint} ); };
        }
        my $msg = "Mounting " . $self->description . ", path: $path to IP $ip on client " . $nfs_args{client};
        NOTE::logcomment($msg);
        $self->{_mntpoint}->{ $nfs_args{client} } = GenericMount::mount( $ip, $path, %nfs_args );
        confess $@ if $@;
        confess "No mountpoint returned from $msg"
          unless $self->{_mntpoint}->{ $nfs_args{client} };
        $self->{_RW}->{ $nfs_args{client} } = $rw_tree;    # Record Tree client is mounted on
    }
    $Trace->exit( msg => "<<exit" );
    return $self->{_mntpoint}->{ $nfs_args{client} };
}

# Replace mount points of this object with the supplied object.
# An internal method to be call be SnapMirror::promote method

sub _replace_mountpoint {
    my ( $self, $src ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # First refresh junction and share from ONTAP
    $self->_refresh_junction_and_share();

    # Second replace the mount points from src volume
    if ( exists $src->{_mntpoint} ) {
        $self->{_mntpoint} = $src->{_mntpoint};
    }
    else {
        delete $self->{_mntpoint}
          if exists $self->{_mntpoint};
    }
}

=head2 ::umount

Unmounts the volume from all clients. Or just from specific client if client name is provided

=over

=item Arguments

B<client>

I<Optional.>

Host name of the client to unmount

=item Example

=back

     $vol->umount( client=>'localhost' );   # just unmount localhost

     $vol->umount;                          # unmount everything

=cut

sub umount {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    $Trace->entry( msg => "entry" );

    # Return if nothing to umount
    return 0 unless exists $self->{_mntpoint};

    my @clients = ( exists $args{client} ? $args{client} : keys %{$self->{'_mntpoint'}} );
    foreach my $client (@clients) {
        $args{client} = $client;
        if ( $self->{_mntpoint}->{$client} ) {
            NOTE::logcomment( 'Unmounting ' . $self->description . " from client $client" );
            if (GenericMount::umount( $self->{_mntpoint}->{$client}, %args )) {
                delete $self->{_mntpoint}->{$client};
                return 1;
            }
            confess ref($self) . ":umount: $@"
                if $@;
        }
    }
    $Trace->exit( msg => "<<exit" );
    return 0;
}

=head2 ::mount_path

Returns the junction path name of the volume.

=over

=item Arguments

None.

=item Example

=back

     $mntpt = $vol->mount_path;

=cut

sub mount_path {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->add_junction()
      unless $self->{_path};

    return $self->{_path};
}

=head2 ::create_share

Create a share for this volume and returns the share name within the volume.

=over

=item Arguments

B<path> => I<path>

The path of the share is created from. Default: Junction of the volume

B<RW> => I<1|0>

Use the RW-Tree or R-Tree of the volume. Default: 0 i.e. the R-Tree

=item Examples

=back

    my $rtree_share = $vol->create_share(); # Create a share for the R-Tree
    my $rwtree_share = $vol->create_share(RW=>1); # Create a share for the RW-Tree

=cut

sub create_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $share;
    my $path;

    if ( exists $self->{_share} ) {
        NOTE::logcomment( $self->description . ": share " . $self->{_share} . " already created." );
    }
    else {
        $share = $self->name;

        if ( $args{path} ) {
            $path = $args{path};
        }
        else {
            $self->add_junction()
              unless exists $self->{_path};
            if ( delete $args{RW} ) {
                $path = "/.admin" . $self->{_path};
            }
            else {
                $path = $self->{_path};
            }
        }

        # setup cifs server
        $self->Vserver->cifs_setup( storage => $self );

        NOTE::logcomment( "CIFS share path will be: ", $path );

        NOTE::logcomment( $self->description . ": Create share" );
        $self->server->vserver_cifs_share_new(
            vserver      => $self->Vserver->name,
            'share-name' => $share,
            path         => $path
        );
        $self->{_share} = $share;
    }
    return $self->{_share};
}

=head2 ::delete_share

Unmaps CIFS share from client if share was mapped and deletes share from vserver.
Returns name of share being deleted if volume was shared. If nothing was mapped
then request is ignorred.

=cut

sub delete_share {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( exists $self->{_share} ) {
        $self->unmap_share() if exists $self->{_drive};
        NOTE::logcomment( $self->description . ": Delete share" );
        $self->server->vserver_cifs_share_delete(
            vserver      => $self->Vserver->name,
            'share-name' => $self->{_share}
        );
        delete $self->{_share};
    }
    else {
        NOTE::logcomment( $self->description . ": no share to delete. Ignoring delete share request" );
    }
}

=head2 ::map_share

Map volume as CIFS share on CIFS client and returns drive letter. Also checks if client is actually
a CIFS client prior to mapping. Compares RSH and RPC environments on windows client and fails if
they mismatch. Creates user mapping between actual windows user and actual unix user in addition to
Administrator to root mapping.

If share on a volume is not yet created then create_share method is called as dependency, there is
no need to call both create_share and map_share method from script. Even more - it's not required
to have CIFS server configured because create_share method will make required calls to create one.


B<LIMITATIONS and WARNINGS:> Current design does not allow to map same volume from multiple clients.
If volume is already mapped then drive letter will be returned immediately without client side
validation!

=over

=item Arguments

B<CIFS_DOMAIN> => I<string>

I<Required>

Domain name to use for authentication. It required to have parameter for method or CIFS_DOMAIN cli
parameter of CIFS_DOMAIN_LONG cli parameter.

B<CIFS_USER> => I<string>

I<Required>

User name to use for authentication with domain server. It can be at least one of CIFS_USER as method
argument, or CIFS_USER a cli parameter.

B<CIFS_PASSWD> => I<string>

I<Required>

Password to use for authentication with domain server. It can be at least one of CIFS_PASSWD as
method argument, or CIFS_PASSWD as cli parameter.

B<client> => I<string>

I<Optional>

Client's host name. Default is localhost.


B<drive> => I<string>

I<Optional>

Drive letter for newly mounted share. Default behavior - let windows assign it automatically.
Be aware that some letters like 'X:' are reserved for NATE itself when assigning manually

=item See also

GNOTE::GX::Volume::create_share and GNOTE::GX::Vserver::cifs_setup

=back

=cut

sub map_share {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->{_drive} if( exists $self->{_drive} );

    my $cliarg = new Options; $cliarg->process();
    my $clihash = $cliarg->get_hash();

    ( my $domain = $args{CIFS_DOMAIN} =~ /\w+/ ? $args{CIFS_DOMAIN} : $clihash->{CIFS_DOMAIN_LONG} || $clihash->{CIFS_DOMAIN} ) =~ s/\..*//;    # get short form of domain from any place possible
    confess( "CIFS_DOMAIN parameter is required" ) unless( $domain =~ /\w+/ );
    my $username = $args{CIFS_USER} =~ /\w+/ ? $args{CIFS_USER} : $clihash->{CIFS_USER};                            # pull CIFS user from args or CLI
    confess( "CIFS_USER parameter is required" ) unless( $username =~ /\w+/ );
    my $password = $args{CIFS_PASSWD} =~ /\w+/ ? $args{CIFS_PASSWD} : $clihash->{CIFS_PASSWD};                          # pull password from args or CLI
    confess( "CIFS_PASSWD parameter is required" ) unless( $password =~ /\w+/ );
    my $client = $args{client} || "localhost";                                                  # set client name
    my $drive  = '*'           || $args{drive};                                                 # and drive name

    confess "Client [$client] is not a windows machine" unless( Tharn::host($client)->hosttype =~ /windows/i );                 # fail if client type is not windows
    my $winrsh = Tharn::connect('#rsh');                                                    # create RSH connection together
    my $winrpc = NOTE::Client::RPC->new( client => $client );                                           # with RPC connection to STAFF

    my $enverror = undef;                                                           # environment error counter
    my %rshdata = map {split('=', $_)} grep(/^(USER|HOMEPATH|LOGONSERVER)/, split( /[\n\r]/, join('',Tharn::say($winrsh, "set"))));                     # pull ENV from RSH
    my %rpcdata = map {split('=', $_)} grep(/^(USER|HOMEPATH|LOGONSERVER)/, split( /[\n\r]/, $winrpc->{_handle}->perl_eval(code=>'GNOTE::Client::Windows::showenv();')));   # and from RPC and
    $winrpc->DESTROY();                                                             # destroy RPC server, don't leave zombies

    map { $enverror++ if $rshdata{$_} ne $rpcdata{$_} } ( keys %rshdata, keys %rpcdata );                           # compare both environments and
    confess "STAF/RSH environment mismatch\n". Dumper( \%rshdata, \%rpcdata ) if $enverror;                         # fail if STAF/RSH data do not match each other

    $self->create_share(); # call create share to export volume, which will also
                           # create CIFS server and adiminstrators user maps

    my ( %usermap, $maxid );                                                        # user map list and max position holder
    map{ $usermap{ $$_{'direction'} }->{$$_{'pattern'}} = $$_{'replacement'} } $self->server->vserver_name_mapping_show( vserver=>$self->Vserver->name );   # create hash of user maps
    map{ $maxid = $$_{'position'} > $maxid ? $$_{'position'} : $maxid } $self->server->vserver_name_mapping_show( vserver=>$self->Vserver->name );      # and calculate max position which must
    $maxid++;                                                                       # be a bit larger then found on filer

    unless( $usermap{'unix-win'}->{$ENV{USER}} eq "$rshdata{'USERDOMAIN'}\\\\$rshdata{'USERNAME'}" ) {                      # create unix to windows mapping
    $self->server->vserver_name_mapping_create(                                     # if such mapping does not exist yet
                            position=>$maxid,                                       # this is mapping between UNIX user
                            vserver=>$self->Vserver->name,                          # which actually running ntest and
                            direction=>'unix-win',                                  # RSH/STAFF user on windows client
                            pattern=>$ENV{USER},
                            replacement=>"$rshdata{'USERDOMAIN'}\\\\$rshdata{'USERNAME'}"
                        );
    }

    unless( $usermap{'win-unix'}->{"$rshdata{'USERDOMAIN'}\\\\$rshdata{'USERNAME'}"} eq $ENV{USER} ) {                      # same mapping should be done for
    $self->server->vserver_name_mapping_create(                                             # windows to unix mapping
                            position=>$maxid,
                            vserver=>$self->Vserver->name,
                            direction=>'win-unix',
                            replacement=>$ENV{USER},
                            pattern=>"$rshdata{'USERDOMAIN'}\\\\$rshdata{'USERNAME'}"
                        );
    }

    my $mapped_drive = MapShare::map_cifs_share(
                                    server   => $self->Vserver->lif_address(),
                                    client   => $client,
                                    share    => $self->{_share},
                                    drive    => $drive,
                                    domain   => $domain,
                                    username => $username,
                                    password => $password
                                );
    confess "map_share error: $@"
      if $@;
    $self->{_cifs_client} = $client;
    $self->{_drive}       = $mapped_drive;
    NOTE::logcomment( $self->description . ": Mapped share to drive " . $self->{_drive} . " on client " . $self->{_cifs_client} );
    return $self->{_drive};
}

=head2 ::unmap_share

Unmap CIFS client if it was mapped. Returns client name share was unmounted from.

=cut

sub unmap_share {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if ( exists $self->{_drive} ) {
        NOTE::logcomment( $self->description . ": Unmap share from drive " . $self->{_drive} );
        MapShare::unmap_cifs_share(
            drive  => $self->{_drive},
            client => $self->{_cifs_client}
        );
        delete $self->{_drive};
        delete $self->{_cifs_client};
    }
    else {
        NOTE::logcomment( $self->description . ": not mapped... Ignoring request to unmap." );
    }
}

=head2 ::Cluster

Returns the GX::Cluster object associated with this Volume

=over

=item Arguments

B<None>

=item Example

=back

    my $clu = $vol->Cluster;

=cut

sub Cluster {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Cluster}
}

=head2 ::Node

Returns the GX::Node object associated with this Volume

=over

=item Arguments

B<None>

=item Example

=back

    my $node = $vol->Node;

=cut

sub Node {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # If Aggregate object exists use it otherwise use self
    # Aggregate object will always exists for non-trad volumes
    return ( exists $self->{_Aggregate} ? $self->Aggregate->Node : $self->{_Node} );
}

=head2 ::server

Returns 'Server.pm' server object assosiated with cluster.
See Server.pm documentation for methods and examples. Used
for direct access to node itself.

=over

=item Example

 $vol->server->system_node_show(fields=>'serialnumber');

=back

=cut

sub server {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->Node->server;
}

=head2 ::Vserver

Returns the GX::Vserver object associated with this Volume

=over

=item Arguments

B<None>

=item Example

=back

    my $vsrv = $vol->Vserver;

=cut

sub Vserver {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->{_Vserver}
}

=head2 ::msid

Return the volume's MSID.

=over

=item Arguments

B<None>

=item Examples

=back

    my $vol_msid = $vol->msid();

=cut

sub msid {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_msid} = $self->status("msid")
      unless exists $self->{_msid};

    return $self->{_msid};
}

=head2 ::fsid

Returns the Volume's FSID (which is MSID).

=over

=item Arguments

B<None>

=item Examples

=back

    my $vol_fsid = $vol->fsid();

=cut

sub fsid {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->msid;
}

sub mirror {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
}

=head2 ::hostpath

Returns the name of the volume
pre-pended by the containing cluster and vserver name.

=over

=item Arguments

None.

=item Example

=back

     $hp = $vol->hostpath;

=cut

sub hostpath {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return $self->Cluster->name . "://" . $self->Vserver->name . "/" . $self->name;
}

=head2 ::fullname

Returns the same thing as hostpath method, namely the volume name
prepended by the containing cluster and vserver name.

=over

=item Arguments

None.

=item Example

=back

     $hp = $vol->fullname;

=cut

sub fullname {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->hostpath;
}

=head2 ::offline, ::online, ::restrict

Sets volume offline, online and restrict respectively

=over

=item Arguments

None

=item Examples

=back

     $vol->offline;

     $vol->online;

     $vol->restrict;

=cut

sub offline {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Volume must be unmounted before it can be offlined
    $self->remove_junction;
    eval { $self->server->volume_modify( volume => $self->name, state => 'offline', vserver => $self->Vserver->name, ); };
    confess $@ if $@;

    $self->{_sm_set}->_suspend( $self, 1 )
      if exists $self->{_sm_set};

    return $self;
}

sub online {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    eval { $self->server->volume_modify( volume => $self->name, vserver => $self->Vserver->name, state => 'online', ); };
    confess $@ if $@;

    $self->{_sm_set}->_suspend( $self, 0 )
      if exists $self->{_sm_set};

    return $self;
}

sub restrict {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Wrap the CDEFS call in an eval so that we can use confess and get a
    # full stack-trace on exits. The "1;" at the end of the eval lets me
    # test whether the eval succeeded, as testing $@ by itself is not
    # reliable.
    my $result = eval {
        $self->server->volume_modify(
            vserver => $self->Vserver->name,
            volume  => $self->name,
            state   => 'restricted'
        );
        1;
    };
    if (! $result) {
        confess $@;
    }

    return $self;
}

=head2 ::restricted

Returns a boolean indicating whether a volume is restricted or not

=over

=item Arguments

B<None>

=item Examples

=back

      $vol->restricted;

=cut

sub restricted {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( shift->status("state") =~ /restricted/i );
}

=head2 ::writeable

Returns a boolean indicating whether a volume is writeable,
i.e. is online and of type 'RW'.

=over

=item Arguments

B<None>

=item Examples

=back

    my $flag = $vol->writable();

=cut

sub writeable {
    my $s = shift->status;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ( $s->{'state'} =~ /online/ && $s->{'type'} =~ /RW/ );
}

=head2 ::readable

Returns boolean indicating whether the volume is readable,
i.e. online.

=over

=item Arguments

B<None>

=item Examples

=back

    my $flag = $vol->readable();

=cut

sub readable {
    my $s = shift->status;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return ( $s->{'state'} =~ /online/ );
}

=head2 ::snapshot_create

Create a snapshot on the volume

=over

=item Argument

B<snapname> - I<name>

Optional

If not set then anonymous name is used to create the snapshot with

=item Examples

=back

     $vol->snapshot_create;

     $vol->snapshot("special");

=cut

sub snapshot_create {
    my ( $self, $snapname, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $snapname = $self->newName("sn") unless ($snapname);

    NOTE::logcomment( "Creating snapshot " . $snapname . " of " . $self->description );
    my $r;
    eval { $r = $self->server->volume_snapshot_new( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapname, %args, ); };
    if( $@ && $@ =~ /could not find the command definition for volume_snapshot_new/si ) {
	eval { $r = $self->server->volume_snapshot_create( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapname, %args, ); };
    }
    confess $@ if ($@);
    return $snapname;
}

=head2 ::snapshot_restore

Restored a named snapshot

=over

=item Arguments

B<snapname> - I<name>

Name of the snapshot to restore

=item Example

=back

     $vol_snapshot_restore("special");

=cut

sub snapshot_restore {
    my ( $self, $snapname, %opts ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Need snapshot name" unless ($snapname);
    NOTE::logcomment( "Restoring " . $self->description . " using snapshot " . $snapname );

    eval { $self->server->volume_snapshot_promote( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapname, %opts, ); };
    confess $@ if ($@);
    return $self;
}

=head2 snapshot_rename

Rename a snapshot on the volume

=over

=item Arguments

B<snapname> = I<name>

Name of the snapshot to be rename.


B<newname> = I<new name>

I<Optional>

New name of the snapshot.

=item Examples

=back

     $vol->snapshot_rename("oldsnapshot, "newsnap");

     $vol->snapshot_rename("oldsnapshot");

=cut

sub snapshot_rename {
    my ( $self, $snapname, $newname ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $newname = $self->newName("sn") unless ($newname);

    NOTE::logcomment( "Renaming snapshot " . $snapname . " to " . $newname . " of " . $self->description );
    my $r;
    eval { $r = $self->server->volume_snapshot_rename( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapname, 'new-name' => $newname, ); };
    confess $@ if ($@);

    return $newname;
}

=head2 ::snapshot_list

Returns an array of the snapshots that exists on this volume

=over

=item Arguments

None.

=item Example

=back

     $vol->snapshot_list;

=cut

sub snapshot_list {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $snaplist;

    eval { $snaplist = $self->server->volume_snapshot_show( vserver => $self->Vserver->name, volume => $self->name, %args, ); };
    confess $@
      if $@ && !( $@ =~ /there are no entries matching your query/i );

    my @list = ();
    for my $snapshot ( @{$snaplist} ) {
        push @list, $snapshot->{snapshot};
    }
    return @list;
}

=head2 ::snapshot_status

Returns a reference to an array of hashref of snapshot status

=over

=item Arguments

None

=item Example

=back

     $vol->snapshot_status;

=cut

sub snapshot_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $status;

    eval { $status = $self->server->volume_snapshot_show( vserver => $self->Vserver->name, volume => $self->name, %args, ); };
    confess $@
      if $@ && !( $@ =~ /there are no entries matching your query/i );

    return $status;
}

=head2 ::snapshot_exists

Returns a boolean to indicate whether a given named snapshot exists or not

=over

=item Arguments

B<snapshot> - I<name>

Name of the snapshot to check

=item Example

=back

     $vol->snapshot_exists('special');

=cut

sub snapshot_exists {
    my ( $self, $snapshot ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    confess "Missing snapshot name..."
      unless defined $snapshot;

    my $ret;
    eval { $ret = $self->server->volume_snapshot_show( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapshot ); };
    if ($@) {
        if ( $@ =~ /there are no entries matching your query/i ) {
            return 0;
        }
        else {
            confess "Snapshot $snapshot, Error: $@";
        }
    }

    return (
        $ret->[0]->{snapshot} =~ /$snapshot/
        ? 1
        : confess "Found different snapshot. Search for $snapshot, found " . $ret->[0]->{snapshot}
    );
}

=head2 ::snapshot_delete

Delete a named snapshot

=over

=item Arguments

B<snapshot> - I<name>

Name of the snapshot to delete

=item Example

=back

      $vol->snapshot_delete('special');

=cut

sub snapshot_delete {
    my ( $self, $snapshot, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    NOTE::logcomment( "Deleting snapshot " . $snapshot );

    eval { $self->server->volume_snapshot_delete( vserver => $self->Vserver->name, volume => $self->name, snapshot => $snapshot, %args, ); };
    confess $@ if ($@);
    return $self;
}

sub _vol_exists {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return eval { $self->status() };
}

sub _status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return eval {
        if ( exists $self->{_parent} )
        {
            return $self->server->volume_member_show(%args);
        }
        else {
            return $self->server->volume_show(%args);
        }
    };
}

sub _vol_aggr {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    shift->aggregate;
}

sub _vol_create {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $r, $path );
    my $type = 'RW';
    my $size = delete $args{size};

    my %opts = (
        policy => 'default',

        # Commenting these out as a workaround for TESTLIB burt 277246
        # user             => 'root',
        # group                => 'root',
        'security-style'   => 'mixed',
        'unix-permissions' => '---rwxrwxrwx',
        'space-guarantee'  => 'volume',
        'snapshot-policy'  => ( $args{'mirror-type'} =~ /(DP|LS)/ ? "none" : 'default' )
    );
    delete $opts{'unix-permissions'}
      if $args{'mirror-type'} =~ /(DP|LS)/;

    if (   ( ( $self->{_type} eq 'striped' ) || ( $self->{_type} eq 'flex' ) )
        && ( ( exists $args{'coral-type'} )  || ( exists $args{'flex-type'} ) )
        && ( ( defined $args{'coral-type'} ) || ( defined $args{'flex-type'} ) )
        && ( ( $args{'coral-type'} eq 'DP' ) || ( $args{'flex-type'} eq 'DP' ) ) )
    {
        $type = $args{'flex-type'}  if ( defined $args{'flex-type'} );
        $type = $args{'coral-type'} if ( defined $args{'coral-type'} );
        $self->{_mirror_type} = $type;
        delete $args{'coral-type'} if ( defined $args{'coral-type'} );
        delete $args{'type'}       if ( defined $args{'flex-type'} );
    }

    if ( ( exists $args{'mirror-type'} ) && ( defined $args{'mirror-type'} ) ) {
        $type = $args{'mirror-type'};
        NOTE::logcomment("Setting _mirror_type to $type");
        $self->{_mirror_type} = $type;
        delete $args{'mirror-type'};
    }

    my $maxfiles;
    if ( exists $args{'files'} && $args{'files'} ) {
        $maxfiles = $args{'files'};
        delete $args{'files'};
    }
    foreach my $keys ( keys %args ) {
        $opts{$keys} = $args{$keys};
    }

    # For LS mirrors junction is not allowed nor for DP mirrors until intialized
    #    if ( $type ne 'LS' && $type ne 'DP' ) {
    #       $path = $self->_junction_path($args{path});
    #       $opts{'junction-path'} = $path;
    #    }
    if ( $type ne 'LS' && $type ne 'DP' ) {
        $path = $self->_junction_path( $args{path} );
        $opts{'junction-path'} = $path;
    }
    elsif ( $type eq 'LS' ) {

        # Save the source name so that we can mount using the source volume name
        # If type is LS and no load-sharing-source has been passed, the volume
        # will fail.
        $self->{_loadshare_src} = $opts{'load-sharing-source'};
    }
    $opts{'-Timeout'} = 108000;

    my $level;
    if ( exists $opts{'stripe-format'} ) {

        # diag mode is needed to create striped volumes
        $level = "diag";
        $self->{_coral} = ( $opts{'stripe-format'} =~ /data/ );
    }
    else {
        $self->{_is_coral} = 0;
    }

    # Need test mode for striped DP volumes for TRICKY_PARITY mode.
    $level = "test"
      if $self->{_type} =~ /striped/i
          && $type =~ /dp/i
          && $self->Node->tricky_parity();
    $self->server->set( privilege => $level, confirmations => "off" )
      if $level;

    eval {
        $r = $self->server->volume_new(
            vserver   => $self->Vserver->name,
            volume    => $self->name,
            aggregate => $self->Aggregate->name,
            state     => 'online',
            size      => $size . 'k',
            'type'    => $type,
            %opts
        );
    };

    if( $@ && $@ =~ /could not find the command definition for volume_new/si ) {    # BURT625461
	eval {
	    $r = $self->server->volume_create(
		vserver   => $self->Vserver->name,
		volume    => $self->name,
		aggregate => $self->Aggregate->name,
		state     => 'online',
		size      => $size . 'k',
		'type'    => $type,
		%opts
	    );
	};
    }
    if ($@) {
        if ( $@ =~ /Insufficient space to enforce guarantee/ ) {
            confess "Not enough space";
        }
        confess $@;
    }

    # Trun off test mode after creation of Coral volumes
    $self->server->set( privilege => "diag", confirmations => "off" )
      if $level;

    # XXX Pull this piece out inside a separate method vol_modify
    if ($maxfiles) {

        # modify fails saying 'vol is offline' if modify is done immediately after
        # creation of a coral volume
        sleep 10;
        eval { my $r = $self->server->volume_modify( vserver => $self->Vserver->name, volume => $self->name, files => $maxfiles, ); };
        if ($@) {
            confess $@;
        }
    }

    # Fix found by JR: For LS & DP volumes, $path is not set. Setting the
    # _path key to this undef value was breaking the mounting of these vols
    # under CIFS. So only set _path for RW vols.
    if ( $type ne 'LS' && $type ne 'DP' ) {
        $self->{_path} = $path;
    }

    return;
}

sub _vol_type {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return shift->volume_style;
}

sub _vol_rename {
    my ( $self, $newname ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    eval { $self->server->volume_rename( vserver => $self->Vserver->name, volume => $self->name, newname => $newname ); };
    confess $@ if $@;
    $self->{_name} = $newname;
}

sub _vol_destroy {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    eval {
        $self->delete_share()
          if exists $self->{_share};
        $self->server->volume_delete(
            vserver => $self->Vserver->name,
            volume  => $self->name,
            %args,
        );
    };
    $self->_processError($@);
    NOTE::logcomment( $self->description . " destroyed" );
}

sub snaplock {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return 0;
}    # not supported in GX

=head2 ::set_loadsharing_src_flag

Deprecated method. Used by LS mirror to mark object belongs to
load sharing. As of now it returns following warning:

 WARNING: '_loadsharing_src' is a deprecated feature. Please use 'RW' parameter to mount the RW-Tree

=cut

sub set_loadsharing_src_flag {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    $self->{_loadsharing_src} = 1;
}

=head2 ::wafliron_start

Start Wafliron on the corresponding aggregate of this volume. By default it
runs in the background. Set I<wait> argument to 1 if foreground execution is
required.

This function will return the result of the wafliron operation if it's being
called with I<wait> argument set to 1. The result is an array of text.

=over

=item Arguments

B<wait> = 1|0

I<Optional>

Wait for wafliron to finish and exit. By default it starts wafliron as
background process. If I<wait> is set to 1, this function will block and wait
for the wafliron operation to finish, extract the reports from
/mroot/etc/messages and /mroot/etc/mlog/messages.log files from all nodes
exercised by the wafliron operation.

B<force> = 1|0

I<Optional>

Indicate whether this function shall continue to start Wafliron operation even
though if it detect that there is/are outstanding wafliron operation/s running
in the targeted aggregate. Setting force to 1 instruct the function to ignore
any outstanding running wafliron sessions, and continue its attempt to start
the operation. Default is 1.

B<include_mirror> = 1|0

I<Optional>

Indicate whether the wafliron operation will run on aggregates with existing
mirror relationship. Default is 1, which means that it will include aggregate
with mirror relationship.

=item Examples

    Example 1 :

        (1) Start Wafliron on the volume (and thus, its corresponding aggregate)
        (2) Block while waiting for the operation to complete
        (3) Get the result from the return value.

    #
    # Code fragment start here .....
    #

    my $result = $vol->walfiron_start (wait=>1);
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("Wafliron report :\n$report");
    }

    Example 2 :

        (1) Start Wafliron on the volume without waiting
        (2) Call wafliron_wait later to wait for the completion
        (3) Get the result from wafliron_wait's return value

    #
    # Code fragment start here .....
    #

    my $status = $vol->walfiron_start (wait=>0);

    #
    # Do something else in between
    #

    my $status = $vol->walfiron_wait ();
    if (@$result) {
        my $report = join "\n", @$result;

        NOTE::logcomment ("Wafliron report :\n$report");
    }

    Example 3 :

        (1) Start Wafliron on the volume without waiting
        (2) Call wafliron_stop later to stop the operation
        (3) Get the result from wafliron_stop's return value

    #
    # Code fragment start here .....
    #

    my $status = $vol->walfiron_start (wait=>0);

    #
    # Do something else in between
    #

    my $status = $vol->walfiron_stop ();
    if (@$result) {
        my $report = join "\n", @$result;

        NOTE::logcomment ("Wafliron report :\n$report");
    }

=back

=cut

sub wafliron_start {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Aggregate->wafliron_start(%args);
}

=head2 ::wafliron_stop

Termintate outstanding Wafliron session on this Volume (and thus, the
corresponding Aggregate).

=over

=item Arguments

B<wait> = 1|0

I<Optional>

Wait for wafliron to stop and exit. If I<wait> is set to 1, this function will
block and wait for the wafliron stop operation to complete, extract information
from various sources, such as  /mroot/etc/messages and /mroot/etc/mlog/messages.log
files, from all nodes exercised by the wafliron operation. Default is 1.


=item Examples

    my $result = $vol->wafliron_stop ();
    if (@$result) {
        my $report = join "\n", @$result;

    NOTE::logcomment ("Wafliron report :\n$report");
    }

=back

=cut

sub wafliron_stop {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Aggregate->wafliron_stop(%args);
}

=head2 ::wafliron_status

Return status of the outstanding or previous Wafliron session

=over

=item Arguments

B<None>

=item Example

    eval {
        $status = $vol->wafliron_status ();
    };
    confess $@ if $@;
    if ($status->{'iron-status'} eq 'not_running') {
        #
        # Some code here
        #
    }

=back

=cut

sub wafliron_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Aggregate->wafliron_status(%args);
}

=head2 ::wafliron_wait

Wait for the outstanding Wafliron session (if exist) to finish, indicated by the "not_running" status of the session.

=over

=item Arguments

B<timeout> = I<number of seconds>

I<Optional>

The elapsed time limit in seconds after which this wait operation will be
terminated and return as error if the corresponding wafliron session is still
not finished. Default is 3600 seconds (i.e. 1 hour).

B<interval> = I<number of seconds>

I<Optional>

The number of seconds to sleep between iterations of status check on the
wafliron session. This value should be smaller than the <timeout> value above.
Default is 30 seconds.

=item Examples

    $vol->wafliron_wait ();
    $vol->wafliron_wait (timeout => 300);
    $vol->wafliron_wait (timeout => 600, interval => 30);

=back

=cut

sub wafliron_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->Aggregate->wafliron_wait(%args);
}

=head2 ::is_stripe

Return 1 if this is a Striped Volume. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($vol->is_stripe) {
        #
        # the $vol instance is a stripe volume
        #
    }

=back

=cut

sub is_stripe {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $self->{_type} =~ /striped/ ) {
        return 1;
    }

    return 0;
}

=head2 ::is_member

Return 1 if this is a Member Volume. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($vol->is_member) {
        #
        # the $aggr instance is a member volume
        #
    }

=back

=cut

sub is_member {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $self->is_stripe ) {
        if ( exists $self->{_parent} ) {
            return 1;
        }
    }

    return 0;
}

=head2 ::is_acro

Return 1 if stripe type of this volume is Acro. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($vol->is_acro) {
        #
        # the $vol instance is a of Acro stripe type
        #
    }

=back

=cut

sub is_acro {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $self->is_stripe ) {
        if ( $self->{_args}->{'stripe-format'} =~ /full/ ) {
            return 1;
        }
    }

    return 0;
}

=head2 ::is_coral

Return 1 if stripe type of this volume is Coral. Otherwise, return 0.

=over

=item Arguments

B<None>

=item Example

    if ($vol->is_coral) {
        #
        # the $vol instance is a of Coral stripe type
        #
    }

=back

=cut

sub is_coral {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $self->is_stripe ) {
        if ( $self->{_args}->{'stripe-format'} =~ /data/ ) {
            return 1;
        }
    }

    return 0;
}

=head2 ::SnapMirrorSet

Sets or gets the NOTE::SnapMirrorSet object that this volume is a destination of.
Returns undef if this volume is not part of a NOTE::SnapMirrorSet object
This method can also be used to determine if this object is part of a SnapMirrorSet object.

=over

=item Arguments

B<sms_obj> = I<NOTE::SnapMirrorSet>

I<Optional>

If passed will record this as the SnapMirrorSet object that this volume is part of.

=item Examples

=back

    $vol->SnapMirrorSet($sms);
    my $sms = $vol->SnapMirrorSet();

=cut

sub SnapMirrorSet {
    my ( $self, $sms_obj ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $old_value = $self->{_sm_set};
    $self->{_sm_set} = $sms_obj
      if defined $sms_obj;
    return $old_value;
}

# Clears the internal SnapMirrorSet Object

sub _sm_clr {
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    delete shift->{_sm_set};
}

=head2 ::qtrees

Returns array of qtree names

=over

=item Arguments

B<None>

=item Example

 my @qtrees = $vol->qtrees;

=back

=cut

sub qtrees {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my @qtrees = ();
    eval { @qtrees = $self->server->volume_qtree_show( vserver => $self->Vserver->name, volume => $self->name ); };
    confess "Failed to get list of qtrees:\n$@" if $@;
    return grep /\w+/, map { $$_{'qtree'} } @qtrees;
}

=head2 ::Qtrees

Returns array of qtree objects

=over

=item Arguments

B<None>

=item Example

 my @QT_OBJs = $vol->QTREES;

=back

=cut

sub Qtrees {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return map { NOTE::Qtree->attach( Volume => $self, name => $_ ) } $self->qtrees;
}

=head2 ::is_qtree

Check volume for qtree identified by name. Returns 1 if qtree was found or 0 if not found

=over

=item Arguments

B<qtree> I<string>

Qtree name

=item Example

 confess "Qtree not found" unless( $vol->is_qtrees(qtree=>$qtree_name) );

=back

=cut

sub is_qtree {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $found = grep /^$args{'qtree'}$/, $self->qtrees;
    return $found;
}

sub DESTROY {
    my ($self) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # delete all the flexcache volumes associated with this volume
    if (defined($self->{_flexcache_name})) {
        $self->flexcache_delete;
    }

    # delete quota policy associated with this volume
    if (defined($self->{'_quota_policy_cleanup'})) {
        for my $policy_name ( keys %{$self->{'_quota_policy_cleanup'}}) {
            $self->quota_policy_delete('policy-name' => $policy_name);
        }
    }

    $self->umount       unless $self->{_keepclient};
    delete $self->{_keepclient} if $self->{_keepclient};

    eval { $self->job_wait()    if exists $self->{_job_id}; };

    $self->_san_cleanup();              # blind clean up SAN if present. This will never fail.

    return $self->SUPER::DESTROY;
}

=head2 ::wex_compare

compare the contents, inode and metadata files between two volumes. The default nature
is to compare the active file system. unless a snapshot name is provided. If differences
are found the routine will confess with all differences found. if no differences
are found the routine will return 1.

NOTE: the compare only shows inodes/files/data missing or changes at dst which
are present on src. If you want a complete mirror check you will need to call
this routine from with in the src and dst objects
    $diff = $src->wex_compare (cmp_vol => $dst, ...);
    $diff = $dst->wex_compare (cmp_vol => $src, ...);
the diffs may be different based on additions on either volume

=over

=item Arguments

B<cmp_vol> I<NOTE::Volume>

A NOTE::Volume object, to be used as the destination which will compared against

B<path> => I<string>

path of the tree which should be compared, if not provided compare from root (/)

I<Optional>

B<metadata> => I<1|0>

determine if metadata file should be compared or not, default 1 - compare

I<Optional>

B<snapshot> => I<string>

Snapshot name which will need to be compared: Default is active file system ( no snapshots )

I<Optional>

B<cifsdata> => I<1|0>

determines if the both inode and data contents of ACLs, Streams etc should be cmp
default is 1 -compare data

I<Optional>

B<verbose> => I<1|0>

enable/disable logging comments during compare
default is 0 - do not log

I<Optional>

=item Example

    eval {$vol->wex_compare(cmp_vol => $dst_vol, path => /x/y/z, metadata => 0, snapshot => "second_snapshot");};
    diffs = $@ if $@;
    $vol->wex_compare(cmp_vol => $dst_vol, path => /x/y/z, cifsdata => 0);
    $vol->wex_compare(cmp_vol => $dst_vol, path => /x/y );
    $vol->wex_compare(cmp_vol => $dst_vol, verbose => 1);

=back

=cut

sub wex_compare {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $diffs, $info, $snap_id );

    if ( ref( $args{cmp_vol} ) !~ /NOTE::.*Volume/ ) {
        confess "cmp_vol must be a object of type NOTE::Volume";
    }
    my $recursive = $args{recursive} || 0;
    my $verbose   = $args{verbose}   || 0;
    my $dst       = $args{cmp_vol};
    my $path      = $args{path}      || "/";
    my $metadata  = exists( $args{metadata} )  ? $args{metadata}  : 1;
    my $cifs_data = exists( $args{cifs_data} ) ? $args{cifs_data} : 1;
    my $snapshot  = $args{snapshot};
    my $vol_name  = $self->name;

    # if the cmp is for a snapshot, tranlate name to logical id and prep path
    if ($snapshot) {
        my $snapinfo = $self->snapshot_status( snapshot => $snapshot, fields => "logical-snap-id" );
        confess " Snapshot $snapshot not found on vol $vol_name. compare not possible "
          unless ($snapinfo);
        $snap_id = @$snapinfo[0]->{"logical-snap-id"};
        $path    = "~$snap_id" . "$path";
    }

    # compare metadata information.
    $diffs .= $self->_wex_metadata_cmp( cmp_vol => $dst, snap_id => $snap_id )
      if ($metadata);

    # walk through the tree and compare inodes
    eval { $info = $self->server->shell( cmd => "wex dir $vol_name$path" ); };    # use wex dir to get contents
    confess "wex dir for " . $path . " failed. Error: $@"                         # report error if it fails
      if $@;
    foreach my $line ( split( /\n/, $info ) ) {                                   # seperate the line of the wex dir output
        chomp $line;                                                              # and get rid of newlines
        if ( $line =~ /entry/ && $line !~ /\".\"/ && $line !~ /\"..\"/ ) {        # this is a leaf/branch
            $line =~ /name\s+\"([^\"]+)/;                                         # get the file/dir name
            my $fname = $1;
            if ( $path =~ /^\/$/ || $path =~ /^~\d+\/$/ ) {                       # check if path contains only "/" or ~<ssnum>/
                $path .= "$fname";                                                # if path is / or ~<ssnum>/ append file name
            }
            else {                                                                # else path is of type /x/y
                $path .= "/$fname";                                               # append /file_name to path /x/y
            }
            my ( @src_vals, @dst_vals );
            @src_vals = $self->wex_info( path => $path );                         # get the inode info on src vol
            eval { @dst_vals = $dst->wex_info( path => $path ); };                # get the inode info on dst vol
            if ($@) {
                $diffs .= $self->_wex_cmp_err_chk( $@, "file/dir", $path );       # handle error if any
            }
            else {
                $diffs .= $self->_cmp_inodes( \@src_vals, \@dst_vals, $verbose );    # cmp inodes based on file type
            }
            $diffs .= $self->_wex_cifsdata_cmp( \@src_vals, \@dst_vals, $dst, "ACL", $verbose, $snap_id )
              if ( $src_vals[0]->{xinode} && $cifs_data );                           # cmp cifs data is file has ACL
            $diffs .= $self->_wex_cifsdata_cmp( \@src_vals, \@dst_vals, $dst, "stream", $verbose, $snap_id )
              if ( $src_vals[0]->{sinode} && $cifs_data );                           # cmp cifs data if file has stream dir
            if ( $src_vals[0]->{type} == 2 ) {                                       # if inode type == 2 its a dir, call recursively
                NOTE::logcomment("sub-dir path:$path") if ($verbose);
                eval {
                    $path =~ s/~$snap_id//;                                          # remove the snapshot id before recursion
                    $diffs .= $self->wex_compare(
                        cmp_vol   => $dst,
                        path      => $path,
                        snapshot  => $snapshot,
                        cifsdata  => $cifs_data,
                        metadata  => 0,
                        recursive => 1,
                        verbose   => $verbose,
                    );
                    $path = "~$snap_id" . $path                                      # add the snapid back to the path
                      if ($snapshot);
                };
                $diffs .= "tree walk through failed for $path :$@ " if $@;
            }
            my $rem = ( split( /\//, $path, ) )[-1];                                 # remove last entry in path
            $path =~ s/\/$rem//;
        }
    }
    return $diffs if ($recursive);                                                   # if recursive call dont confess with diffs just return it
    confess($diffs) if ( $diffs ne "" );                                             # if not a recursive call and diffs are found confess with the diffs
    return 1;                                                                        # if no diffs found return 1
}

# handle wex info errors for wex_compare
sub _wex_cmp_err_chk {
    my ( $self, $error, $data_type, $path ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    if ( $error =~ /unable to find/i ) {
        return (" missing $data_type at $path\n");                                   # log info if file/dir is missing
    }
    else {
        confess("unexpected error finding $data_type on $path:$error\n");
    }
}

# compare the cifs file data contents
sub _wex_cifsdata_cmp {
    my ( $self, $src_inod, $dst_inod, $dst_vol, $type, $verbose, $snap_id ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( %src_strm_hash, %dst_strm_hash );
    my ( @strm_dirs, $info, $diff );
    my $src_name = $self->name;
    my $dst_name = $dst_vol->name;
    NOTE::logcomment("CIFS data compare started ") if ($verbose);

    # check if we are dealing with streams or ACLS
    if ( $type eq "stream" ) {
        my ( @dst_strm_dir, $src_sinod_path, $dst_sinod_path );
        $src_sinod_path = @$src_inod[0]->{sinode};
        $dst_sinod_path = @$dst_inod[0]->{sinode};
        if ($snap_id) {    #if snapshot id is passed append it to the sinode number
            $src_sinod_path = "~$snap_id." . @$src_inod[0]->{sinode};
            $dst_sinod_path = "~$snap_id." . @$src_inod[0]->{sinode};
        }
        my @src_strm_dir = $self->wex_info( path => $src_sinod_path );    # get inode info of src stream dir
        eval {
            @dst_strm_dir = $dst_vol->wex_info( path => $dst_sinod_path );    # get inode info of dst stream dir
        };
        if ($@) {
            $diff .= $self->_wex_cmp_err_chk( $@, "stream dir", $dst_sinod_path );    # get the error if stream dir not found
        }
        else {
            $diff .= $self->_cmp_inodes( \@src_strm_dir, \@dst_strm_dir, $verbose );    # cmp the stream dir inode contents
        }

        %src_strm_hash = $self->_stream_list( $src_inod, $snap_id );                    #get the list of inodes and names of streams
        %dst_strm_hash = $self->_stream_list( $dst_inod, $snap_id );                    #get the list of inodes and names of streams
        foreach my $strm ( keys(%src_strm_hash) ) {
            my @dst_strm_inode;
            my @src_strm_inode = $self->wex_info( path => $src_strm_hash{$strm} );      # get strm inode inf on src vol
            eval {
                @dst_strm_inode = $dst_vol->wex_info( path => $dst_strm_hash{$strm} );    # get strm inode inf on dst vol
            };
            if ($@) {
                $diff .= $self->_wex_cmp_err_chk( $@, "stream", $src_strm_hash{$strm} );
            }
            else {

                #compare the stream's inode atribs and data blocks
                $diff .= $self->_cmp_inodes( \@src_strm_inode, \@dst_strm_inode, $verbose );
                $diff .= $self->_wex_cmp_blocks( \@src_strm_inode, \@dst_strm_inode, $dst_vol, $verbose, $snap_id );
            }
        }
    }
    elsif ( $type eq "ACL" ) {
        my ( @dst_acl_inode, $src_xinode, $dst_xinode );
        $src_xinode = @$src_inod[0]->{xinode};
        $dst_xinode = @$dst_inod[0]->{xinode};
        if ($snap_id) {
            $src_xinode = "~$snap_id." . @$src_inod[0]->{xinode};
            $dst_xinode = "~$snap_id." . @$dst_inod[0]->{xinode};
        }
        my @src_acl_inode = $self->wex_info( path => $src_xinode );
        eval { @dst_acl_inode = $dst_vol->wex_info( path => $dst_xinode ); };
        if ($@) {
            $diff .= $self->_wex_cmp_err_chk( $@, "ACL", $src_xinode );
        }
        else {

            #compare the acls inode atribs and data blocks
            $diff .= $self->_cmp_inodes( \@src_acl_inode, \@dst_acl_inode, $verbose );    # cmp the acl inode contents
            $diff .= $self->_wex_cmp_blocks( \@src_acl_inode, \@dst_acl_inode, $dst_vol, $verbose );
        }
    }
    return $diff;
}

# internal routine to compare the data blocks for a given inode number.
# requires inode numbers for src and dst, dst vol name, snapshot id
sub _wex_cmp_blocks {
    my ( $self, $src_inod, $dst_inod, $dst_vol, $verbose, $snap_id ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $src_data, $dst_data );
    my $diff;
    NOTE::logcomment("Block comparison started ") if ($verbose);

    my $src_name = $self->name;
    my $dst_name = $dst_vol->name;

    # only compare blocks for ACLS, streams and stream dirs
    if ( ( @$src_inod[0]->{type} == 11 ) || ( @$src_inod[0]->{type} == 12 ) || ( @$src_inod[0]->{type} == 13 ) ) {
        for ( my $i = $#{@$src_inod} ; $i >= 0 ; $i-- ) {
            my $end_blk = @$src_inod[$i]->{blockcount} - 1;    # find out how many blocks of data are there
            $end_blk = 1
              if ( $end_blk <= 0 );                            # to handle files with just 1 blk of data
            my $inum_src = "$src_name.@$src_inod[$i]->{inum}";
            my $inum_dst = "$dst_name.@$dst_inod[$i]->{inum}";
            if ($snap_id) {
                $inum_src = "$src_name~$snap_id.@$src_inod[$i]->{inum}";
                $inum_dst = "$dst_name~$snap_id.@$dst_inod[$i]->{inum}";
            }
            eval { $src_data = $self->server->shell( cmd => "wex raw $inum_src\@0b-${end_blk}b" ); };
            confess "wex info for  $inum_src failed. Error: $@"
              if $@;
            eval { $dst_data = $dst_vol->server->shell( cmd => "wex raw $inum_dst\@0b-${end_blk}b" ); };
            confess "wex info for $inum_dst failed. Error: $@"
              if $@;

            #trim and compare data
            my @src_dat = split( /\n/, $src_data );
            my @dst_dat = split( /\n/, $dst_data );
            for ( my $j = @src_dat ; $j >= 0 ; $j-- ) {
                if ( $src_dat[$j] =~ /^\s\s\d+\s/ ) {    # dont cmp dsk id and blk offset and inod details
                    my $src_str = substr( $src_dat[$j], 8 );    #rm the disk offsets
                    my $dst_str = substr( $dst_dat[$j], 8 );    #rm the disk offsets( first 8 chars )
                    $diff .= "src => $src_str \; dst => $dst_str\n"
                      if ( $src_str ne $dst_str );
                }
            }
        }
    }
    else {
        $diff = "file type not a ACL,stream or stream dir: Type = @$src_inod[0]->{type} ";
    }
    return $diff;
}

## compares the inode information for the list of metadata inodes
sub _wex_metadata_cmp {
    my ( $self, %args, $verbose ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $dst = $args{cmp_vol};
    my ( $diff, $i );
    NOTE::logcomment("Metadata data cmp Started") if ($verbose);
    my @check_val   = qw( nlink umask qtree uid gid size xinode i2p owner sinode);
    my @meta_inodes = qw( 86p 500p );                                                #86 -> qtree metafile, 500 -> i2p metafile
    if ( $args{snap_id} ) {
        @meta_inodes = ();
        push( @meta_inodes, "~$args{snap_id}." . "86p" );
        push( @meta_inodes, "~$args{snap_id}." . "500p" );
    }

    foreach my $inod (@meta_inodes) {
        my @src_vals = $self->wex_info( path => $inod );
        my @dst_vals = $dst->wex_info( path => $inod );
        for ( $i = 0 ; $i <= $#src_vals ; $i++ ) {
            foreach (@check_val) {
                $diff .= "$inod $_ => $src_vals[$i]->{$_}, $_ => $dst_vals[$i]->{$_} \n"
                  if ( $src_vals[$i]->{$_} ne $dst_vals[$i]->{$_} );
            }
        }
    }
    NOTE::logcomment("Metadata data cmp Completed") if ($verbose);
    return $diff;
}

## compares the inode information for the different file types
sub _cmp_inodes {
    my ( $self, $src_inodes, $dst_inodes, $verbose ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ( $src, $dst, $diff );
    NOTE::logcomment("Inode contents comparison Started") if ($verbose);

    # based on the file type the following fields need to be compared
    my %txfer_info = (
        0 => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "size",   "xinode", "12p",   "owner", "sinode" ],
        1 => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "size",   "xinode", "owner", "sinode" ],
        2 => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", ],
        3  => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", "rdev" ],
        4  => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", "rdev" ],
        5  => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", ],
        6  => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", ],
        7  => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "xinode", "owner", ],
        11 => [ "generation", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "uid", "gid", "size", ],
        12 => [ "generation", "uid",   "gid",   "qtree", "sinode" ],
        13 => [ "generation", "uid", "gid", "nlink", "qtree", "size",  "sinode" ],
        18 => [ "generation", "uid", "gid", "nlink", "umask", "atime", "ctime", "mtime", "crtime", "qtree", "xinode", "owner", ],
    );
    for ( my $i = $#{@$src_inodes} ; $i >= 0 ; $i-- ) {
        foreach ( @{ $txfer_info{ @$src_inodes[$i]->{type} } } ) {    #get the array from txfer_info hash for file type
            NOTE::logcomment("comparing $_ ") if ($verbose);
            $diff .= "src inode @$src_inodes[$i]->{inum}: $_ => @$src_inodes[$i]->{$_}, dst inode @$dst_inodes[$i]->{inum}: $_ => @$dst_inodes[$i]->{$_} \n"
              if ( @$src_inodes[$i]->{$_} ne @$dst_inodes[$i]->{$_} );
        }
    }
    NOTE::logcomment("Inode contents comparison Completed") if ($verbose);
    return $diff;
}

=head2 ::counter

Returns amount of data read/write to this volume using the 'statistics show'.

command.

=over

=item Arguments

B<type> => I<r|w>

Specify whether read of write counter to return. Default: r

B<delta> => I<boolean>

Returns delta change since the last call to this method.
The default is to return the total count since last reboot.

Default: 0

=item Examples

    my $amt_read = $vol->counter();
    my $amt_wrt  = $vol->counter(type=>"w");

    $vol->counter();
    Tharn::snooze(10);
    my $delta = $vol->counter(delta=>1); # Returns bytes read between the 2 calls

=back

=cut

sub counter {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %type = (
        'r' => "read_data",
        'w' => "write_data"
    );
    $args{type} = "r" unless exists $args{type};
    confess "Invalid flag type. Must be " . ( join ', ', keys %type )
      unless exists $type{ $args{type} };

    my $req_col = ( $args{delta} ? "base-delta" : "value" );

    my $output;
    eval { $output = $self->server->statistics_show( node => $self->Node->name, object => $self->name, catgegory => 'volume', counter => $type{ $args{type} } ); };
    confess "Error obtaining read statistics for " . $self->description . ": $@"
      if $@;
    NOTE::logcomment( $self->description . " " . $type{ $args{type} } . " stats: value (" . $output->[0]->{value} . ") delta (" . $output->[0]->{'base-delta'} . ")" );
    return $output->[0]->{$req_col};
}

=head2 ::wafl_scan

Runs a walf scan on the volume based on the scan option provided. Because
this method only supports wafl scan volume commands only a subset of the
sub-commands are supported, and of those no command options are supported.
The sub-commands supported are listed below, along witht their privilege
level.

Because different sub-commands require different privilege levels certain
commands will require test mode. This method does its best to set the level
appropriate for the specific sub-command so tests using sub-commands not
requiring test mode can be used to test promoted builds. (burt 490236)

Supported sub-commands and their required privilege level:
    'measure_layout' => 'advanced',
    'reallocate' => 'advanced',
    'i2p' => 'advanced',
    'fold' => 'advanced',
    'status' => 'advanced',
    'ownblocks_calc' => 'diag',
    'snaplock' => 'diag',
    'blk_reclaim' => 'diag',
    'redirect' => 'diag',
    'pvbn_check -c' => 'test',
    'kireeti' => 'test',

=over

=item Arguments

B<option> => I<string>

where I<string> is one of the sub-commands listed above

=item Returns

=over

=item B<Success:>

output from command as a scalar (string)

=item B<Failure:>

confesses:
1) if invalid option (sub-command)
2) if can not set required privilege level
3) if the scan fails to start
4) if the volume is read-only

=back

=item Examples

    my $out = $vol->wafl_scan(option => 'measure_layout');
    my $out = $vol->wafl_scan(option => 'pvbn_check -c');

=back

=cut

sub wafl_scan {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # these are the wafl scan sub-commands supported by this method
    # and the privilege level they require. any additional sub-commands
    # being supported should be added to this list, or conversely the
    # method could be changed to issue the command in diag mode and if
    # that fails retry the command in test mode (similar to Node::setflag).
    # this is needed to support promoted builds, where test mode is not
    # available. (burt 490236)
    my %sub_commands = (
            'measure_layout' => 'advanced',
            'reallocate' => 'advanced',
            'i2p' => 'advanced',
            'fold' => 'advanced',
            'status' => 'advanced',
            'ownblocks_calc' => 'diag',
            'snaplock' => 'diag',
            'blk_reclaim' => 'diag',
            'redirect' => 'diag',
            'pvbn_check -c' => 'test',
            'kireeti' => 'test',
        );

    my ( $vol_name, $cmd, $output );

    $vol_name = $self->name;

    if (! exists($sub_commands{$args{option}})) {
        confess("option passed is not supported by wafl_scan: $args{option}");
    }

    # set the correct privilege for this command (burt 490236)
    $cmd = "priv set -q $sub_commands{$args{option}};";
    $cmd .= "wafl scan $args{option} $self->{_name}";

    $output = $self->server->dblade( cmd => $cmd );

    if ($output && ($output =~ /unrecognized privilege/i)) {
        confess('build does not support privilege level required for '
                . "$args{option}: $sub_commands{$args{option}}");
    }

    if ( $output =~ /(unable to start scan|usage)/i ) {
        confess(" incorrect vol name|param passed to wafl scan :$output ");
    }

    if ( $output =~ /(read only|readonly)/i) {
        confess(" wafl scan failed on RO volume : $output ");
    }

    return $output;
}

=head2 sd_test_differ

Run the test-mode C<sd test_differ> command on a node. Triggers a series of
WAFL_SNAP_DIFF messages to diff the named file in the given snapshot.

Returns an array reference of records that are the parsed lines from the
command, or a message string if there are no diffs to parse. Each record is a
hash reference with the keys C<fbn>, C<vbn>, C<count> and C<type>.

=over

=item Arguments

B<path> = string

File path to compare in the diffing operation

B<snap> = string

Snapshot directory under C<.snapshot> to compare to in the diff operation

B<opts> = hash table

I<Optional> hash table of arguments to the C<test_differ> command:

=over

=item B<base_snap> = string

Base snapshot to diff against

=item B<abort_on_corruption> = boolean

Whether to set the C<REP_OK> flag on the B<test_differ> command, which
causes it to abort upon detection of corruption. Default is 0 (false).

=item B<baseline_hole_info> = boolean

Without B<base_snap>, whether to return hole information. Default is 0 (false).

=item B<get_vbn> = boolean

Whether to retrieve VBNs for changed blocks. Default is 0 (false).

=item B<l1_diff> = boolean

Whether to diff only at L1 level. Default is 0 (false).

=item B<start_fbn> = integer

First FBN to process.

=item B<end_fbn> = integer

Last FBN to process.

=item B<max_fbns> = integer

Limit on change records per message. Default is 512.

=item B<expand_lun_clone> = boolean

Whether to expand LUN clones. Default is 0 (false).

=back

=item Examples

To check the changes to a file in the volume against a snapshot:

    # Assume $file is already created on the volume
    $volume->snapshot_create('snap1');
    # Make changes to $file
    my $change_recs = $volume->sd_test_differ($file, 'snap1');
    # The records in @$change_recs should match changes to $file

To check the changes to a file between two specific snapshots:

    # Assume $file is already created on the volume
    $volume->snapshot_create('snap1');
    # Make changes to $file
    $volume->snapshot_create('snap2');
    my $change_recs = $volume->sd_test_differ(
        $file, 'snap2',
        base_snap => 'snap1'
    );
    # The records in @$change_recs should match changes to $file

=back

=cut

sub sd_test_differ { ## no critic (ProhibitExcessComplexity)
    my ($self, $path, $snap, @opts) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $cmd = 'sd test_differ';

    # Validate the params. Fairly easy here, as all are optional. Use
    # the opportunity to set default values to the flags, so that we
    # don't need "defined" tests later on...
    my %opts = validate(
        @opts => {
            base_snap => {
                optional => 1,
                type     => SCALAR,
            },
            abort_on_corruption => {
                optional => 1,
                type     => SCALAR,
                default  => 0,
            },
            baseline_hole_info => {
                optional => 1,
                type     => SCALAR,
                default  => 0,
            },
            get_vbn => {
                optional => 1,
                type     => SCALAR,
                default  => 0,
            },
            l1_diff => {
                optional => 1,
                type     => SCALAR,
                default  => 0,
            },
            start_fbn => {
                optional => 1,
                type     => SCALAR,
                regex    => qr/^\d+$/,
            },
            end_fbn => {
                optional => 1,
                type     => SCALAR,
                regex    => qr/^\d+$/,
            },
            max_fbns => {
                optional => 1,
                type     => SCALAR,
                regex    => qr/^\d+$/,
            },
            expand_lun_clone => {
                optional => 1,
                type     => SCALAR,
                default  => 0,
            },
        }
    );

    # Build up the args to the filer-level "sd test_differ" command
    my $args = q{};
    $args .= ' -r' if ($opts{abort_on_corruption});
    $args .= ' -c' if ($opts{expand_lun_clone});
    $args .= " -e $opts{end_fbn}" if (defined $opts{end_fbn});
    $args .= ' -h' if ($opts{baseline_hole_info});
    $args .= " -m $opts{max_fbns}" if (defined $opts{max_fbns});
    $args .= " -s $opts{start_fbn}" if (defined $opts{start_fbn});
    $args .= ' -t' if ($opts{l1_diff});
    $args .= ' -v' if ($opts{get_vbn});

    if (defined $opts{base_snap}) {
        my $base_path = sprintf '/vol/%s/.snapshot/%s/%s' =>
            $self->name, $opts{base_snap}, $path;
        $args .= " -b $base_path";
    }

    my $incr_path = sprintf '/vol/%s/.snapshot/%s/%s' =>
        $self->name, $snap, $path;
    $args .= " -i $incr_path";

    # XXX: In this form, this has no hope of working on a huge file.
    # We need to parse the output line-by-line as it comes back.
    my $text = $self->server->dblade(cmd => "priv set -q test ; $cmd$args");

    # Parse and validate the header (first line of output). Use regexp
    # matching/substitution to avoid creating copies of $text, since it is
    # likely to be HUGE.
    my $header;
    if ($text =~ s/^(.*)(?:\n|\z)//) {
        $header = $1;
    } else {
        croak 'Regexp failed to extract header from text';
    }

    # Look to see if we got a diagnostic line stemming from the use of the
    # "max_fbns" option. One of the two possibilities is actually an error
    # case:
    if ($header =~ /max_fbns too large, using \d+/) {
        # Slice out the record-lines from the remaining text so that only
        # diagnostic output remains. We'll want to provide all of this in the
        # error message when we croak.
        $text =~ s/^Start\s+FBN.*(?:\n|\z)//gm;
        croak "$cmd failed (max_fbns range): $header\n$text";
    }
    if ($header =~ /max_fbns set to \d+/) {
        # We ignore this one. Just re-load $header with the next line from
        # $text.
        if ($text =~ s/^(.*)(?:\n|\z)//) {
            $header = $1;
        } else {
            croak 'Regexp failed to extract header from text';
        }
    }

    # We passed the error-case, and if the second matched we reloaded $header.
    # Check that the header we got matches the parameters we passed. The
    # header should match this line:
    #
    # Using: start_fbn=0, end_fbn=0, need_vbn=true, differ_type=wsd, max_fbns=512
    #
    # Numbers may vary, need_vbn may be "true" or "false", and differ_type
    # may be "wsd" or "test".
    my ($start, $end, $get_vbn, $type, $max);
    if ($header =~ m{
                        ^ # Start of string
                        Using:\s+
                        start_fbn=(\d+),\s+        # $start
                        end_fbn=(\d+),\s+          # $end
                        need_vbn=(true|false),\s+  # $get_vbn
                        differ_type=(wsd|test),\s+ # $type
                        max_fbns=(\d+)\s*          # $max
                        $ # End of string
                    }x) {
        ($start, $end, $get_vbn, $type, $max) = ($1, $2, $3, $4, $5);
    } else {
        croak "Error parsing '$cmd' header: $header";
    }
    if (! ($start == ((defined $opts{start_fbn}) ? $opts{start_fbn} : 0) &&
           $end == ((defined $opts{end_fbn}) ? $opts{end_fbn} : 0) &&
           $get_vbn eq ($opts{get_vbn} ? 'true' : 'false') &&
           $type eq ($opts{l1_diff} ? 'test' : 'wsd') &&
           $max == ((defined $opts{max_fbns}) ? $opts{max_fbns} : 512))) {
        croak "header from '$cmd' did not match input parameters";
    }

    # Build a regexp to check for an empty record list by enumerating
    # the different messages that could signal that. This approach makes
    # it easier to extend the list without having an explicit RE that is
    # a pain to read.
    my @empty_record_messages = (
        'Files are identical',
        'The file is of 0 length',
        'The file is completely holey',
    );
    my $empty_record_re = join q{|} => @empty_record_messages;
    # Now, check for an empty record list.
    if ($text =~ /^($empty_record_re)$/) {
        return $text;
    }

    # Check whether the command explicitly passed on an error.
    if ($text =~ /^WAFL error: (.*?)$/m) {
        croak "$1\n";
    }

    # Parse each line of output and accumulate the list of records. The
    # logic in the while() loop may look odd, but it is more efficient
    # than splitting the string into lines, as that would allocate a
    # potentially huge array of strings.
    #
    # Lines look like this:
    #
    # Start FBN: 3986, VBN: 13341, Count: 20, DATA_CHANGED
    #
    # Numbers may vary, and the last element may be one of several different
    # text values.
    my @records = ();
    while ($text ne q{} && $text =~ s/\A(.*?)(?:\n|\z)//) {
        my $line = $1;
        if ($line =~ m{
                          ^                  # Start of line
                          Start\s+
                          FBN:\s+(\d+),\s+   # Key "fbn"
                          VBN:\s+(\d+),\s+   # Key "vbn"
                          Count:\s+(\d+),\s+ # Key "count"
                          ([A-Z_]+)\s*       # Key "type"
                          $                  # End of line
                      }x) {
            push @records, { fbn => $1, vbn => $2, count => $3, type => $4 };
        } else {
            croak "error parsing '$cmd' output: $line";
        }
    }
    if (! @records) {
        croak "no records returned from '$cmd'";
    }

    return \@records;
}

=head2 efficiency_on

Enable SIS/Efficiency (dedupe) on the volume this is called on. Does not
start the process, only activates the relevant code. See also
B<efficiency_start>, below.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments:

=over

=item B<compression> = boolean

Here compression means background compression.
If true, then after efficiency is enabled call B<efficiency_modify> to also
activate background compression on the volume.

=item B<inline-compression> = boolean

If true, then after efficiency is enabled call B<efficiency_modify> to activate
background compression and set it to inline operation. If this option is
passed, C<compression> is assumed to be true (a false value for
C<compression> is ignored).

=back

=item Examples

    # Just enable SIS/efficiency:
    $volume->efficiency_on();

    # Enable efficiency and compression at the same time:
    $volume->efficiency_on('compression' => 1);

    # Enable efficiency and compression, and set compression to inline:
    $volume->efficiency_on(
        'compression' => 1,
        'inline-compression'     => 1
    );

    # Same effect as the previous example:
    $volume->efficiency_on('inline-compression' => 1);

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_on
#
#   Description:    Enable SIS/Efficiency on the volume
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume to enable this for
#                   %args     in      hash      Hash table of arguments
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_on {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $result;

    $result = eval {
        $self->server->volume_efficiency_on(
            volume  => $self->name,
            vserver => $self->Vserver->name
        );
        1;
    };
    if (! $result) {
        # Ignore the error if it is just saying that the efficiency engine
        # is already on. We absorb that.
        $self->_processError($@)
            if ($@ !~ /Operation has already been enabled/);
    }

    # The efficiency_modify() method will handle any errors from that CDEF,
    # so no need for eval's here:
    # The parameter background-compression is changed to compression, due to
    # the UI change 492688. However still keeping background-compression
    # parameter in this function, because the function efficiency_modify, which
    # is being called by this function, is handling appropriately.
    if ($args{'inline-compression'}) {
        $self->efficiency_modify(
            'background-compression' => 1,
            'inline-compression'     => 1
        );
    } elsif ($args{'background-compression'} or $args{'compression'}) {
        $self->efficiency_modify('background-compression' => 1);
    }

    return $self;
}

=head2 efficiency_off

Disable SIS/Efficiency (dedupe) on the volume this is called on. It is not
clear what the behaviour is if a dedupe scan is currently active when this is
called. See also B<efficiency_stop>, below. If compression was enabled when
B<efficiency_on> was called, it is disabled before efficiency itself is
disabled.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

None.

=item Examples

    $volume->efficiency_off();

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_off
#
#   Description:    Disable SIS/Efficiency on the volume
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume to disable this for
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_off {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Disable efficiency itself. This is a CDEF call, so we have to do the
    # usual eval-trick around it:
    my $result = eval {
        $self->server->volume_efficiency_off(
            volume  => $self->name,
            vserver => $self->Vserver->name
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 efficiency_start

Start the SIS/Efficiency (dedupe) process on the calling volume. For this to
work, the volume must first have Efficiency functionality enabled for the
volume (see B<efficiency_on>, above).

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments to the C<volume efficiency start> command:

=over

=item B<background> = boolean

If true, return immediately without waiting for the job to finish. This is
most significant when the C<scan> option is set to a true value. The default
is 0 (C<false>), which waits for the job to finish before returning.

=item B<scan-old-data> = boolean

The value is passed to the B<-scan-old-data> parameter of the underlying filer
command. Unlike the rest of the true/false parameters, B<scan-old-data>
defaults to 1 (C<true>), and must be explicitly passed as 0 (C<false>) if you
wish to prevent the scan. See also the previous option, B<background>.

=item B<snapshot-blocks> = boolean

This option also defaults to C<true>, and must be explicitly passed as 0
(C<false>) if you do not want it. When true, this option instructs a scan
to also examine blocks on snapshot volumes, as well as the main volume itself.
When false, only the volume itself is scanned.

=item B<force> = boolean

Triggers the B<-force> parameter to the filer command. If this is passed a
C<true> value, forces the scan operation even ONTAP might have decided it
was unnecessary.

=item B<compression> = boolean

=item B<dedupe> = boolean

These options can be used to control which of the underlying functions
(deduplication and compression) are applied in the scan. Any option that
is passed with a C<true> value will be applied, any that are passed with a
B<false> value are suppressed.

If neither of these options are given, then the functions that are currently
enabled (either via B<efficiency_on> or B<efficiency_modify>) are applied.

=item B<use-checkpoint> = boolean

=item B<del-checkpoint> = boolean

=item B<shared-blocks> = boolean

Each of these, if passed, causes the B<volume efficiency start> parameter of
the same name to be passed, with the corresponding true/false value, based on
the value passed here.

=back

=item Examples

    # Start dedupe, scan, and return after scanning is finished
    $volume->efficiency_start();

    # Start, but return immediately instead of waiting for scan
    $volume->efficiency_start(background => 1);

    # Start, but do not do an initial scan
    $volume->efficiency_start('scan-old-data' => 0);

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_start
#
#   Description:    Start SIS/efficiency scanning on a volume
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume to start this on.
#                                                 Efficiency must already be
#                                                 enabled on volume.
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be passed to the CDEF
#                                                 call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_start {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Convert the boolean-style arguments from Perl booleans to strings:
    for (qw(scan-old-data use-checkpoint del-checkpoint force compression
            dedupe shared-blocks snapshot-blocks)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }
    # Set the defaults for the scan and snapshot-blocks arguments
    $args{'scan-old-data'}   ||= 'true';
    $args{'snapshot-blocks'} ||= 'true';

    # 'background' isn't actually an argument for the command itself, it's
    # just for us:
    my $background = delete $args{background} || 0;

    my $result = eval {
        $self->server->volume_efficiency_start(
            vserver => $self->Vserver->name,
            volume  => $self->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    # If $background is true, then they want to let this run in the background
    # without waiting. And we don't have to do anything. But if it is false,
    # we have to poll ->efficiency_show() until status is idle.
    if (! $background) {
        my $status = $self->efficiency_show();
        while ($status->{'op-status'} !~ /idle/i) {
            Tharn::snooze(1);
            $status = $self->efficiency_show();
        }
    }

    return $self;
}

=head2 efficiency_stop

Stop the SIS/Efficiency scanning process on the calling volume. This does not
disable the functionality (see B<efficiency_off>, above), but merely stops
a running scan.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

None.

=item Examples

    $volume->efficiency_stop();

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_stop
#
#   Description:    Stop SIS/efficiency scanning on a volume
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume to stop this on
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_stop {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $result;

    $result = eval {
        $self->server->volume_efficiency_stop(
            volume  => $self->name,
            vserver => $self->Vserver->name
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 efficiency_modify

Modify the SIS/Efficiency process on the calling volume. The arguments hash
controls what is modified.

Returns the calling object if successful, throws an exception otherwise.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments to the C<volume efficiency config> command:

=over

=item B<schedule> = string

Set the SIS/efficiency schedule to the string passed as the value.

=item B<policy> = string

Set the SIS/efficiency policy to the string passed as the value.

=item B<compression> = boolean

Enable or disable SIS/efficiency background compression on the volume.

=item B<inline-compression> = boolean

Enable or disable C<inline> compression. Background compression must either be
already enabled, or be set to C<true> in this same call. If compression is not
enabled, this will fail and throw an exception.

=back

=item Examples

    # Disable compression on the volume:
    $volume->efficiency_modify('compression' => 0);

    # Both enable compression and set it to inline compression mode:
    $volume->efficiency_modify(
        'compression' => 1,
        'inline-compression'     => 1
    );

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_modify
#
#   Description:    Change setting on SIS/efficiency on the calling volume
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being modified
#                   %args     in      hash      Hash table of arguments to
#                                                 be filtered and passed on to
#                                                 the CDEF call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_modify {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # If user provides background-compression parameter, make it to compression
    # This is because due to the UI change 492688, "background-compression"
    # is the parameter for older builds, and "compression" is the changed
    # parameter. So,the cdef will be called first with "compression" and if it
    # fails, then cdef will be called with old parameter
    # "background-compression".
    if(exists $args{'background-compression'}){
        $args{'compression'} = $args{'background-compression'};
        delete $args{'background-compression'};
    }

    # Convert the compression and inline arguments from Perl booleans to
    # strings:
    for (qw(compression inline-compression)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }

    my $result = eval {
        $self->server->volume_efficiency_modify(
            vserver => $self->Vserver->name,
            volume  => $self->name,
            %args
        );
        1;
    };
    if (! $result) {
        if($@ !~ /invalid\s+argument\s+"-compression"/xi) {
              $self->_processError($@);
        }
        else {
            #Try providing the argument background-compression - UI change 492688
            my $compression = delete $args{'compression'};
            $args{'background-compression'} = $compression;
            my $result2 = eval {
                              $self->server->volume_efficiency_modify(
                                   vserver => $self->Vserver->name,
                                   volume  => $self->name,
                                   %args
                              );
                              1;
                          };
            if (! $result2) {
                $self->_processError($@);
            }
        }
    }

    return $self;
}

=head2 efficiency_show

Run C<volume efficiency show> on the calling volume and return the data as
a hash-reference keyed by abbreviated versions of the column names.

If called on a volume that does not have efficiency data available, it
returns B<undef>. This is the case before efficiency is enabled via
B<efficiency_on>, and also the case after efficiency is rolled back via
B<efficiency_undo>.

If an error occurs, an exception is thrown.

=over

=item Arguments

None.

=item Return Value

The return value is a hash reference with keys that are closely based on the
field names from the human-readable format of the call. The fields are
itemized in the order they appear when C<volume efficiency show> is run from
B<ngsh>:

=over 8

=item vserver

Virtual Server Name field

=item volume

Volume Name field

=item path

Path To Volume field

=item state

State field

=item op-status

Status field (note the difference in name)

=item progress

Progress field

=item type

Type field

=item schedule

Schedule field

=item policy

Efficiency Policy Name field

=item policy-id

Efficiency Policy UUID field

=item min-blocks-shared

Min Blocks Shared field

=item blks-skipped-sharing

Blocks Skipped Sharing field

=item last-op-begin

Last Operation Begin field

=item last-op-end

Last Operation End field

=item last-op-size

Last Operation Size field

=item last-op-error

Last Operation Error field

=item changelog-usage

Changelog Usage field

=item checkpoint-time

Checkpoint Time field

=item checkpoint-op-type

Checkpoint Op Type field

=item checkpoint-stage

Checkpoint Stage field

=item checkpoint-sub-stage

Checkpoint Sub-Stage field

=item checkpoint-progress

Checkpoint Progress field

=item stage

Stage field

=item fingerprints-gathered

Fingerprints Gathered field

=item gathering-begin-time

Gathering Begin field

=item fingerprints-sorted

Fingerprints Sorted field

=item duplicate-blocks-found

Duplicate Blocks Found field

=item sorting-begin-time

Sorting Begin field

=item blocks-deduplicated

Blocks Deduplicated field

=item dedupe-begin-time

De-Duping Begin field

=item fingerprints-deleted

Fingerprints Deleted field

=item checking-begin-time

Checking Begin field

=item compression

Offline Compression field, C<true> or C<false> depending on whether background
compression is currently enabled.

=item inline

Inline Compression field, C<true> or C<false> depending on whether inline
compression is currently enabled.

=back

=item Examples

    # Check the status field to wait for a job to finish:
    while ($volume->efficiency_show->{status} !~ /idle/i) {
        sleep 1;
    }

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_show
#
#   Description:    Return the structure of key/value pairs from running
#                   "volume efficiency show" on the given volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being queried
#
#   Returns:        Success:    hash ref
#                   Failure:    dies
#
###############################################################################
sub efficiency_show {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($result, $value);

    $result = eval {
        $value = $self->server->volume_efficiency_show(
            vserver => $self->Vserver->name,
            volume  => $self->name
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    # The return value from volume_efficiency_show is a list-reference
    # containing one hash-ref, which is what we want to return:
    return $value->[0];
}

=head2 efficiency_stat

Run C<volume efficiency stat> on the calling value and return the data as a
hash-reference keyed by abbreviated versions of the column names. If called
on a volume that does not have efficiency enabled via B<efficiency_on>, it
will throw an exception.

=over

=item Arguments

None.

=item Return Value

Hash-reference.

=item Examples

    # Get the "Maximum Refcount" field from the statistics
    my $max_refcount = $volume->efficiency_stat->{'max-ref-count'};

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_stat
#
#   Description:    Return the structure of key/value pairs from running
#                   "volume efficiency stat" on the given volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being queried
#
#   Returns:        Success:    hash ref
#                   Failure:    dies
#
###############################################################################
sub efficiency_stat {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($result, $value);

    $result = eval {
        $value = $self->server->volume_efficiency_stat(
            vserver => $self->Vserver->name,
            volume  => $self->name
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    # The return value from volume_efficiency_stat is a list-reference
    # containing one hash-ref, which is what we want to return:
    return $value->[0];
}

=head2 efficiency_undo

Undo the work of SIS/Efficiency on the calling volume. The volume must have
efficiency disabled (see B<efficiency_off>, above) before this is called
or an exception will be thrown.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments to pass to the C<volume efficiency undo> command:

=over

=item B<force> = boolean

If passed and a true value, forces the operation

=item B<compression> = boolean

Decompress the data in the volume

=item B<dedupe> = boolean

=item B<deduplication> (deprecated) = boolean

Undo block-sharing in the volume. The option to C<undo> is actually
C<dedupe>, but C<deduplication> is also accepted to maintain backwards 
compatiblity with previous UI changes. If both of these are passed, 
C<dedupe> takes precedence.

=item B<inode> = integer

Inode number of a specific file to undo

=item B<undo-type> = string

One of C<all> or C<wrong>, selective undo

=item B<log> = boolean

Whether to enable logging

=back

If only one of C<compression> or C<dedupe> (or C<deduplication>) is given
(and C<true>), then only that facet of the efficiency engine is undone. If
neither of these options is passed, then whatever was enabled (either by
B<efficiency_on> or B<efficiency_modify>) is undone.

=item Examples

    $volume->efficiency_undo();

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_undo
#
#   Description:    Undo a Storage Efficiency operation. Just issues the
#                   appropriate CDEF call. Waits for it to finish unless
#                   "background" is passed in %args with a true value.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being operated upon
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be filtered and passed
#                                                 to the CDEF call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_undo {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $result;

    # Convert these arguments from Perl booleans to strings:
    for (qw(force compression deduplication dedupe log)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }

    # If the "deduplication" option was passed, copy it over to "dedupe"
    # unless that option is already passed.
    if ($args{deduplication}) {
        my $value = delete $args{deduplication};
        $args{dedupe} = $value if (! $args{dedupe});
    }

    # 'background' isn't actually an argument for the command itself, it's
    # just for us:
    my $background = delete $args{background} || 0;

    $result = eval {
        $self->server->volume_efficiency_undo(
            vserver => $self->Vserver->name,
            volume  => $self->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    # If $background is true, then they want to let this run in the background
    # without waiting. And we don't have to do anything. But if it is false,
    # we have to poll ->efficiency_show() until it returns undef.
    if (! $background) {
        while ($self->efficiency_show) {
            Tharn::snooze(1);
        }
    }

    return $self;
}

=head2 efficiency_policy_create

Create a Storage Efficiency policy, a named entity that schedules SE
activities for volumes on a V-server that have SE enabled. The named policy
is associated with a recurrance schedule (as created/managed by the
C<"job schedule"> command hierarchy) that controls when the SE engine
activates.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments:

=over

=item B<policy> = string

Name of the policy to create, a string of 1 to 32 characters. If the policy
name is longer than 32 characters an exception is thrown. The name must be
unique within the V-server that the policy is created for.

=item B<schedule> = string

The name of a job schedule (see the C<"job schedule show"> command on a
filer) that governs the execution of SE for this policy. The value passed
here must match the name of an existing job schedule.

=item B<duration> = integer

The length of time during which the SE engine should be active before going
back to an inactive state and waiting for the scheduler to next activate it.
The value is an integer number of hours in the range 1 to 999.

=item B<enabled> = boolean

Whether or not to enable this policy immediately upon creating it. The default
is C<true>, to enable the policy. You may pass C<true>, C<false>, 1 or 0.

=item B<comment> = string

A comment regarding this policy, useful for notes about its purpose, etc.
Limited to 255 characters, if the string is longer than this an exception
will be thrown.

=back

=item Examples

    $vol->efficiency_policy_create(
        policy   => 'Nightly',
        duration => '1',
        schedule => 'daily',
        comment  => 'Cron runs this each night at midnight'
    );

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_policy_create
#
#   Description:    Create a new Storage Efficiency policy using the args
#                   passed in.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being operated upon
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be filtered and passed
#                                                 to the CDEF call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_policy_create {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Check that policy exists, as it is required
    if (! $args{policy}) {
        confess q(efficiency_policy_create: Argument 'policy' is required);
    }

    # The schedule argument is also required, so check that it is present.
    # The system will tell us if it refers to an unknown schedule, we don't
    # have to check that:
    if (! $args{schedule}) {
        confess q(efficiency_policy_create: Argument 'schedule' is required);
    }

    # If they passed in a comment, make sure it is properly quoted
    if ($args{comment}) {
        # Force the comment to be "quoted" for use by the CDEF. Not sure if
        # this is a CDEF bug or not...
        $args{comment} =~ s/"/'/g; # Make sure there are no embedded "'s
        $args{comment} = qq("$args{comment}");
    }

    # If they passed the enabled parameter, normalize it to a string value:
    if (exists $args{enabled}) {
        $args{enabled} = ($args{enabled} && $args{enabled} !~ /false/i)
            ? 'true' : 'false';
    }

    my $result = eval {
        $self->server->volume_efficiency_policy_create(
            vserver => $self->Vserver->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 efficiency_policy_delete

Delete an existing Storage Efficiency policy. Takes only one parameter, the
name of the policy to delete. Does not delete the job schedule that the
policy is associated with.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments to the C<volume efficiency policy delete> command:

=over

=item B<policy> = string

The name of the policy to delete. Must match exactly the name of an existing
policy on this V-server.

=back

=item Examples

    $volume->efficiency_policy_delete(policy => 'Nightly');

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_policy_delete
#
#   Description:    Delete an existing Storage Efficiency policy using the
#                   args passed in.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being operated upon
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be filtered and passed
#                                                 to the CDEF call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_policy_delete {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Can't delete a policy if we don't know which one to delete:
    if (! $args{policy}) {
        confess q(efficiency_policy_create: Argument 'policy' is required);
    }

    my $result = eval {
        $self->server->volume_efficiency_policy_delete(
            vserver => $self->Vserver->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 efficiency_policy_modify

Modify an existing Storage Efficiency policy. This allows you to change the
duration, the associated job schedule, etc.

Returns the calling object if successful, throws an exception if it fails.

=over

=item Arguments

B<opts> = hash table

Hash table of arguments. Of the parameters recognized, C<policy> I<must>
be specified (or an exception will be thrown). Any other parameters are
only passed if the corresponding attribute of the policy is being changed.
Any parameters not passed are left unchanged in the policy.

=over

=item B<policy> = string

Name of the policy to modify. This policy must already exist in the table
of SE policys or an exception will be thrown.

=item B<schedule> = string

The name of a job schedule (see the C<"job schedule show"> command on a
filer) that governs the execution of SE for this policy. The value passed
here must match the name of an existing job schedule.

=item B<duration> = integer

The length of time during which the SE engine should be active before going
back to an inactive state and waiting for the scheduler to next activate it.
The value is an integer number of hours, ranging from 1 to 999.

=item B<enabled> = boolean

Whether or not to enable this policy. You may pass C<true>, C<false>, 1 or 0.

=item B<comment> = string

A comment regarding this policy, useful for notes about its purpose, etc.
Limited to 255 characters, if the string is longer than this an exception
will be thrown.

=back

=item Examples

    # Change the previously-defined "Nightly" policy from a 1-hour
    # duration to a 2-hour duration:
    $volume->efficiency_policy_modify(
        policy   => 'Nightly',
        duration => '2'
    );

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_policy_modify
#
#   Description:    Modify an existing Storage Efficiency policy, using the
#                   data in the args passed in to the call.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being operated upon
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be filtered and passed
#                                                 to the CDEF call
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub efficiency_policy_modify {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # First check that our required argument is present, throw an exception
    # if it isn't:
    if (! $args{policy}) {
        confess q(efficiency_policy_modify: Argument 'policy' is required);
    }

    # If they specified the 'enabled' parameter, convert it to a string:
    if (exists $args{enabled}) {
        $args{enabled} = ($args{enabled} && $args{enabled} !~ /false/i)
            ? 'true' : 'false';
    }

    if ($args{comment}) {
        # Force the comment to be "quoted" for use by the CDEF. Not sure if
        # this is a CDEF bug or not...
        $args{comment} =~ s/"/'/g; # Make sure there are no embedded "'s
        $args{comment} = qq("$args{comment}");
    }

    my $result = eval {
        $self->server->volume_efficiency_policy_modify(
            vserver => $self->Vserver->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 efficiency_policy_show

Run C<volume efficiency policy show> and return the results as an
array-reference of hash-references of the data returned.

=over

=item Arguments

B<opts> = hash table of arguments to the C<volume efficiency policy show>
command:

=over

=item B<policy> = string

The name of a specific policy to look up and report on. If this isn't passed,
then all policies are returned. If this is not found in the table of
policies, an empty array is returned.

=item Return Value

The return value is a list-reference with zero or more hash-references
contained within. Each hash reference has the following keys:

=item vserver

The V-server this policy applies to. A policy is applied to all volumes in
the V-server that have SE enabled on them.

=item policy

The name of the policy.

=item schedule

The name of the schedule (from C<jobs schedule show>) that governs this
policy's execution.

=item duration

The duration, in hours, for which the SE engine will run before turning itself
off.

=item enabled

A string, one of C<true> or C<false>, that indicates whether this policy is
currently enabled.

=item comment

The comment for the policy, as set by B<efficiency_policy_create> or
B<efficiency_policy_modify>.

=item policyid

A unique hexadecimal ID string for the policy, generated by the filer system
itself.

=back

=item Examples

    # Get all the current policies:
    my $policies = $volume->efficiency_policy_show();

    # Get the details on the "Nightly" policy:
    my $nightly = $volume->efficiency_policy_show(
        policy => 'Nightly'
    );

=back

=cut

###############################################################################
#
#   Sub Name:       efficiency_policy_show
#
#   Description:    Return a data structure listing all existing Storage
#                   Efficiency policies.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class, the
#                                                 volume being operated upon
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to be filtered and passed
#                                                 to the CDEF call
#
#   Returns:        Success:    arrayref of data
#                   Failure:    dies
#
###############################################################################
sub efficiency_policy_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $data;

    # We don't do any real checking of data integrity here, because these are
    # all query parameters and don't actually update any settings on the
    # filer. Bad data will result in an exception anyway.

    my $result = eval {
        $data = $self->server->volume_efficiency_policy_show(
            vserver => $self->Vserver->name,
            %args
        );
        1;
    };
    if (! $result) {
        $self->_processError($@);
    }

    # If there were no policies, $data is undef. We'd prefer to return an
    # empty array-ref:
    return $data || [];
}

=head2 clone_create

The B<clone_create> command creates a FlexClone volume that is linked to the
volume making the call. The return value is a new B<NOTE::Volume> object that
represents the clone.

=over

=item Arguments

B<%args> = hash table of arguments to the C<volume clone create> command:

=over

=item B<flexclone> = string

Name to give the new volume, the clone of the calling volume. If this parameter
is not given, a random name is generated following the same pattern given to
anonymous volumes.

=item B<parent-snapshot> = string

Optional. The name of a snapshot on the parent to clone, rather than the volume
itself.

=item B<junction-path> = string

Optional. Junction path to assign to the clone.

=item B<junction-active> = boolean

Optional. Indicates whether the junction is to be made active (mounted).

=item B<space-guarantee> = none|volume|file

Optional. Space guarantee style. If not given, the default will be used.

=item B<comment> = text

Optional. Comment text. This will be examined, and if it contains any space
characters and is I<not> enclosed in quotation marks, it will be. Otherwise,
the extra words will be interpreted by the filer CLI and cause an exception.

=item B<foreground> = boolean

Optional. Whether to perform the process in the foreground. The default is
C<true>.

=back

=item Examples

    $clone = $vol_src->clone_create(flexclone => $self->name . '_clone');

=back

=cut

###############################################################################
#
#   Sub Name:       clone_create
#
#   Description:    Issue the CDEF for volume_clone_create and if successful
#                   return a new Volume object that represents the new clone.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %args     in      hash      Arguments key/value pairs
#
#   Returns:        Success:    NOTE::GX::Volume object
#                   Failure:    dies
#
###############################################################################
sub clone_create {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ($clone, $eval_return, $data);

    # This is a non-standard approach, but I don't want every call to this
    # method doing a license-check on the cluster. So I'm using a "super-
    # private" hash key as a sort of cached flag. If it is set, then we've
    # already checked and/or installed the license.
    if (! $self->{__FLEXCLONE_LICENSE_PRESENT}) {
        # The second arg tells license() to install it if it isn't present
        $self->Cluster->license('FlexClone', 1);
        $self->{__FLEXCLONE_LICENSE_PRESENT} = 1;
    }

    # Convert these arguments from Perl booleans to strings:
    for (qw(junction-active foreground)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }

    # The 'flexclone' parameter is needed to name the new volume. If they
    # didn't give us one, generate an anonymous name:
    $args{flexclone} ||= NOTE::Volume->newName('vl');

    # For now, force 'foreground' to be true, as I'm not sure if I can return
    # a ::Volume referent if the clone operation is still in progress. And I
    # don't know what I would return in that case, anyway.
    $args{foreground} = 'true';

    # For the CDEF, just set parent-volume to the calling Volume and let %args
    # provide the rest of the data:
    $eval_return = eval {
        $data = $self->server->volume_clone_create(
            vserver         => $self->Vserver->name,
            'parent-volume' => $self->name,
            %args
        );
        1;
    };
    if (! $eval_return) {
        #BURT 479620
        #check to see if clone is there even if the create failed
        #If so, we also guarantee that the clone volume is destroyed
        #as expected by calling clone_show (which also attaches it)

        $clone = $self->clone_show(flexclone => $args{flexclone});
        if ($clone->[0]) {
            NOTE::logcomment(
                'WARNING: clone create failed, yet clone exists!:' .
                ' volume=' . $self->name . ' clone=' . $clone->[0]->name
            );
        }

        $self->_processError($@);
    }

    # Set up $clone as an attach to the new volume:
    $clone = NOTE::GX::Volume->attach(
        name => $args{flexclone},
        Vserver => $self->Vserver,
        Node => $self->Node
    );
    $clone->is_anonymous = $self->is_anonymous;

    return $clone;
}

=head2 clone_show

Return a list-reference containing zero or more volume objects, each one
representing a clone of the calling volume. Only clones that are direct
children of the calling volume are returned (children of the clones, or
lower levels, are not returned). If the volume has no clones, an empty
list (reference) is returned.

This method has no specific parameters. Any of the options that can be
passed to C<volume clone show> may be passed as key/value pairs here, to
refine the search for clones.

Because this method is used during GNOTE tear-down, it must return valid
data even when the C<volume clone> commands are not available (either not
on the running O/S, or not licensed). Thus, no exceptions are thrown in these
cases. Instead, an empty list-reference is returned.

=over

=item Arguments

None. Options valid for C<volume clone show> may be passed (without the leading
C<->) along with the values to search for.

=item Examples

    # Get all clones of a volume
    $clones = $volume->clone_show();

    # Get only those clones where "space-guarantee-enabled" is true:
    $clones = $volume->clone_show('space-guarantee-enabled' => 1);

=back

=cut

###############################################################################
#
#   Sub Name:       clone_show
#
#   Description:    Return a list of zero or more volumes that are clones of
#                   the calling volume.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Volume object
#                   %args     in      hash      Hash of key/value pairs
#
#   Returns:        Success:    List of 0+ NOTE::GX::Volume objects
#                   Failure:    dies
#
###############################################################################
sub clone_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my (@clones, $eval_return, $data);

    # Convert these arguments from Perl booleans to strings:
    for (qw(junction-active space-guarantee-enabled)) {
        if (exists $args{$_}) {
            $args{$_} = ($args{$_} && $args{$_} !~ /false/i) ? 'true' : 'false';
        }
    }

    # This is a fairly easy one... let the volume_clone_show CDEF do all the
    # real work, and all we'll have to do is convert the list we get back to
    # a list of NOTE::GX::Volume objects.
    $eval_return = eval {
        $data = $self->server->volume_clone_show(
            vserver         => $self->Vserver->name,
            'parent-volume' => $self->name,
            %args
        );
        1;
    };
    # No exceptions thrown from here-- only an empty list returned
    return \@clones if (! $eval_return);

    # If we didn't stop from the previous operation, then $data is a listref
    # with zero or more hashrefs. Use these to create new ::Volume objects
    # that are attached to the relevant clones:
    for my $clone_record (@{$data}) {
        my $clone = NOTE::GX::Volume->attach(
            name => $clone_record->{flexclone},
            Node => $self->Node
        );
        $clone->is_anonymous = $self->is_anonymous;

        push @clones, $clone;
    }

    return \@clones;
}

=head2 clone_split

Split a volume-clone off into an independent (stand-alone) volume. Breaks
the parent-clone relationship between this volume (the clone) and the volume
it is a clone of (the parent).

Returns the calling object upon success, and throws an exception upon error.
Throws an exception if the calling object is not actually a clone.

=over

=item Arguments

None.

=item Examples

    # Create a clone
    my $clone = $vol_src->clone_create(flexclone => 'clone');
    # Split it off
    $clone->clone_split;

=back

=cut

###############################################################################
#
#   Sub Name:       clone_split
#
#   Description:    Split a clone off into a stand-alone volume. Wait for the
#                   operation to complete before returning.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub clone_split {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Just call the async method and immediately chain to job_wait(). Any
    # exceptions get bubbled up if/when they occur:
    $self->clone_split_start->job_wait();

    return $self;
}

=head2 clone_split_start

Begin an asynchronous clone-split operation. This starts the split of the
clone into its own stand-alone volume and immediately returns before the
operation completes (rather than waiting for it to complete).

The job ID of the split operation is stored internally (as with the B<move>,
B<copy>, etc. operations) and thus the object may call B<job_wait> to enter
a wait-loop until the process finishes.

Returns the calling object if the job is successfully initiated, throws an
exception otherwise. Throws an exception if the calling object is not actually
a clone.

=over

=item Arguments

None.

=item Examples

    # Create a clone
    my $clone = $vol_src->clone_create(flexclone => 'clone');
    # Start the split of the clone to a stand-alone volume
    $clone->clone_split_start;
    # Do other things while the split takes place in the background
    ...

=back

=cut

###############################################################################
#
#   Sub Name:       clone_split_start
#
#   Description:    Start an async "volume clone split start" operation and
#                   return, with the job ID stored in $self.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub clone_split_start {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $text;

    # Make sure the calling object is a clone-volume:
    confess 'Cannot call "clone_split_start" on a non-clone volume'
        if (! $self->is_clone);

    my $eval_result = eval {
        $text = $self->server->volume_clone_split_start(
            vserver    => $self->Vserver->name,
            flexclone  => $self->name,
            # This is being explicitly set to 'false' even though the default
            # for foreground is false. This is because the underlying NATE
            # job-control was getting confused on some of the calls.
            foreground => 'false'
        );
        1;
    };
    if (! $eval_result) {
        $self->_processError($@);
    }

    $self->{_job_id} = $self->Node->job_id($text);
    $self->Node->job_wait(id => $self->{_job_id}, state => 'running');

    return $self;
}

=head2 clone_split_stop

Stop a currently-running clone-split operation. This method must be called on
the volume-clone object that is currently being split.

Returns the calling object if the job is successfully stopped, throws an
exception otherwise. Throws an exception if the calling object is not actually
a clone.

=over

=item Arguments

None.

=item Examples

    # Create a clone
    my $clone = $vol_src->clone_create(flexclone => 'clone');
    # Start the split of the clone to a stand-alone volume
    $clone->clone_split_start;
    # Oops, didn't mean to do that...
    $clone->clone_split_stop;
    ...

=back

=cut

###############################################################################
#
#   Sub Name:       clone_split_stop
#
#   Description:    Stop a running "volume clone split" operation.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub clone_split_stop {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    # Make sure the calling object is a clone-volume:
    confess 'Cannot call "clone_split_stop" on a non-clone volume'
        if (! $self->is_clone);

    my $eval_result = eval {
        $self->server->volume_clone_split_stop(
            vserver   => $self->Vserver->name,
            flexclone => $self->name
        );
        1;
    };
    if (! $eval_result) {
        $self->_processError($@);
    }

    return $self;
}

=head2 clone_split_show

Shows the status of a clone-split operation on the volume object. If the
object is not a clone, the return value is C<undef>. Otherwise it is a
hash-reference containing the following keys:

=over

=item B<vserver>

The Vserver this volume is on

=item B<volume>

=item B<flexclone>

At present, these are both the same value: the name of the volume/flexclone
itself that is being split

=item B<blocks-scanned>

Number of blocks scanned so far

=item B<blocks-updated>

Number of blocks updated so far

=item B<inodes-processed>

Number of inodes processed so far

=item B<inodes-total>

Total number of inodes

=item B<inodes-percentage-complete>

Percentage of the total inodes processed, as an integer value

=back

Note that once the split operation is complete, the volume is no longer
considered a clone and the return value will change from a hash-reference
to C<undef>.

=over

=item Arguments

The basic arguments to C<volume clone split show> may be given, and are
passed through unchanged.

=item Examples

    # Get a clone
    $clone = $vol_src->clone_create(
        flexclone => $vol_src->name . '_clone'
    );
    # Start a split operation
    $clone->clone_split_start();
    # Show progress as the split is carried out:
    my $status = $clone->clone_split_show();
    my $timeout = 600;
    while ($status && $timeout) {
        NOTE::logcomment(
            sprintf(
                'Split of %s: %d/%d inodes (%d%%)',
                $clone->name,
                $status->{'inodes-processed'},
                $status->{'inodes-total'},
                $status->{'inodes-percentage-complete'}
            )
        );
        # Sleep for 5 seconds before querying status again:
        Tharn::snooze(5);
        $timeout -= 5;
        $status = $clone->clone_split_show();
    }
    # Did we get here because it completed or because of time-out?
    if ($timeout) {
        NOTE::logcomment(
            'Split of ' . $clone->name . ' completed'
        );
    } else {
        NOTE::logcomment(
            'Split of ' . $clone->name . ' not complete after ' .
            '10 minutes...'
        );
        # Stop it, or whatever you want to do at this point...
        $clone->clone_split_stop();
    }

=back

=cut

###############################################################################
#
#   Sub Name:       clone_split_show
#
#   Description:    Show the progress statistics for a backgrounded "clone
#                   split" operation. Operates on the clone represented by
#                   this object.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %args     in      hash      Additional key/value arg pairs
#
#   Returns:        Success:    hashref or undef
#                   Failure:    dies
#
###############################################################################
sub clone_split_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $data;

    # Make sure the calling object is a clone-volume. Quietly return if it
    # isn't (this effectively returns "undef"):
    return if (! $self->is_clone);

    my $eval_result = eval {
        $data = $self->server->volume_clone_split_show(
            vserver   => $self->Vserver->name,
            flexclone => $self->name,
            %args
        );
        1;
    };
    if (! $eval_result) {
        $self->_processError($@);
    }

    return $data->[0];
}

=head2 is_clone

Test whether the volume calling this method is a clone volume or not. If the
volume is a clone, the return value is the name of the parent volume. If the
volume is not a clone (or it once was, but has since been split from the
parent via B<clone_split> or B<clone_split_start>), then C<undef> is returned.

=over

=item Arguments

None.

=item Examples

    if ($volume->is_clone) {
        # Do something with the clone
    }

=back

=cut

###############################################################################
#
#   Sub Name:       is_clone
#
#   Description:    Predicate that determines if the calling volume is a
#                   clone volume or not. If it is, return the (string) name
#                   of the parent volume. Return undef if it isn't.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    Name of parent volume
#                   Failure:    undef
#
###############################################################################
sub is_clone {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($data, $eval_return);

    # Note that I'm not doing the usual "1;" trick in this eval, as the return
    # value is usable directly as-is. Either the call will succeed and I can
    # pull the 0'th element, or it will fail and $data will remain undef.
    $data = eval {
        $self->server->volume_clone_show(
            vserver   => $self->Vserver->name,
            flexclone => $self->name
        );
    };
    if ($data) {
        # If $data is set, it's a listref and I want the first element
        $data = $data->[0];
    }

    # Auto-vivication will kick in if $data is undef. In that case, a proper
    # return value of undef will be sent back.
    return $data->{'parent-volume'};
}

=head2 file_clone_create

Create a file clone at the filer level, using the C<file clone create> command
by way of a CDEF. You can create a clone of a sub-section of a file using
the C<range> option.

=over

=item Arguments

B<%args> = hash table of arguments to the C<file clone create> command:

=over

=item B<source-path> = string

The path, relative to the root of the volume, of the file to clone. This
parameter is required.

=item B<destination-path> = string

The path, relative to the root of the volume, of the new file to create. This
parameter is required.

=item B<snapshot-name> = string

Optional. The name of a snapshot on the volume to use as the source of the
file, rather than the volume itself.

=item B<range> = string

Optional. A range of blocks to clone when making a clone of just a sub-section
of the file.

=back

Because the typical usage of B<file_clone_create> is to just directly clone a
file from the volume itself (not from a snapshot), there is a shortcut
syntax available as well. In this, the method takes exactly two arguments,
the source and destination paths (in that order). No other arguments are
recognized. See the examples for more on this.

=item Examples

    # Simply clone a file:
    $vol_src->file_clone_create(
        'source-path'      => $source,
        'destination-path' => $destination
    );

    # The same operation:
    $vol_src->file_clone_create($source, $destination);

    # Clone $source from a specific snapshot:
    $vol_src->file_clone_create(
        'source-path'      => $source,
        'destination-path' => $destination,
        'snapshot-name'    => 'snapshot.0'
    );

    # Clone a sub-section of a file:
    $vol_src->file_clone_create(
        'source-path'      => $source,
        'destination-path' => $destination,
        range              => '0:0:1024'
    );

=back

=cut

###############################################################################
#
#   Sub Name:       file_clone_create
#
#   Description:    Execute the "file clone create" file command via CDEF, to
#                   create a clone of a file (or clone of a sub-section of a
#                   file).
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   %args     in      hash      Key/value pairs of arguments
#                                                 to the CDEF.
#
#   Returns:        Success:    $self
#                   Failure:    dies
#
###############################################################################
sub file_clone_create {
    my ($self, @args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my (%args, $data, $eval_return);

    # Installing the same way as in method clone_create.
    # Checking first whether flexclone license is installed or not and then
    # accordingly installing the license
    if (! $self->{__FLEXCLONE_LICENSE_PRESENT}) {
        # The second arg tells license() to install it if it isn't present
        $self->Cluster->license('FlexClone', 1);
        $self->{__FLEXCLONE_LICENSE_PRESENT} = 1;
    }

    # If they passed exactly two arguments, then this is a shortcut-syntax
    # for the general case where you're cloning a whole file from the volume
    # itself (not from a snapshot).
    if (@args == 2) {
        %args = (
            'source-path'      => $args[0],
            'destination-path' => $args[1],
        );
    } else {
        # If they didn't pass exactly two arguments, then assume they passed
        # the standard hash-table arguments:
        %args = @args;
    }

    if (! $args{'source-path'}) {
        confess 'Required parameter "source-path" missing or null';
    } elsif (! $args{'destination-path'}) {
        confess 'Required parameter "destination-path" missing or null';
    } elsif ($args{'source-path'} eq $args{'destination-path'}) {
        confess 'Source and destination paths may not be identical';
    }

    # This is a pretty simple wrap. The optional parameters in %args don't
    # need any special pre-processing and we've already validated the source
    # and dest paths.
    $eval_return = eval {
	$self->server->volume_file_clone_create(    # assuming we're running new style.
            # -Verification => 0 prevents Node.pm from trying to follow-up
            # the file_clone_create call with a file_clone_show call (which
            # doesn't exist). Just leave it in here for now.
            '-Verification' => 0,
            vserver => $self->Vserver->name,
            volume  => $self->name,
            %args
        );
        1;
    };

    if( $@ =~ /is not a recognized command/si ) {   # try old style if error is 'not a recognized command'
	$eval_return = eval {
	    $self->server->file_clone_create(
		# -Verification => 0 prevents Node.pm from trying to follow-up
		# the file_clone_create call with a file_clone_show call (which
		# doesn't exist). Just leave it in here for now.
		'-Verification' => 0,
		vserver => $self->Vserver->name,
		volume  => $self->name,
		%args
	    );
	    1;
	};
    }

    if (! $eval_return) {
        $self->_processError($@);
    }

    return $self;
}

=head2 wreck

The wreck command is used to inject persistent corruption into the file system. This
is useful for testing WAFL_check and wafliron, the file system consistency checkers.
This command is only available at the priv set test command level, and hence, is not
available in promoted builds.
For more information on usage: http://web.nane.netapp.com/mainN/files/man/man1t/na_wreck.1t.html

=over

=item Arguments

B<opts> = hash table

Hash table of arguments:

=item B<inject_comand> = string

The inject_command argument specifies a single, injection command used to corrupt the volume. This is a required parameter and the other hash entries passed to this routine will be used as options for constructing the wreck command.

=item Examples

    $sm->src->wreck(inject_command =>"pseudo_bad", level=>0,inum=>96,bno=>5 );
    $vol->wreck(inject_command =>"inode_tid_equal",inum=>96,inum2=>57);

=back

=cut

sub wreck {
    my ($self,%args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $cmd;
    my $wreck_string;
    my $output;

    confess ("Inject command not specified for wreck")
        if (not defined $args{inject_command});
    NOTE::logcomment("Name ".$self->name.", inject_command:\n".$args{inject_command});

    # constructing the wreck string with its options
    $wreck_string = delete($args{inject_command});
    while(my($key,$value) = each(%args)){
        $wreck_string .= " $key=$value";
    }

    # Constructing the cli command
    $cmd = "wreck $self->{_name} ";
    $cmd = qq( "priv set test;$cmd\'$wreck_string\'");
    $cmd = 'node run -node '.$self->{_Node}->name.$cmd;

    $output = $self->server->cli(cmd => $cmd);
    confess("Wreck failed to corrupt filesystem: $output")
        if ($output !~ /Finished/s);
    # Bringing the volume back online after the wreck command is successfully issued
    $self->online
        if $self->restricted;
}

=head2 ::upgrade_start

Start a 32 to 64-bit upgrade on the calling volume. It does not wait for
the upgrade to complete, that is left to the caller.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reason,
1) the command cannot be issued, represented by the following error messages
   a) Volume is already 64-bit
   b) Not enough space in the volume
2) the upgrade is not started

Returns:
SUCCESS: If mode != check then 0 is returned
         If mode == check then the command output is returned as a string
FAILURE: confesses on all failures

If mode == check then a reference to a hash of the statistics is returned.
Example hash,
    $href->{'volume'}           = Volume name
    $href->{'vserver'}          = Vserver name
    $href->{'used'}             = Bytes of volume used
    $href->{'available'}        = Bytes of volume available
    $href->{'total'}            = Total bytes of volume
    $href->{'capacity'}         = Percent of volume used
    $href->{'additional_space'} = Number of bytes to increase
                                  the volume by so it can be upgraded
    $href->{'fixed_size'}       = 1|0 (TRUE|FALSE)

=over

=item Arguments

I<Optional>

B<mode> => I<check|normal>

B<NOTE: Currently mode = check is not fully supported. The output returned is
a string. This will be corrected in the future.>

=item Examples

    # start an upgrade on the volume
    my $result = $vol->upgrade_start();
    if ($result == 0) {
        # upgrade started successfully
    } else {
        # upgrade failed to start
    }

    # check if the volume has sufficient space to perform an upgrade
    my $href = $vol->upgrade_start( mode => check );

=back

=cut

sub upgrade_start {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $mode, $rtn, $output, $result);

    $mode = $args{mode};
    $name = $self->{_name};

    $args{volume} = $name;
    $args{vserver} = $self->Vserver->name;
    $args{output} = \$output;

    # start the upgrade
    # for upgrade start Server returns a string instead of an array of
    # hash references.
    #   if mode != check then the string is the started message
    #       Started 64-bit upgrade on volume <name>
    #   if mode == check then the string is the upgrade check message
    #       Checking for additional space required to upgrade volume <name>...
    eval { $rtn = $self->server->volume_64bit_upgrade_start( %args ); };

    # check for errors
    # currently there are two known error messages,
    #   Volume is already 64-bit
    #   Not enough space in the volume
    if ($@) {
        confess "ERROR: 'volume 64bit-upgrade start $name' failed\n$@";
    } else {
        # check the output for errors
        if ((!defined($mode)) || ($mode !~ /check/i)) {
            if ($rtn =~ /Start(ed|ing) 64-bit upgrade on volume $name/i) {
                $result = 0;
            } else {
                confess("ERROR: unable to start 64-bit upgrade volume $name\n$rtn");
            }
        } else {
# XXX: need to convert the output from '-mode check' option to a hash - stuw
            $result = $rtn
        }
    }

    return $result;
}

=head2 ::upgrade_status

Get the status for a 64-bit upgrade that has previously been started (via
64bit-upgrade start or 64bit-upgrade scan-start) on the calling volume.

The 64bit-upgrade command is a diag level command so the privileges will be
set to diag before issuing the command and then reset to its previous value
before returning.

The method confesses for the following reasons,
1) the command cannot be issued
2) the volume being requested is not found, represented by the following
   error messages.
   a) This table is currently empty
   b) There are no entries matching your query

Returns:
SUCCESS: If any status is returned by volume_64bit_upgrade_status return
    a reference to an array of hash references. The hash is defined by
    Server::volume_64bit_upgrade_status.
FAILURE: confesses on all failures

    Example of returned hash,
    $href->{'vserver'}               = Vserver name
    $href->{'volume'}                = Volume name
    $href->{'format'}                = Volume format (32-bit, 64-bit, upgrading)
    $href->{'scanner-status'}        = State of scanner (running, stopped)
    $href->{'scan-percent-complete'} = Percent of scan completed (42%)
    $href->{'scan-time-to-complete'} = Estimated time to complete scan (hh:mm)
    $href->{'scan-progress'}         = String stating progress based on inodes
                                       and fbns (fbn 0, inode 1 of 321, Public)

=over

=item Arguments


=item Examples

    # get the status of an upgrade on the volume
    $aref = $vol->upgrade_status();
    $href = $aref->[0];   # the array should only have the status of our volume
    # process array of status hashes

    # get the status of an upgrade on the volume and catch any errors
    eval { $aref = $vol->upgrade_status(); };
    if ($@) {
        # status failed, figure out why...
    }

=back

=cut

sub upgrade_status {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($name, $server, $result);
    my $output;
    my @rtn;

    $name = $self->{_name};
    $server = $self->server;

    $args{volume} = $name;
    $args{vserver} = $self->Vserver->name;
    $args{output} = \$output;

    # get the status
    # the only possible value returned by Server is an array of hash
    # references, see header comment for hash description
    eval { @rtn = $server->volume_64bit_upgrade_status( %args ); };

    # verify the command did not fail
    if ($@) {
        # currently two error messages are known,
        #   This table is currently empty
        #   There are no entries matching your query
        confess "ERROR: 'volume 64bit-upgrade status $name' failed\n$@";
    } else {
        return \@rtn;
    }
}


=head2 ::upgrade_wait

Wait for the volume to finish upgrading


Returns:
0  : upgrade completed before timeout
>1 : upgrade did not complete before timeout

=over

=item Arguments

I<Optional>

B<timeout> => I<seconds to wait for upgrade to complete, default = 300>
B<delay> => I<seconds to wait before polling status, default = 5>


=item Examples

    # wait up to 1 hour for the upgrade to complete polling every 5 seconds
    $status = $vol->upgrade_wait( timeout => 3600 );
    if ($status != 0) {
        confess "Upgrade took too long";
    }

    # wait for 10 minutes for the upgrade to complete polling every 10 seconds
    $status = $vol->upgrade_wait( timeout => 600, delay => 10 );
    if ($status != 0) {
        confess "Upgrade took too long";
    }

=back

=cut

sub upgrade_wait {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $timeout = $args{timeout} || 300;
    my $delay = $args{delay} || 5;

    my ($time, $vols);

    $time = 0;
    do {
        Tharn::snooze($delay);
        $time += $delay;
        $vols = $self->upgrade_status();
    } until (($vols->[0]{'scanner-status'} eq '-') || ($time >= $timeout));

    return (($vols->[0]{'scanner-status'} eq '-') ? 0 : 1);
}


=head2 ::upgrade_create_mixed_format

Create a volume of mixed format, transitioning between 32-bit and 64-bit, by
starting an upgrade on a 32-bit volume and then stopping it.

Confessions:
Any confession coming from the following methods:
1) Volume::upgrade_start
2) Volume::upgrade_status
3) Aggregate::upgrade_scan_stop
As well as the following checks done in this method:
4) the scanner is not running after the upgrade has been started
5) the scanner is not stopped after the upgrade has been stopped
6) the volume format is not upgrading after the scanner has been stopped

Returns:
0 if the volume can be put into mixed format
otherwise confess

=over

=item Arguments

=item Examples

    # turn a volume that is in 32-bit format into a mixed format volume
    $status = $vol->upgrade_create_mixed_format();
    # if we get here the call succeeded

    # try to turn a 32-bit volume into mixed format and catch the failure
    eval { $status = $vol->upgrade_create_mixed_format(); };
    if ($@) {
        # could not creat a mixed format volume
    }

=back

=cut

sub upgrade_create_mixed_format {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $name = $self->{_name};
    my $aggr = $self->{_Aggregate};

    my $result;

    # don't catch any failures from the underlying methods. the caller
    # is trying to create a mixed format volume, if their parameters are
    # not correct to allow that to happen allow the test to fail. if the
    # caller wants to handle the errors they can catch them.

    # start upgrading the volume
    # don't catch the failure if the upgrade cannot be started
    $result = $self->upgrade_start();

    # stop the upgrade
    # no need to check result since scan-stop always succeeds
    $result = $aggr->upgrade_scan_stop();

    # verify the upgrade is stopped and the volume is mixed format
    $result = $self->upgrade_status();
    if ($result->[0]{'scanner-status'} !~ /stop/i) {
        confess("ERROR: scanner not stopped on volume $name: $result->[0]{'scanner-status'}");
    }

    if ($result->[0]{format} !~ /upgrading/i) {
        confess("ERROR: volume not upgrading $name: $result->[0]{format}");
    }

    return 0;
}


=head2 ::is_32bit

Check if the volume block-type (format) is 32-bit and return 1 (TRUE)
if it is


Returns:
1 if volume block-type == '32-bit'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the volume is of the correct format
    if ($vol->is_32bit == 0) {
        confess "volume is not 32-bit format";
    }

=back

=cut

sub is_32bit {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /32-bit/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::is_64bit

Check if the volume block-type (format) is 64-bit and return 1 (TRUE)
if it is


Returns:
1 if volume block-type == '64-bit'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the volume is of the correct format
    if ($vol->is_64bit == 0) {
        confess "volume is not 64-bit format";
    }

=back

=cut

sub is_64bit {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /64-bit/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::is_upgrading

Check if the volume block-type (format) is upgrading and return 1 (TRUE)
if it is


Returns:
1 if the volume block-type == 'upgrading'
0 otherwise

=over

=item Arguments

=item Examples

    # check if the volume is being upgraded, mixed format
    if ($vol->is_upgrading == 0) {
        confess "volume is not being upgraded";
    }

=back

=cut

sub is_upgrading {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->get_block_type();
    if ($fmt =~ /upgrading/i) {
        return 1;
    } else {
        return 0;
    }
}


=head2 ::get_block_type

Get the block-type of the volume


Returns:
The string representing the block-type of the volume,
    '32-bit',
    '64-bit',
    'upgrading'

=over

=item Arguments

=item Examples

    # check if this volume is being upgraded
    if ($vol->get_block_type eq 'upgading') {
        # do whatever while the volume is being upgraded
    }

=back

=cut

sub get_block_type {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $fmt = $self->status('block-type');
    return $fmt;
}

=head2 ::file_punchhole

Punch a hole in an existing file. The hole is defined by a starting block
number and the number of blocks to *punch*.


Returns:
SUCCESS: 0
FAILURE: any non-zero value

confesses
1) caller does not provide a filename

=over

=item Arguments
Arguments are passed by hash.

B<filename> => I<the name of the file to punch, file must already exist>

I<Optional>

B<start_block> => I<starting block number, 0 relative. default = 0>
B<size> => I<number of blocks. default = 1>

=item Notes

1)  The file to be punched must already exist.
2)  This is a 7-mode command so takes uses the old style path,
    /vol/<VolName>/<filename>. Filename should be defined accordingly.

=item Examples

    # create a hole starting at block 2 that is 3 blocks in size
    $status = $vol->file_punchhole( filename => 'SwissCheese',
                                    start_block => 2, size => 3 );
    if ($status != 0) {
        # could not punch a hole in SwissCheese
    }


=back

=cut

sub file_punchhole {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($server, $block_size, $max_blocks);
    my ($index_block, $remaining_blocks, $start_byte, $size_bytes);
    my ($cmd, $out, $status);

    if (!defined($args{filename})) {
        confess 'file_punchhole: caller must provide a filename';
    }

    my $fname = "/vol/$self->{_name}/$args{filename}";
    my $start_block = $args{start_block} || 0;
    my $size_blocks = $args{size} || 1;   # this will not allow size = 0

    $server = $self->server();
    $block_size = fileSystem::BLOCK_SIZE;
    $max_blocks = 1024;  # max # blocks punchhole can do at a time

    # at this time file punchhole does not support sizes > 1024 blocks
    # so split the request up into chunks no larger than 1024
    $status = 0;
    $index_block = $start_block;
    $remaining_blocks = $size_blocks;
    HOLE_CHUNK:
    while (($remaining_blocks > 0) && ($status == 0)) {
        if ($remaining_blocks > $max_blocks) {
            $size_bytes = $max_blocks * $block_size;        # do max_blocks
        } else {
            $size_bytes = $remaining_blocks * $block_size;  # do whats left
        }

        $start_byte = $index_block * $block_size;
	$fname =~ s/\/+/\//g;
        $cmd = "priv set test;file punchhole $fname $start_byte $size_bytes";
        $out = $server->dblade( cmd => $cmd );

        # check for error messages in the output. should this return
        # different statuses based on the error? - stuw
        # possible error messages are:
        #   'Offset and length must be multiple of 4096'
        #   'error punching hole: Invalid argument'
        #   'error opening file /vol/vol0/file: No such file or directory'
        #   'usage: file punchhole <filepath> <offset> <length>
        #   <offset> and  <length> must be multiple of block size'
        if (($out =~ /Offset and length must be multiple of 4096/si) || ($out =~ /error/si) || ($out =~ /usage:/si)) {
	    confess("ERROR: file_punchhole: command failed:\n$cmd\n$out");
        }
	# verifying holes. at this point we just run command to get output into log file, no checking for actual holes
	my $have_holes = undef;
	eval { $have_holes = $self->Node->exec( level => 'test', cmd=>'volume_explore',  format=>'inode', scope=>$self->{_name} . '/' . $fname ); };
	if( $have_holes =~ /ptr\[\d+\]: pvbn 0, vvbn 0/ ) {
	    NOTE::logcomment("Found holes!\n$have_holes");
	} else {
	    NOTE::logcomment("WARNING: no holes found at L1. Check L2 manually");
	}

        $index_block += $max_blocks;
        $remaining_blocks -= $max_blocks;
    }

    return $status;
}

=head2 ::create_sparse_file

Create one or more identical sparse files on a volume and enable/disable
the file reservation guarantee. The density of the file can be controlled
by changing how often blocks are written to the file when it is created,
the skip value.


Returns:
SUCCESS: the number of files correctly created (see notes below)
FAILURE: 0

confessions:
1) required arguments not provided (filename and file_size)
2) cannot set file reservation appropriately

=over

=item Arguments
Arguments are passed by hash.

B<filename> => I<the base name of the file(s) to create>
B<file_size> => I<the size of the file in bytes, can use k, m, g, or t>

I<Optional>

B<file_count> => I<The number of files to create. default = 1>
B<skip> => I<The number of blocks to skip between file writes.
             default = 0, write all blocks>
B<reservation> => I<Enable or disable the file reservation guarantee
                    [enable|disable]. default = enable>
B<timeout> => I<The number of seconds to wait for mkfile to complete.
                default = 90>
B<stream> => I<The name of the stream to create. default = no stream>

=item Notes

1)  If more than one file is to be created mkfile will append '.N' to the
    filename to create unique filenames. See the mkfile wiki page for more
    detail, http://wikid.netapp.com/w/WAFL/Commands/mkfile.
2)  Because mkfile is a 7-mode command it uses the old style directory path,
    /vol/<VolName>/<filename>.
3)  It cannot create sub-directories under /vol/<VolName>.
4)  The value returned is the number of files created correctly. This means,
    if reservation = enable the file has to be created AND its file reservation
    set to 'on'.
    If reservation = disable the number of files created is returned.
    A return value of 0 is considered a failure.
5)  Those files that are created but whose file reservation cannot be set to
    'on' will not be deleted.
6)  This uses mkfile which is a 7-mode test level command and ls which is a
    7-mode advanced level command. The command level will be set appropriately
    for their use and reset to the original level after they have been run.

=item Examples

    # create 100 500m files with only the first block used and
    # file reservation enabled
    $n_files = $vol->create_sparse_file( filename => 'NearlyEmptyFiles',
                                         file_size => '500m',
                                         file_count => 100 );
    if ($n_files != 100) {
        # not all files were created
    }

    # create 50 2g files that are roughly 25% full and
    # file reservation enabled
    $n_files = $vol->create_sparse_file( filename => 'QuarterFiles',
                                         file_size => '2g', file_count => 50,
                                         reservation => 'enable' );
    if ($n_files != 50) {
        # not all files were created
    }

    # create a full 2g file with the file reservation disabled
    eval { $n_files = $vol->create_sparse_file( filename => 'FullFile',
                                                file_size => '2g',
                                                reservation => 'disable' );
    };
    if ($@) {
        # could not disable file reservation
    } elsif ($n_files == 0) {
        # did not create file
    }
    # file was created successfully

=back

=cut

sub create_sparse_file {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($res, $out, $cmd, $n_files);
    my (@a_href, @files);

    if (!defined($args{filename})) {
        confess 'create_sparse_file: caller must provide a filename';
    }
    if (!defined($args{file_size})) {
        confess 'create_sparse_file: caller must provide a file_size';
    }

    my $file_size = $args{file_size};
    my $file_count = $args{file_count} || 1;   # this will not allow count = 0
    my $skip = $args{skip} || 0;
    my $reserve = $args{reservation} || 'enable';
    my $timeout = $args{timeout} || 90;

    my $name = $self->{_name};
    my $vserver = $self->vserver();
    my $fname = "/vol/$name/$args{filename}";

    # enable file level space guarantee on the volume
    $self->server->volume_modify( volume            => $name,
                                  'space-guarantee' => 'file',
                                  vserver           => $vserver );
    @a_href = $self->status();
    if ($a_href[0]->{'space-guarantee'} !~ /file/i) {
        confess "ERROR: create_sparse_file: could not set volume space guarantee to file, space-quarantee = $a_href[0]->{'space-guarantee'}";
    }

    # create file(s) with given parameters
    $cmd = "priv set diag;mkfile $file_size $fname $file_count $skip";
    $out = $self->server->dblade( cmd => $cmd, timeout => $timeout );

    # check for errors
    # possible error messages:
    #   /vol/vol_10/file: file creation failed
    #   /vol/vol_10/file.9: write failed
    if (($out =~ /file creation failed/i) ||
        ($out =~ /write failed/i)) {
        NOTE::logcomment("WARNING: create_sparse_file: create failed:\n$out");
        $n_files = 0
    } else {
        # determine how many files we created
        # get a list of files on the volume
        $cmd = "priv set advanced;ls /vol/$name";
        $out = $self->server->dblade( cmd => $cmd );
        # /vol/vol_10: Can't open directory: Undefined error: 0
        if ($out =~ /Can't open directory:/i) {
            confess "ERROR: create_sparse_file: cannot open directory\n$out";
        }

        # filter out the files we did not create
        @files = split(/\n|\r\n/, $out);
        @files = grep { /$args{filename}/ } @files;

        $n_files = scalar(@files);

        if ($reserve =~ /enable/i) {
            # enable file reservation for each file
            $n_files = 0;
            ENABLE_FILES:
            for my $f (@files) {
                $cmd = "file reservation /vol/$name/$f enable";
                $out = $self->server->dblade( cmd => $cmd );
                # possible error messages
                #   file: Can't find path /vol/vol_10/foobar
                #   file: /vol/vol_10 is not a file
                if (($out =~ /file: Can't find path/i) ||
                    ($out =~ /file: \S+ is not a file/i)) {
                    confess "ERROR: create_sparse_file: set reservation failed on file: /vol/$name/$f";
                }

                # if the volume is out of space there is no reason to continue
                # trying to enable file reservations
                # file: error setting file space reservation state. Not enough free disk space.
                if ($out =~ /Not enough free disk space/i) {
                    last ENABLE_FILES;
                }

                $cmd = "file reservation /vol/$name/$f";
                $out = $self->server->dblade( cmd => $cmd );
                # space reservations for file /vol/vol_10/file: on
                if ($out =~ /space reservations for file \S+: (\S+)/i) {
                    $res = $1;
                    if ($res !~ /on/i) {
                        confess "ERROR: create_sparse_file: reservation failed to enable\n$out";
                    }
                } else {
                    confess "ERROR: create_sparse_file: get reservation failed on file: /vol/$name/$f\n$out";
                }

                # increment the count of files correctly created
                $n_files++;
            }
        }
    }

    return $n_files;
}

=head2 ::get_instance_UUID

Get the instance UUID of this volume

Returns:
The string representing the instance UUID of the volume extracted from
 'volume show'.

=over

=item Arguments

=item Examples

    if ($srcvol->get_instance_UUID ne $srcvol->get_provenace_UUID) {
        # do whatever if the instance uuid and provenance uuid of src
        # volume are not the same
    }

=back

=cut

sub get_instance_UUID {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->status('instance-uuid');
}

=head2 ::get_provenance_UUID

Get the provenance UUID of this volume

Returns:
The string representing the provenance UUID of the volume extracted from
 'volume show'.

=over

=item Arguments

=item Examples

    if ($srcvol->get_instance_UUID ne $srcvol->get_provenace_UUID) {
        # do whatever if the instance uuid and provenance uuid of src
        # volume are not the same after initialization
    }

=back

=cut

sub get_provenance_UUID {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->status('provenance-uuid');
}

=head2 ::showfh

When supplied with a filepath, returns a hash reference to a hash including
the following fields:
1. flags - Any flags that have been set on the file
2. snapid - ID of the snapshot the file is present in, if any (0 if not)
3. fileid - File ID (inode number)
4. gen - Generation number
5. fsid - Filesystem ID (volume identifier)

Returns:
SUCCESS: Hash reference to a hash including the values of the above
fields
FAILURE: Hash reference to a hash with undefined values for the above
fields

Confessions:
1. required argument not supplied (filepath)
2. showfh returned an error (file or directory not found)

=over

=item Arguments
Arguments are passed by hash.

B<filepath> => path to the file (exclude leading slash)

I<Optional>

B<format> => Specifies which format values should be returned in
            [hex|dec]. default hex

B<timeout> => Specifies the number of seconds to wait until timeout
            default=90

=item Notes
1. Showfh is a 7-mode command, so it will be really be looking for
/vol/<Volume>/<filepath>.
2. Showfh is an advanced-level 7-mode command, so the mode will be set
automatically to advanced

=item Examples

my $hash_ref = $volume->showfh( filepath => "file1");
my $hash_ref2 = $vol2->showfh( filepath => "sub1/file2", format => "dec");
my $fileid = $hash_ref->{fileid};
my $flags = $hash_ref2->{flags};

=back

=cut

sub showfh{
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if(! defined $args{filepath}){
        confess 'showfh: caller must include filepath';
    }

    my $filepath = $args{filepath};
    my $format = $args{format} || 'hex';
    my $timeout = $args{timeout} || 90;

    my $volname = $self->name();
    my $cmd = "priv set advanced;showfh /vol/$volname/$filepath";

    my $out;
    my $result = eval{
                    $out = $self->server->dblade( cmd => $cmd, -Timeout => $timeout); 1;
                };
    if(! $result){
        confess $@;
    }

    # Test output for error
    if($out =~ /No such file or directory/){
        confess 'ERROR: showfh: No such file or directory';
    }

    # This array contains the priv set warning and showfh output
    my @lines_out = split /\n|\r\n/, $out;

    # This array will contain all the fields of the format "fileid=0x000"
    my @split_out;

    # Test each line to determine which is the output we want
    foreach my $line (@lines_out){
        if ($line =~ /fileid/){
            @split_out = split /\s/, $line;
        }
    }

    # This is the hash reference that will be returned
    my $hash_ref = {};

    # Now that we have extracted the items from the string, we must
    # separate keys from values and save them in a hash reference
    foreach my $unparsed_pair (@split_out){
        my @field_value_pair = split /=/, $unparsed_pair;

        #Convert to decimal if requested and add to hash
        if($format eq 'dec'){
            $hash_ref->{$field_value_pair[0]} = hex $field_value_pair[1];
        } else {
            $hash_ref->{$field_value_pair[0]} = $field_value_pair[1];
        }
    }

    return $hash_ref;
}

=head2 ::rdfile

When supplied with a filepath, reads and returns the contents of that file
if it exists.

Returns:
SUCCESS: raw output string
FAILURE: confess "ERROR: rdfile: No such file or directory"

Confessions:
1. required argument not supplied (filepath)
2. rdfile returned an error (file or directory not found)

=over

=item Arguments
Arguments are passed by hash.

B<filepath> => path to the file (exclude leading slash)

I<Optional>

B<timeout> => Specifies the number of seconds to wait until timeout
             default=90

=item Notes
1. rdfile is a 7-mode command, so it will be looking for
   /vol/<Volume>/<filepath>.

=item Examples

my $output = $volume->rdfile( filepath => "file1");

=back

=cut

sub rdfile{
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if(! defined $args{filepath}){
        confess 'rdfile: caller must include filepath';
    }

    my $filepath = $args{filepath};
    my $timeout = $args{timeout} || 90;

    my $volname = $self->name();
    my $cmd = "rdfile /vol/$volname/$filepath";

    my $out;
    my $result = eval{
                    $out = $self->server->dblade( cmd => $cmd, -Timeout => $timeout); 1    ;
                };
    if(! $result){
        confess $@;
    }

    #Test output for error
    if($out =~ /No such file or directory/){
        confess 'ERROR: rdfile: No such file or directory';
    }

    return $out;
}

=head2 ::is_bre

Check if volume supports physical replica or not and returns true or false string.

=over

=item Arguments
None

=back

=cut

sub is_bre {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    return ($self->Node->exec( level=>"diag", cmd=>"volume_show", vserver => $self->Vserver->name, volume => $self->name, ))->[0]->{'physical-replica'};
}

=head2 flexcache_create_single

Creates a single flexcache for the calling volume by calling 'volume create'.
A flexcache cannot be added to volumes of 'type LS' or 'state restricted'.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<name> => name for cache volume being created, default: GNOTE generated

=item B<-Timeout> => time in seconds to wait for command to complete,
    default: 240 seconds

=back

=item Returns

=over

=item B<Success:> name of cache volume created

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if the calling volume is of type LS

=item 2 Method will confess if the calling volume is in the restricted state

=item 3 Method will confess if command fails for any reason

=back

=item Notes

=over

=item 1 This method does not support any of the command line options for
    'volume create'. The following options will be provided by GNOTE

=over

=item * flexcache-orgin-volume = calling volume name

=item * aggregate = calling volume aggregate

=item * type = DC

=back

=back

=item Examples

=back

    # create a cache volume for the calling volume
    $vol->flexcache_create_single();

    # create a cache volume, cache_vol, for the calling volume
    $vol->flexcache_create_single( name => 'cache_vol' );

=cut

sub flexcache_create_single {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts;
    my ($ret_eval, $name);

    # Checking for LS volumes
    if ($self->status('type') eq 'LS') {
        confess 'Cannot create Flexcache for Volumes of type LS ';
    }

    # Checking for volume being restricted
    if ($self->restricted) {
        confess 'Cannot create Flexcache for a restricted volume';
    }

    # if no name given, create a random name for the Flexcache
    $name = $args{name} ? $args{name} : NOTE::Volume->newName('fc');

    # Name of the flexcache to be created for this Volume
    $opts{'volume'} = $name;

    $opts{'flexcache-origin-volume'} = $self->name;
    $opts{'aggregate'} = $self->Aggregate->name;
    $opts{'type'} = 'DC';
    # increasing the server wait time, as the command takes long to run
    $opts{'-Timeout'} = $args{'-Timeout'} || 240;
    $opts{'-Verification'} = undef;

    $ret_eval = eval {
        $self->server->volume_create(%opts);
        1;
    };

    if (!$ret_eval) {
        confess "Flexcache creation failed : $@";
    }

    # mark that this volume has a flexcache volume associated with it
    $self->{_flexcache_name} = 1;

    return $name;
}

=head2 flexcache_create

Add cluster-wide cache volumes to the calling volume by calling
'volume flexcache create'. Cache volume cannot be added to volumes of
'type LS' or 'state restricted'.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<-Timeout> => time in seconds to wait for command to complete,
    default: 240 seconds

=back

=item Returns

=over

=item B<Success:> The calling object

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if the calling volume is of type LS

=item 2 Method will confess if the calling volume is in the restricted state

=item 3 Method will confess if command fails for any reason

=back

=item Notes

=over

=item 1 There are only two options to this command, origin-volume and
    vserver. These will be assigned by GNOTE.

=back

=item Examples

=back

    # create cluster-wide cache volumes for the calling volume
    $vol->flexcache_create();

    # create cluster-wide cache volumes for the calling volume waiting
    # 10 minutes (600 seconds) for the command to complete
    $vol->flexcache_create( '-Timeout' => 600 );

=cut

sub flexcache_create {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts;
    my ($rtn_eval, $rtn_msg, $output);

    # Checking for LS volumes
    if ($self->status('type') eq 'LS') {
        confess 'Cannot create Flexcache for Volumes of type LS ';
    }

    # Checking for volume being restricted
    if ($self->restricted) {
        confess 'Cannot create Flexcache for a restricted volume';
    }

    $opts{'origin-volume'} = $self->name;
    $opts{'vserver'} = $self->Vserver->name;

    # increasing the server wait time, as the command takes long to run
    $opts{'-Timeout'} = $args{'-Timeout'} || 240;
    $opts{'-Verification'} = undef;
    $opts{'output'} = \$output;

    $rtn_eval = eval {
        $rtn_msg = $self->server->volume_flexcache_create( %opts );
        1;
    };

    # catch command failures in the form of traps or error messages
    if (!$rtn_eval) {
        confess "Flexcache command failed : $@";
    } elsif ($output =~ /error|fail/i) {
        confess "Flexcache creation failed : $output";
    }

    # set the flexcache flag for this volume
    $self->{_flexcache_name} = 1;


    return $self;
}

=head2 flexcache_delete

Deletes all the cache volumes associated to the calling volume using
the 'volume flexcache delete' command.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<-Timeout> => time in seconds to wait for command to complete,
    default: 240 seconds

=back

=item Returns

=over

=item B<Success:> the calling volume object

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if the calling volume is of type LS

=item 2 Method will confess if command fails for any reason

=back

=item Notes

=over

=item 1 There are only two options to this command, origin-volume and
    vserver. These will be assigned by GNOTE.

=back

=item Examples

    # delete all the cache volumes for $vol
    $vol->flexcache_delete();

    # delete all the cache volumes for $vol waiting 10 minutes
    # (600 seconds) for the command to complete
    $vol->flexcache_delete( '-Timeout' => 600 );

=back

=cut

sub flexcache_delete {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my %opts;
    my $ret_eval;

    if ( !defined($self->{_flexcache_name}) ) {
        confess('There are no flexcache volumes for volume ' . $self->name);
    }

    $opts{'vserver'} = $self->Vserver->name;
    $opts{'origin-volume'} = $self->name;
    # increasing the server wait time, as the command takes long to run
    $opts{'-Timeout'} = $args{'-Timeout'} || 240;
    $opts{'-Verification'} = undef;

    $ret_eval = eval {
        $self->server->volume_flexcache_delete( %opts );
        1;
    };

    if (!$ret_eval) {
        confess "Flexcache volumes for this volume not deleted : $@";
    }

    $self->{_flexcache_name} = undef;

    return $self;
}

=head2 flexcache_show

Shows the status of all the cache volumes associated with the calling volume
by calling 'volume flexcache show'.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

All optional arguments default to not being used in the command unless
stated otherwise.

=over

=item B<fields> => Name of the field(s) to display

=item B<cache-volume> => Name of the cache volume to get status for

=item B<cache-aggregate> => Name of aggregate containing cache volume(s)
    of interest

=item B<cache-size> => Show cache volume(s) of this size only

=item B<cache-state> => I<online|restricted|offline|force-online> Show
    cache volume(s) of this state only

=item B<cach-available> => I<integer[KB|MB|GB|TB|PB]> Show only cache
    volume(s) with this available space

=item B<cache-percent-used> => I<0..100> Show cache volume(s) only with
    this percentage used

=item B<connection-status> => I<text> Show cache volume(s) with only this
    connection status

=item B<origin-aggregate> => Name of aggregate containing the orginal volume(s)
    of interest

=item B<origin-state> => I<online|restricted|offline|force-online> Show only
    cache volume(s) for volume(s) of this state

=item B<all> => I<0|1> If 1 show all the cache volumes in the cluster,
    otherwise show only the cache volume(s) associated with the calling
    volume (set origin-volume to the calling volume name, and vserver
    to the calling volume vserver) [Default: 0]

=back

=item Returns

=over

=item B<Success:> returns a reference to an array of hashes (see Notes
    for a description of the hash), or an empty array if no cache volumes

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if command fails for any reason

=item 2 Method will confess if command reports an error or failure

=over

=item * Failed to create cache volume aFC_d0_v0_cache_aFC_s0 in aggr aFC_s0

=item * Error: command failed: [Job 1056] Job failed: Failed to create the
    volume on node

=back

=back

=item Notes

=over

=item 1 The '-instance' command option is not supported by this method.
    NATE sets the filer output to showallfields = true so it is not needed.

=item 2 There are two options to this command that are not arguments to
    this method, origin-volume and vserver. These will be assigned by GNOTE
    when appropriate (seel 'all' argument above).

=item 3 Though this method accepts all the arguments listed above Server
    seems to only use two arguments, origin-volume and vserver. The others
    will not cause Server to fail they will just not be used in the command.

=item 4 This is the hash representing the status of each cache volume

=over

=item $href->{'cache-state'} => state of the cache volume

=item $href->{'cache-volume'} => name of the cache volume

=item $href->{'vserver'} => name of vserver containing cache volume

=item $href->{'cache-aggregate'} => name of aggregate containing cache volume

=item $href->{'origin-volume'} => name of volume cache represents

=item $href->{'origin-aggregate'} => name of aggregate containing original
    volume

=item $href->{'cache-available'} => free space in cache

=item $href->{'connection-status'} => state of connection to original volume

=item $href->{'cache-size'} => total size of cache

=item $href->{'origin-state'} => state of original volume

=item $href->{'cache-percent-used'} => percent of cache space used

=back

=back

=item Examples

    # get the status for all the cache volumes associated with the $vol
    $a_href = $vol->flexcache_show();

    # get the status for all the cache volumes in the cluster containing $vol
    # note, $vol does not have to have cache volumes itself
    $a_href = $vol->flexcache_show( 'all' => 1 );

    # get the status for cache volume cvol associated with $vol
    $a_href = $vol->flexcache_show( 'cache-volume' => 'cvol' );
    $href_cvol_status = $a_href->[0];

=back

=cut

sub flexcache_show {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($rtn_eval, $rtn_msg, $output, $result, $all);

    # if argument all is 1 then display all the cluster-wide caches
    # otherwise only show the caches associated with this volume
    $all = defined($args{all}) ? delete($args{all}) : 0;
    if (0 == $all) {
        $args{'origin-volume'} = $self->name;
        $args{'vserver'} = $self->Vserver->name;
    }

    # increasing the server wait time, as the command takes long to run
    $args{'-Timeout'} = $args{'-Timeout'} || 240;
    $args{'-Verification'} = undef;
    $args{output} = \$output;   # get stderr to check for error messages

    $rtn_eval = eval {
        $rtn_msg = $self->server->volume_flexcache_show( %args );
        1;
    };

    if ( (!$rtn_eval) &&
         ($output !~ /There are no entries matching your query/i) ) {
        confess "Flexcache for this volume does not exist : $@";
    }

    $result = [];   # return an empty array for no entries

    # if we got some status convert the raw filer output to an array
    # of hashes, what Server would normally return (Server is returning
    # the raw filer output for volume_flexcache_show for some reason)
    if ($output !~ /There are no entries matching your query/i) {
        $result = _server_to_array_of_hashes( output => $rtn_msg );
    }

    return $result;
}

=head2 ::snapshot_tags

Returns an array of the snapshot tags for the given snapshot of the volume.
Returns an empty array, if there are no tags for this snapshot.
Dies, if there is an error getting details for this snapshot.

=over

=item Arguments

B<snapshot> => I<snapshot name>

I<Optional>

None.

=item Example

=back

     my $ss_tags_aref = $vol->snapshot_tags(snapshot => $snapshot_name);

=cut

sub snapshot_tags {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $snapshot = $args{snapshot};

    confess 'snapshot name to be provided' if(!defined $snapshot);

    my $sslist;
    my $eval_return = eval {
                         $sslist = $self->server->volume_snapshot_show( vserver => $self->Vserver->name, volume => $self->name );
                         1;
                      };
    if(!$eval_return){
        #means, $@ is set some value
        confess "Could not get snapshot details: $@";
    }

    my $ss_tags_aref = [];
    for my $ss ( @{$sslist} ) {
        if($ss->{snapshot} eq $snapshot) {
            if ($ss->{tags} ne q{-}) {
                if ( ref($ss->{tags}) eq 'ARRAY' ) {
                    $ss_tags_aref = $ss->{tags};
                }
                else {
                    push @{$ss_tags_aref}, $ss->{tags};
                }
            }
        }
    }
    return $ss_tags_aref;
}

=head2 ::deswizzle_check

Function checks for de-swizzle scanner on this volume.
Will croak if scanner is not running or if volume is offline.

=over

=item Arguments

None

=item Examples

    $vol->deswizzle_check;

=back

=cut

sub deswizzle_check {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my ($command,$out,$ret);

    if(($self->status('state')) eq 'offline') {
        confess 'volume is offline';
    }

    $ret =  eval {$out = $self->wafl_scan( option => 'status' ); 1;};

    if($ret && $out =~ /volume deswizzling/i){
        NOTE::logcomment('de-swizzle scanner triggered as expected');
    } else {
       confess "de-swizzle scanner not triggered as expected due to: $@";
    }
    return;
}

=head2 ::snapshot_autodelete_modify

Modify the snapshot autodelete settings for the given snapshot on the calling volume. The arguments hash
controls what is modified.

Returns the calling object if successful, croaks if there is an error

=over

=item Arguments

B<opts> = hash table

  Hash table of arguments to the C<volume snapshot autodelete modify> command:
  Need to provide atleast one value out of the below mentioned arguments

=item B<enabled> = I<boolean>

autodelete setting enabled or disbaled.true/false.

=item B<commitment> = I<string>

can be one of the values. try, disrupt, destroy.

=item B<defer-delete> = I<string>

can be one of the values. scheduled, user_created, prefix, none.

=item B<delete-order> = I<string>

can be one of the values. newest_first, oldest_first.

=item B<defer-delete-prefix> = I<prefix string>

prefix string.

=item B<target-free-space> = I<integer>

% of target free space on the volume

=item B<trigger> = I<string>

can be one of the values. volume, snap_reserve, space_reserve

=item B<destroy-list> = I<boolean>

can be one of the values. lun_clone, vol_clone, cifs_share,
                          sfsr, file_clone or none

=item Example

=back

     $dst_vol->snapshot_autodelete_modify(enabled => 'false',
                                         commitment => 'disrupt',
                                         defer-delete => 'user_created',
                                         delete-order => 'oldest_first',
                                         defer-delete-prefix => 'abc def',
                                         target-free-space => '30%',
                                         trigger => 'snap_reserve',
                                         destroy-list => 'none',
                                       );

=cut

sub snapshot_autodelete_modify {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    #If user specifies the 'enabled' parameter, convert it to a string:
    if (exists $args{enabled}) {
        $args{enabled} = ($args{enabled} && $args{enabled} !~ /false/i)
            ? 'true' : 'false';
    }

    # If user pass in defer-delete-prefix, make sure it is passed in double quotes
    if ($args{'defer-delete-prefix'}) {
        $args{'defer-delete-prefix'} = qq("$args{'defer-delete-prefix'}");
    }

    my $eval_return = eval {
                     $self->server->volume_snapshot_autodelete_modify(
                             vserver => $self->Vserver->name,
                             volume  => $self->name,
                             %args
                          );
        1;
    };
    if (! $eval_return) {
        $self->_processError($@);
    }

    return $self;
}

=head2 snapshot_autodelete_show

Runs <volume snapshot autodelete show> on the given snapshot on the calling
volume and return the reference to hash containg autodelete settings

croaks if there is an error.

=over

=item Arguments

None.

=item Return Value

Returns a reference to an hash, mentioned below.
   {
     'trigger' => 'snap_reserve',
     'defer-delete-prefix' => '"abc def"',
     'volume' => 'vl19217_Owen_numbered',
     'vserver' => 'vs0',
     'delete-order' => 'oldest_first',
     'destroy-list' => 'none',
     'target-free-space' => '30%',
     'commitment' => 'disrupt',
     'defer-delete' => 'user_created',
     'enabled' => 'false'
   }

=item Examples

     my $autodel_settings = $src_vol->snapshot_autodelete_show();

=back

=cut

sub snapshot_autodelete_show {
    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($eval_return, $autodel_show);

    $eval_return = eval {
        $autodel_show = $self->server->volume_snapshot_autodelete_show(
            vserver => $self->Vserver->name,
            volume  => $self->name
        );
        1;
    };
    if (! $eval_return) {
        $self->_processError($@);
    }

    # The return value from snapshot_autodelete_show is a list-reference
    # that contains one hash refernce, which is to be returned
    return $autodel_show->[0];
}

# build the volume path from the description provided and return the value
# description syntax: [cluster:][//vserver/]volume

sub _build_volume_path {
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($path, $desc);

    $desc = $args{description};

    $path = $self->Cluster->name . ':' if ($desc =~ /cluster/i);
    $path .= '//' . $self->Vserver->name . '/' if ($desc =~ /vserver/i);
    $path .= $self->{_name};

    return $path;
}

=head2 ::language

Returns multi-line language string if no arguments are given. Returns specific language
if language identifier is given and matches one of volume's language identifier or
returns hash reference to all volume's language sets if requested identifier is not
found within returned sets.

=over

=item Arguments

B<language>

I<optional>

Language identifier string or simple boolean 'true' value.

=item Examples

=back

  if( $volA->language ne $volB->language ) {
    call_interpreter_please() || confess "dont sprechen sie deutsch!";
  }

complete working example:

      1 use GNOTE;
      2 use Data::Dumper;
      3 eval {
      4     my $ut = NOTE::SnapMirrorUtil->new();
      5     my ($vl) = $ut->create_volume(tag=>'src');
      6
      7     my %out = (
      8                 'single string of everything'   => $vl->language(),
      9                 'pull specific language id'     => $vl->language('oem-character-set'),
     10                 'comple hash of languages'      => $vl->language(1)
     11               );
     12
     13     print "Language sets:\n". Dumper(\%out);
     14 };
     15 logresult('FAIL',$@) if $@;
     16 NOTE::teardown;

and the output:

 20101110 185301 [0001_language] .         Language sets:
 20101110 185301 [0001_language] .         $VAR1 = {
 20101110 185301 [0001_language] .                   'pull specific language id' => 'cp437',
 20101110 185301 [0001_language] .                   'comple hash of languages' => {
 20101110 185301 [0001_language] .                                                   'oem-character-set' => 'cp437',
 20101110 185301 [0001_language] .                                                   'nfs-character-set' => 'iso-8859-1',
 20101110 185301 [0001_language] .                                                   'language' => 'en_US',
 20101110 185301 [0001_language] .                                                   'language-str' => 'English (US)'
 20101110 185301 [0001_language] .                                                 },
 20101110 185301 [0001_language] .                   'single string of everything' => 'en_US, English (US), cp437, iso-8859-1'
 20101110 185301 [0001_language] .                 };

=cut

sub language {
    my ($self,$lang) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics
    my $lstr = $self->server->cli(cmd =>'set -showallfields false;volume language -vserver ' . $self->Vserver->name . ' -volume ' . $self->name . ';set -showallfields true');  # BURT 460192
    my @lset = $lstr =~ /language is (\S+) \((.*)\) for .*OEM Character set is ([^|]+).*NFS Character set is ([^|]+)/s; # store all language ids as array
    return join(', ',@lset) unless $lang;                                       # return readable string if no specifics requested
    my %lhsh = map{$_,shift @lset} qw/language language-str oem-character-set nfs-character-set/;           # convert array to hash if specific set requested
    return $lhsh{$lang} ? $lhsh{$lang} : \%lhsh;                                    # return given identifier or complete hash ref
}

=head2 ::snapshot_restore_file

restores the given file from the given snapshot, to the same volume.
cannot be restored to a different volume.
dies if error in restoring the file.

=over

=item Arguments

B<file> = I<file>
name of the file. relative path from the volume mountpoint

B<snapshot> = I<snapshot>
name of the snapshot from which file is to be restored

I<optional>

B<restore_path> = I<restore_path>
restore path, relative path from the volume mountpoint.
can be restored to same volume only.

=item Examples

=back

    #Code snippet using this function:
    #Restoring it to the same location
    $vol_src->snapshot_restore_file(file => $file_name, snapshot => $ss);

    #to restore it to a difefrent dir in the same volume
    $vol_src->snapshot_restore_file(file => $file_name,
                                    snapshot => $ss,
                                    restore_path => "$restore_dir/$restore_filename");

=cut

sub snapshot_restore_file {
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my $file     = $args{file};
    my $snapshot = $args{snapshot};
    my $restore_path = $args{restore_path};

    for ('file','snapshot') {
        confess "$_ needs to be provided" if(! $args{$_});
    }

    my $volume_name = $self->name;
    my $path = "/vol/$volume_name/$file";
    if($restore_path){
        $restore_path = "/vol/$volume_name/$restore_path";
    }
    else { $restore_path = $path; }

    my ($eval_return, $restore_output);
    $eval_return = eval {
                      $restore_output = $self->server->volume_snapshot_restore_file(
                                                'volume' => $self->name,
                                                'vserver' => $self->Vserver->name,
                                                'snapshot' => $snapshot,
                                                'path' => $path,
                                                'restore-path' => $restore_path,
                                               );
                                        1;
                   };
    if(!$eval_return) {
        $self->_processError($@);
    }

    return;
}

=head2 ::is_dp

Function checks whether volume is dp volume or not
Will return 1 if volume is dp else returns 0

=over

=item Arguments
None

=back

=over

=item Examples

    if($vol->is_dp) {
        NOTE::logcomment('Volume is dp');
    } else {
        NOTE::logcomment('Volume is not dp');
    }

=back

=cut

sub is_dp {

    my $self = shift;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    return $self->status->{type} =~ /dp/i;

}

###############################################################################
#
#   Sub Name: _server_to_array_of_hashes
#
#   Description:
#       convert the raw filer output to the array of hashes that is
#       normally returned from Server.pm
#
#   Agruments:      NAME        IN/OUT  TYPE        DESCRIPTION
#                   output      IN      scalar      raw filer output
#                   separator   IN      scalar      character(s) used to
#                                                   separate output fields,
#                                                   optional [X;X]
#
#   Globals:
#
#   Environment:
#
#   Returns:        Success:    an array of hashes denoting the output
#                   Failure:    confesses if the number of keys does not
#                               equal the number of values
#
###############################################################################
sub _server_to_array_of_hashes {
    my (%args) = @_;

    my ($s_keys);
    my (@a_value_lines, @a_keys, @a_values, @a_href);

    my $separator = defined($args{separator}) ? $args{separator} : 'X;X';

    # this assumes a standard format for Server.pm output,
    # line 0: hash key values
    # line 1: table column headers
    # line 2-n: hash entry values

    # split the string into its components, keys, headers, and lines of values
    ($s_keys, undef, @a_value_lines) = split(/\n|\r\n/, $args{output});

    # separate the key string into an array of keys
    @a_keys = split(/$separator/, $s_keys);

    # process each value string turning it into a hash and adding it to the
    # array of hashes
    @a_href = ();
    #for (my $i = 0; $i < scalar(@a_value_lines); $i++) {
    foreach my $i (0..$#a_value_lines) {
        @a_values = split(/$separator/, $a_value_lines[$i]);

        if (scalar(@a_values) != scalar(@a_keys)) {
            confess("ERROR: number of keys != number of values\n");
        }

        @{$a_href[$i]}{@a_keys} = @a_values;
    }

    return \@a_href;
}

=head2 get_flexcache_names

Get the names of all the cache volumes for the calling volume.

=over

=item Arguments

Arguments passed by hash table

Hash table of arguments:

I<Optional Arguments>

=over

=item B<-Timeout> => time in seconds to wait for command to complete,
    default: 240 seconds

=back

=item Returns

=over

=item B<Success:> a reference to an array containing the names of the cache
    volumes associated with the calling volume

=item B<Failure:> confesses

=back

=item Confessions

=over

=item 1 Method will confess if the calling volume is of type LS

=item 2 Method will confess if the 'flexcache show' command fails for
    any reason

=back

=item Notes

=item Examples

=back

    # get the cache volume names for $vol
    $a_cvol_names = $vol->get_flexcache_names();

    # get the cache volumes for the $vol waiting 10 minutes (600 seconds)
    # for the command to complete
    $a_cvol_names = $vol->get_flexcache_create_names( '-Timeout' => 600 );

=cut

sub get_flexcache_names {
    my ($self, %args) = @_;

    my $a_href;
    my $a_cvol_names = [];    # initialize to empty array

    $a_href = $self->flexcache_show( %args );

    if (defined($a_href) && (scalar(@{$a_href}) > 0)) {
        @{$a_cvol_names} = map { $_->{'cache-volume'} } @{$a_href};
    }

    return $a_cvol_names;
}

=head2 ::slist

When supplied with a filepath, returns an array of hashes with the following
fields:
1. inode
2. size
3. filepath

Returns:
SUCCESS: Array of hashes for the files related to the filepath (including
the info of the basefile at the start of the array!)
FAILURE: Empty array

Confessions:
1. required argument not supplied (filepath)
2. slist returned an error (file or directory not found)

=over

=item Arguments
Arguments are passed by hash.

B<filepath> => path to the basefile with stream(s) (exclude leading slash)

I<Optional>

B<timeout> => Specifies the number of seconds to wait until timeout
            default=90

=item Notes
1. The first hash in the array corresponds to the info of the basefile
that was searched upon
2. slist is a 7-mode command, so it will be really be looking for
/vol/<Volume>/<filepath>.
3. slist is a test-level 7-mode command, so the mode will be set
automatically to test

=item Examples

my @slist_output = $volume->slist( filepath => "file1");
my @slist_output2 = $vol2->slist( filepath => "dir1/file2");
my $filepath = $slist_output[-1]->{inode};

=back

=cut

sub slist{
    my ($self, %args) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    if(! defined $args{filepath}){
        confess 'slist: caller must include filepath';
    }

    my $filepath = $args{filepath};
    my $timeout = $args{timeout} || 90;

    my $volname = $self->name();
    my $cmd = "priv set -q diag;slist /vol/$volname/$filepath";

    my $out;
    my $result = eval{
                    $out = $self->server->dblade( cmd => $cmd,
                                                  -Timeout => $timeout);
                    1;
                };
    if(! $result){
        confess $@;
    }

    # Test output for error
    if($out =~ /Invalid path/i){
        confess 'ERROR: slist: Invalid path';
    }elsif($out =~ /Invalid parameter/i){
        confess 'ERROR: slist: Invalid parameter';
    }

    # This array contains the priv set warning and slist output
    my @lines_out = split /\n|\r\n$/, $out;

    # This array of hashes will contain the returned fields in the format
    # inode=>x, size=>y, filepath=>z
    # NOTE: The first element of the array will always contain info about
    # the parent file, not a stream!
    my @filter_out;

    # Test each line to determine which is the output we want
    foreach my $line (@lines_out){
        if ($line =~ /^\s+(\d+)\s+(\d+)\s+((\w|\/)+)/){
            push(@filter_out, {inode => $1, size => $2, filepath => $3});
        }
    }

    return @filter_out;
}

=head2 ::mkfile

Create a file on a volume without having to go through a client. For more
advanced option to create a file with a certain block layout, use create_sparse_file

Returns:
SUCCESS: Nothing
FAILURE: Confesses Error string depending on the error

confessions:
1) required arguments not provided (filename and file_size)
2) write fails or file not created

=over

=item Arguments
Arguments are passed by hash.

B<filename> => I<the base name of the file(s) to create>
B<file_size> => I<the size of the file in bytes, can use k, m, g, or t>

I<Optional>

B<timeout> => I<The number of seconds to wait for mkfile to complete.
                default = 90>
B<stream> => I<The name of the stream to create. default = no stream>

=item Notes

1)  See the mkfile wiki page for more detail,
    http://wikid.netapp.com/w/WAFL/Commands/mkfile.
2)  Because mkfile is a 7-mode command it uses the old style directory path,
    /vol/<VolName>/<filename>.
3)  It cannot create sub-directories under /vol/<VolName>.
4)  mkfile is a 7-mode test level command which requires the advanced privilege
    level. The command level will be set appropriately
    for use and reset to the original level after they have been run.

=item Examples

    # Will create a 500m file
    $vol->mkfile( filename => 'file1', file_size => '500m'};

    # Will create a 500m stream using file1 as the basefile
    $vol->mkfile( stream => 'stream_of_file1', filename => 'file1', file_size => '500m'};

=back

=cut

sub mkfile{
    my ( $self, %args ) = @_;
    my $trace = $NOTE::TRACELOG->start_timer();  # log internal metrics

    my ($res, $out, $cmd);
    my (@a_href, @files);

    if (!defined($args{filename})) {
        confess 'mkfile: caller must provide a filename';
    }
    if (!defined($args{file_size})) {
        confess 'mkfile: caller must provide a file_size';
    }

    my $file_size = $args{file_size};
    my $timeout = $args{timeout} || 90;
    my $stream = $args{stream} ? "-x $args{stream}" : q{};  #empty because mkfile syntax reads streamname before filename arg

    my $name = $self->{_name};
    my $vserver = $self->vserver();
    my $fname = "/vol/$name/$args{filename}";

    # create file with given parameters
    $cmd = "priv set diag;mkfile $stream $file_size $fname";
    $out = $self->server->dblade( cmd => $cmd, timeout => $timeout );

    # check for errors
    # possible error messages:
    #   /vol/vol_10/file: file creation failed
    #   /vol/vol_10/file: write failed
    if (($out =~ /file creation failed/i) || ($out =~ /write failed/i)) {
        confess "WARNING: mkfile: create failed:\n$out";
    }
}
1;
