# This is basically -*- perl -*-
# $Id: ontap/main/iconnect/lib/Command_Testing_Usage.pm
#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
# Iconnect/Command_Testing_Usage.pm
##
## @summary
##    Library for Command Testing Scripts which contians the 'ic' command usage
##    statements.
##
## @description
##    This library is for the Command Testing suite of scripts.
##
##    This file contains the usage statements for the 'ic' family of commands.
##    This should not likely be called outside of Command_Testing.pm.
##
## @synopsis
##   use Iconnect::Command_Testing_Usage qw(:ALL);
#
#
##
## @example
##   use Iconnect::Command_Testing qw(:ALL);
##
## @burt 764125 Com_Test:  Need to add SN.2 to Command_Testing_Usage.pm
## @burt 682902 Command_Testing Suite check-in
## @burt 781806 Com_Test_ic: 'ic nvthruput' @ test privilege
## @burt 785024 Move Command_Testing libs to /lib/Iconnect/
##
## @keywords
##   interconnect; ic; Command_Testing
##
## @status public
##
## @author Interconnect QA (dl-interconnect-qa@netapp.com)
## @author Paul Waters (pwaters@netapp.com)

use strict;
use warnings;

package Iconnect::Command_Testing_Usage;
use Exporter 'import';
use TCD;

our @EXPORT_OK = qw(
    %USAGE_TABLE
);
our %EXPORT_TAGS = (
    ALL => [qw(
        %USAGE_TABLE
    )],
);

use Readonly;

################################################################################
#  Name        : NO_COMMANDS_AT_PRIV
#  Description : No usable commands at this privilege level
#  Privilege   : Usually admin
#  Driver      : All
################################################################################
Readonly our %NO_COMMANDS_AT_PRIV => (
    SN0 => {
        DEBUG    => 'cmd_ic: No usable commands at this privilege level.',
        NONDEBUG => q{},
    },
    SN1 => {
        DEBUG    => 'No usable commands at this privilege level.',
        NONDEBUG => 'No usable commands at this privilege level.',
    },
    SN2 => {
        DEBUG    => 'No usable commands at this privilege level.',
        NONDEBUG => 'No usable commands at this privilege level.',
    },
);

################################################################################
#  Name        : COMMAND_NOT_FOUND
#  Description : Error message at admin level for 'ic <subcommand>'
#  Privilege   : Usually admin
#  Driver      : All
################################################################################
Readonly our %COMMAND_NOT_FOUND => (
    SN1 => {
        DEBUG    => 'Error: Command not found.  Type \'?\' for a list of commands',
        NONDEBUG => 'Error: Command not found.  Type \'?\' for a list of commands',
    },
    SN2 => {
        DEBUG    => 'Error: Command not found.  Type \'?\' for a list of commands',
        NONDEBUG => 'Error: Command not found.  Type \'?\' for a list of commands',
    },
);

################################################################################
#  Name        : COMMAND_NOT_SUPPORTED_ON_PLATFORM
#  Description : Command not supported on platform error message
#  Privilege   : All
#  Driver      : All
################################################################################
Readonly our %COMMAND_NOT_SUPPORTED_ON_PLATFORM => (
    SN0 => {
        DEBUG    => 'ERROR: Command not supported on this platform.',
        NONDEBUG => 'ERROR: Command not supported on this platform.',
    },
    SN1 => {
        DEBUG    => 'Error: Command not supported on this platform.',
        NONDEBUG => 'Error: Command not supported on this platform.',
    },
    SN2 => {
        DEBUG    => 'Error: Command not supported on this platform.',
        NONDEBUG => 'Error: Command not supported on this platform.',
    },
);

################################################################################
#  Name        : COMMAND_NOT_SUPPORTED_ON_PLATFORM
#  Description : Command not supported on platform error message
#  Privilege   : All
#  Driver      : All
################################################################################
Readonly our %SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN => (
    DEBUG => join(
        "\n",
        $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
        $NO_COMMANDS_AT_PRIV{SN0}{DEBUG},
    ),
    NONDEBUG => q{},
);

################################################################################
#  Name        : SN0_COMMAND_NOT_SUPPORTED
#  Description : Command not supported error message
#  Privilege   : All
#  Driver      : All
################################################################################
Readonly our $SN0_COMMAND_NOT_SUPPORTED => 'Command not supported.';

################################################################################
#  Name        : SN0_INSUF_PRIV_LEV
#  Description : Insufficient privilege level message.
#  Privilege   : All
#  Driver      : All
# Build       : SN0 only
################################################################################
Readonly our $SN0_INSUF_PRIV_LEV => 'ERROR: Insufficient privilege level.';

################################################################################
#  Name        : INSUF_PRIV_LEV_ADMIN
#  Description : Admin level message
#  Privilege   : admin
#  Driver      : All
################################################################################
Readonly our %INSUF_PRIV_LEV_ADMIN => {
    SN0 => {
        DEBUG => join(
            "\n",
            $SN0_INSUF_PRIV_LEV,
            $NO_COMMANDS_AT_PRIV{SN0}{DEBUG},
        ),
        NONDEBUG => q{},
    },
    SN1 => $NO_COMMANDS_AT_PRIV{SN1},
    SN2 => $NO_COMMANDS_AT_PRIV{SN2},
};

################################################################################
# Name        : INVALID_SYNTAX
# Description : Invalid syntax and Usage lines at top of usage statements
# Privilege   : >admin
# Driver      : all
################################################################################
Readonly our %INVALID_SYNTAX => (
    SN0 => 'Usage:',
    SN1 => join(
        "\n",
        'Error: Invalid syntax.',
        'Usage:',
    ),
    SN2 => join(
        "\n",
        'Error: Invalid syntax.',
        'Usage:',
    ),
);

################################################################################
#  Name        : CMD_NOT_SUPPORTED_ADMIN
#  Description : Command not supported message for admin level.
#  Privilege   : Admin
#  Driver      : All
################################################################################
Readonly our $SN0_CMD_NOT_SUPPORTED_ADMIN => join (
    "\n",
    $SN0_COMMAND_NOT_SUPPORTED,
    $NO_COMMANDS_AT_PRIV{SN0}{DEBUG},
);

################################################################################
#  Name        : SN0_INSUF_ARGS
#  Description : Insufficient arguments message
#  Privilege   : depends
#  Driver      : depends
################################################################################
Readonly our $SN0_INSUF_ARGS => 'ERROR: Insufficient arguments.';

################################################################################
#  Name        : SN0_INSUF_ARGS_SC
#  Description : Insufficient arguments for this sub command message
#  Privilege   : depends
#  Driver      : depends
################################################################################
Readonly our $SN0_INSUF_ARGS_SC => 'ERROR: Insufficient Arguments for this sub-command.';

################################################################################
#  Name        : SN0_INVALID_COMMAND
#  Description : Invalid command message for SN0
#  Privilege   : depends
#  Driver      : depends
################################################################################
Readonly our $SN0_INVALID_COMMAND => 'ERROR: Invalid command.';

################################################################################
#  Name        : SN0_INVALID_OPTION
#  Description : Invalid option message for SN0
#  Privilege   : depends
#  Driver      : depends
################################################################################
Readonly our $SN0_INVALID_OPTION => 'ERROR: Invalid option.';


################################################################################
# Name        : SN0_IC_USAGE
# Description : Usage statements from SN.0 (so we don't have to repeat this below)
# Privilege   : all
# Driver      : all
################################################################################

Readonly our %SN0_IC_USAGE => (
    CONNECTX => { #SN0_IC_USAGE
        admin    => $NO_COMMANDS_AT_PRIV{SN0},
        advanced => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
        },
        diag     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
        },
        test     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
        },
    },
    IWARP    => { #SN0_IC_USAGE
        admin    => $NO_COMMANDS_AT_PRIV{SN0},
        advanced => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
        },
        diag     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
        },
        test     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
        },
    },
    MVIA     => { #SN0_IC_USAGE
        admin    => $NO_COMMANDS_AT_PRIV{SN0},
        advanced => {
            DEBUG    => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic show port',
            ),
        },
        diag     => {
            DEBUG    => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic link <on|off> <0|1>',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
        },
        test     => {
            DEBUG    => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic config show',
                'ic config <interface_name> [ ip_addr <ip> ]',
                'ic status',
                'ic stats port [clear]',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic link [<on|off> <0|1> | toggle]',
                'ic dumpq <handle>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic ulp_fsm_msecs #value',
                'ic dumpNVTOC',
                'ic dump_qp',
            ),
        },
    },
    SINAI    => { #SN0_IC_USAGE
        admin    => $NO_COMMANDS_AT_PRIV{SN0},
        advanced => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic show port',
            ),
        },
        diag     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic dbg <chan_no> [on|off]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
        },
        test     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic [0|1]',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic reset rv',
                'ic reset loop [start|stop]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic dbg <chan_no> [on|off]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic sinai [init| close]',
                'ic controlarea <read|info|set|clear|help>',
                'help: print command usage',
                'read: print all control area fields',
                'info: print control area local and remote addresses',
                'set: set one or more fields in controlarea. Using 0 will ignore field',
                'set operation',
                'expected format: ic controlarea set <version> <status>',
                '<reason> <fm state>',
                'e.g. to only set version to 2, ic set controlarea 2 0 0 0',
                'clear: pass 0 to clear one or more fields in controlarea.',
                'Using 1111 will ignore field clear operation',
                'expected format: ic controlarea clear <version> <status>',
                '<reason> <fm state>',
                'e.g. to only clear version, ic clear controlarea 0 1111 1111 1111',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic log_vi_error [enable| disable]',
                'ic nonnverror [on|off]',
                'ic rvmonitor [on| off]',
                'ic dumpNVTOC',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic ood stats <on|off|clear>',
                'ic ood stats performance show [ -v ] [nv_vi_<chan#>|vi_if_<chan#>]',
                'ic ood status',
                'ic ood stats error show',
                'ic ood diag send',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic [0|1]',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic reset rv',
                'ic reset loop [start|stop]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic sinai [init| close]',
                'ic controlarea <read|info|set|clear|help>',
                'help: print command usage',
                'read: print all control area fields',
                'info: print control area local and remote addresses',
                'set: set one or more fields in controlarea. Using 0 will ignore field',
                'set operation',
                'expected format: ic controlarea set <version> <status>',
                '<reason> <fm state>',
                'e.g. to only set version to 2, ic set controlarea 2 0 0 0',
                'clear: pass 0 to clear one or more fields in controlarea.',
                'Using 1111 will ignore field clear operation',
                'expected format: ic controlarea clear <version> <status>',
                '<reason> <fm state>',
                'e.g. to only clear version, ic clear controlarea 0 1111 1111 1111',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic log_vi_error [enable| disable]',
                'ic nonnverror [on|off]',
                'ic rvmonitor [on| off]',
                'ic dumpNVTOC',
            ),
        },
    },
    TAVOR    => { #SN0_IC_USAGE
        admin    => $NO_COMMANDS_AT_PRIV{SN0},
        advanced => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic show port',
            ),
        },
        diag     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic dbg <chan_no> [on|off]',
                'ic show port',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v ]',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic dump hw_stats {-i}',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
            ),
        },
        test     => {
            DEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic [0|1]',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic reset rv',
                'ic reset loop [start|stop]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic dbg <chan_no> [on|off]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic log_vi_error [enable| disable]',
                'ic nonnverror [on|off]',
                'ic rvmonitor [on| off]',
                'ic dumpNVTOC',
            ),
            NONDEBUG => join(
                "\n",
                'Usage:',
                'ic status',
                'ic stats performance',
                'ic stats histogram',
                'ic stats error [ -v|-z ]',
                'ic dumpq <vinum>',
                'ic nvthruput [-dubmrc:g:t:] [-c count] <size>',
                'ic reset nic [0|1]',
                'ic reset link',
                'ic reset generatesyserror [0|1]',
                'ic reset rv',
                'ic reset loop [start|stop]',
                'ic dump [hw_stats {-i}|dbg_trace <file | print [#msgs]>]',
                'ic flint <query|verify> <flash|image>',
                'ic ifthruput <size> [chan]',
                'ic show port',
                'ic show rnic_rdma_stats',
                'ic channel_state',
                'ic disable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic enable [nic_reset|ofw|kstat|gather_mbuf|nic #nicnum]',
                'ic stall [nic {-f}]',
                'ic unstall nic',
                'ic kill [-h| -p]',
                'ic nvram',
                'ic nvram_ctest [partner <#count> <error {1|2}>|status|stop]',
                'ic download_fw',
                'ic ic_timeout #value',
                'ic log_vi_error [enable| disable]',
                'ic nonnverror [on|off]',
                'ic rvmonitor [on| off]',
                'ic dumpNVTOC',
            ),
        },
    },
);

# Now lets' start putting together usage information for each command, platform,
# and ONTAP version.


################################################################################
# Name        : IC_USG_<priv>_CONNECTX
# Description : Usage message for 'ic' commands
# Privilege   : all
# Driver      : ConnectX
################################################################################
Readonly our %IC_USAGE_ADV_CONNECTX => (
    SN0 => $SN0_IC_USAGE{CONNECTX}{advanced},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic show port',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic show port',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic show port',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic show port',
        ),
    },
);

Readonly our %IC_USAGE_DIAG_CONNECTX => (
    SN0 => $SN0_IC_USAGE{CONNECTX}{diag},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic dump hw_stats [ -i ]',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic dump hw_stats [ -i ]',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic dump hw_stats [ -i ]',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic dump hw_stats [ -i ]',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
        ),
    },
);

Readonly our %IC_USAGE_TEST_CONNECTX => (
    SN0 => $SN0_IC_USAGE{CONNECTX}{test},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset link { 0 | 1 }',
            'ic dump hw_stats [ -i ]',
            'ic dump dbg_trace { file | print <msgs> }',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset link { 0 | 1 }',
            'ic dump hw_stats [ -i ]',
            'ic dump dbg_trace { file | print <msgs> }',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset link { 0 | 1 }',
            'ic dump hw_stats [ -i ]',
            'ic dump dbg_trace { file | print <msgs> }',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats port [ clear ]',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset link { 0 | 1 }',
            'ic dump hw_stats [ -i ]',
            'ic dump dbg_trace { file | print <msgs> }',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic show port',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
);

################################################################################
#  Name        : IC_USG_<priv>_IWARP
#  Description : Usage message for 'ic' commands
#  Privilege   : all
#  Driver      : iWarp
################################################################################
Readonly our %IC_USAGE_ADV_IWARP => (
    SN0 => $SN0_IC_USAGE{IWARP}{advanced},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
        ),
    },
);

Readonly our %IC_USAGE_DIAG_IWARP => (
    SN0 => $SN0_IC_USAGE{IWARP}{diag},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic reset nic [ <nic> ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic reset nic [ <nic> ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic reset nic [ <nic> ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic reset nic [ <nic> ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
);

Readonly our %IC_USAGE_TEST_IWARP => (
    SN0 => $SN0_IC_USAGE{IWARP}{test},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ] [ -f ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ] [ -f ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ] [ -f ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } { 0 | 1 }',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ] [ -f ]',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
);

################################################################################
# Name        : IC_USG_<priv>_MVIA
# Description : Usage message for 'ic' commands
# Privilege   : all
# Driver      : MVIA
################################################################################
Readonly our %IC_USAGE_ADV_MVIA => (
    SN0 => $SN0_IC_USAGE{MVIA}{advanced},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
        ),
    },
);

Readonly our %IC_USAGE_DIAG_MVIA => (
    SN0 => $SN0_IC_USAGE{MVIA}{diag},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
);

Readonly our %IC_USAGE_TEST_MVIA => (
    SN0 => $SN0_IC_USAGE{MVIA}{test},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic stall nic',
            'ic unstall nic',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic stall nic',
            'ic unstall nic',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic stall nic',
            'ic unstall nic',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic config show',
            'ic config <interface_name> [ ip_addr <ip> ]',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic link { on | off } 0',
            'ic dumpq <handle>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic show rnic_rdma_stats',
            'ic channel_state',
            'ic enable { ofw | kstat }',
            'ic disable { ofw | kstat }',
            'ic stall nic',
            'ic unstall nic',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic ic_timeout [ <value> ]',
            'ic ulp_fsm_msecs <value>',
            'ic dumpNVTOC',
            'ic dump_qp',
        ),
    },
);

################################################################################
# Name        : IC_USG_<priv>_SINAI
# Description : Usage message for 'ic' commands
# Privilege   : all
# Driver      : SINAI
################################################################################
Readonly our %IC_USAGE_ADV_SINAI => (
    SN0 => $SN0_IC_USAGE{SINAI}{advanced},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
        ),
    },
);

Readonly our %IC_USAGE_DIAG_SINAI => (
    SN0 => $SN0_IC_USAGE{SINAI}{diag},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
);

Readonly our %IC_USAGE_TEST_SINAI => (
    SN0 => $SN0_IC_USAGE{SINAI}{test},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic sinai { init | close }',
            'ic controlarea { read | info | help }',
            'ic controlarea set <version> <status> <reason> <fm state>',
            'ic controlarea clear <version> <status> <reason> <fm state>',
            'help: Print command usage',
            'read: Print all control area fields',
            'info: Print control area local and remote addresses',
            'set: Set one or more fields in controlarea.',
            'Using 0 will ignore field set operation.',
            'For example, to set only version to 2:',
            '*> ic set controlarea 2 0 0 0',
            'clear: Pass 0 to clear one or more fields in controlarea.',
            'Using 1111 will ignore field clear operation.',
            'For example, to clear only version:',
            '*> ic clear controlarea 0 1111 1111 1111',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic sinai { init | close }',
            'ic controlarea { read | info | help }',
            'ic controlarea set <version> <status> <reason> <fm state>',
            'ic controlarea clear <version> <status> <reason> <fm state>',
            'help: Print command usage',
            'read: Print all control area fields',
            'info: Print control area local and remote addresses',
            'set: Set one or more fields in controlarea.',
            'Using 0 will ignore field set operation.',
            'For example, to set only version to 2:',
            '*> ic set controlarea 2 0 0 0',
            'clear: Pass 0 to clear one or more fields in controlarea.',
            'Using 1111 will ignore field clear operation.',
            'For example, to clear only version:',
            '*> ic clear controlarea 0 1111 1111 1111',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic sinai { init | close }',
            'ic controlarea { read | info | help }',
            'ic controlarea set <version> <status> <reason> <fm state>',
            'ic controlarea clear <version> <status> <reason> <fm state>',
            'help: Print command usage',
            'read: Print all control area fields',
            'info: Print control area local and remote addresses',
            'set: Set one or more fields in controlarea.',
            'Using 0 will ignore field set operation.',
            'For example, to set only version to 2:',
            '*> ic set controlarea 2 0 0 0',
            'clear: Pass 0 to clear one or more fields in controlarea.',
            'Using 1111 will ignore field clear operation.',
            'For example, to clear only version:',
            '*> ic clear controlarea 0 1111 1111 1111',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic ood stats { on | off | clear }',
            'ic ood stats performance show [ -v ] [ <chan_name> ]',
            'ic ood status',
            'ic ood stats error show',
            'ic ood diag send',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic flint { query | verify } { flash | image }',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic sinai { init | close }',
            'ic controlarea { read | info | help }',
            'ic controlarea set <version> <status> <reason> <fm state>',
            'ic controlarea clear <version> <status> <reason> <fm state>',
            'help: Print command usage',
            'read: Print all control area fields',
            'info: Print control area local and remote addresses',
            'set: Set one or more fields in controlarea.',
            'Using 0 will ignore field set operation.',
            'For example, to set only version to 2:',
            '*> ic set controlarea 2 0 0 0',
            'clear: Pass 0 to clear one or more fields in controlarea.',
            'Using 1111 will ignore field clear operation.',
            'For example, to clear only version:',
            '*> ic clear controlarea 0 1111 1111 1111',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
);


################################################################################
# Name        : IC_USG_<priv>_TAVOR
# Description : Usage message for 'ic' commands
# Privilege   : all
# Driver      : TAVOR
################################################################################
Readonly our %IC_USAGE_ADV_TAVOR => (
    SN0 => $SN0_IC_USAGE{TAVOR}{advanced},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
        ),
    },

);

Readonly our %IC_USAGE_DIAG_TAVOR => (
    SN0 => $SN0_IC_USAGE{TAVOR}{diag},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v ]',
            'ic dumpq <vinum>',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
        ),
    },
);

Readonly our %IC_USAGE_TEST_TAVOR => (
    SN0 => $SN0_IC_USAGE{TAVOR}{test},
    SN1 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset link [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
            'ic nv5dma_hang { fix_on | fix_off | sim_hang | test_dma | status }',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset link [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
    SN2 => {
        DEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset link [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
            'ic nv5dma_hang {fix_on|fix_off|hang_req|sim_hang|test_dma|status}',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
        NONDEBUG => join(
            "\n",
            'Usage:',
            'ic status',
            'ic stats performance',
            'ic stats histogram',
            'ic stats error [ -v | -z ]',
            'ic dumpq <vinum>',
            'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
            'ic nvthruput_mt [-oubhmrg:s:] <size>',
            '-o: Insert an Op barrier between transfers',
            '-u: No data placement semantics',
            '-b: Bidirectional test',
            '-h: Add header and footer to simulate WAFL traffic',
            '-m: Mix of transfer sizes',
            '-r: Source transfers from memory',
            '-g <count>: Construct SGL with <count> elements',
            '-s <source threads>: Number of parallel write threads',
            'ic reset nic [ <nic> ]',
            'ic reset link [ <nic> ]',
            'ic reset generatesyserror { 0 | 1 }',
            'ic reset rv',
            'ic reset loop { start | stop }',
            'ic dump hw_stats',
            'ic ifthruput <size> [ <chan> ]',
            'ic dbg <chan_name> { on | off }',
            'ic channel_state',
            'ic enable { nic_reset | nic <nic> }',
            'ic disable { nic_reset | nic <nic> }',
            'ic kill { -h | -p } [<message>]',
            'ic nvram',
            'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
            'ic download_fw',
            'ic ic_timeout [ <value> ]',
            'ic log_vi_error { enable | disable }',
            'ic nonnverror { on | off }',
            'ic rvmonitor { on | off }',
            'ic dumpNVTOC',
#            'ic nvpm dump',
#            'ic nvmm dump',
#            'ic nvmm { enable | online | disable | offline | refresh } <val>',
        ),
    },
);

################################################################################
# Name:       : IC_USAGE_TABLE
# Description : Put the expected output for 'ic' into a table.
################################################################################
Readonly our %IC_USAGE_TABLE => (
    CONNECTX => {
        admin    => \%NO_COMMANDS_AT_PRIV,
        advanced => \%IC_USAGE_ADV_CONNECTX,
        diag     => \%IC_USAGE_DIAG_CONNECTX,
        test     => \%IC_USAGE_TEST_CONNECTX,
    },
    IWARP    => {
        admin    => \%NO_COMMANDS_AT_PRIV,
        advanced => \%IC_USAGE_ADV_IWARP,
        diag     => \%IC_USAGE_DIAG_IWARP,
        test     => \%IC_USAGE_TEST_IWARP,
    },
    MVIA     => {
        admin    => \%NO_COMMANDS_AT_PRIV,
        advanced => \%IC_USAGE_ADV_MVIA,
        diag     => \%IC_USAGE_DIAG_MVIA,
        test     => \%IC_USAGE_TEST_MVIA,
    },
    SINAI    => {
        admin    => \%NO_COMMANDS_AT_PRIV,
        advanced => \%IC_USAGE_ADV_SINAI,
        diag     => \%IC_USAGE_DIAG_SINAI,
        test     => \%IC_USAGE_TEST_SINAI,
    },
    TAVOR    => {
        admin    => \%NO_COMMANDS_AT_PRIV,
        advanced => \%IC_USAGE_ADV_TAVOR,
        diag     => \%IC_USAGE_DIAG_TAVOR,
        test     => \%IC_USAGE_TEST_TAVOR,
    },
);

################################################################################
# Name        : IC_CHANNEL_STATE_USAGE_TABLE
# Description : Usage message for 'ic channel_state' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_CHANNEL_STATE_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        #test is functional - no usage
    }, #End 'ic channel_state' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic channel_state' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic channel_state' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic channel_state' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic channel_state' TAVOR
);


################################################################################
# Name        : IC_CONFIG_USAGE TABLE
# Description : Usage messages for 'ic config' command
# Privilege   : all
# Driver      : all
################################################################################

Readonly our %IC_CONFIG_USAGE_TABLE => (
    CONNECTX => {
        admin    => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        diag     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        test     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
        },
    }, #End 'ic config' CONNECTX
    IWARP    => {
        admin    => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        diag     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        test     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
        },
    }, #End 'ic config' IWARP
    MVIA     => {
        admin    => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        diag     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                ),
            },
        },
        test     => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN0},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic config show',
                    'ic config <interface_name> [ ip_addr <ip> ]',
                ),
            },
        },
    }, #End 'ic config' MVIA
    SINAI    => {
        admin    => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag     => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test     => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic config' SINAI
    TAVOR    => {
        admin    => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag     => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test     => {
            SN0 => {
                DEBUG    => 'UsageError:"ic config" is supported only on iWARP/Infiniband interconnects',
                NONDEBUG => q{},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic config' TAVOR
);

################################################################################
# Name        : IC_CONTROLAREA_USAGE_TABLE
# Description : Usage message for 'ic controlarea' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_CONTROLAREA_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic controlarea' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic controlarea' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic controlarea' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic controlarea { read | info | help }',
                    'ic controlarea set <version> <status> <reason> <fm state>',
                    'ic controlarea clear <version> <status> <reason> <fm state>',
                    'help: Print command usage',
                    'read: Print all control area fields',
                    'info: Print control area local and remote addresses',
                    'set: Set one or more fields in controlarea.',
                    'Using 0 will ignore field set operation.',
                    'For example, to set only version to 2:',
                    '*> ic set controlarea 2 0 0 0',
                    'clear: Pass 0 to clear one or more fields in controlarea.',
                    'Using 1111 will ignore field clear operation.',
                    'For example, to clear only version:',
                    '*> ic clear controlarea 0 1111 1111 1111',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic controlarea { read | info | help }',
                    'ic controlarea set <version> <status> <reason> <fm state>',
                    'ic controlarea clear <version> <status> <reason> <fm state>',
                    'help: Print command usage',
                    'read: Print all control area fields',
                    'info: Print control area local and remote addresses',
                    'set: Set one or more fields in controlarea.',
                    'Using 0 will ignore field set operation.',
                    'For example, to set only version to 2:',
                    '*> ic set controlarea 2 0 0 0',
                    'clear: Pass 0 to clear one or more fields in controlarea.',
                    'Using 1111 will ignore field clear operation.',
                    'For example, to clear only version:',
                    '*> ic clear controlarea 0 1111 1111 1111',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic controlarea { read | info | help }',
                    'ic controlarea set <version> <status> <reason> <fm state>',
                    'ic controlarea clear <version> <status> <reason> <fm state>',
                    'help: Print command usage',
                    'read: Print all control area fields',
                    'info: Print control area local and remote addresses',
                    'set: Set one or more fields in controlarea.',
                    'Using 0 will ignore field set operation.',
                    'For example, to set only version to 2:',
                    '*> ic set controlarea 2 0 0 0',
                    'clear: Pass 0 to clear one or more fields in controlarea.',
                    'Using 1111 will ignore field clear operation.',
                    'For example, to clear only version:',
                    '*> ic clear controlarea 0 1111 1111 1111',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic controlarea { read | info | help }',
                    'ic controlarea set <version> <status> <reason> <fm state>',
                    'ic controlarea clear <version> <status> <reason> <fm state>',
                    'help: Print command usage',
                    'read: Print all control area fields',
                    'info: Print control area local and remote addresses',
                    'set: Set one or more fields in controlarea.',
                    'Using 0 will ignore field set operation.',
                    'For example, to set only version to 2:',
                    '*> ic set controlarea 2 0 0 0',
                    'clear: Pass 0 to clear one or more fields in controlarea.',
                    'Using 1111 will ignore field clear operation.',
                    'For example, to clear only version:',
                    '*> ic clear controlarea 0 1111 1111 1111',
                ),
            },
        },
    }, #End 'ic controlarea' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic controlarea' TAVOR
);


################################################################################
# Name        : IC_DBG_USAGE_TABLE
# Description : Usage message for 'ic dbg' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DBG_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic dbg' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic dbg' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic dbg' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dbg <chan_name> { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dbg <chan_name> { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dbg <chan_name> { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dbg <chan_name> { on | off }',
                ),
            },
        },
    }, #End 'ic dbg' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dbg <chan_name> { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dbg <chan_name> { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dbg <chan_name> { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dbg <chan_name> { on | off }',
                ),
            },
        },
    }, #End 'ic dbg' TAVOR
);

################################################################################
# Name        : IC_DISABLE_USAGE_TABLE
# Description : Usage message for 'ic disable' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DISABLE_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic disable' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic disable' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic disable' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { nic_reset | nic <nic> }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { nic_reset | nic <nic> }',
                ),
            },
        },
    }, #End 'ic disable' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic disable { nic_reset | nic <nic> }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic disable { nic_reset | nic <nic> }',
                ),
            },
        },
    }, #End 'ic disable' TAVOR
);


################################################################################
# Name        : IC_DUMP_USAGE_TABLE
# Description : Usage message for 'ic dump' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DUMP_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats [ -i ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats [ -i ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats [ -i ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats [ -i ]',
                ),
            },

        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats [ -i ]',
                    'ic dump dbg_trace { file | print <msgs> }'
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats [ -i ]',
                    'ic dump dbg_trace { file | print <msgs> }'
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats [ -i ]',
                    'ic dump dbg_trace { file | print <msgs> }'
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats [ -i ]',
                    'ic dump dbg_trace { file | print <msgs> }'
                ),
            },
        },
    }, #End 'ic dump' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
    }, #End 'ic dump' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
    }, #End 'ic dump' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
    }, #End 'ic dump' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dump hw_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dump hw_stats',
                ),
            },
        },
    }, #End 'ic dump' TAVOR
);


################################################################################
# Name        : IC_DUMP_QP_USAGE_TABLE
# Description : Usage message for 'ic dump_qp' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DUMP_QP_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dump_qp' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dump_qp' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dump_qp' MVIA
    SINAI    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dump_qp' SINAI
    TAVOR    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dump_qp' TAVOR
);

################################################################################
# Name        : IC_DUMPNVTOC_USAGE_TABLE
# Description : Usage message for 'ic dumpNVTOC' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DUMPNVTOC_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dumpNVTOC' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dumpNVTOC' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dumpNVTOC' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dumpNVTOC' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic dumpNVTOC' TAVOR
);

################################################################################
# Name        : IC_DUMPQ_USAGE_TABLE
# Description : Usage message for 'ic dumpq' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_DUMPQ_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
    }, #End 'ic dumpq' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
    }, #End 'ic dumpq' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <handle>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <handle>',
                ),
            },
        },
    }, #End 'ic dumpq' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
            },
        },
    }, #End 'ic dumpq' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic dumpq <vinum>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic dumpq <vinum>',
                ),
            },
        },
    }, #End 'ic dumpq' TAVOR
);

################################################################################
# Name        : IC_ENABLE_USAGE_TABLE
# Description : Usage message for 'ic enable' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_ENABLE_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic enable' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic enable' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { ofw | kstat }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { ofw | kstat }',
                ),
            },
        },
    }, #End 'ic enable' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { nic_reset | nic <nic> }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { nic_reset | nic <nic> }',
                ),
            },
        },
    }, #End 'ic enable' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic enable { nic_reset | nic <nic> }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { nic_reset | nic <nic> }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic enable { nic_reset | nic <nic> }',
                ),
            },
        },
    }, #End 'ic enable' TAVOR
);

################################################################################
# Name        : IC_FLINT_USAGE_TABLE
# Description : Usage message for 'ic flint' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_FLINT_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
        },
    }, # End 'ic flint' CONNECTX
    IWARP => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, # End 'ic flint' IWARP
    MVIA => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, # End 'ic flint' MVIA
    SINAI => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic flint { query | verify } { flash | image }',
                ),
            },
        },
    }, # End 'ic flint' SINAI
    TAVOR => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, # End 'ic flint' TAVOR
);

################################################################################
# Name        : IC_IC_TIMEOUT_USAGE_TABLE
# Description : Usage message for 'ic ic_timeout' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_IC_TIMEOUT_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # diag is functional - no usage
        # test is functioanl - no usage
    }, #End 'ic ic_timeout' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic ic_timeout' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic ic_timeout' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic ic_timeout' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic ic_timeout' TAVOR
);

################################################################################
# Name        : IC_IFTHRUPUT_USAGE_TABLE
# Description : Usage message for 'ic ifthruput' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_IFTHRUPUT_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
    }, #End 'ic ifthruput' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
    }, #End 'ic ifthruput' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
    }, #End 'ic ifthruput' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
    }, #End 'ic ifthruput' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ifthruput <size> [ <chan> ]',
                ),
            },
        },
    }, #End 'ic ifthruput' TAVOR
);

################################################################################
# Name        : IC_KILL_USAGE_TABLE
# Description : Usage message for 'ic kill' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_KILL_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic kill' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic kill' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic kill' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic kill' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic kill' TAVOR
);

################################################################################
# Name        : IC_LINK_USAGE_TABLE
# Description : Usage message for 'ic link' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_LINK_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
    }, #End 'ic link' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } { 0 | 1 }',
                ),
            },
        },
    }, #End 'ic link' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic link { on | off } 0',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic link { on | off } 0',
                ),
            },
        },
    }, #End 'ic link' MVIA
    SINAI    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic link' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 =>{
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic link' TAVOR
);

################################################################################
# Name        : IC_LOG_VI_ERROR_USAGE_TABLE
# Description : Usage message for 'ic log_vi_error' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_LOG_VI_ERROR_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic log_vi_error_message' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic log_vi_error_message' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic log_vi_error_message' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic log_vi_error { enable | disable }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic log_vi_error { enable | disable }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic log_vi_error { enable | disable }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic log_vi_error { enable | disable }',
                ),
            },
        },
    }, #End 'ic log_vi_error_message' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic log_vi_error { enable | disable }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic log_vi_error { enable | disable }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic log_vi_error { enable | disable }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic log_vi_error { enable | disable }',
                ),
            },
        },
    }, #End 'ic log_vi_error_message' TAVOR
);

################################################################################
# Name        : IC_NONNVERROR_USAGE_TABLE
# Description : Usage message for 'ic nonnverror' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NONNVERROR_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic nonnverror' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic nonnverror' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic nonnverror' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nonnverror { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nonnverror { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nonnverror { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nonnverror { on | off }',
                ),
            },
        },
    }, #End 'ic nonnverror' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nonnverror { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nonnverror { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nonnverror { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nonnverror { on | off }',
                ),
            },
        },
    }, #End 'ic nonnverror' TAVOR
);

################################################################################
# Name        : IC_NVCT_USAGE_TABLE
# Description : Usage message for 'ic nvct' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVCT_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
    }, #End 'ic nvct' CONNECTX
    IWARP    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
    }, #End 'ic nvct' IWARP
    MVIA     => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
    }, #End 'ic nvct' MVIA
    SINAI    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
    }, #End 'ic nvct' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
    }, #End 'ic nvct' TAVOR
);

################################################################################
# Name        : IC_NVMM_USAGE_TABLE
# Description : Usage message for 'ic nvmm' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVMM_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#            },
        },
    }, #End 'ic nvmm' CONNECTX
    IWARP    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#            },
        },
    }, #End 'ic nvmm' IWARP
    MVIA     => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2}, #{
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#            },
        },
    }, #End 'ic nvmm' MVIA
    SINAI    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#            },
        },
    }, #End 'ic nvmm' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvmm dump',
#                    'ic nvmm { enable | online | disable | offline | refresh } <val>',
#                ),
#            },
        },
    }, #End 'ic nvmm' TAVOR
);

################################################################################
# Name        : IC_NVPM_USAGE_TABLE
# Description : Usage message for 'ic nvpm' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVPM_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#            },
        },
    }, #End 'ic nvpm' CONNECTX
    IWARP    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#            },
        },
    }, #End 'ic nvpm' IWARP
    MVIA     => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2}, #{
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#            },
        },
    }, #End 'ic nvpm' MVIA
    SINAI    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#            },
        },
    }, #End 'ic nvpm' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INVALID_COMMAND,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
#            SN1 => {
#                DEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#                NONDEBUG => join(
#                    "\n",
#                    $INVALID_SYNTAX{SN1},
#                    'ic nvpm dump',
#                ),
#            },
        },
    }, #End 'ic nvpm' TAVOR
);


################################################################################
# Name        : IC_NVRAM_USAGE_TABLE
# Description : Usage message for 'ic nvram' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVRAM_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic nvram' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic nvram' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic nvram' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic nvram' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # test is functional - no usage
    }, #End 'ic nvram' TAVOR
);

################################################################################
# Name        : IC_NVRAM_CTEST_USAGE_TABLE
# Description : Usage message for 'ic nvram_ctest' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVRAM_CTEST_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
        },
    }, #End 'ic nvram_ctest' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
        },
    }, #End 'ic nvram_ctest' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
        },
    }, #End 'ic nvram_ctest' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
        },
    }, #End 'ic nvram_ctest' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic nvram_ctest { partner <count> [ error {1|2} ] | status | stop }',
                ),
            },
        },
    }, #End 'ic nvram_ctest' TAVOR
);

################################################################################
# Name        : IC_NVTHRUPUT_USAGE_TABLE
# Description : Usage message for 'ic nvthruput' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_NVTHRUPUT_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
            SN2 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
        },
    }, #End 'ic nvthruput' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
            SN2 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
        },
    }, #End 'ic nvthruput' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
            SN2 => {
                DEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
        },
    }, #End 'ic nvthruput' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG    => 'This command is not supported on the current platform.',
                NONDEBUG => 'This command is not supported on the current platform.',
            },
            SN1 => {
                DEBUG    => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
            SN2 => {
                DEBUG    => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
        },
        test => {
            SN0 => {
                DEBUG    => 'This command is not supported on the current platform.',
                NONDEBUG => 'This command is not supported on the current platform.',
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
        },
    }, #End 'ic nvthruput' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG    => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
            SN2 => {
                DEBUG    => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
                NONDEBUG => 'Error: This command is no longer supported.  Type \'?\' for a list of commands.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_INVALID_OPTION,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
                ),
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN1},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
                NONDEBUG => join(
                    "\n",
                    'Missing required \'size\' parameter',
                    $INVALID_SYNTAX{SN2},
                    'ic nvthruput [-odubhmrc:g:t:] [-c count] <size>',
                ),
            },
        },
    }, #End 'ic nvthruput' TAVOR
);

################################################################################
# Name        : IC_OOD_USAGE_TABLE
# Description : Usage message for 'ic ood' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_OOD_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ood' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ood' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ood' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ood stats { on | off | clear }',
                    'ic ood stats performance show [ -v ] [ <chan_name> ]',
                    'ic ood status',
                    'ic ood stats error show',
                    'ic ood diag send',
                ),
            },
        },
    }, #End 'ic ood' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ood' TAVOR
);

################################################################################
# Name        : IC_RESET_USAGE_TABLE
# Description : Usage message for 'ic reset' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_RESET_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset link { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset link { 0 | 1 }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset link { 0 | 1 }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset link { 0 | 1 }',
                ),
            },
        },
    }, #End 'ic reset' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ] [ -f ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ] [ -f ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ] [ -f ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ] [ -f ]',
                ),
            },
        },
    }, #End 'ic reset' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG    => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG    => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic reset' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
            },
        },
    }, #End 'ic reset' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                    'ic reset link [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic reset nic [ <nic> ]',
                    'ic reset link [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                    'ic reset link [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic reset nic [ <nic> ]',
                    'ic reset link [ <nic> ]',
                    'ic reset generatesyserror { 0 | 1 }',
                    'ic reset rv',
                    'ic reset loop { start | stop }',
                ),
            },
        },
    }, #End 'ic reset' TAVOR
);

################################################################################
# Name        : IC_RVMONITOR_USAGE_TABLE
# Description : Usage message for 'ic rvmonitor' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_RVMONITOR_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic rvmonitor' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic rvmonitor' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic rvmonitor' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic rvmonitor { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic rvmonitor { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic rvmonitor { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic rvmonitor { on | off }',
                ),
            },
        },
    }, #End 'ic rvmonitor' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic rvmonitor { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic rvmonitor { on | off }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic rvmonitor { on | off }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic rvmonitor { on | off }',
                ),
            },
        },
    }, #End 'ic rvmonitor' TAVOR
);


################################################################################
# Name        : IC_SINAI_USAGE_TABLE
# Description : Usage message for 'ic sinai' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_SINAI_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic sinai' CONNECTX
    IWARP    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic sinai' IWARP
    MVIA     => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic sinai' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic sinai { init | close }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic sinai { init | close }',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic sinai { init | close }',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic sinai { init | close }',
                ),
            },
        },
    }, #End 'ic sinai' SINAI
    TAVOR    => {
        admin => {
            SN0 => \%SN0_COMMAND_NOT_SUPPORTED_ON_PLATFORM_ADMIN,
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG =>$SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic sinai' TAVOR
);

################################################################################
# Name        : IC_SHOW_USAGE_TABLE
# Description : Usage message for 'ic show' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_SHOW_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show port',
                    'ic show rnic_rdma_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show port',
                    'ic show rnic_rdma_stats',
                ),
            },
        },
    }, #End 'ic show' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show rnic_rdma_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show rnic_rdma_stats',
                ),
            },
        },
    }, #End 'ic show' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic show rnic_rdma_stats',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show rnic_rdma_stats',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic show rnic_rdma_stats',
                ),
            },
        },
    }, #End 'ic show' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic show' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic show' TAVOR
);

################################################################################
# Name        : IC_STALL_USAGE_TABLE
# Description : Usage message for 'ic stall' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_STALL_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic stall' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic stall' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stall nic',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stall nic',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stall nic',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stall nic',
                ),
            },
        },
    }, #End 'ic stall' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic stall' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic stall' TAVOR
);


################################################################################
# Name        : IC_STATS_USAGE_TABLE
# Description : Usage message for 'ic stats' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_STATS_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats port [ clear ]',
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
        },
    }, #End 'ic stats' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
        },
    }, #End 'ic stats' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
        },
    }, #End 'ic stats' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
        },
    }, #End 'ic stats' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v ]',
                ),
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic stats performance',
                    'ic stats histogram',
                    'ic stats error [ -v | -z ]',
                ),
            },
        },
    }, #End 'ic stats' TAVOR
);

################################################################################
# Name        : IC_STATUS_USAGE_TABLE
# Description : Usage message for 'ic status' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_STATUS_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # advanced is functional - no usage
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic status' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # advanced is functional - no usage
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic status' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # advanced is functional - no usage
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic status' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # advanced is functional - no usage
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic status' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        # advanced is functional - no usage
        # diag is functional - no usage
        # test is functional - no usage
    }, #End 'ic status' TAVOR
);


################################################################################
# Name        : IC_ULP_FSM_MSECS_USAGE_TABLE
# Description : Usage message for 'ic ulp_fsm_msecs' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_ULP_FSM_MSECS_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
        },
    }, #End 'ic ulp_fsm_msecs' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
        },
    }, #End 'ic ulp_fsm_msecs' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
            SN2 => {
                DEBUG    => 'Error: Invalid syntax.',
                NONDEBUG => 'Error: Invalid syntax.',
            },
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic ulp_fsm_msecs <value>',
                ),
            },
        },
    }, #End 'ic ulp_fsm_msecs' MVIA
    SINAI    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ulp_fsm_msecs' SINAI
    TAVOR    => {
        admin => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{admin}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{admin}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN0}{DEBUG},
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic ulp_fsm_msecs' TAVOR
);

################################################################################
# Name        : IC_UNSTALL_USAGE_TABLE
# Description : Usage message for 'ic unstall' command
# Privilege   : all
# Driver      : all
################################################################################
Readonly our %IC_UNSTALL_USAGE_TABLE => (
    CONNECTX => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{CONNECTX}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{CONNECTX}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{CONNECTX}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic unstall' CONNECTX
    IWARP    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{IWARP}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{IWARP}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{IWARP}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic unstall' IWARP
    MVIA     => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{MVIA}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{MVIA}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{MVIA}{test}{NONDEBUG},
            },
            SN1 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic unstall nic',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN1},
                    'ic unstall nic',
                ),
            },
            SN2 => {
                DEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic unstall nic',
                ),
                NONDEBUG => join(
                    "\n",
                    $INVALID_SYNTAX{SN2},
                    'ic unstall nic',
                ),
            },
        },
    }, #End 'ic unstall' MVIA
    SINAI    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{SINAI}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{SINAI}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{SINAI}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic unstall' SINAI
    TAVOR    => {
        admin => {
            SN0 => $INSUF_PRIV_LEV_ADMIN{SN0},
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        advanced => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{advanced}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{advanced}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        diag => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_PRIV_LEV,
                    $SN0_IC_USAGE{TAVOR}{diag}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{diag}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_FOUND{SN1},
            SN2 => $COMMAND_NOT_FOUND{SN2},
        },
        test => {
            SN0 => {
                DEBUG => join(
                    "\n",
                    $SN0_INSUF_ARGS,
                    $SN0_IC_USAGE{TAVOR}{test}{DEBUG},
                ),
                NONDEBUG => $SN0_IC_USAGE{TAVOR}{test}{NONDEBUG},
            },
            SN1 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN1},
            SN2 => $COMMAND_NOT_SUPPORTED_ON_PLATFORM{SN2},
        },
    }, #End 'ic unstall' TAVOR
);


################################################################################
# Name:       : USAGE_TABLE
# Description : Now, put all of the above work together into one table.
################################################################################
Readonly our %USAGE_TABLE => (
    'ic'               => \%IC_USAGE_TABLE,
    'ic channel_state' => \%IC_CHANNEL_STATE_USAGE_TABLE,
    'ic config'        => \%IC_CONFIG_USAGE_TABLE,
    'ic controlarea'   => \%IC_CONTROLAREA_USAGE_TABLE,
    'ic dbg'           => \%IC_DBG_USAGE_TABLE,
    'ic disable'       => \%IC_DISABLE_USAGE_TABLE,
    'ic dump'          => \%IC_DUMP_USAGE_TABLE,
    'ic dump_qp'       => \%IC_DUMP_QP_USAGE_TABLE,
    'ic dumpNVTOC'     => \%IC_DUMPNVTOC_USAGE_TABLE,
    'ic dumpq'         => \%IC_DUMPQ_USAGE_TABLE,
    'ic enable'        => \%IC_ENABLE_USAGE_TABLE,
    'ic flint'         => \%IC_FLINT_USAGE_TABLE,
    'ic ic_timeout'    => \%IC_IC_TIMEOUT_USAGE_TABLE,
    'ic ifthruput'     => \%IC_IFTHRUPUT_USAGE_TABLE,
    'ic kill'          => \%IC_KILL_USAGE_TABLE,
    'ic link'          => \%IC_LINK_USAGE_TABLE,
    'ic log_vi_error'  => \%IC_LOG_VI_ERROR_USAGE_TABLE,
    'ic nonnverror'    => \%IC_NONNVERROR_USAGE_TABLE,
    'ic nvct'          => \%IC_NVCT_USAGE_TABLE,
    'ic nvmm'          => \%IC_NVMM_USAGE_TABLE,
    'ic nvpm'          => \%IC_NVPM_USAGE_TABLE,
    'ic nvram'         => \%IC_NVRAM_USAGE_TABLE,
    'ic nvram_ctest'   => \%IC_NVRAM_CTEST_USAGE_TABLE,
    'ic nvthruput'     => \%IC_NVTHRUPUT_USAGE_TABLE,
    'ic ood'           => \%IC_OOD_USAGE_TABLE,
    'ic reset'         => \%IC_RESET_USAGE_TABLE,
    'ic rvmonitor'     => \%IC_RVMONITOR_USAGE_TABLE,
    'ic sinai'         => \%IC_SINAI_USAGE_TABLE,
    'ic show'          => \%IC_SHOW_USAGE_TABLE,
    'ic stall'         => \%IC_STALL_USAGE_TABLE,
    'ic stats'         => \%IC_STATS_USAGE_TABLE,
    'ic status'        => \%IC_STATUS_USAGE_TABLE,
    'ic ulp_fsm_msecs' => \%IC_ULP_FSM_MSECS_USAGE_TABLE,
    'ic unstall'       => \%IC_UNSTALL_USAGE_TABLE,
);


1;
## End of module
__END__

=head1 NAME

Iconnect::Command_Testing_Usage - common variables and subroutines for Command Testing scripts

=head1 SYNOPSIS

    use Command_Testing qw(:ALL);                          # this is all you need

=head1 DESCRIPTION

This module provides the following variables:
    B<%USAGE_TABLE>       - Table of expected usage statements

The module also exports the following subroutines:

=item B<compare_usage_statements()>
Input:
    command     = (required) command to execute on controller
    driver      = (required) driver type (CONNECTX, IWARP, etc.)
    ontap_ver   = (required) ONTAP version (FS0, SN1, etc)
    kernel_type = (required) DEBUG or NONDEBUG
    controller  = (required) APISet object for controller
    privs       = (optional) arrayref - list of priv levels to execute against
Returns:
    $TCD::<status>

=item B<get_driver_type()>
Input:
    controller  = APISet object for controller
Returns:
    scalar - string indicating IC Driver
        CONNECTX
        IWARP
        MVIA
        SINAI
        TAVOR

=item B<get_ontap_name()>
Input:
    controller = APISet object for node
Returns:
    scalar - string with shortened codename of ONTAP version (e.g., SN0)


=item B<validate_cdef_parsed_values()>
Input:
    key        = (required) name of parsed_output key being verified
    expected   = (required) hashref - table of Expected return key/values
    returned   = (required) hashref - parsed_output table from CDEF
Returns:
    $TCD::<status>

1;
