# $Id: //depot/prod/test/Rsierranevada/lib/Quota/QCov_Ops.pm#1 $

# this is basically -*- perl -*-

# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.

## @summary	Quota Coverage library
##
## @description
##	This library provides routines to support scripts which test basic
##	quota functionality for various combinations of /etc/quota file entries,
##	constructed from the following:
##
##		default user
##		default group
##		default tree
##		explicit user
##		explicit group
##		explicit tree
##
##	The supplied functions allow construction, modification, and deletion of
##	users, groups, qtrees, files, directories, usermap file entries, and quota
##	file entries. There are also functions to support checking of quota report
##	results from the filer.
##
##	See the global @EXPORT list below for a list of exported functions.
##
## @usage
##	use QCov_Ops;
##
## @author dl-vargas
##
## @status private
##
## @pod here
##

package Quota::QCov_Ops;

#################################### 
#
#	Imports
#
#################################### 

use Tharn;
use strict;
use base qw(Exporter);
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF UNDEF);
use Volume qw(vol_options);
use filer qw(filer_cifs_stop);
#use SPARSE::Util::Filer_Utils qw(setup_cifs);
use CIFS::Setup qw(cifs_setup);

use SPARSE::Util::Logger;
use FS::Util::Utilities qw(uSystem deltree);
### use FS::Util::Utilities qw(:utils);
use FS::Util::Setup qw(create_vol_and_mount mount_root_vol
		       umount_vol common_cleanup);

#use setup qw(is_cifs_setup);
use CIFS::Setup qw(is_cifs_setup);
use Volume qw(vol_exists vol_options vol_offline vol_list vol_root);

use Quota qw(:quota);
use FS::Util::FileIO qw(fioOpen);
use Utils qw(make_conn);
use Users qw(createUsers createGroups delPasswdList listCIFSUsers removeCIFSUsersExcept
		removeCIFSUser addCIFSUser writePasswdFile writeGroupFile delPasswdEntry
		readPasswdFile readGroupFile listCIFSGroups addCIFSGroup removeCIFSGroup
		writeDefaultPasswdFile );

use SnapCmds qw(snapshot_create snapshot_exists snapshot_delete snaprestore_volume);
use Admin::License qw( license is_licensed unlicense );
use QtreeCmds qw(qtreeCreate qtreeSecurity qtreeDelete qtreeStatusVol qtreePresent);
use fileSystem qw(backupFiles restoreFiles);
use filer qw(filer_cifs_access);
use MapShare qw(map_cifs_share unmap_cifs_share map_exists);
use RegistryCmds qw(getOption setOption);
use TestEnv::MinWin qw(execute_cmd minwin_compatible);
use ExportFS qw(share_add share_delete share_exists);

use Data::Dumper;
$Data::Dumper::Maxdepth = 3;

use File::Basename;
use File::Find;
use File::Path qw(rmtree);
use Cwd;

use Carp qw(croak);
use Carp qw(confess);

use SayWrap qw(sayEOF);  ### Used mostly for debugging


#################################### 
#
#	Global vars and constants
#
#################################### 

# $use_map_cifs_share_for_user
# If '1', then re-map CIFS share to new user when creating CIFS files, 
# Otherwise, create file as $CIFS_USER, and then 'chown' the file on
# the CIFS client.
my $use_map_cifs_share_for_user = 1;

# $use_dos_file_utils
# if '1', use DOS 'fileutils' commands to create/modify/delete files/dirs on
# CIFS client
my $use_dos_file_utils = 1;
#my $fsutil_path = "C:\\%windir%\\system32\\fsutil";
my $fsutil_path;

# $_use_num_id_values
# If '1', then 'numeric id' values are used when checking quota report results.
# That is, '-q' option is passed to quota report.
my $_use_num_id_values = 0;

# $verbose
# True if 'LOGLEVEL' parameter > LOGDBG
my $verbose;

# Default size of created files
my $file_size = 102400;


# -- Hashes used to store 'filer state' --
my %nfs_user_info = ();
my %nfs_user_info_by_uid = ();
my %nfs_group_info = ();
my %nfs_group_info_by_gid = ();
my %nfs_file_info = ();
my %cifs_group_info = ();
my %cifs_group_info_by_rid = ();
my %cifs_user_info = ();
my %cifs_user_info_by_rid = ();
my %cifs_file_info = ();
my %usermap_info = ();
my %qtree_info = ();
my %quota_info = ();

# Counters used to 'gensym' new names for CIFS users and groups
my $cifs_user_index = 1;
my $cifs_group_index = 1;

# Used by functions temp_set_qtree_security/restore_qtree_security
my %_prev_qtree_security = ();

# Used by 'check quota report' functions to store most recent
# quota report results in string and hash formats (some checks
# involve comparing 'current' and 'previous' quota report results.
my $prev_qr_str;
my $prev_qr_hash;

# Used by 'check quota report' functions to store most recent
# 'logStep' string -- included in 'FAIL' messages to aid in determining
# where (in test script) failure occured.
my $prev_log_step;

#  Used to track local CIFS users and groups created, so we can delete them in cleanup
my %added_cifs_users = ();
my %added_cifs_groups = ();

my $qcstate = {};
my %QCState = ();


# -- Included for QCov_Ops_test_setup/QCov_Ops_test_cleanup routines --

# If '1', then CIFS is not setup during QCov_Ops_test_setup
my $no_cifs;

my $fc;
my $testvol;
my $mountpoint;
my $vol0_mount;
my $mapped_drive;
my @existing_cifs_users = ();

my @backup_files = ("/etc/passwd", "/etc/quotas", "/etc/usermap.cfg", "/etc/group");
my $testvol_snap = "quota_test_snap";
my $start_time = time();
my ($saved_euid, $saved_egid);
my $cifs_user_added;
my $connhost;
my $vol_filer_path;
my $vol_nosnapdir;


#################################### 
#
#	Exported functions
#
#################################### 

BEGIN
{
  our @EXPORT = qw(
	show_filer_state
	show_quota_state
	logStep
	map_root_users
	unmap_root_users
	map_all_quotas
	unmap_all_quotas
	compare_whole_report
	QCov_Ops_test_setup
	QCov_Ops_test_cleanup
	grant_cifs_user_share_access
	grant_unix_user_share_access
	grant_all_cifs_users_share_access
	grant_all_unix_users_share_access
	grant_everyone_share_access
	grant_win_machine_share_access
	read_existing_nfs_files
	read_existing_cifs_files
	remove_added_cifs_users
	remove_added_cifs_groups
	change_qtree_security
	quota_report_exists
	create_cifs_group
	delete_cifs_group
	delete_quota_entry
	map_quota_entry
	set_quota_entry
	add_quota_entry
	modify_quota_entry
	reset_quotas
	check_quota_report
	create_nfs_file
	create_nfs_dir
	create_cifs_file
	create_cifs_dir
	create_qtree
	delete_qtree
	create_nfs_user
	create_cifs_user
	delete_nfs_user
	delete_cifs_user
	map_users
	unmap_users
	create_nfs_group
	delete_nfs_group
	delete_nfs_file_or_dir
	move_nfs_file_or_dir
	chown_nfs_file_or_dir
	chown_nfs_file
	chgrp_nfs_file_or_dir
	chgrp_cifs_file
	chgrp_cifs_file_or_dir
	delete_cifs_file_or_dir
	move_cifs_file_or_dir
	chown_cifs_file
	register_test_info
  );

  Params::Validate::validation_options {
	on_fail => sub { my @msgs = @_; chomp(@msgs);
	Carp::confess(Tharn::scipt_result(@msgs)) }
  };
};


#########################################################
#
#	test functions
#
#########################################################

#sub QCov_Ops_def_tree ();
sub TEST_create_file_on_cifs_client {
	my $cifs_path = $QCState{mapped_drive} . "\\FooBar";
	my $cifs_user = create_cifs_user();
	create_file_on_cifs_client($cifs_path, 10240, $cifs_user, verify => 1);
}

sub TEST_cifs_dirtree_stat {
	my $cifs_path = 'C:\Documents and Settings';
	my $cifs_client = $QCState{cifs_client};
	my $cifs_stat = cifs_file_stat($cifs_path);
	logError("CIFS_DIRTREE_STAT  \"$cifs_path\"\n\n".Dumper($cifs_stat));
}

sub TEST_nfs_file_or_dirtree_stat {
	my $dirtree_info = nfs_file_or_dirtree_stat("/tmp/Foo");
	$Data::Dumper::Maxdepth = 4;
	logError("DIRTREE_STAT  \"/tmp/Foo\"\n\n".Dumper($dirtree_info));
	$Data::Dumper::Maxdepth = 2;
}

sub TEST_file_conv {
	%qtree_info = ( "/vol/$QCState{testvol}/qt1" => {}, "qt1" => {} );
	my $testvol = $QCState{testvol};
	my $drive = $QCState{mapped_drive};
	my $testvol_mounted_path = $QCState{testvol_mounted_path};
	my $mountpoint = $testvol_mounted_path;
	my $testvol_filer_path = $QCState{testvol_filer_path};
	my @cifs_test_paths = (
	        "$drive:/qt1/dir1/dir2/file1",
	        "$drive:/dir3/file3",
	        "$drive:\\dir3\\file3",
	        "$drive:/file4",
	        "$drive:file4",
	        "$drive:/dir5",
	        "$drive:/",
	        "$drive:\\",
	        "$drive:"
	);
	my @filer_test_paths = (
	        "/vol/$testvol/qt1/dir5/file5",
	        "/vol/$testvol/dir6/dir16/file6",
	        "/vol/$testvol/file7",
	        "/vol/$testvol/qt1",
	        "/vol/$testvol/qt1/file8",
	        "/vol/$testvol/",
	        "/vol/$testvol"
	);
	my @mounted_test_paths = (
	        "$mountpoint/qt1/dir9/file9",
	        "$mountpoint/qt1/dir10",
	        "$mountpoint/qt1/file11",
	        "$mountpoint/qt1",
	        "$mountpoint/dir12/dir22/file12",
	        "$mountpoint/dir13",
	        "$mountpoint/file14",
	        "$mountpoint/",
	        "$mountpoint"
	);
	
	for my $path (@cifs_test_paths) {
	        my $hash = get_cifs_path_parts($path);
		my $new_path = build_cifs_path(%$hash);
		my $filer_path = cvt_cifs_path_to_filer($path);
	        logDbg("GET_CIFS_PATH_PARTS:\nORIG PATH: \"$path\"\nNEW PATH: \"$new_path\"\nFILER PATH: \"$filer_path\"\n\n".
			   Dumper($hash));
	}
	
	for my $path (@filer_test_paths) {
	        my $hash = get_filer_path_parts($path);
		my $new_path = build_filer_path(%$hash);
		my $cifs_path = cvt_filer_path_to_cifs($path);
		my $mounted_path = cvt_filer_path_to_mounted($path);
	        logDbg("GET_FILER_PATH_PARTS:\nORIG PATH: \"$path\"\nNEW PATH: \"$new_path\"\nCIFS PATH: \"$cifs_path\"\n".
			  "MOUNTED PATH: \"$mounted_path\"\n\n\n".Dumper($hash));
	}
	
	for my $path (@mounted_test_paths) {
	        my $hash = get_mounted_path_parts($path);
		my $new_path = build_mounted_path(%$hash);
		my $filer_path = cvt_mounted_path_to_filer($path);
	        logDbg("GET_MOUNTED_PATH_PARTS:\nORIG PATH: \"$path\"\nNEW PATH: \"$new_path\"\nFILER PATH: \"".
			  "$filer_path\"\n\n".Dumper($hash));
	}
}


#########################################################
#
#	public functions
#
#########################################################

=head1 NAME
QCovOps - Quota Coverage Test Library
=head1 SYNOPSIS
use QCovOps;
=head1 FUNCTIONS


##########################################
#
#    Set/Modify Quota Entry public Routines
#
##########################################

=head2 delete_quota_entry($type, %opts)

	Delete quota entrie(s) which match the
	criteria given in '%opts'

=over 4 

=item	$type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.
=back

=cut
sub delete_quota_entry($%) {
	my ($type, %opts) = validate_mixed(@_,
		[{ type => SCALAR }],
		{
			default =>	{ type => SCALAR, optional => 1 },
			user => 	{ type => SCALAR, optional => 1 },
			group => 	{ type => SCALAR, optional => 1 },
			cifs_user => 	{ type => SCALAR, optional => 1 },
			cifs_group => 	{ type => SCALAR, optional => 1 },
			tree => 	{ type => SCALAR, optional => 1 },
			map =>		{ type => SCALAR, optional => 1 },
			disk => 	{ type => SCALAR, optional => 1 },
			files => 	{ type => SCALAR, optional => 1 },
		}
	);
	logDbg("Enter delete_quota_entry  TYPE: \"$type\"  OPTS:\n\n".Dumper(\%opts));
	my $map_opt = $opts{usermap} || $opts{map};
	delete($opts{map});
	$opts{usermap} = $map_opt if (defined($map_opt));
	if ($type eq "cifs_user" || $opts{cifs_user}) {
		my $user = $opts{user} || $opts{cifs_user};
		$user = get_cifs_user_fullname($user);
		$opts{user} = $user;
	} elsif ($type eq "cifs_group" || $opts{cifs_group}) {
		my $group = $opts{group} || $opts{cifs_group};
		$group = get_cifs_user_fullname($group);
		$opts{group} = $group;
	}
	my ($ignore, $quota_args) = create_quota_template($type, %opts);
	logDbg("delete_quota_entry  Calling deleteQuota:\n\n".Dumper($quota_args));
	deleteQuota($QCState{fc}, %$quota_args);
	LogException($@) if ($@);
	my $qf = sayEOF($QCState{fc}, "rdfile /etc/quotas");
	logDbg("Exit delete_quota_entry  OPTS:\n\n".Dumper(\%opts)."\n\nQUOTAS FILE:\n\n$qf");
	clean_local_state();
}


=head2 map_all_quotas()

	Adds 'QUOTA_PERFORM_USER_MAPPING ON' directive at start of /etc/quotas file,
	so that all (current and future) quota entries will be 'mapped'.

	If there already is a 'QUOTA_PERFORM_USER_MAPPING' directive at the start of
	the /etc/quotas file, it is overwritten.

	If any 'QUOTA_PERFORM_USER_MAPPING' directives occur later in the /etc/quotas file
	(not at the start), they are not affected.
=cut
sub map_all_quotas() {
	my $quotas = readQuotaFile($QCState{fc});
	if (scalar(@$quotas)) {
		mapQuota($QCState{fc});
	} else {
		quotaUserMapping($QCState{fc}, "ON");
	}
	my $qf = sayEOF($QCState{fc}, "rdfile /etc/quotas");
	logDbg("Exit map_all_quotas\n\nQUOTAS FILE:\n\n$qf");
}


=head2 unmap_all_quotas()

	Undos 'map_all_quotas' function.
	Removes 'QUOTA_PERFORM_USER_MAPPING ON' directive at start of /etc/quotas file,

	Other 'QUOTA_PERFORM_USER_MAPPING' directives occuring later in the /etc/quotas
	file are not affected.
=cut
sub unmap_all_quotas() {
	my $quotas = readQuotaFile($QCState{fc});
	if (scalar(@$quotas)) {
		unmapQuota($QCState{fc});
	} else {
		quotaUserMapping($QCState{fc}, "OFF");
	}
	my $qf = sayEOF($QCState{fc}, "rdfile /etc/quotas");
	logDbg("Exit unmap_all_quotas\n\nQUOTAS FILE:\n\n$qf");
}


=head2 map_quota_entry($type, %opts)

	"maps" quota entrie(s) using the
	"QUOTA_PERFORM_USER_MAPPING ON" directive.

	The entrie(s) affected are determined by the
	criteria given by '%opts'

=over 4

=item	type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.
=back

=cut
sub map_quota_entry($%) {
	my ($type, %opts) = validate_mixed(@_,
		[{ type => SCALAR }],
		{
			default =>	{ type => SCALAR, optional => 1 },
			user => 	{ type => SCALAR, optional => 1 },
			group => 	{ type => SCALAR, optional => 1 },
			cifs_user => 	{ type => SCALAR, optional => 1 },
			cifs_group => 	{ type => SCALAR, optional => 1 },
			tree => 	{ type => SCALAR, optional => 1 },
			disk => 	{ type => SCALAR, optional => 1 },
			files => 	{ type => SCALAR, optional => 1 },
			map =>		{ type => SCALAR, optional => 1 },
		}
	);
	if ($type eq "cifs_user" || $opts{cifs_user}) {
		my $user = $opts{user} || $opts{cifs_user};
		$user = get_cifs_user_fullname($user);
		$opts{user} = $user;
	} elsif ($type eq "cifs_group" || $opts{cifs_group}) {
		my $group = $opts{group} || $opts{cifs_group};
		$group = get_cifs_user_fullname($group);
		$opts{group} = $group;
	}
	my ($ignore, $quota_args) = create_quota_template($type, %opts);
	mapQuota($QCState{fc}, %$quota_args);
	LogException($@) if ($@);
	clean_local_state();
}


=head2 unmap_quota_entry($type, %opts)

	"unmaps" existing quota entrie(s) by deleting
	it's associated "QUOTA_PERFORM_USER_MAPPING ON"
	directive.
=over 4
	
=item	$type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.

=back

=cut
sub unmap_quota_entry($%) {
	my ($type, %opts) = validate_mixed(@_,
		[{ type => SCALAR }],
		{
			default =>	{ type => SCALAR, optional => 1 },
			user => 	{ type => SCALAR, optional => 1 },
			group => 	{ type => SCALAR, optional => 1 },
			cifs_user => 	{ type => SCALAR, optional => 1 },
			cifs_group => 	{ type => SCALAR, optional => 1 },
			tree => 	{ type => SCALAR, optional => 1 },
			disk => 	{ type => SCALAR, optional => 1 },
			files => 	{ type => SCALAR, optional => 1 },
			map =>		{ type => SCALAR, optional => 1 },
		}
	);
	my ($ignore, $quota_args) = create_quota_template($type, %opts);
	unmapQuota($QCState{fc}, %$quota_args);
	LogException($@) if ($@);
	clean_local_state();
}


=head2 set_quota_entry($type, %opts)

	Create or modify a quota entry
=over 4
	
=item	$type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.

		disk:	Disk quota limit. Default is "-"

		files:	Files quota limit. Default is "-"

		map:  If '1', then quota entry is 'mapped' by wrapping
			it with "QUOTA_PERFORM_USER_MAPPING ON" and
			"QUOTA_PERFORM_USER_MAPPING off" directives.
			Default is '0'.
=back

=cut
sub set_quota_entry($%) {
	my ($type, %opts) = @_;
	logDbg("Enter set_quota_entry  TYPE: \"$type\"  OPTS:\n\n".Dumper(\%opts));
	my $map_opt = $opts{usermap} || $opts{map};
	delete($opts{map});
	$opts{usermap} = $map_opt if (defined($map_opt));
	if ($type eq "cifs_user" || $opts{cifs_user}) {
		my $user = $opts{user} || $opts{cifs_user};
		$user = get_cifs_user_fullname($user);
		$opts{user} = $user;
	} elsif ($type eq "cifs_group" || $opts{cifs_group}) {
		my $group = $opts{group} || $opts{cifs_group};
		$group = get_cifs_user_fullname($group);
		$opts{group} = $group;
	}
	my ($qinfo, $set_quota_args) = create_quota_template($type, %opts);
	logDbg("set_quota_entry:  Returned from create_quota_template   QINFO:\n\n".Dumper($qinfo));
	my $key = $qinfo->{key};
	logDbg("set_quota_entry  KEY: \"$key\"");

	logDbg("Calling set_quota_entry  setSingleQuota:\n\n".Dumper($set_quota_args));
	setSingleQuota($QCState{fc}, %$set_quota_args);
	LogException($@) if ($@);

	my $qf = sayEOF($QCState{fc}, "rdfile /etc/quotas");
	logDbg("Exiting set_quota_entry   QINFO:\n\n".Dumper($qinfo)."\n\nQUOTAS FILE:\n\n$qf");
	clean_local_state();
}


=head2 add_quota_entry($type, %opts)

	Create a new quota entry
=over 4

=item	$type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.

		disk:	Disk quota limit. Default is "-"

		files:	Files quota limit. Default is "-"

		map:	If '1', then quota entry is 'mapped' by wrapping
			it with "QUOTA_PERFORM_USER_MAPPING ON" and
			"QUOTA_PERFORM_USER_MAPPING off" directives.
			Default is '0'.
=back

=cut
sub add_quota_entry($%) {
	my ($type, %opts) = validate_mixed(@_,
		[{ type => SCALAR }],
		{
			default =>	{ type => SCALAR, optional => 1, default => 0 },
			user => 	{ type => SCALAR, optional => 1 },
			group => 	{ type => SCALAR, optional => 1 },
			cifs_user => 	{ type => SCALAR, optional => 1 },
			cifs_group => 	{ type => SCALAR, optional => 1 },
			tree => 	{ type => SCALAR, optional => 1, default => "" },
			disk => 	{ type => SCALAR, optional => 1, default => "-" },
			files => 	{ type => SCALAR, optional => 1, default => "-" },
			map =>		{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	logDbg("Enter add_quota_entry  TYPE: \"$type\"  OPTS:\n\n".Dumper(\%opts));
	set_quota_entry($type, %opts);
}


=head2 modify_quota_entry($type, %opts)

	Modify an existing quota entry
=over 4
	
=item	$type:	One of: 'user', 'group', or 'tree'

=item	Opts:

		default: If '1', indicates this is a 'default'
			 quota (target is '*'). Default is '0'

		user:	Name of user target ('user' type only)

		group:	Name of group target ('group type only)

		tree:	For 'tree' type, name of the tree target.
			May alse be used to specify the tree
			to which the quota is to be applied,
			for 'user' or 'group' types.

		disk:	Disk quota limit. Default is "-"

		files:	Files quota limit. Default is "-"

		map:	If '1', then quota entry is 'mapped' by wrapping
			it with "QUOTA_PERFORM_USER_MAPPING ON" and
			"QUOTA_PERFORM_USER_MAPPING off" directives.
			Default is '0'.
=back

=cut
sub modify_quota_entry($%) {
	my ($type, %opts) = validate_mixed(@_,
		[{ type => SCALAR }],
		{
			default =>	{ type => SCALAR, optional => 1, default => 0 },
			user => 	{ type => SCALAR, optional => 1 },
			group => 	{ type => SCALAR, optional => 1 },
			cifs_user => 	{ type => SCALAR, optional => 1 },
			cifs_group => 	{ type => SCALAR, optional => 1 },
			tree => 	{ type => SCALAR, optional => 1, default => "" },
			disk => 	{ type => SCALAR, optional => 1, default => "-" },
			files => 	{ type => SCALAR, optional => 1, default => "-" },
			map =>		{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	set_quota_entry($type, %opts);
}


=head2 reset_quotas()

	Deletes all quotas for test volume
=cut
sub reset_quotas() {
	deleteQuota($QCState{fc}, vol => $QCState{testvol});
	clean_local_state();
}


##########################################
#
#    File Routines (NFS, CIFS)
#
##########################################

=head2 create_nfs_file($name, $user, $qtree, %opts)

	Create a new NFS file.
=over 4

=item	$name:	Name of file to be created.

=item	$user:	Name of user to own the file.
		This user must have previously been created
		using 'create_nfs_user'

=item	$qtree: Qtree in which to create this file. If value
		is "", then file is created directly under the
		test voluume (no qtree).

=item	Opts:
		size:	Size in bytes of file to be created. Default
			is 100k.

		dir:	Directory to contain the new file. Default
			is none -- create new file directly under
			volume or qtree.

=back
	Returns path (on filer) of new file.
=cut
sub create_nfs_file($$;$%) {
	my ($fname, $user, $qtree, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			size	=>	{ type => SCALAR, optional => 1, default => 102400 },
			dir	=>	{ type => SCALAR, optional => 1, default => "" },
		}
	);
	if (!$fname) {
		LogException("create_nfs_file: No fname given");
		return;
	}
	my $verify = 0;
	$user = "root" unless ($user);
	my $dir = $opts{dir} || "";
	get_filer_nfs_state();
	my $fsize = (defined($opts{size}) ? $opts{size} : 102400);
	if ($fsize =~ /^(\d+)([A-Za-z])$/) {
		my $fsize_orig = $fsize;
		my $numsize = $1;
		my $sfx = $2;
		$sfx = uc($sfx);
		if ($sfx eq 'K') {
			$fsize = $numsize * 1024;
		} elsif ($sfx eq 'M') {
			$fsize = $numsize * 1024 * 1024;
		} elsif ($sfx eq 'G') {
			$fsize = $numsize * 1024 * 1024 * 1024;
		}
	}

	logDbg("Enter create_nfs_file:  QTREE: \"$qtree\"  FNAME: \"$fname\"  USER: \"$user\"");
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	if ($qtname && !is_qtree($qtname)) {
		LogException("create_nfs_file: No such qtree: \"$qtname\"");
		return;
	}
	my $uinfo = $nfs_user_info{$user};
	if (!defined($uinfo)) {
		LogException("create_nfs_file: Could not find user: $user");
		return;
	}
	my $login = $uinfo->{login};
	my $uid = $uinfo->{uid};
	my $gid = $uinfo->{gid};
	my $group = $uinfo->{group};
	my $fpath = build_filer_path(  qtree => $qtname,
				  	dir => $dir,
					name =>$fname);
	my $dirpath = dirname($fpath);
	if (!is_qtree($dirpath)) {
		my $dinfo = $nfs_file_info{$dirpath};
		if (!defined($dinfo)) {
			LogException("create_nfs_file: dir not found: \"$dirpath\"");
			return;
		}
		if ($dinfo->{type} ne 'dir') {
			LogException("create_nfs_file: dir not found: \"$dirpath\"");
			return;
		}
	}
	my $finfo = $nfs_file_info{$fpath};
	if (defined($finfo)) {
		LogException("create_nfs_file: file already exists: \"$fpath\"");
		### return;
	}
	my $mounted_path = cvt_filer_path_to_mounted($fpath);
	my $mounted_dirpath = dirname($mounted_path);
	my $file_obj = fioOpen(vol => $QCState{testvol},
	                       name => $fname,
	                       dir => $mounted_dirpath,
	                       mode => ">");
	$file_obj->WriteSeq($fsize);
	$file_obj->Close();
	if (! -e $mounted_path) {
		LogException("create_nfs_file_or_dir:  Could not create file \"$fpath\"");
	}
	chown($uid, $gid, $mounted_path)
		|| LogException("create_nfs_file: chown \"$mounted_path\": $!");

	if ($verify) {
		my $finfo = nfs_file_stat($fpath);
	}
	clean_local_state();
	logDbg("Exit create_nfs_file:\n\n".Dumper(get_filer_nfs_state())); clean_local_state();
	return($fpath);
}


=head2 create_cifs_file($name, $user, $qtree, %opts)

	Create a new CIFS file.
=over 4

=item	$name:	Name of file to be created.

=item	$user:	Name of user to own the file.
		This user must have previously been created
		using 'create_cifs_user'

=item	$qtree: Qtree in which to create this file. If value
		is "", then file is created directly under the
		test voluume (no qtree).

=item	Opts:
		size:	Size in bytes of file to be created. Default
			is 100k.

		dir:	Directory to contain the new file. Default
			is none -- create new file directly under
			volume or qtree.
=back
	Returns path (on filer) of new file.
=cut
sub create_cifs_file($;$%) {

	die("This test needs to be run as root!")
	    if ($> != 0);

	my ($fname, $user, $qtree, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type  => SCALAR },
		 { type => SCALAR },
		],
		{
			size	=>	{ type => SCALAR, optional => 1, default => 102400 },
			dir	=>	{ type => SCALAR, optional => 1, default => "" },
		}
	);
	die("create_cifs_files: CIFS not setup") if ($no_cifs);
	my $verify = 1;
	if (!$fname) {
		LogException("create_cifs_file: No fname given");
		return;
	}
	my $dir = $opts{dir};
	$user = "administrator" unless ($user);
	get_filer_cifs_state();
	my $fsize = (defined($opts{size}) ? $opts{size} : 102400);
	if ($fsize =~ /^(\d+)([A-Za-z])$/) {
		my $fsize_orig = $fsize;
		my $numsize = $1;
		my $sfx = $2;
		$sfx = uc($sfx);
		if ($sfx eq 'K') {
			$fsize = $numsize * 1024;
		} elsif ($sfx eq 'M') {
			$fsize = $numsize * 1024 * 1024;
		} elsif ($sfx eq 'G') {
			$fsize = $numsize * 1024 * 1024 * 1024;
		}
	}

	logDbg("Enter create_cifs_file:  QTREE: \"$qtree\"  FNAME: \"$fname\"  USER: \"$user\"");
	my $do_chown = !$use_map_cifs_share_for_user;
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	if ($qtname && !is_qtree($qtname)) {
		LogException("create_cifs_file: No such qtree: \"$qtname\"");
		return;
	}
	### temp_set_qtree_security($qtpath, "ntfs");
	my $uinfo = $cifs_user_info{$user};
	if (!defined($uinfo)) {
		LogException("create_cifs_file: Could not find user: $user");
		return;
	}
	my $cifs_user = $uinfo->{login};
	my $cifs_user_fullname = get_cifs_user_fullname($cifs_user);
	my $cifs_group = $uinfo->{group};
	my $cifs_group_fullname;
	$cifs_group_fullname = get_cifs_user_fullname($cifs_group) if ($cifs_group);
	my $fpath = build_filer_path(  qtree => $qtname,
				  	dir => $dir,
					name =>$fname);
	my $dirpath = dirname($fpath);
	if (!is_qtree($dirpath)) {
		my $dinfo = $cifs_file_info{$dirpath};
		if (!defined($dinfo)) {
			LogException("create_cifs_file: dir not found: \"$dirpath\"");
			return;
		}
		if ($dinfo->{type} ne 'dir') {
			LogException("create_cifs_file: dir not found: \"$dirpath\"");
			return;
		}
	}
	my $finfo = $cifs_file_info{$fpath};
	if (defined($finfo)) {
		LogException("create_fs_file: file already exists: \"$fpath\"");
		### return;
	}
	my $cifs_path = cvt_filer_path_to_cifs($fpath);
	if ($use_map_cifs_share_for_user) {
		map_cifs_share_for_user($cifs_user_fullname);
		if ($use_dos_file_utils) {
			dos_create_file($QCState{cifs_client}, $cifs_path, $fsize);
		} else {
			create_file_on_cifs_client($cifs_path, $fsize);
		}
	}
	if ($do_chown) {
		### create_file_on_cifs_client($cifs_path, $fsize);
		logDbg("create_cifs_file  setting owner to \"$cifs_user_fullname\"  \"$cifs_path\"");
		sleep(2);
	        eval { chownCIFSFile($QCState{cifs_client}, $cifs_user_fullname, $cifs_path); die($@) if ($@) };
		logDbg("create_cifs_file  setting group to \"$cifs_user_fullname\"  \"$cifs_path\"");
		if ($@) {
			LogException("create_cifs_file: Error from chownCIFSFile: $@", action => "notify");
		}
		if ($cifs_group) {
			logDbg("create_cifs_file  setting group to \"$cifs_group_fullname\"  \"$cifs_path\"");
			sleep(2);
	        	eval { chgrpCIFSFile($QCState{cifs_client}, $cifs_group_fullname, $cifs_path); die($@) if ($@) };
			if ($@) {
				LogException("create_cifs_file: Error from chgrpCIFSFile: $@", action => "notify");
			}
		}
	}
	if ($verify) {
		my $finfo = cifs_file_stat($cifs_path);
		if (!defined($finfo)) {
			LogSevereException("create_cifs_file: Could not create file \"$cifs_path\"");
		}
		# if ( ($cifs_user ne "") && !cifs_user_match($cifs_user, $finfo->{owner}) ) {
		#	LogSevereException("create_cifs_file: Could not chown  file \"$cifs_path\" to \"$cifs_user_fullname\"");
		# }
		logDbg("Exit create_cifs_file\n\n\"$cifs_path\"\nOwner: \"$cifs_user\"\nActual Owner: \"$finfo->{owner}\"\n".
			  "Group: \"$cifs_group\"\nActual Group: \"$finfo->{group}\"");
	}
	### restore_qtree_security($qtpath);
	clean_local_state();
	return($fpath);
}


=head2 create_nfs_dir($name, $user, $qtree, %opts)

	Create a new directory via NFS
=over 4

=item	$name:	Name of directory to be created.

=item	$user:	Name of user to own the file.
		This user must have previously been created
		using 'create_nfs_user'

=item	$qtree: Qtree in which to create this file. If value
		is "", then directory is created directly under
		the test voluume (no qtree).

=item	Opts:
		dir:	Directory to contain the new file. Default
			is none -- create new file directly under
			volume or qtree.
=back
	Returns path (on filer) of new directory.
=cut
sub create_nfs_dir($$;$%) {
	my ($dname, $user, $qtree, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			dir	=>	{ type => SCALAR, optional => 1, default => "" },
		}
	);
	if (!$dname) {
		LogException("create_nfs_dir: No dname given");
		return;
	}
	my $verify = 0;
	$user = "root" unless ($user);
	my $dir = $opts{dir} || "";
	get_filer_nfs_state();
	logDbg("Enter create_nfs_dir:  QTREE: \"$qtree\"  DNAME: \"$dname\"  USER: \"$user\"");
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	if ($qtname && !is_qtree($qtname)) {
		LogException("create_nfs_dir: No such qtree: \"$qtname\"");
		return;
	}
	my $uinfo = $nfs_user_info{$user};
	if (!defined($uinfo)) {
		LogException("Could not find user: $user");
		return;
	}
	my $login = $uinfo->{login};
	my $uid = $uinfo->{uid};
	my $gid = $uinfo->{gid};
	my $group = $uinfo->{group};
	my $fpath = build_filer_path(  qtree => $qtname,
				  	dir => $dir,
					name => $dname);
	my $dirpath = dirname($fpath);
	if (!is_qtree($dirpath)) {
		my $dinfo = $nfs_file_info{$dirpath};
		if (!defined($dinfo)) {
			LogException("create_nfs_dir: dir not found: \"$dirpath\"");
			return;
		}
		if ($dinfo->{type} ne 'dir') {
			LogException("create_nfs_dir: dir not found: \"$dirpath\"");
			return;
		}
	}
	my $finfo = $nfs_file_info{$fpath};
	if (defined($finfo)) {
		LogException("create_nfs_dir: file already exists: \"$fpath\"");
		### return;
	}
	my $mounted_path = cvt_filer_path_to_mounted($fpath);
	my $mounted_dirpath = dirname($mounted_path);
	mkdir($mounted_path);
	if (! -d $mounted_path) {
		LogException("create_nfs_dir: Could not create directory \"$fpath\"");
	}
	chown($uid, $gid, $mounted_path)
		|| LogException("create_nfs_dir: chown \"$mounted_path\": $!");
	if ($verify) {
		my $finfo = nfs_file_stat($fpath);
	}
	clean_local_state();
	return($fpath);
}


=head2 create_cifs_dir($name, $user, $qtree, %opts)

	Create a new directory via CIFS
=over 4

=item	$name:	Name of directory to be created.

=item	$user:	Name of user to own the file.
		This user must have previously been created
		using 'create_cifs_user'

=item	$qtree: Qtree in which to create this file. If value
		is "", then directory is created directly under
		the test voluume (no qtree).


=item	Opts:
		dir:	Directory to contain the new file. Default
			is none -- create new file directly under
			volume or qtree.
=back
	Returns path (on filer) of new directory.
=cut
sub create_cifs_dir($$;$%) {
	my ($dname, $user, $qtree, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			dir	=>	{ type => SCALAR, optional => 1, default => "" },
		}
	);
	die("create_cifs_dir: CIFS not setup") if ($no_cifs);
	my $verify = 1;
	my $do_chown = !$use_map_cifs_share_for_user;
	if (!$dname) {
		LogException("create_cifs_dir: No dname given");
		return;
	}
	$user = "root" unless ($user);
	my $dir = $opts{dir} || "";
	get_filer_cifs_state();
	logDbg("Enter create_cifs_dir:  QTREE: \"$qtree\"  DNAME: \"$dname\"  USER: \"$user\"");
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	if ($qtname && !is_qtree($qtname)) {
		LogException("create_cifs_dir: No such qtree: \"$qtname\"");
		return;
	}
	my $uinfo = $cifs_user_info{$user};
	if (!defined($uinfo)) {
		LogException("Could not find user: $user");
		return;
	}
	my $cifs_user = $uinfo->{login};
	my $cifs_user_fullname = get_cifs_user_fullname($cifs_user);
	my $cifs_group = $uinfo->{group};
	my $cifs_group_fullname;
	$cifs_group_fullname = get_cifs_user_fullname($cifs_group) if ($cifs_group);
	my $fpath = build_filer_path(  qtree => $qtname,
				  	dir => $dir,
					name => $dname);
	my $dirpath = dirname($fpath);
	if (!is_qtree($dirpath)) {
		my $dinfo = $cifs_file_info{$dirpath};
		if (!defined($dinfo)) {
			LogException("create_cifs_dir: dir not found: \"$dirpath\"");
			return;
		}
		if ($dinfo->{type} ne 'dir') {
			LogException("create_cifs_dir: dir not found: \"$dirpath\"");
			return;
		}
	}
	my $finfo = $cifs_file_info{$fpath};
	if (defined($finfo)) {
		LogException("create_cifs_dir: file already exists: \"$fpath\"");
		### return;
	}
	my $cifs_path = cvt_filer_path_to_cifs($fpath);
	if ($use_map_cifs_share_for_user) {
		map_cifs_share_for_user($cifs_user_fullname);
		if ($use_dos_file_utils) {
			dos_mkdir($QCState{cifs_client}, $cifs_path);
		} else {
			create_file_on_cifs_client($cifs_path, 'dir');
		}
	}
	if ($do_chown) {
		create_file_on_cifs_client($cifs_path, 'dir', $cifs_user_fullname);
		logDbg("create_cifs_dir  setting owner to \"$cifs_user_fullname\"  \"$cifs_path\"");
		sleep(2);
	        eval { chownCIFSFile($QCState{cifs_client}, $cifs_user_fullname, $cifs_path); die($@) if ($@) };
		if ($@) {
			LogException("create_cifs_dir: Error from chownCIFSFile: $@");
		}
		if ($cifs_group) {
			logDbg("create_cifs_dir  setting group to \"$cifs_group_fullname\"  \"$cifs_path\"");
			sleep(2);
	        	eval { chgrpCIFSFile($QCState{cifs_client}, $cifs_group_fullname, $cifs_path); die($@) if ($@) };
			if ($@) {
				LogException("create_cifs_dir: Error from chownCIFSFile: $@");
			}
		}
	}
	if ($verify) {
		my $finfo = cifs_file_stat($cifs_path);
		if (!defined($finfo)) {
			LogSevereException("create_cifs_dir: Could not create directory \"$cifs_path\"");
		}
		# if ( ($cifs_user ne "") && !cifs_user_match($cifs_user, $finfo->{owner}) ) {
		#	LogSevereException("create_cifs_dir: Could not chown  file \"$cifs_path\" to \"$cifs_user\"");
		# }
		logDbg("Exit create_cifs_dir\n\n\"$cifs_path\"\nOwner: \"$cifs_user\"\nActual Owner: \"$finfo->{owner}\"\n".
			   "Group: \"$cifs_group\"\nActual Group: \"$finfo->{group}\"");
	}
	clean_local_state();
	return($fpath);
}


=head2 delete_nfs_file($fpath, %opts)

	Delete existing NFS file or directory.
	If directory, then contents are deleted also.
=over 4

=item	$fpath:	Path (on filer) of file/dir to be deleted.
		This path is the return value given by
		create_nfs_file() or create_nfs_dir().

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then delete the directory tree rooted
				at 'fpath'. Default is 1.
=back

=cut
sub delete_nfs_file {
	my ($fpath, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 1 },
		}
	);
	logDbg("Enter delete_nfs_file_or_dir  fpath: \"$fpath\"");
	my $verify = 0;
	get_filer_nfs_state();
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	logDbg("Enter delete_nfs_file_or_dir\n\nFPATH: \"$fpath\"\nRECURSIVE: \"$recursive\"\n\nCurrent NFS Files:\n\n".
		    Dumper(%nfs_file_info));
	if (is_nondefault_qtree($fpath)) {
		logDbg("delete_nfs_file_or_dir  Line: ".__LINE__."   Calling delete_qtree");
		return( delete_qtree($fpath) );
	}
	my $parts = get_filer_path_parts($fpath);
	my $qtree = $parts->{qtree};
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	my $finfo = $nfs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("delete_nfs_file_or_dir: No such file \"$fpath\"");
		return;
	}
	my $mounted_path = cvt_filer_path_to_mounted($fpath);
	if ($recursive && -d $mounted_path) {
		### Filer::Path::rmtree($mounted_path);
		nfs_rmtree($mounted_path);
	} elsif (-d $mounted_path) {
		rmdir($mounted_path)
			|| LogException("delete_nfs_file_or_dir: rmdir \"$mounted_path\": $!");
	} else {
		unlink($mounted_path)
			|| LogException("delete_nfs_file_or_dir: unlink \"$mounted_path\": $!");
	}
	if (-e $mounted_path) {
		LogException("delete_nfs_file_or_dir: Could not delete \"$mounted_path\"");
	}
	if ($verify) {
		my $finfo = nfs_file_stat($fpath);
	}
	clean_local_state();
	logDbg("Exit delete_nfs_file_or_dir");
}
sub delete_nfs_file_or_dir {
	my ($fpath, %opts) = @_;
	return( delete_nfs_file($fpath, %opts) );
}


=head2 delete_cifs_file($fpath, %opts)

	Delete existing CIFS file or directory.
	If directory, then contents are deleted also.
=over 4

=item	$fpath:	Path (on filer) of file/dir to be deleted.
		This path is the return value given by
		create_cifs_file() or create_cifs_dir().

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then delete the directory tree rooted
				at 'fpath'. Default is 1.
=back

=cut
sub delete_cifs_file {
	my ($fpath, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 1 },
		}
	);
	die("delete_cifs_file: CIFS not setup") if ($no_cifs);
	my $verify = 1;
	logDbg("Enter delete_cifs_file_or_dir  fpath: \"$fpath\"");
	get_filer_cifs_state();
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	if (is_nondefault_qtree($fpath)) {
		return( delete_qtree($fpath) );
	}
	my $parts = get_filer_path_parts($fpath);
	my $qtree = $parts->{qtree};
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	my $finfo = $cifs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("delete_cifs_file_or_dir: No such file \"$fpath\"");
		return;
	}
	$recursive = 0 unless ($finfo->{type} eq "dir");
	my $cifs_path = cvt_filer_path_to_cifs($fpath);
	logDbg("delete_cifs_file_or_dir  Line: ".__LINE__."    FPATH: \"$fpath\"  CIFS_PATH: \"$cifs_path\"");
	if ($use_dos_file_utils) {
		if ($finfo->{type} eq "dir") {
			dos_rmdir($QCState{cifs_client}, $cifs_path, recursive => $recursive);
		} else {
			dos_rmfile($QCState{cifs_client}, $cifs_path);
		}
	} else {
		delete_file_or_dir_on_cifs_client($cifs_path, recursive => $recursive);
	}
	if ($verify) {
		my $finfo = cifs_file_stat($cifs_path);
		logDbg("Exit delete_cifs_file_or_dir\n\nSTAT:\n\n".Dumper($finfo));
		if (defined($finfo)) {
			LogSevereException("delete_file_on_cifs_client: Could not delete \"$cifs_path\"");
		}
	}
	### delete_cifs_file_via_nfs($fpath, recursive => $recursive);
	logDbg("Exit delete_cifs_file_or_dir  fpath: \"$fpath\"");
	clean_local_state();
}
sub delete_cifs_file_or_dir {
	my ($fpath, %opts) = @_;
	return( delete_cifs_file($fpath, %opts) );
}


=head2 move_nfs_file($src_path, $dest_qtree, $dest_name, $dest_dir)

       Move or rename an existing NFS file or directory.
=over 4

=item	src_path:       Current path (on filer) of file or directory.

=item	dest_qtree:	Name of qtree to which file is to be moved.
			If moved directly under volume then this param
			should be "".

=item	dest_name	New name of file. If "", then "src" file name is
			used.

=item	dest_dir	If given, then the directory to which the file is to
			be moved. Default is "" (no directory).
=back
			Returns new path (destination).
=cut
sub move_nfs_file($$;$$) {
	my ($src_path, $dest_qtree, $dest_name, $dest_dir) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR|UNDEF },
		],
		{
		}
	);
	logDbg("Enter move_nfs_file\nSRC_PATH: \"$src_path\"\nDEST_QTREE: \"$dest_qtree\"\nDEST_NAME: \"$dest_name\"\n".
		    "DEST_DIR: \"$dest_dir\"");
	my $verify = 0;
	get_filer_nfs_state();
	my $dest_dir = "" unless ($dest_dir);
	my $src_name = basename($src_path);
	my $dest_qtree = get_qtree_name($dest_qtree);
	my $dest_name = $dest_name || $src_name;
	my $dest_path = build_filer_path( qtree => $dest_qtree,
					  dir => $dest_dir,
					  name =>  $dest_name);

	my $destdir_path = dirname($dest_path);
	my $src_info = $nfs_file_info{$src_path};
	my $dest_info = $nfs_file_info{$dest_path};
	my $destdir_info = $nfs_file_info{$destdir_path};
	my $src_parts = get_filer_path_parts($src_path);
	my $src_qtree = $src_parts->{qtree};
	my $src_qtname = get_qtree_name($src_qtree);
	my $src_qtpath = get_qtree_path($src_qtree);
	if (is_qtree($src_path)) {
		LogException("move_nfs_file_or_dir: \"$src_path\" is a qtree");
		return;
	}
	if (!defined($src_info)) {
		LogException("move_nfs_file_or_dir: No such file \"$src_path\"");
		return;
	}
	if ($dest_qtree && !is_qtree($dest_qtree) ) {
		LogException("move_nfs_file_or_dir: no such qtree \"$dest_qtree\"");
		return;
	}
	if (defined($dest_info)) {
		LogException("move_nfs_file_or_dir: File already exists \"$dest_path\"");
		### return;
	}
	if (! defined($destdir_info)) {
		LogException("move_nfs_file_or_dir: No such directory \"$dest_path\"");
		return;
	}
	if ($destdir_info->{type} ne 'dir') {
		LogException("move_nfs_file_or_dir: Not a directory: \"$dest_path\"");
		return;
	}
	my $dest_info = $nfs_file_info{$dest_path};
	my $src_path_mounted = cvt_filer_path_to_mounted($src_path);
	my $dest_path_mounted = cvt_filer_path_to_mounted($dest_path);
	my $dest_parent = dirname($dest_path);
	my $dest_parent_mounted = cvt_filer_path_to_mounted($dest_parent);
	rename($src_path_mounted, $dest_path_mounted)
		|| LogException("move_nfs_file: rename \"$src_path_mounted\" \"$dest_path_mounted\": $!");
	if ( -e $src_path_mounted || ! -e $dest_path_mounted) {
		LogException("move_nfs_file_or_dir: Could not move \"$src_path\" to \"$dest_path\"");
	}
	if ($verify) {
		my $src_info = nfs_file_stat($src_path);
		my $dest_info = nfs_file_stat($dest_path);
	}
	clean_local_state();
	logDbg("move_nfs_file  Returning: \"$dest_path\"");
	return( $dest_path );
}
sub move_nfs_file_or_dir {
	my ($src_path, $dest_qtree, $dest_name, $dest_dir) = @_;
	return( move_nfs_file($src_path, $dest_qtree, $dest_name, $dest_dir) );
}


=head2 move_cifs_file($src_path, $dest_qtree, $dest_name, $dest_dir)

       Move or rename an existing CIFS file or directory.
=over 4

=item	src_path:       Current path (on filer) of file or directory.

=item	dest_qtree:	Name of qtree to which file is to be moved.
			If moved directly under volume then this param
			should be "".

=item	dest_name	New name of file. If "", then "src" file name is
			used.

=item	dest_dir	If given, then the directory to which the file is to
			be moved. Default is "" (no directory).
=back

=cut
sub move_cifs_file($$;$$) {
	my ($src_path, $dest_qtree, $dest_name, $dest_dir) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR },
		 { type => SCALAR|UNDEF },
		],
		{
		}
	);
	logDbg("Enter move_cifs_file\nSRC_PATH: \"$src_path\"\nDEST_QTREE: \"$dest_qtree\"\nDEST_NAME: \"$dest_name\"\n".
		    "DEST_DIR: \"$dest_dir\"");
	die("move_cifs_file: CIFS not setup") if ($no_cifs);
	get_filer_cifs_state();
	my $verify = 1;
	my $dest_dir = "" unless ($dest_dir);
	my $src_name = basename($src_path);
	my $src_parts = get_filer_path_parts($src_path);
	my $src_qtree = $src_parts->{qtree};
	my $src_qtname = get_qtree_name($src_qtree);
	my $src_qtpath = get_qtree_path($src_qtree);
	my $dest_qtname = get_qtree_name($dest_qtree);
	my $dest_qtpath = get_qtree_path($dest_qtree);
	my $dest_name = $dest_name || $src_name;
	my $dest_path = build_filer_path( qtree => $dest_qtname,
					  dir => $dest_dir,
					  name =>  $dest_name);

	my $destdir_path = dirname($dest_path);
	my $src_info = $cifs_file_info{$src_path};
	my $dest_info = $cifs_file_info{$dest_path};
	my $destdir_info = $cifs_file_info{$destdir_path};
	if (is_qtree($src_path)) {
		LogException("move_cifs_file_or_dir: \"$src_path\" is a qtree");
		return;
	}
	if (!defined($src_info)) {
		LogException("move_cifs_file_or_dir: No such file \"$src_path\"");
		return;
	}
	if ($dest_qtree && !is_qtree($dest_qtree) ) {
		LogException("move_cifs_file_or_dir: no such qtree \"$dest_qtree\"");
		return;
	}
	if (defined($dest_info)) {
		LogException("move_cifs_file_or_dir: File already exists \"$dest_path\"");
		### return;
	}
	if (! defined($destdir_info)) {
		LogException("move_cifs_file_or_dir: No such directory \"$dest_path\"");
	}
	if ($destdir_info->{type} ne 'dir') {
		LogException("move_cifs_file_or_dir: Not a directory: \"$dest_path\"");
		return;
	}
	my $dest_info = $cifs_file_info{$dest_path};
	my $src_cifs_path = cvt_filer_path_to_cifs($src_path);
	my $dest_cifs_path = cvt_filer_path_to_cifs($dest_path);
	my $dest_parent = dirname($dest_path);
	my $dest_parent_cifs = cvt_filer_path_to_cifs($dest_parent);
	if ($use_dos_file_utils) {
		dos_move_file($QCState{cifs_client}, $src_cifs_path, $dest_cifs_path);
	} else {
		move_file_on_cifs_client($src_cifs_path, $dest_cifs_path, verify => 1);
	}
	if ($verify) {
		my $src_finfo = cifs_file_stat($src_cifs_path);
		my $dest_finfo = cifs_file_stat($dest_cifs_path);
		logDbg("Exit move_file_on_cifs_client\n\nSRC STAT:\n\n".Dumper($src_finfo).
			   "\n\nDEST STAT:\n\n".Dumper($dest_finfo));
		if (defined($src_finfo)) {
			LogSevereException("move_file_on_cifs_client: Could not move \"$src_cifs_path\" to \"$dest_cifs_path\"");
		}
		if (!defined($dest_finfo)) {
			LogSevereException("move_file_on_cifs_client: Could not move \"$src_cifs_path\" to \"$dest_cifs_path\"");
		}
	}
	clean_local_state();
	logDbg("move_cifs_file  Returning: \"$dest_path\"");
	return( $dest_path );
}
sub move_cifs_file_or_dir {
	my ($src_path, $dest_qtree, $dest_name, $dest_dir) = @_;
	return( move_cifs_file($src_path, $dest_qtree, $dest_name, $dest_dir) );
}


=head2 chown_nfs_file($fpath, $new_user, %opts)

	"chown" a previously created NFS file or directory.
	Changes "owner" and "group" attributes of a file/dir to
	those associated with the given user.
=over 4

=item	$fpath:		Path (on filer) to existing NFS file/dir.

=item	$new_user:	Name of existing NFS user to be the new owner.

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then chown the directory tree rooted
				at 'fpath'. Default is 0.
=back

=cut
sub chown_nfs_file($$%) {
	my ($fpath, $new_user, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	my $verify = 0;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	get_filer_nfs_state();
	my $finfo = $nfs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("chown_nfs_file: File \"$fpath\" not found");
		return;
	}
	my $uinfo = $nfs_user_info{$new_user};
	if (!defined($uinfo)) {
		LogException("chown_nfs_file: User \"$new_user\" not found");
		return;
	}
	my $new_uid = $uinfo->{uid};
	my $new_gid = $uinfo->{gid};
	my $new_group = $uinfo->{group};
	my $mounted_path =  cvt_filer_path_to_mounted($fpath);

	my $chown_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				$dirname=~ s%/%\\%g;
				$pathname=~ s%/%\\%g;
				if (! -e $basename) {
					LogException("chown_nfs_file  Pathname \"$pathname\" not found");
					return;
				}
				chown($new_uid, $new_gid, $mounted_path)
					|| LogException("chown_nfs_file: chown \"$mounted_path\": $!");
			    };

	if ($recursive && -d $mounted_path) {
		finddepth($chown_fun, $mounted_path);
	} else {
		chown($new_uid, $new_gid, $mounted_path)
			|| LogException("chown_nfs_file: chown \"$mounted_path\": $!");
	}
	if ($verify) {
		my $finfo = nfs_file_stat($fpath);
	}
	clean_local_state();
}
sub chown_nfs_file_or_dir {
	my ($fpath, $new_user, %opts) = @_;
	return( chown_nfs_file($fpath, $new_user, %opts) );
}


=head2 chown_cifs_file($fpath, $new_user, %opts)

	"chown" a previously created CIFS file or directory.
	Changes "owner" and "group" attributes of a file/dir to
	those associated with the given user.
=back

=item	$fpath:		Path (on filer) to existing CIFS file/dir.

=item	$new_user:	Name of existing CIFS user to be the new owner.

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then chown the directory tree rooted
				at 'fpath'. Default is 0.
=back

=cut
sub chown_cifs_file($$%) {
	my ($fpath, $new_user, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	die("chown_cifs_file: CIFS not setup") if ($no_cifs);
	my $verify = 1;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	get_filer_cifs_state();
	my $parts = get_filer_path_parts($fpath);
	my $qtree = $parts->{qtree};
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	my $finfo = $cifs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("chown_cifs_file: File \"$fpath\" not found");
		return;
	}
	my $ftype = $finfo->{type};
	$recursive = 0 unless ($ftype eq "dir");
	my $uinfo = $cifs_user_info{$new_user};
	if (!defined($uinfo)) {
		LogException("chown_cifs_file: User \"$new_user\" not found");
		return;
	}
	my $cifs_user = $uinfo->{login};
	my $cifs_user_fullname = get_cifs_user_fullname($cifs_user);
	my $cifs_path =  cvt_filer_path_to_cifs($fpath);

	logDbg("chown_cifs_file:  Calling chownCIFSFile  USER: \"$cifs_user_fullname\"   PATH: \"$cifs_path\"");
	sleep(2);
	eval { chownCIFSFile($QCState{cifs_client}, $cifs_user_fullname, $cifs_path, recursive => $recursive); die($@) if ($@) };
	if ($@) {
		LogException("chown_cifs_file: Error from chownCIFSFile: $@");
	}
	if ($verify) {
		my $finfo = cifs_file_stat($fpath);
	}
	clean_local_state();
}
sub chown_cifs_file_or_dir {
	my ($fpath, $new_user, %opts) = @_;
	return( chown_cifs_file($fpath, $new_user, %opts) );
}


=head2 chgrp_nfs_file($fpath, $new_group, %opts)

	"chgrp" a previously created NFS file or directory.
	Changes "group" attributes of a file/dir to the
	given group.
=over 4

=item	$fpath:		Path (on filer) to existing NFS file/dir.

=item	$new_group:	Name of existing NFS group.

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then chown the directory tree rooted
				at 'fpath'. Default is 0.
=back

=cut
sub chgrp_nfs_file {
	my ($fpath, $new_group, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	my $verify = 0;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	get_filer_nfs_state();
	my $finfo = $nfs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("chgrp_nfs_file: File \"$fpath\" not found");
		return;
	}
	my $ginfo = $nfs_group_info{$new_group};
	if (!defined($ginfo)) {
		LogException("chgrp_nfs_file: Group \"$new_group\" not found");
		return;
	}
	my $old_uid = $finfo->{uid};
	my $old_gid = $finfo->{gid};
	my $new_gid = $ginfo->{gid};
	my $mounted_path =  cvt_filer_path_to_mounted($fpath);

	my $chown_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				$dirname=~ s%/%\\%g;
				$pathname=~ s%/%\\%g;
				if (! -e $basename) {
					LogException("chgrp_nfs_file  Pathname \"$pathname\" not found");
					return;
				}
				chown($old_uid, $new_gid, $mounted_path)
					|| LogException("chgrp_nfs_file: chown \"$mounted_path\": $!");
			    };

	if ($recursive && -d $mounted_path) {
		finddepth($chown_fun, $mounted_path);
	} else {
		chown($old_uid, $new_gid, $mounted_path)
			|| LogException("chgrp_nfs_file: chown \"$mounted_path\": $!");
	}
	if ($verify) {
		my $finfo = nfs_file_stat($fpath);
	}
	clean_local_state();
}
sub chgrp_nfs_file_or_dir {
	my ($fpath, $new_group, %opts) = @_;
	return( chgrp_nfs_file($fpath, $new_group, %opts) );
}


=head2 chgrp_cifs_file($fpath, $new_group, %opts)

	"chgrp" a previously created CIFS file or directory.
	Changes "group" attributes of a file/dir to the
	given group.
=over 4

=item	$fpath:		Path (on filer) to existing NFS file/dir.

=item	$new_group:	Name of existing NFS group.

=item	Opts:

		recursive	If 1, and 'fpath' is a directory,
				then chown the directory tree rooted
				at 'fpath'. Default is 0.
=back

	------  DEPRECATED  ------

	'local' groups (created using 'useradmin group add') aren't seen by the
	quota subsystem, the are only used for controlling the 'capabilities'
	assigned to local users (eg. 'administrator', 'audit', etc).

	Therefore this function is not needed. It is kept to allow existing
	scripts to run. It simply returns, with no actions performed.

=cut
sub chgrp_cifs_file($$%) {
	my ($fpath, $new_group, %opts) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
			recursive	=>	{ type => SCALAR, optional => 1, default => 0 },
		}
	);
	die("chgrp_cifs_file: CIFS not setup") if ($no_cifs);
	my $verify = 1;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	get_filer_cifs_state();
	my $parts = get_filer_path_parts($fpath);
	my $qtree = $parts->{qtree};
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	my $finfo = $cifs_file_info{$fpath};
	if (!defined($finfo)) {
		LogException("chgrp_cifs_file: File \"$fpath\" not found");
		return;
	}
	my $ftype = $finfo->{type};
	$recursive = 0 unless ($ftype eq "dir");
	my $ginfo = $cifs_group_info{$new_group};
	if (!defined($ginfo)) {
		LogException("chgrp_cifs_file: Group \"$new_group\" not found");
		return;
	}
	my $cifs_group_fullname = get_cifs_user_fullname($new_group);
	my $cifs_path =  cvt_filer_path_to_cifs($fpath);

	logDbg("chgrp_cifs_file:  Calling chgrpCIFSFile  USER: \"$cifs_group_fullname\"   PATH: \"$cifs_path\"");
	sleep(2);
	eval { chgrpCIFSFile($QCState{cifs_client}, $cifs_group_fullname, $cifs_path, recursive => $recursive); die($@) if ($@) };
	if ($@) {
		LogException("chgrp_cifs_file: Error from chgrpCIFSFile: $@");
	}
	if ($verify) {
		my $finfo = cifs_file_stat($fpath);
	}
	clean_local_state();
	return($fpath);
}
sub chgrp_cifs_file_or_dir {
	my ($fpath, $new_group, %opts) = @_;
	return( chgrp_cifs_file($fpath, $new_group, %opts) );
}



##########################################
#
#    Qtree Routines
#
##########################################


=head2 change_qtree_security($qtree, $style)

       Change "security style" for a qtree.
=over 4

=item	qtree:	Path or Name of qtree.
		Use "" for "volume default" qtree.

=item	style:	One of 'cifs', 'unix', 'mixed'.
=back

=cut
sub change_qtree_security($$) {
	my ($qtree, $style) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
		}
	);
	read_existing_qtrees();
	my $qtname = get_qtree_name($qtree);
	my $qtpath = get_qtree_path($qtree);
	if ( !is_qtree($qtpath) ) {
		LogException("change_qtree_security: qtree not found: \"$qtname\"");
		return;
	}
	qtreeSecurity($QCState{fc}, $qtpath, $style);
	LogException($@) if ($@);
	if ($qtname ne "" && ($style eq "mixed" || $style eq "ntfs")) {
		my $cifs_path = cvt_filer_path_to_cifs($qtpath);
		logDbg("change_qtree_security  Calling grant_administrator_access_on_cifs_client:  "
		       ."QTPATH: \"$qtpath\"   CIFS_PATH: \"$cifs_path\"");
		grant_administratator_access_on_cifs_client($cifs_path);
	}
	clean_local_state();
}


=head2 create_qtree($qtree)

	Create a new qtree (under test volume).
=over 4

=item	$qtree:	Name of qtree to be created.
=back

=cut
sub create_qtree($) {
	my ($qtree) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{ }
	);
	logDbg("Enter create_qtree: \"$qtree\"");
	get_filer_nfs_state();
	my $qtpath = get_qtree_path($qtree);
	my $qtname = get_qtree_name($qtree);
	if ( is_qtree($qtpath) ) {
		LogException("create_qtree: qtree \"$qtpath\" already exists");
		return;
	}
	if (!$qtname) {
		LogException("create_qtree: No qtree name given");
		return;
	}
	my $finfo = $nfs_file_info{$qtpath};
	if (defined($finfo)) {
		LogException("create_qtree: file already exists \$qtpath\"");
		return;
	}
	qtreeCreate($QCState{fc}, $qtpath);
	LogException($@) if ($@);
	if ( !qtreePresent($QCState{fc}, $qtpath) ) {
		LogException("create_qtree: could not create qtree \"$qtpath\"");
	}
	clean_local_state();
	logDbg("Exit create_qtree: \"$qtpath\"");
	return($qtpath);
}


=head2 delete_qtree($qtree)

	Delete existing qtree
=over 4

=item	qtree:	Path or Name of qtree to be deleted.
=back

=cut
sub delete_qtree($) {
	my ($qtree) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{ }
	);
	read_existing_qtrees();
	my $qtpath = get_qtree_path($qtree);
	my $qtname = get_qtree_name($qtree);
	logDbg("Enter delete_qtree   QTREE: \"$qtree\"  QTNAME: \"$qtname\"   QTPATH: \"$qtpath\"");
	my $mounted_qtpath =  cvt_filer_path_to_mounted($qtpath);
	logDbg("STAT QTREE:\n\n".`stat $mounted_qtpath`);
	if (!$qtname) {
		LogException("delete_qtree: Can't delete default qtree");
		return;
	}
	my $qinfo = $qtree_info{$qtpath};
	if (!defined($qinfo)) {
		LogException("delete_qtree: No such qtree: \"$qtpath\"");
		return;
	}
	eval { qtreeDelete($QCState{fc}, $qtpath, VOL_MOUNTPATH => $QCState{testvol_mounted_path}); };
	LogException($@) if ($@);
	if ( qtreePresent($QCState{fc}, $qtpath) ) {
		LogException("delete_qtree: Could not delete qtree \"$qtpath\"");
		return;
	}
	clean_local_state();
	logDbg("Exit delete_qtree   QTREE: \"$qtree\"  QTNAME: \"$qtname\"   QTPATH: \"$qtpath\"");
	return($qtpath);
}


##########################################
#
#    User/Group Routines (NFS, CIFS)
#
##########################################


=head2 create_nfs_user(%opts)

	Create new user.
=over 4

=item	%opts:
		login:	Name of user to be created.

		uid:	UID of new user.

		group:	Name of group for user. If
			no group exists with this name
			a new one will be created (/etc/group).
=back

	Returns name of new user
=cut
sub create_nfs_user(%) {
	my (%opts) = validate_mixed(@_,
		[],
		{
			login	=>	{ type => SCALAR, optional => 1 },
			uid	=>	{ type => SCALAR, optional => 1 },
			group	=>	{ type => SCALAR, optional => 1 },
			gid	=>	{ type => SCALAR, optional => 1 },
		}
	);
	logDbg("Enter create_nfs_user  OPTS:\n\n".Dumper(\%opts));
	if ( !defined($opts{min_uid}) && !defined($opts{uid}) &&
	    !defined($opts{max_uid}) ) {
		$opts{min_uid} = 10001;
	}
	read_existing_nfs_users();
	if (defined($opts{login})) {
		my $uinfo = $nfs_user_info{$opts{login}};
		if (defined($uinfo)) {
			LogException("create_nfs_user: nfs user \"$opts{login}\" already exists");
			return;
		}
	}
	if (defined($opts{uid})) {
		my $uinfo = $nfs_user_info_by_uid{$opts{uid}};
		if (defined($uinfo)) {
			LogException("create_nfs_user: nfs user uid \"$opts{uid}\" already exists");
			return;
		}
	}
	my %new_nfs_users = createUsers($QCState{fc}, 1, %opts);
	LogException($@) if ($@);
	my @nfs_user_list = keys(%new_nfs_users);
	my $user = $nfs_user_list[0];
	clean_local_state();
	logDbg("Exit create_nfs_user: \"$user\"");
	return($user);
}


=head2 create_cifs_user(%opts)

	Create new user.
=over 4

=item	%opts:
		login:	Name of new CIFS user to be created.

		group:  Name of local group(s) for the user (comma separated list).
=back

	----  DEPRECATED OPTIONAL ARGS  ----

	The following optional args are allowed but are ignored.

	Local CIFS users don't have anything  that corresponds to
       a (user-settable) UID,  or group, so these values are ignored.

	       uid:    UID of new user.

	       group:  Name of group for user. If
		       no group exists with this name
		       a new one will be created (/etc/group).

=cut
sub create_cifs_user(%) {
	my (%opts) = validate_mixed(@_,
		[],
		{
			login	=>	{ type => SCALAR, optional => 1 },
			uid	=>	{ type => SCALAR, optional => 1 },
			group	=>	{ type => SCALAR, optional => 1 },
			gid	=>	{ type => SCALAR, optional => 1 },
		}
	);
	die("create_cifs_user: CIFS not setup") if ($no_cifs);
	logDbg("Enter create_cifs_user  OPTS:\n\n".Dumper(%opts));
	read_existing_cifs_groups();
	read_existing_cifs_users();
	my $login;
	my $login_prefix = "qcov_ops_cifs_user_";
	my $group = $opts{group};
	$group = "Users" unless (defined($group));
	my $ginfo = $cifs_group_info{$group};
	if (!defined($ginfo)) {
		create_cifs_group(group => $group);
	}
	if (defined($opts{login})) {
		my $uinfo = $cifs_user_info{$opts{login}};
		if (defined($uinfo)) {
			LogException("create_cifs_user: cifs user \"$opts{login}\" already exists");
			return;
		}
		$login = $opts{login};
	} else {
		$login = $login_prefix . $cifs_user_index++;
		while ($cifs_user_info{$login}) {
			$login = $login_prefix . $cifs_user_index++;
		}
	}
	### addCIFSUser($QCState{fc}, $login, $::CIFS_PASSWD, $group, on_client => 1);
	addCIFSUser($QCState{fc}, $login, $::CIFS_PASSWD, $group, on_client => 0);
	LogException($@) if ($@);
	### grant_cifs_user_share_access($login);
	clean_local_state();

	logDbg("Exit create_cifs_user  \"$login\"");

	$added_cifs_users{$login} = 1;
	return($login);
}


=head2 delete_nfs_user($user)

	Delete existing user.
=over 4

=item	$user:	Name of user to delete.
=back

=cut
sub delete_nfs_user($) {
	my ($user) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{ }
	);
	my ($user) = @_;
	read_existing_nfs_users();
	my $uinfo = $nfs_user_info{$user};
	if (!defined($uinfo)) {
		LogException("delete_nfs_user: no such user \"$user\"");
		return;
	}
	delPasswdEntry($QCState{fc}, $user);
	LogException($@) if ($@);
	clean_local_state();
}


=head2 delete_cifs_user($user)

	Delete existing user.
=over 4

=item	$user:	Name of user to delete.
=back

=cut
sub delete_cifs_user($) {
	my ($user) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{ }
	);
	die("delete_cifs_user: CIFS not setup") if ($no_cifs);
	read_existing_cifs_users();
	my $uinfo = $cifs_user_info{$user};
	if (!defined($uinfo)) {
		logWarn("delete_cifs_user: no such user \"$user\"");
	}
	removeCIFSUser($QCState{fc}, $user, $::CIFS_PASSWD, "Users", on_client => 1);
	LogException($@) if ($@);
	clean_local_state();
	delete($added_cifs_users{$user});
}


=head2 map_users($cifs_user, $nfs_user, $direction)

	Maps existing CIFS and NFS users by modifying /etc/usermap.cfg
=over 4

=item	cifs_user:	CIFS user to map.

=item	nfs_user:	NFS user to map.
=back

=cut
sub map_users($$) {
	my ($nfs_user, $cifs_user) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
		}
	);
	die("map_users: CIFS not setup") if ($no_cifs);
	read_existing_nfs_users();
	read_existing_cifs_users();
	read_existing_usermaps();
	logDbg("map_users Calling: addUsermap(\ncifs_user => \"$cifs_user\"\n"
			."nfs_user => \"$nfs_user\"");
	addUsermap($QCState{fc},
		    cifs_user => $cifs_user,
		    nfs_user => $nfs_user,
		    direction => "==");
	LogException($@) if ($@);
	logDbg("Exit map_users  \"$cifs_user\"  \"$nfs_user\"\n\n".
		   sayEOF($QCState{fc}, "rdfile /etc/usermap.cfg"));
	clean_local_state();
}


=head2 unmap_users($cifs_user, $nfs_user, $direction)

	Removes existing /etc/usermap.cfg entry.
=over 4

=item	cifs_user:	CIFS user in existing usermap.

=item	nfs_user:	NFS user in existing usermap.
=back

=cut
sub unmap_users($$) {
	my ($nfs_user, $cifs_user) = validate_mixed(@_,
		[
		 { type => SCALAR },
		 { type => SCALAR },
		],
		{
		}
	);
	die("unmap_users: CIFS not setup") if ($no_cifs);
	read_existing_nfs_users();
	read_existing_cifs_users();
	read_existing_usermaps();
	logDbg("unmap_users Calling: removeUsermap(\ncifs_user => \"$cifs_user\"\n"
			."nfs_user => \"$nfs_user\"");
	removeUsermap($QCState{fc},
		    cifs_user => $cifs_user,
		    nfs_user => $nfs_user);
	LogException($@) if ($@);
	logDbg("Exit unmap_users  \"$cifs_user\"  \"$nfs_user\"\n\n".
		sayEOF($QCState{fc}, "rdfile /etc/usermap.cfg"));
	clean_local_state();
}


=head2 create_nfs_group(%opts)

	Create new group
=over 4

=item	%opts:

		group:	Name of new group.

		gid:	GID of new group.
=back

=cut
sub create_nfs_group(%) {
	my (%opts) = validate_mixed(@_,
		[],
		{
			group	=>	{ type => SCALAR, optional => 1 },
			gid	=>	{ type => SCALAR, optional => 1 },
		}
	);
	logDbg("Enter create_nfs_group  OPTS:\n\n".Dumper(\%opts));
	if ( !defined($opts{min_gid}) && !defined($opts{gid}) &&
	    !defined($opts{max_gid}) ) {
		$opts{min_gid} = 10001;
	}
	my %new_nfs_groups = createGroups($QCState{fc}, 1, %opts);
	LogException($@) if ($@);
	my @nfs_group_list = keys(%new_nfs_groups);
	my $group = $nfs_group_list[0];
	clean_local_state();
	logDbg("Exit create_nfs_group \"$opts{group}\"   GROUP_INFO:\n\n".
		  Dumper(\%nfs_group_info));
	return($group);
}


=head2 delete_nfs_group($group)

	Delete existing group.
=over 4

=item	$group:	Name of group to delete.
=back

=cut
sub delete_nfs_group($) {
	my ($group) = @_;
	read_existing_nfs_groups();
	my $ginfo = $nfs_group_info{$group};
	if (!defined($ginfo)) {
		LogException("delete_nfs_group: no such group \"$group\"");
		return;
	}
	delGroupEntry($QCState{fc}, $group);
	LogException($@) if ($@);
	clean_local_state();
}


=head2 create_cifs_group($group)

	Create new CIFS group
=over 4

=item	%opts:

		group:    Name of new CIFS group

		role	Role(s) for the new group (comma separated list).
=back

=cut
sub create_cifs_group(%) {
	my (%opts) = validate_mixed(@_,
		[],
		{
			group	=>	{ type => SCALAR, optional => 1 },
			gid	=>	{ type => SCALAR, optional => 1 },
		}
	);
	die("create_cifs_group: CIFS not setup") if ($no_cifs);
	my $group = $opts{group};
	my $role = $opts{role} || "admin";
	my $group_prefix = "qcov_ops_cifs_group_";
	read_existing_cifs_groups();
	if (!defined($opts{group})) {
		$group = $group_prefix . $cifs_group_index++;
		while ($cifs_group_info{$group}) {
			$group = $group_prefix . $cifs_group_index++;
		}
	}
	my $ginfo = $cifs_group_info{$group};
	if (defined($ginfo)) {
		LogException("create_cifs_group: group \"$group\" already exists");
		return;
	}
	if ($role) {
		addCIFSGroup($QCState{fc}, $group, role => $role);
	} else {
		addCIFSGroup($QCState{fc}, $group);
	}
	if ($@) {
		LogException("create_cifs_group: Error from addCIFSGroup: \"$@\"");
	}
	$added_cifs_groups{$group} = 1;
	return;
}


=head2 delete_cifs_group($group)

	Delete an existing CIFS group
=over 4

=item	$group:    Name of CIFS group to delete.
=back

=cut
sub delete_cifs_group($) {
	my ($group) = validate_mixed(@_,
		[
		 { type => SCALAR },
		],
		{
		}
	);
	die("delete_cifs_group: CIFS not setup") if ($no_cifs);
	read_existing_cifs_groups();
	if (!defined($cifs_group_info{$group})) {
		LogException("delete_cifs_group: No such CIFS group \"$group\"");
		return;
	}
	removeCIFSGroup($QCState{fc}, $group);
	if ($@) {
		LogException("delete_cifs_group: Error from removeCIFSGroup: \"$@\"");
	}
	delete($added_cifs_groups{$group});
	return;
}



##########################################
#
#    Quota report related public routines
#
##########################################


# compare_qr_str()
#
sub compare_qr_str {
	my ($qr1, $qr2) = @_;
	my @qr1_lines = split(/\n/, $qr1);
	my @qr2_lines = split(/\n/, $qr2);
}


# compare_whole_report()
#
#	Get quota report, and compare it with "expected quota report".
#	
sub compare_whole_report {
	my $msg = $prev_log_step;
	logDbg("Enter compare_whole_report\n\n$msg");
	my $fail;
	my ($curr_qr_hash, $curr_qr_str) = get_quota_report($QCState{fc});
	if (!$prev_qr_hash) {
		$prev_qr_hash = $curr_qr_hash;
		$prev_qr_str = $curr_qr_str;
		($curr_qr_hash, $curr_qr_str) = get_quota_report($QCState{fc});
	}

	my %keys_hash = ();
	foreach my $hash (@$prev_qr_hash) {
		foreach my $k (keys(%$hash)) {
			$keys_hash{$k} = 1;
		}
	}
	foreach my $hash (@$curr_qr_hash) {
		foreach my $k (keys(%$hash)) {
			$keys_hash{$k} = 1;
		}
	}
	my @keylist = keys(%keys_hash);
	my $keys = \@keylist;

	my ($matched_pairs, $unmatched_prev, $unmatched_curr, $dup_prev, $dup_curr) =
	     compare_hash_lists($keys, $prev_qr_hash, $curr_qr_hash);
	if (scalar(@$unmatched_prev) || scalar(@$unmatched_curr)) {
		$fail = "compare_whole_report  Quota Report compare Failed for\n\n$msg\n\n".
				  "Prev Quota Report:\n\n$prev_qr_str\n\n\n\nCurr Quota Report\n\n$curr_qr_str";
		logDbg($fail);
		logresult('FAIL', $fail);
	}

	if ($fail) {
		$@ = $msg;
	} else {
		logcomment("compare_whole_report PASSED for:\n\n$msg");
	}
	clean_local_state();
	return ( !$fail );
}


# check_quota_report()
#
#	Get quota report, and compare it with "expected quota report".
#	
sub check_quota_report {
	my $msg = $prev_log_step;
	if (statusQuota($QCState{fc}, $QCState{testvol}) ne "on") {
		LogException("Error quotas are off for \"$QCState{testvol}\"");
		return;
	}

	my $fail;
	### manual_check_quota_report($msg);
	$fail = verify_quota_report_changes($msg, @_);
	clean_local_state();

	if ($fail) {
		$@ = $msg;
	} else {
		logcomment("check_quota_report PASSED for:\n\n$msg");
	}

	return( $fail == 0 );
}


##########################################
#
#    Miscellaneous public routines
#
##########################################


=head2 remove_added_cifs_users

	Used for "cleanup" routine.
=cut
sub remove_added_cifs_users() {
	die("remove_added_cifs_users: CIFS not setup") if ($no_cifs);
	foreach my $usr (keys(%added_cifs_users)) {
		removeCIFSUser($QCState{fc},
				$usr,
				### on_client => 1,
				on_client => 0,
				cifs_client => $QCState{cifs_client});
		delete($added_cifs_users{$usr});
	}
	clean_local_state();
}


=head2 remove_added_cifs_groups

	Used for "cleanup" routine.

=cut
sub remove_added_cifs_groups() {
	die("remove_added_cifs_groups: CIFS not setup") if ($no_cifs);
	foreach my $grp (keys(%added_cifs_groups)) {
		removeCIFSGroup($QCState{fc}, $grp);
		delete($added_cifs_groups{$grp});
	}
	clean_local_state();
}


=head2 map_root_users

	Maps local 'administrator' account to local Unix 'root' account.
	Also disables option: 'wafl_nt_admin_priv_map_to_root'

=cut
sub map_root_users {
	logDbg("Enter map_root_user");
	die("map_root_users: CIFS not setup") if ($no_cifs);
	addUsermap($QCState{fc},
		    cifs_user => "BUILTIN\\Administrators",
		    nfs_user => "root",
		    direction => "==");
	LogException($@) if ($@);
	setOption($QCState{fc}, "wafl.nt_admin_priv_map_to_root", "off");
	### setOption($QCState{fc}, "wafl.nt_admin_priv_map_to_root", "on");
	die("map_root_users: setOption: $@") if ($@);
}


=head2 unmap_root_users

	Removes map: local 'administrator' account to local Unix 'root' account.
	Also enables option: 'wafl_nt_admin_priv_map_to_root'

=cut
sub unmap_root_users {
	logDbg("Enter unmap_root_users");
	die("unmap_root_users: CIFS not setup") if ($no_cifs);
	removeUsermap($QCState{fc},
		    cifs_user => "BUILTIN\\Administrators",
		    nfs_user => "root");
	LogException($@) if ($@);
	setOption($QCState{fc}, "wafl.nt_admin_priv_map_to_root", "on");
	die("unmap_root_users: setOption: $@") if ($@);
}


=head2 quota_report_exists

	Returns 1 if /etc/quotas contains at least one
	entry for testvol, 0 otherwise

=cut
sub quota_report_exists() {
	my $quotas = readQuotaFile($QCState{fc});
	my $num_quotas = scalar( grep( { $_->{volname} eq $QCState{testvol} }  @$quotas ) );
	return( $num_quotas > 0 );
}


=head2 register_test_info(%opts)

	Register information to be used by the test.
=over 4

=item	opts:
		conn		=>	filer connection (connectrec)
		volume		=>	test volume
		mount_point	=>	NFS mount point on client.
		mapped_drive	=>	mapped drive on CIFS_CLIENT
		cifs_client	=>	CIFS_CLIENT parameter.
=back

=cut
sub register_test_info {
	my (%opts);
	if ($no_cifs) {
		(%opts) = validate_mixed(@_,
			[],
			{
				conn		=>	{ type => SCALAR|HASHREF },
				vol		=>	{ type => SCALAR },
				mount_point	=>	{ type => SCALAR },
			}
		);
	} else {
		(%opts) = validate_mixed(@_,
			[],
			{
				conn		=>	{ type => SCALAR|HASHREF },
				vol		=>	{ type => SCALAR },
				mount_point	=>	{ type => SCALAR },
				cifs_client	=>	{ type => SCALAR },
				mapped_drive	=>	{ type => SCALAR },
			}
		);
	}

	my %logLevels = (
        XXX             => 0,
        describe        => 100,
        step            => 200,
        trc             => 300,
        dbg             => 400,
        info            => 500,
);

logcomment("register_test_info  Line: ".__LINE__);
	die("create_cifs_group: CIFS not setup") if ($no_cifs);
	my $group = $opts{group};
	my (%opts) = @_;
	my $fc = $opts{conn};
	my $connhost = $fc->connhost();
	my $testvol = $opts{vol};
	my $cifs_client = $opts{cifs_client} || $main::CIFS_CLIENT;
	my $testvol_mounted_path = $opts{mount_point};
	my $mapped_drive = $opts{mapped_drive};
	my $cifs_share = $::CIFS_SHARE;
	my $cifs_server = getCIFSServer($fc);
	my $cifs_passwd = $::CIFS_PASSWD;
	my $cifs_user = $::CIFS_USER;
	my $dbg_loglevel = $logLevels{dbg};   ## See "lib/SPARSE/Util/Logger.pm"
	my $loglevel = $logLevels{$ENV{'LOGLEVEL'} || "step"};
	$verbose = ($loglevel >= $dbg_loglevel);
	logDbg("Line: ".__LINE__."    LOGLEVEL: \"$loglevel\"   DBG LOGLEVEL: \"$dbg_loglevel\"  Verbose: \"$verbose\"");
	%QCState = ( fc => $fc,
		     testvol => $testvol,
		     connhost => $connhost,
		     cifs_client => $cifs_client,
		     mapped_drive => $mapped_drive,
		     testvol_filer_path => "/vol/$testvol",
		     testvol_mounted_path => $testvol_mounted_path,
 		     cifs_user => $cifs_user,
 		     cifs_passwd => $cifs_passwd,
 		     cifs_share => $cifs_share,
		     cifs_server => $cifs_server,
		     interactive => eval("$main::INTERACTIVE")
                  );
		logDbg("register_test_info\n\n".Dumper($qcstate));
		# show_qcstate("Exit register_test_info");

#	if ($mapped_drive && $cifs_share) {
#		logDbg("register_test_info Calling grant_all_cifs_users_share_access");
#		grant_all_cifs_users_share_access();
#		logDbg("register_test_info Calling grant_all_unix_users_share_access");
#		grant_all_unix_users_share_access();
#		logDbg("register_test_info Calling grant_win_machine_share_access");
#		grant_win_machine_share_access();
#		logDbg("register_test_info Calling grant_everyone_share_access");
#		grant_everyone_share_access();
#		logDbg("register_test_info Returned from grant_everyone_share_access");
#	}

logcomment("register_test_info  Line: ".__LINE__);
	vol_options($fc, $testvol, nosnapdir => "on");
logcomment("register_test_info  Line: ".__LINE__);

	create_nfs_group(group => "root", gid => 0);
logcomment("register_test_info  Line: ".__LINE__);

	return( { %QCState } );
}




#########################################################
#
#       support functions
#
#########################################################



##########################################
#
#    Quota Report support routines
#
##########################################


my $prev_log_step;

sub logStep($) {
	my ($msg) = @_;
	$prev_log_step = $msg;
	SPARSE::Util::Logger::logStep($msg);
}


sub get_quota_report {
	logDbg("Enter get_quota_report  QUOTAS FILE: \n\n".sayEOF($QCState{fc}, "rdfile /etc/quotas"));
	my $flags;
	if ($_use_num_id_values) {
		$flags = "-qx";
	} else {
		$flags = "-x";
	}
	my $qr_str = sayEOF($QCState{fc}, "quota report -x");
	my $qr_hash = getQuotaReport($QCState{fc}, flag => $flags);
	### my $qr_hash2 =  getQuotaReport($QCState{fc}, flag => "-x");
	return( $qr_hash, $qr_str );
}


# manual_check_quota_report()
#
#	Get quota report, and compare it with "expected quota report".
#	
sub manual_check_quota_report {
	my ($msg) = @_;
	my $fail = 0;
	$prev_qr_str = "" unless ($prev_qr_str);
	my ($ignore, $curr_qr_str) = get_quota_report();
	logDbg("\n\n====================  Check Quota Report  ====================\n\n$msg\n\n".
                       "==============================================================\n\n\n".
		    "Prev Quota Report:\n\n$prev_qr_str\n\n\nNew Quota Report:"
		    ."\n\n$curr_qr_str\n\n\nOkay? [y|n|s] -- [yES | nO | sKIP]\n\n");
	$prev_qr_str = $curr_qr_str;
	my $resp;
	if ($QCState{interactive}) {
		$resp = readline(STDIN);
	} else {
		$resp = "s";
	}
	chomp($resp);
	$resp =~ s%\W%%;
	if ($resp eq "y") {
		### logresult('INFO', "Manual quota report check passed for: \"$msg\"");
	} elsif ($resp eq "n") {
		$fail++;
		logresult('FAIL', "check_quota_report: \"$msg\"");
	} else {
		logresult('WARN', "Manual quota report check SKIPPED for: \"$msg\"");
	}
	return($fail);
}



# sub get_quota_target_num_id($id, $type)
#
#	Given a name of a user, group or tree, and a
#	'type' string, (one of "user", "group", or "tree"),
#	returns a numeric 'id' -- e.g. 'uid' for NFS user,
#	'gid' for NFS group, etc.
#
sub get_quota_target_num_id {
	my ($id, $type) = @_;
	my $num_id = undef;
	$id =~ s/^(.*\\)//;
	return ($id) if ($id eq "*");
	return ($id) unless ($_use_num_id_values);
	if ($type eq "user") {
		read_existing_nfs_users();
		read_existing_cifs_users();
		if (exists($nfs_user_info{$id})) {
			$num_id = $nfs_user_info{$id}{uid};
		} elsif (exists($cifs_user_info{$id})) {
			$num_id = $cifs_user_info{$id}{rid};
		} else {
			LogException("get_quota_target_num_id: No NFS or CIFS user found \"$id\"");
			return(undef);
		}
	} elsif ($type eq "group") {
		read_existing_nfs_groups();
		if (exists($nfs_group_info{$id})) {
			$num_id = $nfs_group_info{$id}{gid};
		} else {
			LogException("get_quota_target_num_id: No NFS group found \"$id\"");
			return(undef);
		}
	} elsif ($type eq "tree") {
		read_existing_qtrees();
		my $qtpath = get_qtree_path($id);
		if (exists($qtree_info{$id})) {
			$num_id = $qtree_info{$id}{$qtpath};
		} else {
			LogException("get_quota_target_num_id: No qtree found \"$id\"");
			return(undef);
		}
	} else {
		die("get_quota_target_num_id: Unknown type \"$type\"");
	}
	return($num_id);
}


# verify_quota_report_changes(@template_list)
# (formerly "check_quota_report_with_hints")
#
#    Given values representing "previous" and "current" quota reports, verify that
#    that the two reports differ as specified in '@template_list'. The 'current' quota
#    report is obtained via get_quota_report. A global var is used to store
#    quota report values, so the subsequent call to verify_quota_report_changes
#    will see the current call's "current quota report" as its "previous quota
#    report".
#
#    Each elt. of '@template_list' is a hash with the keys described below. This hash
#    describes a single quota report enty. The entry may exist one or both of 'prev_qr'
#   and 'curr_qr'. If multiple entries in either quota report match the hash, then
#   the first one is used. The hash values values are absolute, except int the special
#    case of the value of 'exists" being "yes_yes", the 'files_used', 'kbytes_used",
#    'files_limit', 'kbytes_limit' values represent differences in the values between
#    'prev_qr' and 'curr_qr'
#
#    The check passes if all comparisons between entries in 'previous' and 'current'
#    quota reports which are implied by the list @template_list pass.
#
#    Args:
#
#           $prev_qr:    ARRAYREF representing the "previous" quota report
#                             Format is that given by get_quota_report
#
#            $curr_qr:    ARRAYREF representing the "previous" quota report
#                             Format is that given by get_quota_report
#
#     "Entry Array" Hash Keys:
#
#             exists       =>    Possible values are:
#                                "yes_no"   ==>  Entry exists in 'prev_qr'
#                                                but not in 'curr_qr'
#
#                                "yes_yes"  ==>  Entry exists in both 'prev_qr'
#                                                and 'curr_qr'
#
#                                "no_yes"   ==>  Entry exists in 'curr_qr' but
#                                                not in 'prev_qr'
#
#                              Default is "exists_exists"
#
#             type         =>     Quota type, one of 'user', 'tree', group'
#
#             id           =>     "quota target". Equiv to "user" for user quota entries,
#                                 "group" for group quota entries, and "tree" for tree quota
#                                  entries.
#
#             tree         =>    qtree for quota. 'undef' or empty string "" means quota
#                                applies to volume
#
#             files        =>    Either absolute value of 'files_used' in quota entry or
#                                difference in this value between 'prev_qr' and 'curr_qr'
#                                (see above).
#
#             space        =>   Either absolute value of 'kbytes_used' in quota entry or
#                               difference in this value between 'prev_qr' and 'curr_qr'
#                               (see above).
#
#             files_limit  =>   Either absolute value of 'files_limit' in quota entry or
#                                difference in this value between 'prev_qr' and 'curr_qr'
#                                (see above).
#
#             space_limit  =>   Either absolute value of 'kbytes_limit' in quota entry or
#                                difference in this value between 'prev_qr' and 'curr_qr'
#                                (see above).
#
sub verify_quota_report_changes {
	my ($msg, @orig_template_list) = @_;

logDbg("Enter verify_quota_report_changes:\n\n PREV_LOG_STEP: $prev_log_step");
	logDbg("Enter verify_quota_report_changes:\n\n ARGS\n\n".Dumper(\@_));
	if (ref($msg)) {
		push(@orig_template_list, $msg);
		$msg = undef;
	}
	my $saved_dumper_maxdepth = $Data::Dumper::Maxdepth;
	$Data::Dumper::Maxdepth = 6;
	if (!$msg) {
		$msg = $prev_log_step;
	}
	if ($msg) {
		logDbg("Enter verify_quota_report_changes:\n\n Msg: $msg\n\n".Dumper(\@orig_template_list));
	} else {
		logDbg("Enter verify_quota_report_changes:\n\n".Dumper(\@orig_template_list));
	}
	if (scalar(@orig_template_list) == 0) {
		logWarn("verify_quota_report_changes: No templates given");
	}
	my $verify_quota_report_changes_verbose = 1;
	my @fail = ();
	my $check_for_duplicate_matches;
	logDbg("verify_quota_report_changes  Line: ".__LINE__."   Calling get_quota_report");
	my ($curr_qr_hash, $curr_qr_str) = get_quota_report();
	logDbg("verify_quota_report_changes  Line: ".__LINE__."   Returned from get_quota_report");

	if (!$prev_qr_hash) {
		# $prev_qr_hash = $curr_qr_hash;
		# $prev_qr_str = $curr_qr_str;
		# return;
		$prev_qr_str = "";
		$prev_qr_hash = [];
	}

	my $template_to_str = sub {
		my ($tmpx) = @_;
		my ($tmpl) = { %$tmpx };
		foreach my $k ( qw(tree space space_limit files files_limit) ) {
			$tmpl->{$k} = " -  " unless (exists($tmpl->{$k}) && $tmpl->{$k} ne "");
		}
		my $str = 
		"\nType\tId\t\tTree\t\tKBytes Limit\tKBytes Used\tFiles Used\tFiles Limit\tExists\t\tMap\n".
		"-----------------------------------------------------------------------------------".
		"-----------------------------------------------\n";
		$str .=
		    "$tmpl->{type}\t$tmpl->{id}\t\t$tmpl->{tree}\t\t$tmpl->{space}\t\t".
		    "$tmpl->{space_limit}\t\t$tmpl->{files}\t\t$tmpl->{files_limit}\t\t".
		    "$tmpl->{exists}\t\t$tmpl->{map}\n";
		return( $str );
	};

	if (ref($curr_qr_hash) !~ /ARRAY/) {
		die("verify_quota_report_changes: 'curr_qr_hash' is undefined");
	}
	if (ref($prev_qr_hash) !~ /ARRAY/) {
		die("verify_quota_report_changes: 'prev_qr_hash' is undefined");
	}

	my @template_list = ();
	my @template_str_list = ();
	foreach my $template (@orig_template_list) {
		$template = { %$template };
		push(@template_list, $template);
		my $template_str = &$template_to_str($template);
		push(@template_str_list, $template_str);
	}
	logDbg("verify_quota_report_changes  Line: ".__LINE__.
		"\n\nTemplates:\n\n".Dumper(\@template_str_list)."\n\n".Dumper(@template_list));
	logDbg("verify_quota_report_changes  Line: ".__LINE__.
 		"\n\nPrev Quota Report:\n\n$prev_qr_str\n\n".Dumper($prev_qr_hash).
		"\n\nCurr Quota Report\n\n$curr_qr_str\n\n".Dumper($curr_qr_hash));

	my %qr_hash_hash = ( prev => $prev_qr_hash, curr => $curr_qr_hash );
	my %qr_str_hash = ( prev => $prev_qr_str, curr => $curr_qr_str );
	my %qr_keymap =
		(
			# id		=> "quotaspec",
			id		=> "id",
			volume		=> "volume",
			tree		=> "tree",
			type		=> "type",
			space 		=> "kbytes_used",
			space_limit	=> "kbytes_limit",
			files 		=> "files_used",
			files_limit	=> "files_limit",
		);

	my @bare_template_list = ();
	foreach my $template (@template_list) {
		logDbg("verify_quota_report_changes  Line: ".__LINE__.
			"    Checking Template:\n\n".Dumper($template));
		# $template->{tree} = "" unless (defined($template->{tree}));
		$template->{tree} = "-" unless ($template->{tree});
		my $bare_template = { %$template };
		push(@bare_template_list, $bare_template);
		my $exists = $template->{exists};
		my $tmpl_type = $template->{type};
		if ( !defined($template->{volume}) ) {
			$template->{volume} = $QCState{testvol};
		}
		my $tmpl_id = $template->{id};
		my $tmpl_map = $template->{map};
		my $num_tmpl_id = get_quota_target_num_id($tmpl_id, $tmpl_type);
		if (!$tmpl_type) {
			LogException("verify_quota_report:  Line: ".__LINE__.
					"   Required field 'type' is required field for template:\n\n".
			     		Dumper($template));
		}
		if ( ($tmpl_type ne "tree") && !$tmpl_id ) {
			LogException("verify_quota_report:  Line: ".__LINE__.
					"   Required field 'id' is required field for template:\n\n".
			     		Dumper($template));
		}
		logDbg("verify_quota_report_changes  Line: ".__LINE__.
			"\n\nTemplate Type: \"$tmpl_type\"  Template ID: \"$tmpl_id\"  Numeric Template ID: \"$num_tmpl_id\"");
		my $template_str = &$template_to_str($template);
		# logDbg("TEMPLATE:\n\n".Dumper($bare_template)."\n\n$template_str");
		my $map = $template->{map};
		$template->{info}{map}{"prev"} = (($map eq "yes") || ($map =~ /^yes_/));
		$template->{info}{map}{"curr"} = (($map eq "yes") || ($map =~ /^_yes/));
		my $use_delta = ($exists eq "yes_yes");
		my $delta_hash = ();
		foreach my $key ( ("space", "space_limit", "files", "files_limit") ) {
			if (exists($template->{$key})) {
				$delta_hash->{$key} = $template->{$key};
			}
		}

		my @fixed_template_keys = qw( type volume tree id map );
		my %fixed_template_keys_hash = ();
		my $template_keys = [];
		foreach (@fixed_template_keys) {
			$fixed_template_keys_hash{$_} = 1;
			if (exists($template->{$_})) {
				push(@$template_keys, $_);
			}
		}
		foreach (keys(%$template)) {
			if ( !exists($fixed_template_keys_hash{$_}) ) {
				push(@$template_keys, $_);
			}
		}
		### my @template_keys = grep( {$_ ne "type"}  keys(%$template) );

		my $expect_match = {};
		$expect_match->{"prev"} = ($exists =~ /^yes_/);
		$expect_match->{"curr"} = ($exists =~ /_yes$/);
		$template->{info} = { matches => {},
				      expect_match => {},
	       			      template_str => $template_str,
	      			      bare_template => $bare_template,
				      template_keys => $template_keys,
				      delta_hash => $delta_hash,
				      use_delta => $use_delta,
				      expect_match => $expect_match,
				    };
		if ($tmpl_map && $tmpl_type ne "user") {
			LogException("verify_quota_report:  Line: ".__LINE__.
					"  Multiple entries are valid only for 'user' template type:\n\n".
			     		"Template:\n\n$template_str\n\n".Dumper($template));
		}
	}

	foreach my $template (@template_list) {
		my $bare_template = $template->{info}{bare_template};
		my $template_str = $template->{info}{template_str};
		my $use_delta = $template->{info}{use_delta};
		my $delta_hash = $template->{info}{delta_hash};
		my $template_keys = $template->{info}{template_keys};
		my $tmpl_type = $template->{type};
		my $tmpl_map = $template->{map};
		my $tmpl_id = $template->{id};
		logDbg("verify_quota_report_changes  Line: ".__LINE__.
			"Template Keys:\n\n".Dumper($template_keys));
		foreach my $qr_str ( ("prev", "curr") ) {
			my $entry_list = $qr_hash_hash{$qr_str};
			my $matched_entry;
			foreach my $entry (@$entry_list) {
				logDbg("verify_quota_report_changes  Line: ".__LINE__."   QR_STR: \"$qr_str\"".
					"   Comparing Template\n\n".Dumper($bare_template)."\n\nwith Entry:\n\n".Dumper($entry));
				my $match = 1;
				my $failed_tmplkey;
				my $failed_qrkey;
				my $failed_line;
				$entry->{tree} = "-" unless ($template->{tree});
				logDbg("verify_quota_report  Line: ".__LINE__."\nQR_STR: \"$qr_str\"\n\n".
					"Template:\n\n".Dumper($bare_template)."\n\nEntry:\n\n".Dumper($entry));
				foreach my $tmplkey (@$template_keys) {
					logDbg("verify_quota_report_changes  Line: ".__LINE__."   Key: $tmplkey");
					last unless ($match);
					next if ($tmplkey =~ /^(info|exists)/);
					next if ( $use_delta && exists($delta_hash->{$tmplkey}) );
					my $qrkey = $qr_keymap{$tmplkey} || $tmplkey;
					my $template_value;
					my $entry_value;
					if ($tmplkey eq "map") {
						next unless ($template->{map});
						if (!defined($entry->{map})) {
							LogException("verify_quota_report:  Line: ".__LINE__.
									"'map' value for entry is undefined:\n\n".Dumper($entry));
						}
						# $template_value = $template->{info}{map}{$qr_str};
						$template_value = $template->{map};
						$entry_value = $entry->{map};
						next if ($use_delta && $qr_str eq "prev");
						if ($template_value eq $entry_value) {
							logDbg("verify_quota_report_changes  Line: ".__LINE__.
								"QR_STR: \"$qr_str\"\nKey: \"map\"  Template value: ".
								"  \"$template_value\"  Matches entry value");
						} else {
							logDbg("verify_quota_report_changes  Line: ".__LINE__.
								"  QR_STR: \"$qr_str\"\nMatch Failed for  Key: \"map\"\n".
								"  Template value: \"$template_value\"  Entry value: \"$entry_value\"");
							$match = 0;
							$failed_tmplkey = $tmplkey;
							$failed_qrkey = $qrkey;
							$failed_line = __LINE__;
							last;
						}
					} else {
						$template_value = $template->{$tmplkey};
						$entry_value = $entry->{$qrkey};
					}
					if ($verify_quota_report_changes_verbose) {
						logDbg("verify_quota_report  Line: ".__LINE__."  QR_STR: \"$qr_str\"\n".
						"Key: \"$tmplkey\"\nTemplate Value: \"$template_value\"\n".
							"Entry Value: \"$entry_value\"");
					}
					if ($tmplkey eq "id") {
						my @entry_value_list;
						my @template_value_list;
						if ($entry_value eq "*") {
							@entry_value_list = ("*");
						} else {
							@entry_value_list =
							    split(/,/, $entry_value);
							for (my $i=0; $i<=$#entry_value_list; $i++) {
								$entry_value_list[$i] =~ s/^(.*\\)//;
							}
						}
						if ($template_value eq "*") {
							@template_value_list = ("*");
						} else {
							@template_value_list = 
							    split(m%[,/]%, $template_value);
							for (my $i=0; $i<=$#template_value_list; $i++) {
								$template_value_list[$i] =~ s/^(.*\\)//;
							}
						}

						my $entry_value_list_len = scalar(@entry_value_list);
						my $template_value_list_len = scalar(@template_value_list);
						my %entry_values_hash = ();
						$entry->{map} = ($entry_value_list_len > 1) ? 'yes' : 'no';
						logDbg("verify_quota_report_changes  Line: ".__LINE__."   Key: $tmplkey\n\n".
							"Entry Value: \"$entry_value\"    Entry Value List:\n\n".
							Dumper(\@entry_value_list));

						if (!$tmpl_map && ($template_value_list_len > 1)) {
							LogException("verify_quota_report:  Line: ".__LINE__.
							"  Multiple entries for template 'id' field are expected and only".
							" if 'map' is 1  Template map: \"$tmpl_map\"\n\nId List Entries:\n\n".
							Dumper(\@template_value_list));
						}

						my $idx = 0;
						foreach my $eval (@entry_value_list) {
							logDbg("verify_quota_report_changes  Line: ".__LINE__.
								"   Id Entry Value: \"$eval\"");
							$eval =~ s/.*\-(\d+)/$1/;
							$eval =~ s/^(.*\\)//;
							# my $enval = get_quota_target_num_id($eval, $tmpl_type);
							my $enval = $eval;
							$enval =~ s/.*\-(\d+)/$1/ if ($tmpl_type eq "user");
							$entry_values_hash{$enval} = 1;
							logDbg("verify_quota_report_changes  Line: ".__LINE__."\nQR_STR: \"$qr_str\"".
								"    Key: \"$tmplkey\"   Id Entry Value[$idx]: \"$eval\"".
								"  Num Id Entry: \"$enval\"");
							$idx++;
						}
						logDbg("verify_quota_report_changes  Line: ".__LINE__."\nQR_STR: \"$qr_str\"".
							"Key: \"$tmplkey\"\n\nTemplate Value: \"$template_value\"".
							"  Template Value List:\n\n".  Dumper(\@template_value_list).
							"\n\nEntry value: \"$entry_value\"  Entry value hash:\n\n".
							Dumper(\%entry_values_hash));
						foreach my $tval (@template_value_list) {
							my $tnval = get_quota_target_num_id($tval, $tmpl_type);
							logDbg("verify_quota_report_changes  Line: ".__LINE__.
								"  QR_STR: \"$qr_str\"  Template Type:  $tmpl_type\n".
								"   Key: \"$tmplkey\"".
								"   Template Val: \"$tval\"  Num Template Val: \"$tnval\"");
							if ( exists($entry_values_hash{$tnval}) ) {
								logDbg("verify_quota_report_changes  Line: ".__LINE__.
									"  QR_STR: \"$qr_str\"\nTemplate Type:  $tmpl_type".
									"   Key: \"$tmplkey\"  Template Val: \"$tval\"".
									"   Matches Entry value");
							} else {
								logDbg("verify_quota_report  Line: ".__LINE__.
									"  QR_STR: \"$qr_str\"  Match Failed on ID:\n".
									"Tval: \"$tval\"  Template ID: \"$tmpl_id\"\n\n".
									"Template ID List:\n\n".
									Dumper(\@template_value_list).
									"Entry ID: \"$entry_value\"\n\nEntry ID List:\n\n".
									Dumper(\%entry_values_hash));
								$match = 0;
								$failed_tmplkey = $tmplkey;
								$failed_qrkey = $qrkey;
								$failed_line = __LINE__;
								last;
							}
						}

						if ($verify_quota_report_changes_verbose) {
							logDbg("verify_quota_report  Line: ".__LINE__.
								"  QR_STR: \"$qr_str\"  Key: \"$tmplkey\"\n".
								"Multiple ID Values\n\nTemplate: \"$template_value\"\n".
								Dumper(\@template_value_list)."\n\n".
								"Entry: \"$entry_value\"\n".Dumper(\@entry_value_list).
								"\n\nMatch: \"$match\"");
						}
					} elsif ($tmplkey =~ /space/) {
						my $abs_diff = abs($template_value - $entry_value);
						my $rel_diff = $abs_diff / (1 + abs($template_value));
						logDbg("verify_quota_report_changes  Line: ".__LINE__."   Key: \"Space\"".
							"  QR_STR: \"$qr_str\"  Template val: \"$template_value\"  ".
							"  Entry value: \"$entry_value\"  Rel Diff: $rel_diff");
						if ($rel_diff <= .05) {
							logDbg("verify_quota_report_changes  Line: ".__LINE__."   Key: \"Space\"".
								"  QR_STR: \"$qr_str\"  Template val: \"$template_value\"  Matches".
								"  Entry value: \"$entry_value\"   Rel Diff: $rel_diff");
						} else {
							if ($verify_quota_report_changes_verbose) {
								logDbg("Line: ".__LINE__."\nQR_STR: \"$qr_str\"\nMatch ".
									"Failed on key \"$tmplkey/$qrkey\"\n\nTemplate:\n\n".
									Dumper($bare_template)."\n\nEntry:\n\n".
									Dumper($entry));
							}
							$match = 0;
							$failed_tmplkey = $tmplkey;
							$failed_qrkey = $qrkey;
							$failed_line = __LINE__;
							last;
						}
					} else {
						logDbg("verify_quota_report_changes  Line: ".__LINE__."   Key: \"$tmplkey\"".
							"  QR_STR: \"$qr_str\"\n".
							"Template val: \"$template_value\"  Entry value: \"$entry_value\"");
						if ($entry_value eq $template_value) {
							logDbg("verify_quota_report_changes  Line: ".__LINE__.
								"    QR_STR: \"$qr_str\"\nKey: \"$tmplkey\"".
								"  Template val: \"$template_value\"  ".
								"  Matches entry value");
						} else {
							if ($verify_quota_report_changes_verbose) {
								logDbg("Line: ".__LINE__."\nQR_STR: \"$qr_str\"\nMatch Failed ".
									"on key \"$tmplkey/$qrkey\"  QR_STR: $qr_str\n\n".
									"Template value: \"$template_value\"".
									"  Entry value: \"$entry_value\"\n\n".
									"Template:\n\n".Dumper($bare_template)."\n\nEntry:\n\n".
									Dumper($entry));
							}
							$match = 0;
							$failed_tmplkey = $tmplkey;
							$failed_qrkey = $qrkey;
							$failed_line = __LINE__;
							last;
						}
					}
				}
				if ($match) {
					if ($matched_entry) {
						die("verify_quota_report: Duplicate matches found for \"$qr_str\" ".
						      "quota report:\n\nTemplate:\n\n".Dumper($bare_template)."\n\n".
						      Dumper($matched_entry)."\n\n".Dumper($entry));
					} else {
						$matched_entry = $entry;
						$template->{info}{matches}{$qr_str} = $matched_entry;
						if ($verify_quota_report_changes_verbose) {
							logDbg("Line: ".__LINE__."\nQR_STR: \"$qr_str\"\n".
								"Match Found for Template:\n\n".Dumper($bare_template).
								"\n\nEntry:\n\n".Dumper($entry));
						}
					}
					last unless ($check_for_duplicate_matches);
				} else {
					logDbg("Line: ".__LINE__."\nQR_STR: \"$qr_str\"\nMatch Failed on key ".
						    "\"$failed_tmplkey/$failed_qrkey\"\n\n".
						    "Failed at Line: $failed_line\n\nTemplate:\n\n".
						    Dumper($bare_template)."\n\nEntry:\n\n".Dumper($entry));
				}
			}
		}
		foreach my $qr_str ( ("prev", "curr") ) {
			my $expect_match = $template->{info}{expect_match}{$qr_str};
			my $matched_entry = $template->{info}{matches}{$qr_str};
			if ($matched_entry && !$expect_match) {
				push(@fail, "verify_quota_report: Unexpected Match found for ".
							  "\"$qr_str\" quota report:\n\n$msg\n\nTemplate:\n\n".
							  Dumper($bare_template)."\n\n\"$qr_str\" Quota Report ".
							  "Entry:\n\n".Dumper($template->{info}{matches}{$qr_str}));
			} elsif (!$template->{info}{matches}{$qr_str} && $template->{info}{expect_match}{$qr_str}) {
				if ($verbose) {
					push(@fail, "verify_quota_report: Did not find expected match for ".
							  "\"$qr_str\" quota report:\n\n$msg\n\n".
							  "Template:\n\n".Dumper($bare_template)."\n\n\"$qr_str\" ".
							  "Quota Report:\n\n".Dumper($qr_hash_hash{$qr_str}));
				} else {
					push(@fail, "verify_quota_report: Did not find expected match for ".
							  "\"$qr_str\" quota report:\n\n$msg\n\n".
							  "Template:\n\n".Dumper($bare_template));
				}
			}
		}
		my $curr_entry;
		my $prev_entry;
		if ( $use_delta &&
			exists($template->{info}{matches}{"prev"}) &&
			exists($template->{info}{matches}{"curr"}) ) 
		{
			if ( ($template->{files} < 0 && $template->{space} > 0) ||
			     ($template->{files} > 0 && $template->{space} < 0) ) {
				logWarn("verify_quota_report: Signs for 'space' and 'files' ".
					"delta disagreement for template:\n\n".
					Dumper($template)."\n\n$msg");
			}
			$prev_entry = $template->{info}{matches}{"prev"};
			$curr_entry = $template->{info}{matches}{"curr"};
			my $delta_err = "";
			while ( my ($tmplkey, $expected_delta) = each(%$delta_hash) ) {
				next unless ($expected_delta =~ /\d+/);
				my $qrkey = $qr_keymap{$tmplkey};
				my $curr_value = $curr_entry->{$qrkey};
				my $prev_value = $prev_entry->{$qrkey};
				my $actual_delta = ($curr_value - $prev_value);
				if ($verify_quota_report_changes_verbose) {
					logDbg("Line: ".__LINE__."\nChecking 'delta' for Key: \"$tmplkey\"\n".
						"Curr Value: \"$curr_value\"\nPrev Value: \"$prev_value\"\n".
						"Actual Delta: \"$actual_delta\"\nExpected Delta: \"$expected_delta\"");
				}
				if ($tmplkey =~ /space/) {
					my $abs_diff = abs($actual_delta - $expected_delta);
					my $rel_diff = $abs_diff / (1 + abs($expected_delta));
					if ($rel_diff > .05) {
						$delta_err .= "Key: \"$tmplkey\"\nExpected Delta: \"$expected_delta\"\n".
							      "Actual Delta: \"$actual_delta\"\n".
							      "Relative difference: ".(100 * $rel_diff)."\%\n".
							      "Prev value: \"$prev_value\"\nCurr value: ".
							      "\"$curr_value\"\n\n";
					}
				} else {
					if (($curr_value - $prev_value) != $expected_delta) {
						$delta_err .= "Key: \"$tmplkey\"\nExpected Delta: \"$expected_delta\"\n".
							      "Actual Delta: \"$actual_delta\"\n".
							      "Prev value: \"$prev_value\"\nCurr value: ".
							      "\"$curr_value\"\n\n";
					}
				}
			}
			if ($delta_err) {
				push(@fail, "verify_quota_report_changes: Did not see expected 'delta' for\n".
					    "current vs. previous values:\n\n$msg\n\nFailed for template:\n\n".
					    Dumper($bare_template)."\n\n$delta_err");
				if ($verify_quota_report_changes_verbose) {
					logDbg("Line: ".__LINE__."\n\nDelta Error:\n\n$delta_err");
				}
			}
		}
	}
	my $fail = scalar(@fail);
	if ($fail) {
		# LogException("Exit verify_quota_report_changes for\n\n$msg\n\nTemplates:\n\n".
		#	 Dumper(\@template_list)."\n\nPrev Quota Report:\n\n$prev_qr_str\n\n".
		#	"Curr Quota Report\n\n$curr_qr_str\n\n\n".join("\n", @fail));
		LogException("Exit verify_quota_report_changes for\n\n$msg\n\n".
			 "Prev Quota Report:\n\n$prev_qr_str\n\n".
			"Curr Quota Report\n\n$curr_qr_str\n\n\n".join("\n", @fail));
	}
	$Data::Dumper::Maxdepth = $saved_dumper_maxdepth;
	$prev_qr_hash = $curr_qr_hash;
	$prev_qr_str = $curr_qr_str;
	# $@ = $msg;
	# eval { die($msg) };
	return($fail);
}



##########################################
#
#    'grant access' routines
#
#     Functions to perform 'grant access' commands on filer for CIFS shares
#
##########################################


sub grant_administratator_access_on_cifs_client {
	my ($cifs_path) = @_;
	grant_user_access_on_cifs_client($cifs_path, $QCState{CIFS_USER});
	grant_user_access_on_cifs_client($cifs_path, "administrator");
	grant_user_access_on_cifs_client($cifs_path, "BUILTIN\\Administrators");
	grant_user_access_on_cifs_client($cifs_path, "Everyone");
	die("grant_administrator_access_on_client: $@") if ($@);
}

sub cifs_user_match {
	my ($orig_user1, $orig_user2) = @_;
	my $user1 = $orig_user1;
	my $user2 = $orig_user2;
	$user1 =~ s%.*\\%%;
	$user2 =~ s%.*\\%%;
	my $match = ($user1 =~ /^$user2$/i);
	logDbg("cifs_user_match   \"$user1\"  \"$user2\"  Returning: $match");
	return($match);
}

sub get_cifs_user_fullname {
	my ($cifs_user) = @_;
	if ($cifs_user =~ /\\/) {
		return($cifs_user);
	} else {
		my $cifs_server = $QCState{cifs_server};
		return( $cifs_server . "\\" . $cifs_user );
	}
}


sub grant_user_access_on_cifs_client {
	my ($cifs_path, $cifs_user) = @_;
	$cifs_user = get_cifs_user_fullname($cifs_user)
	    unless ($cifs_user =~ /^every/i);
	my $cifs_client = $QCState{cifs_client};
        my $cifsconn = make_conn($cifs_client, "#rsh");
        my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd = "x:\\bin\\subinacl.exe /file $cifs_path /grant=$cifs_user=F /sgrant=$cifs_user=F";
        ### my $cifs_cmmd = "cmd.exe /c cacls $cifs_path /E /G $cifs_user:F";
###     my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
        my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
        chomp($output);
        $output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
        $output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	logDbg("grant_user_access_on_cifs_client:  OUTPUT:\n\n$output");
        return(1);
}

# grant_cifs_user_share_access($cifs_user)
sub grant_cifs_user_share_access {
	my ($cifs_user) = @_;
	### my $rights = "rwx";
	my $rights = "Full Control";
	my $cifs_server = $QCState{cifs_server};
logDbg("grant_share_access_to_cifs_user:\nCIFS_USER: \"$cifs_user\"\nCIFS_SERVER: \"$cifs_server\"\nCIFS_SHARE: \"$QCState{cifs_share}\"");
	if ($cifs_server && $cifs_user !~ m%\\%) {
		$cifs_user = "$cifs_server\\$cifs_user";
	}
	filer_cifs_access($QCState{fc},
			  share => $QCState{cifs_share}, 
			  user => $cifs_user,
			  rights => $rights);
	if ($@) {
		logWarn("grant_share_access_to_cifs_user: filer_cifs_access_error: $@");
	}
}

# grant_unix_user_share_access($cifs_user)
sub grant_unix_user_share_access {
	my ($unix_user) = @_;
	### my $rights = "rwx";
	my $rights = "Full Control";
logDbg("grant_share_access_to_unix_user:\nUNIX_USER: \"$unix_user\"");
	filer_cifs_access($QCState{fc},
			  share => $QCState{cifs_share}, 
			  user => $unix_user,
			  rights => $rights);
	if ($@) {
		logWarn("grant_share_access_to_unix_user: filer_cifs_access_error: $@");
	}
}

# grant_win_machine_share_access
sub grant_win_machine_share_access {
	filer_cifs_access($QCState{fc},
			  share => $QCState{cifs_share}, 
			  win_machine => "yes");
	if ($@) {
		logWarn("grant_share_access_to_win_machine: filer_cifs_access_error: $@");
	}
}

# grant_all_cifs_users_share_access
sub grant_all_cifs_users_share_access {
	my @users_hash_list = listCIFSUsers($QCState{fc});
	my @users = map( { $_->{Name} }  @users_hash_list );
	### my $rights = "rwx";
	my $rights = "Full Control";
logDbg("grant_all_cifs_users_share_access  ALL_CIFS_USERS:\n\n@users");
	foreach my $user (@users) {
		grant_cifs_user_share_access($user);
	}
}

# grant_all_unix_users_share_access
sub grant_all_unix_users_share_access {
	my $users_hash = readPasswdFile($QCState{fc});
	my @users = keys(%$users_hash);
	foreach my $user (@users) {
		grant_unix_user_share_access($user);
	}
}

# grant_everyone_share_access()
sub grant_everyone_share_access {
	### my $rights = "rwx";
	my $rights = "Full Control";
	my $fc = $QCState{fc};
logDbg("Enter grant_everyone_share_access");
	filer_cifs_access($QCState{fc},
			  share => $QCState{cifs_share}, 
			  user => "everyone",
			  group => "yes",
			  rights => $rights);
	if ($@) {
		logWarn("grant_share_access_to_everyone: filer_cifs_access_error: $@");
	}
}



##########################################
#
#    CIFS support routines
#
##########################################

# get_fsutil_path( $cifs_client )
#
sub get_fsutil_path {
	my ($cifs_client) = @_;
	my $default_windir = "C:\\WINDOWS\\";
	my $windir = $default_windir;
        my $cifsconn = make_conn($cifs_client, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd = "echo %windir%";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	if ($output) {
		$windir = $output;
	}
	my $path = "$windir\\system32\\fsutil";
	$cifs_cmmd = "attrib ${path}.exe";
	$output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	if ($output =~ /file not found/i) {
		die("get_fsutil_path: Could not find fsutil: ${path}.exe");
	}
	logDbg("QCov_Ops::get_fsutil_path  CIFS_Client: \"$cifs_client\"     Returning: \"$path\"");
	return($path);
}


# chownCIFSFile($hostspec, $userid, $winpath, %opts)
#
sub chownCIFSFile {
	my ($hostspec, $userid, $winpath, %opts) = @_;
	my $recursive = $opts{recursive} || 0;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	if ($recursive) {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /file $winpath /setowner\=$userid";
	} else {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /subdirectories $winpath /setowner\=$userid";
	}
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output !~ /is the new owner/) {
		 croak("chownCIFSFile:  Could not change owner of\n".
			 "file \"$winpath\" to \"$userid\":\n\n$output");
	}
	return(1);
}


# chgrpCIFSFile($hostspec, $grpid, $winpath, %opts)
#
sub chgrpCIFSFile {
	my ($hostspec, $grpid, $winpath, %opts) = @_;
    	my $recursive = $opts{recursive} || 0;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	if ($recursive) {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /file $winpath /setprimarygroup\=$grpid";
	} else {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /subdirectories $winpath /setowner\=$grpid";
	}
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output !~ /is the new Primary Group/) {
		 croak("chgrpCIFSFile:  Could not change group of\n".
			 "file \"$winpath\" to \"$grpid\":\n\n$output");
	}
	return(1);
}


# dos_list_drives($hostspec)
#  -- uses fsutil
#
sub dos_list_drives {
	my ($hostspec) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "$fsutil_path fsinfo drives";
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output !~ /Drives(:)?\s*(.*)\s*$/) {
		 croak("dos_list_drives:  'fsutil' command failed:\n$output");
	}
	my $drvs = $2;
	$drvs =~ s/[^A-Za-z]//g;
	$drvs =~ s/Drives:\s*//;
	my @drives = split(//, $drvs);
	return(@drives);
}


# dos_create_file($hostspec, $winpath, $size)
#  -- uses fsutil
#
sub dos_create_file {
	my ($hostspec, $winpath, $size) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "$fsutil_path file createnew $winpath $size";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output !~ /is created/) {
		 croak("dos_create_file:  Could not create file \"$winpath\"");
	}
	return(1);
}


# dos_mkdir($hostspec, $winpath)
#
sub dos_mkdir {
	my ($hostspec, $winpath) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "mkdir $winpath";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output) {
		 croak("dos_mkdir:  Could not create directory \"$winpath\"");
	}
	return(1);
}


# dos_mkdir($hostspec, $winpath)
#
sub dos_file_exists {
	my ($hostspec, $winpath) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "attrib $winpath";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s/\s+/ /g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output =~ /file not found/i) {
		return(0);
	} else {
		return(1);
	}
}


#  dos_move_file($hostspec, $srcpath, $destpath);
#
sub dos_move_file {
	my ($hostspec, $srcpath, $destpath) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "move $srcpath $destpath";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output && $output !~ /file\(s\) moved/i) {
		 croak("dos_move_file:  Could not move \"$srcpath\" to \$destpath\":\n$output");
	}
	return(1);
}


#  dos_rmdir($hostspec, $winpath, %opts)
#
sub dos_rmdir {
	my ($hostspec, $winpath, %opts) = @_;
    	my $recursive = $opts{recursive} || 0;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	if ($recursive) {
		$cifs_cmmd = "rmdir /S /Q $winpath";
	} else {
		$cifs_cmmd = "rmdir $winpath";
	}
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output) {
		 croak("dos_move_file:  Could not rmdir \"winpath\"");
	}
	return(1);
}


#  dos_rmfile($hostspec, $winpath, %opts)
#
sub dos_rmfile {
	my ($hostspec, $winpath, %opts) = @_;
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	$cifs_cmmd = "del $winpath";
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s/^\s*//g;
	$output =~ s/\s*$//g;
	$output =~ s%[^:\w\ \t=\\\/\-\\._]%%g;
	if ($output) {
		 croak("dos_move_file:  Could not rmdir \"winpath\"");
	}
	return(1);
}


##########################################
#
#    Miscellaneous support routines
#
##########################################


my $_change_dir_saved_dir;
my $_change_dir_caller;

# change_dir($new_dir, $caller)
#
sub change_dir {
	my ($new_dir, $caller) = @_;
	my $old_dir = cwd();
	$_change_dir_saved_dir = $old_dir;
	$_change_dir_caller = $caller;
	$caller = "change_dir" unless ($caller);
	chdir($new_dir)
		|| LogException("$caller: chdir \"$new_dir\": $!");
	return($old_dir);
}

# restore_dir()
#
sub restore_dir {
	my $caller = $_change_dir_caller || "restore_dir";
	chdir($_change_dir_saved_dir)
		|| LogException("$caller: chdir \"$_change_dir_saved_dir\": $!");
}


# DiagExit()
#
sub DiagExit {
	my ($msg) = @_;
	$main::CLENUP = "none";
	die("Exiting without cleanup: $msg");
}


my $_severe_problemo_die_on_error = 1;
my $_problemo_die_on_error = 0;
my $_problemo_prev_die_on_error = undef;

# get_die_on_error_value()
sub get_die_on_error_value { return( $_problemo_die_on_error ); }

# set_die_on_error($newval)
#
sub set_die_on_error {
	my ($newval) = @_;
	$_problemo_prev_die_on_error = $_problemo_die_on_error;
	$_problemo_die_on_error = $newval;
}

# restore_die_on_error()
#
sub restore_die_on_error {
	$_problemo_die_on_error = $_problemo_prev_die_on_error;
	$_problemo_prev_die_on_error = undef;
}


# LogException($msg, %opt)
#
sub LogException {
	my ($msg, %opt) = @_;
	logcomment("Enter LogException");
	my $action = $opt{action};
	my $die_on_error = $opt{die_on_error};
	my $interactive;
	if (defined($action)) {
		$die_on_error = $action eq "die";
	} elsif (defined($die_on_error)) {
		$die_on_error = 1;
		$action = "die";
	} elsif ($QCState{interactive}) {
		$interactive = 1;
		$die_on_error = 0;
		$action = "query";
	} elsif ($_problemo_die_on_error) {
		$die_on_error = 1;
		$action = "die";
	}
	### action = "warn" unless (defined($action));
	$action = "notify" unless (defined($action));
	if ($interactive) {
		logresult('WARN', "$msg\n\nContinue: [y|n]?");
		my $resp = readline(STDIN);
		chomp($resp);
		$resp =~ s%\W%%;
		if ($resp eq "y") {
			logcomment("Continuing...");
			return;
		} elsif ($resp eq "n") {
			die($msg);
		}
	}
	if ($action eq "die") {
		die($msg);
	} elsif ($action eq "warn") {
		logresult('WARN', $msg);
	} elsif ($action eq "notify") {
		logresult('INFO', $msg);
	}
}

# LogSevereException($msg, %opt)
#
sub LogSevereException {
	my ($msg, %opt) = @_;
	my $die_on_error = ($opt{die_on_error}) ? $opt{die_on_error} : $_severe_problemo_die_on_error;
	LogException($msg, die_on_error => $die_on_error);
}


my $_prev_map_user;

# map_cifs_share_for_user($cifs_user)
#
sub map_cifs_share_for_user {
	my ($cifs_user) = @_;
	$cifs_user =~ s%^.*\\%%;
	my $map_exists;
	my $drive;
	return if ($cifs_user eq $_prev_map_user);
	logDbg("Enter map_cifs_share_for_user: \"$cifs_user\"");
	if ($QCState{mapped_drive}) {
		$map_exists = map_exists( drive => $QCState{mapped_drive},
					     client => $QCState{cifs_client} );
		$drive = $QCState{mapped_drive};
	} else {
		$drive = $::CIFS_DRIVE;
	}
	if ($map_exists) {
		unmap_cifs_share(  drive => $QCState{mapped_drive},
				   client => $QCState{cifs_client},
				   force => 1);
	}

	my $mapped_drive = map_cifs_share(
					   server    => $QCState{connhost},
					   client    => $QCState{cifs_client},
					   drive     => $drive,
					   share     => $QCState{cifs_share},
				      workgroup    => "WORKGROUP",					   
					   password  => $QCState{cifs_passwd},
					   username  => $cifs_user)
	    or LogSevereException("map_cifs_share_for_user: Error from map_cifs_share  DRIVE: \"$QCState{mapped_drive}\"  USER: \"$cifs_user\"");
	# if ($mapped_drive ne $QCState{mapped_drive}) {
	#	LogException("map_cifs_share_for_user: Unexpected value returned from map_cifs_share:\n".
	#		   "Expected \"$QCState{mapped_drive}\"  Got: \"$mapped_drive\"");
	# }
	$_prev_map_user = $cifs_user;
	return(1);
}


# temp_set_qtree_security($qt, $new_sec)
#
sub temp_set_qtree_security {
	my ($qt, $new_sec) = @_;
	my $qtpath = get_qtree_path($qt);
	my $qtname = get_qtree_path($qt);
	my $old_sec = get_qtree_security($qtpath);
	return if ($old_sec eq $new_sec);
	$_prev_qtree_security{$qtpath} = $old_sec;
	change_qtree_security($qtpath, $new_sec);
}

# restore_qtree_security($qt)
#
sub restore_qtree_security {
	my ($qt) = @_;
	my $qtpath = get_qtree_path($qt);
	my $qtname = get_qtree_path($qt);
	my $sec = $_prev_qtree_security{$qtpath};
	return unless ($sec);
	change_qtree_security($qtpath, $sec);
	delete($_prev_qtree_security{$qtpath});
}


#  -- Functions to allow access of files created thru CIFS via NFS --

my %_de_cifs_path_info = ();

sub de_cifs_path {
	my ($fpath) = @_;
	remove_cifs_share();
	my $parts = get_filer_path_parts($fpath);
logDbg("de_cifs_path Line: ".__LINE__."\nFPATH: \"$fpath\"\nPARTS:\n\n".Dumper($parts));
	my $qtree = $parts->{qtree};
	read_existing_qtrees();
	my $qt_security = get_qtree_security($qtree);
	change_qtree_security($qtree, "unix");
	%_de_cifs_path_info = ( path => $fpath, qtree => $qtree, qt_security => $qt_security );
}

sub re_cifs_path {
	my $fpath = $_de_cifs_path_info{path};
	my $qtree = $_de_cifs_path_info{path};
	my $qt_security = $_de_cifs_path_info{qt_security};
	change_qtree_security($qtree, $qt_security);
	%_de_cifs_path_info = ();
	restore_cifs_share();
}

# remove_cifs_share()
#
sub remove_cifs_share {
	unmap_cifs_share(  drive => $QCState{mapped_drive},
			   client => $QCState{cifs_client},
			   force => 1);
	if ($@) {
		LogException("remove_cifs_share: Error from unmap_cifs_share: $@");
	}
}

# restore_cifs_share()
#
sub restore_cifs_share {
	my $drive = map_cifs_share(
				     server => $QCState{connhost},
				     client => $QCState{cifs_client},
				     drive => $QCState{mapped_drive},
				     share => $QCState{cifs_share},
				     password => $QCState{cifs_passwd},
				     username => $QCState{cifs_user}
		);
	if (!$drive) {
	    LogException("restore_cifs_share: Error from map_cifs_share: $@");
	}
}


# delete_cifs_file_via_nfs($fpath, %opts)
#
sub delete_cifs_file_via_nfs {
	my ($fpath, %opts) =@_;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	logDbg("delete_cifs_file_via_nfs  FPATH: \"$fpath\"  RECURSIVE: \"$recursive\"");
	de_cifs_path($fpath);
	%nfs_file_info = ();
	read_existing_nfs_files();
	delete_nfs_file_or_dir($fpath, %opts);
	re_cifs_path();
	%cifs_file_info = ();
	%nfs_file_info = ();
	read_existing_cifs_files();
	read_existing_nfs_files();
}


sub show_qcstate {
	my ($msg) = @_;
	logDbg("$msg\n\nQCSTATE(1): \n\n".Dumper(\%QCState));
}




##########################################
#
#    CIFS client routines
#
#    Functions which create/access/modify/destroy files via CIFS client
#
##########################################


#sub run_perl_cmmd_on_cifs_client() {
#	my ($perl_cmmd) = @_;
#	my $cifs_cmmd = "x:\\perl\\bin\\perl.exe -e \"". quotameta($perl_cmmd) ."\"";
#	my (undef, undef, $output) = say($cifsconn, $cifs_cmmd);
#	return($output);
#}

# run_perl_cmmd_on_cifs_client($perl_cmmd, @args)
#
sub run_perl_cmmd_on_cifs_client {
	my ($perl_cmmd, @args) = @_;

	my $sub = subtest( $perl_cmmd, "-client", $main::CIFS_CLIENT, "--", @args);
	$sub->wait_finish();
	my @results = $sub->results();
	if (!scalar(@results)) {
		return("", 0);
	}
	my $worst = NATE::Result::worst(@results);
	if (!defined($worst)) {
		return("", 0);
	}
	my $severity = $worst->severity();
	if ($severity > 1) {
		logresult($worst);
		LogException("run_perl_cmmd_on_cifs_client: Error returned from subtest");
	} else {
		logresult($worst);
	}
	my $last = pop(@results);
	if (!$last) {
		return("", 0);
	}
	my @messages = $last->messages();
	my $output = join("\n", @messages);
	return($output, $severity);
}


# create_file_on_cifs_client($cifs_path, $fsize_or_type, %opts)
#
sub create_file_on_cifs_client {
	my ($cifs_path, $fsize_or_type, %opts) = @_;
	my $fsize;
	my $ftype;
	if ($fsize_or_type =~ /^\d+$/) {
		$ftype = 'file';
		$fsize = $fsize_or_type;
	} elsif ($fsize_or_type =~ /^(file|dir)$/) {
		$fsize = 0;
		$ftype = $fsize_or_type;
	}
	my $verify = $opts{verify};
	logDbg("create_file_on_cifs_client:\n\nCIFS_PATH: \"$cifs_path\"\nFSIZE: \"$fsize\"\nFTYPE: \"$ftype\"");
	my $perl_cmmd  =<<'__EOF__';
	{
		use IO::File;
		use Tharn;
		use strict;
		use Cwd;
		my ($cifs_path, $ftype, $fsize, $verbose, $die_on_error) = @_;
		$fsize = 0 unless ($fsize);
		logcomment("Enter create_file_on_cifs_client subtest:  CIFS_PATH: \"$cifs_path\"   fsize: $fsize  ftype: \"$ftype\"")
		    if ($verbose);

		sub LogException {
			my ($msg) = @_;
			if ($die_on_error) {
				die($msg);
			} else {
				logresult('WARN', $msg);
			}
		}

		sub cifs_mkfile {
			my ($cifs_path, $fsize) = @_;
			logcomment("Enter cifs_mkfile  CIFS_PATH: \"$cifs_path\"   FSIZE: \"$fsize\"")
			    if ($verbose);
			my $BUFF = "0123456789abcdef" x 256;   # 4k
			my $BUFFSIZE = length($BUFF);
			my $fh = new IO::File(">$cifs_path");
			if (!$fh) {
				LogException("create_file_on_cifs_client: File::IO::new() failed: $!");
			}
			my $rem_len = $fsize;
			my $len_done = 0;
			while ($rem_len) {
				my $towrite = ($rem_len > $BUFFSIZE) ? $BUFFSIZE : $rem_len;
				my $offset = $len_done % $BUFFSIZE;
				my $wlen = syswrite($fh, $BUFF, $towrite, $offset);
				if (!defined($wlen)) {
					LogException("create_file_on_cifs_client: syswrite failed: $!");
				}
				if ($wlen == 0) {
					logresult('WARN', "no data written by syswrite");
				}
				$rem_len -= $wlen;
				$len_done += $wlen;
			}
			$fh->close();
		}

		if ($ftype eq 'file') {
			logcomment("calling cifs_mkfile")
			    if ($verbose);
			cifs_mkfile($cifs_path, $fsize);
			if (! -f $cifs_path) {
				LogException("create_file_on_cifs_client: Could not create file \"$cifs_path\"");
			}
		} elsif ($ftype eq 'dir') {
			logcomment("Calling mkdir")
			    if ($verbose);
			mkdir($cifs_path);
			if (! -d $cifs_path) {
				LogException("create_file_on_cifs_client: Could not create directory \"$cifs_path\"");
			}
		}
		logcomment("Exit create_file_on_cifs_client:\n\nCIFS_PATH: \"$cifs_path\"")
		    if ($verbose);
		logresult('INFO', "Exit create_file_on_cifs_client:\n\nCIFS_PATH: \"$cifs_path\"");
	}
__EOF__

	my @args_list = ($cifs_path, $ftype, $fsize, $verbose, get_die_on_error_value());
	my $output = run_perl_cmmd_on_cifs_client($perl_cmmd, @args_list);
	logcomment("Exit create_file_on_cifs_client\n\nOUTPUT:\n\n$output")
	   if ($verbose);

	if ($verify) {
		my $finfo = cifs_file_stat($cifs_path);
		logcomment("Exit create_file_on_cifs_client\n\nSTAT:\n\n".Dumper($finfo))
		    if ($verbose);
		if (!defined($finfo)) {
			LogSevereException("create_file_on_cifs_client: Could not create file \"$cifs_path\"");
		}
		# if ( ($cifs_user ne "") && !cifs_user_match($cifs_user, $finfo->{owner}) ) {
		#	LogSevereException("create_file_on_cifs_client: Could not chown  file \"$cifs_path\" to \"$cifs_user\"");
		# }
	}

	# logDbg("create_file_on_cifs_client  Calling grant_administrator_access_on_cifs_client");
	# grant_administratator_access_on_cifs_client($cifs_path);

	return(1);
}


# move_file_on_cifs_client($src_cifs_path, $dest_cifs_path, %opts)
#
sub move_file_on_cifs_client {
	my ($src_cifs_path, $dest_cifs_path, %opts) = @_;
	my $verify = $opts{verify};

	logDbg("move_file_on_cifs_client:\n\nSRC_CIFS_PATH: \"$src_cifs_path\"\nDEST_CIFS_PATH: \"$dest_cifs_path\"")
	    if ($verbose);
	my $perl_cmmd  =<<'__EOF__';
	{
		use IO::File;
		use Tharn;
		use strict;
		use Cwd;
		my ($src_cifs_path, $dest_cifs_path, $verbose, $die_on_error) = @_;

		sub LogException {
			my ($msg) = @_;
			if ($die_on_error) {
				die($msg);
			} else {
				logresult('WARN', $msg);
			}
		}

	       logcomment("Enter move_file_on_cifs_client subtest:\nSRC_CIFS_PATH: \"$src_cifs_path\"\n".
			     "DEST_CIFS_PATH: \"$dest_cifs_path\"")
		    if ($verbose);
		rename($src_cifs_path, $dest_cifs_path)
			|| LogException("move_file_on_cifs_client: move \"$src_cifs_path\" \"$dest_cifs_path\": $!");

		if ( -e $src_cifs_path || ! -e $dest_cifs_path) {
			LogException("move_file_on_cifs_client: Could not move \"$src_cifs_path\" to \"$dest_cifs_path\"");
		}

	}
__EOF__

	my @args_list = ($src_cifs_path, $dest_cifs_path, $verbose, get_die_on_error_value());
	my $output = run_perl_cmmd_on_cifs_client($perl_cmmd, @args_list);
	logcomment("move_file_on_cifs_client\n\nOUTPUT:\n\n$output")
	    if ($verbose);
	if ($verify) {
		my $src_finfo = cifs_file_stat($src_cifs_path);
		my $dest_finfo = cifs_file_stat($dest_cifs_path);
		logcomment("Exit move_file_on_cifs_client\n\nSRC STAT:\n\n".Dumper($src_finfo).
			   "\n\nDEST STAT:\n\n".Dumper($dest_finfo))
		    if ($verbose);
		if (defined($src_finfo)) {
			LogSevereException("move_file_on_cifs_client: Could not move \"$src_cifs_path\" to \"$dest_cifs_path\"");
		}
		if (!defined($dest_finfo)) {
			LogSevereException("move_file_on_cifs_client: Could not move \"$src_cifs_path\" to \"$dest_cifs_path\"");
		}
	}
	return(1);
}


# delete_file_or_dir_on_cifs_client($cifs_path, %opts)
#
sub delete_file_or_dir_on_cifs_client {
	my ($cifs_path, %opts) = @_;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 0;
	my $verify = $opts{verify};
	logDbg("Enter delete_file_on_cifs_client:  CIFS_PATH: \"$cifs_path\"   RECURSIVE: \"$recursive\"");
	$Data::Dumper::Maxdepth = 4;

	my $delete_file_or_dir_on_cifs_client_code_block =<<'__EOF__';
	{
		use File::Find;
		use File::Basename;
		use Data::Dumper;
		use Cwd;
		use Tharn;
		use strict;
		my ($cifs_path, $recursive, $mapped_drive, $testvol_filer_path, $verbose, $die_on_error) = @_;

		sub LogException {
			my ($msg) = @_;
			if ($die_on_error) {
				die($msg);
			} else {
				logresult('WARN', $msg);
			}
		}

		sub cvt_cifs_path_to_filer {
			my ($cifs_path) = @_;
			my $filer_path = $cifs_path;
			$filer_path =~ s%\\%/%g;
			if ($filer_path =~ s%^$mapped_drive(\\|/)?%$testvol_filer_path/%) {
				$filer_path =~ s%/$%%;
				return($filer_path);
			} else {
				LogException("cvt_cifs_path_to_filer: unexpected format for cifs_path: $cifs_path");
			}
		}


		logcomment("Enter delete_file_or_dir_on_cifs_client subtest  CIFS_PATH: \"$cifs_path\"  RECURSIVE: \"$recursive\"")
		    if ($verbose);
		if (! -e $cifs_path) {
			LogException("delete_file_or_dir_on_cifs_client subtest:  cifs_path \"$cifs_path\" not found");
			return;
		}
		$cifs_path =~ s%/%\\%g;
		my $dirtree_hash = {};
		my $pathname = $cifs_path;
		my $basename = File::Basename::basename($cifs_path);
		my $dirname = File::Basename::dirname($cifs_path);
		my $filer_pathname = cvt_cifs_path_to_filer($pathname);
		$dirtree_hash->{ROOT} = $filer_pathname;

		my $_change_dir_saved_dir;
		my $_change_dir_caller;
		sub change_dir {
			my ($new_dir, $caller) = @_;
			my $old_dir = cwd();
			$_change_dir_saved_dir = $old_dir;
			$_change_dir_caller = $caller;
			$caller = "change_dir" unless ($caller);
			chdir($new_dir)
				|| LogException("$caller: chdir \"$new_dir\": $!");
			return($old_dir);
		}

		sub restore_dir {
			my $caller = $_change_dir_caller || "restore_dir";
			chdir($_change_dir_saved_dir)
				|| LogException("$caller: chdir \"$_change_dir_saved_dir\": $!");
		}


		my $cifs_rm_fun = sub {
			my ($pathname, $dirname, $basename) = @_;
			$dirname=~ s%/%\\%g;
			$pathname=~ s%/%\\%g;
			if (! -e $pathname) {
				LogException("cifs_rm_fun  Pathname \"$pathname\" not found");
				return;
			}
			my $filer_pathname = cvt_cifs_path_to_filer($pathname);

			logcomment("Enter cifs_rm_fun\nPathname: \"$pathname\"\nBasename: \"$basename\"\n".
				      "Filer Pathname: \"$filer_pathname\"")
			    if ($verbose);
			if (-d $basename) {
				logcomment("cifs_rm_fun\nFile \"$pathname\" is a directory, calling 'rmdir'")
				    if ($verbose);
				### rmdir($pathname) || LogException("cifs_rm_fun: \"$pathname\" rmdir: $!");
				### eval { rmdir($pathname) };
				eval { rmdir($basename) };
				if ($!) {
					logresult('WARN', "cifs_rm_fun: rmdir \"$pathname\": $!");
				}
			} else {
				logcomment("cifs_rm_fun\nFile \"$pathname\" is a file, calling 'unlink'")
				    if ($verbose);
				### unlink($pathname) || LogException("cifs_rm_fun: unlink \"$pathname\": $!");
				unlink($basename)
					|| LogException("cifs_rm_fun: unlink \"$pathname\": $!");
			}
			if (-e $basename) {
				### LogException("cifs_rm_fun: Could not delete file \"$pathname\" (\"$filer_pathname\")");
				logresult('WARN', "cifs_rm_fun: Could not delete file \"$pathname\" (\"$filer_pathname\")");
			}
			logcomment("Exit cifs_rm_fun")
			    if ($verbose);
		};

		my $cifs_rm_find_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				### my $fhandle = _;
				$dirname=~ s%/%\\%g;
				$pathname=~ s%/%\\%g;
				if (! -e $pathname) {
					LogException("cifs_rm_find_fun  Pathname \"$pathname\" not found");
					return;
				}
				return( &$cifs_rm_fun($pathname, $dirname, $basename) );
		};

		if ($recursive) {
			finddepth($cifs_rm_find_fun, $cifs_path);
		 	rmdir($cifs_path) 
				|| LogException("delete_file_or_dir_on_cifs_client subtest: rmdir \"cifs_path\": $!");
		} else {
			change_dir($dirname, "delete_cifs_file");
			&$cifs_rm_fun($pathname, $dirname, $basename, 1);
			restore_dir();
		}
		logcomment("Exit delete_file_or_dir_on_cifs_client subtest  CIFS_PATH: \"$cifs_path\"  RECURSIVE: \"$recursive\"")
		    if ($verbose);
	}
__EOF__

	my @args_list = ($cifs_path,
			  $recursive,
			  $QCState{mapped_drive},
			  $QCState{testvol_filer_path},
			  $verbose,
			  get_die_on_error_value());
	my ($output, $severity) = run_perl_cmmd_on_cifs_client($delete_file_or_dir_on_cifs_client_code_block, @args_list);

	logDbg("Exit delete_file_on_cifs_client:  CIFS_PATH: \"$cifs_path\"   RECURSIVE: \"$recursive\"\n\n$output");
	if ($output) {
		LogException("delete_file_on_cifs_client \"$cifs_path\" FAILED");
	}
	logcomment("Exit delete_file_on_cifs_client\n\nOUTPUT:\n\n$output")
	    if ($verbose);
	logDbg("Exit delete_file_on_cifs_client:  CIFS_PATH: \"$cifs_path\"   RECURSIVE: \"$recursive\"");
	if ($verify) {
		my $finfo = cifs_file_stat($cifs_path);
		logcomment("Exit delete_file_on_cifs_client\n\nSTAT:\n\n".Dumper($finfo))
		    if ($verbose);
		if (defined($finfo)) {
			LogSevereException("delete_file_on_cifs_client: Could not delete \"$cifs_path\"");
		}
	}
	return(1);
}


# cifs_file_or_dirtree_stat($cifs_path, %opts)
#
sub cifs_file_or_dirtree_stat {
	my ($cifs_path, %opts) = @_;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 0;
	logDbg("Enter cifs_file_or_dirtree_stat  CIFS_PATH: \"$cifs_path\"   RECURSIVE: \"$recursive\"");
	$Data::Dumper::Maxdepth = 4;

	my $stat_cifs_file_or_dirtree_code_block =<<'__EOF__';
	{
		use File::Find;
		use File::Basename;
		use Cwd;
		use Data::Dumper;
		use Tharn;
		use strict;
		my ($cifs_path, $recursive, $mapped_drive, $testvol_filer_path, $verbose, $die_on_error) = @_;
		logcomment("Enter cifs_file_or_dirtree subtest  CIFS_PATH: \"$cifs_path\"  RECURSIVE: \"$recursive\"")
		    if ($verbose);

		sub LogException {
			my ($msg) = @_;
			if ($die_on_error) {
				die($msg);
			} else {
				logresult('WARN', $msg);
			}
		}
	
		sub stat_cifs_file {
		        my ($winpath, %opts) = @_;
			logcomment("Enter stat_cifs_file:  \"$winpath\"")
			    if ($verbose);
			$opts{cifs_client} = $main::CIFS_CLIENT unless ($opts{cifs_client});
		        my $cifsconn = make_conn($opts{cifs_client}, "#rsh");
		        my $cifs_hostname = $cifsconn->hostp()->hostname();
			# my $output = Dumper(stat$winpath);
			my $output = minidumper_array(stat$winpath);
			chomp($output);
			my @lines = split(/\n/, $output);
			my @stat_list = (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
			for (@lines) {
				#  Data::Dumper
				# if (/VAR(\d+)\s+=\s+(.*)/) {
				#	$stat_list[$1-1] = $2;
				# }
				#
				# "minidumper"
				if (/ARRAY<(\d+)>=([^#]*)#/) {
					$stat_list[$1]=$2;
				}
			}
			my $s = {};
		        ($s->{dev}, $s->{ino}, $s->{mode}, $s->{nlink}, $s->{uid}, $s->{gid}, $s->{rdev},
		          $s->{size}, $s->{atime}, $s->{mtime}, $s->{ctime}, $s->{blksize}, $s->{blocks}) = @stat_list;
			logcomment("stat_cifs_file   WinPath: \"$winpath\"   Returning:\n\n".Dumper($s))
			     if ($verbose);
			return( $s );
		}


		if (! -e $cifs_path) {
			logcomment("stat_cifs_dirtree subtest:  cifs_path \"$cifs_path\" not found");
			return;
		}
		$cifs_path =~ s%/%\\%g;
		my $children_hash = {};
		my $dirtree_hash = {};
		my $pathname = $cifs_path;
		my $basename = File::Basename::basename($cifs_path);
		my $dirname = File::Basename::dirname($cifs_path);
		my $filer_pathname = cvt_cifs_path_to_filer($pathname);
		$dirtree_hash->{ROOT} = $filer_pathname;
		sub minidumper_hash {
			my (%hash) = @_;
			my $res_str = "";
			while ( my ($k, $v) = each(%hash) ) {
				my $v_str;
				if (!ref($v)) {
					$v_str = $v;
				} elsif (ref($v) =~ /ARRAY/) {
					$v_str = join(",", (@$v, ""));
				} else {
					LogException("minidumper_hash:  illegal value: $v");
				}
				$res_str .= "HASH<$k>=$v_str#\n";
			}
			return($res_str);
		}

		sub minidumper_array {
			my (@array) = @_;
			my $res_str = "";
			for (my $i=0; $i<scalar(@array); $i++) {
				my $v = $array[$i];
				if (ref($v)) {
					LogException("minidumper_array: illegal value: $v");
				}
				$res_str .= "ARRAY<$i>=$v#\n";
			}
			logcomment("minidumper_array:\n\n".Dumper(@array)."\n\nReturning:\n\n$res_str")
			    if ($verbose);
			return($res_str);
		}

		sub make_conn  {
			my ($spec, $conn) = @_;
			$conn = "#rsh" if (!$conn);
			if (ref($spec) =~ /Connectrec/) {
				return $spec;
			} elsif (ref($spec) =~ /Hostrec/) {
				return connect($spec->{id} . $conn);
			} else {
				my $hspec=host($spec);
				return connect($hspec->{id} . $conn);
			}
		}

		sub get_cifs_file_owner {
		        my ($winpath, %opts) = @_;
			logcomment("Enter get_cifs_file_owner:  \"$winpath\"")
			    if ($verbose);
			sleep(2);
			$opts{cifs_client} = $main::CIFS_CLIENT unless ($opts{cifs_client});
		        my $cifsconn = make_conn($opts{cifs_client}, "#rsh");
		        my $cifs_hostname = $cifsconn->hostp()->hostname();
			my $cifs_cmmd = "x:\\bin\\subinacl.exe /file \"$winpath\" /display";
			my (undef, undef, $output) = say($cifsconn, $cifs_cmmd);
			###my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
			###my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
			chomp($output);
			my @lines = split(/\n/, $output);
			my $line;
			my @filtered_lines = ();
			my $file_not_found_error;
			my ($owner, $group);
			foreach (@lines) {
				my $orig_line = $_;
				s/^\s*//g;
				s/\s*$//g;
				s%[^:\w\ \t=\\\/\-\\._]%%g;
				push(@filtered_lines, $_);
				if (m%^\/owner\s+=(.*)%) {
					$owner = $1;
				}
				if (m%^\/primary group\s+=(.*)%) {
					$group = $1;
				}
				last if ($owner && $group);
				if (m%CreateFile\s+Error%i) {
					$file_not_found_error = $_;
				}
			}
			if ($owner) {
				logcomment("get_cifs_file_owner:   \"$winpath\"\n\nReturning:  ".
						"OWNER: \"$owner\"   GROUP: \"$group\"")
				    if ($verbose);
			} elsif ($file_not_found_error) {
				logcomment("get_cifs_file_owner:  File not found error for ".
						"\"$winpath\"\n\n$file_not_found_error")
				    if ($verbose);
			} else {
				logcomment("get_cifs_file_owner:  \"$winpath\"\n\n".
						"No owner found. Filtered Output:\n\n".
						join("\n", @filtered_lines))
				    if ($verbose);
			}
			return($owner, $group);
		}

		sub cvt_cifs_path_to_filer {
			my ($cifs_path) = @_;
			my $filer_path = $cifs_path;
			$filer_path =~ s%\\%/%g;
			if ($filer_path =~ s%^$mapped_drive(\\|/)?%$testvol_filer_path/%) {
				$filer_path =~ s%/$%%;
				return($filer_path);
			} else {
				LogException("cvt_cifs_path_to_filer: unexpected format for cifs_path: $cifs_path");
			}
		}

		my $_change_dir_saved_dir;
		my $_change_dir_caller;
		sub change_dir {
			my ($new_dir, $caller) = @_;
			my $old_dir = cwd();
			$_change_dir_saved_dir = $old_dir;
			$_change_dir_caller = $caller;
			$caller = "change_dir" unless ($caller);
			chdir($new_dir)
				|| LogException("$caller: chdir \"$new_dir\": $!");
			return($old_dir);
		}

		sub restore_dir {
			my $caller = $_change_dir_caller || "restore_dir";
			chdir($_change_dir_saved_dir)
				|| LogException("$caller: chdir \"$_change_dir_saved_dir\": $!");
		}


		my $cifs_stat_fun = sub {
			my ($pathname, $dirname, $basename) = @_;
				$dirname=~ s%/%\\%g;
				$pathname=~ s%/%\\%g;
				if (! -e $pathname) {
					LogException("cifs_stat_fun  Pathname \"$pathname\" not found");
				}
				my $filer_pathname = cvt_cifs_path_to_filer($pathname);
				my @path_components = split(m%/%, $filer_pathname);
				return if ( grep( /^.snapshot$/, @path_components) );

				logcomment("Enter cifs_stat_fun\nPathname: \"$pathname\"\nDirname: \"$dirname\"\nBasename: ".
					   "\"$basename\"\nFiler Pathname: \"$filer_pathname\"\n\nDIRTREE_HASH:\n\n".minidumper_array(keys(%$dirtree_hash)))
				    if ($verbose);

				my $s = stat_cifs_file($pathname);
				$dirtree_hash->{$filer_pathname} = $s;
				$s->{pathname} = $filer_pathname;
				$s->{cifs_pathname} = $pathname;
				$s->{name} = $basename;
				$s->{dirname} = $dirname;
				if (-d $pathname) {
					$s->{type} = 'dir';
				} elsif (-f $pathname) {
					$s->{type} = 'file';
				} elsif (-l $pathname) {
					$s->{type} = 'link';
				}
				my ($owner, $group) = get_cifs_file_owner($pathname);
				logcomment("cifs_stat_fun:   OWNER: \"$owner\"  GROUP: \"$group\"")
				    if ($verbose);
				$s->{owner} = $owner if ($owner);
				$s->{group} = $group if ($group);

				logcomment("cifs_stat_fun  FILER_PATHNAME: \"$filer_pathname\"")
				     if ($verbose);
				my $filer_dirname = cvt_cifs_path_to_filer($dirname);
				my $parents_children = $children_hash->{$filer_dirname};
				if ( !defined($parents_children) ) {
					$children_hash->{$filer_dirname} = [];
					$parents_children = $children_hash->{$filer_dirname};
				}
				push(@$parents_children, $basename) if ($basename ne ".");
				my @childlist = ();
				if ($s->{type} eq 'dir') {
					if (opendir(DIR, $pathname)) {
#						@childlist = grep( { $_ ne '.' and $_ ne '..' }   readdir(DIR) );
#						closedir(DIR);
					} elsif ($! =~ /Permission denied/i) {
						logresult('INFO', "cifs_stat_fun:\n".
							  "'permission denied' error from  opendir (via CIFS) for \"$pathname\": $!");
					} else {
						LogException("cifs__dirtree_stat_fun: opendir \"$pathname\": $!");
					}
					my $child_info = $children_hash->{$filer_pathname};
					if (!defined($child_info)) {
						$child_info = [];
					}
					@childlist = @$child_info;
					my $children = [];
					foreach my $fname (@childlist) {
						my $fpath = "$filer_pathname/$fname";
						push(@$children, $fpath);
						next if (!$recursive);
						my $fhash = $dirtree_hash->{$fpath};
						if (!defined($fhash)) {
							LogException("WARNING: cifs_stat_fun: No info found for \"$fpath\"\n\n");
						}
						$fhash->{parent} = $filer_pathname;
					}
					$s->{children} = $children;
					if ($recursive) {
						my $descendents = [ @$children ];
						foreach my $child (@$children) {
							if ( defined($dirtree_hash->{$child}{descendents}) ) {
								push(@$descendents, @{$dirtree_hash->{$child}{descendents}});
							}
						}
						$s->{descendents} = $descendents;
					}
				}
				logcomment("cifs_stat_fun  Returning   Pathname: \"$pathname\"\n\n".Dumper($s))
				     if ($verbose);
		};

		my $cifs_stat_find_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				### my $fhandle = _;
				$dirname=~ s%/%\\%g;
				$pathname=~ s%/%\\%g;
				if (! -e $pathname) {
					logcomment("cifs_stat_fun  Pathname \"$pathname\" not found");
					return;
				}
				return( &$cifs_stat_fun($pathname, $dirname, $basename) );
		};

		if ($recursive) {
			finddepth($cifs_stat_find_fun, $cifs_path);
		} else {
			change_dir($dirname, "cifs_dirtree_stat");
			&$cifs_stat_fun($pathname, $dirname, $basename, 1);
			restore_dir();
		}

		my $res_str = "ROOT=$dirtree_hash->{ROOT}#\n";

		while ( my ($path, $hash) = each(%$dirtree_hash) ) {
			next if ($path eq "ROOT");
			$res_str .= "PATH=$path#\n";
			$res_str .= minidumper_hash(%$hash);
			# while ( my ($k, $v) = each(%$hash) ) {
			#	$res_str .= "$k=$v\n";
			# }
		}

		logcomment("stat_cifs_dirtree subtest  Returning  DIRTREE_HASH:\n\n".Dumper($dirtree_hash))
		    if ($verbose);
		### logresult("PASS", $res_str);
		logresult("INFO", $res_str);
	}
__EOF__

	my $sub = subtest( $stat_cifs_file_or_dirtree_code_block,
			    "-client", $main::CIFS_CLIENT,
			    "--",
			    $cifs_path,
			    $recursive,
			    $QCState{mapped_drive},
			    $QCState{testvol_filer_path},
                        );

	$sub->wait_finish();
	my @results = $sub->results();
	if (!@results) {
		logcomment("No results returned from stat_cifs_dirtree subtest")
		    if ($verbose);
		return( {} );
	}
	my $worst = NATE::Result::worst(@results);
	if (!defined($worst)) {
		return({});
	}
	if ($worst->severity() != 0) {
		logresult($worst);
		LogException("Error from stat_cifs_dirtree");
	}
	my $last = pop(@results);
	if (!defined($last)) {
		return({});
	}
	my @messages = $last->messages();
	my $message = join("\n", @messages);

	my @lines = split(/\n/, $message);
	my $finfo_hash = {};
	my $finfo;

	for my $line (@lines) {
		if ($line =~ /ROOT=([^#]*)#/) {
			$finfo_hash->{ROOT} = $1;
			next;
		} elsif ($line =~ /HASH<([^>]+)>=([^#]*)#/) {
			my $key = $1;
			my $val_str = $2;
			if ($val_str =~ /,/) {
				my @split_val = split(/,/, $val_str);
				@split_val = grep( { $_ ne "" }  @split_val );
				$finfo->{$key} = [ @split_val ];
			} else {
				$finfo->{$key} = $val_str;
			}
		} elsif ($line =~ /PATH=([^#]*)/) {
			$finfo = {};
			$finfo_hash->{$1} = $finfo;
			next;
		} else {
		}
	}

	while ( my($k,$f) = each(%$finfo_hash) ) {
		next unless (ref($f) =~ /HASH/);
		my $parts = get_filer_path_parts($f->{pathname});
		$f->{name} = $parts->{name};
		$f->{qtree} = $parts->{qtree};
		$f->{dir} = $parts->{dir};
	}

	logDbg("Exit cifs_dirtree_stat   cifs_path: \"$cifs_path\"  Returning:\n\n".Dumper($finfo_hash));
	$Data::Dumper::Maxdepth = 4;
	return( $finfo_hash );
}



##########################################
#
#    'read existing (filer/client) state' routines
#
##########################################

# is_qtree($qtree)
#
sub is_qtree {
        my ($qtree) = @_;
	read_existing_qtrees();
        my $qtree_path = get_qtree_path($qtree);
        ### return( qtreePresent($QCState{fc}, $qtree_path) );
	return( defined($qtree_info{$qtree_path}) );
}

# is_nondefault_qtree($qtree)
#
sub is_nondefault_qtree {
	my ($qtree) = @_;
        my $qtree_path = get_qtree_path($qtree);
	return ( is_qtree($qtree_path) && ($qtree_path ne $QCState{testvol_filer_path}) );
}


# get_filer_state()
#
#	Checks existing filer state, including
#	local users, groups, qtrees, and quotas
#	and the files and directories contained
#	in the test volume.
#
sub get_filer_state {
	logDbg("Enter get_filer_state");
	read_existing_nfs_groups();
	read_existing_nfs_users();
	read_existing_nfs_files();
	read_existing_cifs_users() unless ($no_cifs);
	read_existing_cifs_groups() unless ($no_cifs);
	read_existing_cifs_files() unless ($no_cifs);
	read_existing_usermaps();
	read_existing_qtrees();
	read_existing_quotas();
	logDbg("Exiting get_filer_state");
}


# show_filer_state()
#
sub show_filer_state {
use SayWrap;
	my ($msg) = @_;
	$msg = "SHOW_FILER_STATE" unless ($msg);
	clean_local_state();
	get_filer_state();
	my $qr = eval { getQuotaReport($QCState{fc}, flag => "-x") };
	my $qf = sayEOF($QCState{fc}, "rdfile /etc/quotas");
	my $umf = sayEOF($QCState{fc}, "rdfile /etc/usermap.cfg");
	my $pwf = sayEOF($QCState{fc}, "rdfile /etc/passwd");
	my $grf = sayEOF($QCState{fc}, "rdfile /etc/group");
	my $qrs = eval { sayEOF($QCState{fc}, "quota report -x") };
	my $uas = sayEOF($QCState{fc}, "useradmin user list");
	my $ugs = sayEOF($QCState{fc}, "useradmin group list");
	my %state_hash = (
		NFS_USER_INFO		=> \%nfs_user_info,
		NFS_GROUP_INFO		=> \%nfs_group_info,
		NFS_FILE_INFO		=> \%nfs_file_info,
		CIFS_USER_INFO		=> \%cifs_user_info,
		CIFS_GROUP_INFO		=> \%cifs_group_info,
		CIFS_FILE_INFO		=> \%cifs_file_info,
		USERMAP_INFO		=> \%usermap_info,
		QTREE_INFO		=> \%qtree_info,
		QUOTA_INFO		=> \%quota_info,
	);
	logcomment("\n\n$msg:\n\n".Dumper(\%state_hash)."\n\n"
		   ."LOCAL USERS:\n\n$uas\n\n"
		   ."LOCAL GROUPS:\n\n$ugs\n\n"
		   ."USERMAP FILE:\n\n$umf\n\n"
		   ."PASSWD FILE:\n\n$pwf\n\n"
		   ."GROUP FILE:\n\n$grf\n\n"
		   ."QUOTAS FILE:\n\n$qf\n\n"
		   ."QUOTA REPORT:\n\n$qrs\n\n".Dumper($qr));
}


# show_quota_state()
#
sub show_quota_state {
use SayWrap;
	my ($msg) = @_;
	$msg = "SHOW_FILER_STATE" unless ($msg);
	my $qr = eval { getQuotaReport($QCState{fc}, flag => "-x") };
	my $qf = eval { sayEOF($QCState{fc}, "rdfile /etc/quotas") };
	my $qrs = eval { sayEOF($QCState{fc}, "quota report -x") };
	logcomment("\n\n$msg:\n\nQUOTAS FILE:\n\n$qf\n\n"
		   ."PREVIOUS QUOTA REPORT:\n\n$prev_qr_str\n\n".Dumper($prev_qr_hash)."\n\n"
		   ."CURRENT QUOTA REPORT:\n\n$qrs\n\n".Dumper($qr));
}


# clean_local_state()
#
#   Empties all 'filer state' hashes.
#   Prevents use of 'stale' values.
# 
sub clean_local_state {
	%nfs_user_info = ();
	%nfs_user_info_by_uid = ();
	%nfs_group_info = ();
	%nfs_group_info_by_gid = ();
	%nfs_file_info = ();
	%cifs_user_info = ();
	%cifs_group_info = ();
	%cifs_file_info = ();
	%usermap_info = ();
	%qtree_info = ();
	%quota_info = ();
	%_prev_qtree_security = ();
}


# get_filer_nfs_state()
#
sub get_filer_nfs_state {
	read_existing_qtrees();
	read_existing_nfs_files();
	read_existing_nfs_users();
	read_existing_nfs_groups();
}


# get_filer_cifs_state()
#
sub get_filer_cifs_state {
	return if ($no_cifs);
	read_existing_qtrees();
	read_existing_cifs_files();
	read_existing_cifs_groups();
	read_existing_cifs_users();
}


# read_existing_qtrees()
#
#	Finds existing qtrees,
#	and internal state appropriately.
#
sub read_existing_qtrees {
	if ( keys(%qtree_info) ) {
		return( {%qtree_info} );
	}
	logDbg("Enter read_existing_qtrees");
	my %qthash = qtreeStatusVol($QCState{fc}, $QCState{testvol});
	%qtree_info = %qthash;
	logDbg("Exiting read_existing_qtrees:\n\n".Dumper(\%qthash));
	return( { %qthash } );
}


# get_qtree_security($qtree)
#
#	Returns existing security style for qtree '$qtree'
sub get_qtree_security {
	my ($qtree) = @_;
	my $qtpath = get_qtree_path($qtree);
	my $qtree_hash = read_existing_qtrees;
	my $qt_hash = $qtree_hash->{$qtpath};
	if ($qt_hash) {
		return( $qt_hash->{security} );
	}
	return( undef );
}


# read_existing_usermaps()
#
#	Reads entries from "/etc/usermap.cfg" to determine existing CIFS <--> NFS user mappings
#
sub read_existing_usermaps {
	logDbg("Enter read_existing_usermaps");
        my @usermap_list = readUsermapFile($QCState{fc});
	my $usermap_hash = {};
	foreach my $entry (@usermap_list) {
		my $key = $entry->{cifs_user};
		$usermap_hash->{$key} = $entry;
	}
	%usermap_info = %$usermap_hash;
	logDbg("Exiting read_existing_usermaps:\n\n".Dumper($usermap_hash));
	return( $usermap_hash );
}


# read_existing_nfs_users()
#
#	Reads current /etc/passwd file, and updates
#	internal state appropriately
#
sub read_existing_nfs_users {
	if ( keys(%nfs_user_info) ) {
		return( {%nfs_user_info} );
	}
	logDbg("Enter read_existing_nfs_users");
	if (!keys(%nfs_group_info)) {
		read_existing_nfs_groups();
	}
	my $users_hash = readPasswdFile($QCState{fc});
	%nfs_user_info_by_uid = ();
	foreach my $usrhash (values(%$users_hash)) {
		$nfs_user_info_by_uid{$usrhash->{uid}} = $usrhash;
	}
	%nfs_user_info = %$users_hash;
	logDbg("Exit read_existing_nfs_users:\n\n".Dumper($users_hash));
	return($users_hash);
}


# read_existing_nfs_groups()
#
#	Reads current /etc/group file, and updates
#	internal state appropriately
#
sub read_existing_nfs_groups {
	if ( keys(%nfs_group_info) ) {
		return( {%nfs_group_info} );
	}
	logDbg("Enter read_existing_nfs_groups");
	my $groups_hash = readGroupFile($QCState{fc});
	%nfs_group_info_by_gid = ();
	foreach my $grphash (values(%$groups_hash)) {
		$nfs_group_info_by_gid{$grphash->{gid}} = $grphash;
	}
	%nfs_group_info = %$groups_hash;
	logDbg("Exit read_existing_nfs_groups\n\n".Dumper($groups_hash));
	return( $groups_hash );
}


# read_existing_cifs_users()
#
#	Reads current local (CIFS) users,
#	and modify internal state appropriately
#
sub read_existing_cifs_users {
	if ( keys(%cifs_user_info) ) {
		return( {%cifs_user_info} );
	}
	if (!keys(%cifs_group_info)) {
		read_existing_cifs_groups();
	}
	logDbg("Enter read_existing_cifs_users");
	my @users_hash_list = listCIFSUsers($QCState{fc});
	my $users_hash = {};
	foreach my $usr (@users_hash_list) {
		my $hash = {
				login => $usr->{Name},
				group => $usr->{Groups},
				rid => $usr->{Rid},
			   };
		$users_hash->{$usr->{Name}} = $hash;
	}
	%cifs_user_info_by_rid = ();
	foreach my $uhash (values(%$users_hash)) {
		$cifs_user_info_by_rid{$uhash->{rid}} = $uhash;
	}
	%cifs_user_info = %$users_hash;
	logDbg("Exit read_existing_cifs_users:\n\n".Dumper($users_hash));
	return($users_hash);
}


# read_existing_cifs_groups()
#
#	Reads current local groups,
#	and modify internal state appropriately
#
sub read_existing_cifs_groups {
	if ( keys(%cifs_group_info) ) {
		return( {%cifs_group_info} );
	}
	logDbg("Enter read_existing_cifs_groups");
	my @groups_hash_list = listCIFSGroups($QCState{fc});
	my $groups_hash = {};
	foreach my $grp (@groups_hash_list) {
		my $hash = {
				group => $grp->{Name},
				roles => $grp->{Roles},
				rid => $grp->{Rid},
			   };
		$groups_hash->{$grp->{Name}} = $hash;
	}
	%cifs_group_info_by_rid = ();
	foreach my $uhash (values(%$groups_hash)) {
		$cifs_group_info_by_rid{$uhash->{rid}} = $uhash;
	}
	%cifs_group_info = %$groups_hash;
	logDbg("Exit read_existing_cifs_groups:\n\n".Dumper($groups_hash));
	return($groups_hash);
}


# read_existing_quotas()
#
#	Reads current /etc/quotas file, and updates
#	internal state appropriately
#
sub read_existing_quotas {
	if ( keys(%quota_info) ) {
		return( {%quota_info} );
	}
	logDbg("Enter read_existing_quotas");
	logDbg("read_existing_quotas  Line: ".__LINE__);
	my $quotas = readQuotaFile($QCState{fc});
	%quota_info = ( user => {}, group => {}, tree => {} );
	logDbg("read_existing_quotas  Line: ".__LINE__."  QUOTAS:\n\n".Dumper($quotas));
	for my $q (@$quotas) {
		logDbg("read_existing_quotas  Line: ".__LINE__."  Q:\n\n".Dumper($q));
		my $type = $q->{type};
		$type =~ s/\@.*//;
		my %opts = ();
		my $vol = $q->{volname};
		next unless ($vol eq $QCState{testvol});
		$opts{target} = $q->{target};
		$opts{vol} = $vol;
		$opts{tree} = $q->{treepath};
		$opts{disk} = $q->{disk};
		$opts{files} = $q->{files};
		logDbg("read_existing_quotas  Calling create_quota_template  TYPE: \"$type\"\n\nOPTS:\n\n".Dumper(\%opts));
		my ($qinfo, $ignore) = create_quota_template($type, %opts);
		my $key = $qinfo->{key};
		logDbg("read_existing_quotas:  Returned from create_quota_template   QINFO:\n\n".Dumper($qinfo)."\n\nKEY: \"$key\"");
		if ($type eq 'tree') {
			$quota_info{tree}{$key} = $qinfo;
		} elsif ($type eq 'user') {
			$quota_info{user}{$key} = $qinfo;
		} elsif ($type eq 'group') {
			$quota_info{group}{$key} = $qinfo;
		}
	}
	logDbg("Exit read_existing_quotas:\n\n".Dumper(\%quota_info));
	return( {%quota_info} );
}


# read_existing_nfs_files()
#
#	Finds existing NFS files and directories in the test
#	volume, and internal state appropriately.
#
sub read_existing_nfs_files {
	if ( keys(%nfs_file_info) ) {
		return( {%nfs_file_info} );
	}
	logDbg("Enter read_existing_nfs_files");
	my $stat_hash = nfs_dirtree_stat($QCState{testvol_mounted_path});
	%nfs_file_info = %$stat_hash;
	logDbg("Exiting read_existing_nfs_files:\n\n".Dumper($stat_hash));
	return($stat_hash);
}


# read_existing_cifs_files()
#
#	Finds existing CIFS files and directories in the test
#	volume, and internal state appropriately.
#
sub read_existing_cifs_files {
	if ( keys(%cifs_file_info) ) {
		return( {%cifs_file_info} );
	}
	logDbg("Enter read_existing_cifs_files");
	my $stat_hash = cifs_dirtree_stat($QCState{mapped_drive});
	%cifs_file_info = %$stat_hash;
	logDbg("Exit read_existing_cifs_files:\n\n".Dumper($stat_hash));
	return($stat_hash);
}


# nfs_file_or_dirtree_stat($nfs_path, %opts)
#
sub nfs_file_or_dirtree_stat {
	my ($nfs_path, %opts) = @_;
	my $recursive = $opts{recursive};
	logDbg("Enter nfs_file_or_dirtree_stat\nNFS_PATH: \"$nfs_path\"   RECURSIVE: \"$recursive\"");
	my $dirtree_hash = {};
	my $children_hash = {};
	my $pathname = $nfs_path;
	my $basename = basename($nfs_path);
	my $dirname = dirname($nfs_path);
	my $filer_pathname = cvt_mounted_path_to_filer($pathname);
	$dirtree_hash->{ROOT} = $filer_pathname;

	my $nfs_file_or_dirtree_stat_fun = sub {
			my ($pathname, $dirname, $basename) = @_;
			logDbg("Enter nfs_file_or_dirtree_stat_fun\nPATHNAME: \"$pathname\"\nBASENAME: \"$basename\"\n".
				    "DIRNAME: \"$dirname\"");
			my $filer_pathname = cvt_mounted_path_to_filer($pathname);
			my @path_components = split(m%/%, $filer_pathname);
			return if ( grep( /^.snapshot$/, @path_components) );
			my $s = {};
			($s->{dev}, $s->{ino}, $s->{mode}, $s->{nlink}, $s->{uid},
				$s->{gid}, $s->{rdev}, $s->{size}, $s->{atime},
				$s->{mtime}, $s->{ctime}, $s->{blksize}, $s->{blocks}) =
			     stat($pathname);
			$dirtree_hash->{$filer_pathname} = $s;
			$s->{pathname} = $filer_pathname;
			$s->{mounted_pathname} = $pathname;
			$s->{name} = $basename;
			$s->{dirname} = $dirname;
			my $parts = get_filer_path_parts($filer_pathname);
			$s->{name} = $parts->{name};
			$s->{qtree} = $parts->{qtree};
			$s->{dir} = $parts->{dir};
			if (-d $pathname) {
				$s->{type} = 'dir';
			} elsif (-f $pathname) {
				$s->{type} = 'file';
			} elsif (-l $pathname) {
				$s->{type} = 'link';
			}
			my $filer_dirname = cvt_mounted_path_to_filer($dirname);
			my $parents_children = $children_hash->{$filer_dirname};
			if ( !defined($parents_children) ) {
				$children_hash->{$filer_dirname} = [];
				$parents_children = $children_hash->{$filer_dirname};
			}
			push(@$parents_children, $basename) if ($basename ne ".");
			my @childlist = ();
			if ($s->{type} eq 'dir') {
				logDbg("nfs_file_or_dirtree_stat  Line: ".__LINE__."\n\nls -ld $pathname:\n\n".`ls -l $pathname`);
				if (opendir(DIR, $pathname)) {
					# @childlist = grep( { $_ ne '.' and $_ ne '..' }   readdir(DIR) );
					# closedir(DIR);
				} elsif ($! =~ /Permission denied/i) {
					logresult('INFO', "nfs_file_or_dirtree_stat:\n".
						  "'permission denied' error from  opendir (via NFS) for \"$pathname\": $!");
				} else {
					LogException("Unexpected error -- nfs_file_or_dirtree_stat \"$pathname\" opendir: $!");
				}
				my $child_info = $children_hash->{$filer_pathname};
				if (!defined($child_info)) {
					$child_info = [];
				}
				@childlist = @$child_info;

				my $children = [];
				foreach my $fname (@childlist) {
					my $fpath = "$filer_pathname/$fname";
					push(@$children, $fpath);
					next if (!$recursive);
					my $fhash = $dirtree_hash->{$fpath};
					if (!defined($fhash)) {
						LogException("nfs_dirtree_stat_fun: No info found for \"$fpath\"");
					}
					$fhash->{parent} = $filer_pathname;
				}
				$s->{children} = $children;
				if ($recursive) {
					my $descendents = [ @$children ];
					foreach my $child (@$children) {
						if ( defined($dirtree_hash->{$child}{descendents}) ) {
							push(@$descendents, @{$dirtree_hash->{$child}{descendents}});
						}
					}
					$s->{descendents} = $descendents;
				}
			}
		    };

		my $nfs_dirtree_find_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				### my $fhandle = _;
				if (! -e $basename) {
					logcomment("nfs_file_or_dirtree_stat_fun  Pathname \"$pathname\" not found");
					return;
				}
				return( &$nfs_file_or_dirtree_stat_fun($pathname, $dirname, $basename) );
		};

	if ($recursive) {
		logDbg("Line: ".__LINE__."     Calling (finddepth) nfs_dirtree_find_fun  \"$nfs_path\"");
		finddepth($nfs_dirtree_find_fun, $nfs_path);
	} else {
		change_dir($dirname, "nfs_file_or_dirtree_stat_fun");
		logDbg("Line: ".__LINE__."     Calling nfs_file_or_dirtree_stat_fun  \"$nfs_path\"");
		&$nfs_file_or_dirtree_stat_fun($pathname, $dirname, $basename, 1);
		restore_dir();
	}
	logDbg("Exit nfs_file_or_dirtree_stat  Returning:\n\n".Dumper($dirtree_hash));
	return($dirtree_hash);
}


# nfs_rmtree($nfs_path, %opts)
#
sub nfs_rmtree {
	my ($nfs_path, %opts) = @_;
	my $recursive = (defined($opts{recursive})) ? $opts{recursive} : 1;
	logDbg("Enter nfs_rmtree\nNFS_PATH: \"$nfs_path\"   RECURSIVE: \"$recursive\"");
	my $filer_path = cvt_mounted_path_to_filer($nfs_path);
	my $pathname = $nfs_path;
	my $basename = basename($nfs_path);
	my $dirname = dirname($nfs_path);
	my $this_host = `hostname`;
	chomp($this_host);
	my $curr_dir = cwd();
	logDbg("nfs_rmtree Line: ".__LINE__."   Calling 'chdir'");
	chdir('/') || LogException("nfs_rm_fun: chdir \"$dirname\": $!");
	logDbg("nfs_rmtree Line: ".__LINE__."    CWD: \"".cwd()."\"");

	my $nfs_rm_fun = sub {
			my ($pathname, $dirname, $basename) = @_;
			my $filer_pathname = cvt_mounted_path_to_filer($pathname);
			logDbg("Enter nfs_rm_fun  Line: ".__LINE__."\nPATHNAME: \"$pathname\"\nDIRNAME: \"$dirname\"\n".
				    "BASENAME: \"$basename\"\nCWD: \"".cwd()."\"");
			if ($basename eq ".") {
				logWarn("nfs_rm_fun: Attempt to delete directory '.' for pathname \"$pathname\" -- ignoring");
				return(1);
			}
			logDbg("nfs_rm_fun  Line: ".__LINE__."\n\nCWD: \"".cwd()."\"\n\nls -ld $basename:\n\n".
				     `ls -ld $basename`."\n\n".`ls -l $basename`);
			if (-d $pathname) {
				logDbg("CWD: \"".cwd()."\"");
				logDbg("nfs_rm_fun  Line: ".__LINE__."\nCalling: rmdir(\"$basename\")");
				### rmdir($pathname) || LogException("nfs_rm_fun \"$pathname\" rmdir: $!");
				### my $stat = eval { rmdir($pathname) };
				my $stat = eval { rmdir($basename) };
				if ($!) {
					logWarn("nfs_rm_fun: rmdir \"$pathname\": $!");
				}
			} else {
				logDbg("nfs_rm_fun  Line: ".__LINE__."\Calling: unlink(\"$pathname\")");
				unlink($pathname)
					|| LogException("nfs_rm_fun: unlink \"$pathname\": $!");
			}
			if (-e $pathname) {
				LogException("nfs_rm_fun  Line: ".__LINE__."   Could not delete \"$pathname\"");
			}
			return(1);
		    };

		my $nfs_rmtree_find_fun = sub {
				my $pathname = $File::Find::name;
				my $dirname = $File::Find::dir;
				my $basename = $_;
				logDbg("Enter nfs_rmtree_find_fun  Line: ".__LINE__."\nPATHNAME: \"$pathname\"\nDIRNAME: \"$dirname\"\n".
				    "BASENAME: \"$basename\"\nCWD: \"".cwd()."\"");
				### my $fhandle = _;
				if ($basename eq ".") {
					# Don't try to delete current directory
					return(1);
				}
				if (! -e $basename) {
					LogException("nfs_rmtree_find_fun  Pathname \"$pathname\" not found");
				}
				logDbg("nfs_rmtree_find_fun   Line: ".__LINE__."   Calling nfs_rm_fun");
				return( &$nfs_rm_fun($pathname, $dirname, $basename) );
		};

	if ($recursive) {
		finddepth($nfs_rmtree_find_fun, $nfs_path);
		 rmdir($nfs_path) 
			|| LogException("nfs_rmtree: rmdir \"nfs_path\": $!");
	} else {
		change_dir($dirname, "nfs_rmtree");
		&$nfs_rm_fun($pathname, $dirname, $basename, 1);
		restore_dir();
	}
	if (-e $nfs_path) {
		logDbg("nfs_rmtree:  Could not delete  \"$nfs_path\"");
	} else {
		logDbg("Exit nfs_rmtree: \"$nfs_path\" deleted");
	}
	chdir($curr_dir)
		|| LogException("nfs_rmtree: chdir \"$curr_dir\": $!");
	return(1);
}


# nfs_file_stat($nfs_path)
#
sub nfs_file_stat {
	my ($nfs_path) = @_;
	my $all_hash = nfs_file_or_dirtree_stat($nfs_path, recursive => 0);
	my $stat_hash = $all_hash->{$all_hash->{ROOT}};
	return($stat_hash);
}


# nfs_dirtree_stat($nfs_path)
#
sub nfs_dirtree_stat {
	my ($nfs_path) = @_;
	my $stat_hash = nfs_file_or_dirtree_stat($nfs_path, recursive => 1);
	return($stat_hash);
}


# cifs_file_stat($cifs_path)
#
sub cifs_file_stat {
	my ($cifs_path) = @_;
	my $all_hash = cifs_file_or_dirtree_stat($cifs_path, recursive => 0);
	my $stat_hash = $all_hash->{$all_hash->{ROOT}};
	return($stat_hash);
}

# cifs_dirtree_stat($cifs_path)
#
sub cifs_dirtree_stat {
	my ($cifs_path) = @_;
	my $stat_hash = cifs_file_or_dirtree_stat($cifs_path, recursive => 1);
	return($stat_hash);
}



##########################################
#
#    pathname conversion routines
#
##########################################

# cvt_filer_path_to_mounted($filer_path)
#
sub cvt_filer_path_to_mounted {
	my ($filer_path) = @_;
	my $mounted_path = $filer_path;
	if ($mounted_path =~ s%^$QCState{testvol_filer_path}%$QCState{testvol_mounted_path}%) {
		$filer_path =~ s%/$%%;
		return($mounted_path);
	} else {
		LogException("cvt_filer_path_to_mounted: unexpected format for filer_path: $filer_path");
	}
}


# cvt_filer_path_to_cifs($filer_path)
#
sub cvt_filer_path_to_cifs {
	my ($filer_path) = @_;
	my $cifs_path = $filer_path;
	if ($cifs_path =~ s%^$QCState{testvol_filer_path}%$QCState{mapped_drive}%) {
		$cifs_path =~ s%/%\\%g;
		$filer_path =~ s%\\$%%;
		return($cifs_path);
	} else {
		LogException("cvt_filer_path_to_cifs: unexpected format for filer_path: $filer_path");
	}
}


# cvt_cifs_path_to_filer($cifs_path)
#
sub cvt_cifs_path_to_filer {
	my ($cifs_path) = @_;
	my $filer_path = $cifs_path;
	$filer_path =~ s%\\%/%g;
	if ($filer_path =~ s%^$QCState{mapped_drive}(\\|/)?%$QCState{testvol_filer_path}/%) {
		$filer_path =~ s%/$%%;
		return($filer_path);
	} else {
		LogException("cvt_cifs_path_to_filer: unexpected format for cifs_path: $cifs_path");
	}
}


# cvt_mounted_path_to_filer($mounted_path)
#
sub cvt_mounted_path_to_filer {
	my ($mounted_path) = @_;
	my $filer_path = $mounted_path;
	if ($filer_path =~ s%^$QCState{testvol_mounted_path}%$QCState{testvol_filer_path}%) {
		$filer_path =~ s%/$%%;
		return($filer_path);
	} else {
		LogException("cvt_mounted_path_to_filer: unexpected format for mounted_path: $mounted_path");
	}
}


# get_qtree_path($qtree)
#
sub get_qtree_path {
	my ($qtree) = @_;
	my $qtree_path;
	if (!$qtree) {
		$qtree_path = $QCState{testvol_filer_path};
	} elsif ($qtree =~ m%$QCState{testvol_filer_path}/([^/]+)%) {
		$qtree_path = $qtree;
	} else {
		$qtree_path = "$QCState{testvol_filer_path}/$qtree";
	}
	return($qtree_path);
}


# get_qtree_name($qtree)
#
sub get_qtree_name {
	my ($qtree) = @_;
	my $qtree_name;
	if (!$qtree) {
		$qtree_name = "";
	} elsif ($qtree =~ m%$QCState{testvol_filer_path}/([^/]+)%) {
		$qtree_name = $1;
	} else {
		$qtree_name = $qtree;
	}
	return($qtree_name);
}


#   get_mounted_path_parts($mounted_path)
#
#   returns hashref:
#    {
#	qtree
#	dir
#	name
#    }
#
sub get_mounted_path_parts {
	my ($mounted_path) = @_;
	my $path_hash = {};
	if ($mounted_path =~ m%^$QCState{testvol_mounted_path}/?$%) {
		return( {} );
	} elsif ($mounted_path =~ m%^$QCState{testvol_mounted_path}/(.+)%) {
		my @path_parts = split(m%/%, $1);
		my $qtname = $path_parts[0];
		my $qtpath = "$QCState{testvol_filer_path}/$qtname";
		if (is_qtree($qtpath)) {
			$path_hash->{qtree} = $qtname;
			shift(@path_parts);
		}
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{name} = pop(@path_parts);
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{dir} = join("/", @path_parts);
		return($path_hash);
	}
}


#   get_filer_path_parts($filer_path)
#
#   returns hashref:
#    {
#	qtree
#	dir
#	name
#    }
#
sub get_filer_path_parts {
	my ($filer_path) = @_;
	my $path_hash = {};
	if ($filer_path =~ m%^$QCState{testvol_filer_path}/?$%) {
		return( {} );
	} elsif ($filer_path =~ m%^$QCState{testvol_filer_path}/(.+)%) {
		my @path_parts = split(m%/%, $1);
		my $qtname = $path_parts[0];
		my $qtpath = "$QCState{testvol_filer_path}/$qtname";
		if (is_qtree($qtpath)) {
			$path_hash->{qtree} = $qtname;
			shift(@path_parts);
		}
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{name} = pop(@path_parts);
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{dir} = join("/", @path_parts);
		return($path_hash);
	}
}


#   get_cifs_path_parts($cifs_path)
#
#   returns hashref:
#    {
#	qtree
#	dir
#	name
#    }
#
sub get_cifs_path_parts {
	my ($cifs_path) = @_;
	$cifs_path =~ s%\\%/%g;
	my $path_hash = {};
	if ($cifs_path =~ m%^$QCState{mapped_drive}/?$%) {
		return( {} );
	} elsif ($cifs_path =~ m%^$QCState{mapped_drive}/?(.+)%) {
		my @path_parts = split(m%/%, $1);
		my $qtname = $path_parts[0];
		my $qtpath = "$QCState{testvol_filer_path}/$qtname";
		if (is_qtree($qtpath)) {
			$path_hash->{qtree} = $qtname;
			shift(@path_parts);
		}
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{name} = pop(@path_parts);
		if (!@path_parts) {
			return($path_hash);
		}
		$path_hash->{dir} = join("/", @path_parts);
		return($path_hash);
	}
}


#   build_filer_path(%opts)
#
#   Opts:
#	qtree
#	dir
#	name
#
sub build_filer_path {
	my (%opts) = @_;
	my $filer_path = $QCState{testvol_filer_path};
	if ($opts{qtree}) {
		my $qtree_name = get_qtree_name($opts{qtree});
		$filer_path .= "/$qtree_name";
	}
	if ($opts{dir}) {
		my $dirpath = $opts{dir};
		$dirpath =~ s%\\%/%g;
		$filer_path .= "/$dirpath";
	}
	if ($opts{name}) {
		$filer_path .= "/$opts{name}";
	}
	return($filer_path);
}


#   build_mounted_path(%opts)
#
#   Opts:
#	qtree
#	dir
#	name
#
sub build_mounted_path {
	my (%opts) = @_;
	my $mounted_path = $QCState{testvol_mounted_path};
	if ($opts{qtree}) {
		my $qtree_name = get_qtree_name($opts{qtree});
		$mounted_path .= "/$qtree_name";
	}
	if ($opts{dir}) {
		my $dirpath = $opts{dir};
		$dirpath =~ s%\\%/%g;
		$mounted_path .= "/$dirpath";
	}
	if ($opts{name}) {
		$mounted_path .= "/$opts{name}";
	}
	return($mounted_path);
}


#   build_cifs_path(%opts)
#
#   Opts:
#	qtree
#	dir
#	name
#
sub build_cifs_path {
	my (%opts) = @_;
	my $cifs_path = $QCState{mapped_drive};
	if ($opts{qtree}) {
		my $qtree_name = get_qtree_name($opts{qtree});
		$cifs_path .= "\\$qtree_name";
	}
	if ($opts{dir}) {
		my $dirpath = $opts{dir};
		$dirpath =~ s%\\%/%g;
		$cifs_path .= "\\$dirpath";
	}
	if ($opts{name}) {
		$cifs_path .= "\\$opts{name}";
	}
	return($cifs_path);
}



##########################################
#
#    quota & quota report utilities
#
##########################################


# synth_quota_report($qr_hash_list)
#
#   Generates a 'synthetic' quota report (string) from a hash.
#   The hash has the same format as that returned by getQuotaReport.
#
sub synth_quota_report {
	my ($qr_hash_list) = @_;
	my $qr_str = "\n\t\t\t ------------  Synthetic Quota Report  ------------\n\n".
	     "Type\t\tId\t\tVolume\t\tTree\t\tKBytes Limit\tKBytes Used\tFiles Used\tFiles Limit\tQuota Spec\n".
"-------------------------------------------------------------------------------------------------------------------\n";
	for my $qr (@$qr_hash_list) {
		### my ($qtemplate, $ignore) = create_quota_template($qr->{type}, %opts);
		$qr_str .= "$qr->{type}\t\t$qr->{target}\t$qr->{volume}\t$qr->{tree}\t\t$qr->{kbytes_used}\t".
			   "$qr->{kbytes_limit}\t$qr->{files_used}\t$qr->{files_limit}\t\t$qr->{quotaspec}\n";
	}
	return($qr_str);
}


# create_quota_template($type, %opts)
#
#    Given quota '$type' and '%opts',
#    returns hashes '$qinfo' and '$set_quota_args'
#
#    '$set_quota_args' is suitable for passing to functions such as setSingleQuota.
#    '$qinfo' contains additional info, used by QCov_Ops functions -- for example,
#     by auto_check_quota_report(). The global %quota_info hash is built using the
#     '$qinfo' hash entries.
#
sub create_quota_template {
	my ($type, %opts) = @_;
	logDbg("Enter create_quota_template  TYPE: \"$type\"  OPTS:\n\n".Dumper(\%opts));
	my $user = $opts{user};
	my $tree = $opts{tree};
	my $qtree_path = get_qtree_path($tree);
	my $qtree_name = get_qtree_name($tree);
	my $group = $opts{group};
	my $default = $opts{default};
	my $target = $opts{target};
	my $usermap = $opts{usermap};
	my $files = $opts{files};
	my $disk = $opts{disk};
	if ($type ne 'user' && $type ne 'tree' && $type ne 'group') {
		LogException("unknown quota type: \"$type\"");
	}
	if ($type eq 'user') {
		if ($default || $group eq "*" || $target eq "*") {
			$user = "*";
			$target = "*";
			$default = 1;
		} elsif ($target) {
			$user = $target;
		} elsif ($user) {
			$target = $user;
		} else {
			LogException("no 'user' given");
		}
	} elsif ($type eq 'group') {
		if ($default || $group eq "*" || $target eq "*") {
			$group = "*";
			$target = "*";
			$default = 1;
		} elsif ($target) {
			$group = $target;
		} elsif ($group) {
			$target = $group;
		} else {
			LogException("no 'group' given");
		}
	} elsif ($type eq 'tree') {
		if ($default || $tree eq "*" || $target eq "*") {
			$tree = "*";
			$target = "*";
			$default = 1;
		} elsif (defined($target)) {
			$qtree_path = get_qtree_path($tree);
			$qtree_name = get_qtree_name($tree);
			$tree = $qtree_path;
			$target = $qtree_path;
		} else {
			$tree = $qtree_path;
			$target = $qtree_path;
		}
	} else {
		LogException("unknown quota type: \"$type\"");
	}

	if ($type ne 'user' && $user) {
	}
	if ($type ne 'group' && $group) {
	}

	my $set_quota_args = {};
	$set_quota_args->{type} = $type;
	$set_quota_args->{target} = $target;
	$set_quota_args->{vol} = $QCState{testvol};
	$set_quota_args->{tree} = $qtree_name unless ($tree eq "*");
	$set_quota_args->{files} = $files if ($files);
	$set_quota_args->{disk} = $disk if ($disk);
	$set_quota_args->{default} = $default if (defined($default));
	$set_quota_args->{usermap} = $usermap if (defined($usermap));

	my $qinfo = {};
	$qinfo->{type} = $type;
	$qinfo->{target} = $target;
	$qinfo->{quotaspec} = $target;
	$qinfo->{qtree_name} = $qtree_name;
	$qinfo->{qtree_path} = $qtree_path;
	$qinfo->{user} = $user if ($user);
	$qinfo->{default} = $default if (defined($default));
	$qinfo->{usermap} = $usermap if (defined($usermap));
	$qinfo->{group} = $group if ($group);
	$qinfo->{disk} = $disk if ($disk);
	$qinfo->{files} = $files if ($files);
	$qinfo->{vol} = $QCState{testvol};
	$qinfo->{volume} = $QCState{testvol};
	if ($type eq 'tree') {
		$qinfo->{key} = $qtree_path;
	} elsif ($type eq 'user') {
		$qinfo->{key} = "$user#$qtree_path";
	} elsif ($type eq 'group') {
		$qinfo->{key} = "$group#$qtree_path";
	}
	logDbg("Exiting create_quota_template  QINFO:\n\n".Dumper($qinfo).
		"\n\nSET_QUOTA_ARGS:\n\n".Dumper($set_quota_args));
	return($qinfo, $set_quota_args);
}


# sub get_quota_entry {
#	my ($type, %opts) = @_;
# logDbg("Enter get_quota_entry  TYPE: \"$type\"  OPTS:\n\n".Dumper(\%opts));
#	my ($qtemplate, $set_quota_args) = create_quota_template($type, %opts);
#	my $qtree_path = $qtemplate->{qtree_path};
#	my $user = $qtemplate->{user};
#	my $group = $qtemplate->{group};
#	return($qinfo);
# }



# compare_hash_lists($keys, $hash_a, $hash_b)
#
#     Compares two hash lists '$hash_a', '$hash_b', using keys '$keys'
#
#     Returns (all are ARRAYREF)
#		$matched_pairs
#		$unmatched_a
#		$unmatched_b
#		$dup_a
#		$dup_b
#
#     If the given hash lists match, and no duplicate entries are found, then
#     all returned ARRAYREFs except '$matched_pairs' are empty
#
sub compare_hash_lists {
	my ($keys, $hash_a, $hash_b) = @_;
	my $compare_hash_lists_verbose = 0;
	my $num_a = scalar(@$hash_a);
	my $num_b = scalar(@$hash_b);
	logDbg("Enter compare_hash_lists  KEYS:\n\n".Dumper($keys)."\n\nHASH_A ($num_a)\n\n".Dumper($hash_a).
		       "\n\nHASH_B ($num_b)\n\n".Dumper($hash_b));
	return([], [], [], [], [])  unless ($num_a || $num_b);
	my ($dup_a, $dup_b, $matched_pairs, $unmatched_a, $unmatched_b);
	my $a_info = {};
	my $b_info = {};
	map( { $a_info->{$_} = {} }   @$hash_a );
	map( { $b_info->{$_} = {} }   @$hash_b );
	@$dup_a = ();  @$dup_b = ();
	@$matched_pairs = ();

	for (my $i=0; $i<=$#$hash_a; $i++) {
		for (my $j=0; $j<$i; $j++) {
			my $match =1;
			for my $key (@$keys) {
				if ($$hash_a[$i]->{$key} ne $$hash_a[$j]->{$key}) {
					$match = 0;
				}
			}
			if ($match) {
				if ($compare_hash_lists_verbose) {
					logDbg("compare_hash_lists  Line: ".__LINE__."\n".
						   "DUP HASH_A FOUND:\n\n".Dumper($$hash_a[$i]));
				}
				push(@$dup_a, $$hash_a[$i]);
			}
		}
	}

	for (my $i=0; $i<=$#$hash_b; $i++) {
		for (my $j=0; $j<$i; $j++) {
			my $match =1;
			for my $key (@$keys) {
				if ($$hash_a[$i]->{$key} ne $$hash_a[$j]->{$key}) {
					$match = 0;
				}
			}
			if ($match) {
				if ($compare_hash_lists_verbose) {
					logDbg("compare_hash_lists  Line: ".__LINE__."\n".
						"DUP HASH_A FOUND:\n\n".Dumper($$hash_b[$i]));
				}
				push(@$dup_b, $$hash_b[$i]);
			}
		}
	}

	for my $a (@$hash_a) {
		$a_info->{$a}{best_num_matching} = -1;
		for my $b (@$hash_b) {
			next if ($b_info->{$b}{match});
			if ($compare_hash_lists_verbose) {
				logDbg("compare_hash_lists  Line: ".__LINE__."   Comparing A:\n\n".
					   Dumper($a)."\n\nB:\n\n".Dumper($b));
			}
			$b_info->{$b}{best_num_matching} = -1 unless (defined($b_info->{$b}{best_num_matching}));
			my $all_match = 1;
			my $num_matching = 0;
			my @failed_keys = ();
			for my $key (@$keys) {
				my $key_match = 0;
				if ($compare_hash_lists_verbose) {
					logDbg("Line: ".__LINE__."   Checking Key: \"$key\"");
				}
				if ($key =~ /\W/) {
					my @alt_keys = split(/\W/, $key);
					for my $kx (@alt_keys) {
						if ($a->{$kx} eq $b->{$kx}) {
							$key_match = 1;
							$num_matching++;
							last;
						}
					}
					if ($compare_hash_lists_verbose && !$key_match) {
						logDbg("compare_hash_lists  Line: ".__LINE__."\nMatch failed on ".
							     "alternate key:\"$key\"");
					}
				} else {
					if ($a->{$key} eq $b->{$key}) {
						$key_match = 1;
						$num_matching++;
					} else {
					}
				}
				if (!$key_match) {
					if ($compare_hash_lists_verbose) {
						logDbg("compare_hash_lists  Line: ".__LINE__."\nMatch failed on key: ".
							    "\"$key\"\nA: \"$a->{$key}\"\nB: \"$b->{$key}\"");
					}
					push(@failed_keys, $key) unless ($key_match);
				}
				$all_match &= $key_match;
			}
			if ($all_match) {
				if ($compare_hash_lists_verbose) {
					logDbg("compare_hash_lists  Line: ".__LINE__."   MATCH FOUND   A:\n\n".Dumper($a).
						    "\n\nB:\n\n".Dumper($b));
				}
				push(@$matched_pairs, [$a, $b]);
				$a_info->{$a}{match} = $b;
				$b_info->{$b}{match} = $a;
				delete($a_info->{$a}{best_match});
				delete($b->{$b}{best_match});
				last;
			} else {
				if ( ($num_matching > $a_info->{$a}{best_num_matching}) &&
				     ($num_matching > $b_info->{$b}{best_num_matching}) ) {
					$a_info->{$a}{best_num_matching} = $num_matching;
					$b_info->{$b}{best_num_matching} = $num_matching;
					$a_info->{$a}{best_match} = $b;
					$b_info->{$b}{best_match} = $a;
					if ($compare_hash_lists_verbose) {
						logDbg("compare_hash_lists  Line: ".__LINE__."  Match Failed for keys:\n".
							  "@failed_keys");
					}
				}
			}
		}
	}

	@$unmatched_a = grep( { !($a_info->{$_}{match}) }  @$hash_a );
	@$unmatched_b = grep( { !($b_info->{$_}{match}) }  @$hash_b );
	my $num_matched = scalar(@$matched_pairs);

	if (!$compare_hash_lists_verbose) {
		logDbg("Exit compare_hash_lists\n\nUnmatched A:\n\n".Dumper($unmatched_a).
			    "\n\nUnmatched B:\n\n".Dumper($unmatched_b));
		return($matched_pairs, $unmatched_a, $unmatched_b, $dup_a, $dup_b);
	}

	logDbg("Line: ".__LINE__."   UNMATCHED A:\n\n".Dumper($unmatched_a)."\n\nUNMATCHED B\n\n".Dumper($unmatched_b));
	logDbg("EXITING COMPARE_HASH_LISTS\n\nKEYS:\n\n".Dumper($keys).
		    "\n\nMATCHED_PAIRS($num_matched):\n\n".Dumper($matched_pairs).
		    "\n\nUNMATCHED A:\n\n".Dumper($unmatched_a).  "\n\nUNMATCHED B:\n\n".Dumper($unmatched_b));

	for my $ax (@$unmatched_a) {
		my @failed_keys = ();
		my $best = $a_info->{$ax}{best_match};
		for my $key (@$keys) {
			if ($ax->{$key} ne $best->{$key}) {
				push(@failed_keys, $key);
			}
		}
                logDbg("Hash_A Entry and Best Match differ for keys: @failed_keys\n\n".Dumper($ax).
				"\n\n".Dumper($best));
	}

	for my $bx (@$unmatched_b) {
		my @failed_keys = ();
		my $best = $b_info->{$bx}{best_match};
		for my $key (@$keys) {
			if ($bx->{$key} ne $best->{$key}) {
				push(@failed_keys, $key);
			}
		}
                logDbg("Hash_B Entry and Best Match differ for keys: @failed_keys\n\n".Dumper($bx).
				"\n\n".Dumper($best));
	}

	logDbg("Exit compare_hash_lists\n\nUnmatched A:\n\n".Dumper($unmatched_a)."\n\nUnmatched B:\n\n".Dumper($unmatched_b));
	return($matched_pairs, $unmatched_a, $unmatched_b, $dup_a, $dup_b);
}



##########################################
#
#    Setup / Cleanup routines
#
##########################################


sub QCov_Ops_test_setup
{
	my %opts = @_;
	$no_cifs = $opts{no_cifs};
	my $do_cifs = !$no_cifs;

	logDbg("Verify we are running with root privileges");
	if ($> != 0) {
		die("This test needs to be run as root!");	
	}

	if ($do_cifs) {
		logDbg("check for MinWin compatible CIFS client: \"$::CIFS_CLIENT\"");
		my $cifs_client_name = $::CIFS_CLIENT;
		$cifs_client_name =~ s/;.*//;
		if ( !minwin_compatible($::CIFS_CLIENT) ) {
			die("Windows client $cifs_client_name is not MinWin compatible");
		} else {
			logcomment("Windows client $cifs_client_name is MinWin compatible");
		}
		logDbg("Check for 'fsutil' availability on CIFS client");
		$fsutil_path = get_fsutil_path($::CIFS_CLIENT);
		dos_list_drives($::CIFS_CLIENT);
		if ($@) {
			die("Windows client $cifs_client_name does not have 'fsutil' available");
		} else {
			logcomment("'fsutil' was found on $cifs_client_name");
		}
		logDbg("Check for 'x:\\nate\\' directory on CIFS client");
		my $nate_found = dos_file_exists($::CIFS_CLIENT, "x:\\nate");
		if ($nate_found) {
			logcomment("X:nate directory exists on $cifs_client_name");
		} else {
			logcomment("X:nate directory was not found on $cifs_client_name");
		}
	}

	logDbg("create and NFS mount testvol");
	$fc = make_conn($::FILER, $::FILER_CONN);
	if (vol_exists($fc, $::FILER_VOL)) {
		qtreeSecurity($fc, "/vol/$::FILER_VOL", "unix")
		    or die $@;
	}
	(undef, $testvol, $mountpoint)
		= create_vol_and_mount(
					filer 		=> $::FILER,
					client		=> 'localhost',
					filer_conn 	=> $::FILER_CONN,
					filer_vol 	=> $::FILER_VOL,
					filer_vol_mnt 	=> $::FILER_VOL_MNT,
					mountpoint 	=> $::MOUNTPOINT,
					mountopts 	=> $::MOUNTOPTS,
					vol_type	=> $::VOL_TYPE,
					vol_size	=> $::VOL_SIZE,
					filer_aggr	=> $::FILER_AGGR,
					aggr_size	=> $::AGGR_SIZE,
					extern_setup	=> $::EXTERN_SETUP,
					cmdtimeout	=> $::CMDTIMEOUT );
	die($@) if ($@);

	$fc->timeout($::CMDTIMEOUT);
	logTrc("my fc=$fc, vol=$testvol, mountdir=$mountpoint");
	$connhost = $fc->connhost();
	$vol_filer_path = "/vol/$testvol";
	my $vol_opts = vol_options($fc, $testvol);
	$vol_nosnapdir = $vol_opts->{nosnapdir};
	vol_options($fc, $testvol, nosnapdir => "on");
	disableQuota($fc, $testvol);

	logTrc("Offline all volumes except root volume and test volume");
	my $vol_root = vol_root($fc);
	my @vol_list = vol_list($fc);
	foreach my $vol (@vol_list) {
		eval { disableQuota($fc, $vol); };
		### next if ($vol eq $vol_root || $vol eq $testvol);
		### eval { vol_offline($fc, $vol) };
		if ($@) {
			logresult('WARN', "Error returned from vol_offline for $vol:  $@");
		}
	}

	logDbg("backup config files");
	$vol0_mount = mount_root_vol($::FILER);
	logTrc("backup /etc/ files");
	backupFiles($vol0_mount, @backup_files);
	die($@) if ($@);

	logDbg("Initialize passwd, group, quota, usermap files");
	writeDefaultPasswdFile($fc);
	writeGroupFile($fc);
	writeQuotaFile($fc);
	writeUsermapFile($fc);
	disableQuota($fc, $testvol);
	@existing_cifs_users = listCIFSUsers($fc);
	removeCIFSUsersExcept($fc, "root", "administrator", $::CIFS_USER);

	logDbg("Create snapshot of test volume");
	license($fc, 'snaprestore') or die ($@);
	if (snapshot_exists($fc, $testvol, $testvol_snap)) {
		die("snapshot $testvol_snap already exists");
	}
	snapshot_create($fc, $testvol, $testvol_snap);
	if (!$testvol_snap) {
		die("Could not create snapshot of $testvol before starting test")
	}

	if ($do_cifs) {
		logDbg("setup cifs on the filer");
		logTrc("Call setup_cifs");
		#setup_cifs($::FILER, $::CIFS_PASSWD);
		cifs_setup(host => $::FILER, cifs_pass => $::CIFS_PASSWD, workgroup => "WORKGROUP");		
		logTrc("Returned from setup_cifs");

		logDbg("add cifs user");
		logTrc("Call removeCIFSUser");
		removeCIFSUser($fc, $::CIFS_USER, on_client=>1);
		logTrc("Returned from removeCIFSUser");
		logTrc("Call addCIFSUser");
		$cifs_user_added =
			addCIFSUser($fc, $::CIFS_USER, $::CIFS_PASSWD, "Users", on_client=>1);
		logTrc("Returned from addCIFSUser");
	
		logTrc("Calling share_add");
		share_add($fc, $::CIFS_SHARE, $vol_filer_path, verify=>1)
			or die($@);
		logTrc("Returned from share_add");
	
		logDbg("change security of CIFS share path");
		qtreeSecurity($fc, $vol_filer_path, "mixed")
			or die $@;
	
		logDbg("map drive on CIFS client");
		logTrc("Call map_cifs_share");
		$mapped_drive = map_cifs_share(
				  server    => $connhost,
				  client    => $::CIFS_CLIENT,
				  drive     => $::CIFS_DRIVE,
				  share	    => $::CIFS_SHARE,
				   workgroup => "WORKGROUP",				  
				  password  => $::CIFS_PASSWD,
				  username  => $::CIFS_USER)
			or die $@;
		logTrc("Returned from map_cifs_share");
		
		die "could not map drive"
			if ($mapped_drive eq "");
	
		$mapped_drive .= ":"
			unless $mapped_drive =~ /:$/;
	}

	register_test_info( conn => $fc,
			    vol => $testvol,
			    cifs_client => $::CIFS_CLIENT,
			    mount_point => $mountpoint,
			    mapped_drive => $mapped_drive);

	return ($fc, $testvol);
}


sub QCov_Ops_test_cleanup
{
	logStep("CLEANUP... ($::CLEANUP)");
	$> = $saved_euid;  $) = $saved_egid;
	if (!$fc) {
		$fc = connect($::FILER_CONN, -timeout, $::TIMEOUT);
		$fc->timeout($::CMDTIMEOUT);
	}
	if ($::CLEANUP =~ /none/i) {
		common_cleanup($fc, mnt=>$mountpoint);
		return;
	}

	my $fh =  host($::FILER);
	if (is_cifs_setup(host=> $fh)) {
		logTrc("remove CIFS users created");
		logTrc("unmap/unmount drives");
		if (defined($mapped_drive)) {
			eval { unmap_cifs_share(drive => $mapped_drive,
						client => $::CIFS_CLIENT,
						force=>1)
			} ;
			logWarn($@) if ($@);
		}

	       if (share_exists($fc, $::CIFS_SHARE)) {
			logTrc("delete cifs share");
			eval { share_delete($fc, $::CIFS_SHARE) };
			logWarn($@) if ($@);
		}

		if (is_cifs_setup(host=> $fh)) {
			logTrc("terminate CIFS");
			filer_cifs_stop($fc);
			logWarn("filer_cifs_stop: $@") if ($@);
		}
	}

	logTrc("Remove added CIFS users");
	if ($cifs_user_added) {
		eval { removeCIFSUser($fc, $::CIFS_USER, on_client => 1) };
		logWarn("removeCIFSUser") if ($@);
	}
	logWarn("QCov_Ops::remove_added_cifs_users: $@") if ($@);
	eval { remove_added_cifs_users() };
	logWarn("QCov_Ops::remove_added_cifs_groups: $@") if ($@);
	eval { remove_added_cifs_groups() };
	logTrc("Restore existing CIFS users");
	for my $ux (@existing_cifs_users) {
		my $name = $ux->{Name};
		next if ($name eq "root");
		my $group = $ux->{Groups} || "Users";
		eval { addCIFSUser($fc, $name, $::CIFS_PASSWD, $group) };
		logWarn("addCIFSUser: $@") if ($@);
	}

	if ($testvol) {
		logTrc("Disable quotas on testvol");
		eval { disableQuota($fc, $testvol) };
		logTrc("Set testvol security to 'unix'");
		eval { qtreeSecurity($fc, "/vol/$testvol", "unix") };
		logWarn("qtreeSecurity") if ($@);
                eval { vol_options($fc, $testvol, nosnapdir => $vol_nosnapdir) };
		if (snapshot_exists($fc, $testvol, $testvol_snap)) {
			logTrc("Restore testvol snapshot");
			eval { snaprestore_volume($fc, $testvol, $testvol_snap) };
			logWarn("snaprestore failed: $@") if ($@);
			logTrc("Delete testvol snapshot");
			eval { snapshot_delete($fc, $testvol, $testvol_snap) };
			logWarn("snapshot_delete failed: $@") if ($@);
		}
	}

	if ($vol0_mount) {
		logTrc("Restoring backed-up files");
		eval { restoreFiles($vol0_mount, @backup_files) };
		logWarn("restoreFiles: $@") if ($@);

		eval { umount_vol($vol0_mount) };
		logWarn("umount_vol: $@") if ($@);
	}

	common_cleanup($fc,
			mnt 	=> $mountpoint,
			vol 	=> $::FILER_VOL,
			cleanup	=> $::CLEANUP,
			client 	=> 'localhost')
			
		or logWarn($@);
}


1;
