#
# $Id: //depot/prod/test/Rsierranevada/lib/Quota/QuotaEMS.pm#1 $
#
# Copyright (c) 2009 Network Appliance, Inc.
# All rights reserved.
#
#
#
## @summary Supporting Subroutines for 'quota_ems_rcca_1047_tests.thpl'
##
## @description
## WAFL002-003-012-001
## Verify quota reports correct EMS messages for various quota-related operations and events.
## initiated by various file operations (create, delete chown, move/rename, truncate/expand),
## for different quota target types (explicit/default), (tree, user, multi-user, group) with
## various operations (create file, chown, add user to the group, copy file/nt_stream...)
##
## Also Verify message is reported only when the event happens (no spam).
##
## -- At this time we will implement a design which does not cover all possible combinations
##    but several "points" in this space of combinations.
##
##		{ single-user vs. multi-user, vs. group vs qtree (type) }		x
##		{ hard-limit vs. soft-limit }						x
##		{ files-limit vs. disk-limit }						x
##		{ default-target vs. explicit-target }					x
##		{ volume-wide quota vs. qtree-specific quota }				x
##		{ CIFS-user vs. NFS-user vs. 'Mixed/Mapped' user }			x
##		{ quota-normal-to-quota-exceeded vs. quota-exceeded-to-quota-normal }	x
##		{ create vs. delete vs. move vs. chown vs. chgrp vs. truncate vs. append/expand }
##
##    .... even this partial enumeration of combinations give several hundred possible "points"
##    in the combination space. To enumerate and test all possible combinations, even if theoretically
##    possible, would be prohibitive both in test design and implementation resources, and
##    in test execution time and resources.
##
## @description Objective:
## Verify that quota reports correct EMS messages for Unix user on qtree when soft limit exceeded
##
##
## @author Bob Johnson (bobj@netapp.com)


#------------------------------------------------------------------
#      Package
#------------------------------------------------------------------
package Quota::QuotaEMS;

#------------------------------------------------------------------
#      Compiler directives.
#------------------------------------------------------------------
use strict;
use warnings;
use vars qw( @ISA @EXPORT_OK );

#------------------------------------------------------------------
#      Exported Symbols
#------------------------------------------------------------------
BEGIN {
	require Exporter;
	@ISA = ('Exporter');
	@EXPORT_OK = qw(
			quota_ems_setup
			quota_ems_cleanup
			get_quota_spec
			set_quota
			combine_hashtables
			set_quota
			delete_test_quotas
			delete_quota
			get_testdir_files
			stat_testdir_files
			get_quota_report
			delete_testdir_files
			show_unix_testdir_files
			quick_delete_testdir_files
			move_testdir_files
			delete_file_list
			truncate_testdir_files
			initialize_file_values
			initialize_quota_values
			combine_hashtables
			verify_testdirs_empty
			verify_initial_quota_state
			chown_testdir_files
			chgrp_testdir_files
			move_testdir_files
			write_quota_files
			get_rel_diff_percent
			search_ems_records
			verify_ems_quota_limit_aux
			verify_ems_quota_limit_exceeded
			verify_ems_quota_limit_normal
			verify_ems_quota_limit_unchanged
			create_users_and_groups
			create_qtrees_and_dirs
			calculate_sizes_and_limits
			cvt_quota_spec_to_query
			create_quota_specs
			scenario_quota_resize
			scenario_bad_quotas_file
			scenario_quota_exceeded_create
			scenario_quota_exceeded_move
			scenario_quota_exceeded_chown
			scenario_quota_exceeded_chgrp
			scenario_quota_disk_limit_exceeded_append
			scenario_quota_normal_delete
			scenario_user_quota_normal_chown
			scenario_group_quota_normal_chgrp
			scenario_quota_limit_normal_move
			scenario_quota_disk_limit_normal_truncate
			setup_cifs
		);
};



#------------------------------------------------------------------
#      Module Import.
#------------------------------------------------------------------
use Carp;
use Tharn;
use SPARSE::Util::Logger;
use SayWrap qw(sayEOF);
use Admin::License qw(license);
use FS::Util::Utilities qw( write_file cifs_chown_file cifs_file_getowner);
use FS::Util::Setup qw(mount_root_vol_if_needed create_vol_and_mount umount_vol common_cleanup);
use Utils::Units qw(:constants UNIT_B guess_unit convert_units convert_to_bytes strip_unit_spec);
use Volume qw( :df vol_disk_free vol_exists);
use QtreeCmds qw( qtreeCreate qtreePresent qtreeDelete qtreeSecurity );
use Errno qw(EDQUOT ENOSPC);
use filer qw( filer_version_get filer_cifs_stop );
use fileSystem qw(backupFiles restoreFiles restoreFilesPreserveBackups);
# use FileManip qw( ls_unix_file get_file_size test_file get_dir_entries make_dir remove_dir setattr_file remove_file stat_file );
use FileManip qw( get_file_size test_file get_dir_entries make_dir remove_dir setattr_file remove_file stat_file );
use CIFS::Setup qw(cifs_setup is_cifs_setup);
use ExportFS qw (share_add share_delete share_exists);
use TestEnv::MinWin qw(:all);
use MapShare qw(map_cifs_share unmap_cifs_share map_exists list_maps unmap_cifs_shares_for_server);
use Data::Dumper;

use Users qw( createUsers listCIFSGroups listCIFSUsers addCIFSUser findCIFSUser
		addCIFSGroup removeCIFSUser removeCIFSGroup 
		writeDefaultPasswdFile writeGroupFile );

use Quota qw(  enableQuota disableQuota setSingleQuota deleteQuota
	       mapToCIFSUser unmapToCIFSUser getQuotaReport writeQuotaFile
	       mapQuota unmapQuota addUsermap removeUsermap setMappedQuota
	       mapUsers unmapUser queryQuotaReport writeUsermapFile
	       pollEMSLogLines pollEMSLogRecords markEMSLog closeEMSLog
		getCIFSServer setCIFSUserQuota setQuotasForCIFSUsers );


#------------------------------------------------------------------
#        Global Vars
#------------------------------------------------------------------
BEGIN {
    use vars qw(

      $DISK_QUOTA_LIMIT
      $FILES_QUOTA_LIMIT
      $EMS_MSG_POLL_TIME

      $Fc
      $Fh
      $Vol
      $Mntpt
      $Rootvol_mntpt
      $nfs_client
      $cifs_client
      $cifs_passwd

      @All_users
      @Null_users
      @NFS_users
      @CIFS_users
      @CIFS_users_created
      @Mapped_users
      %User_hash
      %WinUser_hash
      %User_for_uid
      @Users_z
      @Users_a
      @Users_b
      @Users_c
      @Users_d
      @Users_e
      @Users_f

      @All_groups
      @Null_groups
      @NFS_groups
      @CIFS_groups
      @CIFS_groups_created
      @Mapped_groups
      %Group_hash
      %WinGroup_hash
      %Group_for_gid
      %Group_for_code
      $Group_z
      $Group_a
      $Group_b
      $Group_c
      $Group_d
      $Group_e
      $Group_f

      @All_qtrees
      @NFS_qtrees
      @CIFS_qtrees
      @Mixed_qtrees
      @Null_qtrees
      %Qtree_hash
      %Qtree_for_dir
      %Qtree_for_num
      $Qtree_0
      $Qtree_1
      $Qtree_2
      $Qtree_3
      $Qtree_4
      $Qtree_5
      $Qtree_6

      @All_dirs
      @NFS_dirs
      @CIFS_dirs
      @Mixed_dirs
      @Null_dirs
      %Dir_hash
      %Dir_for_code

      @All_quota_specs
      $Bad_quota_spec
      @Test_quotas

      %Client_for_proto
      @All_files
      @All_cifs_shares
      @All_cifs_drives
      @Base_files
      @Trigger_files
      %NFS_paths
      %CIFS_paths
      %Filer_paths
      $Orig_vol_security
      @Backup_files

      $Disk_quota_limit_kb
      $Files_quota_limit
      $Vol_avail_kb
      $Vol_avail_files

      $Default_file_set
      $File_set
      $Large_file_set
      $Small_file_set
      $Jumbo_file_set

      $cifs_client
      $WRITE_BUFFER_SIZE
      $DEBUG
      $MIN_FILE_SIZE
      );

    require "FS/commonParams.pl";
} ## end BEGIN



#------------------------------------------------------------------
#          Function Prototypes
#------------------------------------------------------------------

sub quota_ems_setup(%);
sub quota_ems_cleanup(%);
sub get_quota_spec($);
sub set_quota(%);
sub combine_hashtables(@);
sub delete_test_quotas();
sub delete_quota($$);
sub get_testdir_files(%);
sub stat_testdir_files(%);
sub get_quota_report($%);
sub delete_testdir_files(%);
sub show_unix_testdir_files(%);
sub quick_delete_testdir_files(%);
sub move_testdir_files(%);
sub delete_file_list(%);
sub truncate_testdir_files(%);
sub initialize_file_values(%);
sub initialize_quota_values(%);
sub combine_hashtables(@);
sub verify_testdirs_empty($%);
sub verify_initial_quota_state($%);
sub chown_testdir_files(%);
sub chgrp_testdir_files(%);
sub move_testdir_files(%);
sub write_quota_files(%);
sub get_rel_diff_percent($$);
sub search_ems_records(%);
sub verify_ems_quota_limit_aux(%);
sub verify_ems_quota_limit_exceeded(%);
sub verify_ems_quota_limit_normal(%);
sub verify_ems_quota_limit_unchanged(%);
sub create_users_and_groups($%);
sub create_qtrees_and_dirs(%);
sub calculate_sizes_and_limits(%);
sub cvt_quota_spec_to_query($);
sub create_quota_specs(%);
sub scenario_quota_resize(%);
sub scenario_bad_quotas_file(%);
sub scenario_quota_exceeded_create(%);

sub scenario_quota_exceeded_create(%);
sub scenario_quota_exceeded_move(%);
sub scenario_quota_exceeded_chown(%);
sub scenario_quota_exceeded_chgrp(%);

sub scenario_quota_normal_delete(%);
sub scenario_quota_disk_limit_exceeded_append(%);
sub scenario_quota_normal_delete(%);
sub scenario_user_quota_normal_chown(%);
sub scenario_group_quota_normal_chgrp(%);
sub scenario_quota_limit_normal_move(%);
sub scenario_quota_disk_limit_normal_truncate(%);
sub setup_cifs(%);



#------------------------------------------------------------------
#                Start Auxilliary Subroutines
#------------------------------------------------------------------


sub get_quota_spec($) {
	my ($index) = @_;
	return $All_quota_specs[$index];
}



sub set_quota(%) {
	my (%opts) = @_;
	my $Fc = $opts{Fc};
	my $fail_msg_list = [];
	logcomment("Enter set_quota");
	logDbg("Enter set_quota\n\n".Dumper(\%opts));
	eval {
            my ($msg, $fail_list, $rslt);

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};
	    my %quota_opts = %$quota_spec;
	    delete($quota_opts{Misc});
	    delete($quota_opts{Desc});

	    logcomment("set_quota  Line: ".__LINE__."    Call  'setSingleQuota' ....\n\n".Dumper(\%quota_opts));
	    setSingleQuota($Fc, %quota_opts)
	         or die("setSingleQuota: $@");
	    logcomment("set_quota  Line: ".__LINE__."\n\n.... Returned from 'setSingleQuota'");

	};

	my $fatal_msg = $@;
#        logcomment("Exit set_quota:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
        logcomment("Exit  set_quota\nErrMsg: \"$fatal_msg\"");
	return($fatal_msg);
}



sub get_quota_report($%)
{
	my ($Fc, %qr_query) = @_;
	my ($qr, $msg);
	logcomment("Enter get_quota_report");

	eval {
		my @qr_entries = queryQuotaReport(  $Fc,
						    flag          => 'sx',
						    match_debug   => 0,
						    %qr_query,
				               );
		if (scalar(@qr_entries) != 1) {
			$msg = "Line: ".__LINE__."   Expected exactly one quota report entry, found: ".
			           scalar(@qr_entries) . " matching criteria:\n\n".Dumper(\%qr_query);
		       logcomment($msg);
		       die($msg);
		}
		$qr = $qr_entries[0];
	};

	my $fatal_msg = $@;
#        logcomment("Exit get_quota_report:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
        logcomment("Exit  get_quota_report\nErrMsg: \"$fatal_msg\"");
	return($fatal_msg, $qr);
}



sub stat_testdir_files(%)
{
	my (%opts) = @_;
	logcomment("Enter stat_testdir_files");
	logDbg("Enter stat_testdir_files\n\n".Dumper(\%opts));
	eval {
		my ($msg, $dirpath, $client, $dir, @dir_list, $dirobj, $dirname, $rslt,
		     $file_list, $path_list_in);
		my ($proto) = $opts{proto};
		if (!$proto) {
			$msg = "stat_testdir_files  Line: ".__LINE__.
			           "No 'proto' option specified";
			logcomment($msg);
			die($msg);
		}

		if ($opts{dir}) {
			@dir_list = ( $opts{dir} );
			$dir = $dir_list[0];
		} elsif ( $opts{dir_list} ) {
			@dir_list = ( @{$opts{dir_list}} );
			$dir = $dir_list[0];
		} elsif ( $opts{Dirs} ) {
			@dir_list = ( @{$opts{Dirs}} );
			$dir = $dir_list[0];
		} else {
			$msg = "stat_testdir_files  Line: ".__LINE__.
				   "No 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		if (ref($dir)) {
			$dirobj = $dir;
			$dirname = $dirobj->{name};
		} else {
			$dirname = $dir;
			$dirobj = $Dir_hash{$dirname};
		}

		$file_list = $opts{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			}
		}

		if ($proto eq 'nfs') {
			$dirpath = $dirobj->{nfs_path};
			$client = $nfs_client;
		} elsif ($proto eq 'cifs') {
			$dirpath = $dirobj->{cifs_path};
			$client = $cifs_client;
		} else {
			$msg = "stat_testdir_files  Line: ".__LINE__.
			           "Invalid 'proto' given: \"$proto\"";
			logcomment($msg);
			die($msg);
		}
		logDbg("stat_testdir_files  Line: ".__LINE__."   Calling 'stat_dir_entries' ....");
		my (@fnames) = get_dir_entries(  dir	  => $dirpath,
			             		 client   => $client,
			          );
		if ($@) {
			$msg = "stat_testdir_files  Line: ".__LINE__.
			          "\nstat_dir_entries \"$dirpath\"\nfailed: $@";
			logcomment($msg);
			die($msg);
		}
		logDbg("stat_testdir_files  Line: ".__LINE__.
		          "\n'stat_testdir_files' returned entries:\n\n".Dumper(\@fnames));
		my @paths = map( {$dirpath . "/" . $_}  @fnames );

		my %all_attrs = ();
		foreach (@paths) {
			my %attrs = stat_file( file       => $_,
			                       client     => $client,
				               rtn_hash   => 1,
		                               user       => 'root'
					);
			if (!%attrs || $@) {
				$msg = "stat_testdir_files: Could not 'stat' file: \"$_\"";
				logcomment($msg);
				die($msg);
			}
			$all_attrs{$_} = { %attrs };
			if ($proto eq 'cifs') {
				my $owner = cifs_file_getowner(path=>$_, client=>$client);
				if ($@) {
					$msg = "stat_testdir_files: Could not 'stat' file: \"$_\"";
					logcomment($msg);
					die($msg);
				}
				$attrs{owner} = $owner;
			}
		}

		logDbg("stat_testdir_files  Line: ".__LINE__."\nAttrs:\n\n".Dumper(\%all_attrs));
		# return
	};

	my $fatal_msg = $@;
#        logcomment("Exit stat_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("Exit stat_testdir_files");
	return( $fatal_msg, undef, undef );
}



sub get_testdir_files(%)
{
	my (%opts) = @_;
	logcomment("Enter get_testdir_files");
	logDbg("Enter get_testdir_files\n\n".Dumper(\%opts));
	eval {
		my ($msg, $dirpath, $client, $dir, @dir_list, $dirobj, $dirname, $rslt);
		my ($proto) = $opts{proto};
		if (!$proto) {
			$msg = "get_testdir_files  Line: ".__LINE__.
			           "No 'proto' option specified";
			logcomment($msg);
			die($msg);
		}

		if ($opts{dir}) {
			@dir_list = ( $opts{dir} );
			$dir = $dir_list[0];
		} elsif ( $opts{dir_list} ) {
			@dir_list = ( @{$opts{dir_list}} );
			$dir = $dir_list[0];
		} elsif ( $opts{Dirs} ) {
			@dir_list = ( @{$opts{Dirs}} );
			$dir = $dir_list[0];
		} else {
			$msg = "get_testdir_files  Line: ".__LINE__.
				   "No 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		if (ref($dir)) {
			$dirobj = $dir;
			$dirname = $dirobj->{name};
		} else {
			$dirname = $dir;
			$dirobj = $Dir_hash{$dirname};
		}

		if ($proto eq 'nfs') {
			$dirpath = $dirobj->{nfs_path};
			$client = $nfs_client;
		} elsif ($proto eq 'cifs') {
			$dirpath = $dirobj->{cifs_path};
			$client = $cifs_client;
		} else {
			$msg = "get_testdir_files  Line: ".__LINE__.
			           "Invalid 'proto' given: \"$proto\"";
			logcomment($msg);
			die($msg);
		}
		logDbg("get_testdir_files  Line: ".__LINE__."   Calling 'get_dir_entries' ....");
		my (@fnames) = get_dir_entries(  dir	  => $dirpath,
			             		 client   => $client,
			          );
		if ($@) {
			$msg = "get_testdir_files  Line: ".__LINE__.
			          "\nget_dir_entries \"$dirpath\"\nfailed: $@";
			logcomment($msg);
			die($msg);
		}
		logDbg("get_testdir_files  Line: ".__LINE__."\n'get_testdir_files' returned entries:\n\n".Dumper(\@fnames));
		my @paths = map( {$dirpath . "/" . $_}  @fnames );
		logDbg("get_testdir_files  Line: ".__LINE__."\nReturning file_names:\n\n".
		             Dumper(\@fnames)."\n\nPaths:\n\n".Dumper(\@paths));
		return( "", [ @fnames ], [ @paths ] );
	};

	my $fatal_msg = $@;
#        logcomment("Exit get_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("Exit get_testdir_files");
	return( $fatal_msg, undef, undef );
}



sub delete_path_list(%)
{
	logcomment("Enter delete_path_list");
	my (%opts) = @_;
	eval {
		my ($msg, $client, $rslt);
		my ($proto) = $opts{proto};
		if (!$proto) {
			$msg = "delete_path_list  Line: ".__LINE__.
			           "No 'proto' option specified";
			logcomment($msg);
			die($msg);
		}
		my ($path_list) = $opts{path_list};
		if (!$path_list) {
			$msg = "delete_path_list  Line: ".__LINE__.
			           "No 'path_list' option specified";
			logcomment($msg);
			die($msg);
		}

		if ($proto eq 'nfs') {
			$client = $nfs_client;
		} elsif ($proto eq 'cifs') {
			$client = $cifs_client;
		} else {
			$msg = "delete_path_list  Line: ".__LINE__.
			           "Invalid 'proto' given: \"$proto\"";
			logcomment($msg);
			die($msg);
		}

		for my $path ( @$path_list ) {
			if (!test_file(file=>$path, test=>'f', client=>$client)) {
				$msg = "Line: ".__LINE__."\nFile \"$path\" not found";
				logcomment($msg);
				die($msg);
			}
			if ($proto eq 'nfs') {
				$rslt = remove_file( file              => $path,
				                     client            => $client,
				                     user              => 'root',
				          );
			} else {
				$rslt = remove_file( file              => $path,
				                     client            => $client,
				                     # user              => 'administrator',
				          );
			}
			if (!$rslt || $@) {
				$msg = "Line: ".__LINE__."\nremove_file \"$path\"  failed: $@";
				logcomment($msg);
				die($msg);
			}
		}
    };
    my $fatal_msg = $@;
#        logcomment("Exit delete_path_list:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
    logcomment("Exit  delete_path_list\nErrMsg: \"$fatal_msg\"");
    return($fatal_msg);
}



sub write_quota_files(%)
{
     my (%opts) = @_;
     my $Fc    = $opts{Fc};
     my $Vol   = $opts{vol};

     logcomment("Enter write_quota_files");
     logDbg("Enter write_quota_files\n\n".Dumper(\%opts));
     my $verify_files_written = 1;
     my ( $msg );
     eval {
	     my $stat_hash;
	     my $file_list = $opts{file_list} || [];
	     my $path_list_in = undef;
	     my $path_list_out = undef;

	     if (!$file_list) {
	     } elsif (ref($file_list) =~ /hash/i) {
		     my @keys = keys(%$file_list);
		     if (@keys) {
			     $path_list_in = \@keys;
		     } else {
			     $path_list_out = $file_list;
		     }
	     } elsif (ref($file_list) =~ /array/i) {
		     if (@$file_list) {
			     $path_list_in = $file_list;
		     } else {
		             $path_list_out = $file_list;
		     }
	     }
	     logDbg("write_quota_files  Line: ".__LINE__.
		       "\n\nPath list IN:\n".(defined($path_list_in)?Dumper($path_list_in):"UNDEF").
		       "\n\nPath list OUT:\n".(defined($path_list_out)?Dumper($path_list_out):"UNDEF"));

	     my $dir;
	     my @dir_list;
	     if ($opts{dir}) {
	            $dir  = $opts{dir};
		    @dir_list = ( $dir );
	     } elsif ($opts{dir_list}) {
		    @dir_list = @{$opts{dir_list}};
		    $dir  = $dir_list[0];
	     } else {
		    $msg = "write_quota_files  Line: ".__LINE__."\nNo 'dir' or 'dir_list' option given";
		    logcomment($msg);
		    die($msg);
	     }

#	     if (!$opts{initialize_quota_valus_called}) {
#		    $msg = "write_quota_files  Line: ".__LINE__.i
#		           "\n'initialize_quota_values' must be called ".
#		           "before calling 'write_quota_files'";
#		    logcomment($msg);
#		    die($msg);
#	     }

	     my ($disk, $files, $d_or_f) = ($opts{disk}, $opts{files}, $opts{d_or_f});
	     if ((!$disk && !$files) || !$d_or_f) {
		$msg = "write_quota_files  Line: ".__LINE__."\n".
		       "Must specify one of 'disk', 'files', 'd_or_f'";
		logcomment($msg);
		die($msg);
	     }
	     my %file_opts = ( %opts );
	     foreach ( qw( d_or_f disk files ) ) {
		delete($file_opts{$_});
	     }

#	     if (!$opts{initialize_file_valus_called}) {
#		    $msg = "write_quota_files  Line: ".__LINE__.i
#		           "\n'initialize_file_values' must be called ".
#		           "before calling 'write_quota_files'";
#		    logcomment($msg);
#		    die($msg);
#	     }

	     my $file_set = $opts{file_set};
	     my ($size_bytes, $size_kb);
	     if ($opts{size_bytes}) {
		$size_bytes = $opts{size_bytes};
		$size_kb = int($size_bytes/1024);
	     } elsif ($opts{size}) {
		my $size = $opts{size};
		$size_bytes = convert_to_bytes($size);
		$size_kb = int($size_bytes/1024);
	     } elsif ($opts{size_kb}) {
		$size_kb = $opts{size_kb};
		$size_bytes = 1024 * $size_kb;
	     } elsif ($file_set) {
	        $size_kb = $file_set->{size_kb};
		$size_bytes = 1024 * $size_kb;
	     }
	     if ($size_bytes < $MIN_FILE_SIZE) {
		$size_bytes = $MIN_FILE_SIZE;
		$size_kb = int($size_bytes/1024);
	     }
	     my $num_files          = $opts{num_files};
	     my $file_prefix        = $opts{file_prefix};
             my $expected_alloc_kb  = $opts{expected_alloc_kb};
	     logDbg("write_quota_files  Line: ".__LINE__.
		      "\n\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "File prefix: \"$file_prefix\"\nExpeted alloc: $expected_alloc_kb Kb");


	     my ($client, $proto, @Protos);
	     if ($opts{proto}) {
	            $proto  = $opts{proto};
		    @Protos = ( $proto );
	     } elsif ($opts{Protos}) {
		    @Protos = @{$opts{Protos}};
		    $proto = $Protos[0];
	     } else {
		    $msg = "write_quota_files  Line: ".__LINE__.
			      "\nNo 'proto' or 'Protos' option given";
		    logcomment($msg);
		    die($msg);
	     }

	     my $user;
	     my @Users;
	     if ($opts{user}) {
	            $user  = $opts{user};
		    @Users = ( $user );
	     } elsif ($opts{Users}) {
		    @Users = @{$opts{Users}};
		    $user  = $Users[0];
	     } else {
		    $user = 'root';
		    @Users = ( $user );
	     }

	      my $userobj = $User_hash{$user};
	      my $login = $user;
#	      my $win_login = $userobj->{win_login};
#	      my $login = $userobj->{login};
	      my $uid = $userobj->{uid};
	      my $gid = $userobj->{gid};

	     logcomment("write_quota_files  Line: ".__LINE__.
			   "\n\ndir: \"$dir\"\ndir_list: \"@dir_list\"\n\"".
		           "user: \"$user\"\nUsers: \"@Users\"");

	     my $dirname = $dir;
	     my $dirobj = $Dir_hash{$dirname};
	     my $nfs_dirpath = $dirobj->{nfs_path};
	     my $cifs_dirpath = $dirobj->{cifs_path};

	     if ($proto eq 'cifs') {
		     #
		     #  CIFS/Windows client
		     #

		     $msg = "write_quota_files  Line: ".__LINE__.
			        "   Using CIFS protocol";
		     logcomment($msg);

		     $client = $cifs_client;
		     if (!$CIFS_paths{$dir}) {
			    $msg = "write_quota_files  Line: ".__LINE__.
			               "\nInvalid dir: \"$dir\"";
			    logcomment($msg);
			    die($msg);
		     }
	
			#  Test 'disk' quota, use files
			logcomment("write_quota_files  Line: ".__LINE__."   via CIFS\n\n".
				     "Client: \"$client\"\nNumfiles: $num_files\n".
		                     "Files_quota_limit: \"$Files_quota_limit\"\n".
				     "Disk_quota_limit_kb: \"$Disk_quota_limit_kb\"\n".
				     "Expected_alloc_kb: \"$expected_alloc_kb\"\n".
				     "Vol_avail_kb: $Vol_avail_kb\nVol_avail_files: ".
				     "$Vol_avail_files");
	
			logcomment("write_quota_files  Line: ".__LINE__."\n\n".
				     "Client: \"$client\"\nNumfiles: $num_files\n".
				     "Size: $size_bytes (bytes)");
	
		       ## @step Create set of files, to trigger quota
		       logStep("Create set of files, to trigger quota");

			my $dirpath = $dirobj->{cifs_path};
			my @dir_entries = get_dir_entries( dir=>$dirpath,
	                                                   client=>$client );
			my $dir_path_list =  [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
		        my $local_file_list = [];
			if (!@$dir_path_list && !$path_list_in) {
			       #  We are 'creating' new files, rather than 'overwriting' or 'appending' existing files.
			       my $path = $cifs_dirpath . '/' . $file_prefix;
			       $stat_hash =
				    FS::Util::Utilities::write_file(
						  client          => $cifs_client,
						  numfiles        => $num_files,
						  size            => $size_bytes,
						  path	   	  => $path,
						  file_list	  => $local_file_list,
						  fstat           => 1,
						  buffer_size     => $WRITE_BUFFER_SIZE,
						  dont_overwrite  => 0,
					 );
				if (!$stat_hash || ref($stat_hash) !~ /HASH/) {
					$msg = "write_quota_files  Line: ".__LINE__.
						     "\nFS::Util::Utilities::write_file:\n$@";
					logcomment($msg);
					die($msg);
				}

				$msg = "write_quota_files  Line: ".__LINE__.
					   "\nwrite_file returned:\n\n".Dumper($stat_hash);
				logcomment($msg);
				my @files_stat = keys(%$stat_hash);
				my $num_files_stat = scalar(@files_stat);
				if ($num_files_stat != $num_files) {
					$msg = "write_quota_files  Line: ".__LINE__.
					           "\nExpected to write $num_files files\n".
					           "Wrote: $num_files_stat";
					logcomment($msg);
					die($msg);
				}
				my $num_files_list = scalar(@$local_file_list);
				if ($num_files_list != $num_files) {
					$msg = "write_quota_files  Line: ".__LINE__.
					           "\nExpected to write $num_files files\n".
					           "Wrote: $num_files_list";
					logcomment($msg);
					die($msg);
				}
			} else {   #if (!@$dir_path_list && !$path_list_in)
				if (!@$dir_path_list && $path_list_in) {
					$dir_path_list = $path_list_in;
					$path_list_in = undef;
				}
				if (!@$dir_path_list) {
					# die();
				}
				foreach my $srcpath ( @$dir_path_list ) {
					$local_file_list = [];
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
				        my $local_stat_hash =
					    FS::Util::Utilities::write_file(
							  client          => $cifs_client,
							  numfiles        => 1,
							  size            => $size_bytes,
							  path	     	  => $srcpath,
							  file_list	  => $local_file_list,
							  fstat           => 1,
							  buffer_size     => $WRITE_BUFFER_SIZE,
							  dont_overwrite  => 0,
						 );
					if (!$local_stat_hash || ref($local_stat_hash) !~ /HASH/) {
						$msg = "write_quota_files  Line: ".__LINE__.
							     "\nFS::Util::Utilities::write_file:\n$@";
						logcomment($msg);
						die($msg);
					}
					$msg = "write_quota_files  Line: ".__LINE__.
						   "\nwrite_file returned:\n\n".Dumper($local_stat_hash);
					logcomment($msg);
					my @files_stat = keys(%$local_stat_hash);
					my $num_files_stat = scalar(@files_stat);
					if ($num_files_stat != 1) {
						$msg = "write_quota_files  Line: ".__LINE__.
						           "\nExpected to write 1 file\n".
						           "Wrote: $num_files_stat\n\nStat Hash:\n\n".Dumper($local_stat_hash);
						logcomment($msg);
						die($msg);
					}
					my $file = $files_stat[0];
					$stat_hash->{$file} = $stat_hash;
					my $num_files_list = scalar(@$local_file_list);
					if ($num_files_list != 1) {
						$msg = "write_quota_files  Line: ".__LINE__.
						           "\nExpected to write $num_files files\n".
						           "Wrote: $num_files_list\n\nFiles List:\n\n".Dumper($local_file_list);
						logcomment($msg);
						die($msg);
					}
				}
			}
			if ($path_list_out) {
				if (ref($path_list_out) =~ m/array/i) {
					@$path_list_out = @$local_file_list;
				} elsif (ref($path_list_out) =~ m/hash/i) {
					map( { $path_list_out->{$_} = {} }  @$local_file_list );
				}
			}
			sleep(5);

			if ($verify_files_written) {
				logcomment("write_quota_files  Line: ".__LINE__."\nVerifying files just written ....");
				@dir_entries = get_dir_entries( dir=>$cifs_dirpath,
				                                   client=>$cifs_client);
				my $num_entries = scalar(@dir_entries);
				if ($num_files != $num_entries) {
					$msg = "write_quota_files  Line: ".__LINE__.
						  "\nExpected to write $num_files files\n".
						  "Actually wrote: $num_entries files\n\nDir entries:\n\n".Dumper(\@dir_entries);
					die($msg);
				}
				foreach my $f (@$local_file_list) {
					my $sz = get_file_size(file=>$f, client=>$cifs_client);
					if ($sz != $size_bytes) {
						$msg = "write_quota_files  Line: ".__LINE__.
						            "\nFile \"$f\" size $sz does not match ".
						            "expected size: $size_bytes (bytes)";
						logcomment($msg);
#						$Log->warn($msg);
						die($msg);
					}
				}
				logcomment("write_quota_files  Line: ".__LINE__."\n.... Finished verifying files just written ....");
			}
			$msg = chown_testdir_files( dir           => $dirname,
				                    proto         => 'cifs',
			                            file_list     => $local_file_list,
			                            # dest_user     => $login,
			                            dest_user     => $login
			             );
			if ($msg) {
				$msg = "write_quota_files  Line: ".__LINE__.
				          "\n'chown_testdir_files' failed: $msg";
				logcomment($msg);
				die($msg);
			}
			foreach (@$local_file_list) {
				my $owner = cifs_file_getowner(path=>$_, client=>$client);
				if ($@) {
					$msg = "write_quota_files: Could not 'stat' file: \"$_\"";
					logcomment($msg);
					die($msg);
				}
				logDbg("write_quota_files  Line: ".__LINE__.
				          "\n(CIFS) File: \"$_\"\nOwner: \"$owner\"");
			}

		} else {
		#  $proto eq 'nfs'

		     #
		     #  UNIX/NFS client
		     #

		     $msg = "write_quota_files  Line: ".__LINE__.
			        "   Using NFS protocol";
		     logcomment($msg);

		     $client = $nfs_client;
		     if  (!$NFS_paths{$dir}) {
			    $msg = "write_quota_files  Line: ".__LINE__.
				        "\nInvalid dir: \"$dir\"";
			    die($msg);
		     }
			#  Test 'disk' quota, use files
			logcomment("write_quota_files  Line: ".__LINE__."   via NFS\n\n".
				     "Client: \"$client\"\nNumfiles: $num_files\n".
		                     "Files_quota_limit: \"$Files_quota_limit\"\n".
				     "Disk_quota_limit_kb: \"$Disk_quota_limit_kb\"\n".
				     "Expected Alloc kb: \"$expected_alloc_kb\"\n".
				     "Vol_avail_kb: $Vol_avail_kb\nVol_avail_files: ".
				     "$Vol_avail_files");
	
			logcomment("write_quota_files  Line: ".__LINE__."\n\n".
				     "Client: \"$client\"\nNumfiles: $num_files\n".
				     "Size: $size_bytes (bytes)\nUid: \"$uid\"\nGid: ".
				     "\"$gid\"");
	
		       ## @step Create set of  files, to trigger quota
		       logStep("Create set of files, to trigger quota");
			my $dirpath = $dirobj->{nfs_path};
			my @dir_entries = get_dir_entries( dir=>$dirpath,
	                                                   client=>$client );
			my $dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
		        my $local_file_list = [];
			if (!@$dir_path_list && !$path_list_in) {
			       #  We are 'creating' new files, rather than 'overwriting' or 'appending' existing files.
		     	       my $path = $nfs_dirpath . '/' . $file_prefix;
			       $stat_hash =
				    FS::Util::Utilities::write_file(
					    	  client          => $nfs_client,
						  numfiles        => $num_files,
						  size            => $size_bytes,
						  path	   	  => $path,
					    	  user            => 'root',
					    	  uid		  => $uid,
					     	  gid		  => $gid,
						  file_list	  => $local_file_list,
						  fstat           => 1,
						  buffer_size     => $WRITE_BUFFER_SIZE,
						  dont_overwrite  => 0,
					 );
				if (!$stat_hash || ref($stat_hash) !~ /HASH/) {
					$msg = "write_quota_files  Line: ".__LINE__.
						     "\nFS::Util::Utilities::write_file:\n$@";
					logcomment($msg);
					die($msg);
				}


###########################################################################################################################
logcomment("write_quota_files  Line: ".__LINE__."     Calling get_testdir_files\nDir: \"$dirname\"");
stat_testdir_files(proto=>'nfs', dir=>$dirname);
logcomment("write_quota_files  Line: ".__LINE__."     Returned from get_testdir_files\n\n");
###########################################################################################################################


				$msg = "write_quota_files  Line: ".__LINE__.
					   "\nwrite_file returned:\n\n".Dumper($stat_hash);
				logcomment($msg);
				my @files_stat = keys(%$stat_hash);
				my $num_files_stat = scalar(@files_stat);
				if ($num_files_stat != $num_files) {
					$msg = "write_quota_files  Line: ".__LINE__.
					           "\nExpected to write $num_files files\n".
					           "Wrote: $num_files_stat";
					logcomment($msg);
					die($msg);
				}
				my $num_files_list = scalar(@$local_file_list);
				if ($num_files_list != $num_files) {
					$msg = "write_quota_files  Line: ".__LINE__.
					           "\nExpected to write $num_files files\n".
					           "Wrote: $num_files_list";
					logcomment($msg);
					die($msg);
				}
			} else {
				if (!@$dir_path_list && $path_list_in) {
					$dir_path_list = $path_list_in;
					$path_list_in = undef;
				}
				if (!@$dir_path_list) {
					# die();
				}
				my $stat_hash = {};
				foreach my $srcpath ( @$dir_path_list ) {
					$local_file_list = [];
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
				        my $local_stat_hash =
					    FS::Util::Utilities::write_file(
					    	  	  client          => $nfs_client,
							  numfiles        => 1,
							  size            => $size_bytes,
							  path	     	  => $srcpath,
					    	  	  user            => 'root',
					    	  	  uid		  => $uid,
					     	  	  gid		  => $gid,
							  file_list	  => $local_file_list,
							  fstat           => 1,
							  buffer_size     => $WRITE_BUFFER_SIZE,
							  dont_overwrite  => 0,
						 );
					if (!$local_stat_hash || ref($local_stat_hash) !~ /HASH/) {
						$msg = "write_quota_files  Line: ".__LINE__.
							     "\nFS::Util::Utilities::write_file:\n$@";
						logcomment($msg);
						die($msg);
					}
					$msg = "write_quota_files  Line: ".__LINE__.
						   "\nwrite_file returned:\n\n".Dumper($local_stat_hash);
					logcomment($msg);
					my @files_stat = keys(%$local_stat_hash);
					my $num_files_stat = scalar(@files_stat);
					if ($num_files_stat != 1) {
						$msg = "write_quota_files  Line: ".__LINE__.
						           "\nExpected to write 1 file\n".
						           "Wrote: $num_files_stat\n\nStat Hash:\n\n".Dumper($local_stat_hash);
						logcomment($msg);
						die($msg);
					}
					my $file = $files_stat[0];
					$stat_hash->{$file} = $stat_hash;
					my $num_files_list = scalar(@$local_file_list);
					if ($num_files_list != 1) {
						$msg = "write_quota_files  Line: ".__LINE__.
						           "\nExpected to write $num_files files\n".
						           "Wrote: $num_files_list\n\nFiles List:\n\n".Dumper($local_file_list);
						logcomment($msg);
						die($msg);
					}
				}
			}
			if ($path_list_out) {
				if (ref($path_list_out) =~ /array/i) {
					@$path_list_out = @$local_file_list;
				} elsif (ref($path_list_out) =~ /hash/i) {
					map( { $path_list_out->{$_} = {} }  @$local_file_list );
				}
			}
			sleep(5);

logcomment("write_quota_files  Line: ".__LINE__);   ########################

			if ($verify_files_written) {
				logcomment("write_quota_files  Line: ".__LINE__."\nVerifying files just written ....");
				@dir_entries = get_dir_entries( dir=>$nfs_dirpath,
								   client=>$nfs_client);
				my $num_entries = scalar(@dir_entries);
				if ($num_files != $num_entries) {
					$msg = "write_quota_files  Line: ".__LINE__.
						  "\nExpected to write $num_files files\n".
						  "Actually wrote: $num_entries files\n\nDir entries:\n\n".Dumper(\@dir_entries);
					die($msg);
				}

				foreach my $f (@$local_file_list) {
					my $sz = get_file_size(file=>$f, client=>$nfs_client);
					if ($sz != $size_bytes) {
						$msg = "write_quota_files  Line: ".__LINE__.
						            "\nFile \"$f\" size $sz does not match ".
						            "expected size: $size_bytes";
						logcomment($msg);
#						$Log->warn($msg);
						die($msg);
					}
				}
				logcomment("write_quota_files  Line: ".__LINE__."\n.... Finished verifying files just written ....");
			}
		}
        };

        my $fatal_msg = $@;
#        logcomment("Exit write_quota_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
        logcomment("Exit  write_quota_files\nErrMsg: ".  "\"$fatal_msg\"");
        return ($fatal_msg);
}



sub search_ems_records(%)
{
	my (%opts) = @_;
	my $Fc = $opts{Fc};
        logcomment("Enter ems_records");
        logDbg("Enter search_ems_records:\n\n".Dumper(\%opts));
	my @ems_records;
	if ($opts{records}) {
		@ems_records = @{$opts{records}};
	} else {
		@ems_records = pollEMSLogRecords($Fc);
	}
	my $patt = $opts{patt};
	my @matched_recs = grep( {$_->{event}->{event_type} =~ /$patt/} @ems_records );
        logDbg("Exit search_ems_records:  Returning:\n\n".Dumper(\@matched_recs));
	return([ @matched_recs ]);
}



sub verify_ems_quota_limit_aux(%)
{
	my (%opts) = @_;
	my $fail_msg_list = [];
	my $hash_out;

        logcomment("Enter verify_ems_quota_limit_aux");
        logDbg("Enter verify_ems_quota_limit_aux:\n\n".Dumper(\%opts));
	eval {
		my ( $msg, $fail_list );
		my @expected_options = grep( {$_ =~ /^expected_.*_count$/}  keys(%opts));
		my $msg_in = ($opts{msg_in} || "");
		my $Fc = $opts{Fc}
	                 or die("verify_ems_quota_limit_aux: No 'filer' option given");

		my ($hard, $soft, $h_or_s);
	        if ($opts{hard} || (defined($opts{h_or_s}) && $opts{h_or_s} eq 'hard')) {
		    $h_or_s = 'hard';
		    $hard = 1;
		    $soft = 0;
	        } elsif ($opts{soft} || (defined($opts{h_or_s}) && $opts{h_or_s} eq 'soft')) {
		    $h_or_s = 'soft';
		    $hard = 0;
		    $soft = 1;
	        } else {
		     $msg = "FATAL: verify_ems_quota_limit_aux  Line: ".__LINE__."\n\n".
		                 "Must specify one of 'hard', 'soft', 'd_or_f'";
	       	     logcomment($msg);
		     die($msg);
                }

		my ($disk, $files, $d_or_f);
	        if ($opts{disk} || (defined($opts{d_or_f}) && $opts{d_or_f} eq 'disk')) {
		    $d_or_f = 'disk';
		    $disk = 1;
		    $files = 0;
	        } elsif ($opts{files} || (defined($opts{d_or_f}) && $opts{d_or_f} eq 'files')) {
		    $d_or_f = 'files';
		    $disk = 0;
		    $files = 1;
	        } else {
		     $msg = "FATAL: verify_ems_quota_limit_aux  Line: ".__LINE__."\n\n".
		                 "Must specify one of 'disk', 'files', 'd_or_f'";
	       	     logcomment($msg);
		     die($msg);
                }

	        logcomment("Enter verify_ems_quota_limit_aux");
	        logDbg("Enter verify_ems_quota_limit_aux  Line: ".__LINE__.
	                  "\n\nMsg In: $msg_in\n\n".
			  "Hard: $hard\nSoft: $soft\nH_or_S: \"$h_or_s\"\n".
			  "Disk: $disk\nFiles: $files\nD_or_F: \"$d_or_f\"\n\n".
			  "'Expected' options: @expected_options\n".
			  "Opts:\n\n".Dumper(\%opts));

		unless ($hard || $soft || $h_or_s) {
			$msg = "verify_ems_quota_limit_aux  Line: ".__LINE__.
			          "\n\nMust specify one of 'hard', 'soft'";
			logcomment($msg);
			die($msg);
		}
		unless ($disk || $files || $d_or_f) {
			$msg = "verify_ems_quota_limit_aux  Line: ".__LINE__.
			          "\n\nMust specify one of 'disk', 'files'";
			logcomment($msg);
			die($msg);
		}

		my $quota_patt            = 'quota_';
		my $normal_patt           = 'quota_'  . 'normal';
		my $exceeded_patt         = 'quota_'  . 'exceeded';
		my $h_or_s_normal_patt    = 'quota_' . $h_or_s . 'limit_normal';
		my $h_or_s_exceeded_patt  = 'quota_' . $h_or_s . 'limit_exceeded';

		my @ems_records = pollEMSLogRecords($Fc);

		my @quota_recs =
		    grep( {$_->{event}->{event_type} =~   /$quota_patt/}              @ems_records );
		my @normal_recs =
		    grep( {$_->{event}->{event_type} =~   /$normal_patt/}             @quota_recs );
		my @exceeded_recs =
		    grep( {$_->{event}->{event_type} =~   /$exceeded_patt/}           @quota_recs );
		my @h_or_s_normal_recs =
		    grep( {$_->{event}->{event_type} =~   /$h_or_s_normal_patt/}      @quota_recs );
		my @h_or_s_exceeded_recs =
		    grep( {$_->{event}->{event_type} =~   /$h_or_s_exceeded_patt/}    @quota_recs );

		my $ems_count               = scalar(@ems_records);
		my $quota_count             = scalar(@quota_recs);
		my $normal_count            = scalar(@normal_recs);
		my $exceeded_count          = scalar(@exceeded_recs);
		my $h_or_s_normal_count     = scalar(@h_or_s_normal_recs);
		my $h_or_s_exceeded_count   = scalar(@h_or_s_exceeded_recs);

		$hash_out = {
				        ems_recs			=> [ @ems_records ],
					ems_count			=> $ems_count,
					quota_patt			=> $quota_patt,
					quota_count			=> $quota_count,
					quota_recs			=> [ @quota_recs ],
					normal_patt			=> $normal_patt,
					normal_count			=> $normal_count,
					normal_recs			=> [ @normal_recs ],
					exceeded_count			=> $exceeded_count,
					exceeded_recs			=> [ @exceeded_recs ],
					exceeded_patt			=> $exceeded_patt,
					h_or_s_normal_count		=> $h_or_s_normal_count,
					h_or_s_normal_recs		=> [ @h_or_s_normal_recs ],
					h_or_s_normal_patt		=> $h_or_s_normal_patt,
					h_or_s_exceeded_count		=> $h_or_s_exceeded_count,
					h_or_s_exceeded_recs		=> [ @h_or_s_exceeded_recs ],
					h_or_s_exceeded_patt		=> $h_or_s_exceeded_patt,
				};

#		logcomment("verify_ems_quota_limit_aux  Line: ".__LINE__."\n\nAll EMS Records:\n\n".
#		              Dumper(\@ems_records));
#		logcomment("verify_ems_quota_limit_aux  Line: ".__LINE__.
#			       "\n\nEMS Messages  'h_or_s_exceeded': ($h_or_s_exceeded_count)\n\n" .
#		               Dumper(\@h_or_s_exceeded_ems).
#			       "\n\nEMS Messages  'exceeded': ($exceeded_count)\n\n" .
#		               Dumper(\@exceeded_ems).
#			       "\n\nEMS Messages  'h_or_s_normal': ($h_or_s_normal_count)\n\n" .
#		               Dumper(\@h_or_s_normal_ems).
#			       "\n\nEMS Messages  'normal': ($normal_count)\n\n" .
#		               Dumper(\@normal_ems));

		foreach my $expected_key (@expected_options) {
			my $actual_key = $expected_key;
			$actual_key =~ s/^expected_//;
			my $match = 0;
			if (defined($hash_out->{$actual_key})) {
				logDbg("verify_ems_quota_limit_aux  Line: ".__LINE__.
				         "\nExpected value supplied for '$actual_key'.\n".
				         "Will compare 'expected' and 'actual' for this quantity");
				my $actual_val = $hash_out->{$actual_key};
				my $expected_val = $hash_out->{expected_val};
				my $rel_margin_key = $actual_key . "rel_margin";
				my $abs_margin_key = $actual_key . "abs_margin";
				my $rel_margin = $opts{$rel_margin_key} || 0;
				my $abs_margin = $opts{$abs_margin_key} || 0;
				if ($opts{check_upper_bound_only} && ($actual_val <= $expected_val)) {
						$match = 1;
				} else {
					if (abs($actual_val - $expected_val) <= $abs_margin) {
						$match = 1;
					} elsif ($expected_val != 0.0 && (abs($actual_val - $expected_val)/$expected_val) <= $rel_margin) {
						$match = 1;
					} else {
						$msg = "${msg_in}:\nverify_ems_quota_limit_aux:\n".
						       "Quantity: '$actual_key\n".
						       "Expected value: $expected_val\n".
						       "Actual value: $actual_val";
					       	logcomment($msg);
						push(@$fail_msg_list, $msg);
					}
				}
			} else {
				$msg = "verify_ems_quota_limit_aux  Line: ".__LINE__."\nCould not pair ".
				             "'Expected value' '$expected_key' with any 'actual' key";
				logDbg($msg);
				logcomment($msg);
				die($msg);
			}
		}

#		@other_quota_ems = list_difference(\@quota_ems, \@normal_ems);

		logDbg("verify_ems_quota_linmit_aux  Line: ".__LINE__."\nTime: ".`date`.
		          "\n\nReturning:\n\nHash Out:\n\n".Dumper($hash_out));
	};

        my $fatal_msg = $@;
        logcomment("Exit verify_ems_quota_limit_aux:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	if ($fatal_msg) {
	        logcomment("Exit verify_ems_quota_limit_aux:\n\nMSG: \"$fatal_msg\"");
		return($fail_msg_list, $fatal_msg, undef);
	} else {
		return($fail_msg_list, "", $hash_out);
	}
}


sub verify_ems_quota_limit_exceeded(%)
{
	my $fail_msg_list = [];

	my (%opts) = @_;
        logcomment("Enter verify_ems_quota_limit_exceeded");
        logDbg("Enter verify_ems_quota_limit_exceeded:\n\n".Dumper(\%opts));
	eval {
		my $msg_in = $opts{msg_in} || "";
		my $Fc = $opts{Fc}
	                 or die("verify_ems_quota_limit_exceeded: No 'filer' option given");
		my $hard = ($opts{hard} || 0);
		my $soft = ($opts{soft} || 0);
		my $h_or_s = ($opts{h_or_s} || 0);
		my $disk = ($opts{disk} || 0);
		my $files = ($opts{files} || 0);
		my $d_or_f = ($opts{d_or_f} || 0);

		my ( $fail_list, $msg, $err, $ems_hash );
	        logcomment("Enter verify_ems_quota_limit_exceeded");
	        logDbg("Enter verify_ems_quota_limit_exceeded  Line: ".__LINE__.
	                  "\n\nMsg In: $msg_in\n\n".
			  "Hard: $hard\nSoft: $soft\nH_or_S: \"$h_or_s\"\n".
			  "Disk: $disk\nFiles: $files\nD_or_F: \"$d_or_f\"\n\n".
			  "Opts:\n\n".Dumper(\%opts));

		unless ($hard || $soft || $h_or_s) {
			$msg = "verify_ems_quota_limit_exceeded  Line: ".__LINE__.
			           "\n\nMust specify one of 'hard', 'soft'";
			logcomment($msg);
			die($msg);
		}
		unless ($disk || $files || $d_or_f) {
			$msg = "verify_ems_quota_limit_exceeded  Line: ".__LINE__.
			           "\n\nMust specify one of 'disk', 'files'";
			logcomment($msg);
			die($msg);
		}

		my $rec_count;
		if (defined($opts{rec_count})) {
			$rec_count = $opts{rec_count};
		} else {
			$rec_count = 1;
		}

		logDbg("verify_ems_quota_limit_exceeded  Line: ".__LINE__."   Calling verify_quota_limit_aux ....");
		($fail_list, $err, $ems_hash) =
		     verify_ems_quota_limit_aux(  msg_in   => $msg_in, Fc => $Fc, %opts);
		if ($err) {
			$msg = "verify_ems_quota_limit_exceeded  Line: ".__LINE__.
			           "\nError from 'verify_ems_quota_limit_aux':\n$msg";
			logcomment($msg);
			die($msg);
		}
		logDbg("verify_ems_quota_limit_exceeded  Line: ".__LINE__."   .... Returned from verify_quota_limit_aux".
		         "\n\nFail cnt: \"$fail_list\"\nErr: \"$err\"\n\nEMS Hash:\n\n".Dumper($ems_hash));
		push(@$fail_msg_list, @$fail_list);

		my $quota_count             = $ems_hash->{quota_count};
		my $normal_count            = $ems_hash->{normal_count};
		my $exceeded_count          = $ems_hash->{exceeded_count};
		my $h_or_s_normal_count     = $ems_hash->{h_or_s_normal_count};
		my $h_or_s_exceeded_count   = $ems_hash->{h_or_s_exceeded_count};

		logDbg("verify_ems_quota_limit_exceeded  Line: ".__LINE__.
			"\n\nQuota Count:\t$quota_count\n".
			"Normal Count:\t$normal_count\nExceeded Count:\t$exceeded_count\n".
			"H_or_S_normal_count\t$h_or_s_normal_count\nH_or_S_exceeded_count:\t".
			"$h_or_s_exceeded_count");

		my $quota_recs              = $ems_hash->{quota_recs};
		my $normal_recs             = $ems_hash->{normal_recs};
		my $exceeded_recs           = $ems_hash->{exceeded_recs};
		my $h_or_s_normal_recs      = $ems_hash->{h_or_s_normal_recs};
		my $h_or_s_exceeded_recs    = $ems_hash->{h_or_s_exceeded_recs};


		if ($normal_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
			      "\nExpected no 'quota limit normal' EMS record,\n".
			      "Found: $normal_count EMS 'quota limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($h_or_s_normal_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
			      "\nExpected no 'quota $h_or_s limit normal' EMS record,\n".
			      "Found: $h_or_s_normal_count 'quota $h_or_s limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

		if ($opts{check_upper_bound_only})  {
			if ($exceeded_count > $rec_count) {
				$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
				      "\nExpected at most $rec_count 'quota limit exceeded' EMS record,\n".
				      "Found: $exceeded_count 'quota limit exceeded' records";
			       	logcomment($msg);
				push(@$fail_msg_list, $msg);
			}
		} elsif ($exceeded_count != $rec_count) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
			      "\nExpected exactly $rec_count 'quota limit exceeded' EMS record,\n".
			      "Found: $exceeded_count 'quota limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($opts{check_upper_bound_only})  {
			if ($h_or_s_exceeded_count > $rec_count) {
				$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
				      "\nExpected at most $rec_count 'quota $h_or_s limit exceeded' EMS record,\n".
				      "Found: $h_or_s_exceeded_count 'quota $h_or_s limit exceeded' records";
			       	logcomment($msg);
				push(@$fail_msg_list, $msg);
			}
		} elsif ($h_or_s_exceeded_count != $rec_count) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_exceeded  Line: ".__LINE__.
			      "\nExpected exactly $rec_count 'quota $h_or_s limit exceeded' EMS record,\n".
			      "Found: $h_or_s_exceeded_count 'quota $h_or_s limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

#		@other_quota_ems = list_difference(\@quota_ems, \@exceeded_ems);
	};

        my $fatal_msg = $@;
        logcomment("Exit verify_ems_quota_limit_exceeded:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub verify_ems_quota_limit_normal(%)
{
	my $fail_msg_list = [];

	eval {
		my (%opts) = @_;
		my $msg_in = $opts{msg_in} || "";
		my $Fc = $opts{Fc}
	                 or die("verify_ems_quota_limit_normal: No 'filer' option given");
		my $hard = ($opts{hard} || 0);
		my $soft = ($opts{soft} || 0);
		my $h_or_s = ($opts{h_or_s} || 0);
		my $disk = ($opts{disk} || 0);
		my $files = ($opts{files} || 0);
		my $d_or_f = ($opts{d_or_f} || 0);
	
		my ( $msg );
	        logcomment("Enter verify_ems_quota_limit_normal");
	        logDbg("Enter verify_ems_quota_limit_normal  Line: ".__LINE__.
	                  "\n\nMsg In: $msg_in\n\n".
			  "Hard: $hard\nSoft: $soft\nH_or_S: \"$h_or_s\"\n".
			  "Disk: $disk\nFiles: $files\nD_or_F: \"$d_or_f\"\n\n".
			  "Opts:\n\n".Dumper(\%opts));

		unless ($hard || $soft || $h_or_s) {
			$msg = "verify_ems_quota_limit_normal  Line: ".__LINE__.
			          "\n\nMust specify one of 'hard', 'soft'";
			logcomment($msg);
			die($msg);
		}
		unless ($disk || $files || $d_or_f) {
			$msg = "verify_ems_quota_limit_normal  Line: ".__LINE__.
			          "\n\nMust specify one of 'disk', 'files'";
		}

		my $rec_count;
		if (defined($opts{rec_count})) {
			$rec_count = $opts{rec_count};
		} else {
			$rec_count = 1;
		}

		logDbg("verify_ems_quota_limit_normal  Line: ".__LINE__."   Calling verify_quota_limit_aux ....");
		my ($fail_list, $err, $ems_hash) =
		             verify_ems_quota_limit_aux(msg_in => $msg_in, Fc => $Fc, %opts);
		if ($err) {
			$msg = "verify_ems_quota_limit_normal  Line: ".__LINE__.
			           "\nError from 'verify_ems_quota_limit_aux':\n$msg";
			logcomment($msg);
			die($msg);
		}
		logDbg("verify_ems_quota_limit_normal  Line: ".__LINE__."   .... Returned from verify_quota_limit_aux".
		         "\n\nFail cnt: \"$fail_list\"\nErr: \"$err\"\n\nEMS Hash:\n\n".Dumper($ems_hash));
		push(@$fail_msg_list, @$fail_list);

		my $quota_count             = $ems_hash->{quota_count};
		my $normal_count            = $ems_hash->{normal_count};
		my $exceeded_count          = $ems_hash->{exceeded_count};
		my $h_or_s_normal_count     = $ems_hash->{h_or_s_normal_count};
		my $h_or_s_exceeded_count   = $ems_hash->{h_or_s_exceeded_count};

		logDbg("verify_ems_quota_limit_normal  Line: ".__LINE__.
			"\n\nQuota Count:\t$quota_count\n".
			"Normal Count:\t$normal_count\nExceeded Count:\t$exceeded_count\n".
			"H_or_S_normal_count\t$h_or_s_normal_count\nH_or_S_exceeded_count:\t".
			"$h_or_s_exceeded_count");

		my $quota_recs              = $ems_hash->{quota_recs};
		my $normal_recs             = $ems_hash->{normal_recs};
		my $exceeded_recs           = $ems_hash->{exceeded_recs};
		my $h_or_s_normal_recs      = $ems_hash->{h_or_s_normal_recs};
		my $h_or_s_exceeded_recs    = $ems_hash->{h_or_s_exceeded_recs};

		if ($exceeded_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
			      "\nExpected no 'quota limit exceeded' EMS record,\n".
			      "Found: $exceeded_count 'quota limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($h_or_s_exceeded_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
			      "\nExpected no 'quota $h_or_s limit exceeded' EMS record,\n".
			      "Found: $h_or_s_exceeded_count 'quota $h_or_s limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

		if ($opts{check_upper_bound_only})  {
			if ($normal_count > $rec_count) {
				$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
				      "\nExpected at most $rec_count 'quota limit normal' EMS record,\n".
				      "Found: $normal_count 'quota limit normal' records";
			       	logcomment($msg);
				push(@$fail_msg_list, $msg);
			}
		} elsif ($normal_count != $rec_count) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
			      "\nExpected exactly $rec_count 'quota limit normal' EMS record,\n".
			      "Found: $normal_count 'quota limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($opts{check_upper_bound_only})  {
			if ($h_or_s_normal_count > $rec_count) {
				$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
				      "\nExpected at most $rec_count 'quota $h_or_s limit normal' EMS record,\n".
				      "Found: $h_or_s_normal_count 'quota $h_or_s limit normal' records";
			       	logcomment($msg);
				push(@$fail_msg_list, $msg);
			}
		} elsif ($h_or_s_normal_count != $rec_count) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_normal  Line: ".__LINE__.
			      "\nExpected exactly $rec_count 'quota $h_or_s limit normal' EMS record,\n".
			      "Found: $h_or_s_normal_count 'quota $h_or_s limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

#		@other_quota_ems = list_difference(\@quota_ems, \@exceeded_ems);
	};

        my $fatal_msg = $@;
        logcomment("Exit verify_ems_quota_limit_normal:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub verify_ems_quota_limit_unchanged(%)
{
	my $fail_msg_list = [];

	eval {
		my (%opts) = @_;
		my $msg_in = $opts{msg_in} || "";
		my $Fc = $opts{Fc}
	                 or die("verify_ems_quota_limit_unchanged: No 'filer' option given");
		my $hard = ($opts{hard} || 0);
		my $soft = ($opts{soft} || 0);
		my $h_or_s = ($opts{h_or_s} || 0);
		my $disk = ($opts{disk} || 0);
		my $files = ($opts{files} || 0);
		my $d_or_f = ($opts{d_or_f} || 0);
	
		my ( $msg );
	        logcomment("Enter verify_ems_quota_limit_unchanged");
	        logDbg("Enter verify_ems_quota_limit_unchanged  Line: ".__LINE__.
	                  "\n\nMsg In: $msg_in\n\n".
			  "Hard: $hard\nSoft: $soft\nH_or_S: \"$h_or_s\"\n".
			  "Disk: $disk\nFiles: $files\nD_or_F: \"$d_or_f\"\n\n".
			  "Opts:\n\n".Dumper(\%opts));

		unless ($hard || $soft || $h_or_s) {
			$msg = "verify_ems_quota_limit_unchanged  Line: ".__LINE__.
			          "\n\nMust specify one of 'hard', 'soft'";
			logcomment($msg);
			die($msg);
		}
		unless ($disk || $files || $d_or_f) {
			$msg = "verify_ems_quota_limit_unchanged  Line: ".__LINE__.
			          "\n\nMust specify one of 'disk', 'files'";
		}

		my $rec_count;
		if (defined($opts{rec_count})) {
			$rec_count = $opts{rec_count};
		} else {
			$rec_count = 1;
		}

		logDbg("verify_ems_quota_limit_unchanged  Line: ".__LINE__."   Calling verify_quota_limit_aux ....");
		my ($fail_list, $err, $ems_hash) =
		             verify_ems_quota_limit_aux(msg_in => $msg_in, Fc => $Fc, %opts);
		if ($err) {
			$msg = "verify_ems_quota_limit_unchanged  Line: ".__LINE__.
			           "\nError from 'verify_ems_quota_limit_aux':\n$msg";
			logcomment($msg);
			die($msg);
		}
		logDbg("verify_ems_quota_limit_unchanged  Line: ".__LINE__."   .... Returned from verify_quota_limit_aux".
		         "\n\nFail cnt: \"$fail_list\"\nErr: \"$err\"\n\nEMS Hash:\n\n".Dumper($ems_hash));
		push(@$fail_msg_list, @$fail_list);

		my $quota_count             = $ems_hash->{quota_count};
		my $normal_count            = $ems_hash->{normal_count};
		my $exceeded_count          = $ems_hash->{exceeded_count};
		my $h_or_s_normal_count     = $ems_hash->{h_or_s_normal_count};
		my $h_or_s_exceeded_count   = $ems_hash->{h_or_s_exceeded_count};

		logDbg("verify_ems_quota_limit_unchanged  Line: ".__LINE__.
			"\n\nQuota Count:\t$quota_count\n".
			"Normal Count:\t$normal_count\nExceeded Count:\t$exceeded_count\n".
			"H_or_S_normal_count\t$h_or_s_normal_count\nH_or_S_exceeded_count:\t".
			"$h_or_s_exceeded_count");

		my $quota_recs              = $ems_hash->{quota_recs};
		my $normal_recs             = $ems_hash->{normal_recs};
		my $exceeded_recs           = $ems_hash->{exceeded_recs};
		my $h_or_s_normal_recs      = $ems_hash->{h_or_s_normal_recs};
		my $h_or_s_exceeded_recs    = $ems_hash->{h_or_s_exceeded_recs};

		if ($normal_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_unchanged  Line: ".__LINE__.
			      "\nExpected no 'quota limit normal' EMS record,\n".
			      "Found: $normal_count 'quota limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($h_or_s_normal_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_unchanged  Line: ".__LINE__.
			      "\nExpected no 'quota $h_or_s limit normal' EMS record,\n".
			      "Found: $h_or_s_normal_count 'quota $h_or_s limit normal' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

		if ($exceeded_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_unchanged  Line: ".__LINE__.
			      "\nExpected no 'quota limit exceeded' EMS record,\n".
			      "Found: $exceeded_count 'quota limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($h_or_s_exceeded_count != 0) {
			$msg = "${msg_in}:\nverify_ems_quota_limit_unchanged  Line: ".__LINE__.
			      "\nExpected no 'quota $h_or_s limit exceeded' EMS record,\n".
			      "Found: $h_or_s_exceeded_count 'quota $h_or_s limit exceeded' records";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}

#		@other_quota_ems = list_difference(\@quota_ems, \@exceeded_ems);
	};

        my $fatal_msg = $@;
        logcomment("Exit verify_ems_quota_limit_unchanged:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



#################################################################
#
#   Groups:
#
#	Group 'A'
#          Role:	"'other' group, neither group nor login_list appear in any quota spec"
#          Members:     User[0], User[1], User[2],
#
#       Group 'B':
#	   Role:        "vanilla" NFS
#          Members:      User[3], User[4], User[5]
#
#       Group 'C':
#	   Role:	"multiple user entry" NFS.
#          Members:     User[6], User[7], User[8]
#
#	Group 'D'
#          Role:	"unmapped CIFS users"
#          Members:     User[9], User[10], User[11]
#
#	Group 'E'
#          Role:	"multi-entry CIFS users"
#          Members:     User[12], User[13], User[14]
#
#	Group 'F'
#          Role:	"mapped NFS/CIFS users"
#          Members:     User[15], User[16], User[17]
#
#	Group 'G'
#          Role:	"multi-entry mapped users"
#          Members:     User[18], User[19], User[20]
#
#
#
#################################################################

sub create_users_and_groups($%)
{
	my ($Fc, %opts ) = @_;
	logcomment("Enter create_users_and_groups");
        my $USE_CIFS = $opts{use_cifs};
	my ( $msg, $rslt );
	eval {
		@NFS_groups = ();
		@Null_groups = ();
		@CIFS_groups = ();
		@CIFS_groups_created = ();
		@Mapped_groups = ();
		@Null_users = ();
		@NFS_users = ();
		@CIFS_users = ();
		@CIFS_users_created = ();
		@Mapped_users = ();
		%User_hash = ();
		%WinUser_hash = ();
		%User_for_uid = ();
		%Group_hash = ();
		%WinGroup_hash = ();
		%Group_for_code = ();
	        %Group_for_gid = ();
		my @created_users_list = ();
		my @created_groups_list = ();

		@Users_z = ();
		my %Users_z_hash =
		    createUsers($Fc, 3,
	                          group                 => 'wafl_null_group_z',
	                          login_prefix          => 'wafl_null_user_z_',
	                          created_users_list    => \@Users_z,
	                          created_groups_list   => \@Null_groups);

		die("createUsers: $@") if ($@);
		$Group_z = $Null_groups[$#Null_groups];
		$Group_z->{letter} = 'Z';
		$Group_z->{desc} = 
		    "'other' group, neither group nor login_list appear in any quota spec";
		$Group_for_code{'Z'} = $Group_z;
		$Group_for_code{'a'} = $Group_z;
		@Null_users = @Users_z;

		@Users_a = ();
		my %Users_a_hash =
		    createUsers($Fc, 3,
	                          group                 => 'wafl_nfs_group_a',
	                          login_prefix          => 'wafl_nfs_user_a_',
	                          created_users_list    => \@Users_a,
	                          created_groups_list   => \@NFS_groups);

		die("createUsers: $@") if ($@);
		$Group_a = $NFS_groups[$#NFS_groups];
		$Group_a->{desc} = "Primary NFS users and group";
		$Group_a->{letter} = 'A';
		$Group_for_code{'A'} = $Group_a;
		$Group_for_code{'a'} = $Group_a;

		@Users_b = ();
		my %Users_b_hash =
		    createUsers($Fc, 3,
	                            group                 => 'wafl_nfs_group_b',
	                            login_prefix          => 'wafl_nfs_user_b_',
	                            created_users_list    => \@Users_b,
	                            created_groups_list   => \@NFS_groups);

		die("createUsers: $@") if ($@);
		$Group_b = $NFS_groups[$#NFS_groups];
		$Group_b->{desc} = "Secondary NFS users and group";
		$Group_b->{letter} = 'B';
		$Group_for_code{'B'} = $Group_b;
		$Group_for_code{'b'} = $Group_b;
		@NFS_users = ( @Users_a, @Users_b );

		if ($USE_CIFS) {
			@Users_c = ();
			my %Users_c_hash =
			    createUsers($Fc, 3,
		                           group                 => 'wafl_cifs_group_c',
		                           login_prefix          => 'wafl_cifs_user_c_',
		                           created_users_list    => \@Users_c,
		                           created_groups_list   => \@CIFS_groups);

			die("createUsers: $@") if ($@);
			$Group_c = $CIFS_groups[$#CIFS_groups];
			$Group_c->{desc} = "Primary CIFS users (unmapped)";
			$Group_c->{letter} = 'C';
			$Group_for_code{'C'} = $Group_c;
			$Group_for_code{'c'} = $Group_c;
	
			@Users_d = ();
			my %Users_d_hash =
			    createUsers($Fc, 3,
		                             group                 => 'wafl_cifs_group_d',
		                             login_prefix          => 'wafl_cifs_user_d_',
		                             created_users_list    => \@Users_d,
		                             created_groups_list   => \@CIFS_groups);

			die("createUsers: $@") if ($@);
			$Group_d = $CIFS_groups[$#CIFS_groups];
			$Group_d->{desc} = "Secondary CIFS users";
			$Group_for_code{'D'} = $Group_d;
			$Group_for_code{'d'} = $Group_d;
			$Group_d->{letter} = 'D';
			@CIFS_users = ( @Users_c, @Users_d );
	
			@Users_e = ();
			my %Users_e_hash =
			    createUsers($Fc, 3,
		                             group                 => 'wafl_mapped_group_e',
		                             login_prefix          => 'wafl_mapped_user_e_',
		                             created_users_list    => \@Users_e,
		                             created_groups_list   => \@Mapped_groups);

			die("createUsers: $@") if ($@);
			$Group_e = $Mapped_groups[$#Mapped_groups];
			$Group_e->{desc} = "Primary 'mapped' NFS/CIFS users";
			$Group_for_code{'E'} = $Group_e;
			$Group_for_code{'e'} = $Group_e;
			$Group_e->{letter} = 'E';
	
			@Users_f = ();
			my %Users_f_hash =
			     createUsers($Fc, 3,
		                              group                 => 'wafl_mapped_group_f',
		                              login_prefix          => 'wafl_mapped_user_f_',
		                              created_users_list    => \@Users_f,
		                              created_groups_list   => \@Mapped_groups);

			die("createUsers: $@") if ($@);
			$Group_f = $Mapped_groups[$#Mapped_groups];
			$Group_f->{desc} = "Secondary 'mapped' NFS/CIFS users";
			$Group_for_code{'F'} = $Group_f;
			$Group_for_code{'f'} = $Group_f;
			$Group_f->{letter} = 'F';
			@Mapped_users = ( @Users_e, @Users_f );
		}

		@All_users = ( @Null_users, @NFS_users, @CIFS_users, @Mapped_users );
		@All_groups = ( @Null_groups, @NFS_groups, @CIFS_groups, @Mapped_groups );

		foreach my $u (@All_users) {
			$User_hash{ $u->{login} } = $u;
			$User_for_uid{ $u->{uid} } = $u;
		}

		foreach my $g (@All_groups) {
			$Group_hash{ $g->{group} } = $g;
			$Group_for_gid{ $g->{gid} } = $g;
			$g->{login_list} = $g->{members_list};
			$g->{user_list} = [ map( {$User_hash{$_} }  @{$g->{login_list}} ) ];
			$g->{num_users} = scalar(@{$g->{login_list}});
			delete($g->{members});
			delete($g->{members_list});
		}

		my $cifs_server = getCIFSServer($Fc);
		if (!$cifs_server || $@) {
			$msg = "create_users_and_groups  Line: ".__LINE__.
			             "\ngetCIFSServer failed: $@";
#			$Log->warn($msg);
			logcomment($msg);
			die($msg);
		}

		foreach my $g (@Null_groups) {
			$g->{protos} = [ 'nfs', 'cifs' ];
			foreach my $u ( @{$g->{user_list}} ) {
				$u->{protos} = [ 'nfs', 'cifs' ];
			}
		}
		foreach my $g (@NFS_groups) {
			$g->{protos} = [ 'nfs' ];
			foreach my $u ( @{$g->{user_list}} ) {
				$u->{protos} = [ 'nfs' ];
			}
		}
		foreach my $g (@CIFS_groups) {
			$g->{protos} = [ 'cifs' ];
			foreach my $u ( @{$g->{user_list}} ) {
				$u->{protos} = [ 'cifs' ];
			}
		}
		foreach my $g (@Mapped_groups) {
			$g->{protos} = [ 'nfs', 'cifs' ];
			foreach my $u ( @{$g->{user_list}} ) {
				$u->{protos} = [ 'nfs', 'cifs' ];
			}
		}


		foreach my $g (@CIFS_groups, @Mapped_groups, @Null_groups) {
			my $wg = $g->{group} . "_win";
			$g->{win_group} = $wg;
			$WinGroup_hash{$wg} = $g;
			$rslt = addCIFSGroup($Fc, $wg);
			push(@CIFS_groups_created, $g);
			if (!$rslt || $@) {
				$msg = "create_users_and_groups  Line: ".__LINE__.
				             "\naddCIFSGroup \"$wg\"  failed: $@";
				logcomment($msg);
#				$Log->warn($msg);
				die($msg);
			}
			foreach my $u ( @{$g->{user_list}} ) {
				my $wu = $u->{login} . "_win";
				$u->{win_login} = $wu;
				$u->{full_win_login} = $cifs_server . "\\" . $wu;
				$WinUser_hash{$wu} = $u;
#				$rslt = addCIFSUser($Fc, $wu, $cifs_passwd, "Users");
				$rslt = addCIFSUser($Fc, $wu, $cifs_passwd, $wg);
				if (!$rslt || $@) {
					$msg = "create_users_and_groups  Line: ".__LINE__.
					             "\naddCIFSUser \"$wu\"  failed: $@";
#				$Log->warn($msg);
				logcomment($msg);
					die($msg);
				}
#				if (!findCIFSUser($Fc, $wu)) {
#					$msg = "create_users_and_groups  Line: ".__LINE__.
#					            "\nFailed to add CIFS user \"$wu\"";
#					$Log->warn($msg);
#					logcomment($msg);
#					die($msg);
#				}
				push(@CIFS_users_created, $u);
				if ($DEBUG) {
				        $msg = "Create_users_and_groups  Line: ".__LINE__."\n\n".
				                  "WARNING: 'DEBUG' is set -- Creating only\n".
						  "one CIFS user for group $g->{group}: \"$u->{win_login}\"";
#					$Log->warn($msg);
					logcomment($msg);
					last;
				}
			}
			$g->{win_login_list}        = [ map( {$_->{win_login}}  @{$g->{user_list}} ) ];
			$g->{login_list}   = [ map( {$_->{login}}  @{$g->{user_list}} ) ];
		}
		my @cifs_user_list = listCIFSUsers($Fc);
		my %cifs_user_hash = ();
		for ( @cifs_user_list ) {
			my $elt = { group => $_->{Group}, rid=>$_->{Rid}, name=>$_->{Name}, info=>$_->{Info} };
			$cifs_user_hash{$elt->{name}} = $elt;
		}

### Start DBG
#logcomment("create_users_and_groups  Line: ".__LINE__."\n\nCIFS_USER_HASH:\n\n".Dumper(\@cifs_user_list)."\n\n".Dumper(\%cifs_user_hash));
#die("Line: ".__LINE__);
### End DBG

		foreach my $u (@CIFS_users_created) {
			if (! $cifs_user_hash{$u->{win_login}}) {
				$msg = "create_users_and_groups  Line: ".__LINE__.
				            "\nFailed to add CIFS user \"$u->{win_login}\"";
#				$Log->warn($msg);
				logcomment($msg);
				die($msg);
			}
		}
	};

	my $fatal_msg = $@;
	if ($fatal_msg) {
		logcomment("Exit  create_users_and_groups \nErrMsg: \"$fatal_msg\"");
		return($fatal_msg);
	}

	my $prev_grp = "";

	my $step_msg =
               "\n\n ==========================  Users and Groups  ===========================\n".
		     "\n".
		     "  GROUPS:\n\n";

	$step_msg .= " NULL Groups:\n";
	for (my $i=0; $i<=$#Null_groups; $i++) {
		my $g_ref = $Null_groups[$i];
		$step_msg .= "        Group \'$g_ref->{letter}\':\n";
		$step_msg .= "            Letter:   \'$g_ref->{letter}\'\n";
		$step_msg .= "            Role:     \'$g_ref->{desc}\'\n";
		$step_msg .= "            Group:    \'$g_ref->{group}\'\n";
		$step_msg .= "            GID:      \'$g_ref->{gid}\'\n";
		$step_msg .= "            Members:\n". join("\n",  map( {"\t\t" . $_} @{$g_ref->{login_list}}) );
		$step_msg .= "\n";
		$step_msg .= "\n";
	}
	$step_msg .= "\n";

	$step_msg .= " NFS Groups:\n";
	for (my $i=0; $i<=$#NFS_groups; $i++) {
		my $g_ref = $NFS_groups[$i];
		$step_msg .= "        Group \'$g_ref->{letter}\':\n";
		$step_msg .= "            Letter:   \'$g_ref->{letter}\'\n";
		$step_msg .= "            Role:     \'$g_ref->{desc}\'\n";
		$step_msg .= "            Group:    \'$g_ref->{group}\'\n";
		$step_msg .= "            GID:      \'$g_ref->{gid}\'\n";
		$step_msg .= "            Members:\n". join("\n",  map( {"\t\t" . $_} @{$g_ref->{login_list}}) );
		$step_msg .= "\n";
		$step_msg .= "\n";
	}
	$step_msg .= "\n";

	$step_msg .= " CIFS Groups:\n";
	for (my $i=0; $i<=$#CIFS_groups; $i++) {
		my $g_ref = $CIFS_groups[$i];
		$step_msg .= "        Group \'$g_ref->{letter}\':\n";
		$step_msg .= "            Letter:   \'$g_ref->{letter}\'\n";
		$step_msg .= "            Role:     \'$g_ref->{desc}\'\n";
		$step_msg .= "            Group:    \'$g_ref->{group}\'\n";
		$step_msg .= "            GID:      \'$g_ref->{gid}\'\n";
		$step_msg .= "            Members:\n". join("\n",  map( {"\t\t" . $_} @{$g_ref->{login_list}}) );
		$step_msg .= "\n";
		$step_msg .= "\n";
	}
	$step_msg .= "\n";

	$step_msg .= " Mapped Groups:\n";
	for (my $i=0; $i<=$#Mapped_groups; $i++) {
		my $g_ref = $Mapped_groups[$i];
		$step_msg .= "        Group \'$g_ref->{letter}\':\n";
		$step_msg .= "            Letter:   \'$g_ref->{letter}\'\n";
		$step_msg .= "            Role:     \'$g_ref->{desc}\'\n";
		$step_msg .= "            Group:    \'$g_ref->{group}\'\n";
		$step_msg .= "            GID:      \'$g_ref->{gid}\'\n";
		$step_msg .= "            Members:\n". join("\n",  map( {"\t\t" . $_} @{$g_ref->{login_list}}) );
		$step_msg .= "\n";
		$step_msg .= "\n";
	}
	$step_msg .= "\n";

	$step_msg .= " Null Users:\n";
	$prev_grp = "";
	for (my $i=0; $i<=$#Null_users; $i++) {
		my $u_ref = $Null_users[$i];
		my $curr_grp = $u_ref->{group};
		if ($curr_grp ne $prev_grp) {
			$step_msg .= "   Group: \"$curr_grp\":\n\n";
			$prev_grp = $curr_grp;
		}
		$step_msg .= "     \'$u_ref->{login}\':\n";
		$step_msg .= "         UID:             \'$u_ref->{uid}\'\n";
		$step_msg .= "         GID:             \'$u_ref->{gid}\'\n";
		$step_msg .= "         win_login:       \'$u_ref->{win_login}\'\n";
		$step_msg .= "         full_win_login:  \'$u_ref->{full_win_login}\'\n\n";
	}
	$step_msg .= "\n";

	$step_msg .= " NFS Users:\n";
	$prev_grp = "";
	for (my $i=0; $i<=$#NFS_users; $i++) {
		my $u_ref = $NFS_users[$i];
		my $curr_grp = $u_ref->{group};
		if ($curr_grp ne $prev_grp) {
			$step_msg .= "   Group: \"$curr_grp\":\n\n";
			$prev_grp = $curr_grp;
		}
		$step_msg .= "     \'$u_ref->{login}\':\n";
		$step_msg .= "         UID:             \'$u_ref->{uid}\'\n";
		$step_msg .= "         GID:             \'$u_ref->{gid}\'\n";
		$step_msg .= "         win_login:       \'$u_ref->{win_login}\'\n";
		$step_msg .= "         full_win_login:  \'$u_ref->{full_win_login}\'\n\n";
	}
	$step_msg .= "\n";

	$step_msg .= " CIFS Users:\n";
	$prev_grp = "";
	for (my $i=0; $i<=$#CIFS_users; $i++) {
		my $u_ref = $CIFS_users[$i];
		my $curr_grp = $u_ref->{group};
		if ($curr_grp ne $prev_grp) {
			$step_msg .= "   Group: \"$curr_grp\":\n\n";
			$prev_grp = $curr_grp;
		}
		$step_msg .= "     \'$u_ref->{login}\':\n";
		$step_msg .= "         UID:             \'$u_ref->{uid}\'\n";
		$step_msg .= "         GID:             \'$u_ref->{gid}\'\n";
		$step_msg .= "         win_login:       \'$u_ref->{win_login}\'\n";
		$step_msg .= "         full_win_login:  \'$u_ref->{full_win_login}\'\n\n";
	}
	$step_msg .= "\n";

	$step_msg .= " Mapped Users:\n";
	$prev_grp = "";
	for (my $i=0; $i<=$#Mapped_users; $i++) {
		my $u_ref = $Mapped_users[$i];
		my $curr_grp = $u_ref->{group};
		if ($curr_grp ne $prev_grp) {
			$step_msg .= "   Group: \"$curr_grp\":\n\n";
			$prev_grp = $curr_grp;
		}
		$step_msg .= "     \'$u_ref->{login}\':\n";
		$step_msg .= "         UID:             \'$u_ref->{uid}\'\n";
		$step_msg .= "         GID:             \'$u_ref->{gid}\'\n";
		$step_msg .= "         win_login:       \'$u_ref->{win_login}\'\n";
		$step_msg .= "         full_win_login:  \'$u_ref->{full_win_login}\'\n\n";
	}
	$step_msg .= "\n";


	$step_msg .=
                    "\n".
                    "\n".
		    "    Global Vars:\n".
                    "         \@NFS_users\n".
                    "         \@NFS_groups\n".
                    "         \@CIFS_users\n".
                    "         \@CIFS_groups\n".
                    "         \@Mapped_users\n".
                    "         \@Mapped_groups\n".
                    "         \%User_hash\n".
                    "         \%WinUser_hash\n".
                    "         \%User_for_uid\n".
                    "         \%Group_hash\n".
                    "         \%WinGroup_hash\n".
                    "         \%Group_for_gid\n".
                    "         \%Group_for_code\n".
                    "         \@Users_a\n".
                    "         \@Users_b\n".
                    "         \@Users_c\n".
                    "         \@Users_d\n".
                    "         \@Users_e\n".
                    "         \@Users_f\n".
                    "         \$Group_a\n".
                    "         \$Group_b\n".
                    "         \$Group_c\n".
                    "         \$Group_d\n".
                    "         \$Group_e\n".
                    "         \$Group_f\n".
		    "\n".
                    "=============================================================================";

       	logStep($step_msg);

	logcomment("Exit  create_users_and_groups\nErrMsg: \"$fatal_msg\"");
#    logcomment("Exit create_users_and_groups\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));

	return("");
}



############################################################################################
#
#   Qtree:
#
#	Qtree_0		mixed
#
#	Qtree_1		unix
#	Qtree_2		unix
#	Qtree_3		ntfs
#	Qtree_4		ntfs
#	Qtree_5		mixed
#	Qtree_6		mixed
#
#    Dirs:
#
#	Qtree_0:	testdir_0a   testdir_0b		mixed
#
#	Qtree_1:	testdir_1a   testdir_1b		unix
#
#	Qtree_2:	testdir_2a   testdir_2b		unix
#
#	Qtree_3:	testdir_3a   testdir_3b		ntfs
#
#	Qtree_4:	testdir_4a   testdir_4b		ntfs
#
#	Qtree_5:	testdir_5a   testdir_5b		mixed
#
#	Qtree_6:	testdir_6a   testdir_6b		mixed
#
#
############################################################################################

sub create_qtrees_and_dirs(%)
{
	my (%opts) = @_;
 	my $Fc = $opts{Fc};
        my $Vol = $opts{vol};
	logcomment("Enter create_qtrees_and_dirs");
        my $USE_CIFS = $opts{use_cifs};
	my $Mntpt = $opts{mntpt};
	eval {
		my ($rslt, $msg);
		%Dir_for_code = ();
		@NFS_qtrees = ();
		%Qtree_for_num = ();
		@CIFS_qtrees = ();
		@Mixed_qtrees = ();
		@Null_qtrees = ();
		%Qtree_hash = ();
		%Dir_hash = ();
		@All_qtrees = ();
		%Qtree_for_dir = ();

	        $Qtree_0 = {
				name	   => 'qtree_0',
				dirs	   => [],
				security   => 'mixed',
				number	   => '0',
				protos	   => [ 'nfs', 'cifs' ],
				desc       => "'null' qtree",
			   };
		push(@Null_qtrees, $Qtree_0);

	        $Qtree_1 = {
				name	   => 'qtree_1',
				dirs	   => [],
				security   => 'unix',
				number	   => '1',
				protos	   => [ 'nfs' ],
				desc       => "Primary NFS qtree",
			   };

		push(@NFS_qtrees, $Qtree_1);

	        $Qtree_2 = {
				name	   => 'qtree_2',
				dirs	   => [],
				security   => 'unix',
				number	   => '2',
				protos	   => [ 'nfs' ],
				desc       => "Secondary NFS qtree",
			   };
		push(@NFS_qtrees, $Qtree_2);

		if ($USE_CIFS) {
		        $Qtree_3 = {
					name	   => 'qtree_3',
					dirs	   => [],
					security   => 'ntfs',
					number	   => '3',
					protos	   => [ 'cifs' ],
					desc       => "Primary CIFS qtree",
				   };
			push(@CIFS_qtrees, $Qtree_3);

		        $Qtree_4 = {
					name	   => 'qtree_4',
					dirs	   => [],
					security   => 'ntfs',
					number	   => '4',
					protos	   => [ 'cifs' ],
					desc       => "Secondary CIFS qtree",
				   };
			push(@CIFS_qtrees, $Qtree_4);

		        $Qtree_5 = {
					name	   => 'qtree_5',
					dirs	   => [],
					security   => 'mixed',
					number	   => '5',
					protos	   => [ 'nfs', 'cifs' ],
					desc       => "Primary 'mixed' qtree",
				   };
			push(@Mixed_qtrees, $Qtree_5);

		        $Qtree_6 = {
					name	   => 'qtree_6',
					dirs	   => [],
					security   => 'mixed',
					number	   => '6',
					protos	   => [ 'nfs', 'cifs' ],
					desc       => "Secondary 'mixed' qtree",
				   };
			push(@Mixed_qtrees, $Qtree_6);
		}


		 @All_qtrees = ( @NFS_qtrees, @CIFS_qtrees, @Mixed_qtrees, @Null_qtrees );
		 foreach my $qt (@All_qtrees) {
			 $Qtree_hash{$qt->{name}} = $qt;
			 $Qtree_for_num{$qt->{number}} = $qt;
			 my $npath = $NFS_paths{$Vol} . '/' . $qt->{name};
			 my $fpath = $Filer_paths{$Vol} . '/' . $qt->{name};
			 $Filer_paths{$qt->{name}} = $fpath;
			 $qt->{filer_path} = $fpath;
			 $NFS_paths{$qt->{name}} = $npath;
			 $qt->{nfs_path} = $npath;
			 if ($USE_CIFS) {
				 my $cpath = $CIFS_paths{$Vol} . '/' . $qt->{name};
				 $CIFS_paths{$qt->{name}} = $cpath;
				 $qt->{cifs_path} = $cpath;
			 }
		 }

		foreach my $qt (@All_qtrees) {
			 if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
				 $msg = "create_qtrees_and_dirs  Line: ".__LINE__."\n".
		         	            "Qtree $Filer_paths{$qt->{name}} already exists";
#		         	 $Log->warn($msg);
		         	 logcomment($msg);
			 } else {
				 $rslt = qtreeCreate($Fc, $Filer_paths{$qt->{name}});
				 if (!$rslt || $@) {
				 	$msg = "create_qtrees_and_dirs  Line: ".__LINE__."\n".
				                  "qtreeCreate $qt->{name}: $@";
					die($msg);
				 }
			 }
	         	 $rslt = qtreeSecurity($Fc, $Filer_paths{$qt->{name}}, $qt->{security});
			 if (!$rslt || $@) {
			 	$msg = "create_qtrees_and_dirs  Line: ".__LINE__."\n".
		                            "(Set) qtreeSecurity $qt->{name}: $@";
				 die($msg);
			 }
		 }

	         ## @step Create and setup test dir to contain files
	         logStep("Create and setup test dirs to contain files");

		 #  Null qtrees
		 $Qtree_0->{dirs} = [ { name=>'testdir_0a', code=>'0a', },
		                        { name=>'testdir_0b', code=>'0b' } ];
		 foreach my $qt ( @Null_qtrees ) {
			foreach my $dir ( @{$qt->{dirs}} ) {
				push(@Null_dirs, $dir);
			}
		 }


		 #  NFS qtrees
		 $Qtree_1->{dirs} = [ { name=>'testdir_1a', code=>'1a', },
		                        { name=>'testdir_1b', code=>'1b' } ];
		 $Qtree_2->{dirs} = [ { name=>'testdir_2a', code=>'2a', },
		                        { name=>'testdir_2b', code=>'2b', } ];
		 foreach my $qt ( @NFS_qtrees ) {
			foreach my $dir ( @{$qt->{dirs}} ) {
				push(@NFS_dirs, $dir);
			}
		 }


		 if ($USE_CIFS) {
		 	$Qtree_3->{dirs} = [ { name=>'testdir_3a', code=>'3a', },
			                       { name=>'testdir_3b', code=>'3b', } ];
			$Qtree_4->{dirs} = [ { name=>'testdir_4a', code=>'4a', },
			                       { name=>'testdir_4b', code=>'4b', } ];
			foreach my $qt ( @CIFS_qtrees ) {
				foreach my $dir ( @{$qt->{dirs}} ) {
					push(@CIFS_dirs, $dir);
				}
			}
		 	$Qtree_5->{dirs} = [ { name=>'testdir_5a', code=>'5a', },
			                       { name=>'testdir_5b', code=>'5b', } ];
			$Qtree_6->{dirs} = [ { name=>'testdir_6a', code=>'6a', },
			                       { name=>'testdir_6b', code=>'6b', } ];
			foreach my $qt ( @Mixed_qtrees ) {
				foreach my $dir ( @{$qt->{dirs}} ) {
					push(@Mixed_dirs, $dir);
				}
			}
		 }

		 @All_dirs = ( @NFS_dirs, @CIFS_dirs, @Mixed_dirs, @Null_dirs );

		 foreach my $qt ( @All_qtrees ) {
			foreach my $dir ( @{$qt->{dirs}} ) {
				$Qtree_for_dir{$dir->{name}} = $qt;
				$dir->{qtree} = $qt->{name};
				$dir->{protos} = $qt->{protos};
				$Dir_hash{$dir->{name}} = $dir;
				$Dir_for_code{$dir->{code}} = $dir;
				my $fp = $Filer_paths{$qt->{name}} . '/' . $dir->{name};
				$Filer_paths{$dir->{name}} = $fp;
				$dir->{filer_path} = $fp;
		 		if ($NFS_paths{$qt->{name}}) {
					my $np = $NFS_paths{$qt->{name}} . '/' . $dir->{name};
					$NFS_paths{$dir->{name}} = $np;
					$dir->{nfs_path} = $np;
				}
		 		if ($USE_CIFS && $CIFS_paths{$qt->{name}}) {
					my $cp = $CIFS_paths{$qt->{name}} . '/' . $dir->{name};
					$CIFS_paths{$dir->{name}} = $cp;
					$dir->{cifs_path} = $cp;
				}
				if ($dir->{name} =~ /_(\d\w)/) {
					$Dir_for_code{$1} = $dir;
				} else {
			 	 	$msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					          "\nUnknown Test Directory: \"$dir->{name}\"";
					die($msg);
				}
			}
		 }

	         $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
	                   "\nVerifying client (NFS/CIFS) access to qtrees ....";
		 logcomment($msg);
		 foreach my $qt ( @All_qtrees ) {
			my $mounted_qt;
			if ($qt->{security} eq 'unix'  ||  $qt->{security} eq 'mixed') {
				$mounted_qt = $qt->{nfs_path};
				if (!test_file(file=>$mounted_qt, test=>'d', client=>$nfs_client)) {
				        $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					          "\nCannot access qtree: $qt->{name} via NFS";
					logcomment($msg);
					die($msg);
				}
			}
			if ($qt->{security} eq 'ntfs'  ||  $qt->{security} eq 'mixed') {
				$mounted_qt = $qt->{cifs_path};
				if (!test_file(file=>$mounted_qt, test=>'d', client=>$cifs_client)) {
				        $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					          "\nCannot access qtree: $qt->{name} via CIFS";
					logcomment($msg);
					die($msg);
				}
			}
		 }

	         $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
	                   "\n.... Finished verifying client (NFS/CIFS) access to qtrees";
		 logcomment($msg);


		 foreach my $qt ( @All_qtrees ) {
			foreach my $dir ( @{$qt->{dirs}} ) {
				# my $mounted_dir = $NFS_paths{$dir->{name}};
				next unless ($qt->{security} eq 'unix'  ||  $qt->{security} eq 'mixed');
				my $mounted_dir = $dir->{nfs_path};
			        next if (test_file(file=>$mounted_dir, test=>'d', client=>$nfs_client));
				$rslt = make_dir(  dir              => $mounted_dir,
				                   client           => $nfs_client,
					           okay_if_exists   => 1,
				         );
				if (!$rslt || $@) {
			 	 	$msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					           "\nmake_dir $mounted_dir: $@";
					die($msg);
				}
			        if (!test_file(file=>$mounted_dir, test=>'d', client=>$nfs_client)) {
				       $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					          "\nmake_dir failed to create directory $mounted_dir";
	#			       $Log->warn($msg);
				       logcomment($msg);
				       die($msg);
				 }
				 $dir->{created_via} = 'nfs';
		 	 	 $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					     "\n\nCreated dir \"$dir->{name}\" via NFS: \"$mounted_dir\"";
				 logcomment($msg);
			 }
		 }

		 foreach my $qt ( @All_qtrees ) {
			foreach my $dir ( @{$qt->{dirs}} ) {
				# my $mounted_dir = $CIFS_paths{$dir->{name}};
				next unless ($qt->{security} eq 'ntfs'  ||  $qt->{security} eq 'mixed');
				my $mounted_dir = $dir->{cifs_path};
			        next if (test_file(file=>$mounted_dir, test=>'d', client=>$cifs_client));
				$rslt = make_dir(  dir              => $mounted_dir,
				                   client           => $cifs_client,
					           okay_if_exists   => 1,
				         );
				if (!$rslt || $@) {
			 	 	$msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					           "\nmake_dir $mounted_dir: $@";
					die($msg);
				}
			        if (!test_file(file=>$mounted_dir, test=>'d', client=>$cifs_client)) {
				       $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					          "\nmake_dir failed to create directory $mounted_dir";
	#			       $Log->warn($msg);
				       logcomment($msg);
				       die($msg);
				 }
				 $dir->{created_via} = 'cifs';
		 	 	 $msg = "create_qtrees_and_dirs  Line: ".__LINE__.
					     "\n\nCreated dir \"$dir->{name}\" via CIFS: \"$mounted_dir\"";
				 logcomment($msg);
			 }
		 }

		foreach my $dir ( @All_dirs ) {
			if (!$dir->{created_via}) {
				$msg = "Testdir \"$dir->{name}\" was not created";
	#		        $Log->warn($msg);
			        logcomment($msg);
			        die($msg);
			}
		}

	};

	my $fatal_msg = $@;
	if ($fatal_msg) {
		logcomment("Exit  create_qtrees_and_dirs Line: ".__LINE__."\nErrMsg: \"$fatal_msg\"");
		return ($fatal_msg);
	}

	my $step_msg =
                "\n\n=========================  Qtrees and Directories ==========================\n".
		     "\n".
		     "  QTREES:\n".
		     "\n".
		     "      $Qtree_0->{name}\n".
		     "          Security:       unix\n".
		     "          Access:         NFS or CIFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_0->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_0->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_0->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_0->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_1->{name}\n".
		     "          Security:       unix\n".
		     "          Access:         NFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_1->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_1->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_1->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_1->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_2->{name}\n".
		     "          Security:       unix\n".
		     "          Access:         NFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_2->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_2->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_2->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_2->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_3->{name}\n".
		     "          Security:       ntfs\n".
		     "          Access:         CIFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_3->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_3->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_3->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_3->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_4->{name}\n".
		     "          Security:       unix\n".
		     "          Access:         NFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_4->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_4->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_4->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_4->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_5->{name}\n".
		     "          Security:       mixed\n".
		     "          Access:         NFS or CIFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_5->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_5->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_5->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_5->{dirs}[1]->{name}."\n".
		     "\n".
		     "      $Qtree_6->{name}\n".
		     "          Security:       mixed\n".
		     "          Access:         NFS or CIFS\n".
		     "          Filer Path:     $Filer_paths{$Qtree_6->{name}}\n".
		     "          Mounted Path:   $NFS_paths{$Qtree_6->{name}}\n".
		     "          Test Dirs:\n".
		     "              ".$Qtree_6->{dirs}[0]->{name}."\n".
		     "              ".$Qtree_6->{dirs}[1]->{name}."\n".
		     "\n".
		     "\n".
		     "   TEST DIRS:\n".
		     "\n".
		     "              testdir_0a\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'0a'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'0a'}}\n".
		     "\n".
		     "              testdir_0b\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'0b'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'0b'}}\n".
		     "\n".
		     "              testdir_1a\n".
		     "			Access:	       NFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'1a'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'1a'}}\n".
		     "\n".
		     "              testdir_1b\n".
		     "			Access:	       NFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'1b'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'1b'}}\n".
		     "\n".
		     "              testdir_2a\n".
		     "			Access:	       NFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'2a'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'2a'}}\n".
		     "\n".
		     "              testdir_2b\n".
		     "			Access:	       NFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'2b'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'2b'}}\n".
		     "\n".
		     "              testdir_3a\n".
		     "			Access:	       CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'3a'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'3a'}}\n".
		     "\n".
		     "              testdir_3b\n".
		     "			Access:	       CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'3b'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'3b'}}\n".
		     "\n".
		     "              testdir_4a\n".
		     "			Access:	       CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'4a'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'4a'}}\n".
		     "\n".
		     "              testdir_4b\n".
		     "			Access:	       CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'4b'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'4b'}}\n".
		     "\n".
		     "              testdir_5a\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'5a'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'5a'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'5a'}}\n".
		     "\n".
		     "              testdir_5b\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'5b'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'5b'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'5b'}}\n".
		     "\n".
		     "              testdir_6a\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'6a'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'6a'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'6a'}}\n".
		     "\n".
		     "              testdir_6b\n".
		     "			Access:	       NFS or CIFS\n".
		     "			Filer_Path:    $Filer_paths{$Dir_for_code{'6b'}}\n".
		     "			Mounted_Path:  $NFS_paths{$Dir_for_code{'6b'}}\n".
		     "			Mounted_Path:  $CIFS_paths{$Dir_for_code{'6b'}}\n".
		     "\n".
		     "\n".
		     "    Global Vars:\n".
                     "         \@Null_qtrees\n".
                     "         \@NFS_qtrees\n".
                     "         \@CIFS_qtrees\n".
                     "         \@Mixed_qtrees\n".
                     "         \@All_qtrees\n".
                     "         \%Qtree_hash\n".
                     "         \%Qtree_for_dir\n".
                     "         \%Qtree_for_num\n".
		     "         \%Filer_path\n".
		     "         \%NFS_paths\n".
                     "         \$Mntpt\n".
                     "         \$Rootvol_mntpt\n".
                     "         \%Dir_for_code\n".
                     "         \$Qtree_0\n".
                     "         \$Qtree_1\n".
                     "         \$Qtree_2\n".
                     "         \$Qtree_3\n".
                     "         \$Qtree_4\n".
                     "         \$Qtree_5\n".
                     "         \$Qtree_6\n".
		     "\n".
                     "=============================================================================";

        logStep($step_msg);
#    logcomment("Exit create_qtrees_and_dirs\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("Exit  create_qtrees_and_dirs\nErrMsg: \"$fatal_msg\"");

	return("");
}



sub create_quota_specs(%)
{
	my (%opts) = @_;
	my $Vol        = $opts{vol};
        my $USE_CIFS   = $opts{use_cifs};

        logcomment("Enter create_quota_specs");

###########################################################################################################
#
#   Given:
#
#	Group 'A'
#          Role:	"'other' group, neither group nor login_list appear in any quota spec"
#
#       Group 'B':
#	   Role:        "vanilla" NFS
#
#       Group 'C':
#	   Role:	"multiple user entry" NFS.
#
#	Group 'D'
#          Role:	"unmapped CIFS users"
#
#	Group 'E'
#          Role:	"multiple-entry CIFS users"
#
#	Group 'F'
#          Role:	"mapped NFS/CIFS users"
#
#	Group 'G'
#          Role:	"multi-entry mapped users"
#
#
#    And:
#
#       Qtree                 Dirs                    security
#       ----------------------------------------------------------
#	Qtree_0:	testdir_0a   testdir_0b		mixed
#
#	Qtree_1:	testdir_1a   testdir_1b		unix
#
#	Qtree_2:	testdir_2a   testdir_2b		unix
#
#	Qtree_3:	testdir_3a   testdir_3b		ntfs
#
#	Qtree_4:	testdir_4a   testdir_4b		ntfs
#
#	Qtree_5:	testdir_5a   testdir_5b		mixed
#
#	Qtree_6:	testdir_6a   testdir_6b		mixed
#
#
#  Then:
#
#   Quota[0]		'bad' quota spec -- for 'test_003_bad_quotas_file' only
#   Quota[1]:	   User		Single NFS user;            Qtree
#   Quota[2]:	   User		Single NFS user;            Volume
#   Quota[3]:	   User		Multi-entry NFS users;      Qtree
#   Quota[4]:	   User		Multi-entry NFS users;      Volume
#   Quota[5]:	   Group	NFS group;                  Qtree
#   Quota[6]:	   Group	NFS group;                  Volume
#   Quota[7]:      Qtree	NFS users;	            (Qtree)
#   Quota[8]:	   User		Single CIFS user            Qtree 
#   Quota[9]: 	   User		Single CIFS user;           Volume
#   Quota[10]:      Qtree	CIFS users;	            (Qtree)
#   Quota[11]:     User		Mapped user pair;           Qtree.
#   Quota[12]:     User		Mapped user pair;           Volume
#   Quota[13]:     Qtree	Mapped user pair;      	    (Qtree)
#
#
#   Quota  		Type		Target		  Qtree		    	Misc.
#   ------------------------------------------------------------------------------------------------------------------------
#   Quota[0]		'bad' quota spec -- for 'test_003_bad_quotas_file' only
#   Quota[1]		user		Grp_a[0]	    Qtree_1	  Groups=>[Grp_a;              Dirs=>testdir1a
#   Quota[2]		user		Grp_a[0]    	    -		  Groups=>[Grp_a;              Dirs=>[testdir1a,testdir2a
#   Quota[3]		user		Grp_b[0],Grp_b[1]   Qtree_1	  Groups=>Grp_b;               Dirs=>Tesdir1a
#   Quota[4]		user		Grp_b[0],Grp_b[1]   -             Groups=>Grp_b;               Dirs=>Tesdir1a,testdir2a
#   Quota[5]		group		Grp_a	            Qtree_1 	  Users=>Grp_a[0],Grp_a[1];    Difs=>testdir1a,
#   Quota[6]		group		Grp_a	            -		  Users=>Grp_a[0],Grp_a[1];    Dirs=>testdir1a,testdir2a
#   Quota[7]		tree		Qtree_1		    (Qtree1)	  Users=>Grp_a[0],Grp_a[1];    Dirs=>testdir1a
#   Quota[8]		user		Grp_c[0]	    Qtree1	  Groups=>Grp_c;               Dirs=>testdir3a
#   Quota[9]		user		Grp_c[0]	    -	  	  Dirs=>testdir3a;             Dirs=>testdir3a,testdir4a
#   Quota[10]		tree		Qtree_1		    (Qtree1)	  Users=>Grp_c[0],Grp_c[1];    Dir=>testdir3a
#   Quota[11]		user		Grp_e[0]	    Qtree_1	  Groups=>Grp_e;               Dirs=>testdir5a
#   Quota[12]		user		Grp_e[0]	    -	          Groups=>Grp_e;               Dirs=>testdir5a,testdir6a
#   Quota[13]		tree		Qtree_1		    (Qtree1)	  Userss=>Grp_e[0],Grp_e[1];   Dirs=>testdir5a
#
#
# 'CIFS' tests only:
#
#   Quota[6]		user		WinUsers[0]	  Qtree1	WinUsers[0]; Testdir1a
#
###########################################################################################################


    eval {

	@All_quota_specs = ();


	$All_quota_specs[0] = {
		type  	 => 'user',
		Desc	 => "Quota[0]: User; Single NFS user; Qtree",
		user	 => $Group_a->{login_list}[0],
		tree	 => $Qtree_1->{name},
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_a->{login_list}[0] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[1] = {
		type  	 => 'user',
		Desc	 => "Quota[1]: User; Single NFS user; Qtree",
		user	 => $Group_a->{login_list}[0],
		tree	 => $Qtree_1->{name},
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_a->{login_list}[0] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[2] = {
		type 	 => 'user',
		Desc	 => "Quota[2]: User; Single NFS user; Volume",
		user	 => $Group_a->{login_list}[0],
		tree	 => '-',
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_a->{login_list}[0] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name}, $Dir_for_code{'2a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[3] = {
		type 	 => 'user',
		Desc     => "Quota[3]: User; Multi-entry NFS users; Qtree",
		user	 => "$Group_b->{login_list}[0],$Group_b->{login_list}[1]",
		tree	 => $Qtree_1->{name},
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_b->{login_list}[0], $Group_b->{login_list}[1] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[4] = {
		type 	 => 'user',
		Desc     => "Quota[4]: User; Multi-entry NFS users; Volume",
		user	 => "$Group_b->{login_list}[0],$Group_b->{login_list}[1]",
		tree	 => '-',
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_b->{login_list}[0], $Group_b->{login_list}[1] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name}, $Dir_for_code{'2a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[5] = {
		type 	 => 'group',
		Desc     => "Quota[5]: Group; NFS group; Qtree",
		group	 => $Group_a->{group},
		tree	 => $Qtree_1->{name},
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_a->{login_list}[0], $Group_a->{login_list}[1] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };

	$All_quota_specs[6] = {
		type 	 => 'group',
		Desc     => "Quota[6]: Group; NFS group; Volume",
		group	 => $Group_a->{group},
		tree	 => '-',
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users           => [ $Group_a->{login_list}[0], $Group_a->{login_list}[1] ],
				Dirs            => [ $Dir_for_code{'1a'}->{name}, $Dir_for_code{'2a'}->{name} ],
				Protos          => [ 'nfs' ],
			    },
	     };


	$All_quota_specs[7] = {
		type 	 => 'tree',
		Desc     => "Quota[7]: Qtree; NFS users; (Qtree)",
		tree	 => $Qtree_1->{name},
		default  => 0,
		usermap  => 0,
		vol	 => $Vol,
		Misc	 => {
				Users          => [ $Group_a->{login_list}[0], $Group_a->{login_list}[1] ],
				Dirs           => [ $Dir_for_code{'1a'}->{name} ],
				Protos         => [ 'nfs' ],
			    },
	      };


	if ($USE_CIFS) {

		$All_quota_specs[8] = {
			type  	 => 'user',
			Desc     => "Quota[8]: User; Single CIFS user; Qtree",
			user	 => $Group_c->{login_list}[0],
			tree	 => $Qtree_3->{name},
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_c->{login_list}[0] ],
					Dirs            => [ $Dir_for_code{'3a'}->{name} ],
					Protos          => [ 'cifs' ],
				    },
		     };

		$All_quota_specs[9] = {
			Desc     => "Quota[9]: User; Single CIFS user; Volume",
			type  	 => 'user',
			user	 => $Group_c->{login_list}[0],
			tree	 => '-',
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_c->{login_list}[0] ],
					Dirs            => [ $Dir_for_code{'3a'}->{name}, $Dir_for_code{'4a'}->{name} ],
					Protos          => [ 'cifs' ],
				    },
		     };


		$All_quota_specs[10] = {
			type 	 => 'tree',
			Desc     => "Quota[10]: Qtree; CIFS users; (Qtree)",
			tree	 => $Qtree_3->{name},
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_c->{login_list}[0], $Group_c->{login_list}[1] ],
					Dirs            => [ $Dir_for_code{'3a'}->{name} ],
					Protos          => [ 'cifs' ],
				    },
		      };


		$All_quota_specs[11] = {
			type  	 => 'user',
			Desc     => "Quota[11]: User; Mapped user pair; Qtree",
			user	 => $Group_e->{login_list}[0],
			tree	 => $Qtree_5->{name},
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_e->{login_list}[0] ],
					Dirs            => [ $Dir_for_code{'5a'}->{name} ],
					Protos          => [ 'nfs', 'cifs' ],
				    },
		     };


		$All_quota_specs[12] = {
			type  	 => 'user',
			Desc     => "Quota[12]: User; Mapped user pair; Volume",
			user	 => $Group_e->{login_list}[0],
			tree	 => '-',
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_e->{login_list}[0] ],
					Dirs            => [ $Dir_for_code{'5a'}->{name}, $Dir_for_code{'6a'}->{name} ],
					Protos          => [ 'nfs', 'cifs' ],
				    },
		     };


		$All_quota_specs[13] = {
			type 	 => 'tree',
			Desc     => "Quota[13]: Qtree; Mapped user pair; Qtree",
			tree	 => $Qtree_5->{name},
			default  => 0,
			usermap  => 0,
			vol	 => $Vol,
			Misc	 => {
					Users           => [ $Group_e->{login_list}[0], $Group_e->{login_list}[1] ],
					Dirs            => [ $Dir_for_code{'5a'}->{name} ],
					Protos          => [ 'nfs', 'cifs' ],
				    },
		      };

	}


        foreach my  $qspec (@All_quota_specs) {
		#  Check validity of each quota
		if (!$qspec->{tree}) {
			logDbg("create_quota_specs  Line: ".__LINE__.
			          "\nSetting missing 'tree' option in quota spec to '-'\n\n".Dumper($qspec));
			$qspec->{tree} = '-';
		}
        }


        my $step_msg =
             "\n\n==============================  Quotas  =================================\n".
             "\n".
             "  Quotas\n".
             "\n";

	 for (my $i=0; $i<=$#All_quota_specs; $i++) {
	      $step_msg .= "      \$All_quota_specs[$i]\n\n";
	      $step_msg .= Dumper($All_quota_specs[$i]);
	      $step_msg .= "\n\n";
         }

         $step_msg .=
              "\n=============================================================================";

       	 logStep($step_msg);
    };

#    logcomment("create_quota_specs  Line: ".__LINE__."   ....");
#    for (my $i=1; $i<=$#All_quota_specs; $i++) {
#	$All_quota_specs[$i] = $All_quota_specs[0];
#    }
#    $Bad_quota_spec = $All_quota_specs[0];
#    logcomment("....  create_quota_specs  Line: ".__LINE__);

    my $fatal_msg = $@;
    if ($fatal_msg) {
       logcomment($fatal_msg);
       die($fatal_msg);
    }
#    logcomment("Exit  create_quota_specs\n\nAll_quotas:\n\n".Dumper(\@All_quota_specs));
#    logcomment("Exit create_quota_specs\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
    logcomment("Exit  create_quota_specs");
    return("");
}



sub calculate_sizes_and_limits(%)
{
	my (%opts) = @_;
	my $Fc                  = $opts{Fc};
	my $Vol                 = $opts{vol};
           $DISK_QUOTA_LIMIT    = $opts{disk_quota_limit}   ||  $main::DISK_QUOTA_LIMIT;
           $FILES_QUOTA_LIMIT   = $opts{files_quota_limit}  ||  $main::FILES_QUOTA_LIMIT;
        logcomment("Enter calculate_sizes_and_limits");
	my ( $msg );

	eval {
		 #  For debugging only ....
		 my $skip_avail_space_calc = 0;
		 if ($skip_avail_space_calc) {
        	 	$msg = "calculate_sizes_and_limits  Line: ".__LINE__."\n".
				     "'skip_avail_space_calc' is set";
#			$Log->warn($msg);
			logcomment($msg);
		 }

		my ($Large_files_size_kb, $Large_files_alloc_kb, $Large_files_num, $Large_files_prefix,
	             $Small_files_size_kb, $Small_files_alloc_kb, $Small_files_num, $Small_files_prefix,
	             $Jumbo_files_size_kb, $Jumbo_files_alloc_kb, $Jumbo_files_num, $Jumbo_files_prefix);

		# my ($Disk_quota_limit_kb, $Files_quota_limit);

      		my ($File_size_base_kb, $Min_file_size_kb, $Min_file_num);
                # my ($Vol_avail_kb, $Vol_avail_files);

		$Min_file_size_kb = 128;
		$Min_file_num = 10;
		$Min_file_size_kb -= ($Min_file_size_kb % 32);

		 $Disk_quota_limit_kb = convert_units( strip_unit_spec($DISK_QUOTA_LIMIT),
					                guess_unit($DISK_QUOTA_LIMIT),
						        UNIT_KB );

		 $Disk_quota_limit_kb -= ($Disk_quota_limit_kb % 32);
		 $Files_quota_limit = convert_to_bytes($FILES_QUOTA_LIMIT);
	
		if (!$skip_avail_space_calc) {
		         my ($vol_disk_df, $vol_inode_df);
			 $vol_disk_df = vol_disk_free(
						     $Fc,
						     path          => $Vol,
						     space_type    => DF_SPACE,
						     report_type   => UNIT_KB
		                                  );
			die("vol_disk_free: $@") if ($@);
			$Vol_avail_kb = $vol_disk_df->{avail};
		
			$vol_inode_df = vol_disk_free(
						     $Fc,
						     path          => $Vol,
						     space_type    => DF_INODES,
						     report_type   => UNIT_B
		                                  );
			die("vol_disk_free: $@") if ($@);
			$Vol_avail_files = $vol_inode_df->{ifree};
		
			if ($Disk_quota_limit_kb > int($Vol_avail_kb/2.5)) {
				$Disk_quota_limit_kb = int($Vol_avail_kb/2.5);
			}
			if ($Files_quota_limit > int($Vol_avail_files/2.5)) {
				$Files_quota_limit = int($Vol_avail_files/2.5);
			}
		}
		$Disk_quota_limit_kb -= ($Disk_quota_limit_kb % 32);
	
		$File_size_base_kb = int($Disk_quota_limit_kb / $Files_quota_limit);
	###	$File_size_base_kb -= ($File_size_base_kb % 32);
	
		if ($File_size_base_kb < $Min_file_size_kb) {
			$File_size_base_kb = $Min_file_size_kb;
			$Files_quota_limit = int($Disk_quota_limit_kb / $File_size_base_kb);
		}
		if ($Files_quota_limit < $Min_file_num) {
			$msg = "CONFIG: Line: ".__LINE__.
			            "   Insufficient space available on filer for:\n".
			            "Min files: $Min_file_num   and/or\n".
			            "Min file size: $Min_file_size_kb";
			die($msg);
		}
	
		logcomment("   Line: ".__LINE__.
		    "\n\nDISK_QUOTA_LIMIT: \"$DISK_QUOTA_LIMIT\"\n".
		    "FILES_QUOTA_LIMIT: \"$FILES_QUOTA_LIMIT\"\n".
		    "Min_file_size_kb: \"$Min_file_size_kb\"\n".
		    "Vol_avail_kb: \"$Vol_avail_kb\"\n".
		    "Vol_avail_files: \"$Vol_avail_files\"\n".
	            "File_size_base_kb: \"$File_size_base_kb\"\n".
		    "Files_quota_limit: \"$Files_quota_limit\"\n".
		    "Disk_quota_limit_kb: \"$Disk_quota_limit_kb\"");


		my $max_size_alloc_kb = $File_size_base_kb * $Files_quota_limit * 2;
		$Large_files_size_kb = int(1.3 * $File_size_base_kb);
		$Large_files_size_kb -= ($Large_files_size_kb % 32);
		$Small_files_size_kb = int(0.75 * $File_size_base_kb);
		$Small_files_size_kb -= ($Small_files_size_kb % 32);
		if (($Large_files_size_kb - $File_size_base_kb) < 64) {
			$Large_files_size_kb = $File_size_base_kb + 64;
		}
		if ( ($Small_files_size_kb < 64)  ||
		    (($File_size_base_kb - $Small_files_size_kb) < 64) )  {
			$Small_files_size_kb = 64;
		}
		$Large_files_num = ($Files_quota_limit - 5);
		$Small_files_num = ($Files_quota_limit + 5);

		$Large_files_alloc_kb = ($Large_files_size_kb * $Large_files_num);
		$Small_files_alloc_kb = ($Small_files_size_kb * $Small_files_num);
		if ($Disk_quota_limit_kb > $Large_files_alloc_kb) {
			$Disk_quota_limit_kb = int(0.8 * $Large_files_alloc_kb);
		}
		if ($Files_quota_limit > $Small_files_num) {
			$Files_quota_limit = int(0.8 * $Small_files_num);
		}
		if ($Files_quota_limit > $Small_files_num) {
			die("calculate_sizes_and_limits  Line: ".__LINE__.
			      "\nError calculating sizes and limits\n".
			      "Small_files_num: $Small_files_num\n".
			      "Files_quota_limit: $Files_quota_limit\n");
		}
		if ($Disk_quota_limit_kb > $Large_files_alloc_kb) {
			die("calculate_sizes_and_limits  Line: ".__LINE__.
			      "\nError calculating sizes and limits\n".
			      "Large_files_alloc_kb: $Large_files_alloc_kb\n".
			      "Disk_quota_limit_kb: $Disk_quota_limit_kb\n");
		}
		if ( ($Disk_quota_limit_kb <= 0) || ($Files_quota_limit <= 0) ) {
			die("calculate_sizes_and_limits  Line: ".__LINE__.
			      "\nError calculating sizes and limits\n".
			      "Disk_quota_limit_kb: $Disk_quota_limit_kb\n".
			      "Files_quota_limit: $Files_quota_limit\n");
		}
		if ( ($Small_files_num <= $Large_files_num) ||
		     ($Small_files_num <= 0) || ($Large_files_num <= 0) ) {
			die("calculate_sizes_and_limits  Line: ".__LINE__.
			      "\nError calculating sizes and limits\n".
			      "Small_files_num: $Small_files_num\n".
			      "Large_files_num: $Large_files_num\n");
		}
		if ( ($Small_files_size_kb >= $Large_files_size_kb) ||
		     ($Small_files_size_kb <= 0) || ($Large_files_size_kb <= 0) ) {
			die("calculate_sizes_and_limits  Line: ".__LINE__.
			      "\nError calculating sizes and limits\n".
			      "Small_files_size_kb: $Small_files_size_kb\n".
			      "Large_files_size_kb: $Large_files_size_kb\n");
		}

		$Large_files_prefix = "Large_Files_";
		$Small_files_prefix = "Small_Files_";
		$Jumbo_files_prefix = "Jumbo_Files_";

		$Jumbo_files_num        = $Small_files_num;
		$Jumbo_files_size_kb    = $Large_files_size_kb;
		$Jumbo_files_alloc_kb   = ($Jumbo_files_size_kb * $Jumbo_files_num);


		my $Large_files_list = [];
                $Large_file_set = {
					    name		 => "Large",
					    desc                 => "'large' files set -- trigger 'disk' quota",
				            size_kb              => $Large_files_size_kb,
			 	            num_files            => $Large_files_num,
			 	            expected_alloc_kb    => $Large_files_alloc_kb,
			 	            file_prefix          => $Large_files_prefix,
			 	            list                 => $Large_files_list,
			     };

		my $Small_files_list = [];
                $Small_file_set = {
					    name		 => "Small",
					    desc                 => "'small' files set -- trigger 'files' quota",
				            size_kb              => $Small_files_size_kb,
			 	            num_files            => $Small_files_num,
			 	            expected_alloc_kb    => $Small_files_alloc_kb,
			 	            file_prefix          => $Small_files_prefix,
			 	            list                 => $Small_files_list,
			     };

		my $Jumbo_files_list = [];
                $Jumbo_file_set = {
					    name		 => "Jumbo",
					    desc                 => "'jumbo' files set -- trigger both 'files' and 'disk' quotas",
				            size_kb              => $Jumbo_files_size_kb,
			 	            num_files            => $Jumbo_files_num,
			 	            expected_alloc_kb    => $Jumbo_files_alloc_kb,
			 	            file_prefix          => $Jumbo_files_prefix,
			 	            list                 => $Jumbo_files_list,
			     };

		$Default_file_set = $Jumbo_file_set;
		$File_set = undef;


	logcomment("   Line: ".__LINE__.
		    "\n\nDISK_QUOTA_LIMIT: \"$DISK_QUOTA_LIMIT\"\n".
		    "FILES_QUOTA_LIMIT: \"$FILES_QUOTA_LIMIT\"\n".
		    "Min_file_size_kb: \"$Min_file_size_kb\"\n".
		    "Vol_avail_kb: \"$Vol_avail_kb\"\n".
		    "Vol_avail_files: \"$Vol_avail_files\"\n".
	            "File_size_base_kb: \"$File_size_base_kb\"\n".
		    "Files_quota_limit: \"$Files_quota_limit\"\n".
		    "Disk_quota_limit_kb: \"$Disk_quota_limit_kb\"\n".
		    "Small_files_size_kb: \"$Small_files_size_kb\"\n".
		    "Small_files_num: \"$Small_files_num\"\n".
		    "Small_files_alloc_kb: \"$Small_files_alloc_kb\"\n".
		    "Large_files_size_kb: \"$Large_files_size_kb\"\n".
		    "Large_files_num: \"$Large_files_num\"\n".
		    "Large_files_alloc_kb: \"$Large_files_alloc_kb\"\n".
		    "Jumbo_files_size_kb: \"$Jumbo_files_size_kb\"\n".
		    "Jumbo_files_num: \"$Jumbo_files_num\"\n".
		    "Jumbo_files_alloc_kb: \"$Jumbo_files_alloc_kb\"\n".
		"");	


        my $step_msg =
             "\n\n========================    File 'Sets'   ==========================\n".
             "\n";

         my %fileset_hash = ( "Small" => $Small_file_set, "Large" => $Large_file_set, "Jumbo" => $Jumbo_file_set );
#	 while (my($k,$v) = each(%fileset_hash) ) {
#	      $step_msg .= "      \$k\n\n";
#	      $step_msg .= Dumper($v);
#	      $step_msg .= "\n\n";
#         }
	 $step_msg .= "\n\n" . Dumper(\%fileset_hash) . "\n\n";

         $step_msg .=
              "\n=============================================================================";

       	 logStep($step_msg);
	};

	my $fatal_msg = $@;
	if ($fatal_msg) {
		logcomment("Exit  calculate_sizes_and_limits\nErrMsg: \"$fatal_msg\"");
		return ($fatal_msg);
	}
	logcomment("Exit  calculate_sizes_and_limits\nErrMsg: \"$fatal_msg\"");
#        logcomment("Exit calculate_sizes_and_limits\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return("");
}



sub setup_cifs(%)
{
	my (%opts) = @_;
	logcomment("Enter setup_cifs");

	eval {
		my ($msg, $rslt);
		my $Fc = $opts{Fc}
		    or die("setup_cifs:  'Fc' option must be specified");
		my $Vol               = $opts{vol}               || $main::FILER_VOL;
		   $cifs_passwd       = $opts{cifs_passwd}       || $main::CIFS_PASSWD;
		my $cifs_workgroup    = $opts{cifs_workgroup}    || $main::CIFS_WORKGROUP;
		my $cifs_share        = $opts{cifs_share}        || $main::CIFS_PASSWD;
		my $cifs_client       = $opts{cifs_client}       || $main::CIFS_CLIENT;
		my $cifs_user         = $opts{cifs_user}         || $main::CIFS_USER;

		my $Fh = $Fc->hostp();
		my $Filer_name = $Fh->hostname();
		my $Ip_addr = $Fh->default_ip();
#		@CIFS_groups = ();
#		@CIFS_users = ();
		@All_cifs_shares = ();
		@All_cifs_drives = ();
#		@Mapped_groups = ();
#		@Mapped_users = ();
		my $FILER = 

		logcomment("setup_cifs  Line: ".__LINE__.
		             "   Verify $cifs_client is MinWin compatible");
		if (!minwin_compatible($cifs_client)) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			          "CIFS client $cifs_client is non MinWin compatible";
			die($msg);
		}
		logcomment("setup_cifs  Line: ".__LINE__."   license CIFS");
		$rslt = license($Fc, 'cifs');
		if (!$rslt || $@) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			          "license CIFS failed: $@";
			die($msg);
		}

		logcomment("setup_cifs  Line: ".__LINE__."   Call 'cifs_setup'");
		$rslt = cifs_setup(    host        => $Filer_name,
			               cifs_pass   => $cifs_passwd,
		                       workgroup   => $cifs_workgroup,
		         );
		if (!$rslt || $@) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			          "cifs_setup failed: $@";
			die($msg);
		}

	 	$rslt = writeUsermapFile($Fc, "");
		if (!$rslt || $@) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			         "writeUsermapFile failed: $@";
			die($msg);
		}

		logcomment("setup_cifs  Line: ".__LINE__.
		              "   Mapping 'administrator' to 'root'");
		$rslt = mapUsers($Fc, "administrator", "root");
		if (!$rslt || $@) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			         "mapUers failed: $@";
			die($msg);
		}
		logcomment("setup_cifs  Line: ".__LINE__.
		               "   Setting up CIFS share");
		if (share_exists($Fc, $cifs_share)) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
				   "Share \"$cifs_share\" already exists";
#			die($msg);
#		        # $Log->warn($msg);
#			$rslt = share_delete($Fc, $CIFS_SHARE);
#			if (!$rslt || $@) {
#				$msg = "setup_cifs  Line: ".__LINE__."\n".
#				           "Could not delete share $CIFS_SHARE: $@";
#				die($msg);
#			}
		}

		if (!share_exists($Fc, $cifs_share)) {
			logcomment("setup_cifs  Line: ".__LINE__.
			               "   Call 'share_add'");
			$rslt = share_add( $Fc,
				           $cifs_share,
				           $Filer_paths{$Vol},
				           verify=>1);
			if (!$rslt || $@) {
				$msg = "setup_cifs  Line: ".__LINE__."\n".
				           "share_add failed: $@";
				die($msg);
			}
		}

		push(@All_cifs_shares, $cifs_share);
		logcomment("setup_cifs  Line: ".__LINE__.
		               "   Call 'map_cifs_share'");
		my $new_drive = map_cifs_share(
					server		=> $Ip_addr,
					client		=> $cifs_client,
					drive		=> "*",
					share		=> $cifs_share,
					workgroup	=> $cifs_workgroup,
					password	=> $cifs_passwd,
					username	=> $cifs_user,
				);
		if (!$new_drive) {
			$msg = "setup_cifs  Line: ".__LINE__."\n".
			             "map_cifs_share failed: $@";
			die($msg);
		}

		$new_drive .= ":" unless ($new_drive =~ /:$/);
		push(@All_cifs_drives, $new_drive);
		$CIFS_paths{$Vol} = $new_drive . "/";
		logcomment("setup_cifs  Line: ".__LINE__.
		                 "\n\nNew CIFS drive: \"$new_drive\"");
	};

	my $fatal_msg = $@;
	logcomment("\nExit  setup_cifs\nErrMsg: \"$fatal_msg\"\n");
#        logcomment("Exit setup_cifs\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return $fatal_msg;
}



sub scenario_quota_exceeded_create(%)
{
        my (%opts) = @_;
	my $Fc   = $opts{Fc};
	my $Vol  = $opts{vol};

        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_exceeded_create");

        eval {
	    my (
		$qr, $qtree, @dir_entries,
		   $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
	    unless ($opts{initialize_quota_values_called}) {
	        my $opts_ref = initialize_quota_values(%opts);
	        %opts = %$opts_ref;
            }

 	    my $h_or_s              = $opts{h_or_s};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); #############

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	     unless ($opts{initialize_file_values_called}) {
	        my $opts_ref = initialize_file_values(%opts);
	        %opts = %$opts_ref;
	     }
#	     my $file_set           = $opts{file_set};
	     my $num_files          = $opts{num_files};
	     my $size_bytes         = $opts{size_bytes};
	     my $file_prefix        = $opts{file_prefix};
             my $expected_alloc_kb  = $opts{expected_alloc_kb};
	     logDbg("scenario_quota_exceeded_create  Line: ".__LINE__.
		      "\n\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "File prefix: \"$file_prefix\"\nExpeted alloc: $expected_alloc_kb Kb");


		## @step  Add quota with limits set
		logStep("Add quota with limits set");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'set_quota' ....");
		$msg = set_quota(Fc => $Fc, %opts);
		if ($msg) {
			$msg = "FATAL: scenario_quota_exceeded_create  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'set_quota'");

		## @step  Start EMS logging, and turn on quotas
		logStep("Start EMS logging, and turn on quotas");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'markEMSLog' ....");
		$rslt = markEMSLog($Fc);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_exceeded_create  Line: ".__LINE__.
			           "\nmarkEMSLog:\n$@";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'markEMSLog'");

		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'enableQuota' ....");
		$rslt = enableQuota($Fc, $Vol, reenable=>1);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_exceeded_create  Line: ".__LINE__.
			          "\nenableQuota:\n$@";
		       	logcomment($msg);
			die($msg);
		}
sleep(60); #################################
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'enableQuota'");

		## @step  Verify directories involved in quota report are empty, and quota report values are as expected
		logStep("Verify directories involved in quota report are empty, and quota report values are as expected");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'verify_initial_quota_state' ....");

logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ###############


### Start DBG
		($fail_list, $msg) = verify_initial_quota_state($Fc,
				                          quota_spec=>$quota_spec,
						          hard=>$hard,
							  soft=>$soft,
						          disk=>$disk,
						          files=>$files);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_exceeded_create  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		push(@$fail_msg_list, @$fail_list);
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'verify_initial_quota_state'");
### End DBG
logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ##############


		## @step  Write files to trigger quota'
		logStep("Write files to trigger quota'");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'write_quota_files' ....");
		$msg = write_quota_files( %opts );
		if ($msg) {
		       	$msg = "FATAL:  scenario_quota_exceeded_create  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'write_quota_files'");
		sleep(5);

		## @step  Verify 'quota report' after writing files
		logStep("Verify 'quota report' after writing files");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
	        	$msg = "FATAL:  scenario_quota__exceeded_create  Line: ".__LINE__."\n$msg";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'get_quota_report'");

		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Verify quota limits from report ....");
		if ($qr->{files_used} != $num_files) {
	        	$msg = "FAIL:  quota_limit_exceeded_create  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"$num_files\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if (get_rel_diff_percent($qr->{disk_used}, $expected_alloc_kb) > 5) {
	        	$msg = "FAIL:  quota_limit_exceeded_create  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"$expected_alloc_kb\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Finished verifying quota limits from report");

		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after writing files
		logStep("Verify  'EMS' log after writing files");
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   Call 'verify_ems_quota_limit_exceeded ....");
		my $msg_in = "FAIL: scenario_quota_exceeded_create  Line: ".__LINE__."  After writing files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_exceeded(
						        msg_in    => $msg_in,
							Fc        => $Fc,
							disk      => $disk,
							files     => $files,
							hard      => $hard,
							soft      => $soft,
						);

		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_exceeded_create  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_exceeded_create  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_exceeded");
		sleep(5);


#		$rslt = restoreFilesPreserveBackups($Rootvol_mntpt, '/etc/quotas');
#         	if (!$rslt || $@) {
#			$msg = "FATAL:  scenario_quota_exceeded_create  Line: ".__LINE__."\n$@";
#			die($msg);
#		}

	};

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota__limit_exceeded_create\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub scenario_quota_disk_limit_exceeded_append(%)
{
        my (%opts) = @_;
	my $Fc   = $opts{Fc};
	my $Vol  = $opts{vol};

        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_disk_limit_exceeded_append");

        eval {
	    my (
		$qr, $qtree, @dir_entries,
		   $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
	    unless ($opts{initialize_quota_values_called}) {
	        my $opts_ref = initialize_quota_values(%opts);
	        %opts = %$opts_ref;
            }

 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $file_set       	    = $opts{file_set} || $File_set || $Default_file_set;
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

	    if (!$disk) {
	   	$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n\n".
	                "This scenario only valid for 'disk' quota limits";
	       	logcomment($msg);
		die($msg);
	    }

logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); #############

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	     unless ($opts{initialize_file_values_called}) {
	        my $opts_ref = initialize_file_values(%opts);
	        %opts = %$opts_ref;
	     }
#	     my $file_set            = $opts{file_set};
	     my $num_files          = $opts{num_files};
	     my $size_bytes         = $opts{size_bytes};
	     my $file_prefix        = $opts{file_prefix};
             my $expected_alloc_kb  = $opts{expected_alloc_kb};
	     logDbg("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__.
		      "\n\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "File prefix: \"$file_prefix\"\nExpeted alloc: $expected_alloc_kb Kb");

		## @step  Add quota with limits set
		logStep("Add quota with limits set");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'set_quota' ....");
		$msg = set_quota(Fc => $Fc, %opts);
		if ($msg) {
			$msg = "FATAL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'set_quota'");

		## @step  Start EMS logging, and turn on quotas
		logStep("Start EMS logging, and turn on quotas");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'markEMSLog' ....");
		$rslt = markEMSLog($Fc);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__.
			           "\nmarkEMSLog:\n$@";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'markEMSLog'");

		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'enableQuota' ....");
		$rslt = enableQuota($Fc, $Vol, reenable=>1);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__.
			          "\nenableQuota:\n$@";
		       	logcomment($msg);
			die($msg);
		}
sleep(60); #################################
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'enableQuota'");


		## @step  Verify directories involved in quota report are empty, and quota report values are as expected
		logStep("Verify directories involved in quota report are empty, and quota report values are as expected");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'verify_initial_quota_state' ....");

logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ###############


## Start DBG
		($fail_list, $msg) = verify_initial_quota_state($Fc,
				                          quota_spec=>$quota_spec,
						          hard=>$hard,
							  soft=>$soft,
						          disk=>$disk,
						          files=>$files);
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'verify_initial_quota_state'");
### End DBG
logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ##############


		## @step  Create files initially with zero (or very small) size  -- disk quota will not be exceeded'
		logStep("Write files to trigger quota'");

		my %create_file_opts = %opts;
		$create_file_opts{size} = 1;
		my %append_file_opts = %opts;

		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call(1) 'write_quota_files' ....");
		$msg = write_quota_files( %create_file_opts );
		if ($msg) {
		       	$msg = "FATAL:  scenario_quota_exceeded_append  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned(1) from 'write_quota_files'");
		sleep(5);


	       ## @step Truncate files just created
	       logStep("Truncate files just created");
               logcomment("quota_limit_exceeded_append  Line: ".__LINE__."   Truncate files affected by quota ....");
	       foreach my $dir (@$dir_list) {
	        	logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'truncate_testdir_files' ....");
			$msg = truncate_testdir_files(proto=>$proto, dir=>$dir, size=>0);
			if ($msg) {
				$msg = "FATAL: scenario_quota_disk_limit_exceeded_append  Line: ".
					 __LINE__."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
				die($msg);
		        }
	        	logcomment("scneario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'truncate_testdir_files'");
		}
                logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Finished truncating files affected by quota");
logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Time: ".`date`);
		sleep(5);
logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Time: ".`date`);


		## @step  Verify 'quota report' after creating files
		logStep("Verify 'quota report' after creating files");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
	        	$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n$msg";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'get_quota_report'");

		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Verify quota limits from report ....");
		if ($qr->{files_used} != $num_files) {
	        	$msg = "FAIL:  quota_disk_limit_exceeded_append  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"$num_files\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Finished verifying quota limits from report");
		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after creating files
		logStep("Verify  'EMS' log after writing files");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'verify_ems_quota_limit_unchanged ....");
		my $msg_in_1 = "FAIL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."  After writing files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_unchanged(
						        msg_in     => $msg_in_1,
							Fc         => $Fc,
							disk       => $disk,
							files      => $files,
							soft       => $soft,
							hard       => $hard,
						);

####		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_unchanged");
		sleep(5);

		## @step  Append/grow files to trigger quota'
		logStep("Append/grow files to trigger quota'");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call(2) 'write_quota_files' ....");
		$msg = write_quota_files( %append_file_opts );
		if ($msg) {
		       	$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned(2) from 'write_quota_files'");
		sleep(5);


		## @step  Verify 'quota report' after writing files
		logStep("Verify 'quota report' after writing files");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
	        	$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n$msg";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'get_quota_report'");

		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Verify quota limits from report ....");
		if ($qr->{files_used} != $num_files) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"$num_files\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if (get_rel_diff_percent($qr->{disk_used}, $expected_alloc_kb) > 5) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"$expected_alloc_kb\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Finished verifying quota limits from report");

		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after appending/growing files
		logStep("Verify  'EMS' log after writing files");
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   Call 'verify_ems_quota_limit_exceeded ....");
		my $msg_in_2 = "FAIL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."  After writing files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_exceeded(
						        msg_in      => $msg_in_2,
							Fc          => $Fc,
							disk        => $disk,
							files       => $files,
							hard        => $hard,
							soft        => $soft,
						);

		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_exceeded");
		sleep(5);

#		$rslt = restoreFilesPreserveBackups($Rootvol_mntpt, '/etc/quotas');
#         	if (!$rslt || $@) {
#			$msg = "FATAL:  scenario_quota_disk_limit_exceeded_append  Line: ".__LINE__."\n$@";
#			die($msg);
#		}

	};

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_disk_limit_exceeded_append\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}




sub scenario_quota_exceeded_move(%)
{
	return scenario_quota_exceeded_create(@_);
}

sub scenario_quota_exceeded_chown(%)
{
	return scenario_quota_exceeded_create(@_);
}

sub scenario_quota_exceeded_chgrp(%)
{
	return scenario_quota_exceeded_create(@_);
}



sub scenario_quota_resize(%)
{
        my (%opts) = @_;
	my $Fc                           = $opts{Fc};
	my $Vol                          = $opts{vol};
	my $resize_events     		 = $opts{resize_events} || $opts{resize_event};
        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_resize");

        eval {
	    my (
		$qr, $qtree, @dir_entries,
		   $fail_list, $msg, $fatal_msg, $rslt,
	    );

	        if (!defined($resize_events)) {
			$resize_events = [ 'start', 'finish' ];
		} elsif (ref($resize_events)) {
			if (ref($resize_events) =~ /ARRAY/) {
				if (grep( {($_ ne 'start') && ($_ !~ /finish/)}  @$resize_events ) ) {
					$msg = "scenario_quota_resize  Line: ".__LINE__.
						 "\nExpected 'resize_events'\n".
						  "to be as string, or an ARRAYREF with possible members 'start', ".
						  "and/or 'finis(ed)h'\n";
					logcomment($msg);
					die($msg);
				}
			} else {
				$msg = "scenario_quota_resize  Line: ".__LINE__."\nExpected 'resize_events'\n".
					  "to be a string, or an ARRAYREF with possible members 'start', and/or 'finish(ed)'\n";
				logcomment($msg);
				die($msg);
			}
		} else {
 			if (($resize_events ne 'start') && ($resize_events !~ /finish/)) {
				$msg = "scenario_quota_resize  Line: ".__LINE__."\nExpected 'resize_events'\n".
					  "to be a string, or an ARRAYREF with possible members 'start', and/or 'finish(ed)'\n";
				logcomment($msg);
				die($msg);
			} else {
				$resize_events = [ $resize_events ];
			}
		}


	    #  The calling routine should perform the next statement sometime before calling this routine
	    unless ($opts{initialize_quota_values_called}) {
	        my $opts_ref = initialize_quota_values(%opts);
	        %opts = %$opts_ref;
            }

 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $file_set       	    = $opts{file_set} || $File_set || $Default_file_set;
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

	    if (!$disk) {
	   	$msg = "FATAL:  scenario_quota_resize  Line: ".__LINE__."\n\n".
	                "This scenario only valid for 'disk' quota limits";
	       	logcomment($msg);
		die($msg);
	    }

logcomment("scenario_quota_resize  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); #############

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	     unless ($opts{initialize_file_values_called}) {
	        my $opts_ref = initialize_file_values(%opts);
	        %opts = %$opts_ref;
	     }
#	     my $file_set            = $opts{file_set};
	     my $num_files          = $opts{num_files};
	     my $size_bytes         = $opts{size_bytes};
	     my $file_prefix        = $opts{file_prefix};
             my $expected_alloc_kb  = $opts{expected_alloc_kb};
	     logDbg("scenario_quota_resize  Line: ".__LINE__.
		      "\n\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "File prefix: \"$file_prefix\"\nExpeted alloc: $expected_alloc_kb Kb");

		## @step  Add quota with limits set
		logStep("Add quota with limits set");
		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'set_quota' ....");
		$msg = set_quota(Fc => $Fc, %opts);
		if ($msg) {
			$msg = "FATAL: scenario_quota_resize  Line: ".__LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'set_quota'");

		## @step  Start EMS logging, and turn on quotas
		logStep("Start EMS logging, and turn on quotas");
		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'markEMSLog' ....");
		$rslt = markEMSLog($Fc);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_resize  Line: ".__LINE__.
			           "\nmarkEMSLog:\n$@";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'markEMSLog'");

		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'enableQuota' ....");
		$rslt = enableQuota($Fc, $Vol, reenable=>1);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_resize  Line: ".__LINE__.
			          "\nenableQuota:\n$@";
		       	logcomment($msg);
			die($msg);
		}
sleep(60); #################################
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'enableQuota'");


		## @step  Verify directories involved in quota report are empty, and quota report values are as expected
		logStep("Verify directories involved in quota report are empty, and quota report values are as expected");
		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'verify_initial_quota_state' ....");

logcomment("scenario_quota_resize  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ###############


## Start DBG
		($fail_list, $msg) = verify_initial_quota_state($Fc,
				                          quota_spec=>$quota_spec,
						          hard=>$hard,
							  soft=>$soft,
						          disk=>$disk,
						          files=>$files);
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_resize  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'verify_initial_quota_state'");
### End DBG
logcomment("scenario_quota_resize  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)); ##############



		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'resizeQuota' ....");
		$rslt = resizeQuota($Fc, $Vol);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_quota_resize  Line: ".__LINE__.
			          "\nresizeQuota:\n$@";
		       	logcomment($msg);
			die($msg);
		}
sleep(60); #################################
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'resizeQuota'");

		logcomment("scenario_quota_resize  Line: ".__LINE__."   Call 'search_ems_records' ....");
		my ($matched_recs) = search_ems_records(Fc => $Fc, patt => 'resize');
		logcomment("scenario_quota_resize  Line: ".__LINE__."   .... Returned from 'search_ems_records'".
			       "\n\nReturned:\n\n".Dumper($matched_recs));

        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_resize\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



############################################################################
#
#   Quota::writeQuotaFile($Fc, @quota_lines)
#
#    $Fc             Filer connection
#    @quota_lines    ARRAY of text lines to be written to /etc/quotas
#
#    Returns '1' on success
#
# sub writeQuotaFile($@);
#
############################################################################

sub scenario_bad_quotas_file(%)
{
        my (%opts) = @_;
	my $Fc                           = $opts{Fc};
	my $Vol                          = $opts{vol};
        my $fail_msg_list = [];
        logcomment("Enter scenario_bad_quotas_file");

        eval {
	    my (
		$qr, $qtree, @dir_entries,
		   $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    my $bad_quota = get_quota(0);
	    my %bad_quota_opts = %$bad_quota;

		## @step  Add 'bad' quota
		logStep("Add 'bad' quota");
		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   Call 'writeQuotaFile' ....");
		my $bad_quota_line = "once more into the breach dear friends";
		writeQuotaFile($Fc, $bad_quota_line)
		    or die("scenario_bad_quotas_file  Line: ".__LINE__."\nError from 'writeQuotaFile': $@");

#		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   Call 'set_quota' ....");
#		$msg = set_quota(Fc => $Fc, %bad_quota_opts);
#		if ($msg) {
#			$msg = "scenario_bad_quotas_file  Line: ".__LINE__.
#			           "\nExpected error from setting 'bad' quotas file: $msg";
#			logcomment($msg);
#		}
#		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   .... Returned from 'set_quota'");


		## @step  Start EMS logging, and turn on quotas
		logStep("Start EMS logging, and turn on quotas");
		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   Call 'markEMSLog' ....");
		$rslt = markEMSLog($Fc);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_bad_quotas_file  Line: ".__LINE__.
			           "\nmarkEMSLog:\n$@";
		       	logcomment($msg);
			die($msg);
		}
		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   .... Returned from 'markEMSLog'");

		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   Call 'enableQuota' ....");
		$rslt = enableQuota($Fc, $Vol, reenable=>1);
		if (!$rslt || $@) {
			$msg = "FATAL:  scenario_bad_quotas_file  Line: ".__LINE__.
			          "\nenableQuota:\n$@";
		       	logcomment($msg);
			die($msg);
		}
sleep(60); #################################
		logcomment("scenario_bad_quotas_file  Line: ".__LINE__."   .... Returned from 'enableQuota'");

        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_bad_quotas_file\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub scenario_quota_normal_delete(%)
{
        my (%opts) = @_;
	my $Fc = $opts{Fc};
        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_normal_delete");
        eval {
	    my (
		 $qr, $qtree, @dir_entries,
	          $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};


	    if (!$file_list || (ref($file_list) !~ /array/i) || (scalar(@$file_list) == 0)) {
		$msg = "FATAL:  scenario_quota_normal_delete  Line: ".__LINE__."\n\n".
		         "Must specify 'file_list'";
	       	logcomment($msg);
		die($msg);
	    }


	       ## @step Delete files just created
	       logStep("Delete files just created");
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Deleting testdirs for quota ....");
	       foreach my $dir (@$dir_list) {
			logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Call 'delete_testdir_files' ....");
			$msg = delete_testdir_files(proto=>$proto, dir=>$dir);
			if ($msg) {
				$msg = "FATAL: scenario_quota_normal_delete  Line: ".
					 __LINE__."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
			logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Returned from 'delete_testdir_files'");
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Finished deleting testdirs for quota");
		sleep(5);


		## @step  Verify 'quota report' after deleting
		logStep("Verify 'quota report' after deleting files");
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: scenario_quota_normal_delete  Line: ".
			           __LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Returned from 'get_quota_report'");

		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Verify 'used/allocated' values from quota report ....");
		if ($qr->{files_used} != 0) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Finished verifying 'used/allocated' values from quota report");


		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Finished sleeping");
		logStep("   .... Finished waiting for EMS messages, retrieve them");

		## @step  Verify 'EMS' log after deleting files
		logStep("Verify  'EMS' log after deleting files");
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   Call 'verify_ems_quota_limit_normal ....");
		my $msg_in = "FAIL: scenario_quota_normal_delete   Line: ".__LINE__."   After deleting files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_normal(
						        msg_in      => $msg_in,
							Fc          => $Fc,
							disk        => $disk,
							files       => $files,
							hard        => $hard,
							soft        => $soft,
						 );
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_normal_delete  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_normal");
		sleep(5);

#		## @step  Delete deleted files
#		logStep("Delete deleted files");

        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_limit_normal_delete\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub scenario_group_quota_normal_chgrp(%)
{
        my (%opts) = @_;
	my $Fc = $opts{Fc};
        my $fail_msg_list = [];
        logcomment("Enter scenario_group_quota_normal_chgrp\n\n".Dumper(\%opts));

        eval {
	    my (
		 $qr, $qtree, @dir_entries,
	         $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

#	    my $proto           = $$proto_list[0];
#	    my $group           = $$group_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

#	    my $quota_type = $quota_spec->{type};
 	    my $quota_type      = $opts{quota_type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	    if ($quota_type ne 'group') {
		$msg = "FATAL:  scenario_group_quota_normal_chgrp  Line: ".__LINE__."\n\n".
		       "Quota type must be 'group' (not \"$quota_type\")";
		       	logcomment($msg);
			die($msg);
	    }

	    if (!$file_list || (ref($file_list) !~ /array/i) || (scalar(@$file_list) == 0)) {
		$msg = "FATAL:  scenario_group_quota_normal_chgrp  Line: ".__LINE__."\n\n".
		         "Must specify 'file_list'";
	       	logcomment($msg);
		die($msg);
	    }

	    my $dest_group_list = [ @Null_groups ];
	    my $dest_group = $$dest_group_list[0];

	       ## @step Chgrp files just created
	       logStep("Chgrpfiles just created");
	       logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Chgrp files for quota ....");
	       foreach my $dir (@$dir_list) {
	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Call 'show_unix_testdir_files'(1) ....");
			$msg = show_unix_testdir_files(
						     	    msg_in      => "scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Before 'chgrp'",
						     	    proto       => $proto,
							    client      => $client,
							    dir         => $dir,
							    dest_group  => $dest_group,
							);
			if ($msg) {
				$msg = "FATAL: scenario_group_quota_normal_chgrp  Line: ". __LINE__ ."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Returned from 'show_unix_testdir_files'(1)");

	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Call 'chgrp_testdir_files' ....");
			$msg = chgrp_testdir_files(
						    proto       => $proto,
						    client      => $client,
						    dir         => $dir,
						    dest_group  => $dest_group,
					         );
			if ($msg) {
				$msg = "FATAL: scenario_group_quota_normal_chgrp  Line: ". __LINE__ ."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Returned from 'chgrp_testdir_files'");

	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Call 'show_unix_testdir_files'(2) ....");
			$msg = show_unix_testdir_files(
						     	    msg_in      => "scenario_group_quota_normal_chgrp  Line: ".__LINE__."   After 'chgrp'",
						     	    proto       => $proto,
							    client      => $client,
							    dir         => $dir,
							    dest_group  => $dest_group,
							);
			if ($msg) {
				$msg = "FATAL: scenario_group_quota_normal_chgrp  Line: ". __LINE__ ."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
	       	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Returned from 'chgrp_testdir_files'(2)");

		}
	       logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Finished chgrp-ing files for quota");
		sleep(5);


		## @step  Verify 'quota report' after 'chgrp-ing' files
		logStep("Verify 'quota report' after 'chgrp-ing' files");
	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: scenario_group_quota_normal_chgrp  Line: ".
			           __LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Returned from 'get_quota_report'");


	        logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Verify 'used/allocated' values from quota report ....");
		if ($qr->{files_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chgrp  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  quota_tlimit_exceeded  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Finished verifying 'used/allocated' values from quota report");


		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after 'chgrp-ing' files
		logStep("Verify  'EMS' log after 'chgrp-ing' files");
		logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   Call 'verify_ems_quota_limit_normal ....");
		my $msg_in = "FAIL: scenario_group_quota_normal_chgrp   Line: ".__LINE__."   After 'chgrp-ing' files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_normal(
						        msg_in                   => $msg_in,
							Fc                       => $Fc,
							disk                     => $disk,
							files                    => $files,
							hard                     => $hard,
							soft                     => $soft,
							check_upper_bound_only   => 1,
						 );
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_group_quota_normal_chgrp  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_group_quota_normal_chgrp  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_normal ....");
		sleep(5);


#		## @step  Delete chgrpd files
#		logStep("Delete chgrpd files");

        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_limit_normal_chgrp\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub scenario_user_quota_normal_chown(%)
{
        my (%opts) = @_;
	my $Fc = $opts{Fc};
        my $fail_msg_list = [];
        logcomment("Enter scenario_user_quota_normal_chown");

        eval {
	    my (
		 $qr, $qtree, @dir_entries,
	         $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

#	    my $quota_type = $quota_spec->{type};
 	    my $quota_type          = $opts{quota_type};

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	    if ($quota_type ne 'user') {
		$msg = "FATAL:  scenario_user_quota_normal_chown  Line: ".__LINE__."\n\n".
		       "Quota type must be 'user' (not \"$quota_type\")";
		       	logcomment($msg);
			die($msg);
	    }

	    if (!$file_list || (ref($file_list) !~ /array/i) || (scalar(@$file_list) == 0)) {
		$msg = "FATAL:  scenario_user_quota_normal_chown  Line: ".__LINE__."\n\n".
		         "Must specify 'file_list'";
	       	logcomment($msg);
		die($msg);
	    }

	    my $dest_user_list = [ @Null_users ];
	    my $dest_user = $$dest_user_list[0];

	       ## @step Chown files just created
	       logStep("Chown files just created");
	       logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Chown files for quota ....");
	       foreach my $dir (@$dir_list) {
	       	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Call 'chown_testdir_files' ....");
			$msg = chown_testdir_files(proto=>$proto, dir=>$dir, dest_user=>$dest_user);
			if ($msg) {
				$msg = "FATAL: scenario_user_quota_normal_chown  Line: ".
					 __LINE__."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
	       	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Returned from 'chown_testdir_files'");
		}
	       logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Finished chown-ing files for quota");
		sleep(5);


		## @step  Verify 'quota report' after chown-ing
		logStep("Verify 'quota report' after chown-ing files");
	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: scenario_user_quota_normal_chown  Line: ".
			           __LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Returned from 'get_quota_report'");
	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Verify 'used/allocated' values from quota report ....");
		if ($qr->{files_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chown  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chown  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
	        logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Finished verifying 'used/allocated' values from quota report");


		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after 'chown-ing' files
		logStep("Verify  'EMS' log after deleting files");
		logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   Call 'verify_ems_quota_limit_normal ....");
		my $msg_in = "FAIL: scenario_user_quota_normal_chown   Line: ".__LINE__."   After deleting files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_normal(
						        msg_in                   => $msg_in,
							Fc                       => $Fc,
							disk                     => $disk,
							files                    => $files,
							hard                     => $hard,
							soft                     => $soft,
							check_upper_bound_only   => 1,
						 );
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_user_quota_normal_chown  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
		logcomment("scenario_user_quota_normal_chown  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_normal ....");
		sleep(5);


#		## @step  Delete chownd files
#		logStep("Delete chownd files");

        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_limit_normal_chown\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub scenario_quota_limit_normal_move(%)
{
        my (%opts) = @_;
	my $Fc = $opts{Fc};
        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_limit_normal_move");

        eval {
	    my (
		 $qr, $qtree, @dir_entries,
	          $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];

#	    my $quota_type = $quota_spec->{type};
	    my $quota_type = $opts{quota_type};
	    my $quota_tree = $opts{quota_tree};
	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};

	    if (!$quota_tree) {
                $msg = "FATAL:  scenario_quota_limit_normal_move  Line: ".__LINE__.
		       "\n\nQuota type must be of 'tree' scope (not 'volume')";
	       	logcomment($msg);
		die($msg);
	    }

	    my $dest_dir_list = [ @Null_dirs ];
	    my $dest_dir = $$dest_dir_list[0];

	    if (!$file_list || (ref($file_list) !~ /array/i) || (scalar(@$file_list) == 0)) {
		$msg = "FATAL:  quota_limit_normal_move  Line: ".__LINE__."\n\n".
		         "Must specify 'file_list'";
	       	logcomment($msg);
		die($msg);
	    }


	       ## @step Move files just created
	       logStep("Move files just created");
	       logcomment("quota_limit_normal_move  Line: ".__LINE__."   Move files for quota ....");
	       foreach my $dir (@$dir_list) {
			$msg = move_testdir_files(proto=>$proto, srcdir=>$dir, dstdir=>$dest_dir);
			if ($msg) {
				$msg = "FATAL: scenario_quota_limit_normal_move  Line: ".
					 __LINE__."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
		        }
		}
	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   .... Finished moving files for quota");
		sleep(5);


		## @step  Verify 'quota report' after deleting
		logStep("Verify 'quota report' after deleting files");
	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: scenario_quota_limit_normal_move  Line: ".
			           __LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   .... Returned from 'get_quota_report'");


	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   Verify 'used/allocated' values from quota report ....");
		if ($qr->{files_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chown  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files used\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{files_used}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chown  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("scenario_quota_normal_delete  Line: ".__LINE__."   .... Finished verifying 'used/allocated' values from quota report");


		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("quota_limit_normal_move  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("quota_limit_normal_move  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after deleting files
		logStep("Verify  'EMS' log after deleting files");
	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   Call 'verify_ems_quota_limit_normal' ....");
		my $msg_in = "FAIL: scenario_quota_limit_normal_move   Line: ".__LINE__."   After deleting files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_normal(
						        msg_in                   => $msg_in,
							Fc                       => $Fc,
							disk                     => $disk,
							files			 => $files,
							hard			 => $hard,
							soft			 => $soft,
						 );
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_limit_normal_move  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
	        logcomment("quota_limit_normal_move  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_normal'");
		sleep(5);
        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_limit_normal_move\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}


sub scenario_quota_disk_limit_normal_truncate(%)
{
        my (%opts) = @_;
	my $Fc = $opts{Fc};
        my $fail_msg_list = [];
        logcomment("Enter scenario_quota_disk_limit_normal_truncate");
        eval {
	    my (
		  $qr, $qtree, @dir_entries,
	          $fail_list, $msg, $fatal_msg, $rslt,
	    );

	    #  The calling routine should perform the next statement sometime before calling this routine
#	    my ($hash_out, $misc_opts, $comb_opts) = initialize_quota_values(%opts);
            my $expected_alloc_kb   = $opts{expected_alloc_kb};
 	    my $h_or_s              = $opts{h_or_s};
 	    my $hard                = $opts{hard};
 	    my $soft                = $opts{soft};
 	    my $d_or_f              = $opts{d_or_f};
 	    my $disk                = $opts{disk};
 	    my $files               = $opts{files};
 	    my $qr_query            = $opts{qr_query};
 	    my $user_list           = $opts{user_list};
 	    my $dir_list            = $opts{dir_list};
 	    my $proto_list          = $opts{proto_list} || undef;
 	    my $group_list          = $opts{group_list} || undef;
 	    my $srcdest_user_list   = $opts{srcdest_user_list} || undef;
 	    my $srcdest_group_list  = $opts{srcdest_group_list} || undef;
 	    my $srcdest_dir_list    = $opts{srcdest_dir_list} || undef;
 	    my $quota_spec          = $opts{quota_spec};
 	    my $file_list           = $opts{file_list};

	    if ($d_or_f ne 'disk') {
	   	$msg = "FATAL:  scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."\n\n".
	                "This scenario only valid for 'disk' quota limits";
	       	logcomment($msg);
		die($msg);
	    }

#	    my $proto           = $$proto_list[0];
#	    my $user            = $$user_list[0];
#	    my $group           = $$group_list[0];
#	    my $dir             = $$group_list[0];
#	    my $srcdest_user    = $$user_list[0];
#	    my $srcdest_group   = $$group_list[0];
#	    my $srcdest_dir     = $$group_list[0];

	   my $quota_type = $quota_spec->{type};

	    if (!$file_list || (ref($file_list) !~ /array/i) || (scalar(@$file_list) == 0)) {
		$msg = "FATAL:  scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."\n\n".
		         "Must specify 'file_list'";
	       	logcomment($msg);
		die($msg);
	    }

	    my $proto = $$proto_list[0];
	    my $client = $Client_for_proto{$proto};


	       ## @step Truncate files just created
	       logStep("Truncate files just created");
               logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Truncate files affected by quota ....");
	       foreach my $dir (@$dir_list) {
	        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Call 'truncate_testdir_files' ....");
			$msg = truncate_testdir_files(proto=>$proto, dir=>$dir, size=>0);
			if ($msg) {
				$msg = "FATAL: scenario_quota_disk_limit_normal_truncate  Line: ".
					 __LINE__."\n$msg";
		       	        logcomment($msg);
				push(@$fail_msg_list, $msg);
				die($msg);
		        }
	        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Returned from 'truncate_testdir_files'");
		}
                logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Finished truncating files affected by quota");
		sleep(5);


		## @step  Verify 'quota report' after truncating files
		logStep("Verify 'quota report' after truncating files");
        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Call 'get_quota_report' ....");
		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: scenario_quota_disk_limit_normal_truncate  Line: ".
			           __LINE__."\n$msg";
			logcomment($msg);
			die($msg);
		}
        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Returned from 'get_quota_report'");


        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Verify 'used/allocated values from quota report ....");
		if ($qr->{disk_used} != 0) {
	        	$msg = "FAIL:  scenario_group_quota_normal_chown  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
		                "Disk used (Kb)\n\tExpected: \"0\"\n".
	                        "\tActual: \"$qr->{disk_used}\"\n";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Finished verifying 'used/allocated values from quota report");


		## @step  Wait a while to collect any EMS messages...
		logStep("Wait a while to collect any EMS messages -- $EMS_MSG_POLL_TIME sec ....");
		logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Sleep EMS_MSG_POLL_TIME ($EMS_MSG_POLL_TIME sec.) ....");
		sleep($EMS_MSG_POLL_TIME);
		logcomment("scenariio_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Finished sleeping");
		logStep(".... Finished waiting for EMS messages, retrieve them");


		## @step  Verify 'EMS' log after truncating files
		logStep("Verify  'EMS' log after truncating files");
        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   Call 'verify_ems_quota_limit_normal' ....");
		my $msg_in = "FAIL: scenario_quota_disk_limit_normal_truncate   Line: ".__LINE__."   After truncating files";
		($fail_list, $msg) =
		    verify_ems_quota_limit_normal(
						        msg_in         => $msg_in,
							Fc             => $Fc,
							disk           => 1,
							hard           => $hard,
							soft           => $soft,
							h_or_s         => $h_or_s,
						 );
		push(@$fail_msg_list, @$fail_list);
		if ($msg) {
	       		$msg = "FATAL: scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__ ."\n\n$msg";
			die($msg);
		}
        	logcomment("scenario_quota_disk_limit_normal_truncate  Line: ".__LINE__."   .... Returned from 'verify_ems_quota_limit_normal'");
		sleep(5);


#		## @step  Delete truncated files
#		logStep("Delete truncated files");
        };

	my $fatal_msg = $@;
        logcomment("Exit scenario_quota_disk_limit_normal_truncate\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}


sub verify_initial_quota_state($%)
{
	my ($Fc, %opts) = @_;
	logcomment("Enter verify_initial_quota_state:\n\n".Dumper(\%opts));
	my $fail_msg_list = [];
	eval {
	        my ($fail_list, $msg, $rslt, $fatal_msg);
		my (@dir_entries, $qr, $qr_query, $num_files);
		my $quota_spec = $opts{quota_spec};
		if (!$quota_spec) {
			$msg = "FATAL:  verify_initial_quota_state  Line: ".__LINE__.
					"\nNo 'quota_spec' option given";
		       	logcomment($msg);
		}

		my $misc_opts = $quota_spec->{misc_opts} || {};
		my $dir_list = $opts{dir_list} || $quota_spec->{dir_list} || $misc_opts->{Dirs} || \@All_dirs;
		my @dir_list = @$dir_list;

		my $h_or_s = $opts{h_or_s} || $quota_spec->{h_or_s} || $misc_opts->{h_or_s};
		my $hard = $opts{hard} || $quota_spec->{hard} || $misc_opts->{hard};
		my $soft = $opts{soft} || $quota_spec->{soft} || $misc_opts->{soft};
	        if ($hard || (defined($h_or_s) && $h_or_s eq 'hard')) {
		    $h_or_s = 'hard';
		    $hard = 1;
		    $soft = 0;
	        } elsif ($soft || (defined($h_or_s) && $h_or_s eq 'soft')) {
		    $h_or_s = 'soft';
		    $hard = 0;
		    $soft = 1;
	        } else {
		     $msg = "FATAL:  verify_initial_quota_state  Line: ".__LINE__."\n\n".
		                 "Must specify one of 'hard', 'soft'";
	       	     logcomment($msg);
		     die($msg);
                }

		my $d_or_f = $opts{d_or_f} || $quota_spec->{d_or_f} || $misc_opts->{d_or_f};
		my $disk = $opts{disk} || $quota_spec->{disk} || $misc_opts->{disk};
		my $files = $opts{files} || $quota_spec->{files} || $misc_opts->{files};
	        if ($disk || (defined($d_or_f) && $d_or_f eq 'disk')) {
		    $d_or_f = 'disk';
		    $disk = 1;
		    $files = 0;
	        } elsif ($files || (defined($d_or_f) && $d_or_f eq 'files')) {
		    $d_or_f = 'files';
		    $disk = 0;
		    $files = 1;
	        } else {
		     $msg = "FATAL:  verify_initial_quota_state  Line: ".__LINE__."\n\n".
		                 "Must specify one of 'disk', 'files'";
	       	     logcomment($msg);
		     die($msg);
                }

		## @step  Verify directories involved in quota report are empty
		logStep("Verify directories involved in quota report are empty");
		foreach my $dir (@dir_list) {
			my $proto;
			if (grep( {$_ eq 'nfs'}  @{$dir->{protos}} ) ) {
				$proto = 'nfs';
				@dir_entries = get_dir_entries( dir=>$NFS_paths{$dir->{name}},
			                                          client=>$nfs_client);
				$num_files = scalar(@dir_entries);
				if ($num_files != 0) {
					$msg = "FATAL:  verify_initial_quota_state  Line: ".__LINE__.
						"\nExpected zero files at start of test,\n".
						"Found $num_files";
				       	logcomment($msg);
					die($msg);
				}
			} else {
				$proto = 'cifs';
				@dir_entries = get_dir_entries( dir=>$CIFS_paths{$dir->{name}},
			                                          client=>$cifs_client);
				$num_files = scalar(@dir_entries);
				if ($num_files != 0) {
					$msg = "FATAL:  verify_initial_quota_state  Line: ".__LINE__.
						"\nExpected zero files at start of test,\n".
						"Found $num_files";
				       	logcomment($msg);
					die($msg);
				}
			}
		}

		my ($disk_slimit, $disk_hlimit, $files_slimit, $files_hlimit);
		$disk_slimit = '-';
		$files_slimit = '-';
		$disk_hlimit = '-';
		$files_hlimit = '-';

		if ($soft) {
			if ($disk) {
				$disk_slimit = $Disk_quota_limit_kb;
			}
			if ($files) {
				$files_slimit = $Files_quota_limit;
			}
		} else {
			if ($disk) {
				$disk_hlimit = $Disk_quota_limit_kb;
			}
			if ($files) {
				$files_hlimit = $Files_quota_limit;
			}
		}


		## @step  Get initial 'quota report' before starting test.
		logStep("Get initial 'quota report' before starting test.");
		$qr_query = cvt_quota_spec_to_query($quota_spec);
       	        logcomment("verify_initial_quota_state  Line: ".__LINE__ .
		              "\nCalling get_quota_report .... \n\n".Dumper($qr_query));

		($msg, $qr)  =  get_quota_report($Fc, %$qr_query);
		if ($msg) {
			$msg = "FATAL: verify_initial_quota_state  Line: ".__LINE__.
			          "\nget_quota_report:\n$msg";
		       	logcomment($msg);
			die($msg);
		}
       	        logcomment("verify_initial_quota_state  Line: ".__LINE__ .
		              "\n.... Returned from get_quota_report\n\n".Dumper($qr));

		$fail_list = [];
	        logcomment("verify_initial_quota_state  Line: ".__LINE__."   Verify 'used/allocated' values from quota report ....");
		if ($qr->{disk_soft_limit} ne $disk_slimit) {
	        	$msg = "FAIL:  verify_initial_quota_state  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Disk Soft Limit:\n\tExpected: \"$disk_slimit\"\n".
	                        "\tActual: \"$qr->{disk_soft_limit}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{files_soft_limit} ne $files_slimit) {
	        	$msg = "FAIL:  verify_initial_quota_state  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files Soft Limit:\n\tExpected: \"$files_slimit\"\n".
	                        "\tActual: \"$qr->{files_soft_limit}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{disk_limit} ne $disk_hlimit) {
	        	$msg = "FAIL:  verify_initial_quota_state  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Disk Hard Limit:\n\tExpected: \"$disk_hlimit\"\n".
	                        "\tActual: \"$qr->{disk_limit}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		if ($qr->{files_limit} ne $files_hlimit) {
	        	$msg = "FAIL:  verify_initial_quota_state  Line: ".__LINE__.
			        "\nQuota report returned unexpected values:\n".
			        "Files Hard Limit:\n\tExpected: \"$files_hlimit\"\n".
	                        "\tActual: \"$qr->{files_limit}\"";
		       	logcomment($msg);
			push(@$fail_msg_list, $msg);
		}
		logcomment("verify_initial_quota_state  Line: ".__LINE__."   .... Finished verifying 'used/allocated' values from quota report");
        };

        my $fatal_msg = $@;
        logcomment("Exit verify_initial_quota_state:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}


sub verify_testdirs_empty($%)
{
	my ($Fc, %opts) = @_;
	my $fail_msg_list = [];
	eval {
	        my ($msg, $rslt, $fatal_msg, $fail_list);
		my $num_files;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		my @dir_list;
		my $proto;
		my @dir_entries;
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}
	        my $comb_opts = combine_hashtables(\%opts, $quota_spec, $misc_opts);
		if ($comb_opts->{dir}) {
			@dir_list = ( $comb_opts->{dir} );
		} elsif ( $comb_opts->{dir_list} ) {
			@dir_list = ( @$comb_opts->{dir_list} );
		} elsif ( $comb_opts->{Dirs} ) {
			@dir_list = ( @$comb_opts->{Dirs} );
		} else {
			$msg = "FATAL:  verify_testdirs_empty  Line: ".__LINE__.
				   "No 'dir' 'Dirs' or 'dir_list' option given";
		       	logcomment($msg);
		}
		foreach my $dir (@dir_list) {
			if (grep( {$_ eq 'nfs'}  @{$dir->{protos}} ) ) {
				$proto = 'nfs';
				@dir_entries = get_dir_entries( dir=>$NFS_paths{$dir->{name}},
			                                          client=>$nfs_client);
				$num_files = scalar(@dir_entries);
				if ($num_files != 0) {
					$msg = "FATAL:  verify_testdirs_empty  Line: ".__LINE__.
						"\nExpected zero files at start of test,\n".
						"Found $num_files";
				       	logcomment($msg);
					die($msg);
				}
			} else {
				$proto = 'cifs';
				@dir_entries = get_dir_entries( dir=>$CIFS_paths{$dir->{name}},
			                                          client=>$cifs_client);
				$num_files = scalar(@dir_entries);
				if ($num_files != 0) {
					$msg = "FATAL:  verify_testdirs_empty  Line: ".__LINE__.
						"\nExpected zero files at start of test,\n".
						"Found $num_files";
				       	logcomment($msg);
					die($msg);
				}
			}
		}
	};

        my $fatal_msg = $@;
        logcomment("Exit verify_testdirs_empty:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	return($fail_msg_list, $fatal_msg);
}



sub truncate_testdir_files(%)
{
	my (%opts) = @_;
	my $fail_msg_list = [];
	eval {
	        my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}
	        my $comb_opts = combine_hashtables(\%opts, $quota_spec, $misc_opts);

		my $file_list = $comb_opts->{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($comb_opts->{dir}) {
			@dir_list = ( $comb_opts->{dir} );
		} elsif ( $comb_opts->{dir_list} ) {
			@dir_list = ( @$comb_opts->{dir_list} );
		} elsif ( $comb_opts->{Dirs} ) {
			@dir_list = ( @$comb_opts->{Dirs} );
		} else {
			$msg = "FATAL:  truncate_testdir_files  Line: ".__LINE__.
				   "No 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($comb_opts->{proto}) {
			@proto_list = ( $comb_opts->{proto} );
		} elsif ($comb_opts->{proto_list}) {
			@proto_list = @$comb_opts->{proto_list};
		} elsif ($comb_opts->{Protos}) {
			@proto_list = @$comb_opts->{Protos};
		}

		my $size;
		if (defined($comb_opts->{size})) {
			$size = $comb_opts->{size};
		}
		$size = 0;

		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my $dirname;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};
			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dirobj->{name}};
				$dirpath = $dirobj->{nfs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = setattr_file( file        => $srcpath,
					                      size       => $size,
					                      client     => $client,
						              user       => 'root',
					        );
					if (!$rslt || $@) {
						$msg = "truncate_testdir_files Line: ".__LINE__.
						       "\nCould not truncate file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				}
			} else {
				$proto = 'cifs';
				$client = $cifs_client;
#				$dirpath = $CIFS_paths{$dir->{name}};
				$dirpath = $dirobj->{cifs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = setattr_file( file        => $srcpath,
					                      size       => $size,
					                      client     => $client,
						              # user       => 'administrator',
					        );
					if (!$rslt || $@) {
						$msg = "truncate_testdir_files Line: ".__LINE__."\nCould not truncate file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				}
			}
		}
	};

       my $fatal_msg = $@;
#        logcomment("Exit truncate_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
       logcomment("Exit truncate_testdir_files\nErrMsg: \"$fatal_msg\"");
       return($fatal_msg);
}


sub delete_testdir_files(%)
{
	my (%opts) = @_;
	my $fail_msg_list = [];
	eval {
		my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}
	        my $comb_opts = combine_hashtables(\%opts, $quota_spec, $misc_opts);

		my $file_list = $comb_opts->{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($comb_opts->{dir}) {
			@dir_list = ( $comb_opts->{dir} );
		} elsif ( $comb_opts->{dir_list} ) {
			@dir_list = ( @$comb_opts->{dir_list} );
		} elsif ( $comb_opts->{Dirs} ) {
			@dir_list = ( @$comb_opts->{Dirs} );
		} else {
			$msg = "FATAL:  delete_testdir_files  Line: ".__LINE__.
				   "No 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($comb_opts->{proto}) {
			@proto_list = ( $comb_opts->{proto} );
		} elsif ($comb_opts->{proto_list}) {
			@proto_list = @$comb_opts->{proto_list};
		} elsif ($comb_opts->{Protos}) {
			@proto_list = @$comb_opts->{Protos};
		}

		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my $dirname;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};
			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dir->{name}};
				$dirpath = $dirobj->{nfs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = remove_file( file       => $srcpath,
					                     client     => $client,
						             user       => 'root',
					        );
					if (!$rslt || $@) {
						$msg = "delete_testdir_files Line: ".__LINE__."\nCould not delete file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
				        if (test_file(file=>$srcpath, test=>'f', client=>$client)) {
						$msg = "delete_testdir_files Line: ".__LINE__."\nCould not delete file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					} else {
						if ($path_list_out) {
							if (ref($path_list_out) =~ /array/i) {
								push(@$path_list_out, $srcpath);
							} elsif (ref($path_list_out) =~ /hash/i) {
								$path_list_out->{$srcpath} = {};
							}
						}
					}
				}
			} else {
				$proto = 'cifs';
				$client = $cifs_client;
#				$dirpath = $CIFS_paths{$dir->{name}};
				$dirpath = $dirobj->{cifs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = remove_file( file       => $srcpath,
					                     client     => $client,
						             # user       => 'root',
					        );
					if (!$rslt || $@) {
						$msg = "delete_testdir_files Line: ".__LINE__."\nCould not delete file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
				        if (test_file(file=>$srcpath, test=>'f', client=>$client)) {
						$msg = "delete_testdir_files Line: ".__LINE__."\nCould not delete file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					} else {
						if ($path_list_out) {
							if (ref($path_list_out) =~ /array/i) {
								push(@$path_list_out, $srcpath);
							} elsif (ref($path_list_out) =~ /hash/i) {
								$path_list_out->{$srcpath} = {};
							}
						}
					}
				}
			}
		}
	};

       my $fatal_msg = $@;
#        logcomment("Exit delete_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
       logcomment("Exit delete_testdir_files\nErrMsg: \"$fatal_msg\"");
       return($fatal_msg);
}



#############################################################
sub ls_unix_file(%) {
 my (%opts) = @_;
 my $srcpath = $opts{file};
 return(`ls -l $srcpath`);
}
#############################################################


sub show_unix_testdir_files(%)
{
	my (%opts) = @_;
	my @text_lines = ();
	logDbg("show_unix_testdir_files\n\n".Dumper(\%opts));
	my $fail_msg_list = [];
	eval {
		my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		my $msg_in = $opts{msg_in};
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}
	        my $comb_opts = combine_hashtables(\%opts, $quota_spec, $misc_opts);

		my $file_list = $comb_opts->{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($comb_opts->{dir}) {
			@dir_list = ( $comb_opts->{dir} );
		} elsif ( $comb_opts->{dir_list} ) {
			@dir_list = ( @$comb_opts->{dir_list} );
		} elsif ( $comb_opts->{Dirs} ) {
			@dir_list = ( @$comb_opts->{Dirs} );
		} else {
			$msg = "FATAL:  show_unix_testdir_files  Line: ".__LINE__.
				   "\nNo 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($comb_opts->{proto}) {
			@proto_list = ( $comb_opts->{proto} );
		} elsif ($comb_opts->{proto_list}) {
			@proto_list = @$comb_opts->{proto_list};
		} elsif ($comb_opts->{Protos}) {
			@proto_list = @$comb_opts->{Protos};
		}

		if ($msg_in) {
			push(@text_lines, $msg_in);
			push(@text_lines, "");
		}
		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my $dirname;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};
			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dir->{name}};
				$dirpath = $dirobj->{nfs_path};
				push(@text_lines, "\nDir: $dirpath\n");
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = ls_unix_file(  file       => $srcpath,
					                       client     => $client,
						               user       => 'root',
					        );
					if (!$rslt || $@) {
						$msg = "show_unix_testdir_files Line: ".__LINE__."\nCould not 'ls' file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					} else {
						if ($path_list_out) {
							if (ref($path_list_out) =~ /array/i) {
								push(@$path_list_out, $srcpath);
							} elsif (ref($path_list_out) =~ /hash/i) {
								$path_list_out->{$srcpath} = {};
							}
						}
						push(@text_lines, $rslt);
					}
				}
			}
		}
	};

       my $fatal_msg = $@;
       if ($fatal_msg) {
	       logcomment("Exit show_unix_testdir_files\nErrMsg: \"$fatal_msg\"");
       	       return($fatal_msg);
	} else {
	       logcomment("Exit show_unix_testdir_files:\n\n".join("\n", @text_lines));
       	       return("");
	}
}



sub quick_delete_testdir_files(%)
{
	logcomment("Enter quick_delete_testdir_files");
	my $fail_msg_list = [];
	my (%opts) = @_;
	eval {
		my ($msg, $dirpath, $client, $dirobj, $dirname, $rslt);
		my ($proto) = $opts{proto};
		if (!$proto) {
			$msg = "quick_delete_testdir_files  Line: ".__LINE__.
			           "No 'proto' option specified";
			logcomment($msg);
			die($msg);
		}
		my $dir = $opts{dir};
		if (!$dir) {
			$msg = "quick_delete_testdir_files  Line: ".__LINE__.
			           "No 'dir' option specified";
			logcomment($msg);
			die($msg);
		}
		if (ref($dir)) {
			$dirobj = $dir;
			$dirname = $dirobj->{name};
		} else {
			$dirname = $dir;
			$dirobj = $Dir_hash{$dirname};
		}

		if ($proto eq 'nfs') {
			$dirpath = $dirobj->{nfs_path};
			$client = $nfs_client;
		} elsif ($proto eq 'cifs') {
			$dirpath = $dirobj->{cifs_path};
			$client = $cifs_client;
		} else {
			$msg = "quick_delete_testdir_files  Line: ".__LINE__.
			           "Invalid 'proto' given: \"$proto\"";
			logcomment($msg);
			die($msg);
		}

		logcomment("quick_delete_testdir_files  Line: ".__LINE__."    Call 'remove_dir' ....");
		if ($proto eq 'nfs') {
			$rslt = remove_dir(  dir               => $dirpath,
			                     client            => $client,
			                     recursive         => 1,
			                     contents_only     => 1,
			                     user              => 'root',
			        );
		} else {
			$rslt = remove_dir(  dir               => $dirpath,
			                     client            => $client,
			                     recursive         => 1,
			                     contents_only     => 1,
			                     # user              => 'administrator',
			        );
		}
		if (!$rslt || $@) {
			$msg = "quick_delete_testdir_files  Line: ".__LINE__."\nremove_dir \"$dirpath\"  failed: $@";
			logcomment($msg);
			die($msg);
		}
		logcomment("quick_delete_testdir_files  Line: ".__LINE__."    .... Returned from 'remove_dir'");


	        if (!test_file(file=>$dirpath, test=>'d', client=>$client)) {
			$msg = "quick_delete_testdir_files  Line: ".__LINE__."\nContaining directory \"$dirname\" was deleted";
			logcomment($msg);
#			die($msg);
			logcomment("quick_delete_testdir_files  Line: ".__LINE__."    Call 'make_dir' ....");
			$rslt = make_dir(  dir              => $dirpath,
			                   client           => $client,
				           okay_if_exists   => 1,
			         );
			if (!$rslt || $@) {
		 	 	$msg = "quick_delete_testdir_files  Line: ".__LINE__.
				           "\nmake_dir $dirpath: $@";
				die($msg);
			}
			logcomment("quick_delete_testdir_files  Line: ".__LINE__."    .... Returned from 'make_dir'");

		} else {

			logcomment("quick_delete_testdir_files  Line: ".__LINE__."    Call 'get_testdir_files' ....");
			my ($err, $rem_files, undef) =
			     get_testdir_files(proto=>$proto, dir=>$dirname);
			if ($err) {
				$msg = "quick_delete_testdir_files  Line: ".__LINE__."\nError from 'get_testdir_files:\n$err";
				logcomment($msg);
				die($msg);
			}
			logcomment("quick_delete_testdir_files  Line: ".__LINE__."    .... Returned from 'get_testdir_files'");
			if (@$rem_files) {
				$msg = "quick_delete_testdir_files  Line: ".__LINE__."\nFailed to delete all files";
				logcomment($msg);
				die($msg);
			}
		}
    };

    my $fatal_msg = $@;
#        logcomment("Exit quick_delete_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
    logcomment("Exit quick_delete_testdir_files\nErrMsg: \"$fatal_msg\"");
    return($fatal_msg);
}



sub chown_testdir_files(%)
{
	my (%opts) = @_;
	logcomment("Enter chown_testdir_files\n\n".Dumper(\%opts));
	my $fail_msg_list = [];
	eval {
		my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}

		my $file_list = $opts{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($opts{dir}) {
			@dir_list = ( $opts{dir} );
		} elsif ( $opts{dir_list} ) {
			@dir_list = ( @{$opts{dir_list}} );
		} elsif ( $opts{Dirs} ) {
			@dir_list = ( @{$opts{Dirs}} );
		} else {
			$msg = "FATAL:  chown_testdir_files  Line: ".__LINE__.
				   "\nNo 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($opts{proto}) {
			@proto_list = ( $opts{proto} );
		} elsif ($opts{proto_list}) {
			@proto_list = @{$opts{proto_list}};
		} elsif ($opts{Protos}) {
			@proto_list = @{$opts{Protos}};
		}

		my @dest_user_list;
		if ($opts{dest_user_list}) {
		    @dest_user_list = @{$opts{dest_user_list}};
		} elsif ($opts{dest_users}) {
		    @dest_user_list = @{$opts{dest_users}};
		} elsif ($opts{dest_user}) {
		    my $dest_user = $opts{dest_user};
		    @dest_user_list = ( $dest_user );
		} else {
		    @dest_user_list = @Null_users;
		}

		my @dest_user_login_list = ();
		my @dest_user_obj_list = ();
		foreach my $user ( @dest_user_list ) {
		    my ($user_login, $user_obj, $user_uid);
		    if (ref($user)) {
		    	$user_obj = $user;
			$user_login = $user->{login};
		    } elsif ($user =~ /^\d+/) {
			$user_uid = $user;
			$user_obj = $User_for_uid{$user_uid};
		    } else {
			$user_login = $user;
			$user_obj = $User_hash{$user_login};
		    }
		    if (!$user_obj) {
			$msg = "FATAL:  chown_testdir_files  Line: ".__LINE__.
			         "\nCould not find user \"$user_login\"";
			logcomment($msg);
			die($msg);
		    }
		    push(@dest_user_obj_list, $user_obj);
		    push(@dest_user_login_list, $user_login);
	       }
	       my $dest_user_login = $dest_user_login_list[0];
	       my $dest_user_obj = $dest_user_obj_list[0];

	       ## @step Chown files just created
		my $dest_user_uid = $dest_user_obj->{uid};
#		my $dest_user_win_login = $dest_user_obj->{win_login};
		my $dest_user_win_login = $dest_user_obj->{full_win_login};
logcomment("chown_testdir_files  Line: ".__LINE__."\n\nOPTS:\n\n".Dumper(\%opts));
logcomment("chown_testdir_files  Line: ".__LINE__."\n\nDEST_USER: \"$dest_user_login\"");

		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};

			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dir->{name}};
				$dirpath = $dirobj->{nfs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = setattr_file( file        => $srcpath,
					                       client     => $client,
						               user       => 'root',
							       uid        => $dest_user_uid,
					            );
					if (!$rslt || $@) {
						$msg = "chown_testdir_files: Could not chown  file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					my %attrs = stat_file( file       => $srcpath,
					                       client     => $client,
						               rtn_hash   => 1,
				                               user       => 'root'
							);
					if (!%attrs || $@) {
						$msg = "chown_testdir_files: Could not 'stat' file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					my $uid = $attrs{uid};
					logcomment("chown_testdir_files  Line: ".__LINE__.
					               "\nExpected new UID: \"$dest_user_uid\"\n".
						       "Actual new UID: \"$uid\"\n".
					               "\n\nAttrs after 'chown'\n\n".Dumper($srcpath));
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				}
			} else {
				$proto = 'cifs';
				$client = $cifs_client;
#				$dirpath = $CIFS_paths{$dir->{name}};
				$dirpath = $dirobj->{cifs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
logcomment("chown_testdir_files  Line: ".__LINE__."    Call 'cifs_chown_file'\n\nPATH: \"$srcpath\"\nCLIENT: \"$client\"\nUSER: \"$dest_user_win_login\"");
### Start DBG
logcomment("chown_testdir_files  Line: ".__LINE__);
### End DBG
					$rslt = cifs_chown_file( path        => $srcpath,
					                         client      => $client,
						                 user        => $dest_user_win_login,
					             );
logcomment("chown_testdir_files  Line: ".__LINE__);
					if (!$rslt || $@) {
						$msg = "chown_testdir_files: Could not chown file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
logcomment("chown_testdir_files  Line: ".__LINE__);
					my %attrs = stat_file(  file       => $srcpath,
					                        client     => $client,
						                rtn_hash   => 1,
				                                # user     => 'root',
						      );
logcomment("chown_testdir_files  Line: ".__LINE__);
					if (!%attrs || $@) {
						$msg = "chown_testdir_files: Could not 'stat' file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
logcomment("chown_testdir_files  Line: ".__LINE__."\n\nAttrs:\n\n".Dumper(\%attrs));
					my $new_owner = cifs_file_getowner(path=>$srcpath, client=>$client);
					if ($@) {
						$msg = "chown_testdir_files: Could not 'stat' file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					logcomment("chown_testdir_files  Line: ".__LINE__.
					               "\nExpected new owner: \"$dest_user_win_login\" -- \"$dest_user_win_login\"\n".
					               "Actual new owner: \"$new_owner\n".
					               "\n\nAttrs after 'chown'\n\n".Dumper($srcpath));
					if (uc($new_owner) eq uc($dest_user_win_login)) {
						logcomment("chown_testdir_files  Line: ".__LINE__.
						             "\nFile: \"$srcpath\" owner okay: \"$new_owner\"");
					} else {
						$msg = "chown_testdir_files  Line: ".__LINE__.
						             "\nCould not change owner for File: \"$srcpath\"\n".
							     "Expected owner \"$dest_user_win_login\"  Actual owner: \"$new_owner\"";
						logcomment($msg);
						die($msg);
					}

### Start DBG
logcomment("chown_testdir_files  Line: ".__LINE__);
%attrs = stat_file(file=>$srcpath, client=>$client, rtn_hash=>1);
logcomment("chown_testdir_files  Line: ".__LINE__."\n\nAttrs for \"$srcpath\"\n\n".Dumper(\%attrs));
### End DBG

					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				}
			}
		}
	};

       my $fatal_msg = $@;
#        logcomment("Exit chown_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
       logcomment("Exit chown_testdir_files\nErrMsg: \"$fatal_msg\"");
       return($fatal_msg);
}



sub chgrp_testdir_files(%)
{
	my (%opts) = @_;
	logcomment("Enter chgrp_testdir_files\n\n".Dumper(\%opts));
	my $fail_msg_list = [];
	eval {
		my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		if ($quota_spec) {
			$misc_opts = $quota_spec->{misc_opts};
		}

		my $file_list = $opts{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($opts{dir}) {
			@dir_list = ( $opts{dir} );
		} elsif ( $opts{dir_list} ) {
			@dir_list = ( @{$opts{dir_list}} );
		} elsif ( $opts{Dirs} ) {
			@dir_list = ( @{$opts{Dirs}} );
		} else {
			$msg = "FATAL:  chgrp_testdir_files  Line: ".__LINE__.
				   "\nNo 'dir' 'Dirs' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($opts{proto}) {
			@proto_list = ( $opts{proto} );
		} elsif ($opts{proto_list}) {
			@proto_list = @{$opts{proto_list}};
		} elsif ($opts{Protos}) {
			@proto_list = @{$opts{Protos}};
		}

		my @dest_group_list;
		if ($opts{dest_group_list}) {
		    @dest_group_list = @{$opts{dest_group_list}};
		} elsif ($opts{dest_groups}) {
		    @dest_group_list = @{$opts{dest_groups}};
		} elsif ($opts{dest_group}) {
		    my $dest_group = $opts{dest_group};
		    @dest_group_list = ( $dest_group );
		} else {
		    @dest_group_list = @Null_groups;
		}

	        my $dest_group_gid = undef;
		my @dest_group_gid_list = ();
		foreach my $group ( @dest_group_list ) {
		    my ($group_gid, $group_obj, $group_name);
		    if (ref($group)) {
		    	$group_obj = $group;
			$group_gid = $group_obj->{gid};
		    } elsif ($group =~ /^\d+/) {
			$group_gid = $group;
			$group_obj = $Group_for_gid{$group_gid};
		    } else {
			$group_name = $group;
			$group_obj = $Group_hash{$group_name};
			$group_gid = $group_obj->{gid};
		    }
		    if (!$group_obj) {
			$msg = "FATAL:  chgrp_testdir_files  Line: ".__LINE__.
			         "\nCould not find group \"$group\"";
			logcomment($msg);
			die($msg);
		    }
		    push(@dest_group_gid_list, $group_gid);
	       }
	       if (scalar(@dest_group_gid_list)) {
	           $dest_group_gid = $dest_group_gid_list[0];
	       }
	       $dest_group_gid = $dest_group_gid_list[0];


		my @dest_user_list;
		if ($opts{dest_user_list}) {
		    @dest_user_list = @{$opts{dest_user_list}};
		} elsif ($opts{dest_users}) {
		    @dest_user_list = @{$opts{dest_users}};
		} elsif ($opts{dest_user}) {
		    my $dest_user = $opts{dest_user};
		    @dest_user_list = ( $dest_user );
		} else {
		    @dest_user_list = @Null_users;
		}

	        my $dest_user_gid = undef;
		my @dest_user_gid_list = ();
		foreach my $user ( @dest_user_list ) {
		    my ($user_gid, $user_obj, $user_login, $user_uid);
		    if (ref($user)) {
		    	$user_obj = $user;
			$user_gid = $user_obj->{gid};
		    } elsif ($user =~ /^\d+/) {
			$user_uid = $user;
			$user_obj = $User_for_uid{$user_uid};
			$user_gid = $user_obj->{gid};
		    } else {
			$user_login = $user;
			$user_obj = $User_hash{$user_login};
			$user_gid = $user_obj->{gid};
		    }
		    if (!$user_obj) {
			$msg = "FATAL:  chgrp_testdir_files  Line: ".__LINE__.
			         "\nCould not find user \"$user\"";
			logcomment($msg);
			die($msg);
		    }
		    push(@dest_user_gid_list, $user_gid);
	       }
	       if (scalar(@dest_user_gid_list)) {
	           $dest_user_gid = $dest_user_gid_list[0];
	       }

	       if (!@dest_group_gid_list && !@dest_user_gid_list) {
	            $msg = "FATAL:  chgrp_testdir_files  Line: ".__LINE__.
		             "\n No 'dest_group' or 'dest_user' specified";
		    logcomment($msg);
		    die($msg);
	       }
	       my $dest_gid = (defined($dest_group_gid) ? $dest_group_gid : $dest_user_gid);

	       ## @step Chgrp files just created
logcomment("chgrp_testdir_files  Line: ".__LINE__."\n\nOPTS:\n\n".Dumper(\%opts));
logcomment("chgrp_testdir_files  Line: ".__LINE__."\n\nDEST_GID: \"$dest_gid\"");

		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};

			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dir->{name}};
				$dirpath = $dirobj->{nfs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					$rslt = setattr_file( file        => $srcpath,
					                       client     => $client,
						               user       => 'root',
							       gid        => $dest_gid,
					            );
					if (!$rslt || $@) {
						$msg = "chgrp_testdir_files: Could not chgrp  file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					my %attrs = stat_file( file       => $srcpath,
					                       client     => $client,
						               rtn_hash   => 1,
				                               user       => 'root'
							);
					if (!%attrs || $@) {
						$msg = "chgrp_testdir_files: Could not 'stat' file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					my $uid = $attrs{uid};
					logcomment("chgrp_testdir_files  Line: ".__LINE__.
					               "\nExpected new GID: \"$dest_gid\"\n".
						       "Actual new UID: \"$uid\"\n".
					               "\n\nAttrs after 'chgrp'\n\n".Dumper($srcpath));
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				}
			} else {
				$msg = "chgrp_testdir_files  Line: ".__LINE__.
					"\n'CIFS' protocol does not support 'group' quotas";
				logcomment($msg);
				die($msg);
			}
		}
	};

       my $fatal_msg = $@;
#        logcomment("Exit chgrp_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
       logcomment("Exit chgrp_testdir_files\nErrMsg: \"$fatal_msg\"");
       return($fatal_msg);
}



sub move_testdir_files(%)
{
	my (%opts) = @_;
	my $fail_msg_list = [];
	eval {
		my ($msg, $rslt, $fatal_msg, $fail_list);
		my $path_list_in;
		my $path_list_out;
		my $quota_spec = $opts{quota_spec};
		my $misc_opts;
		if ($quota_spec) {
			my $misc_opts = $quota_spec->{misc_opts};
		}

		my $file_list = $opts{file_list};
		if (!$file_list) {
		} elsif (ref($file_list) =~ /hash/i) {
			my @keys = keys(%$file_list);
			if (@keys) {
				$path_list_in = \@keys;
			} else {
				$path_list_out = $file_list;
			}
		} elsif (ref($file_list) =~ /array/i) {
			if (@$file_list) {
				$path_list_in = $file_list;
			} else {
				$path_list_out = $file_list;
			}
		}

		my @dir_list;
		if ($opts{dir}) {
			@dir_list = ( $opts{dir} );
		} elsif ( $opts{dir_list} ) {
			@dir_list = @{$opts{dir_list}};
		} elsif ( $opts{Dirs} ) {
			@dir_list = @{$opts{Dirs}};
		} else {
			$msg = "FATAL:  move_testdir_files  Line: ".__LINE__.
				   "No 'dir' or 'dir_list' option given";
			logcomment($msg);
			die($msg);
		}
		my @proto_list;
		if ($opts{proto}) {
			@proto_list = ( $opts{proto} );
		} elsif ($opts{proto_list}) {
			@proto_list = @{$opts{proto_list}};
		} elsif ($opts{Protos}) {
			@proto_list = @{$opts{Protos}};
		}

		my @dest_dir_list;
		if ($opts{dest_dir_list}) {
			@dest_dir_list = @{$opts{dest_dir_list}};
		} elsif ($opts{dest_dirs}) {
			@dest_dir_list = @{$opts{dest_dirs}};
		} elsif ($opts{dest_dir}) {
			my $dest_dir = $opts{dest_dir};
			@dest_dir_list = ( $dest_dir );
		} else {
			@dest_dir_list = @Null_dirs;
		}

		my @dest_dir_name_list = ();
		my @dest_dir_obj_list = ();
		foreach my $dir ( @dest_dir_list) {
		    my ($dir_obj, $dir_name);
		    if (ref($dir)) {
		    	$dir_obj = $dir;
			$dir_name = $dir->{name};
		    } else {
			$dir_name = $dir;
			$dir_obj = $Dir_hash{$dir_name};
		    }
		    if (!$dir_obj) {
			$msg = "FATAL:  move_testdir_files  Line: ".__LINE__.
			         "\nCould not find dir \"$dir_name\"";
			logcomment($msg);
			die($msg);
		    }
		    push(@dest_dir_obj_list, $dir_obj);
		    push(@dest_dir_name_list, $dir_name);
		}
		my $dest_dir_name = $dest_dir_name_list[0];
		my $dest_dir_obj = $dest_dir_obj_list[0];

		foreach my $dirname (@dir_list) {
			my $dirobj = $Dir_hash{$dirname};
			my $proto;
			my $client;
			my $dirpath;
			my @dir_entries;
			my $dir_path_list;
			my @protos_for_dir = @{$dirobj->{protos}};
			if (grep( {$_ eq 'nfs'}  @protos_for_dir ) ) {
				$proto = 'nfs';
				$client = $nfs_client;
#				$dirpath = $NFS_paths{$dir->{name}};
				$dirpath = $dirobj->{nfs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					my $dest_path = $srcpath;
					if ($dest_path =~ m%\/$dirname\/%) {
						$msg = "move_testdir_files:Invalid path: \"$srcpath\"\n".
							    "for directory: \"$dirname\"";
						logcomment($msg);
						die($msg);
					}
					$dest_path =~ s/$dirname/$dest_dir_name/;
					$rslt = rename_file(  file        => $srcpath,
					                       client     => $client,
						               user       => 'root',
						               nfile      => $dest_path,
					        );
					if (!$rslt || $@) {
						$msg = "move_testdir_files: Could not move file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}

				        if (test_file(file=>$srcpath, test=>'f', client=>$client)) {
						$msg = "move_testdir_files  Line: ".__LINE__.
							  "\nFailed to move file from ".
							  "\"$srcpath\" to ".
						          "\"$dest_path\"";
						logcomment($msg);
						die($msg);
					}
				        if (!test_file(file=>$dest_path, test=>'f', client=>$client)) {
						$msg = "move_testdir_files  Line: ".__LINE__.
							  "\nFailed to move file from ".
							  "\"$srcpath\" to ".
						          "\"$dest_path\"";
						logcomment($msg);
						die($msg);
					}
				}
			} else {
				$proto = 'cifs';
				$client = $cifs_client;
#				$dirpath = $CIFS_paths{$dir->{name}};
				$dirpath = $dirobj->{cifs_path};
				@dir_entries = get_dir_entries( dir=>$dirpath,
			                                          client=>$client );
				$dir_path_list = [ map ( { $dirpath . '/' . $_ }   @dir_entries ) ];
				foreach my $srcpath ( @$dir_path_list ) {
					if ($path_list_in) {
						next unless ( grep( {$_ eq $srcpath}  @$path_list_in) );
					}
					my $dest_path = $srcpath;
					if ($dest_path =~ m%\/$dirname\/%) {
						$msg = "move_testdir_files:Invalid path: \"$srcpath\"\n".
							    "for directory: \"$dirname\"";
						logcomment($msg);
						die($msg);
					}
					$dest_path =~ s/$dirname/$dest_dir_name/;
					$rslt = rename_file(  file        => $srcpath,
					                       client     => $client,
						               # user       => 'administrator',
						               nfile      => $dest_path,
					        );
					if (!$rslt || $@) {
						$msg = "move_testdir_files: Could not move file: \"$srcpath\"";
						logcomment($msg);
						die($msg);
					}
					if ($path_list_out) {
						if (ref($path_list_out) =~ /array/i) {
							push(@$path_list_out, $srcpath);
						} elsif (ref($path_list_out) =~ /hash/i) {
							$path_list_out->{$srcpath} = {};
						}
					}
				        if (test_file(file=>$srcpath, test=>'f', client=>$client)) {
						$msg = "move_testdir_files  Line: ".__LINE__.
							  "\nFailed to move file from ".
							  "\"$srcpath\" to ".
						          "\"$dest_path\"";
						logcomment($msg);
						die($msg);
					}
				        if (!test_file(file=>$dest_path, test=>'f', client=>$client)) {
						$msg = "move_testdir_files  Line: ".__LINE__.
							  "\nFailed to move file from ".
							  "\"$srcpath\" to ".
						          "\"$dest_path\"";
						logcomment($msg);
						die($msg);
					}
				}
			}
		}
	};

       my $fatal_msg = $@;
#        logcomment("Exit move_testdir_files:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
       logcomment("Exit move_testdir_files\nErrMsg: \"$fatal_msg\"");
       return($fatal_msg);
}


sub combine_hashtables(@)
{
	my (@hashref_list) = @_;
	my $hash_out = {};
	foreach my $h ( @hashref_list ) {
		next unless (ref($h) =~ /hash/i);
		while (my ($k,$v) = each(%$h) ) {
			if (!defined($hash_out->{$k})) {
				$hash_out->{$k} = $v;
			}
		}
	}
	return( $hash_out );
}


sub list_difference($$)
{
	my ($list1, $list2) = @_;
	my @list3 = ();
	foreach my $elt (@$list1) {
		next if (grep({$elt eq $_} @$list2));
		push(@list3, $elt);
	}
	return(@list3);
}


sub get_rel_diff_percent($$)
{
	my ($test_val, $ref_val) = @_;
        logcomment("Enter get_rel_diff_percent");
	if ( !defined($test_val) || !defined($ref_val) || ($ref_val == 0.0) ) {
		return(undef);
	}
	my $abs_diff = abs($test_val - $ref_val);
	my $sign = ($test_val >= $ref_val) ? 1 : -1;
	my $rel_diff = ($abs_diff / $ref_val);
	my $rel_diff_pct = 100 * $rel_diff;
        logcomment("Exit get_rel_diff_percent");
	if (wantarray) {
		return ($rel_diff_pct, $sign);
	} else {
		return ($rel_diff_pct);
	}
}


sub initialize_file_values(%)
{
	my (%orig_opts) = @_;
        my %opts = (%orig_opts);
	logcomment("Enter initialize_file_values\n\n".Dumper(\%orig_opts));

	eval {
	    my ($msg, $rslt, $file_cnt);

	     my ($file_prefix, $num_files,  $size_bytes, $size_kb, $expected_alloc_kb);
	     my $file_set = $opts{file_set} || $File_set || $Default_file_set;
	     if (!$file_set) {
	           $msg = "FATAL: initialize_file_values  Line: ".__LINE__."\n".
		            "No 'file_set' option given, and 'File_set' and 'Default_file_set' are undefined";
		logcomment($msg);
		die($msg);
	     }

	     if ($file_set) {
	           logDbg("initialize_file_values  Line: ".__LINE__."\n'file_set' is specified");
		   $File_set = $file_set;
	     } else {
	           logDbg("initialize_file_values  Line: ".__LINE__."\nNo 'file_set' is specified");
	     }

	     if ($opts{num_files}) {
		$num_files = $opts{num_files};
	     } elsif ($file_set && $file_set->{num_files}) {
		$num_files = $file_set->{num_files};
	     } else {
	        $msg = "FATAL: initialize_file_values  Line: ".__LINE__.
		        "\nNo 'num_file' option given";
		logcomment($msg);
		die($msg);
	     }

	     if ($opts{size_bytes}) {
		$size_bytes = $opts{size_bytes};
		$size_kb = int($size_bytes/1024);
	     } elsif ($opts{size}) {
		my $size = $opts{size};
		$size_bytes = convert_to_bytes($size);
		$size_kb = int($size_bytes/1024);
	     } elsif ($opts{size_kb}) {
		$size_kb = $opts{size_kb};
		$size_bytes = 1024 * $size_kb;
	     } elsif ($file_set && $file_set->{size_kb}) {
	        $size_kb = $file_set->{size_kb};
		$size_bytes = 1024 * $size_kb;
	     } else {
	        $msg = "FATAL: initialize_file_values  Line: ".__LINE__.
		        "\nNo 'size', 'size_bytes', or 'size_kb' options given";
		logcomment($msg);
		die($msg);
	     }
	     if ($size_bytes < $MIN_FILE_SIZE) {
		$size_bytes = $MIN_FILE_SIZE;
		$size_kb = int($size_bytes/1024);
	     }

	     if ($opts{file_prefix}) {
		$file_prefix = $opts{file_prefix};
	     } elsif ($file_set && $file_set->{file_prefix}) {
		$file_prefix = $file_set->{file_prefix};
	     } else {
		$file_prefix = "";
	     }

	     # $expected_alloc_kb   = $size_kb * $num_files;
	     if ($opts{expected_alloc_kb}) {
		$expected_alloc_kb = $opts{expected_alloc_kb};
	     } elsif ($file_set && $file_set->{expected_alloc_kb}) {
		$expected_alloc_kb = $file_set->{expected_alloc_kb};
	     } else {
		$expected_alloc_kb = $num_files * $size_kb;
	     }
	     logDbg("initialize_file_values  Line: ".__LINE__.
		      "\n\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "File prefix: \"$file_prefix\"\nExpeted alloc: $expected_alloc_kb Kb");


	     $opts{file_set}           = $file_set;
	     $opts{size_bytes}         = $size_bytes;
	     $opts{size}               = $size_bytes;
	     $opts{size_kb}            = $size_kb;
	     $opts{num_files}          = $num_files;
	     $opts{file_prefix}        = $file_prefix;
	     $opts{expected_alloc_kb}  = $expected_alloc_kb;
	     $opts{initialize_file_values_called} = 1;

#        logcomment("Exit initialize_file_values:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));

	     logDbg("initialize_file_values  Line: ".__LINE__.
		      "\nSize Kb:  \t$size_kb\nSize bytes:\t$size_bytes\nNum files:\t$num_files\n".
                      "Expeted alloc: $expected_alloc_kb Kb");
        };

        return { %opts };
}



sub initialize_quota_values(%)
{
	my (%opts) = @_;
	my $file_count = 0;
	logcomment("Enter initialize_quota_values\n\n".Dumper(\%opts));
        my ( $hash_out, $misc_opts, $comb_opts);
	eval {
	    my ($msg, $rslt, $file_cnt);

	    my $orig_quota_spec = $opts{quota_spec};
	    if (!$orig_quota_spec || (ref($orig_quota_spec) !~ /hash/i) || !%$orig_quota_spec )  {
		$msg = "FATAL:  initialize_quota_values  Line: ".__LINE__.
		          "\nMust specify 'quota_spec', 'quota_spec' is non-empty hashtable";
	       	logcomment($msg);
		die($msg);
	    }
	    my $quota_spec = { %$orig_quota_spec };
	    $misc_opts = $quota_spec->{Misc};
	    if (!$misc_opts || (ref($misc_opts) !~ /hash/i) || !%$misc_opts )  {
		$msg = "FATAL:  initialize_quota_values  Line: ".__LINE__.
		          "\nMust specify 'quota_spec', 'quota_spec' is non-empty hashtable";
	       	logcomment($msg);
		die($msg);
	    }

	    $misc_opts = { %$misc_opts };
	    $quota_spec->{Misc} = $misc_opts;
	    my $quota_desc = undef;
	    if ($quota_spec->{Desc}) {
		$quota_desc = $quota_spec->{Desc};
	    } elsif ($misc_opts->{Desc}) {
		$quota_desc = $misc_opts->{Desc};
	   }

	   my $Fc = $opts{Fc};
	   my $Vol = $opts{vol};

	    my $quota_type = $quota_spec->{type};
	    if (! grep( {$_ eq $quota_type} qw(tree user group)) ) {
		$msg = "FATAL:  initialize_quota_values  Line: ".__LINE__.
		           "Invalid 'type' value for quota: \"$quota_type\"";
	    }

       	    my $file_list = $opts{file_list};
	    if (!$opts{file_list} || (ref($file_list) !~ /array/i) || scalar(@$file_list) > 0)
#	    if (!$opts{file_list} || (ref($file_list) !~ /hash/i) || scalar(%$file_list) > 0)
	    {
		$msg = "FATAL:  initialize_quota_values  Line: ".__LINE__.
		          "\nMust specify 'file_list', 'file_list' must be empty array ref.";
#		          "\nMust specify 'file_list', 'file_list' must be empty hash ref.";
	       	logcomment($msg);
		die($msg);
	    }

	    $comb_opts = combine_hashtables(\%opts, $quota_spec, $misc_opts);

	    my ($h_or_s, $d_or_f, $disk, $files, $hard, $soft);
	    if ($comb_opts->{hard} || (defined($comb_opts->{h_or_s}) && $comb_opts->{h_or_s} eq 'hard')) {
		$h_or_s = 'hard';
		$soft = 0;
		$hard = 1;
	    } elsif ($comb_opts->{soft} || (defined($comb_opts->{h_or_s}) && $comb_opts->{h_or_s} eq 'soft')) {
		$h_or_s = 'soft';
		$soft = 1;
		$hard = 0;
	    } else {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		          "Must specify one of 'hard', 'soft'";
		logcomment($msg);
		die($msg);
	    }

	   if ($comb_opts->{disk} || (defined($comb_opts->{d_or_f}) && $comb_opts->{d_or_f} eq 'disk')) {
		$d_or_f = 'disk';
		$disk = 1;
		$files = 0;
	   } elsif ($comb_opts->{files} || (defined($comb_opts->{d_or_f}) && $comb_opts->{d_or_f} eq 'files')) {
		$d_or_f = 'files';
		$disk = 0;
		$files = 1;
	   } else {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		           "Must specify one of 'disk', 'files'";
		logcomment($msg);
		die($msg);
	   }

	   my ($proto_list, $proto);
	   if ($comb_opts->{Protos}) {
		$proto_list = $comb_opts->{Protos};
		$proto = $$proto_list[0];
	   } elsif ($comb_opts->{proto_list}) {
		$proto_list = $comb_opts->{proto_list};
		$proto = $$proto_list[0];
	   } elsif ($comb_opts->{proto}) {
		$proto = $comb_opts->{proto};
		$proto_list = [ $proto ];
	   } else {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		           "Must specify one of 'proto', 'proto_list', or 'Protos'";
		logcomment($msg);
		die($msg);
	   }

	   my ($dir_list, $dir);
	   if ($comb_opts->{Dirs}) {
		$dir_list = $comb_opts->{Dirs};
		$dir = $$dir_list[0];
	   } elsif ($comb_opts->{dir_list}) {
		$dir_list = $comb_opts->{dir_list};
		$dir = $$dir_list[0];
	   } elsif ($comb_opts->{dir}) {
		$dir = $comb_opts->{dir};
		$dir_list = [ $dir ];
	   } else {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		           "Must specify one of 'dir', 'dir_list', or 'Dirs'";
		logcomment($msg);
		die($msg);
	   }

	   my ($user_list, $user);
	   if ($comb_opts->{Users}) {
		$user_list = $comb_opts->{Users};
		$user = $$user_list[0];
	   } elsif ($comb_opts->{user_list}) {
		$user_list = $comb_opts->{user_list};
		$user = $$user_list[0];
	   } elsif ($comb_opts->{user}) {
		$user = $comb_opts->{user};
		$user_list = [ $user ];
	   } elsif ($quota_type eq 'user') {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		           "Must specify one of 'user', 'user_list', or 'Users'";
		logcomment($msg);
		die($msg);
	   }

	   my ($group_list, $group);
	   if ($comb_opts->{Groups}) {
		$group_list = $comb_opts->{Users};
		$group = $$group_list[0];
	   } elsif ($comb_opts->{group_list}) {
		$group_list = $comb_opts->{group_list};
		$group = $$group_list[0];
	   } elsif ($comb_opts->{group}) {
		$group = $comb_opts->{group};
		$group_list = [ $group ];
	   }

	   if (!@$user_list && !@$group_list) {
		$msg = "initialize_quota_values  Line: ".__LINE__."\n\n".
		           "Must specify one of 'user', 'user_list', 'group', or 'group_list'";
		logcomment($msg);
		die($msg);
	   }

	    delete($quota_spec->{Misc});
	    delete($quota_spec->{Desc});
	    delete($quota_spec->{disk});
	    delete($quota_spec->{files});
	    delete($quota_spec->{sdisk});
	    delete($quota_spec->{sfiles});

	    if ($soft) {
		if ($disk) {
			$quota_spec->{sdisk} = $Disk_quota_limit_kb;
		}
		if ($files) {
			$quota_spec->{sfiles} = $Files_quota_limit;
		}
	    }

	    if ($hard) {
		if ($disk) {
			$quota_spec->{disk} = $Disk_quota_limit_kb;
		}
		if ($files) {
			$quota_spec->{files} = $Files_quota_limit;
		}
	    }

	        $hash_out = { %opts };
	        $hash_out->{misc_opts} = { %{$misc_opts} };
	        $hash_out->{orig_quota_spec} = $orig_quota_spec;
	        $hash_out->{quota_spec} = $quota_spec;
	        $hash_out->{quota_desc} = $quota_desc;
		$hash_out->{h_or_s} = $h_or_s;
		$hash_out->{hard} = $hard;
		$hash_out->{soft} = $soft;
		$hash_out->{d_or_f} = $d_or_f;

		$hash_out->{disk} = $disk;
		$hash_out->{files} = $files;
		$hash_out->{file_list}            = $file_list;
		$hash_out->{proto_list}           = $proto_list          if ($proto_list && @$proto_list);
		$hash_out->{dir_list}             = $dir_list            if ($dir_list && @$dir_list);
		$hash_out->{user_list}            = $user_list           if ($user_list && @$user_list);
		$hash_out->{group_list}           = $group_list          if ($group_list && @$group_list);

		$hash_out->{proto}           = $$proto_list[0];
		$hash_out->{group}           = $$group_list[0];
		$hash_out->{user}            = $$user_list[0];
		$hash_out->{dir}             = $$dir_list[0];
		$hash_out->{initialize_quota_values_called} = 1;
		$hash_out->{quota_type} = $quota_spec->{type};
		$hash_out->{quota_tree} = $quota_spec->{tree};


		my $quota_target;
		my $quota_target_list;
		if ($quota_type eq 'user') {
		    my $orig_user_str                      = $quota_spec->{user};
		    my @orig_user_list                     = split(/\s*,\s*/, $orig_user_str);
		    if (!grep( { $User_hash{$_} }   @orig_user_list) ) {
			$msg = "FATAL: initialize_quota_values  Line: ".__LINE__."\n\n".
			          "Invalid user login in quota user list:\n\n".Dumpr(\@orig_user_list);
			logcomment($msg);
			die($msg);
		    }
		    $hash_out->{quota_target_user}         = $quota_spec->{user};
		    my $orig_user_list                     = [ @orig_user_list ];
		    $hash_out->{quota_target_user_list}    = $orig_user_list;
		    $quota_target                          = $quota_spec->{user};
		    $quota_target_list                     = $orig_user_list;
		    $quota_spec->{target}                  = $quota_target;
	            $hash_out->{quota_target}              = $quota_target;
	            $hash_out->{quota_target_list}         = $orig_user_list;
	            $hash_out->{quota_user_list}           = $orig_user_list;
		} elsif ($quota_type eq 'group') {
		    my $q_group = $quota_spec->{group};
		    if (!$Group_hash{$q_group}) {
			$msg = "FATAL: initialize_quota_values  Line: ".__LINE__."\n\n".
			          "Invalid group in quota: \"$q_group\"";
			logcomment($msg);
			die($msg);
		    }
		    $hash_out->{quota_target_group}        = $q_group;
		    $quota_target                          = $q_group;
		    $quota_spec->{target}                  = $quota_target;
	            $hash_out->{quota_target}              = $quota_target;
		} elsif ($quota_type eq 'tree') {
		    my $q_tree = $quota_spec->{tree};
		    if (!$Qtree_hash{$q_tree}) {
			$msg = "FATAL: initialize_quota_values  Line: ".__LINE__."\n\n".
			          "Invalid tree in quota: \"$q_tree\"";
			logcomment($msg);
			die($msg);
		    }
		    $hash_out->{quota_target_tree}         = $q_tree;
		    $quota_target                          = $q_tree;
		    $quota_spec->{target}                  = $quota_target;
	            $hash_out->{quota_target}              = $quota_target;
		}
		if (!$quota_target) {
			$msg = "FATAL: initialize_quota_values  Line: ".__LINE__."\n\n".
				"Target for \"$quota_type\" quota is not defined";
			logcomment($msg);
			die($msg);
		}

		if ($quota_type eq 'user' && $proto eq 'cifs') {
			my @user_list                       = @$quota_target_list;
			my @user_obj_list                   = map( {$User_hash{$_}}  @user_list );
			my $win_user_list                   = [ map( {$_->{full_win_login}} @user_obj_list ) ];
		        my $win_user_str                    = join(",",  @$win_user_list);
			$quota_target                       = $win_user_str;
			$quota_spec->{user}                 = $quota_target;
			$quota_spec->{target}               = $quota_target;
	                $hash_out->{quota_target}           = $quota_target;
	                $hash_out->{quota_target_list}      = $win_user_list;
	                $hash_out->{quota_user_list}        = $win_user_list;
		}

		unless ($hash_out->{initialize_file_values_called}) {
		     my $hash_ref = initialize_file_values(%$hash_out);
		     $hash_out = $hash_ref;
		}
	        my $file_set = $hash_out->{file_set};
	        if (!$hash_out->{file_set}) {
	             $msg = "initialize_quota_values  Line: ".__LINE__."\n".
		                "Could not find value for 'file_set'";
		     logcomment($msg);
		     die($msg);
	        }


	        my $qr_query = cvt_quota_spec_to_query($quota_spec);
	        $hash_out->{qr_query} = $qr_query;

# logcomment("initialize_quota_values  Line: ".__LINE__."\n\nQuota Spec:\n\n".Dumper($quota_spec)."\n\n".Dumper($hash_out->{quota_spec}));

	      push(@Test_quotas, $quota_spec);

	};

	my $fatal_msg = $@;
#        logcomment("Exit initialize_quota_values:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("initialize_quota_values  Returning\n\n\nHash Out\n\n".Dumper($hash_out));
	if ($fatal_msg) {
		logcomment($fatal_msg);
		die($fatal_msg);
	}
#        return($hash_out, $misc_opts, $comb_opts);
        return($hash_out);
}



sub cvt_quota_spec_to_query($)
{
	my ($qspec) = @_;
        logcomment("Enter cvt_quota_spec_to_query");

	my $qr_query = {
			    volume  => $qspec->{vol},
			    tree    => $qspec->{tree},
			    type    => $qspec->{type},
			    id      => $qspec->{target} || $qspec->{$qspec->{type}},
		      };
	logcomment("cv_quota_spec_to_query  Line: ".__LINE__.
			"\n\nQuota spec in:\n\n".Dumper($qspec).
			"\n\nReturning:\n\n".Dumper($qr_query));

#        logcomment("Exit cvt_quota_spec_to_query:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
        logcomment("Exit cvt_quota_spec_to_query");
	return($qr_query);
}


sub delete_quota($$)
{
	my ($fc, $qspec) = @_;
	my ( $msg, $rslt );
        logcomment("Enter delete_quota");
        eval {
		if (!$qspec) {
			$msg = "delete_quota  Line: ".__LINE__."\ndeleteQuota: 'quota spec' is undefined";
			die($msg);
		}
		$rslt = deleteQuota($fc, %$qspec);
		if (!$rslt || $@) {
			$msg = "delete_quota  Line: ".__LINE__."\ndeleteQuota: $@";
			die($msg);
		}
		my $qr_query = cvt_quota_spec_to_query($qspec);
		my @qr_list = queryQuotaReport(  $Fc, %$qr_query,);
		if (scalar(@qr_list) != 0) {
			$msg = "delete_quota  Line: ".__LINE__.
			          "\nExpected no results matching:\n\n".Dumper($qspec).
				  "\n\n.... but Found entries:\n\n".Dumper(\@qr_list);
			die($msg);
		}
	};
	my $fatal_msg = $@;
#        logcomment("Exit delete_quota:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("Exit delete_quota");
	return($fatal_msg);
}



sub delete_test_quotas()
{
	my $fail_msg_list = [];
	eval {
		my ($msg, $fail_list, $rslt);
                $rslt = restoreFilesPreserveBackups($Rootvol_mntpt, "/etc/quotas");
		if (!$rslt || $@) {
			$msg = "delete_test_quotas  Line: ".__LINE__.
				   "\nrestoreFilesPreserveBackups failed: $@";
			logcomment($msg);
			die($msg);
		}
	};

	my $fatal_msg = $@;
#        logcomment("Exit delete_test_quotas:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
	logcomment("Exit delete_test_quotas");
	return($fatal_msg);
}



###############################################################################################
###############################################################################################


#-----------------------------------------------------
#  quota_ems_setup()
#
#  Automatically called before executing tests
#  Setup the test-specific environment for the tests
#  Also, signal 'NA' result if test config not appropriate
#  for this test case.
#-----------------------------------------------------
sub quota_ems_setup(%) {
    my (%opts) = @_;
    logcomment("Enter quota_ems_setup");

    eval {
	 my ( $msg, $rslt );

         my $Filer         	= $opts{Fc}        	    || $main::FILER;
            $Vol          	= $opts{vol}          	    || $main::FILER_VOL;
            $DEBUG         	= $opts{debug}        	    || $main::DEBUG;
         my $Aggr          	= $opts{aggr}         	    || $main::FILER_AGGR;
            $nfs_client    	= $opts{client}       	    || $main::CLIENT;
            $cifs_client   	= $opts{cifs_client}  	    || $main::CIFS_CLIENT;
         my $setup_done    	= $opts{setup_done}   	    || $main::SETUP_DONE;
         my $cmdtimeout    	= $opts{cmdtimeout}   	    || $main::CMDTIMEOUT;
	 my $timeout		= $opts{timeout}      	    || $main::TIMEOUT;
	 my $filer_conn 	= $opts{filer_conn}         || $main::FILER_CONN;
            $EMS_MSG_POLL_TIME  = $opts{ems_msg_poll_time}  || $main::EMS_MSG_POLL_TIME;
	    $WRITE_BUFFER_SIZE  = $opts{write_buffer_size}  || $main::WRITE_BUFFER_SIZE;
	    $MIN_FILE_SIZE      = $opts{min_file_size}      || $main::MIN_FILE_SIZE;
 
##########################################################################################
    if ($DEBUG) {
	    #   Test/Debug Only
	    $EMS_MSG_POLL_TIME = 30;
	    $msg = "'setup' Line: ".__LINE__."\n\n".
	                  "WARNING: 'DEBUG' is set -- Test results are invalid\n\n"    .
			  "Setting EMS_MSG_POLL to $EMS_MSG_POLL_TIME sec.\n\n"        .
	                  "Tests:\n" . join("\n",  map({"\t".$_}  @{$main::Test->{tests}}) );
	    $Log->warn($msg);
    } else {
	    $msg = "'setup' Line: ".__LINE__."\n\n".
	                  "Tests:\n" . join("\n",  map({"\t".$_}  @{$main::Test->{tests}}) );
	    logcomment($msg);
    }
##########################################################################################


         @Backup_files = qw( /etc/passwd /etc/group /etc/quotas /etc/usermap.cfg );
	 %NFS_paths = ();
	 %Filer_paths = ();

         ## @step Create test volume if needed
         logStep("Create test volume if needed");
         logcomment("setup  Line: ".__LINE__."\nCall create_vol_and_mount");
         ($Fc, $Vol, $Mntpt) = create_vol_and_mount();
         logcomment("setup  Line: ".__LINE__."\ncreate_vol_and_mount returned:\n".
			"Fc: $Fc\nVol: $Vol\nMntpt: $Mntpt");
         $Fc->timeout($cmdtimeout);

         logcomment("Filer Version: \"" . filer_version_get($Fc) . "\"");

         ## @step Mount rootvol, and  Backup admin files to be modified
         logStep("Mount rootvol, and Backup admin files to be modified");
         logcomment("setup  Line: ".__LINE__."\nBackup admin files");
         $Rootvol_mntpt = mount_root_vol_if_needed(filer => $Fc, localhost => 1 );
         die("Could not mount root volume: $@") unless $Rootvol_mntpt;
         $rslt = backupFiles($Rootvol_mntpt, @Backup_files);
         if (!$rslt || $@) {
	     $msg = "setup  Line: ".__LINE__."\nbackupFiles: $@";
             # Don't 'restore' files if 'backup' failed...
             @Backup_files = ();
             die ($msg);
         }
         logcomment("setup  Line: ".__LINE__."\nFinished backing-up admin files");

         ## @step Iniitialize 'passwd', 'group', 'usermap.cfg' and 'quotas' files
         logStep("Iniitialize 'passwd', 'group', 'usermap.cfg' and 'quotas' files");
         logcomment("setup  Line: ".__LINE__."\nInitialize passwd, group, quota, and usermap files");
	 writeDefaultPasswdFile($Fc)
	    or die("setup:  writeDefaultPasswdFile: $@");
	 writeGroupFile($Fc, "")
	    or die("setup:  writeGroupFile: $@");
	 writeQuotaFile($Fc, "")
	    or die("setup:  writeQuotaFile: $@");
	 writeUsermapFile($Fc, "")
	    or die("setup:  writeUsermapFile: $@");
         logcomment("setup  Line: ".__LINE__."\nFinished initializing passwd, group, quota, and usermap files");

	 $NFS_paths{$Vol} = $Mntpt;
	 $Filer_paths{$Vol} = "/vol/$Vol";

         ## @step Set security style for $Vol
         logStep("Set security style for $Vol");
         logcomment("setup  Line: ".__LINE__."\nSet security style for volume $Vol");
         $Orig_vol_security = qtreeSecurity($Fc, $Filer_paths{$Vol})
               or die("setup  Line: ".__LINE__."\n(Get) qtreeSecurity for $Vol: $@");
	 qtreeSecurity($Fc, $Filer_paths{$Vol}, 'unix')
               or die("setup  Line: ".__LINE__."\n(Set) qtreeSecurity for $Vol: $@");
         logcomment("setup  Line: ".__LINE__."\nFinished setting security style for volume $Vol");

	my $use_cifs = 1;
	if ($use_cifs) {
	         ## @step Setup CIFS and map CIFS shares
	        logStep("Setup CIFS and map CIFS shares");
         	logcomment("setup  Line: ".__LINE__."\nCIFS_CLIENT specified: \"$cifs_client\"\n".
		                  "Setup CIFS share and mapping");
		 $msg = setup_cifs(Fc => $Fc, vol => $Vol);
	         die("setup  Line: ".__LINE__."\n$msg") if ($msg);
		 %Client_for_proto = {  'nfs'=>$nfs_client, 'cifs'=>$cifs_client };
         	 logcomment("setup  Line: ".__LINE__."\nFinished setting up CIFS");
	} else {
         	 logcomment("\nsetup  Line: ".__LINE__."\nNo CIFS_CLIENT specified, skip CIFS setup\n");
		 %Client_for_proto = {  'nfs'=>$nfs_client };
	}

         ## @step Create and setup qtrees
         logStep("Create and setup qtrees");
       	 logcomment("setup  Line: ".__LINE__."\nSetup qtrees and dirs");
	 $msg = create_qtrees_and_dirs(Fc => $Fc, vol=>$Vol, use_cifs=>$use_cifs);
         die("setup  Line: ".__LINE__."\n$msg") if ($msg);
       	 logcomment("setup  Line: ".__LINE__."\nFinished setting-up qtrees and dirs");

	 ## @step Create Quota target Users and Groups
	 logStep("Create Quota target Users and Groups");
       	 logcomment("setup  Line: ".__LINE__."\nSetup users and groups");
	 $msg = create_users_and_groups($Fc, use_cifs=>$use_cifs);
         die($msg) if ($msg);
       	 logcomment("setup  Line: ".__LINE__."\nFinished setting-up users and groups");

         ## @step Calculate quota limits, and file sizes and numbers
         logStep("Calculate quota limits, and file sizes and numbers");
       	 logcomment("setup  Line: ".__LINE__."\nCalculate file and quota sizes and limits");

	 $msg = calculate_sizes_and_limits(Fc => $Fc, vol => $Vol, );
         die("setup  Line: ".__LINE__."\n$msg") if ($msg);
       	 logcomment("setup  Line: ".__LINE__."\nFinished calculating file and quota sizes and limits");

         logcomment("setup  Line: ".__LINE__."   Setting 'File_set' to 'Jumbo_file_set' ....");
	 $File_set = $Jumbo_file_set;
	 logcomment("setup  Line: ".__LINE__."   .... Finished setting 'File_set' to 'Jumbo_file_set'");

	 ## @step Setup 'quota specifications'
         logStep("Setup 'quota specifications'");
       	 logcomment("setup  Line: ".__LINE__."\nSetup quota specs");
	 $msg = create_quota_specs(use_cifs=>$use_cifs, vol => $Vol);
         die("setup  Line: ".__LINE__."\n$msg") if ($msg);
       	 logcomment("setup  Line: ".__LINE__."\nFinished setting-up quota specs");

       	 logcomment("setup  Line: ".__LINE__."\nEmpty 'Test_quotas' list");
	 @Test_quotas = ();

         ## @step Setup 'tracking' quotas");
         logStep("Setup 'tracking' quotas");
       	 logcomment("setup  Line: ".__LINE__."\nAdd 'tracking' quotas");
	 setSingleQuota($Fc, type=>'user', default=>1, vol=>$Vol)
	       or die("setup  Line: ".__LINE__."\nsetSingleQuota: $@");
	 setSingleQuota($Fc, type=>'group', default=>1, vol=>$Vol)
	       or die("setup  Line: ".__LINE__."\nsetSingleQuota: $@");
	 setSingleQuota($Fc, type=>'tree', default=>1, vol=>$Vol)
	       or die("setup  Line: ".__LINE__."\nsetSingleQuota: $@");
       	 logcomment("setup  Line: ".__LINE__."\nFinished adding 'tracking' quotas");
    };

    my $fatal_msg = $@;
#        logcomment("Exit quota_ems_setup:\n\nMSG: \"$fatal_msg\"\nFail List:\n".Dumper($fail_msg_list));
    logcomment("Exit quota_ems_setup  Line: ".__LINE__."\nErrMsg: \"$fatal_msg\"");
    if ($fatal_msg) {
        return($fatal_msg, undef, undef, undef, undef);
    } else {
        return("", $Fc, $Vol, $Mntpt, $Rootvol_mntpt);
     }
}    # end quota_ems_setup()




#-----------------------------------------------------
#  cleanup()
#
# Automatically called after executing tests
#-----------------------------------------------------
sub quota_ems_cleanup(%) {
    my (%opts) = @_;
    my $Fc          = $opts{Fc}          || $main::Fc;
    my $filer       = $opts{filer}       || $main::FILER;
    my $cleanup     = $opts{cleanup}     || $main::CLEANUP;
    my $cmdtimeout  = $opts{cmdtimeout}  || $main::CMDTIMEOUT;
    my $timeout     = $opts{timeout}     || $main::TIMEOUT;
    my $filer_conn  = $opts{filer_conn}  || $main::FILER_CONN;

    logcomment("Enter quota_ems_cleanup");
    eval {
	my ($msg, $rslt);

        ## @step  Cleanup after test execution and/or setup.
	logcomment("\n>"x4 . (" "x8 . "="x20 ."  CLEANUP  " . "="x20 . " "x8 ."\n\n")x5 . "\n>"x4);
        logcomment("CLEANUP == \"$cleanup\"");

	sleep(15);
         if (!$Fc) {
             host($filer);
             $Fc = connect($filer_conn, -timeout, $timeout);
             $Fc->timeout($cmdtimeout);
         }
	 my $Fh = $Fc->hostp();
	 my $default_ip = $Fh->default_ip();

        if ($cleanup ne 'none') {

	     #######################################################
	     #                 Cleanup Test Volume                 *
	     #######################################################
	     if ($Vol) {
		 logcomment("cleanup:  Delete test directories");

		 #   Delete NFS, Mixed test dirs
                 for my $dir ( @NFS_dirs, @Mixed_dirs, @Null_dirs ) {
		     next unless ($NFS_paths{$dir});
		     eval {
			  logcomment("cleanup:  Delete dir: \"$dir\"");
		          if (test_file(file=>$NFS_paths{$dir}, test=>'d', client=>$nfs_client)) {
			       $rslt = remove_dir(   dir                   => $NFS_paths{$dir},
				                     client                => $nfs_client,
			                             recursive             => 1,
				                     okay_if_not_exists    => 1,
			                             user                  => 'root', 
		                       );
				$msg = "cleanup:  remove_dir($NFS_paths{$dir}): $@";
#				$Log->warn($msg) unless ($rslt);
				logcomment($msg) unless ($rslt);
			  }
		      };
		      if (test_file(file=>$NFS_paths{$dir}, test=>'d', client=>$nfs_client)) {
		           $msg = "Line: ".__LINE__."   remove_dir failed to delete ".
			                "directory $NFS_paths{$dir}";
			   logcomment($msg);
		      }
		 }

		 #   Delete CIFS test dirs
                 for my $dir ( @CIFS_dirs ) {
		     next unless ($CIFS_paths{$dir});
		     eval {
			  logcomment("cleanup:  Delete dir: \"$dir\"");
		          if (test_file(file=>$CIFS_paths{$dir}, test=>'d', client=>$cifs_client)) {
			       $rslt = remove_dir(  dir                   => $CIFS_paths{$dir},
				                    client                => $cifs_client,
			                            recursive             => 1,
				                    okay_if_not_exists    => 1,
			                            # user                  => 'administrator', 
		                        );
				$msg = "cleanup:  remove_dir($CIFS_paths{$dir}): $@";
#				$Log->warn($msg) unless ($rslt);
				logcomment($msg) unless ($rslt);
			  }
		      };
		      if (test_file(file=>$CIFS_paths{$dir}, test=>'d', client=>$cifs_client)) {
		           $msg = "Line: ".__LINE__."   remove_dir failed to delete ".
			                "directory $CIFS_paths{$dir}";
#			   $Log->warn($msg);
			   logcomment($msg);
		      }
		 }

		 #   Delete NFS, Mixed qtrees
                 logcomment("cleanup:  Delete qtrees: @NFS_qtrees");
                 for my $qt ( @NFS_qtrees, @Mixed_qtrees, @Null_qtrees) {
		       if ($NFS_paths{$qt->{name}}) {
			     eval {
			          if (test_file(file=>$NFS_paths{$qt->{name}}, test=>'d', client=>$nfs_client)) {
				       $rslt = remove_dir(  dir                   => $NFS_paths{$qt->{name}},
					                    client                => $nfs_client,
				                            recursive             => 1,
					                    okay_if_not_exists    => 1,
				                            user                  => 'root', 
			                       );
			               $msg = "cleanup:  remove_dir($NFS_paths{$qt->{name}}): $@";
#				       $Log->warn($msg) unless ($rslt);
				       logomment($msg) unless ($rslt);
				  }
			      };
			      if (test_file(file=>$NFS_paths{$qt->{name}}, test=>'d', client=>$nfs_client)) {
			           $msg = "Line: ".__LINE__."   remove_dir failed to delete ".
				                "directory $NFS_paths{$qt->{name}}";
#				   $Log->warn($msg);
				   logcomment($msg);
			      }
		        }
		        if ($Filer_paths{$qt->{name}}) {
			      if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
		    		    eval { treeSecurity($Fc, $Filer_paths{$qt->{name}}, 'unix') };
				    eval { qtreeDelete($Fc, $Filer_paths{$qt->{name}}) };
				    if ($@) {
					 $msg = "cleanup:  qtreeDelete:  $@";
#					 $Log->warn($msg);
					 logcomment($msg);
				    }
				    if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
					 $msg = "cleanup:  Could not delete qtree: $Filer_paths{$qt->{name}}";
#					 $Log->warn($msg);
					 logcomment($msg);
				    } else {
				         $msg = "cleanup:  Deleted qtree: $Filer_paths{$qt->{name}}";
#					 $Log->warn($msg);
					 logcomment($msg);
				    }
			    }
			    if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
				$msg = "cleanup:  Could not delete qtree: $Filer_paths{$qt->{name}}";
#				$Log->warn($msg);
				logcomment($msg);
			    } else {
			        $msg = "cleanup:  Deleted qtree: $Filer_paths{$qt->{name}}";
#				$Log->warn($msg);
				logcomment($msg);
			    }
		       }
                 }

		 #   Delete CIFS qtrees
                 logcomment("cleanup:  Delete qtrees: @CIFS_qtrees");
                 for my $qt ( @CIFS_qtrees ) {
		       if ($CIFS_paths{$qt->{name}}) {
			     eval {
			          if (test_file(file=>$CIFS_paths{$qt->{name}}, test=>'d', client=>$cifs_client)) {
				       $rslt = remove_dir(  dir                   => $CIFS_paths{$qt->{name}},
					                    client                => $cifs_client,
				                            recursive             => 1,
					                    okay_if_not_exists    => 1,
				                            # user                  => 'administrator', 
			                       );
			                $msg = "cleanup:  remove_dir($CIFS_paths{$qt->{name}}): $@";
#					$Log->warn($msg) unless($rslt);
					logcomment($msg) unless($rslt);
				  }
			      };
			      if (test_file(file=>$CIFS_paths{$qt->{name}}, test=>'d', client=>$cifs_client)) {
			           $msg = "Line: ".__LINE__."   remove_dir failed to delete ".
				                "directory $CIFS_paths{$qt->{name}}";
#				   $Log->warn($msg);
				   logcomment($msg);
			      }
		        }
		        if ($Filer_paths{$qt->{name}}) {
			      if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
		    		    eval { treeSecurity($Fc, $Filer_paths{$qt->{name}}, 'unix') };
				    eval { qtreeDelete($Fc, $Filer_paths{$qt->{name}}) };
				    if ($@) {
					$msg = "cleanup:  qtreeDelete:  $@";
#					$Log->warn($msg);
					logcomment($msg);
				    }
				    if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
					$msg = "cleanup:  Could not delete qtree: $Filer_paths{$qt->{name}}";
#					$Log->warn($msg);
					logcomment($msg);
				    } else {
				        logcomment("cleanup:  Deleted qtree: $Filer_paths{$qt->{name}}");
				    }
			    }
			    if (qtreePresent($Fc, $Filer_paths{$qt->{name}})) {
				$msg = "cleanup:  Could not delete qtree: $Filer_paths{$qt->{name}}";
#				$Log->warn($msg);
				logcomment($msg);
			    } else {
			        logcomment("cleanup:  Deleted qtree: $Filer_paths{$qt->{name}}");
			    }
		       }
                 }

		 #   Reset volume security style
                 logcomment("cleanup:  Reset security on volume");
		 eval {
		    $rslt = qtreeSecurity($Fc, $Filer_paths{$Vol}, $Orig_vol_security);
                    $msg = "cleanup:  qtreeSecurity:  $@";
#		    $Log->warn($msg) unless ($rslt);
		    logcomment($msg) unless ($rslt);
	         };

#		logcomment("cleanup Line: ".__LINE__."    ls -lR $NFS_paths{$Vol}\n\n.
#			      ".`ls -lR $NFS_paths{$Vol}`);

		 #   Disable 'quotas' for volume
                 logcomment("cleanup:  Disable quota");
                 eval {
	   	     $rslt = disableQuota($Fc, $Vol);
                     $msg = "cleanup:  disableQuota:  $@";
#		     $Log->warn($msg) unless ($rslt);
		     logcomment($msg) unless ($rslt);
                 };
             }
	     #######################################################
	     #             End Cleanup Test Volume                 *
	     #######################################################

	     #######################################################
	     #                 Shutdown CIFS                       *
	     #######################################################
	     if (is_cifs_setup(host=>$Fc)) {

                logcomment("cleanup: Remove cifs users and groups created");
                logcomment("cleanup: Remove cifs users and groups created");

		# Remove CIFS users
		foreach my $u (@CIFS_users_created) {
			my $wu = $u->{win_login};
#			if ($wu && findCIFSUser($Fc, $wu))
			if ($wu)
			{
			    eval {
				my $rslt = removeCIFSUser($Fc, $wu);
				if (!$rslt || $@) {
					my $msg = "cleanup  Line: ".__LINE__.
					             "removeCIFSUser \"$wu\"  failed: $@";
					die($msg);
				}
			    };
			}
		}

		# Remove CIFS groups
		foreach my $g (@CIFS_groups_created) {
			my $wg = $g->{win_group};
#			if ($wg && findCIFSGroup($Fc, $wg))
			if ($wg) {
			    eval {
				my $rslt = removeCIFSGroup($Fc, $wg);
				if (!$rslt || $@) {
					my $msg = "cleanup  Line: ".__LINE__.
					             "removeCIFSGroup \"$wg\"  failed: $@";
					die($msg);
				}
			    };
			}
		}


                logcomment("cleanup:  Unmap cifs shares");

		#  Unmap our shares on CIFS Client
		foreach my $drv (@All_cifs_drives) {
		    eval {
		        $rslt = unmap_cifs_share(client=>$cifs_client, drive=>$drv, force=>1);
			$msg = "cleanup:  unmap_cifs_share: $@";
#			$Log->warn($msg) unless($rslt);
			logcomment($msg) unless($rslt);
		    };
		    logcomment("cleanup  Line: ".__LINE__."   Error: $@") if ($@);
		}
	        $rslt = unmap_cifs_shares_for_server(  client=>$cifs_client,
					               server=>$default_ip,
					               force=>1);
		$msg = "cleanup:  unmap_cifs_share: $@";
#		$Log->warn($msg) unless($rslt);
		logcomment($msg) unless($rslt);
		my @maps = list_maps(client=>$cifs_client);
		$msg = "cleanup  Line: ".__LINE__."\n\n".
                             "After unmapping cifs shares -- CIFS_CLIENT: $cifs_client maps:\n\n".
		              join("\n", @maps);
		logcomment($msg);

		 #  Delete CIFS shares on filer
                 logcomment("cleanup:  Delete cifs shares");
		 foreach (@All_cifs_shares) {
		    eval {
		        $rslt = share_delete($Fc, $_);
			$msg = "cleanup:  share_delete: $@";
#			$Log->warn($msg);
			logcomment($msg);
		    };
		 }
                 $Log->trace("cleanup:  Stop CIFS service");
                 logcomment("cleanup:  Stop CIFS service");

		 #  Terminate CIFS on filer
		 eval {
	   	    $rslt = filer_cifs_stop($Fc);
			    $msg = "cleanup:  filer_cifs_stop: $@";
#			    $Log->warn($msg) unless ($rslt);
			    logcomment($msg) unless ($rslt);
		 };
	     }
	     #######################################################
	     #               End Shutdown CIFS                     *
	     #######################################################


	     #######################################################
	     #             Restore global filer config             #
	     #######################################################
             if ($Rootvol_mntpt) {

		 #  Restore 'backup' files
                 if (@Backup_files) {
                     logcomment("Restore backed-up files");
                     eval {
                         $rslt = restoreFiles($Rootvol_mntpt, @Backup_files);
                         $msg = "cleanup:  restoreFiles:  $@";
#			 $Log->warn($msg) unless ($rslt);
			 logcomment($msg) unless ($rslt);
                     };
		 }

		 #  Unmount root vol.
	         logcomment("Unmount root vol");
	         eval {
	             $rslt = umount_vol($Rootvol_mntpt);
	             $msg = "cleanup:  umount_vol $Rootvol_mntpt:  $@";
#		     $Log->warn($msg) unless ($rslt);
		     logcomment($msg) unless ($rslt);
	         };
	     }
	     if ($Mntpt) {
	         logcomment("Unmount $Vol");
	         eval {
	              $rslt = umount_vol($Mntpt);
	              $msg = "cleanup:  umount_vol $Mntpt:  $@";
#		      $Log->warn($msg) unless ($rslt);
		      logcomment($msg) unless ($rslt);
               	 };
             }
	     #######################################################
	     #           End Restore global filer config           #
	     #######################################################
        }

    };

    if ($@) {
#        $Log->warn("cleanup: $@");
        logcomment("quota_ems_cleanup: $@");
    }
    logcomment("Exit quota_ems_cleanup");
}    # end cleanup()


###############################################################################################
###############################################################################################




#------------------------------------------------------------------
#                End  Auxilliary Subroutines
#------------------------------------------------------------------



1;
