# Copyright (c) 2010-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for fabric management
## @author derricks@netapp.com
## @status public
## @pod here
## @unit_test thpl p4/test/lib/UnitTest/test_San_FabricMgr.thpl
#
# $Id: //depot/prod/test/Rsierranevada/lib/San/FabricMgr.pm#12 $

package San::FabricMgr;

# Compiler directives.
use strict;

# Load resource specific modules
use ParamsMgr;
use SubIPC;
use Tharn;
use Error;
use Data::Dumper;
use Digest::MD5 qw(md5_hex);

# Load San specific modules
use San::SanSwitch::SwitchUtil;
use San::SanSwitch::ZoneMgr;
use San::SanLog;
use San::SanUtil;

# NACL libs
use NACL::STask::Cluster;
use NACL::CS::Vserver;
use NACL::CS::LunIgroup;
use NACL::CS::NetworkInterface;
use NACL::CS::NetworkPort;
use NACL::C::Network;
use NATE::BaseException qw (:try);
use NACL::Exceptions::NoElementsFound;
use NACL::MTask::NetworkUtils;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

###############################################################################
#                               Public Methods
###############################################################################
###############################################################################
# Method:    new
# Objective: Create new FabricMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;

    my $self = {};

    my $class = ref($proto) || $proto;
    bless($self, $class);

    # Clear out the errors
    $self->_clear_err();

    # need to tell Error class to handle errors in a special way
    # this will allow us to invoke set err over and over to create an error stack
    # without overwriting the previous error
    $self->{error_dont_overwrite_errors} = 1;

    # Make sure the stack style is normal
    $self->{error_stack_style} = "normal";

    # store the arguments to the constructor in the args attribute
    $self->{args} = {&SanUtil::strip_dash(%args)};

    &ParamsMgr::process_args(%args);

    # Get some args
    my $parent            = &ParamsMgr::get_arg(-arg => "parent",            -required => 0, -default => bless({},"EnGenX"));
    my $fabric            = &ParamsMgr::get_arg(-arg => "fabric",            -required => 0, -default => "");
    my $xml_file          = &ParamsMgr::get_arg(-arg => "file",              -required => 0, -default => "");
    my $used_for_zoning   = &ParamsMgr::get_arg(-arg => "used_for_zoning",   -required => 0, -default => 1);
    my $switch_objs       = &ParamsMgr::get_arg(-arg => "switch_objs",       -required => 1, -default => []);
    my $client_objs       = &ParamsMgr::get_arg(-arg => "client_objs",       -required => 0, -default => []);
    my $sanfiler7m_objs   = &ParamsMgr::get_arg(-arg => "sanfiler7m_objs",   -required => 0, -default => []);
    my $cluster_objs      = &ParamsMgr::get_arg(-arg => "cluster_objs",      -required => 0, -default => []);
    my $nacl_cluster_objs = &ParamsMgr::get_arg(-arg => "nacl_cluster_objs", -required => 0, -default => []);
    my $vservers          = &ParamsMgr::get_arg(-arg => "vservers",          -required => 0, -default => []);
    my $vserver_map       = &ParamsMgr::get_arg(-arg => "vserver_map",       -required => 0, -default => {});
    my $protocols         = &ParamsMgr::get_arg(-arg => "protocols",
                                                -ref_type => "ARRAY",
                                                -default => ["fcp","iscsi","nfs","cifs","cluster"]);
    my $restrict          = &ParamsMgr::get_arg(-arg => "restrict_zoning_based_on_existing_igroups",
                                                -default => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $self;
    }

    # Process XML specified file if one was passed in
    if (defined($xml_file) and $xml_file ne "") {
        $self->_load_xml(-file => $xml_file);
        if ($self->error()) {
            $self->_set_err("Failed to load the xml file '$xml_file'");
            return $self;
        }
    } ## end if (defined($xml_file)...

    # Initialize some hashes
    my $client_hash        = {};
    my $switch_hash        = {};
    my $sanfiler7m_hash    = {};
    my $cmode_cluster_hash = {};

    # Validate that we actually received an array of switch objects and then store the switch
    # object in a switch hash keyed by the switch name
    unless (scalar(@$switch_objs) > 0) {
        $self->_set_err("Invalid value specified for '-switch_objs'");
        return $self;
    }

    foreach my $obj (@$switch_objs) {
        unless (ref($obj) =~ /sanswitch/i) {
            $self->_set_err("Invalid value specified for '-switch_objs'");
            return $self;
        }

        my $switch = $obj->hostname();
        if ($obj->error()) {
            $self->_set_err("Unable to get the name of the switch", $obj->errmsg());
            return $self;
        }
        unless (defined($switch) and $switch ne "") {
            $self->_set_err("Invalid data returned for switch hostname");
            return $self;
        }

        # Store the switch object in the switch hash
        $switch_hash->{$switch} = $obj;
    } ## end foreach my $obj (@$switch_objs)

    # Validate that we actually received an array of client objects
    if (scalar(@$client_objs) > 0) {
        foreach my $obj (@$client_objs) {
            unless (ref($obj) =~ /sanclient/i) {
                $self->_set_err("Invalid value specified for '-client_objs'");
                return $self;
            }

            my $client = $obj->hostname();
            if ($obj->error()) {
                $self->_set_err("Unable to get the name of the client", $obj->errmsg());
                return $self;
            }
            unless (defined($client) and $client ne "") {
                $self->_set_err("Invalid data returned for client hostname");
                return $self;
            }

            # Store the client object in the client hash
            $client_hash->{$client} = $obj;
        } ## end foreach my $obj (@$client_objs)
    }

    # Validate that we actually received an array of 7mode objects.  Having 7 mode objects is not required.
    if (scalar(@$sanfiler7m_objs) > 0) {
        foreach my $obj (@$sanfiler7m_objs) {
            unless (ref($obj) =~ /sanfiler/i) {
                $self->_set_err("Invalid value specified for '-7mode_objs'");
                return $self;
            }

            my $filer = $obj->get_name();
            if ($obj->error()) {
                $self->_set_err("Unable to get the name of the 7mode filer", $obj->errmsg());
                return $self;
            }
            unless (defined($filer) and $filer ne "") {
                $self->_set_err("Invalid data returned for the 7mode filer hostname");
                return $self;
            }

            # Store the switch object in the switch hash
            $sanfiler7m_hash->{$filer} = $obj;
        } ## end foreach my $obj (@$sanfiler7m_objs)
    } ## end if (scalar(@$sanfiler7m_objs...

    # if we received any nacl cluster objects, verify they are proper, and
    # add them to the cmode cluster hash
    if (scalar(@$nacl_cluster_objs) > 0) {
        foreach my $nacl_cluster_obj (@$nacl_cluster_objs) {
            if (not ref($nacl_cluster_obj) or not $nacl_cluster_obj->isa("NACL::C::Cluster")) {
                $self->_set_err("Invalid NACL::C::Cluster object specified in value of '-nacl_cluster_objs'");
                return $self;
            }

            try {
                # extract the name of the cluster and add it to the cmode cluster hash
                $cmode_cluster_hash->{$nacl_cluster_obj->name()} = $nacl_cluster_obj;
            } otherwise {
                my $e = shift;
                $self->_set_err("NACL Exception caught", $e->text());
            };
            if ($self->error()) {
                return $self;
            }
        }
    }

    # Validate that we actually received an array of controller cluster objects.  Having controller cluster
    # objects is not required.
    if (scalar(@$cluster_objs) > 0) {
        foreach my $obj (@$cluster_objs) {
            if (ref($obj) !~ /San::ControllerCluster/i) {
                $self->_set_err("Invalid value specified for '-cluster_objs'");
                return $self;
            }

            my $cluster = $obj->get_name(-use_cache => 1);
            if ($obj->error()) {
                $self->_set_err("Unable to get the name of the controller cluster", $obj->errmsg());
                return $self;
            }
            unless (defined($cluster) and $cluster ne "") {
                $self->_set_err("Invalid data returned for the controller cluster name");
                return $self;
            }

            # if this cluster already lives in the cmode cluster hash, then skip it
            # there is no need to convert it into a NACL object
            if (exists($cmode_cluster_hash->{$cluster})) {
                next;
            }

            my ($node_name) = $obj->get_node_members();
            if ($obj->error()) {
                $self->_set_err("Failed to obtain node member names for cluster '$cluster'");
                return $self;
            }

            try {
                # convert this cluster into a NACL::STask::Cluster object
                my $nacl_cluster_obj = NACL::STask::Cluster->new(node => $node_name);
                # Store the nacl cluster object in the cmode cluster hash
                $cmode_cluster_hash->{$cluster} = $nacl_cluster_obj;
            } otherwise {
                my $e = shift;
                $self->_set_err("NACL Exception caught", $e->text());
            };
            if ($self->error()) {
                return $self;
            }

        } ## end foreach my $obj (@$cluster_objs)
    } ## end if (scalar(@$cluster_objs...

    # Get the fabric info from the XML
    my $all_fabrics = &ParamsMgr::search(
        -search_path => [],
        -key         => "FABRICS",
        -key_filter  => qr/^FABRIC\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );

    # Massage the all_fabrics hash and create a new hash that looks like:
    #
    # $fabrics =>
    #             <fabric_name> =>
    #                   zones_per_node_per_init => <brocade virtual fabric id> or undef
    #                   master_switch           => switch to run commands on
    #                   prefix                  => prefix
    #                   switches                =>
    #                           <switch_name> =>
    #                                   admin_domain            => <admin domain> or undef
    #                                   virtual_fabric          => <cisco virtual fabric> or undef
    #                                   fabric_id               => <brocade virtual fabric id> or undef
    #                           <switch_name> =>
    #                                   admin_domain            => <admin domain> or undef
    #                                   virtual_fabric          => <cisco virtual fabric> or undef
    #                                   fabric_id               => <brocade virtual fabric id> or undef
    my $fabrics_hash = $self->_transform_fabrics_xml(-all_fabrics => $all_fabrics, -fabric => $fabric);
    if ($self->error()) {
        $self->_set_err("Error transforming fabric hash");
        return $self;
    }

    # Initialize the object attributes
    $self->{parent}                                                      = $parent;
    $self->{switches}                                                    = $switch_hash;
    $self->{clients}                                                     = $client_hash;
    $self->{sanfiler7m}                                                  = $sanfiler7m_hash;
    $self->{cmode_clusters}                                              = $cmode_cluster_hash;
    $self->{cmode_vservers}                                              = $vservers;
    $self->{fabrics}                                                     = $fabrics_hash;
    $self->{resources}                                                   = undef;
    $self->{zonemgr_objs}                                                = undef;
    $self->{switch_info}                                                 = undef;
    $self->{fabric}                                                      = $fabric;
    $self->{xml_file}                                                    = $xml_file;
    $self->{protocols}                                                   = $protocols;
    $self->{initiator_wwpn_to_target_wwpn_map}                           = undef;
    $self->{restrict_zoning_based_on_existing_igroups}                   = $restrict;
    $self->{deleted_wwpns_info}                                          = {};
    $self->{aliases_and_zones_created}                                   = {};
    $self->{aliases_and_zones_created}->{totals_info}->{initiator_aliases_created} = 0;
    $self->{aliases_and_zones_created}->{totals_info}->{initiator_aliases_reused}  = 0;
    $self->{aliases_and_zones_created}->{totals_info}->{target_aliases_created}    = 0;
    $self->{aliases_and_zones_created}->{totals_info}->{target_aliases_reused}     = 0;
    $self->{aliases_and_zones_created}->{totals_info}->{zones_created}   = 0;
    $self->{aliases_and_zones_created}->{totals_info}->{zones_reused}    = 0;
    $self->{used_for_zoning}                                             = $used_for_zoning;

    # See if we were passed in a populated vserver map.  If we weren't, then replace $self->{client_vserver_map}
    if (scalar(keys %$vserver_map) > 0) {
        $self->{client_vserver_map} = $vserver_map;
    } else {
        $self->{client_vserver_map} = undef;
    }

    # Validate the switches and fabrics against each other
    $self->_validate_fabric_switch_objects();
    if ($self->error()) {
        $self->_set_err("There was an issue with switch fabric validation");
        return $self;
    }

    return $self;
} ## end sub new

###############################################################################
# Method:    create_aliases
# Objective: Create all of the aliases on the Fabric
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_aliases {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the prefix of the fabric
        my $prefix = $self->get_prefix(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Error retrieving prefix for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Build a hash of initiator and target hashes
        my $devices = {};
        foreach my $init (sort keys %{$resources->{$fabric}->{initiators}}) {
            if (defined($resources->{$fabric}->{initiators}->{$init}->{targets}) and
                scalar(@{$resources->{$fabric}->{initiators}->{$init}->{targets}}) > 0) {
                $devices->{$init} = $resources->{$fabric}->{initiators}->{$init};
            } else {
                $self->{deleted_wwpns_info}->{sanclients}->{$init}->{device_name} =
                    $resources->{$fabric}->{initiators}->{$init}->{device};
                $self->{deleted_wwpns_info}->{sanclients}->{$init}->{sanclients} =
                    $resources->{$fabric}->{initiators}->{$init}->{name};
                $self->{deleted_wwpns_info}->{sanclients}->{$init}->{error} .=
                    "\nDid not alias initiator $init because there were not targets to zone it with.\n";
            }
        }
        foreach my $target (sort keys %{$resources->{$fabric}->{targets}}) {
            if (defined($resources->{$fabric}->{targets}->{$target}->{initiators}) and
                scalar(@{$resources->{$fabric}->{targets}->{$target}->{initiators}}) > 0) {
                $devices->{$target} = $resources->{$fabric}->{targets}->{$target};
            } else {
                $self->{deleted_wwpns_info}->{vservers}->{$target}->{lif_name} =
                    $resources->{$fabric}->{initiators}->{$target}->{device};
                $self->{deleted_wwpns_info}->{vservers}->{$target}->{cluster_name} =
                    $resources->{$fabric}->{initiators}->{$target}->{cluster_name};
                $self->{deleted_wwpns_info}->{vservers}->{$target}->{curr_node_name} =
                    $resources->{$fabric}->{initiators}->{$target}->{node};
                $self->{deleted_wwpns_info}->{vservers}->{$target}->{vservers} =
                    $resources->{$fabric}->{initiators}->{$target}->{name};
                $self->{deleted_wwpns_info}->{vservers}->{$target}->{error} .=
                    "\nDid not alias target $target because there were no initiators to zone it with.\n";
            }
        }

        # Iterate over the initators and targets and create new alias objects
        foreach my $adapter (sort keys %$devices) {
            my $ref = $devices->{$adapter};
            unless (defined($ref->{name})
                and $ref->{name} ne ""
                and defined($ref->{device})
                and $ref->{device} ne ""
                and defined($ref->{cluster_name})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$adapter'");
                return $SanUtil::FALSE;
            } ## end unless (defined($ref->name...

            # if this is a target and it is destined to be zoned to an initiator
            # that is requesting multi target single initiator zoning and that
            # client is requesting that no target aliases be created for it's zones
            # then skip this target and dont create an alias for it
            if (exists $ref->{initiators}) {
                # this is a target, find all initiator client names that are to be zoned to it
                my $dont_create = 0;
                my $do_create = 0;
                foreach my $init (@{$ref->{initiators}}) {
                    my $client_name = $resources->{$fabric}->{initiators}->{$init}->{name};
                    if ($self->get_client_hash()->{$client_name}->{fcp}->{CONFIG_FCP_DONT_CREATE_TARGET_ALIASES}) {
                        # dont create an alias for this target
                        $dont_create = 1;
                    } else {
                        # do create an alias for this target
                        $do_create = 1;
                    }
                }
                if ($dont_create and $do_create) {
                    # there is a discrepency between initiators, some want it to have an alias, some dont
                    # we need to fail here because this is a bullshit request
                    $self->_set_err("There is a discrepency between multiple initiators on whether " .
                                    "we should create an alias for the target wwpn '$adapter'. This is invalid. " .
                                    "If you have clients that share FCP/FCoE relationships to the same set of " .
                                    "vservers, make sure that those clients share the same fcp_option value for the" .
                                    "'dont_create_target_aliases' option.");
                    return $SanUtil::FALSE;
                } elsif ($dont_create) {
                    # all hosts to be zoned to this target dont want it to have an alias,
                    # so dont make an alias for it
                    $ref->{dont_create_alias} = 1;
                    next;
                } else {
                    # all hosts to be zoned to this target want it to have an alias,
                    # continue on...
                }
            }

            # generate the alias name for this device
            my $alias_name = $self->_get_alias_name(
                -host    => $ref->{name},
                -device  => $ref->{device},
                -cluster => $ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get alias name");
                return $SanUtil::FALSE;
            }
            my $type = $ref->{type};

            # First see if an alias already exists by that name on the fabric
            my $alias_obj = $zonemgr_obj->get_aliases(-alias => $alias_name);
            if ($zonemgr_obj->error()) {
                $self->_set_err(
                    "Error while trying to retrieve the object for alias '$alias_name'", $zonemgr_obj->errmsg());
                return $SanUtil::FALSE;
            }
            if (ref($alias_obj) =~ /alias/i) {

                # Verify the alias
                $self->_verify_alias(
                    -alias  => $alias_name,
                    -member => $adapter,
                    -fabric => $fabric
                );
                if ($self->error()) {
                    # fix the alias if ("Member on alias object ... differs")
                    if ($self->errmsg() =~ /Member on alias object '(\S+?)' differs from fabric definition '(\S+?)'/i) {
                        my $curr_member = $1;
                        my $new_member = $2;

                        &SanUtil::nate_log("-comment", "Alias '$alias_name' already exists on the fabric but has an invalid member ($curr_member), need to replace with ($new_member)");

                        # replace curr member with new member
                        $alias_obj->add_member($new_member);
                        if ($alias_obj->error()) {
                            $self->_set_err("Unable to add member '$new_member' to alias '$alias_name'", $alias_obj->errmsg());
                            return $SanUtil::FALSE;
                        }
                        $alias_obj->remove_member($curr_member);
                        if ($alias_obj->error()) {
                            $self->_set_err("Unable to remove member '$curr_member' from alias '$alias_name'", $alias_obj->errmsg());
                            return $SanUtil::FALSE;
                        }
                        &SanUtil::nate_log("-comment", "Alias '$alias_name' was corrected with alias member ($new_member)");
                        $self->_clear_err();
                    } else {
                        $self->_set_err("Unable to verify alias '$alias_name'");
                        return $SanUtil::FALSE;
                    }
                } else {
                    # Alias was found to be valid so print a message and skip to the next one.
                    &SanUtil::nate_log("-comment", "Alias '$alias_name' already exists on the fabric");
                }

                # create the alias info hash
                my %alias_info_hash = ( alias=>$alias_name,
                                        wwpn=>$adapter);
                # Clients resources will have a 'targets' array to zone with.
                # Vservers resources will an 'initiators' array to zone with.
                # Use the above information to determine whether the ref to the resource
                # is a client or vserver and update the appropriate place in the
                # alias and zones created hash.
                if(exists $ref->{targets}) {
                    push(@{$self->{aliases_and_zones_created}->{aliases}->{clients}->{$ref->{name}}->{$fabric}->{reused}},\%alias_info_hash);
                    $self->{aliases_and_zones_created}->{totals_info}->{initiator_aliases_reused}++;
                } else {
                    push(@{$self->{aliases_and_zones_created}->{aliases}->{vservers}->{$ref->{cluster_name}}->{$ref->{name}}->{$fabric}->{reused}},\%alias_info_hash);
                    $self->{aliases_and_zones_created}->{totals_info}->{target_aliases_reused}++;
                }
                next;
            } else {

                # We didn't get an alias object back so we need to go ahead and create one.
                my $alias_obj = $zonemgr_obj->new_alias(
                    -alias  => $alias_name,
                    -member => $adapter,
                    -type   => $type
                );
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Unable to create object for alias '$alias_name'", $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (ref($alias_obj) =~ /Alias/i) {
                    $self->_set_err("Invalid object returned for alias '$alias_name'");
                    return $SanUtil::FALSE;
                }
            } ## end else [ if (ref($alias_obj) =~...

            # If we have made it this far then we know that the alias doesn't exist on the fabric
            # and that there were no errors.  We will now create the alias.
            $zonemgr_obj->alias_create(-alias => $alias_name);
            if ($zonemgr_obj->error()) {
                $self->_set_err("Unable to create alias '$alias_name' on the fabric", $zonemgr_obj->errmsg());
                return $SanUtil::FALSE;
            }
            &SanUtil::nate_log("-comment", "Alias '$alias_name' created on the fabric");

            # create the alias info hash
            my %alias_info_hash = ( alias=>$alias_name,
                                    wwpn=>$adapter);
            # Clients will have a 'targets' array, vservers will an 'initiators' array instead
            if(exists $ref->{targets}) {
                push(@{$self->{aliases_and_zones_created}->{aliases}->{clients}->{$ref->{name}}->{$fabric}->{new}},\%alias_info_hash);
                $self->{aliases_and_zones_created}->{totals_info}->{initiator_aliases_created}++;
            } else {
                push(@{$self->{aliases_and_zones_created}->{aliases}->{vservers}->{$ref->{cluster_name}}->{$ref->{name}}->{$fabric}->{new}},\%alias_info_hash);
                $self->{aliases_and_zones_created}->{totals_info}->{target_aliases_created}++;
            }
            next;
            } ## end foreach my $adapter (sort keys...
    } ## end foreach my $fabric (keys %$resources)
    return $SanUtil::TRUE;
} ## end sub create_aliases

###############################################################################
# Method:    create_zones
# Objective: Create all of the zones on the fabric for the resources provided
#            and add them to the zoneset specified on the object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_zones {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the prefix or the fabric
        my $prefix = $self->get_prefix(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Error retrieving prefix for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Build a hash of initiator and target hashes
        my $devices = {};
        foreach my $init (sort keys %{$resources->{$fabric}->{initiators}}) {
            if (defined($resources->{$fabric}->{initiators}->{$init}->{targets}) and
                scalar(@{$resources->{$fabric}->{initiators}->{$init}->{targets}}) > 0) {
                $devices->{$init} = $resources->{$fabric}->{initiators}->{$init};
            }
        }
        foreach my $target (sort keys %{$resources->{$fabric}->{targets}}) {
            if (defined($resources->{$fabric}->{targets}->{$target}->{initiators}) and
                scalar(@{$resources->{$fabric}->{targets}->{$target}->{initiators}}) > 0) {
                $devices->{$target} = $resources->{$fabric}->{targets}->{$target};
            }
        }

        # Get the active zoneset
        my $zoneset = $self->get_active_zoneset(-fabric => $fabric);
        unless (defined($zoneset) and $zoneset ne "") {
            $self->_set_err("No zoneset is defined for fabric '$fabric'.  Unable to continue without"
                  . "having an active zoneset already configured on the fabric.");
            return $SanUtil::FALSE;
        }

        # Iterate over the initators and targets and make sure all of the resources
        # exist out on the fabric.  If they don't, then set an error and bail out.
        foreach my $adapter (sort keys %$devices) {
            my $ref = $devices->{$adapter};
            unless (defined($ref->{name})
                and $ref->{name} ne ""
                and defined($ref->{device})
                and $ref->{device} ne ""
                and defined($ref->{cluster_name})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$adapter'");
                return $SanUtil::FALSE;
            } ## end unless (defined($ref->{name...

            # not all devices may have had aliases created for them,
            # skip the verification step for those
            # that explicitly were set to not have aliases
            next if ($ref->{dont_create_alias});

            my $alias_name = $self->_get_alias_name(
                -host    => $ref->{name},
                -device  => $ref->{device},
                -cluster => $ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get alias name");
                return $SanUtil::FALSE;
            }

            my $alias_obj = $zonemgr_obj->get_aliases(-alias => $alias_name);
            if ($zonemgr_obj->error) {
                $self->_set_err(
                    "Error while trying to retrieve the object for alias '$alias_name'", $zonemgr_obj->errmsg());
                return $SanUtil::FALSE;
            }
            unless (ref($alias_obj) =~ /alias/i) {
                # Alias doesn't exist in the cache.
                $self->_set_err("Alias '$alias_name' doesn't exist in the fabric.");
                return $SanUtil::FALSE;
            } ## end unless (ref($alias_obj) =~...
        } ## end foreach my $adapter (sort keys...

        # Iterate over the initators
        foreach my $initiator (sort keys %{$resources->{$fabric}->{initiators}}) {
            # Skip any initiator that doesn't have targets to zone
            unless (defined($resources->{$fabric}->{initiators}->{$initiator}->{targets}) and
                    scalar(@{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) > 0) {
                # No targets so skip this initiator
                next;
            }

            # Get the client name for the initiator
            my ($client, $undef1, $undef2) = $self->get_wwpn_name(-wwpn => $initiator);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the client name for '$initiator'");
                return $SanUtil::FALSE;
            }

            my $client_hash = $self->get_client_hash();
            my $multi_target_zoning = $self->get_client_hash()->{$client}->{fcp}->{CONFIG_FCP_MULTI_TARGET_ZONING};
            my $no_target_aliases   = $self->get_client_hash()->{$client}->{fcp}->{CONFIG_FCP_DONT_CREATE_TARGET_ALIASES};
            my $init_ref = $resources->{$fabric}->{initiators}->{$initiator};

            unless (defined($init_ref->{name})
                and $init_ref->{name} ne ""
                and defined($init_ref->{device})
                and $init_ref->{device} ne ""
                and defined($init_ref->{cluster_name})) {
                    $self->_set_err("Name or device missing from resource hash for wwpn '$initiator'");
                    return $SanUtil::FALSE;
            } ## end unless (defined($init_ref->...

            # Build the alias name.
            my $init_name = $self->_get_alias_name(
                -host    => $init_ref->{name},
                -device  => $init_ref->{device},
                -cluster => $init_ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get the alias name for '$init_ref->{name}'");
                return $SanUtil::FALSE;
            }

            # Iterate over the targets defined on the initiator
            if ($multi_target_zoning) {
                my $targets_ref = $resources->{$fabric}->{initiators}->{$initiator}->{targets};
                my @target_alias_array;
                my @target_wwpn_array;
                foreach my $target (@{$targets_ref}) {
                    # if the current initiator has requested that no target aliases be created/used
                    # for its zones, then push the target wwpn onto the array instead of the
                    # alias name
                    if ($no_target_aliases) {
                        push(@target_wwpn_array, $target);
                    } else {
                        # fetch the alias name for this target
                        my $target_ref = $resources->{$fabric}->{targets}->{$target};
                        # Build the alias name
                        my $target_name = $self->_get_alias_name(
                            -host    => $target_ref->{name},
                            -device  => $target_ref->{device},
                            -cluster => $target_ref->{cluster_name},
                            -prefix  => $prefix);
                        if ($self->error()) {
                            $self->_set_err("Unable to get the alias name for '$target_ref->{name}'");
                            return $SanUtil::FALSE;
                        }
                        push(@target_alias_array, $target_name);
                    }
                }
                # remove the prefix from the zone name
                my $init_zone_suffix = $init_name;
                $init_zone_suffix =~ s/^${prefix}_//;
                my $zone_name = $prefix . "_multi_" . $init_zone_suffix;
                my $zone_obj = $zonemgr_obj->get_zones(-zone => $zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err(
                        "Error while trying to retrieve the object for zone '($zone_name)'",
                        $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                } ## end if ($zonemgr_obj->error...

                if (ref($zone_obj) =~ /zone/i) {
                    &SanUtil::nate_log("-comment", "Zone '($zone_name)' already exists, reusing it.");
                    # Zone was found to be valid check to see if it is in the zoneset for this fabric.
                    my $zonesets_ref = $zonemgr_obj->zone_get_zonesets(-zone => $zone_name, -type => "switch");
                    if ($zonemgr_obj->error()) {
                        $self->_set_err("Unable to get the list of zonesets for zone '($zone_name)'",
                                        $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                    unless (ref($zonesets_ref) eq "HASH") {
                        $self->_set_err("Invalid data type returned from zone_get_zonesets");
                        return $SanUtil::FALSE;
                    }

                    # This multi-zone currently exists, this means that we need to add all elements in
                    # @target_alias_array/@target_wwpn_array to this zone if they are not already members to it.
                    foreach my $alias_name (@target_alias_array) {
                        # if this alias is not a member of the zone, add it
                        if (!$zone_obj->is_member(alias => $alias_name)) {
                            if ($zone_obj->error()) {
                                $self->_set_err("Error while trying to determine if alias $alias_name is a member of zone '$zone_name'",
                                                $zone_obj->errmsg());
                                return $SanUtil::FALSE;
                            }
                            my $alias_obj = $zonemgr_obj->get_aliases(-alias => $alias_name);
                            if ($zonemgr_obj->error()) {
                                $self->_set_err("Error while trying to obtain alias object for alias '$alias_name'",
                                                $zonemgr_obj->errmsg());
                                return $SanUtil::FALSE;
                            }
                            $zone_obj->add_alias(alias => $alias_obj);
                            if ($zone_obj->error()) {
                                $self->_set_err("Error while trying to add alias '$alias_name' to the zone '$zone_name'",
                                                $zone_obj->errmsg());
                                return $SanUtil::FALSE;
                            }
                        }
                    }
                    foreach my $wwpn (@target_wwpn_array) {
                        # if this wwpn is not a member of the zone, add it
                        if (!$zone_obj->is_member(wwpn => $wwpn)) {
                            if ($zone_obj->error()) {
                                $self->_set_err("Error while trying to determine if wwpn $wwpn is a member of zone '$zone_name'",
                                                $zone_obj->errmsg());
                                return $SanUtil::FALSE;
                            }
                            $zone_obj->add_wwpn(wwpn => $wwpn);
                            if ($zone_obj->error()) {
                                $self->_set_err("Error while trying to add wwpn '$wwpn' to the zone '$zone_name'",
                                                $zone_obj->errmsg());
                                return $SanUtil::FALSE;
                            }
                        }
                    }

                    # Check to see if $zoneset is found in the zonesets_ref hash.  If it isn't, add the zone
                    # to that zoneset.
                    if (defined($zonesets_ref->{$zoneset})) {
                        # Zone was found in the zoneset so just log a message and skip it.
                        &SanUtil::nate_log("-comment", "Zone '($zone_name)' already exists on the " .
                                           "fabric in zoneset '$zoneset'");
                    } else {
                        # Add the Zone to the Zoneset
                        $zonemgr_obj->zoneset_add_zone(-zoneset => $zoneset, -zone => $zone_name);
                        if ($zonemgr_obj->error()) {
                            $self->_set_err("Error while trying to add zone '($zone_name)' to " .
                                            "zoneset '$zoneset'", $zonemgr_obj->errmsg());
                            return $SanUtil::FALSE;
                        } ## end if ($zonemgr_obj->error...
                        &SanUtil::nate_log("-comment", "Zone '($zone_name)' added to" .
                                           "zoneset '$zoneset'");
                    }

                    # Initiator Object are for Clients
                    # Updating the count is to detect how many times this is used in the zone
                    # so that we are prepared for more complex many to one zone types.
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{reused}->{$zone_name}->{initiators}->{$init_name}++;
                    foreach my $target (@{$targets_ref}) {
                        my $target_ref = $resources->{$fabric}->{targets}->{$target};
                        $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$zone_name}->{targets}->{$target_ref->{name}}++;
                        # Targets Objects are for Vservers
                        # Updating the count is to detect how many times this is used in the zone
                        # so that we are prepared for more complex many to one zone types.
                        $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$zone_name}->{targets}->{$target_ref->{name}}++;
                        $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$zone_name}->{initiators}->{$init_name}++;
                        $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$zone_name}->{full_name} = $zone_name;
                    }
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{reused}->{$zone_name}->{full_name} = $zone_name;
                    $self->{aliases_and_zones_created}->{zones}->{totals_info}->{zones_reused}++;

                    next;
                }
                # IF WE GOT HERE, MAKE A BRAND NEW ZONE
                #
                #
                #
                my $init_obj = $zonemgr_obj->get_aliases(-alias => $init_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Error while trying to retrieve the object for alias '$init_name'",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (defined($init_obj) and ref($init_obj) =~ /alias/i) {
                    $self->_set_err("Invalid object data for alias '$init_name'");
                    return $SanUtil::FALSE;
                }

                # build the aliases array to use when creating the zone. This will contain all
                # initiator and target aliases that will be members of this zone
                my @aliases = ($init_obj);
                foreach my $target_name (@target_alias_array) {
                    my $target_obj = $zonemgr_obj->get_aliases(-alias => $target_name);
                    if ($zonemgr_obj->error()) {
                        $self->_set_err("Error while trying to retrieve the object for alias '$target_name'",
                                       $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                    unless (defined($target_obj) and ref($target_obj) =~ /alias/i) {
                        $self->_set_err("Invalid object data for alias '$target_name'");
                        return $SanUtil::FALSE;
                    }
                    push(@aliases, $target_obj);
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$zone_name}->{targets}->{$target_name}++;
                }

                # add all the target wwpns that did not have aliases created for them
                # to the zoning history hash
                foreach my $target_wwpn (@target_wwpn_array) {
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$zone_name}->{targets}->{$target_wwpn}++;
                }

                # Create the zone object using all the aliases and wwpns that we set aside for it
                $zone_obj = $zonemgr_obj->new_zone(
                    -zone    => $zone_name,
                    -aliases => \@aliases,
                    -wwpns   => \@target_wwpn_array
                );
                if ($zonemgr_obj->error) {
                    $self->_set_err("Unable to create object for zone '$zone_name'",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (ref($zone_obj) =~ /zone/i) {
                    $self->_set_err("Invalid object returned for zone '($zone_name)'");
                    return $SanUtil::FALSE;
                }

                $zonemgr_obj->zone_create(-zone => $zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Unable to create zone '($zone_name)' on the fabric",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                &SanUtil::nate_log("-comment", "Zone '($zone_name)' created on the fabric");

                # Add the Zone to the Zoneset
                $zonemgr_obj->zoneset_add_zone(-zoneset => $zoneset, -zone => $zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Error while trying to add zone '($zone_name)' to " .
                                    "zoneset '$zoneset'", $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                } ## end if ($zonemgr_obj->error...
                &SanUtil::nate_log("-comment", "Zone '($zone_name)' added to zoneset '$zoneset'");
                # Initiator Object are for Clients
                $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$zone_name}->{initiators}->{$init_name}++;
                $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$zone_name}->{full_name} = $zone_name;
                # Targets Objects are for Vservers
                foreach my $target (@{$targets_ref}) {
                    my $target_ref = $resources->{$fabric}->{targets}->{$target};
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$zone_name}->{targets}->{$target_ref->{name}}++;
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$zone_name}->{initiators}->{$init_name}++;
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$zone_name}->{full_name} = $zone_name;
                }
                $self->{aliases_and_zones_created}->{totals_info}->{zones_created}++;
            } else {
                # single initiator / single target zoning
                foreach my $target (sort @{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) {
                    my $target_ref = $resources->{$fabric}->{targets}->{$target};
                    unless (defined($target_ref->{name})
                    and $init_ref->{name} ne ""
                    and defined($target_ref->{device})
                    and $target_ref->{device} ne ""
                    and defined($target_ref->{cluster_name})) {
                    $self->_set_err("Name or device missing from resource hash for wwpn '$target'");
                    return $SanUtil::FALSE;
                } ## end unless (defined($target_ref...

                # Build the alias name
                my $target_name = $self->_get_alias_name(
                    -host    => $target_ref->{name},
                    -device  => $target_ref->{device},
                    -cluster => $target_ref->{cluster_name},
                    -prefix  => $prefix);
                if ($self->error()) {
                    $self->_set_err("Unable to get the alias name for '$target_ref->{name}'");
                    return $SanUtil::FALSE;
                }

                # Build the zone name
                my ($zone_name, $md5_zone_name) = $self->_get_zone_name( -initiator => $init_name,
                                                                         -target    => $target_name,
                                                                         -prefix    => $prefix,
                                                                       );
                if ($self->error()) {
                    $self->_set_err("Unable to generate the zone name");
                    return $SanUtil::FALSE;
                }
                # Check to see if we have a zone object already for this zone.  If we do, then
                # it was found during discovery and thus we know it exists on the fabric already.
                my $zone_obj = $zonemgr_obj->get_zones(-zone => $md5_zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err(
                        "Error while trying to retrieve the object for zone '$md5_zone_name ($zone_name)'",
                        $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                } ## end if ($zonemgr_obj->error...

                if (ref($zone_obj) =~ /zone/i) {

                     # Verify the zone
                     $self->_verify_zone(
                             -zone              => $zone_name,
                             -md5               => $md5_zone_name,
                             -initiator         => $init_name,
                             -target            => $target_name,
                             -fabric            => $fabric,
                            );
                    if ($self->error()) {
                         $self->_set_err("Unable to verify zone '$md5_zone_name ($zone_name)'");
                         return $SanUtil::FALSE;
                    }
                    # Zone was found to be valid check to see if it is in the zoneset for this fabric.
                    my $zonesets_ref = $zonemgr_obj->zone_get_zonesets(-zone => $md5_zone_name, -type => "switch");
                    if ($zonemgr_obj->error()) {
                        $self->_set_err("Unable to get the list of zonesets for zone '$md5_zone_name ($zone_name)'",
                                        $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                    unless (ref($zonesets_ref) eq "HASH") {
                        $self->_set_err("Invalid data type returned from zone_get_zonesets");
                        return $SanUtil::FALSE;
                    }

                    # Check to see if $zoneset is found in the zonesets_ref hash.  If it isn't, add the zone
                    # to that zoneset.
                    if (defined($zonesets_ref->{$zoneset})) {
                        # Zone was found in the zoneset so just log a message and skip it.
                        &SanUtil::nate_log("-comment", "Zone '$md5_zone_name ($zone_name)' already exists on the " .
                                           "fabric in zoneset '$zoneset'");
                    } else {
                        # Add the Zone to the Zoneset
                        $zonemgr_obj->zoneset_add_zone(-zoneset => $zoneset, -zone => $md5_zone_name);
                        if ($zonemgr_obj->error()) {
                            $self->_set_err("Error while trying to add zone '$md5_zone_name ($zone_name)' to " .
                                            "zoneset '$zoneset'", $zonemgr_obj->errmsg());
                            return $SanUtil::FALSE;
                        } ## end if ($zonemgr_obj->error...
                        &SanUtil::nate_log("-comment", "Zone '$md5_zone_name ($zone_name)' added to" .
                                           "zoneset '$zoneset'");
                    }
                    # Initiator Object are for Clients
                    # Updating the count is to detect how many times this is used in the zone
                    # so that we are prepared for more complex many to one zone types.
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{targets}->{$target_name}++;
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{initiators}->{$init_name}++;
                    $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{full_name} = $zone_name;
                    # Targets Objects are for Vservers
                    # Updating the count is to detect how many times this is used in the zone
                    # so that we are prepared for more complex many to one zone types.
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{targets}->{$target_name}++;
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{initiators}->{$init_name}++;
                    $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{reused}->{$md5_zone_name}->{full_name} = $zone_name;
                    $self->{aliases_and_zones_created}->{zones}->{totals_info}->{zones_reused}++;
                    next;
                } ## end else [ if (ref($zone_obj) =~ ...

                # By reaching this point, we know the zone doesn't exist on the fabric.  Let's create it.
                # Grab the objects for the initiator and target aliases.
                my $init_obj = $zonemgr_obj->get_aliases(-alias => $init_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Error while trying to retrieve the object for alias '$init_name'",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (defined($init_obj) and ref($init_obj) =~ /alias/i) {
                    $self->_set_err("Invalid object data for alias '$init_name'");
                    return $SanUtil::FALSE;
                }

                my $target_obj = $zonemgr_obj->get_aliases(-alias => $target_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Error while trying to retrieve the object for alias '$target_name'",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (defined($target_obj) and ref($target_obj) =~ /alias/i) {
                    $self->_set_err("Invalid object data for alias '$target_name'");
                    return $SanUtil::FALSE;
                }

                # Add the target and initiator alias objects to an array
                my @aliases = ($init_obj, $target_obj);

                # Create the zone object
                $zone_obj = $zonemgr_obj->new_zone(
                    -zone    => $md5_zone_name,
                    -aliases => \@aliases
                );
                if ($zonemgr_obj->error) {
                    $self->_set_err("Unable to create object for zone '$md5_zone_name ($zone_name)'",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                unless (ref($zone_obj) =~ /zone/i) {
                    $self->_set_err("Invalid object returned for zone '$md5_zone_name ($zone_name)'");
                    return $SanUtil::FALSE;
                }

                $zonemgr_obj->zone_create(-zone => $md5_zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Unable to create zone '$md5_zone_name ($zone_name)' on the fabric",
                                    $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                &SanUtil::nate_log("-comment", "Zone '$md5_zone_name ($zone_name)' created on the fabric");

                # Initiator Object are for Clients
                $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{targets}->{$target_name}++;
                $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{initiators}->{$init_name}++;
                $self->{aliases_and_zones_created}->{zones}->{clients}->{$init_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{full_name} = $zone_name;
                # Targets Objects are for Vservers
                $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{targets}->{$target_name}++;
                $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{initiators}->{$init_name}++;
                $self->{aliases_and_zones_created}->{zones}->{vservers}->{$target_ref->{cluster_name}}->{$target_ref->{name}}->{$fabric}->{new}->{$md5_zone_name}->{full_name} = $zone_name;
                $self->{aliases_and_zones_created}->{totals_info}->{zones_created}++;

                # Add the Zone to the Zoneset
                $zonemgr_obj->zoneset_add_zone(-zoneset => $zoneset, -zone => $md5_zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Error while trying to add zone '$md5_zone_name ($zone_name)' to " .
                                    "zoneset '$zoneset'", $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                } ## end if ($zonemgr_obj->error...
                &SanUtil::nate_log("-comment", "Zone '$md5_zone_name ($zone_name)' added to zoneset '$zoneset'");

                # Get the cluster name to use in the insert
                my $cluster_name;
                if ($target_ref->{cluster_name} eq "") {
                    $cluster_name = "None";
                } else {
                    $cluster_name = $target_ref->{cluster_name};
                }
            } ## end foreach my $target (sort keys...
            }
        } ## end foreach my $initiator (sort...
    } ## end foreach my $fabric (keys %$resources)

    return $SanUtil::TRUE;
} ## end sub create_zones

###############################################################################
# Method:    delete_zones
# Objective: Delete all zones from the fabric for the resources provided
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_zones {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the prefix or the fabric
        my $prefix = $self->get_prefix(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Error retrieving prefix for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Iterate over the initators
        foreach my $initiator (sort keys %{$resources->{$fabric}->{initiators}}) {
            # Skip any initiator that doesn't have targets to unzone
            unless (defined($resources->{$fabric}->{initiators}->{$initiator}->{targets}) and
                    scalar(@{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) > 0) {
                # No targets so skip this initiator
                next;
            }

            my $init_ref = $resources->{$fabric}->{initiators}->{$initiator};
            unless (defined($init_ref->{name})
                and $init_ref->{name} ne ""
                and defined($init_ref->{device})
                and $init_ref->{device} ne ""
                and defined($init_ref->{cluster_name})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$initiator'");
                return $SanUtil::FALSE;
            } ## end unless (defined($init_ref->...

            # Build the alias name.
            my $init_name = $self->_get_alias_name(
                -host    => $init_ref->{name},
                -device  => $init_ref->{device},
                -cluster => $init_ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get the alias name for '$init_ref->{name}'");
                return $SanUtil::FALSE;
            }

            # Get the client name for the initiator
            my ($client, $undef1, $undef2) = $self->get_wwpn_name(-wwpn => $initiator);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the client name for '$initiator'");
                return $SanUtil::FALSE;
            }
            my $multi_target_zoning = $self->get_client_hash()->{$client}->{fcp}->{CONFIG_FCP_MULTI_TARGET_ZONING};
            if ($multi_target_zoning) {
                # MULTI TARGET - SINGLE INITIATOR

                my $init_zone_suffix = $init_name;
                $init_zone_suffix =~ s/^${prefix}_//;
                my $zone_name = $prefix . "_multi_" . $init_zone_suffix;
                # Get a list of zonesets that the zone belongs to
                my $zonesets_ref = $zonemgr_obj->zone_get_zonesets(-zone => $zone_name, -type => "switch");
                if ($zonemgr_obj->error()) {
                    $self->error("Unable to get the list of zonesets for zone '($zone_name)'",
                                 $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                # First remove the aliases/wwpn members of the targets related to this client
                # that exist in this multi-target zone.
                my $zone_obj = $zonemgr_obj->get_zones(-zone => $zone_name);
                if ($zonemgr_obj->error()) {
                    $self->error("Unable to get the zone object for zone '$zone_name'",
                                 $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                foreach my $target (@{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) {
                    my $removed = 0;
                    if (!$zone_obj->is_member(wwpn => $target)) {
                        if ($zone_obj->error()) {
                            $self->_set_err("Error while trying to determine if wwpn $target " .
                                            "is a member of zone '$zone_name'",
                                            $zone_obj->errmsg());
                            return $SanUtil::FALSE;
                        }
                    } else {
                        # remove this wwpn
                        $zone_obj->remove_wwpn(wwpn => $target);
                        if ($zone_obj->error()) {
                            $self->_set_err("Error while trying to remove wwpn $target from zone '$zone_name'",
                                            $zone_obj->errmsg());
                            return $SanUtil::FALSE;
                        }
                        $removed = 1;
                    }
                    if (!$removed) {
                        # $target is not a member of this zone in wwpn form, it might be in alias form
                        # find the aliases that this target belongs to
                        my $alias_ref = $zonemgr_obj->member_get_aliases(-member => $target);
                        if ($zonemgr_obj->error) {
                            $zonemgr_obj->_clear_err();
                            # there is no alias for this target apparently, just skip this target
                            next;
                        }

                        # check if any of these aliases are members of this zone
                        # and remove them from the zone if they are
                        foreach my $alias_name (keys %$alias_ref) {
                            my $alias_obj = $alias_ref->{$alias_name};
                            if (!$zone_obj->is_member(alias => $alias_name)) {
                                if ($zone_obj->error()) {
                                    $self->_set_err("Error while trying to determine if alias $alias_name " .
                                                    "is a member of zone '$zone_name'",
                                                    $zone_obj->errmsg());
                                    return $SanUtil::FALSE;
                                }
                            } else {
                                # remove this alias from the zone
                                $zone_obj->remove_alias(alias => $alias_obj);
                                if ($zone_obj->error()) {
                                    $self->_set_err("Error while trying remove alias $alias_name from zone '$zone_name'",
                                                    $zone_obj->errmsg());
                                    return $SanUtil::FALSE;
                                }
                            }
                        }
                    }
                }

                # if after all that is done, if the multi target zone only has one member left in it (the initiator)
                # then remove the zone from zonesets and delete the zone.
                # the initiator is always going to be represented as an alias in the zone
                # the targets can either be wwpn members or alias members
                my $aliases = $zone_obj->get_aliases();
                my $wwpns = $zone_obj->get_wwpns();
                if (scalar(@{$wwpns}) or scalar(keys %{$aliases}) > 1) {
                    &SanUtil::nate_log("-comment", "Not deleting Zone '($zone_name)', it still has targets in it " .
                                       "that are not a part of this test.");
                    next;
                }

                # Iterate over the zonesets.
                foreach my $zoneset (keys %$zonesets_ref) {

                    # Remove the zone from the zoneset
                    $zonemgr_obj->zoneset_remove_zone(
                        -zone    => $zone_name,
                        -zoneset => $zoneset
                    );
                    if ($zonemgr_obj->error()) {
                        $self->_set_err("Unable to remove zone '$zone_name' from zoneset '$zoneset'",
                                        $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                    &SanUtil::nate_log("-comment",
                        "Zone '($zone_name)' removed from " . "zoneset '$zoneset'");
                } ## end foreach my $zoneset (keys %$zonesets_ref)

                # Now that we have removed the zone from the zonesets, go ahead and remove the zone
                $zonemgr_obj->zone_destroy(-zone => $zone_name);
                if ($zonemgr_obj->error()) {
                    $self->_set_err("Unable to remove zone '($zone_name)'", $zonemgr_obj->errmsg());
                    return $SanUtil::FALSE;
                }
                &SanUtil::nate_log("-comment", "Zone '($zone_name)' removed from the fabric");
                next;
            } else {
                # SINGLE TARGET - SINGLE INITIATOR

                # Iterate over the targets defined on the fabric
                foreach my $target (sort keys %{$resources->{$fabric}->{targets}}) {
                    my $target_ref = $resources->{$fabric}->{targets}->{$target};
                    unless (defined($target_ref->{name})
                            and $init_ref->{name} ne ""
                            and defined($target_ref->{device})
                            and $target_ref->{device} ne ""
                            and defined($target_ref->{cluster_name})) {
                        $self->_set_err("Name or device missing from resource hash for wwpn '$target'");
                        return $SanUtil::FALSE;
                    } ## end unless (defined($target_ref...

                    # Build the alias name
                    my $target_name = $self->_get_alias_name(
                        -host    => $target_ref->{name},
                        -device  => $target_ref->{device},
                        -cluster => $target_ref->{cluster_name},
                        -prefix  => $prefix,
                    );
                    if ($self->error()) {
                        $self->_set_err("Unable to get the alias name for '$target_ref->{name}'");
                        return $SanUtil::FALSE;
                    }

                    # Build the zone name
                    my ($zone_name, $md5_zone_name) = $self->_get_zone_name(
                        -initiator => $init_name,
                        -target    => $target_name,
                        -prefix    => $prefix,
                    );
                    if ($self->error()) {
                        $self->_set_err("Unable to generate the zone name");
                        return $SanUtil::FALSE;
                    }

                    # Verify the zone
                    $self->_verify_zone(
                        -zone      => $zone_name,
                        -md5       => $md5_zone_name,
                        -initiator => $init_name,
                        -target    => $target_name,
                        -fabric    => $fabric,
                    );

                    # Check to see if we got an error.  If we did, just skip to the next one.
                    # We don't want to attempt to delete any zones that don't verify.
                    if ($self->error()) {
                        $self->_clear_err();
                        next;
                    }

                    # Get a list of zonesets that the zone belongs to
                    my $zonesets_ref = $zonemgr_obj->zone_get_zonesets(-zone => $md5_zone_name, -type => "switch");
                    if ($zonemgr_obj->error()) {
                        $self->error("Unable to get the list of zonesets for zone '$md5_zone_name ($zone_name)'",
                                     $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }

                    # Iterate over the zonesets.
                    foreach my $zoneset (keys %$zonesets_ref) {

                        # Remove the zone from the zoneset
                        $zonemgr_obj->zoneset_remove_zone(
                            -zone    => $md5_zone_name,
                            -zoneset => $zoneset
                        );
                        if ($zonemgr_obj->error()) {
                            $self->_set_err("Unable to remove zone '$md5_zone_name ($zone_name)' from zoneset '$zoneset'",
                                            $zonemgr_obj->errmsg());
                            return $SanUtil::FALSE;
                        }
                        &SanUtil::nate_log("-comment",
                                           "Zone '$md5_zone_name ($zone_name)' removed from " . "zoneset '$zoneset'");
                    } ## end foreach my $zoneset (keys %$zonesets_ref)

                    # Now that we have removed the zone from the zonesets, go ahead and remove the zone
                    $zonemgr_obj->zone_destroy(-zone => $md5_zone_name);
                    if ($zonemgr_obj->error()) {
                        $self->_set_err("Unable to remove zone '$md5_zone_name ($zone_name)'", $zonemgr_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                    &SanUtil::nate_log("-comment", "Zone '$md5_zone_name ($zone_name)' removed from the fabric");
                } # foreach my $target (sort keys %{$resources->{targets}})
            } # end if (multi) {multi zoning} else {single zoning}
        }    # foreach my $initiator (sort keys %{$resources->{initiators}})
    } ## end foreach my $fabric (keys %$resources)

    return $SanUtil::TRUE;
} ## end sub delete_zones

###############################################################################
# Method:    delete_aliases
# Objective: Delete all of the aliases from the fabric for the resources specified
#            ONLY if they are not in use by other zones not listed on the object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_aliases {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the prefix or the fabric
        my $prefix = $self->get_prefix(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Error retrieving prefix for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Build a hash of initiator and target hashes
        my $devices = {};
        foreach my $init (sort keys %{$resources->{$fabric}->{initiators}}) {
            $devices->{$init} = $resources->{$fabric}->{initiators}->{$init};
        }
        foreach my $target (sort keys %{$resources->{$fabric}->{targets}}) {
            $devices->{$target} = $resources->{$fabric}->{targets}->{$target};
        }

        # Iterate over the initators and targets and create new alias objects
        foreach my $adapter (sort keys %$devices) {
            my $ref = $devices->{$adapter};
            unless (defined($ref->{name})
                and $ref->{name} ne ""
                and defined($ref->{device})
                and $ref->{device} ne ""
                and defined($ref->{device})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$adapter'");
                return $SanUtil::FALSE;
            } ## end unless (defined($ref->{name...
            my $alias_name = $self->_get_alias_name(
                -host    => $ref->{name},
                -device  => $ref->{device},
                -cluster => $ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get alias name");
                return $SanUtil::FALSE;
            }

            # Verify the alias
            $self->_verify_alias(
                -alias  => $alias_name,
                -member => $adapter,
                -fabric => $fabric
            );

            # Check to see if we got an error.  If we did, just skip to the next one.
            # We don't want to attempt to delete any aliases that don't verify.  We will log
            # a message about the alias being skipped.
            if ($self->error()) {
                $self->sanlog(-msg => "Alias '$alias_name' was not found in fabric '$fabric'.  Skipping delete.");
                $self->_clear_err();
                next;
            }

            # Check to see if it belongs to any zones.  If it is, just print out a message
            # and then skip move on.  It is possible that a different config is still using
            # that alias.
            my $zones_ref = $zonemgr_obj->alias_get_zones(-alias => $alias_name, -type => "switch");
            if ($zonemgr_obj->error()) {
                $self->_set_err("Unable to get a list of zones for alias '$alias_name'", $zonemgr_obj->errmsg());
                return $SanUtil::FALSE;
            }

            if (scalar(keys %$zones_ref) > 0) {

                # Looks like the alias is still in at least one zone. Print out a message about
                # it and move on to the next one.
                my @zones = ();
                foreach my $zone (sort keys %$zones_ref) {
                    push(@zones, $zone);
                }
                my $errmsg = join("\n", @zones);
                $self->sanlog(-msg => "Unable to remove alias '$alias_name' since it is still a member of " .
                                      "the following zones:\n${errmsg}");
                next;
            } ## end if (scalar(keys %$zones_ref...

            # At this point we know the alias doesn't belong to any zones so it should be
            # safe to remove.
            $zonemgr_obj->alias_destroy(-alias => $alias_name);
            if ($zonemgr_obj->error()) {
                $self->_set_err("Unale to remove alias '$alias_name'", $zonemgr_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end foreach my $adapter (sort keys...
    } ## end foreach my $fabric (keys %$resources)
    return $SanUtil::TRUE;
} ## end sub delete_aliases

###############################################################################
# Method:    activate_fabric_changes
# Objective: Save and activate the changes to the fabric
# Details:   See POD documentation at the end of this file.
###############################################################################
sub activate_fabric_changes {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the active zoneset
        my $zoneset = $self->get_active_zoneset(-fabric => $fabric);
        unless (defined($zoneset) and $zoneset ne "") {
            $self->_set_err("Unable to retrieve the name of the active zoneset");
            return $SanUtil::FALSE;
        }

        # Save the config
        $zonemgr_obj->config_save();
        if ($zonemgr_obj->error()) {
            $self->_set_err("Error while trying to save fabric config", $zonemgr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Sleep 60 seconds to allow the fabric commands to push around all of the switches
        $self->sanlog(-msg => "Sleeping 60 seconds to allow fabric refresh to complete");
        sleep(60);

        # Enable the config
        $zonemgr_obj->config_enable(-zoneset => $zoneset);
        if ($zonemgr_obj->error) {
            $self->_set_err("Error while trying to enable zoneset '$zoneset' on fabric", $zonemgr_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end foreach my $fabric (keys %$resources)

    return $SanUtil::TRUE;
} ## end sub activate_fabric_changes

###############################################################################
# Method:    verify_fabric
# Objective: Verify the resources provided exist on the fabric
# Details:   See POD documentation at the end of this file.
###############################################################################
sub verify_fabric {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Initialize some variables
    my $resources = $self->get_resource_hash();
    if ($self->error()) {
        $self->_set_err("Error retrieving object attributes");
        return $SanUtil::FALSE;
    }

    # We will store any non fatal errors from the verification in @errors instead of relying on errmsg on the
    # object.  The idea is to capture all of the fabric configuration issues and then  _set_err to those
    # at the end.
    my @errors = ();

    foreach my $fabric (keys %$resources) {

        # Get the zonemgr object for the fabric
        my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Get the prefix or the fabric
        my $prefix = $self->get_prefix(-fabric => $fabric);
        if ($self->error()) {
            $self->_set_err("Error retrieving prefix for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Build a hash of initiator and target hashes
        my $devices = {};
        foreach my $init (sort keys %{$resources->{$fabric}->{initiators}}) {
            if (defined($resources->{$fabric}->{initiators}->{$init}->{targets}) and
                scalar(@{$resources->{$fabric}->{initiators}->{$init}->{targets}}) > 0) {
                $devices->{$init} = $resources->{$fabric}->{initiators}->{$init};
            }
        }
        foreach my $target (sort keys %{$resources->{$fabric}->{targets}}) {
            if (defined($resources->{$fabric}->{targets}->{$target}->{initiators}) and
                scalar(@{$resources->{$fabric}->{targets}->{$target}->{initiators}}) > 0) {
                $devices->{$target} = $resources->{$fabric}->{targets}->{$target};
            }
        }

        # Get the active zoneset
        my $zoneset = $self->get_active_zoneset(-fabric => $fabric);
        unless (defined($zoneset) and $zoneset ne "") {
            $self->_set_err("Unable to retrieve the name of the active zoneset for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # Iterate over the initators and targets validate each alias
        foreach my $adapter (sort keys %$devices) {
            my $ref = $devices->{$adapter};
            unless (defined($ref->{name})
                and $ref->{name} ne ""
                and defined($ref->{device})
                and $ref->{device} ne ""
                and defined($ref->{device})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$adapter'");
                return $SanUtil::FALSE;
            } ## end unless (defined($ref->{name...

            # Generate the alias name
            my $alias_name = $self->_get_alias_name(
                -host    => $ref->{name},
                -device  => $ref->{device},
                -cluster => $ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get alias name");
                return $SanUtil::FALSE;
            }

            # Verify the alias
            $self->_verify_alias(
                -alias  => $alias_name,
                -member => $adapter,
                -fabric => $fabric
            );
            if ($self->error()) {
                push(@errors, "Unable to verify alias '$alias_name' on fabric '$fabric'");
            }
        }    # foreach my $adapter (sort keys %$devices) {

        # Now verify the zones
        # Iterate over the initators
        foreach my $initiator (sort keys %{$resources->{$fabric}->{initiators}}) {
            # Skip any initiator that doesn't have targets to zone
            unless (defined($resources->{$fabric}->{initiators}->{$initiator}->{targets}) and
                    scalar(@{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) > 0) {
                # No targets so skip this initiator
                next;
            }

            # Get the client name for the initiator
            my ($client, $undef1, $undef2) = $self->get_wwpn_name(-wwpn => $initiator);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the client name for '$initiator'");
                return $SanUtil::FALSE;
            }

            my $init_ref = $resources->{$fabric}->{initiators}->{$initiator};
            unless (defined($init_ref->{name})
                and $init_ref->{name} ne ""
                and defined($init_ref->{device})
                and $init_ref->{device} ne ""
                and defined($init_ref->{device})) {
                $self->_set_err("Name or device missing from resource hash for wwpn '$initiator' on fabric '$fabric'");
                return $SanUtil::FALSE;
            } ## end unless (defined($init_ref->...

            # Build the alias name.
            my $init_name = $self->_get_alias_name(
                -host    => $init_ref->{name},
                -device  => $init_ref->{device},
                -cluster => $init_ref->{cluster_name},
                -prefix  => $prefix,
            );
            if ($self->error()) {
                $self->_set_err("Unable to get the alias name for '$init_ref->{name}'");
                return $SanUtil::FALSE;
            }

            # Iterate over the targets defined on the initiator
            foreach my $target (sort @{$resources->{$fabric}->{initiators}->{$initiator}->{targets}}) {
                my $target_ref = $resources->{$fabric}->{targets}->{$target};
                unless (defined($target_ref->{name})
                    and $init_ref->{name} ne ""
                    and defined($target_ref->{device})
                    and $target_ref->{device} ne ""
                    and defined($target_ref->{device})) {
                    $self->_set_err("Name or device missing from resource hash for wwpn '$target' on fabric '$fabric'");
                    return $SanUtil::FALSE;
                } ## end unless (defined($target_ref...

                # By making it this far we know one or more of the following has occurred.
                # 1. The target is on a 7mode object.
                # 2. The target is a controller cluster object and it was found in the vserver map for this client.
                # 3. The target is a controller cluster object and the client has nothing in the vserver
                #    map hash which means we will map that client to all controllers.

                # Build the alias name
                my $target_name = $self->_get_alias_name(
                    -host    => $target_ref->{name},
                    -device  => $target_ref->{device},
                    -cluster => $target_ref->{cluster_name},
                    -prefix  => $prefix,
                );
                if ($self->error()) {
                    $self->_set_err("Unable to get the alias name for '$target_ref->{name}'");
                    return $SanUtil::FALSE;
                }

                # Build the zone name
                my ($zone_name, $md5_zone_name) = $self->_get_zone_name(
                    -initiator => $init_name,
                    -target    => $target_name,
                    -prefix    => $prefix,
                );
                if ($self->error()) {
                    $self->_set_err("Unable to generate the zone name");
                    return $SanUtil::FALSE;
                }

                # Verify the zone
                $self->_verify_zone(
                    -zone      => $zone_name,
                    -md5       => $md5_zone_name,
                    -initiator => $init_name,
                    -target    => $target_name,
                    -fabric    => $fabric,
                );
                if ($self->error()) {
                    push(@errors, "Unable to verify zone '$md5_zone_name ($zone_name)' on fabric '$fabric'");
                }

                # Get a list of zonesets that the zone belongs to
                my $zonesets_ref = $zonemgr_obj->zone_get_zonesets(-zone => $md5_zone_name, -type => "switch");
                if ($zonemgr_obj->error) {
                    push(@errors, "Unable to retrieve zoneset infomation for zone '$md5_zone_name ($zone_name)'");
                }

                # Make sure that the zoneset passed in on the fabric spec is one of the zonesets that have the zone
                unless (scalar(keys %{$zonesets_ref->{$zoneset}}) > 0) {
                    push(@errors, "Zone '$md5_zone_name ($zone_name)' does not exist in zoneset '$zoneset'");
                }
            }    # foreach my $target (sort keys %{$resources->{targets}}) {
        }    # foreach my $initiator (sort keys %{$resources->{initiators}}) {
    } ## end foreach my $fabric (keys %$resources)

    # We didn't return after every error above since we want to return one giant list of issues with the fabric.
    # Test if we set any error and return appropriately.
    if (scalar(@errors) > 0) {
        $self->_set_err(@errors);
        return $SanUtil::FALSE;
    } else {
        return $SanUtil::TRUE;
    }
} ## end sub verify_fabric

###############################################################################
# Method:    build_switch_info_hash()
# Objective: Gather and Populate the switch info hash
# Details:   See POD documentation at the end of this file.
###############################################################################
sub build_switch_info_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Set deleted_wwpns_info to and empty hash since
    # it will be created later in this method.
    $self->{deleted_wwpns_info} = {};

    # Get the switch hash
    my $switch_hash = $self->get_switch_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the switch hash");
        return $SanUtil::FALSE;
    }

    # create a portmgr object on each sanswitch.
    foreach my $switch (keys %$switch_hash) {
        my $switch_obj = $switch_hash->{$switch};

        # create a portmgr object and force the nameservers to
        # get loaded into memory
        $switch_obj->create_portmgr(-load_nameserver => 1);
        if ($switch_obj->error()) {
            $self->_set_err($switch_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # enable command caching on the switch object
        # so nameserver lookups are faster during the gather
        # phase in this method, we will revert this setting on all
        # the switch objects at the very end of this function.
        $switch_obj->enable_sancache();
    } ## end foreach my $switch (keys %$switch_hash)

    # We need to create 2 arrays based on switch type.  One will be all apcon switches,
    # and the other will be all fc and ethernet switches
    my @fc_mac_switches;
    my @apcon_switches;
    foreach my $switch (keys %$switch_hash) {
        my $switch_obj = $switch_hash->{$switch};

        # Get the hosttype
        my $hosttype = $switch_obj->hosttype();

        if ($hosttype =~ /apcon/i) {
            push (@apcon_switches, $switch);
        } else {
            push (@fc_mac_switches, $switch);
        }
    }

    # this will be our internally maintained datastructure for all switch events
    my $switch_info = {};

    # switchinfo maintains a mapping between resource_types, their respective wwpns/macs
    # and switch port information.
    # this gather task will create a mapping between clients & targets (with wwpns & macs) to switches

    # get all sanclients
    my $client_hash = $self->get_client_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the clients hash");
        return $SanUtil::FALSE;
    }

    # get all cmode controller cluster objects
    my $cmode_cluster_hash = $self->get_cmode_cluster_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the cmode clusters hash");
        return $SanUtil::FALSE;
    }

    # get all 7mode sanfilers
    my $sanfiler7m_hash = $self->get_sanfiler7m_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the sanfiler7m hash");
        return $SanUtil::FALSE;
    }

    # get all Cmode sanfilers
    my $nacl_node_cm_hash = {};
    foreach my $cluster (keys %$cmode_cluster_hash) {

        my @nacl_node_objs;
        # Get the objects from the cluster
        try {
            @nacl_node_objs = $cmode_cluster_hash->{$cluster}->node_objs();
        } otherwise {
            my $e = shift;
            $self->_set_err("Unable to get NACL::C::Node objects", $e->text());
        };
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # Now get the name of the controller in the object
        foreach my $nacl_node_obj (@nacl_node_objs) {
            my $node_name = $nacl_node_obj->node();
            if ($self->error()) {
                $self->_set_err("Failed to get name of CMode node");
                return $SanUtil::FALSE;
            }

            # Add the node and object to the sanfiler_cm_hash hash
            $nacl_node_cm_hash->{$node_name}->{cluster} = $cluster;
            $nacl_node_cm_hash->{$node_name}->{object}  = $nacl_node_obj;
        } ## end foreach my $sanfiler_cm_obj...
    } ## end foreach my $cluster (keys %$cmode_cluster_hash)


    # get the wwpns & macs for each sanclient and store in switch info
    foreach my $client_name (keys %$client_hash) {
        # get the sanclient object for this client
        my $sc_obj = $client_hash->{$client_name};

        my $init_hash_ref = $self->_get_client_initiators(-client => $client_name);
        if ($self->error()) {
            $self->_set_err("Failed to get client initiators from xml");
            return $SanUtil::FALSE;
        }

        # handle FCP if this client is doing FCP
        if ($sc_obj->get_protocol() =~ /fcp/i and grep(/fcp/i, $self->get_protocol_list())) {
            my @init_objs = $sc_obj->get_initiators();
            if ($sc_obj->error()) {
                $self->_set_err($sc_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # extract the WWPN from each initiator object and store it in $switch_info
            foreach my $init_obj (@init_objs) {
                my $device_name = $init_obj->get_name();
                my $protocol    = $init_obj->get_protocol();
                my $wwpn        = $init_obj->get_portname();

                if (lc($protocol) ne "fcp") {
                    # Skip this initiator if it isn't fcp
                    next;
                }

                # Format the wwpn to be lowercase and ensure that it includes : as it
                # comes in different formats with and without the :.
                $wwpn = lc($wwpn);
                $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);
                # Make a wwpn without :. We need both formats to check the init_hash_ref because
                # we dont know what format the hash wwpns will be.
                my $stripped_wwpn = $wwpn;
                $stripped_wwpn =~ s/\://g;

                # Check to see if the wwpn is in the list of iniators to use.  If $init_hash_ref is
                # completely empty, then they want to use everything
                unless (scalar(keys %$init_hash_ref) == 0 or defined($init_hash_ref->{$wwpn}) or defined($init_hash_ref->{$stripped_wwpn})) {

                    # Looks like they don't want everything, but the wwpn we found wasn't in the list
                    # to use so skip it.
                    $self->{deleted_wwpns_info}->{sanclients}->{$wwpn}->{resource_name} = $client_name;
                    $self->{deleted_wwpns_info}->{sanclients}->{$wwpn}->{resource_type} = 'sanclients';
                    $self->{deleted_wwpns_info}->{sanclients}->{$wwpn}->{sanclients} = $client_name;
                    $self->{deleted_wwpns_info}->{sanclients}->{$wwpn}->{device_name} = $device_name;
                    $self->{deleted_wwpns_info}->{sanclients}->{$wwpn}->{error} .=
                        "\nWWPN $wwpn was not included in the list of initiators to use for client $client_name.";
                    next;
                } ## end unless (scalar(keys %$init_hash_ref...

                # Add the wwpn to the switch_info hash for later comparison with what is on the switch
                $switch_info->{sanclients}->{$client_name}->{fc_initiator_wwpn}->{$wwpn}->{device_name} = $device_name;
                $switch_info->{sanclients}->{$client_name}->{apcon_initiator_wwpn}->{$wwpn}->{device_name} = $device_name;
            } ## end foreach my $init_obj (@init_objs)
        } ## end if ($sc_obj->get_protocol...

        # handle iSCSI if this client is doing iSCSI
        if ($sc_obj->get_protocol() =~ /iscsi/i and grep(/iscsi/i, $self->get_protocol_list())) {

            # get mac address from ethernet ports if this client is doing iSCSI
            # only get iscsi capable interfaces
            my $ip_info = $sc_obj->get_ip_info(-iscsi_capable => 1);
            if ($sc_obj->error()) {
                $self->_set_err($sc_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # extract the mac addresses
            foreach my $interface (keys %{$ip_info}) {
                my $mac_address = $ip_info->{$interface}->{mac};
                if (defined($mac_address) and $mac_address ne "") {

                    # if this is the management inferface, then skip
                    if ($ip_info->{$interface}->{is_mgmt_interface} == 1) {
                        next;
                    }
                    # store this mac address in switch_info
                    $switch_info->{sanclients}->{$client_name}->{iscsi_initiator_mac}->{$mac_address} = {};
                } ## end if (defined($mac_address...
            } ## end foreach my $interface (keys...
        } ## end if ($sc_obj->get_protocol...
    } ## end foreach my $client_name (keys...

    # get the wwpns and macs for all 7mode controllers
    foreach my $sanfiler7m (keys %$sanfiler7m_hash) {
        my $filer_obj = $sanfiler7m_hash->{$sanfiler7m};

        if (grep(/fcp/i, $self->get_protocol_list())) {
            # get the fcp target objs on this controller
            my @fcp_targetobjs = $filer_obj->get_targets(-dmp => 1);
            foreach my $target_obj (@fcp_targetobjs) {

                # extract the wwpn name for this fcp target
                my $wwpn        = $target_obj->get_portname();
                my $device_name = $target_obj->get_name();

                if ($target_obj->get_status() eq "LINK NOT CONNECTED") {

                    # skip this target if its not connected
                    next;
                }

                # Lowercase the wwpn and add colons
                $wwpn = lc($wwpn);
                $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);

                # store this target wwpn in the switch info hash
                $switch_info->{sanfilers7m}->{$sanfiler7m}->{fc_target_wwpn}->{$wwpn}->{device_name} = $device_name;
                $switch_info->{sanfilers7m}->{$sanfiler7m}->{apcon_target_wwpn}->{$wwpn}->{device_name} = $device_name;
            } ## end foreach my $target_obj (@fcp_targetobjs)
        }

        if (grep(/iscsi/i, $self->get_protocol_list())) {
            # get all the iscsi target objects that are iSCSI enabled, and
            # have valid IP addresses
            my @iscsi_targetobjs = $filer_obj->get_iscsi_targets(
                -state    => "enabled",
                -valid_ip => 1
            );
            if ($filer_obj->error()) {
                $self->_set_err($filer_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # store each iscsi target mac address in the switch info hash
            foreach my $target_obj (@iscsi_targetobjs) {
                my $mac         = $target_obj->get_mac();
                my $device_name = $target_obj->get_name();
                $switch_info->{sanfilers7m}->{$sanfiler7m}->{iscsi_target_mac}->{$mac} = {};
            }
        }
    } ## end foreach my $sanfiler7m (keys...

    if (grep(/fcp/i, $self->get_protocol_list())) {
        # get the wwpns for all fcp LIFs on all CMode clusters
        foreach my $cluster_name (keys %$cmode_cluster_hash) {

            # get the sanfiler object for this filer
            my $cluster_obj = $cmode_cluster_hash->{$cluster_name};

            # get all the vservers on this cluster
            my @vserver_objs;
            try {
                @vserver_objs = NACL::CS::Vserver->fetch(command_interface => $cluster_obj,
                                                         filter => {type => 'data|cluster'});
            } catch NACL::Exceptions::NoElementsFound with {
                @vserver_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("Unable to fetch NACL Vserver Component States", $e->text());
            };
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # Check to see if the caller passed in a list of vservers
            # that we should limit the operations to.
            my $cmode_vservers = $self->get_cmode_vserver_list();
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # get all the fcp wwpns for the lifs on each vserver
            foreach my $vserver_obj (@vserver_objs) {
                my $vserver_name = $vserver_obj->vserver();

                if (scalar(@{$cmode_vservers}) > 0) {
                    # Looks like we received a list of vservers
                    # so see if the current vserver is in that list.
                    if (grep(/^$vserver_name$/, @{$cmode_vservers}) == 0) {
                        # This vserver doesn't exist.  Log a message and
                        # skip it.
                        $self->sanlog(-msg => "Skipping vserver '$vserver_name' since it wasn't " .
                                          "included in the list of vservers passed into FabricMgr",
                                      -msg_level => 3);
                        next;
                    }
                }

                # fetch all fcp lif objs for this vserver
                my @lif_objs;
                try {
                    @lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                  filter => {'data-protocol' => ["fcp"],
                                                                             'vserver' => $vserver_name});
                } catch NACL::Exceptions::NoElementsFound with {
                    @lif_objs = ();
                } otherwise {
                    my $e = shift;
                    $self->_set_err("Failed to obtain lif objects", $e->text());
                };
                if ($self->error()) {
                    return $SanUtil::FALSE;
                }

                # store all the wwpn's of each lif in the switch info hash
                foreach my $lif_obj (@lif_objs) {
                    my $lif_name = $lif_obj->lif();
                    my $wwpn = $lif_obj->wwpn();
                    my $curr_node = $lif_obj->curr_node();

                    # store the wwpns for this vserver in the switch info hash
                    # Lowercase the wwpn and add colons
                    $wwpn = lc($wwpn);
                    $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);

                    $switch_info->{vservers}->{$vserver_name}->{fc_target_lif_wwpn}->{$wwpn}->{cluster_name}
                        = $cluster_name;
                    $switch_info->{vservers}->{$vserver_name}->{apcon_target_lif_wwpn}->{$wwpn}->{cluster_name}
                        = $cluster_name;
                    $switch_info->{vservers}->{$vserver_name}->{fc_target_lif_wwpn}->{$wwpn}->{curr_node_name}
                        = $curr_node;
                    $switch_info->{vservers}->{$vserver_name}->{apcon_target_lif_wwpn}->{$wwpn}->{curr_node_name}
                        = $curr_node;
                    $switch_info->{vservers}->{$vserver_name}->{fc_target_lif_wwpn}->{$wwpn}->{lif_name} = $lif_name;
                    $switch_info->{vservers}->{$vserver_name}->{apcon_target_lif_wwpn}->{$wwpn}->{lif_name} = $lif_name;

                    # store pointers to this switch info entry in the controller clusters table
                    # as well as the specific node that this lif is connected to.
                    $switch_info->{nacl_clusters_cm}->{$cluster_name}->{fc_target_lif_wwpn}->{$wwpn}
                        = $switch_info->{vservers}->{$vserver_name}->{fc_target_lif_wwpn}->{$wwpn};
                    $switch_info->{nacl_clusters_cm}->{$cluster_name}->{apcon_target_lif_wwpn}->{$wwpn}
                        = $switch_info->{vservers}->{$vserver_name}->{apcon_target_lif_wwpn}->{$wwpn};
                    $switch_info->{nacl_nodes_cm}->{$curr_node}->{fc_target_lif_wwpn}->{$wwpn}
                        = $switch_info->{vservers}->{$vserver_name}->{fc_target_lif_wwpn}->{$wwpn};
                    $switch_info->{nacl_nodes_cm}->{$curr_node}->{apcon_target_lif_wwpn}->{$wwpn}
                        = $switch_info->{vservers}->{$vserver_name}->{apcon_target_lif_wwpn}->{$wwpn};
                } ## end foreach my $lif_obj (@lif_objs)
            } ## end foreach my $vserver_obj (@vserver_objs)
        } ## end foreach my $cluster_name (keys...
    }

    # keep a cache of all component state objs keyed by cluster
    my $cs_cache = {};

    # obtain the mac addresses of the cluster ports on all cmode controllers
    foreach my $controller_name (keys %$nacl_node_cm_hash) {

        # Get the cluster object that this node lives in
        my $cluster_name = $nacl_node_cm_hash->{$controller_name}->{cluster};
        my $cluster_obj  = $cmode_cluster_hash->{$cluster_name};

        if (grep(/cluster/i, $self->get_protocol_list())) {
            # corrolate all the cluster port mac addresses that contain active cluster lifs
            my @cluster_port_objs;
            try {
                @cluster_port_objs = NACL::CS::NetworkPort->fetch(command_interface => $cluster_obj,
                                                                  filter => {'node' => $controller_name,
                                                                             'role' => "cluster"});
            } catch NACL::Exceptions::NoElementsFound with {
                @cluster_port_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("Failed to obtain cluster NetworkPort objects", $e->text());
            };
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # get all cluster lif objs
            my @cluster_lif_objs;
            try {
                @cluster_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                      filter => {'role' => "cluster",
                                                                                 'server' => $controller_name});
            } catch NACL::Exceptions::NoElementsFound with {
                @cluster_lif_objs = ();
            } otherwise {
                my $e = shift;
                $self->_set_err("Failed to obtain cluster lif objects", $e->text());
            };
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # add an entry for all cluster ports that have active cluster lifs
            foreach my $cluster_port (@cluster_port_objs) {
                foreach my $cluster_lif (@cluster_lif_objs) {
                    if ($cluster_lif->home_port() eq $cluster_port->port()) {

                        # store the mac address for this cluster port in the hash
                        unless (exists($switch_info->{nacl_nodes_cm}->{$controller_name}->{cluster_port_mac}->{$cluster_port->mac()})) {
                            $switch_info->{nacl_nodes_cm}->{$controller_name}->{cluster_port_mac}->{$cluster_port->mac()} = {};
                        }

                        # store a copy in the cluster lookup table
                        if ($cluster_name ne "") {
                            $switch_info->{nacl_clusters_cm}->{$cluster_name}->{cluster_port_mac}->{$cluster_port->mac()}
                                = $switch_info->{nacl_nodes_cm}->{$controller_name}->{cluster_port_mac}->{$cluster_port->mac()};

                            # keep track of the cluster name
                            $switch_info->{nacl_nodes_cm}->{$controller_name}->{cluster_port_mac}->{$cluster_port->mac()}
                                ->{cluster_name} = $cluster_name;
                        }

                        # "wake up" the ethernet switch's routing table by pinging the broadcast ip
                        # via this lif, this will ensure that when we query the switch for the port locations,
                        # that the mac addresses will be listed.
                        my $broadcast_ip = undef;
                        if ($cluster_lif->address() !~ /:/i) {
                            $broadcast_ip = &SanUtil::get_ip_broadcast(-ip_address   => $cluster_lif->address(),
                                                                       -ipv4_netmask => $cluster_lif->netmask());
                        }

                        my $exception_text = undef;
                        if (defined($broadcast_ip)) {
                            # try to ping the broadcast cluster ip to wake everyone up
                            # ignore any error
                            try {
                                NACL::C::Network->ping(command_interface => $cluster_obj,
                                                       'lif'             => $cluster_lif->lif(),
                                                       'lif-owner'       => $cluster_lif->server(),
                                                       'destination'     => $broadcast_ip,
                                                       'count'           => 1);
                            } otherwise {
                                my $e = shift;
                                $exception_text = "Ping failed because: " . $e->text();
                            };
                            if (defined($exception_text)) {
                                $self->sanlog(-msg => $exception_text);
                            }
                        }
                    }
                } ## end foreach my $cluster_lif (@cluster_lif_objs)
            } ## end foreach my $cluster_port (@cluster_port_objs)
        }

        if (grep(/iscsi|cifs|nfs/i, $self->get_protocol_list())) {
            # get every possible combination of iscsi,nfs,cifs capable lifs
            my @iscsi_lif_objs;

            if (grep(/iscsi/i, $self->get_protocol_list())) {
                # check if we have these cached
                if (exists($cs_cache->{$cluster_name}->{net_int_iscsi})) {
                    @iscsi_lif_objs = @{$cs_cache->{$cluster_name}->{net_int_iscsi}};
                } else {
                    try {
                        @iscsi_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                            filter => {'role' => "data",
                                                                                       'data-protocol' => ["iscsi"]});
                    } catch NACL::Exceptions::NoElementsFound with {
                        @iscsi_lif_objs = ();
                    } otherwise {
                        my $e = shift;
                        $self->_set_err("Failed to obtain iSCSI lif objects", $e->text());
                    };
                    if ($self->error()) {
                        return $SanUtil::FALSE;
                    }
                    # store in the cache for this cluster
                    $cs_cache->{$cluster_name}->{net_int_iscsi} = \@iscsi_lif_objs;
                }
            }

            my @nfs_cifs_lif_objs;
            if (grep(/nfs|cifs/i, $self->get_protocol_list())) {
                # check the cache for the CS objects
                if (exists($cs_cache->{$cluster_name}->{net_int_nfs_cifs})) {
                    @nfs_cifs_lif_objs = @{$cs_cache->{$cluster_name}->{net_int_nfs_cifs}};
                } else {
                    try {
                        @nfs_cifs_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                               filter => {'role' => "data",
                                                                                          'data-protocol' => ["nfs","cifs"]});
                    } catch NACL::Exceptions::NoElementsFound with {
                        @nfs_cifs_lif_objs = ();
                    } otherwise {
                        my $e = shift;
                        $self->_set_err("Failed to obtain NFS,CIFS lif objects", $e->text());
                    };
                    if ($self->error()) {
                        return $SanUtil::FALSE;
                    }
                    # store in the cache for this cluster
                    $cs_cache->{$cluster_name}->{net_int_nfs_cifs} = \@nfs_cifs_lif_objs;
                }
            }

            my @nfs_lif_objs;
            if (grep(/nfs/i, $self->get_protocol_list())) {
                # check the cache for the CS objects
                if (exists($cs_cache->{$cluster_name}->{net_int_nfs})) {
                    @nfs_lif_objs = @{$cs_cache->{$cluster_name}->{net_int_nfs}};
                } else {
                    try {
                        @nfs_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                          filter => {'role' => "data",
                                                                                     'data-protocol' => ["nfs"]});
                    } catch NACL::Exceptions::NoElementsFound with {
                        @nfs_lif_objs = ();
                    } otherwise {
                        my $e = shift;
                        $self->_set_err("Failed to obtain NFS lif objects", $e->text());
                    };
                    if ($self->error()) {
                        return $SanUtil::FALSE;
                    }
                    # store in the cache for this cluster
                    $cs_cache->{$cluster_name}->{net_int_nfs} = \@nfs_lif_objs;
                }
            }

            my @cifs_lif_objs;
            if (grep(/cifs/i, $self->get_protocol_list())) {
                # check the cache for the CS objects
                if (exists($cs_cache->{$cluster_name}->{net_int_cifs})) {
                    @cifs_lif_objs = @{$cs_cache->{$cluster_name}->{net_int_cifs}};
                } else {
                    try {
                        @cifs_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                                           filter => {'role' => "data",
                                                                                      'data-protocol' => ["cifs"]});
                    } catch NACL::Exceptions::NoElementsFound with {
                        @cifs_lif_objs = ();
                    } otherwise {
                        my $e = shift;
                        $self->_set_err("Failed to obtain CIFS lif objects", $e->text());
                    };
                    if ($self->error()) {
                        return $SanUtil::FALSE;
                    }
                    # store in the cache for this cluster
                    $cs_cache->{$cluster_name}->{net_int_cifs} = \@cifs_lif_objs;
                }
            }

            # get all the data ports on this node
            my @data_port_objs;
            try {
                @data_port_objs = NACL::CS::NetworkPort->fetch(command_interface => $cluster_obj,
                                                               filter => {'node' => $controller_name,
                                                                          'role' => "data"});
            } catch NACL::Exceptions::NoElementsFound with {
                @data_port_objs = ();
            } otherwise {
                my $e = shift;
            $self->_set_err("Failed to obtain data NetworkPort objects", $e->text());
            };
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # catalog all the lifs (nas & iscsi) in a hash so we can iterate over
            # them easily
            my %iscsi_nas_lif_output;
            $iscsi_nas_lif_output{iscsi_target_data_port_mac} = \@iscsi_lif_objs;

            # build out the cumulative array for all NAS protocols
            my (@nas_lif_objs) = (@nfs_lif_objs, @cifs_lif_objs, @nfs_cifs_lif_objs);
            $iscsi_nas_lif_output{nas_target_data_port_mac}   = \@nas_lif_objs;

            # If the caller passed in a list of vservers, check
            # this vserver against that list.
            my $cmode_vservers = $self->get_cmode_vserver_list();
            if ($self->error()) {
                return $SanUtil::FALSE;
            }

            # search through all data ports on this node and corrolate which mac address
            # is associated to each lif
            foreach my $data_port (@data_port_objs) {
                # store iscsi and nas lif data in the switch info hash
                foreach my $port_type (keys %iscsi_nas_lif_output) {
                    # map all lifs of the specified port type to their physical port's mac address
                    foreach my $lif_port (@{$iscsi_nas_lif_output{$port_type}}) {
                        if (scalar(@{$cmode_vservers}) > 0) {
                            # Looks like we received a list of vservers
                            # so see if the current vserver is in that list.
                            my $srv = $lif_port->server();
                            if (grep(/^$srv$/, @{$cmode_vservers}) == 0) {
                                $self->sanlog(-msg => "Not including vserver '$srv' in the switch info " .
                                                      "hash since it is not in the list of vservers passed into FabricMgr",
                                              -msg_level => 3);
                                next;
                            }
                        }

                        # check if this lif lives on this physical data port
                        if (    $lif_port->home_port() eq $data_port->port()
                            and $lif_port->home_node() eq $data_port->node()) {

                            # store this data port's mac address in the hash
                            unless (exists($switch_info->{nacl_nodes_cm}->{$controller_name}->{$port_type}->{$data_port->mac()})) {
                                $switch_info->{nacl_nodes_cm}->{$controller_name}->{$port_type}->{$data_port->mac()} = {};
                            }

                            # store a copy to the reference in the vserver lookup table too
                            # (since this vserver is the owner of the lif)
                            $switch_info->{vservers}->{$lif_port->server()}->{$port_type}->{$data_port->mac()}
                                = $switch_info->{nacl_nodes_cm}->{$controller_name}->{$port_type}->{$data_port->mac()};

                            # store a copy in the cluster lookup table
                            if ($cluster_name ne "") {
                                $switch_info->{nacl_clusters_cm}->{$cluster_name}->{$port_type}->{$data_port->mac()}
                                    = $switch_info->{nacl_nodes_cm}->{$controller_name}->{$port_type}->{$data_port->mac()};

                                # keep track of the cluster name
                                $switch_info->{nacl_nodes_cm}->{$controller_name}->{$port_type}->{$data_port->mac()}
                                    ->{cluster_name} = $cluster_name;
                            }

                            # "wake up" the ethernet switch's routing table by pinging the broadcast ip
                            # via this lif, this will ensure that when we query the switch for the port locations,
                            # that the mac addresses will be listed in the output. This is needed
                            # if there is no traffic flowing through the port at this given point in time.
                            my $broadcast_ip = undef;
                            if ($lif_port->address() !~ /:/i) {
                                $broadcast_ip = &SanUtil::get_ip_broadcast(
                                    -ip_address   => $lif_port->address(),
                                    -ipv4_netmask => $lif_port->netmask());
                            }

                            if (defined($broadcast_ip)) {
                                # try to ping the broadcast ip to wake everyone up
                                # ignore any error
                                my $exception_text = undef;
                                try {
                                    NACL::C::Network->ping(command_interface => $cluster_obj,
                                                           'lif'             => $lif_port->lif(),
                                                           'lif-owner'       => $lif_port->server(),
                                                           'destination'     => $broadcast_ip,
                                                           'count'           => 1);
                                } otherwise {
                                    my $e = shift;
                                    $exception_text = "Ping failed because: " . $e->text();
                                };

                                if (defined($exception_text)) {
                                    $self->sanlog(-msg => $exception_text);
                                }
                            } ## end if (defined($broadcast_ip...
                        } ## end if ($lif_port->{'home_port'...
                    } ## end foreach my $lif_port (@{$iscsi_nas_lif_output...
                } ## end foreach my $port_type (keys...
            } ## end foreach my $data_port (@data_port_objs)
        }

        # discover FC Host and FCVI Host Adapter ports on this cluster
        # this is needed for MCC backend switch discovery
        my $cmd = "sysconfig -v";
        my $sysconfig_str = $cluster_obj->apiset()->execute_raw_command(command => "run -node $controller_name -command \"$cmd\"");
        my @sysconfig_arr = split(/\n/, $sysconfig_str);
        if (@sysconfig_arr) {
            my $get_host_card_info = 0;
            my $get_fcvi_card_info = 0;
            my ($wwpn, $wwnn, $switch, $port);
            foreach my $line (@sysconfig_arr) {
                if ($line =~ /slot\s+\d+:\s+FC Host Adapter/i) {
                    # start searching for fc host card info
                    $get_host_card_info = 1;
                } elsif ($line =~ /slot\s+\d+:\s+FCVI Host Adapter/i) {
                    $get_fcvi_card_info = 1;
                } elsif ($line =~ /slot\s+\d+:/i) {
                    # reset the flags because we encountered a new slot
                    $get_host_card_info = 0;
                    $get_fcvi_card_info = 0;
                } elsif ($line =~ /FC Port Name:\s+(\S+)/i) {
                    $wwpn = $1;
                    # skip this if we are not searching for card info yet
                    if (!$get_fcvi_card_info and !$get_host_card_info) {
                        undef $wwpn;
                        next;
                    }

                    # normalize the wwpn
                    $wwpn = lc($wwpn);
                    $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);
                } elsif ($line =~ /FC Node Name:\s+(\S+)/i) {
                    $wwnn = $1;
                    # skip this if we are not searching for card info yet
                    if (!$get_fcvi_card_info and !$get_host_card_info) {
                        undef $wwnn;
                        next;
                    }

                    # normalize the wwnn
                    $wwnn = lc($wwnn);
                    $wwnn = &SanUtil::wwn_add_colons(-wwn => $wwnn);
                } elsif ($line =~ /Switch Port:\s+(\S+):(\S+)/i) {
                    $switch = $1;
                    $port = $2;
                    # skip this if we are not searching for card info yet
                    if (!$get_fcvi_card_info and !$get_host_card_info) {
                        undef $switch;
                        undef $port;
                        next;
                    }
                } else {
                    # skip this line, it's irrelevant
                    next;
                }

                # if we discovered all the fields needed for the fchost/fcvihost cards, populate the switch info
                # entry, we only look at this if the fchost/fcvihost cards is actually logged in to a switch/port.
                # This is because otherwise we would be issuing warnings on all non-mcc configs.
                if ($get_host_card_info and $wwpn and $switch and $port) {
                    $switch_info->{nacl_nodes_cm}->{$controller_name}->{fc_host_target_wwpn}->{$wwpn}->{cluster_name} = $cluster_name;
                    ($get_host_card_info, $wwpn, $wwnn, $switch, $port) = undef;
                } elsif ($get_fcvi_card_info and $wwnn and $switch and $port) {
                    $switch_info->{nacl_nodes_cm}->{$controller_name}->{fcvi_host_target_wwpn}->{$wwnn}->{cluster_name} = $cluster_name;
                    ($get_fcvi_card_info, $wwpn, $wwnn, $switch, $port) = undef;
                }
            }
        }
    } ## end foreach my $controller_name...

    # Iterate over all of the fc switches and invalidate out the nameserver
    # cache so when we call get_nameserver on the portmgr object with the
    # all vfs flag set to true it will re-disover all of the nameserver information
    # and re-cache it.
    foreach my $switch (@fc_mac_switches) {
        my $switch_obj = $switch_hash->{$switch};
        $switch_obj->invalidate_items_cache_data("nameserver");
    }

    # now we have built out the switch info hash with all the MACs and WWPNs that we intend to discover
    # on the switches that exist in the objects passed in.
    # find out which switch and in what port all the MACs and WWPNs are connected to, and
    # store that data in the switch info hash
    foreach my $resource_type (keys %{$switch_info}) {

        # these are the types of ports we will look through given the resource type
        # we are currently working with
        my @port_types = ();
        if ($resource_type eq "vservers") {
            @port_types = ("fc_target_lif_wwpn", "apcon_target_lif_wwpn");
        } elsif ($resource_type eq "nacl_nodes_cm") {
            @port_types = ("iscsi_target_data_port_mac", "nas_target_data_port_mac", "cluster_port_mac",
                           "fc_host_target_wwpn", "fcvi_host_target_wwpn");
        } elsif ($resource_type eq "sanfilers7m") {
            @port_types = ("fc_target_wwpn", "apcon_target_wwpn", "iscsi_target_mac");
        } elsif ($resource_type eq "sanclients") {
            @port_types = ("fc_initiator_wwpn", "apcon_initiator_wwpn", "iscsi_initiator_mac");
        } elsif ($resource_type eq "nacl_clusters_cm") {
            # skip controllercluster resources (they will get populated by their sub resource (vservers/nacl_nodes_cm))
            next;
        }

        # discover the ports for each resource of this specific resource type
        foreach my $resource_name (keys %{$switch_info->{$resource_type}}) {

            # discover each port type this resource type is capable of
            foreach my $port_type (@port_types) {
                # discover the locations for each mac or wwpn that belongs to this resource
                foreach my $mac_wwpn (keys %{$switch_info->{$resource_type}->{$resource_name}->{$port_type}}) {
                    # keep a flag indicator to tell us whether we found the location of this resource
                    my $found_it = 0;

                    # Figure out if we are dealing with apcons or not and use the appropriate
                    # switch array for the for loop
                    my @switch_list;
                    if ($port_type =~ /^apcon/i) {
                        @switch_list = @apcon_switches;
                    } else {
                        @switch_list = @fc_mac_switches;
                    }

                    # from the list of switches we are working with,
                    # figure out what switch this wwpn/mac lives on
                    foreach my $sanswitch (@switch_list) {
                        my $sanswitch_obj = $switch_hash->{$sanswitch};

                        # Get the switch type
                        my $switch_type = $sanswitch_obj->hosttype();

                        # Grab the nameserver info
                        my $ns = $sanswitch_obj->portmgr()->get_nameserver("1");
                        if ($sanswitch_obj->portmgr()->error) {
                            $self->_set_err("Unable to retrieve the nameserver info for switch '$sanswitch':\n", $sanswitch_obj->portmgr()->errmsg());
                            return $SanUtil::FALSE;
                        }

                        # look for this wwpn/mac on this switch
                        my @port_numbers = $sanswitch_obj->portmgr()->get_port_nums(-wwpn => $mac_wwpn);

                        $self->sanlog(-msg => "Looking for $mac_wwpn on switch '$sanswitch' for resource "
                                              . "'$resource_name'",
                                      -msg_level => 3);

                        if (scalar(@port_numbers) and $port_numbers[0] ne "") {
                            # store the location and switch name for this mac/wwpn in the hash
                            $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn}->{port_number}
                              = \@port_numbers;
                            $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn}->{switch_name}
                              = $sanswitch;

                            # We now need to see if the fid matches either the fid, virtual fabric, or
                            # admin domain of the switch since it is possible that the switch could be
                            # in multiple fabrics.
                            my $fabric = "";
                            if (defined($ns->{$port_numbers[0]}->{fid}) and $ns->{$port_numbers[0]}->{fid} ne "") {
                                $fabric = $self->get_switch_fabric(-switch => $sanswitch,
                                                                   -fid    => $ns->{$port_numbers[0]}->{fid},
                                                                   -wwpn   => $mac_wwpn,
                                                                  );
                            } elsif (defined($ns->{$port_numbers[0]}->{admin_domain})
                                and scalar(@{$ns->{$port_numbers[0]}->{admin_domain}}) > 0) {
                                $fabric = $self->get_switch_fabric(-switch => $sanswitch,
                                                                   -ad     => $ns->{$port_numbers[0]}->{admin_domain},
                                                                   -wwpn   => $mac_wwpn,
                                                                  );
                            } else {
                                $fabric = $self->get_switch_fabric(-switch => $sanswitch,
                                                                   -wwpn   => $mac_wwpn,);
                            }

                            if ($self->error()) {
                                # If we get an error, set the fabric to "NONE" since not all switches passed in
                                # will not necessarily belong to a fabric.  Also clear the error.
                                $fabric = "NONE";
                                # We only want to include this information if this FabricMgr object is going
                                # to be used for zoning. If the user says it is not then we will not include
                                # the extra information.
                                if ($self->{used_for_zoning}) {
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn} =
                                        $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn};
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{$resource_type} = $resource_name;
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{error} .=
                                        "\nCould not find fabric for wwpn $mac_wwpn. There was an ".
                                        "error in FabricMgr::get_switch_fabric call.\n" . $self->errmsg();
                                }
                                $self->_clear_err();
                            } ## end if ($self->error())

                            # If $fabric is "" at this point, set it to "NONE";
                            if ($fabric eq "") {
                                $fabric = "NONE";
                                # We only want to include this information if this FabricMgr object is going
                                # to be used for zoning. If the user says it is not then we will not include
                                # the extra information.
                                if ($self->{used_for_zoning}) {
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn} =
                                        $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn};
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{$resource_type} = $resource_name;
                                    $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{error} .=
                                        "\nCould not find fabric for wwpn. Search for:\n".
                                        "Warnings for wwpn $mac_wwpn\n in the Fabric_Config log";
                                }
                            }

                            $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn}->{fabric}
                              = $fabric;

                            # get the vf context id for this port and store it, if virtual fabrics are enabled on the switch
                            if ($sanswitch_obj->is_vf_enabled()) {
                                my $vf_id = $sanswitch_obj->get_current_vf_context_id();
                                $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn}->{vf_context_id} = $vf_id;
                            } else {
                                $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn}
                                  ->{vf_context_id} = "";
                            }

                            # keep a reverse look up table for quicker access of the information from the switch's
                            # point of view.
                            $switch_info->{sanswitches}->{$sanswitch}->{$port_type}->{$mac_wwpn}
                              = $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn};
                            $switch_info->{sanswitches}->{$sanswitch}->{$port_type}->{$mac_wwpn}->{resource_name}
                              = $resource_name;
                            $switch_info->{sanswitches}->{$sanswitch}->{$port_type}->{$mac_wwpn}->{resource_type}
                              = $resource_type;

                            # Keep the reverse lookup for fabric's point of view
                            $switch_info->{fabrics}->{$fabric}->{$port_type}->{$mac_wwpn}
                              = $switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn};

                            # indicate that we found the port and switch and break out of current foreach switch loop
                            $found_it = 1;
                            &SanUtil::nate_log("-comment",
                                    "Found $mac_wwpn on switch '$sanswitch' ports "
                                  . "'@port_numbers' for resource '$resource_name'");
                            last;
                        } else {

                            # did not find wwpn or mac on this switch,
                            # this is not an error, so clear the error on the portmgr
                            $sanswitch_obj->portmgr()->_clear_err();
                        } ## end else [ if (defined($port_number...
                    } ## end foreach my $sanswitch (keys...

                    if (not $found_it) {

                        # At this point we could not find this MAC/WWPN, issue a warning
                        # but only if we actually had the correct type of switch to look
                        # for the port on.  Also, skip any ethernet warnings since we
                        # don't need ehternet switches in the testbed.
                        if ($port_type =~ /apcon/ and scalar(@apcon_switches > 0)
                            or $port_type =~ /^fc_/ and scalar(@fc_mac_switches > 0)) {
                            &SanUtil::nate_log('-warn',
                                               "Could not find '$port_type' $mac_wwpn for " .
                                               "resource '$resource_name' on switch resources");
                        }

                        # Delete the entry since it wasn't found
                        if ($port_type =~ /apcon/i) {
                            delete($switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn});
                        } else {
                            $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn} =
                                delete($switch_info->{$resource_type}->{$resource_name}->{$port_type}->{$mac_wwpn});
                            $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{$resource_type} = $resource_name;
                            $self->{deleted_wwpns_info}->{$resource_type}->{$mac_wwpn}->{error} .= "\nCould not find wwpn on any switch";
                        }
                        # We also need to check to see if this item needs to be deleted
                        # from the controller cluster hash only for lif wwpn information
                        if ($port_type eq "apcon_target_lif_wwpn" or $port_type eq "fc_target_lif_wwpn") {
                            # Cycle through each cluster and check if it is found there since
                            # we created pointers earlier.
                            foreach my $cluster_name (keys %{$switch_info->{nacl_clusters_cm}}) {
                                if (defined($switch_info->{nacl_clusters_cm}->{$cluster_name}->{$port_type}->{$mac_wwpn})) {
                                    delete($switch_info->{nacl_clusters_cm}->{$cluster_name}->{$port_type}->{$mac_wwpn});
                                }
                            }

                            # Cycle through each nacl_nodes_cm node and check if it is found there since
                            # we created pointers earlier.
                            foreach my $node_name (keys %{$switch_info->{nacl_nodes_cm}}) {
                                if (defined($switch_info->{nacl_nodes_cm}->{$node_name}->{$port_type}->{$mac_wwpn})) {
                                    delete($switch_info->{nacl_nodes_cm}->{$node_name}->{$port_type}->{$mac_wwpn});
                                }
                            }
                        }
                    } ## end if (not $found_it)
                }    # end foreach mac/wwpn
            }    # end foreach port type
        }    # end foreach resource
    }    # end foreach resource type

    if (grep(/fcp/i, $self->get_protocol_list())) {
        # Loop throught the switches again, but this time we are going to look for E ports
        foreach my $sanswitch (@fc_mac_switches) {
            my $sanswitch_obj = $switch_hash->{$sanswitch};

            # skip this switch if it is not an FCP switch (based on hosttype)
            next if ($sanswitch_obj->hosttype() !~ /fcp|brocade/i);

            # Grab the nameserver info
            my $ns = $sanswitch_obj->portmgr()->get_nameserver("1");
            if ($sanswitch_obj->portmgr()->error) {
                $self->_set_err("Unable to retrieve the nameserver info for switch '$sanswitch':\n",
                                $sanswitch_obj->portmgr()->errmsg());
                return $SanUtil::FALSE;
            }

            # Walk the ports and find types that match what we are looking for
            foreach my $port (sort keys %$ns) {
                unless (defined($ns->{$port}->{type}) and $ns->{$port}->{type} ne "") {
                    &SanUtil::nate_log('-comment', "Skipping port '$port' on switch '$sanswitch' because " .
                                                   "port 'type' info is not found");
                    next;
                }

                # Normalize the port type name
                my $normalized_type = &SwitchUtil::normalize_type($ns->{$port}->{type});

                # Check for an e port
                if ($normalized_type eq "E_Port") {
                    # Get the fid if one exits
                    my $vf_context_id = "";
                    if (defined($ns->{$port}->{fid}) and $ns->{$port}->{fid} ne "") {
                        $vf_context_id = $ns->{$port}->{fid};
                    }

                    # Get the port wwpn
                    my $e_port_wwpn = "";
                    if (defined($ns->{$port}->{wwpn}) and $ns->{$port}->{wwpn} ne "") {
                        $e_port_wwpn = $ns->{$port}->{wwpn};

                        # if the wwpn is NA, then we need to make it unique
                        if ($e_port_wwpn eq "NA") {
                            $e_port_wwpn = "${sanswitch}__${port}";
                        } else {
                            # Lowercase the wwpn and add colons
                            $e_port_wwpn = lc($e_port_wwpn);
                            $e_port_wwpn = &SanUtil::wwn_add_colons(-wwn => $e_port_wwpn);
                        }

                        # Add it to the hash; a wwpn may be logged into multiple ports so if the
                        # entry already exists, just add the addtional port.
                        push (@{$switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{port_number} }, $port);
                        $switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{port_number}   = [$port];
                        $switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{switch_name}   = $sanswitch;
                        $switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{resource_name} = $sanswitch;
                        $switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{resource_type} = "sanswitches";
                        $switch_info->{sanswitches}->{$sanswitch}->{fc_isl_wwpn}->{$e_port_wwpn}->{vf_context_id} = $vf_context_id;

                    }
                }
            }
        }
    }

    # go through all switches and revert the command caching to its original value
    foreach my $switch_name (keys %$switch_hash) {
        my $switch_obj = $switch_hash->{$switch_name};

        # revert command caching on the switch object
        $switch_obj->revert_sancache();
        if ($switch_obj->error()) {
            $self->_set_err("Failed to revert sancache on SanSwitch '$switch_name'");
            return $SanUtil::FALSE;
        }
    } ## end foreach my $switch_name (keys...

    # Store the hash on self
    $self->{switch_info} = $switch_info;
    return $SanUtil::TRUE;
} ## end sub build_switch_info_hash

###############################################################################
# Method:    build_all_hashes
# Objective: Method to force the build of all hashes (switch_info, resource, etc)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub build_all_hashes {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Build the switch info hash
    $self->build_switch_info_hash();
    if ($self->error) {
        return $SanUtil::FALSE;
    }

    # Build the client vserver map
    $self->build_client_vserver_map();
    if ($self->error) {
        return $SanUtil::FALSE;
    }

    # Build the resource hash
    $self->build_resource_hash();
    if ($self->error) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    dump_aliases_and_zones_create_hash
# Objective: Dump the various fabric manager hashes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_aliases_and_zones_created_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Get the switch info hash
    my $created_info_hash = $self->get_aliases_and_zones_created_hash();
    if ($self->error) {
        return $SanUtil::FALSE;
    }

    if (scalar(keys %$created_info_hash) > 0) {
        $self->sanlog(-msg => "Aliases and Zones Created Hash:\n" .
                              "Description: This contains data on all aliases and zones" .
                              "created by this FabricMgr object.\n" .
                              Dumper($created_info_hash));
    } else {
        $self->sanlog(-msg => "Aliases and Zones Created Hash is Empty. This means we " .
                              "have not created any aliases or zones with this obj.");
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    dump_hashes
# Objective: Dump the various fabric manager hashes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_hashes {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Create variable to hold all the error messages.
    my $err_msg = 'The following is a list of errors from dumping hashes: ';

    # Create variable to count how many errors we get.
    my $failures = 0;

    # Get the switch info hash
    my $switch_info_hash = $self->get_switch_info_hash();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }

    if (scalar(keys %$switch_info_hash) > 0) {
        $self->sanlog(-msg => "Switch Info Hash:\n" .
                              "Description: This contains data on where all " .
                              "clients/clusters/nodes/vservers/switches are plugged in.\n" .
                              Dumper($switch_info_hash));
    } else {
        $self->sanlog(-msg => "Switch Info Hash is Empty. This means we don't know where anything is plugged in, " .
                              "it's possible that the wrong set of switches were passed into the test.");
    }

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }


    if (scalar(keys %$fabric_hash) > 0) {
        $self->sanlog(-msg => "Switch Fabric Hash:\n" .
                              "Description: This describes which switches belong to this fabric\n" .
                              Dumper($fabric_hash));
    } else {
        $self->sanlog(-msg => "Switch Fabric Hash is empty. This means that we " .
                              "don't know what fabrics each FC switch belong to.");
    }

    # Get the resource hash
    my $resource_hash = $self->get_resource_hash();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }


    if (scalar(keys %$resource_hash) > 0) {
        $self->sanlog(-msg => "Fabric Resource Hash:\n" .
                              "Description: This contains mappings between target wwpns and the initiator wwpns " .
                              "they will be zoned against. It also contains a reverse mapping as well.\n" .
                              Dumper($resource_hash));
    } else {
        $self->sanlog(-msg => "Fabric Resource Hash is Empty");
    }


    # Get the client vserver map
    my $map_hash = $self->get_client_vserver_map();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }


    if (scalar(keys %$map_hash) > 0) {
        $self->sanlog(-msg => "Client Vserver Map Hash:\n" .
                              "Description: This contains a mapping of which clients will get zoned to " .
                              "which vservers in this fabric\n" .
                              Dumper($map_hash));
    } else {
        $self->sanlog(-msg => "Client Vserver Map Hash is empty. This means that we won't be zoning anything for this fabric.");
    }

    # Get the init wwpn to target wwpn map
    my $init_wwpn_target_wwpn_map_hash = $self->get_initiator_wwpn_to_target_wwpn_map();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }


    if (scalar(keys %$init_wwpn_target_wwpn_map_hash) > 0) {
        $self->sanlog(-msg => "Dumping the Initiator WWPN to Target WWPN Map hash Hash:\n" .
                              "Description: This is the main lookup table between initiator WWPNs and target WWPNs. " .
                              "It is build based on the current Igroup configuration on the cluster\n" .
                              Dumper($init_wwpn_target_wwpn_map_hash));
    } else {
        $self->sanlog(-msg => "Initiator WWWPN to Target WWPN map hash is empty, this means we did not find any of the " .
                              "initiators in this fabric on any of igroups owned by the vservers in this fabric.");
    }

    # Get the init wwpn to target wwpn map
    my $deleted_wwpns_info_hash = $self->get_deleted_wwpns_info_hash();
    if ($self->error) {
        # update the failure count.
        $failures++;
        # append the error to the err_msg.
        $err_msg .= "\n" . $self->errmsg();
        # Clear out the error.
        $self->_clear_err();
    }

    if (scalar(keys %$deleted_wwpns_info_hash) > 0) {
        $self->sanlog(-msg => "Dumping the Deleted WWPNs Info Hash:\n" .
                              "Description: This is a hash of the WWPNs that were " .
                              "not found and deleted from Switch Info.\n" .
                              Dumper($deleted_wwpns_info_hash));
    } else {
        $self->sanlog(-msg => "Deleted WWPNs Info Hash is empty, this means  " .
                              "all the wwpns should be found on the switches.");
    }

    if($failures) {
        $self->_set_err("Had $failures error(s)\n" .$err_msg);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method:    get_fabric
# Objective: Return the fabric name from the object if one was specified
#            during contructor initialization.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fabric {
    my ($self) = @_;

    if (defined($self->{fabric})) {
        return $self->{fabric};
    } else {
        return "";
    }
}

###############################################################################
# Method:    get_active_zoneset
# Objective: Return the active zoneset for a fabric
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_active_zoneset {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $fabric = &ParamsMgr::get_arg(-arg => "fabric", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return "";
    }

    # Get the zonemgr object for the fabric
    my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
        return "";
    }

    # Get the active zone set object
    my $active_zonesetobj = $zonemgr_obj->get_active_zoneset();
    if ($zonemgr_obj->error()) {
        $self->_set_err("Unable to retrieve the active zoneset object for fabric '$fabric'", $zonemgr_obj->errmsg());
        return "";
    }
    unless (ref($active_zonesetobj) =~ /zoneset/i) {
        $self->_set_err("Invalid data returned for active zoneset object on fabric '$fabric'");
        return "";
    }

    my $active_zoneset = $active_zonesetobj->get_name();
    unless (defined($active_zoneset) and $active_zoneset ne "") {
        $self->_set_err("Unable to retrieve the active zoneset for fabric '$fabric'");
        return "";
    }

    return $active_zoneset;
} ## end sub get_active_zoneset

###############################################################################
# Method:    get_switch_fabric
# Objective: Return the fabric name of a switch
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_switch_fabric {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $wwpn    = &ParamsMgr::get_arg(-arg => "wwpn", -required => 0, -default => "NOT_SENT_TO_METHOD");
    my $switch  = &ParamsMgr::get_arg(-arg => "switch", -required => 1, -default => "");
    my $fid     = &ParamsMgr::get_arg(-arg => "fid", -required => 0, -default => "");
    my $ad      = &ParamsMgr::get_arg(-arg => "ad", -required => 0, -default => []);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return "";
    }

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return "";
    }

    my $warning_string = 'The following are warnings from searching for the fabric for:\n' .
                         "Switch : $switch\n" .
                         "FID    : $fid\n" .
                         "AD     : " . Dumper($ad) . "\n" .
                         "Warnings for wwpn $wwpn:\n";
    foreach my $fabric (keys %$fabric_hash) {
        # Check to see if the switch is defined on the fabric
        if (defined($fabric_hash->{$fabric}->{switches}->{$switch})) {
            # We found a switch that matches on the fabric.  Now let's
            # see if we can match the fid.
            my $fabric_id      = $fabric_hash->{$fabric}->{switches}->{$switch}->{fabric_id};
            my $virtual_fabric = $fabric_hash->{$fabric}->{switches}->{$switch}->{virtual_fabric};
            my $admin_domain   = $fabric_hash->{$fabric}->{switches}->{$switch}->{admin_domain};

            if ($fabric_id eq ""
                and $virtual_fabric eq ""
                and $admin_domain eq ""
                and scalar(@$ad) == 0
                and $fid eq "") {
                # We don't have any FID, VF, or AD defined on this switch
                # and the fid and admin domain list we are looking for was
                # empty so this fabric owns the entire switch
                return $fabric;
            # Do we have a fabric id for the current fabric:switch(fabric_id)?
            } elsif ($fabric_id ne "") {
                # If the fabric_id does not match the input fabric id(fid)
                # this could be an error so add to warnings
                if($fabric_id != $fid) {
                    $warning_string .= "Fabric_ID Mismatch Warning\n" .
                                       "Switch:$switch in Fabric:$fabric has " .
                                       "Fabric_ID:$fabric_id which does not " .
                                       "match the passed in Fabric_ID:$fid\n";
                    next;
                }
                # Else We have fabric_id and it matches the fid so we found the correct fabric
                return $fabric;
            # Do we have a virtual fabric id for the current fabric:switch(virtual_fabric)?
            } elsif ($virtual_fabric ne "") {
                # If the virtual_fabric does not match the input fabric id(fid)
                # this could be an error so add to warnings
                if($virtual_fabric != $fid) {
                    $warning_string .= "Virtual_Fabric Mismatch Warning\n" .
                                       "Switch:$switch in Fabric:$fabric has " .
                                       "Virtual_Fabric:$virtual_fabric which does not " .
                                       "match the passed in Fabric_ID:$fid\n";
                    next;
                }
                # Else We have virtual fabric and it matches the fid so we found the correct fabric
                return $fabric;
            # Do we have an admin domain for the current fabric:switch(admin_domain)?
            } elsif ($admin_domain ne "") {
                # Loop through the passed in admin domains and see if
                # any match the admin domain of the current fabric:switch
                foreach my $domain (@$ad) {
                    if ($domain ne "" and $domain == $admin_domain) {
                        # We have admin domain and it matches the fid
                        return $fabric;
                    }
                }
                # If we get to here it means that we did not find a match
                $warning_string .= "Admin_Domain Mismatch Warning\n" .
                                       "Switch:$switch in Fabric:$fabric has " .
                                       "Admin_Domain:$admin_domain which does not " .
                                       "match the passed in Admin_Domain:" .
                                       Dumper($ad) . "\n";
                next;
            } else {
                # We didn't find a match so check the next fabric
                # and add a warning that inputs were defined but did not
                # match anything in this fabric.
                $warning_string .= "Input Mismatch Warning\n" .
                                   "Switch:$switch in Fabric:$fabric does not " .
                                   "have Admin_Domain or Fabric_ID or Virtual_Fabric " .
                                   "defined but at least 1 input Fabric_ID:$fid" .
                                   "and/or Admin_Domain:" .
                                   Dumper($ad) . " is defined\n";
                next;
            }
        }
    } ## end foreach my $fabric (keys %$fabric_hash)

    # If we made it this far, then the switch isn't defined in the fabric XML
    # return "" but don't set an error. Log a warning so that if this is
    # because of incorrect configuration of the switch it can be triaged.
    # We only want to print this information if this FabricMgr object is going
    # to be used for zoning. If the user says it is not then we will not print
    # the extra information.
    if ($self->{used_for_zoning}) {
        &SanUtil::nate_log('-warn','' . $warning_string . "\n" . Dumper($fabric_hash));
    }
    return "";
} ## end sub get_switch_fabric

###############################################################################
#                               Accessor Methods
###############################################################################

###############################################################################
# Method:    get_prefix
# Objective: Return the prefix off of the FabricMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_prefix {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $fabric = &ParamsMgr::get_arg(-arg => "fabric", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return "";
    }

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return "";
    }

    unless (defined($fabric_hash->{$fabric}->{prefix}) and $fabric_hash->{$fabric}->{prefix} ne "") {
        $self->_set_err("Invalid value for 'prefix' for fabric '$fabric'");
        return "";
    }

    return $fabric_hash->{$fabric}->{prefix};
} ## end sub get_prefix

###############################################################################
# Method:    get_deleted_wwpns_info_hash
# Objective: Return the deleted_wwpns_info hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_deleted_wwpns_info_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{deleted_wwpns_info}) eq "HASH") {
        $self->_set_err("Invalid value for 'deleted_wwpns_info' on FabricMgr object");
        return {};
    }

    return $self->{deleted_wwpns_info};
} ## end sub get_deleted_wwpns_info_hash

###############################################################################
# Method:    get_aliases_and_zones_created_hash
# Objective: Return the aliases_and_zones_created hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aliases_and_zones_created_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{aliases_and_zones_created}) eq "HASH") {
        $self->_set_err("Invalid value for 'aliases_and_zones_created' on FabricMgr object");
        return {};
    }

    return $self->{aliases_and_zones_created};
} ## end sub get_aliases_and_zones_created_hash

###############################################################################
# Method:    get_client_hash
# Objective: Return the client hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_client_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{clients}) eq "HASH") {
        $self->_set_err("Invalid value for 'clients' on FabricMgr object");
        return {};
    }

    return $self->{clients};
} ## end sub get_client_hash

###############################################################################
# Method:    get_cmode_cluster_hash
# Objective: Return the controller cluster hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cmode_cluster_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{cmode_clusters}) eq "HASH") {
        $self->_set_err("Invalid value for 'cmode_clusters' on FabricMgr object");
        return {};
    }

    return $self->{cmode_clusters};
} ## end sub get_cmode_cluster_hash

###############################################################################
# Method:    get_cmode_vserver_list
# Objective: Return the list of vservers off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cmode_vserver_list {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{cmode_vservers}) eq "ARRAY") {
        $self->_set_err("Invalid value for 'cmode_vservers' on FabricMgr object");
        return []
    }

    return $self->{cmode_vservers};
} ## end sub get_cmode_vserver_list

###############################################################################
# Method:    get_sanfiler7m_hash
# Objective: Return the sanfiler7m hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanfiler7m_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{sanfiler7m}) eq "HASH") {
        $self->_set_err("Invalid value for 'sanfiler7m' on FabricMgr object");
        return {};
    }

    return $self->{sanfiler7m};
} ## end sub get_sanfiler7m_hash

###############################################################################
# Method:    get_switch_hash
# Objective: Return the switch hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_switch_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{switches}) eq "HASH") {
        $self->_set_err("Invalid value for 'switches' on FabricMgr object");
        return {};
    }

    return $self->{switches};
} ## end sub get_switch_hash

###############################################################################
# Method:    get_fabric_hash
# Objective: Return the fabric hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fabric_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    unless (ref($self->{fabrics}) eq "HASH") {
        $self->_set_err("Invalid value for 'fabrics' on FabricMgr object");
        return {};
    }

    return $self->{fabrics};
} ## end sub get_fabric_hash

###############################################################################
# Method:    get_resource_hash
# Objective: Return the resource hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_resource_hash {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $refresh = &ParamsMgr::get_arg(-arg => "refresh", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $self;
    }

    # Check to see if the info resource hash is empty or if $refresh = 1.  If either are
    # true, build the hash.
    unless (defined($self->{resources}) and $refresh == 0) {
        $self->build_resource_hash();
        if ($self->error()) {
            $self->_set_err("Failed to build the client resource hash");
            return {};
        }
    } ## end if (scalar(keys %{$self...

    return $self->{resources};
} ## end sub get_resource_hash

###############################################################################
# Method:    are_there_any_igroup_zoning_restrictions
# Objective: Determine if there are any zoning restrictions
# Details:   See POD documentation at the end of this file.
###############################################################################
sub are_there_any_igroup_zoning_restrictions {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{restrict_zoning_based_on_existing_igroups};
}

###############################################################################
# Method:    get_protocol_list
# Objective: Return the list of protocols to use for this fabricmgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_protocol_list {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    return (wantarray()) ? @{$self->{protocols}} : $self->{protocols};
} ## end sub get_switch_hash

###############################################################################
# Method:    get_wwpn_name
# Objective: Return the name of the host or target based on the wwpn if it
#            is on the fabric.  We will return the name, the cluster, and the fabric.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_wwpn_name {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $wwpn = &ParamsMgr::get_arg(-arg => "wwpn", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return ("", "", "");
    }

    # Get the switch info hash
    my $switch_info_hash = $self->get_switch_info_hash();
    if ($self->error()) {
        $self->_set_error("Unable to retrieve the switch info hash off of self");
        return ("", "", "");
    }

    # Lowercase the wwpn and add colons
    $wwpn = lc($wwpn);
    $wwpn = &SanUtil::wwn_add_colons(-wwn => $wwpn);

    # Iterate through the different fabrics
    foreach my $fabric (sort keys %{$switch_info_hash->{fabrics}}) {
        # Grab the fabric sub hash off of the switch info hash.
        my $fabric_hash = $switch_info_hash->{fabrics}->{$fabric};

        # Loop over each wwpn type
        foreach my $wwpn_type (sort keys %$fabric_hash) {
            # Look up the wwpn in the fabric hash
            if (defined($fabric_hash->{$wwpn_type}->{$wwpn})) {
                my ($name, $cluster);

                # Check for the host name
                if (defined($fabric_hash->{$wwpn_type}->{$wwpn}->{resource_name})) {
                    $name = $fabric_hash->{$wwpn_type}->{$wwpn}->{resource_name};
                } else {
                    $name = "";
                }

                # Check for the cluster name
                if (defined($fabric_hash->{$wwpn_type}->{$wwpn}->{cluster_name})) {
                    $cluster = $fabric_hash->{$wwpn_type}->{$wwpn}->{cluster_name};
                } else {
                    $cluster = "";
                }

                # We found the name so return it.
                return ($name, $cluster, $fabric);
            }
        }
    }

    # If we made it this far, return empty
    return ("", "", "");
} ## end sub get_wwpn_name

###############################################################################
# Method:    get_zonemgr_hash
# Objective: Return the zonemgr hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zonemgr_hash {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $refresh = &ParamsMgr::get_arg(-arg => "refresh", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # Check to see if the hash is empty or if $refresh = 1.  If either are true, build the hash
    unless (defined($self->{zonemgr_objs}) and $refresh == 0) {
        $self->build_zonemgr_objects();
        if ($self->error()) {
            $self->_set_err("Failed to build the zonemgr object hash");
            return {};
        }
    } ## end if (scalar(keys %{$self...

    return $self->{zonemgr_objs};
} ## end sub get_zonemgr_hash

###############################################################################
# Method:    get_zonemgr_object
# Objective: Return the zonemgr object for a given fabric
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zonemgr_object {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $fabric = &ParamsMgr::get_arg(-arg => "fabric", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # Get the zonemgr hash
    my $zonemgr_hash = $self->get_zonemgr_hash();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the zonemgr hash off of self");
        return {};
    }

    unless (exists($zonemgr_hash->{$fabric})) {
        $self->_set_err("Unable to retrieve ZoneMgr object for fabric '$fabric'");
        return {};
    }

    return $zonemgr_hash->{$fabric};
} ## end sub get_zonemgr_object

###############################################################################
# Method:    get_switch_info_hash
# Objective: Return the switch info hash off of self
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_switch_info_hash {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $refresh = &ParamsMgr::get_arg(-arg => "refresh", -required => 0, -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $self;
    }

    # Check to see if the info hash is empty or if $refresh = 1.  If either are true, build the hash
    unless (defined($self->{switch_info}) and $refresh == 0) {
        $self->build_switch_info_hash();
        if ($self->error()) {
            $self->_set_err("Failed to build the switch info hash");
            return {};
        }
    } ## end if (scalar(keys %{$self...

    return $self->{switch_info};
} ## end sub get_switch_info_hash

###############################################################################
# Method:    get_vservers_associated_to_client
# Objective: Get the list of vservers that are associated to a specified client
###############################################################################
sub get_vservers_associated_to_client {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $client = &ParamsMgr::get_arg(-arg => "client", -required => 1, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # generate the relationship hash
    my $hash = $self->get_client_vserver_map();
    if ($self->error()) {
        $self->_set_err("Failed to generate vserver client relationship hash");
        return {};
    }

    # pull out all the vservers that are related to this client
    my $vserver_hash = {};
    foreach my $cluster (keys %{$hash->{$client}}) {
        map { $vserver_hash->{$cluster}->{$_} = $hash->{$client}->{$cluster}->{$_} }
          keys %{$hash->{$client}->{$cluster}};
    }

    return $vserver_hash;
} ## end sub get_vservers_associated_to_client

###############################################################################
# Method:    get_clients_associated_to_vserver
# Objective: Get the list of clients that are associated to a specified vserver
###############################################################################
sub get_clients_associated_to_vserver {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $vserver = &ParamsMgr::get_arg(-arg => "vserver", -required => 1, -default => "");
    my $cluster = &ParamsMgr::get_arg(-arg => "cluster", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # generate the relationship hash
    my $hash = $self->get_client_vserver_map();
    if ($self->error()) {
        $self->_set_err("Failed to generate vserver client relationship hash");
        return {};
    }

    # pull out all the clients that are related to this vserver
    my $vserver_hash = {};
    foreach my $client (keys %$hash) {
        if (exists($hash->{$client}->{$cluster}->{$vserver})) {
            $vserver_hash->{$client} = 1;
        }
    }

    return $vserver_hash;
} ## end sub get_vservers_associated_to_client

###############################################################################
# Method:    get_clients_associated_to_lif_wwpn
# Objective: Get the list of clients that are associated to a lif wwpn
###############################################################################
sub get_clients_associated_to_lif_wwpn {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    my $wwpn = &ParamsMgr::get_arg(-arg => "wwpn", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # generate the relationship hash
    my $map_hash = $self->get_client_vserver_map();
    if ($self->error()) {
        $self->_set_err("Failed to generate vserver client relationship hash");
        return {};
    }

    # Get the switch info hash
    my $switch_info_hash = $self->get_switch_info_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the switch info hash.");
        return {};
    }

    # Get the vserver and cluster information for this lif wwpn
    my $client_hash = {};
    foreach my $vserver (sort keys %{$switch_info_hash->{vservers}}) {
        if (exists($switch_info_hash->{vservers}->{$vserver}->{fc_target_lif_wwpn}->{$wwpn})) {
            my $cluster = $switch_info_hash->{vservers}->{$vserver}->{fc_target_lif_wwpn}->{$wwpn}->{cluster_name};

            # pull out all the clients that are related to this vserver
            foreach my $client (keys %$map_hash) {
                if (exists($map_hash->{$client}->{$cluster}->{$vserver})) {
                    $client_hash->{$client} = 1;
                }
            }
        }
    }

    return $client_hash;
} ## end sub get_vservers_associated_to_lif_wwpn

###############################################################################
# Method:    get_client_vserver_map
# Objective: Return a hash mapping clients with vservers they will be connected to.
###############################################################################
sub get_client_vserver_map {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $refresh = &ParamsMgr::get_arg(-arg => "refresh",
                                      -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # if either of the conditions are true then rebuild the client vserver map
    if (not defined($self->{client_vserver_map}) or $refresh == 1) {
        # Looks like the map doesn't exist yet so call the build routine
        $self->build_client_vserver_map();
        if ($self->error()) {
            $self->_set_err("Unable to build the client vserver map, this defines which " .
                            "client gets zoned to which vserver");
            return {};
        }
    }

    return $self->{client_vserver_map};
} ## end sub get_client_vserver_map

###############################################################################
# Method:    build_client_vserver_map
# Objective: Build a hash mapping clients with vservers they will be connected to.
###############################################################################
sub build_client_vserver_map {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    my ($client_vserver_map, $initiator_wwpn_to_target_wwpn_map) = $self->_build_client_vserver_relationship_map_helper();
    if ($self->error()) {
        $self->_set_err("Failed to build client vserver relationship maps");
        return $SanUtil::FALSE;
    }

    # Save the relationship hash in the client_vserver_map for later use.
    $self->{client_vserver_map} = $client_vserver_map;

    # and if the initiator_wwpn_to_target_wwpn_map isn't defined on $self, then store it just
    # to make things go faster
    if (not defined($self->{initiator_wwpn_to_target_wwpn_map})) {
        $self->{initiator_wwpn_to_target_wwpn_map} = $initiator_wwpn_to_target_wwpn_map;
    }

    return $SanUtil::TRUE;
} ## end sub build_client_vserver_map

###############################################################################
# Method:    get_initiator_wwpn_to_target_wwpn_map
# Objective: Return a hash mapping initiator wwpns to target wwpns
###############################################################################
sub get_initiator_wwpn_to_target_wwpn_map {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $refresh = &ParamsMgr::get_arg(-arg => "refresh",
                                      -default => 0);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # if either of the conditions are true then rebuild the map
    if (not defined($self->{initiator_wwpn_to_target_wwpn_map}) or $refresh == 1) {
        # Looks like the map doesn't exist yet so call the build routine
        $self->build_initiator_wwpn_to_target_wwpn_map();
        if ($self->error()) {
            $self->_set_err("Unable to build the initiator_wwpn_to_target_wwpn_map, this defines which " .
                            "client WWPNs gets zoned to which FC target WWPNs");
            return {};
        }
    }

    return $self->{initiator_wwpn_to_target_wwpn_map};
} ## end sub get_initiator_wwpn_to_target_wwpn_map

###############################################################################
# Method:    build_initiator_wwpn_to_target_wwpn_map
# Objective: Build a hash mapping client wwpns to target wwpns
###############################################################################
sub build_initiator_wwpn_to_target_wwpn_map {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    my ($client_vserver_map, $initiator_wwpn_to_target_wwpn_map) = $self->_build_client_vserver_relationship_map_helper();
    if ($self->error()) {
        $self->_set_err("Failed to build client vserver relationship maps");
        return $SanUtil::FALSE;
    }

    # save the init wwpn to target wwpn map
    $self->{initiator_wwpn_to_target_wwpn_map} = $initiator_wwpn_to_target_wwpn_map;

    # and if the client_vserver_map isn't defined on $self, then store it just
    # to make things go faster
    if (not defined($self->{client_vserver_map})) {
        $self->{client_vserver_map} = $client_vserver_map;
    }

    return $SanUtil::TRUE;
} ## end sub build_initiator_wwpn_to_target_wwpn_map

###############################################################################
# Method:    _build_client_vserver_relationship_map_helper
# Objective: Build a hashes mapping clients to vservers, this
# generates hashes that contain the data that indicates which initiator
# will get zoned to which target
###############################################################################
sub _build_client_vserver_relationship_map_helper {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    my $client_vserver_map = {};

    my $initiator_wwpn_to_target_wwpn_map = {};

    # get all cmode controller cluster objects
    my $cmode_cluster_hash = $self->get_cmode_cluster_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the cmode clusters hash");
        return $SanUtil::FALSE;
    }

    # iterate over all clusters
    foreach my $cluster_name (keys %$cmode_cluster_hash) {

        # get the sanfiler object for this filer
        my $cluster_obj = $cmode_cluster_hash->{$cluster_name};

        # get all the vservers on this cluster
        my @vserver_objs;
        try {
            @vserver_objs = NACL::CS::Vserver->fetch(command_interface => $cluster_obj,
                                                     filter => {type => 'data|cluster'});
        } catch NACL::Exceptions::NoElementsFound with {
            @vserver_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Unable to fetch NACL Vserver Component States", $e->text());
        };
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # Check to see if the caller passed in a list of vservers
        # that we should limit the operations to.
        my $cmode_vservers = $self->get_cmode_vserver_list();
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # fetch all the igroup component state objects up front
        my @igroup_objs;
        try {
            @igroup_objs = NACL::CS::LunIgroup->fetch(command_interface => $cluster_obj,
                                                      filter => { 'protocol' => "fcp,mixed" });
        } catch NACL::Exceptions::NoElementsFound with {
            @igroup_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Unable to fetch NACL::CS::LunIgroup objects", $e->text());
        };
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # fetch all the portset component state objects up front
        my @portset_objs;
        try {
            @portset_objs = NACL::CS::LunPortset->fetch(command_interface => $cluster_obj,
                                                        filter => { 'protocol' => "fcp,mixed" });
        } catch NACL::Exceptions::NoElementsFound with {
            @portset_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Unable to fetch NACL::CS::LunPortset objects", $e->text());
        };
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # fetch all fcp lifs on this cluster
        my @fc_lif_objs;
        try {
            @fc_lif_objs = NACL::CS::NetworkInterface->fetch(command_interface => $cluster_obj,
                                                             filter => {'data-protocol' => ["fcp"]});
        } catch NACL::Exceptions::NoElementsFound with {
            @fc_lif_objs = ();
        } otherwise {
            my $e = shift;
            $self->_set_err("Unable to fetch NACL::CS::NetworkInterface objects", $e->text());
        };
        if ($self->error()) {
            return $SanUtil::FALSE;
        }

        # iterate over all vservers
        foreach my $vserver_obj (@vserver_objs) {
            my $vserver_name = $vserver_obj->vserver();

            if (scalar(@{$cmode_vservers}) > 0) {
                # Looks like we received a list of vservers from the user
                # so see if the current vserver is in that list.
                if (grep(/^$vserver_name$/, @{$cmode_vservers}) == 0) {
                    # This vserver doesn't exist.  Log a message and
                    # skip it.
                    $self->sanlog(-msg => "Not including vserver '$vserver_name' in the client-vserver " .
                                          "map since it is not in the list of vservers passed into FabricMgr");
                    next;
                }
            }

            # filter the igroup object list to only those owned by this vserver
            my @igroup_objs_filtered = grep {
                $_->vserver() eq $vserver_name;
            } @igroup_objs;

            # filter the lif objects for this vserver
            my @fc_lif_objs_filtered = grep {
                $_->vserver() eq $vserver_name;
            } @fc_lif_objs;

            # filter the portset objects for this vserver
            my @portset_objs_filtered = grep {
                $_->vserver() eq $vserver_name;
            } @portset_objs;

            # Iterate over the igroups
            foreach my $igroup_obj (@igroup_objs_filtered) {
                my $igroup_name = $igroup_obj->igroup();

                # find the portset that this igroup is bound to
                my $found_portset_obj;
                foreach my $portset_obj (@portset_objs_filtered) {
                    if (grep(/^$igroup_name$/, $portset_obj->igroups())) {
                        $found_portset_obj = $portset_obj;
                        last;
                    }
                }

                my @fc_lif_objs_filtered_by_portset;
                # if it is bound, then obtain the list of lifs in the portset, and
                # filter down the fc_lif_objs_filtered list into a new portset filtered lif list"
                if ($found_portset_obj) {
                    foreach my $lif_name ($found_portset_obj->port_name()) {
                        foreach my $lif_obj (@fc_lif_objs_filtered) {
                            if ($lif_obj->lif() eq $lif_name) {
                                push(@fc_lif_objs_filtered_by_portset, $lif_obj);
                            }
                        }
                    }
                } else {
                    # if its not bound, then no portset lif filtering should be done
                    # for this igroup
                }

                # pull out the initiators
                my @inits = $igroup_obj->initiator();

                # Iterate over each initiator in the igroup
                foreach my $init (@inits) {
                    # Get the client name for the initiator
                    my ($client, $undef1, $undef2) = $self->get_wwpn_name(-wwpn => $init);
                    if ($self->error()) {
                        $self->_set_err("Unable to retrieve the client name for '$init'");
                        return $SanUtil::FALSE;
                    }

                    # Store the vserver info in the hash by init if $client is not empty
                    if ($client ne "") {
                        # determine if this client wants to do fcp portset level filtering.
                        my $do_portset_filtering = $self->get_client_hash()->{$client}->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET};

                        # if this client wants to do portset level filtering, and this igroup is bound to a portset,
                        # then use the portset filtered lif list
                        if ($found_portset_obj and scalar(@fc_lif_objs_filtered_by_portset) and $do_portset_filtering) {
                            # store a mapping between this initiator wwpn and the the wwpns of all portset filtered lifs
                            # this controls which initiators get zoned to which lifs
                            foreach my $fc_lif_obj (@fc_lif_objs_filtered_by_portset) {
                                # this indicates that this init and this target are "zonable"
                                $initiator_wwpn_to_target_wwpn_map->{$init}->{$fc_lif_obj->wwpn()} += 1;

                                # increment the number of possible zones we'll be making between this client and vserver
                                push(@{$client_vserver_map->{$client}->{$cluster_name}->{$vserver_name}}, [$init, $fc_lif_obj->lif()]);
                            }
                        } else {
                            # the client may or have may not requested portset filtering and we dont have any portsets
                            # on this igroup for this initiator, therefore we have to use all the fc lif objs

                            # store a mapping between this initiator wwpn and the the wwpns of all lifs in this vserver
                            # this controls which initiators get zoned to which lifs
                            foreach my $fc_lif_obj (@fc_lif_objs_filtered) {
                                # this indicates that this init and this target are "zonable"
                                $initiator_wwpn_to_target_wwpn_map->{$init}->{$fc_lif_obj->wwpn()} += 1;

                                # increment the number of possible zones we'll be making between this client and vserver
                                # increment the number of possible zones we'll be making between this client and vserver
                                push(@{$client_vserver_map->{$client}->{$cluster_name}->{$vserver_name}}, [$init, $fc_lif_obj->lif()]);
                            }
                        }
                    } else {
                        $self->sanlog(-msg => "Skipping initiator '$init' since it wasn't found in any of the fabrics");
                    }
                }
            } ## end foreach my $igroup_href (@igroups_objs_filtered)
        } ## end foreach my $vserver_obj (@vserver_objs)
    } ## end foreach my $cluster_name (keys...

    return ($client_vserver_map, $initiator_wwpn_to_target_wwpn_map);
} ## end sub _build_client_vserver_relationship_map_helper

###############################################################################
# Method:    should_init_wwpn_be_zoned_to_target_wwpn
# Objective: Determine if this initiator wwpn should be zoned to this target wwpn
# Outputs:   See POD
###############################################################################
sub should_init_wwpn_be_zoned_to_target_wwpn {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $init_wwpn = &ParamsMgr::get_arg(-arg => "init_wwpn",
                                        -required => 1);
    my $target_wwpn = &ParamsMgr::get_arg(-arg => "target_wwpn",
                                        -required => 1);

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;;
    }

    my $initiator_wwpn_to_target_wwpn_map = $self->get_initiator_wwpn_to_target_wwpn_map(-refresh => 0);
    if ($self->error()) {
        $self->_set_err("Failed to obtain Initiator WWPN to target WWPN map");
        return $SanUtil::FALSE;
    }

    # check if there is a mapping between these two wwpns
    if (exists($initiator_wwpn_to_target_wwpn_map->{$init_wwpn}) and
        exists($initiator_wwpn_to_target_wwpn_map->{$init_wwpn}->{$target_wwpn})) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

###############################################################################
# Method:    build_zonemgr_objects
# Objective: create the zonemgr objects for the fabrics provided.
# Outputs:   See POD
###############################################################################
sub build_zonemgr_objects {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return $SanUtil::FALSE;
    }

    # Get the switch hash
    my $switch_hash = $self->get_switch_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the switch hash");
        return $SanUtil::FALSE;
    }

    # For each fabric, pick a switch and create a ZoneMgr object for it.  We only need
    # one ZoneMgr object per fabric.
    foreach my $fabric (keys %$fabric_hash) {

        # Check to see if a master switch was defined. If there is, then use
        # that switch.  If not, then grab one out of the switches list.
        my $switch = "";

        if (defined($fabric_hash->{$fabric}->{master_switch}) and $fabric_hash->{$fabric}->{master_switch} ne "") {
            $switch = $fabric_hash->{$fabric}->{master_switch};
        } else {
            # Get a list of switches
            unless (scalar(keys %{$fabric_hash->{$fabric}->{switches}}) > 0) {
                $self->_set_err("No switches have been defined for fabric '$fabric'");
                return $SanUtil::FALSE;
            }
            my @switches = (keys %{$fabric_hash->{$fabric}->{switches}});

            # Grab the first switch found and use it for the zonemgr object
            $switch     = shift(@switches);
        }

        # Make sure we have switch object for the switch chosen.  Error out if we don't.
        unless (defined($switch_hash->{$switch}) and ref($switch_hash->{$switch}) =~ /sanswitch/i) {
            $self->_set_err("No object found for switch '$switch' on fabric '$fabric'");
            return $SanUtil::FALSE;
        }
        my $switch_obj = $switch_hash->{$switch};

        # Get the switch hosttype
        my $hosttype = $switch_obj->get_connection()->{hostp}{hosttype};

        my $zonemgr_obj;
        if ($hosttype =~ /switch-brocade/i) {

            # Get the fid field
            my $fabric_id = $fabric_hash->{$fabric}->{switches}->{$switch}->{fabric_id};

            if ($fabric_id eq "") {
                # Get the admin_domain field
                my $admin_domain = $fabric_hash->{$fabric}->{switches}->{$switch}->{admin_domain};

                # If we have an AD defined on the fabric, set the ad value on the switch object
                if ($admin_domain ne "") {
                    # Set the admin_domain value on the switch
                    $switch_obj->set_admin_domain($admin_domain);

                    # Switch to the admin domain
                    $switch_obj->switch_to_admin_domain($admin_domain);
                    if ($switch_obj->error) {
                        $self->_set_err("Unable to switch to admin domain " .
                                        "'$fabric_hash->{$fabric}->{admin_domain}':\n", $switch_obj->errmsg);
                        return $SanUtil::FALSE;
                    }
                }

                # Create the Zone manager object without any parameters
                $zonemgr_obj = $switch_obj->create_zonemgr();
            } else {
                &SanUtil::nate_log("-comment", "Switching switch to virtual context '$fabric_id'");
                $switch_obj->set_vf_context($fabric_id);
                if ($switch_obj->error) {
                    $self->_set_err("Unable to change to virtual context '$fabric_id'", $switch_obj->errmsg());
                    return $SanUtil::FALSE;
                } else {
                    &SanUtil::nate_log("-comment", "Virtual Fabric ID set to '$fabric_id''}");
                }

                # Create the Zone manager object with the -vsan parameter
                $zonemgr_obj = $switch_obj->create_zonemgr(-vsan => $fabric_id);
            } ## end else [ if ($fabric_id eq "")
            if ($switch_obj->error) {
                $self->_set_err("Unable to create the ZoneMgr object", $switch_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } elsif ($hosttype =~ /switch-cisco/i) {

            # Get the vf
            my $vf = $fabric_hash->{$fabric}->{switches}->{$switch}->{virtual_fabric};
            if ($vf eq "") {

                # Create the Zone manager object without any parameters
                $zonemgr_obj = $switch_obj->create_zonemgr();
            } else {

                # Create the Zone manager object with the -vsan parameter
                $zonemgr_obj = $switch_obj->create_zonemgr(-vsan => $vf);
            }
            if ($switch_obj->error) {
                $self->_set_err("Unable to create the ZoneMgr object", $switch_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Switch type of '$hosttype' is not supported by FabricMgr");
            return $SanUtil::FALSE;
        }

        # Update Self with the new zonemgr objects
        $self->{zonemgr_objs}->{$fabric} = $zonemgr_obj;
    } ## end foreach my $fabric (keys %$fabric_hash)

    return $SanUtil::TRUE;
} ## end sub build_zonemgr_objects

###############################################################################
# Method:    build_resource_hash
# Objective: Using the objects provided, return the hash of client and targets
#            to be used for the zoning operations.
# Inputs:    None
# Outputs:   Returns a hash reference. Sets an error on self if an error occurs
#            $resource_hash =>
#                    <fabric> =>
#                        initiators =>
#                            <wwpn> =>
#                                name         => <client_name>
#                                cluster_name => <client_name>
#                                device       => <hba device name and instance>
#                                type         => <pwwn|fwwn>
#                                targets      => <array of target wwpns to zone>
#                        targets =>
#                             <wwpn> =>
#                                name       => Vserver Name or Node Name in 7mode
#                                node       => Node Name
#                                device     => <hba device name and instance>
#                                cluster    => Cluster name or ""
#                                type       => <pwwn|fwwn>
#                                initiators => <array of initiator wwpns to zone>
###############################################################################
sub build_resource_hash {
    my ($self) = @_;

    # Clear out the errors
    $self->_clear_err();

    my $switch_info_hash = $self->get_switch_info_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the switch info hash.");
        return $SanUtil::FALSE;
    }

    my $resource_hash = {};

    # Get the hash generated from the xml for the fabric
    my $fabric_xml_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric xml hash.");
        return $SanUtil::FALSE;
    }

    # Get the client vserver map
    my $client_vserver_map = $self->get_client_vserver_map();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the client vserver map.");
        return $SanUtil::FALSE;
    }

    foreach my $fabric (sort keys %{$switch_info_hash->{fabrics}}) {

        # Define the $targets_hash hash reference to store all of the targets for
        # each node in a hash keyed on node with an array ref for the value.
        my $target_hash = {};

        # We want to skip any fabric named "NONE"
        next if ($fabric eq "NONE");

        my $fabric_hash = $switch_info_hash->{fabrics}->{$fabric};

        # Build a list of all targets available for zoning for each client.  We
        # will use the vserver client map for this purpose.  We need to do this so
        # we will know if we have enough targets to meet the user's requirements.
        my $client_lif_hash = {};

        foreach my $init_wwpn (sort keys %{$fabric_hash->{fc_initiator_wwpn}}) {
            my $client_name = $fabric_hash->{fc_initiator_wwpn}->{$init_wwpn}->{resource_name};

            # Cycle through all of the fc lif targets
            foreach my $target_wwpn (sort keys %{$fabric_hash->{fc_target_lif_wwpn}}) {
                my $vserver = $fabric_hash->{fc_target_lif_wwpn}->{$target_wwpn}->{resource_name};
                my $cluster = $fabric_hash->{fc_target_lif_wwpn}->{$target_wwpn}->{cluster_name};

                if (exists($client_vserver_map->{$client_name}->{$cluster}->{$vserver})) {
                    $client_lif_hash->{$init_wwpn}->{$target_wwpn} = 1;
                }
            }
        }

        # Get the value for zones_per_node_per_init for the fabric
        my $zones_per_node_per_init;
        if (defined($fabric_xml_hash->{$fabric}->{zones_per_node_per_init})
            and $fabric_xml_hash->{$fabric}->{zones_per_node_per_init} ne "") {
            $zones_per_node_per_init = $fabric_xml_hash->{$fabric}->{zones_per_node_per_init};
        } else {
            $self->_set_err("Unable to find the fabric attribute 'zones_per_node_per_init' for fabric '$fabric'");
            return $SanUtil::FALSE;
        }

        # We need to get all of the targets and group them by node.  This will allow
        # us to create the correct number of paths based on zones per node per initiator type logic.
        foreach my $target_wwpn (sort keys %{$fabric_hash->{fc_target_wwpn}}) {
            # Convert the target_wwpn so it hash colons in it. We need to do this so that the wwpns represented
            # in the resource hash are in the format used by brocade and cisco alias commands
            my $new_target_wwpn = &SanUtil::wwn_add_colons(-wwn => $target_wwpn);

            # Define the node and "vserver".  In the case of 7mode targets, we will use the node
            # name to represent the vserver. We have to do this so the same code logic later can
            # handle cmode and 7mode.
            my $node          = $fabric_hash->{fc_target_wwpn}->{$target_wwpn}->{resource_name};
            my $resource_type = $fabric_hash->{fc_target_wwpn}->{$target_wwpn}->{resource_type};

            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{name}          = $node;
            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{cluster_name}  = "";
            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{type}          = "pwwn";
            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{node}          = $node;
            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{resource_type} = $resource_type;

            $target_hash->{$node}->{$node}->{$new_target_wwpn}->{device} =
                $fabric_hash->{fc_target_wwpn}->{$target_wwpn}->{device_name};

        }

        # Add all of the fcp lif targets
        foreach my $target_lif_wwpn (keys %{$fabric_hash->{fc_target_lif_wwpn}}) {
            # Define the cluster node and vserver
            my $node          = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{curr_node_name};
            my $cluster_name  = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{cluster_name};
            my $vserver       = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{resource_name};
            my $resource_type = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{resource_type};

            # Get the clients related to the vserver this lif belongs to.  If we don't
            # find any then we will skip this lif.
            my $vserver_map = $self->get_clients_associated_to_vserver( -cluster => $cluster_name,
                                                                        -vserver => $vserver,
                                                                      );
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the clients mapped to vserver '$vserver' " .
                                "on cluster '$cluster_name'");
                return $SanUtil::FALSE;
            }

            unless (scalar(keys %$vserver_map) > 0) {
                # We didn't find any clients for the vserver that owns this
                # LIF so skip it.
                $self->{deleted_wwpns_info}->{$resource_type}->{$target_lif_wwpn} =
                    $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn};
                $self->{deleted_wwpns_info}->{$resource_type}->{$target_lif_wwpn}->{$resource_type} = $vserver;
                $self->{deleted_wwpns_info}->{$resource_type}->{$target_lif_wwpn}->{error} .=
                    "\nSkipping vserver '$vserver' lif '$target_lif_wwpn'" .
                    "because there are no clients mapped to it";
                $self->sanlog(-msg => "Skipping vserver '$vserver' lif '$target_lif_wwpn'" .
                                      "because there are no clients mapped to it");
                next;
            }

            # Convert the target_lif_wwpn so it hash colons in it. We need to do this so that the wwpns represented
            # in the resource hash are in the format used by brocade and cisco alias commands
            my $new_target_lif_wwpn = &SanUtil::wwn_add_colons(-wwn => $target_lif_wwpn);

            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{name}          = $vserver;
            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{node}          = $node;
            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{type}          = "pwwn";
            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{resource_type} = $resource_type;

            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{device}
              = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{lif_name};

            $target_hash->{$node}->{$vserver}->{$new_target_lif_wwpn}->{cluster_name}
              = $fabric_hash->{fc_target_lif_wwpn}->{$target_lif_wwpn}->{cluster_name};
        } ## end foreach my $target_lif_wwpn...

        # make sure that we have enough resources to cover the number of zones per node per initiator
        # that we need unless they have asked for "max"
        my @errors = ();
        if (lc($zones_per_node_per_init) ne "max") {
            foreach my $node (keys %$target_hash) {
                foreach my $vserver (keys %{$target_hash->{$node}}) {
                    # if the number of targets for this vserver on this node is less than the
                    # number of zones requested per node per initiator then bail
                    if (scalar(keys %{$target_hash->{$node}->{$vserver}}) < $zones_per_node_per_init) {
                        push (@errors,
                            "Not enough targets on node '$node' vserver '$vserver' to reach " .
                            "'$zones_per_node_per_init' zone\(s\) per node per initiator on fabric '$fabric'");
                    }
                }
            }
        }

        # set error and return if we had messages in @errors.
        if (scalar(@errors) > 0) {
            my $errmsg = join ("\n", @errors);
            $self->_set_err($errmsg);
            return $SanUtil::FALSE;
        }

        # Add all of the initiators
        foreach my $init_wwpn (keys %{$fabric_hash->{fc_initiator_wwpn}}) {
            # Get the vservers related to the client that owns this init.
            # If we dont' find any, we will skip this init.
            my $client = $fabric_hash->{fc_initiator_wwpn}->{$init_wwpn}->{resource_name};
            my $vserver_map = $self->get_vservers_associated_to_client(-client => $client);
            if ($self->error()) {
                $self->_set_err("Unable to retrieve the vservers mapped to client '$client'");
                return $SanUtil::FALSE;
            }

            # See if we have any resources to zone this client too
            if (scalar(keys %$vserver_map) == 0
                and scalar(keys %{$fabric_hash->{fc_target_wwpn}}) == 0) {
                # We didn't find any vserver for this client nor did we find any
                # fc 7mode targets in the testbed so skip this initiator.
                $self->{deleted_wwpns_info}->{sanclients}->{$init_wwpn} =
                    $fabric_hash->{fc_initiator_wwpn}->{$init_wwpn};
                $self->{deleted_wwpns_info}->{sanclients}->{$init_wwpn}->{sanclients} = $client;
                $self->{deleted_wwpns_info}->{sanclients}->{$init_wwpn}->{error} .=
                    "\nSkipping initiator $init_wwpn and client '$client' ".
                    "because there are no Cmode vservers and no 7mode " .
                    "controllers mapped";
                $self->sanlog(-msg => "Skipping initiator $init_wwpn and client '$client' ".
                                      "because there are no Cmode vservers and no 7mode " .
                                      "controllers mapped");
                next;
            }

            # Convert the init_wwpn so it hash colon in it. We need to do this so that the wwpns represented
            # in the resource hash are in the format used by brocade and cisco alias commands
            my $new_init_wwpn = &SanUtil::wwn_add_colons(-wwn => $init_wwpn);

            $resource_hash->{$fabric}->{initiators}->{$new_init_wwpn}->{name}
              = $fabric_hash->{fc_initiator_wwpn}->{$init_wwpn}->{resource_name};

            $resource_hash->{$fabric}->{initiators}->{$new_init_wwpn}->{device}
              = $fabric_hash->{fc_initiator_wwpn}->{$init_wwpn}->{device_name};

            $resource_hash->{$fabric}->{initiators}->{$new_init_wwpn}->{type}         = "pwwn";
            $resource_hash->{$fabric}->{initiators}->{$new_init_wwpn}->{cluster_name} = "";
            $resource_hash->{$fabric}->{initiators}->{$new_init_wwpn}->{clusters}     = {};
        } ## end foreach my $init_wwpn (keys...

        # We want to make sure that we received at least one target and one initiator
        unless (scalar(keys %{$resource_hash->{$fabric}->{initiators}}) > 0
            and scalar(keys %$target_hash) > 0) {
            $self->_set_err("Invalid number of resources for fabric '$fabric'.  You must have at least one target "
                  . "and one initiator in the fabric.");
            $self->sanlog(-msg => Dumper($resource_hash));
            $self->sanlog(-msg => Dumper($target_hash));
            return $SanUtil::FALSE;
        } ## end unless (scalar(keys %{$resource_hash...

        # We now need to update the targets information in the resource hash to match the number of
        # zones per node per initiator (or really zones per node per vserver per initiator).
        if (lc($zones_per_node_per_init) eq "max") {
            # We want max zones
            foreach my $node (sort keys %$target_hash) {
                # For each of the nodes on the current fabric
                foreach my $vserver (sort keys %{$target_hash->{$node}}) {
                    # For each of the vservers for the current node
                    foreach my $target_wwpn (sort keys %{$target_hash->{$node}->{$vserver}}) {
                        # For each of the targets wwpns on the vserver for the current node
                        foreach my $init_wwpn (sort keys %{$resource_hash->{$fabric}->{initiators}}) {
                            # Make sure we have a client->vserver mapping if using vservers or
                            # Since we want "max" zones, add each initiator to the initiators array for the
                            # current target wwpn on the current vserver on the current node.
                            my $resource_type = $target_hash->{$node}->{$vserver}->{$target_wwpn}->{resource_type};

                            if ($resource_type eq "sanfilers7m") {
                                # not checking igroups for the existence of $init_wwpn to $target_wwpn mappings
                                # for 7mode zoning at this time

                                # dont add duplicates to this target's initiator hash
                                if (!grep(/^$init_wwpn$/, @{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}})) {
                                    push (@{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}}, $init_wwpn);
                                }
                                # dont add duplicates to this initiator's target hash
                                if (!grep(/^$target_wwpn$/, @{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}})) {
                                    push (@{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}}, $target_wwpn);
                                }
                            } else {
                                my $vserver_map = $self->get_clients_associated_to_lif_wwpn(-wwpn => $target_wwpn);
                                if ($self->error()) {
                                    $self->_set_err("Unable to get the list of clients associated to lif '$target_wwpn'");
                                    return $SanUtil::FALSE;
                                }
                                if (scalar(keys %$vserver_map > 0)) {
                                    # we have a mapping between this client and vserver
                                    # first verify if these two wwpns should be zoned together
                                    if ($self->should_init_wwpn_be_zoned_to_target_wwpn(-init_wwpn => $init_wwpn,
                                                                                        -target_wwpn => $target_wwpn) or
                                        not $self->are_there_any_igroup_zoning_restrictions()) {
                                        # dont add duplicates to this target's initiator hash
                                        if (!grep(/^$init_wwpn$/, @{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}})) {
                                            push (@{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}}, $init_wwpn);
                                        }
                                        # dont add duplicates to this initiator's target hash
                                        if (!grep(/^$target_wwpn$/, @{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}})) {
                                            push (@{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}}, $target_wwpn);
                                        }

                                    } elsif ($self->error()) {
                                        $self->_set_err("Failed to determine if initiator $init_wwpn should be zoned to target $target_wwpn");
                                        return $SanUtil::FALSE;
                                    }
                                } else {
                                    # We don't have a mapping for this relationship and we are not dealing with 7mode
                                    # controllers so skip this initiator for this node.
                                    next;
                                }
                            }
                        }
                    }
                }
            }
        } else {
            # We want a specific number of zones
            foreach my $node (sort keys %$target_hash) {
                # For each of the nodes on the current fabric
                foreach my $vserver (sort keys %{$target_hash->{$node}}) {
                    # Get the cluster name and resource_type from one of
                    # the lifs on the vserver
                    my $resource_type = "";
                    my $vserver_lif_wwpn  = "";
                    foreach my $target_wwpn (keys %{$target_hash->{$node}->{$vserver}}) {
                        $resource_type    = $target_hash->{$node}->{$vserver}->{$target_wwpn}->{resource_type};
                        $vserver_lif_wwpn = $target_wwpn;
                        last;
                    }

                    if ($resource_type eq "") {
                        $self->_set_err("Unable to determine the resource type for '$vserver' on node '$node'");
                        return $SanUtil::FALSE;
                    }

                    # Skip the vserver if it has no mappings
                    if ($resource_type =~ /vserver/) {
                        my $clients_hash = $self->get_clients_associated_to_lif_wwpn(-wwpn => $vserver_lif_wwpn);
                        if ($self->error()) {
                            $self->_set_err("Unable to determine the client lif map for  for '$vserver' on node '$node'");
                            return $SanUtil::FALSE;
                        }
                        if (scalar(keys %$clients_hash) == 0) {
                            # This is a vserver and there are no mapped clients
                            # so skip it.
                            next;
                        }
                    }

                    # We have 3 main cases
                    #
                    # 1. Initiators >= Targets = 1
                    # 2. Initiators <= Targets > 1
                    # 3. Initiators > Targets > 1
                    #
                    # Case 1 would be able to achive HA but not necessarily provide optimal paths
                    # during a port block.
                    #
                    # Case 2 and Case 3 would provide HA and Optimal paths.

                    # Create an array of initiators that are orderd by host.  We first need to create
                    # a hash of inits by host, then put them in an array.
                    my $init_ref = {};
                    my @initiators = ();
                    foreach my $init_wwpn (sort keys %{$resource_hash->{$fabric}->{initiators}}) {
                        my $host = $resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{name};
                        push (@{$init_ref->{$host}}, $init_wwpn);
                    }
                    # Now put each initiator into the @initiators array so that they will
                    # be ordered by host.
                    foreach my $host (sort keys %$init_ref) {
                        foreach my $init (sort @{$init_ref->{$host}}) {
                            push (@initiators, $init);
                        }
                    }

                    # Create an array of targets
                    my @targets = ();
                    foreach my $target_wwpn (sort keys %{$target_hash->{$node}->{$vserver}}) {
                        push (@targets, $target_wwpn);
                    }

                    # Figure out the maximum path combinations
                    my $max_paths = scalar(@targets) * scalar(@initiators);

                    # Create a super target array so we have at least as many targets as it will
                    # take to create the maximum number of path combinations.  The end result will
                    # look like "a,b,c,d,a,b,c,d,a,b,c,d"
                    my @super_targets = @targets;
                    while (scalar(@super_targets) < $max_paths) {
                        # Keep adding targets onto super_targets until super_targets is greater than initiators
                        push (@super_targets, @targets);
                    }

                    # For each of the vservers for the current node
                    # Get a count of targets and initiators and save it.
                    if (scalar(@targets) == 0 or scalar(@initiators) == 0) {
                        # Unacceptable config
                        $self->_set_err("Invalid number of targets or initiators found for " .
                                        "node '$node' vserver '$vserver' on fabric '$fabric'");
                        self->sanlog(-msg => Dumper($resource_hash));
                        return $SanUtil::FALSE;
                    } elsif (scalar(@targets) == 1) {
                        # Case 1.  Map each initiator to the target
                        my $target_wwpn = pop(@targets);
                        foreach my $init_wwpn (@initiators) {
                            # Make sure we have a client->vserver mapping.  We can distinguish between
                            # 7mode controllers and Cmode vservers because "cluster_name" will be "" on
                            # 7mode controllers.
                            my $cluster_name = $target_hash->{$node}->{$vserver}->{$target_wwpn}->{cluster_name};
                            my $client_name  = $resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{name};

                            if ($cluster_name eq ""
                                or exists($client_vserver_map->{$client_name}->{$cluster_name}->{$vserver})) {
                                # Looks like this set has a map so add it to the resource hash
                                # first verify if these two wwpns should be zoned together
                                # if the resource type is a 7mode controller, pair them together
                                if ($self->should_init_wwpn_be_zoned_to_target_wwpn(-init_wwpn => $init_wwpn,
                                                                                    -target_wwpn => $target_wwpn) or
                                    not $self->are_there_any_igroup_zoning_restrictions() or
                                    $resource_type eq "sanfilers7m") {
                                    if (!grep(/^$init_wwpn$/, @{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}})) {
                                        push (@{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}}, $init_wwpn);
                                    }
                                    if (!grep(/^$target_wwpn$/, @{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}})) {
                                        push (@{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}}, $target_wwpn);
                                    }
                                } elsif ($self->error()) {
                                    $self->_set_err("Failed to determine if initiator $init_wwpn should be zoned to target $target_wwpn");
                                    return $SanUtil::FALSE;
                                }
                            } else {
                                # We don't have a mapping for this relationship and we are not dealing with 7mode
                                # controllers so skip this initiator for this node.
                                next;
                            }
                        }
                    } elsif (scalar(@initiators) <= scalar(@super_targets)) {
                        # Case 2 and 3
                        foreach my $init_wwpn (@initiators) {
                            my $client_name  = $resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{name};
                            my $client_has_portset_restrictions = $self->get_client_hash()->{$client_name}->{fcp}->{CONFIG_FCP_FILTER_TARGETS_BY_PORTSET};
                            my $client_zones_per_node = $self->get_client_hash()->{$client_name}->{fcp}->{CONFIG_FCP_ZONES_PER_NODE_PER_INIT};
                            my $retries = 0;

                            # client will by default use the global value
                            my $zones_per_node = $zones_per_node_per_init;

                            # if this client wants portset restrictions and did not specifically request a value for
                            # client_zones_per_node, then we will basically set zones
                            # per node to be the max value. This will force each iteration of this loop
                            # to scan every target on this node for zoning eligibility.
                            if ($client_has_portset_restrictions and !$client_zones_per_node) {
                                $client_zones_per_node = 'max';
                            }

                            # if this client wants to override zones per node
                            if ($client_zones_per_node) {
                                if (lc($client_zones_per_node) eq 'max') {
                                    # this client wants max zones per node per init, calculate the zones per node on the fly
                                    # this will be equal to the number of targets that there are on the current node
                                    $zones_per_node = scalar(@targets);
                                } else {
                                    # client is overriding a non-max value
                                    $zones_per_node = $client_zones_per_node;
                                }
                            }

                            # zone this initiator to this node $zones_per_node_per_init times
                            for (my $count = 1; $count <= $zones_per_node; $count++) {
                                # Grab a target
                                my $target_wwpn = pop(@super_targets);
                                # slip it back in the front
                                unshift(@super_targets, $target_wwpn);

                                # Make sure we have a client->vserver mapping.  We can distinguish between
                                # 7mode controllers and Cmode vservers because "cluster_name" will be "" on
                                # 7mode controllers.
                                my $cluster_name = $target_hash->{$node}->{$vserver}->{$target_wwpn}->{cluster_name};

                                if ($cluster_name eq ""
                                    or exists($client_vserver_map->{$client_name}->{$cluster_name}->{$vserver})) {
                                    # Looks like this set has a map so add it to the resource hash
                                    # first verify if these two wwpns should be zoned together
                                    # if the resource type is a 7mode controller, pair them together
                                    if ($self->should_init_wwpn_be_zoned_to_target_wwpn(-init_wwpn => $init_wwpn,
                                                                                        -target_wwpn => $target_wwpn) or
                                        not $self->are_there_any_igroup_zoning_restrictions() or
                                        $resource_type eq "sanfilers7m") {
                                        if (!grep(/^$init_wwpn$/, @{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}})) {
                                            push (@{$target_hash->{$node}->{$vserver}->{$target_wwpn}->{initiators}}, $init_wwpn);
                                        }

                                        if (!grep(/^$target_wwpn$/, @{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}})) {
                                            push (@{$resource_hash->{$fabric}->{initiators}->{$init_wwpn}->{targets}}, $target_wwpn);
                                        }

                                        # bumping up the retry count makes it so we eventually get out of this loop
                                        $retries++;
                                    } elsif ($self->error()) {
                                        # the previous function call hit an error
                                        $self->_set_err("Failed to determine if initiator $init_wwpn should " .
                                                        "be zoned to target $target_wwpn");
                                        return $SanUtil::FALSE;
                                    } else {
                                        # in this case init wwpn and target wwpn should not get zoned together.
                                        # if the current client that owns this init wwpn has portset zone filtering (and we are restricting by igroup)
                                        # then we may need to take into consideration more ports in the @super_targets list
                                        # the only way to do this is to decrement $count so that it retries the loop.
                                        # if we've retried scalar(@targets) times, then we've exhausted this initiator. Give up
                                        # and go to the next initiator
                                        if ($self->are_there_any_igroup_zoning_restrictions() and $client_has_portset_restrictions) {

                                            if ($retries == scalar(@targets)) {
                                                # give up, this initiator wwpn can't be zoned anymore to this node
                                                last;
                                            }
                                            # bump up the retry and subtract the count so that we can try another port in the @super_targets list
                                            $retries++;
                                            $count--;
                                        }
                                    }
                                } else {
                                    # We don't have a mapping for this relationship and we are not dealing with 7mode
                                    # controllers so skip this initiator for this node.
                                    next;
                                }
                            }
                        }
                    } else {
                        # Unacceptable config
                        $self->_set_err("Invalid combination of initiators and targets for " .
                                         "node '$node' vserver '$vserver' on fabric '$fabric'");
                        self->sanlog(-msg => Dumper($resource_hash));
                        return $SanUtil::FALSE;
                    } ##END if (scalar(@targets) == 0 or scalar(@initiators) == 0) {
                } ##END foreach my $vserver (sort keys %{$target_hash->{$node}}) {
            } ##END  foreach my $node (sort keys %$target_hash) {
        } ##END if ($zones_per_node_per_init eq "max") {

        # Add the target_hash to the resource_hash
        foreach my $node (sort keys %{$target_hash}) {
            foreach my $vserver (sort keys %{$target_hash->{$node}}) {
                foreach my $wwpn (sort keys %{$target_hash->{$node}->{$vserver}}) {
                    $resource_hash->{$fabric}->{targets}->{$wwpn} = $target_hash->{$node}->{$vserver}->{$wwpn};
                }
            }
        }
    } ## end foreach my $fabric (sort keys...

    # Set the resource hash on self
    $self->{resources} = $resource_hash;

    return $SanUtil::TRUE;
} ## end sub build_resource_hash

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _verify_alias
# Objective: verify alias information specified for the fabric with an existing
#            alias object.
# Inputs:    -alias => string containing an alias name
#            -member => string containing the member information
# Outputs:   Success:  $SanUtil::TRUE
#            Failure:  $SanUtil::FALSE and set an error
###############################################################################
sub _verify_alias {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $alias_name = &ParamsMgr::get_arg(-arg => "alias",  -required => 1, -default => "");
    my $member     = &ParamsMgr::get_arg(-arg => "member", -required => 1, -default => "");
    my $fabric     = &ParamsMgr::get_arg(-arg => "fabric", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Get the ZoneMgr object for the fabric
    my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
        return $SanUtil::FALSE;
    }

    # Make sure the alias exists on the fabric. If it doesn't, set an error and bail.
    my $alias_obj = $zonemgr_obj->get_aliases(-alias => $alias_name);
    if ($zonemgr_obj->error()) {
        $self->_set_err(
            "Error while trying to retrieve the object for alias '$alias_name'", $zonemgr_obj->errmsg());
        return $SanUtil::FALSE;
    }
    unless (ref($alias_obj) =~ /alias/i) {

        # Alias doesn't exist. Set an error and bail
        $self->_set_err("Alias '$alias_name' was not found on the fabric");
        return $SanUtil::FALSE;
    } ## end unless (ref($alias_obj) =~...

    # Alias object already exists. Check to make sure the member is correct.
    my $object_member = $alias_obj->get_member();
    unless (defined($object_member) and $object_member ne "") {
        $self->_set_err("Invalid member returned for alias '$alias_name'");
        return $SanUtil::FALSE;
    }
    unless ($object_member eq $member) {
        $self->_set_err("Member on alias object '$object_member' differs from fabric definition '$member' for "
              . "alias '$alias_name'");
        return $SanUtil::FALSE;
    }

    # Validate the alias against the switch
    $zonemgr_obj->alias_verify(-alias => $alias_name);
    if ($zonemgr_obj->error) {
        $self->_set_err("Unable to verify alias '$alias_name'", $zonemgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Everything checked out fine
    return $SanUtil::TRUE;
} ## end sub _verify_alias

###############################################################################
# Method:    _verify_zone
# Objective: verify a zone information specified for the fabric with an existing
#            zone object.
# Inputs:    -target    => string containing a target alias
#            -initiator => string containing a initiator alias
#            -md5       => string containing the md5 hash zone name
#            -zone      => string containing the zone name
#            -fabric    => string containing the fabric name
# Outputs:   Success:  $SanUtil::TRUE
#            Failure:  $SanUtil::FALSE and set an error
###############################################################################
sub _verify_zone {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $target    = &ParamsMgr::get_arg(-arg => "target",    -required => 1, -default => "");
    my $initiator = &ParamsMgr::get_arg(-arg => "initiator", -required => 1, -default => "");
    my $zone_name = &ParamsMgr::get_arg(-arg => "zone",      -required => 1, -default => "");
    my $md5_name  = &ParamsMgr::get_arg(-arg => "md5",       -required => 1, -default => "");
    my $fabric    = &ParamsMgr::get_arg(-arg => "fabric",    -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Get the ZoneMgr object for the fabric
    my $zonemgr_obj = $self->get_zonemgr_object(-fabric => $fabric);
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the ZoneMgr object for fabric '$fabric'");
        return $SanUtil::FALSE;
    }

    # Grab the zone object off of the zonemgr object
    my $zone_obj = $zonemgr_obj->get_zones(-zone => $md5_name);
    if ($zonemgr_obj->error()) {
        $self->_set_err("Error while trying to retrieve the object for zone '$md5_name ($zone_name)'",
                        $zonemgr_obj->errmsg());
        return $SanUtil::FALSE;
    }
    unless (ref($zone_obj) =~ /zone/i) {
        $self->_set_err("Zone '$md5_name ($zone_name)' was not found on fabric '$fabric'");
        return $SanUtil::FALSE;
    }

    # We know the zone exists.  Validate the zone members against the fabric specification.
    my $aliases_ref = $zone_obj->get_aliases();
    unless (scalar(keys %{$aliases_ref}) > 0) {
        $self->_set_err("Zone object for zone '$md5_name ($zone_name)' does not have any aliases defined");
        return $SanUtil::FALSE;
    }
    unless (scalar(keys %{$aliases_ref->{$initiator}}) > 0) {
        $self->_set_err("Initiator alias not found in existing zone object for zone '$md5_name ($zone_name)'");
        return $SanUtil::FALSE;
    }
    unless (scalar(keys %{$aliases_ref->{$target}}) > 0) {
        $self->_set_err("Target alias not found in existing zone object for zone '$md5_name ($zone_name)");
        return $SanUtil::FALSE;
    }

    # Aliases look good.  Validate it.
    $zonemgr_obj->zone_verify(-zone => $md5_name);
    if ($zonemgr_obj->error()) {
        $self->_set_err("Unable to verify zone '$md5_name ($zone_name)'", $zonemgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Everything checked out fine
    return $SanUtil::TRUE;
} ## end sub _verify_zone

###############################################################################
# Method:    _get_alias_name
# Objective: Take a host or controller name and device name and create an alias
# Inputs:    -host   => string containing a host or controller name
#            -device => string containing a device name (ie emlxs0, 0a, etc)
# Outputs:   Returns a string containing the alias name.  If there is an error,
#            set error and return ""
###############################################################################
sub _get_alias_name {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $host    = &ParamsMgr::get_arg(-arg => "host",    -required => 1, -default => "");
    my $device  = &ParamsMgr::get_arg(-arg => "device",  -required => 1, -default => "");
    my $prefix  = &ParamsMgr::get_arg(-arg => "prefix",  -required => 1, -default => "");
    my $cluster = &ParamsMgr::get_arg(-arg => "cluster", -required => 0, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return "";
    }

    # Convert dashes and periods to underscores for the host, device, and cluster
    $host    =~ s/-|\./_/g;
    $device  =~ s/-|\./_/g;
    $cluster =~ s/-|\./_/g;

    # Build the alias name.  The format is prefix_host_device or prefex_cluster_host_device;
    my $alias;
    if ($cluster eq "") {
        $alias = $prefix . "_" . $host . "_" . $device;
    } else {
        $alias = $prefix . "_" . $cluster . "_" . $host . "_" . $device;
    }

    return $alias;
} ## end sub _get_alias_name

###############################################################################
# Method:    _get_zone_name
# Objective: Take an initiator alias and a target alias and create the zone name.
# Inputs:    -initiator => string containing an initiator alias
#            -target    => string containing a target alias
#
# Outputs:   Returns a string containing the zone name or "" if there is an error.
###############################################################################
sub _get_zone_name {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $init   = &ParamsMgr::get_arg(-arg => "initiator", -required => 1, -default => "");
    my $target = &ParamsMgr::get_arg(-arg => "target",    -required => 1, -default => "");
    my $prefix = &ParamsMgr::get_arg(-arg => "prefix",    -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return ("", "");
    }

    # Strip the prefix off of the initiator and target
    $init   =~ s/^${prefix}_+//g;
    $target =~ s/^${prefix}_+//g;

    # Convert dashes and periods to underscores for the initiator and target
    $init   =~ s/-|\./_/g;
    $target =~ s/-|\./_/g;

    # Build the zone name.  The format is prefix__initiator_device__target_device;
    my $zone       = $init . "__" . $target;
    my $clear_zone = $prefix . "__" . $zone;

    my $md5;
    if (length($md5 = md5_hex($zone)) == 0) {
        $self->_set_err("Unable to generate MD5 Hash for zone $clear_zone");
        return ("", "");
    }

    # Create the md5 version of the zone
    my $md5_zone = "$prefix" . "_" . $md5;

    return ($clear_zone, $md5_zone);
} ## end sub _get_zone_name

###############################################################################
# Method:    _validate_fabric_switch_objects
# Objective: Make sure that each switch in the fabric hash has an object in the switch hash.
# Inputs:    None
# Outputs:   Returns TRUE on success or sets and error and returns FALSE
###############################################################################
sub _validate_fabric_switch_objects {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    # Get the fabric hash
    my $fabric_hash = $self->get_fabric_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the fabric hash");
        return $SanUtil::FALSE;
    }

    # Get the switch hash
    my $switch_hash = $self->get_switch_hash();
    if ($self->error()) {
        $self->_set_err("Unable to retrieve the switch hash");
        return $SanUtil::FALSE;
    }

    # First check that each fabric switch has an object in the switch hash
    foreach my $fabric (keys %$fabric_hash) {
        foreach my $switch (keys %{$fabric_hash->{$fabric}->{switches}}) {
            unless (defined($switch_hash->{$switch}) and ref($switch_hash->{$switch}) =~ /sanswitch/i) {
                $self->_set_err("No SanSwitch object found for switch $switch");
                return $SanUtil::FALSE;
            }
        } ## end foreach my $switch (keys %{...
    } ## end foreach my $fabric (keys %$fabric_hash)

    return $SanUtil::TRUE;
} ## end sub _validate_fabric_switch_objects

###############################################################################
# Method:    _get_client_initiators
# Objective: Take in a client name and parse search the ParamMgr data looking
#            for any defined client initiators.  Return a hash reference of anything
#            found.  Do not set an error if none are found, just return an empty
#            hash.  Set an error only if invalid arguements are specified.
# Inputs:    -client => string containing a client to search for
#
# Outputs:   Returns a hash reference containing the wwpns of the initiators found
#            or an empty reference.  Sets an error only if an invalid argument is
#            specified
###############################################################################
sub _get_client_initiators {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $client = &ParamsMgr::get_arg(-arg => "client", -required => 1, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # Create a SUPER hash of clients and hypervisors
    my $xml_hash = {};

    # Get the client info from the XML
    $xml_hash->{CLIENTS} = &ParamsMgr::search(
        -search_path => [],
        -key         => "CLIENTS",
        -key_filter  => qr/^CLIENT\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err("Error searching XML", &ParamsMgr::get_error());
        return {};
    }

    # Get the hypervisor info from the XML
    $xml_hash->{HYPERVISORS} = &ParamsMgr::search(
        -search_path => [],
        -key         => "HYPERVISORS",
        -key_filter  => qr/^ESX\d*$|^HYPERV\d*$/,
        paramset => $self->EnGenX::Util::get_paramset()
    );
    if (&ParamsMgr::error()) {
        $self->_set_err("Error searching XML", &ParamsMgr::get_error());
        return {};
    }

    # Process the xml_hash
    foreach my $client_type (sort keys %$xml_hash) {
        foreach my $xml_client (sort keys %{$xml_hash->{$client_type}}) {
            unless (defined($xml_hash->{$client_type}->{$xml_client}->{NAME})
                and $xml_hash->{$client_type}->{$xml_client}->{NAME} eq $client) {
                next;
            }

            # Looks like we found the client so let's see if there are any intiators
            my $initiators_hash = &ParamsMgr::search(
                -search_path => [],
                -exact_key   => 1,
                -key         => "${client_type}.${xml_client}.INITIATORS",
                -key_filter  => qr/^WWPN\d*$/,
                paramset => $self->EnGenX::Util::get_paramset()
            );

            my $hash_ref = {};
            if (scalar(%$initiators_hash) > 0) {

                # We found initiators so process them.
                foreach my $init (keys %$initiators_hash) {
                    if (defined($initiators_hash->{$init}->{VALUE})
                        and $initiators_hash->{$init}->{VALUE} ne "") {

                        my $wwpn = $initiators_hash->{$init}->{VALUE};

                        # Strip out the colons and lowercase the wwpn
                        $wwpn =~ s/\://g;
                        $wwpn = lc($wwpn);

                        # Add the initiator to the hash reference
                        $hash_ref->{$wwpn} = 1;
                    } ## end if (defined($initiators_hash...
                } ## end foreach my $init (keys %$initiators_hash)
            } ## end if (scalar(%$initiators_hash...
            return $hash_ref;
        }
    }

    # We didn't find anything so just return an empty hash reference.
    return {};
} ## end sub _get_client_initiators

###############################################################################
# Method:    _transform_fabrics_xml
# Objective: Take in a ParamsMgr xml hash and transform it into an easier to use
#            hash.
# Inputs:    -all_fabrics => all fabrics hash built from xml
#            -fabric      => fabric name
#
# Outputs:   Returns a hash reference. Sets an error on self if an error occurs
###############################################################################
sub _transform_fabrics_xml {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);

    # Get some args
    my $all_fabrics = &ParamsMgr::get_arg(-arg => "all_fabrics", -required => 1, -default => {});
    my $fabric      = &ParamsMgr::get_arg(-arg => "fabric", -required => 0, -default => "");

    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return {};
    }

    # Massage the all_fabrics hash and create a new hash that looks like:
    #
    # $fabrics =>
    #             <fabric_name> =>
    #                   zones_per_node_per_init => <brocade virtual fabric id> or undef
    #                   master_switch           => switch to run commands on
    #                   prefix                  => prefix
    #                   switches                =>
    #                           <switch_name> =>
    #                                   admin_domain            => <admin domain> or undef
    #                                   virtual_fabric          => <cisco virtual fabric> or undef
    #                                   fabric_id               => <brocade virtual fabric id> or undef
    #                           <switch_name> =>
    #                                   admin_domain            => <admin domain> or undef
    #                                   virtual_fabric          => <cisco virtual fabric> or undef
    #                                   fabric_id               => <brocade virtual fabric id> or undef
    my $fabrics = {};
    foreach my $xml_fabric (sort keys %$all_fabrics) {
        unless (defined($all_fabrics->{$xml_fabric}->{NAME}) and $all_fabrics->{$xml_fabric}->{NAME} ne "") {
            $self->_set_err("Invalid values specified for fabric 'name' tag in XML");
            return {};
        }

        my $name = $all_fabrics->{$xml_fabric}->{NAME};

        # If $fabric is something other than "", check it against $name and only continue
        # if they match.  If it was provided then the caller only wants that fabric in the hash.
        if ($fabric ne "" and $fabric ne $name) {
            next;
        }

        # initialize some variables
        $fabrics->{$name}->{zones_per_node_per_init}  = "1";
        $fabrics->{$name}->{master_switch}            = "";
        $fabrics->{$name}->{prefix}                   = "nate";

        # Store all the attributes in an array so we can iterate over
        # them easier.
        my @attributes = ( "zones_per_node_per_init",
                           "master_switch",
                           "prefix",
                         );

        # Extract the attributes
        #foreach my $key (sort keys %{$all_fabrics->{$xml_fabric}}) {
        foreach my $attr (sort @attributes) {
            # We need to uppercase the attribute because the all_fabrics hash has
            # the keys in uppercase since they are a paramsmgr grab.
            my $uc_attr = uc($attr);
            if (defined($all_fabrics->{$xml_fabric}->{$uc_attr})
                and $all_fabrics->{$xml_fabric}->{$uc_attr} ne "") {
                # Copy the attribute data out of the all_fabrics xml hash to the
                # fabric hash.
                $fabrics->{$name}->{$attr} = $all_fabrics->{$xml_fabric}->{$uc_attr};
            }
        }

        # Check to make sure the prefix didn't end up with an underscore
        $fabrics->{$name}->{prefix} =~ s/_+$//g;

        # Get the switch Information
        my $switches = &ParamsMgr::search(
            -search_path => [],
            -exact_key   => 1,
            -key         => "FABRICS.${xml_fabric}",
            -key_filter  => qr/^SWITCH\d*$/,
            paramset => $self->EnGenX::Util::get_paramset()
        );

        unless (scalar(%$switches) > 0) {
            $self->_set_err("Invalid values specified for 'SWITCH' tag in XML");
            return {};
        }

        # parse through all of the switches in the hash and add them to the fabrics hash
        foreach my $switch (sort keys %$switches) {
            unless (defined($switches->{$switch}->{NAME}) and $switches->{$switch}->{NAME} ne "") {
                $self->_set_err("Invalid values specified for switch 'name' tag in XML");
                return {};
            }
            my $switch_name = $switches->{$switch}->{NAME};

            $fabrics->{$name}->{switches}->{$switch_name}->{admin_domain}   = "";
            $fabrics->{$name}->{switches}->{$switch_name}->{fabric_id}      = "";
            $fabrics->{$name}->{switches}->{$switch_name}->{virtual_fabric} = "";

            my @switch_attributes = ( "admin_domain",
                                      "fabric_id",
                                      "virtual_fabric",
                                    );

            # Check to see if admin_domain, virtual_fabric, or fabric_id have been defined
            # at the switch level.  If so, we will use that value.  If not, check to see
            # if it is defined at the fabric level and use that one instead.
            foreach my $attr (sort @switch_attributes) {
                # We need to uppercase the attribute because the all_fabrics hash has
                # the keys in uppercase since they are a paramsmgr grab.
                my $uc_attr = uc($attr);
                if (defined($switches->{$switch}->{$uc_attr})) {
                    # Copy the attribute data out of switch hash  from the xml file
                    # into the fabric hash fabric hash.
                    $fabrics->{$name}->{switches}->{$switch_name}->{$attr} = $switches->{$switch}->{$uc_attr};
                } elsif (defined($all_fabrics->{$xml_fabric}->{$uc_attr})) {
                    $fabrics->{$name}->{switches}->{$switch_name}->{$attr} = $all_fabrics->{$xml_fabric}->{$uc_attr};
                }
            }
        } ## end foreach my $switch (sort keys...
    } ## end foreach my $xml_fabric (sort keys %$all_fabrics) {

    return $fabrics;
} ## end sub _transform_fabrics_xml

###############################################################################
# Method:    _load_xml
# Objective: Load XML file into ParamSet via ParamsMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _load_xml {
    my ($self, %args) = @_;

    # Clear out the errors
    $self->_clear_err();

    &ParamsMgr::process_args(%args);
    my $file        = &ParamsMgr::get_arg(-arg => "file",        -required => 1, -default => "");
    my $schema_file = &ParamsMgr::get_arg(-arg => "schema_file", -required => 0, -default => "");
    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Load up the XML into ParamsMgr and ParamSet
    &ParamsMgr::load_xml(-xml_file => $file,
                         -display_xml => $SanUtil::FALSE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Args parsing failed", &ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    return $SanMgr::TRUE;
} ## end sub _load_xml

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

FabricMgr.pm

=head1 Author

Derrick Shultz

=head1 Description

FabricMgr.pm provides an API interface to allow for zoning and unzoning fabric components.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<ParamsMgr, SubIPC, San::FabricMgr::DbUtil, San::SanLog, San::SanUtil, Tharn, Error, San::SanSwitch::SwitchUtil,
San::SanSwitch::ZoneMgr, Digest::MD5>

=head1 Attributes

=head2 B<fabric>

string = Fabric Name (optional)

=head2 B<file>

string = XML File to parse (optional)

=head2 B<switch_objs>

array reference = Reference to an array of switch objects

=head2 B<client_objs>

array reference = Reference to an array of client objects

=head2 B<sanfiler7m_objs>

array reference = Reference to an array of SanFiler7m objects

=head2 B<cluster_objs>

array reference = Reference to an array of NACL::C::Cluster objects

=head2 B<protocols>

array reference = Reference to an array of protocols to scope the build switch info
hash phase to only particular protocols.

=head1 Public Methods

=head2 B<new()>

Creates a new FabricMgr object. Discovery is not performed at this time.

=over 4

=item Inputs

=item * I<-fabric>

Individual fabric name to operate on.  If no fabric is specified, all fabrics found under the FABRICS
XML tag will be processed.

=item * I<-file>

XML file to use.  If no file is specied it will expect to find the FABRICS already loaded into the
environment.

=item * I<-switch_objs>

Reference to an array of switch objects. At least one switch object is required.

=item * I<-client_objs>

Reference to an array of client objects.  At least one client object is required.

=item * I<-sanfiler7m_objs>

Reference to an array of SanFiler7m objects. Sanfiler7m objects are not required; however, you must
specify at least one Sanfiler7m object or one ControllerCluster object.

=item * I<-cluster_objs>

Reference to an array of ControllerCluster objects.

=item * I<-nacl_cluster_objs>

Reference to an array of NACL::C::Cluster objects.

=item * I<-protocols>

Reference to an array of protocols to scope FabricMgr discovery to.
Default: ["fcp","iscsi","nfs","cifs","cluster"]

=item Outputs

Returns a FabricMgr object.

=item Examples

 # Create a FabricMgr object with xml found in environment with
 # only Clients, Switches, and ControllerClusters.  Also specify
 # a fabric name.  This is how SanMgr would call it.


 $FabricMgr_Obj = new FabricMgr( -fabric          => "fabric1",
                                 -switch_objs     => \@switch_objects,
                                 -client_objs     => \@client_objects,
                                 -cluster_objs    => \@cluster_objects,
                                );

 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Create a FabricMgr object by specifying and XML file and not limiting
 # it to one fabric.

 $FabricMgr_Obj = new FabricMgr(-file           => "<path_to_xml_file>",
                                -switch_objs     => \@switch_objects,
                                -client_objs     => \@client_objects,
                                -cluster_objs    => \@cluster_objects,
                                -sanfiler7m_objs => \@filer_objects,
                               );

 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<activate_fabric_changes()>

Activate the fabric changes for all of the fabrics found in the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on successful activation.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->activate_fabric_changes();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<build_client_vserver_map()>

Builds a hash representing the relationship between clients and vservers and stores it on the object.
Only supported with C-mode.

The format of the hash is as follows:

$hash->{<wwpn>}->{<cluster_name>}->{vserver_name} = arbitrary count.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->build_client_vserver_map();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<build_initiator_wwpn_to_target_wwpn_map()>

Builds a hash lookup table that describes which initiators should be zoned to which targets.
Only supported with C-mode.

The format of the hash is as follows:

$hash->{<init_wwpn>}->{<target_wwpn>} = arbitrary count

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->build_initiator_wwpn_to_target_wwpn_map();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<build_resource_hash()>

Builds a hash all of the resources in each fabric that can be zoned together.  The hash is then
stored on the the object.

The format of the hash is as follows:

$resource_hash =>
     <fabric_name> =>
         initiators =>
             <wwpn> =>
                  name         => <client_name>
                  cluster_name => <client_name>
                  device       => <hba device name and instance>
                  type         => <pwwn|fwwn>
                  targets      => <array of target wwpns to zone>
         targets =>
             <wwpn> =>
                  name       => Vserver Name or Node Name in 7mode
                  node       => Node Name
                  device     => <hba device name and instance>
                  cluster    => Cluster name or ""
                  type       => <pwwn|fwwn>
                  initiators => <array of initiator wwpns to zone>

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->build_resource_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<build_switch_info_hash()>

Builds a hash all of the resources discovered out on all of the switches.  This information includes
what switch, switch port, fabric, etc the resources are found in.  Multiple views of of the environment
are available from within the hash.  Hash is stored on the object.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->build_switch_info();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<build_zonemgr_objects()>

Builds outs all of the ZoneMgr objects for each fabric being configured.  These objects are then
stored on the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->build_zonemgr_objects();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<create_aliases()>

Creates all of the aliases required on the FabricMgr object.  Any existing aliases that
match the name will be verified. If they are correctly defined, they will be reused.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->create_aliases();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<create_zones()>

Creates all of the zones required on the FabricMgr object.  Any existing zones that
match the name will be verified. If they are correctly defined, they will be reused.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->create_zones();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<delete_aliases()>

Deletes all of the aliases on the FabricMgr object. If an alias is still
in use by a zone, it will not be deleted.  This is not considered an error since
an alias could be used in other zones not in this fabric.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->delete_aliases();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<delete_zones()>

Deletes all of the zones on the FabricMgr object

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->delete_zones();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

head2 B<dump_hashes()>

Dumps the various hashes used by FabricMgr including the switch info hash, the resource hash, client vserver
map, and the fabric hash.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->dump_hashes();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_active_zoneset()>

Returns the active zoneset for the fabric specified.

=over 4

=item Inputs

=item * I<-fabric>

Name of the fabric to use when searching for the active zoneset.

=item Outputs

Returns a string containing the name of the active zoneset.  If there is a problem, an error is raised
and "" is returned

=item Examples

 my $active_zoneset = $FabricMgr_Obj->get_active_zoneset(-fabric => $fabric);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_client_hash()>

Returns the hash of client objects from the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to a hash containing all of the client objects off of the FabricMgr object.  The key is the client
name and the value is the object.  If there is a problem, an error is raised and {} is returned

=item Examples

 my $client_hash = $FabricMgr_Obj->get_client_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_client_vserver_map()>

Returns the client vserver map hash.

=over 4

=item Inputs


=item * I<-refresh> (Bool: 0|1  Default: 0);

If -refresh is specified and set to 1, the client vserver map hash will be rebuilt prior to being returned.

=item Outputs

Returns a reference to a hash containing the client vserver map.  For the format of the hash, see the
build_client_vserver_map method. If there is a problem, an error is raised and {} is returned

=item Examples

 # Grab the client vserver map

 my $client_vserver_hash = $FabricMgr_Obj->get_client_vserver_map();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Grab the client vserver map with a forced refresh

 my $client_vserver_hash = $FabricMgr_Obj->get_client_vserver_map(-refresh => 1);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_initiator_wwpn_to_target_wwpn_map()>

Returns the initiator wwpn to target wwpn map hash. Only supported with C-mode.

=over 4

=item Inputs


=item * I<-refresh> (Bool: 0|1  Default: 0);

If -refresh is specified and set to 1, the map hash will be rebuilt prior to being returned.

=item Outputs

Returns a reference to a hash containing the map.  For the format of the hash, see the
build_intiiator_wwpn_to_target_wwpn_map method. If there is a problem,
an error is raised and {} is returned.

=item Examples

 # Grab the _initiator_wwpn_to_target_wwpn_map

 my $map = $FabricMgr_Obj->get_initiator_wwpn_to_target_wwpn_map();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Grab the map with a forced refresh

 my $map = $FabricMgr_Obj->get_initiator_wwpn_to_target_wwpn_map(-refresh => 1);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<should_init_wwpn_be_zoned_to_target_wwpn()>

Returns $SanUtil::TRUE or $SanUtil::FALSE based on whether these two WWPNs should be zoned.
Only supported with C-mode.

=over 4

=item Inputs


=item * I<-init_wwpn> (Required)

=item * I<-target_wwpn> (Required)

=item Outputs

$SanUtil::TRUE or $SanUtil::FALSE

=item Examples

 # Grab the _initiator_wwpn_to_target_wwpn_map

 if ($FabricMgr_Obj->should_init_wwpn_be_zoned_to_target_wwpn(-init_wwpn => $init_wwpn,
                                                              -target_wwpn => $target_wwpn)) {
     # zone them
 } elsif ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;

 } else {
     # dont zone them
 }

=back

=head2 B<are_there_any_igroup_zoning_restrictions()>

Returns $SanUtil::TRUE or $SanUtil::FALSE based on whether there are any igroup zoning restrictions established by the object

=over 4

=item Inputs

None

=item Outputs

$SanUtil::TRUE or $SanUtil::FALSE

=item Examples

 if ($FabricMgr_Obj->are_there_any_igroup_zoning_restrictions()) {
     # do something
 } else {
     # dont do what you wanted to do
 }

=back

=head2 B<get_cmode_cluster_hash()>

Returns the hash of ControllerCluster objects from the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to a hash containing all of the ControllerCluster objects off of the FabricMgr object.
The key is the cluster name and the value is the object.  If there is a problem, an error is raised and {} is returned.

=item Examples

 my $cmode_hash = $FabricMgr_Obj->get_cmode_cluster_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_cmode_vserver_list()>

Returns the array of cmode vserver names from the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to an array containing all of the CMode Vserver names off of the FabricMgr object.
If there is a problem, an error is raised and [] is returned.

=item Examples

 my $cmode_vservers = $FabricMgr_Obj->get_cmode_vserver_list();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_fabric()>

Returns fabric name off of the FabricMgr object if one was passed in on the constructor.

=over 4

=item Inputs

None

=item Outputs

Returns a string containing the name of the fabric if one was passed in, otherwise it
returns "". No error is set.

=item Examples

 my $fabric = $FabricMgr_Obj->get_fabric();
 if (defined($fabric) and $fabric ne "") {
     print "Fabric is $fabric\n";
     return 1;
 }

=back

=head2 B<get_fabric_hash()>

Returns the hash of fabrics from the FabricMgr object that were discovered in the XML.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to a hash containing all of the fabrics off of the FabricMgr object.
The key is the cluster name and the value is the object.  If there is a problem, an error
is raised and {} is returned.

Format of the hash is:

$hash
    => <fabric_name>
        => targets
            => <wwpn>
                => name
                => type
                => cluster_name
                => device
        => initiators
            => <wwpn>
                => name
                => type
                => clusters
                => cluster_name
                => device

=item Examples

 my $fabric_hash = $FabricMgr_Obj->get_fabric_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_prefix()>

Returns the prefix for the fabric specified.

=over 4

=item Inputs

=item * I<-fabric>

Name of the fabric to use

=item Outputs

Returns a string containing the prefix for the given fabric.  If there is a problem, an
error is raised and "" is returned

=item Examples

 my $prefix = $FabricMgr_Obj->get_prefix(-fabric => $fabric);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_resource_hash()>

Returns the resource hash off of the FabricMgr object.  This hash is created by the build_resource_hash() method.

=over 4

=item Inputs

=item * I<-refresh> (Bool: 0|1  Default: 0);

If -refresh is specified and set to 1, the resource hash will be rebuilt prior to being returned.

=item Outputs

Returns a reference to the resource hash off of the FabricMgr object.
If there is a problem, an error is raised and {} is returned.

=item Examples

 # Grab the resource hash

 my $resource_hash = $FabricMgr_Obj->get_resource_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Grab the resource hash with a forced refresh

 my $resource_hash = $FabricMgr_Obj->get_resource_hash(-refresh => 1);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_sanfiler7m_hash()>

Returns the hash of SanFiler7m objects from the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to a hash containing all of the SanFiler7m objects off of the FabricMgr object.
The key is the filer name and the value is the object.  If there is a problem, an error is raised and {} is returned.

=item Examples

 my $sanfiler_hash = $FabricMgr_Obj->get_sanfiler7m_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_switch_hash()>

Returns the hash of SanSwitch objects from the FabricMgr object.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to a hash containing all of the SanSwitch objects off of the FabricMgr object.
The key is the switch name and the value is the object.  If there is a problem, an error is raised and {} is returned.

=item Examples

 my $switch_hash = $FabricMgr_Obj->get_switch_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_switch_fabric()>

Returns name of the fabric for a given switch.

=over 4

=item Inputs

=item * I<-switch>

Name of the switch to return the fabric for.

=item * I<-fid>

FID or VSAN ID

=item * I<-ad>

Reference to an array of admin domains

=item Outputs

Returns a string containing the name of the fabric for the switch.  If no
fabric is defined, return "". If there is a problem, an error is raised and "" is returned.

=item Examples

 my $fabric = $FabricMgr_Obj->get_switch_fabric(-switch => $switch_name, -fid => 256);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 if (defined($fabric) and $fabric ne "") {
    print "Fabric is $fabric.\n";
 } else {
    print "Fabric is not defined.\n";
 }

=back

=head2 B<get_switch_info_hash()>

Returns the switch info hash off of the FabricMgr object.  This hash is created by
the build_switch_info_hash() method.

=over 4

=item Inputs

=item * I<-refresh> (Bool: 0|1  Default: 0);

If -refresh is specified and set to 1, the switch info hash will be rebuilt prior to being returned.

=item Outputs

Returns a reference to the switch info hash off of the FabricMgr object.
If there is a problem, an error is raised and {} is returned.

=item Examples

 # Grab the switch info hash

 my $info_hash = $FabricMgr_Obj->get_switch_info_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Grab the switch info hash with a forced refresh

 my $info_hash = $FabricMgr_Obj->get_switch_info_hash(-refresh => 1);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_vservers_associated_to_client()>

Returns has hash of vservers for the given initiator.

=over 4

=item Inputs

=item * I<-client>

Client name to return the vserver map for.

=item Outputs

Returns the vserver map hash for the given initiator.  If no vservers are found
for the initiator, then {} is returned. If there is a problem, an error is raised and {} is returned.

The format of the hash is:

$hash->{<cluster_name>}->{<vserver_name>} = integer.

=item Examples

 my $vserver_hash = $FabricMgr_Obj->get_vservers_associated_to_client(-client => $client);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_protocol_list()>

Returns an array or array reference of protocols used by this instance of FabricMgr

=over 4

=item Inputs

None

=item Outputs

Array reference or array of protocols

=item Examples

 if (grep(/iscsi/i, $FabricMgr_Obj->get_protocol_list())) {
     # do iscsi stuff
 }

=back

=head2 B<get_wwpn_name()>

Returns the name of the resource for the given wwpn

=over 4

=item Inputs

=item * I<-wwpn>

WWPN to return the name of.

=item Outputs

Returns a triplet of strings containing the name of the resource with the given wwpn, the cluster
it belongs to, and the fabric.  If no resource is found is defined, then ("", "", "") is returned.
If there is a problem, an error is raised and ("", "", "") is returned.

=item Examples

 my ($name, $cluster, fabric) = $FabricMgr_Obj->get_wwpn_name(-wwpn => $wwpn);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_zonemgr_hash()>

Returns the zonemgr hash off of the FabricMgr object.  This hash is created by
the build_zonemgr_objects() method.

=over 4

=item Inputs

=item * I<-refresh> (Bool: 0|1  Default: 0);

If -refresh is specified and set to 1, the zonemgr hash will be rebuilt prior to being returned.

=item Outputs

Returns a reference to the zonemgr object hash off of the FabricMgr object.
If there is a problem, an error is raised and {} is returned.

=item Examples

 # Grab the zonemgr object hash

 my $zonemgr_hash = $FabricMgr_Obj->get_zonemgr_hash();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

 # Grab the zonemgr object hash with a forced refresh

 my $zonemgr_hash = $FabricMgr_Obj->get_zonemgr_hash(-refresh => 1);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<get_zonemgr_object()>

Returns the zonemgr object for the given fabric.

=over 4

=item Inputs

=item * I<-fabric>

Fabric to return the zonemgr object for.

=item Outputs

Returns the zonemgr object for the given fabric.
If there is a problem, an error is raised and {} is returned.

=item Examples

 my $zonemgr_obj = $FabricMgr_Obj->get_zonemgr_object(-fabric => $fabric);
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=head2 B<verify_fabric()>

Verifies the all of the resources found in the fabric using the ZoneMgr library calls.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE on success.  If there is a problem, an error is raised
and $SanUtil::FALSE is returned.

=item Examples

 $FabricMgr_Obj->verify_fabric();
 if ($FabricMgr_Obj->error) {
     print $FabricMgr_Obj->errmsg;
     return 0;
 }

=back

=cut
