#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/San.pm#1 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a San machine
## @author pressley@netapp.com
## @status public
## @pod here

package San;

#
# Compiler directives.
#
use strict;

#
# Module import.
#
use San::SanClient;
use San::SanFiler;
use San::SanConfig;
use San::SanLog;
use San::SanUtil;
use Utils qw(make_host make_conn);
use Error;
use Exporter;
use File::Basename;
use ha_blocks_filer qw(ha_filer_parse_lun_list);

use vars qw(
          @EXPORT
          @ISA
          );

#
# Inherit from the Error class
#
@ISA = qw(Error SanLog);

########################################################################
#                            Public Methods
########################################################################

########################################################################
# Method:    new
# Objective: Create new San object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub new
{
    my ($proto, $log_handle) = @_;
    my $class = ref($proto) || $proto;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self, $class;

    #
    # A SAN will contain filers clients and switches.  Create empty arrays to
    # for each of these attributes.  The intent is for each of the elements in
    # these arrays to be references to the proper objects.
    #
    $self->{filers}   = [];
    $self->{clients}  = [];
    $self->{switches} = [];

    #
    # Generally, if we set stuff up, we will want to cleanup.  We will
    # put the cleanup actions needed in this array.  Each array element
    # should be of this form:  <client|filer> <name> <action>
    # For example: filer rtpqa-fas3020-3 lun unmap lun_path lun_map
    #
    $self->{cleanup} = [];

    #
    # If we get a list of vols to parse and setup, we are going to
    # parse and then save in the object.  Each array element will be
    # a hash that contains these fields:
    #   vol_count, vol_size, lun_count, lun_size, first_lun_id
    #   vol_type, vol_prefix, aggr, vol_options, qtree_prefix,
    #   qtree_count
    #
    $self->{filer_vol_list} = [];

    #
    # As we create luns on the filer, we store these objects
    # because we will need them when we config the client.
    #
    $self->{config_lun_objs} = [];

    #
    # Set the debug_level.  The higher the level, the more information the user
    # should expect.
    #
    $self->{debug_level}   = 0;

    #
    # Show command mode.  By default when printing, the command
    # issued will be displayed.  This can be turned off by the user.
    #
    $self->{show_cmd} = 1;

    #
    # Allow CLI commands to be formed, but not executed.
    #
    $self->{execute} = 1;

    #
    # Set a default timeout to be 30 minutes
    #
    $self->{timeout} = 1800;

    #
    # Log Handle (TBD how do we default this to STDOUT?)
    #
    $self->{log}       = $log_handle;
    $self->{no_indent} = undef;
    $self->{call_stack_on_error} = 0;
    $self->{setup_igroups} = 1;
    $self->{setup_luns}    = 1;

    #
    # For Shared Lun Creation set default to "no"
    # Shared Luns will be mapped to more than one igroup
    #
    $self->{shared_lun} = "no";

    #
    # Error message is not initially set, this will set the
    # following inherited attributes:
    #   $self->{error}  = 0;
    #   $self->{errmsg} = '';
    #
    $self->_clear_err();

    return $self;
}  # end of new method

########################################################################
# Method:    setup
# Objective: Initiailize objects Filer and Client objects and setup.
# Details:   See POD documentation at the end of this file.
# Notes:
#            We will determine the type of volume to used based on the
#            aggr argument.  If ware given an aggr, we will use a flex
#            vol otherwise we will use a trad vol.
########################################################################
sub setup
{
    my($self,%args) = @_;

    my $filers;               # arg - filer names (filer1:filer2:...)
    my @filerArray;           # Array split from $filers
    my $clients;              # arg - client names (clint1:client2:...)
    my @clientArray;          # Array split from $clients
    my $switches;             # arg - switch names (switch1:switcht2:...)
    my @switchArray;          # Array split from $switches
    my $filer_vol_list;       # arg - vols and luns to create
    my $lun_list;             # arg - luns to create (1,1g:2-4:2g...)
    my $protocol    = "fcp";  # arg - iscsi | fcp (default)
    my $device_type = "raw";  # arg - raw | vm | fs (raw is default)
    my $aggr;                 # arg - aggregate to optionally create
    my $vol;                  # arg - volume to create
    my $num_disks   = 5;      # arg - disks in aggr or trad vol
    my $vol_size    = "100g"; # arg - size of flex vol
    my $skip_setup    = 0;    # arg - do nothing and return?
    my $skip_config   = 1;    # arg - do nothing and return?
    my $setup_igroups = 1;    # arg - create igroups and lun maps?
    my $setup_luns    = 1;    # arg - create luns?
    my $lun_count     = 0;    # how many luns were created per client
    my @orphaned_clients;     # list of clients already cleaned.
    my %lvm_setup_args;       # arguments to send to _lvm_setup()
    my %devmgr_setup_args;    # pass through to DevMgr
    my $mp_type       = "none";
    my $use_vol_list  = 0;
    my $alua_support  = 0;
    my $ntp_sync      = 1;
    my $ntp_time_server = "timehost1.eng.netapp.com";
    my $ntp_time_zone   = "US/Eastern";
    my $client_pass_thru = {};
    my $filer_pass_thru  = {};
    my @this_filer_targets = ();
    my @targets = ();
    my $cf_mode;
    my $vfiler_enable = 0;
    my $any_iscsi;

    $self->{alua_support}    = $alua_support;
    $self->{ntp_sync}        = $ntp_sync;
    $self->{ntp_time_server} = $ntp_time_server;
    $self->{ntp_time_zone}   = $ntp_time_zone;

    my $create_fs = 0;
    my $create_fs_type;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?filers/)
        {
            $filers = $args{$_};
            if (not length($filers))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-filers'");
                return $self;
            }
            next;
        }
        elsif (/-?clients/)
        {
            $clients = $args{$_};
            if (not length($clients))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-clients'");
                return $self;
            }
            next;
        }
        elsif (/-?shared_lun/)
	{
            my $shared_lun = $args{$_};
            if (not length($shared_lun)) {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-shared_lun'");
                return $self;
            }
	    $self->{shared_lun} = $shared_lun;
            next;
        }
        elsif (/-?switches/)
        {
            $switches = $args{$_};
            if (not length($switches))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-switches'");
                return $self;
            }
            next;
        }
        elsif (/-?filer_vol_list/)
        {
            $filer_vol_list = $args{$_};
            $use_vol_list = 1;
            if (not length($filer_vol_list))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-filer_vol_list'");
                return $self;
            }
            next;
        }
        elsif (/-?lun_list/)
        {
            $lun_list = $args{$_};
            if (not length($lun_list))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-lun_list'");
                return $self;
            }
            next;
        }
        elsif (/-?protocol/)
        {
            $protocol = $args{$_};
            if (not length($protocol) or $protocol !~ /fcp|iscsi/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        }
        elsif (/-?device_type/)
        {
            $device_type = $args{$_};
            if (not length($device_type) or $device_type !~ /raw|lv|fs/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-device_type'");
                return $self;
            }
            next;
        }
        elsif (/-?aggr/)
        {
            $aggr = $args{$_};
            next;
        }
        elsif (/-?filer_vol_size/)
        {
            $vol_size = $args{$_};
            if (not length($vol_size))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol_size'");
                return $self;
            }
            next;
        }
        elsif (/-?filer_vol/)
        {
            $vol = $args{$_};
            if (not length($vol))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol'");
                return $self;
            }
            $vol =~ s/\/$//;      # remove trailing "/"
            next;
        }
        elsif (/-?num_disks/)
        {
            $num_disks = $args{$_};
            if (not length($num_disks))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-num_disks'");
                return $self;
            }
            next;
        }
        elsif (/-?skip_setup/)
        {
            $skip_setup = $args{$_};
            if (not length($skip_setup) or $skip_setup !~ /yes|no|1|0/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-skip_setup'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $skip_setup = 1 if ($skip_setup =~ /yes/i);
            $skip_setup = 0 if ($skip_setup =~ /no/i);
            next;
        }
        elsif (/-?skip_config/)
        {
            $skip_config = $args{$_};
            if (not length($skip_config) or $skip_config !~ /yes|no|1|0/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-skip_config'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $skip_config = 1 if ($skip_config =~ /yes/i);
            $skip_config = 0 if ($skip_config =~ /no/i);
            next;
        }
        elsif (/-?setup_igroups/)
        {
            $setup_igroups = $args{$_};
            if (not length($setup_igroups) or $setup_igroups !~ /yes|no|1|0/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-setup_igroups'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $setup_igroups = 1 if ($setup_igroups =~ /yes/i);
            $setup_igroups = 0 if ($setup_igroups =~ /no/i);
            $self->{setup_igroups} = $setup_igroups;
            next;
        }
        elsif (/-?setup_luns/)
        {
            $setup_luns = $args{$_};
            if (not length($setup_luns) or $setup_luns !~ /yes|no|1|0/i)
            {
                $self->_set_err("Invalid value for argument '-setup_luns'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $setup_luns = 1 if ($setup_luns =~ /yes/i);
            $setup_luns = 0 if ($setup_luns =~ /no/i);
            $self->{setup_luns} = $setup_luns;
            next;
        }
        elsif (/-?ntp_sync/)
        {
            $ntp_sync = $args{$_};
            if (not length($ntp_sync) or $ntp_sync !~ /yes|no|1|0/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-ntp_sync'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $ntp_sync = 1 if ($ntp_sync =~ /yes/i);
            $ntp_sync = 0 if ($ntp_sync =~ /no/i);
            $self->{ntp_sync} = $ntp_sync;
            next;
        }
        elsif (/-?ntp_time_server/)
        {
            $ntp_time_server = $args{$_};
            if (not length($ntp_time_server))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-ntp_time_server'");
                return $self;
            }
            $self->{ntp_time_server} = $ntp_time_server;
            next;
        }
        elsif (/-?ntp_time_zone/)
        {
            $ntp_time_zone = $args{$_};
            if (not length($ntp_time_zone))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-ntp_time_zone'");
                return $self;
            }
            $self->{ntp_time_zone} = $ntp_time_zone;
            next;
        }
        elsif (/-?alua_support/)
        {
            $alua_support = $args{$_};
            if (not length($alua_support) or $alua_support !~ /yes|no|1|0/i)
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-alua_support'");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $alua_support = 1 if ($alua_support =~ /yes/i);
            $alua_support = 0 if ($alua_support =~ /no/i);
            $self->{alua_support} = $alua_support;
            next;
        }
        elsif (/-?lvm/)
        {
            $lvm_setup_args{-lvm}=$args{$_};
            if (not length($lvm_setup_args{-lvm}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        }
        elsif (/-?dg_prefix/)
        {
            $lvm_setup_args{-dg_prefix}=$args{$_};
            if (not length($lvm_setup_args{-dg_prefix}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-dg_prefix'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_layouts/)
        {
            $lvm_setup_args{-vol_layouts}=$args{$_};
            if (not length($lvm_setup_args{-vol_layouts}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol_layouts'");
                return $self;
            }
            next;
        }
        elsif (/-?stripe_size/)
        {
            $lvm_setup_args{-stripe_size}=$args{$_};
            if (not length($lvm_setup_args{-stripe_size}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-stripe_size'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_cnts/)
        {
            $lvm_setup_args{-vol_cnts}=$args{$_};
            if (not length($lvm_setup_args{-vol_cnts}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol_cnts'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_sizes/)
        {
            $lvm_setup_args{-vol_sizes}=$args{$_};
            if (not length($lvm_setup_args{-vol_sizes}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol_sizes'");
                return $self;
            }
            next;
        }
        elsif (/-?disk_fmts/)
        {
            my $disk_fmts = "";
            #
            # If the disk_fmts is blank, don't worry about it.
            #
            if (length($disk_fmts = $args{$_}))
            {
                $lvm_setup_args{-disk_fmts}=$disk_fmts;
            };
            next;
        }
        elsif (/-?mp_type/)
        {
            $mp_type=$args{$_};
            if (not length($mp_type))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-mp_type'");
                return $self;
            }
            next;
        }
        elsif (/-?vfiler_enable/)
        {
            $vfiler_enable=$args{$_};
            if (not length($vfiler_enable)or ($vfiler_enable !~ /0|1/))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vfiler_enable'");
                return $self;
            }
            $self->{vfiler_enable} = $vfiler_enable;
            next;
        }
        elsif (/-?client_pass_thru/)
        {
            $client_pass_thru = $args{$_};
            unless(ref($client_pass_thru) =~ /HASH/)
            {
                $self->_set_err("Error: San::setup() "             .
                                "Invalid value for argument type " .
                                "'-client_pass_thru', HASH type expected");
                return $self;
            }
            next;
        }
        elsif (/-?filer_pass_thru/)
        {
            $filer_pass_thru = $args{$_};
            unless(ref($filer_pass_thru) =~ /HASH/)
            {
                $self->_set_err("Error: San::setup() "             .
                                "Invalid value for argument type " .
                                "'-filer_pass_thru', HASH type expected");
                return $self;
            }
            next;
        }
        elsif (/-?create_fs$/)
        {
            $create_fs = $args{$_};
            if ( defined($create_fs) && ($create_fs !~ /^\s*(0|1|yes|no)\s*$/i) ){
                $self->_set_err("Error: San::setup() "             .
                                "Invalid value for argument type " .
                                "'-create_fs', 0,1,yes,no or undef expected");
                return $self;
            }
            #
            # Change any 'yes' or 'no' sent to a 1 or a 0.
            #
            $create_fs = 1 if ($create_fs =~ /1/i);
            $create_fs = 0 if ($create_fs =~ /0/i);
            $create_fs = 1 if ($create_fs =~ /yes/i);
            $create_fs = 0 if ($create_fs =~ /no/i);

        }
        elsif (/-?create_fs_type/)
        {
            $create_fs_type = $args{$_};
            # Nothing to check here. we don't know what filesystems are acceptable
            # yet. Instead, check the argument once the SanClient has been created
            # before attempting the filesystem creation
        }
        elsif (/-?load_devices/)
        {
            $devmgr_setup_args{-load_devices} = $args{$_};

            if (not length($devmgr_setup_args{-load_devices}))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-load_devices'");
                return $self;
            }

            next;
        }
        else
        {
            $self->_set_err("Error: San::setup()",
                            "Invalid parameter specified \'$_\'");
            return $self;
        }
    } # end foreach arg


    $self->sanlog(-msg       => "Completed parsing args checking for required args.",
                  -msg_level => 7);

    #
    # check for existance of required arguments:
    #   filers, clients
    #
    unless (defined $filers and length $filers)
    {
        $self->_set_err("Error: San::setup()",
                        "Missing required argument '-filers'");
                        return $self;
    }
    unless (defined $clients and length $clients)
    {
        $self->_set_err("Error: San::setup()",
                        "Missing required argument '-clients'");
                        return $self;
    }

    #
    # We need either a filer_vol_list OR a lun_list and vol
    #
    unless (defined $filer_vol_list and length $filer_vol_list)
    {
        unless (defined $lun_list and length $lun_list)
        {
            $self->_set_err("Error: San::setup()",
                            "Missing required argument '-lun_list'.  " .
                            "Either a '-filer_vol_list' or a '-lun_list' " .
                            "must be specified.");
                            return $self;
        }
        unless (defined $vol and length $vol)
        {
            $self->_set_err("Error: San::setup()",
                            "Missing required argument '-vol'.  " .
                            "Either a '-filer_vol_list' or a '-vol' " .
                            "must be specified.");
            return $self;
        }
    }

    $self->sanlog(-msg       => "Completed checking for required args.",
                  -msg_level => 7,
                  -indent    => 4);

    #
    # If the user used the old school lun_list, vol, etc. instesad of
    # the, translate the given values to a filer_vol_list.
    #
    unless ($use_vol_list)
    {
        $filer_vol_list = $self->get_filer_vol_list(
                                             -aggr     => $aggr,
                                             -vol      => $vol,
                                             -vol_size => $vol_size,
                                             -lun_list => $lun_list);
        $self->sanlog(-msg_level => 3,
                      -msg => "Filer Vol List is $filer_vol_list");
    }

    #
    # Instantiate the filer objects.
    #
    @filerArray  = split(/:/, $filers);
    my $num_filers = scalar(@filerArray);
    my $num_filer_sims = 0;

    foreach my $filer (@filerArray)
    {
        @this_filer_targets = ();
        $self->sanlog(-msg => "Creating Filer object for filer $filer");
        my $filer_obj= new SanFiler(-filer   =>$filer,
                                    -timeout =>$self->{timeout},
                                    -debug_level =>$self->{debug_level},
                                    %{$filer_pass_thru},
                                   );

        unless (ref($filer_obj))
        {
            $self->error("Filer is neither real nor scsit simulator.  ABORTING");
            return $self;
        }

        if ($filer_obj->error)
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->sanlog(-msg => $errmsg);

            $self->_set_err("Cannot instantiate filer object for $filer.",
                            "$errmsg");
            return $self;
        }
        #
        # If the filer_obj is a scsitSimulator, the game has to
        # change a little, we'll keep track of this
        #
        if (ref($filer_obj) =~ /ScsitSimulator/)
        {
            $num_filer_sims++;
        }

        push @{$self->{filers}}, $filer_obj;

        # get cfmode
        $cf_mode = $filer_obj->get_cfmode();
        if($filer_obj->error())
        {
            my $errmsg = $filer_obj->errmsg();
            $self->_set_err("Could not get cfmode from controller.",
                            "$errmsg");
            return $self;
        }


        # TBD, once we know the protocol, make sure the filer is licensed
    } # end for each filer

    #
    # Check to see if simulated filers and real filers given
    # Error if they are
    #
    if($num_filer_sims > 0 and $num_filer_sims != @filerArray)
    {
        $self->_set_err("Test must not contain both REAL and SIMULATED filers.");
        return $self;
    }

    #
    # Instantiate the client objects.
    #
    @clientArray = split(/:/, $clients);
    foreach my $client (@clientArray)
    {
        my ($mp_type_ret_ref, $any_iscsi_ret_ref);

        if($mp_type =~ /^per_host$/i)
        {
            $mp_type_ret_ref = $self->_get_pc_mp_type($mp_type, $client);

            if($mp_type_ret_ref)
            {
                $mp_type = $mp_type_ret_ref->{new_mp_type};
            }
            else
            {
                $self->_set_err("Error: San::setup() Instantiating client ".
                                "object for $client. CLIENT_VMS set to ".
                                "per_host but PC_CLIENT_VMS (or PC_ANY_ISCSI) is not set ".
                                "in the tharnhost file for $client.");
            }

        }

        $any_iscsi_ret_ref = $self->_get_pc_any_iscsi($client);

        if($any_iscsi_ret_ref)
        {
            $any_iscsi =  $any_iscsi_ret_ref->{new_any_iscsi};
        }
        else
        {
            $self->_set_err("Error: San::setup() Instantiating client ".
                            "object for $client. $client hosttype is unix|linux ".
                            "but PC_ANY_ISCSI is not set ".
                            "in the tharnhost file for $client.");
        }

        my $client_obj;
        if($any_iscsi)
        {
            $client_obj= new SanClient(-host         => $client,
                                       -timeout      => $self->{timeout},
                                       -debug_level  => $self->{debug_level},
                                       -device_type  => $device_type,
                                       -protocol     => $protocol,
                                       -mp_type      => $mp_type,
                                       -any_iscsi    => $any_iscsi,
                                       -conn_type    => "THARNHOST",
                                       %{$client_pass_thru});
        }
        else
        {

            $client_obj= new SanClient(-host         => $client,
                                       -timeout      => $self->{timeout},
                                       -debug_level  => $self->{debug_level},
                                       -device_type  => $device_type,
                                       -protocol     => $protocol,
                                       -mp_type      => $mp_type,
                                       -conn_type    => "THARNHOST",
                                       %{$client_pass_thru});
        }

        if($mp_type_ret_ref)
        {
            &ClientUtil::restore_original_params($mp_type_ret_ref->{old_client_params});
        }
        if($any_iscsi_ret_ref)
        {
            &ClientUtil::restore_original_params($any_iscsi_ret_ref->{old_client_params});
        }

        if ($client_obj->error)
        {
            my $errmsg = join "\n", $client_obj->errmsg();
            $self->sanlog(-msg => $errmsg);

            $self->_set_err("Error: San::setup()".
                            "Instantiating client object for $client.".
                            "$errmsg");
            return $self;
        }

        push @{$self->{clients}}, $client_obj;

    } # end for each client

    #
    # Parse the filer vol list
    #
    unless ($self->_parse_filer_vol_list($filer_vol_list))
    {
        #
        # The function called will have set the error info
        #
        return $self;
    }

    #
    # If skip setup is set, return.
    #
    if ($skip_setup)
    {
        $self->sanlog( -msg => "Setup exiting because skip_setup is set");
        return 1;
    }

    #
    # Skipping the rest of San.pm setup and calling get_devices
    # because the filer simulator doesn't need any real set up.
    # The call will give us the IT nexuses from the scsit simulator.
    #
    if($num_filer_sims){
        $self->sanlog(-msg => "Setup calling get_devices and exiting because filer".
                              "simulators exist"
                      -msg_level => 3
                     );
        $self->get_devices();
        return 1;
    }
    if($any_iscsi){
        $self->sanlog(-msg => "Setup calling get_devices and exiting because ".
                             "any_iscsi"
                      -msg_level => 3
                     );
        $self->{any_iscsi}=1;
    }

    #
    # Create the vol
    #
    unless ($self->_aggr_vol_lun_setup($aggr, $num_disks, $protocol))
    {
        #
        # The function called will have set the error info
        #
        return $self;
    }


    #
    # If auto config is enabled we need target objects to pass
    # to SanClient configure method to config automatically
    #
    unless ($skip_config)
    {
        foreach my $filer_obj (@{$self->{filers}})
        {
            #
            # Get the controller target objects
            #
            if ($protocol =~ /iscsi/i)
            {
                #
                # check iscsi service for 7mode
                #
                if (ref($filer_obj) =~ /SanFiler7M/)
                {
                    #
                    # add iscsi license
                    #
                    $filer_obj->license_iscsi();
                    if ($filer_obj->error())
                    {
                        $self->_set_err("Failed to add iscsi license.",
                                        $filer_obj->errmsg());
                        return $self;
                    }

                    #
                    # get iscsi status
                    #
                    my $status = $filer_obj->get_iscsi_status();
                    if ($filer_obj->error())
                    {
                        $self->_set_err("Failed to get iscsi status.",
                                        $filer_obj->errmsg());
                        return $self;
                    }

                    #
                    # start iscsi service if status is disable
                    #
                    if ($status =~ /is not running/i)
                    {
                        $filer_obj->iscsi_start();
                        if ($filer_obj->error())
                        {
                            $self->_set_err("Failed to start iscsi service.",
                                            $filer_obj->errmsg());
                            return $self;
                        }
                    }
                }
                @this_filer_targets = $filer_obj->get_iscsi_targets(-state => "enabled",
                                                                    -use_nate_params => 1,
                                                                    -valid_ip => 1);
            } else {
                #
                # check fcp service for 7mode
                #
                if (ref($filer_obj) =~ /SanFiler7M/)
                {
                    #
                    # add fcp license
                    #
                    $filer_obj->license_fcp();
                    if ($filer_obj->error())
                    {
                        $self->_set_err("Failed to add fcp license.",
                                        $filer_obj->errmsg());
                        return $self;
                    }

                    #
                    # get fcp status
                    #
                    my $status = $filer_obj->get_fcp_status();
                    if ($filer_obj->error())
                    {
                        $self->_set_err("Failed to get fcp status.",
                                        $filer_obj->errmsg());
                        return $self;
                    }

                    #
                    # start fcp service if status is disable
                    #
                    if ($status =~ /is not running/i)
                    {
                        $filer_obj->fcp_start();
                        if ($filer_obj->error())
                        {
                            $self->_set_err("Failed to start fcp service.",
                                            $filer_obj->errmsg());
                            return $self;
                        }
                    }
                }
                @this_filer_targets = $filer_obj->get_targets(-use_nate_params => 1);
            }
            if ($filer_obj->error())
            {
                $self->_set_err("Could not get $protocol target objects.",
                                $filer_obj->errmsg());
                return $self;
            }
            push(@targets, @this_filer_targets);
        }
    }

    #
    # Now that luns are complete, the filer side setup is done, we move
    # on to client side setup.
    #
    foreach my $client_obj (@{$self->{clients}})
    {
        my $client_name = $client_obj->hostname();
        if ($client_obj->error)
        {
            my $errmsg = $client_obj->errmsg();
            $self->_set_err("Can not get hostname: ",
                            "$errmsg");
            return $self;
        }

        #
        # Need to configure the host here
        #
        if ($skip_config or $any_iscsi)
        {
            $self->sanlog(-msg => "Setup skipping configure because " .
                                  "skip_config is set");
        }
        else
        {
            $self->sanlog(-msg => "Configuring client $client_name");

            #
            # To configure a host we will need to send the host configure
            # method the list of lun object and a list of filer objects.
            #
            $client_obj->configure(-cfo_mode => $cf_mode, -targets => \@targets);

            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Error: San::setup() ",
                                "Can not configure host: ",
                                "$errmsg");
                return $self;
            }

            return $self
                unless ($self->cleanup_add(-type=>"client",
                                           -name=>$client_name,
                                           -cmd=>"configure_cleanup",
                                           -cmd_args=>\@targets));
        } # end else skip_config

        #
        # Discover and label the devices
        #
        $client_obj->discover();
        if ($client_obj->error)
        {
            my $errmsg = join "\n", $client_obj->errmsg();
            $self->_set_err("Error: San::setup() ",
                            "Can not discover devices: ",
                            "$errmsg");
            return $self;
        }

        #
        # If we are setting up raw devices, we should be done.
        #
        if ($device_type =~ /raw/)
        {
            #
            # Register the cleanup action for raw device setup
            #
            return $self
                unless ($self->cleanup_add(
                                  -type=>"client",
                                  -name=>$client_name,
                                  -cmd=>"raw_device_cleanup",
                                  -cmd_args=>"$protocol"));

            #
            # Setup the filesystem on each LUN
            #
            if ($create_fs){
                my $luns = $self->{config_lun_objs};
                my @lun_paths = map {$_->get_filername . ':' . $_->get_path} @$luns;
                $self->sanlog(-msg => "Lun paths we care about: @lun_paths ");

                #
                # MP devices need to be handled differently
                #
                if ($client_obj->{mp_type} =~ /none/i){
                    my $devmgr_obj = $client_obj->{obj}{devmgr};
                    if (!$devmgr_obj){
                        $devmgr_obj = $client_obj->create_devmgr(%devmgr_setup_args);

                        if ($client_obj->error)
                        {
                            my $errmsg = join "\n", $client_obj->errmsg();
                            $self->_set_err("Error: San::setup() ",
                                            "Can not setup raw devices: ",
                                            "$errmsg");
                            return $self;
                        }

                        # TODO: temporary code
                        #$devmgr_obj->{devices} = [];
                        #$devmgr_obj->_update_definition_file;
                        #if ($devmgr_obj->error)
                        #{
                        #    my $errmsg = join "\n", $devmgr_obj->errmsg();
                        #    $self->_set_err("Error: San::setup() ",
                        #                    "Can not update definition file: ",
                        #                    "$errmsg");
                        #    return $self;
                        #}
                        # End temporary code
                    }
                    #
                    # Register the cleanup action for fs device setup
                    #
                    return $self
                        unless ($self->cleanup_add(
                                          -type=>"client",
                                          -name=>$client_name,
                                          -cmd=>"devmgr_cleanup",
                                          -cmd_args=>$devmgr_obj));

                    #
                    # Add devices we care about to the DevMgr
                    #
                    my @devices = $devmgr_obj->get_free_devices();
                    if (!@devices){
                        $self->_set_err("Error: San::setup() ", "No valid devices are available to use");
                        return $self;
                    }
                    my %seen_luns;
                    foreach my $device (@devices){
                        my $name;
                        if ($device =~ /\/dev\/(.+)$/){
                            $name = $1;
                        }
                        else{
                            $self->_set_err("Error: San::setup() ",
                                            "Can not parse device '$device'");
                            return $self;
                        }

                        # Only create a device object if it matches one of the luns
                        # we're tracking or have created ourselves
                        # Luns may show more than once but only keep one device per lun
                        #
                        my $device_lun_path = $client_obj->get_sanlun->get_lun_path(-device => $device);
                        my $match = grep(/$device_lun_path\s*$/, @lun_paths);
                        if (!$match){
                            $self->sanlog(-msg => "Device's LUN '$device_lun_path' not one of ours");
                            next;
                        }
                        if ($seen_luns{$device_lun_path}){
                            $self->sanlog(-msg => "Device's LUN '$device_lun_path' already seen");
                            next;
                        }
                        $seen_luns{$device_lun_path} = 1;
                        $devmgr_obj->create_device(-name=>$device);
                        if ($devmgr_obj->error){
                            my $errmsg = join "\n", $devmgr_obj->errmsg();
                            $self->_set_err("Error: San::setup() ",
                                            "Can not create device object: ",
                                            "$errmsg");
                            return $self;
                        }
                    }

                    #
                    # Create filesystem and mount each device
                    #
                    my @devObjs = $devmgr_obj->get_devices();
                    if (!@devObjs){
                        $self->_set_err("Error: San::setup() ", "No valid device objects are available to use");
                        return $self;
                    }
                    if ($create_fs_type){
                        my @client_supported_fs_types = $client_obj->get_supported_fs_types;
                        # Verify fs_type is supported
                        unless ( scalar grep(/^$create_fs_type$/, @client_supported_fs_types) )
                        {
                            $self->_set_err("Error: San::setup() ", "Unsupported fs type ($create_fs_type) for client");
                            return $self;
                        }
                    }
                    foreach my $device (@devObjs) {
                        $device->load;
                        if ($device->error){
                            my $errmsg = join "\n", $device->errmsg();
                            $self->_set_err("Error: San::setup() ",
                                            "Can not load device object: ",
                                            "$errmsg");
                            return $self;
                        }
                        if ($device->get_fs){
                            $self->sanlog(-msg => "Device already has a filesystem");
                        }
                        else{
                            $device->create_fs(-type => $create_fs_type);
                            if ($device->error()) {
                                my $errmsg = join "\n", $device->errmsg();
                                $self->_set_err("Error: San::setup() create_fs() failed", $errmsg);
                                return $self;
                            }
                        }
                        $device->mount_fs();
                        if ($device->error()) {
                            my $errmsg = join "\n", $device->errmsg();
                            $self->_set_err("Error: San::setup() mount_fs() failed",$errmsg);
                            return $self;
                        }
                    }
                }# end if mp_type none
                else{
                    # TODO
                }

            }# end if create_fs

            $self->sanlog(-msg => "Setup Complete for client " .
                                  $client_name);

            next;
        }

        #
        # Setup the LVM for the client
        #
        if ($device_type =~ /lv/)
        {
            # args: -dg_prefix, -lvm, -vol_layouts, -stripe_size, -vol_cnts, -vol_sizes, -disk_fmts

            $lvm_setup_args{-client}=$client_obj;
            $lvm_setup_args{-filers}=$filers;
            unless ($self->_lvm_setup(%lvm_setup_args))
            {
                #
                # The function called will have set the error info
                #
                return $self;
            }

            if ($create_fs){
                $self->_set_err("Error: San::setup() ",
                                "Option 'create_fs' not supported with device_type 'lv'");
                return $self;
            }

        } # end if lvm

    } # end for each client

    $self->sanlog(-msg => "San Setup Complete");
    return 1;

}  # end of setup method

########################################################################
# Method:    cleanup
# Objective: Cleanup resources created by setup.
# Details:   See POD documentation at the end of this file.
# Notes:     Our policy is to cleanup only what we have setup.
########################################################################
sub cleanup()
{
    my ($self, $skip_cleanup, $cleanup_reused, $max_tolerable_error) = @_;
    my @cleanup_array = @{$self->{cleanup}};
    my $noaggr = 0;

    #
    # If skip cleanup is set, return.
    #
    if ($skip_cleanup)
    {
        $self->sanlog(-msg => "Cleanup exiting because skip_cleanup is set");
        return 1;
    }

    #
    # If the user didn't give us a value for $cleanup_reused, then set
    # it to 0 so by default we don't cleanup reused elements.
    #
    unless (defined ($cleanup_reused))
    {
        $cleanup_reused = 0;
    }

    #
    # If the user didn't give us a value for $max_tolerable_error, then set
    # it to 1 so by default we have compatibility with old behavior
    # This variable controls how many errors we will tolerate before bailing.
    # It provides the ability to continue cleanup after one or more errors
    # are seen.
    #
    unless (defined ($max_tolerable_error))
    {
        $max_tolerable_error = 1;
    }
    $self->sanlog(-msg => "Max error is set to $max_tolerable_error");
    my $error_cnt = 0;

    #
    # If the user gave us "noaggr" $cleanup_reused, then we will set both
    # the reused var and $noaggr and we won't delete the aggr later.
    #
    if ($cleanup_reused =~ /^noaggr$/i)
    {
        $noaggr = 1;
        $cleanup_reused = 1;
    }

    $self->sanlog(-msg => "Cleanup Started $#cleanup_array items on cleanup list");

    #
    # Execute the cleanup script
    #
    my $step = 0;
    for (my $index = $#cleanup_array ; $index >= 0  ; $index--)
    {
        my $cleanup_hash = $cleanup_array[$index];
        my $type     = $cleanup_hash->{-type};
        my $name     = $cleanup_hash->{-name};
        my $cmd      = $cleanup_hash->{-cmd};
        my $cmd_args = $cleanup_hash->{-cmd_args};
        my $reuse    = $cleanup_hash->{-reuse};
        my $vfiler_context = "";
        if ($cleanup_hash->{-vfiler_context}){
            $vfiler_context = $cleanup_hash->{-vfiler_context};
        }
        my $filer_obj;
        my $client_obj;

        $step++;

        #
        # If this element was reused and not freshly created by us,
        # only cleanup it up if the cleanup_reused is set.  If the
        # user requesed 'noaggr' then we don't destroy reused aggrs
        #
        if ($reuse && !($cleanup_reused))
        {
            $self->sanlog(-msg => "Cleanup Step $step" . ":"   .
                                  " Skipping reused element: " .
                                  " $type $name $cmd $cmd_args");
            next;
        }
        elsif ($reuse && $cleanup_reused && $noaggr && ($cmd eq "aggr_destroy"))
        {
            $self->sanlog(-msg => "Cleanup Step $step" . ":"   .
                                  " Skipping aggr element: " .
                                  " $type $name $cmd $cmd_args");
            next;
        }
        elsif ($reuse && $cleanup_reused)
        {
            $self->sanlog(-msg => "Cleanup Step $step" . ":"  .
                                  " Cleanup reused element: " .
                                  " $type $name $cmd $cmd_args");
        }
        else
        {
            $self->sanlog(-msg => "Cleanup Step $step" . ":" .
                                  " $type $name $cmd $cmd_args $vfiler_context");
        }

        #
        # If we are working on a filer, we will need the filer object.
        # Get it here so this code isn't repeated for all filer cases
        #
        if ($type =~ /filer/)
        {
            $self->sanlog(-msg       => "Finding filer object for filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            $filer_obj = $self->get_filer($name);
            if ($self->error)
            {
                return $self;
            }

        }

        #
        # If we are working on a client, we will need the client object.
        #
        if ($type =~ /client/)
        {
            $self->sanlog(-msg       => "Finding client object for client " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            $client_obj = $self->get_client($name);
            if ($self->error)
            {
                return $self;
            }
        }

        if ($type =~ /filer/ and $cmd =~ /lun_unmap/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   vol lun igroup
            #
            (my $vol, my $lun_name, my $igroup_name) = split(" ", $cmd_args);

            #
            # Find the volume object
            #
            $self->sanlog(-msg       => "Finding volume object for $vol " .
                                        "on filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            #
            # Luns need to be unmapped in the vfiler context in which
            # they were mapped.  if $vfiler_context is set, we need to
            # set context into that vfiler in order to do the cleanup
            #
            $filer_obj->set_vfiler_context(-name => $vfiler_context) if ($self->{vfiler_enable});
            if ($filer_obj->error())
            {
                $self->_set_err("Unable to set vfiler context: ",
                                $filer_obj->errmsg());
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            my $short_vol = $self->_get_short_vol($vol);
            my $filer_vol_obj = $filer_obj->get_vol($short_vol);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not get volume: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Find the lun object
            #
            $self->sanlog(-msg       => "Finding lun object for $vol " . "/" .
                                        "$lun_name on filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            my $lun_obj = $filer_vol_obj->get_lun($lun_name);
            if ($filer_vol_obj->error)
            {
                my $errmsg = join "\n", $filer_vol_obj->errmsg();
                $self->_set_err("Can not get lun: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Unmap the lun
            #
            $self->sanlog(-msg    => "Unmap lun for $vol " . "/" .
                                     "$lun_name on filer " . $name,
                          -indent => 4);

            $lun_obj->unmap($igroup_name);
            if ($lun_obj->error)
            {
                my $errmsg = join "\n", $lun_obj->errmsg();
                $self->_set_err("Can not unmap lun: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }

            $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});

        } # end lun unmap case
        elsif ($type =~ /filer/ and $cmd =~ /lun_destroy/)
        {
            #
            # We expect the cmd_args to be of the form:
            # vol lun
            #

            (my $vol, my $lun_name) = split(" ", $cmd_args);

            #
            # If the path to the lun contains a qtree,
            # the volume will be of the form vol/qtree.
            # Extract the volume name if a qtree is present.
            #

            if($vol =~ /(\/vol\/\w*)\/(\w*)/){
                $vol = $1;
            }

            #
            # Find the volume object
            #
            $self->sanlog(-msg       => "Finding volume object for $vol " .
                                        "on filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            $filer_obj->set_vfiler_context(-name => $vfiler_context) if ($self->{vfiler_enable});
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Unable to set vfiler context ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            my $short_vol = $self->_get_short_vol($vol);
            my $filer_vol_obj = $filer_obj->get_vol($short_vol);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not get volume: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Find the lun object
            #
            $self->sanlog(-msg       => "Finding lun object for $vol " . "/" .
                                        "$lun_name on filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            my $lun_obj = $filer_vol_obj->get_lun($lun_name);
            if ($filer_vol_obj->error)
            {
                my $errmsg = join "\n", $filer_vol_obj->errmsg();
                $self->_set_err("Can not get lun: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Destroy the lun
            #
            $self->sanlog(-msg    => "Destroy lun $vol " . "/" .
                                     "$lun_name on filer " . $name,
                          -indent => 4);

            $lun_obj->destroy();
            if ($lun_obj->error)
            {
                my $errmsg = join "\n", $lun_obj->errmsg();
                $self->_set_err("Can not destroy lun: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }

            $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});

        } # end destroy lun case
        elsif ($type =~ /filer/ and $cmd =~ /vfiler_destroy/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   vfiler
            #
            my $vfiler_name = $cmd_args;

            #
            # Destroy the vfiler.
            #
            $self->sanlog(-msg    => "Destroy vfiler $vfiler_name on filer " . $name,
                          -indent => 4);
            $filer_obj->vfiler_destroy($vfiler_name);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not destroy vfiler: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        } # end destroy vfiler case
        elsif ($type =~ /filer/ and $cmd =~ /igroup_destroy/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   igroup
            #
            my $igroup_name = $cmd_args;

            if ($self->{vfiler_enable}) {
                $filer_obj->set_vfiler_context(-name => $vfiler_context);
                if ($filer_obj->error)
                {
                    my $errmsg = join "\n", $filer_obj->errmsg();
                    $self->_set_err("Unable to set vfiler context ",
                                    "$errmsg");
                    return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                    next;
                }
            }

            #
            # Destroy the igroup.  Note that we could send a %args
            # with the "-force" argument if necessary.
            #
            $self->sanlog(-msg    => "Destroy igroup $igroup_name on filer " . $name,
                          -indent => 4);
            $filer_obj->destroy_igroup($igroup_name);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not destroy igroup: ",
                                "$errmsg");
                $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }

            $filer_obj->clear_vfiler_context() if ($self->{vfiler_enable});

        }
        elsif ($type =~ /filer/ and $cmd =~ /vol_destroy/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   vol
            #
            my $vol = $cmd_args;

            #
            # Find the volume object
            #
            $self->sanlog(-msg       => "Finding volume object for $vol " .
                                        "on filer " . $name,
                          -msg_level => 7,
                          -indent    => 4);

            my $short_vol = $self->_get_short_vol($vol);
            my $filer_vol_obj = $filer_obj->get_vol($short_vol);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not get volume: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Offline the volume
            #
            $self->sanlog(-msg    => "Offline volume $vol on filer " . $name,
                          -indent => 4);
            $filer_vol_obj->offline();
            if ($filer_vol_obj->error)
            {
                my $errmsg = join "\n", $filer_vol_obj->errmsg();
                $self->_set_err("Can not offline volume: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Destroy the volume
            #
            $self->sanlog(-msg    => "Destroy volume $vol on filer " . $name,
                          -indent => 4);
            $filer_vol_obj->destroy();
            if ($filer_vol_obj->error)
            {
                my $errmsg = join "\n", $filer_vol_obj->errmsg();
                $self->_set_err("Can not destroy volume: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        } # end destroy vol case
        elsif ($type =~ /filer/ and $cmd =~ /aggr_destroy/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   aggr
            #
            my $aggr = $cmd_args;

            #
            # Offline the aggregate
            #
            $self->sanlog(-msg    => "Offline aggregate $aggr on filer " . $name,
                          -indent => 4);
            $filer_obj->aggr_offline($aggr);
            if ($filer_obj->error())
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not offline aggr: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Destroy the aggregate (note the "1" means force)
            #
            $self->sanlog(-msg    => "Destroy aggregate $aggr on filer " . $name,
                          -indent => 4);
            $filer_obj->aggr_destroy($aggr,1);
            if ($filer_obj->error())
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Can not destroy aggr: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        } # end destroy aggr case
        elsif ($type =~ /filer/ and $cmd =~ /cmd/)
        {
            $self->_set_err("The filer cmd cleanup command is " .
                            "not yet supported");
            return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
        }
        elsif ($type =~ /client/ and $cmd =~ /lvm_cleanup/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   vol_mgr dg_prefix
            #
            my ($vol_mgr, $dg_prefix) = split(" ", $cmd_args);

            #
            # We will create a SanConfig object so we can use it's
            # lvm_cleanup method which takes a list of vg names.
            # We want to create an LVM object.  Then call get_vgnames.
            # For all the vg_names that start with the dg_prefix, we
            # push them on an array to be sent to SanConfig::lvm_setup().
            #

            #
            # Create the LVM object with the appropriate volume manager
            #
            $self->sanlog(-msg => "Create a LVM object ($vol_mgr)",
                          -msg_level => 7);
            my $lvm_obj = $client_obj->create_lvm(-lvm=>$vol_mgr);
            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Can not create lvm object: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Create the SanConfig object
            #
            $self->sanlog(-msg => "Create a SanConfig object",
                          -msg_level => 7);
            my $config_obj = new SanConfig(-clientObj      => $client_obj,
                                           -lvmObj         => $lvm_obj,
                                           -config         => 1,
                                           -devmap_refresh => 1
                                          );
            if ($config_obj->error)
            {
                my $errmsg = join "\n", $config_obj->errmsg();
                $self->_set_err("Can not create SanConfig object: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Get a list of vg names on the client.
            #
            my @all_vg_names = $lvm_obj->get_vgnames();
            if ($lvm_obj->error)
            {
                my $errmsg = join "\n", $lvm_obj->errmsg();
                $self->_set_err("Can not get_vgnames: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
                next;
            }

            #
            # Match the vg_names to the dg_prefix
            #
            my @vg_names;
            foreach my $vg_name (@all_vg_names)
            {
                if ($vg_name =~ /^$dg_prefix\d+$/)
                {
                    push @vg_names, $vg_name;
                }
            }

            #
            # Call the SanConfig lvm_cleanup method
            #
            $self->sanlog(-msg => "Cleanup the lvm using SanConfig");
            $config_obj->lvm_cleanup(@vg_names);
            if ($config_obj->error)
            {
                my $errmsg = join "\n", $config_obj->errmsg();
                $self->_set_err("Call to lvm_cleanup failed: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        } # end of lvm cleanup case
        elsif ($type =~ /client/ and $cmd =~ /raw_device_cleanup/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   protocol
            #
            my $protocol = $cmd_args;

            #
            # Cleanup the host
            #
            $client_obj->raw_device_cleanup($protocol);
            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Call to raw_device_cleanup failed: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        }
        elsif ($type =~/client/ and $cmd =~ /devmgr_cleanup/){
            # Create a device manager object
            my $devmgr_obj = $cmd_args;
            $devmgr_obj->cleanup;
            if ($devmgr_obj->error){
                my $errmsg = join "\n", $devmgr_obj->errmsg();
                $self->_set_err("Call to devMgr::cleanup failed: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        }
        elsif ($type =~ /client/ and $cmd =~ /orphaned_device_cleanup/)
        {
            #
            # We expect the cmd_args to be of the form:
            #   protocol
            #
            my $protocol = $cmd_args;

            #
            # Cleanup the host
            #
            $client_obj->orphaned_device_cleanup($protocol);
            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Orphaned device cleanup failed: ",
                                "$errmsg");
                return $ClientUtil::FALSE if ($error_cnt++ >= $max_tolerable_error);
            }
        }
        elsif ($type =~ /client/ and $cmd =~ /configure_cleanup/)
        {
            #
            # Cleanup all the targetObjs whose configuration is set
            #
            $client_obj->deconfigure(-targets => $cmd_args);
            if ($client_obj->error)
            {
                my $errmsg = $client_obj->errmsg();
                $self->_set_err("Client deconfiguration  cleanup failed:",
                                "$errmsg");
            }
        }
        else
        {
            $self->_set_err("Error: San::cleanup()",
                             "Invalid command specified \'$cmd\'");
            return 0;
        }
    } # end for loop for cleanup array elements

    $self->sanlog(-msg => "San Cleanup Complete");
    return 1;
} # end cleanup method

########################################################################
# Method:    cleanup_add
# Objective: Add and entry to the cleanup stack.
# Details:   See POD documentation at the end of this file.
########################################################################
sub cleanup_add()
{
    my($self,%args) = @_;

    my $type;
    my $name;
    my $cmd;
    my $cmd_args     = "";
    my $reuse        = 0;
    my $cleanup_hash = {};
    my $vfiler_context = "";

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?type/)
        {
            $type = $args{$_};
            if (not length($type))
            {
                $self->_set_err("Error: San::cleanup_add()",
                                "Invalid value for argument '-type'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?name/)
        {
            $name = $args{$_};
            if (not length($name))
            {
                $self->_set_err("Error: San::cleanup_add()",
                                "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?cmd_args/)
        {
            $cmd_args = $args{$_};
            if (not length($cmd_args))
            {
                $self->_set_err("Error: San::cleanup_add()",
                                "Invalid value for argument '-cmd_args'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?cmd/)
        {
            $cmd = $args{$_};
            if (not length($cmd))
            {
                $self->_set_err("Error: San::cleanup_add()",
                                "Invalid value for argument '-cmd'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?reuse/)
        {
            $reuse = $args{$_};
            if (not length($reuse))
            {
                $self->_set_err("Error: San::cleanup_add()",
                                "Invalid value for argument '-reuse'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?vfiler_context/)
        {
            $vfiler_context = $args{$_};
            if (not length($vfiler_context))
            {
                # if context not set, default it to vfiler0
                $vfiler_context = "vfiler0";
            }
            next;
        }
        else
        {
            $self->sanlog(-msg => "San::cleanup_add ignoring argument $_");
        }
    } # end foreach arg

    #
    # Check for required parameters
    #
    unless (defined $type and length $type)
    {
        $self->_set_err("Error: San::cleanup_add()",
                        "Missing required argument '-type'");
                        return $ClientUtil::FALSE;
    }
    unless (defined $name and length $name)
    {
        $self->_set_err("Error: San::cleanup_add()",
                        "Missing required argument '-name'");
                        return $ClientUtil::FALSE;
    }
    unless (defined $cmd and length $cmd)
    {
        $self->_set_err("Error: San::cleanup_add()",
                        "Missing required argument '-cmd'");
                        return $ClientUtil::FALSE;
    }

    #
    # Push the entry onto the stack
    #

    $cleanup_hash->{-type}     = $type;
    $cleanup_hash->{-name}     = $name;
    $cleanup_hash->{-cmd}      = $cmd;
    $cleanup_hash->{-cmd_args} = $cmd_args;
    $cleanup_hash->{-reuse}    = $reuse;
    $cleanup_hash->{-vfiler_context}    = $vfiler_context;
    my $cleanup_entry = "$type $name $cmd $cmd_args";
    push @{$self->{cleanup}}, $cleanup_hash;
    if ($reuse)
    {
        $self->sanlog(-msg => "Added cleanup reuse step: $cleanup_entry",
                      -msg_level => 3);

    }
    else
    {
        $self->sanlog(-msg => "Added cleanup step: $cleanup_entry");
    }

    return 1;
} # end cleanup_add method

########################################################################
# Method:    destroy
# Objective: Undo the new if necessary.
# Details:   See POD documentation at the end of this file.
########################################################################
sub destroy()
{
    return 1;
}

########################################################################
# Method:    get_devices
# Objective: Get a list of devices to send to an I/O application.
# Details:   See POD documentation at the end of this file.
########################################################################
sub get_devices()
{
    my($self,%args) = @_;

    my $filer;               # arg - filer name
    my $client;              # arg - client name
    my $vol;                 # arg - filer volume name
    my $lun;                 # arg - lun string to match for raw
    my $lun_list;            # arg - list of luns to get num
    my $device_type = "raw"; # arg - raw | vm | fs (raw is default)
    my $protocol    = "fcp"; # arg - iscsi | fcp (default)
    my $short_vol   = "";    # will be used to string "/vol/"
    my @device_array;        # place holder for devices
    my %devArgs;             # args for get_free_devices call
    my @luns;                # used if lun_list is set
    my $num_luns;            # used if lun_list is set
    my $dg_prefix = "san_dg_"; # disk group prefix for lvm
    my $vol_mgr    = "nativelvm";
    my $log_string = "San::get_devices() ";
    my $use_filer_vol_list = 1;
    my $with_device_info = 0;
    my @filer_objs;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?filer$/)
        {
            $filer = $args{$_};
            if (not length($filer))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "filer = $filer ";
            next;
        }
        elsif (/-?client$/)
        {
            $client = $args{$_};
            if (not length($client))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-client'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "client = $client ";
            next;
        }
        elsif (/-?use_filer_vol_list$/)
        {
            $use_filer_vol_list = $args{$_};
            if (not length($use_filer_vol_list))
            {
                $self->_set_err("Error: San::get_devices()",
                            "Invalid value for argument '-use_filer_vol_list'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "use_filer_vol_list = $use_filer_vol_list ";
            next;
        }
        elsif (/-?filer_vol$/)
        {
            $vol = $args{$_};
            if (not length($vol))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-filer_vol'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "vol = $vol ";
            next;
        }
        elsif (/-?lun$/)
        {
            $lun = $args{$_};
            if (not length($lun))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "lun = $lun ";
            next;
        }
        elsif (/-?lvm$/) {
            length($vol_mgr = $args{$_}) or
            do {
                $self->_set_err("San::get_devices invalid value for \"lvm\":",
                                "$args{$_}");
                return (wantarray()) ? () : [];
            };
        }
        elsif (/-?dg_prefix$/)
        {
            $dg_prefix = $args{$_};
            if (not length($dg_prefix))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-dg_prefix'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "dg_prefix = $dg_prefix ";
            next;
        }
        elsif (/-?lun_list$/)
        {
            $lun_list = $args{$_};
            if (not length($lun_list))
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-lun_list'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "lun_list = $lun_list ";
            next;
        }
        elsif (/-?protocol$/)
        {
            $protocol = $args{$_};
            if (not length($protocol) or $protocol !~ /fcp|iscsi/i)
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "protocol = $protocol ";
            next;
        }
        elsif (/-?device_type$/)
        {
            $device_type = $args{$_};
            if (not length($device_type) or $device_type !~ /raw|lv|fs/i)
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-device_type'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "device_type = $device_type ";
            next;
        }
        elsif (/-?with_device_info$/)
        {
            $with_device_info = $args{$_};
            if (not length($with_device_info) or $with_device_info !~ /\d/i)
            {
                $self->_set_err("Error: San::get_devices()",
                                "Invalid value for argument '-with_device_info'");
                return (wantarray()) ? () : [];
            }
            $log_string .= "with_device_info = $with_device_info ";
            next;
        }
        else
        {
            $self->_set_err("Error: San::get_devices()",
                            "Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : [];
        }
    } # end foreach arg

    #
    # get the first filer object from the list
    #
    my $pri_filer = $self->{filers}[0];

    #
    # workaround for Cmode due to the lack of GVA
    # filer and lun-path are fixed
    # filer: lun-pathname device filename adapter protocol lun size lun state
    # vs0: /vol/smokejumper/lun0 /dev/sdb host2   FCP     (unknown) GOOD
    #
    if (ref($pri_filer) =~ /Filer10M/)
    {
        #
        # these parameters are fixed in Cmode lun-path
        # we will remove these parameters once GVA is implemented
        #
        $filer = "vs0";

        #
        # we won't use filer_vol_list so disable use_filer_vol_list
        #
        $use_filer_vol_list = 0;

        #
        # get a list of vservers in the cluster to represent filers in Cmode
        #
        @filer_objs = $pri_filer->get_vservers();
        if ($pri_filer->error())
        {
            $self->_set_err("Can not get vservers: ",
                            $pri_filer->errmsg());
            return (wantarray()) ? () : [];
        }
    }
    else
    {
        @filer_objs = @{$self->{filers}};
    }

    #
    # Logical Volume (lv) devices
    #
    if ($device_type =~ /lv/i)
    {

        #
        # Get devices on each client filer
        #
        foreach my $client_obj (@{$self->{clients}})
        {
            #
            # Get the host name
            #
            my $client_name = $client_obj->hostname();
            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Error: San::get_devices() ",
                                "Can not get hostname: ",
                                "$errmsg");
                return (wantarray()) ? () : [];
            }

            #
            # If the user specified a client and this is not that client,
            # go to the next client.
            #
            if (defined $client and ($client !~ /$client_name/))
            {
                $self->sanlog(-msg       => "Not getting devices for " .
                                            "client $client_name.",
                              -msg_level => 3);
                next;
            }

            $self->sanlog(-msg => "Get lv devices for client $client_name");

            #
            # Create the LVM object with the appropriate volume manager
            #
            if ($vol_mgr eq 'nativelvm')
            {
                $vol_mgr = $client_obj->get_lvm_default();
                $self->sanlog(-msg => "Using Native Volume manager $vol_mgr",
                              -msg_level => 3);
            }

            $self->sanlog(-msg_level => 7,
                          -msg => "Create a LVM object " .
                                  "($client_name $vol_mgr)");
            my $lvm_obj = $client_obj->create_lvm(-lvm=>$vol_mgr);
            if ($client_obj->error)
            {
                my $errmsg = join "\n", $client_obj->errmsg();
                $self->_set_err("Error: San::get_devices() ",
                                "Can not create lvm object: ",
                                "$errmsg");
                return (wantarray()) ? () : [];
            }

            #
            # Get the volume group objects for the client and
            # for each vg get the lvm object and the raw lv devices
            #
            my @vg_list;

            @vg_list = $lvm_obj->get_vgs();
            if ($lvm_obj->error)
            {
                my $errmsg = join "\n", $lvm_obj->errmsg();
                $self->_set_err("Error: San::get_devices() ",
                                "Call to get_vgs failed: ",
                                "$errmsg");
                return (wantarray()) ? () : [];
            }

            foreach my $vg_obj (@vg_list)
            {
                my $vol_name = $vg_obj->get_name();

                #
                # Skip disk groups with the wrong prefix
                # (i.e. existing dgs).
                #
                if ($vol_name =~ /^$dg_prefix/)
                {
                    $self->sanlog(-msg => "Get existing volumes in disk " .
                                          "group '".$vol_name."'");
                    $self->sanlog(-msg => $vg_obj->display(),
                                  -msg_level => 7);
                }
                else
                {
                    $self->sanlog(-msg => "Skipping group '".$vol_name."'");
                    next;
                }

                #
                # Get the volume objects
                #
                my @vol_list = $vg_obj->get_vols();
                if ($vg_obj->error)
                {
                    my $errmsg = join "\n", $vg_obj->errmsg();
                    $self->_set_err("Error: San::get_devices() ",
                                    "Call to get_vols failed: ",
                                    "$errmsg");
                    return (wantarray()) ? () : [];
                }

                #
                # Foreach logical volume, get the raw lv device
                # and add it to the device array
                #
                foreach my $vol_obj (@vol_list)
                {
                    my $device = $vol_obj->get_raw_device();

                    #
                    # With the with_device_info flag, get_devices will
                    # return array of hashes with device info:
                    # dev_name, dev_obj, and dev_size
                    #
                    if ($with_device_info)
                    {
                        my %device_info = ();
                        $device_info{dev_name} = $device;
                        $device_info{dev_obj}  = $vol_obj;
                        $device_info{dev_size} = $vol_obj->get_size();
                        if ($vol_obj->error())
                        {
                            $self->_set_err("Failed to get logical volume size ",
                                            $vol_obj->errmsg());
                            return (wantarray()) ? () : [];
                        }

                        push @device_array, \%device_info;
                    }
                    else
                    {
                        push @device_array, $device;
                    }

                    $self->sanlog(-msg_level => 7,
                                  -msg       => "Added device $device");
                } # end foreach vol
            } # end foreach vg
        } # end foreach client
    }
    #
    # Raw (raw) devices
    #
    elsif ($device_type =~ /raw/i)
    {

        #
        # If we were given a vol, format it and add it to the args list
        # for our call to get_free_devices.
        #
        if (defined $vol and (length($vol)))
        {
            $short_vol = $self->_get_short_vol($vol);
            $devArgs{-vol}=$short_vol;
        }

        #
        # If we were given a lun, add it to the args list
        # for our call to get_free_devices.
        #
        if (defined $lun and (length($lun)))
        {
            $devArgs{-lun}=$lun;
        }

        #
        # If we were given a lun list, parse it and add it's count to the args list
        # for our call to get_free_devices.
        #
        if (defined $lun_list and (length($lun_list)))
        {
            @luns = ha_filer_parse_lun_list("list", $lun_list);
            $num_luns = @luns;
            $devArgs{-cnt} = $num_luns;
        }

        $self->sanlog(-msg => $log_string);

        foreach my $filer_obj (@filer_objs)
        {
            #
            # if this filer is a scsit simulator
            # get the devices the easy way
            # and skip the rest
            #
            if(ref($filer_obj) =~ /ScsitSimulator/){
                return $filer_obj->get_scsit_devices();
            }
            if($self->{any_iscsi}){
                my @ip_addr = $filer_obj->get_ip_addresses();
                return (wantarray()) ? @ip_addr : \@ip_addr;
            }

            my $filer_name = $filer_obj->get_name();
            if ($filer_obj->error())
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Error: San::get_devices() ",
                                "Call to get_name failed: ",
                                "$errmsg");
                return (wantarray()) ? () : [];
            }

            #
            # If the user specified a filer and this is not that filer,
            # go to the next filer.
            #
            if (defined $filer and ($filer !~ /$filer_name/))
            {
                $self->sanlog(-msg       => "Not getting devices for " .
                                            "filer $filer_name.",
                              -msg_level => 3);
                next;
            }

            $self->sanlog(-msg => "Get devices for filer $filer_name");

            #
            # Send the filer name to get_free_devices
            #
            $devArgs{-filer}   = $filer_name;
            $devArgs{-refresh} = 1;

            #
            # Get devices for each client for each filer
            #
            foreach my $client_obj (@{$self->{clients}})
            {
                #
                # Get the host name
                #
                my $client_name = $client_obj->hostname();
                if ($client_obj->error)
                {
                    my $errmsg = join "\n", $client_obj->errmsg();
                    $self->_set_err("Error: San::get_devices() ",
                                    "Call to hostname failed: ",
                                    "$errmsg");
                    return (wantarray()) ? () : [];
                }

                #
                # Get the host type
                #
                my $client_type = $client_obj->hosttype();
                if ($client_obj->error)
                {
                    my $errmsg = join "\n", $client_obj->errmsg();
                    $self->_set_err("Error: San::get_devices() ",
                                    "Call to hosttype failed: ",
                                    "$errmsg");
                    return (wantarray()) ? () : [];
                }
                $client_type =~ s/unix-solaris/solaris/; # strip off unix-

                #
                # If the user specified a client and this is not that client,
                # go to the next client.
                #
                if (defined $client and ($client !~ /$client_name/))
                {
                    $self->sanlog(-msg       => "Not getting devices for " .
                                                "client $client_name.",
                                  -msg_level => 3);
                    next;
                }

                $self->sanlog(-msg => "Get raw devices for client $client_name");

                #
                # Create a device manager object.
                #
                my $dev_mgr_obj =
                    $client_obj->create_devmgr(-parent=>$client_obj);
                if ($client_obj->error)
                {
                    my $errmsg = join "\n", $client_obj->errmsg();
                    $self->_set_err("Error: San::get_devices() ",
                                    "Call to create_devmgr failed: ",
                                    "$errmsg");
                    return (wantarray()) ? () : [];
                }

                if ($use_filer_vol_list)
                {
                    #
                    # For each entry in the filer_vol_list, I need to set
                    # devArgs: -vol and -cnt
                    #

                    #
                    # For each vol prefix we want to store the index where we
                    # are to avoid using vols of the same name.
                    #
                    my $vol_prefix_hash    = {};

                    foreach my $vol_entry_hash (@{$self->{filer_vol_list}})
                    {
                        my $vol_count    = $vol_entry_hash->{vol_count};
                        my $vol_prefix   = $vol_entry_hash->{vol_prefix};
                        my $lun_count    = $vol_entry_hash->{lun_count};

                        #
                        # Init the prefix count for this vol_prefix if necessary
                        #
                        $vol_prefix_hash->{$vol_prefix} = 0
                            unless defined $vol_prefix_hash->{$vol_prefix};

                        for (my $index = 1 ; $index <= $vol_count  ; $index++)
                        {
                            $devArgs{-cnt} = $lun_count;
                            $vol_prefix_hash->{$vol_prefix}++;
                            my $vol_name = $vol_prefix .
                                           $vol_prefix_hash->{$vol_prefix};
                            $devArgs{-vol}=$vol_name;

                            $self->sanlog(-msg => "Get raw devices for " .
                                                   "client $client_name" .
                                                   ":$vol_name");
                            my @devices;
                            #
                            # Call get_free_devices to get the actual devices we are
                            # interested in.
                            #
                            # Added retries for windows here: check the
                            # number of devices vs. the lun_count and if there are not
                            # enough devices, retry a few times (you can also check the
                            # errmsg from dev_mgr_obj and make sure it's what is expected:
                            # Only X of Y requested devices were found.  Unlike UNIX,
                            # windows doesn't reliable report it's luns.
                            #
                            my $retries = 0;
                            while ($retries < 5)
                            {
                                @devices =
                                  $dev_mgr_obj->get_free_devices(%devArgs);

                                #
                                # Only windows needs to retry.
                                #
                                if ($client_type ne "windows")
                                {
                                    last;
                                }
                                elsif (scalar(@devices) >= $lun_count)
                                {
                                    #
                                    # If we found all the luns we are done.
                                    #
                                    last;
                                }
                                $retries++;
                                $devArgs{-refresh} = 1;
                                $self->sanlog(-msg =>
                                       "San: Retrying get_free_devices() " .
                                       "call for windows ($retries)");
                                sleep(30);
                            } # end while retries
                            if ($dev_mgr_obj->error)
                            {
                                my $errmsg = join "\n", $dev_mgr_obj->errmsg();
                                $self->_set_err("San::get_devices: Client is "         .
                                                "$client_name, Filer is $filer_name. " .
                                                "$errmsg");
                                return (wantarray()) ? () : [];
                            }

                            #
                            # Now that we have the devices, lets push them on our array.
                            #
                            foreach my $device (@devices)
                            {
                                #
                                # With the with_device_info flag, get_devices
                                # will return array of hashes with device info:
                                # dev_name, dev_obj, and dev_size
                                #
                                if ($with_device_info)
                                {
                                    #
                                    # Get device info
                                    #
                                    my %device_info = $self->_get_device_info(
                                             -dev_mgr_obj => $dev_mgr_obj,
                                             -device      => $device,
                                             -device_type => $device_type,
                                             -client_type => $client_type);
                                    if ($self->error())
                                    {
                                        $self->_set_err("Failed to get device info ",
                                                        $self->errmsg());
                                        return (wantarray()) ? () : [];
                                    }

                                    push @device_array, \%device_info;
                                }
                                else
                                {
                                    my $fmt_device = $self->_format_device_string(
                                                                     $device,
                                                                     $device_type,
                                                                     $client_type);
                                    push @device_array, $fmt_device;
                                }

                                $self->sanlog(
                                  -msg_level => 3,
                                  -msg       => "San::get_devices pushed " .
                                                "device $device filer is $filer_name");
                            } # end foreach device

                            #
                            # We only want to refresh (call sanlun one time)
                            #
                            $devArgs{-refresh} = 0;

                        } # end for each volume

                    } # end for each filer vol list entry
                } # end if use_filer_vol_list
                else
                {
                    #
                    # Call get_free_devices to get the actual devices we are
                    # interested in.
                    #
                    my @devices = $dev_mgr_obj->get_free_devices(%devArgs);
                    if ($dev_mgr_obj->error)
                    {
                        my $errmsg = join "\n", $dev_mgr_obj->errmsg();
                        $self->_set_err("San::get_devices: Client is "         .
                                        "$client_name, Filer is $filer_name. " .
                                        "$errmsg");
                        return (wantarray()) ? () : [];
                    }

                    #
                    # Now that we have the devices, lets push them on our array.
                    #
                    foreach my $device (@devices)
                    {
                        #
                        # With the with_device_info flag, get_devices
                        # will return array of hash with device info:
                        # dev_name, dev_obj, and dev_size
                        #
                        if ($with_device_info)
                        {
                            #
                            # Get device info
                            #
                            my %device_info = $self->_get_device_info(-dev_mgr_obj => $dev_mgr_obj,
                                                                   -device      => $device,
                                                                   -device_type => $device_type,
                                                                   -client_type => $client_type);
                            if ($self->error())
                            {
                                $self->_set_err("Failed to get device info ",
                                                $self->errmsg());
                                return (wantarray()) ? () : [];
                            }

                            push @device_array, \%device_info;
                        }
                        else
                        {
                            my $fmt_device = $self->_format_device_string($device,
                                                                          $device_type,
                                                                          $client_type);

                            push @device_array, $fmt_device;
                        }

                        $self->sanlog(
                          -msg_level => 3,
                          -msg       => "San::get_devices pushed " .
                                        "device $device filer is $filer_name");
                    } # end foreach device
                } # end else, for use_filer_vol_list

                $log_string =
                    "Filer: $filer_name, Client: $client_name, Devices: ";
                if ($with_device_info)
                {
                    foreach my $dev (@device_array)
                    {
                        $log_string .= " " . $dev->{dev_name} .
                                       ":" . $dev->{dev_size};
                    }
                }
                else
                {
                    $log_string .= join(" ", @device_array);
                }
                $self->sanlog(-msg_level => 1,
                              -msg       => $log_string);
            } # end foreach client
        } # end foreach filer
    } # end if device type raw
    else
    {
        $self->_set_err("Error: San::get_devices()",
                        "Device type $device_type not supported.");
        return (wantarray()) ? () : [];
    }

    #
    # If no device is found, set error
    #
    unless (scalar(@device_array))
    {
        $self->_set_err("Unable to find device. " .
                        "Please verify client and target connectivity.");
        return (wantarray()) ? () : [];
    }

    #
    # Return the devices found
    #
    return (wantarray()) ? @device_array : \@device_array;
} # end get_devices()

########################################################################
# Method:    add_filer
# Objective: Given a filer name, create a SanFiler object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub add_filer()
{
    my($self,%args) = @_;
    my $filer_obj;
    my $filer_name;
    my $filer = $args{-filer};
    my $filer_pass_thru = {};
    my $ntp_sync = 0;
    my $ntp_time_server = undef;
    my $ntp_time_zone   = undef;

    $filer_pass_thru = $args{-filer_pass_thru}
            if defined($args{-filer_pass_thru});

    $ntp_sync = $self->{ntp_sync}
            if defined($self->{ntp_sync});

    $ntp_time_server = $self->{ntp_time_server}
            if defined($self->{ntp_time_server});

    $ntp_time_zone = $self->{ntp_time_zone}
            if defined($self->{ntp_time_zone});

    $self->_clear_err();

    #
    # If we are given a SanFiler object, use it.  Otherwise create a new one.
    #
    if (ref($filer) =~ "SanFiler")
    {
       $filer_obj = $filer;
       $filer_name = $filer_obj->get_name();
    }
    else
    {
        $filer_name = $filer;
    }

    #
    # Check to see if we already have a filer object for this filer
    # If so, just return
    #
    my $check_obj = $self->get_filer($filer_name);
    if ($check_obj)
    {
        $self->sanlog(-msg => "San::add_filer(): Using existing filer " .
                              "object for filer $filer_name",
                      -msg_level => 1);
        return 1;
    }

    $self->_clear_err();

    unless (ref($filer) =~ "SanFiler")
    {
        $self->sanlog(-msg => "San::add_filer(): " .
                              "Creating Filer object for filer  $filer_name");
        $filer_obj= new SanFiler(-filer       => $filer_name,
                                 -timeout     => $self->{timeout},
                                 -debug_level => $self->{debug_level},
                                 %{$filer_pass_thru},
                                );
        if ($filer_obj->error)
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->sanlog(-msg => $errmsg);

            $self->_set_err("Error: San::add_filer(): "                    .
                            "Instantiating Filer object for $filer_name. " .
                            "$errmsg");
            return $ClientUtil::FALSE;
        }
    }

    #
    # Time sync filer if specified
    #
    if ($ntp_sync == 1 && defined($ntp_time_server) &&
        defined($ntp_time_zone))
    {
        $self->sanlog(-msg => "Time syncing filer",
                      -msg_level => 1);
                      #-msg_level => LOG_VERBOSE_LOW);
        $filer_obj->ntp_sync(-ntp_server    => $ntp_time_server,
                             -ntp_time_zone => $ntp_time_zone);
        if ($filer_obj->error())
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->sanlog(-msg => $errmsg);

            $self->_set_err("Error: San::add_filer()",
                            "Couldn't time sync Filer object for $filer_name.",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }
    }

    push @{$self->{filers}}, $filer_obj;
    return 1;
} # end add_filer

########################################################################
# Method:    add_client
# Objective: Given a client name, create a SanClient object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub add_client()
{
    my($self,%args) = @_;
    my $client_obj = undef;
    my $client;
    my $client_name;
    my $device_type = "raw";
    my $protocol    = "fcp";
    my $mp_type     = "none";
    my $any_iscsi = undef;
    my $client_pass_thru = {};

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?client$/)
        {
           $client = $args{$_};
           next;
        }
        elsif (/-?device_type$/)
        {
            $device_type = $args{$_};
            if (not length($device_type))
            {
                $self->_set_err("Error: San::add_client()",
                                "Invalid value for argument '-device_type'");
                return $self;
            }
            next;
        }
        elsif (/-?protocol$/)
        {
            $protocol = $args{$_};
            if (not length($protocol))
            {
                $self->_set_err("Error: San::add_client()",
                                "Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        }
        elsif (/-?mp_type/)
        {
            $mp_type=$args{$_};
            if (not length($mp_type))
            {
                $self->_set_err("Error: San::add_client()",
                                "Invalid value for argument '-mp_type'");
                return $self;
            }
            next;
        }
        elsif (/-?client_pass_thru/)
        {
            $client_pass_thru = $args{$_};
            unless(ref($client_pass_thru) =~ /HASH/)
            {
                $self->_set_err("Error: San::setup() "             .
                                "Invalid value for argument type " .
                                "'-client_pass_thru', HASH type expected");
                return $self;
            }
        }
    } # end foreach arg

    #
    # If we are given a SanClient object, use it.  Otherwise create a new one.
    #
    if (ref($client) =~ "SanClient")
    {
       $client_obj = $client;
       $client_name = $client_obj->hostname();
    }
    else
    {
        $client_name = $client;
    }

    #
    # Check to see if we already have a client object for this client
    # If so, just return
    #
    my $check_obj = $self->get_client($client_name);
    if ($check_obj)
    {
        $self->sanlog(-msg => "San::add_client(): Using existing client " .
                              "object for client $client_name",
                      -msg_level => 1);
        return 1;
    }

    $self->_clear_err();

    #
    # Unless we are given a SanClient object, create a new one.
    #
    unless (ref($client) =~ "SanClient")
    {
        $self->sanlog(-msg => "San::add_client(): Creating Client object " .
                              "for client $client_name");

        my ($mp_type_ret_ref, $any_iscsi_ret_ref);

        #
        #Coupled mp_type and any_iscsi here, to use
        #one you have to use both
        #
        if($mp_type =~ /^per_host$/i)
        {
            $mp_type_ret_ref = $self->_get_pc_mp_type($mp_type, $client);

            if($mp_type_ret_ref)
            {
                $mp_type = $mp_type_ret_ref->{new_mp_type};
            }
            else
            {
                $self->_set_err("Error: San::setup() Instantiating client ".
                                "object for $client. CLIENT_VMS set to ".
                                "per_host but PC_CLIENT_VMS is not set ".
                                "in the tharnhost file for $client.");
                return $self;
            }
        }

        #
        #Always check for any_iscsi.  If it's not there, fine,
        #but if it is then we will continue with the any_iscsi
        #implementation of changing the target IQN
        #
        $any_iscsi_ret_ref = $self->_get_pc_any_iscsi($client);

        if($any_iscsi_ret_ref)
        {
            $any_iscsi = $any_iscsi_ret_ref->{new_any_iscsi};
        }
        else
        {
            $self->_set_err("Error: San::setup() Instantiating client ".
                            "object for $client. hosttype is unix/linux".
                            "but PC_ANY_ISCSI param is not set in client".
                            "tharnhost file for $client.");
            return $self;
        }

        if($any_iscsi)
        {

            $client_obj= new SanClient(-host        => $client_name,
                                       -timeout     => $self->{timeout},
                                       -debug_level => $self->{debug_level},
                                       -device_type => $device_type,
                                       -protocol    => $protocol,
                                       -mp_type     => $mp_type,
                                       -any_iscsi    => $any_iscsi,
                                       %{$client_pass_thru});
        }
        else
        {

            $client_obj= new SanClient(-host        => $client_name,
                                       -timeout     => $self->{timeout},
                                       -debug_level => $self->{debug_level},
                                       -device_type => $device_type,
                                       -protocol    => $protocol,
                                       -mp_type     => $mp_type,
                                       %{$client_pass_thru});
        }

        if($mp_type_ret_ref)
        {
            &ClientUtil::restore_original_params($mp_type_ret_ref->{old_client_params});
        }
        if($any_iscsi_ret_ref)
        {
            &ClientUtil::restore_original_params($any_iscsi_ret_ref->{old_client_params});
        }

        if ($client_obj->error)
        {
            my $errmsg = join "\n", $client_obj->errmsg();
            $self->sanlog(-msg => $errmsg);

            $self->_set_err("Error: San::add_client()",
                            "Instantiating client object for $client.",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }
    }

    push @{$self->{clients}}, $client_obj;
    return 1;
} # end add_client

########################################################################
# Method:    get_filer
# Objective: Given a filer name, return a SanFiler object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub get_filer()
{
    my ($self, $name) = @_;

    $self->_clear_err();

    unless(defined $name and length $name)
    {
        $self->_set_err("San::get_filer() filer name is a required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Look at each filer object in the sand and see if the filer name
    # matches the name we have been given.
    #
    foreach my $filer_obj ( @{$self->{filers}})
    {
        if ($filer_obj->get_name() =~ /^$name$/)
        {
            return $filer_obj;
        }
    } # end foreach

    $self->_set_err("San::get_filer() Unable to find filer '$name'");
    return $ClientUtil::FALSE;
}

########################################################################
# Method:    get_client
# Objective: Given a client name, return a SanClient object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub get_client()
{
    my ($self,$name) = @_;

    $self->_clear_err();

    unless(defined $name and length $name)
    {
        $self->_set_err("San::get_client() client name is a " .
                        "required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Look at each client object in the sand and see if the client name
    # matches the name we have been given.
    #
    foreach my $client_obj ( @{$self->{clients}})
    {
        if ($client_obj->hostname() =~ /^$name$/)
        {
            return $client_obj;
        }
    } # end foreach

    $self->_set_err("San::get_client() Unable to find client '$name'");
    return $ClientUtil::FALSE;
}

########################################################################
# Method:    get_filer_vol_list
# Objective: Given an aggr, vol_prefix, vol_size, lun list return a
#            filer vol list.
# Details:  Here is the syntax for a FILER_VOL_LIST:
#
# Colon separated list of volumes where each entry has the following
# syntax:
#     vol_count, vol_size, lun_count, lun_size [,type, prefix, aggr, options]
#
#     vol_count - required number of volumes to create
#     vol_size  - required volume size (for trad vols use number of disks)
#     lun_count - required number of luns to create on each volume
#     lun_size  - required size of luns created
#     type      - optional flex or trad (default is flex)
#     prefix    - optional string at the begging of each volume name
#                          if you are creating 2 vols and the prefix
#                          is "x_vol_", we create x_vol_1 and x_vol_2
#                          (default is san_vol_)
#    aggr       - optional name of the aggr for flex vols
#                          (default is test_aggr)
#    options    - optional string appended to the vol options command
#                           line which will be called after vol create
#                           (default is "")
#    If redundant vol prefix values used, each colon separated list of volumes
#    would have a vol name that is incremented from the previous.
#    For Example:  FILER_VOL_LIST=3,100g,2,10g:1,1000g,1,990g
#      - This would create the following set of vols(luns):
#      san_vol_1(10g,10g), san_vol_2(10g,10g), san_vol_3(10g,10g),
#      san_vol_4(990g)
#
########################################################################
sub get_filer_vol_list()
{
    my($self,%args) = @_;
    my $aggr;
    my $vol;
    my $vol_type = "flex";
    my $vol_size;
    my $lun_list;

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?aggr/)
        {
            $aggr = $args{$_};
            next;
        }
        elsif (/-?vol$/)
        {
            $vol = $args{$_};
            if (not length($vol))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_size$/)
        {
            $vol_size = $args{$_};
            if (not length($vol_size))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-vol_size'");
                return $self;
            }
            next;
        }
        elsif (/-?lun_list$/)
        {
            $lun_list = $args{$_};
            if (not length($lun_list))
            {
                $self->_set_err("Error: San::setup()",
                                "Invalid value for argument '-lun_list'");
                return $self;
            }
            next;
        }
    } # end foreach arg

    $aggr = "" unless (defined($aggr) and length($aggr));

    #
    # If no aggr is given use a trad vol
    #
    unless (defined $aggr and length $aggr)
    {
        $vol_type = "trad";
    }


    #
    # TBD, you may want to split the lun_list by ":" and iterate thru
    # the lun_list entries to have full support of the older lun_list
    # to do this, however you would need to have some way of repeating
    # vol list entries which we don't have so it may not be worth it.
    #
    # vol_count, vol_size, lun_count, lun_size [,type, prefix, aggr, options]
    my $short_vol = $self->_get_short_vol($vol);
    my $vol_list = "1, $vol_size, $lun_list, $vol_type, $short_vol, $aggr,";
    return $vol_list;
} # end get_filer_vol_list


########################################################################
# Method:    get_resources
# Objective: Return an array of hash of used resources
# Details:   See POD documentation at the end of this file.
########################################################################
sub get_resources()
{
    my($self,%args) = @_;

    my $type  = undef;
    my $name  = undef;
    my $cmd   = undef;
    my $reuse = undef;
    my @cleanup_array  = @{$self->{cleanup}};
    my $cleanup_hash   = {};
    my @resource_array = ();

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?type/)
        {
            $type = $args{$_};
            if (not length($type))
            {
                $self->_set_err("Error: San::get_resources()",
                                "Invalid value for argument '-type'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?name/)
        {
            $name = $args{$_};
            if (not length($name))
            {
                $self->_set_err("Error: San::get_resources()",
                                "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?cmd/)
        {
            $cmd = $args{$_};
            if (not length($cmd))
            {
                $self->_set_err("Error: San::get_resources()",
                                "Invalid value for argument '-cmd'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        elsif (/-?reuse/)
        {
            $reuse = $args{$_};
            next;
        }
        else
        {
            $self->sanlog(-msg => "San::get_resources ignoring argument $_");
        }
    } # end foreach arg

    #
    # Push the entry onto the stack
    #
    foreach my $cleanup_hash (@cleanup_array)
    {
        #
        # Translate the command from a cleanup type to a setup action
        #
        my $cmd_temp = $cleanup_hash->{-cmd};
        $cmd_temp =~ s/aggr_destroy/aggr_create/g;
        $cmd_temp =~ s/vol_destroy/vol_create/g;
        $cmd_temp =~ s/lun_destroy/lun_create/g;
        $cmd_temp =~ s/igroup_destroy/igroup_create/g;
        $cmd_temp =~ s/lun_unmap/lun_map/g;

        #
        # Check the type, name, cmd, reuse filters
        #
        next if defined($type)  && $cleanup_hash->{-type} ne "$type";
        next if defined($name)  && $cleanup_hash->{-name} ne "$name";
        next if defined($cmd)   && $cmd_temp ne "$cmd";
        next if defined($reuse) && $cleanup_hash->{-reuse} ne "$reuse";

        my $resource_hash = {};
        $resource_hash->{-type}     = $cleanup_hash->{-type};
        $resource_hash->{-name}     = $cleanup_hash->{-name};
        $resource_hash->{-cmd_args} = $cleanup_hash->{-cmd_args};
        $resource_hash->{-reuse}    = $cleanup_hash->{-reuse};
        $resource_hash->{-cmd} = $cmd_temp;
        $resource_hash->{-vfiler_context} = $cleanup_hash->{-vfiler_context};
        push(@resource_array, $resource_hash);
    } # end for each entry

    return @resource_array;
} # end get_resources

###############################################################################
#                               Private Methods
###############################################################################


###############################################################################
# Method:    _parse_filer_vol_list
# Objective: Create a hash of hash to represent the entries in a filer_vol_list
# Inputs:    filer_vol_list - aggregate to use
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
# Notes:     An attribute named filer_vol_list will be stored as part of the
#            San Object.
###############################################################################
sub _parse_filer_vol_list
{
    my ($self, $filer_vol_list) = @_;
    my $syntax = "vol_count, vol_size, lun_count, lun_size " .
                 "[,type, prefix, aggr, options]";


    $self->sanlog(-msg       => "Filer vol list is $filer_vol_list",
                  -msg_level => 7);
    #
    # The filer vol list is a colon separated list where each entry
    # describes volumes and their luns.
    #
    my @list_entries  = split(/:/, $filer_vol_list);
    foreach my $entry (@list_entries)
    {
        $self->sanlog(-msg       => "Filer vol list entry is $entry",
                      -msg_level => 7);
        #
        # The lun_count can be a simple integer or a range (4-6 for example).
        # If a range is given, one day we might make the lun map respect this
        # value as the lun ID to map to (4-4 would thus be legal).  If an
        # integer is given, we want -1 as the lun id just so we know to take
        # whatever default the filer gives us (which is what we currently do
        # for everything anyway).
        #
        my $first_lun_id = -1;
        my $reserve      =  1;

        my ($vol_count,
            $vol_size,
            $lun_count,
            $lun_size,
            $vol_type,
            $vol_prefix,
            $aggr,
            $vol_options,
            $lun_prefix) = split(/,/, $entry);

        #
        # Check to make sure the first four values are defined because
        # they are required.
        #
        unless (defined($vol_count) and
                defined($vol_size)  and
                defined($lun_count) and
                defined($lun_size)
               )
        {
            my $errmsg = "Error: San::_parse_filer_vol_list(): " .
                         "the filer_vol_list requires the following fields: " .
                         "vol_count, vol_size, lun_count, lun_size\n"         .
                         $syntax;
            $self->sanlog(-msg => $errmsg);
            $self->_set_err($errmsg);
            return 0;
        }

        #
        # If any of the other values are not set, either set them to valid
        # defaults or to the empty string
        #
        $vol_type    = "flex"     unless (defined($vol_type));
        $vol_prefix  = "san_vol_" unless (defined($vol_prefix));
        $vol_options = ""         unless (defined($vol_options));
        if ($vol_type eq "flex")
        {
            $aggr = "test_aggr" unless (defined($aggr));
        }
        else
        {
            $aggr = "" unless (defined($aggr));
        }

        #
        # Take a look at the lun_count and see if it's an integer
        # or a range.  If it's a range, figure out the count and
        # the first_lun_id.
        #
        if ($lun_count =~ /\-/)
        {
            my ($begin, $end) = split(/\-/,$lun_count);
            $first_lun_id = $begin;
            $lun_count = ($end - $begin) + 1;
        }


        #
        # This is a work around to enable/disable the reserve option
        # when creating a lun.  The reserve option will be expected if there is
        # a dash "-" after the lun_size
        #
        if ($lun_size =~ /\-/)
        {
            my ($size, $lun_option) = split(/\-/,$lun_size);

            #
            # make sure lun_option is valid; othervise default to 1 (reserve)
            #
            $reserve  = $lun_option  if ($lun_option =~ /(0|n|no|f|false)/);
            $lun_size = $size;
        }

        #
        # If the vol_prefix has dashes in it, replace with underscores
        # Dashes "-" are not legal and we have tests that use the client
        # name as the vol_prefix.  Client names often have dashes, it's
        # easy to replace here.
        #
        $vol_prefix =~ s/-/_/g;

        #
        # If the vol_prefix is of the form xxx/qqq then we are
        # going to split on the "/" and the xxx will be the
        # vol_prefix and the qqq will be the qtree prefix
        #
        my $qtree_prefix = "";
        my $qtree_count  = 0;
        if ($vol_prefix =~ /\S+\/\S+/)
        {
            $self->sanlog(
                -msg => "The vol_prefix ($vol_prefix) contains a qtree",
                -msg_level => 7);
            ($vol_prefix, $qtree_prefix) = split(/\//, $vol_prefix);
            $qtree_count = 1;
        } # end if qtree name is vol_prefix

        my $entry_hash = {};
        $entry_hash->{vol_count}    = $vol_count;
        $entry_hash->{vol_size}     = $vol_size;
        $entry_hash->{lun_count}    = $lun_count;
        $entry_hash->{lun_size}     = $lun_size;
        $entry_hash->{reserve}      = $reserve;
        $entry_hash->{first_lun_id} = $first_lun_id;
        $entry_hash->{vol_type}     = $vol_type;
        $entry_hash->{vol_prefix}   = $vol_prefix;
        $entry_hash->{vol_options}  = $vol_options;
        $entry_hash->{qtree_count}  = $qtree_count;
        $entry_hash->{qtree_prefix} = $qtree_prefix;
        $entry_hash->{aggr}         = $aggr;
        $entry_hash->{reuse_policy} = "reuse"; # TBD
        $entry_hash->{lun_prefix}   = $lun_prefix;

        $self->sanlog(-msg       => "Filer vol list entry stored:\n" .
                      "vol count    is $vol_count\n"    .
                      "vol size     is $vol_size\n"     .
                      "lun count    is $lun_count\n"    .
                      "lun size     is $lun_size\n"     .
                      "lun reserve  is $reserve\n"      .
                      "lun id       is $first_lun_id\n" .
                      "vol type     is $vol_type\n"     .
                      "vol prefix   is $vol_prefix\n"   .
                      "vol options  is $vol_options\n"  .
                      "qtree prefix is $qtree_prefix\n" .
                      "qtree count  is $qtree_count\n"  .
                      "aggr         is $aggr\n"         .
                      "reuse policy is reuse\n" .
                      "lun prefix   is $lun_prefix\n",

                      -msg_level => 7);

        push @{$self->{filer_vol_list}}, $entry_hash;

    } # end foreach entry
    return 1;
} # end _parse_filer_vol_list

###############################################################################
# Method:    _aggr_vol_lun_setup
# Objective: Create or find the necessary volume.
# Inputs:    aggr - aggregate to use
#            num_disks - disks to use to create aggr or trad vol
#            protocol  - fcp or iscsi
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
# Notes:     This method relies on the _parse_filer_vol_list method having
#            been called and populated the $self->{filer_vol_list} array.
#            This method has been closely replicated in _vfiler_setup, if
#            making changes here, please consider the changes needed in
#            _vfiler_setup.
###############################################################################
sub _aggr_vol_lun_setup
{
    my ($self, $aggr, $num_disks, $protocol) = @_;
    my $aggr_vol_type = "flex"; # Used to determine aggr create.
    my @orphaned_clients;
    my $blocktype;
    my $raidtype;
    my $timeout = 14400;

    $aggr = "" unless (defined($aggr) and length($aggr));

    #
    # If no aggr is given use a trad vol, i.e. don't create
    # an aggr.
    #
    unless (defined $aggr and length $aggr)
    {
        $aggr_vol_type = "trad";
    }

    foreach my $filer_obj (@{$self->{filers}})
    {
        my $filer_name = $filer_obj->get_name();
        if ($filer_obj->error)
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->_set_err("Error: San::_aggr_vol_lun_setup() ",
                            "Call to get_name failed: ",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }

        #
        # If we are using a flex vol, we will create the aggr then the vol.
        #
        if ($aggr_vol_type =~ /flex/)
        {
            my @aggr_array = split(/,\s*/, $aggr);
            my $aggr_name  = $aggr_array[0];

            $self->sanlog(-msg => "$filer_name - setup aggregate " .
                                  "$aggr using "                   .
                                  "$num_disks disks");

            #
            # Try to create the aggregate
            #
            # The aggr valus can also be split like:
            # aggr_name, blocktype = 64, raidtype = raid4, etc.
            # Values are grabbed here.
            #
            for(my $i = 1; $i < scalar(@aggr_array); $i++)
            {
                my ($option, $value) = split(/\s*=\s*/, $aggr_array[$i]);
                if($option =~ /^blocktype/)
                {
                    $blocktype = $value;
                }
                elsif($option =~ /^raidtype/)
                {
                    $raidtype = $value;
                }
                elsif($option =~ /^timeout/)
                {
                    $timeout = $value;
                }
                else
                {
                    $self->sanlog(-msg => "Invalid option: $option passed " .
                                          "through aggr");
                }
            }

            my %args = ("-name",      $aggr_name,
                        "-num_disks", $num_disks,
                        "-timeout",   $timeout);
            $args{"-blocktype"}  = $blocktype  if (defined $blocktype);
            $args{"-raidtype"} = $raidtype if (defined $raidtype);

            $filer_obj->aggr_create(%args);

            if ($filer_obj->error())
            {
                #
                # If we blew past the timeout, it's an error.
                #
                my $errmsg = join "\n", $filer_obj->errmsg();
                if ($errmsg =~ /wait_for_aggr_online(): Timeout/)
                {
                    $self->sanlog(-msg => $errmsg);
                    $self->_set_err("Error: San::_aggr_vol_lun_setup() " .
                                    "Timed out waiting for aggr_create " .
                                    $aggr_name .
                                    " $errmsg");
                    return $ClientUtil::FALSE;
                }

                #
                # Create failed, lets see if it exists
                #
                $filer_obj->aggr_exist($aggr_name);
                if ($filer_obj->error())
                {
                    #
                    # If we get here we are hosed because the create failed
                    # and we have checked and it doesn't exist.
                    #
                    $errmsg = join "\n", $filer_obj->errmsg();
                    $self->sanlog(-msg => $errmsg);
                    $self->_set_err("Error: San::_aggr_vol_lun_setup() " .
                                    "Getting aggr $aggr_name "           .
                                    "$errmsg");
                    return $ClientUtil::FALSE;
                }

                #
                # If we get here we know that the create failed because
                # the aggr already exists.
                #
                $self->sanlog(-msg => "Using existing aggr $aggr_name " .
                                      "on filer " . $filer_name);
                return $ClientUtil::FALSE
                    unless ($self->cleanup_add(-type     => "filer",
                                               -name     => $filer_name,
                                               -cmd      => "aggr_destroy",
                                               -cmd_args => $aggr_name,
                                               -reuse    => 1));
            }
            else
            {
                sleep(30); # wait so aggr is available
                return $ClientUtil::FALSE
                    unless ($self->cleanup_add(-type     => "filer",
                                               -name     => $filer_name,
                                               -cmd      => "aggr_destroy",
                                               -cmd_args => $aggr_name));
            }

            $self->sanlog(-msg => "Waiting for aggr $aggr_name to come online",
                          -msg_level => 1);
                          #-msg_level => $LOG_VERBOSE_LOW);
            $filer_obj->wait_for_aggr_online(-name => $aggr_name);
        } # end if aggr create for flex vol

        #
        # if Vfiler_Enable is set to one, call vfiler_setup to
        # create the vfilers per client.  we need to refresh the
        # iscsi target list after creating the vfilers
        #
        if ($self->{vfiler_enable})
        {
            return $ClientUtil::FALSE
                unless $self->_vfiler_setup(-filer_obj    => $filer_obj,
                                            -protocol     => $protocol,
                                            -orph_clients => \@orphaned_clients);

            return $ClientUtil::FALSE
                unless $filer_obj->refresh_iscsi_targets();
        } # end if Vfiler_Enable
        else
        {

            #
            # Use the San object's filer_vol_list array to create volumes
            # and luns on the default vfiler
            #

            #
            # For each vol prefix we want to store the index where we
            # are to avoid creating vols of the same name.
            #
            my $vol_prefix_hash    = {};

            foreach my $vol_entry_hash (@{$self->{filer_vol_list}})
            {
                my $vol_count    = $vol_entry_hash->{vol_count};
                my $vol_prefix   = $vol_entry_hash->{vol_prefix};

                #
                # If vol_prefix is an existing volume and vol_count is set
                # to 1, don't use prefixes. Assume user wants to use that
                # exact volume
                #
                if ( $vol_count == 1 && $filer_obj->vol_exist($vol_prefix) )
                {
                    return $ClientUtil::FALSE
                        unless ($self->_vol_setup(-filer_obj      => $filer_obj,
                                                  -vol_name       => "/vol/${vol_prefix}",
                                                  -vol_entry_hash => $vol_entry_hash));
                    return $ClientUtil::FALSE
                        unless ($self->_lun_setup(-filer_obj      => $filer_obj,
                                                  -vol_prefix     => "/vol/${vol_prefix}",
                                                  -protocol       => $protocol,
                                                  -orph_clients   => \@orphaned_clients,
                                                  -vol_entry_hash => $vol_entry_hash,
                                                  -clients        => \@{$self->{clients}}));
                    next;
                }

                #
                # Init the prefix count for this vol_prefix if necessary
                #
                $vol_prefix_hash->{$vol_prefix} = 0
                    unless defined $vol_prefix_hash->{$vol_prefix};

                for (my $index = 1 ; $index <= $vol_count  ; $index++)
                {
                    $vol_prefix_hash->{$vol_prefix}++;
                    my $vol_name = "/vol/$vol_prefix" .
                                   $vol_prefix_hash->{$vol_prefix};

                    return $ClientUtil::FALSE
                        unless ($self->_vol_setup(-filer_obj      => $filer_obj,
                                                  -vol_name       => $vol_name,
                                                  -vol_entry_hash => $vol_entry_hash));

                    return $ClientUtil::FALSE
                        unless ($self->_lun_setup(-filer_obj      => $filer_obj,
                                                  -vol_prefix     => $vol_name,
                                                  -protocol       => $protocol,
                                                  -orph_clients   => \@orphaned_clients,
                                                  -vol_entry_hash => $vol_entry_hash,
                                                  -clients        => \@{$self->{clients}}));
                } # end for each volume

            } # end for each filer vol list entry
        } # end else (default vfiler vol/lun setup)
    } # end for each filer

    return 1;
} # end of _aggr_vol_lun_setup method

###############################################################################
# Method:    _vol_setup
# Objective: Create or find the necessary volume.
# Inputs:    vol_name - Name of the vol derived from the prefix
#            vol_hash - Description of the vol
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
###############################################################################
sub _vol_setup
{
    my ($self, %args) = @_;

    my ($filer_obj, $vol_name, $vol_entry_hash);

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?filer_obj/)
        {
            $filer_obj = $args{$_};
            next;
        }
        elsif (/-?vol_name$/)
        {
            $vol_name = $args{$_};
            if (not length($vol_name))
            {
                $self->_set_err("Invalid value for argument '-vol_name'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_entry_hash$/)
        {
            $vol_entry_hash = $args{$_};
            if (not length($vol_entry_hash))
            {
                $self->_set_err("Invalid value for argument '-vol_entry_hash'");
                return $self;
            }
            next;
        }
    } # end foreach arg

    #
    #If the filer object is a ScsitSimulator, then
    #All of this is a no-op, so we'll just return
    #success and skip it
    #
    if(ref($filer_obj) =~ /ScsitSimulator/){
        return 1;
    }
    my $vol_count    = $vol_entry_hash->{vol_count};
    my $vol_size     = $vol_entry_hash->{vol_size};
    my $vol_type     = $vol_entry_hash->{vol_type};
    my $vol_prefix   = $vol_entry_hash->{vol_prefix};
    my $vol_options  = $vol_entry_hash->{vol_options};
    my $vol_aggr     = $vol_entry_hash->{aggr};
    my $vol_reuse    = $vol_entry_hash->{reuse_policy};
    my $qtree_prefix = $vol_entry_hash->{qtree_prefix};
    my $qtree_count  = $vol_entry_hash->{qtree_count};
    my $reserve      = $vol_entry_hash->{reserve};

    my $filer_name = $filer_obj->get_name();

    $self->sanlog(-msg => "In San::_vol_setup with "                    .
                          "$filer_name - setup vol $vol_name "          .
                          "with size $vol_size $vol_type $vol_options " .
                          "$vol_aggr $vol_reuse");

    if ($vol_type eq "flex")
    {
        #
        # Try to create the flex volume
        #
        $filer_obj->create_vol(-name=>$vol_name,
                               -aggr=>$vol_aggr,
                               -size=>$vol_size);

        if ($filer_obj->error())
        {
            #
            # Create failed, lets see if it exists
            #
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->sanlog(-msg => "$filer_name - setup flex vol $vol_name " .
                                  "with size $vol_size error is $errmsg",
                          -msg_level => 5);

            $filer_obj->vol_exist($vol_name);
            if ($filer_obj->error())
            {
                #
                # If we get here we are hosed because the create failed
                # and we have checked and it doesn't exist.
                #
                $errmsg .= join "\n", $filer_obj->errmsg();
                $self->sanlog(-msg => $errmsg);
                $self->_set_err("Error: San::_vol_setup()",
                    "Creating volume $vol_name ",
                    "$errmsg");
                return $ClientUtil::FALSE;
            }

            #
            # If we get here we know that the create failed because
            # the vol already exists.
            #
            $self->sanlog(-msg => "Using existing vol $vol_name " .
                                  "on filer " . $filer_name);
            return $ClientUtil::FALSE
                unless ($self->cleanup_add(
                          -type=>"filer",
                          -name=>$filer_name,
                          -cmd=>"vol_destroy",
                          -cmd_args=>"$vol_name",
                          -reuse=>1));
        }
        else
        {
            #
            # Register the cleanup action for vol create
            #
            return $ClientUtil::FALSE
                unless ($self->cleanup_add(
                          -type=>"filer",
                          -name=>$filer_name,
                          -cmd=>"vol_destroy",
                          -cmd_args=>"$vol_name"));
        }
    } # end if flex vol
    else
    {
        $self->sanlog(-msg => "$filer_name - setup trad vol $vol_name " .
                              "using $vol_size disks");

        #
        # Try to create a trad volume
        #
        $filer_obj->create_vol(-name=>$vol_name,
                               -disk_cnt=>$vol_size);

        if ($filer_obj->error())
        {
            #
            # Create failed, lets see if it exists
            #
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->sanlog(-msg => "$filer_name - setup trad vol $vol_name " .
                                  "with $vol_size disks is $errmsg");

            $filer_obj->vol_exist($vol_name);
            if ($filer_obj->error())
            {
                #
                # If we get here we are hosed because the create failed
                # and we have checked and it doesn't exist.
                #
                $errmsg .= join "\n", $filer_obj->errmsg();
                $self->sanlog(-msg => $errmsg);
                $self->_set_err("Error: San::_vol_setup()",
                                "Creating volume $vol_name ",
                                "$errmsg");
                return $ClientUtil::FALSE;
            }

            #
            # If we get here we know that the create failed because
            # the vol already exists.
            #
            $self->sanlog(-msg => "Using existing vol $vol_name " .
                                  "on filer " . $filer_name);
            return $ClientUtil::FALSE
                unless ($self->cleanup_add(
                          -type=>"filer",
                          -name=>$filer_name,
                          -cmd=>"vol_destroy",
                          -cmd_args=>"$vol_name",
                          -reuse=>1));
        }
        else
        {
            #
            # Register the cleanup action for vol create
            #
            return $ClientUtil::FALSE
                unless ($self->cleanup_add(
                          -type=>"filer",
                          -name=>$filer_name,
                          -cmd=>"vol_destroy",
                          -cmd_args=>"$vol_name"));
        }
    } # end else for trad vol


    #
    # If there are vol options passed in set them
    #
    if (length($vol_options))
    {
        my $short_vol = $self->_get_short_vol($vol_name);
        my $fvol_obj = $filer_obj->get_vol($short_vol);
        if ($filer_obj->error)
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->_set_err("Error: San::_vol_setup() ",
                            "Call to get_vol failed: ",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }

        my @options_array = split(/\s+/, $vol_options);
        while (scalar(@options_array))
        {
            my $parameter = shift(@options_array);
            my $value     = shift(@options_array);

            $self->sanlog(-msg => "Setting options on vol $short_vol: " .
                                  "$parameter $value");
            $parameter = "-" . $parameter;

            $fvol_obj->options($parameter => $value);
            if ($fvol_obj->error)
            {
                my $errmsg = join "\n", $fvol_obj->errmsg();
                $self->_set_err("Error: San::_vol_setup() ",
                                "Can not set volume options: ",
                                "$errmsg");
                return $ClientUtil::FALSE;
            }
        } # end while
    } # end if options

    #
    # If there is a qtree_prefix, use it as a qtree name.  TBD - eventually
    # we can add a loop to handle multiple qtrees
    #
    if (length($qtree_prefix))
    {
        my $qtree_name = $qtree_prefix;
        my $short_vol = $self->_get_short_vol($vol_name);
        my $fvol_obj = $filer_obj->get_vol($short_vol);
        if ($filer_obj->error)
        {
            my $errmsg = join "\n", $filer_obj->errmsg();
            $self->_set_err("Error: San::_vol_setup() ",
                            "Call to get_vol failed: ",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }

        if(ref($filer_obj) =~ /Filer10M/)
        {
            $fvol_obj->create_qtree_obj(-name => $qtree_name);
        }
        else
        {
            $fvol_obj->create_qtree(-name => $qtree_name);
        }

        if ($fvol_obj->error)
        {
            #
            # Create failed, lets see if it exists
            #
            my $errmsg = join "\n", $fvol_obj->errmsg();
            if (($errmsg =~ /\(File exists\)/) or ($errmsg =~ /already exists/))
            {
                $self->sanlog(-msg => "Using existing qtree " .
                                      "$short_vol/$qtree_name on filer " .
                                      $filer_obj->get_name());

            }
            else
            {
                $self->_set_err(
                    "Error: San::_vol_setup() Failed to create qtree " .
                    "$qtree_name on volume $short_vol" . ": $errmsg");
                return $ClientUtil::FALSE;
            }
        }
    } #end if qtree_prefix

    return 1;
} # end of _vol_setup()

###############################################################################
# Method:    _lun_setup
# Objective: Create the necessary luns.
# Inputs:    vol_name - Name of the vol derived from the prefix
#            protocol - fcp or iscsi
#            vol_hash - Description of the vol
#            client_array_ref - Reference to array of clients objs to use the luns
#                             - this is needed for vfiler case
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
###############################################################################
sub _lun_setup
{
    my ($self, %args) = @_;

    my ($filer_obj, $vol_name, $protocol,
        $orphaned_clients_ref, $vol_entry_hash,
        $client_array_ref);

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?filer_obj/)
        {
            $filer_obj = $args{$_};
            next;
        }
        elsif (/-?vol_prefix$/)
        {
            $vol_name = $args{$_};
            if (not length($vol_name))
            {
                $self->_set_err("Invalid value for argument '-vol_name'");
                return $self;
            }
            next;
        }
        elsif (/-?protocol$/)
        {
            $protocol = $args{$_};
            if (not length($protocol))
            {
                $self->_set_err("Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        }
        elsif (/-?orph_clients$/)
        {
            $orphaned_clients_ref = $args{$_};
            if (not length($orphaned_clients_ref))
            {
                $self->_set_err("Invalid value for argument '-orph_clients'");
                return $self;
            }
            next;
        }
        elsif (/-?vol_entry_hash$/)
        {
            $vol_entry_hash = $args{$_};
            if (not length($vol_entry_hash))
            {
                $self->_set_err("Invalid value for argument '-vol_entry_hash'");
                return $self;
            }
            next;
        }
        elsif (/-?clients$/)
        {
            $client_array_ref = $args{$_};
            if (not length($client_array_ref))
            {
                $self->_set_err("Invalid value for argument '-clients'");
                return $self;
            }
            next;
        }
    } # end foreach arg

    #
    #If the filer object is a ScsitSimulator, then
    #All of this is a no-op, so we'll just return
    #success and skip it
    #
    if(ref($filer_obj) =~ /ScsitSimulator/){
        return 1;
    }
    my $lun_count    = $vol_entry_hash->{lun_count};
    my $lun_size     = $vol_entry_hash->{lun_size};
    my $first_lun_id = $vol_entry_hash->{first_lun_id};
    my $qtree_prefix = $vol_entry_hash->{qtree_prefix};
    my $reserve      = $vol_entry_hash->{reserve};
    my $vol_aggr     = $vol_entry_hash->{aggr};
    my $alua_support = 0;

    $alua_support = $self->{alua_support} if defined($self->{alua_support});
    my $lun_prefix   = $vol_entry_hash->{lun_prefix};

    my $filer_name = $filer_obj->get_name();
    my $vfiler_name;
    #
    # We might be calling lun_setup from within a vfiler context
    # get the current context to save in the cleanup stack
    # this will be needed in order to cleanup correctly
    #
    if ($self->{vfiler_enable}) {
        $vfiler_name = $filer_obj->vfiler();
    }

    $self->sanlog(-msg => "In San::_lun_setup with "                    .
                          "$filer_name ");


    #
    # Create the luns.  We have a lun count and a lun size.  We create the
    # described luns one time on this volume for each client
    #
    foreach my $client_obj (@{$client_array_ref})
    {
        #
        # Get the host name
        #
        my $client_name = $client_obj->hostname();
        if ($client_obj->error)
        {
            my $errmsg = join "\n", $client_obj->errmsg();
            $self->_set_err("Can not get hostname: ",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }

        #
        # Get the host type
        #
        my $client_type = $client_obj->hosttype();
        if ($client_obj->error)
        {
            my $errmsg = join "\n", $client_obj->errmsg();
            $self->_set_err("Can not get hosttype: ",
                            "$errmsg");
            return $ClientUtil::FALSE;
        }
        $client_type =~ s/unix-solaris/solaris/;    # strip off unix-

        if (grep /(linux|sanblaze)/, $client_type)
        {
            $client_type = "linux";
        }
        if (grep /hpux/, $client_type)
        {
            $client_type = "hpux";
        }
        if (grep /aix/, $client_type)
        {
            $client_type = "aix";
        }

        $self->sanlog(-msg_level => 3,
                      -msg       => "Creating luns for client ".
                                    $client_name . ":" . $vol_name);

        #
        # Create an igroup object
        #
        my $igroup_name = $client_name . "_" . $protocol;
        if ($self->{setup_igroups})
        {
            $self->sanlog(-msg => "Creating igroup $igroup_name " .
                                  "on filer " . $filer_name);

            my $igroup_obj = $filer_obj->create_igroup(
                             -name=>$igroup_name,
                             -initObjs=>[$client_obj->get_initiators()]);

            if ($filer_obj->error)
            {
                # lets see if it exists
                my $errmsg = join "\n", $filer_obj->errmsg();
                $igroup_obj = $filer_obj->get_igroup(-name=>$igroup_name);
                if ($filer_obj->error())
                {
                    $errmsg .= " and " . join "\n", $filer_obj->errmsg();
                    $self->sanlog(-msg => $errmsg);
                    $self->_set_err("Creating igroup $igroup_name ",
                                    "$errmsg");
                    return $ClientUtil::FALSE;
                }
                $self->sanlog(-msg => "Using existing igroup " .
                                      "$igroup_name on filer " .
                                      $filer_obj->get_name());
                return $ClientUtil::FALSE
                    unless ($self->cleanup_add(-type=>"filer",
                                   -name=>$filer_name,
                                   -cmd=>"igroup_destroy",
                                   -cmd_args=>$igroup_name,
                                   -reuse=>1,
                                   -vfiler_context=>$vfiler_name));
            }
            else
            {
                #
                # Register the cleanup action for igroup create
                #
                return $ClientUtil::FALSE
                    unless ($self->cleanup_add(-type=>"filer",
                                   -name=>$filer_name,
                                   -cmd=>"igroup_destroy",
                                   -cmd_args=>$igroup_name,
                                   -vfiler_context=>$vfiler_name));
            }

            # Add ALUA support if specified
            if (defined($alua_support) and $alua_support == 1)
            {
                if(ref($filer_obj) =~ /Filer10M/)
                {
                    $igroup_obj->set(-attribute  => 'alua',
                                     -value      => 'true',
                                     -force      => 'true',
                                    );

                    if($igroup_obj->error())
                    {
                        my $errmsg = $igroup_obj->errmsg();
                        $self->sanlog(-msg => "setting igroup "          .
                                              "($igroup_name) for alua ($errmsg)");
                        return $ClientUtil::FALSE;
                    }
                }
                else
                {
                    $igroup_obj->set(-attr => 'alua',
                                     -value => 'yes');
                    if($igroup_obj->error())
                    {
                            my $errmsg = $igroup_obj->errmsg();
                            $self->sanlog(-msg => "setting igroup "          .
                                                  "($igroup_name) for alua ($errmsg)");
                            return $ClientUtil::FALSE;
                    }
                }
            }

            my @igroup_detail =  $igroup_obj->display();
            $self->sanlog(-msg => \@igroup_detail, -indent => 4);
        } # end if setup_igroups
        else
        {
            $self->sanlog(-msg => "Skipping igroup creation " .
                                  "(setup_igroups not set)");
        }

        #
        # Check to see if the user wants us to create luns and if they
        # do create and map the lun
        #
        if ($self->{setup_luns})
        {
            #
            # Find the volume object
            #
            $self->sanlog(-msg => "Finding volume object for $vol_name " .
                                  "on filer " . $filer_name,
                          -msg_level => 7);

            my $short_vol = $self->_get_short_vol($vol_name);
            my $filer_vol_obj = $filer_obj->get_vol($short_vol);
            if ($filer_obj->error)
            {
                my $errmsg = join "\n", $filer_obj->errmsg();
                $self->_set_err("Call to get_vol failed: ",
                                "$errmsg");
                return $ClientUtil::FALSE;
            }

            #
            # Create a SanConfig object
            # TBD - is this really necessary?  The cfg_obj doesn't seem to used
            #       SanConfig::new() doesn't really do anything
            #
            $self->sanlog(-msg       => "Creating SanConfig object ",
                          -msg_level => 7);
            my $cfg_obj = new SanConfig(-clientObj=>$client_obj,
                                        -configpaths=>1,
                                        -devmap_refresh=>1);
            if ($cfg_obj->error)
            {
                my $errmsg = join "\n", $cfg_obj->errmsg();
                $self->_set_err("Can not create a SanConfig object: ",
                                "$errmsg");
                return $ClientUtil::FALSE;
            }

            #
            # TBD - should I do the orphaned device cleanup even if
            #       luns are not setup?
            #

            #
            # Register a orphaned_device_cleanup action.
            # Note we only do this once for each client so
            # we push the client name on orphaned_clients
            # and check that list before we register the
            # cleanup action.
            #
            unless (grep /$client_name/, @{$orphaned_clients_ref})
            {
                return $ClientUtil::FALSE
                    unless ($self->cleanup_add(
                                   -type=>"client",
                                   -name=>$client_name,
                                   -cmd=>"orphaned_device_cleanup",
                                   -cmd_args=>$protocol));
                push @{$orphaned_clients_ref}, $client_name;
            }

            #
            # Create and map each lun
            #
            my $current_id = $first_lun_id;

            #
            # for creating shared luns
            #

            my $shared_lun_id = 1;

            if ($current_id == -1)
            {
                $current_id = 1;
            }
            for (my $index = 1 ; $index <= $lun_count ; $index++)
            {
                my $lun_obj;             # single lun object
                my $lun_name;
                if (!$lun_prefix)
                {
                    if ($self->{shared_lun} =~ /yes/i)
                    {
                        $lun_name = "shared_lun_" . $shared_lun_id;
                    }
                    else
                    {
                        $lun_name = $client_name .  "_lun" . $current_id;
                    }
                }
                else
                {
                    #
                    # If there is only one lun asked for and there already
                    # exists a lun named after the prefix, use the prefix
                    # as the name. This allows user to specify preexisting
                    # lun during setup()
                    #
                    if ($lun_count == 1)
                    {
                        $filer_vol_obj->refresh;
                        if ($filer_vol_obj->get_lun($lun_prefix))
                        {
                            $lun_name = $lun_prefix;
                        }
                    }
                    else
                    {
                        $lun_name = $lun_prefix . $current_id;
                    }
                }

                if (length($qtree_prefix))
                {
                    $self->sanlog(-msg => "Create lun $filer_name : "           .
                                          $vol_name .  "/$qtree_prefix/$lun_name $lun_size "  .
                                          "$client_type $igroup_name");
                }
                else
                {
                    $self->sanlog(-msg => "Create lun $filer_name : "           .
                                          $vol_name .  "/$lun_name $lun_size "  .
                                          "$client_type $igroup_name");
                }

                #
                # If there is a qtree_prefix, use it as a qtree name.
                # TBD - eventually we can add a loop to handle multiple qtrees
                #
                if (length($qtree_prefix))
                {
                    #
                    # Create the lun
                    #

                    if(ref($filer_obj) =~ /Filer10M/i)
                    {
                        if($reserve =~ /(0|n|no|f|false)/)
                        {
                            $filer_vol_obj->create_lun(
                                            -path=>$vol_name,
                                            -qtree=>$qtree_prefix,
                                            -size=>$lun_size,
                                            -o => "noreserve",
                                            -type=>$client_type,
                                            -name=>$lun_name);
                        }
                        else
                        {
                            $filer_vol_obj->create_lun(
                                            -path=>$vol_name,
                                            -qtree=>$qtree_prefix,
                                            -size=>$lun_size,
                                            -type=>$client_type,
                                            -name=>$lun_name);
                        }
                    }
                    else
                    {
                        my $path = $vol_name."/".$qtree_prefix;

                        $filer_vol_obj->create_lun(
                                        -path=> $path,
                                        -size=>$lun_size,
                                        -type=>$client_type,
                                        -reserve=>$reserve,
                                        -name=>$lun_name);
                    }
                }
                else
                {
                    if(ref($filer_obj) =~ /Filer10M/i and $reserve =~ /(0|n|no|f|false)/)
                    {
                        $filer_vol_obj->create_lun(
                                        -size=>$lun_size,
                                        -type=>$client_type,
                                        -o => "noreserve",
                                        -name=>$lun_name);
                    }
                    else
                    {
                        #
                        # Create the lun
                        #
                        $filer_vol_obj->create_lun(
                                        -size=>$lun_size,
                                        -type=>$client_type,
                                        -reserve=>$reserve,
                                        -name=>$lun_name);
                    }

                }


                if ($filer_vol_obj->error)
                {
                    # lets see if it exists
                    my $errmsg = join "\n", $filer_vol_obj->errmsg();

                    if(ref($filer_obj) =~ /Filer10M/i)
                    {
                        $lun_obj = $filer_vol_obj->get_lun(-name => $lun_name);
                    }
                    else
                    {
                        $lun_obj = $filer_vol_obj->get_lun($lun_name);
                    }

                    if ($filer_vol_obj->error())
                    {
                        $errmsg .= " and " . join "\n", $filer_vol_obj->errmsg();
                        $self->sanlog(-msg => $errmsg);
                        $self->_set_err("Error: San::setup() ",
                            "Creating lun $lun_name ",
                            "$errmsg");
                        return $ClientUtil::FALSE;
                    }
                    $self->sanlog(-msg => "Using existing lun " .
                                          "$lun_name on filer " .
                                          $filer_obj->get_name());

                    my $cmd_args;

                    if (length($qtree_prefix))
                    {
                        $cmd_args = "$vol_name/$qtree_prefix $lun_name";
                    }
                    else
                    {
                        $cmd_args = "$vol_name $lun_name";
                    }

                    return $ClientUtil::FALSE
                        unless ($self->cleanup_add(-type=>"filer",
                                       -name=>$filer_name,
                                       -cmd=>"lun_destroy",
                                       -cmd_args=>$cmd_args,
                                       -reuse=>1,
                                       -vfiler_context=>$vfiler_name));

                }
                else
                {
                    $lun_obj = $filer_vol_obj->get_lun($lun_name);

                    #
                    # Register the cleanup action for lun create
                    #

                    my $cmd_args;

                    if(length($qtree_prefix))
                    {
                        $cmd_args = "$vol_name/$qtree_prefix $lun_name";
                    }
                    else
                    {
                        $cmd_args = "$vol_name $lun_name";
                    }

                    return $ClientUtil::FALSE
                        unless ($self->cleanup_add(-type=>"filer",
                                       -name=>$filer_name,
                                       -cmd=>"lun_destroy",
                                       -cmd_args=>$cmd_args,
                                       -vfiler_context=>$vfiler_name));
                }

                push @{$self->{config_lun_objs}}, $lun_obj;

                #
                # Map the lun
                #
                if ($self->{setup_igroups})
                {
                    my @lun_objs = $filer_vol_obj->get_luns();
                    if ($filer_vol_obj->error)
                    {
                        my $errmsg = join "\n", $filer_vol_obj->errmsg();
                        $self->_set_err("Call to get_lun failed: ",
                                        "$errmsg");
                        return $ClientUtil::FALSE;
                    }

                    # get_lun() method gets the lun object based on lun name.
                    # If a qtree prefix is present then get all luns
                    # and filter based on lun path. If no qtree is
                    # present then the lun object from get_luns

                    my $lun_path = $vol_name."/".$qtree_prefix."/".$lun_name;

                    foreach my $lun_object (@lun_objs)
                    {
                        if($lun_object->get_path() =~ /$lun_path/)
                        {
                            $lun_obj = $lun_object;
                            last;
                        }
                    }

                    #
                    # Get the name of the igroups to which the lun is mapped
                    # shared luns can be mapped to more than one igroup
                    # check if the lun is mapped to any group at all
                    #

                    my $igroup_mapped = 0;
                    if ($self->{shared_lun} =~ /yes/i)
                    {
                        my @igroups = $lun_obj->get_maps();

                        if ($lun_obj->error)
                        {
                            $self->_set_err("Get Maps failed: ",
                                            "$lun_obj->errmsg()");
                            return $ClientUtil::FALSE;
                        }

                        #
                        # Check if the lun is mapped to an igroup
                        #

                        $self->sanlog(-msg       => "Igroups mapped @igroups",
                                      -msg_level => 3);

                        foreach my $igrp (@igroups)
                        {
                            if ($igrp eq $igroup_name)
                            {
                                $self->sanlog(-msg => "Lun already mapped",
                                              -msg_level => 3);
                                $igroup_mapped = 1;
                                $self->sanlog(-msg =>
                                     "Using existing lun map for lun $lun_name "
                                     . "mapped to igroup $igroup_name "
                                     . "on filer "
                                     . $filer_obj->get_name());
                                return $ClientUtil::FALSE
                                  unless (
                                       $self->cleanup_add(
                                           -type => "filer",
                                           -name => $filer_name,
                                           -cmd  => "lun_unmap",
                                           -cmd_args =>
                                            "$vol_name $lun_name $igroup_name $vol_aggr",
                                           -reuse => 1,
                                           -vfiler_context=>$vfiler_name));
                            }
                        }#end foreach igrp

                        if ($igroup_mapped == 0)
                        {
                            #
                            # map the lun to the igroup if it has not been done already
                            #

                            $lun_obj->map($igroup_name);
                            if ($lun_obj->error)
                            {
                                $self->_set_err("Can not map lun: ",
                                                "$lun_obj->errmsg()");
                                return $ClientUtil::FALSE;
                            }
                            else
                            {

                                #
                                # Register the cleanup action for lun map
                                #
                                return $ClientUtil::FALSE
                                  unless (
                                        $self->cleanup_add(
                                            -type => "filer",
                                            -name => $filer_name,
                                            -cmd  => "lun_unmap",
                                            -cmd_args =>
                                            "$vol_name $lun_name $igroup_name $vol_aggr",
                                            -vfiler_context=>$vfiler_name));
                            }
                        }# end if igroup_mapped
                    }# end if shared_lun
		    else
		    {
	               $lun_obj->map($igroup_name);
	               if ($lun_obj->error)
	               {
                          #
                          # lets see if it's already mapped
                          #

        	          my $current_igroup = $lun_obj->get_igroup();
	                  if ($current_igroup eq $igroup_name)
                	  {
        	              $self->sanlog( -msg =>
	                      		     "Using existing lun map for lun $lun_name " .
	                                     "mapped to igroup $igroup_name "            .
	                                     "on filer " . $filer_obj->get_name());
                              return $ClientUtil::FALSE
                	          unless ($self->cleanup_add(
        	                              -type=>"filer",
	                                      -name=>$filer_name,
                                	      -cmd=>"lun_unmap",
                        	              -cmd_args=>"$vol_name $lun_name $igroup_name $vol_aggr",
                	                      -reuse=>1,
                                              -vfiler_context=>$vfiler_name));
        	           }
	                   else
                           {
        	                    $self->_set_err("Can not map lun: ",
                                	            "$lun_obj->errmsg()");
                        	    return $ClientUtil::FALSE;
                	   }
        	        }
	                else
                	{
        	            #
	                    # Register the cleanup action for lun map
                            #
                     return $ClientUtil::FALSE
        	                unless ($self->cleanup_add(
	                                    -type=>"filer",
                                            -name=>$filer_name,
                                            -cmd=>"lun_unmap",
                                            -cmd_args=>"$vol_name $lun_name $igroup_name $vol_aggr",
                                            -vfiler_context=>$vfiler_name));
                    	}
                    }
                    # Make sure LUN is online
                    my $lun_is_offline = $lun_obj->is_offline;
                    if ($lun_obj->error){
                        my $errmsg = join "\n", $lun_obj->errmsg();
                        $self->_set_err("Can not see if lun is offline: ",
                                        "$errmsg");
                        return $ClientUtil::FALSE;
                    }
                    if ($lun_is_offline){
                        $lun_obj->online;
                        if ($lun_obj->error){
                            my $errmsg = join "\n", $lun_obj->errmsg();
                            $self->_set_err("Unable to online lun: ",
                                            "$errmsg");
                            return $ClientUtil::FALSE;
                        }
                    }
                } # end if setup_igroups
                else
                {
                    $self->sanlog(-msg => "Skipping lun map creation " .
                                          "(setup_igroups not set)");
                }

                #
                # Increment the id count for the next lun.
                # Increment the shared_lun_id
                #
                $current_id++;
                $shared_lun_id++;
            } # end for each lun
        } # end if setup_luns
        else
        {
            $self->sanlog(-msg =>
                "Skipping lun creation (setup_luns not set). Note that the "
               ."orphaned device cleanup step is not being scheduled for this "
               ."client.  If Luns are later created a mapped to this client, "
               ."the orphaned_device_cleanup step should be added to the "
               ."cleanup array before any lun cleanup steps are added");

        }
    } # end for each client

    return 1;
} # end of _lun_setup()

###############################################################################
# Method:    _vfiler_setup
# Objective: Create the necessary vfilers.
# Inputs:    viler_obj - filer object
#            protocol - fcp or iscsi
#            orphaned_clients_ref -
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
###############################################################################
sub _vfiler_setup
{
    my ($self, %args) = @_;

    my ($filer_obj, $protocol, $orphaned_clients_ref);

    #
    # Parse arguments
    #
    foreach (keys %args)
    {
        if (/-?filer_obj/)
        {
            $filer_obj = $args{$_};
            next;
        }
        elsif (/-?protocol$/)
        {
            $protocol = $args{$_};
            if (not length($protocol))
            {
                $self->_set_err("Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        }
        elsif (/-?orph_clients$/)
        {
            $orphaned_clients_ref = $args{$_};
            if (not length($orphaned_clients_ref))
            {
                $self->_set_err("Invalid value for argument '-orph_clients'");
                return $self;
            }
            next;
        }
    } # end foreach arg

    #
    #If the filer object is a ScsitSimulator, then
    #All of this is a no-op, so we'll just return
    #success and skip it
    #
    if(ref($filer_obj) =~ /ScsitSimulator/){
        return 1;
    }

    my $filer_name = $filer_obj->get_name();

    #
    # For each vol prefix we want to store the index where we
    # are to avoid using vols of the same name.
    #
    my $vol_prefix_hash    = {};

    #
    # A vfiler is created for each client
    #
    foreach my $client_obj (@{$self->{clients}})
    {

        my $ip_ref_array;
        my $client = $client_obj->hostname();
        $self->sanlog(-msg => "In _vfiler_setup - working on client $client");

        #
        # Determine whether the filer we are configuring is
        # filerA or filerB.  This is determined by it's index in the
        # @filers array. Get the pc vfiler info for that filer.
        # If there is only one filer in the array, this will always be
        # filerA
        #
        my $temp_count = 0;
        foreach my $temp_filer_obj (@{$self->{filers}}) {
            if ($temp_filer_obj->get_name() eq $filer_name) {
                #
                # This filer in $self->{filers} is the right one
                # figure out if it's A or B (index 0 or 1)
                #
                if ($temp_count == 0) {        # FilerA
                    $ip_ref_array =
                        $self->_get_pc_vfiler_info($client, "FILERA");
                } elsif ($temp_count == 1) {   # FilerB
                    $ip_ref_array =
                        $self->_get_pc_vfiler_info($client, "FILERB");
                } else {                       # Some other filer
                    $self->_set_err("Unable to get per client vfiler info for" .
                                    " filer $filer_name");
                     return $ClientUtil::FALSE;
                } # end else temp_count
            } else {
                $temp_count++;
            } # end else get_name()
        } # end foreach temp_filer_obj

        unless($ip_ref_array)
        {
            $self->_set_err("_get_pc_vfiler_info " .
                            "returned an error " . $self->errmsg());
             return $ClientUtil::FALSE;
        }

        #
        # list of Volumes that this vfiler will use
        #
        my $vol_list;

        $self->sanlog(-msg => "In San::_vfiler_setup with "             .
                              "$filer_name - setup vfiler for client  " .
                              $client_obj->hostname());

        my $vfiler_name = $client_obj->hostname() . "_vfiler";

        #
        # Filer_vol_list is created on each vfiler
        #
        foreach my $vol_entry_hash (@{$self->{filer_vol_list}})
        {
            my $vol_prefix   = $vol_entry_hash->{vol_prefix};
            my $vol_count    = $vol_entry_hash->{vol_count};
            $self->sanlog(-msg => "In _vfiler_setup - working on vol_setup for $client" .
                                  "vol_prefix = $vol_prefix, vol_count = $vol_count");

            #
            # If vol_prefix is an existing volume and vol_count is set
            # to 1, don't use prefixes. Assume user wants to use that
            # exact volume
            #
            if ( $vol_count == 1 && $filer_obj->vol_exist($vol_prefix) )
            {
                return $ClientUtil::FALSE
                    unless ($self->_vol_setup(-filer_obj      => $filer_obj,
                                              -vol_name       => "/vol/${vol_prefix}",
                                              -vol_entry_hash => $vol_entry_hash));
                next;
            }

            for (my $index = 1 ; $index <= $vol_count  ; $index++)
            {
                $vol_prefix_hash->{$vol_prefix}++;
                my $vol_name = "/vol/$vol_prefix" .
                               $vol_prefix_hash->{$vol_prefix};

                $vol_list = $vol_name . " " . $vol_list;

                return $ClientUtil::FALSE
                    unless ($self->_vol_setup(-filer_obj      => $filer_obj,
                                              -vol_name       => $vol_name,
                                              -vol_entry_hash => $vol_entry_hash));

            } # end for each volume

        } # end for each filer vol list entry

        #
        # Just pick out the first ip address to use in creating the
        # vfiler, we'll go in and call setup later to configure all the
        # interfaces specified in the tharnhost file
        #
        my $vfiler_ip_address = @{$ip_ref_array}[0]->{ipaddress};

        #
        # Create the vfiler using the Volumes specified in filer_vol_list
        #
        my $vfiler_obj = $filer_obj->vfiler_create($vfiler_name,
                                                   $vfiler_ip_address,
                                                   $vol_list);

        if ($filer_obj->error){
            $self->_set_err("Unable to create vfiler" .
                            $filer_obj->errmsg());
            return $ClientUtil::FALSE;
        }
        return $ClientUtil::FALSE
            unless ($self->cleanup_add(-type=>"filer",
                                       -name=>$filer_name,
                                       -cmd=>"vfiler_destroy",
                                       -cmd_args=>"$vfiler_name",
                                       -reuse=>0));

        #
        # Setup the vfiler using the ip addresses specified in tharnhost
        #
        $filer_obj->vfiler_setup($vfiler_name,
                                 $ip_ref_array);
        if ($filer_obj->error){
            $self->_set_err("Unable to setup vfiler" .
                            $filer_obj->errmsg());
            return $ClientUtil::FALSE;
        }


        #
        # Then create the luns, igroups, and maps in the vfiler context
        #
        foreach my $vol_entry_hash (@{$self->{filer_vol_list}})
        {
            my $vol_prefix   = $vol_entry_hash->{vol_prefix};
            my $vol_count    = $vol_entry_hash->{vol_count};

            $self->sanlog(-msg => "In _vfiler_setup - working on lun_setup for $client" .
                                  "vol_prefix = $vol_prefix, vol_count = $vol_count");

            #
            # Need to pass in a client object array to _lun_setup
            #
            my @client_obj_array = ();
            push (@client_obj_array, $client_obj);

            #
            # Init the prefix count for this vol_prefix back to starting value
            # for this vfiler (this was incremented by $vol_count when we
            # created the Volumes, need to subtract that amount)
            #
            $vol_prefix_hash->{$vol_prefix} =
                $vol_prefix_hash->{$vol_prefix} - $vol_count;

            #
            # If vol_prefix is an existing volume and vol_count is set
            # to 1, don't use prefixes. Assume user wants to use that
            # exact volume
            #
            if ( $vol_count == 1 && $filer_obj->vol_exist($vol_prefix) )
            {
                #
                # Set the vfiler context to the newly created vfiler
                #
                $filer_obj->set_vfiler_context(-name => $vfiler_name);
                if ($filer_obj->error){
                    $self->_set_err("Unable to set vfiler context" .
                                    $filer_obj->errmsg());
                    return $ClientUtil::FALSE;
                }

                unless ($self->_lun_setup(-filer_obj      => $filer_obj,
                                          -vol_prefix     => "/vol/${vol_prefix}",
                                          -protocol       => $protocol,
                                          -orph_clients   => $orphaned_clients_ref,
                                          -vol_entry_hash => $vol_entry_hash,
                                          -clients        => \@client_obj_array)) {
                    #
                    # Set the vfiler context back to vfiler0
                    #
                    $filer_obj->clear_vfiler_context();
                    if ($filer_obj->error){
                        $self->_set_err("Unable to set vfiler context" .
                                        $filer_obj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    return $ClientUtil::FALSE;
                } # end unless _lun_setup

                #
                # Set the vfiler context back to vfiler0
                #
                $filer_obj->clear_vfiler_context();
                if ($filer_obj->error){
                    $self->_set_err("Unable to set vfiler context" .
                                    $filer_obj->errmsg());
                    return $ClientUtil::FALSE;
                }

                next;
            }

            for (my $index = 1 ; $index <= $vol_count  ; $index++)
            {
                $vol_prefix_hash->{$vol_prefix}++;
                my $vol_name = "/vol/$vol_prefix" .
                               $vol_prefix_hash->{$vol_prefix};

                #
                # Set the vfiler context to the newly created vfiler
                #
                $filer_obj->set_vfiler_context(-name => $vfiler_name);
                if ($filer_obj->error){
                    $self->_set_err("Unable to set vfiler context" .
                                    $filer_obj->errmsg());
                    return $ClientUtil::FALSE;
                }

                unless ($self->_lun_setup(-filer_obj      => $filer_obj,
                                          -vol_prefix     => $vol_name,
                                          -protocol       => $protocol,
                                          -orph_clients   => $orphaned_clients_ref,
                                          -vol_entry_hash => $vol_entry_hash,
                                          -clients        => \@client_obj_array)) {
                    #
                    # Set the vfiler context back to vfiler0
                    #
                    $filer_obj->clear_vfiler_context();
                    if ($filer_obj->error){
                        $self->_set_err("Unable to set vfiler context" .
                                        $filer_obj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    return $ClientUtil::FALSE;
                } # end unless _lun_setup

                #
                # Set the vfiler context back to vfiler0
                #
                $filer_obj->clear_vfiler_context();
                if ($filer_obj->error){
                    $self->_set_err("Unable to clear vfiler context" .
                                    $filer_obj->errmsg());
                    return $ClientUtil::FALSE;
                }

            } # end for each volume
        } # end for each filer vol list entry
    } # end for each client

    return 1;
} # end of _vfiler_setup()

########################################################################
# Method:    get_lun_objs
# Objective: Return an array of lun objects, or a single lun object.
# Details:   See POD documentation at the end of this file.
########################################################################
sub get_lun_objs()
{
    my($self, %args) = @_;
    my $lun_path = $args{-lun_path} if defined($args{-lun_path});

    $self->_clear_err();

    if (defined($lun_path))
    {
        foreach my $lun_obj (@{$self->{config_lun_objs}})
        {
            $self->sanlog("$lun_obj->get_path() =?");
            if($lun_obj->get_path() eq $lun_path)
            {
                return $lun_obj;
            }
        }
        $self->_set_err("Lun $lun_path not found");
        return undef;
    }

    return @{$self->{config_lun_objs}};
} # get_lun_objs

###############################################################################
# Method:    _lvm_setup
# Objective: Create or find the necessary volume.
# Inputs:    arg hash with the following fields:
#              -client
#              -dg_prefix
#              -lvm
#              -vol_layouts
#              -stripe_size
#              -vol_cnts
#              -vol_sizes
#              -disk_fmts
# Outputs:   1 - success  0 - failure
#            On failure the error flag and message will be set
###############################################################################
sub _lvm_setup
{
    my ($self, %args) = @_;
    my $client_obj;
    my %lvm_setup_args;
    my $dg_prefix  = "san_dg_";
    my $dg_devices = 0; # lun count
    my $vol_mgr    = "nativelvm";
    my $vol_layouts;
    my $stripe_size;
    my $vol_cnts;
    my $vol_sizes;
    my $disk_fmts;
    my $filers = "";
    my $filer_vol = "";
    my @filer_vol_objs;

    # args: -client -dg_prefix, -lvm, -vol_layouts, -stripe_size, -vol_cnts, -vol_sizes, -disk_fmts

    $lvm_setup_args{-dg_prefix}=$dg_prefix;

    foreach (keys %args)
    {
        if (/-?client$/) {
            length($client_obj = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"client\":",
                                "$args{$_}");
                return $self;
            };
        }
        elsif (/-?dg_prefix$/) {
            length($dg_prefix = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"dg_prefix\":",
                                "$args{$_}");
                return $self;
            };
            $lvm_setup_args{-dg_prefix}=$dg_prefix;
        }
        elsif (/-?lvm$/) {
            length($vol_mgr = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"lvm\":",
                                "$args{$_}");
                return $self;
            };
        }
        elsif (/-?vol_layouts$/) {
            length($vol_layouts = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"vol_layouts\":",
                                "$args{$_}");
                return $self;
            };
            $lvm_setup_args{-vol_layouts}=$vol_layouts;
        }
        elsif (/-?stripe_size$/) {
            length($stripe_size = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"stripe_size\":",
                                "$args{$_}");
                return $self;
            };
            #
            # Currently unsupported (TBD)
            # $lvm_setup_args{-stripe_size}=$stripe_size;
            #
        }
        elsif (/-?vol_cnts$/) {
            length($vol_cnts = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"vol_cnts\":",
                                "$args{$_}");
                return $self;
            };
            $lvm_setup_args{-vol_cnts}=$vol_cnts;
        }
        elsif (/-?vol_sizes$/) {
            length($vol_sizes = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"vol_sizes\":",
                                "$args{$_}");
                return $self;
            };
            $lvm_setup_args{-vol_sizes}=$vol_sizes;
        }
        elsif (/-?disk_fmts$/) {
            #
            # If the disk_fmts is blank, don't worry about it.
            #
            if (length($disk_fmts = $args{$_}))
            {
                $lvm_setup_args{-disk_fmts}=$disk_fmts;
            };
        }
        elsif (/-?filers$/) {
            length($filers = $args{$_}) or
            do {
                $self->_set_err("San::_lvm_setup invalid value for \"filers\":",
                                "$args{$_}");
                return $self;
            };
        }
        else {
            $self->_set_err("San::_lvm_setup Invalid parameter specified: $_");
            return $self;
        }
    } # end foreach arg

    #
    # Check for required parameters
    #

    unless (defined $client_obj and ref($client_obj) =~ "SanClient" )
    {
        $self->_set_err("Error: San::_lvm_setup()",
                        "Missing or invalid argument '-client'");
        return $self;
    }

    unless (length($filers))
    {
        $self->_set_err("San::_lvm_setup invalid value for " .
                        "required argument \"-filers\"");
        return $self;
    }

    #
    # Get the client name to determine the lun prefix so we know
    # which luns to go find.
    #
    my $client_name = $client_obj->hostname();
    if ($client_obj->error)
    {
        my $errmsg = join "\n", $client_obj->errmsg();
        $self->_set_err("Error: San::_lvm_setup() ",
                        "Can not get hostname: ",
                        "$errmsg");
        return $self;
    }

    my $lun_prefix = $client_name . "_lun";

    #
    # I can send in the list of filers and then use
    # the get_filer method
    #
    my @filerArray  = split(/:/, $filers);
    foreach my $filer (@filerArray)
    {
        #
        # Find the filer object
        #
        my $filer_obj = $self->get_filer($filer);
        if ($self->error)
        {
            my $errmsg = join "\n", $self->errmsg();
            $self->_set_err("Error: San::_lvm_setup() finding filer ",
                            "object $filer",
                            " $errmsg");
            return $self;
        }

        #
        # For each entry in the filer_vol_list, I need to find the
        # vol object for the volume.
        #

        #
        # For each vol prefix we want to store the index where we
        # are to avoid using vols of the same name.
        #
        my $vol_prefix_hash    = {};

        foreach my $vol_entry_hash (@{$self->{filer_vol_list}})
        {
            my $vol_count    = $vol_entry_hash->{vol_count};
            my $vol_prefix   = $vol_entry_hash->{vol_prefix};
            my $lun_count    = $vol_entry_hash->{lun_count};

            if ( $vol_count == 1 && $filer_obj->vol_exist($vol_prefix) )
            {
                $dg_devices += $lun_count;
                my $fvol_obj = $filer_obj->get_vol($vol_prefix);
                if ($filer_obj->error)
                {
                    my $errmsg = join "\n", $filer_obj->errmsg();
                    $self->_set_err("Error: San::_lvm_setup() ",
                                    "Call to get_vol failed: ",
                                    "$errmsg");
                    return $self;
                }

                #
                # Push the filer volume object on the stack to send to
                # SanConfig::lvm_setup()
                #
                $self->sanlog(-msg => "Pushing a filer ($filer) vol " .
                                      $vol_prefix                       .
                                      " for LVM usage");

                push @filer_vol_objs, $fvol_obj;

                next;
            }

            #
            # Init the prefix count for this vol_prefix if necessary
            #
            $vol_prefix_hash->{$vol_prefix} = 0
                unless defined $vol_prefix_hash->{$vol_prefix};

            for (my $index = 1 ; $index <= $vol_count  ; $index++)
            {
                $dg_devices += $lun_count;
                $vol_prefix_hash->{$vol_prefix}++;
                my $vol_name = $vol_prefix .
                               $vol_prefix_hash->{$vol_prefix};

                #
                # Find the volume object
                #
                my $fvol_obj = $filer_obj->get_vol($vol_name);
                if ($filer_obj->error)
                {
                    my $errmsg = join "\n", $filer_obj->errmsg();
                    $self->_set_err("Error: San::_lvm_setup() ",
                                    "Call to get_vol failed: ",
                                    "$errmsg");
                    return $self;
                }

                #
                # Push the filer volume object on the stack to send to
                # SanConfig::lvm_setup()
                #
                $self->sanlog(-msg => "Pushing a filer ($filer) vol " .
                                      $vol_name                       .
                                      " for LVM usage");

                push @filer_vol_objs, $fvol_obj;
            } # end for each volume
        } # end for each filer vol list entry

    } # end foreach filer

    #
    # Create the LVM object with the appropriate volume manager
    #
    if ($vol_mgr eq 'nativelvm')
    {
        $vol_mgr = $client_obj->get_lvm_default();
        $self->sanlog(-msg => "Using Native Volume manager $vol_mgr",
                      -msg_level => 3);
    }

    $self->sanlog(-msg       => "Create a LVM object ($vol_mgr)",
                  -msg_level => 7);
    my $lvm_obj = $client_obj->create_lvm(-lvm     => $vol_mgr,
                                          -loadcfg => 1);
    if ($client_obj->error)
    {
        my $errmsg = join "\n", $client_obj->errmsg();
        $self->_set_err("Error: San::_lvm_setup() ",
                        "Can not create lvm object: ",
                        "$errmsg");
        return $self;
    }

    #
    # Create the SanConfig object
    #
    $self->sanlog(-msg => "Create a SanConfig object");
    my $config_obj = new SanConfig(-clientObj      => $client_obj,
                                   -lvmObj         => $lvm_obj,
                                   -config         => 1,
                                   -devmap_refresh => 1
                                  );
    if ($config_obj->error)
    {
        my $errmsg = join "\n", $config_obj->errmsg();
        $self->_set_err("Error: San::_lvm_setup() ",
                        "Can not create SanConfig object: ",
                        "$errmsg");
        return $self;
    }

    #
    # Setup the LVM on the host
    #
    $lvm_setup_args{-lun_prefix} = $client_name . "_lun";
    $lvm_setup_args{-dg_unique}  = 0;
    $lvm_setup_args{-dg_ids}     = 0;
    $lvm_setup_args{-fs_types}   = "raw";
    $lvm_setup_args{-fvol_objs}  = [$filer_vol_objs[0]];
    $lvm_setup_args{-fvol_objs}  = [@filer_vol_objs];
    $lvm_setup_args{-dg_devices} = $dg_devices;

    $config_obj->lvm_setup(%lvm_setup_args);
    if ($config_obj->error)
    {
        my $errmsg = join "\n", $config_obj->errmsg();
        $self->_set_err("Error: San::_lvm_setup() ",
                        "Call to lvm_setup failed: ",
                        "$errmsg");
        return $self;
    }

    #
    # Register the cleanup action for lvm setup
    #
    return $self
        unless ($self->cleanup_add(-type=>"client",
                                   -name=>$client_name,
                                   -cmd=>"lvm_cleanup",
                                   -cmd_args=>"$vol_mgr $dg_prefix"));
} # end of _lvm_setup method

###############################################################################
# Method:    _format_device_string
# Objective: Format a devices so user can run I/O immediately.
# Inputs:    device       - the device in question
#            device_type  - raw|vm|fs
#            host_type    - type of client (solaris|hpux|aix|...)
#
# Outputs:   updated array of devices
###############################################################################
sub _format_device_string
{
    my ($self, $device, $device_type, $host_type) = @_;

    if ( ($host_type =~ /solaris/) and ($device_type =~ /raw/))
    {
        #
        # If we have a raw device, that's been formatted, lets
        # make sure that it writes to slice 6 instead of slice 2.
        #
        $device =~ s/s2/s6/g;
        unless ($device =~ /^\/dev\/rdsk\//)
        {
            $device = "/dev/rdsk/$device";
        }
    }
    elsif ( ($host_type =~ /linux/) and ($device_type =~ /raw/))
    {
        #
        # Add the "/dev" if it's not there.
        #
        unless ($device =~ /^\/dev\//)
        {
            $device = "/dev/$device";
        }
    }
    elsif ( ($host_type =~ /hpux/) and ($device_type =~ /raw/))
    {
        #
        # Add the "/dev/rdsk/" if it's not there.
        #
        unless ($device =~ /^\/dev\//)
        {
            $device = "/dev/rdsk/$device";
        }
    }
    elsif ( ($host_type =~ /aix/) and ($device_type =~ /raw/))
    {
        #
        # Add the "/dev/rdsk/" if it's not there.
        #
        unless ($device =~ /^\/dev\//)
        {
            $device = "/dev/r" . $device;
        }
    }

    return $device;

} # end of _format_device_string method

###############################################################################
# Method:    _get_short_vol
# Objective: Given a volume name like /vol/xyz/ return the xyz
# Inputs:    volume       - the volume path
# Outputs:   volume name only
###############################################################################
sub _get_short_vol
{
    my ($self, $volume) = @_;

    my $short_vol = $volume;
    $short_vol    =~ s/^\/vol\///; # remove /vol/
    $short_vol    =~ s/\/$//;      # remove trailing "/"

    return $short_vol;
} # end of _get_short_vol


###############################################################################
# Method:    _get_pc_mp_type
# Objective: if PerClient (PC) mp_type is set in the tharnhost file then
#            set the param in the env variables
# Inputs:    mp_type       - Multipathing type passed into setup()
#            client        - name of the client
# Outputs:   if PC_CLIENT_VMS param is found then returns a hash ref with
#            new_mp_type and old_client_params as the keys
#            else returns 0
###############################################################################
sub _get_pc_mp_type
{
    my ($self, $mp_type, $client) = @_;

    my %ret_ref;

    $self->sanlog(-msg => "San::_get_pc_mp_type(): Loading Per Client ".
                          "Params for $client",
                  -msg_level    => 1);
    #
    # set ENABLE_PER_CLIENT_PARAMS, which will ensure that load_client_params
    # will succeed to read in the client tharnhost file and load any values
    # for parameters being overridden.
    #
    &Tharn::param("ENABLE_PER_CLIENT_PARAMS", -set, 1);

    $ret_ref{old_client_params} = &ClientUtil::load_client_params($client);

    #
    # check if CLIENT_VMS is defined
    #
    &Tharn::param("CLIENT_VMS", "-default", 0);

    my $pc_vms = $Tharn::Params{"CLIENT_VMS"}->{value};

    unless($pc_vms)
    {
        $self->_set_err("San::_get_pc_mp_type(): PC_CLIENT_VMS param not ".
                        "set in tharnhost file for $client");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "San::_get_pc_mp_type(): CLIENT_VMS set to: ".
                          "$pc_vms for $client",
                  -msg_level    => 1);
    #
    # CLIENT_VMS has this format "<mp type>-<volume manager>"
    # we are interested in getting to only the "mp type" value for a
    # particular host, because that value determines the type of MP object
    # that SanClient creates.
    #
    if($pc_vms =~ /^(\S+)\-.+$/)
    {

        $ret_ref{new_mp_type} = $1;

        $self->sanlog(-msg => "San::_get_pc_mp_type(): new_mp_type set to: ".
                              $ret_ref{new_mp_type}." for $client",
                      -msg_level    => 1);

        return \%ret_ref;
    }

    return $ClientUtil::FALSE;

}

###############################################################################
# Method:    _get_pc_vfiler_info
# Objective: if vfiler_enable is set to one in the input parameters
#            get the PC_Vfiler_Head_IPList parameters
# Inputs:    client        - name of the client
# Outputs:   if successful, returns reference to IP config for vfiler
#            else returns 0 and errmsg is set
###############################################################################
sub _get_pc_vfiler_info
{
    my ($self, $client, $filer) = @_;

    my @ip_ref_array = ();
    my $per_client_var_name;

    $self->sanlog(-msg => "San::_get_pc_vfiler_info(): Loading Per Client ".
                          "Vfiler Params for $client",
                  -msg_level    => 1);

    &Tharn::param("ENABLE_PER_CLIENT_PARAMS", -set, 1);

    if ($filer eq "FILERA") {
        $per_client_var_name = "VFILER_HEADA_IPLIST";
    } elsif ($filer eq "FILERB") {
        $per_client_var_name = "VFILER_HEADB_IPLIST";
    } else {
        $self->_set_err("Unable to get per client info for filer $filer");
        return $ClientUtil::FALSE;
    }

    &Tharn::param($per_client_var_name, "-default", "0");

    &ClientUtil::load_client_params($client);

    # store the values into the datastructure
    my $pc_vfiler_head_ip =
                $Tharn::Params{$per_client_var_name}->{value};

    unless($pc_vfiler_head_ip)
    {
        $self->_set_err("$per_client_var_name " .
                        "param not set in tharnhost file for $client");
        return $ClientUtil::FALSE;
    }


    my @ip_config_array = split(/:/, $pc_vfiler_head_ip);

    # The per client vfiler ip list should look like this
    # PC_VFILER_HEADA_IPLIST=e0b(10.225.195.9 255.255.254.0)
    # PC_VFILER_HEADB_IPLIST=e0b(10.225.195.6 255.255.254.0)
    foreach my $ip_config (@ip_config_array) {
        my $ip_ref = {};
        $ip_config =~ /(\S+)[(](\S+) (\S+)[)]/;
        $ip_ref->{interface} = $1;
        $ip_ref->{ipaddress} = $2;
        $ip_ref->{netmask}   = $3;
        push (@ip_ref_array, $ip_ref);
    }


    $self->sanlog(-msg => "San::_get_pc_vfiler_type(): $per_client_var_name set to: ".
                          "$pc_vfiler_head_ip for $client",
                  -msg_level    => 1);

    return (\@ip_ref_array);

}

###############################################################################
# Method:    _get_pc_any_iscsi
# Objective: if client hosttype is unix|linux and any_iscsi is set in the tharnhost
#            file then set the param in the env variables
# Inputs:    client        - name of the client
# Outputs:   if PC_ANY_ISCSI param is found then returns a hash ref with
#            new_pc_any_iscsi_type and old_client_params as the keys
#            else returns 0
###############################################################################
sub _get_pc_any_iscsi
{
    my ($self, $client) = @_;

    my %ret_ref;

    $self->sanlog(-msg => "San::_get_pc_any_iscsi(): Loading Per Client ".
                          "Params for $client",
                  -msg_level    => 1);
    #
    # set ENABLE_PER_CLIENT_PARAMS, which will ensure that load_client_params
    # will succeed to read in the client tharnhost file and load any values
    # for parameters being overridden.
    #
    &Tharn::param("ENABLE_PER_CLIENT_PARAMS", -set, 1);

    $ret_ref{old_client_params} = &ClientUtil::load_client_params($client);

    #
    # check if ANY_ISCSI is defined
    #
    &Tharn::param("ANY_ISCSI", "-default", 0);

    my $any_iscsi = $Tharn::Params{"ANY_ISCSI"}->{value};

    unless($any_iscsi)
    {
        $self->sanlog(-msg => "PC_ANY_ISCSI param not ".
                        "set in tharnhost file for $client"
                      -msg_level => 3);
    }

    $self->sanlog(-msg => "San::_get_pc_any_iscsi(): ANY_ISCSI  set to: ".
                          "$any_iscsi for $client",
                  -msg_level    => 1);

    $ret_ref{new_any_iscsi} = $any_iscsi;

    $self->sanlog(-msg => "San::_get_pc_any_iscsi(): new_any_iscsi set to: ".
                          $ret_ref{new_any_iscsi}." for $client",
                  -msg_level    => 1);

        return \%ret_ref;
}

########################################################################
# Method:    _get_device_size
# Objective: Get the size of a device
# Details:   See POD documentation at the end of this file.
########################################################################
sub _get_device_size()
{
    my ($self, %args) = @_;

    my ($dev_obj, $client_type, $size);

    foreach (keys %args)
    {
        if (/-?dev_obj$/)
        {
            $dev_obj = $args{$_};
            if (not length($dev_obj))
            {
                $self->_set_err("Invalid value for argument '-dev_obj'");
                return "";
            }
            next;
        }
        elsif (/-?client_type$/)
        {
            $client_type = $args{$_};
            if (not length($client_type))
            {
                $self->_set_err("Invalid value for argument '-client_type'");
                return "";
            }
            next;
        }
        else
        {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return "";
        }
    }

    #
    # Check mandatory parameter.
    #
    unless (defined $dev_obj and length $dev_obj) {
        $self->_set_err("mandatory parameter \"-dev_obj\" not specified ");
        return "";
    }
    unless (defined $client_type and length $client_type) {
        $self->_set_err("mandatory parameter \"-client_type\" not specified ");
        return "";
    }

    #
    # get actual device size based on client type
    #
    if ($client_type =~ /solaris/i)
    {
        #
        # get slice 6 size because of the prefix
        #
        $size = $dev_obj->get_partition_size(6);
    }
    elsif ($client_type =~ /linux/i)
    {
        #
        # if it's partitioned, we need to get partition 1
        #
        if ($dev_obj->is_partitioned())
        {
            $size = $dev_obj->get_partition_size(1);
        }
        else
        {
            $size = $dev_obj->get_size();
        }
    }
    else
    {
        #
        # other OSes, we just get partition size from device object
        #
        $size = $dev_obj->get_partition_size();
    }

    #
    # Return the size if no error
    #
    if ($dev_obj->error())
    {
        $self->_set_err("Unable to get partition size ",
                        $dev_obj->errmsg());
                        return "";
    }
    else
    {
        return $size;
    }

} # end _get_device_size()

########################################################################
# Method:    _get_device_info
# Objective: Get the device info
# Details:   See POD documentation at the end of this file.
########################################################################
sub _get_device_info()
{
    my ($self, %args) = @_;

    my ($dev_mgr_obj, $device, $client_type, %device_info);
    my $device_type = "raw";

    foreach (keys %args)
    {
        if (/-?dev_mgr_obj$/)
        {
            $dev_mgr_obj = $args{$_};
            if (not length($dev_mgr_obj))
            {
                $self->_set_err("Invalid value for argument '-dev_mgr_obj'");
                return (wantarray()) ? () : {};
            }
            next;
        }
        elsif (/-?device$/)
        {
            $device = $args{$_};
            if (not length($device))
            {
                $self->_set_err("Invalid value for argument '-device'");
                return (wantarray()) ? () : {};
            }
            next;
        }
        elsif (/-?device_type$/)
        {
            $device_type = $args{$_};
            if (not length($device_type))
            {
                $self->_set_err("Invalid value for argument '-device_type'");
                return (wantarray()) ? () : {};
            }
            next;
        }
        elsif (/-?client_type$/)
        {
            $client_type = $args{$_};
            if (not length($client_type))
            {
                $self->_set_err("Invalid value for argument '-client_type'");
                return (wantarray()) ? () : {};
            }
            next;
        }
        else
        {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : {};
        }
    }

    #
    # Check mandatory parameter.
    #
    unless (defined $dev_mgr_obj and length $dev_mgr_obj) {
        $self->_set_err("mandatory parameter \"-dev_mgr_obj\" not specified ");
        return (wantarray()) ? () : {};
    }
    unless (defined $device and length $device) {
        $self->_set_err("mandatory parameter \"-device\" not specified ");
        return (wantarray()) ? () : {};
    }
    unless (defined $client_type and length $client_type) {
        $self->_set_err("mandatory parameter \"-client_type\" not specified ");
        return (wantarray()) ? () : {};
    }

    #
    # Get the formatted device
    #
    $device_info{dev_name} = $self->_format_device_string($device,
                                                          $device_type,
                                                          $client_type);

    #
    # Get the basename of device
    #
    if ($client_type =~ /linux/   or
        $client_type =~ /hpux/    or
        $client_type =~ /solaris/ or
        $client_type =~ /aix/ )
    {
        #
        # remove the "/dev" if it's there.
        #
        if ($device =~ /^\/dev\//)
        {
            $device = basename($device);
        }
    }

    #
    # Get device object
    #
    $device_info{dev_obj}  = $dev_mgr_obj->get_device($device);
    if ($dev_mgr_obj->error)
    {
        $self->_set_err("Unable to get device object for $device ",
                        $dev_mgr_obj->errmsg());
        return (wantarray()) ? () : {};
    }

    #
    # Get device size
    #
    $device_info{dev_size} = $self->_get_device_size(-dev_obj => $device_info{dev_obj},
                                                     -client_type => $client_type);
    if ($self->error())
    {
        $self->_set_err("Failed to get device size ", $self->errmsg());
        return (wantarray()) ? () : {};
    }

    #
    # return device info hash
    #
    return (wantarray()) ? %device_info : \%device_info;
} # end _get_device_info()

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

San.pm

=head1 Author

David Pressley

=head1 Description

San.pm provides an API interface to a San with the main intent to simplify
setup and cleanup.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm>

=head1 Attributes

=head2 B<filers>

array = list of filer objects

=head2 B<clients>

array = list of client objects

=head2 B<switches>

array = list of switch objects

=head2 B<vol_list>

array = list of volume entries to create

=head2 B<cleanup>

array = list of hashs to describe cleanup actions

=head2 B<debug_level>

integer = Current debug level

=head2 B<errmsg> (from Error.pm)

array = list of error strings accessed via Error.pm

=head2 B<error> (from Error.pm)

boolean = <1 | 0> Flag to be set when an error occurs

=head2 B<log>

Filehandle = Filehandle to log output to

=head2 B<no_indent>

boolean = <1 | 0> Flag used when printing log messages

=head2 B<show_cmd>

type = value and description

=head2 B<execute>

boolean = <1 | 0> allow CLI commands to be formed, but not executed

=head2 B<timeout>

timeout = seconds to wait for commands to execute (defaulted to 1800, 30 min)

=head1 Public Methods

=head2 B<new()>

Creates a new San object.  Give default values to all supported attributes.

=over 4

=item Inputs

=item * -log_handle

File descriptor for a log file.

=item Outputs

Returns a San object.

=item Examples

#
# Create a San object
#
my $san_obj = new San(\*ntest::NTEST_STDOUT);
if ($san_obj->error())
{
        logresult('FATAL', "Failed to instantiate San object" . $san_obj->errmsg());
}


=back

=head2 B<setup(%args)>

Create necessary resources on filers and clients.  This includes aggrs,
vols, igroups, luns on the filer and devices, vg, lv, fs, file on the
client.

=over 4

=item Inputs

=item * -filers

Colon separated list of filer names

=item * I<-filer_pass_thru> (defaults to an empty hash)

A perl Hash that is appended to the SanFiler::new() call.
Users can pass arguments through to the SanFiler constructor
using this argument

=item * -clients

Colon separated list of client names

=item * -shared_lun <yes | no> (default: no)

Create Luns which will be shared by clients.
The shared lun created will be mapped to multiple igroups

=item * I<-client_pass_thru> (defaults to an empty hash)

A perl Hash that is appended to the SanClient::new() call.
Users can pass arguments through to the SanClient constructor
using this argument

=item * I<-switches> (defaults to an empty array)

Colon separated list of switches (currently unused)

=item * I<-protocol> <fcp | iscsi> (default: fcp)

Blocks protocol to use.

=item * I<-device_type> <raw | lv | fs> (default: raw)

Type of device to create on the client: Raw, Volume Manager, File System

=item * I<-aggr>

Name of the aggr to use or create and aggr options using the following syntasx:

      aggr_name [,block_type=32|64][raid_type=raid4|raid_dp][,timeout=#]
      Examples: "bigaggr,block_type=64" "my_raid4_aggr, raidtype=raid4"

=item * I<-num_disks> (default: 5)

Number of disks to use to create the aggr or trad vol.

=item * I<-filer_vol_list>

Colon separated list of volume defintions  where each entry has the following syntax:

      vol_count, vol_size, lun_count, lun_size [,type, prefix, aggr, options]
      vol_count - required number of volumes to create
      vol_size  - required volume size (for trad vols use number of disks)
      lun_count - required number of luns to create on each volume
      lun_size  - required size of luns created
                           if this contains a dash, "-" followed by
                           on (1 accepted) or off (0 accepted) then the
                           noreserve option can be set for lun create
                           example: 10g-off
      type      - optional flex or trad (default is flex)
      prefix    - optional string at the beginning of each volume name
                           if you are creating 2 vols and the prefix
                           is "x_vol_", we create x_vol_1 and x_vol_2.
                           Optional qtree name can be added with the
                           following syntax: prefix/qtree. For example
                           "my_vol_/my_qtree"
                           (default is san_vol_)
     aggr       - optional name of the aggr for flex vols
                           (default is test_aggr)
     options    - optional string representing options to be set on the
                           vols the syntax of this string is a space
                           separated list of name value paris:
                               "option value option value".
                           For example:
                               "svo_enable on nosnap on raidtype raid4"
                           (default is "")
     If redundant vol prefix values used, each colon separated list of volumes
     would have a vol name that is incremented from the previous.
     For Example:  FILER_VOL_LIST=3,100g,2,10g:1,1000g,1,990g
       - This would create the following set of vols(luns):
         san_vol_1(10g,10g), san_vol_2(10g,10g), san_vol_3(10g,10g),
         san_vol_4(990g)


=item * I<-lun_list> (no default)

Stanza like list of luns to configure.  E.g.: LUN_LIST="0,2g:1,1g:2,30g:5-10,250m".x
Deprecated, use -filer_vol_list.

=item * I<-filer_vol> (no default)

Name of the volume to use or create.  Deprecated, use -filer_vol_list.

=item * I<-filer_vol_size> (default: 100g)

Size of the volume if flex vol.  Deprecated, use -filer_vol_list.

=item * I<-vol_mgr> <nativelvm| vxvm | sanpath> (default: nativelvm)

Which volume manager to use on the client.

=item * I<-dg_prefix> (default: san_dg_)

Disk group prefix.  We use this to find logical volumes.

=item * I<-vol_layouts> <stripe | concat> (default: stripe)

The layout to use when creating new volumes. This value is provided as
a scalar value containing comma delimited values ('stripe,concat').
Each volume will be created with the next layout in the list. The
list will wrap around if there are more volumes than layout values.
From the example, if 3 volumes were created, the first layout would
be 'stripe', the second 'concat', and the third a 'stripe'.
Which volume manager to use on the client.

=item * I<-stripe_size> (Note: include units 'k,m,g') (default: 64k)

The stripe size to use when creating striped logical volumes.

=item * I<-vol_cnts> (default: 1)

The number of volumes to create for a disk group.  This value is provided
as a scalar value containing comma delimited values ('2,3')  This will
create 2 volumes for the first diskgroup and 3 volumes for the second
diskgroup.  The list will wrap around if there are more dikgroups than
volume counts.  From the above example, a third diskgroup would be
created with 2 volumes.

=item * I<-vol_sizes> (Note: include units 'k,m,g') (default: 500m)

The size to use when creating new volumes. This value is provided as a
scalar value containing comma delimited values ('500m,1g,2g'). Each
volume will be created with the next size in the list. The list will
wrap around if there are more volumes than size values.  From the
example, if 3 volumes were created, the first would be '500m', the
second '1g', and the third '2g'.

=item * I<-disk_fmts> <sliced | cdsdisk> (default: '')

Tells Veritas to use the specified disk format (i.e. cdsdisk, sliced)
when calling vxdisksetup.  By default we will not set this and Vx
will use it's default (cdsdisk for Veritas 4.0, sliced (or none)
for Veritas 3.5 and earlier).

=item * I<-skip_setup> (default: 0)

Do nothing when this is set.  Accepts no|yes|0|1.

=item * I<-skip_config> (default: 0)

When this is set, there are certain client configuration activities
that will be skipped.  For example, on a Solaris host, we will not
update the lpfc.conf and sd.conf file and reboot.  Accepts no|yes|0|1.

=item * I<-setup_igroups> (default: 1)

When this is set, we create an igroup for each client on each filer
and we map the luns to the igroups.  If it is '0' or 'no', we do
not create igroups or map luns.  Note, get_devices() is not expected
to work unless the luns get mapped to a proper igroup for the client.

=item * I<-vfiler_enable> (default: 0)

When this is set, vfilers will be created during setup.  A single vfiler
is created for each client specified.  The filer_vol_list is created per
vfiler.  The ip addresses for the vfilers should be specified in the clients
tharnhost file using the format specified in _get_pc_vfiler_info().

=item * I<-mp_type> <per_host | dmp | mpxio> (default: none)

When this is set, the corresponding MP object for each client will
be constructed. By default MP objects are not created (set to none).
If this value is set to "per_host" then based on the PC_CLIENT_VMS
value in each of the clients' tharnhost file, the appropriate MP
object is constructed. The MP object enables retrieval of pathing
information.

=item * I<-create_fs> <0|no|1|yes> (default: 0)

If set, attempt to create a file system on managed storage devices.
See also: C<-create_fs_type>

=item * I<-create_fs_type> (default: undef)

If C<-create_fs> is set, this argument specifies the file system used
to format a storage device.  If C<-create_fs> is set and this argument
is unspecified, an appropriate, default file system will be created
based on the host type.

=item * I<-load_devices> <0|1> (default: 1)

When this is set, the DevMgr used to manipulate host devices will
initialize any available physical devices.  If unset, the DevMgr will
skip device initialization.

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

    #
    # Instantiate San Object
    #
    my $sanObj = San::new();

    #
    # Get parameters
    #
    param('CLIENTS',     '-mesg',   "List of clients");
    param('FILERS',      '-mesg',   "List of filers");
    param('LUN_LIST',    '-mesg',   "Luns to create");
    param('PROTOCOL',    '-default', "fcp");
    param('LUN_PREFIX',  '-default', "/vol/blocks");
    param('SKIP_SETUP',  '-default', 0);

    #
    # Setup filers and clients
    # Note that giving a vol and a num_disks without an aggr or
    # vol_size will tell us to create a trad vol.
    #
    $sanObj->setup(-filers      => '$FILERS',
                   -clients     => '$CLIENTS');
                   -lun_list    => '$LUN_LIST,
                   -filer_vol   => '$LUN_PREFIX',
                   -num_disks   => 10,
                   -protocol    => '$PROTOCOL',
                   -device_type => 'raw',
                   -skip_setup  => '$SKIP_SETUP');

    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<cleanup(skip_cleanup, <cleanup_reused>)>

Destroy created resources on filers and clients.  This includes aggrs,
vols, igroups, luns on the filer and devices, vg, lv, fs, file on the
client.

=over 4

=item * skip_cleanup

Do nothing when this is set.

=item * cleanup_reused (0|1|noaggr)

If this is set, the cleanup script will cleanup reused resources as well
as resources that have just been created.  If noaggr is used we will
cleanup everything bug aggrs.

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

    param('SKIP_CLEANUP', '-default',    "0");
    $sanObj->cleanup(-skip_cleanup =>  '$SKIP_CLEANUP');
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<cleanup_add(%args)>

Add and entry to the cleanup stack.  As we setup various resources on the
filer and the client, we register their cleanup actions by calling this
method.  The cleanup() method will then just pop off actions from this
stack to cleanup the San (and thus we only cleanup what we have created).

=over 4

=item * -type <filer|client>

Are we cleaning up a filer or a client?

=item * -name

Name of the filer or client.

=item * -cmd

What are we doing?  Examples include: aggr_destroy, vol_destroy,
igroup_destroy, lun_destroy, igroup_destroy, lun_unmap, cmd.  If we are given
cmd, the plan is to just treat "cmd_args" as an executable command line
for the filer and client and execute it.

=item * -cmd_args

Arguments needed to complete the cleanup actions for the various commands.
Note that if the command is "cmd", this argument list will be an actual
command line.

=item * I<-reuse> <0 | 1> (default: 0)

Is the resource being cleaned up something we reused?

=item Outputs

Returns a scalar value of 1 if successful. If an error occured the
error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Examples

    return $self
        unless ($sanObj->cleanup_add(-type=>"filer",
                                     -name=>$filer_name,
                                     -cmd=>"vol_destroy",
                                     -cmd_args=>"$vol"));
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

    return $self
        unless ($self->cleanup_add(-type=>"filer",
                                   -name=>$filer_name,
                                   -cmd=>"lun_unmap",
                                   -cmd_args=>"$vol $lun_name $igroup_name"));
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<destroy()>

Destroy the San object used by this test.

=back

=head2 B<get_devices(%args)>

Find the devices that have been created for a client.

=over 4

=item Inputs

=item * I<-filer>

Name of a filer for raw device.  By default, we'll check all of them.

=item * I<-client>

Name of a client to find devices on.  By default, we'll check all of them.

=item * I<-filer_vol>

Volume to use if you want to filter raw devices by volume (volume filter).
By default, we return devices from any volume.

=item * I<-lun>

Use only luns that have names that match this pattern (lun filter).
By default, we return devices with any lun name.

=item * I<-lun_list>

Stanza like list of luns to configure.  E.g.: LUN_LIST="0,2g:1,1g:2,30g:5-10,250m".
Used to filter raw devices from client.
By default, we return all devices found on the client.

=item * I<-lvm> <nativelvm | vxvm | sanpath> (default: nativelvm)

Logical Volume Manager.

=item * I<-dg_prefix> (default: san_dg_)

Disk group prefix.  We use this to find logical volumes.

=item * I<-device_type> <raw | lv | fs> (default: raw)

Type of device to return from the client: Raw, Volume Manager, File System

=item * I<-protocol> <fcp | iscsi> (default: fcp)

Blocks protocol to use to find raw devices.

=item * I<-with_device_info> <0 | 1> (default: 0)

With this flag enabled, get_devices() will return an array of hashes of devices
that includes dev_name, dev_size, and dev_obj.
For Solaris, it will return the size of partition 6.
For Linux, if it is partitioned, it will return the size of partition 1.
For other OSes, it will return the size of its full partition.

=item Outputs

Returns an array of devices if successful. With the -with_device_info flag
it will return an array of hash that includes device name, device size,
and device object.
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    #
    # Note that this code will typically follow the code in the
    # setup() example above.
    #
    param('CLIENT', '-mesg',    "Name of the client");
    my @devArray = $sanObj->get_devices(-client =>  '$CLIENT');
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }
    my $io_devs = join " ", @devArray;

    #
    # You can send the $io_devs to a function to run sio
    #
    $return_code = client_io_sio( $io_devs,              # client_io_dev
                                  $CLIENT,               # client
                                  ...
                                );

    #
    # example of get_device() with -with_device_info flag enables
    #
    param('CLIENT', '-mesg',    "Name of the client");
    my @devArray = $sanObj->get_devices(-client =>  '$CLIENT',
                                        -with_device_info => 1);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }
    my $io_devs;
    my $cnt = 1;
    foreach my $device (@devArray) {
        if (defined($device->{dev_name}))
        {
            $io_devs .= " " . $device->{dev_name};
        }
    }

    #
    # Get device size from the first device
    #
    my $dev_size = $devArray[0]->{dev_size};

    #
    # You can send the $io_devs to a function to run sio
    #
    $return_code = client_io_sio( $io_devs,              # client_io_dev
                                  $CLIENT,               # client
                                  $dev_size,             # device size
                                  ...
                                );

=back

=head2 B<add_filer(filer)>

Add a filer object or create a new filer object and add it to the San.
If a filer object is given, it's added, otherwise a filer name is
expected and we will create a SanFiler object and store it.

=over 4

=item Inputs

=item * filer

Name of a filer or a SanFiler object.

=item * I<-filer_pass_thru> (defaults to an empty hash)

A perl Hash that is appended to the SanFiler::new() call.
Users can pass arguments through to the SanFiler constructor
using this argument

=item Outputs

Returns a SanFiler object if successful.
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    #
    # Instantiate San Object
    #
    my $sanObj = San::new();

    my $filerObj = $sanObj->add_filer($FILERA);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<get_lun_objs(%args)>

Get a lun object or array of lun objects

=over 4

=item Inputs

=item * I<lun_path>

A specific lun path that you would like to get the object for

=item Outputs

Will return a single lun object if the lun_path is defined, if the lun_path
is not found it will return undef, if no lun_path is defined it will return
an array of all lun objects
Note: Uses an array that only contains the luns CREATED in setup, if
using an existing lun, they will not be added to the list, and will not
be returned by this method

=item Examples

    #
    # Get lun objects
    #
    $self->get_lun_objs(-lun_path => $lun_path);

=back

=head2 B<add_client(client)>

Create a new client object and add it to the San.

=over 4

=item Inputs

=item * client

Name of a client.

=item * I<-client_pass_thru> (defaults to an empty hash)

A perl Hash that is appended to the SanClient::new() call.
Users can pass arguments through to the SanClient constructor
using this argument

=item Outputs

Returns a SanClient object if successful.
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    #
    # Instantiate San Object
    #
    my $sanObj = San::new();

    my $clientObj = $sanObj->add_client($CLIENT);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<get_filer(filer)>

Get a filer object.

=over 4

=item Inputs

=item * filer

Name of a filer.

=item Outputs

Returns a SanFiler object if successful.
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    my $filerObj = $sanObj->get_filer($FILER);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<get_client(client)>

Get a client object.

=over 4

=item Inputs

=item * client

Name of a client.

=item Outputs

Returns a SanClient object if successful.
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    my $clientObj = $sanObj->get_client($CLIENT);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<get_filer_vol_list(%args)>

Create a filer vol list string from the older parameters including lun_list
Note that if you give an aggr you get a filer_vol_list for a flex vol,
otherwise you get a filer_vol_list for a trad vol

=over 4

=item Inputs

=item * I<-aggr> (default: "" - will use trad vol)

Name of an aggregate

=item * -vol

Volume prefix

=item * -vol_size

Volume Size

=item * -lun_list

Lun list of form, num_luns:lun_size

=item Outputs

Returns a string formatted to be a filer_vol_list.

=item Examples

    my $filer_vol_list = $sanObj->get_filer_vol_list(
                                             -aggr     => $aggr,
                                             -vol      => $vol,
                                             -vol_size => $vol_size,
                                             -lun_list => $lun_list);
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }

=back

=head2 B<get_resources>

Get a client object.

=over 4

=item Inputs

=item * I<-type> (filer|client)

If given a type, only return elements that match this type

=item * I<-name>

If given a name (filer for client name), only return elements that have
a name matching this string

=item * I<-cmd>

Only return elements that have this cmd.  Valid cmds are:
aggr_create, vol_create, lun_create, igroup_create, lun_map

=item * I<-reuse> (0|1)

Only return elements that have a reuse flag matching this value.
If you only want to get the resources you have created, set this to 0.
If you only want to get the resources you reused, set this to 1.
If you want all elements, do not set this or set it to undef.

=item Outputs

Returns an array of hash elements if successful.
Each hash will have the following fields:
type, name, cmd, cmd_args, reuse
If an error occured the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Examples

    my @filer_resources = $sanObj->get_resources(-type=>"filer");
    if ($sanObj->error())
    {
        print "$sanObj->errmsg()";
        ...
    }
    foreach my $r_hash (@filer_resources) {
        print "-----------Entry start-------\n";
        foreach my $key (keys(%{$r_hash})) {
            print "Hash key is $key and value is " . $r_hash->{$key} . "\n";
        }
        print "-----------Entry finish-------\n";
    } # end foreach

=back

=cut

1;
