# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author hrivera@netapp.com
## @status private
## @pod ../Host.pod

use strict;
use warnings;

# Package Name
##############
package San::SanClient::Host::ESXHost;

# Package Requirements
######################
use POSIX;
use Data::Dumper;
use Config;
use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanCache;
use San::SanUtil;
use Error;
use vars qw( @EXPORT @ISA );

@ISA = qw(Error SanLog SanCache SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new ESXHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn, $protocol, $iscsi_type, $obj_template);

    $self = {};
    # bless
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for -protocol");
                return $self;
            }
            next;
        } elsif (/-?iscsi_type/) {
            $iscsi_type = $args{$_};
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    # check for required conn argument
    unless (defined $conn and ref($conn) eq "Connectrec") {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    # check for required protocol argument
    unless (defined $protocol) {
        $self->_set_err("Invalid value or missing value for '-protocol'");
        return $self;
    }

    $self->{conn} = $conn;
    $self->{protocol} = $protocol;

    # get the version
    my $cmd = "vmware -v";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if (scalar(@output) or $status != 0) {
        $self->sanlog(-msg => "ESX Version: @output");
    } else {
        $self->_set_err("Could not determine ESX Version");
        return $self;
    }

    # parse out the major version and the build number
    # VMware ESX Server 3.5.0 build-113339
    # VMware ESX 4.0.0 build-140815
    my ($major_version, $build_number) = ($output[0] =~ /(\d+\.\d+\.\d+)\s+build-(\d+)/);

    # store the version & build number on the object
    $self->{os_version} = $major_version;
    $self->{os_build_number} = $build_number;
    $self->{revision} = $major_version;
    $self->{minor_revision} = $build_number;

    # set the architecture.
    $cmd = "uname -p";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    foreach my $line (@output) {
        if ($line =~ /86/) {
            $self->{arch} = "$line";
            $self->sanlog(-msg => "VMware Architecture is set to $line.");
        }
    }

    # need to load config params right now so we can access them
    $self->{iscsi_type} = $iscsi_type;
    $self->_load_config_params();
    if ($self->error()) {
        return $self;
    }

    # update the vmhba adapters in the host object with those currently
    # present on the system
    $self->sanlog(-msg => "Updating vmhba adapters...");
    $self->_update_vmhba_adapters();
    if ($self->error()) {
        return $self;
    }

    # set tools bin path for ESX (really pointer to Linux binaries)
    $self->{TEST_BIN} = "/x/eng/localtest/arch/i386-redhat-linux8/bin";

    # this is the esxcli nmp namespace for all console based ESX hosts
    $self->{esxcli_nmp_namespace} = "nmp";

    # this is the esxcli nmp namespace for all console based ESX hosts
    $self->{esxcli_corestorage_namespace} = "corestorage";

    return $self;
}

################################################################################
# Subroutine Name: asl_version()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub asl_version {
     my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes. Currently no-op
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for ESX, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: configure
# Objective:       configure host machine for fcp/iscsi/mp
# Details:         See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Configure iSCSI protocol
    if ($self->{protocol} =~ /iscsi/i) {
        $self->_iscsi_configure(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # Configure FCP protocol
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_configure(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # No configuration is required for Multipathing on ESX

    # configure the logging tools
    $self->sanlog(-msg => "Configuring Log Settings");
    $self->_logging_configure();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "ESXHost SanClient Configuration is complete");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective: deconfigure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets, $clean_level);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            @targets = @{$args{$_}};
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # call appropriate deconfigure methods
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure(%args);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    if ($self->{protocol} =~ /iscsi/i) {
        # Build the args
        my %cmd_args;

        # Add in targets if defined.
        if (scalar(@targets) != 0) {
            $cmd_args{"-targets"} = \@targets;
        }

        # Add in clean_level if defined.
        if (defined($clean_level)) {
            $cmd_args{"-clean_level"} = $clean_level;
        } else {
            # default it to 1
            $cmd_args{"-clean_level"} = 1;
        }

        # return if clean level is 0
        if ($cmd_args{"-clean_level"} == 0) {
            $self->sanlog(-msg => "iSCSI Deconfigure: Not performing iSCSI deconfiguration because clean_level is '0'");
            return $ClientUtil::TRUE;
        }

        $self->_iscsi_deconfigure(%args);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: config_paths
# Objective:       Will configure any paths for the MP Object
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {

        # only perform this task if the version of ESX is running an older version than ESX4.1u2 (build502767
        # The EHU was deprecated and is not supported anymore by ESX4.1u2 and later
        if ($self->get_os_version() =~ /^(4\.1|5\.|6\.)/ and $self->{os_build_number} < 502767) {
            # dont run config mpath
            return $ClientUtil::TRUE;
        } else {
            $self->sanlog(-msg => "Configuring Paths (running config_mpath)");

            $self->{mp}->config_paths();
            if ($self->{mp}->error()) {
                $self->_set_err($self->{mp}->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: Discover(rescan) all devices
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self, %args) = @_;

    # before we start, rescan every adapter we know of to filter out
    # any orphaned luns that were left behind from another test.
    $self->orphaned_device_cleanup();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # ESX5/ESX4/ESX3 discover
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $self->_discover_esx5(%args);
    } elsif ($self->get_os_version() =~ /^(4)\./) {
        $self->_discover_esx4(%args);
    } else {
        $self->_discover_esx3(%args);
    }

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _discover_esx3
# Objective: Discover new devices for esx3
################################################################################
sub _discover_esx3 {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my ($vmhba_adapter) = "";
    my $num_expected_luns = 0;
    my ($num_rescan_retries, $num_discovered_luns);
    my $total_discovered_luns = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?num_expected_luns/) {
            if (not length ($num_expected_luns = $args{$_})) {
                $self->_set_err("Invalid value for -num_expected_luns => $num_expected_luns");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # rescan each vmhba adapter in the the adapter array reference
    foreach my $vmhba_adapter (@{$self->{vmhba_adapter}}) {
        $self->sanlog(-msg => "Discovering new luns on $vmhba_adapter");

        $cmd = "/usr/sbin/esxcfg-rescan " . $vmhba_adapter;
        $num_rescan_retries = 2;
        $num_discovered_luns = 0;

        # we may have to rescan several times until we find what we want
        for (my $i = 0; $i < $num_rescan_retries; $i++) {
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to rescan using command '$cmd'" , @output);
                return $ClientUtil::FALSE;
            }

            # pull out the line that discovered the luns.
            @output = grep { /^On scsi\d+, adding:/i } @output;

            # count the luns in output[0] that are of X:Y format
            $num_discovered_luns = ($output[0] =~ s/\d+:\d+//g);

            # sleep and rescan again, there is no way to determine
            # which luns are the ones we were looking for, so just rescan a
            # bunch of times.
            &SanUtil::nate_sleep(1);
        }

        # update the total number of discovered luns
        $total_discovered_luns += $num_discovered_luns;
    }

    # check if the number of luns discovered is at least greater than what we
    # were expecting.
    if ($total_discovered_luns >= $num_expected_luns) {
        $self->sanlog(-msg => "SanClient Discover is complete. Discovered".
                              " ($total_discovered_luns) luns. ".
                              "Expected ($num_expected_luns) luns");
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Expected ($num_expected_luns) luns but only ".
                        "($total_discovered_luns) were discovered.");
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: _discover_esx4
# Objective: Discover new devices for esx4
# Caveats: Not using num_expected_luns because rescan output is quiet
################################################################################
sub _discover_esx4 {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my ($vmhba_adapter) = "";
    my $num_expected_luns = 0;
    my ($num_rescan_retries, $num_discovered_luns);
    my $total_discovered_luns = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?num_expected_luns/) {
            if (not length ($num_expected_luns = $args{$_})) {
                $self->_set_err("Invalid value for -num_expected_luns => $num_expected_luns");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # rescan each vmhba adapter in the the adapter array reference
    foreach my $vmhba_adapter (@{$self->{vmhba_adapter}}) {
        $self->sanlog(-msg => "Discovering new luns on $vmhba_adapter");

        $cmd = "esxcfg-rescan " . $vmhba_adapter;
        $num_rescan_retries = 2;
        my $retry_count = 2;

        # we may have to rescan several times until we find what we want
        for (my $i = 0; $i < $num_rescan_retries; $i++) {
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                if (grep(/Scan operation failed/i, @output) and $retry_count > 0) {
                    $retry_count--;
                    &SanUtil::nate_sleep(5);
                    redo;
                }
                $self->_set_err("Failed to rescan using command '$cmd'");
                return $ClientUtil::FALSE;
            } else {
                $retry_count = 2;
            }

            # sleep and rescan again, there is no way to determine
            # which luns are the ones we were looking for, so just rescan a
            # bunch of times.
            &SanUtil::nate_sleep(1);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _discover_esx5
# Objective: Discover new devices for esx5
# Caveats: Not using num_expected_luns because rescan output is quiet
################################################################################
sub _discover_esx5 {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my ($vmhba_adapter) = "";
    my ($num_rescan_retries);

    $self->_clear_err();

    my $corestorage_ns = $self->{esxcli_corestorage_namespace};

    # rescan each vmhba adapter in the the adapter array reference
    foreach my $vmhba_adapter (@{$self->{vmhba_adapter}}) {
        $self->sanlog(-msg => "Discovering new luns on $vmhba_adapter");

        $cmd = "esxcli ${corestorage_ns} adapter rescan --adapter " . $vmhba_adapter;
        $num_rescan_retries = 2;
        my $retry_count = 2;

        # we may have to rescan several times until we find what we want
        for (my $i = 0; $i < $num_rescan_retries; $i++) {
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($status > 0) {
                if (grep(/Scan operation failed/i, @output) and $retry_count > 0) {
                    $retry_count--;
                    &SanUtil::nate_sleep(5);
                    redo;
                }
                $self->_set_err("Failed to rescan using command '$cmd'");
                return $ClientUtil::FALSE;
            } else {
                $retry_count = 2;
            }

            # sleep and rescan again, there is no way to determine
            # which luns are the ones we were looking for, so just rescan a
            # bunch of times.
            &SanUtil::nate_sleep(1);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: power_management()
# Objective:       This method controls what actions to take on Virtual Machine
#                  power management.
# Details:         See POD documentation at the end of this file
################################################################################
sub power_management {
    my ($self, %args) = @_;

    my ($action);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Error: ESXHost::power_management() Invalid " .
                                "value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($action =~ /^power_on$/i) {
        $self->_power_on_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on_all$/i) {
        $self->_power_on_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off$/i) {
        $self->_power_off_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off_all$/i) {
        $self->_power_off_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_reset$/i) {
        $self->_power_reset_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_reset_all$/i) {
        $self->_power_reset_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle$/i) {
        $self->_power_cycle_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle_all$/i) {
        $self->_power_cycle_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Error: ESXHost::power_management() Invalid value for ".
                        "-action : ($action)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: clone_vmdk()
# Objective:       ESX only subroutine, clone the source VMDK to the destination
#                  VMDK.
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_vmdk {
    my ($self, %args) = @_;

    my ($src_file, $dest_file, $disk_format, $adapter_type, $rdm_device);
    $disk_format = "eagerzeroedthick";
    $adapter_type = "lsilogic";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?source_vmdk_file/) {
            if (not length($src_file = $args{$_})) {
                $self->_set_err("Invalid value for argument -src_file");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?destination_vmdk_file/) {
            if (not length($dest_file = $args{$_})) {
                $self->_set_err("Invalid value for argument -dest_file");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?disk_format/) {
            if (not length($disk_format = $args{$_})) {
                $self->_set_err("Invalid value for argument -disk_format");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?rdm_device/) {
            if (not length($rdm_device = $args{$_})) {
                $self->_set_err("Invalid value for argument -rdm_device");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?adapter_type/) {
            if (not length($adapter_type = $args{$_})) {
                $self->_set_err("Invalid value for argument -adapter_type");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # verify that disk format was specified properly
    if ($disk_format !~ /^(rdm|rdmp|eagerzeroedthick|zeroedthick|thin|2gbsparse)$/) {
        $self->_set_err("Invalid -disk_format '$disk_format', must be either 'rdm', 'rdmp', 'eagerzeroedthick', 'zeroedthick', 'thin' or '2gbsparse'");
        return $ClientUtil::FALSE;
    }

    # if the disk format is RDM, then rdm_device must be defined
    # or the rdm device must be specified properly
    if ($disk_format =~ /rdm|rdmp/) {
        if (not defined($rdm_device)) {
            $self->_set_err("-rdm_device must be specified if -disk_format => 'rdm' or -disk_format => 'rdmp'");
            return $ClientUtil::FALSE;
        } elsif ($rdm_device !~ /^naa\.[a-zA-Z0-9]+\s*$|^\/vmfs\/devices\/disks\/naa\.[a-zA-Z0-9]+\s*$/) {
            $self->_set_err("-rdm_device is invalid, must be something like: 'naa.60a980004334652f4f5a655359417955'");
            return $ClientUtil::FALSE;
        }
    }

    # if we are on a vMA host, then we need to convert the src file and dest file to datastore format
    my $format;
    my $uuid;
    if ($self->has_vma_host()) {
        $format = "datastore_path";
        $uuid = 0;
    } else {
        $format = "file_path";
        $uuid = 1;
    }

    $src_file = $self->_convert_to_datastore_path(-file_path => $src_file,
                                                  -format => $format,
                                                  -use_uuid => $uuid);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    $dest_file = $self->_convert_to_datastore_path(-file_path => $dest_file,
                                                   -format => $format,
                                                   -use_uuid => $uuid);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # perform some special option settings if they are cloning to RDM
    if ($disk_format =~ /rdm|rdmp/) {
        # make sure rdm device is the full path to the device
        if ($rdm_device !~ /^\/vmfs\/devices\/disks\//) {
            $rdm_device = "/vmfs/devices/disks/" . $rdm_device;
        }

        # rdm destination cloning has to be specified slightly differently
        $disk_format .= ":${rdm_device}";
    }

    # execute the clone operation
    my $cmd = "vmkfstools --clonevirtualdisk '$src_file' '$dest_file' --diskformat $disk_format -a $adapter_type";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to clone vmdk file '$src_file' to '$dest_file' (format: $disk_format, adaptertype: $adapter_type)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vmdk()
# Objective:       ESX only subroutine, create a VMDK
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vmdk {
    my ($self, %args) = @_;

    my ($vmdk_file_path, $disk_format, $adapter_type, $rdm_device, $vmdk_size_in_kbytes);
    $disk_format = "";
    $adapter_type = "lsilogic";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?vmdk_file_path/) {
            if (not length($vmdk_file_path = $args{$_})) {
                $self->_set_err("Invalid value for argument -vmdk_file_path");
                return ();
            }
            next;
        } elsif (/-?disk_format/) {
            $disk_format = $args{$_};
            next;
        } elsif (/-?rdm_device/) {
            if (not length($rdm_device = $args{$_})) {
                $self->_set_err("Invalid value for argument -rdm_device");
                return ();
            }
            next;
        } elsif (/-?adapter_type/) {
            if (not length($adapter_type = $args{$_})) {
                $self->_set_err("Invalid value for argument -adapter_type");
                return ();
            }
            next;
        } elsif (/-vmdk_size_in_kbytes?/) {
            if (not length($vmdk_size_in_kbytes = $args{$_})) {
                $self->_set_err("Invalid value for argument -vmdk_size_in_kbytes");
                return ();
            }
            next;
        }
    }

    # if the disk format is RDM, then rdm_device must be defined
    # or the rdm device must be specified properly
    if ($disk_format =~ /rdm|rdmp/) {
        if (not defined($rdm_device)) {
            $self->_set_err("-rdm_device must be specified if -disk_format => 'rdm' or -disk_format => 'rdmp'");
            return ();
        } elsif ($rdm_device !~ /^naa\.[a-zA-Z0-9]+\s*$|^\/vmfs\/devices\/disks\/naa\.[a-zA-Z0-9]+\s*$/) {
            $self->_set_err("-rdm_device is invalid, must be something like: 'naa.60a980004334652f4f5a655359417955'");
            return ();
        }
    }

    # if we are on a vMA host, then we need to convert the src file and dest file to datastore format
    my $format;
    my $uuid;
    if ($self->has_vma_host()) {
        $format = "datastore_path";
        $uuid = 0;
    } else {
        $format = "file_path";
        $uuid = 1;
    }

    $vmdk_file_path = $self->_convert_to_datastore_path(-file_path => $vmdk_file_path,
                                                        -format    => $format,
                                                        -use_uuid  => $uuid);
    if ($self->error()) {
        return ();
    }

    # perform some special option settings if they are cloning to RDM
    if ($disk_format =~ /rdm|rdmp/) {
        # make sure rdm device is the full path to the device
        if ($rdm_device !~ /^\/vmfs\/devices\/disks\//) {
            $rdm_device = "/vmfs/devices/disks/" . $rdm_device;
        }

        # rdm destination cloning has to be specified slightly differently
        $disk_format .= ":${rdm_device}";
    }

    # execute the clone operation
    my $cmd = "";
    if ($disk_format =~ /rdmp/) {
        $cmd = "vmkfstools --createrdmpassthru $rdm_device '$vmdk_file_path' -a $adapter_type" #cmd for RDMp vmdks
    } elsif ($disk_format =~ /rdm/) {
        $cmd = "vmkfstools --createrdm $rdm_device '$vmdk_file_path' -a $adapter_type"         #cmd for RDM vmdks
    } elsif (length($disk_format)) {
        $cmd = "vmkfstools -c $vmdk_size_in_kbytes -a $adapter_type --diskformat $disk_format \'$vmdk_file_path\'" #cmd for VMFS vmdks
    } else {
        $cmd = "vmkfstools -c $vmdk_size_in_kbytes -a $adapter_type \'$vmdk_file_path\'"       #cmd for VMFS vmdks
    }

    return $self->_esx_execute_cmd(-cmd => $cmd);
}

################################################################################
# Subroutine Name: delete_vmdk()
# Objective:       ESX only subroutine, delete a specified VMDK file
# Details:         See POD documentation at the end of this file
################################################################################
sub delete_vmdk {
    my ($self, %args) = @_;

    my $vmdk_file;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?vmdk_file/) {
            if (not length($vmdk_file = $args{$_})) {
                $self->_set_err("Invalid value for argument -vmdk_file");
                return $ClientUtil::FALSE;
            }
        }
    }

    if (not defined($vmdk_file)) {
        $self->_set_err("Missing required parameter: -vmdk_file");
        return $ClientUtil::FALSE;
    }

    my $cmd = "vmkfstools -U $vmdk_file";
    my ($status, @output) = $self->_vmkfstools_cmd($cmd);
    if ($status > 0) {
        $self->_set_err("Failed to delete VMDK file $vmdk_file");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_vmdks_to_vm()
# Objective:       ESX only subroutine. Will assign all vmdk files in
#                  @vmfs_vmdk_files and @rdm_vmdk_files  to $vm's
#                  configuration file.
# Details:         See POD documentation at the end of this file
################################################################################
sub add_vmdks_to_vm {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd);
    my ($vm_name, @vmfs_vmdk_files, @rdm_vmdk_files);
    my ($vm_config_file_path, @config_file, $controller, $preclean);

    # by default, preclean the config file
    $preclean = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument -vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vmfs_vmdk_files/) {
            @vmfs_vmdk_files = @{$args{$_}};
            next;
        } elsif (/-?rdm_vmdk_files/) {
            @rdm_vmdk_files = @{$args{$_}};
            next;
        } elsif (/-?controller/) {
            if (not length($controller = $args{$_})) {
                $self->_set_err("Invalid value for argument -controller");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?preclean/) {
            if (not length($preclean = $args{$_})) {
                $self->_set_err("Invalid value for argument -preclean");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter '-vm_name'");
        return $ClientUtil::FALSE;
    }
    if (not @vmfs_vmdk_files and
        not @rdm_vmdk_files) {
        $self->_set_err("You must provide at least one of the " .
                        "following (-vmfs_vmdk_files | -rdm_vmdk_files)");
        return $ClientUtil::FALSE;
    }
    if (not defined($controller)) {
        $self->_set_err("You must provide a -controller parameter," .
                        "use either lsilogic | buslogic | lsisas1068 | pvscsi");
        return $ClientUtil::FALSE;
    }

    # convert all the vmdk file paths to "file_path" format if they are not in it already
    # convert to use the uuid instead of datastore name
    if (scalar(@vmfs_vmdk_files)) {
        @vmfs_vmdk_files = map { $self->_convert_to_datastore_path(-file_path => $_,
                                                                   -format => "file_path",
                                                                   -use_uuid => 1); } @vmfs_vmdk_files;
        if ($self->error()) {
            $self->_set_err("VMFS vmdk conversion to datastore path failed", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }
    if (scalar(@rdm_vmdk_files)) {
        @rdm_vmdk_files = map { $self->_convert_to_datastore_path(-file_path => $_,
                                                                  -format => "file_path",
                                                                  -use_uuid => 1); } @rdm_vmdk_files;
        if ($self->error()) {
            $self->_set_err("RDM vmdk conversion to datastore path failed", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    my $num_vmfs_vmdks = scalar(@vmfs_vmdk_files);
    my $num_rdm_vmdks = scalar(@rdm_vmdk_files);

    $self->sanlog(-msg => "Adding VMDK Files (VMFS:$num_vmfs_vmdks, RDM:$num_rdm_vmdks) to $vm_name\'s config file.\n" .
                          "VMFS:\n " . join("\n ", @vmfs_vmdk_files) . "\n" .
                          "RDM:\n "  . join("\n ", @rdm_vmdk_files));

    # get this vm's configuration file
    $vm_config_file_path = $self->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check if this vm has its config file in the cache.
    if ($self->is_cache_data_valid("$vm_config_file_path")) {
        @output = $self->get_cache_data("$vm_config_file_path");
    } else {
        # cat the config file
        $cmd = "cat \"$vm_config_file_path\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    if ($preclean) {
        # this will filter out any previous changes done by this function from
        # the config file..
        @config_file = $self->_filter_config_file(@output);

        # unshift the #!/usr/bin/vmware on the config file since
        # _filter_config_file() removed all comments from the file.
        unshift(@config_file, "#!/usr/bin/vmware");

    } else {
        @config_file = @output;
    }

    # now add vmdk modifications to the config file
    push(@config_file, " ");
    push(@config_file, " ");
    push(@config_file, "#__BEGIN_NATE_AUTOMATION__#");
    push(@config_file, " ");

    # for each vmfs vmdk file, add it to the config_file array
    foreach my $vmfs_vmdk_file (@vmfs_vmdk_files) {
        # get the next available scsi slot to use
        my ($create_new_bus, $bus_counter, $lun_counter) = $self->_get_next_scsi_id(\$controller, \@config_file);

        # if there does not exist a scsi controller for
        # the current bus, add one.
        if (not grep {/^\s*scsi$bus_counter\./} @config_file or $create_new_bus == 1) {
            # remove any remnants of this scsi bus just in case there is something there
            @config_file = grep {!/^\s*scsi${bus_counter}\./} @config_file;

            push(@config_file, "scsi" . $bus_counter . ".present = \"true\"");
            push(@config_file, "scsi" . $bus_counter . ".sharedBus = \"none\"");
            push(@config_file, "scsi" . $bus_counter . ".virtualDev = \"$controller\"");
            push(@config_file, " ");
        }

        # push the vmdk information on the config_file
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".present = \"true\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".fileName = \"" . $vmfs_vmdk_file . "\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".mode = \"persistent\" ");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".redo = \"\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".deviceType = \"scsi-hardDisk\"");

        push(@config_file, " ");
    }

    # for each rdm vmdk file, add it to the config_file array
    foreach my $rdm_vmdk_file (@rdm_vmdk_files) {
        # get the next available scsi slot to use
        my ($create_new_bus, $bus_counter, $lun_counter) = $self->_get_next_scsi_id(\$controller, \@config_file);

        # if there does not exist a scsi controller for
        # the current bus, add one
        if (not grep {/^\s*scsi$bus_counter\./} @config_file or $create_new_bus == 1) {
            # remove any remnants of this scsi bus just in case there is something there
            @config_file = grep {!/^\s*scsi${bus_counter}\./} @config_file;

            push(@config_file, "scsi" . $bus_counter . ".present = \"true\"");
            push(@config_file, "scsi" . $bus_counter . ".sharedBus = \"none\"");
            push(@config_file, "scsi" . $bus_counter . ".virtualDev = \"$controller\"");
            push(@config_file, " ");
        }

        # push the vmdk information on the config file
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".present = \"true\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".fileName = \"" . $rdm_vmdk_file . "\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".mode = \"independent-persistent\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".redo = \"\"");
        push(@config_file, "scsi" . $bus_counter . ":" . $lun_counter . ".deviceType = \"scsi-hardDisk\"");

        push(@config_file, " ");
    }

    push(@config_file, "#__END_NATE_AUTOMATION__#");

    # Recreate the config file for this vm!
    my $vm_config_file_path_new = $vm_config_file_path . ".tmp." . time();
    $self->sanlog(-msg => "Creating new config file for $vm_name",
                  -msg_level => 3);

    # first touch the temp file that we plan to build, retry this command if it fails
    $cmd = "touch \"$vm_config_file_path_new\"";
    ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 60, 3, "-comment");
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # echo each line into the temp file
    foreach my $line (@config_file) {
        $cmd = "echo \'$line\' >> \"$vm_config_file_path_new\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # now overwrite the old config file with the new one
    $self->sanlog(-msg => "Overwriting $vm_name\'s old config file with new config file.",
                  -msg_level => 3);
    $cmd = "mv --backup --suffix=\".bak\" \"$vm_config_file_path_new\" \"$vm_config_file_path\"";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # give the new config file executable permissions
    $cmd = "chmod +x \"$vm_config_file_path\"";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # update the cache with the new data for this file
    $self->store_cache_data("$vm_config_file_path", @config_file);

    $self->sanlog(-msg => "Updated $vm_name\'s config file.");

    return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: create_virtual_machine_snapshot()
# Objective:       Perform create snapshot operation.
################################################################################
sub create_virtual_machine_snapshot {
    my ($self, %args) = @_;
    my ($vm_name,$snapshot_name,$snapshot_description);
    my ($cmd, $status, @output);
    my $memory = 1;
    my $quiesce = 1;
    my $snapshot_timeout = -1; 
    $self->_clear_err();
    # parse the input arguments
    foreach (keys %args) {
        if (/-?vm_name/) {	
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?snapshot_name/) {
            if (not length($snapshot_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot_name'");
                return $ClientUtil::FALSE;
            }
	    $snapshot_description="snapshotCreation" ;
            next;
        } elsif (/-?memory/) {
            if (not length($memory = $args{$_})) {
	}
            next;
        } elsif (/-?quiesce/) {
            if (not length($quiesce = $args{$_})) {
            }	
            next;
        }
    }
    #return false , if mandatory parameters are not passed
    if($vm_name eq '')
    {
	$self->_set_err("Required argument vm_name not passed ");
        return $ClientUtil::FALSE;
    }
    if($snapshot_name eq '')
    {
	$self->_set_err("Required argument snapshot_name not passed ");
        return $ClientUtil::FALSE;
    }
    # convert vm_name to its vmx file path
    my $vm_config_file_path = $self->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->error()) {
	$self->_set_err("ERROR : Converting vm_name to its vmx file path ");
        return $ClientUtil::FALSE;
    }
    # issue the command to perform snapshot creation
    $cmd = "vmware-cmd ".
		   "${vm_config_file_path} ".
	       "createsnapshot ".
		   "${snapshot_name} ".
		   "${snapshot_description} ".
		   "${quiesce} ".
		   "${memory} ";
    ($status, @output) = $self->_vmware_cmd($cmd);		 
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd) : ERROR : @output ");
        return $ClientUtil::FALSE;
    }else{
	&SanUtil::nate_log('-comment',
	                               "svmotion_w_snapshot_fault.thpl: created snapshot successfully");
	return $ClientUtil::TRUE;
    }
}
################################################################################
# Subroutine Name: cleanup_vm_config_file()
# Objective:       ESX only subroutine. Will cleanup any config file
#                  modifications made to each vm in @vm_lists.
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup_vm_config_file {
    my ($self, %args) = @_;
    my ($status, @output, $cmd);
    my (@vm_list);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Error: ESXHost::cleanup_vm_config_file()" .
                                "Invalid value for argument '-vm_list' ".
                                "=> @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not @vm_list) {
        $self->_set_err("Error: ESXHost::cleanup_vm_config_file()" .
                        "Missing required argument list: -vm_list");
        return $ClientUtil::FALSE;
    }

    foreach my $vm (@vm_list) {
        $self->sanlog(-msg => "Cleaning up $vm\'s config file.");

        # grab this vm's configuration file
        my $vm_config_file_path = $self->_get_vm_config_file(-vm_name => $vm);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # check if this vm has its config file in the cache.
        if ($self->is_cache_data_valid("$vm_config_file_path")) {
            @output = $self->get_cache_data("$vm_config_file_path");
        } else {
            # cat this file from the host
            $cmd = "cat \"$vm_config_file_path\"";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Error: ESXHost::cleanup_vm_config_file() " .
                                "Command ($cmd) returned a bad status " .
                                "($status)");
                return $ClientUtil::FALSE;
            }
        }

        # this will filter out all changes made to the config file done
        # by add_vmdks_to_vm()
        my @config_file = $self->_filter_config_file(@output);

        # unshift the #!/usr/bin/vmware on the config file
        unshift(@config_file, "#!/usr/bin/vmware");

        # Let's recreate the config file for this vm!
        $self->sanlog(-msg => "Creating $vm\'s new config file.",
                      -msg_level => 3);
        my $vm_config_file_path_new = $vm_config_file_path . ".tmp." . time();

        # first touch the temp file that we plan to build, retry this command if it fails
        $cmd = "touch \"$vm_config_file_path_new\"";
        ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 60, 3, "-comment");
        if ($status > 0) {
            $self->_set_err("($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # echo each line into the temp file
        foreach my $line (@config_file) {
            $cmd = "echo \'$line\' >> \"$vm_config_file_path_new\"";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Error: ESXHost::cleanup_vm_config_file() Command ".
                                "($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # now overwrite the old config file with the new one
        $self->sanlog(-msg => "Overwriting $vm\'s old config file with new config file.",
                      -msg_level => 3);
        $cmd = "mv --backup --suffix=\".bak\" \"$vm_config_file_path_new\" \"$vm_config_file_path\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::cleanup_vm_config_file() Command ".
                                "($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # give the config file executable permissions
        $cmd = "chmod +x \"$vm_config_file_path\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::cleanup_vm_config_file() ".
                            "Command ($cmd) returned a bad status " .
                            "($status)");
            return $ClientUtil::FALSE;
        }

        # update the cache with the new data for this file
        $self->store_cache_data("$vm_config_file_path", @config_file);

        $self->sanlog(-msg => "Cleaned up $vm\'s config file.");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_vmdks_from_vm()
# Objective:       ESX only subroutine. Will return a list of vmdk files
#                  in use by the indicated -vm_name of type -lvm_type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vmdks_from_vm {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);
    my ($vm_name, $vmdk_filepath_regex_filter, $vm_config_file_path);
    my (@vmdk_files);
    my $normalize_file_paths = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name' ".
                                "=> $vm_name");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vmdk_filepath_regex_filter/) {
            if (not length($vmdk_filepath_regex_filter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmdk_filepath_regex_filter' ".
                                "=> $vmdk_filepath_regex_filter");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?normalize_file_paths/) {
            if (not length($normalize_file_paths = $args{$_})) {
                $self->_set_err("Invalid value for argument '-normalize_file_paths' ".
                                "=> $normalize_file_paths");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($vm_name)) {
        $self->_set_err("Missing required argument -vm_name");
        return (wantarray()) ? () : [];
    }

    if (not defined($vmdk_filepath_regex_filter)) {
        $self->_set_err("Missing required argument -vmdk_filepath_regex_filter");
        return (wantarray()) ? () : [];
    }

    # get the vm config file for this vm host
    $vm_config_file_path = $self->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    # check if this vm has its config file in the cache.
    if ($self->is_cache_data_valid("$vm_config_file_path")) {
        @output = $self->get_cache_data("$vm_config_file_path");
    } else {
        # cat this file from the host
        $cmd = "cat \"$vm_config_file_path\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }

        # update the cache with the new data for this file
        $self->store_cache_data("$vm_config_file_path", @output);
    }

    # grab all vmdk file paths from @output
    @vmdk_files = grep {/.*scsi\d+:\d+\.fileName\s*\=\s*\"(.+?)\"/i } @output;

    # take out all the vmx syntax around the file name on each vmdk file
    @vmdk_files = map {
        s/.*?fileName\s*\=\s*\"(.+?)\"/$1/i;
        $_;
    } @vmdk_files;

    # filter the vmdk files by the vmdk_filepath_regex_filter
    if (defined($vmdk_filepath_regex_filter)) {
        @vmdk_files = grep {
            if ($_ =~ $vmdk_filepath_regex_filter) {
                1;
            } else {
                0;
            }
        } @vmdk_files;
    }

    if ($normalize_file_paths) {
        @vmdk_files = map {
            $self->_convert_to_datastore_path(-file_path => $_,
                                              -format => "file_path",
                                              -use_uuid => 0);
        } @vmdk_files;
    }

    # return vmdk file list
    return (wantarray()) ? @vmdk_files : \@vmdk_files;
}

################################################################################
# Subroutine Name: format
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self) = @_;

    $self->_clear_err();

    return &ClientUtil::FALSE;
}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       NOOP for ESX
###############################################################################
sub get_mounted_device {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported lvm types for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vmfs rdm);
}

###############################################################################
# Method:    get_mp_default
# Objective: Get the default multipathing stack
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_mp_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(native);
}

###############################################################################
# Method:    get_supported_mps
# Objective: Get the list of supported multipathing stack
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_mps {
    my ($self) = @_;

    $self->_clear_err();

    return qw(native);
}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp iscsi nfs);
}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vmfs3 vmfs2);
}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my (@output);

    push @output,
        "Host name:      ".$self->hostname(),
        "Host type:      ".$self->hosttype(),
        "OS version:     ".$self->os_version(),
        (
            "Sanlun version: ".$self->sanlun_version(),
            "ASL version:    ".$self->asl_version(),
        ) if ($self->get_protocol() =~ /fcp|iscsi/i);

    push @output,"\n";

    return (wantarray()) ? @output: \@output;
}

################################################################################
# Subroutine Name: iostat
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? () : [];
}

################################################################################
# Subroutine Name: list_vmfs_datastores
# Objective:       list all vmfs datastores
# Details:         See POD documentation at the end of this file
################################################################################
sub list_vmfs_datastores {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "esxcfg-scsidevs --vmfs";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute cmd ($cmd)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       get NFS mounts currently mounted on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my @nfs_mounts;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    $cmd = "/usr/sbin/mount";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to execute command: " . $cmd);
        return (wantarray()) ? () : [];
    }

    # Capture the output of mount for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Look through each line of output and determint if there is a filer in the $targets list mounted to this host
        my @mounts = grep(/$target/i, @output);
        if (scalar(@mounts) == 0) {
            next;
        }

        # Go through the found mounts and populate the return array
        foreach my $mount (@mounts) {
            # /mnt/qa/nate_1_1_hp2620_rtp11 on fas2050-rtp03:/vol/nate_1_1_hp2620_rtp11 rsize=32768,wsize=32768,NFSv3,dev=2000004 on Tue Jul 14 12:57:19 2009
            my ($nfs_mount, undef, $nfs_share, $nfs_options) = split(/\s+/, $mount);
            my $nfs_pt;
            $nfs_pt->{mount} = $nfs_mount;
            $nfs_pt->{share} = $nfs_share;
            $nfs_pt->{options} = $nfs_options;

            push(@nfs_mounts, $nfs_pt);
        }
    }

    if (scalar(@nfs_mounts) == 0) {
        $self->_set_err("Did not find any mounts associated with targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_mounts : \@nfs_mounts;
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       get available NFS shares volumes via showmount -e
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my $interfaces = undef;
    my @nfs_shares;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?interfaces/) {
            if (not length($interfaces = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interfaces'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # If the interface list is not passed in, get it
    if (not defined($interfaces)) {
        $interfaces = $self->get_ip_info();
    }

    # Capture the output of showmount -e for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Collect showmount -e
        $cmd = "/usr/sbin/showmount -e " . $target;
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);

        # Look for the two following erros, they failures.
        # showmount: <device_name>: RPC: Program not registered
        # showmount: <device_name>: RPC: Unknown host
        if (grep(/showmount\:.+RPC\:/, @output)) {
            $self->_set_err("RPC error occurred during command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        # Grep through the output and only get the ones that have an ip address in the interface list
        foreach my $interface (keys(%{$interfaces})) {
            my $ipv4 = $interfaces->{$interface}->{ipv4};
            my $ipv6 = $interfaces->{$interface}->{ipv6};
            my @results = ();

            # If we have an ipv4 address specified, add it to the results array
            if (defined($ipv4) and $ipv4 =~ /\d+\.\d+\.\d+\.\d+/) {
                push(@results, grep(/$ipv4/i, @output));
            }

            # If we have an ipv6 address specified, add it to the results array
            if (defined($ipv6)) {
                push(@results, grep(/$ipv6/i, @output));
            }

            if (scalar(@results) == 0) {
                $self->sanlog(-msg => "Did not find any shares for interface: $ipv4 $ipv6");
                next;
            }

            # Parse the output and populate nfs_mounts
            foreach my $line (@results) {
                # /vol/nate_1_1_hp2620_rtp11   10.60.242.103
                # /vol/nate_1_1_hprx2600_rtp14 10.60.242.105
                my ($vol) = split(/\s+/, $line);

                my $nfs_share;
                $nfs_share->{volume} = basename($vol);
                $nfs_share->{path} = $vol;
                $nfs_share->{ipv4_address} = $ipv4;
                $nfs_share->{ipv6_address} = $ipv6;
                $nfs_share->{target} = $target;

                push(@nfs_shares, $nfs_share);
            }
        }
    }

    # See if no NFS shares were found
    if (scalar(@nfs_shares) == 0) {
        $self->_set_err("No NFS shares found for targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_shares : \@nfs_shares;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self) = @_;

    $self->_clear_err();
    $self->_set_err("Error: ESXHost::unmount_fs() This function is not " .
                    "supported");
    return $self;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;
    my ($nfs_host, $nfs_dir_path, $nfs_mount_point);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?nfs_host/) {
            if (not length($nfs_host = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_mount() ".
                                "Invalid value for argument '-nfs_host' ".
                                "=> $nfs_host");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nfs_dir_path/) {
            if (not length($nfs_dir_path = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_mount() ".
                                "Invalid value for argument '-nfs_dir_path' ".
                                "=> $nfs_dir_path");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nfs_mount_point/) {
            if (not length($nfs_mount_point = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_mount() ".
                                "Invalid value for argument '-nfs_mount_point' ".
                                "=> $nfs_mount_point");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($nfs_host)) {
        $self->_set_err("Error: ESXHost::nfs_mount() ".
                        "Missing required argument -nfs_host");
        return $ClientUtil::FALSE;
    } elsif (not defined($nfs_dir_path)) {
        $self->_set_err("Error: ESXHost::nfs_mount() ".
                        "Missing required argument -nfs_dir_path");
        return $ClientUtil::FALSE;
    } elsif (not defined($nfs_mount_point)) {
        $self->_set_err("Error: ESXHost::nfs_mount() ".
                        "Missing required argument -nfs_mount_point");
        return $ClientUtil::FALSE;
    }

    # enable the nfsClient service
    my ($cmd) = "/usr/sbin/esxcfg-firewall --enableService nfsClient";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::nfs_mount() ".
                        "Unable to enable nfsClient service.");
        return $ClientUtil::FALSE;
    }

    # list all the current mounted nfs hosts
    $cmd = "/usr/sbin/esxcfg-nas --list";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::nfs_mount() " .
                        "Could not list current NFS mount points.");
        return $ClientUtil::FALSE;
    }

    # chop off the trailing / on $nfs_mount_point
    chop($nfs_mount_point) if ($nfs_mount_point =~ /\/$/);

    # check if this nfs mount point is already mounted.
    # @output will look like this
    # lethe_share is /vol/rtpdist/ from lethe mounted
    if (grep(/$nfs_mount_point is $nfs_dir_path\/? from $nfs_host mounted/, @output)) {
        # already mounted, return true
        $self->sanlog(-msg => "NFS volume $nfs_host\:$nfs_dir_path is already ".
                              "mounted on '$nfs_mount_point'");
        return $ClientUtil::TRUE;
    }

    # otherwise its not mounted, mount it.
    $cmd = "/usr/sbin/esxcfg-nas --add --host $nfs_host --share $nfs_dir_path $nfs_mount_point";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::nfs_mount() " .
                        "Could not mount NFS host $nfs_host\:$nfs_dir_path ".
                        "on mount point $nfs_mount_point");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Mounted NFS volume $nfs_host\:$nfs_dir_path ".
                          "mounted on '$nfs_mount_point'");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount {
    my ($self, %args) = @_;
    my ($nfs_host, $nfs_dir_path, $nfs_mount_point);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?nfs_host/) {
            if (not length($nfs_host = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_unmount() ".
                                "Invalid value for argument '-nfs_host' ".
                                "=> $nfs_host");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nfs_dir_path/) {
            if (not length($nfs_dir_path = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_unmount() ".
                                "Invalid value for argument '-nfs_dir_path' ".
                                "=> $nfs_dir_path");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nfs_mount_point/) {
            if (not length($nfs_mount_point = $args{$_})) {
                $self->_set_err("Error: ESXHost::nfs_unmount() ".
                                "Invalid value for argument '-nfs_mount_point' ".
                                "=> $nfs_mount_point");

           return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($nfs_host)) {
        $self->_set_err("Error: ESXHost::nfs_unmount() ".
                        "Missing required argument -nfs_host");
        return $ClientUtil::FALSE;
    } elsif (not defined($nfs_dir_path)) {
        $self->_set_err("Error: ESXHost::nfs_unmount() ".
                        "Missing required argument -nfs_dir_path");
        return $ClientUtil::FALSE;
    } elsif (not defined($nfs_mount_point)) {
        $self->_set_err("Error: ESXHost::nfs_unmount() ".
                        "Missing required argument -nfs_mount_point");
        return $ClientUtil::FALSE;
    }

    # enable the nfsClient service
    my ($cmd) = "/usr/sbin/esxcfg-firewall --enableService nfsClient";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::nfs_unmount() ".
                        "Unable to enable nfsClient service.");
        return $ClientUtil::FALSE;
    }

    # list all the current mounted nfs hosts
    $cmd = "/usr/sbin/esxcfg-nas --list";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::nfs_unmount() " .
                        "Could not list current NFS mount points.");
        return $ClientUtil::FALSE;
    }

    # chop off the trailing / on $nfs_mount_point
    chop($nfs_mount_point) if ($nfs_mount_point =~ /\/$/);

    # check if this nfs mount point is mounted.
    # @output will look like this:
    # lethe_share is /vol/rtpdist/ from lethe mounted
    # or this:
    # lethe_share is /vol/rtpdist from lethe mounted
    if (grep(/$nfs_mount_point is $nfs_dir_path\/? from $nfs_host mounted/, @output)) {

        # this nfs host is mounted, unmount the mount point
        $cmd = "/usr/sbin/esxcfg-nas --delete $nfs_mount_point";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::nfs_unmount() " .
                            "Could not list current NFS mount points.");
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Unmounted NFS volume $nfs_host\:$nfs_dir_path ".
                              "mounted on '$nfs_mount_point'");
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Error: ESXHost::nfs_unmount() " .
                        "'$nfs_host:$nfs_dir_path' is not mounted on '$nfs_mount_point'");
        return $ClientUtil::FALSE;
    }
}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    # Shutdown all virtual guests on this host
    $self->power_management(-action => "power_off_all");
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $cmd = "/sbin/shutdown -h now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->sanlog(-msg => "Rebooting the ESX host (" . $self->hostname() .
                          "). Powering off all VMs first.");

    # first power off all VMs cleanly
    $self->power_management(-action => "power_off_all");
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # reboot the esx host
    $self->sanlog(-msg => "Issuing the reboot to (" . $self->hostname() .
                          ") now.");
    $cmd = "touch /reconfigure;reboot";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # detect if the host actually rebooted or not
    $self->detect_reboot(-shutdown_timeout => 300,
                         -bootup_timeout   => 360,
                         -bootup_sleep     => 300);
    return $ClientUtil::FALSE if ($self->error);

    # sleep for a while to make sure the host is all the
    # way back up.  Sometimes ping will work before ssh
    # will work.
    $self->sanlog(-msg => "Sleeping for 120 seconds to allow the ESX Host to " .
                          "finish booting.");
    &SanUtil::nate_sleep(120);

    # reconnect to the client
    $self->reconnect();
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    $cmd = "uname -a";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::reboot() Command ($cmd) " .
                        "returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # power on all the vms that were shutdown from prior to the reboot
    my (@vms_that_were_shutdown) = $self->_get_vms_that_were_shutdown();
    my $vms_that_were_shutdown = join("\n", @vms_that_were_shutdown);
    $self->sanlog(-msg => "Successfully rebooted " . $self->hostname() .
                          ". Will now power on all VMs that were shutdown " .
                          "before the reboot.\n$vms_that_were_shutdown");

    $self->power_management(-action => "power_on",
                            -vm_list => \@vms_that_were_shutdown,
                            -dont_wait => 1);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Reboot is Complete!");
    $self->{reboot} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;
    my ($cmd, @output, $status, $version);

    $self->_clear_err();

    $cmd = "sanlun version";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::sanlun_version() Command ($cmd) " .
                        "returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]){
        $self->_set_err("Error: ESXHost::sanlun_version(): Unable to ".
                        "get filer name");
        return "";
    }

    $version = $output[0];
    $version =~ s/.*\sversion\s(\d+.\d+).*/$1/;

    return $version;
}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For ESXHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       Clean up any orphaned devices
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    # flush all the vmhba adapters
    $self->_flush_vmhba_adapters();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # load the claim rules into the system, this will remove any mask path rules
    # that were loaded previously
    if ($self->get_os_version() =~ /^4\./) {
        my $corestorage_ns = $self->{esxcli_corestorage_namespace};

        # scan through the claimrules and remove and claimrules that are of type file/location/MASK_PATH
        # this will get rid of any stale claimrules that may have been left on the system
        my $cmd = "esxcli ${corestorage_ns} claimrule list";
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                        -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Failed to list out claimrules");
            return $ClientUtil::FALSE;
        }

        my @rules_to_remove;
        foreach my $line (@output) {
            if ($line =~ /^MP\s+(\d+)\s+file\s+location\s+MASK_PATH/i) {
                push(@rules_to_remove, $1);
            }
        }

        foreach my $rule (@rules_to_remove) {
            $cmd = "esxcli ${corestorage_ns} claimrule delete --rule $rule";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not remove claim rule $rule");
                return $ClientUtil::FALSE;
            }
        }

        $cmd = "esxcli ${corestorage_ns} claimrule load";
        ($status) = $self->_esx_execute_cmd(-cmd => $cmd,
                                            -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Failed to load claimrules");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server, $ntp_time_zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: ESXHost::ntp_sync() Invalid value ".
                                "for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: ESXHost::ntp_sync() Invalid value ".
                                "for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # set the time zone
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");
    $cmd = "echo -e \"11\n" . $ntp_time_zone . "\n1\" | tzselect";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not set the timezone to '$ntp_time_zone'.");
        return $ClientUtil::FALSE;
    }

    # sync with the time server
    $self->sanlog(-msg => "Syncing the time with local NTP server: $ntp_server");

    my $new_ntpconf_file = "/etc/ntp.conf.tmp";
    my $new_steptickers_file = "/etc/ntp/step-tickers.tmp";

    # prepare the ntp conf file with the correct syntax
    $cmd = "echo restrict 127.0.0.1 > $new_ntpconf_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }
    $cmd = "echo restrict default kod nomodify notrap >> $new_ntpconf_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # add the time server to the conf file and step-tickers file
    $cmd = "echo server $ntp_server >> $new_ntpconf_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }
    $cmd = "echo $ntp_server > $new_steptickers_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }
    $cmd = "echo driftfile /var/lib/ntp/drift >> $new_ntpconf_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # copy over the new ntp.conf and step-tickers file.
    $cmd = "mv --backup --suffix=\".bak\" $new_ntpconf_file /etc/ntp.conf";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }
    $cmd = "mv --backup --suffix=\".bak\" $new_steptickers_file /etc/ntp/step-tickers";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # open up the firewall for the ntp service
    $cmd = "/usr/sbin/esxcfg-firewall -e ntpClient";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not enable ntpClient");
        return $ClientUtil::FALSE;
    }

    # restart the ntp service
    $cmd = "service ntpd restart";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not start the ntpd service");
        return $ClientUtil::FALSE;
    }

    # make it so ntpd starts automatically.
    $cmd = "chkconfig --level 345 ntpd on";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not set ntpd in level 3 4 and 5 in chkconfig");
        return $ClientUtil::FALSE;
    }

    # sync the hw clock to ntp time
    $cmd = "hwclock --systohc";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not sync the hwclock to NTP time.");
        return $ClientUtil::FALSE;
    }

    # display the date
    $cmd = "date";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::ntp_sync() ".
                        "Could not obtain the date.");
        return ($ClientUtil::FALSE);
    }

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to ESX syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @output, $status);

    # generate the timestamp that we will prepend to the messages
    my @months = qw( Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec );
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);
    $mon = $months[$mon];
    $min = "0" . $hour if (length($hour) == 1);
    $min = "0" . $min  if (length($min)  == 1);
    $sec = "0" . $sec  if (length($sec)  == 1);
    my $timestamp = "${mon} ${mday}  ${hour}:${min}:${sec} " . $self->hostname() . " NATE:";

    foreach my $msg (@message) {
        $cmd = "logger -t \"**NATE**\" -p user.error \"$msg\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 10);
        if ($status > 0) {
            $self->_set_err("Error: Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }

        $cmd = "echo \"$timestamp $msg\" >> /var/log/vmkernel";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 10);
        if ($status > 0) {
            $self->_set_err("Error: Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }

        $cmd = "echo \"$timestamp $msg\" >> /var/log/vmkwarning";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 10);
        if ($status > 0) {
            $self->_set_err("Error: Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Subroutine Name: vmfs_rescan()
# Objective:       This will rescan for new VMFS Volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub vmfs_rescan {
    my ($self) = @_;

    $self->sanlog(-msg => "Rescanning for VMFS Volumes.");

    # rescan the VMFS datastore list a number of times, because occasionaly
    # it fails to connect and displays the message "Failed to connect" with a status of 0
    for (my $i = 0; $i < 3; $i++) {
        my $cmd = "vimsh -n -e hostsvc/storage/vmfs_rescan 2&>1";
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::vmfs_rescan() ".
                            "Could not rescan for VMFS Volumes.");
            return $ClientUtil::FALSE;
        }
        if (grep(/Failed to connect/i, @output)) {
            # command failed, sleep a bit and try again
            &SanUtil::nate_sleep(20);
            next;
        } else {
            # command succeeded
            return $ClientUtil::TRUE;
        }
    }

    # command failed.
    $self->_set_err("Error: ESXHost::vmfs_rescan() ".
                    "Could not rescan for VMFS Volumes.");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: restart_mgmt_vmware_service()
# Objective:       This will restart the vmware service in order to refresh
#                  the state of the VMware Virtual Infrastructure Client.
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_mgmt_vmware_service {
    my ($self) = @_;

    $self->sanlog(-msg => "Refreshing mgmt-vmware service for " . $self->hostname());

    # restart the vmware service
    my $cmd = "service mgmt-vmware restart";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::restart_mgmt_vmware_service() ".
                        "Could not restart the 'mgmt-vmware' service");
        return $ClientUtil::FALSE;
    }

    # give ESX some time to restart.
    &SanUtil::nate_sleep(20);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_os_version
# Objective:       This will return the os revision and build number
################################################################################
sub get_os_version {
    my ($self) = @_;

    return $self->{os_version};
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($mac, $ipv4, $ipv6, $gateway, $subnet_mask, $ipv6_prefix_len);
    my ($set_interface, $mtusize, $broadcast, $iscsi_capable);

    foreach (keys %args) {
        if (/-?iscsi_capable/) {
            if (not length($iscsi_capable = $args{$_})) {
                $self->_set_err("Invalid value for -iscsi_capable");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $start = 0;
    my $finished = 0;
    my $interface = 0;
    my $interface_name = 'NA';
    my $ip_hash = {};

    # get all IP information from host via ifconfig
    my $cmd = "ifconfig -a";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    my $cnt = 0;
    my $num_lines = scalar(@output);
    foreach my $line (@output) {
        chomp($line);
        $cnt++;

        # Look for unique sections describing interface.
        if ($line =~ /^(\S+)\s+Link/ or $num_lines == $cnt) {
            if ($start == 1) {
                $finished = 1;
                $set_interface = $interface;
                $interface = $1;
            } else {
                $interface = $1;
            }
            $start = 1;
        }

        if ($start == 1 and $finished == 1) {
            # don't push in loopback
            if ($set_interface !~ /lo/) {
                $ip_hash->{$set_interface}->{interface_name} = $interface_name;
                $ip_hash->{$set_interface}->{mac} = $mac;
                $ip_hash->{$set_interface}->{description} = "NIC";
                $ip_hash->{$set_interface}->{gateway} = $gateway;
                $ip_hash->{$set_interface}->{broadcast} = $broadcast;
                $ip_hash->{$set_interface}->{subnet_mask} = $subnet_mask;
                $ip_hash->{$set_interface}->{mtusize} = $mtusize;
                $ip_hash->{$set_interface}->{ipv4} = $ipv4;
                $ip_hash->{$set_interface}->{ipv6} = $ipv6;

                # calculate the ipv4 subnet
                if ($ipv4 ne "" and $subnet_mask ne "") {
                    $ip_hash->{$set_interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                                        -ip_version => "ipv4",
                                                                                        -ipv4_netmask => $subnet_mask);
                } else {
                    $ip_hash->{$set_interface}->{ipv4_subnet} = "";
                }

                # calculate the ipv6 subnet
                if ($ipv6 ne "" and $ipv6_prefix_len ne "") {
                    $ip_hash->{$set_interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
                    $ip_hash->{$set_interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                                        -ip_version => "ipv6",
                                                                                        -ipv6_prefix_len => $ipv6_prefix_len);
                } else {
                    $ip_hash->{$set_interface}->{ipv6_prefix_len} = "";
                    $ip_hash->{$set_interface}->{ipv6_subnet} = "";
                }

                # if this is the management interface of the client
                # set an attribute
                my $host_rec = &Tharn::host($self->hostname());
                my $host_ip_default = $host_rec->{'default_ip'};
                if ($ipv4 eq $host_ip_default) {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 1;
                } else {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 0;
                }
            }

            $finished = 0;

            # initialize values
            $mac = $mtusize = $broadcast = $gateway = $subnet_mask = "";
            $ipv4 = $ipv6 = "";
            $ipv6_prefix_len = "";
        }

        # start pushing values into hash
        if ($start == 1) {
            # get IPv4 address
            if ($line =~ /inet addr:\s*(\S+)/) {
                $ipv4 = $1;
            }

            # get IPv6 address
            if ($line =~ /inet6 addr:\s*(\S+)\/(\d+)/) {
                $ipv6 = $1;
                $ipv6_prefix_len = $2;
            }

            # get MAC address
            if ($line =~ /HWaddr\s*(\S+)/i) {
                $mac = $1;
                $mac =~ s/\://g
            }

            # no gateway for UNIX
            $gateway = "";

            # get subnet mask
            if ($line =~ /Mask:\s*(\S+)/i) {
                $subnet_mask = $1;
            }

            # get broadcast
            if ($line =~ /Bcast:\s*(\S+)/i) {
                $broadcast = $1;
            }

            # get mtusize
            if ($line =~ /MTU:\s*(\S+)/i) {
                $mtusize = $1;
            }
        }
    }

    # get vmknic information and append to physical NICs
    $cmd = "esxcfg-vmknic -l";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ESXHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    # parse esxcfg-vmknic -l output:
    # ESX 4.0:
    # Interface  Port Group/DVPort   IP Family IP Address    Netmask        Broadcast       MAC Address       MTU     TSO MSS   Enabled Type
    # vmk0       vmk0                IPv4      192.168.10.1  255.255.255.0  192.168.10.255  00:50:56:77:32:28 9000    65535     true    STATIC
    # vmk1       VMkernel 2          IPv4      192.168.20.2  255.255.255.0  192.168.20.255  00:50:56:79:ee:2e 9000    65535     true    STATIC
    # vmk0       SAN VMkernel        IPv6      fe80::250:56ff:fe77:e14      64              00:50:56:77:0e:14 1500    65535     true    STATIC
    # ESX 3.5:
    # vmk0       VMkernel            192.168.30.32   255.255.255.0   192.168.30.255  00:50:56:73:05:d0 1500    40960     true

    foreach my $line (@output) {
        next if ($line =~ /Interface/);

        # remove justified whitespace to assist parsing.
        $line =~ s/^\s+//;

        # do not need anything before IPv? type and actually
        # get the interface
        $line =~ m/^(\S+)\s+/;
        $set_interface = $1;

        # ESX 4.x has IPV6 and hence column for this, ESX3.5 does not
        my $type;
        if ($self->get_os_version() =~ /^4\./) {
            # will effect parsing because of alias of port group
            # could contain multiple words.
            $line =~ s/^.*(IPv\d)\s+//;
            $type = $1;
        } else {
            # remove up first IPv4 address and then paste back
            # IPv4 address to get full line minus initial interface and
            # port group (which could be multiple words).
            $line =~ s/^.*?(\d+\.\d+\.\d+\.\d+)\s+//;
            $line = $1 . " " . $line;
            $type = "IPv4";
        }

        # Now, we can split by whitespace
        my @info = split(/\s+/, $line);

        # skip interfaces that don't have IP addrs, or netmasks
        next if ($info[0] =~ /N\/A/i);
        next if ($info[1] =~ /N\/A/i);

        # Store in already existing hash
        # store based on IPv4/6 (no broadcast with IPv6 and
        # parsing is different (because of that)
        if ($type =~ /v6/i) {
            $ip_hash->{$set_interface}->{ipv6} = $info[0];
            $ip_hash->{$set_interface}->{mac} = $info[2];
            $ip_hash->{$set_interface}->{description} = "vmkNIC";
            $ip_hash->{$set_interface}->{gateway} = "NA";
            $ip_hash->{$set_interface}->{ipv6_prefix_len} = $info[1];
            $ip_hash->{$set_interface}->{mtusize} = $info[3];
            $ip_hash->{$set_interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $info[0],
                                                                                -ip_version => "ipv6",
                                                                                -ipv6_prefix_len => $info[1]);
        } else {
            $ip_hash->{$set_interface}->{ipv4} = $info[0];
            $ip_hash->{$set_interface}->{mac} = $info[3];
            $ip_hash->{$set_interface}->{description} = "vmkNIC";
            $ip_hash->{$set_interface}->{gateway} = "NA";
            $ip_hash->{$set_interface}->{broadcast} = $info[2];
            $ip_hash->{$set_interface}->{subnet_mask} = $info[1];
            $ip_hash->{$set_interface}->{mtusize} = $info[4];
            $ip_hash->{$set_interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $info[0],
                                                                                -ip_version => "ipv4",
                                                                                -ipv4_netmask => $info[1]);
        }

        # indicate that this interface is iscsi capable
        $ip_hash->{$set_interface}->{iscsi_capable} = 1;

        # if this is the management interface of the client
        # set an attribute
        my $host_rec = &Tharn::host($self->hostname());
        my $host_ip_default = $host_rec->{'default_ip'};
        if ($info[0] eq $host_ip_default) {
            $ip_hash->{$set_interface}->{is_mgmt_interface} = 1;
        } else {
            $ip_hash->{$set_interface}->{is_mgmt_interface} = 0;
        }
    }

    # get hardware iscsi adapter interface data as well
    # for qlogic hw iscsi HBAs
    if (scalar(@{$self->{vmhba_hwiscsi_adapters}})) {
        foreach my $vmhba_hwiscsi_adapter (@{$self->{vmhba_hwiscsi_adapters}}) {

            # skip this section if this is a broadcom iSCSI hw adapter
            # it should have been covered by the section above
            my @raw_vmhba_list = $self->get_cache_data("vmhba_raw_output");
            next if (grep(/^$vmhba_hwiscsi_adapter .*?Broadcom iSCSI/i, @raw_vmhba_list));

            # get ip information
            #[root@vmshe-ibm3550-02 ~]# vmkiscsi-tool vmhba1 -i
            #IP Address   : 192.168.9.50
            #[root@vmshe-ibm3550-02 ~]# vmkiscsi-tool vmhba2 -i
            #IP Address   : 192.168.8.50
            #IP Address   : fe80::2a0:98ff:fe02:ff22
            #IP Address   : fc00:cccc:bbbb:1232:1::
            #IP Address   : ::
            $cmd = "vmkiscsi-tool $vmhba_hwiscsi_adapter -i";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return {};
            }

            # grep ipv4 information
            my @ipv4_output = grep { /IP Address\s+:\s+\d+\.\d+\.\d+\.\d+/i } @output;

            # grep ipv6 information
            my @ipv6_output = grep { /IP Address\s+:\s+.*:/i } @output;

            # get subnet information
            #[root@vmshe-ibm3550-02 tmp]# vmkiscsi-tool vmhba1 -s
            #SUBNET MASK SETTABLE                    : 1
            #SUBNET MASK VALID                       : 1
            #SUBNET MASK                             : 255.255.255.0
            $cmd = "vmkiscsi-tool $vmhba_hwiscsi_adapter -s";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return {};
            }

            # parse out the subnet line
            my @subnet_mask = grep { /SUBNET MASK\s+:\s+\d+\.\d+\.\d+\.\d+/i } @output;

            # store all the values on the ip hash for this hw iscsi interface
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4})          = ($ipv4_output[0] =~ /IP Address\s*:\s*(\d+\.\d+\.\d+\.\d+)/);
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask})   = ($subnet_mask[0] =~ /SUBNET MASK\s*:\s*(\d+\.\d+\.\d+\.\d+)/);
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6})          = ($ipv6_output[0] =~ /IP Address\s*:\s*(\S+)/);
            $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_prefix_len} = "64"; # hack 64 is the most common prefix len for ipv6 addresses
            $ip_hash->{$vmhba_hwiscsi_adapter}->{broadcast}       = "NA";

            # generate the broadcast ip for ipv4 and store it
            if (defined($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4})) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{broadcast} =
                    &SanUtil::get_ip_broadcast(-ip_address => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4},
                                               -ipv4_netmask => $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask});
            }

            # indicate that this adapter is iscsi capable
            $ip_hash->{$vmhba_hwiscsi_adapter}->{iscsi_capable} = 1;

            # get the mac/gateway/mtu
            $cmd = "vimsh -n -e hostsvc/storage/hba_info | grep -B1 -A10 \\\"" . $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4} . "\\\",";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return {};
            }

            # OUTPUT:
            #mac = "00:c0:dd:0b:03:ed",
            #address = "192.168.8.50",
            #dhcpConfigurationEnabled = false,
            #subnetMask = "255.255.255.0",
            #defaultGateway = "192.168.8.1",
            #primaryDnsServerAddress = "0.0.0.0",
            #alternateDnsServerAddress = "0.0.0.0",
            #ipv6Address = <unset>,
            #ipv6SubnetMask = <unset>,
            #ipv6DefaultGateway = <unset>,
            #arpRedirectEnabled = false,
            #mtu = 1500,

            # parse out the mac address
            my @mac_address = grep { /mac\s+=\s+\"\S+?\"/i } @output;

            # parse out the gateway
            my @default_gateway = grep { /defaultGateway\s+=\s+\"\S+?\"/i } @output;

            # parse out the mtu size
            my @mtusize = grep { /mtu\s+=\s+\d+/i } @output;

            # store these values in the ip hash
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{mac})       = ($mac_address[0] =~ /mac\s+=\s+\"(\S+?)\"\,/i);
            $ip_hash->{$vmhba_hwiscsi_adapter}->{description} = "HW iSCSI HBA";
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{gateway})   = ($default_gateway[0] =~ /defaultGateway\s+=\s+\"(\S+?)\"\,/i);
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{mtusize})   = ($mtusize[0] =~ /mtu\s+=\s+(\d+)\,/i);

            # update the subnet values
            if (defined($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4})) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4_subnet} =
                    &SanUtil::get_ip_subnet(-ip_address   => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4},
                                            -ip_version   => "ipv4",
                                            -ipv4_netmask => $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask});
            }
            if (defined($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6})) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_subnet} =
                    &SanUtil::get_ip_subnet(-ip_address      => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6},
                                            -ip_version      => "ipv6",
                                            -ipv6_prefix_len => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_prefix_len});
            }
        }
    }

    # if requesting only iscsi capable interfaces, filter them out here
    if (defined($iscsi_capable) and $iscsi_capable == 1) {
        my $tmp_ip_hash;
        foreach my $int (keys %{$ip_hash}) {
            if ($ip_hash->{$int}->{iscsi_capable} == 1) {
                # copy this interface's data into the tmp ip hash
                my %data = %{$ip_hash->{$int}};
                $tmp_ip_hash->{$int} = \%data;
            }
        }

        # reassign the new ip hash
        $ip_hash = $tmp_ip_hash;
    }

    # return info hash
    return ($ip_hash);
}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    return "NATIVEMP";
}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for ESX
################################################################################
sub rotate_logs {
    my ($self) = @_;

    # rotate the vmkernel log files for esx
    my $cmd = "/usr/sbin/logrotate -f /etc/logrotate.d/vmkernel";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_additional_info()
# Objective:       dump additional OS specific information
################################################################################
sub display_additional_info {
    my ($self) = @_;

    my $os_version = $self->get_os_version();

    my ($cmd, $status, @output);

    # executing these commands for additional logging information
    my $esxcli_nmp_namespace = $self->{esxcli_nmp_namespace};

    if ($os_version =~/^4\./) {
        $cmd = "esxupdate query -v";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 240);
        $cmd = "esxupdate info";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 240);
        $cmd = "esxcli ${esxcli_nmp_namespace} satp  listrules | grep NETAPP";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 240);
    }

    $cmd = "esxcli ${esxcli_nmp_namespace} device list";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # get ESX syetem information to use for audit purposes, do not validate
    # results.
    $cmd = "esxcfg-info -y";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -timeout => 240);

    # get information on the physical system in use (platform/CPU/memory)
    $cmd = "dmidecode";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -timeout => 240);

    $self->sanlog(-msg => "Running multiple switch commands. " .
                          "Ignore failures.");


    # Brocade Broadcom Intel Emulex Qlogic
    $cmd = "cat /proc/scsi/bfa/*; " .
           "cat /proc/scsi/bnx2fc/*; " .
           "cat /proc/scsi/fcoe/*; " .
           "cat /proc/scsi/lpfc*/*; " .
           "cat /proc/scsi/qla*/*";
    ($status, @output) = $self->_esxi_direct_execute(-cmd => $cmd,
                                                     -timeout => 240);
    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _fcp_configure()
# Objective:       Configure FCP
# Details:         See POD documentation at the end of this file
################################################################################
sub _fcp_configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "FCP Configure: Configuring FCP");

    $self->sanlog(-msg => "FCP Configure: Complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsi to include the targets for the filers.
#                  the IP's of all the ports on each filer.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets);

    # validate that iscsi parameters before continuing.
    $self->_iscsi_configure_validate_params();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # if the user does not want to configure iscsi, then leave now.
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} != 1) {
        $self->sanlog(-msg => "iSCSI Configure: Bypassing iSCSI Configuration by user request (CONFIG_ISCSI_DISCOVERY=0)");
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
        if (/^-?iscsi_targets$/) {
            @targets = @{$args{$_}};
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI");

    # if requested to do subnet matching, filter the targets list
    # so that it contains targets that are accessible by this host
    @targets = $self->_filter_targets_by_subnet(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check what ip version is being requested and set the ip mode on all the target objects
    if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /^ipv4$|^ipv6$|^auto$/i) {
        foreach my $target (@targets) {
            # set the ip mode to be either ipv4 or ipv6
            $target->set_ip_mode($self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
            if ($target->error()) {
                $self->_set_err("Could not set the ip mode of target $target->{nodename} to " .
                                 $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n" .
                                "Reason: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    } else {
       $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: " .
                        $self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
       return $ClientUtil::FALSE;
    }

    # print out a summary of what we are going to do for iSCSI Configure
    $self->_iscsi_configure_print_summary(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure iscsi if requested to do so before configuring
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} > 0) {
        $self->_iscsi_deconfigure(-targets => \@targets,
                                  -clean_level => $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # configure each iSCSI vmhba adapter
    foreach my $iscsi_adapter (@{$self->{vmhba_iscsi_adapters}}) {

        # if we are doing sw iscsi, enable the software initiator.
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^software$/i) {
            &SanUtil::nate_sleep(1);
            # check if the sw iscsi initiator is enabled
            $cmd = "/usr/sbin/esxcfg-swiscsi --query";
            ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could not query the software iscsi service");
                return $ClientUtil::FALSE;
            }

            if (grep { /Software iSCSI is not enabled/i } @output) {
                # then enable the iscsi software initiator
                $cmd = "/usr/sbin/esxcfg-swiscsi --enable";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                if ($status > 0) {
                    $self->_set_err("Could not enable the software iscsi service");
                    return $ClientUtil::FALSE;
                }

                # we may need to reboot the esx machine after enabling the iscsi initiator
                if (grep(/reboot/i, @output)) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                }
            }

            # check if the swISCSIClient service is enabled.
            $cmd = "/usr/sbin/esxcfg-firewall --query swISCSIClient";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                # if the service is blocked, then enable it
                if (grep { /Service swISCSIClient is blocked/i } @output) {
                    $cmd = "/usr/sbin/esxcfg-firewall --enableService swISCSIClient";
                    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                    if ($status > 0) {
                        $self->_set_err("Could not enable the swISCSIClient service");
                        return $ClientUtil::FALSE;
                    }
                } else {
                    # unknown error
                    $self->_set_err("Could not query the swISCSIClient service");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # configure the iSCSI option stacking features
        $self->_iscsi_configure_options(-iscsi_adapter => $iscsi_adapter);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # configure for sendtargets or static discovery depending on what was passed in.
        if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^sendtargets$/i) {
            $self->_iscsi_configure_sendtargets_discovery(-iscsi_adapter => $iscsi_adapter,
                                                          -targets => \@targets);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^static$/i) {
            $self->_iscsi_configure_static_discovery(-iscsi_adapter => $iscsi_adapter,
                                                     -targets => \@targets);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid ISCSI Discovery type '" . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "'");
            return $ClientUtil::FALSE;
        }

    }

    $self->sanlog(-msg => "iSCSI Configure: Complete.");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_validate_params()
# Objective:       Validate that the iscsi params are set properly.
################################################################################
sub _iscsi_configure_validate_params {
    my ($self) = @_;

    # check for unsupported iscsi type
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /^software$|^hardware$/i) {
        $self->_set_err("Unsupported iSCSI type '" . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "'");
        return $ClientUtil::FALSE;
    }

    # check for unsupported options
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} !~ /^(unidirectional|bidirectional|none|\s*)$/i) {
        $self->_set_err("Unsupported CHAP Mode: " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE});
        return $ClientUtil::FALSE;
    }


    # validate incompatibilities with esx3.x
    if ($self->get_os_version() !~ /(^4|5|6)\./) {
        # only sendtargets is supported
        if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} !~ /^sendtargets$/i) {
            $self->_set_err("Unsupported iSCSI Discovery Type: " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE});
            return $ClientUtil::FALSE;
        }
        # header and data digests are not supported
        if ($self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} == 1) {
            $self->_set_err("Header digest is an unsupported feature on this ESX version");
            return $ClientUtil::FALSE;
        }
        if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1) {
            $self->_set_err("Data digest is an unsupported feature on this ESX version");
            return $ClientUtil::FALSE;
        }

    }

    # validate that bidirectional chap is not being used with ESX version before 5.0
    if ($self->get_os_version() !~ /^(5|6)\./ and $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
        $self->_set_err("Bidirectional iSCSI CHAP Mode is only valid with ESXi 5.0 and greater");
        return $ClientUtil::FALSE;
    }

    # isns is not supported on ESX
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
        $self->_set_err("Unsupported iSCSI Discovery Type: " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE});
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_print_summary()
# Objective:       print out a summary of what configuration will be done
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_configure_print_summary {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my @target_ips = map { "\n     " . $_->{ip} } @targets;
    my @config_summary;

    push(@config_summary, "iSCSI Configure: Configuring iSCSI with the following options:\n");
    push(@config_summary, "Targets : @target_ips\n");
    push(@config_summary, "IP Version            : " . $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n");
    push(@config_summary, "iSCSI Type            : " . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "\n");
    push(@config_summary, "iSCSI Discovery Type  : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "\n");
    push(@config_summary, "iSCSI Pre-Clean Level : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} . "\n");
    push(@config_summary, "iSCSI Subnet Matching : " . $self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} . "\n");
    push(@config_summary, "CHAP Mode             : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} . "\n");

    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /unidirectional/i) {
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} ne "") {
            push(@config_summary, "CHAP Username         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} . "\n");
        } else {
            # CHAP username defaults to the iqn of the initiator if its empty
            push(@config_summary, "CHAP Username         : (Defaulted to initiator nodename)\n");
        }

        push(@config_summary, "CHAP Password         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD} . "\n");
    }

    push(@config_summary, "Header Digest         : " . $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} . "\n");
    push(@config_summary, "Data Digest           : " . $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} . "\n");

    $self->sanlog(-msg => "@config_summary");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigures iscsi
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            @targets = @{$args{$_}};
            next;
        } elsif (/^-?clean_level$/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI (clean_level=$clean_level)");

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    # check for unsupported iscsi type
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /^software$|^hardware$/i) {
        $self->_set_err("Unsupported iscsi type '" . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "'");
        return $ClientUtil::FALSE;
    }

    # deconfigure each adapter.
    foreach my $adapter (@{$self->{vmhba_iscsi_adapters}}) {
        # if the clean level is set to 2, then deconfigure both sendtargets & static discovery
        # otherwise, deconfigure each individually.
        if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /sendtargets/i or $clean_level > 0) {
            $self->_iscsi_deconfigure_sendtargets_discovery(-targets => \@targets,
                                                            -iscsi_adapter => $adapter,
                                                            -clean_level => $clean_level);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        }
        if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /static/i or $clean_level > 0) {
            $self->_iscsi_deconfigure_static_discovery(-targets => \@targets,
                                                       -iscsi_adapter => $adapter,
                                                       -clean_level => $clean_level);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        }

        # deconfigure the iscsi options
        $self->_iscsi_deconfigure_options(-iscsi_adapter => $adapter);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: iSCSI has been deconfigured.");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_options()
# Objective:       Set up iSCSI Option stacking features.
################################################################################
sub _iscsi_configure_options {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, $iscsi_adapter);

    # parse arguments
    foreach (keys %args) {
        if (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI options.");

    # set the iscsi alias for this adapter to be the hostname, command may not work in esx3.x
    # dont fatal out if it doesnt work.
    my $hostname = $self->hostname();
    $cmd = "vimsh -n -e \"hostsvc/storage/iscsi_set_alias $iscsi_adapter ${hostname}_${iscsi_adapter}\" 2&>1";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);

    # enable CHAP if requested
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /unidirectional/i) {
        # obtain the CHAP Username & CHAP Password
        my $chap_username;
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} ne "") {
            $chap_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
        } else {
            # default the chap username to the initiator nodename of this adapter
            $chap_username = $self->_get_iscsi_node_name_by_adapter($iscsi_adapter);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        }
        my $chap_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD};

        # enable uni-directional CHAP
        $cmd = "vimsh -n -e \"hostsvc/storage/iscsi_enable_chap $iscsi_adapter $chap_username $chap_password\" 2&>1";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Could not enable CHAP on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    # set HeaderDigest (esx4 only)
    if ($self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} == 1) {
        if ($self->get_os_version() =~ /^4\./) {
            # emable header digest
            $cmd = "vmkiscsi-tool $iscsi_adapter -W -a \"header_digest=4\"";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Could not enable Header Digest on the iscsi adapter '$iscsi_adapter'");
                return $ClientUtil::FALSE;
            }
        }
    }

    # set up Data Digest (esx4 only)
    if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1) {
        if ($self->get_os_version() =~ /^4\./) {
            # enable data digest
            $cmd = "vmkiscsi-tool $iscsi_adapter -W -a \"data_digest=4\"";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Could not enable Data Digest on the iscsi adapter '$iscsi_adapter'");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_options()
# Objective:       Deconfigure iSCSI Option stacking features.
################################################################################
sub _iscsi_deconfigure_options {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, $iscsi_adapter);

    # parse arguments
    foreach (keys %args) {
        if (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI options.");

    # disable CHAP
    $cmd = "vimsh -n -e \"hostsvc/storage/iscsi_disable_chap $iscsi_adapter\" 2&>1";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Could not disable CHAP on the iscsi adapter '$iscsi_adapter'");
        return $ClientUtil::FALSE;
    }

    # disable Header/Data Digest (esx4 only)
    if ($self->get_os_version() =~ /^4\./) {
        # disable header digest
        $cmd = "vmkiscsi-tool $iscsi_adapter -W -a \"header_digest=1\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0 and not grep(/Reset needed/i, @output)) {
            $self->_set_err("Could not disable Header Digest on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }

        # disable data digest
        $cmd = "vmkiscsi-tool $iscsi_adapter -W -a \"data_digest=1\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0 and not grep(/Reset needed/i, @output)) {
            $self->_set_err("Could not disable Data Digest on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_static_discovery()
# Objective:       Set up static discovery for the specified targets.
################################################################################
sub _iscsi_configure_static_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Static' discovery.");

    # static discovery is only supported on esx4
    if ($self->get_os_version() !~ /^4\./) {
        $self->_set_err("Static iSCSI Discovery is not supported on this ESX version");
        return $ClientUtil::FALSE;
    }

    # list the static iscsi targets that are currently added
    $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -S -l";
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd);
    # not cecking for error because this command can do the
    # right thing but still have a bad return status.

    # add a target address for each filer ip in @targets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # check if the target already exists.
        if (grep {/^\s*ADDRESS\s*?:\s*($ip)(:3260)?/i } @list) {
            # then this target already exists, skip it
            $self->sanlog(-msg => "iSCSI Configure: Skipping static target '$ip $nodename', it is already added.");
            next;
        }

        $self->sanlog(-msg => "iSCSI Configure: Adding Static Target '$ip $nodename'");

        # add the target ip (this command can take some time...)
        $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -S -a \"${ip}:3260 ${nodename}\"";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status != 0) {
            $self->_set_err("Could not add the ip '$ip' as a static target ".
                            "on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_static_discovery()
# Objective:       Deconfigure static discovery for the specified targets.
################################################################################
sub _iscsi_deconfigure_static_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);
    my $clean_level = 1;

    # static discovery is only supported on esx4, don't set error for non 4.x
    if ($self->get_os_version() !~ /^4\./) {
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Static' discovery.");

    # list the static iscsi targets that are currently added
    $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -S -l";
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd);
    # not cecking for error because this command can do the
    # right thing but still have a bad return status.

    # if told to clean up everything then remove all targets that are in the static list
    if ($clean_level == 2) {
        my $target_iqn;

        foreach my $line (@list) {
            if ($line =~ /\bNAME\s*?:\s*(\S+)/i) {
                $target_iqn = $1;
                next;
            }

            if ($line =~ /(?:ADDRESS|HOSTNAME)\s*?:\s*(\S+)(:3260)?/i) {
                my $ip = $1;

                # concat the port to the ip if its not already there
                $ip .= ":3260" if ($ip !~ /:3260$/);

                # remove the static target ip (this command can take some time...)
                $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -S -r \"$ip $target_iqn\"";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                if ($status > 0) {
                    $self->_set_err("Could not remove the ip '$ip' from the static discovery target list ".
                                    "on the iscsi adapter '$iscsi_adapter'");
                    return $ClientUtil::FALSE;
                }
            }
        }

    } elsif ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # remove a target address for each filer ip in @targets
        foreach my $target (@targets) {
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
            my $target_iqn = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # check if the static target exists
            if (not grep {/^\s*ADDRESS\s*?:\s*($ip)(:3260)?/i } @list) {
                # then this target does not exist, skip it
                next;
            }

            # if the static target ip exists, make sure that the target iqn also exists
            if (not grep {/^\s*NAME\s*?:\s*${target_iqn}/i} @list) {
                # if it doesnt exist, then this is a case where there is a duplicate target ip address that is using the
                # a different target nodename. Error out in this case
                $self->_set_err("Detected a duplicate static target using the IP address '${ip}' that does " .
                                "not belong to the target '${target_iqn}'. Please correct this manually.");
                return $ClientUtil::FALSE;
            }

            # remove the static target ip (this command can take some time...)
            $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -S -r \"${ip}:3260 $target_iqn\"";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Could not remove the ip '$ip' from the static discovery target list ".
                                "on the iscsi adapter '$iscsi_adapter'");
                return $ClientUtil::FALSE;
            }
        }
    } else {
        $self->_set_err("Invalid clean_level: '$clean_level'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_sendtargets_discovery()
# Objective:       Set up sendtargets discovery for the specified targets.
################################################################################
sub _iscsi_configure_sendtargets_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Sendtargets' Discovery");

    # list the iscsi targets that are currently added
    $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -D -l";
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd);
    # not cecking for error because this command can do the
    # right thing but still have a bad return status.

    # add a target address for each filer ip in @targets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # check if the target already exists.
        if (grep {/^\s*DISCOVERY ADDRESS\s*?:\s*($ip)(:3260)?/i } @list) {
            # then this target already exists, skip it
            $self->sanlog(-msg => "iSCSI Configure: Skipping sendtargets target '$ip $nodename', it is already added.");
            next;
        }

        $self->sanlog(-msg => "iSCSI Configure: Added SendTargets Target '$ip $nodename'");
        # add the target ip (this command can take some time...)
        $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -D -a ${ip}:3260";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status != 0) {
            $self->_set_err("Could not add the ip '$ip' as a sendtargets discovery target ".
                            "on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_sendtargets_discovery()
# Objective:       Deconfigure sendtargets discovery for the specified targets.
################################################################################
sub _iscsi_deconfigure_sendtargets_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);
    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required args
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Sendtargets' Discovery.");

    # list the iscsi targets that are currently added
    $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -D -l";
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd);
    # not cecking for error because this command can do the
    # right thing but still have a bad return status.

    # if told to clean up everything then remove all targets that are in the sendtargets list
    if ($clean_level == 2) {
        foreach my $line (@list) {
            if ($line =~ /^\s*DISCOVERY ADDRESS\s*?:\s*(\S+)(:3260)?/i) {
                my $ip = $1;

                # concat the port if its not there already
                $ip .= ":3260" if ($ip !~ /:3260/);

                # remove the target ip
                $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -D -r $ip";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                # only checking for error on esx4
                if ($self->get_os_version() =~ /^4\./ and $status != 0) {
                    $self->_set_err("Failed to remove $ip from sendtargets list'");
                    return $ClientUtil::FALSE;
                }
            }
        }
    } elsif ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # remove each target address for each filer ip in @targets
        foreach my $target (@targets) {
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # check if the target already exists.
            if (not grep {/^\s*DISCOVERY ADDRESS\s*?:\s*($ip)(:3260)?/i } @list) {
                # then this target does not exist, skip it
                next;
            }

            # remove the target ip (this command can take some time...)
            # only checking for error on esx4
            $cmd = "/usr/sbin/vmkiscsi-tool $iscsi_adapter -D -r ${ip}:3260";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($self->get_os_version() =~ /^4\./ and $status != 0) {
                $self->_set_err("Failed to remove $ip from sendtargets list'");
                return $ClientUtil::FALSE;
            }
        }
    } else {
        $self->_set_err("Invalid clean_level: '$clean_level'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _vmware_cmd()
# Objective:       whenever we perform a vmware-cmd on the ESX host, things
#                  can go wrong so this is a wrapper around &ClientUtil::execute
# Details:         See POD documentation at the end of this file
################################################################################
sub _vmware_cmd {
    my ($self, $cmd) = @_;

    my ($status, @output);

    $self->_clear_err();

    # try running $cmd
    for(my $i = 0; $i < 3; $i++) {
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0 or grep(/vim\.fault\.NotFound|Operation cannot be performed/i, @output)) {
            # if the command failed, then try to recover
            if ($self->_recover_from_missing_vm_error(@output) ==
                    $ClientUtil::FALSE) {
                # if the recovery failed, then game over
                $self->_set_err("Error: ESXHost::_vmware_cmd() " .
                                "Command ($cmd) failed. Status = ($status)");
                return ($status, @output);
            } else {
                # otherwise try running $cmd again
                next;
            }
        } else {
            # $cmd succeeded, return
            return ($status, @output);
        }
    }

    # at this point both attempts to run $cmd failed.
    $self->_set_err("Error: ESXHost::_vmware_cmd() " .
                    "Command ($cmd) failed. Status = ($status)");

    return ($status, @output);
}

################################################################################
# Subroutine Name: _recover_from_missing_vm_error()
# Objective:       Sometimes ESX loses track of its vm's and needs it's
#                  management service reloaded.
# Details:         See POD documentation at the end of this file
################################################################################
sub _recover_from_missing_vm_error {
    my ($self, @output) = @_;

    $self->_clear_err();

    # at this point, invalidate the cache for vmware_cmd_l because its data
    # may be innacurate due to this method being called.
    $self->invalidate_items_cache_data("vmware_cmd_l");

    # VNControl error: No Such virtual machine
    # VMControl error -999: Unknown error: SoapError: ServerFaultCode(1604) :
    # (Operation failed since another task is in progress.)
    # VMControl error -10: The specified property or variable is invalid or undefined: Unable to retrieve config state
    # VMControl error -3: Invalid arguments: Login failed.
    # vim.fault.NotFound: No virtual machine found with name:
    #
    # the easiest way to fix these errors is to restart the vmware service and sleep.
    if (grep {/No such virtual machine/i} @output or
        grep {/another task is in progress/i} @output or
        grep {/The specified property or variable is invalid or undefined/i} @output or
        grep {/Invalid arguments: Login failed/i} @output or
        grep {/No virtual machine found with name:/i} @output or
        grep {/retrieve local ticket from session manager failed/i} @output) {
        $self->sanlog(-msg => "A VM was missing, will try to recover it by ".
                              "restarting the mgmt-vmware service. " .
                              "\nThis will cause VMware I.C. to crash, ".
                              "so you will have to log in again.");

        my $cmd = "service mgmt-vmware restart";
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::_recover_from_missing_vm_error() ".
                            "Could not restart the 'mgmt-vmware' service");
            return $ClientUtil::FALSE;
        }

        # give ESX sometime to get the service back online
        &SanUtil::nate_sleep(120);
        return $ClientUtil::TRUE;
    }

    # VMControl error -3: Invalid arguments: Failed to connect or login with server.
    # Solution: sleep an additional 30 seconds.
    if (grep {/Failed to connect or login with server/i} @output) {
        $self->sanlog(-msg => "Giving ESX some time to load... will try again in 30 seconds");
        &SanUtil::nate_sleep(30);
        return $ClientUtil::TRUE;
    }

    # VMControl error -999: Unknown error: The poweron operation failed unexpectedly
    # the only reason this could happen is if:
    # (1) the vmx file is corrupted
    # (2) the boot vmdk file for the vm is messed up.
    # regardless of the cause, automation can't handle it.
    if (grep {/The poweron operation failed unexpectedly/i} @output){
        $self->sanlog(-msg => "Error: ESXHost::_recover_from_missing_vm_error() ".
                              "1) Verify that the .vmx configuration files of ".
                              "your VMs are not corrupted OR 2) a boot disk on ".
                              "a VM may be corrupted and may need to be re-".
                              "configured.");
        return $ClientUtil::FALSE;
    }

    # otherwise its an unknown error
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: _vmkfstools_cmd
# Objective:       whenever we perform a vmkfstools cmd on the ESX host, things
#                  can go wrong so this is a wrapper around &ClientUtil::execute
# Details:         See POD documentation at the end of this file
################################################################################
sub _vmkfstools_cmd {
    my ($self, $cmd) = @_;

    my ($status, @output);

    $self->_clear_err();

    for(my $i = 0; $i < 3; $i++) {
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1,
                                                     -reconnect_on_failure => 1);
        if ($status > 0) {
            if (grep {/Error: Device or resource busy/i} @output) {
                # sleep sometime and try again.
                &SanUtil::nate_sleep(5);
                next;
            } else {
                # unknown error, probably cant be fixed. Set error.
                $self->_set_err("Error: ESXHost::_vmkfstools_cmd() " .
                                "Command ($cmd) failed. Status = ($status)");
                return ($status, @output);
            }
        } else {
            # the command worked, return
            return ($status, @output);
        }
    }

    # at this point the command probably wont work. Set error.
    $self->_set_err("Error: ESXHost::_vmkfstools_cmd() " .
                    "Command ($cmd) failed. Status = ($status)");
    return ($status, @output);
}

################################################################################
# Subroutine Name: _get_vm_config_file()
# Objective:       Returns the filepath to the vm's config file
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_vm_config_file {
    my ($self, %args) = @_;

    my ($vm_name, @this_vm);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not $vm_name = $args{$_}) {
                $self->_set_err("Error: ESXHost::_get_vm_config_file() " .
                                "Invalid value for argument " .
                                "'-vm_name' => $vm_name");
                return "";
            }
            next;
        }
    }

    # check for required parameters
    if (not defined($vm_name)) {
        $self->_set_err("Error: ESXHost::_get_vm_config_file() ".
                        "Missing required parameter '-vm_name'");
        return "";
    }

    my $vm_is_still_missing = 0;

    while (1) {
        # grab all vms on this host
        my @all_vms = sort($self->_get_all_vm_config_files());
        if ($self->error()) {
            return "";
        }

        # grep for $vm_name in @all_vms
        # this line is very dependent on the fact that the dns name of a VM
        # must match the VM name in ESX.
        # We are expected the line to look like
        # "/vmfs/volumes/.../.../...$vm_name.vmx"
        if ($vm_name =~ /\.vmx\s*$/) {
            @this_vm = grep { /$vm_name/ } @all_vms;
        } else {
            @this_vm = grep { /$vm_name\.vmx/ } @all_vms;
        }

        # if the result was empty, try to recover the vm, and try again
        # if that still didnt work, then set an error
        if (scalar(@this_vm) == 0) {
            if ($vm_is_still_missing == 1) {
                $self->_set_err("Error: ESXHost::_get_vm_config_file() ".
                                "$vm_name does not exist or it's config file ".
                                "is corrupt. Restore it's .vmx file from a ".
                                "backup.");
                return "";
            }
            $self->_recover_from_missing_vm_error("No such virtual machine");
            $vm_is_still_missing = 1;
            next;
        }
        last;
    }

    # at this point we are ok,
    # return the vm config filepath that was found
    return $this_vm[0];
}

################################################################################
# Subroutine Name: _get_all_vm_config_files()
# Objective:       Returns an array of Guest OS .vmx configuration files
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_all_vm_config_files {
    my ($self) = @_;

    my ($status, @output);

    # check if the vm config file list cache is available.
    if ($self->is_cache_data_valid("vmware_cmd_l")) {
        @output = $self->get_cache_data("vmware_cmd_l");
        return (wantarray()) ? @output : \@output;
    }

    # list all the vm's on this esx host.
    for (my $i = 0; $i < 3; $i++) {

        $self->_clear_err();

        my $cmd = "/usr/bin/vmware-cmd -l";
        ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            # if we got an error saying we couldnt connect, sleep 20 seconds to
            # give the mgmt-vmware service time to reload.
            if (grep {/Could not connect to server/i} @output) {
                &SanUtil::nate_sleep(20);
                next;
            } else {
                return (wantarray()) ? () : [];
            }
        }

        # if the output has [] in front of every vmx file path then we
        # have to recover from a missing vm error, this will fix it
        if (grep { /^\s*\[\]\s+/ } @output) {
            $self->_recover_from_missing_vm_error("No such virtual machine");
            if ($self->error()) {
                return (wantarray()) ? () : [];
            }
        }

        # Sometimes the output of vmware-cmd -l gets corrupted by this string:
        # "pid file /var/run/vmkiscsid.pid already locked"
        # Remove it from any line that contains the string
        foreach my $line (@output) {
            $line =~ s/\/var\/run\/vmkiscsid.pid already locked//g;

            # sometimes the output of vmware-cmd contains [] in front of each line
            # remove it on each line
            $line =~ s/^\s*\[\]\s+//;
        }

        # remove empty lines
        @output = grep(!/^\s*$/, @output);

        last;
    }

    # if there is an error still set at this point, then we exhausted our
    # retry count for calling _vmware_cmd()
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    # store the output in the vmware-cmd cache
    $self->store_cache_data("vmware_cmd_l", @output);

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: _get_next_scsi_id
# Objective:       Returns the next scsi id available in @config_file
# Details:         See POD documentation at the end of this file
# Returns:         (create_new_bus:bool, bus_count:int, lun_slot_count:int)
################################################################################
sub _get_next_scsi_id {
    my ($self, $controller, $config_file) = @_;

    my $lun_count = 1;
    my $bus_count = 0;

    # --EXAMPLE config file--
    # scsi0.present = "TRUE"
    # scsi0.sharedBus = "none"
    # scsi0.virtualDev = "lsilogic"
    # scsi0:0.present = "TRUE"
    # scsi0:0.fileName = "imagewin2k332_2.vmdk"
    # scsi0:0.mode = "independent-persistent"
    # scsi0:0.deviceType = "scsi-hardDisk"
    # scsi0:0.redo = ""
    # scsi0.pciSlotNumber = "16"
    # scsi1.present = "true"
    # scsi1.sharedBus = "none"
    # scsi1.virtualDev = "pvscsi"
    # scsi1:0.present = "true"
    # scsi1:0.fileName = "/vmfs/volumes/4acc722f-083f0239-7539-001a649864f4/nate_datastore_ibm3550-rtp10_naa.60a9800056724a774e5a534b4d6b4342:0_ibm3550-rtp10_0.vmdk"
    # scsi1:0.mode = "persistent"
    # scsi1:0.redo = ""
    # scsi1:0.deviceType = "scsi-hardDisk"

    # verify that there is a scsi bus at the current location
    while (grep(/^\s*scsi${bus_count}\.virtualdev\s*=\s*\".+?\"/i, @{$config_file})) {
        # if this bus is not turned on, then skip it
        # we cant assume this bus is configured properly
        if (not grep(/^\s*scsi${bus_count}\.present\s*=\s*\"true\"/i, @{$config_file})) {
            $bus_count++;
            $lun_count = 0;
            next;
        }

        # verify that this bus is of the correct controller type
        if (not grep(/^\s*scsi${bus_count}\.virtualdev\s*=\s*\"$${controller}\"/i, @{$config_file})) {

            # check which controller type is being used on this bus
            my $found_sas_controller = 0;
            my $found_pvscsi_controller = 0;
            $found_sas_controller    = 1 if (grep(/^\s*scsi${bus_count}\.virtualdev\s*=\s*\"lsisas1068\"/i, @{$config_file}));
            $found_pvscsi_controller = 1 if (grep(/^\s*scsi${bus_count}\.virtualdev\s*=\s*\"pvscsi\"/i, @{$config_file}));

            # if this scsi bus is the boot bus (IE: scsi0) and is of type lsisas1068 or pvscsi
            # then if the controller specified was lsilogic, it is ok to reuse these buses.
            # In fact, on some VMs it is not possible to boot off of
            # lsisas1068/pvscsi and have other disks on lsilogic scsi controllers
            if (($found_sas_controller or $found_pvscsi_controller) and
                 $$controller =~ /lsilogic/i and $bus_count == 0) {
                $self->sanlog(-msg => "Detected a pvscsi (paravirtual) or lsisas1068 (sas) scsi0 boot controller but was requested " .
                                      "to use lsilogic.\nWill reuse this scsi controller since it is not possible to boot off " .
                                      "these controller types and have other disks using lsilogic.");
                # modify the controller's memory reference with the type of controller that was found
                $$controller = "pvscsi"     if ($found_pvscsi_controller == 1);
                $$controller = "lsisas1068" if ($found_sas_controller    == 1);

                # reuse this bus (let the code continue)
            } else {
                # In this case, this scsi bus cant be used
                # so we need to go to the next bus,
                # increment bus count and reset lun count
                $bus_count++;
                $lun_count = 0;
                next;
            }
        }

        # if it is the correct bus type, then search for an available lun slot in this bus
        # maintain a checker to see if there are any more slots in this bus available
        my $no_more_slots = 0;

        # keep scanning the config file until we find a lun number that is
        # not being used. Increment the bus number when we reach the end of a bus.
        while (grep {/^\s*scsi${bus_count}\:${lun_count}\./} @{$config_file}) {

            # at this point we recognize that this lun count number is possibly in use.
            # verify that by scanning for a .filename entry
            if (not grep(/^\s*scsi${bus_count}\:${lun_count}\.filename/i, @{$config_file})) {
                # could not find a filename entry, so this lun count is not in use, lets use it
                # clean up the config file of any stale entries for this slot
                @{$config_file} = grep {!/^\s*scsi${bus_count}\:${lun_count}\./} @{$config_file};

                return (0, $bus_count, $lun_count);
            } else {
                # otherwise the lun count is in use, go to the next slot
                $lun_count++;

                # We can't use the lun_count value '7' because it is reserved by the
                # scsi controller itself, incremement again if necesessary
                $lun_count++ if ($lun_count == 7);

                # If the lun count has hit 16, then reset the lun count and increment
                # bus count value. A scsi bus can hold a maximum of 15 lun entries,
                # including slot '7'
                if ($lun_count == 16) {
                    # since here are no lun count slots left, go to the next bus iteration
                    $no_more_slots = 1;
                    last;
                }
            }
        }

        # if we ever get here it can mean two things
        # (1) no more slots in the current bus
        # (2) the current slot is not being used
        if ($no_more_slots) {
            # if there are no more slots, then go to the next bus
            # reset lun count and increment bus count
            $lun_count = 0;
            $bus_count++;
            next;
        } else {
            # we found the right bus and lun slot that's not in use
            # clean up the config file of any stale entries for this slot
            @{$config_file} = grep {!/^\s*scsi${bus_count}\:${lun_count}\./} @{$config_file};
            return (0, $bus_count, $lun_count);
        }
    }

    # at this point we are at the right bus:lun slot but the bus does not exist
    # tell the caller to make a new one

    return (1, $bus_count, $lun_count);
}

################################################################################
# Subroutine Name: _filter_config_file
# Objective:       This will filter out @config_file of all changes made to
#                  it by NATE automation, safely.
# Details:         See POD documentation at the end of this file
################################################################################
sub _filter_config_file {
    my ($self, @config_file) = @_;

    my @output = ();

    # clean up comments and blank lines from the config file
    @config_file = grep { !/^\s*\#/ and !/^\s*$/ } @config_file;

    # we will now go line by line in @config_file. All lines that
    # meet the requirements will get pushed onto @output.
    # if we encounter a line that has to be removed from @config_file
    # we have to make sure that we remove lines related to that line
    # that were pushed on to @ouput in a prior iteration.

    # clean up all scsi vmdk entries that match 'nate'
    while (scalar(@config_file) > 0) {
        my $line = shift @config_file;

        # if the line matches a scsi entry with nate
        if ($line =~ /^(scsi[0-9]+\:[0-9]+).+?\".*?nate.*?\"/) {
            # store the scsi id and scsi controller
            my ($scsi_id) = $1;
            my ($scsi_controller) = ($scsi_id =~ /(scsi[0-9]+)\:/);

            # ignore this scsi_id if it is scsi0:0 because this is the boot disk
            push(@output, $line) and next if ($scsi_id =~ /^scsi0:0\./);

            # now remove every line from @config_file and @output that uses
            # that scsi id
            @config_file = grep { !/^$scsi_id\./ } @config_file;
            @output = grep { !/^$scsi_id\./ } @output;

            # if no disks are using $scsi_controller:
            if (not grep {/^$scsi_controller\:[0-9]+/} @config_file) {
                # then remove the scsi_controller.
                # ignore this line if the scsi controller is scsi0
                # because it contains the boot disk.
                next if ($scsi_controller =~ /^scsi0$/);
                @config_file = grep { !/^$scsi_controller\./ } @config_file;
                @output = grep { !/^$scsi_controller\./ } @output;
            }

            # this line should not be added to the output array,
            # so go to the next line
            next;
        }

        # this will make sure that the guest OS is always time sync'd with the
        # ESX host
        if ($line =~ /tools\.syncTime/i) {
            $line = "tools.syncTime = \"TRUE\"";
        }

        # This line has passed all our requirements, add it to the output array
        push (@output, $line);
    }

    # return the modified config file
    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: _update_vmhba_adapters
# Objective:       This will update the Host object with the vmhba adapter
#                  information.
################################################################################
sub _update_vmhba_adapters {
    my ($self) = @_;

    $self->_clear_err();

    # update the vmhba adapters for esx4 or esx3
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $self->_update_vmhba_adapters_esx5();
    } elsif ($self->get_os_version() =~ /^4\./) {
        $self->_update_vmhba_adapters_esx4();
    } else {
        $self->_update_vmhba_adapters_esx3();
    }

    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # initialize the working adapters
    $self->{vmhba_adapter} = [];
    $self->{vmhba_iscsi_adapters} = [];

    # update the current working adapters depending on protocol type
    # the current working adapters may contain multiple protocol types
    if ($self->{protocol} =~ /fcp/i) {
        push(@{$self->{vmhba_adapter}}, @{$self->{vmhba_fcp_adapters}});
    }
    if ($self->{protocol} =~ /iscsi/i) {
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^software$/i) {
            push(@{$self->{vmhba_adapter}}, @{$self->{vmhba_swiscsi_adapters}});
            push(@{$self->{vmhba_iscsi_adapters}}, @{$self->{vmhba_swiscsi_adapters}});
        } elsif ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^hardware$/i) {
            push(@{$self->{vmhba_adapter}}, @{$self->{vmhba_hwiscsi_adapters}});
            push(@{$self->{vmhba_iscsi_adapters}}, @{$self->{vmhba_hwiscsi_adapters}});
        } else {
            $self->_set_err("Invalid iSCSI Type: $self->{iscsi}->{CONFIG_ISCSI_TYPE}");
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _update_vmhba_adapters_esx3
# Objective:       This will update the Host object with the vmhba adapter
#                  information for ESX3.
################################################################################
sub _update_vmhba_adapters_esx3 {

    my ($self) = @_;
    my (@fcp_adapters, @hw_iscsi_adapters, @sw_iscsi_adapters);

    # search for the fcp and iscsi adapters
    my $cmd = "/usr/sbin/esxcfg-mpath -a";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 1) {
        $self->_set_err("Command ($cmd) failed. Status = ($status)");
        return $ClientUtil::FALSE;
    }

    # store the raw output for later use
    $self->store_cache_data("vmhba_raw_output", @output);

    # @output can look like this:
    #
    # vmhba1 210100e08ba17d60 1:8.1
    # vmhba2 210000e08b817d60 1:8.0
    # vmhba3 iqn.2000-04.com.qlogic:qla4052c.fs20550b13228.1 1:6.1
    # vmhba4 iqn.2000-04.com.qlogic:qla4052c.fs20550b13228.2 1:6.3
    # vmhba40 iqn.1998-01.com.vmware:vmshe-dell2650-4-6e52bd8e sw
    # vmhba32 iqn.1998-01.com.vmware:vmshe-dell2650-4-6e52bd8e sw

    # turn the output into a string delimited by newlines and parse the string.
    foreach my $line (@output) {
        # check if this is a FCP adapter
        if ($line !~ /\s*iqn\./ and $line =~ /(vmhba\d+)\s*/) {
            push(@fcp_adapters, $1);
        }
        # check if this is a hw iscsi adapter
        elsif ($line !~ /\s+sw/ and $line =~ /(vmhba\d+)\s*/) {
            push(@hw_iscsi_adapters, $1);
        }
        # check if this is a sw iscsi adapter
        elsif ($line =~ /\s+sw/ and $line =~ /(vmhba\d+)\s*/) {
            push(@sw_iscsi_adapters, $1);
        }
    }

    # apply the attributes to the object
    # we are storing each as an array reference.
    push(@{$self->{vmhba_fcp_adapters}}, @fcp_adapters);
    push(@{$self->{vmhba_hwiscsi_adapters}}, @hw_iscsi_adapters);
    push(@{$self->{vmhba_swiscsi_adapters}}, @sw_iscsi_adapters);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _update_vmhba_adapters_esx4
# Objective:       This will update the Host object with the vmhba adapter
#                  information.
################################################################################
sub _update_vmhba_adapters_esx4 {
    my ($self) = @_;

    $self->_clear_err();

    my (@fcp_adapters, @hw_iscsi_adapters, @sw_iscsi_adapters);

    # search for the fcp and iscsi adapters
    my $cmd = "esxcfg-scsidevs -a";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 1) {
        $self->_set_err("Command ($cmd) failed. Status = ($status)");
        return $ClientUtil::FALSE;
    }

    # store the raw output for later use
    $self->store_cache_data("vmhba_raw_output", @output);

    # @output can look like this:
    #
    #vmhba0  aacraid           link-n/a  pscsi.vmhba0                            (2:0.0) Adaptec ServeRAID 8k/8k-l8
    #vmhba1  qla2xxx           link-up   fc.2000001b3200d037:2100001b3200d037    (7:0.0) QLogic Corp ISP2432-based 4Gb Fibre Channel to PCI Express HBA
    #vmhba2  qla2xxx           link-up   fc.2001001b3220d037:2101001b3220d037    (7:0.1) QLogic Corp ISP2432-based 4Gb Fibre Channel to PCI Express HBA
    #vmhba3  ata_piix          link-n/a  ide.vmhba3                              (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba32 ata_piix          link-n/a  ide.vmhba32                             (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba33 iscsi_vmk         link-n/a  iscsi.vmhba33                           () Software iSCSI

    #vmhba0  cciss             link-n/a  block.cciss/c0d0:0                      (6:0.0) Hewlett-Packard Company Smart Array P400
    #vmhba1  qla4xxx           link-n/a  iqn.2000-04.com.qlogic:qle4062c.gs40748a01798.1(15:1.1) QLogic Corp QLE406Xc iSCSI Host Bus Adapter
    #vmhba2  qla4xxx           link-n/a  iqn.2000-04.com.qlogic:qle4062c.gs40748a01798.2(15:1.3) QLogic Corp QLE406Xc iSCSI Host Bus Adapter
    #vmhba3  qla4xxx           link-n/a  iqn.2000-04.com.qlogic:qle4062c.gs40748a01768.1(24:1.1) QLogic Corp QLE406Xc iSCSI Host Bus Adapter
    #vmhba4  qla4xxx           link-n/a  iqn.2000-04.com.qlogic:qle4062c.gs40748a01768.2(24:1.3) QLogic Corp QLE406Xc iSCSI Host Bus Adapter
    #vmhba5  ata_piix          link-n/a  ide.vmhba5                              (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba32 ata_piix          link-n/a  ide.vmhba32                             (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller

    #vmhba0  cciss             link-n/a  block.cciss/c0d0:0                      (6:0.0) Hewlett-Packard Company Smart Array P400
    #vmhba1  ata_piix          link-n/a  ide.vmhba1                              (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba32 ata_piix          link-n/a  ide.vmhba32                             (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba33 iscsi_vmk         link-n/a  iqn.1998-01.com.vmware:vmshe-hp380-06-0ce4d803() Software iSCSI


    # FCOE cards (qla)
    #vmhba0  mptsas            link-n/a  sas.5000000000000000                    (1:0.0) LSI Logic / Symbios Logic LSI1068E
    #vmhba1  qla2xxx           link-up   fc.200000e08b851dbb:210000e08b851dbb    (13:3.0) QLogic Corp ISP2422-based 4Gb Fibre Channel to PCI-X HBA
    #vmhba2  qla2xxx           link-up   fc.200100e08ba51dbb:210100e08ba51dbb    (13:3.1) QLogic Corp ISP2422-based 4Gb Fibre Channel to PCI-X HBA
    #vmhba5  ata_piix          link-n/a  ide.vmhba5                              (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba32 usb-storage       link-n/a  usb.vmhba32                             () USB
    #vmhba33 usb-storage       link-n/a  usb.vmhba33                             () USB
    #vmhba34 ata_piix          link-n/a  ide.vmhba34                             (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller

    #vmhba0  cciss             link-n/a  block.cciss/c0d0:0                      (6:0.0) Hewlett-Packard Company Smart Array P400
    #vmhba1  lpfc820           link-up   fc.20000000c97bc544:10000000c97bc544    (11:0.0) Emulex Corporation LPe12000 8Gb Fibre Channel Host Adapter
    #vmhba2  lpfc820           link-up   fc.20000000c97bc545:10000000c97bc545    (11:0.1) Emulex Corporation LPe12000 8Gb Fibre Channel Host Adapter
    #vmhba3  lpfc820           link-up   fc.20000000c97328ca:10000000c97328ca    (23:0.0) Emulex Corporation LPe12000 8Gb Fibre Channel Host Adapter
    #vmhba4  lpfc820           link-up   fc.20000000c97328cb:10000000c97328cb    (23:0.1) Emulex Corporation LPe12000 8Gb Fibre Channel Host Adapter
    #vmhba5  ata_piix          link-n/a  ide.vmhba5                              (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba32 ata_piix          link-n/a  ide.vmhba32                             (0:31.1) Intel Corporation 631xESB/632xESB IDE Controller

    # Emulex serverengines hw iscsi offload
    #vmhba2  be2iscsi          online    iqn.1990-07.com.emulex:00-00-c9-bb-1a-df(0:11:0.2) ServerEngines Corp. Emulex OneConnect
    #vmhba3  be2iscsi          online    iqn.1990-07.com.emulex:00-00-c9-bb-1a-df(0:11:0.3) ServerEngines Corp. Emulex OneConnect

    # Fusion I/O Card (should get ignored)
    # vmhba35     vlun           fc.5b4e0cdaabbccdde:5b4e0cdaabbccdde             ()       Unknown


    foreach my $line (@output) {
        # check if this is a FCP adapter
        if ($line =~ /fc\.\S+\:\S+|Fibre|lpfc/ and $line =~ /(vmhba\d+)\s*/ and $line !~ /vlun/) {
            push(@fcp_adapters, $1);
        }
        # check if this is a sw iscsi adapter
        elsif ($line =~ /Software iSCSI|iscsi_vmk/i and $line =~ /(vmhba\d+)\s*/) {
            push(@sw_iscsi_adapters, $1);
        }
        # check if this is a hw iscsi adapter
        # ignore unbound adapters because that means they are not bound to a VMKernel port
        elsif ($line =~ /iSCSI Host Bus Adapter|qla4xxx|Broadcom iSCSI Adapter|be2iscsi|bnx2i/i and
               $line !~ /unbound/ and
               $line =~ /(vmhba\d+)\s*/) {
            push(@hw_iscsi_adapters, $1);
        }
    }

    # apply the attributes to the object
    # we are storing each as an array reference.
    push(@{$self->{vmhba_fcp_adapters}}, @fcp_adapters);
    push(@{$self->{vmhba_hwiscsi_adapters}}, @hw_iscsi_adapters);
    push(@{$self->{vmhba_swiscsi_adapters}}, @sw_iscsi_adapters);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _update_vmhba_adapters_esx5
# Objective:       This will update the Host object with the vmhba adapter
#                  information.
################################################################################
sub _update_vmhba_adapters_esx5 {
    my ($self) = @_;

    $self->_clear_err();

    my (@fcp_adapters, @hw_iscsi_adapters, @sw_iscsi_adapters);

    my $core_storage_namespace = $self->{esxcli_corestorage_namespace};

    # search for the fcp and iscsi adapters
    my $cmd = "esxcli $core_storage_namespace adapter list";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                    -include_host_name => 1);
    if ($status > 1) {
        $self->_set_err("Command ($cmd) failed. Status = ($status)");
        return $ClientUtil::FALSE;
    }

    # store the raw output for later use
    $self->store_cache_data("vmhba_raw_output", @output);

    # @output can look like this:

    #HBA Name  Driver     Link State  UID                                              Description
    #--------  ---------  ----------  -----------------------------------------------  ---------------------------------------------------------------------
    #vmhba0    ata_piix   link-n/a    ide.vmhba0                                       (0:0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba1    aacraid    link-n/a    pscsi.vmhba1                                     (0:2:0.0) Adaptec ServeRAID 8k/8k-l8
    #vmhba2    lpfc820    link-up     fc.20000000c96e4dd6:10000000c96e4dd6             (0:19:0.0) Emulex Corporation LPe11000 4Gb Fibre Channel Host Adapter
    #vmhba3    lpfc820    link-up     fc.20000000c96e4dd7:10000000c96e4dd7             (0:19:0.1) Emulex Corporation LPe11000 4Gb Fibre Channel Host Adapter
    #vmhba32   ata_piix   link-n/a    ide.vmhba32                                      (0:0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba33   iscsi_vmk  online      iqn.1998-01.com.vmware:auto-ibm3550-03-6e1004b8  iSCSI Software Adapter

    #HBA Name  Driver    Link State  UID                                     Description
    #--------  --------  ----------  --------------------------------------  -----------------------------------------------------------
    #vmhba0    ata_piix  link-n/a    ide.vmhba0                              (0:0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba1    aacraid   link-n/a    pscsi.vmhba1                            (0:2:0.0) Adaptec ServeRAID 8k/8k-l8
    #vmhba32   ata_piix  link-n/a    ide.vmhba32                             (0:0:31.1) Intel Corporation 631xESB/632xESB IDE Controller
    #vmhba33   fcoe      link-up     fcoe.1000001b21531dd2:2000001b21531dd2  () Software FCoE
    #vmhba34   fcoe      link-up     fcoe.1000001b21531dd3:2000001b21531dd3  () Software FCoE

    # Emulex serverengines hw iscsi offload
    #vmhba2  be2iscsi          online    iqn.1990-07.com.emulex:00-00-c9-bb-1a-df(0:11:0.2) ServerEngines Corp. Emulex OneConnect
    #vmhba3  be2iscsi          online    iqn.1990-07.com.emulex:00-00-c9-bb-1a-df(0:11:0.3) ServerEngines Corp. Emulex OneConnect

    # Broadcom HW iSCSI adapters
    #vmhba32   bnx2i     unbound     iscsi.vmhba32                                       Broadcom iSCSI Adapter
    #vmhba33   bnx2i     online      iqn.1998-01.com.vmware:auto-hpdl380-03:21739856:33  Broadcom iSCSI Adapter
    #vmhba34   bnx2i     online      iqn.1998-01.com.vmware:auto-hpdl380-03:21739856:34  Broadcom iSCSI Adapter
    #vmhba35   bnx2i     unbound     iscsi.vmhba35                                       Broadcom iSCSI Adapter

    # Fusion I/O Card (should get ignored)
    # vmhba35     vlun           fc.5b4e0cdaabbccdde:5b4e0cdaabbccdde             ()       Unknown

    foreach my $line (@output) {
        # check if this is a FCP adapter (ignore fusion i/o cards)
        if ($line =~ /fc\.\S+\:\S+|Fibre|lpfc/ and $line =~ /(vmhba\d+)\s*/ and $line !~ /vlun/i) {
            push(@fcp_adapters, $1);
        }
        # check if this is a sw FCoE adapter
        elsif ($line =~ /fcoe\.\S+:\S+|Software FCoE/ and $line =~ /(vmhba\d+)\s*/) {
            push(@fcp_adapters, $1);
        }
        # check if this is a sw iscsi adapter
        elsif ($line =~ /Software iSCSI|iscsi_vmk/i and $line =~ /(vmhba\d+)\s*/) {
            push(@sw_iscsi_adapters, $1);
        }
        # check if this is a hw iscsi adapter
        # ignore unbound adapters because that means they are not bound to a VMKernel port
        elsif ($line =~ /iSCSI Host Bus Adapter|qla4xxx|Broadcom iSCSI Adapter|be2iscsi|bnx2i/i and
               $line !~ /unbound/ and
               $line =~ /(vmhba\d+)\s*/) {
            push(@hw_iscsi_adapters, $1);
        }
    }

    # apply the attributes to the object
    # we are storing each as an array reference.
    push(@{$self->{vmhba_fcp_adapters}}, @fcp_adapters);
    push(@{$self->{vmhba_hwiscsi_adapters}}, @hw_iscsi_adapters);
    push(@{$self->{vmhba_swiscsi_adapters}}, @sw_iscsi_adapters);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_iscsi_node_name_by_adapter
# Objective:       this will return the nodename of the iscsi adapter
################################################################################
sub _get_iscsi_node_name_by_adapter {
    my ($self, $adapter) = @_;

    # obtain the initiator node name
    my $cmd = "/usr/sbin/vmkiscsi-tool $adapter -I";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed. Status = ($status)");
        return "";
    }

    # pares out the nodename and return it
    foreach my $line (@output) {
        if ($line =~ /iSCSI Node Name: (iqn\.\S+)/i) {
            return $1;
        }
    }

    # otherwise set an error
    $self->_set_err("Could not determine iSCSI Node name for adapter: $adapter");
    return "";
}

################################################################################
# Subroutine Name: _flush_vmhba_adapters
# Objective:       This will flush all vmhba adapters
#                  information.
# Details:         See POD documentation at the end of this file
################################################################################
sub _flush_vmhba_adapters {
    my ($self) = @_;

    my (@cmds) = ();
    my $retry_count = 0;

    $self->_clear_err();

    # rescan each fcp adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_fcp_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);
        push (@cmds, "/usr/sbin/esxcfg-rescan " . $vmhba_adapter);
    }

    # rescan each hw iscsi adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_hwiscsi_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);
        push (@cmds, "/usr/sbin/esxcfg-rescan " . $vmhba_adapter);
    }

    # rescan each sw iscsi adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_swiscsi_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);

        # we won't flush the sw iscsi adapter if the sw initiator is
        # not enabled
        my $cmd = "/usr/sbin/esxcfg-swiscsi --query";
        my ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # if its enabled, then setup the command to rescan this adapter.
        # otherwise, forget about flushing sw iscsi adapters.
        if (grep(/Software iSCSI is enabled/i, @output)) {
            push (@cmds, "/usr/sbin/esxcfg-rescan " . $vmhba_adapter);
        } else {
            last;
        }
    }

    $self->sanlog(-msg => "Flushing all vmhba adapters...");
    foreach my $cmd (@cmds) {
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ESXHost::_flush_vmhba_adapters() ".
                                "Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }
    $self->sanlog(-msg => "All vmhba adapters have been flushed.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _logging_configure()
# Objective:       This will configure all logging on the ESX Host
# Details:         See POD documentation at the end of this file
################################################################################
sub _logging_configure {
    my ($self) = @_;

    my ($status, @output, $cmd);

    # set the SCSI log settings
    $self->_set_scsi_logging_settings();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # set the max size of the vmkernel logrotater to something big to make
    # log reading easier.
    my @vmkernel_cfg_file;
    push(@vmkernel_cfg_file, "/var/log/vmkernel {");
    push(@vmkernel_cfg_file, "    create 0600 root root");
    push(@vmkernel_cfg_file, "    missingok");
    push(@vmkernel_cfg_file, "    nocompress");
    push(@vmkernel_cfg_file, "    daily");
    push(@vmkernel_cfg_file, "    rotate 5");
    push(@vmkernel_cfg_file, "    size 10M");
    push(@vmkernel_cfg_file, "    sharedscripts");
    push(@vmkernel_cfg_file, "    postrotate");
    push(@vmkernel_cfg_file, "        /bin/kill -HUP \`cat /var/run/syslogd.pid 2> /dev/null\` 2> /dev/null || true");
    push(@vmkernel_cfg_file, "    endscript");
    push(@vmkernel_cfg_file, "}");

    my @vmkwarning_cfg_file;
    push(@vmkwarning_cfg_file, "/var/log/vmkwarning {");
    push(@vmkwarning_cfg_file, "    create 0600 root root");
    push(@vmkwarning_cfg_file, "    missingok");
    push(@vmkwarning_cfg_file, "    daily");
    push(@vmkwarning_cfg_file, "    rotate 5");
    push(@vmkwarning_cfg_file, "    size 10M");
    push(@vmkwarning_cfg_file, "    sharedscripts");
    push(@vmkwarning_cfg_file, "    postrotate");
    push(@vmkwarning_cfg_file, "        /bin/kill -HUP `cat /var/run/syslogd.pid 2> /dev/null` 2> /dev/null || true");
    push(@vmkwarning_cfg_file, "    endscript");
    push(@vmkwarning_cfg_file, "}");

    my @messages_cfg_file;
    push(@messages_cfg_file, "/var/log/messages {");
    push(@messages_cfg_file, "    create 0600 root root");
    push(@messages_cfg_file, "    missingok");
    push(@messages_cfg_file, "    nocompress");
    push(@messages_cfg_file, "    daily");
    push(@messages_cfg_file, "    rotate 5");
    push(@messages_cfg_file, "    size 10M");
    push(@messages_cfg_file, "    sharedscripts");
    push(@messages_cfg_file, "    postrotate");
    push(@messages_cfg_file, "        /bin/kill -HUP \`cat /var/run/syslogd.pid 2> /dev/null\` 2> /dev/null || true");
    push(@messages_cfg_file, "    endscript");
    push(@messages_cfg_file, "}");

    my $dir_path = "/etc/logrotate.d";
    my $vmkernel_file = "${dir_path}/vmkernel";
    my $vmkwarning_file = "${dir_path}/vmkwarning";
    my $messages_file = "${dir_path}/messages";

    # rebuild the vmkernel config file into a temp file
    my $vmkernel_file_new = $vmkernel_file . ".tmp." . time();
    foreach my $line (@vmkernel_cfg_file) {
        $cmd = "echo \'$line\' >> $vmkernel_file_new";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ".
                            "($status)");
            return $ClientUtil::FALSE;
        }
    }

    # now overwrite the old config file with the new one
    $self->sanlog(-msg => "Overwriting the old vmkernel file with the new ".
                          "vmkernel file.",
                  -msg_level => 3);
    $cmd = "mv --backup --suffix=\".bak\" $vmkernel_file_new $vmkernel_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # rebuild the vmkwarning config file into a temp file
    my $vmkwarning_file_new = $vmkwarning_file . ".tmp." . time();
    foreach my $line (@vmkwarning_cfg_file) {
        $cmd = "echo \'$line\' >> $vmkwarning_file_new";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ".
                            "($status)");
            return $ClientUtil::FALSE;
        }
    }

    # now overwrite the old config file with the new one
    $self->sanlog(-msg => "Overwriting the old vmkwarning file with the new ".
                          "vmkwarning file.",
                  -msg_level => 3);
    $cmd = "mv --backup --suffix=\".bak\" $vmkwarning_file_new $vmkwarning_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # rebuild the messages config file into a temp file
    my $messages_file_new = $messages_file . ".tmp." . time();
    foreach my $line (@messages_cfg_file) {
        $cmd = "echo \'$line\' >> $messages_file_new";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ".
                            "($status)");
            return $ClientUtil::FALSE;
        }
    }

    # now overwrite the old config file with the new one
    $self->sanlog(-msg => "Overwriting the old messages file with the new ".
                          "messages file.",
                  -msg_level => 3);
    $cmd = "mv --backup --suffix=\".bak\" $messages_file_new $messages_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # commit the new config file to the logrotate utility
    $cmd = "/usr/sbin/logrotate $vmkernel_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # commit the new config file to the logrotate utility
    $cmd = "/usr/sbin/logrotate $vmkwarning_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # commit the new config file to the logrotate utility
    $cmd = "/usr/sbin/logrotate $messages_file";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_scsi_logging_settings()
# Objective:       Will enable logging for scsi errors in ESX
################################################################################
sub _set_scsi_logging_settings {
    my ($self) = @_;

    # set the scsi logging settings for esx4 or esx3
    if ($self->get_os_version() =~ /^4\./) {
        # setting esx4 scsi logging may get blocked by some asynchronous lock file, retry after
        # sleeping for a bit
        my $retry_count = 0;
        while (1) {
            $self->_set_esx4_scsi_logging_settings();
            if ($self->error()) {
                $retry_count++;

                # kick out if we've exhausted the retry count
                last if ($retry_count >= 3);
                $self->sanlog(-msg => "Retrying ESX4 SCSI Logging settings ($retry_count / 2)");

                # clear the error and sleep before trying again
                $self->_clear_err();
                &SanUtil::nate_sleep(10);
            } else {
                # the command worked, kick out
                last;
            }
        }

        # check if the command failed
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        # configure for esx3
        $self->_set_esx3_scsi_logging_settings();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # set logging levels on all qlogic hbas
    my $cmd = "ls /proc/scsi/qla2*/ | cat";
    my ($status, @qla_output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # sample output:
    # [root@ibm3550-rtp10]# ls /proc/scsi/qla2*/ | cat
    # 4
    # 5

    # configure each hba that contains a digit
    foreach my $line (@qla_output) {
        # skip lines that do not only contain a digit that corresponds to the
        # hba number, machines with no qlogic fcp hbas will flush through this loop
        next if ($line !~ /^\s*\d+\s*$/);

        # enable qla logging on this hba
        $cmd = "echo \"scsi-qlaenable-log\" > /proc/scsi/qla2*/$line";
        ($status) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_esx3_scsi_logging_settings()
# Objective:       Will enable logging for scsi errors in ESX3
################################################################################
sub _set_esx3_scsi_logging_settings {
    my ($self) = @_;

    my ($cmd, $status, @output);

    # the following commands enable verbose logging to gather addition
    # storage errors in the log files.
    $cmd = "/usr/sbin/esxcfg-advcfg -s 1 /Scsi/LogAborts";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/esxcfg-advcfg -s 1 /Scsi/LogMultiPath";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/esxcfg-advcfg -s 1 /Scsi/PrintCmdErrors";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_esx4_scsi_logging_settings()
# Objective:       Will enable logging for scsi errors in ESX4
################################################################################
sub _set_esx4_scsi_logging_settings {
    my ($self) = @_;

    my ($cmd, $status, @output);

    # the following commands enable verbose logging to gather addition
    # storage errors in the log files.
    $cmd = "esxcfg-advcfg -s 1 /Scsi/LogCmdErrors";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
    $cmd = "esxcfg-advcfg -s 1 /Scsi/LogMPCmdErrors";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command '$cmd' returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # only set scsi settings in non vma host environments
    if (not $self->has_vma_host()) {
        # as per burt 351751 to increase scsi logging levels
        $cmd = "echo 0x1000 > /proc/sys/dev/scsi/logging_level";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
        $cmd = "cat /proc/sys/dev/scsi/logging_level";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: qfull_configure()
# Objective:       Configure qfull settings
################################################################################
sub qfull_configure {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    # return if qfull already was configured
    if ($self->{_qfull_configured}) {
        return $ClientUtil::TRUE;
    } else {
        # initialize it to false
        $self->{_qfull_configured} = 0;
    }

    my $os_version = $self->get_os_version();
    my $os_build_number = $self->{os_build_number};
    my $qfull_sample_size = "32";
    my $qfull_threshold = "8";

    # globally set the qfull options for ESX4.x, ESX5.0, ESX5.1u1, ESX5.5, ESX6 and ESX3u4 (build >= 153875)
    if (($os_version =~ /^(4\.|5\.0|5\.5|6\.[0-9])/) or
        ($os_version =~ /^3\./  and $os_build_number >= 153875) or
        ($os_version =~ /^5\.1/ and $os_build_number >= 1065491)) {
        $self->sanlog(-msg => "Configuring QFull Settings");

        # set the QFull sample size to 32
        $cmd = "esxcfg-advcfg -s ${qfull_sample_size} /Disk/QFullSampleSize";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # set the QFull threshold to 8
        $cmd = "esxcfg-advcfg -s ${qfull_threshold} /Disk/QFullThreshold";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
        $self->{_qfull_configured} = 1;
    } elsif ($os_version =~ /^5\.1/ and $os_build_number < 1065491) {
        # vmware screwed up with esxi5.1 since it does not have the option for global qfull settings,
        # we need to set qfull settings on a per device basis
        # check if there is an MPDevMgr object, if there is grab the devices from that
        if (not defined($self->mpdevmgr())) {
            $self->sanlog(-msg => "Skipping Configuration of QFull Settings, no MPDevMgr loaded yet");
            return $ClientUtil::TRUE;
        }

        $self->sanlog(-msg => "Configuring QFull Settings");

        my $corestorage_ns = $self->{esxcli_corestorage_namespace};
        my @dev_objs = $self->mpdevmgr()->get_device_objs();
        foreach my $dev_obj (@dev_objs) {
            my $dev_name = $dev_obj->mp_device_name();
            $cmd = "esxcli ${corestorage_ns} device set --device ${dev_name} --queue-full-sample-size ${qfull_sample_size} --queue-full-threshold ${qfull_threshold}";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1,
                                                         -reconnect_on_failure => 1);
            if ($status > 0) {
                $self->_set_err("Command '$cmd' returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
        $self->{_qfull_configured} = 1;
    } else {
        # unsupported OS for qfull
        $self->sanlog(-msg => "Skipping Configuration of QFull Settings (unsupported OS)");
        $self->{_qfull_configured} = 0;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: lun_target_reset()
# Objective:       Will do target reset or lun reset for a given Lun.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_target_reset{
    my ($self, %args) = @_;
    my (@lun_naa_ids, $reset_count, $reset_type, $conn_temp, $vma_host,);
    my ($cmd, $status, @output);
    # Default value for reset count
    $reset_count = 1;
    # parse arguments
    foreach (keys %args) {
        if (/-?lun_naa_ids/) {
            if (scalar(@{$args{$_}}) < 1) {
                $self->_set_err("Invalid value for -lun_naa_ids => $args{$_}. Expecting an array of lun naa ids");
                return $ClientUtil::FALSE;
            }
            @lun_naa_ids = @{$args{$_}};
            next;
        }
        if (/-?reset_count/) {
            if (not $args{$_} =~ /\d/ || $args{$_} < 1) {
                $self->_set_err("Invalid value for -reset_count => $args{$_}. reset_count value should be 1 or above");
                return $ClientUtil::FALSE;
            }
            $reset_count = $args{$_};
            next;
        }
        if (/-?reset_type/) {
            if (($args{$_} ne "targetreset") and ($args{$_} ne "lunreset")) {
                $self->_set_err("Invalid value for -reset_type => $args{$_}. Expected values are targetreset|lunreset");
                return $ClientUtil::FALSE;
            }
            $reset_type = $args{$_};
            next;
        }
    }
        if ($reset_type !~ /targetreset|lunreset/) {
            $self->_set_err("Invalid reset type - $_");
            return $ClientUtil::FALSE;
        }

        # verifying scalar lun naa ids
        if (!scalar (@lun_naa_ids)) {
             $self->_set_err("Missing argument -lun_naa_ids");
             return $ClientUtil::FALSE;
        }
        $vma_host = $self->{vma_host};
        # Checking if vma is configured to decide on the connection to ESXi/ESX
        if($vma_host ne "") {
        # Checking if ESX 4.x or ESXi 5.x
            if ($self->get_os_version() =~ /^4\./){
                $conn_temp=  $self->_connect($self->{phy_host}, $self->{phy_host_username}, $self->{phy_host_password}, 1800, "THARNHOST");
            }else{
                my $host_obj = Hostrec->new(id => $self->{phy_host});
                $conn_temp= Connectrec->new(host => $host_obj,id => $self->{phy_host}, username=> $self->{phy_host_username}, password => $self->{phy_host_password},conntype => "ssh-persist", shell_type => "none", template => "");
            }
        }else{
            # Connecton is already made directly to ESX/ESXi
            $conn_temp= $self->{conn}
        }
        $self->sanlog(-msg => "Starting $reset_type.");
        # LUN/target reset operation on ESX/ESXi
        for (my $count = 1 ; $count <= $reset_count; $count++){
            $self->sanlog(-msg => "Running $reset_type iteration ($count/$reset_count).");
            #Run through each Lun naa id
            foreach (@lun_naa_ids) {
                $cmd = "vmkfstools -L $reset_type /vmfs/devices/disks/" . $_;
                $self->sanlog(-msg => "$reset_type for LUN $_.");
                ($status, @output) = &ClientUtil::execute($conn_temp, $cmd);
                    if ($status != 0) {
                        $self->_set_err("Error running command ($cmd) : ERROR : @output");
                        return $ClientUtil::FALSE;
                    }
                #Sleep before next reset operation
                sleep 10;
            }
        }

    &Tharn::disconnect($conn_temp, '-ok');

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_vms()
# Objective:       Will turn on specified vms on this ESX
#                  host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_on_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my $dont_wait = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar (@vm_list = @{$args{$_}})) {
                $self->_set_err("Error: ESXHost::_power_on_vms() Invalid " .
                                "value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?dont_wait/) {
            if (not length($dont_wait = $args{$_})) {
                $self->_set_err("Error: ESXHost::_power_on_vms() Invalid " .
                                "value for -dont_wait => $dont_wait");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not @vm_list) {
        $self->_set_err("Error: ESXHost::_power_on_vms() Missing required ".
                        "parameter -vm_list");
        return $ClientUtil::FALSE;
    }

    # get the config file for each $vm in @vm_list and power them on.
    foreach my $vm (@vm_list) {
        my $vm_config_file = $self->_get_vm_config_file(-vm_name => $vm);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # get the vm's power status
        my $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" getstate";
        my ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # if the vm is already on, skip
        if (grep { /on/i } @output) {
            $self->sanlog(-msg => "$vm is already powered on");
            next;
        }

        # otherwise start it up
        $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" start";
        ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Powered on $vm");
    }

    if ($dont_wait == 1) {
        $self->sanlog(-msg => "All VMs powered on. Not waiting for them to boot.");
    } else {
        $self->sanlog(-msg => "All VMs powered on. Waiting for them to boot...");

        # wait for all VMs to boot up. A warning will be displayed if some
        # take too long to boot.
        $self->_wait_for_vms_to_boot(@vm_list);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_vms()
# Objective:       Will turn off specified vms on this ESX
#                  host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_off_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@vm_config_file_waitlist, @vms_that_were_shutdown);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar (@vm_list = @{$args{$_}})) {
                $self->_set_err("Error: ESXHost::_power_off_vms() Invalid " .
                                "value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not @vm_list) {
        $self->_set_err("Error: ESXHost::_power_off_vms() Missing required ".
                        "parameter -vm_list");
        return $ClientUtil::FALSE;
    }

    # get the config file for each $vm in @vm_list and power them off.
    foreach my $vm (@vm_list) {
        my $vm_config_file = $self->_get_vm_config_file(-vm_name => $vm);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # check if the vm is already off
        my $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" getstate";
        my ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # if the vm is already off, skip it
        if (grep { /off/i } @output) {
            $self->sanlog(-msg => "$vm is already powered off");
            next;
        }

        # otherwise shut the machine down cleanly
        if ($self->has_vma_host()) {
            # command is different if under vma control
            $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" stop soft"
        } else {
            $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" stop trysoft";
        }
        ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            # sometimes an error occurs stating that vmware tools
            # is not running. In this case, force the shutdown.
            if (grep { /VMware tools are not running|VMware Tools is not running/i } @output or
                $status == 231) {
                $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" stop hard";
                my ($status, @output) = $self->_vmware_cmd($cmd);
                if ($self->error()) {
                    return $ClientUtil::FALSE;
                }
                $self->sanlog(-msg => "Forced shutdown for $vm");

                next;
            } else {
                return $ClientUtil::FALSE;
            }
        }

        # push this config file on the shutdown wait list
        push(@vm_config_file_waitlist, $vm_config_file);
        $self->sanlog(-msg => "Issued shutdown for $vm");
        push(@vms_that_were_shutdown, $vm);
    }

    # wait for the VMs in the waitlist to shutdown
    $self->sanlog(-msg => "Waiting for VMs to shutdown.");
    $self->_wait_for_vms_to_shutdown(@vm_config_file_waitlist);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "All VMs have shutdown");

    # update the vms_that_were_shutdown attribute so it can be accessed later
    $self->{vms_that_were_shutdown} = \@vms_that_were_shutdown;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_vms()
# Objective:       Will power cycle specified vms on
#                  this ESX host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_cycle_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar (@vm_list = @{$args{$_}})) {
                $self->_set_err("Error: ESXHost::_power_cycle_vms() Invalid " .
                                "value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not @vm_list) {
        $self->_set_err("Error: ESXHost::_power_cycle_vms() Missing required ".
                        " parameter -vm_list");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Power cycling VMs...");
    $self->_power_off_vms(-vm_list => \@vm_list);

    $self->_power_on_vms(-vm_list => \@vm_list);
    $self->sanlog(-msg => "Power cycle complete.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_reset_vms()
# Objective:       Will reset specified vms on this ESX
#                  host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_reset_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar (@vm_list = @{$args{$_}})) {
                $self->_set_err("Error: ESXHost::_power_reset_vms() Invalid ".
                                "value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not @vm_list) {
        $self->_set_err("Error: ESXHost::_power_reset_vms() Missing required ".
                        " parameter -vm_list");
        return $ClientUtil::FALSE;
    }

    # otherwise, get the config file for each $vm in @vm_list
    # and power it on.
    foreach my $vm (@vm_list) {
        my $vm_config_file = $self->_get_vm_config_file(-vm_name => $vm);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        my $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" reset";
        my ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Power reset for $vm");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_all_vms()
# Objective:       Will turn on all vms on this ESX host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_on_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    my (@all_vms) = $self->_get_all_vm_config_files();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # since the elements of @all_vms are the paths to the config files,
    # we can not wait for the guests to boot because we don't know their
    # dns names.
    $self->_power_on_vms(-vm_list => \@all_vms,
                         -dont_wait => 1);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_all_vms()
# Objective:       Will turn off all vms on this ESX host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_off_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    my (@all_vms) = $self->_get_all_vm_config_files();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->_power_off_vms(-vm_list => \@all_vms);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_all_vms()
# Objective:       Will power cycle all vms on this Host
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_cycle_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    my (@all_vms) = $self->_get_all_vm_config_files();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->_power_cycle_vms(-vm_list => \@all_vms);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_reset_all_vms()
# Objective:       Will reset power on all vms
# Details:         See POD documentation at the end of this file
################################################################################
sub _power_reset_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    my (@all_vms) = $self->_get_all_vm_config_files();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->_power_reset_vms(-vm_list => \@all_vms);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _wait_for_vms_to_shutdown()
# Objective:       This will sleep until all vms have properly shut down
# Details:         See POD documentation at the end of this file
################################################################################
sub _wait_for_vms_to_shutdown {
    my ($self, @vm_config_file_list) = @_;

    # number of times to check state until deciding to force a shutdown
    my $num_retries = 14;
    my $num_vms = scalar(@vm_config_file_list);

    # iterate through the vm_config_file_list until either the retry count is
    # zero, or all VMs have been shutdown.
    while (scalar @vm_config_file_list) {

        # if the number of retries is 0
        if ($num_retries <= 0) {
            # then force shutdown on all config files remaining
            # in @vm_config_file_list
            last;
        }

        # pull this vm off the config file list
        my $vm_config_file = shift @vm_config_file_list;
        $num_vms -= 1;

        # Check the state of this VM
        my $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" getstate";
        my ($status, @output) = $self->_vmware_cmd($cmd);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # if this machine is still on, push it back on the list
        # if this machine is not on, then don't push it
        if (grep { /on/i } @output) {
            push(@vm_config_file_list, $vm_config_file);
        }

        # if num_vms has gone to zero, then update its value, decrement the retry
        # count, and sleep 20 seconds before checking the state of all vms still
        # on the list.
        if ($num_vms <= 0) {
            $num_vms = scalar(@vm_config_file_list);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # force shutdown any VMs that are still in the list
    foreach my $vm_config_file (@vm_config_file_list) {
        my $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" stop hard";
        my ($status, @output) = $self->_vmware_cmd($cmd);

        if ($self->error()) {
            my $errmsg = $self->errmsg();
            # verify that the VM is already off if the stop hard failed.
            $cmd = "/usr/bin/vmware-cmd \"$vm_config_file\" getstate";
            ($status, @output) = $self->_vmware_cmd($cmd);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }

            # if the VM is off then this is not a failure, go to the next vm in the list
            if (grep { /off/i } @output) {
                # clear the error and go to the next vm that is in the shutdown list
                $self->_clear_err();
                next;
            } else {
                # otherwise this is an error because the stop hard command failed,
                # make sure the original error message is restored.
                $self->_set_err($errmsg);
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _wait_for_vms_to_boot()
# Objective:       This will sleep until all VMs have finished booting.
# Details:         See POD documentation at the end of this file
################################################################################
sub _wait_for_vms_to_boot {
    my ($self, @vm_list) = @_;

    my (@ip_list) = ();

    # go through each vm and obtain its ip.
    foreach my $vm (@vm_list) {
        my $tharnhost_hsh = &main::host($vm);
        my $ip = $tharnhost_hsh->{default_ip};
        push(@ip_list, $ip);
    }

    my $num_retries = 10;
    my $num_ips = scalar(@ip_list);

    # cycle through the ip_list, pinging them until they boot up
    while (scalar(@ip_list)) {

        # if our retries are up, then break out of this loop
        if ($num_retries <= 0) {
            last;
        }

        # shift off an ip
        my $ip = shift @ip_list;
        my $vm = shift @vm_list;
        $num_ips -= 1;

        # ping the ip
        my $cmd = "ping -c 1 -w 1 $ip";
        my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 1 and $status != 0) {
            $self->_set_err("Error: ESXHost::_wait_for_vms_to_boot() Command ".
                            "($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # if this ip could not be pinged, push it back on the list.
        if (grep { /100% packet loss/ } @output) {
            push (@ip_list, $ip);
            push (@vm_list, $vm);
        }

        # if num_ips has gone to zero, then update its value to the size
        # of the ip list. Decrement the retry count, and sleep.
        if ($num_ips <= 0) {
            $num_ips = scalar(@ip_list);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # if no ips are left in the list, all vms have booted.
    if (scalar(@ip_list) == 0) {
        $self->sanlog(-msg => "All VMs have booted.");
    } else {
        # then these vm's did not finish booting in time, display a warning
        $self->sanlog(-msg => "WARNING: The following VMs took too long ".
                              "to boot: |@vm_list|. Continuing... ");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_vms_that_were_shutdown()
# Objective:       This will obtain the list of vms that were powered off by
#                  _power_off_vms() or _power_off_all_vms()
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_vms_that_were_shutdown {
    my ($self) = @_;

    $self->_clear_err();

    my (@vms_that_were_shutdown) = @{$self->{vms_that_were_shutdown}};

    return (wantarray()) ? @vms_that_were_shutdown : \@vms_that_were_shutdown;
}

################################################################################
# Subroutine Name: _esx_execute_cmd()
# Objective:       Wrapper to execute a command on the ESX Host
################################################################################
sub _esx_execute_cmd {
    my ($self, %args) = @_;

    my $cmd = "";
    my $retries = undef;
    my $retry_failure_type = "-warn";
    my $timeout;
    my $include_host_name = 0;
    my $reconnect_on_failure = 0;

    # parse arguments
    foreach (keys %args) {
        if (/^-?cmd$/) {
            $cmd = $args{-cmd};
        } elsif (/^-?timeout$/) {
            $timeout = $args{-timeout};
        } elsif (/^-?retries$/) {
            $retries = $args{-retries};
        } elsif (/^-?retry_failure_type$/) {
            $retry_failure_type = $args{-retry_failure_type};
        } elsif (/^-?include_host_name$/) {
            $include_host_name = $args{-include_host_name};
        } elsif (/^-?reconnect_on_failure$/) {
            $reconnect_on_failure = $args{-reconnect_on_failure};
        }
    }

    # if we are communicating via vMA host (remote cli)
    # insert the username and password of the server
    # we are communicating to.
    if ($self->has_vma_host()) {
        my (@cmds) = split(/\s+/, $cmd);
        my $username = $self->{phy_host_username};
        my $password = $self->{phy_host_password};
        my $new_cmd;
        $new_cmd = shift(@cmds);

        # if requested to include the host name or if the vma is not installed
        # add the --server flag to this command
        if ($include_host_name or $self->{vma_installed} == $ClientUtil::FALSE) {
            $new_cmd .= " --server " . $self->{esxi_host_fqdn};
        }


        $new_cmd .= " --username ${username} --password ${password}";
        foreach my $val (@cmds) {
            $new_cmd .= " " .$val
        }
        $cmd = $new_cmd;
    }

    # if requested to retry the command on failure, use execute_retry()
    my ($status, @output);
    if (defined($retries)) {
        ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, $timeout, $retries, $retry_failure_type);
    } else {
        # otherwise default to regular execute()
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $timeout);
    }

    # check @output for known error strings
    my $error_regex = "Invalid host|SOAP request error|Can't open session file|Error while testing status of connection|Runtime error:|failed: Connection failure";
    if (grep(/$error_regex/i, @output)) {
        if ($reconnect_on_failure) {
            # this is usually caused by the vMA losing its shit so reconnect if this happens
            $self->reconnect(-disconnect_existing_conn => 0);
            if ($self->error()) {
                # return a really bad status if we can't reconnect
                return (255);
            }
        }

        # rest for a moment
        sleep 1;

        # run the command again
        if (defined($retries)) {
            ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, $timeout, $retries, $retry_failure_type);
        } else {
            # otherwise default to regular execute()
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $timeout);
        }

        # if output still matches the known error strings, modify $status be non zero
        if (grep(/$error_regex/i, @output) and $status == 0) {
            $status = "255";
        }
    }

    return ($status, @output);
}

################################################################################
# Subroutine Name: _convert_to_datastore_path()
# Objective:       Convert the given filepath to a valid ESX datastore path
################################################################################
sub _convert_to_datastore_path {
    my ($self, %args) = @_;

    my $file_path;
    my $format = "datastore_path";
    my $use_uuid = -1;

    # parse arguments
    foreach (keys %args) {
        if (/^-?file_path$/) {
            if (not length ($file_path = $args{$_})) {
                $self->_set_err("Invalid value for -file_path => $file_path");
                return "";
            }
        } elsif (/^-?format$/) {
            if (not length ($format = $args{$_})) {
                $self->_set_err("Invalid value for -format => $format");
                return "";
            }
        } elsif (/^-?use_uuid$/) {
            if (not length ($use_uuid = $args{$_})) {
                $self->_set_err("Invalid value for -use_uuid => $use_uuid");
                return "";
            }
        }
    }

    # setup a callback to handle conversions between UUID and datastore label
    my $callback = sub {
        my ($name) = @_;
        if ($use_uuid == 1) {
            # convert datastore_name to datastore_uuid
            if (defined($self->mpdevmgr())) {
                my ($dev) = $self->mpdevmgr()->get_filtered_devices(-os_specific_data_filter => {datastore_label => $name});
                if (not $self->mpdevmgr()->error() and defined($dev)) {
                    $name = $dev->get_os_specific_data("datastore_uuid");
                } else {
                    $self->mpdevmgr()->_clear_err();
                    # the mpdevmgr did not have any devices that matched this datastore name
                    # could be that the file is stored on a non NETAPP LUN, manually lookup the uuid
                    my $cmd = "esxcfg-scsidevs --vmfs | grep \"$name\"";
                    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                    if ($status > 0) {
                        $self->_set_err("Unable to convert obtain datastore uuid for " .
                                        "datastore $name. Failed running command ($cmd)");
                        return undef;
                    }
                    foreach my $line (@output) {
                        # find the datastore name based on the uuid
                        if ($line =~ /^\s*\S+\s+\S+\s+(\S+)\s+\S+\s+${name}\s*$/) {
                            $name = $1;
                            last;
                        }
                    }
                }
            }
        } elsif ($use_uuid == 0) {
            # convert datastore_uuid to datastore_name
            if (defined($self->mpdevmgr())) {
                my ($dev) = $self->mpdevmgr()->get_filtered_devices(-os_specific_data_filter => {datastore_uuid => $name});
                if (not $self->mpdevmgr()->error() and defined($dev)) {
                    $name = $dev->get_os_specific_data("datastore_label");
                } else {
                    $self->mpdevmgr()->_clear_err();
                    # the mpdevmgr did not have any devices that matched this UUID,
                    # could be that the file is stored on a non NETAPP LUN, manually lookup the uuid
                    my $cmd = "esxcfg-scsidevs --vmfs | grep \"$name\"";
                    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                    if ($status > 0) {
                        $self->_set_err("Unable to convert obtain datastore name for " .
                                        "datastore uuid $name. Failed running command ($cmd)");
                        return undef;
                    }
                    foreach my $line (@output) {
                        # find the datastore name based on the uuid
                        if ($line =~ /^\s*\S+\s+\S+\s+${name}\s+\S+\s+(.+)\s*$/) {
                            $name = $1;
                            last;
                        }
                    }
                }
            }
        } else {
            # otherwise dont do any conversion (default)
        }

        return $name;
    };

    # Datastore Path format: [datastore_name] path/to/file
    # File Path format: /vmfs/volumes/datastore_name/path/to/file

    # convert the file path
    if ($format =~ /datastore_path/i) {
        if ($file_path =~ /^\s*\/vmfs\/volumes\/(.+?)\/(.+)\s*$/) {
            my $datastore_name = $1;
            my $subdir_file = $2;

            $datastore_name = $callback->($datastore_name);
            if ($self->error()) {
                return "";
            }
            my $datastore_path = "[$datastore_name] $subdir_file";

            return $datastore_path;
        } elsif ($file_path =~ /^\s*\[(.+?)\]\s*(.+)/) {
            my $datastore_name = $1;
            my $subdir_file = $2;

            # in this case it is already in the proper datastore format
            # just return it
            $datastore_name = $callback->($datastore_name);
            if ($self->error()) {
                return "";
            }
            my $datastore_path = "[$datastore_name] $subdir_file";

            return $datastore_path;
        } else {
            $self->_set_err("Unable to parse datastore path from file_path '$file_path'");
            return "";
        }
    } elsif ($format =~ /file_path/i) {
        if ($file_path =~ /^\s*\/vmfs\/volumes\/(.+?)\/(.+)\s*$/) {
            my $datastore_name = $1;
            my $subdir_file = $2;
            # in this case it is already in the proper datastore format
            # just return it
            $datastore_name = $callback->($datastore_name);
            if ($self->error()) {
                return "";
            }
            my $datastore_path = "/vmfs/volumes/${datastore_name}/${subdir_file}";

            return $datastore_path;
        } elsif ($file_path =~ /^\s*\[(.+?)\]\s*(.+)/) {
            my $datastore_name = $1;
            my $subdir_file = $2;

            $datastore_name = $callback->($datastore_name);
            if ($self->error()) {
                return "";
            }
            my $datastore_path = "/vmfs/volumes/${datastore_name}/${subdir_file}";

            return $datastore_path;
        } else {
            $self->_set_err("Unable to parse datastore path from file_path '$file_path'");
            return "";
        }
    }
}

1;

=pod

=head1 Package Name

ESXHost.pm

=head1 Author

Robert Hunter

=head1 Description

For description of methods see B<San::SanClient::Host.pod>.

=cut

=item Event Parameters

    Required Event Parameters:
    lun_target_reset()

=item Examples

  # get all netapp test lun naa ids.first the function will collect all scsi luns in the server ( esxcfg-scsidevs -c )
  # then it will filter naa ids of devices which is of NETAPP type and return the value to lun_target_reset()
  # lun_target_reset() funtion will check for the reset type ( lun reset or target reset ) . If reset type is other than
  # lun reset or target reset function will throw error.
  # get_os_version() will check for ESX/ESXi version and will connect to ESX/ESXi host directly and will start reset operation


=back

=cut
