# Copyright (c) 2007-2011 NetApp Inc.
## All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com, rawat@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::LinuxHost;

# Package Requirements
######################
use Net::Ping;
use Tie::IxHash;
use File::Basename;

use San::SanClient;
use San::SanLog;
use San::SanUtil;
use Error;
use Tharn;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanClient SanCache);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new LinuxHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, %hRaw, $iscsi_type);
    my ($self, $conn, $nfs_mount, $cmd, $status, $hosttype, $protocol, @output, $service, $obj_template);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: LinuxHost::new()" . " Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
             if (not length($self = $args{$_})) {
                $self->_set_err("Error: LinuxHost::new()" . " Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: LinuxHost::new()" . " Invalid value for -protocol: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?iscsi_type/) {
            $iscsi_type = $args{$_};
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: LinuxHost::new()", "Invalid value for argument '-conn'");
        return $self;
    }

    if ($protocol != /fcp|iscsi/i) {
        $self->_set_err("Error: LinuxHost::new()", "Invalid value for argument '-protocol'" .
                        " It can be either fcp or iscsi");
        return $self;
    }

    $self->{protocol} = $protocol;
    $self->{conn}     = $conn;
    $self->{nfs_mount} = undef;

    # determine if this is a virtual machine
    $cmd = "dmidecode |  grep -i -A3 \"System Information\"";
    my (undef, @dmidecode_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@dmidecode_output)) {
        $self->{dmidecode_system_info} = \@dmidecode_output;
    } else {
        $self->{dmidecode_system_info} = [];
    }

    $self->store_OS_major_and_minor_revisions();
    if ($self->error) {
        return $self;
    }

    $self->sanlog(-msg => "OS version: " . $self->{major_revision});

    # store TEST_BIN directory
    $self->store_test_bin_dir();
    if ($self->error) {
        return $self;
    }

    # There is difference between iscsi CMDs for rhel4,ol4 & sles9 and
    # sles10,ol5 and rhel5 so We need to store iSCSI
    # version in a variable to differentiate between the iSCSI versions.
    # Depending on the OS version we can use
    # the iscsi CMDs. before sles10, ol5 and rhel5 we are going
    # to use commands for "linux-iscsi" and from rhel5,ol5
    # sles10 and we are going to use cammands "open-iscsi".
    # If OS version are before sles10, ol5 and rhel5 the following
    # variable will have value "linux-iscsi" else it will have value "open-iscsi"
    if ($self->{major_revision} =~ /rhel4|sles9/) {
        $self->{iscsi}->{iscsi_version} = "linux-iscsi";
    } else {
        $self->{iscsi}->{iscsi_version} = "open-iscsi";
    }
    my $iscsi_version = $self->{iscsi}->{iscsi_version};

    # By default 'raw' module is not loaded on SLES. We need to load
    # it. It will create master raw device '/dev/rawctl'.
    # On Redhat there is no raw module and device /dev/rawctl will already
    # be there
    if(($self->hosttype() =~ /suse|sles/i)) {
        $cmd = "modprobe raw";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxHost::_new() Command ($cmd) returned a bad status ($status)");
            return $self;
        }
    }

    #  Create, load, and manage hash of raw device bindings.
    #  Keep load order.
    tie %hRaw, "Tie::IxHash";

    $self->_load_raw_bindings(\%hRaw);
    if ($self->error) {
        return $self;
    }
    $self->{raw_bindings} = \%hRaw;

    # store client kernel version
    $cmd = "uname -r";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    $self->{revision} = $output[0];

    # need to load config params right now so we can access them
    # for decision on software or hardware initiator
    $self->{iscsi_type} = $iscsi_type;
    $self->_load_config_params();
    if ($self->error()) {
        return $self;
    }

    if ($self->{protocol} =~ /iscsi/i and $self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/i) {
        my $running = $self->_iscsi_service(-action => "status");
        if ($self->error) {
            return $self;
        }

        if ($running =~ /offline/) {
            # iscsi will not start if iscsi.conf doesnt contain anything which is usually the case with newly install system
            # HACK : Add a non-empty line (unique) to /etc/iscsi.conf file and iscsi will start
            # We will remove this line during iscsi configure.
            # This is only for linux-iscsi because in open-iscsi case iscsi.conf will not be empty even
            # on the newly installed system
            if ($self->{iscsi}->{iscsi_version} =~ /linux-iscsi/) {
                $cmd = "cat /etc/iscsi.conf |grep -v \"^#\"|grep -v \"^\$\"";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if (scalar(@output) == 0) {
                    my $line = "remove_this_line_during_iscsi_configuration";
                    $cmd = "echo $line >> /etc/iscsi.conf";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Error: LinuxHost::_new() unable to add $line to /etc/iscsi.conf file");
                        return $self;
                    }
                }
            }

            $self->_iscsi_service(-action => 'start');
            if ($self->error) {
                return $self;
            }
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  Linux hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    my ($cmd, $status, @output);
    my ($transfer_length_granularity, $transfer_length);
    my ($minimum_io_size, $optimal_io_size, $block_size, $physical_block_size, $logical_block_size, $f_block_size, $product, $start_block_value);
    my (@prompts_answers, $filer_scsidev_mapping, $devmgrobj);
    my ($match, $before, $after, $pafter, $prompt, @waitstring, $msg);
    my ($scsi_dev, $mpath_dev, @devices, $device_obj, $type, $device_with_path);

    $self->sanlog(-msg => "Verifying BLOCK LIMITS VPD (0xb0) details are reflected on Linux hosts for NetApp LUNs");

    # filer_scsidev_mapping is a hash that has scsi devices pushed into corresponding filers
    $filer_scsidev_mapping = {};
    $self->_clear_err();

    # Prerequisite to verify blocks limits VPD
    # 1. Host OS: RHEL6/OL6 and SLES11 SP1 onwards
    # 2. ONTAP: BR.1 (8.0.1) onwards
    # 3. Target should conform to SPC-3 (SCSI Primary Commands-3)

    if (($self->{major_revision} =~ /rhel6|ol6/i) or ($self->{major_revision} =~ /sles11/i and $self->{minor_revision} >= 2)) {
        # Fail if sg3_utils is not installed. This package is necessary to verify blocks limits VPD
        if ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE) {
            $self->_set_err("sg3_utils is not installed.\n\tThis package is necessary to verify " .
                            "blocks limits VPD");
            return $ClientUtil::FALSE;
        }

        # Use sanlun lun show to get a list of scsi devices
        my $path = $self->get_hu_bin_path();
        $cmd = $path . "/sanlun lun show all";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 203) {
            # No controller LUNs are available. Skip as there is no device to verify
            $self->sanlog(-msg => "No controller LUNs are available. Blocks limits VPD " .
                                  "verification is skipped");
            return $ClientUtil::TRUE;
        } elsif ($status != 0) {
            $self->_set_err("Command $cmd returned a bad status");
            return $ClientUtil::FALSE;
        }

        # sanlun lun show output is parsed to get filers and their corresponding scsi devices
        foreach my $line (@output) {
            $line =~ s/^\s+//;    # strip leading white space
            my ($filer, $lun_path, $sdevice) = split(/\s+/, $line);
            $filer =~ s/\://;
            if ($lun_path !~ /^\/vol\//) {
                next;
            }
            push @{ $filer_scsidev_mapping->{$filer}->{devices} }, $sdevice;
        }

        # Checking whether the targets satisfies the Prerequisite to verify blocks limits VPD
        foreach my $filer_name (keys(%$filer_scsidev_mapping)) {
            $scsi_dev = $filer_scsidev_mapping->{$filer_name}->{devices}->[0];
            $cmd = "sg_inq $scsi_dev";

            # output of this cmd looks like this:
            # standard INQUIRY:
            #  PQual=0  Device_type=0  RMB=0  version=0x05  [SPC-3]
            #  [AERC=0]  [TrmTsk=0]  NormACA=1  HiSUP=1  Resp_data_format=2
            #  SCCS=0  ACC=0  TPGS=1  3PC=0  Protect=0  BQue=0
            #  EncServ=0  MultiP=1 (VS=0)  [MChngr=0]  [ACKREQQ=0]  Addr16=0
            #  [RelAdr=0]  WBus16=0  Sync=0  Linked=0  [TranDis=0]  CmdQue=1
            #  [SPI: Clocking=0x0  QAS=0  IUS=0]
            #  length=117 (0x75)   Peripheral device type: disk
            # Vendor identification: NETAPP
            # Product identification: LUN
            # Product revision level: 8010
            # Unit serial number: Hna3T4WaOCKV

            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }

            my ($filer_major_rev1, $filer_major_rev2, $filer_minor_rev1);
            $filer_major_rev1 = $filer_major_rev2 = $filer_minor_rev1 = 0;
            my $spc3_flag = 0;

            # parse the standard inquiry cmd output
            foreach my $line (@output) {
                if ($line =~ /Product revision level/i) {
                    # get the filer version
                    if ($line =~ /\:\s*(\d)(\d)(\d)/) {
                        $filer_major_rev1 = $1;
                        $filer_major_rev2 = $2;
                        $filer_minor_rev1 = $3;
                    }
                } elsif ($line =~ /SPC-3/) {
                $spc3_flag = 1;
                }
            }

            # check whether target conforms to SPC3
            if ($spc3_flag == 0) {
                $self->_set_err("BLOCK LIMITS VPD is not supported for filer ($filer_name) which is not conformed to SPC3");
                return $ClientUtil::FALSE;
            }

            # Filer version should be greater than 8.0.1 to verify blocks limits vpd
            if (($filer_major_rev1 > 8) or ($filer_major_rev1 == 8 and $filer_major_rev2 > 0)
            or ($filer_major_rev1 == 8 and $filer_major_rev2 == 0 and $filer_minor_rev1 >=1)) {
                # Continue with next filer
                next;
            } else {
                $self->_set_err("BLOCK LIMITS VPD is not supported for filer ($filer_name) which ONTAP version less than 8.0.1");
                return $ClientUtil::FALSE;
            }
        }

        # Getting Device Manager object
        $devmgrobj = $self->{obj}{devmgr};
        $devmgrobj = $self->create_devmgr() if (not defined $devmgrobj);
        if ($self->error()) {
            $self->_set_err("Failed to create Device Manager Object");
            return $ClientUtil::FALSE;
        }

        # Getting multipath devices
        @devices = $devmgrobj->get_devices();
        if ($devmgrobj->error) {
            $self->_set_err($devmgrobj->errmsg());
            return $ClientUtil::FALSE;
        }

        foreach $device_obj (@devices) {
            if ($self->{mp_type} ne "" && $self->{mp_type} !~ /^none$/i) {
                $mpath_dev = $device_obj->{name};
                $scsi_dev = $self->{mp}->{mapping}->{$mpath_dev}->{sdX};
            } else {
                $scsi_dev = $device_obj->{name};
            }
            $scsi_dev = "/dev/"."$scsi_dev";

            # Getting detailed information of a scsi device
            $cmd = "sg_inq -p 0xb0 $scsi_dev";

            # output of this cmd looks like this,
            # sg_inq -p 0xb0 /dev/sdb
            # VPD INQUIRY: Block limits page (SBC)
            #   Optimal transfer length granularity: 8 blocks
            #   Maximum transfer length: 8388607 blocks
            #   Optimal transfer length: 128 blocks
            #   Maximum prefetch, xdread, xdwrite transfer length: 0 blocks
            #   Maximum unmap LBA count: 499712
            #   Maximum unmap block descriptor count: 4
            #   Optimal unmap granularity: 8
            #   Unmap granularity alignment valid: 1
            #   Unmap granularity alignment: 0

            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }

            # parse the output of "sg_inq -p 0xb0 /dev/sdX" cmd
            foreach my $line (@output) {
                if ($line =~ /Optimal transfer length granularity/i) {
                    # Obtain the value of Optimal transfer length granularity
                    if ($line =~ /(\d+)/) {
                        $transfer_length_granularity = $1;
                    }
                } elsif ($line =~ /Optimal transfer length/i) {
                         # Obtain the value of Optimal transfer length
                         if ($line =~ /(\d+)/) {
                             $transfer_length = $1;
                         }
                 }
            }

            if (not defined($transfer_length_granularity)) {
                $self->_set_err("Unable to find Optimal transfer length granularity value " .
                                "from the output of cmd ($cmd)");
                return $ClientUtil::FALSE;
            }

            if (not defined($transfer_length)) {
                $self->_set_err("Unable to find Optimal transfer length value " .
                                "from the output of cmd ($cmd)");
                return $ClientUtil::FALSE;
            }

            # Remove "/dev/" from "/dev/sdX"
            $scsi_dev =~ s/^\/dev\///;

            # Obtain the value of physical block size
            $cmd = "cat /sys/block/$scsi_dev/queue/physical_block_size";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }
            $physical_block_size = $output[0];

            # Obtain the value of logical block size
            $cmd = "cat /sys/block/$scsi_dev/queue/logical_block_size";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }
            $logical_block_size = $output[0];

            # Obtain the value of minimum io size
            $cmd = "cat /sys/block/$scsi_dev/queue/minimum_io_size";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }
            $minimum_io_size = $output[0];

            # Obtain the value of optimal io size
            $cmd = "cat /sys/block/$scsi_dev/queue/optimal_io_size";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }
            $optimal_io_size = $output[0];

            # Here, minimum_io_size corresponds to the 'Optimal transfer length granularity'
            # So, minimum IO size should be equal to product of Optimal Transfer length granularity
            # and physical block size
            $msg = "";
            $product = $transfer_length_granularity * $logical_block_size;
            if ($minimum_io_size != $product) {
                if ($minimum_io_size == $physical_block_size) {
                    $msg = "Minimum IO size scaled upto Physical Block Size \n".
                           "Physical Block Size = $physical_block_size";
                }
                $self->_set_err("Blocks limits VPD verification: Minimum IO size doen't match with " .
                                "Optimal transfer length granularity for device $scsi_dev.\n" .
                                "Minimum IO size = $minimum_io_size \n" .
                                "Optimal transfer length granularity = $product \n".
                                "$msg");
                return $ClientUtil::FALSE;
            }

            # Optimal_io_size corresponds to the 'Optimal transfer length'.
            # So, optimal IO size should be equal to product of Optimal Transfer length
            # and logical block size
            $product = $transfer_length * $logical_block_size;
            if ($optimal_io_size != $product) {
                $self->_set_err("Blocks limits VPD verification: Optimal IO size doen't match with " .
                                "Optimal transfer length for device $scsi_dev.\n" .
                                "Optimal IO size = $optimal_io_size\n" .
                                "Optimal transfer length = $product");
                return $ClientUtil::FALSE;
            }

            # strings that may appear while executing user interactive fdisk command
            @prompts_answers = (
                ".*p   primary .*" ,
                "Partition number .*",
                ".*First sector.*",
                "First cylinder .*",
                ".*for help.*",
            );

            # While executing user interactive fdisk command, we can obtain default offset value when these strings appear
            @waitstring = (
                ".*Device Boot.*",
            );

            $cmd = "fdisk /dev/$scsi_dev";

            # output of this cmd looks like this,
            # Device contains neither a valid DOS partition table, nor Sun, SGI or OSF disklabel
            # Building a new DOS disklabel with disk identifier 0xa2fd5b09.
            # Changes will remain in memory only, until you decide to write them.
            # After that, of course, the previous content won't be recoverable.

            # Warning: invalid flag 0x0000 of partition table 4 will be corrected by w(rite)

            # WARNING: DOS-compatible mode is deprecated. It's strongly recommended to
            #          switch off the mode (command 'c') and change display units to
            #          sectors (command 'u').

            # Command (m for help): c
            # DOS Compatibility flag is not set

            # Command (m for help): u
            # Changing display/entry units to sectors

            # Command (m for help): n
            # Command action
            #      e   extended
            #      p   primary partition (1-4)
            # p
            # Partition number (1-4): 1
            # First sector (2048-4194303, default 2048):
            # Using default value 2048
            # Last sector, +sectors or +size{K,M,G} (2048-4194303, default 4194303):
            # Using default value 4194303

            # Command (m for help): p
            # Disk /dev/sdb: 2147 MB, 2147483648 bytes
            # 67 heads, 62 sectors/track, 1009 cylinders, total 4194304 sectors
            # Units = sectors of 1 * 512 = 512 bytes
            # Sector size (logical/physical): 512 bytes / 512 bytes
            # I/O size (minimum/optimal): 4096 bytes / 65536 bytes
            # Disk identifier: 0xa2fd5b09

            # Device Boot      Start         End      Blocks   Id  System
            # /dev/sdb1            2048     4194303     2096128   83  Linux

            (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, $cmd, '-prompt', \@prompts_answers);
            if ($status != 0) {
                $self->_set_err("Command $cmd returned a bad status");
                return $ClientUtil::FALSE;
            }

            #sequence of user interactive command to be given while giving fdisk command
            if ($prompt =~ /.*for help.*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "c", '-prompt', \@prompts_answers);
            }
            if ($prompt =~ /.*for help.*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "u", '-prompt', \@prompts_answers);
            }
            if ($prompt =~ /.*for help.*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "n", '-prompt', \@prompts_answers);
            }
            if ($prompt =~ /.*p   primary .*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "p", '-prompt', \@prompts_answers);
            }
            if ($prompt =~ /Partition number .*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "1", '-prompt', \@prompts_answers);
            }
            if ($prompt =~ /First .*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "\n", '-prompt',  \@prompts_answers);
            }
            if ($prompt =~ /.*for help.*/) {
               ($match,$before,$after,$prompt) = &Tharn::say($self->{conn}, "p",'-wait', \@waitstring, '-prompt', \@prompts_answers);
            }
            $pafter = $after;
            if ($prompt =~ /.*for help.*/) {
               (undef,undef,undef,$prompt) = &Tharn::say($self->{conn}, "q", '-prompt', \@prompts_answers);
            }

            $pafter =~ s/^\s+//;
            my ($devic, $start_sector, $end, $blocks, $id) = split(/\s+/, $pafter);
            if (not defined($start_sector)) {
                $self->_set_err("Unable to find default offset value " .
                                "from the output of cmd ($cmd)");
                return $ClientUtil::FALSE;
            }

            $start_block_value = (($start_sector * $logical_block_size) % $minimum_io_size);
            if ($start_block_value != 0) {
                $self->_set_err("Blocks limits VPD verification: The default sector for the first partition " .
                                "does not match.\n" .
                                "default offset = $start_sector \n" );
                return $ClientUtil::FALSE;
            }

            # Perform filesystem check, if the device is a multipath device
            if ($self->{mp_type} =~ /DMMP/i) {
                $device_with_path = "/dev/mapper/" . $mpath_dev;

                # Creating a filesystem on the multipath device
                $device_obj->create_fs();
                if ($device_obj->error()) {
                    $self->_set_err("Unable to create file system on device", $device_obj->{name});
                    return $ClientUtil::FALSE;
                }

                # tune2fs command is used to verify block size of the file system created
                $cmd = "tune2fs -l $device_with_path";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("command $cmd returned a bad status ");
                    return $ClientUtil::FALSE;
                }

                # parse tune2fs command to get block size
                foreach my $line (@output) {
                    if ($line =~ /Block size/i) {
                        # Obtain the value of Block size
                        if ($line =~ /(\d+)/) {
                            $f_block_size = $1;
                        }
                    }
                }

                if (not defined($f_block_size)) {
                    $self->_set_err("Unable to find block size value " .
                                    "from the output of cmd ($cmd)");
                    return $ClientUtil::FALSE;
                }
                $product = $transfer_length_granularity * $logical_block_size;
                if ($f_block_size != $product) {
                    $self->_set_err("Blocks limits VPD verification:File System block  size doen't match with " .
                                    "Minimum IO size = $minimum_io_size \n" );
                    return $ClientUtil::FALSE;
                }
            }
        }
   } else {
        $self->_set_err("BLOCK LIMITS VPD is not supported for the Host ($self->{host})");
        return $ClientUtil::FALSE;
   }

   $self->sanlog(-msg => "BLOCK LIMITS VPD verification has passed");

   return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {

    my ($self, %args) = @_;

    my (@mount_points);
    my ($cmd, $status, @output, $hostname);

    $self->_clear_err();

    $hostname = $self->hostname();
    # parse arguments
    foreach (keys %args) {
       if (/-?mount_points/) {
           if (not scalar(@mount_points = @{$args{$_}})) {
               $self->_set_err("Invalid value for argument '-mount_points'");
               return $ClientUtil::FALSE;
           }
       }
    }
    if (not scalar(@mount_points)) {
        $self->_set_err("Missing required argument '-mount_points'");
        return $ClientUtil::FALSE;
    }
    if (($self->{major_revision} =~ /rhel6/i and $self->{minor_revision} >= 4) or
        ($self->{major_revision} =~ /sles11/i and $self->{minor_revision} >= 3)) {

        &SanUtil::nate_log('-comment', "Running fstrim(space reclaim) command on all mount points for client '$hostname'");

        $cmd = "/bin/mount";

        #/dev/mapper/vg_ibmx336210112-lv_root on / type ext4 (rw)
        #proc on /proc type proc (rw)
        #sysfs on /sys type sysfs (rw)
        #devpts on /dev/pts type devpts (rw,gid=5,mode=620)
        #tmpfs on /dev/shm type tmpfs (rw,rootcontext="system_u:object_r:tmpfs_t:s0")
        #/dev/sda1 on /boot type ext4 (rw)
        #/dev/mapper/vg_ibmx336210112-lv_home on /home type ext4 (rw)
        #none on /proc/sys/fs/binfmt_misc type binfmt_misc (rw)
        #sunrpc on /var/lib/nfs/rpc_pipefs type rpc_pipefs (rw)
        #btcinf01.eng.btc.netapp.in:/vol/software/arch/x86_64-redhat-rhel5 on /usr/software type nfs (rw,addr=10.72.10.93)
        #/dev/mapper/n_vg-nvol--1 on /mnt/qa/n_vg/nvol-1 type ext4 (rw)
        #/dev/mapper/n_vg-nvol--2 on /mnt/qa/n_vg/nvol-2 type xfs (rw)
        #/dev/mapper/n_vg-nvol--3 on /mnt/qa/n_vg/nvol-3 type ext4 (rw)

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return $ClientUtil::FALSE;
        }

        foreach my $mount_point (@mount_points) {

           my ($line) = grep (/\s+$mount_point\s+/, @output);
           $line =~ s/^\s+//;    # strip leading white space
           my @dev_arr = split (/\s+/, $line);
           my $fs_type = $dev_arr[4];

           if ($line =~ /ext4|xfs|btrfs/i) {
               my $cmd  = "time fstrim -v $mount_point";
               my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
               if ($status > 0) {
                   $self->_set_err("Command ($cmd) returned a bad status ($status)");
                   return $ClientUtil::FALSE;
               }
            } else {
              &SanUtil::nate_log('-warn', "Client Offline discard is not supported for filesystem of type $fs_type");
            }
        }
    } else {
        &SanUtil::nate_log('-warn', "Client Offline discard is not supported for client '$hostname'");
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
    if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    if ($self->{protocol} =~ /nfs/i) {
        # No need to do any validation for NFS protocol for now
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # if mp_type is not set, do not validate currently
    my $mp_type = $self->{mp_type};
    if (not defined($mp_type) or $mp_type eq "") {
        $self->sanlog(-msg => "Multipathing type (mp_type) is not set. Not validating configuration");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # get the MP stack info
    my $veritas_version = "";
    my $veritas_version_sp = "";
    if ($mp_type =~ /dmmp/i) {
        $mp_type = "dmmp";
    } elsif ($mp_type =~ /dmp|vrts|veritas/i) {
        $mp_type = "veritasdmp";
        # Get the MP object and get the veritas version from it.
        if (exists($self->{mp}) and ref($self->{mp}) =~ /dmp/i) {
            $veritas_version = $self->{mp}->{version};

            # See if we have an SP installed
            if (defined($self->{mp}->{sp}) and $self->{mp}->{sp} ne "") {
                $veritas_version_sp = $self->{mp}->{sp};
                $veritas_version_sp =~ s/SP//g;
            } else {
                $veritas_version_sp = 0;
            }
        } else {
            $self->_set_err("MP object either doesn't exists or is not a DMP object");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
    } elsif ($mp_type =~ /none/i) {
        $mp_type = "none";
    } else {
        $self->_set_err("Unknown mp_type ($mp_type) for Linux. Failing validation");
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    # Make sure the the restore daemon is running for veritas configs only
    if ($mp_type eq "veritasdmp") {
        if ($veritas_version >= 5.1) {
            # We want to look at the vxdmpadm gettune all output
            my $cmd  = "/sbin/vxdmpadm gettune all";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # Now let's check for some settings
            # vxdmpadm gettune all
            # Tunable                           Current Value  Default Value
            # ------------------------------    -------------  -------------
            # dmp_restore_state                   enabled          enabled
            # dmp_fast_recovery                        on               on
            # dmp_log_level                             4                1
            # dmp_lun_retry_timeout                   300                0
            # dmp_path_age                            120              300
            # dmp_probe_idle_lun                      off               on
            # dmp_restore_interval                     60              300
            # dmp_restore_policy             check_disabled   check_disabled
            # dmp_monitor_fabric                       on               on
            my @errors_found = ();
            foreach my $line (@output) {
                if ($line =~ /^dmp_restore_state\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_state'") if ($1 ne "enabled");
                } elsif ($line =~ /^dmp_fast_recovery\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_fast_recovery'") if ($1 ne "on");
                } elsif ($line =~ /^dmp_log_level\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_log_level'") if ($1 != 4);
                } elsif ($line =~ /^dmp_lun_retry_timeout\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_lun_retry_timeout'") if ($1 != 300);
                } elsif ($line =~ /^dmp_probe_idle_lun\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_probe_idle_lun'") if ($1 ne "off");
                } elsif ($line =~ /^dmp_resore_interval\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_resore_interval'") if ($1 != 60);
                } elsif ($line =~ /^dmp_restore_policy\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_policy'") if ($1 ne "check_disabled");
                } elsif ($line =~ /^dmp_monitor_fabric\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_monitor_fabric'") if ($1 ne "on");
                }
            }

            # Now check out the settings that are 5.1SP1 or later
            if ($veritas_version >= 6.0 or ($veritas_version == 5.1 and $veritas_version_sp > 0)) {
                # Tunable                           Current Value  Default Value
                # ------------------------------    -------------  -------------
                # dmp_path_age                            120              300
                foreach my $line (@output) {
                    if ($line =~ /^dmp_path_age\s+(\S+)\s+/) {
                        push(@errors_found, "Invalid value '$1' for 'dmp_path_age'") if ($1 != 120);
                    }
                }
            }

            if (scalar(@errors_found)) {
                my $errors = join("\n", @errors_found);
                $self->sanlog(-msg => $errors);
                $self->_set_err("There were one or more tuneables set improperly, please verify the " .
                                "settings in 'vxdmpadm gettune all'");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        } else {
            # Run vxdmpadm stat restored
            my $cmd  = "/usr/sbin/vxdmpadm stat restored";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # We want to check the output and look for the following:
            # "The number of daemons running : 1".  The number should not be 0.
            if (grep(/daemons running\s+\:\s+0$/, @output) > 0) {
                $self->_set_err("Veritas Restore Daemon does not appear to be running");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        }
    }

    if ($self->{major_revision} =~ /rhel6|rhel7|fedora|ol6/i) {
        # Check if abrt package is installed. This abrt daemon will capture coredumps
        # if any application fails. These dumps will be helpful to debug if some daemon
        # crashes. So, it is really useful to have this package.
        if ($self->_is_rpm_installed(-rpm => "abrt") == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "abrt package is not installed. This abrt daemon will capture coredumps " .
                                  "if any application fails. It is useful to have this package");
        } else {
            # if the package is installed, then check whether daemon is running or not
            my ($cmd, $status, @output);
            $status = $self->_is_service_running(-service => "abrtd");
            if ($status == $ClientUtil::TRUE) {
                # daemon is running
                $self->sanlog(-msg => "abrt daemon is running...");
            } else {
                # start the daemon if it is not running
                if ($self->{major_revision} =~ /rhel7|fedora/) {
                    $cmd = "systemctl start abrtd.service";
                } else {
                    $cmd = "/etc/init.d/abrtd start";
                }
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    &SanUtil::nate_log('-warn', "unable to start abrt daemon");
                } else {
                    $self->sanlog(-msg => "Started abrt daemon and it is running...");
                }
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  store_test_bin_dir
# Objective:       Populate TEST_BIN directory based on system architecture and
#                  store in host object
################################################################################
sub store_test_bin_dir {
    my ($self) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);

    unless (defined($self->{arch})) {
        # get and store the architecture of the system
        $cmd = "uname -p";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
        $self->{arch} = $output[0];
    }

    # determine TEST_BIN based on system architecture
    if ($self->{arch} =~ /ia64/i) {
        $self->{TEST_BIN} = "/x/eng/localtest/arch/ia64-linux-rhel5/bin";
    } elsif (($self->{arch} =~ /x86_64/i) and ($self->{major_revision} !~ /sles10/i)) {
        # 64bit dt in the following directory is not working on SLES10 series
        # and it is tracked on burt610089.
        $self->{TEST_BIN} = "/x/eng/localtest/arch/x86_64-redhat-rhel5/bin";
    } else {
        $self->{TEST_BIN} = "/x/eng/localtest/arch/i386-redhat-linux8/bin";
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    $self->_clear_err();

    # Get the kernel state. Kernel state of CLIENT can only be gather
    # in the case where NATE server is running Linux(RedHat)
    if ($self->{client_linux_enable_sysrq_dump} == 1
        and (&SanUtil::get_NATE_server_OS_type() =~ /linux/i)) {
        $self->_get_client_kernel_state();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{client_linux_enable_sysrq_dump} == 0) {
        $self->sanlog(-msg=> "SysRq dump is not requested by user");
    } else {
        $self->_set_err("Failed to gather \'SysRq\' info. " .
                        "OS on NATE server is not Linux. Make sure you " .
                        "run your test on NATE server which has Linux on it.");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  configure_remote_syslog_on_client
# Objective: make changes in the configuration file of logging service, to enable
#            remote logging.
################################################################################
sub configure_remote_syslog_on_client {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@output, $status, $cmd, @syslog_conf_content, @syslog_conf_new_content, $conn);
    my ($remote_server_ip, $syslog_service, $syslog_conf_file, $rule, @ps_output, $pid, $line_escaped, $escape);

    foreach (keys %args) {
        if (/-?remote_server_ip/) {
            if (not length($remote_server_ip = $args{$_})) {
                $self->_set_err("Invalid value for -remote_server_ip: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    unless (defined($remote_server_ip)) {
        # NATE controller server will be remote syslog server
        my $NATE_server_reachable = 0;

        # Get ip of NATE controller server
        $cmd = "ifconfig";
        ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to run command \'$cmd\' on NATE Server");
            return $ClientUtil::FALSE;
        }

        foreach my $line (@output) {
            # skip the newlines
            next if ($line =~ /^$/);

            # Get the IP address of NATE server
            if ($line =~ /inet\s+addr:\s*(\d+\.\d+\.\d+\.\d+)\s+/) {
                $remote_server_ip = $1;
                if (&SanUtil::ping_host($remote_server_ip) == $ClientUtil::FALSE) {
                    # Interface not pingable go to next line
                    next;
                } else {
                    # Exit as soon as we get ip address of NATE Server that is pingable
                    $NATE_server_reachable = 1;
                    last;
                 }
            }
        }

        # Not able to ping NATE server. Fail out
        if ($NATE_server_reachable == 0) {
            $self->_set_err("Unable to ping NATE server through any of the interfaces");
            return $ClientUtil::FALSE;
        }
    }

    # Find out which syslog service is running on the client
    if ($self->_is_rpm_installed(-rpm => "rsyslog") == $ClientUtil::TRUE) {
        $syslog_service = "rsyslog";
        $self->{syslog_service} = "rsyslog";
    } elsif ($self->_is_rpm_installed(-rpm => "sysklogd") == $ClientUtil::TRUE) {
        $syslog_service = "syslogd";
        $self->{syslog_service} = "syslogd";
    } elsif ($self->_is_rpm_installed(-rpm => "syslog-ng") == $ClientUtil::TRUE) {
        $syslog_service = "syslog-ng";
        $self->{syslog_service} = "syslog-ng";
    } else {
        $self->_set_err("Current syslog service is not supported with NATE for remote logging.\n" .
                        "Supported services: rsyslog, sysklog and syslog-ng");
        return $ClientUtil::FALSE;
    }

    if ($syslog_service =~ /rsyslog/i) {
        if ($self->_is_service_running(-service => "rsyslog") == $ClientUtil::FALSE) {
            $self->_set_err("Syslog service(rsyslog) is not running!");
            return $ClientUtil::FALSE;
        }
        $syslog_conf_file = "/etc/rsyslog.conf";
        $self->{syslog_conf_file} = "/etc/rsyslog.conf";

        # Content to be added to forward all the messages to remote server
        push (@syslog_conf_new_content, "*.*    @@" . $remote_server_ip . ":514" .  " #content added by AUTOMATION\n");
        push (@syslog_conf_new_content, "*.*    @" . $remote_server_ip . " #content added by AUTOMATION\n");
    } elsif ($syslog_service =~ /syslogd/i) {
        if ($self->_is_service_running(-service => "syslog") == $ClientUtil::FALSE) {
            $self->_set_err("Syslog service(syslog) is not running!");
            return $ClientUtil::FALSE;
        }
        $syslog_conf_file = "/etc/syslog.conf";
        $self->{syslog_conf_file} = "/etc/syslog.conf";

        # Content to be added to forward all the messages to remote server
        push (@syslog_conf_new_content, "*.*    @@" . $remote_server_ip . ":514" . " #content added by AUTOMATION\n");
        push (@syslog_conf_new_content, "*.*    @" . $remote_server_ip . " #content added by AUTOMATION\n");
    } elsif ($syslog_service =~ /syslog-ng/i) {
        if ($self->_is_service_running(-service => "syslog") == $ClientUtil::FALSE) {
            $self->_set_err("Syslog service(syslog) is not running!");
            return $ClientUtil::FALSE;
        }
        $syslog_conf_file = "/etc/syslog-ng/syslog-ng.conf";
        $self->{syslog_conf_file} = "/etc/syslog-ng/syslog-ng.conf";

        # Content to be added to forward all the messages to remote server
         push (@syslog_conf_new_content, "destination logserver \{ udp(\"$remote_server_ip\" port(514))\; \}\;\nlog \{ source(src)\; destination(logserver)\; \}\;\n");

    }

    # Getting the process id of the logging service
    $cmd = "ps aux | grep $self->{syslog_service}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command $cmd returned a bad status");
        return $ClientUtil::FALSE;
    }
    foreach my $line (@output) {
        $line =~ s/^\s+//;    # strip leading white space
        @ps_output = split (/\s+/, $line);
        if($ps_output[10] =~ /$self->{syslog_service}/i) {
            $pid = $ps_output[1];
            &SanUtil::nate_log('-comment', "Process ID of $self->{syslog_service} service : $pid");
            last;
        }
    }

    # Special value of -17 is echoed to file /proc/<pid>/oom_adj , so that the process will be killed at last
    my $oom_adj_value = -17;
    $cmd = "echo \'$oom_adj_value\' >> /proc/$pid/oom_adj";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command $cmd returned a bad status");
        return $ClientUtil::FALSE;
    }

    $cmd = "cat /proc/$pid/oom_adj";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command $cmd returned a bad status");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@syslog_conf_new_content) {
       $line_escaped = $line;
       $line_escaped =~ s/([*@().])/\\\1/g;
       my $cmd =  "perl -p -i.bak -e 'BEGIN{\$x=0}; END{exit \$x}; if (\$x==0) { if (s/^\\s*\#\*\\s*\($line_escaped)\\s*/\$1/) {\$x=1}}' $self->{syslog_conf_file}";
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
       if ($status == 1) {
           # then we successfully made the change, show the diff of the changes we made.
           $cmd = "diff $self->{syslog_conf_file}.bak $self->{syslog_conf_file}";
           &ClientUtil::execute($self->{conn}, $cmd);
       } else {
           # otherwise no instances of $param were modified, append the $param and value to the end of the file
           $cmd = "echo \'$line\' >> $self->{syslog_conf_file}";
           ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
           if ($status != 0) {
               $self->_set_err("Command ($cmd) returned status ($status)");
               return $ClientUtil::FALSE;
           }
       }
    }

    # Displaying modified syslog conf in logs. Will be useful for debugging
    $cmd = "cat " . $syslog_conf_file;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Cmd failed to display $syslog_conf_file file");
        return $ClientUtil::FALSE;
    }

    # Restarting the syslog service
    if ($syslog_service =~ /rsyslog/i) {
        if ($self->{major_revision} =~ /rhel7/) {
            $cmd = "systemctl restart rsyslog.service";
        } else {
            $cmd = "/etc/init.d/rsyslog restart";
        }
    } elsif ($syslog_service =~ /syslogd|syslog-ng/i) {
        $cmd = "/etc/init.d/syslog restart";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to restart syslog service");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  deconfigure_remote_syslog_on_client
# Objective:
################################################################################
sub deconfigure_remote_syslog_on_client {
    my ($self) = @_;

    $self->_clear_err();

    my (@output, $status, $cmd, @syslog_conf_content, $conn);
    my ($syslog_service, $syslog_conf_file, @syslog_conf_new_content, $line_escaped);

    $syslog_service = $self->{syslog_service};
    $syslog_conf_file = $self->{syslog_conf_file};

    unless (defined($syslog_service) or defined($syslog_conf_file)) {
        # Find out which syslog service is running on the client
        if ($self->_is_rpm_installed(-rpm => "rsyslog") == $ClientUtil::TRUE) {
            $syslog_service = "rsyslog";
            $syslog_conf_file = "/etc/rsyslog.conf";
            if ($self->_is_service_running(-service => "rsyslog") == $ClientUtil::FALSE) {
                $self->_set_err("Syslog service(rsyslog) is not running!");
                return $ClientUtil::FALSE;
            }
            push (@syslog_conf_new_content, "*.*    @");
        } elsif ($self->_is_rpm_installed(-rpm => "sysklogd") == $ClientUtil::TRUE) {
            $syslog_service = "syslogd";
            $syslog_conf_file = "/etc/syslog.conf";
            if ($self->_is_service_running(-service => "syslog") == $ClientUtil::FALSE) {
                $self->_set_err("Syslog service(syslog) is not running!");
                return $ClientUtil::FALSE;
            }
            push (@syslog_conf_new_content, "*.*    @");
        } elsif ($self->_is_rpm_installed(-rpm => "syslog-ng") == $ClientUtil::TRUE) {
            $syslog_service = "syslog-ng";
            $syslog_conf_file = "/etc/syslog-ng/syslog-ng.conf";
            if ($self->_is_service_running(-service => "syslog") == $ClientUtil::FALSE) {
                $self->_set_err("Syslog service(syslog) is not running!");
                return $ClientUtil::FALSE;
            }
            push (@syslog_conf_new_content, "destination logserver \{ ");
            push (@syslog_conf_new_content, "log \{ source(src)\; destination(logserver)");
        } else {
            $self->_set_err("Current syslog service is not supported with NATE for remote logging.\n" .
                            "Supported services: rsyslog, sysklog and syslog-ng");
            return $ClientUtil::FALSE;
        }
    }

    foreach my $line (@syslog_conf_new_content) {
       $line_escaped = $line;
       $line_escaped =~ s/([*@().])/\\\1/g;
       my $cmd =  "perl -lp -i.bak -e 's/^\\s*($line_escaped.*)\\s*/#\\1/g' $syslog_conf_file";
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
       if ($status != 0) {
           # then we successfully made the change, show the diff of the changes we made.
           $cmd = "diff $syslog_conf_file.bak $syslog_conf_file";
           &ClientUtil::execute($self->{conn}, $cmd);
       }
    }

    # Displaying modified syslog conf in logs. Will be useful for debugging
    $cmd = "cat " . $syslog_conf_file;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Cmd failed to display $syslog_conf_file file");
        return $ClientUtil::FALSE;
    }

    if ($syslog_service =~ /rsyslog/i) {
       if ($self->{major_revision} =~ /rhel7/) {
            $cmd = "systemctl restart rsyslog.service";
        } else {
           $cmd = "/etc/init.d/rsyslog restart";
       }
    } elsif ($syslog_service =~ /syslogd|syslog-ng/i) {
        $cmd = "/etc/init.d/syslog restart";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to restart syslog service");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  enable_kdump
# Objective:       To configure kdump to collect host crash dump
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_kdump {
    my ($self) = @_;

    my ($status, @output, $cmd, $line1, $line2, $line);

    $self->_clear_err();

    # Checking if kdump is operational
    $status = $self->_is_service_running(-service => "kdump");
    if ($status == $ClientUtil::FALSE) {
       # Verifying if the required kdump packages are installed
       $self->_verify_kdump_packages();
       if ($self->error()) {
          return $ClientUtil::FALSE;
       }

       # Add the crashkernel parameter to /boot/grub/menu.lst
       $self->_add_grub_parameter();
       if ($self->error()) {
          $self->_set_err("Couldn't change the crashkernel parameter value");
          return $ClientUtil::FALSE;
       }

       # Make kdump operational
       if ($self->{major_revision} =~ /rhel7|fedora/) {
           $cmd = "systemctl enable kdump.service";
       } else {
           $cmd = "chkconfig kdump on";
       }
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
       if ($status > 0) {
          $self->_set_err("Couldn't make kdump operational");
          return $ClientUtil::FALSE;
       }

       # Reboot the host to reflect the new crash parameters added
       $self->sanlog(-msg => "Rebooting the host to reflect the new crash parameters added");
       $self->reboot();
       if ($self->error()) {
          return $ClientUtil::FALSE;
       }
    }

    # Add kdump log server and location to /etc/kdump.conf
    # file to store the logs
    $self->_modify_kdump_config_file();
    if ($self->error()) {
       $self->_set_err("Failure in adding the net and path information to /etc/kdump.conf");
       return $ClientUtil::FALSE;
    }

    # Restart kdump service to reflect the changes
    if ($self->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl restart kdump.service";
    } else {
        $cmd = "/etc/init.d/kdump restart";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't restart kdump");
       return $ClientUtil::FALSE;
    }

    # Verifying kdump settings
    $self->_verify_kdump_settings();
    if ($self->error()) {
       $self->_set_err("Failure in verifying kdump settings");
       return $ClientUtil::FALSE;
    }

    # After hard reboot if we crash the host, kdump doesn't start
    # automatically. So manually restarting it
    if ($self->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl restart kdump.service";
    } else {
        $cmd = "/etc/init.d/kdump restart";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't restart kdump");
       return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _modify_kdump_config_file
# Objective:       To specify the net and path where the user wants to
#                  store the kdump logs
# Details:         See POD documentation at the end of this file
################################################################################
sub _modify_kdump_config_file {
    my ($self) = @_;

    my ($hostname, $kdump_server, $kdump_log_location, $cmd, $status, @output);
    my (@kdump_output);

    # Specify the nfs server and the location to save the dump
    $hostname = $self->hostname();
    $kdump_server = $self->{client_linux_kdump_server};
    $kdump_log_location = $self->{client_linux_kdump_location};
    if ($kdump_server eq "" or $kdump_log_location eq "" or
        $kdump_log_location eq "NOT_SET" or $kdump_server eq "NOT_SET") {
       $self->_set_err("User has requested for saving the dump on the server but didn't specify" .
                       "a kdump server and/or a kdump log location");
       return $ClientUtil::FALSE;
    }

    # Creating a backup file of /etc/kdump.conf
    $cmd = "cp /etc/kdump.conf /etc/kdump.conf.bak";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't make a backup file");
       return $ClientUtil::FALSE;
    }

    # No need to check the status because this commands
    # always return status as ZERO
    $cmd = "rm -f /etc/kdump.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    $cmd = "touch /etc/kdump.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Error in creating temp file");
       return $ClientUtil::FALSE;
    }

    # The output of command cmd = "cat /etc/kdump.conf" is as follows
    #-------------------------------------------------------------------------------
    # Configures where to put the kdump /proc/vmcore files
    #
    # This file contains a series of commands to perform (in order) when a
    # kernel crash has happened and the kdump kernel has been loaded.  Directives in
    # this file are only applicable to the kdump initramfs, and have no effect if
    # the root filesystem is mounted and the normal init scripts are processed
    #
    # Currently only one dump target and path may be configured at once
    # if the configured dump target fails, the default action will be preformed
    # the default action may be configured with the default directive below.  If the
    # configured dump target succedes
    #
    # See the kdump.conf(5) man page for details of configuration directives
    # #raw /dev/sda5
    # #ext3 /dev/sda3
    # #ext3 LABEL=/boot
    # #ext3 UUID=03138356-5e61-4ab3-b58e-27507ac41937
    # #net my.server.com:/export/tmp
    # #net user@my.server.com
    # #path /var/crash
    # #core_collector makedumpfile -c --message-level 1
    # #link_delay 60
    # #kdump_post /var/crash/scripts/kdump-post.sh
    # #extra_bins /usr/bin/lftp
    # #extra_modules gfs2
    # #options modulename options
    # #default shell
    #----------------------------------------------------------------------------------
    $cmd = "cat /etc/kdump.conf";
    ($status, @kdump_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Couldn't display /etc/kdump.conf file");
        return $ClientUtil::FALSE;
    }

    # Copy /etc/kdump.conf to a temporary file /etc/kdump.conf.tmp
    # The file contains some comments which we do not change
    foreach my $line (@kdump_output) {
        if ($line =~ /^\s*#/) {
           $cmd = "echo '$line' >> /etc/kdump.conf.tmp";
           my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
           if ($status > 0) {
              $self->_set_err("Couldn't append $line to file /etc/kdump.conf");
              return $clientUtil::FALSE;
           }
        }
    }

    # Append the server location into the temporary file /etc/kdump.conf.tmp
    my $server_param = "net $kdump_server:$kdump_log_location";
    $cmd = "echo '$server_param' >> /etc/kdump.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't could echo $server_param to file /etc/kdump.conf");
       return $clientUtil::FALSE;
    }

    # Append the path information to the temporary file /etc/kdump.conf.tmp
    # hostname is the name of the host for which we want to save the kdump
    # The dump is stored at the server location with this hostname
    $cmd = "echo 'path /$hostname' >> /etc/kdump.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't append $hostname to file /etc/kdump.conf");
       return $clientUtil::FALSE;
    }

    # Move the contents of /etc/kdump.conf.tmp  back to /etc/kdump.conf
    $cmd = "mv -f /etc/kdump.conf.tmp  /etc/kdump.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't move /etc/kdump.conf.tmp file to /etc/kdump.conf file");
       return $clientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _verify_kdump_packages
# Objective:       To check if the kexec-tools and kernel-debuginfo and crash
#                  packages have been installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _verify_kdump_packages {
    my ($self) = @_;

    # Check if kexec-tools package is installed
    if ($self->_is_rpm_installed(-rpm => "kexec-tools") == $ClientUtil::FALSE) {
       $self->_set_err("kexec-tools package not installed");
       return $ClientUtil::FALSE;
    }

    # Check if kernel-debuginfo is installed
    # No error message because this is not a necessary package
    if ($self->_is_rpm_installed(-rpm => "kernel-debuginfo") == $ClientUtil::FALSE) {
       $self->sanlog(-msg => "kernel- debuginfo package not installed");
    }

    # Check if crash package is installed
    # No error message becasue it is not a necessary package
    if ($self->_is_rpm_installed(-rpm => "crash") == $ClientUtil::FALSE) {
       $self->sanlog(-msg => "crash package not installed");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _add_grub_parameter
# Objective:       To add the crashkernel parameter into /boot/grub/menu.lst
# Details:         See POD documentation at the end of this file
################################################################################
sub _add_grub_parameter{
    my ($self) = @_;

    my ($status, @output, $cmd, $value, $file);

    # @16M signifies that the crashkernel memory is allocated at 16M
    $value = $self->{client_linux_kdump_kernel_size} . "M\@16M";

    # Creating a backup of the file /boot/grub/menu.lst
    $cmd = "cp /boot/grub/menu.lst  /boot/grub/menu.lst.bak";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't create a backup file");
       return $ClientUtil::FALSE;
    }

    # Check if file is present, if yes then change the modification time to current time
    # else create a new file
    $cmd = "touch /boot/grub/menu.lst.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Error in creating temp file");
       return $ClientUtil::FALSE;
    }

    # Check if file /boot/grub.menu.lst exists
    $cmd = "ls /boot/grub/menu.lst";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't find file /boot/grub/menu.lst");
       return $ClientUtil::FALSE;
    } else {
       $cmd = "cat /boot/grub/menu.lst";
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
       if ($status > 0){
          $self->_set_err("Couldn't display file /boot/grub/menu.lst");
          return $ClientUtil::FALSE;
       } else {
             foreach my $line (@output) {
                if ($line =~ /^\s+kernel\s+/) {
                   if ($line =~ /crashkernel\=(\d+)/) {
                      # Replace the crashkernel value to user defined value
                      $line =~ s/crashkernel=$1/crashkernel=$value/g;
                   } else {
                       chomp($line);
                       # Append the crashkernel parameter if not present
                       $line = $line . " crashkernel=$value" . "\n";
                   }
                }
                $file .= $line . "\n";
             }
       }

       # Copy the contents of /boot/grub/menu.lst to /boot/grub/menu.lst.tmp
       $cmd = "echo '$file' >> /boot/grub/menu.lst.tmp";
       ($status, @output)= &ClientUtil::execute($self->{conn}, $cmd);
       if ($status > 0) {
          $self->_set_err("Couldn't find file /boot/grub/menu.lst.tmp");
          return $ClientUtil::FALSE;
       }
    }

    # Replace the contents of /boot/grub/menu.lst with that of /boot/grub/menu.lst.tmp
    $cmd = "mv -f /boot/grub/menu.lst.tmp  /boot/grub/menu.lst";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
       $self->_set_err("Couldn't move /boot/grub/menu.lst.tmp file to /boot/grub/menu.lst");
       return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

#####################################################################################
# Subroutine Name  _verify_kdump_settings
# Objective:       To check the status of kdump and call function to crash the kernel
# Details:         See POD documentation at the end of this file
#####################################################################################
sub _verify_kdump_settings {
    my ($self) = @_;

    my($status, @output,$cmd);

    $status = $self->_is_service_running(-service => "kdump");
    if ($status == $ClientUtil::FALSE) {
       $self->_set_err("kdump is not operational");
       return $ClientUtil::FALSE;
    }

    # Calling the function to crash the kernel
    $self->_kernel_crash();
    if ($self->error()) {
       $self->_set_err("Couldn't crash the kernel");
       return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

##############################################################################l
# Subroutine Name  _kernel_crash
# Objective:       To crash the kernel
# Details:         See POD documentation at the end of this file
################################################################################
sub _kernel_crash {
    my ($self) = @_;

    my (@output, $status, $cmd, $hostname);

    $self->_clear_err();

    # Crash the system by echoing the magic key c into sysrq-trigger
    $cmd = "echo c > /proc/sysrq-trigger";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    $status = &SanUtil::ping_host($self->hostname());
    if ($status == $ClientUtil::FALSE) {
       $self->_clear_err();
       # Host is down which means that crash was successful
       $self->sanlog(-msg => "Host is down");
       $hostname = $self->hostname();

       # Hard reboot of the system using rpower
       $cmd = "/usr/software/utils/bin/rpower $hostname";
       &SanUtil::nate_log('-comment', "Running command $cmd on NATE control server");
       ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
       if ($status > 0) {
          $self->_set_err("Failed to hard reboot the host");
          return $ClientUtil::FALSE;
        }
    } else {
         $self->sanlog(-msg => "Ping was successful");
         # Host is still up indicating crash didn't work
         $self->_set_err("Host didn't crash");
         return $ClientUtil::FALSE;
    }

    # After we hard reboot the host,it might take sometime to reconnect, hence sleep
    $self->sanlog(-msg => "Sleep for 5 mins before reconnect");
    &SanUtil::nate_sleep(300);

    # reconnect to the client
    $self->reconnect();
    if ($self->error()){
       $self->_set_err("Unable to reconnect");
       return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _get_ol_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is Oracle Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_ol_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    # The output of cat /etc/issue on OL box looks like
    ##-------------------------------------------------------------
    # Oracle Linux Server release 5.6
    ##-------------------------------------------------------------
    # We are not using 'cat /etc/redhat-release' because this has
    # different kernel and lots of thing are different from RHEL
    $cmd = "cat /etc/issue";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ("", "");;
    }

    if ($output[0] =~ /^\s*Oracle\s+Linux\s+Server\s+release\s+\d+/i) {
        ($major, $minor) = ($output[0] =~ /release\s+(\d)\.?(\d*)/i);
    }

    if ($minor) {
        $revision = "ol" . $major . "." . $minor;
    } else {
        $minor = 0;
        $revision = "ol" . $major . "." . $minor;
    }

    return ($revision, $minor);
}
################################################################################
# Subroutine Name  _get_redhat_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is RedHat
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_redhat_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    $cmd = "cat /etc/redhat-release";
    # The output of cat /etc/redhat-release on RHEL4 box looks like
    ##-------------------------------------------------------------
    # Red Hat Enterprise Linux AS release 4 (Nahant Update 8)
    ##-------------------------------------------------------------
    # And on RHEL5/RHEL6/RHEL7 it looks like
    ##-------------------------------------------------------------
    # Red Hat Enterprise Linux Server release 5.4 (Tikanga)
    ##-------------------------------------------------------------
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ("", "");;
    }

    # Case 1: If OS is RHEL5/RHEL6/RHEL7
    if ($output[0] =~ /release\s+5/i or $output[0] =~ /release\s+6/i
       or $output[0] =~ /release\s+7/i ) {
        ($major, $minor) = ($output[0] =~ /release\s+(\d)\.?(\d*)/i);
    } else { # Case 2: If OS is RHEL4
        if ($output[0] =~ /Update/i) {
            ($major, $minor) = ($output[0] =~ /release\s+(\d+)\s+\(\w+\s+Update\s*(\d+)\)/i);
        } else {
            ($major) = ($output[0] =~ /release\s+(\d+)\s+\(\w+\)/i);
        }
    }
    if ($minor) {
        $revision = "rhel" . $major . "." . $minor;
    } else {
        $minor = 0;
        $revision = "rhel" . $major . "." . $minor;
    }

    return ($revision, $minor);
}

################################################################################
# Subroutine Name  _get_fedora_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is fedora
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_fedora_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    # The output of cat /etc/fedora-release looks like
    ##-------------------------------------------------------------
    # Fedora release 18 (Spherical Cow)
    ##-------------------------------------------------------------
    $cmd = "cat /etc/fedora-release";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ("", "");;
    }

    if ($output[0] =~ /^\s*Fedora\s+release\s+\d+/i) {
       ($major, $minor) = ($output[0] =~ /release\s+(\d+)\.?(\d*)/i);
    }

    if ($minor) {
        $revision = "fedora" . $major . "." . $minor;
    } else {
        $minor = 0;
        $revision = "fedora" . $major;
    }

    return ($revision, $minor);
}

################################################################################
# Subroutine Name  _get_ovm_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is OVM
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_ovm_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    # The output of cat /etc/ovs-release looks like
    ##-------------------------------------------------------------
    # Oracle VM server release 2.2.1
    ##-------------------------------------------------------------
    $cmd = "cat /etc/ovs-release";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ("", "");;
        return $ClientUtil::FALSE;
    }
    ($major, $minor) = ($output[0] =~ /Oracle\s+VM\s+server\s+release\s+(\d+\.\d+)\.?(\d*)/i);
    if ($minor) {
        $revision = "ovm" . $major . "." . $minor;
    } else {
        $minor = 0;
        $revision = "ovm" . $major . "." . $minor;
    }

    return ($revision, $minor);
}

################################################################################
# Subroutine Name  _get_suse_major_and_minor_revisions
# Objective:       Will store major and minor revision in self if OS is SuSe
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_suse_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    $cmd = "cat /etc/SuSE-release";
    # The output of cat /etc/SuSE-release looks like
    # SUSE Linux Enterprise Server 11 (x86_64)
    # VERSION = 11
    # PATCHLEVEL = 0
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_new() Command ($cmd) returned a bad status ($status)");
        return ("", "");;
    }

    ($major) = ($output[1] =~ /(\d+)/i);
    ($minor) = ($output[2] =~ /(\d+)/i);

    if ($minor) {
        $revision = "sles" . $major . "." . $minor;
    } else {
        $minor = 0;
        $revision = "sles" . $major . "." . $minor;
    }


    return ($revision, $minor);
}

################################################################################
# Subroutine Name  _get_citrix_xenserver_major_and_minor_revisions
# Objective:       Will store major and minor revision if OS is Citrix XenServer
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_citrix_xenserver_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $major, $minor, $revision);

    # The output of cat /etc/issue on Xen Server looks like
    ##-------------------------------------------------------------
    # Citrix XenServer Host 6.1.0-59235p
    ##-------------------------------------------------------------
    $cmd = "cat /etc/issue";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ("", "");;
    }

    if ($output[0] =~ /^\s*Citrix\s+XenServer\s+Host/i) {
        ($major, $minor) = ($output[0] =~ /Host\s+(\d+)\.?(\d*)/i);
    } else {
        $self->_set_err("Failed to find major and mior revision of Citrix XenServer");
        return ("", "");
    }

    unless (defined($minor)) {
        $minor = 0;
    }

    $revision = "cxs" . $major . "." . $minor;

    return ($revision, $minor);
}

################################################################################
# Subroutine Name  store_OS_major_and_minor_revisions
# Objective:       Will store major and minor revision in self for the client
# Details:         See POD documentation at the end of this file
################################################################################
sub store_OS_major_and_minor_revisions {
    my ($self) = @_;

    my (@output, $status, $cmd, $revision, $minor);

    $self->_clear_err();

    if ($self->hosttype() =~ /redhat|rhel|centos|fedora|citrix|^unix-linux$/i) {
        # Outout of the following command looks like this
        ##-------------------------------------------------------------
        # Red Hat Enterprise Linux Server release 5.4 (Tikanga)
        ##-------------------------------------------------------------
        $cmd = "cat /etc/issue";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
        # We have separated getting OL major/minor revision from RHEL because
        # Oracle ships with three kernels now.
        # 1. RHEL kernel
        # 2. Kernel from upstream
        # 3. RHEL kernel with some patches
        # if '/etc/issue' file has Oracle in it that means the kernel is
        # upstream kernel and lots of packages changes(including multipath)
        # from RHEL kernel
        if ($output[0] =~ /^\s*Oracle\s+Linux/i) {
            ($revision, $minor) = $self->_get_ol_major_and_minor_revisions();
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        } elsif ($output[0] =~ /Citrix\s+XenServer/i) {
            ($revision, $minor) = $self->_get_citrix_xenserver_major_and_minor_revisions();
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        } elsif ($output[0] =~ /CentOS|Red\s+Hat\s+Enterprise\s+Linux|Enterprise\s+Linux\s+Enterprise\s+Linux\s+Server\s+release/i) {
            ($revision, $minor) = $self->_get_redhat_major_and_minor_revisions();
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        } elsif ($output[0] =~ /Fedora/i) {
             ($revision, $minor) = $self->_get_fedora_major_and_minor_revisions();
             if ($self->error) {
                 return $ClientUtil::FALSE;
             }
        }

        # OVM also has /etc/redhat-release file but content of the
        # file looks like this for OVM
        ##------------------------------------------------------------
        # Oracle VM server release 2.2.1
        ##------------------------------------------------------------
        elsif ($output[0] =~ /Oracle\s+VM\s+server/i) {
            ($revision, $minor) = $self->_get_ovm_major_and_minor_revisions();
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        }
    } elsif ($self->hosttype() =~ /suse|sles/i) {
        ($revision, $minor) = $self->_get_suse_major_and_minor_revisions();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    $self->{major_revision} = $revision;
    $self->{minor_revision} = $minor;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_proc_cmdline
# Objective:       verifies whether the requested option exist in /proc/cmdline
################################################################################
sub verify_proc_cmdline {
    my ($self, %args) = @_;

    my ($option, $cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?option/) {
            if (not defined ($option = $args{$_})) {
                $self->_set_err("Invalid value for -option => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    if (not defined($option)) {
        $self->_set_err("option is a required param");
        return $ClientUtil::FALSE;
    }

    $option =~ s/^\s*//;
    $option =~ s/\s*$//;

    $cmd = "cat /proc/cmdline";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to execute the cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        if ($line =~ /$option/) {
            # Requested option exists in /proc/cmdline
            return $ClientUtil::TRUE;
        }
    }

    # This means that option is not exist in /proc/cmdline
    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name  _get_client_kernel_state
# Objective:       Will request client for SysRq dump
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_client_kernel_state {
    my ($self) = @_;

    my (@output, $sysrq_port, $status, $cmd, $pid, $hostname, $proc_info);

    $hostname = $self->hostname();

    # Get the pid for sysrq dump request script running on NATE
    # Control server and check if the script is running for the host
    if (!defined ($self->{sysrq_pid})) {
        unless (open(PID_FILE, "</var/tmp/$hostname-sysrq-pid")){
            $self->_set_err("Error: unable to open file /var/tmp/$hostname-sysrq-pid");
            return $ClientUtil::FALSE;
        }
        if (defined (my $pid = <PID_FILE>)) {
            chomp($pid);
            $self->{sysrq_pid} = $pid;
            close(PID_FILE);
        }
    }

    $pid = $self->{sysrq_pid};

    # If the scripts are not already running start them
    # Checking this on NATE Control Server because at this time doing it on test machine
    # could be an expensive(machine will even hang in sanboot case) operation
    $cmd = "ps -elf | grep \" $pid \" | grep -v grep";
    &SanUtil::nate_log('-comment', "Running cmd $cmd on NATE control server");
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    foreach my $line (@output) {
        my ($current_pid) = ($line =~ /^\s*\w+\s+\w+\s+\w+\s+(\d+)\s+.+/);
        if ($current_pid eq $pid) {
            $proc_info = $line;
            last;
        }
    }
    if (!defined($proc_info)) {
        $self->sanlog(-msg => "SysRq dump script is not running on machine $hostname " .
                              "and NATE Control Server. Starting them");
        $self->{sysrq_port} = 12882;
        $self->_start_sysrq_dump_collection_scripts();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # Sleep for some time before requesting for sysrq dump
    &SanUtil::nate_sleep(2);

    $self->sanlog(-msg => "collecting sysrq dump using pid $pid");
    $cmd = "kill -USR1 $pid";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Unable to send signal to sysrq dump script on NATE Control Server");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _start_sysrq_dump_script_on_NCS
# Objective:       Will start the sysrq dump script on NATE Control Server
# Details:         See POD documentation at the end of this file
################################################################################
sub _start_sysrq_dump_script_on_NCS {
    my ($self) = @_;

    my (@output, $sysrq_port, $status, $cmd, $pid, $hostname, $bin_dir, $log_dir, $proc_info);

    # Port will be used to start the socket connection
    # between client and NATE control server
    $self->{sysrq_port} = "12882";
    $sysrq_port = $self->{sysrq_port};
    $hostname = $self->hostname();

    # The message received from the client will be stored at this location
    # $log_dir/$hostname-sysrq.log
    $log_dir = $self->{nate_logdir};
    $self->sanlog(-msg => "SysRq dump for client ($hostname) will be saved in the file " .
                          "($log_dir/$hostname-sysrq.log)");

    $bin_dir = $self->{TEST_BIN};

    my $nc_port = $self->{nc_port};

    # start the script on NATE Control Server
    $cmd = "perl " . $bin_dir . "/sysrq_dump_request.pl $hostname $sysrq_port $nc_port $log_dir";
    $self->sanlog(-msg => "running cmd $cmd on NATE Control Server in background");
    system("$cmd &");

    # sleeping for 2 secs to give enough time to NATE Control Server to start the script and
    # store its pid in a file for later use
    &SanUtil::nate_sleep(2);

    # Make sure that the sysrq dump script started on NATE Control Server
    unless (open(PID_FILE, "</var/tmp/$hostname-sysrq-pid")){
        $self->_set_err("Error: unable to open file /var/tmp/$hostname-sysrq-pid");
        return $ClientUtil::FALSE;
    }

    if (defined($pid = <PID_FILE>)) {
        chomp($pid);
        close(PID_FILE);

        # Get the proc info of the sysrq_dump_request script
        # Here we are grepping for PID(which is a number) but this number
        # could occur in someone else proc at some other place. So we need
        # to make sure that we get actual proc. Also grepping with the
        # spaces around PID to make sure that this PID is not a substring
        # of another proc's PID
        #
        $cmd = "ps -elf | grep \" $pid \" | grep -v grep";
        &SanUtil::nate_log('-comment', "Running cmd $cmd on NATE control server");
        ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
        foreach my $line (@output) {
            my ($current_pid) = ($line =~ /^\s*\w+\s+\w+\s+\w+\s+(\d+)\s+.+/);
            if ($current_pid eq $pid) {
                $proc_info = $line;
                last;
            }
        }

        if (!defined($proc_info)) {
            $self->_set_err("Error: unable to start sysrq dump script on NATE Control Server");
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Proc info of sysrq_dump_request is \n $proc_info \n" .
                              "And pid of this proc is '$pid'");
    }

    # Store pid
    $self->{sysrq_pid} = $pid;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _start_sysrq_dump_script_on_client
# Objective:       Will start the sysrq dump script on test machine
# Details:         See POD documentation at the end of this file
################################################################################
sub _start_sysrq_dump_script_on_client {
    my ($self, %args) = @_;

    my (@output, $hostname, $sysrq_port, $status, $cmd, $bin_dir, $retry);
    my ($nate_server_ip, $nate_server_mac, $sysrq_dump_info);

    foreach (keys %args) {
        if (/-?sysrq_dump_info/) {
            if (not length($sysrq_dump_info = $args{$_})) {
                $self->_set_err("Invalid value for -sysrq_dump_info: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->{nc_port} = $self->_get_next_available_port_for_nc();;

    # Enable SysRq on the host
    $cmd = "echo 1 > /proc/sys/kernel/sysrq";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to run command \'$cmd\'");
        return $ClientUtil::FALSE;
    }

    # Increase the log level to 8
    $cmd = "echo 8 > /proc/sysrq-trigger";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to run command \'$cmd\'");
        return $ClientUtil::FALSE;
    }

    # Get the ip address and MAC address of NATE control server
    # These values are going to be used by netconsole module
    $cmd = "ifconfig";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to run command \'$cmd\' on NATE Server");
        return $ClientUtil::FALSE;
    }

    my $NATE_server_reachable = 0;
    foreach my $line (@output) {
        # skip the newlines
        next if ($line =~ /^$/);
        # Get the MAC address of NATE server
        if ($line =~ /HWaddr\s+(\S+)/) {
            $nate_server_mac = $1;
        }
        # Get the IP address of NATE server
        if ($line =~ /inet\s+addr:\s*(\d+\.\d+\.\d+\.\d+)\s+/) {
            $nate_server_ip = $1;
            if (&SanUtil::ping_host($nate_server_ip) == $ClientUtil::FALSE) {
                # Interface not pingable go to next line
                next;
            } else {
                # Exit as soon as we get ip address of NATE Server that is pingable
                $NATE_server_reachable = 1;
                last;
            }
        }
    }

    # Not able to ping NATE server. Fail out
    if ($NATE_server_reachable == 0) {
        $self->_set_err("Error: Unable to ping NATE server through any of the interfaces");
        return $ClientUtil::FALSE;
    }

    # Load netconsole module on client
    $self->_load_netconsole_module(-nate_server_ip  => $nate_server_ip,
                                   -nate_server_mac => $nate_server_mac);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # Port will be used to start the socket connection
    # between client and NATE control server
    $self->{sysrq_port} = "12882";

    $sysrq_port = $self->{sysrq_port};
    $bin_dir = $self->{TEST_BIN};

    # Make sure test_bin dir is accessible. Try 5 times and if still
    # not able to access dir then fail out
    $cmd = "ls $bin_dir";
    for ($retry = 0; $retry < 5; $retry++) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        next if (grep (/no\s+such\s+file\s+or\s+directory/i, @output));
        last;
    }

    # Fail if not able to access in 5 tries
    if ($retry == 5) {
        $self->_set_err("Unable to access the location $bin_dir");
        return $ClientUtil::FALSE;
    }

    # Starting the script with very high priority so that it doesnt get killed if
    # there is a resource cruch on the client
    $cmd = "nice -n -19 perl " . $bin_dir . "/sysrq_dump_trigger.pl $sysrq_port $sysrq_dump_info &";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # Process started in background doesnt return proper status
    # Making sure that SysRq dump script has started
    $cmd = "ps -elf | grep sysrq_dump_trigger | grep -v grep";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) == 0) {
        $self->_set_err("Unable to start the sysrq dump script on the Host $self->hostname()");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _exit_sysrq_dump_script_on_client
# Objective:       Will kill the sysrq dump script on test machine
# Details:         See POD documentation at the end of this file
################################################################################
sub _exit_sysrq_dump_script_on_client {
    my ($self) = @_;
    my (@output, $status, $cmd, $pid, $hostname, $conn);

    # Unload the netconsole module
    if ($self->_is_module_loaded(-module => 'netconsole') == $ClientUtil::TRUE) {
        $self->_unload_module(-module => 'netconsole');
        if ($self->error) {
             $self->_set_err("Unable to unload netconsole driver module \'netconsole\'");
             return $ClientUtil::FALSE;
        }
    }

    $conn = $self->{conn};
    $hostname = $self->hostname();

    # Kill previously running version of sysrq dump script on client if its running
    # Get the pid using ps cmd. Since we are forking in the script. There will
    # be two processes running. Need to kill both of them
    #
    # Output of the "ps" cmd will look like this
    # F S      UID   PID  PPID   C PRI NI     ADDR     SZ    WCHAN    STIME TTY         TIME CMD
    # 0 S    rawat 21918 17175   0  50 20        ?    291        ? 20:32:26 pts/31      0:00 grep grep
    #
    $cmd = "ps -efl | grep sysrq_dump_trigger | grep -v grep";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if (scalar (@output) == 0) {
        $self->sanlog(-msg => "SysRq dump trigger script is not running on test machine");
        return $ClientUtil::TRUE;
    }

    foreach my $proc_info (@output) {
        if ($proc_info =~ /^\s*\w+\s+\w+\s+\w+\s+(\d+)\s+.+/) {
            my $pid = $1;
            $self->sanlog(-msg => "Going to kill the process having PID=$pid");
            $cmd = "kill -9 $pid";
            my ($status, @kill_output) = &ClientUtil::execute($conn, $cmd);
            if ($status > 0) {
                $self->sanlog(-msg => "Error: Unable to kill previously running version " .
                                            "of sysrq dump script on client" . $hostname);
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       will collect sysrq dump periodically in regular time intervals
#                  throughout the test
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    my ($client_linux_enable_sysrq_dump_periodic, $client_linux_sysrq_dump_time_interval);
    my ($client_linux_enable_sysrq_dump);

    $self->_clear_err();

    $client_linux_enable_sysrq_dump_periodic = $self->{client_linux_enable_sysrq_dump_periodic};
    $client_linux_sysrq_dump_time_interval = $self->{client_linux_sysrq_dump_time_interval};
    $client_linux_enable_sysrq_dump = $self->{client_linux_enable_sysrq_dump};

    if ($client_linux_enable_sysrq_dump_periodic == 1 and $client_linux_enable_sysrq_dump ==1) {
        while(1) {
            # Gather sysrq dump from the client
            $self->gather_data();
            # Sleep $client_linux_sysrq_dump_time_interval seconds before
            # collecting data again
            &SanUtil::nate_sleep($client_linux_sysrq_dump_time_interval);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    my ($mnt_pt, $unit, $cmd, @output, $status);
    my $disk_usage = {};

    # Parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
            if (not length($mnt_pt = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_pt': $args{$_}");
                return $disk_usage;
            }
            next;
        } elsif (/-?unit/) {
            if (not length($unit = $args{$_}) or $args{$_} !~ /k|m|g|t/i) {
                $self->_set_err("Invalid value for argument '-unit': $args{$_}");
                return $disk_usage;
            }
            next;
        }
    }

    unless (defined($mnt_pt)) {
        $self->_set_err("Missing argument 'mount_pt'");
        return $disk_usage;
    }

    # if unit is not passed, assume they need size in bytes
    unless (defined($unit)) {
        $unit = "";
    }

    $cmd = "df -kP $mnt_pt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command($cmd) returned bad status");
        return $disk_usage;
    }

    # Remover leading white spaces if any
    $output[1] =~ s/^\s*//;

    my ($undef, $total, $used, $avail) = split(/\s+/, $output[1]);

    $total = $total . "k";
    $used = $used . "k";
    $avail = $avail . "k";

    # Convert the sizes which are in kilo bytes to respective units requested by user
    $total = &ClientUtil::convert_to_bytes($total, $unit);
    $used = &ClientUtil::convert_to_bytes($used, $unit);
    $avail = &ClientUtil::convert_to_bytes($avail, $unit);

    $disk_usage->{total} = $total;
    $disk_usage->{used} = $used;
    $disk_usage->{avail} = $avail;

    return $disk_usage;
}

###############################################################################
# Subroutine Name: deconfigure
# Objective:       deconfigure host machine
# Details:         See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets, $clean_level);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            @targets = @{$args{$_}};
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }
    if ($self->{protocol} =~ /iscsi/i) {
        # Build the args
        my %cmd_args;

        # Add in targets if defined.
        if (scalar(@targets) != 0) {
            $cmd_args{"-targets"} = \@targets;
        }

        # Add in clean_level if defined.
        if (defined($clean_level)) {
            $cmd_args{"-clean_level"} = $clean_level;
        } else {
            # default it to 1
            $cmd_args{"-clean_level"} = 1;
        }

        # return if clean level is 0
        if ($cmd_args{"-clean_level"} == 0) {
            $self->sanlog(-msg => "iSCSI Deconfigure: Not performing iSCSI deconfiguration because clean_level is '0'");
            return $ClientUtil::TRUE;
        }

        $self->_iscsi_deconfigure(%cmd_args);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective: returns scsi device size in bytes
# Details: See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($scsi_device, $cmd, $status, @output, $size);

    # parse arguments
    foreach (keys %args) {
        if (/^-?device$/) {
            if (not length($scsi_device = $args{$_})) {
                $self->_set_err("Invalid value for " . "-device => $args{$_}");
                return "";
            }
            next;
        }
    }

    # this cmd will give output in 512-byte blocks
    $cmd = "cat /sys/block/$scsi_device/size";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Not able to find scsi device size using cmd ($cmd)");
        return "";
    }

    # multiply by 512 to convert size into bytes
    $size = $output[0] * 512;

    return $size;

}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective: returns all scsi devices in /dev/ dir.
# Details: See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self) = @_;

    $self->_clear_err();

    my (@scsi_devices, $cmd, $status, @output);

    # get all the scsi devices in /dev directory
    $cmd = "ls /dev | grep sd.*";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 1 and scalar(@output) == 0) {
        # no devices found
        $self->sanlog(-msg => "No scsi devices found");
    } elsif ($status != 0) {
        $self->_set_err("Cmd ($cmd) returned bad status");
        return (wantarray()) ? () : [];
    }

    @scsi_devices = @output;

    return (wantarray()) ? @scsi_devices : \@scsi_devices;
}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    my ($cfo_mode, @output, @output_lspci, $status, $cmd);
    my (@targets, @fcp_targets, @iscsi_targets);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not (@targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?fcp_targets$/) {
            if (not (@fcp_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-fcp_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Checks if user wants to enable kdump on linux host
    if ($self->{client_linux_enable_kdump} == 1) {
       $self->enable_kdump();
       if ($self->error) {
          return $ClientUtil::FALSE;
       }
    } elsif ($self->{client_linux_enable_kdump} == 0) {
        $self->sanlog(-msg => "kdump has not been requested by user");
    }

    # configure for protocol fcp
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_configure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }
    if ($self->{protocol} =~ /iscsi/i) {
        # configure for protocol iscsi
        if (scalar(@iscsi_targets) == 0) {
            $self->_iscsi_configure(-targets => @targets);
        } else {
            $self->_iscsi_configure(-targets => @iscsi_targets);
        }
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # Configure the MP Object
    if (defined $self->{mp} and $self->{mp_type} !~ /""|0|none/i) {
        $self->_MP_configure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # Need to do it again because system got rebooted
    # By default 'raw' module is not loaded on SLES. We need to load
    # it. It will create master raw device '/dev/rawctl'.
    # On Redhat there is no raw module and device /dev/rawctl will already
    # be there
    if ($self->hosttype() =~ /suse|sles/i) {
        $cmd = "modprobe raw";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "LinuxHost::configure() SanClient Config complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_next_available_port_for_nc
# Objective:       Will return the next free port that can be used by nc
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_next_available_port_for_nc {
    my ($self) = @_;

    my (@output, $nc_port, $status, $cmd, $pid);

    my $hostname = $self->hostname();
    $cmd = "sudo mkdir -p /var/tmp/nc";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    for (my $i = 0; $i < 1000; $i++) {
        # We will use integer to use as port from 9000 to 10000. Using rand function
        # to return a random number between 0 and 1000 and adding 9000 to it that
        # will give us require number to use as a port for nc
        $nc_port = int(rand(1000)) + 9000;
        my $file = "/var/tmp/nc/" . $nc_port;
        # We get permission denied error on NATE server if sudo is not used
        $cmd = "sudo ls $file";
        ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
        if ($status > 0 or grep(/no such file or directory/i, $output[0])) {
            $cmd = "sudo touch $file";
            ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
            if ($status > 0) {
                next;
            }

            # Changing the permission to 777
            $cmd = "sudo chmod 777 $file";
            ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to run command \'$cmd\' on NATE Server");
                return "";
            }

            # Hostname is going to be used later to delete this file
            $cmd = "sudo echo $hostname > $file";
            ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to run command \'$cmd\' on NATE Server");
                return "";
            }
            last;
        }
    }

    return $nc_port;
}

################################################################################
# Subroutine Name: _start_sysrq_dump_collection_scripts
# Objective:       Will start the sysrq dump scripts. These scripts are
#                  responsible for gathering the kernel message and sending
#                  them to log directory
# Details:         See POD documentation at the end of this file
################################################################################
sub _start_sysrq_dump_collection_scripts {
    my ($self) = @_;

    my (@output, $sysrq_port, $status, $cmd, $pid, $client_linux_sysrq_dump_info);

    $client_linux_sysrq_dump_info = $self->{client_linux_sysrq_dump_info};

    $self->_start_sysrq_dump_script_on_client(-sysrq_dump_info => $client_linux_sysrq_dump_info);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # Should give enough time for srcipt to start on client
    &SanUtil::nate_sleep(2);

    $self->_start_sysrq_dump_script_on_NCS();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _rescan_scsi_bus ()
# Objective: Will execute /usr/bin/rescan-scsi-bus.sh script. This script is used
#            to rescan the SCSI bus.
################################################################################
sub _rescan_scsi_bus {
    my ($self, %args) = @_;

    my ($option, $protocol, $cmd, $status, @output);

    $self->_clear_err();

    # Parse arguments
    foreach (keys %args) {
         if (/-?option/) {
             if (not length($option = $args{$_})) {
                 $self->_set_err("Missing value for argument '-option': $args{$_}");
                 return $ClientUtil::FALSE;
             }
             next;
         } elsif (/-?protocol/) {
             if (not length($protocol = $args{$_})) {
                 $self->_set_err("Missing value for argument '-protocol': $args{$_}");
                 return $ClientUtil::FALSE;
             }
             next;
         }
    }

    # Default protocol to $self->{protocol} if not set
    if (not defined($protocol) or $protocol eq "") {
        $protocol = $self->{protocol};
    }

    if ($protocol =~ /fcp/i) {
        if ($option =~ /discover_LUNs/i) {
            # To discover newly mapped luns
            $cmd = "/usr/bin/rescan-scsi-bus.sh";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to rescan the FCP sessions");
                return $ClientUtil::FALSE;
            }
        } elsif ($option =~ /remove_LUNs/i) {
            # To delete unmapped LUNs from the host, argument is: "-r"
            $cmd = "/usr/bin/rescan-scsi-bus.sh -r";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to rescan the FCP sessions");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid value for '$option': $option");
            return $ClientUtil::FALSE;
        }
    }

    if ($protocol =~ /iscsi/i) {
        if ($option =~ /remove_LUNs/i) {
            # To delete unmapped LUNs from the host, argument is: "-r"
            $cmd = "/usr/bin/rescan-scsi-bus.sh -r";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to rescan the iSCSI sessions");
                return $ClientUtil::FALSE;
            }
        } else {
            # To discover newly mapped luns
            $cmd = "/usr/bin/rescan-scsi-bus.sh";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to rescan the iSCSI sessions");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_rpm_installed()
# Objective: Will check for the rpm
# Details: See POD documentation at the end of this file
################################################################################
sub _is_rpm_installed {
    my ($self, %args) = @_;

    my ($rpm, $cmd, $status, @output);

    $self->_clear_err();

    # Parse arguments
    foreach (keys %args) {
         if (/-?rpm/) {
             if (not length($rpm = $args{$_})) {
                 $self->_set_err("Missing value for argument '-rpm': $args{$_}");
                 return $ClientUtil::FALSE;
             }
             next;
         }
    }

    if (not defined($rpm) or not length($rpm)) {
        $self->_set_err("'-rpm' is a required argument");
        return $ClientUtil::FALSE;
    }

    $cmd = "rpm -q $rpm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        # Not setting the error because we will not fail out even if
        # this command fails
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _close_encrypted_devices_luks
# Objective:       Will Close LUKS encryption devices
# Details:         See POD documentation at the end of this file
################################################################################
sub _close_encrypted_devices_luks {
    my ($self, %args) = @_;

    my ($cmd, $status, @output, @encrypt_dev);

    $self->_clear_err();

    # Getting the list of LUKS enabled devices
    $cmd = "ls /dev/mapper/ | grep -P \"^crypt-360a9|^crypt-3600a09\"";
    ($status, @encrypt_dev) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@encrypt_dev) == 0) {
        $self->sanlog(-msg => "No encrypted devices found");
        return $ClientUtil::TRUE;
    }
    if ($status > 0) {
        $self->_set_err("Failed to get LUKS encryption devices");
        return $ClientUtil::FALSE;
    }

    # Close LUKS enabled devices
    foreach my $device (@encrypt_dev) {
        my $full_path = "\/dev\/mapper\/" . $device;
        $cmd = "/sbin/cryptsetup luksClose $full_path";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to close LUKS encryption devices");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _fcp_configure ()
# Objective: Will reboot the system and check if driver is loaded. If not try to
#            load it again. If this load fails then error out. Drivers will be
#            loaded depending upon Card type(Qlogic/Emulex) and Driver type
#            (native/vendor). The value will be saved in a ClientObj for future use
# Details: See POD documentation at the end of this file
################################################################################
sub _fcp_configure {
    my ($self, %args) = @_;

    my (@targets, $major, $minor, $cfo_mode, @output, @output_lspci, $status, $cmd);

    $major = $self->{major_revision};
    $minor = $self->{minor_revision};

    # SLES 10.3 and RHEL 5.4 onwards we use /usr/bin/rescan-scsi-bus.sh to rescan without loading and unloading fc driver
    # except for the very first time. We can avoid rebooting the Host.
    if ($major =~ /cxs/i) {
        # This step is not required for Citrix XS
    } elsif (($major =~ /sles9|rhel4/i)
        or (($major =~ /sles10/i) and $minor < 3)
        or (($major =~ /rhel5/i) and $minor < 4)
        or ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE)) {
        $self->sanlog(-msg => "System is rebooting to unload and load fcp driver");
        $self->reboot();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_rescan_scsi_bus(-option => 'discover_LUNs', -protocol => "fcp");
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    $cmd = "lspci";
    ($status, @output_lspci) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: cmd $cmd return bad status");
        return $ClientUtil::FALSE;
    }

    # if CONFIG_FCP_DRIVER_TYPE is not set then defaulting it to vendor.
    if (!defined $self->{fcp}->{CONFIG_FCP_DRIVER_TYPE}) {
        $self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} = "vendor";
    }

    my @module_type;
    # check to see if we have Cisco CNA for UCS
    if (grep(/Cisco.*FCoE/i, @output_lspci)) {
        $cmd = "lsmod | grep fnic";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v fnic";
           ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: unable to load fcp (Cisco) driver");
                return $ClientUtil::FALSE;
            }
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Cisco";
        $self->{fcp_driver_module_type} = "fnic";
    } elsif (grep(/Brocade/i, @output_lspci)) {
        # check to see if we have Brocade card
        $cmd = "lsmod | grep \"^bfa\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v bfa";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: unable to load fcp (Brocade) driver");
                return $ClientUtil::FALSE;
            }
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Brocade";
        $self->{fcp_driver_module_type} = "bfa";
    } elsif (grep(/QLogic/i, @output_lspci)) {
        # check to see if we have QLogic card
        (@module_type) = (grep /QLogic/i,@output_lspci);
        if ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /vendor/i) {
            if ($module_type[0] =~ /23\d\d/) {
                $cmd = "lsmod | grep qla2300";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0 or scalar(@output) == 0) {
                    $cmd = "modprobe -v qla2300";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() unable to load fcp driver " .
                                        "module qla2300 for Qlogic card. Make sure driver gets loaded after reboot");
                        return $ClientUtil::FALSE;
                    } else {
                        $self->sanlog(msg => "fcp driver module qla2300 for Qlogic card was not loaded after reboot but it" .
                                             " got loaded after retry");
                    }
                } else {
                    $self->sanlog(msg => "fcp driver module qla2300 for Qlogic card is reloaded successfully after reboot");
                }

                # Save fcp driver module type for future use
                $self->{fcp_driver_module_type} = "qla2300";
            } elsif ($module_type[0] =~ /24\d\d/) {
                $cmd = "lsmod | grep qla2400";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0 or scalar(@output) == 0) {
                    $cmd = "modprobe -v qla2400";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() unable to load fcp driver ".
                                        "module qla2400 for Qlogic card Make sure driver gets loaded after reboot");
                        return $ClientUtil::FALSE;
                    } else {
                        $self->sanlog(msg => "fcp driver module qla2400 for Qlogic card was not loaded after reboot but it" .
                                             " got loaded after retry");
                    }
                } else {
                    $self->sanlog(msg => "fcp driver module qla2400 for Qlogic card is reloaded successfully after reboot");
                }
            } else {
                $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() currently this Qlogic card ".
                                "is not supported in Automation:\n\t " . join("\n\t", @module_type));
            }

            # Save fcp driver module type for future use
            $self->{fcp_driver_module_type} = "qla2400";
        } elsif ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /native/i) {
            $cmd = "lsmod | grep qla2xxx";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 or scalar(@output) == 0) {
                $cmd = "modprobe -v qla2xxx";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() unable to load fcp driver ".
                                    "module qla2xxx for Qlogic card. Make sure driver gets loaded after reboot");
                    return $ClientUtil::FALSE;
                } else {
                    $self->sanlog(msg => "fcp driver module qla2xxx for Qlogic card was not loaded after reboot but it" .
                                         " got loaded after a try");
                }
            } else {
                    $self->sanlog(msg => "fcp driver module qla2xxx for Qlogic card is reloaded successfully after reboot");
            }

            # Save fcp driver module for future use
            $self->{fcp_driver_module_type} = "qla2xxx";
        } else {
            $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() \"CONFIG_FCP_DRIVER_TYPE\" is not ".
                            "specified. Please specify this parameter in your .include file which is included when ".
                            " tests are invoked using ntest. Values for this param could be \"native\" or \"vendor\"");
            return $ClientUtil::FALSE;
        }

        # Save fcp card type for future use
        $self->{fcp_card_type} = "Qlogic";
    } elsif (grep(/Emulex|Fibre\s+Channel.*ServerEngines/i, @output_lspci)) {
        $cmd = "lsmod |grep lpfc";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v lpfc";
           ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() unable to load fcp (Emulex) driver");
                return $ClientUtil::FALSE;
            }
        }
        $cmd = "modprobe -v lpfcdfc";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 and (!grep /not\s+found/i,@output)) {
            $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() unable to load fcp (Emulex) driver");
            return $ClientUtil::FALSE;
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Emulex";
        $self->{fcp_driver_module_type} = "lpfc";
    } elsif (grep(/Intel\s+Corporation\s+82599/i, @output_lspci)) {
        # check to see if we have Intel 10G card
        $cmd = "lsmod | grep \"^libfcoe\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v libfcoe";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: unable to load Intel driver");
                return $ClientUtil::FALSE;
            }
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Intel";
        $self->{fcp_driver_module_type} = "libfcoe";
    } elsif (grep(/Broadcom.*BCM57712/i, @output_lspci)) {
        # check to see if we have Broadcom 57712 CNA
        $cmd = "lsmod | grep \"^bnx2fc\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v bnx2fc";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: unable to load fcp (Broadcom) driver");
                return $ClientUtil::FALSE;
            }
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Broadcom";
        $self->{fcp_driver_module_type} = "bnx2fc";
    } elsif (grep(/Chelsio Communications.*Unified Wire Storage Controller/i, @output_lspci)) {
        # check to see if we have Chelsio FCoE card
        $cmd = "lsmod | grep \"^csiostor\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $cmd = "modprobe -v csiostor";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: unable to load fcp (Chelsio) driver");
                return $ClientUtil::FALSE;
            }
        }

        # Save fcp card and driver module type for future use
        $self->{fcp_card_type} = "Chelsio";
        $self->{fcp_driver_module_type} = "csiostor";
    } else {
        $self->_set_err("Error: SanClient::Host::LinuxHost::_fcp_configure() You specified protocol = \"fcp\" ".
                        "But no fcp card is present on the system. Install the fcp card first and re-run the test");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;

    my ($major, $minor, $cmd, @devices, $status, @output, @hDevices);

    $major = $self->{major_revision};
    $minor = $self->{minor_revision};
    $self->_clear_err();

    $self->sanlog(-msg => "Discovering all free devices");

    $self->sanlog(-msg       => "Locate devices that are not part of a volume group",
                  -msg_level => 3);

    if ($self->{protocol} =~ /iscsi/i) {
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/i) {
            if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
                # issue a rescan on all sessions if doing open-iscsi
                # From RHEL 5.4 and SLES 10.3 onwards, we don't use iscsiadm to rescan iSCSI luns,
                # but instead we use /usr/bin/rescan-scsi-bus.sh.
                if ((($major =~ /sles10/i) and $minor < 3)
                    or (($major =~ /rhel5/i) and $minor < 4)
                    or ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE)) {
                    # rescan with iscsiadm
                    $cmd = "/sbin/iscsiadm -m session --rescan";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0 and not grep(/No portal found|No session found|No matching sessions found/i, @output)) {
                        $self->_set_err("Failed to rescan the iSCSI sessions");
                        return $ClientUtil::FALSE;
                    }
                } else {
                    # first rescan with iscsiadm just to be safe (burt490436)
                    $cmd = "/sbin/iscsiadm -m session --rescan";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0 and not grep(/No portal found|No session found|No matching sessions found/i, @output)) {
                        $self->_set_err("Failed to rescan the iSCSI sessions");
                        return $ClientUtil::FALSE;
                    }

                    # then rescan using sg3_utils rescan scsi bus tool
                    $self->_rescan_scsi_bus(-protocol => "iscsi");
                    if ($self->error) {
                            return $ClientUtil::FALSE;
                    }
                }
            } elsif ($self->{sanbooted} == 0) {
                # otherwise restart the iscsi service to do the rescan
                $self->_iscsi_service(-action => 'restart');
                if ($self->error) {
                    return $ClientUtil::FALSE;
                }
            }
        } else {
            # Handle iSCSI HW Adapters rescan
            if ($self->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /serverengines/i) {
                $self->_rescan_iscsi_serverengines();
                if ($self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not rescan the ServerEngines iSCSI adapter\n" .
                                    "$errmsg_str\n");
                    return $ClientUtil::FALSE;
                }
            } elsif ($self->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /chelsio/i) {
                $self->_rescan_iscsi_chelsio();
                if ($self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not rescan the Chelsio iSCSI adapter\n" .
                                    "$errmsg_str\n");
                    return $ClientUtil::FALSE;
                }
            } elsif ($self->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /emulex/i) {
                $self->_rescan_iscsi_emulex();
                if ($self->error()) {
                    my $errmsg_str = join("\n", $self->errmsg());
                    $self->_set_err("Could not rescan the Emulex iSCSI adapter\n" .
                                    "$errmsg_str\n");
                    return $ClientUtil::FALSE;
                }
            } else {
                # For generic HW, use the rescan_scsi_bus command
                $self->_rescan_scsi_bus(-protocol => "iscsi");
                if ($self->error) {
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        # Recreating multipath maps
        $self->{mp}->recreate_maps(-conn => $self->{conn});
        if($self->{mp}->error()) {
            $self->_set_err($self->{mp}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    @devices = $self->_get_unlabeled_devices();
    return $ClientUtil::FALSE if ($self->error);
    $self->sanlog(-msg       => scalar(@devices) . " Devices are not part of a volume group",
                  -msg_level => 3);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub format {

    # No-op for Linux
    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    my ($cmd, @output, $status, $version);

    $self->_clear_err();

    my $path = $self->get_hu_bin_path();
    $cmd = $path . "/sanlun version| awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Error: LinuxHost::sanlun_version(): Unable to retrieve sanlun version");
        return "";
    }

    $version = $output[0];

    return $version;
}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my ($hsh, @output);

    $self->_clear_err();

    push @output, "Host name:      " . $self->hostname(), "Host type:      " . $self->hosttype(),
      "OS version:     " . $self->os_version(),     "Vendor release: " . $self->_get_release(),
      "Sanlun version: " . $self->sanlun_version(), "\n";

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;

    my ($interval, $cmd, $status, @output, $cnt, $count, @buf);

    $self->_clear_err();

    #  Default interval is 2 seconds
    $interval = 2;

    #  Default count is 1
    $count = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Error: LinuxHost::iostat() Invalid value for argument '-interval'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Error: LinuxHost::iostat() Missing value for argument '-count'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: LinuxHost::iostat() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    #  Remember that the first iostat cycle is wrong so add an extra cycle.
    #  Minimize the output return since 'z' flag doesn't kick in until second cycle.
    $count += 1;
    $cmd =
      "iostat -x -t $interval $count | grep -v  '0.00   0.00  0.00  0.00    0.00    0.00     0.00     0.00     0.00     0.00    0.00   0.00   0.00'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::iostat() ", "iostat returned a bad status ($status))");
        return (wantarray()) ? () : [];
    }

    #  Clean up the output by removing the first iostat cycle

    $cnt = 0;
    foreach my $line (@output) {
        if ($line =~ /Time/i) {
            $cnt++;
        }
        next if ($cnt == 1);
        push(@buf, $line);

    }

    push @buf, "\n\n";
    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_encrypted_devices
# Objective:       Will return encrypted devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_encrypted_devices {
    my ($self) = @_;

    my (@encrypted_devices);
    my ($cmd, $status, @output);

    $self->_clear_err();

    # Getting the list of LUKS enabled devices
    # Automation always will create encrypted device with crypt
    # prepended to its name
    $cmd = "ls /dev/mapper/ | grep -P \"^crypt-360a9|^crypt-3600a09\"";
    ($status, @encrypted_devices) = &ClientUtil::execute($self->{conn}, $cmd);
    # Fail if no device is found or there is an error
    if (scalar(@encrypted_devices) == 0 or $status > 0) {
        $self->sanlog(-msg => "No encrypted devices found");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @encrypted_devices : \@encrypted_devices;
}

################################################################################
# Subroutine Name: create_encrypted_devices
# Objective:       Will encrypt multipath devices
# Details:         See POD documentation at the end of this file
################################################################################
sub create_encrypted_devices {
    my ($self, %args) = @_;

    my ($devices, @encrypted_devices);
    my ($cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($devices)) {
        $self->_set_err("Missing required argument '-devices'");
        return $ClientUtil::FALSE;
    }

    if ($self->{major_revision} =~ /rhel4|sles9/i) {
        $self->_set_err("Device encryption is not supported in RHEL4 and SLES9 versions of Linux");
        return $ClientUtil::FALSE;
    }

    # Create a file that will be used as encryption key
    if ($self->{encryption_method} =~ /LUKS/i) {
        $cmd = "touch /var/tmp/nate_encryption_luks.key";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to create empty file " .
                            "/var/tmp/nate_encryption_luks.key");
            return $ClientUtil::FALSE;
        }
    }

    # Create encryted devices
    foreach my $device (@$devices) {
        $self->{mp}->create_encrypted_device(-device => $device);
        if ($self->{mp}->error) {
            $self->_set_err($self->{mp}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       get NFS mounts currently mounted on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    my $targets = undef;
    my @nfs_mounts;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    $cmd = "/bin/mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to execute command: " . $cmd);
        return (wantarray()) ? () : [];
    }

    # Capture the output of mount for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Look through each line of output and determint if there is a filer in the $targets list mounted to this host
        my @mounts = grep(/$target/i, @output);
        if (scalar(@mounts) == 0) {
            next;
        }

        # Go through the found mounts and populate the return array
        foreach my $mount (@mounts) {
            # fas2050-rtp04:/vol/nate_1_3_ibm3550_rtp13/nate_qtree_ibm3550-rtp13_linux_2 on /mnt/qa/fas2050-rtp04/nate_qtree_ibm3550-rtp13_linux_2 type nfs (rw,hard,intr,addr=10.60.242.51)
            my ($nfs_share, undef, $nfs_mount, undef, undef, $nfs_options) = split(/\s+/, $mount);
            $nfs_options =~ s/\(|\)//g;

            my $nfs_pt;
            $nfs_pt->{mount} = $nfs_mount;
            $nfs_pt->{share} = $nfs_share;
            $nfs_pt->{options} = $nfs_options;

            push(@nfs_mounts, $nfs_pt);
        }
    }

    if (scalar(@nfs_mounts) == 0) {
        $self->_set_err("Did not find any mounts associated with targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_mounts : \@nfs_mounts;
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       get available NFS shares volumes via showmount -e
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    my $targets = undef;
    my $interfaces = undef;
    my @nfs_shares;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?interfaces/) {
            if (not length($interfaces = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interfaces'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # If the interface list is not passed in, get it
    if (not defined($interfaces)) {
        $interfaces = $self->get_ip_info();
    }

    # Capture the output of showmount -e for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Collect showmount -e
        $cmd = "/usr/sbin/showmount -e " . $target;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Look for the two following erros, they failures.
        # showmount: <device_name>: RPC: Program not registered
        # showmount: <device_name>: RPC: Unknown host
        if (grep(/showmount\:.+RPC\:/, @output)) {
            $self->_set_err("RPC error occurred during command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        # Grep through the output and only get the ones that have an ip address in the interface list
        foreach my $interface (keys(%{$interfaces})) {
            my $ipv4 = $interfaces->{$interface}->{ipv4};
            my $ipv6 = $interfaces->{$interface}->{ipv6};
            my @results = ();

            # If we have an ipv4 address specified, add it to the results array
            if (defined($ipv4) and $ipv4 =~ /\d+\.\d+\.\d+\.\d+/) {
                push(@results, grep(/$ipv4/i, @output));
            }

            # If we have an ipv6 address specified, add it to the results array
            if (defined($ipv6)) {
                push(@results, grep(/$ipv6/i, @output));
            }

            if (scalar(@results) == 0) {
                $self->sanlog(-msg => "Did not find any shares for interface: $ipv4 $ipv6");
                next;
            }

            # Parse the output and populate nfs_mounts
            foreach my $line (@results) {
                # /vol/nate_1_1_hp2620_rtp11   10.60.242.103
                # /vol/nate_1_1_hprx2600_rtp14 10.60.242.105
                my ($vol) = split(/\s+/, $line);

                my $nfs_share;
                $nfs_share->{volume} = basename($vol);
                $nfs_share->{path} = $vol;
                $nfs_share->{ipv4_address} = $ipv4;
                $nfs_share->{ipv6_address} = $ipv6;
                $nfs_share->{target} = $target;

                push(@nfs_shares, $nfs_share);
            }
        }
    }

    # See if no NFS shares were found
    if (scalar(@nfs_shares) == 0) {
        $self->_set_err("No NFS shares found for targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_shares : \@nfs_shares;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server, $ntp_time_zone, $ntp_daemon);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: LinuxHost::ntp_sync() Invalid value for argument '-ntp_server'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: LinuxHost::ntp_sync() Invalid value for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # determine which ntp daemon to use depending on linux version
    my $new_ntpconf_file = "/etc/ntp.conf.tmp";
    my $new_steptickers_file = "/etc/ntp/step-tickers.tmp";
    my $old_ntpconf_file = "/etc/ntp.conf";
    my $old_steptickers_file = "/etc/ntp/step-tickers";

    if ($self->hosttype() =~ /suse|sles/i) {
        # sles uses ntp and has a different location for the step ticker files
        if ($self->{major_revision} =~ /sles9/) {
            $ntp_daemon = "xntpd";
        } else {
            $ntp_daemon = "ntp";
        }
        $new_steptickers_file = "/etc/step-tickers.tmp";
        $old_steptickers_file = "/etc/step-tickers";
    } else {
        # redhat uses ntpd
        $ntp_daemon = "ntpd";
    }

    # set the time zone
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");
    $cmd = "echo -e \"11\n" . $ntp_time_zone . "\n1\" | tzselect";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not set the timezone to '$ntp_time_zone'.");
        return $ClientUtil::FALSE;
    }

    # sync with the time server
    $self->sanlog(-msg => "Syncing the time with local NTP server: $ntp_server");

    # prepare the ntp conf file with the correct syntax
    $cmd = "echo restrict 127.0.0.1 > $new_ntpconf_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    $cmd = "echo restrict default kod nomodify notrap >> $new_ntpconf_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # add the time server to the conf file and step-tickers file
    $cmd = "echo server $ntp_server >> $new_ntpconf_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    $cmd = "echo $ntp_server > $new_steptickers_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    $cmd = "echo driftfile /var/lib/ntp/drift >> $new_ntpconf_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # back up the old ntp conf and step-tickers file
    # not checking for error because it is not critical.
    $cmd = "cp $old_ntpconf_file ${old_ntpconf_file}.bak";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    $cmd = "cp $old_steptickers_file ${old_steptickers_file}.bak";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # copy over the new ntp.conf and step-tickers file.
    $cmd = "mv $new_ntpconf_file $old_ntpconf_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    $cmd = "mv $new_steptickers_file $old_steptickers_file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not run $cmd");
        return $ClientUtil::FALSE;
    }

    # restart the ntp service
    if ($self->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl restart $ntp_daemon" . ".service";
    } else {
        $cmd = "/etc/init.d/$ntp_daemon restart";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not start the ntpd service");
        return $ClientUtil::FALSE;
    }

    # make it so ntpd starts automatically.
    if ($self->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl enable $ntp_daemon" . ".service";
    } else {
        $cmd = "chkconfig --level 345 $ntp_daemon on";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not set ntpd in level 3 4 and 5 in chkconfig");
        return $ClientUtil::FALSE;
    }

    # sync the hw clock to ntp time
    $cmd = "hwclock --systohc";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not sync the hwclock to NTP time.");
        return $ClientUtil::FALSE;
    }

    # display the date
    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not obtain the date.");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       Will update the host partition table. After creation or deletion
#                  of partitions host doesnt update partition table automatically
#                  soon enough. We need to update the table manually to reflect
#                  the changes as soon as possible
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {
    my ($self) = @_;

    my $cmd = "partprobe";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to update the partition table");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: netapp_scsi_show()
# Objective:       displays information of scsi devices
################################################################################
sub netapp_scsi_show {
    my ($self) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd);

    # we use sg_modes tool in the script. Hence, make sure that tool is available on the host.
    $cmd = "sg_modes -h";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("sg_modes tool is not available. It is required to run the script.");
        return (wantarray()) ? () : [];
    }

    # copy the script if it is not there already
    $cmd = "ls /sbin/netapp_scsi_show";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        # obtain the path to the p4 dir being used for this test
        my $p4_dir = &SanCIT::get_p4_dir();
        my $local_file = "${p4_dir}/SANInterop/tools/linux/netapp_scsi_show";
        my $remote_file = "/sbin/netapp_scsi_show";

        # transfer the dmmp discovery script to the remote host.
        if ($self->ClientUtil::scp_file(-local_file => $local_file,
                                        -remote_file => $remote_file) == $ClientUtil::FALSE
            or $self->error()) {
            $self->_set_err("SCP transfer failed", $self->errmsg());
            return (wantarray()) ? () : [];
        }
    }

    $cmd = "perl /sbin/netapp_scsi_show";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to execute the script");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Subroutine Name: display_time()
# Objective:       display the current system time
# Details:         See POD documentation at the end of this file
################################################################################
sub display_time {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd = "date";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/sbin/shutdown -h now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;

    my ($num_tries, $alive, $retry, @output, $status, $cmd, $machine);

    $self->_clear_err();

    $machine = $self->hostname(),

    $self->sanlog(-msg => "Rebooting Client $machine");

    # reboot the test machine
    $cmd = "/sbin/shutdown -r now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # detect if the host actually rebooted or not
    $self->detect_reboot(-shutdown_timeout => 300,
                         -bootup_timeout   => 900,
                         -bootup_sleep     => 300);
    return $ClientUtil::FALSE if ($self->error);

    #
    # After we ping the box, it might be a little while before we
    # stop getting connection refused on the reconnect.
    #
    &SanUtil::nate_sleep(120);

    # reconnect to the client and retry this operation for 10 minutes
    # before giving up and erroring out.
    $self->reconnect();

    my $retries = 10;
    while ($self->error and $retries > 0) {
       $self->sanlog(-msg =>
                     "LinuxHost::reboot() reconnect failed, still ($retries) retries left");
       &SanUtil::nate_sleep(60);
       $retries--;
       $self->reconnect();
    }

    # if we still have an error after 10 minutes of
    # retries, error out.
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    $self->sanlog(-msg       => "Running Uname on $machine\n",
                  -msg_level => 3);
    ($status, @output) = &ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::reboot() Command ($cmd) " . "returned a bad status ($status)");
        return "";
    }

    # Start sysrq dump is requested
    if ($self->{client_linux_enable_sysrq_dump} == 1) {
        # Host must have been rebooted by now so we can start the sysrq dump on the Client
        # Hard coding the port value. This port number will be used to established a socket
        # connection between host and nate control server(NCS). We can use the same port
        # number to connect to each host because socket is combination of port and hostname/ip.
        $self->_start_sysrq_dump_collection_scripts();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "LinuxHost::configure() Started SysRq dump script after host reboot",
                      -log_level => 2);
    }

    $self->sanlog(-msg       => "Hostname=" . join("\n", @output),
                  -msg_level => 3);
    $self->{reboot} = 0;
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_nfs_mounted
# Objective:       Determine if NFS volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_nfs_mounted {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{nfs_mount} and length $self->{nfs_mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    my ($cmd, $directio, $hide, $resource, $mount_pt, $options, $nodename, $vol_name, $name, $status, @output);

    $self->_clear_err();
    $options = "rw,hard,intr";
    $hide    = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?resource/) {
            if (not length($resource = $args{$_})) {
                $self->_set_err("Error: LinuxHost::nfs_mount() Invalid value for argument '-resource'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?directio/) {
            if (not length($directio = $args{$_} or $directio !~ /1|0/)) {
                $self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Error: LinuxHost::nfs_mount() Missing value for argument '-options'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: LinuxHost::nfs_mount() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    # Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource) {
        $self->_set_err("Error: LinuxHost::nfs_mount() ", "Missing required argument '-resource'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt'
    unless (defined $mount_pt and length $mount_pt) {
        #  Use default path if not defined
        ($nodename = $resource) =~ s/(^\S+):\S+/$1/;

        #Check for empty value
        if ($nodename eq "") {
            $self->_set_err("Error: HPUXHost::nfs_mount() Unable to determine filer name from device mount '$resource'",
                            @output);
            return $ClientUtil::FALSE;
        }

        $vol_name = basename($resource);
        $mount_pt = "/mnt/qa/" . $nodename . "/" . $vol_name;
        $self->sanlog(-msg => "Using default mount pt '$mount_pt' for $resource");
    }

    # check if the volume is already mounted
    $self->mount_table_verify($mount_pt);
    if ($self->error) {
        #  Create the mount point and open up dir privs
        $cmd = "/bin/mkdir -p $mount_pt; /bin/chmod -R 777 $mount_pt";

        @output = $self->execute($cmd, -hide => $hide);
        if ($self->error) {
            $self->sanlog(-msg => "WARNING: LinuxHost::nfs_mount() Unable to create mount "
                            . "point directory for volume '$resource'\n"
                            . join "\n",
                          @output);
            $self->_clear_err();
        }

        # Directio is not supported on Redhat versions through mount options. Its
        # the application responsibility for setting up directio connectivity.
        if ($directio) {
            if ($options =~ /forcedirectio|noquota/) {
                $self->sanlog(-msg => "INFO: LinuxHost::nfs_mount() option -forcedirectio and -noquota not supported.");
                $options =~ s/forcedirectio,//g;
                $options =~ s/noquota,//g;
            }
        }

        #  Remove option param if user passed it in as part of the
        #  options string. Lib should add the tag.
        $options =~ s/-o //g;

        # Create the NFS mount command.
        if (defined $options) {
            $cmd = "/bin/mount -t nfs -o $options $resource $mount_pt";
        } else {
            $cmd = "/bin/mount -t nfs $resource $mount_pt";
        }

        @output = $self->execute($cmd, -hide => $hide);
        if ($self->error) {
            $self->_set_err("Error: LinuxHost::nfs_mount() Unable to NFS mount volume", "Code: $status", @output);
            return $ClientUtil::FALSE;
        }
    } else {
        $self->sanlog(-msg => "Warning: LinuxHost::nfs_mount() volume ($mount_pt) is already mounted");
        $self->_clear_err();
        return $ClientUtil::TRUE;
    }

    #  For now lets not add mount points into the fstab files.  Most test folks probably
    #  don't care about that.  However, keep the code around if someone changes mind.
    #
    # $self->mount_entry_add(
    #     -device_mount=>$resource,
    #     -mount_pt=>$mount_pt,
    #     -fs_type=>"nfs",
    #     -mount_options=>$options,
    #     );
    #
    # if ($self->error) {
    #     $self->_set_err($self->errmsg());
    #     return $ClientUtil::FALSE;
    # }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount() {
    my ($self, %args) = @_;

    my ($cmd, $force, $hide, $mount_pt, $name, $status, @output);

    $self->_clear_err();

    $hide = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {

            #  Make sure mount point exists and is not '/'.
            if (not length($mount_pt = $args{$_}) and ($args{$_} != "/")) {
                $self->_set_err("Error: LinuxHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: LinuxHost::nfs_mount() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: LinuxHost::nfs_mount() ", "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    # Create the mount point and open up dir privs
    $cmd = "umount $mount_pt";
    @output = $self->execute($cmd, -hide => $hide);
    if ($self->error) {
        $self->_set_err("Error: LinuxHost::nfs_unmount() " . "Unable to unmount NFS volume " . $mount_pt . " ",
                        @output);
        return $ClientUtil::FALSE;
    }

    #  For now lets not delete mount points in the fstab files
    # $self->mount_entry_delete($mount_pt);
    # if ($self->error) {
    #    $self->_set_err($self->errmsg());
    #    return $ClientUtil::FALSE;
    # }

    #  Create the command to remove the NFS mount point
    $cmd = "rm -Rf " . $mount_pt;
    @output = $self->execute($cmd, -hide => $hide);
    if ($self->error) {
        $self->_set_err("Warning: LinuxHost::nfs_unmount() Unable to remove "
                          . "mount point directory '"
                          . $mount_pt . "'\n"
                          . join "\n",
                        @output);

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       add entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;

    my ($cmd,     $device_mount, $device_fsck, $label,         $mount_pt, $fs_freq,
        $fs_type, $fsck_auto,    $mount_boot,  $mount_options, $status,   @output);

    $self->_clear_err();

    #  initialize variables
    $fs_freq       = "0";
    $fsck_auto     = "0";
    $mount_options = "nocheck";

    # parse arguments
    foreach (keys %args) {
        if (/-?device_mount/) {
            if (not length($device_mount = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Invalid value for argument '-device_mount'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?device_fsck/) {

            #  Not supported
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?label/) {
            if (not length($label = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-label'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?fs_type/) {
            if (not length($fs_type = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-fs_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?fsck_auto/) {
            if (not length($fsck_auto = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-fsck_auto'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_boot/) {

            #  Not supported
            next;
        }

        elsif (/-?mount_options/) {
            if (not length($mount_options = $args{$_})) {
                $self->_set_err("Error: LinuxHost::mount_entry_add() Missing value for argument '-mount_options'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: LinuxHost::mount_entry_add() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    #  Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount) {
        $self->_set_err("Error: LinuxHost::mount_entry_add() ", "Missing required argument '-device_mount'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: LinuxHost::mount_entry_add() ", "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-fs_type' was provided
    unless (defined $fs_type and $fs_type =~ /ext3|nfs/i) {
        $self->_set_err("Error: LinuxHost::mount_entry_add() ", "Missing required argument '-fs_type'");
        return $ClientUtil::FALSE;
    }

    #  Mounting a partition uses different semantics
    if (length $label) {
        $cmd =
            "echo LABEL=$label" . "\t"
          . $mount_pt . "\t"
          . $fs_type . "\t"
          . $mount_options . "\t"
          . $fs_freq . "\t"
          . $fsck_auto . " "
          . ">> /etc/fstab";
    } else {
        $cmd = "echo "
          . $device_mount . "\t"
          . $mount_pt . "\t"
          . $fs_type . "\t"
          . $mount_options . "\t"
          . $fs_freq . "\t"
          . $fsck_auto . " "
          . ">> /etc/fstab";
    }

    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table");

    #  execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::mount_entry_add Unable to add mount entry to /etc/fstab file", @output);
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;

    my ($cmd, $mount, $status, @output);

    $self->_clear_err();

    #  Unmount the volume's filesystem
    $cmd = "umount $mntpt";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:LinuxHost ::unmount_fs() Unable to unmount filesystem", @output);
    }

    $self->sanlog(-msg => "Removing mount point '$mntpt'");

    #  Remove directory head
    $cmd = "rm -rf $self->{mount}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::unmount_fs() Unable to remove mount point '$mount'", @output);
    }

    return $self;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    #  Clear vfstab mount entry
    $self->sanlog(-msg => "Clear /etc/filesystems mount entry");

    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: LinuxHost::mount_entry_delete Missing required paremeter '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    $mount =~ s/\//\\\//g;

    $cmd = "perl -e\'"
      . "open(FH,\"< \/etc\/fstab\") or exit(1);"
      . "open(FD,\"> \/etc\/fstab_new\") or exit(1);"
      . "while(<FH>){\\\n"
      . "print FD \$_ if (\$_ !~ /$mount/);" . "}\\\n"
      . "close FD;"
      . "close FH;\\\n"
      . "rename(\"/etc/fstab_new\", \"/etc/fstab\") or exit(1);" . "\' ";

    #  execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::mount_entry_delete Unable to remove mount entry from /etc/fstab file",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify {
    my ($self, $mount, $dev) = @_;

    my ($cmd, $entry, $idx, $status, $value, @values, @params, @output, @buf);

    $self->_clear_err();

    # initialize
    $entry = "";
    $idx   = 0;

    # get the /etc/fstab contents
    if ($dev) {
        $cmd = "cat /etc/fstab | grep -v '^#' | egrep '^$mount' ";
    } else {
        $cmd = "cat /etc/fstab | grep -v '^#' | egrep '    $mount    | $mount ' ";
    }

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::mount_entry_verify(): Unable to verify mount file entry", @output);
        return $entry;
    }

    if (scalar(@output) == 0) {
        $self->_set_err("Error: LinuxHost::mount_entry_verify(): No mount file entry exists");
    } elsif (scalar(@output) > 1) {
        $self->_set_err("Error: LinuxHost::mount_entry_verify(): Multiple mount file entries exist", @output);
    } else {
        @params = ("Device to mount:", "Mount pt:", "Fs type:", "Mount Options:");

        @values = split(/\s+/, $output[0]);

        foreach $idx (0 .. (scalar(@params) - 2)) {
            push @buf, $params[$idx] . " " . $values[$idx];
        }

        my $startOpt = (scalar(@params) - 1);
        my $endOpt   = (scalar(@values) - 1);

        push @buf, $params[$startOpt] . " @values[$startOpt..$endOpt]";
    }

    (wantarray()) ? @buf : \@buf;
}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;

    my ($status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or length("@output") eq 0) {
        $self->_set_err("Error: LinuxHost::mount_table_verify(): Unable to verify mount table", @output);
        return (wantarray()) ? @output : \@output;
    }

    unless (grep /$mount/, @output) {
        $self->_set_err("Error: LinuxHost::mount_table_verify(): Unable to verify mount table", @output);
        return (wantarray()) ? @output : \@output;
    }

    #  Need to parse the mount into some usefule structure.  For now just show
    #  that something was found.

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);

    # check if the mount point lives on something that is not mounted on /
    $cmd = "df -k $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);

    # Check the status of the mount command
    if ($status != 0) {
        $self->_set_err("Unable to execute command ($cmd)");
        return $ClientUtil::FALSE;
    }

    #[root@auto-rhel6-01 ~]# df -k /mnt/engenx/tb3170_03/tb3170_03_vs/
    #Filesystem           1K-blocks      Used Available Use% Mounted on
    #/dev/mapper/vg_autorhel6template-lv_root
    #                       5716804   4101464   1324936  76% /

    #[root@auto-rhel6-01 ~]# df -k /mnt/engenx/tb3170_03/tb3170_03_vs/192.168.10.231/nas_vol
    #Filesystem           1K-blocks      Used Available Use% Mounted on
    #192.168.10.231:/nas_vol
    #                       1992320       192   1992128   1% /mnt/engenx/tb3170_03/tb3170_03_vs/192.168.10.231/nas_vol

    # make sure that $mount is not mounted on '/' (unless it is /)
    my $match  = grep(/\%\s*\/\s*$/, @output);
    if ($match and $mount eq '/') {
        return $ClientUtil::TRUE;
    } elsif ($match) {
        # the mount point is not an independent mount point (mounted on /)
        $self->_set_err("Unable to locate mount point ($mount)");
        return $ClientUtil::FALSE;
    } else {
        # mount point is not on /
        return $ClientUtil::TRUE;
    }
}

################################################################################
# Subroutine Name: filer_mount_entry_backup()
# Objective:       backup the filer /etc/exports file to /tmp for later restore
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_backup {
    my ($self, $mount_pt) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_restore() mount_pt is a required argument");
        return $ClientUtil::FALSE;
    }

    # Save the original /etc/exports file for restore at cleanup
    $self->sanlog(-msg => "Backup the original /etc/exports file from \"$mount_pt\" for later restore");

    # backup original /etc/exports
    $cmd = "cp -p ${mount_pt}/etc/exports ${mount_pt}/etc/exports.ORIG";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_backup() Unable to backup \"${mount_pt}/etc/exports\"",
                        "to \"${mount_pt}/etc/exports.ORIG\"", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: filer_mount_entry_restore()
# Objective:       restore the filer /etc/exports file from /tmp
# Details:         See POD documentation at the end of this file
################################################################################
sub filer_mount_entry_restore {
    my ($self, $mount_pt) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_restore() mount_pt is a required argument");
        return $ClientUtil::FALSE;
    }

    # Restore the original /etc/exports file
    $self->sanlog(-msg => "Restore the original /etc/exports file from \"$mount_pt\"");

    # restore original /etc/exports and remove old commented lines
    $cmd = "cat ${mount_pt}/etc/exports.ORIG | grep -v '^#/vol' > ${mount_pt}/etc/exports";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::filer_mount_entry_restore() Unable to restore \"${mount_pt}/etc/exports\"",
                        "from \"${mount_pt}/etc/exports.ORIG\"", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self, $mount) = @_;

    my ($cmd, $line, $status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    $cmd = "mount | grep $mount | awk '{print " . '$1' . "}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or length("@output") eq 0) {
        $self->_set_err("Error: LinuxHost::get_mounted_device(): Mount command failed!", @output);
        return undef;
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Error: LinuxHost::get_mounted_device(): Multiple mount entries found!", @output);
        return undef;
    }

    #  if the device is a partition, return the physical disk
    chop($output[0]) if ($output[0] =~ /dev\/sd.+\d$/);

    return $output[0];
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported LVMs for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(nativelvm);

}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(iscsi fcp nfs);

}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs {
    my ($self) = @_;

    $self->_clear_err();

    return (undef);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    $self->_clear_err();

    return qw(ext3 ext2);

}

################################################################################
# Subroutine Name: config_paths
# Objective:       N/A for Linux, provide for compatibility with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop, or call _rescan_scsi_bus() with option 'remove_LUNs'
#                  to delete unmapped LUNs from the host.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self, $protocol) = @_;

    my ($cmd, $major, $minor, $status, @output);

    $major = $self->{major_revision};
    $minor = $self->{minor_revision};

    $self->_clear_err();

    if ($protocol =~ /iscsi/i) {
        # Stop iSCSI if this the SW initator. For hardware
        # assumption is that this support the rescan_scsi_bus.sh
        # script and it can cleanup with that script.
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/i) {
            $self->_iscsi_service(-action => 'stop');
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_rescan_scsi_bus(-option => 'remove_LUNs', -protocol => "iscsi");
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        }
    }
    if ($protocol =~ /fcp/i) {
        if (($major =~ /sles9|rhel4/i)
            or (($major =~ /sles10/i) and $minor < 3)
            or (($major =~ /rhel5/i) and $minor < 4)
            or ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE)) {
            # Unload fcp driver modules
            $self->_unload_fcp_module();
            return $ClientUtil::FALSE if ($self->error);

            # load fcp driver modules so that "sanlun fcp show adapter -v"
            # will work on next run
            $self->_load_fcp_module();
            return $ClientUtil::FALSE if ($self->error);
        } else {
            $self->_rescan_scsi_bus(-option => 'remove_LUNs', -protocol => "fcp");
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup()
# Objective:       When the luns are removed from the filer, we will have
#                  stale entries of sdX devices. This causes big problems.
#                  So these devices should be removed.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    my ($protocol, $major, $minor, $cmd, $status, @output);

    $self->_clear_err();

    $major = $self->{major_revision};
    $minor = $self->{minor_revision};
    $protocol = $self->{protocol};

    if ($protocol =~ /fcp/i) {
        # From SLES 10.3 and RHEL 5.4 onwards, native rescan script which comes with
        # sg3_utils provide option to remove stale entries of sdX devices.
        # Otherwise, we have to reboot the host to do this.
        if (($major =~ /sles9|rhel4/i)
           or (($major =~ /sles10/i) and $minor < 3)
           or (($major =~ /rhel5/i) and $minor < 4)
           or ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE)) {
            $self->reboot();
            if ($self->error) {
                $self->_set_err($self->errmsg);
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_rescan_scsi_bus(-option => 'remove_LUNs', -protocol => "fcp");
            if ($self->error) {
                $self->_set_err($self->errmsg);
                return $ClientUtil::FALSE;
            }
        }
    }

    if ($self->{protocol} =~ /iscsi/i) {
        # From SLES 10.3 and RHEL 5.4 onwards, native rescan script which comes with sg3_utils
        # provide option to remove stale entries of sdX devices.
        # Otherwise, we have to follow some other methods to do this.
        if ((($major =~ /sles10/i) and $minor < 3)
           or (($major =~ /rhel5/i) and $minor < 4)
           or ($self->_is_rpm_installed(-rpm => "sg3_utils") == $ClientUtil::FALSE)) {
            if (($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/i) and ($self->{sanbooted} == 0)) {
                # if it is software iSCSI and it is non SANbooted, restart the iSCSI service
                # to remove stale entries
                $self->_iscsi_service(-action => 'restart');
                if ($self->error) {
                    $self->_set_err($self->errmsg);
                    return $ClientUtil::FALSE;
                }
            } else {
                # HW iSCSI adapter or software iSCSI-SANbooted, reboot the host to remove stale entries
                $self->reboot();
                if ($self->error) {
                    $self->_set_err($self->errmsg);
                    return $ClientUtil::FALSE;
                }
            }
        } else {
            # use native rescan script to remove stale entries
            $self->_rescan_scsi_bus(-option => 'remove_LUNs', -protocol => "iscsi");
            if ($self->error) {
                $self->_set_err($self->errmsg);
                return $ClientUtil::FALSE;
            }
        }
    }

    # flush multipath devices which are not in use for DMMP
    if (defined $self->{mp} and $self->{mp_type} =~ /dmmp/i) {
        $self->{mp}->flush(-conn => $self->{conn});
        if ($self->{mp}->error) {
            # ignore the error
            $self->{mp}->_clear_err();
        }
    }

    # Printing sanlun output for debug purpose to make sure that
    # all the orphanned devices are cleaned up properly
    my $path = $self->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show all";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 and $status != 203) {
        $self->_set_err("Unable to execute $cmd");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup()
# Objective:       Flush multipath devices which are not in use. Before that
#                  encrypted devices should be removed if they are present
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self, %args) = @_;

    my ($use_encryption, $encryption_method);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?use_encryption/) {
            if (not length($use_encryption = $args{$_})) {
                $self->_set_err("Invalid value for argument '-use_encryption'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?encryption_method/) {
            $encryption_method = $args{$_};
            next;
        }
    }

    # Make sure to close any encryption device
    # if encryption is used to create devices
    if ($use_encryption == 1) {
        # Close LUKS encrypted device
        if ($encryption_method =~ /LUKS/i) {
            $self->_close_encrypted_devices_luks();
            if ($self->error) {
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Encryption is enabled and Encryption method ($encryption_method) is not valid");
            return $ClientUtil::FALSE;
        }
    }

    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        # flush multipath devices which are not in use
        $self->{mp}->flush(-conn => $self->{conn});
        if ($self->{mp}->error) {
            my @mp_error = $self->{mp}->errmsg();
            $self->_set_err(-msg => "@mp_error");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to Linux syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    foreach my $msg (@message) {
        $cmd = "logger -t \"**NATE**\" -p user.error \"$msg\"";
        ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 10, 2);
        if ($status != 0) {
            $self->_set_err("Error: LinuxHost::syslog_msg(): Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Method:    get_system_stats()
# Objective: Get the cpu/memory usage percentage over a specified interval
################################################################################
sub get_system_stats {
    my ($self, $interval) = @_;

    $self->_clear_err();

    # collect system statistics for $interval length
    my $cmd = "rm /tmp/nate_sysstats; top -b -d 1 -n $interval > /tmp/nate_sysstats";
    my ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, $interval + 30, 2, "-comment");
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::get_system_stats(): failed to collect system info.");
        return $ClientUtil::FALSE;
    }

    # parse out only the lines relevant to CPU, Memory, and Swap
    $cmd = "cat /tmp/nate_sysstats | grep -P \"Mem|Cpu|Swap\"";
    ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, $interval + 30, 2, "-comment");
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::get_system_stats(): failed to collect system info.");
        return $ClientUtil::FALSE;
    }

    my $system_stats = {};
    $system_stats->{num_cpus} = 1;
    my $cpu_samples = 0;
    my $mem_samples = 0;
    my $cpu_flag = 0;
    my $mem_flag = 0;
    my $mem_line_found = 0;
    my $used = 0;
    my $total = 0;
    my $buffers = 0;

    # @output would like this:
    # Cpu(s):  0.1%us,  0.4%sy,  0.0%ni, 99.4%id,  0.1%wa,  0.0%hi,  0.0%si,  0.0%st
    # Mem:   1922128k total,  1687868k used,   234260k free,    19184k buffers
    # Swap:  4128760k total,      724k used,  4128036k free,  1264108k cached
    # Cpu(s):  0.0%us,  0.5%sy,  0.0%ni,  0.0%id, 99.5%wa,  0.0%hi,  0.0%si,  0.0%st
    # Mem:   1922128k total,  1698904k used,   223224k free,    23116k buffers
    # Swap:  4128760k total,      724k used,  4128036k free,  1270672k cached

    # parse out all cpu and memory samples and tally them up
    foreach my $line (@output) {
        if ($line =~ /Cpu\(s\)\:\s*\S+?\%\s*us,\s*\S+?\%\s*sy,\s*\S+?\%\s*ni,\s*(\d+\.\d+)\%\s*id/i) {
            if ($cpu_flag == 0) {
                # first reading of CPU is skewing the data because it represents an average since boot.
                # So, we are neglecting first reading.
                $cpu_flag = 1;
                next;
            } else {
                # parse out the idle time and subtract it from 100 to get used cpu %
                $system_stats->{cpu_usage_percent} += int(100 - $1);
                $cpu_samples++;
            }
        } elsif ($line =~ /Mem:\s*(\d+)(t|g|m|k)? total,\s*(\d+)(t|g|m|k)? used,.*\s+(\d+)(t|g|m|k)? buffers/i) {
            if ($mem_flag == 0) {
                # first reading of memory is skewing the data because it represents an average since boot.
                # So, we are neglecting first reading.
                $mem_flag = 1;
                next;
            } else {
                $mem_line_found = 1;
                $total = $1;
                $used = $3;
                $buffers = $5;
            }
        } elsif ($line =~ /Swap:.*\s+(\d+)(t|g|m|k)? cached/i) {
            if ($mem_line_found != 1) {
                next;
            } else {
                my $cached = $1;

                # Exclude buffers and cached from used memory
                my $used_memory = $used - $buffers - $cached;

                my $mem_usage = int(($used_memory / $total) * 100);
                $system_stats->{mem_usage_percent} += $mem_usage;
                $mem_samples++;
                $mem_line_found = 0;
            }
        }
    }

    # average out cpu usage and mem usage from all the tallied samples.
    if ($cpu_samples > 0) {
        $system_stats->{cpu_usage_percent} = int($system_stats->{cpu_usage_percent} / $cpu_samples);
    }
    if ($mem_samples > 0) {
        $system_stats->{mem_usage_percent} = int($system_stats->{mem_usage_percent} / $mem_samples);
    }

    return $system_stats;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self) = @_;

    my ($mac, $ipv4, $ipv6, $gateway, $subnet_mask, $ipv6_prefix_len);
    my ($set_interface, $mtusize, $broadcast);
    my $interface_name = 'NA';

    my $start = 0;
    my $finished = 0;
    my $interface = 0;
    my $ip_hash = {};

    $self->_clear_err();

    # get all IP information from host via ifconfig
    my $cmd = "ifconfig -a";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    my @interfaces = split(/\n\n/, join("\n", @output));


    foreach my $interface (@interfaces) {
        my @lines = split(/\n/, $interface);

        # initialize values
        $mac = $mtusize = $broadcast = $gateway = $subnet_mask = "";
        $ipv4 = $ipv6 = "";

        foreach my $line (@lines) {
            # get interface name
            if ($line =~ /^\s*(\S+)\s+flags/) {
                $interface = $1;
                $interface =~ s/\://;
            } elsif ($line =~ /^\s*(\S+)\s+Link/) {
                $interface = $1;
            }

            # If the interface is not set skip to the next line
            if ($interface eq "") {
                next;
            }

            # get IPv4 address
            if ($line =~ /inet addr:\s*(\S+)/) {
                $ipv4 = $1;
            } elsif ($line =~ /inet\s+(\S+)/) {
                $ipv4 = $1;
            }

            # get IPv6 address
            if ($line =~ /inet6 addr:\s*(\S+)\/(\d+)/) {
                $ipv6 = $1;
                $ipv6_prefix_len = $2;
            } elsif ($line =~ /inet6\s*(\S+)\/(\d+)/) {
                $ipv6 = $1;
                $ipv6_prefix_len = $2;
            }

            # get MAC address
            if ($line =~ /HWaddr\s*(\S+)/i) {
                $mac = $1;
                $mac =~ s/\://g ;
            }

            # no gateway for UNIX
            $gateway = "";

            # get subnet mask
            if (($line =~ /Mask:\s*(\S+)/i) or ($line =~ /netmask\s*(\S+)/i)) {
                $subnet_mask = $1;
            }

            # get broadcast
            if (($line =~ /Bcast:\s*(\S+)/i) or ($line =~ /broadcast\s*(\S+)/i)) {
                $broadcast = $1;
            }

            # get mtusize
            if ($line =~ /MTU:\s*(\S+)/i) {
                $mtusize = $1;
            }
        }

        # If the interface is empty skip this and go onto the next interface
        if ($interface eq "" or ($ipv4 eq "" and $ipv6 eq "")) {
            next;
        } elsif ($interface =~ /lo|sit0/) {
            next;
        }

        $ip_hash->{$set_interface}->{interface_name} = $interface_name;
        $ip_hash->{$interface}->{mac} = $mac;
        $ip_hash->{$interface}->{description} = "NIC";
        $ip_hash->{$interface}->{gateway} = $gateway;
        $ip_hash->{$interface}->{broadcast} = $broadcast;
        $ip_hash->{$interface}->{subnet_mask} = $subnet_mask;
        $ip_hash->{$interface}->{mtusize} = $mtusize;
        $ip_hash->{$interface}->{ipv4} = $ipv4;
        $ip_hash->{$interface}->{ipv6} = $ipv6;

        # calculate the ipv4 subnet
        if ($ipv4 ne "" and $subnet_mask ne "") {
            $ip_hash->{$interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                            -ip_version => "ipv4",
                                                                            -ipv4_netmask => $subnet_mask);
        } else {
            $ip_hash->{$interface}->{ipv4_subnet} = "";
        }

        # calculate the ipv6 subnet
        if ($ipv6 ne "" and $ipv6_prefix_len ne "") {
            $ip_hash->{$interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
            $ip_hash->{$interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                            -ip_version => "ipv6",
                                                                            -ipv6_prefix_len => $ipv6_prefix_len);
        } else {
            $ip_hash->{$interface}->{ipv6_prefix_len} = "";
            $ip_hash->{$interface}->{ipv6_subnet} = "";
        }

        # if this is the management interface of the client
        # set an attribute
        my $host_rec = &Tharn::host($self->hostname());
        my $host_ip_default = $host_rec->{'default_ip'};
        if ($ipv4 eq $host_ip_default) {
            $ip_hash->{$interface}->{is_mgmt_interface} = 1;
        } else {
            $ip_hash->{$interface}->{is_mgmt_interface} = 0;
        }

        # convert the MAC for ServerEngines and Emulex iSCSI cards by
        # adding 1 to the current MAC (from ifconfig). Only convert if
        # this is not the management interface
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /hardware/i and
            $self->{iscsi}->{CONFIG_ISCSI_HARDWARE_VENDOR} =~ /serverengines|emulex/i and
            $ip_hash->{$interface}->{is_mgmt_interface} == 0) {
            my $set_mac = $ip_hash->{$interface}->{mac};
            my $mac_last = substr($set_mac, length($set_mac) - 2, 2);
            my $mac_first = substr($set_mac, 0, length($set_mac) - 2);
            my $new_mac_last = hex($mac_last) + 1;
            $new_mac_last = sprintf("%x", $new_mac_last);
            $ip_hash->{$interface}->{mac} = $mac_first . $new_mac_last;
        }
    }

    # return info hash
    return ($ip_hash);

}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    $self->_clear_err();

    # "NONE" is returned for Linux if the device-mapper-multipath cannot
    # be found. If it can, the default will be DMMP.
    my $cmd = "rpm -q device-mapper";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        return("NONE");
    } else {
        return("DMMP");
    }

}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for Linux
################################################################################
sub rotate_logs {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # rotate the syslog files
    $cmd = "/usr/sbin/logrotate -f /etc/logrotate.d/syslog";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # rotate the Veritas logs
    $status = $self->rotate_generic_log(-file => "/etc/vx/dmpevents.log", -minsize => "1024", -keep => "5");
    if ($status == $ClientUtil::TRUE) {
        return $ClientUtil::TRUE;
    } else {
        if (not $self->error) {
            $self->_set_err("An error was received while attempting to rotate the Veritas DMP logs");
        }
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: get_devices()
# Objective:       wrapper to call _get_devices() routine
################################################################################
sub get_devices {
    my ($self) = @_;

    $self->_clear_err();

    return ($self->_get_devices());

}

################################################################################
# Method:    get_iscsi_info
# Objective: Collects iscsi info from the client.
#            Packages the information into a complex data structure.
#            Builds a hash of target nodenames.
#            Under each target nodename, there is a hash for each iscsi cmd:
#            alias      - contains scalar value (Can't get for linux)
#            connection - contains an array of hashes
#            initiator  - contains an array of hashes
#            interface  - contains an array of hashes
#            nodename   - contains scalar
#            security   - contains a hash
#            session    - contains an array of hashes
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
#            Also added this hash entry which does not map to a command:
#            discovery_type - will return sendtargets or isns
# Details:
#     Inputs  - none
#     Returns $iscsi_hash - hash of client iscsi info
#             empty hash (do not set_err)  - no sessions or target portals
#             empty hash (set_err) - error condition
################################################################################
sub get_iscsi_info {
    my ($self) = @_;

    my $iscsi_info;

    $self->_clear_err();

    # Check for iscsi type.
    if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
        $iscsi_info = $self->_get_iscsi_info_openiscsi();
        if ($self->error()) {
            $self->_set_err("_get_iscsi_info_openiscsi failed!",
                            $self->errmsg());
            return {};
        }
    } else {

        # Other types not currently supported
        $self->_set_err("get_iscsi_info does not currently support "
                        . $self->{iscsi}->{iscsi_version});
        return {};
    }

    return $iscsi_info;

} ## end sub get_iscsi_info

################################################################################
# Subroutine Name: display_additional_info()
# Objective:       dump additional OS specific information
################################################################################
sub display_additional_info {
    my ($self) = @_;

    my ($status, @output, $cmd);

    $self->_clear_err();

    # get information on the physical system in use (platform/CPU/memory)
    # do not die if for some reason this does not run
    $cmd = "dmidecode";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # list out install RPMs
    # do not die if for some reason this does not run
    $cmd = "rpm -qa";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($self->{protocol} =~ /iscsi/i) {
        # list out all the sessions with verbosity
        # do not die if for some reason this does not run
        $cmd = "/sbin/iscsiadm -m session -P 3";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    if ($self->{mp_type} =~ /dmmp/i) {
        # Print contents of multipath conf file
        # do not die if for some reason this does not run
        $self->sanlog(-msg => "Printing contents of multipath configuration file");
        $cmd = "cat /etc/multipath.conf";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        $status = $self->{mp}->_is_multipathd_running(-conn => $self->{conn});
        if ($status == $ClientUtil::FALSE) {
            $self->{mp}->start_multipath_daemon();
            if ($self->{mp}->error()) {
                $self->_set_err("Failed to start multipath daemon");
                return $ClientUtil::FALSE;
            }
            # wait for sometime to get daemon started
            &SanUtil::nate_sleep(5);
        }

        # do not die if for some reason this does not run
        $cmd = "multipathd -k\"show config\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_mixed_fs_template()
# Objective:       Returns mixed file systems template for a specific OS version
################################################################################
sub get_mixed_fs_template {
    my ($self) = @_;

    $self->_clear_err();

    my @fs_array;

    if ($self->{major_revision} =~ /rhel6|rhel7|fedora|ol6/i) {
        push (@fs_array, "ext4");
    }

    if (($self->{major_revision} =~ /rhel7|fedora/i) or
       ($self->{major_revision} =~ /rhel6|ol6/i and $self->{minor_revision} >= 4) or
       ($self->{major_revision} =~ /sles11/i and $self->{minor_revision} >= 2)) {
        push (@fs_array, "xfs");
    }

    if ($self->{major_revision} !~ /rhel7|rhel6|fedora|ol6/i) {
        push (@fs_array, "ext3");
    }

    if ($self->{major_revision} =~ /sles11/i and $self->{minor_revision} >= 2) {
        push (@fs_array, "btrfs");
    }

    return (wantarray()) ? @fs_array : \@fs_array;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _MP_configure()
# Objective:       Configure the multipath
# Details:         See POD documentation at the end of this file
################################################################################
sub _MP_configure() {
    my ($self) = @_;

    if ($self->{skip_mp_config} == 1) {
        # Ignore _MP_configure, we are asked to NOT configure anything
        $self->sanlog(-msg => "Bypassing multipath Configuration by user request (CLIENT_SKIP_MP_CONFIG=1)");
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "Configuring MultiPathing");

    # verify that all mp rpms are installed
    $self->{mp}->verify(-conn => $self->{conn});
    if ($self->{mp}->error) {
        my @mp_error = $self->{mp}->errmsg();
        $self->_set_err(-msg => "@mp_error");
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "MultiPathing Verified");

    # No need to configure anything for Veritas DMP
    if ($self->{mp}->{mp_type} !~ /dmp/) {
        $self->{mp}->config(-conn => $self->{conn}, -protocol => $self->{protocol});
        if ($self->{mp}->error) {
            my @mp_error = $self->{mp}->errmsg();
            $self->_set_err(-msg => "@mp_error");
            return $ClientUtil::FALSE;
        }
        $self->sanlog(-msg => "/etc/multipath.conf configured");
    }

    # load multipath modules and start the service
    $self->{mp}->enable(-conn => $self->{conn});
    if ($self->{mp}->error) {
        my @mp_error = $self->{mp}->errmsg();
        $self->_set_err(-msg => "@mp_error");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "MultiPathing enabled and configured");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_raw_binding()
# Objective:       create raw device binding
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_raw_binding {
    my ($self, $block_device) = @_;

    my ($cmd, $found, $hsh, $major, $minor, $raw_device, $status, @output, $hosttype);

    # Defaults
    $found = 0;
    $hsh   = $self->{raw_bindings};
    ($major, $minor) = split(":", $block_device);

    $hosttype = $self->hosttype();

    #  Check if already bound
    foreach my $key (keys %$hsh) {
        if (    $hsh->{$key}{major} == $major
            and $hsh->{$key}{minor} == $minor) {
            $raw_device = $key;
            $found      = 1;
            last;
        }
    }

    #  If not bound get next availabe device
    if (not $found) {
        foreach my $key (keys %$hsh) {
            next if ($hsh->{$key}{major} != 0);
            $hsh->{$key}{major} = $major;
            $hsh->{$key}{minor} = $minor;
            $raw_device         = $key;
            last;
        }

        $cmd = "raw $raw_device $major $minor";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxHost::_create_raw_binding() Command ($cmd) returned a bad status ($status)");
            return "";
        }
    }

    return $raw_device;
}

################################################################################
# Subroutine Name: _delete_raw_binding()
# Objective:       remove raw device binding
# Details:         See POD documentation at the end of this file
################################################################################
sub _delete_raw_binding {
    my ($self, $key) = @_;

    my ($hsh);

    $hsh = $self->{raw_bindings};

    $hsh->{$key}{major} = 0;
    $hsh->{$key}{minor} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices() {
    my ($self) = @_;

    my ($cmd, $status, @buffer, @hDevices, @output, @lvm_output, @realdevs, $hosttype);

    # We use user_friendly_name = no inside /etc/multipath.conf. So we grep all
    # the devices using id inside /dev/mapper. device id starts with 360a9 (7-mode)
    # or 3600a08 (c-mode) in case of NetApp LUN. We dont even have to check if lvm
    # is created over these devices or not because these devices shouldn't have
    # lvm over them. We should remove lvm (lv, vg and pv) when cleaning the host.
    $hosttype = $self->hosttype();
    if (defined $self->{mp} and $self->{mp_type} !~ /""|0|none/i) {
        @hDevices = $self->{mp}->get_mp_devices(-conn => $self->{conn});
        if ($self->{mp}->error) {
            $self->_set_err($self->{mp}->errmsg());
            return (wantarray()) ? () : [] ;
        }

        if (scalar(@hDevices) > 0) {
            $self->sanlog(-msg => "Got devices successfully");
        } else {
            return (wantarray()) ? () : [] ;
        }
    } else {
        # use sanlun lun show to get a list of devices
        my $path = $self->get_hu_bin_path();
        $cmd = $path . "/sanlun lun show all";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 and $status != 203) {
            $self->_set_err("Error: LinuxHost::_get_devices() Unable to execute $cmd");
            return (wantarray()) ? () : [] ;
        }

        # use lvmdiskscan to get information that'll allow us to tell if a lun
        # is in an LV
        $cmd = "lvmdiskscan";
        ($status, @lvm_output) = &ClientUtil::execute($self->{conn}, $cmd);

        # grep only sd.* devices
        @lvm_output = grep (/\/dev\/sd.*/, @lvm_output);

        # Suse will give false errors when it does not have any lvm devices
        # and so we will always skip this check for Suse
        if (($status > 0) && ($hosttype =~ /redhat/)) {
            $self->_set_err("Error: LinuxHost::_get_devices() Command ($cmd) " .
                            "returned a bad status ($status)");
            return (wantarray()) ? () : [] ;
        }

        # old sanlun kit output looks like this:
        #   filer: lun-pathname     device filename   adapter    protocol  lun size        lun state
        #
        # myfiler: /vol/blocks/lun1 /dev/sdb          <unknown>0 iSCSI     5g (5368709120) GOOD
        #
        # new sanlun kit output looks like below,
        # controller(7mode)/                                                 device          host                  lun
        # vserver(Cmode)       lun-pathname                                  filename        adapter    protocol   size    mode
        # ---------------------------------------------------------------------------------------------------------------------
        # fas3170-201-76       /vol/nate_1_1_1/nate_IBMblade-200-218_linux_0 /dev/sde        host22     iSCSI      4g      7

        foreach my $line (@output) {

            # hDevs is a hash to hold each device.  Will be pushed
            # onto hDevices if it's good
            my $hDevs = {};

            $line =~ s/^\s+//;    # strip leading white space
            my ($filer, $lun_path, $device, $adapter, $protocol,
                $lun_size) = split(/\s+/, $line);
            $lun_size =~ s/g/ GB/;
            $lun_size =~ s/m/ MB/;
            $filer=~ s/\://;

            my $short_dev = $device;
            $short_dev =~ s/\/dev\///;
            $self->sanlog(-msg       => "Sanlun lun show Line is $line",
                          -msg_level => 7);
            $self->sanlog(-msg => "Filer    is $filer\n"
                            . "Lun      is $lun_path\n"
                            . "Device   is $device\n"
                            . "Short D  is $short_dev\n"
                            . "Adapter  is $adapter\n"
                            . "Protocol is $protocol\n"
                            . "Size     is $lun_size\n"
                            . "Hosttype is $hosttype",
                          -indent    => 4,
                          -msg_level => 7);

            # Check to make sure we have an iscsi disk and make sure the
            # lun has not ben orphaned
            if ($lun_path !~ /^\/vol\//) {
                next;
            }
            $hDevs->{path}  = "/dev";
            $hDevs->{name}  = $short_dev;
            $hDevs->{type}  = "NETAPP";
            $hDevs->{size}  = $lun_size;
            $hDevs->{id}    = $short_dev;
            $hDevs->{label} = "free";
            $hDevs->{filer_name} = $filer;

            # figure out if the lun is in an lvm we need this
            # for the label field in the device hash.
            foreach my $lvm_line (@lvm_output) {
                $self->sanlog(-msg       => "lvmdiskscan Line is $lvm_line",
                              -msg_level => 7);
                next unless $lvm_line =~ /$device/;
                $self->sanlog(-msg       => "Found device $device in LVM line",
                              -msg_level => 7);

                # format is:
                # lvmdiskscan -- /dev/sdb  [       9.77 GB] free whole disk
                # lvmdiskscan -- /dev/sdc  [       9.77 GB] LVM physical volume
                if ($lvm_line =~ / LVM /) {
                    $hDevs->{label} = "USED";
                }
                last;    # skip to the end of lvm_output
            }

            # return this device
            $self->sanlog(-msg       => "Pushing device $line\n",
                          -msg_level => 3);
            push @hDevices, $hDevs;

        }
    }

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    my (@hDevices, @devices);

    #  Retrieve all host devices
    @hDevices = $self->_get_devices();
    if ($self->error) {
        return (wantarray()) ? () : [] ;
    }

    foreach my $device (@hDevices) {
        push(@devices, $device->{name}) if ($device->{label} =~ /free|new/i);
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: _get_release
# Objective:       Get vendor release
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_release {
    my ($self) = @_;

    my ($cmd, $release, $status, @output, @hDevices);

    $cmd = "cat /etc/redhat-release";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_get_release() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    $release = $output[0];

    return $release;
}

################################################################################
# Subroutine Name: _load_raw_bindings()
# Objective:       load raw device bindings
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_raw_bindings {
    my ($self, $hsh) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    #  Create a hash of raw devices
    foreach (1 .. 255) {
        my $key = "/dev/raw/raw" . $_;
        $hsh->{$key}{major} = 0;
        $hsh->{$key}{minor} = 0;
    }

    # Load current raw device bindings.
    $cmd = "raw -qa";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::load_raw_bindings() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        # Example: /dev/raw/raw1:  bound to major 58, minor 0
        my ($key, $major, $minor) = $line =~ /(\S+):\s+.*major\s+(\d+),\s+minor\s+(\d+)/;
        $hsh->{$key}{major} = $major;
        $hsh->{$key}{minor} = $minor;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_protocol_initiator_loaded()
# Objective:       Returns TRUE if $protocol initiator is loaded on this host.
#                  Returns FALSE if $protocol initiator is not loaded on
#                  this host.
# Details:         See POD documentation at the end of this file
# June 8th 2007
################################################################################
sub _is_protocol_initiator_loaded {
    my ($self, $protocol) = @_;

    $self->_clear_err();

    # check if protocol is defined
    unless (defined($protocol)) {
        return $ClientUtil::FALSE;
    }

    # all we have to do is call the display_adapters function on
    # the sanlun object. If an error was not set by that function, then
    # the $protocol driver/module is loaded, otherwise it is not loaded.
    $self->{sanlun}->display_adapters($protocol);

    if ($self->error == 0) {
        return $ClientUtil::TRUE;
    } else {
        $self->_clear_err();
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: _load_netconsole_module()
# Objective:       Returns TRUE if netconsole module is loaded on this host.
#                  Returns FALSE if an error occurred.
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_netconsole_module {
    my ($self, %args) = @_;

    my ($netconsole_driver_module, $cmd, $status, @output);
    my ($nate_server_ip, $nate_server_mac);

    foreach (keys %args) {
        if (/-?nate_server_ip/) {
            if (not length($nate_server_ip = $args{$_})) {
                $self->_set_err("Invalid value for -nate_server_ip: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nate_server_mac/) {
            if (not length($nate_server_mac = $args{$_})) {
                $self->_set_err("Invalid value for -nate_server_mac: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Make sure module is unloaded before loading it. Module needs to be unload because
    # arguments passed to module depends on the NATE Control Server on which the
    # test is running.
    if ($self->_is_module_loaded(-module => 'netconsole') == $ClientUtil::TRUE) {
        $self->_unload_module(-module => 'netconsole');
        if ($self->error) {
             $self->_set_err("Unable to unload netconsole driver module \'netconsole\'");
             return $ClientUtil::FALSE;
        }
    }

    my $client_ip = $self->{conn}{hostp}{default_ip};
    my $client_ifac = $self->{conn}{hostp}{IFC};
    my $nc_port = $self->{nc_port};

    # Here is the example of command that will run on host
    #---------------------------------------------------------------------------------
    # modprobe netconsole netconsole="514@10.72.200.192/eth0,514@10.72.200.194/00:14:5E:5B:7D:17"
    #---------------------------------------------------------------------------------
    my $arguments = "514\@$client_ip\/$client_ifac,$nc_port\@$nate_server_ip\/$nate_server_mac";
    $cmd = "modprobe netconsole netconsole=\"$arguments\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if($status > 0) {
        $self->_set_err("Unable to load driver module \'netconsole\'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _load_fcp_module()
# Objective:       Returns TRUE if fcp module is loaded on this host.
#                  Returns FALSE if an error occurred.
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_fcp_module {
    my ($self) = @_;

    my ($fcp_driver_module, $cmd, $status, @output);

    $self->_clear_err();

    $fcp_driver_module = $self->{fcp_driver_module_type};

    $cmd = "modprobe -v $fcp_driver_module";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if($status > 0) {
        $self->_set_err("Error: unable to load fcp driver module $fcp_driver_module");
        return $ClientUtil::FALSE;
    }

    # Load lpfcdfc driver in case of emulex for "sanlun fcp show adapter -v"
    # to work
    if ($fcp_driver_module =~ /lpfc/i) {
        $cmd = "modprobe -v lpfcdfc";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # Fail only if driver module is present and unable to load
        if ($status > 0 and (!grep(/module\s+lpfcdfc\s+not\s+found/i, @output))) {
            $self->_set_err("Unable to load lpfcdfc driver module");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _unload_fcp_module()
# Objective:       Returns TRUE if fcp module was unloaded.
#                  Returns FALSE if an error occurred
#                  NOTE: unloading the fcp module will disable the Multipathing
#                  software. Before calling this method, you must unmount any
#                  multipathed devices.
# Details:         See POD documentation at the end of this file
################################################################################
sub _unload_fcp_module {
    my ($self) = @_;

    my ($fcp_driver_module, $cmd, $status, @output);

    $self->_clear_err();

    $fcp_driver_module = $self->{fcp_driver_module_type};

    # Unload of qla2xxx will fail if "qlremote" is running
    # So killing qlremote before unloading qlogic driver
    if ($fcp_driver_module =~ /qlogic/) {
        $cmd = "/usr/bin/pkill qlremote";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # No need to check the status because even if it return 1 it will get killed
    }

    if ($fcp_driver_module =~ /lpfc/i) {
        # Stop hbaanyware if application kit is installed
        $cmd = "/usr/sbin/hbanyware/stop_hbanyware";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to stop hbaanyware");
            return $ClientUtil::FALSE;
        }
        # unload this extra driver in case of emulex
        $cmd = "modprobe -rv lpfcdfc";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # Fail only if driver module is present and unable to unload
        if ($status > 0 and (!grep(/module\s+lpfcdfc\s+not\s+found/i, @output))) {
            $self->_set_err("Unable to unload fc driver module lpfcdfc");
            return $ClientUtil::FALSE;
        }
    }

    # Unload fc driver
    $cmd = "modprobe -rv $fcp_driver_module";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to unload fc driver module $fcp_driver_module");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _unload_module()
# Objective:       Unload driver module
# Details:         See POD documentation at the end of this file
################################################################################
sub _unload_module {
    my ($self, %args) = @_;

    my (@output, $cmd, $module, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?module/) {
            if (not length($module = $args{$_})) {
                $self->_set_err("Invalid value for argument '-module' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Unload the driver module.
    $cmd = "modprobe -r $module";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to unload module $module");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_module_loaded()
# Objective:       check if a given module is loaded on this host
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_module_loaded {
    my ($self, %args) = @_;

    my (@output, $cmd, $module, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?module/) {
            if (not length($module = $args{$_})) {
                $self->_set_err(
                        "Error: LinuxHost::_is_module_loaded() Invalid" . " value for argument '-module' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # run an lsmod command.
    $cmd = 'lsmod';
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # if the module is running, return true. Otherwise, return false.
    if (grep /^$module/, @output) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: _iscsi_verify()
# Objective:       Verify that the correct iscsi initiator tools are installed
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_verify {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    # For SUSE, check for linux-iscsi
    if ($self->hosttype() =~ /unix-linux-suse/i or $self->hosttype() =~ /unix-linux-sles/i) {
        # SLES 10 and above has different iscsi rpms from SLES 9.
        # make sure to check for the correct rpms
        if ($self->{major_revision} =~ /sles9/) {
            $cmd = "rpm -q linux-iscsi";
        } else {
            $cmd = "rpm -q open-iscsi";
        }
    }

    # For RHEL4, check if iscsi-initiator-utils is installed
    #second condition is to check for OL. testbed_setup is not able to determine the hosttype correctly.
    #See the BURT[]
    elsif ($self->hosttype() =~ /unix-linux-redhat/i or $self->hosttype() =~ /unix-linux/i) {
        $cmd = "rpm -q iscsi-initiator-utils";
    } else {

        # Host type is not matching anything we expect
        $self->_set_err("Error: LinuxHost::_iscsi_verify Bad hosttype specified: |" . $self->hosttype() . "|");
        return $ClientUtil::FALSE;
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::_iscsi_verify()" . " iscsi-initiator-utils is not installed");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsi discovery targets
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    my ($status, @output, @output_tmp, $cmd, @targets);

    my $iscsi = "iscsi";

    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} != 1) {
        # Ignore _iscsi_configure, we are asked to NOT configure anything
        $self->sanlog(-msg => "iSCSI Configure: Bypassing iSCSI Configuration by user request (CONFIG_ISCSI_DISCOVERY=0)");
        return $ClientUtil::TRUE;
    }

    # check to make sure configuration is for software initiator, since hardware is currently
    # not supported.
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /software/i) {
        $self->_set_err("iSCSI Configure: iSCSI Configuration is currently only supported with the iSCSI SW initiator,\n" .
                        "iSCSI hardware adapters are currently not supported");
        return $ClientUtil::FALSE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI");

    # validate the iscsi parameters before continuing.
    $self->_iscsi_configure_validate_params();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # verify proper environment for iscsi
    $self->_iscsi_verify();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # need to flush the multipathing layer before configuring iscsi
    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        $self->{mp}->flush(-conn => $self->{conn});
        if ($self->{mp}->error) {
            my @mp_error = $self->{mp}->errmsg();
            $self->_set_err(-msg => "@mp_error");
            return $ClientUtil::FALSE;
        }
    }

    # if requested to do subnet matching, filter the targets list
    # so that it contains targets that are accessible by this host
    @targets = $self->_filter_targets_by_subnet(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check what ip version is being requested and set the ip mode on all the target objects
    if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /^ipv4$|^ipv6$|^auto$/i) {
        foreach my $target (@targets) {
            # set the ip mode to be either ipv4 or ipv6
            $target->set_ip_mode($self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
            if ($target->error()) {
                $self->_set_err("Could not set the ip mode of target $target->{nodename} to " .
                                 $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} .
                                "Reason: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    } else {
       $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: " .
                        $self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
       return $ClientUtil::FALSE
    }

    # print out a summary of what we are going to do for iSCSI Configure
    $self->_iscsi_configure_print_summary(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # if requested to do iscsi pre-cleanup, do so now
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} > 0) {
        $self->_iscsi_deconfigure(-targets => \@targets,
                                  -clean_level => $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # branch out (open-iscsi/linux-iscsi)
    if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
        # configure open-iscsi (iscsiadm)
        $self->_iscsi_configure_openiscsi(-targets => \@targets);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        # configure linux iscsi (iscsi.conf)
        $self->_iscsi_configure_linuxiscsi(-targets => \@targets);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "iSCSI Configure: iSCSI has been configured.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_validate_params()
# Objective:       Validate that the iscsi params are set properly.
################################################################################
sub _iscsi_configure_validate_params {
    my ($self) = @_;

    my ($cmd);

    # check for unsupported options
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} !~ /^(bidirectional|unidirectional|none|\s*)$/) {
        $self->_set_err("Unsupported CHAP Mode: " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE});
        return $ClientUtil::FALSE;
    }

    # only sendtargets, isns, and static are supported in the automation
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} !~ /^sendtargets|isns|static$/i) {
        $self->_set_err("Unsupported CONFIG_ISCSI_DISCOVERY_TYPE: " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} .
                        ", valid values are (sendtargets|isns|static).");
        return $ClientUtil::FALSE;
    }

    # make sure the user is not doing isns with a sanbooted system on open-iscsi
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i and
        $self->{sanbooted} == 1 and
        $self->{iscsi}->{iscsi_version} =~ /open\-iscsi/i) {
        $self->_set_err("Cannot configure iSNS discovery on a sanbooted " .
                        "system with the Open-iSCSI initiator");
        return $ClientUtil::FALSE;
    }

    # an isns server is required if doing isns discovery
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i and
        $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} eq "") {
        $self->_set_err("CONFIG_ISCSI_ISNS_SERVER is required if doing iSNS Discovery.");
        return $ClientUtil::FALSE;
    }

    # Data digest is not supported in redhat's and oracle linux's open-iscsi
    if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1 and
        $self->{major_revision} =~ /rhel|ol|fedora/i and
        $self->{iscsi}->{iscsi_version} =~ /open\-iscsi/i) {
        $self->_set_err("Data Digest is not supported in Open-iSCSI for Redhat and oracle linux");
        return $ClientUtil::FALSE;
    }

    # obtain the iscsi initiator nodename
    if ($self->{major_revision} =~ /rhel4/) {
        $cmd = "cat /etc/initiatorname.iscsi  | grep InitiatorName=";
    } else {
        $cmd = "cat /etc/iscsi/initiatorname.iscsi  | grep InitiatorName=";
    }
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not determine iSCSI initiator name");
        return $ClientUtil::FALSE;
    }
    my $initiator_nodename;
    foreach my $line (@output) {
        if ($line =~ /InitiatorName=(\S+)/i) {
            $initiator_nodename = $1;
        }
    }

    # default the in/out usernames to be the initiator name if its empty
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} eq "") {
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} = $initiator_nodename;
    }
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} eq "") {
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} = $initiator_nodename;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_print_summary()
# Objective:       print out a summary of what iSCSI configuration will be done
################################################################################
sub _iscsi_configure_print_summary {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my @target_ips = map { "\n\t" . $_->{ip} } @targets;
    my @config_summary;

    push(@config_summary, "iSCSI Configure: Configuring iSCSI with the following options:\n");
    push(@config_summary, "Targets : @target_ips\n");
    push(@config_summary, "Linux iSCSI Initiator : " . $self->{iscsi}->{iscsi_version} . "\n");
    push(@config_summary, "IP Version            : " . $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n");
    push(@config_summary, "iSCSI Type            : " . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "\n");
    push(@config_summary, "iSCSI Discovery Type  : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "\n");
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
        push(@config_summary, "iSNS Server           : " . $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} . "\n");
    }
    push(@config_summary, "iSCSI Pre-Clean Level : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} . "\n");
    push(@config_summary, "iSCSI Subnet Matching : " . $self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} . "\n");

    push(@config_summary, "CHAP Mode             : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} . "\n");
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
        push(@config_summary, "CHAP Username         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} . "\n");
        push(@config_summary, "CHAP Password         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD} . "\n");
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
            push(@config_summary, "CHAP Out Username     : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} . "\n");
            push(@config_summary, "CHAP Mutual Secret    : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} . "\n");
        }
    }

    push(@config_summary, "Header Digest         : " . $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} . "\n");
    push(@config_summary, "Data Digest           : " . $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} . "\n");

    $self->sanlog(-msg => "@config_summary");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_openiscsi()
# Objective:       Configure iscsi discovery for 'open-iscsi'
################################################################################
sub _iscsi_configure_openiscsi {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, $status, $startup_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Missing required param -targets");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring '$self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE}' " .
                          "discovery for open-iscsi");

    if (defined $self->{mp} and $self->{mp_type} =~ /dmmp/i) {
        # modify session replacement timeout in the /etc/iscsi/iscsid.conf file
        $self->_change_parameter_value_in_file("node.session.timeo.replacement_timeout", "5", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # configure sendtargets
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /sendtargets/i) {
        $self->_iscsi_configure_openiscsi_sendtargets(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /static/i) {
        # configure static
        $self->_iscsi_configure_openiscsi_static(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /isns/i) {
        # configure isns
        $self->_iscsi_configure_openiscsi_isns();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        # unsupported discovery type
        $self->_set_err("Unsupported CONFIG_ISCSI_DISCOVERY_TYPE '$self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE}'");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring open-iscsi options");

    # configure iscsi options for all the newly added targets
    $self->_iscsi_configure_openiscsi_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Logging in to all targets");

    # issue the logins on all the targets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $tpgt = $target->get_tpgt();
        if ($target->error()) {
            $self->_set_err("Could not obtain tpgt from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        # enable chap for this target if requested
        $cmd = "/sbin/iscsiadm -m node --login -T $nodename -p ${ip}:3260,$tpgt";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # dont fail if the login session to this target already exists, and  no precleanup was performed
        if ($status != 0) {
            # if we didnt do any precleanup and we see "already exists" then dont fail, otherwise fail
            if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} == 0 and grep(/already exists/i, @output)) {
                # dont fail
                $self->sanlog(-msg => "iSCSI Configure: ${nodename} ${ip}:3260,${tpgt} is already logged in.");
            } else {
                # in this case the login failed
                $self->_set_err("Failed to login to $target->{nodename}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # list out all the sessions
    $cmd = "/sbin/iscsiadm -m session";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to list the iSCSI sessions");
        return $ClientUtil::FALSE;
    }

    # Exclude the sessions created by Hardware iscsi (Qlogic cards)
    @output = grep (!/^qla4/, @output);

    # modify the node.startup flag for all sessions of the targets we added to be automatic
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $tpgt = $target->get_tpgt();
        if ($target->error()) {
            $self->_set_err("Could not obtain tpgt from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        # if there is no session for this target, then skip.
        if (not grep(/\s*${ip}\:3260,${tpgt}\s+${nodename}/i, @output)) {
            next;
        }

        # modify startup setting of the sessions
        if (($self->{sanbooted} == 1) and ($self->{major_revision} =~ /sles/i)) {
            $startup_mode = "onboot";
        } else {
            $startup_mode = "automatic";
        }

        $cmd = "/sbin/iscsiadm -m node -T $nodename -p $ip\:3260 -o update -n node.startup -v $startup_mode";
        ($status) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to make the session's startup setting for target " .
                            "'$nodename' at address '$ip'");
            return $ClientUtil::FALSE;
        }

        # Set startup option of the default connection of the initiator node to "automatic/onboot".
        # This will ensure that login of the session happens during iscsi start.
        $cmd = "/sbin/iscsiadm -m node -T $nodename -p $ip\:3260 -o update -n node.conn[0].startup -v $startup_mode";
        ($status) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to make the session's startup setting for target " .
                            "'$nodename' at address '$ip'");
            return $ClientUtil::FALSE;
        }
    }

    # need to recreate the initrd image for sanbooted systems of SLES.
    # this step is not required for RHEL and OL.
    if (($self->{sanbooted} == 1) and ($self->{major_revision} =~ /sles/i)) {
        # Before modifying the initrd image take a backup
        my $kernel_version =  $self->{revision};
        $cmd = "cp -f /boot/initrd-" . $kernel_version . " /boot/initrd-" . $kernel_version . ".bak";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to backup initrd image");
            return $ClientUtil::FALSE;
        }

        # Recreating the initrd image
        if ($self->{major_revision} =~ /sles10/i) {
            $cmd = "mkinitrd -f \"mpath iscsi\"";
        } elsif ($self->{major_revision} =~ /sles11/i) {
            $cmd = "mkinitrd -f \"multipath iscsi\"";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to create initrd image");
            return $ClientUtil::FALSE;
        }
    }

    use NATE::Log qw(log_global);
    $Log = log_global();

    if ($Log->may_debug()) {
        # list out all the sessions with verbosity for debugging purpose
        $cmd = "/sbin/iscsiadm -m session -P 3";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to list the iSCSI sessions");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_openiscsi_options()
# Objective:       Configure iscsi discovery options for 'open-iscsi'
################################################################################
sub _iscsi_configure_openiscsi_options {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Missing required param -targets");
        return $ClientUtil::FALSE;
    }

    # enable CHAP/HD/DD on all targets passed in if requested
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $tpgt = $target->get_tpgt();
        if ($target->error()) {
            $self->_set_err("Could not obtain tpgt from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        # enable chap for this target if requested
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
            my $out_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME};
            my $in_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
            my $out_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD};
            my $in_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD};

            # enable chap
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.auth.authmethod -v CHAP";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # set up the in username
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.auth.username -v $in_username";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # set up the in password
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.auth.password -v $in_password";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # if doing bidirectional, set up the out username & password (open iscsi refers as the 'in' credentials)
            if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
                $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.auth.username_in -v $out_username";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }

                $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.auth.password_in -v $out_password";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # enable header digest if requested
        if ($self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} == 1) {
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.conn[0].iscsi.HeaderDigest -v CRC32C";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # enable data digest if requested (not supported in redhat and oracle linux)
        if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1 and $self->{major_revision} !~ /rhel|ol|fedora/i) {
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.conn[0].iscsi.DataDigest -v CRC32C";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # if doing dmmp we need to modify the default setting of replacement_timeout from 120 to 5, all other
        # targets default to 120.
        my $replacement_timeout;
        if (defined $self->{mp} and $self->{mp_type} =~ /dmmp/i) {
            $replacement_timeout = 5;
        } else {
            $replacement_timeout = 120;
        }
        $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260,${tpgt} -o update -n node.session.timeo.replacement_timeout -v $replacement_timeout";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_openiscsi_sendtargets()
# Objective:       Configure Sendtargets iscsi discovery for 'open-iscsi'
################################################################################
sub _iscsi_configure_openiscsi_sendtargets {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, @list_output, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Missing required param -targets");
        return $ClientUtil::FALSE;
    }

    # first configure the discovery credentials if requested to do CHAP
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {

        my $in_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
        my $in_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD};
        my $out_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME};
        my $out_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD};

        # modify the following lines in the /etc/iscsid.conf file
        $self->_change_parameter_value_in_file("discovery.sendtargets.auth.authmethod", "CHAP", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_change_parameter_value_in_file("discovery.sendtargets.auth.username", "$in_username", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_change_parameter_value_in_file("discovery.sendtargets.auth.password", "$in_password", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # if doing bi directional, update the out credentials
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
            $self->_change_parameter_value_in_file("discovery.sendtargets.auth.username_in", "$out_username", "/etc/iscsi/iscsid.conf");
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
            $self->_change_parameter_value_in_file("discovery.sendtargets.auth.password_in", "$out_password", "/etc/iscsi/iscsid.conf");
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        }
    }

    # list out all the sendtargets currently on the system
    # this command will return a nonzero status when there are no targets in the list.
    # dont fail in this case.
    $cmd = "/sbin/iscsiadm -m discovery";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 and scalar(@list_output) > 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    # add each target with sendtargets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        #--> iscsiadm -m discovery
        #192.168.1.61:3260 via sendtargets
        #[fe80::a0:98ff:fe01:70aa]:3260 via sendtargets

        $self->sanlog(-msg => "iSCSI Configure: Adding SendTarget '$ip'");

        # add the target. It's ok if it is already addded, this will
        # force iscsiadm to rediscover the targets if they were removed at any point (burt494188)
        $cmd = "/sbin/iscsiadm -m discovery -t sendtargets -p ${ip}:3260";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_openiscsi_static()
# Objective:       Configure Static iscsi discovery for 'open-iscsi'
################################################################################
sub _iscsi_configure_openiscsi_static {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, @list_output, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Missing required param -targets");
        return $ClientUtil::FALSE;
    }

    # list out all the static targets currently on the system
    $cmd = "/sbin/iscsiadm -m node";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 and not grep(/no records found/i, @list_output)) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    # add each target manually
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        my $tpgt = $target->get_tpgt();
        if ($target->error()) {
            $self->_set_err("Could not obtain tpgt from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        #--> iscsiadm -m node
        #192.168.1.61:3260,2000 iqn.1992-08.com.netapp:sn.84174947
        #[fe80::a0:98ff:fe01:70aa]:3260,2000 iqn.1992-08.com.netapp:sn.84174947

        # skip if this target is already added.
        if (grep(/^\s*${ip}:3260,${tpgt}\s+${nodename}/i, @list_output)) {
            $self->sanlog(-msg => "iSCSI Configure: Static target '$nodename,$ip,$tpgt' is already added");
            next;
        } else {
            $self->sanlog(-msg => "iSCSI Configure: Adding Static target '$nodename,$ip,$tpgt'");
        }

        # add the target as a static target
        $cmd = "/sbin/iscsiadm -m node -o new -T $nodename -p ${ip}:3260,${tpgt}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_openiscsi_isns()
# Objective:       Configure iSNS iscsi discovery for for 'open-iscsi'
################################################################################
sub _iscsi_configure_openiscsi_isns {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    # first add the isns server and isns port to /etc/iscsid.conf
    my $isns_server = $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER};
    $self->sanlog(-msg => "iSCSI Configure: Adding iSNS Server: '$isns_server'");
    $self->_change_parameter_value_in_file("isns.address", "$isns_server", "/etc/iscsi/iscsid.conf");
    if ($self->error) {
        return $ClientUtil::FALSE;
    }
    my $isns_port = 3205;
    $self->_change_parameter_value_in_file("isns.port", "$isns_port", "/etc/iscsi/iscsid.conf");
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # need to restart open iscsi in order for the isns server to get updated
    $self->_iscsi_service(-action => "restart");
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # then trigger the update from the isns server
    $cmd = "/sbin/iscsiadm -m discovery -t isns";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {

        # If the error is please specify portal, try the command again
        # but with the newer options for RHEL 5u6 and RHEL 6
        if ( grep(/Please specify portal/i, @output) ){
            $cmd = "/sbin/iscsiadm -m discovery -t isns -p $isns_server:$isns_port";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        }
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_linuxiscsi()
# Objective:       Configure iscsi discovery for for 'linux-iscsi'
################################################################################
sub _iscsi_configure_linuxiscsi {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, @list_output, $status);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Missing required param -targets");
        return $ClientUtil::FALSE;
    }

    # prepend DiscoveryAddress= in front of every ip address in targets
    # Add the required parameters to iscsi.conf
    my @iscsi_conf;
    my @iscsi_conf_extra_options;

    # Push options specific to rhel4 only
    if ($self->{major_revision} =~ /rhel4/i) {
        push (@iscsi_conf_extra_options, 'ConnFailTimeout=5');
    }
    # Push options specific to sles9 only in to iscsi.conf array
    elsif ($self->{major_revision} =~ /sles9/i) {
        push (@iscsi_conf_extra_options, 'HeaderDigest=never', 'DataDigest=never', 'Continous=no');

        # Push the following options only if mp is enabled and mp_type is dmmp
        if (defined $self->{mp} and $self->{mp_type} =~ /dmmp/i) {
            push (@iscsi_conf_extra_options, 'Multipath=portal', 'ConnFailTimeout=5');
        }
    }

    foreach my $target (@targets) {
        $target = "DiscoveryAddress=" .  $target->{ip};
        push (@iscsi_conf, $target);
    }
    push (@iscsi_conf,@iscsi_conf_extra_options);

    # These will be our sentinels for our modifications to the iscsi.conf file
    my $begin_sentinel = "#_#_ START iscsi.conf AUTOMATION ##";
    my $end_sentinel   = "#_#_ END iscsi.conf AUTOMATION ##";
    unshift(@iscsi_conf, $begin_sentinel);
    push(@iscsi_conf, $end_sentinel);

    $cmd = "rm -f /etc/iscsi.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::_iscsi_configure() unable to remove /etc/iscsi.conf.tmp");
        return $ClientUtil::FALSE;
    }

    $cmd = "touch /etc/iscsi.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::_iscsi_configure() unable to create empty /etc/iscsi.conf.tmp");
        return $ClientUtil::FALSE;
    }

    # now back up the previous iscsi.conf file in this format,
    # iscsi.conf.bak.<timestamp>
    $cmd = "cp -f /etc/iscsi.conf /etc/iscsi.conf." . time();
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::_iscsi_configure() unable to take the backup of /etc/iscsi.conf");
        return $ClientUtil::FALSE;
    }

    chomp(@iscsi_conf);

    # now write all the lines of elements into the iscsi.conf file on the
    # host.
    foreach my $line (@iscsi_conf) {
        $cmd = "echo \'$line\' >> /etc/iscsi.conf.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: LinuxHost::_iscsi_configure() cmd ($cmd) " . "returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    $cmd = "mv -f /etc/iscsi.conf.tmp /etc/iscsi.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: LinuxHost::_iscsi_configure() cmd ($cmd) " . "returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    $self->_iscsi_service(-action => 'restart');
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI settings: " . $self->{iscsi}->{iscsi_version});

    # branch out (open-iscsi/linux-iscsi)
    if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
        $self->_iscsi_deconfigure_openiscsi(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_iscsi_deconfigure_linuxiscsi(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: iSCSI has been deconfigured");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_openiscsi()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_openiscsi {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # logout from all the targets if clean_level == 2, if we get:
    # No portal found, that is OK, nothing was configured prior
    if ($clean_level == 2) {
        $cmd = "/sbin/iscsiadm -m node -u";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0 and not grep(/No portal found|No session found|No matching sessions found/i, @output)) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    } else {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # otherwise if clean_level == 1, logout from the targets listed in @targets
        foreach my $target (@targets) {
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            my $nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the target ip is ipv6, then add brackets around it.
            if ($ip =~ /:/) {
                $ip = "[" . $ip . "]";
            }

            # logout from this target (this command will not fail if there is no session to the target)
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -u";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0 and not grep(/No portal found|No session found|No matching sessions found/i, @output)) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    # deconfigure iscsi options for open-iscsi
    $self->_iscsi_deconfigure_openiscsi_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure sendtargets
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /sendtargets/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_openiscsi_sendtargets(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # deconfigure static discovery
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /static/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_openiscsi_static(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # deconfigure isns
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /isns/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_openiscsi_isns(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_openiscsi_options()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_openiscsi_options {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @list_output);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # first check if there are any target records, if there are none, then return immediately
    $cmd = "/sbin/iscsiadm -m node";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (grep(/no records found/i, @list_output)) {
        $self->sanlog(-msg => "iSCSI Deconfigure: Skipping openiscsi options deconfigure (no targets to deconfigure)");
        return $ClientUtil::TRUE;
    }
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    # Disable CHAP/HD/DD for all targets if clean_level is 2
    if ($clean_level == 2) {

        # disable CHAP for everything
        $cmd = "/sbin/iscsiadm -m node -o update -n node.session.auth.authmethod -v None";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # reset the CHAP in-credentials
        $cmd = "/sbin/iscsiadm -m node -o update -n node.session.auth.username_in -v \"\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
        $cmd = "/sbin/iscsiadm -m node -o update -n node.session.auth.password_in -v \"\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # reset CHAP out-credentials
        $cmd = "/sbin/iscsiadm -m node -o update -n node.session.auth.username -v \"\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
        $cmd = "/sbin/iscsiadm -m node -o update -n node.session.auth.password -v \"\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # disable header/data digest
        $cmd = "/sbin/iscsiadm -m node -o update -n node.conn[0].iscsi.HeaderDigest -v None";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
        if ($self->{major_revision} !~ /rhel|ol|fedora/i) {
            # Data Digest is not supported in redhat's and oracle linux's open-iscsi, this command will fail if we try it
            $cmd = "/sbin/iscsiadm -m node -o update -n node.conn[0].iscsi.DataDigest -v None";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }

    } else {
        # for clean_level 1, disable CHAP/HD/DD only for @targets
        foreach my $target (@targets) {
            my $nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the target ip is ipv6, then add brackets around it.
            if ($ip =~ /:/) {
                $ip = "[" . $ip . "]";
            }

            # see if this target is in the list output.
            if (not grep(/^\s*${ip}:3260,\d+ ${nodename}\s*$/, @list_output)) {
                # skip this target since it is not in the output
                next;
            }

            # disable CHAP
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.session.auth.authmethod -v None";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # reset the CHAP in-credentials
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.session.auth.username_in -v \"\"";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.session.auth.password_in -v \"\"";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # reset the CHAP out-credentials
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.session.auth.username -v \"\"";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.session.auth.password -v \"\"";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # disable Header/Data Digest
            $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.conn[0].iscsi.HeaderDigest -v None";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
            if ($self->{major_revision} !~ /rhel|ol|fedora/i) {
                # Data Digest is not supported in redhat's and oracle linux's open-iscsi, this command will fail if we try it
                $cmd = "/sbin/iscsiadm -m node -T $nodename -p ${ip}:3260 -o update -n node.conn[0].iscsi.DataDigest -v None";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_openiscsi_sendtargets()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_openiscsi_sendtargets {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @list_output);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # list out all the sendtargets currently on the system
    $cmd = "/sbin/iscsiadm -m discovery";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 and scalar(@list_output) > 0) {
        # status can be non-zero when there are no targets, don't error out
        # if there is nothing in @list_output
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    # remove all sendtargets targets if clean_level == 2
    if ($clean_level == 2) {
        # remove all sendtargets that are listed in @list_output
        foreach my $line (@list_output) {
            if ($line =~ /(\S+):3260 via sendtargets/i) {
                my $target_ip = $1;

                $cmd = "/sbin/iscsiadm -m discovery -o delete -p $target_ip";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # comment out the changes we made to the /etc/iscsid.conf file for discovery chap authentication
        $self->_comment_out_parameter_in_file("discovery.sendtargets.auth.authmethod", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_comment_out_parameter_in_file("discovery.sendtargets.auth.username", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_comment_out_parameter_in_file("discovery.sendtargets.auth.password", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_comment_out_parameter_in_file("discovery.sendtargets.auth.username_in", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        $self->_comment_out_parameter_in_file("discovery.sendtargets.auth.password_in", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

    } else {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # if there are no targets in the output, then return immediately
        if (scalar(@list_output) == 0) {
            return $ClientUtil::TRUE;
        }

        # remove all sendtargets targets in @targets if clean_level == 1
        foreach my $target (@targets) {
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            if (grep(/$ip\:3260 via sendtargets/i, @list_output)) {
                # delete the target because it is in the list of targets currently added
                $cmd = "/sbin/iscsiadm -m discovery -o delete -p $ip";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_openiscsi_static()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_openiscsi_static {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @list_output);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if there are no targets in the list, return immediately
    $cmd = "/sbin/iscsiadm -m node";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (grep(/no records found/i, @list_output)) {
        return $ClientUtil::TRUE;
    }
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    # remove all static targets if clean_level == 2
    if ($clean_level == 2) {
        $cmd = "/sbin/iscsiadm -m node -o delete";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    } else {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # remove all static targets in @targets if clean_level == 1
        foreach my $target (@targets) {
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            my $nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the target ip is ipv6, then add brackets around it.
            if ($ip =~ /:/) {
                $ip = "[" . $ip . "]";
            }

            #--> iscsiadm -m node
            #192.168.1.61:3260,2000 iqn.1992-08.com.netapp:sn.84174947
            #[fe80::a0:98ff:fe01:70aa]:3260,2000 iqn.1992-08.com.netapp:sn.84174947

            # if this target is in the list, remove it
            if (grep(/^\s*${ip}:3260,\d+\s+${nodename}/i, @list_output)) {
                # delete the target because it is in the list of targets currently added
                $cmd = "/sbin/iscsiadm -m node -o delete -T $nodename -p ${ip}:3260";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_openiscsi_isns()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_openiscsi_isns {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @list_output);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # manually edit the iscsid.conf file and remove the isns address field and port
    # isns.address = <ip>
    # isns.port = 3205
    $self->_comment_out_parameter_in_file("isns.address", "/etc/iscsi/iscsid.conf");
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }
    $self->_comment_out_parameter_in_file("isns.port", "/etc/iscsi/iscsid.conf");
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    $cmd = "/sbin/iscsiadm -m node";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);

    if (not grep(/no records found/i, @list_output)) {
        # see if the command failed
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # if clean_level == 2, remove all targets by doing iscsiadm -m node -o delete
        if ($clean_level == 2) {
            $cmd = "/sbin/iscsiadm -m node -o delete";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        } else {
            # check for required parameters
            if (not scalar(@targets)) {
                $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
                return $ClientUtil::FALSE;
            }

            # if clean_level == 1, remove targets in @targets
            foreach my $target (@targets) {
                my $ip = $target->get_ip();
                if ($target->error()) {
                    $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
                    return $ClientUtil::FALSE;
                }

                my $nodename = $target->get_nodename();
                if ($target->error()) {
                    $self->_set_err("Could not obtain nodename from target: " . $target->errmsg());
                    return $ClientUtil::FALSE;
                }

                # if the target ip is ipv6, then add brackets around it.
                if ($ip =~ /:/) {
                    $ip = "[" . $ip . "]";
                }

                #--> iscsiadm -m node
                #192.168.1.61:3260,2000 iqn.1992-08.com.netapp:sn.84174947
                #[fe80::a0:98ff:fe01:70aa]:3260,2000 iqn.1992-08.com.netapp:sn.84174947

                # if this target is in the list, remove it
                if (grep(/^\s*${ip}:3260,\d+\s+${nodename}/, @list_output)) {
                    $cmd = "/sbin/iscsiadm -m node -o delete -T $nodename -p ${ip}:3260";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0) {
                        $self->_set_err("Command ($cmd) failed with status ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    # need to reload the iscsi driver only if not sanbooted and clean level == 2
    # this will make the iscsidaemon forget about the isns server
    if ($self->{sanbooted} == 0 and $clean_level == 2) {
        $self->_iscsi_service(-action => "restart");
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_linuxiscsi()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure_linuxiscsi {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "rm -f /tmp/iscsi_conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_deconfigure_iscsi failed to remove /tmp/iscsi_conf");
        return $ClientUtil::FALSE;
    }

    $cmd = "touch /tmp/iscsi_conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_deconfigure_iscsi failed to create empty file");
        return $ClientUtil::FALSE;
    }

    my @target_ips;
    foreach my $target (@targets) {
        push (@target_ips, $target->{ip});
    }

    $cmd = "cat /etc/iscsi.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_deconfigure_iscsi failed to cat /etc/iscsi.conf");
        return $ClientUtil::FALSE;
    }

    # Delete all the lines DiscoveryAddress=ip if ip is present
    # in the target list passed into this method
    foreach my $line (@output) {
        if ($line =~ /DiscoveryAddress/i) {
            my ($ip) = ($line =~ /DiscoveryAddress\=(\d+\.\d+\.\d+\.\d+)/);
            next if grep (/$ip/, @target_ips);
        }

        $cmd = "echo \'$line\' >> /tmp/iscsi_conf";
        my ($status, @output_append) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxHost::_deconfigure_iscsi" .
                            " failed to append $line to /tmp/iscsi_conf");
            return $ClientUtil::FALSE;
        }
    }
    $cmd = "mv -f /tmp/iscsi_conf /etc/iscsi.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::_deconfigure_iscsi" .
                        " failed to override /etc/iscsi.conf with /tmp/iscsi_conf");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_service()
# Objective:       perform an iscsi service command "action"
#                  i.e. /etc/init.d/iscsi [start|stop|status|restart]
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_service {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, $action, $status_cmd);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Error: LinuxHost::_iscsi_service() Invalid " .
                                "value for argument '-action' => $_");
                return "";
            }
            next;
        }
    }

    # Verify that the $action parameter is a valid action
    if (not defined($action)) {
        $self->_set_err("Error: LinuxHost::_iscsi_service() action parameter " .
                        "is required");
        return "";
    } elsif ($action !~ /^(start|stop|restart|status)$/) {
        $self->_set_err("Error: LinuxHost::_iscsi_service() Invalid action " .
                        "parameter: \'$action\'");
        return "";
    }

    # if its a restart command, make it do a stop, sleep then start up again
    if ($action =~ /^restart$/) {
        if ($self->{major_revision} =~ /sles|cxs/i) {
            if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
                $cmd = "/etc/init.d/open-iscsi stop; sleep 3; " .
                       "/etc/init.d/open-iscsi start";
                $status_cmd = "/etc/init.d/open-iscsi status";
            } else {
                $cmd = "/etc/init.d/iscsi stop; sleep 3; " .
                       "/etc/init.d/iscsi start";
                $status_cmd = "/etc/init.d/iscsi status";
            }
        } else {
            if ($self->{major_revision} =~ /rhel7|fedora/) {
                $cmd = "systemctl restart iscsid.service";
                $status_cmd = "systemctl status iscsid.service";
            } elsif ($self->{major_revision} =~ /rhel6|ol6/i) {
                $cmd = "/etc/init.d/iscsid stop; sleep 3; " .
                    "/etc/init.d/iscsid force-start";
                $status_cmd = "/etc/init.d/iscsid status";
            } else {
                $cmd = "/etc/init.d/iscsi stop; sleep 3; " .
                    "/etc/init.d/iscsi start";
                $status_cmd = "/etc/init.d/iscsi status";
            }
        }
    } else {
        if ($self->{major_revision} =~ /sles|cxs/i) {
            if ($self->{iscsi}->{iscsi_version} =~ /open-iscsi/i) {
                $cmd = "/etc/init.d/open-iscsi $action";
                $status_cmd = "/etc/init.d/open-iscsi status";
            } else {
                $cmd = "/etc/init.d/iscsi $action";
                $status_cmd = "/etc/init.d/iscsi status";
            }
        } else {
            if ($self->{major_revision} =~ /rhel7|fedora/) {
                $cmd = "systemctl $action iscsid.service";
                $status_cmd = "systemctl status iscsid.service";
            } elsif ($self->{major_revision} =~ /rhel6|ol6/i) {
                if ($action =~ /start/) {
                    $action = "force-start";
                }
                $cmd = "/etc/init.d/iscsid $action";
                $status_cmd = "/etc/init.d/iscsid status";
            } else {
                $cmd = "/etc/init.d/iscsi $action";
                $status_cmd = "/etc/init.d/iscsi status";
            }
        }
    }

    # execute the appropriate/formulated command. Do not
    # check status. Look at status of the service to determine
    # pass/fail. There are cases where the service starts, but
    # status is != 0
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # allow time for the action to complete
    &SanUtil::nate_sleep(3);

    # execute the status command to return the current status
    # Don't check status, return information will be validated.
    ($status, @output) = &ClientUtil::execute($self->{conn}, $status_cmd);

    # determine status from output of status command
    if (grep(/running/i, @output)) {
        if ($action =~ /^stop$/) {
            $self->_set_err("iSCSI service command ($cmd) failed\n" .
                            "Issued $action action, but initiator is still running");
            return "";
        } else {
            return "online";
        }
    } else {
        if ($action =~ /^(start|restart)$/) {
            $self->_set_err("iSCSI service command ($cmd) failed\n" .
                            "Issued $action action, but initiator is not running");
            return "";
        } else {
            return "offline";
        }
    }

}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self) = @_;

    # No-op for Linux (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_service_running()
# Objective:       Returns true if service is running else return false
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _is_service_running {
    my ($self, %args) = @_;

    my ($service, $cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?service/) {
            if (not length($service = $args{$_})) {
                $self->_set_err("Invalid value for argument '-service' => $_");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined ($service)) {
        $self->_set_err("'-service' parameter is required");
        return $ClientUtil::FALSE;
    }

    if ($self->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl status $service" . ".service";
    } else {
        $cmd = "/etc/init.d/$service status";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to return the status of service $service");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _comment_out_parameter_in_file()
# Objective:       will comment out all instances of the given parameter in
#                  a config file
################################################################################
sub _comment_out_parameter_in_file {
    my ($self, $param, $file) = @_;

    my $cmd = "perl -p -i.bak -e 'END {exit 0}; s/^\\s*#?\\s*($param\\s*=.*)/#\\1/g' $file";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned status ($status)");
    }

    return ($status, @output);
}

################################################################################
# Subroutine Name: _change_parameter_value_in_file()
# Objective:       Will change the value of a given parameter in a config file
################################################################################
sub _change_parameter_value_in_file {
    my ($self, $param, $value, $file) = @_;

    my ($status, @output) = $self->_comment_out_parameter_in_file($param, $file);
    if ($self->error()) {
        return ($status, @output);
    }

    my $cmd =  "perl -p -i.bak -e 'BEGIN{\$x=0}; END{exit \$x}; if (\$x==0) { if (s/^\\s*#?\\s*($param)\\s*=.+/\\1 = $value/) {\$x=1}}' $file";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 1) {
        # then we successfully made the change, show the diff of the changes we made.
        $cmd = "diff $file.bak $file";
        &ClientUtil::execute($self->{conn}, $cmd);
        return (0, @output);
    } else {
        # otherwise no instances of $param were modified, append the $param and value to the end of the file
        my $line = "$param = $value";
        $cmd = "echo \'$line\' >> $file";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned status ($status)");
            return ($status, @output);
        }
    }

    return ($status, @output);
}

################################################################################
# Subroutine Name _get_wwid()
# Objective:       Will return wwid of the device
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_wwid {
    my ($self, $path) = @_;

    my ($cmd, @wwid, $status, $conn);

    $self->_clear_err();

    $conn = $self->{conn};

    # In SLES11, RHEL6 and OL6 onwards scsi_id is moved to /lib/udev
    if ($self->{major_revision} =~ /sles11|rhel6|rhel7|ol6|fedora/i) {
        $cmd = "/lib/udev/scsi_id -gud /dev/$path";
    } else {
        $cmd = "/sbin/scsi_id -gus /block/$path";
    }
    ($status, @wwid) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Not able to get the wwid of scsi device $path");
        return "";
    }
    return $wwid[0];
}

################################################################################
# Subroutine Name  _rescan_iscsi_serverengines()
# Objective:       Rescan for new LUNs with ServerEngines iSCSI card
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan_iscsi_serverengines {
    my ($self) = @_;

    my ($cmd, $client_tool, $status);
    my (@iscsi_ports, @output, @instance_info, @target_info);

    # make sure that the client tool is installed properly
    $client_tool = "/opt/ServerEngines/BladeEngine2/smclp/client/client";
    $cmd = "ls $client_tool";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not locate client tool ($client_tool)\n" .
                        "Please make sure the ServerEngines client RPMs are installed");
        return $ClientUtil::FALSE;
    }

    # execute client tool to get iSCSI Physical HBAs (ports really)
    $cmd = "$client_tool -x \"show /BladeEngine0/\" | grep iSCSIPhysicalHBA";
    ($status, @iscsi_ports) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not get any iSCSIPHysicalHBAs located on this host");
        return $ClientUtil::FALSE;
    }

    # foreach initiator port, find the logged in targets and rescanLuns
    foreach my $iscsi_hba_port (@iscsi_ports) {
        $iscsi_hba_port =~ s/\s+//g;
        $iscsi_hba_port =~ s/\///g;

        # get the list of targets seen by that initiator port
        $cmd = "$client_tool -x \"show /BladeEngine0/$iscsi_hba_port/LogicalHBA0\" | grep Target";
        ($status, @target_info) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Could not find any targets that have been logged into");
            return $ClientUtil::FALSE;
        }

        foreach my $target (@target_info) {
            $target =~ s/\s+//g;
            $target =~ s/\///g;

            # dump (show) the path to get info on target
            $cmd = "$client_tool -x \"show /BladeEngine0/$iscsi_hba_port/LogicalHBA0/$target\"";
            ($status, @instance_info) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could not dump information on target ($target) on initator ($iscsi_hba_port)");
                return $ClientUtil::FALSE;
            }

            # rescan the LUNs on that target
            $cmd = "$client_tool -x \"OEMSEReScanLUNS /BladeEngine0/$iscsi_hba_port/LogicalHBA0/$target\"";
            ($status, @instance_info) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 or (grep(/Command executed successfully/i, @instance_info) == 0)) {
                $self->_set_err("Could not rescan for LUNs on target ($target) on initiator ($iscsi_hba_port)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _rescan_iscsi_chelsio()
# Objective:       Rescan for new LUNs with Chelsio iSCSI card
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan_iscsi_chelsio {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  _rescan_iscsi_emulex()
# Objective:       Rescan for new LUNs with Emulex iSCSI card
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan_iscsi_emulex {
    my ($self) = @_;

    my ($cmd, $status, @cnas, @output, $potential_mac, $client_tool);

    # make sure that the Emulex hbacmd tool is installed properly
    $client_tool = "/usr/sbin/hbacmd";
    $cmd = "ls $client_tool";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        # if hbacmd tool is not found in /usr/sbin/, try in /usr/sbin/hbanyware/ directory
        $client_tool = "/usr/sbin/hbanyware/hbacmd";
        $cmd = "ls $client_tool";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Could not locate client tool ($client_tool).\nPlease make sure " .
                            "Emulex HBAnywhere/OCmanager RPMs for iSCSI are installed");
            return $ClientUtil::FALSE;
        }
    }

    # get the list of CNAs on the system
    $cmd = "$client_tool listHBAs";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could execute ($cmd) to get list of CNAs on the system");
        return $ClientUtil::FALSE;
    }

    # get CNAs and look for Port Type of only iSCSI (do not grab NICs)
    foreach my $line (@output) {
        if ($line =~ /^(Port MAC|Permanent MAC)/i) {
            $potential_mac = $line;
        }
        if ($line =~ /Port\s+Type\s+:\s+iSCSI/i) {
            push(@cnas, $potential_mac);
        }
    }

    foreach my $cna (@cnas) {
        $cna =~ s/(Port MAC|Permanent MAC)\s*://;
        $cna =~ s/\s+//g;

        # get the list of targets connected for this CNA
        $cmd = "$client_tool ShowTarget $cna";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Could execute ($cmd) to get list of targets for this CNA/MAC ($cna)");
            return $ClientUtil::FALSE;
        }

        my @targets = grep(/^Target iSCSI Name:/, @output);
        foreach my $target (@targets) {
            $target =~ s/Target iSCSI Name\s*:\s*//;
            $target =~ s/\s+//g;

            # rescan for each CNA/Target
            $cmd = "$client_tool GetiSCSILuns $cna $target";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could execute ($cmd) to rescan LUNs from CNA/MAC ($cna) and Target ($target)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _get_iscsi_info_openiscsi
# Objective: Collects iscsi info from the client.
#            Packages the information into a complex data structure.
#            Builds a hash of target nodenames.
#            Under each target nodename, there is a hash for each iscsi cmd:
#            alias      - contains scalar value (Can't get for linux)
#            connection - contains an array of hashes
#            initiator  - contains an array of hashes
#            interface  - contains an array of hashes
#            nodename   - contains scalar
#            security   - contains a hash
#            session    - contains an array of hashes
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
#            Also added this hash entry which does not map to a command:
#            discovery_type - will return sendtargets or isns
#            FYI: If the user performed isns and sendtargets discovery of the
#            same target, linux appears to favor sendtargets. We can't really
#            detect this condition.
# Details:
#     Inputs  - none
#     Returns $iscsi_hash - hash of client iscsi info
#             empty hash (do not set_err)  - no sessions or target portals
#             empty hash (set_err) - error condition
################################################################################
sub _get_iscsi_info_openiscsi {
    my ($self) = @_;

    my $iscsi_hash;        # Hash to store iscsi info
    my $cmd;               # Cmd to execute
    my $status;            # Return status of cmd
    my @output;            # Output of the cmd
    my $init_node_name;    # iscsi node name of initiator
    my $init_alias;        # iscsi alias of the initiator

    $self->_clear_err();

    # Phase 1: grab the initiator node name and alias
    $cmd = "cat /etc/iscsi/initiatorname.iscsi";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("$cmd failed");
        return {};
    }

    # Expected output: (note Alias may not be present)
    # [root@ssan-rhel5u4-20 ~]# cat /etc/iscsi/initiatorname.iscsi
    # InitiatorName=iqn.1994-05.com.redhat:ssan-rhel5u4-64-30
    # InitiatorAlias=ssan-rhel5u4-64-30

    # Save the initiator iqn and alias.
    foreach my $line (@output) {
        if ($line =~ /^InitiatorName=(\S+)/i) {
            $init_node_name = $1;
            next;
        }
        if ($line =~ /^InitiatorAlias=(\S+)/i) {
            $init_alias = $1;
        }
    } ## end foreach my $line (@output)

    # If iqn was not found, exit.
    if (not length($init_node_name)) {
        $self->_set_err("Unable to get initiator IQN");
        return {};
    }

    # If alias was not found, we can calculate the default.
    if (not length($init_alias)) {
        $init_alias = $self->hostname;

        # Add the domain if it exists.
        if (length $self->{conn}{hostp}{DOMAIN}) {
            $init_alias = $init_alias . '.' . $self->{conn}{hostp}{DOMAIN};
        }
    }

    # Debug msgs
    $self->sanlog(-msg => "Initiator Nodename=$init_node_name\n"
                    . "Initiator Alias=$init_alias",
                  -msg_level => 2);

    # Phase 2: Query target nodename, interface, and tpgroup information
    $cmd = "iscsiadm -m node -P 1";

    # execute the command.  Add a check in case no records are found.  This
    # check is needed in case the user did not perform sendtargets discovery
    # or login.
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (($status != 0) and ($output[0] !~ /iscsiadm: no records found!/)) {
        $self->_set_err("$cmd failed");
        return {};
    }

    # Expected output:
    # [root@ssan-rhel5u4-20 ~]# iscsiadm -m node -P 1
    # Target: iqn.1992-08.com.netapp:sn.09d0591fe8fb11dfa530123478563412:vs.4
    #         Portal: 10.228.129.125:3260,1027
    #                 Iface Name: default
    # Target: iqn.1992-08.com.netapp:sn.8957aa75e8f711dfa530123478563412:vs.3
    #         Portal: 10.228.130.125:3260,1
    #                 Iface Name: default
    #         Portal: 10.228.128.125:3260,1026
    #                 Iface Name: default
    #         Portal: 10.228.131.125:3260,1
    #                 Iface Name: default

    # IPv6 results look like this:
    # Target: iqn.1992-08.com.netapp:sn.69414119285711e18888123478563412:vs.3
    #     Portal: [fd20:8b1e:b255:8320::470]:3260,1033
    #             Iface Name: default

    # While there are still lines to process, check for a new target
    while (@output) {

        # This specific pattern will yield a target nodename
        my ($target_node_name) = ($output[0] =~ /^Target:\s+(iqn\S+)$/);

        # Throw out the current line.
        shift(@output);

        # If this is not the start of a Target block, try the next line.
        if (not length($target_node_name)) {
            next;
        }

        # If target_nodename was found, save the target/initiator nodenames.
        # Also initialize arrays.  May need to return some empty arrays.
        $iscsi_hash->{$target_node_name}->{nodename} = $target_node_name;
        $iscsi_hash->{$target_node_name}->{show}->{'target-name'} =
          $target_node_name;
        $iscsi_hash->{$target_node_name}->{security}->{'initiator-name'} =
          $init_node_name;
        $iscsi_hash->{$target_node_name}->{connection} = [];
        $iscsi_hash->{$target_node_name}->{session}    = [];
        $iscsi_hash->{$target_node_name}->{initiator}  = [];
        $iscsi_hash->{$target_node_name}->{interface}  = [];
        $iscsi_hash->{$target_node_name}->{tpgroup}    = [];
        $iscsi_hash->{$target_node_name}->{status}     = {};

        # While we keep finding Portal blocks, save the info. Note we can't
        # use the non-greedy search here because all of the : in the ipv6
        # become problematic with the : following the ip.
        while ($output[0] =~ /^\s+Portal:\s+(\S+):(\S+),(\S+)/i) {

            # We have a Portal to process.
            my $ip     = $1;
            my $port   = $2;
            my $tpgtag = $3;
            my $saved  = 0;

            # If the ip address is ipv6, we need to remove the [].
            if ($ip =~ /^\[(.+)\]$/) {
                $ip = $1;
            }

            # If we've already saved a tpgroup with this tag for this vserver,
            # then we don't need to create another tpgroup.  We can simply save
            # the ip. So scan each existing tpgroup.
            foreach my $tpgroup (@{$iscsi_hash->{$target_node_name}->{tpgroup}})
            {

                # If we find the tag already exists, just save the ip address.
                # Also turn on the saved flag.
                if ($tpgtag eq $tpgroup->{'tpgroup-tag'}) {
                    push(@{$tpgroup->{lif}}, $ip);
                    $saved = 1;
                }
            } ## end foreach my $tpgroup (@{$iscsi_hash...

            # If the ip was not saved, then we need to create a new tpgroup.
            # Be sure to save the tpgtag and ip.
            if (not $saved) {
                my %tpgroup;
                $tpgroup{'tpgroup-tag'} = $tpgtag;
                push(@{$tpgroup{lif}}, $ip);
                push(@{$iscsi_hash->{$target_node_name}->{tpgroup}}, \%tpgroup);
            }

            # Now let's save the interface info
            my %interface;
            $interface{address}       = $ip;
            $interface{'ip-port'}     = $port;
            $interface{'tpgroup-tag'} = $tpgtag;

            # Leaving is-home set to true for now.  There is no
            # lif failover in rollingrock, should always be true.
            $interface{'is-home'} = "true";
            $interface{enabled} = "true";

            # Save interface.
            push(@{$iscsi_hash->{$target_node_name}->{interface}}, \%interface);

            # Remove the "Portal:" and "Iface Name:" lines.  Then the while
            # loop restarts and we search for another possible Portal: block.
            for (1 .. 2) {
                shift(@output);
            }
        } ## end while ($output[0] =~ /^\s+Portal:\s+(\S+):(\S+),(\S+)/i)

    } ## end while (@output)

    # Phase 3: Query session, connection, initiator information
    $cmd = "iscsiadm -m session -P 3";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("$cmd failed");
        return {};
    }

    # Expected output:
    # [root@ssan-rhel5u4-20 ~]# iscsiadm -m session -P 3
    # iSCSI Transport Class version 2.0-871
    # version 2.0-871
    # Target: iqn.1992-08.com.netapp:sn.09d0591fe8fb11dfa530123478563412:vs.4
    #         Current Portal: 10.228.129.125:3260,1027
    #         Persistent Portal: 10.228.129.125:3260,1027
    #                 **********
    #                 Interface:
    #                 **********
    #                 Iface Name: default
    #                 Iface Transport: tcp
    #                 Iface Initiatorname: iqn.1994-05.com.redhat:ssan-rhel5u4-64-30
    #                 Iface IPaddress: 10.229.136.75
    #                 Iface HWaddress: <empty>
    #                 Iface Netdev: <empty>
    #                 SID: 4
    #                 iSCSI Connection State: LOGGED IN
    #                 iSCSI Session State: LOGGED_IN
    #                 Internal iscsid Session State: NO CHANGE
    #                 ************************
    #                 Negotiated iSCSI params:
    #                 ************************
    #                 HeaderDigest: None
    #                 DataDigest: None
    #                 MaxRecvDataSegmentLength: 262144
    #                 MaxXmitDataSegmentLength: 65536
    #                 FirstBurstLength: 65536
    #                 MaxBurstLength: 65536
    #                 ImmediateData: Yes
    #                 InitialR2T: No
    #                 MaxOutstandingR2T: 1
    #                 ************************
    #                 Attached SCSI devices:
    #                 ************************
    #                 Host Number: 29 State: running

    # Note IPv6 will have [] around the portals and IPaddresses in the above section.

    # While there are still lines to process, check for a new target
    while (@output) {

        # This specific pattern will yield a target nodename
        my ($target_node_name) = ($output[0] =~ /^Target:\s+(iqn\S+)$/);

        # Throw out the current line.
        shift(@output);

        # If this is not the start of a Target block, try the next line.
        if (not length($target_node_name)) {
            next;
        }

        # Let's make sure the target node name was previously found when
        # we called iscsiadm -m node -P 1.  If the target node name was
        # not found earlier, we error out.
        if (not exists $iscsi_hash->{$target_node_name}) {
            $self->_set_err("Found a session with $target_node_name, "
                    . "but this target was not found by iscsiadm -m node -P 1");
            return {};
        }

        # While we keep finding Portal blocks, save the info.  Note we can't
        # use the non-greedy search here because all of the : in the ipv6
        # become problematic with the : following the ip.
        while ($output[0] =~ /^\s+Current Portal:\s+(\S+):(\S+),(\S+)/i) {

            # We have a Portal to process.
            my $ip     = $1;
            my $port   = $2;
            my $tpgtag = $3;

            # If the ip address is ipv6, we need to remove the [].
            if ($ip =~ /^\[(.+)\]$/) {
                $ip = $1;
            }

            # If portal was found, initialize session, connection, initiator
            my (%session, %connection, %initiator);

            # Save items
            $connection{'connection-state'} = "Full_Feature_Phase";
            $connection{'has-session'}      = "true";
            $connection{'connection-id'}    = 0;
            $connection{'local-address'}    = $ip;
            $connection{'local-ip-port'}    = $port;
            $connection{'tpgroup-tag'}      = $tpgtag;
            $session{'tpgroup-tag'}         = $tpgtag;
            $session{'session-type'}        = "Normal";
            $session{'initiator-alias'}     = $init_alias;
            $session{'initiator-name'}      = $init_node_name;
            $session{'connection-ids'}      = 0;
            $initiator{'initiator-alias'}   = $init_alias;
            $initiator{'initiator-name'}    = $init_node_name;
            $initiator{'tpgroup-tag'}       = $tpgtag;

            # Remove the next 8 lines. (Current Portal - just collected,
            # Persistent Portal, ***, Interface, ***, Iface Name, Iface
            # Transport, and Iface InitiatorName.)  We skip these lines because
            # there is nothing that we need to collect. Iface IPaddress is the
            # next line we care about.
            for (1 .. 8) {
                shift(@output);
            }

            # Should now be able to find the initiator's ip.
            # Use the ? to make it non-greedy (otherwise the ] will
            # get captured).  Note [] appears for ipv6 only.
            my ($init_ip) = ($output[0] =~ /^\s+Iface IPaddress:\s+\[*(.+?)\]*$/);

            # If we couldn't find the ip address, error.
            if (not length($init_ip)) {
                $self->_set_err("Could not find initiator's ip address!");
                return {};
            }

            # Save the data
            $connection{'remote-address'} = $init_ip;

            # Remove 3 lines. (Iface IPaddress - just collected, Iface
            # HWaddress, Iface Netdev)  SID is the next line that we need to
            # collect.
            for (1 .. 3) {
                shift(@output);
            }

            # Should now be able to find the SID
            my ($init_sid) = ($output[0] =~ /^\s+SID:\s+(\S+)$/);

            # If we couldn't find the SID, error.
            if (not length($init_sid)) {
                $self->_set_err("Could not find initiator's SID!");
                return {};
            }

            # Initiator shows SID in decimal, we need to convert to hex.
            $init_sid = sprintf("%x", $init_sid);

            # Pad with leading zero if needed.
            if (length($init_sid) eq 1) {
                $init_sid = "0$init_sid";
            }

            # If length is > 2, we can remove the extra digits.
            if (length($init_sid) > 2) {
                $init_sid =~ s/.+(..)/$1/;
            }

            # Save the isid
            $session{isid}   = "00:02:3d:$init_sid:00:00";
            $initiator{isid} = "00:02:3d:$init_sid:00:00";

            # Throw out the SID line.
            shift(@output);

            # Should now be able to find the Connection State.  Will save it
            # for later.
            my ($conn_state) =
              ($output[0] =~ /^\s+iSCSI Connection State:\s+(.+)$/);

            # If we couldn't find the Connection State, error.
            if (not length($conn_state)) {
                $self->_set_err("Could not find initiator's connection state!");
                return {};
            }

            # Throw out the connection state line.
            shift(@output);

            # Should now be able to find the Session State.  Will save it for
            # later.
            my ($sess_state) =
              ($output[0] =~ /^\s+iSCSI Session State:\s+(.+)$/);

            # If we couldn't find the Session State, error.
            if (not length($sess_state)) {
                $self->_set_err("Could not find initiator's session state!");
                return {};
            }

            # Remove 5 lines. (iSCSI Session State - just collected, Internal
            # iscsid, ***, Negotiated iSCSI params, ***).  HeaderDigest is the
            # next line we care about.
            for (1 .. 5) {
                shift(@output);
            }

            # Should now be able to find the HeaderDigest
            my ($header_digest) = ($output[0] =~ /^\s+HeaderDigest:\s+(\S+)$/);

            # If we couldn't find the HeaderDigest, error.
            if (not length($header_digest)) {
                $self->_set_err("Could not find initiator's HeaderDigest!");
                return {};
            }

            # Save the info
            if ($header_digest =~ /^None$/i) {
                $connection{'header-digest-enabled'} = "false";
            } else {
                $connection{'header-digest-enabled'} = "true";
            }

            # Throw out the HeaderDigest line.
            shift(@output);

            # Should now be able to find the DataDigest
            my ($data_digest) = ($output[0] =~ /^\s+DataDigest:\s+(\S+)$/);

            # If we couldn't find the DataDigest, error.
            if (not length($data_digest)) {
                $self->_set_err("Could not find initiator's DataDigest!");
                return {};
            }

            # Save the info
            if ($data_digest =~ /^None$/i) {
                $connection{'data-digest-enabled'} = "false";
            } else {
                $connection{'data-digest-enabled'} = "true";
            }

            # Remove the DataDigest and MaxRecvDataSegmentLength lines.
            # The MaxRecvDataSegmentLength is what the initiator proposed, not
            # what was negotiated.
            for (1 .. 2) {
                shift(@output);
            }

            # Should now be able to find the MaxXmitDataSegmentLength
            # this is the TargetMaxRecvDataSegementLength.
            my ($tgt_max_recv_data_seg_len) =
              ($output[0] =~ /^\s+MaxXmitDataSegmentLength:\s+(\S+)$/);

            # If we couldn't find the MaxXmitDataSegmentLength, error.
            if (not length($tgt_max_recv_data_seg_len)) {
                $self->_set_err(
                        "Could not find initiator's MaxXmitDataSegmentLength!");
                return {};
            }

            # Save the MaxXmitDataSegmentLength (TargetMaxRecvDataSegementLength)
            $connection{'target-mrdsl'} = $tgt_max_recv_data_seg_len;

            # Throw out the MaxXmitDataSegmentLength line.
            shift(@output);

            # Should now be able to find the FirstBurstLength
            my ($first_burst_length) =
              ($output[0] =~ /^\s+FirstBurstLength:\s+(\S+)$/);

            # If we couldn't find the FirstBurstLength, error.
            if (not length($first_burst_length)) {
                $self->_set_err("Could not find initiator's FirstBurstLength!");
                return {};
            }

            # Save the FirstBurstLength
            $session{'first-burst-length'} = $first_burst_length;

            # Throw out the FirstBurstLength line.
            shift(@output);

            # Should now be able to find the MaxBurstLength
            my ($max_burst_length) =
              ($output[0] =~ /^\s+MaxBurstLength:\s+(\S+)$/);

            # If we couldn't find the MaxBurstLength, error.
            if (not length($max_burst_length)) {
                $self->_set_err("Could not find initiator's MaxBurstLength!");
                return {};
            }

            # Save the MaxBurstLength
            $session{'max-burst-length'} = $max_burst_length;

            # Throw out the MaxBurstLength line.
            shift(@output);

            # Should now be able to find the ImmediateData
            my ($imm_data) = ($output[0] =~ /^\s+ImmediateData:\s+(\S+)$/);

            # If we couldn't find the ImmediateData, error.
            if (not length($imm_data)) {
                $self->_set_err("Could not find initiator's ImmediateData!");
                return {};
            }

            # Save the ImmediateData
            if ($imm_data =~ /^Yes$/i) {
                $session{'immediate-data-enabled'} = "true";
            } else {
                $session{'immediate-data-enabled'} = "false";
            }

            # Throw out the ImmediateData line.
            shift(@output);

            # Should now be able to find the InitialR2T
            my ($initialR2T) = ($output[0] =~ /^\s+InitialR2T:\s+(\S+)$/);

            # If we couldn't find the InitialR2T, error.
            if (not length($initialR2T)) {
                $self->_set_err("Could not find initiator's InitialR2T!");
                return {};
            }

            # Save the InitialR2T
            if ($initialR2T =~ /^Yes$/i) {
                $session{'initial-r2t-enabled'} = "true";
            } else {
                $session{'initial-r2t-enabled'} = "false";
            }

            # Throw out the InitialR2T.
            shift(@output);

            # Should now be able to find the MaxOutstandingR2T
            my ($maxR2T) = ($output[0] =~ /^\s+MaxOutstandingR2T:\s+(\S+)$/);

            # If we couldn't find the MaxOutstandingR2T, error.
            if (not length($maxR2T)) {
                $self->_set_err(
                               "Could not find initiator's MaxOutstandingR2T!");
                return {};
            }

            # Save the MaxOutstandingR2T
            $session{'max-outstanding-r2t'} = $maxR2T;

            # Throw out the MaxOutstandingR2T.
            shift(@output);

            # If the session and connection states were LOGGED IN, then
            # save everything.
            if (    ($sess_state =~ /^LOGGED_IN$/)
                and ($conn_state =~ /^LOGGED IN$/)) {

                # Update status and show info since we have connections.
                $iscsi_hash->{$target_node_name}->{show}->{'status-admin'} =
                  "up";
                $iscsi_hash->{$target_node_name}->{status}->{'is-available'} =
                  "true";
                $iscsi_hash->{$target_node_name}->{status}->{'status-admin'} =
                  "up";

                # Save connection, session, initiator info
                push(@{$iscsi_hash->{$target_node_name}->{connection}},
                     \%connection);
                push(@{$iscsi_hash->{$target_node_name}->{session}}, \%session);
                push(@{$iscsi_hash->{$target_node_name}->{initiator}},
                     \%initiator);
            } else {

                # Else we've got a bad state, let's bail out and set error.
                $self->_set_err("Found connection state=$conn_state and session state=$sess_state!");
                return {};
            }

            # If next line is not Current Portal: or Target, throw it away.
            # Keep throwing away until we find the start of the next Current
            # Portal or Target section.
            while ($output[0] !~ /^\s+Current Portal:|^Target:/) {

                # If there are no more lines to process, break out of the loop
                if (shift(@output) eq undef) {
                    last;
                }
            }

        } ## end while ($output[0] =~ ...

    } ## end while (@output)

    # Phase 4: Query security information
    # Need security for each target in the hash
    foreach my $target (keys %{$iscsi_hash}) {

        # Let's determine the auth method
        $cmd = "iscsiadm -m node -T $target | grep node.session.auth.authmethod";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output. 1 line per session.
        # [root@ssan-rhel5u4-20 /]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.8957aa75e8f711dfa530123478563412:vs.3 | grep node.session.auth.authmethod
        # node.session.auth.authmethod = CHAP
        #
        # If auth method is none:
        # [root@ssan-rhel5u4-20 /]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.8957aa75e8f711dfa530123478563412:vs.3 | grep node.session.auth.authmethod
        # node.session.auth.authmethod = None

        # Let's scan each of the results and make sure they agree.  They should,
        # but just in case.

        # Check the 1st line for a value
        my ($auth_method) =
          ($output[0] =~ /^node.session.auth.authmethod = (\S+)$/);

        # Make sure we got something
        if (not length($auth_method)) {
            $self->_set_err("Problem retrieving auth_method for $target");
            return {};
        }

        # Throw out the line.
        shift(@output);

        # Make sure all remaining lines have the same value.
        foreach my $line (@output) {
            if ($line !~ /^node.session.auth.authmethod = $auth_method$/) {
                $self->_set_err(
                    "Target: $target, all sessions did not agree on auth method");
                return {};
            }
        }

        # If auth method is None, we need to change it to none so it agrees
        # with ONTAP
        if ($auth_method eq "None") {
            $auth_method = "none";
        }

        # Let's determine the username
        $cmd = "iscsiadm -m node -T $target | grep \"node.session.auth.username \"";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output. 1 line per session.
        # [root@ssan-rhel5u4-20 /]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.8957aa75e8f711dfa530123478563412:vs.3 | grep "node.session.auth.username "
        # node.session.auth.username = mike
        #
        # If empty:
        # node.session.auth.username = <empty>

        # Check the 1st line for a value
        my ($user) = ($output[0] =~ /^node.session.auth.username = (\S+)$/);

        # Make sure we got something
        if (not length($user)) {
            $self->_set_err("Problem retrieving username for $target");
            return {};
        }

        # Throw out the line.
        shift(@output);

        # Make sure all remaining lines have the same value.
        foreach my $line (@output) {
            if ($line !~ /^node.session.auth.username = $user$/) {
                $self->_set_err(
                       "Target: $target, all sessions did not agree on username");
                return {};
            }
        }

        # If user is <empty>, we need to change it to '-' so it agrees
        # with ONTAP
        if ($user eq "<empty>") {
            $user = "-";
        }

        # Let's determine the username_in
        $cmd = "iscsiadm -m node -T $target | grep node.session.auth.username_in";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output. 1 line per session.
        # [root@ssan-rhel5u4-20 /]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.8957aa75e8f711dfa530123478563412:vs.3 | grep node.session.auth.username_in
        # node.session.auth.username_in = mike
        #
        # If empty:
        # node.session.auth.username_in = <empty>

        # Check the 1st line for a value
        my ($user_in) =
          ($output[0] =~ /^node.session.auth.username_in = (\S+)$/);

        # Make sure we got something
        if (not length($user_in)) {
            $self->_set_err("Problem retrieving username_in for $target");
            return {};
        }

        # Throw out the line.
        shift(@output);

        # Make sure all remaining lines have the same value.
        foreach my $line (@output) {
            if ($line !~ /^node.session.auth.username_in = $user_in$/) {
                $self->_set_err(
                    "Target: $target, all sessions did not agree on username_in");
                return {};
            }
        }

        # If user_in is <empty>, we need to change it to '-' so it agrees
        # with ONTAP
        if ($user_in eq "<empty>") {
              $user_in = "-";
        }

        # Now we can fill in the information we know.
        # Start by populating connection info
        foreach my $conn (@{$iscsi_hash->{$target}->{connection}}) {
              $conn->{'authentication-method'} = $auth_method;
        }

        # Then fill in security info
        $iscsi_hash->{$target}->{security}->{'auth-type'} = $auth_method;

        # If auth_method is CHAP, save the retrieved values of user-name and
        # outbound-user-name.  Else the auth_method is none, save '-'.  Do this
        # because it is possible to set auth method to none, but still have a
        # user-name and outbound-user-name defined in linux.
        if ($auth_method =~ /^chap$/i) {

              $iscsi_hash->{$target}->{security}->{'user-name'} = $user;
              $iscsi_hash->{$target}->{security}->{'outbound-user-name'} = $user_in;
        } else {

              $iscsi_hash->{$target}->{security}->{'user-name'} = "-";
              $iscsi_hash->{$target}->{security}->{'outbound-user-name'} = "-";
        }
    } ## end foreach my $target (keys %{$iscsi_hash...

    # Phase 5: Determine discovery type
    # Need discovery type for each target in the hash
    foreach my $target (keys %{$iscsi_hash}) {

        # Let's determine the discovery_type
        $cmd = "iscsiadm -m node -T $target | grep node.discovery_type";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output if send_targets, 1 line per session.
        # [root@ssan-rhel5u4-20 ~]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.c6a740eef11311e0b550123478563412:vs.131 | grep node.discovery_type
        # node.discovery_type = send_targets

        # Expected output if isns, 1 line per session.
        # [root@ssan-rhel5u4-20 ~]# iscsiadm -m node -T iqn.1992-08.com.netapp:sn.c6a740eef11311e0b550123478563412:vs.131 | grep node.discovery_type
        # node.discovery_type = isns

        # Check the 1st line for a value
        my ($disc_type) = ($output[0] =~ /^node.discovery_type = (\S+)$/);

        # Make sure we got send_targets or isns.
        if ($disc_type !~ /^(send_targets|isns)$/i) {
            $self->_set_err(
                    "For $target: discovery_type was not send_targets or isns");
            return {};
        }

        # Throw out the line.
        shift(@output);

        # Let's scan each of the remaining results and make sure they agree.
        # They should, but just in case.
        foreach my $line (@output) {
            if ($line !~ /^node.discovery_type = $disc_type$/) {
                $self->_set_err(
                    "Target: $target, all sessions did not agree on discovery type"
                );
                return {};
            }
        }

        # Save proper discovery type into the return hash.
        if ($disc_type =~ /^send_targets$/i) {
            $iscsi_hash->{"$target"}->{"discovery_type"} = "sendtargets";
        } elsif ($disc_type =~ /^isns$/i) {
            $iscsi_hash->{"$target"}->{"discovery_type"} = "isns";

            # Fix for burt 560365.  So if the target was discovered with
            # isns and it has zero connections, we want to delete this
            # target from the hash.  It's not relevant and breaks IscsiHelperCM
            # verify routines.
            if (scalar(@{$iscsi_hash->{"$target"}->{"connection"}}) eq 0) {
                delete($iscsi_hash->{"$target"});
            }
        }

    } ## end foreach my $target (keys %{...

    return $iscsi_hash;
} ## end sub _get_iscsi_info_openiscsi

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       will set $self->{sanbooted} and will return mp device which has
#                  root(/) on it
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub get_root_dev {
    my ($self) = @_;
    my ($cmd , $status, @output, $root_dm_dev, @root_devices);
    my (@output_mp, $root_dev, $root_vg, $root_lv);

    $self->_clear_err();

    # default sanbooted to 0
    $self->{sanbooted} = 0;
    $self->{root_dev_vg} = 0;
    $cmd = "mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running \'$cmd\'");
        return (wantarray()) ? () : [] ;
    }
    my @lines = grep (/\s+\/(?:boot)?\s+/, @output);
    my @root_boot_devs = map { (split(/\s+/))[0] } @lines;

    foreach my $root_boot_dev (@root_boot_devs) {
        # Now / could be on a LV or it is not on LV.
        # if / is not on LV
        if ((my $minor) = ($root_boot_dev =~ /\/dev\/dm-(\d+)/)) {

            # We have to find out which mp device correspond to this dm device
            # in output of the following cmd the first entry is the mp device and
            # the third entry is the minor number associated with dm device (which is the
            # number that is appened to dm- in dm device name). e.g.

            ########## dmsetup -c info  cmd output ###############
            # Name                             Maj Min Stat Open Targ Event  UUID
            #360a9800050334c4f536f497044336f65 253   2 L--w    0    1     14 mpath-360a9800050334c4f536f497044336f65
            #3609a09800050334c4f536f497044336f 253   2 L--w    0    1     14 mpath-3600a09800050334c4f536f497044336f
            ########################################################

            $cmd = "/sbin/dmsetup -c info";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0){
                $self->_set_err("Error while running \'$cmd\'");
                return (wantarray()) ? () : [] ;
            }
            foreach my $line (@output) {
                my @dmsetup_info_each_dev = split (/\s+/,$line);
                if ($minor == $dmsetup_info_each_dev[2]) {
                    ($root_dev) = ($dmsetup_info_each_dev[0] =~ /^([a-z0-9]{33}|mpath\w+).*/);
                    # Delete the partition part from the device name if exists
                    # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                    $root_dev =~ s/p\d+$//;
                    $root_dev =~ s/_part\d+$//;
                    push (@root_devices, $root_dev);

                    # this is on a multipathed device, so we will assume sanbooted
                    $self->{sanbooted} = 1;

                    # last because root(/) will be only on 1 device
                    last;
                }
            }
        }

        # if / is on LV
        elsif (($root_lv) = ($root_boot_dev =~ /\/dev\/mapper\/(.*)/i)) {
            if ($root_lv =~ /(mpath.*)/i) {
                $cmd = "dmsetup -c info |grep $root_lv";
                my ($status, @tmp_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0){
                    $self->_set_err("Error while running \'$cmd\'");
                    return (wantarray()) ? () : [] ;
                }
                my @dmsetup_info_each_dev = split (/\s+/,$tmp_output[0]);
                my $lun_id = pop (@dmsetup_info_each_dev);
                my @lun_id_arr = split (/\-/,$lun_id);
                $root_dev = pop (@lun_id_arr);

                # adding root device to @root_devices, The root deivce could be in mpath or 360a9
                # or 3600a09 format
                push (@root_devices, $root_lv);

                # Delete the partition part from the device name if exists
                # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                $root_dev =~ s/p\d+$//;
                $root_dev =~ s/_part\d+$//;
                push (@root_devices, $root_dev);

                # the root device is a netapp multipathed device, so it is sanbooted
                $self->{sanbooted} = 1;
            } elsif ($root_lv =~ /(360a9[a-z0-9]{28}|3600a09[a-z0-9]{26}).*/i) {
                $root_dev = $1;
                push (@root_devices, $root_dev);

                # this is a netapp multipathed device so we must assume we are sanbooted
                $self->{sanbooted} = 1;
            } elsif ($root_lv =~ /\w+-(\w+)/i) {
                my ($root_vol) = ($root_lv =~ /\w+-(\w+)/i);
                $self->{root_dev_vg} = 1;

                # Sometime cmd will return input/output error messages
                # if there are stale entries for scsi devices
                # here is an example of the cmd output
                #---------------------------------------------------------------
                # /dev/sdae: read failed after 0 of 4096 at 0: Input/output error
                # /dev/sdp: read failed after 0 of 4096 at 0: Input/output error
                # /dev/sdaf: read failed after 0 of 4096 at 0: Input/output error
                # Found duplicate PV KwmI2cKW7ibqmhZH86N0aT9wEBBoK1tt: using /dev/sdr not /dev/sdks
                # VolGroup00:r/w:772:-1:0:2:2:-1:0:1:1:71565312:32768:2184:2184:0:FGwvrk-Y1yS-W49C-fXSO-N5eL-XPMY-inoTGQ
                #---------------------------------------------------------------
                $cmd = "lvdisplay -c | grep $root_vol";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status == 0) {
                    # remove possible DVD/CD ROM medium errors on RH
                    @output = grep(!/No medium found/i, @output);

                    # Filter out input/outut error messages
                    @output = grep(!/input\/output\s+error/i, @output);

                    # Filter out duplicate PVs
                    @output = grep(!/Found duplicate/i, @output);

                    # invalid file type messages
                    @output = grep(!/invalid file type/i, @output);

                    my @lv_arr = split (/:/, $output[0]);
                    $root_vg = $lv_arr[1];

                    # Need to find underlying dm device
                    $cmd = "pvdisplay -c | grep $root_vg";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    # error out only if status > 0 and there are no input/output
                    # error messages
                    if ($status > 0 and scalar(grep(/input\/output\s+error/i, @output) == 0)) {
                        $self->_set_err("Error while running \'$cmd\'");
                        return (wantarray()) ? () : [] ;
                    } else {
                        # remove possible DVD/CD ROM medium errors on RH
                        @output = grep(!/No\s+medium\s+found/i, @output);

                        # Filter out input/outut error messages
                        @output = grep(!/input\/output\s+error/i, @output);

                        # Filter out duplicate PVs
                        @output = grep(!/Found duplicate/i, @output);

                        # Filer out invalid file type messages
                        @output = grep(!/invalid file type/i, @output);

                        foreach my $line (@output) {
                            if ((my $minor) = ($line =~ /\/dev\/dm-(\d+)/)) {
                                $cmd = "dmsetup -c info";
                                my ($status, @tmp_output) = &ClientUtil::execute($self->{conn}, $cmd);
                                if ($status > 0){
                                    $self->_set_err("Error while running \'$cmd\'");
                                    return (wantarray()) ? () : [] ;
                                }
                                foreach $line (@tmp_output) {
                                    my @dmsetup_info_each_dev = split (/\s+/,$line);
                                    if ($minor == $dmsetup_info_each_dev[2]) {
                                        ($root_dev) = ($dmsetup_info_each_dev[0] =~ /^([a-z0-9]{33}|mpath\w+).*/);
                                        # Delete the partition part from the device name if exists
                                        # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                                        $root_dev =~ s/p\d+$//;
                                        $root_dev =~ s/_part\d+$//;
                                        push (@root_devices, $root_dev);
                                        # In case of SANBoot the root will be on mpath because user_friendly_names will be yes
                                        # by default. When the automation is run for the first time the root will be something like
                                        # mpathXp. But if user_friendly_names is set to no (which we recommend) the host will be
                                        # rebooted by automation and there wont we any entry with mpathX corresponding to root in /dev/mapper
                                        # Instead it will be WWID (something like 360a9 or 3600a09). After reboot automation will try to
                                        # flush the root device because it has mpathX in its root_dev list. So we are getting the WWID and
                                        # storing it instead of storing the user friendly names(like mpath)
                                        if ($self->{config_dmmp_use_user_friendly_names} !~ /^1$|^yes$/i) {
                                            $cmd = "multipath -ll $root_dev | grep -P \"360a9\\w+|3600a09\\w+\"";
                                            my ($status, @tmp_output) = &ClientUtil::execute($self->{conn}, $cmd);
                                            if ($status > 0){
                                                $self->_set_err("Error while running \'$cmd\'");
                                                return (wantarray()) ? () : [] ;
                                            }
                                            ($root_dev) = ($tmp_output[0] =~ /(360a9\w+|3600a09\w+)\)?\s+/i);
                                            push (@root_devices, $root_dev);
                                        }

                                        # root is on a multipathed device, it is sanbooted
                                        $self->{sanbooted} = 1;
                                        last;
                                    }
                                }
                            }
                            # if used user_friendly_name = no and root is on mpathed device
                            # allow paths with /dev/mapper and /dev/mpath
                            elsif ($line =~ /360a9|3600a09/i) {
                                ($root_dev) = ($line =~
                                 /(360a9[a-z0-9]{28}p*\d*|3600a09[a-z0-9]{26}p*\d*|360a9[a-z0-9]{28}(_part\d+)?|3600a09[a-z0-9]{26}(_part\d+)?)\:/i);
                                # Delete the partition part from the device name if exists
                                # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                                $root_dev =~ s/p\d+$//;
                                $root_dev =~ s/_part\d+$//;
                                push (@root_devices, $root_dev);

                                # root is on a multipathed device, it is sanbooted
                                $self->{sanbooted} = 1;
                            }
                            # if used user_friendly_name = yes and root is on mpathed device
                            # allow paths with /dev/mapper and /dev/mpath
                            elsif ($line =~ /\/dev\/mapper\/mpath/i or $line =~ /\/dev\/mpath\/mpath/i) {
                                ($root_dev) = ($line =~ /\/(mpath\w+)\:/i);
                                # Delete the partition part from the device name if exists
                                # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                                $root_dev =~ s/p\d+$//;
                                $root_dev =~ s/_part\d+$//;
                                push (@root_devices, $root_dev);
                                # In case of SANBoot the root will be on mpath because user_friendly_names will be yes
                                # by default. When the automation is run for the first time the root will be something like
                                # mpathXp. But if user_friendly_names is set to no (which we recommend) the host will be
                                # rebooted by automation and there wont we any entry with mpathX corresponding to root in /dev/mapper
                                # Instead it will be WWID (something like 360a9 or 3600a09). After reboot automation will try to
                                # flush the root device because it has mpathX in its root_dev list. So we are getting the WWID and
                                # storing it instead of storing the user friendly names(like mpath)
                                if ($self->{config_dmmp_use_user_friendly_names} !~ /^1$|^yes$/i) {
                                    $cmd = "multipath -ll $root_dev | grep -P \"360a9\\w+|3600a09\\w+\"";
                                    my ($status, @tmp_output) = &ClientUtil::execute($self->{conn}, $cmd);
                                    if ($status > 0){
                                        $self->_set_err("Error while running \'$cmd\'");
                                        return (wantarray()) ? () : [] ;
                                    }
                                    ($root_dev) = ($tmp_output[0] =~ /(360a9\w+|3600a09\w+)\)?\s+/i);
                                    push (@root_devices, $root_dev);
                                }

                                # root is on a multipathed device, it is sanbooted
                                $self->{sanbooted} = 1;
                            }
                            # root is on sd or vd or hd devices
                            elsif (($root_dev) = ($line =~ /\/dev\/(x*[vhs]d\w+)/i)) {
                                # Delete the partition part from the device name if exists
                                # In RHEL and OL, partition ends with pxx. In SLES, it ends with _partxx
                                $root_dev =~ s/\d+$//;
                                $root_dev =~ s/_part\d+$//;
                                push (@root_devices, $root_dev);
                                # dont know if the root dev is on a multipathed device, so can't assume sanbooted
                            }
                        }
                    }
                }
            } elsif (!(grep (/No\s+volume\s+groups\s+found/i,@output))) {
                $self->_set_err("Unable to identify root mp device");
                return (wantarray()) ? () : [] ;
            }
        } elsif ((my $root_dg) = ($root_boot_dev =~ /\/dev\/vx\/dsk\/(.*)/i)) {
            $root_dg =~ s/\/.*//i;
            $cmd = "vxdisk -g $root_dg list | grep $root_dg | awk '{print \$1}'";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Error while running \'$cmd\'");
                return (wantarray()) ? () : [] ;
            }
            push (@root_devices, @output);

            # This is a veritas disk and is a multipathed device
            $self->{sanbooted} = 1;
        } elsif ((my $root_dev) = ($root_boot_dev =~ /\/dev\/(x*[vhs]d[a-z]+)/)) {
            # root is on sd or vd or hd device
            # parse out device name from mount output
            push(@root_devices, $root_dev);
        } elsif ($root_boot_dev =~ /\/dev\/cciss\/\S+/) {
            # we are definitely not sanbooted in this case, this is a cciss raid disk
            # most likely an internal disk.
            $self->{sanbooted} = 0;

            # not returning root devices because many methods use this function
            # to perform operations against these root devices, not needed
            # for non scsi drives.
            return (wantarray()) ? () : [];
        } else {
            # The reason why we dont just remove the set err in "else" case is because there
            # may be legitimate configs that have scsi disks mounted on root that
            # are not accounted for here. If we do not return those devices and
            # not set an error, it is possible libraries using this function
            # may attempt to fdisk the boot drive.
            $self->_set_err("Unable to determine Root Device from mount output");
            return (wantarray()) ? () : [];
        }
    }

    # If we have not yet determined if we are sanbooted
    # and there are root devices, then check the vendors
    if ($self->{sanbooted} == 0 and scalar (@root_devices) > 0) {
        # if this is a vmware platform, then its not sanbooted
        # regardless of whether it is RDM booted
        if (grep(/VMware/i, @{$self->{dmidecode_system_info}})) {
            $self->{sanbooted} = 0;
        } else {
            # need to search through all the root devices, if any are vendor NETAPP
            # then it is sanbooted, any other vendor means it is not sanbooted
            foreach my $dev (@root_devices) {
                # if the device is non scsi device, we can safely assume that it is not sanbooted
                if ($dev !~ /sd[a-z]+/i) {
                    $self->{sanbooted} = 0;
                    last;
                } else {
                    $cmd = "cat /sys/block/${dev}/device/vendor";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0){
                        $self->_set_err("Error while running \'$cmd\'");
                        return (wantarray()) ? () : [] ;
                    }

                    if (grep(/NETAPP/, @output)) {
                        # if netapp is the vendor, then it is sanbooted
                        $self->{sanbooted} = 1;
                    }
                }
            }
        }
    }

    return (wantarray()) ? @root_devices : \@root_devices;
}

1;


=pod

=head1 Package Name

LinuxHost.pm

=head1 Author

dl-blocks-host-automation

=head1 Description

For description of methods see B<San::SanClient::Host.pod>.

=head1 Public Methods

=head2 B<get_iscsi_info()>

Collects all known iscsi information from the client.  Packages the information
into a complex data structure.

 Builds a hash of target nodenames.
 Under each target nodename, there is a hash for each iscsi cmd:
  alias      - contains scalar value (Can't get for linux)
  connection - contains an array of hashes
  initiator  - contains an array of hashes
  interface  - contains an array of hashes
  nodename   - contains scalar
  security   - contains a hash
  session    - contains an array of hashes
  show       - contains a hash
  status     - contains a hash
  tpgroup    - contains an array of hashes
 Each hash contains keys which correspond to the c-mode cmd options

=over 4

=item Inputs

=item * None

=item Outputs

=item * $iscsi_hash - hash of client iscsi info

=item * empty hash (do not set_err)  - no sessions or target portals

=item * empty hash (set_err) - error condition

=item * Sample output:

 $VAR1 = {
           'iqn.1992-08.com.netapp:sn.daa982f5014f11e0955a123478563412:vs.3' => {
                                                                                  'connection' => [
                                                                                                    {
                                                                                                      'has-session' => 'true',
                                                                                                      'header-digest-enabled' => 'false',
                                                                                                      'authentication-method' => 'CHAP',
                                                                                                      'target-mrdsl' => '65536',
                                                                                                      'local-ip-port' => '3260',
                                                                                                      'connection-id' => 0,
                                                                                                      'tpgroup-tag' => '1027',
                                                                                                      'connection-state' => 'Full_Feature_Phase',
                                                                                                      'remote-address' => '10.229.136.75',
                                                                                                      'local-address' => '10.228.129.125',
                                                                                                      'data-digest-enabled' => 'false'
                                                                                                    }
                                                                                                  ],
                                                                                  'show' => {
                                                                                              'status-admin' => 'up',
                                                                                              'target-name' => 'iqn.1992-08.com.netapp:sn.daa982f5014f11e0955a123478563412:vs.3'
                                                                                            },
                                                                                  'nodename' => 'iqn.1992-08.com.netapp:sn.daa982f5014f11e0955a123478563412:vs.3',
                                                                                  'status' => {
                                                                                                'status-admin' => 'up',
                                                                                                'is-available' => 'true'
                                                                                              },
                                                                                  'discovery_type' => 'sendtargets',
                                                                                  'initiator' => [
                                                                                                   {
                                                                                                     'initiator-name' => 'iqn.1994-05.com.redhat:ssan-rhel5u4-20',
                                                                                                     'tpgroup-tag' => '1027',
                                                                                                     'isid' => '00:02:3d:0d:00:00',
                                                                                                     'initiator-alias' => 'ssan-rhel5u4-20.gdl.englab.netapp.com'
                                                                                                   }
                                                                                                 ],
                                                                                  'tpgroup' => [
                                                                                                 {
                                                                                                   'tpgroup-tag' => '1027',
                                                                                                   'lif' => [
                                                                                                              '10.228.129.125'
                                                                                                            ]
                                                                                                 }
                                                                                               ],
                                                                                  'session' => [
                                                                                                 {
                                                                                                   'max-burst-length' => '65536',
                                                                                                   'connection-ids' => 0,
                                                                                                   'initial-r2t-enabled' => 'false',
                                                                                                   'immediate-data-enabled' => 'true',
                                                                                                   'isid' => '00:02:3d:0d:00:00',
                                                                                                   'initiator-alias' => 'ssan-rhel5u4-20.gdl.englab.netapp.com',
                                                                                                   'session-type' => 'Normal',
                                                                                                   'initiator-name' => 'iqn.1994-05.com.redhat:ssan-rhel5u4-20',
                                                                                                   'tpgroup-tag' => '1027',
                                                                                                   'first-burst-length' => '65536',
                                                                                                   'max-outstanding-r2t' => '1'
                                                                                                 }
                                                                                               ],
                                                                                  'security' => {
                                                                                                  'auth-type' => 'CHAP',
                                                                                                  'initiator-name' => 'iqn.1994-05.com.redhat:ssan-rhel5u4-20',
                                                                                                  'outbound-user-name' => '-',
                                                                                                  'user-name' => 'mike'
                                                                                                },
                                                                                  'interface' => [
                                                                                                   {
                                                                                                     'is-home' => 'true',
                                                                                                     'tpgroup-tag' => '1027',
                                                                                                     'address' => '10.228.129.125',
                                                                                                     'enabled' => 'true',
                                                                                                     'ip-port' => '3260'
                                                                                                   }
                                                                                                 ]
                                                                                }
         };

=item Examples

    my $iscsi_info = $Client_Obj->get_iscsi_info();

=back

=cut
