# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com, rroberts@netapp.com, latif@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::SolarisHost;

# Package Requirements
######################
use File::Basename;

use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanUtil;
use San::SanCIT;
use Error;

use vars qw(@EXPORT @ISA $NEW_DEV_LABEL);

$NEW_DEV_LABEL = "NATE";

@ISA = qw(Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new SolarisHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self, $conn, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
        if (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;

    # execute "uname -p" and "uname -r" to check the OS version
    $self->_determine_os_version();

    # obtain information regarding the memory on this system.
    $self->_obtain_memory_info();

    return ($self);
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);
    my ($cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            $cfmode = $args{$_};
        }
    }

    # get protocol
    my $protocol = $self->{protocol};

    # For now, no validation on NFS configs
    if ($protocol =~ /nfs/i) {
        $self->sanlog(-msg => "Protocol is set to NFS and currently no host validation is being performed");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # Get the location of solaris_config.  For pNate, we have a static location.  For regular Nate we will
    # try to determine the location of the user's p4 area and then fallback to an /x/eng/localtest location
    my $solaris_config;
    if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
        $solaris_config = "/opt/NETAPP/bin/solaris_config";
        # check to see if validaion script is available
        $cmd = "ls $solaris_config";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if (grep(/No such file or directory/i, @output) > 0) {
            $self->_set_err("Could not find $solaris_config tool");
            $self->{valid} = 0;
            # not ready to fail if the binary cannot be found in a pNATE install
            return $ClientUtil::TRUE;
        }
    } else {
        # Try to get the p4 directory
        my $p4_dir = SanCIT::get_p4_dir(-fallback_dir => "");
        if (defined($p4_dir) and $p4_dir ne "") {
            # We have a p4 dir so set solaris_config to  ${p4_dir}/SANInterop/tools/solaris/scripts/solaris_config
            $self->sanlog(-msg => "Found P4 directory '$p4_dir'. " .
                                  "Using '${p4_dir}/SANInterop/tools/solaris/scripts/solaris_config'");
            $solaris_config = "${p4_dir}/SANInterop/tools/solaris/scripts/solaris_config";

            # check to see if p4 version is available
            $cmd = "ls $solaris_config";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if (grep(/No such file or directory/i, @output) > 0) {
                # Looks like the P4 version of solaris_config wasn't found.  Try the fallback version
                $self->sanlog(-msg => "Unable to find '$solaris_config'.  Trying the fallback version " .
                                      "'$self->{TEST_BIN}/solaris_config'");
                $solaris_config = "$self->{TEST_BIN}/solaris_config";
                $cmd = "ls $solaris_config";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if (grep(/No such file or directory/i, @output) > 0) {
                    # Looks like the fallback version is also missing so set error and return FALSE
                    $self->_set_err("Could not find $solaris_config tool");
                    $self->{valid} = 0;
                    return $ClientUtil::FALSE;
                }
            }
        } else {
            $self->sanlog(-msg => "P4 directory was not found. Using '$self->{TEST_BIN}/solaris_config'");
            $solaris_config = "$self->{TEST_BIN}/solaris_config";

            # check to see if validaion script is available
            $cmd = "ls $solaris_config";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if (grep(/No such file or directory/i, @output) > 0) {
                # Looks like the the fallback version of solaris_config is not found.
                $self->_set_err("Could not find $solaris_config tool");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        }
    }

    # if my_type is not set, do not validate currently
    my $mp_type = $self->{mp_type};
    if (not defined($mp_type) or $mp_type eq "") {
        $self->sanlog(-msg => "Multipathing type (mp_type) is not set. Not validating configuration");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # default cfmode to "standalone" if not set
    if (not defined($cfmode)) {
        $self->sanlog(-msg => "CF Mode is not defined.  Defaulting to 'standalone'");
        $cfmode = "standalone";
    }

    # get the MP stack info
    my $veritas_version = "";
    my $veritas_version_sp = "";
    my $mp  = "";
    if ($mp_type =~ /mpxio/i) {
        $mp_type = "mpxio";
    } elsif ($mp_type =~ /dmp|vrts|veritas/i) {
        $mp_type = "veritasdmp";
        # Make sure we have an MP object
        if (exists($self->{mp}) and ref($self->{mp}) =~ /dmp/i) {
            $veritas_version = $self->{mp}->{version};
            $mp = $self->{mp}->{mp};
            # Fits the regex /MP\d+/ when parsed out of pkginfo command
            $mp =~ s/MP//;
            if ($mp eq "" or not defined($mp)) {
                $mp = 0;
            }
            # See if we have an SP installed
            if (defined($self->{mp}->{sp}) and $self->{mp}->{sp} ne "") {
                $veritas_version_sp = $self->{mp}->{sp};
                $veritas_version_sp =~ s/SP//g;
            } else {
                $veritas_version_sp = 0;
            }
        } else {
            $self->_set_err("MP object either doesn't exists or is not a DMP object");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Unknown mp_type ($mp_type) for Solaris. Failing validation");
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    # Get the alua support info
    my $alua_support = $self->get_alua_support();
    if (not defined($alua_support) or $alua_support eq "") {
        if ($self->error) {
            $self->_set_err("Unable to determine the ALUA support information:\n", $self->errmsg());
        } else {
            $self->_set_err("Unable to determine the ALUA support information");
        }
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    # Get the LVM name so we can use it to determine if veritas is being used or not
    # for solaris_config.
    my $lvm_obj;
    my $lvm;
    if (defined($self->{obj}{lvm}) and $self->{obj}{lvm} ne "") {
        # Get the lvm object and then get the lvm name
        $lvm_obj = $self->{obj}{lvm};

        # Check for an error during instantiation
        if ($lvm_obj->error()) {
            $self->_set_err("A problem occurred during instantiation:\n",
                            $lvm_obj->errmsg());
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }

        $lvm = $lvm_obj->get_name();

        # Check for an error
        if ($lvm_obj->error()) {
            $self->_set_err("Unable to determine the lvm type for client:\n",
                            $lvm_obj->errmsg());
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
        # Check for an empty return string
        if ($lvm eq "") {
            $self->_set_err("Unable to determine the lvm type for client");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
    } else {
        $lvm = "none";
    }

    # Make sure the the restore daemon is running for veritas configs only.  For 6.0, the process
    # has changed.
    if ($mp_type eq "veritasdmp") {
        if ($veritas_version >= 5.0) {
            # We want to look at the vxdmpadm gettune all output
            my $cmd  = "/sbin/vxdmpadm gettune all";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # Now let's check for some settings
            # vxdmpadm gettune all
            # Tunable                           Current Value  Default Value
            # ------------------------------    -------------  -------------
            # dmp_restore_state                   enabled          enabled
            # dmp_fast_recovery                        on               on
            # dmp_log_level                             4                1
            # dmp_lun_retry_timeout                     0                0
            # dmp_probe_idle_lun                      off               on
            # dmp_restore_interval                     60              300
            # dmp_restore_policy             check_disabled   check_disabled
            # dmp_monitor_fabric                       on               on
            my @errors_found = ();
            foreach my $line (@output) {
                if ($line =~ /^dmp_restore_state\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_state'") if ($1 ne "enabled");
                } elsif ($line =~ /^dmp_fast_recovery\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_fast_recovery'") if ($1 ne "on");
                } elsif ($line =~ /^dmp_log_level\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_log_level'") if ($1 != 4);
                } elsif ($line =~ /^dmp_lun_retry_timeout\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_lun_retry_timeout'") if ($1 != 0);
                } elsif ($line =~ /^dmp_probe_idle_lun\s+(\S+)\s+/ and $veritas_version == 5.0 and $mp <= 1) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_probe_idle_lun'") if ($1 ne "on");
                } elsif ($line =~ /^dmp_probe_idle_lun\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_probe_idle_lun'") if ($1 ne "off");
                } elsif ($line =~ /^dmp_restore_interval\s+(\S+)\s+/ and !($veritas_version == 5.0 and $mp <= 1)) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_interval'") if ($1 != 60);
                } elsif ($line =~ /^dmp_restore_policy\s+(\S+)\s+/ and !($veritas_version == 5.0 and $mp <= 1)) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_policy'") if ($1 ne "check_disabled");
                } elsif ($line =~ /^dmp_monitor_fabric\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_monitor_fabric'") if ($1 ne "on");
                }
            }

            # Now check out the settings that are 5.1SP1 or later
            if ($veritas_version >= 6.0 or ($veritas_version == 5.1 and $veritas_version_sp > 0)) {
                # Tunable                           Current Value  Default Value
                # ------------------------------    -------------  -------------
                # dmp_path_age                            120              300
                foreach my $line (@output) {
                    if ($line =~ /^dmp_path_age\s+(\S+)\s+/) {
                        push(@errors_found, "Invalid value '$1' for 'dmp_path_age'") if ($1 != 120);
                    }
                }
            }

            if (scalar(@errors_found)) {
                my $errors = join("\n", @errors_found);
                $self->sanlog(-msg => $errors);
                $self->_set_err("There were one or more errors in the 'vxdmpadm gettune all' output");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        } else {
            # Run vxdmpadm stat restored
            my $cmd  = "/usr/sbin/vxdmpadm stat restored";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # We want to check the output and look for the following:
            # "The number of daemons running : 1".  The number should not be 0.
            if (grep(/daemons running\s+\:\s+0$/, @output) > 0) {
                $self->_set_err("Veritas Restore Daemon does not appear to be running");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        } ##END if ($veritas_version >= 6.0) {
    } ##END if ($mp_type eq "veritasdmp") {

    my $config_driver_type = "";
    my $stack = "";
    if ($protocol =~ /fcp/i) {
        $protocol = "fcp";
        # get the driver info
        $config_driver_type = $self->{fcp}->{CONFIG_FCP_DRIVER_TYPE};
        if ($config_driver_type eq "" or $config_driver_type =~ /vendor/i) {
            $config_driver_type = "lpfc";
        } elsif ($config_driver_type =~ /native/i) {
            $config_driver_type = "native";
        } else {
            $self->_set_err("Unknown CONFIG_FCP_DRIVER_TYPE setting ($config_driver_type) " .
                            "for Solaris. Failing validation");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }

        # generate string that is input to solaris_config
        if ($lvm =~ /vxvm/i and $mp_type =~ /mpxio/i) {
            $stack = $protocol . "-" . $config_driver_type . "-veritas-mpxio";
        } else {
            $stack = $protocol . "-" . $config_driver_type . "-" . $mp_type;
        }
    } elsif ($protocol =~ /iscsi/i) {
        $protocol = "iscsi";
        if ($lvm =~ /vxvm/i and $mp_type =~ /mpxio/i) {
            $stack = $protocol . "-native-veritas-mpxio";
        } else {
            $stack = $protocol . "-native-" . $mp_type;
        }
    } else {
        $self->sanlog(-msg => "Protocol ($protocol) is not iscsi or fcp, not validating configuration");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # call the solaris_config utility. Return status will dictate
    # number of failures and hence pass/fail
    $cmd = "$solaris_config --cfmode $cfmode --${stack} --alua_support $alua_support";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Errors running solaris_config, check output for errors and correct");
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    $self->{valid} = 1;
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    my ($self) = @_;

    # burt 726583 - deleting zfs boot lun
    my @root_devices = ();

    if ($self->{revision} =~ /5\.11/) { # 5.11

        my $cmd = "zpool status rpool";
        #[root@auto-sol11u1-vm-01--->]zpool status rpool
        #  pool: rpool
        # state: ONLINE
        #  scan: none requested
        #config:
        #
        #        NAME        STATE     READ WRITE CKSUM
        #        rpool       ONLINE       0     0     0
        #          c8t0d0s0  ONLINE       0     0     0
        #
        #errors: No known data errors
        my $status = undef;
        my @output = undef;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        my $zfs_boot_lun = undef;
        foreach my $line (@output) {
            if ($line =~ /(c\d+t\d+d\d+)/) {
                $zfs_boot_lun = $1;
                last;
            }
        }

        push (@root_devices, $zfs_boot_lun);
    }

    return (wantarray()) ? @root_devices : \@root_devices;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes. Currently no-op
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for Solaris, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the NetApp NTP Time Server
#                  (ntp0.eng.netapp.com)
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my (@lines, $ntp_server, $ntp_time_zone);

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Invalid value for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Invalid value for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "mv /etc/inet/ntp.conf /etc/inet/ntp.conf.bak";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # ignore status, since if there is not a conf file, we will
    # just be re-generating in coming steps.

    # create the ntp.conf file with time server set
    @lines = ("server $ntp_server", "driftfile /etc/ntp.drift");
    &ClientUtil::create_remote_file($self->{conn},
                                    "/etc/inet/ntp.conf",
                                    @lines);

    # touch file, just to make sure it exists
    $cmd = "touch /etc/ntp.drift";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not execute $cmd",
                        @output);
        return $ClientUtil::FALSE;
    }

    # perform the online of the ntp service to make persistent.
    if ($self->{revision} =~ /5\.8/ or $self->{revision} =~ /5\.9/) {
        # stop the ntp service daemon - do not check return status
        # not valid - Solaris bug
        $cmd = "/etc/init.d/xntpd stop";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # start the ntp service daemon
        $cmd = "/etc/init.d/xntpd start";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Could not start ntp service",
                            @output);
            return $ClientUtil::FALSE;
        }
    } else {
        # based on current state of service, either restart or enable
        $cmd = "svcs -a | grep \"ntp:default\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Could query ntp service",
                            @output);
            return $ClientUtil::FALSE;
        } else {
            if ($output[0] =~ /online/i) {
                $cmd = "svcadm restart svc:/network/ntp:default";
            } else {
                $cmd = "svcadm enable svc:/network/ntp:default";
            }
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Could not start ntp service",
                                @output);
                return $ClientUtil::FALSE;
            }
        }

        # sleep 10 seconds to make sure service has started
        # sometimes has problems on certain Solaris systems.
        &SanUtil::nate_sleep(10);

        # check to make sure operation works and service is online
        $cmd = "svcs -a | grep \"ntp:default\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($output[0] !~ /online/) {
            $self->_set_err("After trying to enable ntp service, service is still disabled", @output);
            return $ClientUtil::FALSE;
        }
    }

    # print date to have in log
    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Could not execute $cmd",
                        @output);
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: asl_version()
# Objective:       get asl version
# Details:         See POD documentation at the end of this file
################################################################################
sub asl_version {
    my ($self) = @_;

    my ($cmd, @output, $status, $version);

    $self->_clear_err();

    # Determine if we are even running veritas
    $cmd = "/usr/bin/pkginfo -l VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    my $veritas_version = "";
    foreach my $line (@output) {
        if ($line =~ /VERSION:\s+(.*),REV/i) {
            $veritas_version = $1;
            last;
        }
    }

    # Case of 5.1+, support is in VRTS package itself.
    my $pkg;
    if ($self->{version} >= 5.1) {
        $pkg = "VRTSaslapm";
    } elsif ($self->{version} =~ /^5/) {
        $pkg = "VRTSasl";
    } else {
        $pkg = "NTAPasl";
    }

    $cmd = "/usr/bin/pkginfo -l $pkg | grep 'VERSION' | awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Unable to get ASL version");
        return "";
    }

    ($version) = $output[0] =~ /V(.*)/i;

    return $version;
}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    my ($cfo_mode, @targets, @fcp_targets, @iscsi_targets, $skip_reboot);

    $self->_clear_err();

    # parse arguments

    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not (@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?fcp_targets$/) {
            if (not (@fcp_targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-fcp_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?cfo_mode/) {
            if (not defined($cfo_mode = $args{$_})) {
                $self->_set_err("Invalid value for -targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?skip_reboot/) {
            if (not defined($skip_reboot = $args{$_})) {
                $self->_set_err("Invalid value for -skip_reboot => $args{$_}");
                return $ClientUtil::FALSE;
            }

            unless ($skip_reboot =~ /^(1|0)$/) {
                $self->_set_err("Invalid value for -skip_reboot '$skip_reboot'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # If skip_reboot was not passed in then set it to "0"
    if (not defined($skip_reboot) or $skip_reboot eq "") {
        $skip_reboot = 0;
    }

    # If we didn't recieve a cfo_mode, then default to single_image.
    unless (defined($cfo_mode) and $cfo_mode ne "") {
        $self->sanlog(-msg => "Defaulting cfo_mode to 'single_image'");
        $cfo_mode = "single_image";
    }

    # Get the HU toolkit version
    my $hu_version = $self->get_hu_toolkit_version();

    if ($hu_version < 6.0) {
        $self->_set_basic_config(-cfo_mode => $cfo_mode);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
        # Configure for FCP
        if ($self->{protocol} =~ /fcp/i) {
            # run create_bindings for the lpfc stack
            if ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /vendor/i) {
                $self->sanlog(-msg => "Running create_bindings");
                if (scalar(@fcp_targets) > 0) {
                    $self->_create_bindings(-targets => \@fcp_targets);
                } elsif (scalar(@targets) > 0) {
                    $self->_create_bindings(-targets => \@targets);
                } else {
                    $self->_set_err("Unable to find any targets to use when configuring fcp");
                    return $ClientUtil::FALSE;
                }
                return $ClientUtil::FALSE if ($self->error);
            }
        }
    } else {
        $self->_set_host_config();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # reload the sd driver
    $self->_update_drv();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Configure the MP Object. If we don't configure the MP object and we are
    # running iscsi discovery, we will need to reboot.
    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        $self->_MP_configure(-skip_reboot => $skip_reboot);
        return $ClientUtil::FALSE if ($self->error);
    } elsif ($self->{protocol} =~ /iscsi/i and $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} == 1) {
        # reboot because iscsi configure requires it but only if the user didn't tell
        # us not to reboot.
        if ($skip_reboot == 0) {
            $self->reboot();
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "Skipping reboot since it was requested to skip " .
                                  "client reboot");
        }
    }

    # Configure for iSCSI
    if ($self->{protocol} =~ /iscsi/i) {
        if (scalar(@iscsi_targets) > 0) {
            $self->_iscsi_configure(-targets => \@iscsi_targets);
        } elsif (scalar(@targets) > 0) {
            $self->_iscsi_configure(-targets => \@targets);
        } else {
            $self->_set_err("Unable to find any targets to use when configuring iscsi");
            return $ClientUtil::FALSE;
        }
        return $ClientUtil::FALSE if ($self->error);
    }

    $self->sanlog(-msg => "SanClient Configuration is completed");

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: deconfigure
# Objective: deconfigure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets, @fcp_targets, @iscsi_targets, $clean_level);

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not (@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?fcp_targets$/) {
            if (not (@fcp_targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for " . "-fcp_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    if ($self->{protocol} =~ /iscsi/i) {
        # Build the args
        my %cmd_args;

        # Add in targets if defined.
        if (scalar(@iscsi_targets) > 0) {
            $cmd_args{"-targets"} = \@iscsi_targets;
        } elsif (scalar(@targets) > 0) {
            $cmd_args{"-targets"} = \@targets;
        }

        # Add in clean_level if defined.
        if (defined($clean_level)) {
            $cmd_args{"-clean_level"} = $clean_level;
        } else {
            # default it to 1
            $cmd_args{"-clean_level"} = 1;
        }

        # return if clean level is 0
        if ($cmd_args{"-clean_level"} == 0) {
            $self->sanlog(-msg => "iSCSI Deconfigure: Not performing iSCSI deconfiguration because clean_level is '0'");
            return $ClientUtil::TRUE;
        }

        $self->_iscsi_deconfigure(%cmd_args);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: config_paths
# Objective:       N/A for Solaris, provide for compatibility with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;

    my ($cmd, @devices, $status, @output, @hDevices);

    $self->_clear_err();

    $self->sanlog(-msg => "Discover all devices");

    # execute devfsadm to discover new devices
    if ($self->{protocol} =~ /iscsi/i) {
        $cmd = "/usr/sbin/devfsadm -i iscsi";
    } else {
        $cmd = "/usr/sbin/devfsadm";
    }

    # For Solaris 9 and below, devfsadm will discover new
    # LUNs. For Solaris 10, we need to also call cfgadm.
    if ($self->{revision} =~ /5\.8/ or
        $self->{revision} =~ /5\.9/ or
        $self->{protocol} =~ /iscsi/i) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    } else {
        # this will configure the controllers and run devfsadm
        $self->_cfgadm_discover();
        return $ClientUtil::FALSE if ($self->error);
    }

    # sleep 60 seconds to allow operations to complete including MP stack
    # to refresh based on new discovery
    &SanUtil::nate_sleep(60);

    $self->sanlog(-msg => "Locate devices that are not part of a volume group",
                  -msg_level => 3);
    @devices = $self->_get_unlabeled_devices();
    $self->sanlog(-msg => scalar(@devices) .
                          " Devices are not part of a volume group",
                  -msg_level => 3);

    return $ClientUtil::FALSE if ($self->error);

    # fdisk Luns for x86 machines
    if ($self->{arch} =~ /86/i) {
        $self->sanlog(-msg_level => 3,
                      -msg => "Host Architecture is i386.");
        $self->_fdisk_luns(@devices);
        return $ClientUtil::FALSE if ($self->error);
    }

    # label new devices and give them volume names
    if (scalar(@devices) > 0) {
        $self->sanlog(-msg       => "Volume name '$NEW_DEV_LABEL'",
                      -msg_level => 3);

        $self->format(-devices => \@devices, -vol_name => "$NEW_DEV_LABEL");

        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;

    my ($cmd, $devices, $vol_name, $status, @format_cmds);
    my (@used, @output, @hDevices);

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-devices'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?vol_name/) {
            if (not length($vol_name = $args{$_})) {
                $self->_set_err("Missing value for argument '-vol_name'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    ### Check that required parameter '-device' was provided
    unless (defined $devices and ref($devices) eq "ARRAY") {
        $self->_set_err("Missing required argument '-device'");
        return 0;
    }

    ### Check that required parameter '-vol_name' was provided
    unless (defined $vol_name and length $vol_name) {
        $self->_set_err("Missing required argument '-vol_name'");
        return 0;
    }

    # define location of some temporary files to conduct formats on.
    my $filename = "/usr/tmp/format_input." . $$;

    # if device_type is raw, call the _get_format_cmd_raw to offset 1 cyl
    # to prevent trashing label; otherwise call the _get_format_cmds
    if ($self->{device_type} eq 'raw') {
        @format_cmds = $self->_get_format_cmds_raw($vol_name);
    } else {
        @format_cmds = $self->_get_format_cmds($vol_name);
    }
    &ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # iterate through the devices and label
    foreach my $device (@$devices) {
        $self->sanlog(-msg => "Creating volume name on device '$device'",
                      -msg_level => 3);

        # formulate the format command
        $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format -f $filename $device";
        if (defined($self->{debug_level}) && $self->get_debug_level() <= 1) {
            # make the format command silent.
            $cmd =~ s/format/format -s/;
        }

        # format device
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        &SanUtil::nate_sleep(1);
        if ($status > 0) {
            $self->_set_err("format returned a bad status ($status))");

            ### Cleanup temporary format inout files
            &ClientUtil::remove_remote_file($self->{conn}, $filename);
            return 0;
        }

    }

    ### Cleanup temporary format inout files
    &ClientUtil::remove_remote_file($self->{conn}, $filename);

    return 1;

}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self, $mount) = @_;

    my ($cmd, $line, $status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    $cmd = "mount | grep $mount | awk '{print " . '$3' . "}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or length("@output") eq 0) {
        $self->_set_err("Mount command failed!", @output);
        return undef;
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Multiple mount entries found!", @output);
        return undef;
    }

    return $output[0];

}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vxvm dmp);

}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp nfs iscsi);
}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    return qw(ufs vxfs);

}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my (@output,);

    push @output, "Host name:      " . $self->hostname(),
                  "Host type:      " . $self->hosttype(),
                  "OS version:     " . $self->os_version(),
                  "Protocol:       " . $self->{protocol};

    if ($self->{protocol} =~ /fcp|iscsi/i) {
        push @output, "Sanlun version: " . $self->sanlun_version()
    }

    push @output, "\n";

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;

    my ($interval, $cmd, $status, @output, $cnt, $count, @buf);

    $self->_clear_err();

    ### Default interval is 2 seconds
    $interval = 2;

    ### Default count is 1
    $count = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interval'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Missing value for argument '-count'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    ### Remember that the first iostat cycle is wrong so add an extra cycle.
    ### Minimize the output return since 'z' flag doesn't kick in until
    ### second cycle.
    $count += 1;
    $cmd = "iostat -xnzt $interval $count | grep -v '0.0    0.0    0.0    0.0'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("iostat returned a bad status ($status))");
        return (wantarray()) ? () : [];
    }

    ### Clean up the output by removing the first iostat cycle
    $cnt = 0;
    foreach my $line (@output) {
        $cnt++ if ($line =~ /extended device/i);
        if ($cnt > 1) {
            push(@buf, $line);
        }
    }

    push @buf, "\n\n";
    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       get NFS mounts currently mounted on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my @nfs_mounts;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    $cmd = "/usr/sbin/mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to execute command: " . $cmd);
        return (wantarray()) ? () : [];
    }

    # Capture the output of mount for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Look through each line of output and determint if there is a filer in the $targets list mounted to this host
        my @mounts = grep(/$target/i, @output);
        if (scalar(@mounts) == 0) {
            next;
        }

        # Go through the found mounts and populate the return array
        foreach my $mount (@mounts) {
            # /mnt/qa/nate_1_1_hp2620_rtp11 on fas2050-rtp03:/vol/nate_1_1_hp2620_rtp11 rsize=32768,wsize=32768,NFSv3,dev=2000004 on Tue Jul 14 12:57:19 2009
            my ($nfs_mount, undef, $nfs_share, $nfs_options) = split(/\s+/, $mount);
            my $nfs_pt;
            $nfs_pt->{mount} = $nfs_mount;
            $nfs_pt->{share} = $nfs_share;
            $nfs_pt->{options} = $nfs_options;

            push(@nfs_mounts, $nfs_pt);
        }
    }

    if (scalar(@nfs_mounts) == 0) {
        $self->_set_err("Did not find any mounts associated with targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_mounts : \@nfs_mounts;
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       get available NFS shares volumes via showmount -e
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my $interfaces = undef;
    my @nfs_shares;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?interfaces/) {
            if (not length($interfaces = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interfaces'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # If the interface list is not passed in, get it
    if (not defined($interfaces)) {
        $interfaces = $self->get_ip_info();
    }

    # Capture the output of showmount -e for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Collect showmount -e
        $cmd = "/usr/sbin/showmount -e " . $target;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Look for the two following erros, they failures.
        # showmount: <device_name>: RPC: Program not registered
        # showmount: <device_name>: RPC: Unknown host
        if (grep(/showmount\:.+RPC\:/, @output)) {
            $self->_set_err("RPC error occurred during command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        # Grep through the output and only get the ones that have an ip address in the interface list
        foreach my $interface (keys(%{$interfaces})) {
            my $ipv4 = $interfaces->{$interface}->{ipv4};
            my $ipv6 = $interfaces->{$interface}->{ipv6};
            my @results = ();

            # If we have an ipv4 address specified, add it to the results array
            if (defined($ipv4) and $ipv4 =~ /\d+\.\d+\.\d+\.\d+/) {
                push(@results, grep(/$ipv4/i, @output));
            }

            # If we have an ipv6 address specified, add it to the results array
            if (defined($ipv6)) {
                push(@results, grep(/$ipv6/i, @output));
            }

            if (scalar(@results) == 0) {
                $self->sanlog(-msg => "Did not find any shares for interface: $ipv4 $ipv6");
                next;
            }

            # Parse the output and populate nfs_mounts
            foreach my $line (@results) {
                # /vol/nate_1_1_hp2620_rtp11   10.60.242.103
                # /vol/nate_1_1_hprx2600_rtp14 10.60.242.105
                my ($vol) = split(/\s+/, $line);

                my $nfs_share;
                $nfs_share->{volume} = basename($vol);
                $nfs_share->{path} = $vol;
                $nfs_share->{ipv4_address} = $ipv4;
                $nfs_share->{ipv6_address} = $ipv6;
                $nfs_share->{target} = $target;

                push(@nfs_shares, $nfs_share);
            }
        }
    }

    # See if no NFS shares were found
    if (scalar(@nfs_shares) == 0) {
        $self->_set_err("No NFS shares found for targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_shares : \@nfs_shares;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;

    my ($cmd,       $device_mount, $device_fsck,   $mount_pt, $fs_type,
        $fsck_auto, $mount_boot,   $mount_options, $status,   @output);

    $self->_clear_err();

    ### initialize variables
    $fsck_auto     = "-";
    $mount_boot    = "yes";
    $mount_options = "-";

    # parse arguments
    foreach (keys %args) {
        if (/-?device_mount/) {
            if (not length($device_mount = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device_mount'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?device_fsck/) {
            if (not length($device_fsck = $args{$_})) {
                $self->_set_err("Missing value for argument '-device_fsck'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?fs_type/) {
            if (not length($fs_type = $args{$_})) {
                $self->_set_err("Missing value for argument '-fs_type'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?fsck_auto/) {
            if (not length($fsck_auto = $args{$_})) {
                $self->_set_err("Missing value for argument '-fsck_auto'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?mount_boot/) {
            if (not length($mount_boot = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_boot'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?mount_options/) {
            if (not length($mount_options = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_options'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    ### Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount) {
        $self->_set_err("Missing required argument '-device_mount'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $device_fsck and length $device_fsck) {
        $self->_set_err("Missing required argument '-device_fsck'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Missing required argument '-mount_pt'");
        return $ClientUtil::TRUE;
    }

    ### Check that required parameter '-device' was provided
    unless (defined $fs_type and $fs_type =~ /vxfs|ufs/i) {
        $self->_set_err("Missing required argument '-fs_type'");
        return $ClientUtil::TRUE;
    }

    $cmd = "echo "
      . $device_mount . "\t"
      . $device_fsck . "\t"
      . $mount_pt . "\t"
      . $fs_type . "\t"
      . $fsck_auto . "\t"
      . $mount_boot . "\t"
      . $mount_options . " "
      . ">> /etc/vfstab";

    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table ($mount_pt)",
                  -msg_level => 3);

    ### execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to add mount entry to /etc/vfstab file", @output);
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;

    my ($cmd, $mount, $status, @output);

    $self->_clear_err();

    ### Unmount the volume's filesystem
    $cmd = "umount $mntpt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to unmount filesystem", @output);
    }

    $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'",
                  -msg_level => 3);

    ### Remove directory head
    $cmd = "rm -rf $self->{mount}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove mount point '$mount'", @output);
    }

    return $self;

}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    ### Clear vfstab mount entry
    $self->sanlog(-msg => "Clear vfstab mount entry");

    unless (defined $mount and length($mount)) {
        $self->_set_err("Missing required paremeter '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    if ($mount =~ /dev\/vx/i) {
        $self->_set_err("'$mount' reflects device entry and not mount point", @output);
        return $ClientUtil::FALSE;
    }

    $mount =~ s/\//\\\//g;

    $cmd = "perl -e\'"
      . "open(FH,\"< \/etc\/vfstab\") or exit(1);"
      . "open(FD,\"> \/etc\/vfstab_new\") or exit(1);"
      . "while(<FH>){\\\n"
      . "print FD \$_ if (\$_ !~ /$mount\\s/);" . "}\\\n"
      . "close FD;"
      . "close FH;\\\n"
      . "rename(\"/etc/vfstab_new\", \"/etc/vfstab\") or exit(1);" . "\' ";

    ### execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove mount entry from /etc/vfstab file", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify {
    my ($self, $mount) = @_;

    my ($cmd, $entry, $idx, $status, $value, @values, @params, @output, @buf);

    $self->_clear_err();

    # initialize
    $entry = "";
    $idx   = 0;

    # get the vfstabcontent
    $cmd = "cat /etc/vfstab | grep -v '^#' | egrep '    $mount  | $mount ' ";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to verify mount file entry", @output);
        return $entry;
    }

    $self->sanlog(-msg => \@output,
                  -msg_level => 9);

    if (scalar(@output) == 0) {
        $self->_set_err("No mount file entry exists");
    } elsif (scalar(@output) > 1) {
        $self->_set_err("Multiple mount file entries exist", @output);
    } else {
        @params = ("Device to mount:",
                   "Device to fsck:",
                   "Mount pt:",
                   "Fs type:",
                   "Fsck pass:",
                   "Mount at boot:",
                   "Mount Options:");

        @values = split(/\s+/, $output[0]);

        foreach $idx (0 .. (scalar(@params) - 2)) {
            push @buf, $params[$idx] . " " . $values[$idx];
        }

        my $startOpt = (scalar(@params) - 1);
        my $endOpt   = (scalar(@values) - 1);

        push @buf, $params[$startOpt] . " @values[$startOpt..$endOpt]";

    }

    return (wantarray()) ? @buf : \@buf;

}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;

    my ($line, $status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or length("@output") eq 0) {
        $self->_set_err("Unable to verify mount entry", @output);
        return (wantarray()) ? @output : \@output;
    }

    unless (grep /$mount/, @output) {
        $self->_set_err("Unable to verify mount entry", @output);
        return (wantarray()) ? @output : \@output;
    }

    ### Need to parse the mount into some usefule structure.  For now just show
    ### that something was found.

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    my ($cmd, $force, $mount_pt, $options, $name, $status, $nodename, @output,
        $resource, $vol_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?resource/) {
            if (not length($resource = $args{$_})) {
                $self->_set_err("Invalid value for argument '-resource'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Missing value for argument '-options'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    # Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource) {
        $self->_set_err("Missing required argument '-resource'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt'
    unless (defined $mount_pt and length $mount_pt) {
        #  Use default path if not defined
        ($nodename = $resource) =~ s/(^\S+):\S+/$1/;

        #Check for empty value
        if ($nodename eq "") {
            $self->_set_err("Unable to determine filer name from device mount '$resource'", @output);
            return $ClientUtil::FALSE;
        }

        $vol_name = basename($resource);
        $mount_pt = "/mnt/qa/" . $nodename . "/" . $vol_name;
        $self->sanlog(-msg => "Using default mount pt '$mount_pt' for $resource");
    }

    # Create the mount point and open up dir privs
    $cmd = "mkdir -p $mount_pt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create mount point directory for volume '$resource'",
                        "Code: $status", @output);

        return $ClientUtil::FALSE;
    }

    # Create the NFS mount command.
    if (defined $options) {
        $cmd = "mount -F nfs -o $options $resource $mount_pt";
    } else {
        $cmd = "mount -F nfs $resource $mount_pt";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        unless ($status == 33) {
            $self->_set_err("Unable to NFS mount volume", "Code: $status", @output);
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount {
    my ($self, %args) = @_;

    my ($cmd, $force, $mount_pt, $name, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Missing value for argument '-mount_pt'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    # Create the mount point and open up dir privs
    $cmd = "umount $mount_pt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        unless ("@output" =~ /not mounted/i) {
            $self->_set_err("Unable to unmount directory '$mount_pt'",
                            "Code: $status", @output);

            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    my ($cmd, $status, @output);

    # check if the mount point lives on something that is not mounted on /
    $cmd = "df -k $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);

    # Check the status of the mount command
    if ($status != 0) {
        $self->_set_err("Unable to execute command ($cmd)");
        return $ClientUtil::FALSE;
    }

    #> df -k /usr
    #Filesystem            kbytes    used   avail capacity  Mounted on
    #/dev/dsk/c3t0d0s0    16951496 11195633 5586349    67%    /

    #> df -k /usr/software
    #Filesystem            kbytes    used   avail capacity  Mounted on
    #rtpinf01.rtp.netapp.com:/vol/software/arch/sparc-sun-solaris10
    #                     530369744 397235576 133134168    75%    /usr/software

    # make sure that $mount is not mounted on '/' (unless it is /)
    my $match  = grep(/\%\s*\/\s*$/, @output);
    if ($match and $mount eq '/') {
        return $ClientUtil::TRUE;
    } elsif ($match) {
        # the mount point is not an independent mount point (mounted on /)
        $self->_set_err("Unable to locate mount point ($mount)");
        return $ClientUtil::FALSE;
    } else {
        # mount point is not on /
        return $ClientUtil::TRUE;
    }
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/sbin/shutdown -g 0";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    # cleanup device entries
    $self->sanlog(-msg => "SolarisHost::reboot() Cleaning up stale device entries");
    $self->_devfsadm_cleanup();
    return $ClientUtil::FALSE if ($self->error);

    $cmd = "touch /reconfigure; reboot";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    foreach my $line (@output) {
        $self->sanlog(-msg       => "line=$line\n",
                      -indent    => 4,
                      -msg_level => 7);
    }

    # detect if the host actually rebooted or not
    $self->detect_reboot(-shutdown_timeout => 180,
                         -bootup_timeout   => 360,
                         -bootup_sleep     => 180);
    return $ClientUtil::FALSE if ($self->error);

    # wait an initial 60 seconds for telnet/rsh/ssh service to come
    # up after reboot.
    &SanUtil::nate_sleep(60);

    # reconnect to the client and retry this operation for 10 minutes
    # before giving up and erroring out.
    $self->reconnect();

    my $retries = 10;
    while ($self->error and $retries > 0) {
       $self->sanlog(-msg =>
                     "SolarisHost::reboot() reconnect failed, still ($retries) retries left");
       &SanUtil::nate_sleep(60);
       $retries--;
       $self->reconnect();
    }

    # if we still have an error after 10 minutes of
    # retries, error out.
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    # Loop (retry) for 5 minutes with 30 second increments
    $retries = 10;
    $status = 1;
    $cmd = "uname -a";
    while ($status != 0 and $retries > 0) {
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
       &SanUtil::nate_sleep(30);
       $retries--;
    }
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    my $tempmsg = "Uname output = ";
    $tempmsg .= join "\n", @output;
    $self->sanlog(-msg => $tempmsg);

    # cleanup device entries
    $self->sanlog(-msg => "SolarisHost::reboot() Cleaning up stale device entries");
    $self->_devfsadm_cleanup();
    return $ClientUtil::FALSE if ($self->error);

    $self->{reboot} = 0;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    my ($cmd, @output, $status, $version);

    $self->_clear_err();

    my $path = $self->get_hu_bin_path();
    $cmd = $path . "/sanlun version";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Unable to get filer name");
        return "";
    }

    $version = $output[0];
    $version =~ s/.*\sversion\s(\d+.\d+).*/$1/;

    return $version;

}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For Solaris Host
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems.  We should
#                  have stopped iscsi before we removed the luns so we are
#                  going to reboot and do a iscsi restart here.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output, @rm_output);

    # perform rmdisk on left over VRTS devices (or rmdev on
    # hdisk devcies will fail)
    if ($self->{mp_type} =~ /dmp|vxvm/i) {
        $self->sanlog(-msg => "Removing orphaned vxdisk devices");

        # HUP VRTS, sometimes not all devices are visible
        $cmd = "/usr/sbin/vxdctl enable";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        $cmd = "/usr/sbin/vxdisk -o alldgs list";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # Example Output
        # DEVICE       TYPE            DISK         GROUP        STATUS
        # das_fcp_0_0  auto:none       -            -            online invalid
        # das_fcp_0_1  auto:none       -            -            online invalid
        # das_iscs0_0  auto:none       -            -            online invalid
        # das_iscs0_1  auto:none       -            -            online invalid
        foreach my $line (@output) {
            next if ($line =~  /^DEVICE/ or $line =~ /^Disk_/);

            my @disk_info = split(/\s+/, $line);
            my $dev   = $disk_info[0];
            my $disk  = $disk_info[2];
            my $group = $disk_info[3];
            # if group and disk are "-", no DG's using this.
            # also check to make sure device is not "-" which could
            # be the case in faulted state.
            if ($disk eq "-" and $group eq "-" and $dev ne "-") {
                $cmd = "/usr/sbin/vxdisk rm $dev";
                ($status, @rm_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Command ($cmd) returned a bad status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to Solaris syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @output, $status);

    foreach my $msg (@message) {
        $cmd = "logger -t \"**NATE**\" -p user.error \"$msg\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 10);
        if ($status > 0) {
            $self->_set_err("Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Method:    get_system_stats()
# Objective: Get the cpu/memory usage percentage over a specified interval
################################################################################
sub get_system_stats {
    my ($self, $interval) = @_;

    my $system_stats = {};

    # collect system statistics for $interval length , grep for Average, it will
    # contain the average statistics for the interval we collected
    my $cmd = "/usr/sbin/sar -ru 1 $interval | grep Average";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $interval + 60);
    if ($status > 1) {
        $self->_set_err("Failed to collect system info.");
        return $ClientUtil::FALSE;
    }

    # @output will look like this:
    # Average    58954  1372836
    # Average        0       1       0      99

    # first line, first column is memory free (in pages)
    # second line, final column is cpu idle, (99% idle)

    # average out the samples for CPU usage
    foreach my $line (@output) {
        if ($line =~ /^Average\s+(\d+)\s+\d+\s*$/i) {
            # Memory usage, calculate the total amount of free bytes of memory in order to
            # calculate the memory utilization percentage.
            my $mem_pages_free = $1;
            my $mem_free_bytes = $mem_pages_free * $self->{memory_page_size};
            $system_stats->{mem_usage_percent} = int((1 - ($mem_free_bytes / $self->{memory_size_bytes})) * 100);
        } elsif ($line =~ /^Average\s+(\d+)\s+(\d+)\s+\d+\s+\d+\s*$/i) {
            # CPU usage, calculate CPU utilization by adding user and system time
            my $cpu_user_percent = $1;
            my $cpu_system_percent = $2;
            $system_stats->{cpu_usage_percent} =  int($cpu_user_percent + $cpu_system_percent);
        }
    }

    return $system_stats;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self) = @_;

    my ($mac, $ipv4, $ipv6, $gateway, $subnet_mask, $ipv6_prefix_len);
    my ($set_interface, $mtusize, $broadcast);

    my $start = 0;
    my $finished = 0;
    my $interface = 0;
    my $interface_name = 'NA';
    my $ip_hash = {};

    # get all IP information from host via ifconfig
    my $cmd = "ifconfig -a";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return {};
    }

    my $cnt = 0;
    my $num_lines = scalar(@output);
    foreach my $line (@output) {
        chomp($line);
        $cnt++;

        # get IPv4 address
        if ($line =~ /inet (\S+)/) {
            $ipv4 = $1;
        }

        # get IPv6 address
        if ($line =~ /inet6 (\S+)\/(\d+)/) {
            $ipv6 = $1;
            $ipv6_prefix_len = $2;
        }

        # get MAC address
        if ($line =~ /ether (\S+)/i) {
            $mac = $1;
            # noticed some address are not 0 pad'ed
            # pad out properly
            my @marray = split(/:/, $mac);
            $mac = "";
            foreach my $elem (@marray) {
                if (length($elem) == 1) {
                    $mac .= "0" . $elem;
                } else {
                    $mac .=  $elem;
                }
            }
        }

        # no gateway for UNIX
        $gateway = "";

        # get subnet mask
        if ($line =~ /netmask (\S+)/i) {
            $subnet_mask = $1;
        }

        # get broadcast
        if ($line =~ /broadcast (\S+)/i) {
            $broadcast = $1;
        }



        # Look for unique sections describing interface.
        if ($line =~ /^(\S+:\s+flags=\d+)\</ or $cnt == $num_lines) {
            if ($start == 1) {
                $finished = 1;
                $set_interface = $interface;
                $interface = $1;
            } else {
                $interface = $1;

                # get mtusize for very first record
                if ($line =~ /mtu (\S+)/i) {
                    $mtusize = $1;
                }
            }
            $start = 1;
        }

        if ($start == 1 and $finished == 1) {
            # don't push in loopback
            if ($set_interface !~ /lo0/) {
                $ip_hash->{$set_interface}->{interface_name} = $interface_name;
                $ip_hash->{$set_interface}->{mac} = $mac;
                $ip_hash->{$set_interface}->{description} = "NIC";
                $ip_hash->{$set_interface}->{gateway} = $gateway;
                $ip_hash->{$set_interface}->{broadcast} = $broadcast;
                $ip_hash->{$set_interface}->{subnet_mask} = $subnet_mask;
                $ip_hash->{$set_interface}->{mtusize} = $mtusize;
                $ip_hash->{$set_interface}->{ipv4} = $ipv4;
                $ip_hash->{$set_interface}->{ipv6} = $ipv6;

                # calculate the ipv4 subnet
                if ($ipv4 ne "" and $subnet_mask ne "") {
                    $ip_hash->{$set_interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                                        -ip_version => "ipv4",
                                                                                        -ipv4_netmask => $subnet_mask);
                } else {
                    $ip_hash->{$set_interface}->{ipv4_subnet} = "";
                }

                # calculate the ipv6 subnet
                if ($ipv6 ne "" and $ipv6_prefix_len ne "") {
                    $ip_hash->{$set_interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
                    $ip_hash->{$set_interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                                        -ip_version => "ipv6",
                                                                                        -ipv6_prefix_len => $ipv6_prefix_len);
                } else {
                    $ip_hash->{$set_interface}->{ipv6_prefix_len} = "";
                    $ip_hash->{$set_interface}->{ipv6_subnet} = "";
                }

                # if this is the management interface of the client
                # set an attribute
                my $host_rec = &Tharn::host($self->hostname());
                my $host_ip_default = $host_rec->{'default_ip'};
                if ($ipv4 eq $host_ip_default) {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 1;
                } else {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 0;
                }
            }

            $finished = 0;

            # initialize values
            $mac = $mtusize = $broadcast = $gateway = $subnet_mask = "";
            $ipv4 = $ipv6 = "";
            $ipv6_prefix_len = "";

            # get mtusize for next record
            if ($line =~ /mtu (\S+)/i) {
                $mtusize = $1;
            }
        }
    }

    # return info hash
    return ($ip_hash);

}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    # If we are Solaris 10, return "MPXIO".
    # Official (non-PVR) support started in Solaris 10
    if ($self->{revision} =~ /5\.10/) {
        return("MPXIO");
    }

    # If we are Solaris 9 and Veritas is installed,
    # return "DMP"
    if ($self->{revision} =~ /5\.9/) {
        my $cmd = "pkginfo -l VRTSvxvm";
        my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            return("DMP");
        }
    }

    # If we cannot determine, return "NONE"
    return("NONE");
}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for Solaris
################################################################################
sub rotate_logs {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # rotate the messages file
    $cmd = "/usr/sbin/logadm -p now /var/adm/messages";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # rotate the Veritas logs
    $status = $self->rotate_generic_log(-file => "/etc/vx/dmpevents.log", -minsize => "1024", -keep => "5");
    if ($status == $ClientUtil::TRUE) {
        return $ClientUtil::TRUE;
    } else {
        if (not $self->error) {
            $self->_set_err("An error was received while attempting to rotate the Veritas DMP logs");
        }
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: get_devices()
# Objective:       wrapper to call _get_devices() routine
################################################################################
sub get_devices {
    my ($self) = @_;

    return ($self->_get_devices());

}

################################################################################
# Method:    get_iscsi_info
# Objective: Collects iscsi info from the client.
#            Packages the information into a complex data structure.
#            Builds a hash of target nodenames.
#            Under each target nodename, there is a hash for each iscsi cmd:
#            alias      - contains scalar value
#            connection - contains an array of hashes
#            initiator  - contains an array of hashes
#            interface  - contains an array of hashes
#            nodename   - contains scalar
#            security   - contains a hash
#            session    - contains an array of hashes
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
#            Also added this hash entry which does not map to a command:
#            discovery_type - will return sendtargets or isns
# Details:
#     Inputs  - none
#     Returns $iscsi_hash - hash of client iscsi info
#             empty hash (do not set_err)  - no sessions or target portals
#             empty hash (set_err) - error condition
################################################################################
sub get_iscsi_info {
    my ($self) = @_;

    my $iscsi_hash;            # Hash to store iscsi info
    my $cmd;                   # Cmd to execute
    my $status;                # Return status of cmd
    my @output;                # Output of the cmd
    my $init_node_name;        # iscsi node name of initiator
    my $init_alias;            # iscsi alias of the initiator
    my $auth_method;           # authentication method of initiator
    my $user_name;             # chap user name of initiator
    my $outbound_user_name;    # outbound chap user name

    $self->_clear_err();

    # Phase 1: grab the initiator node name and alias
    $cmd = "iscsiadm list initiator-node";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("$cmd failed");
        return {};
    }

    # Expected output with auth set to NONE:
    # bash-3.00# iscsiadm list initiator-node
    # Initiator node name: iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a
    # Initiator node alias: ssan-240v-2
    #         Login Parameters (Default/Configured):
    #                 Header Digest: NONE/-
    #                 Data Digest: NONE/-
    #         Authentication Type: NONE
    #         RADIUS Server: NONE
    #         RADIUS access: unknown
    #         Configured Sessions: 1

    # Expected output with auth set to CHAP:
    # bash-3.00# iscsiadm list initiator-node
    # Initiator node name: iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a
    # Initiator node alias: ssan-240v-2
    #         Login Parameters (Default/Configured):
    #                 Header Digest: NONE/-
    #                 Data Digest: NONE/-
    #         Authentication Type: CHAP
    #                 CHAP Name: iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a
    #         RADIUS Server: NONE
    #         RADIUS access: unknown
    #         Configured Sessions: 1

    # Note: If alias is not defined, it will show as -, which agrees with
    # Data ONTAP.

    # Save the initiator iqn and alias.
    foreach my $line (@output) {
        if ($line =~ /^Initiator node name:\s+(\S+)/i) {
            $init_node_name = $1;
            next;
        }
        if ($line =~ /^Initiator node alias:\s+(\S+)/i) {
            $init_alias = $1;
            next;
        }
        if ($line =~ /^\s+Authentication Type:\s+(\S+)/i) {
            $auth_method = $1;
            next;
        }
        if ($line =~ /^\s+CHAP Name:\s+(\S+)/i) {
            $user_name = $1;
        }
    } ## end foreach my $line (@output)

    # If iqn was not found, exit.
    if (not length($init_node_name)) {
        $self->_set_err("$cmd: Unable to get initiator IQN");
        return {};
    }

    # If alias was not found, exit.
    if (not length($init_alias)) {
        $self->_set_err("$cmd: Unable to get initiator alias");
        return {};
    }

    # If auth_method was not found, exit.
    if (not length($auth_method)) {
        $self->_set_err("$cmd: Unable to get Authentication Type");
        return {};
    }

    # If auth_method is NONE, convert to lower case to agree with ONTAP.
    # Also set user name to '-' since we are not using CHAP.
    if ($auth_method eq "NONE") {
        $auth_method = "none";
        $user_name = "-";
    }

    # If auth_method is CHAP, make sure we found a user name.
    if (($auth_method eq "CHAP") and (not length($user_name))) {
        $self->_set_err(
                   "$cmd: Auth Type set to CHAP, but could not find user name");
        return {};
    }

    # Debug msgs
    $self->sanlog(-msg => "Initiator Nodename=$init_node_name\n"
                    . "Initiator Alias=$init_alias",
                  -msg_level => 2);

    # Phase 2: Query target nodename, interface, and tpgroup information
    # Loop 1 = check for sendtargets discovery.  loop 2 = isns discovery.
    foreach (my $i = 1 ; $i < 3 ; $i++) {

        if ($i == 1) {
            $cmd = "iscsiadm list discovery-address -v";
        } else {
            $cmd = "iscsiadm list isns-server -v";
        }

        # execute the command.
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output for loop 1, sendtargets discovery:
        # bash-3.00# iscsiadm list discovery-address -v
        # Discovery Address: 10.228.128.125:3260
        #         Target name: iqn.1992-08.com.netapp:sn.ec50f18b020611e0955a123478563412:vs.5
        #                 Target address:  10.228.128.125:3260, 1026
        #         Target name: iqn.1992-08.com.netapp:sn.ec50f18b020611e0955a123478563412:vs.5
        #                 Target address:  10.228.131.125:3260, 1029
        # Expected output for loop 2, sendtargets discovery:
        # bash-3.00# iscsiadm list isns-server -v
        # iSNS Server IP Address: 10.229.136.188:3205
        #        Target name: iqn.1992-08.com.netapp:sn.d8e929e0f0f011e0a831123478563412:vs.34
        #                Target address:  10.228.134.127:3260, 1026
        # Note: It's possible that the user may have performed sendtargets to
        # multiple lifs on the same vserver.  In this case, much of the information
        # will be repeated.  Code below has been updated to handle this.

        # While there are still lines to process, check for a new target
        while (@output) {

            # This specific pattern will yield a target nodename
            my ($target_node_name) =
              ($output[0] =~ /^\s+Target name:\s+(iqn\S+)$/);

            # Throw out the current line.
            shift(@output);

            # If this is not the start of a Target block, try the next line.
            if (not length($target_node_name)) {
                next;
            }

            # We want to raise an error if the target has been discovered by
            # both sendtargets and isns.  So if we are in loop 2 (isns), and
            # the this target has already been seen, and the discovery type
            # was recorded as sendtargets, let's raise the error.
            if (    ($i == 2)
                and (exists $iscsi_hash->{$target_node_name})
                and ($iscsi_hash->{$target_node_name}->{discovery_type} eq
                     "sendtargets")
              ) {
                $self->_set_err(
                        "Target was discovered by both sendtargets and isns. "
                      . "Please only use 1 discovery method.  Target = $target_node_name"
                );
                return {};
            } ## end if (($i == 2) and (exists...

            # If we have not seen this target before, let's initialize it.
            if (not exists $iscsi_hash->{$target_node_name}) {

                # Save the target/initiator nodenames and security info.
                # Also initialize arrays.  May need to return some empty arrays.
                if ($i == 1) {
                    $iscsi_hash->{$target_node_name}->{discovery_type} =
                      "sendtargets";
                } else {
                    $iscsi_hash->{$target_node_name}->{discovery_type} = "isns";
                }
                $iscsi_hash->{$target_node_name}->{nodename} =
                  $target_node_name;
                $iscsi_hash->{$target_node_name}->{show}->{'target-name'} =
                  $target_node_name;
                $iscsi_hash->{$target_node_name}->{security}
                  ->{'initiator-name'} = $init_node_name;
                $iscsi_hash->{$target_node_name}->{security}->{'user-name'} =
                  $user_name;
                $iscsi_hash->{$target_node_name}->{security}->{'auth-type'} =
                  $auth_method;
                $iscsi_hash->{$target_node_name}->{connection} = [];
                $iscsi_hash->{$target_node_name}->{session}    = [];
                $iscsi_hash->{$target_node_name}->{initiator}  = [];
                $iscsi_hash->{$target_node_name}->{interface}  = [];
                $iscsi_hash->{$target_node_name}->{tpgroup}    = [];
                $iscsi_hash->{$target_node_name}->{status}     = {};

            } ## end if (not exists $iscsi_hash...

            # While we keep finding Target address blocks, save the info.
            while ($output[0] =~ /^\s+Target address:\s+(\S+):(\S+),\s+(\S+)/i)
            {

                # We have a Portal to process.
                my $ip     = $1;
                my $port   = $2;
                my $tpgtag = $3;
                my $saved  = 0;

                # If we've already saved a tpgroup with this tag for this vserver,
                # then we don't need to create another tpgroup.  We can simply save
                # the ip if it's not already in the list. So scan each existing
                # tpgroup.
                foreach
                  my $tpgroup (@{$iscsi_hash->{$target_node_name}->{tpgroup}}) {

                    # If we find the tag already exists
                    if ($tpgtag eq $tpgroup->{'tpgroup-tag'}) {

                        # We need to scan all of the ips in the list.
                        foreach my $ip_in_list (@{$tpgroup->{lif}}) {

                            # Check if the ip address is already in the list
                            if ($ip eq $ip_in_list) {

                                # If the ip is in the list, then turn on the saved
                                # flag and exit the foreach $ip_in_list loop.
                                $saved = 1;
                                last;
                            }
                        } ## end foreach my $ip_in_list (@{$tpgroup...

                        # If ip was not in the list, then we need to save it.
                        if (not $saved) {
                            push(@{$tpgroup->{lif}}, $ip);
                            $saved = 1;
                        }

                        # If the code reaches here, then the tag already existed.
                        # We have added the ip if it was not in the list.  So we
                        # can go ahead and break out of the for loop.
                        last;

                    } ## end if ($tpgtag eq $tpgroup...
                } ## end foreach my $tpgroup (@{$iscsi_hash...

                # If the ip was not saved, then we need to create a new tpgroup.
                # Be sure to save the tpgtag and ip.
                if (not $saved) {
                    my %tpgroup;
                    $tpgroup{'tpgroup-tag'} = $tpgtag;
                    push(@{$tpgroup{lif}}, $ip);
                    push(@{$iscsi_hash->{$target_node_name}->{tpgroup}},
                         \%tpgroup);
                }

                # Reset the saved flag.
                $saved = 0;

                # Before we save the interface info, let's scan the current
                # entries to make sure it's not already in the list.
                foreach my $interface (
                             @{$iscsi_hash->{$target_node_name}->{interface}}) {

                    # If we find a match.
                    if (    ($interface->{address} eq $ip)
                        and ($interface->{'ip-port'} eq $port)
                        and ($interface->{'tpgroup-tag'} eq $tpgtag)) {

                        # turn on the saved flag, exit the for loop.
                        $saved = 1;
                        last;
                    }
                } ## end foreach my $interface (@{$iscsi_hash...

                # If interface was not already saved, then save it.
                if (not $saved) {

                    # Now let's save the interface info
                    my %interface;
                    $interface{address}       = $ip;
                    $interface{'ip-port'}     = $port;
                    $interface{'tpgroup-tag'} = $tpgtag;

                    # Leaving is-home set to true for now.  There is no
                    # lif failover in rollingrock, should always be true.
                    $interface{'is-home'} = "true";
                    $interface{enabled} = "true";

                    # Save interface.
                    push(@{$iscsi_hash->{$target_node_name}->{interface}},
                         \%interface);
                } ## end if (not $saved)

                # Throw out the Target address line
                shift(@output);

            } ## end while ($output[0] =~ ...

        } ## end while (@output)

    } ## end foreach (my $i = 1 ; $i < 3...

    # Phase 3: Query session, connection, initiator information

    $cmd = "iscsiadm list target -v";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("$cmd failed");
        return {};
    }

    # Expected output. If you have multiple sessions, there will be 1 block per
    # session.
    # bash-3.00# iscsiadm list target -v
    # Target: iqn.1992-08.com.netapp:sn.ec50f18b020611e0955a123478563412:vs.5
    #         Alias: tester
    #         TPGT: 1026
    #         ISID: 4000002a0000
    #         Connections: 1
    #                 CID: 0
    #                   IP address (Local): 10.228.131.247:44750
    #                   IP address (Peer): 10.228.128.125:3260
    #                   Discovery Method: SendTargets
    #                   Login Parameters (Negotiated):
    #                         Data Sequence In Order: yes
    #                         Data PDU In Order: yes
    #                         Default Time To Retain: 20
    #                         Default Time To Wait: 2
    #                         Error Recovery Level: 0
    #                         First Burst Length: 65536
    #                         Immediate Data: yes
    #                         Initial Ready To Transfer (R2T): yes
    #                         Max Burst Length: 65536
    #                         Max Outstanding R2T: 1
    #                         Max Receive Data Segment Length: 65536
    #                         Max Connections: 1
    #                         Header Digest: NONE
    #                         Data Digest: NONE

    # While there are still lines to process, check for a new target
    while (@output) {

        # This specific pattern will yield a target nodename
        my ($target_node_name) = ($output[0] =~ /^Target:\s+(iqn\S+)$/);

        # Throw out the current line.
        shift(@output);

        # If this is not the start of a Target block, try the next line.
        if (not length($target_node_name)) {
            next;
        }

        # Let's make sure the target node name was previously found when
        # we called iscsiadm list discovery-address -v.  If the target node
        # name was not found earlier, we error out.
        if (not exists $iscsi_hash->{$target_node_name}) {
            $self->_set_err("$cmd: Found a session with $target_node_name, but this " .
                            "target was not found by iscsiadm list discovery-address -v");
            return {};
        }

        # Next line should be target alias.
        # Note: If alias is cleared, it shows as '-' which agrees with ONTAP.
        my ($target_alias) = ($output[0] =~ /^\s+Alias:\s+(\S+)/i);

        # If we couldn't find the Alias, error.
        if (not length($target_alias)) {
            $self->_set_err("$cmd: Could not find target alias!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be TPGT.
        my ($tpgtag) = ($output[0] =~ /^\s+TPGT:\s+(\S+)/i);

        # If we couldn't find the TPGT, error.
        if (not length($tpgtag)) {
            $self->_set_err("$cmd: Could not find TPGT!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be ISID.
        my ($isid) = ($output[0] =~ /^\s+ISID:\s+(\S+)/i);

        # If we couldn't find the ISID, error.
        if (not length($isid)) {
            $self->_set_err("$cmd: Could not find ISID!");
            return {};
        }

        # If ISID is not 12 characters, exit.
        if (length($isid) != 12) {
            $self->_set_err("$cmd: ISID was not 12 characters!");
            return {};
        }

        # Add semi-colons into ISID
        $isid =~ s/(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6/;

        # Throw out the current line.
        shift(@output);

        # Next line should be Connections.  If not, exit.
        my ($conn) = ($output[0] =~ /^\s+Connections:\s+(\S+)/i);

        # If we couldn't find the Connections, error.
        if (not length($conn)) {
            $self->_set_err("$cmd: Could not find Connections!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # If we have zero connections, then there is nothing to save.
        if ($conn eq 0) {
            next;
        }

        # Next line should be CID.
        my ($cid) = ($output[0] =~ /^\s+CID:\s+(\S+)/i);

        # If we couldn't find the CID, error.
        if (not length($cid)) {
            $self->_set_err("$cmd: Could not find CID!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Remote IP and port (from target perspective).
        # (It's local to the initiator)
        my ($remote_ip, $remote_port) =
          ($output[0] =~ /^\s+IP address \(Local\):\s+(\S+):(\S+)/i);

        # If we couldn't find the remote IP, error.
        if (not length($remote_ip)) {
            $self->_set_err("$cmd: Could not find initiator's Local IP!");
            return {};
        }

        # If we couldn't find the remote port, error.
        if (not length($remote_port)) {
            $self->_set_err("$cmd: Could not find initiator's Local port!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Local IP and port (from target perspective).
        # (It's remote to the initiator)
        my ($local_ip, $local_port) =
          ($output[0] =~ /^\s+IP address \(Peer\):\s+(\S+):(\S+)/i);

        # If we couldn't find the local IP, error.
        if (not length($local_ip)) {
            $self->_set_err("$cmd: Could not find initiator's Peer IP!");
            return {};
        }

        # If we couldn't find the local port, error.
        if (not length($local_port)) {
            $self->_set_err("$cmd: Could not find initiator's Peer port!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Discovery Method.  If not, exit.
        if ($output[0] !~ /^\s+Discovery Method:\s+(\S+)/i) {
            $self->_set_err("$cmd: Could not find Discovery Method line!");
            return {};
        }

        # Throw out the current line.  Nothing to save.
        shift(@output);

        # Next line should be Login Parameters (Negotiated).  If not, exit.
        if ($output[0] !~ /^\s+Login Parameters \(Negotiated\):/i) {
            $self->_set_err(
                    "$cmd: Could not find Login Parameters (Negotiated) line!");
            return {};
        }

        # Throw out the current line.  Nothing to save.
        shift(@output);

        # Next line should be Data Sequence In Order.
        my ($data_seq) = ($output[0] =~ /^\s+Data Sequence In Order:\s+(\S+)/i);

        # If we couldn't find the Data Sequence In Order, error.
        if (not length($data_seq)) {
            $self->_set_err("$cmd: Could not find Data Sequence In Order!");
            return {};
        }

        # Convert to ONTAP format.
        if ($data_seq =~ /^yes$/i) {
            $data_seq = "true";
        } else {
            $data_seq = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Data PDU In Order.
        my ($data_pdu) = ($output[0] =~ /^\s+Data PDU In Order:\s+(\S+)/i);

        # If we couldn't find the Data PDU In Order, error.
        if (not length($data_pdu)) {
            $self->_set_err("$cmd: Could not find Data PDU In Order!");
            return {};
        }

        # Convert to ONTAP format.
        if ($data_pdu =~ /^yes$/i) {
            $data_pdu = "true";
        } else {
            $data_pdu = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Default Time To Retain.
        my ($def_time_to_retain) =
          ($output[0] =~ /^\s+Default Time To Retain:\s+(\S+)/i);

        # If we couldn't find the Default Time To Retain, error.
        if (not length($def_time_to_retain)) {
            $self->_set_err("$cmd: Could not find Default Time To Retain!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Default Time To Wait.
        my ($def_time_to_wait) =
          ($output[0] =~ /^\s+Default Time To Wait:\s+(\S+)/i);

        # If we couldn't find the Default Time To Wait, error.
        if (not length($def_time_to_wait)) {
            $self->_set_err("$cmd: Could not find Default Time To Wait!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Error Recovery Level.
        my ($err_recovery_level) =
          ($output[0] =~ /^\s+Error Recovery Level:\s+(\S+)/i);

        # If we couldn't find the Error Recovery Level, error.
        if (not length($err_recovery_level)) {
            $self->_set_err("$cmd: Could not find Error Recovery Level!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be First Burst Length.
        my ($first_burst_length) =
          ($output[0] =~ /^\s+First Burst Length:\s+(\S+)/i);

        # If we couldn't find the First Burst Length, error.
        if (not length($first_burst_length)) {
            $self->_set_err("$cmd: Could not find First Burst Length!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Immediate Data.
        my ($imm_data) = ($output[0] =~ /^\s+Immediate Data:\s+(\S+)/i);

        # If we couldn't find the Immediate Data, error.
        if (not length($imm_data)) {
            $self->_set_err("$cmd: Could not find Immediate Data!");
            return {};
        }

        # Convert to ONTAP format.
        if ($imm_data =~ /^yes$/i) {
            $imm_data = "true";
        } else {
            $imm_data = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Initial Ready To Transfer (R2T).
        my ($initialR2T) =
          ($output[0] =~ /^\s+Initial Ready To Transfer \(R2T\):\s+(\S+)/i);

        # If we couldn't find the Initial Ready To Transfer (R2T), error.
        if (not length($initialR2T)) {
            $self->_set_err("$cmd: Could not find Initial Ready To Transfer (R2T)!");
            return {};
        }

        # Convert to ONTAP format.
        if ($initialR2T =~ /^yes$/i) {
            $initialR2T = "true";
        } else {
            $initialR2T = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Max Burst Length.
        my ($max_burst_length) =
          ($output[0] =~ /^\s+Max Burst Length:\s+(\S+)/i);

        # If we couldn't find the Max Burst Length, error.
        if (not length($max_burst_length)) {
            $self->_set_err("$cmd: Could not find Max Burst Length!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Max Outstanding R2T.
        my ($maxR2T) = ($output[0] =~ /^\s+Max Outstanding R2T:\s+(\S+)/i);

        # If we couldn't find the Max Outstanding R2T, error.
        if (not length($maxR2T)) {
            $self->_set_err("$cmd: Could not find Max Outstanding R2T!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be  Max Receive Data Segment Length.
        my ($tgt_max_recv_data_seg_len) =
          ($output[0] =~ /^\s+Max Receive Data Segment Length:\s+(\S+)/i);

        # If we couldn't find the Max Receive Data Segment Length, error.
        if (not length($tgt_max_recv_data_seg_len)) {
            $self->_set_err("$cmd: Could not find Max Receive Data Segment Length!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Max Connections.
        my ($max_conn) = ($output[0] =~ /^\s+Max Connections:\s+(\S+)/i);

        # If we couldn't find the Max Connections, error.
        if (not length($max_conn)) {
            $self->_set_err("$cmd: Could not find Max Connections!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Header Digest.
        my ($header_digest) = ($output[0] =~ /^\s+Header Digest:\s+(\S+)/i);

        # If we couldn't find the Header Digest, error.
        if (not length($header_digest)) {
            $self->_set_err("$cmd: Could not find Header Digest!");
            return {};
        }

        # Convert to ONTAP format.
        if ($header_digest =~ /^CRC32$/i) {
            $header_digest = "true";
        } else {
            $header_digest = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Data Digest.
        my ($data_digest) = ($output[0] =~ /^\s+Data Digest:\s+(\S+)/i);

        # If we couldn't find the Data Digest, error.
        if (not length($data_digest)) {
            $self->_set_err("$cmd: Could not find Data Digest!");
            return {};
        }

        # Convert to ONTAP format.
        if ($data_digest =~ /^CRC32$/i) {
            $data_digest = "true";
        } else {
            $data_digest = "false";
        }

        # Throw out the current line.
        shift(@output);

        # Now we need to save all of the information that we collected.
        # Initialize data structures.
        my (%session, %connection, %initiator);

        # Save what we collected.
        # Note: Solaris doesn't support MCS, so don't worry about pushing
        # multiple cids to to session('connection-ids').
        $connection{'authentication-method'} = $auth_method;
        $connection{'connection-id'}         = $cid;
        $connection{'connection-state'}      = "Full_Feature_Phase";
        $connection{'data-digest-enabled'}   = $data_digest;
        $connection{'has-session'}           = "true";
        $connection{'header-digest-enabled'} = $header_digest;
        $connection{'local-address'}         = $local_ip;
        $connection{'local-ip-port'}         = $local_port;
        $connection{'target-mrdsl'}          = $tgt_max_recv_data_seg_len;
        $connection{'remote-address'}        = $remote_ip;
        $connection{'remote-ip-port'}        = $remote_port;
        $connection{'tpgroup-tag'}           = $tpgtag;

        $session{'connection-ids'}         = $cid;
        $session{'data-sequence-in-order'} = $data_seq;
        $session{'data-pdu-in-order'}      = $data_pdu;
        $session{'default-time-to-retain'} = $def_time_to_retain;
        $session{'default-time-to-wait'}   = $def_time_to_wait;
        $session{'error-recovery-level'}   = $err_recovery_level;
        $session{'first-burst-length'}     = $first_burst_length;
        $session{'immediate-data-enabled'} = $imm_data;
        $session{'initiator-alias'}        = $init_alias;
        $session{'initial-r2t-enabled'}    = $initialR2T;
        $session{'initiator-name'}         = $init_node_name;
        $session{'isid'}                   = $isid;
        $session{'max-burst-length'}       = $max_burst_length;
        $session{'max-connections'}        = $max_conn;
        $session{'max-outstanding-r2t'}    = $maxR2T;
        $session{'session-type'}           = "Normal";
        $session{'tpgroup-tag'}            = $tpgtag;

        $initiator{'initiator-alias'} = $init_alias;
        $initiator{'initiator-name'}  = $init_node_name;
        $initiator{'isid'}            = $isid;
        $initiator{'tpgroup-tag'}     = $tpgtag;

        # Save connection, session, initiator info
        push(@{$iscsi_hash->{$target_node_name}->{connection}}, \%connection);
        push(@{$iscsi_hash->{$target_node_name}->{session}},    \%session);
        push(@{$iscsi_hash->{$target_node_name}->{initiator}},  \%initiator);

        # Update status/show/alias info since we now know connections exist.
        $iscsi_hash->{$target_node_name}->{show}->{'status-admin'}   = "up";
        $iscsi_hash->{$target_node_name}->{show}->{'target-alias'}   = $target_alias;
        $iscsi_hash->{$target_node_name}->{status}->{'is-available'} = "true";
        $iscsi_hash->{$target_node_name}->{status}->{'status-admin'} = "up";
        $iscsi_hash->{$target_node_name}->{alias} = $target_alias;

    } ## end while (@output)

    # Phase 4: Query 2 way CHAP information from target-param
    # Scan each target that we have found.
    foreach my $target (keys %{$iscsi_hash}) {

        # Let's dump the target-params.
        $cmd = "iscsiadm list target-param -v $target";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$cmd failed");
            return {};
        }

        # Expected output if 2 way CHAP configured
        # bash-3.00# iscsiadm list target-param -v
        # Target: iqn.1992-08.com.netapp:sn.c66ed9fa02d411e0955a123478563412:vs.6
        #         Alias: sol_vserver
        #         Bi-directional Authentication: enabled
        #         Authentication Type: CHAP
        #                 CHAP Name: bart
        #         Login Parameters (Default/Configured):
        #                 Data Sequence In Order: yes/-
        #                 Data PDU In Order: yes/-
        #                 Default Time To Retain: 20/-
        #                 Default Time To Wait: 2/-
        #                 Error Recovery Level: 0/-
        #                 First Burst Length: 65536/-
        #                 Immediate Data: yes/-
        #                 Initial Ready To Transfer (R2T): yes/-
        #                 Max Burst Length: 262144/-
        #                 Max Outstanding R2T: 1/-
        #                 Max Receive Data Segment Length: 8192/-
        #                 Max Connections: 1/-
        #                 Header Digest: NONE/-
        #                 Data Digest: NONE/-
        #         Configured Sessions: 1

        # Expected output if it's just one way CHAP or none:
        # bash-3.00# iscsiadm list target-param -v
        # Target: iqn.1992-08.com.netapp:sn.c66ed9fa02d411e0955a123478563412:vs.6
        #         Alias: sol_vserver
        #         Bi-directional Authentication: disabled
        #         Authentication Type: NONE
        # Note: the rest is identical.  Also note it is possible to have
        # bidirectional disabled, and yet have auth type set to CHAP with a
        # user name.

        # Expected output if we have performed discovery, but did not login.
        # iscsiadm list target-param -v iqn.1992-08.com.netapp:sn.d8e929e0f0f011e0a831123478563412:vs.34
        # iqn.1992-08.com.netapp:sn.d8e929e0f0f011e0a831123478563412:vs.34: not found

        # Make sure we got the right target!  If we did not find the target and
        # we did not get a not found err msg, raise error.
        if (    ($output[0] !~ /^Target:\s+$target$/)
            and ($output[0] !~ /^$target:\s+not found$/)) {
            $self->_set_err("$cmd: Did not match Target: $target");
            return {};
        }

        # It's possible that client may have discovered a target but did not login.
        # In that case, this command will return 'not found'.  Skip to the next
        # target because we won't be able to collect any 2 way CHAP info.
        if ($output[0] =~ /^$target:\s+not found$/) {
            next;
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Alias.
        if ($output[0] !~ /^\s+Alias:\s+\S+$/) {
            $self->_set_err("$cmd: Did not match Alias");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Bi-directional Authentication.
        my ($bi_chap) =
          ($output[0] =~ /^\s+Bi-directional Authentication:\s+(\S+)/i);

        # If we couldn't find Bi-directional Authentication, error.
        if (not length($bi_chap)) {
            $self->_set_err("$cmd: Could not find Bi-directional Authentication!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # Next line should be Authentication Type.
        my ($auth_type_target) =
          ($output[0] =~ /^\s+Authentication Type:\s+(\S+)/i);

        # If we couldn't find the Authentication Type, error.
        if (not length($auth_type_target)) {
            $self->_set_err("$cmd: Could not find Authentication Type!");
            return {};
        }

        # Throw out the current line.
        shift(@output);

        # If the initiator-node reported CHAP and
        # Target-param reports bi-directional is enabled and
        # Target-param Authentication Type is set to CHAP,
        # then we can grab the user name.
        if (    ($auth_method =~ /CHAP/i)
            and ($bi_chap =~ /enabled/i)
            and ($auth_type_target =~ /CHAP/i)) {

            # Fetch the CHAP Name.
            ($outbound_user_name) = ($output[0] =~ /^\s+CHAP Name:\s+(\S+)/i);

            # If we couldn't find the CHAP name, error.
            if (not length($outbound_user_name)) {
                $self->_set_err("$cmd: Could not find CHAP Name!");
                return {};
            }

            # Save the outbound user name.
            $iscsi_hash->{$target}->{security}->{'outbound-user-name'} =
              $outbound_user_name;

        } ## end if (($auth_method =~ /CHAP/i...

        # If outbound user name was not saved, then go ahead and set it to '-'.
        if (not length($outbound_user_name)) {
            $iscsi_hash->{$target}->{security}->{'outbound-user-name'} = "-";
        }

    } ## end foreach my $target (keys %{...

    return $iscsi_hash;
} ## end sub get_iscsi_info

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _MP_configure
# Objective:       Configures MP on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _MP_configure {
    my ($self, %args) = @_;

    my ($skip_reboot);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?skip_reboot/) {
            if (not defined($skip_reboot = $args{$_})) {
                $self->_set_err("Invalid value for -skip_reboot => $args{$_}");
                return $ClientUtil::FALSE;
            }

            unless ($skip_reboot =~ /^(1|0)$/) {
                $self->_set_err("Invalid value for -skip_reboot '$skip_reboot'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # If skip_reboot is not set, then set it to 0
    unless (defined($skip_reboot) and $skip_reboot ne "") {
        $skip_reboot = 0;
    }

    if ($self->{mp}->{mp_type} =~ /dmp/i) {
        # Setup sd.conf
        ### UNKNOWN NEED TO GET sd.conf setup
        return $ClientUtil::FALSE if ($self->error);
    }

    # verify that all mp pkgs are installed
    $self->{mp}->verify(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "MP is verified", -msg_level => 2);

    # configure MP file
    $self->{mp}->config(-conn => $self->{conn},
                        -protocol => $self->{protocol});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "MP is configured", -msg_level => 2);

    # load multipath and start the service
    # x86 Sun servers should have mpxio enabled by default, but will
    # still call the enable in case it was disabled for SUN changes to
    # the default (u4 and greater start providing stmsboot)
    if ($self->{revision} =~ /5\.10/i and
        ($self->{arch} =~ /sparc/i or
        ($self->{arch} =~ /86/i and $self->{minor_revision} > 3))) {
        if (($self->{mp}->{mp_type} =~ /mpxio/i) && ($self->{protocol} =~ /fcp/i)) {
            $self->{mp}->enable(-conn => $self->{conn});
            if ($self->{mp}->error) {
                $self->_set_err($self->{mp}->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->sanlog(-msg => "MP enabled", -msg_level => 2);

    # Rebooting the host if the user wants you too
    if ($skip_reboot == 0) {
        $self->reboot();
        return $ClientUtil::FALSE if ($self->error);
    } else {
        $self->sanlog(-msg => "SolarisHost::_MP_configure(): skipping reboot since it was requested to skip " .
                              "client reboot");
    }

    # After rebooting, solaris Machines using native drivers need to have
    # the controllers reconfigured in order to discover devices.
    if ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /native/i) {
        $self->_cfgadm_discover();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _modify_iscsi_mpxio
# Objective:       disables or enables mpxio on the iscsi driver
# Details:         See POD documentation at the end of this file
################################################################################
sub _modify_iscsi_mpxio {
    my ($self, %args) = @_;

    my ($mpxio_state, $cmd, $status, @output, $mpxio_disable);

    $self->_clear_err();

    # Get the HU toolkit version
    my $hu_version = $self->get_hu_toolkit_version();
    unless (defined($hu_version) and $hu_version ne "") {
        $self->_set_err("Unable to determine the Host Utility version");
        return $ClientUtil::FALSE;
    }

    # For HU 6.0 and later we don't want to have
    # automation do this since stmsboot will do it for us now
    if ($hu_version >= 6.0) {
        return $ClientUtil::True;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?state/) {
            if (not defined($mpxio_state = $args{$_})) {
                $self->_set_err("Invalid value for -state => $args{$_}");
                return $ClientUtil::FALSE;
            }

            unless ($mpxio_state =~ /^(enable|disable)$/) {
                $self->_set_err("Invalid value for -state '$mpxio_state'. " .
                                "Validate states are 'enable' and 'disable'.");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Make sure we received an mpxio state
    unless (defined($mpxio_state) and $mpxio_state ne "") {
        $self->_set_err("Missing required arguement '-state'.");
        return $ClientUtil::FALSE;
    }

    # Since the actual parameter in the iscsi.conf file is "mpxio_disable=<yes|no>" we
    # need to convert enable/disable to the appropriate answer"
    if ($mpxio_state eq "enable") {
        $mpxio_disable = "mpxio-disable=\"no\"\;";
    } else {
        $mpxio_disable = "mpxio-disable=\"yes\"\;";
    }

    # Define the iscsi.conf file and backup
    my $iscsi_conf      = "/kernel/drv/iscsi.conf";
    my $iscsi_conf_temp = "/kernel/drv/iscsi.conf.tmp";

    # Make sure the /kernel/drv/iscsi.conf file exists
    unless (&ClientUtil::remote_file_exists($self->{conn}, $iscsi_conf)) {
        $self->_set_err("File '$iscsi_conf' does not exist on the client");
        return $ClientUtil::FALSE;
    }

    # Remove the temp file if it exists
    if (&ClientUtil::remote_file_exists($self->{conn}, $iscsi_conf_temp)) {
        &ClientUtil::remove_remote_file($self->{conn}, $iscsi_conf_temp, $ClientUtil::FALSE);
    }

    # Grab the contents of the /kernel/drv/iscsi.conf file
    $cmd = "cat $iscsi_conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) == 0) {
        $self->_set_err("No output from command '$cmd' (possible timeout)");
        return $ClientUtil::FALSE;
    }

    # Iterate over the file and look for mpxio_disable.  Store all lines in an
    # array and replace the mpxio_disable line with the appropriate yes/no. If
    # it is not found we will add it to the last line.
    my @new_iscsi_conf;
    foreach my $line (@output) {
        chomp($line);
        # Remove any leading whitespace
        $line =~ s/^\s+//g;

        # Skip the mpxio-disable line since we will add it last.
        next if ($line =~ /^mpxio-disable/);

        push(@new_iscsi_conf, $line);
    }

    # Add the mpxio disable line
    push(@new_iscsi_conf, $mpxio_disable);

    # Write out the file to the temp file.
    &ClientUtil::create_remote_file($self->{conn}, $iscsi_conf_temp, @new_iscsi_conf);

    # Make sure the file size of the temp file is greater than zero
    my $filesize = &ClientUtil::remote_get_filesize($self->{conn}, $iscsi_conf_temp);
    unless (defined($filesize)) {
        $self->_set_err("Unable to retrieve the size of file '$iscsi_conf_temp'");
        return $ClientUtil::FALSE;
    }

    if ($filesize == 0) {
        $self->_set_err("The filesize of '$iscsi_conf_temp' is zero");
        return $ClientUtil::FALSE;
    }

    # Move the temp file back to the original file
    $cmd = "/bin/mv $iscsi_conf_temp $iscsi_conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    unless ($status == 0) {
        $self->_set_err("Command '$cmd' returned with a bad status");
        return $ClientUtil::FALSE;
    }

    # Double check that the temp file no longer exists
    if (&ClientUtil::remote_file_exists($self->{conn}, $iscsi_conf_temp)) {
        $self->_set_err("File '$iscsi_conf_temp' still exists after attempting to move it " .
                        "to '$iscsi_conf'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;

    my ($cmd, $cmdstr, $filename, $status, $status_1, $device, $boot_path);
    my (@output, @output_1, @hDevices, @format_cmds);

    $self->_clear_err();

    # execute format (as root)
    $filename = "/tmp/get_luns.format";
    @format_cmds = ("disk", "0", "quit");
    &ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # Ignore status since disk 0 could be the root disk and format will
    # complain and return 1 (but that will be OK since we will already
    # have gotten the list of devices). Look at @output and make sure it
    # does have output (could be timeout and we want to catch this).
    $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format -f $filename";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) == 0) {
        $self->_set_err("No output from format command (possible timeout)");
        return (wantarray()) ? () : [];
    }
    &ClientUtil::remove_remote_file($self->{conn}, $filename);

    # Get boot path/device
    # Ignore errors and clear the error that might be set by _get_boot_path
    $boot_path = $self->_get_boot_path();

    $self->_clear_err();

    # Creates a remote file on the server that runs inquiry command with format.
    $filename    = "/tmp/inquiry.format";
    @format_cmds = $self->_get_inquiry_cmds();
    &ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # Parsing the format command output to find the device names
    for (my $i = 0; $i <= scalar(@output); $i++) {
        my $line = $output[$i];
        chomp $line;
        my $hDevs = {};
        if ($line =~ /^\s+\d+\.\s+(\S+)\s+\<.+\>\s*(\S*).*$/ &&
            $line !~ /drive not available|drive type unknown/i) {
            # Parsing format command output when it is like following examples:
            # 2. c4t0d0 <NETAPP-LUN-0.2 cyl 2046 alt 2 hd 16 sec 128> <Volname>
            # 3. c4t1d1 <IBM-Nseries2866-A20-7200 cyl 5118 alt 2 hd 16 sec 256>
            #           <Volname>
            # 1. c4t60A980004F644C774234435046657A34d0 <DEFAULT cyl 1020 alt 2
            #    hd 64 sec 32>  <volname>
            $hDevs->{name}     = $1;
            $device            = $1;
            $hDevs->{vol_name} = $2;
            $hDevs->{id}       = $1;

        } elsif (   $line =~ /^\-\s+h(.+)\s+\d+\.\s+(\S+)\s+\<.+\>\s*(\S*).*$/
                 && $line !~ /drive not available|drive type unknown/i) {
            # Special condition to account for the case when there are more
            # output entries for format command than that fit in the page.
            # - hit space for more or s to select - \r
            # 11. c4t60A980004F644C774234435046657765d0
            #     <DEFAULT cyl 1020 alt 2 hd 64 sec 32>  <volname>
            $hDevs->{name}     = $2;
            $device            = $2;
            $hDevs->{vol_name} = $3;
            $hDevs->{id}       = $2;
        }

        if ($hDevs->{name} ne "") {
            $hDevs->{path} = "/dev/rdsk";

            # Check to see if the volname is set and if it is one of the ignore values
            if ($hDevs->{vol_name} =~ /root|boot|reserved/i) {
                $self->sanlog(-msg => "SolarisHost::_get_devices(): ignoring device (" . $hDevs->{name} . ") because vol name is " . $hDevs->{vol_name});
                next;
            }

            # Gather the device path, below is an exmaple path we are looking for
            # 3. c6t500A098187197D3Bd1 <NETAPP-LUN-7310 cyl 1022 alt 2 hd 16 sec 128>
            # /pci@8,700000/fibre-channel@2/fp@0,0/ssd@w500a098187197d3b,1
            $hDevs->{lun_path} = $output[$i+1];
            chomp($hDevs->{lun_path});

            if ($hDevs->{lun_path} eq "") {
                $self->sanlog(-msg => "SolarisHost::_get_devices(): ignoring device ( " . $hDevs->{name} . ") because it's lunpath is empty");
                next;
            } elsif ($hDevs->{lun_path} eq $boot_path and $boot_path ne "" and $hDevs->{vol_name} eq "") {
                # Lun paths matches boot path and vol name is NOT set for the lun
                $self->_set_err("Found boot lun (" . $hDevs->{name} . "), but vol name of lun is NOT set " .
                                "properly. Please set vol name to either \"boot\", \"root\", or \"reserved\" " .
                                "to have the automation ignore this lun");
                return (wantarray()) ? () : [];
            } elsif ($hDevs->{lun_path} eq $boot_path and $boot_path ne "") {
                # Lun paths matches boot path, skip this device
                $self->sanlog(-msg => "SolarisHost::_get_devices(): ignoring device ( " . $hDevs->{name} . ") because it's the boot path");
                next;
            }

            # Using format-inquiry command using remote file 'inquiry.format'
            # for obtaining Vendor & Product Ids
            $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format  -f $filename $device";
            ($status_1, @output_1) = &ClientUtil::execute($self->{conn}, $cmd);
            &SanUtil::nate_sleep(1);

            if (grep(/Operation on mounted disks must be interactive/i, @output_1)) {
                # The device is already mounted, we cannot perform the
                # operation on the device Skip it, goto the next device
                next;
            } elsif (grep(/Failed/, @output_1) && $status_1 == 1) {
                # We possibly have an orphaned device, ignore it
                $self->sanlog(-msg =>
                         "SolarisHost::_get_devices():  Bad status reurned by format-inquiry command, ignoring device ("
                         . $hDevs->{name}
                         . ")");
                next;
            } elsif ($status_1 > 0) {
                # format-inquiry doesn't work for internal disk, which is
                # generally first in the output list So, not setting error
                # flag here as otherwise the rest of the loop is not executed
                $self->sanlog(-msg => "SolarisHost::_get_devices():  Bad status reurned by format-inquiry command",
                              -msg_level => 3);
            } elsif ($status_1 == 0) {
                foreach my $myline (@output_1) {
                    if ($myline =~ /Vendor:\s*(\S*)/) {
                        if ($1 =~ /NETAPP/i) {
                            $hDevs->{type} = "NETAPP";
                        } elsif ($1 =~ /IBM/) {
                            foreach my $other_line (@output_1) {
                                if ($other_line =~ /Product:\s*(\S*)/) {
                                    if ($1 =~ /NSeries/i) {
                                        $hDevs->{type} = "IBM-NSeries";
                                    }
                                }
                            }
                        }
                    }
                }
            }
            push (@hDevices, $hDevs);
        }
    }

    $self->sanlog(-msg => scalar(@hDevices) .
                  " devices found (includes multiple paths to devices)");

    return (wantarray()) ? @hDevices : \@hDevices;

}

################################################################################
# Subroutine Name: _get_format_cmds
# Objective:       Get list of commands to format device
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_format_cmds {
    my ($self, $volname) = @_;

    my (@part_commands);

    # create an array with the commands to pass to format

    # write out the volume name and label
    push(@part_commands, "volname $volname");
    push(@part_commands, "label");

    # push(@part_commands, "verify");
    push(@part_commands, "quit");

    # return the list of commands to be executed.
    return (@part_commands);

}

################################################################################
# Subroutine Name: _get_inquiry_cmds
# Objective:       Calls the inquiry command for getting the vendor and
#                  product id.
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_inquiry_cmds {

    my ($self, $vol_name) = @_;

    # create an array with the commands to pass to format
    my (@part_commands);

    # Get the vendor and the product id from the device id
    push(@part_commands, "inquiry");
    push(@part_commands, "quit");

    # Return the list of commands to be executed.
    return (@part_commands);

}

################################################################################
# Subroutine Name: _get_format_cmds_raw
# Objective:       Get list of commands to format device and offset 1 cyl
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_format_cmds_raw {
    my ($self, $volname) = @_;

    my (@part_commands);

    # create an array with the commands to pass to format

    # write out the volume name and label
    # and create free space for s6 with offset 1 cylinder
    # to prevent trashing label
    push(@part_commands, "volname $volname");
    push(@part_commands, "label");

    # push(@part_commands, "verify");
    push(@part_commands, "quit");

    # return the list of commands to be executed.
    return (@part_commands);

}

###############################################################################
# Subroutine Name: _fdisk_luns
# Objective:       Runs fdisk command for x86 host machines
#                  before creating volumes
###############################################################################
sub _fdisk_luns {
    my ($self, @devices) = @_;

    my ($cmd, $status, @output);

    foreach my $device (@devices) {
        $cmd = "fdisk -B $device" . "p0";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Bad Status after fsdisk! -> $status");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _get_lpfc_conf
# Objective:       Get the content of the lpfc.conf file
###############################################################################
sub _get_lpfc_conf {
    my ($self) = @_;

    my ($cmd, @fcp_tags, $format, $status, @output, @target);

    # get the lpfc.conf content
    $cmd = "cat \/kernel\/drv\/lpfc.conf | grep :lpfc | grep -v '#'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Unable to read lpfc.conf file", @output);
        return (undef, undef);
    } elsif ($status == 1) {
        # Inform user if no no target entries are found.
        $self->sanlog(-msg => "Warning: lpfc.conf file has no target entries");
    }

    ### Check to make sure that no more than 1 lpfc config exists
    @fcp_tags = grep { /fcp-bind-/ } @output;

    if (scalar(@fcp_tags) > 1) {
        $self->_set_err("Multiple fcp-bind tags in configuration", @output);
        return (undef, undef);
    }

    foreach my $line (@output) {
        my ($target) = $line =~ /\"(\S+)\"/;
        push @target, $target;
    }

    $format = join("\n", @output);

    return ($format, @target);

}

###############################################################################
# Method:    _get_lpfc_format()
# Objective: retrieve lpfc.conf format for targets
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_lpfc_format {
    my ($self, $last, $target) = @_;

    my ($cnt, $lpfc_str, $targ, $time, @output);

    $time = strftime("%Y-%m-%d %H:%M:%S", localtime());

    ### Comment out the last lpfc string
    foreach $targ (@$last) {
        $cnt++;

        if ($cnt eq 1) {
            $lpfc_str .= sprintf("# fcp-bind-WWPN=\"%s\"%s\n",
                                 $targ, (scalar(@$last) > 1) ? "," : ";");
        } else {
            $lpfc_str .= sprintf("#%s\"%s\"%s\n", " " x 15, $targ,
                                 ($cnt < scalar(@$last)) ? "," : ";");
        }

    }

    ### Append the new lpfc string with time of modification
    $cnt = 0;
    foreach $targ (@$target) {
        $cnt++;

        if ($cnt eq 1) {
            $lpfc_str .= "\n# Updated by SanClient($time)\n";
            $lpfc_str .= sprintf(" fcp-bind-WWPN=\"%s\"%s\n",
                                 $targ, (scalar(@$target) > 1) ? "," : ";");
        } else {
            $lpfc_str .= sprintf("%s\"%s\"%s\n", " " x 15, $targ,
                                 ($cnt < scalar(@$target)) ? "," : ";");
        }
    }

    return $lpfc_str;

}

###############################################################################
# Subroutine Name: _get_sd_conf
# Objective:       Get the content of the sd.conf file
###############################################################################
sub _get_sd_conf {
    my ($self) = @_;

    my ($cmd, $status, @output);

    # get the sd.conf content
    $cmd = "cat /kernel/drv/sd.conf";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to cat sd.conf file", @output);
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a vol_name
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    my (@hDevices, @devices);

    ### Retrieve all host devices
    @hDevices = $self->_get_devices();
    return (wantarray()) ? () : [] if ($self->error);

    ### Determine which devices do not have
    ### a vol_name
    foreach my $device (@hDevices) {
        next if ($device->{type} !~ /NETAPP|IBM-Nseries/);
        next if ($device->{vol_name} =~ /root|boot|reserved/i);
        if ($device->{vol_name} eq "") {
            # Call _set_smi_label on each device deemed for use within automation
            $self->_set_smi_label($device->{name});
            if ($self->error()) {
                return (wantarray()) ? () : [];
            }
            push(@devices, $device->{name});
        } elsif ($device->{vol_name} =~ /NATE/) {
            $self->sanlog(-msg => "SolarisHost::_get_unlabled_devices() The device (" .
                                  $device->{name} . ") already had a label for NATE on the device\n" .
                                  "skipping the labeling process");
        } else {
            $self->sanlog(-msg => "SolarisHost::_get_unlabled_devices() The device (" .
                                  $device->{name} . ") has a label that is unknown to " .
                                  "this automation (" . $device->{vol_name} . ")");
        }
    }

    ### Format the list of devices with vol_name
    return (wantarray()) ? @devices : \@devices;

}

###############################################################################
# Method:    _recreate_sd_conf()
# Objective: recreate sd.conf file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _recreate_sd_conf {
    my ($self, %args) = @_;

    my ($line_start, $line_cnt, $mode, $buffer, $buflen, $bufsize);
    my ($file, $offset, $tot_lines, @buf, @chunk);

    ### Parameter 'bufsize' is the default number of lines for each
    ### file write.  It will be reduced if its length exceeds 'buflen'.
    ### Writes exceeding 2500 bytes are randomly corrupted by Nate
    ### communication mechanism.
    $bufsize = 60;
    $buflen  = 2000;

    $line_start = 0;
    $buffer     = $args{-buffer};
    $file       = $args{-file};

    $tot_lines = scalar(@{$buffer});

    while ($line_start <= $tot_lines) {
        ### Modify buffer size if its bigger
        ### than amount of actual lines to write
        if ($line_start + $bufsize > $tot_lines) {
            $bufsize = $tot_lines - $line_start + 1;
        }

        my $line_cnt = $line_start + $bufsize - 1;

        $offset = 0;

        while (1) {
            @chunk = ();

            push (@chunk,
                  join("\n", @{$buffer}[$line_start .. $line_cnt - $offset]));
            if (length "@chunk" > $buflen) {
                $offset += 5;
                next;
            }
            last;
        }

        $mode = ($line_start == 0) ? ">" : ">>";

        $self->sanlog(-msg       => "CHUNK:",
                      -msg_level => 9);
        $self->sanlog(-msg       => \@chunk,
                      -msg_level => 9,
                      -indent    => 4);

        @buf = $self->execute("echo \'@chunk\' $mode $file", -hide => 1);
        return $self if ($self->error);

        ### Advance the line_cnt to the next chunk
        ### to write
        $line_start += $bufsize - $offset;
    }

}

###############################################################################
# Subroutine Name: _update_drv
# Objective:       Update the driver
###############################################################################
sub _update_drv {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $cmd = "update_drv -f sd";
    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to reload sd driver", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _devfsadm
# Objective:       Update the driver
###############################################################################
sub _devfsadm {
    my ($self) = @_;

    my ($cmd, $status, @output);

    # execute devfsadm (as root)
    $cmd = "/usr/sbin/devfsadm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _devfsadm_cleanup
# Objective:       cleanup dangling /dev links
###############################################################################
sub _devfsadm_cleanup {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $cmd = "/usr/sbin/devfsadm -C";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _cfgadm_discover
# Objective:       Discover LUNs which are mapped to Solaris hosts using
#                  Solaris Native FCP drivers.
###############################################################################
sub _cfgadm_discover {
    my ($self) = @_;

    my ($cmd, $status, @output, @contAry);

    # use cfgadm -al to first discover if the initiators are logged into
    # the fabric.
    $cmd = "/usr/sbin/cfgadm -al";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");

        # this would mean that the host doesn't support cfgadm,
        return $ClientUtil::FALSE;
    }

    foreach my $ln (@output) {
        # in the cfgadm output look for controllers that are for fc-fabric
        # sample lines this regex is trying to parse:
        # "c1                scsi-bus     connected    configured   unknown"
        # "c1                scsi-bus     connected    unconfigured   unknown"
        if ($ln =~ /^(c\d+)\s+fc-fabric\s+\S+\s+unconfigured.*/) {
            if (length $1 > 1) {
                push @contAry, $1;
            } else {
                $self->_set_err("Length of ($1) is less than 1");
            }
        }
    }

    # The contAry will be empty if all of te controllers are already in a
    # configured state.  This will happen if this host has seen these
    # WWPNs before. So if the conAry is empty that is ok.
    foreach my $controller (@contAry) {
        # unconfigure and configure the fc-fabric controllers
        # so that they can discover luns on the traget.

        # First, do a blind confiugre to work around a problem described
        # in burt 276326
        $cmd = "/usr/sbin/cfgadm -c configure $controller";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($output[0] =~ /I\/O error/i) {
            $self->sanlog(-msg => "Working around burt 276326");
        }

        $cmd = "/usr/sbin/cfgadm -c unconfigure $controller";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status) {
            # if this fails, issue sanlog message of warning. Do not FAIL.
            $self->sanlog(-msg => "Warning: SolarisHost::_cfgadm_discover() Command ($cmd) returned a bad status ($status)");
        }

        $cmd = "/usr/sbin/cfgadm -c configure $controller";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status) {
            # if this fails, issue sanlog message of warning. Do not FAIL.
            $self->sanlog(-msg => "Warning: SolarisHost::_cfgadm_discover() Command ($cmd) returned a bad status ($status)");
        }
    }

    # execute devfsadm (as root)
    $cmd = "/usr/sbin/devfsadm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: discoverDevAluaFCP
# Objective:       Discover LUNs which are mapped to ALUA enabled Solaris10
#                  host. This routine is specific for FCP only and is a place
#                  holder untill Interop can provide official support for
#                  Solaris10 MPxIO.
###############################################################################
sub discoverDevAluaFCP {
    my ($self) = @_;

    my ($cmd, $status, @output, @contAry);

    # use cfgadm -al to first discover if the initiators are logged into
    # the fabric.
    $cmd = "/usr/sbin/cfgadm -al";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        # this would mean that the host doesnot support cfgadm,
        # which would mean a non-solaris10 host.
        return $ClientUtil::FALSE;
    }

    foreach my $ln (@output) {
        # in the cfgadm output look for controllers that are for fc-fabric
        if ($ln =~ /^(c\d+)\s+fc-fabric\s+.*/) {
            if (length $1 > 1) {
                push @contAry, $1;
            } else {
                $self->_set_err("Length of ($1) is less than 1");
            }
        }
    }

    if (not scalar(@contAry)) {
        $self->_set_err("fc-fabric Controller Array is empty");
        return $ClientUtil::FALSE;
    }

    foreach my $controller (@contAry) {
        # unconfigure and configure the fc-fabric controllers
        # so that they can discover luns on the traget.
        $cmd = "/usr/sbin/cfgadm -c unconfigure $controller";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # in the future we might start to care if the cmd has failures
        # for now just plow through....
        $cmd = "/usr/sbin/cfgadm -c configure $controller";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # in the future we might start to care if the cmd has failures
        # for now just plow through....

    }

    # execute luxadm
    $cmd = "/usr/sbin/luxadm probe";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Method:    _verify_lpfc_conf()
# Objective: verify/update the lpfc.conf with required target entries
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _verify_lpfc_conf {
    my ($self, %args) = @_;

    my ($change, $cmd, $line, $lpfc_current, $lpfc_new, $initName, $status);
    my ($total, $wwpn, %hLpfcTarg, @cmd, @current, @filers, @initObjs);
    my (@lpfc_current, @lpfc_new, %lpfc_merge, @targObjs, @output);

    $change = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?filer/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY") {
                foreach my $obj (@{$args{$_}}) {
                    unless (ref($obj) eq "SanFiler") {
                        $self->_set_err("Invalid value for argument '-filerObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @filers = @{$args{$_}};
            } else {
                $self->_set_err("Invalid value for argument '-filerObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        }
    }

    unless (scalar(@filers)) {
        $self->_set_err("Argument array is empty");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Retrieve current lpfc.conf targets");

    ### Retrieve current lpfc target format and array of
    ### target entries from config file
    ($lpfc_current, @lpfc_current) = $self->_get_lpfc_conf();
    return $self if ($self->error);

    ### Determine max targ id of current lpfc.conf
    foreach my $targId (@lpfc_current) {
        my ($name, $id) = $targId =~ /:(\S+)t(\d+)$/;
        $hLpfcTarg{$name} = ($id > $hLpfcTarg{$name}) ? $id : $hLpfcTarg{$name};
    }

    ### Retrieve new lpfc targets
    $self->sanlog(-msg => "Retrieve proposed lpfc.conf targets");
    foreach my $filerObj (@filers) {
        ### Retrieve target port objects on filer
        @targObjs = $filerObj->get_targets(-dmp => 1);
        if ($filerObj->error) {
            $self->_set_err($filerObj->errmsg);
            return $ClientUtil::FALSE;
        }

        ### Retrieve initiator port objects on client
        @initObjs = $self->get_initiators();
        return $ClientUtil::FALSE if ($self->error);

        ### initialize hash to hold unique Client Filer Target ids
        my %hctid;

        foreach my $initObj (@initObjs) {
            ### Retrieve initiator information
            $initName = $initObj->get_name();
            $wwpn     = $initObj->get_portname();

            ### Iterate over existing target object and check whether
            ### an lpfc conf entry already exists.
            foreach my $targObj (@targObjs) {
                my $target = undef;

                if (grep(/$wwpn/i, $targObj->get_initiator_wwpns())) {
                    ### Check to see if target already has a persistent
                    ### binding
                    my $portname = $targObj->get_portname();
                    my ($target) = grep /$portname/, @lpfc_current;

                    ### If entry is new then grab the next target id by
                    ### incrementing the current max target id
                    if (not defined $target) {
                        my $targId = (defined $hLpfcTarg{$initName}) ?
                                      $hLpfcTarg{$initName} + 1 : 0;

                        ### Keep track of target ids used for the filer
                        $hctid{$targId} = 1;

                        $target = $targObj->get_portname() . ":" . $initName .
                                  "t" . $targId;

                        $hLpfcTarg{$initName}++;
                    } else {
                        ### A target already exists.  Add this target
                        # to the hash map controller/target.
                        my ($name, $id) = $target =~ /:(\S+)t(\d+)$/;

                        $hctid{$id} = 1;
                    }

                    ### Add target entries to new lpfc array
                    push @lpfc_new, $target;
                }
            }
        }

        push my @cftid, (sort keys %hctid);
        $self->{cfmap}{$filerObj->get_name()} = [@cftid];
    }

    ### Copy content of current config to a hash with
    ### key being such that it can be sorted on by 'c*t*'
    foreach my $item (@lpfc_current) {
        my ($targ, $lpfc) = $item =~ /(\S+):(\S+)/;
        $lpfc_merge{$lpfc . $targ} = $item;
    }

    ### If any new lpfc entry is not in the current file
    ### than flag that a change has occurred
    foreach my $item (@lpfc_new) {
        if (not grep(/^$item$/i, @lpfc_current)) {
            $change = 1;
            ### add the new item to merge hash
            my ($targ, $lpfc) = $item =~ /(\S+):(\S+)/;
            $lpfc_merge{$lpfc . $targ} = $item;
        }
    }

    ### If lpfc needs a change
    if ($change == 1) {
        ### Display to the user the new bindings, as well as identify which
        ### ones are new.
        $self->sanlog(-msg       => "Changes to lpfc.conf targets",
                      -msg_level => 3);
        @lpfc_new = ();
        foreach my $item (sort keys %lpfc_merge) {
            my $value = $lpfc_merge{$item};
            my ($state) = (grep (/$value/, @lpfc_current)) ? "" : "*NEW*";
            push @lpfc_new, $value;

            $self->sanlog(-msg       => $value . " " . $state,
                          -msg_level => 3,
                          -indent    => 4);
        }

        $self->sanlog(" ");

        # set the reboot flag so users know to reboot the client
        $self->{reboot} = 1;
        $self->sanlog(-msg => "New Target entries have been added to the lpfc.conf.  A REBOOT",
                      "will be required to see these new entries\n\n");

        ### Create new lpfc bind string.  The string should comment out old
        ### and append new string.
        $lpfc_new = $self->_get_lpfc_format(\@lpfc_current, \@lpfc_new);

        ### If no bind string exists that append to end of lpfc.conf file
        if (scalar(@lpfc_current) eq 0) {
            @cmd = ("echo '$lpfc_new' >> \/kernel\/drv\/lpfc.conf");
        } else {
            ### Existing binding exists so replace with new
            @cmd = ("perl -e \' ",
                    "open (FH,\"<\/kernel\/drv\/lpfc.conf\") or exit(1); ",
                    "\$buffer.=\$_ while <FH>;",
                    "open (FD,\">\/kernel\/drv\/lpfc.conf\") or exit(1); ",
                    "\$buffer=~" . "s/" . $lpfc_current . "/" .
                    $lpfc_new . "/;",
                    "print FD \$buffer;",
                    "close FD;",
                    "close FH;",
                    "\' ");
        }

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, "@cmd");
        if ($status > 0) {
            $self->_set_err("Unable to modify lpfc.conf file", @output);
            return (wantarray()) ? () : [];
        }

        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Configuration has not changed");
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Method:    _verify_sd_conf()
# Objective: verify/update the sd.conf with required sd entries
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _verify_sd_conf {
    my ($self, %args) = @_;

    my ($change, $cmd, $detail, $status, $sd, %hsh, %hfilerlun, @buf);
    my (@detail, @filerObjs, @header, @ids,    @lunObjs, @output, @tmp);

    ### Initialize
    $change = 0, $detail = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?lunObj/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY") {
                foreach my $obj (@{$args{$_}}) {
                    unless (ref($obj) eq "Lun") {
                        $self->_set_err("Invalid value for argument '-lunObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @lunObjs = @{$args{$_}};
            } else {
                $self->_set_err("Invalid value for argument '-initObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        } elsif (/-?filer/i) {
            if (length($args{$_}) and ref($args{$_}) eq "ARRAY") {
                foreach my $obj (@{$args{$_}}) {
                    unless (ref($obj) eq "SanFiler") {
                        $self->_set_err("Invalid value for argument '-filerObjs'");
                        return $ClientUtil::FALSE;
                    }
                }

                @filerObjs = @{$args{$_}};
            } else {
                $self->_set_err("Invalid value for argument '-filerObjs'");
                return $ClientUtil::FALSE;
            }

            next;
        } elsif (/-?sd/i) {
            if (not length($sd = $args{$_})) {
                $self->_set_err("Invalid value for argument '-sdExtraEntries'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not scalar(@lunObjs)) {
        $self->sanlog(-msg => "Info: SolarisHost::_verify_sd_conf() Verification of sd.conf is ",
                      "not required since no new lun entries were created");
        return $ClientUtil::TRUE;
    }

    unless (scalar(@filerObjs)) {
        $self->_set_err("Required -FilerObj' array is empty");
        return $ClientUtil::FALSE;
    }

    ### Retrieve content of sd.conf file
    @buf = $self->_get_sd_conf();

    ### Create a hash of 'existing' targets and lun numbers
    foreach my $line (@buf) {
        if ($line =~ /name.*parent.*target.*lun.*/) {
            my ($target, $lun) = $line =~ /target=(\d+)\s+lun=(\d+)/;
            $hsh{$target}{$lun} = 1;
            $detail = 1;
        }

        push @header, "$line" unless ($detail);

    }

    ### Seperate luns by what filer they below too.
    if (@lunObjs) {
        $self->sanlog(-msg       => "lun total=" . scalar(@lunObjs),
                      -msg_level => 9);

        foreach my $filerObj (@filerObjs) {
            my $filername = $filerObj->get_name();
            $self->sanlog(-msg       => "Filer: " . $filername,
                          -msg_level => 9);
            my @lunids;
            foreach my $lunObj (@lunObjs) {
                $self->sanlog(-msg       => "lun filername = " .
                                            $lunObj->get_filername(),
                              -indent    => 4,
                              -msg_level => 9);
                if ($lunObj->get_filername() =~ /$filername/i) {
                    $self->sanlog(-msg       => "Lun: " . $lunObj->get_name() .
                                                " Id: " . $lunObj->get_id(),
                                  -indent    => 8,
                                  -msg_level => 9);
                    push @lunids, $lunObj->get_id();
                }

            }
            $hfilerlun{$filername}{cfmap} = $self->{cfmap}{$filername};
            $hfilerlun{$filername}{luns}  = \@lunids;

        }
    }

    ### Add new lun ids to existing hash
    foreach my $filername (keys %hfilerlun) {
        # Determine the max lunid for the set of
        # luns created.
        my $max = 0;

        ### Determine if any of the sd entries are new.
        foreach my $id (@{$hfilerlun{$filername}{luns}}) {
            foreach my $target (@{$hfilerlun{$filername}{cfmap}}) {
                $max = ($id > $max) ? $id : $max;
                next if (exists $hsh{$target}{$id});
                $hsh{$target}{$id} = 0;
                $change = 1;
            }
        }

        ### Add additional '$sd' entries to the sd.conf so a
        # a reboot is nor required when new luns are created.
        if (defined $sd and length($sd)) {
            $self->sanlog(-msg => "Adding $sd additional scsi device entries to ",
                          "each filer '$filername' target");
            foreach my $target (@{$hfilerlun{$filername}{cfmap}}) {
                foreach my $entry (1 .. $sd) {
                    next if (exists $hsh{$target}{$max + $entry});
                    $hsh{$target}{$max + $entry} = 0;
                    $change = 1;
                }
            }
        }
    }

    ### Recreate the sd.conf file if new sd entries are required
    if ($change) {
        my @target = ();
        foreach my $target (sort { ($a <=> $b) } keys %hsh) {
            foreach my $lun (sort { ($a <=> $b) } keys %{$hsh{$target}}) {
                push @target, "name=\"sd\" parent=\"lpfc\" target=$target lun=$lun;";
            }
        }

        # Create copy of existing sd.conf
        $cmd = "cp /kernel/drv/sd.conf /kernel/drv/sd.conf.bck";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to modify sd.conf file", @output);
            return $ClientUtil::FALSE;
        }

        # Recreate the sd.conf file
        $self->_recreate_sd_conf(-buffer => [@header, @target],
                                 -file   => "/kernel/drv/sd.conf",);
        return $ClientUtil::FALSE if ($self->error);

        if ($self->os_version() =~ /5.9/) {
            $self->_update_drv();
            return $ClientUtil::FALSE if ($self->error);

            $self->sanlog(-msg => "Reloading device drivers");
            $self->_devfsadm();
            return $ClientUtil::FALSE if ($self->error);
        } else {
            $self->{reboot} = 1;
            $self->sanlog(-msg => "Additional entries were added to sd.conf - A REBOOT will",
                          "be required to see all additional luns");
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsiadm to include the correct targets for the
#                  filers. Support for dynamic (send-targets), static
#                  and isns discovery.
#                  Assumptions: a targets list is passed into this function with
#                  the IP's of all the ports on each filer.
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    my ($status, @list_output, @output, $cmd, @targets);

    # Get the alua support
    my $alua_support = $self->get_alua_support();
    if (not defined($alua_support) or $alua_support eq "") {
        if ($self->error) {
            $self->_set_err("Unable to determine the ALUA support information:\n", $self->errmsg());
        } else {
            $self->_set_err("Unable to determine the ALUA support information");
        }
        return $ClientUtil::FALSE;
    }

    if ($self->{revision} > 5.10 || ($self->{revision} == 5.10 && $self->{minor_revision} >= 9)) {
        # We only want to lower the conn-login-max to 60 for ALUA configs
        # since we assume that means Cluster-Mode which has secondary paths.
        # For 7-mode non-alua configs we don't have secondary paths so we don't
        # want the session to go offline so we need to make sure the conn-login-max
        # is set to the default of 180.
        my $conn_login_max = "";
        if ($alua_support) {
            $conn_login_max = "60";
        } else {
            $conn_login_max = "180";
        }

        # set the maximum retry timeout for 10U9 or later to 60.
        $cmd = "iscsiadm modify initiator-node -T conn-login-max=" . $conn_login_max;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("commamd ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # check if the user wants to skip iscsi discovery
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} != 1) {
        # Ignore _iscsi_configure, we are asked to NOT configure anything
        $self->sanlog(-msg => "iSCSI Configure: Bypassing iSCSI Configuration by user request (CONFIG_ISCSI_DISCOVERY=0)");
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Invalid value for argument '-targets'");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI");

    # if requested to do subnet matching, filter the targets list
    # so that it contains targets that are accessible by this host
    @targets = $self->_filter_targets_by_subnet(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check what ip version is being requested and set the ip mode on all the target objects
    if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /ipv4|ipv6|auto/i) {
        foreach my $target (@targets) {
            # set the ip mode to be either ipv4 or ipv6
            $target->set_ip_mode($self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
            if ($target->error()) {
                $self->_set_err("Could not set the ip mode of target $target->{nodename} to " .
                                 $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n" .
                                "Reason: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    } else {
       $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: " .
                        $self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
       return $ClientUtil::FALSE
    }

    # update the targets key in the %args hash to be the new value of @targets
    # because we pass %args down into other subroutines to process.
    if (exists $args{-targets}) {
        $args{-targets} = \@targets;
    }
    if (exists $args{targets}) {
        $args{targets} = \@targets;
    }

    # determine the initiator nodename
    $cmd = "iscsiadm list initiator-node";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }
    my $nodename;
    foreach my $line (@output) {
        if ($line =~ /Initiator node name: (\S+)/i) {
            $nodename = $1;
            last;
        }
    }

    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} eq "") {
        # default the chap username (out username) to be the initiator nodename
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} = $nodename;
    }

    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} eq "") {
        # default the chap username (in username) to be the initiator nodename
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} = $nodename
    }

    # print out the summary of all the changes we will make for iscsi configuration
    $self->_iscsi_configure_print_summary(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure iscsi if specified
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} > 0) {
        $self->_iscsi_deconfigure(-targets => \@targets,
                                  -clean_level => $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # set up the iscsi alias for this host
    $cmd = "iscsiadm modify initiator-node --node-alias " . $self->hostname();
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # configure targets for discovery
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /sendtargets/i) {
        # sendtargets discovery
        $self->_iscsi_configure_sendtargets_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /static/i) {
        # static discovery
        $self->_iscsi_configure_static_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /isns/i) {
        # isns discovery
        $self->_iscsi_configure_isns_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # at this point, perform some final validation to make sure that we have all
    # our targets properly set up
    $cmd = "iscsiadm list target";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $target (@targets) {
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        if (not grep(/Target:\s*${nodename}\s*$/, @output)) {
            $self->_set_err("The target ($nodename) is not present in the initiator's target list");
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "iSCSI Configure: iSCSI Configuration complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_print_summary()
# Objective:       print out a summary of what iSCSI configuration will be done
################################################################################
sub _iscsi_configure_print_summary {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my @target_ips = map { "\n\t" . $_->{ip} } @targets;
    my @config_summary;

    push(@config_summary, "iSCSI Configure: Configuring iSCSI with the following options:\n");
    push(@config_summary, "Targets : @target_ips\n");
    push(@config_summary, "IP Version            : " . $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n");
    push(@config_summary, "iSCSI Type            : " . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "\n");
    push(@config_summary, "iSCSI Discovery Type  : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "\n");
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
        push(@config_summary, "iSNS Server           : " . $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} . "\n");
    }
    push(@config_summary, "iSCSI Pre-Clean Level : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} . "\n");
    push(@config_summary, "iSCSI Subnet Matching : " . $self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} . "\n");
    push(@config_summary, "CHAP Mode             : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} . "\n");

    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
        push(@config_summary, "CHAP Username         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} . "\n");
        push(@config_summary, "CHAP Password         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD} . "\n");
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
            push(@config_summary, "CHAP Out Username     : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} . "\n");
            push(@config_summary, "CHAP Mutual Secret    : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} . "\n");
        }
    }

    push(@config_summary, "Header Digest         : " . $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} . "\n");
    push(@config_summary, "Data Digest           : " . $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} . "\n");

    $self->sanlog(-msg => "@config_summary");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigure iSCSI Options, iSCSI (static,sendtargets,isns)
#                  via iscsiadm.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI settings");

    # deconfigure iscsi options, this might have to go before removing targets
    $self->_iscsi_deconfigure_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure sendtargets
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /sendtargets/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_sendtargets_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # deconfigure static
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /static/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_static_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # deconfigure isns
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /isns/i or $clean_level == 2) {
        $self->_iscsi_deconfigure_isns_discovery(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_options()
# Objective:       Configure iSCSI options (CHAP, Digests)
################################################################################
sub _iscsi_configure_options {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI Options");

    # if doing CHAP, enable chap on the initiator and set up the in password and user name
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
        # enable chap on the initiator
        $cmd = "iscsiadm modify initiator-node --authentication CHAP";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("commamd ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }

        # establish the IN NAME for CHAP (defaults to iqn of host) {CONFIG_ISCSI_CHAP_USERNAME}
        $cmd = "iscsiadm modify initiator-node --CHAP-name " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("commamd ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }

        # establish the IN PASSWORD for CHAP {CONFIG_ISCSI_CHAP_IN_PASSWORD},
        # validate that the password is 12 chars
        if (length($self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD}) != 12) {
            $self->_set_err("CONFIG_ISCSI_CHAP_IN_PASSWORD must be 12 characters");
            return $ClientUtil::FALSE;
        }

        # set up the chap password for the initiator
        $self->_iscsi_configure_options_chap_pw("iscsiadm modify initiator-node --CHAP-secret",
                                                $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD});
        if ($self->error()) {
            return $self;
        }
    }

    # foreach unique target nodename (need to know static/send/isns targets),
    # enable CHAP on the target, set up the out password on the
    # target and set up the out username (iqn of initiator)
    # enable header/data digest if needed
    foreach my $target (@targets) {
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # enable CHAP for the target iqn for both uni-directional and bi-directional chap
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
            $cmd = "iscsiadm modify target-param --authentication CHAP " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # if doing bi-directional chap, set up the outpass/outusername
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
            # enable BI-CHAP for the target iqn
            $cmd = "iscsiadm modify target-param --bi-directional-authentication enable " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

            # set up the OUT NAME for the target
            # defaults to iqn of host.
            $cmd = "iscsiadm modify target-param --CHAP-name " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} . " " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

            # set up OUT PASSWORD for CHAP {CONFIG_ISCSI_CHAP_OUT_PASSWORD}
            # validate that the password is 12 chars
            if (length($self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD}) != 12) {
                $self->_set_err("CONFIG_ISCSI_CHAP_OUT_PASSWORD must be 12 characters");
                return $ClientUtil::FALSE;
            }

            # Set up the chap password for this target
            $self->_iscsi_configure_options_chap_pw("iscsiadm modify target-param --CHAP-secret " . $nodename,
                                                    $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD});
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }

        } else {
            # otherwise we are not doing bi directional chap, so disable it for this target iqn
            $cmd = "iscsiadm modify target-param --bi-directional-authentication disable " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # if doing header digest then enable it
        if ($self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} == 1) {
            $cmd = "iscsiadm modify target-param --headerdigest CRC32 " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

        }

        # if doing data digest then enable it
        if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1) {
            $cmd = "iscsiadm modify target-param --datadigest CRC32 " . $nodename;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("commamd ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_options()
# Objective:       Deconfigure iSCSI Options (CHAP, Digests)
################################################################################
sub _iscsi_deconfigure_options {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @list_output);

    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets) and $clean_level == 1) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI Options");

    # list all the targets
    $cmd = "iscsiadm list target";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # if clean_level == 1 (soft approach, only affect the targets in @targets)
    # disable chap/header/data digest for each unique target in the targets list that we are working with
    if ($clean_level == 1) {
        foreach my $target (@targets) {
            my $nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # skip this target if it is not on the host currently
            if (not grep(/Target:\s+${nodename}/, @list_output)) {
                next;
            }

            # disable chap for this target
            $cmd = "iscsiadm modify target-param --authentication none $nodename";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

            # disable bi-directional chap
            $cmd = "iscsiadm modify target-param --bi-directional-authentication disable $nodename";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

            # disable header digest
            $cmd = "iscsiadm modify target-param --headerdigest none $nodename";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }

            # disable data digest
            $cmd = "iscsiadm modify target-param --datadigest none $nodename";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    # if clean_level == 2 (hammer approach, affect all targets)
    elsif ($clean_level == 2) {

        # disable chap/header/data digest for all targets visible on the system
        foreach my $line (@list_output) {
            if ($line =~ /Target: (\S+)/i) {
                my $nodename = $1;

                # disable chap for this target
                $cmd = "iscsiadm modify target-param --authentication none $nodename";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }

                # disable bi-directional chap
                $cmd = "iscsiadm modify target-param --bi-directional-authentication disable $nodename";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }

                # disable header digest for this target
                $cmd = "iscsiadm modify target-param --headerdigest none $nodename";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }

                # disable data digest for this target
                $cmd = "iscsiadm modify target-param --datadigest none $nodename";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # Disable initiator-side CHAP
        $cmd = "iscsiadm modify initiator-node --authentication none";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }

        # disable header digest on the initiator
        $cmd = "iscsiadm modify initiator-node --headerdigest none";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }

        # disable data digest on the initiator
        $cmd = "iscsiadm modify initiator-node --datadigest none";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_options_chap_pw()
# Objective:       Helper function to handle the password entry for iSCSI CHAP
################################################################################
sub _iscsi_configure_options_chap_pw {
    my ($self, $iscsiadm_cmd, $secret) = @_;

    # pull out the connection object that is currently being used.
    my $conn = $self->get_connection();
    my $disconnect_when_done = 0;

    # if the connection type that is currently being used is not ssh or telnet,
    # then open one here:
    if ($conn->conntype() !~ /^ssh$|^telnet$|^ssh-persist$/i) {

        # change the connection retries value
        my $old_retries = $self->{connection_retries};
        $self->{connection_retries} = 2;

        # first try ssh
        $conn = $self->_connect($self->hostname(), undef, undef, undef,
                                "ssh");
        if ($self->error()) {
            # try telnet if ssh failed.
            $self->_clear_err();
            $conn = $self->_connect($self->hostname(), undef, undef, undef,
                                    "telnet");
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        }

        # restore the old value
        $self->{connection_retries} = $old_retries;

        # disconnect the connection we made when we are done.
        $disconnect_when_done = 1;
    } else {
        # otherwise, reuse the current connection and dont disconnect it later on
        $disconnect_when_done = 0;
    }

    # issue the iscsiadm command and answer the prompt password questions with
    # the specified secret
    my (undef, undef, undef, $recv_prompt) = &Tharn::say($conn, $iscsiadm_cmd,
                                                         -prompt, "Enter secret:",
                                                         -timeout, 120);
    if ($recv_prompt =~ /Enter secret:/) {
        (undef ,undef, undef, $recv_prompt) = &Tharn::say($conn, $secret,
                                                          -prompt, "Re-enter secret:",
                                                          -timeout, 120);
        # re-enter the secret
        if ($recv_prompt =~ /Re-enter secret:/) {
            (undef, undef, undef, $recv_prompt) = &Tharn::say($conn, $secret,
                                                              -timeout, 120);
        } else {
            $self->_set_err("Failed to set CHAP secret: $secret");
            &Tharn::disconnect($conn) if ($disconnect_when_done);
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Failed to set CHAP secret: $secret");
        &Tharn::disconnect($conn) if ($disconnect_when_done);
        return $ClientUtil::FALSE;
    }

    # disconnect the connection
    &Tharn::disconnect($conn) if ($disconnect_when_done);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_sendtargets_discovery()
# Objective:       Configure sendtargets discovery by adding all the
#                  targets listed in @targets.
################################################################################
sub _iscsi_configure_sendtargets_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, @list_output);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Sendtargets' Discovery");

    # get config to see if we are already defined
    $cmd = "iscsiadm list discovery-address";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $target (@targets) {
        # check if the discovery address for this target was already added.
        # If it was not added, then add. If it is, then skip.
        my $dynamic_addr = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if ipv6, modify the address to be in brackets
        if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /ipv6/i) {
            $dynamic_addr = "[" . $dynamic_addr . "]";
        }

        if (grep(/$dynamic_addr/i, @list_output) == 0) {
            $self->sanlog(-msg => "Adding iSCSI discovery (send-targets) address $dynamic_addr");
            $cmd = "iscsiadm add discovery-address " . $dynamic_addr;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "Skipping add of iSCSI discovery (send-targets) address $dynamic_addr (already exists)");
        }
    }

    # configure all iscsi options
    $self->_iscsi_configure_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # enable sendtargets discovery
    $cmd = "iscsiadm modify discovery --sendtargets enable";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # list out the sendtargets addresses that were added
    $cmd = "iscsiadm list discovery-address";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # display what discovery mechanism is enabled.
    $cmd = "iscsiadm list discovery";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # validate that the discovery mode is set to what we want.
    if (not grep(/Send Targets: enabled/i, @output)) {
        $self->_set_err("Send Targets was not enabled correctly.");
        return $ClientUtil::FALSE;
    }


    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_sendtargets_discovery()
# Objective:       Deconfigure sendtargets discovery by removing the specified
#                  targets.
################################################################################
sub _iscsi_deconfigure_sendtargets_discovery {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @sendtargets_list);

    my $clean_level = 1;
    my $already_rebooted = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Sendtargets' Discovery");

    # list out all the sendtargets discovery targets
    $cmd = "iscsiadm list discovery-address";
    ($status, @sendtargets_list) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # if clean_level is 1, then clean up only the targets we know about
    if ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        foreach my $target (@targets) {
            # check to see if the ip address for this target is in the list
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the target ip is ipv6, then add brackets around it.
            if ($ip =~ /:/) {
                $ip = "[" . $ip . "]";
            }

            my $sendtargets_addr = $ip . ":3260";
            if ($target->error()) {
                $self->_set_err("Could not get ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if this target is not in the list, then skip the removal
            if (not grep(/$sendtargets_addr/, @sendtargets_list)) {
                next;
            }

            # remove this target from iscsiadm
            $cmd = "iscsiadm remove discovery-address $sendtargets_addr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                    $already_rebooted = 1;
                    redo;
                } elsif (grep(/logical unit in use/i, @output)) {
                    $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                    "Please consult the Solaris Host Team or Netapp Proven Rep.");
                    return $ClientUtil::FALSE;
                } else {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    # if clean_level is 2, then clean up all targets we see.
    # try to disable discovery
    elsif ($clean_level == 2) {
        foreach my $line (@sendtargets_list) {
            if ($line =~ /Discovery Address:\s*(\S+):\d*/i) {
                my $sendtargets_addr = $1;
                # if ipv6, modify the address to be in brackets
                if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /ipv6/i) {
                    $sendtargets_addr = "[" . $sendtargets_addr . "]";
                }
                # remove this static target from iscsiadm
                $cmd = "iscsiadm remove discovery-address $sendtargets_addr";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                        $self->reboot();
                        if ($self->error()) {
                            return $ClientUtil::FALSE;
                        }
                        $already_rebooted = 1;
                        redo;
                    } elsif (grep(/logical unit in use/i, @output)) {
                        $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                        "Please consult the Solaris Host Team or Netapp Proven Rep.");
                        return $ClientUtil::FALSE;
                    } else {
                        $self->_set_err("Command ($cmd) returned a status of ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }

        # try and disable sendtargets discovery, reboot the system if we detect luns are in use
        while (1) {
            $cmd = "iscsiadm modify discovery --sendtargets disable";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                    $already_rebooted = 1;
                    redo;
                } elsif (grep(/logical unit in use/i, @output)) {
                    $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                    "Please consult the Solaris Host Team or Netapp Proven Rep.");
                    return $ClientUtil::FALSE;
                } else {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
            last;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_static_discovery()
# Objective:       Configure static discovery by adding the specified targets
#                  in @targets
################################################################################
sub _iscsi_configure_static_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, @list_output);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Static' Discovery");

    # list the current static configuration
    $cmd = "iscsiadm list static-config";
    ($status, @list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $target (@targets) {
        # define static address for this target and check to see if
        # it was already added. If not added, add, if added, skip.
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $tpgt = $target->get_tpgt();
        if ($target->error()) {
            $self->_set_err("Could not get tpgt from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # if the target ip is ipv6, then add brackets around it.
        if ($ip =~ /:/) {
            $ip = "[" . $ip . "]";
        }

        my $static_addr = $nodename . "," .
                          $ip . ":" . "3260" . "," .
                          $tpgt;
        if (grep(/$static_addr/i, @list_output) == 0) {
            $self->sanlog(-msg => "Adding iSCSI static address $static_addr");
            $cmd = "iscsiadm add static-config " . $static_addr;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "Skipping add of iSCSI static address $static_addr (already exists)");
        }
    }

    # configure all iscsi options
    $self->_iscsi_configure_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # enable static discovery
    $cmd = "iscsiadm modify discovery --static enable";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # list out the static configuration addresses
    $cmd = "iscsiadm list static-config";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # display what discovery mechanism is enabled.
    $cmd = "iscsiadm list discovery";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # validate that the discovery mode is set to what we want.
    if (not grep(/Static: enabled/i, @output)) {
        $self->_set_err("Static was not enabled correctly.");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_static_discovery()
# Objective:       Deconfigure static discovery by removing the targets listed
#                  in the targets array
################################################################################
sub _iscsi_deconfigure_static_discovery {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets, @static_list);

    my $clean_level = 1;
    my $already_rebooted = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Static' Discovery");

    # list out all the static targets
    $cmd = "iscsiadm list static-config";
    ($status, @static_list) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # if clean_level is 1, then clean up only the targets we know about
    if ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        foreach my $target (@targets) {
            my $nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
            my $ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
            my $tpgt = $target->get_tpgt();
            if ($target->error()) {
                $self->_set_err("Could not get tpgt from target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the target ip is ipv6, then add brackets around it.
            if ($ip =~ /:/) {
                $ip = "[" . $ip . "]";
            }

            my $static_addr = $nodename . "," .
                              $ip . ":" . "3260" . "," .
                              $tpgt;

            # check to see if this ip target is in the list,
            # if it is not, then skip the target
            if (not grep(/$static_addr/, @static_list)) {
                next;
            }

            # remove static targets from iscsiadm
            $cmd = "iscsiadm remove static-config $static_addr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                    $already_rebooted = 1;
                    redo;
                } elsif (grep(/logical unit in use/i, @output)) {
                    $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                    "Please consult the Solaris Host Team or Netapp Proven Rep.");
                    return $ClientUtil::FALSE;
                } else {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    # if clean_level is 2, then clean up all targets we see.
    # try to disable static but dont fail if it doesnt work
    elsif ($clean_level == 2) {
        foreach my $line (@static_list) {
            if ($line =~ /Static Configuration Target:\s*(\S+)/i) {
                my $static_addr = $1;

                # remove this static target from iscsiadm
                $cmd = "iscsiadm remove static-config $static_addr";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                        $self->reboot();
                        if ($self->error()) {
                            return $ClientUtil::FALSE;
                        }
                        $already_rebooted = 1;
                        redo;
                    } elsif (grep(/logical unit in use/i, @output)) {
                        $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                        "Please consult the Solaris Host Team or Netapp Proven Rep. \n" .
                                        "If you are not configuring with veritas, you may want to verify that veritas is disabled.");
                        return $ClientUtil::FALSE;
                    } else {
                        $self->_set_err("Command ($cmd) returned a status of ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }

        # try and disable static discovery, it may error because it can fail if LUNs
        # are mapped to the static iscsi initiator, reboot if this is the case.
        while (1) {
            $cmd = "iscsiadm modify discovery --static disable";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                    $already_rebooted = 1;
                    redo;
                } elsif (grep(/logical unit in use/i, @output)) {
                    $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                    "Please consult the Solaris Host Team or Netapp Proven Rep.");
                    return $ClientUtil::FALSE;
                } else {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
            last;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_isns_discovery()
# Objective:       Configure discovery with iSNS.
################################################################################
sub _iscsi_configure_isns_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, @list_output);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'iSNS' Discovery");

    # list out the isns servers
    $cmd = "iscsiadm list isns-server";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # if the isns server is not added added, then add it
    my $isns_server_ip = $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER};
    if (not grep(/iSNS Server IP Address:\s*$isns_server_ip/i, @output)) {
        # add the isns server
        $cmd = "iscsiadm add isns-server $isns_server_ip";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # configure all iscsi options
    $self->_iscsi_configure_options(%args);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # enable isns
    $cmd = "iscsiadm modify discovery --iSNS enable";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # list out the isns server addresses
    $cmd = "iscsiadm list isns-server";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # display what discovery mechanism is enabled.
    $cmd = "iscsiadm list discovery";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("commamd ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    # validate that the discovery mode is set to what we want.
    if (not grep(/iSNS: enabled/i, @output)) {
        $self->_set_err("iSNS was not enabled correctly.");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_isns_discovery()
# Objective:       Deconfigure iSNS Discovery by removing the iSNS Server.
################################################################################
sub _iscsi_deconfigure_isns_discovery {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @isns_list);

    my $clean_level = 1;
    my $disable_isns = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSNS Discovery");

    my $isns_server = $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER};

    # list out all the static targets
    $cmd = "iscsiadm list isns-server";
    ($status, @isns_list) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    my $already_rebooted = 0;

    # if clean_level is 1, then clean up only the isns servers we know about
    if ($clean_level == 1) {
        # if the isns server specified is there, remove it
        if (grep(/iSNS Server IP Address: ${isns_server}\:/, @isns_list)) {
            # perform one loop so that we can perform redos
            for (1..1) {
                # remove the isns server we know of
                $cmd = "iscsiadm remove isns-server $isns_server";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                        $self->reboot();
                        if ($self->error()) {
                            return $ClientUtil::FALSE;
                        }
                        $already_rebooted = 1;
                        redo;
                    } elsif (grep(/logical unit in use/i, @output)) {
                        $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                        "Please consult the Solaris Host Team or Netapp Proven Rep.");
                        return $ClientUtil::FALSE;
                    } else {
                        $self->_set_err("Command ($cmd) returned a status of ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            }
            # need to disable isns
            $disable_isns = 1;
        }
    }

    # if clean_level is 2, then clean up all isns servers we see.
    # try to disable isns but dont fail if it doesnt work
    elsif ($clean_level == 2) {
        foreach my $line (@isns_list) {
            if ($line =~ /iSNS Server IP Address:\s*(\S+)/i) {
                my $isns_addr = $1;

                # need to do one loop so the redo command works
                for (1..1) {
                    # remove this isns server from iscsiadm
                    $cmd = "iscsiadm remove isns-server $isns_addr";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0) {
                        if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                            $self->reboot();
                            if ($self->error()) {
                                return $ClientUtil::FALSE;
                            }
                            $already_rebooted = 1;
                            redo;
                        } elsif (grep(/logical unit in use/i, @output)) {
                            $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                            "Please consult the Solaris Host Team or Netapp Proven Rep.");
                            return $ClientUtil::FALSE;
                        } else {
                            $self->_set_err("Command ($cmd) returned a status of ($status)");
                            return $ClientUtil::FALSE;
                        }
                    }
                }
            }
        }
        $disable_isns = 1;
    }

    if ($disable_isns == 1) {
        # perform one loop so the redo command works in this for loop
        for (1..1) {
            # try and disable isns discovery, ignoring error because it can fail if LUNs
            # are mapped to the iscsi initiator
            $cmd = "iscsiadm modify discovery --iSNS disable";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                if (grep(/logical unit in use/i, @output) and $already_rebooted == 0) {
                    $self->reboot();
                    if ($self->error()) {
                        return $ClientUtil::FALSE;
                    }
                    $already_rebooted = 1;
                    redo;
                } elsif (grep(/logical unit in use/i, @output)) {
                    $self->_set_err("Command ($cmd) failed even after rebooting the system. " .
                                    "Please consult the Solaris Host Team or Netapp Proven Rep.");
                    return $ClientUtil::FALSE;
                } else {
                    $self->_set_err("Command ($cmd) returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_host_config()
# Objective:       Run the host_config script on the Solaris client
# Details:         See POD documentation at the end of this file
################################################################################
sub _set_host_config {
    my ($self, %args) = @_;

    my ($status, @output, $cmd);

    # Get the alua support info
    my $alua_support = $self->get_alua_support();
    if (not defined($alua_support) or $alua_support eq "") {
        if ($self->error) {
            $self->_set_err("Unable to determine the ALUA support information:\n", $self->errmsg());
        } else {
            $self->_set_err("Unable to determine the ALUA support information");
        }
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    my $path = $self->get_hu_bin_path();

    $cmd = $path . "/host_config -setup -noprompt";
    if ($self->{protocol} =~ /fcp/i and $self->{protocol} =~ /iscsi/i) {
        $cmd .= " -protocol mixed";
    }
    elsif ($self->{protocol} =~ /fcp/i) {
        $cmd .= " -protocol fcp";
    } elsif ($self->{protocol} =~ /iscsi/i){
        $cmd .= " -protocol iscsi";
    } else {
        $self->_set_err("Unable to run host_config. protocol is an unsupport type:".$self->{protocol});
        return $ClientUtil::FALSE;
    }

    if ($self->{mp}->{mp_type} =~ /mpxio/i) {
        $cmd .= " -multipath mpxio";
    } elsif ($self->{mp}->{mp_type} =~ /dmp/i){
        $cmd .= " -multipath dmp";
    } elsif ($self->{mp}->{mp_type} =~ /none/i){
        $cmd .= " -multipath none";
    } else {
        $self->_set_err("Unable to run host_config. mp_type is an unsupport type:".$self->{mp}->{mp_type});
        return $ClientUtil::FALSE;
    }

    unless ($alua_support) {
        $cmd .= " -noalua";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Unable to run host_config. Command ($cmd)",
                        "returned a status of ($status)");
        return $ClientUtil::FALSE;
    }

    my @host_config_output = @output;

    # handle stmsboot. Setup, but do not reboot, will be hanlded later
    my $filename = "/tmp/stmsboot_input.txt";
    my @stms_cmds = ( "y", "n" );
    &ClientUtil::create_remote_file($self->{conn}, $filename, @stms_cmds);

    # Parse the output of host_config to see what commands we need to run.
    # We skip the reboot command because it will be done later.
    # Example output:
    # ####################################################################
    # The following lines will be ADDED to the /kernel/drv/sd.conf file
    # ####################################################################
    #
    # sd-config-list="NETAPP  LUN", "netapp-sd-config";
    # netapp-sd-config=1,0x9c01,8,0,0,0,0,0,0,0,0,0,300,30,30,0,0,2,0,0;
    #
    #
    #
    # The original version of file '/kernel/drv/sd.conf' has been saved
    # to file '/kernel/drv/sd.conf.1344429896'
    #
    #
    # No changes are needed for file '/kernel/drv/scsi_vhci.conf'
    #
    # #################################################################
    # To complete the configuration, please run the following commands:
    # #################################################################
    #
    # /usr/sbin/stmsboot -D fp -d     (Do not reboot if prompted)
    # /usr/sbin/stmsboot -D iscsi -d  (Do not reboot if prompted)
    # /usr/sbin/shutdown -y -g 0 -i 6
    foreach my $line (@host_config_output) {
        # Remove a known string that is postpended
        # to the lines containing the stmsboot commands we need to run.
        $line =~ s/\(Do not reboot if prompted\)//;
        $line =~ s/^\s+|\s+$//g;

        if($line =~ /stmsboot/) {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $line . " < $filename");
            if ($status != 0) {
                $self->_set_err("Unable to run command to finish after host_config. ".
                                "Command ($cmd)","returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    &ClientUtil::remove_remote_file($self->{conn}, $filename);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_basic_config()
# Objective:       Run the basic_config script on the Solaris client to set
#                  lpfc.conf
# Details:         See POD documentation at the end of this file
# Notes:           The filer CFMODE needs to be provided, or it will default to
#                  Single Image mode.
################################################################################
sub _set_basic_config {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, $cfo_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not length($cfo_mode = $args{$_})) {
                $self->_set_err("Missing value for argument '-cfmode'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    if (not defined($cfo_mode)) {
        $self->sanlog(-msg => "Defaulting cfo_mode to 'single_image'");
        $cfo_mode = "single_image";
    }

    # Get the alua support info
    my $alua_support = $self->get_alua_support();
    if (not defined($alua_support) or $alua_support eq "") {
        if ($self->error) {
            $self->_set_err("Unable to determine the ALUA support information:\n", $self->errmsg());
        } else {
            $self->_set_err("Unable to determine the ALUA support information");
        }
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    my $path = $self->get_hu_bin_path();

    # Run basic_config based on the protocol requirements
    $self->sanlog(-msg => "Checking the protocol (PROTOCOL) and driver type (CONFIG_FCP_DRIVER_TYPE) for basic_config");

    # For ISCSI, we need to have alua support enabled
    if ($self->{protocol} =~ /iscsi/i and $alua_support == 1) {
        $self->sanlog(-msg => "Performing ISCSI ALUA basic_config operations");

        # If we are running iscsi with alua support for cmode, then we need
        # to also call basic_config with the -sd_set for all x86 configs and
        # non-mpxio configs.  For sparc mpxio we would call basic_config -ssd_set.
        if ($self->{arch} =~ /86/i) {
            $cmd = "yes|" . $path . "/basic_config -sd_set";
        } else {
            if ($self->{mp}->{mp_type} =~ /mpxio/i) {
                $cmd = "yes|" . $path . "/basic_config -ssd_set";
            } else {
                $cmd = "yes|" . $path . "/basic_config -sd_set";
            }
        }

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("basic_config command ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # For Veritas DMP with Native solaris drivers, use -vxdmp_set.
    # For Native drivers used with mpxio use -ssd_set for sparc and -sd_set for x86
    if ($self->{protocol} =~ /fcp/i) {
        $self->sanlog(-msg => "Performing FCP basic_config operations");
        if ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /native/i) {
            if ($self->{mp}->{mp_type} =~ /dmp/i) {
                $cmd = "yes|" . $path . "/basic_config -vxdmp_set";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Unable to run basic_config. Command ($cmd)",
                                    "returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            } else {  # mp type is not "dmp". Use -ssd_set for all others, including mpxio
                if ($self->{arch} =~ /86/i) {
                    $cmd = "yes|" . $path . "/basic_config -sd_set";
                } else {
                    $cmd = "yes|" . $path . "/basic_config -ssd_set";
                }
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Unable to run basic_config. Command ($cmd) ",
                                    "returned a status of ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        } elsif ($self->{fcp}->{CONFIG_FCP_DRIVER_TYPE} =~ /vendor/i) {
            # Native driver is not in use. Must be lpfc driver (until qla
                # driver is supported)
            $cmd = "yes yes|" . $path . "/basic_config -m $cfo_mode";
            ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a status of ($status)");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Basic Config was not run. Be sure the driver type " .
                            "(CONFIG_FCP_DRIVER_TYPE) is set for fcp configs or " .
                            "(ALUA_SUPPORT) is set for iscsi configs");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_bindings()
# Objective:       Generate LPFC persistent bindings in lpfc.conf file.
#                  Use the create_bindings.pl script from HU for this.
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_bindings {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);
    my ($script_name, $target, $filer_name);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => %args{$_}");
            }
            next;
        }
    }

    # grab the first filer from the targets list
    $target = $targets[0];
    $filer_name = $target->{filer_name};
    my $path = $self->get_hu_bin_path();

    # Get the connection type
    my $conn = &SanUtil::get_device_connection(-resource_name => $filer_name);
    if (not defined($conn) or $conn eq "") {
        $self->_set_err("Unable to determine connection info for device ($filer_name)");
        return $ClientUtil::FALSE;
    }

    # Get the username and password
    my ($username, $password) = &SanUtil::get_login_credentials(-resource_name => $filer_name,
                                                                -connection_type => $conn);

    # We need to see if we got a username back.  Depending on the connection type, we may
    # not get back a username or password. If we dont' get a username back, use "root" as the default.
    if (not defined($username) or $username eq "") {
        $username = "root";
    }

    # hack alert - do ls of directory for create_binding.pl and if
    # found, use it, otherwise it is a later HU and the script is
    # create_binding (no .pl extension) compiled for active state
    $cmd = "ls $path" . "/create_binding.pl";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $script_name = "create_binding";
    } else {
        $script_name = "create_binding.pl";
    }

    # formulate and run command with determined script name.  The
    # command will differ depending on whether or not a password
    # is set in the tharnhost file.
    if (defined($password) and $password ne "") {
        # We have a password set
        $cmd = "yes | " . $path . "/$script_name -d -n $filer_name -l ${username}:${password}";
    } else {
        # We don't have a password set
        $cmd = "yes | " . $path . "/$script_name -d -n $filer_name -l ${username}";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a status of ($status). Check login credentials on controllers");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _obtain_memory_info()
# Objective:       Store total memory and page size of this system on the object
################################################################################
sub _obtain_memory_info {
    my ($self) = @_;
    my ($cmd, $status, @output);

    # store the total memory of the system here
    $cmd = "/usr/sbin/prtconf | grep Memory";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    foreach my $line (@output) {
        # parse out the memory size and convert to bytes
        if ($line =~ /Memory size:\s+(\d+)\s+(M|K|G|T)?/i) {
            $self->{memory_size_bytes} = &ClientUtil::convert_to_bytes($1 . $2);
            last;
        }
    }

    # check if we were able to obtain the size of the memory
    if (not defined($self->{memory_size_bytes})) {
        $self->_set_err("Failed to obtain memory size of host.");
        return $self;
    }

    # obtain the page size, should only be one lined entry in the output
    $cmd = "/usr/bin/pagesize";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status > 0 or scalar(@output) != 1) {
        $self->_set_err("Failed to obtain the page size");
        return $self;
    }

    # store the page size
    $self->{memory_page_size} = $output[0];
}

################################################################################
# Subroutine Name: _get_boot_path()
# Objective:       Go out to the system and gather the boot path
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_boot_path {
    my ($self) = @_;

    my ($boot_path, $cmd, $status, @output);

    # Run prtconf command and gather boot path
    $cmd = "prtconf -pv |grep bootpath";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);

    # Cleanup output
    chomp(@output);

    # Check status > 1 because we may NOT find a bootpath on this machine
    if ($status > 1) {
        $self->_set_err("Command ($cmd) returned a status of ($status)");
        return "";
    } elsif ($output[0] =~ /^\s+bootpath\:\s+\'(\S+)\:\S+\s*$/) {
        # Exmaple output of command
        # bootpath:  '/pci@8,600000/SUNW,qlc@4/fp@0,0/disk@w21000004cf6221bd,0:a'
        $boot_path = $1;
    } else {
        $self->_set_err("Unable to determine bootpath from cmd ($cmd) " .
                        "with output of: " . join("\n", @output));
        return "";
    }

    # Return found boot path
    return $boot_path;
}

################################################################################
# Subroutine Name: _set_smi_label
# Objective:       Set smi label type on device
# Details:         See POD documentation at the end of this file
################################################################################
sub _set_smi_label {
    my ($self, $device) = @_;

    my ($filename, $cmd, $status, @output, @format_cmds);

    $filename = "/tmp/set_smi_label.format";

    # We need to check for x86 vs sparc and set the the format_cmds array
    # based on the architecture.  We also need to run fdisk if it is x86.
    if ($self->{arch} =~ /86/i) {
        # Looks like we are x86. Add $device to a temp array so we can pass it
        # to _fdisk_luns and call _fdisk_luns.
        my @temp_devices = ($device);
        $self->_fdisk_luns(@temp_devices);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }

        # Now put a regular label on the disk since by recreating the fdisk
        # partition, the type was put back to SMI.
        @format_cmds = ("type", "0", "label");
        $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format -f $filename $device";
    } else {
        # Looks like we are on a sparc box so execute the normal format command
        @format_cmds = ("type", "0", "label", "0", "no", "yes");
        $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format -e -f $filename $device";
    }

    # Create temporary file with @format_cmds
    &ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # execute format (as root)
   ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
   if (scalar(@output) == 0) {
       $self->_set_err("No output from format command (possible timeout)");
       return $ClientUtil::FALSE;
   }

    # Remove temporary file
    &ClientUtil::remove_remote_file($self->{conn}, $filename);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _determine_os_version
# Objective:       execute "uname -p" and "uname -r" on device
# Details:         See POD documentation at the end of this file
################################################################################
sub _determine_os_version {
    my ($self) = @_;
    my $cmd = "uname -p";
    my ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    foreach my $line (@output) {
        if ($line =~ /86|sparc/i) {
            $self->{arch} = "$line";
            $self->sanlog(-msg => "Solaris Host Architecture is: $line.");
        }
    }

    # store client OS revision
    $cmd = "uname -r";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    $self->{revision} = $output[0];

    # get Solaris 10 minor revision - note this is the update
    # level w/o the 'u' in front.
    if ($self->{revision} =~ /5\.(9|(1[0-1]))/) { # 5.9 | 5.10 | 5.11
        my $minor_version = "NOT_SET";
        # Get major and minor revision
        $cmd = "cat /etc/release";
        ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
        foreach my $line (@output) {
            # Looking for something like one of the following lines:
            # Solaris 9 9/05 HW s9s_u9wos_06b SPARC
            # Solaris 10 10/09 s10s_u8wos_08a SPARC
            # Solaris 10 5/09 s10x_u7wos_08 X86
            # Oracle Solaris 10 8/11 s10s_u10wos_17b SPARC
            # Oracle Solaris 10 8/11 s10x_u10wos_15 X86
            # Oracle Solaris 11 11/11 SPARC
            # Oracle Solaris 11 11/11 X86
            # Oracle Solaris 11.1 SPARC
            # Oracle Solaris 11.1 X86
            #
            if ($line =~ /oracle solaris (\d+) (\d+\/\d+)\s+(sparc|x86)/i) {
                $minor_version = "0";
                last;
            } elsif ($line =~ /oracle solaris (\d+)\.(\d+)\s+(sparc|x86)/i) {
                $minor_version = $2;
                last;
            } elsif ($line =~ /solaris (\d+) \S+ s.*_u(\d+)/i) {
                $minor_version = $2;
                last;
            } elsif ($line =~ /solaris (\d+) \S+ \S+ s.*_u(\d+)/i) { # This accounts for the Solaris 9 9/05 HW s9s_u9wos_06b SPARC
                $minor_version = $2;
                last;
            }
        }
        $self->{minor_revision} = $minor_version;
    }

    $self->sanlog(-msg =>
        "Major Revision: $self->{revision}, Minor Revision: $self->{minor_revision}");

    # Setting up test path for i/o depending on the Host architecture
    if ($self->{arch} =~ /86/i) {
        $self->{TEST_BIN} = "/x/eng/localtest/arch/i386-sun-solaris8/bin";
    } elsif ($self->{arch} =~ /sparc/i) {
        $self->{TEST_BIN} = "/x/eng/localtest/arch/sparc-sun-solaris8/bin";
    } else {
        $self->_set_err("Invalid parameter specified for -arch");
        return $self;
    }
}

1;

=pod

=head1 Package Name

SolarisHost.pm

=head1 Author

Gary Knopp

=head1 Description

For description of methods see B<San::SanClient::Host.pod>.

=head1 Public Methods

=head2 B<get_iscsi_info()>

Collects all known iscsi information from the client.  Packages the information
into a complex data structure.

 Builds a hash of target nodenames.
 Under each target nodename, there is a hash for each iscsi cmd:
  alias      - contains scalar value
  connection - contains an array of hashes
  initiator  - contains an array of hashes
  interface  - contains an array of hashes
  nodename   - contains scalar
  security   - contains a hash
  session    - contains an array of hashes
  show       - contains a hash
  status     - contains a hash
  tpgroup    - contains an array of hashes
 Each hash contains keys which correspond to the c-mode cmd options

=over 4

=item Inputs

=item * None

=item Outputs

=item * $iscsi_hash - hash of client iscsi info

=item * empty hash (do not set_err)  - no sessions or target portals

=item * empty hash (set_err) - error condition

=item * Sample output:

 $VAR1 = {
           'iqn.1992-08.com.netapp:sn.c66ed9fa02d411e0955a123478563412:vs.6' => {
                                                                                  'connection' => [
                                                                                                    {
                                                                                                      'has-session' => 'true',
                                                                                                      'remote-ip-port' => '52260',
                                                                                                      'header-digest-enabled' => 'false',
                                                                                                      'target-mrdsl' => '65536',
                                                                                                      'authentication-method' => 'none',
                                                                                                      'local-ip-port' => '3260',
                                                                                                      'connection-id' => '0',
                                                                                                      'tpgroup-tag' => '1027',
                                                                                                      'remote-address' => '10.228.131.247',
                                                                                                      'connection-state' => 'Full_Feature_Phase',
                                                                                                      'local-address' => '10.228.130.125',
                                                                                                      'data-digest-enabled' => 'false'
                                                                                                    }

                                                                                                  ],
                                                                                  'show' => {
                                                                                              'status-admin' => 'up',
                                                                                              'target-name' => 'iqn.1992-08.com.netapp:sn.c66ed9fa02d411e0955a123478563412:vs.6',
                                                                                              'target-alias' => 'sol_vserver'
                                                                                            },
                                                                                  'nodename' => 'iqn.1992-08.com.netapp:sn.c66ed9fa02d411e0955a123478563412:vs.6',
                                                                                  'status' => {
                                                                                                'status-admin' => 'up',
                                                                                                'is-available' => 'true'
                                                                                              },
                                                                                  'discovery_type' => 'sendtargets',
                                                                                  'initiator' => [
                                                                                                   {
                                                                                                     'initiator-name' => 'iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a',
                                                                                                     'tpgroup-tag' => '1027',
                                                                                                     'isid' => '40:00:00:2a:00:00',
                                                                                                     'initiator-alias' => 'ssan-240v-2'
                                                                                                   }
                                                                                                 ],
                                                                                  'tpgroup' => [
                                                                                                 {
                                                                                                   'tpgroup-tag' => '1027',
                                                                                                   'lif' => [
                                                                                                              '10.228.130.125'
                                                                                                            ]
                                                                                                 }
                                                                                               ],
                                                                                  'session' => [
                                                                                                 {
                                                                                                   'max-burst-length' => '65536',
                                                                                                   'isid' => '40:00:00:2a:00:00',
                                                                                                   'initiator-alias' => 'ssan-240v-2',
                                                                                                   'immediate-data-enabled' => 'true',
                                                                                                   'default-time-to-wait' => '2',
                                                                                                   'max-connections' => '1',
                                                                                                   'tpgroup-tag' => '1027',
                                                                                                   'error-recovery-level' => '0',
                                                                                                   'connection-ids' => '0',
                                                                                                   'initial-r2t-enabled' => 'true',
                                                                                                   'data-pdu-in-order' => 'true',
                                                                                                   'session-type' => 'Normal',
                                                                                                   'initiator-name' => 'iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a',
                                                                                                   'data-sequence-in-order' => 'true',
                                                                                                   'first-burst-length' => '65536',
                                                                                                   'default-time-to-retain' => '20',
                                                                                                   'max-outstanding-r2t' => '1'
                                                                                                 }

                                                                                               ],
                                                                                  'security' => {
                                                                                                  'auth-type' => 'none',
                                                                                                  'initiator-name' => 'iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a',
                                                                                                  'outbound-user-name' => '-',
                                                                                                  'user-name' => '-'
                                                                                                },
                                                                                  'interface' => [
                                                                                                   {
                                                                                                     'is-home' => 'true',
                                                                                                     'tpgroup-tag' => '1027',
                                                                                                     'address' => '10.228.130.125',
                                                                                                     'enabled' => 'true',
                                                                                                     'ip-port' => '3260'
                                                                                                   }
                                                                                                 ],
                                                                                  'alias' => 'sol_vserver'
                                                                                }
         };

=item Examples

    my $iscsi_info = $Client_Obj->get_iscsi_info();

=back

=cut

=cut
