#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/SanFiler/Volume.pm#2 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to  a Filer Volume
## @author knopp@netapp.com
## @status private
## @pod here

# Package Name
##############
package Volume;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;
use Data::Dumper;

use San::SanClient::ClientUtil;
use San::SanFiler::Volume::Lun;
use San::SanFiler::Volume::Qtree;
use San::SanLog;
use San::SanUtil;
use Tie::IxHash;
use Time::HiRes qw(gettimeofday);
use Date::Parse;
use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

###############################################################################
# Method:    autosize()
# Objective: set state, max, increment for volume autosize policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub autosize {
    my ($self, %args) = @_;
    my (@output, $cmd, $state, $max, $increment,);

    $self->_clear_err();

    # default state to on
    $state = "on";

    # parse arguments
    foreach (keys %args) {
        if (/-?state/) {
            if (not length($state = $args{$_}) or $args{$_} !~ /on|off|reset/) {
                $self->_set_err(  "Error: Volume::autosize() "
                                . "Invalid value for argument '-state'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?max/) {
            if (not length($max = $args{$_}) or $args{$_} !~ /\d+.*/) {
                $self->_set_err(  "Error: Volume::autosize() "
                                . "Invalid value for argument '-max'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?increment/) {
            if (not length($increment = $args{$_}) or $args{$_} !~ /\d+.*/) {
                $self->_set_err(  "Error: Volume::autosize() "
                                . "Invalid value for argument '-increment'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(  "Error: Volume::autosize() "
                            . "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # check for required arguments
    unless (defined $state and length $state) {
        $self->_set_err(  "Error: Volume::autosize() "
                        . "Missing required argument '-state'");
        return $ClientUtil::FALSE;
    }

    $cmd    = "vol autosize " . $self->get_name() . " ";
    $cmd    = $cmd . "-m " . $max . " " if (defined($max));
    $cmd    = $cmd . "-i " . $increment . " " if (defined($increment));
    $cmd    = $cmd . $state;
    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless ($output[0] =~ /(updated|reset)/i) {
        my ($errmsg) = $output[0];
        $self->_set_err("Error: Volume::autosize() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end autosize()

################################################################################
# Subroutine Name: clear_sdlock()
# Objective:       Removes snapdrive lock file.
# Details:         See POD documentation at the end of this file
################################################################################
sub clear_sdlock {
    my ($self) = @_;
    my ($cmd, $name, @output,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "priv set advanced; rm /vol/" . $name . "/.snapdrive_lock";

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end clear_sdlock()

################################################################################
# Subroutine Name:    enable_sis()
# Objective:          Enables SIS
# Details:            See POD documentation at the end of this file
################################################################################
sub enable_sis {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(-msg       => "Volume::enable_sis() Enable SIS",
                            -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "zapi-cli") {
        my @args;
        push(@args, "sis");
        push(@args, "on");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } else {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-enable",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name);
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(
             "Error: Volume::enable_sis(): Unable to " . "enable SIS: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(  "Error: Volume::enable_sis() "
                            . "unable to enable SIS $errmsg");
            return $ClientUtil::FALSE;
        }
    } else {
        if ($output !~ /enabled/i) {
            $self->_set_err(  "Error: Volume::enable_sis() unable to "
                            . "enable SIS $output");
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Error: Volume::enable_sis() " . "result check failed");
        return $ClientUtil::FALSE;
    }
}    # end enable_sis

################################################################################
# Subroutine Name:    disable_sis()
# Objective:          Disables SIS
# Details:            See POD documentation at the end of this file
################################################################################
sub disable_sis {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(-msg       => "Volume::disable_sis() Disable SIS",
                            -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "zapi-cli") {
        my @args;
        push(@args, "sis");
        push(@args, "off");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } else {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-disable",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name);
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: Volume::disable_sis(): Unable to "
                        . "disable SIS: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(
                      "Error: Volume::disable_sis() " . "unable to disable SIS",
                      $errmsg);
            return $ClientUtil::FALSE;
        }
    } else {
        if ($output !~ /disabled/i) {
            $self->_set_err(
                      "Error: Volume::disable_sis() unable to " . "disable SIS",
                      $output);
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err(
                      "Error: Volume::disable_sis() " . "results check failed");
        return $ClientUtil::FALSE;
    }
}    # end disable_sis

################################################################################
# Subroutine Name:    config_sis()
# Objective:          Configures SIS
# Details:            See POD documentation at the end of this file
################################################################################
sub config_sis {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;
    my $schedule;
    my $compression;
    my $inline_compression;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(-msg => "Volume::config_sis() " . "Configure SIS",
                            -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    #
    # Parse args hash
    #
    foreach (keys %args) {
        if (/-?schedule/) {
            length($schedule = $args{$_})
              or do {
                $self->_set_err("Volume::config_sis() "
                                  . "invalid value for "
                                  . "schedule: ",
                                "$args{$_}");
                return $self;
              };
        } elsif (/-?compression/) {
            length($compression = $args{$_});
        } elsif (/-?inline_compression/) {
            length($inline_compression = $args{$_});
        }
    } ## end foreach (keys %args)

    unless (defined $schedule and length $schedule) {
        $self->_set_err(
            "Error: Volume::config_sis() " . "schedule is a required argument");
        return $ClientUtil::FALSE;
    }

    $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd                         => "sis-set-config",
                            -cmd_filer                   => $self->{filer},
                            -check_result                => 1,
                            -cmd_interface               => "zapi",
                            -path                        => "/vol/" . $name,
                            -schedule                    => $schedule,
                            -'enable-compression'        => $compression,
                            -'enable-inline-compression' => $inline_compression
                            );

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: Volume::config_sis(): Unable to "
                        . "configure SIS: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    my $output_hash = $self->{parent}->xml2hash($output);
    my $status      = $output_hash->{status};

    if ($status eq "failed") {
        my $errmsg = $output_hash->{reason};
        $self->_set_err(
            "Error: Volume::config_sis() " . "unable to configure SIS $errmsg");
        return $ClientUtil::FALSE;
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err(
                       "Error: Volume::config_sis() " . "results check failed");
        return $ClientUtil::FALSE;
    }
}    # end config_sis

################################################################################
# Subroutine Name:    get_sis_config()
# Objective:          Gets SIS configuration information
# Details:            See POD documentation at the end of this file
################################################################################
sub get_sis_config {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(
                  -msg => "Volume::get_sis_config() Get " . "SIS configuration",
                  -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "zapi-cli") {
        my @args;
        push(@args, "sis");
        push(@args, "config");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } else {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-set-config",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name);
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: Volume::get_sis_config(): Unable to "
                        . "get SIS config: $errmsg");
        return "";
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(  "Error: Volume::get_sis_config() "
                            . "unable to get SIS configuration $errmsg");
            return "";
        }
    } else {
        if (   $output =~ /Invalid path/i
            or $output =~ /^WARNING: SIS is not licensed.*/) {
            $self->_set_err(  "Error: Volume::get_sis_config() unable to "
                            . "get SIS configuration $output");
            return "";
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $output;
    } else {
        $self->_set_err(
                   "Error: Volume::get_sis_config() " . "results check failed");
        return "";
    }
}    # end get_sis_config

################################################################################
# Subroutine Name:    start_sis()
# Objective:          Starts SIS
# Details:            See POD documentation at the end of this file
################################################################################
sub start_sis {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(-msg       => "Volume::start_sis() Start SIS",
                            -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "zapi-cli") {
        my @args;
        push(@args, "sis");
        push(@args, "start");
        push(@args, "-s");
        push(@args, "-f");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } elsif ($self->{parent}->interface() eq "cli") {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-start",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name,
                                -force        => "",
                                -scan         => "");
    } else {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-start",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name,
                                -scan         => "true");
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(
               "Error: Volume::start_sis(): Unable to " . "start SIS: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(
                          "Error: Volume::start_sis() " . "unable to start SIS",
                          $errmsg);
            return $ClientUtil::FALSE;
        }
    } else {
        if ($output !~ /started/i) {
            $self->_set_err(
                          "Error: Volume::start_sis() unable to " . "start SIS",
                          $output);
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Error: Volume::start_sis() " . "results check failed");
        return $ClientUtil::FALSE;
    }
}    # end start_sis

################################################################################
# Subroutine Name:    stop_sis()
# Objective:          Stops SIS
# Details:            See POD documentation at the end of this file
################################################################################
sub stop_sis {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(-msg       => "Volume::stop_sis() Stop SIS",
                            -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "zapi-cli") {
        my @args;
        push(@args, "sis");
        push(@args, "stop");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } else {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-stop",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name);
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(
                 "Error: Volume::stop_sis(): Unable to " . "stop SIS: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err("Error: Volume::stop_sis() " . "unable to stop SIS",
                            $errmsg);
            return $ClientUtil::FALSE;
        }
    } ## end if ($self->{parent}->{...

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Error: Volume::stop_sis() " . "results check failed");
        return $ClientUtil::FALSE;
    }
}    # end stop_sis

################################################################################
# Subroutine Name:    get_sis_status()
# Objective:          Gets SIS status
# Details:            See POD documentation at the end of this file
################################################################################
sub get_sis_status {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    $name = $self->{name};

    $self->{parent}->sanlog(
                         -msg => "Volume::get_sis_status() " . "Get SIS status",
                         -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    if ($self->{parent}->interface() eq "cli") {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-status",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name,
                                -verbose      => "");
    } elsif ($self->{parent}->interface() eq "zapi") {
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "sis-status",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -path         => "/vol/" . $name,
                                -verbose      => "true");
    } else {
        my @args;
        push(@args, "sis");
        push(@args, "status");
        push(@args, "-l");
        push(@args, "/vol/" . $name);

        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd           => "system-cli",
                                -cmd_filer     => $self->{filer},
                                -cmd_interface => "zapi",
                                -check_result  => 1,
                                -args          => \@args,
                                -priv          => "admin");
    } ## end else [ if ($self->{parent}->interface...

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: Volume::get_sis_status(): Unable to "
                        . "get SIS status: $errmsg");
        return "";
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err("Error: Volume::get_sis_status() "
                              . "unable to get SIS status",
                            $errmsg);
            return "";
        }

        # 'last operation error' is not available in zapi, cli only (RR.0)
        # 'checkpoint-*' elements are not available in cli, zapi only (RR.0)
        $output = $output_hash->{'sis-object'}->{'dense-status'};
    } else {
        if (   ($output =~ /Invalid path/i)
            or ($output =~ /^WARNING: SIS is not licensed.*/)
            or ($output =~ /No status entry found/i)) {
            $self->_set_err("Error: Volume::get_sis_status() unable to "
                              . "get SIS status",
                            $output);
            return "";
        }
        my @result = split(/\n/, $output);
        $output = undef;

        # Path:                            /vol/NdvmSrcVol1
        # State:                           Enabled
        # Compression:                     Disabled
        # Inline Compression:              Disabled
        # Status:                          Idle
        # Progress:                        Idle for -
        # Type:                            Regular
        # Schedule:                        sun-sat@0
        # Minimum Blocks Shared:           1
        # Blocks Skipped Sharing:          0
        # Last Operation State:            Success
        # Last Successful Operation Begin: Wed Aug 17 11:40:18 GMT 2011
        # Last Successful Operation End:   Wed Aug 17 11:40:18 GMT 2011
        # Last Operation Begin:            Wed Aug 17 11:40:18 GMT 2011
        # Last Operation End:              Wed Aug 17 11:40:18 GMT 2011
        # Last Operation Size:             0 KB
        # Last Operation Error:            -
        # Change Log Usage:                0%
        # Logical Data:                    192 KB/69 TB (0%)
        # Queued Job:                      -
        # Stale Fingerprints:              0%

        my %multiplier_hash =
          ('B' => 0, 'K' => 1, 'M' => 2, 'G' => 3, 'T' => 4);
        foreach my $line (@result) {
            if ($line =~ /path:\s+(.+)/i) {
                $output->{'path'} = $1;
            } elsif ($line =~ /^state:\s+(.+)/i) {
                $output->{'state'} = $1;
            } elsif ($line =~ /^compression:\s+(.+)/i) {
                $output->{'is-compression-enabled'} =
                  ($1 =~ /enabled/i) ? "true" : "false";
            } elsif ($line =~ /^inline compression:\s+(.+)/i) {
                $output->{'is-inline-compression-enabled'} =
                  ($1 =~ /enabled/i) ? "true" : "false";
            } elsif ($line =~ /status:\s+(.+)/i) {
                $output->{'status'} = $1;
            } elsif ($line =~ /progress:\s+(.+)/i) {
                $output->{'progress'} = $1;
            } elsif ($line =~ /type:\s+(.+)/i) {
                $output->{'type'} = $1;
            } elsif ($line =~ /schedule:\s+(.+)/i) {
                $output->{'schedule'} = $1;
            } elsif ($line =~ /minimum blocks shared:\s+(.+)/i) {
                $output->{'minimum-blocks-shared'} = $1;
            } elsif ($line =~ /blocks skipped sharing:\s+(.+)/i) {
                $output->{'blocks-skipped-sharing'} = $1;
            } elsif ($line =~ /last successful operation begin:\s+(.+)/i) {
                $output->{'last-successful-operation-begin-timestamp'} =
                  str2time($1);
            } elsif ($line =~ /last successful operation end:\s+(.+)/i) {
                $output->{'last-successful-operation-end-timestamp'} =
                  str2time($1);
            } elsif ($line =~ /last operation begin:\s+(.+)/i) {
                $output->{'last-operation-begin-timestamp'} = str2time($1);
            } elsif ($line =~ /last operation end:\s+(.+)/i) {
                $output->{'last-operation-end-timestamp'} = str2time($1);
            } elsif ($line =~ /last operation size:\s+(\S+) (.)/i) {
                $output->{'last-operation-size-bytes'} =
                  $1 * 1024**$multiplier_hash{$2};

                # Since, cli gives only approximated output, the value of
                # 'last-operation-size-bytes' obtained here may not be
                # same as the value obtained from zapi
            } elsif ($line =~ /last operation state:\s+(.+)/i) {
                $output->{'last-operation-state'} = $1;
            } elsif ($line =~ /change log usage:\s+(\d+)/i) {
                $output->{'change-log-usage-percent'} = $1;
            } elsif ($line =~ /queued job:\s+(.+)/i) {
                $output->{'queued-job-type'} = $1;
            } elsif ($line =~ /stale fingerprints:\s+(\d+)/i) {
                $output->{'stale-fingerprint-percent'} = $1;
            } elsif ($line =~ /logical data:\s+(\d+) (.).*\/(\d+) (.).*(\d+)/i)
            {
                $output->{'logical-data'}->{'sis-logical-data'}
                  ->{'logical-data-size'} = $1 * 1024**$multiplier_hash{$2};
                $output->{'logical-data'}->{'sis-logical-data'}
                  ->{'logical-data-limit'} = $3 * 1024**$multiplier_hash{$4};
                $output->{'logical-data'}->{'sis-logical-data'}
                  ->{'logical-data-percent'} = $5;

                # 'logical-data-percent' is not available in zapi
            } elsif ($line =~ /last operation error:\s+(.+)/i) {

                # 'last operation error' is not available in zapi, cli only (RR.0)
                $output->{'last-operation-error'} = $1;
            }
        } ## end foreach my $line (@result)
    } ## end else [ if ($self->{parent}->{...

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $output;
    } else {
        $self->_set_err(
                   "Error: Volume::get_sis_status() " . "results check failed");
        return "";
    }
}    # end get_sis_status

################################################################################
# Subroutine Name:    start_compression()
# Objective:          Starts compression on this volume
# Details:            See POD documentation at the end of this file
################################################################################
sub start_compression {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    my $delete_checkpoint       = 0;
    my $include_shared_blocks   = 0;
    my $include_snapshot_blocks = 0;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?delete_checkpoint/) {
            if (not length($delete_checkpoint = $args{$_})) {
                $self->_set_err(
                             "Invalid value for argument '-delete_checkpoint'");
                return $self;
            }
            next;
        } elsif (/-?include_shared_blocks/) {
            if (not length($include_shared_blocks = $args{$_})) {
                $self->_set_err(
                         "Invalid value for argument '-include_shared_blocks'");
                return $self;
            }
            next;
        } elsif (/-?include_snapshot_blocks/) {
            if (not length($include_snapshot_blocks = $args{$_})) {
                $self->_set_err(
                       "Invalid value for argument '-include_snapshot_blocks'");
                return $self;
            }
            next;
        }
    } ## end foreach (keys %args)

    # get the volume name
    $name = $self->{name};

    $self->{parent}->sanlog(
                        -msg => "Volume::start_compression() Start Compression",
                        -msg_level => 1);

    if ($self->{parent}->interface() eq "zapi") {

        # issue the volume compress start zapi with the passed in options
        $resultsHash_ref =
          $cmd_obj->execute_cmd(
                         -cmd                       => "volume-compress-start",
                         -cmd_filer                 => $self->{filer},
                         -check_result              => 1,
                         '-delete-checkpoint'       => $delete_checkpoint,
                         '-include-shared-blocks'   => $include_shared_blocks,
                         '-include-snapshot-blocks' => $include_snapshot_blocks,
                         -volume                    => $name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to start compression: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}    # end start_compression

################################################################################
# Subroutine Name:    stop_compression()
# Objective:          Stops compression on this volume
# Details:            See POD documentation at the end of this file
################################################################################
sub stop_compression {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    my $delete_checkpoint = 0;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?delete_checkpoint/) {
            if (not length($delete_checkpoint = $args{$_})) {
                $self->_set_err(
                             "Invalid value for argument '-delete_checkpoint'");
                return $self;
            }
            next;
        }
    } ## end foreach (keys %args)

    # get the volume name
    $name = $self->{name};

    $self->{parent}->sanlog(
                          -msg => "Volume::stop_compression() Stop Compression",
                          -msg_level => 1);

    if ($self->{parent}->interface() eq "zapi") {

        # issue the volume compress stop zapi with the passed in options
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd                 => "volume-compress-abort",
                                -cmd_filer           => $self->{filer},
                                -check_result        => 1,
                                '-delete-checkpoint' => $delete_checkpoint,
                                -volume              => $name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to stop compression: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}    # end stop_compression

################################################################################
# Subroutine Name:    start_decompression()
# Objective:          Starts decompression on this volume
# Details:            See POD documentation at the end of this file
################################################################################
sub start_decompression {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    # get the volume name
    $name = $self->{name};

    $self->{parent}->sanlog(
                    -msg => "Volume::start_decompression() Start Decompression",
                    -msg_level => 1);

    if ($self->{parent}->interface() eq "zapi") {

        # issue the volume decompress start zapi with the passed in options
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "volume-decompress-start",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -volume       => $name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to start decompression: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}    # end start_decompression

################################################################################
# Subroutine Name:    stop_decompression()
# Objective:          Stops decompression on this volume
# Details:            See POD documentation at the end of this file
################################################################################
sub stop_decompression {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name;
    my $resultsHash_ref;

    $self->_clear_err();

    # get the volume name
    $name = $self->{name};

    $self->{parent}->sanlog(
                      -msg => "Volume::stop_decompression() Stop Decompression",
                      -msg_level => 1);

    if ($self->{parent}->interface() eq "zapi") {

        # issue the volume decompress stop zapi with the passed in options
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd          => "volume-decompress-abort",
                                -cmd_filer    => $self->{filer},
                                -check_result => 1,
                                -volume       => $name);
    } else {
        $self->_set_err("Unsupported interface. Use interface 'zapi'");
        return $ClientUtil::FALSE;
    }

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to stop decompression: $errmsg");
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Results check failed");
        return $ClientUtil::FALSE;
    }
}    # end stop_decompression

################################################################################
# Subroutine Name: create_lun
# Objective:       create and return a lun object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_lun {
    my ($self, %args) = @_;
    my $lunObj;

    $self->_clear_err();

    if (defined $self->{log}) {

        $lunObj = Lun->_new(-parent => $self,
                            -filer  => $self->{parent},
                            -loghdl => $self->{log},
                            %args);
    } else {
        $lunObj = Lun->_new(-parent => $self,
                            -filer  => $self->{parent},
                            %args);
    }

    if ($lunObj->error) {
        $self->_set_err($lunObj->errmsg());
        return $self;
    }

    #
    # Add the lun object created to our list of cached luns
    #
    push @{$self->{luns}}, $lunObj;

    return $lunObj;
}    # end create_lun()

################################################################################
# Subroutine Name: create_lun_from_file
# Objective:       create a file, then create and return a lun object from file
# Details:         See POD documentation at the end of this file
################################################################################
sub create_lun_from_file {
    my ($self, %args) = @_;
    my $lunObj;
    my $size;
    my $file_name;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?from_file/) {
            if (not length($file_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-from_file'");
                return $self;
            }
            next;
        }
    } ## end foreach (keys %args)

    # check for required arguments
    unless (defined $size and length $size) {
        $self->_set_err("Missing required argument '-size'");
        return $ClientUtil::FALSE;
    }
    unless (defined $file_name and length $file_name) {
        $self->_set_err("Missing required argument '-from_file'");
        return $ClientUtil::FALSE;
    }

    # Need to create the file for the lun create
    $size = &ClientUtil::convert_to_bytes($size);
    my $cmd =
        "priv set diag; mkfile " 
      . $size 
      . " /vol/"
      . $self->{name} . "/"
      . $file_name;
    $self->{parent}->sanlog(-msg => $cmd);

    my @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    if (defined $self->{log}) {

        $lunObj = Lun->_new(-parent => $self,
                            -filer  => $self->{parent},
                            -loghdl => $self->{log},
                            %args);
    } else {
        $lunObj = Lun->_new(-parent => $self,
                            -filer  => $self->{parent},
                            %args);
    }

    # Need to clean up the file after the lun create - do this even if
    # Lun->_new returns an error
    $cmd = "priv set diag; rm /vol/" . $self->{name} . "/" . $file_name;
    $self->{parent}->sanlog(-msg => $cmd);

    @output = $self->{parent}->execute($cmd);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    if ($lunObj->error) {
        $self->_set_err($lunObj->errmsg());
        return $self;
    }

    #
    # Add the lun object created to our list of cached luns
    #
    push @{$self->{luns}}, $lunObj;

    return $lunObj;
}    # end create_lun_from_file()

################################################################################
# Subroutine Name: destroy()
# Objective:       destroy volume
# Details:         See POD documentation at the end of this file
###############################################################################
sub destroy {
    my ($self) = @_;
    my ($cmd, $name, @output,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol destroy " . $name . " -f";

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub destroy

################################################################################
# Subroutine Name: destroy_lun
# Objective:       destroy a named lun object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_lun {
    my ($self, $name, %args) = @_;
    my ($lunObj, @luns, $value, %dup_luns, @uniq_luns,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err(
                "Error: Volume::destroy_lun() lun name is a required argument");
        return $ClientUtil::FALSE;
    }

    # If not regular expression than assume exact match
    ($name =~ /\/\S+\//)
      ? $value = $name =~ s/\/\S+\///
      : $value = $name;

    # Delete the lun object and update the parent lun list.
    foreach $lunObj (@{$self->{luns}}) {
        if ($lunObj->get_name() =~ /$value/) {
            $lunObj->destroy(%args);
            if ($lunObj->error) {
                my $errmsg = join "\n", $lunObj->errmsg();
                $self->_set_err(  "Error: Volume::destroy_lun() "
                                . "could not destory lun: $errmsg");
                return $ClientUtil::FALSE;
            }
        } else {
            push @luns, $lunObj;
        }
    } ## end foreach $lunObj (@{$self->{...

    #
    # @luns may contain duplicate entries of luns. We need to
    # remove those and keep only unique values.
    # using sort function of Perl, we can achieve this.
    #
    @dup_luns{@luns} = ();
    @uniq_luns       = sort keys %dup_luns;
    $self->{luns}    = \@uniq_luns;

    return $ClientUtil::TRUE;
} ## end sub destroy_lun

################################################################################
# Subroutine Name: display()
# Objective:       Display volume information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;
    my ($cmd,     $name,   $header, $lunfilter, $lunsOnly,
        $summary, @output, @buffer, @lunObjs,   @luns,);

    $self->_clear_err();
    $lunsOnly = 0;

    # $self->sanlog(-msg => "REF=".ref($args{-lunfilter})[0]);

    # parse arguments
    foreach (keys %args) {
        if (/-?lunfilter/i) {
            if (not length($lunfilter = $args{$_})
                or ref($args{$_}[0]) ne "Lun") {
                $self->_set_err(
                    "Error: Volume::display() Invalid value for argument '-lunfilter'"
                    );
                return "";
            }

            push @lunObjs, @{$args{$_}};
            next;
        } elsif (/-?lunsOnly/i) {
            $lunsOnly = $args{$_};
            next;
        } else {
            $self->_set_err(
                 "Error: volume::display() Invalid parameter specified \'$_\'");
            return "";
        }
    } ## end foreach (keys %args)

    if ($lunsOnly == 0) {

        $name = $self->{name};

        $cmd = "vol status " . $name;

        @output = $self->{parent}->execute($cmd, -hide => 1);
        if ($self->{parent}->error()) {
            $self->_set_err($self->{parent}->errmsg());
            return "";
        }

        push @buffer, "Volume Summary: " . $self->get_name(), @output, "\n";
    } ## end if ($lunsOnly == 0)

    $header =
      sprintf(
        "Lun Detail:\n"
          . "%-10.10s %-6.6s %-25.25s %-8.8s %-6.6s %-7.7s %-6.6s %-3.3s %-20.20s\n"
          . "%-10.10s %-6.6s %-25.25s %8.8s %6.6s %7.7s %6.6s %3.3s %20.20s",
        "FILERNAME", "VOLUME", "NAME",   "SIZE",   "ACCESS", "STATE",
        "MAPPED",    "ID",     "IGROUP", "=" x 10, "=" x 6,  "=" x 25,
        "=" x 8,     "=" x 6,  "=" x 7,  "=" x 6,  "=" x 3,  "=" x 20);

    push @buffer, $header;

    # Determine if user requested a lun filter
    if (@lunObjs) {
        push @luns, @lunObjs;
    } else {
        push @luns, @{$self->{luns}};
    }

    foreach my $lunObj (@luns) {
        $summary =
          sprintf(
            "%-10.10s %-6.6s %-25.25s %-8.8s %-6.6s %-7.7s %-6.6s %-3.3s %-30.30s",
            $self->{parent}->get_name(), $self->get_name(),
            $lunObj->get_name(),         $lunObj->get_size(),
            $lunObj->get_access(),       $lunObj->get_state(),
            ($lunObj->is_mapped()) ? "Yes" : "No", $lunObj->get_id(),
            $lunObj->get_igroup());

        push @buffer, $summary;

        ### Display any additional igroups the lun is mapped too.
        my $primary_igroup = $lunObj->get_igroup();
        my @maps           = $lunObj->get_maps();
        foreach my $map (@maps) {

            # Skip over first igroup
            next if ($map =~ /^$primary_igroup$/i);
            $summary =
              sprintf(
                "%-10.10s %-6.6s %-25.25s %-8.8s %-6.6s %-7.7s %-6.6s %-3.3s %-20.20s",
                "", "", "", "", "", "", "", $lunObj->get_mapid($map), $map);

            push @buffer, $summary;

        } ## end foreach my $map (@maps)

    } ## end foreach my $lunObj (@luns)

    push @buffer, "\n\n";

    return (wantarray()) ? @buffer : \@buffer;

}    # end display()

################################################################################
# Subroutine Name: is_exported()
# Objective:       Check to see if a volume is exported.
# Details:         See POD documentation at the end of this file
################################################################################
sub is_exported() {
    my ($self) = @_;

    $self->_clear_err();

    my $parent = $self->get_parent();
    my $volume = $self->get_name();

    my $cmd = "exportfs";

    my @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Parse the output looking for the volume
    # Ouput should look like this:
    #
    # /vol/vol0/home\t-sec=sys,rw,nosuid
    # /vol/vol0\t-sec=sys,rw,anon=0,nosuid
    foreach my $line (@output) {
        if ($line =~ /\/vol\/${volume}\s+/) {
            return $ClientUtil::TRUE;
        }
    }
    return $ClientUtil::FALSE;
} ## end sub is_exported()

################################################################################
# Subroutine Name: export_fs()
# Objective:       export a volume to a host
# Details:         See POD documentation at the end of this file
################################################################################
sub export_fs() {
    my ($self, $hostname, %args) = @_;

    my ($parent, $cmd, @output,);

    $self->_clear_err();

    # Initialize some variables
    my $persistent = $ClientUtil::FALSE;

    # Parse the arguments
    foreach (keys %args) {
        if (/-?persistent/) {
            if (length($persistent = $args{$_}) > 0
                and $persistent =~ /YES|TRUE|1/i) {
                $persistent = $ClientUtil::TRUE;
            } else {
                $self->_set_err("Invalid value for argument '-persistent'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $parent = $self->get_parent();

    # Defined the options based on whether or not a hostname was specified
    my $options;
    if (defined($hostname) and $hostname ne "") {
        $options = "sec=sys,anon=0,rw=" . $hostname;
    } else {
        $options = "sec=sys,anon=0,rw";
    }

    # Check to see if the unexport needs to be persistent and adjust the cmd
    # appropriately.
    if ($persistent == $ClientUtil::TRUE) {
        $cmd = "exportfs -p " . $options . " /vol/" . $self->get_name();
    } else {
        $cmd = "exportfs -io " . $options . " /vol/" . $self->get_name();
    }

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub export_fs()

################################################################################
# Subroutine Name: get_filername()
# Objective:       return filername
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filername {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{parent}) {
        return $self->{parent}->get_name();
    } else {
        $self->_set_err(
                  "Error: Volume::get_filername() Volume object has no parent");
        return \{};
    }

    return \{};
} ## end sub get_filername

################################################################################
# Subroutine Name: get_freespace()
# Objective:       Determine amount of space available on volume.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self, %args) = @_;
    my ($cmd, $line, $name, $reserve, $size, $unit, @output,);

    $self->_clear_err();

    # Assume snap reserve is on
    $reserve = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?reserve/) {
            if (not length($reserve = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::get_freespace() Invalid value for argument '-reserve'"
                    );
                return $self;
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::get_freespace() Invalid parameter specified \'$_\'"
                );
            return $self;
        }
    } ## end foreach (keys %args)

    $name = $self->{name};

    $cmd = "lun maxsize /vol/" . $name;

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $line =
      ($reserve =~ /0|f|false|o|off|n|no/i)
      ? $output[1]
      : $output[2];

    if ($line =~ /no space/i) {
        $self->_set_err(
                       "Error: Volume::get_freespace() There is no space left");
        return $ClientUtil::FALSE;
    }

    ($size, $unit) = $line =~ /\:\s+(\d+).*(\w+)\s+\(/;

    return $size . $unit;
} ## end sub get_freespace
################################################################################
# Subroutine Name: get_name
# Objective:       Return Volumes name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err(
                     "Error: Volume::get_name() Volume object is missing name");
        return "";
    }
} ## end sub get_name

################################################################################
# Subroutine Name: get_lun
# Objective:       return a lun object
# Details:         See POD documentation at the end of this file
#                  TBD - this won't support qtrees if you have more than
#                        one qtree per volume
################################################################################
sub get_lun {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err(
                    "Error: Volume::get_lun() lun name is a required argument");
        return $ClientUtil::FALSE;
    }

    # Parse out lun name if sanlun command included
    # warning message with path
    ($name) = $name =~ /WARNING:\s+(\S+)/i if ($name =~ /WARN/);

    # Check if full lun path include filer name was
    # given.  If so extract the lun name.
    if ($name =~ /:|\//) {
        $name = basename($name);
    }

    foreach my $obj (@{$self->{luns}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: Volume::get_lun() Unable to find lun '$name'");

    return {};
} ## end sub get_lun

################################################################################
# Subroutine Name: get_qtree
# Objective:       return a qtree object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_qtree {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Qtree name is a required argument");
        return $ClientUtil::FALSE;
    }

    # Remove any possible trailing slashes at the end of name
    $name =~ s/\/$//g;

    # Check if full qtree path include filer name was
    # given.  If so extract the qtree name.
    if ($name =~ /\:|\//) {
        $name = basename($name);
    }

    foreach my $obj (@{$self->{qtrees}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    # couldn't find the qtree so checking if it exists
    # and then creating an object to represent it if it does
    if ($self->qtree_exist($name)) {
        my $qtree_obj = Qtree->_new(-parent => $self,
                                    -filer  => $self->{parent},
                                    -exist  => 1,
                                    -name   => $name);
        if ($qtree_obj->error) {
            $self->_set_err($qtree_obj->errmsg());
            return {};
        }

        #
        # Add the Qtree object created to our list of cached qtrees
        #
        push @{$self->{qtrees}}, $qtree_obj;

        return $qtree_obj;
    } ## end if ($self->qtree_exist...

    $self->_set_err("Unable to find qtree '$name'.");

    return {};
} ## end sub get_qtree

###############################################################################
# Method:    get_qtree_id()
# Objective: get the Qtree ID
# Details:   See POD documentation at the end of this file.
###############################################################################

sub get_qtree_id {
    my ($self, $qtree_name) = @_;
    my $vol_name = $self->get_name();
    my $cmd_obj  = $self->{parent}->cmd_obj();
    my $output;
    my $errmsg;
    my $resultsHash_ref;
    my $qtreeid;

    # Build the command line and execute the command
    my %cmd_args = (-cmd           => "qtree-list",
                    -cmd_filer     => $self->{filer},
                    -cmd_interface => "cli",
                    -check_result  => 1,
                    -id            => "",
                    -volume        => "$vol_name");

    if (lc($self->{parent}->{interface}) eq "cli") {
        $cmd_args{"-cmd_privilege"} = "test";
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # log results
    $self->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to list qtree info: $errmsg");
        $self->sanlog(-msg => $self->errmsg());
    }

    # get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    my @output = split("\n", $output);

    # check for error
    if (defined $output[0] and length $output[0]) {
        if ($output[0] =~ /qtree status:/) {
            $self->_set_err("CLI failed: ", @output);
            return $ClientUtil::FALSE;
        }
    }

    # parse the output
    for (my $ind = 0 ; $ind < scalar(@output) ; $ind++) {
        $output[$ind] =~ s/[\s\t]+/ /g;

        $self->sanlog(-msg => "Line " . $output[$ind]);

        my @elements = split(" ", $output[$ind]);

        for (my $ind1 = 0 ; $ind1 < scalar(@elements) ; $ind1++) {
            if (    ($elements[0] =~ /$vol_name/)
                and ($elements[1] =~ /$qtree_name/)) {
                $qtreeid = $elements[5];
            }
        }
    } ## end for (my $ind = 0 ; $ind...

    return $qtreeid;
} ## end sub get_qtree_id

################################################################################
# Subroutine Name: get_luns
# Objective:       return array of lun objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_luns {
    my ($self) = @_;
    my (@lunObjs,);

    $self->_clear_err();

    if (defined $self->{luns}) {
        @lunObjs = @{$self->{luns}};
    } else {

        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @lunObjs : \@lunObjs;

} ## end sub get_luns

################################################################################
# Subroutine Name: get_lun_cnt
# Objective:       return number of volume luns
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_cnt {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{luns}) {
        return scalar(@{$self->{luns}});
    } else {
        $self->_set_err(
                 "Error: Volume::get_lun_cnt() Volume object has no lun count");
        return 0;
    }

    return 0;
} ## end sub get_lun_cnt

################################################################################
# Subroutine Name: get_parent
# Objective:       return filers object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_parent {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{parent}) {
        return $self->{parent};
    } else {
        $self->_set_err(
                     "Error: Volume::get_parent() Volume object has no parent");
        return \{};
    }
    return \{};
} ## end sub get_parent

###############################################################################
# Method:    get_state()
# Objective: Get state of volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state {
    my ($self) = @_;
    my ($line, @output, $cmd, $name, $state,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol status $name";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err(
                      "Error: Volume::get_state(): Unable to get volume state");
        return "";
    }

    $line = $output[1];

    ($state) = $line =~ /$name\s+(\w+)/i;

    $self->{state} = $state;

    return $state;

} ## end sub get_state

###############################################################################
# Method:    get_type()
# Objective: Get type of volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_type {
    my ($self) = @_;
    my ($line, @output, $cmd, $name, $type,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol status $name";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Error: Volume::get_type(): Unable to get volume type");
        return "";
    }

    $line = $output[1];

    my ($tail) = $line =~ /$name\s+\S+\s+(.*)/i;

    if ($tail =~ /flex/i) {
        $type = "flex";
    } elsif ($tail =~ /trad/i) {
        $type = "trad";
    } else {
        $type = "unknown";
    }

    return $type;

} ## end sub get_type

###############################################################################
# Method:    get_space()
# Objective: Get spaces of volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_space {
    my ($self, %args) = @_;
    my ($line, @output, $cmd, $name, $type, $unit);

    $self->_clear_err();

    # default
    $unit = "k";
    $type = "total";

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_}) or $args{$_} !~ /volume|snapshot/)
            {
                $self->_set_err(  "Error: Volume::get_space() "
                                . "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})
                or $args{$_} !~ /total|used|avail/) {
                $self->_set_err(  "Error: Volume::get_space() "
                                . "Invalid value for argument '-type'");
                return $self;
            }
            next;
        } elsif (/-?unit/) {
            if (not length($unit = $args{$_}) or $args{$_} !~ /k|m|g|t/i) {
                $self->_set_err(  "Error: Volume::get_space() "
                                . "Invalid value for argument '-unit'");
                return $self;
            }
            next;
        } else {
            $self->_set_err(  "Error: Volume::get_space() "
                            . "Invalid parameter specified \'$_\'");
            return $self;
        }
    } ## end foreach (keys %args)

    $cmd = "df -$unit " . $self->get_name();

    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err(
                 "Error: Volume::get_space(): " . "Unable to get volume space");
        return "";
    }

    $line =
      (not length($name) or $name !~ /snapshot/) ? $output[1] : $output[2];

    my ($total, $used, $avail) = $line =~ /\s+(\d+)\w+\s+(\d+)\w+\s+(\d+)/i;

    return $total . $unit if ($type =~ /total/i);
    return $used . $unit  if ($type =~ /used/i);
    return $avail . $unit if ($type =~ /avail/i);

}    # end get_space()

################################################################################
# Subroutine Name: offline()
# Objective:       Takes a vol offline
# Details:         See POD documentation at the end of this file
################################################################################

sub offline {
    my ($self) = @_;
    my ($cmd, $name, @output,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol offline " . $name;

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{state} = "offline";

    return $ClientUtil::TRUE;
} ## end sub offline

################################################################################
# Subroutine Name: online()
# Objective:       Puts a volume online
# Details:         See POD documentation at the end of this file
################################################################################
sub online {
    my ($self) = @_;
    my ($cmd, $name, @output,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol online " . $name;

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{state} = "online";

    return $ClientUtil::TRUE;
} ## end sub online
################################################################################
# Subroutine Name: restrict()
# Objective:       Puts a volume in restrict
# Details:         See POD documentation at the end of this file
################################################################################
sub restrict {

    my ($self) = @_;
    my ($cmd, $name, @output,);

    $self->_clear_err();

    $name = $self->{name};

    $cmd = "vol restrict " . $name;

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{state} = "restricted";

    return $ClientUtil::TRUE;

} ## end sub restrict
################################################################################
# Subroutine Name: read_only_volume()
# Objective:       Puts a volume online in read only mode
#                  This is cli only and is available only in priv set test mode
# Details:         See POD documentation at the end of this file
################################################################################

sub read_only_volume {
    my ($self)    = @_;
    my $vol_name  = $self->get_name();
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->{interface};
    my $output;
    my $resultsHash_ref;

    # Build the command line and execute the command
    my %cmd_args = (-cmd           => "volume-online",
                    -cmd_filer     => $self->{filer},
                    -cmd_interface => $interface,
                    -check_result  => 1,
                    -name          => "$vol_name",
                    -readonly      => "",);

    if (lc($self->{parent}->{interface}) eq "cli") {
        $cmd_args{"-cmd_privilege"} = "test";
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # log results
    $self->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        $self->_set_err("Unable to online vol:", $cmd_obj->errmsg());
        $self->sanlog(-msg => $self->errmsg());
    }

    # get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    $self->{state} = "online";

    return $ClientUtil::TRUE;
} ## end sub read_only_volume

################################################################################
# Subroutine Name: options()
# Objective:       Puts a volume online
# Details:         See POD documentation at the end of this file
################################################################################
sub options {
    my ($self, %args) = @_;
    my ($cmd, $name, $parameter, $value, @options, @output,);

    $self->_clear_err();
    @options = qw(
      acdirmax            fractional_reserve  nvfail              snaplock_maximum_period
      acregmax            fs_size_fixed       raid_cv             snaplock_minimum_period
      acsymmax            guarantee           raid_zoned          snapmirrored
      actimeo             maxdirsize          raidsize            svo_allow_rman
      compression         metadata_checking   raidtype            svo_checksum
      convert_ucode       minra               read_realloc        svo_enable
      create_ucode        no_atime_update     resyncsnaptime      svo_reject_errors
      extent              no_i2p              root                thorough_scrub
      extent_size         nosnap              schedsnapname       try_first
      flexcache_min_reserve nosnapdir         snaplock_default_period
      snapshot_clone_dependency
      );

    foreach my $parameter (keys %args) {
        $parameter =~ s/-//g;
        unless (scalar(grep { /^$parameter$/ } @options)) {
            $self->_set_err(
                "Error: Volume::options() Invalid parameter supplied '$parameter'"
                );
            return $ClientUtil::FALSE;
        }
    }

    $name = $self->{name};

    foreach my $parameter (keys %args) {
        $value = $args{$parameter};
        $parameter =~ s/-//g;
        $cmd = "vol options " . $name . " " . $parameter . " " . $value;

        @output = $self->{parent}->execute($cmd, -hide => 1);
        if ("@output" =~ /illegal/i) {
            $self->_set_err(@output);
            return $ClientUtil::FALSE;
        }

        $self->{options}{$parameter} = $value;
    } ## end foreach my $parameter (keys...

    return $ClientUtil::TRUE;
} ## end sub options

################################################################################
# Subroutine Name: get_aggr()
# Objective:       Return the containing aggregate
# Details:         See POD documentation at the end of this file
################################################################################
sub get_aggr {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $command_retries = 1;

    # Parse the required/optional arguments
    foreach (keys %args) {
        if (/-?command_retries/) {
            unless (length($command_retries = $args{$_}) > 0) {
                $self->_set_err(
                             "Invalid value for '-command_retries': $args{$_}");
                return "";
            }
        }
    }

    if (not defined($self->{aggr})) {
        my $cmd_obj = $self->{parent}->cmd_obj();
        my $name    = $self->{name};

        # Get the filer version
        my $version = $cmd_obj->filer_version();
        if ($cmd_obj->error()) {
            my $errmsg = join "\n", $cmd_obj->errmsg();
            $self->_set_err(  "Error: Volume::get_aggr(): Unable to "
                            . "get the filer version: $errmsg");
            return "";
        }

        # Enter the command retry loop.  We will try $command_retries
        # times if necessary
        my $attempt = 1;
        while ($attempt <= $command_retries) {

            # Put in some logic that sleeps before retrying for every attempt
            # but the first
            if ($attempt > 1) {

                # Sleep 10 seconds and then retry
                &SanUtil::nate_sleep(10);
            }
            $self->{parent}->sanlog(
                                   -msg => "Volume::get_aggr(): "
                                     . "Getting containing aggregate "
                                     . "- Attempt $attempt of $command_retries",
                                   -msg_level => 1);

            my $resultsHash_ref =
              $cmd_obj->execute_cmd(-cmd               => "volume-list-info",
                                    -cmd_filer         => $self->{filer},
                                    -cmd_interface     => "zapi",
                                    -cmd_filer_version => $version,
                                    -check_result      => 1,
                                    -check_output      => 1,
                                    -volume            => $name);

            if ($cmd_obj->error()) {

                # CmdsDriver failed.  Log an error and retry.
                my $errmsg = join "\n", $cmd_obj->errmsg();
                $self->{parent}->sanlog(-msg => "Volume::get_aggr(): "
                                          . "Unable to get volume "
                                          . "status: $errmsg",
                                        -msg_level => 1);
                $attempt++;
                next;
            } ## end if ($cmd_obj->error())

            $self->{parent}->log_cmd_results($resultsHash_ref);

            unless (   $resultsHash_ref->{result_check}
                    && $resultsHash_ref->{output_check}) {

                # Zapi failed.  Log an error and retry.
                $self->{parent}->sanlog(
                                 -msg => "Volume::get_aggr(): "
                                   . "Zapi volume-list-info failed to retrieve "
                                   . "volume status.  Retrying command.",
                                 -msg_level => 1);
                $attempt++;
                next;
            } ## end unless ($resultsHash_ref->...

            # Get the output
            my $output = $resultsHash_ref->{output};

            # Convert the output to xml
            my $output_hash = $self->{parent}->xml2hash($output);

            # Check to make sure we have a type defined.  If we don't,
            # then the data is bad so retry.
            unless (
                 defined($output_hash->{'volumes'}->{'volume-info'}->{'type'})
                 and $output_hash->{'volumes'}->{'volume-info'}->{'type'} ne "")
            {
                $self->{parent}->sanlog(-msg => "Volume::get_aggr(): "
                                          . "Volume type not found ... "
                                          . "retrying command",
                                        -msg_level => 1);
                $attempt++;
                next;
            } ## end unless (defined($output_hash...

            if ($output_hash->{'volumes'}->{'volume-info'}->{'type'} eq "trad")
            {

                # We have a trad vol so set an error and return ""
                $self->_set_err(  "Error: Volume::get_aggr(): Unable to "
                                . "get containing aggregate because this is "
                                . "a trad vol.");
                return "";
            } elsif (
                 $output_hash->{'volumes'}->{'volume-info'}->{'type'} eq "flex")
            {

                # We have a flex vol so now make sure that containing-aggregate
                # is defined.  If it is not defined or is blank, retry
                # the command.
                if (defined($output_hash->{'volumes'}->{'volume-info'}
                              ->{'containing-aggregate'})
                    and $output_hash->{'volumes'}->{'volume-info'}
                    ->{'containing-aggregate'} ne ""
                  ) {

                    # Update the aggr on self and then return it.
                    $self->{aggr} =
                      $output_hash->{'volumes'}->{'volume-info'}
                      ->{'containing-aggregate'};
                    return $self->{aggr};
                } else {

                    # containing-aggregate was not defined so retry
                    $self->{parent}->sanlog(
                                       -msg => "Volume::get_aggr(): "
                                         . "Containing aggregate not found ... "
                                         . "retrying command",
                                       -msg_level => 1);
                    $attempt++;
                    next;
                } ## end else [ if (defined($output_hash...
            } else {

                # We got some type other than trad or flex. Increase tye counter and try again
                $self->{parent}->sanlog(-msg => "Volume::get_aggr(): "
                                          . "Invalid volume type returned ... "
                                          . "retrying command",
                                        -msg_level => 1);
                $attempt++;
                next;
            } ## end else [ if ($output_hash->{'volumes'...
        } ## end while ($attempt <= $command_retries)

        # If we make it to this point, then we have exhausted our retries.
        # Set an error and return.
        $self->_set_err("Error: Volume::get_aggr(): Unable to get volume info. "
                        . "$command_retries retries exhausted.");
        return "";
    } else {
        return $self->{aggr};
    }
} ## end sub get_aggr

################################################################################
# Subroutine Name: get_info()
# Objective:       Return information about the volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_info() {
    my ($self)  = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $name    = $self->get_name();

    $self->_clear_err();

    #
    # Exectue the volume-list-info zapi and check to see if it executed
    #
    my $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd           => "volume-list-info",
                            -cmd_filer     => $self->{filer},
                            -check_result  => 1,
                            -cmd_interface => "zapi",
                            -volume        => $name);

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: Volume::get_info(): Unable to "
                        . "get filer info: $errmsg");
        return "";
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Put the output into a hash and check to see if the cmd passed
    #
    my $output_hash = $self->{parent}->xml2hash($output);
    my $status      = $output_hash->{status};

    if ($status eq "failed") {
        my $errmsg = $output_hash->{reason};
        $self->_set_err(
                     "Error: Volume::get_info() " . "unable to get volume info",
                     $errmsg);
        return "";
    }

    return $output;
}    # end get_info()

################################################################################
# Subroutine Name: is_offline()
# Objective:       Check the state of the volume
# Details:         See POD documentation at the end of this file
################################################################################
sub is_offline {
    my ($self) = @_;

    $self->_clear_err();

    return ($self->{state} =~ /offline/) ? 1 : 0;

}

################################################################################
# Subroutine Name: is_online()
# Objective:       Check the state of the volume
# Details:         See POD documentation at the end of this file
################################################################################
sub is_online {
    my ($self) = @_;

    $self->_clear_err();

    return ($self->{state} =~ /online/) ? 1 : 0;

}

################################################################################
# Subroutine Name: is_restrict()
# Objective:       Check the state of the volume
# Details:         See POD documentation at the end of this file
################################################################################
sub is_restricted {
    my ($self) = @_;
    $self->_clear_err();
    return ($self->{state} =~ /restricted/)
      ? $ClientUtil::TRUE
      : $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh volume lun objects and aggr information
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Default the refresh to on for luns and aggrs
    my $aggr = 1;
    my $luns = 1;

    # Parse the arguments
    foreach (keys %args) {
        if (/-?aggr/) {
            unless (length($aggr = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-aggr'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?luns/) {
            unless (length($luns = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-luns'");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameters specified for '$_'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # Refresh the lun objects
    if ($luns == 1) {

        # Remove existing vol cache
        $self->{luns} = undef;
        $self->{luns} = [];

        $self->_create_luns();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    } ## end if ($luns == 1)

    # Refresh the aggr info
    if ($aggr == 1) {
        $self->{aggr} = undef;
        $self->get_aggr(-command_retries => 5);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method:    snap_create()
# Objective: delete snap shots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_create {
    my ($self, %args) = @_;
    my (@output, $cmd, $name, $vol,);

    $self->_clear_err();

    $vol = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_create() Invalid value for argument '-name'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_create() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)

    unless (defined $name and length $name) {
        $self->_set_err(
              "Error: Volume::snap_create() Missing required argument '-name'");
        return {};
    }

    $self->sanlog(-msg => "Creating snapshot '$name' for volume '$vol'");

    # Quote snapname since it might contain spaces
    $cmd = "snap create $vol '$name'";

    $self->{parent}->execute($cmd, -hide => 1);

    $self->snap_list(-filter => $name);

    my $snap_check = $self->snap_exist(-snapname => $name);

    if (not $snap_check) {

        $self->_set_err(
            "Error: Volume::snap_create Failed to create snap '$name' on the volume '$vol' "
            );
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub snap_create

###############################################################################
# Method:    snap_restore()
# Objective: restore snap shots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_restore {
    my ($self, %args) = @_;
    my (@output, $cmd, $name, $vol,);

    $self->_clear_err();

    $vol = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_restore() Invalid value for argument '-name'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_restore() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)

    unless (defined $name and length $name) {
        $self->_set_err(
             "Error: Volume::snap_restore() Missing required argument '-name'");
        return {};
    }

    $self->sanlog(-msg => "Restoring snapshot '$name' for volume '$vol'");

    # Quote snapname since it might contain spaces
    $cmd = "snap restore -f -t vol -s '$name' $vol";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    if ("@output" !~ /revert successful/i) {
        my $errmsg = $output[0];
        $self->_set_err(  "Error: Volume::snap_restore(): Unable to restore "
                        . "snapshot '$name' on volume '$vol': $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub snap_restore
###############################################################################
# Method:    single_file_snap_restore()
# Objective: restore snap shots to a file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub single_file_snap_restore {
    my ($self, %args) = @_;

    my $restore_from_path;
    my $restore_to_path;
    my $interface = "cli";
    my $force     = "no";    #cli only
    my $snapname;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my %cmd_args;
    my $check_output = 0;
    my $check_result = 1;
    $self->_clear_err();

    foreach (keys %args) {
        if (/-?restore_from_path/) {
            if (not length($restore_from_path = $args{$_})) {
                $self->_set_err(
                             "Invalid value for argument '-restore_from_path'");
                return $self;
            }
            next;
        } elsif (/-?restore_to_path/) {
            if (not length($restore_to_path = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-restore_to_path'");
                return $self;
            }
            next;
        } elsif (/-?snap_name/) {
            if (not length($snapname = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snap_name'");
                return $self;
            }
            next;
        } elsif (/-?cmd_interface/) {
            if (not length($interface = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cmd_interface'");
                return $self;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $self;
            }
            next;
        } elsif (/-?check_output/) {
            if (not length($check_output = $args{$_})) {
                $self->_set_err("Invalid value for argument '-check_output'");
                return $self;
            }
            next;
        } elsif (/-?check_result/) {
            if (not length($check_result = $args{$_})) {
                $self->_set_err("Invalid value for argument '-check_result'");
                return $self;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $restore_from_path and length $restore_from_path) {
        $self->_set_err(
                  "value for restore from path setting is a required argument");
        return $ClientUtil::FALSE;
    }
    unless (defined $snapname and length $snapname) {
        $self->_set_err(
                      "value for snapshot name setting is a required argument");
        return $ClientUtil::FALSE;
    }

    %cmd_args = ("-cmd"           => "snapshot-restore-file",
                 "-cmd_filer"     => $self->get_filername(),
                 "-check_result"  => $check_result,
                 "-cmd_interface" => $interface,
                 "-path"          => $restore_from_path,
                 "-snapshot"      => $snapname);

    if (defined($restore_to_path)) {
        $cmd_args{"-restore-path"} = "$restore_to_path";
    }
    if (($force =~ /yes/i) && ($interface =~ /cli/i)) {
        $cmd_args{"-force"} = " ";
    }

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};
    $self->sanlog(-msg_level => 3,
                  -msg       => "\n output is: $output");

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        $self->sanlog(-msg => "restore failed");
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "restore is successful");

    return $ClientUtil::TRUE;

} ## end sub single_file_snap_restore

###############################################################################
# Method:    snapshot_partial_restore_file()
# Objective: restore snap shots to a partial file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snapshot_partial_restore_file() {
    my ($self, %args) = @_;
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = "zapi";
    my $resultsHash_ref;
    my $path;
    my $snapshot;
    my $start_byte;
    my $byte_count;
    my $check_output = 0;
    my $check_result = 1;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?path/) {
            if (not length($path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $self;
            }
            next;
        } elsif (/-?snapshot/) {
            if (not length($snapshot = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot'");
                return $self;
            }
            next;
        } elsif (/-?start_byte/) {
            if (not length($start_byte = $args{$_})) {
                $self->_set_err("Invalid value for argument '-start_byte'");
                return $self;
            }
            next;
        } elsif (/-?byte_count/) {
            if (not length($byte_count = $args{$_})) {
                $self->_set_err("Invalid value for argument '-byte_count'");
                return $self;
            }
            next;
        } elsif (/-?check_output/) {
            if (not length($check_output = $args{$_})) {
                $self->_set_err("Invalid value for argument '-check_output'");
                return $self;
            }
            next;
        } elsif (/-?check_result/) {
            if (not length($check_result = $args{$_})) {
                $self->_set_err("Invalid value for argument '-check_result'");
                return $self;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $path and length $path) {
        $self->_set_err("value for path setting is a required argument");
        return $ClientUtil::FALSE;
    }
    unless (defined $snapshot and length $snapshot) {
        $self->_set_err("value for snapshot setting is a required argument");
        return $ClientUtil::FALSE;
    }
    unless (defined $start_byte and length $start_byte) {
        $self->_set_err("value for start_byte setting is a required argument");
        return $ClientUtil::FALSE;
    }
    unless (defined $byte_count and length $byte_count) {
        $self->_set_err("value for byte_count setting is a required argument");
        return $ClientUtil::FALSE;
    }

    # do a partial restore
    $self->sanlog(  -msg => "--- partial restore :----"
                  . " file_path  = "
                  . $path
                  . " snap_name  = "
                  . $snapshot
                  . " start_byte = "
                  . $start_byte
                  . " byte_count = "
                  . $byte_count
                  . " interface  = "
                  . $interface
                  . " filer name = "
                  . $self->get_filername());

    #order:       byte-count path snapshot start-byte

    my %cmd_args = ("-cmd"           => "snapshot-partial-restore-file",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => $check_result,
                    "-check_output"  => $check_output,
                    "-cmd_interface" => $interface,
                    "-byte-count"    => $byte_count,
                    "-path"          => $path,
                    "-snapshot"      => $snapshot,
                    "-start-byte"    => $start_byte);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};
    $self->sanlog(-msg_level => 3,
                  -msg       => "\n output is: $output");

    #
    # Since we checked the result, lets see if our check was OK
    #

    $self->sanlog(
            -msg => "result_check value is : $resultsHash_ref->{result_check}");
    $self->sanlog(-msg => " output is : $resultsHash_ref->{output}");
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        $self->sanlog(-msg => "restore failed");
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "restore is successful");

    return $ClientUtil::TRUE;
} ## end sub snapshot_partial_restore_file()

###############################################################################
# Method:    snapshot_partial_restore_file_list_info()
# Objective: list the snap restore partial file info
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snapshot_partial_restore_file_list_info() {
    my $self      = shift;
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = "zapi";
    my $resultsHash_ref;
    my $result;

    $self->_clear_err();

    # do a partial restore

    $resultsHash_ref =
      $cmd_obj->execute_cmd("-cmd" => "snapshot-partial-restore-file-list-info",
                            "-cmd_interface" => $interface,
                            "-cmd_filer"     => $self->get_filername(),
                            "-check_output"  => 1,
                            "-check_result"  => 1,);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    ($result) = $output =~ /(\d+)/;
    return ($result);
} ## end sub snapshot_partial_restore_file_list_info()

###############################################################################
# Method:    snap_autodelete()
# Objective: set autodelete of snapshots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_autodelete {
    my ($self, %args) = @_;
    my (@output, $cmd, $state,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?state/) {
            if (not length($state = $args{$_}) or $args{$_} !~ /on|off|reset/) {
                $self->_set_err(
                    "Error: Volume::snap_autodelete() Invalid value for argument '-state'"
                    );
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_autodelete() Invalid parameter specified \'$_\'"
                );
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # check for required arguments
    unless (defined $state and length $state) {
        $self->_set_err(
            "Error: Volume::snap_autodelete() Missing required argument '-state'"
            );
        return $ClientUtil::FALSE;
    }

    $cmd = "snap autodelete " . $self->get_name() . " " . $state;
    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless ($output[0] =~ /enabled|disabled|reset/i) {
        my ($errmsg) = $output[0];
        $self->_set_err("Error: Volume::snap_autodelete() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end snap_autodelete()

###############################################################################
# Method:    snap_autodelete_trigger()
# Objective: set autodelete trigger for snapshots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_autodelete_trigger {
    my ($self, %args) = @_;
    my (@output, $cmd, $trigger,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?trigger/) {
            if (not length($trigger = $args{$_})
                or $args{$_} !~ /volume|snap_reserve|space_reserve/) {
                $self->_set_err(  "Error: Volume::snap_autodelete_trigger() "
                                . "Invalid value for argument '-trigger'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(  "Error: Volume::snap_autodelete_trigger() "
                            . "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # check for required arguments
    unless (defined $trigger and length $trigger) {
        $self->_set_err(  "Error: Volume::snap_autodelete_trigger() "
                        . "Missing required argument '-trigger'");
        return $ClientUtil::FALSE;
    }

    $cmd = "snap autodelete " . $self->get_name() . " trigger " . $trigger;
    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless ($output[0] =~ /options\s+set/i) {
        my ($errmsg) = $output[0];
        $self->_set_err("Error: Volume::snap_autodelete_trigger() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end snap_autodelete_trigger()

###############################################################################
# Method:    snap_delete()
# Objective: delete snap shots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_delete {
    my ($self, %args) = @_;
    my (@buffer, @output, @tmp, $cmd, $found, $name, $vol,);

    $self->_clear_err();

    $found = 0;
    $vol   = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_delete() Invalid value for argument '-name'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_delete() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)

    unless (defined $name and length $name) {
        $self->_set_err(
              "Error: Volume::snap_delete() Missing required argument '-name'");
        return {};
    }

    if ($name =~ /\*/) {
        $self->sanlog(-msg => "deleting all snapshots from volume '$vol'");
        $cmd = "snap delete -a -f $vol ";
    } else {
        $self->sanlog(-msg => "deleting snapshot '$name' from volume '$vol'");

        # Quote snapname since it might contain spaces
        $cmd = "snap delete $vol '$name'";
    }

    @output = $self->{parent}->execute($cmd, -hide => 1);

    unless ($output[0] =~ /deleting snapshot\.\.\./) {
        my ($errmsg) = join "\n", @output;
        $self->_set_err("Error: Volume::snap_delete() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end snap_delete()

###############################################################################
# Method:    snap_status()
# Objective: Get status of snapshots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_status {
    my ($self, %args) = @_;
    my (@buffer, @output, @tmp, $cmd, $found, $filter, $vol, $status);

    $self->_clear_err();

    $vol = $self->get_name();

    $found = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?filter/) {
            if (not length($filter = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_status() Invalid value for argument '-filter'"
                    );
                return (wantarray()) ? () : [];
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_status() Invalid parameter specified \'$_\'"
                );
            return (wantarray()) ? () : [];
        }
    } ## end foreach (keys %args)

    Tharn::logcomment("Listing snapshots status for volume '$vol'");

    $cmd = "snap status " . $self->get_name();
    (@output, $status) = &SanUtil::system_zapi($self->{filer}, $cmd);

    my ($line) = grep (/.*No snapshots exist.*/, @output);
    if ($line =~ /No snapshots exist/i) {
        $self->_set_err("Error: Volume::snap_status() $line");
        return (wantarray()) ? () : [];
    }

    # Get the field header
    push @buffer, (@output[1 .. 2]);

    # Nate dumps out if it sees '--' in output
    # Lets substitute with '=' before we log.
    $buffer[1] =~ s/-/=/g;

    # Store the rest of the data here for filtering
    @tmp = (@output[3 .. scalar(@output) - 1]);

    if (defined $filter and length $filter) {
        $filter =~ s/:/\//g;
        $filter = basename($filter);

        foreach my $line (@tmp) {

            # take spaces in account for snapname
            my ($name) = $line =~ /.*:\d+\s+\d+\s+\S+\s+\d+\s+(.*)/;

            if (_trim($name) =~ /$filter/) {
                push @buffer, $line;
                $found = 1;
            }

        } ## end foreach my $line (@tmp)
        if ($found == 0) {
            $self->_set_err(
                "Error: Volume::snap_status() No snapnames with filter '$filter' found"
                );
            return (wantarray()) ? () : [];
        }

    } else {
        push @buffer, @tmp;
    }

    push @buffer, "\n\n";
    return (wantarray()) ? @buffer : \@buffer;
}    # end snap_status()

###############################################################################
# Method:    snap_list()
# Objective: Get filer version
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_list {
    my ($self, %args) = @_;
    my (@buffer, @output, @tmp, $cmd, $found, $filter, $vol,);

    $self->_clear_err();

    $vol = $self->get_name();

    $found = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?filter/) {
            if (not length($filter = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_list() Invalid value for argument '-filter'"
                    );
                return (wantarray()) ? () : [];
            }
            next;
        } else {
            $self->_set_err(
                 "Error: Volume::snap_list() Invalid parameter specified \'$_\'"
                 );
            return (wantarray()) ? () : [];
        }
    } ## end foreach (keys %args)

    Tharn::logcomment("Listing snapshots for volume '$vol'");

    $cmd = "snap list -n " . $self->get_name();

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($output[1] =~ /does not exist/i) {
        my ($errmsg) = $output[1] =~ /\:(.*)/;
        $self->_set_err("Error: Volume::snap_list() $errmsg");
        return (wantarray()) ? () : [];
    }

    # Get the field header
    push @buffer, (@output[3 .. 4]);

    # Nate dumps out if it sees '--' in output
    # Lets substitute with '=' before we log.
    $buffer[1] =~ s/-/=/g;

    # Store the rest of the data here for filtering
    @tmp = (@output[5 .. scalar(@output) - 1]);

    if (defined $filter and length $filter) {
        $filter =~ s/:/\//g;
        $filter = basename($filter);

        foreach my $line (@tmp) {

           # take spaces in account for snapname
           my ($name) = $line =~ /.*:\d+\s+(.*)/;

           if (_trim($name) =~ /$filter/) {
               push @buffer, $line;
               $found = 1;
           }

        } ## end foreach my $line (@tmp)
        if ($found == 0) {
            $self->_set_err(
                "Error: Volume::snap_list() No snapnames with filter '$filter' found"
                );
            return (wantarray()) ? () : [];
        }

    } else {
        push @buffer, @tmp;
    }

    push @buffer, "\n\n";
    return (wantarray()) ? @buffer : \@buffer;
}    # end snap_list()

###############################################################################
# Method:    snap_exist()
# Objective: Checks if a specific snap exists on this volume.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_exist {
    my ($self, %args) = @_;
    my $snapname;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?snapname/) {
            if (not length($snapname = $args{$_})) {
                $self->_set_err(  "Error: Volume:snap_exist() Invalid value "
                                . "for argument '-snapname'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(  "Error: Volume:snap_exist() Invalid parameter "
                            . "specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    if (not defined($snapname) or not length($snapname)) {
        $self->_set_err(  "Error: Volume::snap_exist() "
                        . "Missing required argument '-snapname'");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Volume::snap_exist() Checking if $snapname exists",
                  -msg_level => 1);

    #-msg_level => $LOG_VERBOSE_LOW);

    my @snap_info = $self->snap_list();

    #
    # The regex is expecting lines in the format of:
    # Jun 6 00:00 snapname
    #
    my $snap_exist = 0;
    foreach my $line (@snap_info) {
        if ($line =~ /\s*\w+\s+\d+\s+\d+:\d+\s+(\S+)/) {
            my $sname = $1;

            if ($sname eq $snapname) {
                $snap_exist = 1;
                last;
            }
        }
    }

    if ($snap_exist == 1) {

        for (my $i = 20; $i > 0; $i--) {
             my @snap_status = $self->snap_status();

             #
             # The regex is expecting lines in the format of:
             # 55     complete   May 23 08:00        41     8.0 21057 hourly.0
             #
             foreach my $line (@snap_status) {
                if ($line =~ /\s*\d+\s+(\S+)\s+\w+\s+\d+\s+\d+:\d+\s+\d+\s+\S+\s+\d+\s+(\S+)/) {
                    my $status = $1;
                    my $name = $2;
                    if ($name eq $snapname) {
                        if($status eq "complete") {
                            return $ClientUtil::TRUE;
                        } else {
                            last;
                        }
                    }
                }
             }
             &SanUtil::nate_sleep(3);
        }
    }
    $self->_set_err("Volume::snap_exist(): Snap $snapname not found");
    return $ClientUtil::FALSE;
}    # end snap_exist()

###############################################################################
# Method:    get_snap_names()
# Objective: Get volume snap names
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_snap_names {
    my ($self, %args) = @_;
    my (@buffer, @output, @snapnames, $found, $cmd, $filter, $snapname, $vol,);

    $self->_clear_err();

    $vol = $self->get_name();

    $found = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?filter/) {
            if (not length($filter = $args{$_})) {
                $self->_set_err(
                    "Error: Volume:get_snap_names() Invalid value for argument '-filter'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume:get_snap_names() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)

    Tharn::logcomment("Listing snapshots for volume $vol");

    $cmd = "snap list -n $vol";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    if ($output[1] =~ /does not exist/i) {
        my ($errmsg) = $output[1] =~ /\:(.*)/;
        $self->_set_err("Error: Volume:get_snap_names() $errmsg");
        return (wantarray()) ? () : [];
    }

    @buffer = (@output[5 .. scalar(@output) - 1]);

    #$self->sanlog(-msg => "BUFFER:",@buffer);

    if (defined $filter and length $filter) {
        foreach my $line (@buffer) {

            # take spaces in account for snapname
            my ($month, $day, $time, @name) = split(/\s+/, $line);
            my $name = "@name";
            if ($name =~ /$filter/) {
                push @snapnames, $name;
                $found = 1;
            }
        } ## end foreach my $line (@buffer)

        if ($found == 0) {
            $self->_set_err(
                "Error: Volume:get_snap_names() No snapnames with filter '$filter' found"
                );
            return (wantarray()) ? () : [];
        }
    } ## end if (defined $filter and...

    return (wantarray()) ? @snapnames : \@snapnames;
}    # end get_snap_names()

###############################################################################
# Method:    snap_get_reserve()
# Objective: retrieve pct for snapshot allocation
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_get_reserve {
    my ($self, %args) = @_;
    my (@output, $cmd, $name, $pct, $vol,);

    $self->_clear_err();

    $vol = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        $self->_set_err(
            "Error: SanFiler::snap_get_reserve() Invalid parameter specified \'$_\'"
            );
        return {};

    }

    # Quote volname since it might contain spaces
    $cmd = "snap reserve '$vol'";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    ($pct) = $output[0] =~ /(\d+)\%/;

    if ($output[0] =~ /usage/i) {
        $self->_set_err("Error: SanFiler::snap_get_reserve()", @output);
        return $ClientUtil::FALSE;
    }

    return $pct;
}    # end snap_get_reserve()

###############################################################################
# Method:    snap_reserve()
# Objective: Reserve percentage for snapshots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_reserve {
    my ($self, %args) = @_;
    my (@output, $cmd, $vol, $pct,);

    $self->_clear_err();

    $vol = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?pct/) {
            if (not length($pct = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::snap_reserve() Invalid value for argument '-pct'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::snap_reserve() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)

    # Remove % if supplied
    $pct =~ s/\%//g if (defined $pct);

    # check for required arguments
    unless (defined $pct and $pct =~ /^\d+$/ and $pct >= 0 and $pct <= 100) {
        $self->_set_err(
                      "Error: Volume::snap_reserve() Invalid percentage '$pct'",
                      "Valid values are 0-100");
        return $ClientUtil::FALSE;
    }

    # Quote snapname since it might contain spaces
    $cmd = "snap reserve '$vol' $pct";

    @output = $self->{parent}->execute($cmd, -hide => 1);

    if (defined $output[1] and $output[1] =~ /does not exist/i) {
        my ($errmsg) = $output[1];
        $self->_set_err("Error: Volume::snap_reserve() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end snap_reserve()

###############################################################################
# Method:    snap_schedule()
# Objective: schedule for taking snapshots
# Details:   See POD documentation at the end of this file.
###############################################################################
sub snap_schedule {
    my ($self, %args) = @_;
    my ($output, $cmd, $vol, $schedule,);

    $self->_clear_err();

    $vol = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?schedule/) {
            if (not length($schedule = $args{$_})) {
                $self->_set_err(  "Error: Volume::snap_schedule() "
                                . "Invalid value for argument '-schedule'");
                return {};
            }
            next;
        } else {
            $self->_set_err(  "Error: Volume::snap_schedule() "
                            . "Invalid parameter specified \'$_\'");
            return {};
        }
    } ## end foreach (keys %args)

    # Quote volname since it might contain spaces
    $cmd = "snap sched '$vol' $schedule";

    $output = $self->{parent}->execute($cmd, -hide => 1);

    if (defined $output and $output =~ /usage:|invalid schedule/i) {
        my ($errmsg) = $output;
        $self->_set_err("Error: Volume::snap_schedule() $errmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end snap_schedule()

################################################################################
# Subroutine Name: get_lun_snap_usage()
# Objective:       Return lun clone information for a given snapshot
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_snap_usage() {
    my ($self, %args) = @_;
    my $resultsHash_ref;
    my $snap_name;
    my $format_output = 0;   # 0 - return raw output 1 - parse and return a hash
    my $volume_name = $self->get_name();
    my $cmd_obj     = $self->{parent}->cmd_obj();
    my $interface   = $self->{parent}->interface();
    my $lc_hash     = {};

    #
    # Parse arguements
    #
    foreach (keys %args) {
        if (/-?snap_name/) {
            if (not length($snap_name = $args{$_})) {
                $self->_set_err(  "Error: Volume::get_lun_snap_usage() "
                                . "Invalid value for argument '-snap_name'");
                return 0;
            }
            next;
        } elsif (/-?format_output/) {
            if (not length($format_output = $args{$_})) {
                $self->_set_err( "Error: get_lun_snap_usage() "
                               . "Invalid value for argument '-format_output'");
                return 0;
            }
            $format_output = 1 if ($format_output =~ /yes/i);
            $format_output = 0 if ($format_output =~ /no/i);
            next;
        } else {
            $self->_set_err(  "Error: get_lun_snap_usage() "
                            . "Invalid parameter specified \'$_\'");
            return 0;
        }
    }    # end foreach

    #
    # Execute the lun-snap-usage-list-info command to get the data we
    # want to send back
    #
    $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd           => "lun-snap-usage-list-info",
                            -cmd_interface => $interface,
                            -check_output  => "no",
                            -check_result  => "yes",
                            -volume        => $volume_name,
                            -snapshot      => $snap_name,);

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err(  "Error: get_lun_snap_usage() "
                        . "Could not get lun snap usage list info: $errmsg");
        return 0;
    }

    #
    # Get the output from the results hash
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check} == 0) {
        $self->_set_err(  "Error: get_lun_snap_usage() Result check failed"
                        . " getting lun snap usage list info");
        return 0;
    }

    #
    # If the user wants unformatted output, send it now and be done
    #
    if ($format_output == 0) {
        return ($output);
    }

    #
    # We handle the output from zapi and CLI differently
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err( "Error: get_lun_snap_usage() "
                           . "Could not get lun snap usage list info: $errmsg");
            return 0;
        }

        #
        # We were either returned a single hash for a single entry or
        # an array of hash.  If it's a single hash, store it in the lc_hash
        # and be done.  If it's an array store each one in the lc_hash
        #
        my $entry_hash;
        if (ref($output_hash->{'lun-snap-usage-luns'}
                  ->{'lun-snap-usage-lun-info'}) ne "ARRAY"
          ) {
            $entry_hash =
              $output_hash->{'lun-snap-usage-luns'}
              ->{'lun-snap-usage-lun-info'};
            my $id;
            if ($entry_hash->{'snapshot'} eq "Active") {
                $id = $snap_name;
            } else {
                $id = $entry_hash->{'snapshot'};
            }
            $lc_hash->{$id} = $entry_hash;
        } else {
            foreach $entry_hash (@{  $output_hash->{'lun-snap-usage-luns'}
                                       ->{'lun-snap-usage-lun-info'}}
              ) {
                my $id;
                if ($entry_hash->{'snapshot'} eq "Active") {
                    $id = $snap_name;
                } else {
                    $id = $entry_hash->{'snapshot'};
                }
                $lc_hash->{$id} = $entry_hash;
            } ## end foreach $entry_hash (@{$output_hash...
        } ## end else [ if (ref($output_hash->...
    } else {

        # CLI case

        my @out = split('\n', $output);
        my $id;
        my $value;

        foreach my $line (@out) {
            if ($line =~ /Active:/) {
                $id = $snap_name;
                $lc_hash->{$id}->{'snapshot'} = "Active";
            } elsif ($line =~ /Snapshot/) {
                (undef, $id) = split(/- \s*/, $line);
                $lc_hash->{$id}->{'snapshot'} = $id;
            } elsif ($line =~ /LUN:/) {
                (undef, $value) = split(/:\s*/, $line);
                $lc_hash->{$id}->{'path'} = $value;
            } elsif ($line =~ /Backed By:/) {
                (undef, $value) = split(/:\s*/, $line);
                $lc_hash->{$id}->{'backing-store'} = $value;
            }
        }    # end foreach line
    }    # end CLI case
    return ($lc_hash);
} ## end sub get_lun_snap_usage()

################################################################################
# Subroutine Name: unexport_fs()
# Objective:       unexport a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unexport_fs() {
    my ($self, %args) = @_;

    my ($parent, $cmd, @output,);

    $self->_clear_err();

    # Initialize some variables
    my $persistent = $ClientUtil::FALSE;

    # Parse the arguments
    foreach (keys %args) {
        if (/-?persistent/) {
            if (length($persistent = $args{$_}) > 0
                and $persistent =~ /YES|TRUE|1/i) {
                $persistent = $ClientUtil::TRUE;
            } else {
                $self->_set_err("Invalid value for argument '-persistent'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $parent = $self->get_parent();

    # Check to see if the unexport needs to be persistent and adjust the cmd
    # appropriately.
    if ($persistent == $ClientUtil::TRUE) {
        $cmd = "exportfs -z " . "/vol/" . $self->get_name();
    } else {
        $cmd = "exportfs -u " . "/vol/" . $self->get_name();
    }

    # Run the command
    @output = $self->{parent}->execute($cmd, -hide => 0);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub unexport_fs()

=cut
################################################################################
# Subroutine Name: wafl_abort()
# Objective:       abort wafl scan processes
# Details:         See POD documentation at the end of this file
################################################################################
sub wafl_abort() {
    my ($self,%args) = @_;
    my(
       $action,
       $action_values,
       $parent,
       $cmd,
       $scanid,
       @scanids,
       @output,
       );

    $self->_clear_err();

    $parent=$self->get_parent();
    $action_values="measure_layout|reallocate|i2p|fold";

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action= $args{$_}) or $args{$_} !~ /$action_values/) {
                $self->_set_err("Error: Volume::wafl_abort() Invalid value for argument '-action'",
                                "Legal values are: ".join(",",split("|",$action_values)));
                return 0;
            }
            next;
        }
        else {
            $self->_set_err("Error: Volume::wafl_abort() Invalid parameter specified \'$_\'");
            return 0;
        }
    }

    unless (defined $action and length($action)){
          $self->_set_err("Error: Volume::wafl_sboty() Missing required parameter '$action'",
                          "Legal values are: ".join(",",split("|",$action_values)));
          return 0;
    }

    if (not defined $scanid){
        @ids=$self->_get_wafl_scanids($action);
    }
    else {
        push @ids,$id;
    }

    foreach my $id (@ids){
        $cmd = "wafl scan abort ". $id;
        @output=$self->{parent}->execute($cmd,-hide=>0);
        if ($self->{parent}->error()){
            $self->_set_err($self->{parent}->errmsg(),@output);
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}
=cut

################################################################################
# Subroutine Name: wafl_scan()
# Objective:       perform wafl scan action
# Details:         See POD documentation at the end of this file
################################################################################
sub wafl_scan {
    my ($self, %args) = @_;
    my ($action, $action_values, $cmd, $parent, $scanid, @output, @scanids,);

    $self->_clear_err();

    # intialize variables
    $parent        = $self->get_parent();
    $scanid        = 0;
    $action_values = "abort|measure_layout|reallocate|i2p|fold";

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})
                or $args{$_} !~ /$action_values/) {
                $self->_set_err(
                    "Error: Volume::wafl_scan() Invalid value for argument '-action'",
                    "Legal values are: " . join(",", split("|", $action_values))
                    );
                return 0;
            }
            next;
        } elsif (/-?scanid/) {
            if (not length($scanid = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::wafl_scan Invalid value for argument '-scanid'"
                    );
                return 0;
            }
            next;
        } else {
            $self->_set_err(
                 "Error: Volume::wafl_scan() Invalid parameter specified \'$_\'"
                 );
            return 0;
        }
    } ## end foreach (keys %args)

    unless (defined $action and length($action)) {
        $self->_set_err(
              "Error: Volume::wafl_scan() Missing required parameter '$action'",
              "Legal values are: " . join(",", split("|", $action_values)));
        return 0;
    }

    $cmd = "priv set advanced; wafl scan $action ";
    $cmd .= ($action =~ /abort/) ? $scanid : $self->get_name();

    @output = $self->{parent}->execute($cmd, -hide => 0);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return 0;
    }

    if ($action =~ /abort/) {
        return 1;
    } else {
        @scanids = $self->_get_wafl_scanids($action);

        if (@scanids) {
            $scanid = $scanids[scalar(@scanids) - 1];
        } else {
            $self->_set_err(
                 "Error: Volume::wafl_scan() Unable to retrieve latest scanid");
            return 0;
        }
    } ## end else [ if ($action =~ /abort/)
    return $scanid;
} ## end sub wafl_scan

###############################################################################
# Method:    clone_create()
# Objective: Create a FlexClone of the specified volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clone_create {
    my ($self, %args) = @_;
    my $vol_name  = $self->get_name();
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->interface();
    my $resultsHash_ref;
    my $clone_volume_obj;

    # If Clone name is not specified, then append the parent volume
    # name with "_flexclone"
    my $clone_name =
      (defined $args{-name}) ? $args{-name} : "flexclone_" . $vol_name;

    $self->_clear_err();

    # Check if parent volume exist
    $self->{parent}->vol_exist($vol_name);
    if ($self->{parent}->error) {

        # Parent volume does not exist
        $self->_set_err(
                    "Parent volume does not exist" . $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # Check if clone volume already exist
    $self->{parent}->vol_exist($clone_name);
    if (!$self->{parent}->error) {

        # Parent volume does not exist
        $self->_set_err("Clone volume(FlexClone) already exists"
                        . $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    } else {

        # Clone volume does not exist, so create it...
        $resultsHash_ref =
          $cmd_obj->execute_cmd(-cmd             => "volume-clone-create",
                                -cmd_filer       => $self->{filer},
                                -check_output    => 1,
                                -check_result    => 1,
                                -cmd_interface   => $interface,
                                "-parent-volume" => $vol_name,
                                -volume          => $clone_name);

    } ## end else [ if (!$self->{parent}->error)

    # log reslts
    $self->{parent}->log_cmd_results($resultsHash_ref);

    if ($cmd_obj->error()) {
        $self->_set_err("FlexClone creation failed " . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Obtain clone object from the clone volume name
    $clone_volume_obj = $self->{parent}->get_vol($clone_name);
    if ($self->{parent}->error) {

        # Clone volume object cannot be obtained
        $self->_set_err(
               "Cannot obtain clone volume object" . $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $clone_volume_obj;
} ## end sub clone_create

################################################################################
# Subroutine Name: create_qtree
# Objective:       create and return a qtree object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_qtree {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $QtreeObj;

    $QtreeObj = Qtree->_new(-parent => $self,
                            -filer  => $self->{filer},
                            %args);

    if ($QtreeObj->error) {
        $self->_set_err($QtreeObj->errmsg());
        return $self;
    }

    #
    # Add the Qtree object created to our list of cached qtrees
    #
    push @{$self->{qtrees}}, $QtreeObj;

    return $QtreeObj;
} ## end sub create_qtree

###############################################################################
# Method:    delete_qtree()
# Objective: Delete a qtree in the volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete_qtree {
    my ($self, %args) = @_;
    my $name;
    my $force;
    my $vol_name  = $self->get_name();
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->interface();
    my $output;
    my $resultsHash_ref;
    my %cmd_args;

    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force and $force =~ /1|t|on|true|y|yes/i) {
                if ($interface eq "zapi") {
                    $cmd_args{"-force"} = 1;
                } else {
                    $cmd_args{"-force"} = "";
                }
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Build the command line and execute the command
    %cmd_args = (-cmd           => "qtree-delete",
                 -cmd_filer     => $self->{filer},
                 -check_result  => 1,
                 -cmd_interface => $interface,);

    if (lc($interface) eq "cli") {
        $cmd_args{"-cmd_privilege"} = "test";
        $cmd_args{"-path"}          = "/vol/$vol_name/$name";
    } else {
        $cmd_args{"-qtree"} = "/vol/$vol_name/$name";
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # log reslts
    $self->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # check for command failure
    if (lc($interface) eq "zapi") {

        # If interface is zapi, convert the output from xml to a string.
        my $output_hash = $self->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);

        if (defined $output[0] and length $output[0]) {
            if ($output[0] =~ /qtree:/i) {
                $self->_set_err("CLI failed: @output");
                return $ClientUtil::FALSE;
            }
        }
    } ## end else [ if (lc($interface) eq ...
    return $ClientUtil::TRUE;
}    # end delete_qtree()

################################################################################
# Subroutine Name: qtree_exist()
# Objective:       Determines if qtree exists
# Details:         See POD documentation at the end of this file
################################################################################
sub qtree_exist {
    my ($self, $name) = @_;
    my ($cmd, $out, @output,);

    unless (length $name) {
        $self->_set_err(  "Error: Volume::qtree_exist() "
                        . "Required parameter \$name required");
        return $ClientUtil::FALSE;
    }

    $cmd = "qtree status " . $self->get_name();

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # join output
    $out = join "\n", @output;

    if ($out !~ /$name/i) {
        $self->_set_err("Error: Volume::qtree_exist() $name not found",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end qtree_exist()

################################################################################
# Subroutine Name: lun_list_info()
# Objective:       Get info from filer about all the luns and create a hash
#                  for each of the luns on this volume and return a hash
#                  of these hashes keyed by the luns path
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_list_info {
    my ($self, %args) = @_;
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->interface();
    my $vol_name  = $self->get_name();
    my $resultsHash_ref;
    my $info_hash;
    my $entry_hash;
    my $lun_hash = ();

    $self->sanlog(
        -msg =>
          "Volume::lun_list_info() Retrieve via $interface lun info for volume $vol_name",
        -msg_level => 1);

    #
    # Create the list of arguments for the CmdsDriver call.
    # Note that -verbose is really for the cli and should just
    # be ignored for the zapi
    #
    my %cmd_args = ("-cmd"           => "lun-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 0,
                    "-verbose"       => "",);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to list lun info: $errmsg");
        $self->sanlog(-msg => $self->errmsg());
        return {};
    }

    $self->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Unable to get lun info: result check failed");
        return {};
    }

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If we get here, we have results back from the filer.  If it's a zapi
    # we want to make the XML a hash and look at the "luns" entry -
    #   if there are no luns, there will be no lun-info entry
    #   if there is one lun, lun-info will be a hash
    #   if there are >1 luns, lun-info will be an array
    #
    # If it's cli, we want to parse the data line by line and fill in the
    # fields of the entry_hash matching the field names from the zapi
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->xml2hash($output);

        #
        # if there are no luns, there will be no lun-info entry
        # return an empty hash
        #
        unless (defined $output_hash->{"luns"}->{"lun-info"}) {
            return ();
        }

        #
        # if there is one lun, lun-info will be a hash
        # if there are >1 luns, lun-info will be an array
        #
        if (ref($output_hash->{"luns"}->{"lun-info"}) eq "HASH") {
            $entry_hash = $output_hash->{"luns"}->{"lun-info"};
            $entry_hash->{name} = basename($entry_hash->{path});

            #
            # Check to see if the lun is on this volume
            #
            if ($entry_hash->{path} =~ /\/vol\/$vol_name\/.*/) {
                $lun_hash->{$entry_hash->{path}} = $entry_hash;
                return (%{$lun_hash});
            } else {
                return ();
            }
        } else {
            foreach $entry_hash (@{$output_hash->{'luns'}->{'lun-info'}}) {

                #
                # Check to see if the lun is on this volume
                #
                if ($entry_hash->{path} =~ /\/vol\/$vol_name\/.*/) {
                    $entry_hash->{name} = basename($entry_hash->{path});
                    $lun_hash->{$entry_hash->{path}} = $entry_hash;
                }
            } ## end foreach $entry_hash (@{$output_hash...
            return (%{$lun_hash});
        } ## end else [ if (ref($output_hash->...
    } else {

        #
        # CLI case
        # We are expecting output that looks like this:
        # /vol/blocks_fvt/cloned_lun 8m (8388608) (r/w, offline, mapped, shared)
        #       Serial#: hofBSJMIuVXV
        #       Backed by: /vol/blocks_fvt/.snapshot/clonesnap/mappedonline_lun
        #       Share: none
        #       Space Reservation: enabled
        #       Multiprotocol Type: solaris
        #
        my @output_array = split(/\n/, $output);
        $entry_hash = {};
        my $lun_path = "";
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace

            #
            # If this line doesn't pertain to this volume, skip it
            #
            if (    ($line !~ /^\/vol/)
                and (!defined($lun_hash->{$lun_path}))) {
                next;
            }

            #
            # This will be the first line in a new lun
            #
            if ($line =~ /^\/vol/) {

                #
                # Regex below parses the first line shown above in the example.
                # Using that example:
                # $new_path = /vol/blocks_fvt/cloned_lun
                # $size = 8m, $access = r/w, $state = offline, mapped, shared
                #
                # my ($new_path,$size,$unit, $access,$state)=
                my ($new_path, $size, $unit, $bytes, $access, $state) = $line =~
                  /^\s*(.*)\s+(\d+.?\d{0,})(\w)\s+\((.*)\)\s+\((\w+\/\w+),\s+(.*)\)/;

                $new_path =~ s/\s+$//;    # remove trailing whitespace
                $self->{parent}->sanlog(
                                      -msg => "Volume::lun_list_info(): $line\n"
                                        . "new path is $new_path,\n"
                                        . "size     is $size$unit,\n"
                                        . "bytes    is $bytes,\n"
                                        . "access   is $access,\n"
                                        . "state    is $state\n",
                                      -msg_level => 2);

                $lun_path = $new_path;

                #
                # If this line doesn't pertain to this volume, skip it
                #
                if ($lun_path !~ /\/vol\/$vol_name\/.*/) {
                    next;
                }

                #
                # Check the $state for mapped, online
                # 1 for mapped, otherwise 0.
                #
                my $mapped   = ($state  =~ /mapped/) ? "true"  : "false";
                my $online   = ($state  =~ /online/) ? "true"  : "false";
                my $readonly = ($access =~ /w/)      ? "false" : "true";

                #
                # Load the lun hash with what we know
                #
                $lun_hash->{$lun_path}->{path}        = $new_path;
                $lun_hash->{$lun_path}->{size}        = $bytes;
                $lun_hash->{$lun_path}->{online}      = $online;
                $lun_hash->{$lun_path}->{mapped}      = $mapped;
                $lun_hash->{$lun_path}->{"read-only"} = $readonly;
            } elsif ($line =~ /^Serial/) {
                $line =~ s/Serial#:\s+//;    # strip off the prefix
                $lun_hash->{$lun_path}->{"serial-number"} = $line;
            } elsif ($line =~ /^Share:/) {
                $line =~ s/Share:\s+//;
                $lun_hash->{$lun_path}->{"share-state"} = $line;
            } elsif ($line =~ /^Space/) {

                #
                # If the line contains "enabled" then use true
                #
                $lun_hash->{$lun_path}->{"is-space-reservation-enabled"} =
                  ($line =~ /enabled/) ? "true" : "false";
            } elsif ($line =~ /^Multiprotocol/) {
                $line =~ s/Multiprotocol\s+Type:\s+//;
                $lun_hash->{$lun_path}->{"multiprotocol-type"} = $line;
            } elsif ($line =~ /^Backed/) {
                $line =~ s/Backed\s+by:\s+//;
                $lun_hash->{$lun_path}->{"backing-snapshot"} = $line;
            } elsif ($line =~ /^Maps/) {
                $line =~ s/Maps:\s+//;
                $lun_hash->{$lun_path}->{maps} = $line;
            }
        }    # end foreach line
        return (%{$lun_hash});
    }    # end if CLI
}    # end lun_list_info()

################################################################################
# Subroutine Name: wait_for_online()
# Objective:       Waits for the volume to be in an online state
# Details:         See POD documentation at the end of this file
################################################################################
sub wait_for_online {
    my ($self, %args) = @_;
    my $cmd_obj  = $self->{parent}->cmd_obj();
    my $vol_name = $self->get_name();
    my $vol_info = "";

    $self->_clear_err();

    #
    # If we were given an interval or timeout use them, otherwise use defaults
    # Note 10800 is 3 hours
    #
    my $timeout  = (defined $args{-timeout})  ? $args{-timeout}  : 10800;
    my $interval = (defined $args{-interval}) ? $args{-interval} : 120;

    my $start_time   = int(gettimeofday);
    my $current_time = int(gettimeofday);

    while (($current_time - $start_time) < $timeout) {

        #
        # Get the volmue info - get_info calls volume-list-info zapi
        #
        $vol_info = $self->get_info();

        if ($self->error()) {
            my $errmsg = join "\n", $self->errmsg();
            $self->_set_err("Unalble to get volume info:" . $errmsg);
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg_level => 3,
                      -msg       => "Volume::wait_for_online(): "
                        . "\nCurrent time is "
                        . $current_time
                        . "\nStart time is   "
                        . $start_time
                        . "\nTimeout is      "
                        . $timeout
                        . "\nOutput is       "
                        . $vol_info);

        #
        # If the vol is online, we are done, return.
        #
        if ($vol_info =~ /<state>online<\/state>/) {
            return $ClientUtil::TRUE;
        }

        &SanUtil::nate_sleep($interval);
        $current_time = int(gettimeofday);
    }    # end while

    $self->_set_err("Timeout waiting for volume '$vol_name' to be online");
    return $ClientUtil::FALSE;
}    # end wait_for_online

#############################################################################
# Method:    move_start()
# Objective: starts the vol move with the options available
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_start {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($aggr);

    # Initialize some variables
    my $num_cutover       = 0;
    my $cutover_window    = 300;
    my $override          = 0;
    my $wait_flag         = 0;
    my $max_wait_time     = 1200;
    my $wait_time         = 30;
    my $manual_cutover    = 0;
    my $keep_source       = 0;
    my $log_output        = 0;
    my $filer_obj         = $self->{parent};
    my $interface         = $filer_obj->{interface};
    my $cmd_obj           = $filer_obj->cmd_obj();
    my $volume            = $self->get_name();
    my $pause_retry_count = 5;

    # Parse the arguments
    foreach (keys %args) {
        if (/-?aggr/) {
            unless (length($aggr = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-aggr'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?wait_flag/) {
            unless (length($wait_flag = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-wait_flag'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?interval/) {
            unless (length($wait_time = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-interval'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?max_wait/) {
            unless (length($max_wait_time = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-max_wait'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?keep_source/) {
            unless (length($keep_source = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-keep_source'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?num_cutover/) {
            unless (length($num_cutover = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-num_cutover'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?cutover_window/) {
            unless (length($cutover_window = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-cutover_window'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?manual_cutover/) {
            unless (length($manual_cutover = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-manual_cutover'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?override/) {
            unless (length($override = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-override'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?pause_retry_count/) {
            unless (length($pause_retry_count = $args{$_}) > 0) {
                $self->_set_err(
                             "Invalid value for argument '-pause_retry_count'");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameters specified for '$_'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # check whether destination aggregate is passed or not
    unless (defined($aggr) and $aggr ne "") {
        $self->_set_err("Required destination aggregate name is missing");
        return $ClientUtil::FALSE;
    }

    # Check to see if we have both manual cutover and cutover attempts specified.
    # If we do, set an error and abort.
    if ($manual_cutover > 0 and $num_cutover > 0) {
        $self->_set_err(
            "'-manual_cutover' and '-num_cutover' are mutually exclusive options"
            );
        return $ClientUtil::FALSE;
    }

    # Build the command driver
    my %cmd_args = (-cmd             => "volume-move-start",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,
                    "-dest-aggr"     => $aggr,);

    if ($interface eq "cli") {
        if ($keep_source == 1) {
            $cmd_args{"-is-keep-source"} = "";
        }
        if ($manual_cutover == 1) {
            $cmd_args{"-is-manual-cutover"} = "";
        }
        if ($override == 1) {
            $cmd_args{"-is-override-warnings"} = "";
        }
    } else {
        if ($keep_source == 1) {
            $cmd_args{"-is-keep-source"} = 1;
        }
        if ($manual_cutover == 1) {
            $cmd_args{"-is-manual-cutover"} = 1;
        }
        if ($override == 1) {
            $cmd_args{"-is-override-warnings"} = 1;
        }
    } ## end else [ if ($interface eq "cli")
    if ($num_cutover > 0) {
        $cmd_args{"-cutover-attempts"} = $num_cutover;
    }
    if ($cutover_window > 0) {
        $cmd_args{"-cutover-window"} = $cutover_window;
    }

    $self->sanlog(-comment =>
                  "NDVM Job Parameters - Volume: $volume  Destination:  $aggr");

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
                  "Unable to run cmd 'vol-move-start':\n" . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {
        $self->_set_err(
                   "Unable to start vol move for volume ${volume}:\n${output}");
        return $ClientUtil::FALSE;
    }

    # For non-cli interfaces, check the command status to see if it outright failed.
    # For cli interfaces, we need to look at the output to see if it has a message
    # we aren't expecting.
    if ($interface ne "cli") {

        # Convert the output to xml
        my $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg;
            if (defined($output_hash->{reason})) {
                $errmsg = $output_hash->{reason};
            } else {
                $errmsg = "No reason found";
            }
            $self->_set_err(
                "Error: Unable to start vol move for volume ${volume}:\n$errmsg"
                );
            return $ClientUtil::FALSE;
        } ## end if (defined($output_hash...
    } ## end if ($interface ne "cli")

    # Grab the vol move status
    my $status_hash = $self->move_status();
    if ($self->error()) {
        my $errmsg = join "\n", $self->errmsg();
        $self->_set_err(
            "Error: Unable to get vol move status for volume ${volume}:\n$errmsg"
            );
        return $ClientUtil::FALSE;
    }

    # Make sure the source volume is found in the output
    if (defined($status_hash->{'migration-state'})
        and $status_hash->{'migration-state'} eq "running") {

        # Volume has a state of "running"
        $self->sanlog(-msg => "Vol move started for volume $volume");
    } else {

        # Volume didn't have a migration state of running.
        $self->_set_err("Unable to start vol move for volume $volume");
        return $ClientUtil::FALSE;
    }

    if ($wait_flag == 1) {
        $self->wait_for_move_complete(-aggr              => $aggr,
                                      -interval          => $wait_time,
                                      -max_wait          => $max_wait_time,
                                      -pause_retry_count => $pause_retry_count);
        if ($self->error()) {
            $self->_set_err("Unable complete migration of volume $volume:\n"
                            . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    } ## end if ($wait_flag == 1)

    return $ClientUtil::TRUE;
} ## end sub move_start

#############################################################################
# Method:    move_status()
# Objective: get the status of the vol move operation
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_status {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@result, @result1, @required);

    # Initialize some variables
    my $output_hash = {};
    my $log_output  = 0;
    my $verbose     = 0;
    my $filer_obj   = $self->{parent};
    my $interface   = $filer_obj->{interface};
    my $cmd_obj     = $filer_obj->cmd_obj();
    my $volume      = $self->get_name();

    # Parse the arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            unless (length($verbose = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-verbose'");
                return {};
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return {};
        }
    }    # end foreach

    my $srcaggr = $self->{aggr};
    if (not defined($srcaggr) or $srcaggr eq "") {
        $self->_set_err("Unable get aggregate of volume ${volume}");
        return {};
    }

    # Build the command driver
    my %cmd_args = (-cmd             => "volume-move-status",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,);

    if ($verbose == 1) {
        if ($interface eq "cli") {
            $cmd_args{"-is-verbose"} = "";
        } else {
            $cmd_args{"-is-verbose"} = 1;
        }
    }

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
                 "Unable to run cmd 'vol-move-status':\n" . $cmd_obj->errmsg());
        return {};
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {

        # We need to check to see if the volume just isn't being moved.  The output
        # check will still fail but it isn't technically a failure.
        if ($output =~ /specified source volume is not being moved/i) {
            $self->sanlog(-msg => "Migration is not running for volume $volume",
                          -msg_level => 3);
            $output_hash->{'migration-state'} = "not_running";
            return $output_hash;
        } else {
            $self->_set_err(
                   "Unable to start vol move for volume ${volume}:\n${output}");
            return $ClientUtil::FALSE;
        }
    } ## end unless ($resultsHash_ref->...

    if ($interface eq "cli") {

        # Construct the output hash
        if ($verbose == 1) {

            #
            # constructing the hash which matches the zapi
            # $output_hash->{
            #             'migration-state => 'running',
            #             'vol-move-status-info' => {
            #                         'cutover-attempts' => '6',
            #                         'move-state' => ' setup (paused)',
            #                         'cutover-window' => '88',
            #                         'last-transfer-size' => '0',
            #                         'Last-transfer-duration' => '0',
            #                         'temp-volume' => 'ndm_dstvol_1267950718',
            #                         'dest-aggregate' => 'destaggr',
            #                         'source-volume-name' => 'NdvmSrcVol1',
            #                         'src-aggregate' => 'testaggr',
            #             }
            # }
            @result = split(/\n/, $output);
            $output_hash->{'migration-state'} = "running";
            $output_hash->{'vol-move-status-info'}->{'src-aggregate'} =
              $srcaggr;

            foreach my $line (@result) {
                chomp($line);

                # Remove any leading or trailing white space
                $line =~ s/^\s+|\s+$//g;

                # Output from cli version of "vol move status" will look like the following
                # once we remove leading and trailing white space
                #
                # Source:                 ndvm_vol3
                # Destination:            nate_aggr:ndm_dstvol_1272405084
                # State:                  move
                # Cutover Attempts:       3
                # Cutover Time:           60
                # Last Completed Transfer:
                # Data Transferred = 1 MB         Time Taken = 1 s
                # Current Transfer Size = 0 KB
                #
                if ($line =~ /^Source\:\s+(\S+)/) {
                    $output_hash->{'vol-move-status-info'}
                      ->{'source-volume-name'} = $1;
                } elsif ($line =~ /^Destination\:\s+(\S+)\:(\S+)/) {
                    $output_hash->{'vol-move-status-info'}->{'dest-aggregate'} =
                      $1;
                    $output_hash->{'vol-move-status-info'}->{'temp-volume'} =
                      $2;
                } elsif ($line =~ /^State\:\s+(\S+)/) {
                    $output_hash->{'vol-move-status-info'}->{'move-state'} = $1;
                } elsif ($line =~ /^Cutover Attempts\:\s+(\S+)/) {
                    $output_hash->{'vol-move-status-info'}
                      ->{'cutover-attempts'} = $1;
                } elsif ($line =~ /^Cutover Time\:\s+(\d+)/) {
                    $output_hash->{'vol-move-status-info'}->{'cutover-window'} =
                      $1;
                } elsif ($line =~
                    /^Data Transferred \= (\S+)\s+\S+\s+Time Taken = (\d+)\s+\S+/
                  ) {
                    $output_hash->{'vol-move-status-info'}
                      ->{'last-transfer-size'} = $1;
                    $output_hash->{'vol-move-status-info'}
                      ->{'last-transfer-duration'} = $2;
                } elsif ($line =~ /^Current Transfer \= (\d+)\s+\S+/) {
                    $output_hash->{'vol-move-status-info'}
                      ->{'current-transfer'} = $1;
                }
            } ## end foreach my $line (@result)
        } else {

            # Verbose is not 1 so we have different output to parse
            if ($output !~ /$volume/) {

                # We didn't find the volume in the output
                $self->sanlog(
                           -msg => "Migration is not running on volume $volume",
                           -msg_level => 3);
                $output_hash->{'migration-state'} = "not_running";
                return $output_hash;
            }

            # constructing the hash which matches the zapi
            # this is how hash looks like
            # $output_hash->{
            #             'migration-state => 'running',
            #             'vol-move-status-info' => {
            #                         'cutover-attempts' => '6',
            #                         'move-state' => ' setup (paused)',
            #                         'cutover-window' => '88',
            #                         'dest-aggregate' => 'destaggr',
            #                         'source-volume-name' => 'NdvmSrcVol1',
            #                         'src-aggregate' => 'testaggr',
            #             }
            # }
            #
            # Output from cli version of "vol move status" will look like the following
            #
            # Source                Destination                     CO Attempts    CO Time     State
            # ndvm_vol3             nate_aggr2                      3              60          cutover
            my @result = split(/\n/, $output);
            foreach my $line (@result) {
                chomp($line);

                # Remove any leading or trailing white space
                $line =~ s/^\s+|\s+$//g;

                # Skip the header line
                next if ($line =~ /^Source\s+Destination/);

                # Now we can split what is left.
                my ($source, $destination, $attempts, $window, @state) =
                  split(/\s+/, $line);
                $output_hash->{'migration-state'} = "running";
                $output_hash->{'vol-move-status-info'}->{'src-aggregate'} =
                  $srcaggr;
                $output_hash->{'vol-move-status-info'}->{'source-volume-name'} =
                  $source;
                $output_hash->{'vol-move-status-info'}->{'dest-aggregate'} =
                  $destination;
                $output_hash->{'vol-move-status-info'}->{'cutover-attempts'} =
                  $attempts;
                $output_hash->{'vol-move-status-info'}->{'cutover-window'} =
                  $window;
                $output_hash->{'vol-move-status-info'}->{'move-state'} =
                  join("", @state);
            } ## end foreach my $line (@result)
        } ## end else [ if ($verbose == 1)
    } else {
        $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg;
            if (defined($output_hash->{reason})
                and $output_hash->{reason} ne "") {
                if ($output_hash->{reason} =~
                    /specified source volume is not being moved/i) {

                    # Command succeeded...the volume just isn't being moved. This is not an
                    # error so set the state to not_running and return the output hash
                    $self->sanlog(
                          -msg => "Migration is not running for volume $volume",
                          -msg_level => 3);
                    $output_hash->{'migration-state'} = "not_running";
                } else {

                    # Command actually had other errors so set the error
                    $self->_set_err(
                        "Error: Unable to get vol move status for volume $volume:\n"
                          . $output_hash->{reason});
                    $output_hash->{'migration-state'} = "not_running";
                }
            } else {
                $self->_set_err(
                    "Error: Unable to get vol move status for volume $volume:\n"
                    );
                $output_hash->{'migration-state'} = "not_running";
            }
            return $output_hash;
        } ## end if (defined($output_hash...

        # The command didn't fail so grab the vol-move-status-info out of the status array
        # and copy it to it's own key.
        if (defined($output_hash->{'status'}->[1]->{'vol-move-status-info'})) {
            $output_hash->{'vol-move-status-info'} =
              $output_hash->{'status'}->[1]->{'vol-move-status-info'};
        }

        $output_hash->{'vol-move-status-info'}->{'src-aggregate'} = $srcaggr;
        $output_hash->{'migration-state'} = "running";

        unless (
              defined(
                  $output_hash->{'vol-move-status-info'}->{'source-volume-name'}
              )
              and $output_hash->{'vol-move-status-info'}->{'source-volume-name'}
              eq $volume
          ) {
            $self->sanlog(-msg => "Migration is not running on volume $volume",
                          -msg_level => 3);
            $output_hash->{'migration-state'} = "not_running";
            return $output_hash;
        } ## end unless (defined($output_hash...
    } ## end else [ if ($interface eq "cli")

    return $output_hash;
} ## end sub move_status

#############################################################################
# Method:    move_pause()
# Objective: pause the vol move operation
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_pause {
    my ($self) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $log_output = 0;
    my $filer_obj  = $self->{parent};
    my $interface  = $filer_obj->{interface};
    my $cmd_obj    = $filer_obj->cmd_obj();
    my $volume     = $self->get_name();

    # Build command driver and Execute the command
    my %cmd_args = (-cmd             => "volume-move-pause",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,);

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
                  "Unable to run cmd 'vol-move-pause':\n" . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {
        $self->_set_err(
                   "Unable to pause vol move for volume ${volume}:\n${output}");
        return $ClientUtil::FALSE;
    }

    if ($interface ne "cli") {
        my $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg;
            if (defined($output_hash->{status})) {
                $errmsg = $output_hash->{reason};
            } else {
                $errmsg = "No reason found";
            }
            $self->_set_err(
                           "Error: Unable to pause volume ${volume}:/n$errmsg");
            return $ClientUtil::FALSE;
        } ## end if (defined($output_hash...
    } ## end if ($interface ne "cli")

    $self->sanlog(-msg => "vol move pause succeeded for volume $volume");

    return $ClientUtil::TRUE;
} ## end sub move_pause

#############################################################################
# Method:    move_resume()
# Objective: resume the vol move operation
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_resume {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($aggr);

    # Initialize some variables
    my $num_cutover    = 0;
    my $cutover_window = 300;
    my $override       = 0;
    my $wait_flag      = 0;
    my $manual_cutover = 0;
    my $keep_source    = 0;
    my $log_output     = 0;
    my $wait_time      = 30;
    my $max_wait_time  = 1200;
    my $filer_obj      = $self->{parent};
    my $interface      = $filer_obj->{interface};
    my $cmd_obj        = $filer_obj->cmd_obj();
    my $volume         = $self->get_name();

    # Parse the arguments
    foreach (keys %args) {
        if (/-?aggr/) {
            unless (length($aggr = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-aggr'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?wait_flag/) {
            unless (length($wait_flag = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-wait_flag'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?interval/) {
            unless (length($wait_time = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-interval'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?max_wait/) {
            unless (length($max_wait_time = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-max_wait'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?keep_source/) {
            unless (length($keep_source = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-keep_source'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?num_cutover/) {
            unless (length($num_cutover = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-num_cutover'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?cutover_window/) {
            unless (length($cutover_window = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-cutover_window'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?manual_cutover/) {
            unless (length($manual_cutover = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-manual_cutover'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?override/) {
            unless (length($override = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-override'");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameters specified for '$_'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Check to see if we have both manual cutover and cutover attempts specified.
    # If we do, set an error and abort.
    if ($manual_cutover > 0 and $num_cutover > 0) {
        $self->_set_err(
            "'-manual_cutover' and '-num_cutover' are mutually exclusive options"
            );
        return $ClientUtil::FALSE;
    }

    # Build the command driver
    my %cmd_args = (-cmd             => "volume-move-resume",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,);

    if ($interface eq "cli") {
        if ($keep_source == 1) {
            $cmd_args{"-is-keep-source"} = "";
        }
        if ($manual_cutover == 1) {
            $cmd_args{"-is-manual-cutover"} = "";
        }
        if ($override == 1) {
            $cmd_args{"-is-override-warnings"} = "";
        }
    } else {
        if ($keep_source == 1) {
            $cmd_args{"-is-keep-source"} = 1;
        }
        if ($manual_cutover == 1) {
            $cmd_args{"-is-manual-cutover"} = 1;
        }
        if ($override == 1) {
            $cmd_args{"-is-override-warnings"} = 1;
        }
    } ## end else [ if ($interface eq "cli")
    if ($num_cutover > 0) {
        $cmd_args{"-cutover-attempts"} = $num_cutover;
    }
    if ($cutover_window > 0) {
        $cmd_args{"-cutover-window"} = $cutover_window;
    }

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
                 "Unable to run cmd 'vol-move-resume':\n" . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {
        $self->_set_err(
                  "Unable to resume vol move for volume ${volume}:\n${output}");
        return $ClientUtil::FALSE;
    }

    if ($interface ne "cli") {
        my $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg;
            if (defined($output_hash->{reason})) {
                $errmsg = $output_hash->{reason};
            } else {
                $errmsg = "No reason found";
            }
            $self->_set_err(
                     "Unable to resume vol move on volume ${volume}:/n$errmsg");
            return $ClientUtil::FALSE;
        } ## end if (defined($output_hash...
    } ## end if ($interface ne "cli")

    $self->sanlog(-msg => "Migration Resumed for volume $volume");

    if ($wait_flag == 1) {
        $self->wait_for_move_complete(-aggr     => $aggr,
                                      -interval => $wait_time,
                                      -max_wait => $max_wait_time);
        if ($self->error()) {
            $self->_set_err("Unable complete migration of volume ${volume}:\n"
                            . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    } ## end if ($wait_flag == 1)

    return $ClientUtil::TRUE;
} ## end sub move_resume

#############################################################################
# Method:    move_abort()
# Objective: abort the vol move operation
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_abort {
    my ($self) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $wait_time  = 10;
    my $log_output = 0;
    my $filer_obj  = $self->{parent};
    my $interface  = $filer_obj->{interface};
    my $cmd_obj    = $filer_obj->cmd_obj();
    my $volume     = $self->get_name();

    # Build command driver
    my %cmd_args = (-cmd             => "volume-move-abort",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,);

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
                  "Unable to run cmd 'vol-move-abort':\n" . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {
        $self->_set_err(
                   "Unable to abort vol move for volume ${volume}:\n${output}");
        return $ClientUtil::FALSE;
    }

    if ($interface ne "cli") {
        my $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(
                "Error: Unable to abort vol move for volume ${volume}:\n$errmsg"
                );
            return $ClientUtil::FALSE;
        }
    } ## end if ($interface ne "cli")

    $self->sanlog(-msg => "vol move abort succeeded for volume $volume");

    return $ClientUtil::TRUE;
} ## end sub move_abort

#############################################################################
# Method:    move_cutover()
# Objective: force cutover of the vol move operation
# Details:   See POD documentation at the end of this file
#############################################################################
sub move_cutover {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Initialize some variables
    my $cutover_window = 300;
    my $log_output     = 0;
    my $filer_obj      = $self->{parent};
    my $interface      = $filer_obj->{interface};
    my $cmd_obj        = $filer_obj->cmd_obj();
    my $volume         = $self->get_name();

    # Parse the arguments
    foreach (keys %args) {
        if (/-?cutover_window/) {
            unless (length($cutover_window = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-cutover_window'");
                return $ClientUtil::FALSE;
            }
        } else {

            # invalid parameters
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Build the command driver
    my %cmd_args = (-cmd             => "volume-move-cutover",
                    -cmd_filer       => $filer_obj->get_name(),
                    -cmd_interface   => $interface,
                    -cmd_log_output  => $log_output,
                    -check_output    => 1,
                    -check_result    => 1,
                    "-source-volume" => $volume,);

    # Set the -cutover-window attribute if a cutover window was specified as an argument
    if ($cutover_window > 0) {
        $cmd_args{"-cutover-window"} = $cutover_window;
    }

    # Execute the command
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err(
             "Unable to run cmd 'volume-move-cutover':\n" . $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    my $output = $resultsHash_ref->{output};

    # Check the results and output
    unless (    $resultsHash_ref->{result_check}
            and $resultsHash_ref->{output_check}) {
        $self->_set_err(
               "Unable to perform cutover of for volume ${volume}:\n${output}");
        return $ClientUtil::FALSE;
    }

    if ($interface ne "cli") {
        my $output_hash = $filer_obj->xml2hash($output);
        if (defined($output_hash->{status})
            and $output_hash->{status} eq "failed") {
            my $errmsg = $output_hash->{reason};
            $self->_set_err(
                "Error: Unable to issue manual cutover for volume ${volume}:\n$errmsg"
                );
            return $ClientUtil::FALSE;
        }
    } ## end if ($interface ne "cli")

    return $ClientUtil::TRUE;
} ## end sub move_cutover

#############################################################################
# Method:    wait_for_move_complete()
# Objective: Waits for migration completes and returns successful or not
# Details:   See POD documentation at the end of this file
#############################################################################
sub wait_for_move_complete {
    my ($self, %args) = @_;

    my ($aggr);

    $self->_clear_err();

    # Initialize some variables
    my $wait_flag            = 1;
    my $wait_time            = 30;
    my $max_wait_time        = 1200;
    my $pause_retry_count    = 5;
    my $pause_retry_attempts = 0;
    my $volume               = $self->get_name();

    # Parse the arguments
    foreach (keys %args) {
        if (/-?aggr/) {
            unless (length($aggr = $args{$_}) > 0) {
                $self->_set_err("Invalid value for argument '-aggr'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?interval/) {
            unless (length($wait_time = $args{$_}) > 0 and $wait_time > 0) {
                $self->_set_err("Invalid value for argument '-interval'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?max_wait/) {
            unless (length($max_wait_time = $args{$_}) > 0
                    and $max_wait_time >= 0) {
                $self->_set_err("Invalid value for argument '-max_wait'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?pause_retry_count/) {
            unless (length($pause_retry_count = $args{$_}) > 0) {
                $self->_set_err(
                             "Invalid value for argument '-pause_retry_count'");
                return $ClientUtil::FALSE;
            }
        } else {

            # invalid parameters
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Check the required parameters
    unless (defined($aggr) and $aggr ne "") {
        $self->_set_err("Missing required parameter '-aggr'");
        return $ClientUtil::FALSE;
    }
    my $status_hash = $self->move_status();
    if (   $status_hash->{'vol-move-status-info'}->{'dest-aggregate'} ne $aggr
        && $status_hash->{'migration-state'} eq "running") {
        $self->_set_err("Volume: $volume is not moving to aggr: $aggr");
        return $ClientUtil::FALSE;
    }

    # Keep waiting on the move
    my $current_wait_time = 0;
    while ($wait_flag) {
        my $status_hash = $self->move_status();
        if ($self->error()) {
            $self->_set_err(
                         "Unable to get vol move status for volume ${volume}:\n"
                           . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # Grab the move state and migration state
        my $move_state      = "";
        my $migration_state = "";
        if (defined($status_hash->{'vol-move-status-info'}->{'move-state'})
            and $status_hash->{'vol-move-status-info'}->{'move-state'} ne "") {
            $move_state =
              $status_hash->{'vol-move-status-info'}->{'move-state'};
        }
        if (defined($status_hash->{'migration-state'})
            and $status_hash->{'migration-state'} ne "") {
            $migration_state = $status_hash->{'migration-state'};
        }

        # Check the move state for paused
        if ($move_state =~ /paused/) {
            if (    $pause_retry_count > 0
                and $pause_retry_attempts <= $pause_retry_count) {
                $self->sanlog(
                             -msg => "Volume move for $volume has been paused. "
                               . "Attempting to resume volume move (Total "
                               . "resume attempts: $pause_retry_count, "
                               . "Remaining resume attempts: "
                               . ($pause_retry_count - $pause_retry_attempts)
                               . ")");
                $pause_retry_attempts++;
                $self->move_resume(-aggr => $aggr);
                if ($self->error()) {
                    $self->_set_err("Attempted to resume volume $volume. "
                                    . $self->errmsg());
                    return $ClientUtil::FALSE;
                }
                $self->sanlog(-msg =>
                         "Sleeping for 60 seconds before checking move status");
                &SanUtil::nate_sleep(60);
            } else {
                $self->_set_err(
                       "The migration of volume $volume is in the pause state");
                return $ClientUtil::FALSE;
            }
        } elsif ($migration_state eq "running") {

            # Migration is running.  If we have a status, print it in the message.
            if ($move_state ne "") {
                $self->sanlog(  -msg => "vol move is in progress for volume"
                              . "'$volume' with a move state of "
                              . "'$move_state'");
            } else {
                $self->sanlog(
                          -msg => "vol move is in progress for volume $volume");
            }
        } else {

            # In order to work around burt 448491 in which "vol move status" incorrectly reports
            # back that there is no move in progress during the final stages of cutover, we are
            # going to add in a 60 second sleep prior to doing the aggr check
            $self->sanlog(
                  -msg => "Vol move completed successfully for volume $volume");
            &SanUtil::nate_sleep(60);

            # Get the destination and current aggrs and compare them.
            $self->refresh(-aggr => 1, -luns => 0);
            if ($self->error) {
                $self->_set_err(
                           "Unable to refresh aggregates:\n" . $self->errmsg());
                return $ClientUtil::FALSE;
            }
            my $curraggr = $self->get_aggr(-command_retries => 5);
            if ($self->error) {
                $self->_set_err(
                           "Unable to get the current aggr for volume $volume");
                return $ClientUtil::FALSE;
            }
            if ($curraggr eq $aggr) {
                $self->sanlog(-msg =>
                          "Vol move completed successfully for volume $volume");
                return $ClientUtil::TRUE;
            }
        } ## end else [ if ($move_state =~ /paused/)

        # Sleep the wait period
        &SanUtil::nate_sleep($wait_time);

        # Check to see if $max_wait_time is 0.  If it is 0, then we will wait
        # forever.  If it is > 0, then we wil increment the time.
        if ($max_wait_time > 0) {

            # Increment $current_wait_time by the $wait_time interval
            $current_wait_time += $wait_time;

            # Check to see if we have exceeded our max wait time
            if ($current_wait_time >= $max_wait_time) {
                $self->_set_err(
                          "Max wait time of '$max_wait_time' seconds exceeded");
                return $ClientUtil::FALSE;
            }
        } ## end if ($max_wait_time > 0)
    } ## end while ($wait_flag)
    return $ClientUtil::TRUE;
} ## end sub wait_for_move_complete

###############################################################################
# Method:    volume_vvlabel()
# Objective: Get FSID from vvol raid label fields
# Details:   This is cli only.
# See POD documentation at the end of this file.
###############################################################################

sub volume_vvlabel {
    my ($self)   = @_;
    my $vol_name = $self->get_name();
    my $cmd_obj  = $self->{parent}->cmd_obj();
    my $output;
    my $errmsg;
    my $resultsHash_ref;
    my $fsid;

    # Build the command line and execute the command
    my %cmd_args = (-cmd           => "volume-vvlabel",
                    -cmd_filer     => $self->{filer},
                    -cmd_interface => "cli",
                    -check_result  => 1,
                    -volume        => "$vol_name");

    if (lc($self->{parent}->{interface}) eq "cli") {
        $cmd_args{"-cmd_privilege"} = "test";
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # log results
    $self->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        $self->_set_err("Unable to list vol info:", $cmd_obj->errmsg());
        $self->sanlog(-msg => $self->errmsg());
    }

    # get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    my @output = split("\n", $output);

    # O/p of the vvlabel command
    # rtpqa-fas3050-1*> vol vvlabel stib_vol1
    #                Volume name stib_vol1
    #                      Magic 0x56565244
    #                   Version 22
    #              Buftree type 1
    #               Mirror type 0
    #                     Flags 0x1
    #                      Fsid 33424916
    #                    PathId 0
    #                ID in aggr 3596
    # ID of clone parent in  aggr 3596
    #               Base snapid 0
    # Clone parent's buftree id 0
    #               Volume UUID abebf5ce-6652-11df-95cd-00a09802439e
    #             Ordinal claim 1275865248
    #             Ordinal value 1
    #            Duplicate FSID 0
    #              Nominal size 1310720
    #                  Checksum 0x0
    #               Volume type 0x0
    #            Guarantee type 0x2
    #            CP at creation 2
    #  Upgrade scanner last fbn 1456351
    #     Upgrade scanner flags 0x2
    #          Inherited blocks 0
    #          Container fileid 3596
    #         Raid label fileid 3597
    #      Sparse config fileid 0

    # check for error
    if (defined $output[0] and length $output[0]) {
        if ($output[0] =~ /Volume vvlabel:/) {
            $self->_set_err("CLI failed: ", @output);
            return $ClientUtil::FALSE;
        }
    }

    # parse the output

    for (my $ind = 0 ; $ind < scalar(@output) ; $ind++) {
        if ($output[$ind] =~ /Fsid\s(\d+)/) {
            $fsid = $1;
        }
    }

    return $fsid;
} ## end sub volume_vvlabel

#############################################################################
# Method:             vol_quota_on()
# Objective:          set quota on for the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_on() {
    my ($self) = @_;
    my ($resultsHash_ref, $output_hash, $output);

    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since this is a non-san feature , only zapi was implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "quota-on",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj-> ...
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_on()
############################################################################
# Method:             vol_quota_off()
# Objective:          Set quota off for the volume
# Details:            See POD documentation at the end of this file
############################################################################
sub vol_quota_off() {
    my ($self) = @_;
    my ($resultsHash_ref, $output_hash, $output);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is non-san feature only zapi was implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "quota-off",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface = "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface = "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_off()
#############################################################################
# Method:    vol_quota_resize()
# Objective: resizes the quota for the volume
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_quota_resize() {
    my ($self) = @_;
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is a non-san feature only zapi was implemented.
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "quota-resize",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -cmd_interface => "zapi",
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_resize()
#############################################################################
# Method:             vol_quota_status()
# Objective:          Displays quota status for the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_status() {
    my ($self) = @_;
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is a non-san feature it was implemented in zapi
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "quota-status",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $output_hash;    # returning an empty hash reference
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returning an empty hash reference
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_quota_status()
#############################################################################
# Method:             vol_quota_get_entry()
# Objective:          Get quotas details entry information
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_get_entry() {
    my ($self, %args) = @_;
    my ($qtree, $quota_type, $quota_target, $output, $resultsHash_ref);
    my $output_hash;
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is a non-san feature only zapi is implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?qtree/) {
            if (not length($qtree = $args{$_})) {
                $self->_set_err("Invalid value for argument '-qtree'");
                return $output_hash;    # returning an empty hash reference
            }
            next;
        } elsif (/-?quota_type/) {
            if (not length($quota_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_type'");
                return $output_hash;    # returning an empty hash reference
            }
            next;
        } elsif (/-?quota_target/) {
            if (not length($quota_target = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_target'");
                return $output_hash;    # returning an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $output_hash;        # returning an empty hash reference
        }
    }    # end foreach (keys %args)

    # input parameter validation
    if (!defined($quota_target)) {
        $self->_set_err("quota_target is required argument");
        return $ClientUtil::FALSE;
    }
    if (!defined($quota_type)) {
        $self->_set_err("quota_type is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "quota-get-entry",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    "-quota-target" => $quota_target,
                    "-quota-type"   => $quota_type,
                    -volume         => $self->get_name(),);
    if ($qtree) {
        $cmd_args{"-qtree"} = $qtree;
    } else {
        $cmd_args{"-qtree"} = "";
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $output_hash;    # returning an empty hash reference
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returning an empty hash reference
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_quota_get_entry()
#############################################################################
# Method:         vol_quota_set_entry()
# Objective:      set quota entries for volume
# Details:        See POD documentation at the end of this file
#############################################################################
sub vol_quota_set_entry() {
    my ($self, %args) = @_;
    my ($qtree,     $quota_type, $quota_target,    $disk_limit);
    my ($threshold, $file_limit, $soft_file_limit, $soft_disk_limit);
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is a non-san feature only zapi is implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?qtree/) {
            if (not length($qtree = $args{$_})) {
                $self->_set_err("Invalid value for argument '-qtree'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_type/) {
            if (not length($quota_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_target/) {
            if (not length($quota_target = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_target'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?disk_limit/) {
            if (not length($disk_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?file_limit/) {
            if (not length($file_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_disk_limit/) {
            if (not length($soft_disk_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_file_limit/) {
            if (not length($soft_file_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?threshold/) {
            if (not length($threshold = $args{$_})) {
                $self->_set_err("Invalid value for argument '-threshold'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         # imput parameter validation
    if (!defined($quota_target)) {
        $self->_set_err("quota_target is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "quota-set-entry",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    -"quota-target" => $quota_target,
                    -"quota-type"   => $quota_type,
                    -volume         => $self->get_name(),);
    if ($qtree) {
        $cmd_args{"-qtree"} = $qtree;
    } else {
        $cmd_args{"-qtree"} = "";
    }    # end if ($qtree)
    $cmd_args{"-disk-limit"}      = $disk_limit      if ($disk_limit);
    $cmd_args{"-file-limit"}      = $file_limit      if ($file_limit);
    $cmd_args{"-soft-disk-limit"} = $soft_disk_limit if ($soft_disk_limit);
    $cmd_args{"-soft-file-limit"} = $soft_file_limit if ($soft_file_limit);
    $cmd_args{"-threshold"}       = $threshold       if ($threshold);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_set_entry()
#############################################################################
# Method:             vol_quota_report()
# Objective:          Displays the detial report of quota for volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_report() {
    my ($self) = @_;
    my ($output, $resultsHash_ref, $dumpOut, $output_hash);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # Since it is a non-san function only zapi is implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "quota-report",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $volume,
                    -path          => $path,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $output_hash;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_quota_report()
#############################################################################
# Method:         vol_quota_list_entries()
# Objective:      lists the available entries for a quota on the volume
# Details:        See POD documentation at the end of this file
#############################################################################
sub vol_quota_list_entries() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $include_output_entry, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # Since it is non-san function hence only zapi is implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/^-?include_output_entry/) {
            if (not length($include_output_entry = $args{$_})) {
                $self->_set_err(
                          "Invalid value for argument '-include_output_entry'");
                return $output_hash;    # returning an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $output_hash;        # returning an empty hash reference
        }
    }    # end foreach (keys %args)

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd                    => "quota-list-entries",
                    -cmd_filer              => $self->get_filername(),
                    -check_result           => 1,
                    -cmd_interface          => $interface,
                    "-include-output-entry" => $include_output_entry,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $output_hash;    # returning an empty hash reference
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returning an empty hash reference
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_quota_list_entries()
#############################################################################
# Method:         vol_quota_add_entry()
# Objective:      adds a entry toa quota on the volume
# Details:        See POD documentation at the end of this file
#############################################################################
sub vol_quota_add_entry() {
    my ($self, %args) = @_;
    my ($qtree,     $quota_type, $quota_target,    $disk_limit);
    my ($threshold, $file_limit, $soft_file_limit, $soft_disk_limit);
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is a non-san function only zapi is implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?qtree/) {
            if (not length($qtree = $args{$_})) {
                $self->_set_err("Invalid value for argument '-qtree'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_type/) {
            if (not length($quota_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_target/) {
            if (not length($quota_target = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_target'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?disk_limit/) {
            if (not length($disk_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?file_limit/) {
            if (not length($file_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_disk_limit/) {
            if (not length($soft_disk_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_file_limit/) {
            if (not length($soft_file_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?threshold/) {
            if (not length($threshold = $args{$_})) {
                $self->_set_err("Invalid value for argument '-threshold'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)

    if (!defined($quota_target)) {
        $self->_set_err("quota_target is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "quota-add-entry",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    -"quota-target" => $quota_target,
                    -"quota-type"   => $quota_type,
                    -volume         => $self->get_name(),);
    if ($qtree) {
        $cmd_args{"-qtree"} = $qtree;
    } else {
        $cmd_args{"-qtree"} = "";
    }    # end if ($qtree)
    $cmd_args{"-disk-limit"}      = $disk_limit      if ($disk_limit);
    $cmd_args{"-file-limit"}      = $file_limit      if ($file_limit);
    $cmd_args{"-soft-disk-limit"} = $soft_disk_limit if ($soft_disk_limit);
    $cmd_args{"-soft-file-limit"} = $soft_file_limit if ($soft_file_limit);
    $cmd_args{"-threshold"}       = $threshold       if ($threshold);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         #Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_add_entry()
#############################################################################
# Method:             vol_quota_modify_entry()
# Objective:          Modifies a entry of quota on the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_modify_entry() {
    my ($self, %args) = @_;
    my ($qtree,     $quota_type, $quota_target,    $disk_limit);
    my ($threshold, $file_limit, $soft_file_limit, $soft_disk_limit);
    my ($output, $resultsHash_ref, $output_hash);

    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is a non-san function only zapi was implemented
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?qtree/) {
            if (not length($qtree = $args{$_})) {
                $self->_set_err("Invalid value for argument '-qtree'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_type/) {
            if (not length($quota_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_target/) {
            if (not length($quota_target = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_target'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?disk_limit/) {
            if (not length($disk_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?file_limit/) {
            if (not length($file_limit = $args{$_})) {
                $self->_set_err("Invalid value for argument '-file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_disk_limit/) {
            if (not length($soft_disk_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_disk_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?soft_file_limit/) {
            if (not length($soft_file_limit = $args{$_})) {
                $self->_set_err(
                               "Invalid value for argument '-soft_file_limit'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?threshold/) {
            if (not length($threshold = $args{$_})) {
                $self->_set_err("Invalid value for argument '-threshold'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    if (!defined($quota_target)) {
        $self->_set_err("quota_target is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "quota-modify-entry",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    -"quota-target" => $quota_target,
                    -"quota-type"   => $quota_type,
                    -volume         => $self->get_name(),);
    if ($qtree) {
        $cmd_args{"-qtree"} = $qtree;
    } else {
        $cmd_args{"-qtree"} = "";
    }    # end if ($qtree)
    $cmd_args{"-disk-limit"}      = $disk_limit      if ($disk_limit);
    $cmd_args{"-file-limit"}      = $file_limit      if ($file_limit);
    $cmd_args{"-soft-disk-limit"} = $soft_disk_limit if ($soft_disk_limit);
    $cmd_args{"-soft-file-limit"} = $soft_file_limit if ($soft_file_limit);
    $cmd_args{"-threshold"}       = $threshold       if ($threshold);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_modify_entry()
#############################################################################
# Method:             vol_quota_delete_entry()
# Objective:          Deletes a quota entry  on the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_quota_delete_entry() {
    my ($self, %args) = @_;
    my ($qtree,  $quota_type,      $quota_target);
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san feature it is only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?qtree/) {
            if (not length($qtree = $args{$_})) {
                $self->_set_err("Invalid value for argument '-qtree'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_type/) {
            if (not length($quota_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?quota_target/) {
            if (not length($quota_target = $args{$_})) {
                $self->_set_err("Invalid value for argument '-quota_target'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)

    if (!defined($quota_target)) {
        $self->_set_err("quota_target is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "quota-delete-entry",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    -"quota-target" => $quota_target,
                    -"quota-type"   => $quota_type,
                    -volume         => $self->get_name(),);
    if ($qtree) {
        $cmd_args{"-qtree"} = $qtree;
    } else {
        $cmd_args{"-qtree"} = "";
    }    # end if ($qtree)

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_quota_delete_entry()
#############################################################################
# Method:             vol_reallocate_on()
# Objective:          Set reallocate on for the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_on() {
    my ($self) = @_;
    my ($result, $output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "reallocate-on",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed:", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    #end sub vol_reallocate_on()
#############################################################################
# Method:             vol_reallocate_off()
# Objective:          set reallocate off for the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_off() {
    my ($self) = @_;
    my ($output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "reallocate-off",
                    -cmd_filer     => $self->get_filername(),
                    -check_output  => 1,
                    -check_result  => 1,
                    -check_result  => 1,
                    -cmd_interface => $interface,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    #Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }
    return $ClientUtil::TRUE;
}    ### end sub vol_reallocate_off()
#############################################################################
# Method:             vol_reallocate_start()
# Objective:          starts reallocate on the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_start() {
    my ($self, %args) = @_;
    my ($threshold, $run_once, $preserve_logical_bno, $is_aggregate, $full);
    my ($output, $resultsHash_ref, $output_hash, $interval, $no_check);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # since it is non-san function only zapi implementation
    my $interface = "zapi";

    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/^-?full/) {
            if (not length($full = $args{$_})) {
                $self->_set_err("Invalid value for argument '-full'");
                return $ClientUtil::FALSE;
            }
            if (defined $full and $full =~ /YES|TRUE|1/i) {
                $full = 1;
            } else {
                $full = 0;
            }
            next;
        } elsif (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interval'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?is_aggregate/) {
            if (not length($is_aggregate = $args{$_})) {
                $self->_set_err("Invalid value for argument '-is_aggregate'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?no_check/) {
            if (not length($no_check = $args{$_})) {
                $self->_set_err("Invalid value for argument '-no_check'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?preserve_logical_bno/) {
            if (not length($preserve_logical_bno = $args{$_})) {
                $self->_set_err(
                          "Invalid value for argument '-preserve_logical_bno'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?run_once/) {
            if (not length($run_once = $args{$_})) {
                $self->_set_err("Invalid value for argument '-run_once'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?threshold/) {
            if (not length($threshold = $args{$_})) {
                $self->_set_err("Invalid value for argument '-threshold'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         #
         # Build the command line and execute the command
         #
    my %cmd_args = (-cmd           => "reallocate-start",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,);
    $cmd_args{"-full"}         = $full         if ($full);
    $cmd_args{"-interval"}     = $interval     if ($interval);
    $cmd_args{"-is-aggregate"} = $is_aggregate if ($is_aggregate);
    $cmd_args{"-no-check"}     = $no_check     if ($no_check);
    $cmd_args{"-preserve-logical_bno"} = $preserve_logical_bno
      if ($preserve_logical_bno);
    $cmd_args{"-run-once"}  = $run_once  if ($run_once);
    $cmd_args{"-threshold"} = $threshold if ($threshold);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_reallocate_start()
#############################################################################
# Method:             vol_reallocate_stop()
# Objective:          Stops reallocate for the volume
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_stop() {
    my ($self) = @_;
    my ($resultsHash_ref, $output_hash, $output);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "reallocate-stop",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    #end sub vol_reallocate_stop()
#############################################################################
# Method:           vol_reallocate_quiesce()
# Objective:        Quiesce the reallocate process for the volume
# Details:          See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_quiesce() {
    my ($self) = @_;
    my ($resultsHash_ref, $output_hash, $output);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "reallocate-quiesce",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_reallocate_quiesce()
#############################################################################
# Method:           vol_reallocate_restart()
# Objective:        Restarts the reallocate on the volume
# Details:          See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_restart() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $output_hash, $ignore_checkpoint);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?ignore_checkpoint/) {
            if (not length($ignore_checkpoint = $args{$_})) {
                $self->_set_err(
                             "Invalid value for argument '-ignore_checkpoint'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         #
         # Build the command line and execute the command
         #
    my %cmd_args = (-cmd           => "reallocate-restart",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,);
    $cmd_args{"-ignore-checkpoint"} = $ignore_checkpoint
      if ($ignore_checkpoint);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
         # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_reallocate_restart()
#############################################################################
# Method:    vol_reallocate_list_info()
# Objective: lists the all the reallocate information for the volume
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_reallocate_list_info() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $output_hash, $verbose, $output_hash);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $path    = "/vol/" . $volume;

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Invalid value for argument '-verbose'");
                return $output_hash;    # returining an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $output_hash;        # returining an empty hash reference
        }
    } ## end foreach (keys %args)
        #
        # Build the command line and execute the command
        #
    my %cmd_args = (-cmd           => "reallocate-list-info",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,);
    $cmd_args{"-path"}    = $path    if ($path);
    $cmd_args{"-verbose"} = $verbose if ($verbose);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $output_hash;    # returining an empty hash reference
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returining an empty hash reference
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a hash.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_reallocate_list_info()
#############################################################################
# Method:             vol_snapshots_list_info()
# Objective:          Lists the detailed information of the volume snapshots.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_snapshots_list_info() {
    my ($self, %args) = @_;
    my ($snapowners, $target_name,     $target_type, $terse);
    my ($output,     $resultsHash_ref, $qtree,       $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?snapowners/) {
            if (not length($snapowners = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapowners'");
                return $output_hash;    # returining an empty hash reference
            }
            next;
        } elsif (/-?target_name/) {
            if (not length($target_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-target_name'");
                return $output_hash;    # returining an empty hash reference
            }
            next;
        } elsif (/-?target_type/) {
            if (not length($target_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-target_type'");
                return $output_hash;    # returining an empty hash reference
            }
            next;
        } elsif (/-?terse/) {
            if (not length($terse = $args{$_})) {
                $self->_set_err("Invalid value for argument '-terse'");
                return $output_hash;    # returining an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $output_hash;        # returining an empty hash reference
        }
    } ## end foreach (keys %args)

    if (!defined($target_name)) {
        $self->_set_err("target_name is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "snapshot-list-info",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $self->get_name(),);
    $cmd_args{"-snapowners"}  = $snapowners  if ($snapowners);
    $cmd_args{"-target-type"} = $target_type if ($target_type);
    $cmd_args{"-target-name"} = $target_name if ($target_name);
    $cmd_args{"-terse"}       = $terse       if ($terse);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $output_hash;    # returining an empty hash reference
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returining an empty hash reference
    }    # end unless

    # Since Interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $output_hash;
        }    # end if ($output_hash-> ...
    }
    return $output_hash;
}    ### end sub vol_snapshots_list_info()
#############################################################################
# Method:    vol_snapshot_rename()
# Objective: renames the snapshot name
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_snapshot_rename() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $output_hash, $new_name, $current_name);
    my $interface = $self->{parent}->{interface};
    my $cmd_obj   = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?new_name/) {
            if (not length($new_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-new_name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?current_name/) {
            if (not length($current_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-current_name'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "snapshot-rename",
                    -cmd_filer      => $self->get_filername(),
                    -check_result   => 1,
                    -check_output   => 1,
                    -cmd_interface  => $interface,
                    "-new-name"     => $new_name,
                    "-current-name" => $current_name,
                    -volume         => $self->get_name(),);

    if (!defined($new_name)) {
        $self->_set_err("new_name is required argument");
        return $ClientUtil::FALSE;
    }
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    if ($interface eq "zapi") {

        # Since Interface is zapi, convert the output from xml to a hash.
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }
    return $ClientUtil::TRUE;
}    ### end sub vol_snapshot_rename()
#############################################################################
# Method:    vol_snapshot_set_reserve()
# Objective: set reserve for the volume snapshot
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_snapshot_set_reserve() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $output_hash, $percentage);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?percentage/) {
            if (not length($percentage = $args{$_})) {
                $self->_set_err("Invalid value for argument '-percentage'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)
    if (!defined($percentage)) {
        $self->_set_err("percentage is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "snapshot-set-reserve",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -percentage    => $percentage,
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    if ($interface eq "zapi") {

        # Since Interface is zapi, convert the output from xml to a hash.
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($output_hash-> ...
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_snapshot_set_reserve()
#############################################################################
# Method:    vol_snapshot_reclaimable_info()
# Objective: Displays the volume snapshots reclaimable information
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_snapshot_reclaimable_info() {
    my ($self, %args) = @_;
    my (@snapshot_name, @results);
    my ($resultsHash_ref, $reclaimable_size, $output);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only cli implementation
    my $interface = "cli";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?snapshot_name/) {
            @snapshot_name = @{$args{$_}};
            if (ref(@snapshot_name) ne /ARRAY/) {
                $self->_set_err(  "Invalid value for \"-snapshot_name\":"
                                . "@snapshot_name Need to pass ARRAY ref "
                                . "snapshot_names");
                return $reclaimable_size;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $reclaimable_size;
        }
    }    # end foreach (keys %args)

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "snapshot-reclaimable-info",
                    -cmd_filer     => $self->get_filername(),
                    -cmd_interface => $interface,
                    -snapshots     => "@snapshot_name",
                    -volume        => $self->get_name(),);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $reclaimable_size;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};
    @results = split(/\n/, $output);
    foreach my $line (@results) {
        if ($line =~
            /snap reclaimable: Approximately (\d+) Kbytes would be freed./) {
            $reclaimable_size = $1;
        }    # end if ($line ...
    }    # end foreach

    return $reclaimable_size;
}    ### end sub vol_snapshot_reclaimable_info()
#############################################################################
# Method:    vol_snapshot_delta_info()
# Objective: displays the volume snapshot delta information
# Details:   See POD documentation at the end of this file
#############################################################################
sub vol_snapshot_delta_info() {
    my ($self, %args) = @_;
    my ($snapshot1,     $snapshot2,    $resultsHash_ref, $output_hash);
    my ($consumed_size, $elapsed_time, $output,          $delta_info);
    my @results;
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->{interface};
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?snapshot1/) {
            if (not length($snapshot1 = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot1'");
                return $delta_info;    # returning an empty hash reference
            }
            next;
        } elsif (/-?snapshot2/) {
            if (not length($snapshot2 = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot2'");
                return $delta_info;    # returning an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $delta_info;        # returning an empty hash reference
        }
    } ## end foreach (keys %args)
                                       #input parameter validation
    if (!defined($snapshot1)) {
        $self->_set_err("snapshot1 is required argument");
        return $ClientUtil::FALSE;
    }
    if (!defined($snapshot2)) {
        $self->_set_err("snapshot2 is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd           => "snapshot-delta-info",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $self->get_name(),
                    -snapshot1     => $snapshot1,);
    $cmd_args{"-snapshot2"} = $snapshot2 if ($snapshot2);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $delta_info;    # returning an empty hash reference
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $delta_info;    # returning an empty hash reference
    }    # end unless

    if ($interface eq "zapi") {

        # Since Interface is zapi, convert the output from xml to a string.
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $delta_info;    # returning an empty hash reference
        }    # end if ($output_hash-> ...
        if (defined $output_hash->{'consumed-size'}) {
            $consumed_size = $output_hash->{'consume-size'};
        } else {
            $self->_set_err("Unable to get consumed-size via zapi");
            return $delta_info;    # returning an empty hash reference
        }    # end if (defined $output_hash-> ...

        if (defined $output_hash->{'elapsed-time'}) {
            $elapsed_time = $output_hash->{'elapsed-time'};
        } else {
            $self->_set_err("Unable to get elapsed-time via zapi");
            return $delta_info;    # returning an empty hash reference
        }    # end if (defined $output_hash-> ...
    } else {
        @results = split(/\n/, $output);
        if ($snapshot2) {

            # ndvm_snap1      ndvm_snap2           88                  0s   -
            foreach my $line (@results) {
                if ($line =~ /$snapshot1\s+$snapshot2\s+(\d+)\s+(\w+)\s+/) {
                    $consumed_size = $1;
                    $elapsed_time  = $2;
                }    # end if ($line ...
            }    # end foreach
        } else {

            # ndvm_snap1      Active File System   172      1s   619200.000
            foreach my $line (@results) {
                if ($line =~
                    /$snapshot1\s+\w+\s+\w+\s+\w+\s+(\d+)\s+(\w+)\s+\d+/) {
                    $consumed_size = $1;
                    $elapsed_time  = $2;
                }    # end if ($line ...
            }    # end foreach
        }    # end if ($snapshot2)
    } ## end else [ if ($interface eq "zapi")
    $delta_info->{'consumed_size'} = $consumed_size;
    $delta_info->{'elapsed_time'}  = $elapsed_time;
    return $delta_info;
}    ### end sub vol_snapshot_delta_info()
#############################################################################
# Subroutine Name:    vol_clone_split_estimate()
# Objective:          gives the estimated time required to clone.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_clone_split_estimate() {
    my ($self, %args) = @_;
    my ($resultsHash_ref, $output_hash, $output, $clone_volume);
    my $interface = "zapi";                       # Implemented only for zapi
    my $cmd_obj   = $self->{parent}->cmd_obj();

    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?clone_volume/) {
            if (not length($clone_volume = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_volume'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    my %cmd_args = (
        -cmd           => "volume-clone-split-estimate",
        -cmd_filer     => $self->get_filername(),
        -check_result  => 1,
        -check_output  => 1,
        -cmd_interface => $interface,
        -volume        => $clone_volume,

        #The name of the clone whose split usage is being estimated.
        );

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}    ### end sub vol_clone_split_estimate()
#############################################################################
# Subroutine Name: vol_clone_split_start()
# Objective:       The cloned volume will be splitted from its orginal volume.
# Details:         See POD documentation at the end of this file
#############################################################################
sub vol_clone_split_start() {
    my ($self, %args) = @_;
    my ($resultsHash_ref, $output_hash, $output, $clone_volume);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?clone_volume/) {
            if (not length($clone_volume = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_volume'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)

    my %cmd_args = (-cmd           => "volume-clone-split-start",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $clone_volume,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("ZAPI failed: ", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_clone_split_start()
#############################################################################
# Subroutine Name:    vol_clone_split_status()
# Objective:          gives the staus of vol vlone split.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_clone_split_status() {
    my ($self, %args) = @_;
    my ($clone_volume, $output, $resultsHash_ref, $output_hash);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?clone_volume/) {
            if (not length($clone_volume = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_volume'");
                return $output_hash;    # returning an empty hash reference
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $output_hash;        # returning an empty hash reference
        }
    }    # end foreach (keys %args)
    if (!defined($clone_volume)) {
        $self->_set_err("clone_volume is required argument");
        return $ClientUtil::FALSE;
    }
    my %cmd_args = (-cmd           => "volume-clone-split-status",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,);
    $cmd_args{"-volume"} = $clone_volume if ($clone_volume);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $output_hash;    # returning an empty hash reference
    }    # end if ($cmd_obj->error())
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $output_hash;    # returning an empty hash reference
    }    # end unless

    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Unable to get clone split status",
                            $output_hash->{reason});
            $self->sanlog(-msg => $output_hash->{reason});
            return $output_hash;    # returning an empty hash reference
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $output_hash;
}    ### end sub vol_clone_split_status()
#############################################################################
# Subroutine Name:    vol_clone_split_stop()
# Objective:          stops the vol clone split.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_clone_split_stop() {
    my ($self, %args) = @_;
    my ($resultsHash_ref, $output_hash, $output, $clone_volume);
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?clone_volume/) {
            if (not length($clone_volume = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_volume'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
    if (!defined($clone_volume)) {
        $self->_set_err("clone_volume is required argument");
        return $ClientUtil::FALSE;
    }
    my %cmd_args = (-cmd           => "volume-clone-split-stop",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -volume        => $clone_volume,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # Since we checked the result, lets see if our check was OK
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless

    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);

        if ($output_hash->{status} eq "failed") {
            $self->_set_err(
                 "ZAPI failed: vol_clone_split_stop " . $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_clone_split_stop()
#############################################################################
# Subroutine Name:    vol_file_create()
# Objective:          creates a file on the splecified volume.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_file_create() {
    my ($self, %args) = @_;
    my ($size, $numbers, $filename, $resultsHash_ref, $output);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only cli implementation
    my $interface = "cli";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/^-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-size'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?numbers/) {
            if (not length($numbers = $args{$_})) {
                $self->_set_err("Invalid value for argument '-numbers'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?filename/) {
            if (not length($filename = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filename'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         #input parameter validation
    if (!defined($size)) {
        $self->_set_err("size is required argument");
        return $ClientUtil::FALSE;
    }
    if (!defined($filename)) {
        $self->_set_err("filename is required argument");
        return $ClientUtil::FALSE;
    }
    my $path = "/vol/" . $volume . "/" . $filename;
    my %cmd_args = (-cmd           => "volume-mkfile",
                    -cmd_filer     => $self->get_filername(),
                    -cmd_interface => $interface,
                    -path          => $path,
                    -size          => $size,
                    -numbers       => $numbers,
                    -cmd_privilege => "diag",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    return $ClientUtil::TRUE;
}    ### end sub vol_file_create()
#############################################################################
# Subroutine Name:    vol_file_delete()
# Objective:          deletes the file.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_file_delete() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $output_hash, $filename);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?filename/) {
            if (not length($filename = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filename'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         # input parameter validation
    if (!defined($filename)) {
        $self->_set_err("filename is required argument");
        return $ClientUtil::FALSE;
    }
    my $path = "/vol/" . $volume . "/" . $filename;
    my %cmd_args = (-cmd           => "file-delete-file",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,
                    -cmd_privilege => "diag",);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())
    $output = $resultsHash_ref->{output};
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Unable to delete file ", $output_hash->{reason});
            $self->sanlog(-msg => $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_file_delete()
#############################################################################
# Subroutine Name:    vol_dir_create()
# Objective:          creates a directory.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_dir_create() {
    my ($self, %args) = @_;
    my ($dirname, $perm, $resultsHash_ref, $output_hash, $output);
    my $volume  = $self->get_name();
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?dirname/) {
            if (not length($dirname = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dirname'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?perm/) {
            if (not length($perm = $args{$_})) {
                $self->_set_err("Invalid value for argument '-perm'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)
         #input parameter validation
    if (!defined($dirname)) {
        $self->_set_err("dirname is required argument");
        return $ClientUtil::FALSE;
    }
    my $path = "/vol/" . $volume . "/" . $dirname;
    my %cmd_args = (-cmd           => "file-create-directory",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,
                    -perm          => $perm,
                    -cmd_privilege => "diag",);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    $output = $resultsHash_ref->{output};
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Zapi Failed : Unable to create Directory ",
                            $output_hash->{reason});
            $self->sanlog(-msg => $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_dir_create()
#############################################################################
# Subroutine Name:    vol_dir_delete()
# Objective:          deletes the directory.
# Details:            See POD documentation at the end of this file
#############################################################################
sub vol_dir_delete() {
    my ($self, %args) = @_;
    my ($dirname, $resultsHash_ref, $output_hash, $output);
    my $volume  = $self->{name};
    my $cmd_obj = $self->{parent}->cmd_obj();

    # since it is non-san function only zapi implementation
    my $interface = "zapi";
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?dirname/) {
            if (not length($dirname = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dirname'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach (keys %args)

    # input parameter validation
    if (!defined($dirname)) {
        $self->_set_err("dirname is required argument");
        return $ClientUtil::FALSE;
    }
    my $path = "/vol/" . $volume . "/" . $dirname;
    my %cmd_args = (-cmd           => "file-delete-directory",
                    -cmd_filer     => $self->get_filername(),
                    -check_result  => 1,
                    -check_output  => 1,
                    -cmd_interface => $interface,
                    -path          => $path,
                    -cmd_privilege => "diag",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }    # end if ($cmd_obj->error())

    $output = $resultsHash_ref->{output};
    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Zapi Failed : Unable to destroy directory",
                            $output_hash->{reason});
            $self->sanlog(-msg => $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if ($status eq "failed")
    }    # end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### end sub vol_dir_delete()
#############################################################################
# Subroutine Name:    snap_autodelete_options_set()
# Objective:          set options .
# Details:            See POD documentation at the end of this file
#############################################################################
sub snap_autodelete_options_set() {
    my ($self, %args) = @_;
    my ($option_name, $option_value, $resultsHash_ref, $output, @result,
        @options);
    my $vol_name  = $self->get_name();
    my $cmd_obj   = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->interface();
    $self->sanlog(-msg => $self);
    $self->_clear_err();
    @options = qw(
      commitment
      trigger
      target_free_space
      delete_order
      defer_delete
      prefix
      destroy_list
      );

    foreach my $option_name (keys %args) {
        $option_name =~ s/-//g;
        unless (scalar(grep { /^$option_name$/ } @options)) {
            $self->_set_err(
                "Error: Volume::snap_autodelete_options() Invalid parameter 'option_name'"
                );
            return $ClientUtil::FALSE;
        }
    }
    foreach my $option_name (keys %args) {
        $option_value = $args{$option_name};
        $option_name =~ s/-//g;

        # Build the command driver
        my %cmd_args = (
            -cmd            => "snapshot-autodelete-set-option",
            -cmd_filer      => $self->get_filername(),
            -cmd_interface  => $interface,
            "-option-name"  => $option_name,
            "-option-value" => $option_value,
            -volume         => $vol_name,

            );

        # Execute the command
        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

        # Checking is there error in command driver execution
        if ($cmd_obj->error()) {
            my $errmsg = join "\n", $cmd_obj->errmsg();
            $self->_set_err(
                        "failed to set value for option $option_name: $errmsg");
            $self->sanlog(-msg => $self->errmsg());
            return $ClientUtil::FALSE;
        }    # end if ($cmd_obj-> ...

        # Get the output
        $output = $resultsHash_ref->{output};

        if ($interface eq "cli") {
            push @result, split(/\n/, $output);
            foreach my $line (@result) {
                if ($line !~ /options set/) {
                    $self->_set_err("cli Failed : Unable to set option");
                    return $ClientUtil::FALSE;
                }    # end if ($line ...
            }
        } elsif ($interface eq "zapi") {
            my $output_hash = $self->{parent}->xml2hash($output);
            if ($output_hash->{status} eq "failed") {
                $self->_set_err("Zapi Failed : Unable to destroy directory",
                                $output_hash->{reason});
                $self->sanlog(-msg => $output_hash->{reason});
                return $ClientUtil::FALSE;
            }    # end if ($status eq "failed")
        }    # end if ($interface eq "zapi")
    } ## end foreach my $option_name (keys...
    return $ClientUtil::TRUE;
} ## end sub snap_autodelete_options_set()

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($aggr, $cmd,     $conn, $create, $disk_cnt,
        $size, $devices, $log,  $name,   $parent,
        $self, @output,  $space_reserve,);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::_new() Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::_new() Invalid value for argument '-parent'"
                    );
                return $self;
            }
            next;
        } elsif (/-?create/) {
            if (not length($create = $args{$_}) or $args{$_} !~ /1|0/i) {
                $self->_set_err(
                    "Error: Volume::_new() Invalid value for argument '-create'"
                    );
                return $self;
            }
            next;
        } elsif (/-?loghdl/) {
            if (not length($log = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::_new() Invalid value for argument '-loghdl'"
                    );
                return $self;
            }
            next;
        } elsif (/-?aggr/) {
            if (not length($aggr = $args{$_})) {
                $self->_set_err("Error: Volume::create_vol()",
                                "Invalid value for argument '-aggr'");
                return $self;
            }
            next;
        } elsif (/-?disk_cnt/) {
            if (not length($disk_cnt = $args{$_}) or $args{$_} < 2) {
                $self->_set_err("Error: Volume::create_vol()",
                                "Invalid value for argument '-disk_cnt'");
                return $self;
            }
            next;
        } elsif (/-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: Volume::create_vol()",
                                "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?space-reserve/) {
            if (not length($space_reserve = $args{$_})) {
                $self->_set_err("Error: Volume::create_vol()",
                                "Invalid value for argument '-space-reserve'");
                return $self;
            }
            next;
        } else {
            $self->_set_err("Error: Volume::create_vol()",
                            "Invalid parameter specified \'$_\'");
            return $self;
        }
    }    # end foreach args

    $self->{log} = $log if (defined $log);

    # check for required lun path
    unless (defined $name and length $name) {
        self->_set_err("Error: Volume::_new()",
                       "Required parameter is missing '-name'");
        return $self;
    }

    # check for filer parent object
    unless (defined $parent and length $parent) {
        $self->_set_err("Error: Volume::_new()",
                        "Required parameter parent object");
        return $self;
    }

    # inititialize object attributes
    $self->{name}   = $name;
    $self->{parent} = $parent;
    $self->{filer}  = $parent->get_name();
    $self->{luns}   = [];

    # if we were passed space-reserve, add to command line
    # invocation (with proper syntax).
    if (defined($space_reserve)) {
        $space_reserve = "-s $space_reserve ";
    }

    if ($create) {
        if (defined($aggr)) {
            $self->{aggr} = $aggr;
        } else {
            $self->get_aggr();
            $self->_clear_err();    # need this in case it was a trad vol
                                    # we don't want to fail unless they
                                    # try to get the aggregate of a trad vol
                                    # here we are just getting it if we can
        }

        if ($aggr) {
            $cmd =
                "vol create " 
              . $name
              . " -l en "
              . $space_reserve
              . $aggr . " "
              . $size;
        } else {
            $cmd =
              "vol create " . $name . " -l en " . $space_reserve . $disk_cnt;
            if ($size) {
                $cmd .= "@" . $size;
            }
        }

        @output = $self->{parent}->execute($cmd, -hide => 1);
        if ($aggr) {
            if (    "@output" !~ /has completed/i
                and "@output" !~ /has been initiated/i) {
                $self->_set_err(
                       "Error: Volume::_new(): Unable to create volume '$name'",
                       $output[0]);
                return $self;
            }
        }
    } else {
        if (not $self->{parent}->vol_exist($self->{name})) {
            $self->_set_err($self->{parent}->errmsg());
            return $self;
        }
        $self->get_state();
        $self->_create_luns();
    }

    return $self;
} ## end sub _new

################################################################################
# Subroutine Name: _create_luns
# Objective:       create lun objects
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_luns {
    my ($self) = @_;
    my ($access, $cmd,   $dir,     $hostname, $id,     $igroup,
        $last,   $lunid, $lunname, $lunObj,   $mapped, $path,
        $size,   $state, $status,  @output,   %hMap,);

    my $name = $self->{name};

    #
    # Create the lun Map (hMap hash of hash)
    # Call lun show -v to get lun information to parse.  Sample output:
    #
    #         /vol/blocks_fvt/ncmds_lun2/mapped_offline_shared_all_lun      8m (8388608)       (r/w, offline, mapped, shared)
    #                 Serial#: C4edaJLkAmSv
    #                 Share: write
    #                 Space Reservation: enabled
    #                 Multiprotocol Type: solaris
    #                 Maps: ncmds_lun_fcp_igroup=1
    #
    $cmd = "lun show -v ";
    if ($self->{parent}->is_cache_data_valid($cmd)) {

        # if this command is in the cache, get it from cache
        # NOTE: it's possible that this cache data might be invalid since
        # we never invalidate the cache after lun destroy and lun create commands
        @output = $self->{parent}->get_cache_data($cmd);
    } else {

        # otherwise invoke the command
        @output = $self->{parent}->execute($cmd, -hide => 1);
    }

    if ($self->{parent}->error) {
        $self->_set_err(
            "Error: Volume::_create_vols(): Unable to retrieve lun map information"
            );
        return $ClientUtil::FALSE;
    }

    # store the output in the cache for this command
    $self->{parent}->store_cache_data($cmd, @output);

    #
    # Parse each line of the lun show command
    #
    my $bMap = 0;
    foreach my $line (@output) {

        #
        # skip unless it's a lun on this volume or a Map line
        #
        next if ($line !~ /\/vol\/$name\/|Map/i);

        #
        # Ignore leading white space then grab the first field
        # into $value and the rest of the line into $data.  That
        # $value will either hold the path to the lun or "Maps:"
        #
        my ($value, $data) = $line =~ /\s+(\S+)\s+(.*)/;
        if ($value =~ /^\/vol/) {
            $last = $value;
            $bMap = 0;
        } elsif ($value =~ /map/i and not $bMap) {
            my @igrp_names = split(/\s+/, $data);

            #
            # Preserve hash entry order and create a hash entry keyed by
            # igroup name that holds the lun id
            #
            tie my %hsh, 'Tie::IxHash';
            foreach my $igrpname (@igrp_names) {
                my ($igrp, $id) = split(/=/, $igrpname);
                $hsh{$igrp} = $id;
            }

            #
            # Store that hash we just created into the hMap hash
            #
            $hMap{$last} = \%hsh;

            $bMap = 1;
        } ## end elsif ($value =~ /map/i and...
    } ## end foreach my $line (@output)

    #
    # Execute lun show and parse the output. Output should look like this:
    #
    #  /vol/blocks_fvt/defined1      8m (8388608)       (r/w, online)
    #  /vol/blocks_fvt/ncmds_lun2/defined2      8m (8388608)       (r/w, online)
    #  /vol/blocks_fvt/mapped_offline_shared_all_lun      8m (8388608)       (r/w, offline, mapped, shared)
    #
    # TBD, what are we getting here that we can't get from the lun
    # show -v call above? When we rewite this to use execute_cmd, also
    # rewrite to be a 1 pass algorithm
    #
    $cmd = "lun show";
    if ($self->{parent}->is_cache_data_valid($cmd)) {

        # if this command is in the cache, get it from cache
        # NOTE: it's possible that this cache data might be invalid since we
        # never invalidate the cache after lun destroy and lun create commands
        @output = $self->{parent}->get_cache_data($cmd);
    } else {

        # otherwise invoke the command
        @output = $self->{parent}->execute($cmd, -hide => 1);
    }

    if ($self->{parent}->error) {
        $self->_set_err(
             "Error: Volume::_create_vols(): Unable to retrieve lun information"
             );
        return $ClientUtil::FALSE;
    }

    # store the output in the cache for this command
    $self->{parent}->store_cache_data("$cmd", @output);

    foreach my $line (@output) {
        next if ($line !~ /\/$name\//);

        #
        # Regex below parses each of the lines shown above in the example.
        # Using that example and looking at the third line,
        # $path = /vol/blocks_fvt/mapped_offline_shared_all_lun
        # $size = 8m, $access = r/w, $state = offline, mapped, shared
        #
        my ($path, $size, $unit, $access, $state) = $line =~
          /^\s+(.*)\s+(\d+.?\d{0,})(\w)\s+\(.*\)\s+\((\w+\/\w+),\s+(\w+)/;

        $mapped = ($line =~ /mapped/) ? 1 : 0;
        $path =~ s/\s//g;    # remove whitespace
        $dir = dirname($path);
        ($lunname) = basename($path) =~ /(\S+)/;

        my $igroup = (keys %{$hMap{$path}})[0];
        my $id     = (values %{$hMap{$path}})[0];

        #
        # Create a lun object using the data we just gathered
        # Note this object will be stored by create_lun()
        #
        # TBD - when you rewrite this, make sure the type is
        #       sent.
        my $lunObj = $self->create_lun(-path   => $path,
                                       -dir    => $dir,
                                       -name   => $lunname,
                                       -id     => $id,
                                       -igroup => $igroup,
                                       -size   => $size . $unit,
                                       -access => $access,
                                       -mapped => $mapped,
                                       -state  => $state,
                                       -_exist => 1,
                                       -_maps  => $hMap{$path});

        return $ClientUtil::FALSE if ($self->error);
    }    # end foreach line from lun show

    return $ClientUtil::TRUE;
}    # end create_luns()

################################################################################
# Subroutine Name: vol_64bit_upgrade
# Objective:       Upgrades the specified 32bit vol to 64bit format
# Details:         See POD documentation at the end of this file
################################################################################
sub vol_64bit_upgrade {
    my ( $self, %args ) = @_;
    my ( $wait_for_completion, $mode, $time_out, $sample_interval )
        = ( 0, "normal", 3600, 30 );
    my $cmd_obj  = $self->{parent}->cmd_obj();
    my $vol_name = $self->get_name();
    $self->_clear_err();
    foreach ( keys %args ) {
        if (/-?wait_for_completion/) {
            if ( not length( $wait_for_completion = $args{$_} ) ) {
                $self->_set_err(
                    "Error: Volume::vol_64bit_upgrade(): Invalid value for argument '-wait_for_completion'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( not length( $wait_for_completion...

            # 'wait_for_completion' is not a param for the cli so verify
            # validity of the input here itself.
            unless ( $wait_for_completion == 1 || $wait_for_completion == 0 )
            {
                $self->_set_err(
                    "Error: Volume::vol_64bit_upgrade(): Invalid value for argument '-wait_for_completion'"
                );
                return $ClientUtil::FALSE;
            } ## end unless ( $wait_for_completion...
            next;
        } ## end if (/-?wait_for_completion/)
        if (/-?mode/) {
            if ( not length( $mode = $args{$_} ) ) {
                $self->_set_err(
                    "Error: Volume::vol_64bit_upgrade(): Invalid value for argument '-mode'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( not length( $mode...
            next;
        } ## end if (/-?mode/)
        if (/-?time_out/) {
            if ( ( $time_out = $args{$_} ) < 0 ) {
                $self->_set_err(
                    "Error: Volume::vol_64bit_upgrade(): Invalid value for argument '-time_out'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( ( $time_out = $args...
            next;
        } ## end if (/-?time_out/)
        if (/-?sample_interval/) {
            if ( ( $sample_interval = $args{$_} ) < 0 ) {
                $self->_set_err(
                    "Error: Volume::vol_64bit_upgrade(): Invalid value for argument '-sample_interval'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( ( $sample_interval...
            next;
        } else {
            $self->_set_err(
                "Error: Volume:vol_64bit_upgrade(): Invalid parameter "
                    . "specified \'$_\'" );
            return $ClientUtil::FALSE;
        } ## end else [ if (/-?sample_interval/)
    } ## end foreach ( keys %args )

    my $resultsHash_ref = $cmd_obj->execute_cmd(
        -cmd           => "vol-64bit-upgrade-start",
        -cmd_filer     => $self->get_filername(),
        -cmd_interface => "cli",
        -check_result  => 1,
        -vol           => $vol_name,
        -mode          => $mode,
        -cmd_privilege => "diag"
    );
    if ( $cmd_obj->error() ) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err( "Error: Volume:vol_64bit_upgrade(): Unable to "
                . "start vol upgrade: $errmsg" );
        return $ClientUtil::FALSE;
    } ## end if ( $cmd_obj->error())

    $self->sanlog( -msg => $resultsHash_ref );

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};
    my $expected_output = "Started 64-bit upgrade on volume $vol_name";
    if ( $output !~ /$expected_output/i ) {
        $self->_set_err(
            "Error: Volume:vol_64bit_upgrade(): Failed to start vol upgrade on '$vol_name'"
        );
        return $ClientUtil::FALSE;
    } ## end if ( $output !~ /$expected_output/i)

    unless ( $resultsHash_ref->{result_check} ) {
        $self->_set_err( "Error: Volume:vol_64bit_upgrade(): "
                . "Result check failed" );
        return $ClientUtil::FALSE;
    }

    # waiting for the scanner to start
    &SanUtil::nate_sleep(30);

    if ($wait_for_completion) {
        $self->wait_for_vol_upgrade_completion(
            -time_out        => $time_out,
            -sample_interval => $sample_interval
        );
        if ( $self->error() ) {
            return $ClientUtil::FALSE;
        }
    } ## end if ($wait_for_completion)

    return $ClientUtil::TRUE;
} ## end sub vol_64bit_upgrade

################################################################################
# Subroutine Name: get_vol_upgrade_status
# Objective:       Returns vol upgrade status
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_upgrade_status {
    my $self     = shift;
    my $vol_name = $self->get_name();
    $self->_clear_err();
    #Getting the status of aggregate 64-bit upgrade
    my $op = $self->{'parent'}->get_aggr_upgrade_status(
        -aggr    => $self->get_aggr(),
        -verbose => 1
    );
    if ( $self->{'parent'}->error() ) {
        $self->_set_err( "Error: Volume:get_vol_upgrade_status(): "
                . $self->{'parent'}->errmsg() );
        return "";
    }
    #Getting the status of volume upgrade
    my @status = keys(%{$op->{$vol_name}});
    #Check for the existance of given volume
    if(exists($op->{$vol_name})) {
        if ( defined( $op->{$vol_name}->{$status[0]} ) ) {
	    #Returning the status of volume upgrade
	    return $op->{$vol_name}->{$status[0]};
        } else {
            $self->_set_err( "Error: Volume:get_vol_upgrade_status(): "
                            ."vol upgrade scanner is not in a previously recognised state");
            return "";
        } ## end else
    }
    $self->_set_err( "Error: Volume:get_vol_upgrade_status(): "
            . "vol upgrade scanner status for '$vol_name' is missing from aggr 64bit-upgrade status output"
    );
    return "";
} ## end sub get_vol_upgrade_status

################################################################################
# Subroutine Name: wait_for_vol_upgrade_completion
# Objective:       Waits for aggr upgrade completion
# Details:         See POD documentation at the end of this file
################################################################################
sub wait_for_vol_upgrade_completion {
    my ( $self, %args ) = @_;
    my ( $time_out, $sample_interval ) = ( 3600, 30 );
    my $vol_name = $self->get_name();
    $self->_clear_err();
    foreach ( keys %args ) {
        if (/-?time_out/) {
            if ( ( $time_out = $args{$_} ) < 0 ) {
                $self->_set_err(
                    "Error: Volume::wait_for_vol_upgrade_completion(): Invalid value for argument '-time_out'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( ( $time_out = $args...
            next;
        } ## end if (/-?time_out/)
        if (/-?sample_interval/) {
            if ( ( $sample_interval = $args{$_} ) < 0 ) {
                $self->_set_err(
                    "Error: Volume::wait_for_vol_upgrade_completion(): Invalid value for argument '-sample_interval'"
                );
                return $ClientUtil::FALSE;
            } ## end if ( ( $sample_interval...
            next;
        } else {
            $self->_set_err(
                "Error: Volume::wait_for_vol_upgrade_completion(): Invalid parameter "
                    . "specified \'$_\'" );
            return $ClientUtil::FALSE;
        } ## end else [ if (/-?sample_interval/)
    } ## end foreach ( keys %args )
    while ( $time_out > 0 ) {
	#Getting the status of aggregate 64-bit upgrade
	my $op = $self->{'parent'}->get_aggr_upgrade_status(
		-aggr    => $self->get_aggr(),
		-verbose => 1);
	if ( $self->{'parent'}->error() ) {
            $self->_set_err( "Error: Volume::get_aggr_upgrade_status(): "
                             . $self->{'parent'}->errmsg() );
	    return "";
        }
	#Extracting the status of specified volume upgrade
	my @status = keys(%{$op->{$vol_name}});
        if(exists($op->{$vol_name})) {
	    my @status = keys(%{$op->{$vol_name}});
	    if($op->{$vol_name}->{$status[0]} eq "upgraded" ) {
		#The volume is upgraded to 64-bit
                return $ClientUtil::TRUE;
            }elsif ( $op->{$vol_name}->{$status[0]} ne "running" ) {
		#The upgrade is either stopped or not started 
		$self->_set_err(
                          "Error: Volume::wait_for_vol_upgrade_completion(): "
                        . "aggr upgrade is not running .The status is $op->{$vol_name}->{$status[0]}" );
		return $ClientUtil::FALSE;
            } ## end if ( $op ne "running" )
        } else {
	    #The volume entry is missing in the output
            $self->_set_err( "Error: Volume:get_vol_upgrade_status(): "
                    . "vol upgrade scanner status for '$vol_name' is missing from aggr 64bit-upgrade status output");
	    return "";
        }
	#The upgrade is going on hence sleeping for a while
        &SanUtil::nate_sleep($sample_interval);
        $time_out -= $sample_interval;
    } ## end while ( $time_out > 0 )
    $self->_set_err( "Error: Volume::wait_for_vol_upgrade_completion(): "
            . "wait_for_aggr_upgrade_completion() timed out." );
    return $ClientUtil::FALSE;
} ## end sub wait_for_vol_upgrade_completion

###############################################################################
# Method:    get_bit_type()
# Objective: Get the volume bit type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_bit_type {
    my ($self) = @_;
    my ( $line, @output, $cmd, $name, $bit, );

    $self->_clear_err();
    $name = $self->{name};
    $cmd  = "vol status $name";

    @output = $self->{parent}->execute( $cmd, -hide => 1 );
    foreach my $output (@output) {
        if ( $output =~ /\s+(\d\d\-\w+)/i ) {
            $bit = $1;
            $self->sanlog( -msg => "The bittype is $bit" );
            $self->{bit_type} = $bit;
            return $bit;
        } ## end if ( $output =~ /\s+(\d\d\-\w+)/i)
    } ## end foreach my $output (@output)
    return;
} ## end sub get_bit_type

################################################################################
# Subroutine Name: _exist()
# Objective:       Determines if volume exists
# Details:         See POD documentation at the end of this file
################################################################################
sub _exist {
    my ($self) = @_;
    my ($cmd, $name, @output,);

    $name = $self->{name};

    $cmd = "vol status " . $name;

    @output = $self->{parent}->execute($cmd, -hide => 1);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    if ($output[0] =~ /No volume/i) {
        $self->_set_err("Error: Volume::_exist() no volume found", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub _exist

################################################################################
# Subroutine Name: _get_wafl_scanids
# Objective:       retrieve wafl scan ids
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_wafl_scanids() {
    my ($self, $action) = @_;
    my ($parent, $cmd, @scanids, @output, %hScan,);

    $self->_clear_err();

    $parent = $self->get_parent();

    %hScan = (reallocate     => "reallocating",
              i2p            => "parent",
              measure_layout => "layout",);

    $cmd = "priv set advanced; wafl scan status " . $self->get_name();

    @output = $self->{parent}->execute($cmd, -hide => 0);
    $self->sanlog(-msg => \@output);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    foreach my $line (@output) {
        next if ($line !~ /$hScan{$action}/);
        my ($id) = $line =~ /^\s+(\d+)/;
        push @scanids, $id;
    }

    return (wantarray()) ? @scanids : [@scanids];
} ## end sub _get_wafl_scanids()

###############################################################################
# Method:    _trim
# Objective: remove leading/triling whitespace
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _trim {
    my @out = @_;
    for (@out) {
        s/^\s+//;
        s/\s+$//;
    }
    return wantarray ? @out : $out[0];
}
##############################################################################
# Method    : get_compression_status
# Objective : Returns compression sis data in the form of hash.
# Details   : See POD documentation at the end of this file.
##############################################################################
sub get_compression_status {

    my ($self, $vol_name, %sis_hash) = @_;

    my $cmd = "df -S $vol_name";

    my @output = $self->{parent}->execute($cmd, -hide => 0);

    my @keys   = ($output[0] =~ /(\S+)/g);
    my @values = ($output[1] =~ /(\S+)/g);

    for (my $count = 0 ; $count < @keys ; $count++) {
        $sis_hash{$keys[$count]} = $values[$count];
    }

    return (\%sis_hash);
} ## end sub get_compression_status
###############################################################################
# Method   : cmd_vol_options
# Objective : Executes the command
# Details   : See POD documentation at the end of this file.
#             (Supported on BR)
# Return    : 1 on sucessful execution of command, 0 on failure.
##############################################################################
sub cmd_vol_options {

    my ($self, $vol_name, $option_name, $option_value) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();

    my $resultsHash_ref =
      $cmd_obj->execute_cmd(-cmd            => "volume-set-option",
                            -cmd_filer      => $self->{filer},
                            -check_result   => 1,
                            -volume         => $vol_name,
                            -'option-name'  => $option_name,
                            -'option-value' => $option_value);

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->{parent}->sanlog(-msg => "Failed to execute command : $errmsg");
    }

    my $output = $resultsHash_ref->{output};

    if ($output) {
        return 0;
    }

    return 1;
} ## end sub cmd_vol_options
##############################################################################
# Method    : get_sis_config_status
# Objective : Returns sis configuration status in the form of hash.
# Details   : See POD documentation at the end of this file.
##############################################################################
sub get_sis_config_status {

    my ($self, $vol_name, %config_sis_hash) = @_;

    my $cmd = "sis config /vol/$vol_name";
    # Build the command driver
    my @output = $self->{parent}->execute($cmd, -hide => 0);

    my @keys   = ($output[1] =~ /(\S+)/g);
    my @values = ($output[3] =~ /(\S+)/g);

    for (my $count = 0 ; $count < @keys ; $count++) {
        $config_sis_hash{$keys[$count]} = $values[$count];
    }

    return (\%config_sis_hash);
} ## end sub get_sis_config_status
##############################################################################
# Method    : vol_rename
# Objective : Rename the volume to specified Name.
# Details   : See POD documentation at the end of this file.
##############################################################################
sub vol_rename {

    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->cmd_obj();
    my $interface = $self->{parent}->interface();
    my $resultsHash_ref;
    my $new_vol_name;

    $self->_clear_err();

    my $vol_name = $self->{name};

    # parse arguments
    foreach (keys %args) {
        if (/-?new_vol_name/) {
            if (not length($new_vol_name = $args{$_})) {
                $self->_set_err(
                    "Error: Volume::vol_rename()"
                    ." Invalid value for argument '-new_vol_name'"
                    );
                return {};
            }
            next;
        } else {
            $self->_set_err(
                "Error: Volume::vol_rename() Invalid parameter specified \'$_\'"
                );
            return {};
        }
    } ## end foreach (keys %args)
    $self->sanlog(-msg => "The new volume name is $new_vol_name");
    unless (defined $new_vol_name and length $new_vol_name) {
        $self->_set_err(
            "Error: Volume:: vol_rename() new_vol_name is a required argument");
        return $ClientUtil::FALSE;
    }

    # Build the command driver
    my %cmd_args = (
		    -cmd            => "volume-rename",
		    -cmd_filer      => $self->get_filername(),
		    -cmd_interface  => $interface,
		    -volume         => $vol_name,
	         );

    if (lc($interface) eq "zapi") {
        $cmd_args{"-new-volume-name"} = $new_vol_name;
    } else {
        $cmd_args{"-new-name"} = $new_vol_name;
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }# end if ($cmd_obj->error())

    $self->{name} = $new_vol_name;
    return $ClientUtil::TRUE;

} ## end sub vol_rename
1;
#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Volume.pm

=head1 Authors

=begin html

Gary Knopp

=end html

=head1 Description

Volume provides an API interface to a filer volume.  The package should be
instantiated through the San::SanFiler package create_vol() or get_vol() methods.

=head1 Attributes

=head2 B<name>

string = Volume Name

=head2 B<parent>

reference = SanFiler Object

=head2 B<filer>

string = Filer name

=head2 B<luns>

array = list of LUN Objects

=head2 B<options>

hash = Volume options (parameter=>value)

=head2 B<state>

string = <online | offline>

=head2 B<log>

boolean = <1 | 0>

=head1 Public Methods

=head2 B<autosize(%args)>

Manage the autosize attribute of a volume

=over 4

=item Inputs

=item * -state (on|off|reset)

The state to pass to the vol autosize command.

=item * I<-max>

The maximum size to pass to the vol autosize command.

=item * I<-increment>

The increment value to pass to the vol autosize command.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Set autosize volume'");
    unless ($volObj->autosize(-state => on)){
        $Test->nlog($volObj->errmsg());
    }

=back

=head2 B<create_lun(%args)>

Creates a new lun object.

=over 4

=item Inputs

=item * -name

The name of the lun (w/o volume path)

=item * -size (note: specify units 500k, 1m, 2g)

The size of the volume (note: specify units 500k, 100m, 1g)

=item Outputs

Returns a lun object. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    # Create a volume Group object.
    $Test->log("-- Creating lun object 'mylun'");
    $lunObj = $volObj->create_lun(-name=>"mylun",-size=>'100m');
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<create_lun_from_file(%args)>

Creates a file, then creates a new lun object from that file.

=over 4

=item Inputs

=item * -name

The name of the lun (w/o volume path)

=item * -from_file

The name of the file (w/o volume path)

=item * -size (note: specify units 500k, 1m, 2g)

The size of the volume (note: specify units 500k, 100m, 1g)

=item Outputs

Returns a lun object. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    # Create a volume Group object.
    $Test->log("-- Creating lun object 'mylun'");
    $lunObj = $volObj->create_lun_from_file(-name=>"mylun",
                                            -size=>'100m',
                                            -from_file=>"myfile");
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<destroy()>

Destroy filer volume

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If the lun object does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Destroy volume'");
    if (not $volObj->destroy()){
        $Test->nlog($volObj->errmsg());
    }

=back

=head2 B<destroy_lun($name,%args)>

Destroy a lun object by name

=over 4

=item Inputs

=item * $name

The name of the lun object to destroy

=item * -force (default=0)

Unmaps the lun from an igroup if mapped

=item Outputs

Returns a scalar value of 1 if successful.  If the lun object does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->log("-- Destroy lun 'mylun'");
    if (not $lunObj->destroy_lun("mylun")){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display(%args)>

Display detailed lun information

=over 4

=item Inputs

=item * I<-lunFilter>

Array ref of lunObjects.  If not provided all luns information on
volume is displayed

=item * I<-lunsOnly> (default: 0)

Only display lun information.

=item Outputs

Returns an array or array reference if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->log("-- Retrieving volume information");

    $Test->log($volObj->display());
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<export_fs(I<$hostname>)>

exports the volume

=over 4

=item Inputs

=item * I<$hostname>

Name of host to give r/w access to filer volume.  The default
is to allow world access.

=item Outputs

Returns a scalar value if successful, otherwise "".  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is returned.

=item Supported Interfaces:

cli

=item Examples

    $hostname="jupiter";
    if ($volObj->export_fs($hostname)) {
        $Test->nlog("Volume successfully exported");
    }


=back

=head2 B<get_filername()>

Name of volume

=over 4

=item Inputs

none

=item Outputs

Returns the name of the filer.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces:

All

=item Examples

    print "Filername is '".$volObj->get_filername()."'";

=back

=head2 B<get_freespace(%args)>

Estimate of free disk space (megabytes) in volume group

=over 4

=item Inputs

=item * I<-reserve> (default: 1)
Determines whether space should be set aside for snapshot reserve. A value of 1
sets aside snapshot space.


=item Outputs

Returns a scalar value (including unit) of available free space.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and 0 is returned.

=item Supported Interfaces:

cli

=item Examples

    $size=$volObj->get_freespace(-reserve=>0);
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

    print "Volume has $size available (w/o snapshot reserve)\n";

=back


=head2 B<get_lun($name)>

Retrieve a lun object by name

=over 4

=item Inputs

=item * $name

The name of the lun to retrieve

=item Outputs

Returns a lun object if successful.  If the lun does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

All

=item Examples

    $Test->log("-- Retrieving lun object 'mylun'");
    $lunObj=$volObj->get_lun("mylun");
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_luns()>

Retrieve a list of all lun objects in the filer volume

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of lun objects if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty array is returned.

=item Supported Interfaces:

All

=item Examples

    $Test->log("-- Retrieving volume lun objects");
    @lunObjs=$volObj->get_luns();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_lun_cnt()>

Number of luns in the filer volume

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and 0 is returned.

=item Supported Interfaces:

cli

=item Examples

    $cnt=$volObj->get_lun_cnt();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

    print $volObj->get_name()." has $cnt luns\n";

=back

=head2 B<get_name()>

Name of volume

=over 4

=item Inputs

none

=item Outputs

Returns the name of the filer volume.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces:

All

=item Examples

    print "Filer Volume name is '".$volObj->get_name()."'";

=back

=head2 B<get_parent()>

Retrieve filer parent object

=over 4

=item Inputs

none

=item Outputs

Returns the filer parent objectof volume.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and empty hash ref is returned.

=item Supported Interfaces:

All

=item Examples

    $filerObj=$volObj->get_parent();

=back

=head2 B<get_snap_names(I<%args>)>

Retrieve snapshots names for aparticular volume

=over 4

=item Outputs

=item * I<-filter>

A regular expression to use to filter out which snapnames are returned.  Examples would
be "^snap_1$", "^snap_", "_2$", or "snap" for handling exact matches, starting with, ending
with, and containing, respectivley.

=item Outputs

Returns an array or array reference of snapnames if successful.  If an error occured the error flag
is set, errmsg is set to the failure reason, and empty array or array ref is passed back as the
return values.

=item Supported Interfaces:

cli

=item Examples

 # Display snapshot names starting wiht 'snap'
 @names= $vol_obj->get_snap_names(-filter=>'snap');
 if ($vol_obj->error){
     $Test->nlog($vol_obj->errmsg());
 }

 $Test->nlog(@names);

=back

=head2 B<get_snap_reserve()>

Retrieve percentage of volume storage reseerved for snapshots

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned

=item Supported Interfaces:

cli

=item Examples

    $reserve=$volObj->get_snap_reserve();
    $Test->nlog("Snap reserve percentage is ".$reserve);

=back

=head2 B<snap_schedule()>

Set the schedule for taking snapshots

=over 4

=item Inputs

=item * -schedule

The schedule for taking snapshots

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Disable scheduled snaps");
    if (not $volObj->snap_schedule(-schedule => "0")){
        $Test->nlog($volObj->errmsg());
    }

=back

=head2 B<get_state()>

Returns the volume current state

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful, otherwise "".  If an error occured the error
flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    $state=$vol->get_state();
    $Test->log("state=".$state);


=back

=head2 B<get_type()>

Returns the volume type (flex|trad)

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful, otherwise "".  If an error occured the error
flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    $type=$vol->get_type();
    $Test->log("type=".$type);


=back

=head2 B<get_space(%args)>

Get the volumes disk space usage (uses df)

=over 4

=item Inputs

=item * I<-name (volume|snapshot)>

The resource to measure

=item * I<-type (total|used|avail) default is "total">

The type of information you want returned

=item * I<-unit (k|m|g|t) default is "k">

Unit size multiplier to report

=item Outputs

Returns a representation of volume space if successful.  Examples of what
would be returned are "14g", "64k".  If an error occured the error
flag is set, errmsg is set to the failure reason, and $self is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Get avail space on a  volume'");
    my $vol_space_avail = $volObj->get_space(-name => "volume",
                                             -type => "avail"));
    unless ($volObj->error()) {
        $Test->nlog("-- Avail space on volume is $vol_space_avail");
    }

=back

=head2 B<is_exported>

Checks to see if a volume is exported

=over 4

=item Inputs

None

=item Outputs

Returns a value of $ClientUtil::TRUE if exported, otherwise $ClientUtil::FALSSE.  If an error occured the
error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    if ($volObj->is_exported()) {
        $Test->nlog("Volume is exported");
    } else {
        $Test->nlog("Volume is not exported");
    }

=back

=head2 B<is_offline()>

Determine if a volume is offline

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if offline, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces:

All

=item Examples

     # If offline place online
    if ($Obj->is_offline()){
        $Obj->online()
    }

=back

=head2 B<offline()>

Place a volume offline

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    # Place a volume offline
    $volObj->offline();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<online()>

Place a volume online

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

    # Place a volume online
    $Obj->online();
    if ($Obj->error){
        $Test->log($Obj->errmsg());
        return 0;
    }

=back

=head2 B<options(%args)>

Set volume options

=over 4

=item Inputs

=item * I<-[all volume options]>

Provide filer options and values as hash key/value pairs.  All options supplied are
validated prior to being set.

=item Outputs

Returns a value of 1 if options is set otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

     # set volume options
     $volObj->options(-guarantee=>none,-fractional_reserve=>80);
     if ($volObj->error){
        $Test->nlog($volObj->errmsg());
        $fail_cnt++;
     }

=back

=head2 B<unexport_fs()>

unexports a volume

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value if successful, otherwise "".  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is returned.

=item Supported Interfaces:

cli

=item Examples


    if ($volObj->unexport_fs($hostname)) {
        $Test->nlog("Volume successfully unexported");
    }

=back

=head2 B<refresh()>

Refresh filer volume.  This method will refresh lun object content.

=over 4

=item Inputs

=item * I<-aggr>

Set to 1 to have the aggr information for the volume refreshed. Defaults to 1.  Set to 0 to turn off.

=item * I<-luns>

Set to 1 to have the lun objects for the volume recreated. Defaults to 1.  Set to 0 to turn off.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces:

All

=item Examples

    $Test->nlog("Refresh filer volume");
    if (not $fvolObj->refresh() {
        if ($fvolObj->error){
            $Test->nlog($fvolObj->errmsg());
        }
    }

    $Test->nlog("Refresh filer volume lun info");
    if (not $fvolObj->refresh(-luns => 1, -aggr => 0) {
        if ($fvolObj->error){
            $Test->nlog($fvolObj->errmsg());
        }
    }

    $Test->nlog("Refresh filer volume aggr info");
    if (not $fvolObj->refresh(-aggr => 1, -luns => 0) {
        if ($fvolObj->error){
            $Test->nlog($fvolObj->errmsg());
        }
    }

=back

=head2 B<snap_create(I<%args>)>

Create a snapshot for a volume

=over 4

=item Inputs

=item * -name

The name of the snapshot.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

 # Create a  snapshot named snap3
 if (not $vol_obj->snap_create(-name=>"snap3")){
     $Test->nlog($volObj->errmsg());
 }

=back

=head2 B<snap_restore(I<%args>)>

Restore a snapshot for a volume

=over 4

=item Inputs

=item * -name

The name of the snapshot.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

 # Restore a  snapshot named snap3
 if (not $vol_obj->snap_restore(-name=>"snap3")){
     $Test->nlog($volObj->errmsg());
 }

=back

=head2 B<snap_autodelete(I<%args>)>

Create a  snapshot for a volume

=over 4

=item Inputs

=item * -state (on|off|reset)

The state to pass to the snap autodelete command.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

 # Autodelete on
 if (not $vol_obj->snap_autodelete(-state=>"on")){
     $Test->nlog($volObj->errmsg());
 }

=back

=head2 B<snap_autodelete_trigger(%args)>

Destroy filer volume

=over 4

=item Inputs

=item * -trigger (volume|snap_reserve|space_reserve)

The trigger to use

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Set snap autodelete trigger");
    if (not $volObj->snap_autodelete_trigger(-trigger => "space_reserve"));
        $Test->nlog($volObj->errmsg());
    }

=back

=head2 B<snap_delete(I<%args>)>

Remove snapshots from volume

=over 4

=item Inputs

=item * -name

The name of the snapshot.  Providing the '*' character as the value will delete all snapshots
on the filer volume.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

 # Display snapshot that start with the word 'snap'
 if (not $vol_obj->snap_delete(-name=>"snap3")){
     $Test->nlog($volObj->errmsg());
 }

=back

=head2 B<disable_sis()>

Disable SIS on the volume.

=over 4

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occured
the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli, zapi, and zapi_cli

=item Examples

    $vol_obj->disable_sis();

=back

=head2 B<config_sis(%args)>

Configures SIS on the volume.

=over 4

=item Inputs

=item * -schedule (required)

The schedule string for the SIS operation. The format
of theschedule: day_list[hour_list] or hour_list[day_list]
or - or auto The day_list specifies which days of the week
the SIS operation should run. It is a comma-separated list
of the first three letters of the day: sun, mon,
tue, wed, thu, fri, sat. The names are not case
sensitive. Day ranges such as mon-fri can also be
given. The default day_list is sun-sat. The
hour_list specifies which hours of the day the SIS
operation should run on each scheduled day. The
hour_list is a comma-separated list of the
integers from 0 to 23. Hour ranges such as 8-17
are allowed. Step values can be used in conjuction
with ranges. For example, 0-23/2 means "every two
hours". The default hour_list is 0, i.e. midnight
on the morning of each scheduled day. If "-" is
specified, no schedule is set. When a non-SIS
volume is enabled the first time, an initial
schedule is assigned to the volume. This initial
schedule is sun-sat0, that means "once
everyday at midnight". The "auto" schedule string
means the SIS operation will be triggered by the
amount of new data written to the volume. The
criterion is subject to be changed later.

=over 4

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occured
the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

zapi

=item Examples

    $vol_obj->config_sis(-schedule => $schedule);

=back

=head2 B<get_sis_config()>

Display current SIS configuration.

=over 4

=item Outputs

Returns the output of the sis config command for the current volume if it was
successful, otherwise "".

=item Supported Interfaces:

cli, zapi, and zapi_cli

=item Examples

    $output = $vol_obj->get_sis_config();

=back

=head2 B<get_sis_status()>

Display current SIS status.

=over 4

=item Outputs

Returns the output of the sis status command for the current volume if it was
successful, otherwise "".

=item Supported Interfaces:

cli, zapi, and zapi_cli

=item Examples

    $output = $vol_obj->get_sis_status();

=back

=head2 B<start_sis()>

Start SIS on the volume.

=over 4

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occured
the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli, zapi, and zapi_cli

=item Examples

    $vol_obj->start_sis();

=back

=head2 B<stop_sis()>

Stop SIS on the volume.

=over 4

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occured
the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

cli, zapi, and zapi_cli

=item Examples

    $vol_obj->stop_sis();

=back

=head2 B<get_aggr()>

Get the containing aggregate.

=over 4

=item Outputs

Returns the containing aggregate. Fails if called on a trad vol.

=item Supported Interfaces:

zapi_cli

=item Examples

    $aggr_name = $vol_obj->get_aggr();

=back

=head2 B<get_info()>

Get volume information from the volume-list-info zapi

=over 4

=item Outputs

Returns the XML that was returned from the volume-list-info ZAPI

=item Supported Interfaces:

zapi

=item Examples

    $volume_xml = $vol_obj->get_info();

=back

=head2 B<snap_exist(%args)>

Checks to see if a specific snapshot exists on this volume.

=over 4

=item Inputs

=item * -snapname

The name of the snapshot

=item Outputs

Returns a scalar value of 1 if it exists, otherwise 0 and an error is set.

=item Supported Interfaces:

cli

=item Examples

    if($vol_obj->snap_exist(-snapname => "hourly.0"))
    {
        # do something
    }

=back

=head2 B<snap_list(I<%args>)>

Retrieve information(raw) on snapshots

=over 4

=item Inputs

=item * I<-filter>

A regular expression to use to filter out which snap records are displayed.  Examples would
be "^snap_1$", "^snap_", "_2$", or "snap" for handling exact matches, starting with, ending
with, and containing, respectivley.

=item Outputs

Returns an array or array reference if successful.  If an error occured the error flag
is set, errmsg is set to the failure reason, and empty array ref is passed back as the
return values.

=item Supported Interfaces:

cli

=item Examples

 # Display snapshot that start with the word 'snap'
 @output= $vol_obj->snap_list(-filter=>'^mysnap');
 if ($vol_obj->error){
     $Test->nlog($vol_obj->errmsg());
 }

 $Test->log(@output);

=back

=head2 B<snap_reserve(I<%args>)>

Set percentage of volume storage reserved for snapshots

=over 4

=item Inputs

=item * I<-pct>

Percentage of storage

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.  If an error occured the error flag
is set, errmsg is set to the failure reason, and empty array ref is passed back as the
return values.

=item Supported Interfaces:

cli

=item Examples

 # Display snapshot that start with the word 'snap'
 $vol_obj->snap_reserve(-pct=>20);
 if ($vol_obj->error){
     $Test->nlog($vol_obj->errmsg());
 }

 $Test->log(@output);

=back

=head2 B<wafl_abort(I<$id>)>

Abort wafl scan operation(s)

=over 4

=item Inputs

=item * I<$id>

Wafl scan id of operation to abort.  If no Wafl scan id is provided than all wafl
operations on the volume are aborted.

=item Outputs

Returns a value of 1 if abort is successful, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

     # abort all wafl scan operations
     $volObj->wafl_abort();

=back

=head2 B<wafl_scan(%args)>

Abort wafl scan operation(s)

=over 4

=item Inputs

=item * B<-action> (values: "abort|measure_layout|reallocate|i2p|fold");

    Wafl scan action to perform.

=item * I<-scanid> (action 'abort' only)

    Wafl scan id of action if 'abort'.  If no Wafl scan id is provided than all wafl
    actions in process on the volume are aborted.

=item Outputs

Returns the operations scanid if successful, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces:

cli

=item Examples

     # abort all wafl scan operations
     $volObj->wafl_scan(-action=>"abort");

=back

=head2 B<create_qtree()>

Create a qtree on the volume

=over 4

=item Inputs

=item * -name

The name of the qtree

=item Outputs

Returns a Qtree object if successful. If an error occurs the error flag is set on parent object and
parent object is returned.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Create a qtree");
    $qtreeObj = $volObj->create_qtree(-name => "myqtree");
    if ($volObj->error()) {
        $Test->nlog($volObj->errmsg());
    }

=back

=head2 B<delete_qtree()>

delete a qtree on the volume

=over 4

=item Inputs

=item * -name

The name of the qtree

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Delete a qtree");
    if (not $volObj->delete_qtree(-name => "qtree1")){
        $Test->nlog($volObj->errmsg());
    }

=back


=head2 B<qtree_exist()>

Checks to see if a qtree exists in the volume

=over 4

=item Inputs

=item * -name

The name of the qtree

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured or
if the qtree does not exist, the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces:

cli

=item Examples

    $Test->nlog("-- Create qtree unless it exists");
    unless ($volObj->qtree_exist("admin")){
        if (not $volObj->create_qtree(-name => "admin")){
            $Test->nlog($volObj->errmsg());
        }
    }

=back

=head2 B<lun_list_info()>

Return information on the luns in this volume.

=over 4

=item Inputs

None

=item Outputs

Returns a hash keyed by the path to each lun.  Each hash entry will
be a hash itself containing entries returned. For the ZAPI interface,
we simply translate the XML returned into the hash and return the
data that came back.  For the CLI, we parse the data that came back
and each hash entry will contain fields that match those from the ZAPI.
Specifically, the hash for each lun will contain these fields:
path, size, online, mapped, read-only, serial-number, share-state,
is-space-reservation-enabled, multiprotocol-type, backing-snapshot,
maps.  Note that the maps field only comes back if we use the CLI and
it looks like this: igroup_name=1 (where 1 is the lun id).
If there are no luns, an empty hash is returned.
If a failure occurs, the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces:

cli zapi

=item Examples

    $Test->nlog("-- List mapped luns on a given volume");
    my $luns_hash = $vol_obj->lun_list_info();
    if ($vol_obj->error()) {
        $Test->nlog("Error getting luns from volume: " . $vol_obj->errmsg())
        return $TCD::FAIL;
    }
    foreach my $lun_path (keys(%{$luns_hash})) {
        my $lun_info = $lun_hash->{$lun_path};
        if ($lun_info->{mapped} =~ /yes/) {
            $Test->nlog("Lun $lun_path is mapped");
        }
    }

=back

=head2 B<wait_for_online(I<%args>)>

Wait for the state of a volume to be online

=over 4

=item Inputs

=item * I<-timeout> (default: 10800)

The number of seconds to wait before failure

=item * I<-interval> (default: 120)

The number of seconds to sleep between polling the state of the volume

=item Outputs

Returns a value of 1 if successful and volume is online, otherwise 0.
If an error occured the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces:

zapi

=item Examples

    $filer_obj->interface("zapi");

    # Create the volume
    $vol_objz = $filer_obj->create_vol(-name     => "testVol",
                                       -disk_cnt => $num_disks);

    if($filer_obj->error())
    {
        my $errmsg = join "\n", $filer_obj->errmsg();
        $Test->nlog("Error: Failed volume_wait_for_online(): ".
                    "Could not get create volume testVol" .
                    ": $errmsg");
        return 0;
    }

    # Call wait for online
    $return_code = $vol_objz->wait_for_online(-timeout  => 3700,
                                              -interval => 90);
    unless  ($return_code) {
        $Test->nlog("Error: Failed volume_wait_for_online(): ".
                    "Volume did not come online before timeout");
        reutrn 0;
    }
    return 1;

=back

=head2 B<start_compression(I<%args>)>

Start compression on the volume

=over 4

=item Inputs

=item * I<-delete_checkpoint> (default: 0)

Whether to delete any compression checkpoints when starting compression.

=item * I<-include_shared_blocks> (default: 0)

Whether to take into account deduped blocks when compressing the volume.

=item * I<-include_snapshot_blocks> (default: 0)

Whether to take into account snapshot blocks when compressing the volume.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi

=item Examples

    $filer_obj->interface("zapi");

    # Create the volume
    $vol_obj = $filer_obj->create_vol(-name => "testVol");
    if($filer_obj->error())
    {
        ...
    }

    # start compression on the volume
    $vol_obj->start_compression(-delete_checkpoint => 1,
                                -include_shared_blocks => 0);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<stop_compression(I<%args>)>

Stop compression on the volume

=over 4

=item Inputs

=item * I<-delete_checkpoint> (default: 0)

Whether to delete any decompression checkpoints when stopping compression.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi

=item Examples

    $filer_obj->interface("zapi");

    # Create the volume
    $vol_obj = $filer_obj->create_vol(-name => "testVol");
    if($filer_obj->error())
    {
        ...
    }

    # stop compression on the volume
    $vol_obj->stop_compression(-delete_checkpoint => 0);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<start_decompression(I<%args>)>

Start decompression on the volume

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi

=item Examples

    $filer_obj->interface("zapi");

    # start decompression on the volume
    $vol_obj->start_decompression();
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<stop_decompression(I<%args>)>

Stop decompression on the volume

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi

=item Examples

    $filer_obj->interface("zapi");

    # stop decompression on the volume
    $vol_obj->stop_decompression();
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<move_start()>

Start a non-disruptive move

=over 4

=item Inputs

=item * -aggr

(Required) Destination aggregate for the move

=item * -cutover_window

(Optional) Seconds allowed per cutover attempt. Default: 300

=item * -interval

(Optional) Seconds to sleep between status checks.  Default: 30

=item * -keep_source

(Optional) Keep the source volume in the source aggregate after it is moved.  1 = True,  0 = False.  Default is 0.

=item * -manual_cutover

(Optional) Perform a move with a manual cutover.  1 = True,  0 = False.  Default is 0. This parameter is mutually
exclusive with -wait_flag.

=item * -max_wait

(Optional) Maximum time in seconds to wait for a move and automatic cutover to happen.  Default: 1200.

=item * -num_cutover

(Optional) Number of cutover attempts

=item * -override

(Optional) Ignore the redundancy characteristics of aggregates in metro cluster environments.  1 = True,  0 = False.
Default is 0.

=item * -wait_flag

(Optional) Flag to determine whether the method should wait for the move and cutover to complete before returning
or start the move and immediately return.  1 = True,  0 = False.  Default is 0.  This parameter is mutually exclusive
with -manual_cutover.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Automatic Cutover with Wait

    $vol_obj->move_start( -aggr      => <destination aggr>,
                          -wait_flag => 1);
    if ($vol_obj->error())
        ...
    }

    # Automatic Cutover with Wait and other options

    $vol_obj->move_start( -aggr      => <destination aggr>,
                          -max_wait  => 900,
                          -interval  => 30,
                          -wait_flag => 1);
    if ($vol_obj->error())
        ...
    }

    # Manual Cutover

    $vol_obj->move_start( -aggr           => <destination aggr>,
                           -manual_cutover => 1);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<move_status>

Get the status of a non-disruptive volume move

=over 4

=item Inputs

=item * -verbose

(Optional) Returns verbose or non-verbose status.  1 = True,  0 = False.  Default is 0.

=item Outputs

Returns reference to a hash containing the status output returned from a vol move status cli call or a
volume-move-status zapi command.  The hash reference will be in the form of:

Verbose Status:

$output_hash->{
    'migration-state => 'running',
    'vol-move-status-info' => {
        'cutover-attempts' => '6',
        'move-state' => ' setup (paused)',
        'cutover-window' => '88',
        'last-transfer-size' => '0',
        'Last-transfer-duration' => '0',
        'temp-volume' => 'ndm_dstvol_1267950718',
        'dest-aggregate' => 'destaggr',
        'source-volume-name' => 'NdvmSrcVol1',
        'src-aggregate' => 'testaggr',
    }
}

Non-Verbose Status:

$output_hash->{
    'migration-state => 'running',
    'vol-move-status-info' => {
        'cutover-attempts' => '6',
        'move-state' => ' setup (paused)',
        'cutover-window' => '88',
        'dest-aggregate' => 'destaggr',
        'source-volume-name' => 'NdvmSrcVol1',
        'src-aggregate' => 'testaggr',
    }
}

=item Supported Interfaces:

zapi and cli

=item Examples

    Verbose Status:

    my $output_hash = $vol_obj->move_status(-verbose => 1);
    if ($ndvm_obj->error())
        ...
    }

    unless (defined($output_hash->{'vol-move-status-info'}->{'move-state'})
            and $output_hash->{'vol-move-status-info'}->{'move-state'} =~ /paused/) {
        $self->_set_err("Unable to resume the move of volume $volume.  Move state is currently " .
                        $output_hash->{'vol-move-status-info'}->{'move-state'});
        return $ClientUtil::FALSE;
    }

=back

=head2 B<move_cutover(I<%args>)>

Perform a manual cutover of a non-disruptive volume move

=over 4

=item Inputs

=item * -cutover_window

(Optional) Seconds allowed per cutover attempt. Default: 300

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Cutover with modified  cutover window and unexport

    $vol_obj->move_cutover(-cutover_window => 90);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<move_abort();

Abort a non-disruptive volume move

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Abort a move

    $vol_obj->move_abort();
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<move_pause();

Pause a non-disruptive volume move

=over 4

=item Inputs

None

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Pause a move

    $vol_obj->move_pause();
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<move_resume(I<%args>)>

Resume a paused non-disruptive volume move

=over 4

=item Inputs

=item * -aggr

(Required) Destination aggregate for the move

=item * -cutover_window

(Optional) Seconds allowed per cutover attempt.  Default: 300

=item * -interval

(Optional) Seconds to sleep between status checks.  Default: 30

=item * -keep_source

(Optional) Keep the source volume in the source aggregate after it is moved.  1 = True,  0 = False.  Default is 0.

=item * -manual_cutover

(Optional) Perform a move with a manual cutover.  1 = True,  0 = False.  Default is 0. This parameter is mutually
exclusive with -wait_flag.

=item * -max_wait

(Optional) Maximum time in seconds to wait for a move and automatic cutover to happen.  Default is 1200.

=item * -num_cutover

(Optional) Number of cutover attempts

=item * -override

(Optional) Ignore the redundancy characteristics of aggregates in metro cluster environments.  1 = True,  0 = False.
Default is 0.

=item * -wait_flag

(Optional) Flag to determine whether the method should wait for the move and cutover to complete before returning
or start the move and immediately return.  1 = True,  0 = False.  Default is 0.  This parameter is mutually exclusive
with -manual_cutover.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    # Resume with an automatic cutover and wait

    $vol_obj->move_resume( -aggr      => <destination aggr>,
                           -wait_flag => 1);
    if ($vol_obj->error())
        ...
    }

    # Resume with an automatic cutover, wait and other options

    $vol_obj->move_resume( -aggr            => <destination aggr>,
                           -max_wait        => 900,
                           -interval        => 30,
                           -num_cutover     => 5,
                           -cutover_window  => 90,
                           -wait_flag       => 1);
    if ($vol_obj->error())
        ...
    }

    # Resume with manual Cutover

    $vol_obj->move_resume( -aggr           => <destination aggr>,
                           -manual_cutover => 1);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<wait_for_move_complete(I<%args>)>

Monitor a volume move that is in progress until it completes or the maximum wait time has expired.

=over 4

=item Inputs

=item * -aggr

(Required) Destination aggregate for the move

=item * -interval

(Optional) Seconds to sleep between status checks.  Default: 30

=item * -max_wait

(Optional) Maximum time in seconds to wait for a move and automatic cutover to happen.  Default: 1200.

=item Outputs

Returns $ClientUtil::TRUE if successful, $ClientUtil::FALSE and sets error if not.

=item Supported Interfaces:

zapi and cli

=item Examples

    $vol_obj->wait_for_vol_move_complete( -aggr    => <destination aggr>,
                                         -max_wait => 900,
                                         -interval => 30);
    if ($vol_obj->error())
        ...
    }

=back

=head2 B<vol_quota_on()>

Sets Volume quota state to on.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

zapi

=item Examples

    # set volume quota state to on
    $volObj->vol_quota_on();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_off()>

Set volume quota state to off.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

zapi

=item Examples

    # set volume quota state to off
    $volObj->vol_quota_off();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_resize()>

Set volume quota state to resizeing.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces:

zapi

=item Examples

    # set volume quota state to resizeing
    $volObj->vol_quota_resize();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_status()>

Returns volume quota status.

=over 4

=item Inputs

none

=item Outputs

Returns a Hash  reference which consists of quota status

=item Supported Interfaces:

zapi

=item Examples

    # Returns volume quota status.
    my $quota_status_hash = $volObj->vol_quota_status();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head 2 B<vol_quota_get_entry(I<%args>)>

Obtains a quota entry specified.

=over 4

=item Inputs

=item * I<-qtree>  (default: " ") # it will take an empty string i.e no qtree if not specified

If qtree is exists have to specify that qtree name

=item * I<-quota_type>

The possible values for quota type are : "user", "group", or "tree".

=item * I<-quota_target>

The quota target of the type specified. Possible values are: name, number, or path name.
Multiple targets can be specified by a comma-separated list.

=item Outputs

Returns a Hash reference which contains all the quota values.

=item Supported Interface:

zapi

=item Examples

    # Obtains a quota entry specified by type, target, volume, and tree.
    my $quota_get_entry_hash = $volObj->vol_quota_get_entry(
                                                            -quota_type   => "user",
                                                            -quota_target => $volume,
                                                       );
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_set_entry(I<%args>)>

Sets a quota entry.

=over4

=item Inputs

=item * I<-qtree>  (default: " ") # it will take an empty string i.e no qtree if not specified

If qtree is exists have to specify that qtree name

=item * I<-quota_type>

The possible values for quota type are : "user", "group", or "tree".

=item * I<-quota_target>

The quota target of the type specified. Possible values are: name, number, or path name.
Multiple targets can be specified by a comma-separated list.

=item * I<-disk_limit>

The amount of disk space that is reserved for the the target. The value is expressed in kilobytes.

=item * I<-file_limit>

The number of files that the target can have.

=item * I<-soft_disk_limit>

The amount of disk space the target would have to exceed before a message is logged and an SNMP trap
is generated. The value is expressed in kilobytes.

=item * I<-soft_file_limit>

The number of files the target would have to exceed before a message is logged and an SNMP trap is generated.

=item * I<-threshold>

The amount of disk space the target would have to exceed before a message is logged. The value is expressed in kilobytes.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Sets a quota entry
    $volObj->vol_quota_set_entry(
                                -quota_type   => "user",
                                -quota_target => $volume,
           );
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_report()>

Returns the quota state

=over 4

=item Inputs

none

=item Outputs

Returns a Hash reference which includes all the information.

=item Supported Interfaces:

zapi

=item Examples

    # Returns volume quota state.
    my $quota_report_hash = $volObj->vol_quota_report();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_list_entries(I<%args>)>

Lists all the entries present inside /etc/quotas file.

=over 4

=item Inputs

=item * I<-include_output_entry>

boolean optional value either True or False.

=item Outputs

Returns quota entries from the /etc/quotas file.

=item Supported Interfaces:

zapi

=item Examples
    # Returns Quota entries
    my $quota_list_hash = $volObj->vol_quota_list_entries(-include_output_entry => "true");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_add_entry(I<%args>)>

Add a new entry to the /etc/quotas.

=over4

=item Inputs

=item * I<-qtree>  (default: " ") # it will take an empty string i.e no qtree if not specified

If qtree is exists have to specify that qtree name

=item * I<-quota_type>

The possible values for quota type are : "user", "group", or "tree".

=item * I<-quota_target>

The quota target of the type specified. Possible values are: name, number, or path name.
Multiple targets can be specified by a comma-separated list.

=item * I<-disk_limit>

The amount of disk space that is reserved for the the target. The value is expressed in kilobytes.

=item * I<-file_limit>

The number of files that the target can have.

=item * I<-soft_disk_limit>

The amount of disk space the target would have to exceed before a message is logged and an SNMP trap
is generated. The value is expressed in kilobytes.

=item * I<-soft_file_limit>

The number of files the target would have to exceed before a message is logged and an SNMP trap is generated.

=item * I<-threshold>

The amount of disk space the target would have to exceed before a message is logged. The value is expressed in kilobytes.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Sets a quota entry
    $volObj->vol_quota_add_entry(
                                -quota_type   => "user",
                                -quota_target => $volume,
           );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_modify_entry(I<%args>)>

Modifies the existing entry of /etc/quotas.

=over4

=item Inputs

=item * I<-qtree>  (default: " ") # it will take an empty string i.e no qtree if not specified

If qtree is exists have to specify that qtree name

=item * I<-quota_type>

The possible values for quota type are : "user", "group", or "tree".

=item * I<-quota_target>

The quota target of the type specified. Possible values are: name, number, or path name.
Multiple targets can be specified by a comma-separated list.

=item * I<-disk_limit>

The amount of disk space that is reserved for the the target. The value is expressed in kilobytes.

=item * I<-file_limit>

The number of files that the target can have.

=item * I<-soft_disk_limit>

The amount of disk space the target would have to exceed before a message is logged and an SNMP trap
is generated. The value is expressed in kilobytes.

=item * I<-soft_file_limit>

The number of files the target would have to exceed before a message is logged and an SNMP trap is generated.

=item * I<-threshold>

The amount of disk space the target would have to exceed before a message is logged. The value is expressed in kilobytes.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Sets a quota entry
    $volObj->vol_quota_modify_entry(
                                -quota_type   => "user",
                                -quota_target => $volume,
           );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_quota_delete_entry(I<%args>)>

Deletes the specifed Entry from /etc/quotas.

=over 4

=item Inputs

=item * I<-qtree>  (default: " ") # it will take an empty string i.e no qtree if not specified

If qtree is exists have to specify that qtree name

=item * I<-quota_type>

The possible values for quota type are : "user", "group", or "tree".

=item * I<-quota_target>

The quota target of the type specified. Possible values are: name, number, or path name.
Multiple targets can be specified by a comma-separated list.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # delete a quota entry specified by type, target, volume, and tree.
    $volObj->vol_quota_delete_entry(
                                    -quota_type   => "user",
                                    -quota_target => $volume,
           );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_on()>

Enables the reallocation jobs.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally enables WAFL reallocation jobs
    $volObj->vol_reallocate_on();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_off()>

Disables the reallocation jobs.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally disables WAFL reallocation jobs
    $volObj->vol_reallocate_off();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_start(I<%args>)>

Starts the reallocation jobs.

=over 4

=item Inputs

=item * I<-full>

Boolean, optional Possible values are "true" or "false".

=item * I<-interval>

The interval between repeated scans. The format is numX, where X is m, h or d, indicating minutes, hours or days.

=item * I<-is_aggregate>

If true, reallocate the aggregate indicated by the path. This performs block reallocation, not file reallocation.

=item * I<-no_check>

If true, do not check if the current layout optimization is above the threshold before reallocateing.

=item * I<-preserve_logical_bno>

Type: boolean optional Possible values are "true" or "False"

=item * I<-run_once>

Type: boolean, optional If true, run the scan once then exit.

=item * I<-threshold>

The allocation threshold before reallocation occurs. A larger value indicates the file or LUN can be more fragmented. Range: [3..10].

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally starts WAFL reallocation jobs
    $volObj->vol_reallocate_start();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_stop()>

Stops the reallocation jobs.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally stops WAFL reallocation jobs
    $volObj->vol_reallocate_stop();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_quiesce()>

Quiesce the active reallocation jobs.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally quiesce WAFL reallocation jobs
    $volObj->vol_reallocate_quiesce();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_restart(I<%args>)>

Restarts the WAFL reallocation jobs.

=over 4

=item Inputs

=item * I<-ignore_checkpoint>

Type: boolean, optional This option, if true, ignores the checkpoint and restarts at the beginning.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Globally restarts WAFL reallocation jobs
    $volObj->vol_reallocate_restart();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_reallocate_list_info(I<%args>)>

Lists all WAFL reallocation jobs.

=over 4

=item Inputs

=item * I<-verbose>

Type: boolean, optional If true, verbose output is returned.

=item Outputs

Returns a hash which includes all reallocation jobs

=item Supported Interface:

zapi

=item Examples

    # Globally lists all WAFL reallocation jobs
    my $reallocate_list_hash = $volObj->vol_reallocate_list_info();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshots_list_info(I<%args>)>

Lists all the snapshots present in the volume along with detailed information.

=over 4

=item Inputs

=item * I<-snapowners>

Type: boolean, optional Possible values are "true" or "false".

=item * I<-target_name>

Name of the object on which to list the snaplist information.

=item * I<-target_type>

Type of the object on which to list the snaplist information. Possible values: volume, aggregate.

=item * I<-terse>

Type: boolean, optional Desc: If set to true, the snapshot block ownership values, namely the "total" and "cumulative-total"
outputs, will be omitted.

=item Outputs

Returns a hash which includes detail information of snapshots

=item Supported Interface:

zapi

=item Examples

    # Return snapshot information for a specified volume.
    my $snapshots_list_hash = $volObj->vol_snapshots_list_info();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_rename(I<%args>)>

Renames the snapshot name.

=over 4

=item Inputs

=item * I<-new_name>

A new snapshot name.

=item * I<-current_name>

Current snapshot name which is required to rename

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi/cli

=item Examples

    # Rename a current snapshot name to new snapshot name.
    $volObj->vol_snapshot_rename(
                                -current_name => "cur_snap",
                                -new_name     => "new_snap",
            );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_autodelete_set_option(I<%args>)>

Sets the specific option-name and option value to the specified volume.

=over 4

=item Inputs

=item * I<-option_name>

Name of the option to be set. "state", "commitment", "trigger", "target_free_space", "delete_order", "defer_delete".

=item * I<-option_value>

The value to set the named option.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Set the option named 'option-name' to the value specified by 'option-value' in the autodelete settings of the specified volume.
    $volObj->vol_snapshot_autodelete_set_option(
                                               -option_name => "state",
                                               -new_name    => "on",
            );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_set_reserve(I<%args>)>

Sets the size of the indicated volume`s snapshot reserve.

=over 4

=item Inputs

=item * I<-percentage>

Percentage to set. Range [0-100].

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    # Sets the size of the indicated volume's snapshot reserve to the specified percentage.
    $volObj->vol_snapshot_set_reserve(-percentage => 60);
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_reserve_list_info()>

Returns percentage of disk space is reserced for snapshots.

=over 4

=item Inputs

none

=item Outputs

Gets the percentage of disk space that is reserved for snapshots in the indicated volume. If no volume is specified,
this will return the percentage of disk space reserved for snapshots for each of the volumes in the system.

=item Supported Interface:

zapi/cli

=item Examples

    #Gets the percentage of disk space that is reserved for snapshots in the indicated volume.
    $volObj->vol_snapshot_reserve_list_info();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_reclaimable_info(I<%args>)>

Returns the Reclaimable size of the Volume.

=over 4

=item Inputs

=item * I<-snapshots_name>

Type: snapshot-name[] List of snapshots. A maximum of 255 snapshots can be listed.

=item Outputs

Returns the amount of space that would be freed when a set of snapshots are deleted from a specified volume.

=item Supported Interface:

cli

=item Examples

    #Returns the amount of space that would be freed
    $volObj->vol_snapshot_reclaimable_info(-snapshots_name=> \@snaps);
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_autodelete_list_info(I<%args>)>

Returns all current autodelete settings.

=over 4

=item Inputs

=item * I<cmd_end>

This is onlt in case of cli :string Desc: The tail end of the command. Pass an empty string.

=item Outputs

Returns the current snapshot autodelete settings

=item Supported Interface:

cli/zapi

=item Examples

    #Returns a hash of current snapshot autodelete settings
    my $autodelete_hash = $volObj->vol_snapshot_autodelete_list_info();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_snapshot_delta_info(I<%args>)>

Returns the delta between 2 snapshots.

=over 4

=item Inputs

=item * I<-snapshot1>

Name of snapshot to be compared with snapshot2 for space consumption calculations.

=item * I<-snapshot2>

Name of snapshot to be compared with snapshot1 for space consumption calculations. If the snapshot is
not specified, it is assumed to be Active File System.

=item Outputs

Returns the amount of space consumed between two snapshots or a snapshot and active filesystem.

=item Supported Interfaces:

cli/zapi

=item Examples

    # Returns the amount of space consumed between two snapshots
    my $delta_hash = $volObj->vol_snapshot_delta_info(
                                                    -snapshot1 => "snap1",
                                                    -snapshot2 => "snap2",
                   );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_clone_split_estimate(I<%args>)>

Estimates the space availabale.

=over 4

=item Inputs

=item * I<-clone_volume>

The name of the clone whose split usage is being estimated.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    $volObj->vol_clone_split_estimate(-clone_volume => "vol_clone");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_clone_split_start(I<%args>)>

Starts the clone split from its source volume.

=over 4

=item Inputs

=item * I<-clone_volume>

The name of the clone whose split usage is being estimated.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    $volObj->vol_clone_split_start(-clone_volume => "vol_clone");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_clone_split_status(I<%args>)>

Returns a clone split status

=over 4

=item Inputs

=item * I<-clone_volume>

The name of the clone whose split usage is being estimated.

=item Outputs

Returns a hash whch includes Clone status

=item Supported Interface:

zapi

=item Examples

    # Returns a hash
    my clone_hash = $volObj->vol_clone_split_status(-clone_volume => "vol_clone");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

head 2 B<vol_clone_split_stop(I<%args>)>

=over 4

=item Inputs

=item * I<-clone_volume>

The name of the clone whose split usage is being estimated.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    $volObj->vol_clone_split_stop(-clone_volume => "vol_clone");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

head2 B<vol_file_create(I<%args>)>

Deletes a file.

=over 4

=item Inputs

=item * I<-numbers>

Number of files you want to create

=item * I<-filename>

Name of file which you want to create

=item * I<-size>

Size of the file to be created

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

cli

=item Examples

    $volObj->vol_file_create(
                            -numbers  => 2,
                            -size     => "100m",
                            -filename => "newfile",
           );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_dir_delete(<%args>)>

Creates a directory.

=over 4

=item Inputs

=item * I<-dirname>

Name of directory which you want to create

=item * I<-perm>

Spedify what type of permission you giving to that directory

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    $volObj->vol_dir_create(
                            -dirname => "dirfile",
                            -perm    => 777,
           );
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<vol_dir_delete(<%args>)>

Deletes a directory.

=over 4

=item Inputs

=item * I<-filename>

Name of directory which you want to delete

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    $volObj->vol_dir_delete(-filename => "dirfile");
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return 0;
    }

=back

=head2 B<get_compression_status>

Gives the compression status in the form of Hash

=over 4

=item Inputs

=item * -vol_name

The name of the Volume

=item Outputs

Returns has reference containing volume compression status. 

=head2 B<cmd_vol_options>

Execute command with volume option.

=over 4

=item Inputs

=item * -vol_name

The name of the Volume 

=item option_name

The name of the option to set

=item option_value

The value to be set on option

=item Outputs

Returns 1 on successfull execution of command with volume options.

=item Supported Interface:

cli

=head2 B<get_sis_config_status>

Gives the config sis status in the form of Hash

=over 4

=item Inputs

=item * -vol_name

The name of the Volume

=item Outputs

Returns hash reference containing sis configuration status of volume

=head2 B<vol_rename(%args)>

Rename the volume to Specified Name

=over 4

=item Input

=item * -new_vol_name

Need to pass the new name of the volume

=item Outputs

Returns 1 on successfull execution of command with new volume name .
Returns 0 If an error occurs on failure

=item Supported Interface:

cli/zapi

=item Examples

    $Test->nlog(" Rename the volume ");
    $Vol_Obj->vol_rename(-new_vol_name => $Rename_Vol);
    if ($Vol_Obj->error()) {
        $Test->nlog("failed to set new vol name" . $Vol_Obj->errmsg());
        return $TCD::FAIL;
    }

=back

