#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/San/SanFiler/Volume/Lun.pm#2 $

# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Filer
## @author knopp@netapp.com
## @status private
## @pod here

# Package Name
##############
package Lun;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;
use Time::HiRes qw(gettimeofday);
use XML::Simple;
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanUtil;
use Data::Dumper;
use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_minsize
# Objective:       Returns the minimum size for different types of luns
# Details:         See POD documentation at the end of this file
################################################################################
sub get_minsize {
    my ($self, %args) = @_;
    my ($type, $min_size, $filer, $cmd_obj);
    my ($output, $status, $output_hash, $dumpOut, $resultsHash_ref);

    $self->_clear_err();

    # Supported only in Zapi
    $filer = $self->get_filername();
    if ($self->error()) {
        $self->_set_err("failed to get the filername");
        return $ClientUtil::FALSE;
    }
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # parse arguments
    foreach (keys %args) {

        # the type of lun should have got from the attribute "type" in the lun
        # object. But currently this attritube is not set correctly if san setup
        # uses existing luns. hence type is passed here.
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # Build the command line and execute the command
    my %cmd_args = (-cmd           => "lun-get-minsize",
                    -cmd_filer     => $filer,
                    -check_result  => 1,
                    -check_output  => 0,
                    -cmd_interface => "zapi",
                    -type          => $type,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Get the output
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # Since Interface is zapi, convert the output from xml to a string.
    $output_hash = $self->{parent}->{parent}->xml2hash($output);

    $status = $output_hash->{status};
    if ($status eq "failed") {
        $self->_set_err("ZAPI failed: ", $output_hash->{reason});
        return $ClientUtil::FALSE;
    }
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);
    if (exists $output_hash->{'min-size'}) {
        $min_size = $output_hash->{'min-size'};
    } else {
        $self->_set_err("Unable to get size min-size via zapi");
        return $ClientUtil::FALSE;
    }
    return $min_size;
} ## end sub get_minsize
################################################################################
# Subroutine Name: set_select_attribute
# Objective:       Set the select attribute for the specified LUN
# Details:         See POD documentation at the end of this file
################################################################################
sub set_select_attribute() {
    my ($self, %args) = @_;
    my ($path,   $select_value, $filer,       $cmd_obj);
    my ($output, $status,       $output_hash, $resultsHash_ref);

    $self->_clear_err();

    # Supported only in zapi
    # parse arguments
    foreach (keys %args) {
        if (/-?select_value/) {
            if (not length($select_value = $args{$_})) {
                $self->_set_err("Invalid value for argument '-select_value'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $filer   = $self->get_filername();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $path    = $self->{path};

    #
    # Get the full lun path
    #
    unless (defined $path) {
        $self->_set_err("Lun path is missing from object");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd            => "lun-set-select-attribute",
                    -cmd_filer      => $filer,
                    -check_result   => 1,
                    -cmd_interface  => "zapi",
                    -path           => $path,
                    "-select-value" => $select_value,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # Since Interface is zapi, convert the output from xml to a string.
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    $status      = $output_hash->{status};
    if ($status eq "failed") {
        $self->_set_err("ZAPI failed: ", $output_hash->{reason});
        return $ClientUtil::FALSE;
    }
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);
    return $ClientUtil::TRUE;
} ## end sub set_select_attribute()
################################################################################
# Subroutine Name: destroy
# Objective:       destroy a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;
    my $force     = 0;
    my $path      = $self->{path};
    my $name      = $self->get_name();
    my $filer     = $self->get_filername();
    my $interface = $self->{parent}->{parent}->interface();
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my @luns;

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force and $force =~ /1|t|on|true|y|yes/i) {
                $force = 1;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    #
    # Get the full lun path
    #
    unless (defined $path) {
        $self->_set_err("Lun path is missing from object");
        return $ClientUtil::FALSE;
    }

    # make sure lun is not mapped (and offline)
    unless ($self->{mapped} == 0) {
        if ($force) {

            # offline the lun if it is online
            if ($self->is_online()) {
                $self->offline();
            }

            # check for error
            return $self if ($self->error());
        } else {
            $self->_set_err("Lun is currently mapped must use force or unmap");
            return $ClientUtil::FALSE;
        }
    } ## end unless ($self->{mapped} ==...

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-destroy",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $path,);
    $cmd_args{"-force"} = "1" if ($force and ($interface eq "zapi"));
    $cmd_args{"-force"} = ""  if ($force and ($interface eq "cli"));

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to destroy lun", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);

        # Message of success is not always returned depending on lun size.
        if (defined $output[0] and length $output[0]) {
            if ($output[0] =~ /lun destroy:/i) {
                $self->_set_err("Unable to destroy lun named '$name'", @output);
                return $ClientUtil::FALSE;
            }
        }
    } ## end else [ if ($interface eq "zapi")

    # Remove the lunObj from the filer volume object
    foreach my $lunObj (@{$self->{parent}->{luns}}) {
        if ($lunObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @luns, $lunObj;
        }
    }
    $self->{parent}->{luns} = \@luns;

    return $ClientUtil::TRUE;
} ## end sub destroy

################################################################################
# Subroutine Name: display()
# Objective:       Display Lun information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = delete($args{-path});

    # default the path to the internally set value if it exists
    if (!$path) {
        if ($self->{path}) {
            $path = $self->{path};
        } else {
            $self->_set_err("Missing required attribute '-path'");
            return (wantarray()) ? () : [];
        }
    }

    my $cmd  = "lun show -v " . $path;
    my @output = $self->{filer}->execute($cmd, -hide => 1);
    if ($self->{filer}->error()) {
        $self->_set_err($self->{filer}->errmsg());
        return $ClientUtil::FALSE;
    }

    push @output, "\n";
    return (wantarray()) ? @output : \@output;
} ## end sub display

################################################################################
# Subroutine Name: get_igroup
# Objective:       Return igroup lun is mapped too.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_igroup {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{igroup}) {
        return $self->{igroup};
    } else {
        $self->_set_err("Lun object is missing igroup");
        return $ClientUtil::FALSE;
    }
} ## end sub get_igroup

################################################################################
# Subroutine Name: get_filername
# Objective:       Return lun object filername
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filername {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filer}) {
        return $self->{filer}->get_name();
    } else {
        $self->_set_err("Lun object is missing filer name");
        return $ClientUtil::FALSE;
    }
} ## end sub get_filername

################################################################################
# Subroutine Name: get_geometry
# Objective:       Return lun geometry
# Details:         See POD documentation at the end of this file
################################################################################
sub get_geometry {
    my ($self) = @_;
    my (@output, $cmd, $path, $interface, $resultsHash_ref);
    my ($cmd_obj, $lun_geometry);

    $self->_clear_err();
    $path = $self->get_path();
    $interface = $self->{parent}->{parent}->interface();
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();

    if ($interface eq "cli") {
    	if (not defined $self->{geometry}) {
	    $self->{geometry} = {};

	    $cmd = "priv set diag;lun geometry " . $path;
	    @output = $self->{filer}->execute($cmd, -hide => 1);
	    if ($self->{filer}->error()) {
		$self->_set_err($self->{filer}->errmsg());
		return $ClientUtil::FALSE;
	    }

	    # parse output
	    foreach my $line (@output) {
		if ($line =~ /\s+(\d+)\s+(\S+)/) {
		    my $value   = $1;
		    my $geo_key = $2;

		    # replace / with _
		    $geo_key =~ s/\//_/;
		    $self->{geometry}{$geo_key} = $value;
		}
	    } ## end foreach my $line (@output)
	} ## end if (not defined $self->...

	return $self->{geometry};
    } else {
        # construct the command hash
	my %cmd_args = ("-cmd"           => "lun-get-geometry",
			"-cmd_filer"     => $self->get_filername(),
			"-check_result"  => 1,
			"-cmd_interface" => $interface,
			"-check_output"  => 1,
			"-path"          => $path,
			"-cmd_privilege" => "diag",);

	# Execute the command
	$resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

	#
	# Check for CmdsDriver errors (not command failure)
	#
	if ($cmd_obj->error()) {
	    my $errmsg = join "\n", $cmd_obj->errmsg();
	    $self->_set_err("Unable to get lun geometry: $errmsg");
	    $self->sanlog(-msg => $errmsg);
	    return $ClientUtil::FALSE;
	} # end if ($cmd_...

	$self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

	#
	# Get the output
	#
	my $output = $resultsHash_ref->{output};

	#
	# Since we checked the result, lets see if our check was OK
	#
	unless ($resultsHash_ref->{result_check}) {
	    $self->_set_err("Result check failed: $output");
	    return $ClientUtil::FALSE;
	} # end unless ($resultsHash_...
	unless ($resultsHash_ref->{output_check}) {
	    $self->_set_err("Output check failed: $output");
	    return $ClientUtil::FALSE;
	} # end unless ($resultsHash_...

	#
	# since zapi convert xml to hash
	#
	my $output_hash = $self->{parent}->{parent}->xml2hash($output);

	# preparing hash which is similar to cli
		$lun_geometry->{'cylinders'} = $output_hash->{'cylinders'};
		$lun_geometry->{'max'} = $output_hash->{'max-resize-size'};
		$lun_geometry->{'bytes_sector'} = $output_hash->{'bytes-per-sector'};
		$lun_geometry->{'device'} = $output_hash->{'size'};
		$lun_geometry->{'sectors_track'} = $output_hash->{'sectors-per-track'};
		$lun_geometry->{'tracks_cylinder'} = $output_hash->{'tracks-per-cylinder'};

		return $lun_geometry;
    } # end if ($interface...

} ## end sub get_geometry()

################################################################################
# Subroutine Name: get_stats_counter
# Objective:       Return counter from SCSI error
# Details:         See POD documentation at the end of this file
################################################################################
sub get_stats_counter {
    my ($self) = @_;
    my (@output, $cmd, $path, $cnt, @keys, @values, %stats_counter,);

    $self->_clear_err();

    $path = $self->get_path();

    $cmd = "lun stats -e " . $path;

    @output = $self->{filer}->execute($cmd, -hide => 1);
    if ($self->{filer}->error()) {
        $self->_set_err($self->{filer}->errmsg());
        return $ClientUtil::FALSE;
    }

    # parse output
    foreach my $line (@output) {
        next if ($line =~ /\s+\/vol\//);    # skip line with lun path
                                            # strip beginning space
        $line =~ s/^\s+(.*)/$1/;
        if ($line =~ /WriteP\s+Enospc.*/) {
            @keys = split(/\s+/, $line);
        } elsif ($line =~ /\d+\s+\d+.*/) {
            @values = split(/\s+/, $line);
        }
    } ## end foreach my $line (@output)
    for ($cnt = 0 ; $cnt < scalar(@keys) ; $cnt++) {
        $stats_counter{$keys[$cnt]} = $values[$cnt];
    }

    return \%stats_counter;
} ## end sub get_stats_counter
################################################################################
# Subroutine Name: get_stats
# Objective:       Return lun stats
# Details:         See POD documentation at the end of this file
################################################################################
sub get_stats {
    my ($self) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $info_hash;
    my %lun_stats_hash;
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();

    $self->_clear_err();

    my %cmd_args = ("-cmd"           => "lun-stats-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-extended"      => "",);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to get lun stats: $errmsg");
        $self->sanlog(-msg => $errmsg);
        return 0;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return 0;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return 0;
    }

    #
    # If we get here, we have results back from the filer.  If it's a zapi
    # we want to set the info_hash to the lun-info hash in the luns hash.
    # If it's cli, we want to parse the data line by line and fill in the
    # fields of the info_hash matching the field names from the zapi
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        unless (defined $output_hash->{"lun-stats"}->{"lun-stats-info"}) {
            $self->_set_err(  "Unable to parse output: {lun-stats}->{lun-stats-info} "
                            . "does not exist");
            return "";
        }
        %lun_stats_hash = %{$output_hash->{"lun-stats"}->{"lun-stats-info"}};
    } else {

        #
        # CLI case
        # We are expecting output that looks like this:
        #    /vol/vol0/lun1  (1 hour, 41 minutes, 50 seconds)
        # Read (kbytes)   Write (kbytes)  Read Ops  Write Ops  Other Ops  QFulls  Partner Ops Partner KBytes
        #  0               0               0         0          0          0       0           0
        #
        # In order to make the CLI hash similar to the ZAPI hash, this is returning
        # elements in units of blocks.  ZAPI returns a block size parameter.  In
        # the CLI case, the block-size parameter is set to 1024, since the units
        # are kilobytes.
        #
        my @output_array = split(/\n/, $output);
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace
            if ($line =~ /^\/vol/) {

                # do nothing - this is the LUN line - we already know the path
            } elsif ($line =~ /^Read/) {

                # again do nothing - we want the stats
            } elsif ($line =~
                /(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/)
            {

                # this is the data line - this is what we want
                $lun_stats_hash{'block-size'}     = "1024";
                $lun_stats_hash{'read-blocks'}    = $1;
                $lun_stats_hash{'write-blocks'}   = $2;
                $lun_stats_hash{'read-ops'}       = $3;
                $lun_stats_hash{'write-ops'}      = $4;
                $lun_stats_hash{'other-ops'}      = $5;
                $lun_stats_hash{'qfulls'}         = $6;
                $lun_stats_hash{'partner-ops'}    = $7;
                $lun_stats_hash{'partner-blocks'} = $8;
            } ## end elsif ($line =~ ...
        }    # end foreach line
    }    # end if CLI
    return \%lun_stats_hash;
}    # end get_stats
################################################################################
# Subroutine Name: get_name
# Objective:       Return lun object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Lun object is missing name");
        return $ClientUtil::FALSE;
    }
} ## end sub get_name

################################################################################
# Subroutine Name: get_path
# Objective:       Return lun object path
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{path}) {
        return $self->{path};
    } else {
        $self->_set_err("Lun object is missing path");
        return $ClientUtil::FALSE;
    }
} ## end sub get_path
################################################################################
# Subroutine Name: get_volname
# Objective:       Return lun's volume name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_volname {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{parent}) {
        return $self->{parent}->get_name();
    } else {
        $self->_set_err("Lun object is missing volume name");
        return $ClientUtil::FALSE;
    }
} ## end sub get_volname

################################################################################
# Subroutine Name: get_state
# Objective:       Return lun state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Lun object is missing state");
        return $ClientUtil::FALSE;
    }
} ## end sub get_state

################################################################################
# Subroutine Name: get_access
# Objective:       Return lun access
# Details:         See POD documentation at the end of this file
################################################################################
sub get_access {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{access}) {
        return $self->{access};
    } else {
        $self->_set_err("Lun object is missing access");
        return $ClientUtil::FALSE;
    }
} ## end sub get_access

################################################################################
# Subroutine Name: get_id
# Objective:       Return lun id
# Details:         See POD documentation at the end of this file
################################################################################
sub get_id {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{lunid}) {
        return $self->{lunid};
    } else {
        $self->_set_err("Lun object is not mapped to a host");
        return $ClientUtil::FALSE;
    }
} ## end sub get_id
################################################################################
# Subroutine Name: get_maps
# Objective:       Return igroup names lun is mapped to
# Details:         See POD documentation at the end of this file
################################################################################
sub get_maps {
    my ($self) = @_;
    my (@names);

    $self->_clear_err();

    if (defined $self->{maps}) {

        @names = (keys %{$self->{maps}});
    } else {

        return (wantarray()) ? () : \();
    }

    return (wantarray()) ? @names : \@names;

} ## end sub get_maps

################################################################################
# Subroutine Name: get_mapid
# Objective:       Return lun id igroup is mapped to
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mapid {
    my ($self, $name) = @_;
    my ($id,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("igroup name is a required argument");
        return $ClientUtil::FALSE;
    }

    if (defined $self->{maps}) {

        $id = $self->{maps}{$name};
    } else {

        $id = "";
    }

    return $id;
} ## end sub get_mapid

################################################################################
# Subroutine Name: get_size
# Objective:       Return lun size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Lun object is missing size");
        return $ClientUtil::FALSE;
    }
} ## end sub get_size

################################################################################
# Subroutine Name: get_parent
# Objective:       return parent (Volume) object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_parent {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{parent}) {
        return $self->{parent};
    }
    $self->_set_err("Lun object has no parent");
    return \{};
} ## end sub get_parent

################################################################################
# Subroutine Name: reset_stats_counter
# Objective:       Return true if mapped to a host
# Details:         See POD documentation at the end of this file
################################################################################
sub reset_stats_counter {
    my ($self) = @_;

    $self->_clear_err();

    my $path = $self->{path};
    my $cmd  = "lun stats -z " . $path;

    my @output = $self->{filer}->execute($cmd, -hide => 1);
    if ($self->{filer}->error()) {
        $self->_set_err($self->{filer}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub reset_stats_counter

################################################################################
# Subroutine Name: is_mapped
# Objective:       Return true if mapped to a host or a specified igroup
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mapped {
    my ($self, $igroup_name) = @_;

    $self->_clear_err();

    if (defined $self->{mapped}) {

        # if the user passed in an igroup_name, see if
        # this lun is mapped to that igroup
        if (defined $igroup_name) {
            foreach my $igroup_map (keys %{$self->{maps}}) {
                if (($igroup_name eq $igroup_map)) {
                    return $self->{mapped};
                }
            }

            # if the igroup name is not found in the maps hash, then return false
            return $ClientUtil::FALSE;
        } else {

            # otherwise just return if the lun is mapped or not
            return $self->{mapped};
        }
    } else {
        $self->_set_err("Lun object is not mapped");
        return $ClientUtil::FALSE;
    }
} ## end sub is_mapped

################################################################################
# Subroutine Name: map()
# Objective:       Map a lun to an igroup
# Details:         See POD documentation at the end of this file
################################################################################
sub map {
    my ($self, $igroup, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $lun_id;
    my $force = 0;

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args) {
        if (/-?lun_id/) {
            if (not length($lun_id = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force and $force =~ /1|t|on|true|y|yes/i) {
                $force = 1;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined $igroup and length $igroup) {
        $self->_set_err("igroup name is a required argument");
        return $ClientUtil::FALSE;
    }

    # Build command arguments.
    my %cmd_args = ("-cmd"             => "lun-map",
                    "-cmd_filer"       => $filer,
                    "-cmd_interface"   => $interface,
                    "-check_result"    => 1,
                    "-initiator-group" => $igroup,
                    "-path"            => $path,);

    $cmd_args{"-lun-id"} = $lun_id if (defined $lun_id);
    $cmd_args{"-force"} = "1" if ($force and ($interface eq "zapi"));
    $cmd_args{"-force"} = ""  if ($force and ($interface eq "cli"));

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    if ($interface eq "zapi") {

        # Since zapi, convert the output from xml to a string.
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);

        my $status = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to map lun $path" . " to $igroup ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }

        # set lun as mapped
        $self->{mapped} = 1;

        # retrieve lun id
        $self->{lunid} = $output_hash->{'lun-id-assigned'};

        # store this igroup in the maps hash
        $self->{maps}->{$igroup} = $self->{lunid};

        # since the user passed in the igroup, set it in the lun
        # object
        $self->{igroup} = $igroup;

    } else {

        my @output = split("\n", $output);

        if ($output[0] =~ /no such initiator/i) {
            $self->_set_err("igroup does not exist", @output);
            return $ClientUtil::FALSE;
        }

        if ($output[0] =~ /already mapped/i) {
            $self->_set_err(@output);
            return $ClientUtil::FALSE;
        }

        my @igroup_info = $self->map_list_info();

        if (scalar(@igroup_info) == 0) {
            $self->_set_err("no map list info for lun");
            return $ClientUtil::FALSE;
        }

        foreach my $entry (@igroup_info) {
            if ($entry->{'initiator-group-name'} eq $igroup) {
                if (defined $entry->{'lun-id'}) {
                    $self->{lunid} = $entry->{'lun-id'};
                } else {
                    $self->_set_err("no lun id entry for $igroup");
                    return $ClientUtil::FALSE;
                }
            }
        } ## end foreach my $entry (@igroup_info)

        if (!defined $self->{lunid}) {
            $self->_set_err("no lun id for $igroup");
            return $ClientUtil::FALSE;
        }

        # set lun as mapped
        $self->{mapped} = 1;

        # store this igroup in the maps hash
        $self->{maps}->{$igroup} = $self->{lunid};

        # since the user passed in the igroup, set it in the lun
        # object
        $self->{igroup} = $igroup;
    } ## end else [ if ($interface eq "zapi")

    return $ClientUtil::TRUE;
} ## end sub map
################################################################################
# Subroutine Name: unmap()
# Objective:       Unmap a lun from an igroup
# Details:         See POD documentation at the end of this file
################################################################################
sub unmap {
    my ($self, $igroup) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    unless (defined $igroup and length $igroup) {
        $self->_set_err("igroup is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Unmapping Lun $path from igroup $igroup");

    # Build command arguments.
    my %cmd_args = ("-cmd"             => "lun-unmap",
                    "-cmd_filer"       => $filer,
                    "-cmd_interface"   => $interface,
                    "-check_result"    => 1,
                    "-initiator-group" => $igroup,
                    "-path"            => $path,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to unmap $igroup" . " from  lun $path ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    # Remove the igroup from the objects igroup map
    delete $self->{maps}{$igroup};

    # Check if other maps exist for this lun.  If so
    # assign the first igroup and id to the object
    # metadata.
    if (keys %{$self->{maps}}) {
        $self->{igroup} = (keys %{$self->{maps}})[0];
        $self->{lunid}  = (values %{$self->{maps}})[0];
    } else {
        $self->{lunid}  = "";
        $self->{igroup} = "";
        $self->{mapped} = 0;
    }

    return $ClientUtil::TRUE;
} ## end sub unmap
################################################################################
# Subroutine Name: offline()
# Objective:       Takes a lun offline
# Details:         See POD documentation at the end of this file
################################################################################
sub offline {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    $self->sanlog(-msg => "Setting Lun $path offline");

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-offline",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to set lun " . " $path offline ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    $self->{state} = "offline";

    return $ClientUtil::TRUE;
} ## end sub offline

################################################################################
# Subroutine Name: online()
# Objective:       Puts a lun online
# Details:         See POD documentation at the end of this file
################################################################################
sub online {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $force     = 0;

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force and $force =~ /1|t|on|true|y|yes/i) {
                $force = 1;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $self->sanlog(-msg => "Setting Lun $path online");

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-online",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,);

    $cmd_args{"-force"} = "1" if ($force and ($interface eq "zapi"));
    $cmd_args{"-force"} = ""  if ($force and ($interface eq "cli"));

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to set lun" . " $path online ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } ## end if ($interface eq "zapi")
    $self->{state} = "online";

    return $ClientUtil::TRUE;
} ## end sub online

################################################################################
# Subroutine Name: is_offline()
# Objective:       Check the state of the lun
# Details:         See POD documentation at the end of this file
################################################################################
sub is_offline {
    my ($self) = @_;

    $self->_clear_err();

    my $path = $self->{path};
    my $cmd  = "lun show " . $path;

    my @output = $self->{filer}->execute($cmd, -hide => 1);
    if ($self->{filer}->error()) {
        $self->_set_err($self->{filer}->errmsg());
        return $ClientUtil::FALSE;
    }

    # we need to search the entire output array for the string
    # in the case of vfilers, this string will be several lines down
    if ("$output[0]" =~ /(No such LUN exists|lun pathname should start)/i) {
        $self->_set_err("Invalid LUN or LUN does not exist", @output);
        return $ClientUtil::FALSE;
    }

    # if the output contains the string offline we assume the filer is
    # offline and then save and return the state
    # Need to account for luns that might have "offline" in the lun path
    $self->{state} = ("@output" =~ /\(\S+\s+offline/i) ? "offline" : "online";
    return ($self->{state} =~ /offline/) ? 1 : 0;
} ## end sub is_offline

################################################################################
# Subroutine Name: is_online()
# Objective:       Check the state of the lun
# Details:         See POD documentation at the end of this file
################################################################################
sub is_online {
    my ($self) = @_;

    $self->_clear_err();

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $info_hash;
    my $path = $self->{path};

    #
    # Create the list of arguments for the CmdsDriver call.
    #
    my %cmd_args = ("-cmd"           => "lun-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 0,
                    "-path"          => $path,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # If we get here, we have results back from the filer.  If it's a zapi
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        unless (defined $output_hash->{"luns"}->{"lun-info"}) {
            $self->_set_err(  "Unable to parse output: {luns}->{lun-info} "
                            . "does not exist");
            return $ClientUtil::FALSE;
        }
        $self->{state} =
          ($output_hash->{"luns"}->{"lun-info"}->{"online"} =~ /true/i)
          ? "online"
          : "offline";
    } else {

        #
        # CLI case
        # We are expecting output that looks like this:
        # /vol/blocks_fvt/cloned_lun 8m (8388608) (r/w, offline, mapped, shared)
        #
        # if the output contains the string online we assume the lun is
        # online and then save and return the state
        # need to accommodate case where lun name has "online" in the string
        my @output = split(/\n/, $output);
        if ("@output" =~ /(No such LUN exists|lun pathname should start)/i) {
            $self->_set_err("Invalid LUN or LUN does not exist", @output);
            return $ClientUtil::FALSE;
        }

        $self->{state} =
          ("@output" =~ /\(\S+\s+online/i) ? "online" : "offline";
    }    # end if CLI

    return ($self->{state} =~ /online/) ? 1 : 0;
} ## end sub is_online
################################################################################
# Subroutine Name: move()
# Objective:       Moves lun to a new path or new lun name
# Details:         See POD documentation at the end of this file
################################################################################
no warnings qw(redefine);
sub move {
    my ($self, %args) = @_;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $new_path;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?path/) {
            if (not length($new_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $new_path and length $new_path) {
        $self->_set_err("-path is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Moving lun from $path to $new_path");

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-move",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,
                    "new-path"       => $new_path,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to move lun from" . " $path to $new_path ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    ### Check to see if just the name changed
    my $old_name = basename($path);
    my $new_name = basename($new_path);

    if ($old_name ne $new_name) {
        $self->sanlog(
             -msg => "Moved lun from old name $old_name to new name $new_name");
        $self->{name} = $new_name;
    }

    $self->{path} = $new_path;

    return $ClientUtil::TRUE;
} ## end sub move
use warnings qw(redefine);

################################################################################
# Subroutine Name: start_dense_clone()
# Objective:       Starts a dense clone operation
# Details:         See POD documentation at the end of this file
################################################################################
sub start_dense_clone {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $src_path;
    my $dest_path;
    my $nosnap;
    my $change_log;
    my $snapshot;
    my $blk_ranges;
    my $blk_num_src;
    my $blk_num_dest;
    my $blk_num_cnt;
    my $timeout = 0;    # 0 means don't wait for start completion
    my $clone_op_id;

    $self->_clear_err();

    #
    # Get the lun path
    #
    $src_path = $self->get_path();

    $self->sanlog(-msg => "Lun::start_dense_clone() Starting dense "
                  . "clone for lun $src_path");

    foreach (keys %args) {
        if (/-?dest_path/) {
            if (not length($dest_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dest_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?nosnap/) {
            if (not length($nosnap = $args{$_})) {
                $self->_set_err("Invalid value for argument '-nosnap'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?change_log/) {
            if (not length($change_log = $args{$_})) {
                $self->_set_err("Invalid value for argument '-change_log'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?block_ranges/) {
            if (not length($blk_ranges = $args{$_})) {
                $self->_set_err("Invalid value for argument '-block_ranges'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout'");
                return $ClientUtil::FALSE;
            }
            next;
	} elsif (/-?snapshot/) {
            if (not length($snapshot = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if ($self->{parent}->{parent}->interface() eq "cli") {
        my %cmd_args = ("-cmd",           "clone-start",
                        "-cmd_filer",     $self->get_filername(),
                        "-check_result",  1,
                        "-cmd_interface", "cli",
                        "-check_output",  0,
                        "-source-path",   $src_path,
                        "-cmd_privilege", "advanced",);
        $cmd_args{"-destination-path"} = $dest_path  if (defined $dest_path);
        $cmd_args{"-block-ranges"}     = $blk_ranges if (defined $blk_ranges);
        $cmd_args{"-change-log"}       = ""          if (defined $change_log);
        $cmd_args{"-nosnap"}           = ""          if (defined $nosnap);
	$cmd_args{"-snapshot-name"}    = $snapshot   if (defined $snapshot);
        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } else {
        my @b_ranges;

        #
        # $blk_ranges should have this format:
        # blk_num_src:blk_num_dest:blk_num_cnt ...
        #
        if (defined($blk_ranges)) {
            my @array = split(/\s+/, $blk_ranges);
            foreach my $b_range (@array) {
                ($blk_num_src, $blk_num_dest, $blk_num_cnt) =
                  split(':', $b_range);
                push(@b_ranges,
                     ({"block-count"              => $blk_num_cnt,
                       "destination-block-number" => $blk_num_dest,
                       "source-block-number"      => $blk_num_src}));
            }
        } ## end if (defined($blk_ranges...

        my %b_rangesHash = ();
        $b_rangesHash{"block-range"} = \@b_ranges;

        my %cmd_args = ("-cmd",           "clone-start",
                        "-cmd_filer",     $self->get_filername(),
                        "-check_result",  1,
                        "-cmd_interface", "zapi",
                        "-check_output",  0,
                        "-source-path",   $src_path,
                        "-cmd_privilege", "advanced",);
        $cmd_args{"-destination-path"} = $dest_path if (defined $dest_path);
        $cmd_args{"-block-ranges"} = \%b_rangesHash if (defined $blk_ranges);
        $cmd_args{"-change-log"}   = "true"         if (defined $change_log);
        $cmd_args{"-nosnap"}       = "true"         if (defined $nosnap);
	$cmd_args{"-snapshot-name"}= $snapshot      if (defined $snapshot);
        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } ## end else [ if ($self->{parent}->{...

    if ($cmd_obj->error()) {
        my $errmsg = join "\n", $cmd_obj->errmsg();
        $self->_set_err("Unable to start dense clone ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string in order to grab the $clone_op_id which is used later on
    #
    if ($self->{parent}->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);

        # my $dumpOut = Dumper($output_hash);
        # $self->sanlog(-msg       => "Output Hash is:\n$dumpOut",
        #               -msg_level => 9);
        my $status = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to start dense clone",
                            $output_hash->{reason});
            $self->sanlog(-msg => join("\n", $output_hash->{reason}));
            return $ClientUtil::FALSE;
        }

        #
        # Find the Clone ID in the output
        #
        if ($output =~ /.*<clone-op-id>(\d+)<.*/) {
            $clone_op_id = $1;
        } else {
            $self->_set_err("Unable to start dense clone $output");
            $self->sanlog(-msg => join("\n", $output));
            return $ClientUtil::FALSE;
        }
    } else {
        if (   $output =~ /clone start/i
            or $output =~ /Clone operation failed to start/i) {
            $self->_set_err("Unable to start dense clone $output");
            $self->sanlog(-msg => join("\n", $output));
            return $ClientUtil::FALSE;
        }

        #
        # Find the Clone ID in the output
        #
        if ($output =~
            /.*Clone operation started successfully\.\s+ID:\s+(\d+).*/) {
            $clone_op_id = $1;
        }
    } ## end else [ if ($self->{parent}->{...

    if ($timeout) {
        $self->wait_for_dense_clone_complete(-clone_op_id => $clone_op_id,
                                             -timeout     => $timeout);
        if ($self->error()) {
            $self->_set_err("Unable to start dense clone: ", $self->error());
            $self->sanlog(-msg => "Timed out");
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $clone_op_id;
    } else {
        $self->_set_err("Unable to start dense clone: result check failed");
        $self->sanlog(-msg => "Results did not match");
        return $ClientUtil::FALSE;
    }
}    # end start_dense_clone

################################################################################
# Subroutine Name: stop_dense_clone()
# Objective:       Stops a dense clone operation
# Details:         See POD documentation at the end of this file
################################################################################
sub stop_dense_clone {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $clone_op_id;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?clone_op_id/) {
            if (not length($clone_op_id = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_op_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Check for required clone_op_id
    unless (defined $clone_op_id and length $clone_op_id) {
        $self->_set_err(  "Missing or invalid value for "
                        . "required parameter '-clone_op_id'");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Lun::stop_dense_clone() Stopping dense "
                  . "clone $clone_op_id");

    if ($self->{parent}->{parent}->interface() eq "cli") {
        my %cmd_args = ("-cmd"           => "clone-stop",
                        "-cmd_filer"     => $self->get_filername(),
                        "-check_result"  => 1,
                        "-cmd_interface" => "cli",
                        "-check_output"  => 0,
                        "-volume-name"   => $self->get_volname(),
                        "-clone-id"      => $clone_op_id,
                        "-cmd_privilege" => "advanced",);

        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } else {

        my $vol_uuid;

        #
        # Need to get the uuid for the volume.  Call get_info() and parse
        # out the uuid
        #
        my $vol_info = $self->{parent}->get_info();
        if ($self->{parent}->error()) {
            $self->_set_err("Unable to get dense clone " . " status: ",
                            $self->{parent}->errmsg());
            $self->sanlog(-msg => join("\n", $self->{parent}->errmsg()));
            return $ClientUtil::FALSE;
        }

        if ($vol_info =~ /.*<uuid>(.*)<\/uuid>.*/) {
            $vol_uuid = $1;
        } else {
            $self->_set_err("Unable to get volume uuid");
            return $ClientUtil::FALSE;
        }

        my %cmd_args = ("-cmd",           "clone-stop",
                        "-cmd_filer",     $self->get_filername(),
                        "-check_result",  1,
                        "-cmd_interface", "zapi",
                        "-check_output",  0,
                        "-cmd_privilege", "advanced",);
        $cmd_args{"-clone-id"} = {'clone-id-info' => {
                                                  'volume-uuid' => $vol_uuid,
                                                  'clone-op-id' => $clone_op_id,
                                    }};

        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } ## end else [ if ($self->{parent}->{...

    if ($cmd_obj->error()) {
        $self->_set_err("Unable to stop dense clone: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);

        # my $dumpOut = Dumper($output_hash);
        # $self->sanlog(-msg      => "Output Hash is:\n$dumpOut",
        #               msg_level => 7);
        my $status = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to stop dense clone ",
                            $output_hash->{reason});
            $self->sanlog(-msg => join("\n", $output_hash->{reason}));
            return $ClientUtil::FALSE;
        }
    } else {
        if ($output =~ /clone stop/i) {
            $self->_set_err("Unable to stop dense clone $output");
            $self->sanlog(-msg => join("\n", $output));
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Unable to stop dense clone: result check failed");
        $self->sanlog(-msg => "Results did not match");
        return $ClientUtil::FALSE;
    }
}    # end stop_dense_clone

################################################################################
# Subroutine Name: clear_dense_clone()
# Objective:       Clears a dense clone operation
# Details:         See POD documentation at the end of this file
################################################################################
sub clear_dense_clone {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $clone_op_id;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?clone_op_id/) {
            if (not length($clone_op_id = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_op_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Check for required clone_op_id
    unless (defined $clone_op_id and length $clone_op_id) {
        $self->_set_err(  "Missing or invalid value for "
                        . "required parameter '-clone_op_id'");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Lun::clear_dense_clone() Clearing dense "
                  . "clone $clone_op_id");

    if ($self->{parent}->{parent}->interface() eq "cli") {
        my %cmd_args = ("-cmd"           => "clone-clear",
                        "-cmd_filer"     => $self->get_filername(),
                        "-check_result"  => 1,
                        "-cmd_interface" => "cli",
                        "-check_output"  => 0,
                        "-volume-name"   => $self->get_volname(),
                        "-clone-id"      => $clone_op_id,
                        "-cmd_privilege" => "advanced",);

        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } else {
        my %cmd_args = ("-cmd"           => "clone-clear",
                        "-cmd_filer"     => $self->get_filername(),
                        "-check_result"  => 1,
                        "-cmd_interface" => "zapi",
                        "-check_output"  => 0,
                        "-clone-id"      => $clone_op_id,
                        "-cmd_privilege" => "advanced",);

        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } ## end else [ if ($self->{parent}->{...

    if ($cmd_obj->error()) {
        $self->_set_err("Unable to clear dense clone: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # If the interface is zapi we have to convert the output from xml
    # to a string.
    #
    if ($self->{parent}->{parent}->{interface} eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);

        # my $dumpOut = Dumper($output_hash);
        # $self->sanlog(-msg      => "Output Hash is:\n$dumpOut",
        #               msg_level => 7);
        my $status = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to clear" . " dense clone ",
                            $output_hash->{reason});
            $self->sanlog(-msg => join("\n", $output_hash->{reason}));
            return $ClientUtil::FALSE;
        }
    } else {
        if ($output =~ /clone clear/i) {
            $self->_set_err("Unable to clear dense clone $output");
            $self->sanlog(-msg => join("\n", $output));
            return $ClientUtil::FALSE;
        }
    }

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Unable to clear dense clone: result check failed");
        $self->sanlog(-msg => "Results did not match");
        return $ClientUtil::FALSE;
    }
}    # end clear_dense_clone

################################################################################
# Subroutine Name: get_dense_clone_status()
# Objective:       Gets the status of a dense clone
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dense_clone_status {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $clone_op_id;
    my $return_output = 0;
    my $dc_hash;
    my $entry_hash;
    my $vol_uuid;

    $self->_clear_err();

    #
    # If the clone_op_id is not passed in, we will return the status
    # of all dense clones, however it will only display status of failed
    # or currently busy clones- if you wish to check if a clone managed to
    # start successfully you will have to specify the specific clone id if
    # it has already finished starting. Otherwise it will just return the
    # status of the single dense clone specified.
    #
    foreach (keys %args) {
        if (/-?clone_op_id/) {
            if (not length($clone_op_id = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clone_op_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?return_output/) {
            if (not length($return_output = $args{$_})) {
                $self->_set_err("Invalid value for argument '-return_output'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if (defined($clone_op_id)) {
        $self->sanlog(-msg => "Lun::get_dense_clone_status() Getting "
                      . "dense clone $clone_op_id status");
    } else {
        $self->sanlog(-msg => "Lun::get_dense_clone_status() Getting "
                      . "dense clone status for all dense clones");
    }

    if ($self->{parent}->{parent}->interface() eq "cli") {
        my %cmd_args = ("-cmd",           "clone-list-status",
                        "-cmd_filer",     $self->get_filername(),
                        "-check_result",  1,
                        "-cmd_interface", "cli",
                        "-check_output",  0,
                        "-cmd_privilege", "advanced",);

        #
        # Syntax is clone volname ID for specific clone ID's
        #
        $cmd_args{"-volume-name"} = $self->get_volname()
          if (defined $clone_op_id);
        $cmd_args{"-clone-id"} = $clone_op_id if (defined $clone_op_id);
        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } else {
        my %cmd_args = ("-cmd",           "clone-list-status",
                        "-cmd_filer",     $self->get_filername(),
                        "-check_result",  1,
                        "-cmd_interface", "zapi",
                        "-check_output",  0,
                        "-cmd_privilege", "advanced",);
        if (defined $clone_op_id) {

            #
            # Need to get the uuid for the volume.  Call get_info() and parse
            # out the uuid
            #
            my $vol_info = $self->{parent}->get_info();
            if ($self->{parent}->error()) {
                $self->_set_err("Unable to get dense clone " . " status: ",
                                $self->{parent}->errmsg());
                $self->sanlog(-msg => join("\n", $self->{parent}->errmsg()));
                return $ClientUtil::FALSE;
            }

            if ($vol_info =~ /.*<uuid>(.*)<\/uuid>.*/) {
                $vol_uuid = $1;
            } else {
                $self->_set_err("Unable to get volume uuid");
                return $ClientUtil::FALSE;
            }

            $cmd_args{"-clone-id"} = {'clone-id-info' => {
                                                  'volume-uuid' => $vol_uuid,
                                                  'clone-op-id' => $clone_op_id,
                                        }};
        } ## end if (defined $clone_op_id)
        my $dumpOut = Dumper(%cmd_args);
        $self->sanlog(-msg      => "Cmd Args is:\n$dumpOut",
                      msg_level => 3);
        $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    } ## end else [ if ($self->{parent}->{...

    if ($cmd_obj->error()) {
        $self->_set_err("Unable to get dense clone status: ",
                        $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    if ($self->{parent}->{parent}->interface() eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);

        # my $dumpOut = Dumper($output_hash);
        # $self->sanlog(-msg      => "Output Hash is:\n$dumpOut"),
        #               msg_level => 7);
        my $status = $output_hash->{status};

        if ($status eq "failed") {
            $self->_set_err("Unable to get status ", $output_hash->{reason});
            $self->sanlog(-msg => join("\n", $output_hash->{reason}));
            return $ClientUtil::FALSE;
        }

        #
        # If there is only a single entry handle that
        # The status hash contains a two element array, the first
        # being a string containing the status, and the second a
        # hash with 'ops-info'
        #
        if (ref($output_hash->{status}->[1]->{'ops-info'}) ne "ARRAY") {
            $entry_hash = $output_hash->{status}->[1]->{'ops-info'};
            my $id =
              $entry_hash->{'clone-id'}->{'clone-id-info'}->{'clone-op-id'};

            if ($clone_op_id and !(defined($id))) {
                $dc_hash->{$clone_op_id} = $entry_hash;
            } else {
                $dc_hash->{$id} = $entry_hash;
            }
        } else {

            # it's an array of hash
            foreach $entry_hash (@{$output_hash->{status}->[1]->{'ops-info'}}) {
                my $id =
                  $entry_hash->{'clone-id'}->{'clone-id-info'}->{'clone-op-id'};
                $dc_hash->{$id} = $entry_hash;
            }
        }
    } else {
        my @out = split('\n', $output);
        my $id;
        my $value;

        foreach my $line (@out) {
            if ($line =~ /ID:/) {
                (undef, $id) = split(/:\s*/, $line);
                $dc_hash->{$id}->{'clone-id-info'}->{'clone-op-id'} = $id;
            } elsif ($line =~ /Source:/) {
                (undef, $value) = split(/:\s*/, $line);
                $dc_hash->{$id}->{'source-file'} = $value;
            } elsif ($line =~ /Destination:/) {
                (undef, $value) = split(/:\s*/, $line);
                $dc_hash->{$id}->{'destination-file'} = $value;
            } elsif ($line =~ /Block ranges:/) {
                (undef, $value) = split(/:\s*/, $line);
                $dc_hash->{$id}->{'block-ranges'} = $value;
            } elsif ($line =~ /State:/) {
                my $reason;
                (undef, $value) = split(/:\s*/, $line);
                ($value, $reason) = split(/ \(/, $value);
                $reason =~ s/\)$//;    # get rid of last ')'

                $dc_hash->{$id}->{'clone-state'} = $value;
                $dc_hash->{$id}->{'reason'}      = $reason;
            } elsif ($line =~ /Type:/) {
                (undef, $value) = split(': ', $line);
                $dc_hash->{$id}->{'clone-type'} = $value;
            }
        } ## end foreach my $line (@out)
    } ## end else [ if ($self->{parent}->{...

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {
        if ($return_output == 1) {
            return $output;
        } else {
            return $dc_hash;
        }
    } else {
        $self->_set_err(
                       "Unable to get dense clone status: result check failed");
        $self->sanlog(-msg => "Results did not match");
        return $ClientUtil::FALSE;
    }
}    # end get_dense_clone_status

################################################################################
# Subroutine Name: wait_for_dense_clone_complete()
# Objective:       Waits for a dense clone to be created
# Details:         See POD documentation at the end of this file
################################################################################
sub wait_for_dense_clone_complete {
    my ($self, %args) = @_;

    #
    # If we were given a name and/or timeout use them, otherwise use defaults
    #
    my $clone_op_id = $args{-clone_op_id};
    my $timeout     = (defined $args{-timeout}) ? $args{-timeout} : 1200;
    my $interval    = (defined $args{-interval}) ? $args{-interval} : 10;

    # Check for required clone_op_id
    unless (defined $clone_op_id and length $clone_op_id) {
        $self->_set_err(  "Missing or invalid value for required "
                        . "parameter '-clone_op_id'");
        return $ClientUtil::FALSE;
    }

    my $start_time   = int(gettimeofday);
    my $current_time = int(gettimeofday);

    while (($current_time - $start_time) < $timeout) {
        $current_time = int(gettimeofday);
        &SanUtil::nate_sleep($interval);

        #
        # Currently this method simply uses zapi when getting the status,
        # so first we save the interface, then change it to zapi, do the
        # status call, then change the interface back to what it was
        # originally.
        #
        my $i = $self->{parent}->{parent}->interface();
        $self->{parent}->{parent}->interface("zapi");
        my $output =
          $self->get_dense_clone_status(-clone_op_id => $clone_op_id);
        $self->{parent}->{parent}->interface($i);

        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Lun::wait_for_dense_clone_complete() "
                        . "Current time is "
                        . $current_time
                        . "\nStart time is "
                        . $start_time
                        . "\nTimeout is "
                        . $timeout,
                      -msg_level => 3);

        #
        # If the dense clone exists, we are done, return.
        #
        my $state = $output->{$clone_op_id}->{'clone-state'};

        if ($state eq "completed") {
            return $ClientUtil::TRUE;
        }
    }    # end while

    $self->_set_err(
           "Timeout waiting for dense clone '$clone_op_id' to be " . "created");
    return $ClientUtil::FALSE;
}    # end wait_for_dense_clone_complete

################################################################################
# Subroutine Name: clone_create()
# Objective:       Create a lun clone backed by a snap
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_create {
    my ($self, %args) = @_;
    my $snap_name;
    my @output, my $noreserve = 0;
    my $interface = $self->{parent}->{parent}->interface();
    my $resultsHash_ref;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    $self->_clear_err();

    #
    # Create the default clone_path by prepending the
    # name of the lun in the lun path with "clone_".
    #
    my $lun_name   = $self->get_name();
    my $lun_path   = $self->get_path();
    my $clone_path = $lun_path;
    my $clone_name = "clone_" . "$lun_name";
    $clone_path =~ s/$lun_name/$clone_name/;

    foreach (keys %args) {
        if (/-?path/) {
            if (not length($clone_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?noreserve/) {
            if (not length($noreserve = $args{$_})
                or $noreserve !~ /yes|no|1|0/i) {
                $self->_set_err("Invalid value for argument '-noreserve'");
                return $ClientUtil::FALSE;
            }
            $noreserve = 1 if ($noreserve =~ /yes/i);
            $noreserve = 0 if ($noreserve =~ /no/i);
            next;
        } elsif (/-?snap_name/) {
            if (not length($snap_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snap_name'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # Check for required snap_name
    unless (defined $snap_name and length $snap_name) {
        $self->_set_err(  "Missing or invalid value for "
                        . "required parameter '-snap_name'");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"             => "lun-create-clone",
                    "-cmd_filer"       => $self->get_filername(),
                    "-check_result"    => 1,
                    "-check_output"    => 1,
                    "-cmd_interface"   => $interface,
                    "-path"            => $clone_path,
                    "-parent-snap"     => $snap_name,
                    "-parent-lun-path" => $lun_path,);

    if ($noreserve) {
        if ($interface eq "cli") {
            $cmd_args{"-noreserve"} = "";
        } else {
            $cmd_args{"-space-reservation-enabled"} = "false";    #zapi value
        }
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }

    #
    # Create a lun object for this clone and add it to the parent volume obj
    #
    my $info_hash_ref = $self->list_info(-path => $clone_path);
    my %create_args;
    if (defined $info_hash_ref->{path}) {
        $create_args{-path} = $info_hash_ref->{path};
        $create_args{-dir}  = dirname($info_hash_ref->{path});
        $create_args{-name} = basename($info_hash_ref->{path});
    }
    $create_args{-type} = $info_hash_ref->{"multiprotocol-type"};
    $create_args{-size} = $info_hash_ref->{size} . "b";
    $create_args{-state} =
      ($info_hash_ref->{online} =~ /true/) ? "online" : "offline";
    $create_args{-access} =
      ($info_hash_ref->{"read-only"} =~ /true/) ? "r" : "r/w";
    $create_args{-mapped} = ($info_hash_ref->{"mapped"} =~ /true/) ? "1" : "0";
    $create_args{-_exist} = 1;

    my $vol_obj     = $self->{parent};
    my $new_lun_obj = $vol_obj->create_lun(%create_args);
    if ($vol_obj->error()) {
        $self->_set_err("Could not create lun object for "
                          . $info_hash_ref->{path} . ": ",
                        $vol_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    return $new_lun_obj;
}    # end clone_create()

################################################################################
# Subroutine Name: clone_split_start()
# Objective:       Start to split a clone - it'll no longer be backed by a snap
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_split_start {
    my ($self) = @_;
    my @output, my $resultsHash_ref;
    my $interface  = $self->{parent}->{parent}->interface();
    my $cmd_obj    = $self->{parent}->{parent}->cmd_obj();
    my $clone_path = $self->get_path();

    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-clone-split-start",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $clone_path,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end clone_split_start()

################################################################################
# Subroutine Name: clone_split_stop()
# Objective:       Stop a lun clone split
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_split_stop {
    my ($self) = @_;
    my @output, my $resultsHash_ref;
    my $interface  = $self->{parent}->{parent}->interface();
    my $cmd_obj    = $self->{parent}->{parent}->cmd_obj();
    my $clone_path = $self->get_path();

    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-clone-split-stop",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $clone_path,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end clone_split_stop()

################################################################################
# Subroutine Name: clone_split_status()
# Objective:       Get and return the status for a lun clone
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_split_status {
    my ($self) = @_;
    my @output, my $resultsHash_ref;
    my $interface  = $self->{parent}->{parent}->interface();
    my $cmd_obj    = $self->{parent}->{parent}->cmd_obj();
    my $clone_path = $self->get_path();

    $self->_clear_err();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-clone-split-status-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $clone_path,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    # If the split is done, the result check will fail so we
    # also want to check to see if the output says "LUN is not a clone"
    #
    unless ($resultsHash_ref->{result_check}) {
        if ($output =~ /LUN is not a clone/) {
            $self->sanlog(  -msg => "Lun::clone_split_status(): Lun is not "
                          . "a clone.  Clone split operation "
                          . "probably completed");
        } else {
            $self->_set_err("Result check failed: $output");
            return $ClientUtil::FALSE;
        }
    } ## end unless ($resultsHash_ref->...

    return $output;
}    # end clone_split_status()

################################################################################
# Subroutine Name: clone_split_percent_complete()
# Objective:       Get and return the percent complete for clone split
# Details:         See POD documentation at the end of this file
################################################################################
sub clone_split_percent_complete {
    my ($self)     = @_;
    my $interface  = $self->{parent}->{parent}->interface();
    my $clone_path = $self->get_path();

    #
    # Get the clone split status and use the data to calculate or find % comp
    #
    my $status = $self->clone_split_status();

    my $percent_complete = "";
    if ($interface eq "zapi") {

        #
        # We are expecting ZAPI output to include:
        #   <blocks-completed>32872</blocks-completed>
        #   <blocks-total>255232</blocks-total>
        #
        # We calculate the percent complete by dividing the completed by
        # the total
        #
        my $blocks_complete;
        my $blocks_total;

        #
        # Check to see if it's no longer a clone, if so percent complete is
        # 100.  Note 9048 is returned when lun is not a clone and 9093 is
        # returned when LUN clone is not currently being split
        #
        if (   $status =~ qr/errno="9048"/
            or $status =~ qr/errno="9093"/
            or $status =~ qr/reason="LUN is not a clone"/) {
            $percent_complete = 100;
        } else {
            if ($status =~ m/<blocks-completed>(\d+)<\/blocks-completed>/) {
                $blocks_complete = $1;
            } else {
                $self->_set_err(
                              "Status did not include a blocks-complete value");
                return $ClientUtil::FALSE;
            }
            if ($status =~ m/<blocks-total>(\d+)<\/blocks-total>/) {
                $blocks_total = $1;
            } else {
                $self->_set_err("Status did not include a blocks-total value");
                return $ClientUtil::FALSE;
            }
            if ($blocks_total != 0) {
                $percent_complete =
                  int(($blocks_complete * 100) / $blocks_total);
            } else {

                #
                # If there are no total blocks, we are 100% done
                #
                $percent_complete = 100;
            }
        } ## end else [ if ($status =~ qr/errno="9048"/...
    } else {

        #
        # CLI output should contain the percent complete like this:
        #  lun clone split status: Done 40012 of 255232 blocks (15% complete).
        #
        if ($status =~ m/LUN is not being cloned/) {
            $percent_complete = 100;
        } elsif ($status =~ m/LUN is not a clone/) {
            $percent_complete = 100;
        } else {
            my @info = split(/\r*\n/, $status);
            foreach my $cmd_line (@info) {
                if ($cmd_line =~
                    m/lun clone split status:\s+Done\s+\d+\s+of\s+\d+\s+blocks\s+\((\d+)\% complete\)/
                  ) {
                    $percent_complete = $1;
                }
            }
            if ($percent_complete eq '') {
                $self->_set_err(  "Status did not include a "
                                . "a percent complete value ($status)");
                return $ClientUtil::FALSE;
            }
        } ## end else [ if ($status =~ m/LUN is not being cloned/)
    } ## end else [ if ($interface eq "zapi")
    return ($percent_complete);
}    # end clone_split_percent_complete()

################################################################################
# Subroutine Name: clone_split_wait()
# Objective:       Wait for splitting of clone to complete
# Details:
################################################################################
sub clone_split_wait {
    my ($self, %args) = @_;
    my $percent_complete = 0;
    my $timeout          = (defined $args{-timeout}) ? $args{-timeout} : 3600;
    my $endtime          = time() + $timeout;
    my $lun_path         = $self->get_path();

    $self->_clear_err();

    #
    # Loop until we timeout waiting for the lun clone split to complete
    #
    while ($endtime > time()) {
        $percent_complete = $self->clone_split_percent_complete();
        if ($self->error()) {
            $self->_set_err(  "Could not get "
                            . "clone split percent complete: "
                            . $self->errmsg());
            return $ClientUtil::FALSE;
        }
        $self->sanlog(  -msg => "Lun::clone_split_wait() Clone split for "
                      . "$lun_path is $percent_complete"
                      . "% complete");

        #
        # If we are done, stop waiting
        #
        if ($percent_complete == 100) {
            return $ClientUtil::TRUE;
        }
        &SanUtil::nate_sleep(20);
    }    # end while

    #
    # If we get here, we didn't find the lun clone to have completed
    # its split before the timeout.
    #
    $self->_set_err("Did not complete in $timeout seconds.");
    return $self;
} ## end sub clone_split_wait

################################################################################
# Subroutine Name: lun_clone_list_info()
# Objective:       Lists all LUN clones in the given snapshot.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_clone_list_info {
    my ($self, %args) = @_;

    my ($resultsHash_ref, $snap_name, $lun_clone_list_info);
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    # implemented only in zapi
    my $interface = "zapi";
    my $volume    = $self->get_volname();
    $self->_clear_err();

    $self->sanlog(-msg => "Lun::clone_list_info() Lists all LUN clones");

    #
    # Parse the arguements,
    #
    foreach (keys %args) {
        if (/-?snap_name/) {
            if (not length($snap_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snap_name'");
                return $ClientUtil::FALSE;
            } # end if (not length...
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        } # end if (/-?snap_name...
    } # end foreach...

    #
    # Check for required size argument
    #
    unless (defined($snap_name)) {
        $self->_set_err("Required 'snap_name' parameter not specified");
        return $ClientUtil::FALSE;
    } # end unless (defined....

    # construct the command hash
    my %cmd_args = ("-cmd"           => "lun-clone-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => $interface,
                    "-snapshot"      => $snap_name,
                    "-volume"        => $volume,);

    #
    # Execute the command
    #
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Unable to list of clone luns: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    } # end if ($cmd_...

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output.
    #
    my $output = $resultsHash_ref->{output};

    #
    # If interface is zapi, convert the output from xml to a hash.
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
	        $self->_set_err("Result check failed:\n", $output_hash->{reason});
	        return $ClientUtil::FALSE;
	    } else {
	        $lun_clone_list_info = $output_hash->{'lun-clone-list'}->
	                                {'lun-clone-lists-info'};
	        my $dumpOut = Dumper($lun_clone_list_info);
            $self->sanlog(-msg => "Lun::clone - list info:" . "\n$dumpOut",
                                                -msg_level => 7);
		} # end if ($output_hash...
    } # end if ($interface...

    return $lun_clone_list_info;
} ## end sub lun_clone_list_info()

################################################################################
# Subroutine Name: reset_stats
# Objective:       Resets (zeroes) block-protocol access statistics for LUN
# Details:         See POD documentation at the end of this file
################################################################################
sub reset_stats {
    my ($self) = @_;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();

    # implemented only in zapi
    my $interface = "zapi";
    my $resultsHash_ref;

    # construct the command hash
    my %cmd_args = ("-cmd"           => "lun-reset-stats",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => $interface,);

    #
    # Execute the command
    #
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Failed to reset lun stats: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    } # end if ($cmd_...

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output.
    #
    my $output = $resultsHash_ref->{output};

    #
    # If interface is zapi, convert the output from xml to a hash.
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
	if ($output_hash->{status} eq "failed") {
            $self->_set_err("Result check failed:\n", $output_hash->{reason});
            return $ClientUtil::FALSE;
        } else {
            return $ClientUtil::TRUE;
        } # end if ($output_hash...
    } # end if ($interface...

} ## end sub reset_stats()

################################################################################
# Subroutine Name: lun_snap_usage_list_info()
# Objective: 	   Lists all LUNs backed by data in the specified snapshot.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_snap_usage_list_info {
    my ($self, %args) = @_;

    my ($resultsHash_ref, $snap_name, $lun_snap_usage_info);
	my $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # implemented only in zapi
    my $interface = "zapi";
    my $volume    = $self->get_volname();

    $self->_clear_err();
    #
    # Parse the arguements,
    #
    foreach (keys %args) {
        if (/-?snap_name/) {
            if (not length($snap_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snap_name'");
                return $ClientUtil::FALSE;
            } # end if (not length...
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        } # end if (/-?snap_snap...
    } # end foreach...

    #
    # Check for required size argument
    #
    unless (defined($snap_name)) {
        $self->_set_err("Required 'snap_name' parameter not specified");
        return $ClientUtil::FALSE;
    } # end unless (defined...

    # construct the command hash
    my %cmd_args = ("-cmd"           => "lun-snap-usage-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => $interface,
                    "-snapshot"      => $snap_name,
                    "-volume"        => $volume,);

    #
    # Execute the command
    #
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Failed to list snap usage ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    } # end if ($cmd_...

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output.
    #
    my $output = $resultsHash_ref->{output};

    #
    # If interface is zapi, convert the output from xml to a hash.
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
	if ($output_hash->{status} eq "failed") {
            $self->_set_err("Result check failed:\n", $output_hash->{reason});
            return $ClientUtil::FALSE;
        } else {
            $lun_snap_usage_info = $output_hash->{'lun-snap-usage-luns'};
            my $dumpOut = Dumper($lun_snap_usage_info);
            $self->sanlog(-msg => "Lun::snap_usage - list info:" . "\n$dumpOut",
                                                -msg_level => 7);
		} # end if ($output_hash...
    } # end if ($interface...

    return $lun_snap_usage_info;

} ## end sub lun_snap_usage_list_info()

################################################################################
# Subroutine Name: resize()
# Objective:       Resize the lun
# Details:         See POD documentation at the end of this file
################################################################################
sub resize {
    my ($self, %args) = @_;
    my $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    my $resultsHash_ref;
    my $path  = $self->get_path();
    my $force = 0;
    my $size;
    my $interface = $self->{parent}->{parent}->interface();

    $self->_clear_err();

    $self->sanlog(-msg => "Lun::resize() Resizing lun $path");

    #
    # Parse the arguements, overwrite defautls if necessary
    #
    foreach (keys %args) {
        if (/-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-size'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = 1 if ($force =~ /yes/i);
            $force = 0 if ($force =~ /no/i);
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    #
    # Check for required size argument
    #
    unless (defined($size)) {
        $self->_set_err("Required 'size' parameter not specified");
        return $ClientUtil::FALSE;
    }

    my %cmd_args = ("-cmd"           => "lun-resize",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 0,
                    "-path"          => $path,
                    "-size"          => $size,);

    #
    # Convert size to bytes if zapi
    #
    if (lc($interface) eq "zapi") {

        # if size is not in bytes, convert to bytes
        unless ($size =~ /^\d+$/) {
            $size = &ClientUtil::convert_to_bytes($size);
            $cmd_args{"-size"} = $size;
        }
    }

    $cmd_args{"-force"} = "1" if ($force and ($interface =~ m/zapi/i));
    $cmd_args{"-force"} = ""  if ($force and ($interface =~ m/cli/i));

    #
    # Execute the command
    #
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("Unable to resize lun: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output.
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    if ($resultsHash_ref->{result_check}) {

        #
        # Also, reset the size attribute before we return
        #
        my $info_hash_ref = $self->list_info(-path => $path);

        if ($self->error()) {
            $self->{size} = $size;
        } else {
            unless (defined $info_hash_ref->{size}) {
                my $dumpOut = Dumper($info_hash_ref);
                $self->sanlog(-msg => "Lun::resize - list info:" . "\n$dumpOut",
                              -msg_level => 7);
            }
            $self->{size} = $info_hash_ref->{size} . "b";
        }
        return $ClientUtil::TRUE;
    } else {

        #
        # If interface is zapi, convert the output from xml to a hash.
        #
        if ($interface eq "zapi") {
            my $output_hash = $self->{parent}->{parent}->xml2hash($output);
            $self->_set_err("Result check failed:\n", $output_hash->{reason});
            return $ClientUtil::FALSE;
        } else {
            $self->_set_err("Result check failed:\n", $output);
            return $ClientUtil::FALSE;
        }
    } ## end else [ if ($resultsHash_ref->...
}    # end resize()

################################################################################
# Subroutine Name: get_comment()
# Objective:       Get the comment on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_comment {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $comment;

    $self->_clear_err();

    $self->sanlog(-msg => "Get comment for $path");

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-get-comment",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("unable to get lun comment" . " for $path ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
        $comment = $output_hash->{'comment'};
    } else {
        $comment = $output;
        $comment =~ s/=====\s+\S+//;    #remove vfiler header lines
        $comment =~ s/Comment:\s+//;
        $comment =~ s/^\s+//;           # remove leading whitespace
        $comment =~ s/\s+$//;           # remove trailing whitespace
        if ($comment =~ /^".*"$/) {
            $comment =~ s/^"(.*)"$/$1/;    # remove quotes from comment
        }
    } ## end else [ if ($interface eq "zapi")

    return $comment;
}    # end get_comment

################################################################################
# Subroutine Name: set_comment()
# Objective:       Set the comment on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_comment {
    my ($self, %args) = @_;

    my $comment;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?comment/) {
            if (not length($comment = $args{$_})) {
                $self->_set_err("Invalid value for argument '-comment'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined $comment and length $comment) {
        $self->_set_err("comment is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Setting the comment to $comment on Lun $path");

    my %cmd_args = ("-cmd"           => "lun-set-comment",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-comment"       => $comment,
                    "-com-end"       => "",);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to set lun comment" . " as $comment ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}    # end set_comment

################################################################################
# Subroutine Name: unset_comment()
# Objective:       Unset the comment on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub unset_comment {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    $self->sanlog(-msg => "Unset the comment on Lun $path");

    #my %cmd_args = ("-cmd"           => "lun-set-comment",
    #                "-cmd_filer"     => $filer,
    #                "-check_result"  => 1,
    #                "-cmd_interface" => $interface,
    #                "-check_output"  => 1,
    #                "-path"          => $path,
    #                "-comment"       => "",
    #                "-com-end"       => "",);

	my %cmd_args = (
        "-cmd"           => "lun-unset-attribute",
        "-cmd_filer"     => $filer,
        "-check_result"  => 1,
        "-cmd_interface" => $interface,
        "-check_output"  => 1,
        "-path"          => $path,
        "-name"          => "comment",
        "-cmd_privilege" => "diag",
    );

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end unset_comment

################################################################################
# Subroutine Name: set_attribute()
# Objective:       Set an attribute on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_attribute {
    my ($self, %args) = @_;

    my $attribute_name;
    my $attribute_value;
    my $force;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?name/) {
            if (not length($attribute_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?value/) {
            if (not length($attribute_value = $args{$_})) {
                $self->_set_err("Invalid value for argument '-value'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $attribute_name and length $attribute_name) {
        $self->_set_err("-name is a required argument");
        return $ClientUtil::FALSE;
    }

    unless (defined $attribute_value and length $attribute_value) {
        $self->_set_err("-value is a required argument");
        return $ClientUtil::FALSE;
    }

    my $msg = "Setting an attribute $attribute_name with value"
      . " $attribute_value on Lun $path";
    $self->sanlog(-msg => join("\n", $msg));

    my %cmd_args = ("-cmd"           => "lun-set-attribute",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-name"          => $attribute_name,
                    "-value"         => $attribute_value,
                    "-path"          => $path,);

    if ($interface eq "cli") {
        $cmd_args{"-cmd_privilege"} = "diag";
    }

    if ($force) {
        $cmd_args{"-force"} = "";
    }

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure)
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to set lun attribute "
                              . "$attribute_name with value $attribute_value "
                              . "on $path",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } ## end if ($interface eq "zapi")

    return $ClientUtil::TRUE;

}    # end set_attribute

###############################################################################
# Subroutine Name: unset_attribute()
# Objective:       Unset an attribute on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub unset_attribute {
    my ($self, %args) = @_;

    my $attribute_name;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?name/) {
            if (not length($attribute_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined $attribute_name and length $attribute_name) {
        $self->_set_err("name is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Unset attribute $attribute_name on lun $path");

    my %cmd_args = ("-cmd"           => "lun-unset-attribute",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-name"          => $attribute_name,
                    "-cmd_privilege" => "diag",);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}    # end unset_attribute

################################################################################
# Subroutine Name: get_serial()
# Objective:       Get the serial number on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_serial {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $path      = $self->get_path();
    my $serial_number;

    $self->sanlog(-msg => "Getting lun serial");

    $self->_clear_err();

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-get-serial-number",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,
                    "-cmd_privilege" => "diag",);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    if ($interface eq "zapi") {

        # If interface is zapi, convert the output from xml to a string.
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to get lun serial ",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }

        if (defined $output_hash->{'serial-number'}) {
            $serial_number = $output_hash->{'serial-number'};
        } else {
            $self->_set_err("Unable to obtain lun serial number via zapi");
            return $ClientUtil::FALSE;
        }
    } else {

        # Parse cli output.
        my @output_array = split(/\n/, $output);
        my $lun_hash = {};
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace

            if ($line =~ /^Serial/) {
                $line =~ s/Serial#:\s+//;    # strip off the prefix
                $lun_hash->{"serial-number"} = $line;
            }
        }

        if (defined $lun_hash->{"serial-number"}) {
            $serial_number = $lun_hash->{"serial-number"};
        } else {
            $self->_set_err("Unable to obtain lun serial number via cli");
            return $ClientUtil::FALSE;
        }
    } ## end else [ if ($interface eq "zapi")

    return $serial_number;

}    # end get_serial

################################################################################
# Subroutine Name: get_attribute()
# Objective:       Get an attribure on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_attribute {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $path      = $self->get_path();
    my $attribute_name;
    my $attribute_value;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?name/) {
            if (not length($attribute_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if ($interface eq "zapi") {

        # If the interface is zapi, then an attribute has to be specified
        unless (defined $attribute_name and length $attribute_name) {
            $self->_set_err("-name is a required argument");
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "Get attribute $attribute_name for lun $path");

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-get-attribute",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,);

    if ($interface eq "cli") {
        $cmd_args{"-cmd_privilege"} = "diag";
    }
	# If the attribute name is specified, then add the attribute name
	if (defined $attribute_name and length $attribute_name) {
		$cmd_args{"-name"} = $attribute_name;
	}

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    if ($interface eq "zapi") {

        # If interface is zapi, convert the output from xml to a string.
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to get lun attribute value"
                              . " for $attribute_name "
                              . " via zapi",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }

        if (defined $output_hash->{value}) {
            $attribute_value = $output_hash->{value};
        } else {
            $self->_set_err(  "Unable to get lun attribute value"
                            . " for $attribute_name via zapi");
            return $ClientUtil::FALSE;
        }
    } else {

        # If the attribute name isn't specified, then return the list
        unless (defined $attribute_name and length $attribute_name) {
            return $output;
        }

        # Parse cli output.
        my @output_array = split(/\n/, $output);
        my $lun_hash = {};
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace

            if ($line =~ /^$attribute_name/) {
                $line =~ s/$attribute_name\s+=\s+//;    # strip off the prefix
                $line =~ s/"//g;    # strip off double quotes#  "
                $lun_hash->{value} = $line;
            }
        } ## end foreach my $line (@output_array)

        if (defined $lun_hash->{value}) {
            $attribute_value = $lun_hash->{value};
        } else {
            $self->_set_err(  "Unable to obtain lun attribute"
                            . " value for $attribute_name via cli");
            return $ClientUtil::FALSE;
        }
    } ## end else [ if ($interface eq "zapi")

    return $attribute_value;

}    # end get_attribute

################################################################################
# Subroutine Name: get_space_alloc()
# Objective:       Get the space reservation information on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_space_alloc {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = "cli";
    my $filer     = $self->get_filername();
    my $space_alloc_state;
    my $option_name = "space_alloc";

    $self->_clear_err();

    $self->sanlog(-msg => "Getting the space allocation on Lun $path");

    my %cmd_args = ("-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-name"          => $option_name,);

    $cmd_args{"-cmd"} = "lun-set";

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure)
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Since we checked the result, lets see if our check was OK
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # interface is cli

    $space_alloc_state = $output;
    $space_alloc_state =~ s/.*:\s+//;    # getting reservation state
    $space_alloc_state =~ s/^\s+//;      # remove leading whitespace
    $space_alloc_state =~ s/\s+$//;      # remove trailing whitespace

    return $space_alloc_state;

}    # end get_space_alloc

################################################################################
# Subroutine Name: set_space_alloc()
# Objective:       Set the space reservation on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_space_alloc {
    my ($self, %args) = @_;

    my $value;
    my $cmd_obj     = $self->{parent}->{parent}->cmd_obj();
    my $path        = $self->get_path();
    my $interface   = "cli";
    my $filer       = $self->get_filername();
    my $option_name = "space_alloc";

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?value/) {
            if (not length($value = $args{$_})) {
                $self->_set_err("Invalid value for argument '-value'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $value and length $value) {
        $self->_set_err("value for space_alloc setting is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(
            -msg => "Setting the space allocation" . " to $value on Lun $path");

    my %cmd_args = ("-cmd"           => "lun-set",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-name"          => $option_name,
                    "-value"         => $value,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}    # end set_space_alloc

################################################################################
# Subroutine Name: set_serial()
# Objective:       Set the serial number on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_serial {
    my ($self, %args) = @_;

    my $serial;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?serial/) {
            if (not length($serial = $args{$_})) {
                $self->_set_err("Invalid value for argument '-serial'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $serial and length $serial) {
        $self->_set_err("serial is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(
              -msg => "Setting the serial number" . " to $serial on Lun $path");

    my %cmd_args = ("-cmd"           => "lun-set-serial-number",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-serial-number" => $serial,
                    "-cmd_privilege" => "diag",);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}    # end set_serial

################################################################################
# Subroutine Name: get_share()
# Objective:       Get the share information on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_share {
    my ($self) = @_;

    $self->_clear_err();

    my $info_hash_ref = $self->list_info(-path => $self->{path});
    if (defined $info_hash_ref->{"share-state"}) {
        return $info_hash_ref->{"share-state"};
    } else {
        $self->_set_err("Unable to obtain lun share information");
        return $ClientUtil::FALSE;
    }

}    # end get_share

################################################################################
# Subroutine Name: set_share()
# Objective:       Set the serial number on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_share {
    my ($self, %args) = @_;

    my $share;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?share/) {
            if (not length($share = $args{$_})) {
                $self->_set_err("Invalid value for argument '-share'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $share and length $share) {
        $self->_set_err("share is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Setting the share status to $share on Lun $path");

    my %cmd_args = ("-cmd"           => "lun-set-share",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-share-type"    => $share,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err("Unable to set lun share" . " as $share on $path",
                            $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}    # end set_share

################################################################################
# Subroutine Name: get_reserve()
# Objective:       Get the space reservation information on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_reservation {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $reservation_state;

    $self->_clear_err();

    $self->sanlog(-msg => "Getting the space reservation on Lun $path");

    my %cmd_args = ("-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,);

    $cmd_args{"-cmd"} = "lun-get-space-reservation-info"
      if ($interface eq "zapi");
    $cmd_args{"-cmd"} = "lun-set-space-reservation-info"
      if ($interface eq "cli");
    $cmd_args{"-state"} = "" if ($interface eq "cli");

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure)
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Since we checked the result, lets see if our check was OK
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err(
                           "Unable to get lun space reservation" . " for $path",
                           $output_hash->{reason});
            return $ClientUtil::FALSE;
        }

        if (defined $output_hash->{'is-enabled'}) {
            $reservation_state = $output_hash->{'is-enabled'};
        } else {
            $self->_set_err("Unable to obtain lun reservation via zapi");
            return $ClientUtil::FALSE;
        }
    } else {

        $reservation_state = $output;
        $reservation_state =~ s/.*:\s+//;    # getting reservation state
        $reservation_state =~ s/^\s+//;      # remove leading whitespace
        $reservation_state =~ s/\s+$//;      # remove trailing whitespace
    }

    return $reservation_state;

}    # end get_reservation

################################################################################
# Subroutine Name: set_reservation()
# Objective:       Set the space reservation on a lun
# Details:         See POD documentation at the end of this file
################################################################################
sub set_reservation {
    my ($self, %args) = @_;

    my $reservation;
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?reservation/) {
            if (not length($reservation = $args{$_})) {
                $self->_set_err("Invalid value for argument '-reservation'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    unless (defined $reservation and length $reservation) {
        $self->_set_err("reservation is a required argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Setting the space reservation"
                  . " to $reservation on Lun $path");

    my %cmd_args = ("-cmd"           => "lun-set-space-reservation-info",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 1,
                    "-path"          => $path,);

    $cmd_args{"-enable"} = 1
      if (($reservation eq "enable") and ($interface eq "zapi"));
    $cmd_args{"-enable"} = 0
      if (($reservation eq "disable") and ($interface eq "zapi"));
    $cmd_args{"-state"} = $reservation
      if ($reservation and ($interface eq "cli"));

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err(
                          "Unable to set lun reservation" . " to $reservation ",
                          $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } ## end if ($interface eq "zapi")

    return $ClientUtil::TRUE;

}    # end set_reservation

################################################################################
# Subroutine Name: list_info()
# Objective:       Get info from filer and put it in a hash
# Details:         See POD documentation at the end of this file
################################################################################
sub list_info {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $info_hash;

    #
    # Check to see if the user sent a path override. This is useful
    # if you don't yet have a lun obj for a new lun or clone
    #
    my $path = (defined $args{-path}) ? $args{-path} : $self->get_path();

    #
    # Check to see if the user sent a privilege override. This will allow
    # user to get information at the priv level they want
    #
    my $cmd_privilege = (defined $args{-priv}) ? $args{-priv} : "";

    $self->_clear_err();

    $self->sanlog(-msg => "Retrieve info for lun $path");

    #
    # Create the list of arguments for the CmdsDriver call.
    # Note that -verbose is really for the cli and should just
    # be ignored for the zapi
    #
    my %cmd_args = ("-cmd"           => "lun-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 0,
                    "-path"          => $path,
                    "-verbose"       => "",
                    "-cmd_privilege" => $cmd_privilege,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # If we get here, we have results back from the filer.  If it's a zapi
    # we want to set the info_hash to the lun-info hash in the luns hash.
    # If it's cli, we want to parse the data line by line and fill in the
    # fields of the info_hash matching the field names from the zapi
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        unless (defined $output_hash->{"luns"}->{"lun-info"}) {
            $self->_set_err(  "Unable to parse output: {luns}->{lun-info} "
                            . "does not exist");
            return $ClientUtil::FALSE;
        }
        $info_hash = $output_hash->{"luns"}->{"lun-info"};
    } else {

        #
        # CLI case
        # We are expecting output that looks like this:
        # /vol/blocks_fvt/cloned_lun 8m (8388608) (r/w, offline, mapped, shared)
        #       Serial#: hofBSJMIuVXV
        #       Backed by: /vol/blocks_fvt/.snapshot/clonesnap/mappedonline_lun
        #       Share: none
        #       Space Reservation: enabled
        #       Multiprotocol Type: solaris
        #
        # For BR.1 in test priv, the output is like this:
        #        /vol/vstorage_vol/lun2        13g (13958643712)   (r/w, online, mapped)
        #        Serial#: HnTBh4T0THeR
        #        Share: none
        #        Space Reservation: enabled
        #        Multiprotocol Type: vmware
        #        Maps: vmware_fcp=1
        #        Occupied Size:   12.1g (12947107840)
        #        Creation Time: Tue Nov 10 17:23:47 GMT 2009
        #        Alignment: indeterminate

        my @output_array = split(/\n/, $output);
        my $lun_hash = {};
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace
            if ($line =~ /^\/vol/) {

                #
                # Regex below parses the first line shown above in the example.
                # Using that example:
                # $new_path = /vol/blocks_fvt/cloned_lun
                # $size = 8m, $access = r/w, $state = offline, mapped, shared
                #
                # my ($new_path,$size,$unit, $access,$state)=
                my ($new_path, $size, $unit, $bytes, $access, $state) = $line =~
                  /^\s*(.*)\s+(\d+.?\d{0,})(\w)\s+\((.*)\)\s+\((\w+\/\w+),\s+(\w+)/;
                $self->sanlog(-msg => "Lun::list_info(): $line\n"
                                . "new path is $new_path,\n"
                                . "size     is $size $unit,\n"
                                . "bytes    is $bytes,\n"
                                . "access   is $access,\n"
                                . "state    is $state\n",
                              -msg_level => 3);

                $line =~ /^\s*(.*)\s+(\d+.?\d{0,})(\w)\s+\((.*)\)\s+(.*)/;
                $lun_hash->{"full-state"} = $5;

                $new_path =~ s/\s+$//;    # remove trailing whitespace
                                          #
                                          # Check the $state for mapped, online
                                          # 1 for mapped, otherwise 0.
                                          #
                my $mapped   = ($state  =~ /mapped/) ? "true"  : "false";
                my $online   = ($state  =~ /online/) ? "true"  : "false";
                my $readonly = ($access =~ /w/)      ? "false" : "true";

                #
                # Load the lun hash with what we know
                #
                $lun_hash->{path}        = $new_path;
                $lun_hash->{size}        = $bytes;
                $lun_hash->{online}      = $online;
                $lun_hash->{mapped}      = $mapped;
                $lun_hash->{"read-only"} = $readonly;

            } elsif ($line =~ /^Serial/) {
                $line =~ s/Serial#:\s+//;    # strip off the prefix
                $lun_hash->{"serial-number"} = $line;
            } elsif ($line =~ /^Share:/) {
                $line =~ s/Share:\s+//;
                $lun_hash->{"share-state"} = $line;
            } elsif ($line =~ /^Space/) {

                #
                # If the line contains "enabled" then use true
                #
                $lun_hash->{"is-space-reservation-enabled"} =
                  ($line =~ /enabled/) ? "true" : "false";
                $line =~ s/Space Reservation:\s+//;
                $lun_hash->{space} = $line;
            } elsif ($line =~ /^Multiprotocol/) {
                $line =~ s/Multiprotocol\s+Type:\s+//;
                $lun_hash->{"multiprotocol-type"} = $line;
            } elsif ($line =~ /^Backed/) {
                $line =~ s/Backed\s+by:\s+//;
                $lun_hash->{"backing-snapshot"} = $line;
            } elsif ($line =~ /^Maps/) {
                $line =~ s/Maps:\s+//;
                $lun_hash->{maps} = $line;
            } elsif ($line =~ /^Comment:/) {
                $line =~ s/Comment:\s+//;
                $lun_hash->{comment} = $line;
            } elsif ($line =~ /^Alignment:/) {
                $line =~ s/Alignment:\s+//;
                $lun_hash->{alignment} = $line;
            }

        }    # end foreach line
        $info_hash = $lun_hash;
    }    # end if CLI
    return $info_hash;
}    # end list_info()
################################################################################
# lun_id_swap
################################################################################
sub lun_id_swap() {

    my ($self, %args) = @_;
    my ($slun_path, $dlun_path, $cmd_interface, $cmd_obj, $filer);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?slun_path/) {
            if (not length($slun_path = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::lun_id_swap() ",
                      "Invalid value for argument source lun path'-slun_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?dlun_path/) {
            if (not length($dlun_path = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "lun_id_swap()",
                                "Invalid value for argument destination lun path'-dlun_path'"
                                );
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: " . ref($self) . "lun_id_swap()",
                            "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # must have both qtree and id params
    unless (defined($slun_path) and defined($dlun_path)) {
        $self->_set_err("Error: " . ref($self) . "::lun_id_swap() ",
                        "Both -slun_path and -dlun_path must be specified");
        return $ClientUtil::FALSE;
    }

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command is only supported for zapi
    $cmd_interface = "zapi";
    $filer         = $self->get_filername();

    my %cmd_args = (-cmd                => "lun-id-swap",
                    -cmd_filer          => $filer,
                    -check_output       => 1,
                    -check_result       => 1,
                    -cmd_interface      => $cmd_interface,
                    "-source-path"      => "$slun_path",
                    "-destination-path" => "$dlun_path",);

    $self->sanlog(-msg => "Swap LUN ids between $slun_path and $dlun_path");

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

} ## end sub lun_id_swap()

################################################################################
#snapvault_restore_qtree
################################################################################
sub lun_ndsvr_restore_qtree() {
    my ($self, %args) = @_;
    my ($qtree_path, $transaction_id, $interface, $cmd_obj, $filer);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?qtree_path/) {
            if (not length($qtree_path = $args{$_})) {
                $self->_set_err(
                        "Error: " . ref($self) . "::lun_ndsvr_restore_qtree() ",
                        "Invalid value for argument '-qtree_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?transaction_id/) {
            if (not length($transaction_id = $args{$_}) or $args{$_} !~ /\d+/) {
                $self->_set_err(
                           "Error: " . ref($self) . "lun_ndsvr_restore_qtree()",
                           "Invalid value for argument '-transaction_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(
                           "Error: " . ref($self) . "lun_ndsvr_restore_qtree()",
                           "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # must have both qtree and id params
    unless (defined($qtree_path) and defined($transaction_id)) {
        $self->_set_err("Error: " . ref($self) . "::lun_ndsvr_restore_qtree() ",
                        "Both -qtree_path and -tranaction_id must be specified"
                        );
        return $ClientUtil::FALSE;
    }

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command only supported for zapi
    # set interface to zapi
    $interface = "zapi";
    $filer     = $self->get_filername();

    # build cmd args
    #
    my %cmd_args = ("-cmd"       => "lun-restore-qtree-nondisruptive-restore",
                    "-cmd_filer" => $filer,
                    "-check_result"   => 1,
                    "-cmd_interface"  => $interface,
                    "-check_output"   => 1,
                    "-qtree-path"     => $qtree_path,
                    "-transaction-id" => $transaction_id,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
} ## end sub lun_ndsvr_restore_qtree()
################################################################################
# Subroutine Name: list_verbose()
# Objective:       Get lun info for a particular LUN
# Details:         See POD documentation at the end of this file
################################################################################
sub list_info_i {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $info_hash;
    my @output;

    #
    # Check to see if the user sent a path override. This is useful
    # if you don't yet have a lun obj for a new lun or clone
    #
    my $path = (defined $args{-path}) ? $args{-path} : $self->get_path();

    #
    # TODO: The format_output will be used in the future to return a hashed format
    #       of the output
    #
    my $format_output =
      (defined $args{-formatted_output}) ? $args{-formatted_output} : 1;

    $self->_clear_err();

    $self->sanlog(-msg => "Retrieve info for lun $path");

    #
    # Create the list of arguments for the CmdsDriver call.
    # Note that -verbose is really for the cli and should just
    # be ignored for the zapi
    #
    my %cmd_args = ("-cmd"           => "lun-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => "cli",
                    "-check_output"  => 0,
                    "-path"          => $path,
                    "-info"          => "",
                    "-cmd_privilege" => "diag",);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # Return the unformatted output string to the caller
    if ($format_output == 0) {
        return $resultsHash_ref->{output};
    }
    if ($format_output == "array") {
        @output = split("\n", $resultsHash_ref->{output});
        return \@output;
    }

    # TODO: Need to add another else to return hash output

    return $resultsHash_ref->{output};

}    # end list_verbose()

################################################################################
#snapvault_preserve_qtree
################################################################################
sub lun_ndsvr_preserve_qtree() {
    my ($self, %args) = @_;
    my ($qtree_path, $transaction_id, $interface, $cmd_obj, $filer);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?qtree_path/) {
            if (not length($qtree_path = $args{$_})) {
                $self->_set_err(
                       "Error: " . ref($self) . "::lun_ndsvr_preserve_qtree() ",
                       "Invalid value for argument '-qtree_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?transaction_id/) {
            if (not length($transaction_id = $args{$_}) or $args{$_} !~ /\d+/) {
                $self->_set_err(
                          "Error: " . ref($self) . "lun_ndsvr_preserve_qtree()",
                          "Invalid value for argument '-transaction_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(
                          "Error: " . ref($self) . "lun_ndsvr_preserve_qtree()",
                          "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # must have both qtree and id params
    unless (defined($qtree_path) and defined($transaction_id)) {
        $self->_set_err("Error: " . ref($self) . "lun_ndsvr_preserve_qtree()",
                       "Both -qtree_path and -tranaction_id must be specified");
        return $ClientUtil::FALSE;
    }

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command only supported for zapi
    # set interface to zapi
    $interface = "zapi";
    $filer     = $self->get_filername();

    # build cmd args
    #
    my %cmd_args = ("-cmd"       => "lun-preserve-qtree-nondisruptive-restore",
                    "-cmd_filer" => $filer,
                    "-check_result"   => 1,
                    "-cmd_interface"  => $interface,
                    "-check_output"   => 1,
                    "-qtree-path"     => $qtree_path,
                    "-transaction-id" => $transaction_id,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub lun_ndsvr_preserve_qtree()
################################################################################
# subroutine name: lun_cleanup_qtree_nondisruptive_restore
# objective:       destroy all the remaining luns stored in the
#                  temporary staging area after a non-disruptive restore
################################################################################
sub lun_ndsvr_cleanup_qtree() {

    my ($self, %args) = @_;
    my ($qtree_path, $transaction_id, $interface, $cmd_obj, $filer);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?qtree_path/) {
            if (not length($qtree_path = $args{$_})) {
                $self->_set_err(
                        "Error: " . ref($self) . "::lun_ndsvr_cleanup_qtree() ",
                        "Invalid value for argument '-qtree_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?transaction_id/) {
            if (not length($transaction_id = $args{$_}) or $args{$_} !~ /\d+/) {
                $self->_set_err(
                        "Error: " . ref($self) . "::lun_ndsvr_cleanup_qtree() ",
                        "Invalid value for argument '-transaction_id'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err(
                        "Error: " . ref($self) . "::lun_ndsvr_cleanup_qtree() ",
                        "Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # must have both qtree and id params
    unless (defined($qtree_path) and defined($transaction_id)) {
        $self->_set_err("Error: " . ref($self) . "::lun_ndsvr_cleanup_qtree() ",
                        "Both -qtree_path and -tranaction_id must be specified"
                        );
        return $ClientUtil::FALSE;
    }

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command only supported for zapi
    # set interface to zapi
    $interface = "zapi";
    $filer     = $self->get_filername();

    # build cmd args
    #
    my %cmd_args = ("-cmd"       => "lun-cleanup-qtree-nondisruptive-restore",
                    "-cmd_filer" => $filer,
                    "-check_result"   => 1,
                    "-cmd_interface"  => $interface,
                    "-check_output"   => 1,
                    "-qtree-path"     => $qtree_path,
                    "-transaction-id" => $transaction_id,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
} ## end sub lun_ndsvr_cleanup_qtree()

###############################################################################
# Method:    show_staged_luns
# Objective: display staged lun(s) in the staging area
# Details:   See POD documentation at the end of this file.
###############################################################################
sub show_staged_luns {
    my $self = shift;
    my (@luns, $cmd_obj, $staged_lun);

    $self->_clear_err();

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command only supported for cli
    # Create the list of arguments for the CmdsDriver call.
    my %cmd_args = ("-cmd"           => "lun-list-info",
                    "-cmd_filer"     => $self->{filer},
                    "-filer_version" => "7.3",
                    "-check_result"  => 1,
                    "-cmd_interface" => "cli",
                    "-check_output"  => 0,
                    "-staging"       => "",);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure)
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()), -msg_level => 1);
        return $ClientUtil::FALSE;
    }

    # Get the output
    my $output  = $resultsHash_ref->{output};
    my $dumpOut = Dumper($output);
    $self->sanlog(-msg => "output is:\n$output", -msg_level => 1);

    # Since we checked the result, lets see if our check was OK
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # parse for cli output
    my @info = split(/\n/, $output);

    foreach my $line (@info) {

        #split based on space
        my @temp_arr = split(/ /, $line);

        #the first element of array will be the lun name
        $staged_lun = $temp_arr[0];
        $self->sanlog(-msg => "staged lun path=> $staged_lun", -msg_level => 1);

        #in case there is more than 1 lun, store them in an array
        push(@luns, $staged_lun);
    } ## end foreach my $line (@info)

    # return undef if no staged lun exists
    return (0 eq scalar(@luns)) ? undef : \@luns;

}    # end show_staged_luns
###############################################################################
# Method:    is_lun_staged
# Objective: show whether lun is staged or not
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_lun_staged {
    my $self = shift;
    my $flag = 0;

    #function show_staged_luns() returns refrence of array of staged luns
    my $staged_array_ref = $self->show_staged_luns();

    #get the lun name of the lun which is calling this method
    my $lun_name = $self->get_name();
    my @lun_array = split(/\//, $lun_name);
    $lun_name = $lun_array[-1];
    $self->sanlog(-msg => "$lun_name is the lun name", -msg_level => 3);

    #make sure that the lun name returned by the method show_stages_luns() is
    #the same as that of the calling lun name
    foreach my $staged_lun (@{$staged_array_ref}) {
        if ($staged_lun =~ /$lun_name/) {
            $self->sanlog(-msg       => "staged lun path=> $staged_lun",
                          -msg_level => 1);
            $flag = 1;
            return $ClientUtil::TRUE;
        }
    }

    #if the names does not match, there are no luns staged,i.e., $flag = 0
    if (!$flag) {
        $self->sanlog(-msg => "no staged luns found", -msg_level => 1);
        return $ClientUtil::FALSE;
    }
}    # end is_lun_staged

################################################################################
#lun_set_device_id
################################################################################
sub lun_set_device_id {
    my ($self, %args) = @_;
    my ($id, $path, $rv_ref, $cmd_obj, $interface, $filer);
    $interface = "zapi";
    $self->_clear_err();
    foreach (keys %args) {
        if (/-?id/) {
            if (not length($id = $args{$_}) or $args{$_} !~ /(none|\d+)/) {
                $self->_set_err(
                              "Error: " . ref($self) . "::lun_set_device_id() ",
                              "Invalid value for argument '-id'");
                return $ClientUtil::FAIL;
            }
            next;
        } elsif (/-?path/) {
            if (not length($path = $args{$_}) or $args{$_} !~ /^\/vol\//) {
                $self->_set_err(
                              "Error: " . ref($self) . "::lun_set_device_id() ",
                              "Invalid value for argument '-path'");
                return $ClientUtil::FAIL;
            }
            next;
        } elsif (/-?interface/) {
            if (not length($interface = $args{$_})) {
                $self->_set_err(
                              "Error: " . ref($self) . "::lun_set_device_id() ",
                              "Invalid value for argument '-path'");
                return $ClientUtil::FAIL;
            }
            next;
        } else {
            $self->_set_err("Error: " . ref($self) . "::lun_set_device_id() ",
                            "Invalid parameter specified \'$_\'");
            return $ClientUtil::FAIL;
        }
    } ## end foreach (keys %args)

    # must have id param
    if (not defined($id) and not defined($path)) {
        $self->_set_err("Error: " . ref($self) . "::lun_set_device_id() ",
                        "-id and -path must be specified");
        return $ClientUtil::FAIL;
    }

    # get cmdsDriver obj
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    # this command only supported for zapi
    # set interface to zapi

    $filer = $self->get_filername();

    my %cmd_args = ("-cmd"           => "lun-set-device-id",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-device-id"     => $id,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}    # end lun_set_device_id

################################################################################
# Subroutine Name: map_list_info()
# Objective:       List all of the igroups to which a lun is mapped
# Details:         See POD documentation at the end of this file
################################################################################
sub map_list_info {
    my ($self) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $path      = $self->get_path();
    my $interface = $self->{parent}->{parent}->interface();
    my $filer     = $self->get_filername();
    my $map_list_info;
    my @igroup_info;

    $self->_clear_err();

    # Build command arguments.
    my %cmd_args = ("-cmd"           => "lun-map-list-info",
                    "-cmd_filer"     => $filer,
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-path"          => $path,);

    # Execute command.
    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure).
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Log results.
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.

    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        my $status      = $output_hash->{status};
        if ($status eq "failed") {
            $self->_set_err(  "unable to get map list info"
                            . "for lun $path "
                            . $output_hash->{reason});
            return $ClientUtil::FALSE;
        }

        if (defined $output_hash->{'initiator-groups'}
            ->{'initiator-group-info'}) {
            # since cli is a subset of zapi taking only the subset.
	        # in order to return same output in both cli and zapi.
			# output looks like this and taking only similar things of cli
			# $VAR1 = {
                # 'status' => 'passed',
                # 'initiator-groups' => {
                            # 'initiator-group-info' => {
                                    # 'initiator-group-alua-enabled' => 'false',
                                    # 'initiators' => {
                                            # 'initiator-info' => {
                                                        # 'initiator-name' => 'iqn.1986-03.com.sun:01:0003ba12daa8.49aeddc4'
                                            # }
                                    # },
                                    # 'initiator-group-vsa-enabled' => 'false',
                                    # 'initiator-group-name' => 'ssan-280r-26_iscsi',
                                    # 'initiator-group-uuid' => '43e8e679-47d3-11e1-ba79-123478563412',
                                    # 'initiator-group-os-type' => 'solaris',
                                    # 'initiator-group-report-scsi-name-enabled' => 'true',
                                    # 'initiator-group-throttle-borrow' => 'false',
                                    # 'initiator-group-throttle-reserve' => '0',
                                    # 'lun-id' => '1',
                                    # 'initiator-group-type' => 'iscsi'
                            # }
                # }
        # };
            $map_list_info->{'initiator-group-type'} = $output_hash->
		        {'initiator-groups'}->{'initiator-group-info'}->
					{'initiator-group-type'};
            $map_list_info->{'initiator-group-name'} = $output_hash->
		        {'initiator-groups'}->{'initiator-group-info'}->
			    {'initiator-group-name'};
	        $map_list_info->{'lun-id'} = $output_hash->{'initiator-groups'}->
		        {'initiator-group-info'}->{'lun-id'};
	        push @igroup_info, $map_list_info;
        } else {
            $self->sanlog(-msg => "no lun map info returned from zapi");
            return ();
        }

        unless (@igroup_info) {
            $self->_set_err("unable to parse zapi output");
            return $ClientUtil::FALSE;
        }

    } else {

        if ($output =~ /is not mapped/) {
            $self->sanlog(-msg => "$output");
            @igroup_info = split(/\n/, $output);
            return @igroup_info;
        }

        my @output_array = split(/\n/, $output);

        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace
            if (($line !~ /Mapped to/) && ($line =~ /^\/vol/)) {
                my ($path, $name, $id, $type) = split(/\s+/, $line, 4);
                push(@igroup_info,
                     {'initiator-group-name' => $name,
                      'lun-id'               => $id,
                      'initiator-group-type' => $type});
                next;
            }
            if ($line !~ /Mapped to/) {
                my ($name, $id, $type) = split(/\s+/, $line, 3);
                push(@igroup_info,
                     {'initiator-group-name' => $name,
                      'lun-id'               => $id,
                      'initiator-group-type' => $type});
                next;
            }
        } ## end foreach my $line (@output_array)
    } ## end else [ if ($interface eq "zapi")

    return @igroup_info;
} ## end sub map_list_info
####################################################################################
# Subroutine Name: df
# Objective:       display the contents of lun using df command
# Details:         A cli only command; See POD documentation at the end of this file
####################################################################################
sub df {
    my ($self, %args) = @_;
    my $verbose   = 0;
    my $path      = $self->{path};
    my $name      = $self->get_name();
    my $filer     = $self->get_filername();
    my $interface = "cli";
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my ($block_no, $block_offset);

    $self->_clear_err();

    #
    # Parse arguments
    #
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Invalid value for argument '-verbose'");
                return $ClientUtil::FALSE;
            }
            if (defined $verbose and $verbose =~ /1|t|on|true|y|yes/i) {
                $verbose = 1;
            }
            next;
        } elsif (/-?block_num/) {
            if (not length($block_no = $args{$_})) {
                $self->_set_err("Invalid value for argument '-block_num'");
                return $ClientUtil::FALSE;
            }
            if ($block_no =~ /\D+/) {
                $self->_set_err("Invalid value for argument '-value'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?block_offset/) {
            if (not length($block_offset = $args{$_})) {
                $self->_set_err("Invalid value for argument '-block_offset'");
                return $ClientUtil::FALSE;
            }
            if ($block_offset =~ /\D+/) {
                $self->_set_err("Invalid value for argument '-value'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)
    if (defined($block_no) xor defined($block_offset)) {
        $self->_set_err(
            "block_num and block_offset are mutually inclusive options.\n
                        Provide values for both or none");
        return $ClientUtil::FALSE;
    }

    #
    # Get the full lun path
    #
    unless (defined $path) {
        $self->_set_err("Lun path is missing from object");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-df",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "cli",
                    "-path"          => $path,
                    "-cmd_privilege" => "diag",);

    if ($verbose and defined $block_no) {
        $self->_set_err(
            "Provide either -verbose or (block_num and block_offset options) for hexdump.\n
                        Cannot provide both together");
        return $ClientUtil::FALSE;
    }
    $cmd_args{"-verbose"} = "" if ($verbose);
    $cmd_args{"-hex_dump"} = $block_no . " " . $block_offset
      if length($block_no);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }
    return $output;
} ## end sub df

################################################################################
# Subroutine Name: alignment_list_info()
# Objective:       Get info from filer and put it in a hash
# Details:         See POD documentation at the end of this file
################################################################################
sub alignment_list_info {
    my ($self, %args) = @_;

    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();
    my $info_hash;

    #
    # Check to see if the user sent a path override. This is useful
    # if you don't yet have a lun obj for a new lun or clone
    #
    my $path = (defined $args{-path}) ? $args{-path} : $self->get_path();

    #
    # Check to see if the user sent a privilege override. This will allow
    # user to get information at the priv level they want. default will be diag
    # for this command to work
    #
    my $cmd_privilege = (defined $args{-priv}) ? $args{-priv} : "diag";

    $self->_clear_err();

    $self->sanlog(-msg => "Retrieve lun alignment info for lun $path,"
                      . "interface is $interface.");

    #
    # Create the list of arguments for the CmdsDriver call.
    # Note that -verbose is really for the cli and should just
    # be ignored for the zapi
    #
    my %cmd_args = ("-cmd"           => "lun-alignment-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-check_output"  => 0,
                    "-lun-path"      => $path,
                     "-cmd_privilege" => $cmd_privilege,);

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    #
    # Check for CmdsDriver errors (not command failure)
    #
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    my $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    #
    # If we get here, we have results back from the filer.  If it's a zapi
    # we want to set the info_hash to the lun-alignment-info hash in the luns hash.
    # If it's cli, we want to parse the data line by line and fill in the
    # fields of the info_hash matching the field names from the zapi
    #
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        unless (defined $output_hash->{"luns"}->{"lun-alignment-info"}) {
            $self->_set_err(
                     "Unable to parse output: {luns}->{lun-alignment-info} "
                       . "does not exist");
            return $ClientUtil::FALSE;
        }
        $info_hash = $output_hash->{"luns"}->{"lun-alignment-info"};
    } else {

        #
        # CLI case
        # We are expecting output that looks like this:
        # /vol/vstorage_vol/sol_lun_1
        #        Multiprotocol type: solaris
        #        Alignment: indeterminate
        #        Write alignment histogram percentage: 0, 0, 0, 0, 0, 0, 0, 0
        #        Read alignment histogram percentage: 0, 0, 0, 0, 0, 0, 0, 0
        #        Partial writes percentage: 0
        #        Partial reads percentage: 0

        my @output_array = split(/\n/, $output);
        my $lun_hash = {};
        foreach my $line (@output_array) {
            $line =~ s/^\s+//;    # remove leading whitespace
            $line =~ s/\s+$//;    # remove trailing whitespace
            if ($line =~ /^\/vol/) {
                my $new_path = $line;
                $lun_hash->{path} = $new_path;
            } elsif ($line =~ /^Multiprotocol/) {
                $line =~ s/Multiprotocol\s+type:\s+//;
                $lun_hash->{"multiprotocol-type"} = $line;
            } elsif ($line =~ /^Alignment:/) {
                $line =~ s/Alignment:\s+//;
                $lun_hash->{alignment} = $line;
            } elsif ($line =~ /^Write alignment histogram percentage:/) {
                $line =~ s/^Write alignment histogram percentage:\s+//
                  ; # strip off the prefix
                $lun_hash->{"write-alignment-histogram"} = $line;
            } elsif ($line =~ /^Read alignment histogram percentage:/) {
                $line =~ s/Read alignment histogram percentage:\s+//;
                $lun_hash->{"read-alignment-histogram"} = $line;
            } elsif ($line =~ /^Partial writes percentage:/) {
                $line =~ s/Partial writes percentage:\s+//;
                $lun_hash->{"write-partial-blocks-percentage"} = $line;
            } elsif ($line =~ /^Partial reads percentage:/) {
                $line =~ s/Partial reads percentage:\s+//;
                $lun_hash->{"read-partial-blocks-percentage"} = $line;
            }

        }    # end foreach line
        $info_hash = $lun_hash;
    }    # end if CLI
    return $info_hash;
}    # end alignment_list_info()

###############################################################################
# Method:    lun_get_select_attribute()
# Objective: Get Lun select information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lun_get_select_attribute {
    my ($self)    = @_;
    my $lun_name  = $self->get_name();
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->{interface};
    my $path      = $self->get_path();
    my $output_hash;
    my $ret_ref;
    my $output;

    # Build the command line and execute the command
    my %cmd_args = (-cmd           => "lun-get-select-attribute",
                    -cmd_filer     => $self->get_filername(),
                    -cmd_interface => $interface,
                    -check_result  => 1,
                    -path          => "$path");

    if (lc($interface) eq "cli") {
        $cmd_args{"-cmd_privilege"} = "test";
    }

    my $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # log results
    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # check for CmdsDriver failure
    if ($cmd_obj->error()) {
        $self->_set_err("Unable to get lun select info: " . $cmd_obj->errmsg());
        $self->sanlog(-msg => $self->errmsg());
    }

    # get the output
    $output = $resultsHash_ref->{output};

    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->{parent}->xml2hash($output);

        if ($self->{parent}->{parent}->get_debug_level() >= 10) {
            my $dumpOut = Dumper($output_hash);
            $self->sanlog(-msg       => "Output Hash is:\n$dumpOut",
                          -msg_level => 10);
        }

        $ret_ref = $output_hash;
    } ## end if ($interface eq "zapi")

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        if ($interface eq "zapi") {
            $self->_set_err("Result check failed:\n", $output_hash->{reason});
            return $ClientUtil::FALSE;
        } else {
            $self->_set_err("Result check failed: $output");
            return $ClientUtil::FALSE;
        }
    }

    if ($interface eq "cli") {
        my @output = split(/\n/, $output);

        # check for error
        if (defined $output[0] and length $output[0]) {
            if ($output[0] =~ /lun \select:/) {
                $self->_set_err("CLI failed: ", @output);
                return $ClientUtil::FALSE;
            }
        }

        for (my $ind = 0 ; $ind < scalar(@output) ; $ind++) {
            if ($output[$ind] =~ /Select Attribute:\s(\w+)\s*/) {
                $ret_ref->{'select-attribute'} = $1;
            } elsif ($output[$ind] =~ /Creation Time\s*:\s([\s+\W+\w+]*)\s*/) {
                $ret_ref->{'creation-time'} = $1;
            } elsif ($output[$ind] =~ /Previous Serial#:\s(\w+)\s*/) {
                $ret_ref->{'previous-serial'} = $1;
            }
        }
    } ## end if ($interface eq "cli")

    return $ret_ref;
} ## end sub lun_get_select_attribute
################################################################################
# Subroutine Name: get_holes_lbarange()
# Objective:       Get the lba holes range
# Details:         See POD documentation at the end of this file
################################################################################
sub get_holes_lbarange {
    my ($self, %args) = @_;
    my %cmd_args;
    my ($output_hash, $resultsHash_ref, $output);
    my ($holeinfo, $start_lba, $end_lba, $cmd_obj, $path, $filer);

    # Applicable only for zapi
    $self->_clear_err();

    foreach (keys %args) {
        if (/-?start_lba/) {
            if (not length($start_lba = $args{$_})) {
                $self->_set_err("Invalid value for argument '-start_lba'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?end_lba/) {
            if (not length($end_lba = $args{$_})) {
                $self->_set_err("Invalid value for argument '-end_lba'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if (!defined($start_lba)) {
        $self->_set_err("start_lba is required argument");
        return $ClientUtil::FALSE;
    }
    if (!defined($end_lba)) {
        $self->_set_err("end_lba is required argument");
        return $ClientUtil::FALSE;
    }
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $path    = $self->get_path();
    $filer   = $self->get_filername();

    # Construct the zapi command
    %cmd_args = ("-cmd"           => "lun-lba-hole-range-query",
                 "-end-lba"       => $end_lba,
                 "-start-lba"     => $start_lba,
                 "-lun-path"      => $path,
                 "-check_output"  => 1,
                 "-cmd_interface" => "zapi",
                 "-cmd_filer"     => $filer,);

    # Execute the command
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    $output          = $resultsHash_ref->{output};

    # Convert the xml to hash and extract the data
    $output_hash = $self->{parent}->{parent}->xml2hash($output);

    if ($output_hash->{'status'} eq "failed") {
        $self->_set_err(
               "Failed to get the hole lba range :" . $output_hash->{'reason'});
    } else {
        $holeinfo = $output_hash->{"holes"}->{"lba-hole-range-info"};
        if (ref($holeinfo) eq "HASH") {
            my %tmphash = %{$holeinfo};
            my @arr;
            push(@arr, \%tmphash);
            return \@arr;
        } else {
            return $holeinfo;
        }
    } ## end else [ if ($output_hash->{'status'...
} ## end sub get_holes_lbarange
################################################################################
# Subroutine Name: is_lba_hole()
# Objective:       For a given lba check if it is a hole or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub is_lba_hole {
    my ($self, %args) = @_;
    my %cmd_args;
    my ($output_hash, $resultsHash_ref, $output);
    my ($check_lba, $cmd_obj, $path, $filer);

    # Applicable only for zapi
    $self->_clear_err();

    foreach (keys %args) {
        if (/-?check_lba/) {
            if (not length($check_lba = $args{$_})) {
                $self->_set_err("Invalid value for argument '-check_lba'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    # input parameter validation
    if (!defined($check_lba)) {
        $self->_set_err("check_lba is required argument");
        return $ClientUtil::FALSE;
    }
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $path    = $self->get_path();
    $filer   = $self->get_filername();

    # Construct the zapi command.
    %cmd_args = ("-cmd"           => "lun-lba-is-hole",
                 "-cmd_interface" => "zapi",
                 "-cmd_filer"     => $filer,
                 "-lba"           => $check_lba,
                 "-lun-path"      => $path,
                 "-check_output"  => 1,);

    # Execute the command
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    $output          = $resultsHash_ref->{output};
    $output_hash     = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{'status'} eq "failed") {
        $self->_set_err("Failed to check the lba is hole or not"
                        . $output_hash->{'reason'});
        return $ClientUtil::FALSE;
    }
    return $output_hash;
} ## end sub is_lba_hole
################################################################################
# Subroutine Name: get_lun_occupied_size()
# Objective:       Get the LUN occupied space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_occupied_size() {
    my ($self) = shift;
    my %cmd_args;
    my ($retval, $resultsHash_ref, $output, $output_hash);
    my ($path,   $cmd_obj,         $filer,  $dumpOut);

    # Applicable only for zapi
    $self->_clear_err();

    $path    = $self->get_path();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $filer   = $self->get_filername();

    # Construct the zapi command
    %cmd_args = ("-cmd"           => "lun-get-occupied-size",
                 "-cmd_interface" => "zapi",
                 "-cmd_filer"     => $filer,
                 "-check_output"  => 1,
                 "-path"          => $path,);

    # Execute the command
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    # Check for CmdsDriver errors (not command failure)
    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # Get the output
    $output = $resultsHash_ref->{output};

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Convert the xml to hash and extract the data
    $output_hash = $self->{parent}->{parent}->xml2hash($output);

    # Get the output printed for debug purpose
    $dumpOut = Dumper($output_hash);
    $self->sanlog(-msg => "Output Hash is:\n$dumpOut", -msg_level => 9);

    if ($output_hash->{'status'} eq "failed") {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    } else {
        $retval = $output_hash->{'occupied-size'};
    }
    return $retval;
} ## end sub get_lun_occupied_size()
################################################################################
# Subroutine Name: get_prefix_sufix()
# Objective:       Get the prefix and sufix for each LUN type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_prefix_sufix() {
    my ($self) = @_;
    my $pre_suf_info;
    my $luntype;

    # get the LUN type
    $luntype = $self->get_type();

    #depending upon the luntype determine the prefix and sufix of the lun
    if ($luntype eq "windows") {
        $pre_suf_info->{prefix} = 63;
        $pre_suf_info->{sufix}  = 0;
    } elsif ($luntype eq "solaris") {
        $pre_suf_info->{prefix} = 2048;
        $pre_suf_info->{sufix}  = 4096;
    } elsif ($luntype eq "solaris_efi") {
        $pre_suf_info->{prefix} = 34;
        $pre_suf_info->{sufix}  = 0;
    } elsif ($luntype eq "windows_gpt") {
        $pre_suf_info->{prefix} = 34;
        $pre_suf_info->{sufix}  = 0;
    } else {
        $pre_suf_info->{prefix} = 0;
        $pre_suf_info->{sufix}  = 0;
    }
    return $pre_suf_info;
} ## end sub get_prefix_sufix()
################################################################################
# Subroutine Name: get_type()
# Objective:       Get LUN type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self) = @_;
    my $type;
    $self->_clear_err();
    if (defined $self->{type}) {
        $type = $self->{type};
        return $type;
    } else {
        $self->_set_err("Lun object is missing type");
        return $ClientUtil::FALSE;
    }
} ## end sub get_type
################################################################################
# Subroutine Name: config_check_single_image_info()
# Objective:
# Details:         See POD documentation at the end of this file
################################################################################
sub config_check_single_image_info() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj);
    my $lun_config_single_image_info = {};

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-config-check-single-image-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err("unable to lun-config-check-single-image-info "
                        . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # extact the required information from output_hash
    $lun_config_single_image_info->{'conflicting-initiators'} =
      $output_hash->{'conflicting-initiators'};
    $lun_config_single_image_info->{'conflicting-maps'} =
      $output_hash->{'conflicting-maps'};
    $lun_config_single_image_info->{'invalid-nodename-settings'} =
      $output_hash->{'invalid-nodename-settings'};

    return $lun_config_single_image_info;
}    ### end sub config_check_single_image_info()
################################################################################
# Subroutine Name: get_inquiry_info()
# Objective:
# Details:         See POD documentation at the end of this file
################################################################################
sub get_inquiry_info() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj);
    my $lun_inquiry_info = {};

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"                  => "lun-get-inquiry-info",
                    "-cmd_filer"            => $self->get_filername(),
                    "-check_result"         => 1,
                    "-check_output"         => 1,
                    "-cmd_interface"        => "zapi",
                    "-initiator-group-name" => $self->get_igroup(),);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err("unable to lun inquiry info " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # extact the required information from output_hash
    $lun_inquiry_info->{'firmware-revision'} =
      $output_hash->{'firmware-revision'};
    $lun_inquiry_info->{'product-id'} = $output_hash->{'product-id'};
    $lun_inquiry_info->{'vendor-id'}  = $output_hash->{'vendor-id'};

    return $lun_inquiry_info;
}    ### end sub get_inquiry_info()
################################################################################
# Subroutine Name: lun_config()
# Objective:      Gives the lun config information
#Details:         See POD documentation at the end of this file
################################################################################
sub lun_config() {

    my($self, $filer, $cmd_obj, $resultsHash_ref, %lun_config_output) = @_;

    $filer     = $self->get_filername();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command, only 'cli'
    # is implemented, no 'zapi' support
    #
    my %cmd_args = ("-cmd"           => "lun-config",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "cli");

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output.
    my $output = $resultsHash_ref->{output};

    # It will create the hash from the output as show below
    # f3070-237-113*> lun config
    # copy_offload.state = "true"
    # vaw.state = "true"
    # write_same.state = "true"
    # local.serialnum = "30015696"
    # partner.single_image.key = "118065076"
    # ispfct.local.nodename = "2687598780090a50"

    # Ensure result_check is passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    my @output = split("\n", $output);

    my $flag = 0;

    foreach(@output) {
        if($_ =~ /(\S+) = (\S+)/) {
            $flag = 1;
            $lun_config_output{$1} = $2;
        }
    }

    if(!$flag) {
        return $ClientUtil::FAIL;
    }

    return \%lun_config_output;

}
################################################################################
################################################################################
# Subroutine Name: initiator_logged_in()
# Objective:       Checks whether the given initiator is logged in or not
# Details:         See POD documentation at the end of this file
################################################################################
sub initiator_logged_in() {
    my ($self, %args) = @_;
    my ($output, $output_hash, $resultsHash_ref, $initiator, $cmd_obj);

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?initiator/) {
            if (not length($initiator = $args{$_})) {
                $self->_set_err("Invalid value for argument '-initiator'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if (!defined($initiator)) {
        $self->_set_err("initiator is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-initiator-logged-in",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-initiator"     => $initiator,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # return 1 if initiator is logged in otherwise return 0
    if ($output_hash->{'is-logged-in' =~ /true/}) {
        return $ClientUtil::TRUE;
    } else {
        $ClientUtil::FALSE;
    }
}    ### end sub initiator_logged_in()
################################################################################
# Subroutine Name: get_target_device_id()
# Objective:       Returns the lun target device id
# Details:         See POD documentation at the end of this file
################################################################################
sub get_target_device_id() {
    my ($self, %args) = @_;
    my ($lun_serial_id, $protocol,    $cmd_obj,         $target_device_id);
    my ($output,        $output_hash, $resultsHash_ref, $initiator);

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?lun_serial_id/) {
            if (not length($lun_serial_id = $args{$_})) {
                $self->_set_err("Invalid value for argument -lun_serial_id");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if (!defined($protocol)) {
        $self->_set_err("protocol is required argument");
        return $ClientUtil::FALSE;
    }
    if (!defined($lun_serial_id)) {
        $self->_set_err("lun_serial_id is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-get-target-device-id",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-protocol"      => $protocol,
                    "lun-serial-id"  => $lun_serial_id,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
                "unable to get lun target device id " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # Obtain the required target device id.
    $target_device_id = $output_hash->{'target-device-id'};
    return $target_device_id;
}    ### end sub get_target_device_id()
################################################################################
# Subroutine Name: config_check_info()
# Objective:
# Details:         See POD documentation at the end of this file
################################################################################
sub config_check_info() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj);
    my $lun_config_check_info = {};

    # Verification hence implemented only in zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-config-check-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
              "unable to get lun config check info  " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # extact the required information from output_hash
    $lun_config_check_info->{'alua-setting-mismatch-info'} =
      $output_hash->{'alua-setting-mismatch-info'};
    $lun_config_check_info->{'alua-setting-mismatch-initiators'} =
      $output_hash->{'alua-setting-mismatch-initiators'};
    $lun_config_check_info->{'invalid-vsa-settings'} =
      $output_hash->{'invalid-vsa-settings'};
    $lun_config_check_info->{'mixed-ostype-initiators'} =
      $output_hash->{'mixed-ostype-initiators'};
    $lun_config_check_info->{'mixed-vsa-initiators'} =
      $output_hash->{'mixed-vsa-initiators'};
    $lun_config_check_info->{'fcp-down-hbas'} = $output_hash->{'fcp-down-hbas'};

    return $lun_config_check_info;
}    ### end sub config_check_info()
################################################################################
# Subroutine Name: unset_device_id()
# Objective:       Sets or unset the device id of lun
# Details:         See POD documentation at the end of this file
################################################################################
sub unset_device_id() {
    my ($self, %args) = @_;
    my ($cmd_obj, $interface, $output, $output_hash, $resultsHash_ref);
    my ($device_id, $path, $unset, $option);
    my %cmd_args;

    # supported only in zapi
    $interface = "zapi";
    $self->_clear_err();
    $interface = $self->{parent}->{parent}->interface();
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    $path      = $self->{path};

    #
    # Build the command line and execute the command
    #
    %cmd_args = ("-cmd"           => "lun-unset-device-id",
                 "-cmd_filer"     => $self->get_filername(),
                 "-check_result"  => 1,
                 "-check_output"  => 1,
                 "-cmd_interface" => $interface,
                 "-path"          => $path,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    if ($interface eq "zapi") {

        #
        # Since we checked the result, lets see if our check was OK
        #
        unless ($resultsHash_ref->{result_check}) {
            $self->_set_err("Failed result check: $output ");
            return $ClientUtil::FALSE;
        }    # end unless ...
        unless ($resultsHash_ref->{output_check}) {
            $self->_set_err("Output check failed: $output");
            return $ClientUtil::FALSE;
        }    # end unless ...
        $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err(
                          "unable to set lun device " . $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if($output_hash-> ...
    } ## end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    #### sub unset_device_id()
##############################################################################
# Subroutine Name: start()
# Objective:       Starts block-protocol access to the LUN listed
# Details:         See POD documentation at the end of this file
################################################################################
sub start() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj, $path, $interface);

    # Supported only in zapi
    $interface = "zapi";

    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $path    = $self->{path};

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-start",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => $interface,
                    "-cmd_privilege" => "diag",
                    "-path"          => $path,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    if ($interface eq "zapi") {

        #
        # Since we checked the result, lets see if our check was OK
        #
        unless ($resultsHash_ref->{result_check}) {
            $self->_set_err("Failed result check: $output ");
            return $ClientUtil::FALSE;
        }    # end unless ...
        unless ($resultsHash_ref->{output_check}) {
            $self->_set_err("Output check failed: $output");
            return $ClientUtil::FALSE;
        }    # end unless ...
        $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("unable to get lun config check info "
                            . $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if($output_hash-> ...
    } ## end if ($interface eq "zapi")
    return $ClientUtil::TRUE;
}    ### sub start()
##############################################################################
# Subroutine Name: vtoc_info()
# Objective:       create a new lun using an existing file
# Details:         See POD documentation at the end of this file
##############################################################################
sub vtoc_info() {
    my ($self) = @_;
    my ($output, $resultsHash_ref, $cmd_obj, $path);
    my @result;

    # Applicable only for cli
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-vtoc",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => "cli",
                    "-volume"        => $self->{parent}->get_name(),
                    "-cmd_privilege" => "test",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};
    @result = split(/\n/, $output);
    return @result;
}    ### sub vtoc_info()
##############################################################################
# Subroutine Name: set_svo_offset()
# Objective:       Sets or disable the svo_offset of the lun
# Details:         See POD documentation at the end of this file
##############################################################################
sub set_svo_offset() {
    my ($self, %args) = @_;
    my ($output, $resultsHash_ref, $offset, $cmd_obj);
    my ($disable, $path);

    # Applicable only for cli
    $self->_clear_err();
    $path    = $self->{path};
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?offset/) {
            if (not length($offset = $args{$_})) {
                $self->_set_err("Invalid value for argument '-offset'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?disable/) {
            if (not length($disable = $args{$_})) {
                $self->_set_err("Invalid value for argument '-disable'");
                return $ClientUtil::FALSE;
            }
            if (defined $disable and $disable =~ /yes|TRUE|1/i) {
                $disable = 1;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if ($offset && $disable) {
        $self->_set_err(
            "Both the values cant be give since both are
                                          mutually exclusive"
            );
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-set-unset-svo-offset",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => "cli",
                    "-path"          => $path,);
    if ($offset) {
        $cmd_args{"-offset"} = $offset;
    }

    if ($disable) {
        $cmd_args{"-disable"} = "";
    }
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};
    if ($offset) {
        my @result = split(/\n/, $output);
        ## end foreach my $line (@output)
        foreach my $line1 (@result) {
            if ($line1 =~ /SnapValidator/i) {

                # SnapValidator Offset:      4g (4294967296)
                if ($line1 =~ / \t+\w+\s+\w+\:\s+(\w+)/) {
                    if ($1 =~ $offset) {
                        $self->sanlog(-msg => " the given offset is set ");
                        return $ClientUtil::TRUE;
                    } else {
                        $self->_set_err(" the offset is not set ");
                        return $ClientUtil::FALSE;
                    }
                }
            } ## end if ($line1 =~ /SnapValidator/i)
        } ## end foreach my $line1 (@result)
    } ## end if ($offset)
    return $ClientUtil::TRUE;
}    ### end sub set_svo_offset
#############################################################################
# Subroutine Name: clone_status_list_info()
# Objective:       Get the cloning status of LUN(s).
# Details:         See POD documentation at the end of this file
#############################################################################
sub clone_status_list_info() {
    my ($self, %args) = @_;
    my ($cmd_obj, $output, $output_hash, $resultsHash_ref, $clone_lun_path);
    my $clone_status = {};

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?clone_lun_path/) {
            if (not length($clone_lun_path = $args{$_})) {
                $self->_set_err("Invalid value for argument'-clone_lun_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-clone-status-list-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",);
    $cmd_args{"-path"} = $clone_lun_path if ($clone_lun_path);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    $clone_status->{'clone-status'} = $output_hash->{'clone-status'};
    return $clone_status;
}    # end sub clone_status_list_info()
################################################################################
# Subroutine Name: create_from_snapshot()
# Objective:       create a new lun using an existing file
# Details:         See POD documentation at the end of this file
################################################################################
sub create_from_snapshot() {
    my ($self, %args) = @_;
    my ($type, $reserve, $snapshot, $path, $filer, $cmd_obj);
    my ($output, $output_hash, $dumpOut, $status, $resultsHash_ref);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?reserve/) {
            if (not length($reserve = $args{$_})) {
                $self->_set_err("Invalid value for argument '-reserve'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?snapshot/) {
            if (not length($snapshot = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?path/) {
            if (not length($path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $filer   = $self->get_filername();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (-cmd                 => "lun-create-from-snapshot",
                    -cmd_filer           => $filer,
                    -check_result        => 1,
                    -cmd_interface       => "zapi",
                    -path                => $path,
                    "-snapshot-lun-path" => $snapshot,);

    # Check to see whether or not to pass no reservation parameter
    if ($reserve =~ /0|n|no|f|false/i) {
        $cmd_args{"-space-reservation-enabled"} = "false";    #zapi value
    }
    $cmd_args{"-type"} = $type if defined($type);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # Ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # Since Interface is zapi, convert the output from xml to a string.
    $output_hash = $self->{parent}->{parent}->xml2hash($output);

    #print the output for debug purpose
    $dumpOut = Dumper($output_hash);
    $self->sanlog(-msg       => "Output Hash is:\n$dumpOut",
                  -msg_level => 9);
    $status = $output_hash->{status};
    if ($status eq "failed") {
        $self->_set_err("ZAPI failed: ", $output_hash->{reason});
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
} ## end sub create_from_snapshot()

##############################################################################
# Subroutine Name: get_vdisk_attributes()
# Objective:       Get path, volume name and filer name for the specified LUN
# Details:         See POD documentation at the end of this file
##############################################################################
sub get_vdisk_attributes() {
    my ($self, %args) = @_;
    my ($output, $output_hash, $resultsHash_ref, $serial_no, $cmd_obj);
    my $lun_vdisk_attributes = {};

    # Applicable only for zapi
    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?serial_no/) {
            if (not length($serial_no = $args{$_})) {
                $self->_set_err("Invalid value for argument '-serial_no'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-get-vdisk-attributes",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-serial-number" => $serial_no,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    # Extract the required data only
    $lun_vdisk_attributes->{'filer-name'} = $output_hash->{'filer-name'};
    $lun_vdisk_attributes->{'path'}       = $output_hash->{'path'};
    $lun_vdisk_attributes->{'vdisk-snapshot-name'} =
      $output_hash->{'vdisk-snapshot-name'};

    return $lun_vdisk_attributes;
}    ### end sub get_vdisk_attributes()
##############################################################################
# Subroutine Name: has_scsi_reservations()
# Objective:       Queries for all types of scsi reservations
# Details:         See POD documentation at the end of this file
#############################################################################
sub has_scsi_reservations() {
    my ($self) = @_;
    my ($cmd_obj, $output, $output_hash, $resultsHash_ref);

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-has-scsi-reservations",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-path"          => $self->{path},);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...
    if ($output_hash->{'is-reservation-held'} =~ /true/) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}    # end has_scsi_reservations()
#############################################################################
# Subroutine Name: restore_status()
# Objective:       Get state of LUN restore.
# Details:         See POD documentation at the end of this file
#############################################################################
sub restore_status() {
    my ($self) = @_;
    my ($cmd_obj, $output, $output_hash, $resultsHash_ref);

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-restore-status",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-path"          => $self->{path},);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...
    if ($output_hash->{'is-running'} =~ /true/) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}    # end lun_restore_status()
#############################################################################
# Subroutine Name: port_has_scsi_reservations()
# Objective:       Queries for all types of scsi reservations
# Details:         See POD documentation at the end of this file
##############################################################################
sub port_has_scsi_reservations() {
    my ($self, %args) = @_;
    my ($cmd_obj, $output, $output_hash, $resultsHash_ref, $portname);

    # Applicable only for zapi
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?portname/) {
            if (not length($portname = $args{$_})) {
                $self->_set_err("Invalid value for argument '-portname'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    if (!defined($portname)) {
        $self->_set_err("portname is required argument");
        return $ClientUtil::FALSE;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-port-has-scsi-reservations",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",
                    "-path"          => $self->{path},
                    "-portname"      => $portname,);

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...

    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to get lun alignment list info " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    if ($output_hash->{"is-reservation-held"} =~ /true/) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}    # end sub port_has_scsi_reservations()
##############################################################################
# Subroutine Name: snapshot_restore_file()
# Objective:       Reverts a file to a revision from a specified snapshot.
# Details:         See POD documentation at the end of this file
#############################################################################
sub snapshot_restore_file() {
    my ($self, %args) = @_;
    my ($restore_path, $force, $snapshot, $output, $resultsHash_ref);
    my ($output_hash);
    my $path      = $self->{path};
    my $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    my $interface = $self->{parent}->{parent}->interface();

    $self->_clear_err();

    #
    # Parse the arguments
    #
    foreach (keys %args) {
        if (/-?restore_path/) {
            if (not length($restore_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-restore_path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            if (defined $force and $force =~ /yes|TRUE|1/i) {
                $force = 1;
            }
            next;
        } elsif (/-?snapshot/) {
            if (not length($snapshot = $args{$_})) {
                $self->_set_err("Invalid value for argument '-snapshot'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    }    # end foreach

    # input parameter validation
    if (not length $snapshot) {
        $self->_set_err("snapshot is required argument");
        return $ClientUtil::FALSE;
    }

    # if the restore path is not provided restore on the same path
    if (not length $restore_path) {
        $restore_path = $path;
    }

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "snapshot-restore-file",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $path,
                    "-restore-path"  => $restore_path,
                    "-snapshot"      => $snapshot,);
    if ($interface eq "cli") {
        $cmd_args{"-force"} = " " if ($force);
        $self->offline();
    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    if ($interface eq "zapi") {
        $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("unable to lun  initiator is-logged-in "
                            . $output_hash->{reason});
            return $ClientUtil::FALSE;
        }    # end if($output_hash-> ...
    }

    # just to wait till restore completes
    sleep(10);
    return $ClientUtil::TRUE;
}    ### end sub snapshot_restore_file()
################################################################################
# Subroutine Name: persistent_resv_clear()
# Objective:       Clears all SCSI-2 reservations
# Details:         See POD documentation at the end of this file
################################################################################
sub persistent_resv_clear() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj, $path, $interface);

    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();
    $path    = $self->{path};

    # implemented only in Zapi. need to be implemented in cli BURT NO:511285
    $interface = "zapi";

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-clear-persistent-reservation-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-cmd_privilege" => "diag",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...
    return $ClientUtil::TRUE;

    # cli needs to done
}    ### sub lun_persistent_resv_clear()
################################################################################
# Subroutine Name: persistent_resv_show()
# Objective:       show all SCSI-2 reservations
# Details:         See POD documentation at the end of this file
################################################################################
sub persistent_resv_show() {
    my ($self) = @_;
    my ($output, $output_hash, $resultsHash_ref, $cmd_obj, $path, $interface);

    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #$interface = $self->{parent} ->{parent}->interface();
    # implemented only in Zapi. need to be implemented in cli BURT NO:511285
    $interface = "zapi";
    $path      = $self->{path};

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-get-persistent-reservation-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-cmd_interface" => $interface,
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-path"          => $path,
                    "-cmd_privilege" => "diag",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...
         # my $dumpOut = Dumper($output_hash);
         # $self->sanlog(-msg       => "Output Hash is:\n$dumpOut" );
    return $output_hash

      # cli needs to done
}    ### sub lun_persistent_resv_show()
################################################################################
# Subroutine Name: get_maxsize()
# Objective:       create a new lun using an existing file
# Details:         See POD documentation at the end of this file
################################################################################
sub get_maxsize() {
    my ($self, %args) = @_;
    my ($path, $type, $filer, $cmd_obj, $resultsHash_ref);
    my ($output, $status, $output_hash);
    my $result = {};
    $self->_clear_err();

    # implemented only in Zapi. need to be implemented in cli BURT NO:511285
    # parse arguments
    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?path/) {
            if (not length($path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Invalid parameters specified for '\$_\'");
            return $ClientUtil::FALSE;
        }
    } ## end foreach (keys %args)

    $filer   = $self->get_filername();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = (
        -cmd           => "lun-get-maxsize",
        -cmd_filer     => $filer,
        -check_result  => 1,
        -cmd_interface => "zapi",
        -path          => $path,               # Need to pass path of the volume
        -type          => $type,);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }

    #
    # Get the output
    #
    $output = $resultsHash_ref->{output};

    # ensure result_check passed
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # Since Interface is zapi, convert the output from xml to a string.
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    $status      = $output_hash->{status};
    if ($status eq "failed") {
        $self->_set_err("ZAPI failed: ", $output_hash->{reason});
        return $ClientUtil::FALSE;
    }
    $result = $output_hash;

    return $result;
} ## end sub get_maxsize()
##############################################################################
# Subroutine Name: config_check_cfmode_info()
# Objective:       Returns a list of configuration warnings related to initiator
# Details:         See POD documentation at the end of this file
##############################################################################
sub config_check_cfmode_info() {
    my ($self) = @_;
    my ($cmd_obj, $output, $output_hash, $resultsHash_ref);
    my $config_check_cfmode_info = {};

    # implemented only in Zapi. need to be implemented in cli BURT NO:511285
    $self->_clear_err();
    $cmd_obj = $self->{parent}->{parent}->cmd_obj();

    #
    # Build the command line and execute the command
    #
    my %cmd_args = ("-cmd"           => "lun-config-check-cfmode-info",
                    "-cmd_filer"     => $self->get_filername(),
                    "-check_result"  => 1,
                    "-check_output"  => 1,
                    "-cmd_interface" => "zapi",);
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("Calling execute_cmd: ", $cmd_obj->errmsg());
        $self->sanlog(-msg => join("\n", $cmd_obj->errmsg()));
        return $ClientUtil::FALSE;
    }    # end if($cmd_obj)
         #
         # Get the output
         #
    $output = $resultsHash_ref->{output};

    #
    # Since we checked the result, lets see if our check was OK
    #
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Failed result check: $output ");
        return $ClientUtil::FALSE;
    }    # end unless ...
    unless ($resultsHash_ref->{output_check}) {
        $self->_set_err("Output check failed: $output");
        return $ClientUtil::FALSE;
    }    # end unless ...
    $output_hash = $self->{parent}->{parent}->xml2hash($output);
    if ($output_hash->{status} eq "failed") {
        $self->_set_err(
             "unable to lun  initiator is-logged-in " . $output_hash->{reason});
        return $ClientUtil::FALSE;
    }    # end if($output_hash-> ...

    $config_check_cfmode_info->{'invalid-use-partner-cfmode-settings'} =
      $output_hash->{'invalid-use-partner-cfmode-settings'};
    $config_check_cfmode_info->{'invalid-use-partner-ostype-settings'} =
      $output_hash->{'invalid-use-partner-ostype-settings'};
    $config_check_cfmode_info->{'invalid-cfmode-settings'} =
      $output_hash->{'invalid-cfmode-settings'};
    $config_check_cfmode_info->{'invalid-ostype-cfmode-settings'} =
      $output_hash->{'invalid-ostype-cfmode-settings'};
    $config_check_cfmode_info->{'fcp-cfmode'} = $output_hash->{'fcp-cfmode'};

    #print the output for debug purpose
    # my $dumpOut = Dumper($config_check_cfmode_info);
    # $self->sanlog(-msg       => "config_check_cfmode_info is:\n$dumpOut" );
    return $config_check_cfmode_info;
}    # end sub lun_config_check_cfmode_info()

################################################################################
# Subroutine Name: get_lun_attributes_struct()
# Objective:       Collecting lun attributes of given luns in the form of hash
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_attributes_struct {
    my $self          = shift;
    my @lun_obj_array = @_;
    my ( %lun_attr_hash, %lun_show_hash );
    my $return_code = 1;

    my ( $cmd, $text, $errmsg );
    foreach my $lun_obj (@lun_obj_array) {
        my $lun_path = $lun_obj->get_path();
        if ( $lun_obj->error() ) {
            $errmsg = $lun_obj->errmsg();
            $self->sanlog( -msg => "unable to retrieve lun path" . $errmsg );
            return $ClientUtil::FALSE;
        }
        $text = $lun_obj->get_attribute();
        if ( $lun_obj->error() ) {
            $errmsg = $lun_obj->errmsg();
            $self->sanlog(
                -msg => "unable to retrieve lun attribute" . $errmsg );
            return $ClientUtil::FALSE;
        } ## end if ( $lun_obj->error())
        my @info = split( "\n", $text );
        foreach my $line (@info) {
            if ( ( $line =~ /lun attribute:/ ) || ( $line =~ /usage:/ ) ) {
                ExitSoftError("Unexpected message from ($cmd): $line");
            } elsif ( $line =~ /^\s+enabled\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+enabled =\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'enabled'} = $temp;
            } elsif ( $line =~ /^\s+path_last\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+path_last =\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'path_last'} = $temp;
            } elsif ( $line =~ /^\s+\@Anvil\@FSConsistent\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+\@Anvil\@FSConsistent\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'AnvilFSConsistent'} = $temp;
            } elsif ( $line =~ /^\s+host_stamp\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+host_stamp\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'host_stamp'} = $temp;
            } elsif ( $line =~ /^\s+serial\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+serial\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'serial'} = $temp;
            } elsif ( $line =~ /^\s+type\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+type\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'type'} = $temp;
            } elsif ( $line =~ /^\s+spacereserve\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+spacereserve\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'spacereserve'} = $temp;
            } elsif ( $line =~ /^\s+cylinder_size\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+cylinder_size\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'cylinder_size'} = $temp;
            } elsif ( $line =~ /^\s+extent_size\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+extent_size\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'extent_size'} = $temp;
            } elsif ( $line =~ /^\s+custom\s+=\s/ ) {
                my $temp = $line;
                $temp =~ s/^\s+custom\s+=\s+(.*)$/$1/;
                $lun_attr_hash{$lun_path}{'custom'} = $temp;
            }
        }    # end foreach my $line (@info)
        $lun_show_hash{ $lun_obj->get_path() } = $lun_obj->list_info();
        if ( $lun_obj->error() ) {
            $errmsg = $lun_obj->errmsg();
            $self->sanlog( -msg => "unable to retrieve attribute" . $errmsg );
            return $ClientUtil::FALSE;
        }
    } ## end foreach my $lun_obj (@lun_obj_array)
    return ( \%lun_attr_hash, \%lun_show_hash );
}    # end get_lun_attributes_struct()

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new Lun object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($access, $conn,   $dir,  $exist, $filer,  $log,
        $lun,    $mapped, $maps, $name,  $parent, $path,
        $self,   $size,   $type, $state, $file_name,$from_snapshot,
        @display_output,$prefix);

    $self = {};

    bless $self, $class;

    # Space reservation on by default
    my $reserve = 1;

    #space_alloc is off by default
    my $space_alloc = 0;

    # Assume lunid and igroup have not been assigned.
    my $lunid  = "";
    my $igroup = "";
    $self->sanlog(-msg => "lun creation started ");

    # parse arguments
    foreach (keys %args) {
        if (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun'");
                return $self;
            }
            next;
        } elsif (/-?path/) {
            if (not length($path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $self;
            }
            next;
        } elsif (/-?filer/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filer'");
                return $self;
            }
            next;
        } elsif (/-?reserve/) {
            if (not length($reserve = $args{$_})) {
                $self->_set_err("Invalid value for argument '-reserve'");
                return $self;
            }
            next;
        } elsif (/-?space_alloc/) {
            if (not length($space_alloc = $args{$_})) {
                $self->_set_err("Invalid value for argument '-space_alloc'");
                return $self;
            }
            next;
        } elsif (/-?dir/) {
            if (not length($dir = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dir'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $self;
            }
            next;
        } elsif (/-?state/) {
            if (not length($state = $args{$_})) {
                $self->_set_err("Invalid value for argument '-state'");
                return $self;
            }
            next;
        } elsif (/-?access/) {
            if (not length($access = $args{$_})) {
                $self->_set_err("Invalid value for argument '-access'");
                return $self;
            }
            next;
        } elsif (/-?_exist/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Invalid value for argument '-_exist'");
                return $self;
            }
            next;
        } elsif (/-?mapped/) {
            if (not length($mapped = $args{$_})) {
                $self->_set_err("Invalid value for argument '-mapped'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loghdl/) {
            if (not length($log = $args{$_})) {
                $self->_set_err("Invalid value for argument '-loghdl'");
                return $self;
            }
            next;
        } elsif (/-?_maps/) {
            if (not length($maps = $args{$_})) {
                $self->_set_err("Invalid value for argument '-maps'");
                return $self;
            }

            next;
        } elsif (/-?from_file/) {
            if (not length($file_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-from_file'");
                return $self;
            }

            next;
        } elsif (/-?from_snapshot/) {
			if (not length($from_snapshot = $args{$_})) {
				$self->_set_err("Invalid value for argument '-from_snapshot'");
                return $self;
            }

            next;
        }elsif (/-?prefix/) {
           if (not length($prefix = $args{$_})) {
				$self->_set_err("Invalid value for argument '-prefix'");
                return $self;
            }

            next;
        } elsif (/-?id/) {
            $lunid = $args{$_};
            next;
        } elsif (/-?igroup/) {
            $igroup = $args{$_};
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $self;
        }
    } ## end foreach (keys %args)

    $self->{log} = $log if (defined $log);

    # check for required parameter lun name.  It should not
    # contain path information
    unless (defined $name and length $name and $name !~ /\//) {
        $self->_set_err(
                     "Missing or invalid value for required parameter '-name'");
        return $ClientUtil::FALSE;
    }

    if(!defined $from_snapshot) {
		# check for parameter lun type which is needed is exist isn't set.
		if (!defined $exist and $exist ne 1) {
			unless (defined $type and length $type) {
				$self->_set_err(
						 "Missing or invalid value for required parameter '-type'");
				return $ClientUtil::FALSE;
			}
		}
		#
		# Check for required parameter lun size. It should contain units
		#
		unless (defined $size and length $size and $size =~ /\d+[b|k|m|g|t]?/i) {
			$self->_set_err(
						 "Missing or invalid value for required parameter '-size'");
			return $ClientUtil::FALSE;
		}
	}
	# Check for Volume, parent object.
    unless (defined $parent and length $parent) {
        $self->_set_err(
                "Required parameter parent volume object is missing '-parent'");
        return $ClientUtil::FALSE;
    }

    # Check for SanFiler object.
    unless (defined $filer and length $filer) {
        $self->_set_err("Required parameter filer object is missing '-filer'");
        return $ClientUtil::FALSE;
    }

    # inititialize object variables
    $self->{filer}       = $filer;
    $self->{parent}      = $parent;
    $self->{size}        = $size;
    $self->{name}        = $name;
    $self->{lunid}       = $lunid;
    $self->{igroup}      = $igroup;
    $self->{maps}        = $maps;
    $self->{space_alloc} = $space_alloc;

    #The Lun object has Prefix value if the LUN is created with prefix.
    if (defined $exist and $exist eq 1 and !$prefix) {
       @display_output = $self->display(-path => $path);
       if ($self->error()){
          $self->_set_err("Failed to display the lun list");
          return $ClientUtil::FALSE;
       }
       foreach(@display_output){
          if($_ =~ /Prefix Size:\s+(\w+)/) {
             $self->{prefix}      = $1;
             last;
          }
       }
    } else {
       if (defined($prefix)){
          $self->{prefix}      = $prefix;
       }
    }
    #
    # If we are adding an existing lun then we are just
    # storing lun attributes
    #
	if (defined $exist and $exist eq 1) {
		$self->{type}   = $type if defined($type);
        $self->{path}   = $path;
        $self->{dir}    = $dir;
        $self->{access} = $access;
        $self->{state}  = $state;
        $self->{mapped} = $mapped;
    } else {
        if (defined($path)) {

            # Use a path if we are given one.  Note that this will allow the user to
            # create luns within qtrees.
            $self->{path} = $path . "/" . $name;

        } else {
            $self->{path} = "/vol/" . $self->{parent}->get_name() . "/" . $name;
        }
        $self->{dir}     = "/vol/" . $self->{parent}->get_name();
        $self->{access}  = "r/w";
        $self->{state}   = "online";
        $self->{reserve} = $reserve;
        $self->{mapped}  = 0;
        $self->{type}    = $type;
	if ($from_snapshot) {
	    $self->{from_snapshot} = $from_snapshot;
            $self->_create_from_snapshot();
            return $self if ($self->error);
        }
	else {
	    if ($file_name) {
		$self->{file_path} =
		"/vol/" . $self->{parent}->get_name() . "/" . $file_name;
		$self->_create_from_file();
		return $self if ($self->error);
	    } else {
		$self->_create();
		return $self if ($self->error);
	    }
	}
    } ## end else [ if (defined $exist and...

    return $self;
} ## end sub _new

################################################################################
# Subroutine Name: _create()
# Objective:       create a new lun
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;
    my ($cmd_obj, $path, $prefix, $size, $type,$priv,
        $reserve, $space_alloc, $name, $volname,
        $filer,   $interface,   $resultsHash_ref);

    $path        = $self->{path};
    $size        = $self->{size};
    $type        = $self->{type};
    $name        = $self->{name};
    $reserve     = $self->{reserve};
    $space_alloc = $self->{space_alloc};
    $prefix      = $self->{prefix} ;
    $volname     = $self->{parent}->get_name();
    $filer       = $self->get_filername();
    $interface   = $self->{parent}->{parent}->interface();
    $cmd_obj     = $self->{parent}->{parent}->cmd_obj();

    #Build the command line and execute the command
    my %cmd_args = ("-cmd"           => "lun-create-by-size",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-type"          => $type,
                    "-path"          => $path,
                    );

    # convert size to bytes if zapi
    if (lc($interface) eq "zapi") {
        $size = &ClientUtil::convert_to_bytes($size);
        $cmd_args{"-size"} = $size;
    } else {
        $cmd_args{"-size"} = $size;
    }

    # Check to see whether or not to pass no reservation parameter
    if ($reserve =~ /0|n|no|f|false/i) {
        if ($interface eq "cli") {
            $cmd_args{"-noreserve"} = "";
        } else {
            $cmd_args{"-space-reservation-enabled"} = "false";    #zapi value
        }
    }

    if ($space_alloc =~ /1/) {
        if ($interface eq "cli") {
            $cmd_args{"-space_alloc"} = "";
        } else {
            $self->_set_err("zapi creation for space_alloc not suported");
            return $ClientUtil::FALSE;
        }
    }

    if (defined ($prefix )){
        $cmd_args{"-cmd_privilege"} = "advanced";
        if ($interface eq "cli") {
	    $cmd_args{"-prefix"} = $prefix;
	} else {
	    $cmd_args{"-prefix-size"} = $prefix;
	}

    }

    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Unable to create lun", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
    my @output = split("\n", $output);

        # Message of success is not always returned depending on lun size
        if (defined $output[0] and length $output[0]) {
            if (($output[0] !~ /lun create:/i)&& (not defined $prefix )){
                $self->_set_err("Unable to create $size lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }

            if ($output[0] =~ /lun create: A LUN exists at the path/i) {
                $self->_set_err("Unable to create lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }
        } ## end if (defined $output[0]...
    } ## end else [ if ($interface eq "zapi")

    return $ClientUtil::TRUE;
} ## end sub _create

################################################################################
# Subroutine Name: _create_from_file()
# Objective:       create a new lun using an existing file
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_from_file {
    my ($self) = @_;
    my ($cmd_obj,   $path,            $type,
        $name,      $volname,         $filer,
        $interface, $resultsHash_ref, $file_path,
	$prefix,    $reserve,   $space_alloc);

    $path      = $self->{path};
    $type      = $self->{type};
    $name      = $self->{name};
    $volname   = $self->{parent}->get_name();
    $filer     = $self->get_filername();
    $interface = $self->{parent}->{parent}->interface();
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    $file_path = $self->{file_path};
    $reserve     = $self->{reserve};
    $space_alloc = $self->{space_alloc};
    $prefix      = $self->{prefix} ;

    # Build the command line and execute the command
    my %cmd_args = ("-cmd"           => "lun-create-from-file",
                    "-cmd_filer"     => $filer,
                    "-check_result"  => 1,
                    "-cmd_interface" => $interface,
                    "-path"          => $path,
                    "-file-name"     => $file_path,
                   );
	# Check to see whether or not to pass no reservation parameter
    if ($reserve =~ /0|n|no|f|false/i) {
        if ($interface eq "cli") {
            $cmd_args{"-noreserve"} = "";
        } else {
            $cmd_args{"-space-reservation-enabled"} = "false";    #zapi value
        }
    }
	if ($interface eq "cli") {
		$cmd_args{"-type"} = $type;
	} else {
		$cmd_args{"-ostype"} = $type;    #zapi value
	}

    if ($space_alloc =~ /1/) {
        if ($interface eq "cli") {
            $cmd_args{"-space_alloc"} = "";
        } else {
            $self->_set_err("zapi creation for space_alloc not suported");
            return $ClientUtil::FALSE;
        }
    }
     if (defined ($prefix )){
            if ($interface eq "cli") {
                    $cmd_args{"-prefix"} = $prefix;
                    $cmd_args{"-cmd_privilege"} = "diag";
            } else {
                 $self->_set_err("zapi creation for prefix-size not suported");
            return $ClientUtil::FALSE;

            }

    }


    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
		my $output_hash = $self->{parent}->{parent}->xml2hash($output);
		if ($output_hash->{status} eq "failed") {
            $self->_set_err("Unable to create lun", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);

        # Message of success is not always returned depending on lun size
        if (defined $output[0] and length $output[0]) {
            if (($output[0] !~ /lun create:/i)&& (not defined $prefix )){
                $self->_set_err("Unable to create lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }

            if ($output[0] =~ /lun create: A LUN exists at the path/i) {
                $self->_set_err("Unable to create lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }
        } ## end if (defined $output[0]...
    } ## end else [ if ($interface eq "zapi")

    return $ClientUtil::TRUE;
} ## end sub _create_from_file

################################################################################
# Subroutine Name: _create_from_snapshot()
# Objective:       create a new lun using a snapshot
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_from_snapshot {
	my ($self) = @_;
    my ($cmd_obj,   $path,            $type,
        $name,      $volname,         $filer,
        $interface, $resultsHash_ref, $snapshot,
		$reserve,   $space_alloc);

    $path      = $self->{path};
    $type      = $self->{type};
    $name      = $self->{name};
    $volname   = $self->{parent}->get_name();
    $filer     = $self->get_filername();
    $interface = "zapi"; #$self->{parent}->{parent}->interface();
    $cmd_obj   = $self->{parent}->{parent}->cmd_obj();
    $snapshot  = $self->{from_snapshot};
	$reserve     = $self->{reserve};
    $space_alloc = $self->{space_alloc};

    # Build the command line and execute the command
    my %cmd_args = (-cmd                 => "lun-create-from-snapshot",
                    -cmd_filer           => $filer,
                    -check_result        => 1,
                    -cmd_interface       => $interface,
                    -path                => $path,
                    "-snapshot-lun-path" => $snapshot,);

    $cmd_args{"-type"} = $type if defined($type);

	# Check to see whether or not to pass no reservation parameter
    if ($reserve =~ /0|n|no|f|false/i) {
        if ($interface eq "cli") {
            $cmd_args{"-noreserve"} = "";
        } else {
            $cmd_args{"-space-reservation-enabled"} = "false";    #zapi value
        }
    }

    if ($space_alloc =~ /1/) {
        if ($interface eq "cli") {
            $cmd_args{"-space_alloc"} = "";
        } else {
            $self->_set_err("zapi creation for space_alloc not suported");
            return $ClientUtil::FALSE;
        }
    }

	$self->sanlog(-msg => "Executing command");
    $resultsHash_ref = $cmd_obj->execute_cmd(%cmd_args);

    if ($cmd_obj->error()) {
        $self->_set_err("CmdsDriver failed: ", $cmd_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{parent}->{parent}->log_cmd_results($resultsHash_ref);

    # Get the output
    my $output = $resultsHash_ref->{output};

    # Ensure result_check and output_check passed.
    unless ($resultsHash_ref->{result_check}) {
        $self->_set_err("Result check failed: $output");
        return $ClientUtil::FALSE;
    }

    # If interface is zapi, convert the output from xml to a string.
    if ($interface eq "zapi") {
        my $output_hash = $self->{parent}->{parent}->xml2hash($output);
        if ($output_hash->{status} eq "failed") {
            $self->_set_err("Unable to create lun", $output_hash->{reason});
            return $ClientUtil::FALSE;
        }
    } else {
        my @output = split("\n", $output);

        # Message of success is not always returned depending on lun size
        if (defined $output[0] and length $output[0]) {
            if ($output[0] !~ /lun create:/i) {
                $self->_set_err("Unable to create lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }

            if ($output[0] =~ /lun create: A LUN exists at the path/i) {
                $self->_set_err("Unable to create lun named '$name' in "
                                  . "volume '$volname'",
                                @output);
                return $ClientUtil::FALSE;
            }
        } ## end if (defined $output[0]...
    } ## end else [ if ($interface eq "zapi")

    return $ClientUtil::TRUE;
} ## end sub _create_from_snapshot

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Lun.pm

=head1 Authors

=begin html

Gary Knopp

=end html

=head1 Description

Lun provides an API interface to a filer lun through the use rsh.  The package should
be instantiated through the SanFiler2 package get_lun() or get_luns() methods.

=head1 Attributes

=head2 B<name>

string = Lun Name

=head2 B<parent>

reference = Volume Object

=head2 B<filer>

reference = SanFiler Object

=head2 B<igroup>

string = Initiator Group Name

=head2 B<lunid>

integer =  Lun ID

=head2 B<maps>

hash = Map of Igroup to Lun ID

=head2 B<size>

string = Lun size in unit (k|m|g|t)

=head2 B<path>

string = Lun Path

=head2 B<dir>

string = Lun Directory

=head2 B<access>

string = Lun accessability (default="r/w")

=head2 B<state>

string = <online | offline>

=head2 B<reserve>

string = <enabled | disabled>

=head2 B<mapped>

boolean = <1 | 0>

=head2 B<log>

boolean = <1 | 0>

=head1 Public Methods

=head2 B<destroy(%args)>

Destroy a filer lun

=over 4

=item Inputs

=item * I<-force> (default: 0)

The option will automatically unmap the device (if mapped) so that it can
be destroyed.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces

cli, zapi

=item Examples

    $Test->log("-- Destroy lun");

    $lunObj->destroy();
    if ($lunObj->error()){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed lun information

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces

cli

=item Examples

    $Test->log("-- Retrieving lun information");

    $Test->log($lunObj->display());
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Supported Interfaces

All

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Supported Interfaces

All

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_filername()>

Name of filer lun is on

=over 4

=item Inputs

none

=item Outputs

Returns the name of the filer.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The lun is on filer '".$lunObj->get_filername()."'";

=back

=head2 B<get_geometry()>

Returns the lun geometry

=over 4

=item Inputs

none

=item Outputs

Returns the lun geometry.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and FALSE is returned.

=item Supported Interfaces

cli

=item Examples

    print "The lun geometry is '".$lunObj->get_geometry()."'";

=back

=head2 B<get_stats_counter()>

Get lun stats

=over 4

=item Inputs

none

=item Outputs

Returns a hash with keys from column headers for the lun stats call.
If an error occurs the error flag is set, errmsg is set to the failure
reason, and FALSE is returned.

=item Supported Interfaces

cli

=item Examples

    my $stats_hash_ref = $lunObj->get_stats_counter();
    foreach my $key (keys %{$stats_hash_ref}) {
       print "$key  =  " . $stats_hash_ref->{$key} . "\n";
    }

=back

=head2 B<get_stats()>

Get lun stats

=over 4

=item Inputs

none

=item Outputs

Returns a hash with keys from column headers for the lun stats call.
If an error occurs the error flag is set, errmsg is set to the failure
reason, and FALSE is returned.

=item Supported Interfaces

cli

=item Examples

    my $stats_hash_ref = $lunObj->get_stats();
    foreach my $key (keys %{$stats_hash_ref}) {
       print "$key  =  " . $stats_hash_ref->{$key} . "\n";
    }

=back

=head2 B<get_igroup()>

Name of igroup that lun is mapped too.

=over 4

=item Inputs

none

=item Outputs

Returns the name of the igroup.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The igroup name is '".$lunObj->get_igroup()."'";

=back

=head2 B<get_name()>

Name of filer lun

=over 4

=item Inputs

none

=item Outputs

Returns the name of the lun (w/o path).  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The lun name is '".$lunObj->get_name()."'";

=back

=head2 B<get_path()>

Returns lun path '/vol/[volName]/[lunName]'

=over 4

=item Inputs

none

=item Outputs

Returns the name of the lun (w path).  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The lun path is '".$lunObj->get_path()."'";

=back

=head2 B<get_state()>

Retrieve the current state of the lun.

=over 4

=item Inputs

none

=item Outputs

Returns the state of the lun.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The lun state is '".$lunObj->get_state()."'";

=back

=head2 B<get_access()>

Retrieve lun access permission

=over 4

=item Inputs

none

=item Outputs

Returns the lun access permission.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "Lun access is '".$lunObj->get_access()."'";

=back

=head2 B<get_id()>

Retrieve the lun id

=over 4

=item Inputs

none

=item Outputs

Returns the id that the lun is mapped too.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "Lun id is '".$lunObj->get_id()."'";

=back

=head2 B<get_size()>

Returns the size of the lun with unit of measure.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and 0 is returned.

=item Supported Interfaces

All

=item Examples

    $size=$lunObj->get_size();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

    print "Lun '".$lunObj->get_name()."' has a size of $size n";

=back

=head2 B<get_parent()>

Returns the parent of the lun object

=over 4

=item Inputs

none

=item Outputs

Returns the parent object if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty hash ref is returned.

=item Supported Interfaces

All

=item Examples

    $volObj = $lunObj->get_parent();
    if ($lunObj->error){
        $Test->log("Unable to get lun parent: " . join "\n", $lunObj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<reset_stats_counter()>

Zeros the lun stats

=over 4

=item Inputs

none

=item Outputs

Returns TRUE value if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and FALSE is returned.

=item Supported Interfaces

cli

=item Examples

    my $before_stats_hash_ref = $lunObj->get_stats_counter();
    $lunObj->reset_stats_counter();
    &run_some_io(&lunObj);
    my $after_stats_hash_ref = $lunObj->get_stats_counter();

=back

=head2 B<get_volname()>

Name of volume

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume (w/o path) the lun is on.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and "" is returned.

=item Supported Interfaces

All

=item Examples

    print "The lun is on volume '".$lunObj->get_volname()."'";

=back

=head2 B<is_mapped()>

Determine if a lun is mapped

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if mapped, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces

All

=item Examples

    # If mapped to $igroup_name then unmap from $igroup_name
    if ($Obj->is_mapped($igroup_name)){
        $Obj->unmap($igroup_name);
    }

    # If mapped to any igroup then unmap
    if ($Obj->is_mapped()){
        $Obj->unmap();
    }

=back

=head2 B<is_offline()>

Determine if a lun if offline

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if offline, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Supported Interfaces

All

=item Examples

     # If offline place online
    if ($Obj->is_offline()){
        $Obj->online()
    }

=back


=head2 B<map($name)>

Map a lun to an igroup

=over 4

=item Inputs

=item * $name

The name of the igroup to use.

=item * I<-lun_id> (default: undefined, if defined, will use specified lun id )

This is not a required option.
If supplied, the option will map the lun to a specified lun id;
when not specified, lun_id is automatically assigned by the command.

=item * I<-force> (default: 0)

This is not a required option.
If supplied, the option will force the lun to be mapped.

=item Outputs

Returns a value of 1 if successfully mapped , otherwise 0.
If an error occured the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli

=item Examples

    # Call the map method by only passing in igroup name.
    $lunObj->map("phantom_igroup");
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

    # Call the map method with igroup name and optional -lun_id and -force.
    # Note: Igroup name does not require a -name tag.
    $lunObj->map("phantom_igroup2",
                -lun_id = 2,
                -force =  1);
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<offline()>

Place a lun offline

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces

cli

=item Examples

    # Place a lun offline
    $Obj->offline();
    if ($Obj->error){
        $Test->log($Obj->errmsg());
        return 0;
    }

=back

=head2 B<online()>

Place a lun online

=over 4

=item Inputs

=item * I<-force> (default: 0)

If supplied, the option will force the lun to go offline.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interfaces

cli

=item Examples

    # Place a lun online
    $Obj->online();
    if ($Obj->error){
        $Test->log($Obj->errmsg());
        return 0;
    }

=back

=head2 B<unmap()>

Unmap a lun to an igroup

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if successfully unmapped , otherwise 0.  If an error occured the error
flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli

=item Examples

    $lunObj->unmap();
    if ($lunObj->error){
                $Test->log($lunObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<start_dense_clone(%args)>

Start a dense clone operation

=over 4

=item Inputs

=item * I<-dest_path>

Destination path in format /vol/<volname>/...

=item * I<-nosnap>

Do not use temporary snapshot for cloning

=item * I<-change_log>

Do change logging for clone blocks

=item * I<-block_ranges>

List of block ranges for sub-file/sub-LUN cloning. For sub-LUN cloning
the block range specified will be considered as SCSI LBA range. For one
set of source and destination block range followings are error conditions:
1) If source and destination block ranges overlap.
2) Source or destination block range is out of file. If multiple set of
block ranges are specified in one clone operation, then user is responsible
for different block ranges do not overlap with each other. Each set of block
range will be processed one by one for cloning. So, user may get random data
if different set of block ranges overlap with each other.

The input is a string in this format: <src_fbn>:<dest_fbn>:<fbn_cnt> ...
Where more than one block_range can be set
     <src_fbn>  - Starting fbn of the source block range
     <dest_fbn> - Starting fbn of the destination block range
     <fbn_cnt>  - Number of blocks to be cloned

=item * I<-timeout> (default: 0)

How long you would like to wait for the dense clone to complete, if
not set, will default to zero. If it is zero we don't wait for completion
and return immediately.

=item Outputs

Returns the clone operation id if successful, otherwise "". If an
error occurs the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli, zapi

=item Examples

    $clone_op_id = $lunObj->start_dense_clone(-dest_path => $dest_path);

=back

=head2 B<stop_dense_clone(%args)>

Stop a dense clone operation

=over 4

=item Inputs

=item * -clone_op_id

The id of the dense clone that you want to stop

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an
error occurs the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli, zapi

=item Examples

    $lunObj->stop_dense_clone(-clone_op_id => 25);

=back

=head2 B<clear_dense_clone(%args)>

Clear a dense clone operation

=over 4

=item Inputs

=item * -clone_op_id

The id of the dense clone that you want to clear

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an
error occurs the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli, zapi

=item Examples

    $lunObj->clear_dense_clone(-clone_op_id => 25);

=back

=head2 B<get_dense_clone_status(%args)>

Get the status of 1 or more dense clones.

=over 4

=item Inputs

=item * I<-clone_op_id>

The id of the dense clone that you want to get the status of,
if not set will return the status of all dense clones

=item Outputs

Returns a hash of clone id's which contains a hash of the status information,
otherwise "". If an error occurs the error flag is set, errmsg is set to the
failure reason. The has will either include the clone-id, source-file,
destination-file, block-ranges, state, and type. Or the clone-id and reason
why it failed to start.

=item Supported Interfaces

cli, zapi

=item Examples

    $clone_status_hash = $lunObj->get_dense_clone_status();

    $clone_op_id = $clone_status_hash->{"clone_op_id"};
    $reason = $clone_status_hash->{"reason"};

=back

=head2 B<wait_for_dense_clone_complete(%args)>

Waits for a dense clone's status to be 'completed'.

=over 4

=item Inputs

=item * -clone_op_id

The id of the dense clone that you want to wait until complete

=item * I<-timeout> (default: 1200)

How long to wait (in seconds) for the dense clone before timing out

=item * I<-interval> (default: 10)

How often (in seconds) to check the clone for completed status

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an
error occurs the error flag is set, errmsg is set to the failure reason.

=item Supported Interfaces

cli, zapi

=item Examples

    $lunObj->wait_for_dense_clone_complete(-clone_op_id => 25,
                                           -timeout     => 30);

=back

=head2 B<clone_create(%args)>

Create a lun clone

=over 4

=item Inputs

=item * -snap_name

Name of the snap shot to use for clone creation

=item * I<-path> (default: Path of the lun with the lun having a clone_ prefix).

Path of the clone.

=item * I<-noreserve> (default: 0,no).

Use the -o noreseve option on the lun clone create command line

=item Outputs

Returns the lun object.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli

=item Examples

    $lunObj->clone_create(-snap_name => volX_snap3);
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<resize(%args)>

Resize a lun

=over 4

=item Inputs

=item * -size

Size of the lun.
If cli, the size is designated with the numerical value plus units,(k,m,g,t).
If zapi, the size can be designated in bytes only or with the numerical value
plus units, (k,m,g.t) because the method itself will convert the value to byes,
in the case of the interface being zapi if needed.

=item * I<-force> (default: 0)

Forcibly reduce the size

=item Outputs

Returns 1 on success.  On error, 0 is returned, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli and zapi

=item Examples

    # This example works for both cli or zapi since the size will be
    # converted to bytes if the interface is set to zapi.
    $lunObj->resize(-size => "10g");
    if ($lunObj->error){
        $Test->log("Unable to resize lun" . join "\n", $lunObj->errmsg());
        return $SanClient::FAIL;
    }

    # This example is another way to pass in the size if the interface is zapi.
    $lunObj->resize(-size => "1073741824");
    if ($lunObj->error){
        $Test->log("Unable to resize lun" . join "\n", $lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<clone_split_start>

Start a lun clone split

=over 4

=item Inputs

none

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli, zapi

=item Examples

    $lunObj->clone_split_start();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<clone_split_stop>

Stop a lun clone split

=over 4

=item Inputs

none

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli, zapi

=item Examples

    $lunObj->clone_split_stop();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<clone_split_status>

Get the status of a lun clone split

=over 4

=item Inputs

none

=item Outputs

Returns a string on success.  The zpai version will return XML, cli
will return plain text.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli, zapi

=item Examples

    my $status = $lunObj->clone_split_status();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }
    print "Lun clone split status: $status\n";

=back

=head2 B<clone_split_percent_complete>

Get the status of a lun clone split and returns the percent complete for
the split operation

=over 4

=item Inputs

none

=item Outputs

Returns an integer on success. If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli, zapi

=item Examples

    my $percent_complete = 0;
    while ($percent_complete < 100) {
        $percent_complete = $lunObj->clone_split_percent_complete();
        print "Lun clone split percent complete is $percent_complete\n";
        sleep 20;
    }
    print "Lun clone split complete\n";

=back

=head2 B<clone_split_wait>

Wait for a lun clone split to complete.  Percent complete will be checked
every 20 seconds until either the split operation is hit or the timeout
has expired.  If the timeout is expired, a failure is returned.

=over 4

=item Inputs

=item * I<-timeout> (default: 3600)

Number of seconds to wait before returning a failure

=item Outputs

Returns the lun object.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->clone_split_wait();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<get_comment>

Get the lun comment

=over 4

=item Inputs

None

=item Outputs

Returns the comment for the lun on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    my $lun_comment = $lunObj->get_comment();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_comment>

Set the lun comment

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_comment("This is my lun comment");
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<get_serial>

Get the lun serial number

=over 4

=item Inputs

None

=item Outputs

Returns the lun serial_number on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    my $serial_number = $lunObj->get_serial();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_serial>

Set the lun serial number

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_serial($serial_number);
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<lun_get_select_attribute>

Get lun selection attribute, creation time, and previous serial
number

=item Inputs
Lun path

=item Outputs

Returns Lun select information on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli

=item Examples

    @lun_sel_info = $lunObj->lun_get_select_attribute($lun);
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<get_share>

Get the lun share status

=over 4

=item Inputs

None

=item Outputs

Returns share_status on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    my $share_status = $lunObj->get_share();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_share>

Set the lun share status

=over 4

=item Inputs

Share status: "none", "read", "write", or "all"

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_share("none");
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<get_reservation>

Get the lun reservation

=over 4

=item Inputs

None

=item Outputs

Returns is-reservation-enabled on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    my $reservation_state = $lunObj->get_reservation();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_reservation>

Set the lun reservation

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_reservation();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_attribute>

Set the lun attribute

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_attribute(  -name  => "myattribute",
                             -value => "1" );
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<unset_attribute>

Unset the lun attribute

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->unset_attribute(  -name  => "myattribute" );
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<set_comment>

Set the lun comment

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_comment( -comment => "new" );
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<unset_comment>

Unset the lun comment

=over 4

=item Inputs

None

=item Outputs

Returns 1 on success.  If an error occurs, the error
flag is set and the errmsg contains the reason for the failure.

=item Supported Interfaces

cli,  zapi

=item Examples

    $lunObj->set_comment();
    if ($lunObj->error){
        $Test->log($lunObj->errmsg());
        return $SanClient::FAIL;
    }

=back

=head2 B<list_info(%args)>

Return information on the lun from the filer

=over 4

=item Inputs

=item * I<-path> (default: $self->get_path())

Path override for the call to get information.  Useful if you need information
on a lun and you don't have a lun object for that lun yet.

=item Outputs

Returns a hash with information about the lun. For the ZAPI interface,
we simply translate the XML returned into the hash and return the
data that came back.  For the CLI, we parse the data that came back
and the hash will contain fields that match those from the ZAPI.
Specifically, the hash for each lun will contain these fields:
path, size, online, mapped, read-only, serial-number, share-state,
is-space-reservation-enabled, multiprotocol-type, backing-snapshot,
maps.  Note that the maps field only comes back if we use the CLI and
it looks like this: igroup_name=1 (where 1 is the lun id).
If there are no luns, an undefined hash is returned.
If a failure occurs, the error flag is set, errmsg is set to
the failure reason, and 0 is passed back as the return value.

=item Supported Interfaces:

cli zapi

=item Examples

    $Test->nlog("-- Print if lun is mapped or not");
    my $lun_hash = $lun_obj->list_info();
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun info: " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }
    my $is_mapped = $lun_hash->{mapped};
    $Test->nlog("Lun is mapped") if ($is_mapped =~ /yes/);

=back


=head2 B<get_prefix_sufix()>

Returns the Prefix and Suffix values of the lun.

=over 4

=item Inputs

none

=item Outputs

Returns a hash reference which includes the prefix and suffix values.

=item Supported Interface:

cli zapi

=item Examples

    $Test->nlog("-- obtain the Prefix and siffix values");
    my $lun_hash = $lun_obj->get_prefix_sufix();
    if ($lun_obj->error()) {
        $Test->nlog("Error getting get_prefix_sufix " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<get_type()>

Get lun type.

=over 4

=item Inputs

none

=item Outputs

Returns the type of the lun.

=item Supported Interface:

cli zapi

=item Examples

    $Test->nlog("-- obtain the Lun type");
    my $type = $lun_obj->get_type();
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun type " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<get_holes_lbarange(I<%args>)>

Get the lba hole ranges.

=over 4

=item Inputs

=item * I<-start_lba>

The value after prefix of lun.

=item * I<-end_lba>

The value before suffix of lun.

=item Outputs

Returns the lun hole information.

=item Supported Interface:

zapi

=item Examples

    $Test->nlog("-- the lba range is ");
    my $lba_range = $lun_obj->get_holes_lbarange(
                                                -start_lba => $prefix+1,
                                                -end_lba   =>  $sufix-1,
                    );
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun lba range " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<is_lba_hole(I<%args>)>

For a given lba check if it is a hole or not.

=over 4

=item Inputs

=item * I<-check_lba>

The lba value needs to given.

=item Outputs

Returns a hash reference.

=item Supported Interface:

zapi

=item Examples

    $Test->nlog("-- the lba inforamation is ");
    my $lba_hash = $lun_obj->is_lba_hole(-check_lba => $lba );
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun lba information " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<get_lun_occupied_size()>

Gives the occupied size of the lun.

=over 4

=item Inputs

none

=item Outputs

occupied size of the lun.

=item Supported Interface:

zapi

=item Examples

    $Test->nlog("-- the occuied Size is ");
    my $size = $lun_obj->get_lun_occupied_size();
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun occupied size " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<create_from_snapshot(I<%args>)>

Creates a new lun with existing file.

=over 4

=item Inputs

=item * I<-type>

Specify the Lun type ( examples: "windows" ,"solaris", "linux"...)

=item * I<-reserve>

Reservation Parameters type Boolean Possible values are 0|n|no|f|false  1|y|yes|t|true

=item * I<-snapshot>

path of the snapshot.

=item * I<-path>

Path of the new lun to be created.

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    #$Test->nlog("-- creating a new lun ");
    $lun_obj->create_from_snapshot(
                                    -type     => "solaris",
                                    -path     => "/vol/$volume/new_lun",
                                    -snapshot => "/vol/$volume/.snapshot/snap1",
            );
    if ($lun_obj->error()) {
        $Test->nlog("Error creating lun from snapshot " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back


=head2 B<get_minsize(I<%args>)>

Returns the minimum size for different types of luns.

=over 4

=item Inputs

=item * I<-type>

Specify the Lun type ( examples: "windows" ,"solaris", "linux"...)

=item Outputs

Returns a minimum size of lun.

=item Supported Interface:

zapi

=item Examples

    $Test->nlog(" obtain the lun minimum size ");
    my $min_size = $get_minsize(-type = > "solaris");
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun minsize " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<set_select_attribute(I<%args>)>

Set the select attribute for the specified LUN.

=over 4

=item Inputs

=item * I<-select_value>

The attributes for a lun the Possible values are "copy", "active", "mirror".

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Supported Interface:

zapi

=item Examples

    #$Test->nlog("-- set the selected attribute to lun ");
    $lun_obj->set_select_attribute( -select_value => "copy"    );
    if ($lun_obj->error()) {
        $Test->nlog("Error to set the selected attribute " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back

=head2 B<show_staged_luns()>

Displays staged luns

=over 4

=item Inputs

=none

=item Outputs

Returns reference of an array of staged luns if successful.  E
lse if there are no staged luns,returns undef value

=item Supported Interfaces

cli

=item Examples

    $Test->log("-- Show staged luns");

    $lunObj->show_staged_luns();
    if ($lunObj->error()){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<is_lun_staged()>

Shows whether the lun is staged or not

=over 4

=item Inputs

=none

=item Outputs

Returns a scalar value of 1 if successful, otherwise 0.

=item Supported Interfaces

cli

=item Examples

    $Test->log("-- Show whether lun is staged");

    $lunObj->is_lun_staged();
    if ($lunObj->error()){
        $Test->log($lunObj->errmsg());
        return $TCase::FAIL;
    }

=back
=head2 B<get_maxsize(I<%args>)>

Possible Maxiamum size for a lun.

=over 4

=item Inputs

=item * I<-type>

Specify the Lun type ( examples: "windows" ,"solaris", "linux"...)

=item * I<-path>

Path of the volume where lun exists.

=item Outputs

returns a hash reference.

=item Supported Interface:

zapi

=item Examples

    $Test->nlog("-- get the lun maxsize information ");
    $lun_obj->get_maxsize(
                            -path => "/vol/$volume",
                            -type => "solaris",
            );
    if ($lun_obj->error()) {
        $Test->nlog("Error getting lun maxsize " . $lun_obj->errmsg())
        return $TCD::FAIL;
    }

=back
=head2 B<persistent_resv_clear(I<%args>)>

Clears all SCSI-2 reservations

=over 4

=item Inputs

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interface:

zapi

=item Examples

    $output = $lunObj->persistent_resv_clear();
    if ($lunObj->error()) {
                $Test->nlog("-- ERROR : " . $lunObj->errmsg());
                return $TCD::FAIL;
    }    #end if ($lunObj->error())

=back

=head2 B<persistent_resv_show(I<%args>)>

show all SCSI-2 reservations

=over 4

=item Inputs

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Supported Interface:

zapi

=item Examples

    $output = $lunObj->persistent_resv_show();
    if ($lunObj->error()) {
                $Test->nlog("-- ERROR : " . $lunObj->errmsg());
                return $TCD::FAIL;
    }    #end if ($lunObj->error())

=back

=head2 B<config_check_cfmode_info(I<%args>)>

Returns a list of configuration warnings related to initiator

=over 4

=item Inputs

=item Outputs

Returns a list of configuration warnings related to initiator

=item Supported Interface:

zapi

=item Examples

    $output = $lunObj->config_check_cfmode_info();
    if ($lunObj->error()) {
                $Test->nlog("-- ERROR : " . $lunObj->errmsg());
                return $TCD::FAIL;
    }    #end if ($lunObj->error())

=back

=head2 B<get_lun_attributes_struct(I<%args>)>

Returns lun attributes in the form of hashref

=over 4

=item Inputs

=item * I<@lunobjects>

Provide the Lun Objects of all the lun for collecting the attributes

=item Outputs

Returns lun attributes of the given luns in the form of hashref

=item Supported Interface:

zapi/cli

=item Examples

    ($lun_attr, $lun_show) = $lunObj->get_lun_attributes_struct(@lunobjects);
    if ($lunObj->error()) {
                $Test->nlog("-- ERROR : " . $lunObj->errmsg());
                return $TCD::FAIL;
    }    #end if ($lunObj->error())

=back

=head2 B<lun_clone_list_info(I<%args>)>

Lists all LUN clones with valid backing snapshots in the given snapshot.

=over 4

=item Inputs

=item * I<-snap_name>

Require a snapshot name.

=item Outputs

Returns a hash refernce which includes all the clones lists, on success
on failure it will return a zero.

=item Supported Interface:

zapi

=item Examples

    $output_hash = $lunObj->lun_clone_list_info(-snap_name => "test_snap");
    if ($lunObj->error()) {
        $Test->nlog("-- ERROR : " . $lunObj->errmsg());
        return $TCD::FAIL;
    } #end if ($lunObj->error())

=back

=head2 B<reset_stats()>

Resets (zeroes) block-protocol access statistics for LUN(s).

=over 4

=item Inputs

=item Outputs

Returns true, on success and returns zero on failure.

=item Supported Interface:

zapi

=item Examples

    $output= $lunObj->reset_stats();
    if ($lunObj->error()) {
        $Test->nlog("-- ERROR : " . $lunObj->errmsg());
        return $TCD::FAIL;
    } #end if ($lunObj->error())

=back

=head2 B<lun_snap_usage_list_info(I<%args>)>

Lists all LUNs backed by data in the specified snapshot.
It also lists the corresponding snapshots in which these LUNs exist

=over 4

=item Inputs

=item * I<-snap_name>

Require a snapshot name.

=item Outputs

Returns a hash refernce which includes lun snap usage lists, on success
on failure it will return a zero.

=item Supported Interface:

zapi

=item Examples

    $output_hash = $lunObj->lun_snap_usage_list_info(-snap_name => "test_snap");
    if ($lunObj->error()) {
        $Test->nlog("-- ERROR : " . $lunObj->errmsg());
        return $TCD::FAIL;
    } #end if ($lunObj->error())

=back

=head2 B<lun_config>

executes the 'lun config; command on the filer and
form forms a hash of the output and returns the same

=over 4

=item Inputs

=item Outputs

Returns a hash refernce which includes the command otput info

=item Supported Interface:

cli

=item Examples

    $output_hash = $lunObj->lun_();
    if ($lunObj->error()) {
        $Test->nlog("-- ERROR : " . $lunObj->errmsg());
        return $TCD::FAIL;
    } #end if ($lunObj->error())

=back

