#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to setup controller clusters
## @author porrua@netapp.com
## @status private
## @pod here

package San::SanMgr::Controller::setup;

# Compiler directives.
use strict;
use San::SanLog;
use ParamsMgr;
use San::SanUtil;
use EnGenX::Util;
use San::SanMgr::Controller::ControllerHelper;
use San::SanClient::ClientUtil;
use Digest::MD5 qw(md5_hex);
use Data::Dumper;
use NACL::C::LunBind;
use NACL::C::QosPolicyGroup;
use NACL::CS::QosPolicyGroup;
use NACL::C::Vserver;
use NACL::C::NetworkPortIfgrp;
use NACL::C::NetworkPortVlan;
use NACL::C::VolumeSnapshotAutodelete;
use NACL::C::Mkfile;
use NACL::C::LunPortset;
use NATE::BaseException qw (:try);

use vars qw(@ISA);
@ISA = qw(Error SanLog San::SanMgr::Controller::ControllerHelper);

###############################################################################
# Method:    setup_controller_cluster
# Objective: Setup run against a single controller cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup_controller_cluster {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($resource, $cserver, $aggr_data, $vserver_data) = undef;
    my ($cserver_name) = "";
    my (@cluster_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up cluster"));

    # Process arguments. A resource is nothing more than a controller cluster
    # element which may come from a config file (i.e. XML)
    &ParamsMgr::process_args(%args);

    # The main controller cluster data structure
    $resource = &ParamsMgr::get_arg(-arg      => "resource",
                                    -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
    }

    # First we need to identify/get a cserver on this cluster. This is
    # a cserver object which will be used to create all objects in the library.
    # NOTE: _create_cserver may need to be modified once c-mode libraries add
    # support for cservers
    $cserver = $self->_create_cserver(resource => $resource);
    if ($self->error()) {
        $self->_set_err("Failed to create cserver object");
        &SubIPC::fail(-object => $self);
    }

    # Get the cserver name
    $cserver_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err($cserver->errmsg());
        &SubIPC::fail(-object => $self);
    }

    # check if requested to modify auto support settings on the cluster
    if (exists($resource->{AUTOSUPPORT}) and length($resource->{AUTOSUPPORT})) {
        $cserver->autosupport_modify(node => "*",
                                     support => $resource->{AUTOSUPPORT},
                                     state => $resource->{AUTOSUPPORT},
                                     local_collection => $resource->{AUTOSUPPORT} eq "disable" ? "false" : "true");
        if ($cserver->error()) {
            $self->_set_err($cserver->errmsg());
            &SubIPC::fail(-object => $self);
        }
    }

    # check if requested to modify auto support settings on the cluster.
    # Only call if parm is true, because you are never allowed to set this value
    # to false.  ONTAP will reject it.
    if (exists($resource->{IPV6_ENABLED})
        and $resource->{IPV6_ENABLED} =~ /true/i) {
        $cserver->modify_ipv6(-enabled => $resource->{IPV6_ENABLED});
        if ($cserver->error()) {
            $self->_set_err($cserver->errmsg());
            &SubIPC::fail(-object => $self);
        }
    }

    # Create aggregates first
    $aggr_data = $self->_setup_aggregates(resource => $resource,
                                          cserver => $cserver);
    if ($self->error()) {
        $self->_set_err("Failed to setup aggregates");
        &SubIPC::fail(-object => $self);
    }

    push(@cluster_data, @{$aggr_data});

    # Create ifgrps
    my $ifgrp_data = $self->_setup_ifgrps(resource => $resource,
                                          cserver => $cserver);
    if ($self->error()) {
        $self->_set_err("Failed to setup ifgrps");
        &SubIPC::fail(-object => $self);
    }
    push(@cluster_data, @{$ifgrp_data});

    # Create vlans
    my $vlan_data = $self->_setup_vlans(resource => $resource,
                                        cserver => $cserver);
    if ($self->error()) {
        $self->_set_err("Failed to setup vlans");
        &SubIPC::fail(-object => $self);
    }
    push(@cluster_data, @{$vlan_data});

    # get all node objs
    my @node_objs = $self->_get_node_objs(resource => $resource);
    if ($self->error()) {
        $self->_set_err("Failed to get node objects");
        &SubIPC::fail(-object => $self);
    }

    # Create vservers (and everything inside them)
    $vserver_data = $self->_setup_vservers(resource => $resource,
                                           cserver => $cserver,
                                           node_objs => \@node_objs);

    if ($self->error()) {
        $self->_set_err("Failed to setup vservers");
        &SubIPC::fail(-object => $self);
    }

    # Append the name of the cserver used to create all objects
    foreach my $data (@cluster_data) {
        $data->{cserver} = $cserver_name;
    }

    # Append vserver data to the message that will be written to SanMgr
    push(@cluster_data, @{$vserver_data});

    # Write data to the parent (Controller::setup())
    &SubIPC::write(-msg => "Data containing controller cluster setup objects",
                   -obj_data => \@cluster_data,
                   -type => "setup_cluster_content_complete",
                   -name => "cluster_setup");
    if (&SubIPC::error()) {
        $self->_set_err("SanMgr failed to set cluster contents", &SubIPC::get_error());
        &SubIPC::fail(-object => $self);
    }

    &SubIPC::pass(-msg => "SanMgr finished setting up a cluster.");
} ## end sub setup_controller_cluster

###############################################################################
# PRIVATE METHODS
#
# NOTE: Method details are included in the header section.  The POD contains
# public method info. Don't expect to see any private method described there.
# This module follows a singular / plural convention, where plural methods
# basically iterate through a list and call their corresponding singular
# methods either subtesting each call or executing it serially
###############################################################################

###############################################################################
#                      A G G R E G A T E    M E T H O D S
#
#  Things to consider:
#  1. Aggregates are created serially for now. In the future, if we want things
#     to run in parallel, we'll need to create a subtest intead of calling
#     _create_aggregate directly in _create_aggregates. _create_aggregate is
#     ready for parallelism as it uses _return_if_subtest, however a change is
#     required to pass the aggr data of each child to the parent.
#  2. setup_controller_cluster is responsible for propagating aggr data to
#     Controller::setup, which will propagate it one more level to SanMgr::setup
#
###############################################################################

###############################################################################
# Method:    _setup_aggregates
# Objective: Method responsible for gathering all data corresponding to the
#            aggregates and calling the method responsible for creating them.
# Inputs:    A hash containing:
#            %args = (cserver => filer object
#                     aggregates => hash containing all aggr data
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_aggregates {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($aggregates, $cserver, $data) = undef;
    my (@aggrs, @aggr_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up aggregates"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing aggregate details
    $aggregates = &ParamsMgr::get_arg(-arg => "resource",
                                      -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    $cserver->get_aggregate_objs();
    # Get aggregate data from the resource data structure
    foreach my $key (keys (%{$aggregates})) {
        if ($key =~ /^AGGR\d*$/i) {

            # NOTE: Currently Params Mgr does not expand "looped" or "cloned"
            # elements. So we need to generate a list of copies. generate_copies
            # takes care of renaming things, indexes, prefixes, etc.

            push (@aggrs, $self->_generate_copies(element => $aggregates->{$key},
                                                  type    => "aggr"));
        }
    }

    # Now take all aggrs (copies and normal tags) and create them.
    # NOTE: If for some reason we want to enable parallelism here, it would just
    # be a matter of creating a subtest for each aggr.
    # For now, create them serially (our library does not manage disk objs)
    foreach my $aggr (@aggrs) {
        my $data = $self->_setup_aggregate(cserver   => $cserver,
                                           subtested => $SanMgr::FALSE,
                                           aggr      => $aggr);
        if ($self->error()) {
            # Error is already set, bubble it up
            return [];
        }

        # Collect all aggr data. This will be written to the parent process
        push(@aggr_data, $data);
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up aggregates"));

    return \@aggr_data;
} ## end sub _setup_aggregates

###############################################################################
# Method:    _setup_aggregate
# Objective: Method responsible for creating or reusing aggregate objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     aggr_data => hash containing all args for an aggr
#                     subtested => 1|0 indicating if this is a subtest.)
# Returns:   A hash containing name, reuse flag, cleanup and type of obj.
###############################################################################
sub _setup_aggregate {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($aggr, $cserver, $aggr_obj, $data) = undef;
    my ($state, $owner, $num_disks, $disk_type, $disklist, $name) = "";
    my ($cleanup_policy, $reuse_policy) = "";
    my (@disk_objs, %disk_args, %aggr_args) = ();
    my ($subtested, $aggr_exists, $reused, $use_disk_objs, $wafliron) =  $SanMgr::FALSE;

    # CONSTANTS
    my $DEFAULT_NUM_DISKS  = 5;
    my $DEFAULT_DISK_STATE = "spare";

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Hash containing objects params
    $aggr = &ParamsMgr::get_arg(-arg => "aggr",
                                -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    # Is this object a child process?
    $subtested = &ParamsMgr::get_arg(-arg => "subtested",
                                     -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return $self->_return_if_subtest(subtested => $subtested,
                                         ret_value => {},
                                         status    => "fail",
                                         msg       => $self->errmsg());
    }

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.
    if (defined($aggr->{NAME})) {
        $name = $aggr->{NAME};
    } else {
        $self->_set_err("Missing aggregate name. Check your XML or params.");
        return $self->_return_if_subtest(subtested => $subtested,
                                         ret_value => {},
                                         status    => "fail",
                                         msg       => $self->errmsg());
    }

    $self->sanlog(-msg => $self->_format_log("Setting up aggregate $name"));

    # Get common things like policies and normalize params. NOTE, this does not
    # set an error flag.
    %aggr_args = $self->_process_common_args(%$aggr);
    $cleanup_policy = delete $aggr_args{cleanup_policy};
    $reuse_policy   = delete $aggr_args{reuse_policy};
    $use_disk_objs  = delete $aggr_args{use_disk_objs};
    $wafliron       = delete $aggr_args{wafliron};

    # Need the state of the disks
    if (defined($aggr_args{state})) {
        $state = delete $aggr_args{state};
    } else {
        # set default to "spare"
        $state = $DEFAULT_DISK_STATE;
    }

    # Need the owner of the disk
    if (defined($aggr_args{disk_owner})) {
        $owner = $aggr_args{disk_owner};
    } else {
        # set default to cserver object
        $owner = $cserver->get_name();
        if ($cserver->error()) {
            $self->_set_err("Unable to get cserver name", $cserver->errmsg());
            return $self->_return_if_subtest(subtested => $subtested,
                                             ret_value => {},
                                             status    => "fail",
                                             msg       => $self->errmsg());
        }
    }

    # Finally, get the number of disks required to create this aggr
    # disklist and (the diskcount and type and nodes/disk_owner) are
    # mutually exclusive
    if (defined($aggr_args{disklist})) {
        $disklist = delete $aggr_args{disklist};
        delete $aggr_args{disk_owner};
        %disk_args = (disklist => $disklist);
    } else {
        if (defined($aggr_args{num_disks})) {
            $num_disks = delete $aggr_args{num_disks};
        } else {
            # set default to 5 disks: this is now mandatory since DTD is IMPLIED
            $self->sanlog(-msg => $self->_format_log("Param num_disks was NOT "
                                        . "specified defaulting to: "
                                        . "$DEFAULT_NUM_DISKS disks for "
                                        . "aggr $name."));
            $num_disks = $DEFAULT_NUM_DISKS;
        }
        %disk_args = (%disk_args, count => $num_disks);

        if (defined($aggr_args{disk_type})) {
            $disk_type = $aggr_args{disk_type};
            %disk_args = (%disk_args, type => $disk_type);
        }
    }

    # Build args hash based on user options to get disk objects
    %disk_args = (%disk_args,
                  state => $state,
                  owner => $owner);

    # Remove extra params. Otherwise the libraries will fail.
    foreach my $key (keys(%aggr_args)) {
        # NOTE: There could be more, currently, check nodes and vservers
        if ($key =~ /NODE|VSERVER|PARALLEL|SERIAL/i) {
            delete $aggr_args{$key};
        }
    }

    # NOTE: calling get_aggregates to refresh objects. For some reason
    # the lib is not refreshing the filer obj
    $cserver->refresh();
    if ($cserver->error()) {
        $self->_set_err("Unable to refresh cserver obj", $cserver->errmsg());
        return $self->_return_if_subtest(subtested => $subtested,
                                         ret_value => {},
                                         status    => "fail",
                                         msg       => $self->errmsg());
    }

    # Check if the aggr already exists.
    $aggr_exists = $cserver->aggregate_exist($name);
    if ($cserver->error()) {
        $self->_set_err("Unable to check if aggr exists", $cserver->errmsg());
        return $self->_return_if_subtest(subtested => $subtested,
                                         ret_value => {},
                                         status    => "fail",
                                         msg       => $self->errmsg());
    }

    if ($aggr_exists) {
        if ($reuse_policy =~ /^reuse$/i) {
            $aggr_obj = $cserver->get_aggregate_obj(aggregate => $name);
            if ($cserver->error()) {
                $self->_set_err("Unable to get aggregate object", $cserver->errmsg());
                return $self->_return_if_subtest(subtested => $subtested,
                                                 ret_value => {},
                                                 status    => "fail",
                                                 msg       => $self->errmsg());
            }
            $self->sanlog(-msg => $self->_format_log("Using existing aggregate $name"));
            $reused = $SanMgr::TRUE;

            # Check to see that the aggr is online
            my $state = $aggr_obj->get_state();
            if ($aggr_obj->error()) {
                $self->_set_err("Unable to get aggr $name state",
                                $aggr_obj->errmsg());
                return $self->_return_if_subtest(subtested => $subtested,
                                                 ret_value => {},
                                                 status    => "fail",
                                                 msg       => $self->errmsg());
            }

            if ($state =~ /^offline$/) {
                $self->sanlog(-msg => $self->_format_log(
                       "Aggregate $name is offline, changing state to online"));
                $aggr_obj->modify(-state => "online");
                if ($aggr_obj->error()) {
                    $self->_set_err("Unable to online aggr $name",
                                    $aggr_obj->errmsg());
                    return $self->_return_if_subtest(subtested => $subtested,
                                                     ret_value => {},
                                                     status    => "fail",
                                                     msg       => $self->errmsg());
                }
            }

            # if requested a specific max raid size, modify the field on the existing aggr
            if (defined($aggr_args{maxraidsize})) {
                $self->sanlog(-msg => $self->_format_log(
                              "Changing maxraidsize of aggregate $name to " . $aggr_args{maxraidsize}));
                $aggr_obj->modify(-maxraidsize => $aggr_args{maxraidsize});
                if ($aggr_obj->error()) {
                    $self->_set_err("Unable to modify maxraidsize for aggr $name",
                                    $aggr_obj->errmsg());
                    return $self->_return_if_subtest(subtested => $subtested,
                                                     ret_value => {},
                                                     status    => "fail",
                                                     msg       => $self->errmsg());
                }
            }

        } elsif ($reuse_policy =~ /^check$/i) {
            # TODO
            $reused = $SanMgr::FALSE;
        } elsif ($reuse_policy =~ /^fail$/i) {
            $reused = $SanMgr::FALSE;
            $self->_set_err("Aggr exists and reuse policy is set to \"fail\"");
            return $self->_return_if_subtest(subtested => $subtested,
                                             ret_value => {},
                                             status    => "fail",
                                             msg       => $self->errmsg());
        }
    } else {
        $reused = $SanMgr::FALSE;

        # use disk objs or discount
        if ($use_disk_objs) {
            # Get the disk objects
            @disk_objs = $cserver->get_disk_objs(%disk_args);
            if ($cserver->error() or (scalar(@disk_objs) == 0)) {
                $self->_set_err("Failed to get disks for this aggr.",
                                $cserver->errmsg());
                return $self->_return_if_subtest(subtested => $subtested,
                                                 ret_value => {},
                                                 status    => "fail",
                                                 msg       => $self->errmsg());
            }
            # push disk objects into %aggr_args
            %aggr_args = (%aggr_args, disk_objs => \@disk_objs);
        } else {
            if (length($disklist)) {
                %aggr_args = (%aggr_args, disklist => $disklist);
            } else {
                %aggr_args = (%aggr_args, diskcount => $num_disks);
            }
        }

        $self->sanlog(-msg => $self->_format_log("Creating aggregate $name"));
        $aggr_obj = $cserver->create_aggregate_obj(%aggr_args);
        if ($cserver->error()) {
            $self->_set_err("Unable to create aggregate object", $cserver->errmsg());
            return $self->_return_if_subtest(subtested => $subtested,
                                             ret_value => {},
                                             status    => "fail",
                                             msg       => $self->errmsg());
        }
    }

    $data = {
        reused => $reused,
        name => $name,
        type => "aggregate",
        subtested => $subtested,
        wafliron => $wafliron,
        cleanup => $cleanup_policy
    };

    $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                           . "for aggregate $name"
                                           . " (reused = $reused)"));

    return $self->_return_if_subtest(subtested => $subtested,
                                     ret_value => $data,
                                     status    => "pass",
                                     msg       => "Aggregate setup complete");
} ## end sub _setup_aggregate

###############################################################################
#                       I F G R P    M E T H O D S
#
###############################################################################
###############################################################################
# Method:    _setup_ifgrps
# Objective: Method for setting up all ifgrps
# Inputs:    A hash containing:
#            %args = (a bunch of params from the ifgrp xml tags)
# Returns:   An hash reference of ifgrp data
###############################################################################
sub _setup_ifgrps {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($ifgrps, $cserver, $data) = undef;
    my (@ifgrp_tag_data, @ifgrp_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up ifgrps"));

    # Process ifgrp and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing ifgrp details
    $ifgrps = &ParamsMgr::get_arg(-arg => "resource",
                                  -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Get ifgrp data from the resource data structure
    foreach my $key (keys (%{$ifgrps})) {
        if ($key =~ /^IFGRP\d*$/i) {
            # NOTE: Currently Params Mgr does not expand "looped" or "cloned"
            # elements. So we need to generate a list of copies. generate_copies
            # takes care of renaming things, indexes, prefixes, etc.

            push (@ifgrp_tag_data, $self->_generate_copies(element => $ifgrps->{$key},
                                                           type    => "ifgrp"));
        }
    }

    foreach my $ifgrp (@ifgrp_tag_data) {
        my $data = $self->_setup_ifgrp(cserver   => $cserver,
                                       subtested => $SanMgr::FALSE,
                                       ifgrp     => $ifgrp);
        if ($self->error()) {
            # Error is already set, bubble it up
            return [];
        }

        # Collect all ifgrp data. This will be written to the parent process
        push(@ifgrp_data, $data);
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up ifgrps"));
    return \@ifgrp_data;
}
###############################################################################
# Method:    _setup_ifgrp
# Objective: Method for setting up all one ifgrp
# Inputs:    A hash containing:
#            %args = (a bunch of ifrgp params)
# Returns:   An hash reference of ifgrp data
###############################################################################
sub _setup_ifgrp {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($ifgrp, $cserver, $data) = undef;
    my (@ifgrp_tag_data, @ifgrp_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up ifgrp"));

    # Process ifgrp and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing ifgrp details
    $ifgrp = &ParamsMgr::get_arg(-arg => "ifgrp",
                                  -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    my $cluster_name = $cserver->get_name();
    # FIRST, get the nacl cluster object
    # the $cluster_name could either be a node or a cluster name
    my ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                'names' => [$cluster_name],
                                                'obj_class' => 'nacl');
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
        return [];
    }

    try {
        my $NetworkPortIfgrp_object = NACL::C::NetworkPortIfgrp->create(
                                    command_interface => $nacl_cluster_obj,
                                    ifgrp             => $ifgrp->{NAME},
                                    node              => $ifgrp->{NODE},
                                    mode              => $ifgrp->{MODE},
                                    'distr-func'      => $ifgrp->{'DISTR-FUNC'});
    } otherwise {
        my $exception = shift;
        if  (!($exception =~ /already exist/i)) {
            $self->_set_err("Unable to create ifgrp " . $ifgrp->{NAME} . " "
                        . $exception->text());
        }
    };

    my @ports = split(/,/,$ifgrp->{'PORT-LIST'});

    foreach my $port (@ports) {
        try {
            NACL::C::NetworkPortIfgrp->add_port(
                                        command_interface => $nacl_cluster_obj,
                                        ifgrp             => $ifgrp->{NAME},
                                        node              => $ifgrp->{NODE},
                                        port              => $port);
        } otherwise {
            my $exception = shift;
            unless ($exception =~ /already a member/i) {
                $self->_set_err("Unable to add port $port to  " . $ifgrp->{NAME} . " "
                        . "on node " . $ifgrp->{NODE}  . $exception->text());
            }
        };
    }

    if ($self->error) {
        return $SanMgr::FALSE;
    }
    $data = {
        name         => $ifgrp->{NAME},
        type         => "ifgrp",
        node         => $ifgrp->{NODE},
        cluster_name => $cluster_name
    };

    return $data;
}
###############################################################################
#                        V L A N   M E T H O D S
#
#
###############################################################################
###############################################################################
# Method:    _setup_vlans
# Objective: Method for setting up all vlans
# Inputs:    A hash containing:
#            %args = (a bunch of params from the vlan xml tags)
# Returns:   An hash reference of vlan data
###############################################################################
sub _setup_vlans {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vlans, $cserver, $data) = undef;
    my (@vlan_tag_data, @vlan_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up vlans"));

    # Process vlan and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing vlan details
    $vlans = &ParamsMgr::get_arg(-arg => "resource",
                                      -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # Get vlan data from the resource data structure
    foreach my $key (keys (%{$vlans})) {
        if ($key =~ /^VLAN\d*$/i) {

            # NOTE: Currently Params Mgr does not expand "looped" or "cloned"
            # elements. So we need to generate a list of copies. generate_copies
            # takes care of renaming things, indexes, prefixes, etc.

            push (@vlan_tag_data, $self->_generate_copies(element => $vlans->{$key},
                                                          type    => "vlan"));
        }
    }

    foreach my $vlan (@vlan_tag_data) {
        my $data = $self->_setup_vlan(cserver   => $cserver,
                                      subtested => $SanMgr::FALSE,
                                      vlan      => $vlan);
        if ($self->error()) {
            # Error is already set, bubble it up
            return [];
        }

        # Collect all vlan data. This will be written to the parent process
        push(@vlan_data, $data);
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up vlans"));
    return \@vlan_data;
}
###############################################################################
# Method:    _setup_vlan
# Objective: Method for setting up one vlan
# Inputs:    A hash containing:
#            %args = (a bunch of vlan params)
# Returns:   An hash reference of vlan data
###############################################################################
sub _setup_vlan {

    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vlan, $cserver, $data) = undef;
    my (@vlan_tag_data, @vlan_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up vlan"));

    # Process vlan and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing vlan details
    $vlan = &ParamsMgr::get_arg(-arg => "vlan",
                                -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    my $cluster_name = $cserver->get_name();
    # FIRST, get the nacl cluster object
    # the $cluster_name could either be a node or a cluster name
    my ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                'names' => [$cluster_name],
                                                'obj_class' => 'nacl');
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
        &SubIPC::fail(-object => $self);
        return [];
    }
    # vlan-name doesn't function in the current code.
    # as soon as I get a working build i'll add it back.
    try {
        my $NetworkPortVlan_object = NACL::C::NetworkPortVlan->create(
                                    command_interface => $nacl_cluster_obj,
                                    'vlan-id'         => $vlan->{'VLAN-ID'},
                                    node              => $vlan->{NODE},
                                    port              => $vlan->{PORT});
                                    #'vlan-name'       => $vlan->{'VLAN-NAME'});
    } otherwise {
        my $exception = shift;
        if ( !($exception->text() =~ /duplicate entry/i)) {
            $self->_set_err("Unable to create vlan " . $vlan->{vlan} . " "
                        . $exception->text());
        }
    };
    if ($self->error) {
        return $SanMgr::FALSE;
    }
    $data = {
        'vlan-id'    => $vlan->{'VLAN-ID'},
        port         => $vlan->{PORT},
        node         => $vlan->{NODE},
        type         => "vlan",
        cluster_name => $cluster_name
    };

    return $data;
}

###############################################################################
#                       V S E R V E R    M E T H O D S
#
#  Things to consider:
#  1. Vservers are created in parallel unless users set a "parallel" attribute
#     in their XML or configuration file.
#  2. setup_controller_cluster is responsible for propagating vserver data to
#     Controller::setup, which will propagate it one more level to SanMgr
#  3. _create_vserver creates vserver objects first, then protocol services
#     and then sets up its contents. Then it gets data from each internal object
#     and bubbles it up to Controller::setup.
#
###############################################################################

###############################################################################
# Method:    _setup_vservers
# Objective: Method responsible for creating or reusing vserver objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     vserver_data => hash containing all args for a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_vservers {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $resource, $vs_data_hash) = undef;
    my ($aggr_name, $cserver_name) = "";
    my (@vservers, @vserver_data, @subtests, @messages, @node_objs) = ();
    my ($parallel, $st_status) = $SanMgr::TRUE;

    $self->sanlog(-msg => $self->_format_log("Start setting up vservers"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing vserver details
    $resource = &ParamsMgr::get_arg(-arg => "resource",
                                    -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg => "cserver",
                                   -required => $SanMgr::TRUE);

    @node_objs = &ParamsMgr::get_arg(-arg => "node_objs",
                                     -ref_type => "ARRAY",
                                     -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a vserver
    $cserver_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err("Could not get cserver name", $cserver->errmsg());
        return [];
    }

    # Get all vservers and append their corresponding aggr name.
    # This will be needed in order to create them
    foreach my $vs_key (keys (%{$resource})) {
        if ($vs_key =~ /^VSERVER\d*$/i) {
            $resource->{$vs_key}->{FILER} = $cserver_name;
            push(@vservers, $self->_generate_copies(element => $resource->{$vs_key},
                                                    type    => "vserver"));
        }
    }

    my $global_iscsi_options = &ParamsMgr::search(search_path => [],
                                                  key => "ENVIRONMENT",
                                                  key_filter => qr/^GLOBAL_ISCSI_OPTIONS$/,
                                                  paramset => $self->EnGenX::Util::get_paramset());

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return [];
    }

    # parse out the parameter in the global iscsi options tag
    my $global_isns_address = $global_iscsi_options->{GLOBAL_ISCSI_OPTIONS}->{ISNS_SERVER};

    # create all the vservers sequentially because of burt481837
    # parallel vserver creates are not supported in RR.0
    # once it is supported, we can push the create_vserver()
    # calls into the _setup_vserver() method.
    my $i = 0;

    foreach my $vserver (@vservers) {
        my $vs_name = $vserver->{NAME};

        # default the node to use to provision this vserver to be the cserver
        my $node_obj = $cserver;

        # if requested to do load balanced provisioning, then round robin select one
        if ($resource->{LOAD_BALANCE_PROVISIONING} == 1) {
            $node_obj = $node_objs[$i % scalar(@node_objs)];
        }

        # if global isns server exists and is not overridden by a vserver
        # specific setting, apply it to the vserver args here
        if (defined $global_isns_address) {
            if (not ($global_isns_address eq "") and
                not defined ($vserver->{ISNS_SERVICE}->{'ADDRESS'})
                and $vserver->{ISCSI_ENABLED}) {

                $vserver->{ISNS_SERVICE}->{'ADDRESS'} =
                    $global_isns_address;
            }
        }

        # create this vserver and push its data onto the vserver data
        # array that will get written to the parent. Use the cserver/node object to perform the provisioning
        my $vs_data =  $self->_create_vserver(-vserver => $vserver,
                                              -cserver => $node_obj);
        if ($self->error()) {
            $self->_set_err("Unable to create vserver '$vs_name'");
            return [];
        }

        push(@vserver_data, $vs_data);
        $vs_data_hash->{$vs_name} = $vs_data;
        $i++;
    }

    # reset the counter for the next loop
    $i = 0;

    # Go through all vservers and provision all their objects. By default, SanMgr will create
    # them in parallel.
    foreach my $vserver (@vservers) {
        # Check if users prefer to create this vserver serially
        if (defined($vserver->{PARALLEL})) {
            $parallel = delete $vserver->{PARALLEL};
        } else {
            $parallel = $SanMgr::TRUE;
        }

        # if the user does not want any parallelism on this current vserver, reap any existing
        # setup subtests if they exist
        if ($parallel eq 0 and scalar(@subtests)) {
            push(@vserver_data, $self->_reap_vserver_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A vserver setup subtest failed");
                return [];
            }
            @subtests = ();
        }

        my $subtest = &SubIPC::create(
            -id  => "setup_vserver__" . $vserver->{NAME},
            -run =>
                sub {
                    # if requested to load balance the provision phase across all nodes, pick a node
                    # round robin from the list
                    if ($resource->{LOAD_BALANCE_PROVISIONING} == 1) {
                        $cserver = $node_objs[$i % scalar(@node_objs)];
                    }

                    $self->_setup_vserver(cserver => $cserver,
                                          cluster_name => $cserver_name,
                                          vserver => $vserver,
                                          vs_data => $vs_data_hash->{$vserver->{NAME}});

                    if ($self->error()) {
                        # Error is already set, bubble it up
                        return [];
                    }
                },
            -ipc => $SanMgr::TRUE,
            -bg  => $SanMgr::TRUE
        );

        if (&SubIPC::error()) {
            $self->_set_err("Unable to dispatch subtest", &SubIPC::get_error());
            return [];
        }
        push(@subtests, $subtest);

        $i++;

        # reap the subtest list
        # if the user doesnt want max prallelism and
        #### the user doesnt want any parallelism on this vserver
        #### or
        #### we have hit a max vserver parallel setup count
        if ($parallel !~ /^max$/i and ($parallel eq 0 or scalar(@subtests) == $parallel)) {
            push(@vserver_data, $self->_reap_vserver_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A vserver setup subtest failed");
                return [];
            }
            @subtests = ();
        }
    }

    # if there are still subtests that havn't been reaped, reap them all here:
    if (scalar(@subtests)) {
        push(@vserver_data, $self->_reap_vserver_setup_subtests(-subtests => \@subtests));
        if ($self->error()) {
            $self->_set_err("A vserver setup subtest failed");
            return [];
        }
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up vservers"));

    return \@vserver_data;
} ## end sub _setup_vservers

###############################################################################
# Method:    _reap_vserver_setup_subtests
# Objective: Method responsible for Reaping vserver setup subtests
# Inputs:    A hash containing:
#            %args = (subtest => an array of SubIPC subtests)
# Returns:   An array or array reference of vserver data return by all the
#            subtests
###############################################################################
sub _reap_vserver_setup_subtests {
    my ($self, %args) = @_;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    my @subtests = &ParamsMgr::get_arg(-arg => "subtests",
                                       -ref_type => "ARRAY",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    # Need to wait for ALL vservers to complete.
    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("Unable to wait on subtests to finish", &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @messages = &SubIPC::get_messages(-type => "setup_vserver_complete");
    if (&SubIPC::error()) {
        $self->_set_err("Unable to get 'setup_vserver_complete' messages from children", &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @vserver_data;

    foreach my $message (@messages) {
        # We are interested only in the object data.
        if (defined($message->{extra}->{obj_data})) {
            foreach my $obj_data (@{$message->{extra}->{obj_data}}) {
                push(@vserver_data, $obj_data);

                # Log the entry we are adding to the cleanup stack
                my $dumpout = Dumper($obj_data);
                $self->sanlog(-msg => "Dumper of data getting added to vserver "
                                    . "cleanup stack:\n" . $dumpout,
                              -msg_level  => 7);
            }
        }
    }

    # Get the status of a subtest and report it
    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != $SanMgr::FALSE) {
        $self->sanlog(-msg => "Found a vserver subtest failure:" .
                              "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->_set_err("Failed to setup vservers");
        return (wantarray()) ? () : [];
    } else {
        $self->sanlog(-msg => "All vserver subtests have passed:" .
                              "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->sanlog(-msg => $self->_format_log("Finished vserver setup"));
    }

    # Do some house keeping and remove all vserver subtests.
    &SubIPC::remove_subtest(subtest => \@subtests) if (scalar(@subtests));

    return (wantarray()) ? @vserver_data : \@vserver_data;
} ## end sub _reap_vserver_setup_subtests

###############################################################################
# Method:    _setup_vserver
# Objective: Method responsible for creating or reusing vserver objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     vserver_data => hash containing all args for a vserver
#                     subtested => 1|0 indicating if this is a subtest.)
# Returns:   An array of hashes containing vserver data.
###############################################################################
sub _setup_vserver {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($vserver, $cserver, $vserver_obj, $data)    = undef;
    my ($protocol_content, $vserver_content)        = undef;
    my ($protocol_count, $vserver_qos_content)      = undef;
    my ($qos_policy_group)                          = undef;
    my ($name, $aggregate, $cluster_name)           = "";
    my ($cleanup_policy, $reuse_policy)             = "";
    my (%vserver_args, %create_args, @vserver_data) = ();
    my ($vserver_exists, $reused)                   = $SanMgr::FALSE;
    my ($fcp_enabled, $iscsi_enabled)               = $SanMgr::FALSE;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Hash containing objects params
    $vserver = &ParamsMgr::get_arg(-arg      => "vserver",
                                   -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Cluster name
    $cluster_name = &ParamsMgr::get_arg(-arg      => "cluster_name",
                                        -required => $SanMgr::TRUE);

    # This is the data obtained from _create_vserver()
    $data = &ParamsMgr::get_arg(-arg      => "vs_data",
                                -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # obtain the cserver name
    my $cserver_name = $cserver->get_name();

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.
    if (defined($vserver->{NAME})) {
        $name = delete $vserver->{NAME};
        $vserver->{VSERVER} = $name;
    } else {
        $self->_set_err("Missing vserver name. Check your XML or params.");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    $self->sanlog(-msg => $self->_format_log("Setting up vserver $name"));
    if (defined($vserver->{AGGREGATE})) {
        $aggregate = $vserver->{AGGREGATE};
    } elsif ($vserver->{AGGR}){
        $vserver->{AGGREGATE} = delete $vserver->{AGGR};
        $aggregate = $vserver->{AGGREGATE};
    } else {
        $self->_set_err("Missing aggregate name. Check your XML or params.");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Normalize params and get local policies
    %vserver_args = $self->_process_common_args(%$vserver);

    $cleanup_policy = delete $vserver_args{cleanup_policy};
    $reuse_policy   = delete $vserver_args{reuse_policy};
    $iscsi_enabled  = delete $vserver_args{iscsi_enabled};
    $fcp_enabled    = delete $vserver_args{fcp_enabled};
    $protocol_count = delete $vserver_args{protocol_count};
    delete $vserver_args{load_balance_provisioning};
    delete $vserver_args{vs_type};

    # In order to create the vserver obj, we want to create a hash,
    # create_args, that contains only the args that create will
    # use.  Otherwise the vserver create will fail because of the
    # extra arguments.
    foreach my $key (keys(%vserver_args)) {
        if ($key =~ /PARALLEL|SERIAL/i) {
            delete $vserver_args{$key};
        } elsif ($key =~ /LIF|VOL|IGROUP|PORTSET|TPGROUP|FCP_SERVICE|ISCSI_SERVICE/i) {
            # These are used by SanMgr but not by ControllerCluster
            next;
        } elsif ($key =~ /comment\d/i) {
            delete $vserver_args{$key};
        } elsif ($key =~ /qos-policy-group|use_default_qos_policy_group/i) {
            # This is not supported in a vserver create command
            next;
        } else {
            $create_args{$key} = $vserver_args{$key};
        }
    }

    # Need to use a new connection on each vserver
    $cserver->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
    if ($cserver->error()) {
        $self->_set_err("Unable to reconnect to cserver", $cserver->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    ($vserver_obj) = $cserver->fetch_vserver_objs(-vserver => $name);
    if ($cserver->error() or not defined($vserver_obj)) {
        $self->_set_err("Unable to fetch vserver obj", $cserver->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Now set protocol services on the vserver
    $protocol_content = $self->_setup_protocol_services(
                                    vserver_obj    => $vserver_obj,
                                    vserver_args   => $vserver,
                                    cserver        => $cserver,
                                    cluster_name   => $cluster_name,
                                    iscsi_enabled  => $iscsi_enabled,
                                    fcp_enabled    => $fcp_enabled,
                                    protocol_count => $protocol_count,
                                    reuse_policy   => $reuse_policy,
                                    cleanup_policy => $cleanup_policy);

    # Save the cleanup data returned.  We do this before the error check
    # so that we can return the array on error and get it cleaned up later
    # NOTE: Notice the use of @{} to keep a one-dimension list
    push(@vserver_data, @{$protocol_content});

    if ($self->error()) {
        $self->_set_err("Unable to setup protocol services for vserver");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # FIRST, get the nacl cluster object
    # the $cserver_name could either be a node or a cluster name
    my ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                'names' => [$cserver_name],
                                                'obj_class' => 'nacl');
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain nacl cluster object for " . $cserver_name);
        &SubIPC::fail(-object => $self);
        return [];
    }

    # See if the release is SN or after because qos is only supported in SN or after.
    my $release_SN_or_after;
    try {
        $release_SN_or_after = $nacl_cluster_obj->version_manager()->is_version_ge(release => '8.2');
    } otherwise {
        my $e = shift;
        $self->_set_err("Failed to get whether the release is SN or after", $e->text());
    };
    if ($self->error()) {
        return $SanMgr::FALSE;
    }

    if ($release_SN_or_after) {
        # Setup QOS policy groups for the vserver
        $vserver_qos_content = $self->_setup_qos_policy_groups(
                                                    vserver_obj  => $vserver_obj,
                                                    vserver_args => $vserver,
                                                    vserver_data => $data,
                                                    nacl_cluster_obj => $nacl_cluster_obj);

        # Save the cleanup data returned.  We do this before the error check
        # so that we can return the array on error and get it cleaned up later
        # NOTE: Notice the use of @{} to keep a one-dimension list
        push(@vserver_data, @{$vserver_qos_content});

        if ($self->error()) {
            $self->_set_err("Unable to setup vserver qos policy groups");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        try {
            if ($vserver->{'USE_DEFAULT_QOS_POLICY_GROUP'}) {
                NACL::C::Vserver->modify(
                                    'command_interface' => $nacl_cluster_obj,
                                    'vserver'           => $name,
                                    'qos-policy-group'  => $name . "_pg");
            } elsif ($vserver->{'QOS-POLICY-GROUP'}) {
                NACL::C::Vserver->modify(
                                    'command_interface' => $nacl_cluster_obj,
                                    'vserver'           => $name,
                                    'qos-policy-group'  => $vserver->{'QOS-POLICY-GROUP'});
            }
        } otherwise {
            my $exception = shift;
            $self->_set_err("Unable to modify vserver qos policy group assignment"
                            . $exception->text());
        };
        if ($self->error) {
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Setup vserver content (vols, luns, lifs, etc).
    $vserver_content = $self->_setup_vserver_content(vserver_obj      => $vserver_obj,
                                                     vserver_args     => $vserver,
                                                     aggregate        => $aggregate,
                                                     vserver_data     => $data,
                                                     cserver          => $cserver,
                                                     nacl_cluster_obj => $nacl_cluster_obj);
    if ($self->error()) {
        $self->_set_err("Unable to setup vserver content");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Extract iscsi/fcp targets from the object and append it to a data
    # structure in order to bubble this data to the parent object.
    if (defined($vserver_obj->{iscsi_lif_data})) {
        $data->{iscsi_lif_data} = $vserver_obj->{iscsi_lif_data};
    }

    if (defined($vserver_obj->{fcp_lif_data})) {
        $data->{fcp_lif_data} = $vserver_obj->{fcp_lif_data};
    }

    # Store the vserver content returned from setup
    # NOTE: Notice the use of @{} to keep a one-dimension list
    push(@vserver_data, @{$vserver_content});

    # store the cserver used to create all the vserver content, needed for cleanup
    foreach my $vs_data (@vserver_data) {
        $vs_data->{cserver} = $cserver_name;
    }

    # disconnect the cserver connection
    $cserver->disconnect_conn();

    return $self->_return_if_subtest(
                              subtested      => $SanMgr::TRUE,
                              status         => "pass",
                              ret_value      => \@vserver_data,
                              write_parent   => $SanMgr::TRUE,
                              write_msg_type => "setup_vserver_complete",
                              msg            => "Vserver created successfully");
} ## end sub _setup_vserver

###############################################################################
# Method:    _create_vserver
# Objective: Create (or reuses) the specified vserver
# Input:     A hash containing a vserver object, vserver args, cserver obj,
#            reuse policy and cleanup policy.
# Returns:   Cleanup data.  Will _set_err flag on error.
###############################################################################
sub _create_vserver {
    my ($self, %args) = @_;

    $self->clear_error();

    my ($name);
    my ($reused) = $SanMgr::FALSE;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Hash containing objects params
    my $vserver = &ParamsMgr::get_arg(-arg      => "vserver",
                                      -required => $SanMgr::TRUE);
    # A cserver object (filer obj)
    my $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                      -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return {};
    }

    my %vserver_hash = %{$vserver};

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.
    if (defined($vserver_hash{NAME})) {
        $name = $vserver_hash{NAME};
        $vserver_hash{VSERVER} = $name;
    } else {
        $self->_set_err("Missing vserver name. Check your XML or params.");
        return {};
    }

    $self->sanlog(-msg => $self->_format_log("Setting up vserver $name"));
    my $aggregate;
    if (defined($vserver_hash{AGGREGATE})) {
        $aggregate = $vserver_hash{AGGREGATE};
    } elsif ($vserver_hash{AGGR}){
        $vserver_hash{AGGREGATE} = delete $vserver_hash{AGGR};
        $aggregate = $vserver_hash{AGGREGATE};
    } else {
        $self->_set_err("Missing aggregate name. Check your XML or params.");
        return {};
    }

    # Normalize params and get local policies
    my %vserver_args = $self->_process_common_args(%vserver_hash);

    my $cleanup_policy = delete $vserver_args{cleanup_policy};
    my $reuse_policy   = delete $vserver_args{reuse_policy};
    delete $vserver_args{iscsi_enabled};
    delete $vserver_args{fcp_enabled};
    delete $vserver_args{protocol_count};
    delete $vserver_args{vs_type};

    # In order to create the vserver obj, we want to create a hash,
    # create_args, that contains only the args that create will
    # use.  Otherwise the vserver create will fail because of the
    # extra arguments.
    my %create_args;
    foreach my $key (keys(%vserver_args)) {
        if ($key =~ /PARALLEL|SERIAL/i) {
            delete $vserver_args{$key};
        } elsif ($key =~ /LIF|VOL|IGROUP|PORTSET|TPGROUP|FCP_SERVICE|ISCSI_SERVICE|NAME|ISNS_SERVICE|QOS_POLICY_GROUP/i) {
            # These are used by SanMgr but not by ControllerCluster
            next;
        } elsif ($key =~ /comment\d/i) {
            delete $vserver_args{$key};
        } elsif ($key =~ /qos-policy-group|use_default_qos_policy_group/i) {
            # This is not supported in a vserver create command
            next;
        } else {
            $create_args{$key} = $vserver_args{$key};
        }
    }

    # Check if the vserver already exists.
    my $vserver_exists = $cserver->vserver_exist(vserver => $name);
    if ($cserver->error()) {
        $self->_set_err("Unable to determine if vserver $name exists", $cserver->errmsg());
        return {};
    }

    if ($vserver_exists) {
        # reuse the vserver if it exists
        if ($reuse_policy =~ /^reuse$/i) {
            my $vserver_obj = $cserver->get_vserver_obj(vserver => $name);
            if ($cserver->error()) {
                $self->_set_err("Failed to get vserver $name", $cserver->errmsg());
                return {};
            }
            $self->sanlog(-msg => $self->_format_log("Using existing vserver $name"));
            $reused = $SanMgr::TRUE;
        } elsif ($reuse_policy =~ /^check$/i) {
            $reused = $SanMgr::FALSE;
            # TODO
        } elsif ($reuse_policy =~ /^fail$/i) {
            $reused = $SanMgr::FALSE;
            $self->_set_err("Vserver $name exists, and reuse policy is set to \"fail\"");
            return {};
        }
    } else {
        # create the vserver since it doesnt exist
        $self->sanlog(-msg => $self->_format_log("Creating vserver $name"));
        my $vserver_obj = $cserver->create_vserver_obj(%create_args);
        if ($cserver->error()) {
            $self->_set_err("SanMgr Failed to create vserver: "
                           . $vserver_args{vserver}
                           , $cserver->errmsg());
            return {};
        }
        $reused = $SanMgr::FALSE;
    }

    # build the vserver cleanup setup
    my $data = {
        reused    => $reused,
        name      => $name,
        aggregate => $aggregate,
        type      => "vserver",
        parallel  => $vserver->{PARALLEL},
        cleanup   => $cleanup_policy,
        cserver   => $cserver->get_name()
    };

    $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                           . "for Vserver $name"
                                           . " (reused = $reused)"));

    return $data;
} ## end _create_vserver

###############################################################################
# Method:    _setup_protocol_services
# Objective: Creates fcp or iscsi services on a given vserver.
# Input:     A hash containing a vserver object, vserver args, cserver obj,
#            reuse policy and cleanup policy.
# Returns:   Cleanup data.  Will _set_err flag on error.
###############################################################################
sub _setup_protocol_services {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local variables
    my ($vs_obj, $vs_args, $iscsi_obj, $fcp_obj) = undef;  # Refs
    my ($vs_name, $cluster_name) = "";                     # Scalars
    my (@cleanup_data) = ();                               # Array of Hash
    my ($fcp_enabled, $iscsi_enabled) = $SanMgr::FALSE;    # Booleans
    my ($iscsi_cleanup_hash, $fcp_cleanup_hash) = {};      # Cleanup Data
    my ($cserver, $protocol_count)              = undef;
    my ($cleanup_policy, $reuse_policy);
    my ($iscsi_status_admin, $fcp_status_admin) = undef;

    # Same story, process args
    &ParamsMgr::process_args(%args);

    # Vserver object
    $vs_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                  -required => $SanMgr::TRUE);

    # vserver params/attributes
    $vs_args = &ParamsMgr::get_arg(-arg      => "vserver_args",
                                   -required => $SanMgr::TRUE);

    # A cserver (filer object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # cluster name
    $cluster_name = &ParamsMgr::get_arg(-arg      => "cluster_name",
                                        -required => $SanMgr::TRUE);

    # Iscsi/FCP enabled
    $iscsi_enabled = &ParamsMgr::get_arg(-arg       => "iscsi_enabled",
                                         -required  => $SanMgr::FALSE,
                                         -default   => $SanMgr::FALSE);

    $fcp_enabled   = &ParamsMgr::get_arg(-arg       => "fcp_enabled",
                                         -required  => $SanMgr::FALSE,
                                         -default   => $SanMgr::FALSE);

    # Protocol Count is set if the vserver count was > 1
    $protocol_count = &ParamsMgr::get_arg(-arg      => "protocol_count",
                                          -required => $SanMgr::FALSE,
                                          -default  => "-1");

    # Cleanup Policy
    $cleanup_policy = &ParamsMgr::get_arg(-arg      => "cleanup_policy",
                                          -required => $SanMgr::TRUE);

    # Reuse Policy
    $reuse_policy   = &ParamsMgr::get_arg(-arg      => "reuse_policy",
                                          -required => $SanMgr::TRUE);

    # Check to make sure all the arguments were properly processed
    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get vserver name
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to parse vserver name", $vs_obj->errmsg());
        return [];
    }

    $self->sanlog(-msg => $self->_format_log("Start setting up protocol "
                                           . "services on vserver $vs_name"));

    # NOTE: This might need to be changed once dev enables service licensing

    # If a user specifies iscsi_enabled, turn it on here.
    if ($iscsi_enabled) {

        # Check to see if the iscsi_service is already running.  If it is,
        # we just save a cleanup action and we are done.
        if ($vs_obj->iscsi_exist()) {

            # Check the reuse policy
            if ($reuse_policy =~ /^reuse$/i) {
                $self->sanlog(-msg => $self->_format_log("Using existing iscsi service"));
            } elsif ($reuse_policy =~ /^check$/i) {
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $self->_set_err("iscsi service exists and reuse policy is set to \"fail\"");
                return [];
            }

            # get the iscsi object
            $iscsi_obj = $vs_obj->get_iscsi_obj();
            if ($vs_obj->error()) {
                $self->_set_err("Failed to get iscsi obj from vserver", $vs_obj->errmsg());
                return [];
            }

            # Create the cleanup stack entry
            $iscsi_cleanup_hash = {
                            reused    => $SanMgr::TRUE,
                            vserver   => $vs_name,
                            type      => "iscsi_service",
                            cleanup   => $cleanup_policy,
                            };
        } else {
            # Check to see if the user used the iscsi_service tag to define
            # values to be passed to iscsi create
            my $setup_args;
            $setup_args->{"-vserver"} = $vs_name;

            # Check to see if the vserver args contain an iscsi_service tag
            if (defined $vs_args->{ISCSI_SERVICE}) {
                my $iscsi_data = $vs_args->{ISCSI_SERVICE};
                my $dumpout = Dumper($iscsi_data);
                $self->sanlog(-msg => "iscsi service tag given for "
                                . "vserver $vs_name:\n"
                                . $dumpout,
                          -msg_level  => 7);

                if (defined  $iscsi_data->{"TARGET-NAME"}) {
                    $setup_args->{"-target-name"} = $iscsi_data->{"TARGET-NAME"};

                    # If there are multiple vservers sharing this tag, add
                    # the protocol count as a suffix to the target name
                    # Note, protocol_count is set when we generage copies and
                    # stored in the vserver args.
                    if ($protocol_count >= 0) {
                        $setup_args->{"-target-name"} .= $protocol_count;
                    }
                }
                if (defined  $iscsi_data->{"TARGET-ALIAS"}) {
                    $setup_args->{"-target-alias"} = $iscsi_data->{"TARGET-ALIAS"};

                    # If there are multiple vservers sharing this tag, add
                    # the add the protocol count as a suffix to the target alias
                    if ($protocol_count >= 0) {
                        $setup_args->{"-target-alias"} .= "_" . $protocol_count;
                    }
                }
                if (defined  $iscsi_data->{"STATUS-ADMIN"}) {
                    $setup_args->{"-status-admin"} = $iscsi_data->{"STATUS-ADMIN"};
                    $iscsi_status_admin = $iscsi_data->{"STATUS-ADMIN"};
                }
                if (defined  $iscsi_data->{"MAX-ERROR-RECOVERY-LEVEL"}) {
                    $setup_args->{"-max-error-recovery-level"} = $iscsi_data->{"MAX-ERROR-RECOVERY-LEVEL"};
                }
                if (defined  $iscsi_data->{"RETAIN-TIMEOUT"}) {
                    $setup_args->{"-retain-timeout"} = $iscsi_data->{"RETAIN-TIMEOUT"};
                }
                if (defined  $iscsi_data->{"LOGIN-TIMEOUT"}) {
                    $setup_args->{"-login-timeout"} = $iscsi_data->{"LOGIN-TIMEOUT"};
                }
                if (defined  $iscsi_data->{"MAX-CONN-PER-SESSION"}) {
                    $setup_args->{"-max-conn-per-session"} = $iscsi_data->{"MAX-CONN-PER-SESSION"};
                }
                if (defined  $iscsi_data->{"MAX-IOS-PER-SESSION"}) {
                    $setup_args->{"-max-ios-per-session"} = $iscsi_data->{"MAX-IOS-PER-SESSION"};
                }
                if (defined  $iscsi_data->{"TCP-WINDOW-SIZE"}) {
                    $setup_args->{"-tcp-window-size"} = $iscsi_data->{"TCP-WINDOW-SIZE"};
                }
                if (defined  $iscsi_data->{"CLEANUP_POLICY"}) {
                    $cleanup_policy = delete $iscsi_data->{"CLEANUP_POLICY"};
                }

            } # end if iscsi_service tag found

            # Now that we have the setup args setup, do the actual iscsi create
            $self->sanlog(-msg => $self->_format_log("Creating iscsi service on "
                                                    ."vserver $vs_name"));
            $iscsi_obj = $vs_obj->create_iscsi_obj(%{$setup_args});
            if ($vs_obj->error()) {
                $self->_set_err($vs_obj->errmsg());
                return [];
            }

            # Create the cleanup stack entry
            $iscsi_cleanup_hash = {
                            vserver      => $vs_name,
                            reused       => $SanMgr::FALSE,
                            type         => "iscsi_service",
                            cleanup      => $cleanup_policy,
                            };
            if (defined ($iscsi_status_admin)) {
                $iscsi_cleanup_hash->{"status_admin"} = $iscsi_status_admin;
            }

        } # end of iscsi doesn't exist case

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($iscsi_cleanup_hash);
        $self->sanlog(-msg => "Dumper of iscsi service data getting added"
                            . " to vserver cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);

        # Push the iscsi setup data on the cleanup stack then setup FCP
        # and push the fcp cleanup data on that same stack and return it.
        push(@cleanup_data, $iscsi_cleanup_hash);
        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                                 . "for iscsi service on "
                                                 . "Vserver $vs_name"
                                                 . " (reused = "
                                                 . $iscsi_cleanup_hash->{reused}
                                                 . ")"));

        $self->sanlog(-msg => $self->_format_log("Start setting up iSCSI Security settings "
                                               . "on vserver $vs_name"));

        # setup iSCSI security for all initiators that have it enabled
        my @iscsi_clients = $self->{parent}->get_clients_associated_to_vserver(-cluster_name => $cluster_name,
                                                                               -vserver_names => [$vs_name],
                                                                               -protocols => ["iscsi", "mixed"]);
        if ($self->{parent}->error()) {
            $self->_set_err("Failed to get clients associated to vserver $vs_name", $self->{parent}->errmsg());
            $self->{parent}->_clear_err();
            return [];
        }

        # get the list of initiator security settings up front for this vserver
        my @iscsi_security_show = $iscsi_obj->initiator_auth_list_info();
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed to obtain current iSCSI security settings");
            return [];
        }

        foreach my $client (@iscsi_clients) {

            # if this client matches an iSCSI nodename or a FCP WWPN, then we won't be able to create a specific iSCSI setting for it
            # because we can't accurately map it to a real client. Assume that it will use the default vserver iSCSI security settings
            if ($client =~ /^(iqn\.\d+\-\d+\.\S*|[A-Fa-f0-9]{2}(?::?[A-Fa-f0-9]{2}){7})$/i) {
                next;
            }

            # set up iscsi security for this client if it is doing CHAP
            # get the SanClient object for this client
            my ($client_obj) = $self->get_objects(-types => ["client", "hypervisor"],
                                                  -names => [$client]);
            if ($self->error()) {
                $self->_set_err("Failed to get client object for $client");
                return [];
            }

            # obtain the chap settings from this client object
            my $chap_mode = $client_obj->{iscsi}->{CONFIG_ISCSI_CHAP_MODE};
            my $chap_inusername = $client_obj->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
            my $chap_outusername = $client_obj->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME};
            my $chap_inpassword = $client_obj->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD};
            my $chap_outpassword = $client_obj->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD};

            # fetch the initiator objects for this client, filter out non-iscsi initiators
            my @initiator_objs = $client_obj->get_initiators();
            @initiator_objs = grep { $_->get_protocol() =~ /iscsi/ } @initiator_objs;

            if (not scalar(@initiator_objs)) {
                $self->_set_err("There are no iSCSI initiators for the client '$client'");
                return [];
            }

            # pull out all the nodenames for this initiator
            my @nodenames = map { $_->get_nodename() } @initiator_objs;

            # remove dups from the nodename array
            my %nodename_hash;
            foreach (@nodenames) {
                $nodename_hash{$_}++;
            }
            @nodenames = keys %nodename_hash;

            # if chap_mode is none, then we have to make sure that chap is turned off for this initiator
            if ($chap_mode =~ /none/i) {
                foreach my $nodename (@nodenames) {
                    # remove security settings for this nodename if it exist
                    if (grep { $_->{initiator_name} =~ /^${nodename}$/ } @iscsi_security_show) {
                        # delete the initiator setting for iscsi security
                        $iscsi_obj->initiator_delete_auth('initiator-name' => $nodename);
                        if ($iscsi_obj->error()) {
                            $self->_set_err("Failed to remove the iscsi security setting for nodename '$nodename' on client '$client'",
                                            $iscsi_obj->errmsg());
                            return [];
                        }
                    } else {
                        # otherwise, there is no specific initiator setting, do nothing
                    }
                }
            } else {
                # create the iscsi security setting for this nodename if one does not exist for it yet
                my %modify_args;
                if ($chap_mode =~ /bidirectional/i) {
                    # bidirectional chap
                    %modify_args = (
                        'auth-type' => "CHAP",
                        'user-name' =>  $chap_inusername,
                        'password' => $chap_inpassword,
                        'outbound-password' => $chap_outpassword,
                        'outbound-user-name' => $chap_outusername);
                } else {
                    # unidirectional chap
                    %modify_args = (
                        'auth-type' => "CHAP",
                        'user-name' =>  $chap_inusername,
                        'password' => $chap_inpassword);
                }

                # setup chap for each nodename
                foreach my $nodename (@nodenames) {
                    my $iscsi_security_reused;

                    # default the in username as the nodename
                    if (not defined($chap_inusername) or
                        not length($chap_inusername)) {
                        $modify_args{'user-name'} = $nodename;
                    }
                    # if doing bidirectional, default the outbound user name to the nodename
                    if ($chap_mode =~ /bidirectional/i) {
                        if (not defined($chap_outusername) or
                            not length($chap_outusername)) {
                            $modify_args{'outbound-user-name'} = $nodename;
                        }
                    }

                    # if this nodename already has a security setting, modify it
                    if (grep { $_->{initiator_name} =~ /^${nodename}$/ } @iscsi_security_show) {
                        $modify_args{'clear-outbound'} = 'true' if ($chap_mode =~ /unidirectional/i);

                        # modify the existing security setting for this nodename if it already exists
                        $iscsi_obj->initiator_modify_chap_params('initiator-name' => $nodename,
                                                                 %modify_args);
                        $iscsi_security_reused = $SanMgr::TRUE;
                    } else {
                        # create an iscsi security setting for it
                        $iscsi_obj->initiator_create_auth('initiator-name' => $nodename,
                                                          %modify_args);
                        $iscsi_security_reused = $SanMgr::FALSE;
                    }
                    if ($iscsi_obj->error()) {
                        $self->_set_err("Failed to setup iscsi security settings for nodename '$nodename' on client '$client'",
                                        $iscsi_obj->errmsg());
                        return [];
                    }

                    # Create the cleanup stack entry for this iscsi security setting
                    my $iscsi_security_cleanup_hash = {
                        vserver      => $vs_name,
                        reused       => $iscsi_security_reused,
                        initiator    => $nodename,
                        type         => "iscsi_security_setting",
                        cleanup      => $cleanup_policy,
                    };
                    push(@cleanup_data, $iscsi_security_cleanup_hash);
                    $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                                           . "for iscsi security setting for initiator $nodename on "
                                                           . "Vserver $vs_name"
                                                           . " (reused = "
                                                           . $iscsi_security_cleanup_hash->{reused}
                                                           . ")"));

                }
            }
        }

    } # end for iscsi_enabled

    # Do the same thing for FCP that we did for ISCSI
    if ($fcp_enabled) {
        if ($vs_obj->fcp_exist()) {

            # Check the reuse policy
            if ($reuse_policy =~ /^reuse$/i) {
                $self->sanlog(-msg => $self->_format_log("Using existing fcp service"));
            } elsif ($reuse_policy =~ /^check$/i) {
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $self->_set_err("fcp service exists and reuse policy is set to \"fail\"");
                return \@cleanup_data;
            }

            # Create the cleanup stack entry
            $fcp_cleanup_hash = {
                            reused    => $SanMgr::TRUE,
                            vserver   => $vs_name,
                            type      => "fcp_service",
                            cleanup   => $cleanup_policy,
                            };
        } else {
            # Check to see if the user used the fcp_service tag to define
            # values to be passed to fcp create
            my $setup_args;
            $setup_args->{"-vserver"} = $vs_name;

            # Check to see if the vserver args contain an fcp_service tag
            if (defined $vs_args->{FCP_SERVICE}) {
                my $fcp_data = $vs_args->{FCP_SERVICE};
                my $dumpout = Dumper($fcp_data);
                $self->sanlog(-msg => "fcp service tag given for "
                                . "vserver $vs_name:\n"
                                . $dumpout,
                          -msg_level  => 7);

                if (defined  $fcp_data->{"NODENAME"}) {
                    $setup_args->{"-nodename"} = $fcp_data->{"NODENAME"};
                }
                if (defined  $fcp_data->{"STATUS-ADMIN"}) {
                    $setup_args->{"-status-admin"} = $fcp_data->{"STATUS-ADMIN"};
                    $fcp_status_admin = $fcp_data->{"STATUS-ADMIN"};
                }
                if (defined  $fcp_data->{"FORCE"}) {
                    $setup_args->{"-force"} = $fcp_data->{"FORCE"};
                }
            }

            $self->sanlog(-msg => $self->_format_log("Creating fcp service on "
                                                    ."vserver $vs_name"));
            $fcp_obj = $vs_obj->create_fcp_obj(%{$setup_args});
            if ($vs_obj->error()) {
                $self->_set_err($vs_obj->errmsg());
                return \@cleanup_data;
            }

            # Create the cleanup stack entry
            $fcp_cleanup_hash = {
                            vserver      => $vs_name,
                            reused       => $SanMgr::FALSE,
                            type         => "fcp_service",
                            cleanup      => $cleanup_policy,
                            };
            if (defined ($fcp_status_admin)) {
                $fcp_cleanup_hash->{"status_admin"} = $fcp_status_admin;
            }
        } # end of fcp doesn't exist case

       # Log the entry we are adding to the cleanup stack
       my $dumpout = Dumper($fcp_cleanup_hash);
       $self->sanlog(-msg => "Dumper of fcp service data getting added"
                           . " to vserver cleanup stack:\n"
                           . $dumpout,
                     -msg_level  => 7);

       # Push the fcp setup data on the cleanup stack
       push(@cleanup_data, $fcp_cleanup_hash);
       $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                              . "for fcp service on "
                                              . "Vserver $vs_name"
                                              . " (reused = "
                                              . $fcp_cleanup_hash->{reused}
                                              . ")"));
    } # end for fcp_enabled

    $self->sanlog(-msg => $self->_format_log("Finished setting up protocol "
                                           . "services on vserver $vs_name"));

    return \@cleanup_data;
} ## end sub _setup_protocol_services

###############################################################################
# Method:    _setup_qos_policy_groups
# Objective: Responsible for calling methods to create the qos policy groups of
#            a vserver.
# Input:     A hash containing a vserver object, a cserver object, vserver data
# Returns:   Array reference containing hashes with qos policy group data
###############################################################################
sub _setup_qos_policy_groups {
    my ($self, %args) = @_;
    $self->clear_error();

    # Locals
    my ($vserver_obj, $data, $nacl_cluster_obj) = undef;
    my ($vserver_name, $aggregate) = "";
    my (@qos_content, @qos_policy_groups) = ();
    my ($vserver_data, $volume_data, $lun_data) = undef;

    # Process and validate args
    &ParamsMgr::process_args(%args);

    # The vserver where all objects will be creates
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # A data structure containing Vserver params and content
    $vserver_data = &ParamsMgr::get_arg(-arg      => "vserver_args",
                                        -required => $SanMgr::TRUE);

    # the NACL command interface to use
    $nacl_cluster_obj = &ParamsMgr::get_arg(-arg      => "nacl_cluster_obj",
                                            -required => $SanMgr::TRUE);

    # A hash containing results from vserver setup
    $data = &ParamsMgr::get_arg(-arg      => "vserver_data",
                                -required => $SanMgr::TRUE);


    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return \@qos_content;
    }

    # We need the vserver name
    $vserver_name = $vserver_obj->get_name();
    if ($vserver_obj->error()) {
        $self->_set_err("Unable to parse vserver name", $vserver_obj->errmsg());
        return \@qos_content;
    }

    my $no_empty_qos_name_found = $SanMgr::TRUE;
    # Look for volumes inside the vserver and put them in a separate list.
    # Also process qos_policy_groups and add them in a separate list.
    foreach my $key (keys (%{$vserver_data})) {
        if ($key =~ /^VOL\d*$/i) {
            my @volumes = ();
            push(@volumes,
                $self->_generate_copies(element => $vserver_data->{$key},
                                        type    => "vol"));
            push(@qos_policy_groups,$self->_get_volume_qos_policy_groups(
                                                        volumes => \@volumes,
                                                        vserver_name => $vserver_name));
        } elsif ($key =~ /^QOS_POLICY_GROUP\d*$/i) {
            if (not exists ($vserver_data->{$key}{NAME})) {
                if((not exists($vserver_data->{$key}{COUNT})) or
                  ($vserver_data->{$key}{COUNT} <= 1)) {
                    $no_empty_qos_name_found = $SanMgr::FALSE;
                }
                $vserver_data->{$key}{NAME} = $vserver_name . "_pg";
            }
            push(@qos_policy_groups,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "qos_policy_group"));

        }
    }

    if ($no_empty_qos_name_found and $vserver_data->{'USE_DEFAULT_QOS_POLICY_GROUP'} == 1) {
        push(@qos_policy_groups,
            {NAME    => $vserver_name . "_pg"});
    }

    # fetch all qos policy groups
    my @fetched_qos_objs;
    if(scalar(@qos_policy_groups)) {
        try {
        @fetched_qos_objs = NACL::CS::QosPolicyGroup->fetch(
                                            command_interface => $nacl_cluster_obj,
                                            allow_empty => 1);
        } otherwise {
            my $exception = shift;
            $self->_set_err("Failed to find qos policy groups" . $exception->text());
        };
        if ($self->error) {
            return \@qos_content;
        }
    }

    # Loop through all qos policy groups that we found during the previous
    # code and create them.
    foreach my $qos_hash_ref (@qos_policy_groups) {
        my %qos_args    = $self->_process_common_args(%$qos_hash_ref);
        my $cleanup_policy = delete $qos_args{cleanup_policy};
        my $reuse_policy   = delete $qos_args{reuse_policy};
        my $name_used = $SanMgr::FALSE;
        my $associated_vserver = "";
        my $reused = $SanMgr::FALSE;
        my $internal_duplicate = $SanMgr::FALSE;

        # Check to see if the qos policy group is in the list of
        # qos policy groups that we retrieved from the controller.
        foreach my $fetched_qos_obj (@fetched_qos_objs) {
            next if ($fetched_qos_obj->policy_group() ne $qos_args{name});
            $associated_vserver = $fetched_qos_obj->vserver();
            $name_used = $SanMgr::TRUE;
            last;
        }
        if ($name_used) {
            if ($associated_vserver ne $vserver_name) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("QOS Policy:\n" . Dumper(\%qos_args)
                                . " with this name already exists but it is created"
                                . " under a different vserver: " . $associated_vserver);
                return \@qos_content;
            } elsif ($reuse_policy =~ /^reuse$/i) {
                $reused = $SanMgr::True;
            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::True;
                # TODO Update this to check whether the qos policy group is what is expected.
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("QOS Policy" . Dumper(\%qos_args)
                                . " already exists and Reuse policy is set to \"fail\"");
                return \@qos_content;
            }
        } else {
            # Normalize params and get local policies
            $qos_args{'policy-group'} = delete $qos_args{name};
            $qos_args{'vserver'}      = $vserver_name;
            try {
                NACL::C::QosPolicyGroup->create(
                                        command_interface => $nacl_cluster_obj,
                                        %qos_args);
            } catch NACL::APISet::Exceptions::ResponseException with {
                my $exception = shift;
                unless($exception->text() =~ /Duplicate .*/i){
                    $self->_set_err(
                            "Failed to create qos policy group with qos arguments:\n"
                            . Dumper(\%qos_args)
                            . $exception->text());
                } else {
                    try {
                        my $duplicate_qos_obj = NACL::CS::QosPolicyGroup->fetch(
                                                command_interface => $nacl_cluster_obj,
                                                filter => \%qos_args,
                                                );

                        # If we find it then that means that it is really an internal
                        # duplicate within this code because we showed all of the
                        # qos policy groups because attempting to add more so we could
                        # compare the list to what the user wanted to add.
                        $internal_duplicate = $SanMgr::TRUE;

                    } otherwise {
                        my $exception = shift;
                        $self->_set_err(
                            "QOS Policy :\n" . Dumper(\%qos_args)
                            . " with this name already exists but it is created"
                            . " under a different vserver");
                    };
                }
            };
            if ($self->error) {
                return \@qos_content;
            }
        } # end ($name_used) {
            # Only push onto the cleanup stack if not an internal duplicate,
            # so that we dont attempt to cleanup the same thing twice.
            unless($internal_duplicate) {
                $data = {
                    reused    => $reused,
                    name      => $qos_args{'policy-group'},
                    type      => "qos_policy_group",
                    cleanup   => $cleanup_policy,
                    vserver   => $qos_args{'vserver'}
                };

                $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                                       . "for qos_policy_group "
                                                       . $qos_args{name}
                                                       . " (reused = $reused)"));
                push(@qos_content, $data);
            }
    } # end foreach my $qos_hash_ref (@qos_policy_groups)

    return \@qos_content;
} ## end sub _setup_qos_policy_groups

###############################################################################
# Method:    _get_volume_qos_policy_groups
# Objective: Responsible for extracting/creating qos policy group data for
#            volumes passed in, this includes the policies in the luns contained
#            in the volumes.
# Input:     A hash containing a vserver name, and an array of hashes of
#            volume data.
# Returns:   Array reference containing hashes with qos policy group data
###############################################################################
sub _get_volume_qos_policy_groups {
    my ($self, %args) = @_;
    $self->clear_error();

    # Locals
    my ($vserver_name, $volume_name, $qtree_name) = "";
    my (@qos_policy_groups, @volumes) = ();
    my ($vserver_data, $volume_data, $qtree_data) = undef;

    # Process and validate args
    &ParamsMgr::process_args(%args);
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    @volumes = &ParamsMgr::get_arg(-arg      => "volumes",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return ();
    }

    # Look for luns inside each vol and qtree and put them in a separate list.
    # Also process qos_policy_groups in vols and add them in a separate list.
    foreach $volume_data (@volumes) {
        my $no_empty_qos_name_found = $SanMgr::TRUE;
        my $volume_name = $volume_data->{NAME};
        foreach my $key (keys (%{$volume_data})) {
            if ($key =~ /^QTREE\d*$/i) {
                my @qtrees = ();
                $qtree_name = $volume_data->{$key}{NAME};
                push(@qtrees,
                    $self->_generate_copies(element => $volume_data->{$key},
                                            type    => "qtree"));
                foreach $qtree_data (@qtrees) {
                    my @luns = ();
                    if ($key =~ /^LUN\d*$/i) {
                        push(@luns,
                            $self->_generate_copies(element => $volume_data->{$key},
                                                    type    => "lun"));
                        push(@qos_policy_groups,$self->_get_lun_qos_policy_groups(
                                                            luns         => \@luns,
                                                            vserver_name => $vserver_name,
                                                            volume_name  => $volume_name,
                                                            qtree_name   => $qtree_name));
                    }
                }
            } elsif ($key =~ /^LUN\d*$/i) {
                my @luns = ();
                push(@luns,
                    $self->_generate_copies(element => $volume_data->{$key},
                                            type    => "lun"));
                push(@qos_policy_groups,$self->_get_lun_qos_policy_groups(
                                                    luns => \@luns,
                                                    vserver_name => $vserver_name,
                                                    volume_name  => $volume_name,
                                                    qtree_name   => $qtree_name));
            } elsif ($key =~ /^QOS_POLICY_GROUP\d*$/i) {
                if (not exists ($volume_data->{$key}{NAME})) {
                    if((not exists($volume_data->{$key}{COUNT})) or
                      ($volume_data->{$key}{COUNT} <= 1)) {
                        $no_empty_qos_name_found = $SanMgr::FALSE;
                    }
                    $volume_data->{$key}{NAME} = $vserver_name . "_" .
                                                 $volume_name . "_pg";
                }
                push(@qos_policy_groups,
                    $self->_generate_copies(element => $volume_data->{$key},
                                            type    => "qos_policy_group"));

            }
        }
        if ($no_empty_qos_name_found and $volume_data->{'USE_DEFAULT_QOS_POLICY_GROUP'} == 1) {
            push(@qos_policy_groups,
                {NAME    => $vserver_name . "_" .
                            $volume_name . "_pg"});
        }
    }
    return @qos_policy_groups;
} # end _get_volume_qos_policy_groups

###############################################################################
# Method:    _get_lun_qos_policy_groups
# Objective: Responsible for extracting/creating qos policy group data for luns
#            passed in.
# Input:     A hash containing a vserver name, volume name, lun name,
#            qtree name(optional), and an array of hashes of lun data.
# Returns:   Array reference containing hashes with qos policy group data
###############################################################################
sub _get_lun_qos_policy_groups {
    my ($self, %args) = @_;
    $self->clear_error();

    # Locals
    my ($vserver_name, $volume_name, $qtree_name) = "";
    my (@qos_policy_groups, @luns) = ();
    my ($lun_data) = undef;
    # Process and validate args
    &ParamsMgr::process_args(%args);
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    $volume_name = &ParamsMgr::get_arg(-arg      => "volume_name",
                                        -required => $SanMgr::TRUE);

    $qtree_name = &ParamsMgr::get_arg(-arg      => "qtree_name",
                                      -required => $SanMgr::FALSE,
                                      -default => "");

    @luns = &ParamsMgr::get_arg(-arg      => "luns",
                                   -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return ();
    }

    # Look through luns for qos policy groups defined..
    foreach $lun_data (@luns) {
        my $no_empty_qos_name_found = $SanMgr::TRUE;
        my $name = $lun_data->{NAME};
        foreach my $key (keys (%{$lun_data})) {
            if ($key =~ /^QOS_POLICY_GROUP\d*$/i) {
                if (not exists ($lun_data->{$key}{NAME})) {
                    if((not exists($lun_data->{$key}{COUNT})) or
                      ($lun_data->{$key}{COUNT} <= 1)) {
                        $no_empty_qos_name_found = $SanMgr::FALSE;
                    }
                    if (defined($qtree_name) and length($qtree_name)) {
                        $lun_data->{$key}{NAME} = $vserver_name . "_" .
                                                  $volume_name  . "_" .
                                                  $qtree_name   . "_" .
                                                  $name         . "_pg";
                    } else {
                        $lun_data->{$key}{NAME} = $vserver_name . "_" .
                                                  $volume_name  . "_" .
                                                  $name         . "_pg";
                    }
                }
                push(@qos_policy_groups,
                     $self->_generate_copies(element => $lun_data->{$key},
                                             type    => "qos_policy_group"));

            }
        }
        if ($no_empty_qos_name_found and $lun_data->{'USE_DEFAULT_QOS_POLICY_GROUP'} == 1) {
            if (defined($qtree_name) and length($qtree_name)) {
                push(@qos_policy_groups,
                 {NAME    => $vserver_name . "_" .
                             $volume_name  . "_" .
                             $qtree_name   . "_" .
                             $name         . "_pg"});
            } else {
                push(@qos_policy_groups,
                 {NAME    => $vserver_name . "_" .
                             $volume_name  . "_" .
                             $name         . "_pg"});
            }
        }
    }
    return @qos_policy_groups;
} # end _get_lun_qos_policy_groups

###############################################################################
# Method:    _setup_vserver_content
# Objective: Responsible for calling methods to create the contents of a
#            vserver such as vols, igroups, lifs, etc.
# Input:     A hash containing a vserver object, a cserver object, vserver data
# Returns:   Array reference containing hashes with vserver content data
###############################################################################
sub _setup_vserver_content {
    my ($self, %args) = @_;
    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data) = undef;
    my ($vserver_name, $aggregate) = "";
    my (@vserver_content, @volumes, @luns, @lifs, @portsets) = ();
    my ($vserver_data, $lif_data, $igroup_data, $volume_data) = undef;
    my ($portset_data, $tpgroup_data, $nacl_cluster_obj) = undef;
    my (@tpgroups, @igroups) = ();

    # Process and validate args
    &ParamsMgr::process_args(%args);

    # The vserver where all objects will be creates
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # A data structure containing Vserver params and content
    $vserver_data = &ParamsMgr::get_arg(-arg      => "vserver_args",
                                        -required => $SanMgr::TRUE);

    # A cserver (filer object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # A hash containing results from vserver setup
    $data = &ParamsMgr::get_arg(-arg      => "vserver_data",
                                -required => $SanMgr::TRUE);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);
    # the nacl_cluster_obj
    $nacl_cluster_obj = &ParamsMgr::get_arg(-arg      => "nacl_cluster_obj",
                                            -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # We need the vserver name
    $vserver_name = $vserver_obj->get_name();
    if ($vserver_obj->error()) {
        $self->_set_err("Unable to parse vserver name", $vserver_obj->errmsg());
        return [];
    }

    # Look for vservers inside each aggregate and put them on separate lists.
    foreach my $key (keys (%{$vserver_data})) {
        if ($key =~ /^VOL\d*$/i) {
            push(@volumes,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "vol"));
        } elsif ($key =~ /^LIF\d*$/i) {
            # NOTE: specifying type to tell generate copies to
            # assign different IPs to each element (iSCSI)
            push(@lifs,
                 $self->_generate_copies('element'     => $vserver_data->{$key},
                                         'parent_name' => $vserver_name,
                                         'type'        => "LIF",
                                        )
                 );
            if ($self->error()) {
                $self->_set_err("Failed to generate copy for lif elements");
                return [];
            }
        } elsif ($key =~ /^IGROUP\d*$/i) {
            push(@igroups,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "igroup"));
        } elsif ($key =~ /^PORTSET\d*$/i) {
            push(@portsets,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "portset"));
        } elsif ($key =~ /^TPGROUP\d*$/i) {
            push(@tpgroups,
                 $self->_generate_copies(element => $vserver_data->{$key},
                                         type    => "tpgroup"));
        }

    }

    # There are certain things that do not depend on each other, such as igroups
    # and lifs. These could be created in parallel. However, there's no real
    # benefit from doing that. The order in which elements are created is:
    # lifs, igroups, portsets, tpgroups, volumes, luns

    # TODO: On error, below we always return [].  We should return the
    #       current vserver_content so we can cleanup what's already
    #       been done.

    # Start with lifs
    if (scalar(@lifs)) {
        $lif_data = $self->_setup_lifs(vserver_name => $vserver_name,
                                       vserver_obj => $vserver_obj,
                                       lifs => \@lifs,
                                       cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to setup lifs");
            return [];
        }

        push(@vserver_content, @{$lif_data});

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($lif_data);
        $self->sanlog(-msg => "Dumper of lif data getting added to vserver "
                            . "cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);
    }

    # Check to see if the vserver args contain an ADDRESS tag
    # If so, attempt to set up isns
    if (defined ($vserver_data->{ISNS_SERVICE}) and
        defined ($vserver_data->{ISNS_SERVICE}->{'ADDRESS'})) {

        my ($isns_status_admin, $isns_server_address, $isns_force) = undef;
        my ($isns_setup_args, $isns_modify_args, $isns_update_args) = {};

        my $dumpout = Dumper($vserver_data);
        $self->sanlog(-msg => "isns service tag given for "
                      . "vserver $vserver_name:\n"
                      . $dumpout,
                      -msg_level  => 7);

        $isns_setup_args->{"-address"} = $vserver_data->{ISNS_SERVICE}->{'ADDRESS'};
        $isns_modify_args->{"-address"} = $vserver_data->{ISNS_SERVICE}->{'ADDRESS'};

        if (defined ($vserver_data->{ISNS_SERVICE}->{'FORCE'})){
            $isns_setup_args->{"-force"} = $vserver_data->{ISNS_SERVICE}->{'FORCE'};
        }

        if (defined ($vserver_name)){
            $isns_setup_args->{"-vserver"} = $vserver_name;
            $isns_modify_args->{"-vserver"} = $vserver_name;
            $isns_update_args->{"-vserver"} = $vserver_name;
        }

        my $iscsi_obj = $vserver_obj->get_iscsi_obj();

        if ($iscsi_obj == {}){
            $self->_set_err("Could not get ISCSI object to create ISNS.",
                $vserver_obj->get_err());
            return [];
        }

        # Create iSNS server for iSCSI lun discovery
        $iscsi_obj->isns_create(%{$isns_setup_args});

        if ($iscsi_obj->error()) {
            my $iscsi_err_message = $iscsi_obj->errmsg();

            # If we get an error message about
            # duplicate entries, keep going.
            if ( not ($iscsi_err_message =~ /duplicate|already exists/)){
                $self->_set_err($iscsi_err_message);
                return [];
            }
        }

        if (defined $vserver_data->{ISNS_SERVICE}->{'STATUS-ADMIN'}){
            $isns_modify_args->{"-status-admin"}   = $vserver_data->{ISNS_SERVICE}->{'STATUS-ADMIN'};
        }

        $iscsi_obj->isns_modify(%{$isns_modify_args});

        if ($iscsi_obj->error()) {
            my $iscsi_err_message = $iscsi_obj->errmsg();
            $self->_set_err($iscsi_err_message);
            return [];
        }

        # Update iSNS server for iSCSI lun discovery, if the isns service is not
        # explicitly set to down
        if ((defined $vserver_data->{ISNS_SERVICE}->{'STATUS-ADMIN'} and
            $vserver_data->{ISNS_SERVICE}->{'STATUS-ADMIN'} ne "down") or
            (not defined $vserver_data->{ISNS_SERVICE}->{'STATUS-ADMIN'})){
            $iscsi_obj->isns_update(%{$isns_update_args});
            if ($iscsi_obj->error()) {
                my $iscsi_err_message = $iscsi_obj->errmsg();
                $self->_set_err($iscsi_err_message);
                return [];
            }
        }

        my $isns_cleanup_hash = {
                            vserver      => $vserver_name,
                            reused       => $SanMgr::FALSE,
                            type         => "isns_service"
                            };

        push(@vserver_content, $isns_cleanup_hash);
    }

    my @fetched_igroup_objs = ();

    if (scalar(@igroups) or scalar(@portsets)) {
        # fetch all igroup objects on this vserver
        @fetched_igroup_objs = $cserver->fetch_igroup_objs(-vserver => $vserver_name);
        if ($cserver->error()) {
            $self->_set_err("Failed to fetch igroup objects", $cserver->errmsg());
            return [];
        }
    }

    # Then iGroups
    if (scalar(@igroups)) {
        $igroup_data = $self->_setup_igroups(vserver_name => $vserver_name,
                                             vserver_obj => $vserver_obj,
                                             igroups => \@igroups,
                                             fetched_igroups => \@fetched_igroup_objs,
                                             cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to setup igroups");
            return [];
        }

        push(@vserver_content, @{$igroup_data});

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($igroup_data);
        $self->sanlog(-msg => "Dumper of igroup data getting added to vserver "
                            . "cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);
    }

    # Then portsets
    if (scalar(@portsets)) {
        $portset_data = $self->_setup_portsets(vserver_name => $vserver_name,
                                               vserver_obj => $vserver_obj,
                                               fetched_igroups => \@fetched_igroup_objs,
                                               portsets => \@portsets,
                                               cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to setup portsets");
            return [];
        }

        push(@vserver_content, @{$portset_data});

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($portset_data);
        $self->sanlog(-msg => "Dumper of portset data getting added to vserver "
                            . "cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);
    }

    # Then tpgroups
    if (scalar(@tpgroups)) {
        $tpgroup_data = $self->_setup_tpgroups(vserver_name => $vserver_name,
                                               tpgroups => \@tpgroups,
                                               cserver => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to setup target portal groups");
            return [];
        }

        push(@vserver_content, @{$tpgroup_data});

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($tpgroup_data);
        $self->sanlog(-msg => "Dumper of tpgroup data getting added to vserver "
                            . "cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);
    }

    # Create volumes
    if (scalar(@volumes)) {
        $volume_data = $self->_setup_volumes(vserver_name => $vserver_name,
                                             vserver_obj  => $vserver_obj,
                                             aggregate    => $aggregate,
                                             volumes      => \@volumes,
                                             cserver      => $cserver,
                                             nacl_cluster_obj => $nacl_cluster_obj);
        if ($self->error()) {
            $self->_set_err("Failed to setup volumes");
            return [];
        }

        push(@vserver_content, @{$volume_data});

        # Log the entry we are adding to the cleanup stack
        my $dumpout = Dumper($volume_data);
        $self->sanlog(-msg => "Dumper of volume data getting added to vserver "
                            . "cleanup stack:\n"
                            . $dumpout,
                      -msg_level  => 7);

        my $bind_data = $self->_setup_lunbinds(vserver_name => $vserver_name,
                                               vserver_obj  => $vserver_obj,
                                               volumes      => \@volumes,
                                               cserver      => $cserver);
        if ($self->error()) {
            $self->_set_err("Failed to setup lun binds");
            return [];
        }

        push(@vserver_content, @{$bind_data});

        # Log the entry we are adding to the cleanup stack
        my $binddumpout = Dumper($bind_data);
        $self->sanlog(-msg => "Dumper of lun bind data getting added to vserver "
                            . "cleanup stack:\n"
                            . $binddumpout,
                      -msg_level  => 7);

    }

    return \@vserver_content;
} ## end sub _setup_vserver_content

###############################################################################
#                           L I F   M E T H O D S
#
#  Things to consider:
#  1. Lifs are created serially. So there's no real need for having plural,
#     singular methods.
#  2. Currently, our way of assigning IPs is not clever at all!
#     Need a better way to handle duplicate ip addresses for ISCSI Lifs
#
###############################################################################

###############################################################################
# Method:    _setup_lifs
# Objective: Method responsible for creating or reusing lif objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     lif_data => hash containing all args for a lif
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_lifs {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $lifs, $data, $vserver_obj, $lif_obj) = undef;
    my ($vserver_name, $name, $gateway, $reuse_policy, $cleanup_policy,
        $protocol) = "";
    my ($lif_exists, $reused) =  $SanMgr::FALSE;
    my (@lif_data) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up lifs"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing lif details
    $lifs = &ParamsMgr::get_arg(-arg      => "lifs",
                                -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # The vserver object
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # fetch all lif objects in this vserver
    my @fetched_lif_objs = $cserver->fetch_lif_objs(-vserver => $vserver_name);
    if ($cserver->error()) {
        $self->_set_err("Failed to fetch lif objects", $cserver->errmsg());
        return [];
    }

    # Get lif data from the resource data structure
    foreach my $lif (@{$lifs}) {
        # Normalize params and get local policies
        my %lif_args    = $self->_process_common_args(%$lif);
        $cleanup_policy = delete $lif_args{cleanup_policy};
        $reuse_policy   = delete $lif_args{reuse_policy};
        $gateway        = delete $lif_args{gateway};
        $lif_exists     = $SanMgr::FALSE;
        $reused         = $SanMgr::FALSE;

        # Get lif name
        if (defined($lif_args{name})) {
            $name          = $lif_args{name};
            $lif_args{lif} = delete $lif_args{name};
        } else {
            $self->_set_err("Lif name is missing!");
            return [];
        }

        # Get protocol name
        if (defined($lif_args{'data-protocol'})) {
            $protocol = $lif_args{'data-protocol'};
        }

        $self->sanlog(-msg => $self->_format_log("Setting up $protocol LIF $name"));

        foreach my $key (keys(%lif_args)) {
            if ($key =~ /^ADDRESS_RANGE($|[1-9]$)/i) {
                delete $lif_args{$key};
            }
        }

        # find the lif object
        foreach my $f_lif_obj (@fetched_lif_objs) {
            next if ($f_lif_obj->get_name(-use_cache => $SanMgr::TRUE) ne $name);
            # we found it
            $lif_exists = $SanMgr::TRUE;
            $lif_obj = $f_lif_obj;
            last;
        }

        # Create or reuse lifs according to the policies
        if ($lif_exists) {
            if ($reuse_policy =~ /^reuse$/i) {
                $self->sanlog(-msg => $self->_format_log("Using existing LIF $name"));
                $reused = $SanMgr::TRUE;
            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::FALSE;
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("Lif exists and reuse policy is set to \"fail\"");
                return [];
            }
        } else {
            $self->sanlog(-msg => $self->_format_log("Creating LIF $name"));
            my ($ips, $unassigned_ips) = [];
            my $found_ip = $SanMgr::FALSE;
            my $error_retry = $SanMgr::FALSE;

            # Get the Range of IPs that could be used when creating iSCSI lifs
            if (defined($self->{_objects}->{cluster}->{$vserver_name}->{$name}->{ips})) {
                $ips = $self->{_objects}->{cluster}->{$vserver_name}->{$name}->{ips};
            }
            # Get the ips that are not assigned to this vserver/lif combination.
            # These IPs could be used if the IP is duplicated.
            if (defined($self->{_objects}->{cluster}->{$vserver_name}->{$name}->{unassigned_ips})) {
                $unassigned_ips =
                    $self->{_objects}->{cluster}->{$vserver_name}->{$name}->{unassigned_ips};
            }

            # Loop until a suitable ip address has been found
            while (!$found_ip) {
                # First - get the address to create the lif
                if ($protocol =~ /iscsi|cifs|nfs|none/i) {
                    unless( defined($lif_args{'address'}) and length($lif_args{'address'}) ) {
                        $lif_args{'address'} = $self->_get_next_available_ip('ips' => $ips,
                                                                             'unassigned_ips' => $unassigned_ips
                                                                            );
                        if( $self->error() ) {
                            $self->_set_err("Could not get a valid IP address for Lif[$name]");
                            return [];
                        }
                        # Retry only once if there is an undisclosed error
                        $error_retry = $SanMgr::TRUE;
                    }
                }
                # Second - create the Lif
                $lif_obj = $vserver_obj->create_lif_obj(%lif_args);
                # Third - Check for error
                if ( $vserver_obj->error() ){
                    # Both cli and zapi contain the same fragment on the error message.
                    if ($vserver_obj->errmsg() =~ /Duplicate ip address/i){
                        $self->sanlog(-msg =>"Duplicate IP[". $lif_args{'address'} .
                                      "] address. Retrying...");
                        delete $lif_args{'address'};
                        next;
                    }
                    # Could face an error that is caused by 'Duplicated IP' but the error
                    # statement does not says anything meaninful, so, let's retry one more
                    # time with the same IP and if the error persists, let the error bubble up.
                    elsif ( ($vserver_obj->errmsg() =~ /failed to add lif/i) and ($error_retry) ) {
                        # Get a random number between 5 and 10 (secs), that will be used to sleep.
                        my $rndm_wait = int(rand(5)+6);
                        $self->sanlog(-msg =>"Encountered what could be a Collision Error " .
                                             "(known issue), will retry again in " . $rndm_wait .
                                             " secs, with same IP address " . $lif_args{'address'} );
                        # Only retry once
                        $error_retry = $SanMgr::FALSE;
                        # Need to sleep before retry to unsync the vservers that are trying to create
                        # the LIF with the conflicting IP address.
                        &SanUtil::nate_sleep($rndm_wait);
                        next;
                    } else {
                        # If lif creation fails because ipv6 isn't enabled, print a nice error message.
                        # explaining how to resolve this issue in addition to the normal error.
                        if ($vserver_obj->errmsg() =~ /IPv6 is not enabled/i){
                            $self->sanlog(-warn =>"You have tried to create an IPv6 lif without having " .
                                                 "enabled ipv6 on your cluster. Add the parameter " .
                                                 "ipv6_enabled=\"true\" to the cluster tag in your xml " .
                                                 "to have SanMgr enable ipv6.");
                        }

                        $self->_set_err("Lif create obj command failed.\n" . $vserver_obj->errmsg());
                        return [];
                    }
                # Fourth - No errors
                } else {
                    $found_ip = $SanMgr::TRUE;
                    last;
                }
            }
            $reused = $SanMgr::FALSE;
        }

        # If the user passed gateway, then appropriate route should be added
        if ($gateway) {
            $lif_obj->create_route(-gateway => $gateway);
            if ($lif_obj->error()) {
                $self->_set_err("Failed to create route for gateway '$gateway'",
                                $lif_obj->errmsg());
                return [];
            }
        }

        # Store the LIF attributes, either FCP or iSCSI, in the vserver object
        # This is needed in order to load fcp or iscsi targets using the right
        # data.

        %lif_args = (%lif_args, vserver => $vserver_name);
        my (%fcp_targets, %iscsi_targets);

        if ($protocol =~ /^fcp$/i) {
            # force copy the hash
            %fcp_targets = %lif_args;
            push (@{$vserver_obj->{fcp_lif_data}}, \%fcp_targets);
        } elsif ($protocol =~ /^iscsi$/i) {
            %iscsi_targets = %lif_args;
            push (@{$vserver_obj->{iscsi_lif_data}}, \%iscsi_targets);
        }

        $data = {
            reused        => $reused,
            fcp_targets   => \%fcp_targets,
            iscsi_targets => \%iscsi_targets,
            name          => $name,
            vserver       => $vserver_name,
            type          => "lif",
            cleanup       => $cleanup_policy
        };

        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                               . "for LIF $name"
                                               . " (reused = $reused)"));
        push(@lif_data, $data);
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up lifs"));

    return \@lif_data;
} ## end sub _setup_lifs

###############################################################################
#                         I G R O U P   M E T H O D S
#
#  Things to consider:
#  1. Igroups are created serially. So there's no real need for having plural,
#     singular methods. They are created after lifs.
#  2. TODO: Handle igroup options.
#
###############################################################################

###############################################################################
# Method:    _setup_igroups
# Objective: Method responsible for creating or reusing igroup objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     igroup_data => hash containing all args for an igroup
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_igroups {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $igroups, $fetched_igroups, $data, $igroup_obj, $vserver_obj) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy) = "";
    my ($reused) =  $SanMgr::FALSE;
    my (@igroup_data, %igroup_args) = ();
    my (@initiators) = ();
    my ($alua_state) = ();

    $self->sanlog(-msg => $self->_format_log("Start setting up igroups"));

    # Process agrs and validate mandatory params

    # Name of the vserver
    $vserver_name = exists($args{vserver_name}) ? delete($args{vserver_name}) : undef;
    # vserver object
    $vserver_obj = exists($args{vserver_obj}) ? delete($args{vserver_obj}) : undef;
    # Main data structure containing igroup details
    $igroups = exists($args{igroups}) ? delete($args{igroups}) : undef;
    # fetched igroup objects from this vserver
    $fetched_igroups = exists($args{fetched_igroups}) ? delete($args{fetched_igroups}) : undef;
    # cserver object (SanFiler object)
    $cserver = exists($args{cserver}) ? delete($args{cserver}) : undef;

    my @fetched_igroup_objs = @{$fetched_igroups};

    # Get igroup data from the resource data structure
    foreach my $igroup (@{$igroups}) {
        # Normalize params and get local policies
        %igroup_args    = $self->_process_common_args(%$igroup);
        $cleanup_policy = delete $igroup_args{cleanup_policy};
        $reuse_policy   = delete $igroup_args{reuse_policy};
        $alua_state     = delete $igroup_args{alua};

        # locally declare these arrays since we will store their
        # references in the return message
        my (@inits, @options) = ();
        my (@passthru_initiator_list) = ();

        # reset initiators array
        @initiators = ();

        # Get igroup name
        if (defined($igroup_args{name})) {
            $name = $igroup_args{name};
            $igroup_args{igroup} = delete $igroup_args{name};
        } else {
            $self->_set_err("Igroup name is missing!");
            return [];
        }

        $self->sanlog(-msg => $self->_format_log("Setting up igroup $name"));

        # Remove options and initiators from the hash..
        foreach my $key (keys(%igroup_args)) {
            if ($key =~ /^INITIATOR\d*$/i) {
                push(@initiators, delete $igroup_args{$key});
            } elsif ($key =~ /^INITIATOR_LIST$/i) {
                @passthru_initiator_list = split(/\s*,\s*/, delete($igroup_args{$key}));
            } elsif ($key =~ /OPTIONS/i) {
                push(@options, delete $igroup_args{$key});
            }
        }

        # Check if the igroup exists.
        # extract it's object if we find it in the fetched igroup object list
        my $igroup_exists = $SanMgr::FALSE;
        foreach my $i_obj (@fetched_igroup_objs) {
            next if ($i_obj->get_name(-use_cache => $SanMgr::TRUE) ne $name);
            $igroup_obj = $i_obj;
            $igroup_exists = $SanMgr::TRUE;
            last;
        }

        # Create or reuse igroups according to the policies
        if ($igroup_exists) {
            if ($reuse_policy =~ /^reuse$/i) {
                $self->sanlog(-msg => $self->_format_log("Using existing igroup $name"));
                $reused = $SanMgr::TRUE;
            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::FALSE;
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("iGroup exists and reuse policy is set to " .
                                "\"fail\"");
                return [];
            }
        } else {
            $self->sanlog(-msg => $self->_format_log("Creating igroup $name"));
            $igroup_obj = $cserver->create_igroup(%igroup_args,
                                                  vserver_obj => $vserver_obj,
                                                  vserver => $vserver_name);
            if ($cserver->error()) {
                $self->_set_err("Failed to create igroup $name",
                                $cserver->errmsg());
                return [];
            }

            $reused = $SanMgr::FALSE;
        }

        # if passthru initiators were specified, then process them
        if (scalar(@passthru_initiator_list)) {
            my @passthru_initiator_list_to_add = ();

            foreach my $initiator (@passthru_initiator_list) {
                # get rid of spaces
                $initiator =~ s/\s+//g;

                # intelligently parse the initiator value
                if ($initiator =~ /^iqn\.\S+/) {
                    # this is a valid iscsi iqn nodename
                    push(@passthru_initiator_list_to_add, $initiator);
                } elsif ($initiator =~ /^[a-fA-F0-9]{16}$/) {
                    # this is a valid FCP WWPN with no : chars
                    push(@passthru_initiator_list_to_add, $initiator);
                } elsif ($initiator =~ /^[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:
                                         [a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}$/x) {
                    # this is a valid FCP WWPN with : chars
                    push(@passthru_initiator_list_to_add, $initiator);
                } else {
                    # if all else fails, check to see if they are refering to a client name
                    my ($client_obj) = $self->get_objects(types => ["client", "hypervisor"],
                                                          names => [$initiator]);
                    if ($self->error()) {
                        # there is no client by this name, then this is a failure
                        $self->_set_err("Failed to identify what to do with initiator '$initiator', " .
                                        "verify that it is valid client name or that it is in ".
                                        "the proper FCP WWPN (16 hex chars) or iSCSI nodename (iqn.*) format.");
                        return [];
                    }

                    # add this client to the @initiators array so that the next section can handle this client
                    # we don't want duplicate clients in this list, so verify that its not already there
                    if (not (grep { $_->{NAME} =~ /^$initiator$/ } @initiators)) {
                        # this array is an array of hash references not just client names
                        push(@initiators, { NAME => $initiator });
                    }
                }
            }

            # go through all the passthru initiators that we processed earlier and add them to the igroup
            if (scalar(@passthru_initiator_list_to_add)) {
                foreach my $initiator (@passthru_initiator_list_to_add) {
                    # add colons if the intiator we have does not contain colons
                    if ($initiator =~ /^[A-Fa-f0-9]{16}$/) {
                        $initiator = $self->normalize_fcp_wwn(-wwn => $initiator);
                    }

                    # check to see if this initiator is already in the igroup
                    next if ($igroup_obj->get_initiator(-use_cache => $SanMgr::TRUE) =~ /\b\Q$initiator\E\b/i);

                    $igroup_obj->add(initiator => $initiator);
                    if ($igroup_obj->error()) {
                        # If the initiator is already in the igroup, clear the
                        # error and go to the next intiator.
                        # This can occur when users reuse igroups
                        if ($igroup_obj->errmsg() =~ /Initiator group already contains node/i) {
                            $igroup_obj->_clear_err();
                            next;
                        } else {
                            $self->_set_err("Failed to add $initiator to $igroup",
                                            $igroup_obj->errmsg());
                            return [];
                        }
                    }
                }

                # add each passthru initiator to the inits list
                # this is part of the datastructure that we are sending back up the stack
                # for cleanup purposes
                push(@inits, @passthru_initiator_list_to_add);

                # Log the initiators we are adding
                my $dumpout = Dumper(@passthru_initiator_list_to_add);
                $self->sanlog(-msg => "Dumper of passthru initiators added to igroup "
                                    . " $name:\n"
                                    . $dumpout,
                              -msg_level  => 3);
            }
        }

        # Add initiators to igroups
        foreach my $client (@initiators) {
            my @initiator_list = ();

            if (defined($client->{NAME})) {
                my ($client_obj) = $self->get_objects(types => ["client", "hypervisor"],
                                                      names => [$client->{NAME}]);
                if ($self->error()) {
                    $self->_set_err("Failed to obtain SanClient object for client '" . $client->{NAME} . "'");
                    return [];
                }

                # find the paramset datastructure for $client->{NAME}
                my $all_clients = $self->EnGenX::Util::_get_all_client_paramsets();                
                my $all_hypervisors = &ParamsMgr::search(-search_path => [],
                                                         -key => "HYPERVISORS",
                                                         -key_filter => qr/^(ESX|HYPERV)\d*$/,
                                                         paramset => $self->EnGenX::Util::get_paramset());
                foreach my $data (@{$all_clients}, values %{$all_hypervisors}) {
                    if ($data->{NAME} eq $client->{NAME}) {
                        # found it, update the datastructure object
                        $client = $data;
                        last;
                    }
                }

                # first check to see if this client has an initiators field, if it doesn then use the
                # initiators from that rather than dynamically searching the client object.
                my @fc_override_init_list = ();
                my @iscsi_override_init_list = ();
                my $perform_dynamic_lookup = 1;

                if (exists($client->{INITIATORS}) and ref($client->{INITIATORS}) eq "HASH") {
                    foreach my $key (keys %{$client->{INITIATORS}}) {
                        if ($igroup_args{'protocol'} =~ /mixed|fcp/i) {
                            if ($key =~ /WWPN/) {
                                # pull out the initiator value for this wwpn entry
                                # and push it on the init list
                                push(@fc_override_init_list, $client->{INITIATORS}->{$key}->{VALUE});
                            }
                        } elsif ($igroup_args{'protocol'} =~ /mixed|iscsi/i) {
                            if ($key =~ /ISCSI_NODENAME/) {
                                # pull out the initiator value for this wwpn entry
                                # and push it on the init list
                                push(@iscsi_override_init_list, $client->{INITIATORS}->{$key}->{VALUE});
                            }
                        }
                    }
                    # add the inits to the final initiator list
                    push(@initiator_list, @fc_override_init_list, @iscsi_override_init_list);

                    # check to see if we are being requested to fall back to the dynamic lookup algorithm
                    # in the event that we dont have a wwpn/nodename override for one of the protocols
                    $perform_dynamic_lookup = $client->{INITIATORS}->{DYNAMIC_LOOKUP};
                }

                # perform a dynamic lookup for initiators if we were requested to
                if ($perform_dynamic_lookup == 1) {
                    # search the client dynamically and add any additional initiators that are needed
                    # for fcp/iscsi
                    my @initiator_objs = $client_obj->get_initiators();
                    if ($client_obj->error()) {
                        $self->_set_err("Failed to obtain initators for client " . $client->{NAME},
                                        $client_obj->errmsg());
                        return [];
                    }

                    foreach my $initiator_obj (@initiator_objs) {
                        # What protocol is this initiator?
                        my $initiator_protocol = $initiator_obj->get_protocol();

                        # If it's fcp, get the portname.
                        if ($initiator_protocol eq "fcp") {
                            # skip this initiator if the user specified an override fcp init list
                            next if (scalar(@fc_override_init_list));

                            if ($igroup_args{'protocol'} =~ /mixed|fcp/i) {
                                # store the fcp wwpn if the igroup is fcp/mixed
                                push(@initiator_list, $initiator_obj->get_portname());
                            }
                        } elsif ($initiator_protocol eq "iscsi") {
                            # skip this initiator if the user specified an override iscsi init list
                            next if (scalar(@iscsi_override_init_list));

                            # Else if it's iscsi, get the nodename.
                            if ($igroup_args{'protocol'} =~ /mixed|iscsi/i) {
                                # store the iscsi nodename if the igroup is iscsi/mixed
                                push(@initiator_list, $initiator_obj->get_nodename());
                            }
                        } else {
                            # Else unknown protocol, return error
                            $self->_set_err("Initiator " . $client->{NAME} . " has an interface that " .
                                            "is not using either iSCSI or FCP");
                            return [];
                        }
                    }
                }

                # add all the discovered initiators to the igroup
                foreach my $init (@initiator_list) {
                    # add colons if the intiator we have does not contain colons
                    if ($init =~ /^[A-Fa-f0-9]{16}$/) {
                        $init = lc(&SanUtil::wwn_add_colons(-wwn => $init));
                    }

                    # check to see if this initiator is already in the igroup
                    next if ($igroup_obj->get_initiator(-use_cache => $SanMgr::TRUE) =~ /\b\Q$init\E\b/i);

                    $igroup_obj->add(initiator => $init);
                    if ($igroup_obj->error()) {
                        # If the initiator is already in the igroup, clear the
                        # error and go to the next intiator.
                        # This can occur when users reuse igroups
                        if ($igroup_obj->errmsg() =~ /Initiator group already contains node/i) {
                            $igroup_obj->_clear_err();
                            next;
                        } else {
                            $self->_set_err("Failed to add $init to $igroup",
                                            $igroup_obj->errmsg());
                            return [];
                        }
                    }
                }

                # Log the initiators we are adding
                my $dumpout = Dumper(@initiator_list);
                $self->sanlog(-msg => "Dumper of initiators added to igroup "
                                    . " $name:\n"
                                    . $dumpout,
                              -msg_level  => 3);

                # add all the initiators used to the inits list
                push(@inits, @initiator_list);
            }
        }

        if ($alua_state =~ /true|false/) {
            # Set ALUA according to callers request
            $igroup_obj->set(-attribute => "alua", -value => $alua_state);
            if ($igroup_obj->error()) {
                $self->_set_err("Failed to set alua attribute ($alua_state) on igroup", $igroup_obj->errmsg());
                return [];
            }
        }

        $data = {
            reused    => $reused,
            name      => $name,
            initiator => \@inits,
            options   => \@options,
            vserver   => $vserver_name,
            type      => "igroup",
            cleanup   => $cleanup_policy
        };

        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                               . "for igroup $name"
                                               . " (reused = $reused)"));
        push(@igroup_data, $data);
    } # end foreach igroup

    $self->sanlog(-msg => $self->_format_log("Finished setting up igroups"));

    return \@igroup_data;
} ## end sub _setup_igroups

###############################################################################
#               P O R T S E T  &  T P G R O U P S  M E T H O D S
#
#  Things to consider:
#  1. Portsets are created serially, once all igroups have been created.
#  2. Binding to igroups is done after creation
#  3. For tpgroups, iSCSI service has to be started.
#
###############################################################################

###############################################################################
# Method:    _setup_portsets
# Objective: Method responsible for creating or reusing portset objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     portset_data => hash containing all args for a portset
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_portsets {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $portsets, $fetched_igroups, $data, $vserver_obj, $portset_obj) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy, $igroup_name) = "";
    my ($portset_exists, $reused) =  $SanMgr::FALSE;
    my (@portset_data, %portset_args) = ();
    my ($igroup_obj, $igroup) = undef;

    $self->sanlog(-msg => $self->_format_log("Start setting up portsets"));

    # Process agrs and validate mandatory params

    # Main data structure containing portset details
    $portsets = exists($args{portsets}) ? delete($args{portsets}) : undef;
    # cserver object (SanFiler object)
    $cserver = exists($args{cserver}) ? delete($args{cserver}) : undef;
    # Name of the vserver
    $vserver_name = exists($args{vserver_name}) ? delete($args{vserver_name}) : undef;
    # vserver object
    $vserver_obj = exists($args{vserver_obj}) ? delete($args{vserver_obj}) : undef;
    # fetch all igroup objects on this vserver
    $fetched_igroups = exists($args{fetched_igroups}) ? delete($args{fetched_igroups}) : undef;

    my @fetched_igroup_objs = @{$fetched_igroups};

    my $cluster_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err("Unable to parse cvserver name", $cserver->errmsg());
        return [];
    }

    # FIRST, get the nacl cluster object
    # the $cluster_name could either be a node or a cluster name
    my ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                                'names' => [$cluster_name],
                                                'obj_class' => 'nacl');

    if ($self->error()) {
        $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
        return [];
    }

    my @fetched_portset_objs;
    try {
        @fetched_portset_objs = NACL::CS::LunPortset->fetch(command_interface => $nacl_cluster_obj,
                                                        filter => {vserver => $vserver_name},
                                                        allow_empty => '1');
    } otherwise {
        my $exception = shift;
        $self->_set_err("Exception caught while fetching portsets on vserver $vserver_name" .
                        $exception->text());
    };
    if ($self->error()){
        return [];
    }

    # Get portset data from the resource data structure
    foreach my $portset (@{$portsets}) {
        # Normalize params and get local policies
        %portset_args   = $self->_process_common_args(%$portset);
        $cleanup_policy = delete $portset_args{cleanup_policy};
        $reuse_policy   = delete $portset_args{reuse_policy};

        # Get portset name
        if (defined($portset_args{name})) {
            $name = $portset_args{name};
            $portset_args{portset} = delete $portset_args{name};
        } else {
            $self->_set_err("Portset name is missing!");
            return [];
        }

        $self->sanlog(-msg => $self->_format_log("Setting up portset $name"));

        # Check if the portset exists.
        my $portset_exists = 0;
        my $current_portset = undef;
        foreach my $fetched_portset_obj (@fetched_portset_objs) {
            if ($fetched_portset_obj->portset() eq $name) {
                $portset_exists = 1;
                $current_portset = $fetched_portset_obj;
                last;
            }
        }

        # Remove igroups from hash. Binding is done later.
        my @igroup_names = ();
        foreach my $key (keys(%portset_args)) {
            # For multiple igroup_binds there will be a digit on the end
            if ($key =~ /igroup_bind(\d+)/i) {
                $igroup = delete $portset_args{'igroup_bind' . $1};
                $igroup_name = $igroup->{NAME};
                push(@igroup_names, $igroup_name);
            } elsif ($key =~ /igroup_bind/i) {
                $igroup = delete $portset_args{igroup_bind};
                $igroup_name = $igroup->{NAME};
                push(@igroup_names, $igroup_name);
            }
        }

        # Create or reuse portset according to the policies
        if ($portset_exists) {
            if ($reuse_policy =~ /^reuse$/i) {
                # We need the cserver name to get the nacl cluster objects from SanMgr

                my @lif_names_in_portset = $current_portset->port_name();

                # check all lifs are in the portset
                my $port_name_arg;
                if (defined($portset_args{'port-name'})) {
                    $port_name_arg = $portset_args{'port-name'};
                    my @lifs = split(",",$port_name_arg);
                    foreach my $lif (@lifs) {
                        if (grep (/$lif/, @lif_names_in_portset)) {
                            # lif is already there, so noop
                        } else {
                            # lif is not there, so add it
                            try {
                                NACL::C::LunPortset->add(command_interface => $nacl_cluster_obj,
                                                         'port-name' => [$lif],
                                                         vserver => $vserver_name,
                                                         portset => $name);
                            } otherwise {
                                my $exception = shift;
                                $self->_set_err("Unable to add lif '$lif' to portset '$name'", $exception->text());
                            };
                        }
                    }
                }
                if ($self->error()){
                    return [];
                }
                $reused = $SanMgr::TRUE;
            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::FALSE;
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("Portset exists and reuse policy is set to " .
                                "\"fail\"");
                return [];
            }
        } else {
            $self->sanlog(-msg => $self->_format_log("Creating portset $name"));
            $portset_obj = $vserver_obj->create_portset_obj(%portset_args);
            if ($vserver_obj->error()) {
                $self->_set_err("Unable to create portset '$name'",
                                $vserver_obj->errmsg());
                return [];
            }

            $reused = $SanMgr::FALSE;
        }

        # Loop over each igroup and bind them to the appropriate portset
OUTER:  foreach my $igroup_name (@igroup_names) {
            # Bind them to an igroup only if they specified an
            # port-name attribute, otherwise the lib will fail
            if (defined($igroup) and defined($portset_args{'port-name'})) {

                # Check if the igroup exists.
                # extract it's object if we find it in the fetched igroup object list
                my $igroup_exists = $SanMgr::FALSE;

                foreach my $i_obj (@fetched_igroup_objs) {
                    next if ($i_obj->igroup() ne $igroup_name);
                    $igroup_exists = $SanMgr::TRUE;
                    last;
                }

                next if !$igroup_exists; # go to the next igroup if it does not exist

                # Now try binding it
                # check if portset is being reused
                if ($reused) {
                    # portset is being reused
                    # if igroup is already bound to a portset, skip the igroup
                    foreach my $current_igroup ($current_portset->igroups()) {
                        next OUTER if $current_igroup eq $igroup;
                    }
                }

                # otherwise bind this igroup to this portset
                NACL::C::LunIgroup->bind(command_interface => $nacl_cluster_obj,
                                         vserver           => $vserver_name,
                                         igroup            => $igroup_name,
                                         portset           => $name);
            }
        }
        $data = {
            reused  => $reused,
            name    => $name,
            vserver => $vserver_name,
            type    => "portset",
            cleanup => $cleanup_policy
        };
        
        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                                 . "for portset $name"
                                                 . " (reused = $reused)"));
        push(@portset_data, $data);
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up portsets"));

    return \@portset_data;
} ## end sub _setup_portsets

###############################################################################
# Method:    _setup_tpgroups
# Objective: Method responsible for creating or reusing tpgroup objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     tpgroupdata => hash containing all args for a tpgroup
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_tpgroups {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $tpgroups, $data, $vserver_obj, $tpgroup_obj) = undef;
    my ($vserver_name, $name, $reuse_policy, $cleanup_policy) = "";
    my ($tpgroup_exists, $reused) =  $SanMgr::FALSE;
    my (@tpgroup_data, %tpgroup_args) = ();
    my ($iscsi_obj) = undef;

    $self->sanlog(-msg => $self->_format_log("Start setting up tpgroups"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing tpgroup details
    $tpgroups = &ParamsMgr::get_arg(-arg      => "tpgroups",
                                    -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed",
                        &ParamsMgr::get_error());
        return [];
    }

    # Get tpgroup data from the resource data structure
    foreach my $tpgroup (@{$tpgroups}) {
        # Normalize params and get local policies
        %tpgroup_args = $self->_process_common_args(%$tpgroup);
        $cleanup_policy = delete $tpgroup_args{cleanup_policy};
        $reuse_policy = delete $tpgroup_args{reuse_policy};

        # Get tpgroup name
        if (defined($tpgroup_args{name})) {
            $name = $tpgroup_args{name};
            $tpgroup_args{tpgroup} = delete $tpgroup_args{name};
        } else {
            $self->_set_err("Tpgroup name is missing!");
            return [];
        }

        # Get the LIF names if they exist
        if (defined($tpgroup_args{'lif-list'})) {
            $tpgroup_args{'lif'} = delete $tpgroup_args{'lif-list'};
        }

        $self->sanlog(-msg => $self->_format_log("Setting up tpgroup $name"));

        # For tpgroups, we must get a vserver object
        $vserver_obj = $cserver->get_vserver_obj(name => $vserver_name);
        if ($cserver->error()) {
            $self->_set_err("Unable to get vserver object for '$vserver_name'",
                            $cserver->errmsg());
            return [];
        }

        # Check if an iscsi object exists
        unless ($vserver_obj->iscsi_exist()) {
            $self->_set_err("iSCSI needed for tpgroups",
                            $vserver_obj->errmsg());
            return [];
        }

        # Now get the iscsi object to see if this tpgroup exists.
        $iscsi_obj = $vserver_obj->get_iscsi_obj();
        $vserver_obj->_clear_err();

        $tpgroup_exists = $iscsi_obj->tpgroup_exist(tpgroup => $name);
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed to check if tpgroup exists",
                            $iscsi_obj->errmsg());
            return [];
        }

        # Create or reuse igroups according to the policies
        if ($tpgroup_exists) {
            if ($reuse_policy =~ /^reuse$/i) {
                $tpgroup_obj = $iscsi_obj->get_tpgroup_obj(tpgroup => $name);
                if ($iscsi_obj->error()) {
                    $self->_set_err("Unable to get tpgroup objec for $name",
                                    $iscsi_obj->errmsg());
                    return [];
                }

                $self->sanlog(-msg => $self->_format_log("Using existing tpgroup $name"));
                $reused = $SanMgr::TRUE;
            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::FALSE;
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("Tpgroup exists and reuse policy is set to " .
                                "\"fail\"");
                return [];
            }
        } else {
            $self->sanlog(-msg => $self->_format_log("Creating tpgroup $name"));
            $tpgroup_obj = $iscsi_obj->create_tpgroup_obj(%tpgroup_args);
            if ($iscsi_obj->error()) {
                $self->_set_err("Unable to create tpgroup $name",
                                $iscsi_obj->errmsg());
                return [];
            }
            $reused = $SanMgr::FALSE;
        }

        $data = {
            reused  => $reused,
            name    => $name,
            vserver => $vserver_name,
            type    => "tpgroup",
            cleanup => $cleanup_policy
        };

        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                               . "for tpgroup $name"
                                               . " (reused = $reused)"));

        push(@tpgroup_data, $data);
    } # end foreach tpgroup

    $self->sanlog(-msg => $self->_format_log("Finished setting up tpgroups"));

    return \@tpgroup_data;
} ## end sub _setup_tpgroups

###############################################################################
#                    VO L U M E S  &  L U N S  M E T H O D S
#
#  Things to consider:
#  1. Volumes are created in parallel by default, unless users disable this
#     mode in the XML. Each vol subtest creates luns serially.
#  2. LUNs will be created serially to avoid creating a large number or logs
#
###############################################################################

###############################################################################
# Method:    _setup_volumes
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing:
#            %args = (cserver => a filer object
#                     volume_data => hash containing all args for a volume
#                     vserver_name => name of a vserver)
# Returns:   A reference to an array of hashes containg name, reuse flag,
#            cleanup and type of obj.
###############################################################################
sub _setup_volumes {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vserver, $cserver, $volumes, $vserver_obj) = undef;
    my ($nacl_cluster_obj) = undef;
    my ($cserver_name, $vserver_name, $aggregate) = "";
    my (@volume_data, @subtests, @messages) = ();
    my ($parallel, $st_status) = $SanMgr::TRUE;

    $self->sanlog(-msg => $self->_format_log("Start setting up volumes"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing volume details
    $volumes = &ParamsMgr::get_arg(-arg      => "volumes",
                                   -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Vserver object
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);

    # The nacl_cluster_obj
    $nacl_cluster_obj = &ParamsMgr::get_arg(-arg      => "nacl_cluster_obj",
                                            -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a volume
    $cserver_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err("Could not get cserver name", $cserver->errmsg());
        return [];
    }

    # fetch all volume objects
    my @fetched_vol_objs = $cserver->fetch_volume_objs(-vserver => $vserver_name);
    if ($cserver->error()) {
        $self->_set_err("Could not get fetch volume objects", $cserver->errmsg());
        return [];
    }

    # Go through all volumes and create them. By default, SanMgr will create
    # them in parallel.
    foreach my $volume (@{$volumes}) {
        # Get all the luns, qtrees and options contained on each volume
        my @luns    = ();
        my @qtrees  = ();
        my @options = ();
        my @efficiency_options = ();
        my @snapshot_autodelete_options =();

        foreach my $key (keys(%{$volume})) {
            # We'll need a volume name later, piggy back it here
            if ($key =~ /^LUN\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@luns,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "lun"));
            } elsif ($key =~ /^QTREE\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@qtrees,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "qtree"));
            } elsif ($key =~ /^OPTIONS$/i) {
                push(@options,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "options"));
            } elsif ($key =~ /^EFFICIENCY_OPTIONS$/i) {
                push(@efficiency_options,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "efficiency_options"));
            } elsif ($key =~ /^SNAPSHOT_AUTODELETE_OPTIONS$/i) {
                push(@snapshot_autodelete_options,
                     $self->_generate_copies(element => $volume->{$key},
                                             type    => "snapshot_autodelete_options"));
            }
        }

        # if there's only one volume, hardcode parallel to 1
        # this will force the subtest to reuse the connection rather than create a new one
        if (scalar(@{$volumes}) == 1) {
            $parallel = 1;
        } else {
            # Initialize parallel flag just in case we have a set of mixed tags
            if (defined($volume->{PARALLEL})) {
                $parallel = delete $volume->{PARALLEL};
            } else {
                $parallel = $SanMgr::TRUE;
            }
        }

        # if the user does not want any parallelism on this current volume, reap any existing
        # setup subtests if they exist
        if ($parallel eq 0 and scalar(@subtests)) {
            push(@volume_data, $self->_reap_volume_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A volume setup subtest failed");
                return [];
            }
            @subtests = ();
        }

        # if there exists one, find the volume object we plan to operate on
        # in the fetched volume objs list
        my $vol_obj = undef;
        foreach my $fetched_vol_obj (@fetched_vol_objs) {
            next if ($fetched_vol_obj->get_name(-use_cache => $SanMgr::TRUE) ne $volume->{NAME});
            $vol_obj = $fetched_vol_obj;
            last;
        }

        # Append the aggregate name needed to create the object
        $volume->{AGGREGATE} = $aggregate;

        my $subtest = &SubIPC::create(
            -id => "setup_volume__" . $volume->{NAME} . "__on_vserver__" .
                $vserver_name,
            -run => sub {
                $self->_setup_volume(cserver              => $cserver,
                                     vserver_name         => $vserver_name,
                                     vserver_obj          => $vserver_obj,
                                     volume               => $volume,
                                     aggregate            => $aggregate,
                                     vol_obj              => $vol_obj,
                                     parallel             => $parallel,
                                     luns                 => \@luns,
                                     qtrees               => \@qtrees,
                                     efficiency_options   => \@efficiency_options,
                                     options              => \@options,
                                     snapshot_autodelete_options => \@snapshot_autodelete_options,
                                     nacl_cluster_obj     => $nacl_cluster_obj);
                if ($self->error()) {
                    # Error is already set, bubble it up
                    return [];
                }
            },
            -ipc => $SanMgr::TRUE,
            -bg => $SanMgr::TRUE
        );
        if (&SubIPC::error()) {
            $self->_set_err("Unable to dispatch subtest", &SubIPC::get_error());
            return [];
        }
        push(@subtests, $subtest);

        # reap the subtest list
        # if the user doesnt want max prallelism and
        #### the user doesnt want any parallelism on this volume
        #### or
        #### we have hit a max volume parallel setup count
        if ($parallel !~ /^max$/i and ($parallel eq 0  or scalar(@subtests) == $parallel)) {
            push(@volume_data, $self->_reap_volume_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A volume setup subtest failed");
                return [];
            }
            @subtests = ();
        }

    } # end foreach volume

    # If some volumes subtests are still out there, wait for them and read messages
    if (scalar(@subtests)) {
        push(@volume_data, $self->_reap_volume_setup_subtests(-subtests => \@subtests));
        if ($self->error()) {
            $self->_set_err("A volume setup subtest failed");
            return [];
        }
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up volumes"));

    return \@volume_data;
} ## end sub _setup_volumes

###############################################################################
# Method:    _reap_volume_setup_subtests
# Objective: Method responsible for Reaping volume setup subtests
# Inputs:    A hash containing:
#            %args = (subtest => an array of SubIPC subtests)
# Returns:   An array or array reference of volume data return by all the
#            subtests
###############################################################################
sub _reap_volume_setup_subtests {
    my ($self, %args) = @_;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    my @subtests = &ParamsMgr::get_arg(-arg => "subtests",
                                       -ref_type => "ARRAY",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("Failed while reaping volume setup subtests",
                        &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @messages = &SubIPC::get_messages(-type => "setup_volume_complete");
    if (&SubIPC::error()) {
        $self->_set_err("Unable to get messages 'setup_volume_complete'",
                        &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @volume_data;

    foreach my $message (@messages) {
        if (defined($message->{extra}->{obj_data})) {
            # Log the entry we are adding to the cleanup stack
            my $dumpout = Dumper($message->{extra}->{obj_data});
            $self->sanlog(-msg => "Dumper of volume data getting added to"
                              . " cleanup stack:\n"
                                  . $dumpout,
                          -msg_level  => 7);
            push(@volume_data, @{$message->{extra}->{obj_data}});
        }
    }

    # Get the status of a subtest and report it
    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != $SanMgr::FALSE) {
        $self->sanlog(-msg => "Found a volume subtest failure:" .
                          "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->_set_err("Failed to setup volumes");
        return (wantarray()) ? () : [];
    } else {
        $self->sanlog(-msg => "All volume subtests have passed:" .
                          "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->sanlog(-msg => $self->_format_log("Finished volume setup"));
    }

    # Do some house keeping and remove all vserver subtests.
    &SubIPC::remove_subtest(subtest => \@subtests) if (scalar(@subtests));

    return (wantarray()) ? @volume_data : \@volume_data;
} ## end sub _reap_volume_setup_subtests

###############################################################################
# Method:    _setup_volume
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, subtested flag, luns and volume options.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _setup_volume {

    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data, $volume_obj, $volume) = undef;
    my ($name, $vserver_name, $aggregate, $percent_snapshot_space, $parallel) = "";
    my ($cleanup_policy, $reuse_policy) = "";
    my (%volume_args, %create_args, @volume_data) = ();
    my ($volume_exists, $reused) =  $SanMgr::FALSE;
    my ($snapshot_policy) = "";
    my ($luns, $options, $lun_data, $efficiency_options, $qtrees, $qtree_data) = undef;
    my ($autosize_percent, $autosize, $snapshot_autodelete_options) = undef;
    my ($nacl_cluster_obj) = undef;
    my ($size, $max_autosize,$autosize_percent_used) = 0;
    my ($enable_efficiency) = -1; # by default we won't touch efficiency

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Name of vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Vserver object
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Volume
    $volume = &ParamsMgr::get_arg(-arg      => "volume",
                                  -required => $SanMgr::TRUE);

    # Volume Object (if it pre-exists)
    $volume_obj = &ParamsMgr::get_arg(-arg      => "vol_obj",
                                      -default  => undef);

    # Parallel count
    $parallel = &ParamsMgr::get_arg(-arg      => "parallel",
                                    -required => $SanMgr::TRUE);

    # Luns on this volume
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # Qtrees on this volume
    $qtrees = &ParamsMgr::get_arg(-arg      => "qtrees",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # efficiency_options on this volume
    $efficiency_options = &ParamsMgr::get_arg(
                                -arg      => "efficiency_options",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # snapshot_autodelete_options on this volume
    $snapshot_autodelete_options = &ParamsMgr::get_arg(
                                -arg      => "snapshot_autodelete_options",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # nacl cluster object
    $nacl_cluster_obj = &ParamsMgr::get_arg(-arg      => "nacl_cluster_obj",
                                            -default  => undef);

    # Volume Options
    $options = &ParamsMgr::get_arg(-arg      => "options",
                                   -required => $SanMgr::FALSE,
                                   -default  => []);

    # Name of the aggregate
    $aggregate = &ParamsMgr::get_arg(-arg      => "aggregate",
                                     -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Check if users specify snapshot policy
    if (defined($volume->{'SNAPSHOT-POLICY'})) {
        $snapshot_policy = $volume->{'SNAPSHOT-POLICY'};
    }

    # Check if users specify a %snapshot space
    if (defined($volume->{PERCENT_SNAPSHOT_SPACE})) {
        $percent_snapshot_space = delete ($volume->{PERCENT_SNAPSHOT_SPACE});
    }

    # Check if users specify the efficiency attribute
    if (defined($volume->{EFFICIENCY})) {
        $enable_efficiency = delete ($volume->{EFFICIENCY});
    }

    if (defined($volume->{'AUTOSIZE-PERCENT'})) {
        $autosize_percent = delete ($volume->{'AUTOSIZE-PERCENT'});
    }

    if (defined($volume->{'AUTOSIZE'})) {
        $autosize = $volume->{'AUTOSIZE'};
    }

    # Check if users specify the max-autosize attribute. If they haven't, then
    # check for autosize-percent. max-autosize has precedence and
    # autosize percent will be ignored if it is passed in.
    # If using autosize percent, we correct the values of size and
    # max-autosize according to the percentage given, unless autosize
    # is false.
    if (defined($volume->{'MAX-AUTOSIZE'})) {

        $max_autosize = $volume->{'MAX-AUTOSIZE'};

         # Log a warning since both max-autosize and autosize-percent were
         # used as parameters, autosize will be disregarded.
         if(defined($autosize_percent)) {
            $self->sanlog(-msg => "WARNING: Both max-autosize and " .
                                      "autosize-percent were passed in as " .
                                      "parameters. autosize-percent will" .
                                      "be disregarded.");
         }

    # If autosize is false, do not calculate the autosize percentage.
    } elsif (defined($autosize_percent)
        and ($autosize eq "true")) {

        $size = $volume->{SIZE};
        my ($size_number, $size_quantifier) = ($size =~ /^(\d*)(\w*)$/);

        # If autosize-percent is less than 100, set the max-autosize param
        # to size and size becomes the autosize percentage of max-autosize
        # if autosize-percent happens to be greater than 100, the max-autosize
        # grows to be that percentage of the size param
        if ($autosize_percent <= 100) {
            $volume->{'MAX-AUTOSIZE'} = $size;
            $volume->{SIZE} = (($size_number
                / (100 / ($autosize_percent)) . $size_quantifier));
        } else {
            $volume->{'MAX-AUTOSIZE'} = (($size_number
                * ($autosize_percent / 100)) . $size_quantifier);
        }

        $autosize_percent_used = "true";

    }

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.

    if (defined($volume->{NAME})) {
        $name = delete $volume->{NAME};
        $volume->{VOLUME} = $name;
    } else {
        $self->_set_err("Missing volume name. Check your XML or params.");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    $self->sanlog(-msg => $self->_format_log("Setting up volume $name"));

    # Append the vserver name. Needed by the library to create a volume
    $volume->{VSERVER} = $vserver_name;

    # Normalize params and get local policies
    %volume_args    = $self->_process_common_args(%$volume);
    $cleanup_policy = delete $volume_args{cleanup_policy};
    $reuse_policy   = delete $volume_args{reuse_policy};

    # Remove extra params. Otherwise the vserver library will fail.
    foreach my $key (keys(%volume_args)) {
        if ($key =~ /PARALLEL|SERIAL/i) {
            delete $volume_args{$key};
        } elsif ($key =~ /LUN|OPTIONS|QTREE|QOS_POLICY_GROUP\d*|use_default_qos_policy_group/i) {
            next;
        } else {
            $create_args{$key} = $volume_args{$key};
        }
    }

    # Replace/Create the qos-policy-group creation argument if
    # 'use_default_qos_policy_group' is true(1).
    if ($volume_args{'use_default_qos_policy_group'}) {
        $create_args{'qos-policy-group'} = $vserver_name . "_" .
                                           $name . "_pg";
    }

    # only reconnect if parallel is > 1
    if ($parallel =~ /max/i or $parallel > 1) {
        # Need to use a new connection on each vserver
        $cserver->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
        if ($cserver->error()) {
            $self->_set_err("Unable to reconnect to cserver", $cserver->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Check if the volume exists.
    if (defined($volume_obj)) {
        $volume_exists = $SanMgr::TRUE;
    }

    if ($volume_exists) {

        if ($reuse_policy =~ /^reuse$/i) {

            $self->sanlog(-msg => $self->_format_log("Using existing " .
                " volume $name"));
            $reused = $SanMgr::TRUE;

            # Check to see that the volume is online
            my $state = $volume_obj->get_state(-use_cache => $SanMgr::TRUE);
            if ($volume_obj->error()) {
                $self->_set_err("Unable to get volume $name state",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # Get volume autosize attribute or fail on error.
            my $current_autosize =
                $volume_obj->get_autosize(-use_cache => $SanUtil::TRUE);
            if ($volume_obj->error()) {
                $self->_set_err("Unable to get volume $name autosize",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            # Get volume max-autosize attribute or fail on error.
            my $current_max_autosize =
                $volume_obj->get_max_autosize(-use_cache => $SanUtil::TRUE);
            if ($volume_obj->error()) {
                $self->_set_err("Unable to get volume $name max-autosize",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

             # Get volume size attribute or fail on error.
            my $current_size =
                $volume_obj->get_size(-use_cache => $SanUtil::TRUE);
            if ($volume_obj->error()) {
                $self->_set_err("Unable to get volume $name size",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }

            my $state_modify_args = {};
            # Only change autosize values if the current
            # volume size matches the parameter.
            if ((defined($volume_args{'size'}))
                 and ($self->_size_arg_is_close($current_size, $volume_args{'size'}))) {

                if ((defined($volume_args{'autosize'})
                    and ($current_autosize ne $volume_args{'autosize'}))) {
                    $state_modify_args->{'autosize'} = $volume_args{'autosize'};
                }

                # Only change max-autosize on reuse if autosize percent
                # is not used to calculate max-autosize
                if ($autosize_percent_used ne "true") {

                    # New max-autosize from param
                    if (defined($volume_args{'max-autosize'})
                        and ($current_max_autosize
                            != $volume_args{'max-autosize'})) {

                        $state_modify_args->{'max-autosize'}
                            = $volume_args{'max-autosize'};
                    }
                }

            } else {

                # Log a warning since configured volume size didn't match
                # the param vol size, the volume size will not be manipulated.
                $self->sanlog(-msg => "WARNING: The current volume size does " .
                                      "not match the param volume size. " .
                                      "Therefore, thin provisioning " .
                                      "settings will not be changed.");

            }

            if ($state =~ /^offline$/) {
           $state_modify_args->{'-state'} = "online";
            }

            if (keys %$state_modify_args) {

                $self->sanlog(-msg => $self->_format_log(
                       "Volume $name needs modification, " .
                       "updating thin provisioning and state"));
                $volume_obj->modify(%$state_modify_args);
                if ($volume_obj->error()) {
                    $self->_set_err("Unable to modify volume $name",
                                    $volume_obj->errmsg());
                    &SubIPC::fail(-object => $self);
                    return $SanMgr::FALSE;
                }
            }

        } elsif ($reuse_policy =~ /^check$/i) {
            $reused = $SanMgr::FALSE;
            # TODO
        } elsif ($reuse_policy =~ /^fail$/i) {
            $reused = $SanMgr::FALSE;
            $self->_set_err("Volume $name already exists and Reuse policy is set to \"fail\"");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

    } else {
        $self->sanlog(-msg => $self->_format_log("Creating volume $name"));

        # create the volume, force this method to not perform unnecessary verifications
        $volume_obj = $cserver->create_volume(%create_args,
                                              vserver_obj => $vserver_obj,
                                              verify_vol => $SanMgr::FALSE);
        if ($cserver->error()) {
            $self->_set_err("Unable to create volume $name", $cserver->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        $reused = $SanMgr::FALSE;
    }

    # Only set percent snapshot space when snapshot policy is default
    if ($snapshot_policy eq "default" and length($percent_snapshot_space)) {
        $volume_obj->modify(percent_snapshot_space => $percent_snapshot_space);
        if ($volume_obj->error()) {
            $self->_set_err("Unable to modify percent_snapshot_space for volume $name",
                            $volume_obj->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    if ($enable_efficiency == 1) {
        $self->sanlog(-msg => "Turning on efficiency for vol $name");
        $volume_obj->efficiency_on();
        if ($volume_obj->error()) {
            # If the failure occured because it's already on, it's not a failure
            my $errmsg = join "\n", $volume_obj->errmsg();
            unless ($errmsg =~ /Operation has already been enabled|Operation is currently active/) {
                $self->_set_err("Failed to turn on efficiency for vol $name: ",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
           }
        }

        # Check to see if the user sent in efficiency options, if they
        # did call volume::efficiency_modify() with what we were sent.
        if (defined ($efficiency_options) and (scalar(@{$efficiency_options}) > 0)) {
            my $modify_args;
            foreach my $current_options (@{$efficiency_options}) {
                if (defined  $current_options->{"COMPRESSION"}) {
                    $modify_args->{"-compression"} = $current_options->{"COMPRESSION"};
                }
                if (defined  $current_options->{"SCHEDULE"}) {
                    $modify_args->{"-schedule"} = $current_options->{"SCHEDULE"};
                }
                if (defined  $current_options->{"POLICY"}) {
                    $modify_args->{"-policy"} = $current_options->{"POLICY"};
                }
                if (defined  $current_options->{"MIN-BLOCKS-SHARED"}) {
                    $modify_args->{"-min-blocks-shared"} = $current_options->{"MIN-BLOCKS-SHARED"};
                }
                if (defined  $current_options->{"INLINE-COMPRESSION"}) {
                    $modify_args->{"-inline-compression"} = $current_options->{"INLINE-COMPRESSION"};
                }
                if (defined  $current_options->{"IDD"}) {
                    $modify_args->{"-idd"} = $current_options->{"IDD"};
                }
                if (defined  $current_options->{"QUICK-CHECK-FSIZE"}) {
                    $modify_args->{"-quick-check-fsize"} = $current_options->{"QUICK-CHECK-FSIZE"};
                }
            } # end foreach

            $volume_obj->efficiency_modify(%{$modify_args});
            if ($volume_obj->error()) {
                my $dumpout = Dumper($modify_args);
                $self->_set_err("Failed to modify efficiency for vol $name: ",
                                $dumpout,
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
    } elsif ($enable_efficiency == 0) {

        # If volume efficiency is supposed to be off we will try to turn
        # it off. For an exisitng volume this might pass or fail.  For a
        # new volume it'll always fail.  We check the error message to see
        # if it tells us that the option isn't configured.  If it's disabled
        # or not configured, we ignore the error.
        $self->sanlog(-msg => "Turning off efficiency for vol $name");
        $volume_obj->efficiency_off();
        if ($volume_obj->error()) {
            my $errmsg = join "\n", $volume_obj->errmsg();
            unless (($errmsg =~ /Operation has already been disabled/) or
                    ($errmsg =~ /Operation is not configured|Failed to disable efficiency/)) {
                $self->_set_err("Failed to turn off efficiency for vol $name: ",
                                $volume_obj->errmsg());
                &SubIPC::fail(-object => $self);
                return $SanMgr::FALSE;
            }
        }
    }

    # Check to see if the user sent in snapshot autodelete options, if they
    # did call NACL.
    if (defined ($snapshot_autodelete_options) and (scalar(@{$snapshot_autodelete_options}) > 0)) {
        my $modify_args;
        foreach my $current_option (@{$snapshot_autodelete_options}) {
            if (defined  $current_option->{"ENABLED"}) {
                $modify_args->{"enabled"} = $current_option->{"ENABLED"};
            }
            if (defined  $current_option>{"COMMITMENT"}) {
                $modify_args->{"commitment"} = $current_option->{"COMMITMENT"};
            }
            if (defined  $current_option->{"DEFER-DELETE"}) {
                $modify_args->{"defer-delete"} = $current_option->{"DEFER-DELETE"};
            }
            if (defined  $current_option->{"TARGET-FREE-SPACE"}) {
               $modify_args->{"target-free-space"} = $current_option->{"TARGET-FREE-SPACE"};
            }
            if (defined  $current_option->{"TRIGGER"}) {
                $modify_args->{"trigger"} = $current_option->{"TRIGGER"};
            }
            if (defined  $current_option->{"DESTROY-LIST"}) {
                $modify_args->{"destroy-list"} = $current_option->{"DESTROY-LIST"};
            }
        } # end foreach

        $modify_args->{volume} = $name;
        $modify_args->{vserver} = $vserver_name;
        $modify_args->{command_interface} = $nacl_cluster_obj;

        try {
            NACL::C::VolumeSnapshotAutodelete->modify(%{$modify_args});
        } otherwise {
            my $exception = shift;
            $self->_set_err("Failed to modify snapshot autodelete options for vol $name." .
                            " Caught exception " . $exception->text());
        };
        if ($self->error()){
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # NOTE: appending aggregate name because it's needed for cleanup purposes
    $data = {
        reused    => $reused,
        name      => $name,
        aggregate => $aggregate,
        vserver   => $vserver_name,
        type      => "volume",
        parallel  => $volume->{PARALLEL},
        cleanup   => $cleanup_policy
    };

    $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                           . "for volume $name"
                                           . " (reused = $reused)"));
    push(@volume_data, $data);

    # Create qtrees serially
    $qtree_data = $self->_setup_qtrees(vserver_name => $vserver_name,
                                       cserver      => $cserver,
                                       qtrees       => $qtrees,
                                       volume_obj   => $volume_obj,
                                       volume_name  => $name,
                                       vserver_obj  => $vserver_obj);
    if ($self->error()) {
        return $self->_return_if_subtest(subtested => $SanMgr::TRUE,
                                         status    => "fail",
                                         ret_value => [],
                                         msg       => $self->get_error());
    }
    push(@volume_data, @{$qtree_data});

    # Now create luns serially - Subtesting hundreds of luns will kill a client!
    # Todo: Like the volumes we can subtest these in groups
    $lun_data = $self->_setup_luns(volume       => $volume_obj,
                                   vserver_obj  => $vserver_obj,
                                   cserver      => $cserver,
                                   luns         => $luns);
    if ($self->error()) {
        $self->_set_err("Unable to setup luns for vserver '$vserver_name' for volume '$name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # NOTE: Pushing luns on a single-dimension list to simplify get_objects.
    push(@volume_data, @{$lun_data});

    # only disconnect if parallel is > 1
    if ($parallel =~ /max/i or $parallel > 1) {
        # disconnect the cserver connection
        $cserver->disconnect_conn();
    }

    return $self->_return_if_subtest(subtested => $SanMgr::TRUE,
                                     status => "pass",
                                     ret_value => \@volume_data,
                                     write_parent => $SanMgr::TRUE,
                                     write_msg_type => "setup_volume_complete",
                                     msg => "Volume created successfully");
} ## end sub _setup_volume

###############################################################################
# Method:    _setup_qtrees
# Objective: Method responsible for creating or reusing qtree objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, luns and volume options.
# Returns:   An array of hashes containing qtree data.
###############################################################################
sub _setup_qtrees {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vserver, $cserver, $volume, $qtrees, $vserver_obj) = undef;
    my ($cserver_name, $vserver_name, $volume_name)         = "";
    my (@qtree_data, @subtests, @messages)    = ();
    my ($parallel, $st_status) = $SanMgr::TRUE;

    $self->sanlog(-msg => $self->_format_log("Start setting up qtrees"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing volume details
    $qtrees = &ParamsMgr::get_arg(-arg      => "qtrees",
                                  -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);
    # The vserver obj
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);
    # Name of the volume
    $volume_name = &ParamsMgr::get_arg(-arg      => "volume_name",
                                       -required => $SanMgr::TRUE);
    # Volume Object
    $volume = &ParamsMgr::get_arg(-arg      => "volume_obj",
                                  -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a qtree
    $cserver_name = $cserver->get_name(-use_cache => $SanMgr::TRUE);
    if ($cserver->error()) {
        $self->_set_err("Could not get cserver name", $cserver->errmsg());
        return [];
    }

    # fetch all the current qtree objects in this volume
    my @fetched_qtree_objs = $cserver->fetch_qtree_objs(-vserver => $vserver_name,
                                                        -volume => $volume_name);
    if ($cserver->error()) {
        $self->_set_err("failed to fetch all qtree objects "
                        . "in vserver '$vserver_name'",
                        $cserver->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # Go through all qtrees and create them. By default, SanMgr will create
    # them in parallel.
    foreach my $qtree (@{$qtrees}) {
        $self->sanlog(-msg => $self->_format_log("Setting up qtree "
                                                 . $qtree->{NAME}));
        # Get all the luns contained in each qtree
        my @luns    = ();
        foreach my $key (keys(%{$qtree})) {
            # We'll need a qtree and volume name later, piggy back it here
            if ($key =~ /^LUN\d*$/i) {
                $qtree->{$key}->{VOLUME} = $volume->get_name();
                $qtree->{$key}->{QTREE}  = $qtree->{NAME};
                push(@luns,
                     $self->_generate_copies(element => $qtree->{$key},
                                             type    => "lun"));
            }
        }

        # if there's only one qtree, hardcode parallel to 1
        # this will force the subtest to reuse the connection rather than create a new one
        if (scalar(@{$qtrees}) == 1) {
            $parallel = 1;
        } else {
            # Initialize parallel flag just in case we have a set of mixed tags
            if (defined($qtree->{PARALLEL})) {
                $parallel = delete $qtree->{PARALLEL};
            } else {
                $parallel = $SanMgr::TRUE;
            }
        }

        # if the user does not want any parallelism on this current volume, reap any existing
        # setup subtests if they exist
        if ($parallel eq 0 and scalar(@subtests)) {
            push(@qtree_data, $self->_reap_qtree_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A qtree setup subtest failed");
                return [];
            }
            @subtests = ();
        }

        # if there exists one, find the qtree object we plan to operate on
        # in the fetched qtree objs list
        my $qtree_obj = undef;
        foreach my $fetched_qtree_obj (@fetched_qtree_objs) {
            next if ($fetched_qtree_obj->get_name(-use_cache => $SanMgr::TRUE) ne $qtree->{NAME});
            $qtree_obj = $fetched_qtree_obj;
            last;
        }

        my $subtest = &SubIPC::create(
            -id => "setup_qtree__" . $qtree->{NAME}
                 . "__on_vol__"     . $volume->get_name()
                 . "__on_vserver__" . $vserver_name,
            -run => sub {
                $self->_setup_qtree(cserver              => $cserver,
                                    parallel             => $parallel,
                                    vserver_name         => $vserver_name,
                                    vserver_obj          => $vserver_obj,
                                    qtree_obj            => $qtree_obj,
                                    volume               => $volume,
                                    qtree                => $qtree,
                                    luns                 => \@luns);
                if ($self->error()) {
                    # Error is already set, bubble it up
                    return [];
                }
            },
            -ipc => $SanMgr::TRUE,
            -bg  => $SanMgr::TRUE
        );
        if (&SubIPC::error()) {
            $self->_set_err("Unable to dispatch subtest", &SubIPC::get_error());
            return [];
        }
        push(@subtests, $subtest);

        # reap the subtest list
        # if the user doesnt want max prallelism and
        #### the user doesnt want any parallelism on this volume
        #### or
        #### we have hit a max volume parallel setup count
        if ($parallel !~ /^max$/i and ($parallel eq 0  or scalar(@subtests) == $parallel)) {
            push(@qtree_data, $self->_reap_qtree_setup_subtests(-subtests => \@subtests));
            if ($self->error()) {
                $self->_set_err("A qtree setup subtest failed");
                return [];
            }
            @subtests = ();
        }

    } # end foreach qtree

    # If some qtrees subtests are still out there, wait for them and read messages
    if (scalar(@subtests)) {
        push(@qtree_data, $self->_reap_qtree_setup_subtests(-subtests => \@subtests));
        if ($self->error()) {
            $self->_set_err("A qtree setup subtest failed");
            return [];
        }
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up qtrees"));

    return \@qtree_data;
} ## end sub _setup_qtrees

###############################################################################
# Method:    _reap_qtree_setup_subtests
# Objective: Method responsible for Reaping qtree setup subtests
# Inputs:    A hash containing:
#            %args = (subtest => an array of SubIPC subtests)
# Returns:   An array or array reference of qtree data returned by all the
#            subtests
###############################################################################
sub _reap_qtree_setup_subtests {
    my ($self, %args) = @_;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    my @subtests = &ParamsMgr::get_arg(-arg      => "subtests",
                                       -ref_type => "ARRAY",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err("Failed to parse arguments", &ParamsMgr::get_error());
        return (wantarray()) ? () : [];
    }

    &SubIPC::wait_finish(-subtest => \@subtests);
    if (&SubIPC::error()) {
        $self->_set_err("Failed while reaping qtree setup subtests",
                        &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @messages = &SubIPC::get_messages(-type => "setup_qtree_complete");
    if (&SubIPC::error()) {
        $self->_set_err("Unable to get messages 'setup_qtree_complete'",
                        &SubIPC::get_error());
        return (wantarray()) ? () : [];
    }

    my @qtree_data;

    foreach my $message (@messages) {
        if (defined($message->{extra}->{obj_data})) {
            # Log the entry we are adding to the cleanup stack
            my $dumpout = Dumper($message->{extra}->{obj_data});
            $self->sanlog(-msg => "Dumper of qtree data getting added to"
                              . " cleanup stack:\n"
                                  . $dumpout,
                          -msg_level  => 7);
            push(@qtree_data, @{$message->{extra}->{obj_data}});
        }
    }

    # Get the status of a subtest and report it
    my $st_status = &SubIPC::get_subtest_status(-subtest => \@subtests);
    if ($st_status->{failed} != $SanMgr::FALSE) {
        $self->sanlog(-msg => "Found a qtree subtest failure:" .
                          "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->_set_err("Failed to setup qtrees");
        return (wantarray()) ? () : [];
    } else {
        $self->sanlog(-msg => "All qtree subtests have passed:" .
                          "\n\t\tFailed: " . $st_status->{failed} .
                              "\n\t\tPassed: " . $st_status->{passed});
        $self->sanlog(-msg => $self->_format_log("Finished qtree setup"));
    }

    # Do some house keeping and remove all vserver subtests.
    &SubIPC::remove_subtest(subtest => \@subtests) if (scalar(@subtests));

    return (wantarray()) ? @qtree_data : \@qtree_data;
} ## end sub _reap_qtree_setup_subtests

###############################################################################
# Method:    _setup_qtree
# Objective: Method responsible for creating or reusing volume objects.
# Inputs:    A hash containing cserver name, volume data, aggregate name,
#            vserver name, subtested flag, luns and volume options.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _setup_qtree {
    my ($self, %args) = @_;

    $self->clear_error();

    # Locals
    my ($cserver, $vserver_obj, $data, $qtree_obj, $qtree, $volume) = undef;
    my ($name, $vserver_name, $volume_name, $percent_snapshot_space) = "";
    my ($cleanup_policy, $reuse_policy) = "";
    my (%qtree_args, %create_args, @qtree_data, @qtree_objs) = ();
    my ($qtree_exists, $reused, $qtrees_fetched) =  $SanMgr::FALSE;
    my ($luns, $options, $lun_data, $parallel) = undef;

    # Process args and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Name of vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    # Vserver Object
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # A cserver object (filer obj)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Qtree
    $qtree = &ParamsMgr::get_arg(-arg      => "qtree",
                                 -required => $SanMgr::TRUE);

    # parallel flag
    $parallel = &ParamsMgr::get_arg(-arg      => "parallel",
                                    -required => $SanMgr::TRUE);

    # Qtree object (if it pre-exists)
    $qtree_obj = &ParamsMgr::get_arg(-arg      => "qtree_obj",
                                     -default  => undef);

    # Luns on this qtree
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::FALSE,
                                -default  => []);

    # Volume object
    $volume = &ParamsMgr::get_arg(-arg      => "volume",
                                  -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        $self->_return_if_subtest(subtested => $SanMgr::TRUE,
                                  ret_value => [],
                                  status    => "fail",
                                  msg       => &ParamsMgr::get_error());
    }

    # Start by getting the "name" of the object. If this is not defined
    # everything else will fail.

    if (defined($qtree->{NAME})) {
        $name = delete $qtree->{NAME};
        $qtree->{QTREE} = $name;
    } else {
        $self->_set_err("Missing qtree name. Check your XML or params.");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    $volume_name = $volume->get_name(-use_cache => $SanMgr::TRUE);

    $self->sanlog(-msg => $self->_format_log("Setting up qtree $name on"
                                           . " volume $volume_name"));

    # Append the vserver name. Needed by the library to create a qtree
    $qtree->{VSERVER} = $vserver_name;

    # Normalize params and get local policies
    %qtree_args     = $self->_process_common_args(%$qtree);
    $cleanup_policy = delete $qtree_args{cleanup_policy};
    $reuse_policy   = delete $qtree_args{reuse_policy};

    # Remove extra params. Otherwise the vserver library will fail.
    foreach my $key (keys(%qtree_args)) {
        if ($key =~ /PARALLEL|SERIAL/i) {
            delete $qtree_args{$key};
        } elsif ($key =~ /LUN/i) {
            next;
        } else {
            $create_args{$key} = $qtree_args{$key};
        }
    }

    # only reconnect if parallel is > 1
    if ($parallel =~ /max/i or $parallel > 1) {
        # Need to use a new connection on each qtree
        $cserver->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
        if ($cserver->error()) {
            $self->_set_err("Unable to reconnect to cserver", $cserver->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    }

    # Check if the qtree exists.
    if (defined($qtree_obj)) {
        $qtree_exists = $SanMgr::TRUE;
    }

    # If the qtree exists, we need to check the reuse policy so we know if
    # we should fail or continue.
    if ($qtree_exists) {
        if ($reuse_policy =~ /^reuse$/i) {
            $self->sanlog(-msg => $self->_format_log("Using existing qtree $name"));
            $reused = $SanMgr::TRUE;
        } elsif ($reuse_policy =~ /^check$/i) {
            $reused = $SanMgr::FALSE;
            # TODO
        } elsif ($reuse_policy =~ /^fail$/i) {
            $reused = $SanMgr::FALSE;
            $self->_set_err("Qtree $name already exists and Reuse policy "
                          . "is set to \"fail\"");
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }
    } else {
        # otherwise create the qtree
        $self->sanlog(-msg => $self->_format_log("Creating qtree $name"));
        $qtree_obj = $volume->create_qtree_obj(%create_args);
        if ($volume->error()) {
            $self->_set_err("Unable to create qtree $name", $volume->errmsg());
            &SubIPC::fail(-object => $self);
            return $SanMgr::FALSE;
        }

        $reused = $SanMgr::FALSE;
    }

    # NOTE: appending aggregate name because it's needed for cleanup purposes
    $data = {
        reused    => $reused,
        name      => $name,
        volume    => $volume_name,
        vserver   => $vserver_name,
        type      => "qtree",
        parallel  => $qtree->{PARALLEL},
        cleanup   => $cleanup_policy
    };

    $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                           . "for qtree $name"
                                           . " (reused = $reused)"));
    push(@qtree_data, $data);

    # Now create luns serially - Subtesting hundreds of luns will kill a client!
    # Todo: Like the volumes we can subtest these in groups
    $lun_data = $self->_setup_luns(cserver      => $cserver,
                                   volume       => $volume,
                                   luns         => $luns,
                                   qtree_obj    => $qtree_obj,
                                   qtree_name   => $name,
                                   vserver_obj  => $vserver_obj
                             );
    if ($self->error()) {
        $self->_set_err("Unable to setup luns for vserver '$vserver_name' for "
                      . "qtree '$name'");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # NOTE: Pushing luns on a single-dimension list to simplify get_objects.
    push(@qtree_data, @{$lun_data});

    # disconnect the cserver connection if parallel > 1
    if ($parallel =~ /max/i or $parallel > 1) {
        $cserver->disconnect_conn();
    }

    return $self->_return_if_subtest(subtested      => $SanMgr::TRUE,
                                     status         => "pass",
                                     ret_value      => \@qtree_data,
                                     write_parent   => $SanMgr::TRUE,
                                     write_msg_type => "setup_qtree_complete",
                                     msg => "Qtree created successfully");
} ## end sub _setup_qtree

###############################################################################
# Method:    _setup_luns
# Objective: Method responsible for creating or reusing lun objects.
# Inputs:    A hash containing cserver, volume name, vserver name, lun data.
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _setup_luns {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cserver, $luns, $data, $lun_obj, $volume_obj, $allocation_req, $class, $file_path) = undef;
    my ($qtree_obj, $parent_obj, $vserver_obj) = undef;
    my ($vserver_name, $name, $volume_name, $qtree_name, $current_interface, $bind_list) = "";
    my ($reuse_policy, $cleanup_policy) = "";
    my ($reserve_req, $reserve_state, $allocation_state, $state_req) = "";
    my (@lun_data, %lun_args, @fetched_lun_objs, @lun_mapped_show) = ();
    my ($space_alloc) =  $SanMgr::TRUE;
    my ($lun_exists, $reused) =  $SanMgr::FALSE;
    my ($parent_is_qtree, $luns_fetched) = $SanMgr::FALSE;

    $self->sanlog(-msg => $self->_format_log("Start setting up luns"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing luns details
    $luns = &ParamsMgr::get_arg(-arg      => "luns",
                                -required => $SanMgr::TRUE);

    # Volume object
    $volume_obj = &ParamsMgr::get_arg(-arg      => "volume",
                                      -required => $SanMgr::TRUE);

    # Vserver object
    $vserver_obj = &ParamsMgr::get_arg(-arg      => "vserver_obj",
                                       -required => $SanMgr::TRUE);

    # Optional qtree name
    $qtree_name = &ParamsMgr::get_arg(-arg      => "qtree_name",
                                      -required => $SanMgr::FALSE,
                                      -default  => "");

    # Optional qtree obj
    $qtree_obj = &ParamsMgr::get_arg(-arg      => "qtree_obj",
                                     -required => $SanMgr::FALSE,
                                     -default  => undef);

    # cserver object
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::FALSE,
                                   -default  => "");

    # Optioinal class type
    $class = &ParamsMgr::get_arg(-arg      => "class",
                                 -required => $SanMgr::FALSE,
                                 -default  => "");

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return \@lun_data;
    }

    # We need the cserver name to get the nacl cluster objects from SanMgr
    my $cluster_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err("Unable to parse cvserver name", $cserver->errmsg());
        return \@lun_data;
    }

    # FIRST, get the nacl cluster object
    # the $cluster_name could either be a node or a cluster name
    my ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                              'names' => [$cluster_name],
                                              'obj_class' => 'nacl');
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain nacl cluster object for " . $cluster_name);
        return \@lun_data;
    }

    $volume_name  = $volume_obj->get_name(-use_cache => $SanMgr::TRUE);
    $vserver_name = $vserver_obj->get_name(-use_cache => $SanMgr::TRUE);

    # Make sure we have luns to create before we go any further
    if (not scalar(@{$luns})) {
        if (defined($qtree_name) and length($qtree_name)) {
            $self->sanlog(-msg => $self->_format_log(
              "No luns to setup on volume $volume_name and qtree $qtree_name"));
        } else {
            $self->sanlog(-msg => $self->_format_log(
              "No luns to setup on volume $volume_name"));
        }
        return \@lun_data;
    }

    # If the caller sent an optional qtree_name then we need to get the
    # qtree object from the volume object.  We are going to set the parent
    # object to either the qtree object or the volume object and use it
    # to perform get/create operations.
    if (defined($qtree_name) and length($qtree_name)) {
        $self->sanlog(-msg => $self->_format_log(
               "Setting up luns on volume $volume_name and qtree $qtree_name"));
        $parent_obj = $qtree_obj;
        $parent_is_qtree = $SanMgr::TRUE;

        # fetch the lun objects in this volume & qtree
        @fetched_lun_objs = $cserver->fetch_lun_objs(-vserver => $vserver_name,
                                                     -qtree => $qtree_name,
                                                     -volume => $volume_name);
        if ($cserver->error()) {
            $self->_set_err("Failed to fetch LUN objects", $cserver->errmsg());
            return \@lun_data;
        }

        # obtain the lun mapped show data
        @lun_mapped_show = $vserver_obj->lun_mapped_list_info(-volume => $volume_name,
                                                              -qtree => $qtree_name,
                                                              -cmd_interface => "cli");
        if ($vserver_obj->error()) {
            $self->_set_err("Failed to obtain lun mapped list info", $vserver_obj->errmsg());
            return \@lun_data;
        }
    } else {
        $self->sanlog(-msg => $self->_format_log(
               "Setting up luns on volume $volume_name"));
        $parent_obj = $volume_obj;

        # fetch the lun objects in this volume (need to pass in empty qtree so that it filters qtree luns out)
        @fetched_lun_objs = $cserver->fetch_lun_objs(-vserver => $vserver_name,
                                                     -volume => $volume_name,
                                                     -qtree => "\"\"");
        if ($cserver->error()) {
            $self->_set_err("Failed to fetch LUN objects", $cserver->errmsg());
            return \@lun_data;
        }

        # obtain lun mapped show data
        @lun_mapped_show = $vserver_obj->lun_mapped_list_info(-volume => $volume_name,
                                                              -qtree => "\"\"",
                                                              -cmd_interface => "cli");
        if ($vserver_obj->error()) {
            $self->_set_err("Failed to obtain lun mapped list info", $vserver_obj->errmsg());
            return \@lun_data;
        }
    }

    # Get lun data from the resource data structure
    foreach my $lun (@{$luns}) {
        # Normalize params and get local policies
        %lun_args       = $self->_process_common_args(%$lun);
        $cleanup_policy = delete $lun_args{cleanup_policy};
        $reuse_policy   = delete $lun_args{reuse_policy};
        $state_req      = (defined($lun_args{'state'})) ?
                             delete $lun_args{'state'} : "online";
        $allocation_req = (defined($lun_args{'space-allocation'})) ?
                             delete $lun_args{'space-allocation'} : undef;
        $bind_list = (defined($lun_args{bind_list})) ?
                        delete $lun_args{bind_list} : undef;

        my $global_lun_id = delete $lun_args{'lun-id'};
        my $lun_id_range = delete $lun_args{'lun_id_range'};
        my $require_same_lun_id_per_igroup = delete $lun_args{'require_same_lun_id_per_igroup'};

        # Get lun name
        if (defined($lun_args{name})) {
            $name = $lun_args{name};
            $lun_args{lun} = delete $lun_args{name};
        } else {
            $self->_set_err("Lun name is missing!");
            return \@lun_data;
        }

        # Create the file for creating the lun if the user specified to,
        # add file-path to the args, and create a cleanup obj.
        if(defined $lun_args{create_from_file} and (delete $lun_args{create_from_file}) == 1) {
            if (defined($qtree_name) and length($qtree_name)) {
                $file_path = "/vol/".$volume_name."/".$qtree_name."/file_$name";
            } else {
                $file_path = "/vol/".$volume_name."/file_$name";
            }
            $lun_args{'file-path'} = $file_path;
            try {
                NACL::C::Mkfile->mkfile(command_interface => $nacl_cluster_obj,
                        filename => $file_path,
                        vserver => $vserver_name,
                        determine_appropriate_nodescope_node => 1,
                        size => (delete $lun_args{size}),
                        );
                $reused = $SanMgr::FALSE;
            } otherwise {
                my $e = shift;
                if($e->text() !~ /write failed/) {
                    $self->_set_err("Unable to create file $file_path",
                                    $e->text());
                } elsif ($reuse_policy =~ /^fail$/i) {
                    $reused = $SanMgr::FALSE;
                    $self->_set_err("File exists and reuse policy is set to \"fail\"");
                } else {
                    $reused = $SanMgr::TRUE;
                }
            };
            if ($self->error()) {
                return \@lun_data;
            }

            # Create the cleanup data for this lun and push it on the cleanup stack
            $data = {
                reused   => $reused,
                name     => $file_path,
                vserver  => $vserver_name,
                type     => "file",
                cleanup  => $cleanup_policy
            };

            $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                                   . "for file $file_path"
                                                   . " (reused = $reused)"));
            push(@lun_data, $data);
        }

        # Replace/Create the qos-policy-group creation argument if
        # 'use_default_qos_policy_group' is true(1).
        if ($lun_args{'use_default_qos_policy_group'}) {
            delete $lun_args{'use_default_qos_policy_group'};
            if (defined($qtree_name) and length($qtree_name)) {
                $lun_args{'qos-policy-group'} = $vserver_name . "_" .
                                                $volume_name  . "_" .
                                                $qtree_name   . "_" .
                                                $name         . "_pg";
            } else {
                $lun_args{'qos-policy-group'} = $vserver_name . "_" .
                                                $volume_name  . "_" .
                                                $name         . "_pg";
            }
        }

        $self->sanlog(-msg => $self->_format_log("Setting up lun $name"));
        my @lun_maps = ();

        #get the lun class
        if (defined($lun_args{class})) {
            $self->sanlog(-msg => $self->_format_log("Setting up lun class $class on lun $name"));
        }

        # Get lun maps and remove them from the params
        my $lun_map_id_hash = {};
        foreach my $key (keys(%lun_args)) {
            if ($key =~ /^LUN_MAP\d*$/i) {
                my $lun_map = delete $lun_args{$key};
                if (defined($lun_map->{IGROUP})) {
                    push(@lun_maps, $lun_map->{IGROUP});
                    # store the lun id to use for this igroup
                    if (defined($lun_map->{LUN_ID})) {
                        $lun_map_id_hash->{$lun_map->{IGROUP}} = $lun_map->{LUN_ID};
                    }
                }
            } elsif ($key =~ /^MAP_LIST$/i) {
                my $lun_map_list = delete $lun_args{$key};
                # split the lun map list by commans, and add each entry into the lun_maps array
                my @split_lun_map_list = split(/,/, $lun_map_list);
                push(@lun_maps, @split_lun_map_list);
            } elsif ($key =~ /^QOS_POLICY_GROUP\d*$/i) {
                delete $lun_args{$key};
            }
        }

        # verify that there are no duplicates in the lun_maps array
        if (scalar(@lun_maps)) {
            my %lun_map_dup_hsh;
            foreach (@lun_maps) {
                $lun_map_dup_hsh{$_}++;
            }
            if (scalar(@lun_maps) != scalar(keys(%lun_map_dup_hsh))) {
                $self->_set_err("Detected duplicate lun map entries for "
                              . "lun '$name', please correct your input file");
                return \@lun_data;
            }
        }

        # Check if it already exists. If it does we'll reuse it, upate the current lun obj
        my $lun_obj;
        $lun_exists = $SanMgr::FALSE;
        foreach my $f_lun_obj (@fetched_lun_objs) {
            if ($f_lun_obj->get_name(-use_cache => $SanMgr::TRUE) eq $name) {
                $lun_obj = $f_lun_obj;
                $lun_exists = $SanMgr::TRUE;
                last;
            }
        }

        # Create or reuse luns according to the policies
        if ($lun_exists) {
            if ($reuse_policy =~ /^reuse$/i) {
                $self->sanlog(-msg => $self->_format_log("Using existing lun $name"));
                $reused = $SanMgr::TRUE;

                if ( defined($lun_args{'space-reserve'}) ) {
                    # Get the state of the space-allocation attribute
                    $reserve_req = delete($lun_args{'space-reserve'});
                    $reserve_state = $lun_obj->get_space_resv(-use_cache => $SanMgr::TRUE);
                    if ($lun_obj->error()) {
                        $self->_set_err("Unable to get lun $name space-reserve state" ,
                                        $lun_obj->errmsg());
                        return \@lun_data;
                    }
                    # If requested value for space-reserve is not the actual state, update it
                    if ($reserve_req ne $reserve_state){
                        $lun_obj->set_space_resrv('space-reserve'=> $reserve_req);
                        if ($lun_obj->error()) {
                            $self->_set_err("Unable to set lun $name space-reserve" ,
                                            $lun_obj->errmsg());
                            return \@lun_data;
                        }
                    }
                    # TODO: Create a private method to the compare_update_attribute.
                    #       It is already been done twice (line 3133). Let's avoid
                    #       duplicating code.
                }

            } elsif ($reuse_policy =~ /^check$/i) {
                $reused = $SanMgr::FALSE;
                # TODO
            } elsif ($reuse_policy =~ /^fail$/i) {
                $reused = $SanMgr::FALSE;
                $self->_set_err("Lun exists and reuse policy is set to \"fail\"");
                return \@lun_data;
            }
        } else {
            $lun_obj = $parent_obj->create_lun_obj(%lun_args);
            if ($parent_obj->error()) {
                $self->_set_err("Failed to create lun", $parent_obj->errmsg());
                return \@lun_data;
            }

            $reused = $SanMgr::FALSE;
        }

        # Will always verify if the LUN is actually online unless the user requested
        # to be set to a different state.
        # use the cache only if the lun pre-existed, otherwise we'll have to refresh the obj
        if ($lun_obj->is_online(-use_cache => $lun_exists)) {
            if ($state_req =~ /offline/) {
                $self->sanlog(-msg => $self->_format_log("LUN $name is online, changing ".
                                                         "state to offline upon user request"));
                $lun_obj->offline('lun'    => $name,
                                  'volume' => $volume_name);
                if ($lun_obj->error()) {
                    $self->_set_err("unable to get offline LUN $name",
                                    $lun_obj->errmsg());
                    return \@lun_data;
                }
            } else {
                $self->sanlog(-msg => $self->_format_log("LUN $name is online"));
            }
        } else {# lun is offline
            if($state_req =~ /online/){
                $self->sanlog(-msg => $self->_format_log("Lun $name is offline, " .
                                                         "changing state to online"));
                $lun_obj->online('lun'   => $name,
                                 'volume'=> $volume_name);
                if ($lun_obj->error()) {
                    $self->_set_err("unable to online LUN $name" , $lun_obj->errmsg());
                        return \@lun_data;
                }
            } else {
                $self->sanlog(-msg => $self->_format_log("LUN $name is offline, staying ".
                                                         "offline upon user request"));
            }
        }

        if (defined($allocation_req)) {
            # Get the state of the space-allocation attribute
            $allocation_state = $lun_obj->get_space_allocation();
            if ($lun_obj->error()) {
                $self->_set_err("Unable to get lun $name space-allocation state" ,
                                $lun_obj->errmsg());
                return \@lun_data;
            }
            # If requested value for space-allocation is not the actual state, update it
            if ($allocation_req ne $allocation_state){
                $lun_obj->set_space_allocation('space-allocation'=> $allocation_req);
                if ($lun_obj->error()) {
                    $self->_set_err("Unable to set lun $name space-allocation" ,
                                    $lun_obj->errmsg());
                    return \@lun_data;
                }
            }
        }

        # if there is a global lun id to use for this lun, store that first
        my %lun_map_args;
        if (defined($global_lun_id)) {
            $lun_map_args{'lun-id'} = $global_lun_id;
        }

        # create a random lun id range if the user passed one in
        my $lun_id_range_random;
        if (defined($lun_id_range)) {
            # use the vserver/volume/qtree as the random seed to make this reproducible
            $lun_id_range_random = $self->_generate_lun_id_range($lun_id_range,
                                                                 $vserver_name,
                                                                 $volume_name,
                                                                 $qtree_name,
                                                                 $name);

            # default the global lun id to be the first lun id in the list
            $lun_map_args{'lun-id'} = $lun_id_range_random->[0];
        }

        # if this lun exists:
        # find all igroups that this lun is mapped to and store it in @lun_map_data
        my @lun_map_data;
        if ($lun_exists) {
            my $lun_path = $lun_obj->get_path(-use_cache => $SanMgr::TRUE);
            # parse out the lun map elements that match this lun's path
            @lun_map_data = grep {
                my $elm = $_;
                if ($elm->{path} eq $lun_path) {
                    1;
                } else {
                    0;
                }
            } @lun_mapped_show;
        }

        # check if its necessary to perform each lun map
        my $successful_maps = 0;
        foreach my $igroup (@lun_maps) {
            my $retry = 0;
            my @maps = ();
            if ($lun_exists) {
                # see if this lun is already mapped to this igroup
                @maps = grep { $_->{igroup} =~ /^${igroup}$/ } @lun_map_data;
            }

            # if there is nothing in the maps array, then this lun is not mapped to the igroup
            # so map the lun to it
            if (not scalar(@maps)) {
                # if this igroup was requested to use a specific lun id, then override the lun id argument
                if (defined($lun_map_id_hash->{$igroup})) {
                    $lun_map_args{'lun-id'} = $lun_map_id_hash->{$igroup};
                }
                $lun_obj->map(igroup => $igroup,
                              check_result => 1,
                              %lun_map_args);
                if ($lun_obj->error()) {
                    # (1) if the user did not specify an absolute lun id to use for this igroup (lun_map_id_hash->{$igroup})
                    # (2) we have a lun id range to work with
                    # (3) we havn't exhausted through the lun_id_range yet for this igroup,
                    # then grab another lun id from the list
                    if (not defined($lun_map_id_hash->{$igroup}) and
                            defined($lun_id_range) and
                            $retry < scalar(@{$lun_id_range_random})) {

                        # if we had a previous succesful map and the user requires that the lun use the same lun id for
                        # all igroups in the map list, then bail out
                        if ($successful_maps > 0 and $require_same_lun_id_per_igroup == 1) {
                            $self->_set_err("Was not able to map LUN $name to the all igroups in the map list " .
                                            "to the same lun id '" . $lun_map_args{'lun_id'} . "'", $lun_obj->errmsg());
                            return \@lun_data;
                        }

                        # clear the error on the lun obj
                        $lun_obj->_clear_err();

                        # rotate the lun id range array
                        push(@{$lun_id_range_random}, shift(@{$lun_id_range_random}));

                        # pick a new lun id and increment the retry
                        $lun_map_args{'lun-id'} = $lun_id_range_random->[0];
                        $retry++;
                        redo;
                    } else {
                        $self->_set_err("Failed to map lun", $lun_obj->errmsg());
                        return \@lun_data;
                    }
                } else {
                    $successful_maps++;
                }
            }
        }

        # Create the cleanup data for this lun and push it on the cleanup stack
        $data = {
            reused   => $reused,
            name     => $name,
            lun_maps => \@lun_maps,
            volume   => $volume_name,
            qtree    => $qtree_name,
            vserver  => $vserver_name,
            type     => "lun",
            cleanup  => $cleanup_policy
        };

        $self->sanlog(-msg => $self->_format_log("Adding Cleanup action "
                                               . "for lun $name"
                                               . " (reused = $reused)"));
        push(@lun_data, $data);
    } # end foreach lun

    $self->sanlog(-msg => $self->_format_log("Finished setting up luns"));
    return \@lun_data;
} ## end sub setup_luns

###############################################################################
# Method:    _setup_lunbinds
# Objective: Method responsible for binding vvols to pe.
# Inputs:    A hash containing cserver, volume name, vserver name, lun data. - need to change this
# Returns:   An array of hashes containing volume data.
###############################################################################
sub _setup_lunbinds {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($vserver, $cserver, $volumes, $class, $data, $nacl_cluster_obj, $LunBind_obj) = undef;
    my ($cserver_name, $vserver_name, $aggregate, $qtree, $qtree_name, $volume_name, $lun, $cleanup_policy, $vvol_path, $lun_name) = "";
    my (@volume_data, @subtests, @messages, @lun_binding_show, @lun_bind_data) = ();
    my ($reused) = $SanMgr::FALSE;

    $self->sanlog(-msg => $self->_format_log("Start setting up lun binds"));

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);

    # Main data structure containing volume details
    $volumes = &ParamsMgr::get_arg(-arg      => "volumes",
                                   -required => $SanMgr::TRUE);

    # cserver object (SanFiler object)
    $cserver = &ParamsMgr::get_arg(-arg      => "cserver",
                                   -required => $SanMgr::TRUE);

    # Name of the vserver
    $vserver_name = &ParamsMgr::get_arg(-arg      => "vserver_name",
                                        -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err("Argument parsing failed", &ParamsMgr::get_error());
        return [];
    }

    # Get the cserver name which is needed to create a volume
    $cserver_name = $cserver->get_name();
    if ($cserver->error()) {
        $self->_set_err("Could not get cserver name", $cserver->errmsg());
        return [];
    }

    #Create a NACL cluster object.
    ($nacl_cluster_obj) = $self->get_objects('type' => "node",
                                             'names' => [$cserver_name],
                                             'obj_class' => 'nacl');
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain nacl cluster object for " . $cserver_name);
        return [];
    }

    foreach my $volume (@{$volumes}) {
        my @luns    = ();
        my @qtrees  = ();
        my @options = ();
        my @efficiency_options = ();

        #added in cleanup and reuse for sanmanager cleanup.
        #Second version will implement these methodes.
        $cleanup_policy = "no_aggr";
        $reused = $SanMgr::FALSE;

        #Break everything down into a usable form.
        foreach my $key (keys(%{$volume})) {
            # We'll need a volume name later, piggy back it here
            if ($key =~ /^LUN\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@luns,
                     $self->_generate_copies(element => $volume->{$key}));
            } elsif ($key =~ /^QTREE\d*$/i) {
                $volume->{$key}->{VOLUME} = $volume->{NAME};
                push(@qtrees,
                     $self->_generate_copies(element => $volume->{$key}));
            }elsif ($key =~ /^OPTIONS$/i) {
                push(@options,
                     $self->_generate_copies(element => $volume->{$key}));
            } elsif ($key =~ /^EFFICIENCY_OPTIONS$/i) {
                push(@efficiency_options,
                     $self->_generate_copies(element => $volume->{$key}));
            }
        }
        #If the lun is in a qtree add the qtree name to the lun hash created.
        if (scalar(@qtrees)){
            foreach $qtree (@qtrees){
                foreach my $key (%{$qtree}){
                    if ($key =~ /^LUN\d*$/i) {
                        $qtree->{$key}->{QTREE} = $qtree->{NAME};
                        push(@luns,$self->_generate_copies(element => $qtree->{$key}));
                    }
                }
            }
        }

        #Run through the luns and pick out the information needed.
        foreach $lun (@luns){
            my $lun_name = $lun->{NAME};
            my $qtree_name = "";
            my $volume_name = $volume->{NAME};

            if (exists($lun->{CLASS}) and ($lun->{CLASS} =~ /vvol/i)){

                # remove all spaces from the bind list
                $lun->{BIND_LIST} =~ s/\s+//g;

                # get the list of peps to bind the vvol to
                my @bind_list = split(/,/,$lun->{BIND_LIST});
                my @final_lun_bind_list = ();

                # ensure that there are no duplicate peps in the list.
                if (scalar(@bind_list)) {
                    my %lun_bind_dup_hsh;
                    foreach my $elem (@bind_list) {
                        $lun_bind_dup_hsh{$elem}++;
                    }

                    # error out if the lun bind list for a vvol conatins dups
                    if (scalar(@bind_list) != scalar(keys(%lun_bind_dup_hsh))) {
                        $self->_set_err("Detected duplicate lun bind entries for "
                                      . "lun '$lun_name', please correct your input XML file");
                        return [];
                    } else {
                        # create the path to the vvol if it lives in a qtree.
                        if ($lun->{QTREE}) {
                            $qtree_name = $lun->{QTREE};
                            $vvol_path = "/vol/${volume_name}/${qtree_name}/${lun_name}";
                        } else {
                            # create the vvol path if it is not in a qtree.
                            $vvol_path = "/vol/${volume_name}/${lun_name}";
                        }

                        # bind this vvol to all PEs in the bind list
                        foreach my $pep (@bind_list) {
                            # convert the Protocol endpoint path to an absolute lun path

                            # get rid of spaces
                            $pep =~ s/\s+//g;

                            # get rid of ./ in the bind list param if the user specified it
                            $pep =~ s/^\.\///;

                            if ($pep !~ /^\/vol/) {
                                # $pep is a relative path to a lun
                                # we need to add the vol name to the front of the pe lun
                                # check if it has a qtree in it
                                if (not length($qtree_name)) {
                                    # the current vvol lun is not in a qtree
                                    # this handles implicit and explicit relative qtree pep cases
                                    $pep = "/vol/" . $volume_name . "/" . $pep;
                                } else {
                                    # the current vvol lun is in a qtree
                                    if ($pep =~ /^\.\.\//) {
                                        # remove the ../ from the pep
                                        $pep =~ s/^\.\.\///;
                                        # the user does not want this pep to be in the qtree
                                        # prepend just the vol name
                                        $pep = "/vol/" . $volume_name . "/"  . $pep;
                                    } else {
                                        # this pep is a relative path with an implicit qtree name,
                                        # ie: it does not contain the qtree name, but the vvol is in a qtree
                                        # prepend the curent volume name with the qtree
                                        $pep = "/vol/" . $volume_name . "/" . $qtree_name . "/" . $pep;
                                    }
                                }
                            }

                            # store this protocol endpoint onto the final lun bind list
                            # the final list should contain only the absolute paths
                            push (@final_lun_bind_list, $pep);

                            #NACL calls need try/otherwise blocks to catch errors.
                            try {
                                #do lun binds on a per vvol basis
                                $LunBind_obj = NACL::C::LunBind->create( command_interface => $nacl_cluster_obj,
                                                                         vserver => $vserver_name,
                                                                         'vvol-path' => $vvol_path,
                                                                         'protocol-endpoint-path' => $pep,
                                                                         );
                            } otherwise {
                                my $e = shift;
                                $self->_set_err("NACL Exception caught". $e->text());
                            };
                            if ($self->error()){
                                return [];
                            }
                        }
                    }
                }

                # Create the cleanup data for this lun bind and push it on the cleanup stack
                $data = {
                    reused   => $reused,
                    name     => $lun_name,
                    lun_binds => \@final_lun_bind_list,
                    volume   => $volume_name,
                    qtree    => $qtree_name,
                    vserver  => $vserver_name,
                    type     => "lun_bind",
                    cleanup  => $cleanup_policy
                };

                $self->sanlog(-msg => $self->_format_log("Adding lun bind cleanup action "
                                                       . "for lun $lun->{NAME}"
                                                       . " (reused = $reused)"));
                #push $data on to the lun_bind_data array. Used for cleanup.
                push(@lun_bind_data, $data);
            } else {
                #if the lun is regular do nothing
                $self->sanlog(-msg => $self->_format_log("The Class for lun $lun->{NAME}."),
                              -msg_level => 7);
            }
        }
    }

    $self->sanlog(-msg => $self->_format_log("Finished setting up lun binds"));

    return \@lun_bind_data;
}

###############################################################################
#           HELPER METHODS AND UTILS FOR CONTROLLER CLUSTER SETUP...
###############################################################################

###############################################################################
# Method:    _create_cserver
# Objective: Creates a "cserver" object. NOTE: What this method is really doing
#            is getting all filer objects created when SanMgr built
#            a ControllerCluster object. Then it gets the "cserver" object
#            from the CLUSTER tag and matches it to one of the filer object.
#            This is just a temporary solution and we expect users to specify
#            the node they want to use as a cserver.
# Input:     A hash containing a resource (which is basically a representation
#            of a cluster data structure/hash).
# Returns:   A cserver object (filer object) or an undef reference if something
#            fails and sets the error flag.
###############################################################################
sub _create_cserver {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($controller_cluster_obj, $cserver_obj) = undef;

    # Process parameters
    &ParamsMgr::process_args(%args);

    # Controller Cluster data
    my $resource = &ParamsMgr::get_arg(-arg      => "resource",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return {};
    }

    # Get cluster tag/element
    ($controller_cluster_obj) = $self->get_objects(type   => "cluster",
                                                   -names => [$resource->{NAME}]);
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain cluster object for " . $resource->{NAME});
        return {};
    }

    # Get nodes objects (SanFilers) in the cluster
    $cserver_obj = $controller_cluster_obj->get_cserver_obj();
    if ($controller_cluster_obj->error()) {
        # Bubble up the error message
        $self->_set_err($controller_cluster_obj->errmsg());
        return {};
    }

    # Ensure we've got something, otherwise everything else will blow up!
    unless (defined($cserver_obj)) {
        $self->_set_err("cserver object is not defined.");
        return {};
    }

    return $cserver_obj;
} ## end sub _create_cserver

###############################################################################
# Method:    _get_node_objs
# Objective: What this method is really doing is getting all filer objects
#            created when SanMgr built a ControllerCluster object.
# Input:     A hash containing a resource (which is basically a representation
#            of a cluster data structure/hash).
# Returns:   An array of node objects (filer objects) or an undef reference if something
#            fails and sets the error flag.
###############################################################################
sub _get_node_objs {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my (@controller_cluster, $controller_cluster_obj, $cserver, $name) = undef;
    my (@node_objs) = ();

    # Process parameters
    &ParamsMgr::process_args(%args);

    # Controller Cluster data
    my $resource = &ParamsMgr::get_arg(-arg      => "resource",
                                       -required => $SanMgr::TRUE);
    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        return ();
    }

    # Get cluster tag/element
    ($controller_cluster_obj) = $self->get_objects(type   => "cluster",
                                                   -names => [$resource->{NAME}]);
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Failed to get cluster object for " . $resource->{NAME});
        return ();
    }

    # Get nodes objects (SanFilers) in the cluster
    @node_objs = $controller_cluster_obj->get_node_objs();
    if ($controller_cluster_obj->error()) {
        # Bubble up the error message
        $self->_set_err($controller_cluster_obj->errmsg());
        return ();
    }

    # Ensure there's at least one node
    unless (scalar(@node_objs)) {
        $self->_set_err("Failed to find nodes in the cluster.");
        return ();
    }

    return @node_objs;
}

###############################################################################
# Method:    _size_arg_is_close
# Objective: Determine if the current size of the volume is close enough to the
#             size param to modify. Close enough is +/- 5%.
# Input:     The current size of the volume and the size of the volume
#             described by the input parameter.
# Returns:   0 or 1
###############################################################################
sub _size_arg_is_close {
    my ($self, $current_size, $size_param) = @_;

    my $size_param_in_bytes = &ClientUtil::convert_to_bytes($size_param, "");

    if (($size_param_in_bytes <= ($current_size * 1.05)) and
        ($size_param_in_bytes >= ($current_size * .95))) {
        return 1;
    }

    return 0;
}

##############################################################################
# Method:    setup_intercluster
# Objective: Create the intercluster object to connect the clusters
# Input:     The hash (key intercluster_param) with the details of the cluster's
#           lifs that will be configured. This input is the input expected by
#           Intercluster.pm
#            The cluster object reference.
# Returns:   0 or 1
###############################################################################
sub _setup_intercluster {
    my ($self, %args) = @_;

    $self->clear_error();

    # Local Variables
    my ($cluster_obj,$intercluster_obj) = undef;
    my $resource_ref = '';
    my $intercluster_created = $SanMgr::FALSE;

    # Process agrs and validate mandatory params
    &ParamsMgr::process_args(%args);
    # Main data structure containing aggregate details
    my $cluster_ref = &ParamsMgr::get_arg(-arg => "cluster",
                                          -required => $SanMgr::TRUE);
    my $interclus_params = &ParamsMgr::get_arg(-arg => "interclus_params",
                                               -required => $SanMgr::TRUE);

    if (&ParamsMgr::error()) {
        $self->_set_err(&ParamsMgr::get_error());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # FIRST, get the controller cluster object
    ($cluster_obj) = $self->get_objects('type' => "cluster",
                                        'names' => [$cluster_ref->{NAME}]);
    if ($self->error()) {
        # Error is already set on $self, just return.
        $self->_set_err("Unable to obtain cluster object for " . $cluster_ref->{NAME});
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # SECOND, create the Intercluster object.
    $intercluster_obj = $cluster_obj->intercluster_obj_create(%{$interclus_params});
    if ($cluster_obj->error()) {
        $self->_set_err($cluster_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # THIRD, connect each cluster to it's peer.
    $intercluster_created = $cluster_obj->intercluster_connection_create('intercluster_obj'=> $intercluster_obj);
    if ($cluster_obj->error()) {
        $self->_set_err($cluster_obj->errmsg());
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE;
    }

    # FOURTH, verify that the connection was properly created.
    unless( $intercluster_created ){
        $self->_set_err("Unable to create an intercluster connection");
        &SubIPC::fail(-object => $self);
        return $SanMgr::FALSE
    }

    return $SanMgr::TRUE;
} ## end sub _setup_intercluster

###############################################################################
# Method:    _generate_lun_id_range
# Objective: Given a lun id range, generate an array of all the lun ids, randomized
# based on the vserver vol and qtree name specified
# Returns:   0 or 1
###############################################################################
sub _generate_lun_id_range {
    my ($self, $lun_id_range, $vserver_name, $vol_name, $qtree_name, $lun_name) = @_;

    my @return = ();
    if ($lun_id_range =~ /^(\d+)\-(\d+)$/) {
        my $start = $1;
        my $end = $2;
        for (my $i = $start; $i <= $end; $i++) {
            push(@return, $i);
        }
    }

    # create a unique integer based on the md5 hash of the vserver vol qtree
    my $md5s = Digest::MD5::md5_hex($vserver_name . "__" . $vol_name . "__" . $qtree_name . "__" . $lun_name);
    my $md5strsub = substr($md5s, 0, 8);
    $md5strsub =~ tr/a-f/1-6/;

    # set the seed based on the vserver volume and qtree name
    srand($md5strsub);

    &ClientUtil::randomize_array(\@return);

    return \@return;
}

1;
