# $Id: //depot/prod/test/Rsierranevada/lib/San/ScsiTool/ScsiUtil.pm#3 $
# Copyright (c) 2008 NetApp Inc.
# All rights reserved.
#
## @summary Utility methods for SCSI testing
## @author bhanup@netapp.com
## @status private
## @pod see ScsiTool.pm

package San::ScsiTool::ScsiUtil;

use strict;

# use vars qw(@ISA @EXPORT @EXPORT_OK);
use vars qw(@ISA);

#use Tie::Hash::Indexed;
use Tie::IxHash;
use San::SanLog;
use EventMan;
use San::ControllerCluster;
use NATE::IPC::Socket;
use Error;
use Tharn;
use Data::Dumper;
use San::ScsiTool;
use San::ScsiTool::CDB;
use Params::Validate qw( :all );
use NATE::Events qw(listener_add eventloop_once eventloop_drain);

# call to bigint will overload all the integers to big int
# and xlate functions will fail. Hence it should not be
# included in this lib

#
# Inherit from the Error, SanLog classes
#
@ISA = qw(Error SanLog);

use constant {FAULT_WAIT             => 60,
              FIXED_FORMAT_SENSE_LEN => 18,
              DBG_MASK_1             => 0x0001,
              DBG_MASK_2             => 0x0002,
              DBG_MASK_3             => 0x0004,
              DBG_LEVEL              => 0x00FF,
              MSN                    => 0xF0,
              LSN                    => 0x0F,
              BIT_0                  => 0x01,
              BIT_1                  => 0x02,
              BIT_2                  => 0x04,
              BIT_3                  => 0x08,
              BIT_4                  => 0x10,
              BIT_5                  => 0x20,
              BIT_6                  => 0x40,
              BIT_7                  => 0x80,};

########## get_all_luns  ##############################
#
sub get_all_luns {

    #
    my ($self, %args) = @_;
    $self->_clear_err();

    my ($client_obj, @path_info);

    $self->sanlog(-msg       => ref($self) . "::get_all_luns()\n",
                  -msg_level => 3);

    $client_obj = $self->{s_t_obj}->{client};

    if (defined $client_obj->{mp}
        && $client_obj->{mp}->can("get_path_info")) {

        my $path_info =
          $client_obj->{mp}->get_path_info(-details => $ClientUtil::TRUE);

        if ($client_obj->{mp}->error) {
            my @msg = $client_obj->{mp}->errmsg();
            my $errmsg = join ' ', @msg;
            $self->_set_err("ERROR:: $errmsg");
            return;
        }

        foreach my $mp_dev (keys %$path_info) {
            my $slice_2use = $mp_dev;

            if ($client_obj->hosttype() =~ /solaris/) {

                if (defined $slice_2use && $slice_2use =~ /c\d+t\d+d\d+s(\d+)/)
                {

                    # if the slice being used is 2 change it to 6
                    $slice_2use =~ s/s2/s6/ if ($1 == 2);
                }

                push @path_info, "/dev/rdsk/" . $slice_2use if ($slice_2use);

            } else {

                push @path_info, $slice_2use if ($slice_2use);

            }
        } ## end foreach my $mp_dev (keys %$path_info)

        print Dumper(\@path_info);

        return (scalar @path_info < 1) ? undef : \@path_info;

    } else {
        $self->_set_err(
            "ERROR:: Client and MP stack " . "do not support path information");
        return undef;
    }

}    # end get_all_luns

########## _probe_paths ##############################
#
sub _probe_paths {

    # Based on user input return either primary paths or
    # secondary paths.
    my ($self, %args) = @_;

    my $all_devs;

    $self->_clear_err();

    foreach my $k (keys %args) {
        if ($k =~ /-?host_disc_itls/) {
            unless (ref($args{$k}) =~ /array/i) {
                $self->_set_err(
                          "Need an array ref for '-host_disc_itls': $args{$k}");
                return;
            }
            $all_devs = $args{$k};
        } else {
            $self->_set_err("Unsupported param: $k");
            return;
        }
    } ## end foreach my $k (keys %args)

    unless (scalar @$all_devs) {
        $self->_set_err("Need to have atleast one device");
        return $self;
    }

    my ($ret_val, $s_data);

    # if the paths hash is ready to go then we do not need to go thru
    # the discovery process...

    foreach my $rawdev (@$all_devs) {

        # for each of these devs we are interested in figuring out this
        # set of info
        my ($inq_model, $inq_protocol, $gva_lun,
            $gva_filer, $gva_adapter,  $inq_proxy);

        # send std inq cmd, it should return with NetApp in its data
        $ret_val =
          $self->{s_t_obj}->test_scsi_cmd(-device   => $rawdev,
                                          -cdb_name => "inquiry");

        unless ($ret_val and $ret_val->{SCSI_DATA}) {
            $self->sanlog(-msg       => "Inq FAILED on dev: " . $rawdev,
                          -msg_level => 3);
            next;
        }

        $s_data =
          $self->scsi_bytes_to_strings(-scsi_data_ref => $ret_val->{SCSI_DATA},
                                       -scsi_cmd_name => "inquiry");

        $inq_model = $s_data->{T10_VENDOR_IDENTIFICATION};

        if ($inq_model =~ /NETAPP\s{2}|VSA|IBM\s{4}/i) {

            my $target_port_type = $s_data->{VERSION_DESCRIPTOR_1};

            # send GVA 10h to get the vdisk path info, byte 4 to end should be
            # the path send GVA 00h, rev 00h, shoule have the filer/vserver
            # hostname
            $ret_val =
              $self->{s_t_obj}->test_scsi_cmd(
                                       -device   => $rawdev,
                                       -cdb_name => "gva_virtual_disk_path_page"
              );

            unless ($ret_val and $ret_val->{SCSI_DATA}) {
                $self->_set_err("GVA Virtual Disk Path Page failed");
                return $self;
            }

            $s_data =
              $self->scsi_bytes_to_strings(
                                  -scsi_data_ref => $ret_val->{SCSI_DATA},
                                  -scsi_cmd_name => "gva_virtual_disk_path_page"
              );

            # start by putting in the device against the lun_name key in the
            # luns paths hash
            $gva_lun = $s_data->{VDISK_PATH};
            $gva_lun =~ s/^\s+|\s+$//;

            $ret_val =
              $self->{s_t_obj}->test_scsi_cmd(-device   => $rawdev,
                                              -cdb_name => "gva_filername_page"
                                             );

            unless ($ret_val and $ret_val->{SCSI_DATA}) {
                $self->_set_err("GVA Filername Page failed");
                return $self;
            }

            $s_data =
              $self->scsi_bytes_to_strings(
                                        -scsi_data_ref => $ret_val->{SCSI_DATA},
                                        -scsi_cmd_name => "gva_filername_page");

            $gva_filer = $s_data->{FILER_HOSTNAME};
            $gva_filer =~ s/^\s+|\s+$//;

            # if vserver and vdisk path matches then get Inq data for page
            # 0xc1 if proxy bit is set store as a secondary path else store
            # as primary paths

            $ret_val =
              $self->{s_t_obj}->test_scsi_cmd(-device   => $rawdev,
                                              -cdb_name => "inquiry_page_C1h");

            unless ($ret_val and $ret_val->{SCSI_DATA}) {
                $self->_set_err("Inquiry Page C1h failed");
                return $self;
            }

            $s_data =
              $self->scsi_bytes_to_strings(
                                        -scsi_data_ref => $ret_val->{SCSI_DATA},
                                        -scsi_cmd_name => "inquiry_page_C1h");

            $inq_proxy = $s_data->{PROXY};

            # if vserver and vdisk path matches then get GVA page 21h data
            # which will have the FCP ADAPTER PORT CLI ID of the target
            # port which corresponds to $rawdev

            my $adap_page = "gva_vdisk_fcp_adapter_page";
            $inq_protocol = "fcp";

            if ($target_port_type == 0x0960) {
                $adap_page = "gva_iscsi_adapter_info";
            } elsif ($target_port_type != 0) {
                $inq_protocol = "iscsi";
                $self->_set_err(
                              "Invalid Target Port TYPE: " . $target_port_type);
            }

            $ret_val =
              $self->{s_t_obj}->test_scsi_cmd(-device   => $rawdev,
                                              -cdb_name => $adap_page,);

            unless ($ret_val and $ret_val->{SCSI_DATA}) {
                $self->_set_err("GVA " . $adap_page . " Page failed");
                return $self;
            }

            $s_data =
              $self->scsi_bytes_to_strings(
                                        -scsi_data_ref => $ret_val->{SCSI_DATA},
                                        -scsi_cmd_name => $adap_page);

            if ($adap_page =~ /fcp/i) {
                $s_data->{CLI_PORT_ID} =~ s/^\W+|\W+$//;
                $gva_adapter = $s_data->{CLI_PORT_ID};
            } elsif ($adap_page =~ /iscsi/i) {
                $s_data->{ADAPTER_NAME} =~ s/^\W+|\W+$//;
                $gva_adapter = $s_data->{ADAPTER_NAME};
            }

            $self->{itl_list}->{$gva_lun}{$rawdev}{proxy}    = $inq_proxy;
            $self->{itl_list}->{$gva_lun}{$rawdev}{adapter}  = $gva_adapter;
            $self->{itl_list}->{$gva_lun}{$rawdev}{filer}    = $gva_filer;
            $self->{itl_list}->{$gva_lun}{$rawdev}{protocol} = $inq_protocol;
        } ## end if ($inq_model =~ /NETAPP|VSA|IBM/i)
    } ## end foreach my $rawdev (@$all_devs)

    $self->sanlog(-msg => Dumper(\%{$self->{itl_list}}));

} ## end sub _probe_paths

########## report_luns ##################################

sub report_luns {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($device, $cdb, $lun_type);

    my $cdb_length = 12;

    foreach my $k (keys %args) {
        if ($k =~ /-?device/) {
            $device = $args{$k};
        } elsif ($k =~ /-?lun_type/) {
            $lun_type = $args{$k};
        } else {
            $self->_set_err("Invalid Argument " . $k);
            return;
        }
    }

    my $buf_length = 255;

    my $cdb_ref = new CDB($cdb_length);

    $cdb_ref->set_operation(0xA0);

    $cdb_ref->set_byte_value(-byte_offset => 1,
                             -value       => 0);

    if ($lun_type =~ /vvol/i) {
        $cdb_ref->set_byte_value(-byte_offset => 2,
                                 -value       => 0xFF);
    } elsif ($lun_type =~ /regular|protocol-endpoint/i) {
        $cdb_ref->set_byte_value(-byte_offset => 2,
                                 -value       => 0xFE);
    } else {    # all luns
        $cdb_ref->set_byte_value(-byte_offset => 2,
                                 -value       => 0x00);
    }

    for (my $ind=3; $ind <= 5; $ind++){
       $cdb_ref->set_byte_value(-byte_offset => $ind,
                                -value       => 0);
    }

    if ($buf_length > 0xFFFFFFFF) {
        $self->_set_err("CDB:$buf_length value exceeds allowable range for command type");
        return $self;
    }

    $cdb_ref->{cdb}[6] = $cdb_ref->_byte_shift_right($buf_length & $cdb_ref->BYTE_3, 6);
    $cdb_ref->{cdb}[7] = $cdb_ref->_byte_shift_right($buf_length & $cdb_ref->BYTE_2, 4);
    $cdb_ref->{cdb}[8] = $cdb_ref->_byte_shift_right($buf_length & $cdb_ref->BYTE_1, 2);
    $cdb_ref->{cdb}[9] = $buf_length & $cdb_ref->BYTE_0;

    $cdb = join ' ', $cdb_ref->get_cdb_string("with_spaces");

    my $ret_ref =
      $self->{s_t_obj}->send_cmd(-device  => $device,
                                 -cdb     => $cdb,
                                 -buf_len => $buf_length);

    my $hash_ref =
      $self->scsi_bytes_to_strings(-scsi_data_ref => $ret_ref->{SCSI_DATA},
                                   -scsi_cmd_name => "report luns",);

    # (number of luns * 8 bytes per lun) + 8 byte header
    $buf_length = ($hash_ref->{num_of_luns} * 8) + 8;

    # Scaled config: buffer length required might be more than 255
    if ($buf_length > 255) {
        $cdb_ref->set_transfer_length($buf_length);

        $ret_ref =
          $self->{s_t_obj}->send_cmd(-device  => $device,
                                     -cdb     => $cdb,
                                     -buf_len => $buf_length);
 
        if ($self->{s_t_obj}->error) {
            $self->_set_err($self->{s_t_obj}->errmsg());
            return $self;
        }

        $hash_ref =
          $self->scsi_bytes_to_strings(-scsi_data_ref => $ret_ref->{SCSI_DATA},
                                       -scsi_cmd_name => "report luns",);
    }

    my @report_luns_op = @{$hash_ref->{'all_luns'}}
      if (defined $hash_ref->{'all_luns'});

    return wantarray() ? @report_luns_op : \@report_luns_op;
}

############ _get_iscsi_portal_info ##########################
# Cmode only

sub _get_iscsi_portal_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $vserver_obj;
    my $lun_id = 0;

    foreach my $k (keys %args) {
        if ($k =~ /-?vs_obj/) {
            $vserver_obj = $args{$k};
        }
    }

    unless (defined $vserver_obj) {
        $self->_set_err("Cmode Only: Please specify the vserver object");
        return $self;
    }

    my $iscsi_obj = $vserver_obj->get_iscsi_obj();

    if ($vserver_obj->error) {
        $self->_set_err($vserver_obj->errmsg());
        return $self;
    }

    my @portals = $iscsi_obj->interface_list_info();

    if ($iscsi_obj->error) {
        $self->_set_err($iscsi_obj->errmsg());
        return $self;
    }

    return wantarray() ? @portals : \@portals;
}

########## _get_itl_any_iscsi   ##############################
# Create Dev path in any_iscsi format
#

sub _get_itl_any_iscsi {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($lun_obj);
    my $lun_id = 0;

    foreach my $k (keys %args) {
        if ($k =~ /-?vs_obj/) {
            $lun_obj = $args{$k};
        }
    }

    my @portals = $self->_get_iscsi_portal_info(-vs_obj => $lun_obj);

    if ($self->error) {
        return $self;
    }

    my $device = $portals[0]->{address} . ":0";

    # Send Report luns to any IT nexus
    my @lun_ids;

    @lun_ids = $self->report_luns(-device => $device);

    # any_iscsi requires lif ip address and the lun map id to be able to send commands
    # create the device id of the form <lif_ip_address>:<lun_map_id>

    my @raw_devs;
    foreach my $portal (@portals) {
        foreach my $lun_id (@lun_ids) {
            $device = $portal->{address} . ":" . $lun_id;
            push(@raw_devs, $device);
        }
    }

    $self->sanlog(-msg       => "Raw Device List" . Dumper(\@raw_devs),
                  -msg_level => 3);

    $self->_probe_paths(-host_disc_itls => \@raw_devs);

    return;
}

########## get_paths_to_lun   ##############################
#
sub get_paths_to_lun {

    # Based on user input return either primary paths or
    # secondary paths.
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($client_obj,   $filer_name, $lun_obj, $lun_name,
        $vserver_objs, @sec_paths,  @pri_paths);

    my (@rdsk_devs, $paths_h, $force_probe);

    foreach my $k (keys %args) {
        if ($k =~ /-?lun_obj$/) {
            $lun_obj = $args{$k};
        } elsif ($k =~ /-?lun/) {
            unless (length $args{$k}) {
                $self->_set_err("Invalid value for '-lun': $args{$k}");
                return;
            }
            $lun_name = $args{$k};
        } elsif ($k =~ /-?filer/) {
            $filer_name = $args{$k};
        } elsif ($k =~ /-?force/) {
            if ($args{$k}) {
                $force_probe = 1;
            }
        } elsif ($k =~ /-?vserver_objs$/) {
            $vserver_objs = $args{$k};
        } else {
            $self->_set_err("Unsupported param: $k");
            return;
        }
    } ## end foreach my $k (keys %args)

    $client_obj = $self->{s_t_obj}->{client};

    if (not $self->{itl_list} or $force_probe) {
        if (ref($self->{s_t_obj}) =~ /any/i) {
            foreach my $vsobj (@$vserver_objs) {
                $self->_get_itl_any_iscsi(-vs_obj => $vsobj);
            }
        } else {

            # based on the hosttype of the client obj figure out the correct set
            # of devices to be checked for being NETAPP LUNs

            if ($client_obj->hosttype() =~ /solaris/i) {

                # get a list of devices from /dev/rdsk that match c?t\S+d\d+s6
                @rdsk_devs = $client_obj->execute("ls -1 /dev/rdsk/*s6");
                chomp @rdsk_devs;

                unless (scalar @rdsk_devs) {
                    $self->sanlog(-msg => "ERROR: Didnot find any slice6 paths",
                                  -msg_level => 3);
                }
            } elsif ($client_obj->hosttype() =~ /linux/i) {

                # linux box has scsi devs under the /dev dir as in
                # /dev/sda, /dev/sdb, /dev/sdc, /dev/sdd ... etc
                # but don't want to get any slices for each of the devices
                @rdsk_devs =
                  $client_obj->execute("ls -1 /dev/sd*| egrep -v '*[0-9]$''");
                chomp @rdsk_devs;

                unless (scalar @rdsk_devs) {
                    $self->sanlog(
                                -msg => "ERROR: Didnot find any /dev/sd* paths",
                                -msg_level => 3);
                }
            } elsif ($client_obj->hosttype() =~ /windows/i) {

                # for a windows box we need access to PHYSICALDRIVEs which are paths
                # to LUNs. Use devcon (on minwin compatible SAN configured test hosts)
                # to find out the number of possible drives we should try.
                #
                # C:\Documents and Settings\Administrator\Desktop>devcon listclass diskdrive
                # Listing 3 device(s) for setup class "DiskDrive" (Disk drives).
                # SCSI\DISK&VEN_IBM-ESXS&PROD_MAW3073NC_____FN&REV_C206\6&C99E489&0&000: IBM-ESXSMAW3073NC     FN SCSI Disk Device
                # SCSI\DISK&VEN_NETAPP&PROD_LUN&REV_0.2\5&381FB9E&0&120       : NETAPP LUN SCSI Disk Device
                # SCSI\DISK&VEN_NETAPP&PROD_LUN&REV_0.2\5&381FB9E&0&130       : NETAPP LUN SCSI Disk Device
                #

                my @disks = $client_obj->execute("devcon listclass diskdrive");
                chomp @disks;

                # number of lines with NETAPP LUN in them is the upper limit of
                # PHYSICALDRIVEs configured on the test host.
                my $num_phy_drives =
                  grep(/NETAPP\s+LUN | IBM\s+Nseries/, @disks);

                $num_phy_drives = 16 unless ($num_phy_drives);

                $self->sanlog(
                           -msg => "Using $num_phy_drives to probe of paths to "
                             . "NETAPP LUNs",
                           -msg_level => 3);

                foreach my $idx (0 .. $num_phy_drives) {
                    push @rdsk_devs, "\\\\.\\PHYSICALDRIVE" . $idx;
                }
            } ## end elsif ($client_obj->hosttype...

            $self->sanlog(
                       -msg => "_probe_paths called when lun is : " . $lun_name,
                       -msg_level => 3);

            # go ahead and probe all the paths to get NetApp info
            $self->_probe_paths(-host_disc_itls => \@rdsk_devs);
        }
    } ## end if (not $self->{itl_list...

    foreach my $lun_N (keys %{$self->{itl_list}}) {

        $self->sanlog(-msg => "looking for: $lun_name, dealing w/ Key: $lun_N",
                      -msg_level => 3);

        (my $lun_N_null_stripped = $lun_N) =~ s/[\0]*$//g;

        if ($lun_N_null_stripped =~ /$lun_name/i) {

            # now get the list of devs for this lun and reformat
            # the data for backwards compatibility
            foreach my $dev (keys %{$self->{itl_list}->{$lun_N}}) {

                $self->sanlog(-msg => "Key: $dev, proxy Value: "
                                . $self->{itl_list}->{$lun_N}->{$dev}->{proxy},
                              -msg_level => 3);

                if ($self->{itl_list}->{$lun_N}->{$dev}->{proxy}) {
                    push @{$paths_h->{sec}}, $dev;
                } else {
                    push @{$paths_h->{pri}}, $dev;
                }
                $paths_h->{target_ports}{$dev} =
                  $self->{itl_list}->{$lun_N}->{$dev}->{adapter};
            } ## end foreach my $dev (keys %{$self...
        } ## end if ($lun_N =~ /$lun_name/i)
    } ## end foreach my $lun_N (keys %{$self...

    $self->sanlog(-msg => Dumper($paths_h));

    unless ($paths_h->{pri}) {
        $self->sanlog(-msg => "CONFIG:: No pri paths found");
    }
    unless ($paths_h->{sec}) {
        $self->sanlog(-msg => "CONFIG:: No sec paths found");
    }

    return $paths_h;

} ## end sub get_paths_to_lun

################# get_paths ##################################
#
sub get_paths {

    # Based on user input return either primary paths or
    # secondary paths.
    #
    my ($self, %args) = @_;
    $self->_clear_err();

    my ($client_obj, $priPaths, $secPaths, $filer_name, $lun_name, @path_info);

    #    ($client_obj, $priPaths, $secPaths, $filer_name, $lun_name) = @_;
    $self->sanlog(-msg       => ref($self) . "::get_paths()\n",
                  -msg_level => 3);

    foreach my $k (keys %args) {
        if ($k =~ /-?pri/) {
            unless (ref($args{$k}) =~ /ARRAY/) {
                $self->_set_err("Invalid option for '-pri': $args{$k}");
                return $self;
            }
            $priPaths = $args{$k};
        } elsif ($k =~ /-?sec/) {
            $secPaths = $args{$k};
            unless (ref($args{$k}) =~ /ARRAY/) {
                $self->_set_err("Invalid option for '-sec': $args{$k}");
                return $self;
            }
        } elsif ($k =~ /-?filer/) {
            unless (length $args{$k}) {
                $self->_set_err("Invalid length for '-filer': $args{$k}");
                return $self;
            }
            $filer_name = $args{$k};
        } elsif ($k =~ /-?lun/) {
            unless (length $args{$k}) {
                $self->_set_err("Invalid length for '-lun': $args{$k}");
                return $self;
            }
            $lun_name = $args{$k};
        } elsif ($k =~ /-?client/) {
            unless (ref($args{$k}) =~ /SanClient/i) {
                $self->_set_err("Invalid obj ref for '-client': $args{$k}");
                return $self;
            }
            $client_obj = $args{$k};
        } else {
            $self->_set_err("Unsupported param: $k");
            return $self;
        }
    } ## end foreach my $k (keys %args)

    my $hu_path = $client_obj->get_hu_bin_path();
    my $cmd     = $hu_path . "/sanlun lun show -p";
    my $slOutp  = $client_obj->execute($cmd);
    my ($upStr, $lunFound) = (undef, undef);

    chomp(@{$slOutp});

    $self->sanlog(-msg       => "Finding: $filer_name:$lun_name",
                  -msg_level => 3);

    foreach my $ln (@{$slOutp}) {

        # delete any leading whitespace so that split
        # is not confused
        $ln =~ s/^\s+//;

        # take the sanlun output, search for primary or secondary
        # if we are dealing with the lun that we setup
        if ($ln =~ m#$filer_name:$lun_name#) {
            $self->sanlog(-msg       => "Found: $lun_name",
                          -msg_level => 3);
            $lunFound++;
        }
        elsif ($lunFound
               && (   ($ln =~ m#$filer_name# && $ln !~ m#$filer_name:$lun_name#)
                   || ($ln =~ /^$/))
          ) {
            $self->sanlog(-msg => "Found a lun not created by this script: $ln",
                          -msg_level => 3);
            $lunFound = 0;
        }

        #            elsif($lunFound && $ln =~ /^-+/) {
        #                my @fields = split/\s/,$ln;
        #                if(not defined $upStr) {
        #                    $upStr .= "A".eval {length $_}."x" for(@fields);
        #                    # take out the x at the end of the unpack string
        #                    $upStr =~ s/x$//;
        #                    $self->sanlog(-msg       => "unpack string: $upStr",
        #                                  -msg_level => 3);
        #                }
        #            }
        elsif ($lunFound && $ln =~ /primary|secondary/i) {

            # If present split the line by space. The
            # second element will indicate if primary or
            # secondary and third will have the device name.
            # my ($pState, $pType, $devName, $hba, $tPort, $pPort, $mpPolicy) =

            # my @elems = unpack($upStr,$ln);

            # with unpack we can know if a port is a tPort or pPort
            # but that does not work with latest sanlun so falling
            # back on split, which will update the tPort even
            # if the port is a pPort.
            my @elems = split /\s+/, $ln;

            # take out all leading and trailing spaces
            s/^\s*|\s*$// for (@elems);

            my %pathH = {};
            $pathH{state}    = $elems[0];
            $pathH{type}     = $elems[1];
            $pathH{dev}      = $elems[2];
            $pathH{hba}      = $elems[3];
            $pathH{tPort}    = $elems[4];
            $pathH{pPort}    = undef;
            $pathH{mpPolicy} = $elems[5];

            push @path_info, \%pathH;

            push @$priPaths, $pathH{dev} if ($pathH{type} =~ /primary/i);
            push @$secPaths, $pathH{dev} if ($pathH{type} =~ /secondary/i);

            $self->sanlog(-msg       => "$pathH{type} -- $pathH{dev}\n",
                          -msg_level => 3);
        } ## end elsif ($lunFound && $ln =~...
    } ## end foreach my $ln (@{$slOutp})
    $self->sanlog(-msg       => ref($self) . "::get_paths() done\n",
                  -msg_level => 3);

    return \@path_info;
} ## end sub get_paths

=pod

=head2 B<clear_cc_ua>

=over 4

Invoke this routine with a list of paths (ITL representations) that need to
be cleared of any pending Check Conditions, Unit Attentions.

The assumption here is that there are no more than 4 UAs at any point in
time on an ITL. We are going to use request sense SCSI cmd for clearing
any pending UAs.

In order to minimize the number of request sense cmds being sent out on
each ITL, we quit if we receive a ILLEGAL REQUEST sense key.

=back

=item Inputs

=over 4

=item * I<paths> - required param

A array ref that has all the paths that need to have their UAs cleared.

=item Outputs

=item * O<$self>

=back

=head2 B<usage examples>

$ret_val = $scsi_util_obj->clear_cc_ua (-paths => \@paths_list);

if($scsi_util_obj->error) {

    ....

    $Test->nlog($scsi_util_obj->err_msg());

    ....

}

=cut

sub clear_cc_ua {

    my ($self, %args) = @_;

    $self->_clear_err();

    my ($failCnt, @retVal, $dev2use);

    $dev2use = $args{-paths};

    $self->sanlog(-msg => "====================================\n"
                    . "Clearing pending UA on all ITL,\n"
                    . "looking for SCSI GOOD\n"
                    . "====================================\n",
                  -msg_level => 3);

    for my $devName (@$dev2use) {
        for (my $idx = 0 ; $idx < 4 ; $idx++) {

            my $ret_ref =
              $self->{s_t_obj}->send_cmd(-device  => $devName,
                                         -cdb     => "test_unit_ready",
                                         -buf_len => 32,);
            if ($self->{s_t_obj}->error) {

                $self->sanlog(-msg => "Error: send_cmd failed: "
                                . $self->{s_t_obj}->errmsg(),
                              -msg_level => 3);

            }

            # first check for GOOD status, else check for Illegal request.
            # In either of these cases, the status is considered cleared.
            if ($self->{s_t_obj}->check_status(-send_cmd_output => $ret_ref,
                                               -scsi_status     => "GOOD")
              ) {

                last;

            }
            elsif ($self->{s_t_obj}->check_status(
                                              -send_cmd_output => $ret_ref,
                                              -scsi_status => "CHECK CONDITION",
                                              -sense_key   => "ILLEGAL REQUEST")
              ) {

                last;

            } ## end elsif ($self->{s_t_obj}->check_status...
        } ## end for (my $idx = 0 ; $idx...

    } ## end for my $devName (@$dev2use)

    $self->sanlog(-msg => "====================================\n"
                    . "Cleared pending UA on all ITL,\n"
                    . "====================================\n",
                  -msg_level => 3);
    return $self;

} ## end sub clear_cc_ua

=pod

=head2 B<get_resource_names>

=over 4

San::setup() creates Volumes, LUNs and igroups based on the FILER_VOL_LIST.
The names used when creating these objects is internal to San.pm. But it is
useful to know of exact names used inside San so that we can later do specific
operations on those names, for example offline LUN1 instead of LUN2.

To this extent San::get_resources returns the clean up stack which lists the
cmds used to create/reuse objects. We use the return values to construct the
names of the resources that were created/reused for this particular test
execution instance.

This routine will take that info and piece together the names of the resources
used for this particular test execution.

=back

=item Inputs

=over 4

=item * I<san_object> - required param

The San object ref which was used to perform the setup().

=item Outputs

=item * O<ret_h>

A hash with luns, volumes and igroups as keys. The values in the hash are
array refs, where each array has names of the resources.

=item * O<$TCD::FAIL>

If the san_object is not a San object ref then we return $TCD::FAIL

=back

=head2 B<usage examples>

$ret_val = $scsi_util_obj->get_resource_names (-san_object => $san_obj);

if($scsi_util_obj->error) {

    ....

    $Test->nlog($scsi_util_obj->err_msg());

    ....

}

my @lun_names_array = $ret_val->{luns};

my @igroup_names_array = $ret_val->{igroups};

my @volume_names_array = $ret_val->{volumes};

=cut

sub get_resource_names {

    my ($self, %args) = @_;

    $self->_clear_err();

    my $san_obj;

    foreach my $k (keys %args) {

        if ($k =~ /-?san_object/) {

            unless (ref($args{$k}) =~ /San/) {

                $self->_set_err(  "Invalid option for '-san_object': $args{$k}"
                                . " expecting a San object ref");
                return;
            }

            $san_obj = $args{$k};

        } else {

            $self->_set_err("Unsupported param: $k");

            return;

        }

    } ## end foreach my $k (keys %args)

    unless (ref($san_obj) =~ /San/) {

        $self->_set_err("$san_obj: should be a San object ref");

        return;

    }

    #
    # get the hash refs for the lun and igroup, that were
    # either created or are being reused
    #
    my @lunAR = $san_obj->get_resources(-cmd => "lun_create");

    if ($san_obj->error()) {
        my $error_msg = join "\n", $san_obj->errmsg();
        $self->sanlog(-msg => ref($self) . "::get_resource_names " . $error_msg,
                      -msg_level => 3);
        return;
    }

    my @igroupAR = $san_obj->get_resources(-cmd => "igroup_create");

    if ($san_obj->error()) {
        my $error_msg = join "\n", $san_obj->errmsg();
        $self->sanlog(-msg => ref($self) . "::get_resource_names " . $error_msg,
                      -msg_level => 3);
        return;
    }

    my @volAR = $san_obj->get_resources(-cmd => "vol_create");

    if ($san_obj->error()) {
        my $error_msg = join "\n", $san_obj->errmsg();
        $self->sanlog(-msg => ref($self) . "::get_resource_names " . $error_msg,
                      -msg_level => 3);
        return;
    }

    my (@lun_names, @igroup_names, @vol_names);

    for my $idx (@lunAR) {
        for my $lhKey (keys %{$idx}) {
            my ($lun_name);
            if ($lhKey =~ /-?cmd_args/) {
                $lun_name = join '/', split /\s/, $idx->{$lhKey};
                push @lun_names, $lun_name if ($lun_name);
            }
        }
    }

    for my $ln (@lun_names) {
        $self->sanlog(-msg       => "Name of the LUN created/reused: $ln",
                      -msg_level => 3);
    }

    for my $igx (@igroupAR) {
        for my $igKey (keys %{$igx}) {
            my ($igroup_name);
            $igroup_name = $igx->{$igKey} if ($igKey =~ /-?cmd_args/);
            push @igroup_names, $igroup_name if ($igroup_name);
        }
    }

    for my $ig (@igroup_names) {
        $self->sanlog(-msg       => "Name of the igroup created/reused: $ig",
                      -msg_level => 3);
    }

    for my $vol (@volAR) {
        for my $vhKey (keys %{$vol}) {
            my ($junk, $junk1, $vol_name);
            if ($vhKey =~ /-?cmd_args/) {
                ($junk, $junk1, $vol_name) = split /\//, $vol->{$vhKey};
                push @vol_names, $vol_name if ($vol_name);
            }
        }
    }

    for my $vn (@vol_names) {
        $self->sanlog(-msg       => "Name of the volume created/reused: $vn",
                      -msg_level => 3);
    }

    #
    # store only uniq names, as there is no association
    # between these names
    #
    my %seen = ();

    my (@uniq_luns, @uniq_vols, @uniq_igroups);

    foreach my $i (@lun_names) {
        push(@uniq_luns, $i) unless $seen{$i}++;
    }

    %seen = ();

    foreach my $i (@vol_names) {
        push(@uniq_vols, $i) unless $seen{$i}++;
    }

    %seen = ();

    foreach my $i (@igroup_names) {
        push(@uniq_igroups, $i) unless $seen{$i}++;
    }

    my %ret_h;
    $ret_h{vols}    = \@uniq_vols;
    $ret_h{luns}    = \@uniq_luns;
    $ret_h{igroups} = \@uniq_igroups;

    return \%ret_h;
} ## end sub get_resource_names

sub subtest_filer_fault {

    my ($fail_cnt, $subToExec);
    $fail_cnt = 0;

    my ($self, $proc_params, $params, $proc) = @_;
    $self->_clear_err();

    $self->sanlog(-msg       => ref($self) . "::subtest_filer_fault()",
                  -msg_level => 3);

    my ($faultType, $client_obj, $filerObj, $numOfSubTests) = @{$params};

    $numOfSubTests = 1 unless ($numOfSubTests);

    # make sure the calling proc passed in the sub to call
    # else exit as a script error.
    #
    if (ref($proc) eq "CODE") {
        $subToExec = $proc;
    } else {
        $subToExec = \&exe_scsi_cmd;
    }
    $self->sanlog(-msg       => "The sub to exec() is: " . ref($subToExec),
                  -msg_level => 3);

    my ($subTestHdl, @subTestAry, $curclient_obj);

    # disconnect the client, so that each subtest can havef
    # its own connection without interfering with the parent's
    # connection. After the subtests complete, reconnect the
    # client again.
    $client_obj->destroy() if ($numOfSubTests > 1);

    #

    my $st = $self->_st_obj();
    if ($self->error) {
        my $err_msg = join "\n", $self->errmsg();
        $self->sanlog(-msg       => "ERROR: " . $err_msg,
                      -msg_level => 3);
        return 0;
    }

    for (my $x = 0 ; $x < $numOfSubTests ; $x++) {
        my $t_stamp = time();
        $subTestHdl = subtest(sub { &{$subToExec} },
                              '-runid',
                              "subtest_" . $proc_params->[1] . "_" . $t_stamp,
                              -ipc,
                              -bg,
                              "--",
                              $st,
                              $proc_params);
        push @subTestAry, $subTestHdl;
    } ## end for (my $x = 0 ; $x < $numOfSubTests...

    # check for an ipc msg from the subtest indicating that they have
    # failed the test, for 60 secs. This should give enough time
    # for subtests to start and not run into immediate issues.
    #

    my $subt_running = 1;

    while ($subt_running) {

        # assume that there are no active subtests. If there
        # is one active subtest it will set this to true.
        $subt_running = 0;

        for my $st (@subTestAry) {
            if ($st->is_alive()) {

                # Even if there is one active subtest then
                # keep spinning in the while loop.
                $subt_running = 1;

                if ($st->ipc()->can_read()) {

                    my $msg = $st->ipc()->read_message();
                    $self->sanlog(-msg => "Msg from Subtest "
                                    . $st->runid() . ": "
                                    . $msg,
                                  -msg_level => 3);

                    if ($msg =~ /Error|Config/) {
                        $st->ipc()->write_message("stop test");
                        $self->_set_err($msg);

                        # Kill the subtest if it is still active.
                        if ($st->is_alive()) {
                            $st->kill();
                            $st->wait_finish();
                        }
                    } elsif ($msg =~ /Start Fault/i) {

                        # Start the fault on the filer object.
                        #
                        if ($self->insertFault(-faultType  => $faultType,
                                               -upFilerObj => $filerObj)
                          ) {
                            $st->ipc()
                              ->write_message("Fault Started " . $faultType);
                            if ($faultType eq "takeover") {
                                while (not $filerObj->is_takeover_complete()) {
                                    $self->sanlog(
                                           -msg =>
                                             "Waiting for Takeover to complete",
                                           -msg_level => 3);
                                    sleep 10;
                                }

                                # if the takeover just completed wait for FAULT_WAIT secs
                                # and then send msg to the subtest
                                sleep FAULT_WAIT;
                                $self->sanlog(
                                      -msg =>
                                        "Sending Fault Inserted msg to Subtest "
                                        . $st->runid(),
                                      -msg_level => 4);

                                $st->ipc()
                                  ->write_message(
                                                "Fault Inserted " . $faultType);

                                if ($st->is_alive()) {
                                    $self->sanlog(
                                             -msg => $st->runid() . " is ALIVE",
                                             -msg_level => 4);

                                } else {
                                    $self->sanlog(
                                         -msg => $st->runid() . " is NOT ALIVE",
                                         -msg_level => 4);
                                }

                                $self->sanlog(
                                         -msg =>
                                           "sent Fault Inserted msg to Subtest "
                                           . $st->runid(),
                                         -msg_level => 4);

                                my $msg = $st->ipc()->read_message();

                                $self->sanlog(
                                             -msg => "Msg from subtest " . $msg,
                                             -msg_level => 3);

                            } ## end if ($faultType eq "takeover")
                        } elsif ($self->error) {
                            my $error_msg = join "\n", $self->errmsg();
                            $self->_set_err(
                                    "Inserting Filer fault FAILED: $error_msg");
                            $st->ipc()->write_message("stop test");

                            # Kill the subtest if it is still active.
                            if ($st->is_alive()) {
                                $st->kill();
                                $st->wait_finish();
                            }
                        } else {
                            $st->ipc()->write_message("stop test");
                            $self->_set_err(
                                       "Unhandled error while inserting fault");

                            # Kill the subtest if it is still active.
                            if ($st->is_alive()) {
                                $st->kill();
                                $st->wait_finish();
                            }
                        } ## end else [ if ($self->insertFault...
                    } elsif ($msg =~ /pass/i) {
                        $self->sanlog(
                                 -msg => "Subtest " . $st->runid() . " PASSED ",
                                 -msg_level => 3);

                        # Kill the subtest if it is still active.
                        if ($st->is_alive()) {
                            $st->kill();
                            $st->wait_finish();
                        }
                    } elsif ($msg =~ /hash/i) {

                        # put into the object and provide
                        # an accessor method.
                        $proc_params->[2] = $st->ipc()->read_obj();
                        $self->sanlog(
                             -msg => "Results hash recvd from: " . $st->runid(),
                             -msg_level => 3);
                    } else {
                        $self->sanlog(-msg       => "Unexpected msg...",
                                      -msg_level => 3);
                        $self->_set_err($msg);

                        # Kill the subtest if it is still active.
                        if ($st->is_alive()) {
                            $st->kill();
                            $st->wait_finish();
                        }
                    } ## end else [ if ($msg =~ /Error|Config/)
                } ## end if ($st->ipc()->can_read...
            } else {
                $self->sanlog(-msg       => "NOT Alive " . $st->runid(),
                              -msg_level => 3);
            }
        } ## end for my $st (@subTestAry)
        sleep 1;
    } ## end while ($subt_running)

    # at this stage waiting for subtests is redundant but
    # should return quick so still do it to be sure.
    Subtest::wait_finish(subtests => [@subTestAry]);

    # Now that the subtests are done reconnect the client.
    #
    $client_obj->reconnect() if ($numOfSubTests > 1);

    # Remove the inserted fault
    if ($self->error) {
        $fail_cnt++;
    } elsif ($self->removeFault($faultType, $filerObj)) {
        $self->sanlog(
               -msg => "Removed fault: " . $faultType . ", sleeping for 2 mins",
               -msg_level => 3);
        sleep(120);
    } else {
        my $error_msg = join "\n", $self->errmsg();
        $self->sanlog(-msg       => "Filer Fault removal FAILED: $error_msg",
                      -msg_level => 3);
        $fail_cnt++;
    }

    return $fail_cnt ? 0 : 1;

} ## end sub subtest_filer_fault

sub exe_scsi_cmd {
    my $fail_cnt;

    my ($self, $st_obj, $param_ref) = @_;

    my ($path_list, $cdb_name, $results_ref, $len) = @{$param_ref};

    $len = 255 unless ($len);
    unless ($cdb_name) {
        Subtest::ipc()
          ->write_message("Config: Missing the name of " . "the cdb to send");
        return;
    }
    unless (ref($results_ref) =~ /array/i) {
        Subtest::ipc()
          ->write_message("Config: Missing an array ref for " . "results");
        return;
    }

    &main::tstlog('-comment', "subtest param cdb_name: " . $cdb_name);
    &main::tstlog('-comment', "subtest param buf len: " . $len);

    # unless we pass one round of data collection through all the
    # paths, the fault injection will not start.
    my $config_errors = 1;

    while (1) {
        foreach my $dev (@$path_list) {
            my $ret_ref =
              $st_obj->send_cmd(-device  => $dev,
                                -cdb     => $cdb_name,
                                -buf_len => $len);
            if ($st_obj->error) {
                my $errmsg = join "\n", $st_obj->errmsg();
                Subtest::ipc()->write_message("Error: " . $errmsg);
                &main::tstlog('-comment', "Error: send_cmd failed: " . $errmsg);
                return;
            }

            push @$results_ref, $ret_ref;
        } ## end foreach my $dev (@$path_list)

        # since we were able to go through atleast one round
        # through the paths, let the parent know that it is
        # ok to start the faults.
        #
        if ($config_errors) {
            Subtest::ipc()->write_message("Start Fault");
            $config_errors = 0;
        }

        if (Subtest::ipc()->can_read()) {
            my $msg = Subtest::ipc()->read_message();
            &main::tstlog('-comment', "Msg from parent: " . $msg);

            if ($msg =~ /Fault Started/i) {
                Subtest::ipc()->write_message("ok");
                &main::tstlog('-comment', "Msg to parent: ok");
            } elsif ($msg =~ /Fault Inserted/i) {

                # send an ipc msg to indicate that the test is done
                Subtest::ipc()->write_message("sending result hash");
                &main::tstlog('-comment', "Msg to parent: sending result hash");

                Subtest::ipc()->write_obj($results_ref);
                &main::tstlog('-comment', "Msg to parent: result hash sent");

                Subtest::ipc()->write_message("pass");
                &main::tstlog('-comment', "Msg to parent: pass");

                return 1;

            } elsif ($msg =~ /stop test/i) {
                return 0;
            } else {
                &main::tstlog('-comment', "Unexpected msg");
                Subtest::ipc()->write_message("Error Unexpected msg: " . $msg);
                return 0;
            }
        } ## end if (Subtest::ipc()->can_read...
    } ## end while (1)
} ## end sub exe_scsi_cmd

sub insertFault {

    my ($self, %args) = @_;

    my ($faultType, $upFilerObj, $downFilerObj, $vserver, $interface,
        $wait_time);

    $wait_time = 300;

    foreach my $k (keys %args) {

        if ($k =~ /-?faultType/) {
            $faultType = $args{$k};

        } elsif ($k =~ /-?upFilerObj/) {
            $upFilerObj = $args{$k};

        } elsif ($k =~ /-?downFilerObj/) {
            $downFilerObj = $args{$k};

        } elsif ($k =~ /-?vserver/) {
            $vserver = $args{$k};

        } elsif ($k =~ /-?interface/) {
            $interface = $args{$k};

        } elsif ($k =~ /-?controller_down_time/) {
            $wait_time = $args{$k};

        } else {

            $self->_set_err("Unsupported param: $k");

            return;

        }

    } ## end foreach my $k (keys %args)

    $self->_clear_err();

    my $ret_val;

    $self->sanlog(-msg => ref($self)
                    . "::insertFault() "
                    . "starting fault type: $faultType",
                  -msg_level => 3);

    unless (ref($upFilerObj) =~ /Filer/i) {
        $self->_set_err(  "Error: Need a SanFiler obj to insert faults.\n"
                        . "It is set to: "
                        . ref($upFilerObj));
        return $self;
    }

    if (ref($upFilerObj) =~ /7M/) {
        for ($faultType) {
            if (/takeover/i) {

                # force a takeover so that any failed disks don't fail the test.
                $ret_val = $upFilerObj->takeover("-f");
            } elsif (/giveback/i) {
                if ($upFilerObj->takeover("-f")) {
                    $self->sanlog(-msg => "Waiting for TO complete - "
                                  . "sleeping for 2 mins" -msg_level => 3);
                    sleep(120);
                    while (!$upFilerObj->is_takeover_complete()) {
                        $self->sanlog(-msg => "Waiting for TO complete - "
                                      . "sleeping for 10 secs" -msg_level => 3);
                        sleep(10);
                    }
                } elsif ($upFilerObj->error()) {
                    my $error_msg = join "\n", $upFilerObj->errmsg();
                    $self->_set_err("TAKEOVER FAILED: $error_msg");
                    return $self;
                }

                # force a giveback so that any failed disks don't fail the test.
                $ret_val = $upFilerObj->giveback("-f");
            } elsif (/fcp stop/i) {
                $ret_val = $upFilerObj->fcp_stop();
            } elsif (/iscsi stop/i) {
                $ret_val = $upFilerObj->iscsi_stop();
            } elsif (/ic disable/i) {
                $ret_val = $upFilerObj->ic_disable_nic(0);
            } elsif (/cf disable/i) {
                $ret_val = $upFilerObj->cf_disable();
            } elsif (/iscsi interface disable/i) {
                if (not defined $interface) {
                    $self->_set_err("Interface has to be specified");
                    return $self;
                }

                $ret_val = $upFilerObj->iscsi_interface_disable($interface);

                if ($upFilerObj->error) {
                    $self->_set_err($upFilerObj->errmsg());
                    return $self;
                }
            } elsif (/fcp port offline/i) {
                if (not defined $interface) {
                    $self->_set_err("Interface has to be specified");
                    return $self;
                }

                $ret_val = $upFilerObj->port_offline($interface);

                if ($upFilerObj->error) {
                    $self->_set_err($upFilerObj->errmsg());
                    return $self;
                }
            } else {
                $self->_set_err("trying to insert unknown fault: $faultType");
                return $self;
            }
        }
    } elsif (ref($upFilerObj) =~ /10M/) {

        #create an EventMan
        my $eventMan = EventMan->new();
        $eventMan->add_resource(-name => $upFilerObj->get_name(),
                                -obj  => $upFilerObj,
                                -type => "sanfilersCm");
        if ($eventMan->error()) {
            $self->sanlog(
                      "Failed to create EventMan object" . $eventMan->errmsg());
            return $self;
        }
        $eventMan->add_resource(-name => $downFilerObj->get_name(),
                                -obj  => $downFilerObj,
                                -type => "sanfilersCm");
        if ($eventMan->error()) {
            $self->sanlog(
                      "Failed to create EventMan object" . $eventMan->errmsg());
            return $self;
        }

        for ($faultType) {
            if (/takeover/) {

                # use bigint overloads every scalar integer in a process space as a Math::BigInt object
                # using numify to convert wait time to an integer as
                # Tharn::snooze() requires an integer

                $eventMan->add_event(
                          -event_name           => "sfo_forcedtakeoverCm",
                          -controller_down_time => $wait_time,
                          -sanfilersCm =>
                            [$upFilerObj->get_name(), $downFilerObj->get_name()]
                );

                $eventMan->run_events();

                if ($eventMan->error()) {
                    $self->sanlog("EventMan failed to cause successful takeover"
                                  . $eventMan->errmsg());
                    return $self;
                } else {
                    $ret_val = "true";
                }
            } elsif (/giveback/) {
                $eventMan->add_event(
                          -event_name           => "sfo_forcedtakeforcedgiveCm",
                          -symmetrical_events   => 0,
                          -controller_down_time => $wait_time,
                          -sanfilersCm =>
                            [$upFilerObj->get_name(), $downFilerObj->get_name()]
                );

                $eventMan->run_events();

                if ($eventMan->error()) {
                    $self->_set_err(
                         "EventMan failed to cause successful takeover/giveback"
                           . $eventMan->errmsg());
                    return $self;
                } else {
                    my $conn = $downFilerObj->reconnect();

                    if ($downFilerObj->error()) {
                        $self->_set_err($downFilerObj->errmsg());
                        return $self;
                    }

                    $downFilerObj->set_shell_environment(-type => "ngsh");

                    if ($downFilerObj->error()) {
                        $self->_set_err($downFilerObj->errmsg());
                        return $self;
                    }

                    $ret_val = "true";
                }
            } elsif (/fcp stop/) {
                my $vserverObj =
                  $upFilerObj->get_vserver_obj(-name => $vserver);
                if ($vserverObj->error()) {
                    $self->_set_err(
                              "Could not get vserver object $vserver from filer"
                                . $upFilerObj->get_name());
                    return $self;
                }

                my $fcpServiceObj = $vserverObj->get_fcp_obj();
                if ($fcpServiceObj->error()) {
                    $self->_set_err(
                            "Could not get fcp service associated with $vserver"
                              . $fcpServiceObj->get_name());
                    return $self;
                }

                $fcpServiceObj->stop();
                if ($fcpServiceObj->error()) {
                    $self->_set_err(
                                  "Could not change admin status of fcp service"
                                    . $fcpServiceObj->errmsg());
                    return $self;
                }

            } elsif (/iscsi stop/) {
                my $vserverObj =
                  $upFilerObj->get_vserver_obj(-name => $vserver);
                if ($vserverObj->error()) {
                    $self->_set_err(
                              "Could not get vserver object $vserver from filer"
                                . $upFilerObj->get_name());
                    return $self;
                }

                my $iscsiServiceObj = $vserver->get_iscsi_obj();
                if ($vserver->error()) {
                    $self->_set_err(
                        "Could not get ISCSI service object affiliated with $vserver "
                    );
                    return $self;
                }

                $iscsiServiceObj->stop();

                if ($iscsiServiceObj->error()) {
                    $self->_set_err(
                                "Could not change admin status of iscsi service"
                                  . $iscsiServiceObj->errmsg());
                    return $self;
                }

            } elsif (/ic disable/) {

                #get the cluster lifs from the
                #cluster vservers.  Disable them to
                #imitate ic down behavior.
                #

                my $vserverObj = $upFilerObj->get_vserver_obj(
                                              -name => $upFilerObj->get_name());
                if ($upFilerObj->error()) {
                    $self->_set_err(
                               "Could not get cluster vserver object from filer"
                                 . $upFilerObj->get_name());
                    return $self;
                }

                my @lifs = $vserverObj->get_lif_objs();
                if ($vserverObj->error()) {
                    $self->_set_err("error retrieving lifs from cluster vserver"
                                    . $vserverObj->errmsg());
                    return $self;
                }

                $vserverObj = $downFilerObj->get_vserver_obj(
                                            -name => $downFilerObj->get_name());
                if ($downFilerObj->error()) {
                    $self->_set_err(
                               "Could not get cluster vserver object from filer"
                                 . $downFilerObj->get_name());
                    return $self;
                }

                push(@lifs, $vserverObj->get_lif_objs());

                foreach my $lif (@lifs) {
                    if ($lif->get_role() =~ /cluster/) {
                        $lif->modify(-admin_status => "down");

                        if ($lif->error()) {
                            $self->_set_err(
                                          "Could not change admin status of lif"
                                            . $lif->errmsg());
                            return $self;
                        }
                    }
                }

            } elsif (/cf disable/) {

            } else {

            }
        }
    }

    $self->sanlog(-msg       => "Done with fault insertion, ret val: $ret_val",
                  -msg_level => 3);
    return $ret_val;
} ## end sub insertFault

sub removeFault {
    my ($self, %args) = @_;

    my ($faultType, $upFilerObj, $downFilerObj, $vserver, $interface);

    foreach my $k (keys %args) {

        if ($k =~ /-?faultType/) {
            $faultType = $args{$k};

        } elsif ($k =~ /-?upFilerObj/) {
            $upFilerObj = $args{$k};

        } elsif ($k =~ /-?downFilerObj/) {
            $downFilerObj = $args{$k};

        } elsif ($k =~ /-?vserver/) {
            $vserver = $args{$k};

        } elsif ($k =~ /-?interface/) {
            $interface = $args{$k};

        } else {

            $self->_set_err("Unsupported param: $k");

            return;

        }

    } ## end foreach my $k (keys %args)

    $self->_clear_err();

    unless (ref($upFilerObj) =~ /Filer/i) {
        $self->_set_err(  "Error: Need a Filer obj to remove faults.\n"
                        . "It is set to: "
                        . ref($upFilerObj));
        return $self;
    }

    my $ret_val;

    if (ref($upFilerObj) =~ /7M/) {
        for ($faultType) {
            if (/takeover/i) {

                # force a takeover so that any failed disks don't fail the test.
                $ret_val = $upFilerObj->giveback("-f");
            } elsif (/giveback/i) {
                $ret_val = 1;
            } elsif (/fcp stop/i) {
                $ret_val = $upFilerObj->fcp_start();
            } elsif (/iscsi stop/i) {
                $ret_val = $upFilerObj->iscsi_start();
            } elsif (/cf disable/i) {
                $ret_val = $upFilerObj->cf_enable();
            } elsif (/ic disable/i) {
                $ret_val = $upFilerObj->ic_enable_nic(0);
            } elsif (/iscsi interface disable/i) {
                if (not defined $interface) {
                    $self->_set_err("Interface has to be specified");
                    return $self;
                }

                $ret_val = $upFilerObj->iscsi_interface_enable($interface);

                if ($upFilerObj->error) {
                    $self->_set_err($upFilerObj->errmsg());
                    return $self;
                }
            } elsif (/fcp port offline/i) {
                if (not defined $interface) {
                    $self->_set_err("Interface has to be specified");
                    return $self;
                }

                $ret_val = $upFilerObj->port_online($interface);

                if ($upFilerObj->error) {
                    $self->_set_err($upFilerObj->errmsg());
                    return $self;
                }
            } else {
                $self->_set_err("Trying to remove unknown fault: $faultType");
                return $self;
            }
        }
    } elsif ($upFilerObj =~ /10M/) {

        #
        #Create an event Manager for the Cmode versions
        #of fault events.
        #
        for ($faultType) {
            if (/takeover/) {
                $upFilerObj->sfo_giveback('fromnode' => $upFilerObj->get_name(),
                                          'override-vetoes'         => "true",
                                          'require-partner-waiting' => "true");

                if ($upFilerObj->error()) {
                    $self->sanlog("EventMan failed to cause successful giveback"
                                  . $upFilerObj->errmsg());
                    return $self;
                } else {
                    $ret_val = "true";
                }

                my $wait_time = 600;

                &SanUtil::nate_sleep($wait_time);

                my $conn = $downFilerObj->reconnect();

                if ($downFilerObj->error()) {
                    $self->_set_err($downFilerObj->errmsg());
                    return $self;
                }

                $downFilerObj->set_shell_environment(-type => "ngsh");

                if ($downFilerObj->error()) {
                    $self->_set_err($downFilerObj->errmsg());
                    return $self;
                }
            } elsif (/giveback/i) {
                $ret_val = 1;
            } elsif (/fcp_stop/) {
                my $vserverObj =
                  $upFilerObj->get_vserver_obj(-name => $vserver);
                if ($vserverObj->error()) {
                    $self->_set_err("could not get vserver $vserver from filer"
                                    . $upFilerObj->get_name());
                    return $self;
                }
                my $fcpServiceObj = $vserver->get_fcp_obj();
                if ($fcpServiceObj->error()) {
                    $self->_set_err(
                        "could not get fcp service associated with vserver $vserver"
                    );
                    return $self;
                }

                $fcpServiceObj->start();

                if ($fcpServiceObj->error()) {
                    $self->_set_err(
                                  "Could not change admin status of fcp service"
                                    . $fcpServiceObj->errmsg());
                    return $self;
                }
            } elsif (/iscsi_stop/) {
                my $vserverObj =
                  $upFilerObj->get_vserver_obj(-name => $vserver);
                if ($vserverObj->error()) {
                    $self->_set_err("could not get vserver $vserver from filer"
                                    . $upFilerObj->get_name());
                    return $self;
                }
                my $iscsiServiceObj = $vserver->get_iscsi_obj();
                if ($iscsiServiceObj->error()) {
                    $self->_set_err(
                        "could not get iscsi service associated with vserver $vserver"
                    );
                    return $self;
                }
                $iscsiServiceObj->start();

                if ($iscsiServiceObj->error()) {
                    $self->_set_err(
                                "Could not change admin status of iscsi service"
                                  . $iscsiServiceObj->errmsg());
                    return $self;
                }
            } elsif (/cf disable/i) {

            } elsif (/ic disable/i) {

                #get the cluster lifs from the
                #cluster vservers.  Disable them to
                #imitate ic down behavior.
                #

                my $vserverObj = $upFilerObj->get_vserver_obj(
                                              -name => $upFilerObj->get_name());
                if ($upFilerObj->error()) {
                    $self->_set_err("could not get cluster vserver from filer"
                                    . $upFilerObj->get_name());
                    return $self;
                }
                my @lifs = $vserver->get_lif_objs();
                if ($vserverObj->error()) {
                    $self->_set_err("could not get lif objects from filer"
                                    . $upFilerObj->get_name());
                    return $self;
                }

                $vserver = $downFilerObj->get_vserver_obj(
                                            -name => $downFilerObj->get_name());
                if ($downFilerObj->error()) {
                    $self->_set_err("could not get cluster vserver from filer"
                                    . $downFilerObj->get_name());
                    return $self;
                }

                push(@lifs, $vserver->get_lif_objs());

                if ($vserverObj->error()) {
                    $self->_set_err("could not get lif objects from filer"
                                    . $downFilerObj->get_name());
                    return $self;
                }

                foreach my $lif (@lifs) {
                    if ($lif->get_role() =~ /cluster/) {
                        $lif->modify(-admin_status => "up");

                        if ($lif->error()) {
                            $self->_set_err(
                                          "Could not change admin status of lif"
                                            . $lif->errmsg());
                            return $self;
                        }
                    }
                }

            } else {
                $self->_set_err("Trying to remove unknown fault: $faultType");
                return $self;
            }
        }
    }

    $self->sanlog(-msg       => "Filer Fault: $faultType, removed successfully",
                  -msg_level => 3);
    return $ret_val;
} ## end sub removeFault

##############           write_and_verify_pat              ####################
# Returns:  TCD::FAIL - Test Failed
###############################################################################

sub write_and_verify_pat {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Validate Params

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {device => {type     => SCALAR,
                               required => 1},
                    lba => {type     => SCALAR,
                            required => 1},
                    alloc_len => {type     => SCALAR,
                                  required => 1},
                    pat => {type     => SCALAR,
                            required => 1},});
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $device        = $args{'device'};
    my $lba           = $args{'lba'};
    my $alloc_len     = $args{'alloc_len'};
    my $pattern       = $args{'pat'};
    my $scsi_tool_obj = $self->{'s_t_obj'};

    $self->sanlog(-msg => "$device: Send write16 command and write patternX"
                  . " to one block of data starting from LBA $lba");

    my $result =
      $scsi_tool_obj->read_or_write_lba(-dev        => $device,
                                        -lba        => $lba,
                                        -read_write => "write",
                                        -alloc_len  => $alloc_len,
                                        -pat        => $pattern,);

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return 0;
    }

    $self->sanlog(
                 -msg => "$device: Send read16 to verify that the data has been"
                   . " successfuly written at LBA $lba");

    $result =
      $scsi_tool_obj->read_or_write_lba(-dev        => $device,
                                        -lba        => $lba,
                                        -read_write => "read",
                                        -alloc_len  => $alloc_len,);

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return 0;
    }

    # Confirm whether the pattern was written correctly

    $pattern x= ($alloc_len * 512);

    $result =
      $scsi_tool_obj->check_pattern(-data_buf => $result->{SCSI_DATA},
                                    -pat      => $pattern,);

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return 0;
    }

    return 1;
}

##############            read_and_verify_pat              ####################
# Returns:  TCD::FAIL - Test Failed
###############################################################################

sub read_and_verify_pat {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Validate Params

    eval {
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {device => {type     => SCALAR,
                               required => 1},
                    lba => {type     => SCALAR,
                            required => 1},
                    alloc_len => {type     => SCALAR,
                                  required => 1},
                    pat => {type     => SCALAR,
                            required => 1},});
    };

    if ($@) {
        $self->_set_err($@);
        return 0;
    }

    my $device        = $args{'device'};
    my $lba           = $args{'lba'};
    my $alloc_len     = $args{'alloc_len'};
    my $pattern       = $args{'pat'};
    my $scsi_tool_obj = $self->{'s_t_obj'};

    $self->sanlog(-msg => "$device: Read 16");

    my $result =
      $scsi_tool_obj->read_or_write_lba(-dev        => $device,
                                        -lba        => $lba,
                                        -read_write => "read",
                                        -alloc_len  => $alloc_len);

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return 0;
    }

    $self->sanlog(
                -msg => "$device: Confirm pattern is written at specified LBA");

    $pattern x= ($alloc_len * 512);

    $result =
      $scsi_tool_obj->check_pattern(-data_buf => $result->{SCSI_DATA},
                                    -pat      => $pattern);

    if ($scsi_tool_obj->error) {
        $self->_set_err($scsi_tool_obj->errmsg());
        return 0;
    }

    return 1;
}

sub new {
    my ($caller, %args) = @_;
    my $class = ref($caller) || $caller;

    #
    # Create a hash to store attributes and bless the hash to create an object.
    #
    my $self = {};
    bless $self, $class;

    # add the scsi tool obj debug level.
    #
    $self->set_debug_level(3);

    foreach my $k (keys %args) {
        if ($k =~ /-?scsi_tool_obj/ && ref($args{$k}) =~ /ScsiTool/) {

            # store the scsi_tool object ref.
            $self->{s_t_obj} = $args{$k};

            # override the default debug level
            # with the ScsiTool debug level.
            $self->set_debug_level($self->{s_t_obj}->get_debug_level());
        } else {
            $self->_set_err("ERROR:: unknown param: $k");
            return $self;
        }
    } ## end foreach my $k (keys %args)
    return $self;
} ## end sub new

########################################################################
# Sub Name:  xlate_RTPG_bytes ()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
########################################################################

sub xlate_RTPG_bytes {
    my ($self, $retRef, $hasta) = @_;

    $self->_clear_err();

    # fold the output into one string and remove any white space
    #
    my @data_ary;

    if ($hasta) {

        @data_ary = @$retRef;

    } else {

        s/^\s*|\s*$// for (@{$retRef});

        @data_ary = grep(/^\w\w\s\w\w\s/, @$retRef);

    }

    my $tmp1 = join '', split ' |\n', join '', @data_ary;

    # find the length of the data returned. If RTPG data len is > fcp_dl
    # return a "retry cmd" w/ new fcp_dl
    my $len = hex unpack("A8", $tmp1);

    unless ($len) {
        $self->_set_err("ERROR: length set to 0 for RTPG data");
        return;
    }

    my %tpgH = ();

    $tpgH{len} = $len;

    # skip the next 8chars as that is length field.
    $tmp1 =~ s/^.{8}//;

    # while the $len is valid keep applying the
    # same template to get TPG data and then skip the required
    # number of bytes to get the next TPG.

    my $offset;

    while ($len * 2 > $offset) {

        my ($state, $flags, $tpg, $pcnt) = unpack("xA A2 A4 x7A", $tmp1);

        my ($numP, $tpgN, @pAry);

        # undef $numP; undef $tpgN; undef @pAry;
        $numP = hex $pcnt;

        $tmp1 =~ s/^.{16}//;
        $offset += 16;

        for (my $x = 0 ; $x < $numP ; $x++) {
            push @pAry, hex unpack("A8", $tmp1);
            $tmp1 =~ s/^.{8}//;
        }

        # now skip to the next TPG
        $offset += 8 * $numP;

        # so that we will always have a valid key into
        # the tpgN hash.

        $tpgN = (hex $tpg) + 1;

        $tpgH{$tpgN}{tpg}   = hex $tpg;
        $tpgH{$tpgN}{state} = hex $state;
        $tpgH{$tpgN}{pCnt}  = $numP;
        $tpgH{$tpgN}{flags} = hex $flags;
        $tpgH{$tpgN}{ports} = \@pAry;

    } ## end while ($len * 2 > $offset)

    return \%tpgH;

} ## end sub xlate_RTPG_bytes

########################################################################
# Sub Name:  xlate_RTPG_bytes_2 ()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
########################################################################

sub xlate_RTPG_bytes_2 {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    my (%xlate_h);

    # length of data returned is the first 4 bytes in the data
    # so should take a splice of the array to get the length

    my $len = hex join '', splice @$s_data, 0, 4;

    unless ($len) {
        $self->_set_err("ERROR: length set to 0 for RTPG data");
        return;
    }

    $xlate_h{return_data_length} = $len;

    my $idx = 0;

    while (length @$s_data) {
        my $tpg_id = "TPG_Descriptor_" . $idx++;

        $xlate_h{$tpg_id}{state} = hex unpack("xA", @$s_data[0]);
        shift @$s_data;
        $xlate_h{$tpg_id}{flags}           = hex shift @$s_data;
        $xlate_h{$tpg_id}{tpg}             = hex join '', splice @$s_data, 0, 2;
        $xlate_h{$tpg_id}{reserved}        = hex shift @$s_data;
        $xlate_h{$tpg_id}{status_code}     = hex shift @$s_data;
        $xlate_h{$tpg_id}{vendor_specific} = hex shift @$s_data;

        my $p_count = hex shift @$s_data;

        $xlate_h{$tpg_id}{target_port_count} = $p_count;

        my $tp_idx = 0;
        while ($p_count--) {
            $xlate_h{$tpg_id}{"TP_Descriptor_" . $tp_idx}{obsolete} =
              hex join '',
              splice @$s_data, 0, 2;

            my $rel_tp_id = hex splice @$s_data, 2, 2;
            $xlate_h{$tpg_id}{"TP_Descriptor_" . $tp_idx}
              {"rel_t_p_id_" . $tp_idx} = $rel_tp_id;
            $tp_idx++;
        } ## end while ($p_count--)
    } ## end while (length @$s_data)

    return \%xlate_h;

} ## end sub xlate_RTPG_bytes_2

########################################################################
# Sub Name:  xlate_inq_pg_85h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'PERIPHERAL_QUALIFIER' => 0,
#          'VPD_PAGE_CODE' => 133,
#          'PAGE_LENGTH' => 128,
#          'NETWORK_SERVICE_DESCRIPTOR_LIST' => [
#                                                 {
#                                                   'SERVICE_TYPE' => 1,
#                                                   'ASSOCIATION' => 0,
#                                                   'RESV_B1' => 0,
#                                                   'NETWORK_ADDRESS_LENGTH' => 24,
#                                                   'NETWORK_ADDRESS' => 'http://10.60.169.11/^@^@^@^@'
#                                                 },
#                                                 {
#                                                   'SERVICE_TYPE' => 1,
#                                                   'ASSOCIATION' => 0,
#                                                   'RESV_B1' => 0,
#                                                   'NETWORK_ADDRESS_LENGTH' => 20,
#                                                   'NETWORK_ADDRESS' => 'http://172.16.51.5/^@'
#                                                 },
#                                                 {
#                                                   'SERVICE_TYPE' => 1,
#                                                   'ASSOCIATION' => 0,
#                                                   'RESV_B1' => 0,
#                                                   'NETWORK_ADDRESS_LENGTH' => 20,
#                                                   'NETWORK_ADDRESS' => 'http://172.16.51.6/^@'
#                                                 },
#                                                 {
#                                                   'SERVICE_TYPE' => 1,
#                                                   'ASSOCIATION' => 0,
#                                                   'RESV_B1' => 0,
#                                                   'NETWORK_ADDRESS_LENGTH' => 20,
#                                                   'NETWORK_ADDRESS' => 'http://172.16.51.9/^@'
#                                                 },
#                                                 {
#                                                   'SERVICE_TYPE' => 1,
#                                                   'ASSOCIATION' => 0,
#                                                   'RESV_B1' => 0,
#                                                   'NETWORK_ADDRESS_LENGTH' => 24,
#                                                   'NETWORK_ADDRESS' => 'http://172.16.51.10/'
#                                                 }
#                                               ]
#        };
#
########################################################################

sub xlate_inq_pg_85h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    my ($offset, $idx, @nsd_list);
    $offset = 0;

    while ($offset < $xlate_h{PAGE_LENGTH}) {
        my ($next_byte);
        tie my %ns_desc, 'Tie::IxHash';
        $next_byte = hex shift @$s_data;
        $ns_desc{SERVICE_TYPE} =
          $next_byte & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
        $ns_desc{ASSOCIATION} = ($next_byte & (BIT_5 | BIT_6)) >> 5;

        $ns_desc{RESV_B1} = hex shift @$s_data;

        $ns_desc{NETWORK_ADDRESS_LENGTH} = hex join '', splice @$s_data, 0, 2;

        my $len = $ns_desc{NETWORK_ADDRESS_LENGTH};

        $ns_desc{NETWORK_ADDRESS} =
          pack("H*", join '', splice @$s_data, 0, $len);
        $offset += 4 + $len;

        push @nsd_list, \%ns_desc;
    } ## end while ($offset < $xlate_h...

    $xlate_h{NETWORK_SERVICE_DESCRIPTOR_LIST} = \@nsd_list;

    return \%xlate_h;

} ## end sub xlate_inq_pg_85h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_C0h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_QUALIFIER' => 0,
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'VPD_PAGE_CODE' => 128,
#          'RESERVED' => 0,
#          'PAGE_LENGTH' => 12,
#          'PRODUCT_SERIAL_NUMBER' => 'HnS7XJNi9VeZ'
#        };
#
########################################################################

sub xlate_inq_pg_C0h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    $xlate_h{VPD_PAGE_CODE}     = hex shift @$s_data;
    $xlate_h{RESV_B2}           = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH}       = hex shift @$s_data;
    $xlate_h{LENGTH}            = hex join '', splice @$s_data, 0, 4;
    $xlate_h{NETAPP_MAGIC_CODE} = join '', splice @$s_data, 0, 3;
    $xlate_h{REVISION}          = hex shift @$s_data;
    $xlate_h{GVA_PAGE_CODE}     = hex shift @$s_data;
    $xlate_h{GVA_REVISION}      = hex shift @$s_data;
    $xlate_h{IP_LENGTH}         = hex join '', splice @$s_data, 0, 2;
    $xlate_h{COUNT}             = hex join '', splice @$s_data, 0, 4;

    my ($offset, @ip_list);
    $offset = 0;

    while ($offset < ($xlate_h{COUNT} * 4)) {
        my $ipv4_addr = join '', splice @$s_data, 0, 4;
        push @ip_list, $ipv4_addr;
        $offset += 4;
    }

    $xlate_h{FILER_IP_ADDRESSES_LIST} = \@ip_list;

    return \%xlate_h;

} ## end sub xlate_inq_pg_C0h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_C1h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'PERIPHERAL_QUALIFIER' => 1,
#          'VPD_PAGE_CODE' => 193,
#          'RESV_B2' => 0,
#          'PAGE_LENGTH' => 16,
#          'LENGTH' => 16,
#          'NETAPP_MAGIC_CODE' => '0a980a',
#          'REVISION' => 0,
#          'PAGE_CODE' => 193,
#          'REVISION_1' => 0,
#          'LENGTH_1' => 8,
#          'RESV_B16_17_18' => 0,
#          'TAKEOVER' => 0,
#          'PROXY' => 1,
#          'HI_PROXY' => 0,
#          'STANDBY' => 0,
#          'RESV_BITS_4_7' => 0
#        };
#
########################################################################

sub xlate_inq_pg_C1h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    $xlate_h{VPD_PAGE_CODE}     = hex shift @$s_data;
    $xlate_h{RESV_B2}           = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH}       = hex shift @$s_data;
    $xlate_h{LENGTH}            = hex join '', splice @$s_data, 0, 4;
    $xlate_h{NETAPP_MAGIC_CODE} = join '', splice @$s_data, 0, 3;
    $xlate_h{REVISION}          = hex shift @$s_data;
    $xlate_h{PAGE_CODE}         = hex shift @$s_data;
    $xlate_h{REVISION_1}        = hex shift @$s_data;
    $xlate_h{LENGTH_1}          = hex join '', splice @$s_data, 0, 2;
    $xlate_h{RESV_B16_17_18}    = hex join '', splice @$s_data, 0, 3;

    my $byte19 = hex shift @$s_data;
    $xlate_h{TAKEOVER}      = $byte19 & BIT_0;
    $xlate_h{PROXY}         = ($byte19 & BIT_1) >> 1;
    $xlate_h{HI_PROXY}      = ($byte19 & BIT_2) >> 2;
    $xlate_h{STANDBY}       = ($byte19 & BIT_3) >> 3;
    $xlate_h{RESV_BITS_4_7} = ($byte19 & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;

    return \%xlate_h;

} ## end sub xlate_inq_pg_C1h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_C2h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'PERIPHERAL_QUALIFIER' => 0,
#          'VPD_PAGE_CODE' => 194,
#          'RESV_B2' => 0,
#          'PAGE_LENGTH' => 15,
#          'VTIC' => 1,
#          'RESV_B4_Bits1_7' => 0,
#          'RESV_B5' => 0,
#          'PORT_ID' => 4,
#          'LUN_NUMBER' => 1,
#          'LUN_NUMBER_6LSB' => 0,
#          'ADAPTER_NAME_STRING_LENGTH' => 2,
#          'ADAPTER_NAME_STRING' => '1b'
#        };
#
########################################################################

sub xlate_inq_pg_C2h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;

    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $next_byte & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $xlate_h{RESV_B2}       = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH}   = hex shift @$s_data;

    $next_byte = hex shift @$s_data;
    $xlate_h{VTIC} = $next_byte & BIT_0;
    $xlate_h{RESV_B4_Bits1_7} = $next_byte & (~BIT_0);

    $xlate_h{RESV_B5} = hex shift @$s_data;

    $xlate_h{PORT_ID} = hex join '', splice @$s_data, 0, 2;

    #
    # SCSI LU Number can be 8 bytes. ONTAP SCSI Target only supports
    # flat addressing which uses only the MostSignificant 2 bytes to hold the
    # LU Number. Rest of the 6 bytes should be all 0s
    #
    $xlate_h{LUN_NUMBER}      = hex join '', splice @$s_data, 0, 2;
    $xlate_h{LUN_NUMBER_6LSB} = hex join '', splice @$s_data, 0, 6;

    my $len = hex shift @$s_data;

    $xlate_h{ADAPTER_NAME_STRING_LENGTH} = $len;

    $xlate_h{ADAPTER_NAME_STRING} =
      pack("H*", join '', splice @$s_data, 0, $len);

    return \%xlate_h;

} ## end sub xlate_inq_pg_C2h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_00h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The return hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_QUALIFIER' => 0,
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'VPD_PAGE_CODE' => 0,
#          'RESERVED' => 0,
#          'PAGE_LENGTH' => 7,
#          'SUPPORTED_PAGE_LIST' => [
#                                     0,
#                                     128,
#                                     131,
#                                     133,
#                                     192,
#                                     193,
#                                     194
#                                   ]
#        };
#
########################################################################

sub xlate_inq_pg_00h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    #    unless (ref $s_data =~ /ARRAY/) {
    #        $self->_set_err("ERROR: Need an Array of data for xlation");
    #        return;
    #    }

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $xlate_h{RESERVED}      = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH}   = hex shift @$s_data;

    my $idx = 0;
    my @page_list;

    while ($idx < $xlate_h{PAGE_LENGTH}) {
        push @page_list, hex shift @$s_data;
        $idx++;
    }

    $xlate_h{SUPPORTED_PAGE_LIST} = \@page_list;

    return \%xlate_h;

} ## end sub xlate_inq_pg_00h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_83h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The return hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_QUALIFIER' => 0,
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'VPD_PAGE_CODE' => 0x83,
#          'PAGE_LENGTH' => 0x9c if alua on, 0x8e if not
#          'ID_LIST' => [
#                         LUN ASCII T10 ID DESCRIPTOR
#                         LUN BINARY NAA ID DESCRIPTOR
#                         LUN BINARY EUI-64 ID DESCRIPTOR
#                         LUN PORT BINARY NAA ID DESCRIPTOR
#                         LUN ALUA RELATIVE PORT ID DESCRIPTOR ## ALUA only
#                         LUN ALUA TPG ID DESCRIPTOR ## ALUA only
#                         LUN TARGET_ASCII T10 ID DESCRIPTOR
#                         LUN LOGICAL UNIT GROUP ID DESCRIPTOR
#                       ]
#        };
#
########################################################################

sub xlate_inq_pg_83h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();
    tie my %xlate_h, 'Tie::IxHash';
    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    my ($offset, $idx, @id_list);
    $offset = 0;

    while (($offset < $xlate_h{PAGE_LENGTH})) {
        my $test_byte = hex @$s_data[0];
        my $prot_id   = ($test_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
        my $code_set  = ($test_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0));
        $test_byte = hex @$s_data[1];
        my $assoc   = ($test_byte & (BIT_5 | BIT_4)) >> 4;
        my $id_type = ($test_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0));
        my $id_len  = hex @$s_data[3];

        my $return_data;

        #increment the offset to the header length of the ID
        $offset += (4 + $id_len);
        if (($code_set == 2) && ($id_type == 1) && ($assoc == 0)) {
            $return_data =
              xlate_lun_ascii_id($s_data, $xlate_h{PERIPHERAL_DEVICE_TYPE});
            if ($id_len - 32) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 32");
                return;
            }
        } elsif (($code_set == 1) && ($id_type == 3) && ($assoc == 0)) {
            $return_data = xlate_lun_binary_naa_id($s_data);
            if ($id_len - 16) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 16");
                return;
            }
        } elsif (($code_set == 1) && ($id_type == 2) && ($assoc == 0)) {
            $return_data = $self->xlate_lun_binary_eui64_id($s_data);
            if ($id_len - 16) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 16");
                return;
            }
        } elsif (($code_set == 1) && ($id_type == 3) && ($assoc == 1)) {
            $return_data = $self->xlate_port_binary_naa_id($s_data);
            if ($id_len - 16) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 16");
                return $TCD::FAIL;
            }
        } elsif (($code_set == 1) && ($id_type == 4) && ($assoc == 1)) {
            $return_data = $self->xlate_alua_relative_port_id($s_data);
            if ($id_len - 4) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 4");
                return $TCD::FAIL;
            }
        } elsif (($code_set == 1) && ($id_type == 5) && ($assoc == 1)) {
            $return_data = xlate_alua_tpg_id($s_data);
            if ($id_len - 4) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 4");
                return $TCD::FAIL;
            }
        } elsif (($code_set == 2) && ($id_type == 1) && ($assoc == 2)) {
            $return_data = xlate_target_ascii_id($s_data);
            if ($id_len - 32) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 32");
                return $TCD::FAIL;
            }
        } elsif (($code_set == 1) && ($id_type == 6) && ($assoc == 0)) {
            $return_data = xlate_lu_group_id($s_data);
            if ($id_len - 4) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 4");
                return $TCD::FAIL;
            }
        } elsif (($code_set == 3) && ($id_type == 8) && ($assoc == 2)) {
            $return_data = xlate_target_device_utf8_id($s_data);
            if ($id_len - 36) {
                $self->_set_err(  "ERROR: ID Length: $id_len, does not match "
                                . "expected length: 36");
                return $TCD::FAIL;
            }
        }
        push @id_list, $return_data;
    } ## end while ( ( $offset < $xlate_h...

    $xlate_h{ID_LIST} = \@id_list;

    return \%xlate_h;
} ## end sub xlate_inq_pg_83h_bytes

########################################################################
# Sub Name:  xlate_lun_ascii_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN ASCII T10 ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 2,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 0,
#            'IDENTIFIER_TYPE' => 1,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x20,
#        ## when peripheral device type is 0
#            'ID_BYTE_8_13' => 'NETAPP',
#            'ID_BYTE_14_16' => '   ',
#            'ID_BYTE_17_19' => 'LUN',
#            'ID_BYTE_20' => ' ',
#            'LUN_SERIAL_NUMBER' => 'HnS7XJNi9VeZ',
#            'ID_BYTES_33_39' => 0
#        ## when peripheral device type is 10
#            'ID_BYTE_8_13' => 'NETAPP',
#            'ID_BYTE_14' => '   ',
#            'ID_BYTE_15_17' => 'LUN',
#            'ID_BYTE_18' => ' ',
#            'ID_BYTE_19_22' => ' ',
#            'ID_BYTE_23' => ' ',
#            'LUN_SERIAL_NUMBER' => 'HnS7XJNi9VeZ',
#            'ID_BYTES_36_39' => 0
#        },
#
########################################################################
sub xlate_lun_ascii_id() {
    my ($s_data, $device_type) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    ## This is if the peripheral device type is 00.
    if ($device_type == 0) {

        # Identifier bytes are from 8 to 39.
        $desc_hash{ID_BYTE_8_13}  = pack("H*", join '', splice @$s_data, 0, 6);
        $desc_hash{ID_BYTE_14_16} = pack("H*", join '', splice @$s_data, 0, 3);
        $desc_hash{ID_BYTE_17_19} = pack("H*", join '', splice @$s_data, 0, 3);
        $desc_hash{ID_BYTE_20} = hex shift @$s_data;
        $desc_hash{LUN_SERIAL_NUMBER} =
          pack("H*", join '', splice @$s_data, 0, 12);
        $desc_hash{ID_BYTES_33_39} = hex join '', splice @$s_data, 0, 7;
    } elsif ($device_type == 10) {
        ## This is if the peripheral device type is 0C(VSA controller)
        # Identifier bytes are from 8 to 39.
        $desc_hash{ID_BYTE_8_13}  = pack("H*", join '', splice @$s_data, 0, 6);
        $desc_hash{ID_BYTE_14}    = hex shift @$s_data;
        $desc_hash{ID_BYTE_15_17} = pack("H*", join '', splice @$s_data, 0, 3);
        $desc_hash{ID_BYTE_18}    = hex shift @$s_data;
        $desc_hash{ID_BYTE_19_22} = pack("H*", join '', splice @$s_data, 0, 4);
        $desc_hash{ID_BYTE_23}    = hex shift @$s_data;
        $desc_hash{LUN_SERIAL_NUMBER} =
          pack("H*", join '', splice @$s_data, 0, 12);
        $desc_hash{ID_BYTES_36_39} = hex join '', splice @$s_data, 0, 7;
    } ## end elsif ( $device_type == 10)

    return \%desc_hash;
}    # end xlate_lun_ascii_id

########################################################################
# Sub Name:  xlate_lun_binary_naa_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN BINARY NAA ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 1,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 0,
#            'IDENTIFIER_TYPE' => 3,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x10,
#            'NETAPP_MAGIC_NUMBER' => '60a98000',
#            'LUN_SERIAL_NUMBER_0_11' => 'HnS7XJNi9VeZ'
#        },
#
########################################################################
sub xlate_lun_binary_naa_id() {
    my ($s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{NETAPP_MAGIC_NUMBER} = join '', splice @$s_data, 0, 4;
    $desc_hash{LUN_SERIAL_NUMBER_0_11} =
      pack("H*", join '', splice @$s_data, 0, 12);

    return \%desc_hash;
}    # end xlate_lun_binary_naa_id

########################################################################
# Sub Name:  xlate_lun_binary_eui64_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN BINARY EUI-64 ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 1,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 0,
#            'IDENTIFIER_TYPE' => 2,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x10,
#            'LUN_SERIAL_NUMBER_5_11' => ' ',
#            'NETAPP_MAGIC_NUMBER' => '000a9800',
#            'LUN_SERIAL_NUMBER_0_4' => ' '
#        },
#
########################################################################
sub xlate_lun_binary_eui64_id() {
    my ($self, $s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{LUN_SERIAL_NUMBER_5_11} =
      pack("H*", join '', splice @$s_data, 0, 7);
    $desc_hash{NETAPP_MAGIC_NUMBER} = join '', splice @$s_data, 0, 4;
    $desc_hash{LUN_SERIAL_NUMBER_0_5} =
      pack("H*", join '', splice @$s_data, 0, 5);

    return \%desc_hash;
}    # end sub xlate_lun_binary_eui64_id

########################################################################
# Sub Name:  xlate_port_binary_naa_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN PORT BINARY NAA ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 1,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 1,
#            'IDENTIFIER_TYPE' => 3,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x10,
#            'NETAPP_MAGIC_NUMBER' => '60a98000000000',
#            'BYTE91' => ' ',
#            'BYTE92_99' => ' '
#         },
#
########################################################################
sub xlate_port_binary_naa_id() {
    my ($self, $s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{NETAPP_MAGIC_NUMBER} = join '', splice @$s_data, 0, 7;
    $desc_hash{BYTE91}              = hex shift @$s_data;
    $desc_hash{BYTES_92_99}         = join '', splice @$s_data, 0, 8;

    return \%desc_hash;
}    # end sub xlate_port_binary_naa_id

########################################################################
# Sub Name:  xlate_alua_relative_port_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN ALUA RELATIVE PORT ID DESCRIPTOR ## ALUA only
#$VAR1 = {
#           'PROTOCOL_IDENTIFIER' => 0,
#           'CODE_SET' => 1,
#           'PIV' => 0,
#           'RESV_B5_bit6' => 0,
#           'ASSOCIATION' => 1,
#           'IDENTIFIER_TYPE' => 4,
#           'RESV_B6' => 0,
#           'IDENTIFIER_LENGTH' => 4,
#           'BYTES_104_105' => '0000',
#           'BYTES_106_107' => ' '
#        },
#
########################################################################
sub xlate_alua_relative_port_id() {
    my ($self, $s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{BYTES_104_105} = pack("H*", join '', splice @$s_data, 0, 2);
    $desc_hash{BYTES_106_107} = pack("H*", join '', splice @$s_data, 0, 2);

    return \%desc_hash;
}    # end sub xlate_alua_relative_port_id

########################################################################
# Sub Name:  xlate_alua_tpg_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#            LUN ALUA TPG ID DESCRIPTOR ## ALUA only
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 1,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 1,
#            'IDENTIFIER_TYPE' => 5,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x04,
#            'RESV_B112_113' => ' ',
#            'BYTES_114_115' => ' '
#        }
#
########################################################################
sub xlate_alua_tpg_id() {
    my ($s_data, $version) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{BYTES_112_113} = pack("H*", join '', splice @$s_data, 0, 2);
    $desc_hash{BYTES_114_115} = pack("H*", join '', splice @$s_data, 0, 2);

    return \%desc_hash;
}    # end sub xlate_alua_tpg_id

########################################################################
# Sub Name:  xlate_target_ascii_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#
#            LUN TARGET_ASCII T10 ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 2,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 2,
#            'IDENTIFIER_TYPE' => 1,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x14,
#            'BYTES_104_109' => 'NETAPP',
#            'BYTES_110_111' => ' ',
#            'BYTES_112_117' => 'TARGET'
#            'BYTES_118' => ' '
#            'TARGET_DEVICE_ID' => ' '
#        }
#
########################################################################
sub xlate_target_ascii_id() {
    my ($s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{BYTES_104_109} = pack("H*", join '', splice @$s_data, 0, 6);
    $desc_hash{BYTES_110_111} = pack("H*", join '', splice @$s_data, 0, 2);
    $desc_hash{BYTES_112_117} = pack("H*", join '', splice @$s_data, 0, 6);
    $desc_hash{BYTES_118} = hex shift @$s_data;
    $desc_hash{TARGET_DEVICE_ID} = pack("H*", join '', splice @$s_data, 0, 17);

    return \%desc_hash;
}    # end sub xlate_target_ascii_id

########################################################################
# Sub Name:  xlate_lu_group_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#
#            LUN LOGICAL UNIT GROUP ID DESCRIPTOR
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 1,
#            'PIV' => 0,
#            'RESV_B5_bit6' => 0,
#            'ASSOCIATION' => 0,
#            'IDENTIFIER_TYPE' => 6,
#            'RESV_B6' => 0,
#            'IDENTIFIER_LENGTH' => 0x04,
#            'RESERVED_BYTES_156_157' => ' ',
#            'LOGICAL_UNIT_GROUP_ID' => ' '
#        }
#
########################################################################
sub xlate_lu_group_id() {
    my ($s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B5_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B6} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{RESERVED_BYTES_156_157} =
      pack("H*", join '', splice @$s_data, 0, 2);
    $desc_hash{LOGICAL_UNIT_GROUP_ID} =
      pack("H*", join '', splice @$s_data, 0, 2);

    return \%desc_hash;
}    # end sub xlate_lu_group_id

########################################################################
# Sub Name:  xlate_target_device_utf8_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
#
#            TARGET DEVICE UTF_8 T10 IDENTIFIER
#$VAR1 = {
#            'PROTOCOL_IDENTIFIER' => 0,
#            'CODE_SET' => 3,
#            'PIV' => 0,
#            'RESV_B1_bit6' => 0,
#            'ASSOCIATION' => 2,
#            'IDENTIFIER_TYPE' => 8,
#            'RESV_B2' => 0,
#            'IDENTIFIER_LENGTH' => 0x24,
#            'UTF_8' => naa.600A098,
#            'TARGET_DEVICE_IDENTIFER' => ' ',
#            'REMAINING_BYTES' => ' '
#        }
#
########################################################################
sub xlate_target_device_utf8_id {
    my ($s_data) = @_;

    tie my %desc_hash, 'Tie::IxHash';

    my $next_byte = hex shift @$s_data;
    $desc_hash{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $desc_hash{CODE_SET} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $next_byte = hex shift @$s_data;
    $desc_hash{IDENTIFIER_TYPE} = $next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $desc_hash{ASSOCIATION}  = ($next_byte & (BIT_5 | BIT_4)) >> 4;
    $desc_hash{RESV_B1_bit6} = ($next_byte & (BIT_6)) >> 6;
    $desc_hash{PIV}          = ($next_byte & (BIT_7)) >> 7;

    $desc_hash{RESV_B2} = hex shift @$s_data;

    $desc_hash{IDENTIFIER_LENGTH} = hex shift @$s_data;

    $desc_hash{UTF_8} = pack("H*", join '', splice @$s_data, 0, 11);
    $desc_hash{TARGET_DEVICE_IDENTIFER} =
      pack("H*", join '', splice @$s_data, 0, 13);

    return \%desc_hash;
}    # end subxlate_target_device_utf8_id

########################################################################
# Sub Name:  xlate_inq_pg_80h_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The hash returned will have similar key and values pairs
#
#$VAR1 = {
#          'PERIPHERAL_QUALIFIER' => 0,
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'VPD_PAGE_CODE' => 128,
#          'RESERVED' => 0,
#          'PAGE_LENGTH' => 12,
#          'PRODUCT_SERIAL_NUMBER' => 'HnS7XJNi9VeZ'
#        };
#
########################################################################

sub xlate_inq_pg_80h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    #    unless (ref $s_data =~ /ARRAY/) {
    #        $self->_set_err("ERROR: Need an Array of data for xlation");
    #        return;
    #    }

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);

    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $xlate_h{RESERVED}      = hex shift @$s_data;
    $xlate_h{PAGE_LENGTH}   = hex shift @$s_data;
    $xlate_h{PRODUCT_SERIAL_NUMBER} =
      pack("H*", join '', splice @$s_data, 0, 12);

    return \%xlate_h;

} ## end sub xlate_inq_pg_80h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_87h_bytes()
# Objective: Return a hash with values set for Mode Page Policy VPD page
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-3 Specification - Revision 23
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'PERIPHERAL_QUALIFIER' => 0,
#          'VPD_PAGE_CODE' => 135,
#          'PAGE_LENGTH' => 60
#          'MODE_PAGE_POLICY_DESCRIPTOR_LIST' => [
#                                                 {
#                                                  'RESV_B0_bits6_7' => 0,
#                                                  'POLICY_PAGE_CODE' => 2,
#                                                  'POLICY_SUBPAGE_CODE' => 1,
#                                                  'MLUS' => 0,
#                                                  'RESV_B2_bits2_6' => 0,
#                                                  'MODE_PAGE_POLICY' => 0,
#                                                  'RESV_B3' => 0,
#                                                 },
#                                                 {
#                                                  ...
#                                                 },
#                                                ]
#        };
#
########################################################################

sub xlate_inq_pg_87h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_length = scalar @$s_data;
    if ($data_length <= 1) {
        $self->_set_err("BYTE_0: Insufficient data length: " . $data_length);
        return;
    }
    my $byte0 = hex shift @$s_data;
    $data_length--;

    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    if ($data_length <= 1) {
        $self->_set_err(
                    "VPD_PAGE_CODE: Insufficient data length: " . $data_length);
        return;
    }
    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $data_length--;

    if ($data_length <= 2) {
        $self->_set_err(
                      "PAGE_LENGTH: Insufficient data length: " . $data_length);
        return;
    }
    $xlate_h{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;
    $data_length -= 2;

    unless ($xlate_h{PAGE_LENGTH} == $data_length) {
        $self->_set_err(-msg => "Error: PAGE_LENGTH field does not match "
                        . "actual length of page.");
        $self->sanlog(-msg => "PAGE_LENGTH var: "
                        . $xlate_h{PAGE_LENGTH}
                        . " Actual page length: "
                        . $data_length,
                      -msg_level => 3);
        return;
    }

    my $len_mpp_desc = $xlate_h{PAGE_LENGTH};
    my @desc_list    = [];
    while ($len_mpp_desc) {
        if ($len_mpp_desc < 4) {
            $self->_set_err(  "MODE PAGE POLICY DESCRIPTOR: Insufficient data "
                            . "length: "
                            . $len_mpp_desc);
            return;
        }
        tie my %mpp_desc, 'Tie::IxHash';

        my $next_byte = hex shift @$s_data;
        $mpp_desc{RESV_B0_bits6_7} = (($next_byte & (BIT_7 | BIT_6)) >> 6);
        $mpp_desc{POLICY_PAGE_CODE} =
          ($next_byte & (BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0));
        $mpp_desc{POLICY_SUBPAGE_CODE} = hex shift @$s_data;
        $next_byte                     = hex shift @$s_data;
        $mpp_desc{MLUS}                = (($next_byte & BIT_7) >> 7);
        $mpp_desc{RESV_B2_bits2_6} =
          (($next_byte & (BIT_6 | BIT_5 | BIT_4 | BIT_3 | BIT_2)) >> 2);
        $mpp_desc{MODE_PAGE_POLICY} = ($next_byte & (BIT_1 | BIT_0));
        $mpp_desc{RESV_B3} = hex shift @$s_data;
        push @desc_list, \%mpp_desc;
        $len_mpp_desc -= 4;
    }

    $xlate_h{DESC_LIST} = \@desc_list;

    return \%xlate_h;

}    # end sub xlate_inq_pg_87h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_B0h_bytes()
# Objective: Return a hash with values set for Block Limits VPD page
# Details:   The returned hash will have similar key and value pairs
#            Based on SBC-3 Specification - Revision 22
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE'                             => 0,
#          'PERIPHERAL_QUALIFIER'                               => 0,
#          'VPD_PAGE_CODE'                                      => 0xB0,
#          'RESERVED_BYTE_2'                                    => 0,
#          'PAGE_LENGTH'                                        => 24,
#          'RESERVED_BYTE_4'                                    => 0,
#          'MAXIMUM_COMPARE_AND_WRITE_LENGTH'                   => 255,
#          'OPTIMAL_TRANSFER_LENGTH_GRANULARITY'                => 0,
#          'MAXIMUM_TRANSFER_LENGTH'                            => 0,
#          'OPTIMAL_TRANSFER_LENGTH'                            => 0,
#          'MAXIMUM_PREFETCH_XDREAD_XDWRITE_TRANSFER_LENGTH'    => 0,
#          'MAXIMUM_UNMAP_LBA_COUNT'                            => 0,
#          'MAXIMUM_UNMAP_DESCRIPTOR_COUNT'                     => 0,
#          'OPTIMAL_UNMAP_GRANULARITY'                          => 0,
#          'UGAVALID'                                           => 0,
#          'UNMAP_GRANULARITY_ALIGNMENT'                        => 0,
#          'RESERVED_BYTES_36_63'                               => 0,
#        };
#
########################################################################

sub xlate_inq_pg_B0h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_length = scalar @$s_data;
    if ($data_length < 1) {
        $self->_set_err("BYTE_0: Insufficient data length: " . $data_length);
        return;
    }
    my $byte0 = hex shift @$s_data;
    $data_length--;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    if ($data_length < 1) {
        $self->_set_err(
                    "VPD_PAGE_CODE: Insufficient data length: " . $data_length);
        return;
    }
    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $data_length--;

    if ($data_length < 1) {
        $self->_set_err(
                  "RESERVED_BYTE_2: Insufficient data length: " . $data_length);
        return;
    }
    $xlate_h{RESERVED_BYTE_2} = hex shift @$s_data;
    $data_length--;

    if ($data_length < 1) {
        $self->_set_err(
                      "PAGE_LENGTH: Insufficient data length: " . $data_length);
        return;
    }
    $xlate_h{PAGE_LENGTH} = hex shift @$s_data;
    $data_length--;

    my $test_data_length = scalar(@$s_data);
    unless ($data_length == $test_data_length) {
        $self->_set_err(-msg => "Data_length var does not match length of "
                        . "actual data.");
        $self->sanlog(-msg => "Data_length var: "
                        . $data_length
                        . " Actual data length: "
                        . $test_data_length,
                      -msg_level => 3);
        return;
    }

    if ($data_length < 1) {
        $self->_set_err(
                  "RESERVED_BYTE_4: Insufficient data length: " . $data_length);
        return;
    }

    my $next_byte = hex shift @$s_data;
    $xlate_h{RES_B4_bits_1_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1))
      >> 1;

    $xlate_h{WSNZ} = ($next_byte & BIT_0);
    $data_length--;

    if ($data_length < 1) {
        $self->_set_err(
                  "MAXIMUM_COMPARE_AND_WRITE_LENGTH: Insufficient data length: "
                    . $data_length);
        return;
    }
    $xlate_h{MAXIMUM_COMPARE_AND_WRITE_LENGTH} = hex shift @$s_data;
    $data_length--;

    if ($data_length < 2) {
        $self->_set_err(
               "OPTIMAL_TRANSFER_LENGTH_GRANULARITY: Insufficient data length: "
                 . $data_length);
        return;
    }
    $xlate_h{OPTIMAL_TRANSFER_LENGTH_GRANULARITY} = hex join '',
      splice @$s_data, 0, 2;
    $data_length -= 2;

    if ($data_length < 4) {
        $self->_set_err("MAXIMUM_TRANSFER_LENGTH: Insufficient data length: "
                        . $data_length);
        return;
    }
    $xlate_h{MAXIMUM_TRANSFER_LENGTH} = hex join '', splice @$s_data, 0, 4;
    $data_length -= 4;

    if ($data_length < 4) {
        $self->_set_err("OPTIMAL_TRANSFER_LENGTH: Insufficient data length: "
                        . $data_length);
        return;
    }
    $xlate_h{OPTIMAL_TRANSFER_LENGTH} = hex join '', splice @$s_data, 0, 4;
    $data_length -= 4;

    if ($data_length < 4) {
        $self->_set_err(  "MAXIMUM_PREFETCH_XDREAD_XDWRITE_TRANSFER_LENGTH: "
                        . "Insufficient data length: "
                        . $data_length);
        return;
    }
    $xlate_h{MAXIMUM_PREFETCH_XDREAD_XDWRITE_TRANSFER_LENGTH} = hex join '',
      splice @$s_data, 0, 4;
    $data_length -= 4;

    if ($data_length < 4) {
        $self->_set_err("MAXIMUM_UNMAP_LBA_COUNT: Insufficient data length: "
                        . $data_length);
        return;
    }
    $xlate_h{MAXIMUM_UNMAP_LBA_COUNT} = hex join '', splice @$s_data, 0, 4;
    $data_length -= 4;

    if ($data_length < 4) {
        $self->_set_err(  "MAXIMUM_UNMAP_DESCRIPTOR_COUNT: "
                        . "Insufficient data length: "
                        . $data_length);
        return;
    }
    $xlate_h{MAXIMUM_UNMAP_DESCRIPTOR_COUNT} = hex join '', splice @$s_data, 0,
      4;
    $data_length -= 4;

    ## These are not yet supported, but will be, so from here on out,
    ## return the hash even though the error is set
    if ($data_length < 4) {
        $self->_set_err(  "OPTIMAL_UNMAP_GRANULARITY: "
                        . "Insufficient data length: "
                        . $data_length);
        return \%xlate_h;
    }
    $xlate_h{OPTIMAL_UNMAP_GRANULARITY} = hex join '', splice @$s_data, 0, 4;
    $data_length -= 4;

    if ($data_length < 4) {
        $self->_set_err(
                   "UGA Bytes: " . "Insufficient data length: " . $data_length);
        return \%xlate_h;
    }
    my $UGA_Bytes      = hex join '', splice @$s_data, 0, 4;
    my $UGA_VALID_MASK = 0x80000000;
    my $UGA_MASK       = 0x7FFFFFFF;
    $xlate_h{UGAVALID} = ($UGA_Bytes & $UGA_VALID_MASK) >> 31;
    $xlate_h{UNMAP_GRANULARITY_ALIGNMENT} = ($UGA_Bytes & $UGA_MASK);

    if ($data_length < 8) {
        $self->_set_err(  "MAXIMUM WRITE SAME LENGTH: "
                        . "Insufficient data length: "
                        . $data_length);
        return \%xlate_h;
    }

    $xlate_h{MAX_WRITE_SAME_LEN} = hex join '', splice @$s_data, 0, 8;

    if ($data_length < 20) {
        $self->_set_err(  "MAXIMUM WRITE SAME LENGTH: "
                        . "Insufficient data length: "
                        . $data_length);
        return \%xlate_h;
    }

    $xlate_h{RESERVED_BYTES_44_63} = hex join '', splice @$s_data, 0, 20;

    return \%xlate_h;

}    # end sub xlate_inq_pg_B0h_bytes

########################################################################
# Sub Name:  xlate_inq_pg_B2h_bytes()
# Objective: Return a hash with values set for Logical Block Provisioning
#            VPD page
# Details:   The returned hash will have similar key and value pairs
#
#$VAR1 = {
#          'PERIPHERAL_DEVICE_TYPE'                             => 0,
#          'PERIPHERAL_QUALIFIER'                               => 0,
#          'VPD_PAGE_CODE'                                      => 0xB2,
#          'PAGE_LENGTH'                                        => n-3,
#          'THRESHOLD_EXPONENT'                                 => 0,
#          'LBPU'                                               => 1,
#          'LBPWS'                                              => 1,
#          'RES_B5_bits_2_5'                                    => 0,
#          'ANC_SUP'                                            => 0,
#          'DP'                                                 => 0,
#          'RES_B6_bits_3_7'                                    => 0,
#          'PROVISIONING_TYPE'                                  => 0,
#          'RESERVED_BYTE_7'                                    => 0,
#          'PROVISIONING_GROUP_DESCRIPTOR'                      => 0,
#        };
#
########################################################################

sub xlate_inq_pg_B2h_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_length = scalar @$s_data;
    if ($data_length < 1) {
        $self->_set_err("BYTE_0: Insufficient data length: " . $data_length);
        return;
    }

    $self->sanlog(
                 -msg => "In B2h bytes, data length is: " . $data_length . "\n",
                 -msg_level => 1);

    my $byte0 = hex shift @$s_data;
    $data_length--;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;

    if ($data_length < 1) {
        $self->_set_err(
                 "Insufficient data length for VPD_PAGE_CODE: " . $data_length);
        return;
    }
    $xlate_h{VPD_PAGE_CODE} = hex shift @$s_data;
    $data_length--;

    if ($data_length < 2) {
        $self->_set_err(
                   "Insufficient data length for PAGE_LENGTH: " . $data_length);
        return;
    }
    $xlate_h{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;
    $data_length -= 2;

    if ($data_length < 1) {
        $self->_set_err(
            "Insufficient data length for THRESHOLD_EXPONENT: " . $data_length);
        return;
    }
    $xlate_h{THRESHOLD_EXPONENT} = hex shift @$s_data;
    $data_length--;

    if ($data_length < 1) {
        $self->_set_err("Insufficient data length for BYTE 5: " . $data_length);
        return;
    }
    my $next_byte = hex shift @$s_data;
    $data_length--;
    $xlate_h{LBPU}            = ($next_byte & BIT_7) >> 7;
    $xlate_h{LBPWS}           = ($next_byte & BIT_6) >> 6;
    $xlate_h{RES_B5_bits_3_5} = ($next_byte & (BIT_5 | BIT_4 | BIT_3)) >> 3;
    $xlate_h{LBPRZ}           = ($next_byte & BIT_2) >> 2;
    $xlate_h{ANC_SUP}         = ($next_byte & BIT_1) >> 1;
    $xlate_h{DP}              = $next_byte & BIT_0;

    if ($data_length < 1) {
        $self->_set_err("Insufficient data length for BYTE 6: " . $data_length);
        return;
    }
    $next_byte = hex shift @$s_data;
    $data_length--;
    $xlate_h{RES_B6_bits_3_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4 | BIT_3)) >> 3;
    $xlate_h{PROVISIONING_TYPE} = ($next_byte & (BIT_2 | BIT_1 | BIT_0));

    if ($data_length < 1) {
        $self->_set_err(
                   "Insufficient data length RESERVED_BYTE_7: " . $data_length);
        return;
    }
    $xlate_h{RESERVED_BYTE_7} = hex shift @$s_data;
    $data_length--;

    my $desc_len = $xlate_h{PAGE_LENGTH} - 4;
    if ($data_length < $desc_len) {
        $self->_set_err(
                  "Insufficient data length for PROVISIONING_GROUP_DESCRIPTOR: "
                    . $data_length);
        return;
    }
    $xlate_h{PROVISIONING_GROUP_DESCRIPTOR} = hex join '', splice @$s_data, 0,
      $desc_len;

    return \%xlate_h;
}    # end sub xlate_inq_pg_B2h_bytes

########################################################################
# Sub Name:  xlate_sense
# Objective: Return a hash with values set for Sense Data
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
#            The format of Information, Command Specific Information, and
#            Sense Key Specific varies based on the command opcode and the
#            Sense Key.  See the following methods for formats of sub hashes:
#             - xlate_sense_command_specific_information*,
#             - xlate_sense_information*,
#             - xlate_sense_sks*
#             - xlate_sense_additional*
#
# Params     s_data arrayref each array index holds 1 byte of sense
#            data in a hex string format.  ie 0a, 0b, 0c, etc
#            cmd int the original opcode that caused sense data to
#            to be returned
#
#$VAR1 = {
#          'VALID' => 1,
#          'RESPONSE_CODE' => 112,
#          'FILEMARK' => 1,
#          'EOM' => 0,
#          'ILI' => 1,
#          'SENSE_KEY' => 10,
#          'INFORMATION' => { ... },
#          'ADDITIONAL_SENSE_LENGTH' => 14,
#          'COMMAND_SPECIFIC_INFORMATION' => { ... },
#          'ADDITIONAL_SENSE_CODE' => 32,
#          'ADDITIONAL_SENSE_CODE_QUALIFIER' => 255,
#          'FIELD_REPLACEABLE_UNIT_CODE' => 255,
#          'SKSV' => 1,
#          'SENSE_KEY_SPECIFIC' => { ... },
#          'ADDITIONAL_SENSE_BYTES' => { ... }
#        };
########################################################################
sub xlate_sense {

    my ($self, %args) = @_;
    SanUtil::strip_dash(%args);
    $self->_clear_err();

    my @s_data = @{$args{'s_data'}};
    my $cmd    = $args{'cmd'};

    # Parse the Fixed Format Sense Data (first 18 bytes)
    tie my %xlate_h, 'Tie::IxHash';
    %xlate_h =
      %{$self->xlate_sense_fixed_format('s_data' => \@s_data, 'cmd' => $cmd)};

    if ($self->error()) {
        return {};
    }

    # If Command Opcode is XCcopy
    if ($cmd == 0x83) {

        $self->sanlog(-msg => ' xlate_sense fixed' . Dumper(\%xlate_h));

        my $src_dev_offset =
          $xlate_h{'COMMAND_SPECIFIC_INFORMATION'}->{'SRC_DEV_SENSE_OFFSET'};

        my $dest_dev_offset =
          $xlate_h{'COMMAND_SPECIFIC_INFORMATION'}->{'DEST_DEV_SENSE_OFFSET'};

        # Parse the additional sense bytes following the first 18 bytes
        $xlate_h{'ADDITIONAL_SENSE_BYTES'} =
          $self->xlate_sense_additional_xcopy(
                                           's_data'          => \@s_data,
                                           'src_dev_offset'  => $src_dev_offset,
                                           'dest_dev_offset' => $dest_dev_offset
          );

        if ($self->error()) {
            return {};
        }
    }

    return \%xlate_h;
}

########################################################################
# Sub Name:  xlate_sense_fixed_format
# Objective: Return a hash with values set for Fixed Format Data.  This
#            differs from xlate_sense in that it doesn't attempt to expand
#            the additional sense data; it only expands the first 18 bytes.
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
#            The format of Information, Command Specific Information, and
#            Sense Key Specific varies based on the command opcode and the
#            Sense Key.  See the following methods for formats of sub hashes:
#             - xlate_sense_command_specific_information*,
#             - xlate_sense_information*,
#             - xlate_sense_sks*
#             - xlate_sense_additional*
#
# Params     s_data arrayref each array index holds 1 byte of sense
#            data in a hex string format.  ie 0a, 0b, 0c, etc
#            cmd int the original opcode that caused sense data to
#            to be returned
#
#$VAR1 = {
#          'VALID' => 1,
#          'RESPONSE_CODE' => 112,
#          'FILEMARK' => 1,
#          'EOM' => 0,
#          'ILI' => 1,
#          'SENSE_KEY' => 10,
#          'INFORMATION' => { ... },
#          'ADDITIONAL_SENSE_LENGTH' => 14,
#          'COMMAND_SPECIFIC_INFORMATION' => { ... },
#          'ADDITIONAL_SENSE_CODE' => 32,
#          'ADDITIONAL_SENSE_CODE_QUALIFIER' => 255,
#          'FIELD_REPLACEABLE_UNIT_CODE' => 255,
#          'SKSV' => 1,
#          'SENSE_KEY_SPECIFIC' => { ... },
#          'ADDITIONAL_SENSE_BYTES' => { ... }
#        };
########################################################################
sub xlate_sense_fixed_format {

    my ($self, %args) = @_;
    SanUtil::strip_dash(%args);
    $self->_clear_err();

    my @s_data = @{$args{'s_data'}};
    my $cmd    = $args{'cmd'};

    tie my %xlate_h,  'Tie::IxHash';
    tie my %info_h,   'Tie::IxHash';
    tie my %csinfo_h, 'Tie::IxHash';
    tie my %sks_h,    'Tie::IxHash';

    # Offset 0, Bit 7
    my $byte_0 = hex $s_data[0];
    $xlate_h{'VALID'} = ($byte_0 & (BIT_7)) >> 7;

    # Offset 0, Bits 6 - 0
    $xlate_h{'RESPONSE_CODE'} =
      $byte_0 & (BIT_6 | BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);

    if ($xlate_h{'RESPONSE_CODE'} != 0x70) {
        $self->_set_err('Unexpected RC:  ' . $xlate_h{'RESPONSE_CODE'});
        return {};
    }

    # Offset 1
    my $byte_1 = hex $s_data[1];
    $xlate_h{'OBSOLETE'} = hex $byte_1;

    # Offset 2, Bit 7
    my $byte_2 = hex $s_data[2];
    $xlate_h{'FILEMARK'} = ($byte_2 & (BIT_7)) >> 7;

    # Offset 2, Bit 6
    $xlate_h{'EOM'} = ($byte_2 & (BIT_6)) >> 6;

    # Offset 2, Bit 5
    $xlate_h{'ILI'} = ($byte_2 & (BIT_5)) >> 5;

    # Offset 2, Bit 3-0
    $xlate_h{'SENSE_KEY'} = ($byte_2 & (BIT_3 | BIT_2 | BIT_1 | BIT_0));

    # Offsets 3-6
    my @info_data = @s_data[3 .. 6];

    if ($xlate_h{'VALID'}) {

        if ($cmd == 0x83) {

            %info_h = %{$self->xlate_sense_information_xcopy(\@info_data)};

        } else {

            $info_h{'HEX'} = \@info_data;

        }

    } else {

        $info_h{'HEX'} = \@info_data;

    }
    $xlate_h{'INFORMATION'} = \%info_h;

    # Offset 7
    $xlate_h{'ADDITIONAL_SENSE_LENGTH'} = hex $s_data[7];

    # Offsets 8-11
    my @csinfo_data = @s_data[8 .. 11];

    if ($cmd == 0x83) {

        %csinfo_h = %{
            $self->xlate_sense_command_specific_information_xcopy(\@csinfo_data)
          };

    } else {

        $csinfo_h{'HEX'} = \@csinfo_data;

    }
    $xlate_h{'COMMAND_SPECIFIC_INFORMATION'} = \%csinfo_h;

    # Offset 12
    $xlate_h{'ADDITIONAL_SENSE_CODE'} = hex $s_data[12];

    # Offset 13
    $xlate_h{'ADDITIONAL_SENSE_CODE_QUALIFIER'} = hex $s_data[13];

    # Offset 14 - Format not described by spc4r32
    $xlate_h{'FIELD_REPLACEABLE_UNIT_CODE'} = hex $s_data[14];

    # Offset 15, Bit 7
    $xlate_h{'SKSV'} = ((hex $s_data[15]) & (BIT_7)) >> 7;
    my @sks_data = @s_data[15 .. 17];

    # If SKSV equals 1, the format of Sense Key Specific is defined by SPC4
    if ($xlate_h{'SKSV'}) {

        # Sense Key Specific Fields will difer based on Sense Key
        if ($xlate_h{'SENSE_KEY'} == 0x05) {

            %sks_h = %{$self->xlate_sense_sks_illegal_request(\@sks_data)};

        } elsif ($xlate_h{'SENSE_KEY'} == 0x0A) {

            %sks_h = %{$self->xlate_sense_sks_copy_aborted(\@sks_data)};

        } else {

            # Parsing for other sense keys not implemented
            $sks_h{'HEX'} = \@sks_data;
        }

    } else {

        # SKSV is not set, so format is undefined by spc4
        $sks_h{'HEX'} = \@sks_data;
    }
    $xlate_h{'SENSE_KEY_SPECIFIC'} = \%sks_h;

    # Offsets 18 - n
    my @addtl_sense_bytes = @s_data[18 .. (scalar @s_data - 1)];
    tie my %additional_h, 'Tie::IxHash';
    $additional_h{'HEX'}               = \@addtl_sense_bytes;
    $xlate_h{'ADDITIONAL_SENSE_BYTES'} = \%additional_h;

    return \%xlate_h;
}

########################################################################
# Sub Name:  xlate_sense_additional_xcopy
# Objective: Return a hash with values set for additional sense bytes
#            when the original command was an XCOPY.
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
#
# 'ADDITIONAL_SENSE_BYTES' => {
#      'DEST_DEV_ADDITIONAL_SENSE' => { ... },
#      'DEST_DEV_STATUS_BYTE' => 2,
#      'SRC_DEV_ADDITIONAL_SENSE' => { ... },
#      'SRC_DEV_STATUS_BYTE' => 2,
# }
########################################################################
sub xlate_sense_additional_xcopy {

    my ($self, %args) = @_;
    SanUtil::strip_dash(%args);
    $self->_clear_err();

    tie my %additional_h, 'Tie::IxHash';

    my @s_data = @{$args{'s_data'}};

    my $err_msg;
    my $sense_len            = @s_data;
    my $additional_sense_len = $sense_len - FIXED_FORMAT_SENSE_LEN;
    my $src_dev_offset       = $args{'src_dev_offset'};
    my $dest_dev_offset      = $args{'dest_dev_offset'};

    my $dest_dev_sense_len =
      ($dest_dev_offset == 0) ? 0 : $sense_len - $dest_dev_offset;
    my $src_dev_sense_len =
      ($src_dev_offset == 0)
      ? 0
      : $additional_sense_len - $dest_dev_sense_len;

    if ($src_dev_sense_len == 19) {
        tie my %src_sense_h, 'Tie::IxHash';
        my @src_addtl_data =
          @s_data[$src_dev_offset .. ($src_dev_offset + FIXED_FORMAT_SENSE_LEN)
          ];

        $additional_h{"SRC_DEV_STATUS_BYTE"} = hex shift @src_addtl_data;

        %src_sense_h =
          %{$self->xlate_sense_fixed_format('s_data' => \@src_addtl_data)};

        $additional_h{"SRC_DEV_ADDITIONAL_SENSE"} = \%src_sense_h;

    } elsif ($src_dev_sense_len == 1) {
        $additional_h{"SRC_DEV_STATUS_BYTE"} = hex $s_data[$src_dev_offset];
    }

    if ($dest_dev_sense_len == 19) {
        tie my %dest_sense_h, 'Tie::IxHash';
        my @dest_addtl_data =
          @s_data[$dest_dev_offset .. (
                                    $dest_dev_offset + FIXED_FORMAT_SENSE_LEN)];

        $additional_h{"DEST_DEV_STATUS_BYTE"} = hex shift @dest_addtl_data;

        %dest_sense_h =
          %{$self->xlate_sense_fixed_format('s_data' => \@dest_addtl_data)};

        $additional_h{"DEST_DEV_ADDITIONAL_SENSE"} = \%dest_sense_h;

    } elsif ($dest_dev_sense_len == 1) {
        $additional_h{"DEST_DEV_STATUS_BYTE"} = hex $s_data[$dest_dev_offset];
    }

    print Dumper(\%additional_h);

    return \%additional_h;
}

########################################################################
# Sub Name:  xlate_sense_information_xcopy
# Objective: Return a hash with values set for Information
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
# Params     s_data arrayref each array index holds 1 byte of
#            Information Sense Data in a hex string format.
#            ie (0a, 0b, 0c, 0d)
#
# 'INFORMATION' => {
#     'RESIDUAL' => '11259375'
# }
#
########################################################################
sub xlate_sense_information_xcopy {

    my ($self, $s_data) = @_;
    tie my %info_h, 'Tie::IxHash';

    $info_h{'RESIDUAL'} = hex join '', splice @$s_data, 0, 4;

    return \%info_h;
}

########################################################################
# Sub Name:  xlate_sense_command_specific_information_xcopy
# Objective: Return a hash with values set for Command Specific Information
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
# Params     s_data arrayref each array index holds 1 byte of Command
#            Specific Information Sense Data in a hex string format.
#            ie (0a, 0b, 0c, 0d)
#
# 'COMMAND_SPECIFIC_INFORMATION' => {
#     'SRC_DEV_SENSE_OFFSET' => '240',
#     'DEST_DEV_SENSE_OFFSET' => '163',
#     'SEGMENT_DESC_INDEX' => '42411'
# }
########################################################################
sub xlate_sense_command_specific_information_xcopy {

    my ($self, $s_data) = @_;
    tie my %csinfo_h, 'Tie::IxHash';

    # 1st Byte
    # If error was in source device
    $csinfo_h{'SRC_DEV_SENSE_OFFSET'} = hex shift @$s_data;

    # 2nd byte
    # If error was in destination device
    $csinfo_h{'DEST_DEV_SENSE_OFFSET'} = hex shift @$s_data;

    # 3rd and 4th Bytes
    # Segment Number based on relative position of segment desc in parameter list
    # First Segment Desc is 0, Second is 1, etc
    $csinfo_h{'SEGMENT_DESC_INDEX'} = hex join '', splice @$s_data, 0, 2;

    return \%csinfo_h;
}

########################################################################
# Sub Name:  xlate_sense_sks_illegal_request
# Objective: Return a hash with values set for SKSV when the Sense Key
#            is Illegal Request.
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
# Params     s_data arrayref each array index holds 1 byte of
#            SKSV Sense Data in a hex string format.
#            ie (0a, 0b, 0c, 0d)
#
# 'SENSE_KEY_SPECIFIC' => {
#      'SKSV' => '1',
#      'C_D' => '1',
#      'BPV' => '1',
#      'BIT_POINTER' => '5',
#      'FIELD_POINTER' => '239'
# }
#
########################################################################
sub xlate_sense_sks_illegal_request {
    my ($self, $s_data) = @_;

    tie my %sks_h, 'Tie::IxHash';
    my $byte_0 = hex shift @$s_data;

    $sks_h{'SKSV'} = ($byte_0 & (BIT_7)) >> 7;

    # C/D - Byte 15, Bit 6
    $sks_h{'C_D'} = ($byte_0 & (BIT_6)) >> 6;

    # BPV - Byte 15, Bit 3
    $sks_h{'BPV'} = ($byte_0 & (BIT_3)) >> 3;

    # Bit Pointer - Byte 15, Bit 2-0
    $sks_h{'BIT_POINTER'} = ($byte_0 & (BIT_2 | BIT_1 | BIT_0));

    # Field Pointer - Bytes 16, 17
    $sks_h{'FIELD_POINTER'} = hex join '', splice @$s_data, 0, 2;

    return \%sks_h;
}

########################################################################
# Sub Name:  xlate_sense_sks_copy_aborted
# Objective: Return a hash with values set for SKSV when the Sense Key
#            is Copy Aborted.
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
# Params     s_data arrayref each array index holds 1 byte of
#            SKSV Sense Data in a hex string format.
#            ie (0a, 0b, 0c, 0d)
#
# 'SENSE_KEY_SPECIFIC' => {
#     'SD' => '0',
#     'BPV' => '1',
#     'BIT_POINTER' => '5',
#     'FIELD_POINTER' => '239'
# }
#
########################################################################
sub xlate_sense_sks_copy_aborted {
    my ($self, $s_data) = @_;

    tie my %sks_h, 'Tie::IxHash';
    my $byte_0 = hex shift @$s_data;

    # C/D - Byte 15, Bit 5
    $sks_h{'SD'} = ($byte_0 & (BIT_5)) >> 5;

    # BPV - Byte 15, Bit 3
    $sks_h{'BPV'} = ($byte_0 & (BIT_3)) >> 3;

    # Bit Pointer - Byte 15, Bit 2-0
    $sks_h{'BIT_POINTER'} = ($byte_0 & (BIT_2 | BIT_1 | BIT_0));

    # Field Pointer - Bytes 16, 17
    $sks_h{'FIELD_POINTER'} = hex join '', splice @$s_data, 0, 2;

    return \%sks_h;
}

########################################################################
# Sub Name:  xlate_std_inq_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The hash returned will have similar key and values pairs
#
#$VAR1 = {
#          'PERIPHERAL_QUALIFIER' => 0,
#          'PERIPHERAL_DEVICE_TYPE' => 0,
#          'RMB' => 0,
#          'RESV_B1_7bits' => 0,
#          'VERSION' => 4,
#          'RESPONSE_DATA_FORMAT' => 2,
#          'HISUP' => 1,
#          'NORMACA' => 1,
#          'OBS_B3_bit6' => 0,
#          'OBS_B3_bit7' => 0,
#          'ADDITIONAL_LENGTH' => 112,
#          'PROTECT' => 0,
#          'RESV_B5_bits12' => 0,
#          '3PC' => 0,
#          'TPGS' => 0,
#          'ACC' => 0,
#          'SCCS' => 0,
#          'ADDR16' => 0,
#          'OBS_B6_bit1' => 0,
#          'OBS_B6_bit2' => 0,
#          'MCHNGR' => 0,
#          'MULTIP' => 1,
#          'B6_VS' => 0,
#          'ENCSERV' => 0,
#          'BQUE' => 0,
#          'VS' => 0,
#          'CMDQUE' => 1,
#          'OBS_B7_bit2' => 0,
#          'LINKED' => 0,
#          'SYNC' => 0,
#          'WBUS16' => 0,
#          'OBS_B7_bit6' => 0,
#          'OBS_B7_bit7' => 0,
#          'T10_VENDOR_IDENTIFICATION' => 'NETAPP  ',
#          'PRODUCT_IDENTIFICATION' => 'LUN             ',
#          'PRODUCT_REVISION_LEVEL' => '8000',
#          'VENDOR_SPECIFIC' => 'HnS7XJNi9VeZ^@^@^@^@^@^@^@^@',
#          'IUS' => 0,
#          'QAS' => 0,
#          'CLOCKING' => 0,
#          'RESV_B56_4bits' => 0,
#          'RESV_B57' => 0,
#          'VERSION_DESCRIPTOR_1' => 0,
#          'VERSION_DESCRIPTOR_2' => 0,
#          'VERSION_DESCRIPTOR_3' => 0,
#          'VERSION_DESCRIPTOR_4' => 0,
#          'VERSION_DESCRIPTOR_5' => 0,
#          'VERSION_DESCRIPTOR_6' => 0,
#          'VERSION_DESCRIPTOR_7' => 0,
#          'VERSION_DESCRIPTOR_8' => 0,
#          'RESV_B74_95' => '00000000000000000000000000000000000000000000',
#          'VENDOR_SPECIFIC_FILER' => 'FAS3070         ',
#          'VENDOR_SPECIFIC_ONTAP' => 'ONTAP'
#        };
#
########################################################################

sub xlate_std_inq_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    #tie my %xlate_h, 'Tie::Hash::Indexed';
    tie my %xlate_h, 'Tie::IxHash';

    my $byte0 = hex shift @$s_data;

    $xlate_h{PERIPHERAL_QUALIFIER} = ($byte0 & (BIT_7 | BIT_6 | BIT_5)) >> 5;
    $xlate_h{PERIPHERAL_DEVICE_TYPE} =
      $byte0 & (BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);

    my $byte1 = hex shift @$s_data;
    $xlate_h{RMB}           = ($byte1 & (BIT_7)) >> 7;
    $xlate_h{RESV_B1_7bits} = ($byte1 & (~BIT_7)) >> 1;

    $xlate_h{VERSION} = hex shift @$s_data;

    my $byte3 = hex shift @$s_data;
    $xlate_h{RESPONSE_DATA_FORMAT} = $byte3 & (BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_h{HISUP}                = ($byte3 & BIT_4) >> 4;
    $xlate_h{NORMACA}              = ($byte3 & BIT_5) >> 5;
    $xlate_h{OBS_B3_bit6}          = ($byte3 & BIT_6) >> 6;
    $xlate_h{OBS_B3_bit7}          = ($byte3 & BIT_7) >> 7;

    $xlate_h{ADDITIONAL_LENGTH} = hex shift @$s_data;

    unless ($xlate_h{ADDITIONAL_LENGTH} > 31) {
        $self->_set_err("ERROR: Std Inq Data cannot be less than 36 bytes");
        return;
    }

    my $byte5 = hex shift @$s_data;
    $xlate_h{PROTECT}        = $byte5 & BIT_0;
    $xlate_h{RESV_B5_bits12} = $byte5 & (BIT_2 | BIT_1);
    $xlate_h{'3PC'}          = ($byte5 & BIT_3) >> 3;
    $xlate_h{TPGS}           = ($byte5 & (BIT_4 | BIT_5)) >> 4;
    $xlate_h{ACC}            = ($byte5 & BIT_6) >> 6;
    $xlate_h{SCCS}           = ($byte5 & BIT_7) >> 7;

    my $byte6 = hex shift @$s_data;
    $xlate_h{ADDR16}      = $byte6 & BIT_0;
    $xlate_h{OBS_B6_bit1} = ($byte6 & BIT_1) >> 1;
    $xlate_h{OBS_B6_bit2} = ($byte6 & BIT_2) >> 2;
    $xlate_h{MCHNGR}      = ($byte6 & BIT_3) >> 3;
    $xlate_h{MULTIP}      = ($byte6 & BIT_4) >> 4;
    $xlate_h{B6_VS}       = ($byte6 & BIT_5) >> 5;
    $xlate_h{ENCSERV}     = ($byte6 & BIT_6) >> 6;
    $xlate_h{BQUE}        = ($byte6 & BIT_7) >> 7;

    my $byte7 = hex shift @$s_data;
    $xlate_h{VS}          = $byte7 & BIT_0;
    $xlate_h{CMDQUE}      = ($byte7 & BIT_1) >> 1;
    $xlate_h{OBS_B7_bit2} = ($byte7 & BIT_2) >> 2;
    $xlate_h{LINKED}      = ($byte7 & BIT_3) >> 3;
    $xlate_h{SYNC}        = ($byte7 & BIT_4) >> 4;
    $xlate_h{WBUS16}      = ($byte7 & BIT_5) >> 5;
    $xlate_h{OBS_B7_bit6} = ($byte7 & BIT_6) >> 6;
    $xlate_h{OBS_B7_bit7} = ($byte7 & BIT_7) >> 7;

    $xlate_h{T10_VENDOR_IDENTIFICATION} =
      pack("H*", join '', splice @$s_data, 0, 8);
    $xlate_h{PRODUCT_IDENTIFICATION} =
      pack("H*", join '', splice @$s_data, 0, 16);
    $xlate_h{PRODUCT_REVISION_LEVEL} =
      pack("H*", join '', splice @$s_data, 0, 4);
    $xlate_h{VENDOR_SPECIFIC} = pack("H*", join '', splice @$s_data, 0, 20);

    my $byte56 = hex shift @$s_data;
    $xlate_h{IUS}            = $byte56 & BIT_0;
    $xlate_h{QAS}            = ($byte56 & BIT_1) >> 1;
    $xlate_h{CLOCKING}       = ($byte56 & (BIT_2 | BIT_3)) >> 2;
    $xlate_h{RESV_B56_4bits} = ($byte56 & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;

    $xlate_h{RESV_B57} = hex shift @$s_data;

    $xlate_h{VERSION_DESCRIPTOR_1} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_2} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_3} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_4} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_5} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_6} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_7} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{VERSION_DESCRIPTOR_8} = hex join '', splice @$s_data, 0, 2;

    $xlate_h{RESV_B74_95} = join '', splice @$s_data, 0, 22;

    $xlate_h{VENDOR_SPECIFIC_FILER} =
      pack("H*", join '', splice @$s_data, 0, 16);

    $xlate_h{VENDOR_SPECIFIC_ONTAP} =
      pack("H*", join '', splice @$s_data, 0, 5);

    return \%xlate_h;

} ## end sub xlate_std_inq_bytes

########################################################################
# Sub Name:  xlate_report_luns_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
########################################################################

sub xlate_report_luns_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    my (@tmp, %xlate_str, $idx);

    # length of data returned is the first 4 bytes in the data
    # so should take a splice of the array to get the length

    @tmp = splice @$s_data, 0, 4;

    my $len = hex join '', @tmp;

    unless ($len) {
        $self->_set_err("No luns listed in the data");
        return;
    }

    $xlate_str{num_of_luns} = $len / 8;

    $idx = 0;
    @tmp = splice @$s_data, 0, 4;
    my @report_lun_new = @$s_data;

    # retaining this for backward compatibility
    while (scalar @$s_data) {

        # each lun is 8 bytes long but the first 2 bytes are the only
        # relevant bytes. In the 2 bytes only the lease significant 12 bits
        # should have valid data, while the most significant 4 bits should
        # always be set to zero.
        #
        @tmp = splice @$s_data, 0, 2;

        my $lun_id = hex join '', @tmp;
        $xlate_str{"lun_" . $idx++} = $lun_id;

       # push(@{$xlate_str{'all_luns'}}, $lun_id);

        # make sure that the next 6 bytes are discarded
        @tmp = splice @$s_data, 0, 6;
    } ## end while (scalar @$s_data)

    while (scalar @report_lun_new) {
        @tmp = splice @report_lun_new, 0, 8;
        my $lun_id = hex join '', @tmp;
        push(@{$xlate_str{"all_luns"}}, $lun_id);

    } ## end while (scalar @report_lun_new)

    return \%xlate_str;

} ## end sub xlate_report_luns_bytes

########################################################################
# Sub Name:  xlate_read_capacity_10_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#            Based on SBC-3 Specification - Revision 19
#
#$VAR1 = {
#          'PRETURNED_LOGICAL_BLOCK_ADDRESS' => ???, # 4 bytes
#          'BLOCK_LENGTH_IN_BYTES'           => 512, # 4 bytes
#        };
#
########################################################################
sub xlate_read_capacity_10_bytes {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';
    my $ret_length = scalar(@$s_data);
    $self->sanlog(-msg       => "Length of data: $ret_length",
                  -msg_level => 3);

    # Bytes 0-3: RETURNED_LOGICAL_BLOCK_ADDRESS
    # make sure there's enough data for the next pull
    if ($ret_length >= 4) {
        $ret_length -= 4;
        $xlate_h{RETURNED_LOGICAL_BLOCK_ADDRESS} = hex join '', splice @$s_data,
          0, 4;
    } else {
        $self->_set_err(  "RETURNED_LOGICAL_BLOCK_ADDRESS: Insufficient data "
                        . " length: "
                        . $ret_length);
        return;
    }

    # Bytes 4-7: BLOCK_LENGTH_IN_BYTES
    # make sure there's enough data for the next pull
    if ($ret_length >= 4) {
        $ret_length -= 4;
        $xlate_h{BLOCK_LENGTH_IN_BYTES} = hex join '', splice @$s_data, 0, 4;
    } else {
        $self->_set_err(
             "BLOCK_LENGTH_IN_BYTES: Insufficient data length: " . $ret_length);
    }

    return \%xlate_h;

}    # end of xlate_read_capacity_10_bytes

########################################################################
# Sub Name:  xlate_read_capacity_16_bytes()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#            Based on SBC-3 Specification - Revision 19
#
#$VAR1 = {
#          'PRETURNED_LOGICAL_BLOCK_ADDRESS' 		=> ??,
#          'BLOCK_LENGTH_IN_BYTES' 			=> 512,
#          'RESERVED_BITS_2_7' 				=> 0,
#          'RTO_EN' 					=> 0,
#          'PROT_EN' 					=> 0,
#          'RESERVED_BITS_4_7' 				=> 0,
#          'LOGICAL_BLOCKS_PER_PHYSICAL_BLOCK' 		=> 0,
#          'TPE' 					=> 0,
#          'TPRZ' 					=> 0,
#          'LOWEST_ALLIGNED_LOGICAL_BLOCK_ADDRESS'	=> 0,
#          'RESERVED_BYTES_16_31' 			=> 0,
#        };
#
########################################################################
sub xlate_read_capacity_16_bytes {
    my ($self, $s_data) = @_;
    my $next_byte;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $ret_length = scalar(@$s_data);
    if ($ret_length < 16) {
        $self->_set_err(
                    "Target returned insufficient data length: " . $ret_length);
        return;
    }

    $ret_length -= 8;    #already checked the length
                         # Bytes 0-7: Returned Logical Block Address
    $xlate_h{RETURNED_LOGICAL_BLOCK_ADDRESS} = hex join '', splice @$s_data, 0,
      8;

    # Bytes 8-11: Block Length
    # make sure there's enough data for the next pull
    if ($ret_length >= 4) {
        $ret_length -= 4;
        $xlate_h{BLOCK_LENGTH_IN_BYTES} = hex join '', splice @$s_data, 0, 4;
    } else {
        $self->_set_err(
             "BLOCK_LENGTH_IN_BYTES: Insufficient data length: " . $ret_length);
        return;
    }

    # Byte 12: Reserved/P_TYPE/PROT_EN
    # make sure there's enough data for the next pull
    if ($ret_length >= 1) {
        $ret_length--;
        $next_byte = shift @$s_data;
    } else {
        $self->_set_err("Byte 12: Insufficient data length: " . $ret_length);
        return;
    }
    $xlate_h{RESERVED_BITS_4_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $xlate_h{P_TYPE}  = ($next_byte & BIT_1) >> 1;
    $xlate_h{PROT_EN} = ($next_byte & BIT_0);

    # Byte 13: RESERVED_BITS_4_7/LOGICAL_BLOCKS_PER_PHYSICAL_BLOCK_EXPONENT
    # make sure there's enough data for the next pull
    if ($ret_length >= 1) {
        $ret_length--;
        $next_byte = shift @$s_data;
    } else {
        $self->_set_err("Byte 13: Insufficient data length: " . $ret_length);
        return;
    }
    $xlate_h{RESERVED_BITS_4_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $xlate_h{LOGICAL_BLOCKS_PER_PHYSICAL_BLOCK_EXPONENT} =
      ($next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0));

    # Bytes 14-15: LBPME/LBPRZ/LOWEST_ALIGNED_LOGICAL_BLOCK_ADDRESS
    # make sure there's enough data for the next pull
    if ($ret_length >= 2) {
        $ret_length -= 2;
        $next_byte = hex join '', splice @$s_data, 0, 2;
    } else {
        $self->_set_err(
                       "Bytes 14-15: Insufficient data length: " . $ret_length);
    }

    # Lowest Alligned Logical Block Address uses more than 1 byte so we
    # need to read 2 bytes
    my $LBPME_Mask = 0x8000;    # 1st bit of 2 byte pull
    my $LBPRZ_Mask = 0x4000;    # 2nd bit of 2 byte pull
    my $LALBA_Mask = 0x1FFF;    # the rest of the 2 byte pull
    $xlate_h{LBPME} = ($next_byte & $LBPME_Mask) >> 15;
    $xlate_h{LBPRZ} = ($next_byte & $LBPRZ_Mask) >> 14;
    $xlate_h{LOWEST_ALLIGNED_LOGICAL_BLOCK_ADDRESS} = $next_byte & $LALBA_Mask;

    # Bytes 16-31: RESERVED
    # make sure there's enough data for the next pull
    if ($ret_length >= 16) {
        $ret_length -= 16;
        $xlate_h{RESERVED_BYTES_16_31} = hex join '', splice @$s_data, 0, 16;
    }

    return \%xlate_h;

}    # end of xlate_read_capacity_16_bytes

########################################################################
# Sub Name:  xlate_receive_copy_results()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:   The returned hash will have similar key and value pairs
#            Based on SPC-4 Specification - Revision 32
#
#$VAR1 = {
#        'AVAILABLE_DATA'                    => 42,
#        'RESERVED_4'                        => 0,
#        'SNLID'                             => 1,
#        'RESERVED_5_7'                      => 0,
#        'MAXIMUM_TARGET_DESCRIPTOR_COUNT'   => 4,
#        'MAXIMUM_SEGMENT_DESCRIPTOR_COUNT'  => 4,
#        'MAXIMUM_DESCRIPTOR_LIST_LENGTH'    => 240,
#        'MAXIMUM_SEGMENT_LENGTH'            => 33554432
#        'MAXIMUM_INLINE_DATA_LENGTH'        => 0,
#        'HELD_DATA_LIMIT'                   => 0,
#        'MAXIMUM_STREAM_DATA_TRANSFER_SIZE' => 0,
#        'RESERVED_32_33'                    => 0,
#        'TOTAL_CONCURRENT_COPIES'           => 1000,
#        'MAXIMUM_CONCURRENT_COPIES'         => 0,
#        'DATA_SEGMENT_GRANULARITY'          => 9,
#        'INLINE_DATA_GRANULARITY'           => 0,
#        'HELD_DATA_GRANULARITY'             => 0,
#        'RESERVED_40_42'                    => 0,
#        'IMPLEMENTED_DESCRIPTOR'            => 2,
#        'DESCRIPTORS'                       => [ 0x02, 0xe4 ]
#        };
#
########################################################################
sub xlate_receive_copy_results {
    my ($self, $s_data) = @_;

    $self->_clear_err();
    tie my %xlate_h, 'Tie::IxHash';

    if (scalar(@{$s_data}) != 46) {
        $self->_set_err("RCR output must be exactly 46 bytes.");
        return;
    }

    $xlate_h{'AVAILABLE_DATA'} = hex join '', splice @$s_data, 0, 4;
    my $byte4 = hex shift @$s_data;
    $xlate_h{'RESERVED_4'} =
      ($byte4 & (BIT_7 | BIT_6 | BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1)) >> 1;
    $xlate_h{'SNLID'} = $byte4 & (BIT_0);
    $xlate_h{'RESERVED_5_7'} = hex join '', splice @$s_data, 0, 3;
    $xlate_h{'MAXIMUM_TARGET_DESCRIPTOR_COUNT'} = hex join '', splice @$s_data,
      0, 2;
    $xlate_h{'MAXIMUM_SEGMENT_DESCRIPTOR_COUNT'} = hex join '', splice @$s_data,
      0, 2;
    $xlate_h{'MAXIMUM_DESCRIPTOR_LIST_LENGTH'} = hex join '', splice @$s_data,
      0, 4;
    $xlate_h{'MAXIMUM_SEGMENT_LENGTH'}     = hex join '', splice @$s_data, 0, 4;
    $xlate_h{'MAXIMUM_INLINE_DATA_LENGTH'} = hex join '', splice @$s_data, 0, 4;
    $xlate_h{'HELD_DATA_LIMIT'}            = hex join '', splice @$s_data, 0, 4;
    $xlate_h{'MAXIMUM_STREAM_DATA_TRANSFER_SIZE'} = hex join '',
      splice @$s_data, 0, 4;
    $xlate_h{'RESERVED_32_33'}          = hex join '', splice @$s_data, 0, 2;
    $xlate_h{'TOTAL_CONCURRENT_COPIES'} = hex join '', splice @$s_data, 0, 2;
    $xlate_h{'MAXIMUM_CONCURRENT_COPIES'} = hex shift @$s_data;
    $xlate_h{'DATA_SEGMENT_GRANULARITY'}  = hex shift @$s_data;
    $xlate_h{'INLINE_DATA_GRANULARITY'}   = hex shift @$s_data;
    $xlate_h{'HELD_DATA_GRANULARITY'}     = hex shift @$s_data;
    $xlate_h{'RESERVED_40_42'}            = hex join '', splice @$s_data, 0, 3;
    $xlate_h{'IMPLEMENTED_DESCRIPTOR'}    = hex shift @$s_data;

    my @descriptors;
    for (my $i = 0 ; $i < $xlate_h{'IMPLEMENTED_DESCRIPTOR'} ; $i++) {
        push(@descriptors, hex shift @$s_data);
    }

    $xlate_h{'DESCRIPTORS'} = \@descriptors;

    return \%xlate_h;
}

###############################################################################
# Sub Name:  xlate_protocol_specific_port_mode_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
#$VAR1 = \{
#             'PS'                                  => 0,
#             'SPF'                                 => 0,
#             'PAGE_CODE'                           => 19h,
#             'PAGE_LENGTH'                         => based on protocol,
#             'RESV_B2_bits4_7'                     => 0
#             'PROTOCOL_IDENTIFIER                  => 00h/05h
#             'PROTOCOL_SPECIFIC_MODE_PARAMETERS'   => var
#         };
###############################################################################
sub xlate_protocol_specific_port_mode_page {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_len = scalar(@$s_data);

    # make sure we have at least 2 bytes so we can read the page length
    # since each page will do this, I could do this check before the method
    # call
    unless ($data_len >= 2) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }

    my $next_byte = hex shift @$s_data;
    $data_len--;
    $xlate_h{PS}  = ($next_byte & BIT_7) >> 7;
    $xlate_h{SPF} = ($next_byte & BIT_6) >> 6;
    $xlate_h{PAGE_CODE} =
      ($next_byte & (BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0));
    unless ($data_len) {
        $self->_set_err(
                  "Data length too short for Protocol Specific Port Mode Page");
        return;
    }

    $xlate_h{PAGE_LENGTH} = hex shift @$s_data;

    # make sure we have enough bytes to get the complete page
    unless ($data_len >= $xlate_h{PAGE_LENGTH}) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }

    my $page_len = $xlate_h{PAGE_LENGTH};
    if ($page_len < 1) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }
    $next_byte = hex shift @$s_data;
    $page_len--;
    $xlate_h{RESV_B2_bits4_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $xlate_h{PROTOCOL_IDENTIFIER} =
      ($next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0));

    # The rest of this page is not returned ONLY in the case of C-mode iSCSI.
    # So just grab the rest of the page(PAGE_LENGTH - 1) and let the app sort
    # it out based on C-mode/7-mode/iSCSI/FCP
    if ($page_len) {    # Fibre Channel
        $xlate_h{PROTOCOL_SPECIFIC_MODE_PARAMETERS} = hex join '',
          splice @$s_data, 0, $page_len;
    }

    return \%xlate_h;
}    # end sub xlate_protocol_specific_port_mode_page

###############################################################################
# Sub Name:  xlate_informational_exceptions_control_mode_page()
#            From SPC-3 r23
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
#$VAR1 = \{
#             'PS'                      => 0,
#             'SPF'                     => 0,
#             'PAGE_CODE'               => 1Ch,
#             'PAGE_LENGTH'             => 0Ah,
#             'PERF'                    => 0
#             'RESERVED_B2_bits6        => 0
#             'EBF'                     => 0
#             'EWASC'                   => 0
#             'DEXCPT'                  => 0
#             'TEST'                    => 0
#             'RESERVED_B2_bits1'       => 0
#             'LOGERR'                  => 0
#             'RESV_B3'                 => 0
#             'MRIE'                    => 0
#             'INTERVAL_TIMER'          => 0
#             'REPORT_COUNT'            => 0
#         };
###############################################################################
sub xlate_informational_exceptions_control_mode_page {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_len = scalar(@$s_data);

    # make sure we have at least 2 bytes so we can read the page length
    # since each page will do this, I could do this check before the method
    # call
    unless ($data_len >= 2) {
        $self->_set_err(
                "Data length too short for Informational Exceptions Mode Page");
        return;
    }

    my $next_byte = hex shift @$s_data;
    $data_len--;
    $xlate_h{PS}  = ($next_byte & BIT_7) >> 7;
    $xlate_h{SPF} = ($next_byte & BIT_6) >> 6;
    $xlate_h{PAGE_CODE} =
      ($next_byte & (BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0));

    $xlate_h{PAGE_LENGTH} = hex shift @$s_data;
    $data_len--;

    # make sure we have enough bytes to get the complete page
    unless ($data_len >= $xlate_h{PAGE_LENGTH}) {
        $self->_set_err(
                "Data length too short for Informational Exceptions Mode Page");
        return;
    }

    my $page_len = $xlate_h{PAGE_LENGTH};
    if ($page_len < 1) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }
    $next_byte = hex shift @$s_data;
    $page_len--;
    $xlate_h{PERF}              = ($next_byte & BIT_7) >> 7;
    $xlate_h{RESERVED_B2_bits6} = ($next_byte & BIT_6) >> 6;
    $xlate_h{EBF}               = ($next_byte & BIT_5) >> 5;
    $xlate_h{EWASC}             = ($next_byte & BIT_4) >> 4;
    $xlate_h{DEXCPT}            = ($next_byte & BIT_3) >> 3;
    $xlate_h{TEST}              = ($next_byte & BIT_2) >> 2;
    $xlate_h{RESERVED_B2_bits1} = ($next_byte & BIT_1) >> 1;
    $xlate_h{LOGERR}            = ($next_byte & BIT_0);

    if ($page_len < 1) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }
    $next_byte = hex shift @$s_data;
    $page_len--;
    $xlate_h{RESV_B3} = ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4)) >> 4;
    $xlate_h{MRIE}    = ($next_byte & (BIT_3 | BIT_2 | BIT_1 | BIT_0));

    if ($page_len < 4) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }
    $xlate_h{INTERVAL_TIMER} = hex join '', splice @$s_data, 0, 4;
    $page_len -= 4;

    if ($page_len < 4) {
        $self->_set_err(
                       "Data length too short for Protocol Specific Mode Page");
        return;
    }
    $xlate_h{REPORT_COUNT} = hex join '', splice @$s_data, 0, 4;
    $page_len -= 4;

    unless ($page_len == 0) {
        $self->_set_err(  "Mismatch in the returned page length and the number "
                        . "of bytes translated.");
    }

    return \%xlate_h;
}    # end sub xlate_informational_exceptions_control_mode_page

###############################################################################
# Sub Name:  xlate_power_condition_mode_page()
#
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
#$VAR1 = \{
#             'PS'                      => 0,
#             'RESV_B0_bits6'           => 0,
#             'PAGE_CODE'               => 1Ah,
#             'PAGE_LENGTH'             => 0Ah,
#             'RESV_B2'                 => 0
#             'RESV_B3_bits2_7'         => 0
#             'IDLE'                    => 0
#             'STANDBY'                 => 0
#             'IDLE_CONDITION_TIMER'    => 0
#             'STANDBY_CONDITION_TIMER' => 0
#         };
###############################################################################
sub xlate_power_condition_mode_page {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    tie my %xlate_h, 'Tie::IxHash';

    my $data_len = scalar(@$s_data);

    # make sure we have at least 2 bytes so we can read the page length
    # since each page will do this, I could do this check before the method
    # call
    unless ($data_len >= 2) {
        $self->_set_err("Data length too short for Power Condition Mode Page");
        return;
    }

    my $next_byte = hex shift @$s_data;
    $xlate_h{PS}            = ($next_byte & BIT_7) >> 7;
    $xlate_h{RESV_B0_bits6} = ($next_byte & BIT_6) >> 6;
    $xlate_h{PAGE_CODE} =
      ($next_byte & (BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0));
    $data_len--;
    $xlate_h{PAGE_LENGTH} = hex shift @$s_data;
    $data_len--;

    # make sure we have enough bytes to get the complete page
    unless ($data_len >= $xlate_h{PAGE_LENGTH}) {
        $self->_set_err("Data length too short for Power Condition Mode Page");
        return;
    }

    # We used data_len to ensure we had enough to get to the page_len. Now
    # we'll use it to make sure we read the correct number of bytes.
    $data_len = $xlate_h{PAGE_LENGTH};
    $xlate_h{RESV_B2} = hex shift @$s_data;
    $data_len--;

    $next_byte = hex shift @$s_data;
    $data_len--;

    $xlate_h{RESV_B3_bits2_7} =
      ($next_byte & (BIT_7 | BIT_6 | BIT_5 | BIT_4 | BIT_3 | BIT_2)) >> 2;
    $xlate_h{IDLE}    = ($next_byte & BIT_1) >> 1;
    $xlate_h{STANDBY} = ($next_byte & BIT_0);

    $xlate_h{IDLE_CONDITION_TIMER} = hex join '', splice @$s_data, 0, 4;
    $data_len -= 4;

    $xlate_h{STANDBY_CONDITION_TIMER} = hex join '', splice @$s_data, 0, 4;
    $data_len -= 4;

    unless ($data_len == 0) {
        $self->_set_err(  "Mismatch in the returned page length and the number "
                        . "of bytes translated.");
    }

    return \%xlate_h;
}    # end sub xlate_power_condition_mode_page

########################################################################
# Sub Name:  xlate_scsi_pr_in_read_keys()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
# $VAR1 = \{
#             'ADD_LENGTH' => 8,
#             'GENERATION' => 926,
#             'KEYS' => [
#                         'aa:bb:cc:dd:00:11:22:33'
#                       ]
#           };
########################################################################
sub xlate_scsi_pr_in_read_keys {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    my (@pr_keys, $xlate_str);

    # Translate the page information GENERATION
    my $generation = hex join '', splice @$s_data, 0, 4;

    # Byte 1 => Additional length
    my $additional_length = hex join '', splice @$s_data, 0, 4;

    my $number_keys = $additional_length / 8;

    for (my $index = 0 ; $index < $number_keys ; $index++) {
        my $key = join ':', splice @$s_data, 0, 8;
        push(@pr_keys, $key);
    }

    $xlate_str->{GENERATION} = $generation;
    $xlate_str->{ADD_LENGTH} = $additional_length;
    $xlate_str->{KEYS}       = \@pr_keys;

    return $xlate_str;
} ## end sub xlate_scsi_pr_in_read_keys

########################################################################
# Sub Name:  xlate_scsi_pr_in_read_resv()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
#$VAR1 = \{
#             'RESERVED' => 0,
#             'ADD_LENGTH' => 16,
#             'GENERATION' => 928,
#             'TYPE' => 3,
#             'ADDRESS' => 0,
#             'KEYS' => 'aa:bb:cc:dd:00:11:22:33'
#         };
########################################################################

sub xlate_scsi_pr_in_read_resv {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    my ($xlate_str);

    # Translate the page information GENERATION
    my $generation = hex join '', splice @$s_data, 0, 4;

    # Byte 1 => Additional length
    my $additional_length = hex join '', splice @$s_data, 0, 4;

    my ($key, $scope_spec_addr, $reserved, $type);

    if ($additional_length == 16) {
        $key = join ':', splice @$s_data, 0, 8;
        $scope_spec_addr = hex join '', splice @$s_data, 0, 4;
        $reserved        = hex join '', splice @$s_data, 0, 1;
        $type            = hex join '', splice @$s_data, 0, 1;

        $xlate_str->{KEYS}     = $key;
        $xlate_str->{ADDRESS}  = $scope_spec_addr;
        $xlate_str->{RESERVED} = $reserved;
        $xlate_str->{TYPE}     = $type;
    } ## end if ($additional_length...

    $xlate_str->{GENERATION} = $generation;
    $xlate_str->{ADD_LENGTH} = $additional_length;

    return $xlate_str;
} ## end sub xlate_scsi_pr_in_read_resv

########################################################################
# Sub Name:  xlate_scsi_pr_in_rep_cap()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
#$VAR1 = \{
#             'BYTE2' => 1,
#             'BYTE3' => 128,
#             'LENGTH' => 8,
#             'MASK' => 59905,
#             'RES' => 0
#         };

########################################################################

sub xlate_scsi_pr_in_rep_cap {
    my ($self, $s_data) = @_;

    $self->_clear_err();

    my ($xlate_str);

    # Translate the page information LENGTH
    $xlate_str->{LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 2
    my $fields = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{PTPL_C} = $fields & BIT_0;
    $xlate_str->{ATP_C}  = ($fields & BIT_2) >> 2;
    $xlate_str->{SIP_C}  = ($fields & BIT_3) >> 3;
    $xlate_str->{CRH}    = ($fields & BIT_4) >> 4;

    # Byte 3
    my $tmv_ptpl = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{PTPL_A} = $tmv_ptpl & BIT_0;
    $xlate_str->{TMV}    = ($tmv_ptpl & BIT_7) >> 7;

    # PR mask
    $xlate_str->{MASK} = hex join '', splice @$s_data, 0, 2;

    # reserved
    $xlate_str->{RES} = hex join '', splice @$s_data, 0, 2;

    return $xlate_str;
} ## end sub xlate_scsi_pr_in_rep_cap

########################################################################
# Sub Name:  parse_inquiry_request()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
########################################################################

sub parse_inquiry_request {
    my ($self, $s_cmd, $s_data) = @_;
    my $ret_val;

    if ($s_cmd =~ /^inquiry$/i) {

        # call the unpacking routine for Standard Inquiry data
        $ret_val = $self->xlate_std_inq_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_00h$|inquiry_page_00h/i) {

        # call the unpacking routine for Supported VPD Pages data
        $ret_val = $self->xlate_inq_pg_00h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_80h$|inquiry_page_80h/i) {

        # call the unpacking routine for Unit Serial Number data
        $ret_val = $self->xlate_inq_pg_80h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_83h$|inquiry_page_83h/i) {

        # call the unpacking routine for Get Device ID data
        $ret_val = $self->xlate_inq_pg_83h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_85h$|inquiry_page_85h/i) {

        # call the unpacking routine for Management Network Addresses data
        $ret_val = $self->xlate_inq_pg_85h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_87h$|inquiry_page_87h/i) {

        # call the unpacking routine for Mode Page Policy data
        $ret_val = $self->xlate_inq_pg_87h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_B0h$|inquiry_page_B0h/i) {

        # call the unpacking routine for Block Limits data
        $ret_val = $self->xlate_inq_pg_B0h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_B2h$|inquiry_page_B2h/i) {

        # call the unpacking routine for Logical Block Provisioning data
        $ret_val = $self->xlate_inq_pg_B2h_bytes($s_data);
    } elsif ($s_cmd =~ /^inq_pg_C0h$|inquiry_page_C0h/i) {

        # call the unpacking routine for Vendor Specific-Get Filer IP data
        $ret_val = $self->xlate_inq_pg_C0h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_C1h$|inquiry_page_C1h/i) {

        # call the unpacking routine for Vendor Specific-Proxy Info data
        $ret_val = $self->xlate_inq_pg_C1h_bytes($s_data);

    } elsif ($s_cmd =~ /^inq_pg_C2h$|inquiry_page_C2h/i) {

        # call the unpacking routine for Vendor Specific-Target Port Info data
        $ret_val = $self->xlate_inq_pg_C2h_bytes($s_data);

    }
}    #end sub parse_inquiry_request

########################################################################
# Sub Name:  xlate_gva_vdisk_vol_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 20,
#            'VOL_ID' => 53644959,
#            'VOL_NAME' => 'stib_vol1',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 17,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 28,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_vol_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Length
    $xlate_str->{VOL_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{VOL_NAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 8));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_vol_page

########################################################################
# Sub Name:  xlate_gva_vdisk_vol_page_rev1()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data. C mode only
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 20,
#            'VOL_ID' => 53644959,
#            'VOL_NAME' => 'stib_vol1',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 17,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 28,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_vol_page_rev1 {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Length
    # $xlate_str->{VOL_ID} = hex join '', splice @$s_data, 0, 16;
    my @vol_id = splice @$s_data, 0, 16;

    # @vol_id = reverse(@vol_id);

    $xlate_str->{VOL_ID} = hex join '', @vol_id;

    #bytes 8..N
    $xlate_str->{VOL_NAME} =
      pack("H*", join '', splice @$s_data, 0, scalar(@$s_data));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_vol_page_rev1

########################################################################
# Sub Name:  xlate_gva_vdisk_path_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 44,
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 16,
#            'VDISK_PATH' => '/vol/stib_vol1/rtpqa-sun280r-65_lun1',
#            'REVISION' => 0,
#            'DATA_LENGTH' => 52,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_path_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{VDISK_PATH} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;

} ## end sub xlate_gva_vdisk_path_page

########################################################################
# Sub Name:  xlate_gva_vdisk_lun_ostype_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 12,
#            'OSTYPE' => 'solaris',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 24,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 20,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_lun_ostype_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{OSTYPE} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_lun_ostype_page

########################################################################
# Sub Name:  xlate_gva_vdisk_system_info_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'SYS_SOF_REV' => ' mainngN_090604_1820',
#            'PROTOCOL' => 125,
#            'PAGE_CODE' => 34,
#            'REVISION' => 0,
#            'SYS_SERIAL_NUM' => '08061876',
#            'DATA_LENGTH' => 84,
#            'PAGE_LENGTH' => 76,
#            'PAGE_REVISION' => 0,
#            'MODEL_NAME' => 'FAS3070',
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_system_info_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..27 System Software Revision String
    $xlate_str->{SYS_SOF_REV} = pack("H*", join '', splice @$s_data, 0, 24);

    #bytes 28..51 System Serial Number
    $xlate_str->{SYS_SERIAL_NUM} = pack("H*", join '', splice @$s_data, 0, 24);

    #bytes 52..55 Reserved bytes
    splice @$s_data, 0, 4;

    #bytes 56..63
    $xlate_str->{MODEL_NAME} = pack("H*", join '', splice @$s_data, 0, 8);

    #bytes 64..70 Reserved
    splice @$s_data, 0, 7;

    # Byte 71 => Reserved NFS ISCSI HTTP HA FCP DAFS CIFS

    my $protocol_license = hex splice @$s_data, 0, 1;

    $xlate_str->{CIFS}     = ($protocol_license & BIT_0);
    $xlate_str->{DAFS}     = ($protocol_license & BIT_1) >> 1;
    $xlate_str->{FCP}      = ($protocol_license & BIT_2) >> 2;
    $xlate_str->{HA}       = ($protocol_license & BIT_3) >> 3;
    $xlate_str->{HTTP}     = ($protocol_license & BIT_4) >> 4;
    $xlate_str->{ISCSI}    = ($protocol_license & BIT_5) >> 5;
    $xlate_str->{NFS}      = ($protocol_license & BIT_6) >> 6;
    $xlate_str->{Reserved} = ($protocol_license & BIT_7) >> 7;

    return $xlate_str;

} ## end sub xlate_gva_vdisk_system_info_page

########################################################################
# Sub Name:  xlate_gva_vdisk_mapped_luns_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_CODE' => 22,
#            'REVISION' => 0,
#            'MLUN_COUNT' => 5,
#            'DATA_LENGTH' => 40,
#            'PAGE_LENGTH' => 32,
#            'IGROUP_NAME' => 'rtpqa-sun280r-65_fcp',
#            'PAGE_REVISION' => 0,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_mapped_luns_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Mapped Lun count
    $xlate_str->{MLUN_COUNT} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{IGROUP_NAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 8));

    return $xlate_str;

} ## end sub xlate_gva_vdisk_mapped_luns_page

########################################################################
# Sub Name:  xlate_gva_vdisk_lun_share_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#               'PAGE_LENGTH' => 8,
#               'PAGE_REVISION' => 0,
#               'PAGE_CODE' => 23,
#               'REVISION' => 0,
#               'SHARED' => 1,
#               'DATA_LENGTH' => 16,
#               'MAGIC_NUMBER' => 694282
#             };
########################################################################

sub xlate_gva_vdisk_lun_share_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..6 # Reserved
    splice @$s_data, 0, 3;

    #bytes 7  # Shared Read Write
    my $shared = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{WRITE} = ($shared & BIT_1) >> 1;
    $xlate_str->{READ}  = ($shared & BIT_0);

    return $xlate_str;
} ## end sub xlate_gva_vdisk_lun_share_page

########################################################################
# Sub Name:  xlate_gva_filername_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 24,
#            'FILER_HOSTNAME' => 'rtpqa-fas3070-11',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 0,
#            'REVISION' => 2,
#            'DATA_LENGTH' => 32,
#            'MAGIC_NUMBER' => 694282
#         };

########################################################################

sub xlate_gva_filername_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Extract the page information

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 8 .. N => ADAPTER DESC
    $xlate_str->{FILER_HOSTNAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;
} ## end sub xlate_gva_filername_page

########################################################################
# Sub Name:  xlate_gva_32_bit_ip_addr()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 28,
#            'IP_ADDR' => [
#                           '10.60.169.56',
#                           '10.225.194.25',
#                           '10.225.194.26',
#                           '10.225.194.27',
#                           '10.225.194.28'
#                         ],
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 1,
#            'COUNT' => 5,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 36,
#            'MAGIC_NUMBER' => 694282
#          };

########################################################################

sub xlate_gva_32_bit_ip_addr {
    my ($self, $s_data) = @_;

    my (@tmp_arr2, @ip_addr);

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Extract the page information
    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Bytes 4-7 => Count
    $xlate_str->{COUNT} = hex join '', splice @$s_data, 0, 4;

    #my $total_bytes = $count * 4;

    for (my $index = 0 ; $index < $xlate_str->{COUNT} ; $index++) {
        @tmp_arr2 = splice @$s_data, 0, 4;
        my $ip_addr =
            hex($tmp_arr2[0]) . "."
          . hex($tmp_arr2[1]) . "."
          . hex($tmp_arr2[2]) . "."
          . hex($tmp_arr2[3]);
        $ip_addr[$index] = $ip_addr;
        $self->sanlog(-msg => "IP_addr Array values" . $ip_addr[$index] . "\n",
                      -msg_level => 1);
    } ## end for (my $index = 0 ; $index...

    my $tmp_arr_len = scalar(@ip_addr);

    $xlate_str->{IP_ADDR} = \@ip_addr;

    return $xlate_str;

} ## end sub xlate_gva_32_bit_ip_addr

########################################################################
# Sub Name:  xlate_gva_64_bit_ip_addr()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: In Data ONTAP 7-Mode, this page just returns the same data as
#          Page Code 01h i.e. 32-bit IP Address page.
#          In Data ONTAP 8.1 C-Mode, this page is OBSOLETE.
# The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 28,
#            'IP_ADDR' => [],
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 2,
#            'COUNT' => 5,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 36,
#            'MAGIC_NUMBER' => 694282
#          };

########################################################################

sub xlate_gva_64_bit_ip_addr {
    my ($self, $s_data) = @_;

    my (@tmp_arr2, @ip_addr);

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Extract the page information
    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Bytes 4-7 => Count
    $xlate_str->{COUNT} = hex join '', splice @$s_data, 0, 4;

    #my $total_bytes = $count * 4;

    for (my $index = 0 ; $index < $xlate_str->{COUNT} ; $index++) {
        @tmp_arr2 = splice @$s_data, 0, 4;
        my $ip_addr =
            hex($tmp_arr2[0]) . "."
          . hex($tmp_arr2[1]) . "."
          . hex($tmp_arr2[2]) . "."
          . hex($tmp_arr2[3]);
        $ip_addr[$index] = $ip_addr;
        $self->sanlog(-msg => "IP_addr Array values" . $ip_addr[$index] . "\n",
                      -msg_level => 1);
    } ## end for (my $index = 0 ; $index...

    my $tmp_arr_len = scalar(@ip_addr);

    $xlate_str->{IP_ADDR} = \@ip_addr;

    return $xlate_str;

} ## end sub xlate_gva_64_bit_ip_addr

########################################################################
# Sub Name:  xlate_gva_vdisk_logged_init_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 8,
#            'INIT_COUNT' => 5,
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 37,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 16,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_logged_init_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4..7 => Initiator Count
    $xlate_str->{INIT_COUNT} = hex join '', splice @$s_data, 0, 4;

    return $xlate_str;
} ## end sub xlate_gva_vdisk_logged_init_page

########################################################################
# Sub Name:  xlate_gva_vdisk_fcp_config_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_CODE' => 32,
#            'REVISION' => 0,
#            'FCP_PORT_NAME' => '500a098397e92b73',
#            'DATA_LENGTH' => 32,
#            'PAGE_LENGTH' => 24,
#            'FCP_CFMODE' => 7,
#            'FCP_NODENAME' => '500a098087e92b73',
#            'PAGE_REVISION' => 1,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_fcp_config_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 .. 6 => Reserved
    splice @$s_data, 0, 3;

    # Byte 7 => FCP CFMODE
    $xlate_str->{FCP_CFMODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 8-15 => FCP NODE NAME
    $xlate_str->{FCP_NODENAME} = join '', splice @$s_data, 0, 8;

    # Bytes 16..23 => FCP PORT NAME
    $xlate_str->{FCP_PORT_NAME} = join '', splice @$s_data, 0, 8;

    return $xlate_str;
} ## end sub xlate_gva_vdisk_fcp_config_page

########################################################################
# Sub Name:  xlate_gva_vdisk_fcp_adapter_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 8,
#            'CLI_PORT_ID' => '1a',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 33,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 16,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_fcp_adapter_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 .. N => FCP ADAPTER PORT CLI IDENTIFIER
    $xlate_str->{CLI_PORT_ID} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_fcp_adapter_page

########################################################################
# Sub Name:  xlate_gva_vdisk_fcp_adap_det_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'ADAPTER_DESC' => 'Dual-channel, QLogic 2432 (2464)',
#            'QUEUE_DEPTH' => 2048,
#            'PAGE_CODE' => 36,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 52,
#            'PAGE_LENGTH' => 44,
#            'PAGE_REVISION' => 0,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_fcp_adap_det_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => QUEUE DEPTH
    $xlate_str->{QUEUE_DEPTH} = hex join '', splice @$s_data, 0, 4;

    # Byte 8 .. N => ADAPTER DESC
    $xlate_str->{ADAPTER_DESC} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 8));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_fcp_adap_det_page

########################################################################
# Sub Name:  xlate_gva_vdisk_igroup_info_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'RESERVED' => 0,
#            'THROTTLE_PER' => 4,
#            'THROTTLE_BORROW' => 1,
#            'PROXY_FLAG' => 1,
#            'OS_TYPE' => 1,
#            'PAGE_REVISION' => 0,
#            'VSA_FLAG' => 0,
#            'PROXY_WINDOWS' => 1,
#            'PROXY_LINUX' => 1,
#            'PAGE_CODE' => 38,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 32,
#            'PAGE_LENGTH' => 24,
#            'UNRESERVED_RES' => 1888,
#            'PROXY_HP' => 1,
#            'PROXY_SOLARIS' => 1,
#            'MAGIC_NUMBER' => 694282,
#            'PROXY_AIX' => 1
#          };
########################################################################

sub xlate_gva_vdisk_igroup_info_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => OS Type
    $xlate_str->{OS_TYPE} = hex join '', splice @$s_data, 0, 1;

    # Byte 5  => VSA flag
    $xlate_str->{VSA_FLAG} = hex join '', splice @$s_data, 0, 1;

    # Bytes 6  => USE PROXY FLAG
    $xlate_str->{PROXY_FLAG} = hex join '', splice @$s_data, 0, 1;

    # Byte 7  => USE PROXY - SOLARIS
    $xlate_str->{PROXY_SOLARIS} = hex join '', splice @$s_data, 0, 1;

    # Bytes 8 => USE PROXY - WINDOWS
    $xlate_str->{PROXY_WINDOWS} = hex join '', splice @$s_data, 0, 1;

    # Bytes 9 => USE PROXY - LINUX
    $xlate_str->{PROXY_LINUX} = hex join '', splice @$s_data, 0, 1;

    # Bytes 10 => USE PROXY - HP-UX
    $xlate_str->{PROXY_HP} = hex join '', splice @$s_data, 0, 1;

    # Bytes 11 => USE PROXY - HP-AIX
    $xlate_str->{PROXY_AIX} = hex join '', splice @$s_data, 0, 1;

    # Bytes 12 ..14  => RESERVED
    $xlate_str->{RESERVED} = hex join '', splice @$s_data, 0, 3;

    # Byte 15 => THROTTLE BORROW
    $xlate_str->{THROTTLE_BORROW} = hex join '', splice @$s_data, 0, 1;

    # Bytes 16 .. 19: THROTTLE RESERVE PERCENT
    $xlate_str->{THROTTLE_PER} = hex join '', splice @$s_data, 0, 4;

    # Bytes 20..23: UNRESERVED RESOURCES
    $xlate_str->{UNRESERVED_RES} = hex join '', splice @$s_data, 0, 4;

    return $xlate_str;
} ## end sub xlate_gva_vdisk_igroup_info_page

########################################################################
# Sub Name:  xlate_gva_iscsi_config_info()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 64,
#            'TCP_PORT' => 3260,
#            'IP_ADDR' => '10.225.194.28',
#            'NODE_NAME' => 'iqn.1992-08.com.netapp:sn.118041459',
#            'PAGE_REVISION' => 1,
#            'PAGE_CODE' => 64,
#            'REVISION' => 1,
#            'TPG_TAG' => 1009,
#            'DATA_LENGTH' => 72,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_iscsi_config_info {
    my ($self, $s_data) = @_;

    my $ip_address;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    #translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    #byte 1 => page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    #byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4-7 => TPG TAG
    $xlate_str->{TPG_TAG} = hex join '', splice @$s_data, 0, 4;

    #bytes 8-9 => TCP PORT
    $xlate_str->{TCP_PORT} = hex join '', splice @$s_data, 0, 2;

    if ($xlate_str->{PAGE_REVISION} == 0) {

        #bytes 10-13 => IP address
        my @tmp_arr = splice @$s_data, 0, 4;
        $xlate_str->{IP_ADDR} =
            hex($tmp_arr[0]) . "."
          . hex($tmp_arr[1]) . "."
          . hex($tmp_arr[2]) . "."
          . hex($tmp_arr[3]);

        #bytes 14..N
        $xlate_str->{NODE_NAME} =
          pack("H*", join '', splice @$s_data, 0, scalar(@$s_data));

    } elsif ($xlate_str->{PAGE_REVISION} == 1) {

        #bytes 10 => Address type
        my $address_type = hex join '', splice @$s_data, 0, 1;

        #bytes 11-26 => IP address
        my @tmp_arr = splice @$s_data, 0, 16;

        if ($address_type == 0) {

            #first 4 bytes are populated
            $xlate_str->{IP_ADDR} =
                hex($tmp_arr[0]) . "."
              . hex($tmp_arr[1]) . "."
              . hex($tmp_arr[2]) . "."
              . hex($tmp_arr[3]);
        } else {

            #16 bytes are populated
            # : separated hexadecimal numbers => ipv6 representation
            $xlate_str->{IP_ADDR} =
                $tmp_arr[0]
              . ($tmp_arr[1]) . ":"
              . $tmp_arr[2]
              . ($tmp_arr[3]) . ":"
              . $tmp_arr[4]
              . ($tmp_arr[5]) . ":"
              . $tmp_arr[6]
              . ($tmp_arr[7]) . ":"
              . $tmp_arr[8]
              . ($tmp_arr[9]) . ":"
              . $tmp_arr[10]
              . ($tmp_arr[11]) . ":"
              . $tmp_arr[12]
              . ($tmp_arr[13]) . ":"
              . $tmp_arr[14]
              . ($tmp_arr[15]);
        } ## end else [ if ($address_type == 0)

        #bytes 27..N
        $xlate_str->{NODE_NAME} =
          pack("H*", join '', splice @$s_data, 0, scalar(@$s_data));
    } ## end elsif ($xlate_str->{PAGE_REVISION...

    return $xlate_str;
} ## end sub xlate_gva_iscsi_config_info

########################################################################
# Sub Name:  xlate_gva_iscsi_adapter_info()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_LENGTH' => 12,
#            'ADAPTER_NAME' => 'ism_sn1',
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 65,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 20,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_iscsi_adapter_info {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    #translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    #byte 1 => page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    #byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    $xlate_str->{ADAPTER_NAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;

} ## end sub xlate_gva_iscsi_adapter_info

########################################################################
# Sub Name:  xlate_gva_vdisk_cluster_info_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PARTNER_NAME' => 'rtpqa-fas3070-12',
#            'PAGE_CODE' => 39,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 40,
#            'PAGE_LENGTH' => 32,
#            'CLUSTER_ENABLED' => 1,
#            'CLUSTER_FLG' => 0,
#            'PAGE_REVISION' => 0,
#            'IC_CONNECTED' => 1,
#            'MAGIC_NUMBER' => 694282
#          };
########################################################################

sub xlate_gva_vdisk_cluster_info_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => Cluster Enabled
    $xlate_str->{CLUSTER_ENABLED} = hex join '', splice @$s_data, 0, 1;

    # Byte 5 => IC connected
    $xlate_str->{IC_CONNECTED} = hex join '', splice @$s_data, 0, 1;

    # Byte 6..10 => reserved
    splice @$s_data, 0, 5;

    # Bytes 11 => CLUSTER FLAGS WAIT, ERR, GBST, INTO, TOST, TOSCH, CCON
    my $cluster_flag = hex splice @$s_data, 0, 1;

    $xlate_str->{CCON}     = ($cluster_flag & BIT_0);
    $xlate_str->{TOSCH}    = ($cluster_flag & BIT_1) >> 1;
    $xlate_str->{TOST}     = ($cluster_flag & BIT_2) >> 2;
    $xlate_str->{INTO}     = ($cluster_flag & BIT_3) >> 3;
    $xlate_str->{GBST}     = ($cluster_flag & BIT_4) >> 4;
    $xlate_str->{ERR}      = ($cluster_flag & BIT_5) >> 5;
    $xlate_str->{WAIT}     = ($cluster_flag & BIT_6) >> 6;
    $xlate_str->{RESERVED} = ($cluster_flag & BIT_7) >> 7;

    # Bytes 12..N
    $xlate_str->{PARTNER_NAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 7));

    return $xlate_str;
} ## end sub xlate_gva_vdisk_cluster_info_page

########################################################################
# Sub Name:  xlate_gva_vdisk_lun_select_info_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_CODE' => 50,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 60,
#            'PAGE_LENGTH' => 52,
#            'PREV_LUN_SER_NUM' => 'HnS7XJU9PBo2',
#            'LUN_PAGE_REVISION' => 0,
#            'PAGE_REVISION' => 0,
#            'CREATION_TIME' => ?,
#            'MAGIC_NUMBER' => 694282,
#            'LUN_SELECT' => 1
#          };
########################################################################

sub xlate_gva_vdisk_lun_select_info_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => Lun select page revision
    $xlate_str->{LUN_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 5 => reserved
    splice @$s_data, 0, 1;

    # Byte 6 => Lun select value
    $xlate_str->{LUN_SELECT} = hex join '', splice @$s_data, 0, 1;

    # Byte 7..10 => reserved
    splice @$s_data, 0, 4;

    # Bytes 11..26 => PREVIOUS LUN SERIAL NUM
    $xlate_str->{PREV_LUN_SER_NUM} =
      pack("H*", join '', splice @$s_data, 0, 12);

    # Bytes 27..30 => RESERVED
    splice @$s_data, 0, 8;

    # Bytes 31 ..46 => LUN CREATION TIME
    my @creation_time = splice @$s_data, 0, 4;

    my @ct = reverse(@creation_time);

    $xlate_str->{CREATION_TIME} = hex join '', @ct;

    return $xlate_str;
} ## end sub xlate_gva_vdisk_lun_select_info_page

########################################################################
# Sub Name:  xlate_gva_vdisk_status_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#               'OTHER' => 0,
#               'STATUS' => 0,
#               'PAGE_CODE' => 20,
#               'REVISION' => 0,
#               'DATA_LENGTH' => 16,
#               'PAGE_LENGTH' => 8,
#               'PAGE_REVISION' => 0,
#               'MAGIC_NUMBER' => 694282
#           };

########################################################################

sub xlate_gva_vdisk_status_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4
    $xlate_str->{OTHER} = hex join '', splice @$s_data, 0, 1;

    #bytes 5..6
    my $reserved = hex join '', splice @$s_data, 0, 2;

    #bytes 7 # status bits
    my $status = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{ROFS}    = ($status & BIT_4) >> 4;
    $xlate_str->{NOSPC}   = ($status & BIT_3) >> 3;
    $xlate_str->{NOENT}   = ($status & BIT_2) >> 2;
    $xlate_str->{NVFAIL}  = ($status & BIT_1) >> 1;
    $xlate_str->{OFFLINE} = ($status & BIT_0);

    return $xlate_str;
} ## end sub xlate_gva_vdisk_status_page

########################################################################
# Sub Name:  xlate_gva_vdisk_qtree_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'PAGE_CODE' => 18,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 32,
#            'PAGE_LENGTH' => 24,
#            'PAGE_REVISION' => 0,
#            'QTREE_NAME' => '/vol/stib_vol1',
#            'QTREE_ID' => 0,
#            'MAGIC_NUMBER' => 694282
#         };

########################################################################

sub xlate_gva_vdisk_qtree_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Qtree id
    $xlate_str->{QTREE_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{QTREE_NAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 8));

    return $xlate_str;

} ## end sub xlate_gva_vdisk_qtree_page

########################################################################
# Sub Name:  xlate_gva_vdisk_snapshot_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#            'SNAPSHOT_FILE_PATH' => '^@^@^@^@',
#            'PAGE_LENGTH' => 8,
#            'PAGE_REVISION' => 0,
#            'PAGE_CODE' => 19,
#            'REVISION' => 0,
#            'DATA_LENGTH' => 16,
#            'MAGIC_NUMBER' => 694282
#          };
#########################################################################

sub xlate_gva_vdisk_snapshot_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{SNAPSHOT_FILE_PATH} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    return $xlate_str;

} ## end sub xlate_gva_vdisk_snapshot_page

########################################################################
# Sub Name:  xlate_gva_vdisk_stream_sizes_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
########################################################################

sub xlate_gva_vdisk_stream_sizes_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    #DATA LENGTH is the total length of the data that can be returned
    # including the header and requested page(s).
    # Note that the actual length of the data returned could be
    # restricted by the ALLOCATION LENGTH in the CDB.
    # In this case the allocation length in the CDB is set to 1C

    unless (($xlate_str->{DATA_LENGTH}) == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4-11
    $xlate_str->{MAIN_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 12..19
    $xlate_str->{PREFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 20..27
    $xlate_str->{SUFFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    return $xlate_str;

} ## end sub xlate_gva_vdisk_stream_sizes_page

########################################################################
# Sub Name:  xlate_gva_vdisk_asl_info_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
# $VAR1 = {
#           'ASL_VERSION' => 2,
#           'STATUS' => 0,
#           'PATH_INFO' => '^@^@^@',
#           'PAGE_CODE' => 35,
#           'ISCSI_AP' => 16,
#           'REVISION' => 0,
#           'TPG_TAG' => 1342835072,
#           'DATA_LENGTH' => 76,
#           'ADDRESS' => '80.0.0.0',
#           'PAGE_LENGTH' => 68,
#           'ADDRESS_TYPE' => 43,
#           'TCP_PORT' => 34793,
#           'ENC_SERIAL_NUM' => '09060017^@^@^@^@^@^@^@^@^@^@^@^@^@^@^@^@',
#           'PAGE_REVISION' => 3,
#           'ENC_NAME' => 'FAS3070^@',
#           'MAGIC_NUMBER' => 694282
#         };
#
########################################################################

sub xlate_gva_vdisk_asl_info_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Translate the page information 0
    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => ASL version
    $xlate_str->{ASL_VERSION} = hex join '', splice @$s_data, 0, 1;

    # Byte 5  => reserved
    splice @$s_data, 0, 1;

    # Bytes 6  => ISCSI_AP
    my $flag = hex splice @$s_data, 0, 1;

    $xlate_str->{AP}    = ($flag & BIT_4) >> 4;
    $xlate_str->{ISCSI} = ($flag & BIT_5) >> 5;

    # Byte 7  => reserved
    splice @$s_data, 0, 1;

    # Bytes 8..31
    $xlate_str->{ENC_SERIAL_NUM} = pack("H*", join '', splice @$s_data, 0, 24);

    # Bytes 32..35  => reserved
    splice @$s_data, 0, 4;

    # Bytes 36..43
    $xlate_str->{ENC_NAME} = pack("H*", join '', splice @$s_data, 0, 8);

    # Bytes 44..47  => reserved
    splice @$s_data, 0, 4;

    # Bytes 48..50
    $xlate_str->{PATH_INFO} = pack("H*", join '', splice @$s_data, 0, 3);

    # Bytes 51  => STANDBY HI PROXY PROXY TAKEOVER
    my $status = hex splice @$s_data, 0, 1;
    $xlate_str->{TAKEOVER} = ($status & BIT_0);
    $xlate_str->{PROXY}    = ($status & BIT_1) >> 1;
    $xlate_str->{HI_PROXY} = ($status & BIT_2) >> 2;
    $xlate_str->{STANDBY}  = ($status & BIT_3) >> 3;

    if ($xlate_str->{ISCSI}) {

        # Bytes 52 .. 55: iscsi tpg tag
        $xlate_str->{TPG_TAG} = hex join '', splice @$s_data, 0, 4;

        # Bytes 56 .. 57: TCP port no
        $xlate_str->{TCP_PORT} = hex join '', splice @$s_data, 0, 2;

        # Bytes 58: Address type
        $xlate_str->{ADDRESS_TYPE} = hex join '', splice @$s_data, 0, 1;

        # Bytes 59: Reserved Byte
        splice @$s_data, 0, 1;

        # Bytes 60..75 IP address
        my @tmp_arr;

        if ($xlate_str->{ADDRESS_TYPE}) {

            # If IPv6 i.e. BYTE[58] = 01h; 128-bit address in BYTE[60:75]
            # : separated hexadecimal numbers => ipv6 representation
            $xlate_str->{ADDRESS} =
                $tmp_arr[0]
              . ($tmp_arr[1]) . ":"
              . $tmp_arr[2]
              . ($tmp_arr[3]) . ":"
              . $tmp_arr[4]
              . ($tmp_arr[5]) . ":"
              . $tmp_arr[6]
              . ($tmp_arr[7]) . ":"
              . $tmp_arr[8]
              . ($tmp_arr[9]) . ":"
              . $tmp_arr[10]
              . ($tmp_arr[11]) . ":"
              . $tmp_arr[12]
              . ($tmp_arr[13]) . ":"
              . $tmp_arr[14]
              . ($tmp_arr[15]);
        } else {

            # If IPv4 i.e. BYTE[58] = 00h; 32-bit address in BYTE[60:63]
            @tmp_arr = splice @$s_data, 0, 4;
            $xlate_str->{ADDRESS} =
                hex($tmp_arr[0]) . "."
              . hex($tmp_arr[1]) . "."
              . hex($tmp_arr[2]) . "."
              . hex($tmp_arr[3]);
        }
    } else {

        # BYTE[52:59] - 8-Byte FCP Node Name i.e. WWNN
        $xlate_str->{FCP_NODENAME} = join '', splice @$s_data, 0, 8;

        # BYTE [60:67] - 8-Byte FCP Port Name i.e. WWPN.
        $xlate_str->{FCP_PORT_NAME} = join '', splice @$s_data, 0, 8;
    }

    return $xlate_str;
} ## end sub xlate_gva_vdisk_asl_info_page

########################################################################
# Sub Name:  xlate_gva_super_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_gva_super_page {
    my ($self, $s_data) = @_;

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Extract the page information

    #
    # 00h - Filer Hostname Page
    #

    $xlate_str->{PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-N => Filer Hostname
    $xlate_str->{FILER_HOSTNAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    #
    # 01h : 32-bit IP Address page
    #

    $xlate_str->{PAGE_CODE_IP} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_IP} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_IP} = hex join '', splice @$s_data, 0, 2;

    # Bytes 4-7 => Count
    $xlate_str->{COUNT} = hex join '', splice @$s_data, 0, 4;

    #my $total_bytes = $count * 4;

    my @ip_addr;

    for (my $index = 0 ; $index < $xlate_str->{COUNT} ; $index++) {
        my @tmp_arr2 = splice @$s_data, 0, 4;
        my $ip_addr =
            hex($tmp_arr2[0]) . "."
          . hex($tmp_arr2[1]) . "."
          . hex($tmp_arr2[2]) . "."
          . hex($tmp_arr2[3]);
        $ip_addr[$index] = $ip_addr;
        $self->sanlog(-msg => "IP_addr Array values" . $ip_addr[$index] . "\n",
                      -msg_level => 1);
    } ## end for (my $index = 0 ; $index...

    my $tmp_arr_len = scalar(@ip_addr);

    $xlate_str->{IP_ADDR} = \@ip_addr;

    #
    # 10h : Virtual Disk Path Page
    #

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_PATH} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_PATH} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_PATH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{VDISK_PATH} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{PAGE_LENGTH_PATH} - 4));

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_VOL} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_VOL} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_VOL} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Length
    $xlate_str->{VOL_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{VOL_NAME} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{PAGE_LENGTH_VOL} - 8));

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_QTREE} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_QTREE} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_QTREE} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Qtree id
    $xlate_str->{QTREE_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{QTREE_NAME} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{PAGE_LENGTH_QTREE} - 8));

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_SNAP} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_SNAP} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_SNAP} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{SNAPSHOT_FILE_PATH} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{PAGE_LENGTH_SNAP} - 4));

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_STS} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_STS} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_STS} = hex join '', splice @$s_data, 0, 2;

    #bytes 4
    $xlate_str->{OTHER} = hex join '', splice @$s_data, 0, 1;

    #bytes 5..6
    my $reserved = hex join '', splice @$s_data, 0, 2;

    #bytes 7 # status bits
    $xlate_str->{STATUS} = hex join '', splice @$s_data, 0, 1;

    # Translate the page information 0
    $xlate_str->{PAGE_CODE_STREAM} = hex join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{PAGE_REVISION_STREAM} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{PAGE_LENGTH_STREAM} = hex join '', splice @$s_data, 0, 2;

    #bytes 4-11
    $xlate_str->{MAIN_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 12..19
    $xlate_str->{PREFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 20..27
    $xlate_str->{SUFFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    # Translate the page information 0
    my $page_code = hex join '', splice @$s_data, 0, 1;

    if ($page_code == 32) {
        $xlate_str->{PAGE_CODE_FCP} = $page_code;

        # Byte 1 => Page revision
        $xlate_str->{PAGE_REVISION_FCP} = hex join '', splice @$s_data, 0, 1;

        # Byte 2-3 => Length
        $xlate_str->{PAGE_LENGTH_FCP} = hex join '', splice @$s_data, 0, 2;

        # Byte 4 .. 6 => Reserved
        splice @$s_data, 0, 3;

        # Byte 7 => FCP CFMODE
        $xlate_str->{FCP_CFMODE} = hex join '', splice @$s_data, 0, 1;

        # Byte 8-15 => FCP NODE NAME
        $xlate_str->{FCP_NODENAME} = join '', splice @$s_data, 0, 8;

        # Bytes 16..23 => FCP PORT NAME
        $xlate_str->{FCP_PORT_NAME} = join '', splice @$s_data, 0, 8;
    } elsif ($page_code == 64) {

        #translate the page information 0
        $xlate_str->{PAGE_CODE_ISCSI_CONFIG} = $page_code;

        #byte 1 => page revision
        $xlate_str->{PAGE_REVISION_ISCSI_CONFIG} = hex join '', splice @$s_data,
          0, 1;

        #byte 2-3 => Length
        $xlate_str->{PAGE_LENGTH_ISCSI_CONFIG} = hex join '', splice @$s_data,
          0, 2;

        #bytes 4-7 => TPG TAG
        $xlate_str->{TPG_TAG_ISCSI_CONFIG} = hex join '', splice @$s_data, 0, 4;

        #bytes 8-9 => TCP PORT
        $xlate_str->{TCP_PORT_ISCSI_CONFIG} = hex join '', splice @$s_data, 0,
          2;

        if ($xlate_str->{PAGE_REVISION_ISCSI_CONFIG} == 0) {

            #bytes 10-13 => IP address
            my @tmp_arr = splice @$s_data, 0, 4;
            $xlate_str->{IP_ADDR_ISCSI_CONFIG} =
                hex($tmp_arr[0]) . "."
              . hex($tmp_arr[1]) . "."
              . hex($tmp_arr[2]) . "."
              . hex($tmp_arr[3]);

            #bytes 14..N
            $xlate_str->{NODE_NAME_ISCSI_CONFIG} =
              pack("H*",
                   join '', splice @$s_data,
                   0, ($xlate_str->{PAGE_LENGTH_ISCSI_CONFIG} - 14));

        } elsif ($xlate_str->{PAGE_REVISION_ISCSI_CONFIG} == 1) {

            #bytes 10 => Address type
            my $address_type = hex join '', splice @$s_data, 0, 1;

            #bytes 11-26 => IP address
            my @tmp_arr = splice @$s_data, 0, 16;

            if ($address_type == 0) {

                #first 4 bytes are populated
                $xlate_str->{IP_ADDR_ISCSI_CONFIG} =
                    hex($tmp_arr[0]) . "."
                  . hex($tmp_arr[1]) . "."
                  . hex($tmp_arr[2]) . "."
                  . hex($tmp_arr[3]);

                #bytes 27..N
                $xlate_str->{NODE_NAME_ISCSI_CONFIG} =
                  pack("H*",
                       join '', splice @$s_data,
                       0, ($xlate_str->{PAGE_LENGTH_ISCSI_CONFIG} - 14));
            } else {

                #16 bytes are populated
                # : separated hexadecimal numbers => ipv6 representation
                $xlate_str->{IP_ADDR_ISCSI_CONFIG} =
                    $tmp_arr[0]
                  . ($tmp_arr[1]) . ":"
                  . $tmp_arr[2]
                  . ($tmp_arr[3]) . ":"
                  . $tmp_arr[4]
                  . ($tmp_arr[5]) . ":"
                  . $tmp_arr[6]
                  . ($tmp_arr[7]) . ":"
                  . $tmp_arr[8]
                  . ($tmp_arr[9]) . ":"
                  . $tmp_arr[10]
                  . ($tmp_arr[11]) . ":"
                  . $tmp_arr[12]
                  . ($tmp_arr[13]) . ":"
                  . $tmp_arr[14]
                  . ($tmp_arr[15]);

                #bytes 27..N
                $xlate_str->{NODE_NAME_ISCSI_CONFIG} =
                  pack("H*",
                       join '', splice @$s_data,
                       0, ($xlate_str->{PAGE_LENGTH_ISCSI_CONFIG} - 27));
            } ## end else [ if ($address_type == 0)

        } ## end elsif ($xlate_str->{PAGE_REVISION...
    } else {
        $self->_set_err("Invalid Page Code received: " . $page_code);
        return;
    }

    $page_code = hex join '', splice @$s_data, 0, 1;

    if ($page_code == 33) {

        # Translate the page information 0
        $xlate_str->{PAGE_CODE_ADAP} = $page_code;

        # Byte 1 => Page revision
        $xlate_str->{PAGE_REVISION_ADAP} = hex join '', splice @$s_data, 0, 1;

        # Byte 2-3 => Length
        $xlate_str->{PAGE_LENGTH_ADAP} = hex join '', splice @$s_data, 0, 2;

        # Byte 4 .. N => FCP ADAPTER PORT CLI IDENTIFIER
        $xlate_str->{CLI_PORT_ID} =
          pack("H*",
               join '', splice @$s_data,
               0, ($xlate_str->{PAGE_LENGTH_ADAP} - 4));
    } elsif ($page_code == 65) {

        #translate the page information 0
        $xlate_str->{PAGE_CODE_ISCSI_ADAP} = $page_code;

        #byte 1 => page revision
        $xlate_str->{PAGE_REVISION_ISCSI_ADAP} = hex join '', splice @$s_data,
          0, 1;

        #byte 2-3 => Length
        $xlate_str->{PAGE_LENGTH_ISCSI_ADAP} = hex join '', splice @$s_data, 0,
          2;

        $xlate_str->{ADAPTER_NAME_ISCSI_ADAP} =
          pack("H*",
               join '', splice @$s_data,
               0, $xlate_str->{PAGE_LENGTH_ISCSI_ADAP});
    } else {
        $self->_set_err("Invalid Page Code received: " . $page_code);
        return;
    }

    return $xlate_str;
} ## end sub xlate_gva_super_page

########################################################################
# Sub Name:  xlate_gva_all_pages()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_gva_all_pages {
    my ($self, $s_data) = @_;

    my $s_t_object = $self->{s_t_obj};

    my $protocol = $s_t_object->{client}->{protocol};

    my $len_recvd = scalar(@$s_data);

    # Extract the gva header information
    my $xlate_str = $self->_read_header($s_data);

    if ($self->error) {
        return;
    }

    unless ($xlate_str->{DATA_LENGTH} == $len_recvd) {
        $self->_set_err(  "Mismatch in the data_length returned"
                        . " in the Header and Page Data received");
        return;
    }

    # Extract the page information

    #
    # 00h - Filer Hostname Page
    #

    $self->sanlog(-msg       => "00h - Filer Hostname Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P00_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P00_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P00_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 8 .. N => ADAPTER DESC
    $xlate_str->{P00_FILER_HOSTNAME} =
      pack("H*", join '', splice @$s_data, 0, ($xlate_str->{PAGE_LENGTH} - 4));

    #
    # 01h - 32-bit IP Address page
    #

    $self->sanlog(-msg       => "01h - 32-bit IP Address page",
                  -msg_level => 1);

    $xlate_str->{P01_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P01_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P01_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Bytes 4-7 => Count
    $xlate_str->{P01_COUNT} = hex join '', splice @$s_data, 0, 4;

    #my $total_bytes = $count * 4;

    my @ip_addr;

    for (my $index = 0 ; $index < $xlate_str->{P01_COUNT} ; $index++) {
        my @tmp_arr2 = splice @$s_data, 0, 4;
        my $ip_addr =
            hex($tmp_arr2[0]) . "."
          . hex($tmp_arr2[1]) . "."
          . hex($tmp_arr2[2]) . "."
          . hex($tmp_arr2[3]);
        $ip_addr[$index] = $ip_addr;
        $self->sanlog(-msg => "IP_addr Array values" . $ip_addr[$index] . "\n",
                      -msg_level => 1);
    } ## end for (my $index = 0 ; $index...

    my $tmp_arr_len = scalar(@ip_addr);

    $xlate_str->{P01_IP_ADDR} = \@ip_addr;

    #
    # 02h - 64-bit IP Address page
    # In Data ONTAP 7-Mode, this page just returns the same data as
    # Page Code 01h i.e. 32-bit IP Address page.
    # In Data ONTAP 8.1 C-Mode, this page is OBSOLETE.
    #

    $self->sanlog(-msg       => "02h - 64-bit IP Address page",
                  -msg_level => 1);

    $xlate_str->{P02_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P02_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P02_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Bytes 4-7 => Count
    $xlate_str->{P02_COUNT} = hex join '', splice @$s_data, 0, 4;

    #my $total_bytes = $count * 4;

    my @ip_addr1;

    for (my $index = 0 ; $index < $xlate_str->{P02_COUNT} ; $index++) {
        my @tmp_arr2 = splice @$s_data, 0, 4;
        my $ip_addr =
            hex($tmp_arr2[0]) . "."
          . hex($tmp_arr2[1]) . "."
          . hex($tmp_arr2[2]) . "."
          . hex($tmp_arr2[3]);
        $ip_addr1[$index] = $ip_addr;
        $self->sanlog(-msg => "IP_addr Array values" . $ip_addr1[$index] . "\n",
                      -msg_level => 1);
    } ## end for (my $index = 0 ; $index...

    my $tmp_arr_len1 = scalar(@ip_addr1);

    $xlate_str->{P02_IP_ADDR} = \@ip_addr1;

    #
    # 10h - Virtual Disk Path Page
    #

    $self->sanlog(-msg       => "10h - Virtual Disk Path Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P10_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P10_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P10_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{P10_VDISK_PATH} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{P10_PAGE_LENGTH} - 4));

    #
    # 11h - Virtual Disk Volume Page
    #

    $self->sanlog(-msg       => "11h - Virtual Disk Volume Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P11_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P11_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P11_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Length
    $xlate_str->{P11_VOL_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{P11_VOL_NAME} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{P11_PAGE_LENGTH} - 8));

    #
    # 12h - Virtual Disk Qtree page
    #

    $self->sanlog(-msg       => "12h - Virtual Disk Qtree page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P12_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P12_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P12_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4-7 => Qtree id
    $xlate_str->{P12_QTREE_ID} = hex join '', splice @$s_data, 0, 4;

    #bytes 8..N
    $xlate_str->{P12_QTREE_NAME} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{P12_PAGE_LENGTH} - 8));

    #
    # 13h - Virtual Disk Snapshot Page
    #

    $self->sanlog(-msg       => "13h - Virtual Disk Snapshot Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P13_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P13_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P13_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{P13_SNAPSHOT_FILE_PATH} =
      pack("H*",
           join '', splice @$s_data,
           0, ($xlate_str->{P13_PAGE_LENGTH} - 4));

    #
    # 14h - Virtual Disk Status Page
    #

    $self->sanlog(-msg       => "14h - Virtual Disk Status Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P14_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P14_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P14_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4
    $xlate_str->{P14_OTHER} = hex join '', splice @$s_data, 0, 1;

    #bytes 5..6
    my $reserved = hex join '', splice @$s_data, 0, 2;

    #bytes 7 # status bits
    $xlate_str->{P14_STATUS} = hex join '', splice @$s_data, 0, 1;

    #
    # 15h - Virtual Disk Stream Sizes Page
    #

    $self->sanlog(-msg       => "15h - Virtual Disk Stream Sizes Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P15_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P15_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P15_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4-11
    $xlate_str->{P15_MAIN_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 12..19
    $xlate_str->{P15_PREFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    #bytes 20..27
    $xlate_str->{P15_SUFFIX_STREAM} = hex join '', splice @$s_data, 0, 8;

    #
    # 16h - Mapped LUNs Page
    #

    $self->sanlog(-msg       => "16h - Mapped LUNs Page",
                  -msg_level => 1);

    $xlate_str->{P16_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P16_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P16_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4-11
    $xlate_str->{P16_ML_COUNT} = hex join '', splice @$s_data, 0, 4;

    #bytes 12..19
    $xlate_str->{P16_ML_IGROUP} = pack("H*", join '', splice @$s_data, 0,
                                       $xlate_str->{P16_PAGE_LENGTH} - 8);

    #
    # 17h - LUN Share Page
    #

    $self->sanlog(-msg       => "17h - LUN Share Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P17_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P17_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P17_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..6 # Reserved
    splice @$s_data, 0, 3;

    #bytes 7  # Shared Read Write
    my $shared = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{P17_WRITE} = ($shared & BIT_1) >> 1;
    $xlate_str->{P17_READ}  = ($shared & BIT_0);

    #
    # 18h - LUN Type page
    #

    $self->sanlog(-msg       => "18h - LUN Type page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P18_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P18_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P18_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..N
    $xlate_str->{P18_OSTYPE} = pack("H*", join '', splice @$s_data, 0,
                                    $xlate_str->{P18_PAGE_LENGTH} - 4);

    # Pages 20h and 21h will be displayed only when the protocol is fcp.

    if ($protocol =~ /fcp/i) {

        #
        # 20h - FCP Target Configuration Information Page
        #

        $self->sanlog(-msg => "20h - FCP Target Configuration Information Page",
                      -msg_level => 1);

        # Translate the page information 0
        $xlate_str->{P20_PAGE_CODE} = join '', splice @$s_data, 0, 1;

        # Byte 1 => Page revision
        $xlate_str->{P20_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

        # Byte 2-3 => Length
        $xlate_str->{P20_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

        # Byte 4 .. 6 => Reserved
        splice @$s_data, 0, 3;

        # Byte 7 => FCP CFMODE
        $xlate_str->{P20_FCP_CFMODE} = hex join '', splice @$s_data, 0, 1;

        # Byte 8-15 => FCP NODE NAME
        $xlate_str->{P20_FCP_NODENAME} = join '', splice @$s_data, 0, 8;

        # Bytes 16..23 => FCP PORT NAME
        $xlate_str->{P20_FCP_PORT_NAME} = join '', splice @$s_data, 0, 8;

        #
        # 21h - FCP Target Adapter Information Page
        #

        $self->sanlog(-msg       => "21h - FCP Target Adapter Information Page",
                      -msg_level => 1);

        # Translate the page information 0
        $xlate_str->{P21_PAGE_CODE} = hex join '', splice @$s_data, 0, 1;

        # Byte 1 => Page revision
        $xlate_str->{P21_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

        # Byte 2-3 => Length
        $xlate_str->{P21_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

        # Byte 4 .. N => FCP ADAPTER PORT CLI IDENTIFIER
        $xlate_str->{P21_CLI_PORT_ID} =
          pack("H*",
               join '', splice @$s_data,
               0, ($xlate_str->{P21_PAGE_LENGTH} - 4));
    }

    #
    # 22h - System Information page
    #

    $self->sanlog(-msg       => "22h - System Information page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P22_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P22_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P22_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    #bytes 4..27
    $xlate_str->{P22_SYS_SOF_REV} = pack("H*", join '', splice @$s_data, 0, 24);

    #bytes 56..63
    $xlate_str->{P22_SYS_SERIAL_NUM} =
      pack("H*", join '', splice @$s_data, 0, 24);

    splice @$s_data, 0, 4;

    #bytes 28..51
    $xlate_str->{P22_MODEL_NAME} = pack("H*", join '', splice @$s_data, 0, 8);

    #bytes 64..70 # reserved
    splice @$s_data, 0, 7;

    # Byte 71 => Reserved NFS ISCSI HTTP HA FCP DAFS CIFS

    my $protocol_license = hex splice @$s_data, 0, 1;

    $xlate_str->{P22_CIFS}     = ($protocol_license & BIT_0);
    $xlate_str->{P22_DAFS}     = ($protocol_license & BIT_1) >> 1;
    $xlate_str->{P22_FCP}      = ($protocol_license & BIT_2) >> 2;
    $xlate_str->{P22_HA}       = ($protocol_license & BIT_3) >> 3;
    $xlate_str->{P22_HTTP}     = ($protocol_license & BIT_4) >> 4;
    $xlate_str->{P22_ISCSI}    = ($protocol_license & BIT_5) >> 5;
    $xlate_str->{P22_NFS}      = ($protocol_license & BIT_6) >> 6;
    $xlate_str->{P22_Reserved} = ($protocol_license & BIT_7) >> 7;

    # Reserved Bytes
    splice @$s_data, 0, 4;

    #
    # 23h - ASL Information Page
    #

    $self->sanlog(-msg       => "23h - ASL Information Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P23_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P23_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P23_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => ASL version
    $xlate_str->{P23_ASL_VERSION} = hex join '', splice @$s_data, 0, 1;

    # Byte 5  => reserved
    splice @$s_data, 0, 1;

    # Bytes 6  => ISCSI_AP
    my $flag = hex splice @$s_data, 0, 1;

    $xlate_str->{P23_AP_ASL}    = ($flag & BIT_4) >> 4;
    $xlate_str->{P23_ISCSI_ASL} = ($flag & BIT_5) >> 5;

    # Byte 7  => reserved
    splice @$s_data, 0, 1;

    # Bytes 8..31
    $xlate_str->{P23_ENC_SERIAL_NUM_ASL} =
      pack("H*", join '', splice @$s_data, 0, 24);

    # Bytes 32..35  => reserved
    splice @$s_data, 0, 4;

    # Bytes 36..43
    $xlate_str->{P23_ENC_NAME_ASL} = pack("H*", join '', splice @$s_data, 0, 8);

    # Bytes 44..47  => reserved
    splice @$s_data, 0, 4;

    # Bytes 48..50
    $xlate_str->{P23_PATH_INFO_ASL} =
      pack("H*", join '', splice @$s_data, 0, 3);

    # Bytes 51  => STANDBY HI PROXY PROXY TAKEOVER
    my $status = hex splice @$s_data, 0, 1;

    $xlate_str->{P23_TAKEOVER_ASL} = ($status & BIT_0);
    $xlate_str->{P23_PROXY_ASL}    = ($status & BIT_1) >> 1;
    $xlate_str->{P23_HI_PROXY_ASL} = ($status & BIT_2) >> 2;
    $xlate_str->{P23_STANDBY_ASL}  = ($status & BIT_3) >> 3;

    if ($xlate_str->{P23_ISCSI_ASL}) {

        # Bytes 52 .. 55: iscsi tpg tag
        $xlate_str->{P23_TPG_TAG_ASL} = hex join '', splice @$s_data, 0, 4;

        # Bytes 56 .. 57: TCP port no
        $xlate_str->{P23_TCP_PORT_ASL} = hex join '', splice @$s_data, 0, 2;

        # Bytes 58: Address type
        $xlate_str->{P23_ADDRESS_TYPE_ASL} = hex join '', splice @$s_data, 0, 1;

        splice @$s_data, 0, 1;

        # Bytes 60..75 IP address
        my @tmp_arr2 = splice @$s_data, 0, 16;

        if ($xlate_str->{P23_ADDRESS_TYPE_ASL}) {
            $xlate_str->{P23_ADDRESS_ASL} =
                $tmp_arr2[0]
              . ($tmp_arr2[1]) . ":"
              . $tmp_arr2[2]
              . ($tmp_arr2[3]) . ":"
              . $tmp_arr2[4]
              . ($tmp_arr2[5]) . ":"
              . $tmp_arr2[6]
              . ($tmp_arr2[7]) . ":"
              . $tmp_arr2[8]
              . ($tmp_arr2[9]) . ":"
              . $tmp_arr2[10]
              . ($tmp_arr2[11]) . ":"
              . $tmp_arr2[12]
              . ($tmp_arr2[13]) . ":"
              . $tmp_arr2[14]
              . ($tmp_arr2[15]);
        } else {
            $xlate_str->{P23_ADDRESS_ASL} =
                hex($tmp_arr2[0]) . "."
              . hex($tmp_arr2[1]) . "."
              . hex($tmp_arr2[2]) . "."
              . hex($tmp_arr2[3]);
        }
    } else {

        # FCP NODE NAME
        $xlate_str->{P23_FCP_NODENAME_ASL} = join '', splice @$s_data, 0, 8;

        # FCP PORT NAME
        $xlate_str->{P23_FCP_PORT_NAME_ASL} = join '', splice @$s_data, 0, 8;

    }

    #
    # 24h - FCP Target Adapter Details Page
    # This page is returned only when the protocol is FCP
    #

    $self->sanlog(-msg       => "24h - FCP Target Adapter Details Page",
                  -msg_level => 1);

    if ($protocol =~ /fcp/i) {

        # Translate the page information 0
        $xlate_str->{P24_PAGE_CODE} = join '', splice @$s_data, 0, 1;

        # Byte 1 => Page revision
        $xlate_str->{P24_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

        # Byte 2-3 => Length
        $xlate_str->{P24_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

        # Byte 4-7 => QUEUE DEPTH
        $xlate_str->{P24_QUEUE_DEPTH} = hex join '', splice @$s_data, 0, 4;

        # Byte 8 .. N => ADAPTER DESC
        $xlate_str->{P24_ADAPTER_DESC} =
          pack("H*",
               join '', splice @$s_data,
               0, $xlate_str->{P24_PAGE_LENGTH} - 8);
    }

    #
    # 25h - Logged Initiators Page
    #

    $self->sanlog(-msg       => "25h - Logged Initiators Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P25_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P25_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P25_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4..7 => Initiator Count
    $xlate_str->{P25_INIT_COUNT} = hex join '', splice @$s_data, 0, 4;

    #
    # 26h - Igroup Information Page
    #

    $self->sanlog(-msg       => "26h - Igroup Information Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P26_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P26_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P26_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => OS Type
    $xlate_str->{P26_OS_TYPE} = hex join '', splice @$s_data, 0, 1;

    # Byte 5  => VSA flag
    $xlate_str->{P26_VSA_FLAG} = hex join '', splice @$s_data, 0, 1;

    # Bytes 6  => USE PROXY FLAG
    $xlate_str->{P26_PROXY_FLAG} = hex join '', splice @$s_data, 0, 1;

    # Byte 7  => USE PROXY - SOLARIS
    $xlate_str->{P26_PROXY_SOLARIS} = hex join '', splice @$s_data, 0, 1;

    # Bytes 8 => USE PROXY - WINDOWS
    $xlate_str->{P26_PROXY_WINDOWS} = hex join '', splice @$s_data, 0, 1;

    # Bytes 9 => USE PROXY - LINUX
    $xlate_str->{P26_PROXY_LINUX} = hex join '', splice @$s_data, 0, 1;

    # Bytes 10 => USE PROXY - HP-UX
    $xlate_str->{P26_PROXY_HP} = hex join '', splice @$s_data, 0, 1;

    # Bytes 11 => USE PROXY - HP-AIX
    $xlate_str->{P26_PROXY_AIX} = hex join '', splice @$s_data, 0, 1;

    # Bytes 12 ..14  => RESERVED
    $xlate_str->{P26_RESERVED} = hex join '', splice @$s_data, 0, 3;

    # Byte 15 => THROTTLE BORROW
    $xlate_str->{P26_THROTTLE_BORROW} = hex join '', splice @$s_data, 0, 1;

    # Bytes 16 .. 19: THROTTLE RESERVE PERCENT
    $xlate_str->{P26_THROTTLE_PER} = hex join '', splice @$s_data, 0, 4;

    # Bytes 20..23: UNRESERVED RESOURCES
    $xlate_str->{P26_UNRESERVED_RES} = hex join '', splice @$s_data, 0, 4;

    #
    # 27h - HA Information Page
    #

    $self->sanlog(-msg       => "27h - HA Information Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P27_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P27_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P27_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => Cluster Enabled
    $xlate_str->{P27_CLUSTER_ENABLED} = hex join '', splice @$s_data, 0, 1;

    # Byte 5 => IC connected
    $xlate_str->{P27_IC_CONNECTED} = hex join '', splice @$s_data, 0, 1;

    # Byte 6..10 => reserved
    splice @$s_data, 0, 5;

    # Bytes 11 => CLUSTER FLAGS WAIT, ERR, GBST, INTO, TOST, TOSCH, CCON
    my $cluster_flag = hex splice @$s_data, 0, 1;

    $xlate_str->{P27_CCON}     = ($cluster_flag & BIT_0);
    $xlate_str->{P27_TOSCH}    = ($cluster_flag & BIT_1) >> 1;
    $xlate_str->{P27_TOST}     = ($cluster_flag & BIT_2) >> 2;
    $xlate_str->{P27_INTO}     = ($cluster_flag & BIT_3) >> 3;
    $xlate_str->{P27_GBST}     = ($cluster_flag & BIT_4) >> 4;
    $xlate_str->{P27_ERR}      = ($cluster_flag & BIT_5) >> 5;
    $xlate_str->{P27_WAIT}     = ($cluster_flag & BIT_6) >> 6;
    $xlate_str->{P27_RESERVED} = ($cluster_flag & BIT_7) >> 7;

    # Bytes 12..N
    $xlate_str->{P27_PARTNER_NAME} =
      pack("H*",
           join '', splice @$s_data,
           0, $xlate_str->{P27_PAGE_LENGTH} - 12);

    #
    # 40h - iSCSI Configuration Page
    #

    $self->sanlog(-msg       => "40h - iSCSI Configuration Page",
                  -msg_level => 1);

    if ($protocol =~ /iscsi/i) {

        #translate the page information 0
        $xlate_str->{P40_PAGE_CODE} = join '', splice @$s_data, 0, 1;

        #byte 1 => page revision
        $xlate_str->{P40_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

        #byte 2-3 => Length
        $xlate_str->{P40_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

        #bytes 4-7 => TPG TAG
        $xlate_str->{P40_TPG_TAG} = hex join '', splice @$s_data, 0, 4;

        #bytes 8-9 => TCP PORT
        $xlate_str->{P40_TCP_PORT} = hex join '', splice @$s_data, 0, 2;

        if ($xlate_str->{P40_PAGE_REVISION} == 0) {

            #bytes 10-13 => IP address
            my @tmp_arr = splice @$s_data, 0, 4;
            $xlate_str->{P40_IP_ADDR} =
                hex($tmp_arr[0]) . "."
              . hex($tmp_arr[1]) . "."
              . hex($tmp_arr[2]) . "."
              . hex($tmp_arr[3]);

            #bytes 14..N
            $xlate_str->{P40_NODE_NAME} =
              pack("H*",
                   join '', splice @$s_data,
                   0, $xlate_str->{P40_PAGE_LENGTH} - 14);

        } elsif ($xlate_str->{P40_PAGE_REVISION} == 1) {

            #bytes 10 => Address type
            my $address_type = hex join '', splice @$s_data, 0, 1;

            #bytes 11-26 => IP address
            my @tmp_arr = splice @$s_data, 0, 16;

            my $size;

            if ($address_type == 0) {

                $size = 15;

                #first 4 bytes are populated
                $xlate_str->{P40_IP_ADDR} =
                    hex($tmp_arr[0]) . "."
                  . hex($tmp_arr[1]) . "."
                  . hex($tmp_arr[2]) . "."
                  . hex($tmp_arr[3]);
            } else {
                $size = 27;

                #16 bytes are populated
                # : separated hexadecimal numbers => ipv6 representation
                $xlate_str->{P40_IP_ADDR} =
                    $tmp_arr[0]
                  . ($tmp_arr[1]) . ":"
                  . $tmp_arr[2]
                  . ($tmp_arr[3]) . ":"
                  . $tmp_arr[4]
                  . ($tmp_arr[5]) . ":"
                  . $tmp_arr[6]
                  . ($tmp_arr[7]) . ":"
                  . $tmp_arr[8]
                  . ($tmp_arr[9]) . ":"
                  . $tmp_arr[10]
                  . ($tmp_arr[11]) . ":"
                  . $tmp_arr[12]
                  . ($tmp_arr[13]) . ":"
                  . $tmp_arr[14]
                  . ($tmp_arr[15]);
            } ## end else [ if ($address_type == 0)

            #bytes 27..N
            $xlate_str->{P40_NODE_NAME} =
              pack("H*",
                   join '', splice @$s_data,
                   0, $xlate_str->{P40_PAGE_LENGTH} - $size);
        } ## end elsif ($xlate_str->{PAGE_REVISION...

        #
        # 41h - iSCSI Adapter Information Page
        #

        $self->sanlog(-msg       => "41h - iSCSI Adapter Information Page",
                      -msg_level => 1);

        #translate the page information 0
        $xlate_str->{P41_PAGE_CODE} = join '', splice @$s_data, 0, 1;

        #byte 1 => page revision
        $xlate_str->{P41_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

        #byte 2-3 => Length
        $xlate_str->{P41_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

        $xlate_str->{P41_ADAPTER_NAME} =
          pack("H*",
               join '', splice @$s_data,
               0, $xlate_str->{P41_PAGE_LENGTH} - 4);
    }

    #
    # 32h - LUN Select Information Page
    #

    $self->sanlog(-msg       => "32h - LUN Select Information Page",
                  -msg_level => 1);

    # Translate the page information 0
    $xlate_str->{P32_PAGE_CODE} = join '', splice @$s_data, 0, 1;

    # Byte 1 => Page revision
    $xlate_str->{P32_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 2-3 => Length
    $xlate_str->{P32_PAGE_LENGTH} = hex join '', splice @$s_data, 0, 2;

    # Byte 4 => Lun select page revision
    $xlate_str->{P32_LUN_PAGE_REVISION} = hex join '', splice @$s_data, 0, 1;

    # Byte 5 => reserved
    splice @$s_data, 0, 1;

    # Byte 6 => Lun select value
    $xlate_str->{P32_LUN_SELECT} = join '', splice @$s_data, 0, 1;

    # Byte 7..10 => reserved
    splice @$s_data, 0, 4;

    # Bytes 11..26 => PREVIOUS LUN SERIAL NUM
    $xlate_str->{P32_PREV_LUN_SER_NUM} =
      pack("H*", join '', splice @$s_data, 0, 16);

    # Bytes 27..30 => RESERVED
    splice @$s_data, 0, 4;

    # Bytes 31 ..46 => LUN CREATION TIME
    my @creation_time = splice @$s_data, 0, 4;

    my @ct = reverse(@creation_time);

    $xlate_str->{P32_CREATION_TIME} = hex join '', @ct;

    splice @$s_data, 0, 5;

    return $xlate_str;
} ## end sub xlate_gva_all_pages

########################################################################
# Sub Name:  _read_header()
# Objective: Return a hash with values set appropriately for gva
#            header
########################################################################

sub _read_header {

    my ($self, $tmp) = @_;
    my %xlate_str;

    $self->_clear_err();

    #translate the header
    #bytes 0-3 => Data_Length
    my $data_length = hex join '', splice @$tmp, 0, 4;

    unless ($data_length) {
        $self->_set_err("data_length not specified");
        return;
    }

    #bytes 4-6 => NetApp Magic Number
    my $magic_number = hex join '', splice @$tmp, 0, 3;

    #byte 7 => revision
    my $revision = hex join '', splice @$tmp, 0, 1;

    $xlate_str{DATA_LENGTH}  = $data_length;
    $xlate_str{MAGIC_NUMBER} = $magic_number;
    $xlate_str{REVISION}     = $revision;

    return (\%xlate_str);

} ## end sub _read_header

########################################################################
# Sub Name:  xlate_mode_sense_geometry_page()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_mode_sense_geometry_page {
    my ($self, $s_data) = @_;

    # Extract the mode_sense header information
    my $xlate_str = $self->_read_mode_page_header($s_data);

    if ($self->error) {
        return $self;
    }

    # Extract the mode_sense header information
    $xlate_str = $self->_read_blk_descriptor($s_data);
    if ($self->error) {
        return $self;
    }

    # Byte 0

    my $byte_0 = hex join '', splice @$s_data, 0, 1;

    $xlate_str->{PAGE_CODE} =
      $byte_0 & (BIT_5 | BIT_4 | BIT_3 | BIT_2 | BIT_1 | BIT_0);
    $xlate_str->{PS} = ($byte_0 & BIT_7) >> 7;

    # Byte 1
    $xlate_str->{PAGE_LENGTH} = hex join '', splice @$s_data, 0, 1;

    # Byte 2
    $xlate_str->{NUMBER_OF_CYLINDERS} = hex join '', splice @$s_data, 0, 3;

    # Byte 5
    $xlate_str->{NUMBER_OF_HEADS} = hex join '', splice @$s_data, 0, 1;

    # Byte 6
    $xlate_str->{STARTING_CYLINDER_WRITE_PRECOMENSATION} = hex join '',
      splice @$s_data, 0, 3;

    # Byte 9
    $xlate_str->{STARTING_CYLINDER_REDUCED_WRITE_CURRENT} = hex join '',
      splice @$s_data, 0, 3;

    # Byte 12
    $xlate_str->{DEVICE_STEP_RATE} = hex join '', splice @$s_data, 0, 2;

    # Byte 14
    $xlate_str->{LANDING_ZONE_CYLINDER} = hex join '', splice @$s_data, 0, 3;

    my $byte_17 = hex join '', splice @$s_data, 0, 1;
    $xlate_str->{RPL} = ($byte_17 & BIT_0);

    # Byte 18
    $xlate_str->{ROTATIONAL_OFFSET} = hex join '', splice @$s_data, 0, 1;

    splice @$s_data, 0, 1;

    # Byte 20
    $xlate_str->{MEDIUM_ROTATION_RATE} = hex join '', splice @$s_data, 0, 2;

    return $xlate_str;
}

########################################################################
# Sub Name:  xlate_mode_sense_vendor_specific()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_mode_sense_vendor_specific {
    my ($self, $s_data) = @_;

    # Extract the mode_sense header information
    my $xlate_str = $self->_read_mode_page_header($s_data);

    if ($self->error) {
        return;
    }

    # Extract the mode_sense block descriptor information

    if ($xlate_str->{BLOCK_DESCRIPTOR_LENGTH}) {
        $xlate_str = $self->_read_blk_descriptor($s_data);
    }

    return $xlate_str;
}

########################################################################
# Sub Name:  _read_blk_descriptor()
# Objective: Return a hash with values set appropriately for mode sense
#            10 block descriptor
########################################################################

sub _read_blk_descriptor {

    my ($self, $tmp) = @_;
    my %xlate_str;

    $xlate_str{NUMBER_OF_BLOCKS} = hex join '', splice @$tmp, 0, 4;

    splice @$tmp, 0, 1;

    $xlate_str{BLOCK_LENGTH} = hex join '', splice @$tmp, 0, 3;

    return (\%xlate_str);

}

########################################################################
# Sub Name:  _read_mode_page_header()
# Objective: Return a hash with values set appropriately for mode sense
#            10 header
########################################################################

sub _read_mode_page_header {

    my ($self, $tmp) = @_;
    my %xlate_str;

    $self->_clear_err();

    #translate the header
    #bytes 0-3 => Data_Length
    my $mode_data_length = hex join '', splice @$tmp, 0, 2;

    unless ($mode_data_length) {
        $self->_set_err("mode_data_length");
        return;
    }

    $xlate_str{MODE_DATA_LENGTH} = $mode_data_length;

    my $medium_type = hex join '', splice @$tmp, 0, 1;

    my $byte_3 = hex join '', splice @$tmp, 0, 1;
    $xlate_str{WP}       = ($byte_3 & BIT_7) >> 7;
    $xlate_str{DPOFUA_1} = ($byte_3 & BIT_4) >> 4;

    my $byte_4 = hex join '', splice @$tmp, 0, 1;
    $xlate_str{LONGLBA_2} = ($byte_4 & BIT_0);

    splice @$tmp, 0, 1;

    $xlate_str{BLOCK_DESCRIPTOR_LENGTH} = hex join '', splice @$tmp, 0, 2;

    return (\%xlate_str);

}

########################################################################
# Sub Name:  xlate_get_dev_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_get_dev_id {
    my ($self, $s_data) = @_;

    my %xlate_str;

    $xlate_str{RESERVED} = splice(@$s_data, 0, 2);

    my $id_valid_byte = hex join '', splice @$s_data, 0, 1;

    $xlate_str{ID_VALID_BIT} = $id_valid_byte & BIT_0;

    $xlate_str{ID_LENGTH} = hex join '', splice @$s_data, 0, 1;

    $xlate_str{DEV_ID} = hex join '', splice @$s_data, 0, 4;

    return (\%xlate_str);
}

########################################################################
# Sub Name:  xlate_report_dev_id()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details: The return hash will have similar key and value pairs
#
########################################################################

sub xlate_report_dev_id {
    my ($self, $s_data) = @_;

    my %xlate_str;

    $xlate_str{ID_LENGTH} = hex join '', splice @$s_data, 0, 4;

    $xlate_str{DEV_ID} = hex join '', splice @$s_data, 0, $xlate_str{ID_LENGTH};

    return (\%xlate_str);
}

########################################################################
# Sub Name:  scsi_bytes_to_strings()
# Objective: Return a hash with values set appropriately for each SCSI
#            cmd data.
# Details:
########################################################################

sub scsi_bytes_to_strings {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($s_data, $s_cmd, $ret_val);

    #
    # Parse the arguments sent in the args hash
    #
    foreach my $param (keys %args) {
        if ($param =~ /^-?scsi_data_ref$/) {
            $s_data = $args{$param};
            unless ($s_data) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::scsi_bytes_to_strings "
                                . "Invalid value passed for '-scsi_data_ref "
                                . "argument: $param");
                return $self;
            }
        } elsif ($param =~ /^-?scsi_cmd_name$/) {
            $s_cmd = $args{$param};
            unless (length($s_cmd)) {
                $self->_set_err(  "Error: "
                                . ref($self)
                                . "::scsi_bytes_to_strings "
                                . "Invalid value passed for '-scsi_cmd_name' "
                                . "argument: $param");
                return $self;
            }
        } else {

            #
            # We got an invalid argument, fail
            #
            $self->_set_err(  "Error: "
                            . ref($self)
                            . "::scsi_bytes_to_strings "
                            . "Invalid argument '$param'");
            return $self;
        } ## end else [ if ($param =~ /^-?scsi_data_ref$/)
    } ## end foreach my $param (keys %args)

    unless ($s_data && $s_cmd) {
        $self->_set_err(  "Error: "
                        . ref($self)
                        . "::scsi_bytes_to_strings "
                        . "SCSI_DATA and SCSI Cmd Name needed");
        return $self;
    }

    if ($s_cmd =~ /RTPG|report target port group/i) {

        # call the translating routine for RTPG data
        $ret_val = $self->xlate_RTPG_bytes($s_data);

    } elsif ($s_cmd =~ /RL|report luns/i) {

        # call the unpacking routine for report luns data
        $ret_val = $self->xlate_report_luns_bytes($s_data);

    } elsif ($s_cmd =~ /^inquiry/i) {

        # call the unpacking routine for Standard Inquiry data
        $ret_val = $self->parse_inquiry_request($s_cmd, $s_data);

    } elsif ($s_cmd =~ /^read_capacity_10/i) {

        # call the unpacking routine for Standard Inquiry data
        $ret_val = $self->xlate_read_capacity_10_bytes($s_data);

    } elsif ($s_cmd =~ /^read_capacity_16/i) {

        # call the unpacking routine for Standard Inquiry data
        $ret_val = $self->xlate_read_capacity_16_bytes($s_data);

    } elsif ($s_cmd =~ /^receive_copy_results/i) {
        $ret_val = $self->xlate_receive_copy_results($s_data);

    } elsif ($s_cmd =~ /^scsi_pr_in_read_keys|^5e 00/i) {
        $ret_val = $self->xlate_scsi_pr_in_read_keys($s_data);

    } elsif ($s_cmd =~ /^scsi_pr_in_read_resv|^5e 01/i) {
        $ret_val = $self->xlate_scsi_pr_in_read_resv($s_data);

    } elsif ($s_cmd =~ /^scsi_pr_in_rep_cap/i) {
        $ret_val = $self->xlate_scsi_pr_in_rep_cap($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_vol_page/i) {
        $ret_val = $self->xlate_gva_vdisk_vol_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_vol_rev_1/i) {
        $ret_val = $self->xlate_gva_vdisk_vol_page_rev1($s_data);

    } elsif ($s_cmd =~ /^gva_virtual_disk_path_page/i) {
        $ret_val = $self->xlate_gva_vdisk_path_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_lun_ostype_page/i) {
        $ret_val = $self->xlate_gva_vdisk_lun_ostype_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_system_info_page/i) {
        $ret_val = $self->xlate_gva_vdisk_system_info_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_mapped_luns_page/i) {
        $ret_val = $self->xlate_gva_vdisk_mapped_luns_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_lun_share_page/i) {
        $ret_val = $self->xlate_gva_vdisk_lun_share_page($s_data);

    } elsif ($s_cmd =~ /^gva_filername_page/i) {
        $ret_val = $self->xlate_gva_filername_page($s_data);

    } elsif ($s_cmd =~ /^gva_32_bit_ip_addr/i) {
        $ret_val = $self->xlate_gva_32_bit_ip_addr($s_data);

    } elsif ($s_cmd =~ /^gva_64_bit_ip_addr/i) {
        $ret_val = $self->xlate_gva_64_bit_ip_addr($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_logged_init_page/i) {
        $ret_val = $self->xlate_gva_vdisk_logged_init_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_fcp_config_page/i) {
        $ret_val = $self->xlate_gva_vdisk_fcp_config_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_fcp_adapter_page/i) {
        $ret_val = $self->xlate_gva_vdisk_fcp_adapter_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_fcp_adap_det_page/i) {
        $ret_val = $self->xlate_gva_vdisk_fcp_adap_det_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_igroup_info_page/i) {
        $ret_val = $self->xlate_gva_vdisk_igroup_info_page($s_data);

    } elsif ($s_cmd =~ /^gva_iscsi_config_info/i) {
        $ret_val = $self->xlate_gva_iscsi_config_info($s_data);

    } elsif ($s_cmd =~ /^gva_iscsi_adapter_info/i) {
        $ret_val = $self->xlate_gva_iscsi_adapter_info($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_cluster_info_page/i) {
        $ret_val = $self->xlate_gva_vdisk_cluster_info_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_lun_select_info_page/i) {
        $ret_val = $self->xlate_gva_vdisk_lun_select_info_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_status_page/i) {
        $ret_val = $self->xlate_gva_vdisk_status_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_qtree_page/i) {
        $ret_val = $self->xlate_gva_vdisk_qtree_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_snapshot_page/i) {
        $ret_val = $self->xlate_gva_vdisk_snapshot_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_asl_info_page/i) {
        $ret_val = $self->xlate_gva_vdisk_asl_info_page($s_data);

    } elsif ($s_cmd =~ /^gva_vdisk_stream_sizes_page/i) {
        $ret_val = $self->xlate_gva_vdisk_stream_sizes_page($s_data);

    } elsif ($s_cmd =~ /^gva_super_page_info/i) {
        $ret_val = $self->xlate_gva_super_page($s_data);
    } elsif ($s_cmd =~ /^gva_all_pages/i) {
        $ret_val = $self->xlate_gva_all_pages($s_data);
    } elsif ($s_cmd =~ /^mode_sense_10/i) {
        $ret_val = $self->xlate_mode_sense_vendor_specific($s_data);
    } elsif ($s_cmd =~ /^mode_geometry_page/i) {
        $ret_val = $self->xlate_mode_sense_geometry_page($s_data);
    } elsif ($s_cmd =~ /^get_dev_id/i) {
        $ret_val = $self->xlate_get_dev_id($s_data);
    } elsif ($s_cmd =~ /^report_dev_id/i) {
        $ret_val = $self->xlate_report_dev_id($s_data);
    } else {
        $self->_set_err("ERROR:: Invalid parameter $s_cmd");
        return;
        $self->sanlog(-msg => ref($self) . "Invalid parameter " . $s_cmd . "\n",
                      -msg_level => 3);

    }

    return $ret_val;

} ## end sub scsi_bytes_to_strings

################################################################################
# Method Name:  get_copy_offload_sense_hash
# Objective:    This method builds a hash of sense data that can be compared to
#               hashes built by ScsiUtil xlate_sense subroutines.
# Param:        
#               exp_asc - expected asc
#               exp_ascq - expected ascq
#               exp_sk - expected sense key
#               exp_fp - expected field pointer
#               exp_bp - expected bit pointer
#               exp_sd_index - index of segment in error
#               exp_sksv - expected sense key specific value
#               is_src - boolean indicating whether the error occurred on the
#               src lun
# Return:       hashref of sense data
################################################################################
sub get_copy_offload_sense_hash {

    my $self = shift;
    my %args = @_;

    # Validate Params
    Params::Validate::validation_options(
                  on_fail => sub { NATE::Exceptions::Argument->throw(shift) },);
    %args = validate(@_, 
        {
                      'exp_asc'  => {
                          type     => SCALAR,
                          required => 1
                      },
                      'exp_ascq' => {
                          type     => SCALAR,
                          required => 1
                      },
                      'exp_sk'   => {
                          type     => SCALAR,
                          required => 1
                       },
                      'exp_fp'   => {
                          type     => SCALAR,
                          required => 1
                      },
                      'exp_bp'   => {
                          type     => SCALAR,
                          optional => 1
                      },
                      'exp_sd_index' => {
                          type     => SCALAR | UNDEF,
                          optional => 1
                      },
                      'exp_sksv' => {
                          type     => SCALAR | UNDEF,
                          optional => 1,
			  default => 1
                      },
                      'is_src'   => {
                          type     => SCALAR | UNDEF,
                          optional => 1,
                          default => 0
                      }
        }
    );

    my $exp_asc      = $args{'exp_asc'};
    my $exp_ascq     = $args{'exp_ascq'};
    my $exp_sk       = $args{'exp_sk'};
    my $exp_fp       = $args{'exp_fp'};
    my $exp_bp       = $args{'exp_bp'};
    my $exp_sd_index = $args{'exp_sd_index'};
    my $exp_sksv     = $args{'exp_sksv'};
    my $is_src       = $args{'is_src'};

    tie my %information_hex_h, 'Tie::IxHash';
    tie my %csi_h,             'Tie::IxHash';
    tie my %sks_h,             'Tie::IxHash';
    tie my %sense_h,           'Tie::IxHash';

    # if segment processing is terminated because a CSCD is unreachable or as the result of a failure in a
    # command sent to a CSCD, then the SENSE-KEY SPECIFIC field shall be set as described in 4.6.2.4.5,
    # with the FIELD POINTER field indicating the first byte of the CSCD descriptor that specifies the CSCD

    if ( $exp_sksv == 0 ) {
        %sks_h = ( 'HEX' => [ '00', '00', '00' ] );
    } else {
        if( $exp_sk == ScsiTool::SK_ILLEGAL_REQUEST ) {
           %sks_h = ('SKSV'          => 1,
                 'C_D'           => 0,
                 'BPV'           => defined $args{'exp_bp'} ? 1 : 0,
                 'BIT_POINTER'   => defined $args{'exp_bp'} ? $args{'exp_bp'} : 0,
                 'FIELD_POINTER' => $exp_fp);
        } else {
           %sks_h = (
                 'SD'            => 1,
                 'BPV'           => defined $args{'exp_bp'} ? 1 : 0,
                 'BIT_POINTER'   => defined $args{'exp_bp'} ? $args{'exp_bp'} : 0,
                 'FIELD_POINTER' => $exp_fp);
        }
    }

    # spc4r32 6.3.3 Errors detected during processing of segment descriptors
    # Information field (holds residual of copy) is only populated if the
    # VALID bit is 1.  For this error, SCSI-T doesn't set the VALID bit.

    %information_hex_h = ('HEX' => ['00', '00', '00', '00']);

    # If ILLEGAL REQUEST is expected, it means that we are building the sense
    # data following the rules for Errors detected before processing of segment
    # descriptors.  Otherwise, follow the rules detailed in 
    # spc4r32 6.3.3 Errors detected during processing of segment descriptors
    # No additional sense data is returned
    if( $exp_sk == ScsiTool::SK_ILLEGAL_REQUEST ) {
        %csi_h = (
            'SRC_DEV_SENSE_OFFSET'  => 0x00,
            'DEST_DEV_SENSE_OFFSET' => 0x00,
            'SEGMENT_DESC_INDEX'    => 0x00,
        );

    } else {
        %csi_h = (
            #'SRC_DEV_SENSE_OFFSET'  => ($is_src) ? 0x12 : 0x00,
            #'DEST_DEV_SENSE_OFFSET' => ($is_src) ? 0x00 : 0x12,
            'SRC_DEV_SENSE_OFFSET'  => 0x00,
            'DEST_DEV_SENSE_OFFSET' => 0x00,
            'SEGMENT_DESC_INDEX'    => $exp_sd_index,
        );
    }

    my %ad_sense_bytes;

    # spc4r32 4.5.3 Fixed format sense data
    %sense_h = (
               'VALID'                   => 0,
               'RESPONSE_CODE'           => ScsiTool::RESP_CODE_CURR_INFO_FIXED,
               'OBSOLETE'                => 0,
               'FILEMARK'                => 0,
               'EOM'                     => 0,
               'ILI'                     => 0,
               'SENSE_KEY'               => $exp_sk,
               'INFORMATION'             => \%information_hex_h,
               'ADDITIONAL_SENSE_LENGTH' => ($exp_sk == ScsiTool::SK_ILLEGAL_REQUEST) ? 14 : 10,
               'COMMAND_SPECIFIC_INFORMATION' => \%csi_h,
               'ADDITIONAL_SENSE_CODE' => $exp_asc,
               'ADDITIONAL_SENSE_CODE_QUALIFIER' => $exp_ascq,
               'FIELD_REPLACEABLE_UNIT_CODE' => 0x00,
               'SKSV'                        => $exp_sksv,
               'SENSE_KEY_SPECIFIC'          => \%sks_h,
               'ADDITIONAL_SENSE_BYTES'      => {});

    return \%sense_h;
}

###############################################################################
##                               Private Methods
################################################################################

sub _st_obj {
    my ($self) = shift @_;
    $self->_clear_err();

    return $self->{s_t_obj} if ($self->{s_t_obj});

    $self->_set_err("ScsiTool Obj not found");
    return $self;
}

1;

=pod

=head1 Package Name

ScsiUtil.pm

=head1 Author

Bhanu Palakodeti

=head1 Description

ScsiUtil will provide lib functions which can be used in SCSI testcases.

=over 4

=item Examples

use San::ScsiTool::ScsiUtil

=back

=head1 Public Methods

=head2 B<get_paths()>

Get the Primary and Secondary paths to a lun.

=over 4

=item Inputs

=item * I<pri>

The array ref that will be updated with primary paths to the lun

=item * I<sec>

The array ref that will be updated with secondary paths to the lun

=item Outputs

=item * O<-1>

Return -1 if unhandled parameter has been passed in

=item Examples

	$paths_info = &ScsiUtil::get_paths( client => $client_obj,
	                                    pri    => \@pri_paths,
	                                    sec    => \@sec_paths,
	                                    filer  => $test_filer,
	                                    lun    => $lun_name);
	return $TCD::CONFIG if($paths_info ==  -1);

=back

=head2 B<scsi_bytes_to_strings()>

Given an array of hex bytes and the name of the SCSI Cmd that reported
these values, a hash ref is returned. The keys in the hash will be
similar to the field names listed in the appropriate SCSI spec, the
values will be either Decimal or Strings.

=over 4

=item Inputs

=item * I<pri>

The array ref that will be updated with primary paths to the lun

=item * I<sec>

The array ref that will be updated with secondary paths to the lun

=item Outputs

=item * O<-1>

Return -1 if unhandled parameter has been passed in

=item Examples

	$paths_info = &ScsiUtil::get_paths( client => $client_obj,
	                                    pri    => \@pri_paths,
	                                    sec    => \@sec_paths,
	                                    filer  => $test_filer,
	                                    lun    => $lun_name);
	return $TCD::CONFIG if($paths_info ==  -1);

=back

=cut
