# -*- perl -*-
#  $Id: $
#
# Copyright (c) 2009 - 2011 NetApp Inc.
# All rights reserved.
#
#
## @summary Just a bunch of helper methods for testing C-mode FCP functionality
## @author (sanap, jennchen) dl-ssan-test@netapp.com
## @status Private
## @pod Not here yet.

package San::Task::FcpHelper;

use strict;
use Exporter;
use List::MoreUtils qw(any);
use Error;
use San::ControllerCluster::Filer10M::Vserver::Lif;
use San::ControllerCluster::Filer10M::Hwresource::Fcpadapter;
use San::ControllerCluster;
use San::SanClient;
use San::SanLog;
use San::SanUtil;
use Data::Dumper;
use IOMan;
use EventMan;
use Tharn;

use vars qw( @EXPORT @ISA );

@ISA = qw( Error SanLog SanUtil );

use constant MAXIMUM_FCP_LIF_SUPPORTED_PER_PORT_BCOM => 16;
###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new FcpHelper object.
# Parameters: None.
# Returns:    A reference to the object.
###############################################################################

sub new {
    my ($class, %args) = @_;

    my $self = {};

    bless $self, $class;

    $self->sanlog(-msg => "FcpHelper::new() - Create FcpHelper Object");

    return $self;
} ## end sub new

###############################################################################
# Method:     verify_vserver_lif_status
# Objective:  Given a vserver, this method verifies the following:
#             1) If FCP exists for the vserver
#             2) If the status matches what you expect the status to be
#             3) If lif(s) exist, the lif(s) status matches what you expect
#             the lif(s) status to be
# Parameters:
#             vserver_obj - A reference to the vserver object which you want
#                           to validate
#             expected_status - The status ('up'/'down') that you expect
#                               the FCP status to be
# Returns:    SanUtil::FALSE if the lif/vserver status did not match the
#             existing status
#             SanUtil::TRUE if everything the vserver and lif status matched
#             the expected status. The function returns TRUE if there are no
#             lifs.
###############################################################################

sub verify_vserver_lif_status {
    my ($self, %args) = @_;
    my ($vserver_obj, $expected_status);
    my ($oper_status, $fcp_obj);

    $self->_clear_err();

    $self->sanlog(msg => "In verify_vserver_lif_status");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver'");
                return $SanUtil::FALSE;
            }
            $vserver_obj = $value;
        } elsif ($key =~ /expected_status/i) {
            unless ($value =~ /up|down/i) {
                $self->_set_err("Incorrect value for key 'expected_status'");
                return $SanUtil::FALSE;
            }
            $expected_status = $value;
        } else {
            $self->_set_err("Invalid key " . $key);
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    unless (defined($vserver_obj) and defined($expected_status)) {
        $self->_set_err("Missing key 'vserver_obj' or 'expected_status'");
        return $SanUtil::FALSE;
    }

    $fcp_obj = $vserver_obj->get_fcp_obj();
    if ($vserver_obj->error()) {
        $self->_set_err(  "Failed to get the FCP object for vserver : "
                        . $vserver_obj->get_name() . " : "
                        . $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify the vserver status

    my $status = $fcp_obj->get_status();

    if ($fcp_obj->error()) {
        $self->_set_err(  "Failed to get the FCP status for vserver : "
                        . $vserver_obj->get_name() . " : "
                        . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    unless ($status =~ /$expected_status/i) {
        $self->_set_err(  "FAIL: The status of the FCP service was " 
                        . $status
                        . " instead of "
                        . $expected_status
                        . " for vserver "
                        . $vserver_obj->get_name());
        return $SanUtil::FALSE;
    }

    # Verify the lif status
    my @fcp_lifs = $vserver_obj->get_lifs(-type => "fcp");
    if ($vserver_obj->error()) {
        $self->_set_err(  "Failed to get the lif objects from vserver : "
                        . $vserver_obj->get_name() . " : "
                        . $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    my $expected_lif_status;

    # fcp status is now up or down instead of online or offline

    if ($status eq "up") {
        $expected_lif_status = "up";
    } else {
        $expected_lif_status = "down";
    }

    if (scalar(@fcp_lifs)) {
        foreach my $lif (@fcp_lifs) {
            $oper_status = $lif->get_status_oper();

            $self->sanlog(msg => $lif->get_name() . " : " . $oper_status);
            unless ($oper_status =~ /$expected_lif_status/i) {
                $self->_set_err(  "FAIL: "
                                . $lif->get_name()
                                . " status is "
                                . $oper_status
                                . " when it should be "
                                . " $expected_lif_status");
                return $SanUtil::FALSE;
            }
        } ## end foreach my $lif (@fcp_lifs)
    } ## end if (scalar(@fcp_lifs))

    return $SanUtil::TRUE;
} ## end sub verify_vserver_lif_status

###############################################################################
# Method:     change_vserver_fcp_status
#
# Objective:  Given a vserver, this method uses the 'fcp start/stop'
#             if cli_style is 7m or 'fcp modify up/down' for cm cli
#             to change the fcp status
#
# Parameters: vserver_obj - vs object, required
#             new_status  - string ('up'/'down'), required
#             cli_style   - string ('cm'/'7m'), optional, default 'cm'
#
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub change_vserver_fcp_status {

    my ($self, %args) = @_;
    my $cmd_interface = 'cli';
    my ($vserver_obj, $new_status, $cli_style);
    my $fcp_obj;

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver'");
                return $SanUtil::FALSE;
            }
            $vserver_obj = $value;
        } elsif ($key =~ /new_status/i) {
            unless (length($value) and ($value =~ /up|down/i)) {
                $self->_set_err("Incorrect value for key 'new_status'");
                return $SanUtil::FALSE;
            }
            $new_status = $value;
        } elsif ($key =~ /cli_style/i) {
            unless (length($value) and ($value =~ /cm|7m/i)) {
                $self->_set_err("Incorrect value for key 'new_status'");
                return $SanUtil::FALSE;
            }
            $cli_style = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        } else {
            $self->_set_err("Invalid key " . $key);
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    unless (defined($vserver_obj) and defined($new_status)) {
        $self->_set_err("Missing key 'vserver_obj' or 'new_status'");
        return $SanUtil::FALSE;
    }

    unless (defined($cli_style)) {
        $cli_style = "cm";
    }

    $fcp_obj = $vserver_obj->get_fcp_obj();
    if ($vserver_obj->error()) {
        $self->_set_err(  "Failed to get the FCP object for vserver : "
                        . $vserver_obj->get_name() . " : "
                        . $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Get current status
    my $curr_status = $fcp_obj->get_status();
    if ($cmd_interface eq 'zapi') {
        if ($curr_status eq 'true') {
            $curr_status = 'up';
        } elsif ($curr_status eq 'false') {
            $curr_status = 'down';
        }
    }
    $self->sanlog(
        msg => "Current status " . $curr_status . " New status " . $new_status);
    if ($fcp_obj->error()) {
        $self->_set_err(
                    "FAIL: Error in retriving fcp status" . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Change fcp status using fcp start/stop
    # if new status is different than current status
    if ($new_status =~ /up/i) {
        if ($cli_style =~ /7m/i) {    # do fcp start
            if (!$fcp_obj->start()) {
                $self->_set_err(  "FAIL: Error starting FCP service for "
                                . $vserver_obj->get_name() . " : "
                                . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {                      # do fcp modify up
            if (!$fcp_obj->modify(status_admin => "up")) {
                $self->_set_err(  "FAIL: Error while fcp modify up "
                                . $vserver_obj->get_name() . " : "
                                . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
    } elsif ($new_status =~ /down/i) {
        if ($cli_style =~ /7m/i) {    # do fcp stop
            if (!$fcp_obj->stop()) {
                $self->_set_err(  "FAIL: Error stoping FCP service for "
                                . $vserver_obj->get_name() . " : "
                                . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {                      # do fcp modify down
            if (!$fcp_obj->modify(status_admin => "down")) {
                $self->_set_err(  "FAIL: Error while fcp modify down "
                                . $vserver_obj->get_name() . " : "
                                . $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
    } ## end elsif ($new_status =~ /down/i)

    # Get the new changed status
    $fcp_obj->refresh();
    if ($fcp_obj->error()) {
        $self->_set_err(
                    "FAIL: Error in retriving fcp status" . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $curr_status = $fcp_obj->get_status();
    if ($cmd_interface eq 'zapi') {
        if ($curr_status eq 'true') {
            $curr_status = 'up';
        } elsif ($curr_status eq 'false') {
            $curr_status = 'down';
        }
    }
    if ($fcp_obj->error()) {
        $self->_set_err(
                    "FAIL: Error in retriving fcp status" . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    if ($curr_status != $new_status) {
        $self->_set_err("FAIL: Failed to change the fcp service status");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub change_vserver_fcp_status

################################################################################
# Method:    run_simple_io
#
# Objective: get devices and run IO for given Vserver
#
# Inputs:    vserver      - vs name, optional
#            client objs  - obj client [], required
#            sanmgr obj   - obj sanmgr, required
#            runlength    - string, optional, default - 240 sec
#            debug_level  - int, optional, default 5 (LOW) for IOMan object
#            wait_flag    - int, optional, if user want to wait till IO is done
#                           default 1 (wait till IO is done and show result)
#                           0 (do not wait, just return IOMan obj for)
#            bk_size      - block size for IO, default 4k
#            io_tool      - sio or dt, default sio
#            IOMan_obj    - IOMan obj, optional, if passed will be used
#            client_devs  - hash ref to client devices
#                           (key client_hostname, value array of dev paths)
#            sanmgr_getdev- devices are already present on Host, just rescan them
#
# Output:    on success SanUtil::TRUE else SanUtil::FALSE
################################################################################

sub run_simple_io {

    my ($self, %args) = @_;
    my ($vserver, $clients, $sanmgr, $runlength, $debug_level, $wait_flag);
    my ($bk_size, $io_tool, @IO_Records, $IOMan, %client_devices, $client_devs,
        $sanmgr_getdev);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?vserver$/) {
            $vserver = $args{$_};
            if (not length($vserver)) {
                $self->_set_err("Invalid value for argument \'-vserver\'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?clients$/) {
            $clients = $args{$_};
            if (not length($clients)) {
                $self->_set_err("Invalid value for argument \'-clients\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?sanmgr$/) {
            $sanmgr = $args{$_};
            if (not length($sanmgr)) {
                $self->_set_err("Invalid value for argument \'-sanmgr\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?runlength$/) {
            $runlength = $args{$_};
            if (not length($runlength)) {
                $self->_set_err("Invalid value for argument \'-runlength\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?debug_level$/) {
            $debug_level = $args{$_};
            if (not length($debug_level)) {
                $self->_set_err("Invalid value for argument \'-debug_level\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?wait_flag$/) {
            $wait_flag = $args{$_};
            if (not length($wait_flag)) {
                $self->_set_err("Invalid value for argument \'-wait_flag\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?bk_size$/) {
            $bk_size = $args{$_};
            if (not length($bk_size)) {
                $self->_set_err("Invalid value for argument \'-bk_size\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?io_tool$/) {
            $io_tool = $args{$_};
            if (not length($io_tool)) {
                $self->_set_err("Invalid value for argument \'-io_tool\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?IOMan_obj$/) {
            $IOMan = $args{$_};
            if (not length($IOMan)) {
                $self->_set_err("Invalid value for argument \'-IOMan_obj\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?client_devs$/) {
            $client_devs = $args{$_};
            if (not length($client_devs)) {
                $self->_set_err("Invalid value for argument \'-client_devs\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?sanmgr_getdev$/) {
            $sanmgr_getdev = $args{$_};
            if (not length($sanmgr_getdev)) {
                $self->_set_err(
                               "Invalid value for argument \'-sanmgr_getdev\'");
                return $SanUtil::FALSE;
            }

        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }    # foreach arg

    unless (defined($clients)) {
        $self->_set_err("Required parameter \'-clients\' missing");
        return $SanUtil::FALSE;
    }

    unless (defined($sanmgr)) {
        $self->_set_err("Required parameter \'-sanmgr\' missing");
        return $SanUtil::FALSE;
    }

    unless (defined($runlength)) {
        $runlength = 240;
    }

    unless (defined($debug_level)) {

        # 1 is LOW for IOMan
        # 7 is HIGH for IOMan debugging
        $debug_level = 1;
    }

    unless (defined($wait_flag)) {
        $wait_flag = 1;
    }

    unless (defined($bk_size)) {

        # Use a random blocksize by default from following possible set
        my @random_block_sizes =
          ("512", "1k", "2k", "4k", "8k", "16k", "32k", "64k", "128k", "256k");
        $bk_size = $random_block_sizes[int(rand(scalar(@random_block_sizes)))];
    }

    unless (defined($io_tool)) {
        $io_tool = "sio";
    }

    unless (defined($IOMan)) {

        # create IOMan obj
        $IOMan = IOMan->new(-debug_level => $debug_level);
        if ($IOMan->error()) {
            $self->_set_err("Failed to create IOMan obj ", $IOMan->errmsg());
            return $SanUtil::FALSE;
        }
    }

    if (defined($client_devs)) {

        %client_devices = %{$client_devs};
        $self->sanlog(-msg => "Received Client devices", -msg_level => 5);
        $self->sanlog(-msg => Dumper(%client_devices),   -msg_level => 5);

    } else {

        my $client_devs = $sanmgr->get_client_dev_list();

        foreach my $client (keys %{$client_devs}) {
            my @free_devices = ();
            foreach my $cdl_entry (@{$client_devs->{$client}}) {
                my $path = $cdl_entry->{path};
                push(@free_devices, $path);
            }
            $client_devices{$client} = \@free_devices;
        }

        $self->sanlog(
                 -msg => "Client devices from SanMgr" . Dumper(%client_devices),
                 -msg_level => 5);
    } ## end else [ if (defined($client_devs...

    # if multiple bk_size are given
    my @bk_sizes = split(/:/, $bk_size);

    # Run through all given block sizes
    foreach my $bk_size (@bk_sizes) {

        # Run througl all clients
        foreach my $client (keys %client_devices) {

            # Create argument hash for IOMan object
            my %hash_args = (-client  => $client,
                             -devices => $client_devices{$client},
                             -io_tool => $io_tool,
                             -iotype  => "sequential",
                             -limit   => "400m",
                             -read    => 0,
                             -write   => 1,
                             -runtime => $runlength,
                             -bs      => $bk_size);

            # SIO specific args
            if ($io_tool eq "sio") {
                $hash_args{"-random_perc"} = "0";
                $hash_args{"-read_perc"}   = "50";
                $hash_args{"-pass_thru"} =
                  "-direct -niceoutput -verify -break_on_dc";
            } else {

                # dt specific args
                $hash_args{"-tool_version"} = "current";
                $hash_args{"-oncerr"}       = "abort";
                $hash_args{"-disable"}      = "pstats";
                $hash_args{"-enable"}       = "noprog,syslog";
                $hash_args{"-noprogt"}      = "15s";
                $hash_args{"-alarm"}        = "7s";
                $hash_args{"-pattern"}      = "iot";
                $hash_args{"-errors"}       = 1;
                $hash_args{"-oflags"}       = "trunc";
                $hash_args{"-trigger"} =
                  "\"cmd:X:\\SAN\\bin\\common\\dt_noprog_script.bat\"";
                $hash_args{"-pass_thru"} = "noprogtt=300s dtype=disk";
            } ## end else [ if ($io_tool eq "sio")

            # Create IO Objs
            @IO_Records = $IOMan->create_io_records(%hash_args);
            if ($IOMan->error()) {
                $self->_set_err("Failed to create IO objects ",
                                $IOMan->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end foreach my $client (keys %client_devices)

        # Start IO for all clients
        $IOMan->start();
        if ($IOMan->error()) {
            $self->_set_err("Failed to start IO", $IOMan->errmsg());
            return $SanUtil::FALSE;
        } else {
            $self->sanlog(msg => "IO Started ");
        }

        if ($wait_flag == 1) {
            if ($self->wait_on_io(-IOMan_obj => $IOMan)) {
                $self->sanlog(msg => "IO finished");
            } else {
                $self->_set_err("Failed to wait on IO", $IOMan->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            $self->sanlog(msg => "Returning IOMan obj");
            return $IOMan;
        }
    }    #end of all block sizes
    return $SanUtil::TRUE;
} ## end sub run_simple_io

################################################################################
# Method:    wait_on_io
#
# Objective: wait untill IO is done and then display results then cleanup IOMan
#
# Inputs:    IOMan_obj - obj of IOMan, required
#            EventMan_obj - obj of EventMan, optional, to check event status
#
# Output:    on success SanUtil::TRUE else SanUtil::FALSE
################################################################################

sub wait_on_io {

    my ($self, %args) = @_;
    my ($IOMan, $EventMan, $errmsg);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?IOMan_obj$/) {
            $IOMan = $args{$_};
            if (not length($IOMan)) {
                $self->_set_err("Invalid value for argument \'-IOMan_obj\'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?EventMan_obj$/) {
            $EventMan = $args{$_};
            if (not length($EventMan)) {
                $self->_set_err("Invalid value for argument \'-IOMan_obj\'");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined($IOMan)) {
        $self->_set_err("Required parameter \'-IOMan_obj\' missing");
        return $SanUtil::FALSE;
    }

    # Wait for IO to finish and then display the results
    # or if IO is still running and events are done, stop the IO
    if (defined $EventMan) {
        while ($EventMan->is_running()) {
            sleep 30;
        }
        if ($EventMan->error()) {
            $errmsg = join('\n',
                           $errmsg, "EventMan failed to complete",
                           $EventMan->errmsg());
            $self->_set_err($errmsg);
        }
        if ($IOMan->error()) {
            $errmsg = "IO did not finish successfully " . $IOMan->errmsg();
            $self->_set_err($errmsg);
            return $SanUtil::FALSE;
        }
        $IOMan->cleanup();
        return $SanUtil::TRUE;

    } else {
        while ($IOMan->is_running()) {
            sleep 30;
        }
        if ($IOMan->error()) {
            $errmsg = "IO did not finish successfully " . $IOMan->errmsg();
            $self->_set_err($errmsg);
            return $SanUtil::FALSE;
        }
        $IOMan->cleanup();
        return $SanUtil::TRUE;
    } ## end else [ if (defined $EventMan)
} ## end sub wait_on_io

################################################################################
# Method:    windows_host_setup
#
# Objective: hacks to optimize windows setup for large number of luns
#            Basically make rescan_timeout to be 0 (instead of 2min)
#            and only online disks which are requested by user
#
# Inputs:    clients      - client objs [], required
#            lunspervol   - how many luns to run from each vol, default 2
#            format       - format the disk or not (1/0), default 1
#
# Output:    on success returns a hash client devices
#            Key   => client_hostname
#            value => devices_path
#            else SanUtil::FALSE
################################################################################

sub windows_host_setup {

    my ($self, %args) = @_;
    my (%client_devices, $lunspervol, $clients, $format);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?clients$/) {
            $clients = $args{$_};
            if (not length($clients)) {
                $self->_set_err("Invalid value for argument \'-clients\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?lunspervol$/) {
            $lunspervol = $args{$_};
            if (not length($lunspervol)) {
                $self->_set_err("Invalid value for argument \'-lunspervol\'");
                return $SanUtil::FALSE;
            }

        } elsif (/-?format$/) {
            $format = $args{$_};
            if (not length($format)) {
                $self->_set_err("Invalid value for argument \'-format\'");
                return $SanUtil::FALSE;
            }

        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined($clients)) {
        $self->_set_err("Required parameter \'-clients\' missing");
        return $SanUtil::FALSE;
    }

    unless (defined($lunspervol)) {
        $lunspervol = 2;
    }

    unless (defined($format)) {
        $format = 1;
    }

    %client_devices = ();

    foreach my $cl_obj (@{$clients}) {
        my $dev_info  = {};
        my $cl_name   = $cl_obj->hostname();
        my @dev_paths = ();

        # get_devices on win client obj
        my @win_devices = $cl_obj->get_devices(-rescan_timeout => 0);

        if (scalar @win_devices) {
            foreach my $device_hash (@win_devices) {

                my @lunPath_split =
                  split(/\//, $device_hash->{diskObj}->{lunPath});
                my $vol_name =
                  $device_hash->{filer_name} . "_" . $lunPath_split[2];
                my $device = $device_hash->{path};

                # Converting device_paths to be compatible with dt. Works for sio too
                # dt prefers the device name format to be Titlecase instead of all caps
                $device =~ tr/[H,Y,S,I,C,A,L,R,V,E]/[h,y,s,i,c,a,l,r,v,e]/;

                if (not defined(@{$dev_info->{$vol_name}})) {
                    push(@dev_paths,                $device);
                    push(@{$dev_info->{$vol_name}}, $device);
                    if ($format) {
                        $self->sanlog(-msg => "Online the disk " . $device,
                                      -msg_level => 3);
                        $device_hash->{diskObj}->online();
                    }

                } elsif (scalar(@{$dev_info->{$vol_name}}) < $lunspervol) {
                    push(@dev_paths,                $device);
                    push(@{$dev_info->{$vol_name}}, $device);
                    if ($format) {
                        $self->sanlog(-msg => "Online the disk " . $device,
                                      -msg_level => 3);
                        $device_hash->{diskObj}->online();
                    }
                }
            } ## end foreach my $device_hash (@win_devices)
        } ## end if (scalar @win_devices)

        # store the reference to device array with client name as key
        $client_devices{$cl_name} = \@dev_paths;
    } ## end foreach my $cl_obj (@{$clients...

    $self->sanlog(-msg => "Client devices",        -msg_level => 5);
    $self->sanlog(-msg => Dumper(%client_devices), -msg_level => 5);
    return \%client_devices;
} ## end sub windows_host_setup

################################################################################
# Method:    client_verify_devices
#
# Objective: get_devices for client, then create devmgr and find free devices
#            for given vserver
#
# Inputs:    client   - client obj, required
#            vserver  - vserver name, required
#
# Output:    SanUtil::TRUE if devices for given vservers are found
#            else SanUtil::FALSE
################################################################################

sub client_verify_devices {

    my ($self, %args) = @_;
    my ($client, $vserver);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?client$/) {
            $client = $args{$_};
            if (not length($client)) {
                $self->_set_err("Invalid value for argument \'-client\'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?vserver$/) {
            $vserver = $args{$_};
            if (not length($vserver)) {
                $self->_set_err("Invalid value for argument \'-vserver\'");
                return $SanUtil::FALSE;
            }

        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined($client)) {
        $self->_set_err("Required parameter \'-clients\' missing");
        return $SanUtil::FALSE;
    }
    unless (defined($vserver)) {
        $self->_set_err("Required parameter \'-vserver\' missing");
        return $SanUtil::FALSE;
    }

    my $client_obj = $client;

    # Rediscover luns
    $client_obj->discover();
    if ($client_obj->error()) {
        $self->_set_err("discovery failed " . $client_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Create a new device manager
    my $DMObj = $client_obj->create_devmgr();
    if ($client_obj->error()) {
        $self->_set_err(
                       "device manager create failed " . $client_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the free devices on it
    my @free_devices = ();
    @free_devices =
      $DMObj->get_free_devices(-filer   => $vserver,
                               -refresh => 1);
    if ($DMObj->error) {
        $self->_set_err($DMObj->errmsg());
        return $SanUtil::FALSE;
    }

    if (scalar(@free_devices) < 1) {
        $self->_set_err("No free devices found " . $client_obj->hostname());
        return $SanUtil::FALSE;
    }
    $self->sanlog(msg => "Free devs found");
    $self->sanlog(msg => Dumper(@free_devices));

    return $SanUtil::TRUE;
} ## end sub client_verify_devices

###############################################################################
# Method:     verify_client_login_logout
# Objective:  This method verifies that a client is logged in to a given
#             vserver
# Parameters:
#             vserver_obj - A reference to the vserver object
#             client_obj - A reference to the client object for which we want
#             to verify login
#             expected_state - Do you expect the client to be logged in or not?
#             valid values for this parameter are 'in'/'out'
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub verify_client_login_logout {
    my ($self, %args) = @_;
    my ($vserver_obj, $client_obj, $expected_state);
    my @wwpns;

    $self->_clear_err();
    $self->sanlog(msg => "In verify_client_login_logout");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver'");
                return $SanUtil::FALSE;
            }
            $vserver_obj = $value;
        } elsif ($key =~ /client_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'client'");
                return $SanUtil::FALSE;
            }
            $client_obj = $value;
        } elsif ($key =~ /expected_state/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'expected_state'");
                return $SanUtil::FALSE;
            }
            $expected_state = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    # Get the client wwpns
    my @initiators = $client_obj->get_initiators();
    if ($client_obj->error()) {
        $self->_set_err("Error getting initiators from " . $client_obj);
        return $SanUtil::FALSE;
    }
    $self->sanlog(msg => "List of initiators: ");
    foreach my $initiator (@initiators) {
        my $portname = $initiator->get_portname();
        push(@wwpns, $portname);
        $self->sanlog(msg => $portname . "\n");
    }

    return ($self->verify_initiator_login_logout(
                                               vserver_obj    => $vserver_obj,
                                               wwpns          => @wwpns,
                                               expected_state => $expected_state
              ));
} ## end sub verify_client_login_logout

###############################################################################
# Method:     verify_initiator_login_logout
# Objective:  This method verifies whether a list of given wwpns are logged in
#             or out of a vserver. The method will return FALSE even if one
#             of the wwpns in the list does not confirm with what is expected.
# Parameters:
#             vserver_obj - A reference to the vserver object
#             client_obj - A reference to the client object for which we want
#             to verify login
#             expected_state - Do you expect the client to be logged in or not
#             valid values for this parameter are 'in'/'out'
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub verify_initiator_login_logout {

    my ($self, %args) = @_;
    my ($vserver_obj, @wwpns, $expected_state);

    $self->_clear_err();

    $self->sanlog(msg => "In verify_initiator_login_logout");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver'");
                return $SanUtil::FALSE;
            }
            $vserver_obj = $value;
        } elsif ($key =~ /wwpns/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'expected_wwpn'");
                return $SanUtil::FALSE;
            }
            @wwpns = ${value};
        } elsif ($key =~ /expected_state/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'expected_state'");
                return $SanUtil::FALSE;
            }
            $expected_state = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    # Get FCP object
    my $fcp_obj = $vserver_obj->get_fcp_obj();
    if ($vserver_obj->error()) {
        $self->_set_err("Error getting fcp object " . $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the filer initiators
    my @fcp_initiators = $fcp_obj->get_initiators();
    if ($fcp_obj->error()) {
        $self->_set_err(
            "Error getting fcp initiator from the filer " . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(msg => "Number of initiators from the filer "
                  . scalar(@fcp_initiators));
    $self->sanlog(msg => "Initiators that are currently logged in : ");

    my @filer_wwpns = ();
    foreach my $fcp_initiator (@fcp_initiators) {
        my $portname = $fcp_initiator->{'port_name'};
        push(@filer_wwpns, $portname);
        $self->sanlog(msg => $portname . "\n");
    }

    # Match the wwpns found on the filer with what is expected
    foreach my $wwpn (@wwpns) {
        $self->sanlog(msg => "WWPN: " . $wwpn);
        if (List::MoreUtils::any { $_ =~ /$wwpn/i } @filer_wwpns)
        {    # Found the initiator
            $self->sanlog(  msg => "Found initiator " 
                          . $wwpn
                          . " logged in to "
                          . $vserver_obj->get_name());

            unless ($expected_state =~ /login/i) {

                # Was expecting the client to be logged out but it is logged in
                $self->_set_err(  "The initiator " 
                                . $wwpn
                                . " should not have been "
                                . "logged in to "
                                . $vserver_obj->get_name());
                return $SanUtil::FALSE;
            } ## end unless ($expected_state =~...
        } else {    # Did not find the initiator
            $self->sanlog(  msg => "Did not find the initiator " 
                          . $wwpn
                          . " logged in to "
                          . $vserver_obj->get_name());
            if ($expected_state =~ /login/i) {

                # Was expecting the client to be logged in and it is not
                $self->_set_err(  "The initiator " 
                                . $wwpn
                                . " should have been "
                                . "logged in to "
                                . $vserver_obj->get_name());
                return $SanUtil::FALSE;
            } ## end if ($expected_state =~...
        } ## end else [ if (List::MoreUtils::any...
    } ## end foreach my $wwpn (@wwpns)

    return $SanUtil::TRUE;

} ## end sub verify_initiator_login_logout

###############################################################################
# Method:     get_fcp_path_info_controller
#
# Objective:  create a hash with following info
#             Lun Name - primary key
#             Lun Serial, Lun Path, Target, Total Path, Active-O path,
#             Active-NO path
#
# Parameters: vserver_obj - A reference to the vserver object
#
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub get_fcp_path_info_controller {

    my ($self, %args) = @_;
    my ($vserver_obj);
    my ($fcp_obj);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver'");
                return $SanUtil::FALSE;
            }
            $vserver_obj = $value;
        } else {
            $self->_set_err("Invalid key " . $key);
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    unless (defined($vserver_obj)) {
        $self->_set_err("Missing key 'vserver_obj'");
        return $SanUtil::FALSE;
    }

    # Get FCP object
    $fcp_obj = $vserver_obj->get_fcp_obj();
    if ($vserver_obj->error()) {
        $self->_set_err("Error getting fcp object " . $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the filer initiators
    my @fcp_initiators = $fcp_obj->get_initiators();
    if ($fcp_obj->error()) {
        $self->_set_err(
            "Error getting fcp initiator from the filer " . $fcp_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(msg => "Number of initiators from the filer "
                  . scalar(@fcp_initiators));
    $self->sanlog(msg => "Initiators that are currently logged in : ");

    my @filer_wwpns = ();
    foreach my $fcp_initiator (@fcp_initiators) {
        my $portname  = $fcp_initiator->{'wwpn'};
        my $nodename  = $fcp_initiator->{'wwnn'};
        my $lif       = $fcp_initiator->{'lif'};
        my $vserver   = $fcp_initiator->{'vserver'};
        my $port_addr = $fcp_initiator->{'port_address'};
        push(@filer_wwpns, $portname);
        $self->sanlog(  msg => $portname . " "
                      . $nodename . " "
                      . $lif . " "
                      . $vserver . " "
                      . $port_addr
                      . "\n");
    } ## end foreach my $fcp_initiator (...

} ## end sub get_fcp_path_info_controller

################################################################################
# Method:    run_fault_SFO
#
# Objective: Use Eventman and run SFO fault
#
# Inputs:    sanmgr_obj - SanMgr obj (required)
#
# Output:    SanUtil::TRUE or SanUtil::FALSE
################################################################################

################################################################################
# Method:    run_fault_SFO
#
# Objective: Use Eventman and run SFO fault
#
# Inputs:    sanmgr_obj - SanMgr obj (required)
#
# Output:    EventMan obj or SanUtil::FALSE
################################################################################

sub run_fault_SFO {

    my ($self, %args) = @_;
    my ($sanmgr_obj, $cycle_count);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?sanmgr_obj$/) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument \'-sanmgr_obj\'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?cycle_count$/) {
            $cycle_count = $args{$_};
            if (not length($cycle_count)) {
                $self->_set_err("Invalid value for argument \'-cycle_count\'");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined($sanmgr_obj)) {
        $self->_set_err("Missing key 'sanmgr'");
        return $SanUtil::FALSE;
    }

    unless (defined($cycle_count)) {
        $cycle_count = 1;
    }

    my @filers = $sanmgr_obj->get_objects(-type => "node");

    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to create ControllerCluster object",
                        $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # choose one filer to execute few debug commands
    my $filer = $filers[0];

    # make a controller cluster object for cluster
    my $cluster_obj =
      San::ControllerCluster->new(-cserver => $filer->get_name());
    if ($cluster_obj->error()) {
        $self->_set_err("Failed to create ControllerCluster object",
                        $cluster_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Initialize event man
    my $EventMan_obj = EventMan->new();
    if ($EventMan_obj->error()) {
        $self->_set_err("Failed to create EventMan object",
                        $EventMan_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # register the cluster with eventman
    $EventMan_obj->add_resource(-name => $cluster_obj->get_name(),
                                -obj  => $cluster_obj,
                                -type => "controllerclusters");

    # register the sanclient objects with eventman
    my @clients = $sanmgr_obj->get_objects(-type => "client");

    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get client objects", $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    foreach my $sanClient_obj (@clients) {
        $EventMan_obj->add_resource(-name => $sanClient_obj->hostname(),
                                    -obj  => $sanClient_obj,
                                    -type => "sanclients");
    }

    if ($EventMan_obj->error()) {
        $self->_set_err("Failed to add resource to EventMan object",
                        $EventMan_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # starting sfo takegive event
    $EventMan_obj->add_event(-event_name         => "sfo_takeforcedgiveCm",
                             -sfo_timeout        => 900,
                             -symmetrical_events => 1,
                             -path_validation    => 1,
                             -path_validation_abort_on_error => 0,
                             -path_validation_min_wait_time  => 300,
                             -path_validation_on_takeover    => 0,
                             -lif_validation                 => 1,
                             -lif_validation_abort_on_error  => 0,
                             -controller_down_time           => 300,
                             -cycle_count                    => $cycle_count);

    # then kick off the events in the background
    my $subtest =
      $EventMan_obj->run_events(-async => 1,
                                -runid => "fcp_sfo_tests");
    if ($EventMan_obj->error()) {
        $self->_set_err("Failed to successfully complete the event",
                        $EventMan_obj->errmsg());
        return $SanUtil::FALSE;
    }

    return $EventMan_obj;
} ## end sub run_fault_SFO

################################################################################
# Method:    run_fault_adapter_up_down
#
# Objective: down and up the fcp adapters
#
# Inputs:    sanmgr_obj - SanMgr obj (required)
#            cycle_count - optional, default 1
#
# Output:    SanUtil::TRUE or SanUtil::FALSE
################################################################################

sub run_fault_adapter_up_down {

    my ($self, %args) = @_;
    my ($sanmgr_obj, $cycle_count);

    $self->_clear_err();

    # argument parsing
    foreach (keys %args) {
        if (/-?sanmgr_obj$/) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument \'-sanmgr_obj\'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?cycle_count$/) {
            $cycle_count = $args{$_};
            if (not length($cycle_count)) {
                $self->_set_err("Invalid value for argument \'-cycle_count\'");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    } ## end foreach (keys %args)

    unless (defined($sanmgr_obj)) {
        $self->_set_err("Missing key 'sanmgr'");
        return $SanUtil::FALSE;
    }

    unless (defined($cycle_count)) {
        $cycle_count = 1;
    }

    my $fcp_info = $self->collect_fcp_info(-sanmgr_obj => $sanmgr_obj);
    if ($self->error()) {
        $self->sanlog("get fcp info had failure ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the client objects
    my @clients = $sanmgr_obj->get_objects(type => "client");

    # Level 1: Getting cluster information from fcp_info
    my $cluster      = $fcp_info->{'cluster_obj'};
    my $cluster_name = $cluster->get_name();

    # Level 2: Getting filer information from the cluster level
    my $num_filers = $fcp_info->{$cluster_name}->{'num_filer_objs'};
    my $filers     = $fcp_info->{$cluster_name}->{'filer_objs'};

    while ($cycle_count > 0) {
        for (my $i = 0 ; $i < $num_filers ; $i++) {
            my $filer      = $$filers[$i];
            my $filer_name = $filer->get_name();

            # Level 3: Getting port information from the filer level
            my $ports =
              $fcp_info->{$cluster_name}->{$filer_name}->{'port_objs'};
            my $num_ports =
              $fcp_info->{$cluster_name}->{$filer_name}->{'num_port_objs'};

            for (my $j = 0 ; $j < $num_ports ; $j++) {

                my $port        = $$ports[$j];
                my $port_status = $port->get_status();
                my $port_name   = $port->get_adapter();

                if ($port_status eq "online") {

                    # Level 4: Getting LIFs information from the port level
                    my $num_lifs =
                      $fcp_info->{$cluster_name}->{$filer_name}->{$port_name}
                      ->{'num_lif_objs'};

                    if ($num_lifs > 0) {
                        my $lifs =
                          $fcp_info->{$cluster_name}->{$filer_name}
                          ->{$port_name}->{'lif_objs'};

                        $self->sanlog(  msg => "Down LIFs on "
                                      . $filer_name . " : "
                                      . $port_name);
                        $self->modify_lifs_status(-lif_objs_ref => $lifs,
                                                  -status_admin => "down");
                        if ($self->error) {
                            $self->sanlog(msg => $self->errmsg());
                            $self->_set_err(  "Lif modify down failed "
                                            . $filer_name . " : "
                                            . $port_name);
                            return $SanUtil::FALSE;
                        }

                        $self->sanlog(  msg => "Down the adapter "
                                      . $filer_name . " : "
                                      . $port_name);
                        $self->modify_fcp_adapter_state(-filer_obj => $filer,
                                                        -adapter => $port_name,
                                                        -state   => "down",
                                                        -modify_success => 1);
                        if ($self->error) {
                            $self->sanlog(msg => $self->errmsg());
                            $self->_set_err(  "adapter modify down failed "
                                            . $filer_name . " : "
                                            . $port_name);
                            return $SanUtil::FALSE;
                        }

                        $self->sanlog(msg =>
                                  "Wait for 20 sec before bringing up adapter");
                        sleep(20);

                        $self->sanlog(  msg => "Up the adapter "
                                      . $filer_name . " : "
                                      . $port_name);
                        $self->modify_fcp_adapter_state(-filer_obj => $filer,
                                                        -adapter => $port_name,
                                                        -state   => "up",
                                                        -modify_success => 1);
                        if ($self->error) {
                            $self->sanlog(msg => $self->errmsg());
                            $self->_set_err(  "adapter modify up failed "
                                            . $filer_name . " : "
                                            . $port_name);
                            return $SanUtil::FALSE;
                        }

                        $self->sanlog(  msg => "Up LIFs on "
                                      . $filer_name . " : "
                                      . $port_name);
                        $self->modify_lifs_status(-lif_objs_ref => $lifs,
                                                  -status_admin => "up");
                        if ($self->error) {
                            $self->sanlog(msg => $self->errmsg());
                            $self->_set_err(  "Lif modify up failed "
                                            . $filer_name . " : "
                                            . $port_name);
                            return $SanUtil::FALSE;
                        }

                        # Rescan the host for the paths which just came back up
                        foreach my $client_obj (@clients) {
                            $self->sanlog(msg => "rescanning devices on client "
                                          . $client_obj->hostname());
                            $client_obj->{WinClient}->rescanDisks();
                        }
                    } else {
                        $self->sanlog(  msg => "Adapter "
                                      . $filer_name . " : "
                                      . $port_name
                                      . " has no LIFs");
                    }
                } else {
                    $self->sanlog(  msg => "Adapter "
                                  . $filer_name . " : "
                                  . $port_name
                                  . " is already down");
                }
            } ## end for (my $j = 0 ; $j < $num_ports...
        } ## end for (my $i = 0 ; $i < $num_filers...
        $cycle_count = $cycle_count - 1;
    } ## end while ($cycle_count > 0)
    return $SanUtil::TRUE;
} ## end sub run_fault_adapter_up_down

################################################################################
# Method:    generate_wwns
#
# Objective: to generate possible node/port names
#            algorithm same as of development
# http://wikid.netapp.com/w/Smokejumper/Bcom/Documents/Protocols_Mgmt_DS
#
# Inputs:    sanmgr_obj - SanMgr obj (required)
#            to_gen - Number of wwns to generate (optional, default 10)
#
# Output:    Arrya of (or reference to) valid node/port names on success
#            else SanUtil::FALSE
################################################################################

sub generate_wwns {

    my ($self, %args) = @_;
    my ($sanmgr_obj, $to_gen);
    my (@valid_names);

    $self->_clear_err();

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'sanmgr_obj'");
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        } elsif ($key =~ /to_gen/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'to_gen'");
                return $SanUtil::FALSE;
            }
            $to_gen = $value;
        } else {
            $self->_set_err("Invalid key " . $key);
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    unless (defined($sanmgr_obj)) {
        $self->_set_err("Missing key 'sanmgr_obj'");
        return $SanUtil::FALSE;
    }

    if (!defined($to_gen)) {

        # use default value of 10
        $to_gen = 10;
    }

    # grab filer objs
    # TODO: When we move to SanMgr from FitMgr this call will change to
    # $sanmgr_obj->get_controllers();
    my @filers = $sanmgr_obj->get_objects(-type => "node");

    # choose one filer to execute few debug commands
    my $filer = $filers[0];

    # find out offset and count currently being used
    my $cmd =
      "set diagnostic  -confirmations off; debug smdb table wwnUsedSeedTable show";
    my @buf = $filer->execute(-cmd => $cmd, -raw_data => 1);
    if ($filer->error()) {
        self->_set_err("Unable to execute " . $cmd, $filer->errmgs());
        return $SanUtil::FALSE;
    }

    my @op   = split(/\n/, $buf[0]);
    my @temp = split(/\|/, $op[4]);
    my $count  = $temp[2] + 20;
    my $offset = $temp[1];
    my $mac    = $temp[0];
    my $n      = 0;
    my $next_wwn;

    # Find out which WWPNs are currently in use, this is to fix a burt 718475
    # where if the previous attempt to restore old WWPN is failed and then re-run the
    # same test, we will end up generating same WWNs and the next test run will fail
    # with duplicate WWN assignment error.

    my @vservers = $sanmgr_obj->get_objects(-type => "vserver");
    my @wwns = ();

    foreach my $vs (@vservers) {
	my $fcp = $vs->get_fcp_obj();
	my $fcp_wwnn = $fcp->get_nodename();
	my @fcp_wwpns = $fcp->get_portnames();
	push(@wwns, $fcp_wwnn);

	foreach my $wwpn (@fcp_wwpns) {
	    push(@wwns, $wwpn->{wwpn});	    
	}
    }

    $self->sanlog(msg => "Current WWNs " . scalar(@wwns));
    foreach my $wwn (@wwns) {
	$self->sanlog(msg => $wwn);
    }

    # next wwn = NetApp Unique ID "2" then
    # count (which is how many WWN have already been generated using this eth port)
    # then mac for e0a then if we run out of total possible WWN using e0a move to e0b
    while ($n < $to_gen) {
        $count = $count + 1;
        my $hex_count = sprintf("%x", $count);
        my @hex_split = split("", $hex_count);
        if (@hex_split eq 3) {
            $next_wwn = "2"
              . $hex_split[0] . ":"
              . $hex_split[1]
              . $hex_split[2] . ":"
              . $mac;
        } elsif (@hex_split eq 2) {
            $next_wwn =
              "2" . "0" . ":" . $hex_split[0] . $hex_split[1] . ":" . $mac;
        } elsif (@hex_split eq 1) {
            $next_wwn = "2" . "0" . ":" . "0" . $hex_split[0] . ":" . $mac;
        }

	# make sure new generated WWN is not already present on controllers
	my $duplicate = 0;
	foreach my $wwn (@wwns) {
	    if ($next_wwn eq $wwn) {
		$self->sanlog(msg => "Duplicate WWN generated, $next_wwn skip this one");
		$duplicate = 1;
		last;
	    }
	}
	if ($duplicate) {
	    $self->sanlog(msg => "Duplicate WWN generated - $next_wwn");
	} else {
	    $self->sanlog(msg => "New WWN generated - $next_wwn");
	    push(@valid_names, $next_wwn);
	}
        $n = $n + 1;
    } ## end while ($n < $to_gen)

    return (wantarray()) ? @valid_names : \@valid_names;
} ## end sub generate_wwns

###############################################################################
# Method:     collect_fcp_info
# Objective:  This method goes out and collects information about the fcp
#             resources in the system and arranges the information in a hash
#             The structure of the hash is as follows:
#
#             Level 1: Cluster
#             If $fcp_info is a reference to the hash, then level 1 looks like
#             this - $fcp_info->{cluster_obj}
#                    $fcp_info->{<cluster name>} # Can get the cluster name from the cluster_obj.
#
#             Level 2: Filers
#             This level of hash is referenced by <cluster name> from level
#             1. It looks like this -
#             {<cluster_name>}->{num_filer_objs} # Number of filers in the cluster
#             {<cluster_name>}->{filer_objs} # Array of filer objects
#             {<cluster_name>}->{<filer name>} # One entry for each filer in the array
#
#             Level 3: Ports
#             This level of hash is referenced by each <filer name> from
#             level 2. It looks like this -
#             {<filer name>}->{num_port_objs} # Number of fcp ports on the filer
#             {<filer name>}->{port_objs} # Array of port objects
#             {<filer name>}->{<port name>} # One entry for each port in the array
#
#             Level 4: Lifs
#             This level of hash is referenced by each <vserver name> from
#
# Parameters:
#             sanmgr_obj - Sanmgr object
#
# Returns:    $SanUtil::FALSE if something failed
#             $SanUtil::TRUE if all went well
###############################################################################

sub collect_fcp_info {

    my ($self, %args) = @_;
    my ($sanmgr_obj);
    my $fcp_info = {};

    $self->_clear_err();

    $self->sanlog(msg => "In collect_fcp_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'sanmgr_obj'");
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    # Get a controller cluster object from SanMgr
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    my $cluster_obj = $cluster_objs[0];

    # Put the controller cluster object in the fcp_info hash
    $fcp_info->{'cluster_obj'} = $cluster_obj;
    my $cluster_name = $cluster_obj->get_name();
    $self->sanlog(msg => "Cluster name: " . $cluster_name);

    # Get all the nodes (which in this case are filers) in the cluster
    my @filer_objs = $sanmgr_obj->get_objects(type => "node");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Put the filer information in the fcp_info hash
    $fcp_info->{$cluster_name}->{'num_filer_objs'} = scalar(@filer_objs);
    $fcp_info->{$cluster_name}->{'filer_objs'}     = \@filer_objs;

    foreach my $filer (@filer_objs) {
        my $filer_name = $filer->get_name();
        $self->sanlog(msg => $filer_name);

        # Get the adapters from each filer
        my @all_adapters = $filer->get_fcpadapter_objs();
        if ($filer->error()) {
            $self->_set_err(
                        "Failed to get the fcp adapter objects from the filer");
            return $SanUtil::FALSE;
        }

        # just get adapters for current filer
        my @adapters = ();
        foreach my $adapter (@all_adapters) {

            my $adapter_node = $adapter->get_node();
            $self->sanlog(msg => "Adpter Node ==" . $adapter_node);
            $self->sanlog(msg => "Filer Name == " . $filer_name);
            if ($adapter_node eq $filer_name) {
                $self->sanlog(
                         msg => "Adding adapter == " . $adapter->get_adapter());
                push(@adapters, $adapter);
            }
        } ## end foreach my $adapter (@all_adapters)

        # Put the adapter (in this case ports) information in the fcp_info hash
        $fcp_info->{$cluster_name}->{$filer_name}->{'num_port_objs'} =
          scalar(@adapters);
        $fcp_info->{$cluster_name}->{$filer_name}->{'port_objs'} = \@adapters;

        foreach my $adapter (@adapters) {

            my $adapter_name = $adapter->get_adapter();
            $self->sanlog(msg => "Adapter name == " . $adapter_name);

            my $lifs_per_adapter =
              $self->get_lifs_for_given_port(-filer_obj => $filer,
                                             -adapter   => $adapter_name);

            if ($self->error) {
                $self->sanlog(msg => $self->errmsg());
                $self->_set_err(
                    "Failed to get the fcp lif for given adapter from the filer"
                    );
                return $SanUtil::FALSE;
            }

            if (scalar(@{$lifs_per_adapter})) {
                foreach my $lif (@{$lifs_per_adapter}) {
                    my $lif_name = $lif->get_name();
                    $self->sanlog(msg => $lif_name);
                }
                $fcp_info->{$cluster_name}->{$filer_name}->{$adapter_name}
                  ->{'num_lif_objs'} = scalar(@{$lifs_per_adapter});
                $fcp_info->{$cluster_name}->{$filer_name}->{$adapter_name}
                  ->{'lif_objs'} = $lifs_per_adapter;
            } ## end if (scalar(@{$lifs_per_adapter...
        } ## end foreach my $adapter (@adapters)
    } ## end foreach my $filer (@filer_objs)

    return $fcp_info;
} ## end sub collect_fcp_info

###############################################################################
# Method:     print_fcp_info
# Objective:  This method prints the contents of an fcp hash as created by
#             collect_fcp_info
# Parameters:
#             fcp_info - Address of a reference to a hash as created by collect_fcp_info
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub print_fcp_info {
    my ($self, %args) = @_;
    my $fcp_info = {};

    $self->_clear_err();

    $self->sanlog(msg => "In print_fcp_info");

    while (my ($key, $value) = each(%args)) {
        $self->sanlog(msg => $key);
        if ($key =~ /fcp_info/i) {
            unless (defined($$value)) {
                $self->_set_err("Missing value for key 'fcp_info'");
                return $SanUtil::FALSE;
            }
            $fcp_info = $$value;
        } else {
            $self->_set_err("Invalid key $key");
            return $SanUtil::FALSE;
        }
    } ## end while (my ($key, $value) ...

    # Level 1: Getting cluster information from fcp_info
    $self->sanlog(msg => "CLUSTER: ");
    my $cluster      = $fcp_info->{'cluster_obj'};
    my $cluster_name = $cluster->get_name();
    $self->sanlog(msg => $cluster_name);

    # Level 2: Getting filer information from the cluster level
    $self->sanlog(msg => "FILERS: ");
    my $num_filers = $fcp_info->{$cluster_name}->{'num_filer_objs'};
    my $filers     = $fcp_info->{$cluster_name}->{'filer_objs'};

    for (my $i = 0 ; $i < $num_filers ; $i++) {
        my $filer       = $$filers[$i];
        my $filers_name = $filer->get_name();
        $self->sanlog(msg => " " . $filers_name);

        # Level 3: Getting port information from the filer level
        $self->sanlog(msg => "\tPORTS: ");
        my $ports = $fcp_info->{$cluster_name}->{$filers_name}->{'port_objs'};
        my $num_ports =
          $fcp_info->{$cluster_name}->{$filers_name}->{'num_port_objs'};

        for (my $j = 0 ; $j < $num_ports ; $j++) {
            my $port       = $$ports[$j];
            my $ports_name = $port->get_adapter();
            $self->sanlog(msg => "\t " . $ports_name);

            # Level 4: Getting vserver information from the port level
            $self->sanlog(msg => "\t\tVSERVERS: ");
            my $vservers =
              $fcp_info->{$cluster_name}->{$filers_name}->{$ports_name}
              ->{'vs_objs'};
            my $num_vservers =
              $fcp_info->{$cluster_name}->{$filers_name}->{$ports_name}
              ->{'num_vs_objs'};

            for (my $k = 0 ; $k < $num_vservers ; $k++) {
                my $vserver       = $$vservers[$k];
                my $vservers_name = $vserver->get_name();
                $self->sanlog(msg => "\t\t " . $vservers_name);

                # Level 5: Getting lif information from the vserver level
                $self->sanlog(msg => "\t\t\tLIFS");

                #$Test->nlog($all_cluster_name . $all_filers_name . $all_ports_name . $all_vservers_name);
                my $lifs =
                  $fcp_info->{$cluster_name}->{$filers_name}->{$ports_name}
                  ->{$vservers_name}->{'lif_objs'};
                my $num_lifs =
                  $fcp_info->{$cluster_name}->{$filers_name}->{$ports_name}
                  ->{$vservers_name}->{'num_lif_objs'};

                #$self->sanlog (msg => "Number of lifs = " . $num_lifs);
                foreach my $lif (@{$lifs}) {
                    $self->sanlog(msg => "\t\t\t" . $lif->get_name());
                }
            } ## end for (my $k = 0 ; $k < $num_vservers...
        } ## end for (my $j = 0 ; $j < $num_ports...
    } ## end for (my $i = 0 ; $i < $num_filers...
    return $SanUtil::TRUE;
} ## end sub print_fcp_info

###############################################################################
# Method:     get_objects_from_fcp_info
# Objective:  This method prints the contents of an fcp hash as created by
#             collect_fcp_info
# Parameters:
#             fcp_info - Address of a reference to a hash as created by collect_fcp_info
# Returns:    SanUtil::FALSE if something failed
#             SanUtil::TRUE is everything went well
###############################################################################

sub get_objects_from_fcp_info {

    my ($self, %args) = @_;
    my ($level, $cluster_name, $filer_name, $port_name, $vserver_name);
    my ($lif_name, $fcp_info, $return_hash);

    $self->_clear_err();

    $self->sanlog(msg => "In get_objects_from_fcp_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /level/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'level'");
                return {};
            }
            $level = $value;
        } elsif ($key =~ /cluster_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'cluster_name'");
                return {};
            }
            $cluster_name = $value;
        } elsif ($key =~ /filer_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'filer_name'");
                return {};
            }
            $filer_name = $value;
        } elsif ($key =~ /port_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'port_name'");
                return {};
            }
            $port_name = $value;
        } elsif ($key =~ /vserver_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver_name'");
                return {};
            }
            $vserver_name = $value;
        } elsif ($key =~ /lif_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'lif_name'");
                return {};
            }
            $lif_name = $value;
        } elsif ($key =~ /fcp_info/i) {
            unless (length($$value)) {
                $self->_set_err("Missing value for key 'fcp_info'");
                return {};
            }
            $fcp_info = $$value;
        } else {
            $self->_set_err("Invalid key $key");
            return {};
        }
    } ## end while (my ($key, $value) ...

    unless (    defined($level)
            and defined($fcp_info)) {
        $self->_set_err("Check parameters, missing key-value pair");
        return {};
    }

    if ($level =~ /filer/i) {
        my $filer_info = {};
        if (defined($filer_name)) {

            # Got the filer_name, return a hash with just the required filer
            $filer_info =
              $self->_get_filer_info_from_fcp_info(fcp_info   => $fcp_info,
                                                   filer_name => $filer_name);
        } else {

            # Return all the filers. This is the portion of the hash pointed
            # to by the cluster name.
            $filer_info =
              $self->_get_filer_info_from_fcp_info(fcp_info => $fcp_info);
        }
        return $filer_info;

    } elsif ($level =~ /port/i) {
        unless (defined($filer_name))
        {    # Need the filer name to know which port to get
            $self->_set_err(
                        "Asked for the port, but have not specified the filer");
            return {};
        }

        # Call function to get the filer info
        my $filer_info =
          $self->_get_filer_info_from_fcp_info(fcp_info   => $fcp_info,
                                               filer_name => $filer_name);

        if ($self->error()) {
            return {};
        }
        my $port_info = {};
        if (defined($port_name)) {

            # Call function to get the specific port
            $port_info =
              $self->_get_port_info_from_filer_info(filer_info => $filer_info,
                                                    port_name  => $port_name);
        } else {

            # Return all the ports
            $port_info =
              $self->_get_port_info_from_filer_info(filer_info => $filer_info);
        }
        return $port_info;

    } elsif ($level =~ /vserver/i) {
        unless (defined($filer_name) and defined($port_name)) {
            $self->_set_err("Asked for the vserver, but have not specified the "
                            . "filer and port");
            return {};
        }

        # Call function to get the filer info
        my $filer_info =
          $self->_get_filer_info_from_fcp_info(fcp_info   => $fcp_info,
                                               filer_name => $filer_name);
        if ($self->error()) {
            return {};
        }

        # Call function to get the port
        my $port_info =
          $self->_get_port_info_from_filer_info(filer_info => $filer_info,
                                                port_name  => $port_name);
        if ($self->error()) {
            return {};
        }
        my $vs_info = {};
        if (defined($vserver_name)) {

            # Call function to get the specific vserver
            $vs_info =
              $self->_get_vs_info_from_port_info(port_info    => $port_info,
                                                 vserver_name => $vserver_name);
        } else {

            # Return all the vservers
            $vs_info =
              $self->_get_vs_info_from_port_info(port_info => $port_info);
        }
        return $vs_info;

    } elsif ($level =~ /lif/i) {
        unless (    defined($filer_name)
                and defined($port_name)
                and defined($vserver_name)) {
            $self->_set_err(  "Asked for the lif, but have not specified the "
                            . "filer, port and vserver");
            return {};
        }

        # Call function to get the filer info
        my $filer_info =
          $self->_get_filer_info_from_fcp_info(fcp_info   => $fcp_info,
                                               filer_name => $filer_name);
        if ($self->error()) {
            return {};
        }

        # Call function to get the port
        my $port_info =
          $self->_get_port_info_from_filer_info(filer_info => $filer_info,
                                                port_name  => $port_name);
        if ($self->error()) {
            return {};
        }

        # Call function to get the vserver
        my $vs_info =
          $self->_get_vs_info_from_port_info(port_info    => $port_info,
                                             vserver_name => $vserver_name);
        if ($self->error()) {
            return {};
        }
        my $lif_info = {};
        if (defined($lif_name)) {

            # Call function to get the specific lif
            $lif_info =
              $self->_get_lif_info_from_vs_info(vs_info  => $vs_info,
                                                lif_name => $lif_name);
        } else {

            # Return all lifs
            $lif_info = $self->_get_lif_info_from_vs_info(vs_info => $vs_info);
        }
        return $lif_info;
    } else {
        $self->_set_err("Invalid value of level");
        return {};
    }
    return {};
} ## end sub get_objects_from_fcp_info

###############################################################################
# P R I V A T E   M E T H O D S
###############################################################################

sub _create_io_obj {

    my ($self, %args) = @_;
    my ($client, $read_pct, $rand);

    my $io_blk_size;
    my $io_offset_begin;
    my $io_filesize;
    my $io_pattern;
    my $io_num_threads;
    my $wait_flag;
    my @devices = ();
    my $IO_obj;

    $self->_clear_err();

    # Not validating parameters as it should already have been validated
    # in calling function
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /client_obj/i) {
            $client = $value;
        } elsif ($key =~ /device_objs/i) {
            @devices = @{$value};
        } elsif ($key =~ /read_pct/i) {
            $read_pct = $value;
        } elsif ($key =~ /rand/i) {
            $rand = $value;
        } elsif ($key =~ /block_size/i) {
            $io_blk_size = $value;
        } elsif ($key =~ /offset_begin/i) {
            $io_offset_begin = $value;
        } elsif ($key =~ /file_size/i) {
            $io_filesize = $value;
        } elsif ($key =~ /io_pattern/i) {
            $io_pattern = $value;
        } elsif ($key =~ /num_threads/i) {
            $io_num_threads = $value;
        } elsif ($key =~ /wait_flag/i) {
            $wait_flag = $value;
        }
    } ## end while (my ($key, $value) ...

    # Create the right string format to pass to sanio
    my $io_devs = ();
    foreach my $device (@devices) {
        $io_devs .= " " . $device->{diskObj}->{rawName};
    }

    $self->sanlog(msg => "Device string " . $io_devs);
    $self->sanlog(msg => "Lun size : " . $io_filesize);
    $self->sanlog(msg => "Device string : " . $io_devs);

    $IO_obj = new SanIo(
                       -io_tool        => "sio",
                       -sio_subtest_id => "sio",
                       -san_client_obj => $client,
                       -wait_flag      => $wait_flag,
                       -client_io_dev  => $io_devs,
                       -read_pct       => $read_pct,
                       -random_pct     => $rand,
                       -block_size     => $io_blk_size,
                       -first_byte     => $io_offset_begin,
                       -last_byte      => $io_filesize,
                       -run_time       => 0,
                       -num_threads    => $io_num_threads,
                       -pass_thru => "-verify -fillonce -fixedfill -$io_pattern"
                       );

    if ($IO_obj->error()) {
        $self->_set_err("Failed to create the SanIO object");
    }
    return $IO_obj;
} ## end sub _create_io_obj

###############################################################################
# Method:     _get_filer_info_from_fcp_info
# Objective:  This method extracts the filer information from the fcp_info
#             hash as created by collect_fcp_info
# Parameters:
#             fcp_info - Reference to a hash as created by collect_fcp_info
#             filer_name - This is an optional parameter. If information about
#             a specific filer is required, this parameter can be used to pass
#             a filer name
# Returns:    An undefined reference if there was an error
#             Reference to the filer information if no error occurred
###############################################################################

sub _get_filer_info_from_fcp_info {
    my ($self, %args) = @_;
    my ($fcp_info, $filer_name);
    my $filer_info = {};

    $self->_clear_err();

    $self->sanlog(msg => "In _get_filer_info_from_fcp_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /fcp_info/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'fcp_info'");
                return {};
            }
            $fcp_info = $value;
        } elsif ($key =~ /filer_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'filer_name'");
                return {};
            }
            $filer_name = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return {};
        }
    } ## end while (my ($key, $value) ...

    my $cluster_obj  = $fcp_info->{'cluster_obj'};
    my $cluster_name = $cluster_obj->get_name();
    my $i;
    if (defined($filer_name)) {    # Want specific filer info
        my $filer_objs     = $fcp_info->{$cluster_name}->{'filer_objs'};
        my $num_filer_objs = $fcp_info->{$cluster_name}->{'num_filer_objs'};
        for ($i = 0 ; $i < $num_filer_objs ; $i++) {
            my $filer = $$filer_objs[$i];
            my $name  = $filer->get_name();
            if ($name eq $filer_name) {
                $filer_info->{'filer_obj'} = $filer;
                $filer_info->{$name} = $fcp_info->{$cluster_name}->{$name};
                return $filer_info;

                #print Dumper ($filer_info);
                #return;
            }
        } ## end for ($i = 0 ; $i < $num_filer_objs...
        if ($i == $num_filer_objs) {
            $self->_set_err(
                        "Filer with specified name does not exist in fcp_info");
            return {};
        }
    } else {    # Returning hash of all filers
        $filer_info = $fcp_info->{$cluster_name};
        return $filer_info;

        #print Dumper ($filer_info);
        #return;
    }
    return {};
} ## end sub _get_filer_info_from_fcp_info

###############################################################################
# Method:     _get_port_info_from_fcp_info
# Objective:  This method extracts the port information from the fcp_info
#             hash as created by collect_fcp_info
# Parameters:
#             filer_info - Reference to a hash as created by _get_filer_info_from_fcp_info.
#             (_get_filer_info_from_fcp_info should have been called with filer_name)
#             port_name - This is an optional parameter. If information about
#             a specific port on a specific filer is required, this parameter
#             can be used to pass a port name
# Returns:    An undefined reference if there was an error
#             Reference to the filer information if no error occurred
###############################################################################

sub _get_port_info_from_filer_info {
    my ($self, %args) = @_;
    my ($filer_info, $port_name);
    my $port_info = {};
    my $fcp_info;

    $self->_clear_err();

    $self->sanlog(msg => "In _get_port_info_from_filer_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /filer_info/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'filer_info'");
                return {};
            }
            $filer_info = $value;
        } elsif ($key =~ /port_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'port_name'");
                return {};
            }
            $port_name = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return {};
        }
    } ## end while (my ($key, $value) ...

    my $filer_obj  = $filer_info->{'filer_obj'};
    my $filer_name = $filer_obj->get_name();
    my $i;
    $self->sanlog(msg => "Port name = " . $port_name);
    if (defined($port_name)) {    # Want specific port info
        my $port_objs     = $filer_info->{$filer_name}->{'port_objs'};
        my $num_port_objs = $filer_info->{$filer_name}->{'num_port_objs'};
        for ($i = 0 ; $i < $num_port_objs ; $i++) {
            my $port = $$port_objs[$i];
            my $name = $port->get_adapter();
            if ($name eq $port_name) {
                $self->sanlog(
                        msg => "name = " . $name . "port_name = " . $port_name);
                $port_info->{'port_obj'} = $port;
                $port_info->{$name} = $filer_info->{$filer_name}->{$name};
                return $port_info;
            }
        } ## end for ($i = 0 ; $i < $num_port_objs...
        if ($i == $num_port_objs) {
            $self->_set_err(
                  "Port with specified name does not exist on specified filer");
            return {};
        }
    } else {    # Returning hash of all ports
        $port_info = $filer_info->{$filer_name};
        return $port_info;
    }
    return {};
} ## end sub _get_port_info_from_filer_info

###############################################################################
# Method:     _get_vs_info_from_fcp_info
# Objective:  This method extracts the vserver information from the fcp_info
#             hash as created by collect_fcp_info
# Parameters:
#             port_info - Reference to a hash as created by _get_port_info_from_fcp_info.
#             (_get_port_info_from_fcp_info should have been called with port_name)
#             vserver_name - This is an optional parameter. If information about
#             a specific vserver with lifs on a specific port which is on a specific
#             filer is required, this parameter can be used to pass a vserver name
# Returns:    An undefined reference if there was an error
#             Reference to the filer information if no error occurred
###############################################################################

sub _get_vs_info_from_port_info {
    my ($self, %args) = @_;
    my ($port_info, $vs_name);
    my $vs_info = {};

    $self->_clear_err();

    $self->sanlog(msg => "In _get_vs_info_from_port_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /port_info/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'port_info'");
                return {};
            }
            $port_info = $value;
        } elsif ($key =~ /vserver_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vserver_name'");
                return {};
            }
            $vs_name = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return {};
        }
    } ## end while (my ($key, $value) ...

    my $port_obj  = $port_info->{'port_obj'};
    my $port_name = $port_obj->get_adapter();
    my $i;
    if (defined($vs_name)) {    # Want specific vserver info
        my $vs_objs     = $port_info->{$port_name}->{'vs_objs'};
        my $num_vs_objs = $port_info->{$port_name}->{'num_vs_objs'};
        for ($i = 0 ; $i < $num_vs_objs ; $i++) {
            my $vs   = $$vs_objs[$i];
            my $name = $vs->get_name();
            if ($name eq $vs_name) {
                $vs_info->{'vs_obj'} = $vs;
                $vs_info->{$name} = $port_info->{$port_name}->{$name};
                return $vs_info;
            }
        }
        if ($i == $num_vs_objs) {
            $self->_set_err(
                "Vserver with specified name does not exist on specified port");
            return {};
        }
    } else {    # Returning hash of all vservers
        $vs_info = $port_info->{$port_name};
        return $vs_info;
    }
    return {};
} ## end sub _get_vs_info_from_port_info

###############################################################################
# Method:     _get_lif_info_from_fcp_info
# Objective:  This method extracts the lif information from the fcp_info
#             hash as created by collect_fcp_info
# Parameters:
#             vs_info - Reference to a hash as created by _get_vs_info_from_fcp_info.
#             (_get_vs_info_from_fcp_info should have been called with vserver_name)
#             lif_name - This is an optional parameter. If information about
#             a specific lif on a specific vserver with lifs on a specific port
#             which is on a specific filer is required, this parameter can be
#             used to pass a lif name
# Returns:    An undefined reference if there was an error
#             Reference to the filer information if no error occurred
###############################################################################

sub _get_lif_info_from_vs_info {
    my ($self, %args) = @_;
    my ($vs_info, $lif_name);
    my $lif_info = {};

    $self->_clear_err();

    $self->sanlog(msg => "In _get_lif_info_from_vs_info");

    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vs_info/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'vs_info'");
                return {};
            }
            $vs_info = $value;
        } elsif ($key =~ /lif_name/i) {
            unless (length($value)) {
                $self->_set_err("Missing value for key 'lif_name'");
                return {};
            }
            $lif_name = $value;
        } else {
            $self->_set_err("Invalid key $key");
            return {};
        }
    } ## end while (my ($key, $value) ...

    my $vs_obj  = $vs_info->{'vs_obj'};
    my $vs_name = $vs_obj->get_name();
    my $i;
    if (defined($lif_name)) {    # Want specific lif info
        my $lif_objs     = $vs_info->{$vs_name}->{'lif_objs'};
        my $num_lif_objs = $vs_info->{$vs_name}->{'num_lif_objs'};
        for ($i = 0 ; $i < $num_lif_objs ; $i++) {
            my $lif  = $$lif_objs[$i];
            my $name = $lif->get_name();
            if ($name eq $lif_name) {
                $lif_info->{'lif_obj'} = $lif;

                #$vs_info->{$name} = $port_info->{$port_name}->{$name};
                return $lif_info;
            }
        } ## end for ($i = 0 ; $i < $num_lif_objs...
        if ($i == $num_lif_objs) {
            $self->_set_err(
                 "Lif with specified name does not exist on specified vserver");
            return {};
        }
    } else {    # Returning hash of all lifs
        $lif_info = $vs_info->{$vs_name};
        return $lif_info;
    }
    return {};
} ## end sub _get_lif_info_from_vs_info

################################################################################
#
# Function  - create_lifs
#             create FCP LIFs and verify them on filer
#
# Input     - vserver_objs_ref required
#                              (reference to array of vserver objects
#                              which will be used for LIF verification)
#             vs               required
#                              (vserver name which LIF will be craeted on)
#             node             required (node name)
#             port             required (port number)
#             number           required (number of LIFs need to create)
#             index            required (starting index of LIF)
#             name_prefix      required (LIF name prefix)
#             status_admin     optional (expected LIF status, default is 'up')
#             skip_fcp_check   optional (default value is '0')
# Output    - Reference to array of newly created FCP LIFs
#             SanUtil::FAIL
#
################################################################################
sub create_lifs {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $lif_vs;
    my $lif_node;
    my $lif_port;
    my $lif_number;
    my $lif_index;
    my $lif_name_prefix;
    my $status_admin   = "up";
    my $skip_fcp_check = 0;
    my $cmd_interface  = 'cli';
    my $lif_prefix;
    my @new_lif_objs = ();
    my $result;
    my $log_msg = "FcpHelper::create_lifs: ";

    $self->sanlog(-msg => "FcpHelper::create_lifs - Create LIFs");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /vs/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_vs = $value;
        } elsif ($key =~ /node/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for node: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_node = $value;
        } elsif ($key =~ /port/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for port: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_port = $value;
        } elsif ($key =~ /number/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for number: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_number = $value;
        } elsif ($key =~ /index/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for index: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_index = $value;
        } elsif ($key =~ /name_prefix/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for name_prefix: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_name_prefix = $value;
        } elsif ($key =~ /status_admin/) {
            $status_admin = $value;
        } elsif ($key =~ /skip_fcp_check/) {
            $skip_fcp_check = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_vs)) {
        $self->_set_err("Parameter 'lif_vs' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_node)) {
        $self->_set_err("Parameter 'lif_node' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_port)) {
        $self->_set_err("Parameter 'lif_port' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_number)) {
        $self->_set_err("Parameter 'lif_number' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_index)) {
        $self->_set_err("Parameter 'lif_index' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_name_prefix)) {
        $self->_set_err("Parameter 'lif_name_prefix' not specified");
        return $SanUtil::FAIL;
    }

    for (my $i = $lif_index ; $i < ($lif_index + $lif_number) ; $i++) {
        my $lif_name;

        # This is special case using -1 as lif index to pass in
        # because of the name convention used by FitMgr
        # when only 1 lif created on physical port.
        # There is no number index appended in the lif name.
        # Same rule has to be followed during restoring the test setup for
        # the next test case if error happens in the current test case.
        if ($lif_index == -1) {
            $lif_name = $lif_name_prefix;
        } else {
            $lif_name = $lif_name_prefix . $i;
        }
        my $new_lif_obj =
          new San::ControllerCluster::Filer10M::Vserver::Lif(
                                                  filer           => $lif_node,
                                                  name            => $lif_name,
                                                  'home-node'     => $lif_node,
                                                  interface       => $lif_port,
                                                  role            => "data",
                                                  'data-protocol' => "fcp",
                                                  status_admin => $status_admin,
                                                  vserver      => $lif_vs);

        if ($new_lif_obj->error()) {
            $self->_set_err(  "Failed to create FCP LIF "
                            . $lif_name
                            . " on phyical port "
                            . $lif_port
                            . " of node "
                            . $lif_node
                            . " on vserver "
                            . $lif_vs);
            return $SanUtil::FAIL;
        } ## end if ($new_lif_obj->error...
        $self->sanlog(  -msg => $log_msg
                      . "New FCP LIF "
                      . $lif_name
                      . " is created on phyical port "
                      . $lif_port
                      . " of node "
                      . $lif_node
                      . " on vserver "
                      . $lif_vs);
        push(@new_lif_objs, $new_lif_obj);
    } ## end for (my $i = $lif_index...

    if ($lif_number > MAXIMUM_FCP_LIF_SUPPORTED_PER_PORT_BCOM) {
        &Tharn::sleep(15);
    }
    else {
        &Tharn::sleep(2);  
    }

    # Verify LIFs on filer
    $result = $self->verify_lif_on_filer(vserver_objs_ref => $vserver_objs_ref,
                                         lif_objs_ref     => \@new_lif_objs,
                                         operation        => "create",
                                         skip_fcp_check   => $skip_fcp_check,
                                         cmd_interface    => $cmd_interface);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err("Failed to verify LIFs on filer", $self->errmsg());
        return $SanUtil::FAIL;
    }

    return \@new_lif_objs;
} ## end sub create_lifs

################################################################################
#
# Function  - delete_lifs
#             delete FCP LIFs and verify lifs on filer
#
# Input     - vserver_objs_ref  required
#             (reference to array of vserver objects which will be used
#             for LIF verification)
#             lif_objs_ref      required
#             (reference to array of LIF objects)
#             skip_fcp_check    optional (default value is '0')
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub delete_lifs {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $lif_objs_ref;
    my $skip_fcp_check = 0;
    my $cmd_interface  = 'cli';
    my $lif_name;
    my $lif_node;
    my $lif_port;
    my $lif_vs;
    my @lif_names = ();
    my $result;
    my $log_msg = "FcpHelper::delete_lifs: ";

    $self->sanlog(-msg => "FcpHelper::delete_lifs - Delete LIFs");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /skip_fcp_check/) {
            $skip_fcp_check = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }

    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    foreach my $lif_obj (@$lif_objs_ref) {
        $lif_name = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
        push(@lif_names, $lif_name);

        $lif_node = $lif_obj->get_curr_node();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        $lif_port = $lif_obj->get_curr_port();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        $lif_vs = $lif_obj->get_vserver();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        $lif_obj->delete();
        if ($lif_obj->error()) {
            $self->_set_err("Failed to delete LIF " . $lif_name);
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "FCP LIF "
                      . $lif_name
                      . " is deleted on phyical port "
                      . $lif_port
                      . " of node "
                      . $lif_node
                      . " on vserver "
                      . $lif_vs);
    } ## end foreach my $lif_obj (@$lif_objs_ref)

    # Verify LIFs on filer
    $result = $self->verify_lif_on_filer(vserver_objs_ref => $vserver_objs_ref,
                                         lif_objs_ref     => \@lif_names,
                                         operation        => "delete",
                                         skip_fcp_check   => $skip_fcp_check,
                                         cmd_interface    => $cmd_interface);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err("Failed to verify LIFs on filer", $self->errmsg());
        return $SanUtil::FAIL;
    }

    return $SanUtil::PASS;
} ## end sub delete_lifs

################################################################################
#
# Function  - modify_lifs_status
#             modify FCP LIFs status and verify LIFs status on filer
#
# Input     - lif_objs_ref      required (reference to array of LIF objects)
#             status_admin      required (expected LIF status after modifying)
#             fcp_adapter_state optional (fcp adapter state)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub modify_lifs_status {
    my ($self, %args) = @_;
    my $lif_objs_ref;
    my $status_admin;
    my $status_oper;
    my $fcp_adapter_state  = "up";
    my $fcp_service_status = "up";
    my $result;
    my $log_msg = "FcpHelper::modify_lifs_status: ";

    $self->sanlog(-msg => "FcpHelper::modify_lifs_status - "
                  . "Modify LIFs status and Verify");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /status_admin/) {
            if (not length($value) and $value !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for status_admin: " . $value);
                return $SanUtil::FAIL;
            }
            $status_admin = $value;
        } elsif ($key =~ /fcp_adapter_state/) {
            $fcp_adapter_state = $value;
        } elsif ($key =~ /fcp_service_status/) {
            $fcp_service_status = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($status_admin)) {
        $self->_set_err("Parameter 'status_admin' not specified");
        return $SanUtil::FAIL;
    }

    foreach my $lif_obj (@$lif_objs_ref) {
        my $lif_name = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        $lif_obj->modify(-status_admin => $status_admin);
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
    } ## end foreach my $lif_obj (@$lif_objs_ref)

    # Get expected operation status of LIFs
    if ($fcp_adapter_state eq "up") {
        if ($fcp_service_status eq "up" && $status_admin eq "up") {
            $status_oper = "up";
        } else {
            $status_oper = "down";
        }
    } else {
        $status_oper = "down";
    }

    $result = $self->verify_lif_status(lif_objs_ref => $lif_objs_ref,
                                       status_admin => $status_admin,
                                       status_oper  => $status_oper);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err("Failed to verify LIFs status", $self->errmsg());
        return $SanUtil::FAIL;
    }

    $self->sanlog(
           -msg => "Succeed to modify LIFs status " . "and verify LIFs status");

    return $SanUtil::PASS;
} ## end sub modify_lifs_status

################################################################################
#
# Function  - modify_fcp_adapter_state
#             modify FCP adapter state and verify it
#
# Input     - filer_obj      required
#             adapter        required  (adapter port reside on filer)
#             state          required  (expected state of FCP adapter after
#                                       modifying)
#             modify_success required
#             (Since test case know whether the modification is success or not
#              due to LIF status on port, pass this paramter for verification
#              purpose, possible values are 1/0)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub modify_fcp_adapter_state {
    my ($self, %args) = @_;
    my $filer_obj;
    my $adapter;
    my $state;
    my $modify_success;
    my $cmd_interface = 'cli';
    my $result;
    my $log_msg = "FcpHelper::modify_fcp_adapter_state: ";

    $self->sanlog(-msg => "FcpHelper::modify_fcp_adapter_state - "
                  . "Modify FCP Adapter State and Verify");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /filer_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for filer_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $filer_obj = $value;
        }
        if ($key =~ /adapter/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for adapter: " . $value);
                return $SanUtil::FAIL;
            }
            $adapter = $value;
        } elsif ($key =~ /state/) {
            if (not length($value) and $value !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for state: " . $value);
                return $SanUtil::FAIL;
            }
            $state = $value;
        } elsif ($key =~ /modify_success/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for modify_success: " . $value);
                return $SanUtil::FAIL;
            }
            $modify_success = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($filer_obj)) {
        $self->_set_err("Parameter 'filer_objs' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($adapter)) {
        $self->_set_err("Parameter 'adapter' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($state)) {
        $self->_set_err("Parameter 'state' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($modify_success)) {
        $self->_set_err("Parameter 'modify_success' not specified");
        return $SanUtil::FAIL;
    }

    my $filer_name = $filer_obj->get_name();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my $hw_obj = $filer_obj->get_hwresource_obj();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my @fcpadapter_objs = $hw_obj->get_fcpadapter_objs();
    if ($hw_obj->error()) {
        $self->_set_err($hw_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $fcpadapter_obj (@fcpadapter_objs) {
        my $fcpadapter_node = $fcpadapter_obj->get_node();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $fcpadapter_adapter = $fcpadapter_obj->get_adapter();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if (   ($fcpadapter_node eq $filer_name)
            && ($fcpadapter_adapter eq $adapter)) {

            # FCP adapter original state
            my $fcpadapter_state = $fcpadapter_obj->get_state();
            if ($fcpadapter_obj->error()) {
                $self->_set_err($fcpadapter_obj->errmsg());
                return $SanUtil::FAIL;
            }

            $result =
              $fcpadapter_obj->modify(node    => $filer_obj,
                                      adapter => $adapter,
                                      state   => $state);
            if (($result eq $SanUtil::FALSE) || $fcpadapter_obj->error()) {
                if ($modify_success == 0) {
                    $self->sanlog(-msg => $log_msg
                                . "Expected Error: Failed to modify FCP adapter"
                                . "state since LIFs are currently active");
                } else {
                    $self->_set_err("Failed to modify FCP adapter state",
                                    $fcpadapter_obj->errmsg());
                    return $SanUtil::FAIL;
                }
            } ## end if (($result eq $SanUtil::FALSE...

            &Tharn::sleep(15);

            my $expected_adapter_state = "";
            if ($modify_success == 0) {
                $expected_adapter_state = $fcpadapter_state;
            } else {
                $expected_adapter_state = $state;
            }
            $result =
              $self->verify_fcp_adapter_state(filer_obj => $filer_obj,
                                              adapter   => $adapter,
                                              state => $expected_adapter_state,
                                              cmd_interface => $cmd_interface);
            if ($result eq $SanUtil::FAIL || $self->error()) {
                $self->_set_err("Failed to verify FCP adapter " 
                                  . $adapter
                                  . " on node "
                                  . $filer_name
                                  . " state",
                                $self->errmsg());
                return $SanUtil::FAIL;
            }
        } ## end if (($fcpadapter_node ...
    } ## end foreach my $fcpadapter_obj ...

    if ($modify_success == 0) {
        $self->sanlog(  -msg => "Failed to modify FCP adapter " 
                      . $adapter
                      . " state on node "
                      . $filer_obj->get_name()
                      . " as expected and verify adapter state");
    } else {
        $self->sanlog(  -msg => "Succeed to modify FCP adapter " 
                      . $adapter
                      . " state on node "
                      . $filer_obj->get_name()
                      . " and verify adapter state");
    }

    return $SanUtil::PASS;
} ## end sub modify_fcp_adapter_state

################################################################################
#
# Function  - modify_fcp_adapter_state_for_given_lifs
#             modify FCP adapter state for given LIFs and verify it
#
# Input     - filer_objs_ref required (reference to array of filer objects)
#             lif_objs_ref   required (reference to array of LIF objects)
#             state          required (expected state of FCP adapter after
#                                      modifying)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub modify_fcp_adapter_state_for_given_lifs {
    my ($self, %args) = @_;
    my $filer_objs_ref;
    my $lif_objs_ref;
    my $state;
    my $cmd_interface = 'cli';
    my $result;
    my $log_msg = "FcpHelper::modify_fcp_adapter_state: ";

    $self->sanlog(-msg => "FcpHelper::modify_fcp_adapter_state - "
                  . "Modify FCP Adapter State and Verify");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /filer_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for filer_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $filer_objs_ref = $value;
        }
        if ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /state/) {
            if (not length($value) and $value !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for state: " . $value);
                return $SanUtil::FAIL;
            }
            $state = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($filer_objs_ref)) {
        $self->_set_err("Parameter 'filer_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($state)) {
        $self->_set_err("Parameter 'state' not specified");
        return $SanUtil::FAIL;
    }

    # Hash %adapters
    # (key: fcp adapter obj; values: a hash %adapter_info)
    my %adapters;

    # Hash %adapter_info
    # (keys:   'fcpadapter_obj'/'node'/'adapter';
    #  values: fcpadapter_obj/filer_obj/adapter port)
    my %adapter_info;

    my $lif_curr_node;
    my $lif_curr_port;
    foreach my $lif_obj (@$lif_objs_ref) {
        $lif_curr_node = $lif_obj->get_curr_node();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        $lif_curr_port = $lif_obj->get_curr_port();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Given any filer object from the cluster can give you all
        # the fcp physical adapter objects (i.e. objects for any fcp
        # adapter exsiting on any node in the cluster).
        # But we have to figure out the home-node controller for the
        # given physical fcp adapter so that we can use it in
        # fcpadapter_obj modify call.
        $result = $self->_get_lif_filer_obj($filer_objs_ref, $lif_obj);
        if ($result eq $SanUtil::FAIL || $self->error()) {
            return $SanUtil::FAIL;
        }
        my $lif_filer_obj = $result;

        $result = $self->_get_fcpadapter_objs($lif_filer_obj, $lif_curr_port);
        if ($result eq $SanUtil::FAIL || $self->error()) {
            return $SanUtil::FAIL;
        }
        my @fcpadapter_objs = @$result;
        my $fcpadapter_obj  = $fcpadapter_objs[0];

        my $adapter_node = $fcpadapter_obj->get_node();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $adapter_port = $fcpadapter_obj->get_adapter();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($adapters{$fcpadapter_obj}) {
            %adapter_info = %{$adapters{$adapter_node . '_' . $adapter_port}};
        }
        $adapter_info{'fcpadapter_obj'} = $fcpadapter_obj;
        $adapter_info{'node'}           = $lif_filer_obj;
        $adapter_info{'port'}           = $adapter_port;
        %{$adapters{$fcpadapter_obj}} = %adapter_info;
    } ## end foreach my $lif_obj (@$lif_objs_ref)

    for (keys(%adapters)) {
        my %adapter_info   = %{$adapters{$_}};
        my $fcpadapter_obj = $adapter_info{'fcpadapter_obj'};
        my $filer_obj      = $adapter_info{'node'};
        my $adapter_port   = $adapter_info{'port'};

        # Get adapter original status
        my $adapter_orig_status = $fcpadapter_obj->get_status();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Refresh adapter object
        $fcpadapter_obj->refresh();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $adapter_state;
        if ($cmd_interface eq 'cli') {

            # Get adapter state
            $adapter_state = $fcpadapter_obj->get_state();
            if ($fcpadapter_obj->error()) {
                $self->_set_err($fcpadapter_obj->errmsg());
                return $SanUtil::FAIL;
            }
        } else {
            if ($adapter_orig_status eq 'online') {
                $adapter_state = 'up';
            } else {
                $adapter_state = 'down';
            }
        }

        # Skip modification if adapter state has already been modified
        if ($adapter_state ne $state) {

            # Get LIFs operation status information before modifying
            # adapter state. This is used for verification purpose
            my $lifs_oper_status =
              $self->_get_lifs_oper_status($filer_obj, $adapter_port,
                                           $lif_objs_ref);
            if ($lifs_oper_status eq $SanUtil::FAIL || $self->error()) {
                $self->_set_err("Failed to get LIF status", $self->errmsg());
                return $SanUtil::FAIL;
            }

            my $result =
              $fcpadapter_obj->modify(node    => $filer_obj,
                                      adapter => $adapter_port,
                                      state   => $state);

            if (($result eq $SanUtil::FALSE) || $fcpadapter_obj->error()) {
                if ($lifs_oper_status eq "up") {
                    $self->sanlog(-msg => $log_msg
                                . "Expected Error: Failed to modify FCP adapter"
                                . "state since LIFs are currently active");
                } else {
                    $self->_set_err("Failed to modify FCP adapter state",
                                    $fcpadapter_obj->errmsg());
                    return $SanUtil::FAIL;
                }
            } ## end if (($result eq $SanUtil::FALSE...

            &Tharn::sleep(15);

            # Verify adpater status
            my $adapter_curr_status = $fcpadapter_obj->get_status();
            if ($fcpadapter_obj->error()) {
                $self->_set_err($fcpadapter_obj->errmsg());
                return $SanUtil::FAIL;
            }

            if ($lifs_oper_status eq "up") {
                if ($adapter_curr_status ne $adapter_orig_status) {
                    $self->_set_err("Failed to verify FCP adapter state");
                    return $SanUtil::FAIL;
                }
            } else {
                if ($adapter_curr_status eq $adapter_orig_status) {
                    $self->_set_err("Failed to verify FCP adapter state");
                    return $SanUtil::FAIL;
                }
            }
        } ## end if ($adapter_state ne ...

        $self->sanlog(  -msg => "Succeed to modify FCP adapter "
                      . $adapter_port
                      . " state on node "
                      . $filer_obj->get_name()
                      . " and verify adapter status");
    } ## end for (keys(%adapters))

    return $SanUtil::PASS;
} ## end sub modify_fcp_adapter_state_for_given_lifs

################################################################################
#
# Function  - get_initiator_info_from_clients
#             Get Initiator Information from Clients
#
# Input     - sanmgr_obj   required (reference to SanMgr object)
# Output    - Reference to array of initoators
#             SanUtil::FAIL
#
################################################################################
sub get_initiator_info_from_clients {
    my ($self, %args) = @_;
    my $sanmgr_obj;
    my @client_objs;
    my @total_initiators = ();
    my $log_msg          = "FcpHelper::get_initiator_info_from_clients: ";

    $self->sanlog(-msg => "FcpHelper::get_initiator_info_from_clients - "
                  . "Get Initiator Information from Clients");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for sanmgr_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $sanmgr_obj = $value;
        }
    }

    # check for required arguments
    unless (defined($sanmgr_obj)) {
        $self->_set_err("Parameter 'sanmgr_obj' not specified");
        return $SanUtil::FAIL;
    }

    # Get the client objects
    @client_objs = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }

    # Get portname and nodename for each initiator
    foreach my $client_obj (@client_objs) {
        my @initiators = $client_obj->get_initiators();
        foreach my $initiator (@initiators) {
            my $initiator_wwpn = $initiator->get_portname();
            if ($initiator->error()) {
                $self->_set_err($initiator->errmsg());
                return $SanUtil::FAIL;
            }

            my $initiator_wwnn = $initiator->get_nodename();
            if ($initiator->error()) {
                $self->_set_err($initiator->errmsg());
                return $SanUtil::FAIL;
            }
            push(@total_initiators, $initiator_wwpn . "|" . $initiator_wwnn);
        } ## end foreach my $initiator (@initiators)
    } ## end foreach my $client_obj (@client_objs)

    $self->sanlog(
        -msg => $log_msg . "Succeed to get initiator information from clients");

    return \@total_initiators;
} ## end sub get_initiator_info_from_clients

################################################################################
#
# Function  - get_client_initiators
#             Get client initiator wwpn
#
# Input     - sanmgr_obj   required (reference to SanMgr object)
# Output    - Reference to hash of client initoators
#             (key: client name; value: array of initiator wwpn)
#             SanUtil::FAIL
#
################################################################################
sub get_client_initiators {
    my ($self, %args) = @_;
    my $sanmgr_obj;
    my @client_objs;
    my %client_init_wwpn;
    my $log_msg = "FcpHelper::get_client_initiators: ";

    $self->sanlog(-msg => "FcpHelper::get_client_initiators - "
                  . "Get Client Initiators");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for sanmgr_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $sanmgr_obj = $value;
        }
    }

    # check for required arguments
    unless (defined($sanmgr_obj)) {
        $self->_set_err("Parameter 'sanmgr_obj' not specified");
        return $SanUtil::FAIL;
    }

    # Get the client objects
    @client_objs = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }

    # Get wwpn for each initiator
    foreach my $client_obj (@client_objs) {
        my $host_name = $client_obj->hostname();
        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my @initiators = $client_obj->get_initiators();
        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg());
            return $SanUtil::FAIL;
        }
        my @init_wwpn_array = ();
        foreach my $initiator (@initiators) {
            my $initiator_wwpn = $initiator->get_portname();
            if ($initiator->error()) {
                $self->_set_err($initiator->errmsg());
                return $SanUtil::FAIL;
            }
            push(@init_wwpn_array, $initiator_wwpn);
        }
        @{$client_init_wwpn{$host_name}} = @init_wwpn_array;
    } ## end foreach my $client_obj (@client_objs)

    $self->sanlog(-msg => $log_msg . "Succeed to get client initiators");

    return \%client_init_wwpn;
} ## end sub get_client_initiators

################################################################################
#
# Function  - get_lif_portname
#             Get FCP LIF portname
#
# Input     - lif_obj           required
#             vs_obj            required
# Output    - lif portname
#             SanUtil::FAIL
#
################################################################################
sub get_lif_portname {
    my ($self, %args) = @_;
    my $lif_obj;
    my $vs_obj;
    my $fcp_obj;
    my $lif_name;
    my $lif_portname;
    my $log_msg = "FcpHelper::get_lif_portname";

    $self->sanlog(
             -msg => "FcpHelper::get_lif_portname - " . "Get FCP LIF Portname");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lif_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_obj = $value;
        } elsif ($key =~ /vs_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_obj)) {
        $self->_set_err("Parameter 'lif_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vs_obj' not specified");
        return $SanUtil::FAIL;
    }

    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $lif_name = $lif_obj->get_name();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my @portname_array = $fcp_obj->get_portnames(lif => $lif_name);
    if ($fcp_obj->error()) {
        $self->_set_err($fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $portname (@portname_array) {
        my $lif = $portname->{'lif'};
        if ($lif eq $lif_name) {
            $lif_portname = $portname->{'wwpn'};
            last;
        }
    }

    $self->sanlog(-msg => $log_msg . "Succeed to get LIF portname");

    return $lif_portname;
} ## end sub get_lif_portname

################################################################################
#
# Function  - get_lifs_for_given_port
#             Get array of LIF objects which resides on given adapter port
#
# Input     - filer_obj     required
#             adapter       required
# Output    - Reference to array of LIF objects
#             SanUtil::FAIL
#
################################################################################
sub get_lifs_for_given_port {
    my ($self, %args) = @_;
    my $filer_obj;
    my $adapter_port;
    my @existing_lif_objs = ();
    my $log_msg           = "FcpHelper::get_lifs_for_given_port";

    $self->sanlog(-msg => "FcpHelper::get_lifs_for_given_port - "
                  . "Get array of LIF objects for given adapter port");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /filer_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for filer_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $filer_obj = $value;
        } elsif ($key =~ /adapter/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for adapter: " . $value);
                return $SanUtil::FAIL;
            }
            $adapter_port = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($filer_obj)) {
        $self->_set_err("Parameter 'filer_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($adapter_port)) {
        $self->_set_err("Parameter 'adapter_port' not specified");
        return $SanUtil::FAIL;
    }

    my $filer_name = $filer_obj->get_name();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $TCD::FAIL;
    }

    my @vs_objs = $filer_obj->get_vservers();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $TCD::FAIL;
    }

    foreach my $vs_obj (@vs_objs) {
        my @lif_objs = $vs_obj->get_lif_objs();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $TCD::FAIL;
        }

        foreach my $lif_obj (@lif_objs) {
            my $lif_node = $lif_obj->get_curr_node();
            if ($lif_obj->error()) {
                $self->_set_err($lif_obj->errmsg());
                return $TCD::FAIL;
            }

            my $lif_port = $lif_obj->get_curr_port();
            if ($lif_obj->error()) {
                $self->_set_err($lif_obj->errmsg());
                return $TCD::FAIL;
            }

            if (($lif_node eq $filer_name) && ($lif_port eq $adapter_port)) {
                push(@existing_lif_objs, $lif_obj);
            }
        } ## end foreach my $lif_obj (@lif_objs)
    } ## end foreach my $vs_obj (@vs_objs)

    $self->sanlog(  -msg => $log_msg
                  . "Succeed to get LIF objects for given adapter port "
                  . $adapter_port);

    return \@existing_lif_objs;
} ## end sub get_lifs_for_given_port

################################################################################
#
# Function  - set_fcp_initiator_alias
#             Set FCP initiator alias and verify
#
# Input     - vs_obj           required
#             initiator         required
#             (initiator is a hash value, one of array of hash values got from
#             $fcp->get_initiators())
#             wwpn              required
#             alias             required
#             (alias is a value string)
#             force             optional
#             (force default value is empty string, user can pass 'true' if
#             using -force option to set fcp initiator alias, otherwise no
#             need to pass anything)
#             fcp_init_verify   optional
#             (By default for initiator alias set verification, command
#             'fcp initiator show -instance' will be executed, but for the case
#             LIFs operation status are 'down', we don't need to check this
#             command. fcp_init_verify is a flag and its default value is
#             'true', users can pass 'false' if they want to skip
#             'vserver fcp initiator show' checking, otherwise no need to pass
#             anything)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub set_fcp_initiator_alias {
    my ($self, %args) = @_;
    my $vs_obj;
    my $fcp_obj;
    my $initiator;
    my $wwpn;
    my $alias;
    my $force           = "";
    my $fcp_init_verify = "true";
    my $result;
    my $log_msg = "FcpHelper::set_fcp_initiator_alias: ";

    $self->sanlog(-msg => "FcpHelper::set_fcp_initiator_alias - "
                  . "Set FCP Initiator Alias");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vs_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        } elsif ($key =~ /initiator/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for initiator: " . $value);
                return $SanUtil::FAIL;
            }
            $initiator = $value;
        } elsif ($key =~ /wwpn/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for wwpn: " . $value);
                return $SanUtil::FAIL;
            }
            $wwpn = $value;
        } elsif ($key =~ /alias/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for alias: " . $value);
                return $SanUtil::FAIL;
            }
            $alias = $value;
        } elsif ($key =~ /force/) {
            $force = $value;
        } elsif ($key =~ /fcp_init_verify/) {
            $fcp_init_verify = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vs_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($initiator)) {
        $self->_set_err("Parameter 'initiator' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($wwpn)) {
        $self->_set_err("Parameter 'wwpn' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($alias)) {
        $self->_set_err("Parameter 'alias' not specified");
        return $SanUtil::FAIL;
    }

    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    if ($force eq "") {
        $result =
          $fcp_obj->wwpnalias_set(-wwpn  => $wwpn,
                                  -alias => $alias);
    } else {
        $result =
          $fcp_obj->wwpnalias_set(-wwpn  => $wwpn,
                                  -alias => $alias,
                                  -force => $force);
    }
    if ($fcp_obj->error()) {
        $self->_set_err(  "Fail to set FCP initiator " 
                        . $wwpn
                        . " alias "
                        . $alias . "\n"
                        . $fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to set FCP initiator "
                  . $wwpn
                  . " alias "
                  . $alias);

    # Verify FCP initiator alias by executing three commands
    # 1. 'fcp wwpn-alias show'
    # 2. 'fcp initiator show -instance'
    # (skip if user specifies fcp_init_verify flag as false)
    # 3. 'lun igroup show -instance'
    $result =
      $self->verify_fcp_initiator_alias(vs_obj          => $vs_obj,
                                        wwpn            => $wwpn,
                                        alias           => $alias,
                                        initiator       => $initiator,
                                        operation       => "set",
                                        fcp_init_verify => $fcp_init_verify);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err(
               "Fail to verify FCP initiator " . $wwpn . " alias set " . $alias,
               $self->errmsg());
        return $SanUtil::FAIL;
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify FCP initiator "
                  . $wwpn
                  . " alias "
                  . $alias
                  . " set");

    return $SanUtil::PASS;
} ## end sub set_fcp_initiator_alias

################################################################################
#
# Function  - remove_fcp_initiator_alias
#             Remove FCP initiator alias and verify
#
# Input     - vs_obj           required
#             initiator         required
#             (initiator is a hash value, one of array of hash values got from
#             $fcp->get_initiators())
#             wwpn              required
#             alias             required
#             (alias is a value string)
#             fcp_init_vserify  optional
#             (By default for initiator alias set verification, command
#             'fcp initiator show -instance' will be executed, but for the case
#             LIFs operation status are 'down', we don't need to check this
#             command. fcp_init_verify is a flag and its default value is
#             'true', users can pass 'false' if they want to skip
#             'vserver fcp initiator show' checking, otherwise no need to pass
#             anything)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub remove_fcp_initiator_alias {
    my ($self, %args) = @_;
    my $vs_obj;
    my $fcp_obj;
    my $initiator;
    my $wwpn;
    my $alias;
    my $fcp_init_verify = "true";
    my $result;
    my $log_msg = "FcpHelper::remove_fcp_initiator_alias: ";

    $self->sanlog(-msg => "FcpHelper::remove_fcp_initiator_alias - "
                  . "Remove FCP Initiator Alias");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vs_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        } elsif ($key =~ /initiator/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for initiator: " . $value);
                return $SanUtil::FAIL;
            }
            $initiator = $value;
        } elsif ($key =~ /wwpn/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for wwpn: " . $value);
                return $SanUtil::FAIL;
            }
            $wwpn = $value;
        } elsif ($key =~ /alias/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for alias: " . $value);
                return $SanUtil::FAIL;
            }
            $alias = $value;
        } elsif ($key =~ /fcp_init_verify/) {
            $fcp_init_verify = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vs_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($initiator)) {
        $self->_set_err("Parameter 'initiator' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($wwpn)) {
        $self->_set_err("Parameter 'wwpn' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($alias)) {
        $self->_set_err("Parameter 'alias' not specified");
        return $SanUtil::FAIL;
    }

    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $result = $fcp_obj->wwpnalias_remove(-alias => $alias);
    if ($fcp_obj->error()) {
        $self->_set_err(  "Fail to remove FCP initiator " 
                        . $wwpn
                        . " alias "
                        . $alias . "\n"
                        . $fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to remove FCP initiator "
                  . $wwpn
                  . " alias "
                  . $alias);

    # Verify FCP initiator alias by executing three commands
    # 1. 'fcp wwpn-alias show'
    # 2. 'fcp initiator show -instance'
    # (skip if user specifies fcp_init_verify flag as false)
    # 3. 'lun igroup show -instance'
    $result =
      $self->verify_fcp_initiator_alias(vs_obj          => $vs_obj,
                                        wwpn            => $wwpn,
                                        alias           => $alias,
                                        initiator       => $initiator,
                                        operation       => "remove",
                                        fcp_init_verify => $fcp_init_verify);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err("Fail to verify FCP initiator " 
                          . $wwpn
                          . " alias "
                          . $alias
                          . " remove",
                        $self->errmsg());
        return $SanUtil::FAIL;
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify FCP initiator "
                  . $wwpn
                  . " alias "
                  . $alias
                  . " remove");

    return $SanUtil::PASS;
} ## end sub remove_fcp_initiator_alias

################################################################################
#
# Function  - rename_lifs
#             Rename given FCP LIFs
#
# Input     - vserver_objs_ref   required (reference to array of vserver objects)
#             lif_objs_ref       required (reference to array of lif objects)
#             lif_new_names_ref  required (reference to array of LIF new names)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub rename_lifs {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $lif_objs_ref;
    my $lif_new_names_ref;
    my $cmd_interface = 'cli';
    my @lif_objs;
    my @lif_new_names;
    my $result;
    my $log_msg = "FcpHelper::rename_lifs: ";
    $self->sanlog(-msg => "FcpHelper::rename_lifs - Rename FCP LIFs");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /lif_new_names_ref/) {
            if (not length($value)) {
                $self->_set_err(
                              "Invalid value for lif_new_names_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_new_names_ref = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @lif_objs = @$lif_objs_ref;

    unless (defined($lif_new_names_ref)) {
        $self->_set_err("Parameter 'lif_new_names_ref' not specified");
        return $SanUtil::FAIL;
    }
    @lif_new_names = @$lif_new_names_ref;

    if (scalar(@lif_new_names) != scalar(@lif_objs)) {
        $self->_set_err("Not enough number of LIF new names are provided");
        return $SanUtil::FAIL;
    }

    my $count = 0;
    foreach my $lif_obj (@lif_objs) {
        my $lif_orig_name = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $lif_new_name = $lif_new_names[$count];

        $lif_obj->rename(-newname => $lif_new_name);
        if ($lif_obj->error()) {
            $self->_set_err(  "Failed to rename LIF "
                            . $lif_orig_name
                            . " to name "
                            . $lif_new_name . "\n"
                            . $lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "Succeed to rename LIF "
                      . $lif_orig_name
                      . " to name "
                      . $lif_new_name);

        # Verify rename
        my $lif_curr_name = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
        if ($lif_curr_name ne $lif_new_name) {
            $self->_set_err(  "Failed to verify LIF "
                            . $lif_orig_name
                            . " renamed to name "
                            . $lif_new_name);
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "Succeed to verify LIF "
                      . $lif_orig_name
                      . " renamed to name "
                      . $lif_new_name);
        $count++;
    } ## end foreach my $lif_obj (@lif_objs)

    # Verify LIFs on filer
    $result = $self->verify_lif_on_filer(vserver_objs_ref => $vserver_objs_ref,
                                         lif_objs_ref     => $lif_objs_ref,
                                         operation        => "create",
                                         cmd_interface    => $cmd_interface);
    if ($result eq $SanUtil::FAIL || $self->error()) {
        $self->_set_err("Failed to verify LIFs on filer", $self->errmsg());
        return $SanUtil::FAIL;
    }

    $self->sanlog(
           -msg => $log_msg . "Succeed to rename LIFs to new names and verify");

    return $SanUtil::PASS;
} ## end sub rename_lifs

################################################################################
#
# Function  - rename_vservers
#             Rename given vservers and verify
#
# Input     - vserver_objs_ref          required
#             vserver_new_names_ref     required
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub rename_vservers {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my @vserver_objs;
    my $vserver_new_names_ref;
    my @vserver_new_names;
    my $client_objs_ref;
    my @renamed_vs_objs;
    my $skip_client_verify = 0;
    my $result;
    my $log_msg = "FcpHelper::rename_vservers: ";

    $self->sanlog(-msg => "FcpHelper::rename_vservers - Rename FCP Vservers");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /vserver_new_names_ref/) {
            if (not length($value)) {
                $self->_set_err(
                          "Invalid value for vserver_new_names_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_new_names_ref = $value;
        } elsif ($key =~ /client_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for client_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_objs_ref = $value;
        } elsif ($key =~ /skip_client_verify/) {
            $skip_client_verify = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @vserver_objs = @$vserver_objs_ref;

    unless (defined($vserver_new_names_ref)) {
        $self->_set_err("Parameter 'vserver_new_names_ref' not specified");
        return $SanUtil::FAIL;
    }
    @vserver_new_names = @$vserver_new_names_ref;

    if (scalar(@vserver_new_names) != scalar(@vserver_objs)) {
        $self->_set_err("Not enough number of vserver new names are provided");
        return $SanUtil::FAIL;
    }

    unless (defined($client_objs_ref)) {
        $self->_set_err("Parameter 'client_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    my $count = 0;
    foreach my $vs_obj (@vserver_objs) {
        my $vs_orig_name = $vs_obj->get_name();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $vs_new_name = $vserver_new_names[$count];

        # vserver rename
        $result = $vs_obj->rename(-newname => $vs_new_name);
        if ($vs_obj->error()) {
            $self->_set_err(  "Failed to rename vserver "
                            . $vs_orig_name
                            . " to name "
                            . $vs_new_name . "\n"
                            . $vs_obj->errmsg());
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "Succeed to rename vserver "
                      . $vs_orig_name
                      . " to name "
                      . $vs_new_name);

        # Verify vserver rename on filer
        my $vs_curr_name = $vs_obj->get_name();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $SanUtil::FAIL;
        }
        if ($vs_curr_name ne $vs_new_name) {
            $self->_set_err(  "Failed to verify vserver "
                            . $vs_curr_name
                            . " renamed to name "
                            . $vs_new_name);
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "Succeed to verify vserver "
                      . $vs_curr_name
                      . " renamed to name "
                      . $vs_new_name
                      . " on filer");

        push(@renamed_vs_objs, $vs_obj);

        # Below step is particularly for the newly created FCP service
        # on renamed vserver
        if ($vs_obj->fcp_exist()) {
            my $fcp_obj = $vs_obj->get_fcp_obj();
            $fcp_obj->{'vserver'} = $vs_obj->get_name();
        }

        $count++;
    } ## end foreach my $vs_obj (@vserver_objs)

    if ($skip_client_verify == 0) {

        # Verify vserver rename on client
        $result =
          $self->verify_vserver_on_client(vserver_objs_ref => \@renamed_vs_objs,
                                          client_objs_ref  => $client_objs_ref);
        if ($result eq $SanUtil::FAIL || $self->error()) {
            $self->_set_err("Failed to verify vservers on client",
                            $self->errmsg());
            return $SanUtil::FAIL;
        }
        $self->sanlog(-msg => $log_msg
                  . "Succeed to verify vservers rename to new names on client");
    } ## end if ($skip_client_verify...

    $self->sanlog(-msg => $log_msg
                  . "Succeed to rename vservers to new names and verify");

    return $SanUtil::PASS;
} ## end sub rename_vservers

################################################################################
#
# Function  - set_fcp_lif_portname
#             Set FCP LIFs new portname
#
# Input     - lif_obj               required
#             lif_new_portname      required (type is string, value is wwpn)
#             vs_obj                required
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub set_fcp_lif_portname {
    my ($self, %args) = @_;
    my $lif_obj;
    my $vs_obj;
    my $fcp_obj;
    my $lif_new_portname;
    my $lif_name;
    my $result;
    my $log_msg = "FcpHelper::set_fcp_lif_portname: ";

    $self->sanlog(-msg => "FcpHelper::set_fcp_lif_portname - "
                  . "Set FCP LIF New Portname and Verify");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lif_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_obj = $value;
        } elsif ($key =~ /lif_new_portname/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for lif_new_portname: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_new_portname = $value;
        } elsif ($key =~ /vs_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_obj)) {
        $self->_set_err("Parameter 'lif_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_new_portname)) {
        $self->_set_err("Parameter 'lif_new_portname' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vs_obj' not specified");
        return $SanUtil::FAIL;
    }

    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $lif_name = $lif_obj->get_name();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $result = $fcp_obj->set_portname(wwpn => $lif_new_portname,
                                     lif  => $lif_name);

    if ($fcp_obj->error()) {
        $self->_set_err(  "Failed to set LIF "
                        . $lif_name
                        . " new portname "
                        . $lif_new_portname . "\n"
                        . $fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to set LIF "
                  . $lif_name
                  . " new portname "
                  . $lif_new_portname);

    # Verify LIF new portname
    my @portname_array = $fcp_obj->get_portnames(lif => $lif_name);
    if ($fcp_obj->error()) {
        $self->_set_err($fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }
    foreach my $portname (@portname_array) {
        my $lif = $portname->{'lif'};
        if ($lif eq $lif_name) {
            my $lif_portname = $portname->{'wwpn'};
            if ($lif_portname ne $lif_new_portname) {
                $self->_set_err(  "Failed to verify LIF "
                                . $lif_name
                                . " new portname "
                                . $lif_new_portname);
                return $SanUtil::FAIL;
            }
        } ## end if ($lif eq $lif_name)
    } ## end foreach my $portname (@portname_array)

    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify LIF "
                  . $lif_name
                  . " new portname "
                  . $lif_new_portname);

    return $SanUtil::PASS;
} ## end sub set_fcp_lif_portname

################################################################################
#
# Function  - rescan_disks_on_clients
#
# Input     - client_objs_ref  required (reference to array of client objects)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub rescan_disks_on_clients {
    my ($self, %args) = @_;
    my $client_objs_ref;
    my @client_objs;
    my $cmd;
    my ($status, @output);
    my $result;
    my $log_msg = "FcpHelper::rescan_disks_on_client: ";

    $self->sanlog(-msg => "FcpHelper::rescan_disks_on_client - "
                  . "Rescan disks on Clients");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /client_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for client_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_objs_ref = $value;
        }
    }

    # check for required arguments
    unless (defined($client_objs_ref)) {
        $self->_set_err("Parameter 'client_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @client_objs = @$client_objs_ref;

    # Verify LIF info from client
    foreach my $client_obj (@client_objs) {
        my $win_client = $client_obj->get_WinClient();
        if ($client_obj->error()) {
            $self->_set_err($client_obj->errmsg());
            return $SanUtil::FAIL;
        }
        
        my $host_name = $win_client->{hostname};

        $win_client->rescanDisks();
        if ($win_client->error()) {
            $self->_set_err(  "Failed to rescan disks on client "
                            . $host_name . "\n"
                            . $win_client->error());
            return $SanUtil::FAIL;
        }
    } ## end foreach my $client_obj (@client_objs)

    $self->sanlog(-msg => $log_msg . "Succeed to rescan disks on all clients");

    return $SanUtil::PASS;
} ## end sub rescan_disks_on_clients

################################################################################
#
# Function  - verify_san_objects
#             Verify SAN Objects Created By FitMgr
#
# Input     - sanmgr_obj   required (reference to SanMgr object)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_san_objects {
    my ($self, %args) = @_;
    my $sanmgr_obj;
    my @vservers;
    my @clients;
    my @lifs;
    my @luns;
    my $log_msg = "FcpHelper::verify_san_objects: ";

    $self->sanlog(-msg => "FcpHelper::verify_san_objects - "
                  . "Verify SAN Objects Created By SanMgr");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for sanmgr_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $sanmgr_obj = $value;
        }
    }

    # check for required arguments
    unless (defined($sanmgr_obj)) {
        $self->_set_err("Parameter 'sanmgr_obj' not specified");
        return $SanUtil::FAIL;
    }

    # Get vserver objects
    @vservers = $sanmgr_obj->get_objects(-type => "vserver");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }
    if (scalar(@vservers) <= 0) {
        $self->_set_err("Failed to get vserver objects");
        return $SanUtil::FAIL;
    }

    # Get the client objects
    @clients = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }
    if (scalar(@clients) <= 0) {
        $self->_set_err("Failed to get client objects");
        return $SanUtil::FAIL;
    }

    # Get LIF objects
    @lifs = $sanmgr_obj->get_objects(type => "lif");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }
    if (scalar(@lifs) <= 0) {
        $self->_set_err("Failed to get LIF objects");
        return $SanUtil::FAIL;
    }

    # Get LUN objects
    @luns = $sanmgr_obj->get_objects(-type => "lun");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FAIL;
    }
    if (scalar(@luns) <= 0) {
        $self->_set_err("Failed to get LUN objects");
        return $SanUtil::FAIL;
    }

    $self->sanlog(-msg => $log_msg . "Succeed to verify san objects");

    return $SanUtil::PASS;
} ## end sub verify_san_objects

################################################################################
#
# Function  - verify_lif_on_filer
#             Verify Created/Deleted LIFs on Filer by executing three commands
#             network interface show/fcp port show/fcp portname show and
#             verify LIF properties match
#             (name/vserver/status_admin/status_oper/curr_node/curr_port/wwpn)
#
# Input     - lif_objs_ref      required (reference to array of LIF objects)
#             vserver_objs_ref  required (reference to array of vserver objects)
#             operation         required ('create' or 'delete')
#             skip_fcp_check    optional (default value is '0')
#             (If LIFs are created on vserver which no FCP service exist,
#              or stopped vserver, skip verification on 'vserver fcp port show'
#              and 'vserver fcp portname show')
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_lif_on_filer {
    my ($self, %args) = @_;
    my $lif_objs_ref;
    my $vserver_objs_ref;
    my $operation;
    my $skip_fcp_check = 0;
    my $cmd_interface  = 'cli';
    my @lif_objs;
    my @vserver_objs;
    my $result;
    my %lif_properties;
    my $log_msg = "FcpHelper::verify_lif_on_filer: ";

    $self->sanlog(-msg => "FcpHelper::verify_lif_on_filer - "
                  . "Verify Created/Deleted LIFs on Filer");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /operation/) {
            if (not length($value) and $value !~ /^create$|^delete$/i) {
                $self->_set_err("Invalid value for operation: " . $value);
                return $SanUtil::FAIL;
            }
            $operation = $value;
        } elsif ($key =~ /skip_fcp_check/) {
            $skip_fcp_check = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($operation)) {
        $self->_set_err("Parameter 'operation' not specified");
        return $SanUtil::FAIL;
    }

    @lif_objs     = @$lif_objs_ref;
    @vserver_objs = @$vserver_objs_ref;

    foreach my $lif_obj (@lif_objs) {
        my $lif_name;
        if ($operation eq "create") {
            $lif_name = $lif_obj->get_name();
            if ($lif_obj->error()) {
                $self->_set_err($lif_obj->errmsg());
                return $SanUtil::FAIL;
            }

            my $lif_vs_name = $lif_obj->get_vserver();
            if ($lif_obj->error()) {
                $self->_set_err($lif_obj->errmsg());
                return $SanUtil::FAIL;
            }

            foreach my $vs_obj (@vserver_objs) {
                if ($lif_vs_name eq $vs_obj->get_name()) {

                    # Verify by "network interface show" command
                    $result = $vs_obj->lif_exist(name => $lif_name);
                    if ($vs_obj->error()) {
                        $self->_set_err($vs_obj->errmsg());
                        return $SanUtil::FAIL;
                    }
                    if ($result eq $SanUtil::FALSE) {
                        $self->_set_err(  "LIF "
                                        . $lif_name
                                        . " of vserver "
                                        . $lif_vs_name
                                        . " failed to show up on filer");
                        return $SanUtil::FAIL;
                    }
                    $self->sanlog(  -msg => $log_msg . "LIF "
                                  . $lif_name
                                  . " of vserver "
                                  . $lif_vs_name
                                  . " shows up on filer");

                    # Get LIF properties
                    $result =
                      $self->_get_lif_properties($lif_obj, $cmd_interface);
                    if ($result eq $SanUtil::FAIL || $self->error()) {
                        $self->_set_err(
                              "Failed to get LIF " . $lif_name . " properties");
                        return $SanUtil::FAIL;
                    }
                    %lif_properties = %$result;
                    if ($skip_fcp_check == 0) {
                        my $fcp_obj = $vs_obj->get_fcp_obj();
                        if ($vs_obj->error()) {
                            $self->_set_err($vs_obj->errmsg());
                            return $SanUtil::FAIL;
                        }

                        # Verify by "vserver fcp interface show" command
                        $result =
                          $self->_verify_lif_port(
                                        $fcp_obj,   $lif_name, \%lif_properties,
                                        $operation, $cmd_interface);
                        if ($result eq $SanUtil::FAIL || $self->error()) {
                            $self->_set_err(
                                "Failed to verify port information of " . "LIF "
                                  . $lif_name
                                  . " on filer");
                            return $SanUtil::FAIL;
                        }
                        $self->sanlog(-msg => $log_msg
                                  . "Succeed to verify port information of LIF "
                                  . $lif_name
                                  . " on filer");

                        # Verify by "vserver fcp portname show" command
                        $result =
                          $self->_verify_lif_portname(
                                        $fcp_obj,   $lif_name, \%lif_properties,
                                        $operation, $cmd_interface);
                        if ($result eq $SanUtil::FAIL || $self->error()) {
                            $self->_set_err(
                                     "Failed to verify portname information of "
                                       . "LIF "
                                       . $lif_name
                                       . " on filer");
                            return $SanUtil::FAIL;
                        }
                        $self->sanlog(-msg => $log_msg
                                  . "Succeed to verify portname information of "
                                  . "LIF "
                                  . $lif_name
                                  . " on filer");
                    } ## end if ($skip_fcp_check ==...
                } ## end if ($lif_vs_name eq $vs_obj...
            } ## end foreach my $vs_obj (@vserver_objs)
        } elsif ($operation eq "delete") {
            $lif_name = $lif_obj;
            my $lif_vs_name;
            foreach my $vs_obj (@vserver_objs) {
                $lif_vs_name = $vs_obj->get_name();
                if ($vs_obj->error()) {
                    $self->_set_err($vs_obj->errmsg());
                    return $SanUtil::FAIL;
                }

                # Verify by "network interface show" command
                $result = $vs_obj->lif_exist(name => $lif_name);
                if ($vs_obj->error()) {
                    $self->_set_err($vs_obj->errmsg());
                    return $SanUtil::FAIL;
                }

                if ($result eq $SanUtil::TRUE) {
                    $self->_set_err(  "LIF "
                                    . $lif_name
                                    . " of vserver "
                                    . $lif_vs_name
                                    . " still shows up on filer");
                    return $SanUtil::FAIL;
                }

                if ($skip_fcp_check == 0) {
                    my $fcp_obj = $vs_obj->get_fcp_obj();
                    if ($vs_obj->error()) {
                        $self->_set_err($vs_obj->errmsg());
                        return $SanUtil::FAIL;
                    }

                    # Verify by "vserver fcp interface show" command
                    $result =
                      $self->_verify_lif_port(
                                        $fcp_obj,   $lif_name, \%lif_properties,
                                        $operation, $cmd_interface);
                    if ($result eq $SanUtil::FAIL || $self->error()) {
                        $self->_set_err(
                                     "Failed to verify port information of LIF "
                                       . $lif_name
                                       . " on filer",
                                     $self->errmsg());
                        return $SanUtil::FAIL;
                    }

                    # Verify by "vserver fcp portname show" command
                    $result =
                      $self->_verify_lif_portname(
                                        $fcp_obj,   $lif_name, \%lif_properties,
                                        $operation, $cmd_interface);
                    if ($result eq $SanUtil::FAIL || $self->error()) {
                        $self->_set_err(
                                       "LIF "
                                         . $lif_name
                                         . " of vserver "
                                         . $lif_vs_name
                                         . " should not show portname on filer",
                                       $self->errmsg());
                        return $SanUtil::FAIL;
                    } ## end if ($result eq $SanUtil::FAIL...
                } ## end if ($skip_fcp_check ==...
            } ## end foreach my $vs_obj (@vserver_objs)
            $self->sanlog(  -msg => $log_msg . "LIF "
                          . $lif_name
                          . " of vserver "
                          . $lif_vs_name
                          . " does not show up on filer");
            $self->sanlog(  -msg => $log_msg
                          . "Succeed to verify port information of LIF "
                          . $lif_name
                          . " on filer");
            $self->sanlog(  -msg => $log_msg
                          . "Succeed to verify portname information of LIF "
                          . $lif_name
                          . " on filer");
        } ## end elsif ($operation eq "delete")
    } ## end foreach my $lif_obj (@lif_objs)

    $self->sanlog(-msg => $log_msg . "Succeed to verify the LIFs on filer");

    return $SanUtil::PASS;
} ## end sub verify_lif_on_filer

################################################################################
#
# Function  - verify_lif_initiator_login
#             Verify whether or not LIFs login to initiators
#
# Input     - vserver_objs_ref required (reference to array of vserver objects)
#             lif_objs_ref     required (reference to array of LIF objects)
#             login            required
#             (If LIFs login to initiator, then value for login should be 1,
#             otherwise, value for login is 0)
# Output    - SanUtil::PASS;
#             SanUtil::FAIL
#
################################################################################
sub verify_lif_initiator_login {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my @vserver_objs;
    my $lif_objs_ref;
    my @lif_objs;
    my $login;
    my $result;
    my $log_msg = "FcpHelper::verify_lif_initiator_login: ";

    $self->sanlog(-msg => "FcpHelper::verify_lif_initiator_login - "
                  . "Verify Initiators Login");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vservers_obj_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /login/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for login: " . $value);
                return $SanUtil::FAIL;
            }
            $login = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @vserver_objs = @$vserver_objs_ref;

    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @lif_objs = @$lif_objs_ref;

    unless (defined($login)) {
        $self->_set_err("Parameter 'login' not specified");
        return $SanUtil::FAIL;
    }

    foreach my $vserver_obj (@vserver_objs) {

        # Get vserver name
        my $vs_name = $vserver_obj->get_name();
        if ($vserver_obj->error()) {
            $self->_set_err($vserver_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Get FCP object
        my $fcp_obj = $vserver_obj->get_fcp_obj();
        if ($vserver_obj->error()) {
            $self->_set_err($vserver_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Refresh FCP object
        $fcp_obj->refresh();
        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my @initiators = $fcp_obj->get_initiators();
        if ($fcp_obj->error()) {
            if ($login == 1) {
                $self->_set_err($fcp_obj->errmsg());
                return $SanUtil::FAIL;
            }
        } else {
            if ($login == 0 && scalar(@initiators) > 0) {
                return $SanUtil::FAIL;
            } elsif ($login == 0 && scalar(@initiators) == 0) {
                next;
            } elsif ($login == 1) {
                my $lif_count = 0;
                foreach my $lif_obj (@lif_objs) {
                    my $lif_name = $lif_obj->get_name();
                    if ($lif_obj->error()) {
                        $self->_set_err($lif_obj->errmsg());
                        return $SanUtil::FAIL;
                    }

                    my $lif_vs = $lif_obj->get_vserver();
                    if ($lif_obj->error()) {
                        $self->_set_err($lif_obj->errmsg());
                        return $SanUtil::FAIL;
                    }

                    # Find LIF objects which belongs to given vserver object
                    # (from line 3863) and make sure LIF objects are login to
                    # initiators
                    if ($lif_vs eq $vs_name) {
                        foreach my $initiator (@initiators) {
                            if (   ($initiator->{'igroup'} ne "-")
                                && ($lif_name eq $initiator->{'lif'})
                                && ($lif_vs   eq $initiator->{'vserver'})) {
                                $self->sanlog(-msg => $log_msg
                                    . "FCP LIF $lif_name on vserver $lif_vs login to initiator"
                                    );
                                $lif_count++;
                            }
                        } ## end foreach my $initiator (@initiators)
                    } ## end if ($lif_vs eq $vs_name)
                } ## end foreach my $lif_obj (@lif_objs)

                if ($lif_count == 0) {
                    $self->_set_err(
                                  "Failed to verify the LIFs initiators login");
                    return $SanUtil::FAIL;
                }
            } ## end elsif ($login == 1)
        } ## end else [ if ($fcp_obj->error())
    } ## end foreach my $vserver_obj (@vserver_objs)

    $self->sanlog(
              -msg => $log_msg . "Succeed to verify the LIFs initiators login");

    return $SanUtil::PASS;
} ## end sub verify_lif_initiator_login

################################################################################
#
# Function  - verify_initiator_login
#             Verify initiators login, used for LIFs verification on clients
#             Make sure initoiators which show up on
#             'vserver fcp initiator show' will also show up on clients
#
# Input     - vserver_objs_ref required (reference to array of vserver objects)
#             initiators_ref   required (reference to array of client initiator
#                                       objects)
# Output    - reference to hash (key: initiator wwwpn, value: array of LIF wwpn)
#             SanUtil::FAIL
#
################################################################################
sub verify_initiator_login {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $total_initiators_ref;
    my @vserver_objs;
    my @total_initiators;
    my %lifs_initiators;
    my $result;
    my $log_msg = "FcpHelper::verify_initiator_login: ";

    $self->sanlog(-msg => "FcpHelper::verify_initiator_login - "
                  . "Verify Initiators Login");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vservers_obj_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /initiators_ref/) {
            if (not length($value)) {
                $self->_set_err(
                             "Invalid value for initiators_obj_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $total_initiators_ref = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($total_initiators_ref)) {
        $self->_set_err("Parameter 'total_initiators_ref' not specified");
        return $SanUtil::FAIL;
    }

    @vserver_objs     = @$vserver_objs_ref;
    @total_initiators = @$total_initiators_ref;

    foreach my $vserver_obj (@vserver_objs) {

        # Get vserver name
        my $vs_name = $vserver_obj->get_name();
        if ($vserver_obj->error()) {
            $self->_set_err($vserver_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Get FCP object
        my $fcp_obj = $vserver_obj->get_fcp_obj();
        if ($vserver_obj->error()) {
            $self->_set_err($vserver_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Refresh FCP object
        $fcp_obj->refresh();
        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Get initiator wwpn and wwnn from filer
        my @initiators = $fcp_obj->get_initiators();
        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my @init_array = ();
        foreach my $initiator (@initiators) {
            my $initiator_wwpn = $initiator->{'wwpn'};
            my $initiator_wwnn = $initiator->{'wwnn'};
            my $lif_name       = $initiator->{'lif'};
            my $lif_vs         = $initiator->{'vserver'};

            my $initiator_igroup = $initiator->{'igroup'};
            if ($initiator_igroup ne "-") {
                push(@init_array, $initiator_wwpn . '|' . $initiator_wwnn);
            }

            # Verify initiators which LIFs login to are in the list of
            # client initiators
            $result = $self->_array_compare(\@init_array, \@total_initiators);
            if ($result eq $SanUtil::FAIL || $self->error()) {

                $self->_set_err("Failed to verify LIF initiator login",
                                $self->errmsg());
                return $SanUtil::FAIL;
            }

            # Group LIFs (wwpn) per initiator
            # Hash %lifs_initiators
            # (key: initiator wwpn; value: array of LIF wwpn)
            my @lif_wwpn_array = ();
            if ($lifs_initiators{$initiator_wwpn}) {
                @lif_wwpn_array = @{$lifs_initiators{$initiator_wwpn}};
            }

            # Get LIF wwpn
            # Can not get LIF wwpn directly from 'vserver fcp initiator show'
            # Get LIF wwpn through 'vserver fcp portname show'
            $result = $self->_get_lif_wwpn(\@vserver_objs, $lif_vs, $lif_name);
            if ($result eq $SanUtil::FAIL || $self->error()) {
                $self->_set_err("Failed to get LIF wwpn", $self->errmsg());
                return $SanUtil::FAIL;
            }
            my $lif_wwpn = $result;
            push(@lif_wwpn_array, $lif_wwpn);

            if ($initiator_igroup ne "-") {
                @{$lifs_initiators{$initiator_wwpn}} = @lif_wwpn_array;
            }
        } ## end foreach my $initiator (@initiators)
    } ## end foreach my $vserver_obj (@vserver_objs)

    my @initiators_login = keys %lifs_initiators;
    if (scalar(@initiators_login) == 0) {
        $self->sanlog(-msg => $log_msg . "No LIF login to initiator");
    } else {
        $self->sanlog(
                   -msg => $log_msg . "Succeed to verify the initiators login");
    }

    # Hash %lifs_initiators
    # (key: initiator wwpn; value: array of LIF wwpn which login to initiator)
    return \%lifs_initiators;
} ## end sub verify_initiator_login

################################################################################
#
# Function  - verify_lif_on_client
#              Verify Created/Deleted LIFs on Clients
#
# Input     - vserver_objs_ref   required (reference to array of vserver objects)
#             lif_objs_login_ref required (reference to hash of lif wwpn)
#             client_objs_ref    required (reference to array of client objects)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_lif_on_client {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $lif_objs_login_ref;
    my $client_objs_ref;
    my %lifs_initiators;
    my @vserver_objs;
    my @client_objs;
    my $cmd;
    my ($status, @output);
    my $result;
    my $line;
    my @lif_wwpn_array_from_filer  = ();
    my @lif_wwpn_array_from_client = ();
    my $retry_loop                 = 5;
    my $log_msg                    = "FcpHelper::verify_lif_on_client: ";

    $self->sanlog(-msg => "FcpHelper::verify_lif_on_client - "
                  . "Verify Created/Deleted LIFs on Clients");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /lif_objs_login_ref/) {
            if (not length($value)) {
                $self->_set_err(
                             "Invalid value for lif_objs_login_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_login_ref = $value;
        } elsif ($key =~ /client_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for client_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_objs_ref = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($lif_objs_login_ref)) {
        $self->_set_err("Parameter 'lif_objs_login_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($client_objs_ref)) {
        $self->_set_err("Parameter 'client_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    @vserver_objs = @$vserver_objs_ref;

    # Hash %lifs_initiators
    # (key: initiator wwpn; value: array of LIF wwpn which login to initiator)
    %lifs_initiators = %$lif_objs_login_ref;
    @client_objs     = @$client_objs_ref;

    # Verify LIF info from client
    foreach my $client_obj (@client_objs) {
        my $win_client = $client_obj->get_WinClient();
        if ($client_obj->error()) {
            $self->_set_err("Failed to get WinClient", $client_obj->error());
            return $SanUtil::FAIL;
        }

        # Retry 5 times
        for (my $i = 0 ; $i < $retry_loop ; $i++) {

            # Verify LIF portname and node name information
            my $fcp_info_hash = $client_obj->get_fcp_info();
            
            my %fcp_info = %$fcp_info_hash;
            my @lif_wwpn_array_from_client = ();
            foreach (keys %fcp_info) {
                my $discover_ports_hash = $fcp_info{$_};
                if ($discover_ports_hash->{'discoveredports'}) {
                    my $port_info_hash = $discover_ports_hash->{'discoveredports'};
                    my @port_info = @$port_info_hash;
                    foreach my $lif_info (@port_info) {   
                        if ($lif_info->{'state'} eq 'Operational'){
                            push (@lif_wwpn_array_from_client, $lif_info->{'portname'});    
                        }
                    }
                }    
            }

            for (keys %lifs_initiators) {
                my $initiator_wwpn            = $_;
                my @lif_wwpn_array_from_filer = @{$lifs_initiators{$_}};

                @lif_wwpn_array_from_filer = sort (@lif_wwpn_array_from_filer);
                @lif_wwpn_array_from_client =
                  sort (@lif_wwpn_array_from_client);
                my $tmp_string = '';
                foreach (@lif_wwpn_array_from_filer) {
                    $tmp_string = $tmp_string . "\t$_";
                }
                $self->sanlog(
                            -msg => "FCP LIF wwpn list on filer:\n$tmp_string");

                # Number of lif wwpn shown up on client (hba_info) will be =
                # number of lif wwpn displayed by command
                # "vserver fcp initiator show" * number of LUN created
                # Remove duplicate values in @lif_wwpn_array_from_client;
                my %tmp_hash;
                @lif_wwpn_array_from_client = grep !($tmp_hash{$_}++),
                  @lif_wwpn_array_from_client;
                $tmp_string = '';
                foreach (@lif_wwpn_array_from_client) {
                    $tmp_string = $tmp_string . "\t$_";
                }
                $self->sanlog(
                           -msg => "FCP LIF/port wwpn list on client:\n$tmp_string");

                $result = $self->_array_compare(\@lif_wwpn_array_from_filer,
                                                \@lif_wwpn_array_from_client);
                if ($result == $SanUtil::FAIL || $self->error()) {
                    if ($i == 4) {
                        $self->_set_err(
                            "FCP LIF which logins into initiator does not show up on client"
                            );
                        $self->_set_err("LIF verification failed on client "
                                          . $client_obj->hostname(),
                                        $self->errmsg());
                        return $SanUtil::FAIL;
                    }
                    &Tharn::sleep(60);
                } else {
                    $i = 5;
                }
            } ## end for (keys %lifs_initiators)
            @lif_wwpn_array_from_client = ();
        } ## end for (my $i = 0 ; $i < $retry_loop...
    } ## end foreach my $client_obj (@client_objs)

    $self->sanlog(-msg => $log_msg . "Succeed to verify the LIFs on client");

    return $SanUtil::PASS;
} ## end sub verify_lif_on_client

################################################################################
#
# Function  - verify_lif_on_switch
#             Veryfy created/deleted LIFs on switch
#
# Input     -
# Output    -
#
################################################################################
sub verify_lif_on_switch {

    # Code will be added in when SanMgr module is ready
    return $SanUtil::PASS;
}

################################################################################
#
# Function  - verify_lif_status
#             Verify FCP LIF status are as expected
#
# Input     - lif_objs_ref  required (reference to array of LIF objects)
#             status_admin  required (expected status: 'up' or 'down')
#             status_oper   required (expected status: 'up' or 'down')
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_lif_status {
    my ($self, %args) = @_;
    my $lif_objs_ref;
    my $status_admin;
    my $status_oper;
    my $lif_name;
    my $lif_node;
    my $lif_port;
    my $lif_vs;
    my $log_msg = "FcpHelper::verify_lif_status: ";

    $self->sanlog(-msg => "FcpHelper::verify_lif_status - Verify LIF Status");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /lif_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for lif_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $lif_objs_ref = $value;
        } elsif ($key =~ /status_admin/) {
            if (not length($value) and $status_admin !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for status: " . $value);
                return $SanUtil::FAIL;
            }
            $status_admin = $value;
        } elsif ($key =~ /status_oper/) {
            if (not length($value) and $status_oper !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for status: " . $value);
                return $SanUtil::FAIL;
            }
            $status_oper = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($lif_objs_ref)) {
        $self->_set_err("Parameter 'lif_objs_ref' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($status_admin)) {
        $self->_set_err("Parameter 'status_admin' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($status_oper)) {
        $self->_set_err("Parameter 'status_oper' not specified");
        return $SanUtil::FAIL;
    }

    foreach my $lif_obj (@$lif_objs_ref) {
        my $lif_name = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
        my $vs_name = $lif_obj->get_vserver();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $lif_status_admin = $lif_obj->get_admin_status();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $lif_status_oper = $lif_obj->get_status_oper();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($lif_status_admin ne $status_admin) {
            $self->_set_err("LIF "
                         . $lif_name
                         . " status is "
                         . $lif_status_admin
                         . " which is not consistant with status it should be");
            return $SanUtil::FAIL;
        }

        if ($lif_status_oper ne $status_oper) {
            $self->_set_err("LIF "
                         . $lif_name
                         . " status is "
                         . $lif_status_oper
                         . " which is not consistant with status it should be");
            return $SanUtil::FAIL;
        }
    } ## end foreach my $lif_obj (@$lif_objs_ref)

    $self->sanlog(-msg => $log_msg . "Succeed to verify the LIFs status");

    return $SanUtil::PASS;
} ## end sub verify_lif_status

################################################################################
#
# Function  - verify_fcp_initiator_alias
#             Verify FCP initiator alias by executing three commands
#             1. 'fcp wwpn-alias show'
#             2. 'fcp initiator show -instance'
#             (skip if user specifies fcp_init_verify flag as false)
#             3. 'lun igroup show -instance'
#
# Input     - vs_obj            required
#             wwpn              required
#             alias             required
#             (alias is a string value)
#             initiator         required
#             (initiator is a hash value, one of array of hash values got from
#              $fcp->get_initiators())
#             operation         required
#             fcp_init_verify   required
#             (By default for initiator alias set verification, command
#             'fcp initiator show -instance' will be executed, but for the case
#             LIFs operation status are 'down', we don't need to check this
#             command. fcp_init_verify default value is 'true', user can pass
#             'false' if want to skip 'vserver fcp initiator show' checking,
#             otherwise no need to pass anything)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_fcp_initiator_alias {
    my ($self, %args) = @_;
    my $vs_obj;
    my $fcp_obj;
    my $init_wwpn;
    my $wwpn_alias;
    my $initiator;
    my $operation;
    my $fcp_init_verify;
    my $wwpn;
    my $alias_string;
    my @alias_array;
    my $result;
    my $log_msg = "FcpHelper::verify_fcp_initiator_alias: ";

    $self->sanlog(-msg => "FcpHelper::verify_fcp_initiator_alias - "
                  . "Verify FCP initiator alias");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vs_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vs_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        } elsif ($key =~ /wwpn/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for wwpn: " . $value);
                return $SanUtil::FAIL;
            }
            $init_wwpn = $value;
        } elsif ($key =~ /alias/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for alias: " . $value);
                return $SanUtil::FAIL;
            }
            $wwpn_alias = $value;
        } elsif ($key =~ /initiator/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for initiator: " . $value);
                return $SanUtil::FAIL;
            }
            $initiator = $value;
        } elsif ($key =~ /operation/) {
            if (not length($value) and $value !~ /^set$|^remove$/i) {
                $self->_set_err("Invalid value for operation: " . $value);
                return $SanUtil::FAIL;
            }
            $operation = $value;
        } elsif ($key =~ /fcp_init_verify/) {
            if (not length($value) and $value !~ /^true$|^false$/i) {
                $self->_set_err("Invalid value for fcp_init_verify: " . $value);
                return $SanUtil::FAIL;
            }
            $fcp_init_verify = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vs_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($init_wwpn)) {
        $self->_set_err("Parameter 'wwpn' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($wwpn_alias)) {
        $self->_set_err("Parameter 'wwpn_alias' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($initiator)) {
        $self->_set_err("Parameter 'initiator' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($operation)) {
        $self->_set_err("Parameter 'operation' not specified");
        return $SanUtil::FAIL;
    }

    # Verify by command 'fcp wwpn-alias show'
    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $result = $fcp_obj->wwpnalias_exists(-alias => $wwpn_alias);
    if ($fcp_obj->error()) {
        $self->_set_err($fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }

    # Since verification is needed for both wwpn-alias set
    # and wwpn-alias remove, error should be handled differently
    if ($operation eq "set") {
        if ($result eq $SanUtil::FALSE) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }
    } elsif ($operation eq "remove") {
        if ($result eq $SanUtil::TRUE) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }
    }
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify fcp initiator alias name "
                  . "by command 'fcp-wwpn-alias show'");

    if ($fcp_init_verify eq "true") {

        # Verify by command 'fcp initiator show -instance'
        # Call fcp->get_initiators() again to get latest initiator list
        # with updated objects
        my @initiators = $fcp_obj->get_initiators();
        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # It is possible that controller is logging to multiple
        # initiators (zoning) across multiple clients
        # and only some of clients are specified in the .xml file for testing.
        # Code need to find out the client initiators which test
        # is running against.
        foreach my $init (@initiators) {
            my $initiator_wwpn = $init->{'wwpn'};
            if ($initiator_wwpn eq $init_wwpn) {
                $initiator = $init;
                last;
            }
        }

        # $initiator is a hash object
        # $initiator->{'wwpn'}:  initiator wwpn value
        # $initiator->{'alias'}: a string which contains all alias names
        #                        for given initiator
        $wwpn         = $initiator->{'wwpn'};
        $alias_string = $initiator->{'alias'};
        @alias_array  = split(",", $alias_string);
        $result =
          $self->_verify_fcp_init_wwpn_and_alias(
                                          $init_wwpn, $wwpn_alias, $initiator,
                                          $operation, $wwpn,       \@alias_array
                                          );
        if ($result eq $SanUtil::FAIL || $self->error()) {
            $self->_set_err("Failed to verify FCP initiator "
                              . $init_wwpn
                              . " alias "
                              . $wwpn_alias,
                            $self->errmsg());
            return $SanUtil::FAIL;
        }
        $self->sanlog(  -msg => $log_msg
                      . "Succeed to verify fcp initiator alias name "
                      . "by command 'fcp initiator show -instance'");
    } ## end if ($fcp_init_verify eq...

    # Verify by command 'lun igroup show -instance'
    my @igroup_objs = $vs_obj->get_igroup_objs();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $igroup_obj (@igroup_objs) {

        # $igroup_obj->get_list_info(): get igroup info from igroup object
        # It is hash object. Please refer Igroup.pm
        $result = $igroup_obj->get_list_info();
        if ($igroup_obj->error()) {
            $self->_set_err($igroup_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $init_details = $result->{'init_details'};
        $self->sanlog(-msg => "init_details: " . $init_details);
        my @init_details_array = split(",", $init_details);
        for (@init_details_array) {
            my $init_info_string = (split("\"", $_))[1];

            my @init_info_array = split("\Q(", $init_info_string);

            $wwpn         = (split(" ",    $init_info_array[0]))[0];
            $alias_string = (split("\Q) ", $init_info_array[1]))[0];
            @alias_array  = ();
            @alias_array = split(" ", $alias_string);

            if ($wwpn eq $init_wwpn) {
                $result =
                  $self->_verify_fcp_init_wwpn_and_alias(
                                          $init_wwpn, $wwpn_alias, $initiator,
                                          $operation, $wwpn,       \@alias_array
                                          );
                if ($result eq $SanUtil::FAIL || $self->error()) {
                    $self->_set_err("Failed to verify FCP initiator "
                                      . $init_wwpn
                                      . " alias "
                                      . $wwpn_alias,
                                    $self->errmsg());
                    return $SanUtil::FAIL;
                }
                last;
            } ## end if ($wwpn eq $init_wwpn)
        } ## end for (@init_details_array)
    } ## end foreach my $igroup_obj (@igroup_objs)
    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify fcp initiator alias name "
                  . "by command 'lun igroup show -instance'");

    $self->sanlog(  -msg => $log_msg
                  . "Succeed to verify the FCP initiator "
                  . $init_wwpn
                  . " alias "
                  . $wwpn_alias);

    return $SanUtil::PASS;
} ## end sub verify_fcp_initiator_alias

################################################################################
#
# Function  - verify_vserver_on_client
#              Verify vservers on Clients
#
# Input     - vserver_objs_ref   required (reference to array of vserver objects)
#             client_objs_ref    required (reference to array of client objects)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_vserver_on_client {
    my ($self, %args) = @_;
    my $vserver_objs_ref;
    my $client_objs_ref;
    my @vserver_objs;
    my @client_objs;
    my $vs_obj;
    my $vs_name;
    my @vs_objs_in_clients = ();
    my %client_devices;
    my @devices = ();
    my $log_msg = "FcpHelper::verify_vserver_on_client: ";

    $self->sanlog(-msg => "FcpHelper::verify_vserver_on_client - "
                  . "Verify Vservers on Clients");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs_ref/) {
            if (not length($value)) {
                $self->_set_err(
                               "Invalid value for vserver_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $vserver_objs_ref = $value;
        } elsif ($key =~ /client_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for client_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_objs_ref = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vserver_objs_ref)) {
        $self->_set_err("Parameter 'vserver_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @vserver_objs = @$vserver_objs_ref;

    unless (defined($client_objs_ref)) {
        $self->_set_err("Parameter 'client_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @client_objs = @$client_objs_ref;

    foreach my $vs_obj (@vserver_objs) {
        $vs_name = $vs_obj->get_name();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $fcp_obj = $vs_obj->get_fcp_obj();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my @initiators = $fcp_obj->get_initiators();
        if ($fcp_obj->error()) {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my @init_array = ();
        foreach my $init (@initiators) {

            # Find the initiator which is in the igroup created in the test
            my $init_igroup = $init->{'igroup'};
            my $init_vs     = $init->{'vserver'};
            if ($init_igroup ne "-" && $init_vs == $vs_name) {
                push(@init_array, $init);
            }
        }

        # Choose vserver which shows up on result of 'vserver fcp initiator show'
        if (scalar(@init_array) != 0) {
            push(@vs_objs_in_clients, $vs_obj);
        }
    } ## end foreach my $vs_obj (@vserver_objs)

    foreach $vs_obj (@vs_objs_in_clients) {
        $vs_name = $vs_obj->get_name();
        if ($vs_obj->error()) {
            $self->_set_err($vs_obj->errmsg());
            return $SanUtil::FAIL;
        }

        foreach my $client_obj (@client_objs) {
            my $host_name = $client_obj->hostname();
            if ($client_obj->error()) {
                $self->_set_err($client_obj->errmsg());
                return $SanUtil::FAIL;
            }

            my $host_type = $client_obj->hosttype();
            if ($client_obj->error()) {
                $self->_set_err($client_obj->error());
                return $SanUtil::FAIL;
            }

            if ($host_type eq 'windows') {
                @devices =
                  $client_obj->get_devices(-refresh        => 1,
                                           -rescan_timeout => 0);
                if ($client_obj->error()) {
                    $self->_set_err($client_obj->errmsg());
                    return $SanUtil::FAIL;
                }
            } else {

                # Need to handle code for other OS in the future
            }

            # Verify vserver name got from filer under
            # 'vserver fcp initiatoir show' will also show up on clients
            my $vs_match_count = 0;
            foreach my $dev (@devices) {
                if ($dev->{filer_name} eq $vs_name) {
                    $vs_match_count++;
                }
            }
            if ($vs_match_count == 0) {
                $self->_set_err(
                            "Failed to find vserver " . $vs_name . "on client");
                return $SanUtil::FAIL;
            }

            # Create a hash %client_devices and later on pass it into
            # function call verify_lun_path_on_client() so that we don't
            # need to call $client_obj->get_devices() again since it will
            # discover the disk and is very time consumed
            @{$client_devices{$host_name}} = @devices;
        } ## end foreach my $client_obj (@client_objs)

        # Verify LUN paths on clients
        my $result =
          $self->verify_lun_path_on_client(
                                         vserver_obj        => $vs_obj,
                                         client_objs_ref    => $client_objs_ref,
                                         client_devices_ref => \%client_devices,
                                         );
        if ($result eq $SanUtil::FAIL || $self->error()) {
            return $SanUtil::FAIL;
        }
    } ## end foreach $vs_obj (@vs_objs_in_clients)

    return $SanUtil::PASS;
} ## end sub verify_vserver_on_client

################################################################################
#
# Function  - verify_lun_path_on_client
#             Verify lun paths on clients
#             Make sure lun path show up under 'lun mapped show' on filer
#             will also show up on client 'mpc show luns'
#
# Input     - vserver_obj         required
#             client_devices_ref  required
#             (reference to a hash, key: host_name; values, array of devices)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_lun_path_on_client {
    my ($self, %args) = @_;
    my $vs_obj;
    my $client_objs_ref;
    my @client_objs;
    my $client_devices_ref;
    my %client_devices;
    my $vs_name;
    my $lun_path;
    my @lun_paths = ();
    my $lun_path_on_client;
    my @lun_paths_on_client = ();
    my $result;
    my $log_msg = "FcpHelper::verify_lun_path_on_client: ";

    $self->sanlog(-msg => "FcpHelper::verify_lun_path_on_client - "
                  . "Verify LUN Paths on Clients");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for vserver_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $vs_obj = $value;
        } elsif ($key =~ /client_objs_ref/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for client_objs_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_objs_ref = $value;
        } elsif ($key =~ /client_devices_ref/) {
            if (not length($value)) {
                $self->_set_err(
                             "Invalid value for client_devices_ref: " . $value);
                return $SanUtil::FAIL;
            }
            $client_devices_ref = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($vs_obj)) {
        $self->_set_err("Parameter 'vserver_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($client_objs_ref)) {
        $self->_set_err("Parameter 'client_objs_ref' not specified");
        return $SanUtil::FAIL;
    }
    @client_objs = @$client_objs_ref;

    unless (defined($client_devices_ref)) {
        $self->_set_err("Parameter 'client_devices_ref' not specified");
        return $SanUtil::FAIL;
    }
    %client_devices = %$client_devices_ref;

    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my @volume_objs = $vs_obj->get_volume_objs();
    if ($vs_obj->error()) {
        $self->_set_err($vs_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $volume_obj (@volume_objs) {
        my @lun_objs = $volume_obj->get_luns();
        if ($volume_obj->error()) {
            $self->_set_err($volume_obj->errmsg());
            return $SanUtil::FAIL;
        }

        foreach my $lun_obj (@lun_objs) {
            $lun_path = $lun_obj->get_path();
            if ($lun_obj->error()) {
                $self->_set_err($lun_obj->errmsg());
                return $SanUtil::FAIL;
            }

            # Get initiators (wwpns) that mapped to this lun
            # result from 'lun mapped show'
            my $initiators = $lun_obj->get_initiators();
            if ($lun_obj->error()) {
                $self->_set_err($lun_obj->errmsg());
                return $SanUtil::FAIL;
            }
            my @inits = split(",", $initiators);

            # Choose initiators which shows up on result of
            # 'vserver fcp initiator show'
            $result = $self->_get_login_initiators($vs_obj, \@inits);
            if ($result eq $SanUtil::FAIL) {
                return $SanUtil::FAIL;
            }
            my @login_inits = @$result;

            my $count = 0;
            foreach my $init (@login_inits) {
                foreach my $client_obj (@client_objs) {
                    my $host_name = $client_obj->hostname();
                    if ($client_obj->error()) {
                        $self->_set_err($client_obj->errmsg());
                        return $SanUtil::FAIL;
                    }
                    my @devices = @{$client_devices{$host_name}};

                    # Get all initiator wwpns from clients
                    # including ones link up/down
                    my @initiators = $client_obj->get_initiators();
                    if ($client_obj->error()) {
                        $self->_set_err($client_obj->errmsg());
                        return $SanUtil::FAIL;
                    }
                    my @initiator_wwpn_array = ();
                    foreach my $initiator (@initiators) {
                        my $portname = $initiator->get_portname();
                        if ($initiator->error()) {
                            $self->_set_err($initiator->errmsg());
                            return $SanUtil::FAIL;
                        }
                        push(@initiator_wwpn_array, $portname);
                    }

                    my @match_value_array_1 =
                      grep { $_ eq $init } @initiator_wwpn_array;

                    # Choose client initiators which are logged in
                    if (scalar(@match_value_array_1) != 0) {
                        $count++;
                        foreach my $dev (@devices) {
                            my $disk_obj = $dev->{'diskObj'};
                            $lun_path_on_client = $disk_obj->{'lunPath'};

                            # Make sure only unique LUN paths(no duplication)
                            # in array @lun_paths_on_client
                            my @match_value_array_2 = ();
                            @match_value_array_2 =
                              grep { $_ eq $lun_path_on_client }
                              @lun_paths_on_client;
                            if (scalar(@match_value_array_2) == 0) {
                                push(@lun_paths_on_client, $lun_path_on_client);
                            }
                        } ## end foreach my $dev (@devices)

                        # Compare LUN path information from filer
                        # with LUN path information from client
                        my @match_value_array_3 = ();
                        @match_value_array_3 =
                          grep { $_ eq $lun_path } @lun_paths_on_client;
                        if (scalar(@match_value_array_3) == 0) {
                            $self->_set_err(  "Failed to find lun path "
                                            . $lun_path
                                            . " on client");
                            return $SanUtil::FAIL;
                        }
                    } ## end if (scalar(@match_value_array_1...
                } ## end foreach my $client_obj (@client_objs)
            } ## end foreach my $init (@login_inits)

            if ($count == 0) {

                # All the client initiators are not mapped to the LUN
                $self->_set_err(
                         "Failed to find lun path " . $lun_path . " on client");
                return $SanUtil::FAIL;
            }
        } ## end foreach my $lun_obj (@lun_objs)
    } ## end foreach my $volume_obj (@volume_objs)

    $self->sanlog(-msg => "Succeed to verify LUN paths on clients");

    return $SanUtil::PASS;
} ## end sub verify_lun_path_on_client

################################################################################
#
# Function  - verify_fcp_adapter_state
#             verify FCP adapter state
#
# Input     - filer_obj      required
#             adapter        required  (adapter port reside on filer)
#             state          required  (expected state of FCP adapter)
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub verify_fcp_adapter_state {
    my ($self, %args) = @_;
    my $filer_obj;
    my $adapter;
    my $state;
    my $cmd_interface = 'cli';
    my $fcpadapter_status;
    my $result;
    my $log_msg = "FcpHelper::verify_fcp_adapter_state: ";

    $self->sanlog(-msg => "FcpHelper::verify_fcp_adapter_state - "
                  . "Verify FCP Adapter State");

    $self->_clear_err();

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /filer_obj/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for filer_obj: " . $value);
                return $SanUtil::FAIL;
            }
            $filer_obj = $value;
        }
        if ($key =~ /adapter/) {
            if (not length($value)) {
                $self->_set_err("Invalid value for adapter: " . $value);
                return $SanUtil::FAIL;
            }
            $adapter = $value;
        } elsif ($key =~ /state/) {
            if (not length($value) and $value !~ /^up$|^down$/i) {
                $self->_set_err("Invalid value for state: " . $value);
                return $SanUtil::FAIL;
            }
            $state = $value;
        } elsif ($key =~ /cmd_interface/) {
            $cmd_interface = $value;
        }
    } ## end while (my ($key, $value) ...

    # check for required arguments
    unless (defined($filer_obj)) {
        $self->_set_err("Parameter 'filer_obj' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($adapter)) {
        $self->_set_err("Parameter 'adapter' not specified");
        return $SanUtil::FAIL;
    }

    unless (defined($state)) {
        $self->_set_err("Parameter 'state' not specified");
        return $SanUtil::FAIL;
    }

    if ($state eq 'up') {
        $fcpadapter_status = 'online';
    } elsif ($state eq 'down') {
        $fcpadapter_status = 'offlined by user/system';
    }

    my $filer_name = $filer_obj->get_name();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my $hw_obj = $filer_obj->get_hwresource_obj();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my @fcpadapter_objs = $hw_obj->get_fcpadapter_objs();
    if ($hw_obj->error()) {
        $self->_set_err($hw_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $fcpadapter_obj (@fcpadapter_objs) {
        my $fcpadapter_node = $fcpadapter_obj->get_node();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }
        my $fcpadapter_adapter = $fcpadapter_obj->get_adapter();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if (   ($fcpadapter_node eq $filer_name)
            && ($fcpadapter_adapter eq $adapter)) {
            if ($cmd_interface eq 'zapi') {
                my $adapter_curr_status = $fcpadapter_obj->get_status();
                if ($fcpadapter_obj->error()) {
                    $self->_set_err($fcpadapter_obj->errmsg());
                    return $SanUtil::FAIL;
                }

                if ($adapter_curr_status ne $fcpadapter_status) {
                    $self->_set_err(  "Failed to verify FCP adapter " 
                                    . $adapter
                                    . " on node "
                                    . $filer_name
                                    . " state");
                    return $SanUtil::FAIL;
                }
            } else {
                my $adapter_curr_state = $fcpadapter_obj->get_state();
                if ($fcpadapter_obj->error()) {
                    $self->_set_err($fcpadapter_obj->errmsg());
                    return $SanUtil::FAIL;
                }

                if ($adapter_curr_state ne $state) {
                    $self->_set_err(  "Failed to verify FCP adapter " 
                                    . $adapter
                                    . " on node "
                                    . $filer_name
                                    . " state");
                    return $SanUtil::FAIL;
                } else {
                    if ((   $adapter_curr_state eq 'up'
                         && $fcpadapter_status ne 'online')
                        || (   $adapter_curr_state ne 'up'
                            && $fcpadapter_status eq 'online')
                      ) {
                        $self->_set_err(  "Failed to verify FCP adapter " 
                                        . $adapter
                                        . " on node "
                                        . $filer_name
                                        . " state");
                        return $SanUtil::FAIL;
                    } ## end if (($adapter_curr_state...
                } ## end else [ if ($adapter_curr_state...
            } ## end else [ if ($cmd_interface eq ...
        } ## end if (($fcpadapter_node ...
    } ## end foreach my $fcpadapter_obj ...

    $self->sanlog(  -msg => "Succed to verify FCP adapter " 
                  . $adapter
                  . " on node "
                  . $filer_name
                  . " state");

    return $SanUtil::PASS;
} ## end sub verify_fcp_adapter_state

################################################################################
#                            Private Methods                                   #
################################################################################

################################################################################
#
# Function  - _verify_lif_port
#             Verify LIF properties by command 'vserver fcp port show'
#             properties compared:
#             name/vserver/status_admin/status_oper/curr_node/curr_port/wwpn
#             'create' operation: compare LIF properties
#             'delete' operation: check if there is entry for given LIF
#
# Input     - fcp_obj               required
#             lif_name              required
#             lif_properties_ref    required
#             operation             required
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub _verify_lif_port {
    my ($self, $fcp_obj, $lif_name, $lif_properties_ref, $operation,
        $cmd_interface)
      = @_;
    my %lif_properties = %$lif_properties_ref;
    my $result;

    $self->_clear_err();

    my @records = $fcp_obj->get_ports(lif => $lif_name);
    if (scalar(@records) == 0 || $fcp_obj->error()) {
        if ($operation eq "create") {
            $self->_set_err($fcp_obj->errmsg());
            return $SanUtil::FAIL;
        } elsif ($operation eq "delete") {
            return $SanUtil::PASS;
        }
    } else {
        if ($operation eq "create") {
            my %properties = %{$lif_properties{$lif_name}};
            my $record     = $records[0];

            for (keys(%properties)) {

                # Keys are fields:
                # lif/vserver/status_admin/status_oper/curr_node/curr_port/wwpn
                # Since fcp-interface-get-iter does not provide status_admin/
                # status_oper/wwpn information for FCP LIF, skip the
                # compare if command interface is zapi
                if ($cmd_interface eq 'zapi') {
                    if (   ($_ eq 'status_admin')
                        || ($_ eq 'status_oper')
                        || ($_ eq 'wwpn')) {
                        next;
                    }
                }

                my $value = $properties{$_};
                $self->sanlog(-msg => "LIF field: $_" . " value: $value\n");
                $self->sanlog(-msg => "LIF value in record: " . $record->{$_} . "\n");
                if ($record->{$_} ne $value) {
                    return $SanUtil::FAIL;
                }
            } ## end for (keys(%properties))
        } elsif ($operation eq "delete") {
            return $SanUtil::FAIL;
        }
    } ## end else [ if (scalar(@records) ==...

    return $SanUtil::PASS;
} ## end sub _verify_lif_port

################################################################################
#
# Function  - _verify_lif_portname
#             Verify LIF portname assigned and same as the address
#             properties compared: wwpn
#             'create' operation: compare LIF properties
#             'delete' operation: check if there is entry for given LIF
#
# Input     - fcp_obj               required
#             lif_name              required
#             lif_properties_ref    required
#             operation             required
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub _verify_lif_portname {
    my ($self, $fcp_obj, $lif_name, $lif_properties_ref, $operation,
        $cmd_interface)
      = @_;
    my %lif_properties = %$lif_properties_ref;

    $self->_clear_err();

    my @wwpn = $fcp_obj->get_wwpn(lif => $lif_name);
    if ($fcp_obj->error()) {
        $self->_set_err($fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }

    if (scalar(@wwpn) == 0) {
        if ($operation eq "create") {
            return $SanUtil::FAIL;
        } elsif ($operation eq "delete") {
            return $SanUtil::PASS;
        }
    } else {
        if ($operation eq "create") {

            # Since net-interface-get-iter does not provide wwpn information
            # for FCP LIF, skip the compare if command interface is zapi
            if ($cmd_interface eq 'zapi') {
                return $SanUtil::PASS;
            }

            my $wwpn       = $wwpn[0];
            my %properties = %{$lif_properties{$lif_name}};
            my $lif_wwpn   = $properties{'wwpn'};
            if ($wwpn ne $lif_wwpn) {
                return $SanUtil::FAIL;
            }
        } elsif ($operation eq "delete") {

            # Since net-interface-get-iter does not provide wwpn information
            # for FCP LIF, skip the compare if command interface is zapi
            if ($cmd_interface eq 'zapi') {
                return $SanUtil::PASS;
            }

            return $SanUtil::FAIL;
        } ## end elsif ($operation eq "delete")
    } ## end else [ if (scalar(@wwpn) == 0)

    return $SanUtil::PASS;
} ## end sub _verify_lif_portname

################################################################################
#
# Function  - _verify_fcp_init_wwpn_and_alias
#             Verify FCP initiator wwpn and alias
#
# Input     - init_wwpn             required
#             wwpn_alias            required
#             initiator             required
#             operation             required
#             wwpn                  required (init wwpn from command output)
#             alias_array_ref       required
# Output    - SanUtil::PASS
#             SanUtil::FAIL
#
################################################################################
sub _verify_fcp_init_wwpn_and_alias {
    my ($self, $init_wwpn, $wwpn_alias, $initiator, $operation, $wwpn,
        $alias_array_ref)
      = @_;
    my @alias_array = ();
    my @alias       = ();
    my $result;

    $self->_clear_err();

    # Verify initiator wwpn
    if ($init_wwpn ne $initiator->{'wwpn'}) {
        return $SanUtil::FAIL;
    }

    # Verify alias names
    # Construct two arrays so that code can call _array_compare()
    @alias_array = @$alias_array_ref;
    push(@alias, $wwpn_alias);

    if (($#alias_array + 1) != 0) {
        $result = $self->_array_compare(\@alias, \@alias_array);
        if ($result eq $SanUtil::FAIL || $self->error()) {
            if ($operation eq "set") {
                return $SanUtil::FAIL;
            }
        } else {
            if ($operation eq "remove") {
                return $SanUtil::FAIL;
            }
        }
    } else {
        if ($operation eq "set") {
            return $SanUtil::FAIL;
        }
    }

    return $SanUtil::PASS;
} ## end sub _verify_fcp_init_wwpn_and_alias

################################################################################
#
# Function  -  _array_compare
#              compare two arrays.
#              The logic here is the second array will have more items than
#              first array, only want to make sure the items in array1 will be
#              included in array2, NOT vise verse.
# Input     -  array_ref1   required (reference to array 1)
#              array_ref2   required (reference to array 2)
# Output    -  SanUtil::PASS if all the items in array 1 can be found in array 2
#              SanUtil::FAIL
################################################################################
sub _array_compare {
    my ($self, $array_ref1, $array_ref2) = @_;
    my @array1 = @$array_ref1;
    my @array2 = @$array_ref2;

    $self->_clear_err();

    foreach my $item1 (@array1) {
        my $count = 0;
        foreach my $item2 (@array2) {
            if ($item1 ne $item2) {
                $count++;
            }
        }
        if ($count == scalar(@array2)) {
            return $SanUtil::FAIL;
        }
    } ## end foreach my $item1 (@array1)
    return $SanUtil::PASS;
} ## end sub _array_compare

################################################################################
#
# Function  - _get_lif_wwpn
#             get LIF wwpn by executing command 'vserver fcp portname show'
#
# Input     - vserver_objs_ref  required
#             vs_name           required
#             lif_name          required
# Output    - LIF wwpn
#             SanUtil::FAIL
#
################################################################################
sub _get_lif_wwpn {
    my ($self, $vserver_objs_ref, $vs_name, $lif_name) = @_;

    my @vserver_objs = @$vserver_objs_ref;
    my @lif_wwpn_array;

    $self->_clear_err();

    foreach my $vserver_obj (@vserver_objs) {
        my $fcp_obj = $vserver_obj->get_fcp_obj();
        if ($vserver_obj->error()) {
            $self->_set_err($vserver_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($vs_name eq $vserver_obj->get_name()) {
            if ($vserver_obj->error()) {
                $self->_set_err($vserver_obj->errmsg());
                return $SanUtil::FAIL;
            }
            @lif_wwpn_array = $fcp_obj->get_wwpn(-lif => $lif_name);
            if ($fcp_obj->error()) {
                $self->_set_err($fcp_obj->errmsg());
                return $SanUtil::FAIL;
            }
            last;
        } ## end if ($vs_name eq $vserver_obj...
    } ## end foreach my $vserver_obj (@vserver_objs)

    return $lif_wwpn_array[0];
} ## end sub _get_lif_wwpn

################################################################################
#
# Function  - _get_lif_properties
#             get LIF properties (by command 'network interface show')
#             LIF properties include:
#             name/vserver/status_admin/status_oper/curr_node/curr_port/wwpn
#
# Input     - filer_obj required
#             lif_obj   required
# Output    - A hash which holds LIF properties
#             (key: lif_name; value: a hash of LIF properties)
#             SanUtil::FAIL
#
################################################################################
sub _get_lif_properties {
    my ($self, $lif_obj, $cmd_interface) = @_;
    my %lif_properties;
    my %properties;

    $self->_clear_err();

    my $lif_name = $lif_obj->get_name();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'lif'} = $lif_name;

    my $lif_vs = $lif_obj->get_vserver();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'vserver'} = $lif_vs;

    my $lif_status_admin = $lif_obj->get_admin_status();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'status_admin'} = $lif_status_admin;

    my $lif_status_oper = $lif_obj->get_status_oper();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'status_oper'} = $lif_status_oper;

    my $lif_curr_node = $lif_obj->get_curr_node();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'curr_node'} = $lif_curr_node;

    my $lif_curr_port = $lif_obj->get_curr_port();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'curr_port'} = $lif_curr_port;

    my $lif_wwpn = $lif_obj->get_wwpn();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }
    $properties{'wwpn'} = $lif_wwpn;

    %{$lif_properties{$lif_name}} = %properties;

    return \%lif_properties;
} ## end sub _get_lif_properties

################################################################################
#
# Function  - _get_lif_filer_obj
#             get filer object of LIF
#
# Input     - filer_objs_ref required
#             lif_obj        required
# Output    - A filer object
#             SanUtil::FAIL
#
################################################################################
sub _get_lif_filer_obj {
    my ($self, $filer_objs_ref, $lif_obj) = @_;
    my $lif_curr_node;

    $self->_clear_err();

    $lif_curr_node = $lif_obj->get_curr_node();
    if ($lif_obj->error()) {
        $self->_set_err($lif_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $filer_obj (@$filer_objs_ref) {
        my $filer_name = $filer_obj->get_name();
        if ($filer_obj->error()) {
            $self->_set_err($filer_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($filer_name eq $lif_curr_node) {
            return $filer_obj;
        }
    } ## end foreach my $filer_obj (@$filer_objs_ref)

    # No filer object is found
    return $SanUtil::FAIL;
} ## end sub _get_lif_filer_obj

################################################################################
#
# Function  - _get_fcpadapter_objs
#             get fcpadapter objects on given node and port
#
# Input     - filer_obj required
# Output    - A array of fcpadapter objects
#             SanUtil::FAIL
#
################################################################################
sub _get_fcpadapter_objs {
    my ($self, $filer_obj, $adapter) = @_;
    my $filer_name;
    my @fcpadapter_objs = ();
    my $hw_obj;
    my @fcpadapter_objs_per_filer = ();

    $self->_clear_err();

    $filer_name = $filer_obj->get_name();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    $hw_obj = $filer_obj->get_hwresource_obj();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    @fcpadapter_objs = $hw_obj->get_fcpadapter_objs();
    if ($hw_obj->error()) {
        $self->_set_err($hw_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $fcpadapter_obj (@fcpadapter_objs) {
        my $fcpadapter_node = $fcpadapter_obj->get_node();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        my $fcpadapter_adapter = $fcpadapter_obj->get_adapter();
        if ($fcpadapter_obj->error()) {
            $self->_set_err($fcpadapter_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if (   ($fcpadapter_node eq $filer_name)
            && ($fcpadapter_adapter eq $adapter)) {
            push(@fcpadapter_objs_per_filer, $fcpadapter_obj);
        }
    } ## end foreach my $fcpadapter_obj ...

    return \@fcpadapter_objs_per_filer;
} ## end sub _get_fcpadapter_objs

################################################################################
#
# Function  - _get_lifs_oper_status
#             check if there are LIFs whose operation status are up
#             on given node and port
#
# Input     - lif_objs_ref required
# Output    - 'down'  if all LIFs status are 'down'
#             'up'    if a LIF with 'up' is found
#
################################################################################
sub _get_lifs_oper_status {
    my ($self, $filer_obj, $adapter, $lif_objs_ref) = @_;
    my $filer_name;
    my @lifs_per_node_port;

    $self->_clear_err();

    $filer_name = $filer_obj->get_name();
    if ($filer_obj->error()) {
        $self->_set_err($filer_obj->errmsg());
        return $SanUtil::FAIL;
    }

    foreach my $lif_obj (@$lif_objs_ref) {
        my $lif_node = $lif_obj->get_curr_node();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }
        my $lif_port = $lif_obj->get_curr_port();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        # Find out all the LIFs which resides on given node and port
        if (($lif_node eq $filer_name) && ($lif_port eq $adapter)) {
            push(@lifs_per_node_port, $lif_obj);
        }
    } ## end foreach my $lif_obj (@$lif_objs_ref)

    foreach my $lif_obj (@lifs_per_node_port) {
        my $lif_status_oper = $lif_obj->get_status_oper();
        if ($lif_obj->error()) {
            $self->_set_err($lif_obj->errmsg());
            return $SanUtil::FAIL;
        }

        if ($lif_status_oper eq "up") {
            return "up";
        }
    } ## end foreach my $lif_obj (@lifs_per_node_port)

    return "down";
} ## end sub _get_lifs_oper_status

################################################################################
#
# Function  - _get_login_initiators
#             Get initiators for given vserver which show up on the result of
#             'vserver fcp initiator show'
#
# Input     - vserver_obj       required
#             initiators_ref    required
#             (reference to array of initiators from command output
#             'lun mapped show' $lun_obj->get_initiators())
# Output    - \@login_inits
#             SanUtil::FAIL
#
################################################################################
sub _get_login_initiators {
    my ($self, $vserver_obj, $initiators_ref) = @_;
    my @initiators  = @$initiators_ref;
    my @login_inits = ();

    $self->_clear_err();

    my $fcp_obj = $vserver_obj->get_fcp_obj();
    if ($vserver_obj->error()) {
        $self->_set_err("Error getting fcp object " . $vserver_obj->errmsg());
        return $SanUtil::FAIL;
    }

    my @fcp_initiators = $fcp_obj->get_initiators();
    if ($fcp_obj->error()) {
        $self->_set_err(
            "Error getting fcp initiator from the filer " . $fcp_obj->errmsg());
        return $SanUtil::FAIL;
    }

    # Create a hash %temp_hash to remove repeated wwpns
    my %temp_hash;
    foreach my $fcp_initiator (@fcp_initiators) {
        my $fcp_initiator_wwpn = $fcp_initiator->{'wwpn'};
        $temp_hash{$fcp_initiator_wwpn} = 1;
    }
    my @init_wwpn_array = keys(%temp_hash);

    foreach my $init (@initiators) {
        foreach (@init_wwpn_array) {
            if ($init eq $_) {
                push(@login_inits, $init);
            }
        }
    }

    return \@login_inits;
} ## end sub _get_login_initiators

1;
